
##############################################################################
#
# Copyright (c) 2003-2016 by The University of Queensland
# http://www.uq.edu.au
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# Development until 2012 by Earth Systems Science Computational Center (ESSCC)
# Development 2012-2013 by School of Earth Sciences
# Development from 2014 by Centre for Geoscience Computing (GeoComp)
#
##############################################################################

from __future__ import print_function, division

__copyright__="""Copyright (c) 2003-2016 by The University of Queensland
http://www.uq.edu.au
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://launchpad.net/escript-finley"

"""
basic tests for functions in util.py effecting the spatial distribution 

it is assumed that the domain is the usint square/cube 

not all these test will run for all domains. check the doc string for the assumptions of a particular test

:var __author__: name of author
:var __copyright__: copyrights
:var __license__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""
__author__="Lutz Gross, l.gross@uq.edu.au"

import esys.escriptcore.utestselect as unittest
from esys.escript import *
from numpy import array
import numpy

class Test_Util_SpatialFunctions_noGradOnBoundary_noContact(unittest.TestCase):
   RES_TOL=1.e-8
   def test_x_ofDomain(self):
     """
     test getX() of the domain to be in the [0,1]^dim box
     """
     dim=self.domain.getDim()
     x=self.domain.getX()
     self.assertEqual(x.getShape(),(dim,),"wrong shape of result.")
     self.assertAlmostEqual(inf(x[0]),0.,int(-log10(self.RES_TOL)),"min x0 wrong")
     self.assertAlmostEqual(sup(x[0]),1.,int(-log10(self.RES_TOL)),"max x0 wrong")
     self.assertAlmostEqual(inf(x[1]),0.,int(-log10(self.RES_TOL)),"min x1 wrong")
     self.assertAlmostEqual(sup(x[1]),1.,int(-log10(self.RES_TOL)),"max x1 wrong")
     if dim>2:
       self.assertAlmostEqual(inf(x[2]),0.,int(-log10(self.RES_TOL)),"min x2 wrong")
       self.assertAlmostEqual(sup(x[2]),1.,int(-log10(self.RES_TOL)),"max x2 wrong")

   def test_SolutionOrder(self):
      """
      test the approximation order 
      """
      self.assertEqual(self.order, Solution(self.domain).getApproximationOrder(), "wrong order (Solution)")
      self.assertEqual(self.order, ContinuousFunction(self.domain).getApproximationOrder(), "wrong order (continuous function)")
      self.assertEqual(1, ReducedSolution(self.domain).getApproximationOrder(), "wrong order (ReducedSolution)")
      self.assertEqual(1, ReducedContinuousFunction(self.domain).getApproximationOrder(), "wrong order (Reduced continuous function)")
      for i in range(self.domain.getDim()):
         for k in range(Function(self.domain).getApproximationOrder()+1):
             self.assertAlmostEqual(integrate(Function(self.domain).getX()[i]**k),1./(k+1),8,"wrong integral (i=%s, order = %s)"%(i,k))
         for k in range(ReducedFunction(self.domain).getApproximationOrder()+1):
             self.assertAlmostEqual(integrate(ReducedFunction(self.domain).getX()[i]**k),1./(k+1),8,"wrong integral (i=%s, order = %s (reduced))"%(i,k))


   def test_normal_FunctionOnBoundary(self):
     """
     test getNormal() on boundary

     assumptions: FunctionOnBoundary(self.domain) exists
     """
     dim=self.domain.getDim()
     f=FunctionOnBoundary(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref.setTaggedValue(200,[0,0,1])
         ref.setTaggedValue(100,[0,0,-1])
         ref.setTaggedValue(20,[0,1,0])
         ref.setTaggedValue(10,[0,-1,0])
         ref.setTaggedValue(2,[1,0,0])
         ref.setTaggedValue(1,[-1,0,0])
     else:
         ref.setTaggedValue(2,[1,0])
         ref.setTaggedValue(1,[-1,0])
         ref.setTaggedValue(20, [0,1])
         ref.setTaggedValue(10, [0,-1])

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertLess(Lsup(ref-res), self.RES_TOL, "wrong result")

   def test_normal_ReducedFunctionOnBoundary(self):
     """
     test getNormal() on boundary

     assumptions: FunctionOnBoundary(self.domain) exists
     """
     dim=self.domain.getDim()
     f=ReducedFunctionOnBoundary(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref.setTaggedValue(200,[0,0,1])
         ref.setTaggedValue(100,[0,0,-1])
         ref.setTaggedValue(20,[0,1,0])
         ref.setTaggedValue(10,[0,-1,0])
         ref.setTaggedValue(2,[1,0,0])
         ref.setTaggedValue(1,[-1,0,0])
     else:
         ref.setTaggedValue(2,[1,0])
         ref.setTaggedValue(1,[-1,0])
         ref.setTaggedValue(20, [0,1])
         ref.setTaggedValue(10, [0,-1])

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertLess(Lsup(ref-res), self.RES_TOL, "wrong result")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.465976690369)*x[0]**o+(-0.25136070196)*x[0]+(0.409576841566)*x[1]**o+(0.726707772754)*x[1]
        ref=(0.875553531935)/(o+1.)+(0.237673535397)
      else:
        arg=(0.567785895928)*x[0]**o+(0.928477128396)*x[0]+(-0.625720943088)*x[1]**o+(-0.681460841298)*x[1]+(-0.866801010069)*x[2]**o+(0.159594300637)*x[2]
        ref=(-0.924736057229)/(o+1.)+(0.203305293867)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref), "wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.839020088668)*x[0]**o+(-0.880288035177)*x[0]+(0.260951734364)*x[1]**o+(-0.624208000982)*x[1]
        arg[1]=(0.985421831314)*x[0]**o+(0.660399229451)*x[0]+(-0.514003848134)*x[1]**o+(0.0312162082593)*x[1]
        ref[0]=(-0.578068354304)/(o+1.)+(-0.75224801808)
        ref[1]=(0.47141798318)/(o+1.)+(0.345807718855)
      else:
        arg[0]=(-0.627307316929)*x[0]**o+(0.500432411035)*x[0]+(0.233895908313)*x[1]**o+(-0.125048135432)*x[1]+(0.0695114009632)*x[2]**o+(0.995170977074)*x[2]
        arg[1]=(0.760033715785)*x[0]**o+(0.48895905139)*x[0]+(0.240372023176)*x[1]**o+(0.33946689896)*x[1]+(-0.774638489419)*x[2]**o+(-0.0777240871074)*x[2]
        ref[0]=(-0.323900007653)/(o+1.)+(0.685277626338)
        ref[1]=(0.225767249542)/(o+1.)+(0.375350931622)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref), "wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.844492048883)*x[0]**o+(0.727132679659)*x[0]+(-0.862576284722)*x[1]**o+(0.502937561417)*x[1]
        arg[0,1]=(-0.587157419309)*x[0]**o+(0.00327068097107)*x[0]+(0.422592349683)*x[1]**o+(0.167479615369)*x[1]
        arg[0,2]=(0.771156665672)*x[0]**o+(-0.371091138833)*x[0]+(-0.867750378658)*x[1]**o+(-0.744394290185)*x[1]
        arg[0,3]=(0.669718598855)*x[0]**o+(-0.0825909409121)*x[0]+(0.868872948791)*x[1]**o+(0.141891982204)*x[1]
        arg[0,4]=(-0.441072764653)*x[0]**o+(-0.929654649265)*x[0]+(-0.35883498816)*x[1]**o+(-0.84045480381)*x[1]
        arg[1,0]=(0.555668196622)*x[0]**o+(0.94306681043)*x[0]+(0.586512446245)*x[1]**o+(0.702007276327)*x[1]
        arg[1,1]=(0.792083771005)*x[0]**o+(-0.865830139617)*x[0]+(0.0929971631615)*x[1]**o+(0.620446289121)*x[1]
        arg[1,2]=(-0.201978375472)*x[0]**o+(-0.0777206885918)*x[0]+(0.798755459687)*x[1]**o+(-0.278086468846)*x[1]
        arg[1,3]=(0.063028419572)*x[0]**o+(-0.454786232294)*x[0]+(-0.0481393324293)*x[1]**o+(-0.612657132834)*x[1]
        arg[1,4]=(-0.743283787964)*x[0]**o+(0.854827303846)*x[0]+(0.94241635825)*x[1]**o+(0.436937526521)*x[1]
        arg[2,0]=(-0.814975695669)*x[0]**o+(0.579477783769)*x[0]+(0.00998264005965)*x[1]**o+(-0.913665878657)*x[1]
        arg[2,1]=(-0.526833976369)*x[0]**o+(-0.800386086232)*x[0]+(-0.487600053086)*x[1]**o+(0.00242562519003)*x[1]
        arg[2,2]=(-0.964119591874)*x[0]**o+(0.203212394214)*x[0]+(-0.96850837954)*x[1]**o+(-0.214886132512)*x[1]
        arg[2,3]=(-0.356249104564)*x[0]**o+(0.329769293221)*x[0]+(0.936664928754)*x[1]**o+(0.520603503283)*x[1]
        arg[2,4]=(0.255464219424)*x[0]**o+(-0.157783294516)*x[0]+(0.519451787865)*x[1]**o+(0.00960521397376)*x[1]
        arg[3,0]=(-0.117250486034)*x[0]**o+(0.192945431273)*x[0]+(-0.7580154297)*x[1]**o+(-0.308801650294)*x[1]
        arg[3,1]=(0.243931002845)*x[0]**o+(-0.626714191745)*x[0]+(-0.365870571265)*x[1]**o+(0.652257582021)*x[1]
        arg[3,2]=(-0.769756656122)*x[0]**o+(0.711106390017)*x[0]+(0.150749630653)*x[1]**o+(-0.0441309732448)*x[1]
        arg[3,3]=(0.92677837493)*x[0]**o+(0.507199859313)*x[0]+(-0.542275490828)*x[1]**o+(-0.0907270087228)*x[1]
        arg[3,4]=(-0.0306175885885)*x[0]**o+(-0.434666505018)*x[0]+(-0.815633427562)*x[1]**o+(-0.916424797903)*x[1]
        ref[0,0]=(-1.70706833361)/(o+1.)+(0.615035120538)
        ref[0,1]=(-0.164565069626)/(o+1.)+(0.08537514817)
        ref[0,2]=(-0.0965937129851)/(o+1.)+(-0.557742714509)
        ref[0,3]=(1.53859154765)/(o+1.)+(0.0296505206459)
        ref[0,4]=(-0.799907752814)/(o+1.)+(-0.885054726537)
        ref[1,0]=(1.14218064287)/(o+1.)+(0.822537043379)
        ref[1,1]=(0.885080934167)/(o+1.)+(-0.122691925248)
        ref[1,2]=(0.596777084215)/(o+1.)+(-0.177903578719)
        ref[1,3]=(0.0148890871427)/(o+1.)+(-0.533721682564)
        ref[1,4]=(0.199132570285)/(o+1.)+(0.645882415184)
        ref[2,0]=(-0.80499305561)/(o+1.)+(-0.167094047444)
        ref[2,1]=(-1.01443402945)/(o+1.)+(-0.398980230521)
        ref[2,2]=(-1.93262797141)/(o+1.)+(-0.00583686914904)
        ref[2,3]=(0.580415824189)/(o+1.)+(0.425186398252)
        ref[2,4]=(0.774916007289)/(o+1.)+(-0.074089040271)
        ref[3,0]=(-0.875265915734)/(o+1.)+(-0.0579281095106)
        ref[3,1]=(-0.12193956842)/(o+1.)+(0.0127716951382)
        ref[3,2]=(-0.619007025469)/(o+1.)+(0.333487708386)
        ref[3,3]=(0.384502884102)/(o+1.)+(0.208236425295)
        ref[3,4]=(-0.846251016151)/(o+1.)+(-0.675545651461)
      else:
        arg[0,0]=(0.596538443259)*x[0]**o+(0.479462980373)*x[0]+(-0.334469126403)*x[1]**o+(0.784201085556)*x[1]+(-0.231474140531)*x[2]**o+(-0.368096345589)*x[2]
        arg[0,1]=(0.188945383338)*x[0]**o+(-0.95033024122)*x[0]+(0.567904901047)*x[1]**o+(-0.76703296351)*x[1]+(-0.0116083690376)*x[2]**o+(-0.211176589764)*x[2]
        arg[0,2]=(0.225412552449)*x[0]**o+(-0.731615155924)*x[0]+(-0.362727501637)*x[1]**o+(-0.195442119344)*x[1]+(-0.215444456285)*x[2]**o+(0.412715796392)*x[2]
        arg[0,3]=(-0.00941287481408)*x[0]**o+(-0.207271678463)*x[0]+(0.66582433583)*x[1]**o+(-0.297871484308)*x[1]+(-0.819460465536)*x[2]**o+(-0.3316307033)*x[2]
        arg[0,4]=(-0.172265683581)*x[0]**o+(0.881658164973)*x[0]+(0.0267816673708)*x[1]**o+(-0.526757427521)*x[1]+(-0.666953449252)*x[2]**o+(0.366533690002)*x[2]
        arg[1,0]=(-0.575272180021)*x[0]**o+(-0.707236339163)*x[0]+(-0.734723292219)*x[1]**o+(-0.872019568084)*x[1]+(-0.513269190461)*x[2]**o+(-0.95077853535)*x[2]
        arg[1,1]=(0.687652562435)*x[0]**o+(0.500668143461)*x[0]+(0.127678854367)*x[1]**o+(0.135196530391)*x[1]+(0.219757431874)*x[2]**o+(0.43373842168)*x[2]
        arg[1,2]=(-0.674120744855)*x[0]**o+(-0.792769889708)*x[0]+(0.0414602844391)*x[1]**o+(-0.159887164563)*x[1]+(0.406441555189)*x[2]**o+(-0.662210077269)*x[2]
        arg[1,3]=(-0.623120612358)*x[0]**o+(0.895951757448)*x[0]+(0.957407224559)*x[1]**o+(-0.472225118848)*x[1]+(-0.294551998847)*x[2]**o+(0.503584320886)*x[2]
        arg[1,4]=(-0.304220645545)*x[0]**o+(0.812321371589)*x[0]+(0.932257196386)*x[1]**o+(0.736014913915)*x[1]+(0.119140391025)*x[2]**o+(0.910696940879)*x[2]
        arg[2,0]=(0.275470013797)*x[0]**o+(-0.903754203548)*x[0]+(-0.443211232771)*x[1]**o+(0.78815938507)*x[1]+(-0.0802934106722)*x[2]**o+(0.096532189886)*x[2]
        arg[2,1]=(-0.4505392488)*x[0]**o+(-0.668542586506)*x[0]+(0.487967999984)*x[1]**o+(-0.642310745404)*x[1]+(-0.956530771836)*x[2]**o+(0.517544404873)*x[2]
        arg[2,2]=(-0.984032885692)*x[0]**o+(-0.606688527142)*x[0]+(0.0565483776156)*x[1]**o+(-0.945228520749)*x[1]+(0.433070187921)*x[2]**o+(-0.599983370996)*x[2]
        arg[2,3]=(-0.673469008741)*x[0]**o+(0.612144012035)*x[0]+(-0.169497339443)*x[1]**o+(0.267990057152)*x[1]+(0.431716413548)*x[2]**o+(0.188904436012)*x[2]
        arg[2,4]=(0.320782309844)*x[0]**o+(-0.830068082984)*x[0]+(0.900183436056)*x[1]**o+(-0.343951186189)*x[1]+(-0.00137135137854)*x[2]**o+(-0.655764522249)*x[2]
        arg[3,0]=(-0.936193981767)*x[0]**o+(0.885260284217)*x[0]+(0.00397725971098)*x[1]**o+(-0.0864751234501)*x[1]+(0.25001166345)*x[2]**o+(-0.127859505667)*x[2]
        arg[3,1]=(-0.0841986843231)*x[0]**o+(-0.95678469673)*x[0]+(0.867049114883)*x[1]**o+(0.154499266839)*x[1]+(0.0217146474622)*x[2]**o+(-0.945774319527)*x[2]
        arg[3,2]=(-0.668161122306)*x[0]**o+(0.779140555525)*x[0]+(0.217613826157)*x[1]**o+(0.900077039607)*x[1]+(-0.251881965789)*x[2]**o+(-0.331281723354)*x[2]
        arg[3,3]=(-0.95624079226)*x[0]**o+(-0.403226291225)*x[0]+(0.303340462877)*x[1]**o+(-0.269819286355)*x[1]+(0.605804508829)*x[2]**o+(-0.461404715818)*x[2]
        arg[3,4]=(-0.933059329973)*x[0]**o+(-0.00851461368762)*x[0]+(0.0588544443918)*x[1]**o+(-0.682939702389)*x[1]+(-0.213965742261)*x[2]**o+(-0.102965977964)*x[2]
        ref[0,0]=(0.0305951763245)/(o+1.)+(0.44778386017)
        ref[0,1]=(0.745241915348)/(o+1.)+(-0.964269897247)
        ref[0,2]=(-0.352759405473)/(o+1.)+(-0.257170739438)
        ref[0,3]=(-0.163049004521)/(o+1.)+(-0.418386933036)
        ref[0,4]=(-0.812437465462)/(o+1.)+(0.360717213727)
        ref[1,0]=(-1.8232646627)/(o+1.)+(-1.2650172213)
        ref[1,1]=(1.03508884868)/(o+1.)+(0.534801547766)
        ref[1,2]=(-0.226218905227)/(o+1.)+(-0.80743356577)
        ref[1,3]=(0.0397346133545)/(o+1.)+(0.463655479743)
        ref[1,4]=(0.747176941866)/(o+1.)+(1.22951661319)
        ref[2,0]=(-0.248034629647)/(o+1.)+(-0.00953131429598)
        ref[2,1]=(-0.919102020652)/(o+1.)+(-0.396654463518)
        ref[2,2]=(-0.494414320156)/(o+1.)+(-1.07595020944)
        ref[2,3]=(-0.411249934637)/(o+1.)+(0.534519252599)
        ref[2,4]=(1.21959439452)/(o+1.)+(-0.914891895711)
        ref[3,0]=(-0.682205058606)/(o+1.)+(0.33546282755)
        ref[3,1]=(0.804565078022)/(o+1.)+(-0.874029874709)
        ref[3,2]=(-0.702429261938)/(o+1.)+(0.673967935889)
        ref[3,3]=(-0.0470958205546)/(o+1.)+(-0.567225146699)
        ref[3,4]=(-1.08817062784)/(o+1.)+(-0.39721014702)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.401537399178)*x[0]**o+(-0.868937065568)*x[0]+(0.356734884346)*x[1]**o+(0.0137485089422)*x[1]
        arg[0,0,1]=(0.705383698155)*x[0]**o+(-0.884988282352)*x[0]+(-0.979185994641)*x[1]**o+(0.929576942519)*x[1]
        arg[0,1,0]=(-0.292540086272)*x[0]**o+(-0.223230840915)*x[0]+(0.717291109855)*x[1]**o+(-0.826233591275)*x[1]
        arg[0,1,1]=(0.55739919643)*x[0]**o+(0.2571407205)*x[0]+(0.00434404142287)*x[1]**o+(0.200305144451)*x[1]
        arg[1,0,0]=(0.451841947443)*x[0]**o+(-0.576279050146)*x[0]+(-0.128142118055)*x[1]**o+(0.316406960557)*x[1]
        arg[1,0,1]=(-0.07174792925)*x[0]**o+(-0.392935435009)*x[0]+(-0.343535485172)*x[1]**o+(0.619131193274)*x[1]
        arg[1,1,0]=(-0.667031669611)*x[0]**o+(-0.184806094874)*x[0]+(-0.199375296384)*x[1]**o+(0.449351889815)*x[1]
        arg[1,1,1]=(-0.51924508404)*x[0]**o+(0.870508122509)*x[0]+(-0.274149433148)*x[1]**o+(-0.762367215962)*x[1]
        arg[2,0,0]=(-0.501546909709)*x[0]**o+(0.164122129343)*x[0]+(-0.601370828214)*x[1]**o+(0.888046948888)*x[1]
        arg[2,0,1]=(0.272472486899)*x[0]**o+(-0.958776296738)*x[0]+(-0.334676045163)*x[1]**o+(0.410943596779)*x[1]
        arg[2,1,0]=(0.363478201368)*x[0]**o+(0.753492598803)*x[0]+(-0.663726768274)*x[1]**o+(0.875443816963)*x[1]
        arg[2,1,1]=(0.995788536211)*x[0]**o+(-0.123932097555)*x[0]+(0.143025171293)*x[1]**o+(0.724918059214)*x[1]
        arg[3,0,0]=(-0.578717829755)*x[0]**o+(-0.899523186929)*x[0]+(0.58523203854)*x[1]**o+(0.751620714771)*x[1]
        arg[3,0,1]=(0.179841033744)*x[0]**o+(0.268015839688)*x[0]+(0.860292991989)*x[1]**o+(0.0117969420991)*x[1]
        arg[3,1,0]=(-0.575777000777)*x[0]**o+(0.170047012345)*x[0]+(-0.418743243278)*x[1]**o+(-0.278415849566)*x[1]
        arg[3,1,1]=(-0.913076795648)*x[0]**o+(0.486330450107)*x[0]+(-0.423570363453)*x[1]**o+(0.388575474946)*x[1]
        arg[4,0,0]=(0.16695368621)*x[0]**o+(0.133497460585)*x[0]+(-0.444434659151)*x[1]**o+(-0.895229636275)*x[1]
        arg[4,0,1]=(0.432018868106)*x[0]**o+(-0.281976075625)*x[0]+(-0.0455135209769)*x[1]**o+(0.76870037059)*x[1]
        arg[4,1,0]=(0.238542836112)*x[0]**o+(-0.806931884409)*x[0]+(-0.0365533871189)*x[1]**o+(0.731336634477)*x[1]
        arg[4,1,1]=(0.512035255059)*x[0]**o+(-0.683560978902)*x[0]+(-0.953782277737)*x[1]**o+(-0.311664210765)*x[1]
        arg[5,0,0]=(-0.842399299624)*x[0]**o+(0.460676816656)*x[0]+(-0.251781603841)*x[1]**o+(0.321792359074)*x[1]
        arg[5,0,1]=(-0.900576257062)*x[0]**o+(0.842393710528)*x[0]+(0.630445446843)*x[1]**o+(-0.647745144739)*x[1]
        arg[5,1,0]=(0.136687908652)*x[0]**o+(0.0252907890641)*x[0]+(0.813054611844)*x[1]**o+(-0.459930787056)*x[1]
        arg[5,1,1]=(-0.00242518550941)*x[0]**o+(-0.558785348137)*x[0]+(0.0568659980522)*x[1]**o+(0.689293614051)*x[1]
        ref[0,0,0]=(0.758272283524)/(o+1.)+(-0.427594278313)
        ref[0,0,1]=(-0.273802296487)/(o+1.)+(0.0222943300833)
        ref[0,1,0]=(0.424751023583)/(o+1.)+(-0.524732216095)
        ref[0,1,1]=(0.561743237852)/(o+1.)+(0.228722932476)
        ref[1,0,0]=(0.323699829388)/(o+1.)+(-0.129936044795)
        ref[1,0,1]=(-0.415283414421)/(o+1.)+(0.113097879133)
        ref[1,1,0]=(-0.866406965995)/(o+1.)+(0.132272897471)
        ref[1,1,1]=(-0.793394517189)/(o+1.)+(0.0540704532731)
        ref[2,0,0]=(-1.10291773792)/(o+1.)+(0.526084539115)
        ref[2,0,1]=(-0.0622035582637)/(o+1.)+(-0.273916349979)
        ref[2,1,0]=(-0.300248566906)/(o+1.)+(0.814468207883)
        ref[2,1,1]=(1.1388137075)/(o+1.)+(0.300492980829)
        ref[3,0,0]=(0.00651420878498)/(o+1.)+(-0.073951236079)
        ref[3,0,1]=(1.04013402573)/(o+1.)+(0.139906390894)
        ref[3,1,0]=(-0.994520244055)/(o+1.)+(-0.0541844186105)
        ref[3,1,1]=(-1.3366471591)/(o+1.)+(0.437452962527)
        ref[4,0,0]=(-0.27748097294)/(o+1.)+(-0.380866087845)
        ref[4,0,1]=(0.386505347129)/(o+1.)+(0.243362147483)
        ref[4,1,0]=(0.201989448993)/(o+1.)+(-0.0377976249661)
        ref[4,1,1]=(-0.441747022678)/(o+1.)+(-0.497612594833)
        ref[5,0,0]=(-1.09418090346)/(o+1.)+(0.391234587865)
        ref[5,0,1]=(-0.270130810219)/(o+1.)+(0.0973242828949)
        ref[5,1,0]=(0.949742520497)/(o+1.)+(-0.217319998996)
        ref[5,1,1]=(0.0544408125428)/(o+1.)+(0.0652541329568)
      else:
        arg[0,0,0]=(0.411788818613)*x[0]**o+(0.324151698669)*x[0]+(0.656551349616)*x[1]**o+(0.801577401197)*x[1]+(0.690440431022)*x[2]**o+(0.776409870151)*x[2]
        arg[0,0,1]=(-0.224985947659)*x[0]**o+(-0.564703086188)*x[0]+(-0.37196336116)*x[1]**o+(-0.114979853172)*x[1]+(0.274462080499)*x[2]**o+(-0.988489021983)*x[2]
        arg[0,1,0]=(0.512331530009)*x[0]**o+(0.519617843594)*x[0]+(-0.515174861158)*x[1]**o+(-0.675953902994)*x[1]+(0.796111490716)*x[2]**o+(0.229569489858)*x[2]
        arg[0,1,1]=(0.347765144376)*x[0]**o+(-0.0110194730031)*x[0]+(-0.593936645599)*x[1]**o+(-0.372786558089)*x[1]+(-0.604691321362)*x[2]**o+(-0.888002521077)*x[2]
        arg[1,0,0]=(-0.575954903433)*x[0]**o+(0.0160235655616)*x[0]+(0.305335632724)*x[1]**o+(-0.259402553796)*x[1]+(-0.221276854741)*x[2]**o+(-0.519932341069)*x[2]
        arg[1,0,1]=(-0.674987647238)*x[0]**o+(-0.45852667934)*x[0]+(-0.758937440978)*x[1]**o+(0.80566898197)*x[1]+(-0.961619154322)*x[2]**o+(0.861564430243)*x[2]
        arg[1,1,0]=(-0.297475412437)*x[0]**o+(0.881828705604)*x[0]+(-0.823267615874)*x[1]**o+(-0.0232728101313)*x[1]+(-0.676220001453)*x[2]**o+(-0.382414036981)*x[2]
        arg[1,1,1]=(-0.31008204616)*x[0]**o+(0.805195900401)*x[0]+(-0.758209772143)*x[1]**o+(0.861379709073)*x[1]+(0.0631063866518)*x[2]**o+(0.356607080553)*x[2]
        arg[2,0,0]=(0.305214964591)*x[0]**o+(0.815106526139)*x[0]+(0.138797156005)*x[1]**o+(-0.241921896332)*x[1]+(-0.103783035058)*x[2]**o+(-0.161952044984)*x[2]
        arg[2,0,1]=(-0.925375677635)*x[0]**o+(0.0341347568201)*x[0]+(-0.706905924964)*x[1]**o+(-0.193304368451)*x[1]+(-0.449866757104)*x[2]**o+(-0.894963680869)*x[2]
        arg[2,1,0]=(0.822676224279)*x[0]**o+(-0.28758006682)*x[0]+(-0.869025918917)*x[1]**o+(0.0107602044744)*x[1]+(-0.832322765163)*x[2]**o+(0.748440319389)*x[2]
        arg[2,1,1]=(-0.415326508997)*x[0]**o+(-0.879521424817)*x[0]+(-0.971781441062)*x[1]**o+(0.527586634108)*x[1]+(0.336682669856)*x[2]**o+(-0.948414041149)*x[2]
        arg[3,0,0]=(-0.345788531715)*x[0]**o+(0.626583498807)*x[0]+(-0.186418683649)*x[1]**o+(0.3141628424)*x[1]+(-0.781119787783)*x[2]**o+(0.972708430266)*x[2]
        arg[3,0,1]=(0.175536279095)*x[0]**o+(-0.150027774295)*x[0]+(-0.506653680683)*x[1]**o+(-0.482544397029)*x[1]+(-0.59090996152)*x[2]**o+(0.618123625026)*x[2]
        arg[3,1,0]=(0.737984493124)*x[0]**o+(0.650901680297)*x[0]+(-0.361389200009)*x[1]**o+(-0.784304665396)*x[1]+(-0.954132232142)*x[2]**o+(0.684341934348)*x[2]
        arg[3,1,1]=(-0.385757579645)*x[0]**o+(0.251136919845)*x[0]+(0.95143939933)*x[1]**o+(0.501860043225)*x[1]+(-0.675673011184)*x[2]**o+(0.337315450579)*x[2]
        arg[4,0,0]=(-0.740183402139)*x[0]**o+(-0.828278760743)*x[0]+(0.88000889117)*x[1]**o+(-0.0226801022184)*x[1]+(0.6811397662)*x[2]**o+(-0.254627446498)*x[2]
        arg[4,0,1]=(-0.194128952361)*x[0]**o+(-0.524854896388)*x[0]+(0.341100326577)*x[1]**o+(-0.26072443499)*x[1]+(-0.197396331776)*x[2]**o+(-0.589037632112)*x[2]
        arg[4,1,0]=(0.261995024865)*x[0]**o+(0.214193793699)*x[0]+(0.572024600244)*x[1]**o+(0.230659327198)*x[1]+(0.420971705453)*x[2]**o+(0.333445049827)*x[2]
        arg[4,1,1]=(-0.0679525054395)*x[0]**o+(0.0756646296473)*x[0]+(-0.299865836239)*x[1]**o+(-0.828294764914)*x[1]+(0.606170745631)*x[2]**o+(0.624267827393)*x[2]
        arg[5,0,0]=(0.312565542211)*x[0]**o+(-0.368815496171)*x[0]+(-0.0615973473105)*x[1]**o+(0.205012835971)*x[1]+(0.310818553076)*x[2]**o+(-0.287393908304)*x[2]
        arg[5,0,1]=(-0.409626309069)*x[0]**o+(0.977953573562)*x[0]+(0.583013544138)*x[1]**o+(0.542907128677)*x[1]+(0.488538336366)*x[2]**o+(-0.85823611123)*x[2]
        arg[5,1,0]=(-0.899607109395)*x[0]**o+(-0.0459053040522)*x[0]+(0.938423293041)*x[1]**o+(-0.112431665317)*x[1]+(0.620609328687)*x[2]**o+(-0.663082429319)*x[2]
        arg[5,1,1]=(0.00370253144626)*x[0]**o+(0.503733497956)*x[0]+(0.362457785043)*x[1]**o+(0.417891957303)*x[1]+(0.404543677829)*x[2]**o+(0.117828669122)*x[2]
        ref[0,0,0]=(1.75878059925)/(o+1.)+(0.951069485009)
        ref[0,0,1]=(-0.32248722832)/(o+1.)+(-0.834085980672)
        ref[0,1,0]=(0.793268159567)/(o+1.)+(0.036616715229)
        ref[0,1,1]=(-0.850862822585)/(o+1.)+(-0.635904276085)
        ref[1,0,0]=(-0.49189612545)/(o+1.)+(-0.381655664652)
        ref[1,0,1]=(-2.39554424254)/(o+1.)+(0.604353366437)
        ref[1,1,0]=(-1.79696302976)/(o+1.)+(0.238070929246)
        ref[1,1,1]=(-1.00518543165)/(o+1.)+(1.01159134501)
        ref[2,0,0]=(0.340229085537)/(o+1.)+(0.205616292412)
        ref[2,0,1]=(-2.0821483597)/(o+1.)+(-0.52706664625)
        ref[2,1,0]=(-0.878672459802)/(o+1.)+(0.235810228522)
        ref[2,1,1]=(-1.0504252802)/(o+1.)+(-0.650174415929)
        ref[3,0,0]=(-1.31332700315)/(o+1.)+(0.956727385736)
        ref[3,0,1]=(-0.922027363109)/(o+1.)+(-0.00722427314892)
        ref[3,1,0]=(-0.577536939027)/(o+1.)+(0.275469474624)
        ref[3,1,1]=(-0.109991191499)/(o+1.)+(0.545156206825)
        ref[4,0,0]=(0.820965255231)/(o+1.)+(-0.55279315473)
        ref[4,0,1]=(-0.0504249575602)/(o+1.)+(-0.687308481745)
        ref[4,1,0]=(1.25499133056)/(o+1.)+(0.389149085362)
        ref[4,1,1]=(0.238352403952)/(o+1.)+(-0.0641811539369)
        ref[5,0,0]=(0.561786747976)/(o+1.)+(-0.225598284252)
        ref[5,0,1]=(0.661925571434)/(o+1.)+(0.331312295504)
        ref[5,1,0]=(0.659425512333)/(o+1.)+(-0.410709699344)
        ref[5,1,1]=(0.770703994319)/(o+1.)+(0.519727062191)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.164405890316)*x[0]**o+(-0.323303939764)*x[0]+(-0.0516829984856)*x[1]**o+(-0.220765052533)*x[1]
        arg[0,0,0,1]=(0.152172612108)*x[0]**o+(-0.166242164365)*x[0]+(0.807177808571)*x[1]**o+(0.0404817109573)*x[1]
        arg[0,0,1,0]=(0.016912808567)*x[0]**o+(0.623923316774)*x[0]+(-0.542886934905)*x[1]**o+(0.0110784560205)*x[1]
        arg[0,0,1,1]=(-0.663885786909)*x[0]**o+(-0.55563355485)*x[0]+(0.191993284382)*x[1]**o+(0.734973556043)*x[1]
        arg[0,0,2,0]=(-0.110316754183)*x[0]**o+(-0.584022783672)*x[0]+(0.56471376011)*x[1]**o+(0.846317747143)*x[1]
        arg[0,0,2,1]=(-0.767988009278)*x[0]**o+(0.440997912352)*x[0]+(0.45366348093)*x[1]**o+(0.376620990318)*x[1]
        arg[0,1,0,0]=(-0.138006508051)*x[0]**o+(0.8450564766)*x[0]+(-0.273559897633)*x[1]**o+(0.683678615645)*x[1]
        arg[0,1,0,1]=(-0.467521625474)*x[0]**o+(0.32289306962)*x[0]+(0.340974562152)*x[1]**o+(-0.58109391053)*x[1]
        arg[0,1,1,0]=(0.76867991208)*x[0]**o+(0.683527735131)*x[0]+(0.762986872389)*x[1]**o+(-0.189342734286)*x[1]
        arg[0,1,1,1]=(0.125150175351)*x[0]**o+(-0.175190872876)*x[0]+(-0.569766488076)*x[1]**o+(0.787503755206)*x[1]
        arg[0,1,2,0]=(-0.339695860559)*x[0]**o+(0.119452596206)*x[0]+(0.474378864192)*x[1]**o+(0.845982555265)*x[1]
        arg[0,1,2,1]=(0.820551334836)*x[0]**o+(-0.200024721285)*x[0]+(-0.525871855871)*x[1]**o+(-0.714639297791)*x[1]
        arg[0,2,0,0]=(-0.145788788262)*x[0]**o+(-0.695717636982)*x[0]+(0.736564332801)*x[1]**o+(0.634307016587)*x[1]
        arg[0,2,0,1]=(-0.25142006947)*x[0]**o+(0.027028767127)*x[0]+(-0.614777438028)*x[1]**o+(-0.691417785617)*x[1]
        arg[0,2,1,0]=(0.366081140839)*x[0]**o+(0.477490533426)*x[0]+(-0.116611335116)*x[1]**o+(-0.863552165712)*x[1]
        arg[0,2,1,1]=(-0.467659558308)*x[0]**o+(0.0836299545334)*x[0]+(0.903093252833)*x[1]**o+(-0.956858182533)*x[1]
        arg[0,2,2,0]=(0.870921031579)*x[0]**o+(-0.801681014465)*x[0]+(-0.951819608329)*x[1]**o+(-0.626222917084)*x[1]
        arg[0,2,2,1]=(0.414973183041)*x[0]**o+(-0.444580205703)*x[0]+(-0.346458764208)*x[1]**o+(-0.679140100745)*x[1]
        arg[0,3,0,0]=(-0.725326935158)*x[0]**o+(-0.914034759513)*x[0]+(0.702614377806)*x[1]**o+(0.819004276317)*x[1]
        arg[0,3,0,1]=(0.617099932352)*x[0]**o+(-0.155182850011)*x[0]+(-0.066428715129)*x[1]**o+(-0.286480585435)*x[1]
        arg[0,3,1,0]=(0.710942545488)*x[0]**o+(-0.734509277268)*x[0]+(-0.807474546329)*x[1]**o+(0.307631760449)*x[1]
        arg[0,3,1,1]=(-0.709966924732)*x[0]**o+(-0.462297022374)*x[0]+(0.773787774038)*x[1]**o+(0.762137020991)*x[1]
        arg[0,3,2,0]=(-0.425660932209)*x[0]**o+(-0.562130300181)*x[0]+(-0.516719014647)*x[1]**o+(-0.547832887369)*x[1]
        arg[0,3,2,1]=(0.149573176879)*x[0]**o+(-0.208999629436)*x[0]+(-0.191426033553)*x[1]**o+(-0.119046538873)*x[1]
        arg[0,4,0,0]=(-0.99932317151)*x[0]**o+(-0.832461558068)*x[0]+(-0.847073271676)*x[1]**o+(0.882046887729)*x[1]
        arg[0,4,0,1]=(0.14235557233)*x[0]**o+(-0.352572053276)*x[0]+(-0.231989519255)*x[1]**o+(0.146855135398)*x[1]
        arg[0,4,1,0]=(0.550561449012)*x[0]**o+(0.530745681425)*x[0]+(0.198889689472)*x[1]**o+(0.165503108658)*x[1]
        arg[0,4,1,1]=(0.059663401309)*x[0]**o+(0.168377411751)*x[0]+(-0.224682723925)*x[1]**o+(-0.255406907757)*x[1]
        arg[0,4,2,0]=(-0.747776894683)*x[0]**o+(-0.281763358533)*x[0]+(0.660369838923)*x[1]**o+(0.348541234687)*x[1]
        arg[0,4,2,1]=(-0.886248366716)*x[0]**o+(0.665256341312)*x[0]+(-0.118041122298)*x[1]**o+(0.192284814559)*x[1]
        arg[1,0,0,0]=(-0.212291422817)*x[0]**o+(-0.16123490382)*x[0]+(-0.884220059537)*x[1]**o+(-0.40104775897)*x[1]
        arg[1,0,0,1]=(0.389745305389)*x[0]**o+(0.544302066739)*x[0]+(-0.504770065056)*x[1]**o+(0.903371132654)*x[1]
        arg[1,0,1,0]=(0.983764499357)*x[0]**o+(0.0707766278425)*x[0]+(-0.119851167133)*x[1]**o+(0.668490934432)*x[1]
        arg[1,0,1,1]=(-0.61333738947)*x[0]**o+(-0.515837139916)*x[0]+(-0.825015725806)*x[1]**o+(-0.858420009384)*x[1]
        arg[1,0,2,0]=(0.838488008532)*x[0]**o+(0.320020900158)*x[0]+(0.265140913892)*x[1]**o+(0.493460973766)*x[1]
        arg[1,0,2,1]=(-0.957171379585)*x[0]**o+(0.911667647687)*x[0]+(0.0634495769152)*x[1]**o+(-0.244783653159)*x[1]
        arg[1,1,0,0]=(0.854312398231)*x[0]**o+(0.0177604750376)*x[0]+(-0.263135825468)*x[1]**o+(0.196452389095)*x[1]
        arg[1,1,0,1]=(0.496952473267)*x[0]**o+(-0.505581680291)*x[0]+(0.227653536528)*x[1]**o+(-0.138354150837)*x[1]
        arg[1,1,1,0]=(0.569682536816)*x[0]**o+(-0.523174667456)*x[0]+(-0.434073415962)*x[1]**o+(-0.853667257479)*x[1]
        arg[1,1,1,1]=(0.636882167235)*x[0]**o+(-0.122206427025)*x[0]+(0.934361025063)*x[1]**o+(-0.43037160177)*x[1]
        arg[1,1,2,0]=(0.0798318147504)*x[0]**o+(0.825231320185)*x[0]+(0.810701650808)*x[1]**o+(-0.544356371381)*x[1]
        arg[1,1,2,1]=(-0.426710170704)*x[0]**o+(0.200697079677)*x[0]+(0.242027752153)*x[1]**o+(0.746762119874)*x[1]
        arg[1,2,0,0]=(-0.497091856216)*x[0]**o+(0.0953816987137)*x[0]+(-0.388615292775)*x[1]**o+(0.794962452203)*x[1]
        arg[1,2,0,1]=(0.484769626964)*x[0]**o+(0.700484219762)*x[0]+(-0.230377697673)*x[1]**o+(-0.110327221848)*x[1]
        arg[1,2,1,0]=(-0.791806541129)*x[0]**o+(0.0199911517557)*x[0]+(-0.441657020836)*x[1]**o+(-0.972999930466)*x[1]
        arg[1,2,1,1]=(-0.566750983493)*x[0]**o+(0.193689535753)*x[0]+(0.659310468709)*x[1]**o+(0.209045229813)*x[1]
        arg[1,2,2,0]=(0.852553900341)*x[0]**o+(-0.869275488172)*x[0]+(-0.308429715369)*x[1]**o+(0.096335995297)*x[1]
        arg[1,2,2,1]=(0.169351018917)*x[0]**o+(0.269577471686)*x[0]+(-0.226840840616)*x[1]**o+(-0.94229540867)*x[1]
        arg[1,3,0,0]=(0.481711489234)*x[0]**o+(0.293576608201)*x[0]+(0.101998881767)*x[1]**o+(-0.10730442531)*x[1]
        arg[1,3,0,1]=(-0.824144820539)*x[0]**o+(0.174754759038)*x[0]+(-0.251803196848)*x[1]**o+(-0.389656255459)*x[1]
        arg[1,3,1,0]=(0.438310633462)*x[0]**o+(-0.268051020546)*x[0]+(0.670456978527)*x[1]**o+(-0.547393352784)*x[1]
        arg[1,3,1,1]=(-0.95082246973)*x[0]**o+(0.739061218443)*x[0]+(0.209495253219)*x[1]**o+(-0.604801908581)*x[1]
        arg[1,3,2,0]=(0.696900724687)*x[0]**o+(0.501191311274)*x[0]+(0.982952568975)*x[1]**o+(0.1345704134)*x[1]
        arg[1,3,2,1]=(0.823865642457)*x[0]**o+(0.344980106206)*x[0]+(0.0740071873515)*x[1]**o+(-0.950460516779)*x[1]
        arg[1,4,0,0]=(-0.428525532814)*x[0]**o+(0.496012936411)*x[0]+(-0.0756953876324)*x[1]**o+(0.8382068924)*x[1]
        arg[1,4,0,1]=(0.950388644852)*x[0]**o+(-0.225116488186)*x[0]+(0.515802878592)*x[1]**o+(-0.51004230039)*x[1]
        arg[1,4,1,0]=(0.868733254296)*x[0]**o+(-0.506544709529)*x[0]+(0.28935579533)*x[1]**o+(0.100277736612)*x[1]
        arg[1,4,1,1]=(-0.104533384722)*x[0]**o+(0.598995288066)*x[0]+(-0.738946772839)*x[1]**o+(-0.23245555216)*x[1]
        arg[1,4,2,0]=(-0.286232179213)*x[0]**o+(-0.31267596034)*x[0]+(0.588681250431)*x[1]**o+(-0.579857978487)*x[1]
        arg[1,4,2,1]=(0.405666376198)*x[0]**o+(0.380029842735)*x[0]+(-0.646672003588)*x[1]**o+(0.419606979519)*x[1]
        arg[2,0,0,0]=(0.183075434304)*x[0]**o+(0.0929532345426)*x[0]+(-0.742007638693)*x[1]**o+(-0.521914593386)*x[1]
        arg[2,0,0,1]=(0.292878118412)*x[0]**o+(-0.427165220789)*x[0]+(0.540033289295)*x[1]**o+(0.997945616043)*x[1]
        arg[2,0,1,0]=(0.872801856197)*x[0]**o+(0.0427577769706)*x[0]+(0.595354626662)*x[1]**o+(-0.878459740609)*x[1]
        arg[2,0,1,1]=(-0.259988866684)*x[0]**o+(0.154043636253)*x[0]+(0.0580047878991)*x[1]**o+(0.40696744056)*x[1]
        arg[2,0,2,0]=(-0.602648518466)*x[0]**o+(0.30877596654)*x[0]+(-0.268563459489)*x[1]**o+(-0.526379663585)*x[1]
        arg[2,0,2,1]=(-0.375777994931)*x[0]**o+(-0.175440011709)*x[0]+(-0.578650003477)*x[1]**o+(0.822157683321)*x[1]
        arg[2,1,0,0]=(-0.44389226502)*x[0]**o+(0.683580346916)*x[0]+(0.471857559282)*x[1]**o+(-0.869624665398)*x[1]
        arg[2,1,0,1]=(0.768272315036)*x[0]**o+(-0.709688321973)*x[0]+(0.200727881948)*x[1]**o+(0.617326752785)*x[1]
        arg[2,1,1,0]=(-0.371169228982)*x[0]**o+(-0.0248248405104)*x[0]+(-0.899904749204)*x[1]**o+(0.796779883083)*x[1]
        arg[2,1,1,1]=(0.220088670807)*x[0]**o+(0.691718163435)*x[0]+(-0.0834793048429)*x[1]**o+(-0.748110928988)*x[1]
        arg[2,1,2,0]=(0.246045293748)*x[0]**o+(-0.669660154233)*x[0]+(0.601040129345)*x[1]**o+(-0.0893792369983)*x[1]
        arg[2,1,2,1]=(0.318708374425)*x[0]**o+(-0.973048583674)*x[0]+(-0.575891882633)*x[1]**o+(0.229050058038)*x[1]
        arg[2,2,0,0]=(0.358008410645)*x[0]**o+(-0.834388697702)*x[0]+(0.0931465346237)*x[1]**o+(0.503934981947)*x[1]
        arg[2,2,0,1]=(-0.257597986495)*x[0]**o+(-0.58070568528)*x[0]+(0.735541935653)*x[1]**o+(-0.704666216263)*x[1]
        arg[2,2,1,0]=(0.0207421250729)*x[0]**o+(-0.724644198871)*x[0]+(0.637124335031)*x[1]**o+(-0.547300138808)*x[1]
        arg[2,2,1,1]=(0.280179217797)*x[0]**o+(-0.387279903944)*x[0]+(0.516275341085)*x[1]**o+(0.781250973385)*x[1]
        arg[2,2,2,0]=(-0.97771111517)*x[0]**o+(-0.140997969015)*x[0]+(-0.495112130955)*x[1]**o+(0.151523726543)*x[1]
        arg[2,2,2,1]=(-0.48638258963)*x[0]**o+(-0.251363723104)*x[0]+(-0.484911659879)*x[1]**o+(-0.841285872322)*x[1]
        arg[2,3,0,0]=(-0.70159761565)*x[0]**o+(-0.661658396735)*x[0]+(-0.380920250292)*x[1]**o+(0.682598882025)*x[1]
        arg[2,3,0,1]=(0.477056088757)*x[0]**o+(0.697397619448)*x[0]+(0.28617323153)*x[1]**o+(-0.0443690469658)*x[1]
        arg[2,3,1,0]=(-0.823038323615)*x[0]**o+(-0.742146866184)*x[0]+(0.258604766787)*x[1]**o+(-0.740368164141)*x[1]
        arg[2,3,1,1]=(0.958751077013)*x[0]**o+(0.353976218746)*x[0]+(-0.954344840728)*x[1]**o+(0.436915389196)*x[1]
        arg[2,3,2,0]=(0.432342075923)*x[0]**o+(-0.0699761039653)*x[0]+(0.701490046186)*x[1]**o+(0.0888793443125)*x[1]
        arg[2,3,2,1]=(0.612536667525)*x[0]**o+(0.107433723026)*x[0]+(-0.611874456965)*x[1]**o+(-0.253338899867)*x[1]
        arg[2,4,0,0]=(-0.0234819568317)*x[0]**o+(-0.0642964170457)*x[0]+(0.556679188908)*x[1]**o+(0.865922901373)*x[1]
        arg[2,4,0,1]=(0.669235019483)*x[0]**o+(0.744361910609)*x[0]+(0.383397302403)*x[1]**o+(0.669587709042)*x[1]
        arg[2,4,1,0]=(-0.143461051192)*x[0]**o+(-0.0272388772724)*x[0]+(0.723868767782)*x[1]**o+(0.303617835707)*x[1]
        arg[2,4,1,1]=(-0.829765013546)*x[0]**o+(-0.784193619495)*x[0]+(0.384376849236)*x[1]**o+(0.280904709362)*x[1]
        arg[2,4,2,0]=(0.38737050935)*x[0]**o+(0.749614613361)*x[0]+(-0.340848293488)*x[1]**o+(0.96291850214)*x[1]
        arg[2,4,2,1]=(0.619896872679)*x[0]**o+(0.970694589398)*x[0]+(-0.300698305892)*x[1]**o+(0.794408153506)*x[1]
        arg[3,0,0,0]=(-0.808442463698)*x[0]**o+(-0.527555255968)*x[0]+(0.246776596713)*x[1]**o+(-0.719809811547)*x[1]
        arg[3,0,0,1]=(-0.390016002326)*x[0]**o+(-0.864168521951)*x[0]+(-0.154649478359)*x[1]**o+(0.210766031283)*x[1]
        arg[3,0,1,0]=(-0.941916382966)*x[0]**o+(-0.0352551994705)*x[0]+(-0.660026082442)*x[1]**o+(-0.869241000235)*x[1]
        arg[3,0,1,1]=(-0.287992443961)*x[0]**o+(0.292551506636)*x[0]+(-0.619953648778)*x[1]**o+(0.746095811695)*x[1]
        arg[3,0,2,0]=(-0.233659077726)*x[0]**o+(-0.568647415596)*x[0]+(-0.22048065705)*x[1]**o+(0.658105381132)*x[1]
        arg[3,0,2,1]=(0.871784854897)*x[0]**o+(-0.118389479971)*x[0]+(0.143290959871)*x[1]**o+(0.753198625961)*x[1]
        arg[3,1,0,0]=(0.993988451174)*x[0]**o+(0.451849763406)*x[0]+(-0.972805640824)*x[1]**o+(0.796893857099)*x[1]
        arg[3,1,0,1]=(-0.424353439197)*x[0]**o+(-0.575799061446)*x[0]+(-0.233936106648)*x[1]**o+(-0.451864349448)*x[1]
        arg[3,1,1,0]=(0.639821887614)*x[0]**o+(0.771041944334)*x[0]+(0.420297869214)*x[1]**o+(0.671734395857)*x[1]
        arg[3,1,1,1]=(0.398121897075)*x[0]**o+(0.647679737081)*x[0]+(-0.544457308885)*x[1]**o+(0.58955287395)*x[1]
        arg[3,1,2,0]=(-0.118222351008)*x[0]**o+(-0.943807502172)*x[0]+(0.866866987138)*x[1]**o+(-0.0793340319339)*x[1]
        arg[3,1,2,1]=(-0.157286204244)*x[0]**o+(-0.789214982692)*x[0]+(0.646008960971)*x[1]**o+(0.866687256647)*x[1]
        arg[3,2,0,0]=(-0.620625598995)*x[0]**o+(-0.12874444072)*x[0]+(0.4402928966)*x[1]**o+(-0.0344340173898)*x[1]
        arg[3,2,0,1]=(-0.714597102332)*x[0]**o+(0.730206152989)*x[0]+(0.795717819655)*x[1]**o+(0.0543647593833)*x[1]
        arg[3,2,1,0]=(0.698721346048)*x[0]**o+(0.599204472096)*x[0]+(0.323529910402)*x[1]**o+(0.362678869119)*x[1]
        arg[3,2,1,1]=(0.976107824554)*x[0]**o+(0.559889392243)*x[0]+(0.240016231718)*x[1]**o+(0.343322878078)*x[1]
        arg[3,2,2,0]=(0.609695396838)*x[0]**o+(0.786331471916)*x[0]+(-0.965489152976)*x[1]**o+(0.553653504493)*x[1]
        arg[3,2,2,1]=(0.897952533418)*x[0]**o+(0.227242627929)*x[0]+(-0.121752169524)*x[1]**o+(0.844617689997)*x[1]
        arg[3,3,0,0]=(-0.249698211819)*x[0]**o+(-0.400095108874)*x[0]+(-0.44300562221)*x[1]**o+(0.387777895056)*x[1]
        arg[3,3,0,1]=(0.626469203641)*x[0]**o+(0.741893669108)*x[0]+(0.261882509899)*x[1]**o+(-0.726628233604)*x[1]
        arg[3,3,1,0]=(-0.525223500134)*x[0]**o+(0.918314547123)*x[0]+(0.426441176651)*x[1]**o+(0.338058005974)*x[1]
        arg[3,3,1,1]=(-0.31358142709)*x[0]**o+(-0.411070225036)*x[0]+(0.0902340021221)*x[1]**o+(0.404135075929)*x[1]
        arg[3,3,2,0]=(0.528842663849)*x[0]**o+(0.474768688828)*x[0]+(0.697643730083)*x[1]**o+(-0.170335125341)*x[1]
        arg[3,3,2,1]=(0.875221054674)*x[0]**o+(-0.441830492507)*x[0]+(-0.60312069438)*x[1]**o+(0.664895429753)*x[1]
        arg[3,4,0,0]=(0.563422972203)*x[0]**o+(-0.595004250424)*x[0]+(0.888927367905)*x[1]**o+(0.553640735136)*x[1]
        arg[3,4,0,1]=(0.79563891979)*x[0]**o+(-0.461626171695)*x[0]+(0.700194671524)*x[1]**o+(-0.687505010217)*x[1]
        arg[3,4,1,0]=(0.655248139354)*x[0]**o+(0.588175755435)*x[0]+(-0.29236593461)*x[1]**o+(0.575186295035)*x[1]
        arg[3,4,1,1]=(0.405730566113)*x[0]**o+(-0.161117488254)*x[0]+(0.912224633472)*x[1]**o+(0.0292423764587)*x[1]
        arg[3,4,2,0]=(-0.405819525278)*x[0]**o+(-0.362840757195)*x[0]+(0.961950188511)*x[1]**o+(0.205213542149)*x[1]
        arg[3,4,2,1]=(0.703066280254)*x[0]**o+(-0.936055321446)*x[0]+(0.861070629413)*x[1]**o+(0.0856428368256)*x[1]
        ref[0,0,0,0]=(-0.216088888802)/(o+1.)+(-0.272034496148)
        ref[0,0,0,1]=(0.959350420679)/(o+1.)+(-0.062880226704)
        ref[0,0,1,0]=(-0.525974126338)/(o+1.)+(0.317500886397)
        ref[0,0,1,1]=(-0.471892502528)/(o+1.)+(0.0896700005967)
        ref[0,0,2,0]=(0.454397005927)/(o+1.)+(0.131147481736)
        ref[0,0,2,1]=(-0.314324528348)/(o+1.)+(0.408809451335)
        ref[0,1,0,0]=(-0.411566405684)/(o+1.)+(0.764367546122)
        ref[0,1,0,1]=(-0.126547063321)/(o+1.)+(-0.129100420455)
        ref[0,1,1,0]=(1.53166678447)/(o+1.)+(0.247092500423)
        ref[0,1,1,1]=(-0.444616312725)/(o+1.)+(0.306156441165)
        ref[0,1,2,0]=(0.134683003633)/(o+1.)+(0.482717575736)
        ref[0,1,2,1]=(0.294679478965)/(o+1.)+(-0.457332009538)
        ref[0,2,0,0]=(0.590775544538)/(o+1.)+(-0.0307053101973)
        ref[0,2,0,1]=(-0.866197507498)/(o+1.)+(-0.332194509245)
        ref[0,2,1,0]=(0.249469805723)/(o+1.)+(-0.193030816143)
        ref[0,2,1,1]=(0.435433694525)/(o+1.)+(-0.436614114)
        ref[0,2,2,0]=(-0.0808985767498)/(o+1.)+(-0.713951965774)
        ref[0,2,2,1]=(0.0685144188336)/(o+1.)+(-0.561860153224)
        ref[0,3,0,0]=(-0.0227125573528)/(o+1.)+(-0.0475152415978)
        ref[0,3,0,1]=(0.550671217223)/(o+1.)+(-0.220831717723)
        ref[0,3,1,0]=(-0.0965320008409)/(o+1.)+(-0.213438758409)
        ref[0,3,1,1]=(0.0638208493064)/(o+1.)+(0.149919999309)
        ref[0,3,2,0]=(-0.942379946856)/(o+1.)+(-0.554981593775)
        ref[0,3,2,1]=(-0.0418528566735)/(o+1.)+(-0.164023084154)
        ref[0,4,0,0]=(-1.84639644319)/(o+1.)+(0.0247926648306)
        ref[0,4,0,1]=(-0.0896339469252)/(o+1.)+(-0.102858458939)
        ref[0,4,1,0]=(0.749451138484)/(o+1.)+(0.348124395042)
        ref[0,4,1,1]=(-0.165019322616)/(o+1.)+(-0.0435147480027)
        ref[0,4,2,0]=(-0.0874070557597)/(o+1.)+(0.033388938077)
        ref[0,4,2,1]=(-1.00428948901)/(o+1.)+(0.428770577935)
        ref[1,0,0,0]=(-1.09651148235)/(o+1.)+(-0.281141331395)
        ref[1,0,0,1]=(-0.115024759667)/(o+1.)+(0.723836599697)
        ref[1,0,1,0]=(0.863913332224)/(o+1.)+(0.369633781137)
        ref[1,0,1,1]=(-1.43835311528)/(o+1.)+(-0.68712857465)
        ref[1,0,2,0]=(1.10362892242)/(o+1.)+(0.406740936962)
        ref[1,0,2,1]=(-0.893721802669)/(o+1.)+(0.333441997264)
        ref[1,1,0,0]=(0.591176572763)/(o+1.)+(0.107106432066)
        ref[1,1,0,1]=(0.724606009795)/(o+1.)+(-0.321967915564)
        ref[1,1,1,0]=(0.135609120854)/(o+1.)+(-0.688420962467)
        ref[1,1,1,1]=(1.5712431923)/(o+1.)+(-0.276289014397)
        ref[1,1,2,0]=(0.890533465558)/(o+1.)+(0.140437474402)
        ref[1,1,2,1]=(-0.184682418551)/(o+1.)+(0.473729599776)
        ref[1,2,0,0]=(-0.88570714899)/(o+1.)+(0.445172075458)
        ref[1,2,0,1]=(0.254391929291)/(o+1.)+(0.295078498957)
        ref[1,2,1,0]=(-1.23346356196)/(o+1.)+(-0.476504389355)
        ref[1,2,1,1]=(0.0925594852161)/(o+1.)+(0.201367382783)
        ref[1,2,2,0]=(0.544124184972)/(o+1.)+(-0.386469746437)
        ref[1,2,2,1]=(-0.057489821699)/(o+1.)+(-0.336358968492)
        ref[1,3,0,0]=(0.583710371001)/(o+1.)+(0.0931360914458)
        ref[1,3,0,1]=(-1.07594801739)/(o+1.)+(-0.107450748211)
        ref[1,3,1,0]=(1.10876761199)/(o+1.)+(-0.407722186665)
        ref[1,3,1,1]=(-0.741327216511)/(o+1.)+(0.0671296549307)
        ref[1,3,2,0]=(1.67985329366)/(o+1.)+(0.317880862337)
        ref[1,3,2,1]=(0.897872829808)/(o+1.)+(-0.302740205287)
        ref[1,4,0,0]=(-0.504220920447)/(o+1.)+(0.667109914405)
        ref[1,4,0,1]=(1.46619152344)/(o+1.)+(-0.367579394288)
        ref[1,4,1,0]=(1.15808904963)/(o+1.)+(-0.203133486458)
        ref[1,4,1,1]=(-0.843480157561)/(o+1.)+(0.183269867953)
        ref[1,4,2,0]=(0.302449071217)/(o+1.)+(-0.446266969414)
        ref[1,4,2,1]=(-0.24100562739)/(o+1.)+(0.399818411127)
        ref[2,0,0,0]=(-0.558932204389)/(o+1.)+(-0.214480679421)
        ref[2,0,0,1]=(0.832911407707)/(o+1.)+(0.285390197627)
        ref[2,0,1,0]=(1.46815648286)/(o+1.)+(-0.417850981819)
        ref[2,0,1,1]=(-0.201984078785)/(o+1.)+(0.280505538406)
        ref[2,0,2,0]=(-0.871211977955)/(o+1.)+(-0.108801848522)
        ref[2,0,2,1]=(-0.954427998409)/(o+1.)+(0.323358835806)
        ref[2,1,0,0]=(0.0279652942615)/(o+1.)+(-0.0930221592407)
        ref[2,1,0,1]=(0.969000196983)/(o+1.)+(-0.0461807845941)
        ref[2,1,1,0]=(-1.27107397819)/(o+1.)+(0.385977521287)
        ref[2,1,1,1]=(0.136609365964)/(o+1.)+(-0.0281963827764)
        ref[2,1,2,0]=(0.847085423093)/(o+1.)+(-0.379519695616)
        ref[2,1,2,1]=(-0.257183508208)/(o+1.)+(-0.371999262818)
        ref[2,2,0,0]=(0.451154945268)/(o+1.)+(-0.165226857878)
        ref[2,2,0,1]=(0.477943949158)/(o+1.)+(-0.642685950772)
        ref[2,2,1,0]=(0.657866460103)/(o+1.)+(-0.635972168839)
        ref[2,2,1,1]=(0.796454558882)/(o+1.)+(0.19698553472)
        ref[2,2,2,0]=(-1.47282324613)/(o+1.)+(0.00526287876389)
        ref[2,2,2,1]=(-0.971294249509)/(o+1.)+(-0.546324797713)
        ref[2,3,0,0]=(-1.08251786594)/(o+1.)+(0.0104702426452)
        ref[2,3,0,1]=(0.763229320287)/(o+1.)+(0.326514286241)
        ref[2,3,1,0]=(-0.564433556829)/(o+1.)+(-0.741257515163)
        ref[2,3,1,1]=(0.00440623628526)/(o+1.)+(0.395445803971)
        ref[2,3,2,0]=(1.13383212211)/(o+1.)+(0.0094516201736)
        ref[2,3,2,1]=(0.000662210559809)/(o+1.)+(-0.0729525884203)
        ref[2,4,0,0]=(0.533197232076)/(o+1.)+(0.400813242164)
        ref[2,4,0,1]=(1.05263232189)/(o+1.)+(0.706974809826)
        ref[2,4,1,0]=(0.580407716589)/(o+1.)+(0.138189479217)
        ref[2,4,1,1]=(-0.44538816431)/(o+1.)+(-0.251644455066)
        ref[2,4,2,0]=(0.0465222158616)/(o+1.)+(0.85626655775)
        ref[2,4,2,1]=(0.319198566788)/(o+1.)+(0.882551371452)
        ref[3,0,0,0]=(-0.561665866984)/(o+1.)+(-0.623682533758)
        ref[3,0,0,1]=(-0.544665480685)/(o+1.)+(-0.326701245334)
        ref[3,0,1,0]=(-1.60194246541)/(o+1.)+(-0.452248099853)
        ref[3,0,1,1]=(-0.907946092739)/(o+1.)+(0.519323659165)
        ref[3,0,2,0]=(-0.454139734776)/(o+1.)+(0.0447289827682)
        ref[3,0,2,1]=(1.01507581477)/(o+1.)+(0.317404572995)
        ref[3,1,0,0]=(0.0211828103496)/(o+1.)+(0.624371810253)
        ref[3,1,0,1]=(-0.658289545845)/(o+1.)+(-0.513831705447)
        ref[3,1,1,0]=(1.06011975683)/(o+1.)+(0.721388170095)
        ref[3,1,1,1]=(-0.14633541181)/(o+1.)+(0.618616305516)
        ref[3,1,2,0]=(0.74864463613)/(o+1.)+(-0.511570767053)
        ref[3,1,2,1]=(0.488722756727)/(o+1.)+(0.0387361369774)
        ref[3,2,0,0]=(-0.180332702395)/(o+1.)+(-0.0815892290549)
        ref[3,2,0,1]=(0.0811207173227)/(o+1.)+(0.392285456186)
        ref[3,2,1,0]=(1.02225125645)/(o+1.)+(0.480941670608)
        ref[3,2,1,1]=(1.21612405627)/(o+1.)+(0.45160613516)
        ref[3,2,2,0]=(-0.355793756138)/(o+1.)+(0.669992488205)
        ref[3,2,2,1]=(0.776200363894)/(o+1.)+(0.535930158963)
        ref[3,3,0,0]=(-0.692703834029)/(o+1.)+(-0.00615860690903)
        ref[3,3,0,1]=(0.88835171354)/(o+1.)+(0.00763271775166)
        ref[3,3,1,0]=(-0.0987823234827)/(o+1.)+(0.628186276549)
        ref[3,3,1,1]=(-0.223347424968)/(o+1.)+(-0.00346757455392)
        ref[3,3,2,0]=(1.22648639393)/(o+1.)+(0.152216781743)
        ref[3,3,2,1]=(0.272100360294)/(o+1.)+(0.111532468623)
        ref[3,4,0,0]=(1.45235034011)/(o+1.)+(-0.0206817576441)
        ref[3,4,0,1]=(1.49583359131)/(o+1.)+(-0.574565590956)
        ref[3,4,1,0]=(0.362882204744)/(o+1.)+(0.581681025235)
        ref[3,4,1,1]=(1.31795519959)/(o+1.)+(-0.0659375558974)
        ref[3,4,2,0]=(0.556130663233)/(o+1.)+(-0.078813607523)
        ref[3,4,2,1]=(1.56413690967)/(o+1.)+(-0.42520624231)
      else:
        arg[0,0,0,0]=(0.956854727968)*x[0]**o+(-0.43263548815)*x[0]+(-0.1485543705)*x[1]**o+(-0.411060684134)*x[1]+(-0.306829513954)*x[2]**o+(0.89077924839)*x[2]
        arg[0,0,0,1]=(0.125905642347)*x[0]**o+(-0.645491335766)*x[0]+(-0.237571989123)*x[1]**o+(-0.239301942511)*x[1]+(-0.384215893518)*x[2]**o+(-0.371218179274)*x[2]
        arg[0,0,1,0]=(0.0204273419969)*x[0]**o+(0.354631489501)*x[0]+(0.235086936835)*x[1]**o+(0.647551574475)*x[1]+(-0.729066874572)*x[2]**o+(0.177733791046)*x[2]
        arg[0,0,1,1]=(-0.0938492166558)*x[0]**o+(-0.484339074939)*x[0]+(0.619333193491)*x[1]**o+(0.809157671781)*x[1]+(-0.28367956249)*x[2]**o+(-0.1758249267)*x[2]
        arg[0,0,2,0]=(-0.334944259519)*x[0]**o+(-0.126145567301)*x[0]+(-0.646972616146)*x[1]**o+(0.20287553181)*x[1]+(-0.383816067326)*x[2]**o+(-0.0265591031498)*x[2]
        arg[0,0,2,1]=(0.103308564583)*x[0]**o+(0.016108181951)*x[0]+(-0.367450237136)*x[1]**o+(0.657618788015)*x[1]+(-0.8327532511)*x[2]**o+(-0.267286876849)*x[2]
        arg[0,1,0,0]=(0.688995722152)*x[0]**o+(-0.982012541087)*x[0]+(-0.232339851099)*x[1]**o+(-0.097863461266)*x[1]+(0.25916639806)*x[2]**o+(0.0538637018789)*x[2]
        arg[0,1,0,1]=(0.446698337789)*x[0]**o+(-0.309434573629)*x[0]+(0.617004214016)*x[1]**o+(0.747666594435)*x[1]+(-0.869684246717)*x[2]**o+(0.75016186442)*x[2]
        arg[0,1,1,0]=(-0.559974353252)*x[0]**o+(0.966642794905)*x[0]+(-0.228473000256)*x[1]**o+(-0.435093965632)*x[1]+(0.426986383284)*x[2]**o+(0.131300817321)*x[2]
        arg[0,1,1,1]=(-0.696027598385)*x[0]**o+(-0.804778535506)*x[0]+(0.930171049019)*x[1]**o+(-0.900826020873)*x[1]+(0.844689166523)*x[2]**o+(0.0349718139676)*x[2]
        arg[0,1,2,0]=(0.359170897328)*x[0]**o+(0.992944582308)*x[0]+(0.579162388074)*x[1]**o+(0.743410851741)*x[1]+(-0.204325306226)*x[2]**o+(-0.992005484878)*x[2]
        arg[0,1,2,1]=(-0.883554988936)*x[0]**o+(0.486947073334)*x[0]+(-0.78078860326)*x[1]**o+(-0.134196299284)*x[1]+(-0.980852067031)*x[2]**o+(0.453023187355)*x[2]
        arg[0,2,0,0]=(0.804524524371)*x[0]**o+(0.397924403155)*x[0]+(0.712297039229)*x[1]**o+(0.713485622558)*x[1]+(-0.966783610543)*x[2]**o+(0.497869404576)*x[2]
        arg[0,2,0,1]=(-0.564706930001)*x[0]**o+(-0.482026983244)*x[0]+(0.284926154564)*x[1]**o+(0.844423322149)*x[1]+(0.444203847061)*x[2]**o+(0.333244188639)*x[2]
        arg[0,2,1,0]=(-0.630410083438)*x[0]**o+(-0.831218126629)*x[0]+(0.400237927956)*x[1]**o+(-0.70518378137)*x[1]+(-0.803754805917)*x[2]**o+(0.60059455602)*x[2]
        arg[0,2,1,1]=(0.198272547853)*x[0]**o+(0.0840356091496)*x[0]+(0.114385146155)*x[1]**o+(0.731057038288)*x[1]+(-0.976056408523)*x[2]**o+(-0.132565954304)*x[2]
        arg[0,2,2,0]=(0.983300141657)*x[0]**o+(-0.733168195162)*x[0]+(-0.948802450335)*x[1]**o+(-0.756616078178)*x[1]+(0.450302374372)*x[2]**o+(0.802115535717)*x[2]
        arg[0,2,2,1]=(-0.680853503774)*x[0]**o+(-0.910402862928)*x[0]+(0.583775613868)*x[1]**o+(0.741516645611)*x[1]+(-0.550498467533)*x[2]**o+(-0.610648481959)*x[2]
        arg[0,3,0,0]=(0.981909313325)*x[0]**o+(0.148145112359)*x[0]+(-0.989721117518)*x[1]**o+(0.0246558730659)*x[1]+(0.721439979982)*x[2]**o+(0.896052837889)*x[2]
        arg[0,3,0,1]=(0.109831691542)*x[0]**o+(-0.349555366958)*x[0]+(-0.54822642636)*x[1]**o+(0.121919022172)*x[1]+(-0.580228247481)*x[2]**o+(0.564276455713)*x[2]
        arg[0,3,1,0]=(0.106091093353)*x[0]**o+(0.410111742327)*x[0]+(-0.0669026478824)*x[1]**o+(-0.662684907574)*x[1]+(-0.351815520918)*x[2]**o+(-0.324802958038)*x[2]
        arg[0,3,1,1]=(0.857685803258)*x[0]**o+(-0.782627954043)*x[0]+(-0.555036389439)*x[1]**o+(-0.0126716685608)*x[1]+(0.260461379777)*x[2]**o+(0.117590260591)*x[2]
        arg[0,3,2,0]=(-0.987873674604)*x[0]**o+(0.0779662295512)*x[0]+(-0.928229202803)*x[1]**o+(0.508015712707)*x[1]+(0.248638760509)*x[2]**o+(-0.712158091156)*x[2]
        arg[0,3,2,1]=(-0.977973356138)*x[0]**o+(0.270015563339)*x[0]+(0.811157094127)*x[1]**o+(-0.756460384936)*x[1]+(-0.735557557504)*x[2]**o+(0.780333584929)*x[2]
        arg[0,4,0,0]=(-0.762444411646)*x[0]**o+(0.386894014691)*x[0]+(0.763898738855)*x[1]**o+(-0.422896581278)*x[1]+(0.428591445563)*x[2]**o+(-0.617626926396)*x[2]
        arg[0,4,0,1]=(-0.269524167286)*x[0]**o+(0.996895823788)*x[0]+(0.771396605337)*x[1]**o+(-0.993794940262)*x[1]+(-0.534803587481)*x[2]**o+(-0.928627946122)*x[2]
        arg[0,4,1,0]=(0.978497459584)*x[0]**o+(-0.343496507186)*x[0]+(-0.250581087884)*x[1]**o+(-0.855743445852)*x[1]+(-0.127512246124)*x[2]**o+(-0.133936486447)*x[2]
        arg[0,4,1,1]=(-0.844450818471)*x[0]**o+(0.263329730105)*x[0]+(0.411212783735)*x[1]**o+(-0.816379748952)*x[1]+(-0.443806901857)*x[2]**o+(0.1262969097)*x[2]
        arg[0,4,2,0]=(-0.326938381124)*x[0]**o+(-0.707054728551)*x[0]+(-0.725513874889)*x[1]**o+(0.655254977879)*x[1]+(0.772734170665)*x[2]**o+(-0.963856465024)*x[2]
        arg[0,4,2,1]=(0.896415185133)*x[0]**o+(-0.722431478737)*x[0]+(0.942058336843)*x[1]**o+(0.838230971246)*x[1]+(-0.279821311439)*x[2]**o+(-0.330130961206)*x[2]
        arg[1,0,0,0]=(-0.466537465363)*x[0]**o+(0.86493624537)*x[0]+(-0.943049537777)*x[1]**o+(0.658727518981)*x[1]+(-0.130355508544)*x[2]**o+(-0.158245981876)*x[2]
        arg[1,0,0,1]=(0.247046604605)*x[0]**o+(-0.287683750248)*x[0]+(0.266630203142)*x[1]**o+(0.372409648579)*x[1]+(-0.239411725166)*x[2]**o+(0.537339451105)*x[2]
        arg[1,0,1,0]=(-0.232312387253)*x[0]**o+(-0.0493954012747)*x[0]+(-0.461085479525)*x[1]**o+(-0.0744729096417)*x[1]+(-0.795163064273)*x[2]**o+(0.613075576806)*x[2]
        arg[1,0,1,1]=(-0.68570821748)*x[0]**o+(-0.492507082229)*x[0]+(-0.0673636097828)*x[1]**o+(0.569119992806)*x[1]+(-0.227653595613)*x[2]**o+(0.0337433636837)*x[2]
        arg[1,0,2,0]=(-0.124371489405)*x[0]**o+(-0.726635218442)*x[0]+(-0.835418167414)*x[1]**o+(-0.00414597707753)*x[1]+(0.8365917104)*x[2]**o+(-0.507314639615)*x[2]
        arg[1,0,2,1]=(-0.973228057052)*x[0]**o+(0.123228228923)*x[0]+(0.545712842666)*x[1]**o+(-0.451331566047)*x[1]+(0.299501907067)*x[2]**o+(-0.24898389883)*x[2]
        arg[1,1,0,0]=(-0.940690906836)*x[0]**o+(-0.357571197214)*x[0]+(0.98643983012)*x[1]**o+(-0.82113140872)*x[1]+(0.460009285424)*x[2]**o+(-0.0298714172583)*x[2]
        arg[1,1,0,1]=(-0.881543245921)*x[0]**o+(-0.640741814727)*x[0]+(-0.749244677096)*x[1]**o+(0.47161101909)*x[1]+(0.712893375828)*x[2]**o+(0.600151133959)*x[2]
        arg[1,1,1,0]=(-0.217214692405)*x[0]**o+(-0.504211076658)*x[0]+(0.857127736087)*x[1]**o+(0.0439761116172)*x[1]+(0.956380848189)*x[2]**o+(0.698482155013)*x[2]
        arg[1,1,1,1]=(-0.288969728734)*x[0]**o+(-0.63960886193)*x[0]+(-0.741293494258)*x[1]**o+(-0.125034323955)*x[1]+(-0.608855286755)*x[2]**o+(0.462427737944)*x[2]
        arg[1,1,2,0]=(0.126242043984)*x[0]**o+(0.811846484451)*x[0]+(-0.522920862127)*x[1]**o+(0.678142081935)*x[1]+(0.0392586436898)*x[2]**o+(0.403908304728)*x[2]
        arg[1,1,2,1]=(-0.290103861843)*x[0]**o+(0.336130241928)*x[0]+(-0.872641307678)*x[1]**o+(-0.372020796526)*x[1]+(0.172572077906)*x[2]**o+(-0.789431556966)*x[2]
        arg[1,2,0,0]=(-0.158114079913)*x[0]**o+(0.954919898537)*x[0]+(-0.711834968234)*x[1]**o+(-0.390487270434)*x[1]+(-0.086581514295)*x[2]**o+(0.442047084988)*x[2]
        arg[1,2,0,1]=(-0.914034594356)*x[0]**o+(0.521137372279)*x[0]+(0.12454416569)*x[1]**o+(0.627815188423)*x[1]+(0.770244577882)*x[2]**o+(-0.846691748091)*x[2]
        arg[1,2,1,0]=(0.769631824465)*x[0]**o+(0.988840796717)*x[0]+(-0.401836868027)*x[1]**o+(0.401846957832)*x[1]+(-0.877826997285)*x[2]**o+(0.857725824229)*x[2]
        arg[1,2,1,1]=(0.173539365276)*x[0]**o+(0.373206670706)*x[0]+(-0.597288730366)*x[1]**o+(-0.770220794555)*x[1]+(-0.586510454467)*x[2]**o+(-0.569370842441)*x[2]
        arg[1,2,2,0]=(0.673765155818)*x[0]**o+(-0.642914516709)*x[0]+(0.794438851788)*x[1]**o+(-0.17557754186)*x[1]+(0.997565143863)*x[2]**o+(0.722460144483)*x[2]
        arg[1,2,2,1]=(-0.878452482718)*x[0]**o+(-0.846862156166)*x[0]+(0.462351027504)*x[1]**o+(0.267435606154)*x[1]+(-0.452135086983)*x[2]**o+(0.701577407252)*x[2]
        arg[1,3,0,0]=(-0.0557389135964)*x[0]**o+(-0.781831460089)*x[0]+(0.214846092931)*x[1]**o+(0.597436260353)*x[1]+(0.713968428327)*x[2]**o+(-0.456335676572)*x[2]
        arg[1,3,0,1]=(0.288266531585)*x[0]**o+(-0.259529556521)*x[0]+(0.0286107975184)*x[1]**o+(0.465095455484)*x[1]+(-0.747648909025)*x[2]**o+(0.682866470382)*x[2]
        arg[1,3,1,0]=(-0.720339104668)*x[0]**o+(0.961100526257)*x[0]+(0.0814885381756)*x[1]**o+(-0.448775223649)*x[1]+(0.250972143198)*x[2]**o+(-0.240231439947)*x[2]
        arg[1,3,1,1]=(-0.459305786748)*x[0]**o+(0.272222143865)*x[0]+(-0.499164099401)*x[1]**o+(-0.225162547271)*x[1]+(-0.92261124167)*x[2]**o+(-0.436659790026)*x[2]
        arg[1,3,2,0]=(0.890752966329)*x[0]**o+(-0.775207922011)*x[0]+(0.715960284721)*x[1]**o+(0.843422841789)*x[1]+(0.513788850466)*x[2]**o+(-0.325733710281)*x[2]
        arg[1,3,2,1]=(0.958726172433)*x[0]**o+(-0.743672393249)*x[0]+(-0.523783133945)*x[1]**o+(0.133861976051)*x[1]+(-0.109230940116)*x[2]**o+(0.819716327751)*x[2]
        arg[1,4,0,0]=(0.54439729551)*x[0]**o+(0.0763503792087)*x[0]+(0.0634218657498)*x[1]**o+(-0.405471845044)*x[1]+(0.316327091756)*x[2]**o+(-0.539493863939)*x[2]
        arg[1,4,0,1]=(0.527330373182)*x[0]**o+(0.77531447123)*x[0]+(-0.481869315442)*x[1]**o+(0.556927692362)*x[1]+(-0.267215098309)*x[2]**o+(0.958919406465)*x[2]
        arg[1,4,1,0]=(0.540950372012)*x[0]**o+(-0.95002626611)*x[0]+(-0.523544509893)*x[1]**o+(-0.583056635057)*x[1]+(-0.00288409928193)*x[2]**o+(0.671385098427)*x[2]
        arg[1,4,1,1]=(-0.39802076092)*x[0]**o+(-0.583264022146)*x[0]+(-0.656261495961)*x[1]**o+(-0.527677078418)*x[1]+(-0.2632498932)*x[2]**o+(-0.537632951507)*x[2]
        arg[1,4,2,0]=(-0.77176416566)*x[0]**o+(0.812696440735)*x[0]+(0.0925789242005)*x[1]**o+(0.299698967601)*x[1]+(-0.0185377066587)*x[2]**o+(-0.883247318682)*x[2]
        arg[1,4,2,1]=(0.360949407452)*x[0]**o+(-0.463659754812)*x[0]+(0.799066387132)*x[1]**o+(0.0318151346916)*x[1]+(-0.380675976437)*x[2]**o+(-0.299341342112)*x[2]
        arg[2,0,0,0]=(-0.985332287141)*x[0]**o+(0.33398206098)*x[0]+(0.39204502373)*x[1]**o+(-0.586826360517)*x[1]+(0.868240161369)*x[2]**o+(-0.257897525907)*x[2]
        arg[2,0,0,1]=(-0.801369266967)*x[0]**o+(0.333496542809)*x[0]+(0.459444994717)*x[1]**o+(-0.950297568688)*x[1]+(-0.760722281765)*x[2]**o+(0.0933850221377)*x[2]
        arg[2,0,1,0]=(0.53372811812)*x[0]**o+(0.999551098511)*x[0]+(-0.761650890163)*x[1]**o+(-0.80456940786)*x[1]+(-0.452089806623)*x[2]**o+(-0.40841169158)*x[2]
        arg[2,0,1,1]=(-0.782294374468)*x[0]**o+(0.614736817827)*x[0]+(0.262995910218)*x[1]**o+(0.067349886804)*x[1]+(-0.167494656909)*x[2]**o+(-0.730462378769)*x[2]
        arg[2,0,2,0]=(0.261001857862)*x[0]**o+(0.959318765248)*x[0]+(-0.177224016298)*x[1]**o+(-0.979637664014)*x[1]+(0.858937930292)*x[2]**o+(-0.627116377798)*x[2]
        arg[2,0,2,1]=(-0.238083578675)*x[0]**o+(0.586367991705)*x[0]+(0.747378561289)*x[1]**o+(-0.540178926415)*x[1]+(-0.519781240185)*x[2]**o+(-0.42715014981)*x[2]
        arg[2,1,0,0]=(-0.21022574437)*x[0]**o+(-0.292359099348)*x[0]+(0.449614855785)*x[1]**o+(-0.416904619153)*x[1]+(0.507537288404)*x[2]**o+(0.943584716996)*x[2]
        arg[2,1,0,1]=(-0.410591525695)*x[0]**o+(-0.258796394127)*x[0]+(-0.474714731399)*x[1]**o+(-0.518288451462)*x[1]+(-0.15640903664)*x[2]**o+(0.97242574307)*x[2]
        arg[2,1,1,0]=(-0.930865834892)*x[0]**o+(0.961887630616)*x[0]+(0.688706862295)*x[1]**o+(-0.375163634404)*x[1]+(0.378954001649)*x[2]**o+(-0.747140890466)*x[2]
        arg[2,1,1,1]=(0.0454701384411)*x[0]**o+(-0.37651801332)*x[0]+(0.0808750051583)*x[1]**o+(-0.530355850075)*x[1]+(0.769675417234)*x[2]**o+(-0.767046792016)*x[2]
        arg[2,1,2,0]=(0.179573418696)*x[0]**o+(-0.0547918709025)*x[0]+(0.739792606043)*x[1]**o+(0.849619562235)*x[1]+(-0.92840796838)*x[2]**o+(-0.302941296337)*x[2]
        arg[2,1,2,1]=(0.160657807137)*x[0]**o+(0.551943403654)*x[0]+(0.649164334204)*x[1]**o+(0.298882266215)*x[1]+(0.33872360276)*x[2]**o+(-0.558910868754)*x[2]
        arg[2,2,0,0]=(-0.704330511528)*x[0]**o+(-0.326943167116)*x[0]+(0.626863819473)*x[1]**o+(-0.107147718189)*x[1]+(0.00668670393016)*x[2]**o+(0.572223662188)*x[2]
        arg[2,2,0,1]=(0.879695137799)*x[0]**o+(-0.313488700497)*x[0]+(-0.113134766565)*x[1]**o+(-0.131007759381)*x[1]+(0.55460969593)*x[2]**o+(0.898706374575)*x[2]
        arg[2,2,1,0]=(0.329932124981)*x[0]**o+(-0.624965470976)*x[0]+(0.82906774364)*x[1]**o+(-0.681403786977)*x[1]+(-0.447119491747)*x[2]**o+(0.143763601614)*x[2]
        arg[2,2,1,1]=(0.0367746207528)*x[0]**o+(0.47553259605)*x[0]+(-0.517723045973)*x[1]**o+(0.263932783195)*x[1]+(-0.131032589284)*x[2]**o+(-0.132729291841)*x[2]
        arg[2,2,2,0]=(-0.108349266564)*x[0]**o+(-0.684748849483)*x[0]+(0.203271582231)*x[1]**o+(-0.860842085896)*x[1]+(0.469684577785)*x[2]**o+(0.195736308182)*x[2]
        arg[2,2,2,1]=(0.915666799349)*x[0]**o+(0.383338953756)*x[0]+(-0.554011249579)*x[1]**o+(0.232285573367)*x[1]+(0.39288151674)*x[2]**o+(0.692592233016)*x[2]
        arg[2,3,0,0]=(0.81179207663)*x[0]**o+(0.786809236572)*x[0]+(-0.814758445611)*x[1]**o+(-0.78461653396)*x[1]+(0.171435802778)*x[2]**o+(0.660186909193)*x[2]
        arg[2,3,0,1]=(0.752694523669)*x[0]**o+(0.301473785296)*x[0]+(0.610379955695)*x[1]**o+(0.573729399495)*x[1]+(0.225983751859)*x[2]**o+(-0.349530143446)*x[2]
        arg[2,3,1,0]=(-0.647825490424)*x[0]**o+(-0.232573858642)*x[0]+(0.773519085771)*x[1]**o+(-0.224497979868)*x[1]+(-0.681726885952)*x[2]**o+(-0.217968759388)*x[2]
        arg[2,3,1,1]=(0.391888417642)*x[0]**o+(-0.240556139087)*x[0]+(-0.852397797723)*x[1]**o+(0.318342758913)*x[1]+(-0.773804983598)*x[2]**o+(-0.957590491908)*x[2]
        arg[2,3,2,0]=(0.7126582007)*x[0]**o+(0.806022164546)*x[0]+(0.232348951782)*x[1]**o+(-0.985774366344)*x[1]+(-0.365161907541)*x[2]**o+(0.0218977623299)*x[2]
        arg[2,3,2,1]=(-0.775068843872)*x[0]**o+(-0.595291330879)*x[0]+(0.681122481841)*x[1]**o+(0.152514252732)*x[1]+(0.510976191808)*x[2]**o+(-0.69539067488)*x[2]
        arg[2,4,0,0]=(-0.362761749405)*x[0]**o+(0.494347958562)*x[0]+(0.215244694522)*x[1]**o+(-0.527767969217)*x[1]+(-0.115976071059)*x[2]**o+(-0.998233415841)*x[2]
        arg[2,4,0,1]=(0.620291015098)*x[0]**o+(-0.222644358028)*x[0]+(0.0377714785847)*x[1]**o+(0.663978623913)*x[1]+(0.996481523389)*x[2]**o+(-0.689519418095)*x[2]
        arg[2,4,1,0]=(-0.0374746407394)*x[0]**o+(0.3186355021)*x[0]+(-0.704305640533)*x[1]**o+(-0.632264150766)*x[1]+(-0.166691199609)*x[2]**o+(-0.794128441334)*x[2]
        arg[2,4,1,1]=(0.117806334919)*x[0]**o+(-0.717334572919)*x[0]+(0.665341054571)*x[1]**o+(0.354208273372)*x[1]+(0.349883631586)*x[2]**o+(-0.260626035314)*x[2]
        arg[2,4,2,0]=(0.160337827685)*x[0]**o+(-0.92734430881)*x[0]+(0.872112181363)*x[1]**o+(-0.796372970466)*x[1]+(-0.776882702685)*x[2]**o+(0.246816968474)*x[2]
        arg[2,4,2,1]=(-0.317926812571)*x[0]**o+(-0.795779593996)*x[0]+(0.785287724793)*x[1]**o+(0.839033515634)*x[1]+(0.433187943714)*x[2]**o+(0.308175885606)*x[2]
        arg[3,0,0,0]=(-0.555193292291)*x[0]**o+(-0.0280452948283)*x[0]+(0.988455127227)*x[1]**o+(-0.637707060271)*x[1]+(-0.95062571916)*x[2]**o+(-0.563121211553)*x[2]
        arg[3,0,0,1]=(-0.169558534946)*x[0]**o+(0.886096637835)*x[0]+(-0.666307491933)*x[1]**o+(0.650321801985)*x[1]+(-0.268934932123)*x[2]**o+(-0.406617199906)*x[2]
        arg[3,0,1,0]=(0.497351781449)*x[0]**o+(0.768617651685)*x[0]+(0.200962956226)*x[1]**o+(0.694236078954)*x[1]+(0.0323702864321)*x[2]**o+(-0.666046265832)*x[2]
        arg[3,0,1,1]=(0.425151807387)*x[0]**o+(-0.886432438604)*x[0]+(-0.963257203569)*x[1]**o+(-0.698524427437)*x[1]+(0.0165810713355)*x[2]**o+(-0.142193988951)*x[2]
        arg[3,0,2,0]=(0.965746112788)*x[0]**o+(-0.333068915886)*x[0]+(0.331031372002)*x[1]**o+(-0.384004796757)*x[1]+(0.0144121092048)*x[2]**o+(-0.352607264387)*x[2]
        arg[3,0,2,1]=(-0.391557775789)*x[0]**o+(0.127020330586)*x[0]+(0.758832450081)*x[1]**o+(0.49669414469)*x[1]+(0.358403498634)*x[2]**o+(0.184963946351)*x[2]
        arg[3,1,0,0]=(0.978053897464)*x[0]**o+(-0.211787313494)*x[0]+(0.927706482976)*x[1]**o+(-0.451292956408)*x[1]+(0.592824645174)*x[2]**o+(0.896240907546)*x[2]
        arg[3,1,0,1]=(0.0780568875295)*x[0]**o+(0.130098698523)*x[0]+(-0.281871571918)*x[1]**o+(-0.596362849045)*x[1]+(-0.436455789199)*x[2]**o+(-0.657145053021)*x[2]
        arg[3,1,1,0]=(-0.278111817329)*x[0]**o+(-0.627384158565)*x[0]+(-0.995802044048)*x[1]**o+(0.42485189159)*x[1]+(0.483473275475)*x[2]**o+(-0.177928974458)*x[2]
        arg[3,1,1,1]=(0.127721047557)*x[0]**o+(0.93147711978)*x[0]+(-0.350992052283)*x[1]**o+(0.0993773637672)*x[1]+(0.29111744653)*x[2]**o+(-0.821710278883)*x[2]
        arg[3,1,2,0]=(-0.697941918276)*x[0]**o+(-0.585872232852)*x[0]+(0.697967779925)*x[1]**o+(-0.811186381728)*x[1]+(0.708726108998)*x[2]**o+(0.289602027081)*x[2]
        arg[3,1,2,1]=(-0.461424409147)*x[0]**o+(0.335089386385)*x[0]+(-0.964838280406)*x[1]**o+(0.828627561583)*x[1]+(-0.861416281013)*x[2]**o+(0.898354675269)*x[2]
        arg[3,2,0,0]=(0.752543705358)*x[0]**o+(0.164487088496)*x[0]+(0.756164002406)*x[1]**o+(-0.740019514032)*x[1]+(0.107245489678)*x[2]**o+(0.677707700079)*x[2]
        arg[3,2,0,1]=(0.763792557134)*x[0]**o+(0.725507550672)*x[0]+(0.171323949629)*x[1]**o+(-0.287172075258)*x[1]+(-0.94865226715)*x[2]**o+(-0.163093465336)*x[2]
        arg[3,2,1,0]=(-0.327892721048)*x[0]**o+(0.334145897353)*x[0]+(-0.936827293436)*x[1]**o+(0.0822687872593)*x[1]+(0.259794829604)*x[2]**o+(0.376908280759)*x[2]
        arg[3,2,1,1]=(0.595012545955)*x[0]**o+(0.617490014693)*x[0]+(-0.658036388462)*x[1]**o+(-0.921055393527)*x[1]+(-0.257878085344)*x[2]**o+(-0.623113089161)*x[2]
        arg[3,2,2,0]=(0.961175705937)*x[0]**o+(0.656031660842)*x[0]+(0.964101580317)*x[1]**o+(0.824033157042)*x[1]+(0.510244886658)*x[2]**o+(-0.236501162729)*x[2]
        arg[3,2,2,1]=(-0.128405588923)*x[0]**o+(-0.0327141502894)*x[0]+(0.909941485131)*x[1]**o+(0.79947258121)*x[1]+(0.635883050534)*x[2]**o+(0.841997575796)*x[2]
        arg[3,3,0,0]=(-0.906234907643)*x[0]**o+(-0.316076976009)*x[0]+(0.863208664533)*x[1]**o+(-0.968640170523)*x[1]+(0.791637381686)*x[2]**o+(-0.543353462098)*x[2]
        arg[3,3,0,1]=(-0.586709036832)*x[0]**o+(-0.299493216592)*x[0]+(-0.173966725899)*x[1]**o+(0.0971838711677)*x[1]+(0.385662922244)*x[2]**o+(0.288932625368)*x[2]
        arg[3,3,1,0]=(-0.33865322795)*x[0]**o+(0.442176445777)*x[0]+(0.706183231282)*x[1]**o+(0.62528018966)*x[1]+(-0.377746006979)*x[2]**o+(-0.261455124139)*x[2]
        arg[3,3,1,1]=(0.834855553776)*x[0]**o+(0.755005824764)*x[0]+(0.446589332183)*x[1]**o+(0.41883521757)*x[1]+(0.820177207209)*x[2]**o+(0.603811552083)*x[2]
        arg[3,3,2,0]=(-0.624282266193)*x[0]**o+(0.28109086861)*x[0]+(0.44286011655)*x[1]**o+(-0.117876187343)*x[1]+(-0.267660131821)*x[2]**o+(0.880359406159)*x[2]
        arg[3,3,2,1]=(-0.0722687033317)*x[0]**o+(-0.7013147214)*x[0]+(0.00614303249005)*x[1]**o+(-0.658138946692)*x[1]+(0.0557277088337)*x[2]**o+(0.614362766103)*x[2]
        arg[3,4,0,0]=(0.935244440586)*x[0]**o+(-0.0699305155896)*x[0]+(-0.849303362918)*x[1]**o+(0.728062537708)*x[1]+(0.404621498028)*x[2]**o+(0.611486631437)*x[2]
        arg[3,4,0,1]=(0.168117782017)*x[0]**o+(-0.0454396423181)*x[0]+(0.46021270423)*x[1]**o+(0.597343569492)*x[1]+(0.561215321486)*x[2]**o+(0.715771227777)*x[2]
        arg[3,4,1,0]=(-0.502675682461)*x[0]**o+(0.543910335185)*x[0]+(0.721732643536)*x[1]**o+(-0.862126899566)*x[1]+(0.794704104653)*x[2]**o+(-0.284248276797)*x[2]
        arg[3,4,1,1]=(-0.580751817984)*x[0]**o+(0.482838505281)*x[0]+(-0.426066109367)*x[1]**o+(-0.456724382007)*x[1]+(-0.770655426938)*x[2]**o+(-0.226232224569)*x[2]
        arg[3,4,2,0]=(0.199832861603)*x[0]**o+(-0.586239366436)*x[0]+(0.730019631541)*x[1]**o+(0.185265476259)*x[1]+(0.514243351617)*x[2]**o+(-0.757742857839)*x[2]
        arg[3,4,2,1]=(-0.941843068195)*x[0]**o+(0.211732407161)*x[0]+(-0.99705602485)*x[1]**o+(-0.198037742645)*x[1]+(0.154058794178)*x[2]**o+(0.0675962671791)*x[2]
        ref[0,0,0,0]=(0.501470843513)/(o+1.)+(0.0235415380528)
        ref[0,0,0,1]=(-0.495882240293)/(o+1.)+(-0.628005728776)
        ref[0,0,1,0]=(-0.47355259574)/(o+1.)+(0.589958427511)
        ref[0,0,1,1]=(0.241804414345)/(o+1.)+(0.0744968350706)
        ref[0,0,2,0]=(-1.36573294299)/(o+1.)+(0.0250854306796)
        ref[0,0,2,1]=(-1.09689492365)/(o+1.)+(0.203220046558)
        ref[0,1,0,0]=(0.715822269113)/(o+1.)+(-0.513006150237)
        ref[0,1,0,1]=(0.194018305088)/(o+1.)+(0.594196942613)
        ref[0,1,1,0]=(-0.361460970224)/(o+1.)+(0.331424823297)
        ref[0,1,1,1]=(1.07883261716)/(o+1.)+(-0.835316371206)
        ref[0,1,2,0]=(0.734007979175)/(o+1.)+(0.372174974585)
        ref[0,1,2,1]=(-2.64519565923)/(o+1.)+(0.402886980703)
        ref[0,2,0,0]=(0.550037953057)/(o+1.)+(0.804639715144)
        ref[0,2,0,1]=(0.164423071623)/(o+1.)+(0.347820263772)
        ref[0,2,1,0]=(-1.0339269614)/(o+1.)+(-0.46790367599)
        ref[0,2,1,1]=(-0.663398714515)/(o+1.)+(0.341263346566)
        ref[0,2,2,0]=(0.484800065694)/(o+1.)+(-0.343834368812)
        ref[0,2,2,1]=(-0.647576357439)/(o+1.)+(-0.389767349638)
        ref[0,3,0,0]=(0.713628175789)/(o+1.)+(0.534426911657)
        ref[0,3,0,1]=(-1.0186229823)/(o+1.)+(0.168320055464)
        ref[0,3,1,0]=(-0.312627075447)/(o+1.)+(-0.288688061643)
        ref[0,3,1,1]=(0.563110793595)/(o+1.)+(-0.338854681006)
        ref[0,3,2,0]=(-1.6674641169)/(o+1.)+(-0.0630880744491)
        ref[0,3,2,1]=(-0.902373819515)/(o+1.)+(0.146944381666)
        ref[0,4,0,0]=(0.430045772773)/(o+1.)+(-0.326814746492)
        ref[0,4,0,1]=(-0.0329311494305)/(o+1.)+(-0.462763531298)
        ref[0,4,1,0]=(0.600404125575)/(o+1.)+(-0.666588219743)
        ref[0,4,1,1]=(-0.877044936593)/(o+1.)+(-0.213376554573)
        ref[0,4,2,0]=(-0.279718085349)/(o+1.)+(-0.507828107848)
        ref[0,4,2,1]=(1.55865221054)/(o+1.)+(-0.107165734349)
        ref[1,0,0,0]=(-1.53994251168)/(o+1.)+(0.682708891237)
        ref[1,0,0,1]=(0.27426508258)/(o+1.)+(0.311032674718)
        ref[1,0,1,0]=(-1.48856093105)/(o+1.)+(0.244603632945)
        ref[1,0,1,1]=(-0.980725422876)/(o+1.)+(0.0551781371305)
        ref[1,0,2,0]=(-0.123197946419)/(o+1.)+(-0.619047917567)
        ref[1,0,2,1]=(-0.128013307318)/(o+1.)+(-0.288543617977)
        ref[1,1,0,0]=(0.505758208708)/(o+1.)+(-0.604287011596)
        ref[1,1,0,1]=(-0.917894547188)/(o+1.)+(0.215510169161)
        ref[1,1,1,0]=(1.59629389187)/(o+1.)+(0.119123594986)
        ref[1,1,1,1]=(-1.63911850975)/(o+1.)+(-0.15110772397)
        ref[1,1,2,0]=(-0.357420174452)/(o+1.)+(0.946948435557)
        ref[1,1,2,1]=(-0.990173091615)/(o+1.)+(-0.412661055782)
        ref[1,2,0,0]=(-0.956530562442)/(o+1.)+(0.503239856545)
        ref[1,2,0,1]=(-0.0192458507848)/(o+1.)+(0.151130406306)
        ref[1,2,1,0]=(-0.510032040848)/(o+1.)+(1.12420678939)
        ref[1,2,1,1]=(-1.01025981956)/(o+1.)+(-0.483192483145)
        ref[1,2,2,0]=(2.46576915147)/(o+1.)+(-0.0480159570431)
        ref[1,2,2,1]=(-0.868236542197)/(o+1.)+(0.0610754286198)
        ref[1,3,0,0]=(0.873075607662)/(o+1.)+(-0.320365438154)
        ref[1,3,0,1]=(-0.430771579922)/(o+1.)+(0.444216184672)
        ref[1,3,1,0]=(-0.387878423294)/(o+1.)+(0.13604693133)
        ref[1,3,1,1]=(-1.88108112782)/(o+1.)+(-0.194800096716)
        ref[1,3,2,0]=(2.12050210152)/(o+1.)+(-0.128759395251)
        ref[1,3,2,1]=(0.325712098372)/(o+1.)+(0.104952955276)
        ref[1,4,0,0]=(0.924146253016)/(o+1.)+(-0.434307664887)
        ref[1,4,0,1]=(-0.221754040568)/(o+1.)+(1.14558078503)
        ref[1,4,1,0]=(0.0145217628367)/(o+1.)+(-0.43084890137)
        ref[1,4,1,1]=(-1.31753215008)/(o+1.)+(-0.824287026036)
        ref[1,4,2,0]=(-0.697722948118)/(o+1.)+(0.114574044827)
        ref[1,4,2,1]=(0.779339818147)/(o+1.)+(-0.365592981116)
        ref[2,0,0,0]=(0.274952897958)/(o+1.)+(-0.255370912722)
        ref[2,0,0,1]=(-1.10264655401)/(o+1.)+(-0.261708001871)
        ref[2,0,1,0]=(-0.680012578665)/(o+1.)+(-0.106715000464)
        ref[2,0,1,1]=(-0.686793121159)/(o+1.)+(-0.0241878370688)
        ref[2,0,2,0]=(0.942715771856)/(o+1.)+(-0.323717638282)
        ref[2,0,2,1]=(-0.0104862575713)/(o+1.)+(-0.19048054226)
        ref[2,1,0,0]=(0.746926399818)/(o+1.)+(0.117160499248)
        ref[2,1,0,1]=(-1.04171529373)/(o+1.)+(0.0976704487401)
        ref[2,1,1,0]=(0.136795029052)/(o+1.)+(-0.0802084471266)
        ref[2,1,1,1]=(0.896020560833)/(o+1.)+(-0.836960327706)
        ref[2,1,2,0]=(-0.00904194364113)/(o+1.)+(0.245943197498)
        ref[2,1,2,1]=(1.1485457441)/(o+1.)+(0.145957400558)
        ref[2,2,0,0]=(-0.0707799881243)/(o+1.)+(0.0690663884416)
        ref[2,2,0,1]=(1.32117006716)/(o+1.)+(0.227104957348)
        ref[2,2,1,0]=(0.711880376874)/(o+1.)+(-0.58130282817)
        ref[2,2,1,1]=(-0.611981014504)/(o+1.)+(0.303368043702)
        ref[2,2,2,0]=(0.564606893452)/(o+1.)+(-0.674927313599)
        ref[2,2,2,1]=(0.754537066511)/(o+1.)+(0.654108380069)
        ref[2,3,0,0]=(0.168469433796)/(o+1.)+(0.331189805903)
        ref[2,3,0,1]=(1.58905823122)/(o+1.)+(0.262836520673)
        ref[2,3,1,0]=(-0.556033290604)/(o+1.)+(-0.337520298949)
        ref[2,3,1,1]=(-1.23431436368)/(o+1.)+(-0.439901936041)
        ref[2,3,2,0]=(0.579845244941)/(o+1.)+(-0.0789272197338)
        ref[2,3,2,1]=(0.417029829778)/(o+1.)+(-0.569083876513)
        ref[2,4,0,0]=(-0.263493125943)/(o+1.)+(-0.515826713248)
        ref[2,4,0,1]=(1.65454401707)/(o+1.)+(-0.124092576105)
        ref[2,4,1,0]=(-0.908471480882)/(o+1.)+(-0.553878545)
        ref[2,4,1,1]=(1.13303102108)/(o+1.)+(-0.31187616743)
        ref[2,4,2,0]=(0.255567306363)/(o+1.)+(-0.738450155401)
        ref[2,4,2,1]=(0.900548855936)/(o+1.)+(0.175714903622)
        ref[3,0,0,0]=(-0.517363884224)/(o+1.)+(-0.614436783326)
        ref[3,0,0,1]=(-1.104800959)/(o+1.)+(0.564900619957)
        ref[3,0,1,0]=(0.730685024108)/(o+1.)+(0.398403732404)
        ref[3,0,1,1]=(-0.521524324846)/(o+1.)+(-0.863575427496)
        ref[3,0,2,0]=(1.31118959399)/(o+1.)+(-0.534840488515)
        ref[3,0,2,1]=(0.725678172925)/(o+1.)+(0.404339210814)
        ref[3,1,0,0]=(2.49858502561)/(o+1.)+(0.116580318822)
        ref[3,1,0,1]=(-0.640270473588)/(o+1.)+(-0.561704601771)
        ref[3,1,1,0]=(-0.790440585902)/(o+1.)+(-0.190230620716)
        ref[3,1,1,1]=(0.0678464418036)/(o+1.)+(0.104572102332)
        ref[3,1,2,0]=(0.708751970646)/(o+1.)+(-0.553728293749)
        ref[3,1,2,1]=(-2.28767897057)/(o+1.)+(1.03103581162)
        ref[3,2,0,0]=(1.61595319744)/(o+1.)+(0.0510876372718)
        ref[3,2,0,1]=(-0.0135357603866)/(o+1.)+(0.137621005039)
        ref[3,2,1,0]=(-1.00492518488)/(o+1.)+(0.396661482686)
        ref[3,2,1,1]=(-0.320901927851)/(o+1.)+(-0.463339233997)
        ref[3,2,2,0]=(2.43552217291)/(o+1.)+(0.621781827577)
        ref[3,2,2,1]=(1.41741894674)/(o+1.)+(0.804378003359)
        ref[3,3,0,0]=(0.748611138576)/(o+1.)+(-0.914035304315)
        ref[3,3,0,1]=(-0.375012840487)/(o+1.)+(0.0433116399723)
        ref[3,3,1,0]=(-0.0102160036467)/(o+1.)+(0.403000755649)
        ref[3,3,1,1]=(2.10162209317)/(o+1.)+(0.888826297208)
        ref[3,3,2,0]=(-0.449082281464)/(o+1.)+(0.521787043713)
        ref[3,3,2,1]=(-0.0103979620079)/(o+1.)+(-0.372545450994)
        ref[3,4,0,0]=(0.490562575696)/(o+1.)+(0.634809326777)
        ref[3,4,0,1]=(1.18954580773)/(o+1.)+(0.633837577475)
        ref[3,4,1,0]=(1.01376106573)/(o+1.)+(-0.301232420589)
        ref[3,4,1,1]=(-1.77747335429)/(o+1.)+(-0.100059050648)
        ref[3,4,2,0]=(1.44409584476)/(o+1.)+(-0.579358374008)
        ref[3,4,2,1]=(-1.78484029887)/(o+1.)+(0.0406454658476)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref), "wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.788336609161)*x[0]**o+(-0.46539601764)*x[0]+(0.0970136798824)*x[1]**o+(-0.120665298671)*x[1]
        ref=(-0.691322929278)/(o+1.)+(-0.293030658156)
      else:
        arg=(0.922255261003)*x[0]**o+(0.0862232364707)*x[0]+(-0.94152635039)*x[1]**o+(0.771078486789)*x[1]+(-0.487642191181)*x[2]**o+(0.971754960362)*x[2]
        ref=(-0.506913280568)/(o+1.)+(0.914528341811)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref), "wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.193071015033)*x[0]**o+(-0.281957941612)*x[0]+(0.269334552863)*x[1]**o+(0.0332262192804)*x[1]
        arg[1]=(-0.982634724425)*x[0]**o+(0.872638820737)*x[0]+(0.828028964603)*x[1]**o+(0.70464418136)*x[1]
        ref[0]=(0.0762635378296)/(o+1.)+(-0.124365861166)
        ref[1]=(-0.154605759822)/(o+1.)+(0.788641501048)
      else:
        arg[0]=(0.0746983082972)*x[0]**o+(0.281818436343)*x[0]+(-0.835321029206)*x[1]**o+(-0.349524270524)*x[1]+(-0.471905088601)*x[2]**o+(0.156327167402)*x[2]
        arg[1]=(-0.971935771255)*x[0]**o+(0.943151542057)*x[0]+(-0.60084258169)*x[1]**o+(0.458392855527)*x[1]+(-0.0173384187065)*x[2]**o+(0.538410338758)*x[2]
        ref[0]=(-1.23252780951)/(o+1.)+(0.0443106666103)
        ref[1]=(-1.59011677165)/(o+1.)+(0.969977368171)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref), "wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.568783264759)*x[0]**o+(0.690377280868)*x[0]+(-0.689750922806)*x[1]**o+(-0.262828324126)*x[1]
        arg[0,1]=(-0.981657284673)*x[0]**o+(0.37118874852)*x[0]+(0.733992104146)*x[1]**o+(0.350268735699)*x[1]
        arg[0,2]=(0.713336100765)*x[0]**o+(0.737674447814)*x[0]+(0.239606043833)*x[1]**o+(0.814911366356)*x[1]
        arg[0,3]=(0.171706130967)*x[0]**o+(0.56324398491)*x[0]+(0.328884588679)*x[1]**o+(0.13307521808)*x[1]
        arg[0,4]=(-0.309130885049)*x[0]**o+(0.353596646674)*x[0]+(-0.282764686556)*x[1]**o+(-0.56977508596)*x[1]
        arg[1,0]=(0.974136161109)*x[0]**o+(-0.432477120938)*x[0]+(-0.841212164751)*x[1]**o+(-0.429211937256)*x[1]
        arg[1,1]=(0.0303890146585)*x[0]**o+(-0.712865964696)*x[0]+(0.572471431129)*x[1]**o+(-0.194324510922)*x[1]
        arg[1,2]=(-0.757134870981)*x[0]**o+(-0.018314520594)*x[0]+(0.250423290169)*x[1]**o+(-0.876678823729)*x[1]
        arg[1,3]=(-0.351885929068)*x[0]**o+(-0.561258605321)*x[0]+(0.208995575986)*x[1]**o+(0.428437856056)*x[1]
        arg[1,4]=(-0.313383912526)*x[0]**o+(-0.713484404328)*x[0]+(-0.279619274966)*x[1]**o+(0.72420661494)*x[1]
        arg[2,0]=(-0.686382641827)*x[0]**o+(0.678578408675)*x[0]+(-0.696532353786)*x[1]**o+(-0.154644995634)*x[1]
        arg[2,1]=(0.368548759951)*x[0]**o+(-0.674856247769)*x[0]+(0.908734263899)*x[1]**o+(0.331173572353)*x[1]
        arg[2,2]=(-0.1279136556)*x[0]**o+(-0.161474840876)*x[0]+(-0.775009015881)*x[1]**o+(-0.689552100815)*x[1]
        arg[2,3]=(-0.638450048485)*x[0]**o+(0.755485009392)*x[0]+(-0.793363159431)*x[1]**o+(0.255883452369)*x[1]
        arg[2,4]=(-0.831891512113)*x[0]**o+(-0.811678751836)*x[0]+(-0.487866621558)*x[1]**o+(0.240951890349)*x[1]
        arg[3,0]=(0.448249735424)*x[0]**o+(0.0809467993259)*x[0]+(0.232269153692)*x[1]**o+(0.477512628661)*x[1]
        arg[3,1]=(0.494093407104)*x[0]**o+(0.196761586998)*x[0]+(-0.702155296644)*x[1]**o+(-0.748641693731)*x[1]
        arg[3,2]=(0.432370067164)*x[0]**o+(-0.365671005874)*x[0]+(0.452009309356)*x[1]**o+(-0.951328172558)*x[1]
        arg[3,3]=(-0.993290240028)*x[0]**o+(0.172534714933)*x[0]+(0.00917693958413)*x[1]**o+(0.417515657393)*x[1]
        arg[3,4]=(0.541299019375)*x[0]**o+(0.479118172057)*x[0]+(-0.502475287154)*x[1]**o+(-0.04397899716)*x[1]
        ref[0,0]=(-0.120967658047)/(o+1.)+(0.213774478371)
        ref[0,1]=(-0.247665180527)/(o+1.)+(0.360728742109)
        ref[0,2]=(0.952942144598)/(o+1.)+(0.776292907085)
        ref[0,3]=(0.500590719646)/(o+1.)+(0.348159601495)
        ref[0,4]=(-0.591895571605)/(o+1.)+(-0.108089219643)
        ref[1,0]=(0.132923996358)/(o+1.)+(-0.430844529097)
        ref[1,1]=(0.602860445788)/(o+1.)+(-0.453595237809)
        ref[1,2]=(-0.506711580812)/(o+1.)+(-0.447496672161)
        ref[1,3]=(-0.142890353081)/(o+1.)+(-0.066410374632)
        ref[1,4]=(-0.593003187492)/(o+1.)+(0.00536110530621)
        ref[2,0]=(-1.38291499561)/(o+1.)+(0.261966706521)
        ref[2,1]=(1.27728302385)/(o+1.)+(-0.171841337708)
        ref[2,2]=(-0.902922671481)/(o+1.)+(-0.425513470846)
        ref[2,3]=(-1.43181320792)/(o+1.)+(0.505684230881)
        ref[2,4]=(-1.31975813367)/(o+1.)+(-0.285363430743)
        ref[3,0]=(0.680518889117)/(o+1.)+(0.279229713993)
        ref[3,1]=(-0.20806188954)/(o+1.)+(-0.275940053366)
        ref[3,2]=(0.88437937652)/(o+1.)+(-0.658499589216)
        ref[3,3]=(-0.984113300443)/(o+1.)+(0.295025186163)
        ref[3,4]=(0.0388237322207)/(o+1.)+(0.217569587448)
      else:
        arg[0,0]=(-0.866239193113)*x[0]**o+(0.479559390215)*x[0]+(0.512920648125)*x[1]**o+(-0.504775214511)*x[1]+(0.462210441279)*x[2]**o+(0.000249131516168)*x[2]
        arg[0,1]=(-0.11397206495)*x[0]**o+(0.0762933475692)*x[0]+(0.347716642793)*x[1]**o+(0.373262578958)*x[1]+(0.965468442129)*x[2]**o+(-0.743205128452)*x[2]
        arg[0,2]=(-0.214519408107)*x[0]**o+(-0.663675446512)*x[0]+(-0.424102625804)*x[1]**o+(-0.426644021856)*x[1]+(-0.0731662128704)*x[2]**o+(-0.375185307096)*x[2]
        arg[0,3]=(0.599595484989)*x[0]**o+(-0.303187098474)*x[0]+(0.278524134467)*x[1]**o+(-0.579432425528)*x[1]+(0.006698672684)*x[2]**o+(-0.820721120517)*x[2]
        arg[0,4]=(0.538847126305)*x[0]**o+(0.58093792853)*x[0]+(-0.21189097773)*x[1]**o+(0.637101424777)*x[1]+(0.0155994636183)*x[2]**o+(0.488675370945)*x[2]
        arg[1,0]=(-0.775581481064)*x[0]**o+(-0.968381856562)*x[0]+(-0.660755920498)*x[1]**o+(0.746481830633)*x[1]+(-0.300061615221)*x[2]**o+(-0.720031101537)*x[2]
        arg[1,1]=(0.990001351802)*x[0]**o+(-0.168997300944)*x[0]+(0.0134555819367)*x[1]**o+(0.969355591009)*x[1]+(-0.347553438064)*x[2]**o+(0.648537935984)*x[2]
        arg[1,2]=(-0.327760535516)*x[0]**o+(0.517489199537)*x[0]+(-0.0589046196438)*x[1]**o+(-0.893401439507)*x[1]+(-0.0821634552571)*x[2]**o+(0.861266440343)*x[2]
        arg[1,3]=(-0.808782893446)*x[0]**o+(-0.196363777506)*x[0]+(0.623837845847)*x[1]**o+(-0.05392424847)*x[1]+(0.616011657554)*x[2]**o+(0.800682023434)*x[2]
        arg[1,4]=(-0.425073222968)*x[0]**o+(-0.635248695204)*x[0]+(0.947222621311)*x[1]**o+(0.253198813179)*x[1]+(-0.907568459712)*x[2]**o+(-0.513049744062)*x[2]
        arg[2,0]=(-0.103867107528)*x[0]**o+(0.893031214224)*x[0]+(0.151292631908)*x[1]**o+(-0.407189878657)*x[1]+(0.618120612304)*x[2]**o+(0.983992964931)*x[2]
        arg[2,1]=(0.310432646627)*x[0]**o+(0.235143842057)*x[0]+(0.729683627093)*x[1]**o+(0.891778928523)*x[1]+(-0.148927367367)*x[2]**o+(0.913403960844)*x[2]
        arg[2,2]=(0.610293692912)*x[0]**o+(-0.28497731539)*x[0]+(-0.926462472232)*x[1]**o+(-0.427657211191)*x[1]+(0.467905160751)*x[2]**o+(0.195578740767)*x[2]
        arg[2,3]=(0.854869452114)*x[0]**o+(-0.752530632831)*x[0]+(0.0128220565485)*x[1]**o+(-0.726013651836)*x[1]+(-0.0979888078519)*x[2]**o+(-0.658955703807)*x[2]
        arg[2,4]=(0.536447848289)*x[0]**o+(-0.658205406579)*x[0]+(0.66427484293)*x[1]**o+(-0.250014589496)*x[1]+(-0.0654257159073)*x[2]**o+(0.538196496913)*x[2]
        arg[3,0]=(-0.514665034974)*x[0]**o+(-0.278189747423)*x[0]+(0.402996329453)*x[1]**o+(-0.0362969716175)*x[1]+(-0.707890815004)*x[2]**o+(-0.298762865096)*x[2]
        arg[3,1]=(0.7342439327)*x[0]**o+(-0.387336714116)*x[0]+(-0.67888634151)*x[1]**o+(0.647322640846)*x[1]+(-0.0346274738413)*x[2]**o+(0.985950301913)*x[2]
        arg[3,2]=(0.687555064124)*x[0]**o+(0.345506248144)*x[0]+(0.658692258064)*x[1]**o+(0.810796031447)*x[1]+(0.102608006398)*x[2]**o+(0.349877767777)*x[2]
        arg[3,3]=(-0.456241806837)*x[0]**o+(-0.181224535865)*x[0]+(0.401842456169)*x[1]**o+(-0.413936620519)*x[1]+(0.436794604974)*x[2]**o+(0.95051042082)*x[2]
        arg[3,4]=(-0.330739393408)*x[0]**o+(-0.261876658969)*x[0]+(0.287474017623)*x[1]**o+(0.463691723118)*x[1]+(0.757597823206)*x[2]**o+(0.49336246207)*x[2]
        ref[0,0]=(0.108891896292)/(o+1.)+(-0.01248334639)
        ref[0,1]=(1.19921301997)/(o+1.)+(-0.146824600962)
        ref[0,2]=(-0.711788246781)/(o+1.)+(-0.732752387732)
        ref[0,3]=(0.88481829214)/(o+1.)+(-0.85167032226)
        ref[0,4]=(0.342555612194)/(o+1.)+(0.853357362126)
        ref[1,0]=(-1.73639901678)/(o+1.)+(-0.470965563733)
        ref[1,1]=(0.655903495675)/(o+1.)+(0.724448113024)
        ref[1,2]=(-0.468828610417)/(o+1.)+(0.242677100186)
        ref[1,3]=(0.431066609955)/(o+1.)+(0.275196998729)
        ref[1,4]=(-0.385419061368)/(o+1.)+(-0.447549813044)
        ref[2,0]=(0.665546136685)/(o+1.)+(0.734917150249)
        ref[2,1]=(0.891188906354)/(o+1.)+(1.02016336571)
        ref[2,2]=(0.151736381432)/(o+1.)+(-0.258527892907)
        ref[2,3]=(0.76970270081)/(o+1.)+(-1.06874999424)
        ref[2,4]=(1.13529697531)/(o+1.)+(-0.185011749581)
        ref[3,0]=(-0.819559520525)/(o+1.)+(-0.306624792068)
        ref[3,1]=(0.020730117348)/(o+1.)+(0.622968114321)
        ref[3,2]=(1.44885532859)/(o+1.)+(0.753090023684)
        ref[3,3]=(0.382395254306)/(o+1.)+(0.177674632218)
        ref[3,4]=(0.714332447421)/(o+1.)+(0.34758876311)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.779261125448)*x[0]**o+(-0.640503838419)*x[0]+(0.589085035572)*x[1]**o+(-0.0426687544445)*x[1]
        arg[0,0,1]=(-0.419361292473)*x[0]**o+(0.0850684591011)*x[0]+(-0.440758336688)*x[1]**o+(0.151015611594)*x[1]
        arg[0,1,0]=(-0.727043966709)*x[0]**o+(0.711730063365)*x[0]+(0.657759819512)*x[1]**o+(-0.210981156609)*x[1]
        arg[0,1,1]=(-0.276491768644)*x[0]**o+(-0.324865684188)*x[0]+(-0.921856877627)*x[1]**o+(-0.246833902926)*x[1]
        arg[1,0,0]=(0.720541615152)*x[0]**o+(-0.0406851820986)*x[0]+(0.33976001267)*x[1]**o+(0.937989079378)*x[1]
        arg[1,0,1]=(0.239272962875)*x[0]**o+(0.482026352856)*x[0]+(0.159148474482)*x[1]**o+(0.680831564145)*x[1]
        arg[1,1,0]=(0.55170300456)*x[0]**o+(-0.261357073238)*x[0]+(0.298185649596)*x[1]**o+(-0.540598468483)*x[1]
        arg[1,1,1]=(0.508262701362)*x[0]**o+(-0.531692643817)*x[0]+(0.151424571021)*x[1]**o+(-0.77488755481)*x[1]
        arg[2,0,0]=(0.265201389846)*x[0]**o+(0.377300848247)*x[0]+(0.0402148688087)*x[1]**o+(0.0263806695524)*x[1]
        arg[2,0,1]=(0.321397565411)*x[0]**o+(-0.277650913698)*x[0]+(0.909430556637)*x[1]**o+(-0.43244155765)*x[1]
        arg[2,1,0]=(-0.671417887851)*x[0]**o+(-0.772328065901)*x[0]+(0.0896878386442)*x[1]**o+(0.584773963567)*x[1]
        arg[2,1,1]=(0.318995094412)*x[0]**o+(-0.396578444213)*x[0]+(-0.627767362562)*x[1]**o+(-0.588717342984)*x[1]
        arg[3,0,0]=(-0.154999830633)*x[0]**o+(-0.263434767944)*x[0]+(0.0704862479101)*x[1]**o+(0.216237505495)*x[1]
        arg[3,0,1]=(0.956868643095)*x[0]**o+(-0.0331644426873)*x[0]+(-0.236340025853)*x[1]**o+(-0.814705425837)*x[1]
        arg[3,1,0]=(-0.477985064084)*x[0]**o+(-0.368896833289)*x[0]+(0.208122873207)*x[1]**o+(-0.566525593432)*x[1]
        arg[3,1,1]=(-0.299257275248)*x[0]**o+(0.937022356746)*x[0]+(0.671724793932)*x[1]**o+(0.80705820341)*x[1]
        arg[4,0,0]=(0.857855249735)*x[0]**o+(0.537756541143)*x[0]+(0.611784225987)*x[1]**o+(0.944061991186)*x[1]
        arg[4,0,1]=(0.825779290308)*x[0]**o+(-0.636461636739)*x[0]+(-0.382007450888)*x[1]**o+(0.170178262303)*x[1]
        arg[4,1,0]=(0.275666857657)*x[0]**o+(-0.404983003194)*x[0]+(0.476932065223)*x[1]**o+(0.931645726493)*x[1]
        arg[4,1,1]=(0.956182179652)*x[0]**o+(0.543482655321)*x[0]+(0.60492494312)*x[1]**o+(0.384497838341)*x[1]
        arg[5,0,0]=(0.476937368598)*x[0]**o+(-0.588434801718)*x[0]+(0.865080125217)*x[1]**o+(-0.954493375041)*x[1]
        arg[5,0,1]=(0.151850214007)*x[0]**o+(0.252651124764)*x[0]+(-0.570132867968)*x[1]**o+(-0.0885902498736)*x[1]
        arg[5,1,0]=(-0.69330471578)*x[0]**o+(0.165874518143)*x[0]+(-0.610503496045)*x[1]**o+(-0.871951964138)*x[1]
        arg[5,1,1]=(-0.539191264945)*x[0]**o+(0.558474736927)*x[0]+(0.752045569996)*x[1]**o+(-0.5305308467)*x[1]
        ref[0,0,0]=(1.36834616102)/(o+1.)+(-0.341586296432)
        ref[0,0,1]=(-0.860119629161)/(o+1.)+(0.118042035348)
        ref[0,1,0]=(-0.0692841471964)/(o+1.)+(0.250374453378)
        ref[0,1,1]=(-1.19834864627)/(o+1.)+(-0.285849793557)
        ref[1,0,0]=(1.06030162782)/(o+1.)+(0.448651948639)
        ref[1,0,1]=(0.398421437357)/(o+1.)+(0.5814289585)
        ref[1,1,0]=(0.849888654155)/(o+1.)+(-0.400977770861)
        ref[1,1,1]=(0.659687272383)/(o+1.)+(-0.653290099314)
        ref[2,0,0]=(0.305416258655)/(o+1.)+(0.2018407589)
        ref[2,0,1]=(1.23082812205)/(o+1.)+(-0.355046235674)
        ref[2,1,0]=(-0.581730049206)/(o+1.)+(-0.0937770511668)
        ref[2,1,1]=(-0.30877226815)/(o+1.)+(-0.492647893598)
        ref[3,0,0]=(-0.0845135827233)/(o+1.)+(-0.0235986312242)
        ref[3,0,1]=(0.720528617241)/(o+1.)+(-0.423934934262)
        ref[3,1,0]=(-0.269862190877)/(o+1.)+(-0.46771121336)
        ref[3,1,1]=(0.372467518684)/(o+1.)+(0.872040280078)
        ref[4,0,0]=(1.46963947572)/(o+1.)+(0.740909266165)
        ref[4,0,1]=(0.44377183942)/(o+1.)+(-0.233141687218)
        ref[4,1,0]=(0.75259892288)/(o+1.)+(0.26333136165)
        ref[4,1,1]=(1.56110712277)/(o+1.)+(0.463990246831)
        ref[5,0,0]=(1.34201749382)/(o+1.)+(-0.771464088379)
        ref[5,0,1]=(-0.418282653961)/(o+1.)+(0.0820304374451)
        ref[5,1,0]=(-1.30380821182)/(o+1.)+(-0.353038722997)
        ref[5,1,1]=(0.212854305051)/(o+1.)+(0.0139719451134)
      else:
        arg[0,0,0]=(-0.825926001774)*x[0]**o+(0.418049316138)*x[0]+(0.668695279849)*x[1]**o+(-0.7323605258)*x[1]+(-0.637717742224)*x[2]**o+(0.257065974143)*x[2]
        arg[0,0,1]=(-0.169779448443)*x[0]**o+(-0.648612931029)*x[0]+(-0.643814352995)*x[1]**o+(0.396704595574)*x[1]+(-0.248766540261)*x[2]**o+(-0.736410673282)*x[2]
        arg[0,1,0]=(-0.371929441192)*x[0]**o+(0.438661796459)*x[0]+(-0.893450426264)*x[1]**o+(0.643494077636)*x[1]+(0.252099102122)*x[2]**o+(-0.0857798075076)*x[2]
        arg[0,1,1]=(0.336251711034)*x[0]**o+(-0.317426111079)*x[0]+(-0.801221986791)*x[1]**o+(0.281123033042)*x[1]+(0.921153543163)*x[2]**o+(0.497003835923)*x[2]
        arg[1,0,0]=(0.40897311872)*x[0]**o+(0.326894801733)*x[0]+(0.216268807497)*x[1]**o+(0.946235364974)*x[1]+(-0.480995485096)*x[2]**o+(0.869772937982)*x[2]
        arg[1,0,1]=(0.108715554449)*x[0]**o+(0.0847529231253)*x[0]+(0.866120965844)*x[1]**o+(0.43989772616)*x[1]+(0.67962061792)*x[2]**o+(0.90311988632)*x[2]
        arg[1,1,0]=(0.28571484173)*x[0]**o+(0.593050819853)*x[0]+(-0.384864010146)*x[1]**o+(-0.57920566259)*x[1]+(-0.304024223769)*x[2]**o+(-0.0279518764989)*x[2]
        arg[1,1,1]=(0.249126658892)*x[0]**o+(0.243355410215)*x[0]+(0.154816978135)*x[1]**o+(-0.260343533153)*x[1]+(-0.973133787852)*x[2]**o+(0.752545508531)*x[2]
        arg[2,0,0]=(0.103612497259)*x[0]**o+(0.0919234144108)*x[0]+(0.416394965635)*x[1]**o+(-0.0132359939268)*x[1]+(-0.522775855178)*x[2]**o+(-0.985811066906)*x[2]
        arg[2,0,1]=(-0.592641488144)*x[0]**o+(0.703703144485)*x[0]+(-0.0149046721584)*x[1]**o+(0.220368283581)*x[1]+(-0.656858164281)*x[2]**o+(0.165951102113)*x[2]
        arg[2,1,0]=(0.309492848012)*x[0]**o+(-0.301037623847)*x[0]+(-0.224814259669)*x[1]**o+(-0.950794406688)*x[1]+(-0.296088293363)*x[2]**o+(-0.712930293168)*x[2]
        arg[2,1,1]=(0.0592633594166)*x[0]**o+(-0.179007619512)*x[0]+(-0.922169690831)*x[1]**o+(-0.778228075089)*x[1]+(-0.135749453777)*x[2]**o+(-0.146265051618)*x[2]
        arg[3,0,0]=(-0.71773633584)*x[0]**o+(0.471551385738)*x[0]+(-0.041415831943)*x[1]**o+(0.343380640343)*x[1]+(-0.211821446187)*x[2]**o+(-0.661201893183)*x[2]
        arg[3,0,1]=(-0.159071045836)*x[0]**o+(0.216715127143)*x[0]+(-0.846847485213)*x[1]**o+(0.662226888693)*x[1]+(0.050806337433)*x[2]**o+(0.744071347051)*x[2]
        arg[3,1,0]=(-0.363431877105)*x[0]**o+(-0.461513470653)*x[0]+(-0.130947882739)*x[1]**o+(-0.624729599981)*x[1]+(0.778572567699)*x[2]**o+(0.324584316744)*x[2]
        arg[3,1,1]=(0.395432262346)*x[0]**o+(0.826450301981)*x[0]+(0.176721657975)*x[1]**o+(-0.370954952405)*x[1]+(-0.708437370609)*x[2]**o+(-0.150075114354)*x[2]
        arg[4,0,0]=(-0.18662583335)*x[0]**o+(-0.0120831234457)*x[0]+(0.805686139762)*x[1]**o+(0.30195413136)*x[1]+(0.648067854031)*x[2]**o+(0.28073240632)*x[2]
        arg[4,0,1]=(-0.913625644854)*x[0]**o+(-0.264637827098)*x[0]+(-0.6531590441)*x[1]**o+(0.396422722623)*x[1]+(0.280591271146)*x[2]**o+(0.498346191707)*x[2]
        arg[4,1,0]=(0.208029597907)*x[0]**o+(-0.295747909396)*x[0]+(-0.541509895186)*x[1]**o+(-0.511394562575)*x[1]+(-0.459197562982)*x[2]**o+(-0.749433614666)*x[2]
        arg[4,1,1]=(-0.29700028905)*x[0]**o+(-0.590932598278)*x[0]+(0.36219700213)*x[1]**o+(0.482759242709)*x[1]+(-0.135696240971)*x[2]**o+(0.49658807849)*x[2]
        arg[5,0,0]=(-0.800345624219)*x[0]**o+(0.259709746207)*x[0]+(0.480073493816)*x[1]**o+(0.499332550127)*x[1]+(0.691794186846)*x[2]**o+(-0.319369042249)*x[2]
        arg[5,0,1]=(0.300879052118)*x[0]**o+(-0.985396817441)*x[0]+(0.971136879002)*x[1]**o+(0.806560091745)*x[1]+(0.361308859099)*x[2]**o+(0.264277089732)*x[2]
        arg[5,1,0]=(0.271857906548)*x[0]**o+(0.338260411675)*x[0]+(-0.170292665803)*x[1]**o+(0.675491566142)*x[1]+(0.897921755421)*x[2]**o+(-0.829233864443)*x[2]
        arg[5,1,1]=(0.673366845241)*x[0]**o+(0.121940550422)*x[0]+(0.116054155545)*x[1]**o+(0.0456556938689)*x[1]+(-0.978127590673)*x[2]**o+(0.180985351995)*x[2]
        ref[0,0,0]=(-0.794948464148)/(o+1.)+(-0.0286226177598)
        ref[0,0,1]=(-1.0623603417)/(o+1.)+(-0.494159504368)
        ref[0,1,0]=(-1.01328076533)/(o+1.)+(0.498188033294)
        ref[0,1,1]=(0.456183267406)/(o+1.)+(0.230350378943)
        ref[1,0,0]=(0.144246441122)/(o+1.)+(1.07145155234)
        ref[1,0,1]=(1.65445713821)/(o+1.)+(0.713885267803)
        ref[1,1,0]=(-0.403173392185)/(o+1.)+(-0.0070533596178)
        ref[1,1,1]=(-0.569190150825)/(o+1.)+(0.367778692797)
        ref[2,0,0]=(-0.00276839228366)/(o+1.)+(-0.453561823211)
        ref[2,0,1]=(-1.26440432458)/(o+1.)+(0.545011265089)
        ref[2,1,0]=(-0.21140970502)/(o+1.)+(-0.982381161851)
        ref[2,1,1]=(-0.998655785191)/(o+1.)+(-0.55175037311)
        ref[3,0,0]=(-0.97097361397)/(o+1.)+(0.0768650664488)
        ref[3,0,1]=(-0.955112193616)/(o+1.)+(0.811506681444)
        ref[3,1,0]=(0.284192807855)/(o+1.)+(-0.380829376946)
        ref[3,1,1]=(-0.136283450288)/(o+1.)+(0.152710117611)
        ref[4,0,0]=(1.26712816044)/(o+1.)+(0.285301707117)
        ref[4,0,1]=(-1.28619341781)/(o+1.)+(0.315065543616)
        ref[4,1,0]=(-0.792677860261)/(o+1.)+(-0.778288043318)
        ref[4,1,1]=(-0.0704995278906)/(o+1.)+(0.194207361461)
        ref[5,0,0]=(0.371522056443)/(o+1.)+(0.219836627043)
        ref[5,0,1]=(1.63332479022)/(o+1.)+(0.0427201820182)
        ref[5,1,0]=(0.999486996166)/(o+1.)+(0.0922590566873)
        ref[5,1,1]=(-0.188706589887)/(o+1.)+(0.174290798143)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.647932589381)*x[0]**o+(-0.092686685918)*x[0]+(0.167888464541)*x[1]**o+(0.698445840627)*x[1]
        arg[0,0,0,1]=(0.413204959658)*x[0]**o+(0.772926029512)*x[0]+(-0.63702948026)*x[1]**o+(0.608392823861)*x[1]
        arg[0,0,1,0]=(0.303506932052)*x[0]**o+(0.650309664063)*x[0]+(-0.805238344485)*x[1]**o+(-0.62971680995)*x[1]
        arg[0,0,1,1]=(-0.205701190736)*x[0]**o+(-0.0834042840814)*x[0]+(-0.525839398431)*x[1]**o+(0.779973223005)*x[1]
        arg[0,0,2,0]=(-0.97954633631)*x[0]**o+(-0.298973871114)*x[0]+(0.858022652298)*x[1]**o+(0.667671796632)*x[1]
        arg[0,0,2,1]=(-0.790133712345)*x[0]**o+(-0.733743414808)*x[0]+(0.337185148145)*x[1]**o+(-0.927853051244)*x[1]
        arg[0,1,0,0]=(0.617198112458)*x[0]**o+(0.70438849416)*x[0]+(-0.861661596364)*x[1]**o+(-0.251796116219)*x[1]
        arg[0,1,0,1]=(0.37107453636)*x[0]**o+(-0.490440661153)*x[0]+(0.849252675242)*x[1]**o+(-0.896547775077)*x[1]
        arg[0,1,1,0]=(0.375080744007)*x[0]**o+(-0.609557663503)*x[0]+(-0.930846980977)*x[1]**o+(0.0623096843381)*x[1]
        arg[0,1,1,1]=(-0.809863615805)*x[0]**o+(-0.999724521803)*x[0]+(0.493027030633)*x[1]**o+(-0.672546253752)*x[1]
        arg[0,1,2,0]=(0.219294077412)*x[0]**o+(-0.7854819847)*x[0]+(-0.353026321798)*x[1]**o+(-0.965348418509)*x[1]
        arg[0,1,2,1]=(0.177236321722)*x[0]**o+(-0.860491592895)*x[0]+(-0.458724268899)*x[1]**o+(-0.476508229653)*x[1]
        arg[0,2,0,0]=(0.0760611767074)*x[0]**o+(-0.00296398964042)*x[0]+(-0.972843039436)*x[1]**o+(-0.0422057743881)*x[1]
        arg[0,2,0,1]=(-0.926737777701)*x[0]**o+(-0.895791593378)*x[0]+(-0.776529150964)*x[1]**o+(0.557115706978)*x[1]
        arg[0,2,1,0]=(0.353264702858)*x[0]**o+(0.632583933138)*x[0]+(-0.354112684469)*x[1]**o+(0.0541346174117)*x[1]
        arg[0,2,1,1]=(-0.124181989217)*x[0]**o+(-0.803262099858)*x[0]+(0.733246361984)*x[1]**o+(0.0639821744038)*x[1]
        arg[0,2,2,0]=(-0.446733845136)*x[0]**o+(-0.187373171381)*x[0]+(-0.312309043572)*x[1]**o+(0.830789380477)*x[1]
        arg[0,2,2,1]=(0.403889932995)*x[0]**o+(-0.243474625233)*x[0]+(0.35707779102)*x[1]**o+(0.29725421988)*x[1]
        arg[0,3,0,0]=(-0.856376546696)*x[0]**o+(0.448431504377)*x[0]+(-0.535187100108)*x[1]**o+(-0.90994242894)*x[1]
        arg[0,3,0,1]=(-0.321475754775)*x[0]**o+(0.0389358177737)*x[0]+(-0.0949733279735)*x[1]**o+(0.921937240884)*x[1]
        arg[0,3,1,0]=(0.591276336387)*x[0]**o+(-0.598236438303)*x[0]+(-0.829829995185)*x[1]**o+(0.539669346291)*x[1]
        arg[0,3,1,1]=(-0.944726366074)*x[0]**o+(-0.796928273271)*x[0]+(-0.424606656681)*x[1]**o+(0.35341306063)*x[1]
        arg[0,3,2,0]=(-0.850914599406)*x[0]**o+(0.791179202404)*x[0]+(-0.0220845565851)*x[1]**o+(0.594756136957)*x[1]
        arg[0,3,2,1]=(0.355862680641)*x[0]**o+(0.134261680432)*x[0]+(0.477798827075)*x[1]**o+(0.508452338499)*x[1]
        arg[0,4,0,0]=(0.481974944781)*x[0]**o+(0.259271175551)*x[0]+(0.953827692515)*x[1]**o+(0.364160709544)*x[1]
        arg[0,4,0,1]=(-0.316627315143)*x[0]**o+(-0.856616735533)*x[0]+(0.308508314695)*x[1]**o+(-0.677602758367)*x[1]
        arg[0,4,1,0]=(-0.76069824745)*x[0]**o+(0.282351651183)*x[0]+(-0.15964276552)*x[1]**o+(-0.344276428838)*x[1]
        arg[0,4,1,1]=(9.00669303738e-05)*x[0]**o+(0.948498888115)*x[0]+(0.690370267333)*x[1]**o+(0.619119188327)*x[1]
        arg[0,4,2,0]=(0.957692561039)*x[0]**o+(-0.147005436631)*x[0]+(-0.177840574215)*x[1]**o+(-0.28259452614)*x[1]
        arg[0,4,2,1]=(-0.431771943451)*x[0]**o+(-0.158132166651)*x[0]+(-0.134385316147)*x[1]**o+(0.751243590835)*x[1]
        arg[1,0,0,0]=(-0.816841420645)*x[0]**o+(0.55608373851)*x[0]+(-0.885838014201)*x[1]**o+(-0.467159163867)*x[1]
        arg[1,0,0,1]=(-0.543361117401)*x[0]**o+(0.329490019675)*x[0]+(0.942703992489)*x[1]**o+(-0.261273838803)*x[1]
        arg[1,0,1,0]=(0.865108106801)*x[0]**o+(-0.376972427024)*x[0]+(-0.0257304693919)*x[1]**o+(0.694210909726)*x[1]
        arg[1,0,1,1]=(0.428354498121)*x[0]**o+(0.962704601892)*x[0]+(-0.442636806139)*x[1]**o+(0.424724924149)*x[1]
        arg[1,0,2,0]=(-0.157277242299)*x[0]**o+(0.487588443251)*x[0]+(-0.813089096341)*x[1]**o+(0.0406316718829)*x[1]
        arg[1,0,2,1]=(0.345322702588)*x[0]**o+(0.224429042659)*x[0]+(-0.24109926586)*x[1]**o+(0.433141241931)*x[1]
        arg[1,1,0,0]=(0.956901263519)*x[0]**o+(0.48418062914)*x[0]+(0.14089030013)*x[1]**o+(-0.492352201717)*x[1]
        arg[1,1,0,1]=(-0.679781695411)*x[0]**o+(0.29857196636)*x[0]+(-0.234968984185)*x[1]**o+(0.114575919411)*x[1]
        arg[1,1,1,0]=(0.488378272934)*x[0]**o+(-0.358508327343)*x[0]+(-0.337958846849)*x[1]**o+(0.189303452912)*x[1]
        arg[1,1,1,1]=(-0.506993859574)*x[0]**o+(-0.682478021258)*x[0]+(-0.540344338899)*x[1]**o+(0.366965535976)*x[1]
        arg[1,1,2,0]=(-0.681151638572)*x[0]**o+(0.790837550009)*x[0]+(-0.810785716023)*x[1]**o+(-0.00739097962416)*x[1]
        arg[1,1,2,1]=(-0.355290573677)*x[0]**o+(-0.263294580108)*x[0]+(0.10469018402)*x[1]**o+(-0.650388283116)*x[1]
        arg[1,2,0,0]=(0.906212689905)*x[0]**o+(-0.673956103538)*x[0]+(0.305335278934)*x[1]**o+(-0.0805949628719)*x[1]
        arg[1,2,0,1]=(-0.98311373852)*x[0]**o+(-0.165316704542)*x[0]+(0.197637770603)*x[1]**o+(0.647989184342)*x[1]
        arg[1,2,1,0]=(0.415990736251)*x[0]**o+(-0.710476520822)*x[0]+(0.791512786296)*x[1]**o+(-0.10853188742)*x[1]
        arg[1,2,1,1]=(0.04337234954)*x[0]**o+(0.595650094449)*x[0]+(0.228098665648)*x[1]**o+(-0.868930899722)*x[1]
        arg[1,2,2,0]=(-0.791842482688)*x[0]**o+(-0.611246387621)*x[0]+(0.0504978789412)*x[1]**o+(0.624626300694)*x[1]
        arg[1,2,2,1]=(0.0207375595263)*x[0]**o+(0.560638072838)*x[0]+(0.44126312508)*x[1]**o+(0.169151353084)*x[1]
        arg[1,3,0,0]=(-0.0512069089279)*x[0]**o+(0.51110271913)*x[0]+(0.0785633458626)*x[1]**o+(0.836821453174)*x[1]
        arg[1,3,0,1]=(0.185737468851)*x[0]**o+(0.516071031227)*x[0]+(0.0995388802342)*x[1]**o+(0.25527909092)*x[1]
        arg[1,3,1,0]=(0.318240681435)*x[0]**o+(-0.442725217398)*x[0]+(0.533066616909)*x[1]**o+(0.679570707339)*x[1]
        arg[1,3,1,1]=(0.353367397644)*x[0]**o+(-0.396487189861)*x[0]+(-0.117773711001)*x[1]**o+(0.885897675293)*x[1]
        arg[1,3,2,0]=(-0.647645988694)*x[0]**o+(-0.755258662164)*x[0]+(0.288300257472)*x[1]**o+(0.505342385389)*x[1]
        arg[1,3,2,1]=(0.000781386924262)*x[0]**o+(0.190244787656)*x[0]+(0.589388834071)*x[1]**o+(0.669705413859)*x[1]
        arg[1,4,0,0]=(0.320107896841)*x[0]**o+(0.673508197912)*x[0]+(0.667026895874)*x[1]**o+(-0.818578971562)*x[1]
        arg[1,4,0,1]=(-0.226755915796)*x[0]**o+(0.71800207991)*x[0]+(0.850490212043)*x[1]**o+(0.261708964138)*x[1]
        arg[1,4,1,0]=(-0.0147350282361)*x[0]**o+(0.743889536542)*x[0]+(-0.898802067607)*x[1]**o+(0.711589170571)*x[1]
        arg[1,4,1,1]=(-0.816145493818)*x[0]**o+(0.405931568806)*x[0]+(0.159156576579)*x[1]**o+(0.143576116984)*x[1]
        arg[1,4,2,0]=(0.994791165013)*x[0]**o+(0.65054901122)*x[0]+(0.039218059733)*x[1]**o+(-0.662978657934)*x[1]
        arg[1,4,2,1]=(-0.792277825616)*x[0]**o+(-0.257224872111)*x[0]+(-0.244502392787)*x[1]**o+(-0.276042189495)*x[1]
        arg[2,0,0,0]=(0.133211578016)*x[0]**o+(0.10029141733)*x[0]+(-0.52480547182)*x[1]**o+(0.814652707336)*x[1]
        arg[2,0,0,1]=(0.836429340685)*x[0]**o+(-0.837709322969)*x[0]+(-0.442953015462)*x[1]**o+(-0.984757344044)*x[1]
        arg[2,0,1,0]=(0.893738853578)*x[0]**o+(-0.238050980408)*x[0]+(-0.913668527302)*x[1]**o+(0.895136777323)*x[1]
        arg[2,0,1,1]=(-0.0457361766611)*x[0]**o+(0.414814798208)*x[0]+(-0.57261759439)*x[1]**o+(-0.358969048076)*x[1]
        arg[2,0,2,0]=(-0.320394419991)*x[0]**o+(-0.990129952721)*x[0]+(0.890888895016)*x[1]**o+(0.241063301452)*x[1]
        arg[2,0,2,1]=(-0.604164679932)*x[0]**o+(-0.598939404114)*x[0]+(-0.606633388454)*x[1]**o+(-0.666879580873)*x[1]
        arg[2,1,0,0]=(0.68515547034)*x[0]**o+(0.892247437859)*x[0]+(0.642003383991)*x[1]**o+(0.569240970392)*x[1]
        arg[2,1,0,1]=(0.918713108587)*x[0]**o+(0.865547962284)*x[0]+(0.517952260633)*x[1]**o+(-0.307334010857)*x[1]
        arg[2,1,1,0]=(0.353382468404)*x[0]**o+(-0.542718692644)*x[0]+(-0.957182863451)*x[1]**o+(-0.889352632225)*x[1]
        arg[2,1,1,1]=(0.505185847817)*x[0]**o+(-0.71137143775)*x[0]+(0.853785404375)*x[1]**o+(-0.731028307335)*x[1]
        arg[2,1,2,0]=(-0.0805774093876)*x[0]**o+(0.949831911897)*x[0]+(0.511009248375)*x[1]**o+(-0.0192710206538)*x[1]
        arg[2,1,2,1]=(-0.987709862593)*x[0]**o+(-0.536077508064)*x[0]+(-0.0726882784026)*x[1]**o+(-0.27275457372)*x[1]
        arg[2,2,0,0]=(0.271001953238)*x[0]**o+(0.0446246860776)*x[0]+(0.103657462724)*x[1]**o+(0.359900072694)*x[1]
        arg[2,2,0,1]=(0.613460277623)*x[0]**o+(-0.23324730519)*x[0]+(0.188458167579)*x[1]**o+(-0.574087648909)*x[1]
        arg[2,2,1,0]=(-0.787444207567)*x[0]**o+(0.143772612473)*x[0]+(-0.720451072166)*x[1]**o+(0.936819171338)*x[1]
        arg[2,2,1,1]=(-0.219701379296)*x[0]**o+(-0.343100221149)*x[0]+(-0.972737482064)*x[1]**o+(-0.614208415137)*x[1]
        arg[2,2,2,0]=(0.0292070295438)*x[0]**o+(-0.694263454424)*x[0]+(0.188064440036)*x[1]**o+(0.0901763471697)*x[1]
        arg[2,2,2,1]=(0.407629585245)*x[0]**o+(-0.745619322197)*x[0]+(0.949044983116)*x[1]**o+(-0.316057288128)*x[1]
        arg[2,3,0,0]=(0.124212182643)*x[0]**o+(0.621703244759)*x[0]+(-0.0592136772296)*x[1]**o+(0.571981450786)*x[1]
        arg[2,3,0,1]=(-0.683989210442)*x[0]**o+(-0.318774594157)*x[0]+(0.850235115277)*x[1]**o+(-0.871075179838)*x[1]
        arg[2,3,1,0]=(0.478657634044)*x[0]**o+(-0.43702074673)*x[0]+(0.535572237843)*x[1]**o+(0.22354698199)*x[1]
        arg[2,3,1,1]=(0.0391087869133)*x[0]**o+(0.765140505702)*x[0]+(-0.66279875624)*x[1]**o+(-0.41964006149)*x[1]
        arg[2,3,2,0]=(-0.11123799186)*x[0]**o+(0.533398540082)*x[0]+(-0.931488199925)*x[1]**o+(-0.83080969281)*x[1]
        arg[2,3,2,1]=(-0.138259313941)*x[0]**o+(-0.508937684332)*x[0]+(0.583603445711)*x[1]**o+(-0.0782724872225)*x[1]
        arg[2,4,0,0]=(-0.316223936951)*x[0]**o+(0.329776289164)*x[0]+(-0.692491559115)*x[1]**o+(0.814572730294)*x[1]
        arg[2,4,0,1]=(0.74890960681)*x[0]**o+(-0.79167945549)*x[0]+(0.175328119487)*x[1]**o+(-0.253003240064)*x[1]
        arg[2,4,1,0]=(0.657827625141)*x[0]**o+(-0.268878597055)*x[0]+(0.247374286921)*x[1]**o+(0.414500041415)*x[1]
        arg[2,4,1,1]=(-0.187587046202)*x[0]**o+(-0.594983066039)*x[0]+(-0.203784406879)*x[1]**o+(0.122237566739)*x[1]
        arg[2,4,2,0]=(0.671210794607)*x[0]**o+(0.886706498415)*x[0]+(-0.368997114212)*x[1]**o+(0.0418311781187)*x[1]
        arg[2,4,2,1]=(-0.272424416171)*x[0]**o+(-0.537477247783)*x[0]+(0.0405655834868)*x[1]**o+(-0.179684833703)*x[1]
        arg[3,0,0,0]=(0.457602964193)*x[0]**o+(-0.186625111516)*x[0]+(-0.75244105343)*x[1]**o+(-0.278893445377)*x[1]
        arg[3,0,0,1]=(-0.690378721784)*x[0]**o+(0.0157608350188)*x[0]+(-0.203765434813)*x[1]**o+(0.69771096957)*x[1]
        arg[3,0,1,0]=(-0.884824688493)*x[0]**o+(0.26613451569)*x[0]+(0.226081491639)*x[1]**o+(0.597724298389)*x[1]
        arg[3,0,1,1]=(0.0372124540969)*x[0]**o+(-0.420242551812)*x[0]+(-0.625683613012)*x[1]**o+(0.279682305563)*x[1]
        arg[3,0,2,0]=(-0.570873872458)*x[0]**o+(-0.82361574923)*x[0]+(0.970204188337)*x[1]**o+(-0.242725003569)*x[1]
        arg[3,0,2,1]=(-0.714826095124)*x[0]**o+(-0.334939427182)*x[0]+(-0.730058530908)*x[1]**o+(0.287148154312)*x[1]
        arg[3,1,0,0]=(0.164234307274)*x[0]**o+(-0.388849394356)*x[0]+(0.391518896617)*x[1]**o+(0.00142750126236)*x[1]
        arg[3,1,0,1]=(-0.327700781196)*x[0]**o+(0.0442604682565)*x[0]+(-0.683834466536)*x[1]**o+(-0.665567774985)*x[1]
        arg[3,1,1,0]=(0.135979934041)*x[0]**o+(-0.822614011173)*x[0]+(-0.946061768113)*x[1]**o+(-0.585442102671)*x[1]
        arg[3,1,1,1]=(0.235568769871)*x[0]**o+(0.738233870703)*x[0]+(0.437750339594)*x[1]**o+(-0.44521724321)*x[1]
        arg[3,1,2,0]=(0.449256898185)*x[0]**o+(-0.464211755196)*x[0]+(0.616886411002)*x[1]**o+(0.885500269963)*x[1]
        arg[3,1,2,1]=(0.0851105760559)*x[0]**o+(0.888976499957)*x[0]+(0.0842441111314)*x[1]**o+(-0.366994083394)*x[1]
        arg[3,2,0,0]=(0.77120050858)*x[0]**o+(0.354835584949)*x[0]+(0.406813878866)*x[1]**o+(-0.600042478179)*x[1]
        arg[3,2,0,1]=(-0.747876879184)*x[0]**o+(0.527204926634)*x[0]+(0.134965492713)*x[1]**o+(0.497662739459)*x[1]
        arg[3,2,1,0]=(0.367550591163)*x[0]**o+(-0.700608509353)*x[0]+(0.0737696587931)*x[1]**o+(-0.164466899816)*x[1]
        arg[3,2,1,1]=(0.0908791429433)*x[0]**o+(-0.168468767791)*x[0]+(0.85978873292)*x[1]**o+(-0.785241412884)*x[1]
        arg[3,2,2,0]=(-0.296455818292)*x[0]**o+(-0.0277004420512)*x[0]+(0.409183414933)*x[1]**o+(-0.879831945543)*x[1]
        arg[3,2,2,1]=(0.971512010223)*x[0]**o+(-0.811249982262)*x[0]+(-0.846549204737)*x[1]**o+(0.689986186691)*x[1]
        arg[3,3,0,0]=(-0.800235545702)*x[0]**o+(-0.796021196346)*x[0]+(-0.0660744375539)*x[1]**o+(-0.826632192842)*x[1]
        arg[3,3,0,1]=(0.305678237361)*x[0]**o+(0.203936000462)*x[0]+(-0.583171879217)*x[1]**o+(0.924112859119)*x[1]
        arg[3,3,1,0]=(-0.937148913762)*x[0]**o+(-0.172547991203)*x[0]+(0.928491154238)*x[1]**o+(-0.502876301351)*x[1]
        arg[3,3,1,1]=(0.731413856616)*x[0]**o+(0.837213097155)*x[0]+(-0.650032764358)*x[1]**o+(0.96973961589)*x[1]
        arg[3,3,2,0]=(-0.501994458598)*x[0]**o+(-0.705938508685)*x[0]+(0.664394117366)*x[1]**o+(0.692950029135)*x[1]
        arg[3,3,2,1]=(-0.0948399323409)*x[0]**o+(-0.686744554674)*x[0]+(0.577765681583)*x[1]**o+(0.969310750204)*x[1]
        arg[3,4,0,0]=(-0.484011788432)*x[0]**o+(-0.560744139846)*x[0]+(0.072097608773)*x[1]**o+(-0.103052228904)*x[1]
        arg[3,4,0,1]=(-0.201158934695)*x[0]**o+(0.463105335811)*x[0]+(0.733365706682)*x[1]**o+(-0.330035241794)*x[1]
        arg[3,4,1,0]=(0.252379277102)*x[0]**o+(-0.110596657969)*x[0]+(0.763687257878)*x[1]**o+(-0.345598371263)*x[1]
        arg[3,4,1,1]=(0.0868412855617)*x[0]**o+(0.385098679406)*x[0]+(0.291487288978)*x[1]**o+(-0.112145846209)*x[1]
        arg[3,4,2,0]=(-0.913724685255)*x[0]**o+(-0.0348735884313)*x[0]+(0.550688528861)*x[1]**o+(0.798394152329)*x[1]
        arg[3,4,2,1]=(-0.92870523219)*x[0]**o+(0.177354002469)*x[0]+(0.350393765313)*x[1]**o+(-0.974997198741)*x[1]
        ref[0,0,0,0]=(0.815821053922)/(o+1.)+(0.302879577355)
        ref[0,0,0,1]=(-0.223824520603)/(o+1.)+(0.690659426687)
        ref[0,0,1,0]=(-0.501731412433)/(o+1.)+(0.0102964270565)
        ref[0,0,1,1]=(-0.731540589167)/(o+1.)+(0.348284469462)
        ref[0,0,2,0]=(-0.121523684012)/(o+1.)+(0.184348962759)
        ref[0,0,2,1]=(-0.4529485642)/(o+1.)+(-0.830798233026)
        ref[0,1,0,0]=(-0.244463483906)/(o+1.)+(0.226296188971)
        ref[0,1,0,1]=(1.2203272116)/(o+1.)+(-0.693494218115)
        ref[0,1,1,0]=(-0.55576623697)/(o+1.)+(-0.273623989583)
        ref[0,1,1,1]=(-0.316836585171)/(o+1.)+(-0.836135387777)
        ref[0,1,2,0]=(-0.133732244387)/(o+1.)+(-0.875415201605)
        ref[0,1,2,1]=(-0.281487947177)/(o+1.)+(-0.668499911274)
        ref[0,2,0,0]=(-0.896781862729)/(o+1.)+(-0.0225848820142)
        ref[0,2,0,1]=(-1.70326692867)/(o+1.)+(-0.1693379432)
        ref[0,2,1,0]=(-0.00084798161079)/(o+1.)+(0.343359275275)
        ref[0,2,1,1]=(0.609064372767)/(o+1.)+(-0.369639962727)
        ref[0,2,2,0]=(-0.759042888708)/(o+1.)+(0.321708104548)
        ref[0,2,2,1]=(0.760967724015)/(o+1.)+(0.0268897973234)
        ref[0,3,0,0]=(-1.3915636468)/(o+1.)+(-0.230755462281)
        ref[0,3,0,1]=(-0.416449082748)/(o+1.)+(0.480436529329)
        ref[0,3,1,0]=(-0.238553658798)/(o+1.)+(-0.029283546006)
        ref[0,3,1,1]=(-1.36933302275)/(o+1.)+(-0.221757606321)
        ref[0,3,2,0]=(-0.872999155991)/(o+1.)+(0.69296766968)
        ref[0,3,2,1]=(0.833661507716)/(o+1.)+(0.321357009466)
        ref[0,4,0,0]=(1.4358026373)/(o+1.)+(0.311715942547)
        ref[0,4,0,1]=(-0.00811900044778)/(o+1.)+(-0.76710974695)
        ref[0,4,1,0]=(-0.92034101297)/(o+1.)+(-0.0309623888275)
        ref[0,4,1,1]=(0.690460334263)/(o+1.)+(0.783809038221)
        ref[0,4,2,0]=(0.779851986824)/(o+1.)+(-0.214799981386)
        ref[0,4,2,1]=(-0.566157259598)/(o+1.)+(0.296555712092)
        ref[1,0,0,0]=(-1.70267943485)/(o+1.)+(0.0444622873215)
        ref[1,0,0,1]=(0.399342875087)/(o+1.)+(0.0341080904363)
        ref[1,0,1,0]=(0.839377637409)/(o+1.)+(0.158619241351)
        ref[1,0,1,1]=(-0.0142823080176)/(o+1.)+(0.693714763021)
        ref[1,0,2,0]=(-0.97036633864)/(o+1.)+(0.264110057567)
        ref[1,0,2,1]=(0.104223436728)/(o+1.)+(0.328785142295)
        ref[1,1,0,0]=(1.09779156365)/(o+1.)+(-0.00408578628855)
        ref[1,1,0,1]=(-0.914750679597)/(o+1.)+(0.206573942886)
        ref[1,1,1,0]=(0.150419426085)/(o+1.)+(-0.0846024372156)
        ref[1,1,1,1]=(-1.04733819847)/(o+1.)+(-0.157756242641)
        ref[1,1,2,0]=(-1.49193735459)/(o+1.)+(0.391723285192)
        ref[1,1,2,1]=(-0.250600389657)/(o+1.)+(-0.456841431612)
        ref[1,2,0,0]=(1.21154796884)/(o+1.)+(-0.377275533205)
        ref[1,2,0,1]=(-0.785475967917)/(o+1.)+(0.2413362399)
        ref[1,2,1,0]=(1.20750352255)/(o+1.)+(-0.409504204121)
        ref[1,2,1,1]=(0.271471015188)/(o+1.)+(-0.136640402636)
        ref[1,2,2,0]=(-0.741344603747)/(o+1.)+(0.00668995653618)
        ref[1,2,2,1]=(0.462000684606)/(o+1.)+(0.364894712961)
        ref[1,3,0,0]=(0.0273564369347)/(o+1.)+(0.673962086152)
        ref[1,3,0,1]=(0.285276349085)/(o+1.)+(0.385675061073)
        ref[1,3,1,0]=(0.851307298344)/(o+1.)+(0.118422744971)
        ref[1,3,1,1]=(0.235593686643)/(o+1.)+(0.244705242716)
        ref[1,3,2,0]=(-0.359345731222)/(o+1.)+(-0.124958138388)
        ref[1,3,2,1]=(0.590170220995)/(o+1.)+(0.429975100757)
        ref[1,4,0,0]=(0.987134792715)/(o+1.)+(-0.0725353868253)
        ref[1,4,0,1]=(0.623734296247)/(o+1.)+(0.489855522024)
        ref[1,4,1,0]=(-0.913537095843)/(o+1.)+(0.727739353556)
        ref[1,4,1,1]=(-0.656988917238)/(o+1.)+(0.274753842895)
        ref[1,4,2,0]=(1.03400922475)/(o+1.)+(-0.00621482335664)
        ref[1,4,2,1]=(-1.0367802184)/(o+1.)+(-0.266633530803)
        ref[2,0,0,0]=(-0.391593893803)/(o+1.)+(0.457472062333)
        ref[2,0,0,1]=(0.393476325223)/(o+1.)+(-0.911233333506)
        ref[2,0,1,0]=(-0.0199296737245)/(o+1.)+(0.328542898457)
        ref[2,0,1,1]=(-0.618353771051)/(o+1.)+(0.0279228750664)
        ref[2,0,2,0]=(0.570494475025)/(o+1.)+(-0.374533325635)
        ref[2,0,2,1]=(-1.21079806839)/(o+1.)+(-0.632909492494)
        ref[2,1,0,0]=(1.32715885433)/(o+1.)+(0.730744204126)
        ref[2,1,0,1]=(1.43666536922)/(o+1.)+(0.279106975714)
        ref[2,1,1,0]=(-0.603800395047)/(o+1.)+(-0.716035662434)
        ref[2,1,1,1]=(1.35897125219)/(o+1.)+(-0.721199872542)
        ref[2,1,2,0]=(0.430431838988)/(o+1.)+(0.465280445622)
        ref[2,1,2,1]=(-1.060398141)/(o+1.)+(-0.404416040892)
        ref[2,2,0,0]=(0.374659415961)/(o+1.)+(0.202262379386)
        ref[2,2,0,1]=(0.801918445202)/(o+1.)+(-0.403667477049)
        ref[2,2,1,0]=(-1.50789527973)/(o+1.)+(0.540295891905)
        ref[2,2,1,1]=(-1.19243886136)/(o+1.)+(-0.478654318143)
        ref[2,2,2,0]=(0.21727146958)/(o+1.)+(-0.302043553627)
        ref[2,2,2,1]=(1.35667456836)/(o+1.)+(-0.530838305162)
        ref[2,3,0,0]=(0.0649985054133)/(o+1.)+(0.596842347772)
        ref[2,3,0,1]=(0.166245904835)/(o+1.)+(-0.594924886997)
        ref[2,3,1,0]=(1.01422987189)/(o+1.)+(-0.10673688237)
        ref[2,3,1,1]=(-0.623689969327)/(o+1.)+(0.172750222106)
        ref[2,3,2,0]=(-1.04272619178)/(o+1.)+(-0.148705576364)
        ref[2,3,2,1]=(0.445344131771)/(o+1.)+(-0.293605085777)
        ref[2,4,0,0]=(-1.00871549607)/(o+1.)+(0.572174509729)
        ref[2,4,0,1]=(0.924237726297)/(o+1.)+(-0.522341347777)
        ref[2,4,1,0]=(0.905201912063)/(o+1.)+(0.0728107221801)
        ref[2,4,1,1]=(-0.391371453081)/(o+1.)+(-0.23637274965)
        ref[2,4,2,0]=(0.302213680395)/(o+1.)+(0.464268838267)
        ref[2,4,2,1]=(-0.231858832685)/(o+1.)+(-0.358581040743)
        ref[3,0,0,0]=(-0.294838089237)/(o+1.)+(-0.232759278447)
        ref[3,0,0,1]=(-0.894144156597)/(o+1.)+(0.356735902294)
        ref[3,0,1,0]=(-0.658743196854)/(o+1.)+(0.43192940704)
        ref[3,0,1,1]=(-0.588471158915)/(o+1.)+(-0.0702801231243)
        ref[3,0,2,0]=(0.399330315879)/(o+1.)+(-0.5331703764)
        ref[3,0,2,1]=(-1.44488462603)/(o+1.)+(-0.0238956364351)
        ref[3,1,0,0]=(0.555753203891)/(o+1.)+(-0.193710946547)
        ref[3,1,0,1]=(-1.01153524773)/(o+1.)+(-0.310653653364)
        ref[3,1,1,0]=(-0.810081834072)/(o+1.)+(-0.704028056922)
        ref[3,1,1,1]=(0.673319109465)/(o+1.)+(0.146508313746)
        ref[3,1,2,0]=(1.06614330919)/(o+1.)+(0.210644257384)
        ref[3,1,2,1]=(0.169354687187)/(o+1.)+(0.260991208281)
        ref[3,2,0,0]=(1.17801438745)/(o+1.)+(-0.122603446615)
        ref[3,2,0,1]=(-0.612911386471)/(o+1.)+(0.512433833047)
        ref[3,2,1,0]=(0.441320249956)/(o+1.)+(-0.432537704584)
        ref[3,2,1,1]=(0.950667875863)/(o+1.)+(-0.476855090338)
        ref[3,2,2,0]=(0.112727596641)/(o+1.)+(-0.453766193797)
        ref[3,2,2,1]=(0.124962805486)/(o+1.)+(-0.0606318977858)
        ref[3,3,0,0]=(-0.866309983256)/(o+1.)+(-0.811326694594)
        ref[3,3,0,1]=(-0.277493641857)/(o+1.)+(0.564024429791)
        ref[3,3,1,0]=(-0.00865775952435)/(o+1.)+(-0.337712146277)
        ref[3,3,1,1]=(0.0813810922576)/(o+1.)+(0.903476356523)
        ref[3,3,2,0]=(0.162399658768)/(o+1.)+(-0.00649423977499)
        ref[3,3,2,1]=(0.482925749243)/(o+1.)+(0.141283097765)
        ref[3,4,0,0]=(-0.411914179659)/(o+1.)+(-0.331898184375)
        ref[3,4,0,1]=(0.532206771987)/(o+1.)+(0.0665350470083)
        ref[3,4,1,0]=(1.01606653498)/(o+1.)+(-0.228097514616)
        ref[3,4,1,1]=(0.37832857454)/(o+1.)+(0.136476416599)
        ref[3,4,2,0]=(-0.363036156393)/(o+1.)+(0.381760281949)
        ref[3,4,2,1]=(-0.578311466877)/(o+1.)+(-0.398821598136)
      else:
        arg[0,0,0,0]=(0.860676142569)*x[0]**o+(-0.156046991162)*x[0]+(-0.0462355579656)*x[1]**o+(0.70842844961)*x[1]+(0.617299123548)*x[2]**o+(-0.865174329565)*x[2]
        arg[0,0,0,1]=(-0.284308060865)*x[0]**o+(0.788848579389)*x[0]+(0.629321669042)*x[1]**o+(0.066637849809)*x[1]+(0.114152381009)*x[2]**o+(-0.466687745846)*x[2]
        arg[0,0,1,0]=(-0.690823198517)*x[0]**o+(-0.141510804573)*x[0]+(0.969978831967)*x[1]**o+(0.307926382658)*x[1]+(0.710373009094)*x[2]**o+(-0.494828315907)*x[2]
        arg[0,0,1,1]=(0.953361413913)*x[0]**o+(-0.103742611936)*x[0]+(0.63934955554)*x[1]**o+(-0.0778114476564)*x[1]+(-0.221185122698)*x[2]**o+(-0.287432047404)*x[2]
        arg[0,0,2,0]=(0.378306380558)*x[0]**o+(-0.117743922672)*x[0]+(-0.0606376630208)*x[1]**o+(-0.59439189505)*x[1]+(0.10507320657)*x[2]**o+(0.856411423067)*x[2]
        arg[0,0,2,1]=(0.777734672779)*x[0]**o+(0.286956002007)*x[0]+(-0.931811626583)*x[1]**o+(-0.806626659661)*x[1]+(-0.420390949282)*x[2]**o+(0.854080006578)*x[2]
        arg[0,1,0,0]=(0.403426948964)*x[0]**o+(0.707099419296)*x[0]+(0.593137130456)*x[1]**o+(-0.757560771598)*x[1]+(0.369632217934)*x[2]**o+(-0.348480285397)*x[2]
        arg[0,1,0,1]=(-0.491454189782)*x[0]**o+(0.547276640565)*x[0]+(-0.955045542708)*x[1]**o+(0.529989872531)*x[1]+(-0.161086874799)*x[2]**o+(-0.181810040724)*x[2]
        arg[0,1,1,0]=(0.46282177104)*x[0]**o+(0.254346937251)*x[0]+(0.824797105546)*x[1]**o+(-0.449979162246)*x[1]+(0.22307703071)*x[2]**o+(-0.592072135668)*x[2]
        arg[0,1,1,1]=(0.329230675141)*x[0]**o+(-0.733034697145)*x[0]+(0.762195609347)*x[1]**o+(-0.639097591362)*x[1]+(0.765354927905)*x[2]**o+(0.313571493405)*x[2]
        arg[0,1,2,0]=(0.3874127072)*x[0]**o+(-0.0569429727569)*x[0]+(-0.203121198708)*x[1]**o+(-0.736379499164)*x[1]+(-0.980629375436)*x[2]**o+(0.955793660048)*x[2]
        arg[0,1,2,1]=(-0.656255469455)*x[0]**o+(0.711904808699)*x[0]+(-0.363253678399)*x[1]**o+(0.921875000807)*x[1]+(0.686394816431)*x[2]**o+(0.818500716704)*x[2]
        arg[0,2,0,0]=(-0.931555347846)*x[0]**o+(-0.249573400742)*x[0]+(0.617166716404)*x[1]**o+(-0.929587051108)*x[1]+(0.830928715064)*x[2]**o+(0.461637431613)*x[2]
        arg[0,2,0,1]=(-0.8677719945)*x[0]**o+(-0.176892180445)*x[0]+(-0.881887456464)*x[1]**o+(0.923355677513)*x[1]+(-0.978544668453)*x[2]**o+(-0.427009384287)*x[2]
        arg[0,2,1,0]=(-0.0707937433084)*x[0]**o+(-0.872892141393)*x[0]+(0.606234798381)*x[1]**o+(0.935845076847)*x[1]+(-0.97559052201)*x[2]**o+(-0.31138518209)*x[2]
        arg[0,2,1,1]=(0.450210597865)*x[0]**o+(0.842939440378)*x[0]+(-0.333422856577)*x[1]**o+(-0.492809679443)*x[1]+(-0.254728278043)*x[2]**o+(-0.0702767077952)*x[2]
        arg[0,2,2,0]=(-0.250463216903)*x[0]**o+(0.756576293763)*x[0]+(0.596902629953)*x[1]**o+(-0.887080794947)*x[1]+(0.027319948008)*x[2]**o+(0.552496782776)*x[2]
        arg[0,2,2,1]=(-0.0883102140591)*x[0]**o+(0.988759610101)*x[0]+(-0.208152392508)*x[1]**o+(0.938999108564)*x[1]+(-0.773883611417)*x[2]**o+(0.653413306338)*x[2]
        arg[0,3,0,0]=(-0.399644211606)*x[0]**o+(-0.978595776407)*x[0]+(0.99454650986)*x[1]**o+(0.143313668715)*x[1]+(0.328592750842)*x[2]**o+(-0.958804581101)*x[2]
        arg[0,3,0,1]=(-0.0667966885959)*x[0]**o+(0.332653574704)*x[0]+(-0.762025150916)*x[1]**o+(0.895044013374)*x[1]+(0.627013843362)*x[2]**o+(-0.682392672232)*x[2]
        arg[0,3,1,0]=(-0.834479453773)*x[0]**o+(-0.375930859033)*x[0]+(0.129507966861)*x[1]**o+(0.995364280408)*x[1]+(-0.126081191572)*x[2]**o+(0.0344262779491)*x[2]
        arg[0,3,1,1]=(0.796952126596)*x[0]**o+(0.310562243091)*x[0]+(0.854548623788)*x[1]**o+(-0.719100537284)*x[1]+(-0.236266783438)*x[2]**o+(-0.646647129209)*x[2]
        arg[0,3,2,0]=(0.844073289115)*x[0]**o+(0.423300345347)*x[0]+(0.126354796852)*x[1]**o+(0.180372257606)*x[1]+(-0.564712722146)*x[2]**o+(0.466988430183)*x[2]
        arg[0,3,2,1]=(0.228775824367)*x[0]**o+(0.40701094831)*x[0]+(0.223101054593)*x[1]**o+(-0.894937905383)*x[1]+(0.425392942773)*x[2]**o+(-0.844238667169)*x[2]
        arg[0,4,0,0]=(0.3363848738)*x[0]**o+(-0.985327066062)*x[0]+(-0.107847971678)*x[1]**o+(0.596121801488)*x[1]+(0.655589801663)*x[2]**o+(-0.729650731949)*x[2]
        arg[0,4,0,1]=(-0.176435679219)*x[0]**o+(0.382558711725)*x[0]+(-0.658212492337)*x[1]**o+(0.584904267612)*x[1]+(-0.534684876705)*x[2]**o+(-0.441262614915)*x[2]
        arg[0,4,1,0]=(0.281301737167)*x[0]**o+(-0.642486274791)*x[0]+(0.420753884869)*x[1]**o+(0.482835955591)*x[1]+(0.169431333788)*x[2]**o+(0.922841193825)*x[2]
        arg[0,4,1,1]=(0.235812590604)*x[0]**o+(-0.446233855396)*x[0]+(0.465015868566)*x[1]**o+(-0.780441276888)*x[1]+(0.513632761351)*x[2]**o+(-0.0576482811678)*x[2]
        arg[0,4,2,0]=(-0.617245874462)*x[0]**o+(0.183661473466)*x[0]+(-0.711148911239)*x[1]**o+(0.125228991306)*x[1]+(-0.500218603959)*x[2]**o+(0.70597577229)*x[2]
        arg[0,4,2,1]=(0.56330304117)*x[0]**o+(0.631284986628)*x[0]+(-0.222485175776)*x[1]**o+(0.0639235737228)*x[1]+(-0.276613868747)*x[2]**o+(0.407488564857)*x[2]
        arg[1,0,0,0]=(0.159032146915)*x[0]**o+(-0.716040989693)*x[0]+(0.410709172613)*x[1]**o+(-0.527667963652)*x[1]+(-0.160515156497)*x[2]**o+(0.294448742382)*x[2]
        arg[1,0,0,1]=(0.941139035835)*x[0]**o+(0.707410046094)*x[0]+(-0.172095410803)*x[1]**o+(-0.967035204304)*x[1]+(0.870395272646)*x[2]**o+(0.830870986802)*x[2]
        arg[1,0,1,0]=(0.731015285309)*x[0]**o+(0.450519048697)*x[0]+(-0.484568714675)*x[1]**o+(0.454201220108)*x[1]+(-0.161850459099)*x[2]**o+(-0.617463263396)*x[2]
        arg[1,0,1,1]=(-0.958841628847)*x[0]**o+(-0.0667813400021)*x[0]+(0.659551780082)*x[1]**o+(0.514424015304)*x[1]+(0.920583608125)*x[2]**o+(0.89564899572)*x[2]
        arg[1,0,2,0]=(0.0482630479817)*x[0]**o+(0.911326064322)*x[0]+(-0.275134422092)*x[1]**o+(0.216495232459)*x[1]+(-0.973586171195)*x[2]**o+(0.275694688611)*x[2]
        arg[1,0,2,1]=(0.769733688293)*x[0]**o+(0.771829018421)*x[0]+(-0.22434549079)*x[1]**o+(0.131966885817)*x[1]+(-0.109424705679)*x[2]**o+(-0.369585136379)*x[2]
        arg[1,1,0,0]=(-0.0160180905877)*x[0]**o+(-0.268260755267)*x[0]+(-0.213940412022)*x[1]**o+(0.665987805279)*x[1]+(0.968879828017)*x[2]**o+(-0.449435538913)*x[2]
        arg[1,1,0,1]=(0.361055351825)*x[0]**o+(0.463294478488)*x[0]+(0.692655668486)*x[1]**o+(0.151415573206)*x[1]+(0.902280461515)*x[2]**o+(0.690926934073)*x[2]
        arg[1,1,1,0]=(0.591998602299)*x[0]**o+(0.516482177709)*x[0]+(-0.298511909182)*x[1]**o+(0.281818071894)*x[1]+(0.207089953979)*x[2]**o+(0.917004722388)*x[2]
        arg[1,1,1,1]=(0.479832408236)*x[0]**o+(-0.643374914165)*x[0]+(0.370844835249)*x[1]**o+(0.560458499239)*x[1]+(-0.865852319767)*x[2]**o+(0.560572130123)*x[2]
        arg[1,1,2,0]=(0.709293526635)*x[0]**o+(-0.388347223155)*x[0]+(0.60797330273)*x[1]**o+(-0.847624783099)*x[1]+(-0.0476367756382)*x[2]**o+(0.637841988478)*x[2]
        arg[1,1,2,1]=(0.343458045422)*x[0]**o+(-0.764451449587)*x[0]+(-0.444256801761)*x[1]**o+(0.385385997921)*x[1]+(-0.722095136671)*x[2]**o+(0.677852473207)*x[2]
        arg[1,2,0,0]=(0.710372887392)*x[0]**o+(-0.628494679119)*x[0]+(-0.59513507905)*x[1]**o+(-0.527824874295)*x[1]+(-0.611189170789)*x[2]**o+(0.090651976806)*x[2]
        arg[1,2,0,1]=(-0.980160150289)*x[0]**o+(-0.0286097581598)*x[0]+(-0.407906975953)*x[1]**o+(-0.319422835951)*x[1]+(-0.489787310147)*x[2]**o+(-0.694386341354)*x[2]
        arg[1,2,1,0]=(-0.416867224376)*x[0]**o+(0.631406877489)*x[0]+(-0.93939195942)*x[1]**o+(0.266858600313)*x[1]+(0.982173221612)*x[2]**o+(0.112901364617)*x[2]
        arg[1,2,1,1]=(0.412001480348)*x[0]**o+(0.624281489887)*x[0]+(0.306558457751)*x[1]**o+(0.95864454419)*x[1]+(0.299175627954)*x[2]**o+(-0.0531991017347)*x[2]
        arg[1,2,2,0]=(-0.0342261621741)*x[0]**o+(-0.587245704486)*x[0]+(-0.976407464107)*x[1]**o+(-0.0157215030322)*x[1]+(-0.529857472095)*x[2]**o+(-0.419626697673)*x[2]
        arg[1,2,2,1]=(-0.681737301037)*x[0]**o+(-0.980349460702)*x[0]+(-0.839469631393)*x[1]**o+(-0.412417814907)*x[1]+(0.451717398947)*x[2]**o+(0.962167685987)*x[2]
        arg[1,3,0,0]=(0.283453965913)*x[0]**o+(0.204445403787)*x[0]+(-0.806150376236)*x[1]**o+(0.122187094069)*x[1]+(-0.679291367964)*x[2]**o+(-0.875353966489)*x[2]
        arg[1,3,0,1]=(0.115974475175)*x[0]**o+(0.701473497654)*x[0]+(0.298012120304)*x[1]**o+(0.567597901709)*x[1]+(0.721211831476)*x[2]**o+(-0.255389572775)*x[2]
        arg[1,3,1,0]=(0.393891431092)*x[0]**o+(-0.563382537925)*x[0]+(0.974194829106)*x[1]**o+(0.659688559688)*x[1]+(-0.394266143828)*x[2]**o+(0.875266200191)*x[2]
        arg[1,3,1,1]=(-0.456930975697)*x[0]**o+(0.334590626231)*x[0]+(-0.227176883362)*x[1]**o+(0.76775699173)*x[1]+(0.839371167887)*x[2]**o+(-0.155473916326)*x[2]
        arg[1,3,2,0]=(0.346761981787)*x[0]**o+(-0.625121706322)*x[0]+(0.340655154638)*x[1]**o+(-0.0646323546999)*x[1]+(-0.211347402677)*x[2]**o+(0.233661266919)*x[2]
        arg[1,3,2,1]=(-0.086096644956)*x[0]**o+(-0.845623653209)*x[0]+(-0.0866093549957)*x[1]**o+(-0.467286349011)*x[1]+(-0.0621455515357)*x[2]**o+(0.164779979085)*x[2]
        arg[1,4,0,0]=(0.42022138208)*x[0]**o+(0.136984451355)*x[0]+(0.664770979086)*x[1]**o+(-0.537368896193)*x[1]+(0.874256726691)*x[2]**o+(-0.58614100824)*x[2]
        arg[1,4,0,1]=(-0.805860964792)*x[0]**o+(0.758511304044)*x[0]+(-0.574686863541)*x[1]**o+(-0.410873101247)*x[1]+(0.932015171551)*x[2]**o+(0.439987195351)*x[2]
        arg[1,4,1,0]=(-0.515187049123)*x[0]**o+(0.308078404334)*x[0]+(0.940959771774)*x[1]**o+(0.466269579691)*x[1]+(-0.190617144843)*x[2]**o+(0.738172533582)*x[2]
        arg[1,4,1,1]=(0.283326036586)*x[0]**o+(-0.693631747857)*x[0]+(0.478965080319)*x[1]**o+(-0.0465389396775)*x[1]+(-0.72886286959)*x[2]**o+(-0.9346424149)*x[2]
        arg[1,4,2,0]=(-0.71539766285)*x[0]**o+(0.361875987071)*x[0]+(0.905270678876)*x[1]**o+(-0.675024725702)*x[1]+(0.378259473658)*x[2]**o+(0.0520394226711)*x[2]
        arg[1,4,2,1]=(-0.319600525923)*x[0]**o+(0.0449956327134)*x[0]+(0.707418658521)*x[1]**o+(0.210137972776)*x[1]+(-0.767860728688)*x[2]**o+(0.232340825758)*x[2]
        arg[2,0,0,0]=(0.92844243651)*x[0]**o+(0.956810319504)*x[0]+(0.762198571194)*x[1]**o+(-0.363107634629)*x[1]+(0.0750393505254)*x[2]**o+(-0.43573581041)*x[2]
        arg[2,0,0,1]=(-0.193613121742)*x[0]**o+(0.440027871839)*x[0]+(-0.607905506346)*x[1]**o+(0.894840354958)*x[1]+(0.601279436178)*x[2]**o+(0.46806560762)*x[2]
        arg[2,0,1,0]=(0.370901517567)*x[0]**o+(-0.738769905647)*x[0]+(-0.698692979664)*x[1]**o+(-0.286318840335)*x[1]+(-0.879367315688)*x[2]**o+(-0.959912132991)*x[2]
        arg[2,0,1,1]=(-0.278939533774)*x[0]**o+(0.183795552426)*x[0]+(-0.341943551466)*x[1]**o+(0.491847260291)*x[1]+(-0.141030429384)*x[2]**o+(0.256287855539)*x[2]
        arg[2,0,2,0]=(-0.135930357616)*x[0]**o+(-0.382515749949)*x[0]+(-0.43030873581)*x[1]**o+(-0.0211891159981)*x[1]+(-0.377865120002)*x[2]**o+(0.109942325029)*x[2]
        arg[2,0,2,1]=(0.288827267201)*x[0]**o+(-0.907970544468)*x[0]+(-0.622258309213)*x[1]**o+(-0.763273879894)*x[1]+(0.438715013671)*x[2]**o+(-0.93873140988)*x[2]
        arg[2,1,0,0]=(-0.963487369906)*x[0]**o+(0.899051290288)*x[0]+(-0.268257956548)*x[1]**o+(-0.229092726584)*x[1]+(0.427024112314)*x[2]**o+(0.0435138095138)*x[2]
        arg[2,1,0,1]=(0.89996926597)*x[0]**o+(-0.206500706229)*x[0]+(-0.191769123095)*x[1]**o+(-0.348924155298)*x[1]+(-0.74405653776)*x[2]**o+(0.0724946691864)*x[2]
        arg[2,1,1,0]=(0.338170397901)*x[0]**o+(-0.725761081242)*x[0]+(-0.45928272676)*x[1]**o+(0.795535326783)*x[1]+(-0.316443844644)*x[2]**o+(-0.844205401892)*x[2]
        arg[2,1,1,1]=(0.509131757681)*x[0]**o+(0.319175801535)*x[0]+(-0.807058926279)*x[1]**o+(0.63928269996)*x[1]+(0.219324073432)*x[2]**o+(0.351217517434)*x[2]
        arg[2,1,2,0]=(-0.534090721611)*x[0]**o+(0.27146424294)*x[0]+(0.250778065716)*x[1]**o+(-0.901322994167)*x[1]+(-0.57588768699)*x[2]**o+(0.405043003456)*x[2]
        arg[2,1,2,1]=(0.101699906269)*x[0]**o+(0.361960318738)*x[0]+(0.278974923593)*x[1]**o+(-0.336289352037)*x[1]+(-0.677353936215)*x[2]**o+(0.0726793347753)*x[2]
        arg[2,2,0,0]=(-0.734961283823)*x[0]**o+(-0.593268557578)*x[0]+(0.310896858565)*x[1]**o+(-0.482326137951)*x[1]+(0.886472957099)*x[2]**o+(-0.0403563476581)*x[2]
        arg[2,2,0,1]=(-0.166071725106)*x[0]**o+(-0.454245686818)*x[0]+(-0.892674508991)*x[1]**o+(0.581554190251)*x[1]+(0.995323654908)*x[2]**o+(-0.0336714433349)*x[2]
        arg[2,2,1,0]=(-0.437491848099)*x[0]**o+(-0.195039508213)*x[0]+(0.822011086215)*x[1]**o+(0.771044412863)*x[1]+(0.697534261252)*x[2]**o+(0.666510217725)*x[2]
        arg[2,2,1,1]=(-0.529148276275)*x[0]**o+(-0.093016460024)*x[0]+(0.663641010649)*x[1]**o+(0.940439359843)*x[1]+(0.754151159716)*x[2]**o+(0.852349310591)*x[2]
        arg[2,2,2,0]=(0.859523244682)*x[0]**o+(0.704825125128)*x[0]+(0.221832129598)*x[1]**o+(0.573940020456)*x[1]+(0.935673751591)*x[2]**o+(0.23235170475)*x[2]
        arg[2,2,2,1]=(0.0359047473512)*x[0]**o+(0.836816935443)*x[0]+(0.6411608273)*x[1]**o+(0.291526953908)*x[1]+(0.735564515362)*x[2]**o+(-0.465781621769)*x[2]
        arg[2,3,0,0]=(0.352740040216)*x[0]**o+(0.54224020581)*x[0]+(0.681356174115)*x[1]**o+(-0.97589117869)*x[1]+(0.880755719884)*x[2]**o+(0.562259403362)*x[2]
        arg[2,3,0,1]=(0.563708471054)*x[0]**o+(0.525523219723)*x[0]+(0.423068360374)*x[1]**o+(0.29239160821)*x[1]+(0.874360368885)*x[2]**o+(0.922265851359)*x[2]
        arg[2,3,1,0]=(0.694309936305)*x[0]**o+(-0.409224769672)*x[0]+(-0.748835864803)*x[1]**o+(-0.21841534641)*x[1]+(0.497059735624)*x[2]**o+(-0.753234694148)*x[2]
        arg[2,3,1,1]=(0.109706510352)*x[0]**o+(-0.0160720403678)*x[0]+(0.888778080744)*x[1]**o+(-0.938692716268)*x[1]+(0.529262951478)*x[2]**o+(0.481430680214)*x[2]
        arg[2,3,2,0]=(0.65368659132)*x[0]**o+(-0.988176396392)*x[0]+(0.989546739333)*x[1]**o+(0.23178715846)*x[1]+(-0.557645438328)*x[2]**o+(0.063228513147)*x[2]
        arg[2,3,2,1]=(0.286316141831)*x[0]**o+(0.668138463424)*x[0]+(-0.950430671533)*x[1]**o+(-0.978012107187)*x[1]+(-0.107618382464)*x[2]**o+(-0.684084421276)*x[2]
        arg[2,4,0,0]=(-0.63510659432)*x[0]**o+(0.264242935995)*x[0]+(0.960647241624)*x[1]**o+(0.58483379289)*x[1]+(0.7255270271)*x[2]**o+(-0.476151653985)*x[2]
        arg[2,4,0,1]=(-0.221953908705)*x[0]**o+(-0.994143786509)*x[0]+(-0.359525421944)*x[1]**o+(0.533966197668)*x[1]+(-0.29739666047)*x[2]**o+(-0.0824204124466)*x[2]
        arg[2,4,1,0]=(-0.447043625871)*x[0]**o+(-0.904485300564)*x[0]+(0.26179071803)*x[1]**o+(0.437320719437)*x[1]+(-0.56501466909)*x[2]**o+(0.129662379166)*x[2]
        arg[2,4,1,1]=(0.259809756366)*x[0]**o+(0.446788808701)*x[0]+(-0.143702396023)*x[1]**o+(0.879531090832)*x[1]+(0.410047703156)*x[2]**o+(0.755750006468)*x[2]
        arg[2,4,2,0]=(-0.730881615951)*x[0]**o+(-0.357110792705)*x[0]+(0.552722654621)*x[1]**o+(0.965623848242)*x[1]+(0.555951256049)*x[2]**o+(0.259021370764)*x[2]
        arg[2,4,2,1]=(-0.465880467377)*x[0]**o+(-0.840984822217)*x[0]+(0.665817382039)*x[1]**o+(-0.390932961529)*x[1]+(-0.0857939888315)*x[2]**o+(-0.712185917672)*x[2]
        arg[3,0,0,0]=(0.710281429223)*x[0]**o+(-0.723903639057)*x[0]+(-0.267455509065)*x[1]**o+(-0.731684188612)*x[1]+(-0.287442751768)*x[2]**o+(-0.266871749947)*x[2]
        arg[3,0,0,1]=(0.460723870663)*x[0]**o+(0.0502154310124)*x[0]+(-0.260241363193)*x[1]**o+(-0.516473088648)*x[1]+(-0.925984113039)*x[2]**o+(-0.0297616649018)*x[2]
        arg[3,0,1,0]=(0.523186648278)*x[0]**o+(0.937042293367)*x[0]+(0.811740916657)*x[1]**o+(0.0544805175015)*x[1]+(-0.543484279955)*x[2]**o+(0.244714198306)*x[2]
        arg[3,0,1,1]=(-0.69212630167)*x[0]**o+(0.868708276385)*x[0]+(0.473290025811)*x[1]**o+(0.137043602294)*x[1]+(-0.19291314113)*x[2]**o+(-0.755147036188)*x[2]
        arg[3,0,2,0]=(0.0472638421122)*x[0]**o+(0.403135551064)*x[0]+(0.290083188289)*x[1]**o+(0.96419060449)*x[1]+(-0.240544184077)*x[2]**o+(-0.183028108529)*x[2]
        arg[3,0,2,1]=(0.354253383908)*x[0]**o+(0.628816528774)*x[0]+(0.588194429756)*x[1]**o+(-0.852128734457)*x[1]+(-0.84774778925)*x[2]**o+(0.796310123562)*x[2]
        arg[3,1,0,0]=(0.528880495806)*x[0]**o+(-0.74176589205)*x[0]+(-0.447816938371)*x[1]**o+(0.611297564802)*x[1]+(0.742134198482)*x[2]**o+(-0.625840664202)*x[2]
        arg[3,1,0,1]=(0.419633746259)*x[0]**o+(-0.247742274578)*x[0]+(0.302328885303)*x[1]**o+(-0.629854480502)*x[1]+(0.493919301288)*x[2]**o+(0.902910119017)*x[2]
        arg[3,1,1,0]=(0.509160481656)*x[0]**o+(-0.388129364133)*x[0]+(-0.595185230598)*x[1]**o+(-0.289592836865)*x[1]+(0.615663908134)*x[2]**o+(-0.560687711511)*x[2]
        arg[3,1,1,1]=(0.647782473568)*x[0]**o+(0.935840881725)*x[0]+(0.440348212287)*x[1]**o+(-0.489797988572)*x[1]+(0.766592631787)*x[2]**o+(0.217399649246)*x[2]
        arg[3,1,2,0]=(0.64944026906)*x[0]**o+(0.487409918475)*x[0]+(-0.0322210593784)*x[1]**o+(0.281311337861)*x[1]+(-0.694252407571)*x[2]**o+(-0.659678386121)*x[2]
        arg[3,1,2,1]=(0.806357278285)*x[0]**o+(0.419430374567)*x[0]+(-0.0525796600116)*x[1]**o+(-0.556886341914)*x[1]+(-0.288209814683)*x[2]**o+(-0.309696771495)*x[2]
        arg[3,2,0,0]=(-0.300806388527)*x[0]**o+(-0.39334416908)*x[0]+(0.000711617540648)*x[1]**o+(-0.618539958736)*x[1]+(0.0574098223152)*x[2]**o+(-0.351595384168)*x[2]
        arg[3,2,0,1]=(-0.898675793892)*x[0]**o+(0.904479958277)*x[0]+(0.0781945046018)*x[1]**o+(-0.0461053392586)*x[1]+(0.844798127031)*x[2]**o+(-0.82306838709)*x[2]
        arg[3,2,1,0]=(-0.268742335093)*x[0]**o+(-0.532639944208)*x[0]+(0.0101227109551)*x[1]**o+(-0.438312865518)*x[1]+(0.104288648246)*x[2]**o+(-0.115669027311)*x[2]
        arg[3,2,1,1]=(0.639871308792)*x[0]**o+(0.260847628514)*x[0]+(-0.455670719167)*x[1]**o+(0.210619035618)*x[1]+(-0.401809068119)*x[2]**o+(-0.0625194517026)*x[2]
        arg[3,2,2,0]=(-0.713684223829)*x[0]**o+(0.214281604766)*x[0]+(-0.60152592324)*x[1]**o+(0.249807228316)*x[1]+(-0.939395078261)*x[2]**o+(-0.519744396388)*x[2]
        arg[3,2,2,1]=(-0.340588993443)*x[0]**o+(-0.0165543854997)*x[0]+(-0.366479789338)*x[1]**o+(0.316347630987)*x[1]+(-0.457055798679)*x[2]**o+(-0.132368055553)*x[2]
        arg[3,3,0,0]=(-0.745211736715)*x[0]**o+(0.0106787556884)*x[0]+(-0.448180943446)*x[1]**o+(0.694982281891)*x[1]+(0.924431258221)*x[2]**o+(-0.841633806168)*x[2]
        arg[3,3,0,1]=(0.368054364362)*x[0]**o+(0.19288251992)*x[0]+(0.565854068035)*x[1]**o+(0.747534544754)*x[1]+(0.473165789569)*x[2]**o+(-0.59399986706)*x[2]
        arg[3,3,1,0]=(-0.142707671693)*x[0]**o+(0.177095043535)*x[0]+(-0.348946846847)*x[1]**o+(-0.545430013244)*x[1]+(-0.671405653136)*x[2]**o+(-0.531699969702)*x[2]
        arg[3,3,1,1]=(0.783201055631)*x[0]**o+(0.899495042323)*x[0]+(0.117802924059)*x[1]**o+(-0.277400900172)*x[1]+(-0.351268362305)*x[2]**o+(0.52756480897)*x[2]
        arg[3,3,2,0]=(-0.866758006698)*x[0]**o+(0.293139875238)*x[0]+(-0.318394231317)*x[1]**o+(-0.783701622124)*x[1]+(-0.95137399655)*x[2]**o+(-0.474198736639)*x[2]
        arg[3,3,2,1]=(0.0428398560619)*x[0]**o+(-0.970180601102)*x[0]+(-0.753897573393)*x[1]**o+(-0.0873352699251)*x[1]+(-0.131440532574)*x[2]**o+(0.875241605077)*x[2]
        arg[3,4,0,0]=(0.32399493643)*x[0]**o+(0.67295092524)*x[0]+(-0.29221110671)*x[1]**o+(0.118066079866)*x[1]+(-0.566541154667)*x[2]**o+(-0.806367499079)*x[2]
        arg[3,4,0,1]=(0.278776399755)*x[0]**o+(0.472327826919)*x[0]+(-0.956871861949)*x[1]**o+(0.109236874792)*x[1]+(-0.494095116883)*x[2]**o+(-0.194313775518)*x[2]
        arg[3,4,1,0]=(0.650920037415)*x[0]**o+(0.977698516929)*x[0]+(-0.0730777837998)*x[1]**o+(-0.576543152032)*x[1]+(0.781481934657)*x[2]**o+(-0.389328840554)*x[2]
        arg[3,4,1,1]=(-0.798341529522)*x[0]**o+(-0.630245234712)*x[0]+(-0.444550323489)*x[1]**o+(-0.539189248295)*x[1]+(-0.904964423673)*x[2]**o+(-0.611560462796)*x[2]
        arg[3,4,2,0]=(-0.392991417685)*x[0]**o+(0.313599061567)*x[0]+(-0.215747337398)*x[1]**o+(-0.310316872476)*x[1]+(-0.269452617785)*x[2]**o+(0.684353794645)*x[2]
        arg[3,4,2,1]=(-0.586668781403)*x[0]**o+(-0.0711284937177)*x[0]+(0.408273850302)*x[1]**o+(0.761860299288)*x[1]+(0.902079242449)*x[2]**o+(0.70161017722)*x[2]
        ref[0,0,0,0]=(1.43173970815)/(o+1.)+(-0.156396435558)
        ref[0,0,0,1]=(0.459165989186)/(o+1.)+(0.194399341676)
        ref[0,0,1,0]=(0.989528642545)/(o+1.)+(-0.164206368911)
        ref[0,0,1,1]=(1.37152584676)/(o+1.)+(-0.234493053498)
        ref[0,0,2,0]=(0.422741924108)/(o+1.)+(0.0721378026728)
        ref[0,0,2,1]=(-0.574467903086)/(o+1.)+(0.167204674462)
        ref[0,1,0,0]=(1.36619629735)/(o+1.)+(-0.19947081885)
        ref[0,1,0,1]=(-1.60758660729)/(o+1.)+(0.447728236186)
        ref[0,1,1,0]=(1.5106959073)/(o+1.)+(-0.393852180331)
        ref[0,1,1,1]=(1.85678121239)/(o+1.)+(-0.529280397551)
        ref[0,1,2,0]=(-0.796337866944)/(o+1.)+(0.0812355940636)
        ref[0,1,2,1]=(-0.333114331422)/(o+1.)+(1.22614026311)
        ref[0,2,0,0]=(0.516540083621)/(o+1.)+(-0.358761510118)
        ref[0,2,0,1]=(-2.72820411942)/(o+1.)+(0.15972705639)
        ref[0,2,1,0]=(-0.440149466937)/(o+1.)+(-0.124216123318)
        ref[0,2,1,1]=(-0.137940536755)/(o+1.)+(0.13992652657)
        ref[0,2,2,0]=(0.373759361058)/(o+1.)+(0.210996140796)
        ref[0,2,2,1]=(-1.07034621798)/(o+1.)+(1.2905860125)
        ref[0,3,0,0]=(0.923495049095)/(o+1.)+(-0.897043344396)
        ref[0,3,0,1]=(-0.20180799615)/(o+1.)+(0.272652457923)
        ref[0,3,1,0]=(-0.831052678484)/(o+1.)+(0.326929849662)
        ref[0,3,1,1]=(1.41523396695)/(o+1.)+(-0.5275927117)
        ref[0,3,2,0]=(0.405715363821)/(o+1.)+(0.535330516568)
        ref[0,3,2,1]=(0.877269821733)/(o+1.)+(-0.666082812121)
        ref[0,4,0,0]=(0.884126703786)/(o+1.)+(-0.559427998262)
        ref[0,4,0,1]=(-1.36933304826)/(o+1.)+(0.263100182211)
        ref[0,4,1,0]=(0.871486955825)/(o+1.)+(0.381595437312)
        ref[0,4,1,1]=(1.21446122052)/(o+1.)+(-0.642161706726)
        ref[0,4,2,0]=(-1.82861338966)/(o+1.)+(0.507433118531)
        ref[0,4,2,1]=(0.0642039966469)/(o+1.)+(0.551348562604)
        ref[1,0,0,0]=(0.40922616303)/(o+1.)+(-0.474630105482)
        ref[1,0,0,1]=(1.63943889768)/(o+1.)+(0.285622914296)
        ref[1,0,1,0]=(0.0845961115353)/(o+1.)+(0.143628502705)
        ref[1,0,1,1]=(0.62129375936)/(o+1.)+(0.671645835511)
        ref[1,0,2,0]=(-1.2004575453)/(o+1.)+(0.701757992696)
        ref[1,0,2,1]=(0.435963491825)/(o+1.)+(0.26710538393)
        ref[1,1,0,0]=(0.738921325407)/(o+1.)+(-0.0258542444502)
        ref[1,1,0,1]=(1.95599148183)/(o+1.)+(0.652818492883)
        ref[1,1,1,0]=(0.500576647095)/(o+1.)+(0.857652485995)
        ref[1,1,1,1]=(-0.0151750762821)/(o+1.)+(0.238827857598)
        ref[1,1,2,0]=(1.26963005373)/(o+1.)+(-0.299065008888)
        ref[1,1,2,1]=(-0.82289389301)/(o+1.)+(0.14939351077)
        ref[1,2,0,0]=(-0.495951362447)/(o+1.)+(-0.532833788304)
        ref[1,2,0,1]=(-1.87785443639)/(o+1.)+(-0.521209467732)
        ref[1,2,1,0]=(-0.374085962184)/(o+1.)+(0.505583421209)
        ref[1,2,1,1]=(1.01773556605)/(o+1.)+(0.764863466171)
        ref[1,2,2,0]=(-1.54049109838)/(o+1.)+(-0.511296952595)
        ref[1,2,2,1]=(-1.06948953348)/(o+1.)+(-0.215299794811)
        ref[1,3,0,0]=(-1.20198777829)/(o+1.)+(-0.274360734317)
        ref[1,3,0,1]=(1.13519842695)/(o+1.)+(0.506840913294)
        ref[1,3,1,0]=(0.973820116369)/(o+1.)+(0.485786110977)
        ref[1,3,1,1]=(0.155263308828)/(o+1.)+(0.473436850817)
        ref[1,3,2,0]=(0.476069733748)/(o+1.)+(-0.228046397052)
        ref[1,3,2,1]=(-0.234851551487)/(o+1.)+(-0.574065011567)
        ref[1,4,0,0]=(1.95924908786)/(o+1.)+(-0.493262726539)
        ref[1,4,0,1]=(-0.448532656782)/(o+1.)+(0.393812699074)
        ref[1,4,1,0]=(0.235155577808)/(o+1.)+(0.756260258803)
        ref[1,4,1,1]=(0.0334282473144)/(o+1.)+(-0.837406551217)
        ref[1,4,2,0]=(0.568132489684)/(o+1.)+(-0.13055465798)
        ref[1,4,2,1]=(-0.380042596089)/(o+1.)+(0.243737215624)
        ref[2,0,0,0]=(1.76568035823)/(o+1.)+(0.0789834372326)
        ref[2,0,0,1]=(-0.20023919191)/(o+1.)+(0.901466917208)
        ref[2,0,1,0]=(-1.20715877778)/(o+1.)+(-0.992500439486)
        ref[2,0,1,1]=(-0.761913514624)/(o+1.)+(0.465965334128)
        ref[2,0,2,0]=(-0.944104213428)/(o+1.)+(-0.146881270459)
        ref[2,0,2,1]=(0.105283971659)/(o+1.)+(-1.30498791712)
        ref[2,1,0,0]=(-0.804721214141)/(o+1.)+(0.356736186609)
        ref[2,1,0,1]=(-0.0358563948854)/(o+1.)+(-0.24146509617)
        ref[2,1,1,0]=(-0.437556173502)/(o+1.)+(-0.387215578175)
        ref[2,1,1,1]=(-0.0786030951662)/(o+1.)+(0.654838009465)
        ref[2,1,2,0]=(-0.859200342885)/(o+1.)+(-0.112407873885)
        ref[2,1,2,1]=(-0.296679106354)/(o+1.)+(0.0491751507381)
        ref[2,2,0,0]=(0.462408531841)/(o+1.)+(-0.557975521593)
        ref[2,2,0,1]=(-0.0634225791893)/(o+1.)+(0.0468185300489)
        ref[2,2,1,0]=(1.08205349937)/(o+1.)+(0.621257561187)
        ref[2,2,1,1]=(0.88864389409)/(o+1.)+(0.849886105205)
        ref[2,2,2,0]=(2.01702912587)/(o+1.)+(0.755558425167)
        ref[2,2,2,1]=(1.41263009001)/(o+1.)+(0.331281133791)
        ref[2,3,0,0]=(1.91485193421)/(o+1.)+(0.0643042152414)
        ref[2,3,0,1]=(1.86113720031)/(o+1.)+(0.870090339646)
        ref[2,3,1,0]=(0.442533807126)/(o+1.)+(-0.690437405116)
        ref[2,3,1,1]=(1.52774754257)/(o+1.)+(-0.236667038211)
        ref[2,3,2,0]=(1.08558789233)/(o+1.)+(-0.346580362392)
        ref[2,3,2,1]=(-0.771732912166)/(o+1.)+(-0.49697903252)
        ref[2,4,0,0]=(1.0510676744)/(o+1.)+(0.18646253745)
        ref[2,4,0,1]=(-0.878875991119)/(o+1.)+(-0.271299000644)
        ref[2,4,1,0]=(-0.75026757693)/(o+1.)+(-0.168751100981)
        ref[2,4,1,1]=(0.526155063499)/(o+1.)+(1.041034953)
        ref[2,4,2,0]=(0.377792294719)/(o+1.)+(0.433767213151)
        ref[2,4,2,1]=(0.114142925831)/(o+1.)+(-0.972051850709)
        ref[3,0,0,0]=(0.15538316839)/(o+1.)+(-0.861229788808)
        ref[3,0,0,1]=(-0.725501605569)/(o+1.)+(-0.248009661269)
        ref[3,0,1,0]=(0.791443284981)/(o+1.)+(0.618118504587)
        ref[3,0,1,1]=(-0.411749416988)/(o+1.)+(0.125302421245)
        ref[3,0,2,0]=(0.0968028463245)/(o+1.)+(0.592149023513)
        ref[3,0,2,1]=(0.0947000244134)/(o+1.)+(0.286498958939)
        ref[3,1,0,0]=(0.823197755916)/(o+1.)+(-0.378154495725)
        ref[3,1,0,1]=(1.21588193285)/(o+1.)+(0.0126566819682)
        ref[3,1,1,0]=(0.529639159192)/(o+1.)+(-0.619204956255)
        ref[3,1,1,1]=(1.85472331764)/(o+1.)+(0.3317212712)
        ref[3,1,2,0]=(-0.0770331978898)/(o+1.)+(0.0545214351076)
        ref[3,1,2,1]=(0.46556780359)/(o+1.)+(-0.223576369421)
        ref[3,2,0,0]=(-0.242684948671)/(o+1.)+(-0.681739755992)
        ref[3,2,0,1]=(0.0243168377401)/(o+1.)+(0.0176531159642)
        ref[3,2,1,0]=(-0.154330975892)/(o+1.)+(-0.543310918519)
        ref[3,2,1,1]=(-0.217608478494)/(o+1.)+(0.204473606215)
        ref[3,2,2,0]=(-2.25460522533)/(o+1.)+(-0.0278277816531)
        ref[3,2,2,1]=(-1.16412458146)/(o+1.)+(0.0837125949668)
        ref[3,3,0,0]=(-0.26896142194)/(o+1.)+(-0.0679863842941)
        ref[3,3,0,1]=(1.40707422197)/(o+1.)+(0.173208598807)
        ref[3,3,1,0]=(-1.16306017168)/(o+1.)+(-0.450017469706)
        ref[3,3,1,1]=(0.549735617385)/(o+1.)+(0.57482947556)
        ref[3,3,2,0]=(-2.13652623456)/(o+1.)+(-0.482380241762)
        ref[3,3,2,1]=(-0.842498249905)/(o+1.)+(-0.0911371329748)
        ref[3,4,0,0]=(-0.534757324947)/(o+1.)+(-0.00767524698684)
        ref[3,4,0,1]=(-1.17219057908)/(o+1.)+(0.193625463097)
        ref[3,4,1,0]=(1.35932418827)/(o+1.)+(0.00591326217154)
        ref[3,4,1,1]=(-2.14785627668)/(o+1.)+(-0.890497472901)
        ref[3,4,2,0]=(-0.878191372868)/(o+1.)+(0.343817991868)
        ref[3,4,2,1]=(0.723684311348)/(o+1.)+(0.696170991395)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.217470460227)*x[0]**o+(0.664132905811)*x[0]+(0.493878617743)*x[1]**o+(0.111327715714)*x[1]
        ref=(0.276408157516)/(o+1.)+(0.387730310763)
      else:
        arg=(0.371937838886)*x[0]**o+(0.653846807778)*x[0]+(0.172358058188)*x[1]**o+(-0.928017916614)*x[1]+(0.116156616174)*x[2]**o+(0.371332501889)*x[2]
        ref=(0.660452513249)/(o+1.)+(0.0485806965265)
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.329952046701)*x[0]**o+(0.12757022314)*x[0]+(0.198280362833)*x[1]**o+(0.129179854334)*x[1]
        arg[1]=(0.185895161332)*x[0]**o+(0.270740935936)*x[0]+(0.630122833594)*x[1]**o+(-0.0696403698787)*x[1]
        ref[0]=(-0.131671683869)/(o+1.)+(0.128375038737)
        ref[1]=(0.816017994925)/(o+1.)+(0.100550283029)
      else:
        arg[0]=(0.526257699936)*x[0]**o+(0.0731442796396)*x[0]+(-0.266121642119)*x[1]**o+(-0.929478242136)*x[1]+(0.743092861949)*x[2]**o+(0.22690419858)*x[2]
        arg[1]=(-0.121387039147)*x[0]**o+(-0.0580608263784)*x[0]+(0.761164059949)*x[1]**o+(0.821938972431)*x[1]+(-0.756423175962)*x[2]**o+(0.213207132353)*x[2]
        ref[0]=(1.00322891977)/(o+1.)+(-0.314714881958)
        ref[1]=(-0.11664615516)/(o+1.)+(0.488542639203)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.963318847499)*x[0]**o+(-0.184977238308)*x[0]+(0.153633231879)*x[1]**o+(-0.640028315216)*x[1]
        arg[0,1]=(0.0879103656102)*x[0]**o+(-0.928854425726)*x[0]+(-0.138566176095)*x[1]**o+(-0.670118398936)*x[1]
        arg[0,2]=(0.920856326204)*x[0]**o+(0.461047594595)*x[0]+(-0.914778320363)*x[1]**o+(-0.561644217568)*x[1]
        arg[0,3]=(0.506148289443)*x[0]**o+(-0.474044343536)*x[0]+(0.742127575836)*x[1]**o+(-0.52349314159)*x[1]
        arg[0,4]=(0.0810224543145)*x[0]**o+(-0.307992483498)*x[0]+(0.749673245291)*x[1]**o+(-0.554099873044)*x[1]
        arg[1,0]=(-0.270897099553)*x[0]**o+(-0.222178461886)*x[0]+(0.755729049844)*x[1]**o+(0.740743777894)*x[1]
        arg[1,1]=(0.0220934176819)*x[0]**o+(-0.309641658323)*x[0]+(-0.951833592348)*x[1]**o+(0.430554526918)*x[1]
        arg[1,2]=(0.176033311218)*x[0]**o+(0.728188001962)*x[0]+(-0.428357954006)*x[1]**o+(0.955059500699)*x[1]
        arg[1,3]=(0.946680422474)*x[0]**o+(-0.251968148743)*x[0]+(-0.3704600461)*x[1]**o+(0.692080922297)*x[1]
        arg[1,4]=(0.122655375591)*x[0]**o+(0.617285886533)*x[0]+(0.729184247986)*x[1]**o+(0.30369620264)*x[1]
        arg[2,0]=(-0.957780501271)*x[0]**o+(0.438290429468)*x[0]+(-0.504841431279)*x[1]**o+(0.935350834608)*x[1]
        arg[2,1]=(-0.49410765407)*x[0]**o+(-0.635142349002)*x[0]+(0.700131616572)*x[1]**o+(0.99523209861)*x[1]
        arg[2,2]=(0.0967686936303)*x[0]**o+(0.987754073381)*x[0]+(-0.161220173386)*x[1]**o+(-0.343926097784)*x[1]
        arg[2,3]=(0.560040886628)*x[0]**o+(-0.843778720323)*x[0]+(0.345354034369)*x[1]**o+(0.440584543858)*x[1]
        arg[2,4]=(0.988977815245)*x[0]**o+(0.661922598357)*x[0]+(-0.425421002549)*x[1]**o+(0.624279872655)*x[1]
        arg[3,0]=(-0.997217068098)*x[0]**o+(-0.973497091595)*x[0]+(-0.0980683851598)*x[1]**o+(0.408081557041)*x[1]
        arg[3,1]=(0.0827616522414)*x[0]**o+(0.931022772362)*x[0]+(-0.588688206358)*x[1]**o+(0.900467348992)*x[1]
        arg[3,2]=(0.0933642874383)*x[0]**o+(0.398732664158)*x[0]+(-0.437102212783)*x[1]**o+(-0.126538302466)*x[1]
        arg[3,3]=(0.251989644486)*x[0]**o+(0.765809815972)*x[0]+(-0.325455283605)*x[1]**o+(0.171819372538)*x[1]
        arg[3,4]=(0.649407682145)*x[0]**o+(0.963776897105)*x[0]+(0.671915752131)*x[1]**o+(-0.468729629153)*x[1]
        ref[0,0]=(-0.80968561562)/(o+1.)+(-0.412502776762)
        ref[0,1]=(-0.0506558104848)/(o+1.)+(-0.799486412331)
        ref[0,2]=(0.00607800584081)/(o+1.)+(-0.0502983114867)
        ref[0,3]=(1.24827586528)/(o+1.)+(-0.498768742563)
        ref[0,4]=(0.830695699605)/(o+1.)+(-0.431046178271)
        ref[1,0]=(0.484831950292)/(o+1.)+(0.259282658004)
        ref[1,1]=(-0.929740174666)/(o+1.)+(0.0604564342974)
        ref[1,2]=(-0.252324642788)/(o+1.)+(0.84162375133)
        ref[1,3]=(0.576220376373)/(o+1.)+(0.220056386777)
        ref[1,4]=(0.851839623577)/(o+1.)+(0.460491044586)
        ref[2,0]=(-1.46262193255)/(o+1.)+(0.686820632038)
        ref[2,1]=(0.206023962501)/(o+1.)+(0.180044874804)
        ref[2,2]=(-0.0644514797559)/(o+1.)+(0.321913987798)
        ref[2,3]=(0.905394920997)/(o+1.)+(-0.201597088233)
        ref[2,4]=(0.563556812695)/(o+1.)+(0.643101235506)
        ref[3,0]=(-1.09528545326)/(o+1.)+(-0.282707767277)
        ref[3,1]=(-0.505926554117)/(o+1.)+(0.915745060677)
        ref[3,2]=(-0.343737925345)/(o+1.)+(0.136097180846)
        ref[3,3]=(-0.0734656391194)/(o+1.)+(0.468814594255)
        ref[3,4]=(1.32132343428)/(o+1.)+(0.247523633976)
      else:
        arg[0,0]=(-0.512698606017)*x[0]**o+(-0.156853163498)*x[0]+(0.247997355633)*x[1]**o+(-0.363003068904)*x[1]+(-0.940097989108)*x[2]**o+(-0.661063850288)*x[2]
        arg[0,1]=(-0.886390181577)*x[0]**o+(0.724049511142)*x[0]+(0.260684592163)*x[1]**o+(0.265648885034)*x[1]+(0.123084281693)*x[2]**o+(0.181492979151)*x[2]
        arg[0,2]=(-0.570543998851)*x[0]**o+(0.0541041165388)*x[0]+(0.630994990544)*x[1]**o+(-0.89311079559)*x[1]+(0.750714073617)*x[2]**o+(-0.0684796380527)*x[2]
        arg[0,3]=(0.605486031268)*x[0]**o+(0.0387617487672)*x[0]+(-0.365027101406)*x[1]**o+(0.323275180514)*x[1]+(-0.763142660003)*x[2]**o+(-0.331069480921)*x[2]
        arg[0,4]=(-0.0779035848472)*x[0]**o+(-0.0234904624931)*x[0]+(0.591635786678)*x[1]**o+(-0.793680126332)*x[1]+(-0.874742296757)*x[2]**o+(-0.184528885157)*x[2]
        arg[1,0]=(-0.62480972544)*x[0]**o+(-0.99242215547)*x[0]+(0.634250679117)*x[1]**o+(-0.785448642692)*x[1]+(-0.013335654402)*x[2]**o+(0.52112750417)*x[2]
        arg[1,1]=(0.611541580179)*x[0]**o+(-0.203144044802)*x[0]+(-0.411277184995)*x[1]**o+(0.373487145939)*x[1]+(-0.679339345026)*x[2]**o+(-0.632413536218)*x[2]
        arg[1,2]=(-0.258927016205)*x[0]**o+(0.0242628959926)*x[0]+(-0.596909760383)*x[1]**o+(-0.295787354787)*x[1]+(-0.337930022823)*x[2]**o+(-0.825085100446)*x[2]
        arg[1,3]=(-0.0976663781507)*x[0]**o+(0.716600132838)*x[0]+(-0.682155341898)*x[1]**o+(-0.154484864167)*x[1]+(0.754686874908)*x[2]**o+(-0.102453470268)*x[2]
        arg[1,4]=(0.268714339964)*x[0]**o+(0.974234398844)*x[0]+(-0.370113158384)*x[1]**o+(0.0745228956861)*x[1]+(-0.0674370676047)*x[2]**o+(0.132605448278)*x[2]
        arg[2,0]=(-0.00750893262562)*x[0]**o+(0.507641887134)*x[0]+(0.638483932117)*x[1]**o+(-0.940295540912)*x[1]+(-0.201492519698)*x[2]**o+(0.504878595818)*x[2]
        arg[2,1]=(-0.0283522980593)*x[0]**o+(-0.76577346605)*x[0]+(0.0863472830598)*x[1]**o+(-0.904447859187)*x[1]+(-0.203390005492)*x[2]**o+(-0.741629136645)*x[2]
        arg[2,2]=(-0.0196123831667)*x[0]**o+(-0.320721372799)*x[0]+(0.0670421287749)*x[1]**o+(-0.415029682416)*x[1]+(0.450269143475)*x[2]**o+(-0.735201210502)*x[2]
        arg[2,3]=(-0.621537330758)*x[0]**o+(0.21573881897)*x[0]+(0.0364275955835)*x[1]**o+(0.873908952751)*x[1]+(-0.197046790303)*x[2]**o+(0.655654722165)*x[2]
        arg[2,4]=(0.432531007791)*x[0]**o+(-0.65605251166)*x[0]+(0.728245204833)*x[1]**o+(-0.769998936989)*x[1]+(-0.797879623133)*x[2]**o+(0.36575554332)*x[2]
        arg[3,0]=(0.512219734043)*x[0]**o+(-0.177942735603)*x[0]+(-0.307293350237)*x[1]**o+(-0.780181541911)*x[1]+(-0.510201727643)*x[2]**o+(-0.567980715525)*x[2]
        arg[3,1]=(-0.405123989366)*x[0]**o+(-0.133316566583)*x[0]+(0.552868689995)*x[1]**o+(-0.921149434884)*x[1]+(0.510851078054)*x[2]**o+(0.7207533866)*x[2]
        arg[3,2]=(0.150011788901)*x[0]**o+(-0.55947555552)*x[0]+(-0.353366762442)*x[1]**o+(0.204954197301)*x[1]+(-0.48254632377)*x[2]**o+(-0.272738900127)*x[2]
        arg[3,3]=(-0.690981207461)*x[0]**o+(0.38683432128)*x[0]+(0.655731808738)*x[1]**o+(0.162269870301)*x[1]+(-0.373468659959)*x[2]**o+(-0.841431521009)*x[2]
        arg[3,4]=(0.156455675319)*x[0]**o+(0.501690084272)*x[0]+(-0.339255791314)*x[1]**o+(0.0832260784264)*x[1]+(0.394137401365)*x[2]**o+(-0.92403906326)*x[2]
        ref[0,0]=(-1.20479923949)/(o+1.)+(-0.590460041345)
        ref[0,1]=(-0.502621307721)/(o+1.)+(0.585595687663)
        ref[0,2]=(0.811165065311)/(o+1.)+(-0.453743158552)
        ref[0,3]=(-0.522683730141)/(o+1.)+(0.0154837241797)
        ref[0,4]=(-0.361010094926)/(o+1.)+(-0.500849736991)
        ref[1,0]=(-0.00389470072479)/(o+1.)+(-0.628371646996)
        ref[1,1]=(-0.479074949842)/(o+1.)+(-0.23103521754)
        ref[1,2]=(-1.19376679941)/(o+1.)+(-0.54830477962)
        ref[1,3]=(-0.0251348451402)/(o+1.)+(0.229830899202)
        ref[1,4]=(-0.168835886025)/(o+1.)+(0.590681371404)
        ref[2,0]=(0.429482479793)/(o+1.)+(0.0361124710203)
        ref[2,1]=(-0.145395020492)/(o+1.)+(-1.20592523094)
        ref[2,2]=(0.497698889084)/(o+1.)+(-0.735476132859)
        ref[2,3]=(-0.782156525477)/(o+1.)+(0.872651246943)
        ref[2,4]=(0.362896589491)/(o+1.)+(-0.530147952664)
        ref[3,0]=(-0.305275343837)/(o+1.)+(-0.763052496519)
        ref[3,1]=(0.658595778683)/(o+1.)+(-0.166856307434)
        ref[3,2]=(-0.685901297311)/(o+1.)+(-0.313630129173)
        ref[3,3]=(-0.408718058683)/(o+1.)+(-0.146163664714)
        ref[3,4]=(0.21133728537)/(o+1.)+(-0.169561450281)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.165423911829)*x[0]**o+(0.912542062064)*x[0]+(0.515840721225)*x[1]**o+(-0.855922021017)*x[1]
        arg[0,0,1]=(0.591873802546)*x[0]**o+(-0.039385167187)*x[0]+(0.0261484884274)*x[1]**o+(0.227950240434)*x[1]
        arg[0,1,0]=(-0.825833712046)*x[0]**o+(0.989606797765)*x[0]+(0.975260558791)*x[1]**o+(0.297963603984)*x[1]
        arg[0,1,1]=(-0.658054502482)*x[0]**o+(-0.865505090963)*x[0]+(0.817184084453)*x[1]**o+(-0.409588098899)*x[1]
        arg[1,0,0]=(0.538456987198)*x[0]**o+(-0.0768787509391)*x[0]+(0.424620431835)*x[1]**o+(0.799917948831)*x[1]
        arg[1,0,1]=(0.125945558199)*x[0]**o+(-0.48320049996)*x[0]+(0.105474962327)*x[1]**o+(-0.090713555582)*x[1]
        arg[1,1,0]=(0.835213375418)*x[0]**o+(-0.513407038769)*x[0]+(-0.342890811735)*x[1]**o+(-0.285057736244)*x[1]
        arg[1,1,1]=(0.762782085749)*x[0]**o+(-0.308457087352)*x[0]+(0.23756726753)*x[1]**o+(-0.909365956504)*x[1]
        arg[2,0,0]=(-0.76467980678)*x[0]**o+(-0.265552922006)*x[0]+(-0.261862355539)*x[1]**o+(-0.399407404618)*x[1]
        arg[2,0,1]=(0.79401193617)*x[0]**o+(-0.0300461495216)*x[0]+(-0.978834133508)*x[1]**o+(-0.984845382599)*x[1]
        arg[2,1,0]=(0.898033018534)*x[0]**o+(-0.184513430159)*x[0]+(0.0430634098526)*x[1]**o+(0.62213313954)*x[1]
        arg[2,1,1]=(0.126926439246)*x[0]**o+(0.671259428917)*x[0]+(-0.396062855417)*x[1]**o+(0.0461825691916)*x[1]
        arg[3,0,0]=(0.185756594279)*x[0]**o+(0.570946291973)*x[0]+(0.0673731944956)*x[1]**o+(0.367766610292)*x[1]
        arg[3,0,1]=(0.021269224946)*x[0]**o+(-0.463663279063)*x[0]+(-0.687930988068)*x[1]**o+(0.184714204849)*x[1]
        arg[3,1,0]=(0.754288639303)*x[0]**o+(-0.94612827517)*x[0]+(0.802463196709)*x[1]**o+(0.202355582077)*x[1]
        arg[3,1,1]=(0.840002857786)*x[0]**o+(-0.692207739447)*x[0]+(0.486569391181)*x[1]**o+(-0.713960793192)*x[1]
        arg[4,0,0]=(-0.205794539655)*x[0]**o+(-0.930458992533)*x[0]+(-0.62580205172)*x[1]**o+(-0.884738110673)*x[1]
        arg[4,0,1]=(-0.91433612349)*x[0]**o+(0.694961527804)*x[0]+(-0.510941334746)*x[1]**o+(-0.510725271706)*x[1]
        arg[4,1,0]=(0.852785499366)*x[0]**o+(0.870826628851)*x[0]+(-0.577529754198)*x[1]**o+(-0.583793684857)*x[1]
        arg[4,1,1]=(-0.604869226881)*x[0]**o+(0.482463572402)*x[0]+(0.829654316885)*x[1]**o+(0.597543822455)*x[1]
        arg[5,0,0]=(0.413459469749)*x[0]**o+(0.865143264083)*x[0]+(0.986197644249)*x[1]**o+(-0.2511021346)*x[1]
        arg[5,0,1]=(0.811572168213)*x[0]**o+(-0.110121334437)*x[0]+(-0.985492232242)*x[1]**o+(-0.913436084156)*x[1]
        arg[5,1,0]=(-0.660869051422)*x[0]**o+(-0.743235620427)*x[0]+(0.205393778238)*x[1]**o+(0.296069404687)*x[1]
        arg[5,1,1]=(-0.13285011267)*x[0]**o+(0.659400366834)*x[0]+(-0.889984368942)*x[1]**o+(0.584893563559)*x[1]
        ref[0,0,0]=(0.350416809396)/(o+1.)+(0.0283100205234)
        ref[0,0,1]=(0.618022290973)/(o+1.)+(0.0942825366236)
        ref[0,1,0]=(0.149426846745)/(o+1.)+(0.643785200874)
        ref[0,1,1]=(0.159129581971)/(o+1.)+(-0.637546594931)
        ref[1,0,0]=(0.963077419032)/(o+1.)+(0.361519598946)
        ref[1,0,1]=(0.231420520526)/(o+1.)+(-0.286957027771)
        ref[1,1,0]=(0.492322563683)/(o+1.)+(-0.399232387506)
        ref[1,1,1]=(1.00034935328)/(o+1.)+(-0.608911521928)
        ref[2,0,0]=(-1.02654216232)/(o+1.)+(-0.332480163312)
        ref[2,0,1]=(-0.184822197338)/(o+1.)+(-0.507445766061)
        ref[2,1,0]=(0.941096428387)/(o+1.)+(0.21880985469)
        ref[2,1,1]=(-0.269136416172)/(o+1.)+(0.358720999054)
        ref[3,0,0]=(0.253129788774)/(o+1.)+(0.469356451133)
        ref[3,0,1]=(-0.666661763122)/(o+1.)+(-0.139474537107)
        ref[3,1,0]=(1.55675183601)/(o+1.)+(-0.371886346546)
        ref[3,1,1]=(1.32657224897)/(o+1.)+(-0.703084266319)
        ref[4,0,0]=(-0.831596591375)/(o+1.)+(-0.907598551603)
        ref[4,0,1]=(-1.42527745824)/(o+1.)+(0.0921181280487)
        ref[4,1,0]=(0.275255745168)/(o+1.)+(0.143516471997)
        ref[4,1,1]=(0.224785090004)/(o+1.)+(0.540003697429)
        ref[5,0,0]=(1.399657114)/(o+1.)+(0.307020564742)
        ref[5,0,1]=(-0.173920064028)/(o+1.)+(-0.511778709296)
        ref[5,1,0]=(-0.455475273184)/(o+1.)+(-0.22358310787)
        ref[5,1,1]=(-1.02283448161)/(o+1.)+(0.622146965196)
      else:
        arg[0,0,0]=(-0.564478749422)*x[0]**o+(-0.0357223504179)*x[0]+(0.396603298865)*x[1]**o+(-0.461941327703)*x[1]+(-0.667580657928)*x[2]**o+(-0.589957067263)*x[2]
        arg[0,0,1]=(0.930616008951)*x[0]**o+(-0.764067565635)*x[0]+(-0.71350617205)*x[1]**o+(-0.530942701898)*x[1]+(-0.103797356434)*x[2]**o+(0.410704036673)*x[2]
        arg[0,1,0]=(-0.59127392299)*x[0]**o+(0.452490122131)*x[0]+(0.714793625631)*x[1]**o+(-0.933114051133)*x[1]+(-0.693437840325)*x[2]**o+(0.979267231764)*x[2]
        arg[0,1,1]=(0.179682332267)*x[0]**o+(-0.419960476425)*x[0]+(-0.10986598341)*x[1]**o+(0.912986428497)*x[1]+(0.0886946782573)*x[2]**o+(0.516065958873)*x[2]
        arg[1,0,0]=(0.524446863481)*x[0]**o+(-0.439668193423)*x[0]+(0.681259811488)*x[1]**o+(0.439275262258)*x[1]+(0.84573881706)*x[2]**o+(0.377640553175)*x[2]
        arg[1,0,1]=(-0.007464905203)*x[0]**o+(0.234727832471)*x[0]+(0.0428956356618)*x[1]**o+(0.195772930172)*x[1]+(0.457790934614)*x[2]**o+(0.445309166495)*x[2]
        arg[1,1,0]=(0.821283375827)*x[0]**o+(0.212480005798)*x[0]+(0.148817990961)*x[1]**o+(0.119716150877)*x[1]+(0.658959126865)*x[2]**o+(-0.346216916131)*x[2]
        arg[1,1,1]=(-0.0587222422894)*x[0]**o+(-0.325321661973)*x[0]+(0.149293818689)*x[1]**o+(-0.632073115037)*x[1]+(0.018604271874)*x[2]**o+(-0.412288240166)*x[2]
        arg[2,0,0]=(-0.22588082169)*x[0]**o+(0.197780461217)*x[0]+(0.387336989213)*x[1]**o+(-0.680021661016)*x[1]+(-0.903782254346)*x[2]**o+(-0.273268109853)*x[2]
        arg[2,0,1]=(0.415876394297)*x[0]**o+(0.199596795904)*x[0]+(-0.445817134682)*x[1]**o+(0.672539590684)*x[1]+(0.907825473964)*x[2]**o+(0.459560469691)*x[2]
        arg[2,1,0]=(-0.052146887134)*x[0]**o+(-0.229450574738)*x[0]+(0.0929720151604)*x[1]**o+(-0.893861034491)*x[1]+(-0.294030328982)*x[2]**o+(-0.193256779109)*x[2]
        arg[2,1,1]=(0.361171422066)*x[0]**o+(-0.212022056495)*x[0]+(0.0672221746963)*x[1]**o+(0.57506871515)*x[1]+(0.654894469509)*x[2]**o+(0.0749777260533)*x[2]
        arg[3,0,0]=(0.505355253404)*x[0]**o+(-0.0753563586359)*x[0]+(0.544299279727)*x[1]**o+(-0.781954131328)*x[1]+(0.534390713891)*x[2]**o+(-0.719289344378)*x[2]
        arg[3,0,1]=(-0.589309652827)*x[0]**o+(-0.0557421151178)*x[0]+(-0.275311076837)*x[1]**o+(0.998180518149)*x[1]+(-0.844823856433)*x[2]**o+(-0.988843805162)*x[2]
        arg[3,1,0]=(-0.411075104301)*x[0]**o+(-0.405896828894)*x[0]+(0.242301856034)*x[1]**o+(-0.760648969984)*x[1]+(0.0940367373673)*x[2]**o+(-0.121420230853)*x[2]
        arg[3,1,1]=(-0.201426146967)*x[0]**o+(0.722847134095)*x[0]+(-0.00743580369433)*x[1]**o+(-0.593158423904)*x[1]+(0.630575332064)*x[2]**o+(-0.309839791343)*x[2]
        arg[4,0,0]=(0.370057366738)*x[0]**o+(0.35683022394)*x[0]+(0.886304520049)*x[1]**o+(0.834488899465)*x[1]+(0.27408661832)*x[2]**o+(0.325007409933)*x[2]
        arg[4,0,1]=(0.386153377977)*x[0]**o+(-0.889700309398)*x[0]+(-0.60695970137)*x[1]**o+(0.701208035732)*x[1]+(0.435164510485)*x[2]**o+(0.40223148148)*x[2]
        arg[4,1,0]=(-0.227256041906)*x[0]**o+(-0.0406115125763)*x[0]+(-0.0298974314183)*x[1]**o+(0.887167464033)*x[1]+(0.1004751563)*x[2]**o+(0.826924545719)*x[2]
        arg[4,1,1]=(0.733575640137)*x[0]**o+(-0.376269474248)*x[0]+(-0.841290711112)*x[1]**o+(-0.0633013917179)*x[1]+(0.335730636406)*x[2]**o+(0.913607658221)*x[2]
        arg[5,0,0]=(-0.896019254334)*x[0]**o+(0.465470987762)*x[0]+(0.338543664838)*x[1]**o+(0.182102591196)*x[1]+(0.476317455739)*x[2]**o+(0.191367623163)*x[2]
        arg[5,0,1]=(0.828458939676)*x[0]**o+(-0.37853033744)*x[0]+(-0.540809170906)*x[1]**o+(0.367869598038)*x[1]+(0.321196361443)*x[2]**o+(0.548981819769)*x[2]
        arg[5,1,0]=(-0.0805871335579)*x[0]**o+(-0.554679375424)*x[0]+(-0.259438414137)*x[1]**o+(0.72362534417)*x[1]+(-0.838855028889)*x[2]**o+(0.929802758377)*x[2]
        arg[5,1,1]=(-0.947910259125)*x[0]**o+(0.654012824998)*x[0]+(0.425232602078)*x[1]**o+(-0.649345176159)*x[1]+(-0.946280438348)*x[2]**o+(-0.429547284912)*x[2]
        ref[0,0,0]=(-0.835456108484)/(o+1.)+(-0.543810372692)
        ref[0,0,1]=(0.113312480467)/(o+1.)+(-0.44215311543)
        ref[0,1,0]=(-0.569918137684)/(o+1.)+(0.249321651381)
        ref[0,1,1]=(0.158511027114)/(o+1.)+(0.504545955472)
        ref[1,0,0]=(2.05144549203)/(o+1.)+(0.188623811006)
        ref[1,0,1]=(0.493221665073)/(o+1.)+(0.437904964569)
        ref[1,1,0]=(1.62906049365)/(o+1.)+(-0.00701037972825)
        ref[1,1,1]=(0.109175848273)/(o+1.)+(-0.684841508588)
        ref[2,0,0]=(-0.742326086822)/(o+1.)+(-0.377754654826)
        ref[2,0,1]=(0.877884733578)/(o+1.)+(0.665848428139)
        ref[2,1,0]=(-0.253205200956)/(o+1.)+(-0.658284194169)
        ref[2,1,1]=(1.08328806627)/(o+1.)+(0.219012192354)
        ref[3,0,0]=(1.58404524702)/(o+1.)+(-0.788299917171)
        ref[3,0,1]=(-1.7094445861)/(o+1.)+(-0.0232027010654)
        ref[3,1,0]=(-0.0747365109002)/(o+1.)+(-0.643983014866)
        ref[3,1,1]=(0.421713381403)/(o+1.)+(-0.0900755405762)
        ref[4,0,0]=(1.53044850511)/(o+1.)+(0.758163266669)
        ref[4,0,1]=(0.214358187092)/(o+1.)+(0.106869603907)
        ref[4,1,0]=(-0.156678317024)/(o+1.)+(0.836740248588)
        ref[4,1,1]=(0.228015565432)/(o+1.)+(0.237018396127)
        ref[5,0,0]=(-0.0811581337568)/(o+1.)+(0.419470601061)
        ref[5,0,1]=(0.608846130213)/(o+1.)+(0.269160540183)
        ref[5,1,0]=(-1.17888057658)/(o+1.)+(0.549374363561)
        ref[5,1,1]=(-1.4689580954)/(o+1.)+(-0.212439818036)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.216014106349)*x[0]**o+(0.98826230471)*x[0]+(0.522566177319)*x[1]**o+(-0.672520997452)*x[1]
        arg[0,0,0,1]=(0.368024408084)*x[0]**o+(0.047972971332)*x[0]+(-0.307374327321)*x[1]**o+(0.0745169735457)*x[1]
        arg[0,0,1,0]=(0.438735045452)*x[0]**o+(0.76355958752)*x[0]+(0.200829025437)*x[1]**o+(0.56912407759)*x[1]
        arg[0,0,1,1]=(-0.0328769393712)*x[0]**o+(-0.39701363611)*x[0]+(0.704125532519)*x[1]**o+(-0.6602435084)*x[1]
        arg[0,0,2,0]=(-0.547717992717)*x[0]**o+(-0.844256763776)*x[0]+(0.621300470147)*x[1]**o+(-0.105216181226)*x[1]
        arg[0,0,2,1]=(-0.723474063234)*x[0]**o+(-0.980738972901)*x[0]+(-0.183578659566)*x[1]**o+(0.586038450047)*x[1]
        arg[0,1,0,0]=(0.505379907001)*x[0]**o+(-0.470514711277)*x[0]+(0.799116811747)*x[1]**o+(0.682760846386)*x[1]
        arg[0,1,0,1]=(0.353992609198)*x[0]**o+(-0.848107655585)*x[0]+(-0.774065459855)*x[1]**o+(0.0323137462292)*x[1]
        arg[0,1,1,0]=(0.799068805694)*x[0]**o+(-0.979761821209)*x[0]+(0.642674987867)*x[1]**o+(-0.639029952044)*x[1]
        arg[0,1,1,1]=(0.542556683327)*x[0]**o+(0.876753086356)*x[0]+(-0.387121297643)*x[1]**o+(0.377543179628)*x[1]
        arg[0,1,2,0]=(0.424672899596)*x[0]**o+(-0.134184751569)*x[0]+(-0.561877304114)*x[1]**o+(0.549933699347)*x[1]
        arg[0,1,2,1]=(-0.126173294501)*x[0]**o+(-0.338964866844)*x[0]+(-0.163536115759)*x[1]**o+(0.747551558186)*x[1]
        arg[0,2,0,0]=(0.304334183452)*x[0]**o+(0.927809971187)*x[0]+(0.961504942988)*x[1]**o+(0.0698014252396)*x[1]
        arg[0,2,0,1]=(0.488820088455)*x[0]**o+(0.890735247208)*x[0]+(0.686782901398)*x[1]**o+(0.972954994564)*x[1]
        arg[0,2,1,0]=(-0.737577792766)*x[0]**o+(-0.351698913646)*x[0]+(0.392591489865)*x[1]**o+(0.79238240143)*x[1]
        arg[0,2,1,1]=(-0.997254027544)*x[0]**o+(0.579800166465)*x[0]+(-0.95000690643)*x[1]**o+(-0.166463083476)*x[1]
        arg[0,2,2,0]=(-0.999703749644)*x[0]**o+(0.0415211466468)*x[0]+(0.0816160245139)*x[1]**o+(-0.0921114587195)*x[1]
        arg[0,2,2,1]=(-0.172818721386)*x[0]**o+(0.10475692446)*x[0]+(-0.855815339933)*x[1]**o+(-0.50842580034)*x[1]
        arg[0,3,0,0]=(-0.154563121513)*x[0]**o+(-0.0956305513369)*x[0]+(-0.893783521038)*x[1]**o+(0.10772474638)*x[1]
        arg[0,3,0,1]=(0.365331678349)*x[0]**o+(0.403124820832)*x[0]+(-0.0750500521493)*x[1]**o+(0.62559762808)*x[1]
        arg[0,3,1,0]=(-0.42235875521)*x[0]**o+(0.831133238533)*x[0]+(-0.893322944982)*x[1]**o+(0.936074096319)*x[1]
        arg[0,3,1,1]=(0.587382334689)*x[0]**o+(0.996190710631)*x[0]+(-0.796032791654)*x[1]**o+(0.783491453801)*x[1]
        arg[0,3,2,0]=(-0.369527027508)*x[0]**o+(-0.501417552089)*x[0]+(-0.372105439045)*x[1]**o+(0.653283000152)*x[1]
        arg[0,3,2,1]=(0.914893750314)*x[0]**o+(-0.769183323793)*x[0]+(0.571638486717)*x[1]**o+(-0.939784652259)*x[1]
        arg[0,4,0,0]=(0.134859060251)*x[0]**o+(-0.297363165275)*x[0]+(0.198957028974)*x[1]**o+(-0.828531725127)*x[1]
        arg[0,4,0,1]=(-0.371375370217)*x[0]**o+(-0.0422058718402)*x[0]+(0.502010302278)*x[1]**o+(0.442051802597)*x[1]
        arg[0,4,1,0]=(0.513346571173)*x[0]**o+(0.182640867356)*x[0]+(0.76200283664)*x[1]**o+(0.134640432361)*x[1]
        arg[0,4,1,1]=(-0.815543712756)*x[0]**o+(0.0306226823724)*x[0]+(-0.0211231926242)*x[1]**o+(-0.718800228294)*x[1]
        arg[0,4,2,0]=(-0.980596107555)*x[0]**o+(-0.714237107759)*x[0]+(0.547986215851)*x[1]**o+(0.180131474534)*x[1]
        arg[0,4,2,1]=(-0.460780402239)*x[0]**o+(-0.483771113997)*x[0]+(-0.017140263255)*x[1]**o+(-0.127947184627)*x[1]
        arg[1,0,0,0]=(-0.0187443155707)*x[0]**o+(-0.490914415873)*x[0]+(-0.720022496093)*x[1]**o+(0.0265633770733)*x[1]
        arg[1,0,0,1]=(-0.634855765904)*x[0]**o+(0.345031954366)*x[0]+(-0.0532392363316)*x[1]**o+(0.983295524564)*x[1]
        arg[1,0,1,0]=(0.262182040476)*x[0]**o+(-0.883857758733)*x[0]+(0.880670543535)*x[1]**o+(0.193477695378)*x[1]
        arg[1,0,1,1]=(-0.0825976168371)*x[0]**o+(0.161838720337)*x[0]+(-0.0110587499089)*x[1]**o+(-0.634455402489)*x[1]
        arg[1,0,2,0]=(0.284221958342)*x[0]**o+(-0.948548729029)*x[0]+(0.430914344458)*x[1]**o+(-0.18563638241)*x[1]
        arg[1,0,2,1]=(-0.459558022898)*x[0]**o+(0.431907572866)*x[0]+(-0.0597669070218)*x[1]**o+(0.0694271372225)*x[1]
        arg[1,1,0,0]=(0.54071078447)*x[0]**o+(-0.244570994356)*x[0]+(0.00506219762373)*x[1]**o+(-0.0816378019186)*x[1]
        arg[1,1,0,1]=(-0.100752221379)*x[0]**o+(0.535738345566)*x[0]+(0.730325284553)*x[1]**o+(-0.885044748717)*x[1]
        arg[1,1,1,0]=(-0.0422460353138)*x[0]**o+(-0.986416505995)*x[0]+(-0.111766776997)*x[1]**o+(-0.842179226197)*x[1]
        arg[1,1,1,1]=(0.0690281593169)*x[0]**o+(-0.885054188663)*x[0]+(-0.680349314907)*x[1]**o+(0.00277190934345)*x[1]
        arg[1,1,2,0]=(0.502632146841)*x[0]**o+(-0.197332151163)*x[0]+(-0.0942257361149)*x[1]**o+(0.045171692749)*x[1]
        arg[1,1,2,1]=(0.427910466818)*x[0]**o+(-0.61355947216)*x[0]+(0.465128040517)*x[1]**o+(0.0874536555354)*x[1]
        arg[1,2,0,0]=(0.943819415417)*x[0]**o+(0.884020001869)*x[0]+(-0.769166027477)*x[1]**o+(-0.10032307068)*x[1]
        arg[1,2,0,1]=(0.0670224828564)*x[0]**o+(-0.645424481071)*x[0]+(-0.756361652678)*x[1]**o+(-0.274687916196)*x[1]
        arg[1,2,1,0]=(0.397836742229)*x[0]**o+(-0.708791058745)*x[0]+(0.920869587663)*x[1]**o+(-0.250037069918)*x[1]
        arg[1,2,1,1]=(0.722570486719)*x[0]**o+(-0.358082774936)*x[0]+(0.786344673969)*x[1]**o+(0.310591893637)*x[1]
        arg[1,2,2,0]=(-0.379404573553)*x[0]**o+(-0.969753501645)*x[0]+(-0.0745714915337)*x[1]**o+(-0.262943796911)*x[1]
        arg[1,2,2,1]=(0.817999459632)*x[0]**o+(0.731516351792)*x[0]+(-0.531411613325)*x[1]**o+(-0.824858061154)*x[1]
        arg[1,3,0,0]=(-0.206295444021)*x[0]**o+(-0.781973944918)*x[0]+(-0.173845545382)*x[1]**o+(-0.360122364573)*x[1]
        arg[1,3,0,1]=(-0.0604068005719)*x[0]**o+(0.238691416507)*x[0]+(-0.82262982031)*x[1]**o+(-0.893751755968)*x[1]
        arg[1,3,1,0]=(-0.686409669386)*x[0]**o+(0.711830599399)*x[0]+(0.0839688142166)*x[1]**o+(-0.99463449684)*x[1]
        arg[1,3,1,1]=(-0.0864745975077)*x[0]**o+(0.00068907895313)*x[0]+(-0.766810513534)*x[1]**o+(0.685236202954)*x[1]
        arg[1,3,2,0]=(0.0909338812695)*x[0]**o+(-0.709595654169)*x[0]+(-0.483428504499)*x[1]**o+(-0.0707341191006)*x[1]
        arg[1,3,2,1]=(-0.38342983855)*x[0]**o+(-0.0407824575865)*x[0]+(-0.201329040891)*x[1]**o+(0.0240007937141)*x[1]
        arg[1,4,0,0]=(0.512465803173)*x[0]**o+(-0.371898603608)*x[0]+(-0.79453595861)*x[1]**o+(0.340546986375)*x[1]
        arg[1,4,0,1]=(0.742383198442)*x[0]**o+(-0.603548189912)*x[0]+(-0.270277825546)*x[1]**o+(-0.525100913894)*x[1]
        arg[1,4,1,0]=(-0.693074384029)*x[0]**o+(-0.825037404844)*x[0]+(-0.626116177978)*x[1]**o+(0.628159342479)*x[1]
        arg[1,4,1,1]=(-0.595844943255)*x[0]**o+(0.250758194807)*x[0]+(0.504390052142)*x[1]**o+(-0.124588859702)*x[1]
        arg[1,4,2,0]=(-0.651727490997)*x[0]**o+(0.411138504151)*x[0]+(0.323769160331)*x[1]**o+(0.904379126671)*x[1]
        arg[1,4,2,1]=(-0.71307850178)*x[0]**o+(0.734964608119)*x[0]+(0.348693698506)*x[1]**o+(0.272270639772)*x[1]
        arg[2,0,0,0]=(0.558236567771)*x[0]**o+(0.516985445776)*x[0]+(0.0278915137861)*x[1]**o+(-0.268619977833)*x[1]
        arg[2,0,0,1]=(-0.871181709658)*x[0]**o+(-0.132251987744)*x[0]+(-0.215497073461)*x[1]**o+(0.508008002082)*x[1]
        arg[2,0,1,0]=(0.880795999572)*x[0]**o+(-0.231151166496)*x[0]+(-0.698846681617)*x[1]**o+(0.914483413825)*x[1]
        arg[2,0,1,1]=(0.684792911211)*x[0]**o+(-0.801408782989)*x[0]+(0.689713232168)*x[1]**o+(0.142847173797)*x[1]
        arg[2,0,2,0]=(-0.912566319945)*x[0]**o+(0.811138574657)*x[0]+(-0.884250879071)*x[1]**o+(-0.177206225477)*x[1]
        arg[2,0,2,1]=(-0.898648046678)*x[0]**o+(0.581554373688)*x[0]+(0.567642878443)*x[1]**o+(0.636312832751)*x[1]
        arg[2,1,0,0]=(-0.788853365648)*x[0]**o+(-0.135329327941)*x[0]+(0.180850147594)*x[1]**o+(0.742619029335)*x[1]
        arg[2,1,0,1]=(-0.824539980645)*x[0]**o+(0.857173393615)*x[0]+(0.850455264916)*x[1]**o+(0.975422322464)*x[1]
        arg[2,1,1,0]=(0.462293517317)*x[0]**o+(-0.173835273794)*x[0]+(-0.800072514276)*x[1]**o+(0.140214114627)*x[1]
        arg[2,1,1,1]=(-0.187890814293)*x[0]**o+(0.534472364495)*x[0]+(-0.799640665571)*x[1]**o+(0.301287955302)*x[1]
        arg[2,1,2,0]=(0.278901492946)*x[0]**o+(0.620544326886)*x[0]+(-0.705702443752)*x[1]**o+(-0.489690636328)*x[1]
        arg[2,1,2,1]=(-0.926772900235)*x[0]**o+(0.134654742845)*x[0]+(0.356561975337)*x[1]**o+(-0.919999359857)*x[1]
        arg[2,2,0,0]=(-0.675956301416)*x[0]**o+(0.149582589205)*x[0]+(0.186440827385)*x[1]**o+(0.88381610446)*x[1]
        arg[2,2,0,1]=(-0.697267962915)*x[0]**o+(0.991366596424)*x[0]+(0.310290094618)*x[1]**o+(-0.881771505066)*x[1]
        arg[2,2,1,0]=(0.506027358474)*x[0]**o+(0.294136137354)*x[0]+(0.732003937836)*x[1]**o+(0.633469131504)*x[1]
        arg[2,2,1,1]=(-0.427146039728)*x[0]**o+(-0.975737063903)*x[0]+(-0.934913653772)*x[1]**o+(-0.662568325005)*x[1]
        arg[2,2,2,0]=(0.477849916621)*x[0]**o+(-0.798883883367)*x[0]+(-0.154217684441)*x[1]**o+(-0.175817547912)*x[1]
        arg[2,2,2,1]=(0.935192620387)*x[0]**o+(-0.193802588249)*x[0]+(0.406181949213)*x[1]**o+(0.20422475396)*x[1]
        arg[2,3,0,0]=(0.965944278856)*x[0]**o+(-0.736175411725)*x[0]+(0.92404347386)*x[1]**o+(0.273152788392)*x[1]
        arg[2,3,0,1]=(-0.606944412413)*x[0]**o+(0.692555413476)*x[0]+(0.812085676275)*x[1]**o+(-0.0809920218886)*x[1]
        arg[2,3,1,0]=(-0.314634777406)*x[0]**o+(0.750181618592)*x[0]+(0.0554739779642)*x[1]**o+(-0.333788326726)*x[1]
        arg[2,3,1,1]=(0.85610815703)*x[0]**o+(0.910096005524)*x[0]+(-0.227092814843)*x[1]**o+(-0.0849532598914)*x[1]
        arg[2,3,2,0]=(0.735064128504)*x[0]**o+(-0.253919528943)*x[0]+(0.8457463307)*x[1]**o+(0.865727345925)*x[1]
        arg[2,3,2,1]=(0.520607989389)*x[0]**o+(0.569210317105)*x[0]+(0.500780081677)*x[1]**o+(-0.787630914351)*x[1]
        arg[2,4,0,0]=(0.179774694296)*x[0]**o+(-0.890872128607)*x[0]+(-0.185758177247)*x[1]**o+(0.888231734606)*x[1]
        arg[2,4,0,1]=(0.059863462227)*x[0]**o+(-0.0796463123948)*x[0]+(-0.104564520504)*x[1]**o+(-0.336182528617)*x[1]
        arg[2,4,1,0]=(-0.636828275967)*x[0]**o+(0.156894572218)*x[0]+(-0.884260096115)*x[1]**o+(-0.967431438795)*x[1]
        arg[2,4,1,1]=(-0.958168110389)*x[0]**o+(-0.0424772631477)*x[0]+(0.512363473316)*x[1]**o+(-0.716729419668)*x[1]
        arg[2,4,2,0]=(0.19736654467)*x[0]**o+(-0.288723518002)*x[0]+(0.243674007033)*x[1]**o+(0.911210601142)*x[1]
        arg[2,4,2,1]=(-0.847493343055)*x[0]**o+(0.115003265593)*x[0]+(0.360071936714)*x[1]**o+(-0.0265385061141)*x[1]
        arg[3,0,0,0]=(-0.510550590836)*x[0]**o+(-0.266174606631)*x[0]+(0.771580708159)*x[1]**o+(-0.770554374259)*x[1]
        arg[3,0,0,1]=(-0.65795024738)*x[0]**o+(-0.83394628525)*x[0]+(0.812461325452)*x[1]**o+(0.191593733482)*x[1]
        arg[3,0,1,0]=(0.136878718548)*x[0]**o+(-0.000345061692795)*x[0]+(-0.0780280205008)*x[1]**o+(0.173925537373)*x[1]
        arg[3,0,1,1]=(0.584079470149)*x[0]**o+(0.138674493948)*x[0]+(-0.563553909861)*x[1]**o+(-0.254304838008)*x[1]
        arg[3,0,2,0]=(0.326192887554)*x[0]**o+(0.993615980303)*x[0]+(-0.26995279693)*x[1]**o+(0.216112615287)*x[1]
        arg[3,0,2,1]=(0.684786021819)*x[0]**o+(0.158934017292)*x[0]+(-0.457448199191)*x[1]**o+(0.644722228163)*x[1]
        arg[3,1,0,0]=(-0.198252162262)*x[0]**o+(-0.821832378445)*x[0]+(0.831807441916)*x[1]**o+(-0.70643670471)*x[1]
        arg[3,1,0,1]=(0.361698759016)*x[0]**o+(0.556511107101)*x[0]+(0.892219700601)*x[1]**o+(-0.861598520361)*x[1]
        arg[3,1,1,0]=(-0.107946706634)*x[0]**o+(-0.910191934756)*x[0]+(0.502507607871)*x[1]**o+(0.649089810516)*x[1]
        arg[3,1,1,1]=(-0.92959760672)*x[0]**o+(0.950682520961)*x[0]+(0.103321004187)*x[1]**o+(-0.631562667815)*x[1]
        arg[3,1,2,0]=(0.858565242025)*x[0]**o+(-0.106810606075)*x[0]+(0.236377425048)*x[1]**o+(-0.296320858145)*x[1]
        arg[3,1,2,1]=(0.148558873625)*x[0]**o+(0.458488334082)*x[0]+(-0.0172824905523)*x[1]**o+(-0.321212766581)*x[1]
        arg[3,2,0,0]=(-0.239733890336)*x[0]**o+(0.558819103697)*x[0]+(-0.151683129885)*x[1]**o+(0.153995780645)*x[1]
        arg[3,2,0,1]=(-0.949100270937)*x[0]**o+(0.0687171499139)*x[0]+(0.968028021831)*x[1]**o+(-0.935600499345)*x[1]
        arg[3,2,1,0]=(-0.739823767237)*x[0]**o+(0.417825150389)*x[0]+(0.318612170707)*x[1]**o+(0.627557337421)*x[1]
        arg[3,2,1,1]=(-0.965142167324)*x[0]**o+(0.981446854383)*x[0]+(0.266191185951)*x[1]**o+(-0.581303142841)*x[1]
        arg[3,2,2,0]=(-0.00312606993817)*x[0]**o+(-0.191885030828)*x[0]+(-0.342525068561)*x[1]**o+(-0.953710538626)*x[1]
        arg[3,2,2,1]=(-0.689127187476)*x[0]**o+(-0.107190366063)*x[0]+(-0.0556716250594)*x[1]**o+(-0.609462166754)*x[1]
        arg[3,3,0,0]=(-0.765237197853)*x[0]**o+(-0.0707283944117)*x[0]+(0.891789872287)*x[1]**o+(-0.255939791858)*x[1]
        arg[3,3,0,1]=(-0.367658362103)*x[0]**o+(0.27406309422)*x[0]+(0.866288683947)*x[1]**o+(0.403194030287)*x[1]
        arg[3,3,1,0]=(-0.362795000993)*x[0]**o+(-0.498774415158)*x[0]+(0.0726808331583)*x[1]**o+(-0.871066791136)*x[1]
        arg[3,3,1,1]=(-0.177149059814)*x[0]**o+(0.287463753612)*x[0]+(0.101726559941)*x[1]**o+(0.561797204517)*x[1]
        arg[3,3,2,0]=(0.770615121039)*x[0]**o+(-0.503402148251)*x[0]+(0.319753007634)*x[1]**o+(0.586868894261)*x[1]
        arg[3,3,2,1]=(0.924125904724)*x[0]**o+(-0.508988920423)*x[0]+(-0.866752504824)*x[1]**o+(0.518544700652)*x[1]
        arg[3,4,0,0]=(-0.973660291124)*x[0]**o+(0.746869369283)*x[0]+(-0.634097171844)*x[1]**o+(0.723638944736)*x[1]
        arg[3,4,0,1]=(-0.599366869066)*x[0]**o+(-0.59893829688)*x[0]+(0.124483401866)*x[1]**o+(-0.924820315792)*x[1]
        arg[3,4,1,0]=(-0.707374787729)*x[0]**o+(-0.494748755022)*x[0]+(0.391752209698)*x[1]**o+(0.149710640116)*x[1]
        arg[3,4,1,1]=(0.329587748119)*x[0]**o+(-0.788289803764)*x[0]+(-0.145100163626)*x[1]**o+(0.141493810237)*x[1]
        arg[3,4,2,0]=(0.17496644337)*x[0]**o+(-0.270242104589)*x[0]+(0.996684379179)*x[1]**o+(-0.751697440046)*x[1]
        arg[3,4,2,1]=(0.79046837395)*x[0]**o+(0.568394473062)*x[0]+(0.359419854016)*x[1]**o+(0.630686948244)*x[1]
        ref[0,0,0,0]=(0.306552070969)/(o+1.)+(0.157870653629)
        ref[0,0,0,1]=(0.0606500807631)/(o+1.)+(0.0612449724389)
        ref[0,0,1,0]=(0.639564070889)/(o+1.)+(0.666341832555)
        ref[0,0,1,1]=(0.671248593147)/(o+1.)+(-0.528628572255)
        ref[0,0,2,0]=(0.0735824774293)/(o+1.)+(-0.474736472501)
        ref[0,0,2,1]=(-0.9070527228)/(o+1.)+(-0.197350261427)
        ref[0,1,0,0]=(1.30449671875)/(o+1.)+(0.106123067554)
        ref[0,1,0,1]=(-0.420072850657)/(o+1.)+(-0.407896954678)
        ref[0,1,1,0]=(1.44174379356)/(o+1.)+(-0.809395886627)
        ref[0,1,1,1]=(0.155435385684)/(o+1.)+(0.627148132992)
        ref[0,1,2,0]=(-0.137204404518)/(o+1.)+(0.207874473889)
        ref[0,1,2,1]=(-0.289709410261)/(o+1.)+(0.204293345671)
        ref[0,2,0,0]=(1.26583912644)/(o+1.)+(0.498805698213)
        ref[0,2,0,1]=(1.17560298985)/(o+1.)+(0.931845120886)
        ref[0,2,1,0]=(-0.344986302901)/(o+1.)+(0.220341743892)
        ref[0,2,1,1]=(-1.94726093397)/(o+1.)+(0.206668541494)
        ref[0,2,2,0]=(-0.91808772513)/(o+1.)+(-0.0252951560363)
        ref[0,2,2,1]=(-1.02863406132)/(o+1.)+(-0.20183443794)
        ref[0,3,0,0]=(-1.04834664255)/(o+1.)+(0.00604709752155)
        ref[0,3,0,1]=(0.2902816262)/(o+1.)+(0.514361224456)
        ref[0,3,1,0]=(-1.31568170019)/(o+1.)+(0.883603667426)
        ref[0,3,1,1]=(-0.208650456965)/(o+1.)+(0.889841082216)
        ref[0,3,2,0]=(-0.741632466554)/(o+1.)+(0.0759327240318)
        ref[0,3,2,1]=(1.48653223703)/(o+1.)+(-0.854483988026)
        ref[0,4,0,0]=(0.333816089225)/(o+1.)+(-0.562947445201)
        ref[0,4,0,1]=(0.130634932061)/(o+1.)+(0.199922965378)
        ref[0,4,1,0]=(1.27534940781)/(o+1.)+(0.158640649858)
        ref[0,4,1,1]=(-0.83666690538)/(o+1.)+(-0.344088772961)
        ref[0,4,2,0]=(-0.432609891704)/(o+1.)+(-0.267052816612)
        ref[0,4,2,1]=(-0.477920665494)/(o+1.)+(-0.305859149312)
        ref[1,0,0,0]=(-0.738766811663)/(o+1.)+(-0.2321755194)
        ref[1,0,0,1]=(-0.688095002235)/(o+1.)+(0.664163739465)
        ref[1,0,1,0]=(1.14285258401)/(o+1.)+(-0.345190031677)
        ref[1,0,1,1]=(-0.093656366746)/(o+1.)+(-0.236308341076)
        ref[1,0,2,0]=(0.7151363028)/(o+1.)+(-0.56709255572)
        ref[1,0,2,1]=(-0.51932492992)/(o+1.)+(0.250667355044)
        ref[1,1,0,0]=(0.545772982093)/(o+1.)+(-0.163104398138)
        ref[1,1,0,1]=(0.629573063174)/(o+1.)+(-0.174653201576)
        ref[1,1,1,0]=(-0.154012812311)/(o+1.)+(-0.914297866096)
        ref[1,1,1,1]=(-0.61132115559)/(o+1.)+(-0.44114113966)
        ref[1,1,2,0]=(0.408406410726)/(o+1.)+(-0.076080229207)
        ref[1,1,2,1]=(0.893038507335)/(o+1.)+(-0.263052908312)
        ref[1,2,0,0]=(0.17465338794)/(o+1.)+(0.391848465594)
        ref[1,2,0,1]=(-0.689339169821)/(o+1.)+(-0.460056198634)
        ref[1,2,1,0]=(1.31870632989)/(o+1.)+(-0.479414064332)
        ref[1,2,1,1]=(1.50891516069)/(o+1.)+(-0.0237454406495)
        ref[1,2,2,0]=(-0.453976065087)/(o+1.)+(-0.616348649278)
        ref[1,2,2,1]=(0.286587846307)/(o+1.)+(-0.0466708546809)
        ref[1,3,0,0]=(-0.380140989403)/(o+1.)+(-0.571048154745)
        ref[1,3,0,1]=(-0.883036620882)/(o+1.)+(-0.32753016973)
        ref[1,3,1,0]=(-0.60244085517)/(o+1.)+(-0.141401948721)
        ref[1,3,1,1]=(-0.853285111042)/(o+1.)+(0.342962640954)
        ref[1,3,2,0]=(-0.39249462323)/(o+1.)+(-0.390164886635)
        ref[1,3,2,1]=(-0.584758879441)/(o+1.)+(-0.00839083193621)
        ref[1,4,0,0]=(-0.282070155438)/(o+1.)+(-0.0156758086169)
        ref[1,4,0,1]=(0.472105372896)/(o+1.)+(-0.564324551903)
        ref[1,4,1,0]=(-1.31919056201)/(o+1.)+(-0.0984390311826)
        ref[1,4,1,1]=(-0.0914548911124)/(o+1.)+(0.0630846675525)
        ref[1,4,2,0]=(-0.327958330666)/(o+1.)+(0.657758815411)
        ref[1,4,2,1]=(-0.364384803274)/(o+1.)+(0.503617623946)
        ref[2,0,0,0]=(0.586128081557)/(o+1.)+(0.124182733972)
        ref[2,0,0,1]=(-1.08667878312)/(o+1.)+(0.187878007169)
        ref[2,0,1,0]=(0.181949317955)/(o+1.)+(0.341666123665)
        ref[2,0,1,1]=(1.37450614338)/(o+1.)+(-0.329280804596)
        ref[2,0,2,0]=(-1.79681719902)/(o+1.)+(0.31696617459)
        ref[2,0,2,1]=(-0.331005168235)/(o+1.)+(0.608933603219)
        ref[2,1,0,0]=(-0.608003218053)/(o+1.)+(0.303644850697)
        ref[2,1,0,1]=(0.0259152842714)/(o+1.)+(0.916297858039)
        ref[2,1,1,0]=(-0.33777899696)/(o+1.)+(-0.0168105795836)
        ref[2,1,1,1]=(-0.987531479864)/(o+1.)+(0.417880159899)
        ref[2,1,2,0]=(-0.426800950806)/(o+1.)+(0.0654268452787)
        ref[2,1,2,1]=(-0.570210924898)/(o+1.)+(-0.392672308506)
        ref[2,2,0,0]=(-0.489515474032)/(o+1.)+(0.516699346832)
        ref[2,2,0,1]=(-0.386977868298)/(o+1.)+(0.054797545679)
        ref[2,2,1,0]=(1.23803129631)/(o+1.)+(0.463802634429)
        ref[2,2,1,1]=(-1.3620596935)/(o+1.)+(-0.819152694454)
        ref[2,2,2,0]=(0.32363223218)/(o+1.)+(-0.48735071564)
        ref[2,2,2,1]=(1.3413745696)/(o+1.)+(0.00521108285542)
        ref[2,3,0,0]=(1.88998775272)/(o+1.)+(-0.231511311666)
        ref[2,3,0,1]=(0.205141263861)/(o+1.)+(0.305781695794)
        ref[2,3,1,0]=(-0.259160799442)/(o+1.)+(0.208196645933)
        ref[2,3,1,1]=(0.629015342187)/(o+1.)+(0.412571372816)
        ref[2,3,2,0]=(1.5808104592)/(o+1.)+(0.305903908491)
        ref[2,3,2,1]=(1.02138807107)/(o+1.)+(-0.109210298623)
        ref[2,4,0,0]=(-0.00598348295078)/(o+1.)+(-0.00132019700044)
        ref[2,4,0,1]=(-0.0447010582767)/(o+1.)+(-0.207914420506)
        ref[2,4,1,0]=(-1.52108837208)/(o+1.)+(-0.405268433289)
        ref[2,4,1,1]=(-0.445804637073)/(o+1.)+(-0.379603341408)
        ref[2,4,2,0]=(0.441040551703)/(o+1.)+(0.31124354157)
        ref[2,4,2,1]=(-0.487421406342)/(o+1.)+(0.0442323797396)
        ref[3,0,0,0]=(0.261030117323)/(o+1.)+(-0.518364490445)
        ref[3,0,0,1]=(0.154511078072)/(o+1.)+(-0.321176275884)
        ref[3,0,1,0]=(0.0588506980476)/(o+1.)+(0.0867902378403)
        ref[3,0,1,1]=(0.0205255602871)/(o+1.)+(-0.0578151720304)
        ref[3,0,2,0]=(0.0562400906237)/(o+1.)+(0.604864297795)
        ref[3,0,2,1]=(0.227337822629)/(o+1.)+(0.401828122727)
        ref[3,1,0,0]=(0.633555279654)/(o+1.)+(-0.764134541578)
        ref[3,1,0,1]=(1.25391845962)/(o+1.)+(-0.15254370663)
        ref[3,1,1,0]=(0.394560901238)/(o+1.)+(-0.13055106212)
        ref[3,1,1,1]=(-0.826276602534)/(o+1.)+(0.159559926573)
        ref[3,1,2,0]=(1.09494266707)/(o+1.)+(-0.20156573211)
        ref[3,1,2,1]=(0.131276383073)/(o+1.)+(0.0686377837506)
        ref[3,2,0,0]=(-0.391417020221)/(o+1.)+(0.356407442171)
        ref[3,2,0,1]=(0.0189277508933)/(o+1.)+(-0.433441674716)
        ref[3,2,1,0]=(-0.42121159653)/(o+1.)+(0.522691243905)
        ref[3,2,1,1]=(-0.698950981373)/(o+1.)+(0.200071855771)
        ref[3,2,2,0]=(-0.345651138499)/(o+1.)+(-0.572797784727)
        ref[3,2,2,1]=(-0.744798812536)/(o+1.)+(-0.358326266408)
        ref[3,3,0,0]=(0.126552674434)/(o+1.)+(-0.163334093135)
        ref[3,3,0,1]=(0.498630321844)/(o+1.)+(0.338628562253)
        ref[3,3,1,0]=(-0.290114167835)/(o+1.)+(-0.684920603147)
        ref[3,3,1,1]=(-0.0754224998727)/(o+1.)+(0.424630479064)
        ref[3,3,2,0]=(1.09036812867)/(o+1.)+(0.0417333730049)
        ref[3,3,2,1]=(0.0573733998999)/(o+1.)+(0.00477789011464)
        ref[3,4,0,0]=(-1.60775746297)/(o+1.)+(0.73525415701)
        ref[3,4,0,1]=(-0.4748834672)/(o+1.)+(-0.761879306336)
        ref[3,4,1,0]=(-0.31562257803)/(o+1.)+(-0.172519057453)
        ref[3,4,1,1]=(0.184487584493)/(o+1.)+(-0.323397996764)
        ref[3,4,2,0]=(1.17165082255)/(o+1.)+(-0.510969772317)
        ref[3,4,2,1]=(1.14988822797)/(o+1.)+(0.599540710653)
      else:
        arg[0,0,0,0]=(-0.517412938774)*x[0]**o+(0.824804313111)*x[0]+(-0.681706563374)*x[1]**o+(-0.259540440272)*x[1]+(0.876735283937)*x[2]**o+(0.307072859807)*x[2]
        arg[0,0,0,1]=(-0.545981202251)*x[0]**o+(-0.530080933212)*x[0]+(-0.339289111599)*x[1]**o+(-0.726311605088)*x[1]+(0.334837705444)*x[2]**o+(-0.422667311064)*x[2]
        arg[0,0,1,0]=(-0.94261387328)*x[0]**o+(-0.315825326734)*x[0]+(0.542502801277)*x[1]**o+(-0.927525943646)*x[1]+(-0.802118238213)*x[2]**o+(-0.106512314706)*x[2]
        arg[0,0,1,1]=(-0.877940099286)*x[0]**o+(-0.801256557869)*x[0]+(-0.155983792242)*x[1]**o+(0.200547191496)*x[1]+(-0.960553133432)*x[2]**o+(0.968862792367)*x[2]
        arg[0,0,2,0]=(-0.870342547946)*x[0]**o+(-0.956447402661)*x[0]+(-0.506797803004)*x[1]**o+(0.342841675107)*x[1]+(0.528354159344)*x[2]**o+(-0.824546729564)*x[2]
        arg[0,0,2,1]=(-0.411899137296)*x[0]**o+(0.133736458559)*x[0]+(0.852726131619)*x[1]**o+(0.692795229813)*x[1]+(0.047372576403)*x[2]**o+(0.810485765945)*x[2]
        arg[0,1,0,0]=(0.961758233254)*x[0]**o+(-0.606307924964)*x[0]+(0.0612696412665)*x[1]**o+(-0.947823541802)*x[1]+(-0.113021104365)*x[2]**o+(-0.00802440322155)*x[2]
        arg[0,1,0,1]=(-0.684097052467)*x[0]**o+(0.467140835537)*x[0]+(-0.0742537091205)*x[1]**o+(0.541641916143)*x[1]+(0.987491548505)*x[2]**o+(-0.558380262255)*x[2]
        arg[0,1,1,0]=(-0.901821988855)*x[0]**o+(0.34443714654)*x[0]+(-0.682552654793)*x[1]**o+(0.398276086461)*x[1]+(0.334644264965)*x[2]**o+(0.21454845251)*x[2]
        arg[0,1,1,1]=(0.929614060295)*x[0]**o+(-0.322609307947)*x[0]+(-0.426962200724)*x[1]**o+(0.464659992564)*x[1]+(-0.13983863674)*x[2]**o+(-0.681648817325)*x[2]
        arg[0,1,2,0]=(0.504166508914)*x[0]**o+(-0.978562859041)*x[0]+(-0.769650404806)*x[1]**o+(0.443500137153)*x[1]+(0.969669965936)*x[2]**o+(0.401742207403)*x[2]
        arg[0,1,2,1]=(-0.478561324822)*x[0]**o+(-0.772019838834)*x[0]+(-0.779194844829)*x[1]**o+(-0.747980400065)*x[1]+(0.827885608192)*x[2]**o+(0.135571151528)*x[2]
        arg[0,2,0,0]=(-0.210426221052)*x[0]**o+(-0.464938141903)*x[0]+(-0.776844066203)*x[1]**o+(0.267244555705)*x[1]+(-0.773135929429)*x[2]**o+(-0.00846060888954)*x[2]
        arg[0,2,0,1]=(0.0945858334531)*x[0]**o+(-0.492400895713)*x[0]+(-0.128564669727)*x[1]**o+(-0.0951458189983)*x[1]+(-0.54426715378)*x[2]**o+(0.106545346072)*x[2]
        arg[0,2,1,0]=(-0.482424571392)*x[0]**o+(0.0431208155652)*x[0]+(0.557138601636)*x[1]**o+(-0.725564282551)*x[1]+(0.9081107419)*x[2]**o+(0.0179201585352)*x[2]
        arg[0,2,1,1]=(-0.555731853052)*x[0]**o+(0.447137254671)*x[0]+(0.820958717703)*x[1]**o+(-0.580251434504)*x[1]+(0.227884622965)*x[2]**o+(0.34460189453)*x[2]
        arg[0,2,2,0]=(-0.25868983898)*x[0]**o+(0.759165881486)*x[0]+(-0.499903058341)*x[1]**o+(0.896471093639)*x[1]+(0.351955520633)*x[2]**o+(0.16566756566)*x[2]
        arg[0,2,2,1]=(-0.37859354794)*x[0]**o+(0.925387188779)*x[0]+(0.0111055293719)*x[1]**o+(-0.311998642154)*x[1]+(0.441009182046)*x[2]**o+(0.0968247918271)*x[2]
        arg[0,3,0,0]=(-0.461665900641)*x[0]**o+(0.812002051396)*x[0]+(-0.767456852027)*x[1]**o+(0.809069781575)*x[1]+(0.690416953885)*x[2]**o+(0.171662369216)*x[2]
        arg[0,3,0,1]=(-0.804811324639)*x[0]**o+(-0.0962711523954)*x[0]+(0.980130059825)*x[1]**o+(0.98279180603)*x[1]+(-0.0614353559054)*x[2]**o+(0.645637153538)*x[2]
        arg[0,3,1,0]=(0.97878415369)*x[0]**o+(-0.676321251698)*x[0]+(0.916075195717)*x[1]**o+(0.244509683371)*x[1]+(-0.939563193145)*x[2]**o+(-0.866506108913)*x[2]
        arg[0,3,1,1]=(-0.676965895162)*x[0]**o+(-0.124355338519)*x[0]+(0.680151101195)*x[1]**o+(-0.839412433156)*x[1]+(0.434088587223)*x[2]**o+(-0.275286048166)*x[2]
        arg[0,3,2,0]=(0.414822536626)*x[0]**o+(0.493836498194)*x[0]+(0.967077097745)*x[1]**o+(-0.0115301157869)*x[1]+(-0.307994657929)*x[2]**o+(-0.967766608312)*x[2]
        arg[0,3,2,1]=(-0.34064670348)*x[0]**o+(-0.975121872905)*x[0]+(0.0839457874766)*x[1]**o+(0.0995106978636)*x[1]+(-0.533837237516)*x[2]**o+(0.771268044302)*x[2]
        arg[0,4,0,0]=(0.223197544637)*x[0]**o+(0.712738871576)*x[0]+(-0.386972075747)*x[1]**o+(0.289829507744)*x[1]+(0.153418793839)*x[2]**o+(0.876087048492)*x[2]
        arg[0,4,0,1]=(0.696750948653)*x[0]**o+(0.466905335953)*x[0]+(-0.611476041907)*x[1]**o+(-0.939902457901)*x[1]+(-0.752675432265)*x[2]**o+(-0.869502717044)*x[2]
        arg[0,4,1,0]=(-0.465562201723)*x[0]**o+(-0.388716352051)*x[0]+(0.260044093503)*x[1]**o+(0.919379163294)*x[1]+(-0.902453394797)*x[2]**o+(0.378318744505)*x[2]
        arg[0,4,1,1]=(0.121990842946)*x[0]**o+(-0.691267894153)*x[0]+(-0.703960568755)*x[1]**o+(-0.582404896428)*x[1]+(0.57560180036)*x[2]**o+(0.537099448589)*x[2]
        arg[0,4,2,0]=(-0.495452196513)*x[0]**o+(-0.939389935669)*x[0]+(-0.237205717598)*x[1]**o+(0.61376516207)*x[1]+(-0.506082738728)*x[2]**o+(0.407704148542)*x[2]
        arg[0,4,2,1]=(0.450976180771)*x[0]**o+(-0.00739258898437)*x[0]+(0.210130848093)*x[1]**o+(0.624639007009)*x[1]+(0.173795671727)*x[2]**o+(0.508269170853)*x[2]
        arg[1,0,0,0]=(0.609685942177)*x[0]**o+(0.609299494609)*x[0]+(0.321046388121)*x[1]**o+(0.598197189498)*x[1]+(0.536092177618)*x[2]**o+(0.576427431157)*x[2]
        arg[1,0,0,1]=(0.412500706234)*x[0]**o+(0.905942239017)*x[0]+(-0.974980343169)*x[1]**o+(-0.131245997607)*x[1]+(0.745809068018)*x[2]**o+(0.543470258153)*x[2]
        arg[1,0,1,0]=(-0.903032545947)*x[0]**o+(-0.778693321051)*x[0]+(0.399342164105)*x[1]**o+(0.478212896767)*x[1]+(0.254596269802)*x[2]**o+(0.70308382001)*x[2]
        arg[1,0,1,1]=(-0.281529506793)*x[0]**o+(0.74166036027)*x[0]+(0.629713747725)*x[1]**o+(-0.275164746597)*x[1]+(0.37523909979)*x[2]**o+(-0.878026427978)*x[2]
        arg[1,0,2,0]=(0.236027764061)*x[0]**o+(-0.816910805083)*x[0]+(0.0782011624185)*x[1]**o+(-0.644004245666)*x[1]+(0.293586462602)*x[2]**o+(0.557045894695)*x[2]
        arg[1,0,2,1]=(0.868061727269)*x[0]**o+(0.53329318943)*x[0]+(0.367603915235)*x[1]**o+(0.89280277925)*x[1]+(-0.316595424948)*x[2]**o+(0.593830565147)*x[2]
        arg[1,1,0,0]=(0.886950595996)*x[0]**o+(-0.290323112286)*x[0]+(-0.863347030089)*x[1]**o+(0.545096994541)*x[1]+(0.338200944281)*x[2]**o+(-0.79898415048)*x[2]
        arg[1,1,0,1]=(0.103761462006)*x[0]**o+(-0.046916161864)*x[0]+(0.0312549239467)*x[1]**o+(0.214045118984)*x[1]+(-0.523161766402)*x[2]**o+(0.714207028317)*x[2]
        arg[1,1,1,0]=(0.494943183987)*x[0]**o+(-0.0543096386006)*x[0]+(-0.263875120588)*x[1]**o+(0.776031489925)*x[1]+(0.881224752145)*x[2]**o+(-0.650038283228)*x[2]
        arg[1,1,1,1]=(0.0871887540513)*x[0]**o+(-0.953848841349)*x[0]+(-0.223751459589)*x[1]**o+(0.227400823993)*x[1]+(-0.224200553009)*x[2]**o+(0.594012970899)*x[2]
        arg[1,1,2,0]=(-0.443116951346)*x[0]**o+(0.891806149626)*x[0]+(0.94731530917)*x[1]**o+(0.739499894809)*x[1]+(-0.92613053703)*x[2]**o+(0.464883087528)*x[2]
        arg[1,1,2,1]=(-0.149831145559)*x[0]**o+(0.913999704597)*x[0]+(0.76082978152)*x[1]**o+(-0.441535347622)*x[1]+(0.902841586936)*x[2]**o+(-0.509022004948)*x[2]
        arg[1,2,0,0]=(0.508230188489)*x[0]**o+(0.241669921231)*x[0]+(-0.802811365056)*x[1]**o+(0.298424305948)*x[1]+(-0.949407454532)*x[2]**o+(-0.131426880214)*x[2]
        arg[1,2,0,1]=(0.65372313274)*x[0]**o+(0.195891642771)*x[0]+(-0.160845429219)*x[1]**o+(0.500815823724)*x[1]+(0.875785249339)*x[2]**o+(0.325024339211)*x[2]
        arg[1,2,1,0]=(0.908405682536)*x[0]**o+(-0.146675166477)*x[0]+(0.24033815873)*x[1]**o+(-0.62814219771)*x[1]+(-0.512365788351)*x[2]**o+(-0.103817191415)*x[2]
        arg[1,2,1,1]=(0.824305237501)*x[0]**o+(0.175753992729)*x[0]+(0.992502114126)*x[1]**o+(0.973110627555)*x[1]+(0.355352002177)*x[2]**o+(0.112252263955)*x[2]
        arg[1,2,2,0]=(-0.485996285003)*x[0]**o+(-0.532551123887)*x[0]+(0.708205121405)*x[1]**o+(-0.803269038441)*x[1]+(-0.232049124534)*x[2]**o+(0.55437652866)*x[2]
        arg[1,2,2,1]=(0.551934776224)*x[0]**o+(0.727629664538)*x[0]+(0.197393511049)*x[1]**o+(-0.146619812174)*x[1]+(0.316592306936)*x[2]**o+(0.0263881493809)*x[2]
        arg[1,3,0,0]=(-0.918554166254)*x[0]**o+(-0.839150758884)*x[0]+(0.436713305329)*x[1]**o+(-0.410131731387)*x[1]+(0.321246028791)*x[2]**o+(0.706389384376)*x[2]
        arg[1,3,0,1]=(-0.870168753295)*x[0]**o+(-0.00473364206025)*x[0]+(0.594648540734)*x[1]**o+(-0.924887214906)*x[1]+(0.58858894515)*x[2]**o+(0.0152152159566)*x[2]
        arg[1,3,1,0]=(0.64005604234)*x[0]**o+(0.308265334836)*x[0]+(0.321645661286)*x[1]**o+(0.670017614875)*x[1]+(0.712764136703)*x[2]**o+(0.458986964342)*x[2]
        arg[1,3,1,1]=(-0.201376578944)*x[0]**o+(-0.271793865118)*x[0]+(-0.798181331582)*x[1]**o+(0.690100877945)*x[1]+(0.600350295622)*x[2]**o+(0.341066858436)*x[2]
        arg[1,3,2,0]=(0.459832113137)*x[0]**o+(0.0566780273877)*x[0]+(0.221723079112)*x[1]**o+(-0.681001728734)*x[1]+(0.265505135729)*x[2]**o+(-0.389365671367)*x[2]
        arg[1,3,2,1]=(0.715444384493)*x[0]**o+(-0.705697520786)*x[0]+(0.945208081248)*x[1]**o+(0.0233171867808)*x[1]+(0.740156166802)*x[2]**o+(-0.144314337012)*x[2]
        arg[1,4,0,0]=(-0.490162545806)*x[0]**o+(0.542981268396)*x[0]+(-0.841471465564)*x[1]**o+(0.79091798393)*x[1]+(0.245598142884)*x[2]**o+(0.4740296042)*x[2]
        arg[1,4,0,1]=(-0.504504615322)*x[0]**o+(0.45494067251)*x[0]+(-0.612427248728)*x[1]**o+(0.0880340051743)*x[1]+(0.645906227711)*x[2]**o+(0.772038416622)*x[2]
        arg[1,4,1,0]=(0.945538632492)*x[0]**o+(0.98519603491)*x[0]+(-0.245495451859)*x[1]**o+(-0.70732864026)*x[1]+(0.378733662069)*x[2]**o+(0.579554415018)*x[2]
        arg[1,4,1,1]=(0.886659064398)*x[0]**o+(-0.491870874438)*x[0]+(0.751814742427)*x[1]**o+(0.430589443767)*x[1]+(-0.729193959143)*x[2]**o+(-0.699344758538)*x[2]
        arg[1,4,2,0]=(-0.4190140481)*x[0]**o+(0.84509612452)*x[0]+(-0.263608911857)*x[1]**o+(0.693648282371)*x[1]+(-0.144463932166)*x[2]**o+(0.384293689266)*x[2]
        arg[1,4,2,1]=(-0.266973780787)*x[0]**o+(0.613988367707)*x[0]+(0.686955199789)*x[1]**o+(-0.301585818785)*x[1]+(-0.574601202558)*x[2]**o+(-0.430718616071)*x[2]
        arg[2,0,0,0]=(-0.117390727803)*x[0]**o+(-0.25850139145)*x[0]+(0.324909086031)*x[1]**o+(-0.593112241553)*x[1]+(-0.51016768177)*x[2]**o+(0.392299415334)*x[2]
        arg[2,0,0,1]=(0.290478531696)*x[0]**o+(0.758153023626)*x[0]+(0.531003993361)*x[1]**o+(-0.112481510374)*x[1]+(0.950222158203)*x[2]**o+(-0.742480526204)*x[2]
        arg[2,0,1,0]=(-0.754820921225)*x[0]**o+(-0.408482044077)*x[0]+(-0.229697426486)*x[1]**o+(-0.983010424713)*x[1]+(-0.286618604257)*x[2]**o+(-0.489034108451)*x[2]
        arg[2,0,1,1]=(0.96775757822)*x[0]**o+(0.797765725721)*x[0]+(-0.90936176648)*x[1]**o+(-0.372755885193)*x[1]+(-0.144727874143)*x[2]**o+(0.341624944551)*x[2]
        arg[2,0,2,0]=(0.672779422278)*x[0]**o+(-0.936102224964)*x[0]+(-0.763620352857)*x[1]**o+(0.265802509744)*x[1]+(0.275798875047)*x[2]**o+(0.235397520942)*x[2]
        arg[2,0,2,1]=(0.192866248041)*x[0]**o+(0.656863606616)*x[0]+(-0.902796948663)*x[1]**o+(0.692123089256)*x[1]+(0.311287987388)*x[2]**o+(0.555427507546)*x[2]
        arg[2,1,0,0]=(0.464585537593)*x[0]**o+(0.161744615208)*x[0]+(-0.325569920662)*x[1]**o+(-0.47142644441)*x[1]+(0.885151021784)*x[2]**o+(-0.840595219086)*x[2]
        arg[2,1,0,1]=(-0.659253397488)*x[0]**o+(0.54391314862)*x[0]+(-0.595612529952)*x[1]**o+(-0.0472758114379)*x[1]+(0.876391930188)*x[2]**o+(-0.440313199946)*x[2]
        arg[2,1,1,0]=(-0.445591423364)*x[0]**o+(0.179144841652)*x[0]+(-0.294166502086)*x[1]**o+(-0.373820778739)*x[1]+(0.464609314863)*x[2]**o+(0.407893781978)*x[2]
        arg[2,1,1,1]=(-0.767512700255)*x[0]**o+(0.63198312168)*x[0]+(0.652318598902)*x[1]**o+(0.817874971823)*x[1]+(0.397211574885)*x[2]**o+(0.33174354886)*x[2]
        arg[2,1,2,0]=(0.473801282131)*x[0]**o+(-0.778895701394)*x[0]+(-0.801488721847)*x[1]**o+(-0.585840161723)*x[1]+(-0.805496721223)*x[2]**o+(0.585971840421)*x[2]
        arg[2,1,2,1]=(0.472329620891)*x[0]**o+(-0.0473805663117)*x[0]+(0.0894117727522)*x[1]**o+(-0.308473625733)*x[1]+(-0.658376953956)*x[2]**o+(0.32027592346)*x[2]
        arg[2,2,0,0]=(0.843241796369)*x[0]**o+(0.346857959798)*x[0]+(0.0111223078269)*x[1]**o+(-0.949623227267)*x[1]+(-0.152362393376)*x[2]**o+(-0.886340001141)*x[2]
        arg[2,2,0,1]=(-0.555393018762)*x[0]**o+(0.666364344488)*x[0]+(0.181237902671)*x[1]**o+(0.670959040345)*x[1]+(0.57383285317)*x[2]**o+(-0.451027832775)*x[2]
        arg[2,2,1,0]=(0.249709446431)*x[0]**o+(0.568753063379)*x[0]+(-0.812386750097)*x[1]**o+(0.506805124672)*x[1]+(-0.454844217662)*x[2]**o+(0.843435269288)*x[2]
        arg[2,2,1,1]=(0.753144930025)*x[0]**o+(0.0181642998269)*x[0]+(0.622635241593)*x[1]**o+(-0.372883842874)*x[1]+(0.366920965577)*x[2]**o+(-0.092391524816)*x[2]
        arg[2,2,2,0]=(0.309601906611)*x[0]**o+(-0.502573192965)*x[0]+(-0.420795137333)*x[1]**o+(-0.807109356307)*x[1]+(-0.802887583918)*x[2]**o+(0.552609684817)*x[2]
        arg[2,2,2,1]=(0.320933223342)*x[0]**o+(0.541931544704)*x[0]+(0.988715316931)*x[1]**o+(-0.4197325237)*x[1]+(-0.438705512435)*x[2]**o+(-0.540175093599)*x[2]
        arg[2,3,0,0]=(0.018870015518)*x[0]**o+(-0.615120691773)*x[0]+(0.596295280542)*x[1]**o+(0.210874310884)*x[1]+(-0.652194502321)*x[2]**o+(0.74019575754)*x[2]
        arg[2,3,0,1]=(0.820285141622)*x[0]**o+(0.130985741886)*x[0]+(-0.397387767472)*x[1]**o+(0.975103051154)*x[1]+(0.0887369214611)*x[2]**o+(0.88498731668)*x[2]
        arg[2,3,1,0]=(0.160503336808)*x[0]**o+(0.641180113144)*x[0]+(-0.408109675737)*x[1]**o+(-0.355443951133)*x[1]+(-0.429622016495)*x[2]**o+(0.998729833779)*x[2]
        arg[2,3,1,1]=(-0.646524352112)*x[0]**o+(-0.289894880059)*x[0]+(0.909761879864)*x[1]**o+(0.30453651619)*x[1]+(0.884596709091)*x[2]**o+(0.329408737003)*x[2]
        arg[2,3,2,0]=(0.459385919432)*x[0]**o+(0.416447271916)*x[0]+(-0.400798659819)*x[1]**o+(-0.769043818187)*x[1]+(-0.631682400803)*x[2]**o+(-0.80150361084)*x[2]
        arg[2,3,2,1]=(-0.544824298969)*x[0]**o+(0.191301965562)*x[0]+(0.41061102918)*x[1]**o+(-0.999195759064)*x[1]+(-0.0226105975802)*x[2]**o+(-0.137078691725)*x[2]
        arg[2,4,0,0]=(-0.624033952375)*x[0]**o+(0.0792222332007)*x[0]+(0.766756718455)*x[1]**o+(0.581487379511)*x[1]+(-0.437940391245)*x[2]**o+(0.115617341492)*x[2]
        arg[2,4,0,1]=(-0.300526068838)*x[0]**o+(-0.822844066866)*x[0]+(0.660949310367)*x[1]**o+(-0.0993949179102)*x[1]+(0.599884615655)*x[2]**o+(-0.903177033523)*x[2]
        arg[2,4,1,0]=(-0.654642517091)*x[0]**o+(0.428391621765)*x[0]+(0.932377441464)*x[1]**o+(-0.459525189684)*x[1]+(0.667425199675)*x[2]**o+(-0.937213410686)*x[2]
        arg[2,4,1,1]=(0.258067525957)*x[0]**o+(0.619943768383)*x[0]+(0.458217567511)*x[1]**o+(0.0812191081554)*x[1]+(0.743260637332)*x[2]**o+(-0.796150472293)*x[2]
        arg[2,4,2,0]=(0.782473636556)*x[0]**o+(-0.441405613305)*x[0]+(0.3951595327)*x[1]**o+(-0.449855781187)*x[1]+(-0.964130179212)*x[2]**o+(-0.299199410525)*x[2]
        arg[2,4,2,1]=(-0.878340584397)*x[0]**o+(-0.674575785429)*x[0]+(0.838137461257)*x[1]**o+(-0.498710475715)*x[1]+(-0.117205127994)*x[2]**o+(0.0934669704542)*x[2]
        arg[3,0,0,0]=(0.139631366137)*x[0]**o+(-0.893001605238)*x[0]+(-0.420780773096)*x[1]**o+(0.439055863393)*x[1]+(0.173113363237)*x[2]**o+(0.893867443017)*x[2]
        arg[3,0,0,1]=(-0.804823042124)*x[0]**o+(0.250231181298)*x[0]+(0.358464048439)*x[1]**o+(-0.847391231172)*x[1]+(-0.515391975323)*x[2]**o+(-0.999640302937)*x[2]
        arg[3,0,1,0]=(0.328676968943)*x[0]**o+(-0.0521242139559)*x[0]+(0.861548411537)*x[1]**o+(-0.288979961428)*x[1]+(-0.927901849987)*x[2]**o+(0.857869958971)*x[2]
        arg[3,0,1,1]=(-0.654866876537)*x[0]**o+(-0.0515462321916)*x[0]+(-0.224164685097)*x[1]**o+(-0.84207399757)*x[1]+(-0.442392040219)*x[2]**o+(-0.915734076984)*x[2]
        arg[3,0,2,0]=(-0.322517406382)*x[0]**o+(-0.757146499161)*x[0]+(-0.541690197465)*x[1]**o+(0.931090695413)*x[1]+(-0.355505263744)*x[2]**o+(-0.773185404514)*x[2]
        arg[3,0,2,1]=(0.421030649833)*x[0]**o+(-0.83768939688)*x[0]+(-0.863791071447)*x[1]**o+(0.78393205545)*x[1]+(-0.489893911681)*x[2]**o+(-0.291989245709)*x[2]
        arg[3,1,0,0]=(0.550839051007)*x[0]**o+(-0.883853789938)*x[0]+(-0.339529971923)*x[1]**o+(-0.756936847344)*x[1]+(0.724888858656)*x[2]**o+(-0.981057762224)*x[2]
        arg[3,1,0,1]=(0.711792743156)*x[0]**o+(0.233248429378)*x[0]+(0.526076154408)*x[1]**o+(0.349364804166)*x[1]+(0.271475992486)*x[2]**o+(0.782429402713)*x[2]
        arg[3,1,1,0]=(-0.499072848146)*x[0]**o+(-0.0604888692308)*x[0]+(0.545442893272)*x[1]**o+(-0.688797522515)*x[1]+(0.305341295222)*x[2]**o+(0.130010601331)*x[2]
        arg[3,1,1,1]=(0.940695359452)*x[0]**o+(0.113015472294)*x[0]+(-0.506254395321)*x[1]**o+(-0.252044277039)*x[1]+(0.0172784972094)*x[2]**o+(-0.184256978193)*x[2]
        arg[3,1,2,0]=(0.550472143981)*x[0]**o+(-0.729809590776)*x[0]+(-0.293765999495)*x[1]**o+(-0.378042976721)*x[1]+(0.973242167993)*x[2]**o+(-0.75350391134)*x[2]
        arg[3,1,2,1]=(0.369266178598)*x[0]**o+(-0.764034492366)*x[0]+(-0.713325992328)*x[1]**o+(0.948932375717)*x[1]+(-0.335074977323)*x[2]**o+(0.117829591745)*x[2]
        arg[3,2,0,0]=(0.426458076699)*x[0]**o+(-0.520551329627)*x[0]+(-0.773464312884)*x[1]**o+(0.26332468989)*x[1]+(0.143497828773)*x[2]**o+(0.0676278275986)*x[2]
        arg[3,2,0,1]=(0.85806016531)*x[0]**o+(0.755361032731)*x[0]+(0.403942188112)*x[1]**o+(0.0113665816202)*x[1]+(-0.439835633198)*x[2]**o+(0.620930145183)*x[2]
        arg[3,2,1,0]=(-0.852838624821)*x[0]**o+(-0.918594624263)*x[0]+(-0.572898555139)*x[1]**o+(-0.14353390918)*x[1]+(-0.716010071307)*x[2]**o+(0.134687635305)*x[2]
        arg[3,2,1,1]=(-0.14282590746)*x[0]**o+(-0.169314199229)*x[0]+(0.880745511071)*x[1]**o+(0.540733710441)*x[1]+(-0.131424658303)*x[2]**o+(-0.93209431119)*x[2]
        arg[3,2,2,0]=(-0.864016075688)*x[0]**o+(-0.265794760131)*x[0]+(0.444625188371)*x[1]**o+(0.133693703972)*x[1]+(-0.749799597936)*x[2]**o+(0.886355820615)*x[2]
        arg[3,2,2,1]=(0.676281241167)*x[0]**o+(-0.351800298325)*x[0]+(-0.751498380099)*x[1]**o+(-0.561737786104)*x[1]+(-0.984872672853)*x[2]**o+(0.476840947707)*x[2]
        arg[3,3,0,0]=(0.457874366022)*x[0]**o+(-0.126929649966)*x[0]+(0.666077559139)*x[1]**o+(-0.966437669224)*x[1]+(-0.849000502816)*x[2]**o+(0.990790747926)*x[2]
        arg[3,3,0,1]=(0.45136859326)*x[0]**o+(0.771355456911)*x[0]+(-0.678558472676)*x[1]**o+(0.0366869703264)*x[1]+(-0.257247358952)*x[2]**o+(-0.290852285538)*x[2]
        arg[3,3,1,0]=(-0.623007069616)*x[0]**o+(-0.365255080044)*x[0]+(0.441927420573)*x[1]**o+(0.331640625664)*x[1]+(-0.21153951456)*x[2]**o+(0.297967669123)*x[2]
        arg[3,3,1,1]=(-0.836057742118)*x[0]**o+(0.959873293328)*x[0]+(-0.52034405502)*x[1]**o+(-0.178838686591)*x[1]+(0.317610563306)*x[2]**o+(-0.555134650833)*x[2]
        arg[3,3,2,0]=(0.0819108022024)*x[0]**o+(0.69432296821)*x[0]+(-0.23685434367)*x[1]**o+(-0.0442137744391)*x[1]+(-0.879175086057)*x[2]**o+(-0.28031950012)*x[2]
        arg[3,3,2,1]=(0.340083286763)*x[0]**o+(-0.381942838645)*x[0]+(0.570579443852)*x[1]**o+(0.772192708863)*x[1]+(0.42813556157)*x[2]**o+(0.735683065499)*x[2]
        arg[3,4,0,0]=(-0.866813692418)*x[0]**o+(-0.53715159706)*x[0]+(0.0725900211922)*x[1]**o+(0.0891167848614)*x[1]+(-0.381482922347)*x[2]**o+(-0.521504123034)*x[2]
        arg[3,4,0,1]=(0.487750922723)*x[0]**o+(0.967820473849)*x[0]+(0.305600886352)*x[1]**o+(0.350777657643)*x[1]+(-0.383731485788)*x[2]**o+(0.814016126476)*x[2]
        arg[3,4,1,0]=(-0.186663244344)*x[0]**o+(0.144872424379)*x[0]+(0.844454339304)*x[1]**o+(0.252133708483)*x[1]+(-0.748235508222)*x[2]**o+(0.702952185343)*x[2]
        arg[3,4,1,1]=(-0.962296047927)*x[0]**o+(0.874277032826)*x[0]+(-0.404689253219)*x[1]**o+(-0.174075799004)*x[1]+(0.65086350106)*x[2]**o+(-0.109850650677)*x[2]
        arg[3,4,2,0]=(-0.59515836027)*x[0]**o+(-0.0917296352911)*x[0]+(0.834779238353)*x[1]**o+(0.951252871777)*x[1]+(0.983008684497)*x[2]**o+(-0.784806954891)*x[2]
        arg[3,4,2,1]=(0.603110971708)*x[0]**o+(0.291507042984)*x[0]+(-0.401523286981)*x[1]**o+(-0.994760476862)*x[1]+(0.410516007475)*x[2]**o+(0.368982230001)*x[2]
        ref[0,0,0,0]=(-0.322384218211)/(o+1.)+(0.436168366323)
        ref[0,0,0,1]=(-0.550432608405)/(o+1.)+(-0.839529924682)
        ref[0,0,1,0]=(-1.20222931022)/(o+1.)+(-0.674931792543)
        ref[0,0,1,1]=(-1.99447702496)/(o+1.)+(0.184076712997)
        ref[0,0,2,0]=(-0.848786191606)/(o+1.)+(-0.719076228559)
        ref[0,0,2,1]=(0.488199570726)/(o+1.)+(0.818508727159)
        ref[0,1,0,0]=(0.910006770156)/(o+1.)+(-0.781077934994)
        ref[0,1,0,1]=(0.229140786917)/(o+1.)+(0.225201244712)
        ref[0,1,1,0]=(-1.24973037868)/(o+1.)+(0.478630842756)
        ref[0,1,1,1]=(0.362813222831)/(o+1.)+(-0.269799066354)
        ref[0,1,2,0]=(0.704186070044)/(o+1.)+(-0.0666602572428)
        ref[0,1,2,1]=(-0.429870561459)/(o+1.)+(-0.692214543685)
        ref[0,2,0,0]=(-1.76040621668)/(o+1.)+(-0.103077097544)
        ref[0,2,0,1]=(-0.578245990054)/(o+1.)+(-0.24050068432)
        ref[0,2,1,0]=(0.982824772144)/(o+1.)+(-0.332261654225)
        ref[0,2,1,1]=(0.493111487615)/(o+1.)+(0.105743857349)
        ref[0,2,2,0]=(-0.406637376689)/(o+1.)+(0.910652270393)
        ref[0,2,2,1]=(0.0735211634779)/(o+1.)+(0.355106669226)
        ref[0,3,0,0]=(-0.538705798782)/(o+1.)+(0.896367101094)
        ref[0,3,0,1]=(0.11388337928)/(o+1.)+(0.766078903586)
        ref[0,3,1,0]=(0.955296156262)/(o+1.)+(-0.64915883862)
        ref[0,3,1,1]=(0.437273793256)/(o+1.)+(-0.61952690992)
        ref[0,3,2,0]=(1.07390497644)/(o+1.)+(-0.242730112953)
        ref[0,3,2,1]=(-0.79053815352)/(o+1.)+(-0.0521715653697)
        ref[0,4,0,0]=(-0.0103557372704)/(o+1.)+(0.939327713906)
        ref[0,4,0,1]=(-0.66740052552)/(o+1.)+(-0.671249919496)
        ref[0,4,1,0]=(-1.10797150302)/(o+1.)+(0.454490777873)
        ref[0,4,1,1]=(-0.00636792544909)/(o+1.)+(-0.368286670996)
        ref[0,4,2,0]=(-1.23874065284)/(o+1.)+(0.0410396874718)
        ref[0,4,2,1]=(0.834902700591)/(o+1.)+(0.562757794439)
        ref[1,0,0,0]=(1.46682450791)/(o+1.)+(0.891962057632)
        ref[1,0,0,1]=(0.183329431084)/(o+1.)+(0.659083249781)
        ref[1,0,1,0]=(-0.24909411204)/(o+1.)+(0.201301697863)
        ref[1,0,1,1]=(0.723423340723)/(o+1.)+(-0.205765407152)
        ref[1,0,2,0]=(0.607815389082)/(o+1.)+(-0.451934578027)
        ref[1,0,2,1]=(0.919070217555)/(o+1.)+(1.00996326691)
        ref[1,1,0,0]=(0.361804510188)/(o+1.)+(-0.272105134113)
        ref[1,1,0,1]=(-0.388145380449)/(o+1.)+(0.440667992719)
        ref[1,1,1,0]=(1.11229281554)/(o+1.)+(0.0358417840478)
        ref[1,1,1,1]=(-0.360763258547)/(o+1.)+(-0.0662175232282)
        ref[1,1,2,0]=(-0.421932179206)/(o+1.)+(1.04809456598)
        ref[1,1,2,1]=(1.5138402229)/(o+1.)+(-0.0182788239862)
        ref[1,2,0,0]=(-1.2439886311)/(o+1.)+(0.204333673483)
        ref[1,2,0,1]=(1.36866295286)/(o+1.)+(0.510865902853)
        ref[1,2,1,0]=(0.636378052915)/(o+1.)+(-0.439317277801)
        ref[1,2,1,1]=(2.1721593538)/(o+1.)+(0.63055844212)
        ref[1,2,2,0]=(-0.00984028813171)/(o+1.)+(-0.390721816834)
        ref[1,2,2,1]=(1.06592059421)/(o+1.)+(0.303699000873)
        ref[1,3,0,0]=(-0.160594832134)/(o+1.)+(-0.271446552947)
        ref[1,3,0,1]=(0.313068732589)/(o+1.)+(-0.457202820505)
        ref[1,3,1,0]=(1.67446584033)/(o+1.)+(0.718634957026)
        ref[1,3,1,1]=(-0.399207614903)/(o+1.)+(0.379686935632)
        ref[1,3,2,0]=(0.947060327978)/(o+1.)+(-0.506844686357)
        ref[1,3,2,1]=(2.40080863254)/(o+1.)+(-0.413347335508)
        ref[1,4,0,0]=(-1.08603586849)/(o+1.)+(0.903964428263)
        ref[1,4,0,1]=(-0.471025636339)/(o+1.)+(0.657506547153)
        ref[1,4,1,0]=(1.0787768427)/(o+1.)+(0.428710904834)
        ref[1,4,1,1]=(0.909279847682)/(o+1.)+(-0.380313094604)
        ref[1,4,2,0]=(-0.827086892123)/(o+1.)+(0.961519048079)
        ref[1,4,2,1]=(-0.154619783556)/(o+1.)+(-0.0591580335741)
        ref[2,0,0,0]=(-0.302649323542)/(o+1.)+(-0.229657108834)
        ref[2,0,0,1]=(1.77170468326)/(o+1.)+(-0.0484045064764)
        ref[2,0,1,0]=(-1.27113695197)/(o+1.)+(-0.94026328862)
        ref[2,0,1,1]=(-0.0863320624027)/(o+1.)+(0.383317392539)
        ref[2,0,2,0]=(0.184957944467)/(o+1.)+(-0.217451097139)
        ref[2,0,2,1]=(-0.398642713235)/(o+1.)+(0.952207101709)
        ref[2,1,0,0]=(1.02416663872)/(o+1.)+(-0.575138524144)
        ref[2,1,0,1]=(-0.378473997252)/(o+1.)+(0.0281620686178)
        ref[2,1,1,0]=(-0.275148610588)/(o+1.)+(0.106608922446)
        ref[2,1,1,1]=(0.282017473532)/(o+1.)+(0.890800821181)
        ref[2,1,2,0]=(-1.13318416094)/(o+1.)+(-0.389382011348)
        ref[2,1,2,1]=(-0.0966355603128)/(o+1.)+(-0.0177891342927)
        ref[2,2,0,0]=(0.70200171082)/(o+1.)+(-0.744552634305)
        ref[2,2,0,1]=(0.199677737079)/(o+1.)+(0.443147776029)
        ref[2,2,1,0]=(-1.01752152133)/(o+1.)+(0.959496728669)
        ref[2,2,1,1]=(1.74270113719)/(o+1.)+(-0.223555533932)
        ref[2,2,2,0]=(-0.91408081464)/(o+1.)+(-0.378536432227)
        ref[2,2,2,1]=(0.870943027839)/(o+1.)+(-0.208988036298)
        ref[2,3,0,0]=(-0.0370292062615)/(o+1.)+(0.167974688326)
        ref[2,3,0,1]=(0.51163429561)/(o+1.)+(0.99553805486)
        ref[2,3,1,0]=(-0.677228355424)/(o+1.)+(0.642232997895)
        ref[2,3,1,1]=(1.14783423684)/(o+1.)+(0.172025186567)
        ref[2,3,2,0]=(-0.57309514119)/(o+1.)+(-0.577050078555)
        ref[2,3,2,1]=(-0.156823867369)/(o+1.)+(-0.472486242613)
        ref[2,4,0,0]=(-0.295217625165)/(o+1.)+(0.388163477102)
        ref[2,4,0,1]=(0.960307857184)/(o+1.)+(-0.91270800915)
        ref[2,4,1,0]=(0.945160124047)/(o+1.)+(-0.484173489303)
        ref[2,4,1,1]=(1.4595457308)/(o+1.)+(-0.0474937978771)
        ref[2,4,2,0]=(0.213502990044)/(o+1.)+(-0.595230402509)
        ref[2,4,2,1]=(-0.157408251134)/(o+1.)+(-0.539909645345)
        ref[3,0,0,0]=(-0.108036043722)/(o+1.)+(0.219960850586)
        ref[3,0,0,1]=(-0.961750969008)/(o+1.)+(-0.798400176406)
        ref[3,0,1,0]=(0.262323530493)/(o+1.)+(0.258382891794)
        ref[3,0,1,1]=(-1.32142360185)/(o+1.)+(-0.904677153373)
        ref[3,0,2,0]=(-1.21971286759)/(o+1.)+(-0.299620604131)
        ref[3,0,2,1]=(-0.932654333294)/(o+1.)+(-0.17287329357)
        ref[3,1,0,0]=(0.93619793774)/(o+1.)+(-1.31092419975)
        ref[3,1,0,1]=(1.50934489005)/(o+1.)+(0.682521318129)
        ref[3,1,1,0]=(0.351711340348)/(o+1.)+(-0.309637895207)
        ref[3,1,1,1]=(0.451719461341)/(o+1.)+(-0.161642891469)
        ref[3,1,2,0]=(1.22994831248)/(o+1.)+(-0.930678239419)
        ref[3,1,2,1]=(-0.679134791053)/(o+1.)+(0.151363737548)
        ref[3,2,0,0]=(-0.203508407413)/(o+1.)+(-0.0947994060691)
        ref[3,2,0,1]=(0.822166720224)/(o+1.)+(0.693828879767)
        ref[3,2,1,0]=(-2.14174725127)/(o+1.)+(-0.463720449069)
        ref[3,2,1,1]=(0.606494945308)/(o+1.)+(-0.280337399989)
        ref[3,2,2,0]=(-1.16919048525)/(o+1.)+(0.377127382228)
        ref[3,2,2,1]=(-1.06008981178)/(o+1.)+(-0.218348568361)
        ref[3,3,0,0]=(0.274951422345)/(o+1.)+(-0.0512882856318)
        ref[3,3,0,1]=(-0.484437238367)/(o+1.)+(0.25859507085)
        ref[3,3,1,0]=(-0.392619163603)/(o+1.)+(0.132176607371)
        ref[3,3,1,1]=(-1.03879123383)/(o+1.)+(0.112949977952)
        ref[3,3,2,0]=(-1.03411862752)/(o+1.)+(0.184894846825)
        ref[3,3,2,1]=(1.33879829218)/(o+1.)+(0.562966467859)
        ref[3,4,0,0]=(-1.17570659357)/(o+1.)+(-0.484769467616)
        ref[3,4,0,1]=(0.409620323287)/(o+1.)+(1.06630712898)
        ref[3,4,1,0]=(-0.0904444132627)/(o+1.)+(0.549979159103)
        ref[3,4,1,1]=(-0.716121800087)/(o+1.)+(0.295175291572)
        ref[3,4,2,0]=(1.22262956258)/(o+1.)+(0.0373581407971)
        ref[3,4,2,1]=(0.612103692201)/(o+1.)+(-0.167135601938)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.465976690369)*x[0]**o+(-0.25136070196)*x[0]+(0.409576841566)*x[1]**o+(0.726707772754)*x[1]
        ref=(0.875553531935)/(o+1.)+(0.237673535397)
      else:
        arg=(0.567785895928)*x[0]**o+(0.928477128396)*x[0]+(-0.625720943088)*x[1]**o+(-0.681460841298)*x[1]+(-0.866801010069)*x[2]**o+(0.159594300637)*x[2]
        ref=(-0.924736057229)/(o+1.)+(0.203305293867)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.839020088668)*x[0]**o+(-0.880288035177)*x[0]+(0.260951734364)*x[1]**o+(-0.624208000982)*x[1]
        arg[1]=(0.985421831314)*x[0]**o+(0.660399229451)*x[0]+(-0.514003848134)*x[1]**o+(0.0312162082593)*x[1]
        ref[0]=(-0.578068354304)/(o+1.)+(-0.75224801808)
        ref[1]=(0.47141798318)/(o+1.)+(0.345807718855)
      else:
        arg[0]=(-0.627307316929)*x[0]**o+(0.500432411035)*x[0]+(0.233895908313)*x[1]**o+(-0.125048135432)*x[1]+(0.0695114009632)*x[2]**o+(0.995170977074)*x[2]
        arg[1]=(0.760033715785)*x[0]**o+(0.48895905139)*x[0]+(0.240372023176)*x[1]**o+(0.33946689896)*x[1]+(-0.774638489419)*x[2]**o+(-0.0777240871074)*x[2]
        ref[0]=(-0.323900007653)/(o+1.)+(0.685277626338)
        ref[1]=(0.225767249542)/(o+1.)+(0.375350931622)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.844492048883)*x[0]**o+(0.727132679659)*x[0]+(-0.862576284722)*x[1]**o+(0.502937561417)*x[1]
        arg[0,1]=(-0.587157419309)*x[0]**o+(0.00327068097107)*x[0]+(0.422592349683)*x[1]**o+(0.167479615369)*x[1]
        arg[0,2]=(0.771156665672)*x[0]**o+(-0.371091138833)*x[0]+(-0.867750378658)*x[1]**o+(-0.744394290185)*x[1]
        arg[0,3]=(0.669718598855)*x[0]**o+(-0.0825909409121)*x[0]+(0.868872948791)*x[1]**o+(0.141891982204)*x[1]
        arg[0,4]=(-0.441072764653)*x[0]**o+(-0.929654649265)*x[0]+(-0.35883498816)*x[1]**o+(-0.84045480381)*x[1]
        arg[1,0]=(0.555668196622)*x[0]**o+(0.94306681043)*x[0]+(0.586512446245)*x[1]**o+(0.702007276327)*x[1]
        arg[1,1]=(0.792083771005)*x[0]**o+(-0.865830139617)*x[0]+(0.0929971631615)*x[1]**o+(0.620446289121)*x[1]
        arg[1,2]=(-0.201978375472)*x[0]**o+(-0.0777206885918)*x[0]+(0.798755459687)*x[1]**o+(-0.278086468846)*x[1]
        arg[1,3]=(0.063028419572)*x[0]**o+(-0.454786232294)*x[0]+(-0.0481393324293)*x[1]**o+(-0.612657132834)*x[1]
        arg[1,4]=(-0.743283787964)*x[0]**o+(0.854827303846)*x[0]+(0.94241635825)*x[1]**o+(0.436937526521)*x[1]
        arg[2,0]=(-0.814975695669)*x[0]**o+(0.579477783769)*x[0]+(0.00998264005965)*x[1]**o+(-0.913665878657)*x[1]
        arg[2,1]=(-0.526833976369)*x[0]**o+(-0.800386086232)*x[0]+(-0.487600053086)*x[1]**o+(0.00242562519003)*x[1]
        arg[2,2]=(-0.964119591874)*x[0]**o+(0.203212394214)*x[0]+(-0.96850837954)*x[1]**o+(-0.214886132512)*x[1]
        arg[2,3]=(-0.356249104564)*x[0]**o+(0.329769293221)*x[0]+(0.936664928754)*x[1]**o+(0.520603503283)*x[1]
        arg[2,4]=(0.255464219424)*x[0]**o+(-0.157783294516)*x[0]+(0.519451787865)*x[1]**o+(0.00960521397376)*x[1]
        arg[3,0]=(-0.117250486034)*x[0]**o+(0.192945431273)*x[0]+(-0.7580154297)*x[1]**o+(-0.308801650294)*x[1]
        arg[3,1]=(0.243931002845)*x[0]**o+(-0.626714191745)*x[0]+(-0.365870571265)*x[1]**o+(0.652257582021)*x[1]
        arg[3,2]=(-0.769756656122)*x[0]**o+(0.711106390017)*x[0]+(0.150749630653)*x[1]**o+(-0.0441309732448)*x[1]
        arg[3,3]=(0.92677837493)*x[0]**o+(0.507199859313)*x[0]+(-0.542275490828)*x[1]**o+(-0.0907270087228)*x[1]
        arg[3,4]=(-0.0306175885885)*x[0]**o+(-0.434666505018)*x[0]+(-0.815633427562)*x[1]**o+(-0.916424797903)*x[1]
        ref[0,0]=(-1.70706833361)/(o+1.)+(0.615035120538)
        ref[0,1]=(-0.164565069626)/(o+1.)+(0.08537514817)
        ref[0,2]=(-0.0965937129851)/(o+1.)+(-0.557742714509)
        ref[0,3]=(1.53859154765)/(o+1.)+(0.0296505206459)
        ref[0,4]=(-0.799907752814)/(o+1.)+(-0.885054726537)
        ref[1,0]=(1.14218064287)/(o+1.)+(0.822537043379)
        ref[1,1]=(0.885080934167)/(o+1.)+(-0.122691925248)
        ref[1,2]=(0.596777084215)/(o+1.)+(-0.177903578719)
        ref[1,3]=(0.0148890871427)/(o+1.)+(-0.533721682564)
        ref[1,4]=(0.199132570285)/(o+1.)+(0.645882415184)
        ref[2,0]=(-0.80499305561)/(o+1.)+(-0.167094047444)
        ref[2,1]=(-1.01443402945)/(o+1.)+(-0.398980230521)
        ref[2,2]=(-1.93262797141)/(o+1.)+(-0.00583686914904)
        ref[2,3]=(0.580415824189)/(o+1.)+(0.425186398252)
        ref[2,4]=(0.774916007289)/(o+1.)+(-0.074089040271)
        ref[3,0]=(-0.875265915734)/(o+1.)+(-0.0579281095106)
        ref[3,1]=(-0.12193956842)/(o+1.)+(0.0127716951382)
        ref[3,2]=(-0.619007025469)/(o+1.)+(0.333487708386)
        ref[3,3]=(0.384502884102)/(o+1.)+(0.208236425295)
        ref[3,4]=(-0.846251016151)/(o+1.)+(-0.675545651461)
      else:
        arg[0,0]=(0.596538443259)*x[0]**o+(0.479462980373)*x[0]+(-0.334469126403)*x[1]**o+(0.784201085556)*x[1]+(-0.231474140531)*x[2]**o+(-0.368096345589)*x[2]
        arg[0,1]=(0.188945383338)*x[0]**o+(-0.95033024122)*x[0]+(0.567904901047)*x[1]**o+(-0.76703296351)*x[1]+(-0.0116083690376)*x[2]**o+(-0.211176589764)*x[2]
        arg[0,2]=(0.225412552449)*x[0]**o+(-0.731615155924)*x[0]+(-0.362727501637)*x[1]**o+(-0.195442119344)*x[1]+(-0.215444456285)*x[2]**o+(0.412715796392)*x[2]
        arg[0,3]=(-0.00941287481408)*x[0]**o+(-0.207271678463)*x[0]+(0.66582433583)*x[1]**o+(-0.297871484308)*x[1]+(-0.819460465536)*x[2]**o+(-0.3316307033)*x[2]
        arg[0,4]=(-0.172265683581)*x[0]**o+(0.881658164973)*x[0]+(0.0267816673708)*x[1]**o+(-0.526757427521)*x[1]+(-0.666953449252)*x[2]**o+(0.366533690002)*x[2]
        arg[1,0]=(-0.575272180021)*x[0]**o+(-0.707236339163)*x[0]+(-0.734723292219)*x[1]**o+(-0.872019568084)*x[1]+(-0.513269190461)*x[2]**o+(-0.95077853535)*x[2]
        arg[1,1]=(0.687652562435)*x[0]**o+(0.500668143461)*x[0]+(0.127678854367)*x[1]**o+(0.135196530391)*x[1]+(0.219757431874)*x[2]**o+(0.43373842168)*x[2]
        arg[1,2]=(-0.674120744855)*x[0]**o+(-0.792769889708)*x[0]+(0.0414602844391)*x[1]**o+(-0.159887164563)*x[1]+(0.406441555189)*x[2]**o+(-0.662210077269)*x[2]
        arg[1,3]=(-0.623120612358)*x[0]**o+(0.895951757448)*x[0]+(0.957407224559)*x[1]**o+(-0.472225118848)*x[1]+(-0.294551998847)*x[2]**o+(0.503584320886)*x[2]
        arg[1,4]=(-0.304220645545)*x[0]**o+(0.812321371589)*x[0]+(0.932257196386)*x[1]**o+(0.736014913915)*x[1]+(0.119140391025)*x[2]**o+(0.910696940879)*x[2]
        arg[2,0]=(0.275470013797)*x[0]**o+(-0.903754203548)*x[0]+(-0.443211232771)*x[1]**o+(0.78815938507)*x[1]+(-0.0802934106722)*x[2]**o+(0.096532189886)*x[2]
        arg[2,1]=(-0.4505392488)*x[0]**o+(-0.668542586506)*x[0]+(0.487967999984)*x[1]**o+(-0.642310745404)*x[1]+(-0.956530771836)*x[2]**o+(0.517544404873)*x[2]
        arg[2,2]=(-0.984032885692)*x[0]**o+(-0.606688527142)*x[0]+(0.0565483776156)*x[1]**o+(-0.945228520749)*x[1]+(0.433070187921)*x[2]**o+(-0.599983370996)*x[2]
        arg[2,3]=(-0.673469008741)*x[0]**o+(0.612144012035)*x[0]+(-0.169497339443)*x[1]**o+(0.267990057152)*x[1]+(0.431716413548)*x[2]**o+(0.188904436012)*x[2]
        arg[2,4]=(0.320782309844)*x[0]**o+(-0.830068082984)*x[0]+(0.900183436056)*x[1]**o+(-0.343951186189)*x[1]+(-0.00137135137854)*x[2]**o+(-0.655764522249)*x[2]
        arg[3,0]=(-0.936193981767)*x[0]**o+(0.885260284217)*x[0]+(0.00397725971098)*x[1]**o+(-0.0864751234501)*x[1]+(0.25001166345)*x[2]**o+(-0.127859505667)*x[2]
        arg[3,1]=(-0.0841986843231)*x[0]**o+(-0.95678469673)*x[0]+(0.867049114883)*x[1]**o+(0.154499266839)*x[1]+(0.0217146474622)*x[2]**o+(-0.945774319527)*x[2]
        arg[3,2]=(-0.668161122306)*x[0]**o+(0.779140555525)*x[0]+(0.217613826157)*x[1]**o+(0.900077039607)*x[1]+(-0.251881965789)*x[2]**o+(-0.331281723354)*x[2]
        arg[3,3]=(-0.95624079226)*x[0]**o+(-0.403226291225)*x[0]+(0.303340462877)*x[1]**o+(-0.269819286355)*x[1]+(0.605804508829)*x[2]**o+(-0.461404715818)*x[2]
        arg[3,4]=(-0.933059329973)*x[0]**o+(-0.00851461368762)*x[0]+(0.0588544443918)*x[1]**o+(-0.682939702389)*x[1]+(-0.213965742261)*x[2]**o+(-0.102965977964)*x[2]
        ref[0,0]=(0.0305951763245)/(o+1.)+(0.44778386017)
        ref[0,1]=(0.745241915348)/(o+1.)+(-0.964269897247)
        ref[0,2]=(-0.352759405473)/(o+1.)+(-0.257170739438)
        ref[0,3]=(-0.163049004521)/(o+1.)+(-0.418386933036)
        ref[0,4]=(-0.812437465462)/(o+1.)+(0.360717213727)
        ref[1,0]=(-1.8232646627)/(o+1.)+(-1.2650172213)
        ref[1,1]=(1.03508884868)/(o+1.)+(0.534801547766)
        ref[1,2]=(-0.226218905227)/(o+1.)+(-0.80743356577)
        ref[1,3]=(0.0397346133545)/(o+1.)+(0.463655479743)
        ref[1,4]=(0.747176941866)/(o+1.)+(1.22951661319)
        ref[2,0]=(-0.248034629647)/(o+1.)+(-0.00953131429598)
        ref[2,1]=(-0.919102020652)/(o+1.)+(-0.396654463518)
        ref[2,2]=(-0.494414320156)/(o+1.)+(-1.07595020944)
        ref[2,3]=(-0.411249934637)/(o+1.)+(0.534519252599)
        ref[2,4]=(1.21959439452)/(o+1.)+(-0.914891895711)
        ref[3,0]=(-0.682205058606)/(o+1.)+(0.33546282755)
        ref[3,1]=(0.804565078022)/(o+1.)+(-0.874029874709)
        ref[3,2]=(-0.702429261938)/(o+1.)+(0.673967935889)
        ref[3,3]=(-0.0470958205546)/(o+1.)+(-0.567225146699)
        ref[3,4]=(-1.08817062784)/(o+1.)+(-0.39721014702)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.401537399178)*x[0]**o+(-0.868937065568)*x[0]+(0.356734884346)*x[1]**o+(0.0137485089422)*x[1]
        arg[0,0,1]=(0.705383698155)*x[0]**o+(-0.884988282352)*x[0]+(-0.979185994641)*x[1]**o+(0.929576942519)*x[1]
        arg[0,1,0]=(-0.292540086272)*x[0]**o+(-0.223230840915)*x[0]+(0.717291109855)*x[1]**o+(-0.826233591275)*x[1]
        arg[0,1,1]=(0.55739919643)*x[0]**o+(0.2571407205)*x[0]+(0.00434404142287)*x[1]**o+(0.200305144451)*x[1]
        arg[1,0,0]=(0.451841947443)*x[0]**o+(-0.576279050146)*x[0]+(-0.128142118055)*x[1]**o+(0.316406960557)*x[1]
        arg[1,0,1]=(-0.07174792925)*x[0]**o+(-0.392935435009)*x[0]+(-0.343535485172)*x[1]**o+(0.619131193274)*x[1]
        arg[1,1,0]=(-0.667031669611)*x[0]**o+(-0.184806094874)*x[0]+(-0.199375296384)*x[1]**o+(0.449351889815)*x[1]
        arg[1,1,1]=(-0.51924508404)*x[0]**o+(0.870508122509)*x[0]+(-0.274149433148)*x[1]**o+(-0.762367215962)*x[1]
        arg[2,0,0]=(-0.501546909709)*x[0]**o+(0.164122129343)*x[0]+(-0.601370828214)*x[1]**o+(0.888046948888)*x[1]
        arg[2,0,1]=(0.272472486899)*x[0]**o+(-0.958776296738)*x[0]+(-0.334676045163)*x[1]**o+(0.410943596779)*x[1]
        arg[2,1,0]=(0.363478201368)*x[0]**o+(0.753492598803)*x[0]+(-0.663726768274)*x[1]**o+(0.875443816963)*x[1]
        arg[2,1,1]=(0.995788536211)*x[0]**o+(-0.123932097555)*x[0]+(0.143025171293)*x[1]**o+(0.724918059214)*x[1]
        arg[3,0,0]=(-0.578717829755)*x[0]**o+(-0.899523186929)*x[0]+(0.58523203854)*x[1]**o+(0.751620714771)*x[1]
        arg[3,0,1]=(0.179841033744)*x[0]**o+(0.268015839688)*x[0]+(0.860292991989)*x[1]**o+(0.0117969420991)*x[1]
        arg[3,1,0]=(-0.575777000777)*x[0]**o+(0.170047012345)*x[0]+(-0.418743243278)*x[1]**o+(-0.278415849566)*x[1]
        arg[3,1,1]=(-0.913076795648)*x[0]**o+(0.486330450107)*x[0]+(-0.423570363453)*x[1]**o+(0.388575474946)*x[1]
        arg[4,0,0]=(0.16695368621)*x[0]**o+(0.133497460585)*x[0]+(-0.444434659151)*x[1]**o+(-0.895229636275)*x[1]
        arg[4,0,1]=(0.432018868106)*x[0]**o+(-0.281976075625)*x[0]+(-0.0455135209769)*x[1]**o+(0.76870037059)*x[1]
        arg[4,1,0]=(0.238542836112)*x[0]**o+(-0.806931884409)*x[0]+(-0.0365533871189)*x[1]**o+(0.731336634477)*x[1]
        arg[4,1,1]=(0.512035255059)*x[0]**o+(-0.683560978902)*x[0]+(-0.953782277737)*x[1]**o+(-0.311664210765)*x[1]
        arg[5,0,0]=(-0.842399299624)*x[0]**o+(0.460676816656)*x[0]+(-0.251781603841)*x[1]**o+(0.321792359074)*x[1]
        arg[5,0,1]=(-0.900576257062)*x[0]**o+(0.842393710528)*x[0]+(0.630445446843)*x[1]**o+(-0.647745144739)*x[1]
        arg[5,1,0]=(0.136687908652)*x[0]**o+(0.0252907890641)*x[0]+(0.813054611844)*x[1]**o+(-0.459930787056)*x[1]
        arg[5,1,1]=(-0.00242518550941)*x[0]**o+(-0.558785348137)*x[0]+(0.0568659980522)*x[1]**o+(0.689293614051)*x[1]
        ref[0,0,0]=(0.758272283524)/(o+1.)+(-0.427594278313)
        ref[0,0,1]=(-0.273802296487)/(o+1.)+(0.0222943300833)
        ref[0,1,0]=(0.424751023583)/(o+1.)+(-0.524732216095)
        ref[0,1,1]=(0.561743237852)/(o+1.)+(0.228722932476)
        ref[1,0,0]=(0.323699829388)/(o+1.)+(-0.129936044795)
        ref[1,0,1]=(-0.415283414421)/(o+1.)+(0.113097879133)
        ref[1,1,0]=(-0.866406965995)/(o+1.)+(0.132272897471)
        ref[1,1,1]=(-0.793394517189)/(o+1.)+(0.0540704532731)
        ref[2,0,0]=(-1.10291773792)/(o+1.)+(0.526084539115)
        ref[2,0,1]=(-0.0622035582637)/(o+1.)+(-0.273916349979)
        ref[2,1,0]=(-0.300248566906)/(o+1.)+(0.814468207883)
        ref[2,1,1]=(1.1388137075)/(o+1.)+(0.300492980829)
        ref[3,0,0]=(0.00651420878498)/(o+1.)+(-0.073951236079)
        ref[3,0,1]=(1.04013402573)/(o+1.)+(0.139906390894)
        ref[3,1,0]=(-0.994520244055)/(o+1.)+(-0.0541844186105)
        ref[3,1,1]=(-1.3366471591)/(o+1.)+(0.437452962527)
        ref[4,0,0]=(-0.27748097294)/(o+1.)+(-0.380866087845)
        ref[4,0,1]=(0.386505347129)/(o+1.)+(0.243362147483)
        ref[4,1,0]=(0.201989448993)/(o+1.)+(-0.0377976249661)
        ref[4,1,1]=(-0.441747022678)/(o+1.)+(-0.497612594833)
        ref[5,0,0]=(-1.09418090346)/(o+1.)+(0.391234587865)
        ref[5,0,1]=(-0.270130810219)/(o+1.)+(0.0973242828949)
        ref[5,1,0]=(0.949742520497)/(o+1.)+(-0.217319998996)
        ref[5,1,1]=(0.0544408125428)/(o+1.)+(0.0652541329568)
      else:
        arg[0,0,0]=(0.411788818613)*x[0]**o+(0.324151698669)*x[0]+(0.656551349616)*x[1]**o+(0.801577401197)*x[1]+(0.690440431022)*x[2]**o+(0.776409870151)*x[2]
        arg[0,0,1]=(-0.224985947659)*x[0]**o+(-0.564703086188)*x[0]+(-0.37196336116)*x[1]**o+(-0.114979853172)*x[1]+(0.274462080499)*x[2]**o+(-0.988489021983)*x[2]
        arg[0,1,0]=(0.512331530009)*x[0]**o+(0.519617843594)*x[0]+(-0.515174861158)*x[1]**o+(-0.675953902994)*x[1]+(0.796111490716)*x[2]**o+(0.229569489858)*x[2]
        arg[0,1,1]=(0.347765144376)*x[0]**o+(-0.0110194730031)*x[0]+(-0.593936645599)*x[1]**o+(-0.372786558089)*x[1]+(-0.604691321362)*x[2]**o+(-0.888002521077)*x[2]
        arg[1,0,0]=(-0.575954903433)*x[0]**o+(0.0160235655616)*x[0]+(0.305335632724)*x[1]**o+(-0.259402553796)*x[1]+(-0.221276854741)*x[2]**o+(-0.519932341069)*x[2]
        arg[1,0,1]=(-0.674987647238)*x[0]**o+(-0.45852667934)*x[0]+(-0.758937440978)*x[1]**o+(0.80566898197)*x[1]+(-0.961619154322)*x[2]**o+(0.861564430243)*x[2]
        arg[1,1,0]=(-0.297475412437)*x[0]**o+(0.881828705604)*x[0]+(-0.823267615874)*x[1]**o+(-0.0232728101313)*x[1]+(-0.676220001453)*x[2]**o+(-0.382414036981)*x[2]
        arg[1,1,1]=(-0.31008204616)*x[0]**o+(0.805195900401)*x[0]+(-0.758209772143)*x[1]**o+(0.861379709073)*x[1]+(0.0631063866518)*x[2]**o+(0.356607080553)*x[2]
        arg[2,0,0]=(0.305214964591)*x[0]**o+(0.815106526139)*x[0]+(0.138797156005)*x[1]**o+(-0.241921896332)*x[1]+(-0.103783035058)*x[2]**o+(-0.161952044984)*x[2]
        arg[2,0,1]=(-0.925375677635)*x[0]**o+(0.0341347568201)*x[0]+(-0.706905924964)*x[1]**o+(-0.193304368451)*x[1]+(-0.449866757104)*x[2]**o+(-0.894963680869)*x[2]
        arg[2,1,0]=(0.822676224279)*x[0]**o+(-0.28758006682)*x[0]+(-0.869025918917)*x[1]**o+(0.0107602044744)*x[1]+(-0.832322765163)*x[2]**o+(0.748440319389)*x[2]
        arg[2,1,1]=(-0.415326508997)*x[0]**o+(-0.879521424817)*x[0]+(-0.971781441062)*x[1]**o+(0.527586634108)*x[1]+(0.336682669856)*x[2]**o+(-0.948414041149)*x[2]
        arg[3,0,0]=(-0.345788531715)*x[0]**o+(0.626583498807)*x[0]+(-0.186418683649)*x[1]**o+(0.3141628424)*x[1]+(-0.781119787783)*x[2]**o+(0.972708430266)*x[2]
        arg[3,0,1]=(0.175536279095)*x[0]**o+(-0.150027774295)*x[0]+(-0.506653680683)*x[1]**o+(-0.482544397029)*x[1]+(-0.59090996152)*x[2]**o+(0.618123625026)*x[2]
        arg[3,1,0]=(0.737984493124)*x[0]**o+(0.650901680297)*x[0]+(-0.361389200009)*x[1]**o+(-0.784304665396)*x[1]+(-0.954132232142)*x[2]**o+(0.684341934348)*x[2]
        arg[3,1,1]=(-0.385757579645)*x[0]**o+(0.251136919845)*x[0]+(0.95143939933)*x[1]**o+(0.501860043225)*x[1]+(-0.675673011184)*x[2]**o+(0.337315450579)*x[2]
        arg[4,0,0]=(-0.740183402139)*x[0]**o+(-0.828278760743)*x[0]+(0.88000889117)*x[1]**o+(-0.0226801022184)*x[1]+(0.6811397662)*x[2]**o+(-0.254627446498)*x[2]
        arg[4,0,1]=(-0.194128952361)*x[0]**o+(-0.524854896388)*x[0]+(0.341100326577)*x[1]**o+(-0.26072443499)*x[1]+(-0.197396331776)*x[2]**o+(-0.589037632112)*x[2]
        arg[4,1,0]=(0.261995024865)*x[0]**o+(0.214193793699)*x[0]+(0.572024600244)*x[1]**o+(0.230659327198)*x[1]+(0.420971705453)*x[2]**o+(0.333445049827)*x[2]
        arg[4,1,1]=(-0.0679525054395)*x[0]**o+(0.0756646296473)*x[0]+(-0.299865836239)*x[1]**o+(-0.828294764914)*x[1]+(0.606170745631)*x[2]**o+(0.624267827393)*x[2]
        arg[5,0,0]=(0.312565542211)*x[0]**o+(-0.368815496171)*x[0]+(-0.0615973473105)*x[1]**o+(0.205012835971)*x[1]+(0.310818553076)*x[2]**o+(-0.287393908304)*x[2]
        arg[5,0,1]=(-0.409626309069)*x[0]**o+(0.977953573562)*x[0]+(0.583013544138)*x[1]**o+(0.542907128677)*x[1]+(0.488538336366)*x[2]**o+(-0.85823611123)*x[2]
        arg[5,1,0]=(-0.899607109395)*x[0]**o+(-0.0459053040522)*x[0]+(0.938423293041)*x[1]**o+(-0.112431665317)*x[1]+(0.620609328687)*x[2]**o+(-0.663082429319)*x[2]
        arg[5,1,1]=(0.00370253144626)*x[0]**o+(0.503733497956)*x[0]+(0.362457785043)*x[1]**o+(0.417891957303)*x[1]+(0.404543677829)*x[2]**o+(0.117828669122)*x[2]
        ref[0,0,0]=(1.75878059925)/(o+1.)+(0.951069485009)
        ref[0,0,1]=(-0.32248722832)/(o+1.)+(-0.834085980672)
        ref[0,1,0]=(0.793268159567)/(o+1.)+(0.036616715229)
        ref[0,1,1]=(-0.850862822585)/(o+1.)+(-0.635904276085)
        ref[1,0,0]=(-0.49189612545)/(o+1.)+(-0.381655664652)
        ref[1,0,1]=(-2.39554424254)/(o+1.)+(0.604353366437)
        ref[1,1,0]=(-1.79696302976)/(o+1.)+(0.238070929246)
        ref[1,1,1]=(-1.00518543165)/(o+1.)+(1.01159134501)
        ref[2,0,0]=(0.340229085537)/(o+1.)+(0.205616292412)
        ref[2,0,1]=(-2.0821483597)/(o+1.)+(-0.52706664625)
        ref[2,1,0]=(-0.878672459802)/(o+1.)+(0.235810228522)
        ref[2,1,1]=(-1.0504252802)/(o+1.)+(-0.650174415929)
        ref[3,0,0]=(-1.31332700315)/(o+1.)+(0.956727385736)
        ref[3,0,1]=(-0.922027363109)/(o+1.)+(-0.00722427314892)
        ref[3,1,0]=(-0.577536939027)/(o+1.)+(0.275469474624)
        ref[3,1,1]=(-0.109991191499)/(o+1.)+(0.545156206825)
        ref[4,0,0]=(0.820965255231)/(o+1.)+(-0.55279315473)
        ref[4,0,1]=(-0.0504249575602)/(o+1.)+(-0.687308481745)
        ref[4,1,0]=(1.25499133056)/(o+1.)+(0.389149085362)
        ref[4,1,1]=(0.238352403952)/(o+1.)+(-0.0641811539369)
        ref[5,0,0]=(0.561786747976)/(o+1.)+(-0.225598284252)
        ref[5,0,1]=(0.661925571434)/(o+1.)+(0.331312295504)
        ref[5,1,0]=(0.659425512333)/(o+1.)+(-0.410709699344)
        ref[5,1,1]=(0.770703994319)/(o+1.)+(0.519727062191)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.164405890316)*x[0]**o+(-0.323303939764)*x[0]+(-0.0516829984856)*x[1]**o+(-0.220765052533)*x[1]
        arg[0,0,0,1]=(0.152172612108)*x[0]**o+(-0.166242164365)*x[0]+(0.807177808571)*x[1]**o+(0.0404817109573)*x[1]
        arg[0,0,1,0]=(0.016912808567)*x[0]**o+(0.623923316774)*x[0]+(-0.542886934905)*x[1]**o+(0.0110784560205)*x[1]
        arg[0,0,1,1]=(-0.663885786909)*x[0]**o+(-0.55563355485)*x[0]+(0.191993284382)*x[1]**o+(0.734973556043)*x[1]
        arg[0,0,2,0]=(-0.110316754183)*x[0]**o+(-0.584022783672)*x[0]+(0.56471376011)*x[1]**o+(0.846317747143)*x[1]
        arg[0,0,2,1]=(-0.767988009278)*x[0]**o+(0.440997912352)*x[0]+(0.45366348093)*x[1]**o+(0.376620990318)*x[1]
        arg[0,1,0,0]=(-0.138006508051)*x[0]**o+(0.8450564766)*x[0]+(-0.273559897633)*x[1]**o+(0.683678615645)*x[1]
        arg[0,1,0,1]=(-0.467521625474)*x[0]**o+(0.32289306962)*x[0]+(0.340974562152)*x[1]**o+(-0.58109391053)*x[1]
        arg[0,1,1,0]=(0.76867991208)*x[0]**o+(0.683527735131)*x[0]+(0.762986872389)*x[1]**o+(-0.189342734286)*x[1]
        arg[0,1,1,1]=(0.125150175351)*x[0]**o+(-0.175190872876)*x[0]+(-0.569766488076)*x[1]**o+(0.787503755206)*x[1]
        arg[0,1,2,0]=(-0.339695860559)*x[0]**o+(0.119452596206)*x[0]+(0.474378864192)*x[1]**o+(0.845982555265)*x[1]
        arg[0,1,2,1]=(0.820551334836)*x[0]**o+(-0.200024721285)*x[0]+(-0.525871855871)*x[1]**o+(-0.714639297791)*x[1]
        arg[0,2,0,0]=(-0.145788788262)*x[0]**o+(-0.695717636982)*x[0]+(0.736564332801)*x[1]**o+(0.634307016587)*x[1]
        arg[0,2,0,1]=(-0.25142006947)*x[0]**o+(0.027028767127)*x[0]+(-0.614777438028)*x[1]**o+(-0.691417785617)*x[1]
        arg[0,2,1,0]=(0.366081140839)*x[0]**o+(0.477490533426)*x[0]+(-0.116611335116)*x[1]**o+(-0.863552165712)*x[1]
        arg[0,2,1,1]=(-0.467659558308)*x[0]**o+(0.0836299545334)*x[0]+(0.903093252833)*x[1]**o+(-0.956858182533)*x[1]
        arg[0,2,2,0]=(0.870921031579)*x[0]**o+(-0.801681014465)*x[0]+(-0.951819608329)*x[1]**o+(-0.626222917084)*x[1]
        arg[0,2,2,1]=(0.414973183041)*x[0]**o+(-0.444580205703)*x[0]+(-0.346458764208)*x[1]**o+(-0.679140100745)*x[1]
        arg[0,3,0,0]=(-0.725326935158)*x[0]**o+(-0.914034759513)*x[0]+(0.702614377806)*x[1]**o+(0.819004276317)*x[1]
        arg[0,3,0,1]=(0.617099932352)*x[0]**o+(-0.155182850011)*x[0]+(-0.066428715129)*x[1]**o+(-0.286480585435)*x[1]
        arg[0,3,1,0]=(0.710942545488)*x[0]**o+(-0.734509277268)*x[0]+(-0.807474546329)*x[1]**o+(0.307631760449)*x[1]
        arg[0,3,1,1]=(-0.709966924732)*x[0]**o+(-0.462297022374)*x[0]+(0.773787774038)*x[1]**o+(0.762137020991)*x[1]
        arg[0,3,2,0]=(-0.425660932209)*x[0]**o+(-0.562130300181)*x[0]+(-0.516719014647)*x[1]**o+(-0.547832887369)*x[1]
        arg[0,3,2,1]=(0.149573176879)*x[0]**o+(-0.208999629436)*x[0]+(-0.191426033553)*x[1]**o+(-0.119046538873)*x[1]
        arg[0,4,0,0]=(-0.99932317151)*x[0]**o+(-0.832461558068)*x[0]+(-0.847073271676)*x[1]**o+(0.882046887729)*x[1]
        arg[0,4,0,1]=(0.14235557233)*x[0]**o+(-0.352572053276)*x[0]+(-0.231989519255)*x[1]**o+(0.146855135398)*x[1]
        arg[0,4,1,0]=(0.550561449012)*x[0]**o+(0.530745681425)*x[0]+(0.198889689472)*x[1]**o+(0.165503108658)*x[1]
        arg[0,4,1,1]=(0.059663401309)*x[0]**o+(0.168377411751)*x[0]+(-0.224682723925)*x[1]**o+(-0.255406907757)*x[1]
        arg[0,4,2,0]=(-0.747776894683)*x[0]**o+(-0.281763358533)*x[0]+(0.660369838923)*x[1]**o+(0.348541234687)*x[1]
        arg[0,4,2,1]=(-0.886248366716)*x[0]**o+(0.665256341312)*x[0]+(-0.118041122298)*x[1]**o+(0.192284814559)*x[1]
        arg[1,0,0,0]=(-0.212291422817)*x[0]**o+(-0.16123490382)*x[0]+(-0.884220059537)*x[1]**o+(-0.40104775897)*x[1]
        arg[1,0,0,1]=(0.389745305389)*x[0]**o+(0.544302066739)*x[0]+(-0.504770065056)*x[1]**o+(0.903371132654)*x[1]
        arg[1,0,1,0]=(0.983764499357)*x[0]**o+(0.0707766278425)*x[0]+(-0.119851167133)*x[1]**o+(0.668490934432)*x[1]
        arg[1,0,1,1]=(-0.61333738947)*x[0]**o+(-0.515837139916)*x[0]+(-0.825015725806)*x[1]**o+(-0.858420009384)*x[1]
        arg[1,0,2,0]=(0.838488008532)*x[0]**o+(0.320020900158)*x[0]+(0.265140913892)*x[1]**o+(0.493460973766)*x[1]
        arg[1,0,2,1]=(-0.957171379585)*x[0]**o+(0.911667647687)*x[0]+(0.0634495769152)*x[1]**o+(-0.244783653159)*x[1]
        arg[1,1,0,0]=(0.854312398231)*x[0]**o+(0.0177604750376)*x[0]+(-0.263135825468)*x[1]**o+(0.196452389095)*x[1]
        arg[1,1,0,1]=(0.496952473267)*x[0]**o+(-0.505581680291)*x[0]+(0.227653536528)*x[1]**o+(-0.138354150837)*x[1]
        arg[1,1,1,0]=(0.569682536816)*x[0]**o+(-0.523174667456)*x[0]+(-0.434073415962)*x[1]**o+(-0.853667257479)*x[1]
        arg[1,1,1,1]=(0.636882167235)*x[0]**o+(-0.122206427025)*x[0]+(0.934361025063)*x[1]**o+(-0.43037160177)*x[1]
        arg[1,1,2,0]=(0.0798318147504)*x[0]**o+(0.825231320185)*x[0]+(0.810701650808)*x[1]**o+(-0.544356371381)*x[1]
        arg[1,1,2,1]=(-0.426710170704)*x[0]**o+(0.200697079677)*x[0]+(0.242027752153)*x[1]**o+(0.746762119874)*x[1]
        arg[1,2,0,0]=(-0.497091856216)*x[0]**o+(0.0953816987137)*x[0]+(-0.388615292775)*x[1]**o+(0.794962452203)*x[1]
        arg[1,2,0,1]=(0.484769626964)*x[0]**o+(0.700484219762)*x[0]+(-0.230377697673)*x[1]**o+(-0.110327221848)*x[1]
        arg[1,2,1,0]=(-0.791806541129)*x[0]**o+(0.0199911517557)*x[0]+(-0.441657020836)*x[1]**o+(-0.972999930466)*x[1]
        arg[1,2,1,1]=(-0.566750983493)*x[0]**o+(0.193689535753)*x[0]+(0.659310468709)*x[1]**o+(0.209045229813)*x[1]
        arg[1,2,2,0]=(0.852553900341)*x[0]**o+(-0.869275488172)*x[0]+(-0.308429715369)*x[1]**o+(0.096335995297)*x[1]
        arg[1,2,2,1]=(0.169351018917)*x[0]**o+(0.269577471686)*x[0]+(-0.226840840616)*x[1]**o+(-0.94229540867)*x[1]
        arg[1,3,0,0]=(0.481711489234)*x[0]**o+(0.293576608201)*x[0]+(0.101998881767)*x[1]**o+(-0.10730442531)*x[1]
        arg[1,3,0,1]=(-0.824144820539)*x[0]**o+(0.174754759038)*x[0]+(-0.251803196848)*x[1]**o+(-0.389656255459)*x[1]
        arg[1,3,1,0]=(0.438310633462)*x[0]**o+(-0.268051020546)*x[0]+(0.670456978527)*x[1]**o+(-0.547393352784)*x[1]
        arg[1,3,1,1]=(-0.95082246973)*x[0]**o+(0.739061218443)*x[0]+(0.209495253219)*x[1]**o+(-0.604801908581)*x[1]
        arg[1,3,2,0]=(0.696900724687)*x[0]**o+(0.501191311274)*x[0]+(0.982952568975)*x[1]**o+(0.1345704134)*x[1]
        arg[1,3,2,1]=(0.823865642457)*x[0]**o+(0.344980106206)*x[0]+(0.0740071873515)*x[1]**o+(-0.950460516779)*x[1]
        arg[1,4,0,0]=(-0.428525532814)*x[0]**o+(0.496012936411)*x[0]+(-0.0756953876324)*x[1]**o+(0.8382068924)*x[1]
        arg[1,4,0,1]=(0.950388644852)*x[0]**o+(-0.225116488186)*x[0]+(0.515802878592)*x[1]**o+(-0.51004230039)*x[1]
        arg[1,4,1,0]=(0.868733254296)*x[0]**o+(-0.506544709529)*x[0]+(0.28935579533)*x[1]**o+(0.100277736612)*x[1]
        arg[1,4,1,1]=(-0.104533384722)*x[0]**o+(0.598995288066)*x[0]+(-0.738946772839)*x[1]**o+(-0.23245555216)*x[1]
        arg[1,4,2,0]=(-0.286232179213)*x[0]**o+(-0.31267596034)*x[0]+(0.588681250431)*x[1]**o+(-0.579857978487)*x[1]
        arg[1,4,2,1]=(0.405666376198)*x[0]**o+(0.380029842735)*x[0]+(-0.646672003588)*x[1]**o+(0.419606979519)*x[1]
        arg[2,0,0,0]=(0.183075434304)*x[0]**o+(0.0929532345426)*x[0]+(-0.742007638693)*x[1]**o+(-0.521914593386)*x[1]
        arg[2,0,0,1]=(0.292878118412)*x[0]**o+(-0.427165220789)*x[0]+(0.540033289295)*x[1]**o+(0.997945616043)*x[1]
        arg[2,0,1,0]=(0.872801856197)*x[0]**o+(0.0427577769706)*x[0]+(0.595354626662)*x[1]**o+(-0.878459740609)*x[1]
        arg[2,0,1,1]=(-0.259988866684)*x[0]**o+(0.154043636253)*x[0]+(0.0580047878991)*x[1]**o+(0.40696744056)*x[1]
        arg[2,0,2,0]=(-0.602648518466)*x[0]**o+(0.30877596654)*x[0]+(-0.268563459489)*x[1]**o+(-0.526379663585)*x[1]
        arg[2,0,2,1]=(-0.375777994931)*x[0]**o+(-0.175440011709)*x[0]+(-0.578650003477)*x[1]**o+(0.822157683321)*x[1]
        arg[2,1,0,0]=(-0.44389226502)*x[0]**o+(0.683580346916)*x[0]+(0.471857559282)*x[1]**o+(-0.869624665398)*x[1]
        arg[2,1,0,1]=(0.768272315036)*x[0]**o+(-0.709688321973)*x[0]+(0.200727881948)*x[1]**o+(0.617326752785)*x[1]
        arg[2,1,1,0]=(-0.371169228982)*x[0]**o+(-0.0248248405104)*x[0]+(-0.899904749204)*x[1]**o+(0.796779883083)*x[1]
        arg[2,1,1,1]=(0.220088670807)*x[0]**o+(0.691718163435)*x[0]+(-0.0834793048429)*x[1]**o+(-0.748110928988)*x[1]
        arg[2,1,2,0]=(0.246045293748)*x[0]**o+(-0.669660154233)*x[0]+(0.601040129345)*x[1]**o+(-0.0893792369983)*x[1]
        arg[2,1,2,1]=(0.318708374425)*x[0]**o+(-0.973048583674)*x[0]+(-0.575891882633)*x[1]**o+(0.229050058038)*x[1]
        arg[2,2,0,0]=(0.358008410645)*x[0]**o+(-0.834388697702)*x[0]+(0.0931465346237)*x[1]**o+(0.503934981947)*x[1]
        arg[2,2,0,1]=(-0.257597986495)*x[0]**o+(-0.58070568528)*x[0]+(0.735541935653)*x[1]**o+(-0.704666216263)*x[1]
        arg[2,2,1,0]=(0.0207421250729)*x[0]**o+(-0.724644198871)*x[0]+(0.637124335031)*x[1]**o+(-0.547300138808)*x[1]
        arg[2,2,1,1]=(0.280179217797)*x[0]**o+(-0.387279903944)*x[0]+(0.516275341085)*x[1]**o+(0.781250973385)*x[1]
        arg[2,2,2,0]=(-0.97771111517)*x[0]**o+(-0.140997969015)*x[0]+(-0.495112130955)*x[1]**o+(0.151523726543)*x[1]
        arg[2,2,2,1]=(-0.48638258963)*x[0]**o+(-0.251363723104)*x[0]+(-0.484911659879)*x[1]**o+(-0.841285872322)*x[1]
        arg[2,3,0,0]=(-0.70159761565)*x[0]**o+(-0.661658396735)*x[0]+(-0.380920250292)*x[1]**o+(0.682598882025)*x[1]
        arg[2,3,0,1]=(0.477056088757)*x[0]**o+(0.697397619448)*x[0]+(0.28617323153)*x[1]**o+(-0.0443690469658)*x[1]
        arg[2,3,1,0]=(-0.823038323615)*x[0]**o+(-0.742146866184)*x[0]+(0.258604766787)*x[1]**o+(-0.740368164141)*x[1]
        arg[2,3,1,1]=(0.958751077013)*x[0]**o+(0.353976218746)*x[0]+(-0.954344840728)*x[1]**o+(0.436915389196)*x[1]
        arg[2,3,2,0]=(0.432342075923)*x[0]**o+(-0.0699761039653)*x[0]+(0.701490046186)*x[1]**o+(0.0888793443125)*x[1]
        arg[2,3,2,1]=(0.612536667525)*x[0]**o+(0.107433723026)*x[0]+(-0.611874456965)*x[1]**o+(-0.253338899867)*x[1]
        arg[2,4,0,0]=(-0.0234819568317)*x[0]**o+(-0.0642964170457)*x[0]+(0.556679188908)*x[1]**o+(0.865922901373)*x[1]
        arg[2,4,0,1]=(0.669235019483)*x[0]**o+(0.744361910609)*x[0]+(0.383397302403)*x[1]**o+(0.669587709042)*x[1]
        arg[2,4,1,0]=(-0.143461051192)*x[0]**o+(-0.0272388772724)*x[0]+(0.723868767782)*x[1]**o+(0.303617835707)*x[1]
        arg[2,4,1,1]=(-0.829765013546)*x[0]**o+(-0.784193619495)*x[0]+(0.384376849236)*x[1]**o+(0.280904709362)*x[1]
        arg[2,4,2,0]=(0.38737050935)*x[0]**o+(0.749614613361)*x[0]+(-0.340848293488)*x[1]**o+(0.96291850214)*x[1]
        arg[2,4,2,1]=(0.619896872679)*x[0]**o+(0.970694589398)*x[0]+(-0.300698305892)*x[1]**o+(0.794408153506)*x[1]
        arg[3,0,0,0]=(-0.808442463698)*x[0]**o+(-0.527555255968)*x[0]+(0.246776596713)*x[1]**o+(-0.719809811547)*x[1]
        arg[3,0,0,1]=(-0.390016002326)*x[0]**o+(-0.864168521951)*x[0]+(-0.154649478359)*x[1]**o+(0.210766031283)*x[1]
        arg[3,0,1,0]=(-0.941916382966)*x[0]**o+(-0.0352551994705)*x[0]+(-0.660026082442)*x[1]**o+(-0.869241000235)*x[1]
        arg[3,0,1,1]=(-0.287992443961)*x[0]**o+(0.292551506636)*x[0]+(-0.619953648778)*x[1]**o+(0.746095811695)*x[1]
        arg[3,0,2,0]=(-0.233659077726)*x[0]**o+(-0.568647415596)*x[0]+(-0.22048065705)*x[1]**o+(0.658105381132)*x[1]
        arg[3,0,2,1]=(0.871784854897)*x[0]**o+(-0.118389479971)*x[0]+(0.143290959871)*x[1]**o+(0.753198625961)*x[1]
        arg[3,1,0,0]=(0.993988451174)*x[0]**o+(0.451849763406)*x[0]+(-0.972805640824)*x[1]**o+(0.796893857099)*x[1]
        arg[3,1,0,1]=(-0.424353439197)*x[0]**o+(-0.575799061446)*x[0]+(-0.233936106648)*x[1]**o+(-0.451864349448)*x[1]
        arg[3,1,1,0]=(0.639821887614)*x[0]**o+(0.771041944334)*x[0]+(0.420297869214)*x[1]**o+(0.671734395857)*x[1]
        arg[3,1,1,1]=(0.398121897075)*x[0]**o+(0.647679737081)*x[0]+(-0.544457308885)*x[1]**o+(0.58955287395)*x[1]
        arg[3,1,2,0]=(-0.118222351008)*x[0]**o+(-0.943807502172)*x[0]+(0.866866987138)*x[1]**o+(-0.0793340319339)*x[1]
        arg[3,1,2,1]=(-0.157286204244)*x[0]**o+(-0.789214982692)*x[0]+(0.646008960971)*x[1]**o+(0.866687256647)*x[1]
        arg[3,2,0,0]=(-0.620625598995)*x[0]**o+(-0.12874444072)*x[0]+(0.4402928966)*x[1]**o+(-0.0344340173898)*x[1]
        arg[3,2,0,1]=(-0.714597102332)*x[0]**o+(0.730206152989)*x[0]+(0.795717819655)*x[1]**o+(0.0543647593833)*x[1]
        arg[3,2,1,0]=(0.698721346048)*x[0]**o+(0.599204472096)*x[0]+(0.323529910402)*x[1]**o+(0.362678869119)*x[1]
        arg[3,2,1,1]=(0.976107824554)*x[0]**o+(0.559889392243)*x[0]+(0.240016231718)*x[1]**o+(0.343322878078)*x[1]
        arg[3,2,2,0]=(0.609695396838)*x[0]**o+(0.786331471916)*x[0]+(-0.965489152976)*x[1]**o+(0.553653504493)*x[1]
        arg[3,2,2,1]=(0.897952533418)*x[0]**o+(0.227242627929)*x[0]+(-0.121752169524)*x[1]**o+(0.844617689997)*x[1]
        arg[3,3,0,0]=(-0.249698211819)*x[0]**o+(-0.400095108874)*x[0]+(-0.44300562221)*x[1]**o+(0.387777895056)*x[1]
        arg[3,3,0,1]=(0.626469203641)*x[0]**o+(0.741893669108)*x[0]+(0.261882509899)*x[1]**o+(-0.726628233604)*x[1]
        arg[3,3,1,0]=(-0.525223500134)*x[0]**o+(0.918314547123)*x[0]+(0.426441176651)*x[1]**o+(0.338058005974)*x[1]
        arg[3,3,1,1]=(-0.31358142709)*x[0]**o+(-0.411070225036)*x[0]+(0.0902340021221)*x[1]**o+(0.404135075929)*x[1]
        arg[3,3,2,0]=(0.528842663849)*x[0]**o+(0.474768688828)*x[0]+(0.697643730083)*x[1]**o+(-0.170335125341)*x[1]
        arg[3,3,2,1]=(0.875221054674)*x[0]**o+(-0.441830492507)*x[0]+(-0.60312069438)*x[1]**o+(0.664895429753)*x[1]
        arg[3,4,0,0]=(0.563422972203)*x[0]**o+(-0.595004250424)*x[0]+(0.888927367905)*x[1]**o+(0.553640735136)*x[1]
        arg[3,4,0,1]=(0.79563891979)*x[0]**o+(-0.461626171695)*x[0]+(0.700194671524)*x[1]**o+(-0.687505010217)*x[1]
        arg[3,4,1,0]=(0.655248139354)*x[0]**o+(0.588175755435)*x[0]+(-0.29236593461)*x[1]**o+(0.575186295035)*x[1]
        arg[3,4,1,1]=(0.405730566113)*x[0]**o+(-0.161117488254)*x[0]+(0.912224633472)*x[1]**o+(0.0292423764587)*x[1]
        arg[3,4,2,0]=(-0.405819525278)*x[0]**o+(-0.362840757195)*x[0]+(0.961950188511)*x[1]**o+(0.205213542149)*x[1]
        arg[3,4,2,1]=(0.703066280254)*x[0]**o+(-0.936055321446)*x[0]+(0.861070629413)*x[1]**o+(0.0856428368256)*x[1]
        ref[0,0,0,0]=(-0.216088888802)/(o+1.)+(-0.272034496148)
        ref[0,0,0,1]=(0.959350420679)/(o+1.)+(-0.062880226704)
        ref[0,0,1,0]=(-0.525974126338)/(o+1.)+(0.317500886397)
        ref[0,0,1,1]=(-0.471892502528)/(o+1.)+(0.0896700005967)
        ref[0,0,2,0]=(0.454397005927)/(o+1.)+(0.131147481736)
        ref[0,0,2,1]=(-0.314324528348)/(o+1.)+(0.408809451335)
        ref[0,1,0,0]=(-0.411566405684)/(o+1.)+(0.764367546122)
        ref[0,1,0,1]=(-0.126547063321)/(o+1.)+(-0.129100420455)
        ref[0,1,1,0]=(1.53166678447)/(o+1.)+(0.247092500423)
        ref[0,1,1,1]=(-0.444616312725)/(o+1.)+(0.306156441165)
        ref[0,1,2,0]=(0.134683003633)/(o+1.)+(0.482717575736)
        ref[0,1,2,1]=(0.294679478965)/(o+1.)+(-0.457332009538)
        ref[0,2,0,0]=(0.590775544538)/(o+1.)+(-0.0307053101973)
        ref[0,2,0,1]=(-0.866197507498)/(o+1.)+(-0.332194509245)
        ref[0,2,1,0]=(0.249469805723)/(o+1.)+(-0.193030816143)
        ref[0,2,1,1]=(0.435433694525)/(o+1.)+(-0.436614114)
        ref[0,2,2,0]=(-0.0808985767498)/(o+1.)+(-0.713951965774)
        ref[0,2,2,1]=(0.0685144188336)/(o+1.)+(-0.561860153224)
        ref[0,3,0,0]=(-0.0227125573528)/(o+1.)+(-0.0475152415978)
        ref[0,3,0,1]=(0.550671217223)/(o+1.)+(-0.220831717723)
        ref[0,3,1,0]=(-0.0965320008409)/(o+1.)+(-0.213438758409)
        ref[0,3,1,1]=(0.0638208493064)/(o+1.)+(0.149919999309)
        ref[0,3,2,0]=(-0.942379946856)/(o+1.)+(-0.554981593775)
        ref[0,3,2,1]=(-0.0418528566735)/(o+1.)+(-0.164023084154)
        ref[0,4,0,0]=(-1.84639644319)/(o+1.)+(0.0247926648306)
        ref[0,4,0,1]=(-0.0896339469252)/(o+1.)+(-0.102858458939)
        ref[0,4,1,0]=(0.749451138484)/(o+1.)+(0.348124395042)
        ref[0,4,1,1]=(-0.165019322616)/(o+1.)+(-0.0435147480027)
        ref[0,4,2,0]=(-0.0874070557597)/(o+1.)+(0.033388938077)
        ref[0,4,2,1]=(-1.00428948901)/(o+1.)+(0.428770577935)
        ref[1,0,0,0]=(-1.09651148235)/(o+1.)+(-0.281141331395)
        ref[1,0,0,1]=(-0.115024759667)/(o+1.)+(0.723836599697)
        ref[1,0,1,0]=(0.863913332224)/(o+1.)+(0.369633781137)
        ref[1,0,1,1]=(-1.43835311528)/(o+1.)+(-0.68712857465)
        ref[1,0,2,0]=(1.10362892242)/(o+1.)+(0.406740936962)
        ref[1,0,2,1]=(-0.893721802669)/(o+1.)+(0.333441997264)
        ref[1,1,0,0]=(0.591176572763)/(o+1.)+(0.107106432066)
        ref[1,1,0,1]=(0.724606009795)/(o+1.)+(-0.321967915564)
        ref[1,1,1,0]=(0.135609120854)/(o+1.)+(-0.688420962467)
        ref[1,1,1,1]=(1.5712431923)/(o+1.)+(-0.276289014397)
        ref[1,1,2,0]=(0.890533465558)/(o+1.)+(0.140437474402)
        ref[1,1,2,1]=(-0.184682418551)/(o+1.)+(0.473729599776)
        ref[1,2,0,0]=(-0.88570714899)/(o+1.)+(0.445172075458)
        ref[1,2,0,1]=(0.254391929291)/(o+1.)+(0.295078498957)
        ref[1,2,1,0]=(-1.23346356196)/(o+1.)+(-0.476504389355)
        ref[1,2,1,1]=(0.0925594852161)/(o+1.)+(0.201367382783)
        ref[1,2,2,0]=(0.544124184972)/(o+1.)+(-0.386469746437)
        ref[1,2,2,1]=(-0.057489821699)/(o+1.)+(-0.336358968492)
        ref[1,3,0,0]=(0.583710371001)/(o+1.)+(0.0931360914458)
        ref[1,3,0,1]=(-1.07594801739)/(o+1.)+(-0.107450748211)
        ref[1,3,1,0]=(1.10876761199)/(o+1.)+(-0.407722186665)
        ref[1,3,1,1]=(-0.741327216511)/(o+1.)+(0.0671296549307)
        ref[1,3,2,0]=(1.67985329366)/(o+1.)+(0.317880862337)
        ref[1,3,2,1]=(0.897872829808)/(o+1.)+(-0.302740205287)
        ref[1,4,0,0]=(-0.504220920447)/(o+1.)+(0.667109914405)
        ref[1,4,0,1]=(1.46619152344)/(o+1.)+(-0.367579394288)
        ref[1,4,1,0]=(1.15808904963)/(o+1.)+(-0.203133486458)
        ref[1,4,1,1]=(-0.843480157561)/(o+1.)+(0.183269867953)
        ref[1,4,2,0]=(0.302449071217)/(o+1.)+(-0.446266969414)
        ref[1,4,2,1]=(-0.24100562739)/(o+1.)+(0.399818411127)
        ref[2,0,0,0]=(-0.558932204389)/(o+1.)+(-0.214480679421)
        ref[2,0,0,1]=(0.832911407707)/(o+1.)+(0.285390197627)
        ref[2,0,1,0]=(1.46815648286)/(o+1.)+(-0.417850981819)
        ref[2,0,1,1]=(-0.201984078785)/(o+1.)+(0.280505538406)
        ref[2,0,2,0]=(-0.871211977955)/(o+1.)+(-0.108801848522)
        ref[2,0,2,1]=(-0.954427998409)/(o+1.)+(0.323358835806)
        ref[2,1,0,0]=(0.0279652942615)/(o+1.)+(-0.0930221592407)
        ref[2,1,0,1]=(0.969000196983)/(o+1.)+(-0.0461807845941)
        ref[2,1,1,0]=(-1.27107397819)/(o+1.)+(0.385977521287)
        ref[2,1,1,1]=(0.136609365964)/(o+1.)+(-0.0281963827764)
        ref[2,1,2,0]=(0.847085423093)/(o+1.)+(-0.379519695616)
        ref[2,1,2,1]=(-0.257183508208)/(o+1.)+(-0.371999262818)
        ref[2,2,0,0]=(0.451154945268)/(o+1.)+(-0.165226857878)
        ref[2,2,0,1]=(0.477943949158)/(o+1.)+(-0.642685950772)
        ref[2,2,1,0]=(0.657866460103)/(o+1.)+(-0.635972168839)
        ref[2,2,1,1]=(0.796454558882)/(o+1.)+(0.19698553472)
        ref[2,2,2,0]=(-1.47282324613)/(o+1.)+(0.00526287876389)
        ref[2,2,2,1]=(-0.971294249509)/(o+1.)+(-0.546324797713)
        ref[2,3,0,0]=(-1.08251786594)/(o+1.)+(0.0104702426452)
        ref[2,3,0,1]=(0.763229320287)/(o+1.)+(0.326514286241)
        ref[2,3,1,0]=(-0.564433556829)/(o+1.)+(-0.741257515163)
        ref[2,3,1,1]=(0.00440623628526)/(o+1.)+(0.395445803971)
        ref[2,3,2,0]=(1.13383212211)/(o+1.)+(0.0094516201736)
        ref[2,3,2,1]=(0.000662210559809)/(o+1.)+(-0.0729525884203)
        ref[2,4,0,0]=(0.533197232076)/(o+1.)+(0.400813242164)
        ref[2,4,0,1]=(1.05263232189)/(o+1.)+(0.706974809826)
        ref[2,4,1,0]=(0.580407716589)/(o+1.)+(0.138189479217)
        ref[2,4,1,1]=(-0.44538816431)/(o+1.)+(-0.251644455066)
        ref[2,4,2,0]=(0.0465222158616)/(o+1.)+(0.85626655775)
        ref[2,4,2,1]=(0.319198566788)/(o+1.)+(0.882551371452)
        ref[3,0,0,0]=(-0.561665866984)/(o+1.)+(-0.623682533758)
        ref[3,0,0,1]=(-0.544665480685)/(o+1.)+(-0.326701245334)
        ref[3,0,1,0]=(-1.60194246541)/(o+1.)+(-0.452248099853)
        ref[3,0,1,1]=(-0.907946092739)/(o+1.)+(0.519323659165)
        ref[3,0,2,0]=(-0.454139734776)/(o+1.)+(0.0447289827682)
        ref[3,0,2,1]=(1.01507581477)/(o+1.)+(0.317404572995)
        ref[3,1,0,0]=(0.0211828103496)/(o+1.)+(0.624371810253)
        ref[3,1,0,1]=(-0.658289545845)/(o+1.)+(-0.513831705447)
        ref[3,1,1,0]=(1.06011975683)/(o+1.)+(0.721388170095)
        ref[3,1,1,1]=(-0.14633541181)/(o+1.)+(0.618616305516)
        ref[3,1,2,0]=(0.74864463613)/(o+1.)+(-0.511570767053)
        ref[3,1,2,1]=(0.488722756727)/(o+1.)+(0.0387361369774)
        ref[3,2,0,0]=(-0.180332702395)/(o+1.)+(-0.0815892290549)
        ref[3,2,0,1]=(0.0811207173227)/(o+1.)+(0.392285456186)
        ref[3,2,1,0]=(1.02225125645)/(o+1.)+(0.480941670608)
        ref[3,2,1,1]=(1.21612405627)/(o+1.)+(0.45160613516)
        ref[3,2,2,0]=(-0.355793756138)/(o+1.)+(0.669992488205)
        ref[3,2,2,1]=(0.776200363894)/(o+1.)+(0.535930158963)
        ref[3,3,0,0]=(-0.692703834029)/(o+1.)+(-0.00615860690903)
        ref[3,3,0,1]=(0.88835171354)/(o+1.)+(0.00763271775166)
        ref[3,3,1,0]=(-0.0987823234827)/(o+1.)+(0.628186276549)
        ref[3,3,1,1]=(-0.223347424968)/(o+1.)+(-0.00346757455392)
        ref[3,3,2,0]=(1.22648639393)/(o+1.)+(0.152216781743)
        ref[3,3,2,1]=(0.272100360294)/(o+1.)+(0.111532468623)
        ref[3,4,0,0]=(1.45235034011)/(o+1.)+(-0.0206817576441)
        ref[3,4,0,1]=(1.49583359131)/(o+1.)+(-0.574565590956)
        ref[3,4,1,0]=(0.362882204744)/(o+1.)+(0.581681025235)
        ref[3,4,1,1]=(1.31795519959)/(o+1.)+(-0.0659375558974)
        ref[3,4,2,0]=(0.556130663233)/(o+1.)+(-0.078813607523)
        ref[3,4,2,1]=(1.56413690967)/(o+1.)+(-0.42520624231)
      else:
        arg[0,0,0,0]=(0.956854727968)*x[0]**o+(-0.43263548815)*x[0]+(-0.1485543705)*x[1]**o+(-0.411060684134)*x[1]+(-0.306829513954)*x[2]**o+(0.89077924839)*x[2]
        arg[0,0,0,1]=(0.125905642347)*x[0]**o+(-0.645491335766)*x[0]+(-0.237571989123)*x[1]**o+(-0.239301942511)*x[1]+(-0.384215893518)*x[2]**o+(-0.371218179274)*x[2]
        arg[0,0,1,0]=(0.0204273419969)*x[0]**o+(0.354631489501)*x[0]+(0.235086936835)*x[1]**o+(0.647551574475)*x[1]+(-0.729066874572)*x[2]**o+(0.177733791046)*x[2]
        arg[0,0,1,1]=(-0.0938492166558)*x[0]**o+(-0.484339074939)*x[0]+(0.619333193491)*x[1]**o+(0.809157671781)*x[1]+(-0.28367956249)*x[2]**o+(-0.1758249267)*x[2]
        arg[0,0,2,0]=(-0.334944259519)*x[0]**o+(-0.126145567301)*x[0]+(-0.646972616146)*x[1]**o+(0.20287553181)*x[1]+(-0.383816067326)*x[2]**o+(-0.0265591031498)*x[2]
        arg[0,0,2,1]=(0.103308564583)*x[0]**o+(0.016108181951)*x[0]+(-0.367450237136)*x[1]**o+(0.657618788015)*x[1]+(-0.8327532511)*x[2]**o+(-0.267286876849)*x[2]
        arg[0,1,0,0]=(0.688995722152)*x[0]**o+(-0.982012541087)*x[0]+(-0.232339851099)*x[1]**o+(-0.097863461266)*x[1]+(0.25916639806)*x[2]**o+(0.0538637018789)*x[2]
        arg[0,1,0,1]=(0.446698337789)*x[0]**o+(-0.309434573629)*x[0]+(0.617004214016)*x[1]**o+(0.747666594435)*x[1]+(-0.869684246717)*x[2]**o+(0.75016186442)*x[2]
        arg[0,1,1,0]=(-0.559974353252)*x[0]**o+(0.966642794905)*x[0]+(-0.228473000256)*x[1]**o+(-0.435093965632)*x[1]+(0.426986383284)*x[2]**o+(0.131300817321)*x[2]
        arg[0,1,1,1]=(-0.696027598385)*x[0]**o+(-0.804778535506)*x[0]+(0.930171049019)*x[1]**o+(-0.900826020873)*x[1]+(0.844689166523)*x[2]**o+(0.0349718139676)*x[2]
        arg[0,1,2,0]=(0.359170897328)*x[0]**o+(0.992944582308)*x[0]+(0.579162388074)*x[1]**o+(0.743410851741)*x[1]+(-0.204325306226)*x[2]**o+(-0.992005484878)*x[2]
        arg[0,1,2,1]=(-0.883554988936)*x[0]**o+(0.486947073334)*x[0]+(-0.78078860326)*x[1]**o+(-0.134196299284)*x[1]+(-0.980852067031)*x[2]**o+(0.453023187355)*x[2]
        arg[0,2,0,0]=(0.804524524371)*x[0]**o+(0.397924403155)*x[0]+(0.712297039229)*x[1]**o+(0.713485622558)*x[1]+(-0.966783610543)*x[2]**o+(0.497869404576)*x[2]
        arg[0,2,0,1]=(-0.564706930001)*x[0]**o+(-0.482026983244)*x[0]+(0.284926154564)*x[1]**o+(0.844423322149)*x[1]+(0.444203847061)*x[2]**o+(0.333244188639)*x[2]
        arg[0,2,1,0]=(-0.630410083438)*x[0]**o+(-0.831218126629)*x[0]+(0.400237927956)*x[1]**o+(-0.70518378137)*x[1]+(-0.803754805917)*x[2]**o+(0.60059455602)*x[2]
        arg[0,2,1,1]=(0.198272547853)*x[0]**o+(0.0840356091496)*x[0]+(0.114385146155)*x[1]**o+(0.731057038288)*x[1]+(-0.976056408523)*x[2]**o+(-0.132565954304)*x[2]
        arg[0,2,2,0]=(0.983300141657)*x[0]**o+(-0.733168195162)*x[0]+(-0.948802450335)*x[1]**o+(-0.756616078178)*x[1]+(0.450302374372)*x[2]**o+(0.802115535717)*x[2]
        arg[0,2,2,1]=(-0.680853503774)*x[0]**o+(-0.910402862928)*x[0]+(0.583775613868)*x[1]**o+(0.741516645611)*x[1]+(-0.550498467533)*x[2]**o+(-0.610648481959)*x[2]
        arg[0,3,0,0]=(0.981909313325)*x[0]**o+(0.148145112359)*x[0]+(-0.989721117518)*x[1]**o+(0.0246558730659)*x[1]+(0.721439979982)*x[2]**o+(0.896052837889)*x[2]
        arg[0,3,0,1]=(0.109831691542)*x[0]**o+(-0.349555366958)*x[0]+(-0.54822642636)*x[1]**o+(0.121919022172)*x[1]+(-0.580228247481)*x[2]**o+(0.564276455713)*x[2]
        arg[0,3,1,0]=(0.106091093353)*x[0]**o+(0.410111742327)*x[0]+(-0.0669026478824)*x[1]**o+(-0.662684907574)*x[1]+(-0.351815520918)*x[2]**o+(-0.324802958038)*x[2]
        arg[0,3,1,1]=(0.857685803258)*x[0]**o+(-0.782627954043)*x[0]+(-0.555036389439)*x[1]**o+(-0.0126716685608)*x[1]+(0.260461379777)*x[2]**o+(0.117590260591)*x[2]
        arg[0,3,2,0]=(-0.987873674604)*x[0]**o+(0.0779662295512)*x[0]+(-0.928229202803)*x[1]**o+(0.508015712707)*x[1]+(0.248638760509)*x[2]**o+(-0.712158091156)*x[2]
        arg[0,3,2,1]=(-0.977973356138)*x[0]**o+(0.270015563339)*x[0]+(0.811157094127)*x[1]**o+(-0.756460384936)*x[1]+(-0.735557557504)*x[2]**o+(0.780333584929)*x[2]
        arg[0,4,0,0]=(-0.762444411646)*x[0]**o+(0.386894014691)*x[0]+(0.763898738855)*x[1]**o+(-0.422896581278)*x[1]+(0.428591445563)*x[2]**o+(-0.617626926396)*x[2]
        arg[0,4,0,1]=(-0.269524167286)*x[0]**o+(0.996895823788)*x[0]+(0.771396605337)*x[1]**o+(-0.993794940262)*x[1]+(-0.534803587481)*x[2]**o+(-0.928627946122)*x[2]
        arg[0,4,1,0]=(0.978497459584)*x[0]**o+(-0.343496507186)*x[0]+(-0.250581087884)*x[1]**o+(-0.855743445852)*x[1]+(-0.127512246124)*x[2]**o+(-0.133936486447)*x[2]
        arg[0,4,1,1]=(-0.844450818471)*x[0]**o+(0.263329730105)*x[0]+(0.411212783735)*x[1]**o+(-0.816379748952)*x[1]+(-0.443806901857)*x[2]**o+(0.1262969097)*x[2]
        arg[0,4,2,0]=(-0.326938381124)*x[0]**o+(-0.707054728551)*x[0]+(-0.725513874889)*x[1]**o+(0.655254977879)*x[1]+(0.772734170665)*x[2]**o+(-0.963856465024)*x[2]
        arg[0,4,2,1]=(0.896415185133)*x[0]**o+(-0.722431478737)*x[0]+(0.942058336843)*x[1]**o+(0.838230971246)*x[1]+(-0.279821311439)*x[2]**o+(-0.330130961206)*x[2]
        arg[1,0,0,0]=(-0.466537465363)*x[0]**o+(0.86493624537)*x[0]+(-0.943049537777)*x[1]**o+(0.658727518981)*x[1]+(-0.130355508544)*x[2]**o+(-0.158245981876)*x[2]
        arg[1,0,0,1]=(0.247046604605)*x[0]**o+(-0.287683750248)*x[0]+(0.266630203142)*x[1]**o+(0.372409648579)*x[1]+(-0.239411725166)*x[2]**o+(0.537339451105)*x[2]
        arg[1,0,1,0]=(-0.232312387253)*x[0]**o+(-0.0493954012747)*x[0]+(-0.461085479525)*x[1]**o+(-0.0744729096417)*x[1]+(-0.795163064273)*x[2]**o+(0.613075576806)*x[2]
        arg[1,0,1,1]=(-0.68570821748)*x[0]**o+(-0.492507082229)*x[0]+(-0.0673636097828)*x[1]**o+(0.569119992806)*x[1]+(-0.227653595613)*x[2]**o+(0.0337433636837)*x[2]
        arg[1,0,2,0]=(-0.124371489405)*x[0]**o+(-0.726635218442)*x[0]+(-0.835418167414)*x[1]**o+(-0.00414597707753)*x[1]+(0.8365917104)*x[2]**o+(-0.507314639615)*x[2]
        arg[1,0,2,1]=(-0.973228057052)*x[0]**o+(0.123228228923)*x[0]+(0.545712842666)*x[1]**o+(-0.451331566047)*x[1]+(0.299501907067)*x[2]**o+(-0.24898389883)*x[2]
        arg[1,1,0,0]=(-0.940690906836)*x[0]**o+(-0.357571197214)*x[0]+(0.98643983012)*x[1]**o+(-0.82113140872)*x[1]+(0.460009285424)*x[2]**o+(-0.0298714172583)*x[2]
        arg[1,1,0,1]=(-0.881543245921)*x[0]**o+(-0.640741814727)*x[0]+(-0.749244677096)*x[1]**o+(0.47161101909)*x[1]+(0.712893375828)*x[2]**o+(0.600151133959)*x[2]
        arg[1,1,1,0]=(-0.217214692405)*x[0]**o+(-0.504211076658)*x[0]+(0.857127736087)*x[1]**o+(0.0439761116172)*x[1]+(0.956380848189)*x[2]**o+(0.698482155013)*x[2]
        arg[1,1,1,1]=(-0.288969728734)*x[0]**o+(-0.63960886193)*x[0]+(-0.741293494258)*x[1]**o+(-0.125034323955)*x[1]+(-0.608855286755)*x[2]**o+(0.462427737944)*x[2]
        arg[1,1,2,0]=(0.126242043984)*x[0]**o+(0.811846484451)*x[0]+(-0.522920862127)*x[1]**o+(0.678142081935)*x[1]+(0.0392586436898)*x[2]**o+(0.403908304728)*x[2]
        arg[1,1,2,1]=(-0.290103861843)*x[0]**o+(0.336130241928)*x[0]+(-0.872641307678)*x[1]**o+(-0.372020796526)*x[1]+(0.172572077906)*x[2]**o+(-0.789431556966)*x[2]
        arg[1,2,0,0]=(-0.158114079913)*x[0]**o+(0.954919898537)*x[0]+(-0.711834968234)*x[1]**o+(-0.390487270434)*x[1]+(-0.086581514295)*x[2]**o+(0.442047084988)*x[2]
        arg[1,2,0,1]=(-0.914034594356)*x[0]**o+(0.521137372279)*x[0]+(0.12454416569)*x[1]**o+(0.627815188423)*x[1]+(0.770244577882)*x[2]**o+(-0.846691748091)*x[2]
        arg[1,2,1,0]=(0.769631824465)*x[0]**o+(0.988840796717)*x[0]+(-0.401836868027)*x[1]**o+(0.401846957832)*x[1]+(-0.877826997285)*x[2]**o+(0.857725824229)*x[2]
        arg[1,2,1,1]=(0.173539365276)*x[0]**o+(0.373206670706)*x[0]+(-0.597288730366)*x[1]**o+(-0.770220794555)*x[1]+(-0.586510454467)*x[2]**o+(-0.569370842441)*x[2]
        arg[1,2,2,0]=(0.673765155818)*x[0]**o+(-0.642914516709)*x[0]+(0.794438851788)*x[1]**o+(-0.17557754186)*x[1]+(0.997565143863)*x[2]**o+(0.722460144483)*x[2]
        arg[1,2,2,1]=(-0.878452482718)*x[0]**o+(-0.846862156166)*x[0]+(0.462351027504)*x[1]**o+(0.267435606154)*x[1]+(-0.452135086983)*x[2]**o+(0.701577407252)*x[2]
        arg[1,3,0,0]=(-0.0557389135964)*x[0]**o+(-0.781831460089)*x[0]+(0.214846092931)*x[1]**o+(0.597436260353)*x[1]+(0.713968428327)*x[2]**o+(-0.456335676572)*x[2]
        arg[1,3,0,1]=(0.288266531585)*x[0]**o+(-0.259529556521)*x[0]+(0.0286107975184)*x[1]**o+(0.465095455484)*x[1]+(-0.747648909025)*x[2]**o+(0.682866470382)*x[2]
        arg[1,3,1,0]=(-0.720339104668)*x[0]**o+(0.961100526257)*x[0]+(0.0814885381756)*x[1]**o+(-0.448775223649)*x[1]+(0.250972143198)*x[2]**o+(-0.240231439947)*x[2]
        arg[1,3,1,1]=(-0.459305786748)*x[0]**o+(0.272222143865)*x[0]+(-0.499164099401)*x[1]**o+(-0.225162547271)*x[1]+(-0.92261124167)*x[2]**o+(-0.436659790026)*x[2]
        arg[1,3,2,0]=(0.890752966329)*x[0]**o+(-0.775207922011)*x[0]+(0.715960284721)*x[1]**o+(0.843422841789)*x[1]+(0.513788850466)*x[2]**o+(-0.325733710281)*x[2]
        arg[1,3,2,1]=(0.958726172433)*x[0]**o+(-0.743672393249)*x[0]+(-0.523783133945)*x[1]**o+(0.133861976051)*x[1]+(-0.109230940116)*x[2]**o+(0.819716327751)*x[2]
        arg[1,4,0,0]=(0.54439729551)*x[0]**o+(0.0763503792087)*x[0]+(0.0634218657498)*x[1]**o+(-0.405471845044)*x[1]+(0.316327091756)*x[2]**o+(-0.539493863939)*x[2]
        arg[1,4,0,1]=(0.527330373182)*x[0]**o+(0.77531447123)*x[0]+(-0.481869315442)*x[1]**o+(0.556927692362)*x[1]+(-0.267215098309)*x[2]**o+(0.958919406465)*x[2]
        arg[1,4,1,0]=(0.540950372012)*x[0]**o+(-0.95002626611)*x[0]+(-0.523544509893)*x[1]**o+(-0.583056635057)*x[1]+(-0.00288409928193)*x[2]**o+(0.671385098427)*x[2]
        arg[1,4,1,1]=(-0.39802076092)*x[0]**o+(-0.583264022146)*x[0]+(-0.656261495961)*x[1]**o+(-0.527677078418)*x[1]+(-0.2632498932)*x[2]**o+(-0.537632951507)*x[2]
        arg[1,4,2,0]=(-0.77176416566)*x[0]**o+(0.812696440735)*x[0]+(0.0925789242005)*x[1]**o+(0.299698967601)*x[1]+(-0.0185377066587)*x[2]**o+(-0.883247318682)*x[2]
        arg[1,4,2,1]=(0.360949407452)*x[0]**o+(-0.463659754812)*x[0]+(0.799066387132)*x[1]**o+(0.0318151346916)*x[1]+(-0.380675976437)*x[2]**o+(-0.299341342112)*x[2]
        arg[2,0,0,0]=(-0.985332287141)*x[0]**o+(0.33398206098)*x[0]+(0.39204502373)*x[1]**o+(-0.586826360517)*x[1]+(0.868240161369)*x[2]**o+(-0.257897525907)*x[2]
        arg[2,0,0,1]=(-0.801369266967)*x[0]**o+(0.333496542809)*x[0]+(0.459444994717)*x[1]**o+(-0.950297568688)*x[1]+(-0.760722281765)*x[2]**o+(0.0933850221377)*x[2]
        arg[2,0,1,0]=(0.53372811812)*x[0]**o+(0.999551098511)*x[0]+(-0.761650890163)*x[1]**o+(-0.80456940786)*x[1]+(-0.452089806623)*x[2]**o+(-0.40841169158)*x[2]
        arg[2,0,1,1]=(-0.782294374468)*x[0]**o+(0.614736817827)*x[0]+(0.262995910218)*x[1]**o+(0.067349886804)*x[1]+(-0.167494656909)*x[2]**o+(-0.730462378769)*x[2]
        arg[2,0,2,0]=(0.261001857862)*x[0]**o+(0.959318765248)*x[0]+(-0.177224016298)*x[1]**o+(-0.979637664014)*x[1]+(0.858937930292)*x[2]**o+(-0.627116377798)*x[2]
        arg[2,0,2,1]=(-0.238083578675)*x[0]**o+(0.586367991705)*x[0]+(0.747378561289)*x[1]**o+(-0.540178926415)*x[1]+(-0.519781240185)*x[2]**o+(-0.42715014981)*x[2]
        arg[2,1,0,0]=(-0.21022574437)*x[0]**o+(-0.292359099348)*x[0]+(0.449614855785)*x[1]**o+(-0.416904619153)*x[1]+(0.507537288404)*x[2]**o+(0.943584716996)*x[2]
        arg[2,1,0,1]=(-0.410591525695)*x[0]**o+(-0.258796394127)*x[0]+(-0.474714731399)*x[1]**o+(-0.518288451462)*x[1]+(-0.15640903664)*x[2]**o+(0.97242574307)*x[2]
        arg[2,1,1,0]=(-0.930865834892)*x[0]**o+(0.961887630616)*x[0]+(0.688706862295)*x[1]**o+(-0.375163634404)*x[1]+(0.378954001649)*x[2]**o+(-0.747140890466)*x[2]
        arg[2,1,1,1]=(0.0454701384411)*x[0]**o+(-0.37651801332)*x[0]+(0.0808750051583)*x[1]**o+(-0.530355850075)*x[1]+(0.769675417234)*x[2]**o+(-0.767046792016)*x[2]
        arg[2,1,2,0]=(0.179573418696)*x[0]**o+(-0.0547918709025)*x[0]+(0.739792606043)*x[1]**o+(0.849619562235)*x[1]+(-0.92840796838)*x[2]**o+(-0.302941296337)*x[2]
        arg[2,1,2,1]=(0.160657807137)*x[0]**o+(0.551943403654)*x[0]+(0.649164334204)*x[1]**o+(0.298882266215)*x[1]+(0.33872360276)*x[2]**o+(-0.558910868754)*x[2]
        arg[2,2,0,0]=(-0.704330511528)*x[0]**o+(-0.326943167116)*x[0]+(0.626863819473)*x[1]**o+(-0.107147718189)*x[1]+(0.00668670393016)*x[2]**o+(0.572223662188)*x[2]
        arg[2,2,0,1]=(0.879695137799)*x[0]**o+(-0.313488700497)*x[0]+(-0.113134766565)*x[1]**o+(-0.131007759381)*x[1]+(0.55460969593)*x[2]**o+(0.898706374575)*x[2]
        arg[2,2,1,0]=(0.329932124981)*x[0]**o+(-0.624965470976)*x[0]+(0.82906774364)*x[1]**o+(-0.681403786977)*x[1]+(-0.447119491747)*x[2]**o+(0.143763601614)*x[2]
        arg[2,2,1,1]=(0.0367746207528)*x[0]**o+(0.47553259605)*x[0]+(-0.517723045973)*x[1]**o+(0.263932783195)*x[1]+(-0.131032589284)*x[2]**o+(-0.132729291841)*x[2]
        arg[2,2,2,0]=(-0.108349266564)*x[0]**o+(-0.684748849483)*x[0]+(0.203271582231)*x[1]**o+(-0.860842085896)*x[1]+(0.469684577785)*x[2]**o+(0.195736308182)*x[2]
        arg[2,2,2,1]=(0.915666799349)*x[0]**o+(0.383338953756)*x[0]+(-0.554011249579)*x[1]**o+(0.232285573367)*x[1]+(0.39288151674)*x[2]**o+(0.692592233016)*x[2]
        arg[2,3,0,0]=(0.81179207663)*x[0]**o+(0.786809236572)*x[0]+(-0.814758445611)*x[1]**o+(-0.78461653396)*x[1]+(0.171435802778)*x[2]**o+(0.660186909193)*x[2]
        arg[2,3,0,1]=(0.752694523669)*x[0]**o+(0.301473785296)*x[0]+(0.610379955695)*x[1]**o+(0.573729399495)*x[1]+(0.225983751859)*x[2]**o+(-0.349530143446)*x[2]
        arg[2,3,1,0]=(-0.647825490424)*x[0]**o+(-0.232573858642)*x[0]+(0.773519085771)*x[1]**o+(-0.224497979868)*x[1]+(-0.681726885952)*x[2]**o+(-0.217968759388)*x[2]
        arg[2,3,1,1]=(0.391888417642)*x[0]**o+(-0.240556139087)*x[0]+(-0.852397797723)*x[1]**o+(0.318342758913)*x[1]+(-0.773804983598)*x[2]**o+(-0.957590491908)*x[2]
        arg[2,3,2,0]=(0.7126582007)*x[0]**o+(0.806022164546)*x[0]+(0.232348951782)*x[1]**o+(-0.985774366344)*x[1]+(-0.365161907541)*x[2]**o+(0.0218977623299)*x[2]
        arg[2,3,2,1]=(-0.775068843872)*x[0]**o+(-0.595291330879)*x[0]+(0.681122481841)*x[1]**o+(0.152514252732)*x[1]+(0.510976191808)*x[2]**o+(-0.69539067488)*x[2]
        arg[2,4,0,0]=(-0.362761749405)*x[0]**o+(0.494347958562)*x[0]+(0.215244694522)*x[1]**o+(-0.527767969217)*x[1]+(-0.115976071059)*x[2]**o+(-0.998233415841)*x[2]
        arg[2,4,0,1]=(0.620291015098)*x[0]**o+(-0.222644358028)*x[0]+(0.0377714785847)*x[1]**o+(0.663978623913)*x[1]+(0.996481523389)*x[2]**o+(-0.689519418095)*x[2]
        arg[2,4,1,0]=(-0.0374746407394)*x[0]**o+(0.3186355021)*x[0]+(-0.704305640533)*x[1]**o+(-0.632264150766)*x[1]+(-0.166691199609)*x[2]**o+(-0.794128441334)*x[2]
        arg[2,4,1,1]=(0.117806334919)*x[0]**o+(-0.717334572919)*x[0]+(0.665341054571)*x[1]**o+(0.354208273372)*x[1]+(0.349883631586)*x[2]**o+(-0.260626035314)*x[2]
        arg[2,4,2,0]=(0.160337827685)*x[0]**o+(-0.92734430881)*x[0]+(0.872112181363)*x[1]**o+(-0.796372970466)*x[1]+(-0.776882702685)*x[2]**o+(0.246816968474)*x[2]
        arg[2,4,2,1]=(-0.317926812571)*x[0]**o+(-0.795779593996)*x[0]+(0.785287724793)*x[1]**o+(0.839033515634)*x[1]+(0.433187943714)*x[2]**o+(0.308175885606)*x[2]
        arg[3,0,0,0]=(-0.555193292291)*x[0]**o+(-0.0280452948283)*x[0]+(0.988455127227)*x[1]**o+(-0.637707060271)*x[1]+(-0.95062571916)*x[2]**o+(-0.563121211553)*x[2]
        arg[3,0,0,1]=(-0.169558534946)*x[0]**o+(0.886096637835)*x[0]+(-0.666307491933)*x[1]**o+(0.650321801985)*x[1]+(-0.268934932123)*x[2]**o+(-0.406617199906)*x[2]
        arg[3,0,1,0]=(0.497351781449)*x[0]**o+(0.768617651685)*x[0]+(0.200962956226)*x[1]**o+(0.694236078954)*x[1]+(0.0323702864321)*x[2]**o+(-0.666046265832)*x[2]
        arg[3,0,1,1]=(0.425151807387)*x[0]**o+(-0.886432438604)*x[0]+(-0.963257203569)*x[1]**o+(-0.698524427437)*x[1]+(0.0165810713355)*x[2]**o+(-0.142193988951)*x[2]
        arg[3,0,2,0]=(0.965746112788)*x[0]**o+(-0.333068915886)*x[0]+(0.331031372002)*x[1]**o+(-0.384004796757)*x[1]+(0.0144121092048)*x[2]**o+(-0.352607264387)*x[2]
        arg[3,0,2,1]=(-0.391557775789)*x[0]**o+(0.127020330586)*x[0]+(0.758832450081)*x[1]**o+(0.49669414469)*x[1]+(0.358403498634)*x[2]**o+(0.184963946351)*x[2]
        arg[3,1,0,0]=(0.978053897464)*x[0]**o+(-0.211787313494)*x[0]+(0.927706482976)*x[1]**o+(-0.451292956408)*x[1]+(0.592824645174)*x[2]**o+(0.896240907546)*x[2]
        arg[3,1,0,1]=(0.0780568875295)*x[0]**o+(0.130098698523)*x[0]+(-0.281871571918)*x[1]**o+(-0.596362849045)*x[1]+(-0.436455789199)*x[2]**o+(-0.657145053021)*x[2]
        arg[3,1,1,0]=(-0.278111817329)*x[0]**o+(-0.627384158565)*x[0]+(-0.995802044048)*x[1]**o+(0.42485189159)*x[1]+(0.483473275475)*x[2]**o+(-0.177928974458)*x[2]
        arg[3,1,1,1]=(0.127721047557)*x[0]**o+(0.93147711978)*x[0]+(-0.350992052283)*x[1]**o+(0.0993773637672)*x[1]+(0.29111744653)*x[2]**o+(-0.821710278883)*x[2]
        arg[3,1,2,0]=(-0.697941918276)*x[0]**o+(-0.585872232852)*x[0]+(0.697967779925)*x[1]**o+(-0.811186381728)*x[1]+(0.708726108998)*x[2]**o+(0.289602027081)*x[2]
        arg[3,1,2,1]=(-0.461424409147)*x[0]**o+(0.335089386385)*x[0]+(-0.964838280406)*x[1]**o+(0.828627561583)*x[1]+(-0.861416281013)*x[2]**o+(0.898354675269)*x[2]
        arg[3,2,0,0]=(0.752543705358)*x[0]**o+(0.164487088496)*x[0]+(0.756164002406)*x[1]**o+(-0.740019514032)*x[1]+(0.107245489678)*x[2]**o+(0.677707700079)*x[2]
        arg[3,2,0,1]=(0.763792557134)*x[0]**o+(0.725507550672)*x[0]+(0.171323949629)*x[1]**o+(-0.287172075258)*x[1]+(-0.94865226715)*x[2]**o+(-0.163093465336)*x[2]
        arg[3,2,1,0]=(-0.327892721048)*x[0]**o+(0.334145897353)*x[0]+(-0.936827293436)*x[1]**o+(0.0822687872593)*x[1]+(0.259794829604)*x[2]**o+(0.376908280759)*x[2]
        arg[3,2,1,1]=(0.595012545955)*x[0]**o+(0.617490014693)*x[0]+(-0.658036388462)*x[1]**o+(-0.921055393527)*x[1]+(-0.257878085344)*x[2]**o+(-0.623113089161)*x[2]
        arg[3,2,2,0]=(0.961175705937)*x[0]**o+(0.656031660842)*x[0]+(0.964101580317)*x[1]**o+(0.824033157042)*x[1]+(0.510244886658)*x[2]**o+(-0.236501162729)*x[2]
        arg[3,2,2,1]=(-0.128405588923)*x[0]**o+(-0.0327141502894)*x[0]+(0.909941485131)*x[1]**o+(0.79947258121)*x[1]+(0.635883050534)*x[2]**o+(0.841997575796)*x[2]
        arg[3,3,0,0]=(-0.906234907643)*x[0]**o+(-0.316076976009)*x[0]+(0.863208664533)*x[1]**o+(-0.968640170523)*x[1]+(0.791637381686)*x[2]**o+(-0.543353462098)*x[2]
        arg[3,3,0,1]=(-0.586709036832)*x[0]**o+(-0.299493216592)*x[0]+(-0.173966725899)*x[1]**o+(0.0971838711677)*x[1]+(0.385662922244)*x[2]**o+(0.288932625368)*x[2]
        arg[3,3,1,0]=(-0.33865322795)*x[0]**o+(0.442176445777)*x[0]+(0.706183231282)*x[1]**o+(0.62528018966)*x[1]+(-0.377746006979)*x[2]**o+(-0.261455124139)*x[2]
        arg[3,3,1,1]=(0.834855553776)*x[0]**o+(0.755005824764)*x[0]+(0.446589332183)*x[1]**o+(0.41883521757)*x[1]+(0.820177207209)*x[2]**o+(0.603811552083)*x[2]
        arg[3,3,2,0]=(-0.624282266193)*x[0]**o+(0.28109086861)*x[0]+(0.44286011655)*x[1]**o+(-0.117876187343)*x[1]+(-0.267660131821)*x[2]**o+(0.880359406159)*x[2]
        arg[3,3,2,1]=(-0.0722687033317)*x[0]**o+(-0.7013147214)*x[0]+(0.00614303249005)*x[1]**o+(-0.658138946692)*x[1]+(0.0557277088337)*x[2]**o+(0.614362766103)*x[2]
        arg[3,4,0,0]=(0.935244440586)*x[0]**o+(-0.0699305155896)*x[0]+(-0.849303362918)*x[1]**o+(0.728062537708)*x[1]+(0.404621498028)*x[2]**o+(0.611486631437)*x[2]
        arg[3,4,0,1]=(0.168117782017)*x[0]**o+(-0.0454396423181)*x[0]+(0.46021270423)*x[1]**o+(0.597343569492)*x[1]+(0.561215321486)*x[2]**o+(0.715771227777)*x[2]
        arg[3,4,1,0]=(-0.502675682461)*x[0]**o+(0.543910335185)*x[0]+(0.721732643536)*x[1]**o+(-0.862126899566)*x[1]+(0.794704104653)*x[2]**o+(-0.284248276797)*x[2]
        arg[3,4,1,1]=(-0.580751817984)*x[0]**o+(0.482838505281)*x[0]+(-0.426066109367)*x[1]**o+(-0.456724382007)*x[1]+(-0.770655426938)*x[2]**o+(-0.226232224569)*x[2]
        arg[3,4,2,0]=(0.199832861603)*x[0]**o+(-0.586239366436)*x[0]+(0.730019631541)*x[1]**o+(0.185265476259)*x[1]+(0.514243351617)*x[2]**o+(-0.757742857839)*x[2]
        arg[3,4,2,1]=(-0.941843068195)*x[0]**o+(0.211732407161)*x[0]+(-0.99705602485)*x[1]**o+(-0.198037742645)*x[1]+(0.154058794178)*x[2]**o+(0.0675962671791)*x[2]
        ref[0,0,0,0]=(0.501470843513)/(o+1.)+(0.0235415380528)
        ref[0,0,0,1]=(-0.495882240293)/(o+1.)+(-0.628005728776)
        ref[0,0,1,0]=(-0.47355259574)/(o+1.)+(0.589958427511)
        ref[0,0,1,1]=(0.241804414345)/(o+1.)+(0.0744968350706)
        ref[0,0,2,0]=(-1.36573294299)/(o+1.)+(0.0250854306796)
        ref[0,0,2,1]=(-1.09689492365)/(o+1.)+(0.203220046558)
        ref[0,1,0,0]=(0.715822269113)/(o+1.)+(-0.513006150237)
        ref[0,1,0,1]=(0.194018305088)/(o+1.)+(0.594196942613)
        ref[0,1,1,0]=(-0.361460970224)/(o+1.)+(0.331424823297)
        ref[0,1,1,1]=(1.07883261716)/(o+1.)+(-0.835316371206)
        ref[0,1,2,0]=(0.734007979175)/(o+1.)+(0.372174974585)
        ref[0,1,2,1]=(-2.64519565923)/(o+1.)+(0.402886980703)
        ref[0,2,0,0]=(0.550037953057)/(o+1.)+(0.804639715144)
        ref[0,2,0,1]=(0.164423071623)/(o+1.)+(0.347820263772)
        ref[0,2,1,0]=(-1.0339269614)/(o+1.)+(-0.46790367599)
        ref[0,2,1,1]=(-0.663398714515)/(o+1.)+(0.341263346566)
        ref[0,2,2,0]=(0.484800065694)/(o+1.)+(-0.343834368812)
        ref[0,2,2,1]=(-0.647576357439)/(o+1.)+(-0.389767349638)
        ref[0,3,0,0]=(0.713628175789)/(o+1.)+(0.534426911657)
        ref[0,3,0,1]=(-1.0186229823)/(o+1.)+(0.168320055464)
        ref[0,3,1,0]=(-0.312627075447)/(o+1.)+(-0.288688061643)
        ref[0,3,1,1]=(0.563110793595)/(o+1.)+(-0.338854681006)
        ref[0,3,2,0]=(-1.6674641169)/(o+1.)+(-0.0630880744491)
        ref[0,3,2,1]=(-0.902373819515)/(o+1.)+(0.146944381666)
        ref[0,4,0,0]=(0.430045772773)/(o+1.)+(-0.326814746492)
        ref[0,4,0,1]=(-0.0329311494305)/(o+1.)+(-0.462763531298)
        ref[0,4,1,0]=(0.600404125575)/(o+1.)+(-0.666588219743)
        ref[0,4,1,1]=(-0.877044936593)/(o+1.)+(-0.213376554573)
        ref[0,4,2,0]=(-0.279718085349)/(o+1.)+(-0.507828107848)
        ref[0,4,2,1]=(1.55865221054)/(o+1.)+(-0.107165734349)
        ref[1,0,0,0]=(-1.53994251168)/(o+1.)+(0.682708891237)
        ref[1,0,0,1]=(0.27426508258)/(o+1.)+(0.311032674718)
        ref[1,0,1,0]=(-1.48856093105)/(o+1.)+(0.244603632945)
        ref[1,0,1,1]=(-0.980725422876)/(o+1.)+(0.0551781371305)
        ref[1,0,2,0]=(-0.123197946419)/(o+1.)+(-0.619047917567)
        ref[1,0,2,1]=(-0.128013307318)/(o+1.)+(-0.288543617977)
        ref[1,1,0,0]=(0.505758208708)/(o+1.)+(-0.604287011596)
        ref[1,1,0,1]=(-0.917894547188)/(o+1.)+(0.215510169161)
        ref[1,1,1,0]=(1.59629389187)/(o+1.)+(0.119123594986)
        ref[1,1,1,1]=(-1.63911850975)/(o+1.)+(-0.15110772397)
        ref[1,1,2,0]=(-0.357420174452)/(o+1.)+(0.946948435557)
        ref[1,1,2,1]=(-0.990173091615)/(o+1.)+(-0.412661055782)
        ref[1,2,0,0]=(-0.956530562442)/(o+1.)+(0.503239856545)
        ref[1,2,0,1]=(-0.0192458507848)/(o+1.)+(0.151130406306)
        ref[1,2,1,0]=(-0.510032040848)/(o+1.)+(1.12420678939)
        ref[1,2,1,1]=(-1.01025981956)/(o+1.)+(-0.483192483145)
        ref[1,2,2,0]=(2.46576915147)/(o+1.)+(-0.0480159570431)
        ref[1,2,2,1]=(-0.868236542197)/(o+1.)+(0.0610754286198)
        ref[1,3,0,0]=(0.873075607662)/(o+1.)+(-0.320365438154)
        ref[1,3,0,1]=(-0.430771579922)/(o+1.)+(0.444216184672)
        ref[1,3,1,0]=(-0.387878423294)/(o+1.)+(0.13604693133)
        ref[1,3,1,1]=(-1.88108112782)/(o+1.)+(-0.194800096716)
        ref[1,3,2,0]=(2.12050210152)/(o+1.)+(-0.128759395251)
        ref[1,3,2,1]=(0.325712098372)/(o+1.)+(0.104952955276)
        ref[1,4,0,0]=(0.924146253016)/(o+1.)+(-0.434307664887)
        ref[1,4,0,1]=(-0.221754040568)/(o+1.)+(1.14558078503)
        ref[1,4,1,0]=(0.0145217628367)/(o+1.)+(-0.43084890137)
        ref[1,4,1,1]=(-1.31753215008)/(o+1.)+(-0.824287026036)
        ref[1,4,2,0]=(-0.697722948118)/(o+1.)+(0.114574044827)
        ref[1,4,2,1]=(0.779339818147)/(o+1.)+(-0.365592981116)
        ref[2,0,0,0]=(0.274952897958)/(o+1.)+(-0.255370912722)
        ref[2,0,0,1]=(-1.10264655401)/(o+1.)+(-0.261708001871)
        ref[2,0,1,0]=(-0.680012578665)/(o+1.)+(-0.106715000464)
        ref[2,0,1,1]=(-0.686793121159)/(o+1.)+(-0.0241878370688)
        ref[2,0,2,0]=(0.942715771856)/(o+1.)+(-0.323717638282)
        ref[2,0,2,1]=(-0.0104862575713)/(o+1.)+(-0.19048054226)
        ref[2,1,0,0]=(0.746926399818)/(o+1.)+(0.117160499248)
        ref[2,1,0,1]=(-1.04171529373)/(o+1.)+(0.0976704487401)
        ref[2,1,1,0]=(0.136795029052)/(o+1.)+(-0.0802084471266)
        ref[2,1,1,1]=(0.896020560833)/(o+1.)+(-0.836960327706)
        ref[2,1,2,0]=(-0.00904194364113)/(o+1.)+(0.245943197498)
        ref[2,1,2,1]=(1.1485457441)/(o+1.)+(0.145957400558)
        ref[2,2,0,0]=(-0.0707799881243)/(o+1.)+(0.0690663884416)
        ref[2,2,0,1]=(1.32117006716)/(o+1.)+(0.227104957348)
        ref[2,2,1,0]=(0.711880376874)/(o+1.)+(-0.58130282817)
        ref[2,2,1,1]=(-0.611981014504)/(o+1.)+(0.303368043702)
        ref[2,2,2,0]=(0.564606893452)/(o+1.)+(-0.674927313599)
        ref[2,2,2,1]=(0.754537066511)/(o+1.)+(0.654108380069)
        ref[2,3,0,0]=(0.168469433796)/(o+1.)+(0.331189805903)
        ref[2,3,0,1]=(1.58905823122)/(o+1.)+(0.262836520673)
        ref[2,3,1,0]=(-0.556033290604)/(o+1.)+(-0.337520298949)
        ref[2,3,1,1]=(-1.23431436368)/(o+1.)+(-0.439901936041)
        ref[2,3,2,0]=(0.579845244941)/(o+1.)+(-0.0789272197338)
        ref[2,3,2,1]=(0.417029829778)/(o+1.)+(-0.569083876513)
        ref[2,4,0,0]=(-0.263493125943)/(o+1.)+(-0.515826713248)
        ref[2,4,0,1]=(1.65454401707)/(o+1.)+(-0.124092576105)
        ref[2,4,1,0]=(-0.908471480882)/(o+1.)+(-0.553878545)
        ref[2,4,1,1]=(1.13303102108)/(o+1.)+(-0.31187616743)
        ref[2,4,2,0]=(0.255567306363)/(o+1.)+(-0.738450155401)
        ref[2,4,2,1]=(0.900548855936)/(o+1.)+(0.175714903622)
        ref[3,0,0,0]=(-0.517363884224)/(o+1.)+(-0.614436783326)
        ref[3,0,0,1]=(-1.104800959)/(o+1.)+(0.564900619957)
        ref[3,0,1,0]=(0.730685024108)/(o+1.)+(0.398403732404)
        ref[3,0,1,1]=(-0.521524324846)/(o+1.)+(-0.863575427496)
        ref[3,0,2,0]=(1.31118959399)/(o+1.)+(-0.534840488515)
        ref[3,0,2,1]=(0.725678172925)/(o+1.)+(0.404339210814)
        ref[3,1,0,0]=(2.49858502561)/(o+1.)+(0.116580318822)
        ref[3,1,0,1]=(-0.640270473588)/(o+1.)+(-0.561704601771)
        ref[3,1,1,0]=(-0.790440585902)/(o+1.)+(-0.190230620716)
        ref[3,1,1,1]=(0.0678464418036)/(o+1.)+(0.104572102332)
        ref[3,1,2,0]=(0.708751970646)/(o+1.)+(-0.553728293749)
        ref[3,1,2,1]=(-2.28767897057)/(o+1.)+(1.03103581162)
        ref[3,2,0,0]=(1.61595319744)/(o+1.)+(0.0510876372718)
        ref[3,2,0,1]=(-0.0135357603866)/(o+1.)+(0.137621005039)
        ref[3,2,1,0]=(-1.00492518488)/(o+1.)+(0.396661482686)
        ref[3,2,1,1]=(-0.320901927851)/(o+1.)+(-0.463339233997)
        ref[3,2,2,0]=(2.43552217291)/(o+1.)+(0.621781827577)
        ref[3,2,2,1]=(1.41741894674)/(o+1.)+(0.804378003359)
        ref[3,3,0,0]=(0.748611138576)/(o+1.)+(-0.914035304315)
        ref[3,3,0,1]=(-0.375012840487)/(o+1.)+(0.0433116399723)
        ref[3,3,1,0]=(-0.0102160036467)/(o+1.)+(0.403000755649)
        ref[3,3,1,1]=(2.10162209317)/(o+1.)+(0.888826297208)
        ref[3,3,2,0]=(-0.449082281464)/(o+1.)+(0.521787043713)
        ref[3,3,2,1]=(-0.0103979620079)/(o+1.)+(-0.372545450994)
        ref[3,4,0,0]=(0.490562575696)/(o+1.)+(0.634809326777)
        ref[3,4,0,1]=(1.18954580773)/(o+1.)+(0.633837577475)
        ref[3,4,1,0]=(1.01376106573)/(o+1.)+(-0.301232420589)
        ref[3,4,1,1]=(-1.77747335429)/(o+1.)+(-0.100059050648)
        ref[3,4,2,0]=(1.44409584476)/(o+1.)+(-0.579358374008)
        ref[3,4,2,1]=(-1.78484029887)/(o+1.)+(0.0406454658476)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.788336609161)*x[0]**o+(-0.46539601764)*x[0]+(0.0970136798824)*x[1]**o+(-0.120665298671)*x[1]
        ref=(-0.691322929278)/(o+1.)+(-0.293030658156)
      else:
        arg=(0.922255261003)*x[0]**o+(0.0862232364707)*x[0]+(-0.94152635039)*x[1]**o+(0.771078486789)*x[1]+(-0.487642191181)*x[2]**o+(0.971754960362)*x[2]
        ref=(-0.506913280568)/(o+1.)+(0.914528341811)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.193071015033)*x[0]**o+(-0.281957941612)*x[0]+(0.269334552863)*x[1]**o+(0.0332262192804)*x[1]
        arg[1]=(-0.982634724425)*x[0]**o+(0.872638820737)*x[0]+(0.828028964603)*x[1]**o+(0.70464418136)*x[1]
        ref[0]=(0.0762635378296)/(o+1.)+(-0.124365861166)
        ref[1]=(-0.154605759822)/(o+1.)+(0.788641501048)
      else:
        arg[0]=(0.0746983082972)*x[0]**o+(0.281818436343)*x[0]+(-0.835321029206)*x[1]**o+(-0.349524270524)*x[1]+(-0.471905088601)*x[2]**o+(0.156327167402)*x[2]
        arg[1]=(-0.971935771255)*x[0]**o+(0.943151542057)*x[0]+(-0.60084258169)*x[1]**o+(0.458392855527)*x[1]+(-0.0173384187065)*x[2]**o+(0.538410338758)*x[2]
        ref[0]=(-1.23252780951)/(o+1.)+(0.0443106666103)
        ref[1]=(-1.59011677165)/(o+1.)+(0.969977368171)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.568783264759)*x[0]**o+(0.690377280868)*x[0]+(-0.689750922806)*x[1]**o+(-0.262828324126)*x[1]
        arg[0,1]=(-0.981657284673)*x[0]**o+(0.37118874852)*x[0]+(0.733992104146)*x[1]**o+(0.350268735699)*x[1]
        arg[0,2]=(0.713336100765)*x[0]**o+(0.737674447814)*x[0]+(0.239606043833)*x[1]**o+(0.814911366356)*x[1]
        arg[0,3]=(0.171706130967)*x[0]**o+(0.56324398491)*x[0]+(0.328884588679)*x[1]**o+(0.13307521808)*x[1]
        arg[0,4]=(-0.309130885049)*x[0]**o+(0.353596646674)*x[0]+(-0.282764686556)*x[1]**o+(-0.56977508596)*x[1]
        arg[1,0]=(0.974136161109)*x[0]**o+(-0.432477120938)*x[0]+(-0.841212164751)*x[1]**o+(-0.429211937256)*x[1]
        arg[1,1]=(0.0303890146585)*x[0]**o+(-0.712865964696)*x[0]+(0.572471431129)*x[1]**o+(-0.194324510922)*x[1]
        arg[1,2]=(-0.757134870981)*x[0]**o+(-0.018314520594)*x[0]+(0.250423290169)*x[1]**o+(-0.876678823729)*x[1]
        arg[1,3]=(-0.351885929068)*x[0]**o+(-0.561258605321)*x[0]+(0.208995575986)*x[1]**o+(0.428437856056)*x[1]
        arg[1,4]=(-0.313383912526)*x[0]**o+(-0.713484404328)*x[0]+(-0.279619274966)*x[1]**o+(0.72420661494)*x[1]
        arg[2,0]=(-0.686382641827)*x[0]**o+(0.678578408675)*x[0]+(-0.696532353786)*x[1]**o+(-0.154644995634)*x[1]
        arg[2,1]=(0.368548759951)*x[0]**o+(-0.674856247769)*x[0]+(0.908734263899)*x[1]**o+(0.331173572353)*x[1]
        arg[2,2]=(-0.1279136556)*x[0]**o+(-0.161474840876)*x[0]+(-0.775009015881)*x[1]**o+(-0.689552100815)*x[1]
        arg[2,3]=(-0.638450048485)*x[0]**o+(0.755485009392)*x[0]+(-0.793363159431)*x[1]**o+(0.255883452369)*x[1]
        arg[2,4]=(-0.831891512113)*x[0]**o+(-0.811678751836)*x[0]+(-0.487866621558)*x[1]**o+(0.240951890349)*x[1]
        arg[3,0]=(0.448249735424)*x[0]**o+(0.0809467993259)*x[0]+(0.232269153692)*x[1]**o+(0.477512628661)*x[1]
        arg[3,1]=(0.494093407104)*x[0]**o+(0.196761586998)*x[0]+(-0.702155296644)*x[1]**o+(-0.748641693731)*x[1]
        arg[3,2]=(0.432370067164)*x[0]**o+(-0.365671005874)*x[0]+(0.452009309356)*x[1]**o+(-0.951328172558)*x[1]
        arg[3,3]=(-0.993290240028)*x[0]**o+(0.172534714933)*x[0]+(0.00917693958413)*x[1]**o+(0.417515657393)*x[1]
        arg[3,4]=(0.541299019375)*x[0]**o+(0.479118172057)*x[0]+(-0.502475287154)*x[1]**o+(-0.04397899716)*x[1]
        ref[0,0]=(-0.120967658047)/(o+1.)+(0.213774478371)
        ref[0,1]=(-0.247665180527)/(o+1.)+(0.360728742109)
        ref[0,2]=(0.952942144598)/(o+1.)+(0.776292907085)
        ref[0,3]=(0.500590719646)/(o+1.)+(0.348159601495)
        ref[0,4]=(-0.591895571605)/(o+1.)+(-0.108089219643)
        ref[1,0]=(0.132923996358)/(o+1.)+(-0.430844529097)
        ref[1,1]=(0.602860445788)/(o+1.)+(-0.453595237809)
        ref[1,2]=(-0.506711580812)/(o+1.)+(-0.447496672161)
        ref[1,3]=(-0.142890353081)/(o+1.)+(-0.066410374632)
        ref[1,4]=(-0.593003187492)/(o+1.)+(0.00536110530621)
        ref[2,0]=(-1.38291499561)/(o+1.)+(0.261966706521)
        ref[2,1]=(1.27728302385)/(o+1.)+(-0.171841337708)
        ref[2,2]=(-0.902922671481)/(o+1.)+(-0.425513470846)
        ref[2,3]=(-1.43181320792)/(o+1.)+(0.505684230881)
        ref[2,4]=(-1.31975813367)/(o+1.)+(-0.285363430743)
        ref[3,0]=(0.680518889117)/(o+1.)+(0.279229713993)
        ref[3,1]=(-0.20806188954)/(o+1.)+(-0.275940053366)
        ref[3,2]=(0.88437937652)/(o+1.)+(-0.658499589216)
        ref[3,3]=(-0.984113300443)/(o+1.)+(0.295025186163)
        ref[3,4]=(0.0388237322207)/(o+1.)+(0.217569587448)
      else:
        arg[0,0]=(-0.866239193113)*x[0]**o+(0.479559390215)*x[0]+(0.512920648125)*x[1]**o+(-0.504775214511)*x[1]+(0.462210441279)*x[2]**o+(0.000249131516168)*x[2]
        arg[0,1]=(-0.11397206495)*x[0]**o+(0.0762933475692)*x[0]+(0.347716642793)*x[1]**o+(0.373262578958)*x[1]+(0.965468442129)*x[2]**o+(-0.743205128452)*x[2]
        arg[0,2]=(-0.214519408107)*x[0]**o+(-0.663675446512)*x[0]+(-0.424102625804)*x[1]**o+(-0.426644021856)*x[1]+(-0.0731662128704)*x[2]**o+(-0.375185307096)*x[2]
        arg[0,3]=(0.599595484989)*x[0]**o+(-0.303187098474)*x[0]+(0.278524134467)*x[1]**o+(-0.579432425528)*x[1]+(0.006698672684)*x[2]**o+(-0.820721120517)*x[2]
        arg[0,4]=(0.538847126305)*x[0]**o+(0.58093792853)*x[0]+(-0.21189097773)*x[1]**o+(0.637101424777)*x[1]+(0.0155994636183)*x[2]**o+(0.488675370945)*x[2]
        arg[1,0]=(-0.775581481064)*x[0]**o+(-0.968381856562)*x[0]+(-0.660755920498)*x[1]**o+(0.746481830633)*x[1]+(-0.300061615221)*x[2]**o+(-0.720031101537)*x[2]
        arg[1,1]=(0.990001351802)*x[0]**o+(-0.168997300944)*x[0]+(0.0134555819367)*x[1]**o+(0.969355591009)*x[1]+(-0.347553438064)*x[2]**o+(0.648537935984)*x[2]
        arg[1,2]=(-0.327760535516)*x[0]**o+(0.517489199537)*x[0]+(-0.0589046196438)*x[1]**o+(-0.893401439507)*x[1]+(-0.0821634552571)*x[2]**o+(0.861266440343)*x[2]
        arg[1,3]=(-0.808782893446)*x[0]**o+(-0.196363777506)*x[0]+(0.623837845847)*x[1]**o+(-0.05392424847)*x[1]+(0.616011657554)*x[2]**o+(0.800682023434)*x[2]
        arg[1,4]=(-0.425073222968)*x[0]**o+(-0.635248695204)*x[0]+(0.947222621311)*x[1]**o+(0.253198813179)*x[1]+(-0.907568459712)*x[2]**o+(-0.513049744062)*x[2]
        arg[2,0]=(-0.103867107528)*x[0]**o+(0.893031214224)*x[0]+(0.151292631908)*x[1]**o+(-0.407189878657)*x[1]+(0.618120612304)*x[2]**o+(0.983992964931)*x[2]
        arg[2,1]=(0.310432646627)*x[0]**o+(0.235143842057)*x[0]+(0.729683627093)*x[1]**o+(0.891778928523)*x[1]+(-0.148927367367)*x[2]**o+(0.913403960844)*x[2]
        arg[2,2]=(0.610293692912)*x[0]**o+(-0.28497731539)*x[0]+(-0.926462472232)*x[1]**o+(-0.427657211191)*x[1]+(0.467905160751)*x[2]**o+(0.195578740767)*x[2]
        arg[2,3]=(0.854869452114)*x[0]**o+(-0.752530632831)*x[0]+(0.0128220565485)*x[1]**o+(-0.726013651836)*x[1]+(-0.0979888078519)*x[2]**o+(-0.658955703807)*x[2]
        arg[2,4]=(0.536447848289)*x[0]**o+(-0.658205406579)*x[0]+(0.66427484293)*x[1]**o+(-0.250014589496)*x[1]+(-0.0654257159073)*x[2]**o+(0.538196496913)*x[2]
        arg[3,0]=(-0.514665034974)*x[0]**o+(-0.278189747423)*x[0]+(0.402996329453)*x[1]**o+(-0.0362969716175)*x[1]+(-0.707890815004)*x[2]**o+(-0.298762865096)*x[2]
        arg[3,1]=(0.7342439327)*x[0]**o+(-0.387336714116)*x[0]+(-0.67888634151)*x[1]**o+(0.647322640846)*x[1]+(-0.0346274738413)*x[2]**o+(0.985950301913)*x[2]
        arg[3,2]=(0.687555064124)*x[0]**o+(0.345506248144)*x[0]+(0.658692258064)*x[1]**o+(0.810796031447)*x[1]+(0.102608006398)*x[2]**o+(0.349877767777)*x[2]
        arg[3,3]=(-0.456241806837)*x[0]**o+(-0.181224535865)*x[0]+(0.401842456169)*x[1]**o+(-0.413936620519)*x[1]+(0.436794604974)*x[2]**o+(0.95051042082)*x[2]
        arg[3,4]=(-0.330739393408)*x[0]**o+(-0.261876658969)*x[0]+(0.287474017623)*x[1]**o+(0.463691723118)*x[1]+(0.757597823206)*x[2]**o+(0.49336246207)*x[2]
        ref[0,0]=(0.108891896292)/(o+1.)+(-0.01248334639)
        ref[0,1]=(1.19921301997)/(o+1.)+(-0.146824600962)
        ref[0,2]=(-0.711788246781)/(o+1.)+(-0.732752387732)
        ref[0,3]=(0.88481829214)/(o+1.)+(-0.85167032226)
        ref[0,4]=(0.342555612194)/(o+1.)+(0.853357362126)
        ref[1,0]=(-1.73639901678)/(o+1.)+(-0.470965563733)
        ref[1,1]=(0.655903495675)/(o+1.)+(0.724448113024)
        ref[1,2]=(-0.468828610417)/(o+1.)+(0.242677100186)
        ref[1,3]=(0.431066609955)/(o+1.)+(0.275196998729)
        ref[1,4]=(-0.385419061368)/(o+1.)+(-0.447549813044)
        ref[2,0]=(0.665546136685)/(o+1.)+(0.734917150249)
        ref[2,1]=(0.891188906354)/(o+1.)+(1.02016336571)
        ref[2,2]=(0.151736381432)/(o+1.)+(-0.258527892907)
        ref[2,3]=(0.76970270081)/(o+1.)+(-1.06874999424)
        ref[2,4]=(1.13529697531)/(o+1.)+(-0.185011749581)
        ref[3,0]=(-0.819559520525)/(o+1.)+(-0.306624792068)
        ref[3,1]=(0.020730117348)/(o+1.)+(0.622968114321)
        ref[3,2]=(1.44885532859)/(o+1.)+(0.753090023684)
        ref[3,3]=(0.382395254306)/(o+1.)+(0.177674632218)
        ref[3,4]=(0.714332447421)/(o+1.)+(0.34758876311)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.779261125448)*x[0]**o+(-0.640503838419)*x[0]+(0.589085035572)*x[1]**o+(-0.0426687544445)*x[1]
        arg[0,0,1]=(-0.419361292473)*x[0]**o+(0.0850684591011)*x[0]+(-0.440758336688)*x[1]**o+(0.151015611594)*x[1]
        arg[0,1,0]=(-0.727043966709)*x[0]**o+(0.711730063365)*x[0]+(0.657759819512)*x[1]**o+(-0.210981156609)*x[1]
        arg[0,1,1]=(-0.276491768644)*x[0]**o+(-0.324865684188)*x[0]+(-0.921856877627)*x[1]**o+(-0.246833902926)*x[1]
        arg[1,0,0]=(0.720541615152)*x[0]**o+(-0.0406851820986)*x[0]+(0.33976001267)*x[1]**o+(0.937989079378)*x[1]
        arg[1,0,1]=(0.239272962875)*x[0]**o+(0.482026352856)*x[0]+(0.159148474482)*x[1]**o+(0.680831564145)*x[1]
        arg[1,1,0]=(0.55170300456)*x[0]**o+(-0.261357073238)*x[0]+(0.298185649596)*x[1]**o+(-0.540598468483)*x[1]
        arg[1,1,1]=(0.508262701362)*x[0]**o+(-0.531692643817)*x[0]+(0.151424571021)*x[1]**o+(-0.77488755481)*x[1]
        arg[2,0,0]=(0.265201389846)*x[0]**o+(0.377300848247)*x[0]+(0.0402148688087)*x[1]**o+(0.0263806695524)*x[1]
        arg[2,0,1]=(0.321397565411)*x[0]**o+(-0.277650913698)*x[0]+(0.909430556637)*x[1]**o+(-0.43244155765)*x[1]
        arg[2,1,0]=(-0.671417887851)*x[0]**o+(-0.772328065901)*x[0]+(0.0896878386442)*x[1]**o+(0.584773963567)*x[1]
        arg[2,1,1]=(0.318995094412)*x[0]**o+(-0.396578444213)*x[0]+(-0.627767362562)*x[1]**o+(-0.588717342984)*x[1]
        arg[3,0,0]=(-0.154999830633)*x[0]**o+(-0.263434767944)*x[0]+(0.0704862479101)*x[1]**o+(0.216237505495)*x[1]
        arg[3,0,1]=(0.956868643095)*x[0]**o+(-0.0331644426873)*x[0]+(-0.236340025853)*x[1]**o+(-0.814705425837)*x[1]
        arg[3,1,0]=(-0.477985064084)*x[0]**o+(-0.368896833289)*x[0]+(0.208122873207)*x[1]**o+(-0.566525593432)*x[1]
        arg[3,1,1]=(-0.299257275248)*x[0]**o+(0.937022356746)*x[0]+(0.671724793932)*x[1]**o+(0.80705820341)*x[1]
        arg[4,0,0]=(0.857855249735)*x[0]**o+(0.537756541143)*x[0]+(0.611784225987)*x[1]**o+(0.944061991186)*x[1]
        arg[4,0,1]=(0.825779290308)*x[0]**o+(-0.636461636739)*x[0]+(-0.382007450888)*x[1]**o+(0.170178262303)*x[1]
        arg[4,1,0]=(0.275666857657)*x[0]**o+(-0.404983003194)*x[0]+(0.476932065223)*x[1]**o+(0.931645726493)*x[1]
        arg[4,1,1]=(0.956182179652)*x[0]**o+(0.543482655321)*x[0]+(0.60492494312)*x[1]**o+(0.384497838341)*x[1]
        arg[5,0,0]=(0.476937368598)*x[0]**o+(-0.588434801718)*x[0]+(0.865080125217)*x[1]**o+(-0.954493375041)*x[1]
        arg[5,0,1]=(0.151850214007)*x[0]**o+(0.252651124764)*x[0]+(-0.570132867968)*x[1]**o+(-0.0885902498736)*x[1]
        arg[5,1,0]=(-0.69330471578)*x[0]**o+(0.165874518143)*x[0]+(-0.610503496045)*x[1]**o+(-0.871951964138)*x[1]
        arg[5,1,1]=(-0.539191264945)*x[0]**o+(0.558474736927)*x[0]+(0.752045569996)*x[1]**o+(-0.5305308467)*x[1]
        ref[0,0,0]=(1.36834616102)/(o+1.)+(-0.341586296432)
        ref[0,0,1]=(-0.860119629161)/(o+1.)+(0.118042035348)
        ref[0,1,0]=(-0.0692841471964)/(o+1.)+(0.250374453378)
        ref[0,1,1]=(-1.19834864627)/(o+1.)+(-0.285849793557)
        ref[1,0,0]=(1.06030162782)/(o+1.)+(0.448651948639)
        ref[1,0,1]=(0.398421437357)/(o+1.)+(0.5814289585)
        ref[1,1,0]=(0.849888654155)/(o+1.)+(-0.400977770861)
        ref[1,1,1]=(0.659687272383)/(o+1.)+(-0.653290099314)
        ref[2,0,0]=(0.305416258655)/(o+1.)+(0.2018407589)
        ref[2,0,1]=(1.23082812205)/(o+1.)+(-0.355046235674)
        ref[2,1,0]=(-0.581730049206)/(o+1.)+(-0.0937770511668)
        ref[2,1,1]=(-0.30877226815)/(o+1.)+(-0.492647893598)
        ref[3,0,0]=(-0.0845135827233)/(o+1.)+(-0.0235986312242)
        ref[3,0,1]=(0.720528617241)/(o+1.)+(-0.423934934262)
        ref[3,1,0]=(-0.269862190877)/(o+1.)+(-0.46771121336)
        ref[3,1,1]=(0.372467518684)/(o+1.)+(0.872040280078)
        ref[4,0,0]=(1.46963947572)/(o+1.)+(0.740909266165)
        ref[4,0,1]=(0.44377183942)/(o+1.)+(-0.233141687218)
        ref[4,1,0]=(0.75259892288)/(o+1.)+(0.26333136165)
        ref[4,1,1]=(1.56110712277)/(o+1.)+(0.463990246831)
        ref[5,0,0]=(1.34201749382)/(o+1.)+(-0.771464088379)
        ref[5,0,1]=(-0.418282653961)/(o+1.)+(0.0820304374451)
        ref[5,1,0]=(-1.30380821182)/(o+1.)+(-0.353038722997)
        ref[5,1,1]=(0.212854305051)/(o+1.)+(0.0139719451134)
      else:
        arg[0,0,0]=(-0.825926001774)*x[0]**o+(0.418049316138)*x[0]+(0.668695279849)*x[1]**o+(-0.7323605258)*x[1]+(-0.637717742224)*x[2]**o+(0.257065974143)*x[2]
        arg[0,0,1]=(-0.169779448443)*x[0]**o+(-0.648612931029)*x[0]+(-0.643814352995)*x[1]**o+(0.396704595574)*x[1]+(-0.248766540261)*x[2]**o+(-0.736410673282)*x[2]
        arg[0,1,0]=(-0.371929441192)*x[0]**o+(0.438661796459)*x[0]+(-0.893450426264)*x[1]**o+(0.643494077636)*x[1]+(0.252099102122)*x[2]**o+(-0.0857798075076)*x[2]
        arg[0,1,1]=(0.336251711034)*x[0]**o+(-0.317426111079)*x[0]+(-0.801221986791)*x[1]**o+(0.281123033042)*x[1]+(0.921153543163)*x[2]**o+(0.497003835923)*x[2]
        arg[1,0,0]=(0.40897311872)*x[0]**o+(0.326894801733)*x[0]+(0.216268807497)*x[1]**o+(0.946235364974)*x[1]+(-0.480995485096)*x[2]**o+(0.869772937982)*x[2]
        arg[1,0,1]=(0.108715554449)*x[0]**o+(0.0847529231253)*x[0]+(0.866120965844)*x[1]**o+(0.43989772616)*x[1]+(0.67962061792)*x[2]**o+(0.90311988632)*x[2]
        arg[1,1,0]=(0.28571484173)*x[0]**o+(0.593050819853)*x[0]+(-0.384864010146)*x[1]**o+(-0.57920566259)*x[1]+(-0.304024223769)*x[2]**o+(-0.0279518764989)*x[2]
        arg[1,1,1]=(0.249126658892)*x[0]**o+(0.243355410215)*x[0]+(0.154816978135)*x[1]**o+(-0.260343533153)*x[1]+(-0.973133787852)*x[2]**o+(0.752545508531)*x[2]
        arg[2,0,0]=(0.103612497259)*x[0]**o+(0.0919234144108)*x[0]+(0.416394965635)*x[1]**o+(-0.0132359939268)*x[1]+(-0.522775855178)*x[2]**o+(-0.985811066906)*x[2]
        arg[2,0,1]=(-0.592641488144)*x[0]**o+(0.703703144485)*x[0]+(-0.0149046721584)*x[1]**o+(0.220368283581)*x[1]+(-0.656858164281)*x[2]**o+(0.165951102113)*x[2]
        arg[2,1,0]=(0.309492848012)*x[0]**o+(-0.301037623847)*x[0]+(-0.224814259669)*x[1]**o+(-0.950794406688)*x[1]+(-0.296088293363)*x[2]**o+(-0.712930293168)*x[2]
        arg[2,1,1]=(0.0592633594166)*x[0]**o+(-0.179007619512)*x[0]+(-0.922169690831)*x[1]**o+(-0.778228075089)*x[1]+(-0.135749453777)*x[2]**o+(-0.146265051618)*x[2]
        arg[3,0,0]=(-0.71773633584)*x[0]**o+(0.471551385738)*x[0]+(-0.041415831943)*x[1]**o+(0.343380640343)*x[1]+(-0.211821446187)*x[2]**o+(-0.661201893183)*x[2]
        arg[3,0,1]=(-0.159071045836)*x[0]**o+(0.216715127143)*x[0]+(-0.846847485213)*x[1]**o+(0.662226888693)*x[1]+(0.050806337433)*x[2]**o+(0.744071347051)*x[2]
        arg[3,1,0]=(-0.363431877105)*x[0]**o+(-0.461513470653)*x[0]+(-0.130947882739)*x[1]**o+(-0.624729599981)*x[1]+(0.778572567699)*x[2]**o+(0.324584316744)*x[2]
        arg[3,1,1]=(0.395432262346)*x[0]**o+(0.826450301981)*x[0]+(0.176721657975)*x[1]**o+(-0.370954952405)*x[1]+(-0.708437370609)*x[2]**o+(-0.150075114354)*x[2]
        arg[4,0,0]=(-0.18662583335)*x[0]**o+(-0.0120831234457)*x[0]+(0.805686139762)*x[1]**o+(0.30195413136)*x[1]+(0.648067854031)*x[2]**o+(0.28073240632)*x[2]
        arg[4,0,1]=(-0.913625644854)*x[0]**o+(-0.264637827098)*x[0]+(-0.6531590441)*x[1]**o+(0.396422722623)*x[1]+(0.280591271146)*x[2]**o+(0.498346191707)*x[2]
        arg[4,1,0]=(0.208029597907)*x[0]**o+(-0.295747909396)*x[0]+(-0.541509895186)*x[1]**o+(-0.511394562575)*x[1]+(-0.459197562982)*x[2]**o+(-0.749433614666)*x[2]
        arg[4,1,1]=(-0.29700028905)*x[0]**o+(-0.590932598278)*x[0]+(0.36219700213)*x[1]**o+(0.482759242709)*x[1]+(-0.135696240971)*x[2]**o+(0.49658807849)*x[2]
        arg[5,0,0]=(-0.800345624219)*x[0]**o+(0.259709746207)*x[0]+(0.480073493816)*x[1]**o+(0.499332550127)*x[1]+(0.691794186846)*x[2]**o+(-0.319369042249)*x[2]
        arg[5,0,1]=(0.300879052118)*x[0]**o+(-0.985396817441)*x[0]+(0.971136879002)*x[1]**o+(0.806560091745)*x[1]+(0.361308859099)*x[2]**o+(0.264277089732)*x[2]
        arg[5,1,0]=(0.271857906548)*x[0]**o+(0.338260411675)*x[0]+(-0.170292665803)*x[1]**o+(0.675491566142)*x[1]+(0.897921755421)*x[2]**o+(-0.829233864443)*x[2]
        arg[5,1,1]=(0.673366845241)*x[0]**o+(0.121940550422)*x[0]+(0.116054155545)*x[1]**o+(0.0456556938689)*x[1]+(-0.978127590673)*x[2]**o+(0.180985351995)*x[2]
        ref[0,0,0]=(-0.794948464148)/(o+1.)+(-0.0286226177598)
        ref[0,0,1]=(-1.0623603417)/(o+1.)+(-0.494159504368)
        ref[0,1,0]=(-1.01328076533)/(o+1.)+(0.498188033294)
        ref[0,1,1]=(0.456183267406)/(o+1.)+(0.230350378943)
        ref[1,0,0]=(0.144246441122)/(o+1.)+(1.07145155234)
        ref[1,0,1]=(1.65445713821)/(o+1.)+(0.713885267803)
        ref[1,1,0]=(-0.403173392185)/(o+1.)+(-0.0070533596178)
        ref[1,1,1]=(-0.569190150825)/(o+1.)+(0.367778692797)
        ref[2,0,0]=(-0.00276839228366)/(o+1.)+(-0.453561823211)
        ref[2,0,1]=(-1.26440432458)/(o+1.)+(0.545011265089)
        ref[2,1,0]=(-0.21140970502)/(o+1.)+(-0.982381161851)
        ref[2,1,1]=(-0.998655785191)/(o+1.)+(-0.55175037311)
        ref[3,0,0]=(-0.97097361397)/(o+1.)+(0.0768650664488)
        ref[3,0,1]=(-0.955112193616)/(o+1.)+(0.811506681444)
        ref[3,1,0]=(0.284192807855)/(o+1.)+(-0.380829376946)
        ref[3,1,1]=(-0.136283450288)/(o+1.)+(0.152710117611)
        ref[4,0,0]=(1.26712816044)/(o+1.)+(0.285301707117)
        ref[4,0,1]=(-1.28619341781)/(o+1.)+(0.315065543616)
        ref[4,1,0]=(-0.792677860261)/(o+1.)+(-0.778288043318)
        ref[4,1,1]=(-0.0704995278906)/(o+1.)+(0.194207361461)
        ref[5,0,0]=(0.371522056443)/(o+1.)+(0.219836627043)
        ref[5,0,1]=(1.63332479022)/(o+1.)+(0.0427201820182)
        ref[5,1,0]=(0.999486996166)/(o+1.)+(0.0922590566873)
        ref[5,1,1]=(-0.188706589887)/(o+1.)+(0.174290798143)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.647932589381)*x[0]**o+(-0.092686685918)*x[0]+(0.167888464541)*x[1]**o+(0.698445840627)*x[1]
        arg[0,0,0,1]=(0.413204959658)*x[0]**o+(0.772926029512)*x[0]+(-0.63702948026)*x[1]**o+(0.608392823861)*x[1]
        arg[0,0,1,0]=(0.303506932052)*x[0]**o+(0.650309664063)*x[0]+(-0.805238344485)*x[1]**o+(-0.62971680995)*x[1]
        arg[0,0,1,1]=(-0.205701190736)*x[0]**o+(-0.0834042840814)*x[0]+(-0.525839398431)*x[1]**o+(0.779973223005)*x[1]
        arg[0,0,2,0]=(-0.97954633631)*x[0]**o+(-0.298973871114)*x[0]+(0.858022652298)*x[1]**o+(0.667671796632)*x[1]
        arg[0,0,2,1]=(-0.790133712345)*x[0]**o+(-0.733743414808)*x[0]+(0.337185148145)*x[1]**o+(-0.927853051244)*x[1]
        arg[0,1,0,0]=(0.617198112458)*x[0]**o+(0.70438849416)*x[0]+(-0.861661596364)*x[1]**o+(-0.251796116219)*x[1]
        arg[0,1,0,1]=(0.37107453636)*x[0]**o+(-0.490440661153)*x[0]+(0.849252675242)*x[1]**o+(-0.896547775077)*x[1]
        arg[0,1,1,0]=(0.375080744007)*x[0]**o+(-0.609557663503)*x[0]+(-0.930846980977)*x[1]**o+(0.0623096843381)*x[1]
        arg[0,1,1,1]=(-0.809863615805)*x[0]**o+(-0.999724521803)*x[0]+(0.493027030633)*x[1]**o+(-0.672546253752)*x[1]
        arg[0,1,2,0]=(0.219294077412)*x[0]**o+(-0.7854819847)*x[0]+(-0.353026321798)*x[1]**o+(-0.965348418509)*x[1]
        arg[0,1,2,1]=(0.177236321722)*x[0]**o+(-0.860491592895)*x[0]+(-0.458724268899)*x[1]**o+(-0.476508229653)*x[1]
        arg[0,2,0,0]=(0.0760611767074)*x[0]**o+(-0.00296398964042)*x[0]+(-0.972843039436)*x[1]**o+(-0.0422057743881)*x[1]
        arg[0,2,0,1]=(-0.926737777701)*x[0]**o+(-0.895791593378)*x[0]+(-0.776529150964)*x[1]**o+(0.557115706978)*x[1]
        arg[0,2,1,0]=(0.353264702858)*x[0]**o+(0.632583933138)*x[0]+(-0.354112684469)*x[1]**o+(0.0541346174117)*x[1]
        arg[0,2,1,1]=(-0.124181989217)*x[0]**o+(-0.803262099858)*x[0]+(0.733246361984)*x[1]**o+(0.0639821744038)*x[1]
        arg[0,2,2,0]=(-0.446733845136)*x[0]**o+(-0.187373171381)*x[0]+(-0.312309043572)*x[1]**o+(0.830789380477)*x[1]
        arg[0,2,2,1]=(0.403889932995)*x[0]**o+(-0.243474625233)*x[0]+(0.35707779102)*x[1]**o+(0.29725421988)*x[1]
        arg[0,3,0,0]=(-0.856376546696)*x[0]**o+(0.448431504377)*x[0]+(-0.535187100108)*x[1]**o+(-0.90994242894)*x[1]
        arg[0,3,0,1]=(-0.321475754775)*x[0]**o+(0.0389358177737)*x[0]+(-0.0949733279735)*x[1]**o+(0.921937240884)*x[1]
        arg[0,3,1,0]=(0.591276336387)*x[0]**o+(-0.598236438303)*x[0]+(-0.829829995185)*x[1]**o+(0.539669346291)*x[1]
        arg[0,3,1,1]=(-0.944726366074)*x[0]**o+(-0.796928273271)*x[0]+(-0.424606656681)*x[1]**o+(0.35341306063)*x[1]
        arg[0,3,2,0]=(-0.850914599406)*x[0]**o+(0.791179202404)*x[0]+(-0.0220845565851)*x[1]**o+(0.594756136957)*x[1]
        arg[0,3,2,1]=(0.355862680641)*x[0]**o+(0.134261680432)*x[0]+(0.477798827075)*x[1]**o+(0.508452338499)*x[1]
        arg[0,4,0,0]=(0.481974944781)*x[0]**o+(0.259271175551)*x[0]+(0.953827692515)*x[1]**o+(0.364160709544)*x[1]
        arg[0,4,0,1]=(-0.316627315143)*x[0]**o+(-0.856616735533)*x[0]+(0.308508314695)*x[1]**o+(-0.677602758367)*x[1]
        arg[0,4,1,0]=(-0.76069824745)*x[0]**o+(0.282351651183)*x[0]+(-0.15964276552)*x[1]**o+(-0.344276428838)*x[1]
        arg[0,4,1,1]=(9.00669303738e-05)*x[0]**o+(0.948498888115)*x[0]+(0.690370267333)*x[1]**o+(0.619119188327)*x[1]
        arg[0,4,2,0]=(0.957692561039)*x[0]**o+(-0.147005436631)*x[0]+(-0.177840574215)*x[1]**o+(-0.28259452614)*x[1]
        arg[0,4,2,1]=(-0.431771943451)*x[0]**o+(-0.158132166651)*x[0]+(-0.134385316147)*x[1]**o+(0.751243590835)*x[1]
        arg[1,0,0,0]=(-0.816841420645)*x[0]**o+(0.55608373851)*x[0]+(-0.885838014201)*x[1]**o+(-0.467159163867)*x[1]
        arg[1,0,0,1]=(-0.543361117401)*x[0]**o+(0.329490019675)*x[0]+(0.942703992489)*x[1]**o+(-0.261273838803)*x[1]
        arg[1,0,1,0]=(0.865108106801)*x[0]**o+(-0.376972427024)*x[0]+(-0.0257304693919)*x[1]**o+(0.694210909726)*x[1]
        arg[1,0,1,1]=(0.428354498121)*x[0]**o+(0.962704601892)*x[0]+(-0.442636806139)*x[1]**o+(0.424724924149)*x[1]
        arg[1,0,2,0]=(-0.157277242299)*x[0]**o+(0.487588443251)*x[0]+(-0.813089096341)*x[1]**o+(0.0406316718829)*x[1]
        arg[1,0,2,1]=(0.345322702588)*x[0]**o+(0.224429042659)*x[0]+(-0.24109926586)*x[1]**o+(0.433141241931)*x[1]
        arg[1,1,0,0]=(0.956901263519)*x[0]**o+(0.48418062914)*x[0]+(0.14089030013)*x[1]**o+(-0.492352201717)*x[1]
        arg[1,1,0,1]=(-0.679781695411)*x[0]**o+(0.29857196636)*x[0]+(-0.234968984185)*x[1]**o+(0.114575919411)*x[1]
        arg[1,1,1,0]=(0.488378272934)*x[0]**o+(-0.358508327343)*x[0]+(-0.337958846849)*x[1]**o+(0.189303452912)*x[1]
        arg[1,1,1,1]=(-0.506993859574)*x[0]**o+(-0.682478021258)*x[0]+(-0.540344338899)*x[1]**o+(0.366965535976)*x[1]
        arg[1,1,2,0]=(-0.681151638572)*x[0]**o+(0.790837550009)*x[0]+(-0.810785716023)*x[1]**o+(-0.00739097962416)*x[1]
        arg[1,1,2,1]=(-0.355290573677)*x[0]**o+(-0.263294580108)*x[0]+(0.10469018402)*x[1]**o+(-0.650388283116)*x[1]
        arg[1,2,0,0]=(0.906212689905)*x[0]**o+(-0.673956103538)*x[0]+(0.305335278934)*x[1]**o+(-0.0805949628719)*x[1]
        arg[1,2,0,1]=(-0.98311373852)*x[0]**o+(-0.165316704542)*x[0]+(0.197637770603)*x[1]**o+(0.647989184342)*x[1]
        arg[1,2,1,0]=(0.415990736251)*x[0]**o+(-0.710476520822)*x[0]+(0.791512786296)*x[1]**o+(-0.10853188742)*x[1]
        arg[1,2,1,1]=(0.04337234954)*x[0]**o+(0.595650094449)*x[0]+(0.228098665648)*x[1]**o+(-0.868930899722)*x[1]
        arg[1,2,2,0]=(-0.791842482688)*x[0]**o+(-0.611246387621)*x[0]+(0.0504978789412)*x[1]**o+(0.624626300694)*x[1]
        arg[1,2,2,1]=(0.0207375595263)*x[0]**o+(0.560638072838)*x[0]+(0.44126312508)*x[1]**o+(0.169151353084)*x[1]
        arg[1,3,0,0]=(-0.0512069089279)*x[0]**o+(0.51110271913)*x[0]+(0.0785633458626)*x[1]**o+(0.836821453174)*x[1]
        arg[1,3,0,1]=(0.185737468851)*x[0]**o+(0.516071031227)*x[0]+(0.0995388802342)*x[1]**o+(0.25527909092)*x[1]
        arg[1,3,1,0]=(0.318240681435)*x[0]**o+(-0.442725217398)*x[0]+(0.533066616909)*x[1]**o+(0.679570707339)*x[1]
        arg[1,3,1,1]=(0.353367397644)*x[0]**o+(-0.396487189861)*x[0]+(-0.117773711001)*x[1]**o+(0.885897675293)*x[1]
        arg[1,3,2,0]=(-0.647645988694)*x[0]**o+(-0.755258662164)*x[0]+(0.288300257472)*x[1]**o+(0.505342385389)*x[1]
        arg[1,3,2,1]=(0.000781386924262)*x[0]**o+(0.190244787656)*x[0]+(0.589388834071)*x[1]**o+(0.669705413859)*x[1]
        arg[1,4,0,0]=(0.320107896841)*x[0]**o+(0.673508197912)*x[0]+(0.667026895874)*x[1]**o+(-0.818578971562)*x[1]
        arg[1,4,0,1]=(-0.226755915796)*x[0]**o+(0.71800207991)*x[0]+(0.850490212043)*x[1]**o+(0.261708964138)*x[1]
        arg[1,4,1,0]=(-0.0147350282361)*x[0]**o+(0.743889536542)*x[0]+(-0.898802067607)*x[1]**o+(0.711589170571)*x[1]
        arg[1,4,1,1]=(-0.816145493818)*x[0]**o+(0.405931568806)*x[0]+(0.159156576579)*x[1]**o+(0.143576116984)*x[1]
        arg[1,4,2,0]=(0.994791165013)*x[0]**o+(0.65054901122)*x[0]+(0.039218059733)*x[1]**o+(-0.662978657934)*x[1]
        arg[1,4,2,1]=(-0.792277825616)*x[0]**o+(-0.257224872111)*x[0]+(-0.244502392787)*x[1]**o+(-0.276042189495)*x[1]
        arg[2,0,0,0]=(0.133211578016)*x[0]**o+(0.10029141733)*x[0]+(-0.52480547182)*x[1]**o+(0.814652707336)*x[1]
        arg[2,0,0,1]=(0.836429340685)*x[0]**o+(-0.837709322969)*x[0]+(-0.442953015462)*x[1]**o+(-0.984757344044)*x[1]
        arg[2,0,1,0]=(0.893738853578)*x[0]**o+(-0.238050980408)*x[0]+(-0.913668527302)*x[1]**o+(0.895136777323)*x[1]
        arg[2,0,1,1]=(-0.0457361766611)*x[0]**o+(0.414814798208)*x[0]+(-0.57261759439)*x[1]**o+(-0.358969048076)*x[1]
        arg[2,0,2,0]=(-0.320394419991)*x[0]**o+(-0.990129952721)*x[0]+(0.890888895016)*x[1]**o+(0.241063301452)*x[1]
        arg[2,0,2,1]=(-0.604164679932)*x[0]**o+(-0.598939404114)*x[0]+(-0.606633388454)*x[1]**o+(-0.666879580873)*x[1]
        arg[2,1,0,0]=(0.68515547034)*x[0]**o+(0.892247437859)*x[0]+(0.642003383991)*x[1]**o+(0.569240970392)*x[1]
        arg[2,1,0,1]=(0.918713108587)*x[0]**o+(0.865547962284)*x[0]+(0.517952260633)*x[1]**o+(-0.307334010857)*x[1]
        arg[2,1,1,0]=(0.353382468404)*x[0]**o+(-0.542718692644)*x[0]+(-0.957182863451)*x[1]**o+(-0.889352632225)*x[1]
        arg[2,1,1,1]=(0.505185847817)*x[0]**o+(-0.71137143775)*x[0]+(0.853785404375)*x[1]**o+(-0.731028307335)*x[1]
        arg[2,1,2,0]=(-0.0805774093876)*x[0]**o+(0.949831911897)*x[0]+(0.511009248375)*x[1]**o+(-0.0192710206538)*x[1]
        arg[2,1,2,1]=(-0.987709862593)*x[0]**o+(-0.536077508064)*x[0]+(-0.0726882784026)*x[1]**o+(-0.27275457372)*x[1]
        arg[2,2,0,0]=(0.271001953238)*x[0]**o+(0.0446246860776)*x[0]+(0.103657462724)*x[1]**o+(0.359900072694)*x[1]
        arg[2,2,0,1]=(0.613460277623)*x[0]**o+(-0.23324730519)*x[0]+(0.188458167579)*x[1]**o+(-0.574087648909)*x[1]
        arg[2,2,1,0]=(-0.787444207567)*x[0]**o+(0.143772612473)*x[0]+(-0.720451072166)*x[1]**o+(0.936819171338)*x[1]
        arg[2,2,1,1]=(-0.219701379296)*x[0]**o+(-0.343100221149)*x[0]+(-0.972737482064)*x[1]**o+(-0.614208415137)*x[1]
        arg[2,2,2,0]=(0.0292070295438)*x[0]**o+(-0.694263454424)*x[0]+(0.188064440036)*x[1]**o+(0.0901763471697)*x[1]
        arg[2,2,2,1]=(0.407629585245)*x[0]**o+(-0.745619322197)*x[0]+(0.949044983116)*x[1]**o+(-0.316057288128)*x[1]
        arg[2,3,0,0]=(0.124212182643)*x[0]**o+(0.621703244759)*x[0]+(-0.0592136772296)*x[1]**o+(0.571981450786)*x[1]
        arg[2,3,0,1]=(-0.683989210442)*x[0]**o+(-0.318774594157)*x[0]+(0.850235115277)*x[1]**o+(-0.871075179838)*x[1]
        arg[2,3,1,0]=(0.478657634044)*x[0]**o+(-0.43702074673)*x[0]+(0.535572237843)*x[1]**o+(0.22354698199)*x[1]
        arg[2,3,1,1]=(0.0391087869133)*x[0]**o+(0.765140505702)*x[0]+(-0.66279875624)*x[1]**o+(-0.41964006149)*x[1]
        arg[2,3,2,0]=(-0.11123799186)*x[0]**o+(0.533398540082)*x[0]+(-0.931488199925)*x[1]**o+(-0.83080969281)*x[1]
        arg[2,3,2,1]=(-0.138259313941)*x[0]**o+(-0.508937684332)*x[0]+(0.583603445711)*x[1]**o+(-0.0782724872225)*x[1]
        arg[2,4,0,0]=(-0.316223936951)*x[0]**o+(0.329776289164)*x[0]+(-0.692491559115)*x[1]**o+(0.814572730294)*x[1]
        arg[2,4,0,1]=(0.74890960681)*x[0]**o+(-0.79167945549)*x[0]+(0.175328119487)*x[1]**o+(-0.253003240064)*x[1]
        arg[2,4,1,0]=(0.657827625141)*x[0]**o+(-0.268878597055)*x[0]+(0.247374286921)*x[1]**o+(0.414500041415)*x[1]
        arg[2,4,1,1]=(-0.187587046202)*x[0]**o+(-0.594983066039)*x[0]+(-0.203784406879)*x[1]**o+(0.122237566739)*x[1]
        arg[2,4,2,0]=(0.671210794607)*x[0]**o+(0.886706498415)*x[0]+(-0.368997114212)*x[1]**o+(0.0418311781187)*x[1]
        arg[2,4,2,1]=(-0.272424416171)*x[0]**o+(-0.537477247783)*x[0]+(0.0405655834868)*x[1]**o+(-0.179684833703)*x[1]
        arg[3,0,0,0]=(0.457602964193)*x[0]**o+(-0.186625111516)*x[0]+(-0.75244105343)*x[1]**o+(-0.278893445377)*x[1]
        arg[3,0,0,1]=(-0.690378721784)*x[0]**o+(0.0157608350188)*x[0]+(-0.203765434813)*x[1]**o+(0.69771096957)*x[1]
        arg[3,0,1,0]=(-0.884824688493)*x[0]**o+(0.26613451569)*x[0]+(0.226081491639)*x[1]**o+(0.597724298389)*x[1]
        arg[3,0,1,1]=(0.0372124540969)*x[0]**o+(-0.420242551812)*x[0]+(-0.625683613012)*x[1]**o+(0.279682305563)*x[1]
        arg[3,0,2,0]=(-0.570873872458)*x[0]**o+(-0.82361574923)*x[0]+(0.970204188337)*x[1]**o+(-0.242725003569)*x[1]
        arg[3,0,2,1]=(-0.714826095124)*x[0]**o+(-0.334939427182)*x[0]+(-0.730058530908)*x[1]**o+(0.287148154312)*x[1]
        arg[3,1,0,0]=(0.164234307274)*x[0]**o+(-0.388849394356)*x[0]+(0.391518896617)*x[1]**o+(0.00142750126236)*x[1]
        arg[3,1,0,1]=(-0.327700781196)*x[0]**o+(0.0442604682565)*x[0]+(-0.683834466536)*x[1]**o+(-0.665567774985)*x[1]
        arg[3,1,1,0]=(0.135979934041)*x[0]**o+(-0.822614011173)*x[0]+(-0.946061768113)*x[1]**o+(-0.585442102671)*x[1]
        arg[3,1,1,1]=(0.235568769871)*x[0]**o+(0.738233870703)*x[0]+(0.437750339594)*x[1]**o+(-0.44521724321)*x[1]
        arg[3,1,2,0]=(0.449256898185)*x[0]**o+(-0.464211755196)*x[0]+(0.616886411002)*x[1]**o+(0.885500269963)*x[1]
        arg[3,1,2,1]=(0.0851105760559)*x[0]**o+(0.888976499957)*x[0]+(0.0842441111314)*x[1]**o+(-0.366994083394)*x[1]
        arg[3,2,0,0]=(0.77120050858)*x[0]**o+(0.354835584949)*x[0]+(0.406813878866)*x[1]**o+(-0.600042478179)*x[1]
        arg[3,2,0,1]=(-0.747876879184)*x[0]**o+(0.527204926634)*x[0]+(0.134965492713)*x[1]**o+(0.497662739459)*x[1]
        arg[3,2,1,0]=(0.367550591163)*x[0]**o+(-0.700608509353)*x[0]+(0.0737696587931)*x[1]**o+(-0.164466899816)*x[1]
        arg[3,2,1,1]=(0.0908791429433)*x[0]**o+(-0.168468767791)*x[0]+(0.85978873292)*x[1]**o+(-0.785241412884)*x[1]
        arg[3,2,2,0]=(-0.296455818292)*x[0]**o+(-0.0277004420512)*x[0]+(0.409183414933)*x[1]**o+(-0.879831945543)*x[1]
        arg[3,2,2,1]=(0.971512010223)*x[0]**o+(-0.811249982262)*x[0]+(-0.846549204737)*x[1]**o+(0.689986186691)*x[1]
        arg[3,3,0,0]=(-0.800235545702)*x[0]**o+(-0.796021196346)*x[0]+(-0.0660744375539)*x[1]**o+(-0.826632192842)*x[1]
        arg[3,3,0,1]=(0.305678237361)*x[0]**o+(0.203936000462)*x[0]+(-0.583171879217)*x[1]**o+(0.924112859119)*x[1]
        arg[3,3,1,0]=(-0.937148913762)*x[0]**o+(-0.172547991203)*x[0]+(0.928491154238)*x[1]**o+(-0.502876301351)*x[1]
        arg[3,3,1,1]=(0.731413856616)*x[0]**o+(0.837213097155)*x[0]+(-0.650032764358)*x[1]**o+(0.96973961589)*x[1]
        arg[3,3,2,0]=(-0.501994458598)*x[0]**o+(-0.705938508685)*x[0]+(0.664394117366)*x[1]**o+(0.692950029135)*x[1]
        arg[3,3,2,1]=(-0.0948399323409)*x[0]**o+(-0.686744554674)*x[0]+(0.577765681583)*x[1]**o+(0.969310750204)*x[1]
        arg[3,4,0,0]=(-0.484011788432)*x[0]**o+(-0.560744139846)*x[0]+(0.072097608773)*x[1]**o+(-0.103052228904)*x[1]
        arg[3,4,0,1]=(-0.201158934695)*x[0]**o+(0.463105335811)*x[0]+(0.733365706682)*x[1]**o+(-0.330035241794)*x[1]
        arg[3,4,1,0]=(0.252379277102)*x[0]**o+(-0.110596657969)*x[0]+(0.763687257878)*x[1]**o+(-0.345598371263)*x[1]
        arg[3,4,1,1]=(0.0868412855617)*x[0]**o+(0.385098679406)*x[0]+(0.291487288978)*x[1]**o+(-0.112145846209)*x[1]
        arg[3,4,2,0]=(-0.913724685255)*x[0]**o+(-0.0348735884313)*x[0]+(0.550688528861)*x[1]**o+(0.798394152329)*x[1]
        arg[3,4,2,1]=(-0.92870523219)*x[0]**o+(0.177354002469)*x[0]+(0.350393765313)*x[1]**o+(-0.974997198741)*x[1]
        ref[0,0,0,0]=(0.815821053922)/(o+1.)+(0.302879577355)
        ref[0,0,0,1]=(-0.223824520603)/(o+1.)+(0.690659426687)
        ref[0,0,1,0]=(-0.501731412433)/(o+1.)+(0.0102964270565)
        ref[0,0,1,1]=(-0.731540589167)/(o+1.)+(0.348284469462)
        ref[0,0,2,0]=(-0.121523684012)/(o+1.)+(0.184348962759)
        ref[0,0,2,1]=(-0.4529485642)/(o+1.)+(-0.830798233026)
        ref[0,1,0,0]=(-0.244463483906)/(o+1.)+(0.226296188971)
        ref[0,1,0,1]=(1.2203272116)/(o+1.)+(-0.693494218115)
        ref[0,1,1,0]=(-0.55576623697)/(o+1.)+(-0.273623989583)
        ref[0,1,1,1]=(-0.316836585171)/(o+1.)+(-0.836135387777)
        ref[0,1,2,0]=(-0.133732244387)/(o+1.)+(-0.875415201605)
        ref[0,1,2,1]=(-0.281487947177)/(o+1.)+(-0.668499911274)
        ref[0,2,0,0]=(-0.896781862729)/(o+1.)+(-0.0225848820142)
        ref[0,2,0,1]=(-1.70326692867)/(o+1.)+(-0.1693379432)
        ref[0,2,1,0]=(-0.00084798161079)/(o+1.)+(0.343359275275)
        ref[0,2,1,1]=(0.609064372767)/(o+1.)+(-0.369639962727)
        ref[0,2,2,0]=(-0.759042888708)/(o+1.)+(0.321708104548)
        ref[0,2,2,1]=(0.760967724015)/(o+1.)+(0.0268897973234)
        ref[0,3,0,0]=(-1.3915636468)/(o+1.)+(-0.230755462281)
        ref[0,3,0,1]=(-0.416449082748)/(o+1.)+(0.480436529329)
        ref[0,3,1,0]=(-0.238553658798)/(o+1.)+(-0.029283546006)
        ref[0,3,1,1]=(-1.36933302275)/(o+1.)+(-0.221757606321)
        ref[0,3,2,0]=(-0.872999155991)/(o+1.)+(0.69296766968)
        ref[0,3,2,1]=(0.833661507716)/(o+1.)+(0.321357009466)
        ref[0,4,0,0]=(1.4358026373)/(o+1.)+(0.311715942547)
        ref[0,4,0,1]=(-0.00811900044778)/(o+1.)+(-0.76710974695)
        ref[0,4,1,0]=(-0.92034101297)/(o+1.)+(-0.0309623888275)
        ref[0,4,1,1]=(0.690460334263)/(o+1.)+(0.783809038221)
        ref[0,4,2,0]=(0.779851986824)/(o+1.)+(-0.214799981386)
        ref[0,4,2,1]=(-0.566157259598)/(o+1.)+(0.296555712092)
        ref[1,0,0,0]=(-1.70267943485)/(o+1.)+(0.0444622873215)
        ref[1,0,0,1]=(0.399342875087)/(o+1.)+(0.0341080904363)
        ref[1,0,1,0]=(0.839377637409)/(o+1.)+(0.158619241351)
        ref[1,0,1,1]=(-0.0142823080176)/(o+1.)+(0.693714763021)
        ref[1,0,2,0]=(-0.97036633864)/(o+1.)+(0.264110057567)
        ref[1,0,2,1]=(0.104223436728)/(o+1.)+(0.328785142295)
        ref[1,1,0,0]=(1.09779156365)/(o+1.)+(-0.00408578628855)
        ref[1,1,0,1]=(-0.914750679597)/(o+1.)+(0.206573942886)
        ref[1,1,1,0]=(0.150419426085)/(o+1.)+(-0.0846024372156)
        ref[1,1,1,1]=(-1.04733819847)/(o+1.)+(-0.157756242641)
        ref[1,1,2,0]=(-1.49193735459)/(o+1.)+(0.391723285192)
        ref[1,1,2,1]=(-0.250600389657)/(o+1.)+(-0.456841431612)
        ref[1,2,0,0]=(1.21154796884)/(o+1.)+(-0.377275533205)
        ref[1,2,0,1]=(-0.785475967917)/(o+1.)+(0.2413362399)
        ref[1,2,1,0]=(1.20750352255)/(o+1.)+(-0.409504204121)
        ref[1,2,1,1]=(0.271471015188)/(o+1.)+(-0.136640402636)
        ref[1,2,2,0]=(-0.741344603747)/(o+1.)+(0.00668995653618)
        ref[1,2,2,1]=(0.462000684606)/(o+1.)+(0.364894712961)
        ref[1,3,0,0]=(0.0273564369347)/(o+1.)+(0.673962086152)
        ref[1,3,0,1]=(0.285276349085)/(o+1.)+(0.385675061073)
        ref[1,3,1,0]=(0.851307298344)/(o+1.)+(0.118422744971)
        ref[1,3,1,1]=(0.235593686643)/(o+1.)+(0.244705242716)
        ref[1,3,2,0]=(-0.359345731222)/(o+1.)+(-0.124958138388)
        ref[1,3,2,1]=(0.590170220995)/(o+1.)+(0.429975100757)
        ref[1,4,0,0]=(0.987134792715)/(o+1.)+(-0.0725353868253)
        ref[1,4,0,1]=(0.623734296247)/(o+1.)+(0.489855522024)
        ref[1,4,1,0]=(-0.913537095843)/(o+1.)+(0.727739353556)
        ref[1,4,1,1]=(-0.656988917238)/(o+1.)+(0.274753842895)
        ref[1,4,2,0]=(1.03400922475)/(o+1.)+(-0.00621482335664)
        ref[1,4,2,1]=(-1.0367802184)/(o+1.)+(-0.266633530803)
        ref[2,0,0,0]=(-0.391593893803)/(o+1.)+(0.457472062333)
        ref[2,0,0,1]=(0.393476325223)/(o+1.)+(-0.911233333506)
        ref[2,0,1,0]=(-0.0199296737245)/(o+1.)+(0.328542898457)
        ref[2,0,1,1]=(-0.618353771051)/(o+1.)+(0.0279228750664)
        ref[2,0,2,0]=(0.570494475025)/(o+1.)+(-0.374533325635)
        ref[2,0,2,1]=(-1.21079806839)/(o+1.)+(-0.632909492494)
        ref[2,1,0,0]=(1.32715885433)/(o+1.)+(0.730744204126)
        ref[2,1,0,1]=(1.43666536922)/(o+1.)+(0.279106975714)
        ref[2,1,1,0]=(-0.603800395047)/(o+1.)+(-0.716035662434)
        ref[2,1,1,1]=(1.35897125219)/(o+1.)+(-0.721199872542)
        ref[2,1,2,0]=(0.430431838988)/(o+1.)+(0.465280445622)
        ref[2,1,2,1]=(-1.060398141)/(o+1.)+(-0.404416040892)
        ref[2,2,0,0]=(0.374659415961)/(o+1.)+(0.202262379386)
        ref[2,2,0,1]=(0.801918445202)/(o+1.)+(-0.403667477049)
        ref[2,2,1,0]=(-1.50789527973)/(o+1.)+(0.540295891905)
        ref[2,2,1,1]=(-1.19243886136)/(o+1.)+(-0.478654318143)
        ref[2,2,2,0]=(0.21727146958)/(o+1.)+(-0.302043553627)
        ref[2,2,2,1]=(1.35667456836)/(o+1.)+(-0.530838305162)
        ref[2,3,0,0]=(0.0649985054133)/(o+1.)+(0.596842347772)
        ref[2,3,0,1]=(0.166245904835)/(o+1.)+(-0.594924886997)
        ref[2,3,1,0]=(1.01422987189)/(o+1.)+(-0.10673688237)
        ref[2,3,1,1]=(-0.623689969327)/(o+1.)+(0.172750222106)
        ref[2,3,2,0]=(-1.04272619178)/(o+1.)+(-0.148705576364)
        ref[2,3,2,1]=(0.445344131771)/(o+1.)+(-0.293605085777)
        ref[2,4,0,0]=(-1.00871549607)/(o+1.)+(0.572174509729)
        ref[2,4,0,1]=(0.924237726297)/(o+1.)+(-0.522341347777)
        ref[2,4,1,0]=(0.905201912063)/(o+1.)+(0.0728107221801)
        ref[2,4,1,1]=(-0.391371453081)/(o+1.)+(-0.23637274965)
        ref[2,4,2,0]=(0.302213680395)/(o+1.)+(0.464268838267)
        ref[2,4,2,1]=(-0.231858832685)/(o+1.)+(-0.358581040743)
        ref[3,0,0,0]=(-0.294838089237)/(o+1.)+(-0.232759278447)
        ref[3,0,0,1]=(-0.894144156597)/(o+1.)+(0.356735902294)
        ref[3,0,1,0]=(-0.658743196854)/(o+1.)+(0.43192940704)
        ref[3,0,1,1]=(-0.588471158915)/(o+1.)+(-0.0702801231243)
        ref[3,0,2,0]=(0.399330315879)/(o+1.)+(-0.5331703764)
        ref[3,0,2,1]=(-1.44488462603)/(o+1.)+(-0.0238956364351)
        ref[3,1,0,0]=(0.555753203891)/(o+1.)+(-0.193710946547)
        ref[3,1,0,1]=(-1.01153524773)/(o+1.)+(-0.310653653364)
        ref[3,1,1,0]=(-0.810081834072)/(o+1.)+(-0.704028056922)
        ref[3,1,1,1]=(0.673319109465)/(o+1.)+(0.146508313746)
        ref[3,1,2,0]=(1.06614330919)/(o+1.)+(0.210644257384)
        ref[3,1,2,1]=(0.169354687187)/(o+1.)+(0.260991208281)
        ref[3,2,0,0]=(1.17801438745)/(o+1.)+(-0.122603446615)
        ref[3,2,0,1]=(-0.612911386471)/(o+1.)+(0.512433833047)
        ref[3,2,1,0]=(0.441320249956)/(o+1.)+(-0.432537704584)
        ref[3,2,1,1]=(0.950667875863)/(o+1.)+(-0.476855090338)
        ref[3,2,2,0]=(0.112727596641)/(o+1.)+(-0.453766193797)
        ref[3,2,2,1]=(0.124962805486)/(o+1.)+(-0.0606318977858)
        ref[3,3,0,0]=(-0.866309983256)/(o+1.)+(-0.811326694594)
        ref[3,3,0,1]=(-0.277493641857)/(o+1.)+(0.564024429791)
        ref[3,3,1,0]=(-0.00865775952435)/(o+1.)+(-0.337712146277)
        ref[3,3,1,1]=(0.0813810922576)/(o+1.)+(0.903476356523)
        ref[3,3,2,0]=(0.162399658768)/(o+1.)+(-0.00649423977499)
        ref[3,3,2,1]=(0.482925749243)/(o+1.)+(0.141283097765)
        ref[3,4,0,0]=(-0.411914179659)/(o+1.)+(-0.331898184375)
        ref[3,4,0,1]=(0.532206771987)/(o+1.)+(0.0665350470083)
        ref[3,4,1,0]=(1.01606653498)/(o+1.)+(-0.228097514616)
        ref[3,4,1,1]=(0.37832857454)/(o+1.)+(0.136476416599)
        ref[3,4,2,0]=(-0.363036156393)/(o+1.)+(0.381760281949)
        ref[3,4,2,1]=(-0.578311466877)/(o+1.)+(-0.398821598136)
      else:
        arg[0,0,0,0]=(0.860676142569)*x[0]**o+(-0.156046991162)*x[0]+(-0.0462355579656)*x[1]**o+(0.70842844961)*x[1]+(0.617299123548)*x[2]**o+(-0.865174329565)*x[2]
        arg[0,0,0,1]=(-0.284308060865)*x[0]**o+(0.788848579389)*x[0]+(0.629321669042)*x[1]**o+(0.066637849809)*x[1]+(0.114152381009)*x[2]**o+(-0.466687745846)*x[2]
        arg[0,0,1,0]=(-0.690823198517)*x[0]**o+(-0.141510804573)*x[0]+(0.969978831967)*x[1]**o+(0.307926382658)*x[1]+(0.710373009094)*x[2]**o+(-0.494828315907)*x[2]
        arg[0,0,1,1]=(0.953361413913)*x[0]**o+(-0.103742611936)*x[0]+(0.63934955554)*x[1]**o+(-0.0778114476564)*x[1]+(-0.221185122698)*x[2]**o+(-0.287432047404)*x[2]
        arg[0,0,2,0]=(0.378306380558)*x[0]**o+(-0.117743922672)*x[0]+(-0.0606376630208)*x[1]**o+(-0.59439189505)*x[1]+(0.10507320657)*x[2]**o+(0.856411423067)*x[2]
        arg[0,0,2,1]=(0.777734672779)*x[0]**o+(0.286956002007)*x[0]+(-0.931811626583)*x[1]**o+(-0.806626659661)*x[1]+(-0.420390949282)*x[2]**o+(0.854080006578)*x[2]
        arg[0,1,0,0]=(0.403426948964)*x[0]**o+(0.707099419296)*x[0]+(0.593137130456)*x[1]**o+(-0.757560771598)*x[1]+(0.369632217934)*x[2]**o+(-0.348480285397)*x[2]
        arg[0,1,0,1]=(-0.491454189782)*x[0]**o+(0.547276640565)*x[0]+(-0.955045542708)*x[1]**o+(0.529989872531)*x[1]+(-0.161086874799)*x[2]**o+(-0.181810040724)*x[2]
        arg[0,1,1,0]=(0.46282177104)*x[0]**o+(0.254346937251)*x[0]+(0.824797105546)*x[1]**o+(-0.449979162246)*x[1]+(0.22307703071)*x[2]**o+(-0.592072135668)*x[2]
        arg[0,1,1,1]=(0.329230675141)*x[0]**o+(-0.733034697145)*x[0]+(0.762195609347)*x[1]**o+(-0.639097591362)*x[1]+(0.765354927905)*x[2]**o+(0.313571493405)*x[2]
        arg[0,1,2,0]=(0.3874127072)*x[0]**o+(-0.0569429727569)*x[0]+(-0.203121198708)*x[1]**o+(-0.736379499164)*x[1]+(-0.980629375436)*x[2]**o+(0.955793660048)*x[2]
        arg[0,1,2,1]=(-0.656255469455)*x[0]**o+(0.711904808699)*x[0]+(-0.363253678399)*x[1]**o+(0.921875000807)*x[1]+(0.686394816431)*x[2]**o+(0.818500716704)*x[2]
        arg[0,2,0,0]=(-0.931555347846)*x[0]**o+(-0.249573400742)*x[0]+(0.617166716404)*x[1]**o+(-0.929587051108)*x[1]+(0.830928715064)*x[2]**o+(0.461637431613)*x[2]
        arg[0,2,0,1]=(-0.8677719945)*x[0]**o+(-0.176892180445)*x[0]+(-0.881887456464)*x[1]**o+(0.923355677513)*x[1]+(-0.978544668453)*x[2]**o+(-0.427009384287)*x[2]
        arg[0,2,1,0]=(-0.0707937433084)*x[0]**o+(-0.872892141393)*x[0]+(0.606234798381)*x[1]**o+(0.935845076847)*x[1]+(-0.97559052201)*x[2]**o+(-0.31138518209)*x[2]
        arg[0,2,1,1]=(0.450210597865)*x[0]**o+(0.842939440378)*x[0]+(-0.333422856577)*x[1]**o+(-0.492809679443)*x[1]+(-0.254728278043)*x[2]**o+(-0.0702767077952)*x[2]
        arg[0,2,2,0]=(-0.250463216903)*x[0]**o+(0.756576293763)*x[0]+(0.596902629953)*x[1]**o+(-0.887080794947)*x[1]+(0.027319948008)*x[2]**o+(0.552496782776)*x[2]
        arg[0,2,2,1]=(-0.0883102140591)*x[0]**o+(0.988759610101)*x[0]+(-0.208152392508)*x[1]**o+(0.938999108564)*x[1]+(-0.773883611417)*x[2]**o+(0.653413306338)*x[2]
        arg[0,3,0,0]=(-0.399644211606)*x[0]**o+(-0.978595776407)*x[0]+(0.99454650986)*x[1]**o+(0.143313668715)*x[1]+(0.328592750842)*x[2]**o+(-0.958804581101)*x[2]
        arg[0,3,0,1]=(-0.0667966885959)*x[0]**o+(0.332653574704)*x[0]+(-0.762025150916)*x[1]**o+(0.895044013374)*x[1]+(0.627013843362)*x[2]**o+(-0.682392672232)*x[2]
        arg[0,3,1,0]=(-0.834479453773)*x[0]**o+(-0.375930859033)*x[0]+(0.129507966861)*x[1]**o+(0.995364280408)*x[1]+(-0.126081191572)*x[2]**o+(0.0344262779491)*x[2]
        arg[0,3,1,1]=(0.796952126596)*x[0]**o+(0.310562243091)*x[0]+(0.854548623788)*x[1]**o+(-0.719100537284)*x[1]+(-0.236266783438)*x[2]**o+(-0.646647129209)*x[2]
        arg[0,3,2,0]=(0.844073289115)*x[0]**o+(0.423300345347)*x[0]+(0.126354796852)*x[1]**o+(0.180372257606)*x[1]+(-0.564712722146)*x[2]**o+(0.466988430183)*x[2]
        arg[0,3,2,1]=(0.228775824367)*x[0]**o+(0.40701094831)*x[0]+(0.223101054593)*x[1]**o+(-0.894937905383)*x[1]+(0.425392942773)*x[2]**o+(-0.844238667169)*x[2]
        arg[0,4,0,0]=(0.3363848738)*x[0]**o+(-0.985327066062)*x[0]+(-0.107847971678)*x[1]**o+(0.596121801488)*x[1]+(0.655589801663)*x[2]**o+(-0.729650731949)*x[2]
        arg[0,4,0,1]=(-0.176435679219)*x[0]**o+(0.382558711725)*x[0]+(-0.658212492337)*x[1]**o+(0.584904267612)*x[1]+(-0.534684876705)*x[2]**o+(-0.441262614915)*x[2]
        arg[0,4,1,0]=(0.281301737167)*x[0]**o+(-0.642486274791)*x[0]+(0.420753884869)*x[1]**o+(0.482835955591)*x[1]+(0.169431333788)*x[2]**o+(0.922841193825)*x[2]
        arg[0,4,1,1]=(0.235812590604)*x[0]**o+(-0.446233855396)*x[0]+(0.465015868566)*x[1]**o+(-0.780441276888)*x[1]+(0.513632761351)*x[2]**o+(-0.0576482811678)*x[2]
        arg[0,4,2,0]=(-0.617245874462)*x[0]**o+(0.183661473466)*x[0]+(-0.711148911239)*x[1]**o+(0.125228991306)*x[1]+(-0.500218603959)*x[2]**o+(0.70597577229)*x[2]
        arg[0,4,2,1]=(0.56330304117)*x[0]**o+(0.631284986628)*x[0]+(-0.222485175776)*x[1]**o+(0.0639235737228)*x[1]+(-0.276613868747)*x[2]**o+(0.407488564857)*x[2]
        arg[1,0,0,0]=(0.159032146915)*x[0]**o+(-0.716040989693)*x[0]+(0.410709172613)*x[1]**o+(-0.527667963652)*x[1]+(-0.160515156497)*x[2]**o+(0.294448742382)*x[2]
        arg[1,0,0,1]=(0.941139035835)*x[0]**o+(0.707410046094)*x[0]+(-0.172095410803)*x[1]**o+(-0.967035204304)*x[1]+(0.870395272646)*x[2]**o+(0.830870986802)*x[2]
        arg[1,0,1,0]=(0.731015285309)*x[0]**o+(0.450519048697)*x[0]+(-0.484568714675)*x[1]**o+(0.454201220108)*x[1]+(-0.161850459099)*x[2]**o+(-0.617463263396)*x[2]
        arg[1,0,1,1]=(-0.958841628847)*x[0]**o+(-0.0667813400021)*x[0]+(0.659551780082)*x[1]**o+(0.514424015304)*x[1]+(0.920583608125)*x[2]**o+(0.89564899572)*x[2]
        arg[1,0,2,0]=(0.0482630479817)*x[0]**o+(0.911326064322)*x[0]+(-0.275134422092)*x[1]**o+(0.216495232459)*x[1]+(-0.973586171195)*x[2]**o+(0.275694688611)*x[2]
        arg[1,0,2,1]=(0.769733688293)*x[0]**o+(0.771829018421)*x[0]+(-0.22434549079)*x[1]**o+(0.131966885817)*x[1]+(-0.109424705679)*x[2]**o+(-0.369585136379)*x[2]
        arg[1,1,0,0]=(-0.0160180905877)*x[0]**o+(-0.268260755267)*x[0]+(-0.213940412022)*x[1]**o+(0.665987805279)*x[1]+(0.968879828017)*x[2]**o+(-0.449435538913)*x[2]
        arg[1,1,0,1]=(0.361055351825)*x[0]**o+(0.463294478488)*x[0]+(0.692655668486)*x[1]**o+(0.151415573206)*x[1]+(0.902280461515)*x[2]**o+(0.690926934073)*x[2]
        arg[1,1,1,0]=(0.591998602299)*x[0]**o+(0.516482177709)*x[0]+(-0.298511909182)*x[1]**o+(0.281818071894)*x[1]+(0.207089953979)*x[2]**o+(0.917004722388)*x[2]
        arg[1,1,1,1]=(0.479832408236)*x[0]**o+(-0.643374914165)*x[0]+(0.370844835249)*x[1]**o+(0.560458499239)*x[1]+(-0.865852319767)*x[2]**o+(0.560572130123)*x[2]
        arg[1,1,2,0]=(0.709293526635)*x[0]**o+(-0.388347223155)*x[0]+(0.60797330273)*x[1]**o+(-0.847624783099)*x[1]+(-0.0476367756382)*x[2]**o+(0.637841988478)*x[2]
        arg[1,1,2,1]=(0.343458045422)*x[0]**o+(-0.764451449587)*x[0]+(-0.444256801761)*x[1]**o+(0.385385997921)*x[1]+(-0.722095136671)*x[2]**o+(0.677852473207)*x[2]
        arg[1,2,0,0]=(0.710372887392)*x[0]**o+(-0.628494679119)*x[0]+(-0.59513507905)*x[1]**o+(-0.527824874295)*x[1]+(-0.611189170789)*x[2]**o+(0.090651976806)*x[2]
        arg[1,2,0,1]=(-0.980160150289)*x[0]**o+(-0.0286097581598)*x[0]+(-0.407906975953)*x[1]**o+(-0.319422835951)*x[1]+(-0.489787310147)*x[2]**o+(-0.694386341354)*x[2]
        arg[1,2,1,0]=(-0.416867224376)*x[0]**o+(0.631406877489)*x[0]+(-0.93939195942)*x[1]**o+(0.266858600313)*x[1]+(0.982173221612)*x[2]**o+(0.112901364617)*x[2]
        arg[1,2,1,1]=(0.412001480348)*x[0]**o+(0.624281489887)*x[0]+(0.306558457751)*x[1]**o+(0.95864454419)*x[1]+(0.299175627954)*x[2]**o+(-0.0531991017347)*x[2]
        arg[1,2,2,0]=(-0.0342261621741)*x[0]**o+(-0.587245704486)*x[0]+(-0.976407464107)*x[1]**o+(-0.0157215030322)*x[1]+(-0.529857472095)*x[2]**o+(-0.419626697673)*x[2]
        arg[1,2,2,1]=(-0.681737301037)*x[0]**o+(-0.980349460702)*x[0]+(-0.839469631393)*x[1]**o+(-0.412417814907)*x[1]+(0.451717398947)*x[2]**o+(0.962167685987)*x[2]
        arg[1,3,0,0]=(0.283453965913)*x[0]**o+(0.204445403787)*x[0]+(-0.806150376236)*x[1]**o+(0.122187094069)*x[1]+(-0.679291367964)*x[2]**o+(-0.875353966489)*x[2]
        arg[1,3,0,1]=(0.115974475175)*x[0]**o+(0.701473497654)*x[0]+(0.298012120304)*x[1]**o+(0.567597901709)*x[1]+(0.721211831476)*x[2]**o+(-0.255389572775)*x[2]
        arg[1,3,1,0]=(0.393891431092)*x[0]**o+(-0.563382537925)*x[0]+(0.974194829106)*x[1]**o+(0.659688559688)*x[1]+(-0.394266143828)*x[2]**o+(0.875266200191)*x[2]
        arg[1,3,1,1]=(-0.456930975697)*x[0]**o+(0.334590626231)*x[0]+(-0.227176883362)*x[1]**o+(0.76775699173)*x[1]+(0.839371167887)*x[2]**o+(-0.155473916326)*x[2]
        arg[1,3,2,0]=(0.346761981787)*x[0]**o+(-0.625121706322)*x[0]+(0.340655154638)*x[1]**o+(-0.0646323546999)*x[1]+(-0.211347402677)*x[2]**o+(0.233661266919)*x[2]
        arg[1,3,2,1]=(-0.086096644956)*x[0]**o+(-0.845623653209)*x[0]+(-0.0866093549957)*x[1]**o+(-0.467286349011)*x[1]+(-0.0621455515357)*x[2]**o+(0.164779979085)*x[2]
        arg[1,4,0,0]=(0.42022138208)*x[0]**o+(0.136984451355)*x[0]+(0.664770979086)*x[1]**o+(-0.537368896193)*x[1]+(0.874256726691)*x[2]**o+(-0.58614100824)*x[2]
        arg[1,4,0,1]=(-0.805860964792)*x[0]**o+(0.758511304044)*x[0]+(-0.574686863541)*x[1]**o+(-0.410873101247)*x[1]+(0.932015171551)*x[2]**o+(0.439987195351)*x[2]
        arg[1,4,1,0]=(-0.515187049123)*x[0]**o+(0.308078404334)*x[0]+(0.940959771774)*x[1]**o+(0.466269579691)*x[1]+(-0.190617144843)*x[2]**o+(0.738172533582)*x[2]
        arg[1,4,1,1]=(0.283326036586)*x[0]**o+(-0.693631747857)*x[0]+(0.478965080319)*x[1]**o+(-0.0465389396775)*x[1]+(-0.72886286959)*x[2]**o+(-0.9346424149)*x[2]
        arg[1,4,2,0]=(-0.71539766285)*x[0]**o+(0.361875987071)*x[0]+(0.905270678876)*x[1]**o+(-0.675024725702)*x[1]+(0.378259473658)*x[2]**o+(0.0520394226711)*x[2]
        arg[1,4,2,1]=(-0.319600525923)*x[0]**o+(0.0449956327134)*x[0]+(0.707418658521)*x[1]**o+(0.210137972776)*x[1]+(-0.767860728688)*x[2]**o+(0.232340825758)*x[2]
        arg[2,0,0,0]=(0.92844243651)*x[0]**o+(0.956810319504)*x[0]+(0.762198571194)*x[1]**o+(-0.363107634629)*x[1]+(0.0750393505254)*x[2]**o+(-0.43573581041)*x[2]
        arg[2,0,0,1]=(-0.193613121742)*x[0]**o+(0.440027871839)*x[0]+(-0.607905506346)*x[1]**o+(0.894840354958)*x[1]+(0.601279436178)*x[2]**o+(0.46806560762)*x[2]
        arg[2,0,1,0]=(0.370901517567)*x[0]**o+(-0.738769905647)*x[0]+(-0.698692979664)*x[1]**o+(-0.286318840335)*x[1]+(-0.879367315688)*x[2]**o+(-0.959912132991)*x[2]
        arg[2,0,1,1]=(-0.278939533774)*x[0]**o+(0.183795552426)*x[0]+(-0.341943551466)*x[1]**o+(0.491847260291)*x[1]+(-0.141030429384)*x[2]**o+(0.256287855539)*x[2]
        arg[2,0,2,0]=(-0.135930357616)*x[0]**o+(-0.382515749949)*x[0]+(-0.43030873581)*x[1]**o+(-0.0211891159981)*x[1]+(-0.377865120002)*x[2]**o+(0.109942325029)*x[2]
        arg[2,0,2,1]=(0.288827267201)*x[0]**o+(-0.907970544468)*x[0]+(-0.622258309213)*x[1]**o+(-0.763273879894)*x[1]+(0.438715013671)*x[2]**o+(-0.93873140988)*x[2]
        arg[2,1,0,0]=(-0.963487369906)*x[0]**o+(0.899051290288)*x[0]+(-0.268257956548)*x[1]**o+(-0.229092726584)*x[1]+(0.427024112314)*x[2]**o+(0.0435138095138)*x[2]
        arg[2,1,0,1]=(0.89996926597)*x[0]**o+(-0.206500706229)*x[0]+(-0.191769123095)*x[1]**o+(-0.348924155298)*x[1]+(-0.74405653776)*x[2]**o+(0.0724946691864)*x[2]
        arg[2,1,1,0]=(0.338170397901)*x[0]**o+(-0.725761081242)*x[0]+(-0.45928272676)*x[1]**o+(0.795535326783)*x[1]+(-0.316443844644)*x[2]**o+(-0.844205401892)*x[2]
        arg[2,1,1,1]=(0.509131757681)*x[0]**o+(0.319175801535)*x[0]+(-0.807058926279)*x[1]**o+(0.63928269996)*x[1]+(0.219324073432)*x[2]**o+(0.351217517434)*x[2]
        arg[2,1,2,0]=(-0.534090721611)*x[0]**o+(0.27146424294)*x[0]+(0.250778065716)*x[1]**o+(-0.901322994167)*x[1]+(-0.57588768699)*x[2]**o+(0.405043003456)*x[2]
        arg[2,1,2,1]=(0.101699906269)*x[0]**o+(0.361960318738)*x[0]+(0.278974923593)*x[1]**o+(-0.336289352037)*x[1]+(-0.677353936215)*x[2]**o+(0.0726793347753)*x[2]
        arg[2,2,0,0]=(-0.734961283823)*x[0]**o+(-0.593268557578)*x[0]+(0.310896858565)*x[1]**o+(-0.482326137951)*x[1]+(0.886472957099)*x[2]**o+(-0.0403563476581)*x[2]
        arg[2,2,0,1]=(-0.166071725106)*x[0]**o+(-0.454245686818)*x[0]+(-0.892674508991)*x[1]**o+(0.581554190251)*x[1]+(0.995323654908)*x[2]**o+(-0.0336714433349)*x[2]
        arg[2,2,1,0]=(-0.437491848099)*x[0]**o+(-0.195039508213)*x[0]+(0.822011086215)*x[1]**o+(0.771044412863)*x[1]+(0.697534261252)*x[2]**o+(0.666510217725)*x[2]
        arg[2,2,1,1]=(-0.529148276275)*x[0]**o+(-0.093016460024)*x[0]+(0.663641010649)*x[1]**o+(0.940439359843)*x[1]+(0.754151159716)*x[2]**o+(0.852349310591)*x[2]
        arg[2,2,2,0]=(0.859523244682)*x[0]**o+(0.704825125128)*x[0]+(0.221832129598)*x[1]**o+(0.573940020456)*x[1]+(0.935673751591)*x[2]**o+(0.23235170475)*x[2]
        arg[2,2,2,1]=(0.0359047473512)*x[0]**o+(0.836816935443)*x[0]+(0.6411608273)*x[1]**o+(0.291526953908)*x[1]+(0.735564515362)*x[2]**o+(-0.465781621769)*x[2]
        arg[2,3,0,0]=(0.352740040216)*x[0]**o+(0.54224020581)*x[0]+(0.681356174115)*x[1]**o+(-0.97589117869)*x[1]+(0.880755719884)*x[2]**o+(0.562259403362)*x[2]
        arg[2,3,0,1]=(0.563708471054)*x[0]**o+(0.525523219723)*x[0]+(0.423068360374)*x[1]**o+(0.29239160821)*x[1]+(0.874360368885)*x[2]**o+(0.922265851359)*x[2]
        arg[2,3,1,0]=(0.694309936305)*x[0]**o+(-0.409224769672)*x[0]+(-0.748835864803)*x[1]**o+(-0.21841534641)*x[1]+(0.497059735624)*x[2]**o+(-0.753234694148)*x[2]
        arg[2,3,1,1]=(0.109706510352)*x[0]**o+(-0.0160720403678)*x[0]+(0.888778080744)*x[1]**o+(-0.938692716268)*x[1]+(0.529262951478)*x[2]**o+(0.481430680214)*x[2]
        arg[2,3,2,0]=(0.65368659132)*x[0]**o+(-0.988176396392)*x[0]+(0.989546739333)*x[1]**o+(0.23178715846)*x[1]+(-0.557645438328)*x[2]**o+(0.063228513147)*x[2]
        arg[2,3,2,1]=(0.286316141831)*x[0]**o+(0.668138463424)*x[0]+(-0.950430671533)*x[1]**o+(-0.978012107187)*x[1]+(-0.107618382464)*x[2]**o+(-0.684084421276)*x[2]
        arg[2,4,0,0]=(-0.63510659432)*x[0]**o+(0.264242935995)*x[0]+(0.960647241624)*x[1]**o+(0.58483379289)*x[1]+(0.7255270271)*x[2]**o+(-0.476151653985)*x[2]
        arg[2,4,0,1]=(-0.221953908705)*x[0]**o+(-0.994143786509)*x[0]+(-0.359525421944)*x[1]**o+(0.533966197668)*x[1]+(-0.29739666047)*x[2]**o+(-0.0824204124466)*x[2]
        arg[2,4,1,0]=(-0.447043625871)*x[0]**o+(-0.904485300564)*x[0]+(0.26179071803)*x[1]**o+(0.437320719437)*x[1]+(-0.56501466909)*x[2]**o+(0.129662379166)*x[2]
        arg[2,4,1,1]=(0.259809756366)*x[0]**o+(0.446788808701)*x[0]+(-0.143702396023)*x[1]**o+(0.879531090832)*x[1]+(0.410047703156)*x[2]**o+(0.755750006468)*x[2]
        arg[2,4,2,0]=(-0.730881615951)*x[0]**o+(-0.357110792705)*x[0]+(0.552722654621)*x[1]**o+(0.965623848242)*x[1]+(0.555951256049)*x[2]**o+(0.259021370764)*x[2]
        arg[2,4,2,1]=(-0.465880467377)*x[0]**o+(-0.840984822217)*x[0]+(0.665817382039)*x[1]**o+(-0.390932961529)*x[1]+(-0.0857939888315)*x[2]**o+(-0.712185917672)*x[2]
        arg[3,0,0,0]=(0.710281429223)*x[0]**o+(-0.723903639057)*x[0]+(-0.267455509065)*x[1]**o+(-0.731684188612)*x[1]+(-0.287442751768)*x[2]**o+(-0.266871749947)*x[2]
        arg[3,0,0,1]=(0.460723870663)*x[0]**o+(0.0502154310124)*x[0]+(-0.260241363193)*x[1]**o+(-0.516473088648)*x[1]+(-0.925984113039)*x[2]**o+(-0.0297616649018)*x[2]
        arg[3,0,1,0]=(0.523186648278)*x[0]**o+(0.937042293367)*x[0]+(0.811740916657)*x[1]**o+(0.0544805175015)*x[1]+(-0.543484279955)*x[2]**o+(0.244714198306)*x[2]
        arg[3,0,1,1]=(-0.69212630167)*x[0]**o+(0.868708276385)*x[0]+(0.473290025811)*x[1]**o+(0.137043602294)*x[1]+(-0.19291314113)*x[2]**o+(-0.755147036188)*x[2]
        arg[3,0,2,0]=(0.0472638421122)*x[0]**o+(0.403135551064)*x[0]+(0.290083188289)*x[1]**o+(0.96419060449)*x[1]+(-0.240544184077)*x[2]**o+(-0.183028108529)*x[2]
        arg[3,0,2,1]=(0.354253383908)*x[0]**o+(0.628816528774)*x[0]+(0.588194429756)*x[1]**o+(-0.852128734457)*x[1]+(-0.84774778925)*x[2]**o+(0.796310123562)*x[2]
        arg[3,1,0,0]=(0.528880495806)*x[0]**o+(-0.74176589205)*x[0]+(-0.447816938371)*x[1]**o+(0.611297564802)*x[1]+(0.742134198482)*x[2]**o+(-0.625840664202)*x[2]
        arg[3,1,0,1]=(0.419633746259)*x[0]**o+(-0.247742274578)*x[0]+(0.302328885303)*x[1]**o+(-0.629854480502)*x[1]+(0.493919301288)*x[2]**o+(0.902910119017)*x[2]
        arg[3,1,1,0]=(0.509160481656)*x[0]**o+(-0.388129364133)*x[0]+(-0.595185230598)*x[1]**o+(-0.289592836865)*x[1]+(0.615663908134)*x[2]**o+(-0.560687711511)*x[2]
        arg[3,1,1,1]=(0.647782473568)*x[0]**o+(0.935840881725)*x[0]+(0.440348212287)*x[1]**o+(-0.489797988572)*x[1]+(0.766592631787)*x[2]**o+(0.217399649246)*x[2]
        arg[3,1,2,0]=(0.64944026906)*x[0]**o+(0.487409918475)*x[0]+(-0.0322210593784)*x[1]**o+(0.281311337861)*x[1]+(-0.694252407571)*x[2]**o+(-0.659678386121)*x[2]
        arg[3,1,2,1]=(0.806357278285)*x[0]**o+(0.419430374567)*x[0]+(-0.0525796600116)*x[1]**o+(-0.556886341914)*x[1]+(-0.288209814683)*x[2]**o+(-0.309696771495)*x[2]
        arg[3,2,0,0]=(-0.300806388527)*x[0]**o+(-0.39334416908)*x[0]+(0.000711617540648)*x[1]**o+(-0.618539958736)*x[1]+(0.0574098223152)*x[2]**o+(-0.351595384168)*x[2]
        arg[3,2,0,1]=(-0.898675793892)*x[0]**o+(0.904479958277)*x[0]+(0.0781945046018)*x[1]**o+(-0.0461053392586)*x[1]+(0.844798127031)*x[2]**o+(-0.82306838709)*x[2]
        arg[3,2,1,0]=(-0.268742335093)*x[0]**o+(-0.532639944208)*x[0]+(0.0101227109551)*x[1]**o+(-0.438312865518)*x[1]+(0.104288648246)*x[2]**o+(-0.115669027311)*x[2]
        arg[3,2,1,1]=(0.639871308792)*x[0]**o+(0.260847628514)*x[0]+(-0.455670719167)*x[1]**o+(0.210619035618)*x[1]+(-0.401809068119)*x[2]**o+(-0.0625194517026)*x[2]
        arg[3,2,2,0]=(-0.713684223829)*x[0]**o+(0.214281604766)*x[0]+(-0.60152592324)*x[1]**o+(0.249807228316)*x[1]+(-0.939395078261)*x[2]**o+(-0.519744396388)*x[2]
        arg[3,2,2,1]=(-0.340588993443)*x[0]**o+(-0.0165543854997)*x[0]+(-0.366479789338)*x[1]**o+(0.316347630987)*x[1]+(-0.457055798679)*x[2]**o+(-0.132368055553)*x[2]
        arg[3,3,0,0]=(-0.745211736715)*x[0]**o+(0.0106787556884)*x[0]+(-0.448180943446)*x[1]**o+(0.694982281891)*x[1]+(0.924431258221)*x[2]**o+(-0.841633806168)*x[2]
        arg[3,3,0,1]=(0.368054364362)*x[0]**o+(0.19288251992)*x[0]+(0.565854068035)*x[1]**o+(0.747534544754)*x[1]+(0.473165789569)*x[2]**o+(-0.59399986706)*x[2]
        arg[3,3,1,0]=(-0.142707671693)*x[0]**o+(0.177095043535)*x[0]+(-0.348946846847)*x[1]**o+(-0.545430013244)*x[1]+(-0.671405653136)*x[2]**o+(-0.531699969702)*x[2]
        arg[3,3,1,1]=(0.783201055631)*x[0]**o+(0.899495042323)*x[0]+(0.117802924059)*x[1]**o+(-0.277400900172)*x[1]+(-0.351268362305)*x[2]**o+(0.52756480897)*x[2]
        arg[3,3,2,0]=(-0.866758006698)*x[0]**o+(0.293139875238)*x[0]+(-0.318394231317)*x[1]**o+(-0.783701622124)*x[1]+(-0.95137399655)*x[2]**o+(-0.474198736639)*x[2]
        arg[3,3,2,1]=(0.0428398560619)*x[0]**o+(-0.970180601102)*x[0]+(-0.753897573393)*x[1]**o+(-0.0873352699251)*x[1]+(-0.131440532574)*x[2]**o+(0.875241605077)*x[2]
        arg[3,4,0,0]=(0.32399493643)*x[0]**o+(0.67295092524)*x[0]+(-0.29221110671)*x[1]**o+(0.118066079866)*x[1]+(-0.566541154667)*x[2]**o+(-0.806367499079)*x[2]
        arg[3,4,0,1]=(0.278776399755)*x[0]**o+(0.472327826919)*x[0]+(-0.956871861949)*x[1]**o+(0.109236874792)*x[1]+(-0.494095116883)*x[2]**o+(-0.194313775518)*x[2]
        arg[3,4,1,0]=(0.650920037415)*x[0]**o+(0.977698516929)*x[0]+(-0.0730777837998)*x[1]**o+(-0.576543152032)*x[1]+(0.781481934657)*x[2]**o+(-0.389328840554)*x[2]
        arg[3,4,1,1]=(-0.798341529522)*x[0]**o+(-0.630245234712)*x[0]+(-0.444550323489)*x[1]**o+(-0.539189248295)*x[1]+(-0.904964423673)*x[2]**o+(-0.611560462796)*x[2]
        arg[3,4,2,0]=(-0.392991417685)*x[0]**o+(0.313599061567)*x[0]+(-0.215747337398)*x[1]**o+(-0.310316872476)*x[1]+(-0.269452617785)*x[2]**o+(0.684353794645)*x[2]
        arg[3,4,2,1]=(-0.586668781403)*x[0]**o+(-0.0711284937177)*x[0]+(0.408273850302)*x[1]**o+(0.761860299288)*x[1]+(0.902079242449)*x[2]**o+(0.70161017722)*x[2]
        ref[0,0,0,0]=(1.43173970815)/(o+1.)+(-0.156396435558)
        ref[0,0,0,1]=(0.459165989186)/(o+1.)+(0.194399341676)
        ref[0,0,1,0]=(0.989528642545)/(o+1.)+(-0.164206368911)
        ref[0,0,1,1]=(1.37152584676)/(o+1.)+(-0.234493053498)
        ref[0,0,2,0]=(0.422741924108)/(o+1.)+(0.0721378026728)
        ref[0,0,2,1]=(-0.574467903086)/(o+1.)+(0.167204674462)
        ref[0,1,0,0]=(1.36619629735)/(o+1.)+(-0.19947081885)
        ref[0,1,0,1]=(-1.60758660729)/(o+1.)+(0.447728236186)
        ref[0,1,1,0]=(1.5106959073)/(o+1.)+(-0.393852180331)
        ref[0,1,1,1]=(1.85678121239)/(o+1.)+(-0.529280397551)
        ref[0,1,2,0]=(-0.796337866944)/(o+1.)+(0.0812355940636)
        ref[0,1,2,1]=(-0.333114331422)/(o+1.)+(1.22614026311)
        ref[0,2,0,0]=(0.516540083621)/(o+1.)+(-0.358761510118)
        ref[0,2,0,1]=(-2.72820411942)/(o+1.)+(0.15972705639)
        ref[0,2,1,0]=(-0.440149466937)/(o+1.)+(-0.124216123318)
        ref[0,2,1,1]=(-0.137940536755)/(o+1.)+(0.13992652657)
        ref[0,2,2,0]=(0.373759361058)/(o+1.)+(0.210996140796)
        ref[0,2,2,1]=(-1.07034621798)/(o+1.)+(1.2905860125)
        ref[0,3,0,0]=(0.923495049095)/(o+1.)+(-0.897043344396)
        ref[0,3,0,1]=(-0.20180799615)/(o+1.)+(0.272652457923)
        ref[0,3,1,0]=(-0.831052678484)/(o+1.)+(0.326929849662)
        ref[0,3,1,1]=(1.41523396695)/(o+1.)+(-0.5275927117)
        ref[0,3,2,0]=(0.405715363821)/(o+1.)+(0.535330516568)
        ref[0,3,2,1]=(0.877269821733)/(o+1.)+(-0.666082812121)
        ref[0,4,0,0]=(0.884126703786)/(o+1.)+(-0.559427998262)
        ref[0,4,0,1]=(-1.36933304826)/(o+1.)+(0.263100182211)
        ref[0,4,1,0]=(0.871486955825)/(o+1.)+(0.381595437312)
        ref[0,4,1,1]=(1.21446122052)/(o+1.)+(-0.642161706726)
        ref[0,4,2,0]=(-1.82861338966)/(o+1.)+(0.507433118531)
        ref[0,4,2,1]=(0.0642039966469)/(o+1.)+(0.551348562604)
        ref[1,0,0,0]=(0.40922616303)/(o+1.)+(-0.474630105482)
        ref[1,0,0,1]=(1.63943889768)/(o+1.)+(0.285622914296)
        ref[1,0,1,0]=(0.0845961115353)/(o+1.)+(0.143628502705)
        ref[1,0,1,1]=(0.62129375936)/(o+1.)+(0.671645835511)
        ref[1,0,2,0]=(-1.2004575453)/(o+1.)+(0.701757992696)
        ref[1,0,2,1]=(0.435963491825)/(o+1.)+(0.26710538393)
        ref[1,1,0,0]=(0.738921325407)/(o+1.)+(-0.0258542444502)
        ref[1,1,0,1]=(1.95599148183)/(o+1.)+(0.652818492883)
        ref[1,1,1,0]=(0.500576647095)/(o+1.)+(0.857652485995)
        ref[1,1,1,1]=(-0.0151750762821)/(o+1.)+(0.238827857598)
        ref[1,1,2,0]=(1.26963005373)/(o+1.)+(-0.299065008888)
        ref[1,1,2,1]=(-0.82289389301)/(o+1.)+(0.14939351077)
        ref[1,2,0,0]=(-0.495951362447)/(o+1.)+(-0.532833788304)
        ref[1,2,0,1]=(-1.87785443639)/(o+1.)+(-0.521209467732)
        ref[1,2,1,0]=(-0.374085962184)/(o+1.)+(0.505583421209)
        ref[1,2,1,1]=(1.01773556605)/(o+1.)+(0.764863466171)
        ref[1,2,2,0]=(-1.54049109838)/(o+1.)+(-0.511296952595)
        ref[1,2,2,1]=(-1.06948953348)/(o+1.)+(-0.215299794811)
        ref[1,3,0,0]=(-1.20198777829)/(o+1.)+(-0.274360734317)
        ref[1,3,0,1]=(1.13519842695)/(o+1.)+(0.506840913294)
        ref[1,3,1,0]=(0.973820116369)/(o+1.)+(0.485786110977)
        ref[1,3,1,1]=(0.155263308828)/(o+1.)+(0.473436850817)
        ref[1,3,2,0]=(0.476069733748)/(o+1.)+(-0.228046397052)
        ref[1,3,2,1]=(-0.234851551487)/(o+1.)+(-0.574065011567)
        ref[1,4,0,0]=(1.95924908786)/(o+1.)+(-0.493262726539)
        ref[1,4,0,1]=(-0.448532656782)/(o+1.)+(0.393812699074)
        ref[1,4,1,0]=(0.235155577808)/(o+1.)+(0.756260258803)
        ref[1,4,1,1]=(0.0334282473144)/(o+1.)+(-0.837406551217)
        ref[1,4,2,0]=(0.568132489684)/(o+1.)+(-0.13055465798)
        ref[1,4,2,1]=(-0.380042596089)/(o+1.)+(0.243737215624)
        ref[2,0,0,0]=(1.76568035823)/(o+1.)+(0.0789834372326)
        ref[2,0,0,1]=(-0.20023919191)/(o+1.)+(0.901466917208)
        ref[2,0,1,0]=(-1.20715877778)/(o+1.)+(-0.992500439486)
        ref[2,0,1,1]=(-0.761913514624)/(o+1.)+(0.465965334128)
        ref[2,0,2,0]=(-0.944104213428)/(o+1.)+(-0.146881270459)
        ref[2,0,2,1]=(0.105283971659)/(o+1.)+(-1.30498791712)
        ref[2,1,0,0]=(-0.804721214141)/(o+1.)+(0.356736186609)
        ref[2,1,0,1]=(-0.0358563948854)/(o+1.)+(-0.24146509617)
        ref[2,1,1,0]=(-0.437556173502)/(o+1.)+(-0.387215578175)
        ref[2,1,1,1]=(-0.0786030951662)/(o+1.)+(0.654838009465)
        ref[2,1,2,0]=(-0.859200342885)/(o+1.)+(-0.112407873885)
        ref[2,1,2,1]=(-0.296679106354)/(o+1.)+(0.0491751507381)
        ref[2,2,0,0]=(0.462408531841)/(o+1.)+(-0.557975521593)
        ref[2,2,0,1]=(-0.0634225791893)/(o+1.)+(0.0468185300489)
        ref[2,2,1,0]=(1.08205349937)/(o+1.)+(0.621257561187)
        ref[2,2,1,1]=(0.88864389409)/(o+1.)+(0.849886105205)
        ref[2,2,2,0]=(2.01702912587)/(o+1.)+(0.755558425167)
        ref[2,2,2,1]=(1.41263009001)/(o+1.)+(0.331281133791)
        ref[2,3,0,0]=(1.91485193421)/(o+1.)+(0.0643042152414)
        ref[2,3,0,1]=(1.86113720031)/(o+1.)+(0.870090339646)
        ref[2,3,1,0]=(0.442533807126)/(o+1.)+(-0.690437405116)
        ref[2,3,1,1]=(1.52774754257)/(o+1.)+(-0.236667038211)
        ref[2,3,2,0]=(1.08558789233)/(o+1.)+(-0.346580362392)
        ref[2,3,2,1]=(-0.771732912166)/(o+1.)+(-0.49697903252)
        ref[2,4,0,0]=(1.0510676744)/(o+1.)+(0.18646253745)
        ref[2,4,0,1]=(-0.878875991119)/(o+1.)+(-0.271299000644)
        ref[2,4,1,0]=(-0.75026757693)/(o+1.)+(-0.168751100981)
        ref[2,4,1,1]=(0.526155063499)/(o+1.)+(1.041034953)
        ref[2,4,2,0]=(0.377792294719)/(o+1.)+(0.433767213151)
        ref[2,4,2,1]=(0.114142925831)/(o+1.)+(-0.972051850709)
        ref[3,0,0,0]=(0.15538316839)/(o+1.)+(-0.861229788808)
        ref[3,0,0,1]=(-0.725501605569)/(o+1.)+(-0.248009661269)
        ref[3,0,1,0]=(0.791443284981)/(o+1.)+(0.618118504587)
        ref[3,0,1,1]=(-0.411749416988)/(o+1.)+(0.125302421245)
        ref[3,0,2,0]=(0.0968028463245)/(o+1.)+(0.592149023513)
        ref[3,0,2,1]=(0.0947000244134)/(o+1.)+(0.286498958939)
        ref[3,1,0,0]=(0.823197755916)/(o+1.)+(-0.378154495725)
        ref[3,1,0,1]=(1.21588193285)/(o+1.)+(0.0126566819682)
        ref[3,1,1,0]=(0.529639159192)/(o+1.)+(-0.619204956255)
        ref[3,1,1,1]=(1.85472331764)/(o+1.)+(0.3317212712)
        ref[3,1,2,0]=(-0.0770331978898)/(o+1.)+(0.0545214351076)
        ref[3,1,2,1]=(0.46556780359)/(o+1.)+(-0.223576369421)
        ref[3,2,0,0]=(-0.242684948671)/(o+1.)+(-0.681739755992)
        ref[3,2,0,1]=(0.0243168377401)/(o+1.)+(0.0176531159642)
        ref[3,2,1,0]=(-0.154330975892)/(o+1.)+(-0.543310918519)
        ref[3,2,1,1]=(-0.217608478494)/(o+1.)+(0.204473606215)
        ref[3,2,2,0]=(-2.25460522533)/(o+1.)+(-0.0278277816531)
        ref[3,2,2,1]=(-1.16412458146)/(o+1.)+(0.0837125949668)
        ref[3,3,0,0]=(-0.26896142194)/(o+1.)+(-0.0679863842941)
        ref[3,3,0,1]=(1.40707422197)/(o+1.)+(0.173208598807)
        ref[3,3,1,0]=(-1.16306017168)/(o+1.)+(-0.450017469706)
        ref[3,3,1,1]=(0.549735617385)/(o+1.)+(0.57482947556)
        ref[3,3,2,0]=(-2.13652623456)/(o+1.)+(-0.482380241762)
        ref[3,3,2,1]=(-0.842498249905)/(o+1.)+(-0.0911371329748)
        ref[3,4,0,0]=(-0.534757324947)/(o+1.)+(-0.00767524698684)
        ref[3,4,0,1]=(-1.17219057908)/(o+1.)+(0.193625463097)
        ref[3,4,1,0]=(1.35932418827)/(o+1.)+(0.00591326217154)
        ref[3,4,1,1]=(-2.14785627668)/(o+1.)+(-0.890497472901)
        ref[3,4,2,0]=(-0.878191372868)/(o+1.)+(0.343817991868)
        ref[3,4,2,1]=(0.723684311348)/(o+1.)+(0.696170991395)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.217470460227)*x[0]**o+(0.664132905811)*x[0]+(0.493878617743)*x[1]**o+(0.111327715714)*x[1]
        ref=(0.276408157516)/(o+1.)+(0.387730310763)
      else:
        arg=(0.371937838886)*x[0]**o+(0.653846807778)*x[0]+(0.172358058188)*x[1]**o+(-0.928017916614)*x[1]+(0.116156616174)*x[2]**o+(0.371332501889)*x[2]
        ref=(0.660452513249)/(o+1.)+(0.0485806965265)
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.329952046701)*x[0]**o+(0.12757022314)*x[0]+(0.198280362833)*x[1]**o+(0.129179854334)*x[1]
        arg[1]=(0.185895161332)*x[0]**o+(0.270740935936)*x[0]+(0.630122833594)*x[1]**o+(-0.0696403698787)*x[1]
        ref[0]=(-0.131671683869)/(o+1.)+(0.128375038737)
        ref[1]=(0.816017994925)/(o+1.)+(0.100550283029)
      else:
        arg[0]=(0.526257699936)*x[0]**o+(0.0731442796396)*x[0]+(-0.266121642119)*x[1]**o+(-0.929478242136)*x[1]+(0.743092861949)*x[2]**o+(0.22690419858)*x[2]
        arg[1]=(-0.121387039147)*x[0]**o+(-0.0580608263784)*x[0]+(0.761164059949)*x[1]**o+(0.821938972431)*x[1]+(-0.756423175962)*x[2]**o+(0.213207132353)*x[2]
        ref[0]=(1.00322891977)/(o+1.)+(-0.314714881958)
        ref[1]=(-0.11664615516)/(o+1.)+(0.488542639203)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.963318847499)*x[0]**o+(-0.184977238308)*x[0]+(0.153633231879)*x[1]**o+(-0.640028315216)*x[1]
        arg[0,1]=(0.0879103656102)*x[0]**o+(-0.928854425726)*x[0]+(-0.138566176095)*x[1]**o+(-0.670118398936)*x[1]
        arg[0,2]=(0.920856326204)*x[0]**o+(0.461047594595)*x[0]+(-0.914778320363)*x[1]**o+(-0.561644217568)*x[1]
        arg[0,3]=(0.506148289443)*x[0]**o+(-0.474044343536)*x[0]+(0.742127575836)*x[1]**o+(-0.52349314159)*x[1]
        arg[0,4]=(0.0810224543145)*x[0]**o+(-0.307992483498)*x[0]+(0.749673245291)*x[1]**o+(-0.554099873044)*x[1]
        arg[1,0]=(-0.270897099553)*x[0]**o+(-0.222178461886)*x[0]+(0.755729049844)*x[1]**o+(0.740743777894)*x[1]
        arg[1,1]=(0.0220934176819)*x[0]**o+(-0.309641658323)*x[0]+(-0.951833592348)*x[1]**o+(0.430554526918)*x[1]
        arg[1,2]=(0.176033311218)*x[0]**o+(0.728188001962)*x[0]+(-0.428357954006)*x[1]**o+(0.955059500699)*x[1]
        arg[1,3]=(0.946680422474)*x[0]**o+(-0.251968148743)*x[0]+(-0.3704600461)*x[1]**o+(0.692080922297)*x[1]
        arg[1,4]=(0.122655375591)*x[0]**o+(0.617285886533)*x[0]+(0.729184247986)*x[1]**o+(0.30369620264)*x[1]
        arg[2,0]=(-0.957780501271)*x[0]**o+(0.438290429468)*x[0]+(-0.504841431279)*x[1]**o+(0.935350834608)*x[1]
        arg[2,1]=(-0.49410765407)*x[0]**o+(-0.635142349002)*x[0]+(0.700131616572)*x[1]**o+(0.99523209861)*x[1]
        arg[2,2]=(0.0967686936303)*x[0]**o+(0.987754073381)*x[0]+(-0.161220173386)*x[1]**o+(-0.343926097784)*x[1]
        arg[2,3]=(0.560040886628)*x[0]**o+(-0.843778720323)*x[0]+(0.345354034369)*x[1]**o+(0.440584543858)*x[1]
        arg[2,4]=(0.988977815245)*x[0]**o+(0.661922598357)*x[0]+(-0.425421002549)*x[1]**o+(0.624279872655)*x[1]
        arg[3,0]=(-0.997217068098)*x[0]**o+(-0.973497091595)*x[0]+(-0.0980683851598)*x[1]**o+(0.408081557041)*x[1]
        arg[3,1]=(0.0827616522414)*x[0]**o+(0.931022772362)*x[0]+(-0.588688206358)*x[1]**o+(0.900467348992)*x[1]
        arg[3,2]=(0.0933642874383)*x[0]**o+(0.398732664158)*x[0]+(-0.437102212783)*x[1]**o+(-0.126538302466)*x[1]
        arg[3,3]=(0.251989644486)*x[0]**o+(0.765809815972)*x[0]+(-0.325455283605)*x[1]**o+(0.171819372538)*x[1]
        arg[3,4]=(0.649407682145)*x[0]**o+(0.963776897105)*x[0]+(0.671915752131)*x[1]**o+(-0.468729629153)*x[1]
        ref[0,0]=(-0.80968561562)/(o+1.)+(-0.412502776762)
        ref[0,1]=(-0.0506558104848)/(o+1.)+(-0.799486412331)
        ref[0,2]=(0.00607800584081)/(o+1.)+(-0.0502983114867)
        ref[0,3]=(1.24827586528)/(o+1.)+(-0.498768742563)
        ref[0,4]=(0.830695699605)/(o+1.)+(-0.431046178271)
        ref[1,0]=(0.484831950292)/(o+1.)+(0.259282658004)
        ref[1,1]=(-0.929740174666)/(o+1.)+(0.0604564342974)
        ref[1,2]=(-0.252324642788)/(o+1.)+(0.84162375133)
        ref[1,3]=(0.576220376373)/(o+1.)+(0.220056386777)
        ref[1,4]=(0.851839623577)/(o+1.)+(0.460491044586)
        ref[2,0]=(-1.46262193255)/(o+1.)+(0.686820632038)
        ref[2,1]=(0.206023962501)/(o+1.)+(0.180044874804)
        ref[2,2]=(-0.0644514797559)/(o+1.)+(0.321913987798)
        ref[2,3]=(0.905394920997)/(o+1.)+(-0.201597088233)
        ref[2,4]=(0.563556812695)/(o+1.)+(0.643101235506)
        ref[3,0]=(-1.09528545326)/(o+1.)+(-0.282707767277)
        ref[3,1]=(-0.505926554117)/(o+1.)+(0.915745060677)
        ref[3,2]=(-0.343737925345)/(o+1.)+(0.136097180846)
        ref[3,3]=(-0.0734656391194)/(o+1.)+(0.468814594255)
        ref[3,4]=(1.32132343428)/(o+1.)+(0.247523633976)
      else:
        arg[0,0]=(-0.512698606017)*x[0]**o+(-0.156853163498)*x[0]+(0.247997355633)*x[1]**o+(-0.363003068904)*x[1]+(-0.940097989108)*x[2]**o+(-0.661063850288)*x[2]
        arg[0,1]=(-0.886390181577)*x[0]**o+(0.724049511142)*x[0]+(0.260684592163)*x[1]**o+(0.265648885034)*x[1]+(0.123084281693)*x[2]**o+(0.181492979151)*x[2]
        arg[0,2]=(-0.570543998851)*x[0]**o+(0.0541041165388)*x[0]+(0.630994990544)*x[1]**o+(-0.89311079559)*x[1]+(0.750714073617)*x[2]**o+(-0.0684796380527)*x[2]
        arg[0,3]=(0.605486031268)*x[0]**o+(0.0387617487672)*x[0]+(-0.365027101406)*x[1]**o+(0.323275180514)*x[1]+(-0.763142660003)*x[2]**o+(-0.331069480921)*x[2]
        arg[0,4]=(-0.0779035848472)*x[0]**o+(-0.0234904624931)*x[0]+(0.591635786678)*x[1]**o+(-0.793680126332)*x[1]+(-0.874742296757)*x[2]**o+(-0.184528885157)*x[2]
        arg[1,0]=(-0.62480972544)*x[0]**o+(-0.99242215547)*x[0]+(0.634250679117)*x[1]**o+(-0.785448642692)*x[1]+(-0.013335654402)*x[2]**o+(0.52112750417)*x[2]
        arg[1,1]=(0.611541580179)*x[0]**o+(-0.203144044802)*x[0]+(-0.411277184995)*x[1]**o+(0.373487145939)*x[1]+(-0.679339345026)*x[2]**o+(-0.632413536218)*x[2]
        arg[1,2]=(-0.258927016205)*x[0]**o+(0.0242628959926)*x[0]+(-0.596909760383)*x[1]**o+(-0.295787354787)*x[1]+(-0.337930022823)*x[2]**o+(-0.825085100446)*x[2]
        arg[1,3]=(-0.0976663781507)*x[0]**o+(0.716600132838)*x[0]+(-0.682155341898)*x[1]**o+(-0.154484864167)*x[1]+(0.754686874908)*x[2]**o+(-0.102453470268)*x[2]
        arg[1,4]=(0.268714339964)*x[0]**o+(0.974234398844)*x[0]+(-0.370113158384)*x[1]**o+(0.0745228956861)*x[1]+(-0.0674370676047)*x[2]**o+(0.132605448278)*x[2]
        arg[2,0]=(-0.00750893262562)*x[0]**o+(0.507641887134)*x[0]+(0.638483932117)*x[1]**o+(-0.940295540912)*x[1]+(-0.201492519698)*x[2]**o+(0.504878595818)*x[2]
        arg[2,1]=(-0.0283522980593)*x[0]**o+(-0.76577346605)*x[0]+(0.0863472830598)*x[1]**o+(-0.904447859187)*x[1]+(-0.203390005492)*x[2]**o+(-0.741629136645)*x[2]
        arg[2,2]=(-0.0196123831667)*x[0]**o+(-0.320721372799)*x[0]+(0.0670421287749)*x[1]**o+(-0.415029682416)*x[1]+(0.450269143475)*x[2]**o+(-0.735201210502)*x[2]
        arg[2,3]=(-0.621537330758)*x[0]**o+(0.21573881897)*x[0]+(0.0364275955835)*x[1]**o+(0.873908952751)*x[1]+(-0.197046790303)*x[2]**o+(0.655654722165)*x[2]
        arg[2,4]=(0.432531007791)*x[0]**o+(-0.65605251166)*x[0]+(0.728245204833)*x[1]**o+(-0.769998936989)*x[1]+(-0.797879623133)*x[2]**o+(0.36575554332)*x[2]
        arg[3,0]=(0.512219734043)*x[0]**o+(-0.177942735603)*x[0]+(-0.307293350237)*x[1]**o+(-0.780181541911)*x[1]+(-0.510201727643)*x[2]**o+(-0.567980715525)*x[2]
        arg[3,1]=(-0.405123989366)*x[0]**o+(-0.133316566583)*x[0]+(0.552868689995)*x[1]**o+(-0.921149434884)*x[1]+(0.510851078054)*x[2]**o+(0.7207533866)*x[2]
        arg[3,2]=(0.150011788901)*x[0]**o+(-0.55947555552)*x[0]+(-0.353366762442)*x[1]**o+(0.204954197301)*x[1]+(-0.48254632377)*x[2]**o+(-0.272738900127)*x[2]
        arg[3,3]=(-0.690981207461)*x[0]**o+(0.38683432128)*x[0]+(0.655731808738)*x[1]**o+(0.162269870301)*x[1]+(-0.373468659959)*x[2]**o+(-0.841431521009)*x[2]
        arg[3,4]=(0.156455675319)*x[0]**o+(0.501690084272)*x[0]+(-0.339255791314)*x[1]**o+(0.0832260784264)*x[1]+(0.394137401365)*x[2]**o+(-0.92403906326)*x[2]
        ref[0,0]=(-1.20479923949)/(o+1.)+(-0.590460041345)
        ref[0,1]=(-0.502621307721)/(o+1.)+(0.585595687663)
        ref[0,2]=(0.811165065311)/(o+1.)+(-0.453743158552)
        ref[0,3]=(-0.522683730141)/(o+1.)+(0.0154837241797)
        ref[0,4]=(-0.361010094926)/(o+1.)+(-0.500849736991)
        ref[1,0]=(-0.00389470072479)/(o+1.)+(-0.628371646996)
        ref[1,1]=(-0.479074949842)/(o+1.)+(-0.23103521754)
        ref[1,2]=(-1.19376679941)/(o+1.)+(-0.54830477962)
        ref[1,3]=(-0.0251348451402)/(o+1.)+(0.229830899202)
        ref[1,4]=(-0.168835886025)/(o+1.)+(0.590681371404)
        ref[2,0]=(0.429482479793)/(o+1.)+(0.0361124710203)
        ref[2,1]=(-0.145395020492)/(o+1.)+(-1.20592523094)
        ref[2,2]=(0.497698889084)/(o+1.)+(-0.735476132859)
        ref[2,3]=(-0.782156525477)/(o+1.)+(0.872651246943)
        ref[2,4]=(0.362896589491)/(o+1.)+(-0.530147952664)
        ref[3,0]=(-0.305275343837)/(o+1.)+(-0.763052496519)
        ref[3,1]=(0.658595778683)/(o+1.)+(-0.166856307434)
        ref[3,2]=(-0.685901297311)/(o+1.)+(-0.313630129173)
        ref[3,3]=(-0.408718058683)/(o+1.)+(-0.146163664714)
        ref[3,4]=(0.21133728537)/(o+1.)+(-0.169561450281)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.165423911829)*x[0]**o+(0.912542062064)*x[0]+(0.515840721225)*x[1]**o+(-0.855922021017)*x[1]
        arg[0,0,1]=(0.591873802546)*x[0]**o+(-0.039385167187)*x[0]+(0.0261484884274)*x[1]**o+(0.227950240434)*x[1]
        arg[0,1,0]=(-0.825833712046)*x[0]**o+(0.989606797765)*x[0]+(0.975260558791)*x[1]**o+(0.297963603984)*x[1]
        arg[0,1,1]=(-0.658054502482)*x[0]**o+(-0.865505090963)*x[0]+(0.817184084453)*x[1]**o+(-0.409588098899)*x[1]
        arg[1,0,0]=(0.538456987198)*x[0]**o+(-0.0768787509391)*x[0]+(0.424620431835)*x[1]**o+(0.799917948831)*x[1]
        arg[1,0,1]=(0.125945558199)*x[0]**o+(-0.48320049996)*x[0]+(0.105474962327)*x[1]**o+(-0.090713555582)*x[1]
        arg[1,1,0]=(0.835213375418)*x[0]**o+(-0.513407038769)*x[0]+(-0.342890811735)*x[1]**o+(-0.285057736244)*x[1]
        arg[1,1,1]=(0.762782085749)*x[0]**o+(-0.308457087352)*x[0]+(0.23756726753)*x[1]**o+(-0.909365956504)*x[1]
        arg[2,0,0]=(-0.76467980678)*x[0]**o+(-0.265552922006)*x[0]+(-0.261862355539)*x[1]**o+(-0.399407404618)*x[1]
        arg[2,0,1]=(0.79401193617)*x[0]**o+(-0.0300461495216)*x[0]+(-0.978834133508)*x[1]**o+(-0.984845382599)*x[1]
        arg[2,1,0]=(0.898033018534)*x[0]**o+(-0.184513430159)*x[0]+(0.0430634098526)*x[1]**o+(0.62213313954)*x[1]
        arg[2,1,1]=(0.126926439246)*x[0]**o+(0.671259428917)*x[0]+(-0.396062855417)*x[1]**o+(0.0461825691916)*x[1]
        arg[3,0,0]=(0.185756594279)*x[0]**o+(0.570946291973)*x[0]+(0.0673731944956)*x[1]**o+(0.367766610292)*x[1]
        arg[3,0,1]=(0.021269224946)*x[0]**o+(-0.463663279063)*x[0]+(-0.687930988068)*x[1]**o+(0.184714204849)*x[1]
        arg[3,1,0]=(0.754288639303)*x[0]**o+(-0.94612827517)*x[0]+(0.802463196709)*x[1]**o+(0.202355582077)*x[1]
        arg[3,1,1]=(0.840002857786)*x[0]**o+(-0.692207739447)*x[0]+(0.486569391181)*x[1]**o+(-0.713960793192)*x[1]
        arg[4,0,0]=(-0.205794539655)*x[0]**o+(-0.930458992533)*x[0]+(-0.62580205172)*x[1]**o+(-0.884738110673)*x[1]
        arg[4,0,1]=(-0.91433612349)*x[0]**o+(0.694961527804)*x[0]+(-0.510941334746)*x[1]**o+(-0.510725271706)*x[1]
        arg[4,1,0]=(0.852785499366)*x[0]**o+(0.870826628851)*x[0]+(-0.577529754198)*x[1]**o+(-0.583793684857)*x[1]
        arg[4,1,1]=(-0.604869226881)*x[0]**o+(0.482463572402)*x[0]+(0.829654316885)*x[1]**o+(0.597543822455)*x[1]
        arg[5,0,0]=(0.413459469749)*x[0]**o+(0.865143264083)*x[0]+(0.986197644249)*x[1]**o+(-0.2511021346)*x[1]
        arg[5,0,1]=(0.811572168213)*x[0]**o+(-0.110121334437)*x[0]+(-0.985492232242)*x[1]**o+(-0.913436084156)*x[1]
        arg[5,1,0]=(-0.660869051422)*x[0]**o+(-0.743235620427)*x[0]+(0.205393778238)*x[1]**o+(0.296069404687)*x[1]
        arg[5,1,1]=(-0.13285011267)*x[0]**o+(0.659400366834)*x[0]+(-0.889984368942)*x[1]**o+(0.584893563559)*x[1]
        ref[0,0,0]=(0.350416809396)/(o+1.)+(0.0283100205234)
        ref[0,0,1]=(0.618022290973)/(o+1.)+(0.0942825366236)
        ref[0,1,0]=(0.149426846745)/(o+1.)+(0.643785200874)
        ref[0,1,1]=(0.159129581971)/(o+1.)+(-0.637546594931)
        ref[1,0,0]=(0.963077419032)/(o+1.)+(0.361519598946)
        ref[1,0,1]=(0.231420520526)/(o+1.)+(-0.286957027771)
        ref[1,1,0]=(0.492322563683)/(o+1.)+(-0.399232387506)
        ref[1,1,1]=(1.00034935328)/(o+1.)+(-0.608911521928)
        ref[2,0,0]=(-1.02654216232)/(o+1.)+(-0.332480163312)
        ref[2,0,1]=(-0.184822197338)/(o+1.)+(-0.507445766061)
        ref[2,1,0]=(0.941096428387)/(o+1.)+(0.21880985469)
        ref[2,1,1]=(-0.269136416172)/(o+1.)+(0.358720999054)
        ref[3,0,0]=(0.253129788774)/(o+1.)+(0.469356451133)
        ref[3,0,1]=(-0.666661763122)/(o+1.)+(-0.139474537107)
        ref[3,1,0]=(1.55675183601)/(o+1.)+(-0.371886346546)
        ref[3,1,1]=(1.32657224897)/(o+1.)+(-0.703084266319)
        ref[4,0,0]=(-0.831596591375)/(o+1.)+(-0.907598551603)
        ref[4,0,1]=(-1.42527745824)/(o+1.)+(0.0921181280487)
        ref[4,1,0]=(0.275255745168)/(o+1.)+(0.143516471997)
        ref[4,1,1]=(0.224785090004)/(o+1.)+(0.540003697429)
        ref[5,0,0]=(1.399657114)/(o+1.)+(0.307020564742)
        ref[5,0,1]=(-0.173920064028)/(o+1.)+(-0.511778709296)
        ref[5,1,0]=(-0.455475273184)/(o+1.)+(-0.22358310787)
        ref[5,1,1]=(-1.02283448161)/(o+1.)+(0.622146965196)
      else:
        arg[0,0,0]=(-0.564478749422)*x[0]**o+(-0.0357223504179)*x[0]+(0.396603298865)*x[1]**o+(-0.461941327703)*x[1]+(-0.667580657928)*x[2]**o+(-0.589957067263)*x[2]
        arg[0,0,1]=(0.930616008951)*x[0]**o+(-0.764067565635)*x[0]+(-0.71350617205)*x[1]**o+(-0.530942701898)*x[1]+(-0.103797356434)*x[2]**o+(0.410704036673)*x[2]
        arg[0,1,0]=(-0.59127392299)*x[0]**o+(0.452490122131)*x[0]+(0.714793625631)*x[1]**o+(-0.933114051133)*x[1]+(-0.693437840325)*x[2]**o+(0.979267231764)*x[2]
        arg[0,1,1]=(0.179682332267)*x[0]**o+(-0.419960476425)*x[0]+(-0.10986598341)*x[1]**o+(0.912986428497)*x[1]+(0.0886946782573)*x[2]**o+(0.516065958873)*x[2]
        arg[1,0,0]=(0.524446863481)*x[0]**o+(-0.439668193423)*x[0]+(0.681259811488)*x[1]**o+(0.439275262258)*x[1]+(0.84573881706)*x[2]**o+(0.377640553175)*x[2]
        arg[1,0,1]=(-0.007464905203)*x[0]**o+(0.234727832471)*x[0]+(0.0428956356618)*x[1]**o+(0.195772930172)*x[1]+(0.457790934614)*x[2]**o+(0.445309166495)*x[2]
        arg[1,1,0]=(0.821283375827)*x[0]**o+(0.212480005798)*x[0]+(0.148817990961)*x[1]**o+(0.119716150877)*x[1]+(0.658959126865)*x[2]**o+(-0.346216916131)*x[2]
        arg[1,1,1]=(-0.0587222422894)*x[0]**o+(-0.325321661973)*x[0]+(0.149293818689)*x[1]**o+(-0.632073115037)*x[1]+(0.018604271874)*x[2]**o+(-0.412288240166)*x[2]
        arg[2,0,0]=(-0.22588082169)*x[0]**o+(0.197780461217)*x[0]+(0.387336989213)*x[1]**o+(-0.680021661016)*x[1]+(-0.903782254346)*x[2]**o+(-0.273268109853)*x[2]
        arg[2,0,1]=(0.415876394297)*x[0]**o+(0.199596795904)*x[0]+(-0.445817134682)*x[1]**o+(0.672539590684)*x[1]+(0.907825473964)*x[2]**o+(0.459560469691)*x[2]
        arg[2,1,0]=(-0.052146887134)*x[0]**o+(-0.229450574738)*x[0]+(0.0929720151604)*x[1]**o+(-0.893861034491)*x[1]+(-0.294030328982)*x[2]**o+(-0.193256779109)*x[2]
        arg[2,1,1]=(0.361171422066)*x[0]**o+(-0.212022056495)*x[0]+(0.0672221746963)*x[1]**o+(0.57506871515)*x[1]+(0.654894469509)*x[2]**o+(0.0749777260533)*x[2]
        arg[3,0,0]=(0.505355253404)*x[0]**o+(-0.0753563586359)*x[0]+(0.544299279727)*x[1]**o+(-0.781954131328)*x[1]+(0.534390713891)*x[2]**o+(-0.719289344378)*x[2]
        arg[3,0,1]=(-0.589309652827)*x[0]**o+(-0.0557421151178)*x[0]+(-0.275311076837)*x[1]**o+(0.998180518149)*x[1]+(-0.844823856433)*x[2]**o+(-0.988843805162)*x[2]
        arg[3,1,0]=(-0.411075104301)*x[0]**o+(-0.405896828894)*x[0]+(0.242301856034)*x[1]**o+(-0.760648969984)*x[1]+(0.0940367373673)*x[2]**o+(-0.121420230853)*x[2]
        arg[3,1,1]=(-0.201426146967)*x[0]**o+(0.722847134095)*x[0]+(-0.00743580369433)*x[1]**o+(-0.593158423904)*x[1]+(0.630575332064)*x[2]**o+(-0.309839791343)*x[2]
        arg[4,0,0]=(0.370057366738)*x[0]**o+(0.35683022394)*x[0]+(0.886304520049)*x[1]**o+(0.834488899465)*x[1]+(0.27408661832)*x[2]**o+(0.325007409933)*x[2]
        arg[4,0,1]=(0.386153377977)*x[0]**o+(-0.889700309398)*x[0]+(-0.60695970137)*x[1]**o+(0.701208035732)*x[1]+(0.435164510485)*x[2]**o+(0.40223148148)*x[2]
        arg[4,1,0]=(-0.227256041906)*x[0]**o+(-0.0406115125763)*x[0]+(-0.0298974314183)*x[1]**o+(0.887167464033)*x[1]+(0.1004751563)*x[2]**o+(0.826924545719)*x[2]
        arg[4,1,1]=(0.733575640137)*x[0]**o+(-0.376269474248)*x[0]+(-0.841290711112)*x[1]**o+(-0.0633013917179)*x[1]+(0.335730636406)*x[2]**o+(0.913607658221)*x[2]
        arg[5,0,0]=(-0.896019254334)*x[0]**o+(0.465470987762)*x[0]+(0.338543664838)*x[1]**o+(0.182102591196)*x[1]+(0.476317455739)*x[2]**o+(0.191367623163)*x[2]
        arg[5,0,1]=(0.828458939676)*x[0]**o+(-0.37853033744)*x[0]+(-0.540809170906)*x[1]**o+(0.367869598038)*x[1]+(0.321196361443)*x[2]**o+(0.548981819769)*x[2]
        arg[5,1,0]=(-0.0805871335579)*x[0]**o+(-0.554679375424)*x[0]+(-0.259438414137)*x[1]**o+(0.72362534417)*x[1]+(-0.838855028889)*x[2]**o+(0.929802758377)*x[2]
        arg[5,1,1]=(-0.947910259125)*x[0]**o+(0.654012824998)*x[0]+(0.425232602078)*x[1]**o+(-0.649345176159)*x[1]+(-0.946280438348)*x[2]**o+(-0.429547284912)*x[2]
        ref[0,0,0]=(-0.835456108484)/(o+1.)+(-0.543810372692)
        ref[0,0,1]=(0.113312480467)/(o+1.)+(-0.44215311543)
        ref[0,1,0]=(-0.569918137684)/(o+1.)+(0.249321651381)
        ref[0,1,1]=(0.158511027114)/(o+1.)+(0.504545955472)
        ref[1,0,0]=(2.05144549203)/(o+1.)+(0.188623811006)
        ref[1,0,1]=(0.493221665073)/(o+1.)+(0.437904964569)
        ref[1,1,0]=(1.62906049365)/(o+1.)+(-0.00701037972825)
        ref[1,1,1]=(0.109175848273)/(o+1.)+(-0.684841508588)
        ref[2,0,0]=(-0.742326086822)/(o+1.)+(-0.377754654826)
        ref[2,0,1]=(0.877884733578)/(o+1.)+(0.665848428139)
        ref[2,1,0]=(-0.253205200956)/(o+1.)+(-0.658284194169)
        ref[2,1,1]=(1.08328806627)/(o+1.)+(0.219012192354)
        ref[3,0,0]=(1.58404524702)/(o+1.)+(-0.788299917171)
        ref[3,0,1]=(-1.7094445861)/(o+1.)+(-0.0232027010654)
        ref[3,1,0]=(-0.0747365109002)/(o+1.)+(-0.643983014866)
        ref[3,1,1]=(0.421713381403)/(o+1.)+(-0.0900755405762)
        ref[4,0,0]=(1.53044850511)/(o+1.)+(0.758163266669)
        ref[4,0,1]=(0.214358187092)/(o+1.)+(0.106869603907)
        ref[4,1,0]=(-0.156678317024)/(o+1.)+(0.836740248588)
        ref[4,1,1]=(0.228015565432)/(o+1.)+(0.237018396127)
        ref[5,0,0]=(-0.0811581337568)/(o+1.)+(0.419470601061)
        ref[5,0,1]=(0.608846130213)/(o+1.)+(0.269160540183)
        ref[5,1,0]=(-1.17888057658)/(o+1.)+(0.549374363561)
        ref[5,1,1]=(-1.4689580954)/(o+1.)+(-0.212439818036)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.216014106349)*x[0]**o+(0.98826230471)*x[0]+(0.522566177319)*x[1]**o+(-0.672520997452)*x[1]
        arg[0,0,0,1]=(0.368024408084)*x[0]**o+(0.047972971332)*x[0]+(-0.307374327321)*x[1]**o+(0.0745169735457)*x[1]
        arg[0,0,1,0]=(0.438735045452)*x[0]**o+(0.76355958752)*x[0]+(0.200829025437)*x[1]**o+(0.56912407759)*x[1]
        arg[0,0,1,1]=(-0.0328769393712)*x[0]**o+(-0.39701363611)*x[0]+(0.704125532519)*x[1]**o+(-0.6602435084)*x[1]
        arg[0,0,2,0]=(-0.547717992717)*x[0]**o+(-0.844256763776)*x[0]+(0.621300470147)*x[1]**o+(-0.105216181226)*x[1]
        arg[0,0,2,1]=(-0.723474063234)*x[0]**o+(-0.980738972901)*x[0]+(-0.183578659566)*x[1]**o+(0.586038450047)*x[1]
        arg[0,1,0,0]=(0.505379907001)*x[0]**o+(-0.470514711277)*x[0]+(0.799116811747)*x[1]**o+(0.682760846386)*x[1]
        arg[0,1,0,1]=(0.353992609198)*x[0]**o+(-0.848107655585)*x[0]+(-0.774065459855)*x[1]**o+(0.0323137462292)*x[1]
        arg[0,1,1,0]=(0.799068805694)*x[0]**o+(-0.979761821209)*x[0]+(0.642674987867)*x[1]**o+(-0.639029952044)*x[1]
        arg[0,1,1,1]=(0.542556683327)*x[0]**o+(0.876753086356)*x[0]+(-0.387121297643)*x[1]**o+(0.377543179628)*x[1]
        arg[0,1,2,0]=(0.424672899596)*x[0]**o+(-0.134184751569)*x[0]+(-0.561877304114)*x[1]**o+(0.549933699347)*x[1]
        arg[0,1,2,1]=(-0.126173294501)*x[0]**o+(-0.338964866844)*x[0]+(-0.163536115759)*x[1]**o+(0.747551558186)*x[1]
        arg[0,2,0,0]=(0.304334183452)*x[0]**o+(0.927809971187)*x[0]+(0.961504942988)*x[1]**o+(0.0698014252396)*x[1]
        arg[0,2,0,1]=(0.488820088455)*x[0]**o+(0.890735247208)*x[0]+(0.686782901398)*x[1]**o+(0.972954994564)*x[1]
        arg[0,2,1,0]=(-0.737577792766)*x[0]**o+(-0.351698913646)*x[0]+(0.392591489865)*x[1]**o+(0.79238240143)*x[1]
        arg[0,2,1,1]=(-0.997254027544)*x[0]**o+(0.579800166465)*x[0]+(-0.95000690643)*x[1]**o+(-0.166463083476)*x[1]
        arg[0,2,2,0]=(-0.999703749644)*x[0]**o+(0.0415211466468)*x[0]+(0.0816160245139)*x[1]**o+(-0.0921114587195)*x[1]
        arg[0,2,2,1]=(-0.172818721386)*x[0]**o+(0.10475692446)*x[0]+(-0.855815339933)*x[1]**o+(-0.50842580034)*x[1]
        arg[0,3,0,0]=(-0.154563121513)*x[0]**o+(-0.0956305513369)*x[0]+(-0.893783521038)*x[1]**o+(0.10772474638)*x[1]
        arg[0,3,0,1]=(0.365331678349)*x[0]**o+(0.403124820832)*x[0]+(-0.0750500521493)*x[1]**o+(0.62559762808)*x[1]
        arg[0,3,1,0]=(-0.42235875521)*x[0]**o+(0.831133238533)*x[0]+(-0.893322944982)*x[1]**o+(0.936074096319)*x[1]
        arg[0,3,1,1]=(0.587382334689)*x[0]**o+(0.996190710631)*x[0]+(-0.796032791654)*x[1]**o+(0.783491453801)*x[1]
        arg[0,3,2,0]=(-0.369527027508)*x[0]**o+(-0.501417552089)*x[0]+(-0.372105439045)*x[1]**o+(0.653283000152)*x[1]
        arg[0,3,2,1]=(0.914893750314)*x[0]**o+(-0.769183323793)*x[0]+(0.571638486717)*x[1]**o+(-0.939784652259)*x[1]
        arg[0,4,0,0]=(0.134859060251)*x[0]**o+(-0.297363165275)*x[0]+(0.198957028974)*x[1]**o+(-0.828531725127)*x[1]
        arg[0,4,0,1]=(-0.371375370217)*x[0]**o+(-0.0422058718402)*x[0]+(0.502010302278)*x[1]**o+(0.442051802597)*x[1]
        arg[0,4,1,0]=(0.513346571173)*x[0]**o+(0.182640867356)*x[0]+(0.76200283664)*x[1]**o+(0.134640432361)*x[1]
        arg[0,4,1,1]=(-0.815543712756)*x[0]**o+(0.0306226823724)*x[0]+(-0.0211231926242)*x[1]**o+(-0.718800228294)*x[1]
        arg[0,4,2,0]=(-0.980596107555)*x[0]**o+(-0.714237107759)*x[0]+(0.547986215851)*x[1]**o+(0.180131474534)*x[1]
        arg[0,4,2,1]=(-0.460780402239)*x[0]**o+(-0.483771113997)*x[0]+(-0.017140263255)*x[1]**o+(-0.127947184627)*x[1]
        arg[1,0,0,0]=(-0.0187443155707)*x[0]**o+(-0.490914415873)*x[0]+(-0.720022496093)*x[1]**o+(0.0265633770733)*x[1]
        arg[1,0,0,1]=(-0.634855765904)*x[0]**o+(0.345031954366)*x[0]+(-0.0532392363316)*x[1]**o+(0.983295524564)*x[1]
        arg[1,0,1,0]=(0.262182040476)*x[0]**o+(-0.883857758733)*x[0]+(0.880670543535)*x[1]**o+(0.193477695378)*x[1]
        arg[1,0,1,1]=(-0.0825976168371)*x[0]**o+(0.161838720337)*x[0]+(-0.0110587499089)*x[1]**o+(-0.634455402489)*x[1]
        arg[1,0,2,0]=(0.284221958342)*x[0]**o+(-0.948548729029)*x[0]+(0.430914344458)*x[1]**o+(-0.18563638241)*x[1]
        arg[1,0,2,1]=(-0.459558022898)*x[0]**o+(0.431907572866)*x[0]+(-0.0597669070218)*x[1]**o+(0.0694271372225)*x[1]
        arg[1,1,0,0]=(0.54071078447)*x[0]**o+(-0.244570994356)*x[0]+(0.00506219762373)*x[1]**o+(-0.0816378019186)*x[1]
        arg[1,1,0,1]=(-0.100752221379)*x[0]**o+(0.535738345566)*x[0]+(0.730325284553)*x[1]**o+(-0.885044748717)*x[1]
        arg[1,1,1,0]=(-0.0422460353138)*x[0]**o+(-0.986416505995)*x[0]+(-0.111766776997)*x[1]**o+(-0.842179226197)*x[1]
        arg[1,1,1,1]=(0.0690281593169)*x[0]**o+(-0.885054188663)*x[0]+(-0.680349314907)*x[1]**o+(0.00277190934345)*x[1]
        arg[1,1,2,0]=(0.502632146841)*x[0]**o+(-0.197332151163)*x[0]+(-0.0942257361149)*x[1]**o+(0.045171692749)*x[1]
        arg[1,1,2,1]=(0.427910466818)*x[0]**o+(-0.61355947216)*x[0]+(0.465128040517)*x[1]**o+(0.0874536555354)*x[1]
        arg[1,2,0,0]=(0.943819415417)*x[0]**o+(0.884020001869)*x[0]+(-0.769166027477)*x[1]**o+(-0.10032307068)*x[1]
        arg[1,2,0,1]=(0.0670224828564)*x[0]**o+(-0.645424481071)*x[0]+(-0.756361652678)*x[1]**o+(-0.274687916196)*x[1]
        arg[1,2,1,0]=(0.397836742229)*x[0]**o+(-0.708791058745)*x[0]+(0.920869587663)*x[1]**o+(-0.250037069918)*x[1]
        arg[1,2,1,1]=(0.722570486719)*x[0]**o+(-0.358082774936)*x[0]+(0.786344673969)*x[1]**o+(0.310591893637)*x[1]
        arg[1,2,2,0]=(-0.379404573553)*x[0]**o+(-0.969753501645)*x[0]+(-0.0745714915337)*x[1]**o+(-0.262943796911)*x[1]
        arg[1,2,2,1]=(0.817999459632)*x[0]**o+(0.731516351792)*x[0]+(-0.531411613325)*x[1]**o+(-0.824858061154)*x[1]
        arg[1,3,0,0]=(-0.206295444021)*x[0]**o+(-0.781973944918)*x[0]+(-0.173845545382)*x[1]**o+(-0.360122364573)*x[1]
        arg[1,3,0,1]=(-0.0604068005719)*x[0]**o+(0.238691416507)*x[0]+(-0.82262982031)*x[1]**o+(-0.893751755968)*x[1]
        arg[1,3,1,0]=(-0.686409669386)*x[0]**o+(0.711830599399)*x[0]+(0.0839688142166)*x[1]**o+(-0.99463449684)*x[1]
        arg[1,3,1,1]=(-0.0864745975077)*x[0]**o+(0.00068907895313)*x[0]+(-0.766810513534)*x[1]**o+(0.685236202954)*x[1]
        arg[1,3,2,0]=(0.0909338812695)*x[0]**o+(-0.709595654169)*x[0]+(-0.483428504499)*x[1]**o+(-0.0707341191006)*x[1]
        arg[1,3,2,1]=(-0.38342983855)*x[0]**o+(-0.0407824575865)*x[0]+(-0.201329040891)*x[1]**o+(0.0240007937141)*x[1]
        arg[1,4,0,0]=(0.512465803173)*x[0]**o+(-0.371898603608)*x[0]+(-0.79453595861)*x[1]**o+(0.340546986375)*x[1]
        arg[1,4,0,1]=(0.742383198442)*x[0]**o+(-0.603548189912)*x[0]+(-0.270277825546)*x[1]**o+(-0.525100913894)*x[1]
        arg[1,4,1,0]=(-0.693074384029)*x[0]**o+(-0.825037404844)*x[0]+(-0.626116177978)*x[1]**o+(0.628159342479)*x[1]
        arg[1,4,1,1]=(-0.595844943255)*x[0]**o+(0.250758194807)*x[0]+(0.504390052142)*x[1]**o+(-0.124588859702)*x[1]
        arg[1,4,2,0]=(-0.651727490997)*x[0]**o+(0.411138504151)*x[0]+(0.323769160331)*x[1]**o+(0.904379126671)*x[1]
        arg[1,4,2,1]=(-0.71307850178)*x[0]**o+(0.734964608119)*x[0]+(0.348693698506)*x[1]**o+(0.272270639772)*x[1]
        arg[2,0,0,0]=(0.558236567771)*x[0]**o+(0.516985445776)*x[0]+(0.0278915137861)*x[1]**o+(-0.268619977833)*x[1]
        arg[2,0,0,1]=(-0.871181709658)*x[0]**o+(-0.132251987744)*x[0]+(-0.215497073461)*x[1]**o+(0.508008002082)*x[1]
        arg[2,0,1,0]=(0.880795999572)*x[0]**o+(-0.231151166496)*x[0]+(-0.698846681617)*x[1]**o+(0.914483413825)*x[1]
        arg[2,0,1,1]=(0.684792911211)*x[0]**o+(-0.801408782989)*x[0]+(0.689713232168)*x[1]**o+(0.142847173797)*x[1]
        arg[2,0,2,0]=(-0.912566319945)*x[0]**o+(0.811138574657)*x[0]+(-0.884250879071)*x[1]**o+(-0.177206225477)*x[1]
        arg[2,0,2,1]=(-0.898648046678)*x[0]**o+(0.581554373688)*x[0]+(0.567642878443)*x[1]**o+(0.636312832751)*x[1]
        arg[2,1,0,0]=(-0.788853365648)*x[0]**o+(-0.135329327941)*x[0]+(0.180850147594)*x[1]**o+(0.742619029335)*x[1]
        arg[2,1,0,1]=(-0.824539980645)*x[0]**o+(0.857173393615)*x[0]+(0.850455264916)*x[1]**o+(0.975422322464)*x[1]
        arg[2,1,1,0]=(0.462293517317)*x[0]**o+(-0.173835273794)*x[0]+(-0.800072514276)*x[1]**o+(0.140214114627)*x[1]
        arg[2,1,1,1]=(-0.187890814293)*x[0]**o+(0.534472364495)*x[0]+(-0.799640665571)*x[1]**o+(0.301287955302)*x[1]
        arg[2,1,2,0]=(0.278901492946)*x[0]**o+(0.620544326886)*x[0]+(-0.705702443752)*x[1]**o+(-0.489690636328)*x[1]
        arg[2,1,2,1]=(-0.926772900235)*x[0]**o+(0.134654742845)*x[0]+(0.356561975337)*x[1]**o+(-0.919999359857)*x[1]
        arg[2,2,0,0]=(-0.675956301416)*x[0]**o+(0.149582589205)*x[0]+(0.186440827385)*x[1]**o+(0.88381610446)*x[1]
        arg[2,2,0,1]=(-0.697267962915)*x[0]**o+(0.991366596424)*x[0]+(0.310290094618)*x[1]**o+(-0.881771505066)*x[1]
        arg[2,2,1,0]=(0.506027358474)*x[0]**o+(0.294136137354)*x[0]+(0.732003937836)*x[1]**o+(0.633469131504)*x[1]
        arg[2,2,1,1]=(-0.427146039728)*x[0]**o+(-0.975737063903)*x[0]+(-0.934913653772)*x[1]**o+(-0.662568325005)*x[1]
        arg[2,2,2,0]=(0.477849916621)*x[0]**o+(-0.798883883367)*x[0]+(-0.154217684441)*x[1]**o+(-0.175817547912)*x[1]
        arg[2,2,2,1]=(0.935192620387)*x[0]**o+(-0.193802588249)*x[0]+(0.406181949213)*x[1]**o+(0.20422475396)*x[1]
        arg[2,3,0,0]=(0.965944278856)*x[0]**o+(-0.736175411725)*x[0]+(0.92404347386)*x[1]**o+(0.273152788392)*x[1]
        arg[2,3,0,1]=(-0.606944412413)*x[0]**o+(0.692555413476)*x[0]+(0.812085676275)*x[1]**o+(-0.0809920218886)*x[1]
        arg[2,3,1,0]=(-0.314634777406)*x[0]**o+(0.750181618592)*x[0]+(0.0554739779642)*x[1]**o+(-0.333788326726)*x[1]
        arg[2,3,1,1]=(0.85610815703)*x[0]**o+(0.910096005524)*x[0]+(-0.227092814843)*x[1]**o+(-0.0849532598914)*x[1]
        arg[2,3,2,0]=(0.735064128504)*x[0]**o+(-0.253919528943)*x[0]+(0.8457463307)*x[1]**o+(0.865727345925)*x[1]
        arg[2,3,2,1]=(0.520607989389)*x[0]**o+(0.569210317105)*x[0]+(0.500780081677)*x[1]**o+(-0.787630914351)*x[1]
        arg[2,4,0,0]=(0.179774694296)*x[0]**o+(-0.890872128607)*x[0]+(-0.185758177247)*x[1]**o+(0.888231734606)*x[1]
        arg[2,4,0,1]=(0.059863462227)*x[0]**o+(-0.0796463123948)*x[0]+(-0.104564520504)*x[1]**o+(-0.336182528617)*x[1]
        arg[2,4,1,0]=(-0.636828275967)*x[0]**o+(0.156894572218)*x[0]+(-0.884260096115)*x[1]**o+(-0.967431438795)*x[1]
        arg[2,4,1,1]=(-0.958168110389)*x[0]**o+(-0.0424772631477)*x[0]+(0.512363473316)*x[1]**o+(-0.716729419668)*x[1]
        arg[2,4,2,0]=(0.19736654467)*x[0]**o+(-0.288723518002)*x[0]+(0.243674007033)*x[1]**o+(0.911210601142)*x[1]
        arg[2,4,2,1]=(-0.847493343055)*x[0]**o+(0.115003265593)*x[0]+(0.360071936714)*x[1]**o+(-0.0265385061141)*x[1]
        arg[3,0,0,0]=(-0.510550590836)*x[0]**o+(-0.266174606631)*x[0]+(0.771580708159)*x[1]**o+(-0.770554374259)*x[1]
        arg[3,0,0,1]=(-0.65795024738)*x[0]**o+(-0.83394628525)*x[0]+(0.812461325452)*x[1]**o+(0.191593733482)*x[1]
        arg[3,0,1,0]=(0.136878718548)*x[0]**o+(-0.000345061692795)*x[0]+(-0.0780280205008)*x[1]**o+(0.173925537373)*x[1]
        arg[3,0,1,1]=(0.584079470149)*x[0]**o+(0.138674493948)*x[0]+(-0.563553909861)*x[1]**o+(-0.254304838008)*x[1]
        arg[3,0,2,0]=(0.326192887554)*x[0]**o+(0.993615980303)*x[0]+(-0.26995279693)*x[1]**o+(0.216112615287)*x[1]
        arg[3,0,2,1]=(0.684786021819)*x[0]**o+(0.158934017292)*x[0]+(-0.457448199191)*x[1]**o+(0.644722228163)*x[1]
        arg[3,1,0,0]=(-0.198252162262)*x[0]**o+(-0.821832378445)*x[0]+(0.831807441916)*x[1]**o+(-0.70643670471)*x[1]
        arg[3,1,0,1]=(0.361698759016)*x[0]**o+(0.556511107101)*x[0]+(0.892219700601)*x[1]**o+(-0.861598520361)*x[1]
        arg[3,1,1,0]=(-0.107946706634)*x[0]**o+(-0.910191934756)*x[0]+(0.502507607871)*x[1]**o+(0.649089810516)*x[1]
        arg[3,1,1,1]=(-0.92959760672)*x[0]**o+(0.950682520961)*x[0]+(0.103321004187)*x[1]**o+(-0.631562667815)*x[1]
        arg[3,1,2,0]=(0.858565242025)*x[0]**o+(-0.106810606075)*x[0]+(0.236377425048)*x[1]**o+(-0.296320858145)*x[1]
        arg[3,1,2,1]=(0.148558873625)*x[0]**o+(0.458488334082)*x[0]+(-0.0172824905523)*x[1]**o+(-0.321212766581)*x[1]
        arg[3,2,0,0]=(-0.239733890336)*x[0]**o+(0.558819103697)*x[0]+(-0.151683129885)*x[1]**o+(0.153995780645)*x[1]
        arg[3,2,0,1]=(-0.949100270937)*x[0]**o+(0.0687171499139)*x[0]+(0.968028021831)*x[1]**o+(-0.935600499345)*x[1]
        arg[3,2,1,0]=(-0.739823767237)*x[0]**o+(0.417825150389)*x[0]+(0.318612170707)*x[1]**o+(0.627557337421)*x[1]
        arg[3,2,1,1]=(-0.965142167324)*x[0]**o+(0.981446854383)*x[0]+(0.266191185951)*x[1]**o+(-0.581303142841)*x[1]
        arg[3,2,2,0]=(-0.00312606993817)*x[0]**o+(-0.191885030828)*x[0]+(-0.342525068561)*x[1]**o+(-0.953710538626)*x[1]
        arg[3,2,2,1]=(-0.689127187476)*x[0]**o+(-0.107190366063)*x[0]+(-0.0556716250594)*x[1]**o+(-0.609462166754)*x[1]
        arg[3,3,0,0]=(-0.765237197853)*x[0]**o+(-0.0707283944117)*x[0]+(0.891789872287)*x[1]**o+(-0.255939791858)*x[1]
        arg[3,3,0,1]=(-0.367658362103)*x[0]**o+(0.27406309422)*x[0]+(0.866288683947)*x[1]**o+(0.403194030287)*x[1]
        arg[3,3,1,0]=(-0.362795000993)*x[0]**o+(-0.498774415158)*x[0]+(0.0726808331583)*x[1]**o+(-0.871066791136)*x[1]
        arg[3,3,1,1]=(-0.177149059814)*x[0]**o+(0.287463753612)*x[0]+(0.101726559941)*x[1]**o+(0.561797204517)*x[1]
        arg[3,3,2,0]=(0.770615121039)*x[0]**o+(-0.503402148251)*x[0]+(0.319753007634)*x[1]**o+(0.586868894261)*x[1]
        arg[3,3,2,1]=(0.924125904724)*x[0]**o+(-0.508988920423)*x[0]+(-0.866752504824)*x[1]**o+(0.518544700652)*x[1]
        arg[3,4,0,0]=(-0.973660291124)*x[0]**o+(0.746869369283)*x[0]+(-0.634097171844)*x[1]**o+(0.723638944736)*x[1]
        arg[3,4,0,1]=(-0.599366869066)*x[0]**o+(-0.59893829688)*x[0]+(0.124483401866)*x[1]**o+(-0.924820315792)*x[1]
        arg[3,4,1,0]=(-0.707374787729)*x[0]**o+(-0.494748755022)*x[0]+(0.391752209698)*x[1]**o+(0.149710640116)*x[1]
        arg[3,4,1,1]=(0.329587748119)*x[0]**o+(-0.788289803764)*x[0]+(-0.145100163626)*x[1]**o+(0.141493810237)*x[1]
        arg[3,4,2,0]=(0.17496644337)*x[0]**o+(-0.270242104589)*x[0]+(0.996684379179)*x[1]**o+(-0.751697440046)*x[1]
        arg[3,4,2,1]=(0.79046837395)*x[0]**o+(0.568394473062)*x[0]+(0.359419854016)*x[1]**o+(0.630686948244)*x[1]
        ref[0,0,0,0]=(0.306552070969)/(o+1.)+(0.157870653629)
        ref[0,0,0,1]=(0.0606500807631)/(o+1.)+(0.0612449724389)
        ref[0,0,1,0]=(0.639564070889)/(o+1.)+(0.666341832555)
        ref[0,0,1,1]=(0.671248593147)/(o+1.)+(-0.528628572255)
        ref[0,0,2,0]=(0.0735824774293)/(o+1.)+(-0.474736472501)
        ref[0,0,2,1]=(-0.9070527228)/(o+1.)+(-0.197350261427)
        ref[0,1,0,0]=(1.30449671875)/(o+1.)+(0.106123067554)
        ref[0,1,0,1]=(-0.420072850657)/(o+1.)+(-0.407896954678)
        ref[0,1,1,0]=(1.44174379356)/(o+1.)+(-0.809395886627)
        ref[0,1,1,1]=(0.155435385684)/(o+1.)+(0.627148132992)
        ref[0,1,2,0]=(-0.137204404518)/(o+1.)+(0.207874473889)
        ref[0,1,2,1]=(-0.289709410261)/(o+1.)+(0.204293345671)
        ref[0,2,0,0]=(1.26583912644)/(o+1.)+(0.498805698213)
        ref[0,2,0,1]=(1.17560298985)/(o+1.)+(0.931845120886)
        ref[0,2,1,0]=(-0.344986302901)/(o+1.)+(0.220341743892)
        ref[0,2,1,1]=(-1.94726093397)/(o+1.)+(0.206668541494)
        ref[0,2,2,0]=(-0.91808772513)/(o+1.)+(-0.0252951560363)
        ref[0,2,2,1]=(-1.02863406132)/(o+1.)+(-0.20183443794)
        ref[0,3,0,0]=(-1.04834664255)/(o+1.)+(0.00604709752155)
        ref[0,3,0,1]=(0.2902816262)/(o+1.)+(0.514361224456)
        ref[0,3,1,0]=(-1.31568170019)/(o+1.)+(0.883603667426)
        ref[0,3,1,1]=(-0.208650456965)/(o+1.)+(0.889841082216)
        ref[0,3,2,0]=(-0.741632466554)/(o+1.)+(0.0759327240318)
        ref[0,3,2,1]=(1.48653223703)/(o+1.)+(-0.854483988026)
        ref[0,4,0,0]=(0.333816089225)/(o+1.)+(-0.562947445201)
        ref[0,4,0,1]=(0.130634932061)/(o+1.)+(0.199922965378)
        ref[0,4,1,0]=(1.27534940781)/(o+1.)+(0.158640649858)
        ref[0,4,1,1]=(-0.83666690538)/(o+1.)+(-0.344088772961)
        ref[0,4,2,0]=(-0.432609891704)/(o+1.)+(-0.267052816612)
        ref[0,4,2,1]=(-0.477920665494)/(o+1.)+(-0.305859149312)
        ref[1,0,0,0]=(-0.738766811663)/(o+1.)+(-0.2321755194)
        ref[1,0,0,1]=(-0.688095002235)/(o+1.)+(0.664163739465)
        ref[1,0,1,0]=(1.14285258401)/(o+1.)+(-0.345190031677)
        ref[1,0,1,1]=(-0.093656366746)/(o+1.)+(-0.236308341076)
        ref[1,0,2,0]=(0.7151363028)/(o+1.)+(-0.56709255572)
        ref[1,0,2,1]=(-0.51932492992)/(o+1.)+(0.250667355044)
        ref[1,1,0,0]=(0.545772982093)/(o+1.)+(-0.163104398138)
        ref[1,1,0,1]=(0.629573063174)/(o+1.)+(-0.174653201576)
        ref[1,1,1,0]=(-0.154012812311)/(o+1.)+(-0.914297866096)
        ref[1,1,1,1]=(-0.61132115559)/(o+1.)+(-0.44114113966)
        ref[1,1,2,0]=(0.408406410726)/(o+1.)+(-0.076080229207)
        ref[1,1,2,1]=(0.893038507335)/(o+1.)+(-0.263052908312)
        ref[1,2,0,0]=(0.17465338794)/(o+1.)+(0.391848465594)
        ref[1,2,0,1]=(-0.689339169821)/(o+1.)+(-0.460056198634)
        ref[1,2,1,0]=(1.31870632989)/(o+1.)+(-0.479414064332)
        ref[1,2,1,1]=(1.50891516069)/(o+1.)+(-0.0237454406495)
        ref[1,2,2,0]=(-0.453976065087)/(o+1.)+(-0.616348649278)
        ref[1,2,2,1]=(0.286587846307)/(o+1.)+(-0.0466708546809)
        ref[1,3,0,0]=(-0.380140989403)/(o+1.)+(-0.571048154745)
        ref[1,3,0,1]=(-0.883036620882)/(o+1.)+(-0.32753016973)
        ref[1,3,1,0]=(-0.60244085517)/(o+1.)+(-0.141401948721)
        ref[1,3,1,1]=(-0.853285111042)/(o+1.)+(0.342962640954)
        ref[1,3,2,0]=(-0.39249462323)/(o+1.)+(-0.390164886635)
        ref[1,3,2,1]=(-0.584758879441)/(o+1.)+(-0.00839083193621)
        ref[1,4,0,0]=(-0.282070155438)/(o+1.)+(-0.0156758086169)
        ref[1,4,0,1]=(0.472105372896)/(o+1.)+(-0.564324551903)
        ref[1,4,1,0]=(-1.31919056201)/(o+1.)+(-0.0984390311826)
        ref[1,4,1,1]=(-0.0914548911124)/(o+1.)+(0.0630846675525)
        ref[1,4,2,0]=(-0.327958330666)/(o+1.)+(0.657758815411)
        ref[1,4,2,1]=(-0.364384803274)/(o+1.)+(0.503617623946)
        ref[2,0,0,0]=(0.586128081557)/(o+1.)+(0.124182733972)
        ref[2,0,0,1]=(-1.08667878312)/(o+1.)+(0.187878007169)
        ref[2,0,1,0]=(0.181949317955)/(o+1.)+(0.341666123665)
        ref[2,0,1,1]=(1.37450614338)/(o+1.)+(-0.329280804596)
        ref[2,0,2,0]=(-1.79681719902)/(o+1.)+(0.31696617459)
        ref[2,0,2,1]=(-0.331005168235)/(o+1.)+(0.608933603219)
        ref[2,1,0,0]=(-0.608003218053)/(o+1.)+(0.303644850697)
        ref[2,1,0,1]=(0.0259152842714)/(o+1.)+(0.916297858039)
        ref[2,1,1,0]=(-0.33777899696)/(o+1.)+(-0.0168105795836)
        ref[2,1,1,1]=(-0.987531479864)/(o+1.)+(0.417880159899)
        ref[2,1,2,0]=(-0.426800950806)/(o+1.)+(0.0654268452787)
        ref[2,1,2,1]=(-0.570210924898)/(o+1.)+(-0.392672308506)
        ref[2,2,0,0]=(-0.489515474032)/(o+1.)+(0.516699346832)
        ref[2,2,0,1]=(-0.386977868298)/(o+1.)+(0.054797545679)
        ref[2,2,1,0]=(1.23803129631)/(o+1.)+(0.463802634429)
        ref[2,2,1,1]=(-1.3620596935)/(o+1.)+(-0.819152694454)
        ref[2,2,2,0]=(0.32363223218)/(o+1.)+(-0.48735071564)
        ref[2,2,2,1]=(1.3413745696)/(o+1.)+(0.00521108285542)
        ref[2,3,0,0]=(1.88998775272)/(o+1.)+(-0.231511311666)
        ref[2,3,0,1]=(0.205141263861)/(o+1.)+(0.305781695794)
        ref[2,3,1,0]=(-0.259160799442)/(o+1.)+(0.208196645933)
        ref[2,3,1,1]=(0.629015342187)/(o+1.)+(0.412571372816)
        ref[2,3,2,0]=(1.5808104592)/(o+1.)+(0.305903908491)
        ref[2,3,2,1]=(1.02138807107)/(o+1.)+(-0.109210298623)
        ref[2,4,0,0]=(-0.00598348295078)/(o+1.)+(-0.00132019700044)
        ref[2,4,0,1]=(-0.0447010582767)/(o+1.)+(-0.207914420506)
        ref[2,4,1,0]=(-1.52108837208)/(o+1.)+(-0.405268433289)
        ref[2,4,1,1]=(-0.445804637073)/(o+1.)+(-0.379603341408)
        ref[2,4,2,0]=(0.441040551703)/(o+1.)+(0.31124354157)
        ref[2,4,2,1]=(-0.487421406342)/(o+1.)+(0.0442323797396)
        ref[3,0,0,0]=(0.261030117323)/(o+1.)+(-0.518364490445)
        ref[3,0,0,1]=(0.154511078072)/(o+1.)+(-0.321176275884)
        ref[3,0,1,0]=(0.0588506980476)/(o+1.)+(0.0867902378403)
        ref[3,0,1,1]=(0.0205255602871)/(o+1.)+(-0.0578151720304)
        ref[3,0,2,0]=(0.0562400906237)/(o+1.)+(0.604864297795)
        ref[3,0,2,1]=(0.227337822629)/(o+1.)+(0.401828122727)
        ref[3,1,0,0]=(0.633555279654)/(o+1.)+(-0.764134541578)
        ref[3,1,0,1]=(1.25391845962)/(o+1.)+(-0.15254370663)
        ref[3,1,1,0]=(0.394560901238)/(o+1.)+(-0.13055106212)
        ref[3,1,1,1]=(-0.826276602534)/(o+1.)+(0.159559926573)
        ref[3,1,2,0]=(1.09494266707)/(o+1.)+(-0.20156573211)
        ref[3,1,2,1]=(0.131276383073)/(o+1.)+(0.0686377837506)
        ref[3,2,0,0]=(-0.391417020221)/(o+1.)+(0.356407442171)
        ref[3,2,0,1]=(0.0189277508933)/(o+1.)+(-0.433441674716)
        ref[3,2,1,0]=(-0.42121159653)/(o+1.)+(0.522691243905)
        ref[3,2,1,1]=(-0.698950981373)/(o+1.)+(0.200071855771)
        ref[3,2,2,0]=(-0.345651138499)/(o+1.)+(-0.572797784727)
        ref[3,2,2,1]=(-0.744798812536)/(o+1.)+(-0.358326266408)
        ref[3,3,0,0]=(0.126552674434)/(o+1.)+(-0.163334093135)
        ref[3,3,0,1]=(0.498630321844)/(o+1.)+(0.338628562253)
        ref[3,3,1,0]=(-0.290114167835)/(o+1.)+(-0.684920603147)
        ref[3,3,1,1]=(-0.0754224998727)/(o+1.)+(0.424630479064)
        ref[3,3,2,0]=(1.09036812867)/(o+1.)+(0.0417333730049)
        ref[3,3,2,1]=(0.0573733998999)/(o+1.)+(0.00477789011464)
        ref[3,4,0,0]=(-1.60775746297)/(o+1.)+(0.73525415701)
        ref[3,4,0,1]=(-0.4748834672)/(o+1.)+(-0.761879306336)
        ref[3,4,1,0]=(-0.31562257803)/(o+1.)+(-0.172519057453)
        ref[3,4,1,1]=(0.184487584493)/(o+1.)+(-0.323397996764)
        ref[3,4,2,0]=(1.17165082255)/(o+1.)+(-0.510969772317)
        ref[3,4,2,1]=(1.14988822797)/(o+1.)+(0.599540710653)
      else:
        arg[0,0,0,0]=(-0.517412938774)*x[0]**o+(0.824804313111)*x[0]+(-0.681706563374)*x[1]**o+(-0.259540440272)*x[1]+(0.876735283937)*x[2]**o+(0.307072859807)*x[2]
        arg[0,0,0,1]=(-0.545981202251)*x[0]**o+(-0.530080933212)*x[0]+(-0.339289111599)*x[1]**o+(-0.726311605088)*x[1]+(0.334837705444)*x[2]**o+(-0.422667311064)*x[2]
        arg[0,0,1,0]=(-0.94261387328)*x[0]**o+(-0.315825326734)*x[0]+(0.542502801277)*x[1]**o+(-0.927525943646)*x[1]+(-0.802118238213)*x[2]**o+(-0.106512314706)*x[2]
        arg[0,0,1,1]=(-0.877940099286)*x[0]**o+(-0.801256557869)*x[0]+(-0.155983792242)*x[1]**o+(0.200547191496)*x[1]+(-0.960553133432)*x[2]**o+(0.968862792367)*x[2]
        arg[0,0,2,0]=(-0.870342547946)*x[0]**o+(-0.956447402661)*x[0]+(-0.506797803004)*x[1]**o+(0.342841675107)*x[1]+(0.528354159344)*x[2]**o+(-0.824546729564)*x[2]
        arg[0,0,2,1]=(-0.411899137296)*x[0]**o+(0.133736458559)*x[0]+(0.852726131619)*x[1]**o+(0.692795229813)*x[1]+(0.047372576403)*x[2]**o+(0.810485765945)*x[2]
        arg[0,1,0,0]=(0.961758233254)*x[0]**o+(-0.606307924964)*x[0]+(0.0612696412665)*x[1]**o+(-0.947823541802)*x[1]+(-0.113021104365)*x[2]**o+(-0.00802440322155)*x[2]
        arg[0,1,0,1]=(-0.684097052467)*x[0]**o+(0.467140835537)*x[0]+(-0.0742537091205)*x[1]**o+(0.541641916143)*x[1]+(0.987491548505)*x[2]**o+(-0.558380262255)*x[2]
        arg[0,1,1,0]=(-0.901821988855)*x[0]**o+(0.34443714654)*x[0]+(-0.682552654793)*x[1]**o+(0.398276086461)*x[1]+(0.334644264965)*x[2]**o+(0.21454845251)*x[2]
        arg[0,1,1,1]=(0.929614060295)*x[0]**o+(-0.322609307947)*x[0]+(-0.426962200724)*x[1]**o+(0.464659992564)*x[1]+(-0.13983863674)*x[2]**o+(-0.681648817325)*x[2]
        arg[0,1,2,0]=(0.504166508914)*x[0]**o+(-0.978562859041)*x[0]+(-0.769650404806)*x[1]**o+(0.443500137153)*x[1]+(0.969669965936)*x[2]**o+(0.401742207403)*x[2]
        arg[0,1,2,1]=(-0.478561324822)*x[0]**o+(-0.772019838834)*x[0]+(-0.779194844829)*x[1]**o+(-0.747980400065)*x[1]+(0.827885608192)*x[2]**o+(0.135571151528)*x[2]
        arg[0,2,0,0]=(-0.210426221052)*x[0]**o+(-0.464938141903)*x[0]+(-0.776844066203)*x[1]**o+(0.267244555705)*x[1]+(-0.773135929429)*x[2]**o+(-0.00846060888954)*x[2]
        arg[0,2,0,1]=(0.0945858334531)*x[0]**o+(-0.492400895713)*x[0]+(-0.128564669727)*x[1]**o+(-0.0951458189983)*x[1]+(-0.54426715378)*x[2]**o+(0.106545346072)*x[2]
        arg[0,2,1,0]=(-0.482424571392)*x[0]**o+(0.0431208155652)*x[0]+(0.557138601636)*x[1]**o+(-0.725564282551)*x[1]+(0.9081107419)*x[2]**o+(0.0179201585352)*x[2]
        arg[0,2,1,1]=(-0.555731853052)*x[0]**o+(0.447137254671)*x[0]+(0.820958717703)*x[1]**o+(-0.580251434504)*x[1]+(0.227884622965)*x[2]**o+(0.34460189453)*x[2]
        arg[0,2,2,0]=(-0.25868983898)*x[0]**o+(0.759165881486)*x[0]+(-0.499903058341)*x[1]**o+(0.896471093639)*x[1]+(0.351955520633)*x[2]**o+(0.16566756566)*x[2]
        arg[0,2,2,1]=(-0.37859354794)*x[0]**o+(0.925387188779)*x[0]+(0.0111055293719)*x[1]**o+(-0.311998642154)*x[1]+(0.441009182046)*x[2]**o+(0.0968247918271)*x[2]
        arg[0,3,0,0]=(-0.461665900641)*x[0]**o+(0.812002051396)*x[0]+(-0.767456852027)*x[1]**o+(0.809069781575)*x[1]+(0.690416953885)*x[2]**o+(0.171662369216)*x[2]
        arg[0,3,0,1]=(-0.804811324639)*x[0]**o+(-0.0962711523954)*x[0]+(0.980130059825)*x[1]**o+(0.98279180603)*x[1]+(-0.0614353559054)*x[2]**o+(0.645637153538)*x[2]
        arg[0,3,1,0]=(0.97878415369)*x[0]**o+(-0.676321251698)*x[0]+(0.916075195717)*x[1]**o+(0.244509683371)*x[1]+(-0.939563193145)*x[2]**o+(-0.866506108913)*x[2]
        arg[0,3,1,1]=(-0.676965895162)*x[0]**o+(-0.124355338519)*x[0]+(0.680151101195)*x[1]**o+(-0.839412433156)*x[1]+(0.434088587223)*x[2]**o+(-0.275286048166)*x[2]
        arg[0,3,2,0]=(0.414822536626)*x[0]**o+(0.493836498194)*x[0]+(0.967077097745)*x[1]**o+(-0.0115301157869)*x[1]+(-0.307994657929)*x[2]**o+(-0.967766608312)*x[2]
        arg[0,3,2,1]=(-0.34064670348)*x[0]**o+(-0.975121872905)*x[0]+(0.0839457874766)*x[1]**o+(0.0995106978636)*x[1]+(-0.533837237516)*x[2]**o+(0.771268044302)*x[2]
        arg[0,4,0,0]=(0.223197544637)*x[0]**o+(0.712738871576)*x[0]+(-0.386972075747)*x[1]**o+(0.289829507744)*x[1]+(0.153418793839)*x[2]**o+(0.876087048492)*x[2]
        arg[0,4,0,1]=(0.696750948653)*x[0]**o+(0.466905335953)*x[0]+(-0.611476041907)*x[1]**o+(-0.939902457901)*x[1]+(-0.752675432265)*x[2]**o+(-0.869502717044)*x[2]
        arg[0,4,1,0]=(-0.465562201723)*x[0]**o+(-0.388716352051)*x[0]+(0.260044093503)*x[1]**o+(0.919379163294)*x[1]+(-0.902453394797)*x[2]**o+(0.378318744505)*x[2]
        arg[0,4,1,1]=(0.121990842946)*x[0]**o+(-0.691267894153)*x[0]+(-0.703960568755)*x[1]**o+(-0.582404896428)*x[1]+(0.57560180036)*x[2]**o+(0.537099448589)*x[2]
        arg[0,4,2,0]=(-0.495452196513)*x[0]**o+(-0.939389935669)*x[0]+(-0.237205717598)*x[1]**o+(0.61376516207)*x[1]+(-0.506082738728)*x[2]**o+(0.407704148542)*x[2]
        arg[0,4,2,1]=(0.450976180771)*x[0]**o+(-0.00739258898437)*x[0]+(0.210130848093)*x[1]**o+(0.624639007009)*x[1]+(0.173795671727)*x[2]**o+(0.508269170853)*x[2]
        arg[1,0,0,0]=(0.609685942177)*x[0]**o+(0.609299494609)*x[0]+(0.321046388121)*x[1]**o+(0.598197189498)*x[1]+(0.536092177618)*x[2]**o+(0.576427431157)*x[2]
        arg[1,0,0,1]=(0.412500706234)*x[0]**o+(0.905942239017)*x[0]+(-0.974980343169)*x[1]**o+(-0.131245997607)*x[1]+(0.745809068018)*x[2]**o+(0.543470258153)*x[2]
        arg[1,0,1,0]=(-0.903032545947)*x[0]**o+(-0.778693321051)*x[0]+(0.399342164105)*x[1]**o+(0.478212896767)*x[1]+(0.254596269802)*x[2]**o+(0.70308382001)*x[2]
        arg[1,0,1,1]=(-0.281529506793)*x[0]**o+(0.74166036027)*x[0]+(0.629713747725)*x[1]**o+(-0.275164746597)*x[1]+(0.37523909979)*x[2]**o+(-0.878026427978)*x[2]
        arg[1,0,2,0]=(0.236027764061)*x[0]**o+(-0.816910805083)*x[0]+(0.0782011624185)*x[1]**o+(-0.644004245666)*x[1]+(0.293586462602)*x[2]**o+(0.557045894695)*x[2]
        arg[1,0,2,1]=(0.868061727269)*x[0]**o+(0.53329318943)*x[0]+(0.367603915235)*x[1]**o+(0.89280277925)*x[1]+(-0.316595424948)*x[2]**o+(0.593830565147)*x[2]
        arg[1,1,0,0]=(0.886950595996)*x[0]**o+(-0.290323112286)*x[0]+(-0.863347030089)*x[1]**o+(0.545096994541)*x[1]+(0.338200944281)*x[2]**o+(-0.79898415048)*x[2]
        arg[1,1,0,1]=(0.103761462006)*x[0]**o+(-0.046916161864)*x[0]+(0.0312549239467)*x[1]**o+(0.214045118984)*x[1]+(-0.523161766402)*x[2]**o+(0.714207028317)*x[2]
        arg[1,1,1,0]=(0.494943183987)*x[0]**o+(-0.0543096386006)*x[0]+(-0.263875120588)*x[1]**o+(0.776031489925)*x[1]+(0.881224752145)*x[2]**o+(-0.650038283228)*x[2]
        arg[1,1,1,1]=(0.0871887540513)*x[0]**o+(-0.953848841349)*x[0]+(-0.223751459589)*x[1]**o+(0.227400823993)*x[1]+(-0.224200553009)*x[2]**o+(0.594012970899)*x[2]
        arg[1,1,2,0]=(-0.443116951346)*x[0]**o+(0.891806149626)*x[0]+(0.94731530917)*x[1]**o+(0.739499894809)*x[1]+(-0.92613053703)*x[2]**o+(0.464883087528)*x[2]
        arg[1,1,2,1]=(-0.149831145559)*x[0]**o+(0.913999704597)*x[0]+(0.76082978152)*x[1]**o+(-0.441535347622)*x[1]+(0.902841586936)*x[2]**o+(-0.509022004948)*x[2]
        arg[1,2,0,0]=(0.508230188489)*x[0]**o+(0.241669921231)*x[0]+(-0.802811365056)*x[1]**o+(0.298424305948)*x[1]+(-0.949407454532)*x[2]**o+(-0.131426880214)*x[2]
        arg[1,2,0,1]=(0.65372313274)*x[0]**o+(0.195891642771)*x[0]+(-0.160845429219)*x[1]**o+(0.500815823724)*x[1]+(0.875785249339)*x[2]**o+(0.325024339211)*x[2]
        arg[1,2,1,0]=(0.908405682536)*x[0]**o+(-0.146675166477)*x[0]+(0.24033815873)*x[1]**o+(-0.62814219771)*x[1]+(-0.512365788351)*x[2]**o+(-0.103817191415)*x[2]
        arg[1,2,1,1]=(0.824305237501)*x[0]**o+(0.175753992729)*x[0]+(0.992502114126)*x[1]**o+(0.973110627555)*x[1]+(0.355352002177)*x[2]**o+(0.112252263955)*x[2]
        arg[1,2,2,0]=(-0.485996285003)*x[0]**o+(-0.532551123887)*x[0]+(0.708205121405)*x[1]**o+(-0.803269038441)*x[1]+(-0.232049124534)*x[2]**o+(0.55437652866)*x[2]
        arg[1,2,2,1]=(0.551934776224)*x[0]**o+(0.727629664538)*x[0]+(0.197393511049)*x[1]**o+(-0.146619812174)*x[1]+(0.316592306936)*x[2]**o+(0.0263881493809)*x[2]
        arg[1,3,0,0]=(-0.918554166254)*x[0]**o+(-0.839150758884)*x[0]+(0.436713305329)*x[1]**o+(-0.410131731387)*x[1]+(0.321246028791)*x[2]**o+(0.706389384376)*x[2]
        arg[1,3,0,1]=(-0.870168753295)*x[0]**o+(-0.00473364206025)*x[0]+(0.594648540734)*x[1]**o+(-0.924887214906)*x[1]+(0.58858894515)*x[2]**o+(0.0152152159566)*x[2]
        arg[1,3,1,0]=(0.64005604234)*x[0]**o+(0.308265334836)*x[0]+(0.321645661286)*x[1]**o+(0.670017614875)*x[1]+(0.712764136703)*x[2]**o+(0.458986964342)*x[2]
        arg[1,3,1,1]=(-0.201376578944)*x[0]**o+(-0.271793865118)*x[0]+(-0.798181331582)*x[1]**o+(0.690100877945)*x[1]+(0.600350295622)*x[2]**o+(0.341066858436)*x[2]
        arg[1,3,2,0]=(0.459832113137)*x[0]**o+(0.0566780273877)*x[0]+(0.221723079112)*x[1]**o+(-0.681001728734)*x[1]+(0.265505135729)*x[2]**o+(-0.389365671367)*x[2]
        arg[1,3,2,1]=(0.715444384493)*x[0]**o+(-0.705697520786)*x[0]+(0.945208081248)*x[1]**o+(0.0233171867808)*x[1]+(0.740156166802)*x[2]**o+(-0.144314337012)*x[2]
        arg[1,4,0,0]=(-0.490162545806)*x[0]**o+(0.542981268396)*x[0]+(-0.841471465564)*x[1]**o+(0.79091798393)*x[1]+(0.245598142884)*x[2]**o+(0.4740296042)*x[2]
        arg[1,4,0,1]=(-0.504504615322)*x[0]**o+(0.45494067251)*x[0]+(-0.612427248728)*x[1]**o+(0.0880340051743)*x[1]+(0.645906227711)*x[2]**o+(0.772038416622)*x[2]
        arg[1,4,1,0]=(0.945538632492)*x[0]**o+(0.98519603491)*x[0]+(-0.245495451859)*x[1]**o+(-0.70732864026)*x[1]+(0.378733662069)*x[2]**o+(0.579554415018)*x[2]
        arg[1,4,1,1]=(0.886659064398)*x[0]**o+(-0.491870874438)*x[0]+(0.751814742427)*x[1]**o+(0.430589443767)*x[1]+(-0.729193959143)*x[2]**o+(-0.699344758538)*x[2]
        arg[1,4,2,0]=(-0.4190140481)*x[0]**o+(0.84509612452)*x[0]+(-0.263608911857)*x[1]**o+(0.693648282371)*x[1]+(-0.144463932166)*x[2]**o+(0.384293689266)*x[2]
        arg[1,4,2,1]=(-0.266973780787)*x[0]**o+(0.613988367707)*x[0]+(0.686955199789)*x[1]**o+(-0.301585818785)*x[1]+(-0.574601202558)*x[2]**o+(-0.430718616071)*x[2]
        arg[2,0,0,0]=(-0.117390727803)*x[0]**o+(-0.25850139145)*x[0]+(0.324909086031)*x[1]**o+(-0.593112241553)*x[1]+(-0.51016768177)*x[2]**o+(0.392299415334)*x[2]
        arg[2,0,0,1]=(0.290478531696)*x[0]**o+(0.758153023626)*x[0]+(0.531003993361)*x[1]**o+(-0.112481510374)*x[1]+(0.950222158203)*x[2]**o+(-0.742480526204)*x[2]
        arg[2,0,1,0]=(-0.754820921225)*x[0]**o+(-0.408482044077)*x[0]+(-0.229697426486)*x[1]**o+(-0.983010424713)*x[1]+(-0.286618604257)*x[2]**o+(-0.489034108451)*x[2]
        arg[2,0,1,1]=(0.96775757822)*x[0]**o+(0.797765725721)*x[0]+(-0.90936176648)*x[1]**o+(-0.372755885193)*x[1]+(-0.144727874143)*x[2]**o+(0.341624944551)*x[2]
        arg[2,0,2,0]=(0.672779422278)*x[0]**o+(-0.936102224964)*x[0]+(-0.763620352857)*x[1]**o+(0.265802509744)*x[1]+(0.275798875047)*x[2]**o+(0.235397520942)*x[2]
        arg[2,0,2,1]=(0.192866248041)*x[0]**o+(0.656863606616)*x[0]+(-0.902796948663)*x[1]**o+(0.692123089256)*x[1]+(0.311287987388)*x[2]**o+(0.555427507546)*x[2]
        arg[2,1,0,0]=(0.464585537593)*x[0]**o+(0.161744615208)*x[0]+(-0.325569920662)*x[1]**o+(-0.47142644441)*x[1]+(0.885151021784)*x[2]**o+(-0.840595219086)*x[2]
        arg[2,1,0,1]=(-0.659253397488)*x[0]**o+(0.54391314862)*x[0]+(-0.595612529952)*x[1]**o+(-0.0472758114379)*x[1]+(0.876391930188)*x[2]**o+(-0.440313199946)*x[2]
        arg[2,1,1,0]=(-0.445591423364)*x[0]**o+(0.179144841652)*x[0]+(-0.294166502086)*x[1]**o+(-0.373820778739)*x[1]+(0.464609314863)*x[2]**o+(0.407893781978)*x[2]
        arg[2,1,1,1]=(-0.767512700255)*x[0]**o+(0.63198312168)*x[0]+(0.652318598902)*x[1]**o+(0.817874971823)*x[1]+(0.397211574885)*x[2]**o+(0.33174354886)*x[2]
        arg[2,1,2,0]=(0.473801282131)*x[0]**o+(-0.778895701394)*x[0]+(-0.801488721847)*x[1]**o+(-0.585840161723)*x[1]+(-0.805496721223)*x[2]**o+(0.585971840421)*x[2]
        arg[2,1,2,1]=(0.472329620891)*x[0]**o+(-0.0473805663117)*x[0]+(0.0894117727522)*x[1]**o+(-0.308473625733)*x[1]+(-0.658376953956)*x[2]**o+(0.32027592346)*x[2]
        arg[2,2,0,0]=(0.843241796369)*x[0]**o+(0.346857959798)*x[0]+(0.0111223078269)*x[1]**o+(-0.949623227267)*x[1]+(-0.152362393376)*x[2]**o+(-0.886340001141)*x[2]
        arg[2,2,0,1]=(-0.555393018762)*x[0]**o+(0.666364344488)*x[0]+(0.181237902671)*x[1]**o+(0.670959040345)*x[1]+(0.57383285317)*x[2]**o+(-0.451027832775)*x[2]
        arg[2,2,1,0]=(0.249709446431)*x[0]**o+(0.568753063379)*x[0]+(-0.812386750097)*x[1]**o+(0.506805124672)*x[1]+(-0.454844217662)*x[2]**o+(0.843435269288)*x[2]
        arg[2,2,1,1]=(0.753144930025)*x[0]**o+(0.0181642998269)*x[0]+(0.622635241593)*x[1]**o+(-0.372883842874)*x[1]+(0.366920965577)*x[2]**o+(-0.092391524816)*x[2]
        arg[2,2,2,0]=(0.309601906611)*x[0]**o+(-0.502573192965)*x[0]+(-0.420795137333)*x[1]**o+(-0.807109356307)*x[1]+(-0.802887583918)*x[2]**o+(0.552609684817)*x[2]
        arg[2,2,2,1]=(0.320933223342)*x[0]**o+(0.541931544704)*x[0]+(0.988715316931)*x[1]**o+(-0.4197325237)*x[1]+(-0.438705512435)*x[2]**o+(-0.540175093599)*x[2]
        arg[2,3,0,0]=(0.018870015518)*x[0]**o+(-0.615120691773)*x[0]+(0.596295280542)*x[1]**o+(0.210874310884)*x[1]+(-0.652194502321)*x[2]**o+(0.74019575754)*x[2]
        arg[2,3,0,1]=(0.820285141622)*x[0]**o+(0.130985741886)*x[0]+(-0.397387767472)*x[1]**o+(0.975103051154)*x[1]+(0.0887369214611)*x[2]**o+(0.88498731668)*x[2]
        arg[2,3,1,0]=(0.160503336808)*x[0]**o+(0.641180113144)*x[0]+(-0.408109675737)*x[1]**o+(-0.355443951133)*x[1]+(-0.429622016495)*x[2]**o+(0.998729833779)*x[2]
        arg[2,3,1,1]=(-0.646524352112)*x[0]**o+(-0.289894880059)*x[0]+(0.909761879864)*x[1]**o+(0.30453651619)*x[1]+(0.884596709091)*x[2]**o+(0.329408737003)*x[2]
        arg[2,3,2,0]=(0.459385919432)*x[0]**o+(0.416447271916)*x[0]+(-0.400798659819)*x[1]**o+(-0.769043818187)*x[1]+(-0.631682400803)*x[2]**o+(-0.80150361084)*x[2]
        arg[2,3,2,1]=(-0.544824298969)*x[0]**o+(0.191301965562)*x[0]+(0.41061102918)*x[1]**o+(-0.999195759064)*x[1]+(-0.0226105975802)*x[2]**o+(-0.137078691725)*x[2]
        arg[2,4,0,0]=(-0.624033952375)*x[0]**o+(0.0792222332007)*x[0]+(0.766756718455)*x[1]**o+(0.581487379511)*x[1]+(-0.437940391245)*x[2]**o+(0.115617341492)*x[2]
        arg[2,4,0,1]=(-0.300526068838)*x[0]**o+(-0.822844066866)*x[0]+(0.660949310367)*x[1]**o+(-0.0993949179102)*x[1]+(0.599884615655)*x[2]**o+(-0.903177033523)*x[2]
        arg[2,4,1,0]=(-0.654642517091)*x[0]**o+(0.428391621765)*x[0]+(0.932377441464)*x[1]**o+(-0.459525189684)*x[1]+(0.667425199675)*x[2]**o+(-0.937213410686)*x[2]
        arg[2,4,1,1]=(0.258067525957)*x[0]**o+(0.619943768383)*x[0]+(0.458217567511)*x[1]**o+(0.0812191081554)*x[1]+(0.743260637332)*x[2]**o+(-0.796150472293)*x[2]
        arg[2,4,2,0]=(0.782473636556)*x[0]**o+(-0.441405613305)*x[0]+(0.3951595327)*x[1]**o+(-0.449855781187)*x[1]+(-0.964130179212)*x[2]**o+(-0.299199410525)*x[2]
        arg[2,4,2,1]=(-0.878340584397)*x[0]**o+(-0.674575785429)*x[0]+(0.838137461257)*x[1]**o+(-0.498710475715)*x[1]+(-0.117205127994)*x[2]**o+(0.0934669704542)*x[2]
        arg[3,0,0,0]=(0.139631366137)*x[0]**o+(-0.893001605238)*x[0]+(-0.420780773096)*x[1]**o+(0.439055863393)*x[1]+(0.173113363237)*x[2]**o+(0.893867443017)*x[2]
        arg[3,0,0,1]=(-0.804823042124)*x[0]**o+(0.250231181298)*x[0]+(0.358464048439)*x[1]**o+(-0.847391231172)*x[1]+(-0.515391975323)*x[2]**o+(-0.999640302937)*x[2]
        arg[3,0,1,0]=(0.328676968943)*x[0]**o+(-0.0521242139559)*x[0]+(0.861548411537)*x[1]**o+(-0.288979961428)*x[1]+(-0.927901849987)*x[2]**o+(0.857869958971)*x[2]
        arg[3,0,1,1]=(-0.654866876537)*x[0]**o+(-0.0515462321916)*x[0]+(-0.224164685097)*x[1]**o+(-0.84207399757)*x[1]+(-0.442392040219)*x[2]**o+(-0.915734076984)*x[2]
        arg[3,0,2,0]=(-0.322517406382)*x[0]**o+(-0.757146499161)*x[0]+(-0.541690197465)*x[1]**o+(0.931090695413)*x[1]+(-0.355505263744)*x[2]**o+(-0.773185404514)*x[2]
        arg[3,0,2,1]=(0.421030649833)*x[0]**o+(-0.83768939688)*x[0]+(-0.863791071447)*x[1]**o+(0.78393205545)*x[1]+(-0.489893911681)*x[2]**o+(-0.291989245709)*x[2]
        arg[3,1,0,0]=(0.550839051007)*x[0]**o+(-0.883853789938)*x[0]+(-0.339529971923)*x[1]**o+(-0.756936847344)*x[1]+(0.724888858656)*x[2]**o+(-0.981057762224)*x[2]
        arg[3,1,0,1]=(0.711792743156)*x[0]**o+(0.233248429378)*x[0]+(0.526076154408)*x[1]**o+(0.349364804166)*x[1]+(0.271475992486)*x[2]**o+(0.782429402713)*x[2]
        arg[3,1,1,0]=(-0.499072848146)*x[0]**o+(-0.0604888692308)*x[0]+(0.545442893272)*x[1]**o+(-0.688797522515)*x[1]+(0.305341295222)*x[2]**o+(0.130010601331)*x[2]
        arg[3,1,1,1]=(0.940695359452)*x[0]**o+(0.113015472294)*x[0]+(-0.506254395321)*x[1]**o+(-0.252044277039)*x[1]+(0.0172784972094)*x[2]**o+(-0.184256978193)*x[2]
        arg[3,1,2,0]=(0.550472143981)*x[0]**o+(-0.729809590776)*x[0]+(-0.293765999495)*x[1]**o+(-0.378042976721)*x[1]+(0.973242167993)*x[2]**o+(-0.75350391134)*x[2]
        arg[3,1,2,1]=(0.369266178598)*x[0]**o+(-0.764034492366)*x[0]+(-0.713325992328)*x[1]**o+(0.948932375717)*x[1]+(-0.335074977323)*x[2]**o+(0.117829591745)*x[2]
        arg[3,2,0,0]=(0.426458076699)*x[0]**o+(-0.520551329627)*x[0]+(-0.773464312884)*x[1]**o+(0.26332468989)*x[1]+(0.143497828773)*x[2]**o+(0.0676278275986)*x[2]
        arg[3,2,0,1]=(0.85806016531)*x[0]**o+(0.755361032731)*x[0]+(0.403942188112)*x[1]**o+(0.0113665816202)*x[1]+(-0.439835633198)*x[2]**o+(0.620930145183)*x[2]
        arg[3,2,1,0]=(-0.852838624821)*x[0]**o+(-0.918594624263)*x[0]+(-0.572898555139)*x[1]**o+(-0.14353390918)*x[1]+(-0.716010071307)*x[2]**o+(0.134687635305)*x[2]
        arg[3,2,1,1]=(-0.14282590746)*x[0]**o+(-0.169314199229)*x[0]+(0.880745511071)*x[1]**o+(0.540733710441)*x[1]+(-0.131424658303)*x[2]**o+(-0.93209431119)*x[2]
        arg[3,2,2,0]=(-0.864016075688)*x[0]**o+(-0.265794760131)*x[0]+(0.444625188371)*x[1]**o+(0.133693703972)*x[1]+(-0.749799597936)*x[2]**o+(0.886355820615)*x[2]
        arg[3,2,2,1]=(0.676281241167)*x[0]**o+(-0.351800298325)*x[0]+(-0.751498380099)*x[1]**o+(-0.561737786104)*x[1]+(-0.984872672853)*x[2]**o+(0.476840947707)*x[2]
        arg[3,3,0,0]=(0.457874366022)*x[0]**o+(-0.126929649966)*x[0]+(0.666077559139)*x[1]**o+(-0.966437669224)*x[1]+(-0.849000502816)*x[2]**o+(0.990790747926)*x[2]
        arg[3,3,0,1]=(0.45136859326)*x[0]**o+(0.771355456911)*x[0]+(-0.678558472676)*x[1]**o+(0.0366869703264)*x[1]+(-0.257247358952)*x[2]**o+(-0.290852285538)*x[2]
        arg[3,3,1,0]=(-0.623007069616)*x[0]**o+(-0.365255080044)*x[0]+(0.441927420573)*x[1]**o+(0.331640625664)*x[1]+(-0.21153951456)*x[2]**o+(0.297967669123)*x[2]
        arg[3,3,1,1]=(-0.836057742118)*x[0]**o+(0.959873293328)*x[0]+(-0.52034405502)*x[1]**o+(-0.178838686591)*x[1]+(0.317610563306)*x[2]**o+(-0.555134650833)*x[2]
        arg[3,3,2,0]=(0.0819108022024)*x[0]**o+(0.69432296821)*x[0]+(-0.23685434367)*x[1]**o+(-0.0442137744391)*x[1]+(-0.879175086057)*x[2]**o+(-0.28031950012)*x[2]
        arg[3,3,2,1]=(0.340083286763)*x[0]**o+(-0.381942838645)*x[0]+(0.570579443852)*x[1]**o+(0.772192708863)*x[1]+(0.42813556157)*x[2]**o+(0.735683065499)*x[2]
        arg[3,4,0,0]=(-0.866813692418)*x[0]**o+(-0.53715159706)*x[0]+(0.0725900211922)*x[1]**o+(0.0891167848614)*x[1]+(-0.381482922347)*x[2]**o+(-0.521504123034)*x[2]
        arg[3,4,0,1]=(0.487750922723)*x[0]**o+(0.967820473849)*x[0]+(0.305600886352)*x[1]**o+(0.350777657643)*x[1]+(-0.383731485788)*x[2]**o+(0.814016126476)*x[2]
        arg[3,4,1,0]=(-0.186663244344)*x[0]**o+(0.144872424379)*x[0]+(0.844454339304)*x[1]**o+(0.252133708483)*x[1]+(-0.748235508222)*x[2]**o+(0.702952185343)*x[2]
        arg[3,4,1,1]=(-0.962296047927)*x[0]**o+(0.874277032826)*x[0]+(-0.404689253219)*x[1]**o+(-0.174075799004)*x[1]+(0.65086350106)*x[2]**o+(-0.109850650677)*x[2]
        arg[3,4,2,0]=(-0.59515836027)*x[0]**o+(-0.0917296352911)*x[0]+(0.834779238353)*x[1]**o+(0.951252871777)*x[1]+(0.983008684497)*x[2]**o+(-0.784806954891)*x[2]
        arg[3,4,2,1]=(0.603110971708)*x[0]**o+(0.291507042984)*x[0]+(-0.401523286981)*x[1]**o+(-0.994760476862)*x[1]+(0.410516007475)*x[2]**o+(0.368982230001)*x[2]
        ref[0,0,0,0]=(-0.322384218211)/(o+1.)+(0.436168366323)
        ref[0,0,0,1]=(-0.550432608405)/(o+1.)+(-0.839529924682)
        ref[0,0,1,0]=(-1.20222931022)/(o+1.)+(-0.674931792543)
        ref[0,0,1,1]=(-1.99447702496)/(o+1.)+(0.184076712997)
        ref[0,0,2,0]=(-0.848786191606)/(o+1.)+(-0.719076228559)
        ref[0,0,2,1]=(0.488199570726)/(o+1.)+(0.818508727159)
        ref[0,1,0,0]=(0.910006770156)/(o+1.)+(-0.781077934994)
        ref[0,1,0,1]=(0.229140786917)/(o+1.)+(0.225201244712)
        ref[0,1,1,0]=(-1.24973037868)/(o+1.)+(0.478630842756)
        ref[0,1,1,1]=(0.362813222831)/(o+1.)+(-0.269799066354)
        ref[0,1,2,0]=(0.704186070044)/(o+1.)+(-0.0666602572428)
        ref[0,1,2,1]=(-0.429870561459)/(o+1.)+(-0.692214543685)
        ref[0,2,0,0]=(-1.76040621668)/(o+1.)+(-0.103077097544)
        ref[0,2,0,1]=(-0.578245990054)/(o+1.)+(-0.24050068432)
        ref[0,2,1,0]=(0.982824772144)/(o+1.)+(-0.332261654225)
        ref[0,2,1,1]=(0.493111487615)/(o+1.)+(0.105743857349)
        ref[0,2,2,0]=(-0.406637376689)/(o+1.)+(0.910652270393)
        ref[0,2,2,1]=(0.0735211634779)/(o+1.)+(0.355106669226)
        ref[0,3,0,0]=(-0.538705798782)/(o+1.)+(0.896367101094)
        ref[0,3,0,1]=(0.11388337928)/(o+1.)+(0.766078903586)
        ref[0,3,1,0]=(0.955296156262)/(o+1.)+(-0.64915883862)
        ref[0,3,1,1]=(0.437273793256)/(o+1.)+(-0.61952690992)
        ref[0,3,2,0]=(1.07390497644)/(o+1.)+(-0.242730112953)
        ref[0,3,2,1]=(-0.79053815352)/(o+1.)+(-0.0521715653697)
        ref[0,4,0,0]=(-0.0103557372704)/(o+1.)+(0.939327713906)
        ref[0,4,0,1]=(-0.66740052552)/(o+1.)+(-0.671249919496)
        ref[0,4,1,0]=(-1.10797150302)/(o+1.)+(0.454490777873)
        ref[0,4,1,1]=(-0.00636792544909)/(o+1.)+(-0.368286670996)
        ref[0,4,2,0]=(-1.23874065284)/(o+1.)+(0.0410396874718)
        ref[0,4,2,1]=(0.834902700591)/(o+1.)+(0.562757794439)
        ref[1,0,0,0]=(1.46682450791)/(o+1.)+(0.891962057632)
        ref[1,0,0,1]=(0.183329431084)/(o+1.)+(0.659083249781)
        ref[1,0,1,0]=(-0.24909411204)/(o+1.)+(0.201301697863)
        ref[1,0,1,1]=(0.723423340723)/(o+1.)+(-0.205765407152)
        ref[1,0,2,0]=(0.607815389082)/(o+1.)+(-0.451934578027)
        ref[1,0,2,1]=(0.919070217555)/(o+1.)+(1.00996326691)
        ref[1,1,0,0]=(0.361804510188)/(o+1.)+(-0.272105134113)
        ref[1,1,0,1]=(-0.388145380449)/(o+1.)+(0.440667992719)
        ref[1,1,1,0]=(1.11229281554)/(o+1.)+(0.0358417840478)
        ref[1,1,1,1]=(-0.360763258547)/(o+1.)+(-0.0662175232282)
        ref[1,1,2,0]=(-0.421932179206)/(o+1.)+(1.04809456598)
        ref[1,1,2,1]=(1.5138402229)/(o+1.)+(-0.0182788239862)
        ref[1,2,0,0]=(-1.2439886311)/(o+1.)+(0.204333673483)
        ref[1,2,0,1]=(1.36866295286)/(o+1.)+(0.510865902853)
        ref[1,2,1,0]=(0.636378052915)/(o+1.)+(-0.439317277801)
        ref[1,2,1,1]=(2.1721593538)/(o+1.)+(0.63055844212)
        ref[1,2,2,0]=(-0.00984028813171)/(o+1.)+(-0.390721816834)
        ref[1,2,2,1]=(1.06592059421)/(o+1.)+(0.303699000873)
        ref[1,3,0,0]=(-0.160594832134)/(o+1.)+(-0.271446552947)
        ref[1,3,0,1]=(0.313068732589)/(o+1.)+(-0.457202820505)
        ref[1,3,1,0]=(1.67446584033)/(o+1.)+(0.718634957026)
        ref[1,3,1,1]=(-0.399207614903)/(o+1.)+(0.379686935632)
        ref[1,3,2,0]=(0.947060327978)/(o+1.)+(-0.506844686357)
        ref[1,3,2,1]=(2.40080863254)/(o+1.)+(-0.413347335508)
        ref[1,4,0,0]=(-1.08603586849)/(o+1.)+(0.903964428263)
        ref[1,4,0,1]=(-0.471025636339)/(o+1.)+(0.657506547153)
        ref[1,4,1,0]=(1.0787768427)/(o+1.)+(0.428710904834)
        ref[1,4,1,1]=(0.909279847682)/(o+1.)+(-0.380313094604)
        ref[1,4,2,0]=(-0.827086892123)/(o+1.)+(0.961519048079)
        ref[1,4,2,1]=(-0.154619783556)/(o+1.)+(-0.0591580335741)
        ref[2,0,0,0]=(-0.302649323542)/(o+1.)+(-0.229657108834)
        ref[2,0,0,1]=(1.77170468326)/(o+1.)+(-0.0484045064764)
        ref[2,0,1,0]=(-1.27113695197)/(o+1.)+(-0.94026328862)
        ref[2,0,1,1]=(-0.0863320624027)/(o+1.)+(0.383317392539)
        ref[2,0,2,0]=(0.184957944467)/(o+1.)+(-0.217451097139)
        ref[2,0,2,1]=(-0.398642713235)/(o+1.)+(0.952207101709)
        ref[2,1,0,0]=(1.02416663872)/(o+1.)+(-0.575138524144)
        ref[2,1,0,1]=(-0.378473997252)/(o+1.)+(0.0281620686178)
        ref[2,1,1,0]=(-0.275148610588)/(o+1.)+(0.106608922446)
        ref[2,1,1,1]=(0.282017473532)/(o+1.)+(0.890800821181)
        ref[2,1,2,0]=(-1.13318416094)/(o+1.)+(-0.389382011348)
        ref[2,1,2,1]=(-0.0966355603128)/(o+1.)+(-0.0177891342927)
        ref[2,2,0,0]=(0.70200171082)/(o+1.)+(-0.744552634305)
        ref[2,2,0,1]=(0.199677737079)/(o+1.)+(0.443147776029)
        ref[2,2,1,0]=(-1.01752152133)/(o+1.)+(0.959496728669)
        ref[2,2,1,1]=(1.74270113719)/(o+1.)+(-0.223555533932)
        ref[2,2,2,0]=(-0.91408081464)/(o+1.)+(-0.378536432227)
        ref[2,2,2,1]=(0.870943027839)/(o+1.)+(-0.208988036298)
        ref[2,3,0,0]=(-0.0370292062615)/(o+1.)+(0.167974688326)
        ref[2,3,0,1]=(0.51163429561)/(o+1.)+(0.99553805486)
        ref[2,3,1,0]=(-0.677228355424)/(o+1.)+(0.642232997895)
        ref[2,3,1,1]=(1.14783423684)/(o+1.)+(0.172025186567)
        ref[2,3,2,0]=(-0.57309514119)/(o+1.)+(-0.577050078555)
        ref[2,3,2,1]=(-0.156823867369)/(o+1.)+(-0.472486242613)
        ref[2,4,0,0]=(-0.295217625165)/(o+1.)+(0.388163477102)
        ref[2,4,0,1]=(0.960307857184)/(o+1.)+(-0.91270800915)
        ref[2,4,1,0]=(0.945160124047)/(o+1.)+(-0.484173489303)
        ref[2,4,1,1]=(1.4595457308)/(o+1.)+(-0.0474937978771)
        ref[2,4,2,0]=(0.213502990044)/(o+1.)+(-0.595230402509)
        ref[2,4,2,1]=(-0.157408251134)/(o+1.)+(-0.539909645345)
        ref[3,0,0,0]=(-0.108036043722)/(o+1.)+(0.219960850586)
        ref[3,0,0,1]=(-0.961750969008)/(o+1.)+(-0.798400176406)
        ref[3,0,1,0]=(0.262323530493)/(o+1.)+(0.258382891794)
        ref[3,0,1,1]=(-1.32142360185)/(o+1.)+(-0.904677153373)
        ref[3,0,2,0]=(-1.21971286759)/(o+1.)+(-0.299620604131)
        ref[3,0,2,1]=(-0.932654333294)/(o+1.)+(-0.17287329357)
        ref[3,1,0,0]=(0.93619793774)/(o+1.)+(-1.31092419975)
        ref[3,1,0,1]=(1.50934489005)/(o+1.)+(0.682521318129)
        ref[3,1,1,0]=(0.351711340348)/(o+1.)+(-0.309637895207)
        ref[3,1,1,1]=(0.451719461341)/(o+1.)+(-0.161642891469)
        ref[3,1,2,0]=(1.22994831248)/(o+1.)+(-0.930678239419)
        ref[3,1,2,1]=(-0.679134791053)/(o+1.)+(0.151363737548)
        ref[3,2,0,0]=(-0.203508407413)/(o+1.)+(-0.0947994060691)
        ref[3,2,0,1]=(0.822166720224)/(o+1.)+(0.693828879767)
        ref[3,2,1,0]=(-2.14174725127)/(o+1.)+(-0.463720449069)
        ref[3,2,1,1]=(0.606494945308)/(o+1.)+(-0.280337399989)
        ref[3,2,2,0]=(-1.16919048525)/(o+1.)+(0.377127382228)
        ref[3,2,2,1]=(-1.06008981178)/(o+1.)+(-0.218348568361)
        ref[3,3,0,0]=(0.274951422345)/(o+1.)+(-0.0512882856318)
        ref[3,3,0,1]=(-0.484437238367)/(o+1.)+(0.25859507085)
        ref[3,3,1,0]=(-0.392619163603)/(o+1.)+(0.132176607371)
        ref[3,3,1,1]=(-1.03879123383)/(o+1.)+(0.112949977952)
        ref[3,3,2,0]=(-1.03411862752)/(o+1.)+(0.184894846825)
        ref[3,3,2,1]=(1.33879829218)/(o+1.)+(0.562966467859)
        ref[3,4,0,0]=(-1.17570659357)/(o+1.)+(-0.484769467616)
        ref[3,4,0,1]=(0.409620323287)/(o+1.)+(1.06630712898)
        ref[3,4,1,0]=(-0.0904444132627)/(o+1.)+(0.549979159103)
        ref[3,4,1,1]=(-0.716121800087)/(o+1.)+(0.295175291572)
        ref[3,4,2,0]=(1.22262956258)/(o+1.)+(0.0373581407971)
        ref[3,4,2,1]=(0.612103692201)/(o+1.)+(-0.167135601938)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.506621877215)*x[0]**o+(-0.45560383685)*x[0]+(-0.598137301806)*x[1]**o+(0.784241595615)*x[1]
        ref=(-0.0915154245913)*(1+2.*(dim-1.)/(o+1.))+(0.328637758765)*dim
      else:
        arg=(-0.505489131484)*x[0]**o+(0.58139159259)*x[0]+(0.458086828888)*x[1]**o+(0.549090746672)*x[1]+(-0.0882325936167)*x[2]**o+(0.0170295954808)*x[2]
        ref=(-0.135634896213)*(1+2.*(dim-1.)/(o+1.))+(1.14751193474)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.88779433873)*x[0]**o+(0.507955012301)*x[0]+(0.652739274924)*x[1]**o+(-0.971975755989)*x[1]
        arg[1]=(-0.0812690931139)*x[0]**o+(-0.283880962354)*x[0]+(0.094904517439)*x[1]**o+(-0.940528164353)*x[1]
        ref[0]=(-0.235055063806)*(1+2.*(dim-1.)/(o+1.))+(-0.464020743688)*dim
        ref[1]=(0.0136354243251)*(1+2.*(dim-1.)/(o+1.))+(-1.22440912671)*dim
      else:
        arg[0]=(-0.506412384389)*x[0]**o+(-0.465794405394)*x[0]+(0.346977499982)*x[1]**o+(0.386181176987)*x[1]+(-0.476729572397)*x[2]**o+(-0.956945933508)*x[2]
        arg[1]=(0.267092807416)*x[0]**o+(-0.54211155802)*x[0]+(-0.27815411727)*x[1]**o+(-0.984542450102)*x[1]+(0.656389296076)*x[2]**o+(0.0395863971401)*x[2]
        ref[0]=(-0.636164456804)*(1+2.*(dim-1.)/(o+1.))+(-1.03655916191)*dim
        ref[1]=(0.645327986222)*(1+2.*(dim-1.)/(o+1.))+(-1.48706761098)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.0607636892262)*x[0]**o+(0.740866012668)*x[0]+(0.145223136928)*x[1]**o+(-0.738387773526)*x[1]
        arg[0,1]=(0.0292598484511)*x[0]**o+(-0.189208290889)*x[0]+(0.902021405165)*x[1]**o+(0.423106545559)*x[1]
        arg[0,2]=(0.519658217226)*x[0]**o+(-0.768406018136)*x[0]+(-0.189532527097)*x[1]**o+(0.256417429833)*x[1]
        arg[0,3]=(-0.565327663481)*x[0]**o+(0.876762003673)*x[0]+(-0.243804758063)*x[1]**o+(0.399476036454)*x[1]
        arg[0,4]=(0.0632252409818)*x[0]**o+(0.0951130662538)*x[0]+(0.185662425763)*x[1]**o+(0.907446482506)*x[1]
        arg[1,0]=(-0.430504584794)*x[0]**o+(0.864814314023)*x[0]+(0.47460095501)*x[1]**o+(-0.599679142188)*x[1]
        arg[1,1]=(0.699618175298)*x[0]**o+(-0.561782605517)*x[0]+(-0.304250865015)*x[1]**o+(-0.341836776795)*x[1]
        arg[1,2]=(0.856956767316)*x[0]**o+(0.610026903213)*x[0]+(-0.666749954599)*x[1]**o+(0.885388885234)*x[1]
        arg[1,3]=(0.699377007631)*x[0]**o+(0.699906750287)*x[0]+(0.486690669414)*x[1]**o+(0.167425901549)*x[1]
        arg[1,4]=(0.466701415896)*x[0]**o+(-0.0719752759126)*x[0]+(0.795662047062)*x[1]**o+(0.979862503926)*x[1]
        arg[2,0]=(-0.912218732612)*x[0]**o+(0.950990472894)*x[0]+(-0.527999708902)*x[1]**o+(0.695735593043)*x[1]
        arg[2,1]=(-0.223468270048)*x[0]**o+(-0.882565147673)*x[0]+(0.750516681645)*x[1]**o+(-0.652794590093)*x[1]
        arg[2,2]=(0.897646682412)*x[0]**o+(-0.815401214058)*x[0]+(0.629680741327)*x[1]**o+(-0.9716988156)*x[1]
        arg[2,3]=(0.279739517524)*x[0]**o+(-0.0859797206016)*x[0]+(-0.0454325622757)*x[1]**o+(0.301214753868)*x[1]
        arg[2,4]=(-0.358056827092)*x[0]**o+(0.254207762658)*x[0]+(0.831499219864)*x[1]**o+(0.0930388028811)*x[1]
        arg[3,0]=(0.0761520020047)*x[0]**o+(0.650696394486)*x[0]+(-0.132369660678)*x[1]**o+(-0.321220385971)*x[1]
        arg[3,1]=(0.238715231585)*x[0]**o+(-0.297219391083)*x[0]+(-0.15446552614)*x[1]**o+(0.860560615056)*x[1]
        arg[3,2]=(0.708864754538)*x[0]**o+(-0.221186702958)*x[0]+(-0.478127729891)*x[1]**o+(0.0309604412212)*x[1]
        arg[3,3]=(0.362909366479)*x[0]**o+(0.382279106813)*x[0]+(0.827905062086)*x[1]**o+(-0.303048355949)*x[1]
        arg[3,4]=(0.573736946844)*x[0]**o+(0.645059764766)*x[0]+(0.301945383678)*x[1]**o+(-0.103083649994)*x[1]
        ref[0,0]=(0.205986826154)*(1+2.*(dim-1.)/(o+1.))+(0.0024782391428)*dim
        ref[0,1]=(0.931281253617)*(1+2.*(dim-1.)/(o+1.))+(0.23389825467)*dim
        ref[0,2]=(0.330125690129)*(1+2.*(dim-1.)/(o+1.))+(-0.511988588303)*dim
        ref[0,3]=(-0.809132421544)*(1+2.*(dim-1.)/(o+1.))+(1.27623804013)*dim
        ref[0,4]=(0.248887666745)*(1+2.*(dim-1.)/(o+1.))+(1.00255954876)*dim
        ref[1,0]=(0.044096370216)*(1+2.*(dim-1.)/(o+1.))+(0.265135171835)*dim
        ref[1,1]=(0.395367310283)*(1+2.*(dim-1.)/(o+1.))+(-0.903619382312)*dim
        ref[1,2]=(0.190206812717)*(1+2.*(dim-1.)/(o+1.))+(1.49541578845)*dim
        ref[1,3]=(1.18606767705)*(1+2.*(dim-1.)/(o+1.))+(0.867332651836)*dim
        ref[1,4]=(1.26236346296)*(1+2.*(dim-1.)/(o+1.))+(0.907887228013)*dim
        ref[2,0]=(-1.44021844151)*(1+2.*(dim-1.)/(o+1.))+(1.64672606594)*dim
        ref[2,1]=(0.527048411596)*(1+2.*(dim-1.)/(o+1.))+(-1.53535973777)*dim
        ref[2,2]=(1.52732742374)*(1+2.*(dim-1.)/(o+1.))+(-1.78710002966)*dim
        ref[2,3]=(0.234306955249)*(1+2.*(dim-1.)/(o+1.))+(0.215235033266)*dim
        ref[2,4]=(0.473442392772)*(1+2.*(dim-1.)/(o+1.))+(0.34724656554)*dim
        ref[3,0]=(-0.0562176586729)*(1+2.*(dim-1.)/(o+1.))+(0.329476008516)*dim
        ref[3,1]=(0.0842497054449)*(1+2.*(dim-1.)/(o+1.))+(0.563341223973)*dim
        ref[3,2]=(0.230737024647)*(1+2.*(dim-1.)/(o+1.))+(-0.190226261737)*dim
        ref[3,3]=(1.19081442856)*(1+2.*(dim-1.)/(o+1.))+(0.079230750864)*dim
        ref[3,4]=(0.875682330522)*(1+2.*(dim-1.)/(o+1.))+(0.541976114772)*dim
      else:
        arg[0,0]=(0.0640108068236)*x[0]**o+(0.0322830166043)*x[0]+(-0.703691335255)*x[1]**o+(-0.185228770711)*x[1]+(-0.738714166313)*x[2]**o+(-0.423576647077)*x[2]
        arg[0,1]=(-0.738341351945)*x[0]**o+(0.396306395436)*x[0]+(-0.369974848568)*x[1]**o+(0.750933208783)*x[1]+(0.554069869127)*x[2]**o+(0.634275493051)*x[2]
        arg[0,2]=(-0.259193248708)*x[0]**o+(-0.900007449583)*x[0]+(0.894379436915)*x[1]**o+(0.997814320752)*x[1]+(0.233192116622)*x[2]**o+(0.986616259649)*x[2]
        arg[0,3]=(0.381480250594)*x[0]**o+(0.734823219057)*x[0]+(0.569912440169)*x[1]**o+(-0.441084504695)*x[1]+(0.312539701584)*x[2]**o+(-0.3733004974)*x[2]
        arg[0,4]=(-0.779059807622)*x[0]**o+(0.538777922322)*x[0]+(-0.0695879211612)*x[1]**o+(-0.204862216499)*x[1]+(-0.0109698525672)*x[2]**o+(0.981875117617)*x[2]
        arg[1,0]=(-0.415857630044)*x[0]**o+(-0.356607866913)*x[0]+(-0.628854326903)*x[1]**o+(0.0401233333505)*x[1]+(0.687470169579)*x[2]**o+(-0.565349753114)*x[2]
        arg[1,1]=(-0.438364910402)*x[0]**o+(0.498078194367)*x[0]+(-0.120729597948)*x[1]**o+(-0.622076117713)*x[1]+(0.995130029172)*x[2]**o+(-0.7234601064)*x[2]
        arg[1,2]=(-0.957869585558)*x[0]**o+(0.374693219824)*x[0]+(0.0355192120972)*x[1]**o+(-0.516770140651)*x[1]+(0.521887953791)*x[2]**o+(0.00771564236502)*x[2]
        arg[1,3]=(-0.0495580194062)*x[0]**o+(-0.183895326318)*x[0]+(-0.55304086303)*x[1]**o+(0.920204449879)*x[1]+(0.143165855157)*x[2]**o+(-0.773684620462)*x[2]
        arg[1,4]=(0.795913131087)*x[0]**o+(0.653826129199)*x[0]+(-0.585562760813)*x[1]**o+(0.274083108918)*x[1]+(-0.165275654116)*x[2]**o+(0.358678758212)*x[2]
        arg[2,0]=(-0.0699685857843)*x[0]**o+(-0.42922435931)*x[0]+(-0.0494308981938)*x[1]**o+(-0.956597829176)*x[1]+(-0.912941305307)*x[2]**o+(0.214777332562)*x[2]
        arg[2,1]=(-0.900983338173)*x[0]**o+(-0.779729884378)*x[0]+(-0.860100306508)*x[1]**o+(0.14211537281)*x[1]+(-0.984439938285)*x[2]**o+(-0.549987779842)*x[2]
        arg[2,2]=(0.801489512389)*x[0]**o+(0.896468923826)*x[0]+(-0.155737972887)*x[1]**o+(0.639864330477)*x[1]+(0.798598705038)*x[2]**o+(-0.863019041525)*x[2]
        arg[2,3]=(-0.453612136504)*x[0]**o+(0.522662491688)*x[0]+(-0.463058076486)*x[1]**o+(0.398018809499)*x[1]+(-0.612044961611)*x[2]**o+(-0.481951883217)*x[2]
        arg[2,4]=(-0.0739336519425)*x[0]**o+(-0.241753740154)*x[0]+(0.0176292058898)*x[1]**o+(0.435641132809)*x[1]+(-0.0592828584976)*x[2]**o+(0.430115620801)*x[2]
        arg[3,0]=(0.604784358152)*x[0]**o+(-0.817214179503)*x[0]+(-0.423104272008)*x[1]**o+(0.287787881773)*x[1]+(0.678036469241)*x[2]**o+(0.578347129563)*x[2]
        arg[3,1]=(0.362390778598)*x[0]**o+(-0.389066580119)*x[0]+(-0.934087580396)*x[1]**o+(0.907587485226)*x[1]+(-0.902083043406)*x[2]**o+(0.904707538314)*x[2]
        arg[3,2]=(0.862463207465)*x[0]**o+(0.904923219336)*x[0]+(0.421361696042)*x[1]**o+(-0.779329196171)*x[1]+(0.059521522394)*x[2]**o+(-0.744510067387)*x[2]
        arg[3,3]=(0.412029226851)*x[0]**o+(-0.649730143511)*x[0]+(0.579628017001)*x[1]**o+(-0.949180147129)*x[1]+(-0.073016492309)*x[2]**o+(0.666554170757)*x[2]
        arg[3,4]=(-0.700401396504)*x[0]**o+(-0.89194120812)*x[0]+(0.592387042213)*x[1]**o+(-0.906632787891)*x[1]+(-0.486072860812)*x[2]**o+(-0.505386176523)*x[2]
        ref[0,0]=(-1.37839469474)*(1+2.*(dim-1.)/(o+1.))+(-0.576522401184)*dim
        ref[0,1]=(-0.554246331386)*(1+2.*(dim-1.)/(o+1.))+(1.78151509727)*dim
        ref[0,2]=(0.868378304829)*(1+2.*(dim-1.)/(o+1.))+(1.08442313082)*dim
        ref[0,3]=(1.26393239235)*(1+2.*(dim-1.)/(o+1.))+(-0.0795617830385)*dim
        ref[0,4]=(-0.85961758135)*(1+2.*(dim-1.)/(o+1.))+(1.31579082344)*dim
        ref[1,0]=(-0.357241787368)*(1+2.*(dim-1.)/(o+1.))+(-0.881834286676)*dim
        ref[1,1]=(0.436035520821)*(1+2.*(dim-1.)/(o+1.))+(-0.847458029746)*dim
        ref[1,2]=(-0.40046241967)*(1+2.*(dim-1.)/(o+1.))+(-0.134361278462)*dim
        ref[1,3]=(-0.459433027279)*(1+2.*(dim-1.)/(o+1.))+(-0.0373754969011)*dim
        ref[1,4]=(0.0450747161574)*(1+2.*(dim-1.)/(o+1.))+(1.28658799633)*dim
        ref[2,0]=(-1.03234078929)*(1+2.*(dim-1.)/(o+1.))+(-1.17104485592)*dim
        ref[2,1]=(-2.74552358297)*(1+2.*(dim-1.)/(o+1.))+(-1.18760229141)*dim
        ref[2,2]=(1.44435024454)*(1+2.*(dim-1.)/(o+1.))+(0.673314212777)*dim
        ref[2,3]=(-1.5287151746)*(1+2.*(dim-1.)/(o+1.))+(0.43872941797)*dim
        ref[2,4]=(-0.11558730455)*(1+2.*(dim-1.)/(o+1.))+(0.624003013456)*dim
        ref[3,0]=(0.859716555385)*(1+2.*(dim-1.)/(o+1.))+(0.0489208318321)*dim
        ref[3,1]=(-1.4737798452)*(1+2.*(dim-1.)/(o+1.))+(1.42322844342)*dim
        ref[3,2]=(1.3433464259)*(1+2.*(dim-1.)/(o+1.))+(-0.618916044222)*dim
        ref[3,3]=(0.918640751542)*(1+2.*(dim-1.)/(o+1.))+(-0.932356119883)*dim
        ref[3,4]=(-0.594087215103)*(1+2.*(dim-1.)/(o+1.))+(-2.30396017253)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.662707930283)*x[0]**o+(0.625512077969)*x[0]+(-0.222208840922)*x[1]**o+(0.374614317804)*x[1]
        arg[0,0,1]=(0.685952252808)*x[0]**o+(-0.88219802797)*x[0]+(-0.948918831634)*x[1]**o+(-0.00615202896776)*x[1]
        arg[0,1,0]=(-0.555231206657)*x[0]**o+(0.997112425959)*x[0]+(-0.186247870511)*x[1]**o+(-0.0658892046229)*x[1]
        arg[0,1,1]=(-0.735288285178)*x[0]**o+(0.941234848614)*x[0]+(-0.307761737256)*x[1]**o+(0.933293069433)*x[1]
        arg[1,0,0]=(-0.618191337825)*x[0]**o+(-0.12930895166)*x[0]+(0.76213325083)*x[1]**o+(-0.744049403156)*x[1]
        arg[1,0,1]=(-0.793412570981)*x[0]**o+(0.533175323401)*x[0]+(-0.87398581932)*x[1]**o+(-0.279651754725)*x[1]
        arg[1,1,0]=(-0.0654077907694)*x[0]**o+(0.239040458488)*x[0]+(0.730639927837)*x[1]**o+(0.805020533329)*x[1]
        arg[1,1,1]=(-0.560195748542)*x[0]**o+(-0.265389430135)*x[0]+(-0.295142324741)*x[1]**o+(-0.929699771024)*x[1]
        arg[2,0,0]=(0.935760865577)*x[0]**o+(-0.332458843131)*x[0]+(0.257302695187)*x[1]**o+(-0.347547927033)*x[1]
        arg[2,0,1]=(-0.245770653276)*x[0]**o+(0.126945988409)*x[0]+(0.302793434995)*x[1]**o+(-0.997483200172)*x[1]
        arg[2,1,0]=(0.523494364951)*x[0]**o+(-0.323782740449)*x[0]+(-0.563752593454)*x[1]**o+(-0.358325752838)*x[1]
        arg[2,1,1]=(-0.506330114515)*x[0]**o+(-0.733234050567)*x[0]+(0.734774667778)*x[1]**o+(0.795264566776)*x[1]
        arg[3,0,0]=(-0.21258295235)*x[0]**o+(-0.929331330718)*x[0]+(-0.546061847742)*x[1]**o+(0.0613931491065)*x[1]
        arg[3,0,1]=(-0.357847946656)*x[0]**o+(-0.133795684508)*x[0]+(-0.796646028688)*x[1]**o+(0.875795801429)*x[1]
        arg[3,1,0]=(0.923717804039)*x[0]**o+(-0.176266633621)*x[0]+(-0.605892867534)*x[1]**o+(0.633377612365)*x[1]
        arg[3,1,1]=(0.729025508129)*x[0]**o+(0.412034955257)*x[0]+(0.0933054864485)*x[1]**o+(0.386019085532)*x[1]
        arg[4,0,0]=(-0.0705524918008)*x[0]**o+(-0.869605321304)*x[0]+(0.244351267755)*x[1]**o+(0.898883388102)*x[1]
        arg[4,0,1]=(-0.453297333327)*x[0]**o+(-0.67824185123)*x[0]+(0.0497322840245)*x[1]**o+(0.581677879069)*x[1]
        arg[4,1,0]=(0.343975006358)*x[0]**o+(0.724084357103)*x[0]+(-0.186292335362)*x[1]**o+(-0.257997075018)*x[1]
        arg[4,1,1]=(-0.389731923798)*x[0]**o+(0.0949155578616)*x[0]+(-0.298428138752)*x[1]**o+(0.819510069666)*x[1]
        arg[5,0,0]=(0.827670169513)*x[0]**o+(-0.584607978668)*x[0]+(0.669056948658)*x[1]**o+(-0.393158541985)*x[1]
        arg[5,0,1]=(-0.471551968206)*x[0]**o+(-0.339964831396)*x[0]+(0.354554393554)*x[1]**o+(-0.334773591957)*x[1]
        arg[5,1,0]=(0.113562904864)*x[0]**o+(-0.775045505123)*x[0]+(0.783044874931)*x[1]**o+(0.52265843873)*x[1]
        arg[5,1,1]=(-0.115476036205)*x[0]**o+(-0.290822012728)*x[0]+(-0.398510477854)*x[1]**o+(0.398737205542)*x[1]
        ref[0,0,0]=(0.440499089362)*(1+2.*(dim-1.)/(o+1.))+(1.00012639577)*dim
        ref[0,0,1]=(-0.262966578826)*(1+2.*(dim-1.)/(o+1.))+(-0.888350056938)*dim
        ref[0,1,0]=(-0.741479077167)*(1+2.*(dim-1.)/(o+1.))+(0.931223221336)*dim
        ref[0,1,1]=(-1.04305002243)*(1+2.*(dim-1.)/(o+1.))+(1.87452791805)*dim
        ref[1,0,0]=(0.143941913005)*(1+2.*(dim-1.)/(o+1.))+(-0.873358354816)*dim
        ref[1,0,1]=(-1.6673983903)*(1+2.*(dim-1.)/(o+1.))+(0.253523568676)*dim
        ref[1,1,0]=(0.665232137068)*(1+2.*(dim-1.)/(o+1.))+(1.04406099182)*dim
        ref[1,1,1]=(-0.855338073283)*(1+2.*(dim-1.)/(o+1.))+(-1.19508920116)*dim
        ref[2,0,0]=(1.19306356076)*(1+2.*(dim-1.)/(o+1.))+(-0.680006770164)*dim
        ref[2,0,1]=(0.0570227817187)*(1+2.*(dim-1.)/(o+1.))+(-0.870537211764)*dim
        ref[2,1,0]=(-0.0402582285028)*(1+2.*(dim-1.)/(o+1.))+(-0.682108493287)*dim
        ref[2,1,1]=(0.228444553263)*(1+2.*(dim-1.)/(o+1.))+(0.0620305162089)*dim
        ref[3,0,0]=(-0.758644800093)*(1+2.*(dim-1.)/(o+1.))+(-0.867938181611)*dim
        ref[3,0,1]=(-1.15449397534)*(1+2.*(dim-1.)/(o+1.))+(0.742000116921)*dim
        ref[3,1,0]=(0.317824936505)*(1+2.*(dim-1.)/(o+1.))+(0.457110978744)*dim
        ref[3,1,1]=(0.822330994578)*(1+2.*(dim-1.)/(o+1.))+(0.798054040789)*dim
        ref[4,0,0]=(0.173798775954)*(1+2.*(dim-1.)/(o+1.))+(0.0292780667983)*dim
        ref[4,0,1]=(-0.403565049303)*(1+2.*(dim-1.)/(o+1.))+(-0.0965639721601)*dim
        ref[4,1,0]=(0.157682670995)*(1+2.*(dim-1.)/(o+1.))+(0.466087282085)*dim
        ref[4,1,1]=(-0.68816006255)*(1+2.*(dim-1.)/(o+1.))+(0.914425627528)*dim
        ref[5,0,0]=(1.49672711817)*(1+2.*(dim-1.)/(o+1.))+(-0.977766520654)*dim
        ref[5,0,1]=(-0.116997574651)*(1+2.*(dim-1.)/(o+1.))+(-0.674738423353)*dim
        ref[5,1,0]=(0.896607779795)*(1+2.*(dim-1.)/(o+1.))+(-0.252387066394)*dim
        ref[5,1,1]=(-0.513986514058)*(1+2.*(dim-1.)/(o+1.))+(0.107915192814)*dim
      else:
        arg[0,0,0]=(-0.418671467289)*x[0]**o+(0.498387164484)*x[0]+(0.208896299314)*x[1]**o+(-0.643038678771)*x[1]+(-0.85475466439)*x[2]**o+(0.59586368009)*x[2]
        arg[0,0,1]=(0.149229891253)*x[0]**o+(0.229836799955)*x[0]+(0.557946253001)*x[1]**o+(-0.434223003739)*x[1]+(-0.250979831671)*x[2]**o+(0.762582320269)*x[2]
        arg[0,1,0]=(-0.950902166094)*x[0]**o+(-0.312913424444)*x[0]+(-0.573887973428)*x[1]**o+(-0.372485255069)*x[1]+(-0.18228472562)*x[2]**o+(-0.566679093436)*x[2]
        arg[0,1,1]=(-0.191769651257)*x[0]**o+(-0.344504306057)*x[0]+(0.355926399715)*x[1]**o+(-0.721341598375)*x[1]+(0.362401271289)*x[2]**o+(0.601351093654)*x[2]
        arg[1,0,0]=(-0.375572370375)*x[0]**o+(-0.532723139333)*x[0]+(-0.738166755424)*x[1]**o+(0.785506757134)*x[1]+(0.192667976423)*x[2]**o+(-0.06534893872)*x[2]
        arg[1,0,1]=(-0.353750712453)*x[0]**o+(0.401872866285)*x[0]+(-0.747783222746)*x[1]**o+(0.663116752574)*x[1]+(-0.162954382466)*x[2]**o+(0.318112656687)*x[2]
        arg[1,1,0]=(-0.711345828912)*x[0]**o+(0.764701438614)*x[0]+(-0.232576947703)*x[1]**o+(-0.50541791533)*x[1]+(0.604603682696)*x[2]**o+(0.498828792648)*x[2]
        arg[1,1,1]=(-0.72474879447)*x[0]**o+(0.888266999605)*x[0]+(0.639903282531)*x[1]**o+(-0.777435844156)*x[1]+(-0.0814738680964)*x[2]**o+(-0.0135286853237)*x[2]
        arg[2,0,0]=(0.839793308998)*x[0]**o+(0.374051014226)*x[0]+(0.594653957466)*x[1]**o+(-0.962352782345)*x[1]+(0.888632514336)*x[2]**o+(0.491195067464)*x[2]
        arg[2,0,1]=(-0.827615661598)*x[0]**o+(0.675227575488)*x[0]+(-0.93068580658)*x[1]**o+(0.694434337954)*x[1]+(0.3583179263)*x[2]**o+(0.775147111187)*x[2]
        arg[2,1,0]=(-0.772656999491)*x[0]**o+(-0.464476706435)*x[0]+(0.569832694977)*x[1]**o+(0.500517676542)*x[1]+(-0.272153062279)*x[2]**o+(-0.398436498506)*x[2]
        arg[2,1,1]=(-0.3065198839)*x[0]**o+(-0.454353346315)*x[0]+(-0.539448433869)*x[1]**o+(0.776950547078)*x[1]+(-0.987178147402)*x[2]**o+(0.368935549126)*x[2]
        arg[3,0,0]=(-0.516131622594)*x[0]**o+(-0.680922252714)*x[0]+(0.304951059501)*x[1]**o+(-0.236361364379)*x[1]+(-0.37983306233)*x[2]**o+(0.693426329316)*x[2]
        arg[3,0,1]=(0.85094983056)*x[0]**o+(-0.356782375867)*x[0]+(0.362116478047)*x[1]**o+(0.269126381543)*x[1]+(0.950254707197)*x[2]**o+(0.592028802301)*x[2]
        arg[3,1,0]=(-0.17721437307)*x[0]**o+(-0.127432965766)*x[0]+(-0.754975923735)*x[1]**o+(0.755938625221)*x[1]+(0.776393089777)*x[2]**o+(0.913028629874)*x[2]
        arg[3,1,1]=(-0.779840479711)*x[0]**o+(0.432392796542)*x[0]+(0.0457670482312)*x[1]**o+(0.493363120993)*x[1]+(-0.290308698356)*x[2]**o+(0.263565329052)*x[2]
        arg[4,0,0]=(-0.594371032991)*x[0]**o+(0.673573089348)*x[0]+(0.0301238273919)*x[1]**o+(-0.900766544194)*x[1]+(0.937125579182)*x[2]**o+(-0.360512287398)*x[2]
        arg[4,0,1]=(-0.187561420369)*x[0]**o+(-0.727058667406)*x[0]+(0.643904893485)*x[1]**o+(0.547358609515)*x[1]+(-0.810276689361)*x[2]**o+(0.347967967049)*x[2]
        arg[4,1,0]=(0.0261452776327)*x[0]**o+(0.630125453041)*x[0]+(0.884509651558)*x[1]**o+(-0.580251885597)*x[1]+(0.141608918047)*x[2]**o+(0.417140237181)*x[2]
        arg[4,1,1]=(0.943149763428)*x[0]**o+(0.0414421588123)*x[0]+(-0.453944232191)*x[1]**o+(0.817963360891)*x[1]+(-0.0132464046434)*x[2]**o+(-0.247024507218)*x[2]
        arg[5,0,0]=(0.993356103007)*x[0]**o+(-0.604991443085)*x[0]+(-0.0141942193536)*x[1]**o+(-0.467982818195)*x[1]+(0.907608332852)*x[2]**o+(-0.918494340721)*x[2]
        arg[5,0,1]=(0.782079295393)*x[0]**o+(0.0125642236587)*x[0]+(-0.512764292821)*x[1]**o+(-0.0699705327379)*x[1]+(0.141341885735)*x[2]**o+(0.420508183565)*x[2]
        arg[5,1,0]=(-0.835665486427)*x[0]**o+(-0.536850925449)*x[0]+(-0.0958499285885)*x[1]**o+(-0.541119926637)*x[1]+(0.341886033336)*x[2]**o+(-0.772823771771)*x[2]
        arg[5,1,1]=(-0.391737530391)*x[0]**o+(-0.160045087084)*x[0]+(-0.984350699381)*x[1]**o+(0.303782477449)*x[1]+(-0.279906615513)*x[2]**o+(-0.690872289955)*x[2]
        ref[0,0,0]=(-1.06452983237)*(1+2.*(dim-1.)/(o+1.))+(0.451212165803)*dim
        ref[0,0,1]=(0.456196312583)*(1+2.*(dim-1.)/(o+1.))+(0.558196116486)*dim
        ref[0,1,0]=(-1.70707486514)*(1+2.*(dim-1.)/(o+1.))+(-1.25207777295)*dim
        ref[0,1,1]=(0.526558019747)*(1+2.*(dim-1.)/(o+1.))+(-0.464494810779)*dim
        ref[1,0,0]=(-0.921071149376)*(1+2.*(dim-1.)/(o+1.))+(0.187434679081)*dim
        ref[1,0,1]=(-1.26448831766)*(1+2.*(dim-1.)/(o+1.))+(1.38310227555)*dim
        ref[1,1,0]=(-0.339319093919)*(1+2.*(dim-1.)/(o+1.))+(0.758112315932)*dim
        ref[1,1,1]=(-0.166319380035)*(1+2.*(dim-1.)/(o+1.))+(0.0973024701251)*dim
        ref[2,0,0]=(2.3230797808)*(1+2.*(dim-1.)/(o+1.))+(-0.0971067006546)*dim
        ref[2,0,1]=(-1.39998354188)*(1+2.*(dim-1.)/(o+1.))+(2.14480902463)*dim
        ref[2,1,0]=(-0.474977366793)*(1+2.*(dim-1.)/(o+1.))+(-0.362395528399)*dim
        ref[2,1,1]=(-1.83314646517)*(1+2.*(dim-1.)/(o+1.))+(0.69153274989)*dim
        ref[3,0,0]=(-0.591013625423)*(1+2.*(dim-1.)/(o+1.))+(-0.223857287777)*dim
        ref[3,0,1]=(2.1633210158)*(1+2.*(dim-1.)/(o+1.))+(0.504372807977)*dim
        ref[3,1,0]=(-0.155797207029)*(1+2.*(dim-1.)/(o+1.))+(1.54153428933)*dim
        ref[3,1,1]=(-1.02438212984)*(1+2.*(dim-1.)/(o+1.))+(1.18932124659)*dim
        ref[4,0,0]=(0.372878373583)*(1+2.*(dim-1.)/(o+1.))+(-0.587705742243)*dim
        ref[4,0,1]=(-0.353933216245)*(1+2.*(dim-1.)/(o+1.))+(0.168267909158)*dim
        ref[4,1,0]=(1.05226384724)*(1+2.*(dim-1.)/(o+1.))+(0.467013804625)*dim
        ref[4,1,1]=(0.475959126593)*(1+2.*(dim-1.)/(o+1.))+(0.612381012486)*dim
        ref[5,0,0]=(1.88677021651)*(1+2.*(dim-1.)/(o+1.))+(-1.991468602)*dim
        ref[5,0,1]=(0.410656888307)*(1+2.*(dim-1.)/(o+1.))+(0.363101874486)*dim
        ref[5,1,0]=(-0.58962938168)*(1+2.*(dim-1.)/(o+1.))+(-1.85079462386)*dim
        ref[5,1,1]=(-1.65599484529)*(1+2.*(dim-1.)/(o+1.))+(-0.54713489959)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.606358724075)*x[0]**o+(-0.6326664687)*x[0]+(-0.521101256804)*x[1]**o+(0.702113497865)*x[1]
        arg[0,0,0,1]=(0.528315511137)*x[0]**o+(-0.116566989978)*x[0]+(-0.350847384045)*x[1]**o+(-0.581988878345)*x[1]
        arg[0,0,1,0]=(0.565076347032)*x[0]**o+(0.752532782601)*x[0]+(0.00706163321089)*x[1]**o+(-0.89265244135)*x[1]
        arg[0,0,1,1]=(-0.342136497951)*x[0]**o+(0.753153993866)*x[0]+(-0.199643944066)*x[1]**o+(0.995411752312)*x[1]
        arg[0,0,2,0]=(0.621283193015)*x[0]**o+(0.591186318333)*x[0]+(0.880169584588)*x[1]**o+(-0.11839698224)*x[1]
        arg[0,0,2,1]=(-0.831745199475)*x[0]**o+(0.293656166596)*x[0]+(0.0717745661823)*x[1]**o+(0.827680695965)*x[1]
        arg[0,1,0,0]=(0.686864676707)*x[0]**o+(-0.6552976723)*x[0]+(-0.65589888943)*x[1]**o+(-0.878054891784)*x[1]
        arg[0,1,0,1]=(-0.43359086203)*x[0]**o+(0.723819586663)*x[0]+(0.0741799035834)*x[1]**o+(0.13005305188)*x[1]
        arg[0,1,1,0]=(-0.0605346127445)*x[0]**o+(-0.717729061173)*x[0]+(0.825539660518)*x[1]**o+(0.873033938391)*x[1]
        arg[0,1,1,1]=(-0.286483607332)*x[0]**o+(-0.577255191498)*x[0]+(-0.328362280683)*x[1]**o+(0.499078653745)*x[1]
        arg[0,1,2,0]=(-0.334132932545)*x[0]**o+(0.462461110225)*x[0]+(-0.932311565873)*x[1]**o+(0.934683608666)*x[1]
        arg[0,1,2,1]=(-0.383203086272)*x[0]**o+(0.261035781624)*x[0]+(-0.203936125231)*x[1]**o+(0.657584310755)*x[1]
        arg[0,2,0,0]=(-0.637642146695)*x[0]**o+(0.80524493224)*x[0]+(-0.139137135459)*x[1]**o+(0.947434162736)*x[1]
        arg[0,2,0,1]=(-0.0738996477961)*x[0]**o+(0.0663360260869)*x[0]+(-0.599113718918)*x[1]**o+(-0.349514990363)*x[1]
        arg[0,2,1,0]=(-0.0539211250603)*x[0]**o+(0.262567947534)*x[0]+(0.115386681137)*x[1]**o+(-0.68179682822)*x[1]
        arg[0,2,1,1]=(-0.770114869627)*x[0]**o+(-0.388068404619)*x[0]+(-0.712914716647)*x[1]**o+(-0.810536809871)*x[1]
        arg[0,2,2,0]=(0.274425895455)*x[0]**o+(0.338467327995)*x[0]+(-0.78223636363)*x[1]**o+(0.406149090351)*x[1]
        arg[0,2,2,1]=(-0.748672361572)*x[0]**o+(0.508311939723)*x[0]+(0.875797107098)*x[1]**o+(0.0992795348726)*x[1]
        arg[0,3,0,0]=(-0.547350688209)*x[0]**o+(0.671683747412)*x[0]+(-0.0954212965969)*x[1]**o+(0.983487887304)*x[1]
        arg[0,3,0,1]=(0.626104338883)*x[0]**o+(-0.825311449834)*x[0]+(-0.422521614246)*x[1]**o+(0.304208397592)*x[1]
        arg[0,3,1,0]=(0.698777992012)*x[0]**o+(0.00291670095874)*x[0]+(0.327527809646)*x[1]**o+(0.283289333886)*x[1]
        arg[0,3,1,1]=(-0.0197648150597)*x[0]**o+(-0.902860814776)*x[0]+(0.663961228673)*x[1]**o+(0.675952386365)*x[1]
        arg[0,3,2,0]=(0.286747814345)*x[0]**o+(-0.0641946364041)*x[0]+(0.812019228115)*x[1]**o+(-0.225698609001)*x[1]
        arg[0,3,2,1]=(0.879776322226)*x[0]**o+(-0.251856143047)*x[0]+(-0.252403834757)*x[1]**o+(0.650790406323)*x[1]
        arg[0,4,0,0]=(-0.58782572388)*x[0]**o+(-0.144122228965)*x[0]+(0.521707348991)*x[1]**o+(-0.0460840614459)*x[1]
        arg[0,4,0,1]=(-0.777497837497)*x[0]**o+(-0.229011563089)*x[0]+(0.754657080443)*x[1]**o+(-0.690675521089)*x[1]
        arg[0,4,1,0]=(0.351528557097)*x[0]**o+(0.00192986530026)*x[0]+(-0.414849484269)*x[1]**o+(-0.250205659505)*x[1]
        arg[0,4,1,1]=(-0.814037796843)*x[0]**o+(0.00188512986627)*x[0]+(-0.810799710998)*x[1]**o+(-0.114406174711)*x[1]
        arg[0,4,2,0]=(-0.755511212878)*x[0]**o+(-0.502206951212)*x[0]+(0.555738333365)*x[1]**o+(-0.220359547926)*x[1]
        arg[0,4,2,1]=(0.247108894345)*x[0]**o+(0.479239235148)*x[0]+(0.663958827816)*x[1]**o+(0.220938477625)*x[1]
        arg[1,0,0,0]=(0.818774052595)*x[0]**o+(-0.651584728881)*x[0]+(-0.780674092348)*x[1]**o+(-0.397277502644)*x[1]
        arg[1,0,0,1]=(0.482101508249)*x[0]**o+(-0.523618202277)*x[0]+(0.806406836488)*x[1]**o+(-0.883738297664)*x[1]
        arg[1,0,1,0]=(0.272517667346)*x[0]**o+(-0.700461889638)*x[0]+(0.280736354342)*x[1]**o+(0.182983247089)*x[1]
        arg[1,0,1,1]=(0.566589111451)*x[0]**o+(-0.876744547538)*x[0]+(-0.934741703566)*x[1]**o+(-0.562770667492)*x[1]
        arg[1,0,2,0]=(-0.954128714623)*x[0]**o+(0.115699770127)*x[0]+(-0.580663262426)*x[1]**o+(-0.451619694161)*x[1]
        arg[1,0,2,1]=(0.268605150185)*x[0]**o+(-0.233415174179)*x[0]+(0.477143801066)*x[1]**o+(0.443696994244)*x[1]
        arg[1,1,0,0]=(-0.00240846686485)*x[0]**o+(0.776336851643)*x[0]+(0.488314964094)*x[1]**o+(-0.840196476788)*x[1]
        arg[1,1,0,1]=(-0.511732155638)*x[0]**o+(0.528782039485)*x[0]+(0.51274502262)*x[1]**o+(-0.700213610705)*x[1]
        arg[1,1,1,0]=(0.300359000499)*x[0]**o+(0.405101876401)*x[0]+(-0.47022583299)*x[1]**o+(-0.977634916267)*x[1]
        arg[1,1,1,1]=(-0.494482318948)*x[0]**o+(0.254112728167)*x[0]+(-0.541672595343)*x[1]**o+(-0.270970064522)*x[1]
        arg[1,1,2,0]=(-0.586624294233)*x[0]**o+(0.629601920904)*x[0]+(-0.525092252936)*x[1]**o+(-0.44611635978)*x[1]
        arg[1,1,2,1]=(0.716927085139)*x[0]**o+(0.700193925267)*x[0]+(0.0624921446647)*x[1]**o+(0.109485129877)*x[1]
        arg[1,2,0,0]=(0.325729307898)*x[0]**o+(0.773432634097)*x[0]+(-0.675612597356)*x[1]**o+(-0.0554913733692)*x[1]
        arg[1,2,0,1]=(-0.381606382262)*x[0]**o+(-0.689040382854)*x[0]+(-0.580191973398)*x[1]**o+(0.951689147387)*x[1]
        arg[1,2,1,0]=(-0.527642403744)*x[0]**o+(-0.936915203927)*x[0]+(0.631733997289)*x[1]**o+(-0.715975680015)*x[1]
        arg[1,2,1,1]=(0.0886898373323)*x[0]**o+(-0.0943320873915)*x[0]+(0.470668750443)*x[1]**o+(0.773832960093)*x[1]
        arg[1,2,2,0]=(-0.727720290197)*x[0]**o+(0.422119803173)*x[0]+(0.705882951779)*x[1]**o+(-0.350801865244)*x[1]
        arg[1,2,2,1]=(0.879335475407)*x[0]**o+(-0.0307168471577)*x[0]+(0.458578151544)*x[1]**o+(0.788796576339)*x[1]
        arg[1,3,0,0]=(0.182143508146)*x[0]**o+(-0.680787176288)*x[0]+(0.626099489145)*x[1]**o+(-0.771540378408)*x[1]
        arg[1,3,0,1]=(-0.578341681282)*x[0]**o+(0.856407179963)*x[0]+(0.525319425356)*x[1]**o+(0.345651851179)*x[1]
        arg[1,3,1,0]=(-0.316652838058)*x[0]**o+(0.652409898929)*x[0]+(-0.82123442746)*x[1]**o+(-0.136868558081)*x[1]
        arg[1,3,1,1]=(-0.755530700082)*x[0]**o+(0.354429731374)*x[0]+(-0.14330437792)*x[1]**o+(-0.221454495186)*x[1]
        arg[1,3,2,0]=(-0.520261483769)*x[0]**o+(0.0873707179205)*x[0]+(-0.882995356682)*x[1]**o+(-0.773921687804)*x[1]
        arg[1,3,2,1]=(0.404278952718)*x[0]**o+(0.651166017764)*x[0]+(-0.496695253747)*x[1]**o+(0.625671754182)*x[1]
        arg[1,4,0,0]=(-0.730831539043)*x[0]**o+(0.219467989041)*x[0]+(-0.853784008431)*x[1]**o+(0.274799624897)*x[1]
        arg[1,4,0,1]=(0.893313801855)*x[0]**o+(-0.539687651773)*x[0]+(-0.411582029239)*x[1]**o+(-0.293521483805)*x[1]
        arg[1,4,1,0]=(0.484878049594)*x[0]**o+(0.218930293134)*x[0]+(-0.138362733141)*x[1]**o+(-0.663241354177)*x[1]
        arg[1,4,1,1]=(-0.608410207857)*x[0]**o+(-0.0921006917331)*x[0]+(-0.885705130565)*x[1]**o+(0.103724764404)*x[1]
        arg[1,4,2,0]=(0.150946260009)*x[0]**o+(-0.601834829203)*x[0]+(0.0141091402169)*x[1]**o+(-0.98316187826)*x[1]
        arg[1,4,2,1]=(0.98764918249)*x[0]**o+(-0.841155365687)*x[0]+(0.185525084432)*x[1]**o+(-0.613421841538)*x[1]
        arg[2,0,0,0]=(0.485337970037)*x[0]**o+(-0.898602108837)*x[0]+(0.847812629133)*x[1]**o+(-0.23522051021)*x[1]
        arg[2,0,0,1]=(0.835641029366)*x[0]**o+(-0.0524661277338)*x[0]+(0.41697147238)*x[1]**o+(-0.622813671618)*x[1]
        arg[2,0,1,0]=(0.505181088966)*x[0]**o+(-0.870141755006)*x[0]+(0.451139050304)*x[1]**o+(0.957790881903)*x[1]
        arg[2,0,1,1]=(0.822955028051)*x[0]**o+(-0.103028201548)*x[0]+(-0.604614127349)*x[1]**o+(-0.655334670762)*x[1]
        arg[2,0,2,0]=(-0.166647802545)*x[0]**o+(-0.466885186725)*x[0]+(0.0782333081578)*x[1]**o+(0.814414769115)*x[1]
        arg[2,0,2,1]=(-0.980310280968)*x[0]**o+(-0.884306329761)*x[0]+(-0.895550319566)*x[1]**o+(0.160657671233)*x[1]
        arg[2,1,0,0]=(-0.846405435483)*x[0]**o+(-0.625465093616)*x[0]+(-0.0789995975268)*x[1]**o+(-0.440180761086)*x[1]
        arg[2,1,0,1]=(-0.949855662419)*x[0]**o+(-0.013707742372)*x[0]+(0.0397729185783)*x[1]**o+(-0.836054083952)*x[1]
        arg[2,1,1,0]=(0.427246928002)*x[0]**o+(0.835369889909)*x[0]+(0.925300799558)*x[1]**o+(-0.328266770631)*x[1]
        arg[2,1,1,1]=(-0.36143365614)*x[0]**o+(-0.744107540819)*x[0]+(-0.358554131639)*x[1]**o+(-0.562674543946)*x[1]
        arg[2,1,2,0]=(0.60580594275)*x[0]**o+(0.644796993923)*x[0]+(-0.682407021927)*x[1]**o+(0.938016908061)*x[1]
        arg[2,1,2,1]=(-0.934393274317)*x[0]**o+(-0.718420272664)*x[0]+(0.0357039032065)*x[1]**o+(0.446387943344)*x[1]
        arg[2,2,0,0]=(-0.329105411737)*x[0]**o+(0.588935500072)*x[0]+(-0.833362483432)*x[1]**o+(0.629996961074)*x[1]
        arg[2,2,0,1]=(0.723977530226)*x[0]**o+(0.0698430385616)*x[0]+(-0.503761809156)*x[1]**o+(0.907472279839)*x[1]
        arg[2,2,1,0]=(-0.767545510793)*x[0]**o+(0.615094783682)*x[0]+(0.894594933703)*x[1]**o+(0.922357219049)*x[1]
        arg[2,2,1,1]=(-0.155739055576)*x[0]**o+(0.646142299486)*x[0]+(0.865932681534)*x[1]**o+(-0.126887380871)*x[1]
        arg[2,2,2,0]=(-0.629534320112)*x[0]**o+(-0.0904366301836)*x[0]+(0.193019806731)*x[1]**o+(-0.866080160466)*x[1]
        arg[2,2,2,1]=(-0.714527761053)*x[0]**o+(-0.813712879405)*x[0]+(-0.291734238438)*x[1]**o+(0.232675536415)*x[1]
        arg[2,3,0,0]=(-0.447474292879)*x[0]**o+(-0.639046887788)*x[0]+(0.159616683647)*x[1]**o+(0.564127141623)*x[1]
        arg[2,3,0,1]=(-0.608953936619)*x[0]**o+(0.273179484103)*x[0]+(0.42693156927)*x[1]**o+(-0.983331075647)*x[1]
        arg[2,3,1,0]=(0.855859483301)*x[0]**o+(-0.432495503925)*x[0]+(-0.67928203769)*x[1]**o+(-0.856541006702)*x[1]
        arg[2,3,1,1]=(-0.581185594204)*x[0]**o+(0.947720323693)*x[0]+(-0.520212852628)*x[1]**o+(-0.37563221594)*x[1]
        arg[2,3,2,0]=(0.351358492535)*x[0]**o+(0.367210525846)*x[0]+(0.97870179383)*x[1]**o+(-0.199559052968)*x[1]
        arg[2,3,2,1]=(0.797294518541)*x[0]**o+(0.546397721552)*x[0]+(0.374121851646)*x[1]**o+(0.704436689798)*x[1]
        arg[2,4,0,0]=(-0.756047775149)*x[0]**o+(-0.705403684825)*x[0]+(-0.662421213376)*x[1]**o+(0.390069314234)*x[1]
        arg[2,4,0,1]=(-0.658842889869)*x[0]**o+(0.766273769822)*x[0]+(-0.779201775962)*x[1]**o+(-0.968670270742)*x[1]
        arg[2,4,1,0]=(-0.336220564628)*x[0]**o+(-0.134745604319)*x[0]+(0.391939888428)*x[1]**o+(-0.113113999792)*x[1]
        arg[2,4,1,1]=(-0.8345363851)*x[0]**o+(0.339223929857)*x[0]+(0.628403353717)*x[1]**o+(0.799364167182)*x[1]
        arg[2,4,2,0]=(-0.898050446701)*x[0]**o+(-0.666373288962)*x[0]+(-0.620778021659)*x[1]**o+(-0.990359850049)*x[1]
        arg[2,4,2,1]=(-0.933931027675)*x[0]**o+(-0.119077281033)*x[0]+(0.557313533556)*x[1]**o+(0.160336158304)*x[1]
        arg[3,0,0,0]=(0.972462636314)*x[0]**o+(0.607913740671)*x[0]+(0.490168539257)*x[1]**o+(-0.12515921391)*x[1]
        arg[3,0,0,1]=(0.856370886333)*x[0]**o+(-0.143225009869)*x[0]+(-0.330381845603)*x[1]**o+(-0.271958059018)*x[1]
        arg[3,0,1,0]=(0.86742129112)*x[0]**o+(-0.0259756341823)*x[0]+(-0.0137047849942)*x[1]**o+(-0.0165444161044)*x[1]
        arg[3,0,1,1]=(-0.445996036941)*x[0]**o+(0.247936136111)*x[0]+(-0.381484740124)*x[1]**o+(0.836678417006)*x[1]
        arg[3,0,2,0]=(0.161387456771)*x[0]**o+(0.687533753311)*x[0]+(-0.463845417603)*x[1]**o+(0.665023823118)*x[1]
        arg[3,0,2,1]=(0.243410612742)*x[0]**o+(-0.52431675734)*x[0]+(0.75630192488)*x[1]**o+(-0.703601557317)*x[1]
        arg[3,1,0,0]=(0.793530689056)*x[0]**o+(0.814806726176)*x[0]+(0.582051068232)*x[1]**o+(0.692432923027)*x[1]
        arg[3,1,0,1]=(-0.283506909465)*x[0]**o+(0.21438128538)*x[0]+(-0.327476122861)*x[1]**o+(0.990178870269)*x[1]
        arg[3,1,1,0]=(-0.117386491827)*x[0]**o+(0.747771272481)*x[0]+(-0.439767754004)*x[1]**o+(-0.905218338639)*x[1]
        arg[3,1,1,1]=(-0.544805647706)*x[0]**o+(-0.0484033735345)*x[0]+(0.861443439156)*x[1]**o+(-0.992086899129)*x[1]
        arg[3,1,2,0]=(0.596208599112)*x[0]**o+(-0.276136075021)*x[0]+(0.079604120535)*x[1]**o+(0.124232758054)*x[1]
        arg[3,1,2,1]=(0.809847641841)*x[0]**o+(0.956041370087)*x[0]+(0.65224420189)*x[1]**o+(0.565327743771)*x[1]
        arg[3,2,0,0]=(-0.264599479107)*x[0]**o+(-0.276726556903)*x[0]+(0.577911819561)*x[1]**o+(-0.656157381756)*x[1]
        arg[3,2,0,1]=(-0.0636193103458)*x[0]**o+(0.941482878452)*x[0]+(0.877581030494)*x[1]**o+(-0.254911385455)*x[1]
        arg[3,2,1,0]=(-0.28083673633)*x[0]**o+(0.849095032935)*x[0]+(0.6825284068)*x[1]**o+(0.445605171749)*x[1]
        arg[3,2,1,1]=(-0.729331034619)*x[0]**o+(0.036400895394)*x[0]+(0.365733359214)*x[1]**o+(0.0305324156609)*x[1]
        arg[3,2,2,0]=(-0.70932226428)*x[0]**o+(0.310455571562)*x[0]+(-0.101305643894)*x[1]**o+(0.346873071627)*x[1]
        arg[3,2,2,1]=(0.712053329488)*x[0]**o+(-0.845042122758)*x[0]+(0.909517101069)*x[1]**o+(-0.163109058477)*x[1]
        arg[3,3,0,0]=(0.451744730688)*x[0]**o+(0.308756227526)*x[0]+(0.13963307493)*x[1]**o+(-0.945789640023)*x[1]
        arg[3,3,0,1]=(-0.852609785834)*x[0]**o+(-0.821513824231)*x[0]+(-0.544690113726)*x[1]**o+(0.335547843106)*x[1]
        arg[3,3,1,0]=(0.337600834432)*x[0]**o+(-0.170181468776)*x[0]+(-0.302415145845)*x[1]**o+(0.985425450212)*x[1]
        arg[3,3,1,1]=(-0.638242591048)*x[0]**o+(-0.0221442303187)*x[0]+(0.421311946694)*x[1]**o+(-0.43495081316)*x[1]
        arg[3,3,2,0]=(-0.734168622382)*x[0]**o+(-0.889300572069)*x[0]+(0.808161580285)*x[1]**o+(-0.526348332439)*x[1]
        arg[3,3,2,1]=(0.998918324168)*x[0]**o+(-0.953729666152)*x[0]+(0.665101882861)*x[1]**o+(-0.581118145725)*x[1]
        arg[3,4,0,0]=(-0.251226888098)*x[0]**o+(-0.711615999514)*x[0]+(0.291022154479)*x[1]**o+(0.0772744442604)*x[1]
        arg[3,4,0,1]=(0.162379537091)*x[0]**o+(0.363750962136)*x[0]+(0.635943839083)*x[1]**o+(0.513856953389)*x[1]
        arg[3,4,1,0]=(-0.99052354199)*x[0]**o+(0.289202668813)*x[0]+(-0.759078503012)*x[1]**o+(-0.588801682903)*x[1]
        arg[3,4,1,1]=(-0.942278699306)*x[0]**o+(0.476003512559)*x[0]+(0.873920293292)*x[1]**o+(0.867413991424)*x[1]
        arg[3,4,2,0]=(-0.602386702281)*x[0]**o+(0.694432164577)*x[0]+(0.367008950869)*x[1]**o+(-0.803331549449)*x[1]
        arg[3,4,2,1]=(-0.00204465287104)*x[0]**o+(-0.697572466685)*x[0]+(-0.84755978799)*x[1]**o+(-0.758946055726)*x[1]
        ref[0,0,0,0]=(0.0852574672711)*(1+2.*(dim-1.)/(o+1.))+(0.069447029165)*dim
        ref[0,0,0,1]=(0.177468127092)*(1+2.*(dim-1.)/(o+1.))+(-0.698555868323)*dim
        ref[0,0,1,0]=(0.572137980243)*(1+2.*(dim-1.)/(o+1.))+(-0.140119658749)*dim
        ref[0,0,1,1]=(-0.541780442017)*(1+2.*(dim-1.)/(o+1.))+(1.74856574618)*dim
        ref[0,0,2,0]=(1.5014527776)*(1+2.*(dim-1.)/(o+1.))+(0.472789336092)*dim
        ref[0,0,2,1]=(-0.759970633293)*(1+2.*(dim-1.)/(o+1.))+(1.12133686256)*dim
        ref[0,1,0,0]=(0.030965787278)*(1+2.*(dim-1.)/(o+1.))+(-1.53335256408)*dim
        ref[0,1,0,1]=(-0.359410958446)*(1+2.*(dim-1.)/(o+1.))+(0.853872638544)*dim
        ref[0,1,1,0]=(0.765005047774)*(1+2.*(dim-1.)/(o+1.))+(0.155304877218)*dim
        ref[0,1,1,1]=(-0.614845888015)*(1+2.*(dim-1.)/(o+1.))+(-0.0781765377528)*dim
        ref[0,1,2,0]=(-1.26644449842)*(1+2.*(dim-1.)/(o+1.))+(1.39714471889)*dim
        ref[0,1,2,1]=(-0.587139211502)*(1+2.*(dim-1.)/(o+1.))+(0.918620092379)*dim
        ref[0,2,0,0]=(-0.776779282155)*(1+2.*(dim-1.)/(o+1.))+(1.75267909498)*dim
        ref[0,2,0,1]=(-0.673013366714)*(1+2.*(dim-1.)/(o+1.))+(-0.283178964276)*dim
        ref[0,2,1,0]=(0.0614655560763)*(1+2.*(dim-1.)/(o+1.))+(-0.419228880686)*dim
        ref[0,2,1,1]=(-1.48302958627)*(1+2.*(dim-1.)/(o+1.))+(-1.19860521449)*dim
        ref[0,2,2,0]=(-0.507810468175)*(1+2.*(dim-1.)/(o+1.))+(0.744616418346)*dim
        ref[0,2,2,1]=(0.127124745526)*(1+2.*(dim-1.)/(o+1.))+(0.607591474596)*dim
        ref[0,3,0,0]=(-0.642771984806)*(1+2.*(dim-1.)/(o+1.))+(1.65517163472)*dim
        ref[0,3,0,1]=(0.203582724637)*(1+2.*(dim-1.)/(o+1.))+(-0.521103052242)*dim
        ref[0,3,1,0]=(1.02630580166)*(1+2.*(dim-1.)/(o+1.))+(0.286206034845)*dim
        ref[0,3,1,1]=(0.644196413613)*(1+2.*(dim-1.)/(o+1.))+(-0.226908428411)*dim
        ref[0,3,2,0]=(1.09876704246)*(1+2.*(dim-1.)/(o+1.))+(-0.289893245405)*dim
        ref[0,3,2,1]=(0.627372487469)*(1+2.*(dim-1.)/(o+1.))+(0.398934263276)*dim
        ref[0,4,0,0]=(-0.0661183748894)*(1+2.*(dim-1.)/(o+1.))+(-0.190206290411)*dim
        ref[0,4,0,1]=(-0.022840757054)*(1+2.*(dim-1.)/(o+1.))+(-0.919687084178)*dim
        ref[0,4,1,0]=(-0.063320927172)*(1+2.*(dim-1.)/(o+1.))+(-0.248275794205)*dim
        ref[0,4,1,1]=(-1.62483750784)*(1+2.*(dim-1.)/(o+1.))+(-0.112521044845)*dim
        ref[0,4,2,0]=(-0.199772879513)*(1+2.*(dim-1.)/(o+1.))+(-0.722566499138)*dim
        ref[0,4,2,1]=(0.911067722161)*(1+2.*(dim-1.)/(o+1.))+(0.700177712774)*dim
        ref[1,0,0,0]=(0.0380999602468)*(1+2.*(dim-1.)/(o+1.))+(-1.04886223152)*dim
        ref[1,0,0,1]=(1.28850834474)*(1+2.*(dim-1.)/(o+1.))+(-1.40735649994)*dim
        ref[1,0,1,0]=(0.553254021688)*(1+2.*(dim-1.)/(o+1.))+(-0.517478642549)*dim
        ref[1,0,1,1]=(-0.368152592115)*(1+2.*(dim-1.)/(o+1.))+(-1.43951521503)*dim
        ref[1,0,2,0]=(-1.53479197705)*(1+2.*(dim-1.)/(o+1.))+(-0.335919924034)*dim
        ref[1,0,2,1]=(0.745748951251)*(1+2.*(dim-1.)/(o+1.))+(0.210281820065)*dim
        ref[1,1,0,0]=(0.485906497229)*(1+2.*(dim-1.)/(o+1.))+(-0.0638596251453)*dim
        ref[1,1,0,1]=(0.00101286698233)*(1+2.*(dim-1.)/(o+1.))+(-0.17143157122)*dim
        ref[1,1,1,0]=(-0.169866832491)*(1+2.*(dim-1.)/(o+1.))+(-0.572533039866)*dim
        ref[1,1,1,1]=(-1.03615491429)*(1+2.*(dim-1.)/(o+1.))+(-0.016857336355)*dim
        ref[1,1,2,0]=(-1.11171654717)*(1+2.*(dim-1.)/(o+1.))+(0.183485561125)*dim
        ref[1,1,2,1]=(0.779419229804)*(1+2.*(dim-1.)/(o+1.))+(0.809679055143)*dim
        ref[1,2,0,0]=(-0.349883289458)*(1+2.*(dim-1.)/(o+1.))+(0.717941260728)*dim
        ref[1,2,0,1]=(-0.96179835566)*(1+2.*(dim-1.)/(o+1.))+(0.262648764533)*dim
        ref[1,2,1,0]=(0.104091593546)*(1+2.*(dim-1.)/(o+1.))+(-1.65289088394)*dim
        ref[1,2,1,1]=(0.559358587775)*(1+2.*(dim-1.)/(o+1.))+(0.679500872701)*dim
        ref[1,2,2,0]=(-0.0218373384179)*(1+2.*(dim-1.)/(o+1.))+(0.0713179379289)*dim
        ref[1,2,2,1]=(1.33791362695)*(1+2.*(dim-1.)/(o+1.))+(0.758079729181)*dim
        ref[1,3,0,0]=(0.808242997292)*(1+2.*(dim-1.)/(o+1.))+(-1.4523275547)*dim
        ref[1,3,0,1]=(-0.0530222559262)*(1+2.*(dim-1.)/(o+1.))+(1.20205903114)*dim
        ref[1,3,1,0]=(-1.13788726552)*(1+2.*(dim-1.)/(o+1.))+(0.515541340848)*dim
        ref[1,3,1,1]=(-0.898835078002)*(1+2.*(dim-1.)/(o+1.))+(0.132975236188)*dim
        ref[1,3,2,0]=(-1.40325684045)*(1+2.*(dim-1.)/(o+1.))+(-0.686550969884)*dim
        ref[1,3,2,1]=(-0.0924163010291)*(1+2.*(dim-1.)/(o+1.))+(1.27683777195)*dim
        ref[1,4,0,0]=(-1.58461554747)*(1+2.*(dim-1.)/(o+1.))+(0.494267613937)*dim
        ref[1,4,0,1]=(0.481731772616)*(1+2.*(dim-1.)/(o+1.))+(-0.833209135578)*dim
        ref[1,4,1,0]=(0.346515316453)*(1+2.*(dim-1.)/(o+1.))+(-0.444311061043)*dim
        ref[1,4,1,1]=(-1.49411533842)*(1+2.*(dim-1.)/(o+1.))+(0.0116240726713)*dim
        ref[1,4,2,0]=(0.165055400226)*(1+2.*(dim-1.)/(o+1.))+(-1.58499670746)*dim
        ref[1,4,2,1]=(1.17317426692)*(1+2.*(dim-1.)/(o+1.))+(-1.45457720722)*dim
        ref[2,0,0,0]=(1.33315059917)*(1+2.*(dim-1.)/(o+1.))+(-1.13382261905)*dim
        ref[2,0,0,1]=(1.25261250175)*(1+2.*(dim-1.)/(o+1.))+(-0.675279799352)*dim
        ref[2,0,1,0]=(0.95632013927)*(1+2.*(dim-1.)/(o+1.))+(0.0876491268976)*dim
        ref[2,0,1,1]=(0.218340900702)*(1+2.*(dim-1.)/(o+1.))+(-0.75836287231)*dim
        ref[2,0,2,0]=(-0.0884144943875)*(1+2.*(dim-1.)/(o+1.))+(0.34752958239)*dim
        ref[2,0,2,1]=(-1.87586060053)*(1+2.*(dim-1.)/(o+1.))+(-0.723648658528)*dim
        ref[2,1,0,0]=(-0.92540503301)*(1+2.*(dim-1.)/(o+1.))+(-1.0656458547)*dim
        ref[2,1,0,1]=(-0.910082743841)*(1+2.*(dim-1.)/(o+1.))+(-0.849761826324)*dim
        ref[2,1,1,0]=(1.35254772756)*(1+2.*(dim-1.)/(o+1.))+(0.507103119278)*dim
        ref[2,1,1,1]=(-0.719987787779)*(1+2.*(dim-1.)/(o+1.))+(-1.30678208477)*dim
        ref[2,1,2,0]=(-0.0766010791773)*(1+2.*(dim-1.)/(o+1.))+(1.58281390198)*dim
        ref[2,1,2,1]=(-0.89868937111)*(1+2.*(dim-1.)/(o+1.))+(-0.27203232932)*dim
        ref[2,2,0,0]=(-1.16246789517)*(1+2.*(dim-1.)/(o+1.))+(1.21893246115)*dim
        ref[2,2,0,1]=(0.22021572107)*(1+2.*(dim-1.)/(o+1.))+(0.977315318401)*dim
        ref[2,2,1,0]=(0.12704942291)*(1+2.*(dim-1.)/(o+1.))+(1.53745200273)*dim
        ref[2,2,1,1]=(0.710193625958)*(1+2.*(dim-1.)/(o+1.))+(0.519254918615)*dim
        ref[2,2,2,0]=(-0.43651451338)*(1+2.*(dim-1.)/(o+1.))+(-0.956516790649)*dim
        ref[2,2,2,1]=(-1.00626199949)*(1+2.*(dim-1.)/(o+1.))+(-0.581037342991)*dim
        ref[2,3,0,0]=(-0.287857609232)*(1+2.*(dim-1.)/(o+1.))+(-0.0749197461649)*dim
        ref[2,3,0,1]=(-0.182022367349)*(1+2.*(dim-1.)/(o+1.))+(-0.710151591543)*dim
        ref[2,3,1,0]=(0.176577445611)*(1+2.*(dim-1.)/(o+1.))+(-1.28903651063)*dim
        ref[2,3,1,1]=(-1.10139844683)*(1+2.*(dim-1.)/(o+1.))+(0.572088107754)*dim
        ref[2,3,2,0]=(1.33006028637)*(1+2.*(dim-1.)/(o+1.))+(0.167651472878)*dim
        ref[2,3,2,1]=(1.17141637019)*(1+2.*(dim-1.)/(o+1.))+(1.25083441135)*dim
        ref[2,4,0,0]=(-1.41846898852)*(1+2.*(dim-1.)/(o+1.))+(-0.315334370591)*dim
        ref[2,4,0,1]=(-1.43804466583)*(1+2.*(dim-1.)/(o+1.))+(-0.20239650092)*dim
        ref[2,4,1,0]=(0.0557193237997)*(1+2.*(dim-1.)/(o+1.))+(-0.247859604111)*dim
        ref[2,4,1,1]=(-0.206133031383)*(1+2.*(dim-1.)/(o+1.))+(1.13858809704)*dim
        ref[2,4,2,0]=(-1.51882846836)*(1+2.*(dim-1.)/(o+1.))+(-1.65673313901)*dim
        ref[2,4,2,1]=(-0.376617494119)*(1+2.*(dim-1.)/(o+1.))+(0.041258877271)*dim
        ref[3,0,0,0]=(1.46263117557)*(1+2.*(dim-1.)/(o+1.))+(0.482754526761)*dim
        ref[3,0,0,1]=(0.52598904073)*(1+2.*(dim-1.)/(o+1.))+(-0.415183068887)*dim
        ref[3,0,1,0]=(0.853716506126)*(1+2.*(dim-1.)/(o+1.))+(-0.0425200502867)*dim
        ref[3,0,1,1]=(-0.827480777065)*(1+2.*(dim-1.)/(o+1.))+(1.08461455312)*dim
        ref[3,0,2,0]=(-0.302457960832)*(1+2.*(dim-1.)/(o+1.))+(1.35255757643)*dim
        ref[3,0,2,1]=(0.999712537623)*(1+2.*(dim-1.)/(o+1.))+(-1.22791831466)*dim
        ref[3,1,0,0]=(1.37558175729)*(1+2.*(dim-1.)/(o+1.))+(1.5072396492)*dim
        ref[3,1,0,1]=(-0.610983032326)*(1+2.*(dim-1.)/(o+1.))+(1.20456015565)*dim
        ref[3,1,1,0]=(-0.557154245831)*(1+2.*(dim-1.)/(o+1.))+(-0.157447066158)*dim
        ref[3,1,1,1]=(0.316637791451)*(1+2.*(dim-1.)/(o+1.))+(-1.04049027266)*dim
        ref[3,1,2,0]=(0.675812719647)*(1+2.*(dim-1.)/(o+1.))+(-0.151903316967)*dim
        ref[3,1,2,1]=(1.46209184373)*(1+2.*(dim-1.)/(o+1.))+(1.52136911386)*dim
        ref[3,2,0,0]=(0.313312340454)*(1+2.*(dim-1.)/(o+1.))+(-0.93288393866)*dim
        ref[3,2,0,1]=(0.813961720148)*(1+2.*(dim-1.)/(o+1.))+(0.686571492997)*dim
        ref[3,2,1,0]=(0.40169167047)*(1+2.*(dim-1.)/(o+1.))+(1.29470020468)*dim
        ref[3,2,1,1]=(-0.363597675405)*(1+2.*(dim-1.)/(o+1.))+(0.066933311055)*dim
        ref[3,2,2,0]=(-0.810627908174)*(1+2.*(dim-1.)/(o+1.))+(0.657328643189)*dim
        ref[3,2,2,1]=(1.62157043056)*(1+2.*(dim-1.)/(o+1.))+(-1.00815118124)*dim
        ref[3,3,0,0]=(0.591377805617)*(1+2.*(dim-1.)/(o+1.))+(-0.637033412497)*dim
        ref[3,3,0,1]=(-1.39729989956)*(1+2.*(dim-1.)/(o+1.))+(-0.485965981125)*dim
        ref[3,3,1,0]=(0.0351856885862)*(1+2.*(dim-1.)/(o+1.))+(0.815243981437)*dim
        ref[3,3,1,1]=(-0.216930644354)*(1+2.*(dim-1.)/(o+1.))+(-0.457095043478)*dim
        ref[3,3,2,0]=(0.0739929579029)*(1+2.*(dim-1.)/(o+1.))+(-1.41564890451)*dim
        ref[3,3,2,1]=(1.66402020703)*(1+2.*(dim-1.)/(o+1.))+(-1.53484781188)*dim
        ref[3,4,0,0]=(0.0397952663815)*(1+2.*(dim-1.)/(o+1.))+(-0.634341555254)*dim
        ref[3,4,0,1]=(0.798323376174)*(1+2.*(dim-1.)/(o+1.))+(0.877607915525)*dim
        ref[3,4,1,0]=(-1.749602045)*(1+2.*(dim-1.)/(o+1.))+(-0.29959901409)*dim
        ref[3,4,1,1]=(-0.0683584060144)*(1+2.*(dim-1.)/(o+1.))+(1.34341750398)*dim
        ref[3,4,2,0]=(-0.235377751413)*(1+2.*(dim-1.)/(o+1.))+(-0.108899384872)*dim
        ref[3,4,2,1]=(-0.849604440861)*(1+2.*(dim-1.)/(o+1.))+(-1.45651852241)*dim
      else:
        arg[0,0,0,0]=(-0.229219068739)*x[0]**o+(0.293241016924)*x[0]+(-0.2730739542)*x[1]**o+(0.980438707356)*x[1]+(-0.712463216466)*x[2]**o+(0.201036317883)*x[2]
        arg[0,0,0,1]=(0.626271947422)*x[0]**o+(0.379293769575)*x[0]+(-0.476378592342)*x[1]**o+(-0.0089906260934)*x[1]+(-0.433154385811)*x[2]**o+(0.126766298748)*x[2]
        arg[0,0,1,0]=(-0.517950283035)*x[0]**o+(-0.219780986066)*x[0]+(0.345048580545)*x[1]**o+(0.116833851128)*x[1]+(-0.790641523314)*x[2]**o+(0.40506306417)*x[2]
        arg[0,0,1,1]=(0.910924420538)*x[0]**o+(0.281102231596)*x[0]+(0.115454208619)*x[1]**o+(0.164254056171)*x[1]+(0.44960060103)*x[2]**o+(-0.14238069165)*x[2]
        arg[0,0,2,0]=(0.86195808217)*x[0]**o+(0.775425956811)*x[0]+(0.214559021523)*x[1]**o+(0.400663063038)*x[1]+(0.995663608954)*x[2]**o+(0.228658543268)*x[2]
        arg[0,0,2,1]=(0.9025917492)*x[0]**o+(0.643205662573)*x[0]+(-0.304382459664)*x[1]**o+(-0.653256441532)*x[1]+(0.591142459807)*x[2]**o+(-0.888899030043)*x[2]
        arg[0,1,0,0]=(-0.874720784838)*x[0]**o+(-0.961456282274)*x[0]+(-0.954972993457)*x[1]**o+(0.00963939708746)*x[1]+(-0.884848381277)*x[2]**o+(-0.859924591622)*x[2]
        arg[0,1,0,1]=(0.133592941787)*x[0]**o+(0.12505415793)*x[0]+(-0.872197297559)*x[1]**o+(-0.666936976734)*x[1]+(-0.540166108036)*x[2]**o+(0.739145002101)*x[2]
        arg[0,1,1,0]=(0.543825266564)*x[0]**o+(0.167854064952)*x[0]+(-0.876426395953)*x[1]**o+(0.317538118917)*x[1]+(0.0454531201303)*x[2]**o+(0.822731343286)*x[2]
        arg[0,1,1,1]=(0.294557050513)*x[0]**o+(0.811698886899)*x[0]+(0.906213580147)*x[1]**o+(0.47895749797)*x[1]+(-0.937617554883)*x[2]**o+(0.171907458148)*x[2]
        arg[0,1,2,0]=(-0.843310949654)*x[0]**o+(0.766452648675)*x[0]+(-0.801151193917)*x[1]**o+(-0.373068887299)*x[1]+(-0.908043738284)*x[2]**o+(0.742278324563)*x[2]
        arg[0,1,2,1]=(-0.83259110525)*x[0]**o+(-0.303796015231)*x[0]+(0.153727560059)*x[1]**o+(-0.848393854019)*x[1]+(0.887795931916)*x[2]**o+(0.713486413074)*x[2]
        arg[0,2,0,0]=(-0.928603918182)*x[0]**o+(0.631213545526)*x[0]+(0.0149216054247)*x[1]**o+(0.136454313226)*x[1]+(-0.182080277504)*x[2]**o+(-0.898437757742)*x[2]
        arg[0,2,0,1]=(0.553844129645)*x[0]**o+(0.91439079798)*x[0]+(0.260424446678)*x[1]**o+(0.786629777779)*x[1]+(-0.412353876109)*x[2]**o+(0.496953000487)*x[2]
        arg[0,2,1,0]=(-0.48365422571)*x[0]**o+(-0.772252454664)*x[0]+(0.692592845143)*x[1]**o+(0.861992232175)*x[1]+(0.65169271692)*x[2]**o+(0.889175716524)*x[2]
        arg[0,2,1,1]=(-0.761086015475)*x[0]**o+(-0.616528617148)*x[0]+(-0.949399556926)*x[1]**o+(-0.783355074724)*x[1]+(-0.260470660608)*x[2]**o+(-0.854537542803)*x[2]
        arg[0,2,2,0]=(-0.812649439778)*x[0]**o+(-0.844862957855)*x[0]+(0.719928253134)*x[1]**o+(-0.892803903061)*x[1]+(-0.434753054503)*x[2]**o+(0.814188548451)*x[2]
        arg[0,2,2,1]=(-0.831106881173)*x[0]**o+(0.893305297647)*x[0]+(-0.351357795644)*x[1]**o+(0.0318574503478)*x[1]+(0.721896657873)*x[2]**o+(-0.75215838035)*x[2]
        arg[0,3,0,0]=(-0.868084492214)*x[0]**o+(-0.108398365838)*x[0]+(0.994384983885)*x[1]**o+(0.199445529335)*x[1]+(-0.570120483013)*x[2]**o+(0.992752456282)*x[2]
        arg[0,3,0,1]=(-0.304553428923)*x[0]**o+(-0.589461304318)*x[0]+(-0.570142565817)*x[1]**o+(-0.858530013475)*x[1]+(0.0690574539796)*x[2]**o+(-0.0321587525658)*x[2]
        arg[0,3,1,0]=(-0.729413026954)*x[0]**o+(-0.728248951404)*x[0]+(-0.568598599095)*x[1]**o+(0.0823012989572)*x[1]+(0.248614424573)*x[2]**o+(-0.116584878805)*x[2]
        arg[0,3,1,1]=(0.586846538084)*x[0]**o+(-0.406933529762)*x[0]+(0.839125682244)*x[1]**o+(0.237338453386)*x[1]+(-0.953101889056)*x[2]**o+(0.164147861616)*x[2]
        arg[0,3,2,0]=(-0.603895026116)*x[0]**o+(-0.401045478885)*x[0]+(-0.524739550832)*x[1]**o+(-0.077907095917)*x[1]+(0.535404409567)*x[2]**o+(0.745721528391)*x[2]
        arg[0,3,2,1]=(-0.189151084147)*x[0]**o+(-0.865927541126)*x[0]+(0.55265091657)*x[1]**o+(-0.592847311104)*x[1]+(-0.467509383222)*x[2]**o+(-0.779649276678)*x[2]
        arg[0,4,0,0]=(-0.564283606569)*x[0]**o+(-0.588029571592)*x[0]+(0.544948665159)*x[1]**o+(0.140885576219)*x[1]+(-0.333531842397)*x[2]**o+(-0.579095398292)*x[2]
        arg[0,4,0,1]=(-0.189127238455)*x[0]**o+(-0.693907127162)*x[0]+(0.40947793861)*x[1]**o+(-0.371098494308)*x[1]+(-0.189273002147)*x[2]**o+(-0.863169616331)*x[2]
        arg[0,4,1,0]=(-0.929838141145)*x[0]**o+(-0.0237583231579)*x[0]+(0.721069982341)*x[1]**o+(0.0378559913864)*x[1]+(-0.354729434476)*x[2]**o+(0.486256042969)*x[2]
        arg[0,4,1,1]=(0.349085173473)*x[0]**o+(0.398939355666)*x[0]+(-0.165772805366)*x[1]**o+(-0.532771334584)*x[1]+(-0.971868054413)*x[2]**o+(-0.38430185761)*x[2]
        arg[0,4,2,0]=(0.0330282399287)*x[0]**o+(0.157336131691)*x[0]+(-0.232739262685)*x[1]**o+(-0.280149712596)*x[1]+(0.374058430291)*x[2]**o+(0.594113015973)*x[2]
        arg[0,4,2,1]=(0.841740974112)*x[0]**o+(-0.454359680369)*x[0]+(-0.29970574726)*x[1]**o+(-0.234747232007)*x[1]+(0.590942390095)*x[2]**o+(-0.241807214234)*x[2]
        arg[1,0,0,0]=(0.815585848775)*x[0]**o+(-0.915020392674)*x[0]+(-0.0246530790781)*x[1]**o+(-0.783552074981)*x[1]+(0.539847778607)*x[2]**o+(-0.239345161181)*x[2]
        arg[1,0,0,1]=(-0.34212551354)*x[0]**o+(-0.231769108271)*x[0]+(-0.495709071836)*x[1]**o+(-0.301124823696)*x[1]+(-0.567450536966)*x[2]**o+(0.553640896859)*x[2]
        arg[1,0,1,0]=(0.507091065517)*x[0]**o+(-0.844317985101)*x[0]+(0.488126307787)*x[1]**o+(-0.624165922596)*x[1]+(0.00960189690918)*x[2]**o+(-0.35617062688)*x[2]
        arg[1,0,1,1]=(-0.219877421252)*x[0]**o+(-0.574843930095)*x[0]+(-0.846995760292)*x[1]**o+(-0.778496435041)*x[1]+(-0.817593097977)*x[2]**o+(0.387406829939)*x[2]
        arg[1,0,2,0]=(-0.774922277048)*x[0]**o+(-0.238152671014)*x[0]+(-0.913770877808)*x[1]**o+(-0.737305937434)*x[1]+(-0.13977056904)*x[2]**o+(-0.0156668131016)*x[2]
        arg[1,0,2,1]=(0.565993265911)*x[0]**o+(0.0690768496992)*x[0]+(-0.019845214223)*x[1]**o+(0.16253166701)*x[1]+(0.848117215885)*x[2]**o+(0.838075165243)*x[2]
        arg[1,1,0,0]=(0.871499694148)*x[0]**o+(-0.640851595157)*x[0]+(-0.170827284611)*x[1]**o+(-0.855550549988)*x[1]+(0.796086141249)*x[2]**o+(-0.214410672377)*x[2]
        arg[1,1,0,1]=(-0.903644604832)*x[0]**o+(0.0604780584681)*x[0]+(0.12800839523)*x[1]**o+(-0.594186752891)*x[1]+(0.415537334962)*x[2]**o+(0.122944611099)*x[2]
        arg[1,1,1,0]=(-0.611432775957)*x[0]**o+(0.391078622669)*x[0]+(-0.943091949273)*x[1]**o+(0.952739057317)*x[1]+(0.205104161722)*x[2]**o+(-0.946040001739)*x[2]
        arg[1,1,1,1]=(0.122941344579)*x[0]**o+(0.836981501711)*x[0]+(0.488970739079)*x[1]**o+(0.210062341907)*x[1]+(0.0264620393683)*x[2]**o+(-0.396386059797)*x[2]
        arg[1,1,2,0]=(-0.782658774038)*x[0]**o+(0.00611865020069)*x[0]+(0.806802875462)*x[1]**o+(-0.504323859583)*x[1]+(0.796276800331)*x[2]**o+(-0.792865116401)*x[2]
        arg[1,1,2,1]=(-0.249170081188)*x[0]**o+(-0.0234780571796)*x[0]+(-0.314017467872)*x[1]**o+(-0.0819148249592)*x[1]+(-0.71553368486)*x[2]**o+(0.763463207899)*x[2]
        arg[1,2,0,0]=(-0.722263632571)*x[0]**o+(-0.922870211147)*x[0]+(0.380794419995)*x[1]**o+(-0.76088532378)*x[1]+(0.417253071012)*x[2]**o+(0.152980049966)*x[2]
        arg[1,2,0,1]=(-0.667661997967)*x[0]**o+(-0.520570751147)*x[0]+(0.526368189149)*x[1]**o+(-0.0729376045288)*x[1]+(-0.434764684251)*x[2]**o+(0.124284939267)*x[2]
        arg[1,2,1,0]=(0.610868035553)*x[0]**o+(0.958644088455)*x[0]+(0.876029854699)*x[1]**o+(-0.278348787699)*x[1]+(-0.576695356283)*x[2]**o+(-0.302696556744)*x[2]
        arg[1,2,1,1]=(-0.224897853162)*x[0]**o+(0.89405397732)*x[0]+(-0.865281799513)*x[1]**o+(0.778866760053)*x[1]+(-0.939295291326)*x[2]**o+(0.256725768393)*x[2]
        arg[1,2,2,0]=(0.763785330202)*x[0]**o+(-0.92861265304)*x[0]+(-0.73488309202)*x[1]**o+(0.912711310857)*x[1]+(-0.359622626692)*x[2]**o+(-0.901261216611)*x[2]
        arg[1,2,2,1]=(-0.70687794303)*x[0]**o+(0.186497640523)*x[0]+(-0.587602932017)*x[1]**o+(0.439204174904)*x[1]+(0.0527573129765)*x[2]**o+(-0.121921819047)*x[2]
        arg[1,3,0,0]=(-0.576513727689)*x[0]**o+(0.228971758182)*x[0]+(0.282735129059)*x[1]**o+(-0.375270097915)*x[1]+(0.360087142563)*x[2]**o+(-0.0629946372113)*x[2]
        arg[1,3,0,1]=(-0.934375471561)*x[0]**o+(-0.710481522462)*x[0]+(-0.506438555546)*x[1]**o+(0.515278994703)*x[1]+(-0.789552111338)*x[2]**o+(0.556694887963)*x[2]
        arg[1,3,1,0]=(0.418081142778)*x[0]**o+(-0.990334755424)*x[0]+(0.465998147831)*x[1]**o+(0.85211792131)*x[1]+(0.64293315455)*x[2]**o+(0.418636667223)*x[2]
        arg[1,3,1,1]=(-0.192682640907)*x[0]**o+(-0.367176463246)*x[0]+(-0.512565915338)*x[1]**o+(0.431102518715)*x[1]+(-0.0213613616827)*x[2]**o+(-0.329868944656)*x[2]
        arg[1,3,2,0]=(-0.63277285658)*x[0]**o+(0.679819079734)*x[0]+(0.986520110022)*x[1]**o+(0.988046257154)*x[1]+(0.973419900492)*x[2]**o+(-0.177540534942)*x[2]
        arg[1,3,2,1]=(-0.660456888389)*x[0]**o+(0.546277948387)*x[0]+(-0.437065207396)*x[1]**o+(-0.310754727958)*x[1]+(-0.786950057282)*x[2]**o+(0.959526365177)*x[2]
        arg[1,4,0,0]=(0.764190976082)*x[0]**o+(0.970535029727)*x[0]+(0.389051006993)*x[1]**o+(-0.289346925704)*x[1]+(-0.386263606576)*x[2]**o+(0.624234063589)*x[2]
        arg[1,4,0,1]=(-0.982030683687)*x[0]**o+(-0.310500292581)*x[0]+(-0.559096285575)*x[1]**o+(-0.622541272039)*x[1]+(0.49924233081)*x[2]**o+(-0.219258681197)*x[2]
        arg[1,4,1,0]=(-0.124268932891)*x[0]**o+(0.102696501388)*x[0]+(-0.135123912862)*x[1]**o+(-0.960215132796)*x[1]+(-0.63171550556)*x[2]**o+(-0.493725285999)*x[2]
        arg[1,4,1,1]=(-0.397915436399)*x[0]**o+(-0.629822230732)*x[0]+(-0.65097233947)*x[1]**o+(0.915396047247)*x[1]+(0.835684562564)*x[2]**o+(-0.538096550078)*x[2]
        arg[1,4,2,0]=(-0.946429448216)*x[0]**o+(-0.310412085571)*x[0]+(-0.0984587257278)*x[1]**o+(-0.662090989911)*x[1]+(-0.424711481307)*x[2]**o+(0.00916879493127)*x[2]
        arg[1,4,2,1]=(0.919399549311)*x[0]**o+(-0.890110018315)*x[0]+(-0.685928475328)*x[1]**o+(-0.4731186395)*x[1]+(0.961887038718)*x[2]**o+(0.850957316131)*x[2]
        arg[2,0,0,0]=(0.834851801283)*x[0]**o+(-0.401572429371)*x[0]+(0.313309302528)*x[1]**o+(0.0368590314435)*x[1]+(-0.299807168778)*x[2]**o+(0.250599904818)*x[2]
        arg[2,0,0,1]=(-0.717562866174)*x[0]**o+(0.374797888506)*x[0]+(0.254329847579)*x[1]**o+(-0.987429692718)*x[1]+(-0.886758521324)*x[2]**o+(0.309486900444)*x[2]
        arg[2,0,1,0]=(0.139962143021)*x[0]**o+(-0.134735774213)*x[0]+(-0.231254041503)*x[1]**o+(0.0694785371931)*x[1]+(0.769929630132)*x[2]**o+(0.486646094292)*x[2]
        arg[2,0,1,1]=(0.349691701871)*x[0]**o+(-0.0308285150496)*x[0]+(0.00977143515921)*x[1]**o+(0.0970015800122)*x[1]+(-0.725991176667)*x[2]**o+(0.433663449433)*x[2]
        arg[2,0,2,0]=(0.564474039592)*x[0]**o+(0.950620065312)*x[0]+(0.256658075134)*x[1]**o+(0.774458453189)*x[1]+(0.00181734756534)*x[2]**o+(-0.270638852561)*x[2]
        arg[2,0,2,1]=(-0.175813811707)*x[0]**o+(0.821341326493)*x[0]+(-0.464966990893)*x[1]**o+(-0.310747066257)*x[1]+(-0.413557861251)*x[2]**o+(-0.799160067701)*x[2]
        arg[2,1,0,0]=(-0.449817381621)*x[0]**o+(-0.0573780378383)*x[0]+(-0.22675349133)*x[1]**o+(-0.150314340624)*x[1]+(-0.116693721705)*x[2]**o+(-0.20577625675)*x[2]
        arg[2,1,0,1]=(0.955318430674)*x[0]**o+(-0.148100168239)*x[0]+(0.741705440574)*x[1]**o+(-0.0768022054634)*x[1]+(-0.721807872287)*x[2]**o+(-0.898235041116)*x[2]
        arg[2,1,1,0]=(-0.183120810472)*x[0]**o+(0.307305712226)*x[0]+(0.699835684443)*x[1]**o+(0.964082351856)*x[1]+(-0.0348051023458)*x[2]**o+(0.544435382916)*x[2]
        arg[2,1,1,1]=(0.786862912642)*x[0]**o+(0.37388939814)*x[0]+(0.608937302857)*x[1]**o+(0.122305286217)*x[1]+(-0.892536849678)*x[2]**o+(0.119257478145)*x[2]
        arg[2,1,2,0]=(0.758154908623)*x[0]**o+(0.695188856067)*x[0]+(0.458313343957)*x[1]**o+(0.979157471455)*x[1]+(0.194237063725)*x[2]**o+(0.773001101454)*x[2]
        arg[2,1,2,1]=(0.0754678814151)*x[0]**o+(-0.639556181007)*x[0]+(0.934627102078)*x[1]**o+(0.0832082572611)*x[1]+(0.668822756858)*x[2]**o+(0.627532210297)*x[2]
        arg[2,2,0,0]=(-0.192921889174)*x[0]**o+(0.989703287553)*x[0]+(0.00517568784693)*x[1]**o+(-0.360076937731)*x[1]+(0.947056110784)*x[2]**o+(0.469841550381)*x[2]
        arg[2,2,0,1]=(0.201501669595)*x[0]**o+(-0.959957961834)*x[0]+(0.0229159090152)*x[1]**o+(0.396188263084)*x[1]+(0.186211667989)*x[2]**o+(0.304503474778)*x[2]
        arg[2,2,1,0]=(0.333974995608)*x[0]**o+(0.210202867626)*x[0]+(-0.410565897643)*x[1]**o+(0.52659985644)*x[1]+(0.983960564934)*x[2]**o+(-0.439323691866)*x[2]
        arg[2,2,1,1]=(0.706119074329)*x[0]**o+(-0.00220875175563)*x[0]+(0.143693169123)*x[1]**o+(0.217254593235)*x[1]+(-0.412057975252)*x[2]**o+(-0.967360361422)*x[2]
        arg[2,2,2,0]=(0.217307356597)*x[0]**o+(0.861728452163)*x[0]+(-0.193239462241)*x[1]**o+(0.872273176526)*x[1]+(-0.436414721507)*x[2]**o+(-0.62398425635)*x[2]
        arg[2,2,2,1]=(0.467099668794)*x[0]**o+(0.109742400918)*x[0]+(-0.987522191801)*x[1]**o+(-0.250251964275)*x[1]+(0.442152166523)*x[2]**o+(-0.709157679576)*x[2]
        arg[2,3,0,0]=(0.769666242031)*x[0]**o+(0.2352745864)*x[0]+(-0.302893533611)*x[1]**o+(0.0806677655098)*x[1]+(0.886434893279)*x[2]**o+(0.366016417047)*x[2]
        arg[2,3,0,1]=(-0.612204706187)*x[0]**o+(-0.242961993133)*x[0]+(0.946225816441)*x[1]**o+(0.528642650581)*x[1]+(-0.0166837450337)*x[2]**o+(-0.0889397945206)*x[2]
        arg[2,3,1,0]=(0.218589745288)*x[0]**o+(-0.296125811204)*x[0]+(0.358743045538)*x[1]**o+(-0.945778818656)*x[1]+(-0.33155158646)*x[2]**o+(-0.293038455557)*x[2]
        arg[2,3,1,1]=(0.753737723433)*x[0]**o+(-0.772926141856)*x[0]+(-0.0107183417902)*x[1]**o+(0.125538612735)*x[1]+(0.225277420156)*x[2]**o+(-0.54876651451)*x[2]
        arg[2,3,2,0]=(0.947419453319)*x[0]**o+(-0.101935160067)*x[0]+(0.481841505589)*x[1]**o+(-0.873842146032)*x[1]+(-0.377764402402)*x[2]**o+(-0.361560947647)*x[2]
        arg[2,3,2,1]=(0.745703609187)*x[0]**o+(-0.488401078192)*x[0]+(0.190261500318)*x[1]**o+(-0.22176180212)*x[1]+(0.80182740764)*x[2]**o+(-0.504953098351)*x[2]
        arg[2,4,0,0]=(0.476355625206)*x[0]**o+(0.452225231178)*x[0]+(0.669315429911)*x[1]**o+(0.159511010158)*x[1]+(0.652056592943)*x[2]**o+(0.44515033609)*x[2]
        arg[2,4,0,1]=(0.512194809916)*x[0]**o+(0.654452558318)*x[0]+(-0.526832882528)*x[1]**o+(0.590859897981)*x[1]+(-0.155989159043)*x[2]**o+(-0.689178072559)*x[2]
        arg[2,4,1,0]=(0.0450244624553)*x[0]**o+(0.0790299000335)*x[0]+(-0.690643908994)*x[1]**o+(-0.262426911912)*x[1]+(-0.426935471303)*x[2]**o+(-0.434570506153)*x[2]
        arg[2,4,1,1]=(-0.04445890079)*x[0]**o+(-0.541574770252)*x[0]+(-0.879290270122)*x[1]**o+(0.545302743626)*x[1]+(-0.836472565261)*x[2]**o+(0.66750272736)*x[2]
        arg[2,4,2,0]=(-0.477788509677)*x[0]**o+(0.706195950613)*x[0]+(0.628430029293)*x[1]**o+(-0.297608932813)*x[1]+(0.38364501438)*x[2]**o+(0.855444938831)*x[2]
        arg[2,4,2,1]=(0.123183000451)*x[0]**o+(0.41176132265)*x[0]+(-0.379498883356)*x[1]**o+(-0.53442961658)*x[1]+(-0.101215056317)*x[2]**o+(0.247971839461)*x[2]
        arg[3,0,0,0]=(-0.239170701904)*x[0]**o+(0.369420936359)*x[0]+(0.972913513036)*x[1]**o+(0.775248415403)*x[1]+(-0.850310640223)*x[2]**o+(-0.815345534491)*x[2]
        arg[3,0,0,1]=(0.591124498004)*x[0]**o+(0.529522171858)*x[0]+(-0.974005221641)*x[1]**o+(0.430102201571)*x[1]+(0.570519406976)*x[2]**o+(-0.0915083364671)*x[2]
        arg[3,0,1,0]=(-0.27678065557)*x[0]**o+(-0.53225308602)*x[0]+(0.902356190541)*x[1]**o+(-0.564808732046)*x[1]+(0.0981283775668)*x[2]**o+(-0.64957631107)*x[2]
        arg[3,0,1,1]=(0.230523873408)*x[0]**o+(-0.382745197262)*x[0]+(0.41507101812)*x[1]**o+(0.48492708335)*x[1]+(-0.11672353881)*x[2]**o+(-0.56902283397)*x[2]
        arg[3,0,2,0]=(-0.400023140104)*x[0]**o+(-0.406234012546)*x[0]+(-0.17810623045)*x[1]**o+(-0.833748300281)*x[1]+(-0.463689953863)*x[2]**o+(-0.11095482465)*x[2]
        arg[3,0,2,1]=(0.448885918108)*x[0]**o+(0.0447046792888)*x[0]+(-0.781652457281)*x[1]**o+(-0.893726375109)*x[1]+(0.918099233406)*x[2]**o+(0.560717614604)*x[2]
        arg[3,1,0,0]=(0.046700262853)*x[0]**o+(0.729867044213)*x[0]+(-0.870868205823)*x[1]**o+(-0.711267787629)*x[1]+(0.760550112312)*x[2]**o+(-0.0188334901698)*x[2]
        arg[3,1,0,1]=(0.304305551549)*x[0]**o+(-0.73527291962)*x[0]+(0.722134673974)*x[1]**o+(-0.40822784698)*x[1]+(0.912464122456)*x[2]**o+(-0.857656020349)*x[2]
        arg[3,1,1,0]=(-0.857864508282)*x[0]**o+(-0.22160002809)*x[0]+(-0.0892530574233)*x[1]**o+(-0.547259900671)*x[1]+(0.641858628113)*x[2]**o+(-0.203700292205)*x[2]
        arg[3,1,1,1]=(-0.47064403668)*x[0]**o+(-0.105885697358)*x[0]+(-0.899570240486)*x[1]**o+(0.660942689756)*x[1]+(-0.365836826289)*x[2]**o+(0.9593955249)*x[2]
        arg[3,1,2,0]=(-0.51407433496)*x[0]**o+(0.369092182634)*x[0]+(-0.740281092517)*x[1]**o+(0.383443939366)*x[1]+(-0.140318759419)*x[2]**o+(-0.481274651104)*x[2]
        arg[3,1,2,1]=(-0.938924125636)*x[0]**o+(0.41994781084)*x[0]+(-0.946320511542)*x[1]**o+(-0.86702606561)*x[1]+(0.63547569926)*x[2]**o+(0.923988068696)*x[2]
        arg[3,2,0,0]=(0.212833889165)*x[0]**o+(-0.0823077975238)*x[0]+(0.038483237489)*x[1]**o+(0.393784376623)*x[1]+(0.573190557103)*x[2]**o+(0.403015743632)*x[2]
        arg[3,2,0,1]=(-0.169294531943)*x[0]**o+(-0.769228919251)*x[0]+(-0.907462814057)*x[1]**o+(-0.188574782291)*x[1]+(-0.859793027531)*x[2]**o+(-0.673334626964)*x[2]
        arg[3,2,1,0]=(0.646744388517)*x[0]**o+(0.346749855021)*x[0]+(-0.314382791156)*x[1]**o+(-0.0396545336605)*x[1]+(0.844678566939)*x[2]**o+(0.724721336727)*x[2]
        arg[3,2,1,1]=(-0.471761417345)*x[0]**o+(-0.801631597222)*x[0]+(-0.376444012893)*x[1]**o+(0.920939382926)*x[1]+(-0.728019331398)*x[2]**o+(0.905808341144)*x[2]
        arg[3,2,2,0]=(-0.713759272803)*x[0]**o+(-0.930638013429)*x[0]+(0.864451879958)*x[1]**o+(0.0566794358824)*x[1]+(-0.292092226905)*x[2]**o+(-0.195695530268)*x[2]
        arg[3,2,2,1]=(0.843437618194)*x[0]**o+(0.214911962718)*x[0]+(0.475690162826)*x[1]**o+(0.0744570063618)*x[1]+(0.351711310616)*x[2]**o+(-0.187223450548)*x[2]
        arg[3,3,0,0]=(0.108651500878)*x[0]**o+(-0.402624590847)*x[0]+(0.120700159277)*x[1]**o+(-0.363139807635)*x[1]+(-0.62413441951)*x[2]**o+(-0.474236131596)*x[2]
        arg[3,3,0,1]=(-0.0836636547536)*x[0]**o+(-0.187959781777)*x[0]+(0.756983584308)*x[1]**o+(0.756573839599)*x[1]+(0.195640878564)*x[2]**o+(-0.566596474032)*x[2]
        arg[3,3,1,0]=(-0.171427844449)*x[0]**o+(-0.266807792484)*x[0]+(0.380046037965)*x[1]**o+(-0.129297371793)*x[1]+(0.79139100315)*x[2]**o+(0.223804751067)*x[2]
        arg[3,3,1,1]=(-0.054757303264)*x[0]**o+(-0.823878124196)*x[0]+(0.80695256107)*x[1]**o+(0.583250192211)*x[1]+(0.894937939133)*x[2]**o+(0.878757173892)*x[2]
        arg[3,3,2,0]=(0.13923802434)*x[0]**o+(-0.600732296834)*x[0]+(0.111226716676)*x[1]**o+(-0.474524749496)*x[1]+(0.674290476359)*x[2]**o+(-0.06714767759)*x[2]
        arg[3,3,2,1]=(0.246187769682)*x[0]**o+(-0.871969439162)*x[0]+(0.56748693564)*x[1]**o+(-0.0831774328079)*x[1]+(-0.27380467532)*x[2]**o+(-0.623711953244)*x[2]
        arg[3,4,0,0]=(-0.345947366312)*x[0]**o+(-0.304842609526)*x[0]+(-0.720722450633)*x[1]**o+(0.889712922756)*x[1]+(-0.0451362945237)*x[2]**o+(-0.52344522887)*x[2]
        arg[3,4,0,1]=(-0.830119647904)*x[0]**o+(0.577888363835)*x[0]+(0.180691945154)*x[1]**o+(-0.489554894073)*x[1]+(0.848489044545)*x[2]**o+(-0.243139985297)*x[2]
        arg[3,4,1,0]=(0.846597693999)*x[0]**o+(0.757477455396)*x[0]+(-0.827604908657)*x[1]**o+(-0.108392312717)*x[1]+(0.597454707613)*x[2]**o+(0.35846243312)*x[2]
        arg[3,4,1,1]=(0.107973753109)*x[0]**o+(0.578649460223)*x[0]+(0.609012033524)*x[1]**o+(-0.8428861922)*x[1]+(0.42144759968)*x[2]**o+(-0.938952208084)*x[2]
        arg[3,4,2,0]=(0.286586330618)*x[0]**o+(0.7324646289)*x[0]+(-0.448461377611)*x[1]**o+(0.62562408085)*x[1]+(0.352429283116)*x[2]**o+(0.112047034296)*x[2]
        arg[3,4,2,1]=(-0.431663387082)*x[0]**o+(-0.612934322552)*x[0]+(-0.426644797074)*x[1]**o+(0.768645027094)*x[1]+(-0.453242885757)*x[2]**o+(0.893748345809)*x[2]
        ref[0,0,0,0]=(-1.2147562394)*(1+2.*(dim-1.)/(o+1.))+(1.47471604216)*dim
        ref[0,0,0,1]=(-0.283261030731)*(1+2.*(dim-1.)/(o+1.))+(0.49706944223)*dim
        ref[0,0,1,0]=(-0.963543225803)*(1+2.*(dim-1.)/(o+1.))+(0.302115929232)*dim
        ref[0,0,1,1]=(1.47597923019)*(1+2.*(dim-1.)/(o+1.))+(0.302975596116)*dim
        ref[0,0,2,0]=(2.07218071265)*(1+2.*(dim-1.)/(o+1.))+(1.40474756312)*dim
        ref[0,0,2,1]=(1.18935174934)*(1+2.*(dim-1.)/(o+1.))+(-0.898949809002)*dim
        ref[0,1,0,0]=(-2.71454215957)*(1+2.*(dim-1.)/(o+1.))+(-1.81174147681)*dim
        ref[0,1,0,1]=(-1.27877046381)*(1+2.*(dim-1.)/(o+1.))+(0.197262183297)*dim
        ref[0,1,1,0]=(-0.287148009258)*(1+2.*(dim-1.)/(o+1.))+(1.30812352716)*dim
        ref[0,1,1,1]=(0.263153075777)*(1+2.*(dim-1.)/(o+1.))+(1.46256384302)*dim
        ref[0,1,2,0]=(-2.55250588186)*(1+2.*(dim-1.)/(o+1.))+(1.13566208594)*dim
        ref[0,1,2,1]=(0.208932386724)*(1+2.*(dim-1.)/(o+1.))+(-0.438703456176)*dim
        ref[0,2,0,0]=(-1.09576259026)*(1+2.*(dim-1.)/(o+1.))+(-0.13076989899)*dim
        ref[0,2,0,1]=(0.401914700214)*(1+2.*(dim-1.)/(o+1.))+(2.19797357625)*dim
        ref[0,2,1,0]=(0.860631336354)*(1+2.*(dim-1.)/(o+1.))+(0.978915494035)*dim
        ref[0,2,1,1]=(-1.97095623301)*(1+2.*(dim-1.)/(o+1.))+(-2.25442123467)*dim
        ref[0,2,2,0]=(-0.527474241147)*(1+2.*(dim-1.)/(o+1.))+(-0.923478312465)*dim
        ref[0,2,2,1]=(-0.460568018944)*(1+2.*(dim-1.)/(o+1.))+(0.173004367644)*dim
        ref[0,3,0,0]=(-0.443819991343)*(1+2.*(dim-1.)/(o+1.))+(1.08379961978)*dim
        ref[0,3,0,1]=(-0.80563854076)*(1+2.*(dim-1.)/(o+1.))+(-1.48015007036)*dim
        ref[0,3,1,0]=(-1.04939720148)*(1+2.*(dim-1.)/(o+1.))+(-0.762532531252)*dim
        ref[0,3,1,1]=(0.472870331273)*(1+2.*(dim-1.)/(o+1.))+(-0.00544721475928)*dim
        ref[0,3,2,0]=(-0.593230167381)*(1+2.*(dim-1.)/(o+1.))+(0.26676895359)*dim
        ref[0,3,2,1]=(-0.104009550799)*(1+2.*(dim-1.)/(o+1.))+(-2.23842412891)*dim
        ref[0,4,0,0]=(-0.352866783806)*(1+2.*(dim-1.)/(o+1.))+(-1.02623939367)*dim
        ref[0,4,0,1]=(0.0310776980081)*(1+2.*(dim-1.)/(o+1.))+(-1.9281752378)*dim
        ref[0,4,1,0]=(-0.56349759328)*(1+2.*(dim-1.)/(o+1.))+(0.500353711197)*dim
        ref[0,4,1,1]=(-0.788555686306)*(1+2.*(dim-1.)/(o+1.))+(-0.518133836527)*dim
        ref[0,4,2,0]=(0.174347407535)*(1+2.*(dim-1.)/(o+1.))+(0.471299435068)*dim
        ref[0,4,2,1]=(1.13297761695)*(1+2.*(dim-1.)/(o+1.))+(-0.93091412661)*dim
        ref[1,0,0,0]=(1.3307805483)*(1+2.*(dim-1.)/(o+1.))+(-1.93791762884)*dim
        ref[1,0,0,1]=(-1.40528512234)*(1+2.*(dim-1.)/(o+1.))+(0.0207469648915)*dim
        ref[1,0,1,0]=(1.00481927021)*(1+2.*(dim-1.)/(o+1.))+(-1.82465453458)*dim
        ref[1,0,1,1]=(-1.88446627952)*(1+2.*(dim-1.)/(o+1.))+(-0.965933535196)*dim
        ref[1,0,2,0]=(-1.8284637239)*(1+2.*(dim-1.)/(o+1.))+(-0.991125421549)*dim
        ref[1,0,2,1]=(1.39426526757)*(1+2.*(dim-1.)/(o+1.))+(1.06968368195)*dim
        ref[1,1,0,0]=(1.49675855079)*(1+2.*(dim-1.)/(o+1.))+(-1.71081281752)*dim
        ref[1,1,0,1]=(-0.36009887464)*(1+2.*(dim-1.)/(o+1.))+(-0.410764083324)*dim
        ref[1,1,1,0]=(-1.34942056351)*(1+2.*(dim-1.)/(o+1.))+(0.397777678247)*dim
        ref[1,1,1,1]=(0.638374123027)*(1+2.*(dim-1.)/(o+1.))+(0.650657783821)*dim
        ref[1,1,2,0]=(0.820420901755)*(1+2.*(dim-1.)/(o+1.))+(-1.29107032578)*dim
        ref[1,1,2,1]=(-1.27872123392)*(1+2.*(dim-1.)/(o+1.))+(0.65807032576)*dim
        ref[1,2,0,0]=(0.075783858436)*(1+2.*(dim-1.)/(o+1.))+(-1.53077548496)*dim
        ref[1,2,0,1]=(-0.576058493068)*(1+2.*(dim-1.)/(o+1.))+(-0.469223416409)*dim
        ref[1,2,1,0]=(0.91020253397)*(1+2.*(dim-1.)/(o+1.))+(0.377598744012)*dim
        ref[1,2,1,1]=(-2.029474944)*(1+2.*(dim-1.)/(o+1.))+(1.92964650577)*dim
        ref[1,2,2,0]=(-0.330720388509)*(1+2.*(dim-1.)/(o+1.))+(-0.917162558794)*dim
        ref[1,2,2,1]=(-1.24172356207)*(1+2.*(dim-1.)/(o+1.))+(0.503779996381)*dim
        ref[1,3,0,0]=(0.0663085439332)*(1+2.*(dim-1.)/(o+1.))+(-0.209292976945)*dim
        ref[1,3,0,1]=(-2.23036613844)*(1+2.*(dim-1.)/(o+1.))+(0.361492360204)*dim
        ref[1,3,1,0]=(1.52701244516)*(1+2.*(dim-1.)/(o+1.))+(0.280419833109)*dim
        ref[1,3,1,1]=(-0.726609917928)*(1+2.*(dim-1.)/(o+1.))+(-0.265942889187)*dim
        ref[1,3,2,0]=(1.32716715393)*(1+2.*(dim-1.)/(o+1.))+(1.49032480195)*dim
        ref[1,3,2,1]=(-1.88447215307)*(1+2.*(dim-1.)/(o+1.))+(1.19504958561)*dim
        ref[1,4,0,0]=(0.766978376499)*(1+2.*(dim-1.)/(o+1.))+(1.30542216761)*dim
        ref[1,4,0,1]=(-1.04188463845)*(1+2.*(dim-1.)/(o+1.))+(-1.15230024582)*dim
        ref[1,4,1,0]=(-0.891108351313)*(1+2.*(dim-1.)/(o+1.))+(-1.35124391741)*dim
        ref[1,4,1,1]=(-0.213203213304)*(1+2.*(dim-1.)/(o+1.))+(-0.252522733564)*dim
        ref[1,4,2,0]=(-1.46959965525)*(1+2.*(dim-1.)/(o+1.))+(-0.963334280551)*dim
        ref[1,4,2,1]=(1.1953581127)*(1+2.*(dim-1.)/(o+1.))+(-0.512271341684)*dim
        ref[2,0,0,0]=(0.848353935033)*(1+2.*(dim-1.)/(o+1.))+(-0.11411349311)*dim
        ref[2,0,0,1]=(-1.34999153992)*(1+2.*(dim-1.)/(o+1.))+(-0.303144903768)*dim
        ref[2,0,1,0]=(0.67863773165)*(1+2.*(dim-1.)/(o+1.))+(0.421388857272)*dim
        ref[2,0,1,1]=(-0.366528039637)*(1+2.*(dim-1.)/(o+1.))+(0.499836514396)*dim
        ref[2,0,2,0]=(0.822949462292)*(1+2.*(dim-1.)/(o+1.))+(1.45443966594)*dim
        ref[2,0,2,1]=(-1.05433866385)*(1+2.*(dim-1.)/(o+1.))+(-0.288565807465)*dim
        ref[2,1,0,0]=(-0.793264594657)*(1+2.*(dim-1.)/(o+1.))+(-0.413468635212)*dim
        ref[2,1,0,1]=(0.975215998961)*(1+2.*(dim-1.)/(o+1.))+(-1.12313741482)*dim
        ref[2,1,1,0]=(0.481909771625)*(1+2.*(dim-1.)/(o+1.))+(1.815823447)*dim
        ref[2,1,1,1]=(0.503263365821)*(1+2.*(dim-1.)/(o+1.))+(0.615452162503)*dim
        ref[2,1,2,0]=(1.41070531631)*(1+2.*(dim-1.)/(o+1.))+(2.44734742898)*dim
        ref[2,1,2,1]=(1.67891774035)*(1+2.*(dim-1.)/(o+1.))+(0.0711842865505)*dim
        ref[2,2,0,0]=(0.759309909457)*(1+2.*(dim-1.)/(o+1.))+(1.0994679002)*dim
        ref[2,2,0,1]=(0.410629246599)*(1+2.*(dim-1.)/(o+1.))+(-0.259266223972)*dim
        ref[2,2,1,0]=(0.907369662899)*(1+2.*(dim-1.)/(o+1.))+(0.297479032201)*dim
        ref[2,2,1,1]=(0.437754268199)*(1+2.*(dim-1.)/(o+1.))+(-0.752314519943)*dim
        ref[2,2,2,0]=(-0.412346827151)*(1+2.*(dim-1.)/(o+1.))+(1.11001737234)*dim
        ref[2,2,2,1]=(-0.0782703564833)*(1+2.*(dim-1.)/(o+1.))+(-0.849667242932)*dim
        ref[2,3,0,0]=(1.3532076017)*(1+2.*(dim-1.)/(o+1.))+(0.681958768956)*dim
        ref[2,3,0,1]=(0.317337365221)*(1+2.*(dim-1.)/(o+1.))+(0.196740862928)*dim
        ref[2,3,1,0]=(0.245781204366)*(1+2.*(dim-1.)/(o+1.))+(-1.53494308542)*dim
        ref[2,3,1,1]=(0.968296801799)*(1+2.*(dim-1.)/(o+1.))+(-1.19615404363)*dim
        ref[2,3,2,0]=(1.05149655651)*(1+2.*(dim-1.)/(o+1.))+(-1.33733825375)*dim
        ref[2,3,2,1]=(1.73779251715)*(1+2.*(dim-1.)/(o+1.))+(-1.21511597866)*dim
        ref[2,4,0,0]=(1.79772764806)*(1+2.*(dim-1.)/(o+1.))+(1.05688657743)*dim
        ref[2,4,0,1]=(-0.170627231654)*(1+2.*(dim-1.)/(o+1.))+(0.55613438374)*dim
        ref[2,4,1,0]=(-1.07255491784)*(1+2.*(dim-1.)/(o+1.))+(-0.617967518031)*dim
        ref[2,4,1,1]=(-1.76022173617)*(1+2.*(dim-1.)/(o+1.))+(0.671230700733)*dim
        ref[2,4,2,0]=(0.534286533995)*(1+2.*(dim-1.)/(o+1.))+(1.26403195663)*dim
        ref[2,4,2,1]=(-0.357530939222)*(1+2.*(dim-1.)/(o+1.))+(0.12530354553)*dim
        ref[3,0,0,0]=(-0.11656782909)*(1+2.*(dim-1.)/(o+1.))+(0.32932381727)*dim
        ref[3,0,0,1]=(0.187638683339)*(1+2.*(dim-1.)/(o+1.))+(0.868116036962)*dim
        ref[3,0,1,0]=(0.723703912538)*(1+2.*(dim-1.)/(o+1.))+(-1.74663812914)*dim
        ref[3,0,1,1]=(0.528871352717)*(1+2.*(dim-1.)/(o+1.))+(-0.466840947882)*dim
        ref[3,0,2,0]=(-1.04181932442)*(1+2.*(dim-1.)/(o+1.))+(-1.35093713748)*dim
        ref[3,0,2,1]=(0.585332694234)*(1+2.*(dim-1.)/(o+1.))+(-0.288304081216)*dim
        ref[3,1,0,0]=(-0.0636178306576)*(1+2.*(dim-1.)/(o+1.))+(-0.000234233586678)*dim
        ref[3,1,0,1]=(1.93890434798)*(1+2.*(dim-1.)/(o+1.))+(-2.00115678695)*dim
        ref[3,1,1,0]=(-0.305258937592)*(1+2.*(dim-1.)/(o+1.))+(-0.972560220966)*dim
        ref[3,1,1,1]=(-1.73605110346)*(1+2.*(dim-1.)/(o+1.))+(1.5144525173)*dim
        ref[3,1,2,0]=(-1.3946741869)*(1+2.*(dim-1.)/(o+1.))+(0.271261470896)*dim
        ref[3,1,2,1]=(-1.24976893792)*(1+2.*(dim-1.)/(o+1.))+(0.476909813927)*dim
        ref[3,2,0,0]=(0.824507683756)*(1+2.*(dim-1.)/(o+1.))+(0.71449232273)*dim
        ref[3,2,0,1]=(-1.93655037353)*(1+2.*(dim-1.)/(o+1.))+(-1.63113832851)*dim
        ref[3,2,1,0]=(1.1770401643)*(1+2.*(dim-1.)/(o+1.))+(1.03181665809)*dim
        ref[3,2,1,1]=(-1.57622476164)*(1+2.*(dim-1.)/(o+1.))+(1.02511612685)*dim
        ref[3,2,2,0]=(-0.141399619749)*(1+2.*(dim-1.)/(o+1.))+(-1.06965410781)*dim
        ref[3,2,2,1]=(1.67083909164)*(1+2.*(dim-1.)/(o+1.))+(0.102145518531)*dim
        ref[3,3,0,0]=(-0.394782759355)*(1+2.*(dim-1.)/(o+1.))+(-1.24000053008)*dim
        ref[3,3,0,1]=(0.868960808118)*(1+2.*(dim-1.)/(o+1.))+(0.00201758379073)*dim
        ref[3,3,1,0]=(1.00000919667)*(1+2.*(dim-1.)/(o+1.))+(-0.17230041321)*dim
        ref[3,3,1,1]=(1.64713319694)*(1+2.*(dim-1.)/(o+1.))+(0.638129241907)*dim
        ref[3,3,2,0]=(0.924755217374)*(1+2.*(dim-1.)/(o+1.))+(-1.14240472392)*dim
        ref[3,3,2,1]=(0.539870030002)*(1+2.*(dim-1.)/(o+1.))+(-1.57885882521)*dim
        ref[3,4,0,0]=(-1.11180611147)*(1+2.*(dim-1.)/(o+1.))+(0.0614250843596)*dim
        ref[3,4,0,1]=(0.199061341795)*(1+2.*(dim-1.)/(o+1.))+(-0.154806515534)*dim
        ref[3,4,1,0]=(0.616447492955)*(1+2.*(dim-1.)/(o+1.))+(1.0075475758)*dim
        ref[3,4,1,1]=(1.13843338631)*(1+2.*(dim-1.)/(o+1.))+(-1.20318894006)*dim
        ref[3,4,2,0]=(0.190554236123)*(1+2.*(dim-1.)/(o+1.))+(1.47013574405)*dim
        ref[3,4,2,1]=(-1.31155106991)*(1+2.*(dim-1.)/(o+1.))+(1.04945905035)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.416878137535)*x[0]**o+(0.0809721296208)*x[0]+(-0.275084069478)*x[1]**o+(-0.704226656249)*x[1]
        ref=(0.141794068057)*(1+2.*(dim-1.)/(o+1.))+(-0.623254526628)*dim
      else:
        arg=(0.587142789605)*x[0]**o+(0.894018391929)*x[0]+(-0.417709558814)*x[1]**o+(0.0970658484968)*x[1]+(-0.840982437679)*x[2]**o+(-0.400242562091)*x[2]
        ref=(-0.671549206889)*(1+2.*(dim-1.)/(o+1.))+(0.590841678335)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.051335729494)*x[0]**o+(0.257057962305)*x[0]+(0.00794214446212)*x[1]**o+(0.52307365061)*x[1]
        arg[1]=(0.419915493803)*x[0]**o+(-0.0105826268931)*x[0]+(0.369274332127)*x[1]**o+(0.340796701053)*x[1]
        ref[0]=(0.0592778739561)*(1+2.*(dim-1.)/(o+1.))+(0.780131612915)*dim
        ref[1]=(0.789189825931)*(1+2.*(dim-1.)/(o+1.))+(0.33021407416)*dim
      else:
        arg[0]=(-0.483302491019)*x[0]**o+(0.223892398776)*x[0]+(-0.555612541317)*x[1]**o+(0.75535374219)*x[1]+(-0.208573228788)*x[2]**o+(0.479879212564)*x[2]
        arg[1]=(-0.012341185885)*x[0]**o+(-0.959474230066)*x[0]+(0.397671123434)*x[1]**o+(-0.120340696903)*x[1]+(-0.645757653205)*x[2]**o+(-0.216057959957)*x[2]
        ref[0]=(-1.24748826112)*(1+2.*(dim-1.)/(o+1.))+(1.45912535353)*dim
        ref[1]=(-0.260427715657)*(1+2.*(dim-1.)/(o+1.))+(-1.29587288693)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.493237934301)*x[0]**o+(-0.977176505302)*x[0]+(-0.837740403623)*x[1]**o+(0.204709065496)*x[1]
        arg[0,1]=(0.794064508577)*x[0]**o+(-0.917060945623)*x[0]+(0.0114185601474)*x[1]**o+(0.547108694374)*x[1]
        arg[0,2]=(-0.989778984923)*x[0]**o+(-0.499651095311)*x[0]+(-0.126817505183)*x[1]**o+(0.889077385524)*x[1]
        arg[0,3]=(-0.223453594493)*x[0]**o+(-0.17983229649)*x[0]+(0.352602269148)*x[1]**o+(0.00617724258273)*x[1]
        arg[0,4]=(0.401784356952)*x[0]**o+(0.839308128877)*x[0]+(0.322943402455)*x[1]**o+(0.770206521944)*x[1]
        arg[1,0]=(0.491015078355)*x[0]**o+(0.194909202321)*x[0]+(0.38986016491)*x[1]**o+(0.671106836372)*x[1]
        arg[1,1]=(0.200697856784)*x[0]**o+(0.389746053244)*x[0]+(0.850237988574)*x[1]**o+(-0.677750684271)*x[1]
        arg[1,2]=(0.300444129016)*x[0]**o+(-0.502811932395)*x[0]+(0.390751425646)*x[1]**o+(-0.0234058348564)*x[1]
        arg[1,3]=(-0.451410318068)*x[0]**o+(0.0959092789621)*x[0]+(0.364602756931)*x[1]**o+(-0.637883149677)*x[1]
        arg[1,4]=(0.84148898142)*x[0]**o+(0.447701396385)*x[0]+(-0.925960934959)*x[1]**o+(-0.216915201171)*x[1]
        arg[2,0]=(0.497685540053)*x[0]**o+(0.706445401439)*x[0]+(0.278688226804)*x[1]**o+(0.263560807934)*x[1]
        arg[2,1]=(0.846700886302)*x[0]**o+(0.794459987626)*x[0]+(0.206707548962)*x[1]**o+(0.536938178952)*x[1]
        arg[2,2]=(-0.825060225713)*x[0]**o+(0.00650901742391)*x[0]+(0.500591036904)*x[1]**o+(0.86251606869)*x[1]
        arg[2,3]=(-0.490203731083)*x[0]**o+(-0.758885551614)*x[0]+(0.44212368096)*x[1]**o+(0.422183463853)*x[1]
        arg[2,4]=(0.326180837967)*x[0]**o+(0.0511693925274)*x[0]+(0.450713817213)*x[1]**o+(-0.788372992203)*x[1]
        arg[3,0]=(0.427328335694)*x[0]**o+(-0.549570782632)*x[0]+(0.0484849428866)*x[1]**o+(-0.488921618371)*x[1]
        arg[3,1]=(-0.000395237589224)*x[0]**o+(-0.902889768871)*x[0]+(-0.991821787559)*x[1]**o+(-0.907842202962)*x[1]
        arg[3,2]=(0.737156280391)*x[0]**o+(0.104714673958)*x[0]+(0.102528971835)*x[1]**o+(-0.0982626427735)*x[1]
        arg[3,3]=(0.971460723075)*x[0]**o+(-0.759821212391)*x[0]+(0.385963754923)*x[1]**o+(-0.533924725996)*x[1]
        arg[3,4]=(0.25440424638)*x[0]**o+(-0.82432588605)*x[0]+(0.759285911224)*x[1]**o+(0.328467600746)*x[1]
        ref[0,0]=(-1.33097833792)*(1+2.*(dim-1.)/(o+1.))+(-0.772467439806)*dim
        ref[0,1]=(0.805483068724)*(1+2.*(dim-1.)/(o+1.))+(-0.36995225125)*dim
        ref[0,2]=(-1.11659649011)*(1+2.*(dim-1.)/(o+1.))+(0.389426290213)*dim
        ref[0,3]=(0.129148674655)*(1+2.*(dim-1.)/(o+1.))+(-0.173655053907)*dim
        ref[0,4]=(0.724727759407)*(1+2.*(dim-1.)/(o+1.))+(1.60951465082)*dim
        ref[1,0]=(0.880875243265)*(1+2.*(dim-1.)/(o+1.))+(0.866016038693)*dim
        ref[1,1]=(1.05093584536)*(1+2.*(dim-1.)/(o+1.))+(-0.288004631027)*dim
        ref[1,2]=(0.691195554662)*(1+2.*(dim-1.)/(o+1.))+(-0.526217767251)*dim
        ref[1,3]=(-0.0868075611369)*(1+2.*(dim-1.)/(o+1.))+(-0.541973870715)*dim
        ref[1,4]=(-0.0844719535392)*(1+2.*(dim-1.)/(o+1.))+(0.230786195214)*dim
        ref[2,0]=(0.776373766857)*(1+2.*(dim-1.)/(o+1.))+(0.970006209374)*dim
        ref[2,1]=(1.05340843526)*(1+2.*(dim-1.)/(o+1.))+(1.33139816658)*dim
        ref[2,2]=(-0.324469188808)*(1+2.*(dim-1.)/(o+1.))+(0.869025086114)*dim
        ref[2,3]=(-0.0480800501232)*(1+2.*(dim-1.)/(o+1.))+(-0.336702087761)*dim
        ref[2,4]=(0.776894655179)*(1+2.*(dim-1.)/(o+1.))+(-0.737203599675)*dim
        ref[3,0]=(0.47581327858)*(1+2.*(dim-1.)/(o+1.))+(-1.038492401)*dim
        ref[3,1]=(-0.992217025148)*(1+2.*(dim-1.)/(o+1.))+(-1.81073197183)*dim
        ref[3,2]=(0.839685252226)*(1+2.*(dim-1.)/(o+1.))+(0.00645203118434)*dim
        ref[3,3]=(1.357424478)*(1+2.*(dim-1.)/(o+1.))+(-1.29374593839)*dim
        ref[3,4]=(1.0136901576)*(1+2.*(dim-1.)/(o+1.))+(-0.495858285305)*dim
      else:
        arg[0,0]=(-0.543785449843)*x[0]**o+(-0.593471310422)*x[0]+(0.223732084513)*x[1]**o+(-0.126390494354)*x[1]+(0.602561108649)*x[2]**o+(-0.879650196727)*x[2]
        arg[0,1]=(-0.373708985284)*x[0]**o+(0.274723212925)*x[0]+(0.0901068561926)*x[1]**o+(0.0705878267418)*x[1]+(-0.87124855534)*x[2]**o+(-0.342156987834)*x[2]
        arg[0,2]=(0.197662792758)*x[0]**o+(0.993389348528)*x[0]+(0.300080063659)*x[1]**o+(-0.87329423474)*x[1]+(-0.708861668092)*x[2]**o+(-0.145567510817)*x[2]
        arg[0,3]=(-0.641680604707)*x[0]**o+(0.502870155472)*x[0]+(0.726460257571)*x[1]**o+(-0.796246941548)*x[1]+(-0.034138694789)*x[2]**o+(-0.756773445728)*x[2]
        arg[0,4]=(-0.910301138852)*x[0]**o+(0.370603440687)*x[0]+(0.30908632679)*x[1]**o+(0.399307258883)*x[1]+(0.734329887794)*x[2]**o+(-0.105355323212)*x[2]
        arg[1,0]=(-0.998162269861)*x[0]**o+(0.744436497607)*x[0]+(0.474991680035)*x[1]**o+(-0.523297766774)*x[1]+(-0.147411103004)*x[2]**o+(-0.366260797475)*x[2]
        arg[1,1]=(-0.895068058812)*x[0]**o+(-0.543026567202)*x[0]+(0.681910579939)*x[1]**o+(0.839238385122)*x[1]+(-0.943454697555)*x[2]**o+(-0.942882687085)*x[2]
        arg[1,2]=(-0.0327934757673)*x[0]**o+(-0.551060110327)*x[0]+(0.0246419122114)*x[1]**o+(0.290492943846)*x[1]+(-0.81629260256)*x[2]**o+(-0.0703654573398)*x[2]
        arg[1,3]=(0.0101368428324)*x[0]**o+(0.169228100407)*x[0]+(-0.0720675504637)*x[1]**o+(0.854569517064)*x[1]+(-0.185495614033)*x[2]**o+(0.159402892807)*x[2]
        arg[1,4]=(-0.879056526056)*x[0]**o+(-0.499865495878)*x[0]+(0.389281340553)*x[1]**o+(0.208323326648)*x[1]+(0.201495185869)*x[2]**o+(0.287693342543)*x[2]
        arg[2,0]=(-0.0980367959573)*x[0]**o+(-0.743908923973)*x[0]+(0.612332291133)*x[1]**o+(-0.833162883242)*x[1]+(-0.149234146487)*x[2]**o+(-0.265352382043)*x[2]
        arg[2,1]=(-0.731053054407)*x[0]**o+(-0.328824749845)*x[0]+(-0.270771833154)*x[1]**o+(-0.924718425874)*x[1]+(-0.333737248999)*x[2]**o+(-0.437707530432)*x[2]
        arg[2,2]=(0.0568240255423)*x[0]**o+(0.889929679474)*x[0]+(-0.550346007257)*x[1]**o+(-0.276114383355)*x[1]+(0.818856335741)*x[2]**o+(0.949830759195)*x[2]
        arg[2,3]=(-0.487278099196)*x[0]**o+(0.621833922878)*x[0]+(-0.641187646288)*x[1]**o+(-0.62187991518)*x[1]+(0.193101177186)*x[2]**o+(0.574039256278)*x[2]
        arg[2,4]=(0.411046922073)*x[0]**o+(-0.10196277248)*x[0]+(-0.880065306034)*x[1]**o+(0.629582927604)*x[1]+(-0.364618567951)*x[2]**o+(0.130293365676)*x[2]
        arg[3,0]=(-0.625244224555)*x[0]**o+(0.50345243521)*x[0]+(0.701616120451)*x[1]**o+(0.44401843794)*x[1]+(-0.881068054195)*x[2]**o+(0.337405641389)*x[2]
        arg[3,1]=(0.244113319139)*x[0]**o+(0.739760310044)*x[0]+(0.439528183979)*x[1]**o+(0.808496622595)*x[1]+(0.432069187682)*x[2]**o+(-0.106394550529)*x[2]
        arg[3,2]=(-0.200973139796)*x[0]**o+(-0.373932231384)*x[0]+(-0.438946407503)*x[1]**o+(0.543079830762)*x[1]+(-0.499599010695)*x[2]**o+(0.501967674296)*x[2]
        arg[3,3]=(-0.128263000609)*x[0]**o+(-0.408181851435)*x[0]+(0.403315054177)*x[1]**o+(0.75067593555)*x[1]+(0.361607987217)*x[2]**o+(0.269483801463)*x[2]
        arg[3,4]=(0.821969994275)*x[0]**o+(0.350771424951)*x[0]+(-0.761235507511)*x[1]**o+(0.148329827876)*x[1]+(-0.918493845231)*x[2]**o+(0.592094636272)*x[2]
        ref[0,0]=(0.282507743319)*(1+2.*(dim-1.)/(o+1.))+(-1.5995120015)*dim
        ref[0,1]=(-1.15485068443)*(1+2.*(dim-1.)/(o+1.))+(0.00315405183214)*dim
        ref[0,2]=(-0.211118811675)*(1+2.*(dim-1.)/(o+1.))+(-0.0254723970291)*dim
        ref[0,3]=(0.0506409580755)*(1+2.*(dim-1.)/(o+1.))+(-1.0501502318)*dim
        ref[0,4]=(0.133115075732)*(1+2.*(dim-1.)/(o+1.))+(0.664555376358)*dim
        ref[1,0]=(-0.67058169283)*(1+2.*(dim-1.)/(o+1.))+(-0.145122066642)*dim
        ref[1,1]=(-1.15661217643)*(1+2.*(dim-1.)/(o+1.))+(-0.646670869164)*dim
        ref[1,2]=(-0.824444166116)*(1+2.*(dim-1.)/(o+1.))+(-0.330932623821)*dim
        ref[1,3]=(-0.247426321664)*(1+2.*(dim-1.)/(o+1.))+(1.18320051028)*dim
        ref[1,4]=(-0.288279999634)*(1+2.*(dim-1.)/(o+1.))+(-0.00384882668765)*dim
        ref[2,0]=(0.365061348688)*(1+2.*(dim-1.)/(o+1.))+(-1.84242418926)*dim
        ref[2,1]=(-1.33556213656)*(1+2.*(dim-1.)/(o+1.))+(-1.69125070615)*dim
        ref[2,2]=(0.325334354026)*(1+2.*(dim-1.)/(o+1.))+(1.56364605531)*dim
        ref[2,3]=(-0.935364568298)*(1+2.*(dim-1.)/(o+1.))+(0.573993263976)*dim
        ref[2,4]=(-0.833636951913)*(1+2.*(dim-1.)/(o+1.))+(0.6579135208)*dim
        ref[3,0]=(-0.804696158299)*(1+2.*(dim-1.)/(o+1.))+(1.28487651454)*dim
        ref[3,1]=(1.1157106908)*(1+2.*(dim-1.)/(o+1.))+(1.44186238211)*dim
        ref[3,2]=(-1.13951855799)*(1+2.*(dim-1.)/(o+1.))+(0.671115273674)*dim
        ref[3,3]=(0.636660040784)*(1+2.*(dim-1.)/(o+1.))+(0.611977885578)*dim
        ref[3,4]=(-0.857759358467)*(1+2.*(dim-1.)/(o+1.))+(1.0911958891)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.331954160625)*x[0]**o+(0.713423346586)*x[0]+(0.922333771405)*x[1]**o+(-0.306799323245)*x[1]
        arg[0,0,1]=(-0.837278273696)*x[0]**o+(-0.322068356249)*x[0]+(-0.952056283446)*x[1]**o+(0.78147215355)*x[1]
        arg[0,1,0]=(-0.71217197633)*x[0]**o+(-0.222121067189)*x[0]+(-0.740851323426)*x[1]**o+(-0.664415003166)*x[1]
        arg[0,1,1]=(0.304797844524)*x[0]**o+(-0.644642249615)*x[0]+(-0.942312793933)*x[1]**o+(0.758857258357)*x[1]
        arg[1,0,0]=(-0.978734634026)*x[0]**o+(0.697829375729)*x[0]+(-0.161023313194)*x[1]**o+(0.61960280308)*x[1]
        arg[1,0,1]=(0.65091302745)*x[0]**o+(-0.411592147008)*x[0]+(0.748017319704)*x[1]**o+(0.751661257345)*x[1]
        arg[1,1,0]=(-0.752749816617)*x[0]**o+(-0.889417028079)*x[0]+(0.187075769961)*x[1]**o+(0.96605423194)*x[1]
        arg[1,1,1]=(-0.791276410919)*x[0]**o+(-0.66910236639)*x[0]+(-0.107283654349)*x[1]**o+(0.556315348977)*x[1]
        arg[2,0,0]=(0.0717781211643)*x[0]**o+(-0.895974517834)*x[0]+(-0.388350799289)*x[1]**o+(0.55541171641)*x[1]
        arg[2,0,1]=(0.886475701949)*x[0]**o+(0.0794480579219)*x[0]+(-0.77671944551)*x[1]**o+(-0.29611930088)*x[1]
        arg[2,1,0]=(0.891530897911)*x[0]**o+(-0.767892439984)*x[0]+(-0.692871902002)*x[1]**o+(0.296346783431)*x[1]
        arg[2,1,1]=(0.904417221616)*x[0]**o+(-0.379208335246)*x[0]+(0.491675297059)*x[1]**o+(0.994131453216)*x[1]
        arg[3,0,0]=(-0.278750282129)*x[0]**o+(-0.889235749208)*x[0]+(-0.970025256676)*x[1]**o+(0.459265023633)*x[1]
        arg[3,0,1]=(-0.718548177793)*x[0]**o+(-0.801618785436)*x[0]+(-0.0489351007933)*x[1]**o+(-0.0331281021404)*x[1]
        arg[3,1,0]=(0.838645233036)*x[0]**o+(0.174830460073)*x[0]+(0.343887140274)*x[1]**o+(0.475643625956)*x[1]
        arg[3,1,1]=(-0.376629360434)*x[0]**o+(0.965850026388)*x[0]+(-0.62284011199)*x[1]**o+(-0.548261522438)*x[1]
        arg[4,0,0]=(-0.747109622101)*x[0]**o+(-0.793286186527)*x[0]+(0.700502366132)*x[1]**o+(0.708459489695)*x[1]
        arg[4,0,1]=(-0.372920704339)*x[0]**o+(-0.856662712404)*x[0]+(0.735191558102)*x[1]**o+(0.451959287084)*x[1]
        arg[4,1,0]=(-0.825227447648)*x[0]**o+(-0.978701927456)*x[0]+(0.856562109765)*x[1]**o+(0.226820057686)*x[1]
        arg[4,1,1]=(-0.883951567752)*x[0]**o+(0.346114896878)*x[0]+(0.22103721575)*x[1]**o+(0.766907412924)*x[1]
        arg[5,0,0]=(0.440140408264)*x[0]**o+(0.576442002932)*x[0]+(0.181136518153)*x[1]**o+(0.467290325887)*x[1]
        arg[5,0,1]=(0.323653297993)*x[0]**o+(-0.0169080219964)*x[0]+(-0.790889359324)*x[1]**o+(-0.752609210154)*x[1]
        arg[5,1,0]=(-0.771859095701)*x[0]**o+(0.392074466446)*x[0]+(0.672832544535)*x[1]**o+(-0.518196376899)*x[1]
        arg[5,1,1]=(-0.714106835785)*x[0]**o+(-0.392974811152)*x[0]+(0.477349412757)*x[1]**o+(-0.978738343032)*x[1]
        ref[0,0,0]=(0.59037961078)*(1+2.*(dim-1.)/(o+1.))+(0.406624023341)*dim
        ref[0,0,1]=(-1.78933455714)*(1+2.*(dim-1.)/(o+1.))+(0.459403797301)*dim
        ref[0,1,0]=(-1.45302329976)*(1+2.*(dim-1.)/(o+1.))+(-0.886536070355)*dim
        ref[0,1,1]=(-0.637514949409)*(1+2.*(dim-1.)/(o+1.))+(0.114215008741)*dim
        ref[1,0,0]=(-1.13975794722)*(1+2.*(dim-1.)/(o+1.))+(1.31743217881)*dim
        ref[1,0,1]=(1.39893034715)*(1+2.*(dim-1.)/(o+1.))+(0.340069110337)*dim
        ref[1,1,0]=(-0.565674046656)*(1+2.*(dim-1.)/(o+1.))+(0.076637203861)*dim
        ref[1,1,1]=(-0.898560065268)*(1+2.*(dim-1.)/(o+1.))+(-0.112787017413)*dim
        ref[2,0,0]=(-0.316572678125)*(1+2.*(dim-1.)/(o+1.))+(-0.340562801424)*dim
        ref[2,0,1]=(0.10975625644)*(1+2.*(dim-1.)/(o+1.))+(-0.216671242958)*dim
        ref[2,1,0]=(0.198658995909)*(1+2.*(dim-1.)/(o+1.))+(-0.471545656553)*dim
        ref[2,1,1]=(1.39609251868)*(1+2.*(dim-1.)/(o+1.))+(0.614923117971)*dim
        ref[3,0,0]=(-1.2487755388)*(1+2.*(dim-1.)/(o+1.))+(-0.429970725575)*dim
        ref[3,0,1]=(-0.767483278586)*(1+2.*(dim-1.)/(o+1.))+(-0.834746887577)*dim
        ref[3,1,0]=(1.18253237331)*(1+2.*(dim-1.)/(o+1.))+(0.650474086029)*dim
        ref[3,1,1]=(-0.999469472424)*(1+2.*(dim-1.)/(o+1.))+(0.41758850395)*dim
        ref[4,0,0]=(-0.0466072559697)*(1+2.*(dim-1.)/(o+1.))+(-0.0848266968319)*dim
        ref[4,0,1]=(0.362270853763)*(1+2.*(dim-1.)/(o+1.))+(-0.404703425319)*dim
        ref[4,1,0]=(0.0313346621166)*(1+2.*(dim-1.)/(o+1.))+(-0.75188186977)*dim
        ref[4,1,1]=(-0.662914352001)*(1+2.*(dim-1.)/(o+1.))+(1.1130223098)*dim
        ref[5,0,0]=(0.621276926417)*(1+2.*(dim-1.)/(o+1.))+(1.04373232882)*dim
        ref[5,0,1]=(-0.467236061331)*(1+2.*(dim-1.)/(o+1.))+(-0.76951723215)*dim
        ref[5,1,0]=(-0.0990265511653)*(1+2.*(dim-1.)/(o+1.))+(-0.126121910453)*dim
        ref[5,1,1]=(-0.236757423028)*(1+2.*(dim-1.)/(o+1.))+(-1.37171315418)*dim
      else:
        arg[0,0,0]=(-0.676850371385)*x[0]**o+(0.343194249293)*x[0]+(-0.289569238124)*x[1]**o+(0.546143025789)*x[1]+(-0.292429216951)*x[2]**o+(0.100638805907)*x[2]
        arg[0,0,1]=(-0.558403199885)*x[0]**o+(0.369265961569)*x[0]+(0.098539502761)*x[1]**o+(-0.555747072012)*x[1]+(0.0142505043744)*x[2]**o+(0.110677590143)*x[2]
        arg[0,1,0]=(0.480265324249)*x[0]**o+(-0.0801799981418)*x[0]+(-0.659040110213)*x[1]**o+(0.315685251032)*x[1]+(0.146222780181)*x[2]**o+(-0.307845817615)*x[2]
        arg[0,1,1]=(-0.469643195547)*x[0]**o+(0.631180947887)*x[0]+(0.0873749097691)*x[1]**o+(-0.907805635326)*x[1]+(-0.445238378297)*x[2]**o+(-0.583750385876)*x[2]
        arg[1,0,0]=(-0.519648347947)*x[0]**o+(-0.580046561107)*x[0]+(-0.474702629402)*x[1]**o+(-0.76698196707)*x[1]+(0.631290211156)*x[2]**o+(0.283150656338)*x[2]
        arg[1,0,1]=(0.828049802595)*x[0]**o+(-0.0996271417739)*x[0]+(0.13247948046)*x[1]**o+(0.64657332105)*x[1]+(0.901448888477)*x[2]**o+(0.895379597076)*x[2]
        arg[1,1,0]=(-0.197367099356)*x[0]**o+(-0.933210661759)*x[0]+(-0.0134349749548)*x[1]**o+(-0.651111981269)*x[1]+(0.256945439519)*x[2]**o+(0.528412689094)*x[2]
        arg[1,1,1]=(-0.922943172982)*x[0]**o+(-0.883102479542)*x[0]+(-0.0037345936911)*x[1]**o+(-0.598256159261)*x[1]+(-0.8562089409)*x[2]**o+(-0.298502617143)*x[2]
        arg[2,0,0]=(0.499285053237)*x[0]**o+(-0.347877602065)*x[0]+(0.783117300662)*x[1]**o+(0.0413016895448)*x[1]+(0.22555777414)*x[2]**o+(0.089246166469)*x[2]
        arg[2,0,1]=(-0.180602931202)*x[0]**o+(0.883958054663)*x[0]+(0.904974519387)*x[1]**o+(-0.169618307434)*x[1]+(0.361963890467)*x[2]**o+(-0.364236450353)*x[2]
        arg[2,1,0]=(0.332239090882)*x[0]**o+(-0.392239851855)*x[0]+(0.905142609029)*x[1]**o+(-0.471493540964)*x[1]+(-0.305327788141)*x[2]**o+(0.581213078834)*x[2]
        arg[2,1,1]=(0.481276189253)*x[0]**o+(0.645400809127)*x[0]+(-0.840844470697)*x[1]**o+(-0.47587396865)*x[1]+(0.134145050668)*x[2]**o+(-0.315081369951)*x[2]
        arg[3,0,0]=(-0.480497217552)*x[0]**o+(0.595345393671)*x[0]+(0.776315451392)*x[1]**o+(-0.396110510671)*x[1]+(0.0437422791996)*x[2]**o+(0.348675326203)*x[2]
        arg[3,0,1]=(0.867117243092)*x[0]**o+(-0.222483151231)*x[0]+(0.0856439550139)*x[1]**o+(0.355855801395)*x[1]+(-0.779380984653)*x[2]**o+(0.799440639884)*x[2]
        arg[3,1,0]=(-0.869234399353)*x[0]**o+(-0.747500724089)*x[0]+(-0.83154206968)*x[1]**o+(-0.771906212766)*x[1]+(-0.188734869619)*x[2]**o+(0.7312382704)*x[2]
        arg[3,1,1]=(-0.729677636148)*x[0]**o+(-0.358066533942)*x[0]+(-0.342651052704)*x[1]**o+(-0.531453801913)*x[1]+(-0.898300547827)*x[2]**o+(0.0387220517653)*x[2]
        arg[4,0,0]=(-0.380489041999)*x[0]**o+(0.517823660835)*x[0]+(0.722149782685)*x[1]**o+(0.510169845258)*x[1]+(-0.979340093133)*x[2]**o+(0.846050053391)*x[2]
        arg[4,0,1]=(0.390895580804)*x[0]**o+(0.122461788897)*x[0]+(-0.809243912452)*x[1]**o+(0.871689879803)*x[1]+(-0.991689466458)*x[2]**o+(-0.0373953689911)*x[2]
        arg[4,1,0]=(0.00670611033858)*x[0]**o+(-0.700084486793)*x[0]+(0.365358002826)*x[1]**o+(-0.928347441251)*x[1]+(-0.675081280461)*x[2]**o+(-0.262871553902)*x[2]
        arg[4,1,1]=(-0.566521655794)*x[0]**o+(-0.23325381374)*x[0]+(0.744538505444)*x[1]**o+(0.263970013855)*x[1]+(0.360984907343)*x[2]**o+(0.704696800059)*x[2]
        arg[5,0,0]=(0.754229123633)*x[0]**o+(-0.523353302691)*x[0]+(0.404181763418)*x[1]**o+(0.175414644029)*x[1]+(0.627455124887)*x[2]**o+(-0.450163514418)*x[2]
        arg[5,0,1]=(0.804547370453)*x[0]**o+(-0.748034210243)*x[0]+(-0.943261499953)*x[1]**o+(0.360715335834)*x[1]+(-0.952785406204)*x[2]**o+(0.188838771539)*x[2]
        arg[5,1,0]=(0.653730390007)*x[0]**o+(0.663746229033)*x[0]+(0.454911944273)*x[1]**o+(-0.548110740107)*x[1]+(0.799171265206)*x[2]**o+(-0.646356900812)*x[2]
        arg[5,1,1]=(0.533464266683)*x[0]**o+(-0.0650985333228)*x[0]+(-0.599880569232)*x[1]**o+(0.645120695827)*x[1]+(0.985316456657)*x[2]**o+(0.527848503917)*x[2]
        ref[0,0,0]=(-1.25884882646)*(1+2.*(dim-1.)/(o+1.))+(0.989976080988)*dim
        ref[0,0,1]=(-0.445613192749)*(1+2.*(dim-1.)/(o+1.))+(-0.0758035202994)*dim
        ref[0,1,0]=(-0.0325520057832)*(1+2.*(dim-1.)/(o+1.))+(-0.0723405647253)*dim
        ref[0,1,1]=(-0.827506664075)*(1+2.*(dim-1.)/(o+1.))+(-0.860375073315)*dim
        ref[1,0,0]=(-0.363060766193)*(1+2.*(dim-1.)/(o+1.))+(-1.06387787184)*dim
        ref[1,0,1]=(1.86197817153)*(1+2.*(dim-1.)/(o+1.))+(1.44232577635)*dim
        ref[1,1,0]=(0.0461433652081)*(1+2.*(dim-1.)/(o+1.))+(-1.05590995393)*dim
        ref[1,1,1]=(-1.78288670757)*(1+2.*(dim-1.)/(o+1.))+(-1.77986125595)*dim
        ref[2,0,0]=(1.50796012804)*(1+2.*(dim-1.)/(o+1.))+(-0.217329746051)*dim
        ref[2,0,1]=(1.08633547865)*(1+2.*(dim-1.)/(o+1.))+(0.350103296876)*dim
        ref[2,1,0]=(0.932053911771)*(1+2.*(dim-1.)/(o+1.))+(-0.282520313985)*dim
        ref[2,1,1]=(-0.225423230776)*(1+2.*(dim-1.)/(o+1.))+(-0.145554529474)*dim
        ref[3,0,0]=(0.33956051304)*(1+2.*(dim-1.)/(o+1.))+(0.547910209204)*dim
        ref[3,0,1]=(0.173380213454)*(1+2.*(dim-1.)/(o+1.))+(0.932813290049)*dim
        ref[3,1,0]=(-1.88951133865)*(1+2.*(dim-1.)/(o+1.))+(-0.788168666455)*dim
        ref[3,1,1]=(-1.97062923668)*(1+2.*(dim-1.)/(o+1.))+(-0.850798284089)*dim
        ref[4,0,0]=(-0.637679352447)*(1+2.*(dim-1.)/(o+1.))+(1.87404355948)*dim
        ref[4,0,1]=(-1.41003779811)*(1+2.*(dim-1.)/(o+1.))+(0.956756299709)*dim
        ref[4,1,0]=(-0.303017167297)*(1+2.*(dim-1.)/(o+1.))+(-1.89130348195)*dim
        ref[4,1,1]=(0.539001756993)*(1+2.*(dim-1.)/(o+1.))+(0.735413000173)*dim
        ref[5,0,0]=(1.78586601194)*(1+2.*(dim-1.)/(o+1.))+(-0.79810217308)*dim
        ref[5,0,1]=(-1.0914995357)*(1+2.*(dim-1.)/(o+1.))+(-0.198480102871)*dim
        ref[5,1,0]=(1.90781359949)*(1+2.*(dim-1.)/(o+1.))+(-0.530721411887)*dim
        ref[5,1,1]=(0.918900154108)*(1+2.*(dim-1.)/(o+1.))+(1.10787066642)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.181966612335)*x[0]**o+(-0.461883712153)*x[0]+(-0.420217253531)*x[1]**o+(0.52197109374)*x[1]
        arg[0,0,0,1]=(-0.122704378373)*x[0]**o+(0.549708706786)*x[0]+(-0.728108908074)*x[1]**o+(-0.839262748545)*x[1]
        arg[0,0,1,0]=(0.58557741427)*x[0]**o+(-0.970633305521)*x[0]+(0.215513381389)*x[1]**o+(-0.178877868196)*x[1]
        arg[0,0,1,1]=(0.646118894595)*x[0]**o+(0.839917605008)*x[0]+(-0.608843960266)*x[1]**o+(-0.840842208189)*x[1]
        arg[0,0,2,0]=(-0.126004594384)*x[0]**o+(0.663487355648)*x[0]+(0.801751176464)*x[1]**o+(-0.76620842256)*x[1]
        arg[0,0,2,1]=(-0.244425023768)*x[0]**o+(0.736809513316)*x[0]+(0.736776514098)*x[1]**o+(0.382648973627)*x[1]
        arg[0,1,0,0]=(-0.264460763052)*x[0]**o+(-0.325868872518)*x[0]+(-0.0636203008538)*x[1]**o+(-0.398059740187)*x[1]
        arg[0,1,0,1]=(0.443639751435)*x[0]**o+(0.523903519493)*x[0]+(-0.222575248757)*x[1]**o+(-0.168203019401)*x[1]
        arg[0,1,1,0]=(-0.329523829258)*x[0]**o+(-0.908408437436)*x[0]+(0.695441587034)*x[1]**o+(-0.236957609006)*x[1]
        arg[0,1,1,1]=(-0.57363301023)*x[0]**o+(-0.168677045141)*x[0]+(0.455235475098)*x[1]**o+(-0.498803217861)*x[1]
        arg[0,1,2,0]=(0.09906404243)*x[0]**o+(0.48682122702)*x[0]+(-0.250940180366)*x[1]**o+(-0.641013418008)*x[1]
        arg[0,1,2,1]=(-0.963193555161)*x[0]**o+(0.129868691037)*x[0]+(0.821192174481)*x[1]**o+(-0.245576703223)*x[1]
        arg[0,2,0,0]=(-0.867410251576)*x[0]**o+(0.170194043006)*x[0]+(0.837669366536)*x[1]**o+(0.285772890787)*x[1]
        arg[0,2,0,1]=(0.767089114154)*x[0]**o+(-0.425174482112)*x[0]+(-0.460840645159)*x[1]**o+(-0.314295694392)*x[1]
        arg[0,2,1,0]=(-0.932934920269)*x[0]**o+(0.399552332498)*x[0]+(-0.137764102275)*x[1]**o+(-0.0375128952657)*x[1]
        arg[0,2,1,1]=(0.507999826854)*x[0]**o+(-0.60826518686)*x[0]+(0.766838061511)*x[1]**o+(0.800723296773)*x[1]
        arg[0,2,2,0]=(0.372064864992)*x[0]**o+(0.29029476011)*x[0]+(0.203299629846)*x[1]**o+(-0.386357855884)*x[1]
        arg[0,2,2,1]=(0.661268311956)*x[0]**o+(0.60257115577)*x[0]+(0.88881711536)*x[1]**o+(0.335431084294)*x[1]
        arg[0,3,0,0]=(-0.756295024079)*x[0]**o+(-0.982472476783)*x[0]+(-0.264060825724)*x[1]**o+(0.082684535426)*x[1]
        arg[0,3,0,1]=(0.514580120205)*x[0]**o+(0.524816925461)*x[0]+(-0.745096223094)*x[1]**o+(-0.0796881815577)*x[1]
        arg[0,3,1,0]=(0.597556449974)*x[0]**o+(0.527680495979)*x[0]+(0.517638167762)*x[1]**o+(-0.357836534374)*x[1]
        arg[0,3,1,1]=(0.354366560101)*x[0]**o+(-0.617697325436)*x[0]+(0.339071431286)*x[1]**o+(0.299934132319)*x[1]
        arg[0,3,2,0]=(0.26563455123)*x[0]**o+(0.821495614519)*x[0]+(0.177496963332)*x[1]**o+(-0.154521746257)*x[1]
        arg[0,3,2,1]=(0.983477388632)*x[0]**o+(0.524806318175)*x[0]+(-0.604096007297)*x[1]**o+(0.350360164328)*x[1]
        arg[0,4,0,0]=(0.311369778768)*x[0]**o+(-0.648416457513)*x[0]+(0.463469290711)*x[1]**o+(0.0719628828709)*x[1]
        arg[0,4,0,1]=(-0.743081056195)*x[0]**o+(-0.803573973461)*x[0]+(-0.766982068691)*x[1]**o+(0.387240020398)*x[1]
        arg[0,4,1,0]=(-0.0715117411345)*x[0]**o+(-0.341871715212)*x[0]+(0.0593378446291)*x[1]**o+(-0.10183864979)*x[1]
        arg[0,4,1,1]=(0.538850329978)*x[0]**o+(0.0280201497914)*x[0]+(0.350206859589)*x[1]**o+(0.316831165676)*x[1]
        arg[0,4,2,0]=(-0.327852385434)*x[0]**o+(0.408866306659)*x[0]+(0.940378517318)*x[1]**o+(0.578309623908)*x[1]
        arg[0,4,2,1]=(0.378380112483)*x[0]**o+(0.696828716067)*x[0]+(-0.908233000582)*x[1]**o+(-0.328611483675)*x[1]
        arg[1,0,0,0]=(0.753991558436)*x[0]**o+(0.0542891140783)*x[0]+(-0.428934638249)*x[1]**o+(0.318155063446)*x[1]
        arg[1,0,0,1]=(-0.387609085343)*x[0]**o+(0.711023478158)*x[0]+(-0.969026331169)*x[1]**o+(0.456175434422)*x[1]
        arg[1,0,1,0]=(-0.468504469917)*x[0]**o+(-0.314756281356)*x[0]+(0.0145647044588)*x[1]**o+(0.074035761055)*x[1]
        arg[1,0,1,1]=(0.240749620446)*x[0]**o+(0.906736060537)*x[0]+(0.391320123368)*x[1]**o+(-0.261639387239)*x[1]
        arg[1,0,2,0]=(-0.461436333094)*x[0]**o+(-0.809258428817)*x[0]+(0.0670777162266)*x[1]**o+(-0.445304287721)*x[1]
        arg[1,0,2,1]=(-0.569794448172)*x[0]**o+(-0.0624052432724)*x[0]+(-0.759840810858)*x[1]**o+(-0.609106375016)*x[1]
        arg[1,1,0,0]=(0.751570143287)*x[0]**o+(0.78189919599)*x[0]+(-0.103246458909)*x[1]**o+(0.732156635312)*x[1]
        arg[1,1,0,1]=(-0.0126214545951)*x[0]**o+(-0.612853966263)*x[0]+(-0.238555698245)*x[1]**o+(-0.668009933698)*x[1]
        arg[1,1,1,0]=(0.816325838743)*x[0]**o+(0.545518357738)*x[0]+(0.893296287925)*x[1]**o+(0.694579919864)*x[1]
        arg[1,1,1,1]=(0.927638834795)*x[0]**o+(0.750347688466)*x[0]+(-0.15924462002)*x[1]**o+(-0.952995938892)*x[1]
        arg[1,1,2,0]=(0.970761253678)*x[0]**o+(0.205308852313)*x[0]+(0.909209906751)*x[1]**o+(-0.178635180174)*x[1]
        arg[1,1,2,1]=(-0.632065128479)*x[0]**o+(-0.300905398347)*x[0]+(0.865317638002)*x[1]**o+(0.886398250669)*x[1]
        arg[1,2,0,0]=(-0.0529293620539)*x[0]**o+(-0.909022003179)*x[0]+(0.235231916345)*x[1]**o+(-0.623380752415)*x[1]
        arg[1,2,0,1]=(0.873344987616)*x[0]**o+(0.144231546529)*x[0]+(0.307375566381)*x[1]**o+(-0.201617507084)*x[1]
        arg[1,2,1,0]=(-0.957945047658)*x[0]**o+(-0.432955474381)*x[0]+(0.0742288164869)*x[1]**o+(-0.106080430366)*x[1]
        arg[1,2,1,1]=(-0.664227561512)*x[0]**o+(0.0821007781774)*x[0]+(-0.541506076231)*x[1]**o+(-0.88307957937)*x[1]
        arg[1,2,2,0]=(0.754702675232)*x[0]**o+(0.175176916913)*x[0]+(0.519607893979)*x[1]**o+(-0.24129526204)*x[1]
        arg[1,2,2,1]=(-0.415660998528)*x[0]**o+(-0.964071393069)*x[0]+(0.458656303526)*x[1]**o+(0.752153826846)*x[1]
        arg[1,3,0,0]=(-0.15638393187)*x[0]**o+(0.0828824719)*x[0]+(-0.0846205011509)*x[1]**o+(-0.438939463736)*x[1]
        arg[1,3,0,1]=(0.222994482074)*x[0]**o+(-0.488426101986)*x[0]+(-0.799262651825)*x[1]**o+(0.0638880385288)*x[1]
        arg[1,3,1,0]=(0.690529298963)*x[0]**o+(-0.845248154386)*x[0]+(0.748635110096)*x[1]**o+(0.820882250869)*x[1]
        arg[1,3,1,1]=(-0.0989162396328)*x[0]**o+(-0.171252734498)*x[0]+(-0.363447812361)*x[1]**o+(-0.06937701403)*x[1]
        arg[1,3,2,0]=(-0.739647298124)*x[0]**o+(-0.83384135572)*x[0]+(0.546124768813)*x[1]**o+(0.415761104064)*x[1]
        arg[1,3,2,1]=(0.859487140106)*x[0]**o+(0.481214854724)*x[0]+(0.788685241166)*x[1]**o+(-0.776410333263)*x[1]
        arg[1,4,0,0]=(-0.312428038681)*x[0]**o+(-0.178596293803)*x[0]+(-0.229770546807)*x[1]**o+(-0.43765633326)*x[1]
        arg[1,4,0,1]=(-0.974305286188)*x[0]**o+(-0.978154435248)*x[0]+(-0.671799350916)*x[1]**o+(0.645287291694)*x[1]
        arg[1,4,1,0]=(-0.219711371151)*x[0]**o+(-0.432784911532)*x[0]+(0.552940388883)*x[1]**o+(-0.622742205614)*x[1]
        arg[1,4,1,1]=(0.677289749287)*x[0]**o+(-0.92825132284)*x[0]+(0.818767358259)*x[1]**o+(-0.426000012305)*x[1]
        arg[1,4,2,0]=(-0.507474696573)*x[0]**o+(0.0882151625837)*x[0]+(0.826611187078)*x[1]**o+(0.157884229299)*x[1]
        arg[1,4,2,1]=(0.902398378671)*x[0]**o+(-0.372679984773)*x[0]+(-0.998272135687)*x[1]**o+(0.292294917465)*x[1]
        arg[2,0,0,0]=(0.699035433674)*x[0]**o+(0.380640017665)*x[0]+(-0.650735771214)*x[1]**o+(-0.55187467604)*x[1]
        arg[2,0,0,1]=(0.703438074924)*x[0]**o+(0.795508285134)*x[0]+(-0.0650847718675)*x[1]**o+(-0.0250766383001)*x[1]
        arg[2,0,1,0]=(0.493174434488)*x[0]**o+(-0.775523541313)*x[0]+(0.23184619496)*x[1]**o+(0.343368328234)*x[1]
        arg[2,0,1,1]=(0.404974280918)*x[0]**o+(0.0673627086244)*x[0]+(-0.843957699955)*x[1]**o+(0.645895713789)*x[1]
        arg[2,0,2,0]=(0.985047143685)*x[0]**o+(0.969582300743)*x[0]+(0.929427858098)*x[1]**o+(-0.951787413112)*x[1]
        arg[2,0,2,1]=(0.618505056215)*x[0]**o+(0.685760720523)*x[0]+(0.14594727158)*x[1]**o+(0.816579780815)*x[1]
        arg[2,1,0,0]=(0.0824953418089)*x[0]**o+(-0.896753351387)*x[0]+(0.415841596602)*x[1]**o+(0.349436600222)*x[1]
        arg[2,1,0,1]=(-0.904246301273)*x[0]**o+(0.51403543319)*x[0]+(0.0950147832662)*x[1]**o+(0.0482220256459)*x[1]
        arg[2,1,1,0]=(0.653418547332)*x[0]**o+(0.703094950575)*x[0]+(-0.294924381355)*x[1]**o+(0.889408306649)*x[1]
        arg[2,1,1,1]=(-0.329361050221)*x[0]**o+(-0.93072021487)*x[0]+(0.0845194032793)*x[1]**o+(-0.310617126183)*x[1]
        arg[2,1,2,0]=(-0.282371473471)*x[0]**o+(0.387458491178)*x[0]+(0.643908016757)*x[1]**o+(0.303741991451)*x[1]
        arg[2,1,2,1]=(-0.416872502153)*x[0]**o+(0.373465764073)*x[0]+(-0.90689140487)*x[1]**o+(-0.351052228378)*x[1]
        arg[2,2,0,0]=(0.494198770026)*x[0]**o+(0.744917026846)*x[0]+(-0.806591346974)*x[1]**o+(-0.812396479226)*x[1]
        arg[2,2,0,1]=(0.230573331783)*x[0]**o+(-0.449815323831)*x[0]+(-0.662178492084)*x[1]**o+(0.0281786827584)*x[1]
        arg[2,2,1,0]=(-0.546296599764)*x[0]**o+(-0.457939754236)*x[0]+(-0.0432225723667)*x[1]**o+(0.0431706529852)*x[1]
        arg[2,2,1,1]=(-0.808174460895)*x[0]**o+(-0.473584203382)*x[0]+(0.567514752294)*x[1]**o+(-0.115780229631)*x[1]
        arg[2,2,2,0]=(-0.134425008633)*x[0]**o+(-0.103550646345)*x[0]+(-0.588484951305)*x[1]**o+(0.955148281194)*x[1]
        arg[2,2,2,1]=(0.550095417437)*x[0]**o+(0.943564321046)*x[0]+(-0.990366145789)*x[1]**o+(0.872652302953)*x[1]
        arg[2,3,0,0]=(0.209341033615)*x[0]**o+(-0.285197252362)*x[0]+(0.961111991429)*x[1]**o+(0.810022268034)*x[1]
        arg[2,3,0,1]=(0.238319839582)*x[0]**o+(-0.714990515235)*x[0]+(0.0925371590198)*x[1]**o+(0.938145796375)*x[1]
        arg[2,3,1,0]=(0.889068347007)*x[0]**o+(-0.301904392078)*x[0]+(0.41386941601)*x[1]**o+(-0.973115118701)*x[1]
        arg[2,3,1,1]=(-0.522083695846)*x[0]**o+(-0.582455208417)*x[0]+(0.725431346859)*x[1]**o+(-0.950157996033)*x[1]
        arg[2,3,2,0]=(-0.897838142917)*x[0]**o+(0.745488698358)*x[0]+(-0.38910727894)*x[1]**o+(0.614539961316)*x[1]
        arg[2,3,2,1]=(0.924500645496)*x[0]**o+(-0.68988446876)*x[0]+(-0.466238481805)*x[1]**o+(-0.957703006413)*x[1]
        arg[2,4,0,0]=(0.280692018949)*x[0]**o+(0.600320489083)*x[0]+(-0.402390231845)*x[1]**o+(0.906721159456)*x[1]
        arg[2,4,0,1]=(0.50594574787)*x[0]**o+(0.130986910768)*x[0]+(0.272643351394)*x[1]**o+(0.53790545383)*x[1]
        arg[2,4,1,0]=(-0.627308461808)*x[0]**o+(0.690560028115)*x[0]+(-0.455768647207)*x[1]**o+(0.134926199015)*x[1]
        arg[2,4,1,1]=(-0.501163279392)*x[0]**o+(0.980144824993)*x[0]+(-0.0358930723173)*x[1]**o+(0.720157452357)*x[1]
        arg[2,4,2,0]=(-0.609861117539)*x[0]**o+(0.965699461849)*x[0]+(0.305674136837)*x[1]**o+(-0.578333683729)*x[1]
        arg[2,4,2,1]=(-0.921051416627)*x[0]**o+(-0.703479746028)*x[0]+(-0.0189027931731)*x[1]**o+(0.403415497212)*x[1]
        arg[3,0,0,0]=(-0.846266714886)*x[0]**o+(-0.103621657972)*x[0]+(0.217340457606)*x[1]**o+(-0.537654290693)*x[1]
        arg[3,0,0,1]=(0.186251745403)*x[0]**o+(0.694458841658)*x[0]+(0.126837362883)*x[1]**o+(0.0997502401792)*x[1]
        arg[3,0,1,0]=(0.153737843143)*x[0]**o+(0.811590040899)*x[0]+(0.189363446699)*x[1]**o+(0.00688399857093)*x[1]
        arg[3,0,1,1]=(-0.912444421339)*x[0]**o+(-0.68504171198)*x[0]+(-0.820717153919)*x[1]**o+(-0.00475951426309)*x[1]
        arg[3,0,2,0]=(0.908522439455)*x[0]**o+(-0.911126936858)*x[0]+(0.0424198145962)*x[1]**o+(-0.352097158128)*x[1]
        arg[3,0,2,1]=(0.935127271967)*x[0]**o+(0.154700792222)*x[0]+(-0.283148208173)*x[1]**o+(-0.607541482501)*x[1]
        arg[3,1,0,0]=(-0.131329160543)*x[0]**o+(0.24170180662)*x[0]+(-0.00893591013587)*x[1]**o+(0.262635048958)*x[1]
        arg[3,1,0,1]=(-0.559200924561)*x[0]**o+(0.509310116425)*x[0]+(0.484486720282)*x[1]**o+(-0.89314080386)*x[1]
        arg[3,1,1,0]=(0.355094442114)*x[0]**o+(-0.748398915688)*x[0]+(-0.379187476624)*x[1]**o+(0.447409858108)*x[1]
        arg[3,1,1,1]=(0.170855109103)*x[0]**o+(-0.48925054508)*x[0]+(0.840559727559)*x[1]**o+(0.0359562471889)*x[1]
        arg[3,1,2,0]=(0.827586094521)*x[0]**o+(0.0231820503547)*x[0]+(0.457192172855)*x[1]**o+(0.826960230377)*x[1]
        arg[3,1,2,1]=(-0.962890676345)*x[0]**o+(-0.112439391098)*x[0]+(-0.865510686025)*x[1]**o+(-0.539059426684)*x[1]
        arg[3,2,0,0]=(-0.165030011664)*x[0]**o+(-0.285694090419)*x[0]+(-0.522579583443)*x[1]**o+(0.73946640768)*x[1]
        arg[3,2,0,1]=(0.809351962503)*x[0]**o+(-0.872047628102)*x[0]+(0.61415433255)*x[1]**o+(0.132268088986)*x[1]
        arg[3,2,1,0]=(-0.587823544835)*x[0]**o+(-0.416092975475)*x[0]+(0.829027070431)*x[1]**o+(0.805271620727)*x[1]
        arg[3,2,1,1]=(-0.0107935093527)*x[0]**o+(0.988434287979)*x[0]+(0.941743055822)*x[1]**o+(0.137807879334)*x[1]
        arg[3,2,2,0]=(-0.855931527408)*x[0]**o+(0.575075608849)*x[0]+(0.11916967111)*x[1]**o+(-0.472844336533)*x[1]
        arg[3,2,2,1]=(-0.542112514875)*x[0]**o+(0.081308081083)*x[0]+(-0.934575648869)*x[1]**o+(-0.218208369535)*x[1]
        arg[3,3,0,0]=(0.486635391195)*x[0]**o+(0.995728500002)*x[0]+(-0.686767180456)*x[1]**o+(0.246849483855)*x[1]
        arg[3,3,0,1]=(-0.331571848212)*x[0]**o+(0.801870431206)*x[0]+(-0.91142654055)*x[1]**o+(0.089854528457)*x[1]
        arg[3,3,1,0]=(-0.657455652936)*x[0]**o+(-0.721466128651)*x[0]+(0.0420989631458)*x[1]**o+(-0.982316453325)*x[1]
        arg[3,3,1,1]=(-0.709257565086)*x[0]**o+(0.689028730506)*x[0]+(-0.196516960475)*x[1]**o+(-0.216704841188)*x[1]
        arg[3,3,2,0]=(-0.910547650394)*x[0]**o+(-0.977471949987)*x[0]+(0.37319115353)*x[1]**o+(-0.248978955796)*x[1]
        arg[3,3,2,1]=(-0.71025804576)*x[0]**o+(0.373559860529)*x[0]+(-0.9064074733)*x[1]**o+(-0.542012058073)*x[1]
        arg[3,4,0,0]=(-0.649273580176)*x[0]**o+(0.525784948768)*x[0]+(-0.213929862516)*x[1]**o+(-0.162295172275)*x[1]
        arg[3,4,0,1]=(0.209784328978)*x[0]**o+(-0.972601183002)*x[0]+(-0.730886568265)*x[1]**o+(0.966415518759)*x[1]
        arg[3,4,1,0]=(-0.334060094293)*x[0]**o+(0.29371522739)*x[0]+(0.936156150108)*x[1]**o+(-0.257026769476)*x[1]
        arg[3,4,1,1]=(-0.0465451767064)*x[0]**o+(0.115525862669)*x[0]+(0.891280073228)*x[1]**o+(-0.0436233989979)*x[1]
        arg[3,4,2,0]=(-0.748697989123)*x[0]**o+(0.246672415445)*x[0]+(0.62642302161)*x[1]**o+(-0.411644161038)*x[1]
        arg[3,4,2,1]=(0.850132316298)*x[0]**o+(0.939463534845)*x[0]+(0.0437756618735)*x[1]**o+(-0.955198146074)*x[1]
        ref[0,0,0,0]=(-0.238250641196)*(1+2.*(dim-1.)/(o+1.))+(0.0600873815876)*dim
        ref[0,0,0,1]=(-0.850813286447)*(1+2.*(dim-1.)/(o+1.))+(-0.28955404176)*dim
        ref[0,0,1,0]=(0.801090795659)*(1+2.*(dim-1.)/(o+1.))+(-1.14951117372)*dim
        ref[0,0,1,1]=(0.0372749343294)*(1+2.*(dim-1.)/(o+1.))+(-0.000924603180129)*dim
        ref[0,0,2,0]=(0.675746582079)*(1+2.*(dim-1.)/(o+1.))+(-0.102721066912)*dim
        ref[0,0,2,1]=(0.49235149033)*(1+2.*(dim-1.)/(o+1.))+(1.11945848694)*dim
        ref[0,1,0,0]=(-0.328081063906)*(1+2.*(dim-1.)/(o+1.))+(-0.723928612706)*dim
        ref[0,1,0,1]=(0.221064502678)*(1+2.*(dim-1.)/(o+1.))+(0.355700500092)*dim
        ref[0,1,1,0]=(0.365917757777)*(1+2.*(dim-1.)/(o+1.))+(-1.14536604644)*dim
        ref[0,1,1,1]=(-0.118397535132)*(1+2.*(dim-1.)/(o+1.))+(-0.667480263002)*dim
        ref[0,1,2,0]=(-0.151876137936)*(1+2.*(dim-1.)/(o+1.))+(-0.154192190988)*dim
        ref[0,1,2,1]=(-0.14200138068)*(1+2.*(dim-1.)/(o+1.))+(-0.115708012186)*dim
        ref[0,2,0,0]=(-0.0297408850402)*(1+2.*(dim-1.)/(o+1.))+(0.455966933793)*dim
        ref[0,2,0,1]=(0.306248468995)*(1+2.*(dim-1.)/(o+1.))+(-0.739470176504)*dim
        ref[0,2,1,0]=(-1.07069902254)*(1+2.*(dim-1.)/(o+1.))+(0.362039437233)*dim
        ref[0,2,1,1]=(1.27483788836)*(1+2.*(dim-1.)/(o+1.))+(0.192458109913)*dim
        ref[0,2,2,0]=(0.575364494839)*(1+2.*(dim-1.)/(o+1.))+(-0.0960630957749)*dim
        ref[0,2,2,1]=(1.55008542732)*(1+2.*(dim-1.)/(o+1.))+(0.938002240064)*dim
        ref[0,3,0,0]=(-1.0203558498)*(1+2.*(dim-1.)/(o+1.))+(-0.899787941357)*dim
        ref[0,3,0,1]=(-0.230516102889)*(1+2.*(dim-1.)/(o+1.))+(0.445128743903)*dim
        ref[0,3,1,0]=(1.11519461774)*(1+2.*(dim-1.)/(o+1.))+(0.169843961604)*dim
        ref[0,3,1,1]=(0.693437991388)*(1+2.*(dim-1.)/(o+1.))+(-0.317763193117)*dim
        ref[0,3,2,0]=(0.443131514562)*(1+2.*(dim-1.)/(o+1.))+(0.666973868262)*dim
        ref[0,3,2,1]=(0.379381381335)*(1+2.*(dim-1.)/(o+1.))+(0.875166482503)*dim
        ref[0,4,0,0]=(0.774839069479)*(1+2.*(dim-1.)/(o+1.))+(-0.576453574642)*dim
        ref[0,4,0,1]=(-1.51006312489)*(1+2.*(dim-1.)/(o+1.))+(-0.416333953063)*dim
        ref[0,4,1,0]=(-0.0121738965054)*(1+2.*(dim-1.)/(o+1.))+(-0.443710365002)*dim
        ref[0,4,1,1]=(0.889057189568)*(1+2.*(dim-1.)/(o+1.))+(0.344851315468)*dim
        ref[0,4,2,0]=(0.612526131884)*(1+2.*(dim-1.)/(o+1.))+(0.987175930567)*dim
        ref[0,4,2,1]=(-0.529852888099)*(1+2.*(dim-1.)/(o+1.))+(0.368217232392)*dim
        ref[1,0,0,0]=(0.325056920187)*(1+2.*(dim-1.)/(o+1.))+(0.372444177524)*dim
        ref[1,0,0,1]=(-1.35663541651)*(1+2.*(dim-1.)/(o+1.))+(1.16719891258)*dim
        ref[1,0,1,0]=(-0.453939765458)*(1+2.*(dim-1.)/(o+1.))+(-0.240720520301)*dim
        ref[1,0,1,1]=(0.632069743814)*(1+2.*(dim-1.)/(o+1.))+(0.645096673298)*dim
        ref[1,0,2,0]=(-0.394358616867)*(1+2.*(dim-1.)/(o+1.))+(-1.25456271654)*dim
        ref[1,0,2,1]=(-1.32963525903)*(1+2.*(dim-1.)/(o+1.))+(-0.671511618288)*dim
        ref[1,1,0,0]=(0.648323684378)*(1+2.*(dim-1.)/(o+1.))+(1.5140558313)*dim
        ref[1,1,0,1]=(-0.251177152841)*(1+2.*(dim-1.)/(o+1.))+(-1.28086389996)*dim
        ref[1,1,1,0]=(1.70962212667)*(1+2.*(dim-1.)/(o+1.))+(1.2400982776)*dim
        ref[1,1,1,1]=(0.768394214775)*(1+2.*(dim-1.)/(o+1.))+(-0.202648250425)*dim
        ref[1,1,2,0]=(1.87997116043)*(1+2.*(dim-1.)/(o+1.))+(0.0266736721392)*dim
        ref[1,1,2,1]=(0.233252509524)*(1+2.*(dim-1.)/(o+1.))+(0.585492852322)*dim
        ref[1,2,0,0]=(0.182302554291)*(1+2.*(dim-1.)/(o+1.))+(-1.53240275559)*dim
        ref[1,2,0,1]=(1.180720554)*(1+2.*(dim-1.)/(o+1.))+(-0.0573859605549)*dim
        ref[1,2,1,0]=(-0.883716231171)*(1+2.*(dim-1.)/(o+1.))+(-0.539035904746)*dim
        ref[1,2,1,1]=(-1.20573363774)*(1+2.*(dim-1.)/(o+1.))+(-0.800978801193)*dim
        ref[1,2,2,0]=(1.27431056921)*(1+2.*(dim-1.)/(o+1.))+(-0.066118345127)*dim
        ref[1,2,2,1]=(0.0429953049979)*(1+2.*(dim-1.)/(o+1.))+(-0.211917566223)*dim
        ref[1,3,0,0]=(-0.241004433021)*(1+2.*(dim-1.)/(o+1.))+(-0.356056991836)*dim
        ref[1,3,0,1]=(-0.576268169751)*(1+2.*(dim-1.)/(o+1.))+(-0.424538063457)*dim
        ref[1,3,1,0]=(1.43916440906)*(1+2.*(dim-1.)/(o+1.))+(-0.0243659035174)*dim
        ref[1,3,1,1]=(-0.462364051993)*(1+2.*(dim-1.)/(o+1.))+(-0.240629748528)*dim
        ref[1,3,2,0]=(-0.193522529311)*(1+2.*(dim-1.)/(o+1.))+(-0.418080251656)*dim
        ref[1,3,2,1]=(1.64817238127)*(1+2.*(dim-1.)/(o+1.))+(-0.295195478539)*dim
        ref[1,4,0,0]=(-0.542198585489)*(1+2.*(dim-1.)/(o+1.))+(-0.616252627062)*dim
        ref[1,4,0,1]=(-1.6461046371)*(1+2.*(dim-1.)/(o+1.))+(-0.332867143554)*dim
        ref[1,4,1,0]=(0.333229017732)*(1+2.*(dim-1.)/(o+1.))+(-1.05552711715)*dim
        ref[1,4,1,1]=(1.49605710755)*(1+2.*(dim-1.)/(o+1.))+(-1.35425133515)*dim
        ref[1,4,2,0]=(0.319136490505)*(1+2.*(dim-1.)/(o+1.))+(0.246099391883)*dim
        ref[1,4,2,1]=(-0.0958737570155)*(1+2.*(dim-1.)/(o+1.))+(-0.0803850673082)*dim
        ref[2,0,0,0]=(0.0482996624604)*(1+2.*(dim-1.)/(o+1.))+(-0.171234658375)*dim
        ref[2,0,0,1]=(0.638353303057)*(1+2.*(dim-1.)/(o+1.))+(0.770431646834)*dim
        ref[2,0,1,0]=(0.725020629449)*(1+2.*(dim-1.)/(o+1.))+(-0.432155213079)*dim
        ref[2,0,1,1]=(-0.438983419037)*(1+2.*(dim-1.)/(o+1.))+(0.713258422414)*dim
        ref[2,0,2,0]=(1.91447500178)*(1+2.*(dim-1.)/(o+1.))+(0.0177948876308)*dim
        ref[2,0,2,1]=(0.764452327795)*(1+2.*(dim-1.)/(o+1.))+(1.50234050134)*dim
        ref[2,1,0,0]=(0.498336938411)*(1+2.*(dim-1.)/(o+1.))+(-0.547316751164)*dim
        ref[2,1,0,1]=(-0.809231518007)*(1+2.*(dim-1.)/(o+1.))+(0.562257458836)*dim
        ref[2,1,1,0]=(0.358494165977)*(1+2.*(dim-1.)/(o+1.))+(1.59250325722)*dim
        ref[2,1,1,1]=(-0.244841646941)*(1+2.*(dim-1.)/(o+1.))+(-1.24133734105)*dim
        ref[2,1,2,0]=(0.361536543287)*(1+2.*(dim-1.)/(o+1.))+(0.691200482629)*dim
        ref[2,1,2,1]=(-1.32376390702)*(1+2.*(dim-1.)/(o+1.))+(0.0224135356952)*dim
        ref[2,2,0,0]=(-0.312392576948)*(1+2.*(dim-1.)/(o+1.))+(-0.06747945238)*dim
        ref[2,2,0,1]=(-0.431605160301)*(1+2.*(dim-1.)/(o+1.))+(-0.421636641073)*dim
        ref[2,2,1,0]=(-0.589519172131)*(1+2.*(dim-1.)/(o+1.))+(-0.414769101251)*dim
        ref[2,2,1,1]=(-0.240659708601)*(1+2.*(dim-1.)/(o+1.))+(-0.589364433013)*dim
        ref[2,2,2,0]=(-0.722909959939)*(1+2.*(dim-1.)/(o+1.))+(0.851597634849)*dim
        ref[2,2,2,1]=(-0.440270728352)*(1+2.*(dim-1.)/(o+1.))+(1.816216624)*dim
        ref[2,3,0,0]=(1.17045302504)*(1+2.*(dim-1.)/(o+1.))+(0.524825015672)*dim
        ref[2,3,0,1]=(0.330856998602)*(1+2.*(dim-1.)/(o+1.))+(0.22315528114)*dim
        ref[2,3,1,0]=(1.30293776302)*(1+2.*(dim-1.)/(o+1.))+(-1.27501951078)*dim
        ref[2,3,1,1]=(0.203347651013)*(1+2.*(dim-1.)/(o+1.))+(-1.53261320445)*dim
        ref[2,3,2,0]=(-1.28694542186)*(1+2.*(dim-1.)/(o+1.))+(1.36002865967)*dim
        ref[2,3,2,1]=(0.458262163691)*(1+2.*(dim-1.)/(o+1.))+(-1.64758747517)*dim
        ref[2,4,0,0]=(-0.121698212896)*(1+2.*(dim-1.)/(o+1.))+(1.50704164854)*dim
        ref[2,4,0,1]=(0.778589099264)*(1+2.*(dim-1.)/(o+1.))+(0.668892364598)*dim
        ref[2,4,1,0]=(-1.08307710902)*(1+2.*(dim-1.)/(o+1.))+(0.82548622713)*dim
        ref[2,4,1,1]=(-0.537056351709)*(1+2.*(dim-1.)/(o+1.))+(1.70030227735)*dim
        ref[2,4,2,0]=(-0.304186980702)*(1+2.*(dim-1.)/(o+1.))+(0.38736577812)*dim
        ref[2,4,2,1]=(-0.9399542098)*(1+2.*(dim-1.)/(o+1.))+(-0.300064248815)*dim
        ref[3,0,0,0]=(-0.62892625728)*(1+2.*(dim-1.)/(o+1.))+(-0.641275948665)*dim
        ref[3,0,0,1]=(0.313089108286)*(1+2.*(dim-1.)/(o+1.))+(0.794209081837)*dim
        ref[3,0,1,0]=(0.343101289842)*(1+2.*(dim-1.)/(o+1.))+(0.81847403947)*dim
        ref[3,0,1,1]=(-1.73316157526)*(1+2.*(dim-1.)/(o+1.))+(-0.689801226243)*dim
        ref[3,0,2,0]=(0.950942254051)*(1+2.*(dim-1.)/(o+1.))+(-1.26322409499)*dim
        ref[3,0,2,1]=(0.651979063795)*(1+2.*(dim-1.)/(o+1.))+(-0.452840690279)*dim
        ref[3,1,0,0]=(-0.140265070679)*(1+2.*(dim-1.)/(o+1.))+(0.504336855579)*dim
        ref[3,1,0,1]=(-0.074714204279)*(1+2.*(dim-1.)/(o+1.))+(-0.383830687435)*dim
        ref[3,1,1,0]=(-0.0240930345095)*(1+2.*(dim-1.)/(o+1.))+(-0.30098905758)*dim
        ref[3,1,1,1]=(1.01141483666)*(1+2.*(dim-1.)/(o+1.))+(-0.453294297891)*dim
        ref[3,1,2,0]=(1.28477826738)*(1+2.*(dim-1.)/(o+1.))+(0.850142280731)*dim
        ref[3,1,2,1]=(-1.82840136237)*(1+2.*(dim-1.)/(o+1.))+(-0.651498817783)*dim
        ref[3,2,0,0]=(-0.687609595107)*(1+2.*(dim-1.)/(o+1.))+(0.453772317262)*dim
        ref[3,2,0,1]=(1.42350629505)*(1+2.*(dim-1.)/(o+1.))+(-0.739779539116)*dim
        ref[3,2,1,0]=(0.241203525596)*(1+2.*(dim-1.)/(o+1.))+(0.389178645253)*dim
        ref[3,2,1,1]=(0.930949546469)*(1+2.*(dim-1.)/(o+1.))+(1.12624216731)*dim
        ref[3,2,2,0]=(-0.736761856298)*(1+2.*(dim-1.)/(o+1.))+(0.102231272315)*dim
        ref[3,2,2,1]=(-1.47668816374)*(1+2.*(dim-1.)/(o+1.))+(-0.136900288452)*dim
        ref[3,3,0,0]=(-0.200131789262)*(1+2.*(dim-1.)/(o+1.))+(1.24257798386)*dim
        ref[3,3,0,1]=(-1.24299838876)*(1+2.*(dim-1.)/(o+1.))+(0.891724959663)*dim
        ref[3,3,1,0]=(-0.61535668979)*(1+2.*(dim-1.)/(o+1.))+(-1.70378258198)*dim
        ref[3,3,1,1]=(-0.905774525561)*(1+2.*(dim-1.)/(o+1.))+(0.472323889318)*dim
        ref[3,3,2,0]=(-0.537356496864)*(1+2.*(dim-1.)/(o+1.))+(-1.22645090578)*dim
        ref[3,3,2,1]=(-1.61666551906)*(1+2.*(dim-1.)/(o+1.))+(-0.168452197544)*dim
        ref[3,4,0,0]=(-0.863203442693)*(1+2.*(dim-1.)/(o+1.))+(0.363489776493)*dim
        ref[3,4,0,1]=(-0.521102239287)*(1+2.*(dim-1.)/(o+1.))+(-0.00618566424301)*dim
        ref[3,4,1,0]=(0.602096055815)*(1+2.*(dim-1.)/(o+1.))+(0.0366884579134)*dim
        ref[3,4,1,1]=(0.844734896522)*(1+2.*(dim-1.)/(o+1.))+(0.0719024636709)*dim
        ref[3,4,2,0]=(-0.122274967513)*(1+2.*(dim-1.)/(o+1.))+(-0.164971745593)*dim
        ref[3,4,2,1]=(0.893907978172)*(1+2.*(dim-1.)/(o+1.))+(-0.0157346112292)*dim
      else:
        arg[0,0,0,0]=(-0.0445963085788)*x[0]**o+(0.649401127647)*x[0]+(-0.949035518207)*x[1]**o+(0.620675629012)*x[1]+(0.379727527503)*x[2]**o+(0.0321033403935)*x[2]
        arg[0,0,0,1]=(-0.334483142061)*x[0]**o+(-0.408749824143)*x[0]+(-0.427042441291)*x[1]**o+(-0.263818177378)*x[1]+(-0.0490001897328)*x[2]**o+(0.925792628022)*x[2]
        arg[0,0,1,0]=(0.632498844224)*x[0]**o+(-0.756487058254)*x[0]+(-0.0277391041902)*x[1]**o+(0.615017171892)*x[1]+(-0.0956053270216)*x[2]**o+(-0.0383148218356)*x[2]
        arg[0,0,1,1]=(0.560189295143)*x[0]**o+(-0.565632451768)*x[0]+(0.131609132282)*x[1]**o+(0.689012872903)*x[1]+(-0.0542418297927)*x[2]**o+(-0.816694105058)*x[2]
        arg[0,0,2,0]=(-0.606523401606)*x[0]**o+(-0.129782638438)*x[0]+(0.286554604329)*x[1]**o+(-0.264164004519)*x[1]+(-0.145168356054)*x[2]**o+(0.346147680755)*x[2]
        arg[0,0,2,1]=(-0.847528471559)*x[0]**o+(-0.50641170288)*x[0]+(-0.773325001399)*x[1]**o+(0.517649361249)*x[1]+(-0.531000881434)*x[2]**o+(0.220222077563)*x[2]
        arg[0,1,0,0]=(-0.0946495378384)*x[0]**o+(-0.312650449416)*x[0]+(0.183321918505)*x[1]**o+(-0.359448777446)*x[1]+(0.713219915733)*x[2]**o+(-0.436742264028)*x[2]
        arg[0,1,0,1]=(0.856947337921)*x[0]**o+(0.497228192699)*x[0]+(0.351718755501)*x[1]**o+(-0.928210477058)*x[1]+(0.586853747677)*x[2]**o+(-0.915642724205)*x[2]
        arg[0,1,1,0]=(-0.745330469339)*x[0]**o+(-0.724610807566)*x[0]+(0.821804346007)*x[1]**o+(0.241991559492)*x[1]+(0.508380905493)*x[2]**o+(0.756706999359)*x[2]
        arg[0,1,1,1]=(-0.647952004012)*x[0]**o+(0.940964624139)*x[0]+(-0.620895246845)*x[1]**o+(0.577736885116)*x[1]+(0.228716730257)*x[2]**o+(0.336914427419)*x[2]
        arg[0,1,2,0]=(-0.671541115529)*x[0]**o+(0.153595444626)*x[0]+(0.420649971716)*x[1]**o+(-0.459078302559)*x[1]+(-0.583263239511)*x[2]**o+(-0.927250932344)*x[2]
        arg[0,1,2,1]=(-0.893954406097)*x[0]**o+(0.266297374506)*x[0]+(-0.777463374391)*x[1]**o+(0.0905062188555)*x[1]+(-0.668756964585)*x[2]**o+(-0.580980380472)*x[2]
        arg[0,2,0,0]=(0.424449538697)*x[0]**o+(0.0787502268502)*x[0]+(-0.645996834851)*x[1]**o+(-0.0409730346273)*x[1]+(0.835463212243)*x[2]**o+(0.716052012249)*x[2]
        arg[0,2,0,1]=(0.0241441154932)*x[0]**o+(-0.580242261617)*x[0]+(0.693484316968)*x[1]**o+(0.198598746328)*x[1]+(-0.586283866696)*x[2]**o+(-0.595008846645)*x[2]
        arg[0,2,1,0]=(0.074643270363)*x[0]**o+(0.454053362138)*x[0]+(0.861855819289)*x[1]**o+(-0.0310194739742)*x[1]+(-0.35738290172)*x[2]**o+(-0.400226698584)*x[2]
        arg[0,2,1,1]=(0.61780464269)*x[0]**o+(-0.702818770068)*x[0]+(0.0840653474306)*x[1]**o+(0.786842849442)*x[1]+(0.0704342755849)*x[2]**o+(0.386459053939)*x[2]
        arg[0,2,2,0]=(-0.815363379053)*x[0]**o+(-0.621823615812)*x[0]+(-0.13535005543)*x[1]**o+(-0.127080917834)*x[1]+(-0.178866000441)*x[2]**o+(-0.0753933476244)*x[2]
        arg[0,2,2,1]=(0.400194366678)*x[0]**o+(0.075003395019)*x[0]+(-0.35708995046)*x[1]**o+(-0.370028795924)*x[1]+(-0.674944523932)*x[2]**o+(-0.437373855225)*x[2]
        arg[0,3,0,0]=(-0.736317751792)*x[0]**o+(0.671627391036)*x[0]+(0.0768737429532)*x[1]**o+(-0.90036869218)*x[1]+(0.796767364925)*x[2]**o+(-0.185034722102)*x[2]
        arg[0,3,0,1]=(0.530921403378)*x[0]**o+(0.970385957718)*x[0]+(-0.75659968724)*x[1]**o+(-0.437200812366)*x[1]+(0.0519374612092)*x[2]**o+(0.657323406602)*x[2]
        arg[0,3,1,0]=(-0.109751102222)*x[0]**o+(-0.154990305064)*x[0]+(-0.502004367351)*x[1]**o+(0.535266424739)*x[1]+(-0.968926773395)*x[2]**o+(-0.138032126456)*x[2]
        arg[0,3,1,1]=(0.509506900641)*x[0]**o+(-0.872357145765)*x[0]+(-0.154129018225)*x[1]**o+(0.0190914956768)*x[1]+(0.119094726078)*x[2]**o+(0.29593674915)*x[2]
        arg[0,3,2,0]=(-0.133398979819)*x[0]**o+(-0.780852347888)*x[0]+(-0.346576066687)*x[1]**o+(0.187923380934)*x[1]+(0.939817544138)*x[2]**o+(-0.497081081638)*x[2]
        arg[0,3,2,1]=(-0.908783140711)*x[0]**o+(0.541754659297)*x[0]+(-0.297350066105)*x[1]**o+(0.118676031731)*x[1]+(0.985854358996)*x[2]**o+(-0.424715359194)*x[2]
        arg[0,4,0,0]=(0.341191520307)*x[0]**o+(-0.252381555804)*x[0]+(-0.307289805415)*x[1]**o+(0.762228901652)*x[1]+(0.390901852611)*x[2]**o+(0.147349139229)*x[2]
        arg[0,4,0,1]=(0.236625833503)*x[0]**o+(0.856322905978)*x[0]+(-0.160244298088)*x[1]**o+(-0.971940590481)*x[1]+(-0.0829366716357)*x[2]**o+(-0.687610358703)*x[2]
        arg[0,4,1,0]=(0.619591455739)*x[0]**o+(-0.62589432832)*x[0]+(-0.0275687591477)*x[1]**o+(0.367590198294)*x[1]+(0.745367285277)*x[2]**o+(0.232571513762)*x[2]
        arg[0,4,1,1]=(0.317669904199)*x[0]**o+(0.741734360559)*x[0]+(-0.923644886773)*x[1]**o+(-0.725994528938)*x[1]+(-0.949385908603)*x[2]**o+(-0.0515088731457)*x[2]
        arg[0,4,2,0]=(0.82547889841)*x[0]**o+(-0.926970312436)*x[0]+(0.946941744691)*x[1]**o+(-0.171582097545)*x[1]+(-0.185628570945)*x[2]**o+(0.40495393086)*x[2]
        arg[0,4,2,1]=(0.51879613184)*x[0]**o+(-0.517447781109)*x[0]+(-0.288826585096)*x[1]**o+(-0.736890998485)*x[1]+(-0.167119326999)*x[2]**o+(-0.653082372893)*x[2]
        arg[1,0,0,0]=(-0.879008868811)*x[0]**o+(-0.0613987509179)*x[0]+(-0.671288168832)*x[1]**o+(-0.959517297906)*x[1]+(0.292554039162)*x[2]**o+(0.942246368678)*x[2]
        arg[1,0,0,1]=(0.599956815218)*x[0]**o+(0.731601305574)*x[0]+(-0.61593947163)*x[1]**o+(-0.597924796707)*x[1]+(-0.791620695721)*x[2]**o+(-0.562978718552)*x[2]
        arg[1,0,1,0]=(0.448381207242)*x[0]**o+(-0.178004176549)*x[0]+(0.631841924913)*x[1]**o+(0.311836637044)*x[1]+(0.0619475841251)*x[2]**o+(0.606564490624)*x[2]
        arg[1,0,1,1]=(0.159907398582)*x[0]**o+(-0.278605461684)*x[0]+(0.822914081538)*x[1]**o+(0.75112795022)*x[1]+(0.91389895436)*x[2]**o+(0.929177149059)*x[2]
        arg[1,0,2,0]=(-0.472286972241)*x[0]**o+(0.00950883684535)*x[0]+(0.41110884679)*x[1]**o+(0.109422159345)*x[1]+(-0.707436024011)*x[2]**o+(0.795057014561)*x[2]
        arg[1,0,2,1]=(-0.0156784802605)*x[0]**o+(-0.657283760492)*x[0]+(-0.0422880440635)*x[1]**o+(-0.326436491982)*x[1]+(-0.543917764251)*x[2]**o+(-0.691698129266)*x[2]
        arg[1,1,0,0]=(-0.814424723604)*x[0]**o+(0.746801606056)*x[0]+(0.22025107532)*x[1]**o+(-0.785254149057)*x[1]+(-0.845637273475)*x[2]**o+(-0.468272754144)*x[2]
        arg[1,1,0,1]=(-0.0815177460988)*x[0]**o+(-0.47210691717)*x[0]+(-0.285964111717)*x[1]**o+(0.527387496664)*x[1]+(0.87059583429)*x[2]**o+(-0.543899395468)*x[2]
        arg[1,1,1,0]=(0.424163426295)*x[0]**o+(0.649692204305)*x[0]+(0.0570010318344)*x[1]**o+(-0.426047169951)*x[1]+(0.334750596342)*x[2]**o+(-0.746549570927)*x[2]
        arg[1,1,1,1]=(-0.120493849358)*x[0]**o+(0.567677110538)*x[0]+(-0.355891513526)*x[1]**o+(0.68069794787)*x[1]+(0.786560156246)*x[2]**o+(-0.422804642829)*x[2]
        arg[1,1,2,0]=(-0.80298202932)*x[0]**o+(-0.674766348751)*x[0]+(-0.29945072468)*x[1]**o+(0.385941328636)*x[1]+(-0.756272624651)*x[2]**o+(-0.607541099269)*x[2]
        arg[1,1,2,1]=(0.915020713521)*x[0]**o+(-0.875957199283)*x[0]+(0.0217597029367)*x[1]**o+(-0.0649461013997)*x[1]+(-0.302450259541)*x[2]**o+(0.738498325526)*x[2]
        arg[1,2,0,0]=(0.239905462913)*x[0]**o+(0.197348893262)*x[0]+(-0.151173414577)*x[1]**o+(0.368405558905)*x[1]+(0.891448032058)*x[2]**o+(0.874975769684)*x[2]
        arg[1,2,0,1]=(-0.304971614823)*x[0]**o+(-0.0166065674003)*x[0]+(0.883397157318)*x[1]**o+(-0.918803270424)*x[1]+(0.590875856583)*x[2]**o+(-0.360659400522)*x[2]
        arg[1,2,1,0]=(0.185368540041)*x[0]**o+(0.498866146124)*x[0]+(0.591601152982)*x[1]**o+(-0.986932631453)*x[1]+(-0.525706872664)*x[2]**o+(-0.0462993899877)*x[2]
        arg[1,2,1,1]=(0.260677894495)*x[0]**o+(-0.216847258131)*x[0]+(0.932876363707)*x[1]**o+(0.761630676808)*x[1]+(0.260889221066)*x[2]**o+(0.546259908582)*x[2]
        arg[1,2,2,0]=(-0.316803600202)*x[0]**o+(-0.898044300642)*x[0]+(-0.0865001657852)*x[1]**o+(0.760424275199)*x[1]+(0.771564532427)*x[2]**o+(-0.0852237819025)*x[2]
        arg[1,2,2,1]=(-0.54274371504)*x[0]**o+(0.0822524750896)*x[0]+(0.872655885939)*x[1]**o+(0.594054041006)*x[1]+(0.322150325252)*x[2]**o+(0.774020612668)*x[2]
        arg[1,3,0,0]=(0.034593806358)*x[0]**o+(0.541638857337)*x[0]+(0.291359214696)*x[1]**o+(-0.203872730757)*x[1]+(0.0312260615928)*x[2]**o+(0.62568815669)*x[2]
        arg[1,3,0,1]=(-0.31929050432)*x[0]**o+(-0.931756481553)*x[0]+(0.755537511071)*x[1]**o+(-0.182578278538)*x[1]+(0.00958902784528)*x[2]**o+(0.345520390047)*x[2]
        arg[1,3,1,0]=(-0.435083251076)*x[0]**o+(0.248332963857)*x[0]+(-0.122701495966)*x[1]**o+(0.669350594899)*x[1]+(-0.0758596665333)*x[2]**o+(0.0784997907526)*x[2]
        arg[1,3,1,1]=(-0.725479421981)*x[0]**o+(0.657242746138)*x[0]+(-0.199958263735)*x[1]**o+(-0.784165440236)*x[1]+(-0.550936676647)*x[2]**o+(-0.31907813038)*x[2]
        arg[1,3,2,0]=(0.657177346512)*x[0]**o+(-0.168053866293)*x[0]+(0.782142168258)*x[1]**o+(0.660514922866)*x[1]+(0.788106295165)*x[2]**o+(-0.254290645125)*x[2]
        arg[1,3,2,1]=(-0.215166688682)*x[0]**o+(0.743577328376)*x[0]+(-0.374827404054)*x[1]**o+(0.267885661143)*x[1]+(0.728818301014)*x[2]**o+(0.6828999278)*x[2]
        arg[1,4,0,0]=(0.371646403413)*x[0]**o+(0.00998777673275)*x[0]+(-0.313378345871)*x[1]**o+(-0.952580545501)*x[1]+(-0.680596974717)*x[2]**o+(-0.181426263871)*x[2]
        arg[1,4,0,1]=(-0.828437965469)*x[0]**o+(0.302443878606)*x[0]+(0.581310358652)*x[1]**o+(0.714092786618)*x[1]+(0.650179420347)*x[2]**o+(0.663482256421)*x[2]
        arg[1,4,1,0]=(0.220318191456)*x[0]**o+(0.443108440795)*x[0]+(-0.538413779146)*x[1]**o+(0.660052783252)*x[1]+(-0.846698606343)*x[2]**o+(-0.0447806256913)*x[2]
        arg[1,4,1,1]=(0.866921895372)*x[0]**o+(0.776291161128)*x[0]+(-0.385318933957)*x[1]**o+(0.841514060614)*x[1]+(-0.401465329431)*x[2]**o+(0.474475442638)*x[2]
        arg[1,4,2,0]=(-0.222535436657)*x[0]**o+(0.921341450452)*x[0]+(-0.765043798058)*x[1]**o+(-0.492946667792)*x[1]+(0.758817008648)*x[2]**o+(0.0950595321909)*x[2]
        arg[1,4,2,1]=(0.333884542089)*x[0]**o+(0.671265678974)*x[0]+(-0.719262407869)*x[1]**o+(0.498176251457)*x[1]+(-0.867726438404)*x[2]**o+(-0.872614195608)*x[2]
        arg[2,0,0,0]=(-0.303331399994)*x[0]**o+(-0.563887060818)*x[0]+(0.64537626236)*x[1]**o+(-0.657611314342)*x[1]+(0.800882746828)*x[2]**o+(0.700671892157)*x[2]
        arg[2,0,0,1]=(0.417691239931)*x[0]**o+(-0.822789801933)*x[0]+(0.712214951539)*x[1]**o+(0.160960575173)*x[1]+(0.0400205244239)*x[2]**o+(0.49892939112)*x[2]
        arg[2,0,1,0]=(-0.828677767124)*x[0]**o+(-0.347411290435)*x[0]+(-0.700177402223)*x[1]**o+(-0.42498634463)*x[1]+(0.984172081328)*x[2]**o+(-0.677168643292)*x[2]
        arg[2,0,1,1]=(-0.423091398412)*x[0]**o+(-0.786913239913)*x[0]+(-0.888102258019)*x[1]**o+(-0.119525894432)*x[1]+(-0.158513498081)*x[2]**o+(0.503403674219)*x[2]
        arg[2,0,2,0]=(0.547217868614)*x[0]**o+(-0.737401808325)*x[0]+(-0.865530737659)*x[1]**o+(-0.750671412749)*x[1]+(-0.06731420082)*x[2]**o+(0.658572191108)*x[2]
        arg[2,0,2,1]=(0.501785022047)*x[0]**o+(-0.0542441989122)*x[0]+(-0.379614025724)*x[1]**o+(-0.785618793316)*x[1]+(0.994476365391)*x[2]**o+(-0.436996817494)*x[2]
        arg[2,1,0,0]=(0.0329439903445)*x[0]**o+(0.688097592632)*x[0]+(0.554030860489)*x[1]**o+(0.60949736297)*x[1]+(0.499716171437)*x[2]**o+(0.501464282299)*x[2]
        arg[2,1,0,1]=(0.459629414019)*x[0]**o+(0.810658146454)*x[0]+(-0.902023596001)*x[1]**o+(-0.982998009297)*x[1]+(0.487736529196)*x[2]**o+(-0.509164692477)*x[2]
        arg[2,1,1,0]=(0.0818151896334)*x[0]**o+(-0.931023788966)*x[0]+(-0.217349745187)*x[1]**o+(-0.785586110912)*x[1]+(0.16783052001)*x[2]**o+(-0.515867681053)*x[2]
        arg[2,1,1,1]=(-0.758874975703)*x[0]**o+(0.219408816573)*x[0]+(0.508187203896)*x[1]**o+(0.306700464008)*x[1]+(-0.641891226105)*x[2]**o+(-0.137752888266)*x[2]
        arg[2,1,2,0]=(-0.952430879482)*x[0]**o+(0.179895758684)*x[0]+(0.502285174801)*x[1]**o+(-0.82300826605)*x[1]+(-0.267848001419)*x[2]**o+(-0.882529775244)*x[2]
        arg[2,1,2,1]=(0.371372868601)*x[0]**o+(-0.521801538326)*x[0]+(0.894906517432)*x[1]**o+(0.205799057307)*x[1]+(-0.315840560279)*x[2]**o+(0.65823096555)*x[2]
        arg[2,2,0,0]=(0.585088762354)*x[0]**o+(0.527662713737)*x[0]+(0.83511043292)*x[1]**o+(-0.263597549489)*x[1]+(-0.342210974079)*x[2]**o+(0.532149528802)*x[2]
        arg[2,2,0,1]=(0.985868978475)*x[0]**o+(0.162798954854)*x[0]+(-0.887525429296)*x[1]**o+(0.618905848201)*x[1]+(-0.213643353241)*x[2]**o+(0.0170513098359)*x[2]
        arg[2,2,1,0]=(-0.216402064475)*x[0]**o+(0.350853951704)*x[0]+(0.592219675443)*x[1]**o+(-0.915293916561)*x[1]+(0.160588856435)*x[2]**o+(-0.574686420146)*x[2]
        arg[2,2,1,1]=(0.233118497878)*x[0]**o+(0.482035958353)*x[0]+(-0.581971661478)*x[1]**o+(0.832450520208)*x[1]+(-0.430914444794)*x[2]**o+(-0.791658025494)*x[2]
        arg[2,2,2,0]=(-0.710516844)*x[0]**o+(0.0266391730256)*x[0]+(-0.154541997151)*x[1]**o+(-0.232843778379)*x[1]+(0.453522207498)*x[2]**o+(-0.688939516711)*x[2]
        arg[2,2,2,1]=(0.26860058468)*x[0]**o+(-0.969638661751)*x[0]+(0.623784614857)*x[1]**o+(0.686359051774)*x[1]+(0.677056704385)*x[2]**o+(0.156688176847)*x[2]
        arg[2,3,0,0]=(-0.0620909841534)*x[0]**o+(-0.364946660535)*x[0]+(-0.388275430806)*x[1]**o+(-0.947908160286)*x[1]+(0.0774302410344)*x[2]**o+(0.374469033649)*x[2]
        arg[2,3,0,1]=(0.164845147121)*x[0]**o+(0.480213031852)*x[0]+(0.58157030274)*x[1]**o+(0.882820521347)*x[1]+(0.584714825748)*x[2]**o+(0.0285154925376)*x[2]
        arg[2,3,1,0]=(0.664711234489)*x[0]**o+(0.247415435607)*x[0]+(0.741592287061)*x[1]**o+(0.963505986041)*x[1]+(-0.987844687003)*x[2]**o+(0.998294459659)*x[2]
        arg[2,3,1,1]=(-0.734529316764)*x[0]**o+(-0.177480004713)*x[0]+(0.923555424779)*x[1]**o+(0.727331498256)*x[1]+(0.452312367674)*x[2]**o+(0.00652243396085)*x[2]
        arg[2,3,2,0]=(-0.367377897953)*x[0]**o+(-0.157529442804)*x[0]+(-0.121881528621)*x[1]**o+(0.865116726884)*x[1]+(-0.200087079925)*x[2]**o+(0.930889629788)*x[2]
        arg[2,3,2,1]=(-0.900394562978)*x[0]**o+(-0.762334071565)*x[0]+(-0.723776245294)*x[1]**o+(0.332728994729)*x[1]+(0.390453779523)*x[2]**o+(0.0447350797932)*x[2]
        arg[2,4,0,0]=(0.178230443819)*x[0]**o+(0.0898732089623)*x[0]+(0.808448952186)*x[1]**o+(0.164579824494)*x[1]+(-0.553032263217)*x[2]**o+(-0.782141879825)*x[2]
        arg[2,4,0,1]=(-0.105582488753)*x[0]**o+(0.995415696949)*x[0]+(0.978971670365)*x[1]**o+(-0.977866735017)*x[1]+(0.156938966924)*x[2]**o+(-0.120871488936)*x[2]
        arg[2,4,1,0]=(0.652354601889)*x[0]**o+(0.381186764672)*x[0]+(-0.783479778972)*x[1]**o+(-0.158898829615)*x[1]+(0.794693790129)*x[2]**o+(0.914720093741)*x[2]
        arg[2,4,1,1]=(0.571506419376)*x[0]**o+(0.906086581152)*x[0]+(-0.798519792125)*x[1]**o+(-0.22840897962)*x[1]+(-0.750891721145)*x[2]**o+(-0.591014038221)*x[2]
        arg[2,4,2,0]=(0.515982716306)*x[0]**o+(0.299569326845)*x[0]+(-0.407423180268)*x[1]**o+(-0.252103363573)*x[1]+(-0.0228238994531)*x[2]**o+(-0.282150542844)*x[2]
        arg[2,4,2,1]=(0.5335415942)*x[0]**o+(-0.885728570489)*x[0]+(-0.750810131327)*x[1]**o+(0.604053940679)*x[1]+(-0.0946405866686)*x[2]**o+(-0.423845465234)*x[2]
        arg[3,0,0,0]=(-0.498025341136)*x[0]**o+(-0.296354416633)*x[0]+(-0.600874889655)*x[1]**o+(0.544042300882)*x[1]+(0.318784851038)*x[2]**o+(-0.801109111437)*x[2]
        arg[3,0,0,1]=(-0.492361949745)*x[0]**o+(-0.708188092599)*x[0]+(-0.863737808435)*x[1]**o+(0.674889485553)*x[1]+(-0.982254652233)*x[2]**o+(0.682172862189)*x[2]
        arg[3,0,1,0]=(-0.872692397532)*x[0]**o+(0.563852836655)*x[0]+(-0.0219331486599)*x[1]**o+(-0.858175679559)*x[1]+(0.802432631538)*x[2]**o+(-0.655545287387)*x[2]
        arg[3,0,1,1]=(-0.538117166854)*x[0]**o+(-0.62507518352)*x[0]+(0.101524298816)*x[1]**o+(0.590432448393)*x[1]+(0.668057302006)*x[2]**o+(0.405382330056)*x[2]
        arg[3,0,2,0]=(-0.322916455064)*x[0]**o+(0.950624857669)*x[0]+(0.485623846818)*x[1]**o+(-0.12611522301)*x[1]+(0.503586393973)*x[2]**o+(0.109451719709)*x[2]
        arg[3,0,2,1]=(0.0289965995645)*x[0]**o+(0.0101440661359)*x[0]+(-0.303827119958)*x[1]**o+(0.355269935141)*x[1]+(0.29889427526)*x[2]**o+(-0.707999988781)*x[2]
        arg[3,1,0,0]=(-0.541382521848)*x[0]**o+(0.534108535206)*x[0]+(0.875610773119)*x[1]**o+(-0.575849439451)*x[1]+(-0.0836939606164)*x[2]**o+(0.433678856889)*x[2]
        arg[3,1,0,1]=(0.0322576144027)*x[0]**o+(0.794219359514)*x[0]+(0.629442621265)*x[1]**o+(0.0517726053499)*x[1]+(-0.776050101686)*x[2]**o+(0.697910196708)*x[2]
        arg[3,1,1,0]=(-0.198211682395)*x[0]**o+(-0.926268101941)*x[0]+(-0.904760383993)*x[1]**o+(-0.820075667018)*x[1]+(-0.101262044428)*x[2]**o+(-0.0621221772941)*x[2]
        arg[3,1,1,1]=(-0.531324749521)*x[0]**o+(0.736017207718)*x[0]+(0.487299696259)*x[1]**o+(-0.00604423021253)*x[1]+(-0.635722590088)*x[2]**o+(-0.143501730927)*x[2]
        arg[3,1,2,0]=(-0.731421637848)*x[0]**o+(-0.0371648783851)*x[0]+(-0.290362466891)*x[1]**o+(-0.375048584106)*x[1]+(-0.854622532092)*x[2]**o+(-0.794295905388)*x[2]
        arg[3,1,2,1]=(-0.766383422105)*x[0]**o+(0.149248159984)*x[0]+(0.970103674649)*x[1]**o+(-0.117794368107)*x[1]+(0.0178775633047)*x[2]**o+(-0.221126829278)*x[2]
        arg[3,2,0,0]=(0.706586887601)*x[0]**o+(0.162520023447)*x[0]+(-0.731539880336)*x[1]**o+(-0.992558519713)*x[1]+(0.426569303014)*x[2]**o+(0.0451783819487)*x[2]
        arg[3,2,0,1]=(0.0392295508272)*x[0]**o+(-0.157929005382)*x[0]+(0.0636418899878)*x[1]**o+(0.892501759578)*x[1]+(-0.368161001401)*x[2]**o+(-0.715520794543)*x[2]
        arg[3,2,1,0]=(0.889667707133)*x[0]**o+(0.308604168817)*x[0]+(-0.275558663983)*x[1]**o+(-0.408210095215)*x[1]+(-0.0918000869138)*x[2]**o+(0.142754064645)*x[2]
        arg[3,2,1,1]=(0.585831954407)*x[0]**o+(-0.705004375516)*x[0]+(0.657277099061)*x[1]**o+(-0.972401794257)*x[1]+(-0.658049310396)*x[2]**o+(-0.553961555706)*x[2]
        arg[3,2,2,0]=(0.88916769079)*x[0]**o+(-0.1401162684)*x[0]+(-0.396496633317)*x[1]**o+(0.215029522669)*x[1]+(-0.619023294208)*x[2]**o+(-0.187670706968)*x[2]
        arg[3,2,2,1]=(-0.262983340384)*x[0]**o+(-0.763236528477)*x[0]+(-0.339720289405)*x[1]**o+(-0.283077066751)*x[1]+(0.625600522278)*x[2]**o+(-0.132697432298)*x[2]
        arg[3,3,0,0]=(-0.346408105223)*x[0]**o+(0.299760534041)*x[0]+(0.567951405594)*x[1]**o+(-0.880683463459)*x[1]+(0.784127947661)*x[2]**o+(0.268895744072)*x[2]
        arg[3,3,0,1]=(0.064159345936)*x[0]**o+(-0.685098798117)*x[0]+(-0.182310885124)*x[1]**o+(-0.700081075257)*x[1]+(0.543959312437)*x[2]**o+(0.125104170434)*x[2]
        arg[3,3,1,0]=(0.460375837083)*x[0]**o+(0.82355800923)*x[0]+(-0.186000244328)*x[1]**o+(0.8418355439)*x[1]+(-0.375269292115)*x[2]**o+(0.712431840155)*x[2]
        arg[3,3,1,1]=(0.57260867805)*x[0]**o+(-0.146405851776)*x[0]+(-0.50640779246)*x[1]**o+(0.741760309299)*x[1]+(0.069752390409)*x[2]**o+(-0.376724191471)*x[2]
        arg[3,3,2,0]=(0.795493408403)*x[0]**o+(0.396861648344)*x[0]+(-0.728716068325)*x[1]**o+(0.739330072416)*x[1]+(0.202019645585)*x[2]**o+(0.841240509566)*x[2]
        arg[3,3,2,1]=(-0.967372744621)*x[0]**o+(-0.334044549662)*x[0]+(-0.0686982670779)*x[1]**o+(0.180083293629)*x[1]+(-0.208697793793)*x[2]**o+(0.0305446253751)*x[2]
        arg[3,4,0,0]=(0.94168350996)*x[0]**o+(0.0545532586137)*x[0]+(-0.300430532207)*x[1]**o+(-0.960362884765)*x[1]+(0.738261342527)*x[2]**o+(-0.351477574571)*x[2]
        arg[3,4,0,1]=(-0.525893857873)*x[0]**o+(0.227252001588)*x[0]+(0.179831911446)*x[1]**o+(0.350952712138)*x[1]+(-0.671550635312)*x[2]**o+(-0.338834217909)*x[2]
        arg[3,4,1,0]=(-0.60610056656)*x[0]**o+(0.219192673102)*x[0]+(-0.801595717702)*x[1]**o+(-0.696853420225)*x[1]+(-0.495201990047)*x[2]**o+(-0.500683649414)*x[2]
        arg[3,4,1,1]=(-0.37343020332)*x[0]**o+(-0.585536591527)*x[0]+(0.81794997494)*x[1]**o+(-0.240232527585)*x[1]+(-0.538970681636)*x[2]**o+(-0.146105210473)*x[2]
        arg[3,4,2,0]=(0.857322149358)*x[0]**o+(-0.552847286493)*x[0]+(0.530970378294)*x[1]**o+(0.0746811386922)*x[1]+(0.946170976628)*x[2]**o+(-0.329916962713)*x[2]
        arg[3,4,2,1]=(-0.353323100685)*x[0]**o+(0.808796594618)*x[0]+(0.481450588496)*x[1]**o+(0.1202874944)*x[1]+(0.671293736572)*x[2]**o+(-0.905074286629)*x[2]
        ref[0,0,0,0]=(-0.613904299283)*(1+2.*(dim-1.)/(o+1.))+(1.30218009705)*dim
        ref[0,0,0,1]=(-0.810525773084)*(1+2.*(dim-1.)/(o+1.))+(0.253224626501)*dim
        ref[0,0,1,0]=(0.509154413012)*(1+2.*(dim-1.)/(o+1.))+(-0.179784708198)*dim
        ref[0,0,1,1]=(0.637556597632)*(1+2.*(dim-1.)/(o+1.))+(-0.693313683923)*dim
        ref[0,0,2,0]=(-0.465137153331)*(1+2.*(dim-1.)/(o+1.))+(-0.0477989622018)*dim
        ref[0,0,2,1]=(-2.15185435439)*(1+2.*(dim-1.)/(o+1.))+(0.231459735932)*dim
        ref[0,1,0,0]=(0.8018922964)*(1+2.*(dim-1.)/(o+1.))+(-1.10884149089)*dim
        ref[0,1,0,1]=(1.7955198411)*(1+2.*(dim-1.)/(o+1.))+(-1.34662500856)*dim
        ref[0,1,1,0]=(0.584854782162)*(1+2.*(dim-1.)/(o+1.))+(0.274087751285)*dim
        ref[0,1,1,1]=(-1.0401305206)*(1+2.*(dim-1.)/(o+1.))+(1.85561593667)*dim
        ref[0,1,2,0]=(-0.834154383324)*(1+2.*(dim-1.)/(o+1.))+(-1.23273379028)*dim
        ref[0,1,2,1]=(-2.34017474507)*(1+2.*(dim-1.)/(o+1.))+(-0.224176787111)*dim
        ref[0,2,0,0]=(0.613915916089)*(1+2.*(dim-1.)/(o+1.))+(0.753829204471)*dim
        ref[0,2,0,1]=(0.131344565765)*(1+2.*(dim-1.)/(o+1.))+(-0.976652361934)*dim
        ref[0,2,1,0]=(0.579116187931)*(1+2.*(dim-1.)/(o+1.))+(0.0228071895801)*dim
        ref[0,2,1,1]=(0.772304265705)*(1+2.*(dim-1.)/(o+1.))+(0.470483133313)*dim
        ref[0,2,2,0]=(-1.12957943492)*(1+2.*(dim-1.)/(o+1.))+(-0.82429788127)*dim
        ref[0,2,2,1]=(-0.631840107715)*(1+2.*(dim-1.)/(o+1.))+(-0.732399256131)*dim
        ref[0,3,0,0]=(0.137323356087)*(1+2.*(dim-1.)/(o+1.))+(-0.413776023245)*dim
        ref[0,3,0,1]=(-0.173740822652)*(1+2.*(dim-1.)/(o+1.))+(1.19050855195)*dim
        ref[0,3,1,0]=(-1.58068224297)*(1+2.*(dim-1.)/(o+1.))+(0.24224399322)*dim
        ref[0,3,1,1]=(0.474472608494)*(1+2.*(dim-1.)/(o+1.))+(-0.557328900938)*dim
        ref[0,3,2,0]=(0.459842497633)*(1+2.*(dim-1.)/(o+1.))+(-1.09001004859)*dim
        ref[0,3,2,1]=(-0.22027884782)*(1+2.*(dim-1.)/(o+1.))+(0.235715331835)*dim
        ref[0,4,0,0]=(0.424803567504)*(1+2.*(dim-1.)/(o+1.))+(0.657196485077)*dim
        ref[0,4,0,1]=(-0.00655513622033)*(1+2.*(dim-1.)/(o+1.))+(-0.803228043207)*dim
        ref[0,4,1,0]=(1.33738998187)*(1+2.*(dim-1.)/(o+1.))+(-0.0257326162636)*dim
        ref[0,4,1,1]=(-1.55536089118)*(1+2.*(dim-1.)/(o+1.))+(-0.0357690415245)*dim
        ref[0,4,2,0]=(1.58679207216)*(1+2.*(dim-1.)/(o+1.))+(-0.693598479122)*dim
        ref[0,4,2,1]=(0.0628502197447)*(1+2.*(dim-1.)/(o+1.))+(-1.90742115249)*dim
        ref[1,0,0,0]=(-1.25774299848)*(1+2.*(dim-1.)/(o+1.))+(-0.0786696801461)*dim
        ref[1,0,0,1]=(-0.807603352133)*(1+2.*(dim-1.)/(o+1.))+(-0.429302209684)*dim
        ref[1,0,1,0]=(1.14217071628)*(1+2.*(dim-1.)/(o+1.))+(0.740396951119)*dim
        ref[1,0,1,1]=(1.89672043448)*(1+2.*(dim-1.)/(o+1.))+(1.4016996376)*dim
        ref[1,0,2,0]=(-0.768614149462)*(1+2.*(dim-1.)/(o+1.))+(0.913988010751)*dim
        ref[1,0,2,1]=(-0.601884288575)*(1+2.*(dim-1.)/(o+1.))+(-1.67541838174)*dim
        ref[1,1,0,0]=(-1.43981092176)*(1+2.*(dim-1.)/(o+1.))+(-0.506725297145)*dim
        ref[1,1,0,1]=(0.503113976474)*(1+2.*(dim-1.)/(o+1.))+(-0.488618815974)*dim
        ref[1,1,1,0]=(0.815915054472)*(1+2.*(dim-1.)/(o+1.))+(-0.522904536573)*dim
        ref[1,1,1,1]=(0.310174793362)*(1+2.*(dim-1.)/(o+1.))+(0.825570415579)*dim
        ref[1,1,2,0]=(-1.85870537865)*(1+2.*(dim-1.)/(o+1.))+(-0.896366119383)*dim
        ref[1,1,2,1]=(0.634330156916)*(1+2.*(dim-1.)/(o+1.))+(-0.202404975156)*dim
        ref[1,2,0,0]=(0.980180080393)*(1+2.*(dim-1.)/(o+1.))+(1.44073022185)*dim
        ref[1,2,0,1]=(1.16930139908)*(1+2.*(dim-1.)/(o+1.))+(-1.29606923835)*dim
        ref[1,2,1,0]=(0.251262820359)*(1+2.*(dim-1.)/(o+1.))+(-0.534365875316)*dim
        ref[1,2,1,1]=(1.45444347927)*(1+2.*(dim-1.)/(o+1.))+(1.09104332726)*dim
        ref[1,2,2,0]=(0.36826076644)*(1+2.*(dim-1.)/(o+1.))+(-0.222843807345)*dim
        ref[1,2,2,1]=(0.652062496151)*(1+2.*(dim-1.)/(o+1.))+(1.45032712876)*dim
        ref[1,3,0,0]=(0.357179082647)*(1+2.*(dim-1.)/(o+1.))+(0.96345428327)*dim
        ref[1,3,0,1]=(0.445836034597)*(1+2.*(dim-1.)/(o+1.))+(-0.768814370045)*dim
        ref[1,3,1,0]=(-0.633644413575)*(1+2.*(dim-1.)/(o+1.))+(0.996183349509)*dim
        ref[1,3,1,1]=(-1.47637436236)*(1+2.*(dim-1.)/(o+1.))+(-0.446000824477)*dim
        ref[1,3,2,0]=(2.22742580994)*(1+2.*(dim-1.)/(o+1.))+(0.238170411449)*dim
        ref[1,3,2,1]=(0.138824208277)*(1+2.*(dim-1.)/(o+1.))+(1.69436291732)*dim
        ref[1,4,0,0]=(-0.622328917175)*(1+2.*(dim-1.)/(o+1.))+(-1.12401903264)*dim
        ref[1,4,0,1]=(0.40305181353)*(1+2.*(dim-1.)/(o+1.))+(1.68001892165)*dim
        ref[1,4,1,0]=(-1.16479419403)*(1+2.*(dim-1.)/(o+1.))+(1.05838059836)*dim
        ref[1,4,1,1]=(0.0801376319834)*(1+2.*(dim-1.)/(o+1.))+(2.09228066438)*dim
        ref[1,4,2,0]=(-0.228762226067)*(1+2.*(dim-1.)/(o+1.))+(0.523454314851)*dim
        ref[1,4,2,1]=(-1.25310430418)*(1+2.*(dim-1.)/(o+1.))+(0.296827734822)*dim
        ref[2,0,0,0]=(1.14292760919)*(1+2.*(dim-1.)/(o+1.))+(-0.520826483003)*dim
        ref[2,0,0,1]=(1.16992671589)*(1+2.*(dim-1.)/(o+1.))+(-0.162899835641)*dim
        ref[2,0,1,0]=(-0.544683088018)*(1+2.*(dim-1.)/(o+1.))+(-1.44956627836)*dim
        ref[2,0,1,1]=(-1.46970715451)*(1+2.*(dim-1.)/(o+1.))+(-0.403035460127)*dim
        ref[2,0,2,0]=(-0.385627069865)*(1+2.*(dim-1.)/(o+1.))+(-0.829501029966)*dim
        ref[2,0,2,1]=(1.11664736171)*(1+2.*(dim-1.)/(o+1.))+(-1.27685980972)*dim
        ref[2,1,0,0]=(1.08669102227)*(1+2.*(dim-1.)/(o+1.))+(1.7990592379)*dim
        ref[2,1,0,1]=(0.0453423472143)*(1+2.*(dim-1.)/(o+1.))+(-0.68150455532)*dim
        ref[2,1,1,0]=(0.0322959644566)*(1+2.*(dim-1.)/(o+1.))+(-2.23247758093)*dim
        ref[2,1,1,1]=(-0.892578997912)*(1+2.*(dim-1.)/(o+1.))+(0.388356392315)*dim
        ref[2,1,2,0]=(-0.7179937061)*(1+2.*(dim-1.)/(o+1.))+(-1.52564228261)*dim
        ref[2,1,2,1]=(0.950438825754)*(1+2.*(dim-1.)/(o+1.))+(0.342228484532)*dim
        ref[2,2,0,0]=(1.0779882212)*(1+2.*(dim-1.)/(o+1.))+(0.796214693049)*dim
        ref[2,2,0,1]=(-0.115299804062)*(1+2.*(dim-1.)/(o+1.))+(0.798756112891)*dim
        ref[2,2,1,0]=(0.536406467403)*(1+2.*(dim-1.)/(o+1.))+(-1.139126385)*dim
        ref[2,2,1,1]=(-0.779767608394)*(1+2.*(dim-1.)/(o+1.))+(0.522828453067)*dim
        ref[2,2,2,0]=(-0.411536633652)*(1+2.*(dim-1.)/(o+1.))+(-0.895144122064)*dim
        ref[2,2,2,1]=(1.56944190392)*(1+2.*(dim-1.)/(o+1.))+(-0.12659143313)*dim
        ref[2,3,0,0]=(-0.372936173925)*(1+2.*(dim-1.)/(o+1.))+(-0.938385787171)*dim
        ref[2,3,0,1]=(1.33113027561)*(1+2.*(dim-1.)/(o+1.))+(1.39154904574)*dim
        ref[2,3,1,0]=(0.418458834548)*(1+2.*(dim-1.)/(o+1.))+(2.20921588131)*dim
        ref[2,3,1,1]=(0.64133847569)*(1+2.*(dim-1.)/(o+1.))+(0.556373927504)*dim
        ref[2,3,2,0]=(-0.689346506499)*(1+2.*(dim-1.)/(o+1.))+(1.63847691387)*dim
        ref[2,3,2,1]=(-1.23371702875)*(1+2.*(dim-1.)/(o+1.))+(-0.384869997042)*dim
        ref[2,4,0,0]=(0.433647132788)*(1+2.*(dim-1.)/(o+1.))+(-0.527688846368)*dim
        ref[2,4,0,1]=(1.03032814854)*(1+2.*(dim-1.)/(o+1.))+(-0.103322527004)*dim
        ref[2,4,1,0]=(0.663568613046)*(1+2.*(dim-1.)/(o+1.))+(1.1370080288)*dim
        ref[2,4,1,1]=(-0.977905093893)*(1+2.*(dim-1.)/(o+1.))+(0.0866635633103)*dim
        ref[2,4,2,0]=(0.085735636585)*(1+2.*(dim-1.)/(o+1.))+(-0.234684579572)*dim
        ref[2,4,2,1]=(-0.311909123795)*(1+2.*(dim-1.)/(o+1.))+(-0.705520095044)*dim
        ref[3,0,0,0]=(-0.780115379753)*(1+2.*(dim-1.)/(o+1.))+(-0.553421227188)*dim
        ref[3,0,0,1]=(-2.33835441041)*(1+2.*(dim-1.)/(o+1.))+(0.648874255143)*dim
        ref[3,0,1,0]=(-0.0921929146543)*(1+2.*(dim-1.)/(o+1.))+(-0.949868130291)*dim
        ref[3,0,1,1]=(0.231464433969)*(1+2.*(dim-1.)/(o+1.))+(0.370739594929)*dim
        ref[3,0,2,0]=(0.666293785727)*(1+2.*(dim-1.)/(o+1.))+(0.933961354368)*dim
        ref[3,0,2,1]=(0.0240637548661)*(1+2.*(dim-1.)/(o+1.))+(-0.342585987504)*dim
        ref[3,1,0,0]=(0.250534290654)*(1+2.*(dim-1.)/(o+1.))+(0.391937952644)*dim
        ref[3,1,0,1]=(-0.114349866018)*(1+2.*(dim-1.)/(o+1.))+(1.54390216157)*dim
        ref[3,1,1,0]=(-1.20423411082)*(1+2.*(dim-1.)/(o+1.))+(-1.80846594625)*dim
        ref[3,1,1,1]=(-0.679747643349)*(1+2.*(dim-1.)/(o+1.))+(0.586471246578)*dim
        ref[3,1,2,0]=(-1.87640663683)*(1+2.*(dim-1.)/(o+1.))+(-1.20650936788)*dim
        ref[3,1,2,1]=(0.221597815849)*(1+2.*(dim-1.)/(o+1.))+(-0.189673037401)*dim
        ref[3,2,0,0]=(0.40161631028)*(1+2.*(dim-1.)/(o+1.))+(-0.784860114318)*dim
        ref[3,2,0,1]=(-0.265289560586)*(1+2.*(dim-1.)/(o+1.))+(0.0190519596518)*dim
        ref[3,2,1,0]=(0.522308956237)*(1+2.*(dim-1.)/(o+1.))+(0.0431481382473)*dim
        ref[3,2,1,1]=(0.585059743072)*(1+2.*(dim-1.)/(o+1.))+(-2.23136772548)*dim
        ref[3,2,2,0]=(-0.126352236735)*(1+2.*(dim-1.)/(o+1.))+(-0.112757452698)*dim
        ref[3,2,2,1]=(0.0228968924883)*(1+2.*(dim-1.)/(o+1.))+(-1.17901102753)*dim
        ref[3,3,0,0]=(1.00567124803)*(1+2.*(dim-1.)/(o+1.))+(-0.312027185345)*dim
        ref[3,3,0,1]=(0.425807773249)*(1+2.*(dim-1.)/(o+1.))+(-1.26007570294)*dim
        ref[3,3,1,0]=(-0.10089369936)*(1+2.*(dim-1.)/(o+1.))+(2.37782539328)*dim
        ref[3,3,1,1]=(0.135953275999)*(1+2.*(dim-1.)/(o+1.))+(0.218630266051)*dim
        ref[3,3,2,0]=(0.268796985663)*(1+2.*(dim-1.)/(o+1.))+(1.97743223033)*dim
        ref[3,3,2,1]=(-1.24476880549)*(1+2.*(dim-1.)/(o+1.))+(-0.123416630658)*dim
        ref[3,4,0,0]=(1.37951432028)*(1+2.*(dim-1.)/(o+1.))+(-1.25728720072)*dim
        ref[3,4,0,1]=(-1.01761258174)*(1+2.*(dim-1.)/(o+1.))+(0.239370495817)*dim
        ref[3,4,1,0]=(-1.90289827431)*(1+2.*(dim-1.)/(o+1.))+(-0.978344396537)*dim
        ref[3,4,1,1]=(-0.0944509100161)*(1+2.*(dim-1.)/(o+1.))+(-0.971874329585)*dim
        ref[3,4,2,0]=(2.33446350428)*(1+2.*(dim-1.)/(o+1.))+(-0.808083110513)*dim
        ref[3,4,2,1]=(0.799421224383)*(1+2.*(dim-1.)/(o+1.))+(0.0240098023897)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.626062912046)*x[0]**o+(0.600712856149)*x[0]+(0.0478861882442)*x[1]**o+(0.77008552734)*x[1]
        ref=(0.67394910029)*(1+2.*(dim-1.)/(o+1.))+(1.37079838349)*dim
      else:
        arg=(0.85588880126)*x[0]**o+(-0.898804533462)*x[0]+(-0.270816921088)*x[1]**o+(-0.842823912373)*x[1]+(-0.47559690762)*x[2]**o+(0.751770125129)*x[2]
        ref=(0.109474972552)*(1+2.*(dim-1.)/(o+1.))+(-0.989858320706)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.298913063662)*x[0]**o+(-0.250822875451)*x[0]+(0.707311253225)*x[1]**o+(0.0408091445557)*x[1]
        arg[1]=(-0.70143850042)*x[0]**o+(-0.80121227022)*x[0]+(0.0731514590408)*x[1]**o+(0.621969876113)*x[1]
        ref[0]=(1.00622431689)*(1+2.*(dim-1.)/(o+1.))+(-0.210013730895)*dim
        ref[1]=(-0.628287041379)*(1+2.*(dim-1.)/(o+1.))+(-0.179242394108)*dim
      else:
        arg[0]=(0.676976938336)*x[0]**o+(-0.928207054947)*x[0]+(0.864749399199)*x[1]**o+(0.864358390848)*x[1]+(-0.0255569869984)*x[2]**o+(0.0268536872793)*x[2]
        arg[1]=(0.270542056965)*x[0]**o+(0.0696194435081)*x[0]+(0.401360001043)*x[1]**o+(0.917064528874)*x[1]+(-0.755510144559)*x[2]**o+(0.881244515908)*x[2]
        ref[0]=(1.51616935054)*(1+2.*(dim-1.)/(o+1.))+(-0.0369949768193)*dim
        ref[1]=(-0.083608086551)*(1+2.*(dim-1.)/(o+1.))+(1.86792848829)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.381843423534)*x[0]**o+(-0.563210012057)*x[0]+(0.730631055205)*x[1]**o+(0.13636238698)*x[1]
        arg[0,1]=(0.306357653338)*x[0]**o+(-0.797201232101)*x[0]+(-0.66333793372)*x[1]**o+(-0.943650360464)*x[1]
        arg[0,2]=(-0.34058649734)*x[0]**o+(-0.27047792232)*x[0]+(0.57156687267)*x[1]**o+(-0.659114355468)*x[1]
        arg[0,3]=(0.73319004289)*x[0]**o+(0.229139540294)*x[0]+(-0.51558765225)*x[1]**o+(0.45052521066)*x[1]
        arg[0,4]=(-0.264117198359)*x[0]**o+(0.807950109095)*x[0]+(-0.290558748289)*x[1]**o+(0.161864708045)*x[1]
        arg[1,0]=(0.988732415297)*x[0]**o+(-0.725047414136)*x[0]+(-0.722337599885)*x[1]**o+(0.739130135256)*x[1]
        arg[1,1]=(-0.315086233906)*x[0]**o+(-0.783177846939)*x[0]+(-0.271931454399)*x[1]**o+(-0.935744613016)*x[1]
        arg[1,2]=(-0.787047622811)*x[0]**o+(0.00746242211892)*x[0]+(-0.539716273578)*x[1]**o+(-0.340449115622)*x[1]
        arg[1,3]=(-0.782830405823)*x[0]**o+(0.842318742811)*x[0]+(-0.118894493096)*x[1]**o+(0.645943277412)*x[1]
        arg[1,4]=(0.0594068367647)*x[0]**o+(-0.184711682975)*x[0]+(-0.0643050731295)*x[1]**o+(-0.981593287261)*x[1]
        arg[2,0]=(0.17170754931)*x[0]**o+(0.582651363494)*x[0]+(0.334422749985)*x[1]**o+(0.852747429144)*x[1]
        arg[2,1]=(0.683967516208)*x[0]**o+(-0.612989314204)*x[0]+(0.845175135723)*x[1]**o+(0.374223395314)*x[1]
        arg[2,2]=(0.588917536035)*x[0]**o+(0.0130713873939)*x[0]+(0.881942872071)*x[1]**o+(0.683941046151)*x[1]
        arg[2,3]=(-0.63861264887)*x[0]**o+(0.453919752415)*x[0]+(0.448731342326)*x[1]**o+(-0.259183783496)*x[1]
        arg[2,4]=(-0.132175638549)*x[0]**o+(0.710071566415)*x[0]+(-0.870119332524)*x[1]**o+(0.523394538403)*x[1]
        arg[3,0]=(0.768036036634)*x[0]**o+(-0.218321673662)*x[0]+(0.50224512768)*x[1]**o+(0.694191572559)*x[1]
        arg[3,1]=(-0.298369075336)*x[0]**o+(-0.276776010079)*x[0]+(-0.659309162004)*x[1]**o+(0.436841182692)*x[1]
        arg[3,2]=(-0.00370049871499)*x[0]**o+(-0.41700259949)*x[0]+(0.726806931339)*x[1]**o+(-0.113165181924)*x[1]
        arg[3,3]=(0.288175924931)*x[0]**o+(-0.489667003521)*x[0]+(-0.531842953651)*x[1]**o+(-0.443517285929)*x[1]
        arg[3,4]=(-0.645574931589)*x[0]**o+(0.464738749544)*x[0]+(-0.538213524368)*x[1]**o+(-0.94443292146)*x[1]
        ref[0,0]=(0.348787631671)*(1+2.*(dim-1.)/(o+1.))+(-0.426847625077)*dim
        ref[0,1]=(-0.356980280382)*(1+2.*(dim-1.)/(o+1.))+(-1.74085159257)*dim
        ref[0,2]=(0.23098037533)*(1+2.*(dim-1.)/(o+1.))+(-0.929592277787)*dim
        ref[0,3]=(0.21760239064)*(1+2.*(dim-1.)/(o+1.))+(0.679664750954)*dim
        ref[0,4]=(-0.554675946648)*(1+2.*(dim-1.)/(o+1.))+(0.96981481714)*dim
        ref[1,0]=(0.266394815412)*(1+2.*(dim-1.)/(o+1.))+(0.0140827211195)*dim
        ref[1,1]=(-0.587017688304)*(1+2.*(dim-1.)/(o+1.))+(-1.71892245996)*dim
        ref[1,2]=(-1.32676389639)*(1+2.*(dim-1.)/(o+1.))+(-0.332986693503)*dim
        ref[1,3]=(-0.901724898919)*(1+2.*(dim-1.)/(o+1.))+(1.48826202022)*dim
        ref[1,4]=(-0.00489823636486)*(1+2.*(dim-1.)/(o+1.))+(-1.16630497024)*dim
        ref[2,0]=(0.506130299295)*(1+2.*(dim-1.)/(o+1.))+(1.43539879264)*dim
        ref[2,1]=(1.52914265193)*(1+2.*(dim-1.)/(o+1.))+(-0.238765918889)*dim
        ref[2,2]=(1.47086040811)*(1+2.*(dim-1.)/(o+1.))+(0.697012433545)*dim
        ref[2,3]=(-0.189881306544)*(1+2.*(dim-1.)/(o+1.))+(0.194735968919)*dim
        ref[2,4]=(-1.00229497107)*(1+2.*(dim-1.)/(o+1.))+(1.23346610482)*dim
        ref[3,0]=(1.27028116431)*(1+2.*(dim-1.)/(o+1.))+(0.475869898897)*dim
        ref[3,1]=(-0.957678237339)*(1+2.*(dim-1.)/(o+1.))+(0.160065172613)*dim
        ref[3,2]=(0.723106432624)*(1+2.*(dim-1.)/(o+1.))+(-0.530167781414)*dim
        ref[3,3]=(-0.243667028719)*(1+2.*(dim-1.)/(o+1.))+(-0.93318428945)*dim
        ref[3,4]=(-1.18378845596)*(1+2.*(dim-1.)/(o+1.))+(-0.479694171917)*dim
      else:
        arg[0,0]=(0.751222317885)*x[0]**o+(0.953666267123)*x[0]+(-0.707417048158)*x[1]**o+(0.310686543483)*x[1]+(0.326267797641)*x[2]**o+(-0.722772500437)*x[2]
        arg[0,1]=(-0.038709086264)*x[0]**o+(-0.483307819428)*x[0]+(0.611120835303)*x[1]**o+(0.885935954604)*x[1]+(0.630199910046)*x[2]**o+(-0.101764126615)*x[2]
        arg[0,2]=(0.890606731439)*x[0]**o+(-0.0555491679283)*x[0]+(0.334835177686)*x[1]**o+(0.999030888839)*x[1]+(-0.891653525154)*x[2]**o+(-0.14056280616)*x[2]
        arg[0,3]=(-0.458305223634)*x[0]**o+(0.679075077126)*x[0]+(0.1176620176)*x[1]**o+(0.685223087684)*x[1]+(-0.952262623464)*x[2]**o+(-0.905568306628)*x[2]
        arg[0,4]=(0.623179050934)*x[0]**o+(-0.0150847487218)*x[0]+(0.0823495403769)*x[1]**o+(0.48303169009)*x[1]+(0.610276253905)*x[2]**o+(0.353104308072)*x[2]
        arg[1,0]=(-0.709185342895)*x[0]**o+(0.789980598884)*x[0]+(0.655538102878)*x[1]**o+(0.713913568814)*x[1]+(0.222131732773)*x[2]**o+(0.712644729381)*x[2]
        arg[1,1]=(-0.683646421822)*x[0]**o+(-0.943691516528)*x[0]+(-0.134261002557)*x[1]**o+(0.28225941018)*x[1]+(0.906235993779)*x[2]**o+(-0.819214603609)*x[2]
        arg[1,2]=(0.713781993938)*x[0]**o+(-0.189551367619)*x[0]+(-0.021310841804)*x[1]**o+(-0.873466428012)*x[1]+(0.248337664546)*x[2]**o+(0.911824874935)*x[2]
        arg[1,3]=(-0.393142772637)*x[0]**o+(0.773501237865)*x[0]+(0.759039429922)*x[1]**o+(0.67036876499)*x[1]+(-0.921411979023)*x[2]**o+(-0.649371534276)*x[2]
        arg[1,4]=(0.526708082723)*x[0]**o+(-0.522449586718)*x[0]+(0.951176548558)*x[1]**o+(0.604602039898)*x[1]+(0.345888192325)*x[2]**o+(-0.973729954334)*x[2]
        arg[2,0]=(-0.139931215293)*x[0]**o+(-0.996072915638)*x[0]+(-0.0451731431264)*x[1]**o+(0.433694224874)*x[1]+(-0.955189760302)*x[2]**o+(-0.68361667242)*x[2]
        arg[2,1]=(0.139626552661)*x[0]**o+(-0.91217468751)*x[0]+(-0.854252204911)*x[1]**o+(0.417539913464)*x[1]+(0.557096050124)*x[2]**o+(-0.195272359868)*x[2]
        arg[2,2]=(0.960352725049)*x[0]**o+(0.98274500341)*x[0]+(0.967782565025)*x[1]**o+(0.566011353381)*x[1]+(0.583885094583)*x[2]**o+(-0.943173847936)*x[2]
        arg[2,3]=(0.959388333494)*x[0]**o+(0.531908797155)*x[0]+(-0.62531126203)*x[1]**o+(-0.288763293646)*x[1]+(-0.823398780526)*x[2]**o+(-0.164348871248)*x[2]
        arg[2,4]=(0.73397702179)*x[0]**o+(0.407844592399)*x[0]+(-0.835769355727)*x[1]**o+(0.742360626394)*x[1]+(0.31130186529)*x[2]**o+(0.114221168872)*x[2]
        arg[3,0]=(-0.42278940444)*x[0]**o+(-0.513158064442)*x[0]+(-0.133233894233)*x[1]**o+(0.701381913887)*x[1]+(0.283776333553)*x[2]**o+(-0.805096661642)*x[2]
        arg[3,1]=(-0.417338508269)*x[0]**o+(0.535961900783)*x[0]+(0.0836386907734)*x[1]**o+(0.449680928704)*x[1]+(0.998998322868)*x[2]**o+(0.983057825698)*x[2]
        arg[3,2]=(0.24102382276)*x[0]**o+(0.0291470157398)*x[0]+(-0.539542451698)*x[1]**o+(0.281017043631)*x[1]+(-0.464652133499)*x[2]**o+(0.652828720185)*x[2]
        arg[3,3]=(-0.482577920013)*x[0]**o+(0.931756756394)*x[0]+(0.96648206338)*x[1]**o+(0.73941459549)*x[1]+(0.136743015429)*x[2]**o+(0.41477914722)*x[2]
        arg[3,4]=(0.879881885972)*x[0]**o+(0.208970079585)*x[0]+(-0.200079958414)*x[1]**o+(-0.0291238772241)*x[1]+(-0.333569505857)*x[2]**o+(0.846835729875)*x[2]
        ref[0,0]=(0.370073067368)*(1+2.*(dim-1.)/(o+1.))+(0.541580310169)*dim
        ref[0,1]=(1.20261165909)*(1+2.*(dim-1.)/(o+1.))+(0.300864008561)*dim
        ref[0,2]=(0.333788383971)*(1+2.*(dim-1.)/(o+1.))+(0.80291891475)*dim
        ref[0,3]=(-1.2929058295)*(1+2.*(dim-1.)/(o+1.))+(0.458729858183)*dim
        ref[0,4]=(1.31580484522)*(1+2.*(dim-1.)/(o+1.))+(0.82105124944)*dim
        ref[1,0]=(0.168484492756)*(1+2.*(dim-1.)/(o+1.))+(2.21653889708)*dim
        ref[1,1]=(0.0883285694004)*(1+2.*(dim-1.)/(o+1.))+(-1.48064670996)*dim
        ref[1,2]=(0.940808816681)*(1+2.*(dim-1.)/(o+1.))+(-0.151192920696)*dim
        ref[1,3]=(-0.555515321739)*(1+2.*(dim-1.)/(o+1.))+(0.79449846858)*dim
        ref[1,4]=(1.82377282361)*(1+2.*(dim-1.)/(o+1.))+(-0.891577501154)*dim
        ref[2,0]=(-1.14029411872)*(1+2.*(dim-1.)/(o+1.))+(-1.24599536318)*dim
        ref[2,1]=(-0.157529602127)*(1+2.*(dim-1.)/(o+1.))+(-0.689907133913)*dim
        ref[2,2]=(2.51202038466)*(1+2.*(dim-1.)/(o+1.))+(0.605582508855)*dim
        ref[2,3]=(-0.489321709063)*(1+2.*(dim-1.)/(o+1.))+(0.0787966322607)*dim
        ref[2,4]=(0.209509531353)*(1+2.*(dim-1.)/(o+1.))+(1.26442638766)*dim
        ref[3,0]=(-0.27224696512)*(1+2.*(dim-1.)/(o+1.))+(-0.616872812198)*dim
        ref[3,1]=(0.665298505372)*(1+2.*(dim-1.)/(o+1.))+(1.96870065519)*dim
        ref[3,2]=(-0.763170762437)*(1+2.*(dim-1.)/(o+1.))+(0.962992779555)*dim
        ref[3,3]=(0.620647158797)*(1+2.*(dim-1.)/(o+1.))+(2.0859504991)*dim
        ref[3,4]=(0.346232421701)*(1+2.*(dim-1.)/(o+1.))+(1.02668193224)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.201115186772)*x[0]**o+(0.324758103748)*x[0]+(0.285065583242)*x[1]**o+(0.0264985832782)*x[1]
        arg[0,0,1]=(-0.443023003973)*x[0]**o+(-0.194732927607)*x[0]+(0.64414908519)*x[1]**o+(-0.248261857779)*x[1]
        arg[0,1,0]=(0.587534172195)*x[0]**o+(0.021088951369)*x[0]+(-0.986195557459)*x[1]**o+(0.595659005312)*x[1]
        arg[0,1,1]=(-0.916961145134)*x[0]**o+(-0.766484469605)*x[0]+(-0.281345964683)*x[1]**o+(-0.399113856826)*x[1]
        arg[1,0,0]=(0.175943424949)*x[0]**o+(0.344853386001)*x[0]+(0.175474681848)*x[1]**o+(-0.0775102874697)*x[1]
        arg[1,0,1]=(-0.874074467007)*x[0]**o+(0.903506914387)*x[0]+(0.177699372538)*x[1]**o+(0.969818154968)*x[1]
        arg[1,1,0]=(0.765751472387)*x[0]**o+(-0.0288724203687)*x[0]+(0.655676828008)*x[1]**o+(0.22281307247)*x[1]
        arg[1,1,1]=(0.871009077152)*x[0]**o+(-0.20052353101)*x[0]+(-0.889496767621)*x[1]**o+(0.8458762887)*x[1]
        arg[2,0,0]=(-0.15377513542)*x[0]**o+(0.963714892768)*x[0]+(0.942514815983)*x[1]**o+(0.846111018838)*x[1]
        arg[2,0,1]=(0.000727086938663)*x[0]**o+(0.243136987492)*x[0]+(-0.308674899886)*x[1]**o+(-0.963435664436)*x[1]
        arg[2,1,0]=(0.780912245051)*x[0]**o+(0.227139041778)*x[0]+(-0.336203258384)*x[1]**o+(0.092568284679)*x[1]
        arg[2,1,1]=(0.0464862681168)*x[0]**o+(0.284493375176)*x[0]+(-0.811073509077)*x[1]**o+(0.54280235419)*x[1]
        arg[3,0,0]=(0.605642304612)*x[0]**o+(-0.103442921567)*x[0]+(0.106253324829)*x[1]**o+(0.562480007904)*x[1]
        arg[3,0,1]=(0.497668538008)*x[0]**o+(-0.859679982844)*x[0]+(-0.255499355194)*x[1]**o+(0.0725919239243)*x[1]
        arg[3,1,0]=(0.641826893201)*x[0]**o+(0.889514581887)*x[0]+(-0.211781857836)*x[1]**o+(0.45900873193)*x[1]
        arg[3,1,1]=(-0.29349600565)*x[0]**o+(-0.627542961817)*x[0]+(0.313009642686)*x[1]**o+(0.338028134809)*x[1]
        arg[4,0,0]=(-0.641915241397)*x[0]**o+(-0.682547311604)*x[0]+(-0.106472077451)*x[1]**o+(0.103402239052)*x[1]
        arg[4,0,1]=(-0.822562520978)*x[0]**o+(-0.430539530156)*x[0]+(-0.0225528898923)*x[1]**o+(-0.0908986665674)*x[1]
        arg[4,1,0]=(-0.456160645978)*x[0]**o+(-0.264592336242)*x[0]+(-0.672702978868)*x[1]**o+(-0.893197955639)*x[1]
        arg[4,1,1]=(-0.714689169676)*x[0]**o+(0.513612105215)*x[0]+(-0.3093340377)*x[1]**o+(-0.34467035156)*x[1]
        arg[5,0,0]=(-0.993761664154)*x[0]**o+(0.375953817044)*x[0]+(0.0463240503072)*x[1]**o+(-0.595646829466)*x[1]
        arg[5,0,1]=(0.453145186545)*x[0]**o+(0.331353840672)*x[0]+(-0.908533078632)*x[1]**o+(0.755175839674)*x[1]
        arg[5,1,0]=(-0.736874540101)*x[0]**o+(-0.867437170846)*x[0]+(-0.472395301622)*x[1]**o+(0.78289019661)*x[1]
        arg[5,1,1]=(-0.900966156016)*x[0]**o+(-0.160363876891)*x[0]+(-0.0969222200475)*x[1]**o+(0.0527983470624)*x[1]
        ref[0,0,0]=(0.486180770014)*(1+2.*(dim-1.)/(o+1.))+(0.351256687026)*dim
        ref[0,0,1]=(0.201126081217)*(1+2.*(dim-1.)/(o+1.))+(-0.442994785386)*dim
        ref[0,1,0]=(-0.398661385264)*(1+2.*(dim-1.)/(o+1.))+(0.616747956681)*dim
        ref[0,1,1]=(-1.19830710982)*(1+2.*(dim-1.)/(o+1.))+(-1.16559832643)*dim
        ref[1,0,0]=(0.351418106797)*(1+2.*(dim-1.)/(o+1.))+(0.267343098531)*dim
        ref[1,0,1]=(-0.696375094469)*(1+2.*(dim-1.)/(o+1.))+(1.87332506936)*dim
        ref[1,1,0]=(1.42142830039)*(1+2.*(dim-1.)/(o+1.))+(0.193940652101)*dim
        ref[1,1,1]=(-0.0184876904693)*(1+2.*(dim-1.)/(o+1.))+(0.64535275769)*dim
        ref[2,0,0]=(0.788739680562)*(1+2.*(dim-1.)/(o+1.))+(1.80982591161)*dim
        ref[2,0,1]=(-0.307947812947)*(1+2.*(dim-1.)/(o+1.))+(-0.720298676944)*dim
        ref[2,1,0]=(0.444708986667)*(1+2.*(dim-1.)/(o+1.))+(0.319707326457)*dim
        ref[2,1,1]=(-0.76458724096)*(1+2.*(dim-1.)/(o+1.))+(0.827295729366)*dim
        ref[3,0,0]=(0.711895629441)*(1+2.*(dim-1.)/(o+1.))+(0.459037086337)*dim
        ref[3,0,1]=(0.242169182814)*(1+2.*(dim-1.)/(o+1.))+(-0.78708805892)*dim
        ref[3,1,0]=(0.430045035365)*(1+2.*(dim-1.)/(o+1.))+(1.34852331382)*dim
        ref[3,1,1]=(0.019513637036)*(1+2.*(dim-1.)/(o+1.))+(-0.289514827008)*dim
        ref[4,0,0]=(-0.748387318848)*(1+2.*(dim-1.)/(o+1.))+(-0.579145072551)*dim
        ref[4,0,1]=(-0.845115410871)*(1+2.*(dim-1.)/(o+1.))+(-0.521438196723)*dim
        ref[4,1,0]=(-1.12886362485)*(1+2.*(dim-1.)/(o+1.))+(-1.15779029188)*dim
        ref[4,1,1]=(-1.02402320738)*(1+2.*(dim-1.)/(o+1.))+(0.168941753655)*dim
        ref[5,0,0]=(-0.947437613847)*(1+2.*(dim-1.)/(o+1.))+(-0.219693012422)*dim
        ref[5,0,1]=(-0.455387892087)*(1+2.*(dim-1.)/(o+1.))+(1.08652968035)*dim
        ref[5,1,0]=(-1.20926984172)*(1+2.*(dim-1.)/(o+1.))+(-0.0845469742352)*dim
        ref[5,1,1]=(-0.997888376064)*(1+2.*(dim-1.)/(o+1.))+(-0.107565529828)*dim
      else:
        arg[0,0,0]=(-0.760764230601)*x[0]**o+(-0.955215307904)*x[0]+(0.376251560231)*x[1]**o+(0.147849416477)*x[1]+(0.627037741745)*x[2]**o+(-0.93996244415)*x[2]
        arg[0,0,1]=(0.904449671196)*x[0]**o+(-0.297804849063)*x[0]+(0.773059038176)*x[1]**o+(-0.836104534497)*x[1]+(-0.150040994258)*x[2]**o+(0.837988726131)*x[2]
        arg[0,1,0]=(0.354255190058)*x[0]**o+(0.463977999833)*x[0]+(0.218855388406)*x[1]**o+(-0.163086423536)*x[1]+(0.653790071045)*x[2]**o+(-0.457721486689)*x[2]
        arg[0,1,1]=(0.357317871424)*x[0]**o+(-0.752862019242)*x[0]+(0.674187130836)*x[1]**o+(-0.983503125785)*x[1]+(-0.776153051078)*x[2]**o+(0.390700377058)*x[2]
        arg[1,0,0]=(-0.465423805042)*x[0]**o+(0.629855541232)*x[0]+(-0.405300083887)*x[1]**o+(-0.211254598072)*x[1]+(0.502581881406)*x[2]**o+(-0.853779249469)*x[2]
        arg[1,0,1]=(0.716025450344)*x[0]**o+(0.876021396077)*x[0]+(-0.530324190586)*x[1]**o+(0.700207449985)*x[1]+(-0.788565669024)*x[2]**o+(0.963118484198)*x[2]
        arg[1,1,0]=(0.813943574582)*x[0]**o+(-0.938778607938)*x[0]+(-0.451619613547)*x[1]**o+(-0.916990884105)*x[1]+(-0.671730738367)*x[2]**o+(0.109664560689)*x[2]
        arg[1,1,1]=(-0.557598711518)*x[0]**o+(-0.541691706474)*x[0]+(0.500874328683)*x[1]**o+(-0.959653364589)*x[1]+(0.572057415183)*x[2]**o+(-0.842306660222)*x[2]
        arg[2,0,0]=(0.0470537168687)*x[0]**o+(-0.151521222488)*x[0]+(0.168279496456)*x[1]**o+(0.869948894317)*x[1]+(-0.554196665086)*x[2]**o+(0.444189090676)*x[2]
        arg[2,0,1]=(-0.662480394662)*x[0]**o+(0.506778773041)*x[0]+(0.664368833417)*x[1]**o+(-0.76728005759)*x[1]+(0.729668063743)*x[2]**o+(-0.476499535761)*x[2]
        arg[2,1,0]=(0.40405640406)*x[0]**o+(0.537691275434)*x[0]+(0.675845932799)*x[1]**o+(-0.663037107454)*x[1]+(-0.947246204559)*x[2]**o+(-0.210977364198)*x[2]
        arg[2,1,1]=(-0.460999541334)*x[0]**o+(-0.403551870206)*x[0]+(-0.59751233801)*x[1]**o+(0.389018763902)*x[1]+(-0.916883519483)*x[2]**o+(0.712714107738)*x[2]
        arg[3,0,0]=(-0.547139864054)*x[0]**o+(0.91774354334)*x[0]+(-0.953248927323)*x[1]**o+(0.693673610785)*x[1]+(-0.474082903895)*x[2]**o+(0.187174904886)*x[2]
        arg[3,0,1]=(0.538597533339)*x[0]**o+(0.903368885629)*x[0]+(-0.548058780044)*x[1]**o+(-0.380406303455)*x[1]+(-0.389244091021)*x[2]**o+(0.137402766164)*x[2]
        arg[3,1,0]=(-0.11691167078)*x[0]**o+(-0.812294829383)*x[0]+(0.592192015269)*x[1]**o+(-0.3278056345)*x[1]+(-0.60462640896)*x[2]**o+(-0.701541632465)*x[2]
        arg[3,1,1]=(-0.196382907034)*x[0]**o+(-0.496234735923)*x[0]+(-0.646502033751)*x[1]**o+(-0.18981379172)*x[1]+(0.627608199527)*x[2]**o+(-0.468659552408)*x[2]
        arg[4,0,0]=(0.56789650708)*x[0]**o+(0.854984079971)*x[0]+(-0.641473205573)*x[1]**o+(-0.791240666887)*x[1]+(0.453562162212)*x[2]**o+(0.849111318036)*x[2]
        arg[4,0,1]=(0.90836228457)*x[0]**o+(-0.666684673376)*x[0]+(0.245928632693)*x[1]**o+(0.313580017685)*x[1]+(0.353126203728)*x[2]**o+(0.467530918033)*x[2]
        arg[4,1,0]=(0.670926585272)*x[0]**o+(0.768555336283)*x[0]+(-0.170656824353)*x[1]**o+(-0.346425800834)*x[1]+(-0.969641886683)*x[2]**o+(0.775335418066)*x[2]
        arg[4,1,1]=(-0.152500967035)*x[0]**o+(0.575400101502)*x[0]+(0.901331752685)*x[1]**o+(-0.672576642993)*x[1]+(0.803176375377)*x[2]**o+(0.202785337179)*x[2]
        arg[5,0,0]=(-0.798851736802)*x[0]**o+(-0.221276048656)*x[0]+(0.773295190339)*x[1]**o+(0.0903112854292)*x[1]+(0.884961364642)*x[2]**o+(0.326429585537)*x[2]
        arg[5,0,1]=(0.482124059007)*x[0]**o+(-0.39227677557)*x[0]+(-0.23074922677)*x[1]**o+(0.986783507045)*x[1]+(0.591759402442)*x[2]**o+(-0.680278617308)*x[2]
        arg[5,1,0]=(-0.857456972644)*x[0]**o+(-0.200661025977)*x[0]+(-0.850031395341)*x[1]**o+(0.830863037447)*x[1]+(-0.323765625026)*x[2]**o+(-0.969463605839)*x[2]
        arg[5,1,1]=(-0.582538216697)*x[0]**o+(-0.495303456972)*x[0]+(0.0134139524699)*x[1]**o+(-0.750144603651)*x[1]+(-0.632380446311)*x[2]**o+(0.38782755017)*x[2]
        ref[0,0,0]=(0.242525071375)*(1+2.*(dim-1.)/(o+1.))+(-1.74732833558)*dim
        ref[0,0,1]=(1.52746771511)*(1+2.*(dim-1.)/(o+1.))+(-0.295920657428)*dim
        ref[0,1,0]=(1.22690064951)*(1+2.*(dim-1.)/(o+1.))+(-0.156829910392)*dim
        ref[0,1,1]=(0.255351951182)*(1+2.*(dim-1.)/(o+1.))+(-1.34566476797)*dim
        ref[1,0,0]=(-0.368142007523)*(1+2.*(dim-1.)/(o+1.))+(-0.43517830631)*dim
        ref[1,0,1]=(-0.602864409266)*(1+2.*(dim-1.)/(o+1.))+(2.53934733026)*dim
        ref[1,1,0]=(-0.309406777332)*(1+2.*(dim-1.)/(o+1.))+(-1.74610493135)*dim
        ref[1,1,1]=(0.515333032348)*(1+2.*(dim-1.)/(o+1.))+(-2.34365173129)*dim
        ref[2,0,0]=(-0.338863451761)*(1+2.*(dim-1.)/(o+1.))+(1.16261676251)*dim
        ref[2,0,1]=(0.731556502498)*(1+2.*(dim-1.)/(o+1.))+(-0.73700082031)*dim
        ref[2,1,0]=(0.132656132299)*(1+2.*(dim-1.)/(o+1.))+(-0.336323196217)*dim
        ref[2,1,1]=(-1.97539539883)*(1+2.*(dim-1.)/(o+1.))+(0.698181001434)*dim
        ref[3,0,0]=(-1.97447169527)*(1+2.*(dim-1.)/(o+1.))+(1.79859205901)*dim
        ref[3,0,1]=(-0.398705337727)*(1+2.*(dim-1.)/(o+1.))+(0.660365348338)*dim
        ref[3,1,0]=(-0.129346064471)*(1+2.*(dim-1.)/(o+1.))+(-1.84164209635)*dim
        ref[3,1,1]=(-0.215276741258)*(1+2.*(dim-1.)/(o+1.))+(-1.15470808005)*dim
        ref[4,0,0]=(0.379985463719)*(1+2.*(dim-1.)/(o+1.))+(0.91285473112)*dim
        ref[4,0,1]=(1.50741712099)*(1+2.*(dim-1.)/(o+1.))+(0.114426262342)*dim
        ref[4,1,0]=(-0.469372125764)*(1+2.*(dim-1.)/(o+1.))+(1.19746495352)*dim
        ref[4,1,1]=(1.55200716103)*(1+2.*(dim-1.)/(o+1.))+(0.105608795688)*dim
        ref[5,0,0]=(0.859404818179)*(1+2.*(dim-1.)/(o+1.))+(0.19546482231)*dim
        ref[5,0,1]=(0.843134234679)*(1+2.*(dim-1.)/(o+1.))+(-0.085771885833)*dim
        ref[5,1,0]=(-2.03125399301)*(1+2.*(dim-1.)/(o+1.))+(-0.339261594369)*dim
        ref[5,1,1]=(-1.20150471054)*(1+2.*(dim-1.)/(o+1.))+(-0.857620510454)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.161068279134)*x[0]**o+(-0.20016693624)*x[0]+(0.360967926263)*x[1]**o+(0.519774730324)*x[1]
        arg[0,0,0,1]=(-0.470257945181)*x[0]**o+(0.0652334206708)*x[0]+(0.652139991357)*x[1]**o+(-0.592482756578)*x[1]
        arg[0,0,1,0]=(0.143126570256)*x[0]**o+(0.683474249212)*x[0]+(0.839611756722)*x[1]**o+(-0.436612603526)*x[1]
        arg[0,0,1,1]=(-0.82328367976)*x[0]**o+(-0.848145278113)*x[0]+(-0.649046694126)*x[1]**o+(0.7686615034)*x[1]
        arg[0,0,2,0]=(-0.89676527056)*x[0]**o+(-0.466316806084)*x[0]+(-0.454908744166)*x[1]**o+(0.425524454897)*x[1]
        arg[0,0,2,1]=(-0.720978656424)*x[0]**o+(-0.655449460498)*x[0]+(0.466383340682)*x[1]**o+(-0.609615749634)*x[1]
        arg[0,1,0,0]=(-0.358670380794)*x[0]**o+(-0.764939003508)*x[0]+(-0.550898156906)*x[1]**o+(0.803150103728)*x[1]
        arg[0,1,0,1]=(-0.652241076457)*x[0]**o+(-0.614033274928)*x[0]+(-0.308770011731)*x[1]**o+(0.925847136965)*x[1]
        arg[0,1,1,0]=(0.710539026622)*x[0]**o+(-0.345587804078)*x[0]+(-0.170026226995)*x[1]**o+(0.340557574804)*x[1]
        arg[0,1,1,1]=(0.348069554895)*x[0]**o+(-0.633629515001)*x[0]+(-0.199233786202)*x[1]**o+(0.500291488845)*x[1]
        arg[0,1,2,0]=(0.262179751197)*x[0]**o+(-0.55460029497)*x[0]+(-0.354335283745)*x[1]**o+(0.301130103531)*x[1]
        arg[0,1,2,1]=(0.212709964525)*x[0]**o+(0.205907518129)*x[0]+(0.167040570615)*x[1]**o+(-0.0684155270612)*x[1]
        arg[0,2,0,0]=(-0.272141616042)*x[0]**o+(0.187045318075)*x[0]+(-0.364974531949)*x[1]**o+(-0.969702579161)*x[1]
        arg[0,2,0,1]=(-0.997155483171)*x[0]**o+(0.864439328971)*x[0]+(0.887907754657)*x[1]**o+(-0.323021835147)*x[1]
        arg[0,2,1,0]=(-0.544228085563)*x[0]**o+(-0.56752289714)*x[0]+(-0.389234395916)*x[1]**o+(-0.969287386144)*x[1]
        arg[0,2,1,1]=(-0.786136450866)*x[0]**o+(0.589465339989)*x[0]+(-0.411595778776)*x[1]**o+(0.726975706536)*x[1]
        arg[0,2,2,0]=(-0.788299350517)*x[0]**o+(0.697037609691)*x[0]+(-0.391342918209)*x[1]**o+(0.432368662365)*x[1]
        arg[0,2,2,1]=(-0.390536554506)*x[0]**o+(0.547528359725)*x[0]+(0.58957966188)*x[1]**o+(0.146918361066)*x[1]
        arg[0,3,0,0]=(0.176059579799)*x[0]**o+(-0.774041865433)*x[0]+(0.52369869825)*x[1]**o+(-0.812474996262)*x[1]
        arg[0,3,0,1]=(-0.938727595513)*x[0]**o+(0.734277246158)*x[0]+(-0.500452658276)*x[1]**o+(0.591562924701)*x[1]
        arg[0,3,1,0]=(-0.109970490351)*x[0]**o+(-0.463382333915)*x[0]+(0.430588389365)*x[1]**o+(-0.174978181104)*x[1]
        arg[0,3,1,1]=(-0.564486046162)*x[0]**o+(0.325599920584)*x[0]+(-0.177334934858)*x[1]**o+(-0.789210879368)*x[1]
        arg[0,3,2,0]=(-0.409769653157)*x[0]**o+(-0.674668709075)*x[0]+(0.977465383205)*x[1]**o+(-0.178249094926)*x[1]
        arg[0,3,2,1]=(0.112394584736)*x[0]**o+(-0.229767891006)*x[0]+(0.952422132907)*x[1]**o+(-0.133952947363)*x[1]
        arg[0,4,0,0]=(0.927747625444)*x[0]**o+(-0.147390883234)*x[0]+(-0.0523168758823)*x[1]**o+(0.698643952882)*x[1]
        arg[0,4,0,1]=(0.986627759385)*x[0]**o+(-0.0486375625834)*x[0]+(0.27598075052)*x[1]**o+(0.970272004318)*x[1]
        arg[0,4,1,0]=(-0.461643441984)*x[0]**o+(-0.815261492927)*x[0]+(-0.0806771439228)*x[1]**o+(0.877302861263)*x[1]
        arg[0,4,1,1]=(-0.108403920958)*x[0]**o+(0.651566419087)*x[0]+(-0.629972296369)*x[1]**o+(-0.235413582155)*x[1]
        arg[0,4,2,0]=(-0.177527213798)*x[0]**o+(0.987202381273)*x[0]+(-0.15522298966)*x[1]**o+(0.234268561378)*x[1]
        arg[0,4,2,1]=(0.521954293269)*x[0]**o+(-0.776898814469)*x[0]+(0.783085515988)*x[1]**o+(-0.469921197982)*x[1]
        arg[1,0,0,0]=(0.858157489751)*x[0]**o+(0.444181035326)*x[0]+(0.3426806552)*x[1]**o+(-0.219249318925)*x[1]
        arg[1,0,0,1]=(0.405857394564)*x[0]**o+(0.178361294933)*x[0]+(0.0879917927852)*x[1]**o+(-0.0689581267347)*x[1]
        arg[1,0,1,0]=(0.63175565694)*x[0]**o+(0.0810462363705)*x[0]+(0.0350399072162)*x[1]**o+(-0.269906961338)*x[1]
        arg[1,0,1,1]=(-0.273973794035)*x[0]**o+(0.824519533391)*x[0]+(0.596785621044)*x[1]**o+(-0.243043370261)*x[1]
        arg[1,0,2,0]=(-0.959246118913)*x[0]**o+(0.3570024829)*x[0]+(-0.260098162669)*x[1]**o+(0.341044029457)*x[1]
        arg[1,0,2,1]=(0.966671029776)*x[0]**o+(0.598898116395)*x[0]+(0.392360541053)*x[1]**o+(-0.322723646751)*x[1]
        arg[1,1,0,0]=(-0.658365963559)*x[0]**o+(-0.674107595463)*x[0]+(-0.0103911551403)*x[1]**o+(0.858879803218)*x[1]
        arg[1,1,0,1]=(0.340459206102)*x[0]**o+(-0.681434505015)*x[0]+(0.917412924129)*x[1]**o+(0.703699154381)*x[1]
        arg[1,1,1,0]=(0.0473840209756)*x[0]**o+(-0.787070458106)*x[0]+(0.11979244907)*x[1]**o+(0.541238063489)*x[1]
        arg[1,1,1,1]=(0.184943768213)*x[0]**o+(-0.942961758488)*x[0]+(-0.844785429901)*x[1]**o+(-0.0109779269179)*x[1]
        arg[1,1,2,0]=(0.921469748172)*x[0]**o+(-0.122684636542)*x[0]+(0.149552532716)*x[1]**o+(-0.0102623511735)*x[1]
        arg[1,1,2,1]=(-0.720466710917)*x[0]**o+(-0.755841224189)*x[0]+(0.372635950892)*x[1]**o+(0.271604227274)*x[1]
        arg[1,2,0,0]=(-0.548557725319)*x[0]**o+(0.745340396995)*x[0]+(0.461556064961)*x[1]**o+(0.696460891668)*x[1]
        arg[1,2,0,1]=(-0.738939238644)*x[0]**o+(0.344829756443)*x[0]+(0.272341172248)*x[1]**o+(0.768063359753)*x[1]
        arg[1,2,1,0]=(-0.213161389496)*x[0]**o+(0.458730188555)*x[0]+(-0.783049463147)*x[1]**o+(0.785273989074)*x[1]
        arg[1,2,1,1]=(0.189092239774)*x[0]**o+(-0.0649917239139)*x[0]+(0.0171281685468)*x[1]**o+(0.397065812267)*x[1]
        arg[1,2,2,0]=(-0.809663598067)*x[0]**o+(0.587214021528)*x[0]+(0.912814601228)*x[1]**o+(0.860792008172)*x[1]
        arg[1,2,2,1]=(0.487927123118)*x[0]**o+(-0.760831928598)*x[0]+(-0.770486346056)*x[1]**o+(-0.715077841171)*x[1]
        arg[1,3,0,0]=(0.0202747853647)*x[0]**o+(0.0666642857036)*x[0]+(-0.717756666598)*x[1]**o+(-0.481009806467)*x[1]
        arg[1,3,0,1]=(-0.251164550146)*x[0]**o+(0.0519935051437)*x[0]+(-0.703400138413)*x[1]**o+(-0.990329903234)*x[1]
        arg[1,3,1,0]=(0.970793375169)*x[0]**o+(-0.365675269998)*x[0]+(0.77706784549)*x[1]**o+(-0.237785293481)*x[1]
        arg[1,3,1,1]=(-0.162786953006)*x[0]**o+(0.985402077919)*x[0]+(0.739483663725)*x[1]**o+(0.559554321882)*x[1]
        arg[1,3,2,0]=(0.184701463663)*x[0]**o+(0.946173117295)*x[0]+(-0.451420616712)*x[1]**o+(0.244506578089)*x[1]
        arg[1,3,2,1]=(0.894307549357)*x[0]**o+(0.474983970689)*x[0]+(0.626273940678)*x[1]**o+(0.107985157225)*x[1]
        arg[1,4,0,0]=(-0.180994231135)*x[0]**o+(-0.899808624585)*x[0]+(0.395935138342)*x[1]**o+(-0.387788497342)*x[1]
        arg[1,4,0,1]=(0.0713358452056)*x[0]**o+(0.505276043139)*x[0]+(-0.635565006093)*x[1]**o+(0.956951408169)*x[1]
        arg[1,4,1,0]=(0.355608577722)*x[0]**o+(0.58482326687)*x[0]+(0.654137690443)*x[1]**o+(-0.849016362888)*x[1]
        arg[1,4,1,1]=(-0.275966957458)*x[0]**o+(0.856095068377)*x[0]+(-0.760077050664)*x[1]**o+(-0.904159432699)*x[1]
        arg[1,4,2,0]=(-0.0637454805096)*x[0]**o+(-0.833057387633)*x[0]+(0.279008913569)*x[1]**o+(-0.093356790193)*x[1]
        arg[1,4,2,1]=(0.165869666472)*x[0]**o+(-0.4554345089)*x[0]+(0.697149094381)*x[1]**o+(0.244460087482)*x[1]
        arg[2,0,0,0]=(0.388188511348)*x[0]**o+(-0.00552398040882)*x[0]+(0.0679336322536)*x[1]**o+(-0.281132460216)*x[1]
        arg[2,0,0,1]=(0.968019690109)*x[0]**o+(0.525335591339)*x[0]+(0.660069087546)*x[1]**o+(-0.795388487336)*x[1]
        arg[2,0,1,0]=(-0.444010450901)*x[0]**o+(-0.498771184141)*x[0]+(-0.394999176962)*x[1]**o+(-0.579230388911)*x[1]
        arg[2,0,1,1]=(-0.96211072138)*x[0]**o+(-0.0977519108256)*x[0]+(0.206179333903)*x[1]**o+(-0.206155312014)*x[1]
        arg[2,0,2,0]=(0.131307090845)*x[0]**o+(0.179396626286)*x[0]+(-0.436118623166)*x[1]**o+(0.829727875456)*x[1]
        arg[2,0,2,1]=(-0.206336435516)*x[0]**o+(0.175744210813)*x[0]+(0.818204199205)*x[1]**o+(0.433292546177)*x[1]
        arg[2,1,0,0]=(0.920189410059)*x[0]**o+(-0.629286762249)*x[0]+(-0.14444305096)*x[1]**o+(0.564022686795)*x[1]
        arg[2,1,0,1]=(0.43550707444)*x[0]**o+(-0.0203691253164)*x[0]+(0.632472807002)*x[1]**o+(0.722390213081)*x[1]
        arg[2,1,1,0]=(0.827084071535)*x[0]**o+(0.864139638834)*x[0]+(0.795893591186)*x[1]**o+(0.0492667773115)*x[1]
        arg[2,1,1,1]=(0.863813787478)*x[0]**o+(0.268800194137)*x[0]+(-0.567637801205)*x[1]**o+(0.931221493563)*x[1]
        arg[2,1,2,0]=(-0.613463525024)*x[0]**o+(-0.0688030533166)*x[0]+(-0.786395692457)*x[1]**o+(0.862189706883)*x[1]
        arg[2,1,2,1]=(0.315306814851)*x[0]**o+(0.76879509223)*x[0]+(0.451451342779)*x[1]**o+(-0.478873421268)*x[1]
        arg[2,2,0,0]=(-0.785116047841)*x[0]**o+(0.347614087057)*x[0]+(-0.0106716042015)*x[1]**o+(-0.350898768187)*x[1]
        arg[2,2,0,1]=(-0.987778931187)*x[0]**o+(-0.380379654757)*x[0]+(-0.383604999655)*x[1]**o+(0.022203803886)*x[1]
        arg[2,2,1,0]=(0.508327115069)*x[0]**o+(0.443678958632)*x[0]+(-0.320069211602)*x[1]**o+(-0.778140184171)*x[1]
        arg[2,2,1,1]=(-0.532025577107)*x[0]**o+(-0.0475799386434)*x[0]+(0.650417175266)*x[1]**o+(0.33739661369)*x[1]
        arg[2,2,2,0]=(-0.406412097333)*x[0]**o+(0.0510500610605)*x[0]+(-0.25755890225)*x[1]**o+(-0.350209779392)*x[1]
        arg[2,2,2,1]=(0.401361048956)*x[0]**o+(-0.594793585281)*x[0]+(0.155257431052)*x[1]**o+(-0.674885216524)*x[1]
        arg[2,3,0,0]=(0.352606222236)*x[0]**o+(0.64529574473)*x[0]+(-0.838258091233)*x[1]**o+(0.527911574536)*x[1]
        arg[2,3,0,1]=(-0.62652610251)*x[0]**o+(0.879664434218)*x[0]+(0.730845316192)*x[1]**o+(0.0318659384791)*x[1]
        arg[2,3,1,0]=(-0.311152313179)*x[0]**o+(-0.333154399786)*x[0]+(0.743640958069)*x[1]**o+(0.313394944493)*x[1]
        arg[2,3,1,1]=(-0.728968478378)*x[0]**o+(0.500229380574)*x[0]+(0.756628133572)*x[1]**o+(0.270533146994)*x[1]
        arg[2,3,2,0]=(-0.219775843272)*x[0]**o+(0.882943220401)*x[0]+(0.79341547188)*x[1]**o+(0.503486054551)*x[1]
        arg[2,3,2,1]=(-0.850285149353)*x[0]**o+(0.78820884302)*x[0]+(0.348024963962)*x[1]**o+(0.83743616537)*x[1]
        arg[2,4,0,0]=(0.868649421127)*x[0]**o+(-0.622361223999)*x[0]+(0.836321867516)*x[1]**o+(0.423582994131)*x[1]
        arg[2,4,0,1]=(0.140963355402)*x[0]**o+(0.190496469385)*x[0]+(0.765311231744)*x[1]**o+(0.368696267095)*x[1]
        arg[2,4,1,0]=(-0.680854080076)*x[0]**o+(-0.729641193897)*x[0]+(-0.843347646585)*x[1]**o+(-0.0993036180842)*x[1]
        arg[2,4,1,1]=(-0.130039648356)*x[0]**o+(0.534197300899)*x[0]+(-0.142588861186)*x[1]**o+(-0.308376644249)*x[1]
        arg[2,4,2,0]=(0.331212825202)*x[0]**o+(0.466247840288)*x[0]+(-0.201714210846)*x[1]**o+(0.80521345364)*x[1]
        arg[2,4,2,1]=(0.454299813247)*x[0]**o+(0.646694941459)*x[0]+(-0.929854585458)*x[1]**o+(-0.365252296746)*x[1]
        arg[3,0,0,0]=(-0.623915395721)*x[0]**o+(-0.789827882448)*x[0]+(-0.862568457189)*x[1]**o+(-0.141799958679)*x[1]
        arg[3,0,0,1]=(0.197040281353)*x[0]**o+(0.155785418081)*x[0]+(-0.587898788817)*x[1]**o+(0.700612189317)*x[1]
        arg[3,0,1,0]=(-0.899654954269)*x[0]**o+(-0.801426785588)*x[0]+(-0.528678401571)*x[1]**o+(-0.224185097206)*x[1]
        arg[3,0,1,1]=(0.291298188198)*x[0]**o+(0.786698690825)*x[0]+(0.925634940697)*x[1]**o+(0.617661715197)*x[1]
        arg[3,0,2,0]=(0.788272473274)*x[0]**o+(-0.0400044705274)*x[0]+(-0.669092974469)*x[1]**o+(-0.063697728491)*x[1]
        arg[3,0,2,1]=(-0.0612398565521)*x[0]**o+(-0.884311783009)*x[0]+(0.444689078628)*x[1]**o+(0.302741981611)*x[1]
        arg[3,1,0,0]=(0.483756249722)*x[0]**o+(-0.214185139167)*x[0]+(0.526387014144)*x[1]**o+(-0.172619975799)*x[1]
        arg[3,1,0,1]=(0.0685836489777)*x[0]**o+(-0.425861440836)*x[0]+(0.94669899123)*x[1]**o+(-0.649068988929)*x[1]
        arg[3,1,1,0]=(0.935225203148)*x[0]**o+(-0.649378812349)*x[0]+(-0.791553458732)*x[1]**o+(0.776307740809)*x[1]
        arg[3,1,1,1]=(-0.45319945403)*x[0]**o+(-0.473310763902)*x[0]+(-0.268480872039)*x[1]**o+(0.847382076174)*x[1]
        arg[3,1,2,0]=(-0.573022434421)*x[0]**o+(0.584143976397)*x[0]+(0.131688457808)*x[1]**o+(-0.628894863435)*x[1]
        arg[3,1,2,1]=(0.628566928363)*x[0]**o+(-0.868564569338)*x[0]+(-0.482273553287)*x[1]**o+(0.679652796519)*x[1]
        arg[3,2,0,0]=(-0.64046509323)*x[0]**o+(-0.108928537952)*x[0]+(-0.512288515479)*x[1]**o+(0.683072290041)*x[1]
        arg[3,2,0,1]=(0.472589328326)*x[0]**o+(0.219741638671)*x[0]+(0.136368535311)*x[1]**o+(-0.525113003647)*x[1]
        arg[3,2,1,0]=(-0.45731256203)*x[0]**o+(0.0854410475382)*x[0]+(0.257725845542)*x[1]**o+(0.430430185351)*x[1]
        arg[3,2,1,1]=(0.948699678725)*x[0]**o+(-0.00538658505415)*x[0]+(0.921898600708)*x[1]**o+(-0.448522286404)*x[1]
        arg[3,2,2,0]=(-0.0863353766204)*x[0]**o+(-0.652417730681)*x[0]+(0.755871413533)*x[1]**o+(0.206065623655)*x[1]
        arg[3,2,2,1]=(0.263075920527)*x[0]**o+(0.94162407402)*x[0]+(0.750015836722)*x[1]**o+(-0.815154267718)*x[1]
        arg[3,3,0,0]=(-0.606984299829)*x[0]**o+(0.120704522147)*x[0]+(-0.324211323778)*x[1]**o+(-0.123647081263)*x[1]
        arg[3,3,0,1]=(0.838995936232)*x[0]**o+(0.854147779134)*x[0]+(-0.192271552952)*x[1]**o+(0.0281108740635)*x[1]
        arg[3,3,1,0]=(-0.0343159887291)*x[0]**o+(0.242432659911)*x[0]+(-0.37299830191)*x[1]**o+(-0.456792158625)*x[1]
        arg[3,3,1,1]=(-0.305277044867)*x[0]**o+(-0.930386582491)*x[0]+(-0.979921197617)*x[1]**o+(-0.507736755766)*x[1]
        arg[3,3,2,0]=(0.548675744195)*x[0]**o+(0.755596482286)*x[0]+(0.0722267054576)*x[1]**o+(-0.393768047035)*x[1]
        arg[3,3,2,1]=(0.74781993508)*x[0]**o+(-0.655033306209)*x[0]+(0.480653847364)*x[1]**o+(-0.169400714085)*x[1]
        arg[3,4,0,0]=(0.269186769422)*x[0]**o+(-0.0786214595922)*x[0]+(-0.975508552024)*x[1]**o+(-0.798249509899)*x[1]
        arg[3,4,0,1]=(-0.599201275441)*x[0]**o+(0.473490319504)*x[0]+(0.0277044125859)*x[1]**o+(0.818107843614)*x[1]
        arg[3,4,1,0]=(-0.46272642862)*x[0]**o+(0.740282148493)*x[0]+(0.35998379499)*x[1]**o+(-0.938681489717)*x[1]
        arg[3,4,1,1]=(0.187067286301)*x[0]**o+(-0.284171344521)*x[0]+(0.612229918327)*x[1]**o+(0.3917984881)*x[1]
        arg[3,4,2,0]=(-0.535270910341)*x[0]**o+(-0.00723888213927)*x[0]+(-0.776984534618)*x[1]**o+(-0.0899536033273)*x[1]
        arg[3,4,2,1]=(-0.382290376155)*x[0]**o+(-0.0593490177599)*x[0]+(-0.19716515678)*x[1]**o+(-0.268991693613)*x[1]
        ref[0,0,0,0]=(0.19989964713)*(1+2.*(dim-1.)/(o+1.))+(0.319607794084)*dim
        ref[0,0,0,1]=(0.181882046176)*(1+2.*(dim-1.)/(o+1.))+(-0.527249335907)*dim
        ref[0,0,1,0]=(0.982738326978)*(1+2.*(dim-1.)/(o+1.))+(0.246861645686)*dim
        ref[0,0,1,1]=(-1.47233037389)*(1+2.*(dim-1.)/(o+1.))+(-0.0794837747136)*dim
        ref[0,0,2,0]=(-1.35167401473)*(1+2.*(dim-1.)/(o+1.))+(-0.0407923511861)*dim
        ref[0,0,2,1]=(-0.254595315742)*(1+2.*(dim-1.)/(o+1.))+(-1.26506521013)*dim
        ref[0,1,0,0]=(-0.909568537701)*(1+2.*(dim-1.)/(o+1.))+(0.0382111002203)*dim
        ref[0,1,0,1]=(-0.961011088187)*(1+2.*(dim-1.)/(o+1.))+(0.311813862038)*dim
        ref[0,1,1,0]=(0.540512799627)*(1+2.*(dim-1.)/(o+1.))+(-0.00503022927381)*dim
        ref[0,1,1,1]=(0.148835768693)*(1+2.*(dim-1.)/(o+1.))+(-0.133338026157)*dim
        ref[0,1,2,0]=(-0.0921555325482)*(1+2.*(dim-1.)/(o+1.))+(-0.253470191439)*dim
        ref[0,1,2,1]=(0.37975053514)*(1+2.*(dim-1.)/(o+1.))+(0.137491991068)*dim
        ref[0,2,0,0]=(-0.637116147991)*(1+2.*(dim-1.)/(o+1.))+(-0.782657261086)*dim
        ref[0,2,0,1]=(-0.109247728514)*(1+2.*(dim-1.)/(o+1.))+(0.541417493824)*dim
        ref[0,2,1,0]=(-0.933462481479)*(1+2.*(dim-1.)/(o+1.))+(-1.53681028328)*dim
        ref[0,2,1,1]=(-1.19773222964)*(1+2.*(dim-1.)/(o+1.))+(1.31644104653)*dim
        ref[0,2,2,0]=(-1.17964226873)*(1+2.*(dim-1.)/(o+1.))+(1.12940627206)*dim
        ref[0,2,2,1]=(0.199043107375)*(1+2.*(dim-1.)/(o+1.))+(0.694446720792)*dim
        ref[0,3,0,0]=(0.699758278049)*(1+2.*(dim-1.)/(o+1.))+(-1.5865168617)*dim
        ref[0,3,0,1]=(-1.43918025379)*(1+2.*(dim-1.)/(o+1.))+(1.32584017086)*dim
        ref[0,3,1,0]=(0.320617899014)*(1+2.*(dim-1.)/(o+1.))+(-0.63836051502)*dim
        ref[0,3,1,1]=(-0.741820981021)*(1+2.*(dim-1.)/(o+1.))+(-0.463610958784)*dim
        ref[0,3,2,0]=(0.567695730048)*(1+2.*(dim-1.)/(o+1.))+(-0.852917804)*dim
        ref[0,3,2,1]=(1.06481671764)*(1+2.*(dim-1.)/(o+1.))+(-0.363720838369)*dim
        ref[0,4,0,0]=(0.875430749562)*(1+2.*(dim-1.)/(o+1.))+(0.551253069648)*dim
        ref[0,4,0,1]=(1.26260850991)*(1+2.*(dim-1.)/(o+1.))+(0.921634441735)*dim
        ref[0,4,1,0]=(-0.542320585907)*(1+2.*(dim-1.)/(o+1.))+(0.0620413683361)*dim
        ref[0,4,1,1]=(-0.738376217327)*(1+2.*(dim-1.)/(o+1.))+(0.416152836932)*dim
        ref[0,4,2,0]=(-0.332750203458)*(1+2.*(dim-1.)/(o+1.))+(1.22147094265)*dim
        ref[0,4,2,1]=(1.30503980926)*(1+2.*(dim-1.)/(o+1.))+(-1.24682001245)*dim
        ref[1,0,0,0]=(1.20083814495)*(1+2.*(dim-1.)/(o+1.))+(0.224931716401)*dim
        ref[1,0,0,1]=(0.493849187349)*(1+2.*(dim-1.)/(o+1.))+(0.109403168198)*dim
        ref[1,0,1,0]=(0.666795564157)*(1+2.*(dim-1.)/(o+1.))+(-0.188860724968)*dim
        ref[1,0,1,1]=(0.322811827009)*(1+2.*(dim-1.)/(o+1.))+(0.58147616313)*dim
        ref[1,0,2,0]=(-1.21934428158)*(1+2.*(dim-1.)/(o+1.))+(0.698046512357)*dim
        ref[1,0,2,1]=(1.35903157083)*(1+2.*(dim-1.)/(o+1.))+(0.276174469644)*dim
        ref[1,1,0,0]=(-0.668757118699)*(1+2.*(dim-1.)/(o+1.))+(0.184772207756)*dim
        ref[1,1,0,1]=(1.25787213023)*(1+2.*(dim-1.)/(o+1.))+(0.0222646493659)*dim
        ref[1,1,1,0]=(0.167176470046)*(1+2.*(dim-1.)/(o+1.))+(-0.245832394617)*dim
        ref[1,1,1,1]=(-0.659841661688)*(1+2.*(dim-1.)/(o+1.))+(-0.953939685406)*dim
        ref[1,1,2,0]=(1.07102228089)*(1+2.*(dim-1.)/(o+1.))+(-0.132946987715)*dim
        ref[1,1,2,1]=(-0.347830760025)*(1+2.*(dim-1.)/(o+1.))+(-0.484236996915)*dim
        ref[1,2,0,0]=(-0.087001660358)*(1+2.*(dim-1.)/(o+1.))+(1.44180128866)*dim
        ref[1,2,0,1]=(-0.466598066397)*(1+2.*(dim-1.)/(o+1.))+(1.1128931162)*dim
        ref[1,2,1,0]=(-0.996210852643)*(1+2.*(dim-1.)/(o+1.))+(1.24400417763)*dim
        ref[1,2,1,1]=(0.206220408321)*(1+2.*(dim-1.)/(o+1.))+(0.332074088353)*dim
        ref[1,2,2,0]=(0.103151003161)*(1+2.*(dim-1.)/(o+1.))+(1.4480060297)*dim
        ref[1,2,2,1]=(-0.282559222939)*(1+2.*(dim-1.)/(o+1.))+(-1.47590976977)*dim
        ref[1,3,0,0]=(-0.697481881234)*(1+2.*(dim-1.)/(o+1.))+(-0.414345520764)*dim
        ref[1,3,0,1]=(-0.954564688559)*(1+2.*(dim-1.)/(o+1.))+(-0.93833639809)*dim
        ref[1,3,1,0]=(1.74786122066)*(1+2.*(dim-1.)/(o+1.))+(-0.603460563479)*dim
        ref[1,3,1,1]=(0.576696710719)*(1+2.*(dim-1.)/(o+1.))+(1.5449563998)*dim
        ref[1,3,2,0]=(-0.26671915305)*(1+2.*(dim-1.)/(o+1.))+(1.19067969538)*dim
        ref[1,3,2,1]=(1.52058149003)*(1+2.*(dim-1.)/(o+1.))+(0.582969127914)*dim
        ref[1,4,0,0]=(0.214940907208)*(1+2.*(dim-1.)/(o+1.))+(-1.28759712193)*dim
        ref[1,4,0,1]=(-0.564229160887)*(1+2.*(dim-1.)/(o+1.))+(1.46222745131)*dim
        ref[1,4,1,0]=(1.00974626816)*(1+2.*(dim-1.)/(o+1.))+(-0.264193096019)*dim
        ref[1,4,1,1]=(-1.03604400812)*(1+2.*(dim-1.)/(o+1.))+(-0.048064364322)*dim
        ref[1,4,2,0]=(0.215263433059)*(1+2.*(dim-1.)/(o+1.))+(-0.926414177826)*dim
        ref[1,4,2,1]=(0.863018760853)*(1+2.*(dim-1.)/(o+1.))+(-0.210974421418)*dim
        ref[2,0,0,0]=(0.456122143601)*(1+2.*(dim-1.)/(o+1.))+(-0.286656440625)*dim
        ref[2,0,0,1]=(1.62808877765)*(1+2.*(dim-1.)/(o+1.))+(-0.270052895997)*dim
        ref[2,0,1,0]=(-0.839009627862)*(1+2.*(dim-1.)/(o+1.))+(-1.07800157305)*dim
        ref[2,0,1,1]=(-0.755931387477)*(1+2.*(dim-1.)/(o+1.))+(-0.30390722284)*dim
        ref[2,0,2,0]=(-0.304811532322)*(1+2.*(dim-1.)/(o+1.))+(1.00912450174)*dim
        ref[2,0,2,1]=(0.61186776369)*(1+2.*(dim-1.)/(o+1.))+(0.60903675699)*dim
        ref[2,1,0,0]=(0.775746359099)*(1+2.*(dim-1.)/(o+1.))+(-0.0652640754538)*dim
        ref[2,1,0,1]=(1.06797988144)*(1+2.*(dim-1.)/(o+1.))+(0.702021087765)*dim
        ref[2,1,1,0]=(1.62297766272)*(1+2.*(dim-1.)/(o+1.))+(0.913406416145)*dim
        ref[2,1,1,1]=(0.296175986273)*(1+2.*(dim-1.)/(o+1.))+(1.2000216877)*dim
        ref[2,1,2,0]=(-1.39985921748)*(1+2.*(dim-1.)/(o+1.))+(0.793386653567)*dim
        ref[2,1,2,1]=(0.76675815763)*(1+2.*(dim-1.)/(o+1.))+(0.289921670962)*dim
        ref[2,2,0,0]=(-0.795787652043)*(1+2.*(dim-1.)/(o+1.))+(-0.0032846811301)*dim
        ref[2,2,0,1]=(-1.37138393084)*(1+2.*(dim-1.)/(o+1.))+(-0.358175850871)*dim
        ref[2,2,1,0]=(0.188257903467)*(1+2.*(dim-1.)/(o+1.))+(-0.334461225539)*dim
        ref[2,2,1,1]=(0.118391598158)*(1+2.*(dim-1.)/(o+1.))+(0.289816675047)*dim
        ref[2,2,2,0]=(-0.663970999582)*(1+2.*(dim-1.)/(o+1.))+(-0.299159718332)*dim
        ref[2,2,2,1]=(0.556618480008)*(1+2.*(dim-1.)/(o+1.))+(-1.2696788018)*dim
        ref[2,3,0,0]=(-0.485651868997)*(1+2.*(dim-1.)/(o+1.))+(1.17320731927)*dim
        ref[2,3,0,1]=(0.104319213682)*(1+2.*(dim-1.)/(o+1.))+(0.911530372697)*dim
        ref[2,3,1,0]=(0.43248864489)*(1+2.*(dim-1.)/(o+1.))+(-0.0197594552935)*dim
        ref[2,3,1,1]=(0.0276596551936)*(1+2.*(dim-1.)/(o+1.))+(0.770762527568)*dim
        ref[2,3,2,0]=(0.573639628608)*(1+2.*(dim-1.)/(o+1.))+(1.38642927495)*dim
        ref[2,3,2,1]=(-0.502260185391)*(1+2.*(dim-1.)/(o+1.))+(1.62564500839)*dim
        ref[2,4,0,0]=(1.70497128864)*(1+2.*(dim-1.)/(o+1.))+(-0.198778229868)*dim
        ref[2,4,0,1]=(0.906274587146)*(1+2.*(dim-1.)/(o+1.))+(0.55919273648)*dim
        ref[2,4,1,0]=(-1.52420172666)*(1+2.*(dim-1.)/(o+1.))+(-0.828944811982)*dim
        ref[2,4,1,1]=(-0.272628509541)*(1+2.*(dim-1.)/(o+1.))+(0.225820656649)*dim
        ref[2,4,2,0]=(0.129498614356)*(1+2.*(dim-1.)/(o+1.))+(1.27146129393)*dim
        ref[2,4,2,1]=(-0.475554772211)*(1+2.*(dim-1.)/(o+1.))+(0.281442644713)*dim
        ref[3,0,0,0]=(-1.48648385291)*(1+2.*(dim-1.)/(o+1.))+(-0.931627841127)*dim
        ref[3,0,0,1]=(-0.390858507465)*(1+2.*(dim-1.)/(o+1.))+(0.856397607398)*dim
        ref[3,0,1,0]=(-1.42833335584)*(1+2.*(dim-1.)/(o+1.))+(-1.02561188279)*dim
        ref[3,0,1,1]=(1.2169331289)*(1+2.*(dim-1.)/(o+1.))+(1.40436040602)*dim
        ref[3,0,2,0]=(0.119179498805)*(1+2.*(dim-1.)/(o+1.))+(-0.103702199018)*dim
        ref[3,0,2,1]=(0.383449222076)*(1+2.*(dim-1.)/(o+1.))+(-0.581569801398)*dim
        ref[3,1,0,0]=(1.01014326387)*(1+2.*(dim-1.)/(o+1.))+(-0.386805114966)*dim
        ref[3,1,0,1]=(1.01528264021)*(1+2.*(dim-1.)/(o+1.))+(-1.07493042977)*dim
        ref[3,1,1,0]=(0.143671744416)*(1+2.*(dim-1.)/(o+1.))+(0.12692892846)*dim
        ref[3,1,1,1]=(-0.721680326068)*(1+2.*(dim-1.)/(o+1.))+(0.374071312272)*dim
        ref[3,1,2,0]=(-0.441333976613)*(1+2.*(dim-1.)/(o+1.))+(-0.0447508870373)*dim
        ref[3,1,2,1]=(0.146293375075)*(1+2.*(dim-1.)/(o+1.))+(-0.188911772819)*dim
        ref[3,2,0,0]=(-1.15275360871)*(1+2.*(dim-1.)/(o+1.))+(0.57414375209)*dim
        ref[3,2,0,1]=(0.608957863637)*(1+2.*(dim-1.)/(o+1.))+(-0.305371364976)*dim
        ref[3,2,1,0]=(-0.199586716488)*(1+2.*(dim-1.)/(o+1.))+(0.515871232889)*dim
        ref[3,2,1,1]=(1.87059827943)*(1+2.*(dim-1.)/(o+1.))+(-0.453908871458)*dim
        ref[3,2,2,0]=(0.669536036913)*(1+2.*(dim-1.)/(o+1.))+(-0.446352107026)*dim
        ref[3,2,2,1]=(1.01309175725)*(1+2.*(dim-1.)/(o+1.))+(0.126469806303)*dim
        ref[3,3,0,0]=(-0.931195623607)*(1+2.*(dim-1.)/(o+1.))+(-0.00294255911617)*dim
        ref[3,3,0,1]=(0.64672438328)*(1+2.*(dim-1.)/(o+1.))+(0.882258653197)*dim
        ref[3,3,1,0]=(-0.407314290639)*(1+2.*(dim-1.)/(o+1.))+(-0.214359498715)*dim
        ref[3,3,1,1]=(-1.28519824248)*(1+2.*(dim-1.)/(o+1.))+(-1.43812333826)*dim
        ref[3,3,2,0]=(0.620902449652)*(1+2.*(dim-1.)/(o+1.))+(0.361828435252)*dim
        ref[3,3,2,1]=(1.22847378244)*(1+2.*(dim-1.)/(o+1.))+(-0.824434020294)*dim
        ref[3,4,0,0]=(-0.706321782603)*(1+2.*(dim-1.)/(o+1.))+(-0.876870969492)*dim
        ref[3,4,0,1]=(-0.571496862855)*(1+2.*(dim-1.)/(o+1.))+(1.29159816312)*dim
        ref[3,4,1,0]=(-0.102742633631)*(1+2.*(dim-1.)/(o+1.))+(-0.198399341224)*dim
        ref[3,4,1,1]=(0.799297204627)*(1+2.*(dim-1.)/(o+1.))+(0.107627143578)*dim
        ref[3,4,2,0]=(-1.31225544496)*(1+2.*(dim-1.)/(o+1.))+(-0.0971924854666)*dim
        ref[3,4,2,1]=(-0.579455532935)*(1+2.*(dim-1.)/(o+1.))+(-0.328340711373)*dim
      else:
        arg[0,0,0,0]=(-0.727570048979)*x[0]**o+(-0.769770353069)*x[0]+(-0.532381054852)*x[1]**o+(0.590589994369)*x[1]+(0.775348726284)*x[2]**o+(-0.840239939796)*x[2]
        arg[0,0,0,1]=(-0.441856577751)*x[0]**o+(-0.39234308235)*x[0]+(-0.834378647798)*x[1]**o+(0.84269918161)*x[1]+(0.441102591728)*x[2]**o+(0.166331058665)*x[2]
        arg[0,0,1,0]=(0.391773069521)*x[0]**o+(0.496759920942)*x[0]+(0.555386226902)*x[1]**o+(0.980197783628)*x[1]+(0.549622021652)*x[2]**o+(-0.596179893504)*x[2]
        arg[0,0,1,1]=(-0.104197387892)*x[0]**o+(-0.64636505375)*x[0]+(0.377612339028)*x[1]**o+(-0.215111619231)*x[1]+(0.637249692625)*x[2]**o+(-0.0998227822351)*x[2]
        arg[0,0,2,0]=(0.780693043561)*x[0]**o+(0.571405344342)*x[0]+(-0.793680154472)*x[1]**o+(-0.0481469136436)*x[1]+(0.959561051293)*x[2]**o+(0.0756152497582)*x[2]
        arg[0,0,2,1]=(0.88678762282)*x[0]**o+(0.254753272214)*x[0]+(0.344376798516)*x[1]**o+(-0.908135902735)*x[1]+(-0.3584387726)*x[2]**o+(-0.890225037617)*x[2]
        arg[0,1,0,0]=(0.680262917473)*x[0]**o+(0.65860357305)*x[0]+(0.859832322473)*x[1]**o+(0.326105580592)*x[1]+(0.132170297876)*x[2]**o+(-0.928606849775)*x[2]
        arg[0,1,0,1]=(-0.500503822293)*x[0]**o+(-0.456331012174)*x[0]+(0.900478846187)*x[1]**o+(0.350542503874)*x[1]+(0.755849038617)*x[2]**o+(-0.514240007142)*x[2]
        arg[0,1,1,0]=(0.124404217291)*x[0]**o+(-0.499651000805)*x[0]+(-0.968178694708)*x[1]**o+(0.306780500818)*x[1]+(-0.946470794819)*x[2]**o+(0.919424575523)*x[2]
        arg[0,1,1,1]=(0.722520017474)*x[0]**o+(-0.718216593718)*x[0]+(0.801196051455)*x[1]**o+(-0.791951278546)*x[1]+(-0.675975417529)*x[2]**o+(0.355069676047)*x[2]
        arg[0,1,2,0]=(-0.137017428857)*x[0]**o+(-0.760267879487)*x[0]+(0.625869259413)*x[1]**o+(-0.697626713352)*x[1]+(-0.467876751423)*x[2]**o+(0.456475454123)*x[2]
        arg[0,1,2,1]=(0.646536442797)*x[0]**o+(0.0981671744166)*x[0]+(0.477147756414)*x[1]**o+(-0.23211823966)*x[1]+(-0.553808167147)*x[2]**o+(0.405402676302)*x[2]
        arg[0,2,0,0]=(-0.0604502966371)*x[0]**o+(0.297995946813)*x[0]+(0.837045433858)*x[1]**o+(-0.270340963575)*x[1]+(-0.58462190484)*x[2]**o+(0.292953993287)*x[2]
        arg[0,2,0,1]=(-0.463109088268)*x[0]**o+(0.346260157368)*x[0]+(0.0228615151352)*x[1]**o+(0.905507471559)*x[1]+(-0.423777553665)*x[2]**o+(-0.45964535654)*x[2]
        arg[0,2,1,0]=(-0.50761294542)*x[0]**o+(-0.465260578419)*x[0]+(-0.729953382373)*x[1]**o+(0.903666219263)*x[1]+(-0.756794344711)*x[2]**o+(0.817263008431)*x[2]
        arg[0,2,1,1]=(-0.551253781393)*x[0]**o+(-0.911481117802)*x[0]+(0.180349330394)*x[1]**o+(0.419118491074)*x[1]+(-0.691917875703)*x[2]**o+(-0.835365504347)*x[2]
        arg[0,2,2,0]=(0.32567932803)*x[0]**o+(0.413167839883)*x[0]+(0.781320691089)*x[1]**o+(-0.777223093057)*x[1]+(0.297556831029)*x[2]**o+(-0.952676550814)*x[2]
        arg[0,2,2,1]=(0.226250474873)*x[0]**o+(0.279287242249)*x[0]+(0.421824124143)*x[1]**o+(-0.97664927412)*x[1]+(0.703245736838)*x[2]**o+(-0.629660622476)*x[2]
        arg[0,3,0,0]=(-0.138403445102)*x[0]**o+(0.0979152923187)*x[0]+(-0.177116218507)*x[1]**o+(0.0100523961708)*x[1]+(-0.634159492232)*x[2]**o+(-0.847618220907)*x[2]
        arg[0,3,0,1]=(0.612268738957)*x[0]**o+(-0.462682392204)*x[0]+(-0.146623835506)*x[1]**o+(0.699258469487)*x[1]+(0.0139983038691)*x[2]**o+(0.236461011719)*x[2]
        arg[0,3,1,0]=(-0.0518123102809)*x[0]**o+(0.386822059696)*x[0]+(-0.208858744655)*x[1]**o+(0.971393362011)*x[1]+(-0.171649562077)*x[2]**o+(-0.507479926411)*x[2]
        arg[0,3,1,1]=(0.0140396270391)*x[0]**o+(-0.837711757872)*x[0]+(-0.00108924459289)*x[1]**o+(-0.724087080496)*x[1]+(-0.793071741694)*x[2]**o+(0.38654936043)*x[2]
        arg[0,3,2,0]=(-0.897711437057)*x[0]**o+(0.562533782571)*x[0]+(-0.354013039442)*x[1]**o+(-0.473034213198)*x[1]+(0.4598333464)*x[2]**o+(0.467503992754)*x[2]
        arg[0,3,2,1]=(-0.198441232877)*x[0]**o+(-0.411568849672)*x[0]+(0.0850211273328)*x[1]**o+(-0.333702582633)*x[1]+(0.727888461746)*x[2]**o+(-0.174835541819)*x[2]
        arg[0,4,0,0]=(-0.114679585752)*x[0]**o+(0.0446596462171)*x[0]+(-0.950604674991)*x[1]**o+(-0.307622186566)*x[1]+(0.956712754366)*x[2]**o+(0.334630480874)*x[2]
        arg[0,4,0,1]=(0.345475911118)*x[0]**o+(0.834620576123)*x[0]+(0.903176273684)*x[1]**o+(-0.364211856466)*x[1]+(0.910903954488)*x[2]**o+(0.392248545744)*x[2]
        arg[0,4,1,0]=(-0.593360797686)*x[0]**o+(-0.0199892857836)*x[0]+(0.293839091284)*x[1]**o+(0.580208854742)*x[1]+(0.060499817751)*x[2]**o+(0.907498841261)*x[2]
        arg[0,4,1,1]=(-0.446192713548)*x[0]**o+(0.29870251614)*x[0]+(0.754231657585)*x[1]**o+(0.321710920549)*x[1]+(0.287978396665)*x[2]**o+(-0.963473165492)*x[2]
        arg[0,4,2,0]=(0.780174386478)*x[0]**o+(0.603377914483)*x[0]+(-0.841703120145)*x[1]**o+(0.316522795462)*x[1]+(-0.192916583378)*x[2]**o+(-0.734091913995)*x[2]
        arg[0,4,2,1]=(-0.130567656486)*x[0]**o+(-0.342928842892)*x[0]+(-0.0827376128775)*x[1]**o+(-0.771104057041)*x[1]+(0.733571283734)*x[2]**o+(-0.880531838119)*x[2]
        arg[1,0,0,0]=(-0.634428121043)*x[0]**o+(0.294135783753)*x[0]+(-0.566305561339)*x[1]**o+(0.0332700472211)*x[1]+(0.128130119463)*x[2]**o+(0.0468678470468)*x[2]
        arg[1,0,0,1]=(0.82168858107)*x[0]**o+(-0.75544177649)*x[0]+(0.260115562827)*x[1]**o+(-0.149575433725)*x[1]+(-0.780945910032)*x[2]**o+(-0.0539291857071)*x[2]
        arg[1,0,1,0]=(-0.97885291605)*x[0]**o+(0.777955971883)*x[0]+(0.115355866293)*x[1]**o+(-0.456449594589)*x[1]+(-0.832816352387)*x[2]**o+(-0.418070257158)*x[2]
        arg[1,0,1,1]=(-0.908396506245)*x[0]**o+(-0.982147653872)*x[0]+(0.138657052347)*x[1]**o+(0.633837624386)*x[1]+(-0.529748303776)*x[2]**o+(-0.950225664343)*x[2]
        arg[1,0,2,0]=(0.733974435377)*x[0]**o+(-0.758228777635)*x[0]+(0.841368499977)*x[1]**o+(0.871715889444)*x[1]+(0.541055759071)*x[2]**o+(0.669798500731)*x[2]
        arg[1,0,2,1]=(-0.572994548563)*x[0]**o+(0.806241853906)*x[0]+(0.141217952929)*x[1]**o+(0.832291361477)*x[1]+(0.207737014762)*x[2]**o+(-0.279769695917)*x[2]
        arg[1,1,0,0]=(-0.394247339923)*x[0]**o+(0.548338039327)*x[0]+(-0.845822483858)*x[1]**o+(-0.941161341738)*x[1]+(-0.244513027249)*x[2]**o+(-0.524790843131)*x[2]
        arg[1,1,0,1]=(0.509994957918)*x[0]**o+(0.691470630092)*x[0]+(0.249241759795)*x[1]**o+(-0.995293157716)*x[1]+(0.162652088127)*x[2]**o+(0.873578806104)*x[2]
        arg[1,1,1,0]=(-0.22367137056)*x[0]**o+(-0.896306406631)*x[0]+(0.613214186844)*x[1]**o+(-0.460055562163)*x[1]+(-0.704736634406)*x[2]**o+(-0.517741014186)*x[2]
        arg[1,1,1,1]=(-0.426072795689)*x[0]**o+(-0.13452184053)*x[0]+(0.908778671123)*x[1]**o+(0.194472828672)*x[1]+(-0.658909495204)*x[2]**o+(-0.234260758459)*x[2]
        arg[1,1,2,0]=(0.236086562474)*x[0]**o+(0.33312461561)*x[0]+(0.867089643801)*x[1]**o+(-0.720157305369)*x[1]+(0.812845198077)*x[2]**o+(-0.375694713863)*x[2]
        arg[1,1,2,1]=(-0.561145817489)*x[0]**o+(0.746630137163)*x[0]+(-0.809280840008)*x[1]**o+(-0.690912076356)*x[1]+(0.88933641231)*x[2]**o+(0.246647817564)*x[2]
        arg[1,2,0,0]=(0.701319152729)*x[0]**o+(-0.309074024603)*x[0]+(0.567988628717)*x[1]**o+(-0.0596939135728)*x[1]+(0.242135294918)*x[2]**o+(-0.101136005445)*x[2]
        arg[1,2,0,1]=(0.880665674235)*x[0]**o+(-0.833066777852)*x[0]+(-0.0556340786728)*x[1]**o+(0.620095583249)*x[1]+(-0.390556829051)*x[2]**o+(0.793481461864)*x[2]
        arg[1,2,1,0]=(0.832046852431)*x[0]**o+(0.4276990324)*x[0]+(-0.486292331356)*x[1]**o+(0.990296730852)*x[1]+(-0.381732943927)*x[2]**o+(0.225838720655)*x[2]
        arg[1,2,1,1]=(-0.971477034333)*x[0]**o+(0.819409035213)*x[0]+(-0.610341797903)*x[1]**o+(0.842186877805)*x[1]+(0.571187503044)*x[2]**o+(-0.150565006374)*x[2]
        arg[1,2,2,0]=(0.266080624706)*x[0]**o+(0.893377400845)*x[0]+(0.718525426441)*x[1]**o+(-0.847615622286)*x[1]+(0.614914866)*x[2]**o+(0.826503774974)*x[2]
        arg[1,2,2,1]=(0.998060871997)*x[0]**o+(-0.572523112154)*x[0]+(-0.786785614822)*x[1]**o+(0.556952019953)*x[1]+(0.804849394796)*x[2]**o+(-0.201157281118)*x[2]
        arg[1,3,0,0]=(-0.322048105751)*x[0]**o+(0.215441634353)*x[0]+(0.203298400258)*x[1]**o+(0.452071048161)*x[1]+(0.0134315902731)*x[2]**o+(0.240098266028)*x[2]
        arg[1,3,0,1]=(-0.109845136841)*x[0]**o+(0.0479663457074)*x[0]+(0.899394604931)*x[1]**o+(0.853608214151)*x[1]+(-0.27487885579)*x[2]**o+(-0.0236702866567)*x[2]
        arg[1,3,1,0]=(0.948550474293)*x[0]**o+(0.359077475995)*x[0]+(0.655387467125)*x[1]**o+(-0.0501114084308)*x[1]+(-0.244470769181)*x[2]**o+(0.670276065978)*x[2]
        arg[1,3,1,1]=(-0.860769147822)*x[0]**o+(-0.738449680804)*x[0]+(0.840959189643)*x[1]**o+(0.229930283198)*x[1]+(-0.521034843253)*x[2]**o+(-0.50213233896)*x[2]
        arg[1,3,2,0]=(0.29910216992)*x[0]**o+(0.411953562132)*x[0]+(0.777584507308)*x[1]**o+(0.41482951697)*x[1]+(-0.948751119313)*x[2]**o+(-0.526573791564)*x[2]
        arg[1,3,2,1]=(-0.768167653244)*x[0]**o+(-0.318762199437)*x[0]+(0.7566706966)*x[1]**o+(-0.151347269999)*x[1]+(-0.626690637485)*x[2]**o+(-0.731518248609)*x[2]
        arg[1,4,0,0]=(0.0147733854143)*x[0]**o+(-0.812988077571)*x[0]+(0.815385483312)*x[1]**o+(0.285326283687)*x[1]+(0.872502797762)*x[2]**o+(-0.748392032843)*x[2]
        arg[1,4,0,1]=(0.172931930795)*x[0]**o+(0.425088340505)*x[0]+(0.0122969950852)*x[1]**o+(0.556865378995)*x[1]+(0.614686173604)*x[2]**o+(0.966581266651)*x[2]
        arg[1,4,1,0]=(0.725270959871)*x[0]**o+(-0.672683754253)*x[0]+(0.486696038678)*x[1]**o+(-0.53856418144)*x[1]+(-0.295197744155)*x[2]**o+(0.0610876962829)*x[2]
        arg[1,4,1,1]=(0.942300770794)*x[0]**o+(0.876440453364)*x[0]+(0.432652207093)*x[1]**o+(-0.562629961162)*x[1]+(0.497526473424)*x[2]**o+(-0.165824324383)*x[2]
        arg[1,4,2,0]=(-0.668927175078)*x[0]**o+(0.587283304457)*x[0]+(0.588305038304)*x[1]**o+(0.087906366831)*x[1]+(-0.907721198172)*x[2]**o+(-0.506067369924)*x[2]
        arg[1,4,2,1]=(0.899071894553)*x[0]**o+(0.920678485704)*x[0]+(-0.244107214117)*x[1]**o+(0.196327093214)*x[1]+(-0.538075261013)*x[2]**o+(-0.925646646321)*x[2]
        arg[2,0,0,0]=(-0.113452114076)*x[0]**o+(0.541178389983)*x[0]+(0.501083424074)*x[1]**o+(-0.234654083639)*x[1]+(-0.435569149263)*x[2]**o+(0.671517947867)*x[2]
        arg[2,0,0,1]=(-0.00139653535956)*x[0]**o+(-0.93731587459)*x[0]+(-0.387786776826)*x[1]**o+(0.361445822788)*x[1]+(-0.0958490804452)*x[2]**o+(0.22475163005)*x[2]
        arg[2,0,1,0]=(0.100832420019)*x[0]**o+(-0.957577198313)*x[0]+(0.244479124631)*x[1]**o+(-0.589295650233)*x[1]+(-0.11020792463)*x[2]**o+(0.969023424875)*x[2]
        arg[2,0,1,1]=(-0.82757398282)*x[0]**o+(0.665274113482)*x[0]+(0.219624146627)*x[1]**o+(0.604687589384)*x[1]+(0.539774513376)*x[2]**o+(0.676136495778)*x[2]
        arg[2,0,2,0]=(-0.869539421193)*x[0]**o+(0.449645390105)*x[0]+(-0.216356671763)*x[1]**o+(-0.191616800811)*x[1]+(-0.324126591392)*x[2]**o+(0.726527544226)*x[2]
        arg[2,0,2,1]=(-0.522340761977)*x[0]**o+(-0.253136332372)*x[0]+(-0.993693358789)*x[1]**o+(0.689338979208)*x[1]+(-0.73030598892)*x[2]**o+(0.604868622965)*x[2]
        arg[2,1,0,0]=(-0.87285436276)*x[0]**o+(-0.665917385941)*x[0]+(0.105570880359)*x[1]**o+(0.773780432278)*x[1]+(-0.0932677950401)*x[2]**o+(0.55563257373)*x[2]
        arg[2,1,0,1]=(-0.717823895492)*x[0]**o+(-0.843992380543)*x[0]+(0.348730031558)*x[1]**o+(-0.0641727895105)*x[1]+(-0.575332519937)*x[2]**o+(0.566731137573)*x[2]
        arg[2,1,1,0]=(-0.749216309273)*x[0]**o+(0.945232679736)*x[0]+(-0.984694983452)*x[1]**o+(-0.697422850579)*x[1]+(0.573407778839)*x[2]**o+(0.654548962131)*x[2]
        arg[2,1,1,1]=(-0.951637762959)*x[0]**o+(-0.0299724748286)*x[0]+(-0.784527116834)*x[1]**o+(0.490138786409)*x[1]+(0.945743565011)*x[2]**o+(0.158024494836)*x[2]
        arg[2,1,2,0]=(0.80439224967)*x[0]**o+(0.889509225531)*x[0]+(-0.178335319647)*x[1]**o+(-0.365669987413)*x[1]+(-0.170055755127)*x[2]**o+(0.258843711555)*x[2]
        arg[2,1,2,1]=(0.520706901588)*x[0]**o+(0.897178241272)*x[0]+(-0.128495163664)*x[1]**o+(-0.404742110163)*x[1]+(0.140333496135)*x[2]**o+(0.31429703244)*x[2]
        arg[2,2,0,0]=(-0.933201442136)*x[0]**o+(0.0083650170184)*x[0]+(-0.281807642508)*x[1]**o+(-0.09193002721)*x[1]+(-0.422582781227)*x[2]**o+(0.358915935551)*x[2]
        arg[2,2,0,1]=(-0.09712053324)*x[0]**o+(-0.0752448353447)*x[0]+(0.41291697894)*x[1]**o+(0.593823615119)*x[1]+(0.954183095989)*x[2]**o+(0.479576723309)*x[2]
        arg[2,2,1,0]=(-0.48592273499)*x[0]**o+(-0.555807193072)*x[0]+(0.842926147107)*x[1]**o+(-0.0760474608316)*x[1]+(0.864590424311)*x[2]**o+(-0.854721691042)*x[2]
        arg[2,2,1,1]=(0.238261310086)*x[0]**o+(-0.506114025767)*x[0]+(-0.534154656267)*x[1]**o+(-0.282262066272)*x[1]+(-0.241950678835)*x[2]**o+(-0.596913007205)*x[2]
        arg[2,2,2,0]=(0.35002425052)*x[0]**o+(-0.724235915518)*x[0]+(-0.443311896206)*x[1]**o+(0.0183169190559)*x[1]+(-0.180995873436)*x[2]**o+(0.949105760518)*x[2]
        arg[2,2,2,1]=(-0.524596480672)*x[0]**o+(-0.72511087285)*x[0]+(-0.793159588655)*x[1]**o+(0.197579071177)*x[1]+(0.214510478855)*x[2]**o+(0.592011219443)*x[2]
        arg[2,3,0,0]=(-0.777144207151)*x[0]**o+(0.821815068413)*x[0]+(0.828183715461)*x[1]**o+(-0.77742744879)*x[1]+(0.520088528639)*x[2]**o+(0.564796500907)*x[2]
        arg[2,3,0,1]=(-0.444252706506)*x[0]**o+(-0.999658246414)*x[0]+(0.458490846376)*x[1]**o+(0.206371738627)*x[1]+(-0.5943830165)*x[2]**o+(0.125954475804)*x[2]
        arg[2,3,1,0]=(-0.499555379078)*x[0]**o+(0.667767779032)*x[0]+(-0.169053599596)*x[1]**o+(0.587608309525)*x[1]+(-0.99317016838)*x[2]**o+(-0.580232194372)*x[2]
        arg[2,3,1,1]=(0.881736170108)*x[0]**o+(-0.0221128885544)*x[0]+(-0.239202356219)*x[1]**o+(-0.825250840397)*x[1]+(-0.34138233902)*x[2]**o+(-0.900437209752)*x[2]
        arg[2,3,2,0]=(-0.810415345283)*x[0]**o+(0.671639285402)*x[0]+(0.0577696593549)*x[1]**o+(0.12466039802)*x[1]+(-0.164952472389)*x[2]**o+(-0.881356759483)*x[2]
        arg[2,3,2,1]=(0.641411687742)*x[0]**o+(0.939465110042)*x[0]+(0.182541049996)*x[1]**o+(-0.502132165421)*x[1]+(-0.914212670088)*x[2]**o+(0.361356243622)*x[2]
        arg[2,4,0,0]=(0.696033328872)*x[0]**o+(0.854329989895)*x[0]+(0.66092830442)*x[1]**o+(-0.165623528752)*x[1]+(0.587322879532)*x[2]**o+(-0.441145527759)*x[2]
        arg[2,4,0,1]=(0.885589185979)*x[0]**o+(0.170856006811)*x[0]+(-0.500114387063)*x[1]**o+(-0.772574748562)*x[1]+(0.599300325869)*x[2]**o+(0.0950082175327)*x[2]
        arg[2,4,1,0]=(0.877739402691)*x[0]**o+(0.494715865857)*x[0]+(-0.4548916845)*x[1]**o+(-0.108381552398)*x[1]+(-0.620522177203)*x[2]**o+(-0.375546939131)*x[2]
        arg[2,4,1,1]=(-0.108044365878)*x[0]**o+(-0.592058059304)*x[0]+(-0.64481990676)*x[1]**o+(-0.903674017194)*x[1]+(0.0964052945913)*x[2]**o+(0.357819426568)*x[2]
        arg[2,4,2,0]=(-0.686157836559)*x[0]**o+(0.585514351597)*x[0]+(-0.5623443108)*x[1]**o+(-0.636059380122)*x[1]+(0.469863235081)*x[2]**o+(0.374567303622)*x[2]
        arg[2,4,2,1]=(0.214800008136)*x[0]**o+(0.0680448664819)*x[0]+(0.722930302776)*x[1]**o+(0.107573165145)*x[1]+(0.0027804595356)*x[2]**o+(0.725566605544)*x[2]
        arg[3,0,0,0]=(-0.273256791086)*x[0]**o+(0.801129507153)*x[0]+(-0.840227644476)*x[1]**o+(-0.31594258736)*x[1]+(-0.140806177065)*x[2]**o+(-0.0958103115672)*x[2]
        arg[3,0,0,1]=(0.588534583503)*x[0]**o+(0.242235627811)*x[0]+(-0.541539488722)*x[1]**o+(0.624385574868)*x[1]+(-0.0641371754789)*x[2]**o+(-0.635676751281)*x[2]
        arg[3,0,1,0]=(0.81967100218)*x[0]**o+(0.671597165436)*x[0]+(0.160179001916)*x[1]**o+(-0.226714136475)*x[1]+(-0.523709112534)*x[2]**o+(0.932797007296)*x[2]
        arg[3,0,1,1]=(0.707793115256)*x[0]**o+(-0.0995173271712)*x[0]+(-0.594597614398)*x[1]**o+(-0.72096550461)*x[1]+(-0.373631371949)*x[2]**o+(-0.282151921803)*x[2]
        arg[3,0,2,0]=(0.873307297971)*x[0]**o+(0.0950091191819)*x[0]+(-0.534108539131)*x[1]**o+(0.567381193719)*x[1]+(-0.276691140405)*x[2]**o+(0.84882104254)*x[2]
        arg[3,0,2,1]=(-0.895148878411)*x[0]**o+(-0.880973451462)*x[0]+(0.105602819158)*x[1]**o+(-0.836475687487)*x[1]+(-0.18380164164)*x[2]**o+(-0.391643939001)*x[2]
        arg[3,1,0,0]=(-0.144070349397)*x[0]**o+(-0.247578520564)*x[0]+(0.512469248899)*x[1]**o+(0.0369808259097)*x[1]+(0.67517925106)*x[2]**o+(-0.306122403522)*x[2]
        arg[3,1,0,1]=(0.389023591419)*x[0]**o+(0.0335324297663)*x[0]+(0.166380432563)*x[1]**o+(0.415133512101)*x[1]+(-0.956055398028)*x[2]**o+(-0.069254560597)*x[2]
        arg[3,1,1,0]=(-0.253194755795)*x[0]**o+(-0.432240637655)*x[0]+(0.886730830871)*x[1]**o+(0.94685399231)*x[1]+(-0.0193581886643)*x[2]**o+(-0.828124732153)*x[2]
        arg[3,1,1,1]=(0.948739050057)*x[0]**o+(-0.200402018913)*x[0]+(0.062573402994)*x[1]**o+(-0.408727698463)*x[1]+(-0.0334411087309)*x[2]**o+(-0.676147119957)*x[2]
        arg[3,1,2,0]=(0.249237167322)*x[0]**o+(-0.621387879739)*x[0]+(-0.745335388751)*x[1]**o+(-0.183173004074)*x[1]+(0.56632553768)*x[2]**o+(-0.519909176525)*x[2]
        arg[3,1,2,1]=(-0.660822495815)*x[0]**o+(0.0459275869317)*x[0]+(-0.0853936912232)*x[1]**o+(-0.895502076207)*x[1]+(0.150013238845)*x[2]**o+(0.201527905115)*x[2]
        arg[3,2,0,0]=(0.0846880899978)*x[0]**o+(-0.450721252185)*x[0]+(0.0100224498309)*x[1]**o+(-0.729399043167)*x[1]+(0.249073157161)*x[2]**o+(-0.313056982897)*x[2]
        arg[3,2,0,1]=(0.349123663394)*x[0]**o+(-0.0439266646004)*x[0]+(0.201749457716)*x[1]**o+(-0.29961557135)*x[1]+(0.247657591065)*x[2]**o+(0.702448030656)*x[2]
        arg[3,2,1,0]=(-0.67551758529)*x[0]**o+(-0.856271486186)*x[0]+(-0.337698270178)*x[1]**o+(-0.855526501932)*x[1]+(0.204628923051)*x[2]**o+(0.0425229547861)*x[2]
        arg[3,2,1,1]=(0.913895958603)*x[0]**o+(0.33488814727)*x[0]+(0.725768429772)*x[1]**o+(0.31487179035)*x[1]+(-0.926895768372)*x[2]**o+(-0.303623128744)*x[2]
        arg[3,2,2,0]=(0.695497510732)*x[0]**o+(-0.71819800795)*x[0]+(0.0949892877746)*x[1]**o+(-0.371952684199)*x[1]+(-0.712679288901)*x[2]**o+(-0.839032166941)*x[2]
        arg[3,2,2,1]=(0.847915632309)*x[0]**o+(0.489320767225)*x[0]+(0.884694888545)*x[1]**o+(-0.836734066966)*x[1]+(-0.782643394677)*x[2]**o+(0.998012000062)*x[2]
        arg[3,3,0,0]=(0.400343531732)*x[0]**o+(-0.977399566427)*x[0]+(-0.0491815761683)*x[1]**o+(0.820343030844)*x[1]+(0.145276965371)*x[2]**o+(0.106181410329)*x[2]
        arg[3,3,0,1]=(0.39509119571)*x[0]**o+(-0.532152199068)*x[0]+(0.0961163031465)*x[1]**o+(-0.845734786516)*x[1]+(0.0685207262665)*x[2]**o+(0.10725771579)*x[2]
        arg[3,3,1,0]=(-0.326475146957)*x[0]**o+(0.0491193078417)*x[0]+(-0.193405294452)*x[1]**o+(0.283579426128)*x[1]+(0.885442239373)*x[2]**o+(-0.838544460442)*x[2]
        arg[3,3,1,1]=(-0.502234549601)*x[0]**o+(-0.764617462736)*x[0]+(0.52701849705)*x[1]**o+(-0.423245302614)*x[1]+(0.728096320995)*x[2]**o+(0.143356081954)*x[2]
        arg[3,3,2,0]=(0.924562662058)*x[0]**o+(-0.0120799887424)*x[0]+(-0.342886608544)*x[1]**o+(0.609435784717)*x[1]+(0.392296911228)*x[2]**o+(0.572866548262)*x[2]
        arg[3,3,2,1]=(0.661267105524)*x[0]**o+(-0.71847134372)*x[0]+(-0.913866381825)*x[1]**o+(-0.502556458508)*x[1]+(0.201214639904)*x[2]**o+(-0.431939125167)*x[2]
        arg[3,4,0,0]=(0.391785002874)*x[0]**o+(0.424885787931)*x[0]+(-0.377558346944)*x[1]**o+(-0.556924666311)*x[1]+(0.511852194353)*x[2]**o+(0.514261734074)*x[2]
        arg[3,4,0,1]=(-0.269148156354)*x[0]**o+(0.692033978804)*x[0]+(-0.390604590454)*x[1]**o+(-0.551681907742)*x[1]+(0.0613720420346)*x[2]**o+(-0.508841982991)*x[2]
        arg[3,4,1,0]=(0.593170387694)*x[0]**o+(-0.00676980821753)*x[0]+(0.788564472913)*x[1]**o+(0.823094059865)*x[1]+(0.52895485971)*x[2]**o+(0.88849618397)*x[2]
        arg[3,4,1,1]=(0.772185283881)*x[0]**o+(0.759014894154)*x[0]+(-0.142739578724)*x[1]**o+(-0.141899255898)*x[1]+(0.479086662114)*x[2]**o+(0.68993908918)*x[2]
        arg[3,4,2,0]=(0.998097253401)*x[0]**o+(0.621254339576)*x[0]+(-0.692540865624)*x[1]**o+(0.716545091382)*x[1]+(0.478209703594)*x[2]**o+(0.259063088226)*x[2]
        arg[3,4,2,1]=(-0.321962318028)*x[0]**o+(0.0377531939042)*x[0]+(-0.362041758315)*x[1]**o+(0.632474956936)*x[1]+(0.133972776903)*x[2]**o+(-0.341803258375)*x[2]
        ref[0,0,0,0]=(-0.484602377547)*(1+2.*(dim-1.)/(o+1.))+(-1.0194202985)*dim
        ref[0,0,0,1]=(-0.835132633821)*(1+2.*(dim-1.)/(o+1.))+(0.616687157924)*dim
        ref[0,0,1,0]=(1.49678131808)*(1+2.*(dim-1.)/(o+1.))+(0.880777811066)*dim
        ref[0,0,1,1]=(0.910664643761)*(1+2.*(dim-1.)/(o+1.))+(-0.961299455217)*dim
        ref[0,0,2,0]=(0.946573940383)*(1+2.*(dim-1.)/(o+1.))+(0.598873680457)*dim
        ref[0,0,2,1]=(0.872725648736)*(1+2.*(dim-1.)/(o+1.))+(-1.54360766814)*dim
        ref[0,1,0,0]=(1.67226553782)*(1+2.*(dim-1.)/(o+1.))+(0.0561023038674)*dim
        ref[0,1,0,1]=(1.15582406251)*(1+2.*(dim-1.)/(o+1.))+(-0.620028515442)*dim
        ref[0,1,1,0]=(-1.79024527224)*(1+2.*(dim-1.)/(o+1.))+(0.726554075535)*dim
        ref[0,1,1,1]=(0.8477406514)*(1+2.*(dim-1.)/(o+1.))+(-1.15509819622)*dim
        ref[0,1,2,0]=(0.0209750791326)*(1+2.*(dim-1.)/(o+1.))+(-1.00141913872)*dim
        ref[0,1,2,1]=(0.569876032064)*(1+2.*(dim-1.)/(o+1.))+(0.271451611059)*dim
        ref[0,2,0,0]=(0.19197323238)*(1+2.*(dim-1.)/(o+1.))+(0.320608976525)*dim
        ref[0,2,0,1]=(-0.864025126798)*(1+2.*(dim-1.)/(o+1.))+(0.792122272387)*dim
        ref[0,2,1,0]=(-1.9943606725)*(1+2.*(dim-1.)/(o+1.))+(1.25566864927)*dim
        ref[0,2,1,1]=(-1.0628223267)*(1+2.*(dim-1.)/(o+1.))+(-1.32772813107)*dim
        ref[0,2,2,0]=(1.40455685015)*(1+2.*(dim-1.)/(o+1.))+(-1.31673180399)*dim
        ref[0,2,2,1]=(1.35132033585)*(1+2.*(dim-1.)/(o+1.))+(-1.32702265435)*dim
        ref[0,3,0,0]=(-0.949679155841)*(1+2.*(dim-1.)/(o+1.))+(-0.739650532418)*dim
        ref[0,3,0,1]=(0.47964320732)*(1+2.*(dim-1.)/(o+1.))+(0.473037089003)*dim
        ref[0,3,1,0]=(-0.432320617013)*(1+2.*(dim-1.)/(o+1.))+(0.850735495297)*dim
        ref[0,3,1,1]=(-0.780121359247)*(1+2.*(dim-1.)/(o+1.))+(-1.17524947794)*dim
        ref[0,3,2,0]=(-0.791891130098)*(1+2.*(dim-1.)/(o+1.))+(0.557003562128)*dim
        ref[0,3,2,1]=(0.614468356202)*(1+2.*(dim-1.)/(o+1.))+(-0.920106974125)*dim
        ref[0,4,0,0]=(-0.108571506377)*(1+2.*(dim-1.)/(o+1.))+(0.0716679405256)*dim
        ref[0,4,0,1]=(2.15955613929)*(1+2.*(dim-1.)/(o+1.))+(0.862657265402)*dim
        ref[0,4,1,0]=(-0.239021888652)*(1+2.*(dim-1.)/(o+1.))+(1.46771841022)*dim
        ref[0,4,1,1]=(0.596017340702)*(1+2.*(dim-1.)/(o+1.))+(-0.343059728803)*dim
        ref[0,4,2,0]=(-0.254445317045)*(1+2.*(dim-1.)/(o+1.))+(0.185808795951)*dim
        ref[0,4,2,1]=(0.520266014371)*(1+2.*(dim-1.)/(o+1.))+(-1.99456473805)*dim
        ref[1,0,0,0]=(-1.07260356292)*(1+2.*(dim-1.)/(o+1.))+(0.374273678021)*dim
        ref[1,0,0,1]=(0.300858233864)*(1+2.*(dim-1.)/(o+1.))+(-0.958946395922)*dim
        ref[1,0,1,0]=(-1.69631340214)*(1+2.*(dim-1.)/(o+1.))+(-0.096563879864)*dim
        ref[1,0,1,1]=(-1.29948775767)*(1+2.*(dim-1.)/(o+1.))+(-1.29853569383)*dim
        ref[1,0,2,0]=(2.11639869443)*(1+2.*(dim-1.)/(o+1.))+(0.783285612541)*dim
        ref[1,0,2,1]=(-0.224039580872)*(1+2.*(dim-1.)/(o+1.))+(1.35876351947)*dim
        ref[1,1,0,0]=(-1.48458285103)*(1+2.*(dim-1.)/(o+1.))+(-0.917614145542)*dim
        ref[1,1,0,1]=(0.92188880584)*(1+2.*(dim-1.)/(o+1.))+(0.569756278479)*dim
        ref[1,1,1,0]=(-0.315193818122)*(1+2.*(dim-1.)/(o+1.))+(-1.87410298298)*dim
        ref[1,1,1,1]=(-0.17620361977)*(1+2.*(dim-1.)/(o+1.))+(-0.174309770317)*dim
        ref[1,1,2,0]=(1.91602140435)*(1+2.*(dim-1.)/(o+1.))+(-0.762727403622)*dim
        ref[1,1,2,1]=(-0.481090245186)*(1+2.*(dim-1.)/(o+1.))+(0.302365878371)*dim
        ref[1,2,0,0]=(1.51144307636)*(1+2.*(dim-1.)/(o+1.))+(-0.469903943621)*dim
        ref[1,2,0,1]=(0.434474766511)*(1+2.*(dim-1.)/(o+1.))+(0.580510267262)*dim
        ref[1,2,1,0]=(-0.0359784228521)*(1+2.*(dim-1.)/(o+1.))+(1.64383448391)*dim
        ref[1,2,1,1]=(-1.01063132919)*(1+2.*(dim-1.)/(o+1.))+(1.51103090664)*dim
        ref[1,2,2,0]=(1.59952091715)*(1+2.*(dim-1.)/(o+1.))+(0.872265553532)*dim
        ref[1,2,2,1]=(1.01612465197)*(1+2.*(dim-1.)/(o+1.))+(-0.216728373319)*dim
        ref[1,3,0,0]=(-0.10531811522)*(1+2.*(dim-1.)/(o+1.))+(0.907610948542)*dim
        ref[1,3,0,1]=(0.5146706123)*(1+2.*(dim-1.)/(o+1.))+(0.877904273202)*dim
        ref[1,3,1,0]=(1.35946717224)*(1+2.*(dim-1.)/(o+1.))+(0.979242133542)*dim
        ref[1,3,1,1]=(-0.540844801432)*(1+2.*(dim-1.)/(o+1.))+(-1.01065173657)*dim
        ref[1,3,2,0]=(0.127935557915)*(1+2.*(dim-1.)/(o+1.))+(0.300209287538)*dim
        ref[1,3,2,1]=(-0.63818759413)*(1+2.*(dim-1.)/(o+1.))+(-1.20162771805)*dim
        ref[1,4,0,0]=(1.70266166649)*(1+2.*(dim-1.)/(o+1.))+(-1.27605382673)*dim
        ref[1,4,0,1]=(0.799915099484)*(1+2.*(dim-1.)/(o+1.))+(1.94853498615)*dim
        ref[1,4,1,0]=(0.916769254394)*(1+2.*(dim-1.)/(o+1.))+(-1.15016023941)*dim
        ref[1,4,1,1]=(1.87247945131)*(1+2.*(dim-1.)/(o+1.))+(0.147986167818)*dim
        ref[1,4,2,0]=(-0.988343334946)*(1+2.*(dim-1.)/(o+1.))+(0.169122301365)*dim
        ref[1,4,2,1]=(0.116889419423)*(1+2.*(dim-1.)/(o+1.))+(0.191358932597)*dim
        ref[2,0,0,0]=(-0.0479378392653)*(1+2.*(dim-1.)/(o+1.))+(0.978042254211)*dim
        ref[2,0,0,1]=(-0.485032392631)*(1+2.*(dim-1.)/(o+1.))+(-0.351118421752)*dim
        ref[2,0,1,0]=(0.23510362002)*(1+2.*(dim-1.)/(o+1.))+(-0.577849423671)*dim
        ref[2,0,1,1]=(-0.068175322817)*(1+2.*(dim-1.)/(o+1.))+(1.94609819864)*dim
        ref[2,0,2,0]=(-1.41002268435)*(1+2.*(dim-1.)/(o+1.))+(0.98455613352)*dim
        ref[2,0,2,1]=(-2.24634010969)*(1+2.*(dim-1.)/(o+1.))+(1.0410712698)*dim
        ref[2,1,0,0]=(-0.860551277441)*(1+2.*(dim-1.)/(o+1.))+(0.663495620068)*dim
        ref[2,1,0,1]=(-0.944426383871)*(1+2.*(dim-1.)/(o+1.))+(-0.341434032481)*dim
        ref[2,1,1,0]=(-1.16050351389)*(1+2.*(dim-1.)/(o+1.))+(0.902358791288)*dim
        ref[2,1,1,1]=(-0.790421314781)*(1+2.*(dim-1.)/(o+1.))+(0.618190806416)*dim
        ref[2,1,2,0]=(0.456001174897)*(1+2.*(dim-1.)/(o+1.))+(0.782682949672)*dim
        ref[2,1,2,1]=(0.532545234059)*(1+2.*(dim-1.)/(o+1.))+(0.806733163549)*dim
        ref[2,2,0,0]=(-1.63759186587)*(1+2.*(dim-1.)/(o+1.))+(0.275350925359)*dim
        ref[2,2,0,1]=(1.26997954169)*(1+2.*(dim-1.)/(o+1.))+(0.998155503083)*dim
        ref[2,2,1,0]=(1.22159383643)*(1+2.*(dim-1.)/(o+1.))+(-1.48657634495)*dim
        ref[2,2,1,1]=(-0.537844025016)*(1+2.*(dim-1.)/(o+1.))+(-1.38528909924)*dim
        ref[2,2,2,0]=(-0.274283519122)*(1+2.*(dim-1.)/(o+1.))+(0.243186764056)*dim
        ref[2,2,2,1]=(-1.10324559047)*(1+2.*(dim-1.)/(o+1.))+(0.0644794177698)*dim
        ref[2,3,0,0]=(0.571128036949)*(1+2.*(dim-1.)/(o+1.))+(0.60918412053)*dim
        ref[2,3,0,1]=(-0.580144876629)*(1+2.*(dim-1.)/(o+1.))+(-0.667332031982)*dim
        ref[2,3,1,0]=(-1.66177914705)*(1+2.*(dim-1.)/(o+1.))+(0.675143894185)*dim
        ref[2,3,1,1]=(0.301151474869)*(1+2.*(dim-1.)/(o+1.))+(-1.7478009387)*dim
        ref[2,3,2,0]=(-0.917598158317)*(1+2.*(dim-1.)/(o+1.))+(-0.0850570760611)*dim
        ref[2,3,2,1]=(-0.09025993235)*(1+2.*(dim-1.)/(o+1.))+(0.798689188243)*dim
        ref[2,4,0,0]=(1.94428451282)*(1+2.*(dim-1.)/(o+1.))+(0.247560933384)*dim
        ref[2,4,0,1]=(0.984775124785)*(1+2.*(dim-1.)/(o+1.))+(-0.506710524219)*dim
        ref[2,4,1,0]=(-0.197674459012)*(1+2.*(dim-1.)/(o+1.))+(0.0107873743289)*dim
        ref[2,4,1,1]=(-0.656458978046)*(1+2.*(dim-1.)/(o+1.))+(-1.13791264993)*dim
        ref[2,4,2,0]=(-0.778638912279)*(1+2.*(dim-1.)/(o+1.))+(0.324022275097)*dim
        ref[2,4,2,1]=(0.940510770447)*(1+2.*(dim-1.)/(o+1.))+(0.901184637171)*dim
        ref[3,0,0,0]=(-1.25429061263)*(1+2.*(dim-1.)/(o+1.))+(0.389376608226)*dim
        ref[3,0,0,1]=(-0.0171420806973)*(1+2.*(dim-1.)/(o+1.))+(0.230944451397)*dim
        ref[3,0,1,0]=(0.456140891562)*(1+2.*(dim-1.)/(o+1.))+(1.37768003626)*dim
        ref[3,0,1,1]=(-0.260435871091)*(1+2.*(dim-1.)/(o+1.))+(-1.10263475358)*dim
        ref[3,0,2,0]=(0.0625076184349)*(1+2.*(dim-1.)/(o+1.))+(1.51121135544)*dim
        ref[3,0,2,1]=(-0.973347700893)*(1+2.*(dim-1.)/(o+1.))+(-2.10909307795)*dim
        ref[3,1,0,0]=(1.04357815056)*(1+2.*(dim-1.)/(o+1.))+(-0.516720098176)*dim
        ref[3,1,0,1]=(-0.400651374046)*(1+2.*(dim-1.)/(o+1.))+(0.37941138127)*dim
        ref[3,1,1,0]=(0.614177886412)*(1+2.*(dim-1.)/(o+1.))+(-0.313511377498)*dim
        ref[3,1,1,1]=(0.97787134432)*(1+2.*(dim-1.)/(o+1.))+(-1.28527683733)*dim
        ref[3,1,2,0]=(0.0702273162506)*(1+2.*(dim-1.)/(o+1.))+(-1.32447006034)*dim
        ref[3,1,2,1]=(-0.596202948193)*(1+2.*(dim-1.)/(o+1.))+(-0.648046584161)*dim
        ref[3,2,0,0]=(0.34378369699)*(1+2.*(dim-1.)/(o+1.))+(-1.49317727825)*dim
        ref[3,2,0,1]=(0.798530712175)*(1+2.*(dim-1.)/(o+1.))+(0.358905794706)*dim
        ref[3,2,1,0]=(-0.808586932417)*(1+2.*(dim-1.)/(o+1.))+(-1.66927503333)*dim
        ref[3,2,1,1]=(0.712768620002)*(1+2.*(dim-1.)/(o+1.))+(0.346136808876)*dim
        ref[3,2,2,0]=(0.0778075096053)*(1+2.*(dim-1.)/(o+1.))+(-1.92918285909)*dim
        ref[3,2,2,1]=(0.949967126178)*(1+2.*(dim-1.)/(o+1.))+(0.650598700321)*dim
        ref[3,3,0,0]=(0.496438920934)*(1+2.*(dim-1.)/(o+1.))+(-0.0508751252546)*dim
        ref[3,3,0,1]=(0.559728225123)*(1+2.*(dim-1.)/(o+1.))+(-1.27062926979)*dim
        ref[3,3,1,0]=(0.365561797964)*(1+2.*(dim-1.)/(o+1.))+(-0.505845726473)*dim
        ref[3,3,1,1]=(0.752880268444)*(1+2.*(dim-1.)/(o+1.))+(-1.0445066834)*dim
        ref[3,3,2,0]=(0.973972964742)*(1+2.*(dim-1.)/(o+1.))+(1.17022234424)*dim
        ref[3,3,2,1]=(-0.0513846363967)*(1+2.*(dim-1.)/(o+1.))+(-1.65296692739)*dim
        ref[3,4,0,0]=(0.526078850284)*(1+2.*(dim-1.)/(o+1.))+(0.382222855695)*dim
        ref[3,4,0,1]=(-0.598380704773)*(1+2.*(dim-1.)/(o+1.))+(-0.368489911929)*dim
        ref[3,4,1,0]=(1.91068972032)*(1+2.*(dim-1.)/(o+1.))+(1.70482043562)*dim
        ref[3,4,1,1]=(1.10853236727)*(1+2.*(dim-1.)/(o+1.))+(1.30705472744)*dim
        ref[3,4,2,0]=(0.78376609137)*(1+2.*(dim-1.)/(o+1.))+(1.59686251918)*dim
        ref[3,4,2,1]=(-0.550031299439)*(1+2.*(dim-1.)/(o+1.))+(0.328424892465)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.506621877215)*x[0]**o+(-0.45560383685)*x[0]+(-0.598137301806)*x[1]**o+(0.784241595615)*x[1]
        ref=(-0.0915154245913)*(1+2.*(dim-1.)/(o+1.))+(0.328637758765)*dim
      else:
        arg=(-0.505489131484)*x[0]**o+(0.58139159259)*x[0]+(0.458086828888)*x[1]**o+(0.549090746672)*x[1]+(-0.0882325936167)*x[2]**o+(0.0170295954808)*x[2]
        ref=(-0.135634896213)*(1+2.*(dim-1.)/(o+1.))+(1.14751193474)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.88779433873)*x[0]**o+(0.507955012301)*x[0]+(0.652739274924)*x[1]**o+(-0.971975755989)*x[1]
        arg[1]=(-0.0812690931139)*x[0]**o+(-0.283880962354)*x[0]+(0.094904517439)*x[1]**o+(-0.940528164353)*x[1]
        ref[0]=(-0.235055063806)*(1+2.*(dim-1.)/(o+1.))+(-0.464020743688)*dim
        ref[1]=(0.0136354243251)*(1+2.*(dim-1.)/(o+1.))+(-1.22440912671)*dim
      else:
        arg[0]=(-0.506412384389)*x[0]**o+(-0.465794405394)*x[0]+(0.346977499982)*x[1]**o+(0.386181176987)*x[1]+(-0.476729572397)*x[2]**o+(-0.956945933508)*x[2]
        arg[1]=(0.267092807416)*x[0]**o+(-0.54211155802)*x[0]+(-0.27815411727)*x[1]**o+(-0.984542450102)*x[1]+(0.656389296076)*x[2]**o+(0.0395863971401)*x[2]
        ref[0]=(-0.636164456804)*(1+2.*(dim-1.)/(o+1.))+(-1.03655916191)*dim
        ref[1]=(0.645327986222)*(1+2.*(dim-1.)/(o+1.))+(-1.48706761098)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.0607636892262)*x[0]**o+(0.740866012668)*x[0]+(0.145223136928)*x[1]**o+(-0.738387773526)*x[1]
        arg[0,1]=(0.0292598484511)*x[0]**o+(-0.189208290889)*x[0]+(0.902021405165)*x[1]**o+(0.423106545559)*x[1]
        arg[0,2]=(0.519658217226)*x[0]**o+(-0.768406018136)*x[0]+(-0.189532527097)*x[1]**o+(0.256417429833)*x[1]
        arg[0,3]=(-0.565327663481)*x[0]**o+(0.876762003673)*x[0]+(-0.243804758063)*x[1]**o+(0.399476036454)*x[1]
        arg[0,4]=(0.0632252409818)*x[0]**o+(0.0951130662538)*x[0]+(0.185662425763)*x[1]**o+(0.907446482506)*x[1]
        arg[1,0]=(-0.430504584794)*x[0]**o+(0.864814314023)*x[0]+(0.47460095501)*x[1]**o+(-0.599679142188)*x[1]
        arg[1,1]=(0.699618175298)*x[0]**o+(-0.561782605517)*x[0]+(-0.304250865015)*x[1]**o+(-0.341836776795)*x[1]
        arg[1,2]=(0.856956767316)*x[0]**o+(0.610026903213)*x[0]+(-0.666749954599)*x[1]**o+(0.885388885234)*x[1]
        arg[1,3]=(0.699377007631)*x[0]**o+(0.699906750287)*x[0]+(0.486690669414)*x[1]**o+(0.167425901549)*x[1]
        arg[1,4]=(0.466701415896)*x[0]**o+(-0.0719752759126)*x[0]+(0.795662047062)*x[1]**o+(0.979862503926)*x[1]
        arg[2,0]=(-0.912218732612)*x[0]**o+(0.950990472894)*x[0]+(-0.527999708902)*x[1]**o+(0.695735593043)*x[1]
        arg[2,1]=(-0.223468270048)*x[0]**o+(-0.882565147673)*x[0]+(0.750516681645)*x[1]**o+(-0.652794590093)*x[1]
        arg[2,2]=(0.897646682412)*x[0]**o+(-0.815401214058)*x[0]+(0.629680741327)*x[1]**o+(-0.9716988156)*x[1]
        arg[2,3]=(0.279739517524)*x[0]**o+(-0.0859797206016)*x[0]+(-0.0454325622757)*x[1]**o+(0.301214753868)*x[1]
        arg[2,4]=(-0.358056827092)*x[0]**o+(0.254207762658)*x[0]+(0.831499219864)*x[1]**o+(0.0930388028811)*x[1]
        arg[3,0]=(0.0761520020047)*x[0]**o+(0.650696394486)*x[0]+(-0.132369660678)*x[1]**o+(-0.321220385971)*x[1]
        arg[3,1]=(0.238715231585)*x[0]**o+(-0.297219391083)*x[0]+(-0.15446552614)*x[1]**o+(0.860560615056)*x[1]
        arg[3,2]=(0.708864754538)*x[0]**o+(-0.221186702958)*x[0]+(-0.478127729891)*x[1]**o+(0.0309604412212)*x[1]
        arg[3,3]=(0.362909366479)*x[0]**o+(0.382279106813)*x[0]+(0.827905062086)*x[1]**o+(-0.303048355949)*x[1]
        arg[3,4]=(0.573736946844)*x[0]**o+(0.645059764766)*x[0]+(0.301945383678)*x[1]**o+(-0.103083649994)*x[1]
        ref[0,0]=(0.205986826154)*(1+2.*(dim-1.)/(o+1.))+(0.0024782391428)*dim
        ref[0,1]=(0.931281253617)*(1+2.*(dim-1.)/(o+1.))+(0.23389825467)*dim
        ref[0,2]=(0.330125690129)*(1+2.*(dim-1.)/(o+1.))+(-0.511988588303)*dim
        ref[0,3]=(-0.809132421544)*(1+2.*(dim-1.)/(o+1.))+(1.27623804013)*dim
        ref[0,4]=(0.248887666745)*(1+2.*(dim-1.)/(o+1.))+(1.00255954876)*dim
        ref[1,0]=(0.044096370216)*(1+2.*(dim-1.)/(o+1.))+(0.265135171835)*dim
        ref[1,1]=(0.395367310283)*(1+2.*(dim-1.)/(o+1.))+(-0.903619382312)*dim
        ref[1,2]=(0.190206812717)*(1+2.*(dim-1.)/(o+1.))+(1.49541578845)*dim
        ref[1,3]=(1.18606767705)*(1+2.*(dim-1.)/(o+1.))+(0.867332651836)*dim
        ref[1,4]=(1.26236346296)*(1+2.*(dim-1.)/(o+1.))+(0.907887228013)*dim
        ref[2,0]=(-1.44021844151)*(1+2.*(dim-1.)/(o+1.))+(1.64672606594)*dim
        ref[2,1]=(0.527048411596)*(1+2.*(dim-1.)/(o+1.))+(-1.53535973777)*dim
        ref[2,2]=(1.52732742374)*(1+2.*(dim-1.)/(o+1.))+(-1.78710002966)*dim
        ref[2,3]=(0.234306955249)*(1+2.*(dim-1.)/(o+1.))+(0.215235033266)*dim
        ref[2,4]=(0.473442392772)*(1+2.*(dim-1.)/(o+1.))+(0.34724656554)*dim
        ref[3,0]=(-0.0562176586729)*(1+2.*(dim-1.)/(o+1.))+(0.329476008516)*dim
        ref[3,1]=(0.0842497054449)*(1+2.*(dim-1.)/(o+1.))+(0.563341223973)*dim
        ref[3,2]=(0.230737024647)*(1+2.*(dim-1.)/(o+1.))+(-0.190226261737)*dim
        ref[3,3]=(1.19081442856)*(1+2.*(dim-1.)/(o+1.))+(0.079230750864)*dim
        ref[3,4]=(0.875682330522)*(1+2.*(dim-1.)/(o+1.))+(0.541976114772)*dim
      else:
        arg[0,0]=(0.0640108068236)*x[0]**o+(0.0322830166043)*x[0]+(-0.703691335255)*x[1]**o+(-0.185228770711)*x[1]+(-0.738714166313)*x[2]**o+(-0.423576647077)*x[2]
        arg[0,1]=(-0.738341351945)*x[0]**o+(0.396306395436)*x[0]+(-0.369974848568)*x[1]**o+(0.750933208783)*x[1]+(0.554069869127)*x[2]**o+(0.634275493051)*x[2]
        arg[0,2]=(-0.259193248708)*x[0]**o+(-0.900007449583)*x[0]+(0.894379436915)*x[1]**o+(0.997814320752)*x[1]+(0.233192116622)*x[2]**o+(0.986616259649)*x[2]
        arg[0,3]=(0.381480250594)*x[0]**o+(0.734823219057)*x[0]+(0.569912440169)*x[1]**o+(-0.441084504695)*x[1]+(0.312539701584)*x[2]**o+(-0.3733004974)*x[2]
        arg[0,4]=(-0.779059807622)*x[0]**o+(0.538777922322)*x[0]+(-0.0695879211612)*x[1]**o+(-0.204862216499)*x[1]+(-0.0109698525672)*x[2]**o+(0.981875117617)*x[2]
        arg[1,0]=(-0.415857630044)*x[0]**o+(-0.356607866913)*x[0]+(-0.628854326903)*x[1]**o+(0.0401233333505)*x[1]+(0.687470169579)*x[2]**o+(-0.565349753114)*x[2]
        arg[1,1]=(-0.438364910402)*x[0]**o+(0.498078194367)*x[0]+(-0.120729597948)*x[1]**o+(-0.622076117713)*x[1]+(0.995130029172)*x[2]**o+(-0.7234601064)*x[2]
        arg[1,2]=(-0.957869585558)*x[0]**o+(0.374693219824)*x[0]+(0.0355192120972)*x[1]**o+(-0.516770140651)*x[1]+(0.521887953791)*x[2]**o+(0.00771564236502)*x[2]
        arg[1,3]=(-0.0495580194062)*x[0]**o+(-0.183895326318)*x[0]+(-0.55304086303)*x[1]**o+(0.920204449879)*x[1]+(0.143165855157)*x[2]**o+(-0.773684620462)*x[2]
        arg[1,4]=(0.795913131087)*x[0]**o+(0.653826129199)*x[0]+(-0.585562760813)*x[1]**o+(0.274083108918)*x[1]+(-0.165275654116)*x[2]**o+(0.358678758212)*x[2]
        arg[2,0]=(-0.0699685857843)*x[0]**o+(-0.42922435931)*x[0]+(-0.0494308981938)*x[1]**o+(-0.956597829176)*x[1]+(-0.912941305307)*x[2]**o+(0.214777332562)*x[2]
        arg[2,1]=(-0.900983338173)*x[0]**o+(-0.779729884378)*x[0]+(-0.860100306508)*x[1]**o+(0.14211537281)*x[1]+(-0.984439938285)*x[2]**o+(-0.549987779842)*x[2]
        arg[2,2]=(0.801489512389)*x[0]**o+(0.896468923826)*x[0]+(-0.155737972887)*x[1]**o+(0.639864330477)*x[1]+(0.798598705038)*x[2]**o+(-0.863019041525)*x[2]
        arg[2,3]=(-0.453612136504)*x[0]**o+(0.522662491688)*x[0]+(-0.463058076486)*x[1]**o+(0.398018809499)*x[1]+(-0.612044961611)*x[2]**o+(-0.481951883217)*x[2]
        arg[2,4]=(-0.0739336519425)*x[0]**o+(-0.241753740154)*x[0]+(0.0176292058898)*x[1]**o+(0.435641132809)*x[1]+(-0.0592828584976)*x[2]**o+(0.430115620801)*x[2]
        arg[3,0]=(0.604784358152)*x[0]**o+(-0.817214179503)*x[0]+(-0.423104272008)*x[1]**o+(0.287787881773)*x[1]+(0.678036469241)*x[2]**o+(0.578347129563)*x[2]
        arg[3,1]=(0.362390778598)*x[0]**o+(-0.389066580119)*x[0]+(-0.934087580396)*x[1]**o+(0.907587485226)*x[1]+(-0.902083043406)*x[2]**o+(0.904707538314)*x[2]
        arg[3,2]=(0.862463207465)*x[0]**o+(0.904923219336)*x[0]+(0.421361696042)*x[1]**o+(-0.779329196171)*x[1]+(0.059521522394)*x[2]**o+(-0.744510067387)*x[2]
        arg[3,3]=(0.412029226851)*x[0]**o+(-0.649730143511)*x[0]+(0.579628017001)*x[1]**o+(-0.949180147129)*x[1]+(-0.073016492309)*x[2]**o+(0.666554170757)*x[2]
        arg[3,4]=(-0.700401396504)*x[0]**o+(-0.89194120812)*x[0]+(0.592387042213)*x[1]**o+(-0.906632787891)*x[1]+(-0.486072860812)*x[2]**o+(-0.505386176523)*x[2]
        ref[0,0]=(-1.37839469474)*(1+2.*(dim-1.)/(o+1.))+(-0.576522401184)*dim
        ref[0,1]=(-0.554246331386)*(1+2.*(dim-1.)/(o+1.))+(1.78151509727)*dim
        ref[0,2]=(0.868378304829)*(1+2.*(dim-1.)/(o+1.))+(1.08442313082)*dim
        ref[0,3]=(1.26393239235)*(1+2.*(dim-1.)/(o+1.))+(-0.0795617830385)*dim
        ref[0,4]=(-0.85961758135)*(1+2.*(dim-1.)/(o+1.))+(1.31579082344)*dim
        ref[1,0]=(-0.357241787368)*(1+2.*(dim-1.)/(o+1.))+(-0.881834286676)*dim
        ref[1,1]=(0.436035520821)*(1+2.*(dim-1.)/(o+1.))+(-0.847458029746)*dim
        ref[1,2]=(-0.40046241967)*(1+2.*(dim-1.)/(o+1.))+(-0.134361278462)*dim
        ref[1,3]=(-0.459433027279)*(1+2.*(dim-1.)/(o+1.))+(-0.0373754969011)*dim
        ref[1,4]=(0.0450747161574)*(1+2.*(dim-1.)/(o+1.))+(1.28658799633)*dim
        ref[2,0]=(-1.03234078929)*(1+2.*(dim-1.)/(o+1.))+(-1.17104485592)*dim
        ref[2,1]=(-2.74552358297)*(1+2.*(dim-1.)/(o+1.))+(-1.18760229141)*dim
        ref[2,2]=(1.44435024454)*(1+2.*(dim-1.)/(o+1.))+(0.673314212777)*dim
        ref[2,3]=(-1.5287151746)*(1+2.*(dim-1.)/(o+1.))+(0.43872941797)*dim
        ref[2,4]=(-0.11558730455)*(1+2.*(dim-1.)/(o+1.))+(0.624003013456)*dim
        ref[3,0]=(0.859716555385)*(1+2.*(dim-1.)/(o+1.))+(0.0489208318321)*dim
        ref[3,1]=(-1.4737798452)*(1+2.*(dim-1.)/(o+1.))+(1.42322844342)*dim
        ref[3,2]=(1.3433464259)*(1+2.*(dim-1.)/(o+1.))+(-0.618916044222)*dim
        ref[3,3]=(0.918640751542)*(1+2.*(dim-1.)/(o+1.))+(-0.932356119883)*dim
        ref[3,4]=(-0.594087215103)*(1+2.*(dim-1.)/(o+1.))+(-2.30396017253)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.662707930283)*x[0]**o+(0.625512077969)*x[0]+(-0.222208840922)*x[1]**o+(0.374614317804)*x[1]
        arg[0,0,1]=(0.685952252808)*x[0]**o+(-0.88219802797)*x[0]+(-0.948918831634)*x[1]**o+(-0.00615202896776)*x[1]
        arg[0,1,0]=(-0.555231206657)*x[0]**o+(0.997112425959)*x[0]+(-0.186247870511)*x[1]**o+(-0.0658892046229)*x[1]
        arg[0,1,1]=(-0.735288285178)*x[0]**o+(0.941234848614)*x[0]+(-0.307761737256)*x[1]**o+(0.933293069433)*x[1]
        arg[1,0,0]=(-0.618191337825)*x[0]**o+(-0.12930895166)*x[0]+(0.76213325083)*x[1]**o+(-0.744049403156)*x[1]
        arg[1,0,1]=(-0.793412570981)*x[0]**o+(0.533175323401)*x[0]+(-0.87398581932)*x[1]**o+(-0.279651754725)*x[1]
        arg[1,1,0]=(-0.0654077907694)*x[0]**o+(0.239040458488)*x[0]+(0.730639927837)*x[1]**o+(0.805020533329)*x[1]
        arg[1,1,1]=(-0.560195748542)*x[0]**o+(-0.265389430135)*x[0]+(-0.295142324741)*x[1]**o+(-0.929699771024)*x[1]
        arg[2,0,0]=(0.935760865577)*x[0]**o+(-0.332458843131)*x[0]+(0.257302695187)*x[1]**o+(-0.347547927033)*x[1]
        arg[2,0,1]=(-0.245770653276)*x[0]**o+(0.126945988409)*x[0]+(0.302793434995)*x[1]**o+(-0.997483200172)*x[1]
        arg[2,1,0]=(0.523494364951)*x[0]**o+(-0.323782740449)*x[0]+(-0.563752593454)*x[1]**o+(-0.358325752838)*x[1]
        arg[2,1,1]=(-0.506330114515)*x[0]**o+(-0.733234050567)*x[0]+(0.734774667778)*x[1]**o+(0.795264566776)*x[1]
        arg[3,0,0]=(-0.21258295235)*x[0]**o+(-0.929331330718)*x[0]+(-0.546061847742)*x[1]**o+(0.0613931491065)*x[1]
        arg[3,0,1]=(-0.357847946656)*x[0]**o+(-0.133795684508)*x[0]+(-0.796646028688)*x[1]**o+(0.875795801429)*x[1]
        arg[3,1,0]=(0.923717804039)*x[0]**o+(-0.176266633621)*x[0]+(-0.605892867534)*x[1]**o+(0.633377612365)*x[1]
        arg[3,1,1]=(0.729025508129)*x[0]**o+(0.412034955257)*x[0]+(0.0933054864485)*x[1]**o+(0.386019085532)*x[1]
        arg[4,0,0]=(-0.0705524918008)*x[0]**o+(-0.869605321304)*x[0]+(0.244351267755)*x[1]**o+(0.898883388102)*x[1]
        arg[4,0,1]=(-0.453297333327)*x[0]**o+(-0.67824185123)*x[0]+(0.0497322840245)*x[1]**o+(0.581677879069)*x[1]
        arg[4,1,0]=(0.343975006358)*x[0]**o+(0.724084357103)*x[0]+(-0.186292335362)*x[1]**o+(-0.257997075018)*x[1]
        arg[4,1,1]=(-0.389731923798)*x[0]**o+(0.0949155578616)*x[0]+(-0.298428138752)*x[1]**o+(0.819510069666)*x[1]
        arg[5,0,0]=(0.827670169513)*x[0]**o+(-0.584607978668)*x[0]+(0.669056948658)*x[1]**o+(-0.393158541985)*x[1]
        arg[5,0,1]=(-0.471551968206)*x[0]**o+(-0.339964831396)*x[0]+(0.354554393554)*x[1]**o+(-0.334773591957)*x[1]
        arg[5,1,0]=(0.113562904864)*x[0]**o+(-0.775045505123)*x[0]+(0.783044874931)*x[1]**o+(0.52265843873)*x[1]
        arg[5,1,1]=(-0.115476036205)*x[0]**o+(-0.290822012728)*x[0]+(-0.398510477854)*x[1]**o+(0.398737205542)*x[1]
        ref[0,0,0]=(0.440499089362)*(1+2.*(dim-1.)/(o+1.))+(1.00012639577)*dim
        ref[0,0,1]=(-0.262966578826)*(1+2.*(dim-1.)/(o+1.))+(-0.888350056938)*dim
        ref[0,1,0]=(-0.741479077167)*(1+2.*(dim-1.)/(o+1.))+(0.931223221336)*dim
        ref[0,1,1]=(-1.04305002243)*(1+2.*(dim-1.)/(o+1.))+(1.87452791805)*dim
        ref[1,0,0]=(0.143941913005)*(1+2.*(dim-1.)/(o+1.))+(-0.873358354816)*dim
        ref[1,0,1]=(-1.6673983903)*(1+2.*(dim-1.)/(o+1.))+(0.253523568676)*dim
        ref[1,1,0]=(0.665232137068)*(1+2.*(dim-1.)/(o+1.))+(1.04406099182)*dim
        ref[1,1,1]=(-0.855338073283)*(1+2.*(dim-1.)/(o+1.))+(-1.19508920116)*dim
        ref[2,0,0]=(1.19306356076)*(1+2.*(dim-1.)/(o+1.))+(-0.680006770164)*dim
        ref[2,0,1]=(0.0570227817187)*(1+2.*(dim-1.)/(o+1.))+(-0.870537211764)*dim
        ref[2,1,0]=(-0.0402582285028)*(1+2.*(dim-1.)/(o+1.))+(-0.682108493287)*dim
        ref[2,1,1]=(0.228444553263)*(1+2.*(dim-1.)/(o+1.))+(0.0620305162089)*dim
        ref[3,0,0]=(-0.758644800093)*(1+2.*(dim-1.)/(o+1.))+(-0.867938181611)*dim
        ref[3,0,1]=(-1.15449397534)*(1+2.*(dim-1.)/(o+1.))+(0.742000116921)*dim
        ref[3,1,0]=(0.317824936505)*(1+2.*(dim-1.)/(o+1.))+(0.457110978744)*dim
        ref[3,1,1]=(0.822330994578)*(1+2.*(dim-1.)/(o+1.))+(0.798054040789)*dim
        ref[4,0,0]=(0.173798775954)*(1+2.*(dim-1.)/(o+1.))+(0.0292780667983)*dim
        ref[4,0,1]=(-0.403565049303)*(1+2.*(dim-1.)/(o+1.))+(-0.0965639721601)*dim
        ref[4,1,0]=(0.157682670995)*(1+2.*(dim-1.)/(o+1.))+(0.466087282085)*dim
        ref[4,1,1]=(-0.68816006255)*(1+2.*(dim-1.)/(o+1.))+(0.914425627528)*dim
        ref[5,0,0]=(1.49672711817)*(1+2.*(dim-1.)/(o+1.))+(-0.977766520654)*dim
        ref[5,0,1]=(-0.116997574651)*(1+2.*(dim-1.)/(o+1.))+(-0.674738423353)*dim
        ref[5,1,0]=(0.896607779795)*(1+2.*(dim-1.)/(o+1.))+(-0.252387066394)*dim
        ref[5,1,1]=(-0.513986514058)*(1+2.*(dim-1.)/(o+1.))+(0.107915192814)*dim
      else:
        arg[0,0,0]=(-0.418671467289)*x[0]**o+(0.498387164484)*x[0]+(0.208896299314)*x[1]**o+(-0.643038678771)*x[1]+(-0.85475466439)*x[2]**o+(0.59586368009)*x[2]
        arg[0,0,1]=(0.149229891253)*x[0]**o+(0.229836799955)*x[0]+(0.557946253001)*x[1]**o+(-0.434223003739)*x[1]+(-0.250979831671)*x[2]**o+(0.762582320269)*x[2]
        arg[0,1,0]=(-0.950902166094)*x[0]**o+(-0.312913424444)*x[0]+(-0.573887973428)*x[1]**o+(-0.372485255069)*x[1]+(-0.18228472562)*x[2]**o+(-0.566679093436)*x[2]
        arg[0,1,1]=(-0.191769651257)*x[0]**o+(-0.344504306057)*x[0]+(0.355926399715)*x[1]**o+(-0.721341598375)*x[1]+(0.362401271289)*x[2]**o+(0.601351093654)*x[2]
        arg[1,0,0]=(-0.375572370375)*x[0]**o+(-0.532723139333)*x[0]+(-0.738166755424)*x[1]**o+(0.785506757134)*x[1]+(0.192667976423)*x[2]**o+(-0.06534893872)*x[2]
        arg[1,0,1]=(-0.353750712453)*x[0]**o+(0.401872866285)*x[0]+(-0.747783222746)*x[1]**o+(0.663116752574)*x[1]+(-0.162954382466)*x[2]**o+(0.318112656687)*x[2]
        arg[1,1,0]=(-0.711345828912)*x[0]**o+(0.764701438614)*x[0]+(-0.232576947703)*x[1]**o+(-0.50541791533)*x[1]+(0.604603682696)*x[2]**o+(0.498828792648)*x[2]
        arg[1,1,1]=(-0.72474879447)*x[0]**o+(0.888266999605)*x[0]+(0.639903282531)*x[1]**o+(-0.777435844156)*x[1]+(-0.0814738680964)*x[2]**o+(-0.0135286853237)*x[2]
        arg[2,0,0]=(0.839793308998)*x[0]**o+(0.374051014226)*x[0]+(0.594653957466)*x[1]**o+(-0.962352782345)*x[1]+(0.888632514336)*x[2]**o+(0.491195067464)*x[2]
        arg[2,0,1]=(-0.827615661598)*x[0]**o+(0.675227575488)*x[0]+(-0.93068580658)*x[1]**o+(0.694434337954)*x[1]+(0.3583179263)*x[2]**o+(0.775147111187)*x[2]
        arg[2,1,0]=(-0.772656999491)*x[0]**o+(-0.464476706435)*x[0]+(0.569832694977)*x[1]**o+(0.500517676542)*x[1]+(-0.272153062279)*x[2]**o+(-0.398436498506)*x[2]
        arg[2,1,1]=(-0.3065198839)*x[0]**o+(-0.454353346315)*x[0]+(-0.539448433869)*x[1]**o+(0.776950547078)*x[1]+(-0.987178147402)*x[2]**o+(0.368935549126)*x[2]
        arg[3,0,0]=(-0.516131622594)*x[0]**o+(-0.680922252714)*x[0]+(0.304951059501)*x[1]**o+(-0.236361364379)*x[1]+(-0.37983306233)*x[2]**o+(0.693426329316)*x[2]
        arg[3,0,1]=(0.85094983056)*x[0]**o+(-0.356782375867)*x[0]+(0.362116478047)*x[1]**o+(0.269126381543)*x[1]+(0.950254707197)*x[2]**o+(0.592028802301)*x[2]
        arg[3,1,0]=(-0.17721437307)*x[0]**o+(-0.127432965766)*x[0]+(-0.754975923735)*x[1]**o+(0.755938625221)*x[1]+(0.776393089777)*x[2]**o+(0.913028629874)*x[2]
        arg[3,1,1]=(-0.779840479711)*x[0]**o+(0.432392796542)*x[0]+(0.0457670482312)*x[1]**o+(0.493363120993)*x[1]+(-0.290308698356)*x[2]**o+(0.263565329052)*x[2]
        arg[4,0,0]=(-0.594371032991)*x[0]**o+(0.673573089348)*x[0]+(0.0301238273919)*x[1]**o+(-0.900766544194)*x[1]+(0.937125579182)*x[2]**o+(-0.360512287398)*x[2]
        arg[4,0,1]=(-0.187561420369)*x[0]**o+(-0.727058667406)*x[0]+(0.643904893485)*x[1]**o+(0.547358609515)*x[1]+(-0.810276689361)*x[2]**o+(0.347967967049)*x[2]
        arg[4,1,0]=(0.0261452776327)*x[0]**o+(0.630125453041)*x[0]+(0.884509651558)*x[1]**o+(-0.580251885597)*x[1]+(0.141608918047)*x[2]**o+(0.417140237181)*x[2]
        arg[4,1,1]=(0.943149763428)*x[0]**o+(0.0414421588123)*x[0]+(-0.453944232191)*x[1]**o+(0.817963360891)*x[1]+(-0.0132464046434)*x[2]**o+(-0.247024507218)*x[2]
        arg[5,0,0]=(0.993356103007)*x[0]**o+(-0.604991443085)*x[0]+(-0.0141942193536)*x[1]**o+(-0.467982818195)*x[1]+(0.907608332852)*x[2]**o+(-0.918494340721)*x[2]
        arg[5,0,1]=(0.782079295393)*x[0]**o+(0.0125642236587)*x[0]+(-0.512764292821)*x[1]**o+(-0.0699705327379)*x[1]+(0.141341885735)*x[2]**o+(0.420508183565)*x[2]
        arg[5,1,0]=(-0.835665486427)*x[0]**o+(-0.536850925449)*x[0]+(-0.0958499285885)*x[1]**o+(-0.541119926637)*x[1]+(0.341886033336)*x[2]**o+(-0.772823771771)*x[2]
        arg[5,1,1]=(-0.391737530391)*x[0]**o+(-0.160045087084)*x[0]+(-0.984350699381)*x[1]**o+(0.303782477449)*x[1]+(-0.279906615513)*x[2]**o+(-0.690872289955)*x[2]
        ref[0,0,0]=(-1.06452983237)*(1+2.*(dim-1.)/(o+1.))+(0.451212165803)*dim
        ref[0,0,1]=(0.456196312583)*(1+2.*(dim-1.)/(o+1.))+(0.558196116486)*dim
        ref[0,1,0]=(-1.70707486514)*(1+2.*(dim-1.)/(o+1.))+(-1.25207777295)*dim
        ref[0,1,1]=(0.526558019747)*(1+2.*(dim-1.)/(o+1.))+(-0.464494810779)*dim
        ref[1,0,0]=(-0.921071149376)*(1+2.*(dim-1.)/(o+1.))+(0.187434679081)*dim
        ref[1,0,1]=(-1.26448831766)*(1+2.*(dim-1.)/(o+1.))+(1.38310227555)*dim
        ref[1,1,0]=(-0.339319093919)*(1+2.*(dim-1.)/(o+1.))+(0.758112315932)*dim
        ref[1,1,1]=(-0.166319380035)*(1+2.*(dim-1.)/(o+1.))+(0.0973024701251)*dim
        ref[2,0,0]=(2.3230797808)*(1+2.*(dim-1.)/(o+1.))+(-0.0971067006546)*dim
        ref[2,0,1]=(-1.39998354188)*(1+2.*(dim-1.)/(o+1.))+(2.14480902463)*dim
        ref[2,1,0]=(-0.474977366793)*(1+2.*(dim-1.)/(o+1.))+(-0.362395528399)*dim
        ref[2,1,1]=(-1.83314646517)*(1+2.*(dim-1.)/(o+1.))+(0.69153274989)*dim
        ref[3,0,0]=(-0.591013625423)*(1+2.*(dim-1.)/(o+1.))+(-0.223857287777)*dim
        ref[3,0,1]=(2.1633210158)*(1+2.*(dim-1.)/(o+1.))+(0.504372807977)*dim
        ref[3,1,0]=(-0.155797207029)*(1+2.*(dim-1.)/(o+1.))+(1.54153428933)*dim
        ref[3,1,1]=(-1.02438212984)*(1+2.*(dim-1.)/(o+1.))+(1.18932124659)*dim
        ref[4,0,0]=(0.372878373583)*(1+2.*(dim-1.)/(o+1.))+(-0.587705742243)*dim
        ref[4,0,1]=(-0.353933216245)*(1+2.*(dim-1.)/(o+1.))+(0.168267909158)*dim
        ref[4,1,0]=(1.05226384724)*(1+2.*(dim-1.)/(o+1.))+(0.467013804625)*dim
        ref[4,1,1]=(0.475959126593)*(1+2.*(dim-1.)/(o+1.))+(0.612381012486)*dim
        ref[5,0,0]=(1.88677021651)*(1+2.*(dim-1.)/(o+1.))+(-1.991468602)*dim
        ref[5,0,1]=(0.410656888307)*(1+2.*(dim-1.)/(o+1.))+(0.363101874486)*dim
        ref[5,1,0]=(-0.58962938168)*(1+2.*(dim-1.)/(o+1.))+(-1.85079462386)*dim
        ref[5,1,1]=(-1.65599484529)*(1+2.*(dim-1.)/(o+1.))+(-0.54713489959)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.606358724075)*x[0]**o+(-0.6326664687)*x[0]+(-0.521101256804)*x[1]**o+(0.702113497865)*x[1]
        arg[0,0,0,1]=(0.528315511137)*x[0]**o+(-0.116566989978)*x[0]+(-0.350847384045)*x[1]**o+(-0.581988878345)*x[1]
        arg[0,0,1,0]=(0.565076347032)*x[0]**o+(0.752532782601)*x[0]+(0.00706163321089)*x[1]**o+(-0.89265244135)*x[1]
        arg[0,0,1,1]=(-0.342136497951)*x[0]**o+(0.753153993866)*x[0]+(-0.199643944066)*x[1]**o+(0.995411752312)*x[1]
        arg[0,0,2,0]=(0.621283193015)*x[0]**o+(0.591186318333)*x[0]+(0.880169584588)*x[1]**o+(-0.11839698224)*x[1]
        arg[0,0,2,1]=(-0.831745199475)*x[0]**o+(0.293656166596)*x[0]+(0.0717745661823)*x[1]**o+(0.827680695965)*x[1]
        arg[0,1,0,0]=(0.686864676707)*x[0]**o+(-0.6552976723)*x[0]+(-0.65589888943)*x[1]**o+(-0.878054891784)*x[1]
        arg[0,1,0,1]=(-0.43359086203)*x[0]**o+(0.723819586663)*x[0]+(0.0741799035834)*x[1]**o+(0.13005305188)*x[1]
        arg[0,1,1,0]=(-0.0605346127445)*x[0]**o+(-0.717729061173)*x[0]+(0.825539660518)*x[1]**o+(0.873033938391)*x[1]
        arg[0,1,1,1]=(-0.286483607332)*x[0]**o+(-0.577255191498)*x[0]+(-0.328362280683)*x[1]**o+(0.499078653745)*x[1]
        arg[0,1,2,0]=(-0.334132932545)*x[0]**o+(0.462461110225)*x[0]+(-0.932311565873)*x[1]**o+(0.934683608666)*x[1]
        arg[0,1,2,1]=(-0.383203086272)*x[0]**o+(0.261035781624)*x[0]+(-0.203936125231)*x[1]**o+(0.657584310755)*x[1]
        arg[0,2,0,0]=(-0.637642146695)*x[0]**o+(0.80524493224)*x[0]+(-0.139137135459)*x[1]**o+(0.947434162736)*x[1]
        arg[0,2,0,1]=(-0.0738996477961)*x[0]**o+(0.0663360260869)*x[0]+(-0.599113718918)*x[1]**o+(-0.349514990363)*x[1]
        arg[0,2,1,0]=(-0.0539211250603)*x[0]**o+(0.262567947534)*x[0]+(0.115386681137)*x[1]**o+(-0.68179682822)*x[1]
        arg[0,2,1,1]=(-0.770114869627)*x[0]**o+(-0.388068404619)*x[0]+(-0.712914716647)*x[1]**o+(-0.810536809871)*x[1]
        arg[0,2,2,0]=(0.274425895455)*x[0]**o+(0.338467327995)*x[0]+(-0.78223636363)*x[1]**o+(0.406149090351)*x[1]
        arg[0,2,2,1]=(-0.748672361572)*x[0]**o+(0.508311939723)*x[0]+(0.875797107098)*x[1]**o+(0.0992795348726)*x[1]
        arg[0,3,0,0]=(-0.547350688209)*x[0]**o+(0.671683747412)*x[0]+(-0.0954212965969)*x[1]**o+(0.983487887304)*x[1]
        arg[0,3,0,1]=(0.626104338883)*x[0]**o+(-0.825311449834)*x[0]+(-0.422521614246)*x[1]**o+(0.304208397592)*x[1]
        arg[0,3,1,0]=(0.698777992012)*x[0]**o+(0.00291670095874)*x[0]+(0.327527809646)*x[1]**o+(0.283289333886)*x[1]
        arg[0,3,1,1]=(-0.0197648150597)*x[0]**o+(-0.902860814776)*x[0]+(0.663961228673)*x[1]**o+(0.675952386365)*x[1]
        arg[0,3,2,0]=(0.286747814345)*x[0]**o+(-0.0641946364041)*x[0]+(0.812019228115)*x[1]**o+(-0.225698609001)*x[1]
        arg[0,3,2,1]=(0.879776322226)*x[0]**o+(-0.251856143047)*x[0]+(-0.252403834757)*x[1]**o+(0.650790406323)*x[1]
        arg[0,4,0,0]=(-0.58782572388)*x[0]**o+(-0.144122228965)*x[0]+(0.521707348991)*x[1]**o+(-0.0460840614459)*x[1]
        arg[0,4,0,1]=(-0.777497837497)*x[0]**o+(-0.229011563089)*x[0]+(0.754657080443)*x[1]**o+(-0.690675521089)*x[1]
        arg[0,4,1,0]=(0.351528557097)*x[0]**o+(0.00192986530026)*x[0]+(-0.414849484269)*x[1]**o+(-0.250205659505)*x[1]
        arg[0,4,1,1]=(-0.814037796843)*x[0]**o+(0.00188512986627)*x[0]+(-0.810799710998)*x[1]**o+(-0.114406174711)*x[1]
        arg[0,4,2,0]=(-0.755511212878)*x[0]**o+(-0.502206951212)*x[0]+(0.555738333365)*x[1]**o+(-0.220359547926)*x[1]
        arg[0,4,2,1]=(0.247108894345)*x[0]**o+(0.479239235148)*x[0]+(0.663958827816)*x[1]**o+(0.220938477625)*x[1]
        arg[1,0,0,0]=(0.818774052595)*x[0]**o+(-0.651584728881)*x[0]+(-0.780674092348)*x[1]**o+(-0.397277502644)*x[1]
        arg[1,0,0,1]=(0.482101508249)*x[0]**o+(-0.523618202277)*x[0]+(0.806406836488)*x[1]**o+(-0.883738297664)*x[1]
        arg[1,0,1,0]=(0.272517667346)*x[0]**o+(-0.700461889638)*x[0]+(0.280736354342)*x[1]**o+(0.182983247089)*x[1]
        arg[1,0,1,1]=(0.566589111451)*x[0]**o+(-0.876744547538)*x[0]+(-0.934741703566)*x[1]**o+(-0.562770667492)*x[1]
        arg[1,0,2,0]=(-0.954128714623)*x[0]**o+(0.115699770127)*x[0]+(-0.580663262426)*x[1]**o+(-0.451619694161)*x[1]
        arg[1,0,2,1]=(0.268605150185)*x[0]**o+(-0.233415174179)*x[0]+(0.477143801066)*x[1]**o+(0.443696994244)*x[1]
        arg[1,1,0,0]=(-0.00240846686485)*x[0]**o+(0.776336851643)*x[0]+(0.488314964094)*x[1]**o+(-0.840196476788)*x[1]
        arg[1,1,0,1]=(-0.511732155638)*x[0]**o+(0.528782039485)*x[0]+(0.51274502262)*x[1]**o+(-0.700213610705)*x[1]
        arg[1,1,1,0]=(0.300359000499)*x[0]**o+(0.405101876401)*x[0]+(-0.47022583299)*x[1]**o+(-0.977634916267)*x[1]
        arg[1,1,1,1]=(-0.494482318948)*x[0]**o+(0.254112728167)*x[0]+(-0.541672595343)*x[1]**o+(-0.270970064522)*x[1]
        arg[1,1,2,0]=(-0.586624294233)*x[0]**o+(0.629601920904)*x[0]+(-0.525092252936)*x[1]**o+(-0.44611635978)*x[1]
        arg[1,1,2,1]=(0.716927085139)*x[0]**o+(0.700193925267)*x[0]+(0.0624921446647)*x[1]**o+(0.109485129877)*x[1]
        arg[1,2,0,0]=(0.325729307898)*x[0]**o+(0.773432634097)*x[0]+(-0.675612597356)*x[1]**o+(-0.0554913733692)*x[1]
        arg[1,2,0,1]=(-0.381606382262)*x[0]**o+(-0.689040382854)*x[0]+(-0.580191973398)*x[1]**o+(0.951689147387)*x[1]
        arg[1,2,1,0]=(-0.527642403744)*x[0]**o+(-0.936915203927)*x[0]+(0.631733997289)*x[1]**o+(-0.715975680015)*x[1]
        arg[1,2,1,1]=(0.0886898373323)*x[0]**o+(-0.0943320873915)*x[0]+(0.470668750443)*x[1]**o+(0.773832960093)*x[1]
        arg[1,2,2,0]=(-0.727720290197)*x[0]**o+(0.422119803173)*x[0]+(0.705882951779)*x[1]**o+(-0.350801865244)*x[1]
        arg[1,2,2,1]=(0.879335475407)*x[0]**o+(-0.0307168471577)*x[0]+(0.458578151544)*x[1]**o+(0.788796576339)*x[1]
        arg[1,3,0,0]=(0.182143508146)*x[0]**o+(-0.680787176288)*x[0]+(0.626099489145)*x[1]**o+(-0.771540378408)*x[1]
        arg[1,3,0,1]=(-0.578341681282)*x[0]**o+(0.856407179963)*x[0]+(0.525319425356)*x[1]**o+(0.345651851179)*x[1]
        arg[1,3,1,0]=(-0.316652838058)*x[0]**o+(0.652409898929)*x[0]+(-0.82123442746)*x[1]**o+(-0.136868558081)*x[1]
        arg[1,3,1,1]=(-0.755530700082)*x[0]**o+(0.354429731374)*x[0]+(-0.14330437792)*x[1]**o+(-0.221454495186)*x[1]
        arg[1,3,2,0]=(-0.520261483769)*x[0]**o+(0.0873707179205)*x[0]+(-0.882995356682)*x[1]**o+(-0.773921687804)*x[1]
        arg[1,3,2,1]=(0.404278952718)*x[0]**o+(0.651166017764)*x[0]+(-0.496695253747)*x[1]**o+(0.625671754182)*x[1]
        arg[1,4,0,0]=(-0.730831539043)*x[0]**o+(0.219467989041)*x[0]+(-0.853784008431)*x[1]**o+(0.274799624897)*x[1]
        arg[1,4,0,1]=(0.893313801855)*x[0]**o+(-0.539687651773)*x[0]+(-0.411582029239)*x[1]**o+(-0.293521483805)*x[1]
        arg[1,4,1,0]=(0.484878049594)*x[0]**o+(0.218930293134)*x[0]+(-0.138362733141)*x[1]**o+(-0.663241354177)*x[1]
        arg[1,4,1,1]=(-0.608410207857)*x[0]**o+(-0.0921006917331)*x[0]+(-0.885705130565)*x[1]**o+(0.103724764404)*x[1]
        arg[1,4,2,0]=(0.150946260009)*x[0]**o+(-0.601834829203)*x[0]+(0.0141091402169)*x[1]**o+(-0.98316187826)*x[1]
        arg[1,4,2,1]=(0.98764918249)*x[0]**o+(-0.841155365687)*x[0]+(0.185525084432)*x[1]**o+(-0.613421841538)*x[1]
        arg[2,0,0,0]=(0.485337970037)*x[0]**o+(-0.898602108837)*x[0]+(0.847812629133)*x[1]**o+(-0.23522051021)*x[1]
        arg[2,0,0,1]=(0.835641029366)*x[0]**o+(-0.0524661277338)*x[0]+(0.41697147238)*x[1]**o+(-0.622813671618)*x[1]
        arg[2,0,1,0]=(0.505181088966)*x[0]**o+(-0.870141755006)*x[0]+(0.451139050304)*x[1]**o+(0.957790881903)*x[1]
        arg[2,0,1,1]=(0.822955028051)*x[0]**o+(-0.103028201548)*x[0]+(-0.604614127349)*x[1]**o+(-0.655334670762)*x[1]
        arg[2,0,2,0]=(-0.166647802545)*x[0]**o+(-0.466885186725)*x[0]+(0.0782333081578)*x[1]**o+(0.814414769115)*x[1]
        arg[2,0,2,1]=(-0.980310280968)*x[0]**o+(-0.884306329761)*x[0]+(-0.895550319566)*x[1]**o+(0.160657671233)*x[1]
        arg[2,1,0,0]=(-0.846405435483)*x[0]**o+(-0.625465093616)*x[0]+(-0.0789995975268)*x[1]**o+(-0.440180761086)*x[1]
        arg[2,1,0,1]=(-0.949855662419)*x[0]**o+(-0.013707742372)*x[0]+(0.0397729185783)*x[1]**o+(-0.836054083952)*x[1]
        arg[2,1,1,0]=(0.427246928002)*x[0]**o+(0.835369889909)*x[0]+(0.925300799558)*x[1]**o+(-0.328266770631)*x[1]
        arg[2,1,1,1]=(-0.36143365614)*x[0]**o+(-0.744107540819)*x[0]+(-0.358554131639)*x[1]**o+(-0.562674543946)*x[1]
        arg[2,1,2,0]=(0.60580594275)*x[0]**o+(0.644796993923)*x[0]+(-0.682407021927)*x[1]**o+(0.938016908061)*x[1]
        arg[2,1,2,1]=(-0.934393274317)*x[0]**o+(-0.718420272664)*x[0]+(0.0357039032065)*x[1]**o+(0.446387943344)*x[1]
        arg[2,2,0,0]=(-0.329105411737)*x[0]**o+(0.588935500072)*x[0]+(-0.833362483432)*x[1]**o+(0.629996961074)*x[1]
        arg[2,2,0,1]=(0.723977530226)*x[0]**o+(0.0698430385616)*x[0]+(-0.503761809156)*x[1]**o+(0.907472279839)*x[1]
        arg[2,2,1,0]=(-0.767545510793)*x[0]**o+(0.615094783682)*x[0]+(0.894594933703)*x[1]**o+(0.922357219049)*x[1]
        arg[2,2,1,1]=(-0.155739055576)*x[0]**o+(0.646142299486)*x[0]+(0.865932681534)*x[1]**o+(-0.126887380871)*x[1]
        arg[2,2,2,0]=(-0.629534320112)*x[0]**o+(-0.0904366301836)*x[0]+(0.193019806731)*x[1]**o+(-0.866080160466)*x[1]
        arg[2,2,2,1]=(-0.714527761053)*x[0]**o+(-0.813712879405)*x[0]+(-0.291734238438)*x[1]**o+(0.232675536415)*x[1]
        arg[2,3,0,0]=(-0.447474292879)*x[0]**o+(-0.639046887788)*x[0]+(0.159616683647)*x[1]**o+(0.564127141623)*x[1]
        arg[2,3,0,1]=(-0.608953936619)*x[0]**o+(0.273179484103)*x[0]+(0.42693156927)*x[1]**o+(-0.983331075647)*x[1]
        arg[2,3,1,0]=(0.855859483301)*x[0]**o+(-0.432495503925)*x[0]+(-0.67928203769)*x[1]**o+(-0.856541006702)*x[1]
        arg[2,3,1,1]=(-0.581185594204)*x[0]**o+(0.947720323693)*x[0]+(-0.520212852628)*x[1]**o+(-0.37563221594)*x[1]
        arg[2,3,2,0]=(0.351358492535)*x[0]**o+(0.367210525846)*x[0]+(0.97870179383)*x[1]**o+(-0.199559052968)*x[1]
        arg[2,3,2,1]=(0.797294518541)*x[0]**o+(0.546397721552)*x[0]+(0.374121851646)*x[1]**o+(0.704436689798)*x[1]
        arg[2,4,0,0]=(-0.756047775149)*x[0]**o+(-0.705403684825)*x[0]+(-0.662421213376)*x[1]**o+(0.390069314234)*x[1]
        arg[2,4,0,1]=(-0.658842889869)*x[0]**o+(0.766273769822)*x[0]+(-0.779201775962)*x[1]**o+(-0.968670270742)*x[1]
        arg[2,4,1,0]=(-0.336220564628)*x[0]**o+(-0.134745604319)*x[0]+(0.391939888428)*x[1]**o+(-0.113113999792)*x[1]
        arg[2,4,1,1]=(-0.8345363851)*x[0]**o+(0.339223929857)*x[0]+(0.628403353717)*x[1]**o+(0.799364167182)*x[1]
        arg[2,4,2,0]=(-0.898050446701)*x[0]**o+(-0.666373288962)*x[0]+(-0.620778021659)*x[1]**o+(-0.990359850049)*x[1]
        arg[2,4,2,1]=(-0.933931027675)*x[0]**o+(-0.119077281033)*x[0]+(0.557313533556)*x[1]**o+(0.160336158304)*x[1]
        arg[3,0,0,0]=(0.972462636314)*x[0]**o+(0.607913740671)*x[0]+(0.490168539257)*x[1]**o+(-0.12515921391)*x[1]
        arg[3,0,0,1]=(0.856370886333)*x[0]**o+(-0.143225009869)*x[0]+(-0.330381845603)*x[1]**o+(-0.271958059018)*x[1]
        arg[3,0,1,0]=(0.86742129112)*x[0]**o+(-0.0259756341823)*x[0]+(-0.0137047849942)*x[1]**o+(-0.0165444161044)*x[1]
        arg[3,0,1,1]=(-0.445996036941)*x[0]**o+(0.247936136111)*x[0]+(-0.381484740124)*x[1]**o+(0.836678417006)*x[1]
        arg[3,0,2,0]=(0.161387456771)*x[0]**o+(0.687533753311)*x[0]+(-0.463845417603)*x[1]**o+(0.665023823118)*x[1]
        arg[3,0,2,1]=(0.243410612742)*x[0]**o+(-0.52431675734)*x[0]+(0.75630192488)*x[1]**o+(-0.703601557317)*x[1]
        arg[3,1,0,0]=(0.793530689056)*x[0]**o+(0.814806726176)*x[0]+(0.582051068232)*x[1]**o+(0.692432923027)*x[1]
        arg[3,1,0,1]=(-0.283506909465)*x[0]**o+(0.21438128538)*x[0]+(-0.327476122861)*x[1]**o+(0.990178870269)*x[1]
        arg[3,1,1,0]=(-0.117386491827)*x[0]**o+(0.747771272481)*x[0]+(-0.439767754004)*x[1]**o+(-0.905218338639)*x[1]
        arg[3,1,1,1]=(-0.544805647706)*x[0]**o+(-0.0484033735345)*x[0]+(0.861443439156)*x[1]**o+(-0.992086899129)*x[1]
        arg[3,1,2,0]=(0.596208599112)*x[0]**o+(-0.276136075021)*x[0]+(0.079604120535)*x[1]**o+(0.124232758054)*x[1]
        arg[3,1,2,1]=(0.809847641841)*x[0]**o+(0.956041370087)*x[0]+(0.65224420189)*x[1]**o+(0.565327743771)*x[1]
        arg[3,2,0,0]=(-0.264599479107)*x[0]**o+(-0.276726556903)*x[0]+(0.577911819561)*x[1]**o+(-0.656157381756)*x[1]
        arg[3,2,0,1]=(-0.0636193103458)*x[0]**o+(0.941482878452)*x[0]+(0.877581030494)*x[1]**o+(-0.254911385455)*x[1]
        arg[3,2,1,0]=(-0.28083673633)*x[0]**o+(0.849095032935)*x[0]+(0.6825284068)*x[1]**o+(0.445605171749)*x[1]
        arg[3,2,1,1]=(-0.729331034619)*x[0]**o+(0.036400895394)*x[0]+(0.365733359214)*x[1]**o+(0.0305324156609)*x[1]
        arg[3,2,2,0]=(-0.70932226428)*x[0]**o+(0.310455571562)*x[0]+(-0.101305643894)*x[1]**o+(0.346873071627)*x[1]
        arg[3,2,2,1]=(0.712053329488)*x[0]**o+(-0.845042122758)*x[0]+(0.909517101069)*x[1]**o+(-0.163109058477)*x[1]
        arg[3,3,0,0]=(0.451744730688)*x[0]**o+(0.308756227526)*x[0]+(0.13963307493)*x[1]**o+(-0.945789640023)*x[1]
        arg[3,3,0,1]=(-0.852609785834)*x[0]**o+(-0.821513824231)*x[0]+(-0.544690113726)*x[1]**o+(0.335547843106)*x[1]
        arg[3,3,1,0]=(0.337600834432)*x[0]**o+(-0.170181468776)*x[0]+(-0.302415145845)*x[1]**o+(0.985425450212)*x[1]
        arg[3,3,1,1]=(-0.638242591048)*x[0]**o+(-0.0221442303187)*x[0]+(0.421311946694)*x[1]**o+(-0.43495081316)*x[1]
        arg[3,3,2,0]=(-0.734168622382)*x[0]**o+(-0.889300572069)*x[0]+(0.808161580285)*x[1]**o+(-0.526348332439)*x[1]
        arg[3,3,2,1]=(0.998918324168)*x[0]**o+(-0.953729666152)*x[0]+(0.665101882861)*x[1]**o+(-0.581118145725)*x[1]
        arg[3,4,0,0]=(-0.251226888098)*x[0]**o+(-0.711615999514)*x[0]+(0.291022154479)*x[1]**o+(0.0772744442604)*x[1]
        arg[3,4,0,1]=(0.162379537091)*x[0]**o+(0.363750962136)*x[0]+(0.635943839083)*x[1]**o+(0.513856953389)*x[1]
        arg[3,4,1,0]=(-0.99052354199)*x[0]**o+(0.289202668813)*x[0]+(-0.759078503012)*x[1]**o+(-0.588801682903)*x[1]
        arg[3,4,1,1]=(-0.942278699306)*x[0]**o+(0.476003512559)*x[0]+(0.873920293292)*x[1]**o+(0.867413991424)*x[1]
        arg[3,4,2,0]=(-0.602386702281)*x[0]**o+(0.694432164577)*x[0]+(0.367008950869)*x[1]**o+(-0.803331549449)*x[1]
        arg[3,4,2,1]=(-0.00204465287104)*x[0]**o+(-0.697572466685)*x[0]+(-0.84755978799)*x[1]**o+(-0.758946055726)*x[1]
        ref[0,0,0,0]=(0.0852574672711)*(1+2.*(dim-1.)/(o+1.))+(0.069447029165)*dim
        ref[0,0,0,1]=(0.177468127092)*(1+2.*(dim-1.)/(o+1.))+(-0.698555868323)*dim
        ref[0,0,1,0]=(0.572137980243)*(1+2.*(dim-1.)/(o+1.))+(-0.140119658749)*dim
        ref[0,0,1,1]=(-0.541780442017)*(1+2.*(dim-1.)/(o+1.))+(1.74856574618)*dim
        ref[0,0,2,0]=(1.5014527776)*(1+2.*(dim-1.)/(o+1.))+(0.472789336092)*dim
        ref[0,0,2,1]=(-0.759970633293)*(1+2.*(dim-1.)/(o+1.))+(1.12133686256)*dim
        ref[0,1,0,0]=(0.030965787278)*(1+2.*(dim-1.)/(o+1.))+(-1.53335256408)*dim
        ref[0,1,0,1]=(-0.359410958446)*(1+2.*(dim-1.)/(o+1.))+(0.853872638544)*dim
        ref[0,1,1,0]=(0.765005047774)*(1+2.*(dim-1.)/(o+1.))+(0.155304877218)*dim
        ref[0,1,1,1]=(-0.614845888015)*(1+2.*(dim-1.)/(o+1.))+(-0.0781765377528)*dim
        ref[0,1,2,0]=(-1.26644449842)*(1+2.*(dim-1.)/(o+1.))+(1.39714471889)*dim
        ref[0,1,2,1]=(-0.587139211502)*(1+2.*(dim-1.)/(o+1.))+(0.918620092379)*dim
        ref[0,2,0,0]=(-0.776779282155)*(1+2.*(dim-1.)/(o+1.))+(1.75267909498)*dim
        ref[0,2,0,1]=(-0.673013366714)*(1+2.*(dim-1.)/(o+1.))+(-0.283178964276)*dim
        ref[0,2,1,0]=(0.0614655560763)*(1+2.*(dim-1.)/(o+1.))+(-0.419228880686)*dim
        ref[0,2,1,1]=(-1.48302958627)*(1+2.*(dim-1.)/(o+1.))+(-1.19860521449)*dim
        ref[0,2,2,0]=(-0.507810468175)*(1+2.*(dim-1.)/(o+1.))+(0.744616418346)*dim
        ref[0,2,2,1]=(0.127124745526)*(1+2.*(dim-1.)/(o+1.))+(0.607591474596)*dim
        ref[0,3,0,0]=(-0.642771984806)*(1+2.*(dim-1.)/(o+1.))+(1.65517163472)*dim
        ref[0,3,0,1]=(0.203582724637)*(1+2.*(dim-1.)/(o+1.))+(-0.521103052242)*dim
        ref[0,3,1,0]=(1.02630580166)*(1+2.*(dim-1.)/(o+1.))+(0.286206034845)*dim
        ref[0,3,1,1]=(0.644196413613)*(1+2.*(dim-1.)/(o+1.))+(-0.226908428411)*dim
        ref[0,3,2,0]=(1.09876704246)*(1+2.*(dim-1.)/(o+1.))+(-0.289893245405)*dim
        ref[0,3,2,1]=(0.627372487469)*(1+2.*(dim-1.)/(o+1.))+(0.398934263276)*dim
        ref[0,4,0,0]=(-0.0661183748894)*(1+2.*(dim-1.)/(o+1.))+(-0.190206290411)*dim
        ref[0,4,0,1]=(-0.022840757054)*(1+2.*(dim-1.)/(o+1.))+(-0.919687084178)*dim
        ref[0,4,1,0]=(-0.063320927172)*(1+2.*(dim-1.)/(o+1.))+(-0.248275794205)*dim
        ref[0,4,1,1]=(-1.62483750784)*(1+2.*(dim-1.)/(o+1.))+(-0.112521044845)*dim
        ref[0,4,2,0]=(-0.199772879513)*(1+2.*(dim-1.)/(o+1.))+(-0.722566499138)*dim
        ref[0,4,2,1]=(0.911067722161)*(1+2.*(dim-1.)/(o+1.))+(0.700177712774)*dim
        ref[1,0,0,0]=(0.0380999602468)*(1+2.*(dim-1.)/(o+1.))+(-1.04886223152)*dim
        ref[1,0,0,1]=(1.28850834474)*(1+2.*(dim-1.)/(o+1.))+(-1.40735649994)*dim
        ref[1,0,1,0]=(0.553254021688)*(1+2.*(dim-1.)/(o+1.))+(-0.517478642549)*dim
        ref[1,0,1,1]=(-0.368152592115)*(1+2.*(dim-1.)/(o+1.))+(-1.43951521503)*dim
        ref[1,0,2,0]=(-1.53479197705)*(1+2.*(dim-1.)/(o+1.))+(-0.335919924034)*dim
        ref[1,0,2,1]=(0.745748951251)*(1+2.*(dim-1.)/(o+1.))+(0.210281820065)*dim
        ref[1,1,0,0]=(0.485906497229)*(1+2.*(dim-1.)/(o+1.))+(-0.0638596251453)*dim
        ref[1,1,0,1]=(0.00101286698233)*(1+2.*(dim-1.)/(o+1.))+(-0.17143157122)*dim
        ref[1,1,1,0]=(-0.169866832491)*(1+2.*(dim-1.)/(o+1.))+(-0.572533039866)*dim
        ref[1,1,1,1]=(-1.03615491429)*(1+2.*(dim-1.)/(o+1.))+(-0.016857336355)*dim
        ref[1,1,2,0]=(-1.11171654717)*(1+2.*(dim-1.)/(o+1.))+(0.183485561125)*dim
        ref[1,1,2,1]=(0.779419229804)*(1+2.*(dim-1.)/(o+1.))+(0.809679055143)*dim
        ref[1,2,0,0]=(-0.349883289458)*(1+2.*(dim-1.)/(o+1.))+(0.717941260728)*dim
        ref[1,2,0,1]=(-0.96179835566)*(1+2.*(dim-1.)/(o+1.))+(0.262648764533)*dim
        ref[1,2,1,0]=(0.104091593546)*(1+2.*(dim-1.)/(o+1.))+(-1.65289088394)*dim
        ref[1,2,1,1]=(0.559358587775)*(1+2.*(dim-1.)/(o+1.))+(0.679500872701)*dim
        ref[1,2,2,0]=(-0.0218373384179)*(1+2.*(dim-1.)/(o+1.))+(0.0713179379289)*dim
        ref[1,2,2,1]=(1.33791362695)*(1+2.*(dim-1.)/(o+1.))+(0.758079729181)*dim
        ref[1,3,0,0]=(0.808242997292)*(1+2.*(dim-1.)/(o+1.))+(-1.4523275547)*dim
        ref[1,3,0,1]=(-0.0530222559262)*(1+2.*(dim-1.)/(o+1.))+(1.20205903114)*dim
        ref[1,3,1,0]=(-1.13788726552)*(1+2.*(dim-1.)/(o+1.))+(0.515541340848)*dim
        ref[1,3,1,1]=(-0.898835078002)*(1+2.*(dim-1.)/(o+1.))+(0.132975236188)*dim
        ref[1,3,2,0]=(-1.40325684045)*(1+2.*(dim-1.)/(o+1.))+(-0.686550969884)*dim
        ref[1,3,2,1]=(-0.0924163010291)*(1+2.*(dim-1.)/(o+1.))+(1.27683777195)*dim
        ref[1,4,0,0]=(-1.58461554747)*(1+2.*(dim-1.)/(o+1.))+(0.494267613937)*dim
        ref[1,4,0,1]=(0.481731772616)*(1+2.*(dim-1.)/(o+1.))+(-0.833209135578)*dim
        ref[1,4,1,0]=(0.346515316453)*(1+2.*(dim-1.)/(o+1.))+(-0.444311061043)*dim
        ref[1,4,1,1]=(-1.49411533842)*(1+2.*(dim-1.)/(o+1.))+(0.0116240726713)*dim
        ref[1,4,2,0]=(0.165055400226)*(1+2.*(dim-1.)/(o+1.))+(-1.58499670746)*dim
        ref[1,4,2,1]=(1.17317426692)*(1+2.*(dim-1.)/(o+1.))+(-1.45457720722)*dim
        ref[2,0,0,0]=(1.33315059917)*(1+2.*(dim-1.)/(o+1.))+(-1.13382261905)*dim
        ref[2,0,0,1]=(1.25261250175)*(1+2.*(dim-1.)/(o+1.))+(-0.675279799352)*dim
        ref[2,0,1,0]=(0.95632013927)*(1+2.*(dim-1.)/(o+1.))+(0.0876491268976)*dim
        ref[2,0,1,1]=(0.218340900702)*(1+2.*(dim-1.)/(o+1.))+(-0.75836287231)*dim
        ref[2,0,2,0]=(-0.0884144943875)*(1+2.*(dim-1.)/(o+1.))+(0.34752958239)*dim
        ref[2,0,2,1]=(-1.87586060053)*(1+2.*(dim-1.)/(o+1.))+(-0.723648658528)*dim
        ref[2,1,0,0]=(-0.92540503301)*(1+2.*(dim-1.)/(o+1.))+(-1.0656458547)*dim
        ref[2,1,0,1]=(-0.910082743841)*(1+2.*(dim-1.)/(o+1.))+(-0.849761826324)*dim
        ref[2,1,1,0]=(1.35254772756)*(1+2.*(dim-1.)/(o+1.))+(0.507103119278)*dim
        ref[2,1,1,1]=(-0.719987787779)*(1+2.*(dim-1.)/(o+1.))+(-1.30678208477)*dim
        ref[2,1,2,0]=(-0.0766010791773)*(1+2.*(dim-1.)/(o+1.))+(1.58281390198)*dim
        ref[2,1,2,1]=(-0.89868937111)*(1+2.*(dim-1.)/(o+1.))+(-0.27203232932)*dim
        ref[2,2,0,0]=(-1.16246789517)*(1+2.*(dim-1.)/(o+1.))+(1.21893246115)*dim
        ref[2,2,0,1]=(0.22021572107)*(1+2.*(dim-1.)/(o+1.))+(0.977315318401)*dim
        ref[2,2,1,0]=(0.12704942291)*(1+2.*(dim-1.)/(o+1.))+(1.53745200273)*dim
        ref[2,2,1,1]=(0.710193625958)*(1+2.*(dim-1.)/(o+1.))+(0.519254918615)*dim
        ref[2,2,2,0]=(-0.43651451338)*(1+2.*(dim-1.)/(o+1.))+(-0.956516790649)*dim
        ref[2,2,2,1]=(-1.00626199949)*(1+2.*(dim-1.)/(o+1.))+(-0.581037342991)*dim
        ref[2,3,0,0]=(-0.287857609232)*(1+2.*(dim-1.)/(o+1.))+(-0.0749197461649)*dim
        ref[2,3,0,1]=(-0.182022367349)*(1+2.*(dim-1.)/(o+1.))+(-0.710151591543)*dim
        ref[2,3,1,0]=(0.176577445611)*(1+2.*(dim-1.)/(o+1.))+(-1.28903651063)*dim
        ref[2,3,1,1]=(-1.10139844683)*(1+2.*(dim-1.)/(o+1.))+(0.572088107754)*dim
        ref[2,3,2,0]=(1.33006028637)*(1+2.*(dim-1.)/(o+1.))+(0.167651472878)*dim
        ref[2,3,2,1]=(1.17141637019)*(1+2.*(dim-1.)/(o+1.))+(1.25083441135)*dim
        ref[2,4,0,0]=(-1.41846898852)*(1+2.*(dim-1.)/(o+1.))+(-0.315334370591)*dim
        ref[2,4,0,1]=(-1.43804466583)*(1+2.*(dim-1.)/(o+1.))+(-0.20239650092)*dim
        ref[2,4,1,0]=(0.0557193237997)*(1+2.*(dim-1.)/(o+1.))+(-0.247859604111)*dim
        ref[2,4,1,1]=(-0.206133031383)*(1+2.*(dim-1.)/(o+1.))+(1.13858809704)*dim
        ref[2,4,2,0]=(-1.51882846836)*(1+2.*(dim-1.)/(o+1.))+(-1.65673313901)*dim
        ref[2,4,2,1]=(-0.376617494119)*(1+2.*(dim-1.)/(o+1.))+(0.041258877271)*dim
        ref[3,0,0,0]=(1.46263117557)*(1+2.*(dim-1.)/(o+1.))+(0.482754526761)*dim
        ref[3,0,0,1]=(0.52598904073)*(1+2.*(dim-1.)/(o+1.))+(-0.415183068887)*dim
        ref[3,0,1,0]=(0.853716506126)*(1+2.*(dim-1.)/(o+1.))+(-0.0425200502867)*dim
        ref[3,0,1,1]=(-0.827480777065)*(1+2.*(dim-1.)/(o+1.))+(1.08461455312)*dim
        ref[3,0,2,0]=(-0.302457960832)*(1+2.*(dim-1.)/(o+1.))+(1.35255757643)*dim
        ref[3,0,2,1]=(0.999712537623)*(1+2.*(dim-1.)/(o+1.))+(-1.22791831466)*dim
        ref[3,1,0,0]=(1.37558175729)*(1+2.*(dim-1.)/(o+1.))+(1.5072396492)*dim
        ref[3,1,0,1]=(-0.610983032326)*(1+2.*(dim-1.)/(o+1.))+(1.20456015565)*dim
        ref[3,1,1,0]=(-0.557154245831)*(1+2.*(dim-1.)/(o+1.))+(-0.157447066158)*dim
        ref[3,1,1,1]=(0.316637791451)*(1+2.*(dim-1.)/(o+1.))+(-1.04049027266)*dim
        ref[3,1,2,0]=(0.675812719647)*(1+2.*(dim-1.)/(o+1.))+(-0.151903316967)*dim
        ref[3,1,2,1]=(1.46209184373)*(1+2.*(dim-1.)/(o+1.))+(1.52136911386)*dim
        ref[3,2,0,0]=(0.313312340454)*(1+2.*(dim-1.)/(o+1.))+(-0.93288393866)*dim
        ref[3,2,0,1]=(0.813961720148)*(1+2.*(dim-1.)/(o+1.))+(0.686571492997)*dim
        ref[3,2,1,0]=(0.40169167047)*(1+2.*(dim-1.)/(o+1.))+(1.29470020468)*dim
        ref[3,2,1,1]=(-0.363597675405)*(1+2.*(dim-1.)/(o+1.))+(0.066933311055)*dim
        ref[3,2,2,0]=(-0.810627908174)*(1+2.*(dim-1.)/(o+1.))+(0.657328643189)*dim
        ref[3,2,2,1]=(1.62157043056)*(1+2.*(dim-1.)/(o+1.))+(-1.00815118124)*dim
        ref[3,3,0,0]=(0.591377805617)*(1+2.*(dim-1.)/(o+1.))+(-0.637033412497)*dim
        ref[3,3,0,1]=(-1.39729989956)*(1+2.*(dim-1.)/(o+1.))+(-0.485965981125)*dim
        ref[3,3,1,0]=(0.0351856885862)*(1+2.*(dim-1.)/(o+1.))+(0.815243981437)*dim
        ref[3,3,1,1]=(-0.216930644354)*(1+2.*(dim-1.)/(o+1.))+(-0.457095043478)*dim
        ref[3,3,2,0]=(0.0739929579029)*(1+2.*(dim-1.)/(o+1.))+(-1.41564890451)*dim
        ref[3,3,2,1]=(1.66402020703)*(1+2.*(dim-1.)/(o+1.))+(-1.53484781188)*dim
        ref[3,4,0,0]=(0.0397952663815)*(1+2.*(dim-1.)/(o+1.))+(-0.634341555254)*dim
        ref[3,4,0,1]=(0.798323376174)*(1+2.*(dim-1.)/(o+1.))+(0.877607915525)*dim
        ref[3,4,1,0]=(-1.749602045)*(1+2.*(dim-1.)/(o+1.))+(-0.29959901409)*dim
        ref[3,4,1,1]=(-0.0683584060144)*(1+2.*(dim-1.)/(o+1.))+(1.34341750398)*dim
        ref[3,4,2,0]=(-0.235377751413)*(1+2.*(dim-1.)/(o+1.))+(-0.108899384872)*dim
        ref[3,4,2,1]=(-0.849604440861)*(1+2.*(dim-1.)/(o+1.))+(-1.45651852241)*dim
      else:
        arg[0,0,0,0]=(-0.229219068739)*x[0]**o+(0.293241016924)*x[0]+(-0.2730739542)*x[1]**o+(0.980438707356)*x[1]+(-0.712463216466)*x[2]**o+(0.201036317883)*x[2]
        arg[0,0,0,1]=(0.626271947422)*x[0]**o+(0.379293769575)*x[0]+(-0.476378592342)*x[1]**o+(-0.0089906260934)*x[1]+(-0.433154385811)*x[2]**o+(0.126766298748)*x[2]
        arg[0,0,1,0]=(-0.517950283035)*x[0]**o+(-0.219780986066)*x[0]+(0.345048580545)*x[1]**o+(0.116833851128)*x[1]+(-0.790641523314)*x[2]**o+(0.40506306417)*x[2]
        arg[0,0,1,1]=(0.910924420538)*x[0]**o+(0.281102231596)*x[0]+(0.115454208619)*x[1]**o+(0.164254056171)*x[1]+(0.44960060103)*x[2]**o+(-0.14238069165)*x[2]
        arg[0,0,2,0]=(0.86195808217)*x[0]**o+(0.775425956811)*x[0]+(0.214559021523)*x[1]**o+(0.400663063038)*x[1]+(0.995663608954)*x[2]**o+(0.228658543268)*x[2]
        arg[0,0,2,1]=(0.9025917492)*x[0]**o+(0.643205662573)*x[0]+(-0.304382459664)*x[1]**o+(-0.653256441532)*x[1]+(0.591142459807)*x[2]**o+(-0.888899030043)*x[2]
        arg[0,1,0,0]=(-0.874720784838)*x[0]**o+(-0.961456282274)*x[0]+(-0.954972993457)*x[1]**o+(0.00963939708746)*x[1]+(-0.884848381277)*x[2]**o+(-0.859924591622)*x[2]
        arg[0,1,0,1]=(0.133592941787)*x[0]**o+(0.12505415793)*x[0]+(-0.872197297559)*x[1]**o+(-0.666936976734)*x[1]+(-0.540166108036)*x[2]**o+(0.739145002101)*x[2]
        arg[0,1,1,0]=(0.543825266564)*x[0]**o+(0.167854064952)*x[0]+(-0.876426395953)*x[1]**o+(0.317538118917)*x[1]+(0.0454531201303)*x[2]**o+(0.822731343286)*x[2]
        arg[0,1,1,1]=(0.294557050513)*x[0]**o+(0.811698886899)*x[0]+(0.906213580147)*x[1]**o+(0.47895749797)*x[1]+(-0.937617554883)*x[2]**o+(0.171907458148)*x[2]
        arg[0,1,2,0]=(-0.843310949654)*x[0]**o+(0.766452648675)*x[0]+(-0.801151193917)*x[1]**o+(-0.373068887299)*x[1]+(-0.908043738284)*x[2]**o+(0.742278324563)*x[2]
        arg[0,1,2,1]=(-0.83259110525)*x[0]**o+(-0.303796015231)*x[0]+(0.153727560059)*x[1]**o+(-0.848393854019)*x[1]+(0.887795931916)*x[2]**o+(0.713486413074)*x[2]
        arg[0,2,0,0]=(-0.928603918182)*x[0]**o+(0.631213545526)*x[0]+(0.0149216054247)*x[1]**o+(0.136454313226)*x[1]+(-0.182080277504)*x[2]**o+(-0.898437757742)*x[2]
        arg[0,2,0,1]=(0.553844129645)*x[0]**o+(0.91439079798)*x[0]+(0.260424446678)*x[1]**o+(0.786629777779)*x[1]+(-0.412353876109)*x[2]**o+(0.496953000487)*x[2]
        arg[0,2,1,0]=(-0.48365422571)*x[0]**o+(-0.772252454664)*x[0]+(0.692592845143)*x[1]**o+(0.861992232175)*x[1]+(0.65169271692)*x[2]**o+(0.889175716524)*x[2]
        arg[0,2,1,1]=(-0.761086015475)*x[0]**o+(-0.616528617148)*x[0]+(-0.949399556926)*x[1]**o+(-0.783355074724)*x[1]+(-0.260470660608)*x[2]**o+(-0.854537542803)*x[2]
        arg[0,2,2,0]=(-0.812649439778)*x[0]**o+(-0.844862957855)*x[0]+(0.719928253134)*x[1]**o+(-0.892803903061)*x[1]+(-0.434753054503)*x[2]**o+(0.814188548451)*x[2]
        arg[0,2,2,1]=(-0.831106881173)*x[0]**o+(0.893305297647)*x[0]+(-0.351357795644)*x[1]**o+(0.0318574503478)*x[1]+(0.721896657873)*x[2]**o+(-0.75215838035)*x[2]
        arg[0,3,0,0]=(-0.868084492214)*x[0]**o+(-0.108398365838)*x[0]+(0.994384983885)*x[1]**o+(0.199445529335)*x[1]+(-0.570120483013)*x[2]**o+(0.992752456282)*x[2]
        arg[0,3,0,1]=(-0.304553428923)*x[0]**o+(-0.589461304318)*x[0]+(-0.570142565817)*x[1]**o+(-0.858530013475)*x[1]+(0.0690574539796)*x[2]**o+(-0.0321587525658)*x[2]
        arg[0,3,1,0]=(-0.729413026954)*x[0]**o+(-0.728248951404)*x[0]+(-0.568598599095)*x[1]**o+(0.0823012989572)*x[1]+(0.248614424573)*x[2]**o+(-0.116584878805)*x[2]
        arg[0,3,1,1]=(0.586846538084)*x[0]**o+(-0.406933529762)*x[0]+(0.839125682244)*x[1]**o+(0.237338453386)*x[1]+(-0.953101889056)*x[2]**o+(0.164147861616)*x[2]
        arg[0,3,2,0]=(-0.603895026116)*x[0]**o+(-0.401045478885)*x[0]+(-0.524739550832)*x[1]**o+(-0.077907095917)*x[1]+(0.535404409567)*x[2]**o+(0.745721528391)*x[2]
        arg[0,3,2,1]=(-0.189151084147)*x[0]**o+(-0.865927541126)*x[0]+(0.55265091657)*x[1]**o+(-0.592847311104)*x[1]+(-0.467509383222)*x[2]**o+(-0.779649276678)*x[2]
        arg[0,4,0,0]=(-0.564283606569)*x[0]**o+(-0.588029571592)*x[0]+(0.544948665159)*x[1]**o+(0.140885576219)*x[1]+(-0.333531842397)*x[2]**o+(-0.579095398292)*x[2]
        arg[0,4,0,1]=(-0.189127238455)*x[0]**o+(-0.693907127162)*x[0]+(0.40947793861)*x[1]**o+(-0.371098494308)*x[1]+(-0.189273002147)*x[2]**o+(-0.863169616331)*x[2]
        arg[0,4,1,0]=(-0.929838141145)*x[0]**o+(-0.0237583231579)*x[0]+(0.721069982341)*x[1]**o+(0.0378559913864)*x[1]+(-0.354729434476)*x[2]**o+(0.486256042969)*x[2]
        arg[0,4,1,1]=(0.349085173473)*x[0]**o+(0.398939355666)*x[0]+(-0.165772805366)*x[1]**o+(-0.532771334584)*x[1]+(-0.971868054413)*x[2]**o+(-0.38430185761)*x[2]
        arg[0,4,2,0]=(0.0330282399287)*x[0]**o+(0.157336131691)*x[0]+(-0.232739262685)*x[1]**o+(-0.280149712596)*x[1]+(0.374058430291)*x[2]**o+(0.594113015973)*x[2]
        arg[0,4,2,1]=(0.841740974112)*x[0]**o+(-0.454359680369)*x[0]+(-0.29970574726)*x[1]**o+(-0.234747232007)*x[1]+(0.590942390095)*x[2]**o+(-0.241807214234)*x[2]
        arg[1,0,0,0]=(0.815585848775)*x[0]**o+(-0.915020392674)*x[0]+(-0.0246530790781)*x[1]**o+(-0.783552074981)*x[1]+(0.539847778607)*x[2]**o+(-0.239345161181)*x[2]
        arg[1,0,0,1]=(-0.34212551354)*x[0]**o+(-0.231769108271)*x[0]+(-0.495709071836)*x[1]**o+(-0.301124823696)*x[1]+(-0.567450536966)*x[2]**o+(0.553640896859)*x[2]
        arg[1,0,1,0]=(0.507091065517)*x[0]**o+(-0.844317985101)*x[0]+(0.488126307787)*x[1]**o+(-0.624165922596)*x[1]+(0.00960189690918)*x[2]**o+(-0.35617062688)*x[2]
        arg[1,0,1,1]=(-0.219877421252)*x[0]**o+(-0.574843930095)*x[0]+(-0.846995760292)*x[1]**o+(-0.778496435041)*x[1]+(-0.817593097977)*x[2]**o+(0.387406829939)*x[2]
        arg[1,0,2,0]=(-0.774922277048)*x[0]**o+(-0.238152671014)*x[0]+(-0.913770877808)*x[1]**o+(-0.737305937434)*x[1]+(-0.13977056904)*x[2]**o+(-0.0156668131016)*x[2]
        arg[1,0,2,1]=(0.565993265911)*x[0]**o+(0.0690768496992)*x[0]+(-0.019845214223)*x[1]**o+(0.16253166701)*x[1]+(0.848117215885)*x[2]**o+(0.838075165243)*x[2]
        arg[1,1,0,0]=(0.871499694148)*x[0]**o+(-0.640851595157)*x[0]+(-0.170827284611)*x[1]**o+(-0.855550549988)*x[1]+(0.796086141249)*x[2]**o+(-0.214410672377)*x[2]
        arg[1,1,0,1]=(-0.903644604832)*x[0]**o+(0.0604780584681)*x[0]+(0.12800839523)*x[1]**o+(-0.594186752891)*x[1]+(0.415537334962)*x[2]**o+(0.122944611099)*x[2]
        arg[1,1,1,0]=(-0.611432775957)*x[0]**o+(0.391078622669)*x[0]+(-0.943091949273)*x[1]**o+(0.952739057317)*x[1]+(0.205104161722)*x[2]**o+(-0.946040001739)*x[2]
        arg[1,1,1,1]=(0.122941344579)*x[0]**o+(0.836981501711)*x[0]+(0.488970739079)*x[1]**o+(0.210062341907)*x[1]+(0.0264620393683)*x[2]**o+(-0.396386059797)*x[2]
        arg[1,1,2,0]=(-0.782658774038)*x[0]**o+(0.00611865020069)*x[0]+(0.806802875462)*x[1]**o+(-0.504323859583)*x[1]+(0.796276800331)*x[2]**o+(-0.792865116401)*x[2]
        arg[1,1,2,1]=(-0.249170081188)*x[0]**o+(-0.0234780571796)*x[0]+(-0.314017467872)*x[1]**o+(-0.0819148249592)*x[1]+(-0.71553368486)*x[2]**o+(0.763463207899)*x[2]
        arg[1,2,0,0]=(-0.722263632571)*x[0]**o+(-0.922870211147)*x[0]+(0.380794419995)*x[1]**o+(-0.76088532378)*x[1]+(0.417253071012)*x[2]**o+(0.152980049966)*x[2]
        arg[1,2,0,1]=(-0.667661997967)*x[0]**o+(-0.520570751147)*x[0]+(0.526368189149)*x[1]**o+(-0.0729376045288)*x[1]+(-0.434764684251)*x[2]**o+(0.124284939267)*x[2]
        arg[1,2,1,0]=(0.610868035553)*x[0]**o+(0.958644088455)*x[0]+(0.876029854699)*x[1]**o+(-0.278348787699)*x[1]+(-0.576695356283)*x[2]**o+(-0.302696556744)*x[2]
        arg[1,2,1,1]=(-0.224897853162)*x[0]**o+(0.89405397732)*x[0]+(-0.865281799513)*x[1]**o+(0.778866760053)*x[1]+(-0.939295291326)*x[2]**o+(0.256725768393)*x[2]
        arg[1,2,2,0]=(0.763785330202)*x[0]**o+(-0.92861265304)*x[0]+(-0.73488309202)*x[1]**o+(0.912711310857)*x[1]+(-0.359622626692)*x[2]**o+(-0.901261216611)*x[2]
        arg[1,2,2,1]=(-0.70687794303)*x[0]**o+(0.186497640523)*x[0]+(-0.587602932017)*x[1]**o+(0.439204174904)*x[1]+(0.0527573129765)*x[2]**o+(-0.121921819047)*x[2]
        arg[1,3,0,0]=(-0.576513727689)*x[0]**o+(0.228971758182)*x[0]+(0.282735129059)*x[1]**o+(-0.375270097915)*x[1]+(0.360087142563)*x[2]**o+(-0.0629946372113)*x[2]
        arg[1,3,0,1]=(-0.934375471561)*x[0]**o+(-0.710481522462)*x[0]+(-0.506438555546)*x[1]**o+(0.515278994703)*x[1]+(-0.789552111338)*x[2]**o+(0.556694887963)*x[2]
        arg[1,3,1,0]=(0.418081142778)*x[0]**o+(-0.990334755424)*x[0]+(0.465998147831)*x[1]**o+(0.85211792131)*x[1]+(0.64293315455)*x[2]**o+(0.418636667223)*x[2]
        arg[1,3,1,1]=(-0.192682640907)*x[0]**o+(-0.367176463246)*x[0]+(-0.512565915338)*x[1]**o+(0.431102518715)*x[1]+(-0.0213613616827)*x[2]**o+(-0.329868944656)*x[2]
        arg[1,3,2,0]=(-0.63277285658)*x[0]**o+(0.679819079734)*x[0]+(0.986520110022)*x[1]**o+(0.988046257154)*x[1]+(0.973419900492)*x[2]**o+(-0.177540534942)*x[2]
        arg[1,3,2,1]=(-0.660456888389)*x[0]**o+(0.546277948387)*x[0]+(-0.437065207396)*x[1]**o+(-0.310754727958)*x[1]+(-0.786950057282)*x[2]**o+(0.959526365177)*x[2]
        arg[1,4,0,0]=(0.764190976082)*x[0]**o+(0.970535029727)*x[0]+(0.389051006993)*x[1]**o+(-0.289346925704)*x[1]+(-0.386263606576)*x[2]**o+(0.624234063589)*x[2]
        arg[1,4,0,1]=(-0.982030683687)*x[0]**o+(-0.310500292581)*x[0]+(-0.559096285575)*x[1]**o+(-0.622541272039)*x[1]+(0.49924233081)*x[2]**o+(-0.219258681197)*x[2]
        arg[1,4,1,0]=(-0.124268932891)*x[0]**o+(0.102696501388)*x[0]+(-0.135123912862)*x[1]**o+(-0.960215132796)*x[1]+(-0.63171550556)*x[2]**o+(-0.493725285999)*x[2]
        arg[1,4,1,1]=(-0.397915436399)*x[0]**o+(-0.629822230732)*x[0]+(-0.65097233947)*x[1]**o+(0.915396047247)*x[1]+(0.835684562564)*x[2]**o+(-0.538096550078)*x[2]
        arg[1,4,2,0]=(-0.946429448216)*x[0]**o+(-0.310412085571)*x[0]+(-0.0984587257278)*x[1]**o+(-0.662090989911)*x[1]+(-0.424711481307)*x[2]**o+(0.00916879493127)*x[2]
        arg[1,4,2,1]=(0.919399549311)*x[0]**o+(-0.890110018315)*x[0]+(-0.685928475328)*x[1]**o+(-0.4731186395)*x[1]+(0.961887038718)*x[2]**o+(0.850957316131)*x[2]
        arg[2,0,0,0]=(0.834851801283)*x[0]**o+(-0.401572429371)*x[0]+(0.313309302528)*x[1]**o+(0.0368590314435)*x[1]+(-0.299807168778)*x[2]**o+(0.250599904818)*x[2]
        arg[2,0,0,1]=(-0.717562866174)*x[0]**o+(0.374797888506)*x[0]+(0.254329847579)*x[1]**o+(-0.987429692718)*x[1]+(-0.886758521324)*x[2]**o+(0.309486900444)*x[2]
        arg[2,0,1,0]=(0.139962143021)*x[0]**o+(-0.134735774213)*x[0]+(-0.231254041503)*x[1]**o+(0.0694785371931)*x[1]+(0.769929630132)*x[2]**o+(0.486646094292)*x[2]
        arg[2,0,1,1]=(0.349691701871)*x[0]**o+(-0.0308285150496)*x[0]+(0.00977143515921)*x[1]**o+(0.0970015800122)*x[1]+(-0.725991176667)*x[2]**o+(0.433663449433)*x[2]
        arg[2,0,2,0]=(0.564474039592)*x[0]**o+(0.950620065312)*x[0]+(0.256658075134)*x[1]**o+(0.774458453189)*x[1]+(0.00181734756534)*x[2]**o+(-0.270638852561)*x[2]
        arg[2,0,2,1]=(-0.175813811707)*x[0]**o+(0.821341326493)*x[0]+(-0.464966990893)*x[1]**o+(-0.310747066257)*x[1]+(-0.413557861251)*x[2]**o+(-0.799160067701)*x[2]
        arg[2,1,0,0]=(-0.449817381621)*x[0]**o+(-0.0573780378383)*x[0]+(-0.22675349133)*x[1]**o+(-0.150314340624)*x[1]+(-0.116693721705)*x[2]**o+(-0.20577625675)*x[2]
        arg[2,1,0,1]=(0.955318430674)*x[0]**o+(-0.148100168239)*x[0]+(0.741705440574)*x[1]**o+(-0.0768022054634)*x[1]+(-0.721807872287)*x[2]**o+(-0.898235041116)*x[2]
        arg[2,1,1,0]=(-0.183120810472)*x[0]**o+(0.307305712226)*x[0]+(0.699835684443)*x[1]**o+(0.964082351856)*x[1]+(-0.0348051023458)*x[2]**o+(0.544435382916)*x[2]
        arg[2,1,1,1]=(0.786862912642)*x[0]**o+(0.37388939814)*x[0]+(0.608937302857)*x[1]**o+(0.122305286217)*x[1]+(-0.892536849678)*x[2]**o+(0.119257478145)*x[2]
        arg[2,1,2,0]=(0.758154908623)*x[0]**o+(0.695188856067)*x[0]+(0.458313343957)*x[1]**o+(0.979157471455)*x[1]+(0.194237063725)*x[2]**o+(0.773001101454)*x[2]
        arg[2,1,2,1]=(0.0754678814151)*x[0]**o+(-0.639556181007)*x[0]+(0.934627102078)*x[1]**o+(0.0832082572611)*x[1]+(0.668822756858)*x[2]**o+(0.627532210297)*x[2]
        arg[2,2,0,0]=(-0.192921889174)*x[0]**o+(0.989703287553)*x[0]+(0.00517568784693)*x[1]**o+(-0.360076937731)*x[1]+(0.947056110784)*x[2]**o+(0.469841550381)*x[2]
        arg[2,2,0,1]=(0.201501669595)*x[0]**o+(-0.959957961834)*x[0]+(0.0229159090152)*x[1]**o+(0.396188263084)*x[1]+(0.186211667989)*x[2]**o+(0.304503474778)*x[2]
        arg[2,2,1,0]=(0.333974995608)*x[0]**o+(0.210202867626)*x[0]+(-0.410565897643)*x[1]**o+(0.52659985644)*x[1]+(0.983960564934)*x[2]**o+(-0.439323691866)*x[2]
        arg[2,2,1,1]=(0.706119074329)*x[0]**o+(-0.00220875175563)*x[0]+(0.143693169123)*x[1]**o+(0.217254593235)*x[1]+(-0.412057975252)*x[2]**o+(-0.967360361422)*x[2]
        arg[2,2,2,0]=(0.217307356597)*x[0]**o+(0.861728452163)*x[0]+(-0.193239462241)*x[1]**o+(0.872273176526)*x[1]+(-0.436414721507)*x[2]**o+(-0.62398425635)*x[2]
        arg[2,2,2,1]=(0.467099668794)*x[0]**o+(0.109742400918)*x[0]+(-0.987522191801)*x[1]**o+(-0.250251964275)*x[1]+(0.442152166523)*x[2]**o+(-0.709157679576)*x[2]
        arg[2,3,0,0]=(0.769666242031)*x[0]**o+(0.2352745864)*x[0]+(-0.302893533611)*x[1]**o+(0.0806677655098)*x[1]+(0.886434893279)*x[2]**o+(0.366016417047)*x[2]
        arg[2,3,0,1]=(-0.612204706187)*x[0]**o+(-0.242961993133)*x[0]+(0.946225816441)*x[1]**o+(0.528642650581)*x[1]+(-0.0166837450337)*x[2]**o+(-0.0889397945206)*x[2]
        arg[2,3,1,0]=(0.218589745288)*x[0]**o+(-0.296125811204)*x[0]+(0.358743045538)*x[1]**o+(-0.945778818656)*x[1]+(-0.33155158646)*x[2]**o+(-0.293038455557)*x[2]
        arg[2,3,1,1]=(0.753737723433)*x[0]**o+(-0.772926141856)*x[0]+(-0.0107183417902)*x[1]**o+(0.125538612735)*x[1]+(0.225277420156)*x[2]**o+(-0.54876651451)*x[2]
        arg[2,3,2,0]=(0.947419453319)*x[0]**o+(-0.101935160067)*x[0]+(0.481841505589)*x[1]**o+(-0.873842146032)*x[1]+(-0.377764402402)*x[2]**o+(-0.361560947647)*x[2]
        arg[2,3,2,1]=(0.745703609187)*x[0]**o+(-0.488401078192)*x[0]+(0.190261500318)*x[1]**o+(-0.22176180212)*x[1]+(0.80182740764)*x[2]**o+(-0.504953098351)*x[2]
        arg[2,4,0,0]=(0.476355625206)*x[0]**o+(0.452225231178)*x[0]+(0.669315429911)*x[1]**o+(0.159511010158)*x[1]+(0.652056592943)*x[2]**o+(0.44515033609)*x[2]
        arg[2,4,0,1]=(0.512194809916)*x[0]**o+(0.654452558318)*x[0]+(-0.526832882528)*x[1]**o+(0.590859897981)*x[1]+(-0.155989159043)*x[2]**o+(-0.689178072559)*x[2]
        arg[2,4,1,0]=(0.0450244624553)*x[0]**o+(0.0790299000335)*x[0]+(-0.690643908994)*x[1]**o+(-0.262426911912)*x[1]+(-0.426935471303)*x[2]**o+(-0.434570506153)*x[2]
        arg[2,4,1,1]=(-0.04445890079)*x[0]**o+(-0.541574770252)*x[0]+(-0.879290270122)*x[1]**o+(0.545302743626)*x[1]+(-0.836472565261)*x[2]**o+(0.66750272736)*x[2]
        arg[2,4,2,0]=(-0.477788509677)*x[0]**o+(0.706195950613)*x[0]+(0.628430029293)*x[1]**o+(-0.297608932813)*x[1]+(0.38364501438)*x[2]**o+(0.855444938831)*x[2]
        arg[2,4,2,1]=(0.123183000451)*x[0]**o+(0.41176132265)*x[0]+(-0.379498883356)*x[1]**o+(-0.53442961658)*x[1]+(-0.101215056317)*x[2]**o+(0.247971839461)*x[2]
        arg[3,0,0,0]=(-0.239170701904)*x[0]**o+(0.369420936359)*x[0]+(0.972913513036)*x[1]**o+(0.775248415403)*x[1]+(-0.850310640223)*x[2]**o+(-0.815345534491)*x[2]
        arg[3,0,0,1]=(0.591124498004)*x[0]**o+(0.529522171858)*x[0]+(-0.974005221641)*x[1]**o+(0.430102201571)*x[1]+(0.570519406976)*x[2]**o+(-0.0915083364671)*x[2]
        arg[3,0,1,0]=(-0.27678065557)*x[0]**o+(-0.53225308602)*x[0]+(0.902356190541)*x[1]**o+(-0.564808732046)*x[1]+(0.0981283775668)*x[2]**o+(-0.64957631107)*x[2]
        arg[3,0,1,1]=(0.230523873408)*x[0]**o+(-0.382745197262)*x[0]+(0.41507101812)*x[1]**o+(0.48492708335)*x[1]+(-0.11672353881)*x[2]**o+(-0.56902283397)*x[2]
        arg[3,0,2,0]=(-0.400023140104)*x[0]**o+(-0.406234012546)*x[0]+(-0.17810623045)*x[1]**o+(-0.833748300281)*x[1]+(-0.463689953863)*x[2]**o+(-0.11095482465)*x[2]
        arg[3,0,2,1]=(0.448885918108)*x[0]**o+(0.0447046792888)*x[0]+(-0.781652457281)*x[1]**o+(-0.893726375109)*x[1]+(0.918099233406)*x[2]**o+(0.560717614604)*x[2]
        arg[3,1,0,0]=(0.046700262853)*x[0]**o+(0.729867044213)*x[0]+(-0.870868205823)*x[1]**o+(-0.711267787629)*x[1]+(0.760550112312)*x[2]**o+(-0.0188334901698)*x[2]
        arg[3,1,0,1]=(0.304305551549)*x[0]**o+(-0.73527291962)*x[0]+(0.722134673974)*x[1]**o+(-0.40822784698)*x[1]+(0.912464122456)*x[2]**o+(-0.857656020349)*x[2]
        arg[3,1,1,0]=(-0.857864508282)*x[0]**o+(-0.22160002809)*x[0]+(-0.0892530574233)*x[1]**o+(-0.547259900671)*x[1]+(0.641858628113)*x[2]**o+(-0.203700292205)*x[2]
        arg[3,1,1,1]=(-0.47064403668)*x[0]**o+(-0.105885697358)*x[0]+(-0.899570240486)*x[1]**o+(0.660942689756)*x[1]+(-0.365836826289)*x[2]**o+(0.9593955249)*x[2]
        arg[3,1,2,0]=(-0.51407433496)*x[0]**o+(0.369092182634)*x[0]+(-0.740281092517)*x[1]**o+(0.383443939366)*x[1]+(-0.140318759419)*x[2]**o+(-0.481274651104)*x[2]
        arg[3,1,2,1]=(-0.938924125636)*x[0]**o+(0.41994781084)*x[0]+(-0.946320511542)*x[1]**o+(-0.86702606561)*x[1]+(0.63547569926)*x[2]**o+(0.923988068696)*x[2]
        arg[3,2,0,0]=(0.212833889165)*x[0]**o+(-0.0823077975238)*x[0]+(0.038483237489)*x[1]**o+(0.393784376623)*x[1]+(0.573190557103)*x[2]**o+(0.403015743632)*x[2]
        arg[3,2,0,1]=(-0.169294531943)*x[0]**o+(-0.769228919251)*x[0]+(-0.907462814057)*x[1]**o+(-0.188574782291)*x[1]+(-0.859793027531)*x[2]**o+(-0.673334626964)*x[2]
        arg[3,2,1,0]=(0.646744388517)*x[0]**o+(0.346749855021)*x[0]+(-0.314382791156)*x[1]**o+(-0.0396545336605)*x[1]+(0.844678566939)*x[2]**o+(0.724721336727)*x[2]
        arg[3,2,1,1]=(-0.471761417345)*x[0]**o+(-0.801631597222)*x[0]+(-0.376444012893)*x[1]**o+(0.920939382926)*x[1]+(-0.728019331398)*x[2]**o+(0.905808341144)*x[2]
        arg[3,2,2,0]=(-0.713759272803)*x[0]**o+(-0.930638013429)*x[0]+(0.864451879958)*x[1]**o+(0.0566794358824)*x[1]+(-0.292092226905)*x[2]**o+(-0.195695530268)*x[2]
        arg[3,2,2,1]=(0.843437618194)*x[0]**o+(0.214911962718)*x[0]+(0.475690162826)*x[1]**o+(0.0744570063618)*x[1]+(0.351711310616)*x[2]**o+(-0.187223450548)*x[2]
        arg[3,3,0,0]=(0.108651500878)*x[0]**o+(-0.402624590847)*x[0]+(0.120700159277)*x[1]**o+(-0.363139807635)*x[1]+(-0.62413441951)*x[2]**o+(-0.474236131596)*x[2]
        arg[3,3,0,1]=(-0.0836636547536)*x[0]**o+(-0.187959781777)*x[0]+(0.756983584308)*x[1]**o+(0.756573839599)*x[1]+(0.195640878564)*x[2]**o+(-0.566596474032)*x[2]
        arg[3,3,1,0]=(-0.171427844449)*x[0]**o+(-0.266807792484)*x[0]+(0.380046037965)*x[1]**o+(-0.129297371793)*x[1]+(0.79139100315)*x[2]**o+(0.223804751067)*x[2]
        arg[3,3,1,1]=(-0.054757303264)*x[0]**o+(-0.823878124196)*x[0]+(0.80695256107)*x[1]**o+(0.583250192211)*x[1]+(0.894937939133)*x[2]**o+(0.878757173892)*x[2]
        arg[3,3,2,0]=(0.13923802434)*x[0]**o+(-0.600732296834)*x[0]+(0.111226716676)*x[1]**o+(-0.474524749496)*x[1]+(0.674290476359)*x[2]**o+(-0.06714767759)*x[2]
        arg[3,3,2,1]=(0.246187769682)*x[0]**o+(-0.871969439162)*x[0]+(0.56748693564)*x[1]**o+(-0.0831774328079)*x[1]+(-0.27380467532)*x[2]**o+(-0.623711953244)*x[2]
        arg[3,4,0,0]=(-0.345947366312)*x[0]**o+(-0.304842609526)*x[0]+(-0.720722450633)*x[1]**o+(0.889712922756)*x[1]+(-0.0451362945237)*x[2]**o+(-0.52344522887)*x[2]
        arg[3,4,0,1]=(-0.830119647904)*x[0]**o+(0.577888363835)*x[0]+(0.180691945154)*x[1]**o+(-0.489554894073)*x[1]+(0.848489044545)*x[2]**o+(-0.243139985297)*x[2]
        arg[3,4,1,0]=(0.846597693999)*x[0]**o+(0.757477455396)*x[0]+(-0.827604908657)*x[1]**o+(-0.108392312717)*x[1]+(0.597454707613)*x[2]**o+(0.35846243312)*x[2]
        arg[3,4,1,1]=(0.107973753109)*x[0]**o+(0.578649460223)*x[0]+(0.609012033524)*x[1]**o+(-0.8428861922)*x[1]+(0.42144759968)*x[2]**o+(-0.938952208084)*x[2]
        arg[3,4,2,0]=(0.286586330618)*x[0]**o+(0.7324646289)*x[0]+(-0.448461377611)*x[1]**o+(0.62562408085)*x[1]+(0.352429283116)*x[2]**o+(0.112047034296)*x[2]
        arg[3,4,2,1]=(-0.431663387082)*x[0]**o+(-0.612934322552)*x[0]+(-0.426644797074)*x[1]**o+(0.768645027094)*x[1]+(-0.453242885757)*x[2]**o+(0.893748345809)*x[2]
        ref[0,0,0,0]=(-1.2147562394)*(1+2.*(dim-1.)/(o+1.))+(1.47471604216)*dim
        ref[0,0,0,1]=(-0.283261030731)*(1+2.*(dim-1.)/(o+1.))+(0.49706944223)*dim
        ref[0,0,1,0]=(-0.963543225803)*(1+2.*(dim-1.)/(o+1.))+(0.302115929232)*dim
        ref[0,0,1,1]=(1.47597923019)*(1+2.*(dim-1.)/(o+1.))+(0.302975596116)*dim
        ref[0,0,2,0]=(2.07218071265)*(1+2.*(dim-1.)/(o+1.))+(1.40474756312)*dim
        ref[0,0,2,1]=(1.18935174934)*(1+2.*(dim-1.)/(o+1.))+(-0.898949809002)*dim
        ref[0,1,0,0]=(-2.71454215957)*(1+2.*(dim-1.)/(o+1.))+(-1.81174147681)*dim
        ref[0,1,0,1]=(-1.27877046381)*(1+2.*(dim-1.)/(o+1.))+(0.197262183297)*dim
        ref[0,1,1,0]=(-0.287148009258)*(1+2.*(dim-1.)/(o+1.))+(1.30812352716)*dim
        ref[0,1,1,1]=(0.263153075777)*(1+2.*(dim-1.)/(o+1.))+(1.46256384302)*dim
        ref[0,1,2,0]=(-2.55250588186)*(1+2.*(dim-1.)/(o+1.))+(1.13566208594)*dim
        ref[0,1,2,1]=(0.208932386724)*(1+2.*(dim-1.)/(o+1.))+(-0.438703456176)*dim
        ref[0,2,0,0]=(-1.09576259026)*(1+2.*(dim-1.)/(o+1.))+(-0.13076989899)*dim
        ref[0,2,0,1]=(0.401914700214)*(1+2.*(dim-1.)/(o+1.))+(2.19797357625)*dim
        ref[0,2,1,0]=(0.860631336354)*(1+2.*(dim-1.)/(o+1.))+(0.978915494035)*dim
        ref[0,2,1,1]=(-1.97095623301)*(1+2.*(dim-1.)/(o+1.))+(-2.25442123467)*dim
        ref[0,2,2,0]=(-0.527474241147)*(1+2.*(dim-1.)/(o+1.))+(-0.923478312465)*dim
        ref[0,2,2,1]=(-0.460568018944)*(1+2.*(dim-1.)/(o+1.))+(0.173004367644)*dim
        ref[0,3,0,0]=(-0.443819991343)*(1+2.*(dim-1.)/(o+1.))+(1.08379961978)*dim
        ref[0,3,0,1]=(-0.80563854076)*(1+2.*(dim-1.)/(o+1.))+(-1.48015007036)*dim
        ref[0,3,1,0]=(-1.04939720148)*(1+2.*(dim-1.)/(o+1.))+(-0.762532531252)*dim
        ref[0,3,1,1]=(0.472870331273)*(1+2.*(dim-1.)/(o+1.))+(-0.00544721475928)*dim
        ref[0,3,2,0]=(-0.593230167381)*(1+2.*(dim-1.)/(o+1.))+(0.26676895359)*dim
        ref[0,3,2,1]=(-0.104009550799)*(1+2.*(dim-1.)/(o+1.))+(-2.23842412891)*dim
        ref[0,4,0,0]=(-0.352866783806)*(1+2.*(dim-1.)/(o+1.))+(-1.02623939367)*dim
        ref[0,4,0,1]=(0.0310776980081)*(1+2.*(dim-1.)/(o+1.))+(-1.9281752378)*dim
        ref[0,4,1,0]=(-0.56349759328)*(1+2.*(dim-1.)/(o+1.))+(0.500353711197)*dim
        ref[0,4,1,1]=(-0.788555686306)*(1+2.*(dim-1.)/(o+1.))+(-0.518133836527)*dim
        ref[0,4,2,0]=(0.174347407535)*(1+2.*(dim-1.)/(o+1.))+(0.471299435068)*dim
        ref[0,4,2,1]=(1.13297761695)*(1+2.*(dim-1.)/(o+1.))+(-0.93091412661)*dim
        ref[1,0,0,0]=(1.3307805483)*(1+2.*(dim-1.)/(o+1.))+(-1.93791762884)*dim
        ref[1,0,0,1]=(-1.40528512234)*(1+2.*(dim-1.)/(o+1.))+(0.0207469648915)*dim
        ref[1,0,1,0]=(1.00481927021)*(1+2.*(dim-1.)/(o+1.))+(-1.82465453458)*dim
        ref[1,0,1,1]=(-1.88446627952)*(1+2.*(dim-1.)/(o+1.))+(-0.965933535196)*dim
        ref[1,0,2,0]=(-1.8284637239)*(1+2.*(dim-1.)/(o+1.))+(-0.991125421549)*dim
        ref[1,0,2,1]=(1.39426526757)*(1+2.*(dim-1.)/(o+1.))+(1.06968368195)*dim
        ref[1,1,0,0]=(1.49675855079)*(1+2.*(dim-1.)/(o+1.))+(-1.71081281752)*dim
        ref[1,1,0,1]=(-0.36009887464)*(1+2.*(dim-1.)/(o+1.))+(-0.410764083324)*dim
        ref[1,1,1,0]=(-1.34942056351)*(1+2.*(dim-1.)/(o+1.))+(0.397777678247)*dim
        ref[1,1,1,1]=(0.638374123027)*(1+2.*(dim-1.)/(o+1.))+(0.650657783821)*dim
        ref[1,1,2,0]=(0.820420901755)*(1+2.*(dim-1.)/(o+1.))+(-1.29107032578)*dim
        ref[1,1,2,1]=(-1.27872123392)*(1+2.*(dim-1.)/(o+1.))+(0.65807032576)*dim
        ref[1,2,0,0]=(0.075783858436)*(1+2.*(dim-1.)/(o+1.))+(-1.53077548496)*dim
        ref[1,2,0,1]=(-0.576058493068)*(1+2.*(dim-1.)/(o+1.))+(-0.469223416409)*dim
        ref[1,2,1,0]=(0.91020253397)*(1+2.*(dim-1.)/(o+1.))+(0.377598744012)*dim
        ref[1,2,1,1]=(-2.029474944)*(1+2.*(dim-1.)/(o+1.))+(1.92964650577)*dim
        ref[1,2,2,0]=(-0.330720388509)*(1+2.*(dim-1.)/(o+1.))+(-0.917162558794)*dim
        ref[1,2,2,1]=(-1.24172356207)*(1+2.*(dim-1.)/(o+1.))+(0.503779996381)*dim
        ref[1,3,0,0]=(0.0663085439332)*(1+2.*(dim-1.)/(o+1.))+(-0.209292976945)*dim
        ref[1,3,0,1]=(-2.23036613844)*(1+2.*(dim-1.)/(o+1.))+(0.361492360204)*dim
        ref[1,3,1,0]=(1.52701244516)*(1+2.*(dim-1.)/(o+1.))+(0.280419833109)*dim
        ref[1,3,1,1]=(-0.726609917928)*(1+2.*(dim-1.)/(o+1.))+(-0.265942889187)*dim
        ref[1,3,2,0]=(1.32716715393)*(1+2.*(dim-1.)/(o+1.))+(1.49032480195)*dim
        ref[1,3,2,1]=(-1.88447215307)*(1+2.*(dim-1.)/(o+1.))+(1.19504958561)*dim
        ref[1,4,0,0]=(0.766978376499)*(1+2.*(dim-1.)/(o+1.))+(1.30542216761)*dim
        ref[1,4,0,1]=(-1.04188463845)*(1+2.*(dim-1.)/(o+1.))+(-1.15230024582)*dim
        ref[1,4,1,0]=(-0.891108351313)*(1+2.*(dim-1.)/(o+1.))+(-1.35124391741)*dim
        ref[1,4,1,1]=(-0.213203213304)*(1+2.*(dim-1.)/(o+1.))+(-0.252522733564)*dim
        ref[1,4,2,0]=(-1.46959965525)*(1+2.*(dim-1.)/(o+1.))+(-0.963334280551)*dim
        ref[1,4,2,1]=(1.1953581127)*(1+2.*(dim-1.)/(o+1.))+(-0.512271341684)*dim
        ref[2,0,0,0]=(0.848353935033)*(1+2.*(dim-1.)/(o+1.))+(-0.11411349311)*dim
        ref[2,0,0,1]=(-1.34999153992)*(1+2.*(dim-1.)/(o+1.))+(-0.303144903768)*dim
        ref[2,0,1,0]=(0.67863773165)*(1+2.*(dim-1.)/(o+1.))+(0.421388857272)*dim
        ref[2,0,1,1]=(-0.366528039637)*(1+2.*(dim-1.)/(o+1.))+(0.499836514396)*dim
        ref[2,0,2,0]=(0.822949462292)*(1+2.*(dim-1.)/(o+1.))+(1.45443966594)*dim
        ref[2,0,2,1]=(-1.05433866385)*(1+2.*(dim-1.)/(o+1.))+(-0.288565807465)*dim
        ref[2,1,0,0]=(-0.793264594657)*(1+2.*(dim-1.)/(o+1.))+(-0.413468635212)*dim
        ref[2,1,0,1]=(0.975215998961)*(1+2.*(dim-1.)/(o+1.))+(-1.12313741482)*dim
        ref[2,1,1,0]=(0.481909771625)*(1+2.*(dim-1.)/(o+1.))+(1.815823447)*dim
        ref[2,1,1,1]=(0.503263365821)*(1+2.*(dim-1.)/(o+1.))+(0.615452162503)*dim
        ref[2,1,2,0]=(1.41070531631)*(1+2.*(dim-1.)/(o+1.))+(2.44734742898)*dim
        ref[2,1,2,1]=(1.67891774035)*(1+2.*(dim-1.)/(o+1.))+(0.0711842865505)*dim
        ref[2,2,0,0]=(0.759309909457)*(1+2.*(dim-1.)/(o+1.))+(1.0994679002)*dim
        ref[2,2,0,1]=(0.410629246599)*(1+2.*(dim-1.)/(o+1.))+(-0.259266223972)*dim
        ref[2,2,1,0]=(0.907369662899)*(1+2.*(dim-1.)/(o+1.))+(0.297479032201)*dim
        ref[2,2,1,1]=(0.437754268199)*(1+2.*(dim-1.)/(o+1.))+(-0.752314519943)*dim
        ref[2,2,2,0]=(-0.412346827151)*(1+2.*(dim-1.)/(o+1.))+(1.11001737234)*dim
        ref[2,2,2,1]=(-0.0782703564833)*(1+2.*(dim-1.)/(o+1.))+(-0.849667242932)*dim
        ref[2,3,0,0]=(1.3532076017)*(1+2.*(dim-1.)/(o+1.))+(0.681958768956)*dim
        ref[2,3,0,1]=(0.317337365221)*(1+2.*(dim-1.)/(o+1.))+(0.196740862928)*dim
        ref[2,3,1,0]=(0.245781204366)*(1+2.*(dim-1.)/(o+1.))+(-1.53494308542)*dim
        ref[2,3,1,1]=(0.968296801799)*(1+2.*(dim-1.)/(o+1.))+(-1.19615404363)*dim
        ref[2,3,2,0]=(1.05149655651)*(1+2.*(dim-1.)/(o+1.))+(-1.33733825375)*dim
        ref[2,3,2,1]=(1.73779251715)*(1+2.*(dim-1.)/(o+1.))+(-1.21511597866)*dim
        ref[2,4,0,0]=(1.79772764806)*(1+2.*(dim-1.)/(o+1.))+(1.05688657743)*dim
        ref[2,4,0,1]=(-0.170627231654)*(1+2.*(dim-1.)/(o+1.))+(0.55613438374)*dim
        ref[2,4,1,0]=(-1.07255491784)*(1+2.*(dim-1.)/(o+1.))+(-0.617967518031)*dim
        ref[2,4,1,1]=(-1.76022173617)*(1+2.*(dim-1.)/(o+1.))+(0.671230700733)*dim
        ref[2,4,2,0]=(0.534286533995)*(1+2.*(dim-1.)/(o+1.))+(1.26403195663)*dim
        ref[2,4,2,1]=(-0.357530939222)*(1+2.*(dim-1.)/(o+1.))+(0.12530354553)*dim
        ref[3,0,0,0]=(-0.11656782909)*(1+2.*(dim-1.)/(o+1.))+(0.32932381727)*dim
        ref[3,0,0,1]=(0.187638683339)*(1+2.*(dim-1.)/(o+1.))+(0.868116036962)*dim
        ref[3,0,1,0]=(0.723703912538)*(1+2.*(dim-1.)/(o+1.))+(-1.74663812914)*dim
        ref[3,0,1,1]=(0.528871352717)*(1+2.*(dim-1.)/(o+1.))+(-0.466840947882)*dim
        ref[3,0,2,0]=(-1.04181932442)*(1+2.*(dim-1.)/(o+1.))+(-1.35093713748)*dim
        ref[3,0,2,1]=(0.585332694234)*(1+2.*(dim-1.)/(o+1.))+(-0.288304081216)*dim
        ref[3,1,0,0]=(-0.0636178306576)*(1+2.*(dim-1.)/(o+1.))+(-0.000234233586678)*dim
        ref[3,1,0,1]=(1.93890434798)*(1+2.*(dim-1.)/(o+1.))+(-2.00115678695)*dim
        ref[3,1,1,0]=(-0.305258937592)*(1+2.*(dim-1.)/(o+1.))+(-0.972560220966)*dim
        ref[3,1,1,1]=(-1.73605110346)*(1+2.*(dim-1.)/(o+1.))+(1.5144525173)*dim
        ref[3,1,2,0]=(-1.3946741869)*(1+2.*(dim-1.)/(o+1.))+(0.271261470896)*dim
        ref[3,1,2,1]=(-1.24976893792)*(1+2.*(dim-1.)/(o+1.))+(0.476909813927)*dim
        ref[3,2,0,0]=(0.824507683756)*(1+2.*(dim-1.)/(o+1.))+(0.71449232273)*dim
        ref[3,2,0,1]=(-1.93655037353)*(1+2.*(dim-1.)/(o+1.))+(-1.63113832851)*dim
        ref[3,2,1,0]=(1.1770401643)*(1+2.*(dim-1.)/(o+1.))+(1.03181665809)*dim
        ref[3,2,1,1]=(-1.57622476164)*(1+2.*(dim-1.)/(o+1.))+(1.02511612685)*dim
        ref[3,2,2,0]=(-0.141399619749)*(1+2.*(dim-1.)/(o+1.))+(-1.06965410781)*dim
        ref[3,2,2,1]=(1.67083909164)*(1+2.*(dim-1.)/(o+1.))+(0.102145518531)*dim
        ref[3,3,0,0]=(-0.394782759355)*(1+2.*(dim-1.)/(o+1.))+(-1.24000053008)*dim
        ref[3,3,0,1]=(0.868960808118)*(1+2.*(dim-1.)/(o+1.))+(0.00201758379073)*dim
        ref[3,3,1,0]=(1.00000919667)*(1+2.*(dim-1.)/(o+1.))+(-0.17230041321)*dim
        ref[3,3,1,1]=(1.64713319694)*(1+2.*(dim-1.)/(o+1.))+(0.638129241907)*dim
        ref[3,3,2,0]=(0.924755217374)*(1+2.*(dim-1.)/(o+1.))+(-1.14240472392)*dim
        ref[3,3,2,1]=(0.539870030002)*(1+2.*(dim-1.)/(o+1.))+(-1.57885882521)*dim
        ref[3,4,0,0]=(-1.11180611147)*(1+2.*(dim-1.)/(o+1.))+(0.0614250843596)*dim
        ref[3,4,0,1]=(0.199061341795)*(1+2.*(dim-1.)/(o+1.))+(-0.154806515534)*dim
        ref[3,4,1,0]=(0.616447492955)*(1+2.*(dim-1.)/(o+1.))+(1.0075475758)*dim
        ref[3,4,1,1]=(1.13843338631)*(1+2.*(dim-1.)/(o+1.))+(-1.20318894006)*dim
        ref[3,4,2,0]=(0.190554236123)*(1+2.*(dim-1.)/(o+1.))+(1.47013574405)*dim
        ref[3,4,2,1]=(-1.31155106991)*(1+2.*(dim-1.)/(o+1.))+(1.04945905035)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.416878137535)*x[0]**o+(0.0809721296208)*x[0]+(-0.275084069478)*x[1]**o+(-0.704226656249)*x[1]
        ref=(0.141794068057)*(1+2.*(dim-1.)/(o+1.))+(-0.623254526628)*dim
      else:
        arg=(0.587142789605)*x[0]**o+(0.894018391929)*x[0]+(-0.417709558814)*x[1]**o+(0.0970658484968)*x[1]+(-0.840982437679)*x[2]**o+(-0.400242562091)*x[2]
        ref=(-0.671549206889)*(1+2.*(dim-1.)/(o+1.))+(0.590841678335)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.051335729494)*x[0]**o+(0.257057962305)*x[0]+(0.00794214446212)*x[1]**o+(0.52307365061)*x[1]
        arg[1]=(0.419915493803)*x[0]**o+(-0.0105826268931)*x[0]+(0.369274332127)*x[1]**o+(0.340796701053)*x[1]
        ref[0]=(0.0592778739561)*(1+2.*(dim-1.)/(o+1.))+(0.780131612915)*dim
        ref[1]=(0.789189825931)*(1+2.*(dim-1.)/(o+1.))+(0.33021407416)*dim
      else:
        arg[0]=(-0.483302491019)*x[0]**o+(0.223892398776)*x[0]+(-0.555612541317)*x[1]**o+(0.75535374219)*x[1]+(-0.208573228788)*x[2]**o+(0.479879212564)*x[2]
        arg[1]=(-0.012341185885)*x[0]**o+(-0.959474230066)*x[0]+(0.397671123434)*x[1]**o+(-0.120340696903)*x[1]+(-0.645757653205)*x[2]**o+(-0.216057959957)*x[2]
        ref[0]=(-1.24748826112)*(1+2.*(dim-1.)/(o+1.))+(1.45912535353)*dim
        ref[1]=(-0.260427715657)*(1+2.*(dim-1.)/(o+1.))+(-1.29587288693)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.493237934301)*x[0]**o+(-0.977176505302)*x[0]+(-0.837740403623)*x[1]**o+(0.204709065496)*x[1]
        arg[0,1]=(0.794064508577)*x[0]**o+(-0.917060945623)*x[0]+(0.0114185601474)*x[1]**o+(0.547108694374)*x[1]
        arg[0,2]=(-0.989778984923)*x[0]**o+(-0.499651095311)*x[0]+(-0.126817505183)*x[1]**o+(0.889077385524)*x[1]
        arg[0,3]=(-0.223453594493)*x[0]**o+(-0.17983229649)*x[0]+(0.352602269148)*x[1]**o+(0.00617724258273)*x[1]
        arg[0,4]=(0.401784356952)*x[0]**o+(0.839308128877)*x[0]+(0.322943402455)*x[1]**o+(0.770206521944)*x[1]
        arg[1,0]=(0.491015078355)*x[0]**o+(0.194909202321)*x[0]+(0.38986016491)*x[1]**o+(0.671106836372)*x[1]
        arg[1,1]=(0.200697856784)*x[0]**o+(0.389746053244)*x[0]+(0.850237988574)*x[1]**o+(-0.677750684271)*x[1]
        arg[1,2]=(0.300444129016)*x[0]**o+(-0.502811932395)*x[0]+(0.390751425646)*x[1]**o+(-0.0234058348564)*x[1]
        arg[1,3]=(-0.451410318068)*x[0]**o+(0.0959092789621)*x[0]+(0.364602756931)*x[1]**o+(-0.637883149677)*x[1]
        arg[1,4]=(0.84148898142)*x[0]**o+(0.447701396385)*x[0]+(-0.925960934959)*x[1]**o+(-0.216915201171)*x[1]
        arg[2,0]=(0.497685540053)*x[0]**o+(0.706445401439)*x[0]+(0.278688226804)*x[1]**o+(0.263560807934)*x[1]
        arg[2,1]=(0.846700886302)*x[0]**o+(0.794459987626)*x[0]+(0.206707548962)*x[1]**o+(0.536938178952)*x[1]
        arg[2,2]=(-0.825060225713)*x[0]**o+(0.00650901742391)*x[0]+(0.500591036904)*x[1]**o+(0.86251606869)*x[1]
        arg[2,3]=(-0.490203731083)*x[0]**o+(-0.758885551614)*x[0]+(0.44212368096)*x[1]**o+(0.422183463853)*x[1]
        arg[2,4]=(0.326180837967)*x[0]**o+(0.0511693925274)*x[0]+(0.450713817213)*x[1]**o+(-0.788372992203)*x[1]
        arg[3,0]=(0.427328335694)*x[0]**o+(-0.549570782632)*x[0]+(0.0484849428866)*x[1]**o+(-0.488921618371)*x[1]
        arg[3,1]=(-0.000395237589224)*x[0]**o+(-0.902889768871)*x[0]+(-0.991821787559)*x[1]**o+(-0.907842202962)*x[1]
        arg[3,2]=(0.737156280391)*x[0]**o+(0.104714673958)*x[0]+(0.102528971835)*x[1]**o+(-0.0982626427735)*x[1]
        arg[3,3]=(0.971460723075)*x[0]**o+(-0.759821212391)*x[0]+(0.385963754923)*x[1]**o+(-0.533924725996)*x[1]
        arg[3,4]=(0.25440424638)*x[0]**o+(-0.82432588605)*x[0]+(0.759285911224)*x[1]**o+(0.328467600746)*x[1]
        ref[0,0]=(-1.33097833792)*(1+2.*(dim-1.)/(o+1.))+(-0.772467439806)*dim
        ref[0,1]=(0.805483068724)*(1+2.*(dim-1.)/(o+1.))+(-0.36995225125)*dim
        ref[0,2]=(-1.11659649011)*(1+2.*(dim-1.)/(o+1.))+(0.389426290213)*dim
        ref[0,3]=(0.129148674655)*(1+2.*(dim-1.)/(o+1.))+(-0.173655053907)*dim
        ref[0,4]=(0.724727759407)*(1+2.*(dim-1.)/(o+1.))+(1.60951465082)*dim
        ref[1,0]=(0.880875243265)*(1+2.*(dim-1.)/(o+1.))+(0.866016038693)*dim
        ref[1,1]=(1.05093584536)*(1+2.*(dim-1.)/(o+1.))+(-0.288004631027)*dim
        ref[1,2]=(0.691195554662)*(1+2.*(dim-1.)/(o+1.))+(-0.526217767251)*dim
        ref[1,3]=(-0.0868075611369)*(1+2.*(dim-1.)/(o+1.))+(-0.541973870715)*dim
        ref[1,4]=(-0.0844719535392)*(1+2.*(dim-1.)/(o+1.))+(0.230786195214)*dim
        ref[2,0]=(0.776373766857)*(1+2.*(dim-1.)/(o+1.))+(0.970006209374)*dim
        ref[2,1]=(1.05340843526)*(1+2.*(dim-1.)/(o+1.))+(1.33139816658)*dim
        ref[2,2]=(-0.324469188808)*(1+2.*(dim-1.)/(o+1.))+(0.869025086114)*dim
        ref[2,3]=(-0.0480800501232)*(1+2.*(dim-1.)/(o+1.))+(-0.336702087761)*dim
        ref[2,4]=(0.776894655179)*(1+2.*(dim-1.)/(o+1.))+(-0.737203599675)*dim
        ref[3,0]=(0.47581327858)*(1+2.*(dim-1.)/(o+1.))+(-1.038492401)*dim
        ref[3,1]=(-0.992217025148)*(1+2.*(dim-1.)/(o+1.))+(-1.81073197183)*dim
        ref[3,2]=(0.839685252226)*(1+2.*(dim-1.)/(o+1.))+(0.00645203118434)*dim
        ref[3,3]=(1.357424478)*(1+2.*(dim-1.)/(o+1.))+(-1.29374593839)*dim
        ref[3,4]=(1.0136901576)*(1+2.*(dim-1.)/(o+1.))+(-0.495858285305)*dim
      else:
        arg[0,0]=(-0.543785449843)*x[0]**o+(-0.593471310422)*x[0]+(0.223732084513)*x[1]**o+(-0.126390494354)*x[1]+(0.602561108649)*x[2]**o+(-0.879650196727)*x[2]
        arg[0,1]=(-0.373708985284)*x[0]**o+(0.274723212925)*x[0]+(0.0901068561926)*x[1]**o+(0.0705878267418)*x[1]+(-0.87124855534)*x[2]**o+(-0.342156987834)*x[2]
        arg[0,2]=(0.197662792758)*x[0]**o+(0.993389348528)*x[0]+(0.300080063659)*x[1]**o+(-0.87329423474)*x[1]+(-0.708861668092)*x[2]**o+(-0.145567510817)*x[2]
        arg[0,3]=(-0.641680604707)*x[0]**o+(0.502870155472)*x[0]+(0.726460257571)*x[1]**o+(-0.796246941548)*x[1]+(-0.034138694789)*x[2]**o+(-0.756773445728)*x[2]
        arg[0,4]=(-0.910301138852)*x[0]**o+(0.370603440687)*x[0]+(0.30908632679)*x[1]**o+(0.399307258883)*x[1]+(0.734329887794)*x[2]**o+(-0.105355323212)*x[2]
        arg[1,0]=(-0.998162269861)*x[0]**o+(0.744436497607)*x[0]+(0.474991680035)*x[1]**o+(-0.523297766774)*x[1]+(-0.147411103004)*x[2]**o+(-0.366260797475)*x[2]
        arg[1,1]=(-0.895068058812)*x[0]**o+(-0.543026567202)*x[0]+(0.681910579939)*x[1]**o+(0.839238385122)*x[1]+(-0.943454697555)*x[2]**o+(-0.942882687085)*x[2]
        arg[1,2]=(-0.0327934757673)*x[0]**o+(-0.551060110327)*x[0]+(0.0246419122114)*x[1]**o+(0.290492943846)*x[1]+(-0.81629260256)*x[2]**o+(-0.0703654573398)*x[2]
        arg[1,3]=(0.0101368428324)*x[0]**o+(0.169228100407)*x[0]+(-0.0720675504637)*x[1]**o+(0.854569517064)*x[1]+(-0.185495614033)*x[2]**o+(0.159402892807)*x[2]
        arg[1,4]=(-0.879056526056)*x[0]**o+(-0.499865495878)*x[0]+(0.389281340553)*x[1]**o+(0.208323326648)*x[1]+(0.201495185869)*x[2]**o+(0.287693342543)*x[2]
        arg[2,0]=(-0.0980367959573)*x[0]**o+(-0.743908923973)*x[0]+(0.612332291133)*x[1]**o+(-0.833162883242)*x[1]+(-0.149234146487)*x[2]**o+(-0.265352382043)*x[2]
        arg[2,1]=(-0.731053054407)*x[0]**o+(-0.328824749845)*x[0]+(-0.270771833154)*x[1]**o+(-0.924718425874)*x[1]+(-0.333737248999)*x[2]**o+(-0.437707530432)*x[2]
        arg[2,2]=(0.0568240255423)*x[0]**o+(0.889929679474)*x[0]+(-0.550346007257)*x[1]**o+(-0.276114383355)*x[1]+(0.818856335741)*x[2]**o+(0.949830759195)*x[2]
        arg[2,3]=(-0.487278099196)*x[0]**o+(0.621833922878)*x[0]+(-0.641187646288)*x[1]**o+(-0.62187991518)*x[1]+(0.193101177186)*x[2]**o+(0.574039256278)*x[2]
        arg[2,4]=(0.411046922073)*x[0]**o+(-0.10196277248)*x[0]+(-0.880065306034)*x[1]**o+(0.629582927604)*x[1]+(-0.364618567951)*x[2]**o+(0.130293365676)*x[2]
        arg[3,0]=(-0.625244224555)*x[0]**o+(0.50345243521)*x[0]+(0.701616120451)*x[1]**o+(0.44401843794)*x[1]+(-0.881068054195)*x[2]**o+(0.337405641389)*x[2]
        arg[3,1]=(0.244113319139)*x[0]**o+(0.739760310044)*x[0]+(0.439528183979)*x[1]**o+(0.808496622595)*x[1]+(0.432069187682)*x[2]**o+(-0.106394550529)*x[2]
        arg[3,2]=(-0.200973139796)*x[0]**o+(-0.373932231384)*x[0]+(-0.438946407503)*x[1]**o+(0.543079830762)*x[1]+(-0.499599010695)*x[2]**o+(0.501967674296)*x[2]
        arg[3,3]=(-0.128263000609)*x[0]**o+(-0.408181851435)*x[0]+(0.403315054177)*x[1]**o+(0.75067593555)*x[1]+(0.361607987217)*x[2]**o+(0.269483801463)*x[2]
        arg[3,4]=(0.821969994275)*x[0]**o+(0.350771424951)*x[0]+(-0.761235507511)*x[1]**o+(0.148329827876)*x[1]+(-0.918493845231)*x[2]**o+(0.592094636272)*x[2]
        ref[0,0]=(0.282507743319)*(1+2.*(dim-1.)/(o+1.))+(-1.5995120015)*dim
        ref[0,1]=(-1.15485068443)*(1+2.*(dim-1.)/(o+1.))+(0.00315405183214)*dim
        ref[0,2]=(-0.211118811675)*(1+2.*(dim-1.)/(o+1.))+(-0.0254723970291)*dim
        ref[0,3]=(0.0506409580755)*(1+2.*(dim-1.)/(o+1.))+(-1.0501502318)*dim
        ref[0,4]=(0.133115075732)*(1+2.*(dim-1.)/(o+1.))+(0.664555376358)*dim
        ref[1,0]=(-0.67058169283)*(1+2.*(dim-1.)/(o+1.))+(-0.145122066642)*dim
        ref[1,1]=(-1.15661217643)*(1+2.*(dim-1.)/(o+1.))+(-0.646670869164)*dim
        ref[1,2]=(-0.824444166116)*(1+2.*(dim-1.)/(o+1.))+(-0.330932623821)*dim
        ref[1,3]=(-0.247426321664)*(1+2.*(dim-1.)/(o+1.))+(1.18320051028)*dim
        ref[1,4]=(-0.288279999634)*(1+2.*(dim-1.)/(o+1.))+(-0.00384882668765)*dim
        ref[2,0]=(0.365061348688)*(1+2.*(dim-1.)/(o+1.))+(-1.84242418926)*dim
        ref[2,1]=(-1.33556213656)*(1+2.*(dim-1.)/(o+1.))+(-1.69125070615)*dim
        ref[2,2]=(0.325334354026)*(1+2.*(dim-1.)/(o+1.))+(1.56364605531)*dim
        ref[2,3]=(-0.935364568298)*(1+2.*(dim-1.)/(o+1.))+(0.573993263976)*dim
        ref[2,4]=(-0.833636951913)*(1+2.*(dim-1.)/(o+1.))+(0.6579135208)*dim
        ref[3,0]=(-0.804696158299)*(1+2.*(dim-1.)/(o+1.))+(1.28487651454)*dim
        ref[3,1]=(1.1157106908)*(1+2.*(dim-1.)/(o+1.))+(1.44186238211)*dim
        ref[3,2]=(-1.13951855799)*(1+2.*(dim-1.)/(o+1.))+(0.671115273674)*dim
        ref[3,3]=(0.636660040784)*(1+2.*(dim-1.)/(o+1.))+(0.611977885578)*dim
        ref[3,4]=(-0.857759358467)*(1+2.*(dim-1.)/(o+1.))+(1.0911958891)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.331954160625)*x[0]**o+(0.713423346586)*x[0]+(0.922333771405)*x[1]**o+(-0.306799323245)*x[1]
        arg[0,0,1]=(-0.837278273696)*x[0]**o+(-0.322068356249)*x[0]+(-0.952056283446)*x[1]**o+(0.78147215355)*x[1]
        arg[0,1,0]=(-0.71217197633)*x[0]**o+(-0.222121067189)*x[0]+(-0.740851323426)*x[1]**o+(-0.664415003166)*x[1]
        arg[0,1,1]=(0.304797844524)*x[0]**o+(-0.644642249615)*x[0]+(-0.942312793933)*x[1]**o+(0.758857258357)*x[1]
        arg[1,0,0]=(-0.978734634026)*x[0]**o+(0.697829375729)*x[0]+(-0.161023313194)*x[1]**o+(0.61960280308)*x[1]
        arg[1,0,1]=(0.65091302745)*x[0]**o+(-0.411592147008)*x[0]+(0.748017319704)*x[1]**o+(0.751661257345)*x[1]
        arg[1,1,0]=(-0.752749816617)*x[0]**o+(-0.889417028079)*x[0]+(0.187075769961)*x[1]**o+(0.96605423194)*x[1]
        arg[1,1,1]=(-0.791276410919)*x[0]**o+(-0.66910236639)*x[0]+(-0.107283654349)*x[1]**o+(0.556315348977)*x[1]
        arg[2,0,0]=(0.0717781211643)*x[0]**o+(-0.895974517834)*x[0]+(-0.388350799289)*x[1]**o+(0.55541171641)*x[1]
        arg[2,0,1]=(0.886475701949)*x[0]**o+(0.0794480579219)*x[0]+(-0.77671944551)*x[1]**o+(-0.29611930088)*x[1]
        arg[2,1,0]=(0.891530897911)*x[0]**o+(-0.767892439984)*x[0]+(-0.692871902002)*x[1]**o+(0.296346783431)*x[1]
        arg[2,1,1]=(0.904417221616)*x[0]**o+(-0.379208335246)*x[0]+(0.491675297059)*x[1]**o+(0.994131453216)*x[1]
        arg[3,0,0]=(-0.278750282129)*x[0]**o+(-0.889235749208)*x[0]+(-0.970025256676)*x[1]**o+(0.459265023633)*x[1]
        arg[3,0,1]=(-0.718548177793)*x[0]**o+(-0.801618785436)*x[0]+(-0.0489351007933)*x[1]**o+(-0.0331281021404)*x[1]
        arg[3,1,0]=(0.838645233036)*x[0]**o+(0.174830460073)*x[0]+(0.343887140274)*x[1]**o+(0.475643625956)*x[1]
        arg[3,1,1]=(-0.376629360434)*x[0]**o+(0.965850026388)*x[0]+(-0.62284011199)*x[1]**o+(-0.548261522438)*x[1]
        arg[4,0,0]=(-0.747109622101)*x[0]**o+(-0.793286186527)*x[0]+(0.700502366132)*x[1]**o+(0.708459489695)*x[1]
        arg[4,0,1]=(-0.372920704339)*x[0]**o+(-0.856662712404)*x[0]+(0.735191558102)*x[1]**o+(0.451959287084)*x[1]
        arg[4,1,0]=(-0.825227447648)*x[0]**o+(-0.978701927456)*x[0]+(0.856562109765)*x[1]**o+(0.226820057686)*x[1]
        arg[4,1,1]=(-0.883951567752)*x[0]**o+(0.346114896878)*x[0]+(0.22103721575)*x[1]**o+(0.766907412924)*x[1]
        arg[5,0,0]=(0.440140408264)*x[0]**o+(0.576442002932)*x[0]+(0.181136518153)*x[1]**o+(0.467290325887)*x[1]
        arg[5,0,1]=(0.323653297993)*x[0]**o+(-0.0169080219964)*x[0]+(-0.790889359324)*x[1]**o+(-0.752609210154)*x[1]
        arg[5,1,0]=(-0.771859095701)*x[0]**o+(0.392074466446)*x[0]+(0.672832544535)*x[1]**o+(-0.518196376899)*x[1]
        arg[5,1,1]=(-0.714106835785)*x[0]**o+(-0.392974811152)*x[0]+(0.477349412757)*x[1]**o+(-0.978738343032)*x[1]
        ref[0,0,0]=(0.59037961078)*(1+2.*(dim-1.)/(o+1.))+(0.406624023341)*dim
        ref[0,0,1]=(-1.78933455714)*(1+2.*(dim-1.)/(o+1.))+(0.459403797301)*dim
        ref[0,1,0]=(-1.45302329976)*(1+2.*(dim-1.)/(o+1.))+(-0.886536070355)*dim
        ref[0,1,1]=(-0.637514949409)*(1+2.*(dim-1.)/(o+1.))+(0.114215008741)*dim
        ref[1,0,0]=(-1.13975794722)*(1+2.*(dim-1.)/(o+1.))+(1.31743217881)*dim
        ref[1,0,1]=(1.39893034715)*(1+2.*(dim-1.)/(o+1.))+(0.340069110337)*dim
        ref[1,1,0]=(-0.565674046656)*(1+2.*(dim-1.)/(o+1.))+(0.076637203861)*dim
        ref[1,1,1]=(-0.898560065268)*(1+2.*(dim-1.)/(o+1.))+(-0.112787017413)*dim
        ref[2,0,0]=(-0.316572678125)*(1+2.*(dim-1.)/(o+1.))+(-0.340562801424)*dim
        ref[2,0,1]=(0.10975625644)*(1+2.*(dim-1.)/(o+1.))+(-0.216671242958)*dim
        ref[2,1,0]=(0.198658995909)*(1+2.*(dim-1.)/(o+1.))+(-0.471545656553)*dim
        ref[2,1,1]=(1.39609251868)*(1+2.*(dim-1.)/(o+1.))+(0.614923117971)*dim
        ref[3,0,0]=(-1.2487755388)*(1+2.*(dim-1.)/(o+1.))+(-0.429970725575)*dim
        ref[3,0,1]=(-0.767483278586)*(1+2.*(dim-1.)/(o+1.))+(-0.834746887577)*dim
        ref[3,1,0]=(1.18253237331)*(1+2.*(dim-1.)/(o+1.))+(0.650474086029)*dim
        ref[3,1,1]=(-0.999469472424)*(1+2.*(dim-1.)/(o+1.))+(0.41758850395)*dim
        ref[4,0,0]=(-0.0466072559697)*(1+2.*(dim-1.)/(o+1.))+(-0.0848266968319)*dim
        ref[4,0,1]=(0.362270853763)*(1+2.*(dim-1.)/(o+1.))+(-0.404703425319)*dim
        ref[4,1,0]=(0.0313346621166)*(1+2.*(dim-1.)/(o+1.))+(-0.75188186977)*dim
        ref[4,1,1]=(-0.662914352001)*(1+2.*(dim-1.)/(o+1.))+(1.1130223098)*dim
        ref[5,0,0]=(0.621276926417)*(1+2.*(dim-1.)/(o+1.))+(1.04373232882)*dim
        ref[5,0,1]=(-0.467236061331)*(1+2.*(dim-1.)/(o+1.))+(-0.76951723215)*dim
        ref[5,1,0]=(-0.0990265511653)*(1+2.*(dim-1.)/(o+1.))+(-0.126121910453)*dim
        ref[5,1,1]=(-0.236757423028)*(1+2.*(dim-1.)/(o+1.))+(-1.37171315418)*dim
      else:
        arg[0,0,0]=(-0.676850371385)*x[0]**o+(0.343194249293)*x[0]+(-0.289569238124)*x[1]**o+(0.546143025789)*x[1]+(-0.292429216951)*x[2]**o+(0.100638805907)*x[2]
        arg[0,0,1]=(-0.558403199885)*x[0]**o+(0.369265961569)*x[0]+(0.098539502761)*x[1]**o+(-0.555747072012)*x[1]+(0.0142505043744)*x[2]**o+(0.110677590143)*x[2]
        arg[0,1,0]=(0.480265324249)*x[0]**o+(-0.0801799981418)*x[0]+(-0.659040110213)*x[1]**o+(0.315685251032)*x[1]+(0.146222780181)*x[2]**o+(-0.307845817615)*x[2]
        arg[0,1,1]=(-0.469643195547)*x[0]**o+(0.631180947887)*x[0]+(0.0873749097691)*x[1]**o+(-0.907805635326)*x[1]+(-0.445238378297)*x[2]**o+(-0.583750385876)*x[2]
        arg[1,0,0]=(-0.519648347947)*x[0]**o+(-0.580046561107)*x[0]+(-0.474702629402)*x[1]**o+(-0.76698196707)*x[1]+(0.631290211156)*x[2]**o+(0.283150656338)*x[2]
        arg[1,0,1]=(0.828049802595)*x[0]**o+(-0.0996271417739)*x[0]+(0.13247948046)*x[1]**o+(0.64657332105)*x[1]+(0.901448888477)*x[2]**o+(0.895379597076)*x[2]
        arg[1,1,0]=(-0.197367099356)*x[0]**o+(-0.933210661759)*x[0]+(-0.0134349749548)*x[1]**o+(-0.651111981269)*x[1]+(0.256945439519)*x[2]**o+(0.528412689094)*x[2]
        arg[1,1,1]=(-0.922943172982)*x[0]**o+(-0.883102479542)*x[0]+(-0.0037345936911)*x[1]**o+(-0.598256159261)*x[1]+(-0.8562089409)*x[2]**o+(-0.298502617143)*x[2]
        arg[2,0,0]=(0.499285053237)*x[0]**o+(-0.347877602065)*x[0]+(0.783117300662)*x[1]**o+(0.0413016895448)*x[1]+(0.22555777414)*x[2]**o+(0.089246166469)*x[2]
        arg[2,0,1]=(-0.180602931202)*x[0]**o+(0.883958054663)*x[0]+(0.904974519387)*x[1]**o+(-0.169618307434)*x[1]+(0.361963890467)*x[2]**o+(-0.364236450353)*x[2]
        arg[2,1,0]=(0.332239090882)*x[0]**o+(-0.392239851855)*x[0]+(0.905142609029)*x[1]**o+(-0.471493540964)*x[1]+(-0.305327788141)*x[2]**o+(0.581213078834)*x[2]
        arg[2,1,1]=(0.481276189253)*x[0]**o+(0.645400809127)*x[0]+(-0.840844470697)*x[1]**o+(-0.47587396865)*x[1]+(0.134145050668)*x[2]**o+(-0.315081369951)*x[2]
        arg[3,0,0]=(-0.480497217552)*x[0]**o+(0.595345393671)*x[0]+(0.776315451392)*x[1]**o+(-0.396110510671)*x[1]+(0.0437422791996)*x[2]**o+(0.348675326203)*x[2]
        arg[3,0,1]=(0.867117243092)*x[0]**o+(-0.222483151231)*x[0]+(0.0856439550139)*x[1]**o+(0.355855801395)*x[1]+(-0.779380984653)*x[2]**o+(0.799440639884)*x[2]
        arg[3,1,0]=(-0.869234399353)*x[0]**o+(-0.747500724089)*x[0]+(-0.83154206968)*x[1]**o+(-0.771906212766)*x[1]+(-0.188734869619)*x[2]**o+(0.7312382704)*x[2]
        arg[3,1,1]=(-0.729677636148)*x[0]**o+(-0.358066533942)*x[0]+(-0.342651052704)*x[1]**o+(-0.531453801913)*x[1]+(-0.898300547827)*x[2]**o+(0.0387220517653)*x[2]
        arg[4,0,0]=(-0.380489041999)*x[0]**o+(0.517823660835)*x[0]+(0.722149782685)*x[1]**o+(0.510169845258)*x[1]+(-0.979340093133)*x[2]**o+(0.846050053391)*x[2]
        arg[4,0,1]=(0.390895580804)*x[0]**o+(0.122461788897)*x[0]+(-0.809243912452)*x[1]**o+(0.871689879803)*x[1]+(-0.991689466458)*x[2]**o+(-0.0373953689911)*x[2]
        arg[4,1,0]=(0.00670611033858)*x[0]**o+(-0.700084486793)*x[0]+(0.365358002826)*x[1]**o+(-0.928347441251)*x[1]+(-0.675081280461)*x[2]**o+(-0.262871553902)*x[2]
        arg[4,1,1]=(-0.566521655794)*x[0]**o+(-0.23325381374)*x[0]+(0.744538505444)*x[1]**o+(0.263970013855)*x[1]+(0.360984907343)*x[2]**o+(0.704696800059)*x[2]
        arg[5,0,0]=(0.754229123633)*x[0]**o+(-0.523353302691)*x[0]+(0.404181763418)*x[1]**o+(0.175414644029)*x[1]+(0.627455124887)*x[2]**o+(-0.450163514418)*x[2]
        arg[5,0,1]=(0.804547370453)*x[0]**o+(-0.748034210243)*x[0]+(-0.943261499953)*x[1]**o+(0.360715335834)*x[1]+(-0.952785406204)*x[2]**o+(0.188838771539)*x[2]
        arg[5,1,0]=(0.653730390007)*x[0]**o+(0.663746229033)*x[0]+(0.454911944273)*x[1]**o+(-0.548110740107)*x[1]+(0.799171265206)*x[2]**o+(-0.646356900812)*x[2]
        arg[5,1,1]=(0.533464266683)*x[0]**o+(-0.0650985333228)*x[0]+(-0.599880569232)*x[1]**o+(0.645120695827)*x[1]+(0.985316456657)*x[2]**o+(0.527848503917)*x[2]
        ref[0,0,0]=(-1.25884882646)*(1+2.*(dim-1.)/(o+1.))+(0.989976080988)*dim
        ref[0,0,1]=(-0.445613192749)*(1+2.*(dim-1.)/(o+1.))+(-0.0758035202994)*dim
        ref[0,1,0]=(-0.0325520057832)*(1+2.*(dim-1.)/(o+1.))+(-0.0723405647253)*dim
        ref[0,1,1]=(-0.827506664075)*(1+2.*(dim-1.)/(o+1.))+(-0.860375073315)*dim
        ref[1,0,0]=(-0.363060766193)*(1+2.*(dim-1.)/(o+1.))+(-1.06387787184)*dim
        ref[1,0,1]=(1.86197817153)*(1+2.*(dim-1.)/(o+1.))+(1.44232577635)*dim
        ref[1,1,0]=(0.0461433652081)*(1+2.*(dim-1.)/(o+1.))+(-1.05590995393)*dim
        ref[1,1,1]=(-1.78288670757)*(1+2.*(dim-1.)/(o+1.))+(-1.77986125595)*dim
        ref[2,0,0]=(1.50796012804)*(1+2.*(dim-1.)/(o+1.))+(-0.217329746051)*dim
        ref[2,0,1]=(1.08633547865)*(1+2.*(dim-1.)/(o+1.))+(0.350103296876)*dim
        ref[2,1,0]=(0.932053911771)*(1+2.*(dim-1.)/(o+1.))+(-0.282520313985)*dim
        ref[2,1,1]=(-0.225423230776)*(1+2.*(dim-1.)/(o+1.))+(-0.145554529474)*dim
        ref[3,0,0]=(0.33956051304)*(1+2.*(dim-1.)/(o+1.))+(0.547910209204)*dim
        ref[3,0,1]=(0.173380213454)*(1+2.*(dim-1.)/(o+1.))+(0.932813290049)*dim
        ref[3,1,0]=(-1.88951133865)*(1+2.*(dim-1.)/(o+1.))+(-0.788168666455)*dim
        ref[3,1,1]=(-1.97062923668)*(1+2.*(dim-1.)/(o+1.))+(-0.850798284089)*dim
        ref[4,0,0]=(-0.637679352447)*(1+2.*(dim-1.)/(o+1.))+(1.87404355948)*dim
        ref[4,0,1]=(-1.41003779811)*(1+2.*(dim-1.)/(o+1.))+(0.956756299709)*dim
        ref[4,1,0]=(-0.303017167297)*(1+2.*(dim-1.)/(o+1.))+(-1.89130348195)*dim
        ref[4,1,1]=(0.539001756993)*(1+2.*(dim-1.)/(o+1.))+(0.735413000173)*dim
        ref[5,0,0]=(1.78586601194)*(1+2.*(dim-1.)/(o+1.))+(-0.79810217308)*dim
        ref[5,0,1]=(-1.0914995357)*(1+2.*(dim-1.)/(o+1.))+(-0.198480102871)*dim
        ref[5,1,0]=(1.90781359949)*(1+2.*(dim-1.)/(o+1.))+(-0.530721411887)*dim
        ref[5,1,1]=(0.918900154108)*(1+2.*(dim-1.)/(o+1.))+(1.10787066642)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.181966612335)*x[0]**o+(-0.461883712153)*x[0]+(-0.420217253531)*x[1]**o+(0.52197109374)*x[1]
        arg[0,0,0,1]=(-0.122704378373)*x[0]**o+(0.549708706786)*x[0]+(-0.728108908074)*x[1]**o+(-0.839262748545)*x[1]
        arg[0,0,1,0]=(0.58557741427)*x[0]**o+(-0.970633305521)*x[0]+(0.215513381389)*x[1]**o+(-0.178877868196)*x[1]
        arg[0,0,1,1]=(0.646118894595)*x[0]**o+(0.839917605008)*x[0]+(-0.608843960266)*x[1]**o+(-0.840842208189)*x[1]
        arg[0,0,2,0]=(-0.126004594384)*x[0]**o+(0.663487355648)*x[0]+(0.801751176464)*x[1]**o+(-0.76620842256)*x[1]
        arg[0,0,2,1]=(-0.244425023768)*x[0]**o+(0.736809513316)*x[0]+(0.736776514098)*x[1]**o+(0.382648973627)*x[1]
        arg[0,1,0,0]=(-0.264460763052)*x[0]**o+(-0.325868872518)*x[0]+(-0.0636203008538)*x[1]**o+(-0.398059740187)*x[1]
        arg[0,1,0,1]=(0.443639751435)*x[0]**o+(0.523903519493)*x[0]+(-0.222575248757)*x[1]**o+(-0.168203019401)*x[1]
        arg[0,1,1,0]=(-0.329523829258)*x[0]**o+(-0.908408437436)*x[0]+(0.695441587034)*x[1]**o+(-0.236957609006)*x[1]
        arg[0,1,1,1]=(-0.57363301023)*x[0]**o+(-0.168677045141)*x[0]+(0.455235475098)*x[1]**o+(-0.498803217861)*x[1]
        arg[0,1,2,0]=(0.09906404243)*x[0]**o+(0.48682122702)*x[0]+(-0.250940180366)*x[1]**o+(-0.641013418008)*x[1]
        arg[0,1,2,1]=(-0.963193555161)*x[0]**o+(0.129868691037)*x[0]+(0.821192174481)*x[1]**o+(-0.245576703223)*x[1]
        arg[0,2,0,0]=(-0.867410251576)*x[0]**o+(0.170194043006)*x[0]+(0.837669366536)*x[1]**o+(0.285772890787)*x[1]
        arg[0,2,0,1]=(0.767089114154)*x[0]**o+(-0.425174482112)*x[0]+(-0.460840645159)*x[1]**o+(-0.314295694392)*x[1]
        arg[0,2,1,0]=(-0.932934920269)*x[0]**o+(0.399552332498)*x[0]+(-0.137764102275)*x[1]**o+(-0.0375128952657)*x[1]
        arg[0,2,1,1]=(0.507999826854)*x[0]**o+(-0.60826518686)*x[0]+(0.766838061511)*x[1]**o+(0.800723296773)*x[1]
        arg[0,2,2,0]=(0.372064864992)*x[0]**o+(0.29029476011)*x[0]+(0.203299629846)*x[1]**o+(-0.386357855884)*x[1]
        arg[0,2,2,1]=(0.661268311956)*x[0]**o+(0.60257115577)*x[0]+(0.88881711536)*x[1]**o+(0.335431084294)*x[1]
        arg[0,3,0,0]=(-0.756295024079)*x[0]**o+(-0.982472476783)*x[0]+(-0.264060825724)*x[1]**o+(0.082684535426)*x[1]
        arg[0,3,0,1]=(0.514580120205)*x[0]**o+(0.524816925461)*x[0]+(-0.745096223094)*x[1]**o+(-0.0796881815577)*x[1]
        arg[0,3,1,0]=(0.597556449974)*x[0]**o+(0.527680495979)*x[0]+(0.517638167762)*x[1]**o+(-0.357836534374)*x[1]
        arg[0,3,1,1]=(0.354366560101)*x[0]**o+(-0.617697325436)*x[0]+(0.339071431286)*x[1]**o+(0.299934132319)*x[1]
        arg[0,3,2,0]=(0.26563455123)*x[0]**o+(0.821495614519)*x[0]+(0.177496963332)*x[1]**o+(-0.154521746257)*x[1]
        arg[0,3,2,1]=(0.983477388632)*x[0]**o+(0.524806318175)*x[0]+(-0.604096007297)*x[1]**o+(0.350360164328)*x[1]
        arg[0,4,0,0]=(0.311369778768)*x[0]**o+(-0.648416457513)*x[0]+(0.463469290711)*x[1]**o+(0.0719628828709)*x[1]
        arg[0,4,0,1]=(-0.743081056195)*x[0]**o+(-0.803573973461)*x[0]+(-0.766982068691)*x[1]**o+(0.387240020398)*x[1]
        arg[0,4,1,0]=(-0.0715117411345)*x[0]**o+(-0.341871715212)*x[0]+(0.0593378446291)*x[1]**o+(-0.10183864979)*x[1]
        arg[0,4,1,1]=(0.538850329978)*x[0]**o+(0.0280201497914)*x[0]+(0.350206859589)*x[1]**o+(0.316831165676)*x[1]
        arg[0,4,2,0]=(-0.327852385434)*x[0]**o+(0.408866306659)*x[0]+(0.940378517318)*x[1]**o+(0.578309623908)*x[1]
        arg[0,4,2,1]=(0.378380112483)*x[0]**o+(0.696828716067)*x[0]+(-0.908233000582)*x[1]**o+(-0.328611483675)*x[1]
        arg[1,0,0,0]=(0.753991558436)*x[0]**o+(0.0542891140783)*x[0]+(-0.428934638249)*x[1]**o+(0.318155063446)*x[1]
        arg[1,0,0,1]=(-0.387609085343)*x[0]**o+(0.711023478158)*x[0]+(-0.969026331169)*x[1]**o+(0.456175434422)*x[1]
        arg[1,0,1,0]=(-0.468504469917)*x[0]**o+(-0.314756281356)*x[0]+(0.0145647044588)*x[1]**o+(0.074035761055)*x[1]
        arg[1,0,1,1]=(0.240749620446)*x[0]**o+(0.906736060537)*x[0]+(0.391320123368)*x[1]**o+(-0.261639387239)*x[1]
        arg[1,0,2,0]=(-0.461436333094)*x[0]**o+(-0.809258428817)*x[0]+(0.0670777162266)*x[1]**o+(-0.445304287721)*x[1]
        arg[1,0,2,1]=(-0.569794448172)*x[0]**o+(-0.0624052432724)*x[0]+(-0.759840810858)*x[1]**o+(-0.609106375016)*x[1]
        arg[1,1,0,0]=(0.751570143287)*x[0]**o+(0.78189919599)*x[0]+(-0.103246458909)*x[1]**o+(0.732156635312)*x[1]
        arg[1,1,0,1]=(-0.0126214545951)*x[0]**o+(-0.612853966263)*x[0]+(-0.238555698245)*x[1]**o+(-0.668009933698)*x[1]
        arg[1,1,1,0]=(0.816325838743)*x[0]**o+(0.545518357738)*x[0]+(0.893296287925)*x[1]**o+(0.694579919864)*x[1]
        arg[1,1,1,1]=(0.927638834795)*x[0]**o+(0.750347688466)*x[0]+(-0.15924462002)*x[1]**o+(-0.952995938892)*x[1]
        arg[1,1,2,0]=(0.970761253678)*x[0]**o+(0.205308852313)*x[0]+(0.909209906751)*x[1]**o+(-0.178635180174)*x[1]
        arg[1,1,2,1]=(-0.632065128479)*x[0]**o+(-0.300905398347)*x[0]+(0.865317638002)*x[1]**o+(0.886398250669)*x[1]
        arg[1,2,0,0]=(-0.0529293620539)*x[0]**o+(-0.909022003179)*x[0]+(0.235231916345)*x[1]**o+(-0.623380752415)*x[1]
        arg[1,2,0,1]=(0.873344987616)*x[0]**o+(0.144231546529)*x[0]+(0.307375566381)*x[1]**o+(-0.201617507084)*x[1]
        arg[1,2,1,0]=(-0.957945047658)*x[0]**o+(-0.432955474381)*x[0]+(0.0742288164869)*x[1]**o+(-0.106080430366)*x[1]
        arg[1,2,1,1]=(-0.664227561512)*x[0]**o+(0.0821007781774)*x[0]+(-0.541506076231)*x[1]**o+(-0.88307957937)*x[1]
        arg[1,2,2,0]=(0.754702675232)*x[0]**o+(0.175176916913)*x[0]+(0.519607893979)*x[1]**o+(-0.24129526204)*x[1]
        arg[1,2,2,1]=(-0.415660998528)*x[0]**o+(-0.964071393069)*x[0]+(0.458656303526)*x[1]**o+(0.752153826846)*x[1]
        arg[1,3,0,0]=(-0.15638393187)*x[0]**o+(0.0828824719)*x[0]+(-0.0846205011509)*x[1]**o+(-0.438939463736)*x[1]
        arg[1,3,0,1]=(0.222994482074)*x[0]**o+(-0.488426101986)*x[0]+(-0.799262651825)*x[1]**o+(0.0638880385288)*x[1]
        arg[1,3,1,0]=(0.690529298963)*x[0]**o+(-0.845248154386)*x[0]+(0.748635110096)*x[1]**o+(0.820882250869)*x[1]
        arg[1,3,1,1]=(-0.0989162396328)*x[0]**o+(-0.171252734498)*x[0]+(-0.363447812361)*x[1]**o+(-0.06937701403)*x[1]
        arg[1,3,2,0]=(-0.739647298124)*x[0]**o+(-0.83384135572)*x[0]+(0.546124768813)*x[1]**o+(0.415761104064)*x[1]
        arg[1,3,2,1]=(0.859487140106)*x[0]**o+(0.481214854724)*x[0]+(0.788685241166)*x[1]**o+(-0.776410333263)*x[1]
        arg[1,4,0,0]=(-0.312428038681)*x[0]**o+(-0.178596293803)*x[0]+(-0.229770546807)*x[1]**o+(-0.43765633326)*x[1]
        arg[1,4,0,1]=(-0.974305286188)*x[0]**o+(-0.978154435248)*x[0]+(-0.671799350916)*x[1]**o+(0.645287291694)*x[1]
        arg[1,4,1,0]=(-0.219711371151)*x[0]**o+(-0.432784911532)*x[0]+(0.552940388883)*x[1]**o+(-0.622742205614)*x[1]
        arg[1,4,1,1]=(0.677289749287)*x[0]**o+(-0.92825132284)*x[0]+(0.818767358259)*x[1]**o+(-0.426000012305)*x[1]
        arg[1,4,2,0]=(-0.507474696573)*x[0]**o+(0.0882151625837)*x[0]+(0.826611187078)*x[1]**o+(0.157884229299)*x[1]
        arg[1,4,2,1]=(0.902398378671)*x[0]**o+(-0.372679984773)*x[0]+(-0.998272135687)*x[1]**o+(0.292294917465)*x[1]
        arg[2,0,0,0]=(0.699035433674)*x[0]**o+(0.380640017665)*x[0]+(-0.650735771214)*x[1]**o+(-0.55187467604)*x[1]
        arg[2,0,0,1]=(0.703438074924)*x[0]**o+(0.795508285134)*x[0]+(-0.0650847718675)*x[1]**o+(-0.0250766383001)*x[1]
        arg[2,0,1,0]=(0.493174434488)*x[0]**o+(-0.775523541313)*x[0]+(0.23184619496)*x[1]**o+(0.343368328234)*x[1]
        arg[2,0,1,1]=(0.404974280918)*x[0]**o+(0.0673627086244)*x[0]+(-0.843957699955)*x[1]**o+(0.645895713789)*x[1]
        arg[2,0,2,0]=(0.985047143685)*x[0]**o+(0.969582300743)*x[0]+(0.929427858098)*x[1]**o+(-0.951787413112)*x[1]
        arg[2,0,2,1]=(0.618505056215)*x[0]**o+(0.685760720523)*x[0]+(0.14594727158)*x[1]**o+(0.816579780815)*x[1]
        arg[2,1,0,0]=(0.0824953418089)*x[0]**o+(-0.896753351387)*x[0]+(0.415841596602)*x[1]**o+(0.349436600222)*x[1]
        arg[2,1,0,1]=(-0.904246301273)*x[0]**o+(0.51403543319)*x[0]+(0.0950147832662)*x[1]**o+(0.0482220256459)*x[1]
        arg[2,1,1,0]=(0.653418547332)*x[0]**o+(0.703094950575)*x[0]+(-0.294924381355)*x[1]**o+(0.889408306649)*x[1]
        arg[2,1,1,1]=(-0.329361050221)*x[0]**o+(-0.93072021487)*x[0]+(0.0845194032793)*x[1]**o+(-0.310617126183)*x[1]
        arg[2,1,2,0]=(-0.282371473471)*x[0]**o+(0.387458491178)*x[0]+(0.643908016757)*x[1]**o+(0.303741991451)*x[1]
        arg[2,1,2,1]=(-0.416872502153)*x[0]**o+(0.373465764073)*x[0]+(-0.90689140487)*x[1]**o+(-0.351052228378)*x[1]
        arg[2,2,0,0]=(0.494198770026)*x[0]**o+(0.744917026846)*x[0]+(-0.806591346974)*x[1]**o+(-0.812396479226)*x[1]
        arg[2,2,0,1]=(0.230573331783)*x[0]**o+(-0.449815323831)*x[0]+(-0.662178492084)*x[1]**o+(0.0281786827584)*x[1]
        arg[2,2,1,0]=(-0.546296599764)*x[0]**o+(-0.457939754236)*x[0]+(-0.0432225723667)*x[1]**o+(0.0431706529852)*x[1]
        arg[2,2,1,1]=(-0.808174460895)*x[0]**o+(-0.473584203382)*x[0]+(0.567514752294)*x[1]**o+(-0.115780229631)*x[1]
        arg[2,2,2,0]=(-0.134425008633)*x[0]**o+(-0.103550646345)*x[0]+(-0.588484951305)*x[1]**o+(0.955148281194)*x[1]
        arg[2,2,2,1]=(0.550095417437)*x[0]**o+(0.943564321046)*x[0]+(-0.990366145789)*x[1]**o+(0.872652302953)*x[1]
        arg[2,3,0,0]=(0.209341033615)*x[0]**o+(-0.285197252362)*x[0]+(0.961111991429)*x[1]**o+(0.810022268034)*x[1]
        arg[2,3,0,1]=(0.238319839582)*x[0]**o+(-0.714990515235)*x[0]+(0.0925371590198)*x[1]**o+(0.938145796375)*x[1]
        arg[2,3,1,0]=(0.889068347007)*x[0]**o+(-0.301904392078)*x[0]+(0.41386941601)*x[1]**o+(-0.973115118701)*x[1]
        arg[2,3,1,1]=(-0.522083695846)*x[0]**o+(-0.582455208417)*x[0]+(0.725431346859)*x[1]**o+(-0.950157996033)*x[1]
        arg[2,3,2,0]=(-0.897838142917)*x[0]**o+(0.745488698358)*x[0]+(-0.38910727894)*x[1]**o+(0.614539961316)*x[1]
        arg[2,3,2,1]=(0.924500645496)*x[0]**o+(-0.68988446876)*x[0]+(-0.466238481805)*x[1]**o+(-0.957703006413)*x[1]
        arg[2,4,0,0]=(0.280692018949)*x[0]**o+(0.600320489083)*x[0]+(-0.402390231845)*x[1]**o+(0.906721159456)*x[1]
        arg[2,4,0,1]=(0.50594574787)*x[0]**o+(0.130986910768)*x[0]+(0.272643351394)*x[1]**o+(0.53790545383)*x[1]
        arg[2,4,1,0]=(-0.627308461808)*x[0]**o+(0.690560028115)*x[0]+(-0.455768647207)*x[1]**o+(0.134926199015)*x[1]
        arg[2,4,1,1]=(-0.501163279392)*x[0]**o+(0.980144824993)*x[0]+(-0.0358930723173)*x[1]**o+(0.720157452357)*x[1]
        arg[2,4,2,0]=(-0.609861117539)*x[0]**o+(0.965699461849)*x[0]+(0.305674136837)*x[1]**o+(-0.578333683729)*x[1]
        arg[2,4,2,1]=(-0.921051416627)*x[0]**o+(-0.703479746028)*x[0]+(-0.0189027931731)*x[1]**o+(0.403415497212)*x[1]
        arg[3,0,0,0]=(-0.846266714886)*x[0]**o+(-0.103621657972)*x[0]+(0.217340457606)*x[1]**o+(-0.537654290693)*x[1]
        arg[3,0,0,1]=(0.186251745403)*x[0]**o+(0.694458841658)*x[0]+(0.126837362883)*x[1]**o+(0.0997502401792)*x[1]
        arg[3,0,1,0]=(0.153737843143)*x[0]**o+(0.811590040899)*x[0]+(0.189363446699)*x[1]**o+(0.00688399857093)*x[1]
        arg[3,0,1,1]=(-0.912444421339)*x[0]**o+(-0.68504171198)*x[0]+(-0.820717153919)*x[1]**o+(-0.00475951426309)*x[1]
        arg[3,0,2,0]=(0.908522439455)*x[0]**o+(-0.911126936858)*x[0]+(0.0424198145962)*x[1]**o+(-0.352097158128)*x[1]
        arg[3,0,2,1]=(0.935127271967)*x[0]**o+(0.154700792222)*x[0]+(-0.283148208173)*x[1]**o+(-0.607541482501)*x[1]
        arg[3,1,0,0]=(-0.131329160543)*x[0]**o+(0.24170180662)*x[0]+(-0.00893591013587)*x[1]**o+(0.262635048958)*x[1]
        arg[3,1,0,1]=(-0.559200924561)*x[0]**o+(0.509310116425)*x[0]+(0.484486720282)*x[1]**o+(-0.89314080386)*x[1]
        arg[3,1,1,0]=(0.355094442114)*x[0]**o+(-0.748398915688)*x[0]+(-0.379187476624)*x[1]**o+(0.447409858108)*x[1]
        arg[3,1,1,1]=(0.170855109103)*x[0]**o+(-0.48925054508)*x[0]+(0.840559727559)*x[1]**o+(0.0359562471889)*x[1]
        arg[3,1,2,0]=(0.827586094521)*x[0]**o+(0.0231820503547)*x[0]+(0.457192172855)*x[1]**o+(0.826960230377)*x[1]
        arg[3,1,2,1]=(-0.962890676345)*x[0]**o+(-0.112439391098)*x[0]+(-0.865510686025)*x[1]**o+(-0.539059426684)*x[1]
        arg[3,2,0,0]=(-0.165030011664)*x[0]**o+(-0.285694090419)*x[0]+(-0.522579583443)*x[1]**o+(0.73946640768)*x[1]
        arg[3,2,0,1]=(0.809351962503)*x[0]**o+(-0.872047628102)*x[0]+(0.61415433255)*x[1]**o+(0.132268088986)*x[1]
        arg[3,2,1,0]=(-0.587823544835)*x[0]**o+(-0.416092975475)*x[0]+(0.829027070431)*x[1]**o+(0.805271620727)*x[1]
        arg[3,2,1,1]=(-0.0107935093527)*x[0]**o+(0.988434287979)*x[0]+(0.941743055822)*x[1]**o+(0.137807879334)*x[1]
        arg[3,2,2,0]=(-0.855931527408)*x[0]**o+(0.575075608849)*x[0]+(0.11916967111)*x[1]**o+(-0.472844336533)*x[1]
        arg[3,2,2,1]=(-0.542112514875)*x[0]**o+(0.081308081083)*x[0]+(-0.934575648869)*x[1]**o+(-0.218208369535)*x[1]
        arg[3,3,0,0]=(0.486635391195)*x[0]**o+(0.995728500002)*x[0]+(-0.686767180456)*x[1]**o+(0.246849483855)*x[1]
        arg[3,3,0,1]=(-0.331571848212)*x[0]**o+(0.801870431206)*x[0]+(-0.91142654055)*x[1]**o+(0.089854528457)*x[1]
        arg[3,3,1,0]=(-0.657455652936)*x[0]**o+(-0.721466128651)*x[0]+(0.0420989631458)*x[1]**o+(-0.982316453325)*x[1]
        arg[3,3,1,1]=(-0.709257565086)*x[0]**o+(0.689028730506)*x[0]+(-0.196516960475)*x[1]**o+(-0.216704841188)*x[1]
        arg[3,3,2,0]=(-0.910547650394)*x[0]**o+(-0.977471949987)*x[0]+(0.37319115353)*x[1]**o+(-0.248978955796)*x[1]
        arg[3,3,2,1]=(-0.71025804576)*x[0]**o+(0.373559860529)*x[0]+(-0.9064074733)*x[1]**o+(-0.542012058073)*x[1]
        arg[3,4,0,0]=(-0.649273580176)*x[0]**o+(0.525784948768)*x[0]+(-0.213929862516)*x[1]**o+(-0.162295172275)*x[1]
        arg[3,4,0,1]=(0.209784328978)*x[0]**o+(-0.972601183002)*x[0]+(-0.730886568265)*x[1]**o+(0.966415518759)*x[1]
        arg[3,4,1,0]=(-0.334060094293)*x[0]**o+(0.29371522739)*x[0]+(0.936156150108)*x[1]**o+(-0.257026769476)*x[1]
        arg[3,4,1,1]=(-0.0465451767064)*x[0]**o+(0.115525862669)*x[0]+(0.891280073228)*x[1]**o+(-0.0436233989979)*x[1]
        arg[3,4,2,0]=(-0.748697989123)*x[0]**o+(0.246672415445)*x[0]+(0.62642302161)*x[1]**o+(-0.411644161038)*x[1]
        arg[3,4,2,1]=(0.850132316298)*x[0]**o+(0.939463534845)*x[0]+(0.0437756618735)*x[1]**o+(-0.955198146074)*x[1]
        ref[0,0,0,0]=(-0.238250641196)*(1+2.*(dim-1.)/(o+1.))+(0.0600873815876)*dim
        ref[0,0,0,1]=(-0.850813286447)*(1+2.*(dim-1.)/(o+1.))+(-0.28955404176)*dim
        ref[0,0,1,0]=(0.801090795659)*(1+2.*(dim-1.)/(o+1.))+(-1.14951117372)*dim
        ref[0,0,1,1]=(0.0372749343294)*(1+2.*(dim-1.)/(o+1.))+(-0.000924603180129)*dim
        ref[0,0,2,0]=(0.675746582079)*(1+2.*(dim-1.)/(o+1.))+(-0.102721066912)*dim
        ref[0,0,2,1]=(0.49235149033)*(1+2.*(dim-1.)/(o+1.))+(1.11945848694)*dim
        ref[0,1,0,0]=(-0.328081063906)*(1+2.*(dim-1.)/(o+1.))+(-0.723928612706)*dim
        ref[0,1,0,1]=(0.221064502678)*(1+2.*(dim-1.)/(o+1.))+(0.355700500092)*dim
        ref[0,1,1,0]=(0.365917757777)*(1+2.*(dim-1.)/(o+1.))+(-1.14536604644)*dim
        ref[0,1,1,1]=(-0.118397535132)*(1+2.*(dim-1.)/(o+1.))+(-0.667480263002)*dim
        ref[0,1,2,0]=(-0.151876137936)*(1+2.*(dim-1.)/(o+1.))+(-0.154192190988)*dim
        ref[0,1,2,1]=(-0.14200138068)*(1+2.*(dim-1.)/(o+1.))+(-0.115708012186)*dim
        ref[0,2,0,0]=(-0.0297408850402)*(1+2.*(dim-1.)/(o+1.))+(0.455966933793)*dim
        ref[0,2,0,1]=(0.306248468995)*(1+2.*(dim-1.)/(o+1.))+(-0.739470176504)*dim
        ref[0,2,1,0]=(-1.07069902254)*(1+2.*(dim-1.)/(o+1.))+(0.362039437233)*dim
        ref[0,2,1,1]=(1.27483788836)*(1+2.*(dim-1.)/(o+1.))+(0.192458109913)*dim
        ref[0,2,2,0]=(0.575364494839)*(1+2.*(dim-1.)/(o+1.))+(-0.0960630957749)*dim
        ref[0,2,2,1]=(1.55008542732)*(1+2.*(dim-1.)/(o+1.))+(0.938002240064)*dim
        ref[0,3,0,0]=(-1.0203558498)*(1+2.*(dim-1.)/(o+1.))+(-0.899787941357)*dim
        ref[0,3,0,1]=(-0.230516102889)*(1+2.*(dim-1.)/(o+1.))+(0.445128743903)*dim
        ref[0,3,1,0]=(1.11519461774)*(1+2.*(dim-1.)/(o+1.))+(0.169843961604)*dim
        ref[0,3,1,1]=(0.693437991388)*(1+2.*(dim-1.)/(o+1.))+(-0.317763193117)*dim
        ref[0,3,2,0]=(0.443131514562)*(1+2.*(dim-1.)/(o+1.))+(0.666973868262)*dim
        ref[0,3,2,1]=(0.379381381335)*(1+2.*(dim-1.)/(o+1.))+(0.875166482503)*dim
        ref[0,4,0,0]=(0.774839069479)*(1+2.*(dim-1.)/(o+1.))+(-0.576453574642)*dim
        ref[0,4,0,1]=(-1.51006312489)*(1+2.*(dim-1.)/(o+1.))+(-0.416333953063)*dim
        ref[0,4,1,0]=(-0.0121738965054)*(1+2.*(dim-1.)/(o+1.))+(-0.443710365002)*dim
        ref[0,4,1,1]=(0.889057189568)*(1+2.*(dim-1.)/(o+1.))+(0.344851315468)*dim
        ref[0,4,2,0]=(0.612526131884)*(1+2.*(dim-1.)/(o+1.))+(0.987175930567)*dim
        ref[0,4,2,1]=(-0.529852888099)*(1+2.*(dim-1.)/(o+1.))+(0.368217232392)*dim
        ref[1,0,0,0]=(0.325056920187)*(1+2.*(dim-1.)/(o+1.))+(0.372444177524)*dim
        ref[1,0,0,1]=(-1.35663541651)*(1+2.*(dim-1.)/(o+1.))+(1.16719891258)*dim
        ref[1,0,1,0]=(-0.453939765458)*(1+2.*(dim-1.)/(o+1.))+(-0.240720520301)*dim
        ref[1,0,1,1]=(0.632069743814)*(1+2.*(dim-1.)/(o+1.))+(0.645096673298)*dim
        ref[1,0,2,0]=(-0.394358616867)*(1+2.*(dim-1.)/(o+1.))+(-1.25456271654)*dim
        ref[1,0,2,1]=(-1.32963525903)*(1+2.*(dim-1.)/(o+1.))+(-0.671511618288)*dim
        ref[1,1,0,0]=(0.648323684378)*(1+2.*(dim-1.)/(o+1.))+(1.5140558313)*dim
        ref[1,1,0,1]=(-0.251177152841)*(1+2.*(dim-1.)/(o+1.))+(-1.28086389996)*dim
        ref[1,1,1,0]=(1.70962212667)*(1+2.*(dim-1.)/(o+1.))+(1.2400982776)*dim
        ref[1,1,1,1]=(0.768394214775)*(1+2.*(dim-1.)/(o+1.))+(-0.202648250425)*dim
        ref[1,1,2,0]=(1.87997116043)*(1+2.*(dim-1.)/(o+1.))+(0.0266736721392)*dim
        ref[1,1,2,1]=(0.233252509524)*(1+2.*(dim-1.)/(o+1.))+(0.585492852322)*dim
        ref[1,2,0,0]=(0.182302554291)*(1+2.*(dim-1.)/(o+1.))+(-1.53240275559)*dim
        ref[1,2,0,1]=(1.180720554)*(1+2.*(dim-1.)/(o+1.))+(-0.0573859605549)*dim
        ref[1,2,1,0]=(-0.883716231171)*(1+2.*(dim-1.)/(o+1.))+(-0.539035904746)*dim
        ref[1,2,1,1]=(-1.20573363774)*(1+2.*(dim-1.)/(o+1.))+(-0.800978801193)*dim
        ref[1,2,2,0]=(1.27431056921)*(1+2.*(dim-1.)/(o+1.))+(-0.066118345127)*dim
        ref[1,2,2,1]=(0.0429953049979)*(1+2.*(dim-1.)/(o+1.))+(-0.211917566223)*dim
        ref[1,3,0,0]=(-0.241004433021)*(1+2.*(dim-1.)/(o+1.))+(-0.356056991836)*dim
        ref[1,3,0,1]=(-0.576268169751)*(1+2.*(dim-1.)/(o+1.))+(-0.424538063457)*dim
        ref[1,3,1,0]=(1.43916440906)*(1+2.*(dim-1.)/(o+1.))+(-0.0243659035174)*dim
        ref[1,3,1,1]=(-0.462364051993)*(1+2.*(dim-1.)/(o+1.))+(-0.240629748528)*dim
        ref[1,3,2,0]=(-0.193522529311)*(1+2.*(dim-1.)/(o+1.))+(-0.418080251656)*dim
        ref[1,3,2,1]=(1.64817238127)*(1+2.*(dim-1.)/(o+1.))+(-0.295195478539)*dim
        ref[1,4,0,0]=(-0.542198585489)*(1+2.*(dim-1.)/(o+1.))+(-0.616252627062)*dim
        ref[1,4,0,1]=(-1.6461046371)*(1+2.*(dim-1.)/(o+1.))+(-0.332867143554)*dim
        ref[1,4,1,0]=(0.333229017732)*(1+2.*(dim-1.)/(o+1.))+(-1.05552711715)*dim
        ref[1,4,1,1]=(1.49605710755)*(1+2.*(dim-1.)/(o+1.))+(-1.35425133515)*dim
        ref[1,4,2,0]=(0.319136490505)*(1+2.*(dim-1.)/(o+1.))+(0.246099391883)*dim
        ref[1,4,2,1]=(-0.0958737570155)*(1+2.*(dim-1.)/(o+1.))+(-0.0803850673082)*dim
        ref[2,0,0,0]=(0.0482996624604)*(1+2.*(dim-1.)/(o+1.))+(-0.171234658375)*dim
        ref[2,0,0,1]=(0.638353303057)*(1+2.*(dim-1.)/(o+1.))+(0.770431646834)*dim
        ref[2,0,1,0]=(0.725020629449)*(1+2.*(dim-1.)/(o+1.))+(-0.432155213079)*dim
        ref[2,0,1,1]=(-0.438983419037)*(1+2.*(dim-1.)/(o+1.))+(0.713258422414)*dim
        ref[2,0,2,0]=(1.91447500178)*(1+2.*(dim-1.)/(o+1.))+(0.0177948876308)*dim
        ref[2,0,2,1]=(0.764452327795)*(1+2.*(dim-1.)/(o+1.))+(1.50234050134)*dim
        ref[2,1,0,0]=(0.498336938411)*(1+2.*(dim-1.)/(o+1.))+(-0.547316751164)*dim
        ref[2,1,0,1]=(-0.809231518007)*(1+2.*(dim-1.)/(o+1.))+(0.562257458836)*dim
        ref[2,1,1,0]=(0.358494165977)*(1+2.*(dim-1.)/(o+1.))+(1.59250325722)*dim
        ref[2,1,1,1]=(-0.244841646941)*(1+2.*(dim-1.)/(o+1.))+(-1.24133734105)*dim
        ref[2,1,2,0]=(0.361536543287)*(1+2.*(dim-1.)/(o+1.))+(0.691200482629)*dim
        ref[2,1,2,1]=(-1.32376390702)*(1+2.*(dim-1.)/(o+1.))+(0.0224135356952)*dim
        ref[2,2,0,0]=(-0.312392576948)*(1+2.*(dim-1.)/(o+1.))+(-0.06747945238)*dim
        ref[2,2,0,1]=(-0.431605160301)*(1+2.*(dim-1.)/(o+1.))+(-0.421636641073)*dim
        ref[2,2,1,0]=(-0.589519172131)*(1+2.*(dim-1.)/(o+1.))+(-0.414769101251)*dim
        ref[2,2,1,1]=(-0.240659708601)*(1+2.*(dim-1.)/(o+1.))+(-0.589364433013)*dim
        ref[2,2,2,0]=(-0.722909959939)*(1+2.*(dim-1.)/(o+1.))+(0.851597634849)*dim
        ref[2,2,2,1]=(-0.440270728352)*(1+2.*(dim-1.)/(o+1.))+(1.816216624)*dim
        ref[2,3,0,0]=(1.17045302504)*(1+2.*(dim-1.)/(o+1.))+(0.524825015672)*dim
        ref[2,3,0,1]=(0.330856998602)*(1+2.*(dim-1.)/(o+1.))+(0.22315528114)*dim
        ref[2,3,1,0]=(1.30293776302)*(1+2.*(dim-1.)/(o+1.))+(-1.27501951078)*dim
        ref[2,3,1,1]=(0.203347651013)*(1+2.*(dim-1.)/(o+1.))+(-1.53261320445)*dim
        ref[2,3,2,0]=(-1.28694542186)*(1+2.*(dim-1.)/(o+1.))+(1.36002865967)*dim
        ref[2,3,2,1]=(0.458262163691)*(1+2.*(dim-1.)/(o+1.))+(-1.64758747517)*dim
        ref[2,4,0,0]=(-0.121698212896)*(1+2.*(dim-1.)/(o+1.))+(1.50704164854)*dim
        ref[2,4,0,1]=(0.778589099264)*(1+2.*(dim-1.)/(o+1.))+(0.668892364598)*dim
        ref[2,4,1,0]=(-1.08307710902)*(1+2.*(dim-1.)/(o+1.))+(0.82548622713)*dim
        ref[2,4,1,1]=(-0.537056351709)*(1+2.*(dim-1.)/(o+1.))+(1.70030227735)*dim
        ref[2,4,2,0]=(-0.304186980702)*(1+2.*(dim-1.)/(o+1.))+(0.38736577812)*dim
        ref[2,4,2,1]=(-0.9399542098)*(1+2.*(dim-1.)/(o+1.))+(-0.300064248815)*dim
        ref[3,0,0,0]=(-0.62892625728)*(1+2.*(dim-1.)/(o+1.))+(-0.641275948665)*dim
        ref[3,0,0,1]=(0.313089108286)*(1+2.*(dim-1.)/(o+1.))+(0.794209081837)*dim
        ref[3,0,1,0]=(0.343101289842)*(1+2.*(dim-1.)/(o+1.))+(0.81847403947)*dim
        ref[3,0,1,1]=(-1.73316157526)*(1+2.*(dim-1.)/(o+1.))+(-0.689801226243)*dim
        ref[3,0,2,0]=(0.950942254051)*(1+2.*(dim-1.)/(o+1.))+(-1.26322409499)*dim
        ref[3,0,2,1]=(0.651979063795)*(1+2.*(dim-1.)/(o+1.))+(-0.452840690279)*dim
        ref[3,1,0,0]=(-0.140265070679)*(1+2.*(dim-1.)/(o+1.))+(0.504336855579)*dim
        ref[3,1,0,1]=(-0.074714204279)*(1+2.*(dim-1.)/(o+1.))+(-0.383830687435)*dim
        ref[3,1,1,0]=(-0.0240930345095)*(1+2.*(dim-1.)/(o+1.))+(-0.30098905758)*dim
        ref[3,1,1,1]=(1.01141483666)*(1+2.*(dim-1.)/(o+1.))+(-0.453294297891)*dim
        ref[3,1,2,0]=(1.28477826738)*(1+2.*(dim-1.)/(o+1.))+(0.850142280731)*dim
        ref[3,1,2,1]=(-1.82840136237)*(1+2.*(dim-1.)/(o+1.))+(-0.651498817783)*dim
        ref[3,2,0,0]=(-0.687609595107)*(1+2.*(dim-1.)/(o+1.))+(0.453772317262)*dim
        ref[3,2,0,1]=(1.42350629505)*(1+2.*(dim-1.)/(o+1.))+(-0.739779539116)*dim
        ref[3,2,1,0]=(0.241203525596)*(1+2.*(dim-1.)/(o+1.))+(0.389178645253)*dim
        ref[3,2,1,1]=(0.930949546469)*(1+2.*(dim-1.)/(o+1.))+(1.12624216731)*dim
        ref[3,2,2,0]=(-0.736761856298)*(1+2.*(dim-1.)/(o+1.))+(0.102231272315)*dim
        ref[3,2,2,1]=(-1.47668816374)*(1+2.*(dim-1.)/(o+1.))+(-0.136900288452)*dim
        ref[3,3,0,0]=(-0.200131789262)*(1+2.*(dim-1.)/(o+1.))+(1.24257798386)*dim
        ref[3,3,0,1]=(-1.24299838876)*(1+2.*(dim-1.)/(o+1.))+(0.891724959663)*dim
        ref[3,3,1,0]=(-0.61535668979)*(1+2.*(dim-1.)/(o+1.))+(-1.70378258198)*dim
        ref[3,3,1,1]=(-0.905774525561)*(1+2.*(dim-1.)/(o+1.))+(0.472323889318)*dim
        ref[3,3,2,0]=(-0.537356496864)*(1+2.*(dim-1.)/(o+1.))+(-1.22645090578)*dim
        ref[3,3,2,1]=(-1.61666551906)*(1+2.*(dim-1.)/(o+1.))+(-0.168452197544)*dim
        ref[3,4,0,0]=(-0.863203442693)*(1+2.*(dim-1.)/(o+1.))+(0.363489776493)*dim
        ref[3,4,0,1]=(-0.521102239287)*(1+2.*(dim-1.)/(o+1.))+(-0.00618566424301)*dim
        ref[3,4,1,0]=(0.602096055815)*(1+2.*(dim-1.)/(o+1.))+(0.0366884579134)*dim
        ref[3,4,1,1]=(0.844734896522)*(1+2.*(dim-1.)/(o+1.))+(0.0719024636709)*dim
        ref[3,4,2,0]=(-0.122274967513)*(1+2.*(dim-1.)/(o+1.))+(-0.164971745593)*dim
        ref[3,4,2,1]=(0.893907978172)*(1+2.*(dim-1.)/(o+1.))+(-0.0157346112292)*dim
      else:
        arg[0,0,0,0]=(-0.0445963085788)*x[0]**o+(0.649401127647)*x[0]+(-0.949035518207)*x[1]**o+(0.620675629012)*x[1]+(0.379727527503)*x[2]**o+(0.0321033403935)*x[2]
        arg[0,0,0,1]=(-0.334483142061)*x[0]**o+(-0.408749824143)*x[0]+(-0.427042441291)*x[1]**o+(-0.263818177378)*x[1]+(-0.0490001897328)*x[2]**o+(0.925792628022)*x[2]
        arg[0,0,1,0]=(0.632498844224)*x[0]**o+(-0.756487058254)*x[0]+(-0.0277391041902)*x[1]**o+(0.615017171892)*x[1]+(-0.0956053270216)*x[2]**o+(-0.0383148218356)*x[2]
        arg[0,0,1,1]=(0.560189295143)*x[0]**o+(-0.565632451768)*x[0]+(0.131609132282)*x[1]**o+(0.689012872903)*x[1]+(-0.0542418297927)*x[2]**o+(-0.816694105058)*x[2]
        arg[0,0,2,0]=(-0.606523401606)*x[0]**o+(-0.129782638438)*x[0]+(0.286554604329)*x[1]**o+(-0.264164004519)*x[1]+(-0.145168356054)*x[2]**o+(0.346147680755)*x[2]
        arg[0,0,2,1]=(-0.847528471559)*x[0]**o+(-0.50641170288)*x[0]+(-0.773325001399)*x[1]**o+(0.517649361249)*x[1]+(-0.531000881434)*x[2]**o+(0.220222077563)*x[2]
        arg[0,1,0,0]=(-0.0946495378384)*x[0]**o+(-0.312650449416)*x[0]+(0.183321918505)*x[1]**o+(-0.359448777446)*x[1]+(0.713219915733)*x[2]**o+(-0.436742264028)*x[2]
        arg[0,1,0,1]=(0.856947337921)*x[0]**o+(0.497228192699)*x[0]+(0.351718755501)*x[1]**o+(-0.928210477058)*x[1]+(0.586853747677)*x[2]**o+(-0.915642724205)*x[2]
        arg[0,1,1,0]=(-0.745330469339)*x[0]**o+(-0.724610807566)*x[0]+(0.821804346007)*x[1]**o+(0.241991559492)*x[1]+(0.508380905493)*x[2]**o+(0.756706999359)*x[2]
        arg[0,1,1,1]=(-0.647952004012)*x[0]**o+(0.940964624139)*x[0]+(-0.620895246845)*x[1]**o+(0.577736885116)*x[1]+(0.228716730257)*x[2]**o+(0.336914427419)*x[2]
        arg[0,1,2,0]=(-0.671541115529)*x[0]**o+(0.153595444626)*x[0]+(0.420649971716)*x[1]**o+(-0.459078302559)*x[1]+(-0.583263239511)*x[2]**o+(-0.927250932344)*x[2]
        arg[0,1,2,1]=(-0.893954406097)*x[0]**o+(0.266297374506)*x[0]+(-0.777463374391)*x[1]**o+(0.0905062188555)*x[1]+(-0.668756964585)*x[2]**o+(-0.580980380472)*x[2]
        arg[0,2,0,0]=(0.424449538697)*x[0]**o+(0.0787502268502)*x[0]+(-0.645996834851)*x[1]**o+(-0.0409730346273)*x[1]+(0.835463212243)*x[2]**o+(0.716052012249)*x[2]
        arg[0,2,0,1]=(0.0241441154932)*x[0]**o+(-0.580242261617)*x[0]+(0.693484316968)*x[1]**o+(0.198598746328)*x[1]+(-0.586283866696)*x[2]**o+(-0.595008846645)*x[2]
        arg[0,2,1,0]=(0.074643270363)*x[0]**o+(0.454053362138)*x[0]+(0.861855819289)*x[1]**o+(-0.0310194739742)*x[1]+(-0.35738290172)*x[2]**o+(-0.400226698584)*x[2]
        arg[0,2,1,1]=(0.61780464269)*x[0]**o+(-0.702818770068)*x[0]+(0.0840653474306)*x[1]**o+(0.786842849442)*x[1]+(0.0704342755849)*x[2]**o+(0.386459053939)*x[2]
        arg[0,2,2,0]=(-0.815363379053)*x[0]**o+(-0.621823615812)*x[0]+(-0.13535005543)*x[1]**o+(-0.127080917834)*x[1]+(-0.178866000441)*x[2]**o+(-0.0753933476244)*x[2]
        arg[0,2,2,1]=(0.400194366678)*x[0]**o+(0.075003395019)*x[0]+(-0.35708995046)*x[1]**o+(-0.370028795924)*x[1]+(-0.674944523932)*x[2]**o+(-0.437373855225)*x[2]
        arg[0,3,0,0]=(-0.736317751792)*x[0]**o+(0.671627391036)*x[0]+(0.0768737429532)*x[1]**o+(-0.90036869218)*x[1]+(0.796767364925)*x[2]**o+(-0.185034722102)*x[2]
        arg[0,3,0,1]=(0.530921403378)*x[0]**o+(0.970385957718)*x[0]+(-0.75659968724)*x[1]**o+(-0.437200812366)*x[1]+(0.0519374612092)*x[2]**o+(0.657323406602)*x[2]
        arg[0,3,1,0]=(-0.109751102222)*x[0]**o+(-0.154990305064)*x[0]+(-0.502004367351)*x[1]**o+(0.535266424739)*x[1]+(-0.968926773395)*x[2]**o+(-0.138032126456)*x[2]
        arg[0,3,1,1]=(0.509506900641)*x[0]**o+(-0.872357145765)*x[0]+(-0.154129018225)*x[1]**o+(0.0190914956768)*x[1]+(0.119094726078)*x[2]**o+(0.29593674915)*x[2]
        arg[0,3,2,0]=(-0.133398979819)*x[0]**o+(-0.780852347888)*x[0]+(-0.346576066687)*x[1]**o+(0.187923380934)*x[1]+(0.939817544138)*x[2]**o+(-0.497081081638)*x[2]
        arg[0,3,2,1]=(-0.908783140711)*x[0]**o+(0.541754659297)*x[0]+(-0.297350066105)*x[1]**o+(0.118676031731)*x[1]+(0.985854358996)*x[2]**o+(-0.424715359194)*x[2]
        arg[0,4,0,0]=(0.341191520307)*x[0]**o+(-0.252381555804)*x[0]+(-0.307289805415)*x[1]**o+(0.762228901652)*x[1]+(0.390901852611)*x[2]**o+(0.147349139229)*x[2]
        arg[0,4,0,1]=(0.236625833503)*x[0]**o+(0.856322905978)*x[0]+(-0.160244298088)*x[1]**o+(-0.971940590481)*x[1]+(-0.0829366716357)*x[2]**o+(-0.687610358703)*x[2]
        arg[0,4,1,0]=(0.619591455739)*x[0]**o+(-0.62589432832)*x[0]+(-0.0275687591477)*x[1]**o+(0.367590198294)*x[1]+(0.745367285277)*x[2]**o+(0.232571513762)*x[2]
        arg[0,4,1,1]=(0.317669904199)*x[0]**o+(0.741734360559)*x[0]+(-0.923644886773)*x[1]**o+(-0.725994528938)*x[1]+(-0.949385908603)*x[2]**o+(-0.0515088731457)*x[2]
        arg[0,4,2,0]=(0.82547889841)*x[0]**o+(-0.926970312436)*x[0]+(0.946941744691)*x[1]**o+(-0.171582097545)*x[1]+(-0.185628570945)*x[2]**o+(0.40495393086)*x[2]
        arg[0,4,2,1]=(0.51879613184)*x[0]**o+(-0.517447781109)*x[0]+(-0.288826585096)*x[1]**o+(-0.736890998485)*x[1]+(-0.167119326999)*x[2]**o+(-0.653082372893)*x[2]
        arg[1,0,0,0]=(-0.879008868811)*x[0]**o+(-0.0613987509179)*x[0]+(-0.671288168832)*x[1]**o+(-0.959517297906)*x[1]+(0.292554039162)*x[2]**o+(0.942246368678)*x[2]
        arg[1,0,0,1]=(0.599956815218)*x[0]**o+(0.731601305574)*x[0]+(-0.61593947163)*x[1]**o+(-0.597924796707)*x[1]+(-0.791620695721)*x[2]**o+(-0.562978718552)*x[2]
        arg[1,0,1,0]=(0.448381207242)*x[0]**o+(-0.178004176549)*x[0]+(0.631841924913)*x[1]**o+(0.311836637044)*x[1]+(0.0619475841251)*x[2]**o+(0.606564490624)*x[2]
        arg[1,0,1,1]=(0.159907398582)*x[0]**o+(-0.278605461684)*x[0]+(0.822914081538)*x[1]**o+(0.75112795022)*x[1]+(0.91389895436)*x[2]**o+(0.929177149059)*x[2]
        arg[1,0,2,0]=(-0.472286972241)*x[0]**o+(0.00950883684535)*x[0]+(0.41110884679)*x[1]**o+(0.109422159345)*x[1]+(-0.707436024011)*x[2]**o+(0.795057014561)*x[2]
        arg[1,0,2,1]=(-0.0156784802605)*x[0]**o+(-0.657283760492)*x[0]+(-0.0422880440635)*x[1]**o+(-0.326436491982)*x[1]+(-0.543917764251)*x[2]**o+(-0.691698129266)*x[2]
        arg[1,1,0,0]=(-0.814424723604)*x[0]**o+(0.746801606056)*x[0]+(0.22025107532)*x[1]**o+(-0.785254149057)*x[1]+(-0.845637273475)*x[2]**o+(-0.468272754144)*x[2]
        arg[1,1,0,1]=(-0.0815177460988)*x[0]**o+(-0.47210691717)*x[0]+(-0.285964111717)*x[1]**o+(0.527387496664)*x[1]+(0.87059583429)*x[2]**o+(-0.543899395468)*x[2]
        arg[1,1,1,0]=(0.424163426295)*x[0]**o+(0.649692204305)*x[0]+(0.0570010318344)*x[1]**o+(-0.426047169951)*x[1]+(0.334750596342)*x[2]**o+(-0.746549570927)*x[2]
        arg[1,1,1,1]=(-0.120493849358)*x[0]**o+(0.567677110538)*x[0]+(-0.355891513526)*x[1]**o+(0.68069794787)*x[1]+(0.786560156246)*x[2]**o+(-0.422804642829)*x[2]
        arg[1,1,2,0]=(-0.80298202932)*x[0]**o+(-0.674766348751)*x[0]+(-0.29945072468)*x[1]**o+(0.385941328636)*x[1]+(-0.756272624651)*x[2]**o+(-0.607541099269)*x[2]
        arg[1,1,2,1]=(0.915020713521)*x[0]**o+(-0.875957199283)*x[0]+(0.0217597029367)*x[1]**o+(-0.0649461013997)*x[1]+(-0.302450259541)*x[2]**o+(0.738498325526)*x[2]
        arg[1,2,0,0]=(0.239905462913)*x[0]**o+(0.197348893262)*x[0]+(-0.151173414577)*x[1]**o+(0.368405558905)*x[1]+(0.891448032058)*x[2]**o+(0.874975769684)*x[2]
        arg[1,2,0,1]=(-0.304971614823)*x[0]**o+(-0.0166065674003)*x[0]+(0.883397157318)*x[1]**o+(-0.918803270424)*x[1]+(0.590875856583)*x[2]**o+(-0.360659400522)*x[2]
        arg[1,2,1,0]=(0.185368540041)*x[0]**o+(0.498866146124)*x[0]+(0.591601152982)*x[1]**o+(-0.986932631453)*x[1]+(-0.525706872664)*x[2]**o+(-0.0462993899877)*x[2]
        arg[1,2,1,1]=(0.260677894495)*x[0]**o+(-0.216847258131)*x[0]+(0.932876363707)*x[1]**o+(0.761630676808)*x[1]+(0.260889221066)*x[2]**o+(0.546259908582)*x[2]
        arg[1,2,2,0]=(-0.316803600202)*x[0]**o+(-0.898044300642)*x[0]+(-0.0865001657852)*x[1]**o+(0.760424275199)*x[1]+(0.771564532427)*x[2]**o+(-0.0852237819025)*x[2]
        arg[1,2,2,1]=(-0.54274371504)*x[0]**o+(0.0822524750896)*x[0]+(0.872655885939)*x[1]**o+(0.594054041006)*x[1]+(0.322150325252)*x[2]**o+(0.774020612668)*x[2]
        arg[1,3,0,0]=(0.034593806358)*x[0]**o+(0.541638857337)*x[0]+(0.291359214696)*x[1]**o+(-0.203872730757)*x[1]+(0.0312260615928)*x[2]**o+(0.62568815669)*x[2]
        arg[1,3,0,1]=(-0.31929050432)*x[0]**o+(-0.931756481553)*x[0]+(0.755537511071)*x[1]**o+(-0.182578278538)*x[1]+(0.00958902784528)*x[2]**o+(0.345520390047)*x[2]
        arg[1,3,1,0]=(-0.435083251076)*x[0]**o+(0.248332963857)*x[0]+(-0.122701495966)*x[1]**o+(0.669350594899)*x[1]+(-0.0758596665333)*x[2]**o+(0.0784997907526)*x[2]
        arg[1,3,1,1]=(-0.725479421981)*x[0]**o+(0.657242746138)*x[0]+(-0.199958263735)*x[1]**o+(-0.784165440236)*x[1]+(-0.550936676647)*x[2]**o+(-0.31907813038)*x[2]
        arg[1,3,2,0]=(0.657177346512)*x[0]**o+(-0.168053866293)*x[0]+(0.782142168258)*x[1]**o+(0.660514922866)*x[1]+(0.788106295165)*x[2]**o+(-0.254290645125)*x[2]
        arg[1,3,2,1]=(-0.215166688682)*x[0]**o+(0.743577328376)*x[0]+(-0.374827404054)*x[1]**o+(0.267885661143)*x[1]+(0.728818301014)*x[2]**o+(0.6828999278)*x[2]
        arg[1,4,0,0]=(0.371646403413)*x[0]**o+(0.00998777673275)*x[0]+(-0.313378345871)*x[1]**o+(-0.952580545501)*x[1]+(-0.680596974717)*x[2]**o+(-0.181426263871)*x[2]
        arg[1,4,0,1]=(-0.828437965469)*x[0]**o+(0.302443878606)*x[0]+(0.581310358652)*x[1]**o+(0.714092786618)*x[1]+(0.650179420347)*x[2]**o+(0.663482256421)*x[2]
        arg[1,4,1,0]=(0.220318191456)*x[0]**o+(0.443108440795)*x[0]+(-0.538413779146)*x[1]**o+(0.660052783252)*x[1]+(-0.846698606343)*x[2]**o+(-0.0447806256913)*x[2]
        arg[1,4,1,1]=(0.866921895372)*x[0]**o+(0.776291161128)*x[0]+(-0.385318933957)*x[1]**o+(0.841514060614)*x[1]+(-0.401465329431)*x[2]**o+(0.474475442638)*x[2]
        arg[1,4,2,0]=(-0.222535436657)*x[0]**o+(0.921341450452)*x[0]+(-0.765043798058)*x[1]**o+(-0.492946667792)*x[1]+(0.758817008648)*x[2]**o+(0.0950595321909)*x[2]
        arg[1,4,2,1]=(0.333884542089)*x[0]**o+(0.671265678974)*x[0]+(-0.719262407869)*x[1]**o+(0.498176251457)*x[1]+(-0.867726438404)*x[2]**o+(-0.872614195608)*x[2]
        arg[2,0,0,0]=(-0.303331399994)*x[0]**o+(-0.563887060818)*x[0]+(0.64537626236)*x[1]**o+(-0.657611314342)*x[1]+(0.800882746828)*x[2]**o+(0.700671892157)*x[2]
        arg[2,0,0,1]=(0.417691239931)*x[0]**o+(-0.822789801933)*x[0]+(0.712214951539)*x[1]**o+(0.160960575173)*x[1]+(0.0400205244239)*x[2]**o+(0.49892939112)*x[2]
        arg[2,0,1,0]=(-0.828677767124)*x[0]**o+(-0.347411290435)*x[0]+(-0.700177402223)*x[1]**o+(-0.42498634463)*x[1]+(0.984172081328)*x[2]**o+(-0.677168643292)*x[2]
        arg[2,0,1,1]=(-0.423091398412)*x[0]**o+(-0.786913239913)*x[0]+(-0.888102258019)*x[1]**o+(-0.119525894432)*x[1]+(-0.158513498081)*x[2]**o+(0.503403674219)*x[2]
        arg[2,0,2,0]=(0.547217868614)*x[0]**o+(-0.737401808325)*x[0]+(-0.865530737659)*x[1]**o+(-0.750671412749)*x[1]+(-0.06731420082)*x[2]**o+(0.658572191108)*x[2]
        arg[2,0,2,1]=(0.501785022047)*x[0]**o+(-0.0542441989122)*x[0]+(-0.379614025724)*x[1]**o+(-0.785618793316)*x[1]+(0.994476365391)*x[2]**o+(-0.436996817494)*x[2]
        arg[2,1,0,0]=(0.0329439903445)*x[0]**o+(0.688097592632)*x[0]+(0.554030860489)*x[1]**o+(0.60949736297)*x[1]+(0.499716171437)*x[2]**o+(0.501464282299)*x[2]
        arg[2,1,0,1]=(0.459629414019)*x[0]**o+(0.810658146454)*x[0]+(-0.902023596001)*x[1]**o+(-0.982998009297)*x[1]+(0.487736529196)*x[2]**o+(-0.509164692477)*x[2]
        arg[2,1,1,0]=(0.0818151896334)*x[0]**o+(-0.931023788966)*x[0]+(-0.217349745187)*x[1]**o+(-0.785586110912)*x[1]+(0.16783052001)*x[2]**o+(-0.515867681053)*x[2]
        arg[2,1,1,1]=(-0.758874975703)*x[0]**o+(0.219408816573)*x[0]+(0.508187203896)*x[1]**o+(0.306700464008)*x[1]+(-0.641891226105)*x[2]**o+(-0.137752888266)*x[2]
        arg[2,1,2,0]=(-0.952430879482)*x[0]**o+(0.179895758684)*x[0]+(0.502285174801)*x[1]**o+(-0.82300826605)*x[1]+(-0.267848001419)*x[2]**o+(-0.882529775244)*x[2]
        arg[2,1,2,1]=(0.371372868601)*x[0]**o+(-0.521801538326)*x[0]+(0.894906517432)*x[1]**o+(0.205799057307)*x[1]+(-0.315840560279)*x[2]**o+(0.65823096555)*x[2]
        arg[2,2,0,0]=(0.585088762354)*x[0]**o+(0.527662713737)*x[0]+(0.83511043292)*x[1]**o+(-0.263597549489)*x[1]+(-0.342210974079)*x[2]**o+(0.532149528802)*x[2]
        arg[2,2,0,1]=(0.985868978475)*x[0]**o+(0.162798954854)*x[0]+(-0.887525429296)*x[1]**o+(0.618905848201)*x[1]+(-0.213643353241)*x[2]**o+(0.0170513098359)*x[2]
        arg[2,2,1,0]=(-0.216402064475)*x[0]**o+(0.350853951704)*x[0]+(0.592219675443)*x[1]**o+(-0.915293916561)*x[1]+(0.160588856435)*x[2]**o+(-0.574686420146)*x[2]
        arg[2,2,1,1]=(0.233118497878)*x[0]**o+(0.482035958353)*x[0]+(-0.581971661478)*x[1]**o+(0.832450520208)*x[1]+(-0.430914444794)*x[2]**o+(-0.791658025494)*x[2]
        arg[2,2,2,0]=(-0.710516844)*x[0]**o+(0.0266391730256)*x[0]+(-0.154541997151)*x[1]**o+(-0.232843778379)*x[1]+(0.453522207498)*x[2]**o+(-0.688939516711)*x[2]
        arg[2,2,2,1]=(0.26860058468)*x[0]**o+(-0.969638661751)*x[0]+(0.623784614857)*x[1]**o+(0.686359051774)*x[1]+(0.677056704385)*x[2]**o+(0.156688176847)*x[2]
        arg[2,3,0,0]=(-0.0620909841534)*x[0]**o+(-0.364946660535)*x[0]+(-0.388275430806)*x[1]**o+(-0.947908160286)*x[1]+(0.0774302410344)*x[2]**o+(0.374469033649)*x[2]
        arg[2,3,0,1]=(0.164845147121)*x[0]**o+(0.480213031852)*x[0]+(0.58157030274)*x[1]**o+(0.882820521347)*x[1]+(0.584714825748)*x[2]**o+(0.0285154925376)*x[2]
        arg[2,3,1,0]=(0.664711234489)*x[0]**o+(0.247415435607)*x[0]+(0.741592287061)*x[1]**o+(0.963505986041)*x[1]+(-0.987844687003)*x[2]**o+(0.998294459659)*x[2]
        arg[2,3,1,1]=(-0.734529316764)*x[0]**o+(-0.177480004713)*x[0]+(0.923555424779)*x[1]**o+(0.727331498256)*x[1]+(0.452312367674)*x[2]**o+(0.00652243396085)*x[2]
        arg[2,3,2,0]=(-0.367377897953)*x[0]**o+(-0.157529442804)*x[0]+(-0.121881528621)*x[1]**o+(0.865116726884)*x[1]+(-0.200087079925)*x[2]**o+(0.930889629788)*x[2]
        arg[2,3,2,1]=(-0.900394562978)*x[0]**o+(-0.762334071565)*x[0]+(-0.723776245294)*x[1]**o+(0.332728994729)*x[1]+(0.390453779523)*x[2]**o+(0.0447350797932)*x[2]
        arg[2,4,0,0]=(0.178230443819)*x[0]**o+(0.0898732089623)*x[0]+(0.808448952186)*x[1]**o+(0.164579824494)*x[1]+(-0.553032263217)*x[2]**o+(-0.782141879825)*x[2]
        arg[2,4,0,1]=(-0.105582488753)*x[0]**o+(0.995415696949)*x[0]+(0.978971670365)*x[1]**o+(-0.977866735017)*x[1]+(0.156938966924)*x[2]**o+(-0.120871488936)*x[2]
        arg[2,4,1,0]=(0.652354601889)*x[0]**o+(0.381186764672)*x[0]+(-0.783479778972)*x[1]**o+(-0.158898829615)*x[1]+(0.794693790129)*x[2]**o+(0.914720093741)*x[2]
        arg[2,4,1,1]=(0.571506419376)*x[0]**o+(0.906086581152)*x[0]+(-0.798519792125)*x[1]**o+(-0.22840897962)*x[1]+(-0.750891721145)*x[2]**o+(-0.591014038221)*x[2]
        arg[2,4,2,0]=(0.515982716306)*x[0]**o+(0.299569326845)*x[0]+(-0.407423180268)*x[1]**o+(-0.252103363573)*x[1]+(-0.0228238994531)*x[2]**o+(-0.282150542844)*x[2]
        arg[2,4,2,1]=(0.5335415942)*x[0]**o+(-0.885728570489)*x[0]+(-0.750810131327)*x[1]**o+(0.604053940679)*x[1]+(-0.0946405866686)*x[2]**o+(-0.423845465234)*x[2]
        arg[3,0,0,0]=(-0.498025341136)*x[0]**o+(-0.296354416633)*x[0]+(-0.600874889655)*x[1]**o+(0.544042300882)*x[1]+(0.318784851038)*x[2]**o+(-0.801109111437)*x[2]
        arg[3,0,0,1]=(-0.492361949745)*x[0]**o+(-0.708188092599)*x[0]+(-0.863737808435)*x[1]**o+(0.674889485553)*x[1]+(-0.982254652233)*x[2]**o+(0.682172862189)*x[2]
        arg[3,0,1,0]=(-0.872692397532)*x[0]**o+(0.563852836655)*x[0]+(-0.0219331486599)*x[1]**o+(-0.858175679559)*x[1]+(0.802432631538)*x[2]**o+(-0.655545287387)*x[2]
        arg[3,0,1,1]=(-0.538117166854)*x[0]**o+(-0.62507518352)*x[0]+(0.101524298816)*x[1]**o+(0.590432448393)*x[1]+(0.668057302006)*x[2]**o+(0.405382330056)*x[2]
        arg[3,0,2,0]=(-0.322916455064)*x[0]**o+(0.950624857669)*x[0]+(0.485623846818)*x[1]**o+(-0.12611522301)*x[1]+(0.503586393973)*x[2]**o+(0.109451719709)*x[2]
        arg[3,0,2,1]=(0.0289965995645)*x[0]**o+(0.0101440661359)*x[0]+(-0.303827119958)*x[1]**o+(0.355269935141)*x[1]+(0.29889427526)*x[2]**o+(-0.707999988781)*x[2]
        arg[3,1,0,0]=(-0.541382521848)*x[0]**o+(0.534108535206)*x[0]+(0.875610773119)*x[1]**o+(-0.575849439451)*x[1]+(-0.0836939606164)*x[2]**o+(0.433678856889)*x[2]
        arg[3,1,0,1]=(0.0322576144027)*x[0]**o+(0.794219359514)*x[0]+(0.629442621265)*x[1]**o+(0.0517726053499)*x[1]+(-0.776050101686)*x[2]**o+(0.697910196708)*x[2]
        arg[3,1,1,0]=(-0.198211682395)*x[0]**o+(-0.926268101941)*x[0]+(-0.904760383993)*x[1]**o+(-0.820075667018)*x[1]+(-0.101262044428)*x[2]**o+(-0.0621221772941)*x[2]
        arg[3,1,1,1]=(-0.531324749521)*x[0]**o+(0.736017207718)*x[0]+(0.487299696259)*x[1]**o+(-0.00604423021253)*x[1]+(-0.635722590088)*x[2]**o+(-0.143501730927)*x[2]
        arg[3,1,2,0]=(-0.731421637848)*x[0]**o+(-0.0371648783851)*x[0]+(-0.290362466891)*x[1]**o+(-0.375048584106)*x[1]+(-0.854622532092)*x[2]**o+(-0.794295905388)*x[2]
        arg[3,1,2,1]=(-0.766383422105)*x[0]**o+(0.149248159984)*x[0]+(0.970103674649)*x[1]**o+(-0.117794368107)*x[1]+(0.0178775633047)*x[2]**o+(-0.221126829278)*x[2]
        arg[3,2,0,0]=(0.706586887601)*x[0]**o+(0.162520023447)*x[0]+(-0.731539880336)*x[1]**o+(-0.992558519713)*x[1]+(0.426569303014)*x[2]**o+(0.0451783819487)*x[2]
        arg[3,2,0,1]=(0.0392295508272)*x[0]**o+(-0.157929005382)*x[0]+(0.0636418899878)*x[1]**o+(0.892501759578)*x[1]+(-0.368161001401)*x[2]**o+(-0.715520794543)*x[2]
        arg[3,2,1,0]=(0.889667707133)*x[0]**o+(0.308604168817)*x[0]+(-0.275558663983)*x[1]**o+(-0.408210095215)*x[1]+(-0.0918000869138)*x[2]**o+(0.142754064645)*x[2]
        arg[3,2,1,1]=(0.585831954407)*x[0]**o+(-0.705004375516)*x[0]+(0.657277099061)*x[1]**o+(-0.972401794257)*x[1]+(-0.658049310396)*x[2]**o+(-0.553961555706)*x[2]
        arg[3,2,2,0]=(0.88916769079)*x[0]**o+(-0.1401162684)*x[0]+(-0.396496633317)*x[1]**o+(0.215029522669)*x[1]+(-0.619023294208)*x[2]**o+(-0.187670706968)*x[2]
        arg[3,2,2,1]=(-0.262983340384)*x[0]**o+(-0.763236528477)*x[0]+(-0.339720289405)*x[1]**o+(-0.283077066751)*x[1]+(0.625600522278)*x[2]**o+(-0.132697432298)*x[2]
        arg[3,3,0,0]=(-0.346408105223)*x[0]**o+(0.299760534041)*x[0]+(0.567951405594)*x[1]**o+(-0.880683463459)*x[1]+(0.784127947661)*x[2]**o+(0.268895744072)*x[2]
        arg[3,3,0,1]=(0.064159345936)*x[0]**o+(-0.685098798117)*x[0]+(-0.182310885124)*x[1]**o+(-0.700081075257)*x[1]+(0.543959312437)*x[2]**o+(0.125104170434)*x[2]
        arg[3,3,1,0]=(0.460375837083)*x[0]**o+(0.82355800923)*x[0]+(-0.186000244328)*x[1]**o+(0.8418355439)*x[1]+(-0.375269292115)*x[2]**o+(0.712431840155)*x[2]
        arg[3,3,1,1]=(0.57260867805)*x[0]**o+(-0.146405851776)*x[0]+(-0.50640779246)*x[1]**o+(0.741760309299)*x[1]+(0.069752390409)*x[2]**o+(-0.376724191471)*x[2]
        arg[3,3,2,0]=(0.795493408403)*x[0]**o+(0.396861648344)*x[0]+(-0.728716068325)*x[1]**o+(0.739330072416)*x[1]+(0.202019645585)*x[2]**o+(0.841240509566)*x[2]
        arg[3,3,2,1]=(-0.967372744621)*x[0]**o+(-0.334044549662)*x[0]+(-0.0686982670779)*x[1]**o+(0.180083293629)*x[1]+(-0.208697793793)*x[2]**o+(0.0305446253751)*x[2]
        arg[3,4,0,0]=(0.94168350996)*x[0]**o+(0.0545532586137)*x[0]+(-0.300430532207)*x[1]**o+(-0.960362884765)*x[1]+(0.738261342527)*x[2]**o+(-0.351477574571)*x[2]
        arg[3,4,0,1]=(-0.525893857873)*x[0]**o+(0.227252001588)*x[0]+(0.179831911446)*x[1]**o+(0.350952712138)*x[1]+(-0.671550635312)*x[2]**o+(-0.338834217909)*x[2]
        arg[3,4,1,0]=(-0.60610056656)*x[0]**o+(0.219192673102)*x[0]+(-0.801595717702)*x[1]**o+(-0.696853420225)*x[1]+(-0.495201990047)*x[2]**o+(-0.500683649414)*x[2]
        arg[3,4,1,1]=(-0.37343020332)*x[0]**o+(-0.585536591527)*x[0]+(0.81794997494)*x[1]**o+(-0.240232527585)*x[1]+(-0.538970681636)*x[2]**o+(-0.146105210473)*x[2]
        arg[3,4,2,0]=(0.857322149358)*x[0]**o+(-0.552847286493)*x[0]+(0.530970378294)*x[1]**o+(0.0746811386922)*x[1]+(0.946170976628)*x[2]**o+(-0.329916962713)*x[2]
        arg[3,4,2,1]=(-0.353323100685)*x[0]**o+(0.808796594618)*x[0]+(0.481450588496)*x[1]**o+(0.1202874944)*x[1]+(0.671293736572)*x[2]**o+(-0.905074286629)*x[2]
        ref[0,0,0,0]=(-0.613904299283)*(1+2.*(dim-1.)/(o+1.))+(1.30218009705)*dim
        ref[0,0,0,1]=(-0.810525773084)*(1+2.*(dim-1.)/(o+1.))+(0.253224626501)*dim
        ref[0,0,1,0]=(0.509154413012)*(1+2.*(dim-1.)/(o+1.))+(-0.179784708198)*dim
        ref[0,0,1,1]=(0.637556597632)*(1+2.*(dim-1.)/(o+1.))+(-0.693313683923)*dim
        ref[0,0,2,0]=(-0.465137153331)*(1+2.*(dim-1.)/(o+1.))+(-0.0477989622018)*dim
        ref[0,0,2,1]=(-2.15185435439)*(1+2.*(dim-1.)/(o+1.))+(0.231459735932)*dim
        ref[0,1,0,0]=(0.8018922964)*(1+2.*(dim-1.)/(o+1.))+(-1.10884149089)*dim
        ref[0,1,0,1]=(1.7955198411)*(1+2.*(dim-1.)/(o+1.))+(-1.34662500856)*dim
        ref[0,1,1,0]=(0.584854782162)*(1+2.*(dim-1.)/(o+1.))+(0.274087751285)*dim
        ref[0,1,1,1]=(-1.0401305206)*(1+2.*(dim-1.)/(o+1.))+(1.85561593667)*dim
        ref[0,1,2,0]=(-0.834154383324)*(1+2.*(dim-1.)/(o+1.))+(-1.23273379028)*dim
        ref[0,1,2,1]=(-2.34017474507)*(1+2.*(dim-1.)/(o+1.))+(-0.224176787111)*dim
        ref[0,2,0,0]=(0.613915916089)*(1+2.*(dim-1.)/(o+1.))+(0.753829204471)*dim
        ref[0,2,0,1]=(0.131344565765)*(1+2.*(dim-1.)/(o+1.))+(-0.976652361934)*dim
        ref[0,2,1,0]=(0.579116187931)*(1+2.*(dim-1.)/(o+1.))+(0.0228071895801)*dim
        ref[0,2,1,1]=(0.772304265705)*(1+2.*(dim-1.)/(o+1.))+(0.470483133313)*dim
        ref[0,2,2,0]=(-1.12957943492)*(1+2.*(dim-1.)/(o+1.))+(-0.82429788127)*dim
        ref[0,2,2,1]=(-0.631840107715)*(1+2.*(dim-1.)/(o+1.))+(-0.732399256131)*dim
        ref[0,3,0,0]=(0.137323356087)*(1+2.*(dim-1.)/(o+1.))+(-0.413776023245)*dim
        ref[0,3,0,1]=(-0.173740822652)*(1+2.*(dim-1.)/(o+1.))+(1.19050855195)*dim
        ref[0,3,1,0]=(-1.58068224297)*(1+2.*(dim-1.)/(o+1.))+(0.24224399322)*dim
        ref[0,3,1,1]=(0.474472608494)*(1+2.*(dim-1.)/(o+1.))+(-0.557328900938)*dim
        ref[0,3,2,0]=(0.459842497633)*(1+2.*(dim-1.)/(o+1.))+(-1.09001004859)*dim
        ref[0,3,2,1]=(-0.22027884782)*(1+2.*(dim-1.)/(o+1.))+(0.235715331835)*dim
        ref[0,4,0,0]=(0.424803567504)*(1+2.*(dim-1.)/(o+1.))+(0.657196485077)*dim
        ref[0,4,0,1]=(-0.00655513622033)*(1+2.*(dim-1.)/(o+1.))+(-0.803228043207)*dim
        ref[0,4,1,0]=(1.33738998187)*(1+2.*(dim-1.)/(o+1.))+(-0.0257326162636)*dim
        ref[0,4,1,1]=(-1.55536089118)*(1+2.*(dim-1.)/(o+1.))+(-0.0357690415245)*dim
        ref[0,4,2,0]=(1.58679207216)*(1+2.*(dim-1.)/(o+1.))+(-0.693598479122)*dim
        ref[0,4,2,1]=(0.0628502197447)*(1+2.*(dim-1.)/(o+1.))+(-1.90742115249)*dim
        ref[1,0,0,0]=(-1.25774299848)*(1+2.*(dim-1.)/(o+1.))+(-0.0786696801461)*dim
        ref[1,0,0,1]=(-0.807603352133)*(1+2.*(dim-1.)/(o+1.))+(-0.429302209684)*dim
        ref[1,0,1,0]=(1.14217071628)*(1+2.*(dim-1.)/(o+1.))+(0.740396951119)*dim
        ref[1,0,1,1]=(1.89672043448)*(1+2.*(dim-1.)/(o+1.))+(1.4016996376)*dim
        ref[1,0,2,0]=(-0.768614149462)*(1+2.*(dim-1.)/(o+1.))+(0.913988010751)*dim
        ref[1,0,2,1]=(-0.601884288575)*(1+2.*(dim-1.)/(o+1.))+(-1.67541838174)*dim
        ref[1,1,0,0]=(-1.43981092176)*(1+2.*(dim-1.)/(o+1.))+(-0.506725297145)*dim
        ref[1,1,0,1]=(0.503113976474)*(1+2.*(dim-1.)/(o+1.))+(-0.488618815974)*dim
        ref[1,1,1,0]=(0.815915054472)*(1+2.*(dim-1.)/(o+1.))+(-0.522904536573)*dim
        ref[1,1,1,1]=(0.310174793362)*(1+2.*(dim-1.)/(o+1.))+(0.825570415579)*dim
        ref[1,1,2,0]=(-1.85870537865)*(1+2.*(dim-1.)/(o+1.))+(-0.896366119383)*dim
        ref[1,1,2,1]=(0.634330156916)*(1+2.*(dim-1.)/(o+1.))+(-0.202404975156)*dim
        ref[1,2,0,0]=(0.980180080393)*(1+2.*(dim-1.)/(o+1.))+(1.44073022185)*dim
        ref[1,2,0,1]=(1.16930139908)*(1+2.*(dim-1.)/(o+1.))+(-1.29606923835)*dim
        ref[1,2,1,0]=(0.251262820359)*(1+2.*(dim-1.)/(o+1.))+(-0.534365875316)*dim
        ref[1,2,1,1]=(1.45444347927)*(1+2.*(dim-1.)/(o+1.))+(1.09104332726)*dim
        ref[1,2,2,0]=(0.36826076644)*(1+2.*(dim-1.)/(o+1.))+(-0.222843807345)*dim
        ref[1,2,2,1]=(0.652062496151)*(1+2.*(dim-1.)/(o+1.))+(1.45032712876)*dim
        ref[1,3,0,0]=(0.357179082647)*(1+2.*(dim-1.)/(o+1.))+(0.96345428327)*dim
        ref[1,3,0,1]=(0.445836034597)*(1+2.*(dim-1.)/(o+1.))+(-0.768814370045)*dim
        ref[1,3,1,0]=(-0.633644413575)*(1+2.*(dim-1.)/(o+1.))+(0.996183349509)*dim
        ref[1,3,1,1]=(-1.47637436236)*(1+2.*(dim-1.)/(o+1.))+(-0.446000824477)*dim
        ref[1,3,2,0]=(2.22742580994)*(1+2.*(dim-1.)/(o+1.))+(0.238170411449)*dim
        ref[1,3,2,1]=(0.138824208277)*(1+2.*(dim-1.)/(o+1.))+(1.69436291732)*dim
        ref[1,4,0,0]=(-0.622328917175)*(1+2.*(dim-1.)/(o+1.))+(-1.12401903264)*dim
        ref[1,4,0,1]=(0.40305181353)*(1+2.*(dim-1.)/(o+1.))+(1.68001892165)*dim
        ref[1,4,1,0]=(-1.16479419403)*(1+2.*(dim-1.)/(o+1.))+(1.05838059836)*dim
        ref[1,4,1,1]=(0.0801376319834)*(1+2.*(dim-1.)/(o+1.))+(2.09228066438)*dim
        ref[1,4,2,0]=(-0.228762226067)*(1+2.*(dim-1.)/(o+1.))+(0.523454314851)*dim
        ref[1,4,2,1]=(-1.25310430418)*(1+2.*(dim-1.)/(o+1.))+(0.296827734822)*dim
        ref[2,0,0,0]=(1.14292760919)*(1+2.*(dim-1.)/(o+1.))+(-0.520826483003)*dim
        ref[2,0,0,1]=(1.16992671589)*(1+2.*(dim-1.)/(o+1.))+(-0.162899835641)*dim
        ref[2,0,1,0]=(-0.544683088018)*(1+2.*(dim-1.)/(o+1.))+(-1.44956627836)*dim
        ref[2,0,1,1]=(-1.46970715451)*(1+2.*(dim-1.)/(o+1.))+(-0.403035460127)*dim
        ref[2,0,2,0]=(-0.385627069865)*(1+2.*(dim-1.)/(o+1.))+(-0.829501029966)*dim
        ref[2,0,2,1]=(1.11664736171)*(1+2.*(dim-1.)/(o+1.))+(-1.27685980972)*dim
        ref[2,1,0,0]=(1.08669102227)*(1+2.*(dim-1.)/(o+1.))+(1.7990592379)*dim
        ref[2,1,0,1]=(0.0453423472143)*(1+2.*(dim-1.)/(o+1.))+(-0.68150455532)*dim
        ref[2,1,1,0]=(0.0322959644566)*(1+2.*(dim-1.)/(o+1.))+(-2.23247758093)*dim
        ref[2,1,1,1]=(-0.892578997912)*(1+2.*(dim-1.)/(o+1.))+(0.388356392315)*dim
        ref[2,1,2,0]=(-0.7179937061)*(1+2.*(dim-1.)/(o+1.))+(-1.52564228261)*dim
        ref[2,1,2,1]=(0.950438825754)*(1+2.*(dim-1.)/(o+1.))+(0.342228484532)*dim
        ref[2,2,0,0]=(1.0779882212)*(1+2.*(dim-1.)/(o+1.))+(0.796214693049)*dim
        ref[2,2,0,1]=(-0.115299804062)*(1+2.*(dim-1.)/(o+1.))+(0.798756112891)*dim
        ref[2,2,1,0]=(0.536406467403)*(1+2.*(dim-1.)/(o+1.))+(-1.139126385)*dim
        ref[2,2,1,1]=(-0.779767608394)*(1+2.*(dim-1.)/(o+1.))+(0.522828453067)*dim
        ref[2,2,2,0]=(-0.411536633652)*(1+2.*(dim-1.)/(o+1.))+(-0.895144122064)*dim
        ref[2,2,2,1]=(1.56944190392)*(1+2.*(dim-1.)/(o+1.))+(-0.12659143313)*dim
        ref[2,3,0,0]=(-0.372936173925)*(1+2.*(dim-1.)/(o+1.))+(-0.938385787171)*dim
        ref[2,3,0,1]=(1.33113027561)*(1+2.*(dim-1.)/(o+1.))+(1.39154904574)*dim
        ref[2,3,1,0]=(0.418458834548)*(1+2.*(dim-1.)/(o+1.))+(2.20921588131)*dim
        ref[2,3,1,1]=(0.64133847569)*(1+2.*(dim-1.)/(o+1.))+(0.556373927504)*dim
        ref[2,3,2,0]=(-0.689346506499)*(1+2.*(dim-1.)/(o+1.))+(1.63847691387)*dim
        ref[2,3,2,1]=(-1.23371702875)*(1+2.*(dim-1.)/(o+1.))+(-0.384869997042)*dim
        ref[2,4,0,0]=(0.433647132788)*(1+2.*(dim-1.)/(o+1.))+(-0.527688846368)*dim
        ref[2,4,0,1]=(1.03032814854)*(1+2.*(dim-1.)/(o+1.))+(-0.103322527004)*dim
        ref[2,4,1,0]=(0.663568613046)*(1+2.*(dim-1.)/(o+1.))+(1.1370080288)*dim
        ref[2,4,1,1]=(-0.977905093893)*(1+2.*(dim-1.)/(o+1.))+(0.0866635633103)*dim
        ref[2,4,2,0]=(0.085735636585)*(1+2.*(dim-1.)/(o+1.))+(-0.234684579572)*dim
        ref[2,4,2,1]=(-0.311909123795)*(1+2.*(dim-1.)/(o+1.))+(-0.705520095044)*dim
        ref[3,0,0,0]=(-0.780115379753)*(1+2.*(dim-1.)/(o+1.))+(-0.553421227188)*dim
        ref[3,0,0,1]=(-2.33835441041)*(1+2.*(dim-1.)/(o+1.))+(0.648874255143)*dim
        ref[3,0,1,0]=(-0.0921929146543)*(1+2.*(dim-1.)/(o+1.))+(-0.949868130291)*dim
        ref[3,0,1,1]=(0.231464433969)*(1+2.*(dim-1.)/(o+1.))+(0.370739594929)*dim
        ref[3,0,2,0]=(0.666293785727)*(1+2.*(dim-1.)/(o+1.))+(0.933961354368)*dim
        ref[3,0,2,1]=(0.0240637548661)*(1+2.*(dim-1.)/(o+1.))+(-0.342585987504)*dim
        ref[3,1,0,0]=(0.250534290654)*(1+2.*(dim-1.)/(o+1.))+(0.391937952644)*dim
        ref[3,1,0,1]=(-0.114349866018)*(1+2.*(dim-1.)/(o+1.))+(1.54390216157)*dim
        ref[3,1,1,0]=(-1.20423411082)*(1+2.*(dim-1.)/(o+1.))+(-1.80846594625)*dim
        ref[3,1,1,1]=(-0.679747643349)*(1+2.*(dim-1.)/(o+1.))+(0.586471246578)*dim
        ref[3,1,2,0]=(-1.87640663683)*(1+2.*(dim-1.)/(o+1.))+(-1.20650936788)*dim
        ref[3,1,2,1]=(0.221597815849)*(1+2.*(dim-1.)/(o+1.))+(-0.189673037401)*dim
        ref[3,2,0,0]=(0.40161631028)*(1+2.*(dim-1.)/(o+1.))+(-0.784860114318)*dim
        ref[3,2,0,1]=(-0.265289560586)*(1+2.*(dim-1.)/(o+1.))+(0.0190519596518)*dim
        ref[3,2,1,0]=(0.522308956237)*(1+2.*(dim-1.)/(o+1.))+(0.0431481382473)*dim
        ref[3,2,1,1]=(0.585059743072)*(1+2.*(dim-1.)/(o+1.))+(-2.23136772548)*dim
        ref[3,2,2,0]=(-0.126352236735)*(1+2.*(dim-1.)/(o+1.))+(-0.112757452698)*dim
        ref[3,2,2,1]=(0.0228968924883)*(1+2.*(dim-1.)/(o+1.))+(-1.17901102753)*dim
        ref[3,3,0,0]=(1.00567124803)*(1+2.*(dim-1.)/(o+1.))+(-0.312027185345)*dim
        ref[3,3,0,1]=(0.425807773249)*(1+2.*(dim-1.)/(o+1.))+(-1.26007570294)*dim
        ref[3,3,1,0]=(-0.10089369936)*(1+2.*(dim-1.)/(o+1.))+(2.37782539328)*dim
        ref[3,3,1,1]=(0.135953275999)*(1+2.*(dim-1.)/(o+1.))+(0.218630266051)*dim
        ref[3,3,2,0]=(0.268796985663)*(1+2.*(dim-1.)/(o+1.))+(1.97743223033)*dim
        ref[3,3,2,1]=(-1.24476880549)*(1+2.*(dim-1.)/(o+1.))+(-0.123416630658)*dim
        ref[3,4,0,0]=(1.37951432028)*(1+2.*(dim-1.)/(o+1.))+(-1.25728720072)*dim
        ref[3,4,0,1]=(-1.01761258174)*(1+2.*(dim-1.)/(o+1.))+(0.239370495817)*dim
        ref[3,4,1,0]=(-1.90289827431)*(1+2.*(dim-1.)/(o+1.))+(-0.978344396537)*dim
        ref[3,4,1,1]=(-0.0944509100161)*(1+2.*(dim-1.)/(o+1.))+(-0.971874329585)*dim
        ref[3,4,2,0]=(2.33446350428)*(1+2.*(dim-1.)/(o+1.))+(-0.808083110513)*dim
        ref[3,4,2,1]=(0.799421224383)*(1+2.*(dim-1.)/(o+1.))+(0.0240098023897)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.626062912046)*x[0]**o+(0.600712856149)*x[0]+(0.0478861882442)*x[1]**o+(0.77008552734)*x[1]
        ref=(0.67394910029)*(1+2.*(dim-1.)/(o+1.))+(1.37079838349)*dim
      else:
        arg=(0.85588880126)*x[0]**o+(-0.898804533462)*x[0]+(-0.270816921088)*x[1]**o+(-0.842823912373)*x[1]+(-0.47559690762)*x[2]**o+(0.751770125129)*x[2]
        ref=(0.109474972552)*(1+2.*(dim-1.)/(o+1.))+(-0.989858320706)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.298913063662)*x[0]**o+(-0.250822875451)*x[0]+(0.707311253225)*x[1]**o+(0.0408091445557)*x[1]
        arg[1]=(-0.70143850042)*x[0]**o+(-0.80121227022)*x[0]+(0.0731514590408)*x[1]**o+(0.621969876113)*x[1]
        ref[0]=(1.00622431689)*(1+2.*(dim-1.)/(o+1.))+(-0.210013730895)*dim
        ref[1]=(-0.628287041379)*(1+2.*(dim-1.)/(o+1.))+(-0.179242394108)*dim
      else:
        arg[0]=(0.676976938336)*x[0]**o+(-0.928207054947)*x[0]+(0.864749399199)*x[1]**o+(0.864358390848)*x[1]+(-0.0255569869984)*x[2]**o+(0.0268536872793)*x[2]
        arg[1]=(0.270542056965)*x[0]**o+(0.0696194435081)*x[0]+(0.401360001043)*x[1]**o+(0.917064528874)*x[1]+(-0.755510144559)*x[2]**o+(0.881244515908)*x[2]
        ref[0]=(1.51616935054)*(1+2.*(dim-1.)/(o+1.))+(-0.0369949768193)*dim
        ref[1]=(-0.083608086551)*(1+2.*(dim-1.)/(o+1.))+(1.86792848829)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.381843423534)*x[0]**o+(-0.563210012057)*x[0]+(0.730631055205)*x[1]**o+(0.13636238698)*x[1]
        arg[0,1]=(0.306357653338)*x[0]**o+(-0.797201232101)*x[0]+(-0.66333793372)*x[1]**o+(-0.943650360464)*x[1]
        arg[0,2]=(-0.34058649734)*x[0]**o+(-0.27047792232)*x[0]+(0.57156687267)*x[1]**o+(-0.659114355468)*x[1]
        arg[0,3]=(0.73319004289)*x[0]**o+(0.229139540294)*x[0]+(-0.51558765225)*x[1]**o+(0.45052521066)*x[1]
        arg[0,4]=(-0.264117198359)*x[0]**o+(0.807950109095)*x[0]+(-0.290558748289)*x[1]**o+(0.161864708045)*x[1]
        arg[1,0]=(0.988732415297)*x[0]**o+(-0.725047414136)*x[0]+(-0.722337599885)*x[1]**o+(0.739130135256)*x[1]
        arg[1,1]=(-0.315086233906)*x[0]**o+(-0.783177846939)*x[0]+(-0.271931454399)*x[1]**o+(-0.935744613016)*x[1]
        arg[1,2]=(-0.787047622811)*x[0]**o+(0.00746242211892)*x[0]+(-0.539716273578)*x[1]**o+(-0.340449115622)*x[1]
        arg[1,3]=(-0.782830405823)*x[0]**o+(0.842318742811)*x[0]+(-0.118894493096)*x[1]**o+(0.645943277412)*x[1]
        arg[1,4]=(0.0594068367647)*x[0]**o+(-0.184711682975)*x[0]+(-0.0643050731295)*x[1]**o+(-0.981593287261)*x[1]
        arg[2,0]=(0.17170754931)*x[0]**o+(0.582651363494)*x[0]+(0.334422749985)*x[1]**o+(0.852747429144)*x[1]
        arg[2,1]=(0.683967516208)*x[0]**o+(-0.612989314204)*x[0]+(0.845175135723)*x[1]**o+(0.374223395314)*x[1]
        arg[2,2]=(0.588917536035)*x[0]**o+(0.0130713873939)*x[0]+(0.881942872071)*x[1]**o+(0.683941046151)*x[1]
        arg[2,3]=(-0.63861264887)*x[0]**o+(0.453919752415)*x[0]+(0.448731342326)*x[1]**o+(-0.259183783496)*x[1]
        arg[2,4]=(-0.132175638549)*x[0]**o+(0.710071566415)*x[0]+(-0.870119332524)*x[1]**o+(0.523394538403)*x[1]
        arg[3,0]=(0.768036036634)*x[0]**o+(-0.218321673662)*x[0]+(0.50224512768)*x[1]**o+(0.694191572559)*x[1]
        arg[3,1]=(-0.298369075336)*x[0]**o+(-0.276776010079)*x[0]+(-0.659309162004)*x[1]**o+(0.436841182692)*x[1]
        arg[3,2]=(-0.00370049871499)*x[0]**o+(-0.41700259949)*x[0]+(0.726806931339)*x[1]**o+(-0.113165181924)*x[1]
        arg[3,3]=(0.288175924931)*x[0]**o+(-0.489667003521)*x[0]+(-0.531842953651)*x[1]**o+(-0.443517285929)*x[1]
        arg[3,4]=(-0.645574931589)*x[0]**o+(0.464738749544)*x[0]+(-0.538213524368)*x[1]**o+(-0.94443292146)*x[1]
        ref[0,0]=(0.348787631671)*(1+2.*(dim-1.)/(o+1.))+(-0.426847625077)*dim
        ref[0,1]=(-0.356980280382)*(1+2.*(dim-1.)/(o+1.))+(-1.74085159257)*dim
        ref[0,2]=(0.23098037533)*(1+2.*(dim-1.)/(o+1.))+(-0.929592277787)*dim
        ref[0,3]=(0.21760239064)*(1+2.*(dim-1.)/(o+1.))+(0.679664750954)*dim
        ref[0,4]=(-0.554675946648)*(1+2.*(dim-1.)/(o+1.))+(0.96981481714)*dim
        ref[1,0]=(0.266394815412)*(1+2.*(dim-1.)/(o+1.))+(0.0140827211195)*dim
        ref[1,1]=(-0.587017688304)*(1+2.*(dim-1.)/(o+1.))+(-1.71892245996)*dim
        ref[1,2]=(-1.32676389639)*(1+2.*(dim-1.)/(o+1.))+(-0.332986693503)*dim
        ref[1,3]=(-0.901724898919)*(1+2.*(dim-1.)/(o+1.))+(1.48826202022)*dim
        ref[1,4]=(-0.00489823636486)*(1+2.*(dim-1.)/(o+1.))+(-1.16630497024)*dim
        ref[2,0]=(0.506130299295)*(1+2.*(dim-1.)/(o+1.))+(1.43539879264)*dim
        ref[2,1]=(1.52914265193)*(1+2.*(dim-1.)/(o+1.))+(-0.238765918889)*dim
        ref[2,2]=(1.47086040811)*(1+2.*(dim-1.)/(o+1.))+(0.697012433545)*dim
        ref[2,3]=(-0.189881306544)*(1+2.*(dim-1.)/(o+1.))+(0.194735968919)*dim
        ref[2,4]=(-1.00229497107)*(1+2.*(dim-1.)/(o+1.))+(1.23346610482)*dim
        ref[3,0]=(1.27028116431)*(1+2.*(dim-1.)/(o+1.))+(0.475869898897)*dim
        ref[3,1]=(-0.957678237339)*(1+2.*(dim-1.)/(o+1.))+(0.160065172613)*dim
        ref[3,2]=(0.723106432624)*(1+2.*(dim-1.)/(o+1.))+(-0.530167781414)*dim
        ref[3,3]=(-0.243667028719)*(1+2.*(dim-1.)/(o+1.))+(-0.93318428945)*dim
        ref[3,4]=(-1.18378845596)*(1+2.*(dim-1.)/(o+1.))+(-0.479694171917)*dim
      else:
        arg[0,0]=(0.751222317885)*x[0]**o+(0.953666267123)*x[0]+(-0.707417048158)*x[1]**o+(0.310686543483)*x[1]+(0.326267797641)*x[2]**o+(-0.722772500437)*x[2]
        arg[0,1]=(-0.038709086264)*x[0]**o+(-0.483307819428)*x[0]+(0.611120835303)*x[1]**o+(0.885935954604)*x[1]+(0.630199910046)*x[2]**o+(-0.101764126615)*x[2]
        arg[0,2]=(0.890606731439)*x[0]**o+(-0.0555491679283)*x[0]+(0.334835177686)*x[1]**o+(0.999030888839)*x[1]+(-0.891653525154)*x[2]**o+(-0.14056280616)*x[2]
        arg[0,3]=(-0.458305223634)*x[0]**o+(0.679075077126)*x[0]+(0.1176620176)*x[1]**o+(0.685223087684)*x[1]+(-0.952262623464)*x[2]**o+(-0.905568306628)*x[2]
        arg[0,4]=(0.623179050934)*x[0]**o+(-0.0150847487218)*x[0]+(0.0823495403769)*x[1]**o+(0.48303169009)*x[1]+(0.610276253905)*x[2]**o+(0.353104308072)*x[2]
        arg[1,0]=(-0.709185342895)*x[0]**o+(0.789980598884)*x[0]+(0.655538102878)*x[1]**o+(0.713913568814)*x[1]+(0.222131732773)*x[2]**o+(0.712644729381)*x[2]
        arg[1,1]=(-0.683646421822)*x[0]**o+(-0.943691516528)*x[0]+(-0.134261002557)*x[1]**o+(0.28225941018)*x[1]+(0.906235993779)*x[2]**o+(-0.819214603609)*x[2]
        arg[1,2]=(0.713781993938)*x[0]**o+(-0.189551367619)*x[0]+(-0.021310841804)*x[1]**o+(-0.873466428012)*x[1]+(0.248337664546)*x[2]**o+(0.911824874935)*x[2]
        arg[1,3]=(-0.393142772637)*x[0]**o+(0.773501237865)*x[0]+(0.759039429922)*x[1]**o+(0.67036876499)*x[1]+(-0.921411979023)*x[2]**o+(-0.649371534276)*x[2]
        arg[1,4]=(0.526708082723)*x[0]**o+(-0.522449586718)*x[0]+(0.951176548558)*x[1]**o+(0.604602039898)*x[1]+(0.345888192325)*x[2]**o+(-0.973729954334)*x[2]
        arg[2,0]=(-0.139931215293)*x[0]**o+(-0.996072915638)*x[0]+(-0.0451731431264)*x[1]**o+(0.433694224874)*x[1]+(-0.955189760302)*x[2]**o+(-0.68361667242)*x[2]
        arg[2,1]=(0.139626552661)*x[0]**o+(-0.91217468751)*x[0]+(-0.854252204911)*x[1]**o+(0.417539913464)*x[1]+(0.557096050124)*x[2]**o+(-0.195272359868)*x[2]
        arg[2,2]=(0.960352725049)*x[0]**o+(0.98274500341)*x[0]+(0.967782565025)*x[1]**o+(0.566011353381)*x[1]+(0.583885094583)*x[2]**o+(-0.943173847936)*x[2]
        arg[2,3]=(0.959388333494)*x[0]**o+(0.531908797155)*x[0]+(-0.62531126203)*x[1]**o+(-0.288763293646)*x[1]+(-0.823398780526)*x[2]**o+(-0.164348871248)*x[2]
        arg[2,4]=(0.73397702179)*x[0]**o+(0.407844592399)*x[0]+(-0.835769355727)*x[1]**o+(0.742360626394)*x[1]+(0.31130186529)*x[2]**o+(0.114221168872)*x[2]
        arg[3,0]=(-0.42278940444)*x[0]**o+(-0.513158064442)*x[0]+(-0.133233894233)*x[1]**o+(0.701381913887)*x[1]+(0.283776333553)*x[2]**o+(-0.805096661642)*x[2]
        arg[3,1]=(-0.417338508269)*x[0]**o+(0.535961900783)*x[0]+(0.0836386907734)*x[1]**o+(0.449680928704)*x[1]+(0.998998322868)*x[2]**o+(0.983057825698)*x[2]
        arg[3,2]=(0.24102382276)*x[0]**o+(0.0291470157398)*x[0]+(-0.539542451698)*x[1]**o+(0.281017043631)*x[1]+(-0.464652133499)*x[2]**o+(0.652828720185)*x[2]
        arg[3,3]=(-0.482577920013)*x[0]**o+(0.931756756394)*x[0]+(0.96648206338)*x[1]**o+(0.73941459549)*x[1]+(0.136743015429)*x[2]**o+(0.41477914722)*x[2]
        arg[3,4]=(0.879881885972)*x[0]**o+(0.208970079585)*x[0]+(-0.200079958414)*x[1]**o+(-0.0291238772241)*x[1]+(-0.333569505857)*x[2]**o+(0.846835729875)*x[2]
        ref[0,0]=(0.370073067368)*(1+2.*(dim-1.)/(o+1.))+(0.541580310169)*dim
        ref[0,1]=(1.20261165909)*(1+2.*(dim-1.)/(o+1.))+(0.300864008561)*dim
        ref[0,2]=(0.333788383971)*(1+2.*(dim-1.)/(o+1.))+(0.80291891475)*dim
        ref[0,3]=(-1.2929058295)*(1+2.*(dim-1.)/(o+1.))+(0.458729858183)*dim
        ref[0,4]=(1.31580484522)*(1+2.*(dim-1.)/(o+1.))+(0.82105124944)*dim
        ref[1,0]=(0.168484492756)*(1+2.*(dim-1.)/(o+1.))+(2.21653889708)*dim
        ref[1,1]=(0.0883285694004)*(1+2.*(dim-1.)/(o+1.))+(-1.48064670996)*dim
        ref[1,2]=(0.940808816681)*(1+2.*(dim-1.)/(o+1.))+(-0.151192920696)*dim
        ref[1,3]=(-0.555515321739)*(1+2.*(dim-1.)/(o+1.))+(0.79449846858)*dim
        ref[1,4]=(1.82377282361)*(1+2.*(dim-1.)/(o+1.))+(-0.891577501154)*dim
        ref[2,0]=(-1.14029411872)*(1+2.*(dim-1.)/(o+1.))+(-1.24599536318)*dim
        ref[2,1]=(-0.157529602127)*(1+2.*(dim-1.)/(o+1.))+(-0.689907133913)*dim
        ref[2,2]=(2.51202038466)*(1+2.*(dim-1.)/(o+1.))+(0.605582508855)*dim
        ref[2,3]=(-0.489321709063)*(1+2.*(dim-1.)/(o+1.))+(0.0787966322607)*dim
        ref[2,4]=(0.209509531353)*(1+2.*(dim-1.)/(o+1.))+(1.26442638766)*dim
        ref[3,0]=(-0.27224696512)*(1+2.*(dim-1.)/(o+1.))+(-0.616872812198)*dim
        ref[3,1]=(0.665298505372)*(1+2.*(dim-1.)/(o+1.))+(1.96870065519)*dim
        ref[3,2]=(-0.763170762437)*(1+2.*(dim-1.)/(o+1.))+(0.962992779555)*dim
        ref[3,3]=(0.620647158797)*(1+2.*(dim-1.)/(o+1.))+(2.0859504991)*dim
        ref[3,4]=(0.346232421701)*(1+2.*(dim-1.)/(o+1.))+(1.02668193224)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.201115186772)*x[0]**o+(0.324758103748)*x[0]+(0.285065583242)*x[1]**o+(0.0264985832782)*x[1]
        arg[0,0,1]=(-0.443023003973)*x[0]**o+(-0.194732927607)*x[0]+(0.64414908519)*x[1]**o+(-0.248261857779)*x[1]
        arg[0,1,0]=(0.587534172195)*x[0]**o+(0.021088951369)*x[0]+(-0.986195557459)*x[1]**o+(0.595659005312)*x[1]
        arg[0,1,1]=(-0.916961145134)*x[0]**o+(-0.766484469605)*x[0]+(-0.281345964683)*x[1]**o+(-0.399113856826)*x[1]
        arg[1,0,0]=(0.175943424949)*x[0]**o+(0.344853386001)*x[0]+(0.175474681848)*x[1]**o+(-0.0775102874697)*x[1]
        arg[1,0,1]=(-0.874074467007)*x[0]**o+(0.903506914387)*x[0]+(0.177699372538)*x[1]**o+(0.969818154968)*x[1]
        arg[1,1,0]=(0.765751472387)*x[0]**o+(-0.0288724203687)*x[0]+(0.655676828008)*x[1]**o+(0.22281307247)*x[1]
        arg[1,1,1]=(0.871009077152)*x[0]**o+(-0.20052353101)*x[0]+(-0.889496767621)*x[1]**o+(0.8458762887)*x[1]
        arg[2,0,0]=(-0.15377513542)*x[0]**o+(0.963714892768)*x[0]+(0.942514815983)*x[1]**o+(0.846111018838)*x[1]
        arg[2,0,1]=(0.000727086938663)*x[0]**o+(0.243136987492)*x[0]+(-0.308674899886)*x[1]**o+(-0.963435664436)*x[1]
        arg[2,1,0]=(0.780912245051)*x[0]**o+(0.227139041778)*x[0]+(-0.336203258384)*x[1]**o+(0.092568284679)*x[1]
        arg[2,1,1]=(0.0464862681168)*x[0]**o+(0.284493375176)*x[0]+(-0.811073509077)*x[1]**o+(0.54280235419)*x[1]
        arg[3,0,0]=(0.605642304612)*x[0]**o+(-0.103442921567)*x[0]+(0.106253324829)*x[1]**o+(0.562480007904)*x[1]
        arg[3,0,1]=(0.497668538008)*x[0]**o+(-0.859679982844)*x[0]+(-0.255499355194)*x[1]**o+(0.0725919239243)*x[1]
        arg[3,1,0]=(0.641826893201)*x[0]**o+(0.889514581887)*x[0]+(-0.211781857836)*x[1]**o+(0.45900873193)*x[1]
        arg[3,1,1]=(-0.29349600565)*x[0]**o+(-0.627542961817)*x[0]+(0.313009642686)*x[1]**o+(0.338028134809)*x[1]
        arg[4,0,0]=(-0.641915241397)*x[0]**o+(-0.682547311604)*x[0]+(-0.106472077451)*x[1]**o+(0.103402239052)*x[1]
        arg[4,0,1]=(-0.822562520978)*x[0]**o+(-0.430539530156)*x[0]+(-0.0225528898923)*x[1]**o+(-0.0908986665674)*x[1]
        arg[4,1,0]=(-0.456160645978)*x[0]**o+(-0.264592336242)*x[0]+(-0.672702978868)*x[1]**o+(-0.893197955639)*x[1]
        arg[4,1,1]=(-0.714689169676)*x[0]**o+(0.513612105215)*x[0]+(-0.3093340377)*x[1]**o+(-0.34467035156)*x[1]
        arg[5,0,0]=(-0.993761664154)*x[0]**o+(0.375953817044)*x[0]+(0.0463240503072)*x[1]**o+(-0.595646829466)*x[1]
        arg[5,0,1]=(0.453145186545)*x[0]**o+(0.331353840672)*x[0]+(-0.908533078632)*x[1]**o+(0.755175839674)*x[1]
        arg[5,1,0]=(-0.736874540101)*x[0]**o+(-0.867437170846)*x[0]+(-0.472395301622)*x[1]**o+(0.78289019661)*x[1]
        arg[5,1,1]=(-0.900966156016)*x[0]**o+(-0.160363876891)*x[0]+(-0.0969222200475)*x[1]**o+(0.0527983470624)*x[1]
        ref[0,0,0]=(0.486180770014)*(1+2.*(dim-1.)/(o+1.))+(0.351256687026)*dim
        ref[0,0,1]=(0.201126081217)*(1+2.*(dim-1.)/(o+1.))+(-0.442994785386)*dim
        ref[0,1,0]=(-0.398661385264)*(1+2.*(dim-1.)/(o+1.))+(0.616747956681)*dim
        ref[0,1,1]=(-1.19830710982)*(1+2.*(dim-1.)/(o+1.))+(-1.16559832643)*dim
        ref[1,0,0]=(0.351418106797)*(1+2.*(dim-1.)/(o+1.))+(0.267343098531)*dim
        ref[1,0,1]=(-0.696375094469)*(1+2.*(dim-1.)/(o+1.))+(1.87332506936)*dim
        ref[1,1,0]=(1.42142830039)*(1+2.*(dim-1.)/(o+1.))+(0.193940652101)*dim
        ref[1,1,1]=(-0.0184876904693)*(1+2.*(dim-1.)/(o+1.))+(0.64535275769)*dim
        ref[2,0,0]=(0.788739680562)*(1+2.*(dim-1.)/(o+1.))+(1.80982591161)*dim
        ref[2,0,1]=(-0.307947812947)*(1+2.*(dim-1.)/(o+1.))+(-0.720298676944)*dim
        ref[2,1,0]=(0.444708986667)*(1+2.*(dim-1.)/(o+1.))+(0.319707326457)*dim
        ref[2,1,1]=(-0.76458724096)*(1+2.*(dim-1.)/(o+1.))+(0.827295729366)*dim
        ref[3,0,0]=(0.711895629441)*(1+2.*(dim-1.)/(o+1.))+(0.459037086337)*dim
        ref[3,0,1]=(0.242169182814)*(1+2.*(dim-1.)/(o+1.))+(-0.78708805892)*dim
        ref[3,1,0]=(0.430045035365)*(1+2.*(dim-1.)/(o+1.))+(1.34852331382)*dim
        ref[3,1,1]=(0.019513637036)*(1+2.*(dim-1.)/(o+1.))+(-0.289514827008)*dim
        ref[4,0,0]=(-0.748387318848)*(1+2.*(dim-1.)/(o+1.))+(-0.579145072551)*dim
        ref[4,0,1]=(-0.845115410871)*(1+2.*(dim-1.)/(o+1.))+(-0.521438196723)*dim
        ref[4,1,0]=(-1.12886362485)*(1+2.*(dim-1.)/(o+1.))+(-1.15779029188)*dim
        ref[4,1,1]=(-1.02402320738)*(1+2.*(dim-1.)/(o+1.))+(0.168941753655)*dim
        ref[5,0,0]=(-0.947437613847)*(1+2.*(dim-1.)/(o+1.))+(-0.219693012422)*dim
        ref[5,0,1]=(-0.455387892087)*(1+2.*(dim-1.)/(o+1.))+(1.08652968035)*dim
        ref[5,1,0]=(-1.20926984172)*(1+2.*(dim-1.)/(o+1.))+(-0.0845469742352)*dim
        ref[5,1,1]=(-0.997888376064)*(1+2.*(dim-1.)/(o+1.))+(-0.107565529828)*dim
      else:
        arg[0,0,0]=(-0.760764230601)*x[0]**o+(-0.955215307904)*x[0]+(0.376251560231)*x[1]**o+(0.147849416477)*x[1]+(0.627037741745)*x[2]**o+(-0.93996244415)*x[2]
        arg[0,0,1]=(0.904449671196)*x[0]**o+(-0.297804849063)*x[0]+(0.773059038176)*x[1]**o+(-0.836104534497)*x[1]+(-0.150040994258)*x[2]**o+(0.837988726131)*x[2]
        arg[0,1,0]=(0.354255190058)*x[0]**o+(0.463977999833)*x[0]+(0.218855388406)*x[1]**o+(-0.163086423536)*x[1]+(0.653790071045)*x[2]**o+(-0.457721486689)*x[2]
        arg[0,1,1]=(0.357317871424)*x[0]**o+(-0.752862019242)*x[0]+(0.674187130836)*x[1]**o+(-0.983503125785)*x[1]+(-0.776153051078)*x[2]**o+(0.390700377058)*x[2]
        arg[1,0,0]=(-0.465423805042)*x[0]**o+(0.629855541232)*x[0]+(-0.405300083887)*x[1]**o+(-0.211254598072)*x[1]+(0.502581881406)*x[2]**o+(-0.853779249469)*x[2]
        arg[1,0,1]=(0.716025450344)*x[0]**o+(0.876021396077)*x[0]+(-0.530324190586)*x[1]**o+(0.700207449985)*x[1]+(-0.788565669024)*x[2]**o+(0.963118484198)*x[2]
        arg[1,1,0]=(0.813943574582)*x[0]**o+(-0.938778607938)*x[0]+(-0.451619613547)*x[1]**o+(-0.916990884105)*x[1]+(-0.671730738367)*x[2]**o+(0.109664560689)*x[2]
        arg[1,1,1]=(-0.557598711518)*x[0]**o+(-0.541691706474)*x[0]+(0.500874328683)*x[1]**o+(-0.959653364589)*x[1]+(0.572057415183)*x[2]**o+(-0.842306660222)*x[2]
        arg[2,0,0]=(0.0470537168687)*x[0]**o+(-0.151521222488)*x[0]+(0.168279496456)*x[1]**o+(0.869948894317)*x[1]+(-0.554196665086)*x[2]**o+(0.444189090676)*x[2]
        arg[2,0,1]=(-0.662480394662)*x[0]**o+(0.506778773041)*x[0]+(0.664368833417)*x[1]**o+(-0.76728005759)*x[1]+(0.729668063743)*x[2]**o+(-0.476499535761)*x[2]
        arg[2,1,0]=(0.40405640406)*x[0]**o+(0.537691275434)*x[0]+(0.675845932799)*x[1]**o+(-0.663037107454)*x[1]+(-0.947246204559)*x[2]**o+(-0.210977364198)*x[2]
        arg[2,1,1]=(-0.460999541334)*x[0]**o+(-0.403551870206)*x[0]+(-0.59751233801)*x[1]**o+(0.389018763902)*x[1]+(-0.916883519483)*x[2]**o+(0.712714107738)*x[2]
        arg[3,0,0]=(-0.547139864054)*x[0]**o+(0.91774354334)*x[0]+(-0.953248927323)*x[1]**o+(0.693673610785)*x[1]+(-0.474082903895)*x[2]**o+(0.187174904886)*x[2]
        arg[3,0,1]=(0.538597533339)*x[0]**o+(0.903368885629)*x[0]+(-0.548058780044)*x[1]**o+(-0.380406303455)*x[1]+(-0.389244091021)*x[2]**o+(0.137402766164)*x[2]
        arg[3,1,0]=(-0.11691167078)*x[0]**o+(-0.812294829383)*x[0]+(0.592192015269)*x[1]**o+(-0.3278056345)*x[1]+(-0.60462640896)*x[2]**o+(-0.701541632465)*x[2]
        arg[3,1,1]=(-0.196382907034)*x[0]**o+(-0.496234735923)*x[0]+(-0.646502033751)*x[1]**o+(-0.18981379172)*x[1]+(0.627608199527)*x[2]**o+(-0.468659552408)*x[2]
        arg[4,0,0]=(0.56789650708)*x[0]**o+(0.854984079971)*x[0]+(-0.641473205573)*x[1]**o+(-0.791240666887)*x[1]+(0.453562162212)*x[2]**o+(0.849111318036)*x[2]
        arg[4,0,1]=(0.90836228457)*x[0]**o+(-0.666684673376)*x[0]+(0.245928632693)*x[1]**o+(0.313580017685)*x[1]+(0.353126203728)*x[2]**o+(0.467530918033)*x[2]
        arg[4,1,0]=(0.670926585272)*x[0]**o+(0.768555336283)*x[0]+(-0.170656824353)*x[1]**o+(-0.346425800834)*x[1]+(-0.969641886683)*x[2]**o+(0.775335418066)*x[2]
        arg[4,1,1]=(-0.152500967035)*x[0]**o+(0.575400101502)*x[0]+(0.901331752685)*x[1]**o+(-0.672576642993)*x[1]+(0.803176375377)*x[2]**o+(0.202785337179)*x[2]
        arg[5,0,0]=(-0.798851736802)*x[0]**o+(-0.221276048656)*x[0]+(0.773295190339)*x[1]**o+(0.0903112854292)*x[1]+(0.884961364642)*x[2]**o+(0.326429585537)*x[2]
        arg[5,0,1]=(0.482124059007)*x[0]**o+(-0.39227677557)*x[0]+(-0.23074922677)*x[1]**o+(0.986783507045)*x[1]+(0.591759402442)*x[2]**o+(-0.680278617308)*x[2]
        arg[5,1,0]=(-0.857456972644)*x[0]**o+(-0.200661025977)*x[0]+(-0.850031395341)*x[1]**o+(0.830863037447)*x[1]+(-0.323765625026)*x[2]**o+(-0.969463605839)*x[2]
        arg[5,1,1]=(-0.582538216697)*x[0]**o+(-0.495303456972)*x[0]+(0.0134139524699)*x[1]**o+(-0.750144603651)*x[1]+(-0.632380446311)*x[2]**o+(0.38782755017)*x[2]
        ref[0,0,0]=(0.242525071375)*(1+2.*(dim-1.)/(o+1.))+(-1.74732833558)*dim
        ref[0,0,1]=(1.52746771511)*(1+2.*(dim-1.)/(o+1.))+(-0.295920657428)*dim
        ref[0,1,0]=(1.22690064951)*(1+2.*(dim-1.)/(o+1.))+(-0.156829910392)*dim
        ref[0,1,1]=(0.255351951182)*(1+2.*(dim-1.)/(o+1.))+(-1.34566476797)*dim
        ref[1,0,0]=(-0.368142007523)*(1+2.*(dim-1.)/(o+1.))+(-0.43517830631)*dim
        ref[1,0,1]=(-0.602864409266)*(1+2.*(dim-1.)/(o+1.))+(2.53934733026)*dim
        ref[1,1,0]=(-0.309406777332)*(1+2.*(dim-1.)/(o+1.))+(-1.74610493135)*dim
        ref[1,1,1]=(0.515333032348)*(1+2.*(dim-1.)/(o+1.))+(-2.34365173129)*dim
        ref[2,0,0]=(-0.338863451761)*(1+2.*(dim-1.)/(o+1.))+(1.16261676251)*dim
        ref[2,0,1]=(0.731556502498)*(1+2.*(dim-1.)/(o+1.))+(-0.73700082031)*dim
        ref[2,1,0]=(0.132656132299)*(1+2.*(dim-1.)/(o+1.))+(-0.336323196217)*dim
        ref[2,1,1]=(-1.97539539883)*(1+2.*(dim-1.)/(o+1.))+(0.698181001434)*dim
        ref[3,0,0]=(-1.97447169527)*(1+2.*(dim-1.)/(o+1.))+(1.79859205901)*dim
        ref[3,0,1]=(-0.398705337727)*(1+2.*(dim-1.)/(o+1.))+(0.660365348338)*dim
        ref[3,1,0]=(-0.129346064471)*(1+2.*(dim-1.)/(o+1.))+(-1.84164209635)*dim
        ref[3,1,1]=(-0.215276741258)*(1+2.*(dim-1.)/(o+1.))+(-1.15470808005)*dim
        ref[4,0,0]=(0.379985463719)*(1+2.*(dim-1.)/(o+1.))+(0.91285473112)*dim
        ref[4,0,1]=(1.50741712099)*(1+2.*(dim-1.)/(o+1.))+(0.114426262342)*dim
        ref[4,1,0]=(-0.469372125764)*(1+2.*(dim-1.)/(o+1.))+(1.19746495352)*dim
        ref[4,1,1]=(1.55200716103)*(1+2.*(dim-1.)/(o+1.))+(0.105608795688)*dim
        ref[5,0,0]=(0.859404818179)*(1+2.*(dim-1.)/(o+1.))+(0.19546482231)*dim
        ref[5,0,1]=(0.843134234679)*(1+2.*(dim-1.)/(o+1.))+(-0.085771885833)*dim
        ref[5,1,0]=(-2.03125399301)*(1+2.*(dim-1.)/(o+1.))+(-0.339261594369)*dim
        ref[5,1,1]=(-1.20150471054)*(1+2.*(dim-1.)/(o+1.))+(-0.857620510454)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.161068279134)*x[0]**o+(-0.20016693624)*x[0]+(0.360967926263)*x[1]**o+(0.519774730324)*x[1]
        arg[0,0,0,1]=(-0.470257945181)*x[0]**o+(0.0652334206708)*x[0]+(0.652139991357)*x[1]**o+(-0.592482756578)*x[1]
        arg[0,0,1,0]=(0.143126570256)*x[0]**o+(0.683474249212)*x[0]+(0.839611756722)*x[1]**o+(-0.436612603526)*x[1]
        arg[0,0,1,1]=(-0.82328367976)*x[0]**o+(-0.848145278113)*x[0]+(-0.649046694126)*x[1]**o+(0.7686615034)*x[1]
        arg[0,0,2,0]=(-0.89676527056)*x[0]**o+(-0.466316806084)*x[0]+(-0.454908744166)*x[1]**o+(0.425524454897)*x[1]
        arg[0,0,2,1]=(-0.720978656424)*x[0]**o+(-0.655449460498)*x[0]+(0.466383340682)*x[1]**o+(-0.609615749634)*x[1]
        arg[0,1,0,0]=(-0.358670380794)*x[0]**o+(-0.764939003508)*x[0]+(-0.550898156906)*x[1]**o+(0.803150103728)*x[1]
        arg[0,1,0,1]=(-0.652241076457)*x[0]**o+(-0.614033274928)*x[0]+(-0.308770011731)*x[1]**o+(0.925847136965)*x[1]
        arg[0,1,1,0]=(0.710539026622)*x[0]**o+(-0.345587804078)*x[0]+(-0.170026226995)*x[1]**o+(0.340557574804)*x[1]
        arg[0,1,1,1]=(0.348069554895)*x[0]**o+(-0.633629515001)*x[0]+(-0.199233786202)*x[1]**o+(0.500291488845)*x[1]
        arg[0,1,2,0]=(0.262179751197)*x[0]**o+(-0.55460029497)*x[0]+(-0.354335283745)*x[1]**o+(0.301130103531)*x[1]
        arg[0,1,2,1]=(0.212709964525)*x[0]**o+(0.205907518129)*x[0]+(0.167040570615)*x[1]**o+(-0.0684155270612)*x[1]
        arg[0,2,0,0]=(-0.272141616042)*x[0]**o+(0.187045318075)*x[0]+(-0.364974531949)*x[1]**o+(-0.969702579161)*x[1]
        arg[0,2,0,1]=(-0.997155483171)*x[0]**o+(0.864439328971)*x[0]+(0.887907754657)*x[1]**o+(-0.323021835147)*x[1]
        arg[0,2,1,0]=(-0.544228085563)*x[0]**o+(-0.56752289714)*x[0]+(-0.389234395916)*x[1]**o+(-0.969287386144)*x[1]
        arg[0,2,1,1]=(-0.786136450866)*x[0]**o+(0.589465339989)*x[0]+(-0.411595778776)*x[1]**o+(0.726975706536)*x[1]
        arg[0,2,2,0]=(-0.788299350517)*x[0]**o+(0.697037609691)*x[0]+(-0.391342918209)*x[1]**o+(0.432368662365)*x[1]
        arg[0,2,2,1]=(-0.390536554506)*x[0]**o+(0.547528359725)*x[0]+(0.58957966188)*x[1]**o+(0.146918361066)*x[1]
        arg[0,3,0,0]=(0.176059579799)*x[0]**o+(-0.774041865433)*x[0]+(0.52369869825)*x[1]**o+(-0.812474996262)*x[1]
        arg[0,3,0,1]=(-0.938727595513)*x[0]**o+(0.734277246158)*x[0]+(-0.500452658276)*x[1]**o+(0.591562924701)*x[1]
        arg[0,3,1,0]=(-0.109970490351)*x[0]**o+(-0.463382333915)*x[0]+(0.430588389365)*x[1]**o+(-0.174978181104)*x[1]
        arg[0,3,1,1]=(-0.564486046162)*x[0]**o+(0.325599920584)*x[0]+(-0.177334934858)*x[1]**o+(-0.789210879368)*x[1]
        arg[0,3,2,0]=(-0.409769653157)*x[0]**o+(-0.674668709075)*x[0]+(0.977465383205)*x[1]**o+(-0.178249094926)*x[1]
        arg[0,3,2,1]=(0.112394584736)*x[0]**o+(-0.229767891006)*x[0]+(0.952422132907)*x[1]**o+(-0.133952947363)*x[1]
        arg[0,4,0,0]=(0.927747625444)*x[0]**o+(-0.147390883234)*x[0]+(-0.0523168758823)*x[1]**o+(0.698643952882)*x[1]
        arg[0,4,0,1]=(0.986627759385)*x[0]**o+(-0.0486375625834)*x[0]+(0.27598075052)*x[1]**o+(0.970272004318)*x[1]
        arg[0,4,1,0]=(-0.461643441984)*x[0]**o+(-0.815261492927)*x[0]+(-0.0806771439228)*x[1]**o+(0.877302861263)*x[1]
        arg[0,4,1,1]=(-0.108403920958)*x[0]**o+(0.651566419087)*x[0]+(-0.629972296369)*x[1]**o+(-0.235413582155)*x[1]
        arg[0,4,2,0]=(-0.177527213798)*x[0]**o+(0.987202381273)*x[0]+(-0.15522298966)*x[1]**o+(0.234268561378)*x[1]
        arg[0,4,2,1]=(0.521954293269)*x[0]**o+(-0.776898814469)*x[0]+(0.783085515988)*x[1]**o+(-0.469921197982)*x[1]
        arg[1,0,0,0]=(0.858157489751)*x[0]**o+(0.444181035326)*x[0]+(0.3426806552)*x[1]**o+(-0.219249318925)*x[1]
        arg[1,0,0,1]=(0.405857394564)*x[0]**o+(0.178361294933)*x[0]+(0.0879917927852)*x[1]**o+(-0.0689581267347)*x[1]
        arg[1,0,1,0]=(0.63175565694)*x[0]**o+(0.0810462363705)*x[0]+(0.0350399072162)*x[1]**o+(-0.269906961338)*x[1]
        arg[1,0,1,1]=(-0.273973794035)*x[0]**o+(0.824519533391)*x[0]+(0.596785621044)*x[1]**o+(-0.243043370261)*x[1]
        arg[1,0,2,0]=(-0.959246118913)*x[0]**o+(0.3570024829)*x[0]+(-0.260098162669)*x[1]**o+(0.341044029457)*x[1]
        arg[1,0,2,1]=(0.966671029776)*x[0]**o+(0.598898116395)*x[0]+(0.392360541053)*x[1]**o+(-0.322723646751)*x[1]
        arg[1,1,0,0]=(-0.658365963559)*x[0]**o+(-0.674107595463)*x[0]+(-0.0103911551403)*x[1]**o+(0.858879803218)*x[1]
        arg[1,1,0,1]=(0.340459206102)*x[0]**o+(-0.681434505015)*x[0]+(0.917412924129)*x[1]**o+(0.703699154381)*x[1]
        arg[1,1,1,0]=(0.0473840209756)*x[0]**o+(-0.787070458106)*x[0]+(0.11979244907)*x[1]**o+(0.541238063489)*x[1]
        arg[1,1,1,1]=(0.184943768213)*x[0]**o+(-0.942961758488)*x[0]+(-0.844785429901)*x[1]**o+(-0.0109779269179)*x[1]
        arg[1,1,2,0]=(0.921469748172)*x[0]**o+(-0.122684636542)*x[0]+(0.149552532716)*x[1]**o+(-0.0102623511735)*x[1]
        arg[1,1,2,1]=(-0.720466710917)*x[0]**o+(-0.755841224189)*x[0]+(0.372635950892)*x[1]**o+(0.271604227274)*x[1]
        arg[1,2,0,0]=(-0.548557725319)*x[0]**o+(0.745340396995)*x[0]+(0.461556064961)*x[1]**o+(0.696460891668)*x[1]
        arg[1,2,0,1]=(-0.738939238644)*x[0]**o+(0.344829756443)*x[0]+(0.272341172248)*x[1]**o+(0.768063359753)*x[1]
        arg[1,2,1,0]=(-0.213161389496)*x[0]**o+(0.458730188555)*x[0]+(-0.783049463147)*x[1]**o+(0.785273989074)*x[1]
        arg[1,2,1,1]=(0.189092239774)*x[0]**o+(-0.0649917239139)*x[0]+(0.0171281685468)*x[1]**o+(0.397065812267)*x[1]
        arg[1,2,2,0]=(-0.809663598067)*x[0]**o+(0.587214021528)*x[0]+(0.912814601228)*x[1]**o+(0.860792008172)*x[1]
        arg[1,2,2,1]=(0.487927123118)*x[0]**o+(-0.760831928598)*x[0]+(-0.770486346056)*x[1]**o+(-0.715077841171)*x[1]
        arg[1,3,0,0]=(0.0202747853647)*x[0]**o+(0.0666642857036)*x[0]+(-0.717756666598)*x[1]**o+(-0.481009806467)*x[1]
        arg[1,3,0,1]=(-0.251164550146)*x[0]**o+(0.0519935051437)*x[0]+(-0.703400138413)*x[1]**o+(-0.990329903234)*x[1]
        arg[1,3,1,0]=(0.970793375169)*x[0]**o+(-0.365675269998)*x[0]+(0.77706784549)*x[1]**o+(-0.237785293481)*x[1]
        arg[1,3,1,1]=(-0.162786953006)*x[0]**o+(0.985402077919)*x[0]+(0.739483663725)*x[1]**o+(0.559554321882)*x[1]
        arg[1,3,2,0]=(0.184701463663)*x[0]**o+(0.946173117295)*x[0]+(-0.451420616712)*x[1]**o+(0.244506578089)*x[1]
        arg[1,3,2,1]=(0.894307549357)*x[0]**o+(0.474983970689)*x[0]+(0.626273940678)*x[1]**o+(0.107985157225)*x[1]
        arg[1,4,0,0]=(-0.180994231135)*x[0]**o+(-0.899808624585)*x[0]+(0.395935138342)*x[1]**o+(-0.387788497342)*x[1]
        arg[1,4,0,1]=(0.0713358452056)*x[0]**o+(0.505276043139)*x[0]+(-0.635565006093)*x[1]**o+(0.956951408169)*x[1]
        arg[1,4,1,0]=(0.355608577722)*x[0]**o+(0.58482326687)*x[0]+(0.654137690443)*x[1]**o+(-0.849016362888)*x[1]
        arg[1,4,1,1]=(-0.275966957458)*x[0]**o+(0.856095068377)*x[0]+(-0.760077050664)*x[1]**o+(-0.904159432699)*x[1]
        arg[1,4,2,0]=(-0.0637454805096)*x[0]**o+(-0.833057387633)*x[0]+(0.279008913569)*x[1]**o+(-0.093356790193)*x[1]
        arg[1,4,2,1]=(0.165869666472)*x[0]**o+(-0.4554345089)*x[0]+(0.697149094381)*x[1]**o+(0.244460087482)*x[1]
        arg[2,0,0,0]=(0.388188511348)*x[0]**o+(-0.00552398040882)*x[0]+(0.0679336322536)*x[1]**o+(-0.281132460216)*x[1]
        arg[2,0,0,1]=(0.968019690109)*x[0]**o+(0.525335591339)*x[0]+(0.660069087546)*x[1]**o+(-0.795388487336)*x[1]
        arg[2,0,1,0]=(-0.444010450901)*x[0]**o+(-0.498771184141)*x[0]+(-0.394999176962)*x[1]**o+(-0.579230388911)*x[1]
        arg[2,0,1,1]=(-0.96211072138)*x[0]**o+(-0.0977519108256)*x[0]+(0.206179333903)*x[1]**o+(-0.206155312014)*x[1]
        arg[2,0,2,0]=(0.131307090845)*x[0]**o+(0.179396626286)*x[0]+(-0.436118623166)*x[1]**o+(0.829727875456)*x[1]
        arg[2,0,2,1]=(-0.206336435516)*x[0]**o+(0.175744210813)*x[0]+(0.818204199205)*x[1]**o+(0.433292546177)*x[1]
        arg[2,1,0,0]=(0.920189410059)*x[0]**o+(-0.629286762249)*x[0]+(-0.14444305096)*x[1]**o+(0.564022686795)*x[1]
        arg[2,1,0,1]=(0.43550707444)*x[0]**o+(-0.0203691253164)*x[0]+(0.632472807002)*x[1]**o+(0.722390213081)*x[1]
        arg[2,1,1,0]=(0.827084071535)*x[0]**o+(0.864139638834)*x[0]+(0.795893591186)*x[1]**o+(0.0492667773115)*x[1]
        arg[2,1,1,1]=(0.863813787478)*x[0]**o+(0.268800194137)*x[0]+(-0.567637801205)*x[1]**o+(0.931221493563)*x[1]
        arg[2,1,2,0]=(-0.613463525024)*x[0]**o+(-0.0688030533166)*x[0]+(-0.786395692457)*x[1]**o+(0.862189706883)*x[1]
        arg[2,1,2,1]=(0.315306814851)*x[0]**o+(0.76879509223)*x[0]+(0.451451342779)*x[1]**o+(-0.478873421268)*x[1]
        arg[2,2,0,0]=(-0.785116047841)*x[0]**o+(0.347614087057)*x[0]+(-0.0106716042015)*x[1]**o+(-0.350898768187)*x[1]
        arg[2,2,0,1]=(-0.987778931187)*x[0]**o+(-0.380379654757)*x[0]+(-0.383604999655)*x[1]**o+(0.022203803886)*x[1]
        arg[2,2,1,0]=(0.508327115069)*x[0]**o+(0.443678958632)*x[0]+(-0.320069211602)*x[1]**o+(-0.778140184171)*x[1]
        arg[2,2,1,1]=(-0.532025577107)*x[0]**o+(-0.0475799386434)*x[0]+(0.650417175266)*x[1]**o+(0.33739661369)*x[1]
        arg[2,2,2,0]=(-0.406412097333)*x[0]**o+(0.0510500610605)*x[0]+(-0.25755890225)*x[1]**o+(-0.350209779392)*x[1]
        arg[2,2,2,1]=(0.401361048956)*x[0]**o+(-0.594793585281)*x[0]+(0.155257431052)*x[1]**o+(-0.674885216524)*x[1]
        arg[2,3,0,0]=(0.352606222236)*x[0]**o+(0.64529574473)*x[0]+(-0.838258091233)*x[1]**o+(0.527911574536)*x[1]
        arg[2,3,0,1]=(-0.62652610251)*x[0]**o+(0.879664434218)*x[0]+(0.730845316192)*x[1]**o+(0.0318659384791)*x[1]
        arg[2,3,1,0]=(-0.311152313179)*x[0]**o+(-0.333154399786)*x[0]+(0.743640958069)*x[1]**o+(0.313394944493)*x[1]
        arg[2,3,1,1]=(-0.728968478378)*x[0]**o+(0.500229380574)*x[0]+(0.756628133572)*x[1]**o+(0.270533146994)*x[1]
        arg[2,3,2,0]=(-0.219775843272)*x[0]**o+(0.882943220401)*x[0]+(0.79341547188)*x[1]**o+(0.503486054551)*x[1]
        arg[2,3,2,1]=(-0.850285149353)*x[0]**o+(0.78820884302)*x[0]+(0.348024963962)*x[1]**o+(0.83743616537)*x[1]
        arg[2,4,0,0]=(0.868649421127)*x[0]**o+(-0.622361223999)*x[0]+(0.836321867516)*x[1]**o+(0.423582994131)*x[1]
        arg[2,4,0,1]=(0.140963355402)*x[0]**o+(0.190496469385)*x[0]+(0.765311231744)*x[1]**o+(0.368696267095)*x[1]
        arg[2,4,1,0]=(-0.680854080076)*x[0]**o+(-0.729641193897)*x[0]+(-0.843347646585)*x[1]**o+(-0.0993036180842)*x[1]
        arg[2,4,1,1]=(-0.130039648356)*x[0]**o+(0.534197300899)*x[0]+(-0.142588861186)*x[1]**o+(-0.308376644249)*x[1]
        arg[2,4,2,0]=(0.331212825202)*x[0]**o+(0.466247840288)*x[0]+(-0.201714210846)*x[1]**o+(0.80521345364)*x[1]
        arg[2,4,2,1]=(0.454299813247)*x[0]**o+(0.646694941459)*x[0]+(-0.929854585458)*x[1]**o+(-0.365252296746)*x[1]
        arg[3,0,0,0]=(-0.623915395721)*x[0]**o+(-0.789827882448)*x[0]+(-0.862568457189)*x[1]**o+(-0.141799958679)*x[1]
        arg[3,0,0,1]=(0.197040281353)*x[0]**o+(0.155785418081)*x[0]+(-0.587898788817)*x[1]**o+(0.700612189317)*x[1]
        arg[3,0,1,0]=(-0.899654954269)*x[0]**o+(-0.801426785588)*x[0]+(-0.528678401571)*x[1]**o+(-0.224185097206)*x[1]
        arg[3,0,1,1]=(0.291298188198)*x[0]**o+(0.786698690825)*x[0]+(0.925634940697)*x[1]**o+(0.617661715197)*x[1]
        arg[3,0,2,0]=(0.788272473274)*x[0]**o+(-0.0400044705274)*x[0]+(-0.669092974469)*x[1]**o+(-0.063697728491)*x[1]
        arg[3,0,2,1]=(-0.0612398565521)*x[0]**o+(-0.884311783009)*x[0]+(0.444689078628)*x[1]**o+(0.302741981611)*x[1]
        arg[3,1,0,0]=(0.483756249722)*x[0]**o+(-0.214185139167)*x[0]+(0.526387014144)*x[1]**o+(-0.172619975799)*x[1]
        arg[3,1,0,1]=(0.0685836489777)*x[0]**o+(-0.425861440836)*x[0]+(0.94669899123)*x[1]**o+(-0.649068988929)*x[1]
        arg[3,1,1,0]=(0.935225203148)*x[0]**o+(-0.649378812349)*x[0]+(-0.791553458732)*x[1]**o+(0.776307740809)*x[1]
        arg[3,1,1,1]=(-0.45319945403)*x[0]**o+(-0.473310763902)*x[0]+(-0.268480872039)*x[1]**o+(0.847382076174)*x[1]
        arg[3,1,2,0]=(-0.573022434421)*x[0]**o+(0.584143976397)*x[0]+(0.131688457808)*x[1]**o+(-0.628894863435)*x[1]
        arg[3,1,2,1]=(0.628566928363)*x[0]**o+(-0.868564569338)*x[0]+(-0.482273553287)*x[1]**o+(0.679652796519)*x[1]
        arg[3,2,0,0]=(-0.64046509323)*x[0]**o+(-0.108928537952)*x[0]+(-0.512288515479)*x[1]**o+(0.683072290041)*x[1]
        arg[3,2,0,1]=(0.472589328326)*x[0]**o+(0.219741638671)*x[0]+(0.136368535311)*x[1]**o+(-0.525113003647)*x[1]
        arg[3,2,1,0]=(-0.45731256203)*x[0]**o+(0.0854410475382)*x[0]+(0.257725845542)*x[1]**o+(0.430430185351)*x[1]
        arg[3,2,1,1]=(0.948699678725)*x[0]**o+(-0.00538658505415)*x[0]+(0.921898600708)*x[1]**o+(-0.448522286404)*x[1]
        arg[3,2,2,0]=(-0.0863353766204)*x[0]**o+(-0.652417730681)*x[0]+(0.755871413533)*x[1]**o+(0.206065623655)*x[1]
        arg[3,2,2,1]=(0.263075920527)*x[0]**o+(0.94162407402)*x[0]+(0.750015836722)*x[1]**o+(-0.815154267718)*x[1]
        arg[3,3,0,0]=(-0.606984299829)*x[0]**o+(0.120704522147)*x[0]+(-0.324211323778)*x[1]**o+(-0.123647081263)*x[1]
        arg[3,3,0,1]=(0.838995936232)*x[0]**o+(0.854147779134)*x[0]+(-0.192271552952)*x[1]**o+(0.0281108740635)*x[1]
        arg[3,3,1,0]=(-0.0343159887291)*x[0]**o+(0.242432659911)*x[0]+(-0.37299830191)*x[1]**o+(-0.456792158625)*x[1]
        arg[3,3,1,1]=(-0.305277044867)*x[0]**o+(-0.930386582491)*x[0]+(-0.979921197617)*x[1]**o+(-0.507736755766)*x[1]
        arg[3,3,2,0]=(0.548675744195)*x[0]**o+(0.755596482286)*x[0]+(0.0722267054576)*x[1]**o+(-0.393768047035)*x[1]
        arg[3,3,2,1]=(0.74781993508)*x[0]**o+(-0.655033306209)*x[0]+(0.480653847364)*x[1]**o+(-0.169400714085)*x[1]
        arg[3,4,0,0]=(0.269186769422)*x[0]**o+(-0.0786214595922)*x[0]+(-0.975508552024)*x[1]**o+(-0.798249509899)*x[1]
        arg[3,4,0,1]=(-0.599201275441)*x[0]**o+(0.473490319504)*x[0]+(0.0277044125859)*x[1]**o+(0.818107843614)*x[1]
        arg[3,4,1,0]=(-0.46272642862)*x[0]**o+(0.740282148493)*x[0]+(0.35998379499)*x[1]**o+(-0.938681489717)*x[1]
        arg[3,4,1,1]=(0.187067286301)*x[0]**o+(-0.284171344521)*x[0]+(0.612229918327)*x[1]**o+(0.3917984881)*x[1]
        arg[3,4,2,0]=(-0.535270910341)*x[0]**o+(-0.00723888213927)*x[0]+(-0.776984534618)*x[1]**o+(-0.0899536033273)*x[1]
        arg[3,4,2,1]=(-0.382290376155)*x[0]**o+(-0.0593490177599)*x[0]+(-0.19716515678)*x[1]**o+(-0.268991693613)*x[1]
        ref[0,0,0,0]=(0.19989964713)*(1+2.*(dim-1.)/(o+1.))+(0.319607794084)*dim
        ref[0,0,0,1]=(0.181882046176)*(1+2.*(dim-1.)/(o+1.))+(-0.527249335907)*dim
        ref[0,0,1,0]=(0.982738326978)*(1+2.*(dim-1.)/(o+1.))+(0.246861645686)*dim
        ref[0,0,1,1]=(-1.47233037389)*(1+2.*(dim-1.)/(o+1.))+(-0.0794837747136)*dim
        ref[0,0,2,0]=(-1.35167401473)*(1+2.*(dim-1.)/(o+1.))+(-0.0407923511861)*dim
        ref[0,0,2,1]=(-0.254595315742)*(1+2.*(dim-1.)/(o+1.))+(-1.26506521013)*dim
        ref[0,1,0,0]=(-0.909568537701)*(1+2.*(dim-1.)/(o+1.))+(0.0382111002203)*dim
        ref[0,1,0,1]=(-0.961011088187)*(1+2.*(dim-1.)/(o+1.))+(0.311813862038)*dim
        ref[0,1,1,0]=(0.540512799627)*(1+2.*(dim-1.)/(o+1.))+(-0.00503022927381)*dim
        ref[0,1,1,1]=(0.148835768693)*(1+2.*(dim-1.)/(o+1.))+(-0.133338026157)*dim
        ref[0,1,2,0]=(-0.0921555325482)*(1+2.*(dim-1.)/(o+1.))+(-0.253470191439)*dim
        ref[0,1,2,1]=(0.37975053514)*(1+2.*(dim-1.)/(o+1.))+(0.137491991068)*dim
        ref[0,2,0,0]=(-0.637116147991)*(1+2.*(dim-1.)/(o+1.))+(-0.782657261086)*dim
        ref[0,2,0,1]=(-0.109247728514)*(1+2.*(dim-1.)/(o+1.))+(0.541417493824)*dim
        ref[0,2,1,0]=(-0.933462481479)*(1+2.*(dim-1.)/(o+1.))+(-1.53681028328)*dim
        ref[0,2,1,1]=(-1.19773222964)*(1+2.*(dim-1.)/(o+1.))+(1.31644104653)*dim
        ref[0,2,2,0]=(-1.17964226873)*(1+2.*(dim-1.)/(o+1.))+(1.12940627206)*dim
        ref[0,2,2,1]=(0.199043107375)*(1+2.*(dim-1.)/(o+1.))+(0.694446720792)*dim
        ref[0,3,0,0]=(0.699758278049)*(1+2.*(dim-1.)/(o+1.))+(-1.5865168617)*dim
        ref[0,3,0,1]=(-1.43918025379)*(1+2.*(dim-1.)/(o+1.))+(1.32584017086)*dim
        ref[0,3,1,0]=(0.320617899014)*(1+2.*(dim-1.)/(o+1.))+(-0.63836051502)*dim
        ref[0,3,1,1]=(-0.741820981021)*(1+2.*(dim-1.)/(o+1.))+(-0.463610958784)*dim
        ref[0,3,2,0]=(0.567695730048)*(1+2.*(dim-1.)/(o+1.))+(-0.852917804)*dim
        ref[0,3,2,1]=(1.06481671764)*(1+2.*(dim-1.)/(o+1.))+(-0.363720838369)*dim
        ref[0,4,0,0]=(0.875430749562)*(1+2.*(dim-1.)/(o+1.))+(0.551253069648)*dim
        ref[0,4,0,1]=(1.26260850991)*(1+2.*(dim-1.)/(o+1.))+(0.921634441735)*dim
        ref[0,4,1,0]=(-0.542320585907)*(1+2.*(dim-1.)/(o+1.))+(0.0620413683361)*dim
        ref[0,4,1,1]=(-0.738376217327)*(1+2.*(dim-1.)/(o+1.))+(0.416152836932)*dim
        ref[0,4,2,0]=(-0.332750203458)*(1+2.*(dim-1.)/(o+1.))+(1.22147094265)*dim
        ref[0,4,2,1]=(1.30503980926)*(1+2.*(dim-1.)/(o+1.))+(-1.24682001245)*dim
        ref[1,0,0,0]=(1.20083814495)*(1+2.*(dim-1.)/(o+1.))+(0.224931716401)*dim
        ref[1,0,0,1]=(0.493849187349)*(1+2.*(dim-1.)/(o+1.))+(0.109403168198)*dim
        ref[1,0,1,0]=(0.666795564157)*(1+2.*(dim-1.)/(o+1.))+(-0.188860724968)*dim
        ref[1,0,1,1]=(0.322811827009)*(1+2.*(dim-1.)/(o+1.))+(0.58147616313)*dim
        ref[1,0,2,0]=(-1.21934428158)*(1+2.*(dim-1.)/(o+1.))+(0.698046512357)*dim
        ref[1,0,2,1]=(1.35903157083)*(1+2.*(dim-1.)/(o+1.))+(0.276174469644)*dim
        ref[1,1,0,0]=(-0.668757118699)*(1+2.*(dim-1.)/(o+1.))+(0.184772207756)*dim
        ref[1,1,0,1]=(1.25787213023)*(1+2.*(dim-1.)/(o+1.))+(0.0222646493659)*dim
        ref[1,1,1,0]=(0.167176470046)*(1+2.*(dim-1.)/(o+1.))+(-0.245832394617)*dim
        ref[1,1,1,1]=(-0.659841661688)*(1+2.*(dim-1.)/(o+1.))+(-0.953939685406)*dim
        ref[1,1,2,0]=(1.07102228089)*(1+2.*(dim-1.)/(o+1.))+(-0.132946987715)*dim
        ref[1,1,2,1]=(-0.347830760025)*(1+2.*(dim-1.)/(o+1.))+(-0.484236996915)*dim
        ref[1,2,0,0]=(-0.087001660358)*(1+2.*(dim-1.)/(o+1.))+(1.44180128866)*dim
        ref[1,2,0,1]=(-0.466598066397)*(1+2.*(dim-1.)/(o+1.))+(1.1128931162)*dim
        ref[1,2,1,0]=(-0.996210852643)*(1+2.*(dim-1.)/(o+1.))+(1.24400417763)*dim
        ref[1,2,1,1]=(0.206220408321)*(1+2.*(dim-1.)/(o+1.))+(0.332074088353)*dim
        ref[1,2,2,0]=(0.103151003161)*(1+2.*(dim-1.)/(o+1.))+(1.4480060297)*dim
        ref[1,2,2,1]=(-0.282559222939)*(1+2.*(dim-1.)/(o+1.))+(-1.47590976977)*dim
        ref[1,3,0,0]=(-0.697481881234)*(1+2.*(dim-1.)/(o+1.))+(-0.414345520764)*dim
        ref[1,3,0,1]=(-0.954564688559)*(1+2.*(dim-1.)/(o+1.))+(-0.93833639809)*dim
        ref[1,3,1,0]=(1.74786122066)*(1+2.*(dim-1.)/(o+1.))+(-0.603460563479)*dim
        ref[1,3,1,1]=(0.576696710719)*(1+2.*(dim-1.)/(o+1.))+(1.5449563998)*dim
        ref[1,3,2,0]=(-0.26671915305)*(1+2.*(dim-1.)/(o+1.))+(1.19067969538)*dim
        ref[1,3,2,1]=(1.52058149003)*(1+2.*(dim-1.)/(o+1.))+(0.582969127914)*dim
        ref[1,4,0,0]=(0.214940907208)*(1+2.*(dim-1.)/(o+1.))+(-1.28759712193)*dim
        ref[1,4,0,1]=(-0.564229160887)*(1+2.*(dim-1.)/(o+1.))+(1.46222745131)*dim
        ref[1,4,1,0]=(1.00974626816)*(1+2.*(dim-1.)/(o+1.))+(-0.264193096019)*dim
        ref[1,4,1,1]=(-1.03604400812)*(1+2.*(dim-1.)/(o+1.))+(-0.048064364322)*dim
        ref[1,4,2,0]=(0.215263433059)*(1+2.*(dim-1.)/(o+1.))+(-0.926414177826)*dim
        ref[1,4,2,1]=(0.863018760853)*(1+2.*(dim-1.)/(o+1.))+(-0.210974421418)*dim
        ref[2,0,0,0]=(0.456122143601)*(1+2.*(dim-1.)/(o+1.))+(-0.286656440625)*dim
        ref[2,0,0,1]=(1.62808877765)*(1+2.*(dim-1.)/(o+1.))+(-0.270052895997)*dim
        ref[2,0,1,0]=(-0.839009627862)*(1+2.*(dim-1.)/(o+1.))+(-1.07800157305)*dim
        ref[2,0,1,1]=(-0.755931387477)*(1+2.*(dim-1.)/(o+1.))+(-0.30390722284)*dim
        ref[2,0,2,0]=(-0.304811532322)*(1+2.*(dim-1.)/(o+1.))+(1.00912450174)*dim
        ref[2,0,2,1]=(0.61186776369)*(1+2.*(dim-1.)/(o+1.))+(0.60903675699)*dim
        ref[2,1,0,0]=(0.775746359099)*(1+2.*(dim-1.)/(o+1.))+(-0.0652640754538)*dim
        ref[2,1,0,1]=(1.06797988144)*(1+2.*(dim-1.)/(o+1.))+(0.702021087765)*dim
        ref[2,1,1,0]=(1.62297766272)*(1+2.*(dim-1.)/(o+1.))+(0.913406416145)*dim
        ref[2,1,1,1]=(0.296175986273)*(1+2.*(dim-1.)/(o+1.))+(1.2000216877)*dim
        ref[2,1,2,0]=(-1.39985921748)*(1+2.*(dim-1.)/(o+1.))+(0.793386653567)*dim
        ref[2,1,2,1]=(0.76675815763)*(1+2.*(dim-1.)/(o+1.))+(0.289921670962)*dim
        ref[2,2,0,0]=(-0.795787652043)*(1+2.*(dim-1.)/(o+1.))+(-0.0032846811301)*dim
        ref[2,2,0,1]=(-1.37138393084)*(1+2.*(dim-1.)/(o+1.))+(-0.358175850871)*dim
        ref[2,2,1,0]=(0.188257903467)*(1+2.*(dim-1.)/(o+1.))+(-0.334461225539)*dim
        ref[2,2,1,1]=(0.118391598158)*(1+2.*(dim-1.)/(o+1.))+(0.289816675047)*dim
        ref[2,2,2,0]=(-0.663970999582)*(1+2.*(dim-1.)/(o+1.))+(-0.299159718332)*dim
        ref[2,2,2,1]=(0.556618480008)*(1+2.*(dim-1.)/(o+1.))+(-1.2696788018)*dim
        ref[2,3,0,0]=(-0.485651868997)*(1+2.*(dim-1.)/(o+1.))+(1.17320731927)*dim
        ref[2,3,0,1]=(0.104319213682)*(1+2.*(dim-1.)/(o+1.))+(0.911530372697)*dim
        ref[2,3,1,0]=(0.43248864489)*(1+2.*(dim-1.)/(o+1.))+(-0.0197594552935)*dim
        ref[2,3,1,1]=(0.0276596551936)*(1+2.*(dim-1.)/(o+1.))+(0.770762527568)*dim
        ref[2,3,2,0]=(0.573639628608)*(1+2.*(dim-1.)/(o+1.))+(1.38642927495)*dim
        ref[2,3,2,1]=(-0.502260185391)*(1+2.*(dim-1.)/(o+1.))+(1.62564500839)*dim
        ref[2,4,0,0]=(1.70497128864)*(1+2.*(dim-1.)/(o+1.))+(-0.198778229868)*dim
        ref[2,4,0,1]=(0.906274587146)*(1+2.*(dim-1.)/(o+1.))+(0.55919273648)*dim
        ref[2,4,1,0]=(-1.52420172666)*(1+2.*(dim-1.)/(o+1.))+(-0.828944811982)*dim
        ref[2,4,1,1]=(-0.272628509541)*(1+2.*(dim-1.)/(o+1.))+(0.225820656649)*dim
        ref[2,4,2,0]=(0.129498614356)*(1+2.*(dim-1.)/(o+1.))+(1.27146129393)*dim
        ref[2,4,2,1]=(-0.475554772211)*(1+2.*(dim-1.)/(o+1.))+(0.281442644713)*dim
        ref[3,0,0,0]=(-1.48648385291)*(1+2.*(dim-1.)/(o+1.))+(-0.931627841127)*dim
        ref[3,0,0,1]=(-0.390858507465)*(1+2.*(dim-1.)/(o+1.))+(0.856397607398)*dim
        ref[3,0,1,0]=(-1.42833335584)*(1+2.*(dim-1.)/(o+1.))+(-1.02561188279)*dim
        ref[3,0,1,1]=(1.2169331289)*(1+2.*(dim-1.)/(o+1.))+(1.40436040602)*dim
        ref[3,0,2,0]=(0.119179498805)*(1+2.*(dim-1.)/(o+1.))+(-0.103702199018)*dim
        ref[3,0,2,1]=(0.383449222076)*(1+2.*(dim-1.)/(o+1.))+(-0.581569801398)*dim
        ref[3,1,0,0]=(1.01014326387)*(1+2.*(dim-1.)/(o+1.))+(-0.386805114966)*dim
        ref[3,1,0,1]=(1.01528264021)*(1+2.*(dim-1.)/(o+1.))+(-1.07493042977)*dim
        ref[3,1,1,0]=(0.143671744416)*(1+2.*(dim-1.)/(o+1.))+(0.12692892846)*dim
        ref[3,1,1,1]=(-0.721680326068)*(1+2.*(dim-1.)/(o+1.))+(0.374071312272)*dim
        ref[3,1,2,0]=(-0.441333976613)*(1+2.*(dim-1.)/(o+1.))+(-0.0447508870373)*dim
        ref[3,1,2,1]=(0.146293375075)*(1+2.*(dim-1.)/(o+1.))+(-0.188911772819)*dim
        ref[3,2,0,0]=(-1.15275360871)*(1+2.*(dim-1.)/(o+1.))+(0.57414375209)*dim
        ref[3,2,0,1]=(0.608957863637)*(1+2.*(dim-1.)/(o+1.))+(-0.305371364976)*dim
        ref[3,2,1,0]=(-0.199586716488)*(1+2.*(dim-1.)/(o+1.))+(0.515871232889)*dim
        ref[3,2,1,1]=(1.87059827943)*(1+2.*(dim-1.)/(o+1.))+(-0.453908871458)*dim
        ref[3,2,2,0]=(0.669536036913)*(1+2.*(dim-1.)/(o+1.))+(-0.446352107026)*dim
        ref[3,2,2,1]=(1.01309175725)*(1+2.*(dim-1.)/(o+1.))+(0.126469806303)*dim
        ref[3,3,0,0]=(-0.931195623607)*(1+2.*(dim-1.)/(o+1.))+(-0.00294255911617)*dim
        ref[3,3,0,1]=(0.64672438328)*(1+2.*(dim-1.)/(o+1.))+(0.882258653197)*dim
        ref[3,3,1,0]=(-0.407314290639)*(1+2.*(dim-1.)/(o+1.))+(-0.214359498715)*dim
        ref[3,3,1,1]=(-1.28519824248)*(1+2.*(dim-1.)/(o+1.))+(-1.43812333826)*dim
        ref[3,3,2,0]=(0.620902449652)*(1+2.*(dim-1.)/(o+1.))+(0.361828435252)*dim
        ref[3,3,2,1]=(1.22847378244)*(1+2.*(dim-1.)/(o+1.))+(-0.824434020294)*dim
        ref[3,4,0,0]=(-0.706321782603)*(1+2.*(dim-1.)/(o+1.))+(-0.876870969492)*dim
        ref[3,4,0,1]=(-0.571496862855)*(1+2.*(dim-1.)/(o+1.))+(1.29159816312)*dim
        ref[3,4,1,0]=(-0.102742633631)*(1+2.*(dim-1.)/(o+1.))+(-0.198399341224)*dim
        ref[3,4,1,1]=(0.799297204627)*(1+2.*(dim-1.)/(o+1.))+(0.107627143578)*dim
        ref[3,4,2,0]=(-1.31225544496)*(1+2.*(dim-1.)/(o+1.))+(-0.0971924854666)*dim
        ref[3,4,2,1]=(-0.579455532935)*(1+2.*(dim-1.)/(o+1.))+(-0.328340711373)*dim
      else:
        arg[0,0,0,0]=(-0.727570048979)*x[0]**o+(-0.769770353069)*x[0]+(-0.532381054852)*x[1]**o+(0.590589994369)*x[1]+(0.775348726284)*x[2]**o+(-0.840239939796)*x[2]
        arg[0,0,0,1]=(-0.441856577751)*x[0]**o+(-0.39234308235)*x[0]+(-0.834378647798)*x[1]**o+(0.84269918161)*x[1]+(0.441102591728)*x[2]**o+(0.166331058665)*x[2]
        arg[0,0,1,0]=(0.391773069521)*x[0]**o+(0.496759920942)*x[0]+(0.555386226902)*x[1]**o+(0.980197783628)*x[1]+(0.549622021652)*x[2]**o+(-0.596179893504)*x[2]
        arg[0,0,1,1]=(-0.104197387892)*x[0]**o+(-0.64636505375)*x[0]+(0.377612339028)*x[1]**o+(-0.215111619231)*x[1]+(0.637249692625)*x[2]**o+(-0.0998227822351)*x[2]
        arg[0,0,2,0]=(0.780693043561)*x[0]**o+(0.571405344342)*x[0]+(-0.793680154472)*x[1]**o+(-0.0481469136436)*x[1]+(0.959561051293)*x[2]**o+(0.0756152497582)*x[2]
        arg[0,0,2,1]=(0.88678762282)*x[0]**o+(0.254753272214)*x[0]+(0.344376798516)*x[1]**o+(-0.908135902735)*x[1]+(-0.3584387726)*x[2]**o+(-0.890225037617)*x[2]
        arg[0,1,0,0]=(0.680262917473)*x[0]**o+(0.65860357305)*x[0]+(0.859832322473)*x[1]**o+(0.326105580592)*x[1]+(0.132170297876)*x[2]**o+(-0.928606849775)*x[2]
        arg[0,1,0,1]=(-0.500503822293)*x[0]**o+(-0.456331012174)*x[0]+(0.900478846187)*x[1]**o+(0.350542503874)*x[1]+(0.755849038617)*x[2]**o+(-0.514240007142)*x[2]
        arg[0,1,1,0]=(0.124404217291)*x[0]**o+(-0.499651000805)*x[0]+(-0.968178694708)*x[1]**o+(0.306780500818)*x[1]+(-0.946470794819)*x[2]**o+(0.919424575523)*x[2]
        arg[0,1,1,1]=(0.722520017474)*x[0]**o+(-0.718216593718)*x[0]+(0.801196051455)*x[1]**o+(-0.791951278546)*x[1]+(-0.675975417529)*x[2]**o+(0.355069676047)*x[2]
        arg[0,1,2,0]=(-0.137017428857)*x[0]**o+(-0.760267879487)*x[0]+(0.625869259413)*x[1]**o+(-0.697626713352)*x[1]+(-0.467876751423)*x[2]**o+(0.456475454123)*x[2]
        arg[0,1,2,1]=(0.646536442797)*x[0]**o+(0.0981671744166)*x[0]+(0.477147756414)*x[1]**o+(-0.23211823966)*x[1]+(-0.553808167147)*x[2]**o+(0.405402676302)*x[2]
        arg[0,2,0,0]=(-0.0604502966371)*x[0]**o+(0.297995946813)*x[0]+(0.837045433858)*x[1]**o+(-0.270340963575)*x[1]+(-0.58462190484)*x[2]**o+(0.292953993287)*x[2]
        arg[0,2,0,1]=(-0.463109088268)*x[0]**o+(0.346260157368)*x[0]+(0.0228615151352)*x[1]**o+(0.905507471559)*x[1]+(-0.423777553665)*x[2]**o+(-0.45964535654)*x[2]
        arg[0,2,1,0]=(-0.50761294542)*x[0]**o+(-0.465260578419)*x[0]+(-0.729953382373)*x[1]**o+(0.903666219263)*x[1]+(-0.756794344711)*x[2]**o+(0.817263008431)*x[2]
        arg[0,2,1,1]=(-0.551253781393)*x[0]**o+(-0.911481117802)*x[0]+(0.180349330394)*x[1]**o+(0.419118491074)*x[1]+(-0.691917875703)*x[2]**o+(-0.835365504347)*x[2]
        arg[0,2,2,0]=(0.32567932803)*x[0]**o+(0.413167839883)*x[0]+(0.781320691089)*x[1]**o+(-0.777223093057)*x[1]+(0.297556831029)*x[2]**o+(-0.952676550814)*x[2]
        arg[0,2,2,1]=(0.226250474873)*x[0]**o+(0.279287242249)*x[0]+(0.421824124143)*x[1]**o+(-0.97664927412)*x[1]+(0.703245736838)*x[2]**o+(-0.629660622476)*x[2]
        arg[0,3,0,0]=(-0.138403445102)*x[0]**o+(0.0979152923187)*x[0]+(-0.177116218507)*x[1]**o+(0.0100523961708)*x[1]+(-0.634159492232)*x[2]**o+(-0.847618220907)*x[2]
        arg[0,3,0,1]=(0.612268738957)*x[0]**o+(-0.462682392204)*x[0]+(-0.146623835506)*x[1]**o+(0.699258469487)*x[1]+(0.0139983038691)*x[2]**o+(0.236461011719)*x[2]
        arg[0,3,1,0]=(-0.0518123102809)*x[0]**o+(0.386822059696)*x[0]+(-0.208858744655)*x[1]**o+(0.971393362011)*x[1]+(-0.171649562077)*x[2]**o+(-0.507479926411)*x[2]
        arg[0,3,1,1]=(0.0140396270391)*x[0]**o+(-0.837711757872)*x[0]+(-0.00108924459289)*x[1]**o+(-0.724087080496)*x[1]+(-0.793071741694)*x[2]**o+(0.38654936043)*x[2]
        arg[0,3,2,0]=(-0.897711437057)*x[0]**o+(0.562533782571)*x[0]+(-0.354013039442)*x[1]**o+(-0.473034213198)*x[1]+(0.4598333464)*x[2]**o+(0.467503992754)*x[2]
        arg[0,3,2,1]=(-0.198441232877)*x[0]**o+(-0.411568849672)*x[0]+(0.0850211273328)*x[1]**o+(-0.333702582633)*x[1]+(0.727888461746)*x[2]**o+(-0.174835541819)*x[2]
        arg[0,4,0,0]=(-0.114679585752)*x[0]**o+(0.0446596462171)*x[0]+(-0.950604674991)*x[1]**o+(-0.307622186566)*x[1]+(0.956712754366)*x[2]**o+(0.334630480874)*x[2]
        arg[0,4,0,1]=(0.345475911118)*x[0]**o+(0.834620576123)*x[0]+(0.903176273684)*x[1]**o+(-0.364211856466)*x[1]+(0.910903954488)*x[2]**o+(0.392248545744)*x[2]
        arg[0,4,1,0]=(-0.593360797686)*x[0]**o+(-0.0199892857836)*x[0]+(0.293839091284)*x[1]**o+(0.580208854742)*x[1]+(0.060499817751)*x[2]**o+(0.907498841261)*x[2]
        arg[0,4,1,1]=(-0.446192713548)*x[0]**o+(0.29870251614)*x[0]+(0.754231657585)*x[1]**o+(0.321710920549)*x[1]+(0.287978396665)*x[2]**o+(-0.963473165492)*x[2]
        arg[0,4,2,0]=(0.780174386478)*x[0]**o+(0.603377914483)*x[0]+(-0.841703120145)*x[1]**o+(0.316522795462)*x[1]+(-0.192916583378)*x[2]**o+(-0.734091913995)*x[2]
        arg[0,4,2,1]=(-0.130567656486)*x[0]**o+(-0.342928842892)*x[0]+(-0.0827376128775)*x[1]**o+(-0.771104057041)*x[1]+(0.733571283734)*x[2]**o+(-0.880531838119)*x[2]
        arg[1,0,0,0]=(-0.634428121043)*x[0]**o+(0.294135783753)*x[0]+(-0.566305561339)*x[1]**o+(0.0332700472211)*x[1]+(0.128130119463)*x[2]**o+(0.0468678470468)*x[2]
        arg[1,0,0,1]=(0.82168858107)*x[0]**o+(-0.75544177649)*x[0]+(0.260115562827)*x[1]**o+(-0.149575433725)*x[1]+(-0.780945910032)*x[2]**o+(-0.0539291857071)*x[2]
        arg[1,0,1,0]=(-0.97885291605)*x[0]**o+(0.777955971883)*x[0]+(0.115355866293)*x[1]**o+(-0.456449594589)*x[1]+(-0.832816352387)*x[2]**o+(-0.418070257158)*x[2]
        arg[1,0,1,1]=(-0.908396506245)*x[0]**o+(-0.982147653872)*x[0]+(0.138657052347)*x[1]**o+(0.633837624386)*x[1]+(-0.529748303776)*x[2]**o+(-0.950225664343)*x[2]
        arg[1,0,2,0]=(0.733974435377)*x[0]**o+(-0.758228777635)*x[0]+(0.841368499977)*x[1]**o+(0.871715889444)*x[1]+(0.541055759071)*x[2]**o+(0.669798500731)*x[2]
        arg[1,0,2,1]=(-0.572994548563)*x[0]**o+(0.806241853906)*x[0]+(0.141217952929)*x[1]**o+(0.832291361477)*x[1]+(0.207737014762)*x[2]**o+(-0.279769695917)*x[2]
        arg[1,1,0,0]=(-0.394247339923)*x[0]**o+(0.548338039327)*x[0]+(-0.845822483858)*x[1]**o+(-0.941161341738)*x[1]+(-0.244513027249)*x[2]**o+(-0.524790843131)*x[2]
        arg[1,1,0,1]=(0.509994957918)*x[0]**o+(0.691470630092)*x[0]+(0.249241759795)*x[1]**o+(-0.995293157716)*x[1]+(0.162652088127)*x[2]**o+(0.873578806104)*x[2]
        arg[1,1,1,0]=(-0.22367137056)*x[0]**o+(-0.896306406631)*x[0]+(0.613214186844)*x[1]**o+(-0.460055562163)*x[1]+(-0.704736634406)*x[2]**o+(-0.517741014186)*x[2]
        arg[1,1,1,1]=(-0.426072795689)*x[0]**o+(-0.13452184053)*x[0]+(0.908778671123)*x[1]**o+(0.194472828672)*x[1]+(-0.658909495204)*x[2]**o+(-0.234260758459)*x[2]
        arg[1,1,2,0]=(0.236086562474)*x[0]**o+(0.33312461561)*x[0]+(0.867089643801)*x[1]**o+(-0.720157305369)*x[1]+(0.812845198077)*x[2]**o+(-0.375694713863)*x[2]
        arg[1,1,2,1]=(-0.561145817489)*x[0]**o+(0.746630137163)*x[0]+(-0.809280840008)*x[1]**o+(-0.690912076356)*x[1]+(0.88933641231)*x[2]**o+(0.246647817564)*x[2]
        arg[1,2,0,0]=(0.701319152729)*x[0]**o+(-0.309074024603)*x[0]+(0.567988628717)*x[1]**o+(-0.0596939135728)*x[1]+(0.242135294918)*x[2]**o+(-0.101136005445)*x[2]
        arg[1,2,0,1]=(0.880665674235)*x[0]**o+(-0.833066777852)*x[0]+(-0.0556340786728)*x[1]**o+(0.620095583249)*x[1]+(-0.390556829051)*x[2]**o+(0.793481461864)*x[2]
        arg[1,2,1,0]=(0.832046852431)*x[0]**o+(0.4276990324)*x[0]+(-0.486292331356)*x[1]**o+(0.990296730852)*x[1]+(-0.381732943927)*x[2]**o+(0.225838720655)*x[2]
        arg[1,2,1,1]=(-0.971477034333)*x[0]**o+(0.819409035213)*x[0]+(-0.610341797903)*x[1]**o+(0.842186877805)*x[1]+(0.571187503044)*x[2]**o+(-0.150565006374)*x[2]
        arg[1,2,2,0]=(0.266080624706)*x[0]**o+(0.893377400845)*x[0]+(0.718525426441)*x[1]**o+(-0.847615622286)*x[1]+(0.614914866)*x[2]**o+(0.826503774974)*x[2]
        arg[1,2,2,1]=(0.998060871997)*x[0]**o+(-0.572523112154)*x[0]+(-0.786785614822)*x[1]**o+(0.556952019953)*x[1]+(0.804849394796)*x[2]**o+(-0.201157281118)*x[2]
        arg[1,3,0,0]=(-0.322048105751)*x[0]**o+(0.215441634353)*x[0]+(0.203298400258)*x[1]**o+(0.452071048161)*x[1]+(0.0134315902731)*x[2]**o+(0.240098266028)*x[2]
        arg[1,3,0,1]=(-0.109845136841)*x[0]**o+(0.0479663457074)*x[0]+(0.899394604931)*x[1]**o+(0.853608214151)*x[1]+(-0.27487885579)*x[2]**o+(-0.0236702866567)*x[2]
        arg[1,3,1,0]=(0.948550474293)*x[0]**o+(0.359077475995)*x[0]+(0.655387467125)*x[1]**o+(-0.0501114084308)*x[1]+(-0.244470769181)*x[2]**o+(0.670276065978)*x[2]
        arg[1,3,1,1]=(-0.860769147822)*x[0]**o+(-0.738449680804)*x[0]+(0.840959189643)*x[1]**o+(0.229930283198)*x[1]+(-0.521034843253)*x[2]**o+(-0.50213233896)*x[2]
        arg[1,3,2,0]=(0.29910216992)*x[0]**o+(0.411953562132)*x[0]+(0.777584507308)*x[1]**o+(0.41482951697)*x[1]+(-0.948751119313)*x[2]**o+(-0.526573791564)*x[2]
        arg[1,3,2,1]=(-0.768167653244)*x[0]**o+(-0.318762199437)*x[0]+(0.7566706966)*x[1]**o+(-0.151347269999)*x[1]+(-0.626690637485)*x[2]**o+(-0.731518248609)*x[2]
        arg[1,4,0,0]=(0.0147733854143)*x[0]**o+(-0.812988077571)*x[0]+(0.815385483312)*x[1]**o+(0.285326283687)*x[1]+(0.872502797762)*x[2]**o+(-0.748392032843)*x[2]
        arg[1,4,0,1]=(0.172931930795)*x[0]**o+(0.425088340505)*x[0]+(0.0122969950852)*x[1]**o+(0.556865378995)*x[1]+(0.614686173604)*x[2]**o+(0.966581266651)*x[2]
        arg[1,4,1,0]=(0.725270959871)*x[0]**o+(-0.672683754253)*x[0]+(0.486696038678)*x[1]**o+(-0.53856418144)*x[1]+(-0.295197744155)*x[2]**o+(0.0610876962829)*x[2]
        arg[1,4,1,1]=(0.942300770794)*x[0]**o+(0.876440453364)*x[0]+(0.432652207093)*x[1]**o+(-0.562629961162)*x[1]+(0.497526473424)*x[2]**o+(-0.165824324383)*x[2]
        arg[1,4,2,0]=(-0.668927175078)*x[0]**o+(0.587283304457)*x[0]+(0.588305038304)*x[1]**o+(0.087906366831)*x[1]+(-0.907721198172)*x[2]**o+(-0.506067369924)*x[2]
        arg[1,4,2,1]=(0.899071894553)*x[0]**o+(0.920678485704)*x[0]+(-0.244107214117)*x[1]**o+(0.196327093214)*x[1]+(-0.538075261013)*x[2]**o+(-0.925646646321)*x[2]
        arg[2,0,0,0]=(-0.113452114076)*x[0]**o+(0.541178389983)*x[0]+(0.501083424074)*x[1]**o+(-0.234654083639)*x[1]+(-0.435569149263)*x[2]**o+(0.671517947867)*x[2]
        arg[2,0,0,1]=(-0.00139653535956)*x[0]**o+(-0.93731587459)*x[0]+(-0.387786776826)*x[1]**o+(0.361445822788)*x[1]+(-0.0958490804452)*x[2]**o+(0.22475163005)*x[2]
        arg[2,0,1,0]=(0.100832420019)*x[0]**o+(-0.957577198313)*x[0]+(0.244479124631)*x[1]**o+(-0.589295650233)*x[1]+(-0.11020792463)*x[2]**o+(0.969023424875)*x[2]
        arg[2,0,1,1]=(-0.82757398282)*x[0]**o+(0.665274113482)*x[0]+(0.219624146627)*x[1]**o+(0.604687589384)*x[1]+(0.539774513376)*x[2]**o+(0.676136495778)*x[2]
        arg[2,0,2,0]=(-0.869539421193)*x[0]**o+(0.449645390105)*x[0]+(-0.216356671763)*x[1]**o+(-0.191616800811)*x[1]+(-0.324126591392)*x[2]**o+(0.726527544226)*x[2]
        arg[2,0,2,1]=(-0.522340761977)*x[0]**o+(-0.253136332372)*x[0]+(-0.993693358789)*x[1]**o+(0.689338979208)*x[1]+(-0.73030598892)*x[2]**o+(0.604868622965)*x[2]
        arg[2,1,0,0]=(-0.87285436276)*x[0]**o+(-0.665917385941)*x[0]+(0.105570880359)*x[1]**o+(0.773780432278)*x[1]+(-0.0932677950401)*x[2]**o+(0.55563257373)*x[2]
        arg[2,1,0,1]=(-0.717823895492)*x[0]**o+(-0.843992380543)*x[0]+(0.348730031558)*x[1]**o+(-0.0641727895105)*x[1]+(-0.575332519937)*x[2]**o+(0.566731137573)*x[2]
        arg[2,1,1,0]=(-0.749216309273)*x[0]**o+(0.945232679736)*x[0]+(-0.984694983452)*x[1]**o+(-0.697422850579)*x[1]+(0.573407778839)*x[2]**o+(0.654548962131)*x[2]
        arg[2,1,1,1]=(-0.951637762959)*x[0]**o+(-0.0299724748286)*x[0]+(-0.784527116834)*x[1]**o+(0.490138786409)*x[1]+(0.945743565011)*x[2]**o+(0.158024494836)*x[2]
        arg[2,1,2,0]=(0.80439224967)*x[0]**o+(0.889509225531)*x[0]+(-0.178335319647)*x[1]**o+(-0.365669987413)*x[1]+(-0.170055755127)*x[2]**o+(0.258843711555)*x[2]
        arg[2,1,2,1]=(0.520706901588)*x[0]**o+(0.897178241272)*x[0]+(-0.128495163664)*x[1]**o+(-0.404742110163)*x[1]+(0.140333496135)*x[2]**o+(0.31429703244)*x[2]
        arg[2,2,0,0]=(-0.933201442136)*x[0]**o+(0.0083650170184)*x[0]+(-0.281807642508)*x[1]**o+(-0.09193002721)*x[1]+(-0.422582781227)*x[2]**o+(0.358915935551)*x[2]
        arg[2,2,0,1]=(-0.09712053324)*x[0]**o+(-0.0752448353447)*x[0]+(0.41291697894)*x[1]**o+(0.593823615119)*x[1]+(0.954183095989)*x[2]**o+(0.479576723309)*x[2]
        arg[2,2,1,0]=(-0.48592273499)*x[0]**o+(-0.555807193072)*x[0]+(0.842926147107)*x[1]**o+(-0.0760474608316)*x[1]+(0.864590424311)*x[2]**o+(-0.854721691042)*x[2]
        arg[2,2,1,1]=(0.238261310086)*x[0]**o+(-0.506114025767)*x[0]+(-0.534154656267)*x[1]**o+(-0.282262066272)*x[1]+(-0.241950678835)*x[2]**o+(-0.596913007205)*x[2]
        arg[2,2,2,0]=(0.35002425052)*x[0]**o+(-0.724235915518)*x[0]+(-0.443311896206)*x[1]**o+(0.0183169190559)*x[1]+(-0.180995873436)*x[2]**o+(0.949105760518)*x[2]
        arg[2,2,2,1]=(-0.524596480672)*x[0]**o+(-0.72511087285)*x[0]+(-0.793159588655)*x[1]**o+(0.197579071177)*x[1]+(0.214510478855)*x[2]**o+(0.592011219443)*x[2]
        arg[2,3,0,0]=(-0.777144207151)*x[0]**o+(0.821815068413)*x[0]+(0.828183715461)*x[1]**o+(-0.77742744879)*x[1]+(0.520088528639)*x[2]**o+(0.564796500907)*x[2]
        arg[2,3,0,1]=(-0.444252706506)*x[0]**o+(-0.999658246414)*x[0]+(0.458490846376)*x[1]**o+(0.206371738627)*x[1]+(-0.5943830165)*x[2]**o+(0.125954475804)*x[2]
        arg[2,3,1,0]=(-0.499555379078)*x[0]**o+(0.667767779032)*x[0]+(-0.169053599596)*x[1]**o+(0.587608309525)*x[1]+(-0.99317016838)*x[2]**o+(-0.580232194372)*x[2]
        arg[2,3,1,1]=(0.881736170108)*x[0]**o+(-0.0221128885544)*x[0]+(-0.239202356219)*x[1]**o+(-0.825250840397)*x[1]+(-0.34138233902)*x[2]**o+(-0.900437209752)*x[2]
        arg[2,3,2,0]=(-0.810415345283)*x[0]**o+(0.671639285402)*x[0]+(0.0577696593549)*x[1]**o+(0.12466039802)*x[1]+(-0.164952472389)*x[2]**o+(-0.881356759483)*x[2]
        arg[2,3,2,1]=(0.641411687742)*x[0]**o+(0.939465110042)*x[0]+(0.182541049996)*x[1]**o+(-0.502132165421)*x[1]+(-0.914212670088)*x[2]**o+(0.361356243622)*x[2]
        arg[2,4,0,0]=(0.696033328872)*x[0]**o+(0.854329989895)*x[0]+(0.66092830442)*x[1]**o+(-0.165623528752)*x[1]+(0.587322879532)*x[2]**o+(-0.441145527759)*x[2]
        arg[2,4,0,1]=(0.885589185979)*x[0]**o+(0.170856006811)*x[0]+(-0.500114387063)*x[1]**o+(-0.772574748562)*x[1]+(0.599300325869)*x[2]**o+(0.0950082175327)*x[2]
        arg[2,4,1,0]=(0.877739402691)*x[0]**o+(0.494715865857)*x[0]+(-0.4548916845)*x[1]**o+(-0.108381552398)*x[1]+(-0.620522177203)*x[2]**o+(-0.375546939131)*x[2]
        arg[2,4,1,1]=(-0.108044365878)*x[0]**o+(-0.592058059304)*x[0]+(-0.64481990676)*x[1]**o+(-0.903674017194)*x[1]+(0.0964052945913)*x[2]**o+(0.357819426568)*x[2]
        arg[2,4,2,0]=(-0.686157836559)*x[0]**o+(0.585514351597)*x[0]+(-0.5623443108)*x[1]**o+(-0.636059380122)*x[1]+(0.469863235081)*x[2]**o+(0.374567303622)*x[2]
        arg[2,4,2,1]=(0.214800008136)*x[0]**o+(0.0680448664819)*x[0]+(0.722930302776)*x[1]**o+(0.107573165145)*x[1]+(0.0027804595356)*x[2]**o+(0.725566605544)*x[2]
        arg[3,0,0,0]=(-0.273256791086)*x[0]**o+(0.801129507153)*x[0]+(-0.840227644476)*x[1]**o+(-0.31594258736)*x[1]+(-0.140806177065)*x[2]**o+(-0.0958103115672)*x[2]
        arg[3,0,0,1]=(0.588534583503)*x[0]**o+(0.242235627811)*x[0]+(-0.541539488722)*x[1]**o+(0.624385574868)*x[1]+(-0.0641371754789)*x[2]**o+(-0.635676751281)*x[2]
        arg[3,0,1,0]=(0.81967100218)*x[0]**o+(0.671597165436)*x[0]+(0.160179001916)*x[1]**o+(-0.226714136475)*x[1]+(-0.523709112534)*x[2]**o+(0.932797007296)*x[2]
        arg[3,0,1,1]=(0.707793115256)*x[0]**o+(-0.0995173271712)*x[0]+(-0.594597614398)*x[1]**o+(-0.72096550461)*x[1]+(-0.373631371949)*x[2]**o+(-0.282151921803)*x[2]
        arg[3,0,2,0]=(0.873307297971)*x[0]**o+(0.0950091191819)*x[0]+(-0.534108539131)*x[1]**o+(0.567381193719)*x[1]+(-0.276691140405)*x[2]**o+(0.84882104254)*x[2]
        arg[3,0,2,1]=(-0.895148878411)*x[0]**o+(-0.880973451462)*x[0]+(0.105602819158)*x[1]**o+(-0.836475687487)*x[1]+(-0.18380164164)*x[2]**o+(-0.391643939001)*x[2]
        arg[3,1,0,0]=(-0.144070349397)*x[0]**o+(-0.247578520564)*x[0]+(0.512469248899)*x[1]**o+(0.0369808259097)*x[1]+(0.67517925106)*x[2]**o+(-0.306122403522)*x[2]
        arg[3,1,0,1]=(0.389023591419)*x[0]**o+(0.0335324297663)*x[0]+(0.166380432563)*x[1]**o+(0.415133512101)*x[1]+(-0.956055398028)*x[2]**o+(-0.069254560597)*x[2]
        arg[3,1,1,0]=(-0.253194755795)*x[0]**o+(-0.432240637655)*x[0]+(0.886730830871)*x[1]**o+(0.94685399231)*x[1]+(-0.0193581886643)*x[2]**o+(-0.828124732153)*x[2]
        arg[3,1,1,1]=(0.948739050057)*x[0]**o+(-0.200402018913)*x[0]+(0.062573402994)*x[1]**o+(-0.408727698463)*x[1]+(-0.0334411087309)*x[2]**o+(-0.676147119957)*x[2]
        arg[3,1,2,0]=(0.249237167322)*x[0]**o+(-0.621387879739)*x[0]+(-0.745335388751)*x[1]**o+(-0.183173004074)*x[1]+(0.56632553768)*x[2]**o+(-0.519909176525)*x[2]
        arg[3,1,2,1]=(-0.660822495815)*x[0]**o+(0.0459275869317)*x[0]+(-0.0853936912232)*x[1]**o+(-0.895502076207)*x[1]+(0.150013238845)*x[2]**o+(0.201527905115)*x[2]
        arg[3,2,0,0]=(0.0846880899978)*x[0]**o+(-0.450721252185)*x[0]+(0.0100224498309)*x[1]**o+(-0.729399043167)*x[1]+(0.249073157161)*x[2]**o+(-0.313056982897)*x[2]
        arg[3,2,0,1]=(0.349123663394)*x[0]**o+(-0.0439266646004)*x[0]+(0.201749457716)*x[1]**o+(-0.29961557135)*x[1]+(0.247657591065)*x[2]**o+(0.702448030656)*x[2]
        arg[3,2,1,0]=(-0.67551758529)*x[0]**o+(-0.856271486186)*x[0]+(-0.337698270178)*x[1]**o+(-0.855526501932)*x[1]+(0.204628923051)*x[2]**o+(0.0425229547861)*x[2]
        arg[3,2,1,1]=(0.913895958603)*x[0]**o+(0.33488814727)*x[0]+(0.725768429772)*x[1]**o+(0.31487179035)*x[1]+(-0.926895768372)*x[2]**o+(-0.303623128744)*x[2]
        arg[3,2,2,0]=(0.695497510732)*x[0]**o+(-0.71819800795)*x[0]+(0.0949892877746)*x[1]**o+(-0.371952684199)*x[1]+(-0.712679288901)*x[2]**o+(-0.839032166941)*x[2]
        arg[3,2,2,1]=(0.847915632309)*x[0]**o+(0.489320767225)*x[0]+(0.884694888545)*x[1]**o+(-0.836734066966)*x[1]+(-0.782643394677)*x[2]**o+(0.998012000062)*x[2]
        arg[3,3,0,0]=(0.400343531732)*x[0]**o+(-0.977399566427)*x[0]+(-0.0491815761683)*x[1]**o+(0.820343030844)*x[1]+(0.145276965371)*x[2]**o+(0.106181410329)*x[2]
        arg[3,3,0,1]=(0.39509119571)*x[0]**o+(-0.532152199068)*x[0]+(0.0961163031465)*x[1]**o+(-0.845734786516)*x[1]+(0.0685207262665)*x[2]**o+(0.10725771579)*x[2]
        arg[3,3,1,0]=(-0.326475146957)*x[0]**o+(0.0491193078417)*x[0]+(-0.193405294452)*x[1]**o+(0.283579426128)*x[1]+(0.885442239373)*x[2]**o+(-0.838544460442)*x[2]
        arg[3,3,1,1]=(-0.502234549601)*x[0]**o+(-0.764617462736)*x[0]+(0.52701849705)*x[1]**o+(-0.423245302614)*x[1]+(0.728096320995)*x[2]**o+(0.143356081954)*x[2]
        arg[3,3,2,0]=(0.924562662058)*x[0]**o+(-0.0120799887424)*x[0]+(-0.342886608544)*x[1]**o+(0.609435784717)*x[1]+(0.392296911228)*x[2]**o+(0.572866548262)*x[2]
        arg[3,3,2,1]=(0.661267105524)*x[0]**o+(-0.71847134372)*x[0]+(-0.913866381825)*x[1]**o+(-0.502556458508)*x[1]+(0.201214639904)*x[2]**o+(-0.431939125167)*x[2]
        arg[3,4,0,0]=(0.391785002874)*x[0]**o+(0.424885787931)*x[0]+(-0.377558346944)*x[1]**o+(-0.556924666311)*x[1]+(0.511852194353)*x[2]**o+(0.514261734074)*x[2]
        arg[3,4,0,1]=(-0.269148156354)*x[0]**o+(0.692033978804)*x[0]+(-0.390604590454)*x[1]**o+(-0.551681907742)*x[1]+(0.0613720420346)*x[2]**o+(-0.508841982991)*x[2]
        arg[3,4,1,0]=(0.593170387694)*x[0]**o+(-0.00676980821753)*x[0]+(0.788564472913)*x[1]**o+(0.823094059865)*x[1]+(0.52895485971)*x[2]**o+(0.88849618397)*x[2]
        arg[3,4,1,1]=(0.772185283881)*x[0]**o+(0.759014894154)*x[0]+(-0.142739578724)*x[1]**o+(-0.141899255898)*x[1]+(0.479086662114)*x[2]**o+(0.68993908918)*x[2]
        arg[3,4,2,0]=(0.998097253401)*x[0]**o+(0.621254339576)*x[0]+(-0.692540865624)*x[1]**o+(0.716545091382)*x[1]+(0.478209703594)*x[2]**o+(0.259063088226)*x[2]
        arg[3,4,2,1]=(-0.321962318028)*x[0]**o+(0.0377531939042)*x[0]+(-0.362041758315)*x[1]**o+(0.632474956936)*x[1]+(0.133972776903)*x[2]**o+(-0.341803258375)*x[2]
        ref[0,0,0,0]=(-0.484602377547)*(1+2.*(dim-1.)/(o+1.))+(-1.0194202985)*dim
        ref[0,0,0,1]=(-0.835132633821)*(1+2.*(dim-1.)/(o+1.))+(0.616687157924)*dim
        ref[0,0,1,0]=(1.49678131808)*(1+2.*(dim-1.)/(o+1.))+(0.880777811066)*dim
        ref[0,0,1,1]=(0.910664643761)*(1+2.*(dim-1.)/(o+1.))+(-0.961299455217)*dim
        ref[0,0,2,0]=(0.946573940383)*(1+2.*(dim-1.)/(o+1.))+(0.598873680457)*dim
        ref[0,0,2,1]=(0.872725648736)*(1+2.*(dim-1.)/(o+1.))+(-1.54360766814)*dim
        ref[0,1,0,0]=(1.67226553782)*(1+2.*(dim-1.)/(o+1.))+(0.0561023038674)*dim
        ref[0,1,0,1]=(1.15582406251)*(1+2.*(dim-1.)/(o+1.))+(-0.620028515442)*dim
        ref[0,1,1,0]=(-1.79024527224)*(1+2.*(dim-1.)/(o+1.))+(0.726554075535)*dim
        ref[0,1,1,1]=(0.8477406514)*(1+2.*(dim-1.)/(o+1.))+(-1.15509819622)*dim
        ref[0,1,2,0]=(0.0209750791326)*(1+2.*(dim-1.)/(o+1.))+(-1.00141913872)*dim
        ref[0,1,2,1]=(0.569876032064)*(1+2.*(dim-1.)/(o+1.))+(0.271451611059)*dim
        ref[0,2,0,0]=(0.19197323238)*(1+2.*(dim-1.)/(o+1.))+(0.320608976525)*dim
        ref[0,2,0,1]=(-0.864025126798)*(1+2.*(dim-1.)/(o+1.))+(0.792122272387)*dim
        ref[0,2,1,0]=(-1.9943606725)*(1+2.*(dim-1.)/(o+1.))+(1.25566864927)*dim
        ref[0,2,1,1]=(-1.0628223267)*(1+2.*(dim-1.)/(o+1.))+(-1.32772813107)*dim
        ref[0,2,2,0]=(1.40455685015)*(1+2.*(dim-1.)/(o+1.))+(-1.31673180399)*dim
        ref[0,2,2,1]=(1.35132033585)*(1+2.*(dim-1.)/(o+1.))+(-1.32702265435)*dim
        ref[0,3,0,0]=(-0.949679155841)*(1+2.*(dim-1.)/(o+1.))+(-0.739650532418)*dim
        ref[0,3,0,1]=(0.47964320732)*(1+2.*(dim-1.)/(o+1.))+(0.473037089003)*dim
        ref[0,3,1,0]=(-0.432320617013)*(1+2.*(dim-1.)/(o+1.))+(0.850735495297)*dim
        ref[0,3,1,1]=(-0.780121359247)*(1+2.*(dim-1.)/(o+1.))+(-1.17524947794)*dim
        ref[0,3,2,0]=(-0.791891130098)*(1+2.*(dim-1.)/(o+1.))+(0.557003562128)*dim
        ref[0,3,2,1]=(0.614468356202)*(1+2.*(dim-1.)/(o+1.))+(-0.920106974125)*dim
        ref[0,4,0,0]=(-0.108571506377)*(1+2.*(dim-1.)/(o+1.))+(0.0716679405256)*dim
        ref[0,4,0,1]=(2.15955613929)*(1+2.*(dim-1.)/(o+1.))+(0.862657265402)*dim
        ref[0,4,1,0]=(-0.239021888652)*(1+2.*(dim-1.)/(o+1.))+(1.46771841022)*dim
        ref[0,4,1,1]=(0.596017340702)*(1+2.*(dim-1.)/(o+1.))+(-0.343059728803)*dim
        ref[0,4,2,0]=(-0.254445317045)*(1+2.*(dim-1.)/(o+1.))+(0.185808795951)*dim
        ref[0,4,2,1]=(0.520266014371)*(1+2.*(dim-1.)/(o+1.))+(-1.99456473805)*dim
        ref[1,0,0,0]=(-1.07260356292)*(1+2.*(dim-1.)/(o+1.))+(0.374273678021)*dim
        ref[1,0,0,1]=(0.300858233864)*(1+2.*(dim-1.)/(o+1.))+(-0.958946395922)*dim
        ref[1,0,1,0]=(-1.69631340214)*(1+2.*(dim-1.)/(o+1.))+(-0.096563879864)*dim
        ref[1,0,1,1]=(-1.29948775767)*(1+2.*(dim-1.)/(o+1.))+(-1.29853569383)*dim
        ref[1,0,2,0]=(2.11639869443)*(1+2.*(dim-1.)/(o+1.))+(0.783285612541)*dim
        ref[1,0,2,1]=(-0.224039580872)*(1+2.*(dim-1.)/(o+1.))+(1.35876351947)*dim
        ref[1,1,0,0]=(-1.48458285103)*(1+2.*(dim-1.)/(o+1.))+(-0.917614145542)*dim
        ref[1,1,0,1]=(0.92188880584)*(1+2.*(dim-1.)/(o+1.))+(0.569756278479)*dim
        ref[1,1,1,0]=(-0.315193818122)*(1+2.*(dim-1.)/(o+1.))+(-1.87410298298)*dim
        ref[1,1,1,1]=(-0.17620361977)*(1+2.*(dim-1.)/(o+1.))+(-0.174309770317)*dim
        ref[1,1,2,0]=(1.91602140435)*(1+2.*(dim-1.)/(o+1.))+(-0.762727403622)*dim
        ref[1,1,2,1]=(-0.481090245186)*(1+2.*(dim-1.)/(o+1.))+(0.302365878371)*dim
        ref[1,2,0,0]=(1.51144307636)*(1+2.*(dim-1.)/(o+1.))+(-0.469903943621)*dim
        ref[1,2,0,1]=(0.434474766511)*(1+2.*(dim-1.)/(o+1.))+(0.580510267262)*dim
        ref[1,2,1,0]=(-0.0359784228521)*(1+2.*(dim-1.)/(o+1.))+(1.64383448391)*dim
        ref[1,2,1,1]=(-1.01063132919)*(1+2.*(dim-1.)/(o+1.))+(1.51103090664)*dim
        ref[1,2,2,0]=(1.59952091715)*(1+2.*(dim-1.)/(o+1.))+(0.872265553532)*dim
        ref[1,2,2,1]=(1.01612465197)*(1+2.*(dim-1.)/(o+1.))+(-0.216728373319)*dim
        ref[1,3,0,0]=(-0.10531811522)*(1+2.*(dim-1.)/(o+1.))+(0.907610948542)*dim
        ref[1,3,0,1]=(0.5146706123)*(1+2.*(dim-1.)/(o+1.))+(0.877904273202)*dim
        ref[1,3,1,0]=(1.35946717224)*(1+2.*(dim-1.)/(o+1.))+(0.979242133542)*dim
        ref[1,3,1,1]=(-0.540844801432)*(1+2.*(dim-1.)/(o+1.))+(-1.01065173657)*dim
        ref[1,3,2,0]=(0.127935557915)*(1+2.*(dim-1.)/(o+1.))+(0.300209287538)*dim
        ref[1,3,2,1]=(-0.63818759413)*(1+2.*(dim-1.)/(o+1.))+(-1.20162771805)*dim
        ref[1,4,0,0]=(1.70266166649)*(1+2.*(dim-1.)/(o+1.))+(-1.27605382673)*dim
        ref[1,4,0,1]=(0.799915099484)*(1+2.*(dim-1.)/(o+1.))+(1.94853498615)*dim
        ref[1,4,1,0]=(0.916769254394)*(1+2.*(dim-1.)/(o+1.))+(-1.15016023941)*dim
        ref[1,4,1,1]=(1.87247945131)*(1+2.*(dim-1.)/(o+1.))+(0.147986167818)*dim
        ref[1,4,2,0]=(-0.988343334946)*(1+2.*(dim-1.)/(o+1.))+(0.169122301365)*dim
        ref[1,4,2,1]=(0.116889419423)*(1+2.*(dim-1.)/(o+1.))+(0.191358932597)*dim
        ref[2,0,0,0]=(-0.0479378392653)*(1+2.*(dim-1.)/(o+1.))+(0.978042254211)*dim
        ref[2,0,0,1]=(-0.485032392631)*(1+2.*(dim-1.)/(o+1.))+(-0.351118421752)*dim
        ref[2,0,1,0]=(0.23510362002)*(1+2.*(dim-1.)/(o+1.))+(-0.577849423671)*dim
        ref[2,0,1,1]=(-0.068175322817)*(1+2.*(dim-1.)/(o+1.))+(1.94609819864)*dim
        ref[2,0,2,0]=(-1.41002268435)*(1+2.*(dim-1.)/(o+1.))+(0.98455613352)*dim
        ref[2,0,2,1]=(-2.24634010969)*(1+2.*(dim-1.)/(o+1.))+(1.0410712698)*dim
        ref[2,1,0,0]=(-0.860551277441)*(1+2.*(dim-1.)/(o+1.))+(0.663495620068)*dim
        ref[2,1,0,1]=(-0.944426383871)*(1+2.*(dim-1.)/(o+1.))+(-0.341434032481)*dim
        ref[2,1,1,0]=(-1.16050351389)*(1+2.*(dim-1.)/(o+1.))+(0.902358791288)*dim
        ref[2,1,1,1]=(-0.790421314781)*(1+2.*(dim-1.)/(o+1.))+(0.618190806416)*dim
        ref[2,1,2,0]=(0.456001174897)*(1+2.*(dim-1.)/(o+1.))+(0.782682949672)*dim
        ref[2,1,2,1]=(0.532545234059)*(1+2.*(dim-1.)/(o+1.))+(0.806733163549)*dim
        ref[2,2,0,0]=(-1.63759186587)*(1+2.*(dim-1.)/(o+1.))+(0.275350925359)*dim
        ref[2,2,0,1]=(1.26997954169)*(1+2.*(dim-1.)/(o+1.))+(0.998155503083)*dim
        ref[2,2,1,0]=(1.22159383643)*(1+2.*(dim-1.)/(o+1.))+(-1.48657634495)*dim
        ref[2,2,1,1]=(-0.537844025016)*(1+2.*(dim-1.)/(o+1.))+(-1.38528909924)*dim
        ref[2,2,2,0]=(-0.274283519122)*(1+2.*(dim-1.)/(o+1.))+(0.243186764056)*dim
        ref[2,2,2,1]=(-1.10324559047)*(1+2.*(dim-1.)/(o+1.))+(0.0644794177698)*dim
        ref[2,3,0,0]=(0.571128036949)*(1+2.*(dim-1.)/(o+1.))+(0.60918412053)*dim
        ref[2,3,0,1]=(-0.580144876629)*(1+2.*(dim-1.)/(o+1.))+(-0.667332031982)*dim
        ref[2,3,1,0]=(-1.66177914705)*(1+2.*(dim-1.)/(o+1.))+(0.675143894185)*dim
        ref[2,3,1,1]=(0.301151474869)*(1+2.*(dim-1.)/(o+1.))+(-1.7478009387)*dim
        ref[2,3,2,0]=(-0.917598158317)*(1+2.*(dim-1.)/(o+1.))+(-0.0850570760611)*dim
        ref[2,3,2,1]=(-0.09025993235)*(1+2.*(dim-1.)/(o+1.))+(0.798689188243)*dim
        ref[2,4,0,0]=(1.94428451282)*(1+2.*(dim-1.)/(o+1.))+(0.247560933384)*dim
        ref[2,4,0,1]=(0.984775124785)*(1+2.*(dim-1.)/(o+1.))+(-0.506710524219)*dim
        ref[2,4,1,0]=(-0.197674459012)*(1+2.*(dim-1.)/(o+1.))+(0.0107873743289)*dim
        ref[2,4,1,1]=(-0.656458978046)*(1+2.*(dim-1.)/(o+1.))+(-1.13791264993)*dim
        ref[2,4,2,0]=(-0.778638912279)*(1+2.*(dim-1.)/(o+1.))+(0.324022275097)*dim
        ref[2,4,2,1]=(0.940510770447)*(1+2.*(dim-1.)/(o+1.))+(0.901184637171)*dim
        ref[3,0,0,0]=(-1.25429061263)*(1+2.*(dim-1.)/(o+1.))+(0.389376608226)*dim
        ref[3,0,0,1]=(-0.0171420806973)*(1+2.*(dim-1.)/(o+1.))+(0.230944451397)*dim
        ref[3,0,1,0]=(0.456140891562)*(1+2.*(dim-1.)/(o+1.))+(1.37768003626)*dim
        ref[3,0,1,1]=(-0.260435871091)*(1+2.*(dim-1.)/(o+1.))+(-1.10263475358)*dim
        ref[3,0,2,0]=(0.0625076184349)*(1+2.*(dim-1.)/(o+1.))+(1.51121135544)*dim
        ref[3,0,2,1]=(-0.973347700893)*(1+2.*(dim-1.)/(o+1.))+(-2.10909307795)*dim
        ref[3,1,0,0]=(1.04357815056)*(1+2.*(dim-1.)/(o+1.))+(-0.516720098176)*dim
        ref[3,1,0,1]=(-0.400651374046)*(1+2.*(dim-1.)/(o+1.))+(0.37941138127)*dim
        ref[3,1,1,0]=(0.614177886412)*(1+2.*(dim-1.)/(o+1.))+(-0.313511377498)*dim
        ref[3,1,1,1]=(0.97787134432)*(1+2.*(dim-1.)/(o+1.))+(-1.28527683733)*dim
        ref[3,1,2,0]=(0.0702273162506)*(1+2.*(dim-1.)/(o+1.))+(-1.32447006034)*dim
        ref[3,1,2,1]=(-0.596202948193)*(1+2.*(dim-1.)/(o+1.))+(-0.648046584161)*dim
        ref[3,2,0,0]=(0.34378369699)*(1+2.*(dim-1.)/(o+1.))+(-1.49317727825)*dim
        ref[3,2,0,1]=(0.798530712175)*(1+2.*(dim-1.)/(o+1.))+(0.358905794706)*dim
        ref[3,2,1,0]=(-0.808586932417)*(1+2.*(dim-1.)/(o+1.))+(-1.66927503333)*dim
        ref[3,2,1,1]=(0.712768620002)*(1+2.*(dim-1.)/(o+1.))+(0.346136808876)*dim
        ref[3,2,2,0]=(0.0778075096053)*(1+2.*(dim-1.)/(o+1.))+(-1.92918285909)*dim
        ref[3,2,2,1]=(0.949967126178)*(1+2.*(dim-1.)/(o+1.))+(0.650598700321)*dim
        ref[3,3,0,0]=(0.496438920934)*(1+2.*(dim-1.)/(o+1.))+(-0.0508751252546)*dim
        ref[3,3,0,1]=(0.559728225123)*(1+2.*(dim-1.)/(o+1.))+(-1.27062926979)*dim
        ref[3,3,1,0]=(0.365561797964)*(1+2.*(dim-1.)/(o+1.))+(-0.505845726473)*dim
        ref[3,3,1,1]=(0.752880268444)*(1+2.*(dim-1.)/(o+1.))+(-1.0445066834)*dim
        ref[3,3,2,0]=(0.973972964742)*(1+2.*(dim-1.)/(o+1.))+(1.17022234424)*dim
        ref[3,3,2,1]=(-0.0513846363967)*(1+2.*(dim-1.)/(o+1.))+(-1.65296692739)*dim
        ref[3,4,0,0]=(0.526078850284)*(1+2.*(dim-1.)/(o+1.))+(0.382222855695)*dim
        ref[3,4,0,1]=(-0.598380704773)*(1+2.*(dim-1.)/(o+1.))+(-0.368489911929)*dim
        ref[3,4,1,0]=(1.91068972032)*(1+2.*(dim-1.)/(o+1.))+(1.70482043562)*dim
        ref[3,4,1,1]=(1.10853236727)*(1+2.*(dim-1.)/(o+1.))+(1.30705472744)*dim
        ref[3,4,2,0]=(0.78376609137)*(1+2.*(dim-1.)/(o+1.))+(1.59686251918)*dim
        ref[3,4,2,1]=(-0.550031299439)*(1+2.*(dim-1.)/(o+1.))+(0.328424892465)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.260967575068)*x[0]**o+(-0.456559015376)*x[0]+(0.663023090785)*x[1]**o+(-0.000383723402374)*x[1]
        ref[0]=o*(0.260967575068)*x_ref[0]**(o-1)+(-0.456559015376)
        ref[1]=o*(0.663023090785)*x_ref[1]**(o-1)+(-0.000383723402374)
      else:
        arg=(-0.679453642968)*x[0]**o+(-0.836558889087)*x[0]+(0.938200435255)*x[1]**o+(-0.368159735907)*x[1]+(-0.0546850112561)*x[2]**o+(0.931986983035)*x[2]
        ref[0]=o*(-0.679453642968)*x_ref[0]**(o-1)+(-0.836558889087)
        ref[1]=o*(0.938200435255)*x_ref[1]**(o-1)+(-0.368159735907)
        ref[2]=o*(-0.0546850112561)*x_ref[2]**(o-1)+(0.931986983035)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.221278607378)*x[0]**o+(-0.25051444588)*x[0]+(0.752161895582)*x[1]**o+(0.662434618125)*x[1]
        arg[1]=(0.984829669888)*x[0]**o+(0.00500258390598)*x[0]+(-0.292523259369)*x[1]**o+(-0.238277884466)*x[1]
        ref[0,0]=o*(0.221278607378)*x_ref[0]**(o-1)+(-0.25051444588)
        ref[0,1]=o*(0.752161895582)*x_ref[1]**(o-1)+(0.662434618125)
        ref[1,0]=o*(0.984829669888)*x_ref[0]**(o-1)+(0.00500258390598)
        ref[1,1]=o*(-0.292523259369)*x_ref[1]**(o-1)+(-0.238277884466)
      else:
        arg[0]=(0.815886350601)*x[0]**o+(-0.41721938077)*x[0]+(0.478358181839)*x[1]**o+(-0.458860964588)*x[1]+(0.575706872631)*x[2]**o+(-0.965426161177)*x[2]
        arg[1]=(-0.159203003073)*x[0]**o+(-0.985356409774)*x[0]+(0.504125935966)*x[1]**o+(0.703505637664)*x[1]+(-0.480016288837)*x[2]**o+(0.544400209204)*x[2]
        ref[0,0]=o*(0.815886350601)*x_ref[0]**(o-1)+(-0.41721938077)
        ref[0,1]=o*(0.478358181839)*x_ref[1]**(o-1)+(-0.458860964588)
        ref[0,2]=o*(0.575706872631)*x_ref[2]**(o-1)+(-0.965426161177)
        ref[1,0]=o*(-0.159203003073)*x_ref[0]**(o-1)+(-0.985356409774)
        ref[1,1]=o*(0.504125935966)*x_ref[1]**(o-1)+(0.703505637664)
        ref[1,2]=o*(-0.480016288837)*x_ref[2]**(o-1)+(0.544400209204)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(-0.722004066753)*x[0]**o+(-0.911439550186)*x[0]+(0.0793361579222)*x[1]**o+(0.958526722957)*x[1]
        arg[0,1]=(-0.0237328054252)*x[0]**o+(0.832621528642)*x[0]+(0.168484408262)*x[1]**o+(-0.181293214021)*x[1]
        arg[0,2]=(0.0284992392733)*x[0]**o+(0.884411090784)*x[0]+(-0.615596355135)*x[1]**o+(-0.564819619851)*x[1]
        arg[0,3]=(0.37408647134)*x[0]**o+(-0.177901551674)*x[0]+(-0.466165048356)*x[1]**o+(-0.278511961588)*x[1]
        arg[0,4]=(-0.24695435531)*x[0]**o+(-0.624082916034)*x[0]+(-0.343849543401)*x[1]**o+(0.661610450445)*x[1]
        arg[1,0]=(-0.176271330569)*x[0]**o+(-0.661612282712)*x[0]+(-0.937163166613)*x[1]**o+(0.376473179587)*x[1]
        arg[1,1]=(0.964111839461)*x[0]**o+(0.576837069403)*x[0]+(-0.718562476684)*x[1]**o+(0.154883495525)*x[1]
        arg[1,2]=(-0.771556809734)*x[0]**o+(0.720054023732)*x[0]+(-0.390732462938)*x[1]**o+(-0.484359019049)*x[1]
        arg[1,3]=(0.565910084656)*x[0]**o+(0.69105192657)*x[0]+(-0.449408955892)*x[1]**o+(-0.789780518242)*x[1]
        arg[1,4]=(0.602485034177)*x[0]**o+(-0.345139282323)*x[0]+(-0.951334152921)*x[1]**o+(-0.478700081579)*x[1]
        arg[2,0]=(0.00114729937901)*x[0]**o+(0.295356090995)*x[0]+(0.982261544035)*x[1]**o+(0.357076141518)*x[1]
        arg[2,1]=(0.653029330145)*x[0]**o+(-0.0178804068723)*x[0]+(-0.0238303862422)*x[1]**o+(0.983033975535)*x[1]
        arg[2,2]=(0.680878878444)*x[0]**o+(-0.404330541218)*x[0]+(0.856544883866)*x[1]**o+(0.71899071821)*x[1]
        arg[2,3]=(0.0674687202072)*x[0]**o+(-0.748990717637)*x[0]+(0.185597242309)*x[1]**o+(0.623849127358)*x[1]
        arg[2,4]=(0.302869699915)*x[0]**o+(-0.852621552847)*x[0]+(0.115557112172)*x[1]**o+(-0.0827214208715)*x[1]
        arg[3,0]=(0.0929072570643)*x[0]**o+(0.533152382923)*x[0]+(-0.0913754984708)*x[1]**o+(-0.25865969337)*x[1]
        arg[3,1]=(-0.794799886368)*x[0]**o+(-0.865263962776)*x[0]+(0.441343356932)*x[1]**o+(-0.541975483415)*x[1]
        arg[3,2]=(0.261341948686)*x[0]**o+(0.378142729621)*x[0]+(0.0556054089673)*x[1]**o+(-0.854195169318)*x[1]
        arg[3,3]=(0.177222467454)*x[0]**o+(-0.400820197499)*x[0]+(0.88513075927)*x[1]**o+(0.0420082325044)*x[1]
        arg[3,4]=(0.307178422145)*x[0]**o+(-0.203765349758)*x[0]+(-0.546052990937)*x[1]**o+(-0.960312136292)*x[1]
        ref[0,0,0]=o*(-0.722004066753)*x_ref[0]**(o-1)+(-0.911439550186)
        ref[0,0,1]=o*(0.0793361579222)*x_ref[1]**(o-1)+(0.958526722957)
        ref[0,1,0]=o*(-0.0237328054252)*x_ref[0]**(o-1)+(0.832621528642)
        ref[0,1,1]=o*(0.168484408262)*x_ref[1]**(o-1)+(-0.181293214021)
        ref[0,2,0]=o*(0.0284992392733)*x_ref[0]**(o-1)+(0.884411090784)
        ref[0,2,1]=o*(-0.615596355135)*x_ref[1]**(o-1)+(-0.564819619851)
        ref[0,3,0]=o*(0.37408647134)*x_ref[0]**(o-1)+(-0.177901551674)
        ref[0,3,1]=o*(-0.466165048356)*x_ref[1]**(o-1)+(-0.278511961588)
        ref[0,4,0]=o*(-0.24695435531)*x_ref[0]**(o-1)+(-0.624082916034)
        ref[0,4,1]=o*(-0.343849543401)*x_ref[1]**(o-1)+(0.661610450445)
        ref[1,0,0]=o*(-0.176271330569)*x_ref[0]**(o-1)+(-0.661612282712)
        ref[1,0,1]=o*(-0.937163166613)*x_ref[1]**(o-1)+(0.376473179587)
        ref[1,1,0]=o*(0.964111839461)*x_ref[0]**(o-1)+(0.576837069403)
        ref[1,1,1]=o*(-0.718562476684)*x_ref[1]**(o-1)+(0.154883495525)
        ref[1,2,0]=o*(-0.771556809734)*x_ref[0]**(o-1)+(0.720054023732)
        ref[1,2,1]=o*(-0.390732462938)*x_ref[1]**(o-1)+(-0.484359019049)
        ref[1,3,0]=o*(0.565910084656)*x_ref[0]**(o-1)+(0.69105192657)
        ref[1,3,1]=o*(-0.449408955892)*x_ref[1]**(o-1)+(-0.789780518242)
        ref[1,4,0]=o*(0.602485034177)*x_ref[0]**(o-1)+(-0.345139282323)
        ref[1,4,1]=o*(-0.951334152921)*x_ref[1]**(o-1)+(-0.478700081579)
        ref[2,0,0]=o*(0.00114729937901)*x_ref[0]**(o-1)+(0.295356090995)
        ref[2,0,1]=o*(0.982261544035)*x_ref[1]**(o-1)+(0.357076141518)
        ref[2,1,0]=o*(0.653029330145)*x_ref[0]**(o-1)+(-0.0178804068723)
        ref[2,1,1]=o*(-0.0238303862422)*x_ref[1]**(o-1)+(0.983033975535)
        ref[2,2,0]=o*(0.680878878444)*x_ref[0]**(o-1)+(-0.404330541218)
        ref[2,2,1]=o*(0.856544883866)*x_ref[1]**(o-1)+(0.71899071821)
        ref[2,3,0]=o*(0.0674687202072)*x_ref[0]**(o-1)+(-0.748990717637)
        ref[2,3,1]=o*(0.185597242309)*x_ref[1]**(o-1)+(0.623849127358)
        ref[2,4,0]=o*(0.302869699915)*x_ref[0]**(o-1)+(-0.852621552847)
        ref[2,4,1]=o*(0.115557112172)*x_ref[1]**(o-1)+(-0.0827214208715)
        ref[3,0,0]=o*(0.0929072570643)*x_ref[0]**(o-1)+(0.533152382923)
        ref[3,0,1]=o*(-0.0913754984708)*x_ref[1]**(o-1)+(-0.25865969337)
        ref[3,1,0]=o*(-0.794799886368)*x_ref[0]**(o-1)+(-0.865263962776)
        ref[3,1,1]=o*(0.441343356932)*x_ref[1]**(o-1)+(-0.541975483415)
        ref[3,2,0]=o*(0.261341948686)*x_ref[0]**(o-1)+(0.378142729621)
        ref[3,2,1]=o*(0.0556054089673)*x_ref[1]**(o-1)+(-0.854195169318)
        ref[3,3,0]=o*(0.177222467454)*x_ref[0]**(o-1)+(-0.400820197499)
        ref[3,3,1]=o*(0.88513075927)*x_ref[1]**(o-1)+(0.0420082325044)
        ref[3,4,0]=o*(0.307178422145)*x_ref[0]**(o-1)+(-0.203765349758)
        ref[3,4,1]=o*(-0.546052990937)*x_ref[1]**(o-1)+(-0.960312136292)
      else:
        arg[0,0]=(0.767891828291)*x[0]**o+(-0.0840162180616)*x[0]+(-0.942354055114)*x[1]**o+(-0.458324281794)*x[1]+(-0.837589052489)*x[2]**o+(-0.49741065871)*x[2]
        arg[0,1]=(-0.770574732815)*x[0]**o+(0.949362675967)*x[0]+(0.334062389906)*x[1]**o+(0.589134582043)*x[1]+(0.611876443107)*x[2]**o+(-0.476746595186)*x[2]
        arg[0,2]=(-0.113076234857)*x[0]**o+(0.207116555982)*x[0]+(-0.434423505296)*x[1]**o+(0.582451788291)*x[1]+(-0.858451242875)*x[2]**o+(-0.229934112115)*x[2]
        arg[0,3]=(0.55494281628)*x[0]**o+(0.53116989281)*x[0]+(0.150549353151)*x[1]**o+(0.70886785918)*x[1]+(-0.148436740612)*x[2]**o+(0.755584812901)*x[2]
        arg[0,4]=(0.0855991607862)*x[0]**o+(0.509120320151)*x[0]+(-0.303374431746)*x[1]**o+(0.1813545791)*x[1]+(-0.0434667822132)*x[2]**o+(0.395447849194)*x[2]
        arg[1,0]=(-0.779720617706)*x[0]**o+(0.730668258583)*x[0]+(-0.0315884698035)*x[1]**o+(-0.0929845369431)*x[1]+(0.0853332239352)*x[2]**o+(0.303372304789)*x[2]
        arg[1,1]=(-0.687021672626)*x[0]**o+(-0.380139153436)*x[0]+(-0.624774641851)*x[1]**o+(0.331833510141)*x[1]+(-0.747115143343)*x[2]**o+(-0.901384102505)*x[2]
        arg[1,2]=(-0.0528543338921)*x[0]**o+(-0.683764012224)*x[0]+(0.498886779776)*x[1]**o+(-0.396777215862)*x[1]+(0.777051418657)*x[2]**o+(-0.786345142641)*x[2]
        arg[1,3]=(-0.823112017253)*x[0]**o+(0.278818264102)*x[0]+(-0.79372160065)*x[1]**o+(0.254044599609)*x[1]+(-0.372925254827)*x[2]**o+(-0.85054008508)*x[2]
        arg[1,4]=(-0.247156845703)*x[0]**o+(-0.944335229412)*x[0]+(-0.572532409179)*x[1]**o+(0.289091364421)*x[1]+(0.848874739954)*x[2]**o+(0.705098999827)*x[2]
        arg[2,0]=(0.0958493700534)*x[0]**o+(-0.334070579338)*x[0]+(-0.74117408785)*x[1]**o+(0.0609646344894)*x[1]+(-0.328888690751)*x[2]**o+(-0.0430772626841)*x[2]
        arg[2,1]=(0.846871401827)*x[0]**o+(0.379912211333)*x[0]+(-0.197881866966)*x[1]**o+(-0.551498190852)*x[1]+(0.416977164384)*x[2]**o+(0.09398324743)*x[2]
        arg[2,2]=(0.737261473925)*x[0]**o+(0.696028328637)*x[0]+(-0.735110324145)*x[1]**o+(0.949453963809)*x[1]+(-0.894722167055)*x[2]**o+(-0.16124057159)*x[2]
        arg[2,3]=(0.267556059297)*x[0]**o+(-0.633581895464)*x[0]+(-0.176232065927)*x[1]**o+(0.98533106984)*x[1]+(0.404671434387)*x[2]**o+(0.343560595562)*x[2]
        arg[2,4]=(0.502981833324)*x[0]**o+(0.716659442696)*x[0]+(-0.349106283426)*x[1]**o+(0.0726954534053)*x[1]+(0.468516366697)*x[2]**o+(0.328926164383)*x[2]
        arg[3,0]=(0.846509624251)*x[0]**o+(-0.34161660121)*x[0]+(0.25149102353)*x[1]**o+(-0.0130496471853)*x[1]+(-0.556826239055)*x[2]**o+(0.774658667066)*x[2]
        arg[3,1]=(-0.0777319662414)*x[0]**o+(0.992622365536)*x[0]+(0.720504365851)*x[1]**o+(0.0701085839568)*x[1]+(0.475573965193)*x[2]**o+(0.402599619829)*x[2]
        arg[3,2]=(-0.311228599853)*x[0]**o+(-0.732684576916)*x[0]+(0.308213192232)*x[1]**o+(-0.131325871896)*x[1]+(0.964631653839)*x[2]**o+(-0.600646614776)*x[2]
        arg[3,3]=(0.656589254036)*x[0]**o+(-0.275449493207)*x[0]+(-0.611075965309)*x[1]**o+(-0.460879628412)*x[1]+(-0.820861932527)*x[2]**o+(0.424066222721)*x[2]
        arg[3,4]=(0.303269333663)*x[0]**o+(-0.745774722275)*x[0]+(0.82933539351)*x[1]**o+(-0.295464064484)*x[1]+(0.818135297647)*x[2]**o+(-0.989600307642)*x[2]
        ref[0,0,0]=o*(0.767891828291)*x_ref[0]**(o-1)+(-0.0840162180616)
        ref[0,0,1]=o*(-0.942354055114)*x_ref[1]**(o-1)+(-0.458324281794)
        ref[0,0,2]=o*(-0.837589052489)*x_ref[2]**(o-1)+(-0.49741065871)
        ref[0,1,0]=o*(-0.770574732815)*x_ref[0]**(o-1)+(0.949362675967)
        ref[0,1,1]=o*(0.334062389906)*x_ref[1]**(o-1)+(0.589134582043)
        ref[0,1,2]=o*(0.611876443107)*x_ref[2]**(o-1)+(-0.476746595186)
        ref[0,2,0]=o*(-0.113076234857)*x_ref[0]**(o-1)+(0.207116555982)
        ref[0,2,1]=o*(-0.434423505296)*x_ref[1]**(o-1)+(0.582451788291)
        ref[0,2,2]=o*(-0.858451242875)*x_ref[2]**(o-1)+(-0.229934112115)
        ref[0,3,0]=o*(0.55494281628)*x_ref[0]**(o-1)+(0.53116989281)
        ref[0,3,1]=o*(0.150549353151)*x_ref[1]**(o-1)+(0.70886785918)
        ref[0,3,2]=o*(-0.148436740612)*x_ref[2]**(o-1)+(0.755584812901)
        ref[0,4,0]=o*(0.0855991607862)*x_ref[0]**(o-1)+(0.509120320151)
        ref[0,4,1]=o*(-0.303374431746)*x_ref[1]**(o-1)+(0.1813545791)
        ref[0,4,2]=o*(-0.0434667822132)*x_ref[2]**(o-1)+(0.395447849194)
        ref[1,0,0]=o*(-0.779720617706)*x_ref[0]**(o-1)+(0.730668258583)
        ref[1,0,1]=o*(-0.0315884698035)*x_ref[1]**(o-1)+(-0.0929845369431)
        ref[1,0,2]=o*(0.0853332239352)*x_ref[2]**(o-1)+(0.303372304789)
        ref[1,1,0]=o*(-0.687021672626)*x_ref[0]**(o-1)+(-0.380139153436)
        ref[1,1,1]=o*(-0.624774641851)*x_ref[1]**(o-1)+(0.331833510141)
        ref[1,1,2]=o*(-0.747115143343)*x_ref[2]**(o-1)+(-0.901384102505)
        ref[1,2,0]=o*(-0.0528543338921)*x_ref[0]**(o-1)+(-0.683764012224)
        ref[1,2,1]=o*(0.498886779776)*x_ref[1]**(o-1)+(-0.396777215862)
        ref[1,2,2]=o*(0.777051418657)*x_ref[2]**(o-1)+(-0.786345142641)
        ref[1,3,0]=o*(-0.823112017253)*x_ref[0]**(o-1)+(0.278818264102)
        ref[1,3,1]=o*(-0.79372160065)*x_ref[1]**(o-1)+(0.254044599609)
        ref[1,3,2]=o*(-0.372925254827)*x_ref[2]**(o-1)+(-0.85054008508)
        ref[1,4,0]=o*(-0.247156845703)*x_ref[0]**(o-1)+(-0.944335229412)
        ref[1,4,1]=o*(-0.572532409179)*x_ref[1]**(o-1)+(0.289091364421)
        ref[1,4,2]=o*(0.848874739954)*x_ref[2]**(o-1)+(0.705098999827)
        ref[2,0,0]=o*(0.0958493700534)*x_ref[0]**(o-1)+(-0.334070579338)
        ref[2,0,1]=o*(-0.74117408785)*x_ref[1]**(o-1)+(0.0609646344894)
        ref[2,0,2]=o*(-0.328888690751)*x_ref[2]**(o-1)+(-0.0430772626841)
        ref[2,1,0]=o*(0.846871401827)*x_ref[0]**(o-1)+(0.379912211333)
        ref[2,1,1]=o*(-0.197881866966)*x_ref[1]**(o-1)+(-0.551498190852)
        ref[2,1,2]=o*(0.416977164384)*x_ref[2]**(o-1)+(0.09398324743)
        ref[2,2,0]=o*(0.737261473925)*x_ref[0]**(o-1)+(0.696028328637)
        ref[2,2,1]=o*(-0.735110324145)*x_ref[1]**(o-1)+(0.949453963809)
        ref[2,2,2]=o*(-0.894722167055)*x_ref[2]**(o-1)+(-0.16124057159)
        ref[2,3,0]=o*(0.267556059297)*x_ref[0]**(o-1)+(-0.633581895464)
        ref[2,3,1]=o*(-0.176232065927)*x_ref[1]**(o-1)+(0.98533106984)
        ref[2,3,2]=o*(0.404671434387)*x_ref[2]**(o-1)+(0.343560595562)
        ref[2,4,0]=o*(0.502981833324)*x_ref[0]**(o-1)+(0.716659442696)
        ref[2,4,1]=o*(-0.349106283426)*x_ref[1]**(o-1)+(0.0726954534053)
        ref[2,4,2]=o*(0.468516366697)*x_ref[2]**(o-1)+(0.328926164383)
        ref[3,0,0]=o*(0.846509624251)*x_ref[0]**(o-1)+(-0.34161660121)
        ref[3,0,1]=o*(0.25149102353)*x_ref[1]**(o-1)+(-0.0130496471853)
        ref[3,0,2]=o*(-0.556826239055)*x_ref[2]**(o-1)+(0.774658667066)
        ref[3,1,0]=o*(-0.0777319662414)*x_ref[0]**(o-1)+(0.992622365536)
        ref[3,1,1]=o*(0.720504365851)*x_ref[1]**(o-1)+(0.0701085839568)
        ref[3,1,2]=o*(0.475573965193)*x_ref[2]**(o-1)+(0.402599619829)
        ref[3,2,0]=o*(-0.311228599853)*x_ref[0]**(o-1)+(-0.732684576916)
        ref[3,2,1]=o*(0.308213192232)*x_ref[1]**(o-1)+(-0.131325871896)
        ref[3,2,2]=o*(0.964631653839)*x_ref[2]**(o-1)+(-0.600646614776)
        ref[3,3,0]=o*(0.656589254036)*x_ref[0]**(o-1)+(-0.275449493207)
        ref[3,3,1]=o*(-0.611075965309)*x_ref[1]**(o-1)+(-0.460879628412)
        ref[3,3,2]=o*(-0.820861932527)*x_ref[2]**(o-1)+(0.424066222721)
        ref[3,4,0]=o*(0.303269333663)*x_ref[0]**(o-1)+(-0.745774722275)
        ref[3,4,1]=o*(0.82933539351)*x_ref[1]**(o-1)+(-0.295464064484)
        ref[3,4,2]=o*(0.818135297647)*x_ref[2]**(o-1)+(-0.989600307642)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.572693829631)*x[0]**o+(0.583556749477)*x[0]+(0.22594627003)*x[1]**o+(0.0199162626119)*x[1]
        arg[0,0,1]=(0.196266833125)*x[0]**o+(0.544824498113)*x[0]+(0.150736772157)*x[1]**o+(0.362174902716)*x[1]
        arg[0,1,0]=(0.169481539273)*x[0]**o+(0.0355504217105)*x[0]+(-0.156746013218)*x[1]**o+(0.879476257954)*x[1]
        arg[0,1,1]=(0.824826637612)*x[0]**o+(-0.732645177858)*x[0]+(0.732873648361)*x[1]**o+(-0.40090515696)*x[1]
        arg[1,0,0]=(0.77348720999)*x[0]**o+(0.229475702878)*x[0]+(0.686324947328)*x[1]**o+(0.313197971717)*x[1]
        arg[1,0,1]=(-0.966626441904)*x[0]**o+(0.989778282376)*x[0]+(0.0447214020611)*x[1]**o+(0.439337525343)*x[1]
        arg[1,1,0]=(-0.774806567769)*x[0]**o+(0.634591112025)*x[0]+(0.797720133135)*x[1]**o+(0.579137230562)*x[1]
        arg[1,1,1]=(-0.118703583742)*x[0]**o+(0.482822282284)*x[0]+(0.141839968056)*x[1]**o+(0.916615496002)*x[1]
        arg[2,0,0]=(0.7419378407)*x[0]**o+(0.754978933397)*x[0]+(-0.839586800033)*x[1]**o+(-0.525830008753)*x[1]
        arg[2,0,1]=(-0.855040380309)*x[0]**o+(0.423661210447)*x[0]+(0.538277902172)*x[1]**o+(0.127617006384)*x[1]
        arg[2,1,0]=(0.107349578938)*x[0]**o+(0.697738123598)*x[0]+(0.862106802027)*x[1]**o+(-0.829827651082)*x[1]
        arg[2,1,1]=(-0.970208311818)*x[0]**o+(0.806698181512)*x[0]+(-0.374007524144)*x[1]**o+(-0.0600501198736)*x[1]
        arg[3,0,0]=(0.223312188411)*x[0]**o+(-0.130776319489)*x[0]+(0.616021313401)*x[1]**o+(-0.552848902138)*x[1]
        arg[3,0,1]=(-0.461837450406)*x[0]**o+(-0.895912538242)*x[0]+(0.477606324718)*x[1]**o+(0.220148409572)*x[1]
        arg[3,1,0]=(0.0783005249695)*x[0]**o+(-0.762350707337)*x[0]+(0.616470467018)*x[1]**o+(-0.154358859634)*x[1]
        arg[3,1,1]=(-0.604945998091)*x[0]**o+(0.23801834237)*x[0]+(0.500666300529)*x[1]**o+(0.709482270038)*x[1]
        arg[4,0,0]=(0.746627054515)*x[0]**o+(0.550282853502)*x[0]+(-0.223527434147)*x[1]**o+(-0.291559486198)*x[1]
        arg[4,0,1]=(0.512462457431)*x[0]**o+(0.0908594735188)*x[0]+(-0.888862229845)*x[1]**o+(-0.969985508695)*x[1]
        arg[4,1,0]=(0.606213701819)*x[0]**o+(0.147049558356)*x[0]+(-0.0365908849884)*x[1]**o+(-0.584510320466)*x[1]
        arg[4,1,1]=(-0.110180223906)*x[0]**o+(-0.83442554862)*x[0]+(-0.0516097061459)*x[1]**o+(-0.155615857347)*x[1]
        arg[5,0,0]=(0.603753207781)*x[0]**o+(0.302032475639)*x[0]+(-0.68719165993)*x[1]**o+(-0.375829665239)*x[1]
        arg[5,0,1]=(-0.599357977805)*x[0]**o+(0.562486062075)*x[0]+(0.102714751356)*x[1]**o+(-0.114179799185)*x[1]
        arg[5,1,0]=(0.874756501256)*x[0]**o+(0.550416007969)*x[0]+(0.669721605367)*x[1]**o+(0.24372600222)*x[1]
        arg[5,1,1]=(0.651356889701)*x[0]**o+(0.178366196323)*x[0]+(0.762268544189)*x[1]**o+(0.515014711815)*x[1]
        ref[0,0,0,0]=o*(-0.572693829631)*x_ref[0]**(o-1)+(0.583556749477)
        ref[0,0,0,1]=o*(0.22594627003)*x_ref[1]**(o-1)+(0.0199162626119)
        ref[0,0,1,0]=o*(0.196266833125)*x_ref[0]**(o-1)+(0.544824498113)
        ref[0,0,1,1]=o*(0.150736772157)*x_ref[1]**(o-1)+(0.362174902716)
        ref[0,1,0,0]=o*(0.169481539273)*x_ref[0]**(o-1)+(0.0355504217105)
        ref[0,1,0,1]=o*(-0.156746013218)*x_ref[1]**(o-1)+(0.879476257954)
        ref[0,1,1,0]=o*(0.824826637612)*x_ref[0]**(o-1)+(-0.732645177858)
        ref[0,1,1,1]=o*(0.732873648361)*x_ref[1]**(o-1)+(-0.40090515696)
        ref[1,0,0,0]=o*(0.77348720999)*x_ref[0]**(o-1)+(0.229475702878)
        ref[1,0,0,1]=o*(0.686324947328)*x_ref[1]**(o-1)+(0.313197971717)
        ref[1,0,1,0]=o*(-0.966626441904)*x_ref[0]**(o-1)+(0.989778282376)
        ref[1,0,1,1]=o*(0.0447214020611)*x_ref[1]**(o-1)+(0.439337525343)
        ref[1,1,0,0]=o*(-0.774806567769)*x_ref[0]**(o-1)+(0.634591112025)
        ref[1,1,0,1]=o*(0.797720133135)*x_ref[1]**(o-1)+(0.579137230562)
        ref[1,1,1,0]=o*(-0.118703583742)*x_ref[0]**(o-1)+(0.482822282284)
        ref[1,1,1,1]=o*(0.141839968056)*x_ref[1]**(o-1)+(0.916615496002)
        ref[2,0,0,0]=o*(0.7419378407)*x_ref[0]**(o-1)+(0.754978933397)
        ref[2,0,0,1]=o*(-0.839586800033)*x_ref[1]**(o-1)+(-0.525830008753)
        ref[2,0,1,0]=o*(-0.855040380309)*x_ref[0]**(o-1)+(0.423661210447)
        ref[2,0,1,1]=o*(0.538277902172)*x_ref[1]**(o-1)+(0.127617006384)
        ref[2,1,0,0]=o*(0.107349578938)*x_ref[0]**(o-1)+(0.697738123598)
        ref[2,1,0,1]=o*(0.862106802027)*x_ref[1]**(o-1)+(-0.829827651082)
        ref[2,1,1,0]=o*(-0.970208311818)*x_ref[0]**(o-1)+(0.806698181512)
        ref[2,1,1,1]=o*(-0.374007524144)*x_ref[1]**(o-1)+(-0.0600501198736)
        ref[3,0,0,0]=o*(0.223312188411)*x_ref[0]**(o-1)+(-0.130776319489)
        ref[3,0,0,1]=o*(0.616021313401)*x_ref[1]**(o-1)+(-0.552848902138)
        ref[3,0,1,0]=o*(-0.461837450406)*x_ref[0]**(o-1)+(-0.895912538242)
        ref[3,0,1,1]=o*(0.477606324718)*x_ref[1]**(o-1)+(0.220148409572)
        ref[3,1,0,0]=o*(0.0783005249695)*x_ref[0]**(o-1)+(-0.762350707337)
        ref[3,1,0,1]=o*(0.616470467018)*x_ref[1]**(o-1)+(-0.154358859634)
        ref[3,1,1,0]=o*(-0.604945998091)*x_ref[0]**(o-1)+(0.23801834237)
        ref[3,1,1,1]=o*(0.500666300529)*x_ref[1]**(o-1)+(0.709482270038)
        ref[4,0,0,0]=o*(0.746627054515)*x_ref[0]**(o-1)+(0.550282853502)
        ref[4,0,0,1]=o*(-0.223527434147)*x_ref[1]**(o-1)+(-0.291559486198)
        ref[4,0,1,0]=o*(0.512462457431)*x_ref[0]**(o-1)+(0.0908594735188)
        ref[4,0,1,1]=o*(-0.888862229845)*x_ref[1]**(o-1)+(-0.969985508695)
        ref[4,1,0,0]=o*(0.606213701819)*x_ref[0]**(o-1)+(0.147049558356)
        ref[4,1,0,1]=o*(-0.0365908849884)*x_ref[1]**(o-1)+(-0.584510320466)
        ref[4,1,1,0]=o*(-0.110180223906)*x_ref[0]**(o-1)+(-0.83442554862)
        ref[4,1,1,1]=o*(-0.0516097061459)*x_ref[1]**(o-1)+(-0.155615857347)
        ref[5,0,0,0]=o*(0.603753207781)*x_ref[0]**(o-1)+(0.302032475639)
        ref[5,0,0,1]=o*(-0.68719165993)*x_ref[1]**(o-1)+(-0.375829665239)
        ref[5,0,1,0]=o*(-0.599357977805)*x_ref[0]**(o-1)+(0.562486062075)
        ref[5,0,1,1]=o*(0.102714751356)*x_ref[1]**(o-1)+(-0.114179799185)
        ref[5,1,0,0]=o*(0.874756501256)*x_ref[0]**(o-1)+(0.550416007969)
        ref[5,1,0,1]=o*(0.669721605367)*x_ref[1]**(o-1)+(0.24372600222)
        ref[5,1,1,0]=o*(0.651356889701)*x_ref[0]**(o-1)+(0.178366196323)
        ref[5,1,1,1]=o*(0.762268544189)*x_ref[1]**(o-1)+(0.515014711815)
      else:
        arg[0,0,0]=(-0.634739027866)*x[0]**o+(-0.696370463772)*x[0]+(0.503815931458)*x[1]**o+(-0.457824888032)*x[1]+(0.655074800235)*x[2]**o+(-0.456111009206)*x[2]
        arg[0,0,1]=(-0.533710147141)*x[0]**o+(-0.0814085290298)*x[0]+(-0.0894565935545)*x[1]**o+(-0.3775655102)*x[1]+(-0.134051544262)*x[2]**o+(0.000713749886013)*x[2]
        arg[0,1,0]=(-0.64394535332)*x[0]**o+(0.430741797171)*x[0]+(0.985157009863)*x[1]**o+(-0.655331266432)*x[1]+(-0.436629135988)*x[2]**o+(-0.404480211702)*x[2]
        arg[0,1,1]=(-0.0858143551662)*x[0]**o+(0.0452769288327)*x[0]+(0.930452744895)*x[1]**o+(-0.877149214078)*x[1]+(0.675472222413)*x[2]**o+(0.917497654616)*x[2]
        arg[1,0,0]=(0.88346309442)*x[0]**o+(0.892116796676)*x[0]+(0.22765643526)*x[1]**o+(-0.934442833938)*x[1]+(0.0781755340006)*x[2]**o+(0.106009847768)*x[2]
        arg[1,0,1]=(0.20824563298)*x[0]**o+(0.613584602599)*x[0]+(-0.482386438919)*x[1]**o+(-0.155840561301)*x[1]+(0.817890842096)*x[2]**o+(0.798519843541)*x[2]
        arg[1,1,0]=(-0.915853569389)*x[0]**o+(0.573041720657)*x[0]+(0.236967108991)*x[1]**o+(0.421805849528)*x[1]+(-0.576407564364)*x[2]**o+(0.0709741076428)*x[2]
        arg[1,1,1]=(0.530417188741)*x[0]**o+(0.872073272277)*x[0]+(-0.21047376763)*x[1]**o+(0.88490064655)*x[1]+(0.523085602451)*x[2]**o+(-0.926133382905)*x[2]
        arg[2,0,0]=(-0.974186019136)*x[0]**o+(-0.332433414417)*x[0]+(-0.500938330608)*x[1]**o+(-0.113188574497)*x[1]+(0.228656301416)*x[2]**o+(-0.375756425313)*x[2]
        arg[2,0,1]=(0.246057421673)*x[0]**o+(0.231587929687)*x[0]+(0.398413883486)*x[1]**o+(-0.390998026525)*x[1]+(0.273401149062)*x[2]**o+(-0.296142647578)*x[2]
        arg[2,1,0]=(0.283717204372)*x[0]**o+(-0.545852417255)*x[0]+(-0.616345461169)*x[1]**o+(-0.705051061202)*x[1]+(0.664377876346)*x[2]**o+(0.338382503272)*x[2]
        arg[2,1,1]=(-0.433297970157)*x[0]**o+(-0.716600154712)*x[0]+(0.836194932687)*x[1]**o+(0.0903372116119)*x[1]+(0.113374499353)*x[2]**o+(-0.588810627425)*x[2]
        arg[3,0,0]=(0.404388255474)*x[0]**o+(0.521816615546)*x[0]+(-0.0976581170689)*x[1]**o+(-0.678563989037)*x[1]+(0.550953471125)*x[2]**o+(-0.668793844235)*x[2]
        arg[3,0,1]=(-0.66023047868)*x[0]**o+(0.822794930168)*x[0]+(-0.284215677966)*x[1]**o+(-0.779237447318)*x[1]+(-0.191727752719)*x[2]**o+(-0.748596484262)*x[2]
        arg[3,1,0]=(-0.0610807571905)*x[0]**o+(0.044623781518)*x[0]+(0.851226664753)*x[1]**o+(0.921090244094)*x[1]+(0.494414571241)*x[2]**o+(-0.473831333684)*x[2]
        arg[3,1,1]=(0.54651807228)*x[0]**o+(-0.580525990778)*x[0]+(-0.822936606301)*x[1]**o+(-0.0169562878937)*x[1]+(-0.732219249555)*x[2]**o+(-0.435156155008)*x[2]
        arg[4,0,0]=(0.830021032139)*x[0]**o+(0.976836986062)*x[0]+(0.534162405111)*x[1]**o+(-0.286292103539)*x[1]+(0.820276894761)*x[2]**o+(0.245897387706)*x[2]
        arg[4,0,1]=(0.362201493311)*x[0]**o+(-0.186328147108)*x[0]+(-0.91611458572)*x[1]**o+(-0.965258904175)*x[1]+(0.61850292372)*x[2]**o+(0.410229835478)*x[2]
        arg[4,1,0]=(0.0726313013515)*x[0]**o+(-0.974714544914)*x[0]+(0.46313543268)*x[1]**o+(0.100486704072)*x[1]+(-0.149124857233)*x[2]**o+(0.455733236337)*x[2]
        arg[4,1,1]=(-0.875220220201)*x[0]**o+(-0.416152301789)*x[0]+(-0.0145627466423)*x[1]**o+(0.267856009119)*x[1]+(-0.604260174049)*x[2]**o+(0.689107997051)*x[2]
        arg[5,0,0]=(0.196182127971)*x[0]**o+(0.337209503833)*x[0]+(0.595342239158)*x[1]**o+(0.507423300857)*x[1]+(0.934760469389)*x[2]**o+(-0.198686443328)*x[2]
        arg[5,0,1]=(0.77210909967)*x[0]**o+(-0.310778136545)*x[0]+(0.512326103142)*x[1]**o+(-0.501056206243)*x[1]+(0.98755961257)*x[2]**o+(-0.658414340038)*x[2]
        arg[5,1,0]=(0.375277702198)*x[0]**o+(-0.603850047717)*x[0]+(0.677227868069)*x[1]**o+(-0.784640615159)*x[1]+(0.177350721639)*x[2]**o+(0.241053657333)*x[2]
        arg[5,1,1]=(-0.368625522736)*x[0]**o+(0.196977476909)*x[0]+(-0.582451887206)*x[1]**o+(0.598240904158)*x[1]+(0.102240230796)*x[2]**o+(0.487322529158)*x[2]
        ref[0,0,0,0]=o*(-0.634739027866)*x_ref[0]**(o-1)+(-0.696370463772)
        ref[0,0,0,1]=o*(0.503815931458)*x_ref[1]**(o-1)+(-0.457824888032)
        ref[0,0,0,2]=o*(0.655074800235)*x_ref[2]**(o-1)+(-0.456111009206)
        ref[0,0,1,0]=o*(-0.533710147141)*x_ref[0]**(o-1)+(-0.0814085290298)
        ref[0,0,1,1]=o*(-0.0894565935545)*x_ref[1]**(o-1)+(-0.3775655102)
        ref[0,0,1,2]=o*(-0.134051544262)*x_ref[2]**(o-1)+(0.000713749886013)
        ref[0,1,0,0]=o*(-0.64394535332)*x_ref[0]**(o-1)+(0.430741797171)
        ref[0,1,0,1]=o*(0.985157009863)*x_ref[1]**(o-1)+(-0.655331266432)
        ref[0,1,0,2]=o*(-0.436629135988)*x_ref[2]**(o-1)+(-0.404480211702)
        ref[0,1,1,0]=o*(-0.0858143551662)*x_ref[0]**(o-1)+(0.0452769288327)
        ref[0,1,1,1]=o*(0.930452744895)*x_ref[1]**(o-1)+(-0.877149214078)
        ref[0,1,1,2]=o*(0.675472222413)*x_ref[2]**(o-1)+(0.917497654616)
        ref[1,0,0,0]=o*(0.88346309442)*x_ref[0]**(o-1)+(0.892116796676)
        ref[1,0,0,1]=o*(0.22765643526)*x_ref[1]**(o-1)+(-0.934442833938)
        ref[1,0,0,2]=o*(0.0781755340006)*x_ref[2]**(o-1)+(0.106009847768)
        ref[1,0,1,0]=o*(0.20824563298)*x_ref[0]**(o-1)+(0.613584602599)
        ref[1,0,1,1]=o*(-0.482386438919)*x_ref[1]**(o-1)+(-0.155840561301)
        ref[1,0,1,2]=o*(0.817890842096)*x_ref[2]**(o-1)+(0.798519843541)
        ref[1,1,0,0]=o*(-0.915853569389)*x_ref[0]**(o-1)+(0.573041720657)
        ref[1,1,0,1]=o*(0.236967108991)*x_ref[1]**(o-1)+(0.421805849528)
        ref[1,1,0,2]=o*(-0.576407564364)*x_ref[2]**(o-1)+(0.0709741076428)
        ref[1,1,1,0]=o*(0.530417188741)*x_ref[0]**(o-1)+(0.872073272277)
        ref[1,1,1,1]=o*(-0.21047376763)*x_ref[1]**(o-1)+(0.88490064655)
        ref[1,1,1,2]=o*(0.523085602451)*x_ref[2]**(o-1)+(-0.926133382905)
        ref[2,0,0,0]=o*(-0.974186019136)*x_ref[0]**(o-1)+(-0.332433414417)
        ref[2,0,0,1]=o*(-0.500938330608)*x_ref[1]**(o-1)+(-0.113188574497)
        ref[2,0,0,2]=o*(0.228656301416)*x_ref[2]**(o-1)+(-0.375756425313)
        ref[2,0,1,0]=o*(0.246057421673)*x_ref[0]**(o-1)+(0.231587929687)
        ref[2,0,1,1]=o*(0.398413883486)*x_ref[1]**(o-1)+(-0.390998026525)
        ref[2,0,1,2]=o*(0.273401149062)*x_ref[2]**(o-1)+(-0.296142647578)
        ref[2,1,0,0]=o*(0.283717204372)*x_ref[0]**(o-1)+(-0.545852417255)
        ref[2,1,0,1]=o*(-0.616345461169)*x_ref[1]**(o-1)+(-0.705051061202)
        ref[2,1,0,2]=o*(0.664377876346)*x_ref[2]**(o-1)+(0.338382503272)
        ref[2,1,1,0]=o*(-0.433297970157)*x_ref[0]**(o-1)+(-0.716600154712)
        ref[2,1,1,1]=o*(0.836194932687)*x_ref[1]**(o-1)+(0.0903372116119)
        ref[2,1,1,2]=o*(0.113374499353)*x_ref[2]**(o-1)+(-0.588810627425)
        ref[3,0,0,0]=o*(0.404388255474)*x_ref[0]**(o-1)+(0.521816615546)
        ref[3,0,0,1]=o*(-0.0976581170689)*x_ref[1]**(o-1)+(-0.678563989037)
        ref[3,0,0,2]=o*(0.550953471125)*x_ref[2]**(o-1)+(-0.668793844235)
        ref[3,0,1,0]=o*(-0.66023047868)*x_ref[0]**(o-1)+(0.822794930168)
        ref[3,0,1,1]=o*(-0.284215677966)*x_ref[1]**(o-1)+(-0.779237447318)
        ref[3,0,1,2]=o*(-0.191727752719)*x_ref[2]**(o-1)+(-0.748596484262)
        ref[3,1,0,0]=o*(-0.0610807571905)*x_ref[0]**(o-1)+(0.044623781518)
        ref[3,1,0,1]=o*(0.851226664753)*x_ref[1]**(o-1)+(0.921090244094)
        ref[3,1,0,2]=o*(0.494414571241)*x_ref[2]**(o-1)+(-0.473831333684)
        ref[3,1,1,0]=o*(0.54651807228)*x_ref[0]**(o-1)+(-0.580525990778)
        ref[3,1,1,1]=o*(-0.822936606301)*x_ref[1]**(o-1)+(-0.0169562878937)
        ref[3,1,1,2]=o*(-0.732219249555)*x_ref[2]**(o-1)+(-0.435156155008)
        ref[4,0,0,0]=o*(0.830021032139)*x_ref[0]**(o-1)+(0.976836986062)
        ref[4,0,0,1]=o*(0.534162405111)*x_ref[1]**(o-1)+(-0.286292103539)
        ref[4,0,0,2]=o*(0.820276894761)*x_ref[2]**(o-1)+(0.245897387706)
        ref[4,0,1,0]=o*(0.362201493311)*x_ref[0]**(o-1)+(-0.186328147108)
        ref[4,0,1,1]=o*(-0.91611458572)*x_ref[1]**(o-1)+(-0.965258904175)
        ref[4,0,1,2]=o*(0.61850292372)*x_ref[2]**(o-1)+(0.410229835478)
        ref[4,1,0,0]=o*(0.0726313013515)*x_ref[0]**(o-1)+(-0.974714544914)
        ref[4,1,0,1]=o*(0.46313543268)*x_ref[1]**(o-1)+(0.100486704072)
        ref[4,1,0,2]=o*(-0.149124857233)*x_ref[2]**(o-1)+(0.455733236337)
        ref[4,1,1,0]=o*(-0.875220220201)*x_ref[0]**(o-1)+(-0.416152301789)
        ref[4,1,1,1]=o*(-0.0145627466423)*x_ref[1]**(o-1)+(0.267856009119)
        ref[4,1,1,2]=o*(-0.604260174049)*x_ref[2]**(o-1)+(0.689107997051)
        ref[5,0,0,0]=o*(0.196182127971)*x_ref[0]**(o-1)+(0.337209503833)
        ref[5,0,0,1]=o*(0.595342239158)*x_ref[1]**(o-1)+(0.507423300857)
        ref[5,0,0,2]=o*(0.934760469389)*x_ref[2]**(o-1)+(-0.198686443328)
        ref[5,0,1,0]=o*(0.77210909967)*x_ref[0]**(o-1)+(-0.310778136545)
        ref[5,0,1,1]=o*(0.512326103142)*x_ref[1]**(o-1)+(-0.501056206243)
        ref[5,0,1,2]=o*(0.98755961257)*x_ref[2]**(o-1)+(-0.658414340038)
        ref[5,1,0,0]=o*(0.375277702198)*x_ref[0]**(o-1)+(-0.603850047717)
        ref[5,1,0,1]=o*(0.677227868069)*x_ref[1]**(o-1)+(-0.784640615159)
        ref[5,1,0,2]=o*(0.177350721639)*x_ref[2]**(o-1)+(0.241053657333)
        ref[5,1,1,0]=o*(-0.368625522736)*x_ref[0]**(o-1)+(0.196977476909)
        ref[5,1,1,1]=o*(-0.582451887206)*x_ref[1]**(o-1)+(0.598240904158)
        ref[5,1,1,2]=o*(0.102240230796)*x_ref[2]**(o-1)+(0.487322529158)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.851643614423)*x[0]**o+(-0.77638754265)*x[0]+(-0.216746360457)*x[1]**o+(-0.117807364973)*x[1]
        ref[0]=o*(0.851643614423)*x_ref[0]**(o-1)+(-0.77638754265)
        ref[1]=o*(-0.216746360457)*x_ref[1]**(o-1)+(-0.117807364973)
      else:
        arg=(0.846271449213)*x[0]**o+(0.895578416904)*x[0]+(-0.518297490388)*x[1]**o+(0.272255265951)*x[1]+(0.537822992528)*x[2]**o+(0.74226977002)*x[2]
        ref[0]=o*(0.846271449213)*x_ref[0]**(o-1)+(0.895578416904)
        ref[1]=o*(-0.518297490388)*x_ref[1]**(o-1)+(0.272255265951)
        ref[2]=o*(0.537822992528)*x_ref[2]**(o-1)+(0.74226977002)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.365601318395)*x[0]**o+(0.295624555802)*x[0]+(0.540618182217)*x[1]**o+(0.0458416078386)*x[1]
        arg[1]=(0.332276039998)*x[0]**o+(0.422606936253)*x[0]+(-0.221603594854)*x[1]**o+(0.879863420258)*x[1]
        ref[0,0]=o*(0.365601318395)*x_ref[0]**(o-1)+(0.295624555802)
        ref[0,1]=o*(0.540618182217)*x_ref[1]**(o-1)+(0.0458416078386)
        ref[1,0]=o*(0.332276039998)*x_ref[0]**(o-1)+(0.422606936253)
        ref[1,1]=o*(-0.221603594854)*x_ref[1]**(o-1)+(0.879863420258)
      else:
        arg[0]=(-0.258072277129)*x[0]**o+(-0.833085462518)*x[0]+(-0.124240965858)*x[1]**o+(-0.121853073047)*x[1]+(-0.877974471569)*x[2]**o+(0.929660451331)*x[2]
        arg[1]=(0.306520203535)*x[0]**o+(0.635444320723)*x[0]+(-0.766579509915)*x[1]**o+(-0.381189007175)*x[1]+(-0.76608952351)*x[2]**o+(0.00718583364881)*x[2]
        ref[0,0]=o*(-0.258072277129)*x_ref[0]**(o-1)+(-0.833085462518)
        ref[0,1]=o*(-0.124240965858)*x_ref[1]**(o-1)+(-0.121853073047)
        ref[0,2]=o*(-0.877974471569)*x_ref[2]**(o-1)+(0.929660451331)
        ref[1,0]=o*(0.306520203535)*x_ref[0]**(o-1)+(0.635444320723)
        ref[1,1]=o*(-0.766579509915)*x_ref[1]**(o-1)+(-0.381189007175)
        ref[1,2]=o*(-0.76608952351)*x_ref[2]**(o-1)+(0.00718583364881)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.403088506806)*x[0]**o+(-0.436942773988)*x[0]+(-0.0594011537346)*x[1]**o+(0.855063195595)*x[1]
        arg[0,1]=(-0.516992303892)*x[0]**o+(-0.132586737299)*x[0]+(-0.854639239618)*x[1]**o+(0.241200454455)*x[1]
        arg[0,2]=(-0.373941155498)*x[0]**o+(0.933328225549)*x[0]+(-0.986870460316)*x[1]**o+(-0.051265908773)*x[1]
        arg[0,3]=(-0.175701653328)*x[0]**o+(-0.726542005356)*x[0]+(-0.553505635409)*x[1]**o+(-0.199806969718)*x[1]
        arg[0,4]=(-0.750832127973)*x[0]**o+(0.404511387161)*x[0]+(0.874423404897)*x[1]**o+(0.35360129734)*x[1]
        arg[1,0]=(-0.000191271001285)*x[0]**o+(-0.319365399066)*x[0]+(0.398444976089)*x[1]**o+(-0.966391658006)*x[1]
        arg[1,1]=(-0.58562890965)*x[0]**o+(-0.623992315262)*x[0]+(-0.913704774126)*x[1]**o+(0.910305409081)*x[1]
        arg[1,2]=(-0.480491011401)*x[0]**o+(-0.387015926056)*x[0]+(0.845826744282)*x[1]**o+(0.380626980034)*x[1]
        arg[1,3]=(-0.284167706438)*x[0]**o+(0.730574128045)*x[0]+(0.823389953051)*x[1]**o+(0.59483032515)*x[1]
        arg[1,4]=(-0.545147810354)*x[0]**o+(-0.632923743497)*x[0]+(0.619596244096)*x[1]**o+(-0.816596365827)*x[1]
        arg[2,0]=(0.161289895106)*x[0]**o+(0.435472177094)*x[0]+(0.125591455076)*x[1]**o+(0.0207320895937)*x[1]
        arg[2,1]=(0.639457164801)*x[0]**o+(0.2631408826)*x[0]+(0.00255095677899)*x[1]**o+(0.732164720004)*x[1]
        arg[2,2]=(-0.152014092979)*x[0]**o+(0.347024615036)*x[0]+(0.94212294017)*x[1]**o+(0.173939724203)*x[1]
        arg[2,3]=(-0.0324872553975)*x[0]**o+(-0.969492827039)*x[0]+(-0.830876162378)*x[1]**o+(0.601308513636)*x[1]
        arg[2,4]=(0.932999309515)*x[0]**o+(0.0320937041249)*x[0]+(-0.354492894455)*x[1]**o+(-0.752837146132)*x[1]
        arg[3,0]=(0.995147840382)*x[0]**o+(-0.276974278934)*x[0]+(0.905272259272)*x[1]**o+(-0.433178159448)*x[1]
        arg[3,1]=(-0.291708633096)*x[0]**o+(-0.542000555721)*x[0]+(0.372404748364)*x[1]**o+(0.23140215774)*x[1]
        arg[3,2]=(0.88604771071)*x[0]**o+(-0.753078356803)*x[0]+(-0.265302521749)*x[1]**o+(-0.948898850214)*x[1]
        arg[3,3]=(-0.429496844082)*x[0]**o+(0.0731695917195)*x[0]+(0.291387878783)*x[1]**o+(-0.078863825039)*x[1]
        arg[3,4]=(-0.0067999635984)*x[0]**o+(0.488018496019)*x[0]+(-0.875646054486)*x[1]**o+(0.629803284064)*x[1]
        ref[0,0,0]=o*(0.403088506806)*x_ref[0]**(o-1)+(-0.436942773988)
        ref[0,0,1]=o*(-0.0594011537346)*x_ref[1]**(o-1)+(0.855063195595)
        ref[0,1,0]=o*(-0.516992303892)*x_ref[0]**(o-1)+(-0.132586737299)
        ref[0,1,1]=o*(-0.854639239618)*x_ref[1]**(o-1)+(0.241200454455)
        ref[0,2,0]=o*(-0.373941155498)*x_ref[0]**(o-1)+(0.933328225549)
        ref[0,2,1]=o*(-0.986870460316)*x_ref[1]**(o-1)+(-0.051265908773)
        ref[0,3,0]=o*(-0.175701653328)*x_ref[0]**(o-1)+(-0.726542005356)
        ref[0,3,1]=o*(-0.553505635409)*x_ref[1]**(o-1)+(-0.199806969718)
        ref[0,4,0]=o*(-0.750832127973)*x_ref[0]**(o-1)+(0.404511387161)
        ref[0,4,1]=o*(0.874423404897)*x_ref[1]**(o-1)+(0.35360129734)
        ref[1,0,0]=o*(-0.000191271001285)*x_ref[0]**(o-1)+(-0.319365399066)
        ref[1,0,1]=o*(0.398444976089)*x_ref[1]**(o-1)+(-0.966391658006)
        ref[1,1,0]=o*(-0.58562890965)*x_ref[0]**(o-1)+(-0.623992315262)
        ref[1,1,1]=o*(-0.913704774126)*x_ref[1]**(o-1)+(0.910305409081)
        ref[1,2,0]=o*(-0.480491011401)*x_ref[0]**(o-1)+(-0.387015926056)
        ref[1,2,1]=o*(0.845826744282)*x_ref[1]**(o-1)+(0.380626980034)
        ref[1,3,0]=o*(-0.284167706438)*x_ref[0]**(o-1)+(0.730574128045)
        ref[1,3,1]=o*(0.823389953051)*x_ref[1]**(o-1)+(0.59483032515)
        ref[1,4,0]=o*(-0.545147810354)*x_ref[0]**(o-1)+(-0.632923743497)
        ref[1,4,1]=o*(0.619596244096)*x_ref[1]**(o-1)+(-0.816596365827)
        ref[2,0,0]=o*(0.161289895106)*x_ref[0]**(o-1)+(0.435472177094)
        ref[2,0,1]=o*(0.125591455076)*x_ref[1]**(o-1)+(0.0207320895937)
        ref[2,1,0]=o*(0.639457164801)*x_ref[0]**(o-1)+(0.2631408826)
        ref[2,1,1]=o*(0.00255095677899)*x_ref[1]**(o-1)+(0.732164720004)
        ref[2,2,0]=o*(-0.152014092979)*x_ref[0]**(o-1)+(0.347024615036)
        ref[2,2,1]=o*(0.94212294017)*x_ref[1]**(o-1)+(0.173939724203)
        ref[2,3,0]=o*(-0.0324872553975)*x_ref[0]**(o-1)+(-0.969492827039)
        ref[2,3,1]=o*(-0.830876162378)*x_ref[1]**(o-1)+(0.601308513636)
        ref[2,4,0]=o*(0.932999309515)*x_ref[0]**(o-1)+(0.0320937041249)
        ref[2,4,1]=o*(-0.354492894455)*x_ref[1]**(o-1)+(-0.752837146132)
        ref[3,0,0]=o*(0.995147840382)*x_ref[0]**(o-1)+(-0.276974278934)
        ref[3,0,1]=o*(0.905272259272)*x_ref[1]**(o-1)+(-0.433178159448)
        ref[3,1,0]=o*(-0.291708633096)*x_ref[0]**(o-1)+(-0.542000555721)
        ref[3,1,1]=o*(0.372404748364)*x_ref[1]**(o-1)+(0.23140215774)
        ref[3,2,0]=o*(0.88604771071)*x_ref[0]**(o-1)+(-0.753078356803)
        ref[3,2,1]=o*(-0.265302521749)*x_ref[1]**(o-1)+(-0.948898850214)
        ref[3,3,0]=o*(-0.429496844082)*x_ref[0]**(o-1)+(0.0731695917195)
        ref[3,3,1]=o*(0.291387878783)*x_ref[1]**(o-1)+(-0.078863825039)
        ref[3,4,0]=o*(-0.0067999635984)*x_ref[0]**(o-1)+(0.488018496019)
        ref[3,4,1]=o*(-0.875646054486)*x_ref[1]**(o-1)+(0.629803284064)
      else:
        arg[0,0]=(-0.512668741406)*x[0]**o+(-0.73142104842)*x[0]+(0.489050993226)*x[1]**o+(0.771941571419)*x[1]+(-0.538092512577)*x[2]**o+(-0.457812389328)*x[2]
        arg[0,1]=(0.258566436386)*x[0]**o+(-0.572006860118)*x[0]+(-0.0370305840623)*x[1]**o+(0.328810934717)*x[1]+(-0.652873123265)*x[2]**o+(0.734929914374)*x[2]
        arg[0,2]=(-0.228462027625)*x[0]**o+(-0.873359484139)*x[0]+(0.902575701641)*x[1]**o+(0.428758160478)*x[1]+(-0.377042847702)*x[2]**o+(-0.161364604019)*x[2]
        arg[0,3]=(0.31338185054)*x[0]**o+(0.973849539932)*x[0]+(-0.403569231167)*x[1]**o+(-0.110026269257)*x[1]+(0.603101392345)*x[2]**o+(0.499232034173)*x[2]
        arg[0,4]=(0.468416317974)*x[0]**o+(-0.0154203909264)*x[0]+(0.409228888268)*x[1]**o+(0.358886480877)*x[1]+(-0.00677343401301)*x[2]**o+(-0.0567153195302)*x[2]
        arg[1,0]=(0.614929795628)*x[0]**o+(0.54911679279)*x[0]+(0.527138394555)*x[1]**o+(-0.84070465171)*x[1]+(0.421928294706)*x[2]**o+(-0.871734463848)*x[2]
        arg[1,1]=(-0.498770327357)*x[0]**o+(0.294610855431)*x[0]+(-0.815129074288)*x[1]**o+(0.395332176796)*x[1]+(-0.555947305527)*x[2]**o+(-0.623550814988)*x[2]
        arg[1,2]=(-0.369404109593)*x[0]**o+(-0.662927702362)*x[0]+(0.886789099892)*x[1]**o+(-0.815728539359)*x[1]+(-0.843861162574)*x[2]**o+(-0.510039607856)*x[2]
        arg[1,3]=(-0.747305678574)*x[0]**o+(0.350337618157)*x[0]+(0.313491051893)*x[1]**o+(-0.554840960712)*x[1]+(-0.259760094369)*x[2]**o+(0.904246223768)*x[2]
        arg[1,4]=(0.485859115418)*x[0]**o+(-0.751900240375)*x[0]+(-0.233384727142)*x[1]**o+(-0.466677891749)*x[1]+(-0.771311310708)*x[2]**o+(0.380515497846)*x[2]
        arg[2,0]=(0.628529887016)*x[0]**o+(0.778810609149)*x[0]+(-0.95647127374)*x[1]**o+(-0.887895309984)*x[1]+(-0.926017554164)*x[2]**o+(0.0922106889932)*x[2]
        arg[2,1]=(-0.981382750453)*x[0]**o+(0.858805367823)*x[0]+(-0.202352133378)*x[1]**o+(0.527229765084)*x[1]+(-0.0862956097426)*x[2]**o+(-0.76269735579)*x[2]
        arg[2,2]=(0.591276459941)*x[0]**o+(0.554086306469)*x[0]+(-0.175045397917)*x[1]**o+(0.319542375346)*x[1]+(0.328810955576)*x[2]**o+(-0.245381809232)*x[2]
        arg[2,3]=(-0.826436985287)*x[0]**o+(-0.931352535769)*x[0]+(-0.480801956568)*x[1]**o+(0.830773509938)*x[1]+(0.157556801982)*x[2]**o+(0.982505166946)*x[2]
        arg[2,4]=(-0.0992736644463)*x[0]**o+(-0.147525710611)*x[0]+(-0.723322157973)*x[1]**o+(0.0527438662093)*x[1]+(-0.338847227522)*x[2]**o+(-0.704585609897)*x[2]
        arg[3,0]=(-0.45250460769)*x[0]**o+(-0.323299292565)*x[0]+(-0.435540041421)*x[1]**o+(0.418330181239)*x[1]+(0.526739853977)*x[2]**o+(-0.00513414974056)*x[2]
        arg[3,1]=(0.289534600598)*x[0]**o+(0.467498987256)*x[0]+(0.27851903721)*x[1]**o+(0.908249187588)*x[1]+(-0.84695489717)*x[2]**o+(-0.799372710877)*x[2]
        arg[3,2]=(0.00338011092566)*x[0]**o+(0.819216072868)*x[0]+(0.115437140937)*x[1]**o+(0.0968835984948)*x[1]+(-0.633253222581)*x[2]**o+(-0.0224890801282)*x[2]
        arg[3,3]=(0.578331574145)*x[0]**o+(0.328247456626)*x[0]+(-0.725914955131)*x[1]**o+(-0.863562766028)*x[1]+(-0.283057959979)*x[2]**o+(-0.521563360583)*x[2]
        arg[3,4]=(0.345725352698)*x[0]**o+(0.614912618718)*x[0]+(-0.218084952915)*x[1]**o+(-0.629514485943)*x[1]+(-0.207605980168)*x[2]**o+(-0.0950042458608)*x[2]
        ref[0,0,0]=o*(-0.512668741406)*x_ref[0]**(o-1)+(-0.73142104842)
        ref[0,0,1]=o*(0.489050993226)*x_ref[1]**(o-1)+(0.771941571419)
        ref[0,0,2]=o*(-0.538092512577)*x_ref[2]**(o-1)+(-0.457812389328)
        ref[0,1,0]=o*(0.258566436386)*x_ref[0]**(o-1)+(-0.572006860118)
        ref[0,1,1]=o*(-0.0370305840623)*x_ref[1]**(o-1)+(0.328810934717)
        ref[0,1,2]=o*(-0.652873123265)*x_ref[2]**(o-1)+(0.734929914374)
        ref[0,2,0]=o*(-0.228462027625)*x_ref[0]**(o-1)+(-0.873359484139)
        ref[0,2,1]=o*(0.902575701641)*x_ref[1]**(o-1)+(0.428758160478)
        ref[0,2,2]=o*(-0.377042847702)*x_ref[2]**(o-1)+(-0.161364604019)
        ref[0,3,0]=o*(0.31338185054)*x_ref[0]**(o-1)+(0.973849539932)
        ref[0,3,1]=o*(-0.403569231167)*x_ref[1]**(o-1)+(-0.110026269257)
        ref[0,3,2]=o*(0.603101392345)*x_ref[2]**(o-1)+(0.499232034173)
        ref[0,4,0]=o*(0.468416317974)*x_ref[0]**(o-1)+(-0.0154203909264)
        ref[0,4,1]=o*(0.409228888268)*x_ref[1]**(o-1)+(0.358886480877)
        ref[0,4,2]=o*(-0.00677343401301)*x_ref[2]**(o-1)+(-0.0567153195302)
        ref[1,0,0]=o*(0.614929795628)*x_ref[0]**(o-1)+(0.54911679279)
        ref[1,0,1]=o*(0.527138394555)*x_ref[1]**(o-1)+(-0.84070465171)
        ref[1,0,2]=o*(0.421928294706)*x_ref[2]**(o-1)+(-0.871734463848)
        ref[1,1,0]=o*(-0.498770327357)*x_ref[0]**(o-1)+(0.294610855431)
        ref[1,1,1]=o*(-0.815129074288)*x_ref[1]**(o-1)+(0.395332176796)
        ref[1,1,2]=o*(-0.555947305527)*x_ref[2]**(o-1)+(-0.623550814988)
        ref[1,2,0]=o*(-0.369404109593)*x_ref[0]**(o-1)+(-0.662927702362)
        ref[1,2,1]=o*(0.886789099892)*x_ref[1]**(o-1)+(-0.815728539359)
        ref[1,2,2]=o*(-0.843861162574)*x_ref[2]**(o-1)+(-0.510039607856)
        ref[1,3,0]=o*(-0.747305678574)*x_ref[0]**(o-1)+(0.350337618157)
        ref[1,3,1]=o*(0.313491051893)*x_ref[1]**(o-1)+(-0.554840960712)
        ref[1,3,2]=o*(-0.259760094369)*x_ref[2]**(o-1)+(0.904246223768)
        ref[1,4,0]=o*(0.485859115418)*x_ref[0]**(o-1)+(-0.751900240375)
        ref[1,4,1]=o*(-0.233384727142)*x_ref[1]**(o-1)+(-0.466677891749)
        ref[1,4,2]=o*(-0.771311310708)*x_ref[2]**(o-1)+(0.380515497846)
        ref[2,0,0]=o*(0.628529887016)*x_ref[0]**(o-1)+(0.778810609149)
        ref[2,0,1]=o*(-0.95647127374)*x_ref[1]**(o-1)+(-0.887895309984)
        ref[2,0,2]=o*(-0.926017554164)*x_ref[2]**(o-1)+(0.0922106889932)
        ref[2,1,0]=o*(-0.981382750453)*x_ref[0]**(o-1)+(0.858805367823)
        ref[2,1,1]=o*(-0.202352133378)*x_ref[1]**(o-1)+(0.527229765084)
        ref[2,1,2]=o*(-0.0862956097426)*x_ref[2]**(o-1)+(-0.76269735579)
        ref[2,2,0]=o*(0.591276459941)*x_ref[0]**(o-1)+(0.554086306469)
        ref[2,2,1]=o*(-0.175045397917)*x_ref[1]**(o-1)+(0.319542375346)
        ref[2,2,2]=o*(0.328810955576)*x_ref[2]**(o-1)+(-0.245381809232)
        ref[2,3,0]=o*(-0.826436985287)*x_ref[0]**(o-1)+(-0.931352535769)
        ref[2,3,1]=o*(-0.480801956568)*x_ref[1]**(o-1)+(0.830773509938)
        ref[2,3,2]=o*(0.157556801982)*x_ref[2]**(o-1)+(0.982505166946)
        ref[2,4,0]=o*(-0.0992736644463)*x_ref[0]**(o-1)+(-0.147525710611)
        ref[2,4,1]=o*(-0.723322157973)*x_ref[1]**(o-1)+(0.0527438662093)
        ref[2,4,2]=o*(-0.338847227522)*x_ref[2]**(o-1)+(-0.704585609897)
        ref[3,0,0]=o*(-0.45250460769)*x_ref[0]**(o-1)+(-0.323299292565)
        ref[3,0,1]=o*(-0.435540041421)*x_ref[1]**(o-1)+(0.418330181239)
        ref[3,0,2]=o*(0.526739853977)*x_ref[2]**(o-1)+(-0.00513414974056)
        ref[3,1,0]=o*(0.289534600598)*x_ref[0]**(o-1)+(0.467498987256)
        ref[3,1,1]=o*(0.27851903721)*x_ref[1]**(o-1)+(0.908249187588)
        ref[3,1,2]=o*(-0.84695489717)*x_ref[2]**(o-1)+(-0.799372710877)
        ref[3,2,0]=o*(0.00338011092566)*x_ref[0]**(o-1)+(0.819216072868)
        ref[3,2,1]=o*(0.115437140937)*x_ref[1]**(o-1)+(0.0968835984948)
        ref[3,2,2]=o*(-0.633253222581)*x_ref[2]**(o-1)+(-0.0224890801282)
        ref[3,3,0]=o*(0.578331574145)*x_ref[0]**(o-1)+(0.328247456626)
        ref[3,3,1]=o*(-0.725914955131)*x_ref[1]**(o-1)+(-0.863562766028)
        ref[3,3,2]=o*(-0.283057959979)*x_ref[2]**(o-1)+(-0.521563360583)
        ref[3,4,0]=o*(0.345725352698)*x_ref[0]**(o-1)+(0.614912618718)
        ref[3,4,1]=o*(-0.218084952915)*x_ref[1]**(o-1)+(-0.629514485943)
        ref[3,4,2]=o*(-0.207605980168)*x_ref[2]**(o-1)+(-0.0950042458608)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.353390112117)*x[0]**o+(0.121267964347)*x[0]+(0.095925646757)*x[1]**o+(-0.00141503525661)*x[1]
        arg[0,0,1]=(0.610556742471)*x[0]**o+(0.542948343923)*x[0]+(0.124762431991)*x[1]**o+(0.00216142978003)*x[1]
        arg[0,1,0]=(0.261559018753)*x[0]**o+(-0.881689419767)*x[0]+(-0.009352418091)*x[1]**o+(0.333357441723)*x[1]
        arg[0,1,1]=(-0.25379927884)*x[0]**o+(-0.124076833523)*x[0]+(-0.167690772835)*x[1]**o+(0.89804632709)*x[1]
        arg[1,0,0]=(-0.39353207502)*x[0]**o+(0.838815579522)*x[0]+(0.829467918805)*x[1]**o+(0.614403367394)*x[1]
        arg[1,0,1]=(-0.698946975112)*x[0]**o+(-0.551313390457)*x[0]+(0.965127390549)*x[1]**o+(-0.819660095397)*x[1]
        arg[1,1,0]=(-0.656237793952)*x[0]**o+(0.727261124295)*x[0]+(0.602446266985)*x[1]**o+(-0.320796826351)*x[1]
        arg[1,1,1]=(0.111875111713)*x[0]**o+(0.000360626785094)*x[0]+(-0.802516890555)*x[1]**o+(0.420759758492)*x[1]
        arg[2,0,0]=(-0.421920068352)*x[0]**o+(-0.022970702876)*x[0]+(-0.0257219638216)*x[1]**o+(-0.36544576073)*x[1]
        arg[2,0,1]=(-0.459431228769)*x[0]**o+(0.150702477357)*x[0]+(-0.869898945031)*x[1]**o+(0.261651620176)*x[1]
        arg[2,1,0]=(-0.866165026127)*x[0]**o+(0.149307656472)*x[0]+(-0.270058622191)*x[1]**o+(0.0873163553474)*x[1]
        arg[2,1,1]=(-0.871128617394)*x[0]**o+(-0.774356000237)*x[0]+(0.337383167457)*x[1]**o+(-0.453335744484)*x[1]
        arg[3,0,0]=(0.404195477481)*x[0]**o+(-0.350041445198)*x[0]+(-0.583794909281)*x[1]**o+(-0.895439618872)*x[1]
        arg[3,0,1]=(0.693567867164)*x[0]**o+(0.989642722856)*x[0]+(0.814104467522)*x[1]**o+(-0.126302588221)*x[1]
        arg[3,1,0]=(0.935509747673)*x[0]**o+(0.0104612091188)*x[0]+(0.304974305936)*x[1]**o+(0.658660310426)*x[1]
        arg[3,1,1]=(0.677420202064)*x[0]**o+(-0.160834685914)*x[0]+(0.241578146196)*x[1]**o+(-0.182223867353)*x[1]
        arg[4,0,0]=(-0.908542015864)*x[0]**o+(0.125595344482)*x[0]+(-0.396864254557)*x[1]**o+(0.408182849245)*x[1]
        arg[4,0,1]=(0.436101221257)*x[0]**o+(0.946667075656)*x[0]+(-0.293523459017)*x[1]**o+(0.363130416513)*x[1]
        arg[4,1,0]=(0.583416010605)*x[0]**o+(0.531883961319)*x[0]+(-0.418905217485)*x[1]**o+(0.271442034667)*x[1]
        arg[4,1,1]=(-0.708766397827)*x[0]**o+(0.0266251980073)*x[0]+(-0.0929293524384)*x[1]**o+(-0.661095295605)*x[1]
        arg[5,0,0]=(0.693424423443)*x[0]**o+(-0.201539674621)*x[0]+(-0.221922493999)*x[1]**o+(-0.218359043699)*x[1]
        arg[5,0,1]=(-0.478707796293)*x[0]**o+(0.791533578559)*x[0]+(0.280213321002)*x[1]**o+(-0.60809908629)*x[1]
        arg[5,1,0]=(-0.517992572451)*x[0]**o+(0.591995363281)*x[0]+(0.480220427965)*x[1]**o+(0.134542063762)*x[1]
        arg[5,1,1]=(0.415861674391)*x[0]**o+(0.773614372731)*x[0]+(-0.984515861984)*x[1]**o+(0.971916685266)*x[1]
        ref[0,0,0,0]=o*(-0.353390112117)*x_ref[0]**(o-1)+(0.121267964347)
        ref[0,0,0,1]=o*(0.095925646757)*x_ref[1]**(o-1)+(-0.00141503525661)
        ref[0,0,1,0]=o*(0.610556742471)*x_ref[0]**(o-1)+(0.542948343923)
        ref[0,0,1,1]=o*(0.124762431991)*x_ref[1]**(o-1)+(0.00216142978003)
        ref[0,1,0,0]=o*(0.261559018753)*x_ref[0]**(o-1)+(-0.881689419767)
        ref[0,1,0,1]=o*(-0.009352418091)*x_ref[1]**(o-1)+(0.333357441723)
        ref[0,1,1,0]=o*(-0.25379927884)*x_ref[0]**(o-1)+(-0.124076833523)
        ref[0,1,1,1]=o*(-0.167690772835)*x_ref[1]**(o-1)+(0.89804632709)
        ref[1,0,0,0]=o*(-0.39353207502)*x_ref[0]**(o-1)+(0.838815579522)
        ref[1,0,0,1]=o*(0.829467918805)*x_ref[1]**(o-1)+(0.614403367394)
        ref[1,0,1,0]=o*(-0.698946975112)*x_ref[0]**(o-1)+(-0.551313390457)
        ref[1,0,1,1]=o*(0.965127390549)*x_ref[1]**(o-1)+(-0.819660095397)
        ref[1,1,0,0]=o*(-0.656237793952)*x_ref[0]**(o-1)+(0.727261124295)
        ref[1,1,0,1]=o*(0.602446266985)*x_ref[1]**(o-1)+(-0.320796826351)
        ref[1,1,1,0]=o*(0.111875111713)*x_ref[0]**(o-1)+(0.000360626785094)
        ref[1,1,1,1]=o*(-0.802516890555)*x_ref[1]**(o-1)+(0.420759758492)
        ref[2,0,0,0]=o*(-0.421920068352)*x_ref[0]**(o-1)+(-0.022970702876)
        ref[2,0,0,1]=o*(-0.0257219638216)*x_ref[1]**(o-1)+(-0.36544576073)
        ref[2,0,1,0]=o*(-0.459431228769)*x_ref[0]**(o-1)+(0.150702477357)
        ref[2,0,1,1]=o*(-0.869898945031)*x_ref[1]**(o-1)+(0.261651620176)
        ref[2,1,0,0]=o*(-0.866165026127)*x_ref[0]**(o-1)+(0.149307656472)
        ref[2,1,0,1]=o*(-0.270058622191)*x_ref[1]**(o-1)+(0.0873163553474)
        ref[2,1,1,0]=o*(-0.871128617394)*x_ref[0]**(o-1)+(-0.774356000237)
        ref[2,1,1,1]=o*(0.337383167457)*x_ref[1]**(o-1)+(-0.453335744484)
        ref[3,0,0,0]=o*(0.404195477481)*x_ref[0]**(o-1)+(-0.350041445198)
        ref[3,0,0,1]=o*(-0.583794909281)*x_ref[1]**(o-1)+(-0.895439618872)
        ref[3,0,1,0]=o*(0.693567867164)*x_ref[0]**(o-1)+(0.989642722856)
        ref[3,0,1,1]=o*(0.814104467522)*x_ref[1]**(o-1)+(-0.126302588221)
        ref[3,1,0,0]=o*(0.935509747673)*x_ref[0]**(o-1)+(0.0104612091188)
        ref[3,1,0,1]=o*(0.304974305936)*x_ref[1]**(o-1)+(0.658660310426)
        ref[3,1,1,0]=o*(0.677420202064)*x_ref[0]**(o-1)+(-0.160834685914)
        ref[3,1,1,1]=o*(0.241578146196)*x_ref[1]**(o-1)+(-0.182223867353)
        ref[4,0,0,0]=o*(-0.908542015864)*x_ref[0]**(o-1)+(0.125595344482)
        ref[4,0,0,1]=o*(-0.396864254557)*x_ref[1]**(o-1)+(0.408182849245)
        ref[4,0,1,0]=o*(0.436101221257)*x_ref[0]**(o-1)+(0.946667075656)
        ref[4,0,1,1]=o*(-0.293523459017)*x_ref[1]**(o-1)+(0.363130416513)
        ref[4,1,0,0]=o*(0.583416010605)*x_ref[0]**(o-1)+(0.531883961319)
        ref[4,1,0,1]=o*(-0.418905217485)*x_ref[1]**(o-1)+(0.271442034667)
        ref[4,1,1,0]=o*(-0.708766397827)*x_ref[0]**(o-1)+(0.0266251980073)
        ref[4,1,1,1]=o*(-0.0929293524384)*x_ref[1]**(o-1)+(-0.661095295605)
        ref[5,0,0,0]=o*(0.693424423443)*x_ref[0]**(o-1)+(-0.201539674621)
        ref[5,0,0,1]=o*(-0.221922493999)*x_ref[1]**(o-1)+(-0.218359043699)
        ref[5,0,1,0]=o*(-0.478707796293)*x_ref[0]**(o-1)+(0.791533578559)
        ref[5,0,1,1]=o*(0.280213321002)*x_ref[1]**(o-1)+(-0.60809908629)
        ref[5,1,0,0]=o*(-0.517992572451)*x_ref[0]**(o-1)+(0.591995363281)
        ref[5,1,0,1]=o*(0.480220427965)*x_ref[1]**(o-1)+(0.134542063762)
        ref[5,1,1,0]=o*(0.415861674391)*x_ref[0]**(o-1)+(0.773614372731)
        ref[5,1,1,1]=o*(-0.984515861984)*x_ref[1]**(o-1)+(0.971916685266)
      else:
        arg[0,0,0]=(0.157678109652)*x[0]**o+(-0.939095069874)*x[0]+(0.340072659801)*x[1]**o+(0.806723712633)*x[1]+(0.0175654201162)*x[2]**o+(-0.695887679405)*x[2]
        arg[0,0,1]=(0.148444957808)*x[0]**o+(0.769780554196)*x[0]+(-0.999778710821)*x[1]**o+(0.765998508134)*x[1]+(0.489498280892)*x[2]**o+(0.272011172233)*x[2]
        arg[0,1,0]=(-0.78853623669)*x[0]**o+(-0.889333747099)*x[0]+(0.897684283327)*x[1]**o+(0.878046029595)*x[1]+(-0.355566201991)*x[2]**o+(-0.656187599294)*x[2]
        arg[0,1,1]=(-0.903496685895)*x[0]**o+(0.335762730702)*x[0]+(0.264330545204)*x[1]**o+(-0.369300150725)*x[1]+(0.00819339769162)*x[2]**o+(0.0466274018566)*x[2]
        arg[1,0,0]=(0.490893502666)*x[0]**o+(0.853253169019)*x[0]+(-0.706120536551)*x[1]**o+(-0.810665092749)*x[1]+(0.877124690496)*x[2]**o+(-0.663316848595)*x[2]
        arg[1,0,1]=(-0.897301110118)*x[0]**o+(-0.571577796347)*x[0]+(-0.00912216314091)*x[1]**o+(0.479091266837)*x[1]+(-0.951994896347)*x[2]**o+(0.293801256264)*x[2]
        arg[1,1,0]=(-0.0843442242335)*x[0]**o+(0.436592099426)*x[0]+(-0.961835284742)*x[1]**o+(0.453758903497)*x[1]+(0.759916173417)*x[2]**o+(-0.356674191439)*x[2]
        arg[1,1,1]=(0.368337095889)*x[0]**o+(-0.180552461333)*x[0]+(-0.814727889141)*x[1]**o+(-0.226678543654)*x[1]+(0.629233877979)*x[2]**o+(0.934452743806)*x[2]
        arg[2,0,0]=(0.335801632975)*x[0]**o+(0.4853793666)*x[0]+(-0.603075878381)*x[1]**o+(-0.464873655094)*x[1]+(-0.304107114487)*x[2]**o+(0.768564020927)*x[2]
        arg[2,0,1]=(-0.843867260693)*x[0]**o+(-0.758493294188)*x[0]+(0.801820428023)*x[1]**o+(0.0909753752871)*x[1]+(-0.208134327922)*x[2]**o+(0.343325582627)*x[2]
        arg[2,1,0]=(-0.143364423799)*x[0]**o+(-0.191201452642)*x[0]+(0.356480498629)*x[1]**o+(0.0853546566948)*x[1]+(0.234781870759)*x[2]**o+(-0.702906217789)*x[2]
        arg[2,1,1]=(0.81324792898)*x[0]**o+(0.565754643122)*x[0]+(-0.0246314241495)*x[1]**o+(-0.83548496516)*x[1]+(0.730246143504)*x[2]**o+(-0.835539994852)*x[2]
        arg[3,0,0]=(0.877992836393)*x[0]**o+(-0.137135423487)*x[0]+(0.534511286076)*x[1]**o+(-0.0584370047297)*x[1]+(-0.0497236607574)*x[2]**o+(0.408414670632)*x[2]
        arg[3,0,1]=(-0.959771372305)*x[0]**o+(0.587216149992)*x[0]+(-0.873216249139)*x[1]**o+(0.0703350087653)*x[1]+(0.452900901257)*x[2]**o+(-0.180788315472)*x[2]
        arg[3,1,0]=(0.477701526969)*x[0]**o+(0.16165566827)*x[0]+(0.219901428845)*x[1]**o+(-0.0816687269195)*x[1]+(-0.379553348394)*x[2]**o+(-0.764672519612)*x[2]
        arg[3,1,1]=(0.274518799686)*x[0]**o+(0.945009722336)*x[0]+(-0.979787014238)*x[1]**o+(0.587704213162)*x[1]+(0.7959640984)*x[2]**o+(-0.0811112488533)*x[2]
        arg[4,0,0]=(-0.579121056378)*x[0]**o+(0.971799195206)*x[0]+(-0.646271217985)*x[1]**o+(0.369751123548)*x[1]+(0.792425864156)*x[2]**o+(-0.170799235786)*x[2]
        arg[4,0,1]=(0.0939380562355)*x[0]**o+(0.857041359029)*x[0]+(-0.379025882204)*x[1]**o+(-0.571968622949)*x[1]+(-0.156853183697)*x[2]**o+(0.643214485063)*x[2]
        arg[4,1,0]=(-0.715210364871)*x[0]**o+(0.150341666571)*x[0]+(0.325165277323)*x[1]**o+(-0.322744240034)*x[1]+(-0.622094139461)*x[2]**o+(-0.830479146327)*x[2]
        arg[4,1,1]=(-0.697109424406)*x[0]**o+(-0.730318259042)*x[0]+(-0.135119012822)*x[1]**o+(0.410502029229)*x[1]+(0.547987909743)*x[2]**o+(0.144877544254)*x[2]
        arg[5,0,0]=(0.787147597087)*x[0]**o+(-0.528180541295)*x[0]+(-0.251578089108)*x[1]**o+(-0.170766337864)*x[1]+(-0.123341129221)*x[2]**o+(-0.815622069347)*x[2]
        arg[5,0,1]=(0.193873484005)*x[0]**o+(0.270190252915)*x[0]+(0.807946985621)*x[1]**o+(0.184522999461)*x[1]+(0.97429597406)*x[2]**o+(0.532931028021)*x[2]
        arg[5,1,0]=(0.555644676015)*x[0]**o+(-0.225593442089)*x[0]+(0.877869009767)*x[1]**o+(-0.750894245506)*x[1]+(0.888501611997)*x[2]**o+(-0.84027021889)*x[2]
        arg[5,1,1]=(0.76909422867)*x[0]**o+(0.406697645173)*x[0]+(-0.233710263631)*x[1]**o+(-0.432942345121)*x[1]+(0.270484113666)*x[2]**o+(0.707302418115)*x[2]
        ref[0,0,0,0]=o*(0.157678109652)*x_ref[0]**(o-1)+(-0.939095069874)
        ref[0,0,0,1]=o*(0.340072659801)*x_ref[1]**(o-1)+(0.806723712633)
        ref[0,0,0,2]=o*(0.0175654201162)*x_ref[2]**(o-1)+(-0.695887679405)
        ref[0,0,1,0]=o*(0.148444957808)*x_ref[0]**(o-1)+(0.769780554196)
        ref[0,0,1,1]=o*(-0.999778710821)*x_ref[1]**(o-1)+(0.765998508134)
        ref[0,0,1,2]=o*(0.489498280892)*x_ref[2]**(o-1)+(0.272011172233)
        ref[0,1,0,0]=o*(-0.78853623669)*x_ref[0]**(o-1)+(-0.889333747099)
        ref[0,1,0,1]=o*(0.897684283327)*x_ref[1]**(o-1)+(0.878046029595)
        ref[0,1,0,2]=o*(-0.355566201991)*x_ref[2]**(o-1)+(-0.656187599294)
        ref[0,1,1,0]=o*(-0.903496685895)*x_ref[0]**(o-1)+(0.335762730702)
        ref[0,1,1,1]=o*(0.264330545204)*x_ref[1]**(o-1)+(-0.369300150725)
        ref[0,1,1,2]=o*(0.00819339769162)*x_ref[2]**(o-1)+(0.0466274018566)
        ref[1,0,0,0]=o*(0.490893502666)*x_ref[0]**(o-1)+(0.853253169019)
        ref[1,0,0,1]=o*(-0.706120536551)*x_ref[1]**(o-1)+(-0.810665092749)
        ref[1,0,0,2]=o*(0.877124690496)*x_ref[2]**(o-1)+(-0.663316848595)
        ref[1,0,1,0]=o*(-0.897301110118)*x_ref[0]**(o-1)+(-0.571577796347)
        ref[1,0,1,1]=o*(-0.00912216314091)*x_ref[1]**(o-1)+(0.479091266837)
        ref[1,0,1,2]=o*(-0.951994896347)*x_ref[2]**(o-1)+(0.293801256264)
        ref[1,1,0,0]=o*(-0.0843442242335)*x_ref[0]**(o-1)+(0.436592099426)
        ref[1,1,0,1]=o*(-0.961835284742)*x_ref[1]**(o-1)+(0.453758903497)
        ref[1,1,0,2]=o*(0.759916173417)*x_ref[2]**(o-1)+(-0.356674191439)
        ref[1,1,1,0]=o*(0.368337095889)*x_ref[0]**(o-1)+(-0.180552461333)
        ref[1,1,1,1]=o*(-0.814727889141)*x_ref[1]**(o-1)+(-0.226678543654)
        ref[1,1,1,2]=o*(0.629233877979)*x_ref[2]**(o-1)+(0.934452743806)
        ref[2,0,0,0]=o*(0.335801632975)*x_ref[0]**(o-1)+(0.4853793666)
        ref[2,0,0,1]=o*(-0.603075878381)*x_ref[1]**(o-1)+(-0.464873655094)
        ref[2,0,0,2]=o*(-0.304107114487)*x_ref[2]**(o-1)+(0.768564020927)
        ref[2,0,1,0]=o*(-0.843867260693)*x_ref[0]**(o-1)+(-0.758493294188)
        ref[2,0,1,1]=o*(0.801820428023)*x_ref[1]**(o-1)+(0.0909753752871)
        ref[2,0,1,2]=o*(-0.208134327922)*x_ref[2]**(o-1)+(0.343325582627)
        ref[2,1,0,0]=o*(-0.143364423799)*x_ref[0]**(o-1)+(-0.191201452642)
        ref[2,1,0,1]=o*(0.356480498629)*x_ref[1]**(o-1)+(0.0853546566948)
        ref[2,1,0,2]=o*(0.234781870759)*x_ref[2]**(o-1)+(-0.702906217789)
        ref[2,1,1,0]=o*(0.81324792898)*x_ref[0]**(o-1)+(0.565754643122)
        ref[2,1,1,1]=o*(-0.0246314241495)*x_ref[1]**(o-1)+(-0.83548496516)
        ref[2,1,1,2]=o*(0.730246143504)*x_ref[2]**(o-1)+(-0.835539994852)
        ref[3,0,0,0]=o*(0.877992836393)*x_ref[0]**(o-1)+(-0.137135423487)
        ref[3,0,0,1]=o*(0.534511286076)*x_ref[1]**(o-1)+(-0.0584370047297)
        ref[3,0,0,2]=o*(-0.0497236607574)*x_ref[2]**(o-1)+(0.408414670632)
        ref[3,0,1,0]=o*(-0.959771372305)*x_ref[0]**(o-1)+(0.587216149992)
        ref[3,0,1,1]=o*(-0.873216249139)*x_ref[1]**(o-1)+(0.0703350087653)
        ref[3,0,1,2]=o*(0.452900901257)*x_ref[2]**(o-1)+(-0.180788315472)
        ref[3,1,0,0]=o*(0.477701526969)*x_ref[0]**(o-1)+(0.16165566827)
        ref[3,1,0,1]=o*(0.219901428845)*x_ref[1]**(o-1)+(-0.0816687269195)
        ref[3,1,0,2]=o*(-0.379553348394)*x_ref[2]**(o-1)+(-0.764672519612)
        ref[3,1,1,0]=o*(0.274518799686)*x_ref[0]**(o-1)+(0.945009722336)
        ref[3,1,1,1]=o*(-0.979787014238)*x_ref[1]**(o-1)+(0.587704213162)
        ref[3,1,1,2]=o*(0.7959640984)*x_ref[2]**(o-1)+(-0.0811112488533)
        ref[4,0,0,0]=o*(-0.579121056378)*x_ref[0]**(o-1)+(0.971799195206)
        ref[4,0,0,1]=o*(-0.646271217985)*x_ref[1]**(o-1)+(0.369751123548)
        ref[4,0,0,2]=o*(0.792425864156)*x_ref[2]**(o-1)+(-0.170799235786)
        ref[4,0,1,0]=o*(0.0939380562355)*x_ref[0]**(o-1)+(0.857041359029)
        ref[4,0,1,1]=o*(-0.379025882204)*x_ref[1]**(o-1)+(-0.571968622949)
        ref[4,0,1,2]=o*(-0.156853183697)*x_ref[2]**(o-1)+(0.643214485063)
        ref[4,1,0,0]=o*(-0.715210364871)*x_ref[0]**(o-1)+(0.150341666571)
        ref[4,1,0,1]=o*(0.325165277323)*x_ref[1]**(o-1)+(-0.322744240034)
        ref[4,1,0,2]=o*(-0.622094139461)*x_ref[2]**(o-1)+(-0.830479146327)
        ref[4,1,1,0]=o*(-0.697109424406)*x_ref[0]**(o-1)+(-0.730318259042)
        ref[4,1,1,1]=o*(-0.135119012822)*x_ref[1]**(o-1)+(0.410502029229)
        ref[4,1,1,2]=o*(0.547987909743)*x_ref[2]**(o-1)+(0.144877544254)
        ref[5,0,0,0]=o*(0.787147597087)*x_ref[0]**(o-1)+(-0.528180541295)
        ref[5,0,0,1]=o*(-0.251578089108)*x_ref[1]**(o-1)+(-0.170766337864)
        ref[5,0,0,2]=o*(-0.123341129221)*x_ref[2]**(o-1)+(-0.815622069347)
        ref[5,0,1,0]=o*(0.193873484005)*x_ref[0]**(o-1)+(0.270190252915)
        ref[5,0,1,1]=o*(0.807946985621)*x_ref[1]**(o-1)+(0.184522999461)
        ref[5,0,1,2]=o*(0.97429597406)*x_ref[2]**(o-1)+(0.532931028021)
        ref[5,1,0,0]=o*(0.555644676015)*x_ref[0]**(o-1)+(-0.225593442089)
        ref[5,1,0,1]=o*(0.877869009767)*x_ref[1]**(o-1)+(-0.750894245506)
        ref[5,1,0,2]=o*(0.888501611997)*x_ref[2]**(o-1)+(-0.84027021889)
        ref[5,1,1,0]=o*(0.76909422867)*x_ref[0]**(o-1)+(0.406697645173)
        ref[5,1,1,1]=o*(-0.233710263631)*x_ref[1]**(o-1)+(-0.432942345121)
        ref[5,1,1,2]=o*(0.270484113666)*x_ref[2]**(o-1)+(0.707302418115)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.260967575068)*x[0]**o+(-0.456559015376)*x[0]+(0.663023090785)*x[1]**o+(-0.000383723402374)*x[1]
        ref[0]=o*(0.260967575068)*x_ref[0]**(o-1)+(-0.456559015376)
        ref[1]=o*(0.663023090785)*x_ref[1]**(o-1)+(-0.000383723402374)
      else:
        arg=(-0.679453642968)*x[0]**o+(-0.836558889087)*x[0]+(0.938200435255)*x[1]**o+(-0.368159735907)*x[1]+(-0.0546850112561)*x[2]**o+(0.931986983035)*x[2]
        ref[0]=o*(-0.679453642968)*x_ref[0]**(o-1)+(-0.836558889087)
        ref[1]=o*(0.938200435255)*x_ref[1]**(o-1)+(-0.368159735907)
        ref[2]=o*(-0.0546850112561)*x_ref[2]**(o-1)+(0.931986983035)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.221278607378)*x[0]**o+(-0.25051444588)*x[0]+(0.752161895582)*x[1]**o+(0.662434618125)*x[1]
        arg[1]=(0.984829669888)*x[0]**o+(0.00500258390598)*x[0]+(-0.292523259369)*x[1]**o+(-0.238277884466)*x[1]
        ref[0,0]=o*(0.221278607378)*x_ref[0]**(o-1)+(-0.25051444588)
        ref[0,1]=o*(0.752161895582)*x_ref[1]**(o-1)+(0.662434618125)
        ref[1,0]=o*(0.984829669888)*x_ref[0]**(o-1)+(0.00500258390598)
        ref[1,1]=o*(-0.292523259369)*x_ref[1]**(o-1)+(-0.238277884466)
      else:
        arg[0]=(0.815886350601)*x[0]**o+(-0.41721938077)*x[0]+(0.478358181839)*x[1]**o+(-0.458860964588)*x[1]+(0.575706872631)*x[2]**o+(-0.965426161177)*x[2]
        arg[1]=(-0.159203003073)*x[0]**o+(-0.985356409774)*x[0]+(0.504125935966)*x[1]**o+(0.703505637664)*x[1]+(-0.480016288837)*x[2]**o+(0.544400209204)*x[2]
        ref[0,0]=o*(0.815886350601)*x_ref[0]**(o-1)+(-0.41721938077)
        ref[0,1]=o*(0.478358181839)*x_ref[1]**(o-1)+(-0.458860964588)
        ref[0,2]=o*(0.575706872631)*x_ref[2]**(o-1)+(-0.965426161177)
        ref[1,0]=o*(-0.159203003073)*x_ref[0]**(o-1)+(-0.985356409774)
        ref[1,1]=o*(0.504125935966)*x_ref[1]**(o-1)+(0.703505637664)
        ref[1,2]=o*(-0.480016288837)*x_ref[2]**(o-1)+(0.544400209204)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(-0.722004066753)*x[0]**o+(-0.911439550186)*x[0]+(0.0793361579222)*x[1]**o+(0.958526722957)*x[1]
        arg[0,1]=(-0.0237328054252)*x[0]**o+(0.832621528642)*x[0]+(0.168484408262)*x[1]**o+(-0.181293214021)*x[1]
        arg[0,2]=(0.0284992392733)*x[0]**o+(0.884411090784)*x[0]+(-0.615596355135)*x[1]**o+(-0.564819619851)*x[1]
        arg[0,3]=(0.37408647134)*x[0]**o+(-0.177901551674)*x[0]+(-0.466165048356)*x[1]**o+(-0.278511961588)*x[1]
        arg[0,4]=(-0.24695435531)*x[0]**o+(-0.624082916034)*x[0]+(-0.343849543401)*x[1]**o+(0.661610450445)*x[1]
        arg[1,0]=(-0.176271330569)*x[0]**o+(-0.661612282712)*x[0]+(-0.937163166613)*x[1]**o+(0.376473179587)*x[1]
        arg[1,1]=(0.964111839461)*x[0]**o+(0.576837069403)*x[0]+(-0.718562476684)*x[1]**o+(0.154883495525)*x[1]
        arg[1,2]=(-0.771556809734)*x[0]**o+(0.720054023732)*x[0]+(-0.390732462938)*x[1]**o+(-0.484359019049)*x[1]
        arg[1,3]=(0.565910084656)*x[0]**o+(0.69105192657)*x[0]+(-0.449408955892)*x[1]**o+(-0.789780518242)*x[1]
        arg[1,4]=(0.602485034177)*x[0]**o+(-0.345139282323)*x[0]+(-0.951334152921)*x[1]**o+(-0.478700081579)*x[1]
        arg[2,0]=(0.00114729937901)*x[0]**o+(0.295356090995)*x[0]+(0.982261544035)*x[1]**o+(0.357076141518)*x[1]
        arg[2,1]=(0.653029330145)*x[0]**o+(-0.0178804068723)*x[0]+(-0.0238303862422)*x[1]**o+(0.983033975535)*x[1]
        arg[2,2]=(0.680878878444)*x[0]**o+(-0.404330541218)*x[0]+(0.856544883866)*x[1]**o+(0.71899071821)*x[1]
        arg[2,3]=(0.0674687202072)*x[0]**o+(-0.748990717637)*x[0]+(0.185597242309)*x[1]**o+(0.623849127358)*x[1]
        arg[2,4]=(0.302869699915)*x[0]**o+(-0.852621552847)*x[0]+(0.115557112172)*x[1]**o+(-0.0827214208715)*x[1]
        arg[3,0]=(0.0929072570643)*x[0]**o+(0.533152382923)*x[0]+(-0.0913754984708)*x[1]**o+(-0.25865969337)*x[1]
        arg[3,1]=(-0.794799886368)*x[0]**o+(-0.865263962776)*x[0]+(0.441343356932)*x[1]**o+(-0.541975483415)*x[1]
        arg[3,2]=(0.261341948686)*x[0]**o+(0.378142729621)*x[0]+(0.0556054089673)*x[1]**o+(-0.854195169318)*x[1]
        arg[3,3]=(0.177222467454)*x[0]**o+(-0.400820197499)*x[0]+(0.88513075927)*x[1]**o+(0.0420082325044)*x[1]
        arg[3,4]=(0.307178422145)*x[0]**o+(-0.203765349758)*x[0]+(-0.546052990937)*x[1]**o+(-0.960312136292)*x[1]
        ref[0,0,0]=o*(-0.722004066753)*x_ref[0]**(o-1)+(-0.911439550186)
        ref[0,0,1]=o*(0.0793361579222)*x_ref[1]**(o-1)+(0.958526722957)
        ref[0,1,0]=o*(-0.0237328054252)*x_ref[0]**(o-1)+(0.832621528642)
        ref[0,1,1]=o*(0.168484408262)*x_ref[1]**(o-1)+(-0.181293214021)
        ref[0,2,0]=o*(0.0284992392733)*x_ref[0]**(o-1)+(0.884411090784)
        ref[0,2,1]=o*(-0.615596355135)*x_ref[1]**(o-1)+(-0.564819619851)
        ref[0,3,0]=o*(0.37408647134)*x_ref[0]**(o-1)+(-0.177901551674)
        ref[0,3,1]=o*(-0.466165048356)*x_ref[1]**(o-1)+(-0.278511961588)
        ref[0,4,0]=o*(-0.24695435531)*x_ref[0]**(o-1)+(-0.624082916034)
        ref[0,4,1]=o*(-0.343849543401)*x_ref[1]**(o-1)+(0.661610450445)
        ref[1,0,0]=o*(-0.176271330569)*x_ref[0]**(o-1)+(-0.661612282712)
        ref[1,0,1]=o*(-0.937163166613)*x_ref[1]**(o-1)+(0.376473179587)
        ref[1,1,0]=o*(0.964111839461)*x_ref[0]**(o-1)+(0.576837069403)
        ref[1,1,1]=o*(-0.718562476684)*x_ref[1]**(o-1)+(0.154883495525)
        ref[1,2,0]=o*(-0.771556809734)*x_ref[0]**(o-1)+(0.720054023732)
        ref[1,2,1]=o*(-0.390732462938)*x_ref[1]**(o-1)+(-0.484359019049)
        ref[1,3,0]=o*(0.565910084656)*x_ref[0]**(o-1)+(0.69105192657)
        ref[1,3,1]=o*(-0.449408955892)*x_ref[1]**(o-1)+(-0.789780518242)
        ref[1,4,0]=o*(0.602485034177)*x_ref[0]**(o-1)+(-0.345139282323)
        ref[1,4,1]=o*(-0.951334152921)*x_ref[1]**(o-1)+(-0.478700081579)
        ref[2,0,0]=o*(0.00114729937901)*x_ref[0]**(o-1)+(0.295356090995)
        ref[2,0,1]=o*(0.982261544035)*x_ref[1]**(o-1)+(0.357076141518)
        ref[2,1,0]=o*(0.653029330145)*x_ref[0]**(o-1)+(-0.0178804068723)
        ref[2,1,1]=o*(-0.0238303862422)*x_ref[1]**(o-1)+(0.983033975535)
        ref[2,2,0]=o*(0.680878878444)*x_ref[0]**(o-1)+(-0.404330541218)
        ref[2,2,1]=o*(0.856544883866)*x_ref[1]**(o-1)+(0.71899071821)
        ref[2,3,0]=o*(0.0674687202072)*x_ref[0]**(o-1)+(-0.748990717637)
        ref[2,3,1]=o*(0.185597242309)*x_ref[1]**(o-1)+(0.623849127358)
        ref[2,4,0]=o*(0.302869699915)*x_ref[0]**(o-1)+(-0.852621552847)
        ref[2,4,1]=o*(0.115557112172)*x_ref[1]**(o-1)+(-0.0827214208715)
        ref[3,0,0]=o*(0.0929072570643)*x_ref[0]**(o-1)+(0.533152382923)
        ref[3,0,1]=o*(-0.0913754984708)*x_ref[1]**(o-1)+(-0.25865969337)
        ref[3,1,0]=o*(-0.794799886368)*x_ref[0]**(o-1)+(-0.865263962776)
        ref[3,1,1]=o*(0.441343356932)*x_ref[1]**(o-1)+(-0.541975483415)
        ref[3,2,0]=o*(0.261341948686)*x_ref[0]**(o-1)+(0.378142729621)
        ref[3,2,1]=o*(0.0556054089673)*x_ref[1]**(o-1)+(-0.854195169318)
        ref[3,3,0]=o*(0.177222467454)*x_ref[0]**(o-1)+(-0.400820197499)
        ref[3,3,1]=o*(0.88513075927)*x_ref[1]**(o-1)+(0.0420082325044)
        ref[3,4,0]=o*(0.307178422145)*x_ref[0]**(o-1)+(-0.203765349758)
        ref[3,4,1]=o*(-0.546052990937)*x_ref[1]**(o-1)+(-0.960312136292)
      else:
        arg[0,0]=(0.767891828291)*x[0]**o+(-0.0840162180616)*x[0]+(-0.942354055114)*x[1]**o+(-0.458324281794)*x[1]+(-0.837589052489)*x[2]**o+(-0.49741065871)*x[2]
        arg[0,1]=(-0.770574732815)*x[0]**o+(0.949362675967)*x[0]+(0.334062389906)*x[1]**o+(0.589134582043)*x[1]+(0.611876443107)*x[2]**o+(-0.476746595186)*x[2]
        arg[0,2]=(-0.113076234857)*x[0]**o+(0.207116555982)*x[0]+(-0.434423505296)*x[1]**o+(0.582451788291)*x[1]+(-0.858451242875)*x[2]**o+(-0.229934112115)*x[2]
        arg[0,3]=(0.55494281628)*x[0]**o+(0.53116989281)*x[0]+(0.150549353151)*x[1]**o+(0.70886785918)*x[1]+(-0.148436740612)*x[2]**o+(0.755584812901)*x[2]
        arg[0,4]=(0.0855991607862)*x[0]**o+(0.509120320151)*x[0]+(-0.303374431746)*x[1]**o+(0.1813545791)*x[1]+(-0.0434667822132)*x[2]**o+(0.395447849194)*x[2]
        arg[1,0]=(-0.779720617706)*x[0]**o+(0.730668258583)*x[0]+(-0.0315884698035)*x[1]**o+(-0.0929845369431)*x[1]+(0.0853332239352)*x[2]**o+(0.303372304789)*x[2]
        arg[1,1]=(-0.687021672626)*x[0]**o+(-0.380139153436)*x[0]+(-0.624774641851)*x[1]**o+(0.331833510141)*x[1]+(-0.747115143343)*x[2]**o+(-0.901384102505)*x[2]
        arg[1,2]=(-0.0528543338921)*x[0]**o+(-0.683764012224)*x[0]+(0.498886779776)*x[1]**o+(-0.396777215862)*x[1]+(0.777051418657)*x[2]**o+(-0.786345142641)*x[2]
        arg[1,3]=(-0.823112017253)*x[0]**o+(0.278818264102)*x[0]+(-0.79372160065)*x[1]**o+(0.254044599609)*x[1]+(-0.372925254827)*x[2]**o+(-0.85054008508)*x[2]
        arg[1,4]=(-0.247156845703)*x[0]**o+(-0.944335229412)*x[0]+(-0.572532409179)*x[1]**o+(0.289091364421)*x[1]+(0.848874739954)*x[2]**o+(0.705098999827)*x[2]
        arg[2,0]=(0.0958493700534)*x[0]**o+(-0.334070579338)*x[0]+(-0.74117408785)*x[1]**o+(0.0609646344894)*x[1]+(-0.328888690751)*x[2]**o+(-0.0430772626841)*x[2]
        arg[2,1]=(0.846871401827)*x[0]**o+(0.379912211333)*x[0]+(-0.197881866966)*x[1]**o+(-0.551498190852)*x[1]+(0.416977164384)*x[2]**o+(0.09398324743)*x[2]
        arg[2,2]=(0.737261473925)*x[0]**o+(0.696028328637)*x[0]+(-0.735110324145)*x[1]**o+(0.949453963809)*x[1]+(-0.894722167055)*x[2]**o+(-0.16124057159)*x[2]
        arg[2,3]=(0.267556059297)*x[0]**o+(-0.633581895464)*x[0]+(-0.176232065927)*x[1]**o+(0.98533106984)*x[1]+(0.404671434387)*x[2]**o+(0.343560595562)*x[2]
        arg[2,4]=(0.502981833324)*x[0]**o+(0.716659442696)*x[0]+(-0.349106283426)*x[1]**o+(0.0726954534053)*x[1]+(0.468516366697)*x[2]**o+(0.328926164383)*x[2]
        arg[3,0]=(0.846509624251)*x[0]**o+(-0.34161660121)*x[0]+(0.25149102353)*x[1]**o+(-0.0130496471853)*x[1]+(-0.556826239055)*x[2]**o+(0.774658667066)*x[2]
        arg[3,1]=(-0.0777319662414)*x[0]**o+(0.992622365536)*x[0]+(0.720504365851)*x[1]**o+(0.0701085839568)*x[1]+(0.475573965193)*x[2]**o+(0.402599619829)*x[2]
        arg[3,2]=(-0.311228599853)*x[0]**o+(-0.732684576916)*x[0]+(0.308213192232)*x[1]**o+(-0.131325871896)*x[1]+(0.964631653839)*x[2]**o+(-0.600646614776)*x[2]
        arg[3,3]=(0.656589254036)*x[0]**o+(-0.275449493207)*x[0]+(-0.611075965309)*x[1]**o+(-0.460879628412)*x[1]+(-0.820861932527)*x[2]**o+(0.424066222721)*x[2]
        arg[3,4]=(0.303269333663)*x[0]**o+(-0.745774722275)*x[0]+(0.82933539351)*x[1]**o+(-0.295464064484)*x[1]+(0.818135297647)*x[2]**o+(-0.989600307642)*x[2]
        ref[0,0,0]=o*(0.767891828291)*x_ref[0]**(o-1)+(-0.0840162180616)
        ref[0,0,1]=o*(-0.942354055114)*x_ref[1]**(o-1)+(-0.458324281794)
        ref[0,0,2]=o*(-0.837589052489)*x_ref[2]**(o-1)+(-0.49741065871)
        ref[0,1,0]=o*(-0.770574732815)*x_ref[0]**(o-1)+(0.949362675967)
        ref[0,1,1]=o*(0.334062389906)*x_ref[1]**(o-1)+(0.589134582043)
        ref[0,1,2]=o*(0.611876443107)*x_ref[2]**(o-1)+(-0.476746595186)
        ref[0,2,0]=o*(-0.113076234857)*x_ref[0]**(o-1)+(0.207116555982)
        ref[0,2,1]=o*(-0.434423505296)*x_ref[1]**(o-1)+(0.582451788291)
        ref[0,2,2]=o*(-0.858451242875)*x_ref[2]**(o-1)+(-0.229934112115)
        ref[0,3,0]=o*(0.55494281628)*x_ref[0]**(o-1)+(0.53116989281)
        ref[0,3,1]=o*(0.150549353151)*x_ref[1]**(o-1)+(0.70886785918)
        ref[0,3,2]=o*(-0.148436740612)*x_ref[2]**(o-1)+(0.755584812901)
        ref[0,4,0]=o*(0.0855991607862)*x_ref[0]**(o-1)+(0.509120320151)
        ref[0,4,1]=o*(-0.303374431746)*x_ref[1]**(o-1)+(0.1813545791)
        ref[0,4,2]=o*(-0.0434667822132)*x_ref[2]**(o-1)+(0.395447849194)
        ref[1,0,0]=o*(-0.779720617706)*x_ref[0]**(o-1)+(0.730668258583)
        ref[1,0,1]=o*(-0.0315884698035)*x_ref[1]**(o-1)+(-0.0929845369431)
        ref[1,0,2]=o*(0.0853332239352)*x_ref[2]**(o-1)+(0.303372304789)
        ref[1,1,0]=o*(-0.687021672626)*x_ref[0]**(o-1)+(-0.380139153436)
        ref[1,1,1]=o*(-0.624774641851)*x_ref[1]**(o-1)+(0.331833510141)
        ref[1,1,2]=o*(-0.747115143343)*x_ref[2]**(o-1)+(-0.901384102505)
        ref[1,2,0]=o*(-0.0528543338921)*x_ref[0]**(o-1)+(-0.683764012224)
        ref[1,2,1]=o*(0.498886779776)*x_ref[1]**(o-1)+(-0.396777215862)
        ref[1,2,2]=o*(0.777051418657)*x_ref[2]**(o-1)+(-0.786345142641)
        ref[1,3,0]=o*(-0.823112017253)*x_ref[0]**(o-1)+(0.278818264102)
        ref[1,3,1]=o*(-0.79372160065)*x_ref[1]**(o-1)+(0.254044599609)
        ref[1,3,2]=o*(-0.372925254827)*x_ref[2]**(o-1)+(-0.85054008508)
        ref[1,4,0]=o*(-0.247156845703)*x_ref[0]**(o-1)+(-0.944335229412)
        ref[1,4,1]=o*(-0.572532409179)*x_ref[1]**(o-1)+(0.289091364421)
        ref[1,4,2]=o*(0.848874739954)*x_ref[2]**(o-1)+(0.705098999827)
        ref[2,0,0]=o*(0.0958493700534)*x_ref[0]**(o-1)+(-0.334070579338)
        ref[2,0,1]=o*(-0.74117408785)*x_ref[1]**(o-1)+(0.0609646344894)
        ref[2,0,2]=o*(-0.328888690751)*x_ref[2]**(o-1)+(-0.0430772626841)
        ref[2,1,0]=o*(0.846871401827)*x_ref[0]**(o-1)+(0.379912211333)
        ref[2,1,1]=o*(-0.197881866966)*x_ref[1]**(o-1)+(-0.551498190852)
        ref[2,1,2]=o*(0.416977164384)*x_ref[2]**(o-1)+(0.09398324743)
        ref[2,2,0]=o*(0.737261473925)*x_ref[0]**(o-1)+(0.696028328637)
        ref[2,2,1]=o*(-0.735110324145)*x_ref[1]**(o-1)+(0.949453963809)
        ref[2,2,2]=o*(-0.894722167055)*x_ref[2]**(o-1)+(-0.16124057159)
        ref[2,3,0]=o*(0.267556059297)*x_ref[0]**(o-1)+(-0.633581895464)
        ref[2,3,1]=o*(-0.176232065927)*x_ref[1]**(o-1)+(0.98533106984)
        ref[2,3,2]=o*(0.404671434387)*x_ref[2]**(o-1)+(0.343560595562)
        ref[2,4,0]=o*(0.502981833324)*x_ref[0]**(o-1)+(0.716659442696)
        ref[2,4,1]=o*(-0.349106283426)*x_ref[1]**(o-1)+(0.0726954534053)
        ref[2,4,2]=o*(0.468516366697)*x_ref[2]**(o-1)+(0.328926164383)
        ref[3,0,0]=o*(0.846509624251)*x_ref[0]**(o-1)+(-0.34161660121)
        ref[3,0,1]=o*(0.25149102353)*x_ref[1]**(o-1)+(-0.0130496471853)
        ref[3,0,2]=o*(-0.556826239055)*x_ref[2]**(o-1)+(0.774658667066)
        ref[3,1,0]=o*(-0.0777319662414)*x_ref[0]**(o-1)+(0.992622365536)
        ref[3,1,1]=o*(0.720504365851)*x_ref[1]**(o-1)+(0.0701085839568)
        ref[3,1,2]=o*(0.475573965193)*x_ref[2]**(o-1)+(0.402599619829)
        ref[3,2,0]=o*(-0.311228599853)*x_ref[0]**(o-1)+(-0.732684576916)
        ref[3,2,1]=o*(0.308213192232)*x_ref[1]**(o-1)+(-0.131325871896)
        ref[3,2,2]=o*(0.964631653839)*x_ref[2]**(o-1)+(-0.600646614776)
        ref[3,3,0]=o*(0.656589254036)*x_ref[0]**(o-1)+(-0.275449493207)
        ref[3,3,1]=o*(-0.611075965309)*x_ref[1]**(o-1)+(-0.460879628412)
        ref[3,3,2]=o*(-0.820861932527)*x_ref[2]**(o-1)+(0.424066222721)
        ref[3,4,0]=o*(0.303269333663)*x_ref[0]**(o-1)+(-0.745774722275)
        ref[3,4,1]=o*(0.82933539351)*x_ref[1]**(o-1)+(-0.295464064484)
        ref[3,4,2]=o*(0.818135297647)*x_ref[2]**(o-1)+(-0.989600307642)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.572693829631)*x[0]**o+(0.583556749477)*x[0]+(0.22594627003)*x[1]**o+(0.0199162626119)*x[1]
        arg[0,0,1]=(0.196266833125)*x[0]**o+(0.544824498113)*x[0]+(0.150736772157)*x[1]**o+(0.362174902716)*x[1]
        arg[0,1,0]=(0.169481539273)*x[0]**o+(0.0355504217105)*x[0]+(-0.156746013218)*x[1]**o+(0.879476257954)*x[1]
        arg[0,1,1]=(0.824826637612)*x[0]**o+(-0.732645177858)*x[0]+(0.732873648361)*x[1]**o+(-0.40090515696)*x[1]
        arg[1,0,0]=(0.77348720999)*x[0]**o+(0.229475702878)*x[0]+(0.686324947328)*x[1]**o+(0.313197971717)*x[1]
        arg[1,0,1]=(-0.966626441904)*x[0]**o+(0.989778282376)*x[0]+(0.0447214020611)*x[1]**o+(0.439337525343)*x[1]
        arg[1,1,0]=(-0.774806567769)*x[0]**o+(0.634591112025)*x[0]+(0.797720133135)*x[1]**o+(0.579137230562)*x[1]
        arg[1,1,1]=(-0.118703583742)*x[0]**o+(0.482822282284)*x[0]+(0.141839968056)*x[1]**o+(0.916615496002)*x[1]
        arg[2,0,0]=(0.7419378407)*x[0]**o+(0.754978933397)*x[0]+(-0.839586800033)*x[1]**o+(-0.525830008753)*x[1]
        arg[2,0,1]=(-0.855040380309)*x[0]**o+(0.423661210447)*x[0]+(0.538277902172)*x[1]**o+(0.127617006384)*x[1]
        arg[2,1,0]=(0.107349578938)*x[0]**o+(0.697738123598)*x[0]+(0.862106802027)*x[1]**o+(-0.829827651082)*x[1]
        arg[2,1,1]=(-0.970208311818)*x[0]**o+(0.806698181512)*x[0]+(-0.374007524144)*x[1]**o+(-0.0600501198736)*x[1]
        arg[3,0,0]=(0.223312188411)*x[0]**o+(-0.130776319489)*x[0]+(0.616021313401)*x[1]**o+(-0.552848902138)*x[1]
        arg[3,0,1]=(-0.461837450406)*x[0]**o+(-0.895912538242)*x[0]+(0.477606324718)*x[1]**o+(0.220148409572)*x[1]
        arg[3,1,0]=(0.0783005249695)*x[0]**o+(-0.762350707337)*x[0]+(0.616470467018)*x[1]**o+(-0.154358859634)*x[1]
        arg[3,1,1]=(-0.604945998091)*x[0]**o+(0.23801834237)*x[0]+(0.500666300529)*x[1]**o+(0.709482270038)*x[1]
        arg[4,0,0]=(0.746627054515)*x[0]**o+(0.550282853502)*x[0]+(-0.223527434147)*x[1]**o+(-0.291559486198)*x[1]
        arg[4,0,1]=(0.512462457431)*x[0]**o+(0.0908594735188)*x[0]+(-0.888862229845)*x[1]**o+(-0.969985508695)*x[1]
        arg[4,1,0]=(0.606213701819)*x[0]**o+(0.147049558356)*x[0]+(-0.0365908849884)*x[1]**o+(-0.584510320466)*x[1]
        arg[4,1,1]=(-0.110180223906)*x[0]**o+(-0.83442554862)*x[0]+(-0.0516097061459)*x[1]**o+(-0.155615857347)*x[1]
        arg[5,0,0]=(0.603753207781)*x[0]**o+(0.302032475639)*x[0]+(-0.68719165993)*x[1]**o+(-0.375829665239)*x[1]
        arg[5,0,1]=(-0.599357977805)*x[0]**o+(0.562486062075)*x[0]+(0.102714751356)*x[1]**o+(-0.114179799185)*x[1]
        arg[5,1,0]=(0.874756501256)*x[0]**o+(0.550416007969)*x[0]+(0.669721605367)*x[1]**o+(0.24372600222)*x[1]
        arg[5,1,1]=(0.651356889701)*x[0]**o+(0.178366196323)*x[0]+(0.762268544189)*x[1]**o+(0.515014711815)*x[1]
        ref[0,0,0,0]=o*(-0.572693829631)*x_ref[0]**(o-1)+(0.583556749477)
        ref[0,0,0,1]=o*(0.22594627003)*x_ref[1]**(o-1)+(0.0199162626119)
        ref[0,0,1,0]=o*(0.196266833125)*x_ref[0]**(o-1)+(0.544824498113)
        ref[0,0,1,1]=o*(0.150736772157)*x_ref[1]**(o-1)+(0.362174902716)
        ref[0,1,0,0]=o*(0.169481539273)*x_ref[0]**(o-1)+(0.0355504217105)
        ref[0,1,0,1]=o*(-0.156746013218)*x_ref[1]**(o-1)+(0.879476257954)
        ref[0,1,1,0]=o*(0.824826637612)*x_ref[0]**(o-1)+(-0.732645177858)
        ref[0,1,1,1]=o*(0.732873648361)*x_ref[1]**(o-1)+(-0.40090515696)
        ref[1,0,0,0]=o*(0.77348720999)*x_ref[0]**(o-1)+(0.229475702878)
        ref[1,0,0,1]=o*(0.686324947328)*x_ref[1]**(o-1)+(0.313197971717)
        ref[1,0,1,0]=o*(-0.966626441904)*x_ref[0]**(o-1)+(0.989778282376)
        ref[1,0,1,1]=o*(0.0447214020611)*x_ref[1]**(o-1)+(0.439337525343)
        ref[1,1,0,0]=o*(-0.774806567769)*x_ref[0]**(o-1)+(0.634591112025)
        ref[1,1,0,1]=o*(0.797720133135)*x_ref[1]**(o-1)+(0.579137230562)
        ref[1,1,1,0]=o*(-0.118703583742)*x_ref[0]**(o-1)+(0.482822282284)
        ref[1,1,1,1]=o*(0.141839968056)*x_ref[1]**(o-1)+(0.916615496002)
        ref[2,0,0,0]=o*(0.7419378407)*x_ref[0]**(o-1)+(0.754978933397)
        ref[2,0,0,1]=o*(-0.839586800033)*x_ref[1]**(o-1)+(-0.525830008753)
        ref[2,0,1,0]=o*(-0.855040380309)*x_ref[0]**(o-1)+(0.423661210447)
        ref[2,0,1,1]=o*(0.538277902172)*x_ref[1]**(o-1)+(0.127617006384)
        ref[2,1,0,0]=o*(0.107349578938)*x_ref[0]**(o-1)+(0.697738123598)
        ref[2,1,0,1]=o*(0.862106802027)*x_ref[1]**(o-1)+(-0.829827651082)
        ref[2,1,1,0]=o*(-0.970208311818)*x_ref[0]**(o-1)+(0.806698181512)
        ref[2,1,1,1]=o*(-0.374007524144)*x_ref[1]**(o-1)+(-0.0600501198736)
        ref[3,0,0,0]=o*(0.223312188411)*x_ref[0]**(o-1)+(-0.130776319489)
        ref[3,0,0,1]=o*(0.616021313401)*x_ref[1]**(o-1)+(-0.552848902138)
        ref[3,0,1,0]=o*(-0.461837450406)*x_ref[0]**(o-1)+(-0.895912538242)
        ref[3,0,1,1]=o*(0.477606324718)*x_ref[1]**(o-1)+(0.220148409572)
        ref[3,1,0,0]=o*(0.0783005249695)*x_ref[0]**(o-1)+(-0.762350707337)
        ref[3,1,0,1]=o*(0.616470467018)*x_ref[1]**(o-1)+(-0.154358859634)
        ref[3,1,1,0]=o*(-0.604945998091)*x_ref[0]**(o-1)+(0.23801834237)
        ref[3,1,1,1]=o*(0.500666300529)*x_ref[1]**(o-1)+(0.709482270038)
        ref[4,0,0,0]=o*(0.746627054515)*x_ref[0]**(o-1)+(0.550282853502)
        ref[4,0,0,1]=o*(-0.223527434147)*x_ref[1]**(o-1)+(-0.291559486198)
        ref[4,0,1,0]=o*(0.512462457431)*x_ref[0]**(o-1)+(0.0908594735188)
        ref[4,0,1,1]=o*(-0.888862229845)*x_ref[1]**(o-1)+(-0.969985508695)
        ref[4,1,0,0]=o*(0.606213701819)*x_ref[0]**(o-1)+(0.147049558356)
        ref[4,1,0,1]=o*(-0.0365908849884)*x_ref[1]**(o-1)+(-0.584510320466)
        ref[4,1,1,0]=o*(-0.110180223906)*x_ref[0]**(o-1)+(-0.83442554862)
        ref[4,1,1,1]=o*(-0.0516097061459)*x_ref[1]**(o-1)+(-0.155615857347)
        ref[5,0,0,0]=o*(0.603753207781)*x_ref[0]**(o-1)+(0.302032475639)
        ref[5,0,0,1]=o*(-0.68719165993)*x_ref[1]**(o-1)+(-0.375829665239)
        ref[5,0,1,0]=o*(-0.599357977805)*x_ref[0]**(o-1)+(0.562486062075)
        ref[5,0,1,1]=o*(0.102714751356)*x_ref[1]**(o-1)+(-0.114179799185)
        ref[5,1,0,0]=o*(0.874756501256)*x_ref[0]**(o-1)+(0.550416007969)
        ref[5,1,0,1]=o*(0.669721605367)*x_ref[1]**(o-1)+(0.24372600222)
        ref[5,1,1,0]=o*(0.651356889701)*x_ref[0]**(o-1)+(0.178366196323)
        ref[5,1,1,1]=o*(0.762268544189)*x_ref[1]**(o-1)+(0.515014711815)
      else:
        arg[0,0,0]=(-0.634739027866)*x[0]**o+(-0.696370463772)*x[0]+(0.503815931458)*x[1]**o+(-0.457824888032)*x[1]+(0.655074800235)*x[2]**o+(-0.456111009206)*x[2]
        arg[0,0,1]=(-0.533710147141)*x[0]**o+(-0.0814085290298)*x[0]+(-0.0894565935545)*x[1]**o+(-0.3775655102)*x[1]+(-0.134051544262)*x[2]**o+(0.000713749886013)*x[2]
        arg[0,1,0]=(-0.64394535332)*x[0]**o+(0.430741797171)*x[0]+(0.985157009863)*x[1]**o+(-0.655331266432)*x[1]+(-0.436629135988)*x[2]**o+(-0.404480211702)*x[2]
        arg[0,1,1]=(-0.0858143551662)*x[0]**o+(0.0452769288327)*x[0]+(0.930452744895)*x[1]**o+(-0.877149214078)*x[1]+(0.675472222413)*x[2]**o+(0.917497654616)*x[2]
        arg[1,0,0]=(0.88346309442)*x[0]**o+(0.892116796676)*x[0]+(0.22765643526)*x[1]**o+(-0.934442833938)*x[1]+(0.0781755340006)*x[2]**o+(0.106009847768)*x[2]
        arg[1,0,1]=(0.20824563298)*x[0]**o+(0.613584602599)*x[0]+(-0.482386438919)*x[1]**o+(-0.155840561301)*x[1]+(0.817890842096)*x[2]**o+(0.798519843541)*x[2]
        arg[1,1,0]=(-0.915853569389)*x[0]**o+(0.573041720657)*x[0]+(0.236967108991)*x[1]**o+(0.421805849528)*x[1]+(-0.576407564364)*x[2]**o+(0.0709741076428)*x[2]
        arg[1,1,1]=(0.530417188741)*x[0]**o+(0.872073272277)*x[0]+(-0.21047376763)*x[1]**o+(0.88490064655)*x[1]+(0.523085602451)*x[2]**o+(-0.926133382905)*x[2]
        arg[2,0,0]=(-0.974186019136)*x[0]**o+(-0.332433414417)*x[0]+(-0.500938330608)*x[1]**o+(-0.113188574497)*x[1]+(0.228656301416)*x[2]**o+(-0.375756425313)*x[2]
        arg[2,0,1]=(0.246057421673)*x[0]**o+(0.231587929687)*x[0]+(0.398413883486)*x[1]**o+(-0.390998026525)*x[1]+(0.273401149062)*x[2]**o+(-0.296142647578)*x[2]
        arg[2,1,0]=(0.283717204372)*x[0]**o+(-0.545852417255)*x[0]+(-0.616345461169)*x[1]**o+(-0.705051061202)*x[1]+(0.664377876346)*x[2]**o+(0.338382503272)*x[2]
        arg[2,1,1]=(-0.433297970157)*x[0]**o+(-0.716600154712)*x[0]+(0.836194932687)*x[1]**o+(0.0903372116119)*x[1]+(0.113374499353)*x[2]**o+(-0.588810627425)*x[2]
        arg[3,0,0]=(0.404388255474)*x[0]**o+(0.521816615546)*x[0]+(-0.0976581170689)*x[1]**o+(-0.678563989037)*x[1]+(0.550953471125)*x[2]**o+(-0.668793844235)*x[2]
        arg[3,0,1]=(-0.66023047868)*x[0]**o+(0.822794930168)*x[0]+(-0.284215677966)*x[1]**o+(-0.779237447318)*x[1]+(-0.191727752719)*x[2]**o+(-0.748596484262)*x[2]
        arg[3,1,0]=(-0.0610807571905)*x[0]**o+(0.044623781518)*x[0]+(0.851226664753)*x[1]**o+(0.921090244094)*x[1]+(0.494414571241)*x[2]**o+(-0.473831333684)*x[2]
        arg[3,1,1]=(0.54651807228)*x[0]**o+(-0.580525990778)*x[0]+(-0.822936606301)*x[1]**o+(-0.0169562878937)*x[1]+(-0.732219249555)*x[2]**o+(-0.435156155008)*x[2]
        arg[4,0,0]=(0.830021032139)*x[0]**o+(0.976836986062)*x[0]+(0.534162405111)*x[1]**o+(-0.286292103539)*x[1]+(0.820276894761)*x[2]**o+(0.245897387706)*x[2]
        arg[4,0,1]=(0.362201493311)*x[0]**o+(-0.186328147108)*x[0]+(-0.91611458572)*x[1]**o+(-0.965258904175)*x[1]+(0.61850292372)*x[2]**o+(0.410229835478)*x[2]
        arg[4,1,0]=(0.0726313013515)*x[0]**o+(-0.974714544914)*x[0]+(0.46313543268)*x[1]**o+(0.100486704072)*x[1]+(-0.149124857233)*x[2]**o+(0.455733236337)*x[2]
        arg[4,1,1]=(-0.875220220201)*x[0]**o+(-0.416152301789)*x[0]+(-0.0145627466423)*x[1]**o+(0.267856009119)*x[1]+(-0.604260174049)*x[2]**o+(0.689107997051)*x[2]
        arg[5,0,0]=(0.196182127971)*x[0]**o+(0.337209503833)*x[0]+(0.595342239158)*x[1]**o+(0.507423300857)*x[1]+(0.934760469389)*x[2]**o+(-0.198686443328)*x[2]
        arg[5,0,1]=(0.77210909967)*x[0]**o+(-0.310778136545)*x[0]+(0.512326103142)*x[1]**o+(-0.501056206243)*x[1]+(0.98755961257)*x[2]**o+(-0.658414340038)*x[2]
        arg[5,1,0]=(0.375277702198)*x[0]**o+(-0.603850047717)*x[0]+(0.677227868069)*x[1]**o+(-0.784640615159)*x[1]+(0.177350721639)*x[2]**o+(0.241053657333)*x[2]
        arg[5,1,1]=(-0.368625522736)*x[0]**o+(0.196977476909)*x[0]+(-0.582451887206)*x[1]**o+(0.598240904158)*x[1]+(0.102240230796)*x[2]**o+(0.487322529158)*x[2]
        ref[0,0,0,0]=o*(-0.634739027866)*x_ref[0]**(o-1)+(-0.696370463772)
        ref[0,0,0,1]=o*(0.503815931458)*x_ref[1]**(o-1)+(-0.457824888032)
        ref[0,0,0,2]=o*(0.655074800235)*x_ref[2]**(o-1)+(-0.456111009206)
        ref[0,0,1,0]=o*(-0.533710147141)*x_ref[0]**(o-1)+(-0.0814085290298)
        ref[0,0,1,1]=o*(-0.0894565935545)*x_ref[1]**(o-1)+(-0.3775655102)
        ref[0,0,1,2]=o*(-0.134051544262)*x_ref[2]**(o-1)+(0.000713749886013)
        ref[0,1,0,0]=o*(-0.64394535332)*x_ref[0]**(o-1)+(0.430741797171)
        ref[0,1,0,1]=o*(0.985157009863)*x_ref[1]**(o-1)+(-0.655331266432)
        ref[0,1,0,2]=o*(-0.436629135988)*x_ref[2]**(o-1)+(-0.404480211702)
        ref[0,1,1,0]=o*(-0.0858143551662)*x_ref[0]**(o-1)+(0.0452769288327)
        ref[0,1,1,1]=o*(0.930452744895)*x_ref[1]**(o-1)+(-0.877149214078)
        ref[0,1,1,2]=o*(0.675472222413)*x_ref[2]**(o-1)+(0.917497654616)
        ref[1,0,0,0]=o*(0.88346309442)*x_ref[0]**(o-1)+(0.892116796676)
        ref[1,0,0,1]=o*(0.22765643526)*x_ref[1]**(o-1)+(-0.934442833938)
        ref[1,0,0,2]=o*(0.0781755340006)*x_ref[2]**(o-1)+(0.106009847768)
        ref[1,0,1,0]=o*(0.20824563298)*x_ref[0]**(o-1)+(0.613584602599)
        ref[1,0,1,1]=o*(-0.482386438919)*x_ref[1]**(o-1)+(-0.155840561301)
        ref[1,0,1,2]=o*(0.817890842096)*x_ref[2]**(o-1)+(0.798519843541)
        ref[1,1,0,0]=o*(-0.915853569389)*x_ref[0]**(o-1)+(0.573041720657)
        ref[1,1,0,1]=o*(0.236967108991)*x_ref[1]**(o-1)+(0.421805849528)
        ref[1,1,0,2]=o*(-0.576407564364)*x_ref[2]**(o-1)+(0.0709741076428)
        ref[1,1,1,0]=o*(0.530417188741)*x_ref[0]**(o-1)+(0.872073272277)
        ref[1,1,1,1]=o*(-0.21047376763)*x_ref[1]**(o-1)+(0.88490064655)
        ref[1,1,1,2]=o*(0.523085602451)*x_ref[2]**(o-1)+(-0.926133382905)
        ref[2,0,0,0]=o*(-0.974186019136)*x_ref[0]**(o-1)+(-0.332433414417)
        ref[2,0,0,1]=o*(-0.500938330608)*x_ref[1]**(o-1)+(-0.113188574497)
        ref[2,0,0,2]=o*(0.228656301416)*x_ref[2]**(o-1)+(-0.375756425313)
        ref[2,0,1,0]=o*(0.246057421673)*x_ref[0]**(o-1)+(0.231587929687)
        ref[2,0,1,1]=o*(0.398413883486)*x_ref[1]**(o-1)+(-0.390998026525)
        ref[2,0,1,2]=o*(0.273401149062)*x_ref[2]**(o-1)+(-0.296142647578)
        ref[2,1,0,0]=o*(0.283717204372)*x_ref[0]**(o-1)+(-0.545852417255)
        ref[2,1,0,1]=o*(-0.616345461169)*x_ref[1]**(o-1)+(-0.705051061202)
        ref[2,1,0,2]=o*(0.664377876346)*x_ref[2]**(o-1)+(0.338382503272)
        ref[2,1,1,0]=o*(-0.433297970157)*x_ref[0]**(o-1)+(-0.716600154712)
        ref[2,1,1,1]=o*(0.836194932687)*x_ref[1]**(o-1)+(0.0903372116119)
        ref[2,1,1,2]=o*(0.113374499353)*x_ref[2]**(o-1)+(-0.588810627425)
        ref[3,0,0,0]=o*(0.404388255474)*x_ref[0]**(o-1)+(0.521816615546)
        ref[3,0,0,1]=o*(-0.0976581170689)*x_ref[1]**(o-1)+(-0.678563989037)
        ref[3,0,0,2]=o*(0.550953471125)*x_ref[2]**(o-1)+(-0.668793844235)
        ref[3,0,1,0]=o*(-0.66023047868)*x_ref[0]**(o-1)+(0.822794930168)
        ref[3,0,1,1]=o*(-0.284215677966)*x_ref[1]**(o-1)+(-0.779237447318)
        ref[3,0,1,2]=o*(-0.191727752719)*x_ref[2]**(o-1)+(-0.748596484262)
        ref[3,1,0,0]=o*(-0.0610807571905)*x_ref[0]**(o-1)+(0.044623781518)
        ref[3,1,0,1]=o*(0.851226664753)*x_ref[1]**(o-1)+(0.921090244094)
        ref[3,1,0,2]=o*(0.494414571241)*x_ref[2]**(o-1)+(-0.473831333684)
        ref[3,1,1,0]=o*(0.54651807228)*x_ref[0]**(o-1)+(-0.580525990778)
        ref[3,1,1,1]=o*(-0.822936606301)*x_ref[1]**(o-1)+(-0.0169562878937)
        ref[3,1,1,2]=o*(-0.732219249555)*x_ref[2]**(o-1)+(-0.435156155008)
        ref[4,0,0,0]=o*(0.830021032139)*x_ref[0]**(o-1)+(0.976836986062)
        ref[4,0,0,1]=o*(0.534162405111)*x_ref[1]**(o-1)+(-0.286292103539)
        ref[4,0,0,2]=o*(0.820276894761)*x_ref[2]**(o-1)+(0.245897387706)
        ref[4,0,1,0]=o*(0.362201493311)*x_ref[0]**(o-1)+(-0.186328147108)
        ref[4,0,1,1]=o*(-0.91611458572)*x_ref[1]**(o-1)+(-0.965258904175)
        ref[4,0,1,2]=o*(0.61850292372)*x_ref[2]**(o-1)+(0.410229835478)
        ref[4,1,0,0]=o*(0.0726313013515)*x_ref[0]**(o-1)+(-0.974714544914)
        ref[4,1,0,1]=o*(0.46313543268)*x_ref[1]**(o-1)+(0.100486704072)
        ref[4,1,0,2]=o*(-0.149124857233)*x_ref[2]**(o-1)+(0.455733236337)
        ref[4,1,1,0]=o*(-0.875220220201)*x_ref[0]**(o-1)+(-0.416152301789)
        ref[4,1,1,1]=o*(-0.0145627466423)*x_ref[1]**(o-1)+(0.267856009119)
        ref[4,1,1,2]=o*(-0.604260174049)*x_ref[2]**(o-1)+(0.689107997051)
        ref[5,0,0,0]=o*(0.196182127971)*x_ref[0]**(o-1)+(0.337209503833)
        ref[5,0,0,1]=o*(0.595342239158)*x_ref[1]**(o-1)+(0.507423300857)
        ref[5,0,0,2]=o*(0.934760469389)*x_ref[2]**(o-1)+(-0.198686443328)
        ref[5,0,1,0]=o*(0.77210909967)*x_ref[0]**(o-1)+(-0.310778136545)
        ref[5,0,1,1]=o*(0.512326103142)*x_ref[1]**(o-1)+(-0.501056206243)
        ref[5,0,1,2]=o*(0.98755961257)*x_ref[2]**(o-1)+(-0.658414340038)
        ref[5,1,0,0]=o*(0.375277702198)*x_ref[0]**(o-1)+(-0.603850047717)
        ref[5,1,0,1]=o*(0.677227868069)*x_ref[1]**(o-1)+(-0.784640615159)
        ref[5,1,0,2]=o*(0.177350721639)*x_ref[2]**(o-1)+(0.241053657333)
        ref[5,1,1,0]=o*(-0.368625522736)*x_ref[0]**(o-1)+(0.196977476909)
        ref[5,1,1,1]=o*(-0.582451887206)*x_ref[1]**(o-1)+(0.598240904158)
        ref[5,1,1,2]=o*(0.102240230796)*x_ref[2]**(o-1)+(0.487322529158)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.851643614423)*x[0]**o+(-0.77638754265)*x[0]+(-0.216746360457)*x[1]**o+(-0.117807364973)*x[1]
        ref[0]=o*(0.851643614423)*x_ref[0]**(o-1)+(-0.77638754265)
        ref[1]=o*(-0.216746360457)*x_ref[1]**(o-1)+(-0.117807364973)
      else:
        arg=(0.846271449213)*x[0]**o+(0.895578416904)*x[0]+(-0.518297490388)*x[1]**o+(0.272255265951)*x[1]+(0.537822992528)*x[2]**o+(0.74226977002)*x[2]
        ref[0]=o*(0.846271449213)*x_ref[0]**(o-1)+(0.895578416904)
        ref[1]=o*(-0.518297490388)*x_ref[1]**(o-1)+(0.272255265951)
        ref[2]=o*(0.537822992528)*x_ref[2]**(o-1)+(0.74226977002)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.365601318395)*x[0]**o+(0.295624555802)*x[0]+(0.540618182217)*x[1]**o+(0.0458416078386)*x[1]
        arg[1]=(0.332276039998)*x[0]**o+(0.422606936253)*x[0]+(-0.221603594854)*x[1]**o+(0.879863420258)*x[1]
        ref[0,0]=o*(0.365601318395)*x_ref[0]**(o-1)+(0.295624555802)
        ref[0,1]=o*(0.540618182217)*x_ref[1]**(o-1)+(0.0458416078386)
        ref[1,0]=o*(0.332276039998)*x_ref[0]**(o-1)+(0.422606936253)
        ref[1,1]=o*(-0.221603594854)*x_ref[1]**(o-1)+(0.879863420258)
      else:
        arg[0]=(-0.258072277129)*x[0]**o+(-0.833085462518)*x[0]+(-0.124240965858)*x[1]**o+(-0.121853073047)*x[1]+(-0.877974471569)*x[2]**o+(0.929660451331)*x[2]
        arg[1]=(0.306520203535)*x[0]**o+(0.635444320723)*x[0]+(-0.766579509915)*x[1]**o+(-0.381189007175)*x[1]+(-0.76608952351)*x[2]**o+(0.00718583364881)*x[2]
        ref[0,0]=o*(-0.258072277129)*x_ref[0]**(o-1)+(-0.833085462518)
        ref[0,1]=o*(-0.124240965858)*x_ref[1]**(o-1)+(-0.121853073047)
        ref[0,2]=o*(-0.877974471569)*x_ref[2]**(o-1)+(0.929660451331)
        ref[1,0]=o*(0.306520203535)*x_ref[0]**(o-1)+(0.635444320723)
        ref[1,1]=o*(-0.766579509915)*x_ref[1]**(o-1)+(-0.381189007175)
        ref[1,2]=o*(-0.76608952351)*x_ref[2]**(o-1)+(0.00718583364881)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.403088506806)*x[0]**o+(-0.436942773988)*x[0]+(-0.0594011537346)*x[1]**o+(0.855063195595)*x[1]
        arg[0,1]=(-0.516992303892)*x[0]**o+(-0.132586737299)*x[0]+(-0.854639239618)*x[1]**o+(0.241200454455)*x[1]
        arg[0,2]=(-0.373941155498)*x[0]**o+(0.933328225549)*x[0]+(-0.986870460316)*x[1]**o+(-0.051265908773)*x[1]
        arg[0,3]=(-0.175701653328)*x[0]**o+(-0.726542005356)*x[0]+(-0.553505635409)*x[1]**o+(-0.199806969718)*x[1]
        arg[0,4]=(-0.750832127973)*x[0]**o+(0.404511387161)*x[0]+(0.874423404897)*x[1]**o+(0.35360129734)*x[1]
        arg[1,0]=(-0.000191271001285)*x[0]**o+(-0.319365399066)*x[0]+(0.398444976089)*x[1]**o+(-0.966391658006)*x[1]
        arg[1,1]=(-0.58562890965)*x[0]**o+(-0.623992315262)*x[0]+(-0.913704774126)*x[1]**o+(0.910305409081)*x[1]
        arg[1,2]=(-0.480491011401)*x[0]**o+(-0.387015926056)*x[0]+(0.845826744282)*x[1]**o+(0.380626980034)*x[1]
        arg[1,3]=(-0.284167706438)*x[0]**o+(0.730574128045)*x[0]+(0.823389953051)*x[1]**o+(0.59483032515)*x[1]
        arg[1,4]=(-0.545147810354)*x[0]**o+(-0.632923743497)*x[0]+(0.619596244096)*x[1]**o+(-0.816596365827)*x[1]
        arg[2,0]=(0.161289895106)*x[0]**o+(0.435472177094)*x[0]+(0.125591455076)*x[1]**o+(0.0207320895937)*x[1]
        arg[2,1]=(0.639457164801)*x[0]**o+(0.2631408826)*x[0]+(0.00255095677899)*x[1]**o+(0.732164720004)*x[1]
        arg[2,2]=(-0.152014092979)*x[0]**o+(0.347024615036)*x[0]+(0.94212294017)*x[1]**o+(0.173939724203)*x[1]
        arg[2,3]=(-0.0324872553975)*x[0]**o+(-0.969492827039)*x[0]+(-0.830876162378)*x[1]**o+(0.601308513636)*x[1]
        arg[2,4]=(0.932999309515)*x[0]**o+(0.0320937041249)*x[0]+(-0.354492894455)*x[1]**o+(-0.752837146132)*x[1]
        arg[3,0]=(0.995147840382)*x[0]**o+(-0.276974278934)*x[0]+(0.905272259272)*x[1]**o+(-0.433178159448)*x[1]
        arg[3,1]=(-0.291708633096)*x[0]**o+(-0.542000555721)*x[0]+(0.372404748364)*x[1]**o+(0.23140215774)*x[1]
        arg[3,2]=(0.88604771071)*x[0]**o+(-0.753078356803)*x[0]+(-0.265302521749)*x[1]**o+(-0.948898850214)*x[1]
        arg[3,3]=(-0.429496844082)*x[0]**o+(0.0731695917195)*x[0]+(0.291387878783)*x[1]**o+(-0.078863825039)*x[1]
        arg[3,4]=(-0.0067999635984)*x[0]**o+(0.488018496019)*x[0]+(-0.875646054486)*x[1]**o+(0.629803284064)*x[1]
        ref[0,0,0]=o*(0.403088506806)*x_ref[0]**(o-1)+(-0.436942773988)
        ref[0,0,1]=o*(-0.0594011537346)*x_ref[1]**(o-1)+(0.855063195595)
        ref[0,1,0]=o*(-0.516992303892)*x_ref[0]**(o-1)+(-0.132586737299)
        ref[0,1,1]=o*(-0.854639239618)*x_ref[1]**(o-1)+(0.241200454455)
        ref[0,2,0]=o*(-0.373941155498)*x_ref[0]**(o-1)+(0.933328225549)
        ref[0,2,1]=o*(-0.986870460316)*x_ref[1]**(o-1)+(-0.051265908773)
        ref[0,3,0]=o*(-0.175701653328)*x_ref[0]**(o-1)+(-0.726542005356)
        ref[0,3,1]=o*(-0.553505635409)*x_ref[1]**(o-1)+(-0.199806969718)
        ref[0,4,0]=o*(-0.750832127973)*x_ref[0]**(o-1)+(0.404511387161)
        ref[0,4,1]=o*(0.874423404897)*x_ref[1]**(o-1)+(0.35360129734)
        ref[1,0,0]=o*(-0.000191271001285)*x_ref[0]**(o-1)+(-0.319365399066)
        ref[1,0,1]=o*(0.398444976089)*x_ref[1]**(o-1)+(-0.966391658006)
        ref[1,1,0]=o*(-0.58562890965)*x_ref[0]**(o-1)+(-0.623992315262)
        ref[1,1,1]=o*(-0.913704774126)*x_ref[1]**(o-1)+(0.910305409081)
        ref[1,2,0]=o*(-0.480491011401)*x_ref[0]**(o-1)+(-0.387015926056)
        ref[1,2,1]=o*(0.845826744282)*x_ref[1]**(o-1)+(0.380626980034)
        ref[1,3,0]=o*(-0.284167706438)*x_ref[0]**(o-1)+(0.730574128045)
        ref[1,3,1]=o*(0.823389953051)*x_ref[1]**(o-1)+(0.59483032515)
        ref[1,4,0]=o*(-0.545147810354)*x_ref[0]**(o-1)+(-0.632923743497)
        ref[1,4,1]=o*(0.619596244096)*x_ref[1]**(o-1)+(-0.816596365827)
        ref[2,0,0]=o*(0.161289895106)*x_ref[0]**(o-1)+(0.435472177094)
        ref[2,0,1]=o*(0.125591455076)*x_ref[1]**(o-1)+(0.0207320895937)
        ref[2,1,0]=o*(0.639457164801)*x_ref[0]**(o-1)+(0.2631408826)
        ref[2,1,1]=o*(0.00255095677899)*x_ref[1]**(o-1)+(0.732164720004)
        ref[2,2,0]=o*(-0.152014092979)*x_ref[0]**(o-1)+(0.347024615036)
        ref[2,2,1]=o*(0.94212294017)*x_ref[1]**(o-1)+(0.173939724203)
        ref[2,3,0]=o*(-0.0324872553975)*x_ref[0]**(o-1)+(-0.969492827039)
        ref[2,3,1]=o*(-0.830876162378)*x_ref[1]**(o-1)+(0.601308513636)
        ref[2,4,0]=o*(0.932999309515)*x_ref[0]**(o-1)+(0.0320937041249)
        ref[2,4,1]=o*(-0.354492894455)*x_ref[1]**(o-1)+(-0.752837146132)
        ref[3,0,0]=o*(0.995147840382)*x_ref[0]**(o-1)+(-0.276974278934)
        ref[3,0,1]=o*(0.905272259272)*x_ref[1]**(o-1)+(-0.433178159448)
        ref[3,1,0]=o*(-0.291708633096)*x_ref[0]**(o-1)+(-0.542000555721)
        ref[3,1,1]=o*(0.372404748364)*x_ref[1]**(o-1)+(0.23140215774)
        ref[3,2,0]=o*(0.88604771071)*x_ref[0]**(o-1)+(-0.753078356803)
        ref[3,2,1]=o*(-0.265302521749)*x_ref[1]**(o-1)+(-0.948898850214)
        ref[3,3,0]=o*(-0.429496844082)*x_ref[0]**(o-1)+(0.0731695917195)
        ref[3,3,1]=o*(0.291387878783)*x_ref[1]**(o-1)+(-0.078863825039)
        ref[3,4,0]=o*(-0.0067999635984)*x_ref[0]**(o-1)+(0.488018496019)
        ref[3,4,1]=o*(-0.875646054486)*x_ref[1]**(o-1)+(0.629803284064)
      else:
        arg[0,0]=(-0.512668741406)*x[0]**o+(-0.73142104842)*x[0]+(0.489050993226)*x[1]**o+(0.771941571419)*x[1]+(-0.538092512577)*x[2]**o+(-0.457812389328)*x[2]
        arg[0,1]=(0.258566436386)*x[0]**o+(-0.572006860118)*x[0]+(-0.0370305840623)*x[1]**o+(0.328810934717)*x[1]+(-0.652873123265)*x[2]**o+(0.734929914374)*x[2]
        arg[0,2]=(-0.228462027625)*x[0]**o+(-0.873359484139)*x[0]+(0.902575701641)*x[1]**o+(0.428758160478)*x[1]+(-0.377042847702)*x[2]**o+(-0.161364604019)*x[2]
        arg[0,3]=(0.31338185054)*x[0]**o+(0.973849539932)*x[0]+(-0.403569231167)*x[1]**o+(-0.110026269257)*x[1]+(0.603101392345)*x[2]**o+(0.499232034173)*x[2]
        arg[0,4]=(0.468416317974)*x[0]**o+(-0.0154203909264)*x[0]+(0.409228888268)*x[1]**o+(0.358886480877)*x[1]+(-0.00677343401301)*x[2]**o+(-0.0567153195302)*x[2]
        arg[1,0]=(0.614929795628)*x[0]**o+(0.54911679279)*x[0]+(0.527138394555)*x[1]**o+(-0.84070465171)*x[1]+(0.421928294706)*x[2]**o+(-0.871734463848)*x[2]
        arg[1,1]=(-0.498770327357)*x[0]**o+(0.294610855431)*x[0]+(-0.815129074288)*x[1]**o+(0.395332176796)*x[1]+(-0.555947305527)*x[2]**o+(-0.623550814988)*x[2]
        arg[1,2]=(-0.369404109593)*x[0]**o+(-0.662927702362)*x[0]+(0.886789099892)*x[1]**o+(-0.815728539359)*x[1]+(-0.843861162574)*x[2]**o+(-0.510039607856)*x[2]
        arg[1,3]=(-0.747305678574)*x[0]**o+(0.350337618157)*x[0]+(0.313491051893)*x[1]**o+(-0.554840960712)*x[1]+(-0.259760094369)*x[2]**o+(0.904246223768)*x[2]
        arg[1,4]=(0.485859115418)*x[0]**o+(-0.751900240375)*x[0]+(-0.233384727142)*x[1]**o+(-0.466677891749)*x[1]+(-0.771311310708)*x[2]**o+(0.380515497846)*x[2]
        arg[2,0]=(0.628529887016)*x[0]**o+(0.778810609149)*x[0]+(-0.95647127374)*x[1]**o+(-0.887895309984)*x[1]+(-0.926017554164)*x[2]**o+(0.0922106889932)*x[2]
        arg[2,1]=(-0.981382750453)*x[0]**o+(0.858805367823)*x[0]+(-0.202352133378)*x[1]**o+(0.527229765084)*x[1]+(-0.0862956097426)*x[2]**o+(-0.76269735579)*x[2]
        arg[2,2]=(0.591276459941)*x[0]**o+(0.554086306469)*x[0]+(-0.175045397917)*x[1]**o+(0.319542375346)*x[1]+(0.328810955576)*x[2]**o+(-0.245381809232)*x[2]
        arg[2,3]=(-0.826436985287)*x[0]**o+(-0.931352535769)*x[0]+(-0.480801956568)*x[1]**o+(0.830773509938)*x[1]+(0.157556801982)*x[2]**o+(0.982505166946)*x[2]
        arg[2,4]=(-0.0992736644463)*x[0]**o+(-0.147525710611)*x[0]+(-0.723322157973)*x[1]**o+(0.0527438662093)*x[1]+(-0.338847227522)*x[2]**o+(-0.704585609897)*x[2]
        arg[3,0]=(-0.45250460769)*x[0]**o+(-0.323299292565)*x[0]+(-0.435540041421)*x[1]**o+(0.418330181239)*x[1]+(0.526739853977)*x[2]**o+(-0.00513414974056)*x[2]
        arg[3,1]=(0.289534600598)*x[0]**o+(0.467498987256)*x[0]+(0.27851903721)*x[1]**o+(0.908249187588)*x[1]+(-0.84695489717)*x[2]**o+(-0.799372710877)*x[2]
        arg[3,2]=(0.00338011092566)*x[0]**o+(0.819216072868)*x[0]+(0.115437140937)*x[1]**o+(0.0968835984948)*x[1]+(-0.633253222581)*x[2]**o+(-0.0224890801282)*x[2]
        arg[3,3]=(0.578331574145)*x[0]**o+(0.328247456626)*x[0]+(-0.725914955131)*x[1]**o+(-0.863562766028)*x[1]+(-0.283057959979)*x[2]**o+(-0.521563360583)*x[2]
        arg[3,4]=(0.345725352698)*x[0]**o+(0.614912618718)*x[0]+(-0.218084952915)*x[1]**o+(-0.629514485943)*x[1]+(-0.207605980168)*x[2]**o+(-0.0950042458608)*x[2]
        ref[0,0,0]=o*(-0.512668741406)*x_ref[0]**(o-1)+(-0.73142104842)
        ref[0,0,1]=o*(0.489050993226)*x_ref[1]**(o-1)+(0.771941571419)
        ref[0,0,2]=o*(-0.538092512577)*x_ref[2]**(o-1)+(-0.457812389328)
        ref[0,1,0]=o*(0.258566436386)*x_ref[0]**(o-1)+(-0.572006860118)
        ref[0,1,1]=o*(-0.0370305840623)*x_ref[1]**(o-1)+(0.328810934717)
        ref[0,1,2]=o*(-0.652873123265)*x_ref[2]**(o-1)+(0.734929914374)
        ref[0,2,0]=o*(-0.228462027625)*x_ref[0]**(o-1)+(-0.873359484139)
        ref[0,2,1]=o*(0.902575701641)*x_ref[1]**(o-1)+(0.428758160478)
        ref[0,2,2]=o*(-0.377042847702)*x_ref[2]**(o-1)+(-0.161364604019)
        ref[0,3,0]=o*(0.31338185054)*x_ref[0]**(o-1)+(0.973849539932)
        ref[0,3,1]=o*(-0.403569231167)*x_ref[1]**(o-1)+(-0.110026269257)
        ref[0,3,2]=o*(0.603101392345)*x_ref[2]**(o-1)+(0.499232034173)
        ref[0,4,0]=o*(0.468416317974)*x_ref[0]**(o-1)+(-0.0154203909264)
        ref[0,4,1]=o*(0.409228888268)*x_ref[1]**(o-1)+(0.358886480877)
        ref[0,4,2]=o*(-0.00677343401301)*x_ref[2]**(o-1)+(-0.0567153195302)
        ref[1,0,0]=o*(0.614929795628)*x_ref[0]**(o-1)+(0.54911679279)
        ref[1,0,1]=o*(0.527138394555)*x_ref[1]**(o-1)+(-0.84070465171)
        ref[1,0,2]=o*(0.421928294706)*x_ref[2]**(o-1)+(-0.871734463848)
        ref[1,1,0]=o*(-0.498770327357)*x_ref[0]**(o-1)+(0.294610855431)
        ref[1,1,1]=o*(-0.815129074288)*x_ref[1]**(o-1)+(0.395332176796)
        ref[1,1,2]=o*(-0.555947305527)*x_ref[2]**(o-1)+(-0.623550814988)
        ref[1,2,0]=o*(-0.369404109593)*x_ref[0]**(o-1)+(-0.662927702362)
        ref[1,2,1]=o*(0.886789099892)*x_ref[1]**(o-1)+(-0.815728539359)
        ref[1,2,2]=o*(-0.843861162574)*x_ref[2]**(o-1)+(-0.510039607856)
        ref[1,3,0]=o*(-0.747305678574)*x_ref[0]**(o-1)+(0.350337618157)
        ref[1,3,1]=o*(0.313491051893)*x_ref[1]**(o-1)+(-0.554840960712)
        ref[1,3,2]=o*(-0.259760094369)*x_ref[2]**(o-1)+(0.904246223768)
        ref[1,4,0]=o*(0.485859115418)*x_ref[0]**(o-1)+(-0.751900240375)
        ref[1,4,1]=o*(-0.233384727142)*x_ref[1]**(o-1)+(-0.466677891749)
        ref[1,4,2]=o*(-0.771311310708)*x_ref[2]**(o-1)+(0.380515497846)
        ref[2,0,0]=o*(0.628529887016)*x_ref[0]**(o-1)+(0.778810609149)
        ref[2,0,1]=o*(-0.95647127374)*x_ref[1]**(o-1)+(-0.887895309984)
        ref[2,0,2]=o*(-0.926017554164)*x_ref[2]**(o-1)+(0.0922106889932)
        ref[2,1,0]=o*(-0.981382750453)*x_ref[0]**(o-1)+(0.858805367823)
        ref[2,1,1]=o*(-0.202352133378)*x_ref[1]**(o-1)+(0.527229765084)
        ref[2,1,2]=o*(-0.0862956097426)*x_ref[2]**(o-1)+(-0.76269735579)
        ref[2,2,0]=o*(0.591276459941)*x_ref[0]**(o-1)+(0.554086306469)
        ref[2,2,1]=o*(-0.175045397917)*x_ref[1]**(o-1)+(0.319542375346)
        ref[2,2,2]=o*(0.328810955576)*x_ref[2]**(o-1)+(-0.245381809232)
        ref[2,3,0]=o*(-0.826436985287)*x_ref[0]**(o-1)+(-0.931352535769)
        ref[2,3,1]=o*(-0.480801956568)*x_ref[1]**(o-1)+(0.830773509938)
        ref[2,3,2]=o*(0.157556801982)*x_ref[2]**(o-1)+(0.982505166946)
        ref[2,4,0]=o*(-0.0992736644463)*x_ref[0]**(o-1)+(-0.147525710611)
        ref[2,4,1]=o*(-0.723322157973)*x_ref[1]**(o-1)+(0.0527438662093)
        ref[2,4,2]=o*(-0.338847227522)*x_ref[2]**(o-1)+(-0.704585609897)
        ref[3,0,0]=o*(-0.45250460769)*x_ref[0]**(o-1)+(-0.323299292565)
        ref[3,0,1]=o*(-0.435540041421)*x_ref[1]**(o-1)+(0.418330181239)
        ref[3,0,2]=o*(0.526739853977)*x_ref[2]**(o-1)+(-0.00513414974056)
        ref[3,1,0]=o*(0.289534600598)*x_ref[0]**(o-1)+(0.467498987256)
        ref[3,1,1]=o*(0.27851903721)*x_ref[1]**(o-1)+(0.908249187588)
        ref[3,1,2]=o*(-0.84695489717)*x_ref[2]**(o-1)+(-0.799372710877)
        ref[3,2,0]=o*(0.00338011092566)*x_ref[0]**(o-1)+(0.819216072868)
        ref[3,2,1]=o*(0.115437140937)*x_ref[1]**(o-1)+(0.0968835984948)
        ref[3,2,2]=o*(-0.633253222581)*x_ref[2]**(o-1)+(-0.0224890801282)
        ref[3,3,0]=o*(0.578331574145)*x_ref[0]**(o-1)+(0.328247456626)
        ref[3,3,1]=o*(-0.725914955131)*x_ref[1]**(o-1)+(-0.863562766028)
        ref[3,3,2]=o*(-0.283057959979)*x_ref[2]**(o-1)+(-0.521563360583)
        ref[3,4,0]=o*(0.345725352698)*x_ref[0]**(o-1)+(0.614912618718)
        ref[3,4,1]=o*(-0.218084952915)*x_ref[1]**(o-1)+(-0.629514485943)
        ref[3,4,2]=o*(-0.207605980168)*x_ref[2]**(o-1)+(-0.0950042458608)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.353390112117)*x[0]**o+(0.121267964347)*x[0]+(0.095925646757)*x[1]**o+(-0.00141503525661)*x[1]
        arg[0,0,1]=(0.610556742471)*x[0]**o+(0.542948343923)*x[0]+(0.124762431991)*x[1]**o+(0.00216142978003)*x[1]
        arg[0,1,0]=(0.261559018753)*x[0]**o+(-0.881689419767)*x[0]+(-0.009352418091)*x[1]**o+(0.333357441723)*x[1]
        arg[0,1,1]=(-0.25379927884)*x[0]**o+(-0.124076833523)*x[0]+(-0.167690772835)*x[1]**o+(0.89804632709)*x[1]
        arg[1,0,0]=(-0.39353207502)*x[0]**o+(0.838815579522)*x[0]+(0.829467918805)*x[1]**o+(0.614403367394)*x[1]
        arg[1,0,1]=(-0.698946975112)*x[0]**o+(-0.551313390457)*x[0]+(0.965127390549)*x[1]**o+(-0.819660095397)*x[1]
        arg[1,1,0]=(-0.656237793952)*x[0]**o+(0.727261124295)*x[0]+(0.602446266985)*x[1]**o+(-0.320796826351)*x[1]
        arg[1,1,1]=(0.111875111713)*x[0]**o+(0.000360626785094)*x[0]+(-0.802516890555)*x[1]**o+(0.420759758492)*x[1]
        arg[2,0,0]=(-0.421920068352)*x[0]**o+(-0.022970702876)*x[0]+(-0.0257219638216)*x[1]**o+(-0.36544576073)*x[1]
        arg[2,0,1]=(-0.459431228769)*x[0]**o+(0.150702477357)*x[0]+(-0.869898945031)*x[1]**o+(0.261651620176)*x[1]
        arg[2,1,0]=(-0.866165026127)*x[0]**o+(0.149307656472)*x[0]+(-0.270058622191)*x[1]**o+(0.0873163553474)*x[1]
        arg[2,1,1]=(-0.871128617394)*x[0]**o+(-0.774356000237)*x[0]+(0.337383167457)*x[1]**o+(-0.453335744484)*x[1]
        arg[3,0,0]=(0.404195477481)*x[0]**o+(-0.350041445198)*x[0]+(-0.583794909281)*x[1]**o+(-0.895439618872)*x[1]
        arg[3,0,1]=(0.693567867164)*x[0]**o+(0.989642722856)*x[0]+(0.814104467522)*x[1]**o+(-0.126302588221)*x[1]
        arg[3,1,0]=(0.935509747673)*x[0]**o+(0.0104612091188)*x[0]+(0.304974305936)*x[1]**o+(0.658660310426)*x[1]
        arg[3,1,1]=(0.677420202064)*x[0]**o+(-0.160834685914)*x[0]+(0.241578146196)*x[1]**o+(-0.182223867353)*x[1]
        arg[4,0,0]=(-0.908542015864)*x[0]**o+(0.125595344482)*x[0]+(-0.396864254557)*x[1]**o+(0.408182849245)*x[1]
        arg[4,0,1]=(0.436101221257)*x[0]**o+(0.946667075656)*x[0]+(-0.293523459017)*x[1]**o+(0.363130416513)*x[1]
        arg[4,1,0]=(0.583416010605)*x[0]**o+(0.531883961319)*x[0]+(-0.418905217485)*x[1]**o+(0.271442034667)*x[1]
        arg[4,1,1]=(-0.708766397827)*x[0]**o+(0.0266251980073)*x[0]+(-0.0929293524384)*x[1]**o+(-0.661095295605)*x[1]
        arg[5,0,0]=(0.693424423443)*x[0]**o+(-0.201539674621)*x[0]+(-0.221922493999)*x[1]**o+(-0.218359043699)*x[1]
        arg[5,0,1]=(-0.478707796293)*x[0]**o+(0.791533578559)*x[0]+(0.280213321002)*x[1]**o+(-0.60809908629)*x[1]
        arg[5,1,0]=(-0.517992572451)*x[0]**o+(0.591995363281)*x[0]+(0.480220427965)*x[1]**o+(0.134542063762)*x[1]
        arg[5,1,1]=(0.415861674391)*x[0]**o+(0.773614372731)*x[0]+(-0.984515861984)*x[1]**o+(0.971916685266)*x[1]
        ref[0,0,0,0]=o*(-0.353390112117)*x_ref[0]**(o-1)+(0.121267964347)
        ref[0,0,0,1]=o*(0.095925646757)*x_ref[1]**(o-1)+(-0.00141503525661)
        ref[0,0,1,0]=o*(0.610556742471)*x_ref[0]**(o-1)+(0.542948343923)
        ref[0,0,1,1]=o*(0.124762431991)*x_ref[1]**(o-1)+(0.00216142978003)
        ref[0,1,0,0]=o*(0.261559018753)*x_ref[0]**(o-1)+(-0.881689419767)
        ref[0,1,0,1]=o*(-0.009352418091)*x_ref[1]**(o-1)+(0.333357441723)
        ref[0,1,1,0]=o*(-0.25379927884)*x_ref[0]**(o-1)+(-0.124076833523)
        ref[0,1,1,1]=o*(-0.167690772835)*x_ref[1]**(o-1)+(0.89804632709)
        ref[1,0,0,0]=o*(-0.39353207502)*x_ref[0]**(o-1)+(0.838815579522)
        ref[1,0,0,1]=o*(0.829467918805)*x_ref[1]**(o-1)+(0.614403367394)
        ref[1,0,1,0]=o*(-0.698946975112)*x_ref[0]**(o-1)+(-0.551313390457)
        ref[1,0,1,1]=o*(0.965127390549)*x_ref[1]**(o-1)+(-0.819660095397)
        ref[1,1,0,0]=o*(-0.656237793952)*x_ref[0]**(o-1)+(0.727261124295)
        ref[1,1,0,1]=o*(0.602446266985)*x_ref[1]**(o-1)+(-0.320796826351)
        ref[1,1,1,0]=o*(0.111875111713)*x_ref[0]**(o-1)+(0.000360626785094)
        ref[1,1,1,1]=o*(-0.802516890555)*x_ref[1]**(o-1)+(0.420759758492)
        ref[2,0,0,0]=o*(-0.421920068352)*x_ref[0]**(o-1)+(-0.022970702876)
        ref[2,0,0,1]=o*(-0.0257219638216)*x_ref[1]**(o-1)+(-0.36544576073)
        ref[2,0,1,0]=o*(-0.459431228769)*x_ref[0]**(o-1)+(0.150702477357)
        ref[2,0,1,1]=o*(-0.869898945031)*x_ref[1]**(o-1)+(0.261651620176)
        ref[2,1,0,0]=o*(-0.866165026127)*x_ref[0]**(o-1)+(0.149307656472)
        ref[2,1,0,1]=o*(-0.270058622191)*x_ref[1]**(o-1)+(0.0873163553474)
        ref[2,1,1,0]=o*(-0.871128617394)*x_ref[0]**(o-1)+(-0.774356000237)
        ref[2,1,1,1]=o*(0.337383167457)*x_ref[1]**(o-1)+(-0.453335744484)
        ref[3,0,0,0]=o*(0.404195477481)*x_ref[0]**(o-1)+(-0.350041445198)
        ref[3,0,0,1]=o*(-0.583794909281)*x_ref[1]**(o-1)+(-0.895439618872)
        ref[3,0,1,0]=o*(0.693567867164)*x_ref[0]**(o-1)+(0.989642722856)
        ref[3,0,1,1]=o*(0.814104467522)*x_ref[1]**(o-1)+(-0.126302588221)
        ref[3,1,0,0]=o*(0.935509747673)*x_ref[0]**(o-1)+(0.0104612091188)
        ref[3,1,0,1]=o*(0.304974305936)*x_ref[1]**(o-1)+(0.658660310426)
        ref[3,1,1,0]=o*(0.677420202064)*x_ref[0]**(o-1)+(-0.160834685914)
        ref[3,1,1,1]=o*(0.241578146196)*x_ref[1]**(o-1)+(-0.182223867353)
        ref[4,0,0,0]=o*(-0.908542015864)*x_ref[0]**(o-1)+(0.125595344482)
        ref[4,0,0,1]=o*(-0.396864254557)*x_ref[1]**(o-1)+(0.408182849245)
        ref[4,0,1,0]=o*(0.436101221257)*x_ref[0]**(o-1)+(0.946667075656)
        ref[4,0,1,1]=o*(-0.293523459017)*x_ref[1]**(o-1)+(0.363130416513)
        ref[4,1,0,0]=o*(0.583416010605)*x_ref[0]**(o-1)+(0.531883961319)
        ref[4,1,0,1]=o*(-0.418905217485)*x_ref[1]**(o-1)+(0.271442034667)
        ref[4,1,1,0]=o*(-0.708766397827)*x_ref[0]**(o-1)+(0.0266251980073)
        ref[4,1,1,1]=o*(-0.0929293524384)*x_ref[1]**(o-1)+(-0.661095295605)
        ref[5,0,0,0]=o*(0.693424423443)*x_ref[0]**(o-1)+(-0.201539674621)
        ref[5,0,0,1]=o*(-0.221922493999)*x_ref[1]**(o-1)+(-0.218359043699)
        ref[5,0,1,0]=o*(-0.478707796293)*x_ref[0]**(o-1)+(0.791533578559)
        ref[5,0,1,1]=o*(0.280213321002)*x_ref[1]**(o-1)+(-0.60809908629)
        ref[5,1,0,0]=o*(-0.517992572451)*x_ref[0]**(o-1)+(0.591995363281)
        ref[5,1,0,1]=o*(0.480220427965)*x_ref[1]**(o-1)+(0.134542063762)
        ref[5,1,1,0]=o*(0.415861674391)*x_ref[0]**(o-1)+(0.773614372731)
        ref[5,1,1,1]=o*(-0.984515861984)*x_ref[1]**(o-1)+(0.971916685266)
      else:
        arg[0,0,0]=(0.157678109652)*x[0]**o+(-0.939095069874)*x[0]+(0.340072659801)*x[1]**o+(0.806723712633)*x[1]+(0.0175654201162)*x[2]**o+(-0.695887679405)*x[2]
        arg[0,0,1]=(0.148444957808)*x[0]**o+(0.769780554196)*x[0]+(-0.999778710821)*x[1]**o+(0.765998508134)*x[1]+(0.489498280892)*x[2]**o+(0.272011172233)*x[2]
        arg[0,1,0]=(-0.78853623669)*x[0]**o+(-0.889333747099)*x[0]+(0.897684283327)*x[1]**o+(0.878046029595)*x[1]+(-0.355566201991)*x[2]**o+(-0.656187599294)*x[2]
        arg[0,1,1]=(-0.903496685895)*x[0]**o+(0.335762730702)*x[0]+(0.264330545204)*x[1]**o+(-0.369300150725)*x[1]+(0.00819339769162)*x[2]**o+(0.0466274018566)*x[2]
        arg[1,0,0]=(0.490893502666)*x[0]**o+(0.853253169019)*x[0]+(-0.706120536551)*x[1]**o+(-0.810665092749)*x[1]+(0.877124690496)*x[2]**o+(-0.663316848595)*x[2]
        arg[1,0,1]=(-0.897301110118)*x[0]**o+(-0.571577796347)*x[0]+(-0.00912216314091)*x[1]**o+(0.479091266837)*x[1]+(-0.951994896347)*x[2]**o+(0.293801256264)*x[2]
        arg[1,1,0]=(-0.0843442242335)*x[0]**o+(0.436592099426)*x[0]+(-0.961835284742)*x[1]**o+(0.453758903497)*x[1]+(0.759916173417)*x[2]**o+(-0.356674191439)*x[2]
        arg[1,1,1]=(0.368337095889)*x[0]**o+(-0.180552461333)*x[0]+(-0.814727889141)*x[1]**o+(-0.226678543654)*x[1]+(0.629233877979)*x[2]**o+(0.934452743806)*x[2]
        arg[2,0,0]=(0.335801632975)*x[0]**o+(0.4853793666)*x[0]+(-0.603075878381)*x[1]**o+(-0.464873655094)*x[1]+(-0.304107114487)*x[2]**o+(0.768564020927)*x[2]
        arg[2,0,1]=(-0.843867260693)*x[0]**o+(-0.758493294188)*x[0]+(0.801820428023)*x[1]**o+(0.0909753752871)*x[1]+(-0.208134327922)*x[2]**o+(0.343325582627)*x[2]
        arg[2,1,0]=(-0.143364423799)*x[0]**o+(-0.191201452642)*x[0]+(0.356480498629)*x[1]**o+(0.0853546566948)*x[1]+(0.234781870759)*x[2]**o+(-0.702906217789)*x[2]
        arg[2,1,1]=(0.81324792898)*x[0]**o+(0.565754643122)*x[0]+(-0.0246314241495)*x[1]**o+(-0.83548496516)*x[1]+(0.730246143504)*x[2]**o+(-0.835539994852)*x[2]
        arg[3,0,0]=(0.877992836393)*x[0]**o+(-0.137135423487)*x[0]+(0.534511286076)*x[1]**o+(-0.0584370047297)*x[1]+(-0.0497236607574)*x[2]**o+(0.408414670632)*x[2]
        arg[3,0,1]=(-0.959771372305)*x[0]**o+(0.587216149992)*x[0]+(-0.873216249139)*x[1]**o+(0.0703350087653)*x[1]+(0.452900901257)*x[2]**o+(-0.180788315472)*x[2]
        arg[3,1,0]=(0.477701526969)*x[0]**o+(0.16165566827)*x[0]+(0.219901428845)*x[1]**o+(-0.0816687269195)*x[1]+(-0.379553348394)*x[2]**o+(-0.764672519612)*x[2]
        arg[3,1,1]=(0.274518799686)*x[0]**o+(0.945009722336)*x[0]+(-0.979787014238)*x[1]**o+(0.587704213162)*x[1]+(0.7959640984)*x[2]**o+(-0.0811112488533)*x[2]
        arg[4,0,0]=(-0.579121056378)*x[0]**o+(0.971799195206)*x[0]+(-0.646271217985)*x[1]**o+(0.369751123548)*x[1]+(0.792425864156)*x[2]**o+(-0.170799235786)*x[2]
        arg[4,0,1]=(0.0939380562355)*x[0]**o+(0.857041359029)*x[0]+(-0.379025882204)*x[1]**o+(-0.571968622949)*x[1]+(-0.156853183697)*x[2]**o+(0.643214485063)*x[2]
        arg[4,1,0]=(-0.715210364871)*x[0]**o+(0.150341666571)*x[0]+(0.325165277323)*x[1]**o+(-0.322744240034)*x[1]+(-0.622094139461)*x[2]**o+(-0.830479146327)*x[2]
        arg[4,1,1]=(-0.697109424406)*x[0]**o+(-0.730318259042)*x[0]+(-0.135119012822)*x[1]**o+(0.410502029229)*x[1]+(0.547987909743)*x[2]**o+(0.144877544254)*x[2]
        arg[5,0,0]=(0.787147597087)*x[0]**o+(-0.528180541295)*x[0]+(-0.251578089108)*x[1]**o+(-0.170766337864)*x[1]+(-0.123341129221)*x[2]**o+(-0.815622069347)*x[2]
        arg[5,0,1]=(0.193873484005)*x[0]**o+(0.270190252915)*x[0]+(0.807946985621)*x[1]**o+(0.184522999461)*x[1]+(0.97429597406)*x[2]**o+(0.532931028021)*x[2]
        arg[5,1,0]=(0.555644676015)*x[0]**o+(-0.225593442089)*x[0]+(0.877869009767)*x[1]**o+(-0.750894245506)*x[1]+(0.888501611997)*x[2]**o+(-0.84027021889)*x[2]
        arg[5,1,1]=(0.76909422867)*x[0]**o+(0.406697645173)*x[0]+(-0.233710263631)*x[1]**o+(-0.432942345121)*x[1]+(0.270484113666)*x[2]**o+(0.707302418115)*x[2]
        ref[0,0,0,0]=o*(0.157678109652)*x_ref[0]**(o-1)+(-0.939095069874)
        ref[0,0,0,1]=o*(0.340072659801)*x_ref[1]**(o-1)+(0.806723712633)
        ref[0,0,0,2]=o*(0.0175654201162)*x_ref[2]**(o-1)+(-0.695887679405)
        ref[0,0,1,0]=o*(0.148444957808)*x_ref[0]**(o-1)+(0.769780554196)
        ref[0,0,1,1]=o*(-0.999778710821)*x_ref[1]**(o-1)+(0.765998508134)
        ref[0,0,1,2]=o*(0.489498280892)*x_ref[2]**(o-1)+(0.272011172233)
        ref[0,1,0,0]=o*(-0.78853623669)*x_ref[0]**(o-1)+(-0.889333747099)
        ref[0,1,0,1]=o*(0.897684283327)*x_ref[1]**(o-1)+(0.878046029595)
        ref[0,1,0,2]=o*(-0.355566201991)*x_ref[2]**(o-1)+(-0.656187599294)
        ref[0,1,1,0]=o*(-0.903496685895)*x_ref[0]**(o-1)+(0.335762730702)
        ref[0,1,1,1]=o*(0.264330545204)*x_ref[1]**(o-1)+(-0.369300150725)
        ref[0,1,1,2]=o*(0.00819339769162)*x_ref[2]**(o-1)+(0.0466274018566)
        ref[1,0,0,0]=o*(0.490893502666)*x_ref[0]**(o-1)+(0.853253169019)
        ref[1,0,0,1]=o*(-0.706120536551)*x_ref[1]**(o-1)+(-0.810665092749)
        ref[1,0,0,2]=o*(0.877124690496)*x_ref[2]**(o-1)+(-0.663316848595)
        ref[1,0,1,0]=o*(-0.897301110118)*x_ref[0]**(o-1)+(-0.571577796347)
        ref[1,0,1,1]=o*(-0.00912216314091)*x_ref[1]**(o-1)+(0.479091266837)
        ref[1,0,1,2]=o*(-0.951994896347)*x_ref[2]**(o-1)+(0.293801256264)
        ref[1,1,0,0]=o*(-0.0843442242335)*x_ref[0]**(o-1)+(0.436592099426)
        ref[1,1,0,1]=o*(-0.961835284742)*x_ref[1]**(o-1)+(0.453758903497)
        ref[1,1,0,2]=o*(0.759916173417)*x_ref[2]**(o-1)+(-0.356674191439)
        ref[1,1,1,0]=o*(0.368337095889)*x_ref[0]**(o-1)+(-0.180552461333)
        ref[1,1,1,1]=o*(-0.814727889141)*x_ref[1]**(o-1)+(-0.226678543654)
        ref[1,1,1,2]=o*(0.629233877979)*x_ref[2]**(o-1)+(0.934452743806)
        ref[2,0,0,0]=o*(0.335801632975)*x_ref[0]**(o-1)+(0.4853793666)
        ref[2,0,0,1]=o*(-0.603075878381)*x_ref[1]**(o-1)+(-0.464873655094)
        ref[2,0,0,2]=o*(-0.304107114487)*x_ref[2]**(o-1)+(0.768564020927)
        ref[2,0,1,0]=o*(-0.843867260693)*x_ref[0]**(o-1)+(-0.758493294188)
        ref[2,0,1,1]=o*(0.801820428023)*x_ref[1]**(o-1)+(0.0909753752871)
        ref[2,0,1,2]=o*(-0.208134327922)*x_ref[2]**(o-1)+(0.343325582627)
        ref[2,1,0,0]=o*(-0.143364423799)*x_ref[0]**(o-1)+(-0.191201452642)
        ref[2,1,0,1]=o*(0.356480498629)*x_ref[1]**(o-1)+(0.0853546566948)
        ref[2,1,0,2]=o*(0.234781870759)*x_ref[2]**(o-1)+(-0.702906217789)
        ref[2,1,1,0]=o*(0.81324792898)*x_ref[0]**(o-1)+(0.565754643122)
        ref[2,1,1,1]=o*(-0.0246314241495)*x_ref[1]**(o-1)+(-0.83548496516)
        ref[2,1,1,2]=o*(0.730246143504)*x_ref[2]**(o-1)+(-0.835539994852)
        ref[3,0,0,0]=o*(0.877992836393)*x_ref[0]**(o-1)+(-0.137135423487)
        ref[3,0,0,1]=o*(0.534511286076)*x_ref[1]**(o-1)+(-0.0584370047297)
        ref[3,0,0,2]=o*(-0.0497236607574)*x_ref[2]**(o-1)+(0.408414670632)
        ref[3,0,1,0]=o*(-0.959771372305)*x_ref[0]**(o-1)+(0.587216149992)
        ref[3,0,1,1]=o*(-0.873216249139)*x_ref[1]**(o-1)+(0.0703350087653)
        ref[3,0,1,2]=o*(0.452900901257)*x_ref[2]**(o-1)+(-0.180788315472)
        ref[3,1,0,0]=o*(0.477701526969)*x_ref[0]**(o-1)+(0.16165566827)
        ref[3,1,0,1]=o*(0.219901428845)*x_ref[1]**(o-1)+(-0.0816687269195)
        ref[3,1,0,2]=o*(-0.379553348394)*x_ref[2]**(o-1)+(-0.764672519612)
        ref[3,1,1,0]=o*(0.274518799686)*x_ref[0]**(o-1)+(0.945009722336)
        ref[3,1,1,1]=o*(-0.979787014238)*x_ref[1]**(o-1)+(0.587704213162)
        ref[3,1,1,2]=o*(0.7959640984)*x_ref[2]**(o-1)+(-0.0811112488533)
        ref[4,0,0,0]=o*(-0.579121056378)*x_ref[0]**(o-1)+(0.971799195206)
        ref[4,0,0,1]=o*(-0.646271217985)*x_ref[1]**(o-1)+(0.369751123548)
        ref[4,0,0,2]=o*(0.792425864156)*x_ref[2]**(o-1)+(-0.170799235786)
        ref[4,0,1,0]=o*(0.0939380562355)*x_ref[0]**(o-1)+(0.857041359029)
        ref[4,0,1,1]=o*(-0.379025882204)*x_ref[1]**(o-1)+(-0.571968622949)
        ref[4,0,1,2]=o*(-0.156853183697)*x_ref[2]**(o-1)+(0.643214485063)
        ref[4,1,0,0]=o*(-0.715210364871)*x_ref[0]**(o-1)+(0.150341666571)
        ref[4,1,0,1]=o*(0.325165277323)*x_ref[1]**(o-1)+(-0.322744240034)
        ref[4,1,0,2]=o*(-0.622094139461)*x_ref[2]**(o-1)+(-0.830479146327)
        ref[4,1,1,0]=o*(-0.697109424406)*x_ref[0]**(o-1)+(-0.730318259042)
        ref[4,1,1,1]=o*(-0.135119012822)*x_ref[1]**(o-1)+(0.410502029229)
        ref[4,1,1,2]=o*(0.547987909743)*x_ref[2]**(o-1)+(0.144877544254)
        ref[5,0,0,0]=o*(0.787147597087)*x_ref[0]**(o-1)+(-0.528180541295)
        ref[5,0,0,1]=o*(-0.251578089108)*x_ref[1]**(o-1)+(-0.170766337864)
        ref[5,0,0,2]=o*(-0.123341129221)*x_ref[2]**(o-1)+(-0.815622069347)
        ref[5,0,1,0]=o*(0.193873484005)*x_ref[0]**(o-1)+(0.270190252915)
        ref[5,0,1,1]=o*(0.807946985621)*x_ref[1]**(o-1)+(0.184522999461)
        ref[5,0,1,2]=o*(0.97429597406)*x_ref[2]**(o-1)+(0.532931028021)
        ref[5,1,0,0]=o*(0.555644676015)*x_ref[0]**(o-1)+(-0.225593442089)
        ref[5,1,0,1]=o*(0.877869009767)*x_ref[1]**(o-1)+(-0.750894245506)
        ref[5,1,0,2]=o*(0.888501611997)*x_ref[2]**(o-1)+(-0.84027021889)
        ref[5,1,1,0]=o*(0.76909422867)*x_ref[0]**(o-1)+(0.406697645173)
        ref[5,1,1,1]=o*(-0.233710263631)*x_ref[1]**(o-1)+(-0.432942345121)
        ref[5,1,1,2]=o*(0.270484113666)*x_ref[2]**(o-1)+(0.707302418115)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gradient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.350515208335)*x[0]**o+(-0.604596801095)*x[0]+(0.145267491062)*x[1]**o+(-0.422271555369)*x[1]
        ref=(-0.350515208335)*x_ref[0]**o+(-0.604596801095)*x_ref[0]+(0.145267491062)*x_ref[1]**o+(-0.422271555369)*x_ref[1]
      else:
        arg=(-0.162304501731)*x[0]**o+(0.342262454165)*x[0]+(0.693988751897)*x[1]**o+(-0.315225717754)*x[1]+(-0.938405302425)*x[2]**o+(-0.0929881908885)*x[2]
        ref=(-0.162304501731)*x_ref[0]**o+(0.342262454165)*x_ref[0]+(0.693988751897)*x_ref[1]**o+(-0.315225717754)*x_ref[1]+(-0.938405302425)*x_ref[2]**o+(-0.0929881908885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.104854565192)*x[0]**o+(-0.596688162793)*x[0]+(-0.0286471957779)*x[1]**o+(-0.157282508059)*x[1]
        arg[1]=(0.619817759027)*x[0]**o+(-0.832815965127)*x[0]+(-0.638126497014)*x[1]**o+(0.771271149341)*x[1]
        ref[0]=(0.104854565192)*x_ref[0]**o+(-0.596688162793)*x_ref[0]+(-0.0286471957779)*x_ref[1]**o+(-0.157282508059)*x_ref[1]
        ref[1]=(0.619817759027)*x_ref[0]**o+(-0.832815965127)*x_ref[0]+(-0.638126497014)*x_ref[1]**o+(0.771271149341)*x_ref[1]
      else:
        arg[0]=(-0.358142932888)*x[0]**o+(0.758685606027)*x[0]+(-0.0716050574096)*x[1]**o+(-0.723587074286)*x[1]+(0.387240173671)*x[2]**o+(0.544679811421)*x[2]
        arg[1]=(0.864154242034)*x[0]**o+(-0.0740598679188)*x[0]+(-0.0865611132887)*x[1]**o+(0.0414815431481)*x[1]+(-0.897395775334)*x[2]**o+(-0.580401889318)*x[2]
        ref[0]=(-0.358142932888)*x_ref[0]**o+(0.758685606027)*x_ref[0]+(-0.0716050574096)*x_ref[1]**o+(-0.723587074286)*x_ref[1]+(0.387240173671)*x_ref[2]**o+(0.544679811421)*x_ref[2]
        ref[1]=(0.864154242034)*x_ref[0]**o+(-0.0740598679188)*x_ref[0]+(-0.0865611132887)*x_ref[1]**o+(0.0414815431481)*x_ref[1]+(-0.897395775334)*x_ref[2]**o+(-0.580401889318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.0571385613076)*x[0]**o+(-0.296048600622)*x[0]+(0.300914212927)*x[1]**o+(0.81578919689)*x[1]
        arg[0,1]=(0.136796732287)*x[0]**o+(0.169384864236)*x[0]+(-0.874585047899)*x[1]**o+(-0.67613629157)*x[1]
        arg[0,2]=(-0.947048556597)*x[0]**o+(-0.807647782941)*x[0]+(-0.53233790031)*x[1]**o+(0.618861393596)*x[1]
        arg[0,3]=(-0.533009370674)*x[0]**o+(-0.732258917532)*x[0]+(0.735058784793)*x[1]**o+(0.313284184128)*x[1]
        arg[0,4]=(0.678291805831)*x[0]**o+(0.892272189132)*x[0]+(-0.902740106267)*x[1]**o+(0.902593764933)*x[1]
        arg[1,0]=(0.101446968367)*x[0]**o+(0.758270058166)*x[0]+(-0.511416064913)*x[1]**o+(-0.464147847068)*x[1]
        arg[1,1]=(-0.0131419066862)*x[0]**o+(-0.942460391354)*x[0]+(0.479955107486)*x[1]**o+(-0.311304181244)*x[1]
        arg[1,2]=(0.623958732361)*x[0]**o+(-0.313196338837)*x[0]+(0.413428996617)*x[1]**o+(0.348327566672)*x[1]
        arg[1,3]=(-0.8413636876)*x[0]**o+(-0.572573098032)*x[0]+(-0.678663954356)*x[1]**o+(-0.419716403314)*x[1]
        arg[1,4]=(-0.580466216486)*x[0]**o+(0.384226238557)*x[0]+(-0.732722613215)*x[1]**o+(-0.0238970013828)*x[1]
        arg[2,0]=(-0.248986933717)*x[0]**o+(0.264019654506)*x[0]+(0.0314811616292)*x[1]**o+(0.631481081422)*x[1]
        arg[2,1]=(0.643576426484)*x[0]**o+(-0.313599735242)*x[0]+(0.908780876791)*x[1]**o+(-0.698972268201)*x[1]
        arg[2,2]=(-0.623208321993)*x[0]**o+(0.649637800083)*x[0]+(-0.878484988208)*x[1]**o+(0.368793479691)*x[1]
        arg[2,3]=(0.239333407828)*x[0]**o+(-0.293558915555)*x[0]+(0.271630238733)*x[1]**o+(0.741304510722)*x[1]
        arg[2,4]=(0.175914942104)*x[0]**o+(0.336116863935)*x[0]+(-0.799168065807)*x[1]**o+(0.0264962830729)*x[1]
        arg[3,0]=(0.777665875019)*x[0]**o+(-0.13232137001)*x[0]+(0.682046389381)*x[1]**o+(0.589069623109)*x[1]
        arg[3,1]=(0.7607621575)*x[0]**o+(-0.81285460551)*x[0]+(0.0523010659976)*x[1]**o+(-0.913765717778)*x[1]
        arg[3,2]=(0.822107753118)*x[0]**o+(0.508869893322)*x[0]+(0.679583348422)*x[1]**o+(-0.597814039932)*x[1]
        arg[3,3]=(-0.77699176705)*x[0]**o+(-0.267758166676)*x[0]+(0.661073127871)*x[1]**o+(-0.41434545147)*x[1]
        arg[3,4]=(0.223013380161)*x[0]**o+(-0.558077040245)*x[0]+(-0.450831764747)*x[1]**o+(-0.0897765635191)*x[1]
        ref[0,0]=(-0.0571385613076)*x_ref[0]**o+(-0.296048600622)*x_ref[0]+(0.300914212927)*x_ref[1]**o+(0.81578919689)*x_ref[1]
        ref[0,1]=(0.136796732287)*x_ref[0]**o+(0.169384864236)*x_ref[0]+(-0.874585047899)*x_ref[1]**o+(-0.67613629157)*x_ref[1]
        ref[0,2]=(-0.947048556597)*x_ref[0]**o+(-0.807647782941)*x_ref[0]+(-0.53233790031)*x_ref[1]**o+(0.618861393596)*x_ref[1]
        ref[0,3]=(-0.533009370674)*x_ref[0]**o+(-0.732258917532)*x_ref[0]+(0.735058784793)*x_ref[1]**o+(0.313284184128)*x_ref[1]
        ref[0,4]=(0.678291805831)*x_ref[0]**o+(0.892272189132)*x_ref[0]+(-0.902740106267)*x_ref[1]**o+(0.902593764933)*x_ref[1]
        ref[1,0]=(0.101446968367)*x_ref[0]**o+(0.758270058166)*x_ref[0]+(-0.511416064913)*x_ref[1]**o+(-0.464147847068)*x_ref[1]
        ref[1,1]=(-0.0131419066862)*x_ref[0]**o+(-0.942460391354)*x_ref[0]+(0.479955107486)*x_ref[1]**o+(-0.311304181244)*x_ref[1]
        ref[1,2]=(0.623958732361)*x_ref[0]**o+(-0.313196338837)*x_ref[0]+(0.413428996617)*x_ref[1]**o+(0.348327566672)*x_ref[1]
        ref[1,3]=(-0.8413636876)*x_ref[0]**o+(-0.572573098032)*x_ref[0]+(-0.678663954356)*x_ref[1]**o+(-0.419716403314)*x_ref[1]
        ref[1,4]=(-0.580466216486)*x_ref[0]**o+(0.384226238557)*x_ref[0]+(-0.732722613215)*x_ref[1]**o+(-0.0238970013828)*x_ref[1]
        ref[2,0]=(-0.248986933717)*x_ref[0]**o+(0.264019654506)*x_ref[0]+(0.0314811616292)*x_ref[1]**o+(0.631481081422)*x_ref[1]
        ref[2,1]=(0.643576426484)*x_ref[0]**o+(-0.313599735242)*x_ref[0]+(0.908780876791)*x_ref[1]**o+(-0.698972268201)*x_ref[1]
        ref[2,2]=(-0.623208321993)*x_ref[0]**o+(0.649637800083)*x_ref[0]+(-0.878484988208)*x_ref[1]**o+(0.368793479691)*x_ref[1]
        ref[2,3]=(0.239333407828)*x_ref[0]**o+(-0.293558915555)*x_ref[0]+(0.271630238733)*x_ref[1]**o+(0.741304510722)*x_ref[1]
        ref[2,4]=(0.175914942104)*x_ref[0]**o+(0.336116863935)*x_ref[0]+(-0.799168065807)*x_ref[1]**o+(0.0264962830729)*x_ref[1]
        ref[3,0]=(0.777665875019)*x_ref[0]**o+(-0.13232137001)*x_ref[0]+(0.682046389381)*x_ref[1]**o+(0.589069623109)*x_ref[1]
        ref[3,1]=(0.7607621575)*x_ref[0]**o+(-0.81285460551)*x_ref[0]+(0.0523010659976)*x_ref[1]**o+(-0.913765717778)*x_ref[1]
        ref[3,2]=(0.822107753118)*x_ref[0]**o+(0.508869893322)*x_ref[0]+(0.679583348422)*x_ref[1]**o+(-0.597814039932)*x_ref[1]
        ref[3,3]=(-0.77699176705)*x_ref[0]**o+(-0.267758166676)*x_ref[0]+(0.661073127871)*x_ref[1]**o+(-0.41434545147)*x_ref[1]
        ref[3,4]=(0.223013380161)*x_ref[0]**o+(-0.558077040245)*x_ref[0]+(-0.450831764747)*x_ref[1]**o+(-0.0897765635191)*x_ref[1]
      else:
        arg[0,0]=(0.487697656095)*x[0]**o+(0.362509734966)*x[0]+(-0.40438484582)*x[1]**o+(-0.358984481681)*x[1]+(0.157670938098)*x[2]**o+(-0.631406375093)*x[2]
        arg[0,1]=(-0.00251154956677)*x[0]**o+(0.671447769464)*x[0]+(0.535541351526)*x[1]**o+(0.433508112485)*x[1]+(-0.897668718638)*x[2]**o+(0.141331586102)*x[2]
        arg[0,2]=(-0.964778505658)*x[0]**o+(0.522815935313)*x[0]+(-0.750318077186)*x[1]**o+(0.0595240411196)*x[1]+(-0.384021192511)*x[2]**o+(-0.802529657745)*x[2]
        arg[0,3]=(0.264781810221)*x[0]**o+(-0.654277693956)*x[0]+(0.890264421741)*x[1]**o+(-0.657743774786)*x[1]+(0.320234835664)*x[2]**o+(0.0637070094165)*x[2]
        arg[0,4]=(0.712229359775)*x[0]**o+(-0.646586719826)*x[0]+(0.876519277606)*x[1]**o+(0.473649630238)*x[1]+(0.332259187685)*x[2]**o+(-0.331489699179)*x[2]
        arg[1,0]=(0.808534274392)*x[0]**o+(-0.296462057802)*x[0]+(0.152536716655)*x[1]**o+(-0.990267302005)*x[1]+(-0.823462709316)*x[2]**o+(-0.51608634537)*x[2]
        arg[1,1]=(-0.745238552081)*x[0]**o+(-0.892086591023)*x[0]+(0.691429994644)*x[1]**o+(-0.52538570831)*x[1]+(-0.668959934842)*x[2]**o+(-0.5178179109)*x[2]
        arg[1,2]=(0.10173204456)*x[0]**o+(-0.853001564748)*x[0]+(0.617252846191)*x[1]**o+(-0.518228673297)*x[1]+(-0.995447484095)*x[2]**o+(0.56679145486)*x[2]
        arg[1,3]=(0.285145125389)*x[0]**o+(0.773362877732)*x[0]+(-0.457364007088)*x[1]**o+(0.226257217659)*x[1]+(0.439832733067)*x[2]**o+(0.872812128578)*x[2]
        arg[1,4]=(-0.581535304023)*x[0]**o+(-0.944599120493)*x[0]+(-0.720009453699)*x[1]**o+(-0.29624047891)*x[1]+(0.966528522407)*x[2]**o+(0.818945905214)*x[2]
        arg[2,0]=(-0.650732348774)*x[0]**o+(0.368909306978)*x[0]+(-0.608327978493)*x[1]**o+(0.301600904054)*x[1]+(-0.747648472883)*x[2]**o+(0.846568601061)*x[2]
        arg[2,1]=(-0.0221803961531)*x[0]**o+(-0.858842969825)*x[0]+(-0.40141447334)*x[1]**o+(0.408164536045)*x[1]+(-0.0131899768237)*x[2]**o+(-0.0983376041104)*x[2]
        arg[2,2]=(0.0855129797683)*x[0]**o+(0.0645200342375)*x[0]+(0.295916169508)*x[1]**o+(0.343169626483)*x[1]+(0.973477621396)*x[2]**o+(0.179505855846)*x[2]
        arg[2,3]=(-0.115714157482)*x[0]**o+(-0.00785628190789)*x[0]+(0.406498238895)*x[1]**o+(-0.954353977843)*x[1]+(0.751544148524)*x[2]**o+(-0.092307218389)*x[2]
        arg[2,4]=(0.807583813259)*x[0]**o+(-0.0150562208564)*x[0]+(0.967290753425)*x[1]**o+(0.960049095802)*x[1]+(0.41935259669)*x[2]**o+(0.878823415691)*x[2]
        arg[3,0]=(-0.162221321908)*x[0]**o+(0.0142404015457)*x[0]+(0.769189546147)*x[1]**o+(-0.251240080116)*x[1]+(0.980441953151)*x[2]**o+(-0.183639594778)*x[2]
        arg[3,1]=(0.270231471303)*x[0]**o+(-0.527471406108)*x[0]+(-0.255144134035)*x[1]**o+(0.268061599615)*x[1]+(-0.396373327005)*x[2]**o+(0.105905036411)*x[2]
        arg[3,2]=(0.5159632953)*x[0]**o+(-0.87397274133)*x[0]+(0.493876004047)*x[1]**o+(0.320996363656)*x[1]+(0.884345148533)*x[2]**o+(0.116262647026)*x[2]
        arg[3,3]=(-0.354053396048)*x[0]**o+(-0.130121845174)*x[0]+(-0.33114435645)*x[1]**o+(-0.73998189126)*x[1]+(0.362417530484)*x[2]**o+(0.676820875084)*x[2]
        arg[3,4]=(-0.53273881792)*x[0]**o+(-0.0624208514424)*x[0]+(0.680486778939)*x[1]**o+(-0.370666453954)*x[1]+(0.294345859395)*x[2]**o+(-0.679169213305)*x[2]
        ref[0,0]=(0.487697656095)*x_ref[0]**o+(0.362509734966)*x_ref[0]+(-0.40438484582)*x_ref[1]**o+(-0.358984481681)*x_ref[1]+(0.157670938098)*x_ref[2]**o+(-0.631406375093)*x_ref[2]
        ref[0,1]=(-0.00251154956677)*x_ref[0]**o+(0.671447769464)*x_ref[0]+(0.535541351526)*x_ref[1]**o+(0.433508112485)*x_ref[1]+(-0.897668718638)*x_ref[2]**o+(0.141331586102)*x_ref[2]
        ref[0,2]=(-0.964778505658)*x_ref[0]**o+(0.522815935313)*x_ref[0]+(-0.750318077186)*x_ref[1]**o+(0.0595240411196)*x_ref[1]+(-0.384021192511)*x_ref[2]**o+(-0.802529657745)*x_ref[2]
        ref[0,3]=(0.264781810221)*x_ref[0]**o+(-0.654277693956)*x_ref[0]+(0.890264421741)*x_ref[1]**o+(-0.657743774786)*x_ref[1]+(0.320234835664)*x_ref[2]**o+(0.0637070094165)*x_ref[2]
        ref[0,4]=(0.712229359775)*x_ref[0]**o+(-0.646586719826)*x_ref[0]+(0.876519277606)*x_ref[1]**o+(0.473649630238)*x_ref[1]+(0.332259187685)*x_ref[2]**o+(-0.331489699179)*x_ref[2]
        ref[1,0]=(0.808534274392)*x_ref[0]**o+(-0.296462057802)*x_ref[0]+(0.152536716655)*x_ref[1]**o+(-0.990267302005)*x_ref[1]+(-0.823462709316)*x_ref[2]**o+(-0.51608634537)*x_ref[2]
        ref[1,1]=(-0.745238552081)*x_ref[0]**o+(-0.892086591023)*x_ref[0]+(0.691429994644)*x_ref[1]**o+(-0.52538570831)*x_ref[1]+(-0.668959934842)*x_ref[2]**o+(-0.5178179109)*x_ref[2]
        ref[1,2]=(0.10173204456)*x_ref[0]**o+(-0.853001564748)*x_ref[0]+(0.617252846191)*x_ref[1]**o+(-0.518228673297)*x_ref[1]+(-0.995447484095)*x_ref[2]**o+(0.56679145486)*x_ref[2]
        ref[1,3]=(0.285145125389)*x_ref[0]**o+(0.773362877732)*x_ref[0]+(-0.457364007088)*x_ref[1]**o+(0.226257217659)*x_ref[1]+(0.439832733067)*x_ref[2]**o+(0.872812128578)*x_ref[2]
        ref[1,4]=(-0.581535304023)*x_ref[0]**o+(-0.944599120493)*x_ref[0]+(-0.720009453699)*x_ref[1]**o+(-0.29624047891)*x_ref[1]+(0.966528522407)*x_ref[2]**o+(0.818945905214)*x_ref[2]
        ref[2,0]=(-0.650732348774)*x_ref[0]**o+(0.368909306978)*x_ref[0]+(-0.608327978493)*x_ref[1]**o+(0.301600904054)*x_ref[1]+(-0.747648472883)*x_ref[2]**o+(0.846568601061)*x_ref[2]
        ref[2,1]=(-0.0221803961531)*x_ref[0]**o+(-0.858842969825)*x_ref[0]+(-0.40141447334)*x_ref[1]**o+(0.408164536045)*x_ref[1]+(-0.0131899768237)*x_ref[2]**o+(-0.0983376041104)*x_ref[2]
        ref[2,2]=(0.0855129797683)*x_ref[0]**o+(0.0645200342375)*x_ref[0]+(0.295916169508)*x_ref[1]**o+(0.343169626483)*x_ref[1]+(0.973477621396)*x_ref[2]**o+(0.179505855846)*x_ref[2]
        ref[2,3]=(-0.115714157482)*x_ref[0]**o+(-0.00785628190789)*x_ref[0]+(0.406498238895)*x_ref[1]**o+(-0.954353977843)*x_ref[1]+(0.751544148524)*x_ref[2]**o+(-0.092307218389)*x_ref[2]
        ref[2,4]=(0.807583813259)*x_ref[0]**o+(-0.0150562208564)*x_ref[0]+(0.967290753425)*x_ref[1]**o+(0.960049095802)*x_ref[1]+(0.41935259669)*x_ref[2]**o+(0.878823415691)*x_ref[2]
        ref[3,0]=(-0.162221321908)*x_ref[0]**o+(0.0142404015457)*x_ref[0]+(0.769189546147)*x_ref[1]**o+(-0.251240080116)*x_ref[1]+(0.980441953151)*x_ref[2]**o+(-0.183639594778)*x_ref[2]
        ref[3,1]=(0.270231471303)*x_ref[0]**o+(-0.527471406108)*x_ref[0]+(-0.255144134035)*x_ref[1]**o+(0.268061599615)*x_ref[1]+(-0.396373327005)*x_ref[2]**o+(0.105905036411)*x_ref[2]
        ref[3,2]=(0.5159632953)*x_ref[0]**o+(-0.87397274133)*x_ref[0]+(0.493876004047)*x_ref[1]**o+(0.320996363656)*x_ref[1]+(0.884345148533)*x_ref[2]**o+(0.116262647026)*x_ref[2]
        ref[3,3]=(-0.354053396048)*x_ref[0]**o+(-0.130121845174)*x_ref[0]+(-0.33114435645)*x_ref[1]**o+(-0.73998189126)*x_ref[1]+(0.362417530484)*x_ref[2]**o+(0.676820875084)*x_ref[2]
        ref[3,4]=(-0.53273881792)*x_ref[0]**o+(-0.0624208514424)*x_ref[0]+(0.680486778939)*x_ref[1]**o+(-0.370666453954)*x_ref[1]+(0.294345859395)*x_ref[2]**o+(-0.679169213305)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0906971676514)*x[0]**o+(-0.0640252128559)*x[0]+(-0.0108470120234)*x[1]**o+(-0.62531206416)*x[1]
        arg[0,0,1]=(-0.0186703191671)*x[0]**o+(0.256044044276)*x[0]+(-0.420130556042)*x[1]**o+(-0.241203381189)*x[1]
        arg[0,1,0]=(-0.287587258734)*x[0]**o+(-0.152487596827)*x[0]+(-0.554975489568)*x[1]**o+(0.713650618367)*x[1]
        arg[0,1,1]=(0.467597882001)*x[0]**o+(0.617291271522)*x[0]+(0.839893220642)*x[1]**o+(0.477553624935)*x[1]
        arg[1,0,0]=(0.292433268891)*x[0]**o+(0.0420372671524)*x[0]+(-0.342734694042)*x[1]**o+(0.892580432212)*x[1]
        arg[1,0,1]=(0.191399101606)*x[0]**o+(0.413781516804)*x[0]+(-0.346483382584)*x[1]**o+(0.805718474386)*x[1]
        arg[1,1,0]=(0.917644938285)*x[0]**o+(0.60457272962)*x[0]+(-0.0949129141155)*x[1]**o+(0.705971564397)*x[1]
        arg[1,1,1]=(0.327738300229)*x[0]**o+(-0.738951673731)*x[0]+(0.335869509312)*x[1]**o+(0.392167761004)*x[1]
        arg[2,0,0]=(0.460678533261)*x[0]**o+(-0.10665029868)*x[0]+(0.658148262201)*x[1]**o+(-0.75975084995)*x[1]
        arg[2,0,1]=(0.307057582515)*x[0]**o+(0.0755776536138)*x[0]+(0.866743758656)*x[1]**o+(-0.463537240812)*x[1]
        arg[2,1,0]=(-0.470241867396)*x[0]**o+(-0.0316070952691)*x[0]+(0.877786692055)*x[1]**o+(-0.750713976717)*x[1]
        arg[2,1,1]=(0.755903505788)*x[0]**o+(0.398472877018)*x[0]+(-0.372159129627)*x[1]**o+(0.170870288116)*x[1]
        arg[3,0,0]=(0.0310115714805)*x[0]**o+(0.822261427932)*x[0]+(-0.0979068243755)*x[1]**o+(0.61848414895)*x[1]
        arg[3,0,1]=(-0.442666645023)*x[0]**o+(0.759423291727)*x[0]+(0.771826722418)*x[1]**o+(-0.342283533203)*x[1]
        arg[3,1,0]=(0.292608821995)*x[0]**o+(0.301553824784)*x[0]+(0.445054135435)*x[1]**o+(0.990249711651)*x[1]
        arg[3,1,1]=(0.835714722542)*x[0]**o+(0.888569910076)*x[0]+(0.76442321847)*x[1]**o+(-0.272880995105)*x[1]
        arg[4,0,0]=(0.434738367563)*x[0]**o+(-0.801063604881)*x[0]+(-0.0152782263323)*x[1]**o+(-0.309467614762)*x[1]
        arg[4,0,1]=(-0.647140151015)*x[0]**o+(-0.0810020795911)*x[0]+(0.356402263616)*x[1]**o+(0.861692124435)*x[1]
        arg[4,1,0]=(-0.958501155541)*x[0]**o+(0.611720259342)*x[0]+(-0.631450146062)*x[1]**o+(-0.512668667768)*x[1]
        arg[4,1,1]=(0.771988547023)*x[0]**o+(0.466474334634)*x[0]+(-0.380283529607)*x[1]**o+(-0.619497160805)*x[1]
        arg[5,0,0]=(0.427386565443)*x[0]**o+(-0.178637075872)*x[0]+(0.81988184595)*x[1]**o+(-0.407798714033)*x[1]
        arg[5,0,1]=(-0.850161084571)*x[0]**o+(0.183440360862)*x[0]+(-0.83848298192)*x[1]**o+(-0.222248160259)*x[1]
        arg[5,1,0]=(-0.930623680879)*x[0]**o+(0.694991147173)*x[0]+(-0.299904374239)*x[1]**o+(0.552324198856)*x[1]
        arg[5,1,1]=(0.740715083219)*x[0]**o+(0.438768785937)*x[0]+(0.977523774483)*x[1]**o+(-0.668723413125)*x[1]
        ref[0,0,0]=(-0.0906971676514)*x_ref[0]**o+(-0.0640252128559)*x_ref[0]+(-0.0108470120234)*x_ref[1]**o+(-0.62531206416)*x_ref[1]
        ref[0,0,1]=(-0.0186703191671)*x_ref[0]**o+(0.256044044276)*x_ref[0]+(-0.420130556042)*x_ref[1]**o+(-0.241203381189)*x_ref[1]
        ref[0,1,0]=(-0.287587258734)*x_ref[0]**o+(-0.152487596827)*x_ref[0]+(-0.554975489568)*x_ref[1]**o+(0.713650618367)*x_ref[1]
        ref[0,1,1]=(0.467597882001)*x_ref[0]**o+(0.617291271522)*x_ref[0]+(0.839893220642)*x_ref[1]**o+(0.477553624935)*x_ref[1]
        ref[1,0,0]=(0.292433268891)*x_ref[0]**o+(0.0420372671524)*x_ref[0]+(-0.342734694042)*x_ref[1]**o+(0.892580432212)*x_ref[1]
        ref[1,0,1]=(0.191399101606)*x_ref[0]**o+(0.413781516804)*x_ref[0]+(-0.346483382584)*x_ref[1]**o+(0.805718474386)*x_ref[1]
        ref[1,1,0]=(0.917644938285)*x_ref[0]**o+(0.60457272962)*x_ref[0]+(-0.0949129141155)*x_ref[1]**o+(0.705971564397)*x_ref[1]
        ref[1,1,1]=(0.327738300229)*x_ref[0]**o+(-0.738951673731)*x_ref[0]+(0.335869509312)*x_ref[1]**o+(0.392167761004)*x_ref[1]
        ref[2,0,0]=(0.460678533261)*x_ref[0]**o+(-0.10665029868)*x_ref[0]+(0.658148262201)*x_ref[1]**o+(-0.75975084995)*x_ref[1]
        ref[2,0,1]=(0.307057582515)*x_ref[0]**o+(0.0755776536138)*x_ref[0]+(0.866743758656)*x_ref[1]**o+(-0.463537240812)*x_ref[1]
        ref[2,1,0]=(-0.470241867396)*x_ref[0]**o+(-0.0316070952691)*x_ref[0]+(0.877786692055)*x_ref[1]**o+(-0.750713976717)*x_ref[1]
        ref[2,1,1]=(0.755903505788)*x_ref[0]**o+(0.398472877018)*x_ref[0]+(-0.372159129627)*x_ref[1]**o+(0.170870288116)*x_ref[1]
        ref[3,0,0]=(0.0310115714805)*x_ref[0]**o+(0.822261427932)*x_ref[0]+(-0.0979068243755)*x_ref[1]**o+(0.61848414895)*x_ref[1]
        ref[3,0,1]=(-0.442666645023)*x_ref[0]**o+(0.759423291727)*x_ref[0]+(0.771826722418)*x_ref[1]**o+(-0.342283533203)*x_ref[1]
        ref[3,1,0]=(0.292608821995)*x_ref[0]**o+(0.301553824784)*x_ref[0]+(0.445054135435)*x_ref[1]**o+(0.990249711651)*x_ref[1]
        ref[3,1,1]=(0.835714722542)*x_ref[0]**o+(0.888569910076)*x_ref[0]+(0.76442321847)*x_ref[1]**o+(-0.272880995105)*x_ref[1]
        ref[4,0,0]=(0.434738367563)*x_ref[0]**o+(-0.801063604881)*x_ref[0]+(-0.0152782263323)*x_ref[1]**o+(-0.309467614762)*x_ref[1]
        ref[4,0,1]=(-0.647140151015)*x_ref[0]**o+(-0.0810020795911)*x_ref[0]+(0.356402263616)*x_ref[1]**o+(0.861692124435)*x_ref[1]
        ref[4,1,0]=(-0.958501155541)*x_ref[0]**o+(0.611720259342)*x_ref[0]+(-0.631450146062)*x_ref[1]**o+(-0.512668667768)*x_ref[1]
        ref[4,1,1]=(0.771988547023)*x_ref[0]**o+(0.466474334634)*x_ref[0]+(-0.380283529607)*x_ref[1]**o+(-0.619497160805)*x_ref[1]
        ref[5,0,0]=(0.427386565443)*x_ref[0]**o+(-0.178637075872)*x_ref[0]+(0.81988184595)*x_ref[1]**o+(-0.407798714033)*x_ref[1]
        ref[5,0,1]=(-0.850161084571)*x_ref[0]**o+(0.183440360862)*x_ref[0]+(-0.83848298192)*x_ref[1]**o+(-0.222248160259)*x_ref[1]
        ref[5,1,0]=(-0.930623680879)*x_ref[0]**o+(0.694991147173)*x_ref[0]+(-0.299904374239)*x_ref[1]**o+(0.552324198856)*x_ref[1]
        ref[5,1,1]=(0.740715083219)*x_ref[0]**o+(0.438768785937)*x_ref[0]+(0.977523774483)*x_ref[1]**o+(-0.668723413125)*x_ref[1]
      else:
        arg[0,0,0]=(0.0408197213005)*x[0]**o+(-0.992920054273)*x[0]+(0.122468233435)*x[1]**o+(-0.435942398199)*x[1]+(0.822235014621)*x[2]**o+(0.425925287623)*x[2]
        arg[0,0,1]=(0.474354220259)*x[0]**o+(0.287892840646)*x[0]+(-0.519818501933)*x[1]**o+(0.412301993774)*x[1]+(-0.242320576454)*x[2]**o+(0.758556860596)*x[2]
        arg[0,1,0]=(0.588735088154)*x[0]**o+(-0.948279493327)*x[0]+(0.194404465814)*x[1]**o+(-0.00761227943336)*x[1]+(-0.117173968067)*x[2]**o+(-0.228746139749)*x[2]
        arg[0,1,1]=(0.166448081217)*x[0]**o+(-0.228754542424)*x[0]+(-0.87739172573)*x[1]**o+(-0.592282861607)*x[1]+(-0.836732991313)*x[2]**o+(-0.20226535214)*x[2]
        arg[1,0,0]=(0.2683388552)*x[0]**o+(-0.733516131653)*x[0]+(-0.900887340495)*x[1]**o+(-0.614908158304)*x[1]+(0.422189467695)*x[2]**o+(0.642276887435)*x[2]
        arg[1,0,1]=(0.965294152134)*x[0]**o+(-0.326731862045)*x[0]+(-0.797754723728)*x[1]**o+(-0.678993698958)*x[1]+(0.256815891879)*x[2]**o+(0.48610716642)*x[2]
        arg[1,1,0]=(0.71383480118)*x[0]**o+(-0.949206151211)*x[0]+(0.40802633055)*x[1]**o+(-0.512457451544)*x[1]+(0.83647020826)*x[2]**o+(0.330550196742)*x[2]
        arg[1,1,1]=(-0.108602934331)*x[0]**o+(-0.585532075482)*x[0]+(0.302790940309)*x[1]**o+(-0.201628116557)*x[1]+(0.523440894937)*x[2]**o+(-0.421867218852)*x[2]
        arg[2,0,0]=(-0.911096130025)*x[0]**o+(-0.944970520542)*x[0]+(0.597179186191)*x[1]**o+(0.530539729209)*x[1]+(-0.538576718159)*x[2]**o+(-0.908863106348)*x[2]
        arg[2,0,1]=(0.828174522282)*x[0]**o+(-0.132534369339)*x[0]+(0.851849613713)*x[1]**o+(0.191494183049)*x[1]+(-0.953488097214)*x[2]**o+(0.423553741843)*x[2]
        arg[2,1,0]=(0.458631350156)*x[0]**o+(0.21755431391)*x[0]+(-0.128541259216)*x[1]**o+(-0.172831640065)*x[1]+(-0.62567145453)*x[2]**o+(-0.345626622399)*x[2]
        arg[2,1,1]=(0.49890508395)*x[0]**o+(-0.395084188812)*x[0]+(-0.653164498387)*x[1]**o+(-0.656416446654)*x[1]+(-0.844855683273)*x[2]**o+(-0.813755115861)*x[2]
        arg[3,0,0]=(-0.252710828698)*x[0]**o+(-0.0100429888666)*x[0]+(0.952930817098)*x[1]**o+(-0.950777038172)*x[1]+(0.244588721191)*x[2]**o+(-0.400442332557)*x[2]
        arg[3,0,1]=(-0.505331423555)*x[0]**o+(-0.535163478725)*x[0]+(0.924129836434)*x[1]**o+(-0.285813201016)*x[1]+(0.225037301542)*x[2]**o+(-0.331424134564)*x[2]
        arg[3,1,0]=(-0.556397106971)*x[0]**o+(0.26005371138)*x[0]+(0.902941320217)*x[1]**o+(0.70375648044)*x[1]+(-0.559371940837)*x[2]**o+(-0.628051748121)*x[2]
        arg[3,1,1]=(-0.644440017615)*x[0]**o+(-0.13415915731)*x[0]+(-0.752973804206)*x[1]**o+(-0.430304389409)*x[1]+(0.892979644749)*x[2]**o+(0.0618885081881)*x[2]
        arg[4,0,0]=(0.388902949971)*x[0]**o+(-0.413207160819)*x[0]+(-0.597892565551)*x[1]**o+(-0.433017605538)*x[1]+(0.369120396114)*x[2]**o+(-0.714185662873)*x[2]
        arg[4,0,1]=(0.140621693105)*x[0]**o+(0.910612147355)*x[0]+(0.404844175588)*x[1]**o+(0.963351416048)*x[1]+(0.545142156468)*x[2]**o+(-0.513154196146)*x[2]
        arg[4,1,0]=(-0.6045448854)*x[0]**o+(0.364599037279)*x[0]+(-0.165333666972)*x[1]**o+(-0.454715984949)*x[1]+(-0.123749300084)*x[2]**o+(0.23666517892)*x[2]
        arg[4,1,1]=(-0.701819332028)*x[0]**o+(-0.945269094128)*x[0]+(0.313745376139)*x[1]**o+(-0.0508160648325)*x[1]+(0.273150632944)*x[2]**o+(0.363230489878)*x[2]
        arg[5,0,0]=(-0.899477778287)*x[0]**o+(0.408300340496)*x[0]+(-0.501090863132)*x[1]**o+(-0.142172524799)*x[1]+(0.840930927028)*x[2]**o+(0.0786174984564)*x[2]
        arg[5,0,1]=(0.816863720765)*x[0]**o+(0.421613919321)*x[0]+(0.619636513942)*x[1]**o+(-0.867504595488)*x[1]+(-0.377345269009)*x[2]**o+(-0.884630348781)*x[2]
        arg[5,1,0]=(0.972307172208)*x[0]**o+(0.704673083035)*x[0]+(0.546566839332)*x[1]**o+(-0.303145527888)*x[1]+(-0.203985539635)*x[2]**o+(0.984097584463)*x[2]
        arg[5,1,1]=(0.67971485539)*x[0]**o+(0.492284313707)*x[0]+(0.139113989973)*x[1]**o+(-0.981414708174)*x[1]+(0.585926444541)*x[2]**o+(-0.790952419012)*x[2]
        ref[0,0,0]=(0.0408197213005)*x_ref[0]**o+(-0.992920054273)*x_ref[0]+(0.122468233435)*x_ref[1]**o+(-0.435942398199)*x_ref[1]+(0.822235014621)*x_ref[2]**o+(0.425925287623)*x_ref[2]
        ref[0,0,1]=(0.474354220259)*x_ref[0]**o+(0.287892840646)*x_ref[0]+(-0.519818501933)*x_ref[1]**o+(0.412301993774)*x_ref[1]+(-0.242320576454)*x_ref[2]**o+(0.758556860596)*x_ref[2]
        ref[0,1,0]=(0.588735088154)*x_ref[0]**o+(-0.948279493327)*x_ref[0]+(0.194404465814)*x_ref[1]**o+(-0.00761227943336)*x_ref[1]+(-0.117173968067)*x_ref[2]**o+(-0.228746139749)*x_ref[2]
        ref[0,1,1]=(0.166448081217)*x_ref[0]**o+(-0.228754542424)*x_ref[0]+(-0.87739172573)*x_ref[1]**o+(-0.592282861607)*x_ref[1]+(-0.836732991313)*x_ref[2]**o+(-0.20226535214)*x_ref[2]
        ref[1,0,0]=(0.2683388552)*x_ref[0]**o+(-0.733516131653)*x_ref[0]+(-0.900887340495)*x_ref[1]**o+(-0.614908158304)*x_ref[1]+(0.422189467695)*x_ref[2]**o+(0.642276887435)*x_ref[2]
        ref[1,0,1]=(0.965294152134)*x_ref[0]**o+(-0.326731862045)*x_ref[0]+(-0.797754723728)*x_ref[1]**o+(-0.678993698958)*x_ref[1]+(0.256815891879)*x_ref[2]**o+(0.48610716642)*x_ref[2]
        ref[1,1,0]=(0.71383480118)*x_ref[0]**o+(-0.949206151211)*x_ref[0]+(0.40802633055)*x_ref[1]**o+(-0.512457451544)*x_ref[1]+(0.83647020826)*x_ref[2]**o+(0.330550196742)*x_ref[2]
        ref[1,1,1]=(-0.108602934331)*x_ref[0]**o+(-0.585532075482)*x_ref[0]+(0.302790940309)*x_ref[1]**o+(-0.201628116557)*x_ref[1]+(0.523440894937)*x_ref[2]**o+(-0.421867218852)*x_ref[2]
        ref[2,0,0]=(-0.911096130025)*x_ref[0]**o+(-0.944970520542)*x_ref[0]+(0.597179186191)*x_ref[1]**o+(0.530539729209)*x_ref[1]+(-0.538576718159)*x_ref[2]**o+(-0.908863106348)*x_ref[2]
        ref[2,0,1]=(0.828174522282)*x_ref[0]**o+(-0.132534369339)*x_ref[0]+(0.851849613713)*x_ref[1]**o+(0.191494183049)*x_ref[1]+(-0.953488097214)*x_ref[2]**o+(0.423553741843)*x_ref[2]
        ref[2,1,0]=(0.458631350156)*x_ref[0]**o+(0.21755431391)*x_ref[0]+(-0.128541259216)*x_ref[1]**o+(-0.172831640065)*x_ref[1]+(-0.62567145453)*x_ref[2]**o+(-0.345626622399)*x_ref[2]
        ref[2,1,1]=(0.49890508395)*x_ref[0]**o+(-0.395084188812)*x_ref[0]+(-0.653164498387)*x_ref[1]**o+(-0.656416446654)*x_ref[1]+(-0.844855683273)*x_ref[2]**o+(-0.813755115861)*x_ref[2]
        ref[3,0,0]=(-0.252710828698)*x_ref[0]**o+(-0.0100429888666)*x_ref[0]+(0.952930817098)*x_ref[1]**o+(-0.950777038172)*x_ref[1]+(0.244588721191)*x_ref[2]**o+(-0.400442332557)*x_ref[2]
        ref[3,0,1]=(-0.505331423555)*x_ref[0]**o+(-0.535163478725)*x_ref[0]+(0.924129836434)*x_ref[1]**o+(-0.285813201016)*x_ref[1]+(0.225037301542)*x_ref[2]**o+(-0.331424134564)*x_ref[2]
        ref[3,1,0]=(-0.556397106971)*x_ref[0]**o+(0.26005371138)*x_ref[0]+(0.902941320217)*x_ref[1]**o+(0.70375648044)*x_ref[1]+(-0.559371940837)*x_ref[2]**o+(-0.628051748121)*x_ref[2]
        ref[3,1,1]=(-0.644440017615)*x_ref[0]**o+(-0.13415915731)*x_ref[0]+(-0.752973804206)*x_ref[1]**o+(-0.430304389409)*x_ref[1]+(0.892979644749)*x_ref[2]**o+(0.0618885081881)*x_ref[2]
        ref[4,0,0]=(0.388902949971)*x_ref[0]**o+(-0.413207160819)*x_ref[0]+(-0.597892565551)*x_ref[1]**o+(-0.433017605538)*x_ref[1]+(0.369120396114)*x_ref[2]**o+(-0.714185662873)*x_ref[2]
        ref[4,0,1]=(0.140621693105)*x_ref[0]**o+(0.910612147355)*x_ref[0]+(0.404844175588)*x_ref[1]**o+(0.963351416048)*x_ref[1]+(0.545142156468)*x_ref[2]**o+(-0.513154196146)*x_ref[2]
        ref[4,1,0]=(-0.6045448854)*x_ref[0]**o+(0.364599037279)*x_ref[0]+(-0.165333666972)*x_ref[1]**o+(-0.454715984949)*x_ref[1]+(-0.123749300084)*x_ref[2]**o+(0.23666517892)*x_ref[2]
        ref[4,1,1]=(-0.701819332028)*x_ref[0]**o+(-0.945269094128)*x_ref[0]+(0.313745376139)*x_ref[1]**o+(-0.0508160648325)*x_ref[1]+(0.273150632944)*x_ref[2]**o+(0.363230489878)*x_ref[2]
        ref[5,0,0]=(-0.899477778287)*x_ref[0]**o+(0.408300340496)*x_ref[0]+(-0.501090863132)*x_ref[1]**o+(-0.142172524799)*x_ref[1]+(0.840930927028)*x_ref[2]**o+(0.0786174984564)*x_ref[2]
        ref[5,0,1]=(0.816863720765)*x_ref[0]**o+(0.421613919321)*x_ref[0]+(0.619636513942)*x_ref[1]**o+(-0.867504595488)*x_ref[1]+(-0.377345269009)*x_ref[2]**o+(-0.884630348781)*x_ref[2]
        ref[5,1,0]=(0.972307172208)*x_ref[0]**o+(0.704673083035)*x_ref[0]+(0.546566839332)*x_ref[1]**o+(-0.303145527888)*x_ref[1]+(-0.203985539635)*x_ref[2]**o+(0.984097584463)*x_ref[2]
        ref[5,1,1]=(0.67971485539)*x_ref[0]**o+(0.492284313707)*x_ref[0]+(0.139113989973)*x_ref[1]**o+(-0.981414708174)*x_ref[1]+(0.585926444541)*x_ref[2]**o+(-0.790952419012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.717741572212)*x[0]**o+(-0.771801794802)*x[0]+(-0.782201793954)*x[1]**o+(0.125253178673)*x[1]
        arg[0,0,0,1]=(-0.483339499341)*x[0]**o+(0.0177945970437)*x[0]+(0.233532060933)*x[1]**o+(-0.465948890703)*x[1]
        arg[0,0,1,0]=(0.140544455071)*x[0]**o+(-0.120145399579)*x[0]+(-0.0732045826929)*x[1]**o+(-0.738165902415)*x[1]
        arg[0,0,1,1]=(0.98057407477)*x[0]**o+(0.346125329427)*x[0]+(0.697386820108)*x[1]**o+(0.402406370862)*x[1]
        arg[0,0,2,0]=(0.347641628204)*x[0]**o+(-0.217938924727)*x[0]+(-0.0572727430966)*x[1]**o+(0.170631149745)*x[1]
        arg[0,0,2,1]=(0.621840702108)*x[0]**o+(-0.109219941839)*x[0]+(0.894052962811)*x[1]**o+(0.423894680808)*x[1]
        arg[0,1,0,0]=(0.0339315069567)*x[0]**o+(-0.655546201957)*x[0]+(-0.271776247317)*x[1]**o+(0.4760527815)*x[1]
        arg[0,1,0,1]=(0.904216237912)*x[0]**o+(0.974869309983)*x[0]+(0.678379481202)*x[1]**o+(0.602534020957)*x[1]
        arg[0,1,1,0]=(-0.98805028173)*x[0]**o+(0.255251214419)*x[0]+(0.659010647624)*x[1]**o+(-0.45360685971)*x[1]
        arg[0,1,1,1]=(0.0884339867328)*x[0]**o+(-0.325188714625)*x[0]+(0.158562643552)*x[1]**o+(-0.788600388567)*x[1]
        arg[0,1,2,0]=(0.976404407593)*x[0]**o+(0.353231650963)*x[0]+(0.546951413477)*x[1]**o+(0.0738098551899)*x[1]
        arg[0,1,2,1]=(-0.812156797654)*x[0]**o+(-0.0180014270879)*x[0]+(0.686011517957)*x[1]**o+(0.148522434365)*x[1]
        arg[0,2,0,0]=(-0.40018516407)*x[0]**o+(-0.0512695503575)*x[0]+(-0.715512784143)*x[1]**o+(0.111064716031)*x[1]
        arg[0,2,0,1]=(-0.963397739298)*x[0]**o+(0.406951519848)*x[0]+(-0.293539028967)*x[1]**o+(-0.549328925146)*x[1]
        arg[0,2,1,0]=(-0.00827402449089)*x[0]**o+(-0.657982352244)*x[0]+(-0.515966831052)*x[1]**o+(0.531853371259)*x[1]
        arg[0,2,1,1]=(0.0379862601491)*x[0]**o+(0.988691147833)*x[0]+(0.820623911484)*x[1]**o+(0.689064076022)*x[1]
        arg[0,2,2,0]=(-0.857638654259)*x[0]**o+(0.664968676908)*x[0]+(-0.228424264597)*x[1]**o+(0.740003679791)*x[1]
        arg[0,2,2,1]=(0.456486069586)*x[0]**o+(-0.510130176146)*x[0]+(-0.189525160745)*x[1]**o+(0.497975981413)*x[1]
        arg[0,3,0,0]=(-0.981633737105)*x[0]**o+(0.25636883881)*x[0]+(-0.0477676975467)*x[1]**o+(0.654475880014)*x[1]
        arg[0,3,0,1]=(-0.0710284929969)*x[0]**o+(0.338590765191)*x[0]+(0.639048913997)*x[1]**o+(-0.24999606762)*x[1]
        arg[0,3,1,0]=(-0.731484482659)*x[0]**o+(-0.808253846389)*x[0]+(-0.911723431862)*x[1]**o+(-0.266029942425)*x[1]
        arg[0,3,1,1]=(0.909037857376)*x[0]**o+(-0.164525039433)*x[0]+(-0.924332370345)*x[1]**o+(0.476479399362)*x[1]
        arg[0,3,2,0]=(0.00534173833653)*x[0]**o+(-0.435160515142)*x[0]+(0.402701511687)*x[1]**o+(0.161631276653)*x[1]
        arg[0,3,2,1]=(-0.595791669027)*x[0]**o+(0.170580206948)*x[0]+(0.468802250813)*x[1]**o+(-0.881472636722)*x[1]
        arg[0,4,0,0]=(-0.0739506349513)*x[0]**o+(0.513385505957)*x[0]+(0.046977740371)*x[1]**o+(0.95763159987)*x[1]
        arg[0,4,0,1]=(-0.286789006223)*x[0]**o+(0.017476293519)*x[0]+(0.637378781092)*x[1]**o+(-0.670201122126)*x[1]
        arg[0,4,1,0]=(0.159607100455)*x[0]**o+(0.715341910227)*x[0]+(0.949546645819)*x[1]**o+(0.362982182775)*x[1]
        arg[0,4,1,1]=(-0.237702024358)*x[0]**o+(-0.520488841454)*x[0]+(0.606333933569)*x[1]**o+(-0.00662757701049)*x[1]
        arg[0,4,2,0]=(-0.415753817901)*x[0]**o+(0.731285464784)*x[0]+(0.15987336144)*x[1]**o+(0.0697706236338)*x[1]
        arg[0,4,2,1]=(0.0467482158427)*x[0]**o+(0.306004722736)*x[0]+(-0.905184725692)*x[1]**o+(-0.158017889366)*x[1]
        arg[1,0,0,0]=(-0.785535082426)*x[0]**o+(-0.189285100452)*x[0]+(-0.766349940057)*x[1]**o+(0.531469308995)*x[1]
        arg[1,0,0,1]=(0.522213196477)*x[0]**o+(-0.661381741437)*x[0]+(0.816089886956)*x[1]**o+(0.139829382305)*x[1]
        arg[1,0,1,0]=(0.0957295570417)*x[0]**o+(-0.105252674536)*x[0]+(-0.569429953382)*x[1]**o+(-0.801934666657)*x[1]
        arg[1,0,1,1]=(-0.557307051058)*x[0]**o+(0.218852576416)*x[0]+(0.430167216464)*x[1]**o+(-0.112285681761)*x[1]
        arg[1,0,2,0]=(0.288472422477)*x[0]**o+(0.294377994763)*x[0]+(-0.809503961511)*x[1]**o+(0.357020226991)*x[1]
        arg[1,0,2,1]=(-0.148113477705)*x[0]**o+(0.164468646294)*x[0]+(0.7977142476)*x[1]**o+(-0.407274933384)*x[1]
        arg[1,1,0,0]=(-0.926374434822)*x[0]**o+(-0.107713463193)*x[0]+(-0.511635723869)*x[1]**o+(0.259319717689)*x[1]
        arg[1,1,0,1]=(-0.18636582243)*x[0]**o+(0.698258973628)*x[0]+(0.234379281397)*x[1]**o+(0.596195875495)*x[1]
        arg[1,1,1,0]=(-0.180395892006)*x[0]**o+(0.631248500084)*x[0]+(0.104751616377)*x[1]**o+(0.204500206653)*x[1]
        arg[1,1,1,1]=(-0.517119564238)*x[0]**o+(-0.16634015634)*x[0]+(0.556785586675)*x[1]**o+(0.42659306392)*x[1]
        arg[1,1,2,0]=(0.163807238958)*x[0]**o+(-0.534619478446)*x[0]+(-0.700353460753)*x[1]**o+(0.121393131648)*x[1]
        arg[1,1,2,1]=(0.964276278961)*x[0]**o+(0.846030751304)*x[0]+(-0.745301894392)*x[1]**o+(0.33821485558)*x[1]
        arg[1,2,0,0]=(0.333173101086)*x[0]**o+(0.774848048743)*x[0]+(-0.659091916357)*x[1]**o+(-0.797247144497)*x[1]
        arg[1,2,0,1]=(0.0581394534285)*x[0]**o+(-0.524771663031)*x[0]+(-0.519465753317)*x[1]**o+(0.71748795226)*x[1]
        arg[1,2,1,0]=(0.224935262833)*x[0]**o+(-0.393399819306)*x[0]+(0.199349075224)*x[1]**o+(-0.808847009412)*x[1]
        arg[1,2,1,1]=(0.0385750795521)*x[0]**o+(-8.09949553151e-05)*x[0]+(0.496860964506)*x[1]**o+(0.0119507407829)*x[1]
        arg[1,2,2,0]=(0.876341556167)*x[0]**o+(0.18082766372)*x[0]+(0.401349871827)*x[1]**o+(-0.543439037303)*x[1]
        arg[1,2,2,1]=(-0.894525004507)*x[0]**o+(-0.516771143695)*x[0]+(0.0410916490263)*x[1]**o+(0.753841618439)*x[1]
        arg[1,3,0,0]=(0.251979970924)*x[0]**o+(0.790276550438)*x[0]+(-0.155708793888)*x[1]**o+(0.28971278336)*x[1]
        arg[1,3,0,1]=(0.258336265436)*x[0]**o+(0.696642875149)*x[0]+(0.0561355915649)*x[1]**o+(0.601839744417)*x[1]
        arg[1,3,1,0]=(-0.978729102785)*x[0]**o+(0.335280613262)*x[0]+(0.89228798232)*x[1]**o+(-0.693050679348)*x[1]
        arg[1,3,1,1]=(-0.885817753431)*x[0]**o+(-0.0641571535708)*x[0]+(-0.649578576181)*x[1]**o+(-0.95773137602)*x[1]
        arg[1,3,2,0]=(-0.333891001461)*x[0]**o+(-0.496064308564)*x[0]+(0.438549603377)*x[1]**o+(0.0772468142672)*x[1]
        arg[1,3,2,1]=(-0.16171788255)*x[0]**o+(-0.939293453222)*x[0]+(-0.284522648546)*x[1]**o+(0.744642328375)*x[1]
        arg[1,4,0,0]=(0.806133262332)*x[0]**o+(-0.0225320206432)*x[0]+(0.125958642606)*x[1]**o+(0.610630249467)*x[1]
        arg[1,4,0,1]=(0.657971870946)*x[0]**o+(0.698894701525)*x[0]+(-0.595952612496)*x[1]**o+(-0.38268841292)*x[1]
        arg[1,4,1,0]=(-0.277274411331)*x[0]**o+(-0.483183004514)*x[0]+(0.77656657009)*x[1]**o+(-0.959290466645)*x[1]
        arg[1,4,1,1]=(-0.96384356305)*x[0]**o+(-0.00383301781476)*x[0]+(0.788010409608)*x[1]**o+(0.135378744316)*x[1]
        arg[1,4,2,0]=(0.986761958017)*x[0]**o+(-0.570276378038)*x[0]+(-0.515784028315)*x[1]**o+(0.487125123439)*x[1]
        arg[1,4,2,1]=(-0.954735054702)*x[0]**o+(-0.0610323074217)*x[0]+(-0.255284859281)*x[1]**o+(0.277795725698)*x[1]
        arg[2,0,0,0]=(-0.708301219324)*x[0]**o+(-0.970166781553)*x[0]+(-0.0836451425855)*x[1]**o+(0.525854202995)*x[1]
        arg[2,0,0,1]=(-0.0995393865183)*x[0]**o+(-0.0156624820324)*x[0]+(0.225529316197)*x[1]**o+(0.142617834056)*x[1]
        arg[2,0,1,0]=(-0.585329524103)*x[0]**o+(0.590285924441)*x[0]+(0.201483726322)*x[1]**o+(-0.735262197689)*x[1]
        arg[2,0,1,1]=(-0.682877182963)*x[0]**o+(-0.162682262799)*x[0]+(-0.890051459843)*x[1]**o+(-0.0188015866765)*x[1]
        arg[2,0,2,0]=(0.868935476294)*x[0]**o+(-0.629891068854)*x[0]+(0.335632865663)*x[1]**o+(-0.955849055716)*x[1]
        arg[2,0,2,1]=(-0.907933228238)*x[0]**o+(-0.105861987974)*x[0]+(-0.131372383042)*x[1]**o+(-0.970114136854)*x[1]
        arg[2,1,0,0]=(-0.275788501814)*x[0]**o+(-0.992861867033)*x[0]+(-0.80676324962)*x[1]**o+(0.269244680224)*x[1]
        arg[2,1,0,1]=(-0.704704138774)*x[0]**o+(-0.371803193009)*x[0]+(-0.0190586405922)*x[1]**o+(-0.0828837624698)*x[1]
        arg[2,1,1,0]=(-0.532645566013)*x[0]**o+(0.571754809868)*x[0]+(0.656737493773)*x[1]**o+(-0.548229591842)*x[1]
        arg[2,1,1,1]=(0.23836921087)*x[0]**o+(0.707475737785)*x[0]+(-0.59660155968)*x[1]**o+(-0.450204951071)*x[1]
        arg[2,1,2,0]=(-0.534027348014)*x[0]**o+(0.332100193727)*x[0]+(-0.864157138567)*x[1]**o+(-0.41833267066)*x[1]
        arg[2,1,2,1]=(0.459878593269)*x[0]**o+(0.446259140228)*x[0]+(0.571242887502)*x[1]**o+(-0.0390525669486)*x[1]
        arg[2,2,0,0]=(0.322052799527)*x[0]**o+(-0.915265209304)*x[0]+(-0.99942007618)*x[1]**o+(0.482294513743)*x[1]
        arg[2,2,0,1]=(0.558489194456)*x[0]**o+(0.00400576554086)*x[0]+(-0.87484330165)*x[1]**o+(-0.664504218977)*x[1]
        arg[2,2,1,0]=(0.669621607461)*x[0]**o+(-0.0450960554319)*x[0]+(-0.221151480246)*x[1]**o+(0.0123621737211)*x[1]
        arg[2,2,1,1]=(0.476743013029)*x[0]**o+(0.11118371822)*x[0]+(-0.202074315677)*x[1]**o+(0.377649433458)*x[1]
        arg[2,2,2,0]=(0.667910543701)*x[0]**o+(0.035703922547)*x[0]+(-0.274733967095)*x[1]**o+(-0.998934720182)*x[1]
        arg[2,2,2,1]=(0.578909660087)*x[0]**o+(0.149900086514)*x[0]+(-0.226597651858)*x[1]**o+(0.169175812287)*x[1]
        arg[2,3,0,0]=(0.158370246586)*x[0]**o+(0.595643112285)*x[0]+(-0.476875564052)*x[1]**o+(0.869752657517)*x[1]
        arg[2,3,0,1]=(0.99718847714)*x[0]**o+(0.581531822297)*x[0]+(-0.607102775564)*x[1]**o+(-0.8317970284)*x[1]
        arg[2,3,1,0]=(-0.27390195745)*x[0]**o+(0.0512853330414)*x[0]+(0.465190666302)*x[1]**o+(0.042533388788)*x[1]
        arg[2,3,1,1]=(-0.183679504962)*x[0]**o+(-0.0522782076121)*x[0]+(0.921423118529)*x[1]**o+(-0.0315080275225)*x[1]
        arg[2,3,2,0]=(0.0421842348778)*x[0]**o+(0.462023049702)*x[0]+(-0.888342606803)*x[1]**o+(-0.956503447884)*x[1]
        arg[2,3,2,1]=(-0.950820288444)*x[0]**o+(0.104271990168)*x[0]+(-0.607212909582)*x[1]**o+(-0.16283931259)*x[1]
        arg[2,4,0,0]=(0.113505933533)*x[0]**o+(0.27873514806)*x[0]+(0.736699443523)*x[1]**o+(-0.951737555693)*x[1]
        arg[2,4,0,1]=(0.264170584849)*x[0]**o+(-0.226649236617)*x[0]+(0.489902180874)*x[1]**o+(-0.191536591762)*x[1]
        arg[2,4,1,0]=(0.712322285985)*x[0]**o+(-0.323395455815)*x[0]+(-0.298356681586)*x[1]**o+(0.699891974541)*x[1]
        arg[2,4,1,1]=(-0.631057763255)*x[0]**o+(-0.489374933445)*x[0]+(-0.021591806874)*x[1]**o+(-0.692800124585)*x[1]
        arg[2,4,2,0]=(0.916068935184)*x[0]**o+(0.33779688858)*x[0]+(0.06300443559)*x[1]**o+(-0.231385203399)*x[1]
        arg[2,4,2,1]=(-0.283485044625)*x[0]**o+(0.548366574442)*x[0]+(-0.543953132872)*x[1]**o+(-0.956662823037)*x[1]
        arg[3,0,0,0]=(0.709349274221)*x[0]**o+(-0.342981009263)*x[0]+(-0.682331761929)*x[1]**o+(0.673169884813)*x[1]
        arg[3,0,0,1]=(0.236249836593)*x[0]**o+(-0.682845054434)*x[0]+(-0.349169988218)*x[1]**o+(0.502684425824)*x[1]
        arg[3,0,1,0]=(-0.561655748793)*x[0]**o+(0.120183772326)*x[0]+(-0.966535325583)*x[1]**o+(-0.0585117264949)*x[1]
        arg[3,0,1,1]=(-0.661030287189)*x[0]**o+(-0.322924086505)*x[0]+(-0.986892059552)*x[1]**o+(-0.778969361365)*x[1]
        arg[3,0,2,0]=(-0.166307107307)*x[0]**o+(0.612931303617)*x[0]+(-0.283925165456)*x[1]**o+(0.5998038814)*x[1]
        arg[3,0,2,1]=(0.205099355916)*x[0]**o+(-0.520970536278)*x[0]+(0.900416833183)*x[1]**o+(0.981826306069)*x[1]
        arg[3,1,0,0]=(-0.875306666443)*x[0]**o+(0.0189358479861)*x[0]+(0.101187857445)*x[1]**o+(-0.873584307534)*x[1]
        arg[3,1,0,1]=(-0.0203656235757)*x[0]**o+(-0.579229777042)*x[0]+(-0.0799345631826)*x[1]**o+(0.896315444069)*x[1]
        arg[3,1,1,0]=(0.576791459982)*x[0]**o+(-0.86541815506)*x[0]+(-0.582764176207)*x[1]**o+(-0.249437395011)*x[1]
        arg[3,1,1,1]=(0.525953680497)*x[0]**o+(-0.222445717402)*x[0]+(-0.87710202063)*x[1]**o+(-0.762921821973)*x[1]
        arg[3,1,2,0]=(-0.883737185908)*x[0]**o+(0.254053241277)*x[0]+(0.559803298059)*x[1]**o+(-0.577376956856)*x[1]
        arg[3,1,2,1]=(-0.125792083387)*x[0]**o+(-0.771549307509)*x[0]+(0.142493480552)*x[1]**o+(0.459996962548)*x[1]
        arg[3,2,0,0]=(0.0531598927049)*x[0]**o+(-0.950343229037)*x[0]+(-0.748240453154)*x[1]**o+(0.850555892342)*x[1]
        arg[3,2,0,1]=(-0.660852152041)*x[0]**o+(-0.763119278604)*x[0]+(-0.811925409663)*x[1]**o+(-0.178765071495)*x[1]
        arg[3,2,1,0]=(-0.189124223305)*x[0]**o+(-0.410392909042)*x[0]+(0.993247167522)*x[1]**o+(0.291451796298)*x[1]
        arg[3,2,1,1]=(0.83496133421)*x[0]**o+(0.695147480265)*x[0]+(0.444420524946)*x[1]**o+(0.107454620114)*x[1]
        arg[3,2,2,0]=(0.94687479313)*x[0]**o+(-0.140702324395)*x[0]+(0.149203502428)*x[1]**o+(-0.338246126345)*x[1]
        arg[3,2,2,1]=(0.147609394709)*x[0]**o+(-0.549219765239)*x[0]+(0.397308993347)*x[1]**o+(0.0402189312685)*x[1]
        arg[3,3,0,0]=(0.27400981409)*x[0]**o+(-0.516367266257)*x[0]+(-0.784095952814)*x[1]**o+(0.32858899009)*x[1]
        arg[3,3,0,1]=(-0.393457382163)*x[0]**o+(0.590406794051)*x[0]+(0.328523918357)*x[1]**o+(0.899537513355)*x[1]
        arg[3,3,1,0]=(-0.851038623366)*x[0]**o+(0.596974864726)*x[0]+(-0.338630687551)*x[1]**o+(0.634961435042)*x[1]
        arg[3,3,1,1]=(0.569166143918)*x[0]**o+(-0.388162282491)*x[0]+(0.764830536733)*x[1]**o+(-0.426737342542)*x[1]
        arg[3,3,2,0]=(0.502221903448)*x[0]**o+(0.940382225897)*x[0]+(0.563444730579)*x[1]**o+(0.023210869151)*x[1]
        arg[3,3,2,1]=(0.061695403852)*x[0]**o+(0.868302188758)*x[0]+(0.0855803325758)*x[1]**o+(-0.24366173638)*x[1]
        arg[3,4,0,0]=(-0.533038182387)*x[0]**o+(0.0222388915737)*x[0]+(0.529037752769)*x[1]**o+(-0.0665320895494)*x[1]
        arg[3,4,0,1]=(0.714374055253)*x[0]**o+(-0.623670743333)*x[0]+(-0.866024792929)*x[1]**o+(-0.0673124896111)*x[1]
        arg[3,4,1,0]=(-0.0523466615045)*x[0]**o+(-0.472563426622)*x[0]+(0.139892673626)*x[1]**o+(0.539655417178)*x[1]
        arg[3,4,1,1]=(-0.473032580514)*x[0]**o+(0.692812093618)*x[0]+(0.664400517742)*x[1]**o+(0.184125423364)*x[1]
        arg[3,4,2,0]=(-0.0340798058725)*x[0]**o+(-0.58900562863)*x[0]+(0.396856384717)*x[1]**o+(-0.559175428019)*x[1]
        arg[3,4,2,1]=(-0.217679376634)*x[0]**o+(-0.90389036667)*x[0]+(0.967372942873)*x[1]**o+(-0.482913619277)*x[1]
        ref[0,0,0,0]=(-0.717741572212)*x_ref[0]**o+(-0.771801794802)*x_ref[0]+(-0.782201793954)*x_ref[1]**o+(0.125253178673)*x_ref[1]
        ref[0,0,0,1]=(-0.483339499341)*x_ref[0]**o+(0.0177945970437)*x_ref[0]+(0.233532060933)*x_ref[1]**o+(-0.465948890703)*x_ref[1]
        ref[0,0,1,0]=(0.140544455071)*x_ref[0]**o+(-0.120145399579)*x_ref[0]+(-0.0732045826929)*x_ref[1]**o+(-0.738165902415)*x_ref[1]
        ref[0,0,1,1]=(0.98057407477)*x_ref[0]**o+(0.346125329427)*x_ref[0]+(0.697386820108)*x_ref[1]**o+(0.402406370862)*x_ref[1]
        ref[0,0,2,0]=(0.347641628204)*x_ref[0]**o+(-0.217938924727)*x_ref[0]+(-0.0572727430966)*x_ref[1]**o+(0.170631149745)*x_ref[1]
        ref[0,0,2,1]=(0.621840702108)*x_ref[0]**o+(-0.109219941839)*x_ref[0]+(0.894052962811)*x_ref[1]**o+(0.423894680808)*x_ref[1]
        ref[0,1,0,0]=(0.0339315069567)*x_ref[0]**o+(-0.655546201957)*x_ref[0]+(-0.271776247317)*x_ref[1]**o+(0.4760527815)*x_ref[1]
        ref[0,1,0,1]=(0.904216237912)*x_ref[0]**o+(0.974869309983)*x_ref[0]+(0.678379481202)*x_ref[1]**o+(0.602534020957)*x_ref[1]
        ref[0,1,1,0]=(-0.98805028173)*x_ref[0]**o+(0.255251214419)*x_ref[0]+(0.659010647624)*x_ref[1]**o+(-0.45360685971)*x_ref[1]
        ref[0,1,1,1]=(0.0884339867328)*x_ref[0]**o+(-0.325188714625)*x_ref[0]+(0.158562643552)*x_ref[1]**o+(-0.788600388567)*x_ref[1]
        ref[0,1,2,0]=(0.976404407593)*x_ref[0]**o+(0.353231650963)*x_ref[0]+(0.546951413477)*x_ref[1]**o+(0.0738098551899)*x_ref[1]
        ref[0,1,2,1]=(-0.812156797654)*x_ref[0]**o+(-0.0180014270879)*x_ref[0]+(0.686011517957)*x_ref[1]**o+(0.148522434365)*x_ref[1]
        ref[0,2,0,0]=(-0.40018516407)*x_ref[0]**o+(-0.0512695503575)*x_ref[0]+(-0.715512784143)*x_ref[1]**o+(0.111064716031)*x_ref[1]
        ref[0,2,0,1]=(-0.963397739298)*x_ref[0]**o+(0.406951519848)*x_ref[0]+(-0.293539028967)*x_ref[1]**o+(-0.549328925146)*x_ref[1]
        ref[0,2,1,0]=(-0.00827402449089)*x_ref[0]**o+(-0.657982352244)*x_ref[0]+(-0.515966831052)*x_ref[1]**o+(0.531853371259)*x_ref[1]
        ref[0,2,1,1]=(0.0379862601491)*x_ref[0]**o+(0.988691147833)*x_ref[0]+(0.820623911484)*x_ref[1]**o+(0.689064076022)*x_ref[1]
        ref[0,2,2,0]=(-0.857638654259)*x_ref[0]**o+(0.664968676908)*x_ref[0]+(-0.228424264597)*x_ref[1]**o+(0.740003679791)*x_ref[1]
        ref[0,2,2,1]=(0.456486069586)*x_ref[0]**o+(-0.510130176146)*x_ref[0]+(-0.189525160745)*x_ref[1]**o+(0.497975981413)*x_ref[1]
        ref[0,3,0,0]=(-0.981633737105)*x_ref[0]**o+(0.25636883881)*x_ref[0]+(-0.0477676975467)*x_ref[1]**o+(0.654475880014)*x_ref[1]
        ref[0,3,0,1]=(-0.0710284929969)*x_ref[0]**o+(0.338590765191)*x_ref[0]+(0.639048913997)*x_ref[1]**o+(-0.24999606762)*x_ref[1]
        ref[0,3,1,0]=(-0.731484482659)*x_ref[0]**o+(-0.808253846389)*x_ref[0]+(-0.911723431862)*x_ref[1]**o+(-0.266029942425)*x_ref[1]
        ref[0,3,1,1]=(0.909037857376)*x_ref[0]**o+(-0.164525039433)*x_ref[0]+(-0.924332370345)*x_ref[1]**o+(0.476479399362)*x_ref[1]
        ref[0,3,2,0]=(0.00534173833653)*x_ref[0]**o+(-0.435160515142)*x_ref[0]+(0.402701511687)*x_ref[1]**o+(0.161631276653)*x_ref[1]
        ref[0,3,2,1]=(-0.595791669027)*x_ref[0]**o+(0.170580206948)*x_ref[0]+(0.468802250813)*x_ref[1]**o+(-0.881472636722)*x_ref[1]
        ref[0,4,0,0]=(-0.0739506349513)*x_ref[0]**o+(0.513385505957)*x_ref[0]+(0.046977740371)*x_ref[1]**o+(0.95763159987)*x_ref[1]
        ref[0,4,0,1]=(-0.286789006223)*x_ref[0]**o+(0.017476293519)*x_ref[0]+(0.637378781092)*x_ref[1]**o+(-0.670201122126)*x_ref[1]
        ref[0,4,1,0]=(0.159607100455)*x_ref[0]**o+(0.715341910227)*x_ref[0]+(0.949546645819)*x_ref[1]**o+(0.362982182775)*x_ref[1]
        ref[0,4,1,1]=(-0.237702024358)*x_ref[0]**o+(-0.520488841454)*x_ref[0]+(0.606333933569)*x_ref[1]**o+(-0.00662757701049)*x_ref[1]
        ref[0,4,2,0]=(-0.415753817901)*x_ref[0]**o+(0.731285464784)*x_ref[0]+(0.15987336144)*x_ref[1]**o+(0.0697706236338)*x_ref[1]
        ref[0,4,2,1]=(0.0467482158427)*x_ref[0]**o+(0.306004722736)*x_ref[0]+(-0.905184725692)*x_ref[1]**o+(-0.158017889366)*x_ref[1]
        ref[1,0,0,0]=(-0.785535082426)*x_ref[0]**o+(-0.189285100452)*x_ref[0]+(-0.766349940057)*x_ref[1]**o+(0.531469308995)*x_ref[1]
        ref[1,0,0,1]=(0.522213196477)*x_ref[0]**o+(-0.661381741437)*x_ref[0]+(0.816089886956)*x_ref[1]**o+(0.139829382305)*x_ref[1]
        ref[1,0,1,0]=(0.0957295570417)*x_ref[0]**o+(-0.105252674536)*x_ref[0]+(-0.569429953382)*x_ref[1]**o+(-0.801934666657)*x_ref[1]
        ref[1,0,1,1]=(-0.557307051058)*x_ref[0]**o+(0.218852576416)*x_ref[0]+(0.430167216464)*x_ref[1]**o+(-0.112285681761)*x_ref[1]
        ref[1,0,2,0]=(0.288472422477)*x_ref[0]**o+(0.294377994763)*x_ref[0]+(-0.809503961511)*x_ref[1]**o+(0.357020226991)*x_ref[1]
        ref[1,0,2,1]=(-0.148113477705)*x_ref[0]**o+(0.164468646294)*x_ref[0]+(0.7977142476)*x_ref[1]**o+(-0.407274933384)*x_ref[1]
        ref[1,1,0,0]=(-0.926374434822)*x_ref[0]**o+(-0.107713463193)*x_ref[0]+(-0.511635723869)*x_ref[1]**o+(0.259319717689)*x_ref[1]
        ref[1,1,0,1]=(-0.18636582243)*x_ref[0]**o+(0.698258973628)*x_ref[0]+(0.234379281397)*x_ref[1]**o+(0.596195875495)*x_ref[1]
        ref[1,1,1,0]=(-0.180395892006)*x_ref[0]**o+(0.631248500084)*x_ref[0]+(0.104751616377)*x_ref[1]**o+(0.204500206653)*x_ref[1]
        ref[1,1,1,1]=(-0.517119564238)*x_ref[0]**o+(-0.16634015634)*x_ref[0]+(0.556785586675)*x_ref[1]**o+(0.42659306392)*x_ref[1]
        ref[1,1,2,0]=(0.163807238958)*x_ref[0]**o+(-0.534619478446)*x_ref[0]+(-0.700353460753)*x_ref[1]**o+(0.121393131648)*x_ref[1]
        ref[1,1,2,1]=(0.964276278961)*x_ref[0]**o+(0.846030751304)*x_ref[0]+(-0.745301894392)*x_ref[1]**o+(0.33821485558)*x_ref[1]
        ref[1,2,0,0]=(0.333173101086)*x_ref[0]**o+(0.774848048743)*x_ref[0]+(-0.659091916357)*x_ref[1]**o+(-0.797247144497)*x_ref[1]
        ref[1,2,0,1]=(0.0581394534285)*x_ref[0]**o+(-0.524771663031)*x_ref[0]+(-0.519465753317)*x_ref[1]**o+(0.71748795226)*x_ref[1]
        ref[1,2,1,0]=(0.224935262833)*x_ref[0]**o+(-0.393399819306)*x_ref[0]+(0.199349075224)*x_ref[1]**o+(-0.808847009412)*x_ref[1]
        ref[1,2,1,1]=(0.0385750795521)*x_ref[0]**o+(-8.09949553151e-05)*x_ref[0]+(0.496860964506)*x_ref[1]**o+(0.0119507407829)*x_ref[1]
        ref[1,2,2,0]=(0.876341556167)*x_ref[0]**o+(0.18082766372)*x_ref[0]+(0.401349871827)*x_ref[1]**o+(-0.543439037303)*x_ref[1]
        ref[1,2,2,1]=(-0.894525004507)*x_ref[0]**o+(-0.516771143695)*x_ref[0]+(0.0410916490263)*x_ref[1]**o+(0.753841618439)*x_ref[1]
        ref[1,3,0,0]=(0.251979970924)*x_ref[0]**o+(0.790276550438)*x_ref[0]+(-0.155708793888)*x_ref[1]**o+(0.28971278336)*x_ref[1]
        ref[1,3,0,1]=(0.258336265436)*x_ref[0]**o+(0.696642875149)*x_ref[0]+(0.0561355915649)*x_ref[1]**o+(0.601839744417)*x_ref[1]
        ref[1,3,1,0]=(-0.978729102785)*x_ref[0]**o+(0.335280613262)*x_ref[0]+(0.89228798232)*x_ref[1]**o+(-0.693050679348)*x_ref[1]
        ref[1,3,1,1]=(-0.885817753431)*x_ref[0]**o+(-0.0641571535708)*x_ref[0]+(-0.649578576181)*x_ref[1]**o+(-0.95773137602)*x_ref[1]
        ref[1,3,2,0]=(-0.333891001461)*x_ref[0]**o+(-0.496064308564)*x_ref[0]+(0.438549603377)*x_ref[1]**o+(0.0772468142672)*x_ref[1]
        ref[1,3,2,1]=(-0.16171788255)*x_ref[0]**o+(-0.939293453222)*x_ref[0]+(-0.284522648546)*x_ref[1]**o+(0.744642328375)*x_ref[1]
        ref[1,4,0,0]=(0.806133262332)*x_ref[0]**o+(-0.0225320206432)*x_ref[0]+(0.125958642606)*x_ref[1]**o+(0.610630249467)*x_ref[1]
        ref[1,4,0,1]=(0.657971870946)*x_ref[0]**o+(0.698894701525)*x_ref[0]+(-0.595952612496)*x_ref[1]**o+(-0.38268841292)*x_ref[1]
        ref[1,4,1,0]=(-0.277274411331)*x_ref[0]**o+(-0.483183004514)*x_ref[0]+(0.77656657009)*x_ref[1]**o+(-0.959290466645)*x_ref[1]
        ref[1,4,1,1]=(-0.96384356305)*x_ref[0]**o+(-0.00383301781476)*x_ref[0]+(0.788010409608)*x_ref[1]**o+(0.135378744316)*x_ref[1]
        ref[1,4,2,0]=(0.986761958017)*x_ref[0]**o+(-0.570276378038)*x_ref[0]+(-0.515784028315)*x_ref[1]**o+(0.487125123439)*x_ref[1]
        ref[1,4,2,1]=(-0.954735054702)*x_ref[0]**o+(-0.0610323074217)*x_ref[0]+(-0.255284859281)*x_ref[1]**o+(0.277795725698)*x_ref[1]
        ref[2,0,0,0]=(-0.708301219324)*x_ref[0]**o+(-0.970166781553)*x_ref[0]+(-0.0836451425855)*x_ref[1]**o+(0.525854202995)*x_ref[1]
        ref[2,0,0,1]=(-0.0995393865183)*x_ref[0]**o+(-0.0156624820324)*x_ref[0]+(0.225529316197)*x_ref[1]**o+(0.142617834056)*x_ref[1]
        ref[2,0,1,0]=(-0.585329524103)*x_ref[0]**o+(0.590285924441)*x_ref[0]+(0.201483726322)*x_ref[1]**o+(-0.735262197689)*x_ref[1]
        ref[2,0,1,1]=(-0.682877182963)*x_ref[0]**o+(-0.162682262799)*x_ref[0]+(-0.890051459843)*x_ref[1]**o+(-0.0188015866765)*x_ref[1]
        ref[2,0,2,0]=(0.868935476294)*x_ref[0]**o+(-0.629891068854)*x_ref[0]+(0.335632865663)*x_ref[1]**o+(-0.955849055716)*x_ref[1]
        ref[2,0,2,1]=(-0.907933228238)*x_ref[0]**o+(-0.105861987974)*x_ref[0]+(-0.131372383042)*x_ref[1]**o+(-0.970114136854)*x_ref[1]
        ref[2,1,0,0]=(-0.275788501814)*x_ref[0]**o+(-0.992861867033)*x_ref[0]+(-0.80676324962)*x_ref[1]**o+(0.269244680224)*x_ref[1]
        ref[2,1,0,1]=(-0.704704138774)*x_ref[0]**o+(-0.371803193009)*x_ref[0]+(-0.0190586405922)*x_ref[1]**o+(-0.0828837624698)*x_ref[1]
        ref[2,1,1,0]=(-0.532645566013)*x_ref[0]**o+(0.571754809868)*x_ref[0]+(0.656737493773)*x_ref[1]**o+(-0.548229591842)*x_ref[1]
        ref[2,1,1,1]=(0.23836921087)*x_ref[0]**o+(0.707475737785)*x_ref[0]+(-0.59660155968)*x_ref[1]**o+(-0.450204951071)*x_ref[1]
        ref[2,1,2,0]=(-0.534027348014)*x_ref[0]**o+(0.332100193727)*x_ref[0]+(-0.864157138567)*x_ref[1]**o+(-0.41833267066)*x_ref[1]
        ref[2,1,2,1]=(0.459878593269)*x_ref[0]**o+(0.446259140228)*x_ref[0]+(0.571242887502)*x_ref[1]**o+(-0.0390525669486)*x_ref[1]
        ref[2,2,0,0]=(0.322052799527)*x_ref[0]**o+(-0.915265209304)*x_ref[0]+(-0.99942007618)*x_ref[1]**o+(0.482294513743)*x_ref[1]
        ref[2,2,0,1]=(0.558489194456)*x_ref[0]**o+(0.00400576554086)*x_ref[0]+(-0.87484330165)*x_ref[1]**o+(-0.664504218977)*x_ref[1]
        ref[2,2,1,0]=(0.669621607461)*x_ref[0]**o+(-0.0450960554319)*x_ref[0]+(-0.221151480246)*x_ref[1]**o+(0.0123621737211)*x_ref[1]
        ref[2,2,1,1]=(0.476743013029)*x_ref[0]**o+(0.11118371822)*x_ref[0]+(-0.202074315677)*x_ref[1]**o+(0.377649433458)*x_ref[1]
        ref[2,2,2,0]=(0.667910543701)*x_ref[0]**o+(0.035703922547)*x_ref[0]+(-0.274733967095)*x_ref[1]**o+(-0.998934720182)*x_ref[1]
        ref[2,2,2,1]=(0.578909660087)*x_ref[0]**o+(0.149900086514)*x_ref[0]+(-0.226597651858)*x_ref[1]**o+(0.169175812287)*x_ref[1]
        ref[2,3,0,0]=(0.158370246586)*x_ref[0]**o+(0.595643112285)*x_ref[0]+(-0.476875564052)*x_ref[1]**o+(0.869752657517)*x_ref[1]
        ref[2,3,0,1]=(0.99718847714)*x_ref[0]**o+(0.581531822297)*x_ref[0]+(-0.607102775564)*x_ref[1]**o+(-0.8317970284)*x_ref[1]
        ref[2,3,1,0]=(-0.27390195745)*x_ref[0]**o+(0.0512853330414)*x_ref[0]+(0.465190666302)*x_ref[1]**o+(0.042533388788)*x_ref[1]
        ref[2,3,1,1]=(-0.183679504962)*x_ref[0]**o+(-0.0522782076121)*x_ref[0]+(0.921423118529)*x_ref[1]**o+(-0.0315080275225)*x_ref[1]
        ref[2,3,2,0]=(0.0421842348778)*x_ref[0]**o+(0.462023049702)*x_ref[0]+(-0.888342606803)*x_ref[1]**o+(-0.956503447884)*x_ref[1]
        ref[2,3,2,1]=(-0.950820288444)*x_ref[0]**o+(0.104271990168)*x_ref[0]+(-0.607212909582)*x_ref[1]**o+(-0.16283931259)*x_ref[1]
        ref[2,4,0,0]=(0.113505933533)*x_ref[0]**o+(0.27873514806)*x_ref[0]+(0.736699443523)*x_ref[1]**o+(-0.951737555693)*x_ref[1]
        ref[2,4,0,1]=(0.264170584849)*x_ref[0]**o+(-0.226649236617)*x_ref[0]+(0.489902180874)*x_ref[1]**o+(-0.191536591762)*x_ref[1]
        ref[2,4,1,0]=(0.712322285985)*x_ref[0]**o+(-0.323395455815)*x_ref[0]+(-0.298356681586)*x_ref[1]**o+(0.699891974541)*x_ref[1]
        ref[2,4,1,1]=(-0.631057763255)*x_ref[0]**o+(-0.489374933445)*x_ref[0]+(-0.021591806874)*x_ref[1]**o+(-0.692800124585)*x_ref[1]
        ref[2,4,2,0]=(0.916068935184)*x_ref[0]**o+(0.33779688858)*x_ref[0]+(0.06300443559)*x_ref[1]**o+(-0.231385203399)*x_ref[1]
        ref[2,4,2,1]=(-0.283485044625)*x_ref[0]**o+(0.548366574442)*x_ref[0]+(-0.543953132872)*x_ref[1]**o+(-0.956662823037)*x_ref[1]
        ref[3,0,0,0]=(0.709349274221)*x_ref[0]**o+(-0.342981009263)*x_ref[0]+(-0.682331761929)*x_ref[1]**o+(0.673169884813)*x_ref[1]
        ref[3,0,0,1]=(0.236249836593)*x_ref[0]**o+(-0.682845054434)*x_ref[0]+(-0.349169988218)*x_ref[1]**o+(0.502684425824)*x_ref[1]
        ref[3,0,1,0]=(-0.561655748793)*x_ref[0]**o+(0.120183772326)*x_ref[0]+(-0.966535325583)*x_ref[1]**o+(-0.0585117264949)*x_ref[1]
        ref[3,0,1,1]=(-0.661030287189)*x_ref[0]**o+(-0.322924086505)*x_ref[0]+(-0.986892059552)*x_ref[1]**o+(-0.778969361365)*x_ref[1]
        ref[3,0,2,0]=(-0.166307107307)*x_ref[0]**o+(0.612931303617)*x_ref[0]+(-0.283925165456)*x_ref[1]**o+(0.5998038814)*x_ref[1]
        ref[3,0,2,1]=(0.205099355916)*x_ref[0]**o+(-0.520970536278)*x_ref[0]+(0.900416833183)*x_ref[1]**o+(0.981826306069)*x_ref[1]
        ref[3,1,0,0]=(-0.875306666443)*x_ref[0]**o+(0.0189358479861)*x_ref[0]+(0.101187857445)*x_ref[1]**o+(-0.873584307534)*x_ref[1]
        ref[3,1,0,1]=(-0.0203656235757)*x_ref[0]**o+(-0.579229777042)*x_ref[0]+(-0.0799345631826)*x_ref[1]**o+(0.896315444069)*x_ref[1]
        ref[3,1,1,0]=(0.576791459982)*x_ref[0]**o+(-0.86541815506)*x_ref[0]+(-0.582764176207)*x_ref[1]**o+(-0.249437395011)*x_ref[1]
        ref[3,1,1,1]=(0.525953680497)*x_ref[0]**o+(-0.222445717402)*x_ref[0]+(-0.87710202063)*x_ref[1]**o+(-0.762921821973)*x_ref[1]
        ref[3,1,2,0]=(-0.883737185908)*x_ref[0]**o+(0.254053241277)*x_ref[0]+(0.559803298059)*x_ref[1]**o+(-0.577376956856)*x_ref[1]
        ref[3,1,2,1]=(-0.125792083387)*x_ref[0]**o+(-0.771549307509)*x_ref[0]+(0.142493480552)*x_ref[1]**o+(0.459996962548)*x_ref[1]
        ref[3,2,0,0]=(0.0531598927049)*x_ref[0]**o+(-0.950343229037)*x_ref[0]+(-0.748240453154)*x_ref[1]**o+(0.850555892342)*x_ref[1]
        ref[3,2,0,1]=(-0.660852152041)*x_ref[0]**o+(-0.763119278604)*x_ref[0]+(-0.811925409663)*x_ref[1]**o+(-0.178765071495)*x_ref[1]
        ref[3,2,1,0]=(-0.189124223305)*x_ref[0]**o+(-0.410392909042)*x_ref[0]+(0.993247167522)*x_ref[1]**o+(0.291451796298)*x_ref[1]
        ref[3,2,1,1]=(0.83496133421)*x_ref[0]**o+(0.695147480265)*x_ref[0]+(0.444420524946)*x_ref[1]**o+(0.107454620114)*x_ref[1]
        ref[3,2,2,0]=(0.94687479313)*x_ref[0]**o+(-0.140702324395)*x_ref[0]+(0.149203502428)*x_ref[1]**o+(-0.338246126345)*x_ref[1]
        ref[3,2,2,1]=(0.147609394709)*x_ref[0]**o+(-0.549219765239)*x_ref[0]+(0.397308993347)*x_ref[1]**o+(0.0402189312685)*x_ref[1]
        ref[3,3,0,0]=(0.27400981409)*x_ref[0]**o+(-0.516367266257)*x_ref[0]+(-0.784095952814)*x_ref[1]**o+(0.32858899009)*x_ref[1]
        ref[3,3,0,1]=(-0.393457382163)*x_ref[0]**o+(0.590406794051)*x_ref[0]+(0.328523918357)*x_ref[1]**o+(0.899537513355)*x_ref[1]
        ref[3,3,1,0]=(-0.851038623366)*x_ref[0]**o+(0.596974864726)*x_ref[0]+(-0.338630687551)*x_ref[1]**o+(0.634961435042)*x_ref[1]
        ref[3,3,1,1]=(0.569166143918)*x_ref[0]**o+(-0.388162282491)*x_ref[0]+(0.764830536733)*x_ref[1]**o+(-0.426737342542)*x_ref[1]
        ref[3,3,2,0]=(0.502221903448)*x_ref[0]**o+(0.940382225897)*x_ref[0]+(0.563444730579)*x_ref[1]**o+(0.023210869151)*x_ref[1]
        ref[3,3,2,1]=(0.061695403852)*x_ref[0]**o+(0.868302188758)*x_ref[0]+(0.0855803325758)*x_ref[1]**o+(-0.24366173638)*x_ref[1]
        ref[3,4,0,0]=(-0.533038182387)*x_ref[0]**o+(0.0222388915737)*x_ref[0]+(0.529037752769)*x_ref[1]**o+(-0.0665320895494)*x_ref[1]
        ref[3,4,0,1]=(0.714374055253)*x_ref[0]**o+(-0.623670743333)*x_ref[0]+(-0.866024792929)*x_ref[1]**o+(-0.0673124896111)*x_ref[1]
        ref[3,4,1,0]=(-0.0523466615045)*x_ref[0]**o+(-0.472563426622)*x_ref[0]+(0.139892673626)*x_ref[1]**o+(0.539655417178)*x_ref[1]
        ref[3,4,1,1]=(-0.473032580514)*x_ref[0]**o+(0.692812093618)*x_ref[0]+(0.664400517742)*x_ref[1]**o+(0.184125423364)*x_ref[1]
        ref[3,4,2,0]=(-0.0340798058725)*x_ref[0]**o+(-0.58900562863)*x_ref[0]+(0.396856384717)*x_ref[1]**o+(-0.559175428019)*x_ref[1]
        ref[3,4,2,1]=(-0.217679376634)*x_ref[0]**o+(-0.90389036667)*x_ref[0]+(0.967372942873)*x_ref[1]**o+(-0.482913619277)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.93229223906)*x[0]**o+(0.176839936782)*x[0]+(0.986271000359)*x[1]**o+(-0.449322385723)*x[1]+(-0.856450437823)*x[2]**o+(0.31473579221)*x[2]
        arg[0,0,0,1]=(0.714047232765)*x[0]**o+(-0.323471323003)*x[0]+(0.628080480343)*x[1]**o+(-0.992007912045)*x[1]+(-0.11019723051)*x[2]**o+(-0.104735408233)*x[2]
        arg[0,0,1,0]=(-0.437250443294)*x[0]**o+(-0.741469652781)*x[0]+(0.650921038465)*x[1]**o+(0.679622741243)*x[1]+(0.604005894895)*x[2]**o+(0.598155624876)*x[2]
        arg[0,0,1,1]=(-0.574872168761)*x[0]**o+(0.796207672391)*x[0]+(-0.478046646159)*x[1]**o+(0.604661326741)*x[1]+(-0.898457020186)*x[2]**o+(0.43234750827)*x[2]
        arg[0,0,2,0]=(-0.674008357742)*x[0]**o+(0.699614371016)*x[0]+(0.441321981559)*x[1]**o+(-0.619688689942)*x[1]+(-0.0874579208832)*x[2]**o+(-0.925753840555)*x[2]
        arg[0,0,2,1]=(-0.114329474892)*x[0]**o+(0.699501071503)*x[0]+(-0.618807420683)*x[1]**o+(0.411584695002)*x[1]+(-0.234620261556)*x[2]**o+(0.585720100967)*x[2]
        arg[0,1,0,0]=(0.617432097253)*x[0]**o+(-0.00304868863676)*x[0]+(0.571021785645)*x[1]**o+(0.457197205728)*x[1]+(0.462895867813)*x[2]**o+(0.278022956316)*x[2]
        arg[0,1,0,1]=(-0.659136231188)*x[0]**o+(-0.388212920906)*x[0]+(-0.982152608232)*x[1]**o+(-0.665797156013)*x[1]+(-0.76141890649)*x[2]**o+(-0.591458323836)*x[2]
        arg[0,1,1,0]=(-0.488016386893)*x[0]**o+(0.965660401944)*x[0]+(0.178400189819)*x[1]**o+(0.578417314499)*x[1]+(-0.398083955217)*x[2]**o+(0.460338187535)*x[2]
        arg[0,1,1,1]=(-0.182086458235)*x[0]**o+(0.0842368498629)*x[0]+(-0.642008026184)*x[1]**o+(-0.0449408131957)*x[1]+(-0.927397573018)*x[2]**o+(0.854313729816)*x[2]
        arg[0,1,2,0]=(-0.201475872769)*x[0]**o+(0.098190765292)*x[0]+(0.808624244922)*x[1]**o+(0.0726268899898)*x[1]+(-0.605334327027)*x[2]**o+(-0.999151255934)*x[2]
        arg[0,1,2,1]=(0.672344024218)*x[0]**o+(-0.638554833481)*x[0]+(-0.212231801634)*x[1]**o+(-0.492765366968)*x[1]+(0.311273771921)*x[2]**o+(-0.504370655006)*x[2]
        arg[0,2,0,0]=(0.9944935907)*x[0]**o+(-0.365261996611)*x[0]+(-0.105667939789)*x[1]**o+(0.886935751977)*x[1]+(0.263336157757)*x[2]**o+(-0.880316523113)*x[2]
        arg[0,2,0,1]=(-0.659779300787)*x[0]**o+(-0.728590180152)*x[0]+(0.732078508037)*x[1]**o+(0.684963412098)*x[1]+(0.822970700012)*x[2]**o+(-0.680001879674)*x[2]
        arg[0,2,1,0]=(-0.207855412893)*x[0]**o+(0.438275426529)*x[0]+(0.87046682066)*x[1]**o+(-0.960322013848)*x[1]+(-0.273728198625)*x[2]**o+(-0.960137228004)*x[2]
        arg[0,2,1,1]=(-0.189232680002)*x[0]**o+(0.866417302002)*x[0]+(-0.93403311248)*x[1]**o+(-0.55773104878)*x[1]+(0.94062895773)*x[2]**o+(-0.758052194805)*x[2]
        arg[0,2,2,0]=(0.100859944394)*x[0]**o+(-0.20292135909)*x[0]+(-0.713339941261)*x[1]**o+(-0.615106454411)*x[1]+(0.791245995152)*x[2]**o+(-0.81051791594)*x[2]
        arg[0,2,2,1]=(0.595743424517)*x[0]**o+(-0.179075876404)*x[0]+(-0.799005601715)*x[1]**o+(-0.0554635073569)*x[1]+(-0.850510701985)*x[2]**o+(0.234516538169)*x[2]
        arg[0,3,0,0]=(-0.361667738861)*x[0]**o+(-0.551837498696)*x[0]+(-0.834377202094)*x[1]**o+(0.814168224265)*x[1]+(-0.00479533268377)*x[2]**o+(0.843007652951)*x[2]
        arg[0,3,0,1]=(-0.448417439465)*x[0]**o+(0.431496158388)*x[0]+(-0.547416443433)*x[1]**o+(-0.0746785704695)*x[1]+(-0.646238182944)*x[2]**o+(-0.318608714595)*x[2]
        arg[0,3,1,0]=(-0.275647366007)*x[0]**o+(-0.410594785047)*x[0]+(-0.858282378196)*x[1]**o+(-0.755586690837)*x[1]+(-0.420481091654)*x[2]**o+(-0.0996487371831)*x[2]
        arg[0,3,1,1]=(-0.395961249009)*x[0]**o+(0.0860896467863)*x[0]+(0.482719525748)*x[1]**o+(0.319368178133)*x[1]+(0.516449153414)*x[2]**o+(-0.246749862576)*x[2]
        arg[0,3,2,0]=(-0.367322027122)*x[0]**o+(0.687206412524)*x[0]+(0.616265233972)*x[1]**o+(-0.32660589112)*x[1]+(-0.944607343096)*x[2]**o+(0.123982537951)*x[2]
        arg[0,3,2,1]=(0.700497166804)*x[0]**o+(0.553961329865)*x[0]+(-0.34020003754)*x[1]**o+(0.295870992783)*x[1]+(-0.220275880046)*x[2]**o+(0.766121035825)*x[2]
        arg[0,4,0,0]=(0.405469789242)*x[0]**o+(-0.338347662873)*x[0]+(0.0761796434233)*x[1]**o+(-0.120599175033)*x[1]+(0.107865585379)*x[2]**o+(0.0711941007838)*x[2]
        arg[0,4,0,1]=(0.614166154899)*x[0]**o+(0.195780702963)*x[0]+(0.718935985215)*x[1]**o+(0.214577296792)*x[1]+(-0.366985242107)*x[2]**o+(-0.509584787633)*x[2]
        arg[0,4,1,0]=(-0.638646370981)*x[0]**o+(-0.769955899284)*x[0]+(-0.887407540503)*x[1]**o+(0.0901245429998)*x[1]+(0.786085641321)*x[2]**o+(0.6248068552)*x[2]
        arg[0,4,1,1]=(-0.145426222043)*x[0]**o+(-0.494114545016)*x[0]+(0.422153721335)*x[1]**o+(0.617022143901)*x[1]+(-0.665972324723)*x[2]**o+(-0.476724432799)*x[2]
        arg[0,4,2,0]=(0.556105252092)*x[0]**o+(-0.893268696809)*x[0]+(0.146441889489)*x[1]**o+(-0.170512113283)*x[1]+(-0.869313262875)*x[2]**o+(-0.976874109392)*x[2]
        arg[0,4,2,1]=(0.858927206499)*x[0]**o+(0.198998355146)*x[0]+(0.278125296049)*x[1]**o+(-0.0442518828093)*x[1]+(0.274852724203)*x[2]**o+(-0.167649193314)*x[2]
        arg[1,0,0,0]=(-0.0429420289353)*x[0]**o+(-0.15764693929)*x[0]+(-0.859599917512)*x[1]**o+(-0.904178643575)*x[1]+(0.519560711493)*x[2]**o+(0.258644717907)*x[2]
        arg[1,0,0,1]=(-0.185321114264)*x[0]**o+(-0.794493830004)*x[0]+(0.136084421162)*x[1]**o+(0.926912021419)*x[1]+(0.346392120388)*x[2]**o+(-0.183908371454)*x[2]
        arg[1,0,1,0]=(-0.730675509877)*x[0]**o+(0.865034802778)*x[0]+(-0.506315431878)*x[1]**o+(-0.94607761051)*x[1]+(0.195194792801)*x[2]**o+(-0.865129258879)*x[2]
        arg[1,0,1,1]=(-0.0597026648107)*x[0]**o+(-0.884790497)*x[0]+(-0.0631278742098)*x[1]**o+(-0.808364940737)*x[1]+(-0.968952604837)*x[2]**o+(0.142246702632)*x[2]
        arg[1,0,2,0]=(0.648472894563)*x[0]**o+(0.145861644073)*x[0]+(0.501227520574)*x[1]**o+(0.265886482306)*x[1]+(-0.0472772173573)*x[2]**o+(-0.364169655969)*x[2]
        arg[1,0,2,1]=(0.847863600695)*x[0]**o+(-0.288841685199)*x[0]+(-0.500273502795)*x[1]**o+(0.627379277592)*x[1]+(0.326814024625)*x[2]**o+(-0.514712825895)*x[2]
        arg[1,1,0,0]=(-0.947845713712)*x[0]**o+(-0.372115616305)*x[0]+(-0.572858279998)*x[1]**o+(0.226237627692)*x[1]+(0.210201947182)*x[2]**o+(0.0147616538594)*x[2]
        arg[1,1,0,1]=(-0.39813630466)*x[0]**o+(-0.658496395731)*x[0]+(-0.363906465005)*x[1]**o+(-0.816719661238)*x[1]+(-0.478106835032)*x[2]**o+(0.0819042309589)*x[2]
        arg[1,1,1,0]=(-0.287160278593)*x[0]**o+(-0.0281402640978)*x[0]+(0.41559795378)*x[1]**o+(0.158274458097)*x[1]+(-0.798429481502)*x[2]**o+(-0.278585334624)*x[2]
        arg[1,1,1,1]=(0.809550480973)*x[0]**o+(0.192026915854)*x[0]+(-0.891168838599)*x[1]**o+(0.0482621192882)*x[1]+(-0.35339708159)*x[2]**o+(-0.133161551553)*x[2]
        arg[1,1,2,0]=(0.939582328872)*x[0]**o+(0.561963790516)*x[0]+(0.885264722949)*x[1]**o+(-0.468042305266)*x[1]+(0.232880002429)*x[2]**o+(-0.772896005701)*x[2]
        arg[1,1,2,1]=(-0.297333303209)*x[0]**o+(-0.414050924207)*x[0]+(-0.278903698873)*x[1]**o+(0.345370620869)*x[1]+(-0.0570151259027)*x[2]**o+(0.786449641378)*x[2]
        arg[1,2,0,0]=(0.644047138589)*x[0]**o+(-0.480660438838)*x[0]+(0.32526499462)*x[1]**o+(-0.775988282127)*x[1]+(0.108374143932)*x[2]**o+(0.955463667181)*x[2]
        arg[1,2,0,1]=(-0.468236790339)*x[0]**o+(0.844197711724)*x[0]+(0.808973429831)*x[1]**o+(-0.647699550005)*x[1]+(0.552555959813)*x[2]**o+(-0.3403079515)*x[2]
        arg[1,2,1,0]=(0.864686362308)*x[0]**o+(0.248923646116)*x[0]+(0.0287042863842)*x[1]**o+(0.404974282153)*x[1]+(0.113501438234)*x[2]**o+(-0.619364894397)*x[2]
        arg[1,2,1,1]=(-0.625146992499)*x[0]**o+(-0.0480602452983)*x[0]+(0.168910550914)*x[1]**o+(-0.891584218123)*x[1]+(0.0609357618263)*x[2]**o+(0.0451921002161)*x[2]
        arg[1,2,2,0]=(-0.478728628752)*x[0]**o+(0.428376547798)*x[0]+(0.817203376818)*x[1]**o+(-0.856056266953)*x[1]+(0.354960313382)*x[2]**o+(-0.611468960461)*x[2]
        arg[1,2,2,1]=(0.49500742113)*x[0]**o+(0.902948130263)*x[0]+(-0.810748543744)*x[1]**o+(-0.696150924858)*x[1]+(-0.748278516161)*x[2]**o+(0.694815527058)*x[2]
        arg[1,3,0,0]=(-0.361960150296)*x[0]**o+(-0.845139780461)*x[0]+(-0.0624499309665)*x[1]**o+(-0.0515558058489)*x[1]+(0.400690983017)*x[2]**o+(-0.744608407974)*x[2]
        arg[1,3,0,1]=(-0.452453617467)*x[0]**o+(0.845396139593)*x[0]+(0.292319469936)*x[1]**o+(-0.833547518634)*x[1]+(-0.511797700353)*x[2]**o+(0.857891454773)*x[2]
        arg[1,3,1,0]=(0.779504092912)*x[0]**o+(0.500524773615)*x[0]+(0.590542659382)*x[1]**o+(0.479310630519)*x[1]+(0.509960641956)*x[2]**o+(-0.564980432959)*x[2]
        arg[1,3,1,1]=(-0.145745189298)*x[0]**o+(-0.256064576359)*x[0]+(0.640890837126)*x[1]**o+(-0.718802035152)*x[1]+(-0.322216873583)*x[2]**o+(-0.0246526526334)*x[2]
        arg[1,3,2,0]=(-0.463434285098)*x[0]**o+(0.116722448643)*x[0]+(-0.690378381342)*x[1]**o+(0.429197465885)*x[1]+(-0.353019731384)*x[2]**o+(0.57397712918)*x[2]
        arg[1,3,2,1]=(0.169676853383)*x[0]**o+(0.815225472623)*x[0]+(-0.11125527552)*x[1]**o+(0.391513831525)*x[1]+(-0.587041361585)*x[2]**o+(0.255194151616)*x[2]
        arg[1,4,0,0]=(0.325351608628)*x[0]**o+(-0.57751715453)*x[0]+(0.145167511303)*x[1]**o+(-0.969286030742)*x[1]+(0.917370398478)*x[2]**o+(0.691483691598)*x[2]
        arg[1,4,0,1]=(0.0997998298271)*x[0]**o+(-0.101248202247)*x[0]+(-0.650649653848)*x[1]**o+(-0.157887947288)*x[1]+(0.517501747765)*x[2]**o+(-0.0852921926298)*x[2]
        arg[1,4,1,0]=(0.984595420407)*x[0]**o+(0.499293499076)*x[0]+(0.352193115874)*x[1]**o+(-0.684522909647)*x[1]+(-0.391538373605)*x[2]**o+(0.404635595162)*x[2]
        arg[1,4,1,1]=(-0.231310482173)*x[0]**o+(-0.0385089836559)*x[0]+(0.247671691888)*x[1]**o+(0.883971955027)*x[1]+(-0.0856624709734)*x[2]**o+(0.942940875904)*x[2]
        arg[1,4,2,0]=(-0.471467257434)*x[0]**o+(0.835634240897)*x[0]+(-0.594421294562)*x[1]**o+(-0.187584387111)*x[1]+(0.859867472765)*x[2]**o+(0.930951154576)*x[2]
        arg[1,4,2,1]=(0.504416975601)*x[0]**o+(0.435692923635)*x[0]+(-0.296945944313)*x[1]**o+(0.815156218765)*x[1]+(0.464464247442)*x[2]**o+(-0.378809453041)*x[2]
        arg[2,0,0,0]=(0.213087423107)*x[0]**o+(-0.504681665801)*x[0]+(0.796352190405)*x[1]**o+(0.546752751136)*x[1]+(0.57347644664)*x[2]**o+(0.715856963195)*x[2]
        arg[2,0,0,1]=(-0.28455266769)*x[0]**o+(0.853162220766)*x[0]+(0.164930982863)*x[1]**o+(-0.0136996498765)*x[1]+(0.293717915079)*x[2]**o+(-0.216124158473)*x[2]
        arg[2,0,1,0]=(-0.328115989441)*x[0]**o+(-0.529441496488)*x[0]+(-0.43579479677)*x[1]**o+(-0.851333801194)*x[1]+(-0.278406102972)*x[2]**o+(0.947777471824)*x[2]
        arg[2,0,1,1]=(-0.756438938789)*x[0]**o+(-0.624332272735)*x[0]+(-0.591638429565)*x[1]**o+(-0.0608281771566)*x[1]+(-0.852078784085)*x[2]**o+(-0.526558964866)*x[2]
        arg[2,0,2,0]=(-0.923892426969)*x[0]**o+(0.788121899878)*x[0]+(-0.446006331202)*x[1]**o+(0.997210871574)*x[1]+(0.356998606103)*x[2]**o+(0.972643822823)*x[2]
        arg[2,0,2,1]=(0.225183308571)*x[0]**o+(-0.757505414777)*x[0]+(-0.59573192074)*x[1]**o+(0.57166154987)*x[1]+(0.622343198809)*x[2]**o+(0.60343970503)*x[2]
        arg[2,1,0,0]=(0.51027702934)*x[0]**o+(-0.711737679522)*x[0]+(-0.211712579627)*x[1]**o+(-0.495817605314)*x[1]+(0.924425804272)*x[2]**o+(0.753424835275)*x[2]
        arg[2,1,0,1]=(0.476218252577)*x[0]**o+(0.187579891557)*x[0]+(0.727359583623)*x[1]**o+(-0.389875559955)*x[1]+(-0.352889326477)*x[2]**o+(0.365143943637)*x[2]
        arg[2,1,1,0]=(-0.815704312943)*x[0]**o+(0.315217470245)*x[0]+(-0.535234555065)*x[1]**o+(-0.433408988501)*x[1]+(-0.676694087389)*x[2]**o+(0.384097177718)*x[2]
        arg[2,1,1,1]=(-0.0368452411345)*x[0]**o+(0.514540686444)*x[0]+(-0.527100066965)*x[1]**o+(0.0291015562912)*x[1]+(-0.605646905004)*x[2]**o+(0.931528599981)*x[2]
        arg[2,1,2,0]=(-0.0994015945266)*x[0]**o+(-0.762310347434)*x[0]+(-0.845934950466)*x[1]**o+(-0.235758967923)*x[1]+(0.755531422698)*x[2]**o+(0.108353867777)*x[2]
        arg[2,1,2,1]=(-0.868779052889)*x[0]**o+(0.783878941075)*x[0]+(0.900089560473)*x[1]**o+(-0.836779405877)*x[1]+(-0.799571882449)*x[2]**o+(0.999710029157)*x[2]
        arg[2,2,0,0]=(0.541465277143)*x[0]**o+(-0.840477797945)*x[0]+(-0.362577721098)*x[1]**o+(0.00680131806718)*x[1]+(-0.338485239867)*x[2]**o+(0.828213756751)*x[2]
        arg[2,2,0,1]=(-0.927347423175)*x[0]**o+(-0.936191125785)*x[0]+(-0.411819804833)*x[1]**o+(-0.0222067637269)*x[1]+(-0.104411832586)*x[2]**o+(-0.79937203081)*x[2]
        arg[2,2,1,0]=(0.273051712848)*x[0]**o+(-0.802468557957)*x[0]+(-0.019392898867)*x[1]**o+(-0.612803074784)*x[1]+(-0.269705464223)*x[2]**o+(-0.244405147706)*x[2]
        arg[2,2,1,1]=(-0.08869649035)*x[0]**o+(0.809761782447)*x[0]+(0.761420080249)*x[1]**o+(0.16590038557)*x[1]+(-0.405657593445)*x[2]**o+(0.123593603037)*x[2]
        arg[2,2,2,0]=(-0.799691160987)*x[0]**o+(0.834835690475)*x[0]+(-0.0863569221737)*x[1]**o+(0.988343215152)*x[1]+(0.294423283474)*x[2]**o+(0.325725138203)*x[2]
        arg[2,2,2,1]=(0.648897496592)*x[0]**o+(-0.474329387314)*x[0]+(-0.904613697104)*x[1]**o+(0.382201388286)*x[1]+(-0.596889148339)*x[2]**o+(0.705132157684)*x[2]
        arg[2,3,0,0]=(0.906738144045)*x[0]**o+(-0.365392026426)*x[0]+(0.730557428733)*x[1]**o+(-0.857864471189)*x[1]+(-0.160930657186)*x[2]**o+(0.885067383334)*x[2]
        arg[2,3,0,1]=(0.992764702472)*x[0]**o+(0.720688532948)*x[0]+(0.318869772818)*x[1]**o+(0.995508347248)*x[1]+(0.108569456996)*x[2]**o+(-0.990939292995)*x[2]
        arg[2,3,1,0]=(0.690584992096)*x[0]**o+(0.605885242246)*x[0]+(0.72371794265)*x[1]**o+(-0.632609325879)*x[1]+(0.520273588734)*x[2]**o+(0.885804143155)*x[2]
        arg[2,3,1,1]=(0.485231542482)*x[0]**o+(0.866920225418)*x[0]+(-0.814964945972)*x[1]**o+(0.544132513209)*x[1]+(-0.155071376494)*x[2]**o+(-0.941300746237)*x[2]
        arg[2,3,2,0]=(0.748232238388)*x[0]**o+(0.179397724419)*x[0]+(-0.43881886248)*x[1]**o+(-0.124941799624)*x[1]+(0.14679335819)*x[2]**o+(-0.758738846688)*x[2]
        arg[2,3,2,1]=(-0.0951333655337)*x[0]**o+(0.37270972476)*x[0]+(-0.651977554115)*x[1]**o+(-0.88918597476)*x[1]+(-0.730858685576)*x[2]**o+(0.685881924459)*x[2]
        arg[2,4,0,0]=(-0.823487466802)*x[0]**o+(-0.583630855643)*x[0]+(-0.664814562747)*x[1]**o+(0.860952605956)*x[1]+(-0.249114322668)*x[2]**o+(-0.968356584237)*x[2]
        arg[2,4,0,1]=(0.44040146252)*x[0]**o+(-0.246827565902)*x[0]+(0.596300980733)*x[1]**o+(-0.825818200414)*x[1]+(0.0715512875967)*x[2]**o+(0.996064679803)*x[2]
        arg[2,4,1,0]=(-0.287806453021)*x[0]**o+(0.704331542927)*x[0]+(-0.346139754521)*x[1]**o+(-0.766667037125)*x[1]+(0.283874132105)*x[2]**o+(-0.613145979204)*x[2]
        arg[2,4,1,1]=(-0.116649875156)*x[0]**o+(0.291380048239)*x[0]+(0.689809221905)*x[1]**o+(0.758935910592)*x[1]+(0.505142787603)*x[2]**o+(-0.113611480503)*x[2]
        arg[2,4,2,0]=(-0.836908572692)*x[0]**o+(-0.21892828725)*x[0]+(-0.935780103868)*x[1]**o+(0.106204643268)*x[1]+(-0.894474423807)*x[2]**o+(0.296554773217)*x[2]
        arg[2,4,2,1]=(-0.89413495023)*x[0]**o+(0.6080813058)*x[0]+(0.087863889878)*x[1]**o+(0.213373665612)*x[1]+(0.17085345025)*x[2]**o+(-0.967625138712)*x[2]
        arg[3,0,0,0]=(0.327537677409)*x[0]**o+(-0.299373059373)*x[0]+(-0.43223648542)*x[1]**o+(-0.499115781051)*x[1]+(0.145464567449)*x[2]**o+(0.580930584563)*x[2]
        arg[3,0,0,1]=(-0.122622543351)*x[0]**o+(-0.220547821081)*x[0]+(0.892679827803)*x[1]**o+(0.436584604381)*x[1]+(-0.156093580288)*x[2]**o+(0.464069069075)*x[2]
        arg[3,0,1,0]=(0.590288663203)*x[0]**o+(-0.212929445251)*x[0]+(0.0899671639061)*x[1]**o+(-0.87729854872)*x[1]+(0.789506558716)*x[2]**o+(-0.68867690907)*x[2]
        arg[3,0,1,1]=(-0.913651849012)*x[0]**o+(-0.430938918892)*x[0]+(-0.351186012371)*x[1]**o+(0.500306994748)*x[1]+(0.145381845281)*x[2]**o+(0.624993552333)*x[2]
        arg[3,0,2,0]=(0.281393881265)*x[0]**o+(-0.704727129826)*x[0]+(-0.0102091836449)*x[1]**o+(0.311857501353)*x[1]+(0.300732947904)*x[2]**o+(0.923973499548)*x[2]
        arg[3,0,2,1]=(-0.584907526993)*x[0]**o+(0.625732316822)*x[0]+(-0.609691930408)*x[1]**o+(0.056160226956)*x[1]+(-0.766583561947)*x[2]**o+(0.307692190194)*x[2]
        arg[3,1,0,0]=(0.731231379445)*x[0]**o+(-0.0485980527115)*x[0]+(0.908932758852)*x[1]**o+(0.188918035886)*x[1]+(0.685837743688)*x[2]**o+(-0.299987553989)*x[2]
        arg[3,1,0,1]=(-0.784006561231)*x[0]**o+(-0.809789430022)*x[0]+(-0.846391792735)*x[1]**o+(-0.211583304405)*x[1]+(-0.00198590063687)*x[2]**o+(-0.853253255224)*x[2]
        arg[3,1,1,0]=(0.595567299132)*x[0]**o+(-0.581734416878)*x[0]+(0.88144481486)*x[1]**o+(0.776876564509)*x[1]+(-0.367477971386)*x[2]**o+(0.0806320164453)*x[2]
        arg[3,1,1,1]=(0.440836187611)*x[0]**o+(-0.578925091757)*x[0]+(0.923367189832)*x[1]**o+(-0.993888173929)*x[1]+(-0.645514380411)*x[2]**o+(0.0267439044199)*x[2]
        arg[3,1,2,0]=(-0.305182288788)*x[0]**o+(-0.503372502553)*x[0]+(0.740948657389)*x[1]**o+(0.285661052275)*x[1]+(-0.634510984218)*x[2]**o+(-0.123979746456)*x[2]
        arg[3,1,2,1]=(-0.897274928495)*x[0]**o+(0.971356796164)*x[0]+(0.940768290313)*x[1]**o+(-0.552941985667)*x[1]+(0.172857451423)*x[2]**o+(0.844758237523)*x[2]
        arg[3,2,0,0]=(-0.903542725333)*x[0]**o+(0.936704774551)*x[0]+(-0.398085875228)*x[1]**o+(-0.304061573512)*x[1]+(-0.13362037229)*x[2]**o+(-0.405043350728)*x[2]
        arg[3,2,0,1]=(-0.0208590751801)*x[0]**o+(-0.354086381885)*x[0]+(0.510884698744)*x[1]**o+(-0.275504880518)*x[1]+(0.459476496356)*x[2]**o+(0.25358097976)*x[2]
        arg[3,2,1,0]=(-0.0171081877107)*x[0]**o+(-0.856081626112)*x[0]+(-0.0698686571004)*x[1]**o+(0.0954918361122)*x[1]+(0.938494954069)*x[2]**o+(-0.824928688248)*x[2]
        arg[3,2,1,1]=(-0.277307915098)*x[0]**o+(0.737855779658)*x[0]+(-0.466761948855)*x[1]**o+(-0.821498435809)*x[1]+(-0.391009601442)*x[2]**o+(0.643834542048)*x[2]
        arg[3,2,2,0]=(0.379015007885)*x[0]**o+(0.348646179575)*x[0]+(0.733765133901)*x[1]**o+(-0.722363571789)*x[1]+(0.488531791931)*x[2]**o+(0.279813923679)*x[2]
        arg[3,2,2,1]=(0.516708382119)*x[0]**o+(0.0426665103314)*x[0]+(-0.168751315718)*x[1]**o+(-0.124069236608)*x[1]+(-0.778866001674)*x[2]**o+(-0.362031001279)*x[2]
        arg[3,3,0,0]=(0.995284803465)*x[0]**o+(-0.442327755879)*x[0]+(-0.13110270925)*x[1]**o+(-0.172857543102)*x[1]+(-0.910522664379)*x[2]**o+(0.516931833163)*x[2]
        arg[3,3,0,1]=(-0.889320295626)*x[0]**o+(-0.728330379534)*x[0]+(-0.584091700627)*x[1]**o+(0.469914082337)*x[1]+(0.18874127891)*x[2]**o+(0.0846532039249)*x[2]
        arg[3,3,1,0]=(0.911523535462)*x[0]**o+(0.0924361220354)*x[0]+(-0.0822504494228)*x[1]**o+(0.00630249023591)*x[1]+(-0.180497330322)*x[2]**o+(0.817405577154)*x[2]
        arg[3,3,1,1]=(0.681872747813)*x[0]**o+(0.840680190206)*x[0]+(-0.477660295206)*x[1]**o+(0.0399086968056)*x[1]+(0.440403204247)*x[2]**o+(-0.220656197016)*x[2]
        arg[3,3,2,0]=(-0.967524407336)*x[0]**o+(0.896652274462)*x[0]+(0.735269328519)*x[1]**o+(-0.249017384614)*x[1]+(-0.573222311563)*x[2]**o+(-0.653800334954)*x[2]
        arg[3,3,2,1]=(0.10041632804)*x[0]**o+(0.167657178348)*x[0]+(0.990621134594)*x[1]**o+(-0.309063473348)*x[1]+(-0.688206827575)*x[2]**o+(0.0926806818663)*x[2]
        arg[3,4,0,0]=(-0.128678339518)*x[0]**o+(-0.974238160059)*x[0]+(-0.762599464772)*x[1]**o+(0.0233286358698)*x[1]+(-0.622481697973)*x[2]**o+(-0.080495303798)*x[2]
        arg[3,4,0,1]=(-0.380318230058)*x[0]**o+(0.568425682425)*x[0]+(0.787369916262)*x[1]**o+(0.296745245123)*x[1]+(-0.917621002414)*x[2]**o+(-0.696209404954)*x[2]
        arg[3,4,1,0]=(-0.382961468898)*x[0]**o+(0.672614453621)*x[0]+(0.565371503874)*x[1]**o+(-0.597331999912)*x[1]+(-0.153637846704)*x[2]**o+(-0.438159587906)*x[2]
        arg[3,4,1,1]=(-0.541043007799)*x[0]**o+(0.415600628159)*x[0]+(-0.793716270785)*x[1]**o+(0.780110667203)*x[1]+(-0.101309486185)*x[2]**o+(-0.138868352939)*x[2]
        arg[3,4,2,0]=(0.679815917569)*x[0]**o+(-0.800657537776)*x[0]+(-0.711956069893)*x[1]**o+(-0.575626146908)*x[1]+(0.184247083249)*x[2]**o+(-0.728118084396)*x[2]
        arg[3,4,2,1]=(-0.608089432748)*x[0]**o+(0.356378798828)*x[0]+(-0.575957599599)*x[1]**o+(0.599586825137)*x[1]+(-0.160475743299)*x[2]**o+(-0.070762356177)*x[2]
        ref[0,0,0,0]=(-0.93229223906)*x_ref[0]**o+(0.176839936782)*x_ref[0]+(0.986271000359)*x_ref[1]**o+(-0.449322385723)*x_ref[1]+(-0.856450437823)*x_ref[2]**o+(0.31473579221)*x_ref[2]
        ref[0,0,0,1]=(0.714047232765)*x_ref[0]**o+(-0.323471323003)*x_ref[0]+(0.628080480343)*x_ref[1]**o+(-0.992007912045)*x_ref[1]+(-0.11019723051)*x_ref[2]**o+(-0.104735408233)*x_ref[2]
        ref[0,0,1,0]=(-0.437250443294)*x_ref[0]**o+(-0.741469652781)*x_ref[0]+(0.650921038465)*x_ref[1]**o+(0.679622741243)*x_ref[1]+(0.604005894895)*x_ref[2]**o+(0.598155624876)*x_ref[2]
        ref[0,0,1,1]=(-0.574872168761)*x_ref[0]**o+(0.796207672391)*x_ref[0]+(-0.478046646159)*x_ref[1]**o+(0.604661326741)*x_ref[1]+(-0.898457020186)*x_ref[2]**o+(0.43234750827)*x_ref[2]
        ref[0,0,2,0]=(-0.674008357742)*x_ref[0]**o+(0.699614371016)*x_ref[0]+(0.441321981559)*x_ref[1]**o+(-0.619688689942)*x_ref[1]+(-0.0874579208832)*x_ref[2]**o+(-0.925753840555)*x_ref[2]
        ref[0,0,2,1]=(-0.114329474892)*x_ref[0]**o+(0.699501071503)*x_ref[0]+(-0.618807420683)*x_ref[1]**o+(0.411584695002)*x_ref[1]+(-0.234620261556)*x_ref[2]**o+(0.585720100967)*x_ref[2]
        ref[0,1,0,0]=(0.617432097253)*x_ref[0]**o+(-0.00304868863676)*x_ref[0]+(0.571021785645)*x_ref[1]**o+(0.457197205728)*x_ref[1]+(0.462895867813)*x_ref[2]**o+(0.278022956316)*x_ref[2]
        ref[0,1,0,1]=(-0.659136231188)*x_ref[0]**o+(-0.388212920906)*x_ref[0]+(-0.982152608232)*x_ref[1]**o+(-0.665797156013)*x_ref[1]+(-0.76141890649)*x_ref[2]**o+(-0.591458323836)*x_ref[2]
        ref[0,1,1,0]=(-0.488016386893)*x_ref[0]**o+(0.965660401944)*x_ref[0]+(0.178400189819)*x_ref[1]**o+(0.578417314499)*x_ref[1]+(-0.398083955217)*x_ref[2]**o+(0.460338187535)*x_ref[2]
        ref[0,1,1,1]=(-0.182086458235)*x_ref[0]**o+(0.0842368498629)*x_ref[0]+(-0.642008026184)*x_ref[1]**o+(-0.0449408131957)*x_ref[1]+(-0.927397573018)*x_ref[2]**o+(0.854313729816)*x_ref[2]
        ref[0,1,2,0]=(-0.201475872769)*x_ref[0]**o+(0.098190765292)*x_ref[0]+(0.808624244922)*x_ref[1]**o+(0.0726268899898)*x_ref[1]+(-0.605334327027)*x_ref[2]**o+(-0.999151255934)*x_ref[2]
        ref[0,1,2,1]=(0.672344024218)*x_ref[0]**o+(-0.638554833481)*x_ref[0]+(-0.212231801634)*x_ref[1]**o+(-0.492765366968)*x_ref[1]+(0.311273771921)*x_ref[2]**o+(-0.504370655006)*x_ref[2]
        ref[0,2,0,0]=(0.9944935907)*x_ref[0]**o+(-0.365261996611)*x_ref[0]+(-0.105667939789)*x_ref[1]**o+(0.886935751977)*x_ref[1]+(0.263336157757)*x_ref[2]**o+(-0.880316523113)*x_ref[2]
        ref[0,2,0,1]=(-0.659779300787)*x_ref[0]**o+(-0.728590180152)*x_ref[0]+(0.732078508037)*x_ref[1]**o+(0.684963412098)*x_ref[1]+(0.822970700012)*x_ref[2]**o+(-0.680001879674)*x_ref[2]
        ref[0,2,1,0]=(-0.207855412893)*x_ref[0]**o+(0.438275426529)*x_ref[0]+(0.87046682066)*x_ref[1]**o+(-0.960322013848)*x_ref[1]+(-0.273728198625)*x_ref[2]**o+(-0.960137228004)*x_ref[2]
        ref[0,2,1,1]=(-0.189232680002)*x_ref[0]**o+(0.866417302002)*x_ref[0]+(-0.93403311248)*x_ref[1]**o+(-0.55773104878)*x_ref[1]+(0.94062895773)*x_ref[2]**o+(-0.758052194805)*x_ref[2]
        ref[0,2,2,0]=(0.100859944394)*x_ref[0]**o+(-0.20292135909)*x_ref[0]+(-0.713339941261)*x_ref[1]**o+(-0.615106454411)*x_ref[1]+(0.791245995152)*x_ref[2]**o+(-0.81051791594)*x_ref[2]
        ref[0,2,2,1]=(0.595743424517)*x_ref[0]**o+(-0.179075876404)*x_ref[0]+(-0.799005601715)*x_ref[1]**o+(-0.0554635073569)*x_ref[1]+(-0.850510701985)*x_ref[2]**o+(0.234516538169)*x_ref[2]
        ref[0,3,0,0]=(-0.361667738861)*x_ref[0]**o+(-0.551837498696)*x_ref[0]+(-0.834377202094)*x_ref[1]**o+(0.814168224265)*x_ref[1]+(-0.00479533268377)*x_ref[2]**o+(0.843007652951)*x_ref[2]
        ref[0,3,0,1]=(-0.448417439465)*x_ref[0]**o+(0.431496158388)*x_ref[0]+(-0.547416443433)*x_ref[1]**o+(-0.0746785704695)*x_ref[1]+(-0.646238182944)*x_ref[2]**o+(-0.318608714595)*x_ref[2]
        ref[0,3,1,0]=(-0.275647366007)*x_ref[0]**o+(-0.410594785047)*x_ref[0]+(-0.858282378196)*x_ref[1]**o+(-0.755586690837)*x_ref[1]+(-0.420481091654)*x_ref[2]**o+(-0.0996487371831)*x_ref[2]
        ref[0,3,1,1]=(-0.395961249009)*x_ref[0]**o+(0.0860896467863)*x_ref[0]+(0.482719525748)*x_ref[1]**o+(0.319368178133)*x_ref[1]+(0.516449153414)*x_ref[2]**o+(-0.246749862576)*x_ref[2]
        ref[0,3,2,0]=(-0.367322027122)*x_ref[0]**o+(0.687206412524)*x_ref[0]+(0.616265233972)*x_ref[1]**o+(-0.32660589112)*x_ref[1]+(-0.944607343096)*x_ref[2]**o+(0.123982537951)*x_ref[2]
        ref[0,3,2,1]=(0.700497166804)*x_ref[0]**o+(0.553961329865)*x_ref[0]+(-0.34020003754)*x_ref[1]**o+(0.295870992783)*x_ref[1]+(-0.220275880046)*x_ref[2]**o+(0.766121035825)*x_ref[2]
        ref[0,4,0,0]=(0.405469789242)*x_ref[0]**o+(-0.338347662873)*x_ref[0]+(0.0761796434233)*x_ref[1]**o+(-0.120599175033)*x_ref[1]+(0.107865585379)*x_ref[2]**o+(0.0711941007838)*x_ref[2]
        ref[0,4,0,1]=(0.614166154899)*x_ref[0]**o+(0.195780702963)*x_ref[0]+(0.718935985215)*x_ref[1]**o+(0.214577296792)*x_ref[1]+(-0.366985242107)*x_ref[2]**o+(-0.509584787633)*x_ref[2]
        ref[0,4,1,0]=(-0.638646370981)*x_ref[0]**o+(-0.769955899284)*x_ref[0]+(-0.887407540503)*x_ref[1]**o+(0.0901245429998)*x_ref[1]+(0.786085641321)*x_ref[2]**o+(0.6248068552)*x_ref[2]
        ref[0,4,1,1]=(-0.145426222043)*x_ref[0]**o+(-0.494114545016)*x_ref[0]+(0.422153721335)*x_ref[1]**o+(0.617022143901)*x_ref[1]+(-0.665972324723)*x_ref[2]**o+(-0.476724432799)*x_ref[2]
        ref[0,4,2,0]=(0.556105252092)*x_ref[0]**o+(-0.893268696809)*x_ref[0]+(0.146441889489)*x_ref[1]**o+(-0.170512113283)*x_ref[1]+(-0.869313262875)*x_ref[2]**o+(-0.976874109392)*x_ref[2]
        ref[0,4,2,1]=(0.858927206499)*x_ref[0]**o+(0.198998355146)*x_ref[0]+(0.278125296049)*x_ref[1]**o+(-0.0442518828093)*x_ref[1]+(0.274852724203)*x_ref[2]**o+(-0.167649193314)*x_ref[2]
        ref[1,0,0,0]=(-0.0429420289353)*x_ref[0]**o+(-0.15764693929)*x_ref[0]+(-0.859599917512)*x_ref[1]**o+(-0.904178643575)*x_ref[1]+(0.519560711493)*x_ref[2]**o+(0.258644717907)*x_ref[2]
        ref[1,0,0,1]=(-0.185321114264)*x_ref[0]**o+(-0.794493830004)*x_ref[0]+(0.136084421162)*x_ref[1]**o+(0.926912021419)*x_ref[1]+(0.346392120388)*x_ref[2]**o+(-0.183908371454)*x_ref[2]
        ref[1,0,1,0]=(-0.730675509877)*x_ref[0]**o+(0.865034802778)*x_ref[0]+(-0.506315431878)*x_ref[1]**o+(-0.94607761051)*x_ref[1]+(0.195194792801)*x_ref[2]**o+(-0.865129258879)*x_ref[2]
        ref[1,0,1,1]=(-0.0597026648107)*x_ref[0]**o+(-0.884790497)*x_ref[0]+(-0.0631278742098)*x_ref[1]**o+(-0.808364940737)*x_ref[1]+(-0.968952604837)*x_ref[2]**o+(0.142246702632)*x_ref[2]
        ref[1,0,2,0]=(0.648472894563)*x_ref[0]**o+(0.145861644073)*x_ref[0]+(0.501227520574)*x_ref[1]**o+(0.265886482306)*x_ref[1]+(-0.0472772173573)*x_ref[2]**o+(-0.364169655969)*x_ref[2]
        ref[1,0,2,1]=(0.847863600695)*x_ref[0]**o+(-0.288841685199)*x_ref[0]+(-0.500273502795)*x_ref[1]**o+(0.627379277592)*x_ref[1]+(0.326814024625)*x_ref[2]**o+(-0.514712825895)*x_ref[2]
        ref[1,1,0,0]=(-0.947845713712)*x_ref[0]**o+(-0.372115616305)*x_ref[0]+(-0.572858279998)*x_ref[1]**o+(0.226237627692)*x_ref[1]+(0.210201947182)*x_ref[2]**o+(0.0147616538594)*x_ref[2]
        ref[1,1,0,1]=(-0.39813630466)*x_ref[0]**o+(-0.658496395731)*x_ref[0]+(-0.363906465005)*x_ref[1]**o+(-0.816719661238)*x_ref[1]+(-0.478106835032)*x_ref[2]**o+(0.0819042309589)*x_ref[2]
        ref[1,1,1,0]=(-0.287160278593)*x_ref[0]**o+(-0.0281402640978)*x_ref[0]+(0.41559795378)*x_ref[1]**o+(0.158274458097)*x_ref[1]+(-0.798429481502)*x_ref[2]**o+(-0.278585334624)*x_ref[2]
        ref[1,1,1,1]=(0.809550480973)*x_ref[0]**o+(0.192026915854)*x_ref[0]+(-0.891168838599)*x_ref[1]**o+(0.0482621192882)*x_ref[1]+(-0.35339708159)*x_ref[2]**o+(-0.133161551553)*x_ref[2]
        ref[1,1,2,0]=(0.939582328872)*x_ref[0]**o+(0.561963790516)*x_ref[0]+(0.885264722949)*x_ref[1]**o+(-0.468042305266)*x_ref[1]+(0.232880002429)*x_ref[2]**o+(-0.772896005701)*x_ref[2]
        ref[1,1,2,1]=(-0.297333303209)*x_ref[0]**o+(-0.414050924207)*x_ref[0]+(-0.278903698873)*x_ref[1]**o+(0.345370620869)*x_ref[1]+(-0.0570151259027)*x_ref[2]**o+(0.786449641378)*x_ref[2]
        ref[1,2,0,0]=(0.644047138589)*x_ref[0]**o+(-0.480660438838)*x_ref[0]+(0.32526499462)*x_ref[1]**o+(-0.775988282127)*x_ref[1]+(0.108374143932)*x_ref[2]**o+(0.955463667181)*x_ref[2]
        ref[1,2,0,1]=(-0.468236790339)*x_ref[0]**o+(0.844197711724)*x_ref[0]+(0.808973429831)*x_ref[1]**o+(-0.647699550005)*x_ref[1]+(0.552555959813)*x_ref[2]**o+(-0.3403079515)*x_ref[2]
        ref[1,2,1,0]=(0.864686362308)*x_ref[0]**o+(0.248923646116)*x_ref[0]+(0.0287042863842)*x_ref[1]**o+(0.404974282153)*x_ref[1]+(0.113501438234)*x_ref[2]**o+(-0.619364894397)*x_ref[2]
        ref[1,2,1,1]=(-0.625146992499)*x_ref[0]**o+(-0.0480602452983)*x_ref[0]+(0.168910550914)*x_ref[1]**o+(-0.891584218123)*x_ref[1]+(0.0609357618263)*x_ref[2]**o+(0.0451921002161)*x_ref[2]
        ref[1,2,2,0]=(-0.478728628752)*x_ref[0]**o+(0.428376547798)*x_ref[0]+(0.817203376818)*x_ref[1]**o+(-0.856056266953)*x_ref[1]+(0.354960313382)*x_ref[2]**o+(-0.611468960461)*x_ref[2]
        ref[1,2,2,1]=(0.49500742113)*x_ref[0]**o+(0.902948130263)*x_ref[0]+(-0.810748543744)*x_ref[1]**o+(-0.696150924858)*x_ref[1]+(-0.748278516161)*x_ref[2]**o+(0.694815527058)*x_ref[2]
        ref[1,3,0,0]=(-0.361960150296)*x_ref[0]**o+(-0.845139780461)*x_ref[0]+(-0.0624499309665)*x_ref[1]**o+(-0.0515558058489)*x_ref[1]+(0.400690983017)*x_ref[2]**o+(-0.744608407974)*x_ref[2]
        ref[1,3,0,1]=(-0.452453617467)*x_ref[0]**o+(0.845396139593)*x_ref[0]+(0.292319469936)*x_ref[1]**o+(-0.833547518634)*x_ref[1]+(-0.511797700353)*x_ref[2]**o+(0.857891454773)*x_ref[2]
        ref[1,3,1,0]=(0.779504092912)*x_ref[0]**o+(0.500524773615)*x_ref[0]+(0.590542659382)*x_ref[1]**o+(0.479310630519)*x_ref[1]+(0.509960641956)*x_ref[2]**o+(-0.564980432959)*x_ref[2]
        ref[1,3,1,1]=(-0.145745189298)*x_ref[0]**o+(-0.256064576359)*x_ref[0]+(0.640890837126)*x_ref[1]**o+(-0.718802035152)*x_ref[1]+(-0.322216873583)*x_ref[2]**o+(-0.0246526526334)*x_ref[2]
        ref[1,3,2,0]=(-0.463434285098)*x_ref[0]**o+(0.116722448643)*x_ref[0]+(-0.690378381342)*x_ref[1]**o+(0.429197465885)*x_ref[1]+(-0.353019731384)*x_ref[2]**o+(0.57397712918)*x_ref[2]
        ref[1,3,2,1]=(0.169676853383)*x_ref[0]**o+(0.815225472623)*x_ref[0]+(-0.11125527552)*x_ref[1]**o+(0.391513831525)*x_ref[1]+(-0.587041361585)*x_ref[2]**o+(0.255194151616)*x_ref[2]
        ref[1,4,0,0]=(0.325351608628)*x_ref[0]**o+(-0.57751715453)*x_ref[0]+(0.145167511303)*x_ref[1]**o+(-0.969286030742)*x_ref[1]+(0.917370398478)*x_ref[2]**o+(0.691483691598)*x_ref[2]
        ref[1,4,0,1]=(0.0997998298271)*x_ref[0]**o+(-0.101248202247)*x_ref[0]+(-0.650649653848)*x_ref[1]**o+(-0.157887947288)*x_ref[1]+(0.517501747765)*x_ref[2]**o+(-0.0852921926298)*x_ref[2]
        ref[1,4,1,0]=(0.984595420407)*x_ref[0]**o+(0.499293499076)*x_ref[0]+(0.352193115874)*x_ref[1]**o+(-0.684522909647)*x_ref[1]+(-0.391538373605)*x_ref[2]**o+(0.404635595162)*x_ref[2]
        ref[1,4,1,1]=(-0.231310482173)*x_ref[0]**o+(-0.0385089836559)*x_ref[0]+(0.247671691888)*x_ref[1]**o+(0.883971955027)*x_ref[1]+(-0.0856624709734)*x_ref[2]**o+(0.942940875904)*x_ref[2]
        ref[1,4,2,0]=(-0.471467257434)*x_ref[0]**o+(0.835634240897)*x_ref[0]+(-0.594421294562)*x_ref[1]**o+(-0.187584387111)*x_ref[1]+(0.859867472765)*x_ref[2]**o+(0.930951154576)*x_ref[2]
        ref[1,4,2,1]=(0.504416975601)*x_ref[0]**o+(0.435692923635)*x_ref[0]+(-0.296945944313)*x_ref[1]**o+(0.815156218765)*x_ref[1]+(0.464464247442)*x_ref[2]**o+(-0.378809453041)*x_ref[2]
        ref[2,0,0,0]=(0.213087423107)*x_ref[0]**o+(-0.504681665801)*x_ref[0]+(0.796352190405)*x_ref[1]**o+(0.546752751136)*x_ref[1]+(0.57347644664)*x_ref[2]**o+(0.715856963195)*x_ref[2]
        ref[2,0,0,1]=(-0.28455266769)*x_ref[0]**o+(0.853162220766)*x_ref[0]+(0.164930982863)*x_ref[1]**o+(-0.0136996498765)*x_ref[1]+(0.293717915079)*x_ref[2]**o+(-0.216124158473)*x_ref[2]
        ref[2,0,1,0]=(-0.328115989441)*x_ref[0]**o+(-0.529441496488)*x_ref[0]+(-0.43579479677)*x_ref[1]**o+(-0.851333801194)*x_ref[1]+(-0.278406102972)*x_ref[2]**o+(0.947777471824)*x_ref[2]
        ref[2,0,1,1]=(-0.756438938789)*x_ref[0]**o+(-0.624332272735)*x_ref[0]+(-0.591638429565)*x_ref[1]**o+(-0.0608281771566)*x_ref[1]+(-0.852078784085)*x_ref[2]**o+(-0.526558964866)*x_ref[2]
        ref[2,0,2,0]=(-0.923892426969)*x_ref[0]**o+(0.788121899878)*x_ref[0]+(-0.446006331202)*x_ref[1]**o+(0.997210871574)*x_ref[1]+(0.356998606103)*x_ref[2]**o+(0.972643822823)*x_ref[2]
        ref[2,0,2,1]=(0.225183308571)*x_ref[0]**o+(-0.757505414777)*x_ref[0]+(-0.59573192074)*x_ref[1]**o+(0.57166154987)*x_ref[1]+(0.622343198809)*x_ref[2]**o+(0.60343970503)*x_ref[2]
        ref[2,1,0,0]=(0.51027702934)*x_ref[0]**o+(-0.711737679522)*x_ref[0]+(-0.211712579627)*x_ref[1]**o+(-0.495817605314)*x_ref[1]+(0.924425804272)*x_ref[2]**o+(0.753424835275)*x_ref[2]
        ref[2,1,0,1]=(0.476218252577)*x_ref[0]**o+(0.187579891557)*x_ref[0]+(0.727359583623)*x_ref[1]**o+(-0.389875559955)*x_ref[1]+(-0.352889326477)*x_ref[2]**o+(0.365143943637)*x_ref[2]
        ref[2,1,1,0]=(-0.815704312943)*x_ref[0]**o+(0.315217470245)*x_ref[0]+(-0.535234555065)*x_ref[1]**o+(-0.433408988501)*x_ref[1]+(-0.676694087389)*x_ref[2]**o+(0.384097177718)*x_ref[2]
        ref[2,1,1,1]=(-0.0368452411345)*x_ref[0]**o+(0.514540686444)*x_ref[0]+(-0.527100066965)*x_ref[1]**o+(0.0291015562912)*x_ref[1]+(-0.605646905004)*x_ref[2]**o+(0.931528599981)*x_ref[2]
        ref[2,1,2,0]=(-0.0994015945266)*x_ref[0]**o+(-0.762310347434)*x_ref[0]+(-0.845934950466)*x_ref[1]**o+(-0.235758967923)*x_ref[1]+(0.755531422698)*x_ref[2]**o+(0.108353867777)*x_ref[2]
        ref[2,1,2,1]=(-0.868779052889)*x_ref[0]**o+(0.783878941075)*x_ref[0]+(0.900089560473)*x_ref[1]**o+(-0.836779405877)*x_ref[1]+(-0.799571882449)*x_ref[2]**o+(0.999710029157)*x_ref[2]
        ref[2,2,0,0]=(0.541465277143)*x_ref[0]**o+(-0.840477797945)*x_ref[0]+(-0.362577721098)*x_ref[1]**o+(0.00680131806718)*x_ref[1]+(-0.338485239867)*x_ref[2]**o+(0.828213756751)*x_ref[2]
        ref[2,2,0,1]=(-0.927347423175)*x_ref[0]**o+(-0.936191125785)*x_ref[0]+(-0.411819804833)*x_ref[1]**o+(-0.0222067637269)*x_ref[1]+(-0.104411832586)*x_ref[2]**o+(-0.79937203081)*x_ref[2]
        ref[2,2,1,0]=(0.273051712848)*x_ref[0]**o+(-0.802468557957)*x_ref[0]+(-0.019392898867)*x_ref[1]**o+(-0.612803074784)*x_ref[1]+(-0.269705464223)*x_ref[2]**o+(-0.244405147706)*x_ref[2]
        ref[2,2,1,1]=(-0.08869649035)*x_ref[0]**o+(0.809761782447)*x_ref[0]+(0.761420080249)*x_ref[1]**o+(0.16590038557)*x_ref[1]+(-0.405657593445)*x_ref[2]**o+(0.123593603037)*x_ref[2]
        ref[2,2,2,0]=(-0.799691160987)*x_ref[0]**o+(0.834835690475)*x_ref[0]+(-0.0863569221737)*x_ref[1]**o+(0.988343215152)*x_ref[1]+(0.294423283474)*x_ref[2]**o+(0.325725138203)*x_ref[2]
        ref[2,2,2,1]=(0.648897496592)*x_ref[0]**o+(-0.474329387314)*x_ref[0]+(-0.904613697104)*x_ref[1]**o+(0.382201388286)*x_ref[1]+(-0.596889148339)*x_ref[2]**o+(0.705132157684)*x_ref[2]
        ref[2,3,0,0]=(0.906738144045)*x_ref[0]**o+(-0.365392026426)*x_ref[0]+(0.730557428733)*x_ref[1]**o+(-0.857864471189)*x_ref[1]+(-0.160930657186)*x_ref[2]**o+(0.885067383334)*x_ref[2]
        ref[2,3,0,1]=(0.992764702472)*x_ref[0]**o+(0.720688532948)*x_ref[0]+(0.318869772818)*x_ref[1]**o+(0.995508347248)*x_ref[1]+(0.108569456996)*x_ref[2]**o+(-0.990939292995)*x_ref[2]
        ref[2,3,1,0]=(0.690584992096)*x_ref[0]**o+(0.605885242246)*x_ref[0]+(0.72371794265)*x_ref[1]**o+(-0.632609325879)*x_ref[1]+(0.520273588734)*x_ref[2]**o+(0.885804143155)*x_ref[2]
        ref[2,3,1,1]=(0.485231542482)*x_ref[0]**o+(0.866920225418)*x_ref[0]+(-0.814964945972)*x_ref[1]**o+(0.544132513209)*x_ref[1]+(-0.155071376494)*x_ref[2]**o+(-0.941300746237)*x_ref[2]
        ref[2,3,2,0]=(0.748232238388)*x_ref[0]**o+(0.179397724419)*x_ref[0]+(-0.43881886248)*x_ref[1]**o+(-0.124941799624)*x_ref[1]+(0.14679335819)*x_ref[2]**o+(-0.758738846688)*x_ref[2]
        ref[2,3,2,1]=(-0.0951333655337)*x_ref[0]**o+(0.37270972476)*x_ref[0]+(-0.651977554115)*x_ref[1]**o+(-0.88918597476)*x_ref[1]+(-0.730858685576)*x_ref[2]**o+(0.685881924459)*x_ref[2]
        ref[2,4,0,0]=(-0.823487466802)*x_ref[0]**o+(-0.583630855643)*x_ref[0]+(-0.664814562747)*x_ref[1]**o+(0.860952605956)*x_ref[1]+(-0.249114322668)*x_ref[2]**o+(-0.968356584237)*x_ref[2]
        ref[2,4,0,1]=(0.44040146252)*x_ref[0]**o+(-0.246827565902)*x_ref[0]+(0.596300980733)*x_ref[1]**o+(-0.825818200414)*x_ref[1]+(0.0715512875967)*x_ref[2]**o+(0.996064679803)*x_ref[2]
        ref[2,4,1,0]=(-0.287806453021)*x_ref[0]**o+(0.704331542927)*x_ref[0]+(-0.346139754521)*x_ref[1]**o+(-0.766667037125)*x_ref[1]+(0.283874132105)*x_ref[2]**o+(-0.613145979204)*x_ref[2]
        ref[2,4,1,1]=(-0.116649875156)*x_ref[0]**o+(0.291380048239)*x_ref[0]+(0.689809221905)*x_ref[1]**o+(0.758935910592)*x_ref[1]+(0.505142787603)*x_ref[2]**o+(-0.113611480503)*x_ref[2]
        ref[2,4,2,0]=(-0.836908572692)*x_ref[0]**o+(-0.21892828725)*x_ref[0]+(-0.935780103868)*x_ref[1]**o+(0.106204643268)*x_ref[1]+(-0.894474423807)*x_ref[2]**o+(0.296554773217)*x_ref[2]
        ref[2,4,2,1]=(-0.89413495023)*x_ref[0]**o+(0.6080813058)*x_ref[0]+(0.087863889878)*x_ref[1]**o+(0.213373665612)*x_ref[1]+(0.17085345025)*x_ref[2]**o+(-0.967625138712)*x_ref[2]
        ref[3,0,0,0]=(0.327537677409)*x_ref[0]**o+(-0.299373059373)*x_ref[0]+(-0.43223648542)*x_ref[1]**o+(-0.499115781051)*x_ref[1]+(0.145464567449)*x_ref[2]**o+(0.580930584563)*x_ref[2]
        ref[3,0,0,1]=(-0.122622543351)*x_ref[0]**o+(-0.220547821081)*x_ref[0]+(0.892679827803)*x_ref[1]**o+(0.436584604381)*x_ref[1]+(-0.156093580288)*x_ref[2]**o+(0.464069069075)*x_ref[2]
        ref[3,0,1,0]=(0.590288663203)*x_ref[0]**o+(-0.212929445251)*x_ref[0]+(0.0899671639061)*x_ref[1]**o+(-0.87729854872)*x_ref[1]+(0.789506558716)*x_ref[2]**o+(-0.68867690907)*x_ref[2]
        ref[3,0,1,1]=(-0.913651849012)*x_ref[0]**o+(-0.430938918892)*x_ref[0]+(-0.351186012371)*x_ref[1]**o+(0.500306994748)*x_ref[1]+(0.145381845281)*x_ref[2]**o+(0.624993552333)*x_ref[2]
        ref[3,0,2,0]=(0.281393881265)*x_ref[0]**o+(-0.704727129826)*x_ref[0]+(-0.0102091836449)*x_ref[1]**o+(0.311857501353)*x_ref[1]+(0.300732947904)*x_ref[2]**o+(0.923973499548)*x_ref[2]
        ref[3,0,2,1]=(-0.584907526993)*x_ref[0]**o+(0.625732316822)*x_ref[0]+(-0.609691930408)*x_ref[1]**o+(0.056160226956)*x_ref[1]+(-0.766583561947)*x_ref[2]**o+(0.307692190194)*x_ref[2]
        ref[3,1,0,0]=(0.731231379445)*x_ref[0]**o+(-0.0485980527115)*x_ref[0]+(0.908932758852)*x_ref[1]**o+(0.188918035886)*x_ref[1]+(0.685837743688)*x_ref[2]**o+(-0.299987553989)*x_ref[2]
        ref[3,1,0,1]=(-0.784006561231)*x_ref[0]**o+(-0.809789430022)*x_ref[0]+(-0.846391792735)*x_ref[1]**o+(-0.211583304405)*x_ref[1]+(-0.00198590063687)*x_ref[2]**o+(-0.853253255224)*x_ref[2]
        ref[3,1,1,0]=(0.595567299132)*x_ref[0]**o+(-0.581734416878)*x_ref[0]+(0.88144481486)*x_ref[1]**o+(0.776876564509)*x_ref[1]+(-0.367477971386)*x_ref[2]**o+(0.0806320164453)*x_ref[2]
        ref[3,1,1,1]=(0.440836187611)*x_ref[0]**o+(-0.578925091757)*x_ref[0]+(0.923367189832)*x_ref[1]**o+(-0.993888173929)*x_ref[1]+(-0.645514380411)*x_ref[2]**o+(0.0267439044199)*x_ref[2]
        ref[3,1,2,0]=(-0.305182288788)*x_ref[0]**o+(-0.503372502553)*x_ref[0]+(0.740948657389)*x_ref[1]**o+(0.285661052275)*x_ref[1]+(-0.634510984218)*x_ref[2]**o+(-0.123979746456)*x_ref[2]
        ref[3,1,2,1]=(-0.897274928495)*x_ref[0]**o+(0.971356796164)*x_ref[0]+(0.940768290313)*x_ref[1]**o+(-0.552941985667)*x_ref[1]+(0.172857451423)*x_ref[2]**o+(0.844758237523)*x_ref[2]
        ref[3,2,0,0]=(-0.903542725333)*x_ref[0]**o+(0.936704774551)*x_ref[0]+(-0.398085875228)*x_ref[1]**o+(-0.304061573512)*x_ref[1]+(-0.13362037229)*x_ref[2]**o+(-0.405043350728)*x_ref[2]
        ref[3,2,0,1]=(-0.0208590751801)*x_ref[0]**o+(-0.354086381885)*x_ref[0]+(0.510884698744)*x_ref[1]**o+(-0.275504880518)*x_ref[1]+(0.459476496356)*x_ref[2]**o+(0.25358097976)*x_ref[2]
        ref[3,2,1,0]=(-0.0171081877107)*x_ref[0]**o+(-0.856081626112)*x_ref[0]+(-0.0698686571004)*x_ref[1]**o+(0.0954918361122)*x_ref[1]+(0.938494954069)*x_ref[2]**o+(-0.824928688248)*x_ref[2]
        ref[3,2,1,1]=(-0.277307915098)*x_ref[0]**o+(0.737855779658)*x_ref[0]+(-0.466761948855)*x_ref[1]**o+(-0.821498435809)*x_ref[1]+(-0.391009601442)*x_ref[2]**o+(0.643834542048)*x_ref[2]
        ref[3,2,2,0]=(0.379015007885)*x_ref[0]**o+(0.348646179575)*x_ref[0]+(0.733765133901)*x_ref[1]**o+(-0.722363571789)*x_ref[1]+(0.488531791931)*x_ref[2]**o+(0.279813923679)*x_ref[2]
        ref[3,2,2,1]=(0.516708382119)*x_ref[0]**o+(0.0426665103314)*x_ref[0]+(-0.168751315718)*x_ref[1]**o+(-0.124069236608)*x_ref[1]+(-0.778866001674)*x_ref[2]**o+(-0.362031001279)*x_ref[2]
        ref[3,3,0,0]=(0.995284803465)*x_ref[0]**o+(-0.442327755879)*x_ref[0]+(-0.13110270925)*x_ref[1]**o+(-0.172857543102)*x_ref[1]+(-0.910522664379)*x_ref[2]**o+(0.516931833163)*x_ref[2]
        ref[3,3,0,1]=(-0.889320295626)*x_ref[0]**o+(-0.728330379534)*x_ref[0]+(-0.584091700627)*x_ref[1]**o+(0.469914082337)*x_ref[1]+(0.18874127891)*x_ref[2]**o+(0.0846532039249)*x_ref[2]
        ref[3,3,1,0]=(0.911523535462)*x_ref[0]**o+(0.0924361220354)*x_ref[0]+(-0.0822504494228)*x_ref[1]**o+(0.00630249023591)*x_ref[1]+(-0.180497330322)*x_ref[2]**o+(0.817405577154)*x_ref[2]
        ref[3,3,1,1]=(0.681872747813)*x_ref[0]**o+(0.840680190206)*x_ref[0]+(-0.477660295206)*x_ref[1]**o+(0.0399086968056)*x_ref[1]+(0.440403204247)*x_ref[2]**o+(-0.220656197016)*x_ref[2]
        ref[3,3,2,0]=(-0.967524407336)*x_ref[0]**o+(0.896652274462)*x_ref[0]+(0.735269328519)*x_ref[1]**o+(-0.249017384614)*x_ref[1]+(-0.573222311563)*x_ref[2]**o+(-0.653800334954)*x_ref[2]
        ref[3,3,2,1]=(0.10041632804)*x_ref[0]**o+(0.167657178348)*x_ref[0]+(0.990621134594)*x_ref[1]**o+(-0.309063473348)*x_ref[1]+(-0.688206827575)*x_ref[2]**o+(0.0926806818663)*x_ref[2]
        ref[3,4,0,0]=(-0.128678339518)*x_ref[0]**o+(-0.974238160059)*x_ref[0]+(-0.762599464772)*x_ref[1]**o+(0.0233286358698)*x_ref[1]+(-0.622481697973)*x_ref[2]**o+(-0.080495303798)*x_ref[2]
        ref[3,4,0,1]=(-0.380318230058)*x_ref[0]**o+(0.568425682425)*x_ref[0]+(0.787369916262)*x_ref[1]**o+(0.296745245123)*x_ref[1]+(-0.917621002414)*x_ref[2]**o+(-0.696209404954)*x_ref[2]
        ref[3,4,1,0]=(-0.382961468898)*x_ref[0]**o+(0.672614453621)*x_ref[0]+(0.565371503874)*x_ref[1]**o+(-0.597331999912)*x_ref[1]+(-0.153637846704)*x_ref[2]**o+(-0.438159587906)*x_ref[2]
        ref[3,4,1,1]=(-0.541043007799)*x_ref[0]**o+(0.415600628159)*x_ref[0]+(-0.793716270785)*x_ref[1]**o+(0.780110667203)*x_ref[1]+(-0.101309486185)*x_ref[2]**o+(-0.138868352939)*x_ref[2]
        ref[3,4,2,0]=(0.679815917569)*x_ref[0]**o+(-0.800657537776)*x_ref[0]+(-0.711956069893)*x_ref[1]**o+(-0.575626146908)*x_ref[1]+(0.184247083249)*x_ref[2]**o+(-0.728118084396)*x_ref[2]
        ref[3,4,2,1]=(-0.608089432748)*x_ref[0]**o+(0.356378798828)*x_ref[0]+(-0.575957599599)*x_ref[1]**o+(0.599586825137)*x_ref[1]+(-0.160475743299)*x_ref[2]**o+(-0.070762356177)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.52738736047)*x[0]**o+(-0.513351179641)*x[0]+(0.466974564277)*x[1]**o+(0.247665236281)*x[1]
        ref=(0.52738736047)*x_ref[0]**o+(-0.513351179641)*x_ref[0]+(0.466974564277)*x_ref[1]**o+(0.247665236281)*x_ref[1]
      else:
        arg=(0.345612544612)*x[0]**o+(0.480207854215)*x[0]+(0.196672493786)*x[1]**o+(-0.891043131902)*x[1]+(0.106095952827)*x[2]**o+(0.436049931061)*x[2]
        ref=(0.345612544612)*x_ref[0]**o+(0.480207854215)*x_ref[0]+(0.196672493786)*x_ref[1]**o+(-0.891043131902)*x_ref[1]+(0.106095952827)*x_ref[2]**o+(0.436049931061)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.738262816349)*x[0]**o+(-0.650523560236)*x[0]+(-0.344136056546)*x[1]**o+(-0.0693168099812)*x[1]
        arg[1]=(-0.565224044511)*x[0]**o+(0.94461728731)*x[0]+(0.234800891741)*x[1]**o+(0.836341748852)*x[1]
        ref[0]=(-0.738262816349)*x_ref[0]**o+(-0.650523560236)*x_ref[0]+(-0.344136056546)*x_ref[1]**o+(-0.0693168099812)*x_ref[1]
        ref[1]=(-0.565224044511)*x_ref[0]**o+(0.94461728731)*x_ref[0]+(0.234800891741)*x_ref[1]**o+(0.836341748852)*x_ref[1]
      else:
        arg[0]=(-0.76925788678)*x[0]**o+(-0.391133765534)*x[0]+(0.89961543807)*x[1]**o+(-0.355090830059)*x[1]+(-0.663969382551)*x[2]**o+(0.0324773110608)*x[2]
        arg[1]=(-0.640271335706)*x[0]**o+(-0.0771899290795)*x[0]+(0.923847889924)*x[1]**o+(-0.732484895713)*x[1]+(-0.782800597476)*x[2]**o+(-0.402230990251)*x[2]
        ref[0]=(-0.76925788678)*x_ref[0]**o+(-0.391133765534)*x_ref[0]+(0.89961543807)*x_ref[1]**o+(-0.355090830059)*x_ref[1]+(-0.663969382551)*x_ref[2]**o+(0.0324773110608)*x_ref[2]
        ref[1]=(-0.640271335706)*x_ref[0]**o+(-0.0771899290795)*x_ref[0]+(0.923847889924)*x_ref[1]**o+(-0.732484895713)*x_ref[1]+(-0.782800597476)*x_ref[2]**o+(-0.402230990251)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.631349529615)*x[0]**o+(0.79216890406)*x[0]+(0.835421397222)*x[1]**o+(-0.209368192458)*x[1]
        arg[0,1]=(-0.166321945093)*x[0]**o+(0.701821475941)*x[0]+(-0.141804772049)*x[1]**o+(-0.860056917696)*x[1]
        arg[0,2]=(0.761659971379)*x[0]**o+(-0.39989119778)*x[0]+(-0.761140588562)*x[1]**o+(-0.595816184788)*x[1]
        arg[0,3]=(0.378518095375)*x[0]**o+(-0.880144120304)*x[0]+(0.814409656431)*x[1]**o+(0.0464425920173)*x[1]
        arg[0,4]=(-0.24431693596)*x[0]**o+(-0.840538735294)*x[0]+(0.0157782225732)*x[1]**o+(0.951311376091)*x[1]
        arg[1,0]=(0.0710042795971)*x[0]**o+(0.936106708209)*x[0]+(-0.580035969982)*x[1]**o+(0.740790378647)*x[1]
        arg[1,1]=(-0.807938035414)*x[0]**o+(0.8723236137)*x[0]+(-0.600615307712)*x[1]**o+(0.71483191921)*x[1]
        arg[1,2]=(0.454239081542)*x[0]**o+(0.966475397079)*x[0]+(-0.184649384525)*x[1]**o+(0.417686859721)*x[1]
        arg[1,3]=(-0.394999073966)*x[0]**o+(-0.0445825855756)*x[0]+(0.0389091203001)*x[1]**o+(-0.265569544304)*x[1]
        arg[1,4]=(0.625419856529)*x[0]**o+(-0.297955079048)*x[0]+(0.0092181343855)*x[1]**o+(-0.423768484981)*x[1]
        arg[2,0]=(0.836466109456)*x[0]**o+(-0.12515777587)*x[0]+(-0.0908815675065)*x[1]**o+(0.825332480799)*x[1]
        arg[2,1]=(-0.454612353892)*x[0]**o+(0.952942366642)*x[0]+(0.538638303761)*x[1]**o+(0.377957206795)*x[1]
        arg[2,2]=(-0.229405757447)*x[0]**o+(-0.125249012236)*x[0]+(0.194040765277)*x[1]**o+(-0.77353897054)*x[1]
        arg[2,3]=(-0.220514078279)*x[0]**o+(-0.937043129926)*x[0]+(0.661270617907)*x[1]**o+(0.342981486241)*x[1]
        arg[2,4]=(0.58858161977)*x[0]**o+(0.324813259161)*x[0]+(0.00439371077545)*x[1]**o+(0.517353738278)*x[1]
        arg[3,0]=(0.130001725331)*x[0]**o+(0.512570729718)*x[0]+(0.971292679928)*x[1]**o+(0.302530140373)*x[1]
        arg[3,1]=(0.999138842976)*x[0]**o+(0.325794749172)*x[0]+(0.94900982582)*x[1]**o+(0.56933825455)*x[1]
        arg[3,2]=(0.965201839092)*x[0]**o+(0.434187660328)*x[0]+(0.842238794411)*x[1]**o+(-0.754540335284)*x[1]
        arg[3,3]=(-0.539162954093)*x[0]**o+(-0.373779084363)*x[0]+(-0.426189858865)*x[1]**o+(-0.812351658629)*x[1]
        arg[3,4]=(0.871516777541)*x[0]**o+(-0.357393696308)*x[0]+(0.267358318776)*x[1]**o+(0.796701523616)*x[1]
        ref[0,0]=(-0.631349529615)*x_ref[0]**o+(0.79216890406)*x_ref[0]+(0.835421397222)*x_ref[1]**o+(-0.209368192458)*x_ref[1]
        ref[0,1]=(-0.166321945093)*x_ref[0]**o+(0.701821475941)*x_ref[0]+(-0.141804772049)*x_ref[1]**o+(-0.860056917696)*x_ref[1]
        ref[0,2]=(0.761659971379)*x_ref[0]**o+(-0.39989119778)*x_ref[0]+(-0.761140588562)*x_ref[1]**o+(-0.595816184788)*x_ref[1]
        ref[0,3]=(0.378518095375)*x_ref[0]**o+(-0.880144120304)*x_ref[0]+(0.814409656431)*x_ref[1]**o+(0.0464425920173)*x_ref[1]
        ref[0,4]=(-0.24431693596)*x_ref[0]**o+(-0.840538735294)*x_ref[0]+(0.0157782225732)*x_ref[1]**o+(0.951311376091)*x_ref[1]
        ref[1,0]=(0.0710042795971)*x_ref[0]**o+(0.936106708209)*x_ref[0]+(-0.580035969982)*x_ref[1]**o+(0.740790378647)*x_ref[1]
        ref[1,1]=(-0.807938035414)*x_ref[0]**o+(0.8723236137)*x_ref[0]+(-0.600615307712)*x_ref[1]**o+(0.71483191921)*x_ref[1]
        ref[1,2]=(0.454239081542)*x_ref[0]**o+(0.966475397079)*x_ref[0]+(-0.184649384525)*x_ref[1]**o+(0.417686859721)*x_ref[1]
        ref[1,3]=(-0.394999073966)*x_ref[0]**o+(-0.0445825855756)*x_ref[0]+(0.0389091203001)*x_ref[1]**o+(-0.265569544304)*x_ref[1]
        ref[1,4]=(0.625419856529)*x_ref[0]**o+(-0.297955079048)*x_ref[0]+(0.0092181343855)*x_ref[1]**o+(-0.423768484981)*x_ref[1]
        ref[2,0]=(0.836466109456)*x_ref[0]**o+(-0.12515777587)*x_ref[0]+(-0.0908815675065)*x_ref[1]**o+(0.825332480799)*x_ref[1]
        ref[2,1]=(-0.454612353892)*x_ref[0]**o+(0.952942366642)*x_ref[0]+(0.538638303761)*x_ref[1]**o+(0.377957206795)*x_ref[1]
        ref[2,2]=(-0.229405757447)*x_ref[0]**o+(-0.125249012236)*x_ref[0]+(0.194040765277)*x_ref[1]**o+(-0.77353897054)*x_ref[1]
        ref[2,3]=(-0.220514078279)*x_ref[0]**o+(-0.937043129926)*x_ref[0]+(0.661270617907)*x_ref[1]**o+(0.342981486241)*x_ref[1]
        ref[2,4]=(0.58858161977)*x_ref[0]**o+(0.324813259161)*x_ref[0]+(0.00439371077545)*x_ref[1]**o+(0.517353738278)*x_ref[1]
        ref[3,0]=(0.130001725331)*x_ref[0]**o+(0.512570729718)*x_ref[0]+(0.971292679928)*x_ref[1]**o+(0.302530140373)*x_ref[1]
        ref[3,1]=(0.999138842976)*x_ref[0]**o+(0.325794749172)*x_ref[0]+(0.94900982582)*x_ref[1]**o+(0.56933825455)*x_ref[1]
        ref[3,2]=(0.965201839092)*x_ref[0]**o+(0.434187660328)*x_ref[0]+(0.842238794411)*x_ref[1]**o+(-0.754540335284)*x_ref[1]
        ref[3,3]=(-0.539162954093)*x_ref[0]**o+(-0.373779084363)*x_ref[0]+(-0.426189858865)*x_ref[1]**o+(-0.812351658629)*x_ref[1]
        ref[3,4]=(0.871516777541)*x_ref[0]**o+(-0.357393696308)*x_ref[0]+(0.267358318776)*x_ref[1]**o+(0.796701523616)*x_ref[1]
      else:
        arg[0,0]=(-0.811142944724)*x[0]**o+(-0.65659407492)*x[0]+(-0.534808740191)*x[1]**o+(0.602252100451)*x[1]+(-0.158845636056)*x[2]**o+(-0.781826300399)*x[2]
        arg[0,1]=(-0.396870942248)*x[0]**o+(0.285698519063)*x[0]+(-0.962418369859)*x[1]**o+(-0.729381989807)*x[1]+(-0.0500038851534)*x[2]**o+(-0.567828852526)*x[2]
        arg[0,2]=(0.346399694372)*x[0]**o+(-0.170517233373)*x[0]+(-0.443275920724)*x[1]**o+(0.581464075295)*x[1]+(0.197230389253)*x[2]**o+(-0.613092248913)*x[2]
        arg[0,3]=(0.920174147113)*x[0]**o+(0.429847401539)*x[0]+(0.112888577662)*x[1]**o+(-0.0540034503714)*x[1]+(-0.0323244631044)*x[2]**o+(-0.293609544741)*x[2]
        arg[0,4]=(-0.0177585560206)*x[0]**o+(-0.806033755558)*x[0]+(-0.363921399918)*x[1]**o+(-0.415721267967)*x[1]+(0.832302676307)*x[2]**o+(-0.102510351741)*x[2]
        arg[1,0]=(-0.769760277763)*x[0]**o+(0.818187369213)*x[0]+(0.454400795864)*x[1]**o+(-0.220119708184)*x[1]+(-0.637117949075)*x[2]**o+(-0.661398404641)*x[2]
        arg[1,1]=(0.828424952723)*x[0]**o+(0.677022266764)*x[0]+(-0.958176510992)*x[1]**o+(-0.732014776269)*x[1]+(-0.495308153354)*x[2]**o+(-0.30698327016)*x[2]
        arg[1,2]=(-0.454106379068)*x[0]**o+(0.686920350455)*x[0]+(0.560344054206)*x[1]**o+(0.355495451775)*x[1]+(0.256969568039)*x[2]**o+(-0.737279585211)*x[2]
        arg[1,3]=(-0.976488579746)*x[0]**o+(-0.652812816758)*x[0]+(-0.996243459964)*x[1]**o+(0.631565668939)*x[1]+(0.996259991179)*x[2]**o+(-0.796610599165)*x[2]
        arg[1,4]=(0.676024716135)*x[0]**o+(-0.179835226483)*x[0]+(0.497541872254)*x[1]**o+(-0.678345104068)*x[1]+(0.945491022682)*x[2]**o+(-0.358403760669)*x[2]
        arg[2,0]=(0.166901676041)*x[0]**o+(0.833434173485)*x[0]+(0.622153751777)*x[1]**o+(0.767679765059)*x[1]+(-0.82271335395)*x[2]**o+(0.667651471482)*x[2]
        arg[2,1]=(-0.405015366041)*x[0]**o+(0.0548291583154)*x[0]+(0.894878845805)*x[1]**o+(-0.0955982667798)*x[1]+(0.468927799304)*x[2]**o+(0.547421481564)*x[2]
        arg[2,2]=(0.893319250823)*x[0]**o+(0.771073524295)*x[0]+(0.993876715991)*x[1]**o+(0.0832975372419)*x[1]+(-0.717061635157)*x[2]**o+(0.303760934227)*x[2]
        arg[2,3]=(0.247006018272)*x[0]**o+(0.329618310414)*x[0]+(-0.824176211089)*x[1]**o+(-0.678238702323)*x[1]+(0.218736302876)*x[2]**o+(0.8652063227)*x[2]
        arg[2,4]=(0.502380810223)*x[0]**o+(-0.94917091185)*x[0]+(-0.0392659831743)*x[1]**o+(-0.308291910503)*x[1]+(-0.498112428439)*x[2]**o+(0.217915791465)*x[2]
        arg[3,0]=(-0.018497051159)*x[0]**o+(-0.543005709743)*x[0]+(-0.433773552307)*x[1]**o+(-0.604850191015)*x[1]+(0.285970270117)*x[2]**o+(0.150938308283)*x[2]
        arg[3,1]=(0.185537796564)*x[0]**o+(0.663947686119)*x[0]+(-0.846597672873)*x[1]**o+(-0.798399069258)*x[1]+(0.0215106260597)*x[2]**o+(-0.388760513164)*x[2]
        arg[3,2]=(-0.378320218313)*x[0]**o+(-0.385079307834)*x[0]+(-0.392389046498)*x[1]**o+(0.323305501943)*x[1]+(0.496285813211)*x[2]**o+(-0.0269835921214)*x[2]
        arg[3,3]=(0.779877220393)*x[0]**o+(0.968536241033)*x[0]+(0.786166009582)*x[1]**o+(-0.974838905578)*x[1]+(0.4579382269)*x[2]**o+(0.178464213473)*x[2]
        arg[3,4]=(-0.149895758057)*x[0]**o+(0.3529889311)*x[0]+(0.223924981441)*x[1]**o+(0.10143440581)*x[1]+(0.162222746832)*x[2]**o+(0.421433640986)*x[2]
        ref[0,0]=(-0.811142944724)*x_ref[0]**o+(-0.65659407492)*x_ref[0]+(-0.534808740191)*x_ref[1]**o+(0.602252100451)*x_ref[1]+(-0.158845636056)*x_ref[2]**o+(-0.781826300399)*x_ref[2]
        ref[0,1]=(-0.396870942248)*x_ref[0]**o+(0.285698519063)*x_ref[0]+(-0.962418369859)*x_ref[1]**o+(-0.729381989807)*x_ref[1]+(-0.0500038851534)*x_ref[2]**o+(-0.567828852526)*x_ref[2]
        ref[0,2]=(0.346399694372)*x_ref[0]**o+(-0.170517233373)*x_ref[0]+(-0.443275920724)*x_ref[1]**o+(0.581464075295)*x_ref[1]+(0.197230389253)*x_ref[2]**o+(-0.613092248913)*x_ref[2]
        ref[0,3]=(0.920174147113)*x_ref[0]**o+(0.429847401539)*x_ref[0]+(0.112888577662)*x_ref[1]**o+(-0.0540034503714)*x_ref[1]+(-0.0323244631044)*x_ref[2]**o+(-0.293609544741)*x_ref[2]
        ref[0,4]=(-0.0177585560206)*x_ref[0]**o+(-0.806033755558)*x_ref[0]+(-0.363921399918)*x_ref[1]**o+(-0.415721267967)*x_ref[1]+(0.832302676307)*x_ref[2]**o+(-0.102510351741)*x_ref[2]
        ref[1,0]=(-0.769760277763)*x_ref[0]**o+(0.818187369213)*x_ref[0]+(0.454400795864)*x_ref[1]**o+(-0.220119708184)*x_ref[1]+(-0.637117949075)*x_ref[2]**o+(-0.661398404641)*x_ref[2]
        ref[1,1]=(0.828424952723)*x_ref[0]**o+(0.677022266764)*x_ref[0]+(-0.958176510992)*x_ref[1]**o+(-0.732014776269)*x_ref[1]+(-0.495308153354)*x_ref[2]**o+(-0.30698327016)*x_ref[2]
        ref[1,2]=(-0.454106379068)*x_ref[0]**o+(0.686920350455)*x_ref[0]+(0.560344054206)*x_ref[1]**o+(0.355495451775)*x_ref[1]+(0.256969568039)*x_ref[2]**o+(-0.737279585211)*x_ref[2]
        ref[1,3]=(-0.976488579746)*x_ref[0]**o+(-0.652812816758)*x_ref[0]+(-0.996243459964)*x_ref[1]**o+(0.631565668939)*x_ref[1]+(0.996259991179)*x_ref[2]**o+(-0.796610599165)*x_ref[2]
        ref[1,4]=(0.676024716135)*x_ref[0]**o+(-0.179835226483)*x_ref[0]+(0.497541872254)*x_ref[1]**o+(-0.678345104068)*x_ref[1]+(0.945491022682)*x_ref[2]**o+(-0.358403760669)*x_ref[2]
        ref[2,0]=(0.166901676041)*x_ref[0]**o+(0.833434173485)*x_ref[0]+(0.622153751777)*x_ref[1]**o+(0.767679765059)*x_ref[1]+(-0.82271335395)*x_ref[2]**o+(0.667651471482)*x_ref[2]
        ref[2,1]=(-0.405015366041)*x_ref[0]**o+(0.0548291583154)*x_ref[0]+(0.894878845805)*x_ref[1]**o+(-0.0955982667798)*x_ref[1]+(0.468927799304)*x_ref[2]**o+(0.547421481564)*x_ref[2]
        ref[2,2]=(0.893319250823)*x_ref[0]**o+(0.771073524295)*x_ref[0]+(0.993876715991)*x_ref[1]**o+(0.0832975372419)*x_ref[1]+(-0.717061635157)*x_ref[2]**o+(0.303760934227)*x_ref[2]
        ref[2,3]=(0.247006018272)*x_ref[0]**o+(0.329618310414)*x_ref[0]+(-0.824176211089)*x_ref[1]**o+(-0.678238702323)*x_ref[1]+(0.218736302876)*x_ref[2]**o+(0.8652063227)*x_ref[2]
        ref[2,4]=(0.502380810223)*x_ref[0]**o+(-0.94917091185)*x_ref[0]+(-0.0392659831743)*x_ref[1]**o+(-0.308291910503)*x_ref[1]+(-0.498112428439)*x_ref[2]**o+(0.217915791465)*x_ref[2]
        ref[3,0]=(-0.018497051159)*x_ref[0]**o+(-0.543005709743)*x_ref[0]+(-0.433773552307)*x_ref[1]**o+(-0.604850191015)*x_ref[1]+(0.285970270117)*x_ref[2]**o+(0.150938308283)*x_ref[2]
        ref[3,1]=(0.185537796564)*x_ref[0]**o+(0.663947686119)*x_ref[0]+(-0.846597672873)*x_ref[1]**o+(-0.798399069258)*x_ref[1]+(0.0215106260597)*x_ref[2]**o+(-0.388760513164)*x_ref[2]
        ref[3,2]=(-0.378320218313)*x_ref[0]**o+(-0.385079307834)*x_ref[0]+(-0.392389046498)*x_ref[1]**o+(0.323305501943)*x_ref[1]+(0.496285813211)*x_ref[2]**o+(-0.0269835921214)*x_ref[2]
        ref[3,3]=(0.779877220393)*x_ref[0]**o+(0.968536241033)*x_ref[0]+(0.786166009582)*x_ref[1]**o+(-0.974838905578)*x_ref[1]+(0.4579382269)*x_ref[2]**o+(0.178464213473)*x_ref[2]
        ref[3,4]=(-0.149895758057)*x_ref[0]**o+(0.3529889311)*x_ref[0]+(0.223924981441)*x_ref[1]**o+(0.10143440581)*x_ref[1]+(0.162222746832)*x_ref[2]**o+(0.421433640986)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.232971584272)*x[0]**o+(0.391351964909)*x[0]+(-0.832254448983)*x[1]**o+(0.880762034713)*x[1]
        arg[0,0,1]=(0.473097471101)*x[0]**o+(-0.660862772376)*x[0]+(0.72974853865)*x[1]**o+(-0.328183435687)*x[1]
        arg[0,1,0]=(-0.749483736482)*x[0]**o+(0.0872454020187)*x[0]+(0.6916024322)*x[1]**o+(0.82757019273)*x[1]
        arg[0,1,1]=(-0.325601366346)*x[0]**o+(0.918246327277)*x[0]+(0.855963838252)*x[1]**o+(0.448567464706)*x[1]
        arg[1,0,0]=(-0.6642143628)*x[0]**o+(-0.82649833447)*x[0]+(0.822778465465)*x[1]**o+(-0.926669226842)*x[1]
        arg[1,0,1]=(-0.644490899373)*x[0]**o+(0.0644054772987)*x[0]+(-0.914922609343)*x[1]**o+(0.898880463842)*x[1]
        arg[1,1,0]=(0.40963096867)*x[0]**o+(-0.347418076162)*x[0]+(0.857966073118)*x[1]**o+(-0.284521442085)*x[1]
        arg[1,1,1]=(-0.00998062171478)*x[0]**o+(-0.234122404174)*x[0]+(-0.149500951208)*x[1]**o+(0.680038328937)*x[1]
        arg[2,0,0]=(-0.963777985833)*x[0]**o+(0.710727608341)*x[0]+(-0.646898391724)*x[1]**o+(-0.464095155503)*x[1]
        arg[2,0,1]=(-0.617670394851)*x[0]**o+(0.592337321587)*x[0]+(-0.0881426217074)*x[1]**o+(-0.00431912882662)*x[1]
        arg[2,1,0]=(0.871279828121)*x[0]**o+(-0.858267190204)*x[0]+(0.440863109038)*x[1]**o+(-0.912521398099)*x[1]
        arg[2,1,1]=(0.17802472174)*x[0]**o+(0.948587960042)*x[0]+(-0.837945850976)*x[1]**o+(0.13308730872)*x[1]
        arg[3,0,0]=(0.0482409149433)*x[0]**o+(-0.198100088194)*x[0]+(0.504789334224)*x[1]**o+(0.296276808608)*x[1]
        arg[3,0,1]=(-0.30515363985)*x[0]**o+(-0.355293167256)*x[0]+(-0.577592997685)*x[1]**o+(-0.882408725943)*x[1]
        arg[3,1,0]=(0.76102937294)*x[0]**o+(-0.210169673144)*x[0]+(-0.181755140891)*x[1]**o+(0.865448484187)*x[1]
        arg[3,1,1]=(-0.894294091863)*x[0]**o+(-0.712245666103)*x[0]+(0.151657455329)*x[1]**o+(-0.54019436763)*x[1]
        arg[4,0,0]=(0.860662861318)*x[0]**o+(-0.36787473166)*x[0]+(0.240406823985)*x[1]**o+(-0.381822992473)*x[1]
        arg[4,0,1]=(-0.109197125112)*x[0]**o+(0.962595232385)*x[0]+(-0.323876506847)*x[1]**o+(0.178613363564)*x[1]
        arg[4,1,0]=(0.486980313424)*x[0]**o+(-0.611294566285)*x[0]+(-0.774813150316)*x[1]**o+(0.335411840405)*x[1]
        arg[4,1,1]=(-0.622163740867)*x[0]**o+(0.426305234663)*x[0]+(0.243557551101)*x[1]**o+(0.930601340004)*x[1]
        arg[5,0,0]=(-0.232795962027)*x[0]**o+(0.49010767268)*x[0]+(-0.503693305594)*x[1]**o+(-0.915931668186)*x[1]
        arg[5,0,1]=(-0.304614759138)*x[0]**o+(0.797445765321)*x[0]+(-0.00820825778888)*x[1]**o+(-0.196762463972)*x[1]
        arg[5,1,0]=(-0.516700339456)*x[0]**o+(-0.440953905414)*x[0]+(0.875266018165)*x[1]**o+(0.780895354543)*x[1]
        arg[5,1,1]=(0.219827931399)*x[0]**o+(-0.950514094131)*x[0]+(-0.120763970425)*x[1]**o+(0.0941571726596)*x[1]
        ref[0,0,0]=(-0.232971584272)*x_ref[0]**o+(0.391351964909)*x_ref[0]+(-0.832254448983)*x_ref[1]**o+(0.880762034713)*x_ref[1]
        ref[0,0,1]=(0.473097471101)*x_ref[0]**o+(-0.660862772376)*x_ref[0]+(0.72974853865)*x_ref[1]**o+(-0.328183435687)*x_ref[1]
        ref[0,1,0]=(-0.749483736482)*x_ref[0]**o+(0.0872454020187)*x_ref[0]+(0.6916024322)*x_ref[1]**o+(0.82757019273)*x_ref[1]
        ref[0,1,1]=(-0.325601366346)*x_ref[0]**o+(0.918246327277)*x_ref[0]+(0.855963838252)*x_ref[1]**o+(0.448567464706)*x_ref[1]
        ref[1,0,0]=(-0.6642143628)*x_ref[0]**o+(-0.82649833447)*x_ref[0]+(0.822778465465)*x_ref[1]**o+(-0.926669226842)*x_ref[1]
        ref[1,0,1]=(-0.644490899373)*x_ref[0]**o+(0.0644054772987)*x_ref[0]+(-0.914922609343)*x_ref[1]**o+(0.898880463842)*x_ref[1]
        ref[1,1,0]=(0.40963096867)*x_ref[0]**o+(-0.347418076162)*x_ref[0]+(0.857966073118)*x_ref[1]**o+(-0.284521442085)*x_ref[1]
        ref[1,1,1]=(-0.00998062171478)*x_ref[0]**o+(-0.234122404174)*x_ref[0]+(-0.149500951208)*x_ref[1]**o+(0.680038328937)*x_ref[1]
        ref[2,0,0]=(-0.963777985833)*x_ref[0]**o+(0.710727608341)*x_ref[0]+(-0.646898391724)*x_ref[1]**o+(-0.464095155503)*x_ref[1]
        ref[2,0,1]=(-0.617670394851)*x_ref[0]**o+(0.592337321587)*x_ref[0]+(-0.0881426217074)*x_ref[1]**o+(-0.00431912882662)*x_ref[1]
        ref[2,1,0]=(0.871279828121)*x_ref[0]**o+(-0.858267190204)*x_ref[0]+(0.440863109038)*x_ref[1]**o+(-0.912521398099)*x_ref[1]
        ref[2,1,1]=(0.17802472174)*x_ref[0]**o+(0.948587960042)*x_ref[0]+(-0.837945850976)*x_ref[1]**o+(0.13308730872)*x_ref[1]
        ref[3,0,0]=(0.0482409149433)*x_ref[0]**o+(-0.198100088194)*x_ref[0]+(0.504789334224)*x_ref[1]**o+(0.296276808608)*x_ref[1]
        ref[3,0,1]=(-0.30515363985)*x_ref[0]**o+(-0.355293167256)*x_ref[0]+(-0.577592997685)*x_ref[1]**o+(-0.882408725943)*x_ref[1]
        ref[3,1,0]=(0.76102937294)*x_ref[0]**o+(-0.210169673144)*x_ref[0]+(-0.181755140891)*x_ref[1]**o+(0.865448484187)*x_ref[1]
        ref[3,1,1]=(-0.894294091863)*x_ref[0]**o+(-0.712245666103)*x_ref[0]+(0.151657455329)*x_ref[1]**o+(-0.54019436763)*x_ref[1]
        ref[4,0,0]=(0.860662861318)*x_ref[0]**o+(-0.36787473166)*x_ref[0]+(0.240406823985)*x_ref[1]**o+(-0.381822992473)*x_ref[1]
        ref[4,0,1]=(-0.109197125112)*x_ref[0]**o+(0.962595232385)*x_ref[0]+(-0.323876506847)*x_ref[1]**o+(0.178613363564)*x_ref[1]
        ref[4,1,0]=(0.486980313424)*x_ref[0]**o+(-0.611294566285)*x_ref[0]+(-0.774813150316)*x_ref[1]**o+(0.335411840405)*x_ref[1]
        ref[4,1,1]=(-0.622163740867)*x_ref[0]**o+(0.426305234663)*x_ref[0]+(0.243557551101)*x_ref[1]**o+(0.930601340004)*x_ref[1]
        ref[5,0,0]=(-0.232795962027)*x_ref[0]**o+(0.49010767268)*x_ref[0]+(-0.503693305594)*x_ref[1]**o+(-0.915931668186)*x_ref[1]
        ref[5,0,1]=(-0.304614759138)*x_ref[0]**o+(0.797445765321)*x_ref[0]+(-0.00820825778888)*x_ref[1]**o+(-0.196762463972)*x_ref[1]
        ref[5,1,0]=(-0.516700339456)*x_ref[0]**o+(-0.440953905414)*x_ref[0]+(0.875266018165)*x_ref[1]**o+(0.780895354543)*x_ref[1]
        ref[5,1,1]=(0.219827931399)*x_ref[0]**o+(-0.950514094131)*x_ref[0]+(-0.120763970425)*x_ref[1]**o+(0.0941571726596)*x_ref[1]
      else:
        arg[0,0,0]=(0.772105181243)*x[0]**o+(-0.548341656718)*x[0]+(0.358327458717)*x[1]**o+(0.07838537297)*x[1]+(-0.221432264133)*x[2]**o+(-0.0831433214109)*x[2]
        arg[0,0,1]=(-0.920280406855)*x[0]**o+(0.288008584982)*x[0]+(0.273175197348)*x[1]**o+(0.71005891586)*x[1]+(-0.372322860924)*x[2]**o+(0.801820758328)*x[2]
        arg[0,1,0]=(0.610498050037)*x[0]**o+(-0.539393441621)*x[0]+(0.587033652815)*x[1]**o+(0.237738229824)*x[1]+(-0.79411765083)*x[2]**o+(-0.519989054973)*x[2]
        arg[0,1,1]=(0.5412774887)*x[0]**o+(-0.851503637634)*x[0]+(-0.71956793056)*x[1]**o+(0.166039550456)*x[1]+(-0.274290267921)*x[2]**o+(-0.909146653718)*x[2]
        arg[1,0,0]=(-0.00458109759775)*x[0]**o+(0.487263914082)*x[0]+(0.0171216930245)*x[1]**o+(0.691063106679)*x[1]+(-0.96202326744)*x[2]**o+(0.361419732405)*x[2]
        arg[1,0,1]=(-0.784311451677)*x[0]**o+(-0.862560770078)*x[0]+(-0.0702620499924)*x[1]**o+(0.183400389223)*x[1]+(0.364905596839)*x[2]**o+(0.874230430841)*x[2]
        arg[1,1,0]=(0.315952660065)*x[0]**o+(-0.0168150653709)*x[0]+(0.372807793658)*x[1]**o+(0.350959428104)*x[1]+(0.292074907124)*x[2]**o+(0.38176271099)*x[2]
        arg[1,1,1]=(-0.941288490366)*x[0]**o+(0.159864446187)*x[0]+(0.616516779388)*x[1]**o+(0.0940731931915)*x[1]+(-0.966416702566)*x[2]**o+(-0.716669694143)*x[2]
        arg[2,0,0]=(0.308619329706)*x[0]**o+(0.68203622136)*x[0]+(-0.732163723834)*x[1]**o+(-0.222214324226)*x[1]+(0.89540478356)*x[2]**o+(-0.0648727721753)*x[2]
        arg[2,0,1]=(-0.829002921476)*x[0]**o+(0.274738860807)*x[0]+(0.944302912883)*x[1]**o+(0.67434413206)*x[1]+(-0.861085567199)*x[2]**o+(0.994494977369)*x[2]
        arg[2,1,0]=(-0.871499634896)*x[0]**o+(-0.716060125296)*x[0]+(-0.655428500508)*x[1]**o+(-0.478135345667)*x[1]+(-0.858927907345)*x[2]**o+(0.718884168928)*x[2]
        arg[2,1,1]=(0.843612813106)*x[0]**o+(-0.221966655825)*x[0]+(-0.0770757351)*x[1]**o+(-0.678171902341)*x[1]+(0.814209490483)*x[2]**o+(0.409183122676)*x[2]
        arg[3,0,0]=(-0.226810706884)*x[0]**o+(-0.167181963701)*x[0]+(0.810478050432)*x[1]**o+(-0.335099187817)*x[1]+(0.210621363393)*x[2]**o+(0.429297116788)*x[2]
        arg[3,0,1]=(0.597117204298)*x[0]**o+(-0.000147371125322)*x[0]+(-0.557842700291)*x[1]**o+(0.329529225331)*x[1]+(-0.208265244071)*x[2]**o+(-0.977698555357)*x[2]
        arg[3,1,0]=(-0.163068070947)*x[0]**o+(0.640075503516)*x[0]+(0.425653990651)*x[1]**o+(0.650655294789)*x[1]+(-0.342488862403)*x[2]**o+(0.227438788286)*x[2]
        arg[3,1,1]=(-0.511981920097)*x[0]**o+(-0.18099315756)*x[0]+(-0.969881140182)*x[1]**o+(-0.418742703662)*x[1]+(-0.271422367765)*x[2]**o+(0.0812058381877)*x[2]
        arg[4,0,0]=(0.569514749505)*x[0]**o+(-0.711561480184)*x[0]+(-0.587159226794)*x[1]**o+(-0.240459786489)*x[1]+(-0.380983629494)*x[2]**o+(-0.550385464258)*x[2]
        arg[4,0,1]=(0.257375332759)*x[0]**o+(-0.477083226492)*x[0]+(0.0180233249756)*x[1]**o+(0.612470082967)*x[1]+(0.99566078545)*x[2]**o+(-0.551544637583)*x[2]
        arg[4,1,0]=(0.702787194057)*x[0]**o+(0.922262573534)*x[0]+(-0.465910368685)*x[1]**o+(0.748743004032)*x[1]+(0.642852066952)*x[2]**o+(0.24283802317)*x[2]
        arg[4,1,1]=(0.745762578199)*x[0]**o+(0.240631495429)*x[0]+(0.413077167171)*x[1]**o+(0.692304442104)*x[1]+(0.150058687546)*x[2]**o+(-0.133413399526)*x[2]
        arg[5,0,0]=(0.0244935245274)*x[0]**o+(-0.866653903557)*x[0]+(-0.682252235104)*x[1]**o+(-0.525092961647)*x[1]+(-0.87552740837)*x[2]**o+(-0.780785295196)*x[2]
        arg[5,0,1]=(0.403317147145)*x[0]**o+(-0.846638525869)*x[0]+(-0.798622632135)*x[1]**o+(-0.199698251819)*x[1]+(-0.852399643056)*x[2]**o+(-0.0618243339514)*x[2]
        arg[5,1,0]=(-0.94578445284)*x[0]**o+(-0.715570943968)*x[0]+(-0.193533970382)*x[1]**o+(0.0325400011289)*x[1]+(0.137922635356)*x[2]**o+(0.0192967461384)*x[2]
        arg[5,1,1]=(-0.706542583366)*x[0]**o+(0.412536423598)*x[0]+(-0.27322227526)*x[1]**o+(0.339116123776)*x[1]+(-0.856871339939)*x[2]**o+(0.140603890427)*x[2]
        ref[0,0,0]=(0.772105181243)*x_ref[0]**o+(-0.548341656718)*x_ref[0]+(0.358327458717)*x_ref[1]**o+(0.07838537297)*x_ref[1]+(-0.221432264133)*x_ref[2]**o+(-0.0831433214109)*x_ref[2]
        ref[0,0,1]=(-0.920280406855)*x_ref[0]**o+(0.288008584982)*x_ref[0]+(0.273175197348)*x_ref[1]**o+(0.71005891586)*x_ref[1]+(-0.372322860924)*x_ref[2]**o+(0.801820758328)*x_ref[2]
        ref[0,1,0]=(0.610498050037)*x_ref[0]**o+(-0.539393441621)*x_ref[0]+(0.587033652815)*x_ref[1]**o+(0.237738229824)*x_ref[1]+(-0.79411765083)*x_ref[2]**o+(-0.519989054973)*x_ref[2]
        ref[0,1,1]=(0.5412774887)*x_ref[0]**o+(-0.851503637634)*x_ref[0]+(-0.71956793056)*x_ref[1]**o+(0.166039550456)*x_ref[1]+(-0.274290267921)*x_ref[2]**o+(-0.909146653718)*x_ref[2]
        ref[1,0,0]=(-0.00458109759775)*x_ref[0]**o+(0.487263914082)*x_ref[0]+(0.0171216930245)*x_ref[1]**o+(0.691063106679)*x_ref[1]+(-0.96202326744)*x_ref[2]**o+(0.361419732405)*x_ref[2]
        ref[1,0,1]=(-0.784311451677)*x_ref[0]**o+(-0.862560770078)*x_ref[0]+(-0.0702620499924)*x_ref[1]**o+(0.183400389223)*x_ref[1]+(0.364905596839)*x_ref[2]**o+(0.874230430841)*x_ref[2]
        ref[1,1,0]=(0.315952660065)*x_ref[0]**o+(-0.0168150653709)*x_ref[0]+(0.372807793658)*x_ref[1]**o+(0.350959428104)*x_ref[1]+(0.292074907124)*x_ref[2]**o+(0.38176271099)*x_ref[2]
        ref[1,1,1]=(-0.941288490366)*x_ref[0]**o+(0.159864446187)*x_ref[0]+(0.616516779388)*x_ref[1]**o+(0.0940731931915)*x_ref[1]+(-0.966416702566)*x_ref[2]**o+(-0.716669694143)*x_ref[2]
        ref[2,0,0]=(0.308619329706)*x_ref[0]**o+(0.68203622136)*x_ref[0]+(-0.732163723834)*x_ref[1]**o+(-0.222214324226)*x_ref[1]+(0.89540478356)*x_ref[2]**o+(-0.0648727721753)*x_ref[2]
        ref[2,0,1]=(-0.829002921476)*x_ref[0]**o+(0.274738860807)*x_ref[0]+(0.944302912883)*x_ref[1]**o+(0.67434413206)*x_ref[1]+(-0.861085567199)*x_ref[2]**o+(0.994494977369)*x_ref[2]
        ref[2,1,0]=(-0.871499634896)*x_ref[0]**o+(-0.716060125296)*x_ref[0]+(-0.655428500508)*x_ref[1]**o+(-0.478135345667)*x_ref[1]+(-0.858927907345)*x_ref[2]**o+(0.718884168928)*x_ref[2]
        ref[2,1,1]=(0.843612813106)*x_ref[0]**o+(-0.221966655825)*x_ref[0]+(-0.0770757351)*x_ref[1]**o+(-0.678171902341)*x_ref[1]+(0.814209490483)*x_ref[2]**o+(0.409183122676)*x_ref[2]
        ref[3,0,0]=(-0.226810706884)*x_ref[0]**o+(-0.167181963701)*x_ref[0]+(0.810478050432)*x_ref[1]**o+(-0.335099187817)*x_ref[1]+(0.210621363393)*x_ref[2]**o+(0.429297116788)*x_ref[2]
        ref[3,0,1]=(0.597117204298)*x_ref[0]**o+(-0.000147371125322)*x_ref[0]+(-0.557842700291)*x_ref[1]**o+(0.329529225331)*x_ref[1]+(-0.208265244071)*x_ref[2]**o+(-0.977698555357)*x_ref[2]
        ref[3,1,0]=(-0.163068070947)*x_ref[0]**o+(0.640075503516)*x_ref[0]+(0.425653990651)*x_ref[1]**o+(0.650655294789)*x_ref[1]+(-0.342488862403)*x_ref[2]**o+(0.227438788286)*x_ref[2]
        ref[3,1,1]=(-0.511981920097)*x_ref[0]**o+(-0.18099315756)*x_ref[0]+(-0.969881140182)*x_ref[1]**o+(-0.418742703662)*x_ref[1]+(-0.271422367765)*x_ref[2]**o+(0.0812058381877)*x_ref[2]
        ref[4,0,0]=(0.569514749505)*x_ref[0]**o+(-0.711561480184)*x_ref[0]+(-0.587159226794)*x_ref[1]**o+(-0.240459786489)*x_ref[1]+(-0.380983629494)*x_ref[2]**o+(-0.550385464258)*x_ref[2]
        ref[4,0,1]=(0.257375332759)*x_ref[0]**o+(-0.477083226492)*x_ref[0]+(0.0180233249756)*x_ref[1]**o+(0.612470082967)*x_ref[1]+(0.99566078545)*x_ref[2]**o+(-0.551544637583)*x_ref[2]
        ref[4,1,0]=(0.702787194057)*x_ref[0]**o+(0.922262573534)*x_ref[0]+(-0.465910368685)*x_ref[1]**o+(0.748743004032)*x_ref[1]+(0.642852066952)*x_ref[2]**o+(0.24283802317)*x_ref[2]
        ref[4,1,1]=(0.745762578199)*x_ref[0]**o+(0.240631495429)*x_ref[0]+(0.413077167171)*x_ref[1]**o+(0.692304442104)*x_ref[1]+(0.150058687546)*x_ref[2]**o+(-0.133413399526)*x_ref[2]
        ref[5,0,0]=(0.0244935245274)*x_ref[0]**o+(-0.866653903557)*x_ref[0]+(-0.682252235104)*x_ref[1]**o+(-0.525092961647)*x_ref[1]+(-0.87552740837)*x_ref[2]**o+(-0.780785295196)*x_ref[2]
        ref[5,0,1]=(0.403317147145)*x_ref[0]**o+(-0.846638525869)*x_ref[0]+(-0.798622632135)*x_ref[1]**o+(-0.199698251819)*x_ref[1]+(-0.852399643056)*x_ref[2]**o+(-0.0618243339514)*x_ref[2]
        ref[5,1,0]=(-0.94578445284)*x_ref[0]**o+(-0.715570943968)*x_ref[0]+(-0.193533970382)*x_ref[1]**o+(0.0325400011289)*x_ref[1]+(0.137922635356)*x_ref[2]**o+(0.0192967461384)*x_ref[2]
        ref[5,1,1]=(-0.706542583366)*x_ref[0]**o+(0.412536423598)*x_ref[0]+(-0.27322227526)*x_ref[1]**o+(0.339116123776)*x_ref[1]+(-0.856871339939)*x_ref[2]**o+(0.140603890427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0577426785287)*x[0]**o+(0.469845292021)*x[0]+(-0.641706722819)*x[1]**o+(0.129652001842)*x[1]
        arg[0,0,0,1]=(-0.87686951076)*x[0]**o+(-0.981093040999)*x[0]+(0.654637319552)*x[1]**o+(-0.953104888811)*x[1]
        arg[0,0,1,0]=(-0.486386436415)*x[0]**o+(0.91083470956)*x[0]+(-0.392432937401)*x[1]**o+(0.494919250884)*x[1]
        arg[0,0,1,1]=(0.299928502076)*x[0]**o+(-0.0787848191157)*x[0]+(-0.541863592475)*x[1]**o+(0.435934816687)*x[1]
        arg[0,0,2,0]=(-0.0500083438365)*x[0]**o+(-0.882581886623)*x[0]+(-0.965347357063)*x[1]**o+(-0.270254041774)*x[1]
        arg[0,0,2,1]=(-0.18974861517)*x[0]**o+(0.341406736671)*x[0]+(0.847020774229)*x[1]**o+(0.650201215849)*x[1]
        arg[0,1,0,0]=(0.0823017084719)*x[0]**o+(0.964585359235)*x[0]+(0.357658660436)*x[1]**o+(0.182585271812)*x[1]
        arg[0,1,0,1]=(0.863411821905)*x[0]**o+(0.313374566426)*x[0]+(-0.273547644827)*x[1]**o+(0.429051854479)*x[1]
        arg[0,1,1,0]=(0.167608645068)*x[0]**o+(0.926908094967)*x[0]+(-0.853415240465)*x[1]**o+(-0.376634525241)*x[1]
        arg[0,1,1,1]=(-0.555491961591)*x[0]**o+(-0.64621137308)*x[0]+(-0.315877124991)*x[1]**o+(-0.962839286133)*x[1]
        arg[0,1,2,0]=(0.909012065667)*x[0]**o+(-0.656739022976)*x[0]+(-0.458377773547)*x[1]**o+(0.394223574009)*x[1]
        arg[0,1,2,1]=(-0.550560496714)*x[0]**o+(0.216941875936)*x[0]+(-0.634745794924)*x[1]**o+(0.887912543217)*x[1]
        arg[0,2,0,0]=(-0.961535901493)*x[0]**o+(0.980336384835)*x[0]+(-0.151755000128)*x[1]**o+(0.0402841925797)*x[1]
        arg[0,2,0,1]=(0.571527088727)*x[0]**o+(-0.475013184811)*x[0]+(0.80051818192)*x[1]**o+(-0.589499747694)*x[1]
        arg[0,2,1,0]=(0.545840648612)*x[0]**o+(0.888313209612)*x[0]+(-0.568938733446)*x[1]**o+(-0.289030374342)*x[1]
        arg[0,2,1,1]=(-0.0631638607979)*x[0]**o+(-0.173754439841)*x[0]+(0.143503554071)*x[1]**o+(-0.0850087835084)*x[1]
        arg[0,2,2,0]=(0.316299661412)*x[0]**o+(0.260890197379)*x[0]+(-0.803519867776)*x[1]**o+(0.36434810185)*x[1]
        arg[0,2,2,1]=(-0.656013716479)*x[0]**o+(0.810499942514)*x[0]+(0.783455480815)*x[1]**o+(-0.943085466999)*x[1]
        arg[0,3,0,0]=(0.128153439722)*x[0]**o+(-0.404740042302)*x[0]+(0.785717360548)*x[1]**o+(-0.710818900841)*x[1]
        arg[0,3,0,1]=(-0.097059274787)*x[0]**o+(0.578757823228)*x[0]+(0.186146155178)*x[1]**o+(0.274602030707)*x[1]
        arg[0,3,1,0]=(0.00522392753954)*x[0]**o+(0.015773606982)*x[0]+(-0.327110065344)*x[1]**o+(0.15571120542)*x[1]
        arg[0,3,1,1]=(-0.14009116305)*x[0]**o+(-0.659975337811)*x[0]+(-0.297582833489)*x[1]**o+(-0.600935015071)*x[1]
        arg[0,3,2,0]=(0.749388604357)*x[0]**o+(0.762968650778)*x[0]+(-0.623477518339)*x[1]**o+(-0.931762943135)*x[1]
        arg[0,3,2,1]=(0.748396100632)*x[0]**o+(-0.841314360401)*x[0]+(0.432649446216)*x[1]**o+(-0.971300843628)*x[1]
        arg[0,4,0,0]=(-0.217249090357)*x[0]**o+(-0.833633675601)*x[0]+(0.169984305627)*x[1]**o+(-0.831738272666)*x[1]
        arg[0,4,0,1]=(0.501571101669)*x[0]**o+(-0.223897334788)*x[0]+(0.24178181976)*x[1]**o+(0.307921789017)*x[1]
        arg[0,4,1,0]=(-0.162064610564)*x[0]**o+(0.217669674603)*x[0]+(0.995517866562)*x[1]**o+(0.903061734589)*x[1]
        arg[0,4,1,1]=(-0.837150436928)*x[0]**o+(0.465269468525)*x[0]+(0.522034182082)*x[1]**o+(-0.725469921625)*x[1]
        arg[0,4,2,0]=(-0.295739906169)*x[0]**o+(0.961959220618)*x[0]+(0.577799918942)*x[1]**o+(-0.313255108148)*x[1]
        arg[0,4,2,1]=(-0.455635927062)*x[0]**o+(0.667038701641)*x[0]+(-0.151183772961)*x[1]**o+(-0.550197828316)*x[1]
        arg[1,0,0,0]=(0.585510684472)*x[0]**o+(-0.64754219493)*x[0]+(0.781495111863)*x[1]**o+(-0.792246214695)*x[1]
        arg[1,0,0,1]=(0.586346471397)*x[0]**o+(0.291794158138)*x[0]+(-0.614723564286)*x[1]**o+(-0.151451723805)*x[1]
        arg[1,0,1,0]=(-0.472877220222)*x[0]**o+(-0.0118252391089)*x[0]+(-0.596943135354)*x[1]**o+(-0.413992986066)*x[1]
        arg[1,0,1,1]=(-0.715007718152)*x[0]**o+(0.032964784403)*x[0]+(-0.350680745946)*x[1]**o+(0.678841285054)*x[1]
        arg[1,0,2,0]=(0.673910318085)*x[0]**o+(-0.15303471314)*x[0]+(-0.0889279944344)*x[1]**o+(-0.0356474498002)*x[1]
        arg[1,0,2,1]=(-0.875691348779)*x[0]**o+(0.547300853671)*x[0]+(-0.362646169132)*x[1]**o+(-0.588521964671)*x[1]
        arg[1,1,0,0]=(-0.0988912551145)*x[0]**o+(-0.710944394432)*x[0]+(0.153875249337)*x[1]**o+(0.274427434926)*x[1]
        arg[1,1,0,1]=(-0.0854272185591)*x[0]**o+(-0.21840195791)*x[0]+(0.906982245924)*x[1]**o+(0.570990116211)*x[1]
        arg[1,1,1,0]=(-0.851259116331)*x[0]**o+(0.309829955487)*x[0]+(0.115725082696)*x[1]**o+(-0.91590292189)*x[1]
        arg[1,1,1,1]=(-0.637464462703)*x[0]**o+(-0.280971672162)*x[0]+(0.270985403603)*x[1]**o+(-0.534115311397)*x[1]
        arg[1,1,2,0]=(0.670179973397)*x[0]**o+(0.847483793904)*x[0]+(0.0681088634526)*x[1]**o+(0.412176359729)*x[1]
        arg[1,1,2,1]=(0.0378205172485)*x[0]**o+(0.974150495292)*x[0]+(-0.720821275299)*x[1]**o+(-0.552588296494)*x[1]
        arg[1,2,0,0]=(0.151367609975)*x[0]**o+(0.832981586349)*x[0]+(0.419488381843)*x[1]**o+(0.561052438283)*x[1]
        arg[1,2,0,1]=(0.267839010425)*x[0]**o+(0.47455974064)*x[0]+(0.498826652567)*x[1]**o+(-0.51906391205)*x[1]
        arg[1,2,1,0]=(-0.350105107527)*x[0]**o+(-0.489326689724)*x[0]+(-0.156970199109)*x[1]**o+(-0.175565230472)*x[1]
        arg[1,2,1,1]=(0.754379973459)*x[0]**o+(-0.70691373234)*x[0]+(-0.964497356137)*x[1]**o+(-0.694989442192)*x[1]
        arg[1,2,2,0]=(-0.491090361655)*x[0]**o+(0.154379441501)*x[0]+(0.403809695039)*x[1]**o+(-0.084183660554)*x[1]
        arg[1,2,2,1]=(-0.164538235223)*x[0]**o+(0.628058472821)*x[0]+(0.331956969519)*x[1]**o+(0.828755643202)*x[1]
        arg[1,3,0,0]=(0.103479416566)*x[0]**o+(0.380691196177)*x[0]+(-0.350375360146)*x[1]**o+(0.18528492785)*x[1]
        arg[1,3,0,1]=(-0.0744423760559)*x[0]**o+(0.912275380092)*x[0]+(0.867254399179)*x[1]**o+(0.410307184344)*x[1]
        arg[1,3,1,0]=(-0.480324052626)*x[0]**o+(-0.438544962062)*x[0]+(0.749575808465)*x[1]**o+(-0.0689669475685)*x[1]
        arg[1,3,1,1]=(0.723325499611)*x[0]**o+(-0.368668363075)*x[0]+(0.404302895836)*x[1]**o+(0.66913878408)*x[1]
        arg[1,3,2,0]=(-0.122189168877)*x[0]**o+(0.603362404944)*x[0]+(0.184163482697)*x[1]**o+(0.728136755361)*x[1]
        arg[1,3,2,1]=(-0.149430048386)*x[0]**o+(-0.478108072433)*x[0]+(0.342479834611)*x[1]**o+(0.596757508109)*x[1]
        arg[1,4,0,0]=(0.334694167334)*x[0]**o+(-0.815793032292)*x[0]+(0.394250112116)*x[1]**o+(0.269197166808)*x[1]
        arg[1,4,0,1]=(-0.19210954081)*x[0]**o+(0.416736181257)*x[0]+(0.146217029416)*x[1]**o+(0.906316717458)*x[1]
        arg[1,4,1,0]=(-0.708061018673)*x[0]**o+(0.299200707691)*x[0]+(0.665111947965)*x[1]**o+(-0.0340170473314)*x[1]
        arg[1,4,1,1]=(0.0353929744202)*x[0]**o+(0.0493194235067)*x[0]+(-0.12215424813)*x[1]**o+(-0.758677440837)*x[1]
        arg[1,4,2,0]=(0.769702957701)*x[0]**o+(0.901805654682)*x[0]+(0.260222241019)*x[1]**o+(-0.471997705392)*x[1]
        arg[1,4,2,1]=(0.204258068316)*x[0]**o+(-0.173085114573)*x[0]+(-0.0253164178046)*x[1]**o+(0.84254484259)*x[1]
        arg[2,0,0,0]=(-0.688170206644)*x[0]**o+(-0.718673702314)*x[0]+(0.834265379201)*x[1]**o+(-0.879925159193)*x[1]
        arg[2,0,0,1]=(-0.515401334666)*x[0]**o+(-0.810523180967)*x[0]+(0.725182741017)*x[1]**o+(-0.74319471007)*x[1]
        arg[2,0,1,0]=(-0.799698108602)*x[0]**o+(0.289222327549)*x[0]+(0.0589402409999)*x[1]**o+(-0.437669765839)*x[1]
        arg[2,0,1,1]=(-0.734933454418)*x[0]**o+(0.371391146956)*x[0]+(-0.17797293763)*x[1]**o+(-0.499412028554)*x[1]
        arg[2,0,2,0]=(0.217283159505)*x[0]**o+(-0.184871188378)*x[0]+(-0.764095670777)*x[1]**o+(0.103705507755)*x[1]
        arg[2,0,2,1]=(0.578071527416)*x[0]**o+(0.845365737459)*x[0]+(-0.173187973137)*x[1]**o+(-0.827819925403)*x[1]
        arg[2,1,0,0]=(-0.258505119864)*x[0]**o+(0.640873475074)*x[0]+(-0.993841847675)*x[1]**o+(0.347490565123)*x[1]
        arg[2,1,0,1]=(0.224893445547)*x[0]**o+(0.827208591218)*x[0]+(0.0391563892525)*x[1]**o+(-0.320939200596)*x[1]
        arg[2,1,1,0]=(-0.993044884903)*x[0]**o+(-0.395235414086)*x[0]+(-0.447460653655)*x[1]**o+(-0.114585622808)*x[1]
        arg[2,1,1,1]=(0.908711735866)*x[0]**o+(-0.911405008413)*x[0]+(-0.798601192961)*x[1]**o+(-0.432937813587)*x[1]
        arg[2,1,2,0]=(-0.270424920067)*x[0]**o+(-0.533648944143)*x[0]+(0.184307763968)*x[1]**o+(-0.802221601505)*x[1]
        arg[2,1,2,1]=(0.154432841861)*x[0]**o+(0.306458633794)*x[0]+(0.0343421658454)*x[1]**o+(-0.768559457414)*x[1]
        arg[2,2,0,0]=(-0.497828132445)*x[0]**o+(0.484920248315)*x[0]+(-0.188239341875)*x[1]**o+(0.100360622012)*x[1]
        arg[2,2,0,1]=(-0.0433518780446)*x[0]**o+(0.119470819101)*x[0]+(0.344025017428)*x[1]**o+(0.667394126898)*x[1]
        arg[2,2,1,0]=(0.512271795846)*x[0]**o+(0.567024910623)*x[0]+(-0.316191920076)*x[1]**o+(-0.202135136851)*x[1]
        arg[2,2,1,1]=(0.781678855797)*x[0]**o+(0.159208533718)*x[0]+(-0.930150837465)*x[1]**o+(-0.666903895504)*x[1]
        arg[2,2,2,0]=(0.292270965732)*x[0]**o+(-0.112877441048)*x[0]+(-0.643355582049)*x[1]**o+(-0.0482560996171)*x[1]
        arg[2,2,2,1]=(0.638688065872)*x[0]**o+(0.787719224138)*x[0]+(0.815312681592)*x[1]**o+(0.334735776221)*x[1]
        arg[2,3,0,0]=(0.409920519672)*x[0]**o+(-0.191886243537)*x[0]+(-0.348205383522)*x[1]**o+(-0.651065510647)*x[1]
        arg[2,3,0,1]=(0.824428500178)*x[0]**o+(-0.909533490456)*x[0]+(-0.455079541262)*x[1]**o+(-0.613994310127)*x[1]
        arg[2,3,1,0]=(0.225908995914)*x[0]**o+(0.258098741048)*x[0]+(-0.858905205327)*x[1]**o+(0.535845140874)*x[1]
        arg[2,3,1,1]=(-0.183600641837)*x[0]**o+(-0.0436432564818)*x[0]+(0.306524301109)*x[1]**o+(0.662513699164)*x[1]
        arg[2,3,2,0]=(-0.678973103555)*x[0]**o+(0.674110571991)*x[0]+(-0.129743880618)*x[1]**o+(-0.210759231178)*x[1]
        arg[2,3,2,1]=(-0.594989617102)*x[0]**o+(0.0557145839856)*x[0]+(-0.379516802094)*x[1]**o+(-0.496127170036)*x[1]
        arg[2,4,0,0]=(0.401088886175)*x[0]**o+(-0.208580896076)*x[0]+(-0.655582275521)*x[1]**o+(-0.522892532154)*x[1]
        arg[2,4,0,1]=(0.609111916382)*x[0]**o+(-0.864603847696)*x[0]+(-0.373482060442)*x[1]**o+(0.048777005392)*x[1]
        arg[2,4,1,0]=(-0.503935244799)*x[0]**o+(0.320517360825)*x[0]+(-0.658994963813)*x[1]**o+(0.00465823412972)*x[1]
        arg[2,4,1,1]=(-0.202767966863)*x[0]**o+(0.737239608437)*x[0]+(-0.633206522406)*x[1]**o+(-0.408531988781)*x[1]
        arg[2,4,2,0]=(0.660048575091)*x[0]**o+(-0.399137043056)*x[0]+(-0.0977181117895)*x[1]**o+(0.148903352776)*x[1]
        arg[2,4,2,1]=(-0.396768795712)*x[0]**o+(-0.550388681028)*x[0]+(0.0946256561554)*x[1]**o+(-0.0566941754267)*x[1]
        arg[3,0,0,0]=(0.0647180650687)*x[0]**o+(0.920788893299)*x[0]+(0.131306350033)*x[1]**o+(-0.840527983774)*x[1]
        arg[3,0,0,1]=(0.451700262969)*x[0]**o+(-0.16104676139)*x[0]+(0.114145999713)*x[1]**o+(-0.949265482931)*x[1]
        arg[3,0,1,0]=(-0.219444617083)*x[0]**o+(0.846470673388)*x[0]+(-0.341392225388)*x[1]**o+(-0.721912717627)*x[1]
        arg[3,0,1,1]=(-0.199502120929)*x[0]**o+(-0.00662410675554)*x[0]+(0.0529013037694)*x[1]**o+(-0.022278580228)*x[1]
        arg[3,0,2,0]=(0.803500732535)*x[0]**o+(0.754138333448)*x[0]+(0.822404687478)*x[1]**o+(0.900099584934)*x[1]
        arg[3,0,2,1]=(0.84674757393)*x[0]**o+(0.326823539079)*x[0]+(-0.824446392056)*x[1]**o+(-0.506583766552)*x[1]
        arg[3,1,0,0]=(0.816775849327)*x[0]**o+(0.957523430402)*x[0]+(-0.822535686534)*x[1]**o+(-0.583389123083)*x[1]
        arg[3,1,0,1]=(0.518120693337)*x[0]**o+(0.0946135115511)*x[0]+(0.577364260402)*x[1]**o+(-0.677435109716)*x[1]
        arg[3,1,1,0]=(-0.559041961891)*x[0]**o+(0.436791134026)*x[0]+(0.287609046653)*x[1]**o+(-0.484573207793)*x[1]
        arg[3,1,1,1]=(-0.0786041693053)*x[0]**o+(0.793526931222)*x[0]+(0.055780633982)*x[1]**o+(0.0447074319839)*x[1]
        arg[3,1,2,0]=(-0.190231018951)*x[0]**o+(-0.61318830169)*x[0]+(-0.157202158534)*x[1]**o+(-0.823043159126)*x[1]
        arg[3,1,2,1]=(-0.502637901786)*x[0]**o+(0.781083145697)*x[0]+(-0.663013942651)*x[1]**o+(-0.305726874525)*x[1]
        arg[3,2,0,0]=(0.566955929517)*x[0]**o+(-0.506981068575)*x[0]+(-0.831439289654)*x[1]**o+(-0.337491442765)*x[1]
        arg[3,2,0,1]=(-0.713648844422)*x[0]**o+(-0.00815764297609)*x[0]+(0.829706206618)*x[1]**o+(0.96471376965)*x[1]
        arg[3,2,1,0]=(-0.944705561715)*x[0]**o+(-0.87962603172)*x[0]+(-0.0525190846342)*x[1]**o+(-0.0457463400515)*x[1]
        arg[3,2,1,1]=(0.54922446571)*x[0]**o+(0.651923573438)*x[0]+(-0.00229398468376)*x[1]**o+(-0.00170863787749)*x[1]
        arg[3,2,2,0]=(0.622372079685)*x[0]**o+(-0.553578221845)*x[0]+(-0.737877450857)*x[1]**o+(0.459729854002)*x[1]
        arg[3,2,2,1]=(0.826155732508)*x[0]**o+(-0.000468898258991)*x[0]+(-0.255179342686)*x[1]**o+(-0.80957312045)*x[1]
        arg[3,3,0,0]=(-0.195212469839)*x[0]**o+(-0.0235216449547)*x[0]+(-0.175192639485)*x[1]**o+(-0.58937661213)*x[1]
        arg[3,3,0,1]=(-0.931538910404)*x[0]**o+(0.649282880007)*x[0]+(-0.237841482637)*x[1]**o+(-0.0220516220802)*x[1]
        arg[3,3,1,0]=(0.420911057166)*x[0]**o+(0.200126728184)*x[0]+(-0.829130861656)*x[1]**o+(-0.563462916661)*x[1]
        arg[3,3,1,1]=(-0.148199974575)*x[0]**o+(-0.248825389533)*x[0]+(-0.346424894584)*x[1]**o+(0.339758232978)*x[1]
        arg[3,3,2,0]=(0.249372496215)*x[0]**o+(-0.343626915746)*x[0]+(0.0961351845087)*x[1]**o+(0.612679249068)*x[1]
        arg[3,3,2,1]=(0.985380582259)*x[0]**o+(-0.603514318626)*x[0]+(-0.131105546364)*x[1]**o+(-0.0798158697238)*x[1]
        arg[3,4,0,0]=(0.478635365792)*x[0]**o+(-0.964424450881)*x[0]+(-0.469995535544)*x[1]**o+(-0.72871149515)*x[1]
        arg[3,4,0,1]=(-0.264802297415)*x[0]**o+(-0.926518282507)*x[0]+(0.803390182149)*x[1]**o+(0.387910762706)*x[1]
        arg[3,4,1,0]=(0.0452448579909)*x[0]**o+(-0.0482221515826)*x[0]+(-0.728020494982)*x[1]**o+(-0.288823633249)*x[1]
        arg[3,4,1,1]=(-0.640459595265)*x[0]**o+(0.759525453728)*x[0]+(-0.0280300707884)*x[1]**o+(0.079438187123)*x[1]
        arg[3,4,2,0]=(-0.14185171171)*x[0]**o+(-0.810989314052)*x[0]+(0.921902325773)*x[1]**o+(0.885920658784)*x[1]
        arg[3,4,2,1]=(-0.0101259901393)*x[0]**o+(-0.696541046542)*x[0]+(0.127932822496)*x[1]**o+(-0.822265796323)*x[1]
        ref[0,0,0,0]=(0.0577426785287)*x_ref[0]**o+(0.469845292021)*x_ref[0]+(-0.641706722819)*x_ref[1]**o+(0.129652001842)*x_ref[1]
        ref[0,0,0,1]=(-0.87686951076)*x_ref[0]**o+(-0.981093040999)*x_ref[0]+(0.654637319552)*x_ref[1]**o+(-0.953104888811)*x_ref[1]
        ref[0,0,1,0]=(-0.486386436415)*x_ref[0]**o+(0.91083470956)*x_ref[0]+(-0.392432937401)*x_ref[1]**o+(0.494919250884)*x_ref[1]
        ref[0,0,1,1]=(0.299928502076)*x_ref[0]**o+(-0.0787848191157)*x_ref[0]+(-0.541863592475)*x_ref[1]**o+(0.435934816687)*x_ref[1]
        ref[0,0,2,0]=(-0.0500083438365)*x_ref[0]**o+(-0.882581886623)*x_ref[0]+(-0.965347357063)*x_ref[1]**o+(-0.270254041774)*x_ref[1]
        ref[0,0,2,1]=(-0.18974861517)*x_ref[0]**o+(0.341406736671)*x_ref[0]+(0.847020774229)*x_ref[1]**o+(0.650201215849)*x_ref[1]
        ref[0,1,0,0]=(0.0823017084719)*x_ref[0]**o+(0.964585359235)*x_ref[0]+(0.357658660436)*x_ref[1]**o+(0.182585271812)*x_ref[1]
        ref[0,1,0,1]=(0.863411821905)*x_ref[0]**o+(0.313374566426)*x_ref[0]+(-0.273547644827)*x_ref[1]**o+(0.429051854479)*x_ref[1]
        ref[0,1,1,0]=(0.167608645068)*x_ref[0]**o+(0.926908094967)*x_ref[0]+(-0.853415240465)*x_ref[1]**o+(-0.376634525241)*x_ref[1]
        ref[0,1,1,1]=(-0.555491961591)*x_ref[0]**o+(-0.64621137308)*x_ref[0]+(-0.315877124991)*x_ref[1]**o+(-0.962839286133)*x_ref[1]
        ref[0,1,2,0]=(0.909012065667)*x_ref[0]**o+(-0.656739022976)*x_ref[0]+(-0.458377773547)*x_ref[1]**o+(0.394223574009)*x_ref[1]
        ref[0,1,2,1]=(-0.550560496714)*x_ref[0]**o+(0.216941875936)*x_ref[0]+(-0.634745794924)*x_ref[1]**o+(0.887912543217)*x_ref[1]
        ref[0,2,0,0]=(-0.961535901493)*x_ref[0]**o+(0.980336384835)*x_ref[0]+(-0.151755000128)*x_ref[1]**o+(0.0402841925797)*x_ref[1]
        ref[0,2,0,1]=(0.571527088727)*x_ref[0]**o+(-0.475013184811)*x_ref[0]+(0.80051818192)*x_ref[1]**o+(-0.589499747694)*x_ref[1]
        ref[0,2,1,0]=(0.545840648612)*x_ref[0]**o+(0.888313209612)*x_ref[0]+(-0.568938733446)*x_ref[1]**o+(-0.289030374342)*x_ref[1]
        ref[0,2,1,1]=(-0.0631638607979)*x_ref[0]**o+(-0.173754439841)*x_ref[0]+(0.143503554071)*x_ref[1]**o+(-0.0850087835084)*x_ref[1]
        ref[0,2,2,0]=(0.316299661412)*x_ref[0]**o+(0.260890197379)*x_ref[0]+(-0.803519867776)*x_ref[1]**o+(0.36434810185)*x_ref[1]
        ref[0,2,2,1]=(-0.656013716479)*x_ref[0]**o+(0.810499942514)*x_ref[0]+(0.783455480815)*x_ref[1]**o+(-0.943085466999)*x_ref[1]
        ref[0,3,0,0]=(0.128153439722)*x_ref[0]**o+(-0.404740042302)*x_ref[0]+(0.785717360548)*x_ref[1]**o+(-0.710818900841)*x_ref[1]
        ref[0,3,0,1]=(-0.097059274787)*x_ref[0]**o+(0.578757823228)*x_ref[0]+(0.186146155178)*x_ref[1]**o+(0.274602030707)*x_ref[1]
        ref[0,3,1,0]=(0.00522392753954)*x_ref[0]**o+(0.015773606982)*x_ref[0]+(-0.327110065344)*x_ref[1]**o+(0.15571120542)*x_ref[1]
        ref[0,3,1,1]=(-0.14009116305)*x_ref[0]**o+(-0.659975337811)*x_ref[0]+(-0.297582833489)*x_ref[1]**o+(-0.600935015071)*x_ref[1]
        ref[0,3,2,0]=(0.749388604357)*x_ref[0]**o+(0.762968650778)*x_ref[0]+(-0.623477518339)*x_ref[1]**o+(-0.931762943135)*x_ref[1]
        ref[0,3,2,1]=(0.748396100632)*x_ref[0]**o+(-0.841314360401)*x_ref[0]+(0.432649446216)*x_ref[1]**o+(-0.971300843628)*x_ref[1]
        ref[0,4,0,0]=(-0.217249090357)*x_ref[0]**o+(-0.833633675601)*x_ref[0]+(0.169984305627)*x_ref[1]**o+(-0.831738272666)*x_ref[1]
        ref[0,4,0,1]=(0.501571101669)*x_ref[0]**o+(-0.223897334788)*x_ref[0]+(0.24178181976)*x_ref[1]**o+(0.307921789017)*x_ref[1]
        ref[0,4,1,0]=(-0.162064610564)*x_ref[0]**o+(0.217669674603)*x_ref[0]+(0.995517866562)*x_ref[1]**o+(0.903061734589)*x_ref[1]
        ref[0,4,1,1]=(-0.837150436928)*x_ref[0]**o+(0.465269468525)*x_ref[0]+(0.522034182082)*x_ref[1]**o+(-0.725469921625)*x_ref[1]
        ref[0,4,2,0]=(-0.295739906169)*x_ref[0]**o+(0.961959220618)*x_ref[0]+(0.577799918942)*x_ref[1]**o+(-0.313255108148)*x_ref[1]
        ref[0,4,2,1]=(-0.455635927062)*x_ref[0]**o+(0.667038701641)*x_ref[0]+(-0.151183772961)*x_ref[1]**o+(-0.550197828316)*x_ref[1]
        ref[1,0,0,0]=(0.585510684472)*x_ref[0]**o+(-0.64754219493)*x_ref[0]+(0.781495111863)*x_ref[1]**o+(-0.792246214695)*x_ref[1]
        ref[1,0,0,1]=(0.586346471397)*x_ref[0]**o+(0.291794158138)*x_ref[0]+(-0.614723564286)*x_ref[1]**o+(-0.151451723805)*x_ref[1]
        ref[1,0,1,0]=(-0.472877220222)*x_ref[0]**o+(-0.0118252391089)*x_ref[0]+(-0.596943135354)*x_ref[1]**o+(-0.413992986066)*x_ref[1]
        ref[1,0,1,1]=(-0.715007718152)*x_ref[0]**o+(0.032964784403)*x_ref[0]+(-0.350680745946)*x_ref[1]**o+(0.678841285054)*x_ref[1]
        ref[1,0,2,0]=(0.673910318085)*x_ref[0]**o+(-0.15303471314)*x_ref[0]+(-0.0889279944344)*x_ref[1]**o+(-0.0356474498002)*x_ref[1]
        ref[1,0,2,1]=(-0.875691348779)*x_ref[0]**o+(0.547300853671)*x_ref[0]+(-0.362646169132)*x_ref[1]**o+(-0.588521964671)*x_ref[1]
        ref[1,1,0,0]=(-0.0988912551145)*x_ref[0]**o+(-0.710944394432)*x_ref[0]+(0.153875249337)*x_ref[1]**o+(0.274427434926)*x_ref[1]
        ref[1,1,0,1]=(-0.0854272185591)*x_ref[0]**o+(-0.21840195791)*x_ref[0]+(0.906982245924)*x_ref[1]**o+(0.570990116211)*x_ref[1]
        ref[1,1,1,0]=(-0.851259116331)*x_ref[0]**o+(0.309829955487)*x_ref[0]+(0.115725082696)*x_ref[1]**o+(-0.91590292189)*x_ref[1]
        ref[1,1,1,1]=(-0.637464462703)*x_ref[0]**o+(-0.280971672162)*x_ref[0]+(0.270985403603)*x_ref[1]**o+(-0.534115311397)*x_ref[1]
        ref[1,1,2,0]=(0.670179973397)*x_ref[0]**o+(0.847483793904)*x_ref[0]+(0.0681088634526)*x_ref[1]**o+(0.412176359729)*x_ref[1]
        ref[1,1,2,1]=(0.0378205172485)*x_ref[0]**o+(0.974150495292)*x_ref[0]+(-0.720821275299)*x_ref[1]**o+(-0.552588296494)*x_ref[1]
        ref[1,2,0,0]=(0.151367609975)*x_ref[0]**o+(0.832981586349)*x_ref[0]+(0.419488381843)*x_ref[1]**o+(0.561052438283)*x_ref[1]
        ref[1,2,0,1]=(0.267839010425)*x_ref[0]**o+(0.47455974064)*x_ref[0]+(0.498826652567)*x_ref[1]**o+(-0.51906391205)*x_ref[1]
        ref[1,2,1,0]=(-0.350105107527)*x_ref[0]**o+(-0.489326689724)*x_ref[0]+(-0.156970199109)*x_ref[1]**o+(-0.175565230472)*x_ref[1]
        ref[1,2,1,1]=(0.754379973459)*x_ref[0]**o+(-0.70691373234)*x_ref[0]+(-0.964497356137)*x_ref[1]**o+(-0.694989442192)*x_ref[1]
        ref[1,2,2,0]=(-0.491090361655)*x_ref[0]**o+(0.154379441501)*x_ref[0]+(0.403809695039)*x_ref[1]**o+(-0.084183660554)*x_ref[1]
        ref[1,2,2,1]=(-0.164538235223)*x_ref[0]**o+(0.628058472821)*x_ref[0]+(0.331956969519)*x_ref[1]**o+(0.828755643202)*x_ref[1]
        ref[1,3,0,0]=(0.103479416566)*x_ref[0]**o+(0.380691196177)*x_ref[0]+(-0.350375360146)*x_ref[1]**o+(0.18528492785)*x_ref[1]
        ref[1,3,0,1]=(-0.0744423760559)*x_ref[0]**o+(0.912275380092)*x_ref[0]+(0.867254399179)*x_ref[1]**o+(0.410307184344)*x_ref[1]
        ref[1,3,1,0]=(-0.480324052626)*x_ref[0]**o+(-0.438544962062)*x_ref[0]+(0.749575808465)*x_ref[1]**o+(-0.0689669475685)*x_ref[1]
        ref[1,3,1,1]=(0.723325499611)*x_ref[0]**o+(-0.368668363075)*x_ref[0]+(0.404302895836)*x_ref[1]**o+(0.66913878408)*x_ref[1]
        ref[1,3,2,0]=(-0.122189168877)*x_ref[0]**o+(0.603362404944)*x_ref[0]+(0.184163482697)*x_ref[1]**o+(0.728136755361)*x_ref[1]
        ref[1,3,2,1]=(-0.149430048386)*x_ref[0]**o+(-0.478108072433)*x_ref[0]+(0.342479834611)*x_ref[1]**o+(0.596757508109)*x_ref[1]
        ref[1,4,0,0]=(0.334694167334)*x_ref[0]**o+(-0.815793032292)*x_ref[0]+(0.394250112116)*x_ref[1]**o+(0.269197166808)*x_ref[1]
        ref[1,4,0,1]=(-0.19210954081)*x_ref[0]**o+(0.416736181257)*x_ref[0]+(0.146217029416)*x_ref[1]**o+(0.906316717458)*x_ref[1]
        ref[1,4,1,0]=(-0.708061018673)*x_ref[0]**o+(0.299200707691)*x_ref[0]+(0.665111947965)*x_ref[1]**o+(-0.0340170473314)*x_ref[1]
        ref[1,4,1,1]=(0.0353929744202)*x_ref[0]**o+(0.0493194235067)*x_ref[0]+(-0.12215424813)*x_ref[1]**o+(-0.758677440837)*x_ref[1]
        ref[1,4,2,0]=(0.769702957701)*x_ref[0]**o+(0.901805654682)*x_ref[0]+(0.260222241019)*x_ref[1]**o+(-0.471997705392)*x_ref[1]
        ref[1,4,2,1]=(0.204258068316)*x_ref[0]**o+(-0.173085114573)*x_ref[0]+(-0.0253164178046)*x_ref[1]**o+(0.84254484259)*x_ref[1]
        ref[2,0,0,0]=(-0.688170206644)*x_ref[0]**o+(-0.718673702314)*x_ref[0]+(0.834265379201)*x_ref[1]**o+(-0.879925159193)*x_ref[1]
        ref[2,0,0,1]=(-0.515401334666)*x_ref[0]**o+(-0.810523180967)*x_ref[0]+(0.725182741017)*x_ref[1]**o+(-0.74319471007)*x_ref[1]
        ref[2,0,1,0]=(-0.799698108602)*x_ref[0]**o+(0.289222327549)*x_ref[0]+(0.0589402409999)*x_ref[1]**o+(-0.437669765839)*x_ref[1]
        ref[2,0,1,1]=(-0.734933454418)*x_ref[0]**o+(0.371391146956)*x_ref[0]+(-0.17797293763)*x_ref[1]**o+(-0.499412028554)*x_ref[1]
        ref[2,0,2,0]=(0.217283159505)*x_ref[0]**o+(-0.184871188378)*x_ref[0]+(-0.764095670777)*x_ref[1]**o+(0.103705507755)*x_ref[1]
        ref[2,0,2,1]=(0.578071527416)*x_ref[0]**o+(0.845365737459)*x_ref[0]+(-0.173187973137)*x_ref[1]**o+(-0.827819925403)*x_ref[1]
        ref[2,1,0,0]=(-0.258505119864)*x_ref[0]**o+(0.640873475074)*x_ref[0]+(-0.993841847675)*x_ref[1]**o+(0.347490565123)*x_ref[1]
        ref[2,1,0,1]=(0.224893445547)*x_ref[0]**o+(0.827208591218)*x_ref[0]+(0.0391563892525)*x_ref[1]**o+(-0.320939200596)*x_ref[1]
        ref[2,1,1,0]=(-0.993044884903)*x_ref[0]**o+(-0.395235414086)*x_ref[0]+(-0.447460653655)*x_ref[1]**o+(-0.114585622808)*x_ref[1]
        ref[2,1,1,1]=(0.908711735866)*x_ref[0]**o+(-0.911405008413)*x_ref[0]+(-0.798601192961)*x_ref[1]**o+(-0.432937813587)*x_ref[1]
        ref[2,1,2,0]=(-0.270424920067)*x_ref[0]**o+(-0.533648944143)*x_ref[0]+(0.184307763968)*x_ref[1]**o+(-0.802221601505)*x_ref[1]
        ref[2,1,2,1]=(0.154432841861)*x_ref[0]**o+(0.306458633794)*x_ref[0]+(0.0343421658454)*x_ref[1]**o+(-0.768559457414)*x_ref[1]
        ref[2,2,0,0]=(-0.497828132445)*x_ref[0]**o+(0.484920248315)*x_ref[0]+(-0.188239341875)*x_ref[1]**o+(0.100360622012)*x_ref[1]
        ref[2,2,0,1]=(-0.0433518780446)*x_ref[0]**o+(0.119470819101)*x_ref[0]+(0.344025017428)*x_ref[1]**o+(0.667394126898)*x_ref[1]
        ref[2,2,1,0]=(0.512271795846)*x_ref[0]**o+(0.567024910623)*x_ref[0]+(-0.316191920076)*x_ref[1]**o+(-0.202135136851)*x_ref[1]
        ref[2,2,1,1]=(0.781678855797)*x_ref[0]**o+(0.159208533718)*x_ref[0]+(-0.930150837465)*x_ref[1]**o+(-0.666903895504)*x_ref[1]
        ref[2,2,2,0]=(0.292270965732)*x_ref[0]**o+(-0.112877441048)*x_ref[0]+(-0.643355582049)*x_ref[1]**o+(-0.0482560996171)*x_ref[1]
        ref[2,2,2,1]=(0.638688065872)*x_ref[0]**o+(0.787719224138)*x_ref[0]+(0.815312681592)*x_ref[1]**o+(0.334735776221)*x_ref[1]
        ref[2,3,0,0]=(0.409920519672)*x_ref[0]**o+(-0.191886243537)*x_ref[0]+(-0.348205383522)*x_ref[1]**o+(-0.651065510647)*x_ref[1]
        ref[2,3,0,1]=(0.824428500178)*x_ref[0]**o+(-0.909533490456)*x_ref[0]+(-0.455079541262)*x_ref[1]**o+(-0.613994310127)*x_ref[1]
        ref[2,3,1,0]=(0.225908995914)*x_ref[0]**o+(0.258098741048)*x_ref[0]+(-0.858905205327)*x_ref[1]**o+(0.535845140874)*x_ref[1]
        ref[2,3,1,1]=(-0.183600641837)*x_ref[0]**o+(-0.0436432564818)*x_ref[0]+(0.306524301109)*x_ref[1]**o+(0.662513699164)*x_ref[1]
        ref[2,3,2,0]=(-0.678973103555)*x_ref[0]**o+(0.674110571991)*x_ref[0]+(-0.129743880618)*x_ref[1]**o+(-0.210759231178)*x_ref[1]
        ref[2,3,2,1]=(-0.594989617102)*x_ref[0]**o+(0.0557145839856)*x_ref[0]+(-0.379516802094)*x_ref[1]**o+(-0.496127170036)*x_ref[1]
        ref[2,4,0,0]=(0.401088886175)*x_ref[0]**o+(-0.208580896076)*x_ref[0]+(-0.655582275521)*x_ref[1]**o+(-0.522892532154)*x_ref[1]
        ref[2,4,0,1]=(0.609111916382)*x_ref[0]**o+(-0.864603847696)*x_ref[0]+(-0.373482060442)*x_ref[1]**o+(0.048777005392)*x_ref[1]
        ref[2,4,1,0]=(-0.503935244799)*x_ref[0]**o+(0.320517360825)*x_ref[0]+(-0.658994963813)*x_ref[1]**o+(0.00465823412972)*x_ref[1]
        ref[2,4,1,1]=(-0.202767966863)*x_ref[0]**o+(0.737239608437)*x_ref[0]+(-0.633206522406)*x_ref[1]**o+(-0.408531988781)*x_ref[1]
        ref[2,4,2,0]=(0.660048575091)*x_ref[0]**o+(-0.399137043056)*x_ref[0]+(-0.0977181117895)*x_ref[1]**o+(0.148903352776)*x_ref[1]
        ref[2,4,2,1]=(-0.396768795712)*x_ref[0]**o+(-0.550388681028)*x_ref[0]+(0.0946256561554)*x_ref[1]**o+(-0.0566941754267)*x_ref[1]
        ref[3,0,0,0]=(0.0647180650687)*x_ref[0]**o+(0.920788893299)*x_ref[0]+(0.131306350033)*x_ref[1]**o+(-0.840527983774)*x_ref[1]
        ref[3,0,0,1]=(0.451700262969)*x_ref[0]**o+(-0.16104676139)*x_ref[0]+(0.114145999713)*x_ref[1]**o+(-0.949265482931)*x_ref[1]
        ref[3,0,1,0]=(-0.219444617083)*x_ref[0]**o+(0.846470673388)*x_ref[0]+(-0.341392225388)*x_ref[1]**o+(-0.721912717627)*x_ref[1]
        ref[3,0,1,1]=(-0.199502120929)*x_ref[0]**o+(-0.00662410675554)*x_ref[0]+(0.0529013037694)*x_ref[1]**o+(-0.022278580228)*x_ref[1]
        ref[3,0,2,0]=(0.803500732535)*x_ref[0]**o+(0.754138333448)*x_ref[0]+(0.822404687478)*x_ref[1]**o+(0.900099584934)*x_ref[1]
        ref[3,0,2,1]=(0.84674757393)*x_ref[0]**o+(0.326823539079)*x_ref[0]+(-0.824446392056)*x_ref[1]**o+(-0.506583766552)*x_ref[1]
        ref[3,1,0,0]=(0.816775849327)*x_ref[0]**o+(0.957523430402)*x_ref[0]+(-0.822535686534)*x_ref[1]**o+(-0.583389123083)*x_ref[1]
        ref[3,1,0,1]=(0.518120693337)*x_ref[0]**o+(0.0946135115511)*x_ref[0]+(0.577364260402)*x_ref[1]**o+(-0.677435109716)*x_ref[1]
        ref[3,1,1,0]=(-0.559041961891)*x_ref[0]**o+(0.436791134026)*x_ref[0]+(0.287609046653)*x_ref[1]**o+(-0.484573207793)*x_ref[1]
        ref[3,1,1,1]=(-0.0786041693053)*x_ref[0]**o+(0.793526931222)*x_ref[0]+(0.055780633982)*x_ref[1]**o+(0.0447074319839)*x_ref[1]
        ref[3,1,2,0]=(-0.190231018951)*x_ref[0]**o+(-0.61318830169)*x_ref[0]+(-0.157202158534)*x_ref[1]**o+(-0.823043159126)*x_ref[1]
        ref[3,1,2,1]=(-0.502637901786)*x_ref[0]**o+(0.781083145697)*x_ref[0]+(-0.663013942651)*x_ref[1]**o+(-0.305726874525)*x_ref[1]
        ref[3,2,0,0]=(0.566955929517)*x_ref[0]**o+(-0.506981068575)*x_ref[0]+(-0.831439289654)*x_ref[1]**o+(-0.337491442765)*x_ref[1]
        ref[3,2,0,1]=(-0.713648844422)*x_ref[0]**o+(-0.00815764297609)*x_ref[0]+(0.829706206618)*x_ref[1]**o+(0.96471376965)*x_ref[1]
        ref[3,2,1,0]=(-0.944705561715)*x_ref[0]**o+(-0.87962603172)*x_ref[0]+(-0.0525190846342)*x_ref[1]**o+(-0.0457463400515)*x_ref[1]
        ref[3,2,1,1]=(0.54922446571)*x_ref[0]**o+(0.651923573438)*x_ref[0]+(-0.00229398468376)*x_ref[1]**o+(-0.00170863787749)*x_ref[1]
        ref[3,2,2,0]=(0.622372079685)*x_ref[0]**o+(-0.553578221845)*x_ref[0]+(-0.737877450857)*x_ref[1]**o+(0.459729854002)*x_ref[1]
        ref[3,2,2,1]=(0.826155732508)*x_ref[0]**o+(-0.000468898258991)*x_ref[0]+(-0.255179342686)*x_ref[1]**o+(-0.80957312045)*x_ref[1]
        ref[3,3,0,0]=(-0.195212469839)*x_ref[0]**o+(-0.0235216449547)*x_ref[0]+(-0.175192639485)*x_ref[1]**o+(-0.58937661213)*x_ref[1]
        ref[3,3,0,1]=(-0.931538910404)*x_ref[0]**o+(0.649282880007)*x_ref[0]+(-0.237841482637)*x_ref[1]**o+(-0.0220516220802)*x_ref[1]
        ref[3,3,1,0]=(0.420911057166)*x_ref[0]**o+(0.200126728184)*x_ref[0]+(-0.829130861656)*x_ref[1]**o+(-0.563462916661)*x_ref[1]
        ref[3,3,1,1]=(-0.148199974575)*x_ref[0]**o+(-0.248825389533)*x_ref[0]+(-0.346424894584)*x_ref[1]**o+(0.339758232978)*x_ref[1]
        ref[3,3,2,0]=(0.249372496215)*x_ref[0]**o+(-0.343626915746)*x_ref[0]+(0.0961351845087)*x_ref[1]**o+(0.612679249068)*x_ref[1]
        ref[3,3,2,1]=(0.985380582259)*x_ref[0]**o+(-0.603514318626)*x_ref[0]+(-0.131105546364)*x_ref[1]**o+(-0.0798158697238)*x_ref[1]
        ref[3,4,0,0]=(0.478635365792)*x_ref[0]**o+(-0.964424450881)*x_ref[0]+(-0.469995535544)*x_ref[1]**o+(-0.72871149515)*x_ref[1]
        ref[3,4,0,1]=(-0.264802297415)*x_ref[0]**o+(-0.926518282507)*x_ref[0]+(0.803390182149)*x_ref[1]**o+(0.387910762706)*x_ref[1]
        ref[3,4,1,0]=(0.0452448579909)*x_ref[0]**o+(-0.0482221515826)*x_ref[0]+(-0.728020494982)*x_ref[1]**o+(-0.288823633249)*x_ref[1]
        ref[3,4,1,1]=(-0.640459595265)*x_ref[0]**o+(0.759525453728)*x_ref[0]+(-0.0280300707884)*x_ref[1]**o+(0.079438187123)*x_ref[1]
        ref[3,4,2,0]=(-0.14185171171)*x_ref[0]**o+(-0.810989314052)*x_ref[0]+(0.921902325773)*x_ref[1]**o+(0.885920658784)*x_ref[1]
        ref[3,4,2,1]=(-0.0101259901393)*x_ref[0]**o+(-0.696541046542)*x_ref[0]+(0.127932822496)*x_ref[1]**o+(-0.822265796323)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.657703355307)*x[0]**o+(0.397813831843)*x[0]+(0.364460538717)*x[1]**o+(-0.845626669617)*x[1]+(-0.987144254126)*x[2]**o+(-0.732384980726)*x[2]
        arg[0,0,0,1]=(-0.13280304886)*x[0]**o+(-0.953796807758)*x[0]+(0.875723746654)*x[1]**o+(0.0764745117242)*x[1]+(-0.536085676221)*x[2]**o+(-0.520611959678)*x[2]
        arg[0,0,1,0]=(0.716429446578)*x[0]**o+(-0.729164693441)*x[0]+(-0.832505557428)*x[1]**o+(-0.329200605611)*x[1]+(0.0310162060964)*x[2]**o+(-0.869978941999)*x[2]
        arg[0,0,1,1]=(0.110253530584)*x[0]**o+(0.995528486822)*x[0]+(0.838577797395)*x[1]**o+(0.443987752225)*x[1]+(0.387337959057)*x[2]**o+(0.548448752612)*x[2]
        arg[0,0,2,0]=(-0.995737515431)*x[0]**o+(-0.899807241717)*x[0]+(-0.0803287764138)*x[1]**o+(0.242680051556)*x[1]+(0.58974836823)*x[2]**o+(-0.498450188944)*x[2]
        arg[0,0,2,1]=(0.341561541555)*x[0]**o+(-0.00541988184597)*x[0]+(-0.204832898701)*x[1]**o+(-0.957247935865)*x[1]+(-0.441002902064)*x[2]**o+(0.756843480304)*x[2]
        arg[0,1,0,0]=(0.136201526244)*x[0]**o+(0.517677304193)*x[0]+(0.523031610748)*x[1]**o+(-0.908998053383)*x[1]+(-0.629649678416)*x[2]**o+(0.762977917512)*x[2]
        arg[0,1,0,1]=(-0.103521080864)*x[0]**o+(-0.434410222595)*x[0]+(0.418239530241)*x[1]**o+(0.65638475119)*x[1]+(0.413036393129)*x[2]**o+(-0.220365518353)*x[2]
        arg[0,1,1,0]=(-0.415690201483)*x[0]**o+(-0.969323394689)*x[0]+(-0.295639420195)*x[1]**o+(-0.24915002749)*x[1]+(-0.00927627101467)*x[2]**o+(0.520335261634)*x[2]
        arg[0,1,1,1]=(-0.262653741797)*x[0]**o+(-0.463590591666)*x[0]+(0.568893042923)*x[1]**o+(-0.869851332898)*x[1]+(-0.463261399415)*x[2]**o+(0.0281915686202)*x[2]
        arg[0,1,2,0]=(0.377159503058)*x[0]**o+(0.484512261943)*x[0]+(0.0303594628925)*x[1]**o+(0.482051243004)*x[1]+(0.240643013949)*x[2]**o+(-0.943908617705)*x[2]
        arg[0,1,2,1]=(0.843854614097)*x[0]**o+(0.018041534553)*x[0]+(0.264198703361)*x[1]**o+(0.732889517628)*x[1]+(-0.154807384394)*x[2]**o+(0.60513622187)*x[2]
        arg[0,2,0,0]=(-0.234758118286)*x[0]**o+(-0.77807396119)*x[0]+(-0.97783091103)*x[1]**o+(-0.716185223828)*x[1]+(-0.573126474259)*x[2]**o+(0.403110701555)*x[2]
        arg[0,2,0,1]=(-0.853878344804)*x[0]**o+(-0.482644826561)*x[0]+(-0.826041149012)*x[1]**o+(0.596376234798)*x[1]+(-0.216400517274)*x[2]**o+(-0.323127194259)*x[2]
        arg[0,2,1,0]=(0.931551121492)*x[0]**o+(-0.527132174379)*x[0]+(0.373496704463)*x[1]**o+(-0.375215826463)*x[1]+(-0.639167297984)*x[2]**o+(0.143608733189)*x[2]
        arg[0,2,1,1]=(-0.982206000879)*x[0]**o+(0.836589596534)*x[0]+(0.606275000138)*x[1]**o+(-0.821469453892)*x[1]+(-0.633828380948)*x[2]**o+(-0.637688777216)*x[2]
        arg[0,2,2,0]=(0.441649557529)*x[0]**o+(-0.34978365051)*x[0]+(0.514175365405)*x[1]**o+(-0.903212271234)*x[1]+(0.70236887156)*x[2]**o+(0.72953045189)*x[2]
        arg[0,2,2,1]=(0.302787896978)*x[0]**o+(0.157598071021)*x[0]+(-0.357101947632)*x[1]**o+(0.381098383156)*x[1]+(-0.552546036689)*x[2]**o+(0.448181721984)*x[2]
        arg[0,3,0,0]=(0.662181077399)*x[0]**o+(-0.283755239518)*x[0]+(0.33694741278)*x[1]**o+(0.275801685868)*x[1]+(0.526155259101)*x[2]**o+(-0.378241855559)*x[2]
        arg[0,3,0,1]=(0.157822133581)*x[0]**o+(0.25402072747)*x[0]+(0.522842123099)*x[1]**o+(0.510595773584)*x[1]+(-0.829239126666)*x[2]**o+(0.990879342044)*x[2]
        arg[0,3,1,0]=(-0.775925280059)*x[0]**o+(-0.726971388023)*x[0]+(-0.630088730753)*x[1]**o+(-0.731024178237)*x[1]+(-0.288643541196)*x[2]**o+(-0.283708428793)*x[2]
        arg[0,3,1,1]=(-0.883176640445)*x[0]**o+(-0.0810211976762)*x[0]+(-0.102919217784)*x[1]**o+(0.271384866385)*x[1]+(0.980796900477)*x[2]**o+(0.439294038672)*x[2]
        arg[0,3,2,0]=(-0.826688221502)*x[0]**o+(-0.909388347347)*x[0]+(0.164406927722)*x[1]**o+(0.60157189364)*x[1]+(-0.885578176748)*x[2]**o+(0.610008674247)*x[2]
        arg[0,3,2,1]=(0.130154197193)*x[0]**o+(-0.754703602975)*x[0]+(0.00463710497017)*x[1]**o+(0.502162916123)*x[1]+(-0.749124548541)*x[2]**o+(0.946592962092)*x[2]
        arg[0,4,0,0]=(-0.0478347448906)*x[0]**o+(-0.812937871882)*x[0]+(-0.460982012551)*x[1]**o+(0.317485998991)*x[1]+(-0.593671683477)*x[2]**o+(0.551385718668)*x[2]
        arg[0,4,0,1]=(0.281998767502)*x[0]**o+(-0.0613830311997)*x[0]+(0.981429151019)*x[1]**o+(-0.580567714008)*x[1]+(0.0369483005331)*x[2]**o+(-0.74132358588)*x[2]
        arg[0,4,1,0]=(0.525845217483)*x[0]**o+(0.907803915216)*x[0]+(0.111429079263)*x[1]**o+(-0.653030401377)*x[1]+(0.0814793913696)*x[2]**o+(-0.419201413875)*x[2]
        arg[0,4,1,1]=(0.862892518463)*x[0]**o+(0.34033875398)*x[0]+(0.0371376953103)*x[1]**o+(0.568160310726)*x[1]+(-0.237231036717)*x[2]**o+(-0.130046899302)*x[2]
        arg[0,4,2,0]=(0.223651181698)*x[0]**o+(0.730991398203)*x[0]+(0.133551525027)*x[1]**o+(0.685783540126)*x[1]+(0.339586395814)*x[2]**o+(-0.524371194863)*x[2]
        arg[0,4,2,1]=(0.220110871383)*x[0]**o+(-0.37385773877)*x[0]+(0.773372000545)*x[1]**o+(-0.4261356565)*x[1]+(-0.619438438945)*x[2]**o+(0.126288949127)*x[2]
        arg[1,0,0,0]=(0.727680196631)*x[0]**o+(-0.876280912599)*x[0]+(0.576553782915)*x[1]**o+(-0.0497780322367)*x[1]+(-0.690257457282)*x[2]**o+(-0.11130953054)*x[2]
        arg[1,0,0,1]=(0.331863815938)*x[0]**o+(0.169379824389)*x[0]+(0.665507756012)*x[1]**o+(0.94933357273)*x[1]+(-0.610506070093)*x[2]**o+(0.752476708027)*x[2]
        arg[1,0,1,0]=(0.851320583415)*x[0]**o+(-0.697235996161)*x[0]+(0.235192165807)*x[1]**o+(0.325473517022)*x[1]+(-0.424683549673)*x[2]**o+(0.228953796486)*x[2]
        arg[1,0,1,1]=(0.804192098517)*x[0]**o+(-0.684337197059)*x[0]+(-0.349251615715)*x[1]**o+(0.596165705743)*x[1]+(-0.636033221182)*x[2]**o+(-0.304367076307)*x[2]
        arg[1,0,2,0]=(-0.237305477187)*x[0]**o+(-0.997679652719)*x[0]+(0.485994068389)*x[1]**o+(0.44508530534)*x[1]+(0.940139935628)*x[2]**o+(0.0789221540062)*x[2]
        arg[1,0,2,1]=(0.814928463364)*x[0]**o+(-0.257693939789)*x[0]+(0.179643785703)*x[1]**o+(-0.696436530567)*x[1]+(-0.592226945554)*x[2]**o+(-0.454620411499)*x[2]
        arg[1,1,0,0]=(-0.933925672439)*x[0]**o+(0.212298180817)*x[0]+(-0.447836316801)*x[1]**o+(0.180796376138)*x[1]+(-0.749910125701)*x[2]**o+(0.134540433851)*x[2]
        arg[1,1,0,1]=(-0.649009346978)*x[0]**o+(0.0316819485953)*x[0]+(-0.718001767835)*x[1]**o+(0.289608435124)*x[1]+(-0.906045493384)*x[2]**o+(0.851021484125)*x[2]
        arg[1,1,1,0]=(-0.216265404566)*x[0]**o+(0.657073282172)*x[0]+(-0.796155541438)*x[1]**o+(0.338980498824)*x[1]+(0.477444448193)*x[2]**o+(0.674278909522)*x[2]
        arg[1,1,1,1]=(-0.67605942954)*x[0]**o+(0.457997253216)*x[0]+(-0.973576381948)*x[1]**o+(0.430432753773)*x[1]+(0.0329169564451)*x[2]**o+(-0.422512702485)*x[2]
        arg[1,1,2,0]=(-0.32458256139)*x[0]**o+(0.289568625922)*x[0]+(-0.529564027476)*x[1]**o+(0.596449833118)*x[1]+(0.960461031989)*x[2]**o+(0.00801425321069)*x[2]
        arg[1,1,2,1]=(-0.0425883615923)*x[0]**o+(-0.313926807333)*x[0]+(0.766062773671)*x[1]**o+(-0.857784560309)*x[1]+(0.914867381682)*x[2]**o+(-0.312493334639)*x[2]
        arg[1,2,0,0]=(-0.637151767355)*x[0]**o+(-0.679947136952)*x[0]+(-0.0838349105494)*x[1]**o+(0.00800835879666)*x[1]+(0.385024194519)*x[2]**o+(-0.627828236571)*x[2]
        arg[1,2,0,1]=(-0.62580153007)*x[0]**o+(-0.183834415542)*x[0]+(0.322406823575)*x[1]**o+(0.302908755868)*x[1]+(0.727234001542)*x[2]**o+(-0.933944878425)*x[2]
        arg[1,2,1,0]=(-0.897198312615)*x[0]**o+(0.825086526945)*x[0]+(-0.031001758991)*x[1]**o+(0.987845551304)*x[1]+(-0.551212418179)*x[2]**o+(0.657813241764)*x[2]
        arg[1,2,1,1]=(-0.544595822278)*x[0]**o+(0.595858429591)*x[0]+(-0.557503238514)*x[1]**o+(0.507560179011)*x[1]+(-0.519438270771)*x[2]**o+(-0.471678264995)*x[2]
        arg[1,2,2,0]=(0.135616278099)*x[0]**o+(0.40676743342)*x[0]+(0.458810719213)*x[1]**o+(-0.881620410228)*x[1]+(-0.897267732846)*x[2]**o+(0.665458992487)*x[2]
        arg[1,2,2,1]=(0.83854610457)*x[0]**o+(0.670263196782)*x[0]+(0.551767734306)*x[1]**o+(0.883934821215)*x[1]+(0.455375261768)*x[2]**o+(0.95604480045)*x[2]
        arg[1,3,0,0]=(0.705853698882)*x[0]**o+(0.84681314245)*x[0]+(-0.0984020252408)*x[1]**o+(0.943165990555)*x[1]+(-0.428955841327)*x[2]**o+(0.0961618993003)*x[2]
        arg[1,3,0,1]=(0.583731800337)*x[0]**o+(-0.738079255003)*x[0]+(0.315556768944)*x[1]**o+(-0.489631118687)*x[1]+(0.371775779104)*x[2]**o+(0.82029490872)*x[2]
        arg[1,3,1,0]=(-0.885657545142)*x[0]**o+(0.542706113286)*x[0]+(0.800319452136)*x[1]**o+(0.314194184897)*x[1]+(-0.320905532427)*x[2]**o+(-0.314307365683)*x[2]
        arg[1,3,1,1]=(-0.217030771566)*x[0]**o+(0.260108519031)*x[0]+(-0.686624725975)*x[1]**o+(-0.869322439609)*x[1]+(0.558806239315)*x[2]**o+(-0.440661121246)*x[2]
        arg[1,3,2,0]=(-0.839335541382)*x[0]**o+(0.228726621639)*x[0]+(0.354403444002)*x[1]**o+(0.215616048959)*x[1]+(0.49622647562)*x[2]**o+(-0.725639112729)*x[2]
        arg[1,3,2,1]=(-0.439687404632)*x[0]**o+(-0.820583921269)*x[0]+(-0.863730409631)*x[1]**o+(0.119773401911)*x[1]+(0.460158483187)*x[2]**o+(0.757153436592)*x[2]
        arg[1,4,0,0]=(-0.77288839701)*x[0]**o+(-0.570203714622)*x[0]+(0.93832772305)*x[1]**o+(-0.759089944596)*x[1]+(0.000174937961704)*x[2]**o+(0.812917484286)*x[2]
        arg[1,4,0,1]=(-0.25535378432)*x[0]**o+(-0.0938207170929)*x[0]+(0.0545321717685)*x[1]**o+(-0.414414792674)*x[1]+(-0.297698896712)*x[2]**o+(-0.367166660571)*x[2]
        arg[1,4,1,0]=(0.743408112248)*x[0]**o+(0.852863814785)*x[0]+(0.259483951249)*x[1]**o+(-0.519288864163)*x[1]+(-0.305772267727)*x[2]**o+(-0.401384318883)*x[2]
        arg[1,4,1,1]=(0.194889815352)*x[0]**o+(0.522093986804)*x[0]+(-0.343320977396)*x[1]**o+(0.21840236597)*x[1]+(-0.672107409136)*x[2]**o+(-0.600203805002)*x[2]
        arg[1,4,2,0]=(0.420860556361)*x[0]**o+(0.789879869049)*x[0]+(0.100987805744)*x[1]**o+(0.802794815123)*x[1]+(-0.69466030201)*x[2]**o+(-0.746043845966)*x[2]
        arg[1,4,2,1]=(-0.306621579211)*x[0]**o+(-0.0676982346933)*x[0]+(0.568861573715)*x[1]**o+(0.0823366334578)*x[1]+(0.501187255641)*x[2]**o+(-0.538615164478)*x[2]
        arg[2,0,0,0]=(-0.231052916706)*x[0]**o+(0.127790831537)*x[0]+(-0.475600435038)*x[1]**o+(0.431586151473)*x[1]+(-0.882365547235)*x[2]**o+(0.778027701879)*x[2]
        arg[2,0,0,1]=(0.0709490835498)*x[0]**o+(0.442614693755)*x[0]+(0.870405247692)*x[1]**o+(-0.648538779515)*x[1]+(0.751107652054)*x[2]**o+(0.0122956931127)*x[2]
        arg[2,0,1,0]=(0.703128329145)*x[0]**o+(0.259775843631)*x[0]+(0.311958888594)*x[1]**o+(-0.0445092402277)*x[1]+(0.338222457799)*x[2]**o+(0.845995300456)*x[2]
        arg[2,0,1,1]=(-0.653977898919)*x[0]**o+(0.0495539350857)*x[0]+(0.4013623408)*x[1]**o+(0.490747039887)*x[1]+(-0.480487636729)*x[2]**o+(0.361660115775)*x[2]
        arg[2,0,2,0]=(0.215778506945)*x[0]**o+(0.316853875053)*x[0]+(0.556770639829)*x[1]**o+(-0.499729882355)*x[1]+(0.692012327582)*x[2]**o+(0.00681314946195)*x[2]
        arg[2,0,2,1]=(0.432077857223)*x[0]**o+(-0.0497579036988)*x[0]+(-0.465403408565)*x[1]**o+(0.125421502956)*x[1]+(-0.313541879877)*x[2]**o+(0.113957941731)*x[2]
        arg[2,1,0,0]=(-0.528126103192)*x[0]**o+(-0.141306512096)*x[0]+(-0.505835313198)*x[1]**o+(-0.0433171534408)*x[1]+(-0.522465687628)*x[2]**o+(-0.16728285732)*x[2]
        arg[2,1,0,1]=(0.892019660948)*x[0]**o+(0.646944130193)*x[0]+(0.865495756748)*x[1]**o+(0.949095276008)*x[1]+(0.786394755068)*x[2]**o+(0.482441081687)*x[2]
        arg[2,1,1,0]=(-0.0485295711204)*x[0]**o+(0.250163927911)*x[0]+(0.0387861125998)*x[1]**o+(-0.0789121929817)*x[1]+(-0.982330126583)*x[2]**o+(0.66328870814)*x[2]
        arg[2,1,1,1]=(-0.494726556678)*x[0]**o+(-0.961342091528)*x[0]+(0.211410563083)*x[1]**o+(-0.884287887286)*x[1]+(-0.583401908135)*x[2]**o+(-0.361423905219)*x[2]
        arg[2,1,2,0]=(-0.326785246692)*x[0]**o+(0.596272945325)*x[0]+(0.625039539234)*x[1]**o+(0.88699688986)*x[1]+(-0.649803810419)*x[2]**o+(0.890718271469)*x[2]
        arg[2,1,2,1]=(0.702110072391)*x[0]**o+(0.72009439563)*x[0]+(-0.930941572511)*x[1]**o+(0.955877158334)*x[1]+(0.1211281332)*x[2]**o+(0.658738626852)*x[2]
        arg[2,2,0,0]=(-0.177402589697)*x[0]**o+(0.701090894256)*x[0]+(-0.00205333079339)*x[1]**o+(-0.427862310859)*x[1]+(0.529186374803)*x[2]**o+(-0.110462035551)*x[2]
        arg[2,2,0,1]=(-0.104112165367)*x[0]**o+(0.043539351452)*x[0]+(0.757384935306)*x[1]**o+(-0.451086492844)*x[1]+(0.0549627297033)*x[2]**o+(-0.311718955315)*x[2]
        arg[2,2,1,0]=(-0.579986090808)*x[0]**o+(-0.922872917924)*x[0]+(0.66422700207)*x[1]**o+(0.947304780214)*x[1]+(-0.0492553517704)*x[2]**o+(0.817225619153)*x[2]
        arg[2,2,1,1]=(0.306298476233)*x[0]**o+(-0.620190960783)*x[0]+(-0.684876074013)*x[1]**o+(-0.371829616982)*x[1]+(0.904709743321)*x[2]**o+(0.170802273852)*x[2]
        arg[2,2,2,0]=(0.317943399758)*x[0]**o+(0.518026054212)*x[0]+(-0.801462311167)*x[1]**o+(0.767424157898)*x[1]+(-0.547271995492)*x[2]**o+(-0.912619177188)*x[2]
        arg[2,2,2,1]=(-0.848711258358)*x[0]**o+(0.456489097468)*x[0]+(-0.310979078367)*x[1]**o+(-0.946717007726)*x[1]+(0.454363833646)*x[2]**o+(-0.715701991077)*x[2]
        arg[2,3,0,0]=(0.605493533014)*x[0]**o+(0.478437880368)*x[0]+(-0.672727491605)*x[1]**o+(0.831259256009)*x[1]+(-0.434153354239)*x[2]**o+(-0.505363218578)*x[2]
        arg[2,3,0,1]=(-0.0703539676643)*x[0]**o+(-0.452352344215)*x[0]+(0.0403568720903)*x[1]**o+(0.959415939826)*x[1]+(-0.784519698477)*x[2]**o+(-0.691664442958)*x[2]
        arg[2,3,1,0]=(0.696353768719)*x[0]**o+(0.295477297618)*x[0]+(0.580540484825)*x[1]**o+(0.098954967023)*x[1]+(-0.547594444517)*x[2]**o+(-0.574783665389)*x[2]
        arg[2,3,1,1]=(0.997841887408)*x[0]**o+(0.857864040549)*x[0]+(-0.413188343971)*x[1]**o+(0.344770631687)*x[1]+(-0.127257058552)*x[2]**o+(0.132618819456)*x[2]
        arg[2,3,2,0]=(0.174045751083)*x[0]**o+(0.866724759298)*x[0]+(0.0818516103455)*x[1]**o+(-0.891670609599)*x[1]+(0.982721717044)*x[2]**o+(0.378710155856)*x[2]
        arg[2,3,2,1]=(0.7760778906)*x[0]**o+(-0.710104290654)*x[0]+(-0.471259845977)*x[1]**o+(-0.791979211724)*x[1]+(-0.845652753706)*x[2]**o+(-0.301795837852)*x[2]
        arg[2,4,0,0]=(-0.698193658421)*x[0]**o+(-0.831956135385)*x[0]+(-0.442637915133)*x[1]**o+(0.108715813637)*x[1]+(-0.888348149694)*x[2]**o+(-0.90115888496)*x[2]
        arg[2,4,0,1]=(0.770237705577)*x[0]**o+(0.889304763618)*x[0]+(-0.139196183564)*x[1]**o+(0.332516232239)*x[1]+(-0.0218103253052)*x[2]**o+(-0.368033928854)*x[2]
        arg[2,4,1,0]=(0.675479351844)*x[0]**o+(0.880998418303)*x[0]+(0.408381837105)*x[1]**o+(0.717402364084)*x[1]+(-0.303902947422)*x[2]**o+(-0.322350049962)*x[2]
        arg[2,4,1,1]=(-0.480881673289)*x[0]**o+(0.30290141529)*x[0]+(-0.538411390886)*x[1]**o+(0.260090848702)*x[1]+(0.638130764808)*x[2]**o+(0.756338975575)*x[2]
        arg[2,4,2,0]=(-0.299233221889)*x[0]**o+(-0.26471083761)*x[0]+(0.717191172884)*x[1]**o+(0.488553216118)*x[1]+(0.718956726644)*x[2]**o+(0.370889025703)*x[2]
        arg[2,4,2,1]=(-0.989020272269)*x[0]**o+(-0.891423884362)*x[0]+(0.486153100693)*x[1]**o+(-0.971171395246)*x[1]+(-0.766942968184)*x[2]**o+(0.728728968862)*x[2]
        arg[3,0,0,0]=(0.756186246227)*x[0]**o+(0.76507856268)*x[0]+(0.771165000414)*x[1]**o+(0.673968873856)*x[1]+(0.827045846253)*x[2]**o+(-0.788265752412)*x[2]
        arg[3,0,0,1]=(-0.66925731439)*x[0]**o+(-0.0110127596435)*x[0]+(0.944388255414)*x[1]**o+(-0.613832072486)*x[1]+(0.0977542120724)*x[2]**o+(0.100776856689)*x[2]
        arg[3,0,1,0]=(0.905138713064)*x[0]**o+(0.673051338153)*x[0]+(0.240614253653)*x[1]**o+(-0.964132330033)*x[1]+(-0.114961018838)*x[2]**o+(-0.438756962988)*x[2]
        arg[3,0,1,1]=(-0.192259182094)*x[0]**o+(0.12315590119)*x[0]+(-0.907744420226)*x[1]**o+(-0.483981269809)*x[1]+(-0.169699431472)*x[2]**o+(-0.45231635406)*x[2]
        arg[3,0,2,0]=(0.895387255885)*x[0]**o+(0.23892997411)*x[0]+(-0.396010446618)*x[1]**o+(0.94853984413)*x[1]+(-0.0165259637241)*x[2]**o+(-0.312558539149)*x[2]
        arg[3,0,2,1]=(0.710478713361)*x[0]**o+(-0.109316102216)*x[0]+(-0.042711568986)*x[1]**o+(-0.448508085742)*x[1]+(-0.0979193346011)*x[2]**o+(0.93156672211)*x[2]
        arg[3,1,0,0]=(-0.21605108221)*x[0]**o+(0.410466481712)*x[0]+(0.799288386783)*x[1]**o+(0.622866017113)*x[1]+(-0.236698938159)*x[2]**o+(0.246625241928)*x[2]
        arg[3,1,0,1]=(0.989568527762)*x[0]**o+(0.0212906378151)*x[0]+(0.311279947324)*x[1]**o+(-0.231412902902)*x[1]+(-0.444760786142)*x[2]**o+(0.908171985421)*x[2]
        arg[3,1,1,0]=(0.42605234231)*x[0]**o+(-0.58629332944)*x[0]+(0.322063640761)*x[1]**o+(0.837771339825)*x[1]+(-0.911083211394)*x[2]**o+(-0.0699607581992)*x[2]
        arg[3,1,1,1]=(0.75449445591)*x[0]**o+(0.722970800307)*x[0]+(-0.582970919465)*x[1]**o+(-0.868818538212)*x[1]+(0.940067636181)*x[2]**o+(-0.659469743617)*x[2]
        arg[3,1,2,0]=(-0.391224924663)*x[0]**o+(-0.768173177737)*x[0]+(-0.856876347997)*x[1]**o+(-0.0451180499647)*x[1]+(0.198590679246)*x[2]**o+(0.187329095476)*x[2]
        arg[3,1,2,1]=(0.483309063092)*x[0]**o+(-0.0777241026893)*x[0]+(0.0438427457286)*x[1]**o+(-0.73195704867)*x[1]+(-0.439660745725)*x[2]**o+(0.279501672861)*x[2]
        arg[3,2,0,0]=(0.799935089139)*x[0]**o+(-0.306860809857)*x[0]+(0.169104975039)*x[1]**o+(-0.3480717525)*x[1]+(0.215231688)*x[2]**o+(-0.528775376945)*x[2]
        arg[3,2,0,1]=(0.917815789587)*x[0]**o+(0.0672910338102)*x[0]+(-0.186550440011)*x[1]**o+(0.922579714252)*x[1]+(0.924162379857)*x[2]**o+(-0.769320483618)*x[2]
        arg[3,2,1,0]=(0.604324468058)*x[0]**o+(-0.0109731419053)*x[0]+(0.229447318325)*x[1]**o+(-0.480693435232)*x[1]+(-0.881779256999)*x[2]**o+(0.288656394771)*x[2]
        arg[3,2,1,1]=(0.934680723352)*x[0]**o+(0.993838528376)*x[0]+(0.957781588502)*x[1]**o+(-0.859837057866)*x[1]+(-0.702139964493)*x[2]**o+(0.0538177124609)*x[2]
        arg[3,2,2,0]=(-0.142809047814)*x[0]**o+(-0.472594415134)*x[0]+(0.0381782768377)*x[1]**o+(0.727331491769)*x[1]+(0.0910992897581)*x[2]**o+(-0.845316063851)*x[2]
        arg[3,2,2,1]=(-0.217483358272)*x[0]**o+(0.991535092285)*x[0]+(-0.399175896468)*x[1]**o+(-0.665836512895)*x[1]+(0.0983700911493)*x[2]**o+(0.321811300155)*x[2]
        arg[3,3,0,0]=(0.797049585464)*x[0]**o+(-0.724467959299)*x[0]+(0.19971589973)*x[1]**o+(0.845928461864)*x[1]+(-0.365158866359)*x[2]**o+(-0.698274270132)*x[2]
        arg[3,3,0,1]=(-0.638474704878)*x[0]**o+(0.551866433352)*x[0]+(0.192309452979)*x[1]**o+(-0.703649325501)*x[1]+(0.823910578138)*x[2]**o+(-0.917446989634)*x[2]
        arg[3,3,1,0]=(0.217577282169)*x[0]**o+(0.521716803668)*x[0]+(-0.116587783856)*x[1]**o+(0.529672391417)*x[1]+(-0.798918158284)*x[2]**o+(-0.0996139511955)*x[2]
        arg[3,3,1,1]=(0.498965049079)*x[0]**o+(0.316489921186)*x[0]+(-0.0988355471817)*x[1]**o+(-0.0477508287681)*x[1]+(-0.778914523993)*x[2]**o+(0.189397812723)*x[2]
        arg[3,3,2,0]=(0.203823698152)*x[0]**o+(-0.340499414266)*x[0]+(0.353374693517)*x[1]**o+(-0.91035264494)*x[1]+(-0.779204084198)*x[2]**o+(-0.00276454165248)*x[2]
        arg[3,3,2,1]=(0.295232007123)*x[0]**o+(0.271766544502)*x[0]+(0.431956430336)*x[1]**o+(0.896481311536)*x[1]+(-0.325014902487)*x[2]**o+(0.70836170284)*x[2]
        arg[3,4,0,0]=(-0.407535202468)*x[0]**o+(-0.33055681091)*x[0]+(0.129762143151)*x[1]**o+(0.352087436228)*x[1]+(-0.538770401565)*x[2]**o+(-0.519071077045)*x[2]
        arg[3,4,0,1]=(0.846011435499)*x[0]**o+(0.172822133672)*x[0]+(-0.740732553847)*x[1]**o+(0.230062837231)*x[1]+(-0.231810007237)*x[2]**o+(-0.568366338923)*x[2]
        arg[3,4,1,0]=(-0.198091119057)*x[0]**o+(0.957989508794)*x[0]+(0.861739071175)*x[1]**o+(0.923450697883)*x[1]+(-0.755262504937)*x[2]**o+(0.102897579936)*x[2]
        arg[3,4,1,1]=(0.198829545353)*x[0]**o+(-0.757182006818)*x[0]+(0.254092337808)*x[1]**o+(-0.00815072620097)*x[1]+(0.347421988308)*x[2]**o+(-0.0215277616923)*x[2]
        arg[3,4,2,0]=(0.775062112529)*x[0]**o+(-0.77746792774)*x[0]+(-0.666490294869)*x[1]**o+(-0.679971069878)*x[1]+(-0.972450779894)*x[2]**o+(-0.705322282477)*x[2]
        arg[3,4,2,1]=(-0.0994927587382)*x[0]**o+(0.293071774786)*x[0]+(0.370784399397)*x[1]**o+(0.931411225946)*x[1]+(-0.55789163294)*x[2]**o+(0.0831890658592)*x[2]
        ref[0,0,0,0]=(-0.657703355307)*x_ref[0]**o+(0.397813831843)*x_ref[0]+(0.364460538717)*x_ref[1]**o+(-0.845626669617)*x_ref[1]+(-0.987144254126)*x_ref[2]**o+(-0.732384980726)*x_ref[2]
        ref[0,0,0,1]=(-0.13280304886)*x_ref[0]**o+(-0.953796807758)*x_ref[0]+(0.875723746654)*x_ref[1]**o+(0.0764745117242)*x_ref[1]+(-0.536085676221)*x_ref[2]**o+(-0.520611959678)*x_ref[2]
        ref[0,0,1,0]=(0.716429446578)*x_ref[0]**o+(-0.729164693441)*x_ref[0]+(-0.832505557428)*x_ref[1]**o+(-0.329200605611)*x_ref[1]+(0.0310162060964)*x_ref[2]**o+(-0.869978941999)*x_ref[2]
        ref[0,0,1,1]=(0.110253530584)*x_ref[0]**o+(0.995528486822)*x_ref[0]+(0.838577797395)*x_ref[1]**o+(0.443987752225)*x_ref[1]+(0.387337959057)*x_ref[2]**o+(0.548448752612)*x_ref[2]
        ref[0,0,2,0]=(-0.995737515431)*x_ref[0]**o+(-0.899807241717)*x_ref[0]+(-0.0803287764138)*x_ref[1]**o+(0.242680051556)*x_ref[1]+(0.58974836823)*x_ref[2]**o+(-0.498450188944)*x_ref[2]
        ref[0,0,2,1]=(0.341561541555)*x_ref[0]**o+(-0.00541988184597)*x_ref[0]+(-0.204832898701)*x_ref[1]**o+(-0.957247935865)*x_ref[1]+(-0.441002902064)*x_ref[2]**o+(0.756843480304)*x_ref[2]
        ref[0,1,0,0]=(0.136201526244)*x_ref[0]**o+(0.517677304193)*x_ref[0]+(0.523031610748)*x_ref[1]**o+(-0.908998053383)*x_ref[1]+(-0.629649678416)*x_ref[2]**o+(0.762977917512)*x_ref[2]
        ref[0,1,0,1]=(-0.103521080864)*x_ref[0]**o+(-0.434410222595)*x_ref[0]+(0.418239530241)*x_ref[1]**o+(0.65638475119)*x_ref[1]+(0.413036393129)*x_ref[2]**o+(-0.220365518353)*x_ref[2]
        ref[0,1,1,0]=(-0.415690201483)*x_ref[0]**o+(-0.969323394689)*x_ref[0]+(-0.295639420195)*x_ref[1]**o+(-0.24915002749)*x_ref[1]+(-0.00927627101467)*x_ref[2]**o+(0.520335261634)*x_ref[2]
        ref[0,1,1,1]=(-0.262653741797)*x_ref[0]**o+(-0.463590591666)*x_ref[0]+(0.568893042923)*x_ref[1]**o+(-0.869851332898)*x_ref[1]+(-0.463261399415)*x_ref[2]**o+(0.0281915686202)*x_ref[2]
        ref[0,1,2,0]=(0.377159503058)*x_ref[0]**o+(0.484512261943)*x_ref[0]+(0.0303594628925)*x_ref[1]**o+(0.482051243004)*x_ref[1]+(0.240643013949)*x_ref[2]**o+(-0.943908617705)*x_ref[2]
        ref[0,1,2,1]=(0.843854614097)*x_ref[0]**o+(0.018041534553)*x_ref[0]+(0.264198703361)*x_ref[1]**o+(0.732889517628)*x_ref[1]+(-0.154807384394)*x_ref[2]**o+(0.60513622187)*x_ref[2]
        ref[0,2,0,0]=(-0.234758118286)*x_ref[0]**o+(-0.77807396119)*x_ref[0]+(-0.97783091103)*x_ref[1]**o+(-0.716185223828)*x_ref[1]+(-0.573126474259)*x_ref[2]**o+(0.403110701555)*x_ref[2]
        ref[0,2,0,1]=(-0.853878344804)*x_ref[0]**o+(-0.482644826561)*x_ref[0]+(-0.826041149012)*x_ref[1]**o+(0.596376234798)*x_ref[1]+(-0.216400517274)*x_ref[2]**o+(-0.323127194259)*x_ref[2]
        ref[0,2,1,0]=(0.931551121492)*x_ref[0]**o+(-0.527132174379)*x_ref[0]+(0.373496704463)*x_ref[1]**o+(-0.375215826463)*x_ref[1]+(-0.639167297984)*x_ref[2]**o+(0.143608733189)*x_ref[2]
        ref[0,2,1,1]=(-0.982206000879)*x_ref[0]**o+(0.836589596534)*x_ref[0]+(0.606275000138)*x_ref[1]**o+(-0.821469453892)*x_ref[1]+(-0.633828380948)*x_ref[2]**o+(-0.637688777216)*x_ref[2]
        ref[0,2,2,0]=(0.441649557529)*x_ref[0]**o+(-0.34978365051)*x_ref[0]+(0.514175365405)*x_ref[1]**o+(-0.903212271234)*x_ref[1]+(0.70236887156)*x_ref[2]**o+(0.72953045189)*x_ref[2]
        ref[0,2,2,1]=(0.302787896978)*x_ref[0]**o+(0.157598071021)*x_ref[0]+(-0.357101947632)*x_ref[1]**o+(0.381098383156)*x_ref[1]+(-0.552546036689)*x_ref[2]**o+(0.448181721984)*x_ref[2]
        ref[0,3,0,0]=(0.662181077399)*x_ref[0]**o+(-0.283755239518)*x_ref[0]+(0.33694741278)*x_ref[1]**o+(0.275801685868)*x_ref[1]+(0.526155259101)*x_ref[2]**o+(-0.378241855559)*x_ref[2]
        ref[0,3,0,1]=(0.157822133581)*x_ref[0]**o+(0.25402072747)*x_ref[0]+(0.522842123099)*x_ref[1]**o+(0.510595773584)*x_ref[1]+(-0.829239126666)*x_ref[2]**o+(0.990879342044)*x_ref[2]
        ref[0,3,1,0]=(-0.775925280059)*x_ref[0]**o+(-0.726971388023)*x_ref[0]+(-0.630088730753)*x_ref[1]**o+(-0.731024178237)*x_ref[1]+(-0.288643541196)*x_ref[2]**o+(-0.283708428793)*x_ref[2]
        ref[0,3,1,1]=(-0.883176640445)*x_ref[0]**o+(-0.0810211976762)*x_ref[0]+(-0.102919217784)*x_ref[1]**o+(0.271384866385)*x_ref[1]+(0.980796900477)*x_ref[2]**o+(0.439294038672)*x_ref[2]
        ref[0,3,2,0]=(-0.826688221502)*x_ref[0]**o+(-0.909388347347)*x_ref[0]+(0.164406927722)*x_ref[1]**o+(0.60157189364)*x_ref[1]+(-0.885578176748)*x_ref[2]**o+(0.610008674247)*x_ref[2]
        ref[0,3,2,1]=(0.130154197193)*x_ref[0]**o+(-0.754703602975)*x_ref[0]+(0.00463710497017)*x_ref[1]**o+(0.502162916123)*x_ref[1]+(-0.749124548541)*x_ref[2]**o+(0.946592962092)*x_ref[2]
        ref[0,4,0,0]=(-0.0478347448906)*x_ref[0]**o+(-0.812937871882)*x_ref[0]+(-0.460982012551)*x_ref[1]**o+(0.317485998991)*x_ref[1]+(-0.593671683477)*x_ref[2]**o+(0.551385718668)*x_ref[2]
        ref[0,4,0,1]=(0.281998767502)*x_ref[0]**o+(-0.0613830311997)*x_ref[0]+(0.981429151019)*x_ref[1]**o+(-0.580567714008)*x_ref[1]+(0.0369483005331)*x_ref[2]**o+(-0.74132358588)*x_ref[2]
        ref[0,4,1,0]=(0.525845217483)*x_ref[0]**o+(0.907803915216)*x_ref[0]+(0.111429079263)*x_ref[1]**o+(-0.653030401377)*x_ref[1]+(0.0814793913696)*x_ref[2]**o+(-0.419201413875)*x_ref[2]
        ref[0,4,1,1]=(0.862892518463)*x_ref[0]**o+(0.34033875398)*x_ref[0]+(0.0371376953103)*x_ref[1]**o+(0.568160310726)*x_ref[1]+(-0.237231036717)*x_ref[2]**o+(-0.130046899302)*x_ref[2]
        ref[0,4,2,0]=(0.223651181698)*x_ref[0]**o+(0.730991398203)*x_ref[0]+(0.133551525027)*x_ref[1]**o+(0.685783540126)*x_ref[1]+(0.339586395814)*x_ref[2]**o+(-0.524371194863)*x_ref[2]
        ref[0,4,2,1]=(0.220110871383)*x_ref[0]**o+(-0.37385773877)*x_ref[0]+(0.773372000545)*x_ref[1]**o+(-0.4261356565)*x_ref[1]+(-0.619438438945)*x_ref[2]**o+(0.126288949127)*x_ref[2]
        ref[1,0,0,0]=(0.727680196631)*x_ref[0]**o+(-0.876280912599)*x_ref[0]+(0.576553782915)*x_ref[1]**o+(-0.0497780322367)*x_ref[1]+(-0.690257457282)*x_ref[2]**o+(-0.11130953054)*x_ref[2]
        ref[1,0,0,1]=(0.331863815938)*x_ref[0]**o+(0.169379824389)*x_ref[0]+(0.665507756012)*x_ref[1]**o+(0.94933357273)*x_ref[1]+(-0.610506070093)*x_ref[2]**o+(0.752476708027)*x_ref[2]
        ref[1,0,1,0]=(0.851320583415)*x_ref[0]**o+(-0.697235996161)*x_ref[0]+(0.235192165807)*x_ref[1]**o+(0.325473517022)*x_ref[1]+(-0.424683549673)*x_ref[2]**o+(0.228953796486)*x_ref[2]
        ref[1,0,1,1]=(0.804192098517)*x_ref[0]**o+(-0.684337197059)*x_ref[0]+(-0.349251615715)*x_ref[1]**o+(0.596165705743)*x_ref[1]+(-0.636033221182)*x_ref[2]**o+(-0.304367076307)*x_ref[2]
        ref[1,0,2,0]=(-0.237305477187)*x_ref[0]**o+(-0.997679652719)*x_ref[0]+(0.485994068389)*x_ref[1]**o+(0.44508530534)*x_ref[1]+(0.940139935628)*x_ref[2]**o+(0.0789221540062)*x_ref[2]
        ref[1,0,2,1]=(0.814928463364)*x_ref[0]**o+(-0.257693939789)*x_ref[0]+(0.179643785703)*x_ref[1]**o+(-0.696436530567)*x_ref[1]+(-0.592226945554)*x_ref[2]**o+(-0.454620411499)*x_ref[2]
        ref[1,1,0,0]=(-0.933925672439)*x_ref[0]**o+(0.212298180817)*x_ref[0]+(-0.447836316801)*x_ref[1]**o+(0.180796376138)*x_ref[1]+(-0.749910125701)*x_ref[2]**o+(0.134540433851)*x_ref[2]
        ref[1,1,0,1]=(-0.649009346978)*x_ref[0]**o+(0.0316819485953)*x_ref[0]+(-0.718001767835)*x_ref[1]**o+(0.289608435124)*x_ref[1]+(-0.906045493384)*x_ref[2]**o+(0.851021484125)*x_ref[2]
        ref[1,1,1,0]=(-0.216265404566)*x_ref[0]**o+(0.657073282172)*x_ref[0]+(-0.796155541438)*x_ref[1]**o+(0.338980498824)*x_ref[1]+(0.477444448193)*x_ref[2]**o+(0.674278909522)*x_ref[2]
        ref[1,1,1,1]=(-0.67605942954)*x_ref[0]**o+(0.457997253216)*x_ref[0]+(-0.973576381948)*x_ref[1]**o+(0.430432753773)*x_ref[1]+(0.0329169564451)*x_ref[2]**o+(-0.422512702485)*x_ref[2]
        ref[1,1,2,0]=(-0.32458256139)*x_ref[0]**o+(0.289568625922)*x_ref[0]+(-0.529564027476)*x_ref[1]**o+(0.596449833118)*x_ref[1]+(0.960461031989)*x_ref[2]**o+(0.00801425321069)*x_ref[2]
        ref[1,1,2,1]=(-0.0425883615923)*x_ref[0]**o+(-0.313926807333)*x_ref[0]+(0.766062773671)*x_ref[1]**o+(-0.857784560309)*x_ref[1]+(0.914867381682)*x_ref[2]**o+(-0.312493334639)*x_ref[2]
        ref[1,2,0,0]=(-0.637151767355)*x_ref[0]**o+(-0.679947136952)*x_ref[0]+(-0.0838349105494)*x_ref[1]**o+(0.00800835879666)*x_ref[1]+(0.385024194519)*x_ref[2]**o+(-0.627828236571)*x_ref[2]
        ref[1,2,0,1]=(-0.62580153007)*x_ref[0]**o+(-0.183834415542)*x_ref[0]+(0.322406823575)*x_ref[1]**o+(0.302908755868)*x_ref[1]+(0.727234001542)*x_ref[2]**o+(-0.933944878425)*x_ref[2]
        ref[1,2,1,0]=(-0.897198312615)*x_ref[0]**o+(0.825086526945)*x_ref[0]+(-0.031001758991)*x_ref[1]**o+(0.987845551304)*x_ref[1]+(-0.551212418179)*x_ref[2]**o+(0.657813241764)*x_ref[2]
        ref[1,2,1,1]=(-0.544595822278)*x_ref[0]**o+(0.595858429591)*x_ref[0]+(-0.557503238514)*x_ref[1]**o+(0.507560179011)*x_ref[1]+(-0.519438270771)*x_ref[2]**o+(-0.471678264995)*x_ref[2]
        ref[1,2,2,0]=(0.135616278099)*x_ref[0]**o+(0.40676743342)*x_ref[0]+(0.458810719213)*x_ref[1]**o+(-0.881620410228)*x_ref[1]+(-0.897267732846)*x_ref[2]**o+(0.665458992487)*x_ref[2]
        ref[1,2,2,1]=(0.83854610457)*x_ref[0]**o+(0.670263196782)*x_ref[0]+(0.551767734306)*x_ref[1]**o+(0.883934821215)*x_ref[1]+(0.455375261768)*x_ref[2]**o+(0.95604480045)*x_ref[2]
        ref[1,3,0,0]=(0.705853698882)*x_ref[0]**o+(0.84681314245)*x_ref[0]+(-0.0984020252408)*x_ref[1]**o+(0.943165990555)*x_ref[1]+(-0.428955841327)*x_ref[2]**o+(0.0961618993003)*x_ref[2]
        ref[1,3,0,1]=(0.583731800337)*x_ref[0]**o+(-0.738079255003)*x_ref[0]+(0.315556768944)*x_ref[1]**o+(-0.489631118687)*x_ref[1]+(0.371775779104)*x_ref[2]**o+(0.82029490872)*x_ref[2]
        ref[1,3,1,0]=(-0.885657545142)*x_ref[0]**o+(0.542706113286)*x_ref[0]+(0.800319452136)*x_ref[1]**o+(0.314194184897)*x_ref[1]+(-0.320905532427)*x_ref[2]**o+(-0.314307365683)*x_ref[2]
        ref[1,3,1,1]=(-0.217030771566)*x_ref[0]**o+(0.260108519031)*x_ref[0]+(-0.686624725975)*x_ref[1]**o+(-0.869322439609)*x_ref[1]+(0.558806239315)*x_ref[2]**o+(-0.440661121246)*x_ref[2]
        ref[1,3,2,0]=(-0.839335541382)*x_ref[0]**o+(0.228726621639)*x_ref[0]+(0.354403444002)*x_ref[1]**o+(0.215616048959)*x_ref[1]+(0.49622647562)*x_ref[2]**o+(-0.725639112729)*x_ref[2]
        ref[1,3,2,1]=(-0.439687404632)*x_ref[0]**o+(-0.820583921269)*x_ref[0]+(-0.863730409631)*x_ref[1]**o+(0.119773401911)*x_ref[1]+(0.460158483187)*x_ref[2]**o+(0.757153436592)*x_ref[2]
        ref[1,4,0,0]=(-0.77288839701)*x_ref[0]**o+(-0.570203714622)*x_ref[0]+(0.93832772305)*x_ref[1]**o+(-0.759089944596)*x_ref[1]+(0.000174937961704)*x_ref[2]**o+(0.812917484286)*x_ref[2]
        ref[1,4,0,1]=(-0.25535378432)*x_ref[0]**o+(-0.0938207170929)*x_ref[0]+(0.0545321717685)*x_ref[1]**o+(-0.414414792674)*x_ref[1]+(-0.297698896712)*x_ref[2]**o+(-0.367166660571)*x_ref[2]
        ref[1,4,1,0]=(0.743408112248)*x_ref[0]**o+(0.852863814785)*x_ref[0]+(0.259483951249)*x_ref[1]**o+(-0.519288864163)*x_ref[1]+(-0.305772267727)*x_ref[2]**o+(-0.401384318883)*x_ref[2]
        ref[1,4,1,1]=(0.194889815352)*x_ref[0]**o+(0.522093986804)*x_ref[0]+(-0.343320977396)*x_ref[1]**o+(0.21840236597)*x_ref[1]+(-0.672107409136)*x_ref[2]**o+(-0.600203805002)*x_ref[2]
        ref[1,4,2,0]=(0.420860556361)*x_ref[0]**o+(0.789879869049)*x_ref[0]+(0.100987805744)*x_ref[1]**o+(0.802794815123)*x_ref[1]+(-0.69466030201)*x_ref[2]**o+(-0.746043845966)*x_ref[2]
        ref[1,4,2,1]=(-0.306621579211)*x_ref[0]**o+(-0.0676982346933)*x_ref[0]+(0.568861573715)*x_ref[1]**o+(0.0823366334578)*x_ref[1]+(0.501187255641)*x_ref[2]**o+(-0.538615164478)*x_ref[2]
        ref[2,0,0,0]=(-0.231052916706)*x_ref[0]**o+(0.127790831537)*x_ref[0]+(-0.475600435038)*x_ref[1]**o+(0.431586151473)*x_ref[1]+(-0.882365547235)*x_ref[2]**o+(0.778027701879)*x_ref[2]
        ref[2,0,0,1]=(0.0709490835498)*x_ref[0]**o+(0.442614693755)*x_ref[0]+(0.870405247692)*x_ref[1]**o+(-0.648538779515)*x_ref[1]+(0.751107652054)*x_ref[2]**o+(0.0122956931127)*x_ref[2]
        ref[2,0,1,0]=(0.703128329145)*x_ref[0]**o+(0.259775843631)*x_ref[0]+(0.311958888594)*x_ref[1]**o+(-0.0445092402277)*x_ref[1]+(0.338222457799)*x_ref[2]**o+(0.845995300456)*x_ref[2]
        ref[2,0,1,1]=(-0.653977898919)*x_ref[0]**o+(0.0495539350857)*x_ref[0]+(0.4013623408)*x_ref[1]**o+(0.490747039887)*x_ref[1]+(-0.480487636729)*x_ref[2]**o+(0.361660115775)*x_ref[2]
        ref[2,0,2,0]=(0.215778506945)*x_ref[0]**o+(0.316853875053)*x_ref[0]+(0.556770639829)*x_ref[1]**o+(-0.499729882355)*x_ref[1]+(0.692012327582)*x_ref[2]**o+(0.00681314946195)*x_ref[2]
        ref[2,0,2,1]=(0.432077857223)*x_ref[0]**o+(-0.0497579036988)*x_ref[0]+(-0.465403408565)*x_ref[1]**o+(0.125421502956)*x_ref[1]+(-0.313541879877)*x_ref[2]**o+(0.113957941731)*x_ref[2]
        ref[2,1,0,0]=(-0.528126103192)*x_ref[0]**o+(-0.141306512096)*x_ref[0]+(-0.505835313198)*x_ref[1]**o+(-0.0433171534408)*x_ref[1]+(-0.522465687628)*x_ref[2]**o+(-0.16728285732)*x_ref[2]
        ref[2,1,0,1]=(0.892019660948)*x_ref[0]**o+(0.646944130193)*x_ref[0]+(0.865495756748)*x_ref[1]**o+(0.949095276008)*x_ref[1]+(0.786394755068)*x_ref[2]**o+(0.482441081687)*x_ref[2]
        ref[2,1,1,0]=(-0.0485295711204)*x_ref[0]**o+(0.250163927911)*x_ref[0]+(0.0387861125998)*x_ref[1]**o+(-0.0789121929817)*x_ref[1]+(-0.982330126583)*x_ref[2]**o+(0.66328870814)*x_ref[2]
        ref[2,1,1,1]=(-0.494726556678)*x_ref[0]**o+(-0.961342091528)*x_ref[0]+(0.211410563083)*x_ref[1]**o+(-0.884287887286)*x_ref[1]+(-0.583401908135)*x_ref[2]**o+(-0.361423905219)*x_ref[2]
        ref[2,1,2,0]=(-0.326785246692)*x_ref[0]**o+(0.596272945325)*x_ref[0]+(0.625039539234)*x_ref[1]**o+(0.88699688986)*x_ref[1]+(-0.649803810419)*x_ref[2]**o+(0.890718271469)*x_ref[2]
        ref[2,1,2,1]=(0.702110072391)*x_ref[0]**o+(0.72009439563)*x_ref[0]+(-0.930941572511)*x_ref[1]**o+(0.955877158334)*x_ref[1]+(0.1211281332)*x_ref[2]**o+(0.658738626852)*x_ref[2]
        ref[2,2,0,0]=(-0.177402589697)*x_ref[0]**o+(0.701090894256)*x_ref[0]+(-0.00205333079339)*x_ref[1]**o+(-0.427862310859)*x_ref[1]+(0.529186374803)*x_ref[2]**o+(-0.110462035551)*x_ref[2]
        ref[2,2,0,1]=(-0.104112165367)*x_ref[0]**o+(0.043539351452)*x_ref[0]+(0.757384935306)*x_ref[1]**o+(-0.451086492844)*x_ref[1]+(0.0549627297033)*x_ref[2]**o+(-0.311718955315)*x_ref[2]
        ref[2,2,1,0]=(-0.579986090808)*x_ref[0]**o+(-0.922872917924)*x_ref[0]+(0.66422700207)*x_ref[1]**o+(0.947304780214)*x_ref[1]+(-0.0492553517704)*x_ref[2]**o+(0.817225619153)*x_ref[2]
        ref[2,2,1,1]=(0.306298476233)*x_ref[0]**o+(-0.620190960783)*x_ref[0]+(-0.684876074013)*x_ref[1]**o+(-0.371829616982)*x_ref[1]+(0.904709743321)*x_ref[2]**o+(0.170802273852)*x_ref[2]
        ref[2,2,2,0]=(0.317943399758)*x_ref[0]**o+(0.518026054212)*x_ref[0]+(-0.801462311167)*x_ref[1]**o+(0.767424157898)*x_ref[1]+(-0.547271995492)*x_ref[2]**o+(-0.912619177188)*x_ref[2]
        ref[2,2,2,1]=(-0.848711258358)*x_ref[0]**o+(0.456489097468)*x_ref[0]+(-0.310979078367)*x_ref[1]**o+(-0.946717007726)*x_ref[1]+(0.454363833646)*x_ref[2]**o+(-0.715701991077)*x_ref[2]
        ref[2,3,0,0]=(0.605493533014)*x_ref[0]**o+(0.478437880368)*x_ref[0]+(-0.672727491605)*x_ref[1]**o+(0.831259256009)*x_ref[1]+(-0.434153354239)*x_ref[2]**o+(-0.505363218578)*x_ref[2]
        ref[2,3,0,1]=(-0.0703539676643)*x_ref[0]**o+(-0.452352344215)*x_ref[0]+(0.0403568720903)*x_ref[1]**o+(0.959415939826)*x_ref[1]+(-0.784519698477)*x_ref[2]**o+(-0.691664442958)*x_ref[2]
        ref[2,3,1,0]=(0.696353768719)*x_ref[0]**o+(0.295477297618)*x_ref[0]+(0.580540484825)*x_ref[1]**o+(0.098954967023)*x_ref[1]+(-0.547594444517)*x_ref[2]**o+(-0.574783665389)*x_ref[2]
        ref[2,3,1,1]=(0.997841887408)*x_ref[0]**o+(0.857864040549)*x_ref[0]+(-0.413188343971)*x_ref[1]**o+(0.344770631687)*x_ref[1]+(-0.127257058552)*x_ref[2]**o+(0.132618819456)*x_ref[2]
        ref[2,3,2,0]=(0.174045751083)*x_ref[0]**o+(0.866724759298)*x_ref[0]+(0.0818516103455)*x_ref[1]**o+(-0.891670609599)*x_ref[1]+(0.982721717044)*x_ref[2]**o+(0.378710155856)*x_ref[2]
        ref[2,3,2,1]=(0.7760778906)*x_ref[0]**o+(-0.710104290654)*x_ref[0]+(-0.471259845977)*x_ref[1]**o+(-0.791979211724)*x_ref[1]+(-0.845652753706)*x_ref[2]**o+(-0.301795837852)*x_ref[2]
        ref[2,4,0,0]=(-0.698193658421)*x_ref[0]**o+(-0.831956135385)*x_ref[0]+(-0.442637915133)*x_ref[1]**o+(0.108715813637)*x_ref[1]+(-0.888348149694)*x_ref[2]**o+(-0.90115888496)*x_ref[2]
        ref[2,4,0,1]=(0.770237705577)*x_ref[0]**o+(0.889304763618)*x_ref[0]+(-0.139196183564)*x_ref[1]**o+(0.332516232239)*x_ref[1]+(-0.0218103253052)*x_ref[2]**o+(-0.368033928854)*x_ref[2]
        ref[2,4,1,0]=(0.675479351844)*x_ref[0]**o+(0.880998418303)*x_ref[0]+(0.408381837105)*x_ref[1]**o+(0.717402364084)*x_ref[1]+(-0.303902947422)*x_ref[2]**o+(-0.322350049962)*x_ref[2]
        ref[2,4,1,1]=(-0.480881673289)*x_ref[0]**o+(0.30290141529)*x_ref[0]+(-0.538411390886)*x_ref[1]**o+(0.260090848702)*x_ref[1]+(0.638130764808)*x_ref[2]**o+(0.756338975575)*x_ref[2]
        ref[2,4,2,0]=(-0.299233221889)*x_ref[0]**o+(-0.26471083761)*x_ref[0]+(0.717191172884)*x_ref[1]**o+(0.488553216118)*x_ref[1]+(0.718956726644)*x_ref[2]**o+(0.370889025703)*x_ref[2]
        ref[2,4,2,1]=(-0.989020272269)*x_ref[0]**o+(-0.891423884362)*x_ref[0]+(0.486153100693)*x_ref[1]**o+(-0.971171395246)*x_ref[1]+(-0.766942968184)*x_ref[2]**o+(0.728728968862)*x_ref[2]
        ref[3,0,0,0]=(0.756186246227)*x_ref[0]**o+(0.76507856268)*x_ref[0]+(0.771165000414)*x_ref[1]**o+(0.673968873856)*x_ref[1]+(0.827045846253)*x_ref[2]**o+(-0.788265752412)*x_ref[2]
        ref[3,0,0,1]=(-0.66925731439)*x_ref[0]**o+(-0.0110127596435)*x_ref[0]+(0.944388255414)*x_ref[1]**o+(-0.613832072486)*x_ref[1]+(0.0977542120724)*x_ref[2]**o+(0.100776856689)*x_ref[2]
        ref[3,0,1,0]=(0.905138713064)*x_ref[0]**o+(0.673051338153)*x_ref[0]+(0.240614253653)*x_ref[1]**o+(-0.964132330033)*x_ref[1]+(-0.114961018838)*x_ref[2]**o+(-0.438756962988)*x_ref[2]
        ref[3,0,1,1]=(-0.192259182094)*x_ref[0]**o+(0.12315590119)*x_ref[0]+(-0.907744420226)*x_ref[1]**o+(-0.483981269809)*x_ref[1]+(-0.169699431472)*x_ref[2]**o+(-0.45231635406)*x_ref[2]
        ref[3,0,2,0]=(0.895387255885)*x_ref[0]**o+(0.23892997411)*x_ref[0]+(-0.396010446618)*x_ref[1]**o+(0.94853984413)*x_ref[1]+(-0.0165259637241)*x_ref[2]**o+(-0.312558539149)*x_ref[2]
        ref[3,0,2,1]=(0.710478713361)*x_ref[0]**o+(-0.109316102216)*x_ref[0]+(-0.042711568986)*x_ref[1]**o+(-0.448508085742)*x_ref[1]+(-0.0979193346011)*x_ref[2]**o+(0.93156672211)*x_ref[2]
        ref[3,1,0,0]=(-0.21605108221)*x_ref[0]**o+(0.410466481712)*x_ref[0]+(0.799288386783)*x_ref[1]**o+(0.622866017113)*x_ref[1]+(-0.236698938159)*x_ref[2]**o+(0.246625241928)*x_ref[2]
        ref[3,1,0,1]=(0.989568527762)*x_ref[0]**o+(0.0212906378151)*x_ref[0]+(0.311279947324)*x_ref[1]**o+(-0.231412902902)*x_ref[1]+(-0.444760786142)*x_ref[2]**o+(0.908171985421)*x_ref[2]
        ref[3,1,1,0]=(0.42605234231)*x_ref[0]**o+(-0.58629332944)*x_ref[0]+(0.322063640761)*x_ref[1]**o+(0.837771339825)*x_ref[1]+(-0.911083211394)*x_ref[2]**o+(-0.0699607581992)*x_ref[2]
        ref[3,1,1,1]=(0.75449445591)*x_ref[0]**o+(0.722970800307)*x_ref[0]+(-0.582970919465)*x_ref[1]**o+(-0.868818538212)*x_ref[1]+(0.940067636181)*x_ref[2]**o+(-0.659469743617)*x_ref[2]
        ref[3,1,2,0]=(-0.391224924663)*x_ref[0]**o+(-0.768173177737)*x_ref[0]+(-0.856876347997)*x_ref[1]**o+(-0.0451180499647)*x_ref[1]+(0.198590679246)*x_ref[2]**o+(0.187329095476)*x_ref[2]
        ref[3,1,2,1]=(0.483309063092)*x_ref[0]**o+(-0.0777241026893)*x_ref[0]+(0.0438427457286)*x_ref[1]**o+(-0.73195704867)*x_ref[1]+(-0.439660745725)*x_ref[2]**o+(0.279501672861)*x_ref[2]
        ref[3,2,0,0]=(0.799935089139)*x_ref[0]**o+(-0.306860809857)*x_ref[0]+(0.169104975039)*x_ref[1]**o+(-0.3480717525)*x_ref[1]+(0.215231688)*x_ref[2]**o+(-0.528775376945)*x_ref[2]
        ref[3,2,0,1]=(0.917815789587)*x_ref[0]**o+(0.0672910338102)*x_ref[0]+(-0.186550440011)*x_ref[1]**o+(0.922579714252)*x_ref[1]+(0.924162379857)*x_ref[2]**o+(-0.769320483618)*x_ref[2]
        ref[3,2,1,0]=(0.604324468058)*x_ref[0]**o+(-0.0109731419053)*x_ref[0]+(0.229447318325)*x_ref[1]**o+(-0.480693435232)*x_ref[1]+(-0.881779256999)*x_ref[2]**o+(0.288656394771)*x_ref[2]
        ref[3,2,1,1]=(0.934680723352)*x_ref[0]**o+(0.993838528376)*x_ref[0]+(0.957781588502)*x_ref[1]**o+(-0.859837057866)*x_ref[1]+(-0.702139964493)*x_ref[2]**o+(0.0538177124609)*x_ref[2]
        ref[3,2,2,0]=(-0.142809047814)*x_ref[0]**o+(-0.472594415134)*x_ref[0]+(0.0381782768377)*x_ref[1]**o+(0.727331491769)*x_ref[1]+(0.0910992897581)*x_ref[2]**o+(-0.845316063851)*x_ref[2]
        ref[3,2,2,1]=(-0.217483358272)*x_ref[0]**o+(0.991535092285)*x_ref[0]+(-0.399175896468)*x_ref[1]**o+(-0.665836512895)*x_ref[1]+(0.0983700911493)*x_ref[2]**o+(0.321811300155)*x_ref[2]
        ref[3,3,0,0]=(0.797049585464)*x_ref[0]**o+(-0.724467959299)*x_ref[0]+(0.19971589973)*x_ref[1]**o+(0.845928461864)*x_ref[1]+(-0.365158866359)*x_ref[2]**o+(-0.698274270132)*x_ref[2]
        ref[3,3,0,1]=(-0.638474704878)*x_ref[0]**o+(0.551866433352)*x_ref[0]+(0.192309452979)*x_ref[1]**o+(-0.703649325501)*x_ref[1]+(0.823910578138)*x_ref[2]**o+(-0.917446989634)*x_ref[2]
        ref[3,3,1,0]=(0.217577282169)*x_ref[0]**o+(0.521716803668)*x_ref[0]+(-0.116587783856)*x_ref[1]**o+(0.529672391417)*x_ref[1]+(-0.798918158284)*x_ref[2]**o+(-0.0996139511955)*x_ref[2]
        ref[3,3,1,1]=(0.498965049079)*x_ref[0]**o+(0.316489921186)*x_ref[0]+(-0.0988355471817)*x_ref[1]**o+(-0.0477508287681)*x_ref[1]+(-0.778914523993)*x_ref[2]**o+(0.189397812723)*x_ref[2]
        ref[3,3,2,0]=(0.203823698152)*x_ref[0]**o+(-0.340499414266)*x_ref[0]+(0.353374693517)*x_ref[1]**o+(-0.91035264494)*x_ref[1]+(-0.779204084198)*x_ref[2]**o+(-0.00276454165248)*x_ref[2]
        ref[3,3,2,1]=(0.295232007123)*x_ref[0]**o+(0.271766544502)*x_ref[0]+(0.431956430336)*x_ref[1]**o+(0.896481311536)*x_ref[1]+(-0.325014902487)*x_ref[2]**o+(0.70836170284)*x_ref[2]
        ref[3,4,0,0]=(-0.407535202468)*x_ref[0]**o+(-0.33055681091)*x_ref[0]+(0.129762143151)*x_ref[1]**o+(0.352087436228)*x_ref[1]+(-0.538770401565)*x_ref[2]**o+(-0.519071077045)*x_ref[2]
        ref[3,4,0,1]=(0.846011435499)*x_ref[0]**o+(0.172822133672)*x_ref[0]+(-0.740732553847)*x_ref[1]**o+(0.230062837231)*x_ref[1]+(-0.231810007237)*x_ref[2]**o+(-0.568366338923)*x_ref[2]
        ref[3,4,1,0]=(-0.198091119057)*x_ref[0]**o+(0.957989508794)*x_ref[0]+(0.861739071175)*x_ref[1]**o+(0.923450697883)*x_ref[1]+(-0.755262504937)*x_ref[2]**o+(0.102897579936)*x_ref[2]
        ref[3,4,1,1]=(0.198829545353)*x_ref[0]**o+(-0.757182006818)*x_ref[0]+(0.254092337808)*x_ref[1]**o+(-0.00815072620097)*x_ref[1]+(0.347421988308)*x_ref[2]**o+(-0.0215277616923)*x_ref[2]
        ref[3,4,2,0]=(0.775062112529)*x_ref[0]**o+(-0.77746792774)*x_ref[0]+(-0.666490294869)*x_ref[1]**o+(-0.679971069878)*x_ref[1]+(-0.972450779894)*x_ref[2]**o+(-0.705322282477)*x_ref[2]
        ref[3,4,2,1]=(-0.0994927587382)*x_ref[0]**o+(0.293071774786)*x_ref[0]+(0.370784399397)*x_ref[1]**o+(0.931411225946)*x_ref[1]+(-0.55789163294)*x_ref[2]**o+(0.0831890658592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref),self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0969659653911)*x[0]**o+(-0.0778467995236)*x[0]+(-0.896520347956)*x[1]**o+(0.452579859858)*x[1]
        ref=(0.0969659653911)*x_ref[0]**o+(-0.0778467995236)*x_ref[0]+(-0.896520347956)*x_ref[1]**o+(0.452579859858)*x_ref[1]
      else:
        arg=(-0.62506591451)*x[0]**o+(0.970735975624)*x[0]+(-0.563055321362)*x[1]**o+(-0.281482915069)*x[1]+(-0.6461258967)*x[2]**o+(0.786610350609)*x[2]
        ref=(-0.62506591451)*x_ref[0]**o+(0.970735975624)*x_ref[0]+(-0.563055321362)*x_ref[1]**o+(-0.281482915069)*x_ref[1]+(-0.6461258967)*x_ref[2]**o+(0.786610350609)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.725240039321)*x[0]**o+(0.326302203613)*x[0]+(0.356072156898)*x[1]**o+(0.980066388307)*x[1]
        arg[1]=(0.171551389095)*x[0]**o+(0.379133470853)*x[0]+(0.120608077039)*x[1]**o+(-0.413701037132)*x[1]
        ref[0]=(0.725240039321)*x_ref[0]**o+(0.326302203613)*x_ref[0]+(0.356072156898)*x_ref[1]**o+(0.980066388307)*x_ref[1]
        ref[1]=(0.171551389095)*x_ref[0]**o+(0.379133470853)*x_ref[0]+(0.120608077039)*x_ref[1]**o+(-0.413701037132)*x_ref[1]
      else:
        arg[0]=(-0.656095745624)*x[0]**o+(-0.540430373724)*x[0]+(-0.109962817674)*x[1]**o+(-0.969692783268)*x[1]+(-0.377994841279)*x[2]**o+(-0.664035030664)*x[2]
        arg[1]=(0.768498465456)*x[0]**o+(0.647294499552)*x[0]+(0.824672861058)*x[1]**o+(-0.848088780836)*x[1]+(0.045558138597)*x[2]**o+(-0.209365900036)*x[2]
        ref[0]=(-0.656095745624)*x_ref[0]**o+(-0.540430373724)*x_ref[0]+(-0.109962817674)*x_ref[1]**o+(-0.969692783268)*x_ref[1]+(-0.377994841279)*x_ref[2]**o+(-0.664035030664)*x_ref[2]
        ref[1]=(0.768498465456)*x_ref[0]**o+(0.647294499552)*x_ref[0]+(0.824672861058)*x_ref[1]**o+(-0.848088780836)*x_ref[1]+(0.045558138597)*x_ref[2]**o+(-0.209365900036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.694866182164)*x[0]**o+(0.759641527396)*x[0]+(-0.696641139757)*x[1]**o+(0.617978953453)*x[1]
        arg[0,1]=(0.75058922213)*x[0]**o+(0.580832851854)*x[0]+(-0.237039031011)*x[1]**o+(0.53518450519)*x[1]
        arg[0,2]=(-0.191933019173)*x[0]**o+(-0.0638088686782)*x[0]+(0.871645336149)*x[1]**o+(0.256502442034)*x[1]
        arg[0,3]=(0.829189580591)*x[0]**o+(0.186064016504)*x[0]+(-0.204496850484)*x[1]**o+(-0.314535598074)*x[1]
        arg[0,4]=(0.170968249419)*x[0]**o+(-0.719959835105)*x[0]+(0.367000223059)*x[1]**o+(-0.210239222752)*x[1]
        arg[1,0]=(-0.424311493894)*x[0]**o+(-0.381097143579)*x[0]+(-0.54925292526)*x[1]**o+(-0.393349131169)*x[1]
        arg[1,1]=(0.148181508143)*x[0]**o+(-0.224530937841)*x[0]+(0.778748719265)*x[1]**o+(-0.522104185968)*x[1]
        arg[1,2]=(0.134819103337)*x[0]**o+(0.0388165865124)*x[0]+(-0.772318490115)*x[1]**o+(-0.928611227188)*x[1]
        arg[1,3]=(0.790807030518)*x[0]**o+(-0.849964960983)*x[0]+(-0.79407080953)*x[1]**o+(-0.626315155005)*x[1]
        arg[1,4]=(-0.998842104726)*x[0]**o+(-0.325938405818)*x[0]+(-0.81445420822)*x[1]**o+(0.868281547664)*x[1]
        arg[2,0]=(-0.975572739854)*x[0]**o+(-0.0630975707316)*x[0]+(0.0688031679656)*x[1]**o+(0.557152464527)*x[1]
        arg[2,1]=(0.389814261767)*x[0]**o+(0.523891881188)*x[0]+(-0.558388242102)*x[1]**o+(-0.45720905598)*x[1]
        arg[2,2]=(0.763327517145)*x[0]**o+(0.637954909904)*x[0]+(0.407602154355)*x[1]**o+(0.707118271445)*x[1]
        arg[2,3]=(-0.823371482965)*x[0]**o+(0.918685041464)*x[0]+(0.991991462022)*x[1]**o+(0.489892661208)*x[1]
        arg[2,4]=(-0.661567455889)*x[0]**o+(-0.00566795402284)*x[0]+(0.86705493299)*x[1]**o+(0.00158321121677)*x[1]
        arg[3,0]=(0.365456232671)*x[0]**o+(-0.587373381962)*x[0]+(0.751291631507)*x[1]**o+(-0.0605248375222)*x[1]
        arg[3,1]=(0.546630294768)*x[0]**o+(-0.903068297271)*x[0]+(0.486313468693)*x[1]**o+(0.355322343068)*x[1]
        arg[3,2]=(0.233755056849)*x[0]**o+(-0.350375425427)*x[0]+(0.0115435986629)*x[1]**o+(-0.713895312694)*x[1]
        arg[3,3]=(0.385411302842)*x[0]**o+(0.0733743700003)*x[0]+(0.439219708301)*x[1]**o+(0.129391539278)*x[1]
        arg[3,4]=(-0.194268604537)*x[0]**o+(0.175829932738)*x[0]+(0.866342013931)*x[1]**o+(-0.213891435501)*x[1]
        ref[0,0]=(0.694866182164)*x_ref[0]**o+(0.759641527396)*x_ref[0]+(-0.696641139757)*x_ref[1]**o+(0.617978953453)*x_ref[1]
        ref[0,1]=(0.75058922213)*x_ref[0]**o+(0.580832851854)*x_ref[0]+(-0.237039031011)*x_ref[1]**o+(0.53518450519)*x_ref[1]
        ref[0,2]=(-0.191933019173)*x_ref[0]**o+(-0.0638088686782)*x_ref[0]+(0.871645336149)*x_ref[1]**o+(0.256502442034)*x_ref[1]
        ref[0,3]=(0.829189580591)*x_ref[0]**o+(0.186064016504)*x_ref[0]+(-0.204496850484)*x_ref[1]**o+(-0.314535598074)*x_ref[1]
        ref[0,4]=(0.170968249419)*x_ref[0]**o+(-0.719959835105)*x_ref[0]+(0.367000223059)*x_ref[1]**o+(-0.210239222752)*x_ref[1]
        ref[1,0]=(-0.424311493894)*x_ref[0]**o+(-0.381097143579)*x_ref[0]+(-0.54925292526)*x_ref[1]**o+(-0.393349131169)*x_ref[1]
        ref[1,1]=(0.148181508143)*x_ref[0]**o+(-0.224530937841)*x_ref[0]+(0.778748719265)*x_ref[1]**o+(-0.522104185968)*x_ref[1]
        ref[1,2]=(0.134819103337)*x_ref[0]**o+(0.0388165865124)*x_ref[0]+(-0.772318490115)*x_ref[1]**o+(-0.928611227188)*x_ref[1]
        ref[1,3]=(0.790807030518)*x_ref[0]**o+(-0.849964960983)*x_ref[0]+(-0.79407080953)*x_ref[1]**o+(-0.626315155005)*x_ref[1]
        ref[1,4]=(-0.998842104726)*x_ref[0]**o+(-0.325938405818)*x_ref[0]+(-0.81445420822)*x_ref[1]**o+(0.868281547664)*x_ref[1]
        ref[2,0]=(-0.975572739854)*x_ref[0]**o+(-0.0630975707316)*x_ref[0]+(0.0688031679656)*x_ref[1]**o+(0.557152464527)*x_ref[1]
        ref[2,1]=(0.389814261767)*x_ref[0]**o+(0.523891881188)*x_ref[0]+(-0.558388242102)*x_ref[1]**o+(-0.45720905598)*x_ref[1]
        ref[2,2]=(0.763327517145)*x_ref[0]**o+(0.637954909904)*x_ref[0]+(0.407602154355)*x_ref[1]**o+(0.707118271445)*x_ref[1]
        ref[2,3]=(-0.823371482965)*x_ref[0]**o+(0.918685041464)*x_ref[0]+(0.991991462022)*x_ref[1]**o+(0.489892661208)*x_ref[1]
        ref[2,4]=(-0.661567455889)*x_ref[0]**o+(-0.00566795402284)*x_ref[0]+(0.86705493299)*x_ref[1]**o+(0.00158321121677)*x_ref[1]
        ref[3,0]=(0.365456232671)*x_ref[0]**o+(-0.587373381962)*x_ref[0]+(0.751291631507)*x_ref[1]**o+(-0.0605248375222)*x_ref[1]
        ref[3,1]=(0.546630294768)*x_ref[0]**o+(-0.903068297271)*x_ref[0]+(0.486313468693)*x_ref[1]**o+(0.355322343068)*x_ref[1]
        ref[3,2]=(0.233755056849)*x_ref[0]**o+(-0.350375425427)*x_ref[0]+(0.0115435986629)*x_ref[1]**o+(-0.713895312694)*x_ref[1]
        ref[3,3]=(0.385411302842)*x_ref[0]**o+(0.0733743700003)*x_ref[0]+(0.439219708301)*x_ref[1]**o+(0.129391539278)*x_ref[1]
        ref[3,4]=(-0.194268604537)*x_ref[0]**o+(0.175829932738)*x_ref[0]+(0.866342013931)*x_ref[1]**o+(-0.213891435501)*x_ref[1]
      else:
        arg[0,0]=(-0.479138946804)*x[0]**o+(0.240408443007)*x[0]+(0.309155796205)*x[1]**o+(-0.124360991957)*x[1]+(0.00301047542037)*x[2]**o+(-0.039439802799)*x[2]
        arg[0,1]=(0.41979402771)*x[0]**o+(0.798738902847)*x[0]+(0.619960746922)*x[1]**o+(-0.551662130664)*x[1]+(0.914293623695)*x[2]**o+(0.6859768812)*x[2]
        arg[0,2]=(0.744593492696)*x[0]**o+(-0.31743031745)*x[0]+(0.389545123971)*x[1]**o+(0.592024987288)*x[1]+(0.63456095022)*x[2]**o+(-0.746571796061)*x[2]
        arg[0,3]=(-0.632549503649)*x[0]**o+(-0.687745599598)*x[0]+(0.794176042123)*x[1]**o+(-0.971466375953)*x[1]+(-0.086939632062)*x[2]**o+(0.373736042024)*x[2]
        arg[0,4]=(-0.966879538217)*x[0]**o+(-0.609291970974)*x[0]+(0.000409196978095)*x[1]**o+(0.623934484362)*x[1]+(-0.1694749999)*x[2]**o+(0.131778917979)*x[2]
        arg[1,0]=(0.763452444611)*x[0]**o+(0.805147582411)*x[0]+(0.999704124146)*x[1]**o+(0.0635253753138)*x[1]+(-0.717889410548)*x[2]**o+(-0.840765968423)*x[2]
        arg[1,1]=(-0.26910479702)*x[0]**o+(0.594093239771)*x[0]+(-0.536150372834)*x[1]**o+(-0.768506263316)*x[1]+(-0.924510063594)*x[2]**o+(-0.950568839784)*x[2]
        arg[1,2]=(-0.452317334239)*x[0]**o+(0.186343827048)*x[0]+(0.921015823436)*x[1]**o+(-0.656809197675)*x[1]+(-0.841898648461)*x[2]**o+(0.267827136605)*x[2]
        arg[1,3]=(-0.880873336233)*x[0]**o+(-0.915350352959)*x[0]+(0.92197459758)*x[1]**o+(0.20861139921)*x[1]+(0.776743964515)*x[2]**o+(0.858569288507)*x[2]
        arg[1,4]=(0.422773256864)*x[0]**o+(0.397743860671)*x[0]+(0.726470981309)*x[1]**o+(-0.233534612019)*x[1]+(-0.350345927201)*x[2]**o+(-0.192294374344)*x[2]
        arg[2,0]=(0.0706785590906)*x[0]**o+(-0.102098736751)*x[0]+(-0.858298802695)*x[1]**o+(-0.1390708698)*x[1]+(0.221129390442)*x[2]**o+(0.770148745174)*x[2]
        arg[2,1]=(0.619559141603)*x[0]**o+(-0.969165124767)*x[0]+(0.36662839455)*x[1]**o+(0.505630485999)*x[1]+(0.170716012487)*x[2]**o+(0.355457960368)*x[2]
        arg[2,2]=(0.254098998564)*x[0]**o+(0.725662782597)*x[0]+(0.274108937869)*x[1]**o+(0.682336753679)*x[1]+(-0.858192561327)*x[2]**o+(-0.538785359078)*x[2]
        arg[2,3]=(0.25513729674)*x[0]**o+(-0.594968324608)*x[0]+(-0.306612686559)*x[1]**o+(-0.953891015684)*x[1]+(-0.00631658753287)*x[2]**o+(-0.889694464702)*x[2]
        arg[2,4]=(0.654463492789)*x[0]**o+(-0.42220702047)*x[0]+(-0.690125229151)*x[1]**o+(0.318923837754)*x[1]+(-0.691341299676)*x[2]**o+(0.824631778823)*x[2]
        arg[3,0]=(-0.576563725107)*x[0]**o+(-0.0354219245402)*x[0]+(0.650854296808)*x[1]**o+(0.930072736771)*x[1]+(0.8383657117)*x[2]**o+(0.191293272146)*x[2]
        arg[3,1]=(0.104382485594)*x[0]**o+(-0.718025390561)*x[0]+(-0.822188446684)*x[1]**o+(-0.0772502664782)*x[1]+(-0.513331680922)*x[2]**o+(0.164340503698)*x[2]
        arg[3,2]=(0.685178273481)*x[0]**o+(-0.241726920352)*x[0]+(-0.891512838953)*x[1]**o+(-0.428833037662)*x[1]+(-0.910220899771)*x[2]**o+(-0.0109246344661)*x[2]
        arg[3,3]=(-0.294769406995)*x[0]**o+(0.643502505933)*x[0]+(0.309444282121)*x[1]**o+(-0.935318923262)*x[1]+(0.242426961867)*x[2]**o+(-0.417268637067)*x[2]
        arg[3,4]=(0.0382306334841)*x[0]**o+(-0.564858278904)*x[0]+(-0.594566652327)*x[1]**o+(0.911894434818)*x[1]+(-0.314178793024)*x[2]**o+(0.302730333267)*x[2]
        ref[0,0]=(-0.479138946804)*x_ref[0]**o+(0.240408443007)*x_ref[0]+(0.309155796205)*x_ref[1]**o+(-0.124360991957)*x_ref[1]+(0.00301047542037)*x_ref[2]**o+(-0.039439802799)*x_ref[2]
        ref[0,1]=(0.41979402771)*x_ref[0]**o+(0.798738902847)*x_ref[0]+(0.619960746922)*x_ref[1]**o+(-0.551662130664)*x_ref[1]+(0.914293623695)*x_ref[2]**o+(0.6859768812)*x_ref[2]
        ref[0,2]=(0.744593492696)*x_ref[0]**o+(-0.31743031745)*x_ref[0]+(0.389545123971)*x_ref[1]**o+(0.592024987288)*x_ref[1]+(0.63456095022)*x_ref[2]**o+(-0.746571796061)*x_ref[2]
        ref[0,3]=(-0.632549503649)*x_ref[0]**o+(-0.687745599598)*x_ref[0]+(0.794176042123)*x_ref[1]**o+(-0.971466375953)*x_ref[1]+(-0.086939632062)*x_ref[2]**o+(0.373736042024)*x_ref[2]
        ref[0,4]=(-0.966879538217)*x_ref[0]**o+(-0.609291970974)*x_ref[0]+(0.000409196978095)*x_ref[1]**o+(0.623934484362)*x_ref[1]+(-0.1694749999)*x_ref[2]**o+(0.131778917979)*x_ref[2]
        ref[1,0]=(0.763452444611)*x_ref[0]**o+(0.805147582411)*x_ref[0]+(0.999704124146)*x_ref[1]**o+(0.0635253753138)*x_ref[1]+(-0.717889410548)*x_ref[2]**o+(-0.840765968423)*x_ref[2]
        ref[1,1]=(-0.26910479702)*x_ref[0]**o+(0.594093239771)*x_ref[0]+(-0.536150372834)*x_ref[1]**o+(-0.768506263316)*x_ref[1]+(-0.924510063594)*x_ref[2]**o+(-0.950568839784)*x_ref[2]
        ref[1,2]=(-0.452317334239)*x_ref[0]**o+(0.186343827048)*x_ref[0]+(0.921015823436)*x_ref[1]**o+(-0.656809197675)*x_ref[1]+(-0.841898648461)*x_ref[2]**o+(0.267827136605)*x_ref[2]
        ref[1,3]=(-0.880873336233)*x_ref[0]**o+(-0.915350352959)*x_ref[0]+(0.92197459758)*x_ref[1]**o+(0.20861139921)*x_ref[1]+(0.776743964515)*x_ref[2]**o+(0.858569288507)*x_ref[2]
        ref[1,4]=(0.422773256864)*x_ref[0]**o+(0.397743860671)*x_ref[0]+(0.726470981309)*x_ref[1]**o+(-0.233534612019)*x_ref[1]+(-0.350345927201)*x_ref[2]**o+(-0.192294374344)*x_ref[2]
        ref[2,0]=(0.0706785590906)*x_ref[0]**o+(-0.102098736751)*x_ref[0]+(-0.858298802695)*x_ref[1]**o+(-0.1390708698)*x_ref[1]+(0.221129390442)*x_ref[2]**o+(0.770148745174)*x_ref[2]
        ref[2,1]=(0.619559141603)*x_ref[0]**o+(-0.969165124767)*x_ref[0]+(0.36662839455)*x_ref[1]**o+(0.505630485999)*x_ref[1]+(0.170716012487)*x_ref[2]**o+(0.355457960368)*x_ref[2]
        ref[2,2]=(0.254098998564)*x_ref[0]**o+(0.725662782597)*x_ref[0]+(0.274108937869)*x_ref[1]**o+(0.682336753679)*x_ref[1]+(-0.858192561327)*x_ref[2]**o+(-0.538785359078)*x_ref[2]
        ref[2,3]=(0.25513729674)*x_ref[0]**o+(-0.594968324608)*x_ref[0]+(-0.306612686559)*x_ref[1]**o+(-0.953891015684)*x_ref[1]+(-0.00631658753287)*x_ref[2]**o+(-0.889694464702)*x_ref[2]
        ref[2,4]=(0.654463492789)*x_ref[0]**o+(-0.42220702047)*x_ref[0]+(-0.690125229151)*x_ref[1]**o+(0.318923837754)*x_ref[1]+(-0.691341299676)*x_ref[2]**o+(0.824631778823)*x_ref[2]
        ref[3,0]=(-0.576563725107)*x_ref[0]**o+(-0.0354219245402)*x_ref[0]+(0.650854296808)*x_ref[1]**o+(0.930072736771)*x_ref[1]+(0.8383657117)*x_ref[2]**o+(0.191293272146)*x_ref[2]
        ref[3,1]=(0.104382485594)*x_ref[0]**o+(-0.718025390561)*x_ref[0]+(-0.822188446684)*x_ref[1]**o+(-0.0772502664782)*x_ref[1]+(-0.513331680922)*x_ref[2]**o+(0.164340503698)*x_ref[2]
        ref[3,2]=(0.685178273481)*x_ref[0]**o+(-0.241726920352)*x_ref[0]+(-0.891512838953)*x_ref[1]**o+(-0.428833037662)*x_ref[1]+(-0.910220899771)*x_ref[2]**o+(-0.0109246344661)*x_ref[2]
        ref[3,3]=(-0.294769406995)*x_ref[0]**o+(0.643502505933)*x_ref[0]+(0.309444282121)*x_ref[1]**o+(-0.935318923262)*x_ref[1]+(0.242426961867)*x_ref[2]**o+(-0.417268637067)*x_ref[2]
        ref[3,4]=(0.0382306334841)*x_ref[0]**o+(-0.564858278904)*x_ref[0]+(-0.594566652327)*x_ref[1]**o+(0.911894434818)*x_ref[1]+(-0.314178793024)*x_ref[2]**o+(0.302730333267)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.298635657968)*x[0]**o+(0.917751825722)*x[0]+(-0.000157823722007)*x[1]**o+(-0.202721972906)*x[1]
        arg[0,0,1]=(0.164880328673)*x[0]**o+(-0.310233508297)*x[0]+(-0.976270095138)*x[1]**o+(-0.86689594754)*x[1]
        arg[0,1,0]=(-0.130835262631)*x[0]**o+(-0.485555484157)*x[0]+(0.055655815912)*x[1]**o+(-0.488201458657)*x[1]
        arg[0,1,1]=(-0.339982923518)*x[0]**o+(0.435464515598)*x[0]+(-0.986955506076)*x[1]**o+(0.590966000031)*x[1]
        arg[1,0,0]=(0.150280127463)*x[0]**o+(0.827183876844)*x[0]+(0.468401701763)*x[1]**o+(-0.879871376091)*x[1]
        arg[1,0,1]=(0.48434831025)*x[0]**o+(-0.623958069024)*x[0]+(-0.781911271849)*x[1]**o+(0.34398679645)*x[1]
        arg[1,1,0]=(0.124273770782)*x[0]**o+(-0.683096046651)*x[0]+(0.769636378285)*x[1]**o+(-0.471970832437)*x[1]
        arg[1,1,1]=(0.769777912195)*x[0]**o+(-0.740802020319)*x[0]+(0.659109761568)*x[1]**o+(-0.495444655614)*x[1]
        arg[2,0,0]=(0.270063777356)*x[0]**o+(0.558849439723)*x[0]+(0.128946408169)*x[1]**o+(-0.962495994445)*x[1]
        arg[2,0,1]=(-0.0403460883009)*x[0]**o+(0.084983334662)*x[0]+(0.297573700577)*x[1]**o+(0.911774703678)*x[1]
        arg[2,1,0]=(0.821680378088)*x[0]**o+(0.10823172035)*x[0]+(-0.150213295203)*x[1]**o+(0.310471151961)*x[1]
        arg[2,1,1]=(-0.407033195691)*x[0]**o+(0.716262964777)*x[0]+(-0.262475432397)*x[1]**o+(-0.234158898682)*x[1]
        arg[3,0,0]=(-0.109193787139)*x[0]**o+(0.403892847875)*x[0]+(0.794082322801)*x[1]**o+(0.669070572253)*x[1]
        arg[3,0,1]=(0.969770217228)*x[0]**o+(0.735280190411)*x[0]+(-0.339891263613)*x[1]**o+(0.859989056027)*x[1]
        arg[3,1,0]=(-0.906193851712)*x[0]**o+(-0.330178132059)*x[0]+(-0.48035114847)*x[1]**o+(-0.48050067188)*x[1]
        arg[3,1,1]=(0.922778761752)*x[0]**o+(0.899145572602)*x[0]+(-0.991922151176)*x[1]**o+(0.615744852656)*x[1]
        arg[4,0,0]=(0.457092295352)*x[0]**o+(-0.120904430848)*x[0]+(0.575186782188)*x[1]**o+(-0.685522788659)*x[1]
        arg[4,0,1]=(0.254934861907)*x[0]**o+(-0.208850206761)*x[0]+(-0.0933987467304)*x[1]**o+(0.687715419212)*x[1]
        arg[4,1,0]=(0.483890412252)*x[0]**o+(-0.192055355216)*x[0]+(0.0891595367214)*x[1]**o+(-0.835900121805)*x[1]
        arg[4,1,1]=(-0.281109655935)*x[0]**o+(0.279867178407)*x[0]+(0.843094337157)*x[1]**o+(-0.457583293166)*x[1]
        arg[5,0,0]=(0.595078633674)*x[0]**o+(-0.994652086537)*x[0]+(0.111354924733)*x[1]**o+(0.715580298167)*x[1]
        arg[5,0,1]=(0.823409068766)*x[0]**o+(0.41047089933)*x[0]+(-0.610763277348)*x[1]**o+(0.0991108428868)*x[1]
        arg[5,1,0]=(-0.0291539407811)*x[0]**o+(0.371799932333)*x[0]+(0.631702366805)*x[1]**o+(-0.858415923309)*x[1]
        arg[5,1,1]=(-0.636146064991)*x[0]**o+(0.407248062196)*x[0]+(-0.507069143933)*x[1]**o+(-0.0509951705166)*x[1]
        ref[0,0,0]=(0.298635657968)*x_ref[0]**o+(0.917751825722)*x_ref[0]+(-0.000157823722007)*x_ref[1]**o+(-0.202721972906)*x_ref[1]
        ref[0,0,1]=(0.164880328673)*x_ref[0]**o+(-0.310233508297)*x_ref[0]+(-0.976270095138)*x_ref[1]**o+(-0.86689594754)*x_ref[1]
        ref[0,1,0]=(-0.130835262631)*x_ref[0]**o+(-0.485555484157)*x_ref[0]+(0.055655815912)*x_ref[1]**o+(-0.488201458657)*x_ref[1]
        ref[0,1,1]=(-0.339982923518)*x_ref[0]**o+(0.435464515598)*x_ref[0]+(-0.986955506076)*x_ref[1]**o+(0.590966000031)*x_ref[1]
        ref[1,0,0]=(0.150280127463)*x_ref[0]**o+(0.827183876844)*x_ref[0]+(0.468401701763)*x_ref[1]**o+(-0.879871376091)*x_ref[1]
        ref[1,0,1]=(0.48434831025)*x_ref[0]**o+(-0.623958069024)*x_ref[0]+(-0.781911271849)*x_ref[1]**o+(0.34398679645)*x_ref[1]
        ref[1,1,0]=(0.124273770782)*x_ref[0]**o+(-0.683096046651)*x_ref[0]+(0.769636378285)*x_ref[1]**o+(-0.471970832437)*x_ref[1]
        ref[1,1,1]=(0.769777912195)*x_ref[0]**o+(-0.740802020319)*x_ref[0]+(0.659109761568)*x_ref[1]**o+(-0.495444655614)*x_ref[1]
        ref[2,0,0]=(0.270063777356)*x_ref[0]**o+(0.558849439723)*x_ref[0]+(0.128946408169)*x_ref[1]**o+(-0.962495994445)*x_ref[1]
        ref[2,0,1]=(-0.0403460883009)*x_ref[0]**o+(0.084983334662)*x_ref[0]+(0.297573700577)*x_ref[1]**o+(0.911774703678)*x_ref[1]
        ref[2,1,0]=(0.821680378088)*x_ref[0]**o+(0.10823172035)*x_ref[0]+(-0.150213295203)*x_ref[1]**o+(0.310471151961)*x_ref[1]
        ref[2,1,1]=(-0.407033195691)*x_ref[0]**o+(0.716262964777)*x_ref[0]+(-0.262475432397)*x_ref[1]**o+(-0.234158898682)*x_ref[1]
        ref[3,0,0]=(-0.109193787139)*x_ref[0]**o+(0.403892847875)*x_ref[0]+(0.794082322801)*x_ref[1]**o+(0.669070572253)*x_ref[1]
        ref[3,0,1]=(0.969770217228)*x_ref[0]**o+(0.735280190411)*x_ref[0]+(-0.339891263613)*x_ref[1]**o+(0.859989056027)*x_ref[1]
        ref[3,1,0]=(-0.906193851712)*x_ref[0]**o+(-0.330178132059)*x_ref[0]+(-0.48035114847)*x_ref[1]**o+(-0.48050067188)*x_ref[1]
        ref[3,1,1]=(0.922778761752)*x_ref[0]**o+(0.899145572602)*x_ref[0]+(-0.991922151176)*x_ref[1]**o+(0.615744852656)*x_ref[1]
        ref[4,0,0]=(0.457092295352)*x_ref[0]**o+(-0.120904430848)*x_ref[0]+(0.575186782188)*x_ref[1]**o+(-0.685522788659)*x_ref[1]
        ref[4,0,1]=(0.254934861907)*x_ref[0]**o+(-0.208850206761)*x_ref[0]+(-0.0933987467304)*x_ref[1]**o+(0.687715419212)*x_ref[1]
        ref[4,1,0]=(0.483890412252)*x_ref[0]**o+(-0.192055355216)*x_ref[0]+(0.0891595367214)*x_ref[1]**o+(-0.835900121805)*x_ref[1]
        ref[4,1,1]=(-0.281109655935)*x_ref[0]**o+(0.279867178407)*x_ref[0]+(0.843094337157)*x_ref[1]**o+(-0.457583293166)*x_ref[1]
        ref[5,0,0]=(0.595078633674)*x_ref[0]**o+(-0.994652086537)*x_ref[0]+(0.111354924733)*x_ref[1]**o+(0.715580298167)*x_ref[1]
        ref[5,0,1]=(0.823409068766)*x_ref[0]**o+(0.41047089933)*x_ref[0]+(-0.610763277348)*x_ref[1]**o+(0.0991108428868)*x_ref[1]
        ref[5,1,0]=(-0.0291539407811)*x_ref[0]**o+(0.371799932333)*x_ref[0]+(0.631702366805)*x_ref[1]**o+(-0.858415923309)*x_ref[1]
        ref[5,1,1]=(-0.636146064991)*x_ref[0]**o+(0.407248062196)*x_ref[0]+(-0.507069143933)*x_ref[1]**o+(-0.0509951705166)*x_ref[1]
      else:
        arg[0,0,0]=(-0.662043178282)*x[0]**o+(0.600130780119)*x[0]+(-0.52233323628)*x[1]**o+(-0.463302513654)*x[1]+(0.714996279403)*x[2]**o+(0.113219137203)*x[2]
        arg[0,0,1]=(0.94317347224)*x[0]**o+(0.987584533548)*x[0]+(0.116577593127)*x[1]**o+(-0.515267022711)*x[1]+(0.856578953903)*x[2]**o+(0.70372338917)*x[2]
        arg[0,1,0]=(-0.405265998966)*x[0]**o+(-0.656404120416)*x[0]+(0.270655604701)*x[1]**o+(0.0638534875176)*x[1]+(0.816834973788)*x[2]**o+(0.918436000225)*x[2]
        arg[0,1,1]=(-0.37299950332)*x[0]**o+(-0.0367737761004)*x[0]+(0.176531372245)*x[1]**o+(0.849773143988)*x[1]+(-0.572911529869)*x[2]**o+(-0.140163625628)*x[2]
        arg[1,0,0]=(0.263616091516)*x[0]**o+(-0.65550051998)*x[0]+(-0.118310211881)*x[1]**o+(-0.695263983587)*x[1]+(-0.49011649182)*x[2]**o+(-0.553284002669)*x[2]
        arg[1,0,1]=(0.859640205193)*x[0]**o+(-0.569681205214)*x[0]+(0.459979003108)*x[1]**o+(0.672561515814)*x[1]+(-0.854289983441)*x[2]**o+(-0.445344129294)*x[2]
        arg[1,1,0]=(-0.904355809101)*x[0]**o+(0.540513356972)*x[0]+(-0.377422260181)*x[1]**o+(0.937285151058)*x[1]+(0.148943597435)*x[2]**o+(0.992256197076)*x[2]
        arg[1,1,1]=(-0.868336238952)*x[0]**o+(-0.713337888363)*x[0]+(0.800427728006)*x[1]**o+(-0.83502690568)*x[1]+(0.365707922084)*x[2]**o+(0.218740577318)*x[2]
        arg[2,0,0]=(-0.0382672594668)*x[0]**o+(0.463831755073)*x[0]+(-0.104957678792)*x[1]**o+(-0.561738607653)*x[1]+(-0.660128314406)*x[2]**o+(0.594649133189)*x[2]
        arg[2,0,1]=(-0.399297921751)*x[0]**o+(0.550876554224)*x[0]+(0.150599666563)*x[1]**o+(-0.748400310268)*x[1]+(0.658318134651)*x[2]**o+(0.49075091442)*x[2]
        arg[2,1,0]=(-0.393769876286)*x[0]**o+(0.328125610064)*x[0]+(-0.703385111404)*x[1]**o+(0.0418077854723)*x[1]+(0.918774663106)*x[2]**o+(-0.119085951508)*x[2]
        arg[2,1,1]=(-0.549250924758)*x[0]**o+(0.169148558793)*x[0]+(0.664162931811)*x[1]**o+(0.207136816188)*x[1]+(-0.17207817969)*x[2]**o+(0.34258020774)*x[2]
        arg[3,0,0]=(0.446467472517)*x[0]**o+(-0.358128380843)*x[0]+(0.146916048359)*x[1]**o+(-0.606692124285)*x[1]+(0.828655704727)*x[2]**o+(0.067139299396)*x[2]
        arg[3,0,1]=(-0.357052242351)*x[0]**o+(-0.528255164843)*x[0]+(0.140204673142)*x[1]**o+(0.674282235672)*x[1]+(-0.469118454039)*x[2]**o+(0.388103121466)*x[2]
        arg[3,1,0]=(-0.359629791333)*x[0]**o+(-0.625108929158)*x[0]+(0.855034791748)*x[1]**o+(-0.343451063341)*x[1]+(-0.681557972105)*x[2]**o+(0.576310870238)*x[2]
        arg[3,1,1]=(0.00546530278132)*x[0]**o+(0.359221089344)*x[0]+(-0.57416247885)*x[1]**o+(-0.521306138323)*x[1]+(-0.618281954095)*x[2]**o+(-0.862162207352)*x[2]
        arg[4,0,0]=(-0.196356423488)*x[0]**o+(-0.871482508533)*x[0]+(-0.367390886081)*x[1]**o+(0.580310108591)*x[1]+(0.857970671482)*x[2]**o+(0.123685673359)*x[2]
        arg[4,0,1]=(0.715104674022)*x[0]**o+(-0.00444438628932)*x[0]+(0.538883210054)*x[1]**o+(0.179340705463)*x[1]+(-0.605249436909)*x[2]**o+(-0.744253826562)*x[2]
        arg[4,1,0]=(-0.315453893647)*x[0]**o+(-0.560946546082)*x[0]+(-0.674448928458)*x[1]**o+(-0.510243444449)*x[1]+(0.592615876779)*x[2]**o+(0.218024801263)*x[2]
        arg[4,1,1]=(-0.46902473326)*x[0]**o+(-0.573926691452)*x[0]+(0.827898114814)*x[1]**o+(-0.644208082857)*x[1]+(-0.956169254426)*x[2]**o+(-0.297101737561)*x[2]
        arg[5,0,0]=(0.45818164662)*x[0]**o+(0.219667021464)*x[0]+(-0.453125753418)*x[1]**o+(-0.201214929832)*x[1]+(0.135363067088)*x[2]**o+(0.850636247323)*x[2]
        arg[5,0,1]=(0.296120304926)*x[0]**o+(-0.139971296351)*x[0]+(0.448115780401)*x[1]**o+(-0.300602343399)*x[1]+(-0.165568753274)*x[2]**o+(0.984788641639)*x[2]
        arg[5,1,0]=(0.321875942152)*x[0]**o+(-0.403122417353)*x[0]+(-0.148664639927)*x[1]**o+(0.644143662303)*x[1]+(-0.718311175542)*x[2]**o+(-0.850248996913)*x[2]
        arg[5,1,1]=(0.44869757406)*x[0]**o+(0.0292094503086)*x[0]+(-0.159644215279)*x[1]**o+(0.695976060359)*x[1]+(-0.221444458713)*x[2]**o+(-0.42974111206)*x[2]
        ref[0,0,0]=(-0.662043178282)*x_ref[0]**o+(0.600130780119)*x_ref[0]+(-0.52233323628)*x_ref[1]**o+(-0.463302513654)*x_ref[1]+(0.714996279403)*x_ref[2]**o+(0.113219137203)*x_ref[2]
        ref[0,0,1]=(0.94317347224)*x_ref[0]**o+(0.987584533548)*x_ref[0]+(0.116577593127)*x_ref[1]**o+(-0.515267022711)*x_ref[1]+(0.856578953903)*x_ref[2]**o+(0.70372338917)*x_ref[2]
        ref[0,1,0]=(-0.405265998966)*x_ref[0]**o+(-0.656404120416)*x_ref[0]+(0.270655604701)*x_ref[1]**o+(0.0638534875176)*x_ref[1]+(0.816834973788)*x_ref[2]**o+(0.918436000225)*x_ref[2]
        ref[0,1,1]=(-0.37299950332)*x_ref[0]**o+(-0.0367737761004)*x_ref[0]+(0.176531372245)*x_ref[1]**o+(0.849773143988)*x_ref[1]+(-0.572911529869)*x_ref[2]**o+(-0.140163625628)*x_ref[2]
        ref[1,0,0]=(0.263616091516)*x_ref[0]**o+(-0.65550051998)*x_ref[0]+(-0.118310211881)*x_ref[1]**o+(-0.695263983587)*x_ref[1]+(-0.49011649182)*x_ref[2]**o+(-0.553284002669)*x_ref[2]
        ref[1,0,1]=(0.859640205193)*x_ref[0]**o+(-0.569681205214)*x_ref[0]+(0.459979003108)*x_ref[1]**o+(0.672561515814)*x_ref[1]+(-0.854289983441)*x_ref[2]**o+(-0.445344129294)*x_ref[2]
        ref[1,1,0]=(-0.904355809101)*x_ref[0]**o+(0.540513356972)*x_ref[0]+(-0.377422260181)*x_ref[1]**o+(0.937285151058)*x_ref[1]+(0.148943597435)*x_ref[2]**o+(0.992256197076)*x_ref[2]
        ref[1,1,1]=(-0.868336238952)*x_ref[0]**o+(-0.713337888363)*x_ref[0]+(0.800427728006)*x_ref[1]**o+(-0.83502690568)*x_ref[1]+(0.365707922084)*x_ref[2]**o+(0.218740577318)*x_ref[2]
        ref[2,0,0]=(-0.0382672594668)*x_ref[0]**o+(0.463831755073)*x_ref[0]+(-0.104957678792)*x_ref[1]**o+(-0.561738607653)*x_ref[1]+(-0.660128314406)*x_ref[2]**o+(0.594649133189)*x_ref[2]
        ref[2,0,1]=(-0.399297921751)*x_ref[0]**o+(0.550876554224)*x_ref[0]+(0.150599666563)*x_ref[1]**o+(-0.748400310268)*x_ref[1]+(0.658318134651)*x_ref[2]**o+(0.49075091442)*x_ref[2]
        ref[2,1,0]=(-0.393769876286)*x_ref[0]**o+(0.328125610064)*x_ref[0]+(-0.703385111404)*x_ref[1]**o+(0.0418077854723)*x_ref[1]+(0.918774663106)*x_ref[2]**o+(-0.119085951508)*x_ref[2]
        ref[2,1,1]=(-0.549250924758)*x_ref[0]**o+(0.169148558793)*x_ref[0]+(0.664162931811)*x_ref[1]**o+(0.207136816188)*x_ref[1]+(-0.17207817969)*x_ref[2]**o+(0.34258020774)*x_ref[2]
        ref[3,0,0]=(0.446467472517)*x_ref[0]**o+(-0.358128380843)*x_ref[0]+(0.146916048359)*x_ref[1]**o+(-0.606692124285)*x_ref[1]+(0.828655704727)*x_ref[2]**o+(0.067139299396)*x_ref[2]
        ref[3,0,1]=(-0.357052242351)*x_ref[0]**o+(-0.528255164843)*x_ref[0]+(0.140204673142)*x_ref[1]**o+(0.674282235672)*x_ref[1]+(-0.469118454039)*x_ref[2]**o+(0.388103121466)*x_ref[2]
        ref[3,1,0]=(-0.359629791333)*x_ref[0]**o+(-0.625108929158)*x_ref[0]+(0.855034791748)*x_ref[1]**o+(-0.343451063341)*x_ref[1]+(-0.681557972105)*x_ref[2]**o+(0.576310870238)*x_ref[2]
        ref[3,1,1]=(0.00546530278132)*x_ref[0]**o+(0.359221089344)*x_ref[0]+(-0.57416247885)*x_ref[1]**o+(-0.521306138323)*x_ref[1]+(-0.618281954095)*x_ref[2]**o+(-0.862162207352)*x_ref[2]
        ref[4,0,0]=(-0.196356423488)*x_ref[0]**o+(-0.871482508533)*x_ref[0]+(-0.367390886081)*x_ref[1]**o+(0.580310108591)*x_ref[1]+(0.857970671482)*x_ref[2]**o+(0.123685673359)*x_ref[2]
        ref[4,0,1]=(0.715104674022)*x_ref[0]**o+(-0.00444438628932)*x_ref[0]+(0.538883210054)*x_ref[1]**o+(0.179340705463)*x_ref[1]+(-0.605249436909)*x_ref[2]**o+(-0.744253826562)*x_ref[2]
        ref[4,1,0]=(-0.315453893647)*x_ref[0]**o+(-0.560946546082)*x_ref[0]+(-0.674448928458)*x_ref[1]**o+(-0.510243444449)*x_ref[1]+(0.592615876779)*x_ref[2]**o+(0.218024801263)*x_ref[2]
        ref[4,1,1]=(-0.46902473326)*x_ref[0]**o+(-0.573926691452)*x_ref[0]+(0.827898114814)*x_ref[1]**o+(-0.644208082857)*x_ref[1]+(-0.956169254426)*x_ref[2]**o+(-0.297101737561)*x_ref[2]
        ref[5,0,0]=(0.45818164662)*x_ref[0]**o+(0.219667021464)*x_ref[0]+(-0.453125753418)*x_ref[1]**o+(-0.201214929832)*x_ref[1]+(0.135363067088)*x_ref[2]**o+(0.850636247323)*x_ref[2]
        ref[5,0,1]=(0.296120304926)*x_ref[0]**o+(-0.139971296351)*x_ref[0]+(0.448115780401)*x_ref[1]**o+(-0.300602343399)*x_ref[1]+(-0.165568753274)*x_ref[2]**o+(0.984788641639)*x_ref[2]
        ref[5,1,0]=(0.321875942152)*x_ref[0]**o+(-0.403122417353)*x_ref[0]+(-0.148664639927)*x_ref[1]**o+(0.644143662303)*x_ref[1]+(-0.718311175542)*x_ref[2]**o+(-0.850248996913)*x_ref[2]
        ref[5,1,1]=(0.44869757406)*x_ref[0]**o+(0.0292094503086)*x_ref[0]+(-0.159644215279)*x_ref[1]**o+(0.695976060359)*x_ref[1]+(-0.221444458713)*x_ref[2]**o+(-0.42974111206)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0737428272134)*x[0]**o+(0.841780036645)*x[0]+(-0.110824257531)*x[1]**o+(-0.957726444468)*x[1]
        arg[0,0,0,1]=(-0.934840410671)*x[0]**o+(-0.83301064321)*x[0]+(-0.352287650266)*x[1]**o+(-0.961111564077)*x[1]
        arg[0,0,1,0]=(-0.829500940226)*x[0]**o+(0.644171710931)*x[0]+(0.232545739722)*x[1]**o+(0.584336012862)*x[1]
        arg[0,0,1,1]=(0.426298830976)*x[0]**o+(-0.615281197268)*x[0]+(0.206334026192)*x[1]**o+(-0.443698947044)*x[1]
        arg[0,0,2,0]=(-0.236495891617)*x[0]**o+(0.0883376851179)*x[0]+(-0.176541434356)*x[1]**o+(0.297175301617)*x[1]
        arg[0,0,2,1]=(-0.879955324373)*x[0]**o+(-0.244778913771)*x[0]+(0.244131188594)*x[1]**o+(-0.334739112198)*x[1]
        arg[0,1,0,0]=(-0.64887128006)*x[0]**o+(-0.489951454377)*x[0]+(-0.328402433668)*x[1]**o+(0.229814237117)*x[1]
        arg[0,1,0,1]=(-0.339451656097)*x[0]**o+(-0.409320682304)*x[0]+(0.216482338677)*x[1]**o+(-0.10594373055)*x[1]
        arg[0,1,1,0]=(-0.350538479425)*x[0]**o+(0.466450046985)*x[0]+(-0.779124795292)*x[1]**o+(-0.654562596704)*x[1]
        arg[0,1,1,1]=(0.00350092879868)*x[0]**o+(-0.214643088647)*x[0]+(0.786934081411)*x[1]**o+(-0.296853380043)*x[1]
        arg[0,1,2,0]=(-0.640183955648)*x[0]**o+(0.568647968237)*x[0]+(0.575533476901)*x[1]**o+(0.854540594085)*x[1]
        arg[0,1,2,1]=(0.633674899518)*x[0]**o+(0.829255679443)*x[0]+(0.537888424303)*x[1]**o+(-0.495082587617)*x[1]
        arg[0,2,0,0]=(-0.700436798342)*x[0]**o+(-0.732581859026)*x[0]+(-0.508905353053)*x[1]**o+(0.664601151676)*x[1]
        arg[0,2,0,1]=(0.720335764302)*x[0]**o+(0.941159917988)*x[0]+(0.2817440547)*x[1]**o+(0.553712160058)*x[1]
        arg[0,2,1,0]=(-0.937114312262)*x[0]**o+(-0.115517396201)*x[0]+(-0.672697125905)*x[1]**o+(-0.221320095369)*x[1]
        arg[0,2,1,1]=(-0.52059341062)*x[0]**o+(-0.653834647475)*x[0]+(-0.0851100861135)*x[1]**o+(-0.668300210067)*x[1]
        arg[0,2,2,0]=(-0.943613035509)*x[0]**o+(-0.294843145822)*x[0]+(0.496278748831)*x[1]**o+(0.901154419425)*x[1]
        arg[0,2,2,1]=(-0.421698818203)*x[0]**o+(-0.721918177544)*x[0]+(-0.197855702215)*x[1]**o+(0.50882827715)*x[1]
        arg[0,3,0,0]=(0.931990780298)*x[0]**o+(-0.550332606951)*x[0]+(0.401584654642)*x[1]**o+(0.0965259747601)*x[1]
        arg[0,3,0,1]=(0.881357708445)*x[0]**o+(0.564624034197)*x[0]+(-0.458191897924)*x[1]**o+(-0.427701112807)*x[1]
        arg[0,3,1,0]=(0.624777088081)*x[0]**o+(0.862654721208)*x[0]+(-0.707618093268)*x[1]**o+(0.743304953313)*x[1]
        arg[0,3,1,1]=(0.4510836974)*x[0]**o+(-0.357023031453)*x[0]+(-0.00503924010273)*x[1]**o+(0.545024336011)*x[1]
        arg[0,3,2,0]=(0.109003851094)*x[0]**o+(0.168659376881)*x[0]+(-0.0578953780655)*x[1]**o+(-0.481212462712)*x[1]
        arg[0,3,2,1]=(-0.191183501636)*x[0]**o+(-0.316145516377)*x[0]+(0.402460103016)*x[1]**o+(-0.163612464438)*x[1]
        arg[0,4,0,0]=(0.489815345033)*x[0]**o+(0.999138280765)*x[0]+(-0.626627766812)*x[1]**o+(-0.339653992568)*x[1]
        arg[0,4,0,1]=(0.616350228791)*x[0]**o+(0.933446438765)*x[0]+(0.944244035512)*x[1]**o+(0.829407789353)*x[1]
        arg[0,4,1,0]=(0.619918547863)*x[0]**o+(0.995488719729)*x[0]+(-0.78847429581)*x[1]**o+(0.542645674462)*x[1]
        arg[0,4,1,1]=(-0.149244819992)*x[0]**o+(0.921766299925)*x[0]+(0.435255339567)*x[1]**o+(-0.0596561300741)*x[1]
        arg[0,4,2,0]=(-0.816364336598)*x[0]**o+(0.701160246658)*x[0]+(-0.616385505652)*x[1]**o+(-0.558625482105)*x[1]
        arg[0,4,2,1]=(-0.36944364693)*x[0]**o+(0.243426977623)*x[0]+(-0.61578435391)*x[1]**o+(0.932570326139)*x[1]
        arg[1,0,0,0]=(-0.36839962807)*x[0]**o+(0.590534439534)*x[0]+(-0.615158899289)*x[1]**o+(0.973649158296)*x[1]
        arg[1,0,0,1]=(0.761515772771)*x[0]**o+(0.706406307012)*x[0]+(0.328712319043)*x[1]**o+(0.438600140517)*x[1]
        arg[1,0,1,0]=(0.895080579207)*x[0]**o+(-0.166789328225)*x[0]+(0.807540897371)*x[1]**o+(0.0345522692723)*x[1]
        arg[1,0,1,1]=(-0.460062661172)*x[0]**o+(0.505167337562)*x[0]+(0.25155527144)*x[1]**o+(0.601093887451)*x[1]
        arg[1,0,2,0]=(0.382129818527)*x[0]**o+(0.694676536542)*x[0]+(0.702794119083)*x[1]**o+(0.704094013056)*x[1]
        arg[1,0,2,1]=(0.265476982913)*x[0]**o+(-0.134338744881)*x[0]+(-0.0938910335356)*x[1]**o+(-0.560073316097)*x[1]
        arg[1,1,0,0]=(0.977368328792)*x[0]**o+(0.084482890741)*x[0]+(0.57068389287)*x[1]**o+(-0.242087071526)*x[1]
        arg[1,1,0,1]=(-0.36067775897)*x[0]**o+(0.768187470286)*x[0]+(-0.881150642619)*x[1]**o+(-0.15839498619)*x[1]
        arg[1,1,1,0]=(-0.808263913645)*x[0]**o+(0.788403321664)*x[0]+(0.976206012343)*x[1]**o+(-0.636933353579)*x[1]
        arg[1,1,1,1]=(0.699185445546)*x[0]**o+(0.251811206876)*x[0]+(-0.146320762812)*x[1]**o+(-0.248033048423)*x[1]
        arg[1,1,2,0]=(-0.782940880812)*x[0]**o+(0.789401686528)*x[0]+(-0.806883904562)*x[1]**o+(0.83689696772)*x[1]
        arg[1,1,2,1]=(0.652404993342)*x[0]**o+(-0.135318598212)*x[0]+(0.138928760301)*x[1]**o+(0.675593554805)*x[1]
        arg[1,2,0,0]=(0.471746379585)*x[0]**o+(-0.989783415847)*x[0]+(-0.0935052296418)*x[1]**o+(-0.532245102898)*x[1]
        arg[1,2,0,1]=(-0.44541059979)*x[0]**o+(0.593734266824)*x[0]+(-0.616647011751)*x[1]**o+(0.461140297024)*x[1]
        arg[1,2,1,0]=(-0.9144210459)*x[0]**o+(0.116806032197)*x[0]+(0.935610395306)*x[1]**o+(0.550884822213)*x[1]
        arg[1,2,1,1]=(-0.962444455944)*x[0]**o+(-0.654328013652)*x[0]+(-0.545330994236)*x[1]**o+(-0.919705456162)*x[1]
        arg[1,2,2,0]=(-0.724491374462)*x[0]**o+(-0.820402375498)*x[0]+(-0.953956431336)*x[1]**o+(-0.201450523182)*x[1]
        arg[1,2,2,1]=(0.723637497881)*x[0]**o+(-0.144483238523)*x[0]+(0.00379745988102)*x[1]**o+(-0.838616087543)*x[1]
        arg[1,3,0,0]=(-0.962256046042)*x[0]**o+(-0.453580322779)*x[0]+(0.887446007424)*x[1]**o+(-0.404454768131)*x[1]
        arg[1,3,0,1]=(0.829189872775)*x[0]**o+(0.327985850399)*x[0]+(0.113600094307)*x[1]**o+(-0.338775377551)*x[1]
        arg[1,3,1,0]=(0.912320027833)*x[0]**o+(0.0658152563951)*x[0]+(-0.843491517568)*x[1]**o+(0.858344434434)*x[1]
        arg[1,3,1,1]=(-0.133918678317)*x[0]**o+(0.644272729149)*x[0]+(0.140663927965)*x[1]**o+(0.853385355163)*x[1]
        arg[1,3,2,0]=(0.37733357322)*x[0]**o+(0.255105934063)*x[0]+(0.88881892183)*x[1]**o+(-0.722661948279)*x[1]
        arg[1,3,2,1]=(-0.2827466345)*x[0]**o+(-0.518815029512)*x[0]+(-0.995043824929)*x[1]**o+(0.973473354481)*x[1]
        arg[1,4,0,0]=(0.044904732821)*x[0]**o+(-0.0373604593083)*x[0]+(0.184254626748)*x[1]**o+(0.031832590233)*x[1]
        arg[1,4,0,1]=(0.388699053911)*x[0]**o+(-0.0619606394414)*x[0]+(0.477947819745)*x[1]**o+(0.982909216221)*x[1]
        arg[1,4,1,0]=(0.854348679973)*x[0]**o+(-0.822052612379)*x[0]+(-0.950305089968)*x[1]**o+(0.0100277256384)*x[1]
        arg[1,4,1,1]=(0.265075277179)*x[0]**o+(0.489846993407)*x[0]+(-0.961111499722)*x[1]**o+(-0.0457909055602)*x[1]
        arg[1,4,2,0]=(0.701951191951)*x[0]**o+(0.675232620259)*x[0]+(-0.141848599227)*x[1]**o+(0.737150257564)*x[1]
        arg[1,4,2,1]=(-0.450684564415)*x[0]**o+(0.854849922021)*x[0]+(0.729670746723)*x[1]**o+(0.530372628879)*x[1]
        arg[2,0,0,0]=(-0.131153262675)*x[0]**o+(-0.477683407979)*x[0]+(0.231416752516)*x[1]**o+(0.790897064402)*x[1]
        arg[2,0,0,1]=(0.0956557004382)*x[0]**o+(0.209084913471)*x[0]+(-0.328544695733)*x[1]**o+(0.701245770328)*x[1]
        arg[2,0,1,0]=(-0.476403718361)*x[0]**o+(0.689023533158)*x[0]+(0.108732189071)*x[1]**o+(0.243379977453)*x[1]
        arg[2,0,1,1]=(0.414917801157)*x[0]**o+(-0.843544596785)*x[0]+(0.772236792935)*x[1]**o+(0.836028831232)*x[1]
        arg[2,0,2,0]=(0.754970600617)*x[0]**o+(-0.422318891122)*x[0]+(0.770767911534)*x[1]**o+(-0.446737894095)*x[1]
        arg[2,0,2,1]=(0.0139989331471)*x[0]**o+(0.262286042291)*x[0]+(-0.440109329114)*x[1]**o+(-0.844965114722)*x[1]
        arg[2,1,0,0]=(-0.758977153009)*x[0]**o+(-0.991192011039)*x[0]+(0.642267239475)*x[1]**o+(-0.701360491163)*x[1]
        arg[2,1,0,1]=(-0.123019007509)*x[0]**o+(-0.622323252979)*x[0]+(0.441364031664)*x[1]**o+(0.311433222821)*x[1]
        arg[2,1,1,0]=(-0.0446644182928)*x[0]**o+(0.614952704493)*x[0]+(-0.228284494629)*x[1]**o+(0.077603612991)*x[1]
        arg[2,1,1,1]=(0.712924108225)*x[0]**o+(-0.153966300537)*x[0]+(-0.584450014127)*x[1]**o+(0.476329149833)*x[1]
        arg[2,1,2,0]=(0.554026516037)*x[0]**o+(0.730575821847)*x[0]+(-0.247529915107)*x[1]**o+(0.210060454223)*x[1]
        arg[2,1,2,1]=(0.877017208752)*x[0]**o+(-0.800916200997)*x[0]+(0.0996710064295)*x[1]**o+(-0.0890375735808)*x[1]
        arg[2,2,0,0]=(-0.198026186062)*x[0]**o+(-0.416160224681)*x[0]+(-0.626491103997)*x[1]**o+(-0.950177752043)*x[1]
        arg[2,2,0,1]=(-0.401319234479)*x[0]**o+(0.948120359627)*x[0]+(0.420351950404)*x[1]**o+(-0.242850312594)*x[1]
        arg[2,2,1,0]=(-0.239652069063)*x[0]**o+(-0.894811076801)*x[0]+(0.979721413509)*x[1]**o+(-0.712493893566)*x[1]
        arg[2,2,1,1]=(0.239416494209)*x[0]**o+(0.449047649538)*x[0]+(0.0503751073583)*x[1]**o+(0.739949331521)*x[1]
        arg[2,2,2,0]=(-0.318067701034)*x[0]**o+(-0.00208209268356)*x[0]+(-0.513821624843)*x[1]**o+(-0.202688815501)*x[1]
        arg[2,2,2,1]=(-0.194111524544)*x[0]**o+(-0.843213241929)*x[0]+(0.124223747087)*x[1]**o+(-0.988305059412)*x[1]
        arg[2,3,0,0]=(0.450894655873)*x[0]**o+(-0.809351657482)*x[0]+(-0.80555128185)*x[1]**o+(-0.922123806508)*x[1]
        arg[2,3,0,1]=(-0.462822602909)*x[0]**o+(0.504388766027)*x[0]+(-0.152725064915)*x[1]**o+(0.988745142638)*x[1]
        arg[2,3,1,0]=(-0.717873944493)*x[0]**o+(0.987011714613)*x[0]+(0.908500737929)*x[1]**o+(0.612497707425)*x[1]
        arg[2,3,1,1]=(0.221645363399)*x[0]**o+(0.219147687681)*x[0]+(-0.195831118343)*x[1]**o+(-0.59541920919)*x[1]
        arg[2,3,2,0]=(-0.289487342206)*x[0]**o+(-0.96904981397)*x[0]+(-0.844880792969)*x[1]**o+(0.803145388038)*x[1]
        arg[2,3,2,1]=(-0.0234295826132)*x[0]**o+(-0.839069887016)*x[0]+(0.591231341444)*x[1]**o+(-0.678605318787)*x[1]
        arg[2,4,0,0]=(0.717943320641)*x[0]**o+(0.625728011485)*x[0]+(0.456836257053)*x[1]**o+(-0.00523685585287)*x[1]
        arg[2,4,0,1]=(-0.549526374185)*x[0]**o+(0.57441917876)*x[0]+(-0.913459106781)*x[1]**o+(-0.610369375195)*x[1]
        arg[2,4,1,0]=(-0.27166563459)*x[0]**o+(0.146710581194)*x[0]+(0.297934124718)*x[1]**o+(0.989561209048)*x[1]
        arg[2,4,1,1]=(-0.0973858709778)*x[0]**o+(-0.909133394381)*x[0]+(0.634182571287)*x[1]**o+(0.477163052831)*x[1]
        arg[2,4,2,0]=(0.0878998038301)*x[0]**o+(-0.90568487021)*x[0]+(0.798978355456)*x[1]**o+(0.977402478037)*x[1]
        arg[2,4,2,1]=(-0.456868538667)*x[0]**o+(0.0723272702119)*x[0]+(0.143461337462)*x[1]**o+(-0.923713568343)*x[1]
        arg[3,0,0,0]=(-0.154410689788)*x[0]**o+(0.634925838303)*x[0]+(-0.0508765762892)*x[1]**o+(-0.889731180834)*x[1]
        arg[3,0,0,1]=(-0.611336323518)*x[0]**o+(0.748058372943)*x[0]+(-0.395060568704)*x[1]**o+(-0.413772262434)*x[1]
        arg[3,0,1,0]=(-0.76924573047)*x[0]**o+(0.61286303551)*x[0]+(0.810883211884)*x[1]**o+(-0.908861022244)*x[1]
        arg[3,0,1,1]=(0.269999661235)*x[0]**o+(0.0330954575695)*x[0]+(0.560272546054)*x[1]**o+(0.877730763728)*x[1]
        arg[3,0,2,0]=(-0.617658215856)*x[0]**o+(0.758414649277)*x[0]+(-0.223497113863)*x[1]**o+(-0.0706589108357)*x[1]
        arg[3,0,2,1]=(0.47956571804)*x[0]**o+(0.938375191345)*x[0]+(-0.193997511219)*x[1]**o+(0.500063249143)*x[1]
        arg[3,1,0,0]=(0.994033773649)*x[0]**o+(0.203022945646)*x[0]+(0.439296640709)*x[1]**o+(0.283679657893)*x[1]
        arg[3,1,0,1]=(0.32495149734)*x[0]**o+(-0.871627347694)*x[0]+(-0.438159387872)*x[1]**o+(0.169112665588)*x[1]
        arg[3,1,1,0]=(-0.0979322945554)*x[0]**o+(-0.936928842822)*x[0]+(0.0402738011011)*x[1]**o+(-0.445031618539)*x[1]
        arg[3,1,1,1]=(-0.511218949907)*x[0]**o+(0.255146779315)*x[0]+(-0.802012163396)*x[1]**o+(0.361819705056)*x[1]
        arg[3,1,2,0]=(-0.902499203011)*x[0]**o+(0.76589952141)*x[0]+(0.52498929732)*x[1]**o+(0.493894292321)*x[1]
        arg[3,1,2,1]=(-0.475117609018)*x[0]**o+(0.682205309303)*x[0]+(-0.661824086649)*x[1]**o+(0.985521553774)*x[1]
        arg[3,2,0,0]=(0.937321686473)*x[0]**o+(-0.518305775925)*x[0]+(0.171689574942)*x[1]**o+(-0.751579960819)*x[1]
        arg[3,2,0,1]=(-0.607004597985)*x[0]**o+(0.342330208344)*x[0]+(0.692094243326)*x[1]**o+(-0.457756057357)*x[1]
        arg[3,2,1,0]=(0.126171621354)*x[0]**o+(0.0499938423847)*x[0]+(0.572006209576)*x[1]**o+(0.121601182464)*x[1]
        arg[3,2,1,1]=(0.337936456181)*x[0]**o+(-0.940186907514)*x[0]+(-0.965354931124)*x[1]**o+(-0.0503295532873)*x[1]
        arg[3,2,2,0]=(0.439590816231)*x[0]**o+(-0.145314466072)*x[0]+(-0.847175364201)*x[1]**o+(-0.570000251495)*x[1]
        arg[3,2,2,1]=(0.401069335123)*x[0]**o+(-0.117139167508)*x[0]+(0.192907527853)*x[1]**o+(0.573283391581)*x[1]
        arg[3,3,0,0]=(-0.434029411666)*x[0]**o+(-0.689115317009)*x[0]+(0.63450218888)*x[1]**o+(0.492858023519)*x[1]
        arg[3,3,0,1]=(-0.678722283369)*x[0]**o+(-0.62031826393)*x[0]+(-0.458646022327)*x[1]**o+(-0.650857410927)*x[1]
        arg[3,3,1,0]=(0.413477874719)*x[0]**o+(-0.903116096879)*x[0]+(-0.505961511722)*x[1]**o+(0.649554735459)*x[1]
        arg[3,3,1,1]=(-0.58459645458)*x[0]**o+(0.861318565628)*x[0]+(0.378511063123)*x[1]**o+(-0.694820011654)*x[1]
        arg[3,3,2,0]=(0.82844733258)*x[0]**o+(-0.151929131446)*x[0]+(0.322848411081)*x[1]**o+(0.436062233263)*x[1]
        arg[3,3,2,1]=(0.269627291792)*x[0]**o+(0.618140998132)*x[0]+(0.851057431407)*x[1]**o+(0.716088832502)*x[1]
        arg[3,4,0,0]=(0.0557605432613)*x[0]**o+(0.0689569775656)*x[0]+(0.489781121774)*x[1]**o+(0.214459998733)*x[1]
        arg[3,4,0,1]=(0.429111134289)*x[0]**o+(0.648324951051)*x[0]+(0.00596799425797)*x[1]**o+(-0.752637714965)*x[1]
        arg[3,4,1,0]=(-0.0394648097179)*x[0]**o+(0.842965554058)*x[0]+(-0.124493160479)*x[1]**o+(0.81914053244)*x[1]
        arg[3,4,1,1]=(0.309465325874)*x[0]**o+(0.631381272937)*x[0]+(-0.164415508845)*x[1]**o+(0.316908911599)*x[1]
        arg[3,4,2,0]=(-0.615108314792)*x[0]**o+(0.300936829452)*x[0]+(-0.289514252333)*x[1]**o+(0.844949030767)*x[1]
        arg[3,4,2,1]=(0.80117516909)*x[0]**o+(-0.602632221639)*x[0]+(-0.68193581457)*x[1]**o+(0.203319693067)*x[1]
        ref[0,0,0,0]=(0.0737428272134)*x_ref[0]**o+(0.841780036645)*x_ref[0]+(-0.110824257531)*x_ref[1]**o+(-0.957726444468)*x_ref[1]
        ref[0,0,0,1]=(-0.934840410671)*x_ref[0]**o+(-0.83301064321)*x_ref[0]+(-0.352287650266)*x_ref[1]**o+(-0.961111564077)*x_ref[1]
        ref[0,0,1,0]=(-0.829500940226)*x_ref[0]**o+(0.644171710931)*x_ref[0]+(0.232545739722)*x_ref[1]**o+(0.584336012862)*x_ref[1]
        ref[0,0,1,1]=(0.426298830976)*x_ref[0]**o+(-0.615281197268)*x_ref[0]+(0.206334026192)*x_ref[1]**o+(-0.443698947044)*x_ref[1]
        ref[0,0,2,0]=(-0.236495891617)*x_ref[0]**o+(0.0883376851179)*x_ref[0]+(-0.176541434356)*x_ref[1]**o+(0.297175301617)*x_ref[1]
        ref[0,0,2,1]=(-0.879955324373)*x_ref[0]**o+(-0.244778913771)*x_ref[0]+(0.244131188594)*x_ref[1]**o+(-0.334739112198)*x_ref[1]
        ref[0,1,0,0]=(-0.64887128006)*x_ref[0]**o+(-0.489951454377)*x_ref[0]+(-0.328402433668)*x_ref[1]**o+(0.229814237117)*x_ref[1]
        ref[0,1,0,1]=(-0.339451656097)*x_ref[0]**o+(-0.409320682304)*x_ref[0]+(0.216482338677)*x_ref[1]**o+(-0.10594373055)*x_ref[1]
        ref[0,1,1,0]=(-0.350538479425)*x_ref[0]**o+(0.466450046985)*x_ref[0]+(-0.779124795292)*x_ref[1]**o+(-0.654562596704)*x_ref[1]
        ref[0,1,1,1]=(0.00350092879868)*x_ref[0]**o+(-0.214643088647)*x_ref[0]+(0.786934081411)*x_ref[1]**o+(-0.296853380043)*x_ref[1]
        ref[0,1,2,0]=(-0.640183955648)*x_ref[0]**o+(0.568647968237)*x_ref[0]+(0.575533476901)*x_ref[1]**o+(0.854540594085)*x_ref[1]
        ref[0,1,2,1]=(0.633674899518)*x_ref[0]**o+(0.829255679443)*x_ref[0]+(0.537888424303)*x_ref[1]**o+(-0.495082587617)*x_ref[1]
        ref[0,2,0,0]=(-0.700436798342)*x_ref[0]**o+(-0.732581859026)*x_ref[0]+(-0.508905353053)*x_ref[1]**o+(0.664601151676)*x_ref[1]
        ref[0,2,0,1]=(0.720335764302)*x_ref[0]**o+(0.941159917988)*x_ref[0]+(0.2817440547)*x_ref[1]**o+(0.553712160058)*x_ref[1]
        ref[0,2,1,0]=(-0.937114312262)*x_ref[0]**o+(-0.115517396201)*x_ref[0]+(-0.672697125905)*x_ref[1]**o+(-0.221320095369)*x_ref[1]
        ref[0,2,1,1]=(-0.52059341062)*x_ref[0]**o+(-0.653834647475)*x_ref[0]+(-0.0851100861135)*x_ref[1]**o+(-0.668300210067)*x_ref[1]
        ref[0,2,2,0]=(-0.943613035509)*x_ref[0]**o+(-0.294843145822)*x_ref[0]+(0.496278748831)*x_ref[1]**o+(0.901154419425)*x_ref[1]
        ref[0,2,2,1]=(-0.421698818203)*x_ref[0]**o+(-0.721918177544)*x_ref[0]+(-0.197855702215)*x_ref[1]**o+(0.50882827715)*x_ref[1]
        ref[0,3,0,0]=(0.931990780298)*x_ref[0]**o+(-0.550332606951)*x_ref[0]+(0.401584654642)*x_ref[1]**o+(0.0965259747601)*x_ref[1]
        ref[0,3,0,1]=(0.881357708445)*x_ref[0]**o+(0.564624034197)*x_ref[0]+(-0.458191897924)*x_ref[1]**o+(-0.427701112807)*x_ref[1]
        ref[0,3,1,0]=(0.624777088081)*x_ref[0]**o+(0.862654721208)*x_ref[0]+(-0.707618093268)*x_ref[1]**o+(0.743304953313)*x_ref[1]
        ref[0,3,1,1]=(0.4510836974)*x_ref[0]**o+(-0.357023031453)*x_ref[0]+(-0.00503924010273)*x_ref[1]**o+(0.545024336011)*x_ref[1]
        ref[0,3,2,0]=(0.109003851094)*x_ref[0]**o+(0.168659376881)*x_ref[0]+(-0.0578953780655)*x_ref[1]**o+(-0.481212462712)*x_ref[1]
        ref[0,3,2,1]=(-0.191183501636)*x_ref[0]**o+(-0.316145516377)*x_ref[0]+(0.402460103016)*x_ref[1]**o+(-0.163612464438)*x_ref[1]
        ref[0,4,0,0]=(0.489815345033)*x_ref[0]**o+(0.999138280765)*x_ref[0]+(-0.626627766812)*x_ref[1]**o+(-0.339653992568)*x_ref[1]
        ref[0,4,0,1]=(0.616350228791)*x_ref[0]**o+(0.933446438765)*x_ref[0]+(0.944244035512)*x_ref[1]**o+(0.829407789353)*x_ref[1]
        ref[0,4,1,0]=(0.619918547863)*x_ref[0]**o+(0.995488719729)*x_ref[0]+(-0.78847429581)*x_ref[1]**o+(0.542645674462)*x_ref[1]
        ref[0,4,1,1]=(-0.149244819992)*x_ref[0]**o+(0.921766299925)*x_ref[0]+(0.435255339567)*x_ref[1]**o+(-0.0596561300741)*x_ref[1]
        ref[0,4,2,0]=(-0.816364336598)*x_ref[0]**o+(0.701160246658)*x_ref[0]+(-0.616385505652)*x_ref[1]**o+(-0.558625482105)*x_ref[1]
        ref[0,4,2,1]=(-0.36944364693)*x_ref[0]**o+(0.243426977623)*x_ref[0]+(-0.61578435391)*x_ref[1]**o+(0.932570326139)*x_ref[1]
        ref[1,0,0,0]=(-0.36839962807)*x_ref[0]**o+(0.590534439534)*x_ref[0]+(-0.615158899289)*x_ref[1]**o+(0.973649158296)*x_ref[1]
        ref[1,0,0,1]=(0.761515772771)*x_ref[0]**o+(0.706406307012)*x_ref[0]+(0.328712319043)*x_ref[1]**o+(0.438600140517)*x_ref[1]
        ref[1,0,1,0]=(0.895080579207)*x_ref[0]**o+(-0.166789328225)*x_ref[0]+(0.807540897371)*x_ref[1]**o+(0.0345522692723)*x_ref[1]
        ref[1,0,1,1]=(-0.460062661172)*x_ref[0]**o+(0.505167337562)*x_ref[0]+(0.25155527144)*x_ref[1]**o+(0.601093887451)*x_ref[1]
        ref[1,0,2,0]=(0.382129818527)*x_ref[0]**o+(0.694676536542)*x_ref[0]+(0.702794119083)*x_ref[1]**o+(0.704094013056)*x_ref[1]
        ref[1,0,2,1]=(0.265476982913)*x_ref[0]**o+(-0.134338744881)*x_ref[0]+(-0.0938910335356)*x_ref[1]**o+(-0.560073316097)*x_ref[1]
        ref[1,1,0,0]=(0.977368328792)*x_ref[0]**o+(0.084482890741)*x_ref[0]+(0.57068389287)*x_ref[1]**o+(-0.242087071526)*x_ref[1]
        ref[1,1,0,1]=(-0.36067775897)*x_ref[0]**o+(0.768187470286)*x_ref[0]+(-0.881150642619)*x_ref[1]**o+(-0.15839498619)*x_ref[1]
        ref[1,1,1,0]=(-0.808263913645)*x_ref[0]**o+(0.788403321664)*x_ref[0]+(0.976206012343)*x_ref[1]**o+(-0.636933353579)*x_ref[1]
        ref[1,1,1,1]=(0.699185445546)*x_ref[0]**o+(0.251811206876)*x_ref[0]+(-0.146320762812)*x_ref[1]**o+(-0.248033048423)*x_ref[1]
        ref[1,1,2,0]=(-0.782940880812)*x_ref[0]**o+(0.789401686528)*x_ref[0]+(-0.806883904562)*x_ref[1]**o+(0.83689696772)*x_ref[1]
        ref[1,1,2,1]=(0.652404993342)*x_ref[0]**o+(-0.135318598212)*x_ref[0]+(0.138928760301)*x_ref[1]**o+(0.675593554805)*x_ref[1]
        ref[1,2,0,0]=(0.471746379585)*x_ref[0]**o+(-0.989783415847)*x_ref[0]+(-0.0935052296418)*x_ref[1]**o+(-0.532245102898)*x_ref[1]
        ref[1,2,0,1]=(-0.44541059979)*x_ref[0]**o+(0.593734266824)*x_ref[0]+(-0.616647011751)*x_ref[1]**o+(0.461140297024)*x_ref[1]
        ref[1,2,1,0]=(-0.9144210459)*x_ref[0]**o+(0.116806032197)*x_ref[0]+(0.935610395306)*x_ref[1]**o+(0.550884822213)*x_ref[1]
        ref[1,2,1,1]=(-0.962444455944)*x_ref[0]**o+(-0.654328013652)*x_ref[0]+(-0.545330994236)*x_ref[1]**o+(-0.919705456162)*x_ref[1]
        ref[1,2,2,0]=(-0.724491374462)*x_ref[0]**o+(-0.820402375498)*x_ref[0]+(-0.953956431336)*x_ref[1]**o+(-0.201450523182)*x_ref[1]
        ref[1,2,2,1]=(0.723637497881)*x_ref[0]**o+(-0.144483238523)*x_ref[0]+(0.00379745988102)*x_ref[1]**o+(-0.838616087543)*x_ref[1]
        ref[1,3,0,0]=(-0.962256046042)*x_ref[0]**o+(-0.453580322779)*x_ref[0]+(0.887446007424)*x_ref[1]**o+(-0.404454768131)*x_ref[1]
        ref[1,3,0,1]=(0.829189872775)*x_ref[0]**o+(0.327985850399)*x_ref[0]+(0.113600094307)*x_ref[1]**o+(-0.338775377551)*x_ref[1]
        ref[1,3,1,0]=(0.912320027833)*x_ref[0]**o+(0.0658152563951)*x_ref[0]+(-0.843491517568)*x_ref[1]**o+(0.858344434434)*x_ref[1]
        ref[1,3,1,1]=(-0.133918678317)*x_ref[0]**o+(0.644272729149)*x_ref[0]+(0.140663927965)*x_ref[1]**o+(0.853385355163)*x_ref[1]
        ref[1,3,2,0]=(0.37733357322)*x_ref[0]**o+(0.255105934063)*x_ref[0]+(0.88881892183)*x_ref[1]**o+(-0.722661948279)*x_ref[1]
        ref[1,3,2,1]=(-0.2827466345)*x_ref[0]**o+(-0.518815029512)*x_ref[0]+(-0.995043824929)*x_ref[1]**o+(0.973473354481)*x_ref[1]
        ref[1,4,0,0]=(0.044904732821)*x_ref[0]**o+(-0.0373604593083)*x_ref[0]+(0.184254626748)*x_ref[1]**o+(0.031832590233)*x_ref[1]
        ref[1,4,0,1]=(0.388699053911)*x_ref[0]**o+(-0.0619606394414)*x_ref[0]+(0.477947819745)*x_ref[1]**o+(0.982909216221)*x_ref[1]
        ref[1,4,1,0]=(0.854348679973)*x_ref[0]**o+(-0.822052612379)*x_ref[0]+(-0.950305089968)*x_ref[1]**o+(0.0100277256384)*x_ref[1]
        ref[1,4,1,1]=(0.265075277179)*x_ref[0]**o+(0.489846993407)*x_ref[0]+(-0.961111499722)*x_ref[1]**o+(-0.0457909055602)*x_ref[1]
        ref[1,4,2,0]=(0.701951191951)*x_ref[0]**o+(0.675232620259)*x_ref[0]+(-0.141848599227)*x_ref[1]**o+(0.737150257564)*x_ref[1]
        ref[1,4,2,1]=(-0.450684564415)*x_ref[0]**o+(0.854849922021)*x_ref[0]+(0.729670746723)*x_ref[1]**o+(0.530372628879)*x_ref[1]
        ref[2,0,0,0]=(-0.131153262675)*x_ref[0]**o+(-0.477683407979)*x_ref[0]+(0.231416752516)*x_ref[1]**o+(0.790897064402)*x_ref[1]
        ref[2,0,0,1]=(0.0956557004382)*x_ref[0]**o+(0.209084913471)*x_ref[0]+(-0.328544695733)*x_ref[1]**o+(0.701245770328)*x_ref[1]
        ref[2,0,1,0]=(-0.476403718361)*x_ref[0]**o+(0.689023533158)*x_ref[0]+(0.108732189071)*x_ref[1]**o+(0.243379977453)*x_ref[1]
        ref[2,0,1,1]=(0.414917801157)*x_ref[0]**o+(-0.843544596785)*x_ref[0]+(0.772236792935)*x_ref[1]**o+(0.836028831232)*x_ref[1]
        ref[2,0,2,0]=(0.754970600617)*x_ref[0]**o+(-0.422318891122)*x_ref[0]+(0.770767911534)*x_ref[1]**o+(-0.446737894095)*x_ref[1]
        ref[2,0,2,1]=(0.0139989331471)*x_ref[0]**o+(0.262286042291)*x_ref[0]+(-0.440109329114)*x_ref[1]**o+(-0.844965114722)*x_ref[1]
        ref[2,1,0,0]=(-0.758977153009)*x_ref[0]**o+(-0.991192011039)*x_ref[0]+(0.642267239475)*x_ref[1]**o+(-0.701360491163)*x_ref[1]
        ref[2,1,0,1]=(-0.123019007509)*x_ref[0]**o+(-0.622323252979)*x_ref[0]+(0.441364031664)*x_ref[1]**o+(0.311433222821)*x_ref[1]
        ref[2,1,1,0]=(-0.0446644182928)*x_ref[0]**o+(0.614952704493)*x_ref[0]+(-0.228284494629)*x_ref[1]**o+(0.077603612991)*x_ref[1]
        ref[2,1,1,1]=(0.712924108225)*x_ref[0]**o+(-0.153966300537)*x_ref[0]+(-0.584450014127)*x_ref[1]**o+(0.476329149833)*x_ref[1]
        ref[2,1,2,0]=(0.554026516037)*x_ref[0]**o+(0.730575821847)*x_ref[0]+(-0.247529915107)*x_ref[1]**o+(0.210060454223)*x_ref[1]
        ref[2,1,2,1]=(0.877017208752)*x_ref[0]**o+(-0.800916200997)*x_ref[0]+(0.0996710064295)*x_ref[1]**o+(-0.0890375735808)*x_ref[1]
        ref[2,2,0,0]=(-0.198026186062)*x_ref[0]**o+(-0.416160224681)*x_ref[0]+(-0.626491103997)*x_ref[1]**o+(-0.950177752043)*x_ref[1]
        ref[2,2,0,1]=(-0.401319234479)*x_ref[0]**o+(0.948120359627)*x_ref[0]+(0.420351950404)*x_ref[1]**o+(-0.242850312594)*x_ref[1]
        ref[2,2,1,0]=(-0.239652069063)*x_ref[0]**o+(-0.894811076801)*x_ref[0]+(0.979721413509)*x_ref[1]**o+(-0.712493893566)*x_ref[1]
        ref[2,2,1,1]=(0.239416494209)*x_ref[0]**o+(0.449047649538)*x_ref[0]+(0.0503751073583)*x_ref[1]**o+(0.739949331521)*x_ref[1]
        ref[2,2,2,0]=(-0.318067701034)*x_ref[0]**o+(-0.00208209268356)*x_ref[0]+(-0.513821624843)*x_ref[1]**o+(-0.202688815501)*x_ref[1]
        ref[2,2,2,1]=(-0.194111524544)*x_ref[0]**o+(-0.843213241929)*x_ref[0]+(0.124223747087)*x_ref[1]**o+(-0.988305059412)*x_ref[1]
        ref[2,3,0,0]=(0.450894655873)*x_ref[0]**o+(-0.809351657482)*x_ref[0]+(-0.80555128185)*x_ref[1]**o+(-0.922123806508)*x_ref[1]
        ref[2,3,0,1]=(-0.462822602909)*x_ref[0]**o+(0.504388766027)*x_ref[0]+(-0.152725064915)*x_ref[1]**o+(0.988745142638)*x_ref[1]
        ref[2,3,1,0]=(-0.717873944493)*x_ref[0]**o+(0.987011714613)*x_ref[0]+(0.908500737929)*x_ref[1]**o+(0.612497707425)*x_ref[1]
        ref[2,3,1,1]=(0.221645363399)*x_ref[0]**o+(0.219147687681)*x_ref[0]+(-0.195831118343)*x_ref[1]**o+(-0.59541920919)*x_ref[1]
        ref[2,3,2,0]=(-0.289487342206)*x_ref[0]**o+(-0.96904981397)*x_ref[0]+(-0.844880792969)*x_ref[1]**o+(0.803145388038)*x_ref[1]
        ref[2,3,2,1]=(-0.0234295826132)*x_ref[0]**o+(-0.839069887016)*x_ref[0]+(0.591231341444)*x_ref[1]**o+(-0.678605318787)*x_ref[1]
        ref[2,4,0,0]=(0.717943320641)*x_ref[0]**o+(0.625728011485)*x_ref[0]+(0.456836257053)*x_ref[1]**o+(-0.00523685585287)*x_ref[1]
        ref[2,4,0,1]=(-0.549526374185)*x_ref[0]**o+(0.57441917876)*x_ref[0]+(-0.913459106781)*x_ref[1]**o+(-0.610369375195)*x_ref[1]
        ref[2,4,1,0]=(-0.27166563459)*x_ref[0]**o+(0.146710581194)*x_ref[0]+(0.297934124718)*x_ref[1]**o+(0.989561209048)*x_ref[1]
        ref[2,4,1,1]=(-0.0973858709778)*x_ref[0]**o+(-0.909133394381)*x_ref[0]+(0.634182571287)*x_ref[1]**o+(0.477163052831)*x_ref[1]
        ref[2,4,2,0]=(0.0878998038301)*x_ref[0]**o+(-0.90568487021)*x_ref[0]+(0.798978355456)*x_ref[1]**o+(0.977402478037)*x_ref[1]
        ref[2,4,2,1]=(-0.456868538667)*x_ref[0]**o+(0.0723272702119)*x_ref[0]+(0.143461337462)*x_ref[1]**o+(-0.923713568343)*x_ref[1]
        ref[3,0,0,0]=(-0.154410689788)*x_ref[0]**o+(0.634925838303)*x_ref[0]+(-0.0508765762892)*x_ref[1]**o+(-0.889731180834)*x_ref[1]
        ref[3,0,0,1]=(-0.611336323518)*x_ref[0]**o+(0.748058372943)*x_ref[0]+(-0.395060568704)*x_ref[1]**o+(-0.413772262434)*x_ref[1]
        ref[3,0,1,0]=(-0.76924573047)*x_ref[0]**o+(0.61286303551)*x_ref[0]+(0.810883211884)*x_ref[1]**o+(-0.908861022244)*x_ref[1]
        ref[3,0,1,1]=(0.269999661235)*x_ref[0]**o+(0.0330954575695)*x_ref[0]+(0.560272546054)*x_ref[1]**o+(0.877730763728)*x_ref[1]
        ref[3,0,2,0]=(-0.617658215856)*x_ref[0]**o+(0.758414649277)*x_ref[0]+(-0.223497113863)*x_ref[1]**o+(-0.0706589108357)*x_ref[1]
        ref[3,0,2,1]=(0.47956571804)*x_ref[0]**o+(0.938375191345)*x_ref[0]+(-0.193997511219)*x_ref[1]**o+(0.500063249143)*x_ref[1]
        ref[3,1,0,0]=(0.994033773649)*x_ref[0]**o+(0.203022945646)*x_ref[0]+(0.439296640709)*x_ref[1]**o+(0.283679657893)*x_ref[1]
        ref[3,1,0,1]=(0.32495149734)*x_ref[0]**o+(-0.871627347694)*x_ref[0]+(-0.438159387872)*x_ref[1]**o+(0.169112665588)*x_ref[1]
        ref[3,1,1,0]=(-0.0979322945554)*x_ref[0]**o+(-0.936928842822)*x_ref[0]+(0.0402738011011)*x_ref[1]**o+(-0.445031618539)*x_ref[1]
        ref[3,1,1,1]=(-0.511218949907)*x_ref[0]**o+(0.255146779315)*x_ref[0]+(-0.802012163396)*x_ref[1]**o+(0.361819705056)*x_ref[1]
        ref[3,1,2,0]=(-0.902499203011)*x_ref[0]**o+(0.76589952141)*x_ref[0]+(0.52498929732)*x_ref[1]**o+(0.493894292321)*x_ref[1]
        ref[3,1,2,1]=(-0.475117609018)*x_ref[0]**o+(0.682205309303)*x_ref[0]+(-0.661824086649)*x_ref[1]**o+(0.985521553774)*x_ref[1]
        ref[3,2,0,0]=(0.937321686473)*x_ref[0]**o+(-0.518305775925)*x_ref[0]+(0.171689574942)*x_ref[1]**o+(-0.751579960819)*x_ref[1]
        ref[3,2,0,1]=(-0.607004597985)*x_ref[0]**o+(0.342330208344)*x_ref[0]+(0.692094243326)*x_ref[1]**o+(-0.457756057357)*x_ref[1]
        ref[3,2,1,0]=(0.126171621354)*x_ref[0]**o+(0.0499938423847)*x_ref[0]+(0.572006209576)*x_ref[1]**o+(0.121601182464)*x_ref[1]
        ref[3,2,1,1]=(0.337936456181)*x_ref[0]**o+(-0.940186907514)*x_ref[0]+(-0.965354931124)*x_ref[1]**o+(-0.0503295532873)*x_ref[1]
        ref[3,2,2,0]=(0.439590816231)*x_ref[0]**o+(-0.145314466072)*x_ref[0]+(-0.847175364201)*x_ref[1]**o+(-0.570000251495)*x_ref[1]
        ref[3,2,2,1]=(0.401069335123)*x_ref[0]**o+(-0.117139167508)*x_ref[0]+(0.192907527853)*x_ref[1]**o+(0.573283391581)*x_ref[1]
        ref[3,3,0,0]=(-0.434029411666)*x_ref[0]**o+(-0.689115317009)*x_ref[0]+(0.63450218888)*x_ref[1]**o+(0.492858023519)*x_ref[1]
        ref[3,3,0,1]=(-0.678722283369)*x_ref[0]**o+(-0.62031826393)*x_ref[0]+(-0.458646022327)*x_ref[1]**o+(-0.650857410927)*x_ref[1]
        ref[3,3,1,0]=(0.413477874719)*x_ref[0]**o+(-0.903116096879)*x_ref[0]+(-0.505961511722)*x_ref[1]**o+(0.649554735459)*x_ref[1]
        ref[3,3,1,1]=(-0.58459645458)*x_ref[0]**o+(0.861318565628)*x_ref[0]+(0.378511063123)*x_ref[1]**o+(-0.694820011654)*x_ref[1]
        ref[3,3,2,0]=(0.82844733258)*x_ref[0]**o+(-0.151929131446)*x_ref[0]+(0.322848411081)*x_ref[1]**o+(0.436062233263)*x_ref[1]
        ref[3,3,2,1]=(0.269627291792)*x_ref[0]**o+(0.618140998132)*x_ref[0]+(0.851057431407)*x_ref[1]**o+(0.716088832502)*x_ref[1]
        ref[3,4,0,0]=(0.0557605432613)*x_ref[0]**o+(0.0689569775656)*x_ref[0]+(0.489781121774)*x_ref[1]**o+(0.214459998733)*x_ref[1]
        ref[3,4,0,1]=(0.429111134289)*x_ref[0]**o+(0.648324951051)*x_ref[0]+(0.00596799425797)*x_ref[1]**o+(-0.752637714965)*x_ref[1]
        ref[3,4,1,0]=(-0.0394648097179)*x_ref[0]**o+(0.842965554058)*x_ref[0]+(-0.124493160479)*x_ref[1]**o+(0.81914053244)*x_ref[1]
        ref[3,4,1,1]=(0.309465325874)*x_ref[0]**o+(0.631381272937)*x_ref[0]+(-0.164415508845)*x_ref[1]**o+(0.316908911599)*x_ref[1]
        ref[3,4,2,0]=(-0.615108314792)*x_ref[0]**o+(0.300936829452)*x_ref[0]+(-0.289514252333)*x_ref[1]**o+(0.844949030767)*x_ref[1]
        ref[3,4,2,1]=(0.80117516909)*x_ref[0]**o+(-0.602632221639)*x_ref[0]+(-0.68193581457)*x_ref[1]**o+(0.203319693067)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.349437323395)*x[0]**o+(0.817821891642)*x[0]+(-0.357998520848)*x[1]**o+(0.349311985429)*x[1]+(0.470351907942)*x[2]**o+(0.264007843212)*x[2]
        arg[0,0,0,1]=(-0.747824148035)*x[0]**o+(0.551817215468)*x[0]+(0.980476067966)*x[1]**o+(0.521120126285)*x[1]+(-0.472357720779)*x[2]**o+(-0.379502066444)*x[2]
        arg[0,0,1,0]=(0.302984899868)*x[0]**o+(0.556601505796)*x[0]+(-0.698948476338)*x[1]**o+(-0.587379572024)*x[1]+(-0.625760359909)*x[2]**o+(0.231956543432)*x[2]
        arg[0,0,1,1]=(-0.335943560463)*x[0]**o+(-0.497643152155)*x[0]+(-0.849829024888)*x[1]**o+(-0.388751135318)*x[1]+(-0.517696637743)*x[2]**o+(-0.966210901)*x[2]
        arg[0,0,2,0]=(-0.0953058405321)*x[0]**o+(-0.864580918175)*x[0]+(-0.732086619685)*x[1]**o+(0.872044654816)*x[1]+(-0.225850391375)*x[2]**o+(-0.502203654371)*x[2]
        arg[0,0,2,1]=(0.228408069101)*x[0]**o+(0.267590362458)*x[0]+(0.127397041031)*x[1]**o+(0.0142393720782)*x[1]+(-0.257563412478)*x[2]**o+(0.0211877877507)*x[2]
        arg[0,1,0,0]=(0.921125039994)*x[0]**o+(0.372548304521)*x[0]+(-0.66574364308)*x[1]**o+(0.0895600219155)*x[1]+(-0.656870562037)*x[2]**o+(-0.608819638268)*x[2]
        arg[0,1,0,1]=(-0.396625252579)*x[0]**o+(0.655542320616)*x[0]+(-0.542629502059)*x[1]**o+(0.587317299384)*x[1]+(0.96380848967)*x[2]**o+(0.121341827766)*x[2]
        arg[0,1,1,0]=(0.0908208346003)*x[0]**o+(-0.0398227321951)*x[0]+(0.907607555288)*x[1]**o+(0.527230589013)*x[1]+(0.519287416653)*x[2]**o+(0.103180358249)*x[2]
        arg[0,1,1,1]=(0.0615195625837)*x[0]**o+(0.695720276474)*x[0]+(0.0862101684049)*x[1]**o+(0.517150419955)*x[1]+(0.511207669575)*x[2]**o+(-0.9556472362)*x[2]
        arg[0,1,2,0]=(0.70040395652)*x[0]**o+(-0.576266027865)*x[0]+(-0.133947982034)*x[1]**o+(-0.764976368456)*x[1]+(-0.860780506782)*x[2]**o+(-0.137946340663)*x[2]
        arg[0,1,2,1]=(0.356017596832)*x[0]**o+(0.94912783604)*x[0]+(0.325951760294)*x[1]**o+(0.0623695610721)*x[1]+(-0.106031051895)*x[2]**o+(-0.339314970581)*x[2]
        arg[0,2,0,0]=(-0.124954395016)*x[0]**o+(0.558233271782)*x[0]+(-0.956846820394)*x[1]**o+(-0.599332648704)*x[1]+(-0.322444201477)*x[2]**o+(0.0833108732532)*x[2]
        arg[0,2,0,1]=(-0.713065259589)*x[0]**o+(-0.597746726477)*x[0]+(-0.976654446591)*x[1]**o+(0.977779698697)*x[1]+(0.694427679422)*x[2]**o+(-0.364463986712)*x[2]
        arg[0,2,1,0]=(-0.415861468352)*x[0]**o+(-0.159802964908)*x[0]+(-0.132331280778)*x[1]**o+(0.120829594848)*x[1]+(-0.114181951611)*x[2]**o+(-0.810187313564)*x[2]
        arg[0,2,1,1]=(-0.869925695451)*x[0]**o+(-0.535444611026)*x[0]+(0.62908837594)*x[1]**o+(-0.488583607314)*x[1]+(0.8461615049)*x[2]**o+(-0.558010294727)*x[2]
        arg[0,2,2,0]=(-0.0822232217466)*x[0]**o+(-0.673631382141)*x[0]+(0.803717792107)*x[1]**o+(0.339934715615)*x[1]+(-0.886365924996)*x[2]**o+(0.0206517750753)*x[2]
        arg[0,2,2,1]=(0.367656249869)*x[0]**o+(0.000987963028539)*x[0]+(0.399182215451)*x[1]**o+(0.923259635138)*x[1]+(0.8573338122)*x[2]**o+(0.679082213415)*x[2]
        arg[0,3,0,0]=(-0.722680496644)*x[0]**o+(0.199427912651)*x[0]+(-0.351799245099)*x[1]**o+(0.240285416231)*x[1]+(-0.701704868242)*x[2]**o+(0.265610436018)*x[2]
        arg[0,3,0,1]=(-0.14479043328)*x[0]**o+(-0.192670959504)*x[0]+(0.327412080296)*x[1]**o+(-0.390894637576)*x[1]+(-0.401862813677)*x[2]**o+(-0.87150071561)*x[2]
        arg[0,3,1,0]=(0.430039400239)*x[0]**o+(-0.0691443536291)*x[0]+(0.205492039071)*x[1]**o+(-0.391580258314)*x[1]+(0.791748125642)*x[2]**o+(-0.0896722225431)*x[2]
        arg[0,3,1,1]=(-0.177554766913)*x[0]**o+(-0.399389195454)*x[0]+(-0.304228953477)*x[1]**o+(0.00462595697167)*x[1]+(0.271814579884)*x[2]**o+(-0.17305639368)*x[2]
        arg[0,3,2,0]=(-0.105869589157)*x[0]**o+(0.803488342924)*x[0]+(-0.65117861655)*x[1]**o+(-0.27529059731)*x[1]+(-0.777107387582)*x[2]**o+(-0.690980328369)*x[2]
        arg[0,3,2,1]=(-0.13792230295)*x[0]**o+(0.649721420509)*x[0]+(0.401441756822)*x[1]**o+(0.468949007973)*x[1]+(-0.762015699412)*x[2]**o+(0.389781367434)*x[2]
        arg[0,4,0,0]=(0.873653819021)*x[0]**o+(0.539406798913)*x[0]+(0.802177801924)*x[1]**o+(-0.861314989922)*x[1]+(0.847509960118)*x[2]**o+(-0.41381965291)*x[2]
        arg[0,4,0,1]=(0.0416858647633)*x[0]**o+(0.866825467268)*x[0]+(0.89776163679)*x[1]**o+(-0.978326218858)*x[1]+(-0.409795413277)*x[2]**o+(0.982449327977)*x[2]
        arg[0,4,1,0]=(-0.114530966075)*x[0]**o+(-0.182414280034)*x[0]+(0.154385814218)*x[1]**o+(0.869624332271)*x[1]+(0.477213291128)*x[2]**o+(-0.137123375612)*x[2]
        arg[0,4,1,1]=(-0.52963019184)*x[0]**o+(0.0217955878025)*x[0]+(-0.174531798693)*x[1]**o+(-0.738947178141)*x[1]+(0.198130667502)*x[2]**o+(-0.449952166994)*x[2]
        arg[0,4,2,0]=(0.998209708224)*x[0]**o+(0.734906841374)*x[0]+(0.225596363429)*x[1]**o+(0.862772101878)*x[1]+(-0.133563132638)*x[2]**o+(-0.78059821331)*x[2]
        arg[0,4,2,1]=(0.384580893901)*x[0]**o+(-0.881918368669)*x[0]+(-0.950768236608)*x[1]**o+(-0.452275072093)*x[1]+(0.592779023744)*x[2]**o+(0.798735695084)*x[2]
        arg[1,0,0,0]=(0.786827894179)*x[0]**o+(-0.233241290567)*x[0]+(0.946717698992)*x[1]**o+(0.179368479607)*x[1]+(-0.10409968587)*x[2]**o+(-0.770115022756)*x[2]
        arg[1,0,0,1]=(-0.355389699793)*x[0]**o+(0.356407960494)*x[0]+(-0.371831767713)*x[1]**o+(0.139120259823)*x[1]+(0.910761795881)*x[2]**o+(0.725966050775)*x[2]
        arg[1,0,1,0]=(0.316372320948)*x[0]**o+(0.432472869369)*x[0]+(0.733187356771)*x[1]**o+(0.133915829584)*x[1]+(0.380713482112)*x[2]**o+(-0.874827460114)*x[2]
        arg[1,0,1,1]=(-0.0625786527083)*x[0]**o+(-0.469518973019)*x[0]+(-0.318610570327)*x[1]**o+(-0.304412247656)*x[1]+(0.844057224107)*x[2]**o+(0.508926494536)*x[2]
        arg[1,0,2,0]=(-0.759697281694)*x[0]**o+(0.701456398418)*x[0]+(-0.671612899124)*x[1]**o+(0.292474197717)*x[1]+(-0.445886723952)*x[2]**o+(-0.319953406782)*x[2]
        arg[1,0,2,1]=(-0.865703619781)*x[0]**o+(0.115639820869)*x[0]+(0.493634056223)*x[1]**o+(0.0343245670418)*x[1]+(0.748200872819)*x[2]**o+(-0.208438395054)*x[2]
        arg[1,1,0,0]=(0.767661930656)*x[0]**o+(-0.578385615919)*x[0]+(0.241313268557)*x[1]**o+(0.74276605359)*x[1]+(0.373638986644)*x[2]**o+(0.475479466784)*x[2]
        arg[1,1,0,1]=(0.723228882826)*x[0]**o+(-0.988524686322)*x[0]+(0.333664096597)*x[1]**o+(-0.756435323592)*x[1]+(0.771173510221)*x[2]**o+(-0.879337556561)*x[2]
        arg[1,1,1,0]=(0.939682105701)*x[0]**o+(-0.172311578817)*x[0]+(-0.407241492831)*x[1]**o+(0.446660064834)*x[1]+(0.0371203367053)*x[2]**o+(-0.302853430561)*x[2]
        arg[1,1,1,1]=(0.1093213068)*x[0]**o+(0.31232933407)*x[0]+(0.47474581408)*x[1]**o+(-0.602763201317)*x[1]+(0.803198523872)*x[2]**o+(-0.178088233667)*x[2]
        arg[1,1,2,0]=(-0.217141686114)*x[0]**o+(0.287363815362)*x[0]+(-0.352552947551)*x[1]**o+(0.106445580983)*x[1]+(0.340814562712)*x[2]**o+(-0.963691249733)*x[2]
        arg[1,1,2,1]=(0.612302241126)*x[0]**o+(-0.771765095612)*x[0]+(-0.18662763339)*x[1]**o+(0.10259882275)*x[1]+(-0.29535785913)*x[2]**o+(-0.577876877465)*x[2]
        arg[1,2,0,0]=(0.497131459554)*x[0]**o+(0.912133134267)*x[0]+(0.944369633113)*x[1]**o+(0.163408823815)*x[1]+(0.85636766551)*x[2]**o+(-0.00282458078743)*x[2]
        arg[1,2,0,1]=(0.362490806338)*x[0]**o+(0.22490192994)*x[0]+(-0.913665309243)*x[1]**o+(-0.88010395132)*x[1]+(-0.926205930269)*x[2]**o+(-0.231830082111)*x[2]
        arg[1,2,1,0]=(-0.974792784505)*x[0]**o+(-0.765605618923)*x[0]+(0.241470054253)*x[1]**o+(0.445863114014)*x[1]+(0.561062749944)*x[2]**o+(-0.215674406206)*x[2]
        arg[1,2,1,1]=(0.827013456823)*x[0]**o+(-0.191864800977)*x[0]+(-0.158099224794)*x[1]**o+(-0.921959562711)*x[1]+(0.930693535629)*x[2]**o+(-0.352967183792)*x[2]
        arg[1,2,2,0]=(0.258970603305)*x[0]**o+(0.586779837275)*x[0]+(0.82707087137)*x[1]**o+(-0.215210535461)*x[1]+(0.36890299705)*x[2]**o+(-0.724385890088)*x[2]
        arg[1,2,2,1]=(-0.125969269863)*x[0]**o+(-0.505216539273)*x[0]+(0.43242008053)*x[1]**o+(0.212079268112)*x[1]+(0.345759190477)*x[2]**o+(-0.98505468199)*x[2]
        arg[1,3,0,0]=(-0.647118215098)*x[0]**o+(0.318566502446)*x[0]+(-0.151406735047)*x[1]**o+(0.253202745125)*x[1]+(-0.404823111329)*x[2]**o+(-0.520797366557)*x[2]
        arg[1,3,0,1]=(-0.946886311645)*x[0]**o+(0.733699524424)*x[0]+(-0.615470980807)*x[1]**o+(-0.103976314414)*x[1]+(-0.932104431883)*x[2]**o+(0.157563806559)*x[2]
        arg[1,3,1,0]=(-0.289471503181)*x[0]**o+(0.749124853135)*x[0]+(-0.578861403092)*x[1]**o+(-0.136231378985)*x[1]+(0.532719246315)*x[2]**o+(0.542455831657)*x[2]
        arg[1,3,1,1]=(-0.343842490684)*x[0]**o+(0.110039182926)*x[0]+(0.901234650534)*x[1]**o+(-0.513362876241)*x[1]+(0.371952189126)*x[2]**o+(-0.169400349862)*x[2]
        arg[1,3,2,0]=(-0.766791944503)*x[0]**o+(0.0267491122166)*x[0]+(-0.699293509597)*x[1]**o+(-0.697363669971)*x[1]+(0.567854170308)*x[2]**o+(-0.516301389629)*x[2]
        arg[1,3,2,1]=(-0.288508579358)*x[0]**o+(0.239321880772)*x[0]+(0.677249635795)*x[1]**o+(0.48369084768)*x[1]+(0.681460633595)*x[2]**o+(-0.80218734437)*x[2]
        arg[1,4,0,0]=(0.510273388936)*x[0]**o+(0.35600020965)*x[0]+(0.271077500616)*x[1]**o+(-0.00479335810576)*x[1]+(-0.348457583464)*x[2]**o+(-0.965895122152)*x[2]
        arg[1,4,0,1]=(0.0134532748813)*x[0]**o+(-0.396954380686)*x[0]+(-0.176682002946)*x[1]**o+(0.971224261138)*x[1]+(-0.359602180261)*x[2]**o+(0.373748703363)*x[2]
        arg[1,4,1,0]=(-0.519413160925)*x[0]**o+(0.856346255965)*x[0]+(0.159167243935)*x[1]**o+(0.92068779838)*x[1]+(0.415122182947)*x[2]**o+(-0.595702311496)*x[2]
        arg[1,4,1,1]=(-0.608969598184)*x[0]**o+(-0.434750765676)*x[0]+(0.74471475701)*x[1]**o+(-0.229333246615)*x[1]+(-0.678138032736)*x[2]**o+(-0.514112926798)*x[2]
        arg[1,4,2,0]=(0.0633453561735)*x[0]**o+(-0.278569473082)*x[0]+(-0.842227144473)*x[1]**o+(-0.00289799815652)*x[1]+(0.153392291113)*x[2]**o+(-0.61448940993)*x[2]
        arg[1,4,2,1]=(0.0897863401674)*x[0]**o+(-0.487876787706)*x[0]+(-0.464008634014)*x[1]**o+(0.576466896344)*x[1]+(0.962345361537)*x[2]**o+(0.580878819156)*x[2]
        arg[2,0,0,0]=(0.316582275434)*x[0]**o+(0.0607872538221)*x[0]+(0.642492366848)*x[1]**o+(0.200300167843)*x[1]+(-0.402802396184)*x[2]**o+(-0.834254866233)*x[2]
        arg[2,0,0,1]=(0.644614184272)*x[0]**o+(-0.086367798694)*x[0]+(-0.428671462705)*x[1]**o+(-0.517827373575)*x[1]+(0.528197945165)*x[2]**o+(-0.523092888005)*x[2]
        arg[2,0,1,0]=(0.41376145869)*x[0]**o+(-0.27425322418)*x[0]+(0.816942168928)*x[1]**o+(-0.882784454814)*x[1]+(-0.270205535673)*x[2]**o+(0.642372102899)*x[2]
        arg[2,0,1,1]=(0.584931736844)*x[0]**o+(0.622511792702)*x[0]+(0.76174625736)*x[1]**o+(0.0561486448557)*x[1]+(-0.588861269552)*x[2]**o+(-0.341896323569)*x[2]
        arg[2,0,2,0]=(0.635215108566)*x[0]**o+(0.299769133671)*x[0]+(0.141053572388)*x[1]**o+(0.102844255879)*x[1]+(-0.995154777534)*x[2]**o+(0.501319388994)*x[2]
        arg[2,0,2,1]=(0.750089643836)*x[0]**o+(0.57570708561)*x[0]+(-0.449935823319)*x[1]**o+(-0.902982109411)*x[1]+(0.561185740222)*x[2]**o+(0.422654056847)*x[2]
        arg[2,1,0,0]=(0.953986415316)*x[0]**o+(-0.7883326473)*x[0]+(-0.723690234059)*x[1]**o+(-0.537242781471)*x[1]+(-0.503466717519)*x[2]**o+(-0.777865169241)*x[2]
        arg[2,1,0,1]=(0.317357554441)*x[0]**o+(-0.326828957855)*x[0]+(-0.609992651237)*x[1]**o+(0.811066934743)*x[1]+(-0.90668441447)*x[2]**o+(0.92581926687)*x[2]
        arg[2,1,1,0]=(-0.0801236189368)*x[0]**o+(0.49151264813)*x[0]+(-0.0770654446151)*x[1]**o+(-0.314981217276)*x[1]+(-0.743180837478)*x[2]**o+(-0.554491413146)*x[2]
        arg[2,1,1,1]=(0.831017344937)*x[0]**o+(-0.61064793935)*x[0]+(0.652976737027)*x[1]**o+(0.242104127969)*x[1]+(0.593330958678)*x[2]**o+(-0.35899584298)*x[2]
        arg[2,1,2,0]=(-0.580618798556)*x[0]**o+(-0.936176474082)*x[0]+(-0.768716477288)*x[1]**o+(-0.353687473159)*x[1]+(-0.261509314019)*x[2]**o+(-0.442139199257)*x[2]
        arg[2,1,2,1]=(0.945760793889)*x[0]**o+(-0.646958670891)*x[0]+(-0.440341726063)*x[1]**o+(0.223283751014)*x[1]+(0.799798944743)*x[2]**o+(0.591902067633)*x[2]
        arg[2,2,0,0]=(-0.267699746405)*x[0]**o+(0.271554172399)*x[0]+(0.747278903642)*x[1]**o+(-0.832877459834)*x[1]+(-0.640373164716)*x[2]**o+(-0.666487164763)*x[2]
        arg[2,2,0,1]=(0.795630715924)*x[0]**o+(0.682973866186)*x[0]+(-0.750818874926)*x[1]**o+(-0.904782408217)*x[1]+(-0.578138718379)*x[2]**o+(-0.369846516992)*x[2]
        arg[2,2,1,0]=(0.165654497062)*x[0]**o+(-0.329827871165)*x[0]+(-0.808981530672)*x[1]**o+(0.350106763785)*x[1]+(0.106638087689)*x[2]**o+(0.425516820398)*x[2]
        arg[2,2,1,1]=(0.017419119421)*x[0]**o+(0.500359047996)*x[0]+(-0.888509940049)*x[1]**o+(-0.375469124891)*x[1]+(0.863718970618)*x[2]**o+(-0.807667604763)*x[2]
        arg[2,2,2,0]=(0.860662652165)*x[0]**o+(0.162530730499)*x[0]+(-0.657510966961)*x[1]**o+(0.215905840286)*x[1]+(0.97380371264)*x[2]**o+(-0.431027671897)*x[2]
        arg[2,2,2,1]=(-0.649151503991)*x[0]**o+(0.484589779517)*x[0]+(0.905080495553)*x[1]**o+(0.997504119204)*x[1]+(-0.4268802512)*x[2]**o+(-0.687236308612)*x[2]
        arg[2,3,0,0]=(-0.60485902589)*x[0]**o+(0.49500777286)*x[0]+(0.33977532104)*x[1]**o+(-0.219209845338)*x[1]+(-0.197311627683)*x[2]**o+(0.285859591044)*x[2]
        arg[2,3,0,1]=(0.117438340483)*x[0]**o+(-0.562954894877)*x[0]+(-0.302982842797)*x[1]**o+(0.977923453436)*x[1]+(0.300127749784)*x[2]**o+(0.45784201373)*x[2]
        arg[2,3,1,0]=(-0.910061442266)*x[0]**o+(0.723114553174)*x[0]+(-0.435693478652)*x[1]**o+(-0.584326584996)*x[1]+(-0.51000911177)*x[2]**o+(0.506615483889)*x[2]
        arg[2,3,1,1]=(-0.50179741763)*x[0]**o+(-0.954551290001)*x[0]+(0.375839266192)*x[1]**o+(0.122120198059)*x[1]+(-0.593050674641)*x[2]**o+(0.549989594312)*x[2]
        arg[2,3,2,0]=(-0.889519666705)*x[0]**o+(-0.815572507674)*x[0]+(0.453677651057)*x[1]**o+(0.450910828217)*x[1]+(-0.00465373008682)*x[2]**o+(0.731899900165)*x[2]
        arg[2,3,2,1]=(-0.969971151589)*x[0]**o+(-0.156948104522)*x[0]+(0.492885029353)*x[1]**o+(-0.0310677847833)*x[1]+(0.327614233837)*x[2]**o+(-0.274974040707)*x[2]
        arg[2,4,0,0]=(0.844667900826)*x[0]**o+(0.863374870522)*x[0]+(0.251456923241)*x[1]**o+(0.434916005183)*x[1]+(-0.139023370677)*x[2]**o+(0.0655176616053)*x[2]
        arg[2,4,0,1]=(0.831755059318)*x[0]**o+(0.0897642674136)*x[0]+(0.190414878812)*x[1]**o+(0.976565327031)*x[1]+(0.413321767371)*x[2]**o+(0.755395415341)*x[2]
        arg[2,4,1,0]=(-0.790364876377)*x[0]**o+(0.829442094248)*x[0]+(0.556251403891)*x[1]**o+(-0.953487711658)*x[1]+(-0.610781905401)*x[2]**o+(-0.320764072326)*x[2]
        arg[2,4,1,1]=(-0.757101565136)*x[0]**o+(0.789340839395)*x[0]+(-0.615189514438)*x[1]**o+(-0.910992676978)*x[1]+(0.693736813401)*x[2]**o+(0.0538707588266)*x[2]
        arg[2,4,2,0]=(0.000335625616611)*x[0]**o+(-0.195171446476)*x[0]+(0.0281149133705)*x[1]**o+(-0.362294159243)*x[1]+(0.830520288743)*x[2]**o+(-0.431967694247)*x[2]
        arg[2,4,2,1]=(-0.232668364843)*x[0]**o+(0.655302452302)*x[0]+(-0.0125806408337)*x[1]**o+(-0.103979202109)*x[1]+(-0.277450383033)*x[2]**o+(0.708914971885)*x[2]
        arg[3,0,0,0]=(-0.813447875186)*x[0]**o+(0.133491078101)*x[0]+(-0.675510725453)*x[1]**o+(-0.161489202717)*x[1]+(0.0474778506607)*x[2]**o+(0.513729029652)*x[2]
        arg[3,0,0,1]=(0.45388386743)*x[0]**o+(-0.909074944715)*x[0]+(-0.540576382376)*x[1]**o+(-0.849753706068)*x[1]+(-0.496320191697)*x[2]**o+(-0.930725196363)*x[2]
        arg[3,0,1,0]=(-0.145526830518)*x[0]**o+(-0.34717641097)*x[0]+(-0.818886792022)*x[1]**o+(0.574315644784)*x[1]+(-0.780431329222)*x[2]**o+(0.892840747651)*x[2]
        arg[3,0,1,1]=(-0.415342962745)*x[0]**o+(-0.899768945518)*x[0]+(0.798573910731)*x[1]**o+(-0.0652893933479)*x[1]+(-0.444459519602)*x[2]**o+(-0.0100131702565)*x[2]
        arg[3,0,2,0]=(-0.997936676013)*x[0]**o+(0.822000858069)*x[0]+(-0.553794408949)*x[1]**o+(0.829037287423)*x[1]+(-0.700288437806)*x[2]**o+(-0.449584600891)*x[2]
        arg[3,0,2,1]=(-0.70979868598)*x[0]**o+(0.689868687655)*x[0]+(0.151257146397)*x[1]**o+(0.32895215144)*x[1]+(0.286043871383)*x[2]**o+(-0.604421567781)*x[2]
        arg[3,1,0,0]=(0.60431849542)*x[0]**o+(0.518618291627)*x[0]+(-0.471868097013)*x[1]**o+(-0.639110899151)*x[1]+(-0.0904184061601)*x[2]**o+(0.113762528304)*x[2]
        arg[3,1,0,1]=(0.307624103157)*x[0]**o+(-0.388281302162)*x[0]+(0.772429578102)*x[1]**o+(-0.197799629585)*x[1]+(-0.149120667036)*x[2]**o+(0.830154780264)*x[2]
        arg[3,1,1,0]=(0.957127413133)*x[0]**o+(0.730080945389)*x[0]+(0.497191948774)*x[1]**o+(-0.586439221663)*x[1]+(-0.56223071093)*x[2]**o+(0.162899395804)*x[2]
        arg[3,1,1,1]=(-0.285740818451)*x[0]**o+(-0.573919970411)*x[0]+(-0.400833211669)*x[1]**o+(0.84414534634)*x[1]+(-0.124352096731)*x[2]**o+(0.907013800834)*x[2]
        arg[3,1,2,0]=(0.186947347989)*x[0]**o+(0.499582449213)*x[0]+(0.525473636695)*x[1]**o+(-0.926500953998)*x[1]+(-0.0392615366415)*x[2]**o+(0.909002911192)*x[2]
        arg[3,1,2,1]=(0.634350440683)*x[0]**o+(0.901699291665)*x[0]+(-0.683899667575)*x[1]**o+(0.411642037744)*x[1]+(-0.72791611585)*x[2]**o+(-0.013293689839)*x[2]
        arg[3,2,0,0]=(0.185164429507)*x[0]**o+(-0.671011652257)*x[0]+(0.973361040829)*x[1]**o+(0.989738565899)*x[1]+(-0.501245887091)*x[2]**o+(-0.405445444386)*x[2]
        arg[3,2,0,1]=(-0.311554020884)*x[0]**o+(-0.929020990333)*x[0]+(0.366016375797)*x[1]**o+(0.838234804773)*x[1]+(0.939602632409)*x[2]**o+(-0.885524260372)*x[2]
        arg[3,2,1,0]=(-0.0101604655589)*x[0]**o+(0.670274078493)*x[0]+(0.198135855354)*x[1]**o+(-0.000275422963644)*x[1]+(0.136674233404)*x[2]**o+(-0.442510118497)*x[2]
        arg[3,2,1,1]=(-0.677948769785)*x[0]**o+(-0.165403610813)*x[0]+(-0.219353502981)*x[1]**o+(0.664297154009)*x[1]+(0.158536302892)*x[2]**o+(0.293940884925)*x[2]
        arg[3,2,2,0]=(0.666669534708)*x[0]**o+(0.609459298502)*x[0]+(0.297058070582)*x[1]**o+(-0.0809555825821)*x[1]+(0.139430791042)*x[2]**o+(-0.253514861624)*x[2]
        arg[3,2,2,1]=(-0.577932413274)*x[0]**o+(-0.583353108895)*x[0]+(-0.839262034394)*x[1]**o+(0.795642236106)*x[1]+(0.0882627381543)*x[2]**o+(-0.382970623122)*x[2]
        arg[3,3,0,0]=(0.187941627429)*x[0]**o+(0.552040688531)*x[0]+(0.837458832716)*x[1]**o+(0.296793121897)*x[1]+(-0.665699511216)*x[2]**o+(-0.318565991052)*x[2]
        arg[3,3,0,1]=(-0.78770176804)*x[0]**o+(0.807509260396)*x[0]+(0.46241086085)*x[1]**o+(0.668506908751)*x[1]+(-0.988954457343)*x[2]**o+(0.0896365679915)*x[2]
        arg[3,3,1,0]=(0.727665746433)*x[0]**o+(-0.819391676739)*x[0]+(-0.118648934607)*x[1]**o+(-0.433862939031)*x[1]+(0.827788711313)*x[2]**o+(-0.615751446784)*x[2]
        arg[3,3,1,1]=(-0.00427901841482)*x[0]**o+(-0.38545713036)*x[0]+(0.570160083404)*x[1]**o+(-0.235505162795)*x[1]+(0.514578304371)*x[2]**o+(-0.152692113279)*x[2]
        arg[3,3,2,0]=(-0.396900613831)*x[0]**o+(0.161875314543)*x[0]+(0.28471295984)*x[1]**o+(-0.623276596728)*x[1]+(-0.748772373219)*x[2]**o+(0.584246898704)*x[2]
        arg[3,3,2,1]=(0.210026027009)*x[0]**o+(-0.530437530387)*x[0]+(0.437155825716)*x[1]**o+(0.984102407204)*x[1]+(0.125030773605)*x[2]**o+(-0.821764907318)*x[2]
        arg[3,4,0,0]=(-0.819129946783)*x[0]**o+(-0.729517565451)*x[0]+(-0.126013244602)*x[1]**o+(0.592832125254)*x[1]+(-0.850426487488)*x[2]**o+(0.271550912174)*x[2]
        arg[3,4,0,1]=(-0.573554440921)*x[0]**o+(0.232106829584)*x[0]+(0.810201614042)*x[1]**o+(-0.85518164211)*x[1]+(0.249535563949)*x[2]**o+(-0.69728963017)*x[2]
        arg[3,4,1,0]=(0.785661643628)*x[0]**o+(0.936424600424)*x[0]+(0.321363811055)*x[1]**o+(0.362341458592)*x[1]+(-0.593576579363)*x[2]**o+(0.0601278435924)*x[2]
        arg[3,4,1,1]=(-0.854420686885)*x[0]**o+(0.993352097501)*x[0]+(-0.572116112224)*x[1]**o+(0.308008845182)*x[1]+(0.0443339578639)*x[2]**o+(-0.1469857285)*x[2]
        arg[3,4,2,0]=(0.696373642151)*x[0]**o+(0.82425126832)*x[0]+(-0.0891735358455)*x[1]**o+(-0.22218993613)*x[1]+(-0.238393149017)*x[2]**o+(-0.631352718194)*x[2]
        arg[3,4,2,1]=(-0.0932690431928)*x[0]**o+(-0.391900483919)*x[0]+(0.236843565361)*x[1]**o+(-0.164197648532)*x[1]+(0.235502596353)*x[2]**o+(0.0110131867917)*x[2]
        ref[0,0,0,0]=(0.349437323395)*x_ref[0]**o+(0.817821891642)*x_ref[0]+(-0.357998520848)*x_ref[1]**o+(0.349311985429)*x_ref[1]+(0.470351907942)*x_ref[2]**o+(0.264007843212)*x_ref[2]
        ref[0,0,0,1]=(-0.747824148035)*x_ref[0]**o+(0.551817215468)*x_ref[0]+(0.980476067966)*x_ref[1]**o+(0.521120126285)*x_ref[1]+(-0.472357720779)*x_ref[2]**o+(-0.379502066444)*x_ref[2]
        ref[0,0,1,0]=(0.302984899868)*x_ref[0]**o+(0.556601505796)*x_ref[0]+(-0.698948476338)*x_ref[1]**o+(-0.587379572024)*x_ref[1]+(-0.625760359909)*x_ref[2]**o+(0.231956543432)*x_ref[2]
        ref[0,0,1,1]=(-0.335943560463)*x_ref[0]**o+(-0.497643152155)*x_ref[0]+(-0.849829024888)*x_ref[1]**o+(-0.388751135318)*x_ref[1]+(-0.517696637743)*x_ref[2]**o+(-0.966210901)*x_ref[2]
        ref[0,0,2,0]=(-0.0953058405321)*x_ref[0]**o+(-0.864580918175)*x_ref[0]+(-0.732086619685)*x_ref[1]**o+(0.872044654816)*x_ref[1]+(-0.225850391375)*x_ref[2]**o+(-0.502203654371)*x_ref[2]
        ref[0,0,2,1]=(0.228408069101)*x_ref[0]**o+(0.267590362458)*x_ref[0]+(0.127397041031)*x_ref[1]**o+(0.0142393720782)*x_ref[1]+(-0.257563412478)*x_ref[2]**o+(0.0211877877507)*x_ref[2]
        ref[0,1,0,0]=(0.921125039994)*x_ref[0]**o+(0.372548304521)*x_ref[0]+(-0.66574364308)*x_ref[1]**o+(0.0895600219155)*x_ref[1]+(-0.656870562037)*x_ref[2]**o+(-0.608819638268)*x_ref[2]
        ref[0,1,0,1]=(-0.396625252579)*x_ref[0]**o+(0.655542320616)*x_ref[0]+(-0.542629502059)*x_ref[1]**o+(0.587317299384)*x_ref[1]+(0.96380848967)*x_ref[2]**o+(0.121341827766)*x_ref[2]
        ref[0,1,1,0]=(0.0908208346003)*x_ref[0]**o+(-0.0398227321951)*x_ref[0]+(0.907607555288)*x_ref[1]**o+(0.527230589013)*x_ref[1]+(0.519287416653)*x_ref[2]**o+(0.103180358249)*x_ref[2]
        ref[0,1,1,1]=(0.0615195625837)*x_ref[0]**o+(0.695720276474)*x_ref[0]+(0.0862101684049)*x_ref[1]**o+(0.517150419955)*x_ref[1]+(0.511207669575)*x_ref[2]**o+(-0.9556472362)*x_ref[2]
        ref[0,1,2,0]=(0.70040395652)*x_ref[0]**o+(-0.576266027865)*x_ref[0]+(-0.133947982034)*x_ref[1]**o+(-0.764976368456)*x_ref[1]+(-0.860780506782)*x_ref[2]**o+(-0.137946340663)*x_ref[2]
        ref[0,1,2,1]=(0.356017596832)*x_ref[0]**o+(0.94912783604)*x_ref[0]+(0.325951760294)*x_ref[1]**o+(0.0623695610721)*x_ref[1]+(-0.106031051895)*x_ref[2]**o+(-0.339314970581)*x_ref[2]
        ref[0,2,0,0]=(-0.124954395016)*x_ref[0]**o+(0.558233271782)*x_ref[0]+(-0.956846820394)*x_ref[1]**o+(-0.599332648704)*x_ref[1]+(-0.322444201477)*x_ref[2]**o+(0.0833108732532)*x_ref[2]
        ref[0,2,0,1]=(-0.713065259589)*x_ref[0]**o+(-0.597746726477)*x_ref[0]+(-0.976654446591)*x_ref[1]**o+(0.977779698697)*x_ref[1]+(0.694427679422)*x_ref[2]**o+(-0.364463986712)*x_ref[2]
        ref[0,2,1,0]=(-0.415861468352)*x_ref[0]**o+(-0.159802964908)*x_ref[0]+(-0.132331280778)*x_ref[1]**o+(0.120829594848)*x_ref[1]+(-0.114181951611)*x_ref[2]**o+(-0.810187313564)*x_ref[2]
        ref[0,2,1,1]=(-0.869925695451)*x_ref[0]**o+(-0.535444611026)*x_ref[0]+(0.62908837594)*x_ref[1]**o+(-0.488583607314)*x_ref[1]+(0.8461615049)*x_ref[2]**o+(-0.558010294727)*x_ref[2]
        ref[0,2,2,0]=(-0.0822232217466)*x_ref[0]**o+(-0.673631382141)*x_ref[0]+(0.803717792107)*x_ref[1]**o+(0.339934715615)*x_ref[1]+(-0.886365924996)*x_ref[2]**o+(0.0206517750753)*x_ref[2]
        ref[0,2,2,1]=(0.367656249869)*x_ref[0]**o+(0.000987963028539)*x_ref[0]+(0.399182215451)*x_ref[1]**o+(0.923259635138)*x_ref[1]+(0.8573338122)*x_ref[2]**o+(0.679082213415)*x_ref[2]
        ref[0,3,0,0]=(-0.722680496644)*x_ref[0]**o+(0.199427912651)*x_ref[0]+(-0.351799245099)*x_ref[1]**o+(0.240285416231)*x_ref[1]+(-0.701704868242)*x_ref[2]**o+(0.265610436018)*x_ref[2]
        ref[0,3,0,1]=(-0.14479043328)*x_ref[0]**o+(-0.192670959504)*x_ref[0]+(0.327412080296)*x_ref[1]**o+(-0.390894637576)*x_ref[1]+(-0.401862813677)*x_ref[2]**o+(-0.87150071561)*x_ref[2]
        ref[0,3,1,0]=(0.430039400239)*x_ref[0]**o+(-0.0691443536291)*x_ref[0]+(0.205492039071)*x_ref[1]**o+(-0.391580258314)*x_ref[1]+(0.791748125642)*x_ref[2]**o+(-0.0896722225431)*x_ref[2]
        ref[0,3,1,1]=(-0.177554766913)*x_ref[0]**o+(-0.399389195454)*x_ref[0]+(-0.304228953477)*x_ref[1]**o+(0.00462595697167)*x_ref[1]+(0.271814579884)*x_ref[2]**o+(-0.17305639368)*x_ref[2]
        ref[0,3,2,0]=(-0.105869589157)*x_ref[0]**o+(0.803488342924)*x_ref[0]+(-0.65117861655)*x_ref[1]**o+(-0.27529059731)*x_ref[1]+(-0.777107387582)*x_ref[2]**o+(-0.690980328369)*x_ref[2]
        ref[0,3,2,1]=(-0.13792230295)*x_ref[0]**o+(0.649721420509)*x_ref[0]+(0.401441756822)*x_ref[1]**o+(0.468949007973)*x_ref[1]+(-0.762015699412)*x_ref[2]**o+(0.389781367434)*x_ref[2]
        ref[0,4,0,0]=(0.873653819021)*x_ref[0]**o+(0.539406798913)*x_ref[0]+(0.802177801924)*x_ref[1]**o+(-0.861314989922)*x_ref[1]+(0.847509960118)*x_ref[2]**o+(-0.41381965291)*x_ref[2]
        ref[0,4,0,1]=(0.0416858647633)*x_ref[0]**o+(0.866825467268)*x_ref[0]+(0.89776163679)*x_ref[1]**o+(-0.978326218858)*x_ref[1]+(-0.409795413277)*x_ref[2]**o+(0.982449327977)*x_ref[2]
        ref[0,4,1,0]=(-0.114530966075)*x_ref[0]**o+(-0.182414280034)*x_ref[0]+(0.154385814218)*x_ref[1]**o+(0.869624332271)*x_ref[1]+(0.477213291128)*x_ref[2]**o+(-0.137123375612)*x_ref[2]
        ref[0,4,1,1]=(-0.52963019184)*x_ref[0]**o+(0.0217955878025)*x_ref[0]+(-0.174531798693)*x_ref[1]**o+(-0.738947178141)*x_ref[1]+(0.198130667502)*x_ref[2]**o+(-0.449952166994)*x_ref[2]
        ref[0,4,2,0]=(0.998209708224)*x_ref[0]**o+(0.734906841374)*x_ref[0]+(0.225596363429)*x_ref[1]**o+(0.862772101878)*x_ref[1]+(-0.133563132638)*x_ref[2]**o+(-0.78059821331)*x_ref[2]
        ref[0,4,2,1]=(0.384580893901)*x_ref[0]**o+(-0.881918368669)*x_ref[0]+(-0.950768236608)*x_ref[1]**o+(-0.452275072093)*x_ref[1]+(0.592779023744)*x_ref[2]**o+(0.798735695084)*x_ref[2]
        ref[1,0,0,0]=(0.786827894179)*x_ref[0]**o+(-0.233241290567)*x_ref[0]+(0.946717698992)*x_ref[1]**o+(0.179368479607)*x_ref[1]+(-0.10409968587)*x_ref[2]**o+(-0.770115022756)*x_ref[2]
        ref[1,0,0,1]=(-0.355389699793)*x_ref[0]**o+(0.356407960494)*x_ref[0]+(-0.371831767713)*x_ref[1]**o+(0.139120259823)*x_ref[1]+(0.910761795881)*x_ref[2]**o+(0.725966050775)*x_ref[2]
        ref[1,0,1,0]=(0.316372320948)*x_ref[0]**o+(0.432472869369)*x_ref[0]+(0.733187356771)*x_ref[1]**o+(0.133915829584)*x_ref[1]+(0.380713482112)*x_ref[2]**o+(-0.874827460114)*x_ref[2]
        ref[1,0,1,1]=(-0.0625786527083)*x_ref[0]**o+(-0.469518973019)*x_ref[0]+(-0.318610570327)*x_ref[1]**o+(-0.304412247656)*x_ref[1]+(0.844057224107)*x_ref[2]**o+(0.508926494536)*x_ref[2]
        ref[1,0,2,0]=(-0.759697281694)*x_ref[0]**o+(0.701456398418)*x_ref[0]+(-0.671612899124)*x_ref[1]**o+(0.292474197717)*x_ref[1]+(-0.445886723952)*x_ref[2]**o+(-0.319953406782)*x_ref[2]
        ref[1,0,2,1]=(-0.865703619781)*x_ref[0]**o+(0.115639820869)*x_ref[0]+(0.493634056223)*x_ref[1]**o+(0.0343245670418)*x_ref[1]+(0.748200872819)*x_ref[2]**o+(-0.208438395054)*x_ref[2]
        ref[1,1,0,0]=(0.767661930656)*x_ref[0]**o+(-0.578385615919)*x_ref[0]+(0.241313268557)*x_ref[1]**o+(0.74276605359)*x_ref[1]+(0.373638986644)*x_ref[2]**o+(0.475479466784)*x_ref[2]
        ref[1,1,0,1]=(0.723228882826)*x_ref[0]**o+(-0.988524686322)*x_ref[0]+(0.333664096597)*x_ref[1]**o+(-0.756435323592)*x_ref[1]+(0.771173510221)*x_ref[2]**o+(-0.879337556561)*x_ref[2]
        ref[1,1,1,0]=(0.939682105701)*x_ref[0]**o+(-0.172311578817)*x_ref[0]+(-0.407241492831)*x_ref[1]**o+(0.446660064834)*x_ref[1]+(0.0371203367053)*x_ref[2]**o+(-0.302853430561)*x_ref[2]
        ref[1,1,1,1]=(0.1093213068)*x_ref[0]**o+(0.31232933407)*x_ref[0]+(0.47474581408)*x_ref[1]**o+(-0.602763201317)*x_ref[1]+(0.803198523872)*x_ref[2]**o+(-0.178088233667)*x_ref[2]
        ref[1,1,2,0]=(-0.217141686114)*x_ref[0]**o+(0.287363815362)*x_ref[0]+(-0.352552947551)*x_ref[1]**o+(0.106445580983)*x_ref[1]+(0.340814562712)*x_ref[2]**o+(-0.963691249733)*x_ref[2]
        ref[1,1,2,1]=(0.612302241126)*x_ref[0]**o+(-0.771765095612)*x_ref[0]+(-0.18662763339)*x_ref[1]**o+(0.10259882275)*x_ref[1]+(-0.29535785913)*x_ref[2]**o+(-0.577876877465)*x_ref[2]
        ref[1,2,0,0]=(0.497131459554)*x_ref[0]**o+(0.912133134267)*x_ref[0]+(0.944369633113)*x_ref[1]**o+(0.163408823815)*x_ref[1]+(0.85636766551)*x_ref[2]**o+(-0.00282458078743)*x_ref[2]
        ref[1,2,0,1]=(0.362490806338)*x_ref[0]**o+(0.22490192994)*x_ref[0]+(-0.913665309243)*x_ref[1]**o+(-0.88010395132)*x_ref[1]+(-0.926205930269)*x_ref[2]**o+(-0.231830082111)*x_ref[2]
        ref[1,2,1,0]=(-0.974792784505)*x_ref[0]**o+(-0.765605618923)*x_ref[0]+(0.241470054253)*x_ref[1]**o+(0.445863114014)*x_ref[1]+(0.561062749944)*x_ref[2]**o+(-0.215674406206)*x_ref[2]
        ref[1,2,1,1]=(0.827013456823)*x_ref[0]**o+(-0.191864800977)*x_ref[0]+(-0.158099224794)*x_ref[1]**o+(-0.921959562711)*x_ref[1]+(0.930693535629)*x_ref[2]**o+(-0.352967183792)*x_ref[2]
        ref[1,2,2,0]=(0.258970603305)*x_ref[0]**o+(0.586779837275)*x_ref[0]+(0.82707087137)*x_ref[1]**o+(-0.215210535461)*x_ref[1]+(0.36890299705)*x_ref[2]**o+(-0.724385890088)*x_ref[2]
        ref[1,2,2,1]=(-0.125969269863)*x_ref[0]**o+(-0.505216539273)*x_ref[0]+(0.43242008053)*x_ref[1]**o+(0.212079268112)*x_ref[1]+(0.345759190477)*x_ref[2]**o+(-0.98505468199)*x_ref[2]
        ref[1,3,0,0]=(-0.647118215098)*x_ref[0]**o+(0.318566502446)*x_ref[0]+(-0.151406735047)*x_ref[1]**o+(0.253202745125)*x_ref[1]+(-0.404823111329)*x_ref[2]**o+(-0.520797366557)*x_ref[2]
        ref[1,3,0,1]=(-0.946886311645)*x_ref[0]**o+(0.733699524424)*x_ref[0]+(-0.615470980807)*x_ref[1]**o+(-0.103976314414)*x_ref[1]+(-0.932104431883)*x_ref[2]**o+(0.157563806559)*x_ref[2]
        ref[1,3,1,0]=(-0.289471503181)*x_ref[0]**o+(0.749124853135)*x_ref[0]+(-0.578861403092)*x_ref[1]**o+(-0.136231378985)*x_ref[1]+(0.532719246315)*x_ref[2]**o+(0.542455831657)*x_ref[2]
        ref[1,3,1,1]=(-0.343842490684)*x_ref[0]**o+(0.110039182926)*x_ref[0]+(0.901234650534)*x_ref[1]**o+(-0.513362876241)*x_ref[1]+(0.371952189126)*x_ref[2]**o+(-0.169400349862)*x_ref[2]
        ref[1,3,2,0]=(-0.766791944503)*x_ref[0]**o+(0.0267491122166)*x_ref[0]+(-0.699293509597)*x_ref[1]**o+(-0.697363669971)*x_ref[1]+(0.567854170308)*x_ref[2]**o+(-0.516301389629)*x_ref[2]
        ref[1,3,2,1]=(-0.288508579358)*x_ref[0]**o+(0.239321880772)*x_ref[0]+(0.677249635795)*x_ref[1]**o+(0.48369084768)*x_ref[1]+(0.681460633595)*x_ref[2]**o+(-0.80218734437)*x_ref[2]
        ref[1,4,0,0]=(0.510273388936)*x_ref[0]**o+(0.35600020965)*x_ref[0]+(0.271077500616)*x_ref[1]**o+(-0.00479335810576)*x_ref[1]+(-0.348457583464)*x_ref[2]**o+(-0.965895122152)*x_ref[2]
        ref[1,4,0,1]=(0.0134532748813)*x_ref[0]**o+(-0.396954380686)*x_ref[0]+(-0.176682002946)*x_ref[1]**o+(0.971224261138)*x_ref[1]+(-0.359602180261)*x_ref[2]**o+(0.373748703363)*x_ref[2]
        ref[1,4,1,0]=(-0.519413160925)*x_ref[0]**o+(0.856346255965)*x_ref[0]+(0.159167243935)*x_ref[1]**o+(0.92068779838)*x_ref[1]+(0.415122182947)*x_ref[2]**o+(-0.595702311496)*x_ref[2]
        ref[1,4,1,1]=(-0.608969598184)*x_ref[0]**o+(-0.434750765676)*x_ref[0]+(0.74471475701)*x_ref[1]**o+(-0.229333246615)*x_ref[1]+(-0.678138032736)*x_ref[2]**o+(-0.514112926798)*x_ref[2]
        ref[1,4,2,0]=(0.0633453561735)*x_ref[0]**o+(-0.278569473082)*x_ref[0]+(-0.842227144473)*x_ref[1]**o+(-0.00289799815652)*x_ref[1]+(0.153392291113)*x_ref[2]**o+(-0.61448940993)*x_ref[2]
        ref[1,4,2,1]=(0.0897863401674)*x_ref[0]**o+(-0.487876787706)*x_ref[0]+(-0.464008634014)*x_ref[1]**o+(0.576466896344)*x_ref[1]+(0.962345361537)*x_ref[2]**o+(0.580878819156)*x_ref[2]
        ref[2,0,0,0]=(0.316582275434)*x_ref[0]**o+(0.0607872538221)*x_ref[0]+(0.642492366848)*x_ref[1]**o+(0.200300167843)*x_ref[1]+(-0.402802396184)*x_ref[2]**o+(-0.834254866233)*x_ref[2]
        ref[2,0,0,1]=(0.644614184272)*x_ref[0]**o+(-0.086367798694)*x_ref[0]+(-0.428671462705)*x_ref[1]**o+(-0.517827373575)*x_ref[1]+(0.528197945165)*x_ref[2]**o+(-0.523092888005)*x_ref[2]
        ref[2,0,1,0]=(0.41376145869)*x_ref[0]**o+(-0.27425322418)*x_ref[0]+(0.816942168928)*x_ref[1]**o+(-0.882784454814)*x_ref[1]+(-0.270205535673)*x_ref[2]**o+(0.642372102899)*x_ref[2]
        ref[2,0,1,1]=(0.584931736844)*x_ref[0]**o+(0.622511792702)*x_ref[0]+(0.76174625736)*x_ref[1]**o+(0.0561486448557)*x_ref[1]+(-0.588861269552)*x_ref[2]**o+(-0.341896323569)*x_ref[2]
        ref[2,0,2,0]=(0.635215108566)*x_ref[0]**o+(0.299769133671)*x_ref[0]+(0.141053572388)*x_ref[1]**o+(0.102844255879)*x_ref[1]+(-0.995154777534)*x_ref[2]**o+(0.501319388994)*x_ref[2]
        ref[2,0,2,1]=(0.750089643836)*x_ref[0]**o+(0.57570708561)*x_ref[0]+(-0.449935823319)*x_ref[1]**o+(-0.902982109411)*x_ref[1]+(0.561185740222)*x_ref[2]**o+(0.422654056847)*x_ref[2]
        ref[2,1,0,0]=(0.953986415316)*x_ref[0]**o+(-0.7883326473)*x_ref[0]+(-0.723690234059)*x_ref[1]**o+(-0.537242781471)*x_ref[1]+(-0.503466717519)*x_ref[2]**o+(-0.777865169241)*x_ref[2]
        ref[2,1,0,1]=(0.317357554441)*x_ref[0]**o+(-0.326828957855)*x_ref[0]+(-0.609992651237)*x_ref[1]**o+(0.811066934743)*x_ref[1]+(-0.90668441447)*x_ref[2]**o+(0.92581926687)*x_ref[2]
        ref[2,1,1,0]=(-0.0801236189368)*x_ref[0]**o+(0.49151264813)*x_ref[0]+(-0.0770654446151)*x_ref[1]**o+(-0.314981217276)*x_ref[1]+(-0.743180837478)*x_ref[2]**o+(-0.554491413146)*x_ref[2]
        ref[2,1,1,1]=(0.831017344937)*x_ref[0]**o+(-0.61064793935)*x_ref[0]+(0.652976737027)*x_ref[1]**o+(0.242104127969)*x_ref[1]+(0.593330958678)*x_ref[2]**o+(-0.35899584298)*x_ref[2]
        ref[2,1,2,0]=(-0.580618798556)*x_ref[0]**o+(-0.936176474082)*x_ref[0]+(-0.768716477288)*x_ref[1]**o+(-0.353687473159)*x_ref[1]+(-0.261509314019)*x_ref[2]**o+(-0.442139199257)*x_ref[2]
        ref[2,1,2,1]=(0.945760793889)*x_ref[0]**o+(-0.646958670891)*x_ref[0]+(-0.440341726063)*x_ref[1]**o+(0.223283751014)*x_ref[1]+(0.799798944743)*x_ref[2]**o+(0.591902067633)*x_ref[2]
        ref[2,2,0,0]=(-0.267699746405)*x_ref[0]**o+(0.271554172399)*x_ref[0]+(0.747278903642)*x_ref[1]**o+(-0.832877459834)*x_ref[1]+(-0.640373164716)*x_ref[2]**o+(-0.666487164763)*x_ref[2]
        ref[2,2,0,1]=(0.795630715924)*x_ref[0]**o+(0.682973866186)*x_ref[0]+(-0.750818874926)*x_ref[1]**o+(-0.904782408217)*x_ref[1]+(-0.578138718379)*x_ref[2]**o+(-0.369846516992)*x_ref[2]
        ref[2,2,1,0]=(0.165654497062)*x_ref[0]**o+(-0.329827871165)*x_ref[0]+(-0.808981530672)*x_ref[1]**o+(0.350106763785)*x_ref[1]+(0.106638087689)*x_ref[2]**o+(0.425516820398)*x_ref[2]
        ref[2,2,1,1]=(0.017419119421)*x_ref[0]**o+(0.500359047996)*x_ref[0]+(-0.888509940049)*x_ref[1]**o+(-0.375469124891)*x_ref[1]+(0.863718970618)*x_ref[2]**o+(-0.807667604763)*x_ref[2]
        ref[2,2,2,0]=(0.860662652165)*x_ref[0]**o+(0.162530730499)*x_ref[0]+(-0.657510966961)*x_ref[1]**o+(0.215905840286)*x_ref[1]+(0.97380371264)*x_ref[2]**o+(-0.431027671897)*x_ref[2]
        ref[2,2,2,1]=(-0.649151503991)*x_ref[0]**o+(0.484589779517)*x_ref[0]+(0.905080495553)*x_ref[1]**o+(0.997504119204)*x_ref[1]+(-0.4268802512)*x_ref[2]**o+(-0.687236308612)*x_ref[2]
        ref[2,3,0,0]=(-0.60485902589)*x_ref[0]**o+(0.49500777286)*x_ref[0]+(0.33977532104)*x_ref[1]**o+(-0.219209845338)*x_ref[1]+(-0.197311627683)*x_ref[2]**o+(0.285859591044)*x_ref[2]
        ref[2,3,0,1]=(0.117438340483)*x_ref[0]**o+(-0.562954894877)*x_ref[0]+(-0.302982842797)*x_ref[1]**o+(0.977923453436)*x_ref[1]+(0.300127749784)*x_ref[2]**o+(0.45784201373)*x_ref[2]
        ref[2,3,1,0]=(-0.910061442266)*x_ref[0]**o+(0.723114553174)*x_ref[0]+(-0.435693478652)*x_ref[1]**o+(-0.584326584996)*x_ref[1]+(-0.51000911177)*x_ref[2]**o+(0.506615483889)*x_ref[2]
        ref[2,3,1,1]=(-0.50179741763)*x_ref[0]**o+(-0.954551290001)*x_ref[0]+(0.375839266192)*x_ref[1]**o+(0.122120198059)*x_ref[1]+(-0.593050674641)*x_ref[2]**o+(0.549989594312)*x_ref[2]
        ref[2,3,2,0]=(-0.889519666705)*x_ref[0]**o+(-0.815572507674)*x_ref[0]+(0.453677651057)*x_ref[1]**o+(0.450910828217)*x_ref[1]+(-0.00465373008682)*x_ref[2]**o+(0.731899900165)*x_ref[2]
        ref[2,3,2,1]=(-0.969971151589)*x_ref[0]**o+(-0.156948104522)*x_ref[0]+(0.492885029353)*x_ref[1]**o+(-0.0310677847833)*x_ref[1]+(0.327614233837)*x_ref[2]**o+(-0.274974040707)*x_ref[2]
        ref[2,4,0,0]=(0.844667900826)*x_ref[0]**o+(0.863374870522)*x_ref[0]+(0.251456923241)*x_ref[1]**o+(0.434916005183)*x_ref[1]+(-0.139023370677)*x_ref[2]**o+(0.0655176616053)*x_ref[2]
        ref[2,4,0,1]=(0.831755059318)*x_ref[0]**o+(0.0897642674136)*x_ref[0]+(0.190414878812)*x_ref[1]**o+(0.976565327031)*x_ref[1]+(0.413321767371)*x_ref[2]**o+(0.755395415341)*x_ref[2]
        ref[2,4,1,0]=(-0.790364876377)*x_ref[0]**o+(0.829442094248)*x_ref[0]+(0.556251403891)*x_ref[1]**o+(-0.953487711658)*x_ref[1]+(-0.610781905401)*x_ref[2]**o+(-0.320764072326)*x_ref[2]
        ref[2,4,1,1]=(-0.757101565136)*x_ref[0]**o+(0.789340839395)*x_ref[0]+(-0.615189514438)*x_ref[1]**o+(-0.910992676978)*x_ref[1]+(0.693736813401)*x_ref[2]**o+(0.0538707588266)*x_ref[2]
        ref[2,4,2,0]=(0.000335625616611)*x_ref[0]**o+(-0.195171446476)*x_ref[0]+(0.0281149133705)*x_ref[1]**o+(-0.362294159243)*x_ref[1]+(0.830520288743)*x_ref[2]**o+(-0.431967694247)*x_ref[2]
        ref[2,4,2,1]=(-0.232668364843)*x_ref[0]**o+(0.655302452302)*x_ref[0]+(-0.0125806408337)*x_ref[1]**o+(-0.103979202109)*x_ref[1]+(-0.277450383033)*x_ref[2]**o+(0.708914971885)*x_ref[2]
        ref[3,0,0,0]=(-0.813447875186)*x_ref[0]**o+(0.133491078101)*x_ref[0]+(-0.675510725453)*x_ref[1]**o+(-0.161489202717)*x_ref[1]+(0.0474778506607)*x_ref[2]**o+(0.513729029652)*x_ref[2]
        ref[3,0,0,1]=(0.45388386743)*x_ref[0]**o+(-0.909074944715)*x_ref[0]+(-0.540576382376)*x_ref[1]**o+(-0.849753706068)*x_ref[1]+(-0.496320191697)*x_ref[2]**o+(-0.930725196363)*x_ref[2]
        ref[3,0,1,0]=(-0.145526830518)*x_ref[0]**o+(-0.34717641097)*x_ref[0]+(-0.818886792022)*x_ref[1]**o+(0.574315644784)*x_ref[1]+(-0.780431329222)*x_ref[2]**o+(0.892840747651)*x_ref[2]
        ref[3,0,1,1]=(-0.415342962745)*x_ref[0]**o+(-0.899768945518)*x_ref[0]+(0.798573910731)*x_ref[1]**o+(-0.0652893933479)*x_ref[1]+(-0.444459519602)*x_ref[2]**o+(-0.0100131702565)*x_ref[2]
        ref[3,0,2,0]=(-0.997936676013)*x_ref[0]**o+(0.822000858069)*x_ref[0]+(-0.553794408949)*x_ref[1]**o+(0.829037287423)*x_ref[1]+(-0.700288437806)*x_ref[2]**o+(-0.449584600891)*x_ref[2]
        ref[3,0,2,1]=(-0.70979868598)*x_ref[0]**o+(0.689868687655)*x_ref[0]+(0.151257146397)*x_ref[1]**o+(0.32895215144)*x_ref[1]+(0.286043871383)*x_ref[2]**o+(-0.604421567781)*x_ref[2]
        ref[3,1,0,0]=(0.60431849542)*x_ref[0]**o+(0.518618291627)*x_ref[0]+(-0.471868097013)*x_ref[1]**o+(-0.639110899151)*x_ref[1]+(-0.0904184061601)*x_ref[2]**o+(0.113762528304)*x_ref[2]
        ref[3,1,0,1]=(0.307624103157)*x_ref[0]**o+(-0.388281302162)*x_ref[0]+(0.772429578102)*x_ref[1]**o+(-0.197799629585)*x_ref[1]+(-0.149120667036)*x_ref[2]**o+(0.830154780264)*x_ref[2]
        ref[3,1,1,0]=(0.957127413133)*x_ref[0]**o+(0.730080945389)*x_ref[0]+(0.497191948774)*x_ref[1]**o+(-0.586439221663)*x_ref[1]+(-0.56223071093)*x_ref[2]**o+(0.162899395804)*x_ref[2]
        ref[3,1,1,1]=(-0.285740818451)*x_ref[0]**o+(-0.573919970411)*x_ref[0]+(-0.400833211669)*x_ref[1]**o+(0.84414534634)*x_ref[1]+(-0.124352096731)*x_ref[2]**o+(0.907013800834)*x_ref[2]
        ref[3,1,2,0]=(0.186947347989)*x_ref[0]**o+(0.499582449213)*x_ref[0]+(0.525473636695)*x_ref[1]**o+(-0.926500953998)*x_ref[1]+(-0.0392615366415)*x_ref[2]**o+(0.909002911192)*x_ref[2]
        ref[3,1,2,1]=(0.634350440683)*x_ref[0]**o+(0.901699291665)*x_ref[0]+(-0.683899667575)*x_ref[1]**o+(0.411642037744)*x_ref[1]+(-0.72791611585)*x_ref[2]**o+(-0.013293689839)*x_ref[2]
        ref[3,2,0,0]=(0.185164429507)*x_ref[0]**o+(-0.671011652257)*x_ref[0]+(0.973361040829)*x_ref[1]**o+(0.989738565899)*x_ref[1]+(-0.501245887091)*x_ref[2]**o+(-0.405445444386)*x_ref[2]
        ref[3,2,0,1]=(-0.311554020884)*x_ref[0]**o+(-0.929020990333)*x_ref[0]+(0.366016375797)*x_ref[1]**o+(0.838234804773)*x_ref[1]+(0.939602632409)*x_ref[2]**o+(-0.885524260372)*x_ref[2]
        ref[3,2,1,0]=(-0.0101604655589)*x_ref[0]**o+(0.670274078493)*x_ref[0]+(0.198135855354)*x_ref[1]**o+(-0.000275422963644)*x_ref[1]+(0.136674233404)*x_ref[2]**o+(-0.442510118497)*x_ref[2]
        ref[3,2,1,1]=(-0.677948769785)*x_ref[0]**o+(-0.165403610813)*x_ref[0]+(-0.219353502981)*x_ref[1]**o+(0.664297154009)*x_ref[1]+(0.158536302892)*x_ref[2]**o+(0.293940884925)*x_ref[2]
        ref[3,2,2,0]=(0.666669534708)*x_ref[0]**o+(0.609459298502)*x_ref[0]+(0.297058070582)*x_ref[1]**o+(-0.0809555825821)*x_ref[1]+(0.139430791042)*x_ref[2]**o+(-0.253514861624)*x_ref[2]
        ref[3,2,2,1]=(-0.577932413274)*x_ref[0]**o+(-0.583353108895)*x_ref[0]+(-0.839262034394)*x_ref[1]**o+(0.795642236106)*x_ref[1]+(0.0882627381543)*x_ref[2]**o+(-0.382970623122)*x_ref[2]
        ref[3,3,0,0]=(0.187941627429)*x_ref[0]**o+(0.552040688531)*x_ref[0]+(0.837458832716)*x_ref[1]**o+(0.296793121897)*x_ref[1]+(-0.665699511216)*x_ref[2]**o+(-0.318565991052)*x_ref[2]
        ref[3,3,0,1]=(-0.78770176804)*x_ref[0]**o+(0.807509260396)*x_ref[0]+(0.46241086085)*x_ref[1]**o+(0.668506908751)*x_ref[1]+(-0.988954457343)*x_ref[2]**o+(0.0896365679915)*x_ref[2]
        ref[3,3,1,0]=(0.727665746433)*x_ref[0]**o+(-0.819391676739)*x_ref[0]+(-0.118648934607)*x_ref[1]**o+(-0.433862939031)*x_ref[1]+(0.827788711313)*x_ref[2]**o+(-0.615751446784)*x_ref[2]
        ref[3,3,1,1]=(-0.00427901841482)*x_ref[0]**o+(-0.38545713036)*x_ref[0]+(0.570160083404)*x_ref[1]**o+(-0.235505162795)*x_ref[1]+(0.514578304371)*x_ref[2]**o+(-0.152692113279)*x_ref[2]
        ref[3,3,2,0]=(-0.396900613831)*x_ref[0]**o+(0.161875314543)*x_ref[0]+(0.28471295984)*x_ref[1]**o+(-0.623276596728)*x_ref[1]+(-0.748772373219)*x_ref[2]**o+(0.584246898704)*x_ref[2]
        ref[3,3,2,1]=(0.210026027009)*x_ref[0]**o+(-0.530437530387)*x_ref[0]+(0.437155825716)*x_ref[1]**o+(0.984102407204)*x_ref[1]+(0.125030773605)*x_ref[2]**o+(-0.821764907318)*x_ref[2]
        ref[3,4,0,0]=(-0.819129946783)*x_ref[0]**o+(-0.729517565451)*x_ref[0]+(-0.126013244602)*x_ref[1]**o+(0.592832125254)*x_ref[1]+(-0.850426487488)*x_ref[2]**o+(0.271550912174)*x_ref[2]
        ref[3,4,0,1]=(-0.573554440921)*x_ref[0]**o+(0.232106829584)*x_ref[0]+(0.810201614042)*x_ref[1]**o+(-0.85518164211)*x_ref[1]+(0.249535563949)*x_ref[2]**o+(-0.69728963017)*x_ref[2]
        ref[3,4,1,0]=(0.785661643628)*x_ref[0]**o+(0.936424600424)*x_ref[0]+(0.321363811055)*x_ref[1]**o+(0.362341458592)*x_ref[1]+(-0.593576579363)*x_ref[2]**o+(0.0601278435924)*x_ref[2]
        ref[3,4,1,1]=(-0.854420686885)*x_ref[0]**o+(0.993352097501)*x_ref[0]+(-0.572116112224)*x_ref[1]**o+(0.308008845182)*x_ref[1]+(0.0443339578639)*x_ref[2]**o+(-0.1469857285)*x_ref[2]
        ref[3,4,2,0]=(0.696373642151)*x_ref[0]**o+(0.82425126832)*x_ref[0]+(-0.0891735358455)*x_ref[1]**o+(-0.22218993613)*x_ref[1]+(-0.238393149017)*x_ref[2]**o+(-0.631352718194)*x_ref[2]
        ref[3,4,2,1]=(-0.0932690431928)*x_ref[0]**o+(-0.391900483919)*x_ref[0]+(0.236843565361)*x_ref[1]**o+(-0.164197648532)*x_ref[1]+(0.235502596353)*x_ref[2]**o+(0.0110131867917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.521673635599)*x[0]**o+(-0.288664980886)*x[0]+(0.57565949318)*x[1]**o+(-0.555613874185)*x[1]
        ref=(-0.521673635599)*x_ref[0]**o+(-0.288664980886)*x_ref[0]+(0.57565949318)*x_ref[1]**o+(-0.555613874185)*x_ref[1]
      else:
        arg=(0.520429804914)*x[0]**o+(-0.000683147004841)*x[0]+(0.676879975826)*x[1]**o+(0.760707578765)*x[1]+(0.479225734268)*x[2]**o+(0.132006876687)*x[2]
        ref=(0.520429804914)*x_ref[0]**o+(-0.000683147004841)*x_ref[0]+(0.676879975826)*x_ref[1]**o+(0.760707578765)*x_ref[1]+(0.479225734268)*x_ref[2]**o+(0.132006876687)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.350515208335)*x[0]**o+(-0.604596801095)*x[0]+(0.145267491062)*x[1]**o+(-0.422271555369)*x[1]
        ref=(-0.350515208335)*x_ref[0]**o+(-0.604596801095)*x_ref[0]+(0.145267491062)*x_ref[1]**o+(-0.422271555369)*x_ref[1]
      else:
        arg=(-0.162304501731)*x[0]**o+(0.342262454165)*x[0]+(0.693988751897)*x[1]**o+(-0.315225717754)*x[1]+(-0.938405302425)*x[2]**o+(-0.0929881908885)*x[2]
        ref=(-0.162304501731)*x_ref[0]**o+(0.342262454165)*x_ref[0]+(0.693988751897)*x_ref[1]**o+(-0.315225717754)*x_ref[1]+(-0.938405302425)*x_ref[2]**o+(-0.0929881908885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.104854565192)*x[0]**o+(-0.596688162793)*x[0]+(-0.0286471957779)*x[1]**o+(-0.157282508059)*x[1]
        arg[1]=(0.619817759027)*x[0]**o+(-0.832815965127)*x[0]+(-0.638126497014)*x[1]**o+(0.771271149341)*x[1]
        ref[0]=(0.104854565192)*x_ref[0]**o+(-0.596688162793)*x_ref[0]+(-0.0286471957779)*x_ref[1]**o+(-0.157282508059)*x_ref[1]
        ref[1]=(0.619817759027)*x_ref[0]**o+(-0.832815965127)*x_ref[0]+(-0.638126497014)*x_ref[1]**o+(0.771271149341)*x_ref[1]
      else:
        arg[0]=(-0.358142932888)*x[0]**o+(0.758685606027)*x[0]+(-0.0716050574096)*x[1]**o+(-0.723587074286)*x[1]+(0.387240173671)*x[2]**o+(0.544679811421)*x[2]
        arg[1]=(0.864154242034)*x[0]**o+(-0.0740598679188)*x[0]+(-0.0865611132887)*x[1]**o+(0.0414815431481)*x[1]+(-0.897395775334)*x[2]**o+(-0.580401889318)*x[2]
        ref[0]=(-0.358142932888)*x_ref[0]**o+(0.758685606027)*x_ref[0]+(-0.0716050574096)*x_ref[1]**o+(-0.723587074286)*x_ref[1]+(0.387240173671)*x_ref[2]**o+(0.544679811421)*x_ref[2]
        ref[1]=(0.864154242034)*x_ref[0]**o+(-0.0740598679188)*x_ref[0]+(-0.0865611132887)*x_ref[1]**o+(0.0414815431481)*x_ref[1]+(-0.897395775334)*x_ref[2]**o+(-0.580401889318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.0571385613076)*x[0]**o+(-0.296048600622)*x[0]+(0.300914212927)*x[1]**o+(0.81578919689)*x[1]
        arg[0,1]=(0.136796732287)*x[0]**o+(0.169384864236)*x[0]+(-0.874585047899)*x[1]**o+(-0.67613629157)*x[1]
        arg[0,2]=(-0.947048556597)*x[0]**o+(-0.807647782941)*x[0]+(-0.53233790031)*x[1]**o+(0.618861393596)*x[1]
        arg[0,3]=(-0.533009370674)*x[0]**o+(-0.732258917532)*x[0]+(0.735058784793)*x[1]**o+(0.313284184128)*x[1]
        arg[0,4]=(0.678291805831)*x[0]**o+(0.892272189132)*x[0]+(-0.902740106267)*x[1]**o+(0.902593764933)*x[1]
        arg[1,0]=(0.101446968367)*x[0]**o+(0.758270058166)*x[0]+(-0.511416064913)*x[1]**o+(-0.464147847068)*x[1]
        arg[1,1]=(-0.0131419066862)*x[0]**o+(-0.942460391354)*x[0]+(0.479955107486)*x[1]**o+(-0.311304181244)*x[1]
        arg[1,2]=(0.623958732361)*x[0]**o+(-0.313196338837)*x[0]+(0.413428996617)*x[1]**o+(0.348327566672)*x[1]
        arg[1,3]=(-0.8413636876)*x[0]**o+(-0.572573098032)*x[0]+(-0.678663954356)*x[1]**o+(-0.419716403314)*x[1]
        arg[1,4]=(-0.580466216486)*x[0]**o+(0.384226238557)*x[0]+(-0.732722613215)*x[1]**o+(-0.0238970013828)*x[1]
        arg[2,0]=(-0.248986933717)*x[0]**o+(0.264019654506)*x[0]+(0.0314811616292)*x[1]**o+(0.631481081422)*x[1]
        arg[2,1]=(0.643576426484)*x[0]**o+(-0.313599735242)*x[0]+(0.908780876791)*x[1]**o+(-0.698972268201)*x[1]
        arg[2,2]=(-0.623208321993)*x[0]**o+(0.649637800083)*x[0]+(-0.878484988208)*x[1]**o+(0.368793479691)*x[1]
        arg[2,3]=(0.239333407828)*x[0]**o+(-0.293558915555)*x[0]+(0.271630238733)*x[1]**o+(0.741304510722)*x[1]
        arg[2,4]=(0.175914942104)*x[0]**o+(0.336116863935)*x[0]+(-0.799168065807)*x[1]**o+(0.0264962830729)*x[1]
        arg[3,0]=(0.777665875019)*x[0]**o+(-0.13232137001)*x[0]+(0.682046389381)*x[1]**o+(0.589069623109)*x[1]
        arg[3,1]=(0.7607621575)*x[0]**o+(-0.81285460551)*x[0]+(0.0523010659976)*x[1]**o+(-0.913765717778)*x[1]
        arg[3,2]=(0.822107753118)*x[0]**o+(0.508869893322)*x[0]+(0.679583348422)*x[1]**o+(-0.597814039932)*x[1]
        arg[3,3]=(-0.77699176705)*x[0]**o+(-0.267758166676)*x[0]+(0.661073127871)*x[1]**o+(-0.41434545147)*x[1]
        arg[3,4]=(0.223013380161)*x[0]**o+(-0.558077040245)*x[0]+(-0.450831764747)*x[1]**o+(-0.0897765635191)*x[1]
        ref[0,0]=(-0.0571385613076)*x_ref[0]**o+(-0.296048600622)*x_ref[0]+(0.300914212927)*x_ref[1]**o+(0.81578919689)*x_ref[1]
        ref[0,1]=(0.136796732287)*x_ref[0]**o+(0.169384864236)*x_ref[0]+(-0.874585047899)*x_ref[1]**o+(-0.67613629157)*x_ref[1]
        ref[0,2]=(-0.947048556597)*x_ref[0]**o+(-0.807647782941)*x_ref[0]+(-0.53233790031)*x_ref[1]**o+(0.618861393596)*x_ref[1]
        ref[0,3]=(-0.533009370674)*x_ref[0]**o+(-0.732258917532)*x_ref[0]+(0.735058784793)*x_ref[1]**o+(0.313284184128)*x_ref[1]
        ref[0,4]=(0.678291805831)*x_ref[0]**o+(0.892272189132)*x_ref[0]+(-0.902740106267)*x_ref[1]**o+(0.902593764933)*x_ref[1]
        ref[1,0]=(0.101446968367)*x_ref[0]**o+(0.758270058166)*x_ref[0]+(-0.511416064913)*x_ref[1]**o+(-0.464147847068)*x_ref[1]
        ref[1,1]=(-0.0131419066862)*x_ref[0]**o+(-0.942460391354)*x_ref[0]+(0.479955107486)*x_ref[1]**o+(-0.311304181244)*x_ref[1]
        ref[1,2]=(0.623958732361)*x_ref[0]**o+(-0.313196338837)*x_ref[0]+(0.413428996617)*x_ref[1]**o+(0.348327566672)*x_ref[1]
        ref[1,3]=(-0.8413636876)*x_ref[0]**o+(-0.572573098032)*x_ref[0]+(-0.678663954356)*x_ref[1]**o+(-0.419716403314)*x_ref[1]
        ref[1,4]=(-0.580466216486)*x_ref[0]**o+(0.384226238557)*x_ref[0]+(-0.732722613215)*x_ref[1]**o+(-0.0238970013828)*x_ref[1]
        ref[2,0]=(-0.248986933717)*x_ref[0]**o+(0.264019654506)*x_ref[0]+(0.0314811616292)*x_ref[1]**o+(0.631481081422)*x_ref[1]
        ref[2,1]=(0.643576426484)*x_ref[0]**o+(-0.313599735242)*x_ref[0]+(0.908780876791)*x_ref[1]**o+(-0.698972268201)*x_ref[1]
        ref[2,2]=(-0.623208321993)*x_ref[0]**o+(0.649637800083)*x_ref[0]+(-0.878484988208)*x_ref[1]**o+(0.368793479691)*x_ref[1]
        ref[2,3]=(0.239333407828)*x_ref[0]**o+(-0.293558915555)*x_ref[0]+(0.271630238733)*x_ref[1]**o+(0.741304510722)*x_ref[1]
        ref[2,4]=(0.175914942104)*x_ref[0]**o+(0.336116863935)*x_ref[0]+(-0.799168065807)*x_ref[1]**o+(0.0264962830729)*x_ref[1]
        ref[3,0]=(0.777665875019)*x_ref[0]**o+(-0.13232137001)*x_ref[0]+(0.682046389381)*x_ref[1]**o+(0.589069623109)*x_ref[1]
        ref[3,1]=(0.7607621575)*x_ref[0]**o+(-0.81285460551)*x_ref[0]+(0.0523010659976)*x_ref[1]**o+(-0.913765717778)*x_ref[1]
        ref[3,2]=(0.822107753118)*x_ref[0]**o+(0.508869893322)*x_ref[0]+(0.679583348422)*x_ref[1]**o+(-0.597814039932)*x_ref[1]
        ref[3,3]=(-0.77699176705)*x_ref[0]**o+(-0.267758166676)*x_ref[0]+(0.661073127871)*x_ref[1]**o+(-0.41434545147)*x_ref[1]
        ref[3,4]=(0.223013380161)*x_ref[0]**o+(-0.558077040245)*x_ref[0]+(-0.450831764747)*x_ref[1]**o+(-0.0897765635191)*x_ref[1]
      else:
        arg[0,0]=(0.487697656095)*x[0]**o+(0.362509734966)*x[0]+(-0.40438484582)*x[1]**o+(-0.358984481681)*x[1]+(0.157670938098)*x[2]**o+(-0.631406375093)*x[2]
        arg[0,1]=(-0.00251154956677)*x[0]**o+(0.671447769464)*x[0]+(0.535541351526)*x[1]**o+(0.433508112485)*x[1]+(-0.897668718638)*x[2]**o+(0.141331586102)*x[2]
        arg[0,2]=(-0.964778505658)*x[0]**o+(0.522815935313)*x[0]+(-0.750318077186)*x[1]**o+(0.0595240411196)*x[1]+(-0.384021192511)*x[2]**o+(-0.802529657745)*x[2]
        arg[0,3]=(0.264781810221)*x[0]**o+(-0.654277693956)*x[0]+(0.890264421741)*x[1]**o+(-0.657743774786)*x[1]+(0.320234835664)*x[2]**o+(0.0637070094165)*x[2]
        arg[0,4]=(0.712229359775)*x[0]**o+(-0.646586719826)*x[0]+(0.876519277606)*x[1]**o+(0.473649630238)*x[1]+(0.332259187685)*x[2]**o+(-0.331489699179)*x[2]
        arg[1,0]=(0.808534274392)*x[0]**o+(-0.296462057802)*x[0]+(0.152536716655)*x[1]**o+(-0.990267302005)*x[1]+(-0.823462709316)*x[2]**o+(-0.51608634537)*x[2]
        arg[1,1]=(-0.745238552081)*x[0]**o+(-0.892086591023)*x[0]+(0.691429994644)*x[1]**o+(-0.52538570831)*x[1]+(-0.668959934842)*x[2]**o+(-0.5178179109)*x[2]
        arg[1,2]=(0.10173204456)*x[0]**o+(-0.853001564748)*x[0]+(0.617252846191)*x[1]**o+(-0.518228673297)*x[1]+(-0.995447484095)*x[2]**o+(0.56679145486)*x[2]
        arg[1,3]=(0.285145125389)*x[0]**o+(0.773362877732)*x[0]+(-0.457364007088)*x[1]**o+(0.226257217659)*x[1]+(0.439832733067)*x[2]**o+(0.872812128578)*x[2]
        arg[1,4]=(-0.581535304023)*x[0]**o+(-0.944599120493)*x[0]+(-0.720009453699)*x[1]**o+(-0.29624047891)*x[1]+(0.966528522407)*x[2]**o+(0.818945905214)*x[2]
        arg[2,0]=(-0.650732348774)*x[0]**o+(0.368909306978)*x[0]+(-0.608327978493)*x[1]**o+(0.301600904054)*x[1]+(-0.747648472883)*x[2]**o+(0.846568601061)*x[2]
        arg[2,1]=(-0.0221803961531)*x[0]**o+(-0.858842969825)*x[0]+(-0.40141447334)*x[1]**o+(0.408164536045)*x[1]+(-0.0131899768237)*x[2]**o+(-0.0983376041104)*x[2]
        arg[2,2]=(0.0855129797683)*x[0]**o+(0.0645200342375)*x[0]+(0.295916169508)*x[1]**o+(0.343169626483)*x[1]+(0.973477621396)*x[2]**o+(0.179505855846)*x[2]
        arg[2,3]=(-0.115714157482)*x[0]**o+(-0.00785628190789)*x[0]+(0.406498238895)*x[1]**o+(-0.954353977843)*x[1]+(0.751544148524)*x[2]**o+(-0.092307218389)*x[2]
        arg[2,4]=(0.807583813259)*x[0]**o+(-0.0150562208564)*x[0]+(0.967290753425)*x[1]**o+(0.960049095802)*x[1]+(0.41935259669)*x[2]**o+(0.878823415691)*x[2]
        arg[3,0]=(-0.162221321908)*x[0]**o+(0.0142404015457)*x[0]+(0.769189546147)*x[1]**o+(-0.251240080116)*x[1]+(0.980441953151)*x[2]**o+(-0.183639594778)*x[2]
        arg[3,1]=(0.270231471303)*x[0]**o+(-0.527471406108)*x[0]+(-0.255144134035)*x[1]**o+(0.268061599615)*x[1]+(-0.396373327005)*x[2]**o+(0.105905036411)*x[2]
        arg[3,2]=(0.5159632953)*x[0]**o+(-0.87397274133)*x[0]+(0.493876004047)*x[1]**o+(0.320996363656)*x[1]+(0.884345148533)*x[2]**o+(0.116262647026)*x[2]
        arg[3,3]=(-0.354053396048)*x[0]**o+(-0.130121845174)*x[0]+(-0.33114435645)*x[1]**o+(-0.73998189126)*x[1]+(0.362417530484)*x[2]**o+(0.676820875084)*x[2]
        arg[3,4]=(-0.53273881792)*x[0]**o+(-0.0624208514424)*x[0]+(0.680486778939)*x[1]**o+(-0.370666453954)*x[1]+(0.294345859395)*x[2]**o+(-0.679169213305)*x[2]
        ref[0,0]=(0.487697656095)*x_ref[0]**o+(0.362509734966)*x_ref[0]+(-0.40438484582)*x_ref[1]**o+(-0.358984481681)*x_ref[1]+(0.157670938098)*x_ref[2]**o+(-0.631406375093)*x_ref[2]
        ref[0,1]=(-0.00251154956677)*x_ref[0]**o+(0.671447769464)*x_ref[0]+(0.535541351526)*x_ref[1]**o+(0.433508112485)*x_ref[1]+(-0.897668718638)*x_ref[2]**o+(0.141331586102)*x_ref[2]
        ref[0,2]=(-0.964778505658)*x_ref[0]**o+(0.522815935313)*x_ref[0]+(-0.750318077186)*x_ref[1]**o+(0.0595240411196)*x_ref[1]+(-0.384021192511)*x_ref[2]**o+(-0.802529657745)*x_ref[2]
        ref[0,3]=(0.264781810221)*x_ref[0]**o+(-0.654277693956)*x_ref[0]+(0.890264421741)*x_ref[1]**o+(-0.657743774786)*x_ref[1]+(0.320234835664)*x_ref[2]**o+(0.0637070094165)*x_ref[2]
        ref[0,4]=(0.712229359775)*x_ref[0]**o+(-0.646586719826)*x_ref[0]+(0.876519277606)*x_ref[1]**o+(0.473649630238)*x_ref[1]+(0.332259187685)*x_ref[2]**o+(-0.331489699179)*x_ref[2]
        ref[1,0]=(0.808534274392)*x_ref[0]**o+(-0.296462057802)*x_ref[0]+(0.152536716655)*x_ref[1]**o+(-0.990267302005)*x_ref[1]+(-0.823462709316)*x_ref[2]**o+(-0.51608634537)*x_ref[2]
        ref[1,1]=(-0.745238552081)*x_ref[0]**o+(-0.892086591023)*x_ref[0]+(0.691429994644)*x_ref[1]**o+(-0.52538570831)*x_ref[1]+(-0.668959934842)*x_ref[2]**o+(-0.5178179109)*x_ref[2]
        ref[1,2]=(0.10173204456)*x_ref[0]**o+(-0.853001564748)*x_ref[0]+(0.617252846191)*x_ref[1]**o+(-0.518228673297)*x_ref[1]+(-0.995447484095)*x_ref[2]**o+(0.56679145486)*x_ref[2]
        ref[1,3]=(0.285145125389)*x_ref[0]**o+(0.773362877732)*x_ref[0]+(-0.457364007088)*x_ref[1]**o+(0.226257217659)*x_ref[1]+(0.439832733067)*x_ref[2]**o+(0.872812128578)*x_ref[2]
        ref[1,4]=(-0.581535304023)*x_ref[0]**o+(-0.944599120493)*x_ref[0]+(-0.720009453699)*x_ref[1]**o+(-0.29624047891)*x_ref[1]+(0.966528522407)*x_ref[2]**o+(0.818945905214)*x_ref[2]
        ref[2,0]=(-0.650732348774)*x_ref[0]**o+(0.368909306978)*x_ref[0]+(-0.608327978493)*x_ref[1]**o+(0.301600904054)*x_ref[1]+(-0.747648472883)*x_ref[2]**o+(0.846568601061)*x_ref[2]
        ref[2,1]=(-0.0221803961531)*x_ref[0]**o+(-0.858842969825)*x_ref[0]+(-0.40141447334)*x_ref[1]**o+(0.408164536045)*x_ref[1]+(-0.0131899768237)*x_ref[2]**o+(-0.0983376041104)*x_ref[2]
        ref[2,2]=(0.0855129797683)*x_ref[0]**o+(0.0645200342375)*x_ref[0]+(0.295916169508)*x_ref[1]**o+(0.343169626483)*x_ref[1]+(0.973477621396)*x_ref[2]**o+(0.179505855846)*x_ref[2]
        ref[2,3]=(-0.115714157482)*x_ref[0]**o+(-0.00785628190789)*x_ref[0]+(0.406498238895)*x_ref[1]**o+(-0.954353977843)*x_ref[1]+(0.751544148524)*x_ref[2]**o+(-0.092307218389)*x_ref[2]
        ref[2,4]=(0.807583813259)*x_ref[0]**o+(-0.0150562208564)*x_ref[0]+(0.967290753425)*x_ref[1]**o+(0.960049095802)*x_ref[1]+(0.41935259669)*x_ref[2]**o+(0.878823415691)*x_ref[2]
        ref[3,0]=(-0.162221321908)*x_ref[0]**o+(0.0142404015457)*x_ref[0]+(0.769189546147)*x_ref[1]**o+(-0.251240080116)*x_ref[1]+(0.980441953151)*x_ref[2]**o+(-0.183639594778)*x_ref[2]
        ref[3,1]=(0.270231471303)*x_ref[0]**o+(-0.527471406108)*x_ref[0]+(-0.255144134035)*x_ref[1]**o+(0.268061599615)*x_ref[1]+(-0.396373327005)*x_ref[2]**o+(0.105905036411)*x_ref[2]
        ref[3,2]=(0.5159632953)*x_ref[0]**o+(-0.87397274133)*x_ref[0]+(0.493876004047)*x_ref[1]**o+(0.320996363656)*x_ref[1]+(0.884345148533)*x_ref[2]**o+(0.116262647026)*x_ref[2]
        ref[3,3]=(-0.354053396048)*x_ref[0]**o+(-0.130121845174)*x_ref[0]+(-0.33114435645)*x_ref[1]**o+(-0.73998189126)*x_ref[1]+(0.362417530484)*x_ref[2]**o+(0.676820875084)*x_ref[2]
        ref[3,4]=(-0.53273881792)*x_ref[0]**o+(-0.0624208514424)*x_ref[0]+(0.680486778939)*x_ref[1]**o+(-0.370666453954)*x_ref[1]+(0.294345859395)*x_ref[2]**o+(-0.679169213305)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0906971676514)*x[0]**o+(-0.0640252128559)*x[0]+(-0.0108470120234)*x[1]**o+(-0.62531206416)*x[1]
        arg[0,0,1]=(-0.0186703191671)*x[0]**o+(0.256044044276)*x[0]+(-0.420130556042)*x[1]**o+(-0.241203381189)*x[1]
        arg[0,1,0]=(-0.287587258734)*x[0]**o+(-0.152487596827)*x[0]+(-0.554975489568)*x[1]**o+(0.713650618367)*x[1]
        arg[0,1,1]=(0.467597882001)*x[0]**o+(0.617291271522)*x[0]+(0.839893220642)*x[1]**o+(0.477553624935)*x[1]
        arg[1,0,0]=(0.292433268891)*x[0]**o+(0.0420372671524)*x[0]+(-0.342734694042)*x[1]**o+(0.892580432212)*x[1]
        arg[1,0,1]=(0.191399101606)*x[0]**o+(0.413781516804)*x[0]+(-0.346483382584)*x[1]**o+(0.805718474386)*x[1]
        arg[1,1,0]=(0.917644938285)*x[0]**o+(0.60457272962)*x[0]+(-0.0949129141155)*x[1]**o+(0.705971564397)*x[1]
        arg[1,1,1]=(0.327738300229)*x[0]**o+(-0.738951673731)*x[0]+(0.335869509312)*x[1]**o+(0.392167761004)*x[1]
        arg[2,0,0]=(0.460678533261)*x[0]**o+(-0.10665029868)*x[0]+(0.658148262201)*x[1]**o+(-0.75975084995)*x[1]
        arg[2,0,1]=(0.307057582515)*x[0]**o+(0.0755776536138)*x[0]+(0.866743758656)*x[1]**o+(-0.463537240812)*x[1]
        arg[2,1,0]=(-0.470241867396)*x[0]**o+(-0.0316070952691)*x[0]+(0.877786692055)*x[1]**o+(-0.750713976717)*x[1]
        arg[2,1,1]=(0.755903505788)*x[0]**o+(0.398472877018)*x[0]+(-0.372159129627)*x[1]**o+(0.170870288116)*x[1]
        arg[3,0,0]=(0.0310115714805)*x[0]**o+(0.822261427932)*x[0]+(-0.0979068243755)*x[1]**o+(0.61848414895)*x[1]
        arg[3,0,1]=(-0.442666645023)*x[0]**o+(0.759423291727)*x[0]+(0.771826722418)*x[1]**o+(-0.342283533203)*x[1]
        arg[3,1,0]=(0.292608821995)*x[0]**o+(0.301553824784)*x[0]+(0.445054135435)*x[1]**o+(0.990249711651)*x[1]
        arg[3,1,1]=(0.835714722542)*x[0]**o+(0.888569910076)*x[0]+(0.76442321847)*x[1]**o+(-0.272880995105)*x[1]
        arg[4,0,0]=(0.434738367563)*x[0]**o+(-0.801063604881)*x[0]+(-0.0152782263323)*x[1]**o+(-0.309467614762)*x[1]
        arg[4,0,1]=(-0.647140151015)*x[0]**o+(-0.0810020795911)*x[0]+(0.356402263616)*x[1]**o+(0.861692124435)*x[1]
        arg[4,1,0]=(-0.958501155541)*x[0]**o+(0.611720259342)*x[0]+(-0.631450146062)*x[1]**o+(-0.512668667768)*x[1]
        arg[4,1,1]=(0.771988547023)*x[0]**o+(0.466474334634)*x[0]+(-0.380283529607)*x[1]**o+(-0.619497160805)*x[1]
        arg[5,0,0]=(0.427386565443)*x[0]**o+(-0.178637075872)*x[0]+(0.81988184595)*x[1]**o+(-0.407798714033)*x[1]
        arg[5,0,1]=(-0.850161084571)*x[0]**o+(0.183440360862)*x[0]+(-0.83848298192)*x[1]**o+(-0.222248160259)*x[1]
        arg[5,1,0]=(-0.930623680879)*x[0]**o+(0.694991147173)*x[0]+(-0.299904374239)*x[1]**o+(0.552324198856)*x[1]
        arg[5,1,1]=(0.740715083219)*x[0]**o+(0.438768785937)*x[0]+(0.977523774483)*x[1]**o+(-0.668723413125)*x[1]
        ref[0,0,0]=(-0.0906971676514)*x_ref[0]**o+(-0.0640252128559)*x_ref[0]+(-0.0108470120234)*x_ref[1]**o+(-0.62531206416)*x_ref[1]
        ref[0,0,1]=(-0.0186703191671)*x_ref[0]**o+(0.256044044276)*x_ref[0]+(-0.420130556042)*x_ref[1]**o+(-0.241203381189)*x_ref[1]
        ref[0,1,0]=(-0.287587258734)*x_ref[0]**o+(-0.152487596827)*x_ref[0]+(-0.554975489568)*x_ref[1]**o+(0.713650618367)*x_ref[1]
        ref[0,1,1]=(0.467597882001)*x_ref[0]**o+(0.617291271522)*x_ref[0]+(0.839893220642)*x_ref[1]**o+(0.477553624935)*x_ref[1]
        ref[1,0,0]=(0.292433268891)*x_ref[0]**o+(0.0420372671524)*x_ref[0]+(-0.342734694042)*x_ref[1]**o+(0.892580432212)*x_ref[1]
        ref[1,0,1]=(0.191399101606)*x_ref[0]**o+(0.413781516804)*x_ref[0]+(-0.346483382584)*x_ref[1]**o+(0.805718474386)*x_ref[1]
        ref[1,1,0]=(0.917644938285)*x_ref[0]**o+(0.60457272962)*x_ref[0]+(-0.0949129141155)*x_ref[1]**o+(0.705971564397)*x_ref[1]
        ref[1,1,1]=(0.327738300229)*x_ref[0]**o+(-0.738951673731)*x_ref[0]+(0.335869509312)*x_ref[1]**o+(0.392167761004)*x_ref[1]
        ref[2,0,0]=(0.460678533261)*x_ref[0]**o+(-0.10665029868)*x_ref[0]+(0.658148262201)*x_ref[1]**o+(-0.75975084995)*x_ref[1]
        ref[2,0,1]=(0.307057582515)*x_ref[0]**o+(0.0755776536138)*x_ref[0]+(0.866743758656)*x_ref[1]**o+(-0.463537240812)*x_ref[1]
        ref[2,1,0]=(-0.470241867396)*x_ref[0]**o+(-0.0316070952691)*x_ref[0]+(0.877786692055)*x_ref[1]**o+(-0.750713976717)*x_ref[1]
        ref[2,1,1]=(0.755903505788)*x_ref[0]**o+(0.398472877018)*x_ref[0]+(-0.372159129627)*x_ref[1]**o+(0.170870288116)*x_ref[1]
        ref[3,0,0]=(0.0310115714805)*x_ref[0]**o+(0.822261427932)*x_ref[0]+(-0.0979068243755)*x_ref[1]**o+(0.61848414895)*x_ref[1]
        ref[3,0,1]=(-0.442666645023)*x_ref[0]**o+(0.759423291727)*x_ref[0]+(0.771826722418)*x_ref[1]**o+(-0.342283533203)*x_ref[1]
        ref[3,1,0]=(0.292608821995)*x_ref[0]**o+(0.301553824784)*x_ref[0]+(0.445054135435)*x_ref[1]**o+(0.990249711651)*x_ref[1]
        ref[3,1,1]=(0.835714722542)*x_ref[0]**o+(0.888569910076)*x_ref[0]+(0.76442321847)*x_ref[1]**o+(-0.272880995105)*x_ref[1]
        ref[4,0,0]=(0.434738367563)*x_ref[0]**o+(-0.801063604881)*x_ref[0]+(-0.0152782263323)*x_ref[1]**o+(-0.309467614762)*x_ref[1]
        ref[4,0,1]=(-0.647140151015)*x_ref[0]**o+(-0.0810020795911)*x_ref[0]+(0.356402263616)*x_ref[1]**o+(0.861692124435)*x_ref[1]
        ref[4,1,0]=(-0.958501155541)*x_ref[0]**o+(0.611720259342)*x_ref[0]+(-0.631450146062)*x_ref[1]**o+(-0.512668667768)*x_ref[1]
        ref[4,1,1]=(0.771988547023)*x_ref[0]**o+(0.466474334634)*x_ref[0]+(-0.380283529607)*x_ref[1]**o+(-0.619497160805)*x_ref[1]
        ref[5,0,0]=(0.427386565443)*x_ref[0]**o+(-0.178637075872)*x_ref[0]+(0.81988184595)*x_ref[1]**o+(-0.407798714033)*x_ref[1]
        ref[5,0,1]=(-0.850161084571)*x_ref[0]**o+(0.183440360862)*x_ref[0]+(-0.83848298192)*x_ref[1]**o+(-0.222248160259)*x_ref[1]
        ref[5,1,0]=(-0.930623680879)*x_ref[0]**o+(0.694991147173)*x_ref[0]+(-0.299904374239)*x_ref[1]**o+(0.552324198856)*x_ref[1]
        ref[5,1,1]=(0.740715083219)*x_ref[0]**o+(0.438768785937)*x_ref[0]+(0.977523774483)*x_ref[1]**o+(-0.668723413125)*x_ref[1]
      else:
        arg[0,0,0]=(0.0408197213005)*x[0]**o+(-0.992920054273)*x[0]+(0.122468233435)*x[1]**o+(-0.435942398199)*x[1]+(0.822235014621)*x[2]**o+(0.425925287623)*x[2]
        arg[0,0,1]=(0.474354220259)*x[0]**o+(0.287892840646)*x[0]+(-0.519818501933)*x[1]**o+(0.412301993774)*x[1]+(-0.242320576454)*x[2]**o+(0.758556860596)*x[2]
        arg[0,1,0]=(0.588735088154)*x[0]**o+(-0.948279493327)*x[0]+(0.194404465814)*x[1]**o+(-0.00761227943336)*x[1]+(-0.117173968067)*x[2]**o+(-0.228746139749)*x[2]
        arg[0,1,1]=(0.166448081217)*x[0]**o+(-0.228754542424)*x[0]+(-0.87739172573)*x[1]**o+(-0.592282861607)*x[1]+(-0.836732991313)*x[2]**o+(-0.20226535214)*x[2]
        arg[1,0,0]=(0.2683388552)*x[0]**o+(-0.733516131653)*x[0]+(-0.900887340495)*x[1]**o+(-0.614908158304)*x[1]+(0.422189467695)*x[2]**o+(0.642276887435)*x[2]
        arg[1,0,1]=(0.965294152134)*x[0]**o+(-0.326731862045)*x[0]+(-0.797754723728)*x[1]**o+(-0.678993698958)*x[1]+(0.256815891879)*x[2]**o+(0.48610716642)*x[2]
        arg[1,1,0]=(0.71383480118)*x[0]**o+(-0.949206151211)*x[0]+(0.40802633055)*x[1]**o+(-0.512457451544)*x[1]+(0.83647020826)*x[2]**o+(0.330550196742)*x[2]
        arg[1,1,1]=(-0.108602934331)*x[0]**o+(-0.585532075482)*x[0]+(0.302790940309)*x[1]**o+(-0.201628116557)*x[1]+(0.523440894937)*x[2]**o+(-0.421867218852)*x[2]
        arg[2,0,0]=(-0.911096130025)*x[0]**o+(-0.944970520542)*x[0]+(0.597179186191)*x[1]**o+(0.530539729209)*x[1]+(-0.538576718159)*x[2]**o+(-0.908863106348)*x[2]
        arg[2,0,1]=(0.828174522282)*x[0]**o+(-0.132534369339)*x[0]+(0.851849613713)*x[1]**o+(0.191494183049)*x[1]+(-0.953488097214)*x[2]**o+(0.423553741843)*x[2]
        arg[2,1,0]=(0.458631350156)*x[0]**o+(0.21755431391)*x[0]+(-0.128541259216)*x[1]**o+(-0.172831640065)*x[1]+(-0.62567145453)*x[2]**o+(-0.345626622399)*x[2]
        arg[2,1,1]=(0.49890508395)*x[0]**o+(-0.395084188812)*x[0]+(-0.653164498387)*x[1]**o+(-0.656416446654)*x[1]+(-0.844855683273)*x[2]**o+(-0.813755115861)*x[2]
        arg[3,0,0]=(-0.252710828698)*x[0]**o+(-0.0100429888666)*x[0]+(0.952930817098)*x[1]**o+(-0.950777038172)*x[1]+(0.244588721191)*x[2]**o+(-0.400442332557)*x[2]
        arg[3,0,1]=(-0.505331423555)*x[0]**o+(-0.535163478725)*x[0]+(0.924129836434)*x[1]**o+(-0.285813201016)*x[1]+(0.225037301542)*x[2]**o+(-0.331424134564)*x[2]
        arg[3,1,0]=(-0.556397106971)*x[0]**o+(0.26005371138)*x[0]+(0.902941320217)*x[1]**o+(0.70375648044)*x[1]+(-0.559371940837)*x[2]**o+(-0.628051748121)*x[2]
        arg[3,1,1]=(-0.644440017615)*x[0]**o+(-0.13415915731)*x[0]+(-0.752973804206)*x[1]**o+(-0.430304389409)*x[1]+(0.892979644749)*x[2]**o+(0.0618885081881)*x[2]
        arg[4,0,0]=(0.388902949971)*x[0]**o+(-0.413207160819)*x[0]+(-0.597892565551)*x[1]**o+(-0.433017605538)*x[1]+(0.369120396114)*x[2]**o+(-0.714185662873)*x[2]
        arg[4,0,1]=(0.140621693105)*x[0]**o+(0.910612147355)*x[0]+(0.404844175588)*x[1]**o+(0.963351416048)*x[1]+(0.545142156468)*x[2]**o+(-0.513154196146)*x[2]
        arg[4,1,0]=(-0.6045448854)*x[0]**o+(0.364599037279)*x[0]+(-0.165333666972)*x[1]**o+(-0.454715984949)*x[1]+(-0.123749300084)*x[2]**o+(0.23666517892)*x[2]
        arg[4,1,1]=(-0.701819332028)*x[0]**o+(-0.945269094128)*x[0]+(0.313745376139)*x[1]**o+(-0.0508160648325)*x[1]+(0.273150632944)*x[2]**o+(0.363230489878)*x[2]
        arg[5,0,0]=(-0.899477778287)*x[0]**o+(0.408300340496)*x[0]+(-0.501090863132)*x[1]**o+(-0.142172524799)*x[1]+(0.840930927028)*x[2]**o+(0.0786174984564)*x[2]
        arg[5,0,1]=(0.816863720765)*x[0]**o+(0.421613919321)*x[0]+(0.619636513942)*x[1]**o+(-0.867504595488)*x[1]+(-0.377345269009)*x[2]**o+(-0.884630348781)*x[2]
        arg[5,1,0]=(0.972307172208)*x[0]**o+(0.704673083035)*x[0]+(0.546566839332)*x[1]**o+(-0.303145527888)*x[1]+(-0.203985539635)*x[2]**o+(0.984097584463)*x[2]
        arg[5,1,1]=(0.67971485539)*x[0]**o+(0.492284313707)*x[0]+(0.139113989973)*x[1]**o+(-0.981414708174)*x[1]+(0.585926444541)*x[2]**o+(-0.790952419012)*x[2]
        ref[0,0,0]=(0.0408197213005)*x_ref[0]**o+(-0.992920054273)*x_ref[0]+(0.122468233435)*x_ref[1]**o+(-0.435942398199)*x_ref[1]+(0.822235014621)*x_ref[2]**o+(0.425925287623)*x_ref[2]
        ref[0,0,1]=(0.474354220259)*x_ref[0]**o+(0.287892840646)*x_ref[0]+(-0.519818501933)*x_ref[1]**o+(0.412301993774)*x_ref[1]+(-0.242320576454)*x_ref[2]**o+(0.758556860596)*x_ref[2]
        ref[0,1,0]=(0.588735088154)*x_ref[0]**o+(-0.948279493327)*x_ref[0]+(0.194404465814)*x_ref[1]**o+(-0.00761227943336)*x_ref[1]+(-0.117173968067)*x_ref[2]**o+(-0.228746139749)*x_ref[2]
        ref[0,1,1]=(0.166448081217)*x_ref[0]**o+(-0.228754542424)*x_ref[0]+(-0.87739172573)*x_ref[1]**o+(-0.592282861607)*x_ref[1]+(-0.836732991313)*x_ref[2]**o+(-0.20226535214)*x_ref[2]
        ref[1,0,0]=(0.2683388552)*x_ref[0]**o+(-0.733516131653)*x_ref[0]+(-0.900887340495)*x_ref[1]**o+(-0.614908158304)*x_ref[1]+(0.422189467695)*x_ref[2]**o+(0.642276887435)*x_ref[2]
        ref[1,0,1]=(0.965294152134)*x_ref[0]**o+(-0.326731862045)*x_ref[0]+(-0.797754723728)*x_ref[1]**o+(-0.678993698958)*x_ref[1]+(0.256815891879)*x_ref[2]**o+(0.48610716642)*x_ref[2]
        ref[1,1,0]=(0.71383480118)*x_ref[0]**o+(-0.949206151211)*x_ref[0]+(0.40802633055)*x_ref[1]**o+(-0.512457451544)*x_ref[1]+(0.83647020826)*x_ref[2]**o+(0.330550196742)*x_ref[2]
        ref[1,1,1]=(-0.108602934331)*x_ref[0]**o+(-0.585532075482)*x_ref[0]+(0.302790940309)*x_ref[1]**o+(-0.201628116557)*x_ref[1]+(0.523440894937)*x_ref[2]**o+(-0.421867218852)*x_ref[2]
        ref[2,0,0]=(-0.911096130025)*x_ref[0]**o+(-0.944970520542)*x_ref[0]+(0.597179186191)*x_ref[1]**o+(0.530539729209)*x_ref[1]+(-0.538576718159)*x_ref[2]**o+(-0.908863106348)*x_ref[2]
        ref[2,0,1]=(0.828174522282)*x_ref[0]**o+(-0.132534369339)*x_ref[0]+(0.851849613713)*x_ref[1]**o+(0.191494183049)*x_ref[1]+(-0.953488097214)*x_ref[2]**o+(0.423553741843)*x_ref[2]
        ref[2,1,0]=(0.458631350156)*x_ref[0]**o+(0.21755431391)*x_ref[0]+(-0.128541259216)*x_ref[1]**o+(-0.172831640065)*x_ref[1]+(-0.62567145453)*x_ref[2]**o+(-0.345626622399)*x_ref[2]
        ref[2,1,1]=(0.49890508395)*x_ref[0]**o+(-0.395084188812)*x_ref[0]+(-0.653164498387)*x_ref[1]**o+(-0.656416446654)*x_ref[1]+(-0.844855683273)*x_ref[2]**o+(-0.813755115861)*x_ref[2]
        ref[3,0,0]=(-0.252710828698)*x_ref[0]**o+(-0.0100429888666)*x_ref[0]+(0.952930817098)*x_ref[1]**o+(-0.950777038172)*x_ref[1]+(0.244588721191)*x_ref[2]**o+(-0.400442332557)*x_ref[2]
        ref[3,0,1]=(-0.505331423555)*x_ref[0]**o+(-0.535163478725)*x_ref[0]+(0.924129836434)*x_ref[1]**o+(-0.285813201016)*x_ref[1]+(0.225037301542)*x_ref[2]**o+(-0.331424134564)*x_ref[2]
        ref[3,1,0]=(-0.556397106971)*x_ref[0]**o+(0.26005371138)*x_ref[0]+(0.902941320217)*x_ref[1]**o+(0.70375648044)*x_ref[1]+(-0.559371940837)*x_ref[2]**o+(-0.628051748121)*x_ref[2]
        ref[3,1,1]=(-0.644440017615)*x_ref[0]**o+(-0.13415915731)*x_ref[0]+(-0.752973804206)*x_ref[1]**o+(-0.430304389409)*x_ref[1]+(0.892979644749)*x_ref[2]**o+(0.0618885081881)*x_ref[2]
        ref[4,0,0]=(0.388902949971)*x_ref[0]**o+(-0.413207160819)*x_ref[0]+(-0.597892565551)*x_ref[1]**o+(-0.433017605538)*x_ref[1]+(0.369120396114)*x_ref[2]**o+(-0.714185662873)*x_ref[2]
        ref[4,0,1]=(0.140621693105)*x_ref[0]**o+(0.910612147355)*x_ref[0]+(0.404844175588)*x_ref[1]**o+(0.963351416048)*x_ref[1]+(0.545142156468)*x_ref[2]**o+(-0.513154196146)*x_ref[2]
        ref[4,1,0]=(-0.6045448854)*x_ref[0]**o+(0.364599037279)*x_ref[0]+(-0.165333666972)*x_ref[1]**o+(-0.454715984949)*x_ref[1]+(-0.123749300084)*x_ref[2]**o+(0.23666517892)*x_ref[2]
        ref[4,1,1]=(-0.701819332028)*x_ref[0]**o+(-0.945269094128)*x_ref[0]+(0.313745376139)*x_ref[1]**o+(-0.0508160648325)*x_ref[1]+(0.273150632944)*x_ref[2]**o+(0.363230489878)*x_ref[2]
        ref[5,0,0]=(-0.899477778287)*x_ref[0]**o+(0.408300340496)*x_ref[0]+(-0.501090863132)*x_ref[1]**o+(-0.142172524799)*x_ref[1]+(0.840930927028)*x_ref[2]**o+(0.0786174984564)*x_ref[2]
        ref[5,0,1]=(0.816863720765)*x_ref[0]**o+(0.421613919321)*x_ref[0]+(0.619636513942)*x_ref[1]**o+(-0.867504595488)*x_ref[1]+(-0.377345269009)*x_ref[2]**o+(-0.884630348781)*x_ref[2]
        ref[5,1,0]=(0.972307172208)*x_ref[0]**o+(0.704673083035)*x_ref[0]+(0.546566839332)*x_ref[1]**o+(-0.303145527888)*x_ref[1]+(-0.203985539635)*x_ref[2]**o+(0.984097584463)*x_ref[2]
        ref[5,1,1]=(0.67971485539)*x_ref[0]**o+(0.492284313707)*x_ref[0]+(0.139113989973)*x_ref[1]**o+(-0.981414708174)*x_ref[1]+(0.585926444541)*x_ref[2]**o+(-0.790952419012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.717741572212)*x[0]**o+(-0.771801794802)*x[0]+(-0.782201793954)*x[1]**o+(0.125253178673)*x[1]
        arg[0,0,0,1]=(-0.483339499341)*x[0]**o+(0.0177945970437)*x[0]+(0.233532060933)*x[1]**o+(-0.465948890703)*x[1]
        arg[0,0,1,0]=(0.140544455071)*x[0]**o+(-0.120145399579)*x[0]+(-0.0732045826929)*x[1]**o+(-0.738165902415)*x[1]
        arg[0,0,1,1]=(0.98057407477)*x[0]**o+(0.346125329427)*x[0]+(0.697386820108)*x[1]**o+(0.402406370862)*x[1]
        arg[0,0,2,0]=(0.347641628204)*x[0]**o+(-0.217938924727)*x[0]+(-0.0572727430966)*x[1]**o+(0.170631149745)*x[1]
        arg[0,0,2,1]=(0.621840702108)*x[0]**o+(-0.109219941839)*x[0]+(0.894052962811)*x[1]**o+(0.423894680808)*x[1]
        arg[0,1,0,0]=(0.0339315069567)*x[0]**o+(-0.655546201957)*x[0]+(-0.271776247317)*x[1]**o+(0.4760527815)*x[1]
        arg[0,1,0,1]=(0.904216237912)*x[0]**o+(0.974869309983)*x[0]+(0.678379481202)*x[1]**o+(0.602534020957)*x[1]
        arg[0,1,1,0]=(-0.98805028173)*x[0]**o+(0.255251214419)*x[0]+(0.659010647624)*x[1]**o+(-0.45360685971)*x[1]
        arg[0,1,1,1]=(0.0884339867328)*x[0]**o+(-0.325188714625)*x[0]+(0.158562643552)*x[1]**o+(-0.788600388567)*x[1]
        arg[0,1,2,0]=(0.976404407593)*x[0]**o+(0.353231650963)*x[0]+(0.546951413477)*x[1]**o+(0.0738098551899)*x[1]
        arg[0,1,2,1]=(-0.812156797654)*x[0]**o+(-0.0180014270879)*x[0]+(0.686011517957)*x[1]**o+(0.148522434365)*x[1]
        arg[0,2,0,0]=(-0.40018516407)*x[0]**o+(-0.0512695503575)*x[0]+(-0.715512784143)*x[1]**o+(0.111064716031)*x[1]
        arg[0,2,0,1]=(-0.963397739298)*x[0]**o+(0.406951519848)*x[0]+(-0.293539028967)*x[1]**o+(-0.549328925146)*x[1]
        arg[0,2,1,0]=(-0.00827402449089)*x[0]**o+(-0.657982352244)*x[0]+(-0.515966831052)*x[1]**o+(0.531853371259)*x[1]
        arg[0,2,1,1]=(0.0379862601491)*x[0]**o+(0.988691147833)*x[0]+(0.820623911484)*x[1]**o+(0.689064076022)*x[1]
        arg[0,2,2,0]=(-0.857638654259)*x[0]**o+(0.664968676908)*x[0]+(-0.228424264597)*x[1]**o+(0.740003679791)*x[1]
        arg[0,2,2,1]=(0.456486069586)*x[0]**o+(-0.510130176146)*x[0]+(-0.189525160745)*x[1]**o+(0.497975981413)*x[1]
        arg[0,3,0,0]=(-0.981633737105)*x[0]**o+(0.25636883881)*x[0]+(-0.0477676975467)*x[1]**o+(0.654475880014)*x[1]
        arg[0,3,0,1]=(-0.0710284929969)*x[0]**o+(0.338590765191)*x[0]+(0.639048913997)*x[1]**o+(-0.24999606762)*x[1]
        arg[0,3,1,0]=(-0.731484482659)*x[0]**o+(-0.808253846389)*x[0]+(-0.911723431862)*x[1]**o+(-0.266029942425)*x[1]
        arg[0,3,1,1]=(0.909037857376)*x[0]**o+(-0.164525039433)*x[0]+(-0.924332370345)*x[1]**o+(0.476479399362)*x[1]
        arg[0,3,2,0]=(0.00534173833653)*x[0]**o+(-0.435160515142)*x[0]+(0.402701511687)*x[1]**o+(0.161631276653)*x[1]
        arg[0,3,2,1]=(-0.595791669027)*x[0]**o+(0.170580206948)*x[0]+(0.468802250813)*x[1]**o+(-0.881472636722)*x[1]
        arg[0,4,0,0]=(-0.0739506349513)*x[0]**o+(0.513385505957)*x[0]+(0.046977740371)*x[1]**o+(0.95763159987)*x[1]
        arg[0,4,0,1]=(-0.286789006223)*x[0]**o+(0.017476293519)*x[0]+(0.637378781092)*x[1]**o+(-0.670201122126)*x[1]
        arg[0,4,1,0]=(0.159607100455)*x[0]**o+(0.715341910227)*x[0]+(0.949546645819)*x[1]**o+(0.362982182775)*x[1]
        arg[0,4,1,1]=(-0.237702024358)*x[0]**o+(-0.520488841454)*x[0]+(0.606333933569)*x[1]**o+(-0.00662757701049)*x[1]
        arg[0,4,2,0]=(-0.415753817901)*x[0]**o+(0.731285464784)*x[0]+(0.15987336144)*x[1]**o+(0.0697706236338)*x[1]
        arg[0,4,2,1]=(0.0467482158427)*x[0]**o+(0.306004722736)*x[0]+(-0.905184725692)*x[1]**o+(-0.158017889366)*x[1]
        arg[1,0,0,0]=(-0.785535082426)*x[0]**o+(-0.189285100452)*x[0]+(-0.766349940057)*x[1]**o+(0.531469308995)*x[1]
        arg[1,0,0,1]=(0.522213196477)*x[0]**o+(-0.661381741437)*x[0]+(0.816089886956)*x[1]**o+(0.139829382305)*x[1]
        arg[1,0,1,0]=(0.0957295570417)*x[0]**o+(-0.105252674536)*x[0]+(-0.569429953382)*x[1]**o+(-0.801934666657)*x[1]
        arg[1,0,1,1]=(-0.557307051058)*x[0]**o+(0.218852576416)*x[0]+(0.430167216464)*x[1]**o+(-0.112285681761)*x[1]
        arg[1,0,2,0]=(0.288472422477)*x[0]**o+(0.294377994763)*x[0]+(-0.809503961511)*x[1]**o+(0.357020226991)*x[1]
        arg[1,0,2,1]=(-0.148113477705)*x[0]**o+(0.164468646294)*x[0]+(0.7977142476)*x[1]**o+(-0.407274933384)*x[1]
        arg[1,1,0,0]=(-0.926374434822)*x[0]**o+(-0.107713463193)*x[0]+(-0.511635723869)*x[1]**o+(0.259319717689)*x[1]
        arg[1,1,0,1]=(-0.18636582243)*x[0]**o+(0.698258973628)*x[0]+(0.234379281397)*x[1]**o+(0.596195875495)*x[1]
        arg[1,1,1,0]=(-0.180395892006)*x[0]**o+(0.631248500084)*x[0]+(0.104751616377)*x[1]**o+(0.204500206653)*x[1]
        arg[1,1,1,1]=(-0.517119564238)*x[0]**o+(-0.16634015634)*x[0]+(0.556785586675)*x[1]**o+(0.42659306392)*x[1]
        arg[1,1,2,0]=(0.163807238958)*x[0]**o+(-0.534619478446)*x[0]+(-0.700353460753)*x[1]**o+(0.121393131648)*x[1]
        arg[1,1,2,1]=(0.964276278961)*x[0]**o+(0.846030751304)*x[0]+(-0.745301894392)*x[1]**o+(0.33821485558)*x[1]
        arg[1,2,0,0]=(0.333173101086)*x[0]**o+(0.774848048743)*x[0]+(-0.659091916357)*x[1]**o+(-0.797247144497)*x[1]
        arg[1,2,0,1]=(0.0581394534285)*x[0]**o+(-0.524771663031)*x[0]+(-0.519465753317)*x[1]**o+(0.71748795226)*x[1]
        arg[1,2,1,0]=(0.224935262833)*x[0]**o+(-0.393399819306)*x[0]+(0.199349075224)*x[1]**o+(-0.808847009412)*x[1]
        arg[1,2,1,1]=(0.0385750795521)*x[0]**o+(-8.09949553151e-05)*x[0]+(0.496860964506)*x[1]**o+(0.0119507407829)*x[1]
        arg[1,2,2,0]=(0.876341556167)*x[0]**o+(0.18082766372)*x[0]+(0.401349871827)*x[1]**o+(-0.543439037303)*x[1]
        arg[1,2,2,1]=(-0.894525004507)*x[0]**o+(-0.516771143695)*x[0]+(0.0410916490263)*x[1]**o+(0.753841618439)*x[1]
        arg[1,3,0,0]=(0.251979970924)*x[0]**o+(0.790276550438)*x[0]+(-0.155708793888)*x[1]**o+(0.28971278336)*x[1]
        arg[1,3,0,1]=(0.258336265436)*x[0]**o+(0.696642875149)*x[0]+(0.0561355915649)*x[1]**o+(0.601839744417)*x[1]
        arg[1,3,1,0]=(-0.978729102785)*x[0]**o+(0.335280613262)*x[0]+(0.89228798232)*x[1]**o+(-0.693050679348)*x[1]
        arg[1,3,1,1]=(-0.885817753431)*x[0]**o+(-0.0641571535708)*x[0]+(-0.649578576181)*x[1]**o+(-0.95773137602)*x[1]
        arg[1,3,2,0]=(-0.333891001461)*x[0]**o+(-0.496064308564)*x[0]+(0.438549603377)*x[1]**o+(0.0772468142672)*x[1]
        arg[1,3,2,1]=(-0.16171788255)*x[0]**o+(-0.939293453222)*x[0]+(-0.284522648546)*x[1]**o+(0.744642328375)*x[1]
        arg[1,4,0,0]=(0.806133262332)*x[0]**o+(-0.0225320206432)*x[0]+(0.125958642606)*x[1]**o+(0.610630249467)*x[1]
        arg[1,4,0,1]=(0.657971870946)*x[0]**o+(0.698894701525)*x[0]+(-0.595952612496)*x[1]**o+(-0.38268841292)*x[1]
        arg[1,4,1,0]=(-0.277274411331)*x[0]**o+(-0.483183004514)*x[0]+(0.77656657009)*x[1]**o+(-0.959290466645)*x[1]
        arg[1,4,1,1]=(-0.96384356305)*x[0]**o+(-0.00383301781476)*x[0]+(0.788010409608)*x[1]**o+(0.135378744316)*x[1]
        arg[1,4,2,0]=(0.986761958017)*x[0]**o+(-0.570276378038)*x[0]+(-0.515784028315)*x[1]**o+(0.487125123439)*x[1]
        arg[1,4,2,1]=(-0.954735054702)*x[0]**o+(-0.0610323074217)*x[0]+(-0.255284859281)*x[1]**o+(0.277795725698)*x[1]
        arg[2,0,0,0]=(-0.708301219324)*x[0]**o+(-0.970166781553)*x[0]+(-0.0836451425855)*x[1]**o+(0.525854202995)*x[1]
        arg[2,0,0,1]=(-0.0995393865183)*x[0]**o+(-0.0156624820324)*x[0]+(0.225529316197)*x[1]**o+(0.142617834056)*x[1]
        arg[2,0,1,0]=(-0.585329524103)*x[0]**o+(0.590285924441)*x[0]+(0.201483726322)*x[1]**o+(-0.735262197689)*x[1]
        arg[2,0,1,1]=(-0.682877182963)*x[0]**o+(-0.162682262799)*x[0]+(-0.890051459843)*x[1]**o+(-0.0188015866765)*x[1]
        arg[2,0,2,0]=(0.868935476294)*x[0]**o+(-0.629891068854)*x[0]+(0.335632865663)*x[1]**o+(-0.955849055716)*x[1]
        arg[2,0,2,1]=(-0.907933228238)*x[0]**o+(-0.105861987974)*x[0]+(-0.131372383042)*x[1]**o+(-0.970114136854)*x[1]
        arg[2,1,0,0]=(-0.275788501814)*x[0]**o+(-0.992861867033)*x[0]+(-0.80676324962)*x[1]**o+(0.269244680224)*x[1]
        arg[2,1,0,1]=(-0.704704138774)*x[0]**o+(-0.371803193009)*x[0]+(-0.0190586405922)*x[1]**o+(-0.0828837624698)*x[1]
        arg[2,1,1,0]=(-0.532645566013)*x[0]**o+(0.571754809868)*x[0]+(0.656737493773)*x[1]**o+(-0.548229591842)*x[1]
        arg[2,1,1,1]=(0.23836921087)*x[0]**o+(0.707475737785)*x[0]+(-0.59660155968)*x[1]**o+(-0.450204951071)*x[1]
        arg[2,1,2,0]=(-0.534027348014)*x[0]**o+(0.332100193727)*x[0]+(-0.864157138567)*x[1]**o+(-0.41833267066)*x[1]
        arg[2,1,2,1]=(0.459878593269)*x[0]**o+(0.446259140228)*x[0]+(0.571242887502)*x[1]**o+(-0.0390525669486)*x[1]
        arg[2,2,0,0]=(0.322052799527)*x[0]**o+(-0.915265209304)*x[0]+(-0.99942007618)*x[1]**o+(0.482294513743)*x[1]
        arg[2,2,0,1]=(0.558489194456)*x[0]**o+(0.00400576554086)*x[0]+(-0.87484330165)*x[1]**o+(-0.664504218977)*x[1]
        arg[2,2,1,0]=(0.669621607461)*x[0]**o+(-0.0450960554319)*x[0]+(-0.221151480246)*x[1]**o+(0.0123621737211)*x[1]
        arg[2,2,1,1]=(0.476743013029)*x[0]**o+(0.11118371822)*x[0]+(-0.202074315677)*x[1]**o+(0.377649433458)*x[1]
        arg[2,2,2,0]=(0.667910543701)*x[0]**o+(0.035703922547)*x[0]+(-0.274733967095)*x[1]**o+(-0.998934720182)*x[1]
        arg[2,2,2,1]=(0.578909660087)*x[0]**o+(0.149900086514)*x[0]+(-0.226597651858)*x[1]**o+(0.169175812287)*x[1]
        arg[2,3,0,0]=(0.158370246586)*x[0]**o+(0.595643112285)*x[0]+(-0.476875564052)*x[1]**o+(0.869752657517)*x[1]
        arg[2,3,0,1]=(0.99718847714)*x[0]**o+(0.581531822297)*x[0]+(-0.607102775564)*x[1]**o+(-0.8317970284)*x[1]
        arg[2,3,1,0]=(-0.27390195745)*x[0]**o+(0.0512853330414)*x[0]+(0.465190666302)*x[1]**o+(0.042533388788)*x[1]
        arg[2,3,1,1]=(-0.183679504962)*x[0]**o+(-0.0522782076121)*x[0]+(0.921423118529)*x[1]**o+(-0.0315080275225)*x[1]
        arg[2,3,2,0]=(0.0421842348778)*x[0]**o+(0.462023049702)*x[0]+(-0.888342606803)*x[1]**o+(-0.956503447884)*x[1]
        arg[2,3,2,1]=(-0.950820288444)*x[0]**o+(0.104271990168)*x[0]+(-0.607212909582)*x[1]**o+(-0.16283931259)*x[1]
        arg[2,4,0,0]=(0.113505933533)*x[0]**o+(0.27873514806)*x[0]+(0.736699443523)*x[1]**o+(-0.951737555693)*x[1]
        arg[2,4,0,1]=(0.264170584849)*x[0]**o+(-0.226649236617)*x[0]+(0.489902180874)*x[1]**o+(-0.191536591762)*x[1]
        arg[2,4,1,0]=(0.712322285985)*x[0]**o+(-0.323395455815)*x[0]+(-0.298356681586)*x[1]**o+(0.699891974541)*x[1]
        arg[2,4,1,1]=(-0.631057763255)*x[0]**o+(-0.489374933445)*x[0]+(-0.021591806874)*x[1]**o+(-0.692800124585)*x[1]
        arg[2,4,2,0]=(0.916068935184)*x[0]**o+(0.33779688858)*x[0]+(0.06300443559)*x[1]**o+(-0.231385203399)*x[1]
        arg[2,4,2,1]=(-0.283485044625)*x[0]**o+(0.548366574442)*x[0]+(-0.543953132872)*x[1]**o+(-0.956662823037)*x[1]
        arg[3,0,0,0]=(0.709349274221)*x[0]**o+(-0.342981009263)*x[0]+(-0.682331761929)*x[1]**o+(0.673169884813)*x[1]
        arg[3,0,0,1]=(0.236249836593)*x[0]**o+(-0.682845054434)*x[0]+(-0.349169988218)*x[1]**o+(0.502684425824)*x[1]
        arg[3,0,1,0]=(-0.561655748793)*x[0]**o+(0.120183772326)*x[0]+(-0.966535325583)*x[1]**o+(-0.0585117264949)*x[1]
        arg[3,0,1,1]=(-0.661030287189)*x[0]**o+(-0.322924086505)*x[0]+(-0.986892059552)*x[1]**o+(-0.778969361365)*x[1]
        arg[3,0,2,0]=(-0.166307107307)*x[0]**o+(0.612931303617)*x[0]+(-0.283925165456)*x[1]**o+(0.5998038814)*x[1]
        arg[3,0,2,1]=(0.205099355916)*x[0]**o+(-0.520970536278)*x[0]+(0.900416833183)*x[1]**o+(0.981826306069)*x[1]
        arg[3,1,0,0]=(-0.875306666443)*x[0]**o+(0.0189358479861)*x[0]+(0.101187857445)*x[1]**o+(-0.873584307534)*x[1]
        arg[3,1,0,1]=(-0.0203656235757)*x[0]**o+(-0.579229777042)*x[0]+(-0.0799345631826)*x[1]**o+(0.896315444069)*x[1]
        arg[3,1,1,0]=(0.576791459982)*x[0]**o+(-0.86541815506)*x[0]+(-0.582764176207)*x[1]**o+(-0.249437395011)*x[1]
        arg[3,1,1,1]=(0.525953680497)*x[0]**o+(-0.222445717402)*x[0]+(-0.87710202063)*x[1]**o+(-0.762921821973)*x[1]
        arg[3,1,2,0]=(-0.883737185908)*x[0]**o+(0.254053241277)*x[0]+(0.559803298059)*x[1]**o+(-0.577376956856)*x[1]
        arg[3,1,2,1]=(-0.125792083387)*x[0]**o+(-0.771549307509)*x[0]+(0.142493480552)*x[1]**o+(0.459996962548)*x[1]
        arg[3,2,0,0]=(0.0531598927049)*x[0]**o+(-0.950343229037)*x[0]+(-0.748240453154)*x[1]**o+(0.850555892342)*x[1]
        arg[3,2,0,1]=(-0.660852152041)*x[0]**o+(-0.763119278604)*x[0]+(-0.811925409663)*x[1]**o+(-0.178765071495)*x[1]
        arg[3,2,1,0]=(-0.189124223305)*x[0]**o+(-0.410392909042)*x[0]+(0.993247167522)*x[1]**o+(0.291451796298)*x[1]
        arg[3,2,1,1]=(0.83496133421)*x[0]**o+(0.695147480265)*x[0]+(0.444420524946)*x[1]**o+(0.107454620114)*x[1]
        arg[3,2,2,0]=(0.94687479313)*x[0]**o+(-0.140702324395)*x[0]+(0.149203502428)*x[1]**o+(-0.338246126345)*x[1]
        arg[3,2,2,1]=(0.147609394709)*x[0]**o+(-0.549219765239)*x[0]+(0.397308993347)*x[1]**o+(0.0402189312685)*x[1]
        arg[3,3,0,0]=(0.27400981409)*x[0]**o+(-0.516367266257)*x[0]+(-0.784095952814)*x[1]**o+(0.32858899009)*x[1]
        arg[3,3,0,1]=(-0.393457382163)*x[0]**o+(0.590406794051)*x[0]+(0.328523918357)*x[1]**o+(0.899537513355)*x[1]
        arg[3,3,1,0]=(-0.851038623366)*x[0]**o+(0.596974864726)*x[0]+(-0.338630687551)*x[1]**o+(0.634961435042)*x[1]
        arg[3,3,1,1]=(0.569166143918)*x[0]**o+(-0.388162282491)*x[0]+(0.764830536733)*x[1]**o+(-0.426737342542)*x[1]
        arg[3,3,2,0]=(0.502221903448)*x[0]**o+(0.940382225897)*x[0]+(0.563444730579)*x[1]**o+(0.023210869151)*x[1]
        arg[3,3,2,1]=(0.061695403852)*x[0]**o+(0.868302188758)*x[0]+(0.0855803325758)*x[1]**o+(-0.24366173638)*x[1]
        arg[3,4,0,0]=(-0.533038182387)*x[0]**o+(0.0222388915737)*x[0]+(0.529037752769)*x[1]**o+(-0.0665320895494)*x[1]
        arg[3,4,0,1]=(0.714374055253)*x[0]**o+(-0.623670743333)*x[0]+(-0.866024792929)*x[1]**o+(-0.0673124896111)*x[1]
        arg[3,4,1,0]=(-0.0523466615045)*x[0]**o+(-0.472563426622)*x[0]+(0.139892673626)*x[1]**o+(0.539655417178)*x[1]
        arg[3,4,1,1]=(-0.473032580514)*x[0]**o+(0.692812093618)*x[0]+(0.664400517742)*x[1]**o+(0.184125423364)*x[1]
        arg[3,4,2,0]=(-0.0340798058725)*x[0]**o+(-0.58900562863)*x[0]+(0.396856384717)*x[1]**o+(-0.559175428019)*x[1]
        arg[3,4,2,1]=(-0.217679376634)*x[0]**o+(-0.90389036667)*x[0]+(0.967372942873)*x[1]**o+(-0.482913619277)*x[1]
        ref[0,0,0,0]=(-0.717741572212)*x_ref[0]**o+(-0.771801794802)*x_ref[0]+(-0.782201793954)*x_ref[1]**o+(0.125253178673)*x_ref[1]
        ref[0,0,0,1]=(-0.483339499341)*x_ref[0]**o+(0.0177945970437)*x_ref[0]+(0.233532060933)*x_ref[1]**o+(-0.465948890703)*x_ref[1]
        ref[0,0,1,0]=(0.140544455071)*x_ref[0]**o+(-0.120145399579)*x_ref[0]+(-0.0732045826929)*x_ref[1]**o+(-0.738165902415)*x_ref[1]
        ref[0,0,1,1]=(0.98057407477)*x_ref[0]**o+(0.346125329427)*x_ref[0]+(0.697386820108)*x_ref[1]**o+(0.402406370862)*x_ref[1]
        ref[0,0,2,0]=(0.347641628204)*x_ref[0]**o+(-0.217938924727)*x_ref[0]+(-0.0572727430966)*x_ref[1]**o+(0.170631149745)*x_ref[1]
        ref[0,0,2,1]=(0.621840702108)*x_ref[0]**o+(-0.109219941839)*x_ref[0]+(0.894052962811)*x_ref[1]**o+(0.423894680808)*x_ref[1]
        ref[0,1,0,0]=(0.0339315069567)*x_ref[0]**o+(-0.655546201957)*x_ref[0]+(-0.271776247317)*x_ref[1]**o+(0.4760527815)*x_ref[1]
        ref[0,1,0,1]=(0.904216237912)*x_ref[0]**o+(0.974869309983)*x_ref[0]+(0.678379481202)*x_ref[1]**o+(0.602534020957)*x_ref[1]
        ref[0,1,1,0]=(-0.98805028173)*x_ref[0]**o+(0.255251214419)*x_ref[0]+(0.659010647624)*x_ref[1]**o+(-0.45360685971)*x_ref[1]
        ref[0,1,1,1]=(0.0884339867328)*x_ref[0]**o+(-0.325188714625)*x_ref[0]+(0.158562643552)*x_ref[1]**o+(-0.788600388567)*x_ref[1]
        ref[0,1,2,0]=(0.976404407593)*x_ref[0]**o+(0.353231650963)*x_ref[0]+(0.546951413477)*x_ref[1]**o+(0.0738098551899)*x_ref[1]
        ref[0,1,2,1]=(-0.812156797654)*x_ref[0]**o+(-0.0180014270879)*x_ref[0]+(0.686011517957)*x_ref[1]**o+(0.148522434365)*x_ref[1]
        ref[0,2,0,0]=(-0.40018516407)*x_ref[0]**o+(-0.0512695503575)*x_ref[0]+(-0.715512784143)*x_ref[1]**o+(0.111064716031)*x_ref[1]
        ref[0,2,0,1]=(-0.963397739298)*x_ref[0]**o+(0.406951519848)*x_ref[0]+(-0.293539028967)*x_ref[1]**o+(-0.549328925146)*x_ref[1]
        ref[0,2,1,0]=(-0.00827402449089)*x_ref[0]**o+(-0.657982352244)*x_ref[0]+(-0.515966831052)*x_ref[1]**o+(0.531853371259)*x_ref[1]
        ref[0,2,1,1]=(0.0379862601491)*x_ref[0]**o+(0.988691147833)*x_ref[0]+(0.820623911484)*x_ref[1]**o+(0.689064076022)*x_ref[1]
        ref[0,2,2,0]=(-0.857638654259)*x_ref[0]**o+(0.664968676908)*x_ref[0]+(-0.228424264597)*x_ref[1]**o+(0.740003679791)*x_ref[1]
        ref[0,2,2,1]=(0.456486069586)*x_ref[0]**o+(-0.510130176146)*x_ref[0]+(-0.189525160745)*x_ref[1]**o+(0.497975981413)*x_ref[1]
        ref[0,3,0,0]=(-0.981633737105)*x_ref[0]**o+(0.25636883881)*x_ref[0]+(-0.0477676975467)*x_ref[1]**o+(0.654475880014)*x_ref[1]
        ref[0,3,0,1]=(-0.0710284929969)*x_ref[0]**o+(0.338590765191)*x_ref[0]+(0.639048913997)*x_ref[1]**o+(-0.24999606762)*x_ref[1]
        ref[0,3,1,0]=(-0.731484482659)*x_ref[0]**o+(-0.808253846389)*x_ref[0]+(-0.911723431862)*x_ref[1]**o+(-0.266029942425)*x_ref[1]
        ref[0,3,1,1]=(0.909037857376)*x_ref[0]**o+(-0.164525039433)*x_ref[0]+(-0.924332370345)*x_ref[1]**o+(0.476479399362)*x_ref[1]
        ref[0,3,2,0]=(0.00534173833653)*x_ref[0]**o+(-0.435160515142)*x_ref[0]+(0.402701511687)*x_ref[1]**o+(0.161631276653)*x_ref[1]
        ref[0,3,2,1]=(-0.595791669027)*x_ref[0]**o+(0.170580206948)*x_ref[0]+(0.468802250813)*x_ref[1]**o+(-0.881472636722)*x_ref[1]
        ref[0,4,0,0]=(-0.0739506349513)*x_ref[0]**o+(0.513385505957)*x_ref[0]+(0.046977740371)*x_ref[1]**o+(0.95763159987)*x_ref[1]
        ref[0,4,0,1]=(-0.286789006223)*x_ref[0]**o+(0.017476293519)*x_ref[0]+(0.637378781092)*x_ref[1]**o+(-0.670201122126)*x_ref[1]
        ref[0,4,1,0]=(0.159607100455)*x_ref[0]**o+(0.715341910227)*x_ref[0]+(0.949546645819)*x_ref[1]**o+(0.362982182775)*x_ref[1]
        ref[0,4,1,1]=(-0.237702024358)*x_ref[0]**o+(-0.520488841454)*x_ref[0]+(0.606333933569)*x_ref[1]**o+(-0.00662757701049)*x_ref[1]
        ref[0,4,2,0]=(-0.415753817901)*x_ref[0]**o+(0.731285464784)*x_ref[0]+(0.15987336144)*x_ref[1]**o+(0.0697706236338)*x_ref[1]
        ref[0,4,2,1]=(0.0467482158427)*x_ref[0]**o+(0.306004722736)*x_ref[0]+(-0.905184725692)*x_ref[1]**o+(-0.158017889366)*x_ref[1]
        ref[1,0,0,0]=(-0.785535082426)*x_ref[0]**o+(-0.189285100452)*x_ref[0]+(-0.766349940057)*x_ref[1]**o+(0.531469308995)*x_ref[1]
        ref[1,0,0,1]=(0.522213196477)*x_ref[0]**o+(-0.661381741437)*x_ref[0]+(0.816089886956)*x_ref[1]**o+(0.139829382305)*x_ref[1]
        ref[1,0,1,0]=(0.0957295570417)*x_ref[0]**o+(-0.105252674536)*x_ref[0]+(-0.569429953382)*x_ref[1]**o+(-0.801934666657)*x_ref[1]
        ref[1,0,1,1]=(-0.557307051058)*x_ref[0]**o+(0.218852576416)*x_ref[0]+(0.430167216464)*x_ref[1]**o+(-0.112285681761)*x_ref[1]
        ref[1,0,2,0]=(0.288472422477)*x_ref[0]**o+(0.294377994763)*x_ref[0]+(-0.809503961511)*x_ref[1]**o+(0.357020226991)*x_ref[1]
        ref[1,0,2,1]=(-0.148113477705)*x_ref[0]**o+(0.164468646294)*x_ref[0]+(0.7977142476)*x_ref[1]**o+(-0.407274933384)*x_ref[1]
        ref[1,1,0,0]=(-0.926374434822)*x_ref[0]**o+(-0.107713463193)*x_ref[0]+(-0.511635723869)*x_ref[1]**o+(0.259319717689)*x_ref[1]
        ref[1,1,0,1]=(-0.18636582243)*x_ref[0]**o+(0.698258973628)*x_ref[0]+(0.234379281397)*x_ref[1]**o+(0.596195875495)*x_ref[1]
        ref[1,1,1,0]=(-0.180395892006)*x_ref[0]**o+(0.631248500084)*x_ref[0]+(0.104751616377)*x_ref[1]**o+(0.204500206653)*x_ref[1]
        ref[1,1,1,1]=(-0.517119564238)*x_ref[0]**o+(-0.16634015634)*x_ref[0]+(0.556785586675)*x_ref[1]**o+(0.42659306392)*x_ref[1]
        ref[1,1,2,0]=(0.163807238958)*x_ref[0]**o+(-0.534619478446)*x_ref[0]+(-0.700353460753)*x_ref[1]**o+(0.121393131648)*x_ref[1]
        ref[1,1,2,1]=(0.964276278961)*x_ref[0]**o+(0.846030751304)*x_ref[0]+(-0.745301894392)*x_ref[1]**o+(0.33821485558)*x_ref[1]
        ref[1,2,0,0]=(0.333173101086)*x_ref[0]**o+(0.774848048743)*x_ref[0]+(-0.659091916357)*x_ref[1]**o+(-0.797247144497)*x_ref[1]
        ref[1,2,0,1]=(0.0581394534285)*x_ref[0]**o+(-0.524771663031)*x_ref[0]+(-0.519465753317)*x_ref[1]**o+(0.71748795226)*x_ref[1]
        ref[1,2,1,0]=(0.224935262833)*x_ref[0]**o+(-0.393399819306)*x_ref[0]+(0.199349075224)*x_ref[1]**o+(-0.808847009412)*x_ref[1]
        ref[1,2,1,1]=(0.0385750795521)*x_ref[0]**o+(-8.09949553151e-05)*x_ref[0]+(0.496860964506)*x_ref[1]**o+(0.0119507407829)*x_ref[1]
        ref[1,2,2,0]=(0.876341556167)*x_ref[0]**o+(0.18082766372)*x_ref[0]+(0.401349871827)*x_ref[1]**o+(-0.543439037303)*x_ref[1]
        ref[1,2,2,1]=(-0.894525004507)*x_ref[0]**o+(-0.516771143695)*x_ref[0]+(0.0410916490263)*x_ref[1]**o+(0.753841618439)*x_ref[1]
        ref[1,3,0,0]=(0.251979970924)*x_ref[0]**o+(0.790276550438)*x_ref[0]+(-0.155708793888)*x_ref[1]**o+(0.28971278336)*x_ref[1]
        ref[1,3,0,1]=(0.258336265436)*x_ref[0]**o+(0.696642875149)*x_ref[0]+(0.0561355915649)*x_ref[1]**o+(0.601839744417)*x_ref[1]
        ref[1,3,1,0]=(-0.978729102785)*x_ref[0]**o+(0.335280613262)*x_ref[0]+(0.89228798232)*x_ref[1]**o+(-0.693050679348)*x_ref[1]
        ref[1,3,1,1]=(-0.885817753431)*x_ref[0]**o+(-0.0641571535708)*x_ref[0]+(-0.649578576181)*x_ref[1]**o+(-0.95773137602)*x_ref[1]
        ref[1,3,2,0]=(-0.333891001461)*x_ref[0]**o+(-0.496064308564)*x_ref[0]+(0.438549603377)*x_ref[1]**o+(0.0772468142672)*x_ref[1]
        ref[1,3,2,1]=(-0.16171788255)*x_ref[0]**o+(-0.939293453222)*x_ref[0]+(-0.284522648546)*x_ref[1]**o+(0.744642328375)*x_ref[1]
        ref[1,4,0,0]=(0.806133262332)*x_ref[0]**o+(-0.0225320206432)*x_ref[0]+(0.125958642606)*x_ref[1]**o+(0.610630249467)*x_ref[1]
        ref[1,4,0,1]=(0.657971870946)*x_ref[0]**o+(0.698894701525)*x_ref[0]+(-0.595952612496)*x_ref[1]**o+(-0.38268841292)*x_ref[1]
        ref[1,4,1,0]=(-0.277274411331)*x_ref[0]**o+(-0.483183004514)*x_ref[0]+(0.77656657009)*x_ref[1]**o+(-0.959290466645)*x_ref[1]
        ref[1,4,1,1]=(-0.96384356305)*x_ref[0]**o+(-0.00383301781476)*x_ref[0]+(0.788010409608)*x_ref[1]**o+(0.135378744316)*x_ref[1]
        ref[1,4,2,0]=(0.986761958017)*x_ref[0]**o+(-0.570276378038)*x_ref[0]+(-0.515784028315)*x_ref[1]**o+(0.487125123439)*x_ref[1]
        ref[1,4,2,1]=(-0.954735054702)*x_ref[0]**o+(-0.0610323074217)*x_ref[0]+(-0.255284859281)*x_ref[1]**o+(0.277795725698)*x_ref[1]
        ref[2,0,0,0]=(-0.708301219324)*x_ref[0]**o+(-0.970166781553)*x_ref[0]+(-0.0836451425855)*x_ref[1]**o+(0.525854202995)*x_ref[1]
        ref[2,0,0,1]=(-0.0995393865183)*x_ref[0]**o+(-0.0156624820324)*x_ref[0]+(0.225529316197)*x_ref[1]**o+(0.142617834056)*x_ref[1]
        ref[2,0,1,0]=(-0.585329524103)*x_ref[0]**o+(0.590285924441)*x_ref[0]+(0.201483726322)*x_ref[1]**o+(-0.735262197689)*x_ref[1]
        ref[2,0,1,1]=(-0.682877182963)*x_ref[0]**o+(-0.162682262799)*x_ref[0]+(-0.890051459843)*x_ref[1]**o+(-0.0188015866765)*x_ref[1]
        ref[2,0,2,0]=(0.868935476294)*x_ref[0]**o+(-0.629891068854)*x_ref[0]+(0.335632865663)*x_ref[1]**o+(-0.955849055716)*x_ref[1]
        ref[2,0,2,1]=(-0.907933228238)*x_ref[0]**o+(-0.105861987974)*x_ref[0]+(-0.131372383042)*x_ref[1]**o+(-0.970114136854)*x_ref[1]
        ref[2,1,0,0]=(-0.275788501814)*x_ref[0]**o+(-0.992861867033)*x_ref[0]+(-0.80676324962)*x_ref[1]**o+(0.269244680224)*x_ref[1]
        ref[2,1,0,1]=(-0.704704138774)*x_ref[0]**o+(-0.371803193009)*x_ref[0]+(-0.0190586405922)*x_ref[1]**o+(-0.0828837624698)*x_ref[1]
        ref[2,1,1,0]=(-0.532645566013)*x_ref[0]**o+(0.571754809868)*x_ref[0]+(0.656737493773)*x_ref[1]**o+(-0.548229591842)*x_ref[1]
        ref[2,1,1,1]=(0.23836921087)*x_ref[0]**o+(0.707475737785)*x_ref[0]+(-0.59660155968)*x_ref[1]**o+(-0.450204951071)*x_ref[1]
        ref[2,1,2,0]=(-0.534027348014)*x_ref[0]**o+(0.332100193727)*x_ref[0]+(-0.864157138567)*x_ref[1]**o+(-0.41833267066)*x_ref[1]
        ref[2,1,2,1]=(0.459878593269)*x_ref[0]**o+(0.446259140228)*x_ref[0]+(0.571242887502)*x_ref[1]**o+(-0.0390525669486)*x_ref[1]
        ref[2,2,0,0]=(0.322052799527)*x_ref[0]**o+(-0.915265209304)*x_ref[0]+(-0.99942007618)*x_ref[1]**o+(0.482294513743)*x_ref[1]
        ref[2,2,0,1]=(0.558489194456)*x_ref[0]**o+(0.00400576554086)*x_ref[0]+(-0.87484330165)*x_ref[1]**o+(-0.664504218977)*x_ref[1]
        ref[2,2,1,0]=(0.669621607461)*x_ref[0]**o+(-0.0450960554319)*x_ref[0]+(-0.221151480246)*x_ref[1]**o+(0.0123621737211)*x_ref[1]
        ref[2,2,1,1]=(0.476743013029)*x_ref[0]**o+(0.11118371822)*x_ref[0]+(-0.202074315677)*x_ref[1]**o+(0.377649433458)*x_ref[1]
        ref[2,2,2,0]=(0.667910543701)*x_ref[0]**o+(0.035703922547)*x_ref[0]+(-0.274733967095)*x_ref[1]**o+(-0.998934720182)*x_ref[1]
        ref[2,2,2,1]=(0.578909660087)*x_ref[0]**o+(0.149900086514)*x_ref[0]+(-0.226597651858)*x_ref[1]**o+(0.169175812287)*x_ref[1]
        ref[2,3,0,0]=(0.158370246586)*x_ref[0]**o+(0.595643112285)*x_ref[0]+(-0.476875564052)*x_ref[1]**o+(0.869752657517)*x_ref[1]
        ref[2,3,0,1]=(0.99718847714)*x_ref[0]**o+(0.581531822297)*x_ref[0]+(-0.607102775564)*x_ref[1]**o+(-0.8317970284)*x_ref[1]
        ref[2,3,1,0]=(-0.27390195745)*x_ref[0]**o+(0.0512853330414)*x_ref[0]+(0.465190666302)*x_ref[1]**o+(0.042533388788)*x_ref[1]
        ref[2,3,1,1]=(-0.183679504962)*x_ref[0]**o+(-0.0522782076121)*x_ref[0]+(0.921423118529)*x_ref[1]**o+(-0.0315080275225)*x_ref[1]
        ref[2,3,2,0]=(0.0421842348778)*x_ref[0]**o+(0.462023049702)*x_ref[0]+(-0.888342606803)*x_ref[1]**o+(-0.956503447884)*x_ref[1]
        ref[2,3,2,1]=(-0.950820288444)*x_ref[0]**o+(0.104271990168)*x_ref[0]+(-0.607212909582)*x_ref[1]**o+(-0.16283931259)*x_ref[1]
        ref[2,4,0,0]=(0.113505933533)*x_ref[0]**o+(0.27873514806)*x_ref[0]+(0.736699443523)*x_ref[1]**o+(-0.951737555693)*x_ref[1]
        ref[2,4,0,1]=(0.264170584849)*x_ref[0]**o+(-0.226649236617)*x_ref[0]+(0.489902180874)*x_ref[1]**o+(-0.191536591762)*x_ref[1]
        ref[2,4,1,0]=(0.712322285985)*x_ref[0]**o+(-0.323395455815)*x_ref[0]+(-0.298356681586)*x_ref[1]**o+(0.699891974541)*x_ref[1]
        ref[2,4,1,1]=(-0.631057763255)*x_ref[0]**o+(-0.489374933445)*x_ref[0]+(-0.021591806874)*x_ref[1]**o+(-0.692800124585)*x_ref[1]
        ref[2,4,2,0]=(0.916068935184)*x_ref[0]**o+(0.33779688858)*x_ref[0]+(0.06300443559)*x_ref[1]**o+(-0.231385203399)*x_ref[1]
        ref[2,4,2,1]=(-0.283485044625)*x_ref[0]**o+(0.548366574442)*x_ref[0]+(-0.543953132872)*x_ref[1]**o+(-0.956662823037)*x_ref[1]
        ref[3,0,0,0]=(0.709349274221)*x_ref[0]**o+(-0.342981009263)*x_ref[0]+(-0.682331761929)*x_ref[1]**o+(0.673169884813)*x_ref[1]
        ref[3,0,0,1]=(0.236249836593)*x_ref[0]**o+(-0.682845054434)*x_ref[0]+(-0.349169988218)*x_ref[1]**o+(0.502684425824)*x_ref[1]
        ref[3,0,1,0]=(-0.561655748793)*x_ref[0]**o+(0.120183772326)*x_ref[0]+(-0.966535325583)*x_ref[1]**o+(-0.0585117264949)*x_ref[1]
        ref[3,0,1,1]=(-0.661030287189)*x_ref[0]**o+(-0.322924086505)*x_ref[0]+(-0.986892059552)*x_ref[1]**o+(-0.778969361365)*x_ref[1]
        ref[3,0,2,0]=(-0.166307107307)*x_ref[0]**o+(0.612931303617)*x_ref[0]+(-0.283925165456)*x_ref[1]**o+(0.5998038814)*x_ref[1]
        ref[3,0,2,1]=(0.205099355916)*x_ref[0]**o+(-0.520970536278)*x_ref[0]+(0.900416833183)*x_ref[1]**o+(0.981826306069)*x_ref[1]
        ref[3,1,0,0]=(-0.875306666443)*x_ref[0]**o+(0.0189358479861)*x_ref[0]+(0.101187857445)*x_ref[1]**o+(-0.873584307534)*x_ref[1]
        ref[3,1,0,1]=(-0.0203656235757)*x_ref[0]**o+(-0.579229777042)*x_ref[0]+(-0.0799345631826)*x_ref[1]**o+(0.896315444069)*x_ref[1]
        ref[3,1,1,0]=(0.576791459982)*x_ref[0]**o+(-0.86541815506)*x_ref[0]+(-0.582764176207)*x_ref[1]**o+(-0.249437395011)*x_ref[1]
        ref[3,1,1,1]=(0.525953680497)*x_ref[0]**o+(-0.222445717402)*x_ref[0]+(-0.87710202063)*x_ref[1]**o+(-0.762921821973)*x_ref[1]
        ref[3,1,2,0]=(-0.883737185908)*x_ref[0]**o+(0.254053241277)*x_ref[0]+(0.559803298059)*x_ref[1]**o+(-0.577376956856)*x_ref[1]
        ref[3,1,2,1]=(-0.125792083387)*x_ref[0]**o+(-0.771549307509)*x_ref[0]+(0.142493480552)*x_ref[1]**o+(0.459996962548)*x_ref[1]
        ref[3,2,0,0]=(0.0531598927049)*x_ref[0]**o+(-0.950343229037)*x_ref[0]+(-0.748240453154)*x_ref[1]**o+(0.850555892342)*x_ref[1]
        ref[3,2,0,1]=(-0.660852152041)*x_ref[0]**o+(-0.763119278604)*x_ref[0]+(-0.811925409663)*x_ref[1]**o+(-0.178765071495)*x_ref[1]
        ref[3,2,1,0]=(-0.189124223305)*x_ref[0]**o+(-0.410392909042)*x_ref[0]+(0.993247167522)*x_ref[1]**o+(0.291451796298)*x_ref[1]
        ref[3,2,1,1]=(0.83496133421)*x_ref[0]**o+(0.695147480265)*x_ref[0]+(0.444420524946)*x_ref[1]**o+(0.107454620114)*x_ref[1]
        ref[3,2,2,0]=(0.94687479313)*x_ref[0]**o+(-0.140702324395)*x_ref[0]+(0.149203502428)*x_ref[1]**o+(-0.338246126345)*x_ref[1]
        ref[3,2,2,1]=(0.147609394709)*x_ref[0]**o+(-0.549219765239)*x_ref[0]+(0.397308993347)*x_ref[1]**o+(0.0402189312685)*x_ref[1]
        ref[3,3,0,0]=(0.27400981409)*x_ref[0]**o+(-0.516367266257)*x_ref[0]+(-0.784095952814)*x_ref[1]**o+(0.32858899009)*x_ref[1]
        ref[3,3,0,1]=(-0.393457382163)*x_ref[0]**o+(0.590406794051)*x_ref[0]+(0.328523918357)*x_ref[1]**o+(0.899537513355)*x_ref[1]
        ref[3,3,1,0]=(-0.851038623366)*x_ref[0]**o+(0.596974864726)*x_ref[0]+(-0.338630687551)*x_ref[1]**o+(0.634961435042)*x_ref[1]
        ref[3,3,1,1]=(0.569166143918)*x_ref[0]**o+(-0.388162282491)*x_ref[0]+(0.764830536733)*x_ref[1]**o+(-0.426737342542)*x_ref[1]
        ref[3,3,2,0]=(0.502221903448)*x_ref[0]**o+(0.940382225897)*x_ref[0]+(0.563444730579)*x_ref[1]**o+(0.023210869151)*x_ref[1]
        ref[3,3,2,1]=(0.061695403852)*x_ref[0]**o+(0.868302188758)*x_ref[0]+(0.0855803325758)*x_ref[1]**o+(-0.24366173638)*x_ref[1]
        ref[3,4,0,0]=(-0.533038182387)*x_ref[0]**o+(0.0222388915737)*x_ref[0]+(0.529037752769)*x_ref[1]**o+(-0.0665320895494)*x_ref[1]
        ref[3,4,0,1]=(0.714374055253)*x_ref[0]**o+(-0.623670743333)*x_ref[0]+(-0.866024792929)*x_ref[1]**o+(-0.0673124896111)*x_ref[1]
        ref[3,4,1,0]=(-0.0523466615045)*x_ref[0]**o+(-0.472563426622)*x_ref[0]+(0.139892673626)*x_ref[1]**o+(0.539655417178)*x_ref[1]
        ref[3,4,1,1]=(-0.473032580514)*x_ref[0]**o+(0.692812093618)*x_ref[0]+(0.664400517742)*x_ref[1]**o+(0.184125423364)*x_ref[1]
        ref[3,4,2,0]=(-0.0340798058725)*x_ref[0]**o+(-0.58900562863)*x_ref[0]+(0.396856384717)*x_ref[1]**o+(-0.559175428019)*x_ref[1]
        ref[3,4,2,1]=(-0.217679376634)*x_ref[0]**o+(-0.90389036667)*x_ref[0]+(0.967372942873)*x_ref[1]**o+(-0.482913619277)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.93229223906)*x[0]**o+(0.176839936782)*x[0]+(0.986271000359)*x[1]**o+(-0.449322385723)*x[1]+(-0.856450437823)*x[2]**o+(0.31473579221)*x[2]
        arg[0,0,0,1]=(0.714047232765)*x[0]**o+(-0.323471323003)*x[0]+(0.628080480343)*x[1]**o+(-0.992007912045)*x[1]+(-0.11019723051)*x[2]**o+(-0.104735408233)*x[2]
        arg[0,0,1,0]=(-0.437250443294)*x[0]**o+(-0.741469652781)*x[0]+(0.650921038465)*x[1]**o+(0.679622741243)*x[1]+(0.604005894895)*x[2]**o+(0.598155624876)*x[2]
        arg[0,0,1,1]=(-0.574872168761)*x[0]**o+(0.796207672391)*x[0]+(-0.478046646159)*x[1]**o+(0.604661326741)*x[1]+(-0.898457020186)*x[2]**o+(0.43234750827)*x[2]
        arg[0,0,2,0]=(-0.674008357742)*x[0]**o+(0.699614371016)*x[0]+(0.441321981559)*x[1]**o+(-0.619688689942)*x[1]+(-0.0874579208832)*x[2]**o+(-0.925753840555)*x[2]
        arg[0,0,2,1]=(-0.114329474892)*x[0]**o+(0.699501071503)*x[0]+(-0.618807420683)*x[1]**o+(0.411584695002)*x[1]+(-0.234620261556)*x[2]**o+(0.585720100967)*x[2]
        arg[0,1,0,0]=(0.617432097253)*x[0]**o+(-0.00304868863676)*x[0]+(0.571021785645)*x[1]**o+(0.457197205728)*x[1]+(0.462895867813)*x[2]**o+(0.278022956316)*x[2]
        arg[0,1,0,1]=(-0.659136231188)*x[0]**o+(-0.388212920906)*x[0]+(-0.982152608232)*x[1]**o+(-0.665797156013)*x[1]+(-0.76141890649)*x[2]**o+(-0.591458323836)*x[2]
        arg[0,1,1,0]=(-0.488016386893)*x[0]**o+(0.965660401944)*x[0]+(0.178400189819)*x[1]**o+(0.578417314499)*x[1]+(-0.398083955217)*x[2]**o+(0.460338187535)*x[2]
        arg[0,1,1,1]=(-0.182086458235)*x[0]**o+(0.0842368498629)*x[0]+(-0.642008026184)*x[1]**o+(-0.0449408131957)*x[1]+(-0.927397573018)*x[2]**o+(0.854313729816)*x[2]
        arg[0,1,2,0]=(-0.201475872769)*x[0]**o+(0.098190765292)*x[0]+(0.808624244922)*x[1]**o+(0.0726268899898)*x[1]+(-0.605334327027)*x[2]**o+(-0.999151255934)*x[2]
        arg[0,1,2,1]=(0.672344024218)*x[0]**o+(-0.638554833481)*x[0]+(-0.212231801634)*x[1]**o+(-0.492765366968)*x[1]+(0.311273771921)*x[2]**o+(-0.504370655006)*x[2]
        arg[0,2,0,0]=(0.9944935907)*x[0]**o+(-0.365261996611)*x[0]+(-0.105667939789)*x[1]**o+(0.886935751977)*x[1]+(0.263336157757)*x[2]**o+(-0.880316523113)*x[2]
        arg[0,2,0,1]=(-0.659779300787)*x[0]**o+(-0.728590180152)*x[0]+(0.732078508037)*x[1]**o+(0.684963412098)*x[1]+(0.822970700012)*x[2]**o+(-0.680001879674)*x[2]
        arg[0,2,1,0]=(-0.207855412893)*x[0]**o+(0.438275426529)*x[0]+(0.87046682066)*x[1]**o+(-0.960322013848)*x[1]+(-0.273728198625)*x[2]**o+(-0.960137228004)*x[2]
        arg[0,2,1,1]=(-0.189232680002)*x[0]**o+(0.866417302002)*x[0]+(-0.93403311248)*x[1]**o+(-0.55773104878)*x[1]+(0.94062895773)*x[2]**o+(-0.758052194805)*x[2]
        arg[0,2,2,0]=(0.100859944394)*x[0]**o+(-0.20292135909)*x[0]+(-0.713339941261)*x[1]**o+(-0.615106454411)*x[1]+(0.791245995152)*x[2]**o+(-0.81051791594)*x[2]
        arg[0,2,2,1]=(0.595743424517)*x[0]**o+(-0.179075876404)*x[0]+(-0.799005601715)*x[1]**o+(-0.0554635073569)*x[1]+(-0.850510701985)*x[2]**o+(0.234516538169)*x[2]
        arg[0,3,0,0]=(-0.361667738861)*x[0]**o+(-0.551837498696)*x[0]+(-0.834377202094)*x[1]**o+(0.814168224265)*x[1]+(-0.00479533268377)*x[2]**o+(0.843007652951)*x[2]
        arg[0,3,0,1]=(-0.448417439465)*x[0]**o+(0.431496158388)*x[0]+(-0.547416443433)*x[1]**o+(-0.0746785704695)*x[1]+(-0.646238182944)*x[2]**o+(-0.318608714595)*x[2]
        arg[0,3,1,0]=(-0.275647366007)*x[0]**o+(-0.410594785047)*x[0]+(-0.858282378196)*x[1]**o+(-0.755586690837)*x[1]+(-0.420481091654)*x[2]**o+(-0.0996487371831)*x[2]
        arg[0,3,1,1]=(-0.395961249009)*x[0]**o+(0.0860896467863)*x[0]+(0.482719525748)*x[1]**o+(0.319368178133)*x[1]+(0.516449153414)*x[2]**o+(-0.246749862576)*x[2]
        arg[0,3,2,0]=(-0.367322027122)*x[0]**o+(0.687206412524)*x[0]+(0.616265233972)*x[1]**o+(-0.32660589112)*x[1]+(-0.944607343096)*x[2]**o+(0.123982537951)*x[2]
        arg[0,3,2,1]=(0.700497166804)*x[0]**o+(0.553961329865)*x[0]+(-0.34020003754)*x[1]**o+(0.295870992783)*x[1]+(-0.220275880046)*x[2]**o+(0.766121035825)*x[2]
        arg[0,4,0,0]=(0.405469789242)*x[0]**o+(-0.338347662873)*x[0]+(0.0761796434233)*x[1]**o+(-0.120599175033)*x[1]+(0.107865585379)*x[2]**o+(0.0711941007838)*x[2]
        arg[0,4,0,1]=(0.614166154899)*x[0]**o+(0.195780702963)*x[0]+(0.718935985215)*x[1]**o+(0.214577296792)*x[1]+(-0.366985242107)*x[2]**o+(-0.509584787633)*x[2]
        arg[0,4,1,0]=(-0.638646370981)*x[0]**o+(-0.769955899284)*x[0]+(-0.887407540503)*x[1]**o+(0.0901245429998)*x[1]+(0.786085641321)*x[2]**o+(0.6248068552)*x[2]
        arg[0,4,1,1]=(-0.145426222043)*x[0]**o+(-0.494114545016)*x[0]+(0.422153721335)*x[1]**o+(0.617022143901)*x[1]+(-0.665972324723)*x[2]**o+(-0.476724432799)*x[2]
        arg[0,4,2,0]=(0.556105252092)*x[0]**o+(-0.893268696809)*x[0]+(0.146441889489)*x[1]**o+(-0.170512113283)*x[1]+(-0.869313262875)*x[2]**o+(-0.976874109392)*x[2]
        arg[0,4,2,1]=(0.858927206499)*x[0]**o+(0.198998355146)*x[0]+(0.278125296049)*x[1]**o+(-0.0442518828093)*x[1]+(0.274852724203)*x[2]**o+(-0.167649193314)*x[2]
        arg[1,0,0,0]=(-0.0429420289353)*x[0]**o+(-0.15764693929)*x[0]+(-0.859599917512)*x[1]**o+(-0.904178643575)*x[1]+(0.519560711493)*x[2]**o+(0.258644717907)*x[2]
        arg[1,0,0,1]=(-0.185321114264)*x[0]**o+(-0.794493830004)*x[0]+(0.136084421162)*x[1]**o+(0.926912021419)*x[1]+(0.346392120388)*x[2]**o+(-0.183908371454)*x[2]
        arg[1,0,1,0]=(-0.730675509877)*x[0]**o+(0.865034802778)*x[0]+(-0.506315431878)*x[1]**o+(-0.94607761051)*x[1]+(0.195194792801)*x[2]**o+(-0.865129258879)*x[2]
        arg[1,0,1,1]=(-0.0597026648107)*x[0]**o+(-0.884790497)*x[0]+(-0.0631278742098)*x[1]**o+(-0.808364940737)*x[1]+(-0.968952604837)*x[2]**o+(0.142246702632)*x[2]
        arg[1,0,2,0]=(0.648472894563)*x[0]**o+(0.145861644073)*x[0]+(0.501227520574)*x[1]**o+(0.265886482306)*x[1]+(-0.0472772173573)*x[2]**o+(-0.364169655969)*x[2]
        arg[1,0,2,1]=(0.847863600695)*x[0]**o+(-0.288841685199)*x[0]+(-0.500273502795)*x[1]**o+(0.627379277592)*x[1]+(0.326814024625)*x[2]**o+(-0.514712825895)*x[2]
        arg[1,1,0,0]=(-0.947845713712)*x[0]**o+(-0.372115616305)*x[0]+(-0.572858279998)*x[1]**o+(0.226237627692)*x[1]+(0.210201947182)*x[2]**o+(0.0147616538594)*x[2]
        arg[1,1,0,1]=(-0.39813630466)*x[0]**o+(-0.658496395731)*x[0]+(-0.363906465005)*x[1]**o+(-0.816719661238)*x[1]+(-0.478106835032)*x[2]**o+(0.0819042309589)*x[2]
        arg[1,1,1,0]=(-0.287160278593)*x[0]**o+(-0.0281402640978)*x[0]+(0.41559795378)*x[1]**o+(0.158274458097)*x[1]+(-0.798429481502)*x[2]**o+(-0.278585334624)*x[2]
        arg[1,1,1,1]=(0.809550480973)*x[0]**o+(0.192026915854)*x[0]+(-0.891168838599)*x[1]**o+(0.0482621192882)*x[1]+(-0.35339708159)*x[2]**o+(-0.133161551553)*x[2]
        arg[1,1,2,0]=(0.939582328872)*x[0]**o+(0.561963790516)*x[0]+(0.885264722949)*x[1]**o+(-0.468042305266)*x[1]+(0.232880002429)*x[2]**o+(-0.772896005701)*x[2]
        arg[1,1,2,1]=(-0.297333303209)*x[0]**o+(-0.414050924207)*x[0]+(-0.278903698873)*x[1]**o+(0.345370620869)*x[1]+(-0.0570151259027)*x[2]**o+(0.786449641378)*x[2]
        arg[1,2,0,0]=(0.644047138589)*x[0]**o+(-0.480660438838)*x[0]+(0.32526499462)*x[1]**o+(-0.775988282127)*x[1]+(0.108374143932)*x[2]**o+(0.955463667181)*x[2]
        arg[1,2,0,1]=(-0.468236790339)*x[0]**o+(0.844197711724)*x[0]+(0.808973429831)*x[1]**o+(-0.647699550005)*x[1]+(0.552555959813)*x[2]**o+(-0.3403079515)*x[2]
        arg[1,2,1,0]=(0.864686362308)*x[0]**o+(0.248923646116)*x[0]+(0.0287042863842)*x[1]**o+(0.404974282153)*x[1]+(0.113501438234)*x[2]**o+(-0.619364894397)*x[2]
        arg[1,2,1,1]=(-0.625146992499)*x[0]**o+(-0.0480602452983)*x[0]+(0.168910550914)*x[1]**o+(-0.891584218123)*x[1]+(0.0609357618263)*x[2]**o+(0.0451921002161)*x[2]
        arg[1,2,2,0]=(-0.478728628752)*x[0]**o+(0.428376547798)*x[0]+(0.817203376818)*x[1]**o+(-0.856056266953)*x[1]+(0.354960313382)*x[2]**o+(-0.611468960461)*x[2]
        arg[1,2,2,1]=(0.49500742113)*x[0]**o+(0.902948130263)*x[0]+(-0.810748543744)*x[1]**o+(-0.696150924858)*x[1]+(-0.748278516161)*x[2]**o+(0.694815527058)*x[2]
        arg[1,3,0,0]=(-0.361960150296)*x[0]**o+(-0.845139780461)*x[0]+(-0.0624499309665)*x[1]**o+(-0.0515558058489)*x[1]+(0.400690983017)*x[2]**o+(-0.744608407974)*x[2]
        arg[1,3,0,1]=(-0.452453617467)*x[0]**o+(0.845396139593)*x[0]+(0.292319469936)*x[1]**o+(-0.833547518634)*x[1]+(-0.511797700353)*x[2]**o+(0.857891454773)*x[2]
        arg[1,3,1,0]=(0.779504092912)*x[0]**o+(0.500524773615)*x[0]+(0.590542659382)*x[1]**o+(0.479310630519)*x[1]+(0.509960641956)*x[2]**o+(-0.564980432959)*x[2]
        arg[1,3,1,1]=(-0.145745189298)*x[0]**o+(-0.256064576359)*x[0]+(0.640890837126)*x[1]**o+(-0.718802035152)*x[1]+(-0.322216873583)*x[2]**o+(-0.0246526526334)*x[2]
        arg[1,3,2,0]=(-0.463434285098)*x[0]**o+(0.116722448643)*x[0]+(-0.690378381342)*x[1]**o+(0.429197465885)*x[1]+(-0.353019731384)*x[2]**o+(0.57397712918)*x[2]
        arg[1,3,2,1]=(0.169676853383)*x[0]**o+(0.815225472623)*x[0]+(-0.11125527552)*x[1]**o+(0.391513831525)*x[1]+(-0.587041361585)*x[2]**o+(0.255194151616)*x[2]
        arg[1,4,0,0]=(0.325351608628)*x[0]**o+(-0.57751715453)*x[0]+(0.145167511303)*x[1]**o+(-0.969286030742)*x[1]+(0.917370398478)*x[2]**o+(0.691483691598)*x[2]
        arg[1,4,0,1]=(0.0997998298271)*x[0]**o+(-0.101248202247)*x[0]+(-0.650649653848)*x[1]**o+(-0.157887947288)*x[1]+(0.517501747765)*x[2]**o+(-0.0852921926298)*x[2]
        arg[1,4,1,0]=(0.984595420407)*x[0]**o+(0.499293499076)*x[0]+(0.352193115874)*x[1]**o+(-0.684522909647)*x[1]+(-0.391538373605)*x[2]**o+(0.404635595162)*x[2]
        arg[1,4,1,1]=(-0.231310482173)*x[0]**o+(-0.0385089836559)*x[0]+(0.247671691888)*x[1]**o+(0.883971955027)*x[1]+(-0.0856624709734)*x[2]**o+(0.942940875904)*x[2]
        arg[1,4,2,0]=(-0.471467257434)*x[0]**o+(0.835634240897)*x[0]+(-0.594421294562)*x[1]**o+(-0.187584387111)*x[1]+(0.859867472765)*x[2]**o+(0.930951154576)*x[2]
        arg[1,4,2,1]=(0.504416975601)*x[0]**o+(0.435692923635)*x[0]+(-0.296945944313)*x[1]**o+(0.815156218765)*x[1]+(0.464464247442)*x[2]**o+(-0.378809453041)*x[2]
        arg[2,0,0,0]=(0.213087423107)*x[0]**o+(-0.504681665801)*x[0]+(0.796352190405)*x[1]**o+(0.546752751136)*x[1]+(0.57347644664)*x[2]**o+(0.715856963195)*x[2]
        arg[2,0,0,1]=(-0.28455266769)*x[0]**o+(0.853162220766)*x[0]+(0.164930982863)*x[1]**o+(-0.0136996498765)*x[1]+(0.293717915079)*x[2]**o+(-0.216124158473)*x[2]
        arg[2,0,1,0]=(-0.328115989441)*x[0]**o+(-0.529441496488)*x[0]+(-0.43579479677)*x[1]**o+(-0.851333801194)*x[1]+(-0.278406102972)*x[2]**o+(0.947777471824)*x[2]
        arg[2,0,1,1]=(-0.756438938789)*x[0]**o+(-0.624332272735)*x[0]+(-0.591638429565)*x[1]**o+(-0.0608281771566)*x[1]+(-0.852078784085)*x[2]**o+(-0.526558964866)*x[2]
        arg[2,0,2,0]=(-0.923892426969)*x[0]**o+(0.788121899878)*x[0]+(-0.446006331202)*x[1]**o+(0.997210871574)*x[1]+(0.356998606103)*x[2]**o+(0.972643822823)*x[2]
        arg[2,0,2,1]=(0.225183308571)*x[0]**o+(-0.757505414777)*x[0]+(-0.59573192074)*x[1]**o+(0.57166154987)*x[1]+(0.622343198809)*x[2]**o+(0.60343970503)*x[2]
        arg[2,1,0,0]=(0.51027702934)*x[0]**o+(-0.711737679522)*x[0]+(-0.211712579627)*x[1]**o+(-0.495817605314)*x[1]+(0.924425804272)*x[2]**o+(0.753424835275)*x[2]
        arg[2,1,0,1]=(0.476218252577)*x[0]**o+(0.187579891557)*x[0]+(0.727359583623)*x[1]**o+(-0.389875559955)*x[1]+(-0.352889326477)*x[2]**o+(0.365143943637)*x[2]
        arg[2,1,1,0]=(-0.815704312943)*x[0]**o+(0.315217470245)*x[0]+(-0.535234555065)*x[1]**o+(-0.433408988501)*x[1]+(-0.676694087389)*x[2]**o+(0.384097177718)*x[2]
        arg[2,1,1,1]=(-0.0368452411345)*x[0]**o+(0.514540686444)*x[0]+(-0.527100066965)*x[1]**o+(0.0291015562912)*x[1]+(-0.605646905004)*x[2]**o+(0.931528599981)*x[2]
        arg[2,1,2,0]=(-0.0994015945266)*x[0]**o+(-0.762310347434)*x[0]+(-0.845934950466)*x[1]**o+(-0.235758967923)*x[1]+(0.755531422698)*x[2]**o+(0.108353867777)*x[2]
        arg[2,1,2,1]=(-0.868779052889)*x[0]**o+(0.783878941075)*x[0]+(0.900089560473)*x[1]**o+(-0.836779405877)*x[1]+(-0.799571882449)*x[2]**o+(0.999710029157)*x[2]
        arg[2,2,0,0]=(0.541465277143)*x[0]**o+(-0.840477797945)*x[0]+(-0.362577721098)*x[1]**o+(0.00680131806718)*x[1]+(-0.338485239867)*x[2]**o+(0.828213756751)*x[2]
        arg[2,2,0,1]=(-0.927347423175)*x[0]**o+(-0.936191125785)*x[0]+(-0.411819804833)*x[1]**o+(-0.0222067637269)*x[1]+(-0.104411832586)*x[2]**o+(-0.79937203081)*x[2]
        arg[2,2,1,0]=(0.273051712848)*x[0]**o+(-0.802468557957)*x[0]+(-0.019392898867)*x[1]**o+(-0.612803074784)*x[1]+(-0.269705464223)*x[2]**o+(-0.244405147706)*x[2]
        arg[2,2,1,1]=(-0.08869649035)*x[0]**o+(0.809761782447)*x[0]+(0.761420080249)*x[1]**o+(0.16590038557)*x[1]+(-0.405657593445)*x[2]**o+(0.123593603037)*x[2]
        arg[2,2,2,0]=(-0.799691160987)*x[0]**o+(0.834835690475)*x[0]+(-0.0863569221737)*x[1]**o+(0.988343215152)*x[1]+(0.294423283474)*x[2]**o+(0.325725138203)*x[2]
        arg[2,2,2,1]=(0.648897496592)*x[0]**o+(-0.474329387314)*x[0]+(-0.904613697104)*x[1]**o+(0.382201388286)*x[1]+(-0.596889148339)*x[2]**o+(0.705132157684)*x[2]
        arg[2,3,0,0]=(0.906738144045)*x[0]**o+(-0.365392026426)*x[0]+(0.730557428733)*x[1]**o+(-0.857864471189)*x[1]+(-0.160930657186)*x[2]**o+(0.885067383334)*x[2]
        arg[2,3,0,1]=(0.992764702472)*x[0]**o+(0.720688532948)*x[0]+(0.318869772818)*x[1]**o+(0.995508347248)*x[1]+(0.108569456996)*x[2]**o+(-0.990939292995)*x[2]
        arg[2,3,1,0]=(0.690584992096)*x[0]**o+(0.605885242246)*x[0]+(0.72371794265)*x[1]**o+(-0.632609325879)*x[1]+(0.520273588734)*x[2]**o+(0.885804143155)*x[2]
        arg[2,3,1,1]=(0.485231542482)*x[0]**o+(0.866920225418)*x[0]+(-0.814964945972)*x[1]**o+(0.544132513209)*x[1]+(-0.155071376494)*x[2]**o+(-0.941300746237)*x[2]
        arg[2,3,2,0]=(0.748232238388)*x[0]**o+(0.179397724419)*x[0]+(-0.43881886248)*x[1]**o+(-0.124941799624)*x[1]+(0.14679335819)*x[2]**o+(-0.758738846688)*x[2]
        arg[2,3,2,1]=(-0.0951333655337)*x[0]**o+(0.37270972476)*x[0]+(-0.651977554115)*x[1]**o+(-0.88918597476)*x[1]+(-0.730858685576)*x[2]**o+(0.685881924459)*x[2]
        arg[2,4,0,0]=(-0.823487466802)*x[0]**o+(-0.583630855643)*x[0]+(-0.664814562747)*x[1]**o+(0.860952605956)*x[1]+(-0.249114322668)*x[2]**o+(-0.968356584237)*x[2]
        arg[2,4,0,1]=(0.44040146252)*x[0]**o+(-0.246827565902)*x[0]+(0.596300980733)*x[1]**o+(-0.825818200414)*x[1]+(0.0715512875967)*x[2]**o+(0.996064679803)*x[2]
        arg[2,4,1,0]=(-0.287806453021)*x[0]**o+(0.704331542927)*x[0]+(-0.346139754521)*x[1]**o+(-0.766667037125)*x[1]+(0.283874132105)*x[2]**o+(-0.613145979204)*x[2]
        arg[2,4,1,1]=(-0.116649875156)*x[0]**o+(0.291380048239)*x[0]+(0.689809221905)*x[1]**o+(0.758935910592)*x[1]+(0.505142787603)*x[2]**o+(-0.113611480503)*x[2]
        arg[2,4,2,0]=(-0.836908572692)*x[0]**o+(-0.21892828725)*x[0]+(-0.935780103868)*x[1]**o+(0.106204643268)*x[1]+(-0.894474423807)*x[2]**o+(0.296554773217)*x[2]
        arg[2,4,2,1]=(-0.89413495023)*x[0]**o+(0.6080813058)*x[0]+(0.087863889878)*x[1]**o+(0.213373665612)*x[1]+(0.17085345025)*x[2]**o+(-0.967625138712)*x[2]
        arg[3,0,0,0]=(0.327537677409)*x[0]**o+(-0.299373059373)*x[0]+(-0.43223648542)*x[1]**o+(-0.499115781051)*x[1]+(0.145464567449)*x[2]**o+(0.580930584563)*x[2]
        arg[3,0,0,1]=(-0.122622543351)*x[0]**o+(-0.220547821081)*x[0]+(0.892679827803)*x[1]**o+(0.436584604381)*x[1]+(-0.156093580288)*x[2]**o+(0.464069069075)*x[2]
        arg[3,0,1,0]=(0.590288663203)*x[0]**o+(-0.212929445251)*x[0]+(0.0899671639061)*x[1]**o+(-0.87729854872)*x[1]+(0.789506558716)*x[2]**o+(-0.68867690907)*x[2]
        arg[3,0,1,1]=(-0.913651849012)*x[0]**o+(-0.430938918892)*x[0]+(-0.351186012371)*x[1]**o+(0.500306994748)*x[1]+(0.145381845281)*x[2]**o+(0.624993552333)*x[2]
        arg[3,0,2,0]=(0.281393881265)*x[0]**o+(-0.704727129826)*x[0]+(-0.0102091836449)*x[1]**o+(0.311857501353)*x[1]+(0.300732947904)*x[2]**o+(0.923973499548)*x[2]
        arg[3,0,2,1]=(-0.584907526993)*x[0]**o+(0.625732316822)*x[0]+(-0.609691930408)*x[1]**o+(0.056160226956)*x[1]+(-0.766583561947)*x[2]**o+(0.307692190194)*x[2]
        arg[3,1,0,0]=(0.731231379445)*x[0]**o+(-0.0485980527115)*x[0]+(0.908932758852)*x[1]**o+(0.188918035886)*x[1]+(0.685837743688)*x[2]**o+(-0.299987553989)*x[2]
        arg[3,1,0,1]=(-0.784006561231)*x[0]**o+(-0.809789430022)*x[0]+(-0.846391792735)*x[1]**o+(-0.211583304405)*x[1]+(-0.00198590063687)*x[2]**o+(-0.853253255224)*x[2]
        arg[3,1,1,0]=(0.595567299132)*x[0]**o+(-0.581734416878)*x[0]+(0.88144481486)*x[1]**o+(0.776876564509)*x[1]+(-0.367477971386)*x[2]**o+(0.0806320164453)*x[2]
        arg[3,1,1,1]=(0.440836187611)*x[0]**o+(-0.578925091757)*x[0]+(0.923367189832)*x[1]**o+(-0.993888173929)*x[1]+(-0.645514380411)*x[2]**o+(0.0267439044199)*x[2]
        arg[3,1,2,0]=(-0.305182288788)*x[0]**o+(-0.503372502553)*x[0]+(0.740948657389)*x[1]**o+(0.285661052275)*x[1]+(-0.634510984218)*x[2]**o+(-0.123979746456)*x[2]
        arg[3,1,2,1]=(-0.897274928495)*x[0]**o+(0.971356796164)*x[0]+(0.940768290313)*x[1]**o+(-0.552941985667)*x[1]+(0.172857451423)*x[2]**o+(0.844758237523)*x[2]
        arg[3,2,0,0]=(-0.903542725333)*x[0]**o+(0.936704774551)*x[0]+(-0.398085875228)*x[1]**o+(-0.304061573512)*x[1]+(-0.13362037229)*x[2]**o+(-0.405043350728)*x[2]
        arg[3,2,0,1]=(-0.0208590751801)*x[0]**o+(-0.354086381885)*x[0]+(0.510884698744)*x[1]**o+(-0.275504880518)*x[1]+(0.459476496356)*x[2]**o+(0.25358097976)*x[2]
        arg[3,2,1,0]=(-0.0171081877107)*x[0]**o+(-0.856081626112)*x[0]+(-0.0698686571004)*x[1]**o+(0.0954918361122)*x[1]+(0.938494954069)*x[2]**o+(-0.824928688248)*x[2]
        arg[3,2,1,1]=(-0.277307915098)*x[0]**o+(0.737855779658)*x[0]+(-0.466761948855)*x[1]**o+(-0.821498435809)*x[1]+(-0.391009601442)*x[2]**o+(0.643834542048)*x[2]
        arg[3,2,2,0]=(0.379015007885)*x[0]**o+(0.348646179575)*x[0]+(0.733765133901)*x[1]**o+(-0.722363571789)*x[1]+(0.488531791931)*x[2]**o+(0.279813923679)*x[2]
        arg[3,2,2,1]=(0.516708382119)*x[0]**o+(0.0426665103314)*x[0]+(-0.168751315718)*x[1]**o+(-0.124069236608)*x[1]+(-0.778866001674)*x[2]**o+(-0.362031001279)*x[2]
        arg[3,3,0,0]=(0.995284803465)*x[0]**o+(-0.442327755879)*x[0]+(-0.13110270925)*x[1]**o+(-0.172857543102)*x[1]+(-0.910522664379)*x[2]**o+(0.516931833163)*x[2]
        arg[3,3,0,1]=(-0.889320295626)*x[0]**o+(-0.728330379534)*x[0]+(-0.584091700627)*x[1]**o+(0.469914082337)*x[1]+(0.18874127891)*x[2]**o+(0.0846532039249)*x[2]
        arg[3,3,1,0]=(0.911523535462)*x[0]**o+(0.0924361220354)*x[0]+(-0.0822504494228)*x[1]**o+(0.00630249023591)*x[1]+(-0.180497330322)*x[2]**o+(0.817405577154)*x[2]
        arg[3,3,1,1]=(0.681872747813)*x[0]**o+(0.840680190206)*x[0]+(-0.477660295206)*x[1]**o+(0.0399086968056)*x[1]+(0.440403204247)*x[2]**o+(-0.220656197016)*x[2]
        arg[3,3,2,0]=(-0.967524407336)*x[0]**o+(0.896652274462)*x[0]+(0.735269328519)*x[1]**o+(-0.249017384614)*x[1]+(-0.573222311563)*x[2]**o+(-0.653800334954)*x[2]
        arg[3,3,2,1]=(0.10041632804)*x[0]**o+(0.167657178348)*x[0]+(0.990621134594)*x[1]**o+(-0.309063473348)*x[1]+(-0.688206827575)*x[2]**o+(0.0926806818663)*x[2]
        arg[3,4,0,0]=(-0.128678339518)*x[0]**o+(-0.974238160059)*x[0]+(-0.762599464772)*x[1]**o+(0.0233286358698)*x[1]+(-0.622481697973)*x[2]**o+(-0.080495303798)*x[2]
        arg[3,4,0,1]=(-0.380318230058)*x[0]**o+(0.568425682425)*x[0]+(0.787369916262)*x[1]**o+(0.296745245123)*x[1]+(-0.917621002414)*x[2]**o+(-0.696209404954)*x[2]
        arg[3,4,1,0]=(-0.382961468898)*x[0]**o+(0.672614453621)*x[0]+(0.565371503874)*x[1]**o+(-0.597331999912)*x[1]+(-0.153637846704)*x[2]**o+(-0.438159587906)*x[2]
        arg[3,4,1,1]=(-0.541043007799)*x[0]**o+(0.415600628159)*x[0]+(-0.793716270785)*x[1]**o+(0.780110667203)*x[1]+(-0.101309486185)*x[2]**o+(-0.138868352939)*x[2]
        arg[3,4,2,0]=(0.679815917569)*x[0]**o+(-0.800657537776)*x[0]+(-0.711956069893)*x[1]**o+(-0.575626146908)*x[1]+(0.184247083249)*x[2]**o+(-0.728118084396)*x[2]
        arg[3,4,2,1]=(-0.608089432748)*x[0]**o+(0.356378798828)*x[0]+(-0.575957599599)*x[1]**o+(0.599586825137)*x[1]+(-0.160475743299)*x[2]**o+(-0.070762356177)*x[2]
        ref[0,0,0,0]=(-0.93229223906)*x_ref[0]**o+(0.176839936782)*x_ref[0]+(0.986271000359)*x_ref[1]**o+(-0.449322385723)*x_ref[1]+(-0.856450437823)*x_ref[2]**o+(0.31473579221)*x_ref[2]
        ref[0,0,0,1]=(0.714047232765)*x_ref[0]**o+(-0.323471323003)*x_ref[0]+(0.628080480343)*x_ref[1]**o+(-0.992007912045)*x_ref[1]+(-0.11019723051)*x_ref[2]**o+(-0.104735408233)*x_ref[2]
        ref[0,0,1,0]=(-0.437250443294)*x_ref[0]**o+(-0.741469652781)*x_ref[0]+(0.650921038465)*x_ref[1]**o+(0.679622741243)*x_ref[1]+(0.604005894895)*x_ref[2]**o+(0.598155624876)*x_ref[2]
        ref[0,0,1,1]=(-0.574872168761)*x_ref[0]**o+(0.796207672391)*x_ref[0]+(-0.478046646159)*x_ref[1]**o+(0.604661326741)*x_ref[1]+(-0.898457020186)*x_ref[2]**o+(0.43234750827)*x_ref[2]
        ref[0,0,2,0]=(-0.674008357742)*x_ref[0]**o+(0.699614371016)*x_ref[0]+(0.441321981559)*x_ref[1]**o+(-0.619688689942)*x_ref[1]+(-0.0874579208832)*x_ref[2]**o+(-0.925753840555)*x_ref[2]
        ref[0,0,2,1]=(-0.114329474892)*x_ref[0]**o+(0.699501071503)*x_ref[0]+(-0.618807420683)*x_ref[1]**o+(0.411584695002)*x_ref[1]+(-0.234620261556)*x_ref[2]**o+(0.585720100967)*x_ref[2]
        ref[0,1,0,0]=(0.617432097253)*x_ref[0]**o+(-0.00304868863676)*x_ref[0]+(0.571021785645)*x_ref[1]**o+(0.457197205728)*x_ref[1]+(0.462895867813)*x_ref[2]**o+(0.278022956316)*x_ref[2]
        ref[0,1,0,1]=(-0.659136231188)*x_ref[0]**o+(-0.388212920906)*x_ref[0]+(-0.982152608232)*x_ref[1]**o+(-0.665797156013)*x_ref[1]+(-0.76141890649)*x_ref[2]**o+(-0.591458323836)*x_ref[2]
        ref[0,1,1,0]=(-0.488016386893)*x_ref[0]**o+(0.965660401944)*x_ref[0]+(0.178400189819)*x_ref[1]**o+(0.578417314499)*x_ref[1]+(-0.398083955217)*x_ref[2]**o+(0.460338187535)*x_ref[2]
        ref[0,1,1,1]=(-0.182086458235)*x_ref[0]**o+(0.0842368498629)*x_ref[0]+(-0.642008026184)*x_ref[1]**o+(-0.0449408131957)*x_ref[1]+(-0.927397573018)*x_ref[2]**o+(0.854313729816)*x_ref[2]
        ref[0,1,2,0]=(-0.201475872769)*x_ref[0]**o+(0.098190765292)*x_ref[0]+(0.808624244922)*x_ref[1]**o+(0.0726268899898)*x_ref[1]+(-0.605334327027)*x_ref[2]**o+(-0.999151255934)*x_ref[2]
        ref[0,1,2,1]=(0.672344024218)*x_ref[0]**o+(-0.638554833481)*x_ref[0]+(-0.212231801634)*x_ref[1]**o+(-0.492765366968)*x_ref[1]+(0.311273771921)*x_ref[2]**o+(-0.504370655006)*x_ref[2]
        ref[0,2,0,0]=(0.9944935907)*x_ref[0]**o+(-0.365261996611)*x_ref[0]+(-0.105667939789)*x_ref[1]**o+(0.886935751977)*x_ref[1]+(0.263336157757)*x_ref[2]**o+(-0.880316523113)*x_ref[2]
        ref[0,2,0,1]=(-0.659779300787)*x_ref[0]**o+(-0.728590180152)*x_ref[0]+(0.732078508037)*x_ref[1]**o+(0.684963412098)*x_ref[1]+(0.822970700012)*x_ref[2]**o+(-0.680001879674)*x_ref[2]
        ref[0,2,1,0]=(-0.207855412893)*x_ref[0]**o+(0.438275426529)*x_ref[0]+(0.87046682066)*x_ref[1]**o+(-0.960322013848)*x_ref[1]+(-0.273728198625)*x_ref[2]**o+(-0.960137228004)*x_ref[2]
        ref[0,2,1,1]=(-0.189232680002)*x_ref[0]**o+(0.866417302002)*x_ref[0]+(-0.93403311248)*x_ref[1]**o+(-0.55773104878)*x_ref[1]+(0.94062895773)*x_ref[2]**o+(-0.758052194805)*x_ref[2]
        ref[0,2,2,0]=(0.100859944394)*x_ref[0]**o+(-0.20292135909)*x_ref[0]+(-0.713339941261)*x_ref[1]**o+(-0.615106454411)*x_ref[1]+(0.791245995152)*x_ref[2]**o+(-0.81051791594)*x_ref[2]
        ref[0,2,2,1]=(0.595743424517)*x_ref[0]**o+(-0.179075876404)*x_ref[0]+(-0.799005601715)*x_ref[1]**o+(-0.0554635073569)*x_ref[1]+(-0.850510701985)*x_ref[2]**o+(0.234516538169)*x_ref[2]
        ref[0,3,0,0]=(-0.361667738861)*x_ref[0]**o+(-0.551837498696)*x_ref[0]+(-0.834377202094)*x_ref[1]**o+(0.814168224265)*x_ref[1]+(-0.00479533268377)*x_ref[2]**o+(0.843007652951)*x_ref[2]
        ref[0,3,0,1]=(-0.448417439465)*x_ref[0]**o+(0.431496158388)*x_ref[0]+(-0.547416443433)*x_ref[1]**o+(-0.0746785704695)*x_ref[1]+(-0.646238182944)*x_ref[2]**o+(-0.318608714595)*x_ref[2]
        ref[0,3,1,0]=(-0.275647366007)*x_ref[0]**o+(-0.410594785047)*x_ref[0]+(-0.858282378196)*x_ref[1]**o+(-0.755586690837)*x_ref[1]+(-0.420481091654)*x_ref[2]**o+(-0.0996487371831)*x_ref[2]
        ref[0,3,1,1]=(-0.395961249009)*x_ref[0]**o+(0.0860896467863)*x_ref[0]+(0.482719525748)*x_ref[1]**o+(0.319368178133)*x_ref[1]+(0.516449153414)*x_ref[2]**o+(-0.246749862576)*x_ref[2]
        ref[0,3,2,0]=(-0.367322027122)*x_ref[0]**o+(0.687206412524)*x_ref[0]+(0.616265233972)*x_ref[1]**o+(-0.32660589112)*x_ref[1]+(-0.944607343096)*x_ref[2]**o+(0.123982537951)*x_ref[2]
        ref[0,3,2,1]=(0.700497166804)*x_ref[0]**o+(0.553961329865)*x_ref[0]+(-0.34020003754)*x_ref[1]**o+(0.295870992783)*x_ref[1]+(-0.220275880046)*x_ref[2]**o+(0.766121035825)*x_ref[2]
        ref[0,4,0,0]=(0.405469789242)*x_ref[0]**o+(-0.338347662873)*x_ref[0]+(0.0761796434233)*x_ref[1]**o+(-0.120599175033)*x_ref[1]+(0.107865585379)*x_ref[2]**o+(0.0711941007838)*x_ref[2]
        ref[0,4,0,1]=(0.614166154899)*x_ref[0]**o+(0.195780702963)*x_ref[0]+(0.718935985215)*x_ref[1]**o+(0.214577296792)*x_ref[1]+(-0.366985242107)*x_ref[2]**o+(-0.509584787633)*x_ref[2]
        ref[0,4,1,0]=(-0.638646370981)*x_ref[0]**o+(-0.769955899284)*x_ref[0]+(-0.887407540503)*x_ref[1]**o+(0.0901245429998)*x_ref[1]+(0.786085641321)*x_ref[2]**o+(0.6248068552)*x_ref[2]
        ref[0,4,1,1]=(-0.145426222043)*x_ref[0]**o+(-0.494114545016)*x_ref[0]+(0.422153721335)*x_ref[1]**o+(0.617022143901)*x_ref[1]+(-0.665972324723)*x_ref[2]**o+(-0.476724432799)*x_ref[2]
        ref[0,4,2,0]=(0.556105252092)*x_ref[0]**o+(-0.893268696809)*x_ref[0]+(0.146441889489)*x_ref[1]**o+(-0.170512113283)*x_ref[1]+(-0.869313262875)*x_ref[2]**o+(-0.976874109392)*x_ref[2]
        ref[0,4,2,1]=(0.858927206499)*x_ref[0]**o+(0.198998355146)*x_ref[0]+(0.278125296049)*x_ref[1]**o+(-0.0442518828093)*x_ref[1]+(0.274852724203)*x_ref[2]**o+(-0.167649193314)*x_ref[2]
        ref[1,0,0,0]=(-0.0429420289353)*x_ref[0]**o+(-0.15764693929)*x_ref[0]+(-0.859599917512)*x_ref[1]**o+(-0.904178643575)*x_ref[1]+(0.519560711493)*x_ref[2]**o+(0.258644717907)*x_ref[2]
        ref[1,0,0,1]=(-0.185321114264)*x_ref[0]**o+(-0.794493830004)*x_ref[0]+(0.136084421162)*x_ref[1]**o+(0.926912021419)*x_ref[1]+(0.346392120388)*x_ref[2]**o+(-0.183908371454)*x_ref[2]
        ref[1,0,1,0]=(-0.730675509877)*x_ref[0]**o+(0.865034802778)*x_ref[0]+(-0.506315431878)*x_ref[1]**o+(-0.94607761051)*x_ref[1]+(0.195194792801)*x_ref[2]**o+(-0.865129258879)*x_ref[2]
        ref[1,0,1,1]=(-0.0597026648107)*x_ref[0]**o+(-0.884790497)*x_ref[0]+(-0.0631278742098)*x_ref[1]**o+(-0.808364940737)*x_ref[1]+(-0.968952604837)*x_ref[2]**o+(0.142246702632)*x_ref[2]
        ref[1,0,2,0]=(0.648472894563)*x_ref[0]**o+(0.145861644073)*x_ref[0]+(0.501227520574)*x_ref[1]**o+(0.265886482306)*x_ref[1]+(-0.0472772173573)*x_ref[2]**o+(-0.364169655969)*x_ref[2]
        ref[1,0,2,1]=(0.847863600695)*x_ref[0]**o+(-0.288841685199)*x_ref[0]+(-0.500273502795)*x_ref[1]**o+(0.627379277592)*x_ref[1]+(0.326814024625)*x_ref[2]**o+(-0.514712825895)*x_ref[2]
        ref[1,1,0,0]=(-0.947845713712)*x_ref[0]**o+(-0.372115616305)*x_ref[0]+(-0.572858279998)*x_ref[1]**o+(0.226237627692)*x_ref[1]+(0.210201947182)*x_ref[2]**o+(0.0147616538594)*x_ref[2]
        ref[1,1,0,1]=(-0.39813630466)*x_ref[0]**o+(-0.658496395731)*x_ref[0]+(-0.363906465005)*x_ref[1]**o+(-0.816719661238)*x_ref[1]+(-0.478106835032)*x_ref[2]**o+(0.0819042309589)*x_ref[2]
        ref[1,1,1,0]=(-0.287160278593)*x_ref[0]**o+(-0.0281402640978)*x_ref[0]+(0.41559795378)*x_ref[1]**o+(0.158274458097)*x_ref[1]+(-0.798429481502)*x_ref[2]**o+(-0.278585334624)*x_ref[2]
        ref[1,1,1,1]=(0.809550480973)*x_ref[0]**o+(0.192026915854)*x_ref[0]+(-0.891168838599)*x_ref[1]**o+(0.0482621192882)*x_ref[1]+(-0.35339708159)*x_ref[2]**o+(-0.133161551553)*x_ref[2]
        ref[1,1,2,0]=(0.939582328872)*x_ref[0]**o+(0.561963790516)*x_ref[0]+(0.885264722949)*x_ref[1]**o+(-0.468042305266)*x_ref[1]+(0.232880002429)*x_ref[2]**o+(-0.772896005701)*x_ref[2]
        ref[1,1,2,1]=(-0.297333303209)*x_ref[0]**o+(-0.414050924207)*x_ref[0]+(-0.278903698873)*x_ref[1]**o+(0.345370620869)*x_ref[1]+(-0.0570151259027)*x_ref[2]**o+(0.786449641378)*x_ref[2]
        ref[1,2,0,0]=(0.644047138589)*x_ref[0]**o+(-0.480660438838)*x_ref[0]+(0.32526499462)*x_ref[1]**o+(-0.775988282127)*x_ref[1]+(0.108374143932)*x_ref[2]**o+(0.955463667181)*x_ref[2]
        ref[1,2,0,1]=(-0.468236790339)*x_ref[0]**o+(0.844197711724)*x_ref[0]+(0.808973429831)*x_ref[1]**o+(-0.647699550005)*x_ref[1]+(0.552555959813)*x_ref[2]**o+(-0.3403079515)*x_ref[2]
        ref[1,2,1,0]=(0.864686362308)*x_ref[0]**o+(0.248923646116)*x_ref[0]+(0.0287042863842)*x_ref[1]**o+(0.404974282153)*x_ref[1]+(0.113501438234)*x_ref[2]**o+(-0.619364894397)*x_ref[2]
        ref[1,2,1,1]=(-0.625146992499)*x_ref[0]**o+(-0.0480602452983)*x_ref[0]+(0.168910550914)*x_ref[1]**o+(-0.891584218123)*x_ref[1]+(0.0609357618263)*x_ref[2]**o+(0.0451921002161)*x_ref[2]
        ref[1,2,2,0]=(-0.478728628752)*x_ref[0]**o+(0.428376547798)*x_ref[0]+(0.817203376818)*x_ref[1]**o+(-0.856056266953)*x_ref[1]+(0.354960313382)*x_ref[2]**o+(-0.611468960461)*x_ref[2]
        ref[1,2,2,1]=(0.49500742113)*x_ref[0]**o+(0.902948130263)*x_ref[0]+(-0.810748543744)*x_ref[1]**o+(-0.696150924858)*x_ref[1]+(-0.748278516161)*x_ref[2]**o+(0.694815527058)*x_ref[2]
        ref[1,3,0,0]=(-0.361960150296)*x_ref[0]**o+(-0.845139780461)*x_ref[0]+(-0.0624499309665)*x_ref[1]**o+(-0.0515558058489)*x_ref[1]+(0.400690983017)*x_ref[2]**o+(-0.744608407974)*x_ref[2]
        ref[1,3,0,1]=(-0.452453617467)*x_ref[0]**o+(0.845396139593)*x_ref[0]+(0.292319469936)*x_ref[1]**o+(-0.833547518634)*x_ref[1]+(-0.511797700353)*x_ref[2]**o+(0.857891454773)*x_ref[2]
        ref[1,3,1,0]=(0.779504092912)*x_ref[0]**o+(0.500524773615)*x_ref[0]+(0.590542659382)*x_ref[1]**o+(0.479310630519)*x_ref[1]+(0.509960641956)*x_ref[2]**o+(-0.564980432959)*x_ref[2]
        ref[1,3,1,1]=(-0.145745189298)*x_ref[0]**o+(-0.256064576359)*x_ref[0]+(0.640890837126)*x_ref[1]**o+(-0.718802035152)*x_ref[1]+(-0.322216873583)*x_ref[2]**o+(-0.0246526526334)*x_ref[2]
        ref[1,3,2,0]=(-0.463434285098)*x_ref[0]**o+(0.116722448643)*x_ref[0]+(-0.690378381342)*x_ref[1]**o+(0.429197465885)*x_ref[1]+(-0.353019731384)*x_ref[2]**o+(0.57397712918)*x_ref[2]
        ref[1,3,2,1]=(0.169676853383)*x_ref[0]**o+(0.815225472623)*x_ref[0]+(-0.11125527552)*x_ref[1]**o+(0.391513831525)*x_ref[1]+(-0.587041361585)*x_ref[2]**o+(0.255194151616)*x_ref[2]
        ref[1,4,0,0]=(0.325351608628)*x_ref[0]**o+(-0.57751715453)*x_ref[0]+(0.145167511303)*x_ref[1]**o+(-0.969286030742)*x_ref[1]+(0.917370398478)*x_ref[2]**o+(0.691483691598)*x_ref[2]
        ref[1,4,0,1]=(0.0997998298271)*x_ref[0]**o+(-0.101248202247)*x_ref[0]+(-0.650649653848)*x_ref[1]**o+(-0.157887947288)*x_ref[1]+(0.517501747765)*x_ref[2]**o+(-0.0852921926298)*x_ref[2]
        ref[1,4,1,0]=(0.984595420407)*x_ref[0]**o+(0.499293499076)*x_ref[0]+(0.352193115874)*x_ref[1]**o+(-0.684522909647)*x_ref[1]+(-0.391538373605)*x_ref[2]**o+(0.404635595162)*x_ref[2]
        ref[1,4,1,1]=(-0.231310482173)*x_ref[0]**o+(-0.0385089836559)*x_ref[0]+(0.247671691888)*x_ref[1]**o+(0.883971955027)*x_ref[1]+(-0.0856624709734)*x_ref[2]**o+(0.942940875904)*x_ref[2]
        ref[1,4,2,0]=(-0.471467257434)*x_ref[0]**o+(0.835634240897)*x_ref[0]+(-0.594421294562)*x_ref[1]**o+(-0.187584387111)*x_ref[1]+(0.859867472765)*x_ref[2]**o+(0.930951154576)*x_ref[2]
        ref[1,4,2,1]=(0.504416975601)*x_ref[0]**o+(0.435692923635)*x_ref[0]+(-0.296945944313)*x_ref[1]**o+(0.815156218765)*x_ref[1]+(0.464464247442)*x_ref[2]**o+(-0.378809453041)*x_ref[2]
        ref[2,0,0,0]=(0.213087423107)*x_ref[0]**o+(-0.504681665801)*x_ref[0]+(0.796352190405)*x_ref[1]**o+(0.546752751136)*x_ref[1]+(0.57347644664)*x_ref[2]**o+(0.715856963195)*x_ref[2]
        ref[2,0,0,1]=(-0.28455266769)*x_ref[0]**o+(0.853162220766)*x_ref[0]+(0.164930982863)*x_ref[1]**o+(-0.0136996498765)*x_ref[1]+(0.293717915079)*x_ref[2]**o+(-0.216124158473)*x_ref[2]
        ref[2,0,1,0]=(-0.328115989441)*x_ref[0]**o+(-0.529441496488)*x_ref[0]+(-0.43579479677)*x_ref[1]**o+(-0.851333801194)*x_ref[1]+(-0.278406102972)*x_ref[2]**o+(0.947777471824)*x_ref[2]
        ref[2,0,1,1]=(-0.756438938789)*x_ref[0]**o+(-0.624332272735)*x_ref[0]+(-0.591638429565)*x_ref[1]**o+(-0.0608281771566)*x_ref[1]+(-0.852078784085)*x_ref[2]**o+(-0.526558964866)*x_ref[2]
        ref[2,0,2,0]=(-0.923892426969)*x_ref[0]**o+(0.788121899878)*x_ref[0]+(-0.446006331202)*x_ref[1]**o+(0.997210871574)*x_ref[1]+(0.356998606103)*x_ref[2]**o+(0.972643822823)*x_ref[2]
        ref[2,0,2,1]=(0.225183308571)*x_ref[0]**o+(-0.757505414777)*x_ref[0]+(-0.59573192074)*x_ref[1]**o+(0.57166154987)*x_ref[1]+(0.622343198809)*x_ref[2]**o+(0.60343970503)*x_ref[2]
        ref[2,1,0,0]=(0.51027702934)*x_ref[0]**o+(-0.711737679522)*x_ref[0]+(-0.211712579627)*x_ref[1]**o+(-0.495817605314)*x_ref[1]+(0.924425804272)*x_ref[2]**o+(0.753424835275)*x_ref[2]
        ref[2,1,0,1]=(0.476218252577)*x_ref[0]**o+(0.187579891557)*x_ref[0]+(0.727359583623)*x_ref[1]**o+(-0.389875559955)*x_ref[1]+(-0.352889326477)*x_ref[2]**o+(0.365143943637)*x_ref[2]
        ref[2,1,1,0]=(-0.815704312943)*x_ref[0]**o+(0.315217470245)*x_ref[0]+(-0.535234555065)*x_ref[1]**o+(-0.433408988501)*x_ref[1]+(-0.676694087389)*x_ref[2]**o+(0.384097177718)*x_ref[2]
        ref[2,1,1,1]=(-0.0368452411345)*x_ref[0]**o+(0.514540686444)*x_ref[0]+(-0.527100066965)*x_ref[1]**o+(0.0291015562912)*x_ref[1]+(-0.605646905004)*x_ref[2]**o+(0.931528599981)*x_ref[2]
        ref[2,1,2,0]=(-0.0994015945266)*x_ref[0]**o+(-0.762310347434)*x_ref[0]+(-0.845934950466)*x_ref[1]**o+(-0.235758967923)*x_ref[1]+(0.755531422698)*x_ref[2]**o+(0.108353867777)*x_ref[2]
        ref[2,1,2,1]=(-0.868779052889)*x_ref[0]**o+(0.783878941075)*x_ref[0]+(0.900089560473)*x_ref[1]**o+(-0.836779405877)*x_ref[1]+(-0.799571882449)*x_ref[2]**o+(0.999710029157)*x_ref[2]
        ref[2,2,0,0]=(0.541465277143)*x_ref[0]**o+(-0.840477797945)*x_ref[0]+(-0.362577721098)*x_ref[1]**o+(0.00680131806718)*x_ref[1]+(-0.338485239867)*x_ref[2]**o+(0.828213756751)*x_ref[2]
        ref[2,2,0,1]=(-0.927347423175)*x_ref[0]**o+(-0.936191125785)*x_ref[0]+(-0.411819804833)*x_ref[1]**o+(-0.0222067637269)*x_ref[1]+(-0.104411832586)*x_ref[2]**o+(-0.79937203081)*x_ref[2]
        ref[2,2,1,0]=(0.273051712848)*x_ref[0]**o+(-0.802468557957)*x_ref[0]+(-0.019392898867)*x_ref[1]**o+(-0.612803074784)*x_ref[1]+(-0.269705464223)*x_ref[2]**o+(-0.244405147706)*x_ref[2]
        ref[2,2,1,1]=(-0.08869649035)*x_ref[0]**o+(0.809761782447)*x_ref[0]+(0.761420080249)*x_ref[1]**o+(0.16590038557)*x_ref[1]+(-0.405657593445)*x_ref[2]**o+(0.123593603037)*x_ref[2]
        ref[2,2,2,0]=(-0.799691160987)*x_ref[0]**o+(0.834835690475)*x_ref[0]+(-0.0863569221737)*x_ref[1]**o+(0.988343215152)*x_ref[1]+(0.294423283474)*x_ref[2]**o+(0.325725138203)*x_ref[2]
        ref[2,2,2,1]=(0.648897496592)*x_ref[0]**o+(-0.474329387314)*x_ref[0]+(-0.904613697104)*x_ref[1]**o+(0.382201388286)*x_ref[1]+(-0.596889148339)*x_ref[2]**o+(0.705132157684)*x_ref[2]
        ref[2,3,0,0]=(0.906738144045)*x_ref[0]**o+(-0.365392026426)*x_ref[0]+(0.730557428733)*x_ref[1]**o+(-0.857864471189)*x_ref[1]+(-0.160930657186)*x_ref[2]**o+(0.885067383334)*x_ref[2]
        ref[2,3,0,1]=(0.992764702472)*x_ref[0]**o+(0.720688532948)*x_ref[0]+(0.318869772818)*x_ref[1]**o+(0.995508347248)*x_ref[1]+(0.108569456996)*x_ref[2]**o+(-0.990939292995)*x_ref[2]
        ref[2,3,1,0]=(0.690584992096)*x_ref[0]**o+(0.605885242246)*x_ref[0]+(0.72371794265)*x_ref[1]**o+(-0.632609325879)*x_ref[1]+(0.520273588734)*x_ref[2]**o+(0.885804143155)*x_ref[2]
        ref[2,3,1,1]=(0.485231542482)*x_ref[0]**o+(0.866920225418)*x_ref[0]+(-0.814964945972)*x_ref[1]**o+(0.544132513209)*x_ref[1]+(-0.155071376494)*x_ref[2]**o+(-0.941300746237)*x_ref[2]
        ref[2,3,2,0]=(0.748232238388)*x_ref[0]**o+(0.179397724419)*x_ref[0]+(-0.43881886248)*x_ref[1]**o+(-0.124941799624)*x_ref[1]+(0.14679335819)*x_ref[2]**o+(-0.758738846688)*x_ref[2]
        ref[2,3,2,1]=(-0.0951333655337)*x_ref[0]**o+(0.37270972476)*x_ref[0]+(-0.651977554115)*x_ref[1]**o+(-0.88918597476)*x_ref[1]+(-0.730858685576)*x_ref[2]**o+(0.685881924459)*x_ref[2]
        ref[2,4,0,0]=(-0.823487466802)*x_ref[0]**o+(-0.583630855643)*x_ref[0]+(-0.664814562747)*x_ref[1]**o+(0.860952605956)*x_ref[1]+(-0.249114322668)*x_ref[2]**o+(-0.968356584237)*x_ref[2]
        ref[2,4,0,1]=(0.44040146252)*x_ref[0]**o+(-0.246827565902)*x_ref[0]+(0.596300980733)*x_ref[1]**o+(-0.825818200414)*x_ref[1]+(0.0715512875967)*x_ref[2]**o+(0.996064679803)*x_ref[2]
        ref[2,4,1,0]=(-0.287806453021)*x_ref[0]**o+(0.704331542927)*x_ref[0]+(-0.346139754521)*x_ref[1]**o+(-0.766667037125)*x_ref[1]+(0.283874132105)*x_ref[2]**o+(-0.613145979204)*x_ref[2]
        ref[2,4,1,1]=(-0.116649875156)*x_ref[0]**o+(0.291380048239)*x_ref[0]+(0.689809221905)*x_ref[1]**o+(0.758935910592)*x_ref[1]+(0.505142787603)*x_ref[2]**o+(-0.113611480503)*x_ref[2]
        ref[2,4,2,0]=(-0.836908572692)*x_ref[0]**o+(-0.21892828725)*x_ref[0]+(-0.935780103868)*x_ref[1]**o+(0.106204643268)*x_ref[1]+(-0.894474423807)*x_ref[2]**o+(0.296554773217)*x_ref[2]
        ref[2,4,2,1]=(-0.89413495023)*x_ref[0]**o+(0.6080813058)*x_ref[0]+(0.087863889878)*x_ref[1]**o+(0.213373665612)*x_ref[1]+(0.17085345025)*x_ref[2]**o+(-0.967625138712)*x_ref[2]
        ref[3,0,0,0]=(0.327537677409)*x_ref[0]**o+(-0.299373059373)*x_ref[0]+(-0.43223648542)*x_ref[1]**o+(-0.499115781051)*x_ref[1]+(0.145464567449)*x_ref[2]**o+(0.580930584563)*x_ref[2]
        ref[3,0,0,1]=(-0.122622543351)*x_ref[0]**o+(-0.220547821081)*x_ref[0]+(0.892679827803)*x_ref[1]**o+(0.436584604381)*x_ref[1]+(-0.156093580288)*x_ref[2]**o+(0.464069069075)*x_ref[2]
        ref[3,0,1,0]=(0.590288663203)*x_ref[0]**o+(-0.212929445251)*x_ref[0]+(0.0899671639061)*x_ref[1]**o+(-0.87729854872)*x_ref[1]+(0.789506558716)*x_ref[2]**o+(-0.68867690907)*x_ref[2]
        ref[3,0,1,1]=(-0.913651849012)*x_ref[0]**o+(-0.430938918892)*x_ref[0]+(-0.351186012371)*x_ref[1]**o+(0.500306994748)*x_ref[1]+(0.145381845281)*x_ref[2]**o+(0.624993552333)*x_ref[2]
        ref[3,0,2,0]=(0.281393881265)*x_ref[0]**o+(-0.704727129826)*x_ref[0]+(-0.0102091836449)*x_ref[1]**o+(0.311857501353)*x_ref[1]+(0.300732947904)*x_ref[2]**o+(0.923973499548)*x_ref[2]
        ref[3,0,2,1]=(-0.584907526993)*x_ref[0]**o+(0.625732316822)*x_ref[0]+(-0.609691930408)*x_ref[1]**o+(0.056160226956)*x_ref[1]+(-0.766583561947)*x_ref[2]**o+(0.307692190194)*x_ref[2]
        ref[3,1,0,0]=(0.731231379445)*x_ref[0]**o+(-0.0485980527115)*x_ref[0]+(0.908932758852)*x_ref[1]**o+(0.188918035886)*x_ref[1]+(0.685837743688)*x_ref[2]**o+(-0.299987553989)*x_ref[2]
        ref[3,1,0,1]=(-0.784006561231)*x_ref[0]**o+(-0.809789430022)*x_ref[0]+(-0.846391792735)*x_ref[1]**o+(-0.211583304405)*x_ref[1]+(-0.00198590063687)*x_ref[2]**o+(-0.853253255224)*x_ref[2]
        ref[3,1,1,0]=(0.595567299132)*x_ref[0]**o+(-0.581734416878)*x_ref[0]+(0.88144481486)*x_ref[1]**o+(0.776876564509)*x_ref[1]+(-0.367477971386)*x_ref[2]**o+(0.0806320164453)*x_ref[2]
        ref[3,1,1,1]=(0.440836187611)*x_ref[0]**o+(-0.578925091757)*x_ref[0]+(0.923367189832)*x_ref[1]**o+(-0.993888173929)*x_ref[1]+(-0.645514380411)*x_ref[2]**o+(0.0267439044199)*x_ref[2]
        ref[3,1,2,0]=(-0.305182288788)*x_ref[0]**o+(-0.503372502553)*x_ref[0]+(0.740948657389)*x_ref[1]**o+(0.285661052275)*x_ref[1]+(-0.634510984218)*x_ref[2]**o+(-0.123979746456)*x_ref[2]
        ref[3,1,2,1]=(-0.897274928495)*x_ref[0]**o+(0.971356796164)*x_ref[0]+(0.940768290313)*x_ref[1]**o+(-0.552941985667)*x_ref[1]+(0.172857451423)*x_ref[2]**o+(0.844758237523)*x_ref[2]
        ref[3,2,0,0]=(-0.903542725333)*x_ref[0]**o+(0.936704774551)*x_ref[0]+(-0.398085875228)*x_ref[1]**o+(-0.304061573512)*x_ref[1]+(-0.13362037229)*x_ref[2]**o+(-0.405043350728)*x_ref[2]
        ref[3,2,0,1]=(-0.0208590751801)*x_ref[0]**o+(-0.354086381885)*x_ref[0]+(0.510884698744)*x_ref[1]**o+(-0.275504880518)*x_ref[1]+(0.459476496356)*x_ref[2]**o+(0.25358097976)*x_ref[2]
        ref[3,2,1,0]=(-0.0171081877107)*x_ref[0]**o+(-0.856081626112)*x_ref[0]+(-0.0698686571004)*x_ref[1]**o+(0.0954918361122)*x_ref[1]+(0.938494954069)*x_ref[2]**o+(-0.824928688248)*x_ref[2]
        ref[3,2,1,1]=(-0.277307915098)*x_ref[0]**o+(0.737855779658)*x_ref[0]+(-0.466761948855)*x_ref[1]**o+(-0.821498435809)*x_ref[1]+(-0.391009601442)*x_ref[2]**o+(0.643834542048)*x_ref[2]
        ref[3,2,2,0]=(0.379015007885)*x_ref[0]**o+(0.348646179575)*x_ref[0]+(0.733765133901)*x_ref[1]**o+(-0.722363571789)*x_ref[1]+(0.488531791931)*x_ref[2]**o+(0.279813923679)*x_ref[2]
        ref[3,2,2,1]=(0.516708382119)*x_ref[0]**o+(0.0426665103314)*x_ref[0]+(-0.168751315718)*x_ref[1]**o+(-0.124069236608)*x_ref[1]+(-0.778866001674)*x_ref[2]**o+(-0.362031001279)*x_ref[2]
        ref[3,3,0,0]=(0.995284803465)*x_ref[0]**o+(-0.442327755879)*x_ref[0]+(-0.13110270925)*x_ref[1]**o+(-0.172857543102)*x_ref[1]+(-0.910522664379)*x_ref[2]**o+(0.516931833163)*x_ref[2]
        ref[3,3,0,1]=(-0.889320295626)*x_ref[0]**o+(-0.728330379534)*x_ref[0]+(-0.584091700627)*x_ref[1]**o+(0.469914082337)*x_ref[1]+(0.18874127891)*x_ref[2]**o+(0.0846532039249)*x_ref[2]
        ref[3,3,1,0]=(0.911523535462)*x_ref[0]**o+(0.0924361220354)*x_ref[0]+(-0.0822504494228)*x_ref[1]**o+(0.00630249023591)*x_ref[1]+(-0.180497330322)*x_ref[2]**o+(0.817405577154)*x_ref[2]
        ref[3,3,1,1]=(0.681872747813)*x_ref[0]**o+(0.840680190206)*x_ref[0]+(-0.477660295206)*x_ref[1]**o+(0.0399086968056)*x_ref[1]+(0.440403204247)*x_ref[2]**o+(-0.220656197016)*x_ref[2]
        ref[3,3,2,0]=(-0.967524407336)*x_ref[0]**o+(0.896652274462)*x_ref[0]+(0.735269328519)*x_ref[1]**o+(-0.249017384614)*x_ref[1]+(-0.573222311563)*x_ref[2]**o+(-0.653800334954)*x_ref[2]
        ref[3,3,2,1]=(0.10041632804)*x_ref[0]**o+(0.167657178348)*x_ref[0]+(0.990621134594)*x_ref[1]**o+(-0.309063473348)*x_ref[1]+(-0.688206827575)*x_ref[2]**o+(0.0926806818663)*x_ref[2]
        ref[3,4,0,0]=(-0.128678339518)*x_ref[0]**o+(-0.974238160059)*x_ref[0]+(-0.762599464772)*x_ref[1]**o+(0.0233286358698)*x_ref[1]+(-0.622481697973)*x_ref[2]**o+(-0.080495303798)*x_ref[2]
        ref[3,4,0,1]=(-0.380318230058)*x_ref[0]**o+(0.568425682425)*x_ref[0]+(0.787369916262)*x_ref[1]**o+(0.296745245123)*x_ref[1]+(-0.917621002414)*x_ref[2]**o+(-0.696209404954)*x_ref[2]
        ref[3,4,1,0]=(-0.382961468898)*x_ref[0]**o+(0.672614453621)*x_ref[0]+(0.565371503874)*x_ref[1]**o+(-0.597331999912)*x_ref[1]+(-0.153637846704)*x_ref[2]**o+(-0.438159587906)*x_ref[2]
        ref[3,4,1,1]=(-0.541043007799)*x_ref[0]**o+(0.415600628159)*x_ref[0]+(-0.793716270785)*x_ref[1]**o+(0.780110667203)*x_ref[1]+(-0.101309486185)*x_ref[2]**o+(-0.138868352939)*x_ref[2]
        ref[3,4,2,0]=(0.679815917569)*x_ref[0]**o+(-0.800657537776)*x_ref[0]+(-0.711956069893)*x_ref[1]**o+(-0.575626146908)*x_ref[1]+(0.184247083249)*x_ref[2]**o+(-0.728118084396)*x_ref[2]
        ref[3,4,2,1]=(-0.608089432748)*x_ref[0]**o+(0.356378798828)*x_ref[0]+(-0.575957599599)*x_ref[1]**o+(0.599586825137)*x_ref[1]+(-0.160475743299)*x_ref[2]**o+(-0.070762356177)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.52738736047)*x[0]**o+(-0.513351179641)*x[0]+(0.466974564277)*x[1]**o+(0.247665236281)*x[1]
        ref=(0.52738736047)*x_ref[0]**o+(-0.513351179641)*x_ref[0]+(0.466974564277)*x_ref[1]**o+(0.247665236281)*x_ref[1]
      else:
        arg=(0.345612544612)*x[0]**o+(0.480207854215)*x[0]+(0.196672493786)*x[1]**o+(-0.891043131902)*x[1]+(0.106095952827)*x[2]**o+(0.436049931061)*x[2]
        ref=(0.345612544612)*x_ref[0]**o+(0.480207854215)*x_ref[0]+(0.196672493786)*x_ref[1]**o+(-0.891043131902)*x_ref[1]+(0.106095952827)*x_ref[2]**o+(0.436049931061)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.738262816349)*x[0]**o+(-0.650523560236)*x[0]+(-0.344136056546)*x[1]**o+(-0.0693168099812)*x[1]
        arg[1]=(-0.565224044511)*x[0]**o+(0.94461728731)*x[0]+(0.234800891741)*x[1]**o+(0.836341748852)*x[1]
        ref[0]=(-0.738262816349)*x_ref[0]**o+(-0.650523560236)*x_ref[0]+(-0.344136056546)*x_ref[1]**o+(-0.0693168099812)*x_ref[1]
        ref[1]=(-0.565224044511)*x_ref[0]**o+(0.94461728731)*x_ref[0]+(0.234800891741)*x_ref[1]**o+(0.836341748852)*x_ref[1]
      else:
        arg[0]=(-0.76925788678)*x[0]**o+(-0.391133765534)*x[0]+(0.89961543807)*x[1]**o+(-0.355090830059)*x[1]+(-0.663969382551)*x[2]**o+(0.0324773110608)*x[2]
        arg[1]=(-0.640271335706)*x[0]**o+(-0.0771899290795)*x[0]+(0.923847889924)*x[1]**o+(-0.732484895713)*x[1]+(-0.782800597476)*x[2]**o+(-0.402230990251)*x[2]
        ref[0]=(-0.76925788678)*x_ref[0]**o+(-0.391133765534)*x_ref[0]+(0.89961543807)*x_ref[1]**o+(-0.355090830059)*x_ref[1]+(-0.663969382551)*x_ref[2]**o+(0.0324773110608)*x_ref[2]
        ref[1]=(-0.640271335706)*x_ref[0]**o+(-0.0771899290795)*x_ref[0]+(0.923847889924)*x_ref[1]**o+(-0.732484895713)*x_ref[1]+(-0.782800597476)*x_ref[2]**o+(-0.402230990251)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.631349529615)*x[0]**o+(0.79216890406)*x[0]+(0.835421397222)*x[1]**o+(-0.209368192458)*x[1]
        arg[0,1]=(-0.166321945093)*x[0]**o+(0.701821475941)*x[0]+(-0.141804772049)*x[1]**o+(-0.860056917696)*x[1]
        arg[0,2]=(0.761659971379)*x[0]**o+(-0.39989119778)*x[0]+(-0.761140588562)*x[1]**o+(-0.595816184788)*x[1]
        arg[0,3]=(0.378518095375)*x[0]**o+(-0.880144120304)*x[0]+(0.814409656431)*x[1]**o+(0.0464425920173)*x[1]
        arg[0,4]=(-0.24431693596)*x[0]**o+(-0.840538735294)*x[0]+(0.0157782225732)*x[1]**o+(0.951311376091)*x[1]
        arg[1,0]=(0.0710042795971)*x[0]**o+(0.936106708209)*x[0]+(-0.580035969982)*x[1]**o+(0.740790378647)*x[1]
        arg[1,1]=(-0.807938035414)*x[0]**o+(0.8723236137)*x[0]+(-0.600615307712)*x[1]**o+(0.71483191921)*x[1]
        arg[1,2]=(0.454239081542)*x[0]**o+(0.966475397079)*x[0]+(-0.184649384525)*x[1]**o+(0.417686859721)*x[1]
        arg[1,3]=(-0.394999073966)*x[0]**o+(-0.0445825855756)*x[0]+(0.0389091203001)*x[1]**o+(-0.265569544304)*x[1]
        arg[1,4]=(0.625419856529)*x[0]**o+(-0.297955079048)*x[0]+(0.0092181343855)*x[1]**o+(-0.423768484981)*x[1]
        arg[2,0]=(0.836466109456)*x[0]**o+(-0.12515777587)*x[0]+(-0.0908815675065)*x[1]**o+(0.825332480799)*x[1]
        arg[2,1]=(-0.454612353892)*x[0]**o+(0.952942366642)*x[0]+(0.538638303761)*x[1]**o+(0.377957206795)*x[1]
        arg[2,2]=(-0.229405757447)*x[0]**o+(-0.125249012236)*x[0]+(0.194040765277)*x[1]**o+(-0.77353897054)*x[1]
        arg[2,3]=(-0.220514078279)*x[0]**o+(-0.937043129926)*x[0]+(0.661270617907)*x[1]**o+(0.342981486241)*x[1]
        arg[2,4]=(0.58858161977)*x[0]**o+(0.324813259161)*x[0]+(0.00439371077545)*x[1]**o+(0.517353738278)*x[1]
        arg[3,0]=(0.130001725331)*x[0]**o+(0.512570729718)*x[0]+(0.971292679928)*x[1]**o+(0.302530140373)*x[1]
        arg[3,1]=(0.999138842976)*x[0]**o+(0.325794749172)*x[0]+(0.94900982582)*x[1]**o+(0.56933825455)*x[1]
        arg[3,2]=(0.965201839092)*x[0]**o+(0.434187660328)*x[0]+(0.842238794411)*x[1]**o+(-0.754540335284)*x[1]
        arg[3,3]=(-0.539162954093)*x[0]**o+(-0.373779084363)*x[0]+(-0.426189858865)*x[1]**o+(-0.812351658629)*x[1]
        arg[3,4]=(0.871516777541)*x[0]**o+(-0.357393696308)*x[0]+(0.267358318776)*x[1]**o+(0.796701523616)*x[1]
        ref[0,0]=(-0.631349529615)*x_ref[0]**o+(0.79216890406)*x_ref[0]+(0.835421397222)*x_ref[1]**o+(-0.209368192458)*x_ref[1]
        ref[0,1]=(-0.166321945093)*x_ref[0]**o+(0.701821475941)*x_ref[0]+(-0.141804772049)*x_ref[1]**o+(-0.860056917696)*x_ref[1]
        ref[0,2]=(0.761659971379)*x_ref[0]**o+(-0.39989119778)*x_ref[0]+(-0.761140588562)*x_ref[1]**o+(-0.595816184788)*x_ref[1]
        ref[0,3]=(0.378518095375)*x_ref[0]**o+(-0.880144120304)*x_ref[0]+(0.814409656431)*x_ref[1]**o+(0.0464425920173)*x_ref[1]
        ref[0,4]=(-0.24431693596)*x_ref[0]**o+(-0.840538735294)*x_ref[0]+(0.0157782225732)*x_ref[1]**o+(0.951311376091)*x_ref[1]
        ref[1,0]=(0.0710042795971)*x_ref[0]**o+(0.936106708209)*x_ref[0]+(-0.580035969982)*x_ref[1]**o+(0.740790378647)*x_ref[1]
        ref[1,1]=(-0.807938035414)*x_ref[0]**o+(0.8723236137)*x_ref[0]+(-0.600615307712)*x_ref[1]**o+(0.71483191921)*x_ref[1]
        ref[1,2]=(0.454239081542)*x_ref[0]**o+(0.966475397079)*x_ref[0]+(-0.184649384525)*x_ref[1]**o+(0.417686859721)*x_ref[1]
        ref[1,3]=(-0.394999073966)*x_ref[0]**o+(-0.0445825855756)*x_ref[0]+(0.0389091203001)*x_ref[1]**o+(-0.265569544304)*x_ref[1]
        ref[1,4]=(0.625419856529)*x_ref[0]**o+(-0.297955079048)*x_ref[0]+(0.0092181343855)*x_ref[1]**o+(-0.423768484981)*x_ref[1]
        ref[2,0]=(0.836466109456)*x_ref[0]**o+(-0.12515777587)*x_ref[0]+(-0.0908815675065)*x_ref[1]**o+(0.825332480799)*x_ref[1]
        ref[2,1]=(-0.454612353892)*x_ref[0]**o+(0.952942366642)*x_ref[0]+(0.538638303761)*x_ref[1]**o+(0.377957206795)*x_ref[1]
        ref[2,2]=(-0.229405757447)*x_ref[0]**o+(-0.125249012236)*x_ref[0]+(0.194040765277)*x_ref[1]**o+(-0.77353897054)*x_ref[1]
        ref[2,3]=(-0.220514078279)*x_ref[0]**o+(-0.937043129926)*x_ref[0]+(0.661270617907)*x_ref[1]**o+(0.342981486241)*x_ref[1]
        ref[2,4]=(0.58858161977)*x_ref[0]**o+(0.324813259161)*x_ref[0]+(0.00439371077545)*x_ref[1]**o+(0.517353738278)*x_ref[1]
        ref[3,0]=(0.130001725331)*x_ref[0]**o+(0.512570729718)*x_ref[0]+(0.971292679928)*x_ref[1]**o+(0.302530140373)*x_ref[1]
        ref[3,1]=(0.999138842976)*x_ref[0]**o+(0.325794749172)*x_ref[0]+(0.94900982582)*x_ref[1]**o+(0.56933825455)*x_ref[1]
        ref[3,2]=(0.965201839092)*x_ref[0]**o+(0.434187660328)*x_ref[0]+(0.842238794411)*x_ref[1]**o+(-0.754540335284)*x_ref[1]
        ref[3,3]=(-0.539162954093)*x_ref[0]**o+(-0.373779084363)*x_ref[0]+(-0.426189858865)*x_ref[1]**o+(-0.812351658629)*x_ref[1]
        ref[3,4]=(0.871516777541)*x_ref[0]**o+(-0.357393696308)*x_ref[0]+(0.267358318776)*x_ref[1]**o+(0.796701523616)*x_ref[1]
      else:
        arg[0,0]=(-0.811142944724)*x[0]**o+(-0.65659407492)*x[0]+(-0.534808740191)*x[1]**o+(0.602252100451)*x[1]+(-0.158845636056)*x[2]**o+(-0.781826300399)*x[2]
        arg[0,1]=(-0.396870942248)*x[0]**o+(0.285698519063)*x[0]+(-0.962418369859)*x[1]**o+(-0.729381989807)*x[1]+(-0.0500038851534)*x[2]**o+(-0.567828852526)*x[2]
        arg[0,2]=(0.346399694372)*x[0]**o+(-0.170517233373)*x[0]+(-0.443275920724)*x[1]**o+(0.581464075295)*x[1]+(0.197230389253)*x[2]**o+(-0.613092248913)*x[2]
        arg[0,3]=(0.920174147113)*x[0]**o+(0.429847401539)*x[0]+(0.112888577662)*x[1]**o+(-0.0540034503714)*x[1]+(-0.0323244631044)*x[2]**o+(-0.293609544741)*x[2]
        arg[0,4]=(-0.0177585560206)*x[0]**o+(-0.806033755558)*x[0]+(-0.363921399918)*x[1]**o+(-0.415721267967)*x[1]+(0.832302676307)*x[2]**o+(-0.102510351741)*x[2]
        arg[1,0]=(-0.769760277763)*x[0]**o+(0.818187369213)*x[0]+(0.454400795864)*x[1]**o+(-0.220119708184)*x[1]+(-0.637117949075)*x[2]**o+(-0.661398404641)*x[2]
        arg[1,1]=(0.828424952723)*x[0]**o+(0.677022266764)*x[0]+(-0.958176510992)*x[1]**o+(-0.732014776269)*x[1]+(-0.495308153354)*x[2]**o+(-0.30698327016)*x[2]
        arg[1,2]=(-0.454106379068)*x[0]**o+(0.686920350455)*x[0]+(0.560344054206)*x[1]**o+(0.355495451775)*x[1]+(0.256969568039)*x[2]**o+(-0.737279585211)*x[2]
        arg[1,3]=(-0.976488579746)*x[0]**o+(-0.652812816758)*x[0]+(-0.996243459964)*x[1]**o+(0.631565668939)*x[1]+(0.996259991179)*x[2]**o+(-0.796610599165)*x[2]
        arg[1,4]=(0.676024716135)*x[0]**o+(-0.179835226483)*x[0]+(0.497541872254)*x[1]**o+(-0.678345104068)*x[1]+(0.945491022682)*x[2]**o+(-0.358403760669)*x[2]
        arg[2,0]=(0.166901676041)*x[0]**o+(0.833434173485)*x[0]+(0.622153751777)*x[1]**o+(0.767679765059)*x[1]+(-0.82271335395)*x[2]**o+(0.667651471482)*x[2]
        arg[2,1]=(-0.405015366041)*x[0]**o+(0.0548291583154)*x[0]+(0.894878845805)*x[1]**o+(-0.0955982667798)*x[1]+(0.468927799304)*x[2]**o+(0.547421481564)*x[2]
        arg[2,2]=(0.893319250823)*x[0]**o+(0.771073524295)*x[0]+(0.993876715991)*x[1]**o+(0.0832975372419)*x[1]+(-0.717061635157)*x[2]**o+(0.303760934227)*x[2]
        arg[2,3]=(0.247006018272)*x[0]**o+(0.329618310414)*x[0]+(-0.824176211089)*x[1]**o+(-0.678238702323)*x[1]+(0.218736302876)*x[2]**o+(0.8652063227)*x[2]
        arg[2,4]=(0.502380810223)*x[0]**o+(-0.94917091185)*x[0]+(-0.0392659831743)*x[1]**o+(-0.308291910503)*x[1]+(-0.498112428439)*x[2]**o+(0.217915791465)*x[2]
        arg[3,0]=(-0.018497051159)*x[0]**o+(-0.543005709743)*x[0]+(-0.433773552307)*x[1]**o+(-0.604850191015)*x[1]+(0.285970270117)*x[2]**o+(0.150938308283)*x[2]
        arg[3,1]=(0.185537796564)*x[0]**o+(0.663947686119)*x[0]+(-0.846597672873)*x[1]**o+(-0.798399069258)*x[1]+(0.0215106260597)*x[2]**o+(-0.388760513164)*x[2]
        arg[3,2]=(-0.378320218313)*x[0]**o+(-0.385079307834)*x[0]+(-0.392389046498)*x[1]**o+(0.323305501943)*x[1]+(0.496285813211)*x[2]**o+(-0.0269835921214)*x[2]
        arg[3,3]=(0.779877220393)*x[0]**o+(0.968536241033)*x[0]+(0.786166009582)*x[1]**o+(-0.974838905578)*x[1]+(0.4579382269)*x[2]**o+(0.178464213473)*x[2]
        arg[3,4]=(-0.149895758057)*x[0]**o+(0.3529889311)*x[0]+(0.223924981441)*x[1]**o+(0.10143440581)*x[1]+(0.162222746832)*x[2]**o+(0.421433640986)*x[2]
        ref[0,0]=(-0.811142944724)*x_ref[0]**o+(-0.65659407492)*x_ref[0]+(-0.534808740191)*x_ref[1]**o+(0.602252100451)*x_ref[1]+(-0.158845636056)*x_ref[2]**o+(-0.781826300399)*x_ref[2]
        ref[0,1]=(-0.396870942248)*x_ref[0]**o+(0.285698519063)*x_ref[0]+(-0.962418369859)*x_ref[1]**o+(-0.729381989807)*x_ref[1]+(-0.0500038851534)*x_ref[2]**o+(-0.567828852526)*x_ref[2]
        ref[0,2]=(0.346399694372)*x_ref[0]**o+(-0.170517233373)*x_ref[0]+(-0.443275920724)*x_ref[1]**o+(0.581464075295)*x_ref[1]+(0.197230389253)*x_ref[2]**o+(-0.613092248913)*x_ref[2]
        ref[0,3]=(0.920174147113)*x_ref[0]**o+(0.429847401539)*x_ref[0]+(0.112888577662)*x_ref[1]**o+(-0.0540034503714)*x_ref[1]+(-0.0323244631044)*x_ref[2]**o+(-0.293609544741)*x_ref[2]
        ref[0,4]=(-0.0177585560206)*x_ref[0]**o+(-0.806033755558)*x_ref[0]+(-0.363921399918)*x_ref[1]**o+(-0.415721267967)*x_ref[1]+(0.832302676307)*x_ref[2]**o+(-0.102510351741)*x_ref[2]
        ref[1,0]=(-0.769760277763)*x_ref[0]**o+(0.818187369213)*x_ref[0]+(0.454400795864)*x_ref[1]**o+(-0.220119708184)*x_ref[1]+(-0.637117949075)*x_ref[2]**o+(-0.661398404641)*x_ref[2]
        ref[1,1]=(0.828424952723)*x_ref[0]**o+(0.677022266764)*x_ref[0]+(-0.958176510992)*x_ref[1]**o+(-0.732014776269)*x_ref[1]+(-0.495308153354)*x_ref[2]**o+(-0.30698327016)*x_ref[2]
        ref[1,2]=(-0.454106379068)*x_ref[0]**o+(0.686920350455)*x_ref[0]+(0.560344054206)*x_ref[1]**o+(0.355495451775)*x_ref[1]+(0.256969568039)*x_ref[2]**o+(-0.737279585211)*x_ref[2]
        ref[1,3]=(-0.976488579746)*x_ref[0]**o+(-0.652812816758)*x_ref[0]+(-0.996243459964)*x_ref[1]**o+(0.631565668939)*x_ref[1]+(0.996259991179)*x_ref[2]**o+(-0.796610599165)*x_ref[2]
        ref[1,4]=(0.676024716135)*x_ref[0]**o+(-0.179835226483)*x_ref[0]+(0.497541872254)*x_ref[1]**o+(-0.678345104068)*x_ref[1]+(0.945491022682)*x_ref[2]**o+(-0.358403760669)*x_ref[2]
        ref[2,0]=(0.166901676041)*x_ref[0]**o+(0.833434173485)*x_ref[0]+(0.622153751777)*x_ref[1]**o+(0.767679765059)*x_ref[1]+(-0.82271335395)*x_ref[2]**o+(0.667651471482)*x_ref[2]
        ref[2,1]=(-0.405015366041)*x_ref[0]**o+(0.0548291583154)*x_ref[0]+(0.894878845805)*x_ref[1]**o+(-0.0955982667798)*x_ref[1]+(0.468927799304)*x_ref[2]**o+(0.547421481564)*x_ref[2]
        ref[2,2]=(0.893319250823)*x_ref[0]**o+(0.771073524295)*x_ref[0]+(0.993876715991)*x_ref[1]**o+(0.0832975372419)*x_ref[1]+(-0.717061635157)*x_ref[2]**o+(0.303760934227)*x_ref[2]
        ref[2,3]=(0.247006018272)*x_ref[0]**o+(0.329618310414)*x_ref[0]+(-0.824176211089)*x_ref[1]**o+(-0.678238702323)*x_ref[1]+(0.218736302876)*x_ref[2]**o+(0.8652063227)*x_ref[2]
        ref[2,4]=(0.502380810223)*x_ref[0]**o+(-0.94917091185)*x_ref[0]+(-0.0392659831743)*x_ref[1]**o+(-0.308291910503)*x_ref[1]+(-0.498112428439)*x_ref[2]**o+(0.217915791465)*x_ref[2]
        ref[3,0]=(-0.018497051159)*x_ref[0]**o+(-0.543005709743)*x_ref[0]+(-0.433773552307)*x_ref[1]**o+(-0.604850191015)*x_ref[1]+(0.285970270117)*x_ref[2]**o+(0.150938308283)*x_ref[2]
        ref[3,1]=(0.185537796564)*x_ref[0]**o+(0.663947686119)*x_ref[0]+(-0.846597672873)*x_ref[1]**o+(-0.798399069258)*x_ref[1]+(0.0215106260597)*x_ref[2]**o+(-0.388760513164)*x_ref[2]
        ref[3,2]=(-0.378320218313)*x_ref[0]**o+(-0.385079307834)*x_ref[0]+(-0.392389046498)*x_ref[1]**o+(0.323305501943)*x_ref[1]+(0.496285813211)*x_ref[2]**o+(-0.0269835921214)*x_ref[2]
        ref[3,3]=(0.779877220393)*x_ref[0]**o+(0.968536241033)*x_ref[0]+(0.786166009582)*x_ref[1]**o+(-0.974838905578)*x_ref[1]+(0.4579382269)*x_ref[2]**o+(0.178464213473)*x_ref[2]
        ref[3,4]=(-0.149895758057)*x_ref[0]**o+(0.3529889311)*x_ref[0]+(0.223924981441)*x_ref[1]**o+(0.10143440581)*x_ref[1]+(0.162222746832)*x_ref[2]**o+(0.421433640986)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.232971584272)*x[0]**o+(0.391351964909)*x[0]+(-0.832254448983)*x[1]**o+(0.880762034713)*x[1]
        arg[0,0,1]=(0.473097471101)*x[0]**o+(-0.660862772376)*x[0]+(0.72974853865)*x[1]**o+(-0.328183435687)*x[1]
        arg[0,1,0]=(-0.749483736482)*x[0]**o+(0.0872454020187)*x[0]+(0.6916024322)*x[1]**o+(0.82757019273)*x[1]
        arg[0,1,1]=(-0.325601366346)*x[0]**o+(0.918246327277)*x[0]+(0.855963838252)*x[1]**o+(0.448567464706)*x[1]
        arg[1,0,0]=(-0.6642143628)*x[0]**o+(-0.82649833447)*x[0]+(0.822778465465)*x[1]**o+(-0.926669226842)*x[1]
        arg[1,0,1]=(-0.644490899373)*x[0]**o+(0.0644054772987)*x[0]+(-0.914922609343)*x[1]**o+(0.898880463842)*x[1]
        arg[1,1,0]=(0.40963096867)*x[0]**o+(-0.347418076162)*x[0]+(0.857966073118)*x[1]**o+(-0.284521442085)*x[1]
        arg[1,1,1]=(-0.00998062171478)*x[0]**o+(-0.234122404174)*x[0]+(-0.149500951208)*x[1]**o+(0.680038328937)*x[1]
        arg[2,0,0]=(-0.963777985833)*x[0]**o+(0.710727608341)*x[0]+(-0.646898391724)*x[1]**o+(-0.464095155503)*x[1]
        arg[2,0,1]=(-0.617670394851)*x[0]**o+(0.592337321587)*x[0]+(-0.0881426217074)*x[1]**o+(-0.00431912882662)*x[1]
        arg[2,1,0]=(0.871279828121)*x[0]**o+(-0.858267190204)*x[0]+(0.440863109038)*x[1]**o+(-0.912521398099)*x[1]
        arg[2,1,1]=(0.17802472174)*x[0]**o+(0.948587960042)*x[0]+(-0.837945850976)*x[1]**o+(0.13308730872)*x[1]
        arg[3,0,0]=(0.0482409149433)*x[0]**o+(-0.198100088194)*x[0]+(0.504789334224)*x[1]**o+(0.296276808608)*x[1]
        arg[3,0,1]=(-0.30515363985)*x[0]**o+(-0.355293167256)*x[0]+(-0.577592997685)*x[1]**o+(-0.882408725943)*x[1]
        arg[3,1,0]=(0.76102937294)*x[0]**o+(-0.210169673144)*x[0]+(-0.181755140891)*x[1]**o+(0.865448484187)*x[1]
        arg[3,1,1]=(-0.894294091863)*x[0]**o+(-0.712245666103)*x[0]+(0.151657455329)*x[1]**o+(-0.54019436763)*x[1]
        arg[4,0,0]=(0.860662861318)*x[0]**o+(-0.36787473166)*x[0]+(0.240406823985)*x[1]**o+(-0.381822992473)*x[1]
        arg[4,0,1]=(-0.109197125112)*x[0]**o+(0.962595232385)*x[0]+(-0.323876506847)*x[1]**o+(0.178613363564)*x[1]
        arg[4,1,0]=(0.486980313424)*x[0]**o+(-0.611294566285)*x[0]+(-0.774813150316)*x[1]**o+(0.335411840405)*x[1]
        arg[4,1,1]=(-0.622163740867)*x[0]**o+(0.426305234663)*x[0]+(0.243557551101)*x[1]**o+(0.930601340004)*x[1]
        arg[5,0,0]=(-0.232795962027)*x[0]**o+(0.49010767268)*x[0]+(-0.503693305594)*x[1]**o+(-0.915931668186)*x[1]
        arg[5,0,1]=(-0.304614759138)*x[0]**o+(0.797445765321)*x[0]+(-0.00820825778888)*x[1]**o+(-0.196762463972)*x[1]
        arg[5,1,0]=(-0.516700339456)*x[0]**o+(-0.440953905414)*x[0]+(0.875266018165)*x[1]**o+(0.780895354543)*x[1]
        arg[5,1,1]=(0.219827931399)*x[0]**o+(-0.950514094131)*x[0]+(-0.120763970425)*x[1]**o+(0.0941571726596)*x[1]
        ref[0,0,0]=(-0.232971584272)*x_ref[0]**o+(0.391351964909)*x_ref[0]+(-0.832254448983)*x_ref[1]**o+(0.880762034713)*x_ref[1]
        ref[0,0,1]=(0.473097471101)*x_ref[0]**o+(-0.660862772376)*x_ref[0]+(0.72974853865)*x_ref[1]**o+(-0.328183435687)*x_ref[1]
        ref[0,1,0]=(-0.749483736482)*x_ref[0]**o+(0.0872454020187)*x_ref[0]+(0.6916024322)*x_ref[1]**o+(0.82757019273)*x_ref[1]
        ref[0,1,1]=(-0.325601366346)*x_ref[0]**o+(0.918246327277)*x_ref[0]+(0.855963838252)*x_ref[1]**o+(0.448567464706)*x_ref[1]
        ref[1,0,0]=(-0.6642143628)*x_ref[0]**o+(-0.82649833447)*x_ref[0]+(0.822778465465)*x_ref[1]**o+(-0.926669226842)*x_ref[1]
        ref[1,0,1]=(-0.644490899373)*x_ref[0]**o+(0.0644054772987)*x_ref[0]+(-0.914922609343)*x_ref[1]**o+(0.898880463842)*x_ref[1]
        ref[1,1,0]=(0.40963096867)*x_ref[0]**o+(-0.347418076162)*x_ref[0]+(0.857966073118)*x_ref[1]**o+(-0.284521442085)*x_ref[1]
        ref[1,1,1]=(-0.00998062171478)*x_ref[0]**o+(-0.234122404174)*x_ref[0]+(-0.149500951208)*x_ref[1]**o+(0.680038328937)*x_ref[1]
        ref[2,0,0]=(-0.963777985833)*x_ref[0]**o+(0.710727608341)*x_ref[0]+(-0.646898391724)*x_ref[1]**o+(-0.464095155503)*x_ref[1]
        ref[2,0,1]=(-0.617670394851)*x_ref[0]**o+(0.592337321587)*x_ref[0]+(-0.0881426217074)*x_ref[1]**o+(-0.00431912882662)*x_ref[1]
        ref[2,1,0]=(0.871279828121)*x_ref[0]**o+(-0.858267190204)*x_ref[0]+(0.440863109038)*x_ref[1]**o+(-0.912521398099)*x_ref[1]
        ref[2,1,1]=(0.17802472174)*x_ref[0]**o+(0.948587960042)*x_ref[0]+(-0.837945850976)*x_ref[1]**o+(0.13308730872)*x_ref[1]
        ref[3,0,0]=(0.0482409149433)*x_ref[0]**o+(-0.198100088194)*x_ref[0]+(0.504789334224)*x_ref[1]**o+(0.296276808608)*x_ref[1]
        ref[3,0,1]=(-0.30515363985)*x_ref[0]**o+(-0.355293167256)*x_ref[0]+(-0.577592997685)*x_ref[1]**o+(-0.882408725943)*x_ref[1]
        ref[3,1,0]=(0.76102937294)*x_ref[0]**o+(-0.210169673144)*x_ref[0]+(-0.181755140891)*x_ref[1]**o+(0.865448484187)*x_ref[1]
        ref[3,1,1]=(-0.894294091863)*x_ref[0]**o+(-0.712245666103)*x_ref[0]+(0.151657455329)*x_ref[1]**o+(-0.54019436763)*x_ref[1]
        ref[4,0,0]=(0.860662861318)*x_ref[0]**o+(-0.36787473166)*x_ref[0]+(0.240406823985)*x_ref[1]**o+(-0.381822992473)*x_ref[1]
        ref[4,0,1]=(-0.109197125112)*x_ref[0]**o+(0.962595232385)*x_ref[0]+(-0.323876506847)*x_ref[1]**o+(0.178613363564)*x_ref[1]
        ref[4,1,0]=(0.486980313424)*x_ref[0]**o+(-0.611294566285)*x_ref[0]+(-0.774813150316)*x_ref[1]**o+(0.335411840405)*x_ref[1]
        ref[4,1,1]=(-0.622163740867)*x_ref[0]**o+(0.426305234663)*x_ref[0]+(0.243557551101)*x_ref[1]**o+(0.930601340004)*x_ref[1]
        ref[5,0,0]=(-0.232795962027)*x_ref[0]**o+(0.49010767268)*x_ref[0]+(-0.503693305594)*x_ref[1]**o+(-0.915931668186)*x_ref[1]
        ref[5,0,1]=(-0.304614759138)*x_ref[0]**o+(0.797445765321)*x_ref[0]+(-0.00820825778888)*x_ref[1]**o+(-0.196762463972)*x_ref[1]
        ref[5,1,0]=(-0.516700339456)*x_ref[0]**o+(-0.440953905414)*x_ref[0]+(0.875266018165)*x_ref[1]**o+(0.780895354543)*x_ref[1]
        ref[5,1,1]=(0.219827931399)*x_ref[0]**o+(-0.950514094131)*x_ref[0]+(-0.120763970425)*x_ref[1]**o+(0.0941571726596)*x_ref[1]
      else:
        arg[0,0,0]=(0.772105181243)*x[0]**o+(-0.548341656718)*x[0]+(0.358327458717)*x[1]**o+(0.07838537297)*x[1]+(-0.221432264133)*x[2]**o+(-0.0831433214109)*x[2]
        arg[0,0,1]=(-0.920280406855)*x[0]**o+(0.288008584982)*x[0]+(0.273175197348)*x[1]**o+(0.71005891586)*x[1]+(-0.372322860924)*x[2]**o+(0.801820758328)*x[2]
        arg[0,1,0]=(0.610498050037)*x[0]**o+(-0.539393441621)*x[0]+(0.587033652815)*x[1]**o+(0.237738229824)*x[1]+(-0.79411765083)*x[2]**o+(-0.519989054973)*x[2]
        arg[0,1,1]=(0.5412774887)*x[0]**o+(-0.851503637634)*x[0]+(-0.71956793056)*x[1]**o+(0.166039550456)*x[1]+(-0.274290267921)*x[2]**o+(-0.909146653718)*x[2]
        arg[1,0,0]=(-0.00458109759775)*x[0]**o+(0.487263914082)*x[0]+(0.0171216930245)*x[1]**o+(0.691063106679)*x[1]+(-0.96202326744)*x[2]**o+(0.361419732405)*x[2]
        arg[1,0,1]=(-0.784311451677)*x[0]**o+(-0.862560770078)*x[0]+(-0.0702620499924)*x[1]**o+(0.183400389223)*x[1]+(0.364905596839)*x[2]**o+(0.874230430841)*x[2]
        arg[1,1,0]=(0.315952660065)*x[0]**o+(-0.0168150653709)*x[0]+(0.372807793658)*x[1]**o+(0.350959428104)*x[1]+(0.292074907124)*x[2]**o+(0.38176271099)*x[2]
        arg[1,1,1]=(-0.941288490366)*x[0]**o+(0.159864446187)*x[0]+(0.616516779388)*x[1]**o+(0.0940731931915)*x[1]+(-0.966416702566)*x[2]**o+(-0.716669694143)*x[2]
        arg[2,0,0]=(0.308619329706)*x[0]**o+(0.68203622136)*x[0]+(-0.732163723834)*x[1]**o+(-0.222214324226)*x[1]+(0.89540478356)*x[2]**o+(-0.0648727721753)*x[2]
        arg[2,0,1]=(-0.829002921476)*x[0]**o+(0.274738860807)*x[0]+(0.944302912883)*x[1]**o+(0.67434413206)*x[1]+(-0.861085567199)*x[2]**o+(0.994494977369)*x[2]
        arg[2,1,0]=(-0.871499634896)*x[0]**o+(-0.716060125296)*x[0]+(-0.655428500508)*x[1]**o+(-0.478135345667)*x[1]+(-0.858927907345)*x[2]**o+(0.718884168928)*x[2]
        arg[2,1,1]=(0.843612813106)*x[0]**o+(-0.221966655825)*x[0]+(-0.0770757351)*x[1]**o+(-0.678171902341)*x[1]+(0.814209490483)*x[2]**o+(0.409183122676)*x[2]
        arg[3,0,0]=(-0.226810706884)*x[0]**o+(-0.167181963701)*x[0]+(0.810478050432)*x[1]**o+(-0.335099187817)*x[1]+(0.210621363393)*x[2]**o+(0.429297116788)*x[2]
        arg[3,0,1]=(0.597117204298)*x[0]**o+(-0.000147371125322)*x[0]+(-0.557842700291)*x[1]**o+(0.329529225331)*x[1]+(-0.208265244071)*x[2]**o+(-0.977698555357)*x[2]
        arg[3,1,0]=(-0.163068070947)*x[0]**o+(0.640075503516)*x[0]+(0.425653990651)*x[1]**o+(0.650655294789)*x[1]+(-0.342488862403)*x[2]**o+(0.227438788286)*x[2]
        arg[3,1,1]=(-0.511981920097)*x[0]**o+(-0.18099315756)*x[0]+(-0.969881140182)*x[1]**o+(-0.418742703662)*x[1]+(-0.271422367765)*x[2]**o+(0.0812058381877)*x[2]
        arg[4,0,0]=(0.569514749505)*x[0]**o+(-0.711561480184)*x[0]+(-0.587159226794)*x[1]**o+(-0.240459786489)*x[1]+(-0.380983629494)*x[2]**o+(-0.550385464258)*x[2]
        arg[4,0,1]=(0.257375332759)*x[0]**o+(-0.477083226492)*x[0]+(0.0180233249756)*x[1]**o+(0.612470082967)*x[1]+(0.99566078545)*x[2]**o+(-0.551544637583)*x[2]
        arg[4,1,0]=(0.702787194057)*x[0]**o+(0.922262573534)*x[0]+(-0.465910368685)*x[1]**o+(0.748743004032)*x[1]+(0.642852066952)*x[2]**o+(0.24283802317)*x[2]
        arg[4,1,1]=(0.745762578199)*x[0]**o+(0.240631495429)*x[0]+(0.413077167171)*x[1]**o+(0.692304442104)*x[1]+(0.150058687546)*x[2]**o+(-0.133413399526)*x[2]
        arg[5,0,0]=(0.0244935245274)*x[0]**o+(-0.866653903557)*x[0]+(-0.682252235104)*x[1]**o+(-0.525092961647)*x[1]+(-0.87552740837)*x[2]**o+(-0.780785295196)*x[2]
        arg[5,0,1]=(0.403317147145)*x[0]**o+(-0.846638525869)*x[0]+(-0.798622632135)*x[1]**o+(-0.199698251819)*x[1]+(-0.852399643056)*x[2]**o+(-0.0618243339514)*x[2]
        arg[5,1,0]=(-0.94578445284)*x[0]**o+(-0.715570943968)*x[0]+(-0.193533970382)*x[1]**o+(0.0325400011289)*x[1]+(0.137922635356)*x[2]**o+(0.0192967461384)*x[2]
        arg[5,1,1]=(-0.706542583366)*x[0]**o+(0.412536423598)*x[0]+(-0.27322227526)*x[1]**o+(0.339116123776)*x[1]+(-0.856871339939)*x[2]**o+(0.140603890427)*x[2]
        ref[0,0,0]=(0.772105181243)*x_ref[0]**o+(-0.548341656718)*x_ref[0]+(0.358327458717)*x_ref[1]**o+(0.07838537297)*x_ref[1]+(-0.221432264133)*x_ref[2]**o+(-0.0831433214109)*x_ref[2]
        ref[0,0,1]=(-0.920280406855)*x_ref[0]**o+(0.288008584982)*x_ref[0]+(0.273175197348)*x_ref[1]**o+(0.71005891586)*x_ref[1]+(-0.372322860924)*x_ref[2]**o+(0.801820758328)*x_ref[2]
        ref[0,1,0]=(0.610498050037)*x_ref[0]**o+(-0.539393441621)*x_ref[0]+(0.587033652815)*x_ref[1]**o+(0.237738229824)*x_ref[1]+(-0.79411765083)*x_ref[2]**o+(-0.519989054973)*x_ref[2]
        ref[0,1,1]=(0.5412774887)*x_ref[0]**o+(-0.851503637634)*x_ref[0]+(-0.71956793056)*x_ref[1]**o+(0.166039550456)*x_ref[1]+(-0.274290267921)*x_ref[2]**o+(-0.909146653718)*x_ref[2]
        ref[1,0,0]=(-0.00458109759775)*x_ref[0]**o+(0.487263914082)*x_ref[0]+(0.0171216930245)*x_ref[1]**o+(0.691063106679)*x_ref[1]+(-0.96202326744)*x_ref[2]**o+(0.361419732405)*x_ref[2]
        ref[1,0,1]=(-0.784311451677)*x_ref[0]**o+(-0.862560770078)*x_ref[0]+(-0.0702620499924)*x_ref[1]**o+(0.183400389223)*x_ref[1]+(0.364905596839)*x_ref[2]**o+(0.874230430841)*x_ref[2]
        ref[1,1,0]=(0.315952660065)*x_ref[0]**o+(-0.0168150653709)*x_ref[0]+(0.372807793658)*x_ref[1]**o+(0.350959428104)*x_ref[1]+(0.292074907124)*x_ref[2]**o+(0.38176271099)*x_ref[2]
        ref[1,1,1]=(-0.941288490366)*x_ref[0]**o+(0.159864446187)*x_ref[0]+(0.616516779388)*x_ref[1]**o+(0.0940731931915)*x_ref[1]+(-0.966416702566)*x_ref[2]**o+(-0.716669694143)*x_ref[2]
        ref[2,0,0]=(0.308619329706)*x_ref[0]**o+(0.68203622136)*x_ref[0]+(-0.732163723834)*x_ref[1]**o+(-0.222214324226)*x_ref[1]+(0.89540478356)*x_ref[2]**o+(-0.0648727721753)*x_ref[2]
        ref[2,0,1]=(-0.829002921476)*x_ref[0]**o+(0.274738860807)*x_ref[0]+(0.944302912883)*x_ref[1]**o+(0.67434413206)*x_ref[1]+(-0.861085567199)*x_ref[2]**o+(0.994494977369)*x_ref[2]
        ref[2,1,0]=(-0.871499634896)*x_ref[0]**o+(-0.716060125296)*x_ref[0]+(-0.655428500508)*x_ref[1]**o+(-0.478135345667)*x_ref[1]+(-0.858927907345)*x_ref[2]**o+(0.718884168928)*x_ref[2]
        ref[2,1,1]=(0.843612813106)*x_ref[0]**o+(-0.221966655825)*x_ref[0]+(-0.0770757351)*x_ref[1]**o+(-0.678171902341)*x_ref[1]+(0.814209490483)*x_ref[2]**o+(0.409183122676)*x_ref[2]
        ref[3,0,0]=(-0.226810706884)*x_ref[0]**o+(-0.167181963701)*x_ref[0]+(0.810478050432)*x_ref[1]**o+(-0.335099187817)*x_ref[1]+(0.210621363393)*x_ref[2]**o+(0.429297116788)*x_ref[2]
        ref[3,0,1]=(0.597117204298)*x_ref[0]**o+(-0.000147371125322)*x_ref[0]+(-0.557842700291)*x_ref[1]**o+(0.329529225331)*x_ref[1]+(-0.208265244071)*x_ref[2]**o+(-0.977698555357)*x_ref[2]
        ref[3,1,0]=(-0.163068070947)*x_ref[0]**o+(0.640075503516)*x_ref[0]+(0.425653990651)*x_ref[1]**o+(0.650655294789)*x_ref[1]+(-0.342488862403)*x_ref[2]**o+(0.227438788286)*x_ref[2]
        ref[3,1,1]=(-0.511981920097)*x_ref[0]**o+(-0.18099315756)*x_ref[0]+(-0.969881140182)*x_ref[1]**o+(-0.418742703662)*x_ref[1]+(-0.271422367765)*x_ref[2]**o+(0.0812058381877)*x_ref[2]
        ref[4,0,0]=(0.569514749505)*x_ref[0]**o+(-0.711561480184)*x_ref[0]+(-0.587159226794)*x_ref[1]**o+(-0.240459786489)*x_ref[1]+(-0.380983629494)*x_ref[2]**o+(-0.550385464258)*x_ref[2]
        ref[4,0,1]=(0.257375332759)*x_ref[0]**o+(-0.477083226492)*x_ref[0]+(0.0180233249756)*x_ref[1]**o+(0.612470082967)*x_ref[1]+(0.99566078545)*x_ref[2]**o+(-0.551544637583)*x_ref[2]
        ref[4,1,0]=(0.702787194057)*x_ref[0]**o+(0.922262573534)*x_ref[0]+(-0.465910368685)*x_ref[1]**o+(0.748743004032)*x_ref[1]+(0.642852066952)*x_ref[2]**o+(0.24283802317)*x_ref[2]
        ref[4,1,1]=(0.745762578199)*x_ref[0]**o+(0.240631495429)*x_ref[0]+(0.413077167171)*x_ref[1]**o+(0.692304442104)*x_ref[1]+(0.150058687546)*x_ref[2]**o+(-0.133413399526)*x_ref[2]
        ref[5,0,0]=(0.0244935245274)*x_ref[0]**o+(-0.866653903557)*x_ref[0]+(-0.682252235104)*x_ref[1]**o+(-0.525092961647)*x_ref[1]+(-0.87552740837)*x_ref[2]**o+(-0.780785295196)*x_ref[2]
        ref[5,0,1]=(0.403317147145)*x_ref[0]**o+(-0.846638525869)*x_ref[0]+(-0.798622632135)*x_ref[1]**o+(-0.199698251819)*x_ref[1]+(-0.852399643056)*x_ref[2]**o+(-0.0618243339514)*x_ref[2]
        ref[5,1,0]=(-0.94578445284)*x_ref[0]**o+(-0.715570943968)*x_ref[0]+(-0.193533970382)*x_ref[1]**o+(0.0325400011289)*x_ref[1]+(0.137922635356)*x_ref[2]**o+(0.0192967461384)*x_ref[2]
        ref[5,1,1]=(-0.706542583366)*x_ref[0]**o+(0.412536423598)*x_ref[0]+(-0.27322227526)*x_ref[1]**o+(0.339116123776)*x_ref[1]+(-0.856871339939)*x_ref[2]**o+(0.140603890427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0577426785287)*x[0]**o+(0.469845292021)*x[0]+(-0.641706722819)*x[1]**o+(0.129652001842)*x[1]
        arg[0,0,0,1]=(-0.87686951076)*x[0]**o+(-0.981093040999)*x[0]+(0.654637319552)*x[1]**o+(-0.953104888811)*x[1]
        arg[0,0,1,0]=(-0.486386436415)*x[0]**o+(0.91083470956)*x[0]+(-0.392432937401)*x[1]**o+(0.494919250884)*x[1]
        arg[0,0,1,1]=(0.299928502076)*x[0]**o+(-0.0787848191157)*x[0]+(-0.541863592475)*x[1]**o+(0.435934816687)*x[1]
        arg[0,0,2,0]=(-0.0500083438365)*x[0]**o+(-0.882581886623)*x[0]+(-0.965347357063)*x[1]**o+(-0.270254041774)*x[1]
        arg[0,0,2,1]=(-0.18974861517)*x[0]**o+(0.341406736671)*x[0]+(0.847020774229)*x[1]**o+(0.650201215849)*x[1]
        arg[0,1,0,0]=(0.0823017084719)*x[0]**o+(0.964585359235)*x[0]+(0.357658660436)*x[1]**o+(0.182585271812)*x[1]
        arg[0,1,0,1]=(0.863411821905)*x[0]**o+(0.313374566426)*x[0]+(-0.273547644827)*x[1]**o+(0.429051854479)*x[1]
        arg[0,1,1,0]=(0.167608645068)*x[0]**o+(0.926908094967)*x[0]+(-0.853415240465)*x[1]**o+(-0.376634525241)*x[1]
        arg[0,1,1,1]=(-0.555491961591)*x[0]**o+(-0.64621137308)*x[0]+(-0.315877124991)*x[1]**o+(-0.962839286133)*x[1]
        arg[0,1,2,0]=(0.909012065667)*x[0]**o+(-0.656739022976)*x[0]+(-0.458377773547)*x[1]**o+(0.394223574009)*x[1]
        arg[0,1,2,1]=(-0.550560496714)*x[0]**o+(0.216941875936)*x[0]+(-0.634745794924)*x[1]**o+(0.887912543217)*x[1]
        arg[0,2,0,0]=(-0.961535901493)*x[0]**o+(0.980336384835)*x[0]+(-0.151755000128)*x[1]**o+(0.0402841925797)*x[1]
        arg[0,2,0,1]=(0.571527088727)*x[0]**o+(-0.475013184811)*x[0]+(0.80051818192)*x[1]**o+(-0.589499747694)*x[1]
        arg[0,2,1,0]=(0.545840648612)*x[0]**o+(0.888313209612)*x[0]+(-0.568938733446)*x[1]**o+(-0.289030374342)*x[1]
        arg[0,2,1,1]=(-0.0631638607979)*x[0]**o+(-0.173754439841)*x[0]+(0.143503554071)*x[1]**o+(-0.0850087835084)*x[1]
        arg[0,2,2,0]=(0.316299661412)*x[0]**o+(0.260890197379)*x[0]+(-0.803519867776)*x[1]**o+(0.36434810185)*x[1]
        arg[0,2,2,1]=(-0.656013716479)*x[0]**o+(0.810499942514)*x[0]+(0.783455480815)*x[1]**o+(-0.943085466999)*x[1]
        arg[0,3,0,0]=(0.128153439722)*x[0]**o+(-0.404740042302)*x[0]+(0.785717360548)*x[1]**o+(-0.710818900841)*x[1]
        arg[0,3,0,1]=(-0.097059274787)*x[0]**o+(0.578757823228)*x[0]+(0.186146155178)*x[1]**o+(0.274602030707)*x[1]
        arg[0,3,1,0]=(0.00522392753954)*x[0]**o+(0.015773606982)*x[0]+(-0.327110065344)*x[1]**o+(0.15571120542)*x[1]
        arg[0,3,1,1]=(-0.14009116305)*x[0]**o+(-0.659975337811)*x[0]+(-0.297582833489)*x[1]**o+(-0.600935015071)*x[1]
        arg[0,3,2,0]=(0.749388604357)*x[0]**o+(0.762968650778)*x[0]+(-0.623477518339)*x[1]**o+(-0.931762943135)*x[1]
        arg[0,3,2,1]=(0.748396100632)*x[0]**o+(-0.841314360401)*x[0]+(0.432649446216)*x[1]**o+(-0.971300843628)*x[1]
        arg[0,4,0,0]=(-0.217249090357)*x[0]**o+(-0.833633675601)*x[0]+(0.169984305627)*x[1]**o+(-0.831738272666)*x[1]
        arg[0,4,0,1]=(0.501571101669)*x[0]**o+(-0.223897334788)*x[0]+(0.24178181976)*x[1]**o+(0.307921789017)*x[1]
        arg[0,4,1,0]=(-0.162064610564)*x[0]**o+(0.217669674603)*x[0]+(0.995517866562)*x[1]**o+(0.903061734589)*x[1]
        arg[0,4,1,1]=(-0.837150436928)*x[0]**o+(0.465269468525)*x[0]+(0.522034182082)*x[1]**o+(-0.725469921625)*x[1]
        arg[0,4,2,0]=(-0.295739906169)*x[0]**o+(0.961959220618)*x[0]+(0.577799918942)*x[1]**o+(-0.313255108148)*x[1]
        arg[0,4,2,1]=(-0.455635927062)*x[0]**o+(0.667038701641)*x[0]+(-0.151183772961)*x[1]**o+(-0.550197828316)*x[1]
        arg[1,0,0,0]=(0.585510684472)*x[0]**o+(-0.64754219493)*x[0]+(0.781495111863)*x[1]**o+(-0.792246214695)*x[1]
        arg[1,0,0,1]=(0.586346471397)*x[0]**o+(0.291794158138)*x[0]+(-0.614723564286)*x[1]**o+(-0.151451723805)*x[1]
        arg[1,0,1,0]=(-0.472877220222)*x[0]**o+(-0.0118252391089)*x[0]+(-0.596943135354)*x[1]**o+(-0.413992986066)*x[1]
        arg[1,0,1,1]=(-0.715007718152)*x[0]**o+(0.032964784403)*x[0]+(-0.350680745946)*x[1]**o+(0.678841285054)*x[1]
        arg[1,0,2,0]=(0.673910318085)*x[0]**o+(-0.15303471314)*x[0]+(-0.0889279944344)*x[1]**o+(-0.0356474498002)*x[1]
        arg[1,0,2,1]=(-0.875691348779)*x[0]**o+(0.547300853671)*x[0]+(-0.362646169132)*x[1]**o+(-0.588521964671)*x[1]
        arg[1,1,0,0]=(-0.0988912551145)*x[0]**o+(-0.710944394432)*x[0]+(0.153875249337)*x[1]**o+(0.274427434926)*x[1]
        arg[1,1,0,1]=(-0.0854272185591)*x[0]**o+(-0.21840195791)*x[0]+(0.906982245924)*x[1]**o+(0.570990116211)*x[1]
        arg[1,1,1,0]=(-0.851259116331)*x[0]**o+(0.309829955487)*x[0]+(0.115725082696)*x[1]**o+(-0.91590292189)*x[1]
        arg[1,1,1,1]=(-0.637464462703)*x[0]**o+(-0.280971672162)*x[0]+(0.270985403603)*x[1]**o+(-0.534115311397)*x[1]
        arg[1,1,2,0]=(0.670179973397)*x[0]**o+(0.847483793904)*x[0]+(0.0681088634526)*x[1]**o+(0.412176359729)*x[1]
        arg[1,1,2,1]=(0.0378205172485)*x[0]**o+(0.974150495292)*x[0]+(-0.720821275299)*x[1]**o+(-0.552588296494)*x[1]
        arg[1,2,0,0]=(0.151367609975)*x[0]**o+(0.832981586349)*x[0]+(0.419488381843)*x[1]**o+(0.561052438283)*x[1]
        arg[1,2,0,1]=(0.267839010425)*x[0]**o+(0.47455974064)*x[0]+(0.498826652567)*x[1]**o+(-0.51906391205)*x[1]
        arg[1,2,1,0]=(-0.350105107527)*x[0]**o+(-0.489326689724)*x[0]+(-0.156970199109)*x[1]**o+(-0.175565230472)*x[1]
        arg[1,2,1,1]=(0.754379973459)*x[0]**o+(-0.70691373234)*x[0]+(-0.964497356137)*x[1]**o+(-0.694989442192)*x[1]
        arg[1,2,2,0]=(-0.491090361655)*x[0]**o+(0.154379441501)*x[0]+(0.403809695039)*x[1]**o+(-0.084183660554)*x[1]
        arg[1,2,2,1]=(-0.164538235223)*x[0]**o+(0.628058472821)*x[0]+(0.331956969519)*x[1]**o+(0.828755643202)*x[1]
        arg[1,3,0,0]=(0.103479416566)*x[0]**o+(0.380691196177)*x[0]+(-0.350375360146)*x[1]**o+(0.18528492785)*x[1]
        arg[1,3,0,1]=(-0.0744423760559)*x[0]**o+(0.912275380092)*x[0]+(0.867254399179)*x[1]**o+(0.410307184344)*x[1]
        arg[1,3,1,0]=(-0.480324052626)*x[0]**o+(-0.438544962062)*x[0]+(0.749575808465)*x[1]**o+(-0.0689669475685)*x[1]
        arg[1,3,1,1]=(0.723325499611)*x[0]**o+(-0.368668363075)*x[0]+(0.404302895836)*x[1]**o+(0.66913878408)*x[1]
        arg[1,3,2,0]=(-0.122189168877)*x[0]**o+(0.603362404944)*x[0]+(0.184163482697)*x[1]**o+(0.728136755361)*x[1]
        arg[1,3,2,1]=(-0.149430048386)*x[0]**o+(-0.478108072433)*x[0]+(0.342479834611)*x[1]**o+(0.596757508109)*x[1]
        arg[1,4,0,0]=(0.334694167334)*x[0]**o+(-0.815793032292)*x[0]+(0.394250112116)*x[1]**o+(0.269197166808)*x[1]
        arg[1,4,0,1]=(-0.19210954081)*x[0]**o+(0.416736181257)*x[0]+(0.146217029416)*x[1]**o+(0.906316717458)*x[1]
        arg[1,4,1,0]=(-0.708061018673)*x[0]**o+(0.299200707691)*x[0]+(0.665111947965)*x[1]**o+(-0.0340170473314)*x[1]
        arg[1,4,1,1]=(0.0353929744202)*x[0]**o+(0.0493194235067)*x[0]+(-0.12215424813)*x[1]**o+(-0.758677440837)*x[1]
        arg[1,4,2,0]=(0.769702957701)*x[0]**o+(0.901805654682)*x[0]+(0.260222241019)*x[1]**o+(-0.471997705392)*x[1]
        arg[1,4,2,1]=(0.204258068316)*x[0]**o+(-0.173085114573)*x[0]+(-0.0253164178046)*x[1]**o+(0.84254484259)*x[1]
        arg[2,0,0,0]=(-0.688170206644)*x[0]**o+(-0.718673702314)*x[0]+(0.834265379201)*x[1]**o+(-0.879925159193)*x[1]
        arg[2,0,0,1]=(-0.515401334666)*x[0]**o+(-0.810523180967)*x[0]+(0.725182741017)*x[1]**o+(-0.74319471007)*x[1]
        arg[2,0,1,0]=(-0.799698108602)*x[0]**o+(0.289222327549)*x[0]+(0.0589402409999)*x[1]**o+(-0.437669765839)*x[1]
        arg[2,0,1,1]=(-0.734933454418)*x[0]**o+(0.371391146956)*x[0]+(-0.17797293763)*x[1]**o+(-0.499412028554)*x[1]
        arg[2,0,2,0]=(0.217283159505)*x[0]**o+(-0.184871188378)*x[0]+(-0.764095670777)*x[1]**o+(0.103705507755)*x[1]
        arg[2,0,2,1]=(0.578071527416)*x[0]**o+(0.845365737459)*x[0]+(-0.173187973137)*x[1]**o+(-0.827819925403)*x[1]
        arg[2,1,0,0]=(-0.258505119864)*x[0]**o+(0.640873475074)*x[0]+(-0.993841847675)*x[1]**o+(0.347490565123)*x[1]
        arg[2,1,0,1]=(0.224893445547)*x[0]**o+(0.827208591218)*x[0]+(0.0391563892525)*x[1]**o+(-0.320939200596)*x[1]
        arg[2,1,1,0]=(-0.993044884903)*x[0]**o+(-0.395235414086)*x[0]+(-0.447460653655)*x[1]**o+(-0.114585622808)*x[1]
        arg[2,1,1,1]=(0.908711735866)*x[0]**o+(-0.911405008413)*x[0]+(-0.798601192961)*x[1]**o+(-0.432937813587)*x[1]
        arg[2,1,2,0]=(-0.270424920067)*x[0]**o+(-0.533648944143)*x[0]+(0.184307763968)*x[1]**o+(-0.802221601505)*x[1]
        arg[2,1,2,1]=(0.154432841861)*x[0]**o+(0.306458633794)*x[0]+(0.0343421658454)*x[1]**o+(-0.768559457414)*x[1]
        arg[2,2,0,0]=(-0.497828132445)*x[0]**o+(0.484920248315)*x[0]+(-0.188239341875)*x[1]**o+(0.100360622012)*x[1]
        arg[2,2,0,1]=(-0.0433518780446)*x[0]**o+(0.119470819101)*x[0]+(0.344025017428)*x[1]**o+(0.667394126898)*x[1]
        arg[2,2,1,0]=(0.512271795846)*x[0]**o+(0.567024910623)*x[0]+(-0.316191920076)*x[1]**o+(-0.202135136851)*x[1]
        arg[2,2,1,1]=(0.781678855797)*x[0]**o+(0.159208533718)*x[0]+(-0.930150837465)*x[1]**o+(-0.666903895504)*x[1]
        arg[2,2,2,0]=(0.292270965732)*x[0]**o+(-0.112877441048)*x[0]+(-0.643355582049)*x[1]**o+(-0.0482560996171)*x[1]
        arg[2,2,2,1]=(0.638688065872)*x[0]**o+(0.787719224138)*x[0]+(0.815312681592)*x[1]**o+(0.334735776221)*x[1]
        arg[2,3,0,0]=(0.409920519672)*x[0]**o+(-0.191886243537)*x[0]+(-0.348205383522)*x[1]**o+(-0.651065510647)*x[1]
        arg[2,3,0,1]=(0.824428500178)*x[0]**o+(-0.909533490456)*x[0]+(-0.455079541262)*x[1]**o+(-0.613994310127)*x[1]
        arg[2,3,1,0]=(0.225908995914)*x[0]**o+(0.258098741048)*x[0]+(-0.858905205327)*x[1]**o+(0.535845140874)*x[1]
        arg[2,3,1,1]=(-0.183600641837)*x[0]**o+(-0.0436432564818)*x[0]+(0.306524301109)*x[1]**o+(0.662513699164)*x[1]
        arg[2,3,2,0]=(-0.678973103555)*x[0]**o+(0.674110571991)*x[0]+(-0.129743880618)*x[1]**o+(-0.210759231178)*x[1]
        arg[2,3,2,1]=(-0.594989617102)*x[0]**o+(0.0557145839856)*x[0]+(-0.379516802094)*x[1]**o+(-0.496127170036)*x[1]
        arg[2,4,0,0]=(0.401088886175)*x[0]**o+(-0.208580896076)*x[0]+(-0.655582275521)*x[1]**o+(-0.522892532154)*x[1]
        arg[2,4,0,1]=(0.609111916382)*x[0]**o+(-0.864603847696)*x[0]+(-0.373482060442)*x[1]**o+(0.048777005392)*x[1]
        arg[2,4,1,0]=(-0.503935244799)*x[0]**o+(0.320517360825)*x[0]+(-0.658994963813)*x[1]**o+(0.00465823412972)*x[1]
        arg[2,4,1,1]=(-0.202767966863)*x[0]**o+(0.737239608437)*x[0]+(-0.633206522406)*x[1]**o+(-0.408531988781)*x[1]
        arg[2,4,2,0]=(0.660048575091)*x[0]**o+(-0.399137043056)*x[0]+(-0.0977181117895)*x[1]**o+(0.148903352776)*x[1]
        arg[2,4,2,1]=(-0.396768795712)*x[0]**o+(-0.550388681028)*x[0]+(0.0946256561554)*x[1]**o+(-0.0566941754267)*x[1]
        arg[3,0,0,0]=(0.0647180650687)*x[0]**o+(0.920788893299)*x[0]+(0.131306350033)*x[1]**o+(-0.840527983774)*x[1]
        arg[3,0,0,1]=(0.451700262969)*x[0]**o+(-0.16104676139)*x[0]+(0.114145999713)*x[1]**o+(-0.949265482931)*x[1]
        arg[3,0,1,0]=(-0.219444617083)*x[0]**o+(0.846470673388)*x[0]+(-0.341392225388)*x[1]**o+(-0.721912717627)*x[1]
        arg[3,0,1,1]=(-0.199502120929)*x[0]**o+(-0.00662410675554)*x[0]+(0.0529013037694)*x[1]**o+(-0.022278580228)*x[1]
        arg[3,0,2,0]=(0.803500732535)*x[0]**o+(0.754138333448)*x[0]+(0.822404687478)*x[1]**o+(0.900099584934)*x[1]
        arg[3,0,2,1]=(0.84674757393)*x[0]**o+(0.326823539079)*x[0]+(-0.824446392056)*x[1]**o+(-0.506583766552)*x[1]
        arg[3,1,0,0]=(0.816775849327)*x[0]**o+(0.957523430402)*x[0]+(-0.822535686534)*x[1]**o+(-0.583389123083)*x[1]
        arg[3,1,0,1]=(0.518120693337)*x[0]**o+(0.0946135115511)*x[0]+(0.577364260402)*x[1]**o+(-0.677435109716)*x[1]
        arg[3,1,1,0]=(-0.559041961891)*x[0]**o+(0.436791134026)*x[0]+(0.287609046653)*x[1]**o+(-0.484573207793)*x[1]
        arg[3,1,1,1]=(-0.0786041693053)*x[0]**o+(0.793526931222)*x[0]+(0.055780633982)*x[1]**o+(0.0447074319839)*x[1]
        arg[3,1,2,0]=(-0.190231018951)*x[0]**o+(-0.61318830169)*x[0]+(-0.157202158534)*x[1]**o+(-0.823043159126)*x[1]
        arg[3,1,2,1]=(-0.502637901786)*x[0]**o+(0.781083145697)*x[0]+(-0.663013942651)*x[1]**o+(-0.305726874525)*x[1]
        arg[3,2,0,0]=(0.566955929517)*x[0]**o+(-0.506981068575)*x[0]+(-0.831439289654)*x[1]**o+(-0.337491442765)*x[1]
        arg[3,2,0,1]=(-0.713648844422)*x[0]**o+(-0.00815764297609)*x[0]+(0.829706206618)*x[1]**o+(0.96471376965)*x[1]
        arg[3,2,1,0]=(-0.944705561715)*x[0]**o+(-0.87962603172)*x[0]+(-0.0525190846342)*x[1]**o+(-0.0457463400515)*x[1]
        arg[3,2,1,1]=(0.54922446571)*x[0]**o+(0.651923573438)*x[0]+(-0.00229398468376)*x[1]**o+(-0.00170863787749)*x[1]
        arg[3,2,2,0]=(0.622372079685)*x[0]**o+(-0.553578221845)*x[0]+(-0.737877450857)*x[1]**o+(0.459729854002)*x[1]
        arg[3,2,2,1]=(0.826155732508)*x[0]**o+(-0.000468898258991)*x[0]+(-0.255179342686)*x[1]**o+(-0.80957312045)*x[1]
        arg[3,3,0,0]=(-0.195212469839)*x[0]**o+(-0.0235216449547)*x[0]+(-0.175192639485)*x[1]**o+(-0.58937661213)*x[1]
        arg[3,3,0,1]=(-0.931538910404)*x[0]**o+(0.649282880007)*x[0]+(-0.237841482637)*x[1]**o+(-0.0220516220802)*x[1]
        arg[3,3,1,0]=(0.420911057166)*x[0]**o+(0.200126728184)*x[0]+(-0.829130861656)*x[1]**o+(-0.563462916661)*x[1]
        arg[3,3,1,1]=(-0.148199974575)*x[0]**o+(-0.248825389533)*x[0]+(-0.346424894584)*x[1]**o+(0.339758232978)*x[1]
        arg[3,3,2,0]=(0.249372496215)*x[0]**o+(-0.343626915746)*x[0]+(0.0961351845087)*x[1]**o+(0.612679249068)*x[1]
        arg[3,3,2,1]=(0.985380582259)*x[0]**o+(-0.603514318626)*x[0]+(-0.131105546364)*x[1]**o+(-0.0798158697238)*x[1]
        arg[3,4,0,0]=(0.478635365792)*x[0]**o+(-0.964424450881)*x[0]+(-0.469995535544)*x[1]**o+(-0.72871149515)*x[1]
        arg[3,4,0,1]=(-0.264802297415)*x[0]**o+(-0.926518282507)*x[0]+(0.803390182149)*x[1]**o+(0.387910762706)*x[1]
        arg[3,4,1,0]=(0.0452448579909)*x[0]**o+(-0.0482221515826)*x[0]+(-0.728020494982)*x[1]**o+(-0.288823633249)*x[1]
        arg[3,4,1,1]=(-0.640459595265)*x[0]**o+(0.759525453728)*x[0]+(-0.0280300707884)*x[1]**o+(0.079438187123)*x[1]
        arg[3,4,2,0]=(-0.14185171171)*x[0]**o+(-0.810989314052)*x[0]+(0.921902325773)*x[1]**o+(0.885920658784)*x[1]
        arg[3,4,2,1]=(-0.0101259901393)*x[0]**o+(-0.696541046542)*x[0]+(0.127932822496)*x[1]**o+(-0.822265796323)*x[1]
        ref[0,0,0,0]=(0.0577426785287)*x_ref[0]**o+(0.469845292021)*x_ref[0]+(-0.641706722819)*x_ref[1]**o+(0.129652001842)*x_ref[1]
        ref[0,0,0,1]=(-0.87686951076)*x_ref[0]**o+(-0.981093040999)*x_ref[0]+(0.654637319552)*x_ref[1]**o+(-0.953104888811)*x_ref[1]
        ref[0,0,1,0]=(-0.486386436415)*x_ref[0]**o+(0.91083470956)*x_ref[0]+(-0.392432937401)*x_ref[1]**o+(0.494919250884)*x_ref[1]
        ref[0,0,1,1]=(0.299928502076)*x_ref[0]**o+(-0.0787848191157)*x_ref[0]+(-0.541863592475)*x_ref[1]**o+(0.435934816687)*x_ref[1]
        ref[0,0,2,0]=(-0.0500083438365)*x_ref[0]**o+(-0.882581886623)*x_ref[0]+(-0.965347357063)*x_ref[1]**o+(-0.270254041774)*x_ref[1]
        ref[0,0,2,1]=(-0.18974861517)*x_ref[0]**o+(0.341406736671)*x_ref[0]+(0.847020774229)*x_ref[1]**o+(0.650201215849)*x_ref[1]
        ref[0,1,0,0]=(0.0823017084719)*x_ref[0]**o+(0.964585359235)*x_ref[0]+(0.357658660436)*x_ref[1]**o+(0.182585271812)*x_ref[1]
        ref[0,1,0,1]=(0.863411821905)*x_ref[0]**o+(0.313374566426)*x_ref[0]+(-0.273547644827)*x_ref[1]**o+(0.429051854479)*x_ref[1]
        ref[0,1,1,0]=(0.167608645068)*x_ref[0]**o+(0.926908094967)*x_ref[0]+(-0.853415240465)*x_ref[1]**o+(-0.376634525241)*x_ref[1]
        ref[0,1,1,1]=(-0.555491961591)*x_ref[0]**o+(-0.64621137308)*x_ref[0]+(-0.315877124991)*x_ref[1]**o+(-0.962839286133)*x_ref[1]
        ref[0,1,2,0]=(0.909012065667)*x_ref[0]**o+(-0.656739022976)*x_ref[0]+(-0.458377773547)*x_ref[1]**o+(0.394223574009)*x_ref[1]
        ref[0,1,2,1]=(-0.550560496714)*x_ref[0]**o+(0.216941875936)*x_ref[0]+(-0.634745794924)*x_ref[1]**o+(0.887912543217)*x_ref[1]
        ref[0,2,0,0]=(-0.961535901493)*x_ref[0]**o+(0.980336384835)*x_ref[0]+(-0.151755000128)*x_ref[1]**o+(0.0402841925797)*x_ref[1]
        ref[0,2,0,1]=(0.571527088727)*x_ref[0]**o+(-0.475013184811)*x_ref[0]+(0.80051818192)*x_ref[1]**o+(-0.589499747694)*x_ref[1]
        ref[0,2,1,0]=(0.545840648612)*x_ref[0]**o+(0.888313209612)*x_ref[0]+(-0.568938733446)*x_ref[1]**o+(-0.289030374342)*x_ref[1]
        ref[0,2,1,1]=(-0.0631638607979)*x_ref[0]**o+(-0.173754439841)*x_ref[0]+(0.143503554071)*x_ref[1]**o+(-0.0850087835084)*x_ref[1]
        ref[0,2,2,0]=(0.316299661412)*x_ref[0]**o+(0.260890197379)*x_ref[0]+(-0.803519867776)*x_ref[1]**o+(0.36434810185)*x_ref[1]
        ref[0,2,2,1]=(-0.656013716479)*x_ref[0]**o+(0.810499942514)*x_ref[0]+(0.783455480815)*x_ref[1]**o+(-0.943085466999)*x_ref[1]
        ref[0,3,0,0]=(0.128153439722)*x_ref[0]**o+(-0.404740042302)*x_ref[0]+(0.785717360548)*x_ref[1]**o+(-0.710818900841)*x_ref[1]
        ref[0,3,0,1]=(-0.097059274787)*x_ref[0]**o+(0.578757823228)*x_ref[0]+(0.186146155178)*x_ref[1]**o+(0.274602030707)*x_ref[1]
        ref[0,3,1,0]=(0.00522392753954)*x_ref[0]**o+(0.015773606982)*x_ref[0]+(-0.327110065344)*x_ref[1]**o+(0.15571120542)*x_ref[1]
        ref[0,3,1,1]=(-0.14009116305)*x_ref[0]**o+(-0.659975337811)*x_ref[0]+(-0.297582833489)*x_ref[1]**o+(-0.600935015071)*x_ref[1]
        ref[0,3,2,0]=(0.749388604357)*x_ref[0]**o+(0.762968650778)*x_ref[0]+(-0.623477518339)*x_ref[1]**o+(-0.931762943135)*x_ref[1]
        ref[0,3,2,1]=(0.748396100632)*x_ref[0]**o+(-0.841314360401)*x_ref[0]+(0.432649446216)*x_ref[1]**o+(-0.971300843628)*x_ref[1]
        ref[0,4,0,0]=(-0.217249090357)*x_ref[0]**o+(-0.833633675601)*x_ref[0]+(0.169984305627)*x_ref[1]**o+(-0.831738272666)*x_ref[1]
        ref[0,4,0,1]=(0.501571101669)*x_ref[0]**o+(-0.223897334788)*x_ref[0]+(0.24178181976)*x_ref[1]**o+(0.307921789017)*x_ref[1]
        ref[0,4,1,0]=(-0.162064610564)*x_ref[0]**o+(0.217669674603)*x_ref[0]+(0.995517866562)*x_ref[1]**o+(0.903061734589)*x_ref[1]
        ref[0,4,1,1]=(-0.837150436928)*x_ref[0]**o+(0.465269468525)*x_ref[0]+(0.522034182082)*x_ref[1]**o+(-0.725469921625)*x_ref[1]
        ref[0,4,2,0]=(-0.295739906169)*x_ref[0]**o+(0.961959220618)*x_ref[0]+(0.577799918942)*x_ref[1]**o+(-0.313255108148)*x_ref[1]
        ref[0,4,2,1]=(-0.455635927062)*x_ref[0]**o+(0.667038701641)*x_ref[0]+(-0.151183772961)*x_ref[1]**o+(-0.550197828316)*x_ref[1]
        ref[1,0,0,0]=(0.585510684472)*x_ref[0]**o+(-0.64754219493)*x_ref[0]+(0.781495111863)*x_ref[1]**o+(-0.792246214695)*x_ref[1]
        ref[1,0,0,1]=(0.586346471397)*x_ref[0]**o+(0.291794158138)*x_ref[0]+(-0.614723564286)*x_ref[1]**o+(-0.151451723805)*x_ref[1]
        ref[1,0,1,0]=(-0.472877220222)*x_ref[0]**o+(-0.0118252391089)*x_ref[0]+(-0.596943135354)*x_ref[1]**o+(-0.413992986066)*x_ref[1]
        ref[1,0,1,1]=(-0.715007718152)*x_ref[0]**o+(0.032964784403)*x_ref[0]+(-0.350680745946)*x_ref[1]**o+(0.678841285054)*x_ref[1]
        ref[1,0,2,0]=(0.673910318085)*x_ref[0]**o+(-0.15303471314)*x_ref[0]+(-0.0889279944344)*x_ref[1]**o+(-0.0356474498002)*x_ref[1]
        ref[1,0,2,1]=(-0.875691348779)*x_ref[0]**o+(0.547300853671)*x_ref[0]+(-0.362646169132)*x_ref[1]**o+(-0.588521964671)*x_ref[1]
        ref[1,1,0,0]=(-0.0988912551145)*x_ref[0]**o+(-0.710944394432)*x_ref[0]+(0.153875249337)*x_ref[1]**o+(0.274427434926)*x_ref[1]
        ref[1,1,0,1]=(-0.0854272185591)*x_ref[0]**o+(-0.21840195791)*x_ref[0]+(0.906982245924)*x_ref[1]**o+(0.570990116211)*x_ref[1]
        ref[1,1,1,0]=(-0.851259116331)*x_ref[0]**o+(0.309829955487)*x_ref[0]+(0.115725082696)*x_ref[1]**o+(-0.91590292189)*x_ref[1]
        ref[1,1,1,1]=(-0.637464462703)*x_ref[0]**o+(-0.280971672162)*x_ref[0]+(0.270985403603)*x_ref[1]**o+(-0.534115311397)*x_ref[1]
        ref[1,1,2,0]=(0.670179973397)*x_ref[0]**o+(0.847483793904)*x_ref[0]+(0.0681088634526)*x_ref[1]**o+(0.412176359729)*x_ref[1]
        ref[1,1,2,1]=(0.0378205172485)*x_ref[0]**o+(0.974150495292)*x_ref[0]+(-0.720821275299)*x_ref[1]**o+(-0.552588296494)*x_ref[1]
        ref[1,2,0,0]=(0.151367609975)*x_ref[0]**o+(0.832981586349)*x_ref[0]+(0.419488381843)*x_ref[1]**o+(0.561052438283)*x_ref[1]
        ref[1,2,0,1]=(0.267839010425)*x_ref[0]**o+(0.47455974064)*x_ref[0]+(0.498826652567)*x_ref[1]**o+(-0.51906391205)*x_ref[1]
        ref[1,2,1,0]=(-0.350105107527)*x_ref[0]**o+(-0.489326689724)*x_ref[0]+(-0.156970199109)*x_ref[1]**o+(-0.175565230472)*x_ref[1]
        ref[1,2,1,1]=(0.754379973459)*x_ref[0]**o+(-0.70691373234)*x_ref[0]+(-0.964497356137)*x_ref[1]**o+(-0.694989442192)*x_ref[1]
        ref[1,2,2,0]=(-0.491090361655)*x_ref[0]**o+(0.154379441501)*x_ref[0]+(0.403809695039)*x_ref[1]**o+(-0.084183660554)*x_ref[1]
        ref[1,2,2,1]=(-0.164538235223)*x_ref[0]**o+(0.628058472821)*x_ref[0]+(0.331956969519)*x_ref[1]**o+(0.828755643202)*x_ref[1]
        ref[1,3,0,0]=(0.103479416566)*x_ref[0]**o+(0.380691196177)*x_ref[0]+(-0.350375360146)*x_ref[1]**o+(0.18528492785)*x_ref[1]
        ref[1,3,0,1]=(-0.0744423760559)*x_ref[0]**o+(0.912275380092)*x_ref[0]+(0.867254399179)*x_ref[1]**o+(0.410307184344)*x_ref[1]
        ref[1,3,1,0]=(-0.480324052626)*x_ref[0]**o+(-0.438544962062)*x_ref[0]+(0.749575808465)*x_ref[1]**o+(-0.0689669475685)*x_ref[1]
        ref[1,3,1,1]=(0.723325499611)*x_ref[0]**o+(-0.368668363075)*x_ref[0]+(0.404302895836)*x_ref[1]**o+(0.66913878408)*x_ref[1]
        ref[1,3,2,0]=(-0.122189168877)*x_ref[0]**o+(0.603362404944)*x_ref[0]+(0.184163482697)*x_ref[1]**o+(0.728136755361)*x_ref[1]
        ref[1,3,2,1]=(-0.149430048386)*x_ref[0]**o+(-0.478108072433)*x_ref[0]+(0.342479834611)*x_ref[1]**o+(0.596757508109)*x_ref[1]
        ref[1,4,0,0]=(0.334694167334)*x_ref[0]**o+(-0.815793032292)*x_ref[0]+(0.394250112116)*x_ref[1]**o+(0.269197166808)*x_ref[1]
        ref[1,4,0,1]=(-0.19210954081)*x_ref[0]**o+(0.416736181257)*x_ref[0]+(0.146217029416)*x_ref[1]**o+(0.906316717458)*x_ref[1]
        ref[1,4,1,0]=(-0.708061018673)*x_ref[0]**o+(0.299200707691)*x_ref[0]+(0.665111947965)*x_ref[1]**o+(-0.0340170473314)*x_ref[1]
        ref[1,4,1,1]=(0.0353929744202)*x_ref[0]**o+(0.0493194235067)*x_ref[0]+(-0.12215424813)*x_ref[1]**o+(-0.758677440837)*x_ref[1]
        ref[1,4,2,0]=(0.769702957701)*x_ref[0]**o+(0.901805654682)*x_ref[0]+(0.260222241019)*x_ref[1]**o+(-0.471997705392)*x_ref[1]
        ref[1,4,2,1]=(0.204258068316)*x_ref[0]**o+(-0.173085114573)*x_ref[0]+(-0.0253164178046)*x_ref[1]**o+(0.84254484259)*x_ref[1]
        ref[2,0,0,0]=(-0.688170206644)*x_ref[0]**o+(-0.718673702314)*x_ref[0]+(0.834265379201)*x_ref[1]**o+(-0.879925159193)*x_ref[1]
        ref[2,0,0,1]=(-0.515401334666)*x_ref[0]**o+(-0.810523180967)*x_ref[0]+(0.725182741017)*x_ref[1]**o+(-0.74319471007)*x_ref[1]
        ref[2,0,1,0]=(-0.799698108602)*x_ref[0]**o+(0.289222327549)*x_ref[0]+(0.0589402409999)*x_ref[1]**o+(-0.437669765839)*x_ref[1]
        ref[2,0,1,1]=(-0.734933454418)*x_ref[0]**o+(0.371391146956)*x_ref[0]+(-0.17797293763)*x_ref[1]**o+(-0.499412028554)*x_ref[1]
        ref[2,0,2,0]=(0.217283159505)*x_ref[0]**o+(-0.184871188378)*x_ref[0]+(-0.764095670777)*x_ref[1]**o+(0.103705507755)*x_ref[1]
        ref[2,0,2,1]=(0.578071527416)*x_ref[0]**o+(0.845365737459)*x_ref[0]+(-0.173187973137)*x_ref[1]**o+(-0.827819925403)*x_ref[1]
        ref[2,1,0,0]=(-0.258505119864)*x_ref[0]**o+(0.640873475074)*x_ref[0]+(-0.993841847675)*x_ref[1]**o+(0.347490565123)*x_ref[1]
        ref[2,1,0,1]=(0.224893445547)*x_ref[0]**o+(0.827208591218)*x_ref[0]+(0.0391563892525)*x_ref[1]**o+(-0.320939200596)*x_ref[1]
        ref[2,1,1,0]=(-0.993044884903)*x_ref[0]**o+(-0.395235414086)*x_ref[0]+(-0.447460653655)*x_ref[1]**o+(-0.114585622808)*x_ref[1]
        ref[2,1,1,1]=(0.908711735866)*x_ref[0]**o+(-0.911405008413)*x_ref[0]+(-0.798601192961)*x_ref[1]**o+(-0.432937813587)*x_ref[1]
        ref[2,1,2,0]=(-0.270424920067)*x_ref[0]**o+(-0.533648944143)*x_ref[0]+(0.184307763968)*x_ref[1]**o+(-0.802221601505)*x_ref[1]
        ref[2,1,2,1]=(0.154432841861)*x_ref[0]**o+(0.306458633794)*x_ref[0]+(0.0343421658454)*x_ref[1]**o+(-0.768559457414)*x_ref[1]
        ref[2,2,0,0]=(-0.497828132445)*x_ref[0]**o+(0.484920248315)*x_ref[0]+(-0.188239341875)*x_ref[1]**o+(0.100360622012)*x_ref[1]
        ref[2,2,0,1]=(-0.0433518780446)*x_ref[0]**o+(0.119470819101)*x_ref[0]+(0.344025017428)*x_ref[1]**o+(0.667394126898)*x_ref[1]
        ref[2,2,1,0]=(0.512271795846)*x_ref[0]**o+(0.567024910623)*x_ref[0]+(-0.316191920076)*x_ref[1]**o+(-0.202135136851)*x_ref[1]
        ref[2,2,1,1]=(0.781678855797)*x_ref[0]**o+(0.159208533718)*x_ref[0]+(-0.930150837465)*x_ref[1]**o+(-0.666903895504)*x_ref[1]
        ref[2,2,2,0]=(0.292270965732)*x_ref[0]**o+(-0.112877441048)*x_ref[0]+(-0.643355582049)*x_ref[1]**o+(-0.0482560996171)*x_ref[1]
        ref[2,2,2,1]=(0.638688065872)*x_ref[0]**o+(0.787719224138)*x_ref[0]+(0.815312681592)*x_ref[1]**o+(0.334735776221)*x_ref[1]
        ref[2,3,0,0]=(0.409920519672)*x_ref[0]**o+(-0.191886243537)*x_ref[0]+(-0.348205383522)*x_ref[1]**o+(-0.651065510647)*x_ref[1]
        ref[2,3,0,1]=(0.824428500178)*x_ref[0]**o+(-0.909533490456)*x_ref[0]+(-0.455079541262)*x_ref[1]**o+(-0.613994310127)*x_ref[1]
        ref[2,3,1,0]=(0.225908995914)*x_ref[0]**o+(0.258098741048)*x_ref[0]+(-0.858905205327)*x_ref[1]**o+(0.535845140874)*x_ref[1]
        ref[2,3,1,1]=(-0.183600641837)*x_ref[0]**o+(-0.0436432564818)*x_ref[0]+(0.306524301109)*x_ref[1]**o+(0.662513699164)*x_ref[1]
        ref[2,3,2,0]=(-0.678973103555)*x_ref[0]**o+(0.674110571991)*x_ref[0]+(-0.129743880618)*x_ref[1]**o+(-0.210759231178)*x_ref[1]
        ref[2,3,2,1]=(-0.594989617102)*x_ref[0]**o+(0.0557145839856)*x_ref[0]+(-0.379516802094)*x_ref[1]**o+(-0.496127170036)*x_ref[1]
        ref[2,4,0,0]=(0.401088886175)*x_ref[0]**o+(-0.208580896076)*x_ref[0]+(-0.655582275521)*x_ref[1]**o+(-0.522892532154)*x_ref[1]
        ref[2,4,0,1]=(0.609111916382)*x_ref[0]**o+(-0.864603847696)*x_ref[0]+(-0.373482060442)*x_ref[1]**o+(0.048777005392)*x_ref[1]
        ref[2,4,1,0]=(-0.503935244799)*x_ref[0]**o+(0.320517360825)*x_ref[0]+(-0.658994963813)*x_ref[1]**o+(0.00465823412972)*x_ref[1]
        ref[2,4,1,1]=(-0.202767966863)*x_ref[0]**o+(0.737239608437)*x_ref[0]+(-0.633206522406)*x_ref[1]**o+(-0.408531988781)*x_ref[1]
        ref[2,4,2,0]=(0.660048575091)*x_ref[0]**o+(-0.399137043056)*x_ref[0]+(-0.0977181117895)*x_ref[1]**o+(0.148903352776)*x_ref[1]
        ref[2,4,2,1]=(-0.396768795712)*x_ref[0]**o+(-0.550388681028)*x_ref[0]+(0.0946256561554)*x_ref[1]**o+(-0.0566941754267)*x_ref[1]
        ref[3,0,0,0]=(0.0647180650687)*x_ref[0]**o+(0.920788893299)*x_ref[0]+(0.131306350033)*x_ref[1]**o+(-0.840527983774)*x_ref[1]
        ref[3,0,0,1]=(0.451700262969)*x_ref[0]**o+(-0.16104676139)*x_ref[0]+(0.114145999713)*x_ref[1]**o+(-0.949265482931)*x_ref[1]
        ref[3,0,1,0]=(-0.219444617083)*x_ref[0]**o+(0.846470673388)*x_ref[0]+(-0.341392225388)*x_ref[1]**o+(-0.721912717627)*x_ref[1]
        ref[3,0,1,1]=(-0.199502120929)*x_ref[0]**o+(-0.00662410675554)*x_ref[0]+(0.0529013037694)*x_ref[1]**o+(-0.022278580228)*x_ref[1]
        ref[3,0,2,0]=(0.803500732535)*x_ref[0]**o+(0.754138333448)*x_ref[0]+(0.822404687478)*x_ref[1]**o+(0.900099584934)*x_ref[1]
        ref[3,0,2,1]=(0.84674757393)*x_ref[0]**o+(0.326823539079)*x_ref[0]+(-0.824446392056)*x_ref[1]**o+(-0.506583766552)*x_ref[1]
        ref[3,1,0,0]=(0.816775849327)*x_ref[0]**o+(0.957523430402)*x_ref[0]+(-0.822535686534)*x_ref[1]**o+(-0.583389123083)*x_ref[1]
        ref[3,1,0,1]=(0.518120693337)*x_ref[0]**o+(0.0946135115511)*x_ref[0]+(0.577364260402)*x_ref[1]**o+(-0.677435109716)*x_ref[1]
        ref[3,1,1,0]=(-0.559041961891)*x_ref[0]**o+(0.436791134026)*x_ref[0]+(0.287609046653)*x_ref[1]**o+(-0.484573207793)*x_ref[1]
        ref[3,1,1,1]=(-0.0786041693053)*x_ref[0]**o+(0.793526931222)*x_ref[0]+(0.055780633982)*x_ref[1]**o+(0.0447074319839)*x_ref[1]
        ref[3,1,2,0]=(-0.190231018951)*x_ref[0]**o+(-0.61318830169)*x_ref[0]+(-0.157202158534)*x_ref[1]**o+(-0.823043159126)*x_ref[1]
        ref[3,1,2,1]=(-0.502637901786)*x_ref[0]**o+(0.781083145697)*x_ref[0]+(-0.663013942651)*x_ref[1]**o+(-0.305726874525)*x_ref[1]
        ref[3,2,0,0]=(0.566955929517)*x_ref[0]**o+(-0.506981068575)*x_ref[0]+(-0.831439289654)*x_ref[1]**o+(-0.337491442765)*x_ref[1]
        ref[3,2,0,1]=(-0.713648844422)*x_ref[0]**o+(-0.00815764297609)*x_ref[0]+(0.829706206618)*x_ref[1]**o+(0.96471376965)*x_ref[1]
        ref[3,2,1,0]=(-0.944705561715)*x_ref[0]**o+(-0.87962603172)*x_ref[0]+(-0.0525190846342)*x_ref[1]**o+(-0.0457463400515)*x_ref[1]
        ref[3,2,1,1]=(0.54922446571)*x_ref[0]**o+(0.651923573438)*x_ref[0]+(-0.00229398468376)*x_ref[1]**o+(-0.00170863787749)*x_ref[1]
        ref[3,2,2,0]=(0.622372079685)*x_ref[0]**o+(-0.553578221845)*x_ref[0]+(-0.737877450857)*x_ref[1]**o+(0.459729854002)*x_ref[1]
        ref[3,2,2,1]=(0.826155732508)*x_ref[0]**o+(-0.000468898258991)*x_ref[0]+(-0.255179342686)*x_ref[1]**o+(-0.80957312045)*x_ref[1]
        ref[3,3,0,0]=(-0.195212469839)*x_ref[0]**o+(-0.0235216449547)*x_ref[0]+(-0.175192639485)*x_ref[1]**o+(-0.58937661213)*x_ref[1]
        ref[3,3,0,1]=(-0.931538910404)*x_ref[0]**o+(0.649282880007)*x_ref[0]+(-0.237841482637)*x_ref[1]**o+(-0.0220516220802)*x_ref[1]
        ref[3,3,1,0]=(0.420911057166)*x_ref[0]**o+(0.200126728184)*x_ref[0]+(-0.829130861656)*x_ref[1]**o+(-0.563462916661)*x_ref[1]
        ref[3,3,1,1]=(-0.148199974575)*x_ref[0]**o+(-0.248825389533)*x_ref[0]+(-0.346424894584)*x_ref[1]**o+(0.339758232978)*x_ref[1]
        ref[3,3,2,0]=(0.249372496215)*x_ref[0]**o+(-0.343626915746)*x_ref[0]+(0.0961351845087)*x_ref[1]**o+(0.612679249068)*x_ref[1]
        ref[3,3,2,1]=(0.985380582259)*x_ref[0]**o+(-0.603514318626)*x_ref[0]+(-0.131105546364)*x_ref[1]**o+(-0.0798158697238)*x_ref[1]
        ref[3,4,0,0]=(0.478635365792)*x_ref[0]**o+(-0.964424450881)*x_ref[0]+(-0.469995535544)*x_ref[1]**o+(-0.72871149515)*x_ref[1]
        ref[3,4,0,1]=(-0.264802297415)*x_ref[0]**o+(-0.926518282507)*x_ref[0]+(0.803390182149)*x_ref[1]**o+(0.387910762706)*x_ref[1]
        ref[3,4,1,0]=(0.0452448579909)*x_ref[0]**o+(-0.0482221515826)*x_ref[0]+(-0.728020494982)*x_ref[1]**o+(-0.288823633249)*x_ref[1]
        ref[3,4,1,1]=(-0.640459595265)*x_ref[0]**o+(0.759525453728)*x_ref[0]+(-0.0280300707884)*x_ref[1]**o+(0.079438187123)*x_ref[1]
        ref[3,4,2,0]=(-0.14185171171)*x_ref[0]**o+(-0.810989314052)*x_ref[0]+(0.921902325773)*x_ref[1]**o+(0.885920658784)*x_ref[1]
        ref[3,4,2,1]=(-0.0101259901393)*x_ref[0]**o+(-0.696541046542)*x_ref[0]+(0.127932822496)*x_ref[1]**o+(-0.822265796323)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.657703355307)*x[0]**o+(0.397813831843)*x[0]+(0.364460538717)*x[1]**o+(-0.845626669617)*x[1]+(-0.987144254126)*x[2]**o+(-0.732384980726)*x[2]
        arg[0,0,0,1]=(-0.13280304886)*x[0]**o+(-0.953796807758)*x[0]+(0.875723746654)*x[1]**o+(0.0764745117242)*x[1]+(-0.536085676221)*x[2]**o+(-0.520611959678)*x[2]
        arg[0,0,1,0]=(0.716429446578)*x[0]**o+(-0.729164693441)*x[0]+(-0.832505557428)*x[1]**o+(-0.329200605611)*x[1]+(0.0310162060964)*x[2]**o+(-0.869978941999)*x[2]
        arg[0,0,1,1]=(0.110253530584)*x[0]**o+(0.995528486822)*x[0]+(0.838577797395)*x[1]**o+(0.443987752225)*x[1]+(0.387337959057)*x[2]**o+(0.548448752612)*x[2]
        arg[0,0,2,0]=(-0.995737515431)*x[0]**o+(-0.899807241717)*x[0]+(-0.0803287764138)*x[1]**o+(0.242680051556)*x[1]+(0.58974836823)*x[2]**o+(-0.498450188944)*x[2]
        arg[0,0,2,1]=(0.341561541555)*x[0]**o+(-0.00541988184597)*x[0]+(-0.204832898701)*x[1]**o+(-0.957247935865)*x[1]+(-0.441002902064)*x[2]**o+(0.756843480304)*x[2]
        arg[0,1,0,0]=(0.136201526244)*x[0]**o+(0.517677304193)*x[0]+(0.523031610748)*x[1]**o+(-0.908998053383)*x[1]+(-0.629649678416)*x[2]**o+(0.762977917512)*x[2]
        arg[0,1,0,1]=(-0.103521080864)*x[0]**o+(-0.434410222595)*x[0]+(0.418239530241)*x[1]**o+(0.65638475119)*x[1]+(0.413036393129)*x[2]**o+(-0.220365518353)*x[2]
        arg[0,1,1,0]=(-0.415690201483)*x[0]**o+(-0.969323394689)*x[0]+(-0.295639420195)*x[1]**o+(-0.24915002749)*x[1]+(-0.00927627101467)*x[2]**o+(0.520335261634)*x[2]
        arg[0,1,1,1]=(-0.262653741797)*x[0]**o+(-0.463590591666)*x[0]+(0.568893042923)*x[1]**o+(-0.869851332898)*x[1]+(-0.463261399415)*x[2]**o+(0.0281915686202)*x[2]
        arg[0,1,2,0]=(0.377159503058)*x[0]**o+(0.484512261943)*x[0]+(0.0303594628925)*x[1]**o+(0.482051243004)*x[1]+(0.240643013949)*x[2]**o+(-0.943908617705)*x[2]
        arg[0,1,2,1]=(0.843854614097)*x[0]**o+(0.018041534553)*x[0]+(0.264198703361)*x[1]**o+(0.732889517628)*x[1]+(-0.154807384394)*x[2]**o+(0.60513622187)*x[2]
        arg[0,2,0,0]=(-0.234758118286)*x[0]**o+(-0.77807396119)*x[0]+(-0.97783091103)*x[1]**o+(-0.716185223828)*x[1]+(-0.573126474259)*x[2]**o+(0.403110701555)*x[2]
        arg[0,2,0,1]=(-0.853878344804)*x[0]**o+(-0.482644826561)*x[0]+(-0.826041149012)*x[1]**o+(0.596376234798)*x[1]+(-0.216400517274)*x[2]**o+(-0.323127194259)*x[2]
        arg[0,2,1,0]=(0.931551121492)*x[0]**o+(-0.527132174379)*x[0]+(0.373496704463)*x[1]**o+(-0.375215826463)*x[1]+(-0.639167297984)*x[2]**o+(0.143608733189)*x[2]
        arg[0,2,1,1]=(-0.982206000879)*x[0]**o+(0.836589596534)*x[0]+(0.606275000138)*x[1]**o+(-0.821469453892)*x[1]+(-0.633828380948)*x[2]**o+(-0.637688777216)*x[2]
        arg[0,2,2,0]=(0.441649557529)*x[0]**o+(-0.34978365051)*x[0]+(0.514175365405)*x[1]**o+(-0.903212271234)*x[1]+(0.70236887156)*x[2]**o+(0.72953045189)*x[2]
        arg[0,2,2,1]=(0.302787896978)*x[0]**o+(0.157598071021)*x[0]+(-0.357101947632)*x[1]**o+(0.381098383156)*x[1]+(-0.552546036689)*x[2]**o+(0.448181721984)*x[2]
        arg[0,3,0,0]=(0.662181077399)*x[0]**o+(-0.283755239518)*x[0]+(0.33694741278)*x[1]**o+(0.275801685868)*x[1]+(0.526155259101)*x[2]**o+(-0.378241855559)*x[2]
        arg[0,3,0,1]=(0.157822133581)*x[0]**o+(0.25402072747)*x[0]+(0.522842123099)*x[1]**o+(0.510595773584)*x[1]+(-0.829239126666)*x[2]**o+(0.990879342044)*x[2]
        arg[0,3,1,0]=(-0.775925280059)*x[0]**o+(-0.726971388023)*x[0]+(-0.630088730753)*x[1]**o+(-0.731024178237)*x[1]+(-0.288643541196)*x[2]**o+(-0.283708428793)*x[2]
        arg[0,3,1,1]=(-0.883176640445)*x[0]**o+(-0.0810211976762)*x[0]+(-0.102919217784)*x[1]**o+(0.271384866385)*x[1]+(0.980796900477)*x[2]**o+(0.439294038672)*x[2]
        arg[0,3,2,0]=(-0.826688221502)*x[0]**o+(-0.909388347347)*x[0]+(0.164406927722)*x[1]**o+(0.60157189364)*x[1]+(-0.885578176748)*x[2]**o+(0.610008674247)*x[2]
        arg[0,3,2,1]=(0.130154197193)*x[0]**o+(-0.754703602975)*x[0]+(0.00463710497017)*x[1]**o+(0.502162916123)*x[1]+(-0.749124548541)*x[2]**o+(0.946592962092)*x[2]
        arg[0,4,0,0]=(-0.0478347448906)*x[0]**o+(-0.812937871882)*x[0]+(-0.460982012551)*x[1]**o+(0.317485998991)*x[1]+(-0.593671683477)*x[2]**o+(0.551385718668)*x[2]
        arg[0,4,0,1]=(0.281998767502)*x[0]**o+(-0.0613830311997)*x[0]+(0.981429151019)*x[1]**o+(-0.580567714008)*x[1]+(0.0369483005331)*x[2]**o+(-0.74132358588)*x[2]
        arg[0,4,1,0]=(0.525845217483)*x[0]**o+(0.907803915216)*x[0]+(0.111429079263)*x[1]**o+(-0.653030401377)*x[1]+(0.0814793913696)*x[2]**o+(-0.419201413875)*x[2]
        arg[0,4,1,1]=(0.862892518463)*x[0]**o+(0.34033875398)*x[0]+(0.0371376953103)*x[1]**o+(0.568160310726)*x[1]+(-0.237231036717)*x[2]**o+(-0.130046899302)*x[2]
        arg[0,4,2,0]=(0.223651181698)*x[0]**o+(0.730991398203)*x[0]+(0.133551525027)*x[1]**o+(0.685783540126)*x[1]+(0.339586395814)*x[2]**o+(-0.524371194863)*x[2]
        arg[0,4,2,1]=(0.220110871383)*x[0]**o+(-0.37385773877)*x[0]+(0.773372000545)*x[1]**o+(-0.4261356565)*x[1]+(-0.619438438945)*x[2]**o+(0.126288949127)*x[2]
        arg[1,0,0,0]=(0.727680196631)*x[0]**o+(-0.876280912599)*x[0]+(0.576553782915)*x[1]**o+(-0.0497780322367)*x[1]+(-0.690257457282)*x[2]**o+(-0.11130953054)*x[2]
        arg[1,0,0,1]=(0.331863815938)*x[0]**o+(0.169379824389)*x[0]+(0.665507756012)*x[1]**o+(0.94933357273)*x[1]+(-0.610506070093)*x[2]**o+(0.752476708027)*x[2]
        arg[1,0,1,0]=(0.851320583415)*x[0]**o+(-0.697235996161)*x[0]+(0.235192165807)*x[1]**o+(0.325473517022)*x[1]+(-0.424683549673)*x[2]**o+(0.228953796486)*x[2]
        arg[1,0,1,1]=(0.804192098517)*x[0]**o+(-0.684337197059)*x[0]+(-0.349251615715)*x[1]**o+(0.596165705743)*x[1]+(-0.636033221182)*x[2]**o+(-0.304367076307)*x[2]
        arg[1,0,2,0]=(-0.237305477187)*x[0]**o+(-0.997679652719)*x[0]+(0.485994068389)*x[1]**o+(0.44508530534)*x[1]+(0.940139935628)*x[2]**o+(0.0789221540062)*x[2]
        arg[1,0,2,1]=(0.814928463364)*x[0]**o+(-0.257693939789)*x[0]+(0.179643785703)*x[1]**o+(-0.696436530567)*x[1]+(-0.592226945554)*x[2]**o+(-0.454620411499)*x[2]
        arg[1,1,0,0]=(-0.933925672439)*x[0]**o+(0.212298180817)*x[0]+(-0.447836316801)*x[1]**o+(0.180796376138)*x[1]+(-0.749910125701)*x[2]**o+(0.134540433851)*x[2]
        arg[1,1,0,1]=(-0.649009346978)*x[0]**o+(0.0316819485953)*x[0]+(-0.718001767835)*x[1]**o+(0.289608435124)*x[1]+(-0.906045493384)*x[2]**o+(0.851021484125)*x[2]
        arg[1,1,1,0]=(-0.216265404566)*x[0]**o+(0.657073282172)*x[0]+(-0.796155541438)*x[1]**o+(0.338980498824)*x[1]+(0.477444448193)*x[2]**o+(0.674278909522)*x[2]
        arg[1,1,1,1]=(-0.67605942954)*x[0]**o+(0.457997253216)*x[0]+(-0.973576381948)*x[1]**o+(0.430432753773)*x[1]+(0.0329169564451)*x[2]**o+(-0.422512702485)*x[2]
        arg[1,1,2,0]=(-0.32458256139)*x[0]**o+(0.289568625922)*x[0]+(-0.529564027476)*x[1]**o+(0.596449833118)*x[1]+(0.960461031989)*x[2]**o+(0.00801425321069)*x[2]
        arg[1,1,2,1]=(-0.0425883615923)*x[0]**o+(-0.313926807333)*x[0]+(0.766062773671)*x[1]**o+(-0.857784560309)*x[1]+(0.914867381682)*x[2]**o+(-0.312493334639)*x[2]
        arg[1,2,0,0]=(-0.637151767355)*x[0]**o+(-0.679947136952)*x[0]+(-0.0838349105494)*x[1]**o+(0.00800835879666)*x[1]+(0.385024194519)*x[2]**o+(-0.627828236571)*x[2]
        arg[1,2,0,1]=(-0.62580153007)*x[0]**o+(-0.183834415542)*x[0]+(0.322406823575)*x[1]**o+(0.302908755868)*x[1]+(0.727234001542)*x[2]**o+(-0.933944878425)*x[2]
        arg[1,2,1,0]=(-0.897198312615)*x[0]**o+(0.825086526945)*x[0]+(-0.031001758991)*x[1]**o+(0.987845551304)*x[1]+(-0.551212418179)*x[2]**o+(0.657813241764)*x[2]
        arg[1,2,1,1]=(-0.544595822278)*x[0]**o+(0.595858429591)*x[0]+(-0.557503238514)*x[1]**o+(0.507560179011)*x[1]+(-0.519438270771)*x[2]**o+(-0.471678264995)*x[2]
        arg[1,2,2,0]=(0.135616278099)*x[0]**o+(0.40676743342)*x[0]+(0.458810719213)*x[1]**o+(-0.881620410228)*x[1]+(-0.897267732846)*x[2]**o+(0.665458992487)*x[2]
        arg[1,2,2,1]=(0.83854610457)*x[0]**o+(0.670263196782)*x[0]+(0.551767734306)*x[1]**o+(0.883934821215)*x[1]+(0.455375261768)*x[2]**o+(0.95604480045)*x[2]
        arg[1,3,0,0]=(0.705853698882)*x[0]**o+(0.84681314245)*x[0]+(-0.0984020252408)*x[1]**o+(0.943165990555)*x[1]+(-0.428955841327)*x[2]**o+(0.0961618993003)*x[2]
        arg[1,3,0,1]=(0.583731800337)*x[0]**o+(-0.738079255003)*x[0]+(0.315556768944)*x[1]**o+(-0.489631118687)*x[1]+(0.371775779104)*x[2]**o+(0.82029490872)*x[2]
        arg[1,3,1,0]=(-0.885657545142)*x[0]**o+(0.542706113286)*x[0]+(0.800319452136)*x[1]**o+(0.314194184897)*x[1]+(-0.320905532427)*x[2]**o+(-0.314307365683)*x[2]
        arg[1,3,1,1]=(-0.217030771566)*x[0]**o+(0.260108519031)*x[0]+(-0.686624725975)*x[1]**o+(-0.869322439609)*x[1]+(0.558806239315)*x[2]**o+(-0.440661121246)*x[2]
        arg[1,3,2,0]=(-0.839335541382)*x[0]**o+(0.228726621639)*x[0]+(0.354403444002)*x[1]**o+(0.215616048959)*x[1]+(0.49622647562)*x[2]**o+(-0.725639112729)*x[2]
        arg[1,3,2,1]=(-0.439687404632)*x[0]**o+(-0.820583921269)*x[0]+(-0.863730409631)*x[1]**o+(0.119773401911)*x[1]+(0.460158483187)*x[2]**o+(0.757153436592)*x[2]
        arg[1,4,0,0]=(-0.77288839701)*x[0]**o+(-0.570203714622)*x[0]+(0.93832772305)*x[1]**o+(-0.759089944596)*x[1]+(0.000174937961704)*x[2]**o+(0.812917484286)*x[2]
        arg[1,4,0,1]=(-0.25535378432)*x[0]**o+(-0.0938207170929)*x[0]+(0.0545321717685)*x[1]**o+(-0.414414792674)*x[1]+(-0.297698896712)*x[2]**o+(-0.367166660571)*x[2]
        arg[1,4,1,0]=(0.743408112248)*x[0]**o+(0.852863814785)*x[0]+(0.259483951249)*x[1]**o+(-0.519288864163)*x[1]+(-0.305772267727)*x[2]**o+(-0.401384318883)*x[2]
        arg[1,4,1,1]=(0.194889815352)*x[0]**o+(0.522093986804)*x[0]+(-0.343320977396)*x[1]**o+(0.21840236597)*x[1]+(-0.672107409136)*x[2]**o+(-0.600203805002)*x[2]
        arg[1,4,2,0]=(0.420860556361)*x[0]**o+(0.789879869049)*x[0]+(0.100987805744)*x[1]**o+(0.802794815123)*x[1]+(-0.69466030201)*x[2]**o+(-0.746043845966)*x[2]
        arg[1,4,2,1]=(-0.306621579211)*x[0]**o+(-0.0676982346933)*x[0]+(0.568861573715)*x[1]**o+(0.0823366334578)*x[1]+(0.501187255641)*x[2]**o+(-0.538615164478)*x[2]
        arg[2,0,0,0]=(-0.231052916706)*x[0]**o+(0.127790831537)*x[0]+(-0.475600435038)*x[1]**o+(0.431586151473)*x[1]+(-0.882365547235)*x[2]**o+(0.778027701879)*x[2]
        arg[2,0,0,1]=(0.0709490835498)*x[0]**o+(0.442614693755)*x[0]+(0.870405247692)*x[1]**o+(-0.648538779515)*x[1]+(0.751107652054)*x[2]**o+(0.0122956931127)*x[2]
        arg[2,0,1,0]=(0.703128329145)*x[0]**o+(0.259775843631)*x[0]+(0.311958888594)*x[1]**o+(-0.0445092402277)*x[1]+(0.338222457799)*x[2]**o+(0.845995300456)*x[2]
        arg[2,0,1,1]=(-0.653977898919)*x[0]**o+(0.0495539350857)*x[0]+(0.4013623408)*x[1]**o+(0.490747039887)*x[1]+(-0.480487636729)*x[2]**o+(0.361660115775)*x[2]
        arg[2,0,2,0]=(0.215778506945)*x[0]**o+(0.316853875053)*x[0]+(0.556770639829)*x[1]**o+(-0.499729882355)*x[1]+(0.692012327582)*x[2]**o+(0.00681314946195)*x[2]
        arg[2,0,2,1]=(0.432077857223)*x[0]**o+(-0.0497579036988)*x[0]+(-0.465403408565)*x[1]**o+(0.125421502956)*x[1]+(-0.313541879877)*x[2]**o+(0.113957941731)*x[2]
        arg[2,1,0,0]=(-0.528126103192)*x[0]**o+(-0.141306512096)*x[0]+(-0.505835313198)*x[1]**o+(-0.0433171534408)*x[1]+(-0.522465687628)*x[2]**o+(-0.16728285732)*x[2]
        arg[2,1,0,1]=(0.892019660948)*x[0]**o+(0.646944130193)*x[0]+(0.865495756748)*x[1]**o+(0.949095276008)*x[1]+(0.786394755068)*x[2]**o+(0.482441081687)*x[2]
        arg[2,1,1,0]=(-0.0485295711204)*x[0]**o+(0.250163927911)*x[0]+(0.0387861125998)*x[1]**o+(-0.0789121929817)*x[1]+(-0.982330126583)*x[2]**o+(0.66328870814)*x[2]
        arg[2,1,1,1]=(-0.494726556678)*x[0]**o+(-0.961342091528)*x[0]+(0.211410563083)*x[1]**o+(-0.884287887286)*x[1]+(-0.583401908135)*x[2]**o+(-0.361423905219)*x[2]
        arg[2,1,2,0]=(-0.326785246692)*x[0]**o+(0.596272945325)*x[0]+(0.625039539234)*x[1]**o+(0.88699688986)*x[1]+(-0.649803810419)*x[2]**o+(0.890718271469)*x[2]
        arg[2,1,2,1]=(0.702110072391)*x[0]**o+(0.72009439563)*x[0]+(-0.930941572511)*x[1]**o+(0.955877158334)*x[1]+(0.1211281332)*x[2]**o+(0.658738626852)*x[2]
        arg[2,2,0,0]=(-0.177402589697)*x[0]**o+(0.701090894256)*x[0]+(-0.00205333079339)*x[1]**o+(-0.427862310859)*x[1]+(0.529186374803)*x[2]**o+(-0.110462035551)*x[2]
        arg[2,2,0,1]=(-0.104112165367)*x[0]**o+(0.043539351452)*x[0]+(0.757384935306)*x[1]**o+(-0.451086492844)*x[1]+(0.0549627297033)*x[2]**o+(-0.311718955315)*x[2]
        arg[2,2,1,0]=(-0.579986090808)*x[0]**o+(-0.922872917924)*x[0]+(0.66422700207)*x[1]**o+(0.947304780214)*x[1]+(-0.0492553517704)*x[2]**o+(0.817225619153)*x[2]
        arg[2,2,1,1]=(0.306298476233)*x[0]**o+(-0.620190960783)*x[0]+(-0.684876074013)*x[1]**o+(-0.371829616982)*x[1]+(0.904709743321)*x[2]**o+(0.170802273852)*x[2]
        arg[2,2,2,0]=(0.317943399758)*x[0]**o+(0.518026054212)*x[0]+(-0.801462311167)*x[1]**o+(0.767424157898)*x[1]+(-0.547271995492)*x[2]**o+(-0.912619177188)*x[2]
        arg[2,2,2,1]=(-0.848711258358)*x[0]**o+(0.456489097468)*x[0]+(-0.310979078367)*x[1]**o+(-0.946717007726)*x[1]+(0.454363833646)*x[2]**o+(-0.715701991077)*x[2]
        arg[2,3,0,0]=(0.605493533014)*x[0]**o+(0.478437880368)*x[0]+(-0.672727491605)*x[1]**o+(0.831259256009)*x[1]+(-0.434153354239)*x[2]**o+(-0.505363218578)*x[2]
        arg[2,3,0,1]=(-0.0703539676643)*x[0]**o+(-0.452352344215)*x[0]+(0.0403568720903)*x[1]**o+(0.959415939826)*x[1]+(-0.784519698477)*x[2]**o+(-0.691664442958)*x[2]
        arg[2,3,1,0]=(0.696353768719)*x[0]**o+(0.295477297618)*x[0]+(0.580540484825)*x[1]**o+(0.098954967023)*x[1]+(-0.547594444517)*x[2]**o+(-0.574783665389)*x[2]
        arg[2,3,1,1]=(0.997841887408)*x[0]**o+(0.857864040549)*x[0]+(-0.413188343971)*x[1]**o+(0.344770631687)*x[1]+(-0.127257058552)*x[2]**o+(0.132618819456)*x[2]
        arg[2,3,2,0]=(0.174045751083)*x[0]**o+(0.866724759298)*x[0]+(0.0818516103455)*x[1]**o+(-0.891670609599)*x[1]+(0.982721717044)*x[2]**o+(0.378710155856)*x[2]
        arg[2,3,2,1]=(0.7760778906)*x[0]**o+(-0.710104290654)*x[0]+(-0.471259845977)*x[1]**o+(-0.791979211724)*x[1]+(-0.845652753706)*x[2]**o+(-0.301795837852)*x[2]
        arg[2,4,0,0]=(-0.698193658421)*x[0]**o+(-0.831956135385)*x[0]+(-0.442637915133)*x[1]**o+(0.108715813637)*x[1]+(-0.888348149694)*x[2]**o+(-0.90115888496)*x[2]
        arg[2,4,0,1]=(0.770237705577)*x[0]**o+(0.889304763618)*x[0]+(-0.139196183564)*x[1]**o+(0.332516232239)*x[1]+(-0.0218103253052)*x[2]**o+(-0.368033928854)*x[2]
        arg[2,4,1,0]=(0.675479351844)*x[0]**o+(0.880998418303)*x[0]+(0.408381837105)*x[1]**o+(0.717402364084)*x[1]+(-0.303902947422)*x[2]**o+(-0.322350049962)*x[2]
        arg[2,4,1,1]=(-0.480881673289)*x[0]**o+(0.30290141529)*x[0]+(-0.538411390886)*x[1]**o+(0.260090848702)*x[1]+(0.638130764808)*x[2]**o+(0.756338975575)*x[2]
        arg[2,4,2,0]=(-0.299233221889)*x[0]**o+(-0.26471083761)*x[0]+(0.717191172884)*x[1]**o+(0.488553216118)*x[1]+(0.718956726644)*x[2]**o+(0.370889025703)*x[2]
        arg[2,4,2,1]=(-0.989020272269)*x[0]**o+(-0.891423884362)*x[0]+(0.486153100693)*x[1]**o+(-0.971171395246)*x[1]+(-0.766942968184)*x[2]**o+(0.728728968862)*x[2]
        arg[3,0,0,0]=(0.756186246227)*x[0]**o+(0.76507856268)*x[0]+(0.771165000414)*x[1]**o+(0.673968873856)*x[1]+(0.827045846253)*x[2]**o+(-0.788265752412)*x[2]
        arg[3,0,0,1]=(-0.66925731439)*x[0]**o+(-0.0110127596435)*x[0]+(0.944388255414)*x[1]**o+(-0.613832072486)*x[1]+(0.0977542120724)*x[2]**o+(0.100776856689)*x[2]
        arg[3,0,1,0]=(0.905138713064)*x[0]**o+(0.673051338153)*x[0]+(0.240614253653)*x[1]**o+(-0.964132330033)*x[1]+(-0.114961018838)*x[2]**o+(-0.438756962988)*x[2]
        arg[3,0,1,1]=(-0.192259182094)*x[0]**o+(0.12315590119)*x[0]+(-0.907744420226)*x[1]**o+(-0.483981269809)*x[1]+(-0.169699431472)*x[2]**o+(-0.45231635406)*x[2]
        arg[3,0,2,0]=(0.895387255885)*x[0]**o+(0.23892997411)*x[0]+(-0.396010446618)*x[1]**o+(0.94853984413)*x[1]+(-0.0165259637241)*x[2]**o+(-0.312558539149)*x[2]
        arg[3,0,2,1]=(0.710478713361)*x[0]**o+(-0.109316102216)*x[0]+(-0.042711568986)*x[1]**o+(-0.448508085742)*x[1]+(-0.0979193346011)*x[2]**o+(0.93156672211)*x[2]
        arg[3,1,0,0]=(-0.21605108221)*x[0]**o+(0.410466481712)*x[0]+(0.799288386783)*x[1]**o+(0.622866017113)*x[1]+(-0.236698938159)*x[2]**o+(0.246625241928)*x[2]
        arg[3,1,0,1]=(0.989568527762)*x[0]**o+(0.0212906378151)*x[0]+(0.311279947324)*x[1]**o+(-0.231412902902)*x[1]+(-0.444760786142)*x[2]**o+(0.908171985421)*x[2]
        arg[3,1,1,0]=(0.42605234231)*x[0]**o+(-0.58629332944)*x[0]+(0.322063640761)*x[1]**o+(0.837771339825)*x[1]+(-0.911083211394)*x[2]**o+(-0.0699607581992)*x[2]
        arg[3,1,1,1]=(0.75449445591)*x[0]**o+(0.722970800307)*x[0]+(-0.582970919465)*x[1]**o+(-0.868818538212)*x[1]+(0.940067636181)*x[2]**o+(-0.659469743617)*x[2]
        arg[3,1,2,0]=(-0.391224924663)*x[0]**o+(-0.768173177737)*x[0]+(-0.856876347997)*x[1]**o+(-0.0451180499647)*x[1]+(0.198590679246)*x[2]**o+(0.187329095476)*x[2]
        arg[3,1,2,1]=(0.483309063092)*x[0]**o+(-0.0777241026893)*x[0]+(0.0438427457286)*x[1]**o+(-0.73195704867)*x[1]+(-0.439660745725)*x[2]**o+(0.279501672861)*x[2]
        arg[3,2,0,0]=(0.799935089139)*x[0]**o+(-0.306860809857)*x[0]+(0.169104975039)*x[1]**o+(-0.3480717525)*x[1]+(0.215231688)*x[2]**o+(-0.528775376945)*x[2]
        arg[3,2,0,1]=(0.917815789587)*x[0]**o+(0.0672910338102)*x[0]+(-0.186550440011)*x[1]**o+(0.922579714252)*x[1]+(0.924162379857)*x[2]**o+(-0.769320483618)*x[2]
        arg[3,2,1,0]=(0.604324468058)*x[0]**o+(-0.0109731419053)*x[0]+(0.229447318325)*x[1]**o+(-0.480693435232)*x[1]+(-0.881779256999)*x[2]**o+(0.288656394771)*x[2]
        arg[3,2,1,1]=(0.934680723352)*x[0]**o+(0.993838528376)*x[0]+(0.957781588502)*x[1]**o+(-0.859837057866)*x[1]+(-0.702139964493)*x[2]**o+(0.0538177124609)*x[2]
        arg[3,2,2,0]=(-0.142809047814)*x[0]**o+(-0.472594415134)*x[0]+(0.0381782768377)*x[1]**o+(0.727331491769)*x[1]+(0.0910992897581)*x[2]**o+(-0.845316063851)*x[2]
        arg[3,2,2,1]=(-0.217483358272)*x[0]**o+(0.991535092285)*x[0]+(-0.399175896468)*x[1]**o+(-0.665836512895)*x[1]+(0.0983700911493)*x[2]**o+(0.321811300155)*x[2]
        arg[3,3,0,0]=(0.797049585464)*x[0]**o+(-0.724467959299)*x[0]+(0.19971589973)*x[1]**o+(0.845928461864)*x[1]+(-0.365158866359)*x[2]**o+(-0.698274270132)*x[2]
        arg[3,3,0,1]=(-0.638474704878)*x[0]**o+(0.551866433352)*x[0]+(0.192309452979)*x[1]**o+(-0.703649325501)*x[1]+(0.823910578138)*x[2]**o+(-0.917446989634)*x[2]
        arg[3,3,1,0]=(0.217577282169)*x[0]**o+(0.521716803668)*x[0]+(-0.116587783856)*x[1]**o+(0.529672391417)*x[1]+(-0.798918158284)*x[2]**o+(-0.0996139511955)*x[2]
        arg[3,3,1,1]=(0.498965049079)*x[0]**o+(0.316489921186)*x[0]+(-0.0988355471817)*x[1]**o+(-0.0477508287681)*x[1]+(-0.778914523993)*x[2]**o+(0.189397812723)*x[2]
        arg[3,3,2,0]=(0.203823698152)*x[0]**o+(-0.340499414266)*x[0]+(0.353374693517)*x[1]**o+(-0.91035264494)*x[1]+(-0.779204084198)*x[2]**o+(-0.00276454165248)*x[2]
        arg[3,3,2,1]=(0.295232007123)*x[0]**o+(0.271766544502)*x[0]+(0.431956430336)*x[1]**o+(0.896481311536)*x[1]+(-0.325014902487)*x[2]**o+(0.70836170284)*x[2]
        arg[3,4,0,0]=(-0.407535202468)*x[0]**o+(-0.33055681091)*x[0]+(0.129762143151)*x[1]**o+(0.352087436228)*x[1]+(-0.538770401565)*x[2]**o+(-0.519071077045)*x[2]
        arg[3,4,0,1]=(0.846011435499)*x[0]**o+(0.172822133672)*x[0]+(-0.740732553847)*x[1]**o+(0.230062837231)*x[1]+(-0.231810007237)*x[2]**o+(-0.568366338923)*x[2]
        arg[3,4,1,0]=(-0.198091119057)*x[0]**o+(0.957989508794)*x[0]+(0.861739071175)*x[1]**o+(0.923450697883)*x[1]+(-0.755262504937)*x[2]**o+(0.102897579936)*x[2]
        arg[3,4,1,1]=(0.198829545353)*x[0]**o+(-0.757182006818)*x[0]+(0.254092337808)*x[1]**o+(-0.00815072620097)*x[1]+(0.347421988308)*x[2]**o+(-0.0215277616923)*x[2]
        arg[3,4,2,0]=(0.775062112529)*x[0]**o+(-0.77746792774)*x[0]+(-0.666490294869)*x[1]**o+(-0.679971069878)*x[1]+(-0.972450779894)*x[2]**o+(-0.705322282477)*x[2]
        arg[3,4,2,1]=(-0.0994927587382)*x[0]**o+(0.293071774786)*x[0]+(0.370784399397)*x[1]**o+(0.931411225946)*x[1]+(-0.55789163294)*x[2]**o+(0.0831890658592)*x[2]
        ref[0,0,0,0]=(-0.657703355307)*x_ref[0]**o+(0.397813831843)*x_ref[0]+(0.364460538717)*x_ref[1]**o+(-0.845626669617)*x_ref[1]+(-0.987144254126)*x_ref[2]**o+(-0.732384980726)*x_ref[2]
        ref[0,0,0,1]=(-0.13280304886)*x_ref[0]**o+(-0.953796807758)*x_ref[0]+(0.875723746654)*x_ref[1]**o+(0.0764745117242)*x_ref[1]+(-0.536085676221)*x_ref[2]**o+(-0.520611959678)*x_ref[2]
        ref[0,0,1,0]=(0.716429446578)*x_ref[0]**o+(-0.729164693441)*x_ref[0]+(-0.832505557428)*x_ref[1]**o+(-0.329200605611)*x_ref[1]+(0.0310162060964)*x_ref[2]**o+(-0.869978941999)*x_ref[2]
        ref[0,0,1,1]=(0.110253530584)*x_ref[0]**o+(0.995528486822)*x_ref[0]+(0.838577797395)*x_ref[1]**o+(0.443987752225)*x_ref[1]+(0.387337959057)*x_ref[2]**o+(0.548448752612)*x_ref[2]
        ref[0,0,2,0]=(-0.995737515431)*x_ref[0]**o+(-0.899807241717)*x_ref[0]+(-0.0803287764138)*x_ref[1]**o+(0.242680051556)*x_ref[1]+(0.58974836823)*x_ref[2]**o+(-0.498450188944)*x_ref[2]
        ref[0,0,2,1]=(0.341561541555)*x_ref[0]**o+(-0.00541988184597)*x_ref[0]+(-0.204832898701)*x_ref[1]**o+(-0.957247935865)*x_ref[1]+(-0.441002902064)*x_ref[2]**o+(0.756843480304)*x_ref[2]
        ref[0,1,0,0]=(0.136201526244)*x_ref[0]**o+(0.517677304193)*x_ref[0]+(0.523031610748)*x_ref[1]**o+(-0.908998053383)*x_ref[1]+(-0.629649678416)*x_ref[2]**o+(0.762977917512)*x_ref[2]
        ref[0,1,0,1]=(-0.103521080864)*x_ref[0]**o+(-0.434410222595)*x_ref[0]+(0.418239530241)*x_ref[1]**o+(0.65638475119)*x_ref[1]+(0.413036393129)*x_ref[2]**o+(-0.220365518353)*x_ref[2]
        ref[0,1,1,0]=(-0.415690201483)*x_ref[0]**o+(-0.969323394689)*x_ref[0]+(-0.295639420195)*x_ref[1]**o+(-0.24915002749)*x_ref[1]+(-0.00927627101467)*x_ref[2]**o+(0.520335261634)*x_ref[2]
        ref[0,1,1,1]=(-0.262653741797)*x_ref[0]**o+(-0.463590591666)*x_ref[0]+(0.568893042923)*x_ref[1]**o+(-0.869851332898)*x_ref[1]+(-0.463261399415)*x_ref[2]**o+(0.0281915686202)*x_ref[2]
        ref[0,1,2,0]=(0.377159503058)*x_ref[0]**o+(0.484512261943)*x_ref[0]+(0.0303594628925)*x_ref[1]**o+(0.482051243004)*x_ref[1]+(0.240643013949)*x_ref[2]**o+(-0.943908617705)*x_ref[2]
        ref[0,1,2,1]=(0.843854614097)*x_ref[0]**o+(0.018041534553)*x_ref[0]+(0.264198703361)*x_ref[1]**o+(0.732889517628)*x_ref[1]+(-0.154807384394)*x_ref[2]**o+(0.60513622187)*x_ref[2]
        ref[0,2,0,0]=(-0.234758118286)*x_ref[0]**o+(-0.77807396119)*x_ref[0]+(-0.97783091103)*x_ref[1]**o+(-0.716185223828)*x_ref[1]+(-0.573126474259)*x_ref[2]**o+(0.403110701555)*x_ref[2]
        ref[0,2,0,1]=(-0.853878344804)*x_ref[0]**o+(-0.482644826561)*x_ref[0]+(-0.826041149012)*x_ref[1]**o+(0.596376234798)*x_ref[1]+(-0.216400517274)*x_ref[2]**o+(-0.323127194259)*x_ref[2]
        ref[0,2,1,0]=(0.931551121492)*x_ref[0]**o+(-0.527132174379)*x_ref[0]+(0.373496704463)*x_ref[1]**o+(-0.375215826463)*x_ref[1]+(-0.639167297984)*x_ref[2]**o+(0.143608733189)*x_ref[2]
        ref[0,2,1,1]=(-0.982206000879)*x_ref[0]**o+(0.836589596534)*x_ref[0]+(0.606275000138)*x_ref[1]**o+(-0.821469453892)*x_ref[1]+(-0.633828380948)*x_ref[2]**o+(-0.637688777216)*x_ref[2]
        ref[0,2,2,0]=(0.441649557529)*x_ref[0]**o+(-0.34978365051)*x_ref[0]+(0.514175365405)*x_ref[1]**o+(-0.903212271234)*x_ref[1]+(0.70236887156)*x_ref[2]**o+(0.72953045189)*x_ref[2]
        ref[0,2,2,1]=(0.302787896978)*x_ref[0]**o+(0.157598071021)*x_ref[0]+(-0.357101947632)*x_ref[1]**o+(0.381098383156)*x_ref[1]+(-0.552546036689)*x_ref[2]**o+(0.448181721984)*x_ref[2]
        ref[0,3,0,0]=(0.662181077399)*x_ref[0]**o+(-0.283755239518)*x_ref[0]+(0.33694741278)*x_ref[1]**o+(0.275801685868)*x_ref[1]+(0.526155259101)*x_ref[2]**o+(-0.378241855559)*x_ref[2]
        ref[0,3,0,1]=(0.157822133581)*x_ref[0]**o+(0.25402072747)*x_ref[0]+(0.522842123099)*x_ref[1]**o+(0.510595773584)*x_ref[1]+(-0.829239126666)*x_ref[2]**o+(0.990879342044)*x_ref[2]
        ref[0,3,1,0]=(-0.775925280059)*x_ref[0]**o+(-0.726971388023)*x_ref[0]+(-0.630088730753)*x_ref[1]**o+(-0.731024178237)*x_ref[1]+(-0.288643541196)*x_ref[2]**o+(-0.283708428793)*x_ref[2]
        ref[0,3,1,1]=(-0.883176640445)*x_ref[0]**o+(-0.0810211976762)*x_ref[0]+(-0.102919217784)*x_ref[1]**o+(0.271384866385)*x_ref[1]+(0.980796900477)*x_ref[2]**o+(0.439294038672)*x_ref[2]
        ref[0,3,2,0]=(-0.826688221502)*x_ref[0]**o+(-0.909388347347)*x_ref[0]+(0.164406927722)*x_ref[1]**o+(0.60157189364)*x_ref[1]+(-0.885578176748)*x_ref[2]**o+(0.610008674247)*x_ref[2]
        ref[0,3,2,1]=(0.130154197193)*x_ref[0]**o+(-0.754703602975)*x_ref[0]+(0.00463710497017)*x_ref[1]**o+(0.502162916123)*x_ref[1]+(-0.749124548541)*x_ref[2]**o+(0.946592962092)*x_ref[2]
        ref[0,4,0,0]=(-0.0478347448906)*x_ref[0]**o+(-0.812937871882)*x_ref[0]+(-0.460982012551)*x_ref[1]**o+(0.317485998991)*x_ref[1]+(-0.593671683477)*x_ref[2]**o+(0.551385718668)*x_ref[2]
        ref[0,4,0,1]=(0.281998767502)*x_ref[0]**o+(-0.0613830311997)*x_ref[0]+(0.981429151019)*x_ref[1]**o+(-0.580567714008)*x_ref[1]+(0.0369483005331)*x_ref[2]**o+(-0.74132358588)*x_ref[2]
        ref[0,4,1,0]=(0.525845217483)*x_ref[0]**o+(0.907803915216)*x_ref[0]+(0.111429079263)*x_ref[1]**o+(-0.653030401377)*x_ref[1]+(0.0814793913696)*x_ref[2]**o+(-0.419201413875)*x_ref[2]
        ref[0,4,1,1]=(0.862892518463)*x_ref[0]**o+(0.34033875398)*x_ref[0]+(0.0371376953103)*x_ref[1]**o+(0.568160310726)*x_ref[1]+(-0.237231036717)*x_ref[2]**o+(-0.130046899302)*x_ref[2]
        ref[0,4,2,0]=(0.223651181698)*x_ref[0]**o+(0.730991398203)*x_ref[0]+(0.133551525027)*x_ref[1]**o+(0.685783540126)*x_ref[1]+(0.339586395814)*x_ref[2]**o+(-0.524371194863)*x_ref[2]
        ref[0,4,2,1]=(0.220110871383)*x_ref[0]**o+(-0.37385773877)*x_ref[0]+(0.773372000545)*x_ref[1]**o+(-0.4261356565)*x_ref[1]+(-0.619438438945)*x_ref[2]**o+(0.126288949127)*x_ref[2]
        ref[1,0,0,0]=(0.727680196631)*x_ref[0]**o+(-0.876280912599)*x_ref[0]+(0.576553782915)*x_ref[1]**o+(-0.0497780322367)*x_ref[1]+(-0.690257457282)*x_ref[2]**o+(-0.11130953054)*x_ref[2]
        ref[1,0,0,1]=(0.331863815938)*x_ref[0]**o+(0.169379824389)*x_ref[0]+(0.665507756012)*x_ref[1]**o+(0.94933357273)*x_ref[1]+(-0.610506070093)*x_ref[2]**o+(0.752476708027)*x_ref[2]
        ref[1,0,1,0]=(0.851320583415)*x_ref[0]**o+(-0.697235996161)*x_ref[0]+(0.235192165807)*x_ref[1]**o+(0.325473517022)*x_ref[1]+(-0.424683549673)*x_ref[2]**o+(0.228953796486)*x_ref[2]
        ref[1,0,1,1]=(0.804192098517)*x_ref[0]**o+(-0.684337197059)*x_ref[0]+(-0.349251615715)*x_ref[1]**o+(0.596165705743)*x_ref[1]+(-0.636033221182)*x_ref[2]**o+(-0.304367076307)*x_ref[2]
        ref[1,0,2,0]=(-0.237305477187)*x_ref[0]**o+(-0.997679652719)*x_ref[0]+(0.485994068389)*x_ref[1]**o+(0.44508530534)*x_ref[1]+(0.940139935628)*x_ref[2]**o+(0.0789221540062)*x_ref[2]
        ref[1,0,2,1]=(0.814928463364)*x_ref[0]**o+(-0.257693939789)*x_ref[0]+(0.179643785703)*x_ref[1]**o+(-0.696436530567)*x_ref[1]+(-0.592226945554)*x_ref[2]**o+(-0.454620411499)*x_ref[2]
        ref[1,1,0,0]=(-0.933925672439)*x_ref[0]**o+(0.212298180817)*x_ref[0]+(-0.447836316801)*x_ref[1]**o+(0.180796376138)*x_ref[1]+(-0.749910125701)*x_ref[2]**o+(0.134540433851)*x_ref[2]
        ref[1,1,0,1]=(-0.649009346978)*x_ref[0]**o+(0.0316819485953)*x_ref[0]+(-0.718001767835)*x_ref[1]**o+(0.289608435124)*x_ref[1]+(-0.906045493384)*x_ref[2]**o+(0.851021484125)*x_ref[2]
        ref[1,1,1,0]=(-0.216265404566)*x_ref[0]**o+(0.657073282172)*x_ref[0]+(-0.796155541438)*x_ref[1]**o+(0.338980498824)*x_ref[1]+(0.477444448193)*x_ref[2]**o+(0.674278909522)*x_ref[2]
        ref[1,1,1,1]=(-0.67605942954)*x_ref[0]**o+(0.457997253216)*x_ref[0]+(-0.973576381948)*x_ref[1]**o+(0.430432753773)*x_ref[1]+(0.0329169564451)*x_ref[2]**o+(-0.422512702485)*x_ref[2]
        ref[1,1,2,0]=(-0.32458256139)*x_ref[0]**o+(0.289568625922)*x_ref[0]+(-0.529564027476)*x_ref[1]**o+(0.596449833118)*x_ref[1]+(0.960461031989)*x_ref[2]**o+(0.00801425321069)*x_ref[2]
        ref[1,1,2,1]=(-0.0425883615923)*x_ref[0]**o+(-0.313926807333)*x_ref[0]+(0.766062773671)*x_ref[1]**o+(-0.857784560309)*x_ref[1]+(0.914867381682)*x_ref[2]**o+(-0.312493334639)*x_ref[2]
        ref[1,2,0,0]=(-0.637151767355)*x_ref[0]**o+(-0.679947136952)*x_ref[0]+(-0.0838349105494)*x_ref[1]**o+(0.00800835879666)*x_ref[1]+(0.385024194519)*x_ref[2]**o+(-0.627828236571)*x_ref[2]
        ref[1,2,0,1]=(-0.62580153007)*x_ref[0]**o+(-0.183834415542)*x_ref[0]+(0.322406823575)*x_ref[1]**o+(0.302908755868)*x_ref[1]+(0.727234001542)*x_ref[2]**o+(-0.933944878425)*x_ref[2]
        ref[1,2,1,0]=(-0.897198312615)*x_ref[0]**o+(0.825086526945)*x_ref[0]+(-0.031001758991)*x_ref[1]**o+(0.987845551304)*x_ref[1]+(-0.551212418179)*x_ref[2]**o+(0.657813241764)*x_ref[2]
        ref[1,2,1,1]=(-0.544595822278)*x_ref[0]**o+(0.595858429591)*x_ref[0]+(-0.557503238514)*x_ref[1]**o+(0.507560179011)*x_ref[1]+(-0.519438270771)*x_ref[2]**o+(-0.471678264995)*x_ref[2]
        ref[1,2,2,0]=(0.135616278099)*x_ref[0]**o+(0.40676743342)*x_ref[0]+(0.458810719213)*x_ref[1]**o+(-0.881620410228)*x_ref[1]+(-0.897267732846)*x_ref[2]**o+(0.665458992487)*x_ref[2]
        ref[1,2,2,1]=(0.83854610457)*x_ref[0]**o+(0.670263196782)*x_ref[0]+(0.551767734306)*x_ref[1]**o+(0.883934821215)*x_ref[1]+(0.455375261768)*x_ref[2]**o+(0.95604480045)*x_ref[2]
        ref[1,3,0,0]=(0.705853698882)*x_ref[0]**o+(0.84681314245)*x_ref[0]+(-0.0984020252408)*x_ref[1]**o+(0.943165990555)*x_ref[1]+(-0.428955841327)*x_ref[2]**o+(0.0961618993003)*x_ref[2]
        ref[1,3,0,1]=(0.583731800337)*x_ref[0]**o+(-0.738079255003)*x_ref[0]+(0.315556768944)*x_ref[1]**o+(-0.489631118687)*x_ref[1]+(0.371775779104)*x_ref[2]**o+(0.82029490872)*x_ref[2]
        ref[1,3,1,0]=(-0.885657545142)*x_ref[0]**o+(0.542706113286)*x_ref[0]+(0.800319452136)*x_ref[1]**o+(0.314194184897)*x_ref[1]+(-0.320905532427)*x_ref[2]**o+(-0.314307365683)*x_ref[2]
        ref[1,3,1,1]=(-0.217030771566)*x_ref[0]**o+(0.260108519031)*x_ref[0]+(-0.686624725975)*x_ref[1]**o+(-0.869322439609)*x_ref[1]+(0.558806239315)*x_ref[2]**o+(-0.440661121246)*x_ref[2]
        ref[1,3,2,0]=(-0.839335541382)*x_ref[0]**o+(0.228726621639)*x_ref[0]+(0.354403444002)*x_ref[1]**o+(0.215616048959)*x_ref[1]+(0.49622647562)*x_ref[2]**o+(-0.725639112729)*x_ref[2]
        ref[1,3,2,1]=(-0.439687404632)*x_ref[0]**o+(-0.820583921269)*x_ref[0]+(-0.863730409631)*x_ref[1]**o+(0.119773401911)*x_ref[1]+(0.460158483187)*x_ref[2]**o+(0.757153436592)*x_ref[2]
        ref[1,4,0,0]=(-0.77288839701)*x_ref[0]**o+(-0.570203714622)*x_ref[0]+(0.93832772305)*x_ref[1]**o+(-0.759089944596)*x_ref[1]+(0.000174937961704)*x_ref[2]**o+(0.812917484286)*x_ref[2]
        ref[1,4,0,1]=(-0.25535378432)*x_ref[0]**o+(-0.0938207170929)*x_ref[0]+(0.0545321717685)*x_ref[1]**o+(-0.414414792674)*x_ref[1]+(-0.297698896712)*x_ref[2]**o+(-0.367166660571)*x_ref[2]
        ref[1,4,1,0]=(0.743408112248)*x_ref[0]**o+(0.852863814785)*x_ref[0]+(0.259483951249)*x_ref[1]**o+(-0.519288864163)*x_ref[1]+(-0.305772267727)*x_ref[2]**o+(-0.401384318883)*x_ref[2]
        ref[1,4,1,1]=(0.194889815352)*x_ref[0]**o+(0.522093986804)*x_ref[0]+(-0.343320977396)*x_ref[1]**o+(0.21840236597)*x_ref[1]+(-0.672107409136)*x_ref[2]**o+(-0.600203805002)*x_ref[2]
        ref[1,4,2,0]=(0.420860556361)*x_ref[0]**o+(0.789879869049)*x_ref[0]+(0.100987805744)*x_ref[1]**o+(0.802794815123)*x_ref[1]+(-0.69466030201)*x_ref[2]**o+(-0.746043845966)*x_ref[2]
        ref[1,4,2,1]=(-0.306621579211)*x_ref[0]**o+(-0.0676982346933)*x_ref[0]+(0.568861573715)*x_ref[1]**o+(0.0823366334578)*x_ref[1]+(0.501187255641)*x_ref[2]**o+(-0.538615164478)*x_ref[2]
        ref[2,0,0,0]=(-0.231052916706)*x_ref[0]**o+(0.127790831537)*x_ref[0]+(-0.475600435038)*x_ref[1]**o+(0.431586151473)*x_ref[1]+(-0.882365547235)*x_ref[2]**o+(0.778027701879)*x_ref[2]
        ref[2,0,0,1]=(0.0709490835498)*x_ref[0]**o+(0.442614693755)*x_ref[0]+(0.870405247692)*x_ref[1]**o+(-0.648538779515)*x_ref[1]+(0.751107652054)*x_ref[2]**o+(0.0122956931127)*x_ref[2]
        ref[2,0,1,0]=(0.703128329145)*x_ref[0]**o+(0.259775843631)*x_ref[0]+(0.311958888594)*x_ref[1]**o+(-0.0445092402277)*x_ref[1]+(0.338222457799)*x_ref[2]**o+(0.845995300456)*x_ref[2]
        ref[2,0,1,1]=(-0.653977898919)*x_ref[0]**o+(0.0495539350857)*x_ref[0]+(0.4013623408)*x_ref[1]**o+(0.490747039887)*x_ref[1]+(-0.480487636729)*x_ref[2]**o+(0.361660115775)*x_ref[2]
        ref[2,0,2,0]=(0.215778506945)*x_ref[0]**o+(0.316853875053)*x_ref[0]+(0.556770639829)*x_ref[1]**o+(-0.499729882355)*x_ref[1]+(0.692012327582)*x_ref[2]**o+(0.00681314946195)*x_ref[2]
        ref[2,0,2,1]=(0.432077857223)*x_ref[0]**o+(-0.0497579036988)*x_ref[0]+(-0.465403408565)*x_ref[1]**o+(0.125421502956)*x_ref[1]+(-0.313541879877)*x_ref[2]**o+(0.113957941731)*x_ref[2]
        ref[2,1,0,0]=(-0.528126103192)*x_ref[0]**o+(-0.141306512096)*x_ref[0]+(-0.505835313198)*x_ref[1]**o+(-0.0433171534408)*x_ref[1]+(-0.522465687628)*x_ref[2]**o+(-0.16728285732)*x_ref[2]
        ref[2,1,0,1]=(0.892019660948)*x_ref[0]**o+(0.646944130193)*x_ref[0]+(0.865495756748)*x_ref[1]**o+(0.949095276008)*x_ref[1]+(0.786394755068)*x_ref[2]**o+(0.482441081687)*x_ref[2]
        ref[2,1,1,0]=(-0.0485295711204)*x_ref[0]**o+(0.250163927911)*x_ref[0]+(0.0387861125998)*x_ref[1]**o+(-0.0789121929817)*x_ref[1]+(-0.982330126583)*x_ref[2]**o+(0.66328870814)*x_ref[2]
        ref[2,1,1,1]=(-0.494726556678)*x_ref[0]**o+(-0.961342091528)*x_ref[0]+(0.211410563083)*x_ref[1]**o+(-0.884287887286)*x_ref[1]+(-0.583401908135)*x_ref[2]**o+(-0.361423905219)*x_ref[2]
        ref[2,1,2,0]=(-0.326785246692)*x_ref[0]**o+(0.596272945325)*x_ref[0]+(0.625039539234)*x_ref[1]**o+(0.88699688986)*x_ref[1]+(-0.649803810419)*x_ref[2]**o+(0.890718271469)*x_ref[2]
        ref[2,1,2,1]=(0.702110072391)*x_ref[0]**o+(0.72009439563)*x_ref[0]+(-0.930941572511)*x_ref[1]**o+(0.955877158334)*x_ref[1]+(0.1211281332)*x_ref[2]**o+(0.658738626852)*x_ref[2]
        ref[2,2,0,0]=(-0.177402589697)*x_ref[0]**o+(0.701090894256)*x_ref[0]+(-0.00205333079339)*x_ref[1]**o+(-0.427862310859)*x_ref[1]+(0.529186374803)*x_ref[2]**o+(-0.110462035551)*x_ref[2]
        ref[2,2,0,1]=(-0.104112165367)*x_ref[0]**o+(0.043539351452)*x_ref[0]+(0.757384935306)*x_ref[1]**o+(-0.451086492844)*x_ref[1]+(0.0549627297033)*x_ref[2]**o+(-0.311718955315)*x_ref[2]
        ref[2,2,1,0]=(-0.579986090808)*x_ref[0]**o+(-0.922872917924)*x_ref[0]+(0.66422700207)*x_ref[1]**o+(0.947304780214)*x_ref[1]+(-0.0492553517704)*x_ref[2]**o+(0.817225619153)*x_ref[2]
        ref[2,2,1,1]=(0.306298476233)*x_ref[0]**o+(-0.620190960783)*x_ref[0]+(-0.684876074013)*x_ref[1]**o+(-0.371829616982)*x_ref[1]+(0.904709743321)*x_ref[2]**o+(0.170802273852)*x_ref[2]
        ref[2,2,2,0]=(0.317943399758)*x_ref[0]**o+(0.518026054212)*x_ref[0]+(-0.801462311167)*x_ref[1]**o+(0.767424157898)*x_ref[1]+(-0.547271995492)*x_ref[2]**o+(-0.912619177188)*x_ref[2]
        ref[2,2,2,1]=(-0.848711258358)*x_ref[0]**o+(0.456489097468)*x_ref[0]+(-0.310979078367)*x_ref[1]**o+(-0.946717007726)*x_ref[1]+(0.454363833646)*x_ref[2]**o+(-0.715701991077)*x_ref[2]
        ref[2,3,0,0]=(0.605493533014)*x_ref[0]**o+(0.478437880368)*x_ref[0]+(-0.672727491605)*x_ref[1]**o+(0.831259256009)*x_ref[1]+(-0.434153354239)*x_ref[2]**o+(-0.505363218578)*x_ref[2]
        ref[2,3,0,1]=(-0.0703539676643)*x_ref[0]**o+(-0.452352344215)*x_ref[0]+(0.0403568720903)*x_ref[1]**o+(0.959415939826)*x_ref[1]+(-0.784519698477)*x_ref[2]**o+(-0.691664442958)*x_ref[2]
        ref[2,3,1,0]=(0.696353768719)*x_ref[0]**o+(0.295477297618)*x_ref[0]+(0.580540484825)*x_ref[1]**o+(0.098954967023)*x_ref[1]+(-0.547594444517)*x_ref[2]**o+(-0.574783665389)*x_ref[2]
        ref[2,3,1,1]=(0.997841887408)*x_ref[0]**o+(0.857864040549)*x_ref[0]+(-0.413188343971)*x_ref[1]**o+(0.344770631687)*x_ref[1]+(-0.127257058552)*x_ref[2]**o+(0.132618819456)*x_ref[2]
        ref[2,3,2,0]=(0.174045751083)*x_ref[0]**o+(0.866724759298)*x_ref[0]+(0.0818516103455)*x_ref[1]**o+(-0.891670609599)*x_ref[1]+(0.982721717044)*x_ref[2]**o+(0.378710155856)*x_ref[2]
        ref[2,3,2,1]=(0.7760778906)*x_ref[0]**o+(-0.710104290654)*x_ref[0]+(-0.471259845977)*x_ref[1]**o+(-0.791979211724)*x_ref[1]+(-0.845652753706)*x_ref[2]**o+(-0.301795837852)*x_ref[2]
        ref[2,4,0,0]=(-0.698193658421)*x_ref[0]**o+(-0.831956135385)*x_ref[0]+(-0.442637915133)*x_ref[1]**o+(0.108715813637)*x_ref[1]+(-0.888348149694)*x_ref[2]**o+(-0.90115888496)*x_ref[2]
        ref[2,4,0,1]=(0.770237705577)*x_ref[0]**o+(0.889304763618)*x_ref[0]+(-0.139196183564)*x_ref[1]**o+(0.332516232239)*x_ref[1]+(-0.0218103253052)*x_ref[2]**o+(-0.368033928854)*x_ref[2]
        ref[2,4,1,0]=(0.675479351844)*x_ref[0]**o+(0.880998418303)*x_ref[0]+(0.408381837105)*x_ref[1]**o+(0.717402364084)*x_ref[1]+(-0.303902947422)*x_ref[2]**o+(-0.322350049962)*x_ref[2]
        ref[2,4,1,1]=(-0.480881673289)*x_ref[0]**o+(0.30290141529)*x_ref[0]+(-0.538411390886)*x_ref[1]**o+(0.260090848702)*x_ref[1]+(0.638130764808)*x_ref[2]**o+(0.756338975575)*x_ref[2]
        ref[2,4,2,0]=(-0.299233221889)*x_ref[0]**o+(-0.26471083761)*x_ref[0]+(0.717191172884)*x_ref[1]**o+(0.488553216118)*x_ref[1]+(0.718956726644)*x_ref[2]**o+(0.370889025703)*x_ref[2]
        ref[2,4,2,1]=(-0.989020272269)*x_ref[0]**o+(-0.891423884362)*x_ref[0]+(0.486153100693)*x_ref[1]**o+(-0.971171395246)*x_ref[1]+(-0.766942968184)*x_ref[2]**o+(0.728728968862)*x_ref[2]
        ref[3,0,0,0]=(0.756186246227)*x_ref[0]**o+(0.76507856268)*x_ref[0]+(0.771165000414)*x_ref[1]**o+(0.673968873856)*x_ref[1]+(0.827045846253)*x_ref[2]**o+(-0.788265752412)*x_ref[2]
        ref[3,0,0,1]=(-0.66925731439)*x_ref[0]**o+(-0.0110127596435)*x_ref[0]+(0.944388255414)*x_ref[1]**o+(-0.613832072486)*x_ref[1]+(0.0977542120724)*x_ref[2]**o+(0.100776856689)*x_ref[2]
        ref[3,0,1,0]=(0.905138713064)*x_ref[0]**o+(0.673051338153)*x_ref[0]+(0.240614253653)*x_ref[1]**o+(-0.964132330033)*x_ref[1]+(-0.114961018838)*x_ref[2]**o+(-0.438756962988)*x_ref[2]
        ref[3,0,1,1]=(-0.192259182094)*x_ref[0]**o+(0.12315590119)*x_ref[0]+(-0.907744420226)*x_ref[1]**o+(-0.483981269809)*x_ref[1]+(-0.169699431472)*x_ref[2]**o+(-0.45231635406)*x_ref[2]
        ref[3,0,2,0]=(0.895387255885)*x_ref[0]**o+(0.23892997411)*x_ref[0]+(-0.396010446618)*x_ref[1]**o+(0.94853984413)*x_ref[1]+(-0.0165259637241)*x_ref[2]**o+(-0.312558539149)*x_ref[2]
        ref[3,0,2,1]=(0.710478713361)*x_ref[0]**o+(-0.109316102216)*x_ref[0]+(-0.042711568986)*x_ref[1]**o+(-0.448508085742)*x_ref[1]+(-0.0979193346011)*x_ref[2]**o+(0.93156672211)*x_ref[2]
        ref[3,1,0,0]=(-0.21605108221)*x_ref[0]**o+(0.410466481712)*x_ref[0]+(0.799288386783)*x_ref[1]**o+(0.622866017113)*x_ref[1]+(-0.236698938159)*x_ref[2]**o+(0.246625241928)*x_ref[2]
        ref[3,1,0,1]=(0.989568527762)*x_ref[0]**o+(0.0212906378151)*x_ref[0]+(0.311279947324)*x_ref[1]**o+(-0.231412902902)*x_ref[1]+(-0.444760786142)*x_ref[2]**o+(0.908171985421)*x_ref[2]
        ref[3,1,1,0]=(0.42605234231)*x_ref[0]**o+(-0.58629332944)*x_ref[0]+(0.322063640761)*x_ref[1]**o+(0.837771339825)*x_ref[1]+(-0.911083211394)*x_ref[2]**o+(-0.0699607581992)*x_ref[2]
        ref[3,1,1,1]=(0.75449445591)*x_ref[0]**o+(0.722970800307)*x_ref[0]+(-0.582970919465)*x_ref[1]**o+(-0.868818538212)*x_ref[1]+(0.940067636181)*x_ref[2]**o+(-0.659469743617)*x_ref[2]
        ref[3,1,2,0]=(-0.391224924663)*x_ref[0]**o+(-0.768173177737)*x_ref[0]+(-0.856876347997)*x_ref[1]**o+(-0.0451180499647)*x_ref[1]+(0.198590679246)*x_ref[2]**o+(0.187329095476)*x_ref[2]
        ref[3,1,2,1]=(0.483309063092)*x_ref[0]**o+(-0.0777241026893)*x_ref[0]+(0.0438427457286)*x_ref[1]**o+(-0.73195704867)*x_ref[1]+(-0.439660745725)*x_ref[2]**o+(0.279501672861)*x_ref[2]
        ref[3,2,0,0]=(0.799935089139)*x_ref[0]**o+(-0.306860809857)*x_ref[0]+(0.169104975039)*x_ref[1]**o+(-0.3480717525)*x_ref[1]+(0.215231688)*x_ref[2]**o+(-0.528775376945)*x_ref[2]
        ref[3,2,0,1]=(0.917815789587)*x_ref[0]**o+(0.0672910338102)*x_ref[0]+(-0.186550440011)*x_ref[1]**o+(0.922579714252)*x_ref[1]+(0.924162379857)*x_ref[2]**o+(-0.769320483618)*x_ref[2]
        ref[3,2,1,0]=(0.604324468058)*x_ref[0]**o+(-0.0109731419053)*x_ref[0]+(0.229447318325)*x_ref[1]**o+(-0.480693435232)*x_ref[1]+(-0.881779256999)*x_ref[2]**o+(0.288656394771)*x_ref[2]
        ref[3,2,1,1]=(0.934680723352)*x_ref[0]**o+(0.993838528376)*x_ref[0]+(0.957781588502)*x_ref[1]**o+(-0.859837057866)*x_ref[1]+(-0.702139964493)*x_ref[2]**o+(0.0538177124609)*x_ref[2]
        ref[3,2,2,0]=(-0.142809047814)*x_ref[0]**o+(-0.472594415134)*x_ref[0]+(0.0381782768377)*x_ref[1]**o+(0.727331491769)*x_ref[1]+(0.0910992897581)*x_ref[2]**o+(-0.845316063851)*x_ref[2]
        ref[3,2,2,1]=(-0.217483358272)*x_ref[0]**o+(0.991535092285)*x_ref[0]+(-0.399175896468)*x_ref[1]**o+(-0.665836512895)*x_ref[1]+(0.0983700911493)*x_ref[2]**o+(0.321811300155)*x_ref[2]
        ref[3,3,0,0]=(0.797049585464)*x_ref[0]**o+(-0.724467959299)*x_ref[0]+(0.19971589973)*x_ref[1]**o+(0.845928461864)*x_ref[1]+(-0.365158866359)*x_ref[2]**o+(-0.698274270132)*x_ref[2]
        ref[3,3,0,1]=(-0.638474704878)*x_ref[0]**o+(0.551866433352)*x_ref[0]+(0.192309452979)*x_ref[1]**o+(-0.703649325501)*x_ref[1]+(0.823910578138)*x_ref[2]**o+(-0.917446989634)*x_ref[2]
        ref[3,3,1,0]=(0.217577282169)*x_ref[0]**o+(0.521716803668)*x_ref[0]+(-0.116587783856)*x_ref[1]**o+(0.529672391417)*x_ref[1]+(-0.798918158284)*x_ref[2]**o+(-0.0996139511955)*x_ref[2]
        ref[3,3,1,1]=(0.498965049079)*x_ref[0]**o+(0.316489921186)*x_ref[0]+(-0.0988355471817)*x_ref[1]**o+(-0.0477508287681)*x_ref[1]+(-0.778914523993)*x_ref[2]**o+(0.189397812723)*x_ref[2]
        ref[3,3,2,0]=(0.203823698152)*x_ref[0]**o+(-0.340499414266)*x_ref[0]+(0.353374693517)*x_ref[1]**o+(-0.91035264494)*x_ref[1]+(-0.779204084198)*x_ref[2]**o+(-0.00276454165248)*x_ref[2]
        ref[3,3,2,1]=(0.295232007123)*x_ref[0]**o+(0.271766544502)*x_ref[0]+(0.431956430336)*x_ref[1]**o+(0.896481311536)*x_ref[1]+(-0.325014902487)*x_ref[2]**o+(0.70836170284)*x_ref[2]
        ref[3,4,0,0]=(-0.407535202468)*x_ref[0]**o+(-0.33055681091)*x_ref[0]+(0.129762143151)*x_ref[1]**o+(0.352087436228)*x_ref[1]+(-0.538770401565)*x_ref[2]**o+(-0.519071077045)*x_ref[2]
        ref[3,4,0,1]=(0.846011435499)*x_ref[0]**o+(0.172822133672)*x_ref[0]+(-0.740732553847)*x_ref[1]**o+(0.230062837231)*x_ref[1]+(-0.231810007237)*x_ref[2]**o+(-0.568366338923)*x_ref[2]
        ref[3,4,1,0]=(-0.198091119057)*x_ref[0]**o+(0.957989508794)*x_ref[0]+(0.861739071175)*x_ref[1]**o+(0.923450697883)*x_ref[1]+(-0.755262504937)*x_ref[2]**o+(0.102897579936)*x_ref[2]
        ref[3,4,1,1]=(0.198829545353)*x_ref[0]**o+(-0.757182006818)*x_ref[0]+(0.254092337808)*x_ref[1]**o+(-0.00815072620097)*x_ref[1]+(0.347421988308)*x_ref[2]**o+(-0.0215277616923)*x_ref[2]
        ref[3,4,2,0]=(0.775062112529)*x_ref[0]**o+(-0.77746792774)*x_ref[0]+(-0.666490294869)*x_ref[1]**o+(-0.679971069878)*x_ref[1]+(-0.972450779894)*x_ref[2]**o+(-0.705322282477)*x_ref[2]
        ref[3,4,2,1]=(-0.0994927587382)*x_ref[0]**o+(0.293071774786)*x_ref[0]+(0.370784399397)*x_ref[1]**o+(0.931411225946)*x_ref[1]+(-0.55789163294)*x_ref[2]**o+(0.0831890658592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0969659653911)*x[0]**o+(-0.0778467995236)*x[0]+(-0.896520347956)*x[1]**o+(0.452579859858)*x[1]
        ref=(0.0969659653911)*x_ref[0]**o+(-0.0778467995236)*x_ref[0]+(-0.896520347956)*x_ref[1]**o+(0.452579859858)*x_ref[1]
      else:
        arg=(-0.62506591451)*x[0]**o+(0.970735975624)*x[0]+(-0.563055321362)*x[1]**o+(-0.281482915069)*x[1]+(-0.6461258967)*x[2]**o+(0.786610350609)*x[2]
        ref=(-0.62506591451)*x_ref[0]**o+(0.970735975624)*x_ref[0]+(-0.563055321362)*x_ref[1]**o+(-0.281482915069)*x_ref[1]+(-0.6461258967)*x_ref[2]**o+(0.786610350609)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.725240039321)*x[0]**o+(0.326302203613)*x[0]+(0.356072156898)*x[1]**o+(0.980066388307)*x[1]
        arg[1]=(0.171551389095)*x[0]**o+(0.379133470853)*x[0]+(0.120608077039)*x[1]**o+(-0.413701037132)*x[1]
        ref[0]=(0.725240039321)*x_ref[0]**o+(0.326302203613)*x_ref[0]+(0.356072156898)*x_ref[1]**o+(0.980066388307)*x_ref[1]
        ref[1]=(0.171551389095)*x_ref[0]**o+(0.379133470853)*x_ref[0]+(0.120608077039)*x_ref[1]**o+(-0.413701037132)*x_ref[1]
      else:
        arg[0]=(-0.656095745624)*x[0]**o+(-0.540430373724)*x[0]+(-0.109962817674)*x[1]**o+(-0.969692783268)*x[1]+(-0.377994841279)*x[2]**o+(-0.664035030664)*x[2]
        arg[1]=(0.768498465456)*x[0]**o+(0.647294499552)*x[0]+(0.824672861058)*x[1]**o+(-0.848088780836)*x[1]+(0.045558138597)*x[2]**o+(-0.209365900036)*x[2]
        ref[0]=(-0.656095745624)*x_ref[0]**o+(-0.540430373724)*x_ref[0]+(-0.109962817674)*x_ref[1]**o+(-0.969692783268)*x_ref[1]+(-0.377994841279)*x_ref[2]**o+(-0.664035030664)*x_ref[2]
        ref[1]=(0.768498465456)*x_ref[0]**o+(0.647294499552)*x_ref[0]+(0.824672861058)*x_ref[1]**o+(-0.848088780836)*x_ref[1]+(0.045558138597)*x_ref[2]**o+(-0.209365900036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.694866182164)*x[0]**o+(0.759641527396)*x[0]+(-0.696641139757)*x[1]**o+(0.617978953453)*x[1]
        arg[0,1]=(0.75058922213)*x[0]**o+(0.580832851854)*x[0]+(-0.237039031011)*x[1]**o+(0.53518450519)*x[1]
        arg[0,2]=(-0.191933019173)*x[0]**o+(-0.0638088686782)*x[0]+(0.871645336149)*x[1]**o+(0.256502442034)*x[1]
        arg[0,3]=(0.829189580591)*x[0]**o+(0.186064016504)*x[0]+(-0.204496850484)*x[1]**o+(-0.314535598074)*x[1]
        arg[0,4]=(0.170968249419)*x[0]**o+(-0.719959835105)*x[0]+(0.367000223059)*x[1]**o+(-0.210239222752)*x[1]
        arg[1,0]=(-0.424311493894)*x[0]**o+(-0.381097143579)*x[0]+(-0.54925292526)*x[1]**o+(-0.393349131169)*x[1]
        arg[1,1]=(0.148181508143)*x[0]**o+(-0.224530937841)*x[0]+(0.778748719265)*x[1]**o+(-0.522104185968)*x[1]
        arg[1,2]=(0.134819103337)*x[0]**o+(0.0388165865124)*x[0]+(-0.772318490115)*x[1]**o+(-0.928611227188)*x[1]
        arg[1,3]=(0.790807030518)*x[0]**o+(-0.849964960983)*x[0]+(-0.79407080953)*x[1]**o+(-0.626315155005)*x[1]
        arg[1,4]=(-0.998842104726)*x[0]**o+(-0.325938405818)*x[0]+(-0.81445420822)*x[1]**o+(0.868281547664)*x[1]
        arg[2,0]=(-0.975572739854)*x[0]**o+(-0.0630975707316)*x[0]+(0.0688031679656)*x[1]**o+(0.557152464527)*x[1]
        arg[2,1]=(0.389814261767)*x[0]**o+(0.523891881188)*x[0]+(-0.558388242102)*x[1]**o+(-0.45720905598)*x[1]
        arg[2,2]=(0.763327517145)*x[0]**o+(0.637954909904)*x[0]+(0.407602154355)*x[1]**o+(0.707118271445)*x[1]
        arg[2,3]=(-0.823371482965)*x[0]**o+(0.918685041464)*x[0]+(0.991991462022)*x[1]**o+(0.489892661208)*x[1]
        arg[2,4]=(-0.661567455889)*x[0]**o+(-0.00566795402284)*x[0]+(0.86705493299)*x[1]**o+(0.00158321121677)*x[1]
        arg[3,0]=(0.365456232671)*x[0]**o+(-0.587373381962)*x[0]+(0.751291631507)*x[1]**o+(-0.0605248375222)*x[1]
        arg[3,1]=(0.546630294768)*x[0]**o+(-0.903068297271)*x[0]+(0.486313468693)*x[1]**o+(0.355322343068)*x[1]
        arg[3,2]=(0.233755056849)*x[0]**o+(-0.350375425427)*x[0]+(0.0115435986629)*x[1]**o+(-0.713895312694)*x[1]
        arg[3,3]=(0.385411302842)*x[0]**o+(0.0733743700003)*x[0]+(0.439219708301)*x[1]**o+(0.129391539278)*x[1]
        arg[3,4]=(-0.194268604537)*x[0]**o+(0.175829932738)*x[0]+(0.866342013931)*x[1]**o+(-0.213891435501)*x[1]
        ref[0,0]=(0.694866182164)*x_ref[0]**o+(0.759641527396)*x_ref[0]+(-0.696641139757)*x_ref[1]**o+(0.617978953453)*x_ref[1]
        ref[0,1]=(0.75058922213)*x_ref[0]**o+(0.580832851854)*x_ref[0]+(-0.237039031011)*x_ref[1]**o+(0.53518450519)*x_ref[1]
        ref[0,2]=(-0.191933019173)*x_ref[0]**o+(-0.0638088686782)*x_ref[0]+(0.871645336149)*x_ref[1]**o+(0.256502442034)*x_ref[1]
        ref[0,3]=(0.829189580591)*x_ref[0]**o+(0.186064016504)*x_ref[0]+(-0.204496850484)*x_ref[1]**o+(-0.314535598074)*x_ref[1]
        ref[0,4]=(0.170968249419)*x_ref[0]**o+(-0.719959835105)*x_ref[0]+(0.367000223059)*x_ref[1]**o+(-0.210239222752)*x_ref[1]
        ref[1,0]=(-0.424311493894)*x_ref[0]**o+(-0.381097143579)*x_ref[0]+(-0.54925292526)*x_ref[1]**o+(-0.393349131169)*x_ref[1]
        ref[1,1]=(0.148181508143)*x_ref[0]**o+(-0.224530937841)*x_ref[0]+(0.778748719265)*x_ref[1]**o+(-0.522104185968)*x_ref[1]
        ref[1,2]=(0.134819103337)*x_ref[0]**o+(0.0388165865124)*x_ref[0]+(-0.772318490115)*x_ref[1]**o+(-0.928611227188)*x_ref[1]
        ref[1,3]=(0.790807030518)*x_ref[0]**o+(-0.849964960983)*x_ref[0]+(-0.79407080953)*x_ref[1]**o+(-0.626315155005)*x_ref[1]
        ref[1,4]=(-0.998842104726)*x_ref[0]**o+(-0.325938405818)*x_ref[0]+(-0.81445420822)*x_ref[1]**o+(0.868281547664)*x_ref[1]
        ref[2,0]=(-0.975572739854)*x_ref[0]**o+(-0.0630975707316)*x_ref[0]+(0.0688031679656)*x_ref[1]**o+(0.557152464527)*x_ref[1]
        ref[2,1]=(0.389814261767)*x_ref[0]**o+(0.523891881188)*x_ref[0]+(-0.558388242102)*x_ref[1]**o+(-0.45720905598)*x_ref[1]
        ref[2,2]=(0.763327517145)*x_ref[0]**o+(0.637954909904)*x_ref[0]+(0.407602154355)*x_ref[1]**o+(0.707118271445)*x_ref[1]
        ref[2,3]=(-0.823371482965)*x_ref[0]**o+(0.918685041464)*x_ref[0]+(0.991991462022)*x_ref[1]**o+(0.489892661208)*x_ref[1]
        ref[2,4]=(-0.661567455889)*x_ref[0]**o+(-0.00566795402284)*x_ref[0]+(0.86705493299)*x_ref[1]**o+(0.00158321121677)*x_ref[1]
        ref[3,0]=(0.365456232671)*x_ref[0]**o+(-0.587373381962)*x_ref[0]+(0.751291631507)*x_ref[1]**o+(-0.0605248375222)*x_ref[1]
        ref[3,1]=(0.546630294768)*x_ref[0]**o+(-0.903068297271)*x_ref[0]+(0.486313468693)*x_ref[1]**o+(0.355322343068)*x_ref[1]
        ref[3,2]=(0.233755056849)*x_ref[0]**o+(-0.350375425427)*x_ref[0]+(0.0115435986629)*x_ref[1]**o+(-0.713895312694)*x_ref[1]
        ref[3,3]=(0.385411302842)*x_ref[0]**o+(0.0733743700003)*x_ref[0]+(0.439219708301)*x_ref[1]**o+(0.129391539278)*x_ref[1]
        ref[3,4]=(-0.194268604537)*x_ref[0]**o+(0.175829932738)*x_ref[0]+(0.866342013931)*x_ref[1]**o+(-0.213891435501)*x_ref[1]
      else:
        arg[0,0]=(-0.479138946804)*x[0]**o+(0.240408443007)*x[0]+(0.309155796205)*x[1]**o+(-0.124360991957)*x[1]+(0.00301047542037)*x[2]**o+(-0.039439802799)*x[2]
        arg[0,1]=(0.41979402771)*x[0]**o+(0.798738902847)*x[0]+(0.619960746922)*x[1]**o+(-0.551662130664)*x[1]+(0.914293623695)*x[2]**o+(0.6859768812)*x[2]
        arg[0,2]=(0.744593492696)*x[0]**o+(-0.31743031745)*x[0]+(0.389545123971)*x[1]**o+(0.592024987288)*x[1]+(0.63456095022)*x[2]**o+(-0.746571796061)*x[2]
        arg[0,3]=(-0.632549503649)*x[0]**o+(-0.687745599598)*x[0]+(0.794176042123)*x[1]**o+(-0.971466375953)*x[1]+(-0.086939632062)*x[2]**o+(0.373736042024)*x[2]
        arg[0,4]=(-0.966879538217)*x[0]**o+(-0.609291970974)*x[0]+(0.000409196978095)*x[1]**o+(0.623934484362)*x[1]+(-0.1694749999)*x[2]**o+(0.131778917979)*x[2]
        arg[1,0]=(0.763452444611)*x[0]**o+(0.805147582411)*x[0]+(0.999704124146)*x[1]**o+(0.0635253753138)*x[1]+(-0.717889410548)*x[2]**o+(-0.840765968423)*x[2]
        arg[1,1]=(-0.26910479702)*x[0]**o+(0.594093239771)*x[0]+(-0.536150372834)*x[1]**o+(-0.768506263316)*x[1]+(-0.924510063594)*x[2]**o+(-0.950568839784)*x[2]
        arg[1,2]=(-0.452317334239)*x[0]**o+(0.186343827048)*x[0]+(0.921015823436)*x[1]**o+(-0.656809197675)*x[1]+(-0.841898648461)*x[2]**o+(0.267827136605)*x[2]
        arg[1,3]=(-0.880873336233)*x[0]**o+(-0.915350352959)*x[0]+(0.92197459758)*x[1]**o+(0.20861139921)*x[1]+(0.776743964515)*x[2]**o+(0.858569288507)*x[2]
        arg[1,4]=(0.422773256864)*x[0]**o+(0.397743860671)*x[0]+(0.726470981309)*x[1]**o+(-0.233534612019)*x[1]+(-0.350345927201)*x[2]**o+(-0.192294374344)*x[2]
        arg[2,0]=(0.0706785590906)*x[0]**o+(-0.102098736751)*x[0]+(-0.858298802695)*x[1]**o+(-0.1390708698)*x[1]+(0.221129390442)*x[2]**o+(0.770148745174)*x[2]
        arg[2,1]=(0.619559141603)*x[0]**o+(-0.969165124767)*x[0]+(0.36662839455)*x[1]**o+(0.505630485999)*x[1]+(0.170716012487)*x[2]**o+(0.355457960368)*x[2]
        arg[2,2]=(0.254098998564)*x[0]**o+(0.725662782597)*x[0]+(0.274108937869)*x[1]**o+(0.682336753679)*x[1]+(-0.858192561327)*x[2]**o+(-0.538785359078)*x[2]
        arg[2,3]=(0.25513729674)*x[0]**o+(-0.594968324608)*x[0]+(-0.306612686559)*x[1]**o+(-0.953891015684)*x[1]+(-0.00631658753287)*x[2]**o+(-0.889694464702)*x[2]
        arg[2,4]=(0.654463492789)*x[0]**o+(-0.42220702047)*x[0]+(-0.690125229151)*x[1]**o+(0.318923837754)*x[1]+(-0.691341299676)*x[2]**o+(0.824631778823)*x[2]
        arg[3,0]=(-0.576563725107)*x[0]**o+(-0.0354219245402)*x[0]+(0.650854296808)*x[1]**o+(0.930072736771)*x[1]+(0.8383657117)*x[2]**o+(0.191293272146)*x[2]
        arg[3,1]=(0.104382485594)*x[0]**o+(-0.718025390561)*x[0]+(-0.822188446684)*x[1]**o+(-0.0772502664782)*x[1]+(-0.513331680922)*x[2]**o+(0.164340503698)*x[2]
        arg[3,2]=(0.685178273481)*x[0]**o+(-0.241726920352)*x[0]+(-0.891512838953)*x[1]**o+(-0.428833037662)*x[1]+(-0.910220899771)*x[2]**o+(-0.0109246344661)*x[2]
        arg[3,3]=(-0.294769406995)*x[0]**o+(0.643502505933)*x[0]+(0.309444282121)*x[1]**o+(-0.935318923262)*x[1]+(0.242426961867)*x[2]**o+(-0.417268637067)*x[2]
        arg[3,4]=(0.0382306334841)*x[0]**o+(-0.564858278904)*x[0]+(-0.594566652327)*x[1]**o+(0.911894434818)*x[1]+(-0.314178793024)*x[2]**o+(0.302730333267)*x[2]
        ref[0,0]=(-0.479138946804)*x_ref[0]**o+(0.240408443007)*x_ref[0]+(0.309155796205)*x_ref[1]**o+(-0.124360991957)*x_ref[1]+(0.00301047542037)*x_ref[2]**o+(-0.039439802799)*x_ref[2]
        ref[0,1]=(0.41979402771)*x_ref[0]**o+(0.798738902847)*x_ref[0]+(0.619960746922)*x_ref[1]**o+(-0.551662130664)*x_ref[1]+(0.914293623695)*x_ref[2]**o+(0.6859768812)*x_ref[2]
        ref[0,2]=(0.744593492696)*x_ref[0]**o+(-0.31743031745)*x_ref[0]+(0.389545123971)*x_ref[1]**o+(0.592024987288)*x_ref[1]+(0.63456095022)*x_ref[2]**o+(-0.746571796061)*x_ref[2]
        ref[0,3]=(-0.632549503649)*x_ref[0]**o+(-0.687745599598)*x_ref[0]+(0.794176042123)*x_ref[1]**o+(-0.971466375953)*x_ref[1]+(-0.086939632062)*x_ref[2]**o+(0.373736042024)*x_ref[2]
        ref[0,4]=(-0.966879538217)*x_ref[0]**o+(-0.609291970974)*x_ref[0]+(0.000409196978095)*x_ref[1]**o+(0.623934484362)*x_ref[1]+(-0.1694749999)*x_ref[2]**o+(0.131778917979)*x_ref[2]
        ref[1,0]=(0.763452444611)*x_ref[0]**o+(0.805147582411)*x_ref[0]+(0.999704124146)*x_ref[1]**o+(0.0635253753138)*x_ref[1]+(-0.717889410548)*x_ref[2]**o+(-0.840765968423)*x_ref[2]
        ref[1,1]=(-0.26910479702)*x_ref[0]**o+(0.594093239771)*x_ref[0]+(-0.536150372834)*x_ref[1]**o+(-0.768506263316)*x_ref[1]+(-0.924510063594)*x_ref[2]**o+(-0.950568839784)*x_ref[2]
        ref[1,2]=(-0.452317334239)*x_ref[0]**o+(0.186343827048)*x_ref[0]+(0.921015823436)*x_ref[1]**o+(-0.656809197675)*x_ref[1]+(-0.841898648461)*x_ref[2]**o+(0.267827136605)*x_ref[2]
        ref[1,3]=(-0.880873336233)*x_ref[0]**o+(-0.915350352959)*x_ref[0]+(0.92197459758)*x_ref[1]**o+(0.20861139921)*x_ref[1]+(0.776743964515)*x_ref[2]**o+(0.858569288507)*x_ref[2]
        ref[1,4]=(0.422773256864)*x_ref[0]**o+(0.397743860671)*x_ref[0]+(0.726470981309)*x_ref[1]**o+(-0.233534612019)*x_ref[1]+(-0.350345927201)*x_ref[2]**o+(-0.192294374344)*x_ref[2]
        ref[2,0]=(0.0706785590906)*x_ref[0]**o+(-0.102098736751)*x_ref[0]+(-0.858298802695)*x_ref[1]**o+(-0.1390708698)*x_ref[1]+(0.221129390442)*x_ref[2]**o+(0.770148745174)*x_ref[2]
        ref[2,1]=(0.619559141603)*x_ref[0]**o+(-0.969165124767)*x_ref[0]+(0.36662839455)*x_ref[1]**o+(0.505630485999)*x_ref[1]+(0.170716012487)*x_ref[2]**o+(0.355457960368)*x_ref[2]
        ref[2,2]=(0.254098998564)*x_ref[0]**o+(0.725662782597)*x_ref[0]+(0.274108937869)*x_ref[1]**o+(0.682336753679)*x_ref[1]+(-0.858192561327)*x_ref[2]**o+(-0.538785359078)*x_ref[2]
        ref[2,3]=(0.25513729674)*x_ref[0]**o+(-0.594968324608)*x_ref[0]+(-0.306612686559)*x_ref[1]**o+(-0.953891015684)*x_ref[1]+(-0.00631658753287)*x_ref[2]**o+(-0.889694464702)*x_ref[2]
        ref[2,4]=(0.654463492789)*x_ref[0]**o+(-0.42220702047)*x_ref[0]+(-0.690125229151)*x_ref[1]**o+(0.318923837754)*x_ref[1]+(-0.691341299676)*x_ref[2]**o+(0.824631778823)*x_ref[2]
        ref[3,0]=(-0.576563725107)*x_ref[0]**o+(-0.0354219245402)*x_ref[0]+(0.650854296808)*x_ref[1]**o+(0.930072736771)*x_ref[1]+(0.8383657117)*x_ref[2]**o+(0.191293272146)*x_ref[2]
        ref[3,1]=(0.104382485594)*x_ref[0]**o+(-0.718025390561)*x_ref[0]+(-0.822188446684)*x_ref[1]**o+(-0.0772502664782)*x_ref[1]+(-0.513331680922)*x_ref[2]**o+(0.164340503698)*x_ref[2]
        ref[3,2]=(0.685178273481)*x_ref[0]**o+(-0.241726920352)*x_ref[0]+(-0.891512838953)*x_ref[1]**o+(-0.428833037662)*x_ref[1]+(-0.910220899771)*x_ref[2]**o+(-0.0109246344661)*x_ref[2]
        ref[3,3]=(-0.294769406995)*x_ref[0]**o+(0.643502505933)*x_ref[0]+(0.309444282121)*x_ref[1]**o+(-0.935318923262)*x_ref[1]+(0.242426961867)*x_ref[2]**o+(-0.417268637067)*x_ref[2]
        ref[3,4]=(0.0382306334841)*x_ref[0]**o+(-0.564858278904)*x_ref[0]+(-0.594566652327)*x_ref[1]**o+(0.911894434818)*x_ref[1]+(-0.314178793024)*x_ref[2]**o+(0.302730333267)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.298635657968)*x[0]**o+(0.917751825722)*x[0]+(-0.000157823722007)*x[1]**o+(-0.202721972906)*x[1]
        arg[0,0,1]=(0.164880328673)*x[0]**o+(-0.310233508297)*x[0]+(-0.976270095138)*x[1]**o+(-0.86689594754)*x[1]
        arg[0,1,0]=(-0.130835262631)*x[0]**o+(-0.485555484157)*x[0]+(0.055655815912)*x[1]**o+(-0.488201458657)*x[1]
        arg[0,1,1]=(-0.339982923518)*x[0]**o+(0.435464515598)*x[0]+(-0.986955506076)*x[1]**o+(0.590966000031)*x[1]
        arg[1,0,0]=(0.150280127463)*x[0]**o+(0.827183876844)*x[0]+(0.468401701763)*x[1]**o+(-0.879871376091)*x[1]
        arg[1,0,1]=(0.48434831025)*x[0]**o+(-0.623958069024)*x[0]+(-0.781911271849)*x[1]**o+(0.34398679645)*x[1]
        arg[1,1,0]=(0.124273770782)*x[0]**o+(-0.683096046651)*x[0]+(0.769636378285)*x[1]**o+(-0.471970832437)*x[1]
        arg[1,1,1]=(0.769777912195)*x[0]**o+(-0.740802020319)*x[0]+(0.659109761568)*x[1]**o+(-0.495444655614)*x[1]
        arg[2,0,0]=(0.270063777356)*x[0]**o+(0.558849439723)*x[0]+(0.128946408169)*x[1]**o+(-0.962495994445)*x[1]
        arg[2,0,1]=(-0.0403460883009)*x[0]**o+(0.084983334662)*x[0]+(0.297573700577)*x[1]**o+(0.911774703678)*x[1]
        arg[2,1,0]=(0.821680378088)*x[0]**o+(0.10823172035)*x[0]+(-0.150213295203)*x[1]**o+(0.310471151961)*x[1]
        arg[2,1,1]=(-0.407033195691)*x[0]**o+(0.716262964777)*x[0]+(-0.262475432397)*x[1]**o+(-0.234158898682)*x[1]
        arg[3,0,0]=(-0.109193787139)*x[0]**o+(0.403892847875)*x[0]+(0.794082322801)*x[1]**o+(0.669070572253)*x[1]
        arg[3,0,1]=(0.969770217228)*x[0]**o+(0.735280190411)*x[0]+(-0.339891263613)*x[1]**o+(0.859989056027)*x[1]
        arg[3,1,0]=(-0.906193851712)*x[0]**o+(-0.330178132059)*x[0]+(-0.48035114847)*x[1]**o+(-0.48050067188)*x[1]
        arg[3,1,1]=(0.922778761752)*x[0]**o+(0.899145572602)*x[0]+(-0.991922151176)*x[1]**o+(0.615744852656)*x[1]
        arg[4,0,0]=(0.457092295352)*x[0]**o+(-0.120904430848)*x[0]+(0.575186782188)*x[1]**o+(-0.685522788659)*x[1]
        arg[4,0,1]=(0.254934861907)*x[0]**o+(-0.208850206761)*x[0]+(-0.0933987467304)*x[1]**o+(0.687715419212)*x[1]
        arg[4,1,0]=(0.483890412252)*x[0]**o+(-0.192055355216)*x[0]+(0.0891595367214)*x[1]**o+(-0.835900121805)*x[1]
        arg[4,1,1]=(-0.281109655935)*x[0]**o+(0.279867178407)*x[0]+(0.843094337157)*x[1]**o+(-0.457583293166)*x[1]
        arg[5,0,0]=(0.595078633674)*x[0]**o+(-0.994652086537)*x[0]+(0.111354924733)*x[1]**o+(0.715580298167)*x[1]
        arg[5,0,1]=(0.823409068766)*x[0]**o+(0.41047089933)*x[0]+(-0.610763277348)*x[1]**o+(0.0991108428868)*x[1]
        arg[5,1,0]=(-0.0291539407811)*x[0]**o+(0.371799932333)*x[0]+(0.631702366805)*x[1]**o+(-0.858415923309)*x[1]
        arg[5,1,1]=(-0.636146064991)*x[0]**o+(0.407248062196)*x[0]+(-0.507069143933)*x[1]**o+(-0.0509951705166)*x[1]
        ref[0,0,0]=(0.298635657968)*x_ref[0]**o+(0.917751825722)*x_ref[0]+(-0.000157823722007)*x_ref[1]**o+(-0.202721972906)*x_ref[1]
        ref[0,0,1]=(0.164880328673)*x_ref[0]**o+(-0.310233508297)*x_ref[0]+(-0.976270095138)*x_ref[1]**o+(-0.86689594754)*x_ref[1]
        ref[0,1,0]=(-0.130835262631)*x_ref[0]**o+(-0.485555484157)*x_ref[0]+(0.055655815912)*x_ref[1]**o+(-0.488201458657)*x_ref[1]
        ref[0,1,1]=(-0.339982923518)*x_ref[0]**o+(0.435464515598)*x_ref[0]+(-0.986955506076)*x_ref[1]**o+(0.590966000031)*x_ref[1]
        ref[1,0,0]=(0.150280127463)*x_ref[0]**o+(0.827183876844)*x_ref[0]+(0.468401701763)*x_ref[1]**o+(-0.879871376091)*x_ref[1]
        ref[1,0,1]=(0.48434831025)*x_ref[0]**o+(-0.623958069024)*x_ref[0]+(-0.781911271849)*x_ref[1]**o+(0.34398679645)*x_ref[1]
        ref[1,1,0]=(0.124273770782)*x_ref[0]**o+(-0.683096046651)*x_ref[0]+(0.769636378285)*x_ref[1]**o+(-0.471970832437)*x_ref[1]
        ref[1,1,1]=(0.769777912195)*x_ref[0]**o+(-0.740802020319)*x_ref[0]+(0.659109761568)*x_ref[1]**o+(-0.495444655614)*x_ref[1]
        ref[2,0,0]=(0.270063777356)*x_ref[0]**o+(0.558849439723)*x_ref[0]+(0.128946408169)*x_ref[1]**o+(-0.962495994445)*x_ref[1]
        ref[2,0,1]=(-0.0403460883009)*x_ref[0]**o+(0.084983334662)*x_ref[0]+(0.297573700577)*x_ref[1]**o+(0.911774703678)*x_ref[1]
        ref[2,1,0]=(0.821680378088)*x_ref[0]**o+(0.10823172035)*x_ref[0]+(-0.150213295203)*x_ref[1]**o+(0.310471151961)*x_ref[1]
        ref[2,1,1]=(-0.407033195691)*x_ref[0]**o+(0.716262964777)*x_ref[0]+(-0.262475432397)*x_ref[1]**o+(-0.234158898682)*x_ref[1]
        ref[3,0,0]=(-0.109193787139)*x_ref[0]**o+(0.403892847875)*x_ref[0]+(0.794082322801)*x_ref[1]**o+(0.669070572253)*x_ref[1]
        ref[3,0,1]=(0.969770217228)*x_ref[0]**o+(0.735280190411)*x_ref[0]+(-0.339891263613)*x_ref[1]**o+(0.859989056027)*x_ref[1]
        ref[3,1,0]=(-0.906193851712)*x_ref[0]**o+(-0.330178132059)*x_ref[0]+(-0.48035114847)*x_ref[1]**o+(-0.48050067188)*x_ref[1]
        ref[3,1,1]=(0.922778761752)*x_ref[0]**o+(0.899145572602)*x_ref[0]+(-0.991922151176)*x_ref[1]**o+(0.615744852656)*x_ref[1]
        ref[4,0,0]=(0.457092295352)*x_ref[0]**o+(-0.120904430848)*x_ref[0]+(0.575186782188)*x_ref[1]**o+(-0.685522788659)*x_ref[1]
        ref[4,0,1]=(0.254934861907)*x_ref[0]**o+(-0.208850206761)*x_ref[0]+(-0.0933987467304)*x_ref[1]**o+(0.687715419212)*x_ref[1]
        ref[4,1,0]=(0.483890412252)*x_ref[0]**o+(-0.192055355216)*x_ref[0]+(0.0891595367214)*x_ref[1]**o+(-0.835900121805)*x_ref[1]
        ref[4,1,1]=(-0.281109655935)*x_ref[0]**o+(0.279867178407)*x_ref[0]+(0.843094337157)*x_ref[1]**o+(-0.457583293166)*x_ref[1]
        ref[5,0,0]=(0.595078633674)*x_ref[0]**o+(-0.994652086537)*x_ref[0]+(0.111354924733)*x_ref[1]**o+(0.715580298167)*x_ref[1]
        ref[5,0,1]=(0.823409068766)*x_ref[0]**o+(0.41047089933)*x_ref[0]+(-0.610763277348)*x_ref[1]**o+(0.0991108428868)*x_ref[1]
        ref[5,1,0]=(-0.0291539407811)*x_ref[0]**o+(0.371799932333)*x_ref[0]+(0.631702366805)*x_ref[1]**o+(-0.858415923309)*x_ref[1]
        ref[5,1,1]=(-0.636146064991)*x_ref[0]**o+(0.407248062196)*x_ref[0]+(-0.507069143933)*x_ref[1]**o+(-0.0509951705166)*x_ref[1]
      else:
        arg[0,0,0]=(-0.662043178282)*x[0]**o+(0.600130780119)*x[0]+(-0.52233323628)*x[1]**o+(-0.463302513654)*x[1]+(0.714996279403)*x[2]**o+(0.113219137203)*x[2]
        arg[0,0,1]=(0.94317347224)*x[0]**o+(0.987584533548)*x[0]+(0.116577593127)*x[1]**o+(-0.515267022711)*x[1]+(0.856578953903)*x[2]**o+(0.70372338917)*x[2]
        arg[0,1,0]=(-0.405265998966)*x[0]**o+(-0.656404120416)*x[0]+(0.270655604701)*x[1]**o+(0.0638534875176)*x[1]+(0.816834973788)*x[2]**o+(0.918436000225)*x[2]
        arg[0,1,1]=(-0.37299950332)*x[0]**o+(-0.0367737761004)*x[0]+(0.176531372245)*x[1]**o+(0.849773143988)*x[1]+(-0.572911529869)*x[2]**o+(-0.140163625628)*x[2]
        arg[1,0,0]=(0.263616091516)*x[0]**o+(-0.65550051998)*x[0]+(-0.118310211881)*x[1]**o+(-0.695263983587)*x[1]+(-0.49011649182)*x[2]**o+(-0.553284002669)*x[2]
        arg[1,0,1]=(0.859640205193)*x[0]**o+(-0.569681205214)*x[0]+(0.459979003108)*x[1]**o+(0.672561515814)*x[1]+(-0.854289983441)*x[2]**o+(-0.445344129294)*x[2]
        arg[1,1,0]=(-0.904355809101)*x[0]**o+(0.540513356972)*x[0]+(-0.377422260181)*x[1]**o+(0.937285151058)*x[1]+(0.148943597435)*x[2]**o+(0.992256197076)*x[2]
        arg[1,1,1]=(-0.868336238952)*x[0]**o+(-0.713337888363)*x[0]+(0.800427728006)*x[1]**o+(-0.83502690568)*x[1]+(0.365707922084)*x[2]**o+(0.218740577318)*x[2]
        arg[2,0,0]=(-0.0382672594668)*x[0]**o+(0.463831755073)*x[0]+(-0.104957678792)*x[1]**o+(-0.561738607653)*x[1]+(-0.660128314406)*x[2]**o+(0.594649133189)*x[2]
        arg[2,0,1]=(-0.399297921751)*x[0]**o+(0.550876554224)*x[0]+(0.150599666563)*x[1]**o+(-0.748400310268)*x[1]+(0.658318134651)*x[2]**o+(0.49075091442)*x[2]
        arg[2,1,0]=(-0.393769876286)*x[0]**o+(0.328125610064)*x[0]+(-0.703385111404)*x[1]**o+(0.0418077854723)*x[1]+(0.918774663106)*x[2]**o+(-0.119085951508)*x[2]
        arg[2,1,1]=(-0.549250924758)*x[0]**o+(0.169148558793)*x[0]+(0.664162931811)*x[1]**o+(0.207136816188)*x[1]+(-0.17207817969)*x[2]**o+(0.34258020774)*x[2]
        arg[3,0,0]=(0.446467472517)*x[0]**o+(-0.358128380843)*x[0]+(0.146916048359)*x[1]**o+(-0.606692124285)*x[1]+(0.828655704727)*x[2]**o+(0.067139299396)*x[2]
        arg[3,0,1]=(-0.357052242351)*x[0]**o+(-0.528255164843)*x[0]+(0.140204673142)*x[1]**o+(0.674282235672)*x[1]+(-0.469118454039)*x[2]**o+(0.388103121466)*x[2]
        arg[3,1,0]=(-0.359629791333)*x[0]**o+(-0.625108929158)*x[0]+(0.855034791748)*x[1]**o+(-0.343451063341)*x[1]+(-0.681557972105)*x[2]**o+(0.576310870238)*x[2]
        arg[3,1,1]=(0.00546530278132)*x[0]**o+(0.359221089344)*x[0]+(-0.57416247885)*x[1]**o+(-0.521306138323)*x[1]+(-0.618281954095)*x[2]**o+(-0.862162207352)*x[2]
        arg[4,0,0]=(-0.196356423488)*x[0]**o+(-0.871482508533)*x[0]+(-0.367390886081)*x[1]**o+(0.580310108591)*x[1]+(0.857970671482)*x[2]**o+(0.123685673359)*x[2]
        arg[4,0,1]=(0.715104674022)*x[0]**o+(-0.00444438628932)*x[0]+(0.538883210054)*x[1]**o+(0.179340705463)*x[1]+(-0.605249436909)*x[2]**o+(-0.744253826562)*x[2]
        arg[4,1,0]=(-0.315453893647)*x[0]**o+(-0.560946546082)*x[0]+(-0.674448928458)*x[1]**o+(-0.510243444449)*x[1]+(0.592615876779)*x[2]**o+(0.218024801263)*x[2]
        arg[4,1,1]=(-0.46902473326)*x[0]**o+(-0.573926691452)*x[0]+(0.827898114814)*x[1]**o+(-0.644208082857)*x[1]+(-0.956169254426)*x[2]**o+(-0.297101737561)*x[2]
        arg[5,0,0]=(0.45818164662)*x[0]**o+(0.219667021464)*x[0]+(-0.453125753418)*x[1]**o+(-0.201214929832)*x[1]+(0.135363067088)*x[2]**o+(0.850636247323)*x[2]
        arg[5,0,1]=(0.296120304926)*x[0]**o+(-0.139971296351)*x[0]+(0.448115780401)*x[1]**o+(-0.300602343399)*x[1]+(-0.165568753274)*x[2]**o+(0.984788641639)*x[2]
        arg[5,1,0]=(0.321875942152)*x[0]**o+(-0.403122417353)*x[0]+(-0.148664639927)*x[1]**o+(0.644143662303)*x[1]+(-0.718311175542)*x[2]**o+(-0.850248996913)*x[2]
        arg[5,1,1]=(0.44869757406)*x[0]**o+(0.0292094503086)*x[0]+(-0.159644215279)*x[1]**o+(0.695976060359)*x[1]+(-0.221444458713)*x[2]**o+(-0.42974111206)*x[2]
        ref[0,0,0]=(-0.662043178282)*x_ref[0]**o+(0.600130780119)*x_ref[0]+(-0.52233323628)*x_ref[1]**o+(-0.463302513654)*x_ref[1]+(0.714996279403)*x_ref[2]**o+(0.113219137203)*x_ref[2]
        ref[0,0,1]=(0.94317347224)*x_ref[0]**o+(0.987584533548)*x_ref[0]+(0.116577593127)*x_ref[1]**o+(-0.515267022711)*x_ref[1]+(0.856578953903)*x_ref[2]**o+(0.70372338917)*x_ref[2]
        ref[0,1,0]=(-0.405265998966)*x_ref[0]**o+(-0.656404120416)*x_ref[0]+(0.270655604701)*x_ref[1]**o+(0.0638534875176)*x_ref[1]+(0.816834973788)*x_ref[2]**o+(0.918436000225)*x_ref[2]
        ref[0,1,1]=(-0.37299950332)*x_ref[0]**o+(-0.0367737761004)*x_ref[0]+(0.176531372245)*x_ref[1]**o+(0.849773143988)*x_ref[1]+(-0.572911529869)*x_ref[2]**o+(-0.140163625628)*x_ref[2]
        ref[1,0,0]=(0.263616091516)*x_ref[0]**o+(-0.65550051998)*x_ref[0]+(-0.118310211881)*x_ref[1]**o+(-0.695263983587)*x_ref[1]+(-0.49011649182)*x_ref[2]**o+(-0.553284002669)*x_ref[2]
        ref[1,0,1]=(0.859640205193)*x_ref[0]**o+(-0.569681205214)*x_ref[0]+(0.459979003108)*x_ref[1]**o+(0.672561515814)*x_ref[1]+(-0.854289983441)*x_ref[2]**o+(-0.445344129294)*x_ref[2]
        ref[1,1,0]=(-0.904355809101)*x_ref[0]**o+(0.540513356972)*x_ref[0]+(-0.377422260181)*x_ref[1]**o+(0.937285151058)*x_ref[1]+(0.148943597435)*x_ref[2]**o+(0.992256197076)*x_ref[2]
        ref[1,1,1]=(-0.868336238952)*x_ref[0]**o+(-0.713337888363)*x_ref[0]+(0.800427728006)*x_ref[1]**o+(-0.83502690568)*x_ref[1]+(0.365707922084)*x_ref[2]**o+(0.218740577318)*x_ref[2]
        ref[2,0,0]=(-0.0382672594668)*x_ref[0]**o+(0.463831755073)*x_ref[0]+(-0.104957678792)*x_ref[1]**o+(-0.561738607653)*x_ref[1]+(-0.660128314406)*x_ref[2]**o+(0.594649133189)*x_ref[2]
        ref[2,0,1]=(-0.399297921751)*x_ref[0]**o+(0.550876554224)*x_ref[0]+(0.150599666563)*x_ref[1]**o+(-0.748400310268)*x_ref[1]+(0.658318134651)*x_ref[2]**o+(0.49075091442)*x_ref[2]
        ref[2,1,0]=(-0.393769876286)*x_ref[0]**o+(0.328125610064)*x_ref[0]+(-0.703385111404)*x_ref[1]**o+(0.0418077854723)*x_ref[1]+(0.918774663106)*x_ref[2]**o+(-0.119085951508)*x_ref[2]
        ref[2,1,1]=(-0.549250924758)*x_ref[0]**o+(0.169148558793)*x_ref[0]+(0.664162931811)*x_ref[1]**o+(0.207136816188)*x_ref[1]+(-0.17207817969)*x_ref[2]**o+(0.34258020774)*x_ref[2]
        ref[3,0,0]=(0.446467472517)*x_ref[0]**o+(-0.358128380843)*x_ref[0]+(0.146916048359)*x_ref[1]**o+(-0.606692124285)*x_ref[1]+(0.828655704727)*x_ref[2]**o+(0.067139299396)*x_ref[2]
        ref[3,0,1]=(-0.357052242351)*x_ref[0]**o+(-0.528255164843)*x_ref[0]+(0.140204673142)*x_ref[1]**o+(0.674282235672)*x_ref[1]+(-0.469118454039)*x_ref[2]**o+(0.388103121466)*x_ref[2]
        ref[3,1,0]=(-0.359629791333)*x_ref[0]**o+(-0.625108929158)*x_ref[0]+(0.855034791748)*x_ref[1]**o+(-0.343451063341)*x_ref[1]+(-0.681557972105)*x_ref[2]**o+(0.576310870238)*x_ref[2]
        ref[3,1,1]=(0.00546530278132)*x_ref[0]**o+(0.359221089344)*x_ref[0]+(-0.57416247885)*x_ref[1]**o+(-0.521306138323)*x_ref[1]+(-0.618281954095)*x_ref[2]**o+(-0.862162207352)*x_ref[2]
        ref[4,0,0]=(-0.196356423488)*x_ref[0]**o+(-0.871482508533)*x_ref[0]+(-0.367390886081)*x_ref[1]**o+(0.580310108591)*x_ref[1]+(0.857970671482)*x_ref[2]**o+(0.123685673359)*x_ref[2]
        ref[4,0,1]=(0.715104674022)*x_ref[0]**o+(-0.00444438628932)*x_ref[0]+(0.538883210054)*x_ref[1]**o+(0.179340705463)*x_ref[1]+(-0.605249436909)*x_ref[2]**o+(-0.744253826562)*x_ref[2]
        ref[4,1,0]=(-0.315453893647)*x_ref[0]**o+(-0.560946546082)*x_ref[0]+(-0.674448928458)*x_ref[1]**o+(-0.510243444449)*x_ref[1]+(0.592615876779)*x_ref[2]**o+(0.218024801263)*x_ref[2]
        ref[4,1,1]=(-0.46902473326)*x_ref[0]**o+(-0.573926691452)*x_ref[0]+(0.827898114814)*x_ref[1]**o+(-0.644208082857)*x_ref[1]+(-0.956169254426)*x_ref[2]**o+(-0.297101737561)*x_ref[2]
        ref[5,0,0]=(0.45818164662)*x_ref[0]**o+(0.219667021464)*x_ref[0]+(-0.453125753418)*x_ref[1]**o+(-0.201214929832)*x_ref[1]+(0.135363067088)*x_ref[2]**o+(0.850636247323)*x_ref[2]
        ref[5,0,1]=(0.296120304926)*x_ref[0]**o+(-0.139971296351)*x_ref[0]+(0.448115780401)*x_ref[1]**o+(-0.300602343399)*x_ref[1]+(-0.165568753274)*x_ref[2]**o+(0.984788641639)*x_ref[2]
        ref[5,1,0]=(0.321875942152)*x_ref[0]**o+(-0.403122417353)*x_ref[0]+(-0.148664639927)*x_ref[1]**o+(0.644143662303)*x_ref[1]+(-0.718311175542)*x_ref[2]**o+(-0.850248996913)*x_ref[2]
        ref[5,1,1]=(0.44869757406)*x_ref[0]**o+(0.0292094503086)*x_ref[0]+(-0.159644215279)*x_ref[1]**o+(0.695976060359)*x_ref[1]+(-0.221444458713)*x_ref[2]**o+(-0.42974111206)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0737428272134)*x[0]**o+(0.841780036645)*x[0]+(-0.110824257531)*x[1]**o+(-0.957726444468)*x[1]
        arg[0,0,0,1]=(-0.934840410671)*x[0]**o+(-0.83301064321)*x[0]+(-0.352287650266)*x[1]**o+(-0.961111564077)*x[1]
        arg[0,0,1,0]=(-0.829500940226)*x[0]**o+(0.644171710931)*x[0]+(0.232545739722)*x[1]**o+(0.584336012862)*x[1]
        arg[0,0,1,1]=(0.426298830976)*x[0]**o+(-0.615281197268)*x[0]+(0.206334026192)*x[1]**o+(-0.443698947044)*x[1]
        arg[0,0,2,0]=(-0.236495891617)*x[0]**o+(0.0883376851179)*x[0]+(-0.176541434356)*x[1]**o+(0.297175301617)*x[1]
        arg[0,0,2,1]=(-0.879955324373)*x[0]**o+(-0.244778913771)*x[0]+(0.244131188594)*x[1]**o+(-0.334739112198)*x[1]
        arg[0,1,0,0]=(-0.64887128006)*x[0]**o+(-0.489951454377)*x[0]+(-0.328402433668)*x[1]**o+(0.229814237117)*x[1]
        arg[0,1,0,1]=(-0.339451656097)*x[0]**o+(-0.409320682304)*x[0]+(0.216482338677)*x[1]**o+(-0.10594373055)*x[1]
        arg[0,1,1,0]=(-0.350538479425)*x[0]**o+(0.466450046985)*x[0]+(-0.779124795292)*x[1]**o+(-0.654562596704)*x[1]
        arg[0,1,1,1]=(0.00350092879868)*x[0]**o+(-0.214643088647)*x[0]+(0.786934081411)*x[1]**o+(-0.296853380043)*x[1]
        arg[0,1,2,0]=(-0.640183955648)*x[0]**o+(0.568647968237)*x[0]+(0.575533476901)*x[1]**o+(0.854540594085)*x[1]
        arg[0,1,2,1]=(0.633674899518)*x[0]**o+(0.829255679443)*x[0]+(0.537888424303)*x[1]**o+(-0.495082587617)*x[1]
        arg[0,2,0,0]=(-0.700436798342)*x[0]**o+(-0.732581859026)*x[0]+(-0.508905353053)*x[1]**o+(0.664601151676)*x[1]
        arg[0,2,0,1]=(0.720335764302)*x[0]**o+(0.941159917988)*x[0]+(0.2817440547)*x[1]**o+(0.553712160058)*x[1]
        arg[0,2,1,0]=(-0.937114312262)*x[0]**o+(-0.115517396201)*x[0]+(-0.672697125905)*x[1]**o+(-0.221320095369)*x[1]
        arg[0,2,1,1]=(-0.52059341062)*x[0]**o+(-0.653834647475)*x[0]+(-0.0851100861135)*x[1]**o+(-0.668300210067)*x[1]
        arg[0,2,2,0]=(-0.943613035509)*x[0]**o+(-0.294843145822)*x[0]+(0.496278748831)*x[1]**o+(0.901154419425)*x[1]
        arg[0,2,2,1]=(-0.421698818203)*x[0]**o+(-0.721918177544)*x[0]+(-0.197855702215)*x[1]**o+(0.50882827715)*x[1]
        arg[0,3,0,0]=(0.931990780298)*x[0]**o+(-0.550332606951)*x[0]+(0.401584654642)*x[1]**o+(0.0965259747601)*x[1]
        arg[0,3,0,1]=(0.881357708445)*x[0]**o+(0.564624034197)*x[0]+(-0.458191897924)*x[1]**o+(-0.427701112807)*x[1]
        arg[0,3,1,0]=(0.624777088081)*x[0]**o+(0.862654721208)*x[0]+(-0.707618093268)*x[1]**o+(0.743304953313)*x[1]
        arg[0,3,1,1]=(0.4510836974)*x[0]**o+(-0.357023031453)*x[0]+(-0.00503924010273)*x[1]**o+(0.545024336011)*x[1]
        arg[0,3,2,0]=(0.109003851094)*x[0]**o+(0.168659376881)*x[0]+(-0.0578953780655)*x[1]**o+(-0.481212462712)*x[1]
        arg[0,3,2,1]=(-0.191183501636)*x[0]**o+(-0.316145516377)*x[0]+(0.402460103016)*x[1]**o+(-0.163612464438)*x[1]
        arg[0,4,0,0]=(0.489815345033)*x[0]**o+(0.999138280765)*x[0]+(-0.626627766812)*x[1]**o+(-0.339653992568)*x[1]
        arg[0,4,0,1]=(0.616350228791)*x[0]**o+(0.933446438765)*x[0]+(0.944244035512)*x[1]**o+(0.829407789353)*x[1]
        arg[0,4,1,0]=(0.619918547863)*x[0]**o+(0.995488719729)*x[0]+(-0.78847429581)*x[1]**o+(0.542645674462)*x[1]
        arg[0,4,1,1]=(-0.149244819992)*x[0]**o+(0.921766299925)*x[0]+(0.435255339567)*x[1]**o+(-0.0596561300741)*x[1]
        arg[0,4,2,0]=(-0.816364336598)*x[0]**o+(0.701160246658)*x[0]+(-0.616385505652)*x[1]**o+(-0.558625482105)*x[1]
        arg[0,4,2,1]=(-0.36944364693)*x[0]**o+(0.243426977623)*x[0]+(-0.61578435391)*x[1]**o+(0.932570326139)*x[1]
        arg[1,0,0,0]=(-0.36839962807)*x[0]**o+(0.590534439534)*x[0]+(-0.615158899289)*x[1]**o+(0.973649158296)*x[1]
        arg[1,0,0,1]=(0.761515772771)*x[0]**o+(0.706406307012)*x[0]+(0.328712319043)*x[1]**o+(0.438600140517)*x[1]
        arg[1,0,1,0]=(0.895080579207)*x[0]**o+(-0.166789328225)*x[0]+(0.807540897371)*x[1]**o+(0.0345522692723)*x[1]
        arg[1,0,1,1]=(-0.460062661172)*x[0]**o+(0.505167337562)*x[0]+(0.25155527144)*x[1]**o+(0.601093887451)*x[1]
        arg[1,0,2,0]=(0.382129818527)*x[0]**o+(0.694676536542)*x[0]+(0.702794119083)*x[1]**o+(0.704094013056)*x[1]
        arg[1,0,2,1]=(0.265476982913)*x[0]**o+(-0.134338744881)*x[0]+(-0.0938910335356)*x[1]**o+(-0.560073316097)*x[1]
        arg[1,1,0,0]=(0.977368328792)*x[0]**o+(0.084482890741)*x[0]+(0.57068389287)*x[1]**o+(-0.242087071526)*x[1]
        arg[1,1,0,1]=(-0.36067775897)*x[0]**o+(0.768187470286)*x[0]+(-0.881150642619)*x[1]**o+(-0.15839498619)*x[1]
        arg[1,1,1,0]=(-0.808263913645)*x[0]**o+(0.788403321664)*x[0]+(0.976206012343)*x[1]**o+(-0.636933353579)*x[1]
        arg[1,1,1,1]=(0.699185445546)*x[0]**o+(0.251811206876)*x[0]+(-0.146320762812)*x[1]**o+(-0.248033048423)*x[1]
        arg[1,1,2,0]=(-0.782940880812)*x[0]**o+(0.789401686528)*x[0]+(-0.806883904562)*x[1]**o+(0.83689696772)*x[1]
        arg[1,1,2,1]=(0.652404993342)*x[0]**o+(-0.135318598212)*x[0]+(0.138928760301)*x[1]**o+(0.675593554805)*x[1]
        arg[1,2,0,0]=(0.471746379585)*x[0]**o+(-0.989783415847)*x[0]+(-0.0935052296418)*x[1]**o+(-0.532245102898)*x[1]
        arg[1,2,0,1]=(-0.44541059979)*x[0]**o+(0.593734266824)*x[0]+(-0.616647011751)*x[1]**o+(0.461140297024)*x[1]
        arg[1,2,1,0]=(-0.9144210459)*x[0]**o+(0.116806032197)*x[0]+(0.935610395306)*x[1]**o+(0.550884822213)*x[1]
        arg[1,2,1,1]=(-0.962444455944)*x[0]**o+(-0.654328013652)*x[0]+(-0.545330994236)*x[1]**o+(-0.919705456162)*x[1]
        arg[1,2,2,0]=(-0.724491374462)*x[0]**o+(-0.820402375498)*x[0]+(-0.953956431336)*x[1]**o+(-0.201450523182)*x[1]
        arg[1,2,2,1]=(0.723637497881)*x[0]**o+(-0.144483238523)*x[0]+(0.00379745988102)*x[1]**o+(-0.838616087543)*x[1]
        arg[1,3,0,0]=(-0.962256046042)*x[0]**o+(-0.453580322779)*x[0]+(0.887446007424)*x[1]**o+(-0.404454768131)*x[1]
        arg[1,3,0,1]=(0.829189872775)*x[0]**o+(0.327985850399)*x[0]+(0.113600094307)*x[1]**o+(-0.338775377551)*x[1]
        arg[1,3,1,0]=(0.912320027833)*x[0]**o+(0.0658152563951)*x[0]+(-0.843491517568)*x[1]**o+(0.858344434434)*x[1]
        arg[1,3,1,1]=(-0.133918678317)*x[0]**o+(0.644272729149)*x[0]+(0.140663927965)*x[1]**o+(0.853385355163)*x[1]
        arg[1,3,2,0]=(0.37733357322)*x[0]**o+(0.255105934063)*x[0]+(0.88881892183)*x[1]**o+(-0.722661948279)*x[1]
        arg[1,3,2,1]=(-0.2827466345)*x[0]**o+(-0.518815029512)*x[0]+(-0.995043824929)*x[1]**o+(0.973473354481)*x[1]
        arg[1,4,0,0]=(0.044904732821)*x[0]**o+(-0.0373604593083)*x[0]+(0.184254626748)*x[1]**o+(0.031832590233)*x[1]
        arg[1,4,0,1]=(0.388699053911)*x[0]**o+(-0.0619606394414)*x[0]+(0.477947819745)*x[1]**o+(0.982909216221)*x[1]
        arg[1,4,1,0]=(0.854348679973)*x[0]**o+(-0.822052612379)*x[0]+(-0.950305089968)*x[1]**o+(0.0100277256384)*x[1]
        arg[1,4,1,1]=(0.265075277179)*x[0]**o+(0.489846993407)*x[0]+(-0.961111499722)*x[1]**o+(-0.0457909055602)*x[1]
        arg[1,4,2,0]=(0.701951191951)*x[0]**o+(0.675232620259)*x[0]+(-0.141848599227)*x[1]**o+(0.737150257564)*x[1]
        arg[1,4,2,1]=(-0.450684564415)*x[0]**o+(0.854849922021)*x[0]+(0.729670746723)*x[1]**o+(0.530372628879)*x[1]
        arg[2,0,0,0]=(-0.131153262675)*x[0]**o+(-0.477683407979)*x[0]+(0.231416752516)*x[1]**o+(0.790897064402)*x[1]
        arg[2,0,0,1]=(0.0956557004382)*x[0]**o+(0.209084913471)*x[0]+(-0.328544695733)*x[1]**o+(0.701245770328)*x[1]
        arg[2,0,1,0]=(-0.476403718361)*x[0]**o+(0.689023533158)*x[0]+(0.108732189071)*x[1]**o+(0.243379977453)*x[1]
        arg[2,0,1,1]=(0.414917801157)*x[0]**o+(-0.843544596785)*x[0]+(0.772236792935)*x[1]**o+(0.836028831232)*x[1]
        arg[2,0,2,0]=(0.754970600617)*x[0]**o+(-0.422318891122)*x[0]+(0.770767911534)*x[1]**o+(-0.446737894095)*x[1]
        arg[2,0,2,1]=(0.0139989331471)*x[0]**o+(0.262286042291)*x[0]+(-0.440109329114)*x[1]**o+(-0.844965114722)*x[1]
        arg[2,1,0,0]=(-0.758977153009)*x[0]**o+(-0.991192011039)*x[0]+(0.642267239475)*x[1]**o+(-0.701360491163)*x[1]
        arg[2,1,0,1]=(-0.123019007509)*x[0]**o+(-0.622323252979)*x[0]+(0.441364031664)*x[1]**o+(0.311433222821)*x[1]
        arg[2,1,1,0]=(-0.0446644182928)*x[0]**o+(0.614952704493)*x[0]+(-0.228284494629)*x[1]**o+(0.077603612991)*x[1]
        arg[2,1,1,1]=(0.712924108225)*x[0]**o+(-0.153966300537)*x[0]+(-0.584450014127)*x[1]**o+(0.476329149833)*x[1]
        arg[2,1,2,0]=(0.554026516037)*x[0]**o+(0.730575821847)*x[0]+(-0.247529915107)*x[1]**o+(0.210060454223)*x[1]
        arg[2,1,2,1]=(0.877017208752)*x[0]**o+(-0.800916200997)*x[0]+(0.0996710064295)*x[1]**o+(-0.0890375735808)*x[1]
        arg[2,2,0,0]=(-0.198026186062)*x[0]**o+(-0.416160224681)*x[0]+(-0.626491103997)*x[1]**o+(-0.950177752043)*x[1]
        arg[2,2,0,1]=(-0.401319234479)*x[0]**o+(0.948120359627)*x[0]+(0.420351950404)*x[1]**o+(-0.242850312594)*x[1]
        arg[2,2,1,0]=(-0.239652069063)*x[0]**o+(-0.894811076801)*x[0]+(0.979721413509)*x[1]**o+(-0.712493893566)*x[1]
        arg[2,2,1,1]=(0.239416494209)*x[0]**o+(0.449047649538)*x[0]+(0.0503751073583)*x[1]**o+(0.739949331521)*x[1]
        arg[2,2,2,0]=(-0.318067701034)*x[0]**o+(-0.00208209268356)*x[0]+(-0.513821624843)*x[1]**o+(-0.202688815501)*x[1]
        arg[2,2,2,1]=(-0.194111524544)*x[0]**o+(-0.843213241929)*x[0]+(0.124223747087)*x[1]**o+(-0.988305059412)*x[1]
        arg[2,3,0,0]=(0.450894655873)*x[0]**o+(-0.809351657482)*x[0]+(-0.80555128185)*x[1]**o+(-0.922123806508)*x[1]
        arg[2,3,0,1]=(-0.462822602909)*x[0]**o+(0.504388766027)*x[0]+(-0.152725064915)*x[1]**o+(0.988745142638)*x[1]
        arg[2,3,1,0]=(-0.717873944493)*x[0]**o+(0.987011714613)*x[0]+(0.908500737929)*x[1]**o+(0.612497707425)*x[1]
        arg[2,3,1,1]=(0.221645363399)*x[0]**o+(0.219147687681)*x[0]+(-0.195831118343)*x[1]**o+(-0.59541920919)*x[1]
        arg[2,3,2,0]=(-0.289487342206)*x[0]**o+(-0.96904981397)*x[0]+(-0.844880792969)*x[1]**o+(0.803145388038)*x[1]
        arg[2,3,2,1]=(-0.0234295826132)*x[0]**o+(-0.839069887016)*x[0]+(0.591231341444)*x[1]**o+(-0.678605318787)*x[1]
        arg[2,4,0,0]=(0.717943320641)*x[0]**o+(0.625728011485)*x[0]+(0.456836257053)*x[1]**o+(-0.00523685585287)*x[1]
        arg[2,4,0,1]=(-0.549526374185)*x[0]**o+(0.57441917876)*x[0]+(-0.913459106781)*x[1]**o+(-0.610369375195)*x[1]
        arg[2,4,1,0]=(-0.27166563459)*x[0]**o+(0.146710581194)*x[0]+(0.297934124718)*x[1]**o+(0.989561209048)*x[1]
        arg[2,4,1,1]=(-0.0973858709778)*x[0]**o+(-0.909133394381)*x[0]+(0.634182571287)*x[1]**o+(0.477163052831)*x[1]
        arg[2,4,2,0]=(0.0878998038301)*x[0]**o+(-0.90568487021)*x[0]+(0.798978355456)*x[1]**o+(0.977402478037)*x[1]
        arg[2,4,2,1]=(-0.456868538667)*x[0]**o+(0.0723272702119)*x[0]+(0.143461337462)*x[1]**o+(-0.923713568343)*x[1]
        arg[3,0,0,0]=(-0.154410689788)*x[0]**o+(0.634925838303)*x[0]+(-0.0508765762892)*x[1]**o+(-0.889731180834)*x[1]
        arg[3,0,0,1]=(-0.611336323518)*x[0]**o+(0.748058372943)*x[0]+(-0.395060568704)*x[1]**o+(-0.413772262434)*x[1]
        arg[3,0,1,0]=(-0.76924573047)*x[0]**o+(0.61286303551)*x[0]+(0.810883211884)*x[1]**o+(-0.908861022244)*x[1]
        arg[3,0,1,1]=(0.269999661235)*x[0]**o+(0.0330954575695)*x[0]+(0.560272546054)*x[1]**o+(0.877730763728)*x[1]
        arg[3,0,2,0]=(-0.617658215856)*x[0]**o+(0.758414649277)*x[0]+(-0.223497113863)*x[1]**o+(-0.0706589108357)*x[1]
        arg[3,0,2,1]=(0.47956571804)*x[0]**o+(0.938375191345)*x[0]+(-0.193997511219)*x[1]**o+(0.500063249143)*x[1]
        arg[3,1,0,0]=(0.994033773649)*x[0]**o+(0.203022945646)*x[0]+(0.439296640709)*x[1]**o+(0.283679657893)*x[1]
        arg[3,1,0,1]=(0.32495149734)*x[0]**o+(-0.871627347694)*x[0]+(-0.438159387872)*x[1]**o+(0.169112665588)*x[1]
        arg[3,1,1,0]=(-0.0979322945554)*x[0]**o+(-0.936928842822)*x[0]+(0.0402738011011)*x[1]**o+(-0.445031618539)*x[1]
        arg[3,1,1,1]=(-0.511218949907)*x[0]**o+(0.255146779315)*x[0]+(-0.802012163396)*x[1]**o+(0.361819705056)*x[1]
        arg[3,1,2,0]=(-0.902499203011)*x[0]**o+(0.76589952141)*x[0]+(0.52498929732)*x[1]**o+(0.493894292321)*x[1]
        arg[3,1,2,1]=(-0.475117609018)*x[0]**o+(0.682205309303)*x[0]+(-0.661824086649)*x[1]**o+(0.985521553774)*x[1]
        arg[3,2,0,0]=(0.937321686473)*x[0]**o+(-0.518305775925)*x[0]+(0.171689574942)*x[1]**o+(-0.751579960819)*x[1]
        arg[3,2,0,1]=(-0.607004597985)*x[0]**o+(0.342330208344)*x[0]+(0.692094243326)*x[1]**o+(-0.457756057357)*x[1]
        arg[3,2,1,0]=(0.126171621354)*x[0]**o+(0.0499938423847)*x[0]+(0.572006209576)*x[1]**o+(0.121601182464)*x[1]
        arg[3,2,1,1]=(0.337936456181)*x[0]**o+(-0.940186907514)*x[0]+(-0.965354931124)*x[1]**o+(-0.0503295532873)*x[1]
        arg[3,2,2,0]=(0.439590816231)*x[0]**o+(-0.145314466072)*x[0]+(-0.847175364201)*x[1]**o+(-0.570000251495)*x[1]
        arg[3,2,2,1]=(0.401069335123)*x[0]**o+(-0.117139167508)*x[0]+(0.192907527853)*x[1]**o+(0.573283391581)*x[1]
        arg[3,3,0,0]=(-0.434029411666)*x[0]**o+(-0.689115317009)*x[0]+(0.63450218888)*x[1]**o+(0.492858023519)*x[1]
        arg[3,3,0,1]=(-0.678722283369)*x[0]**o+(-0.62031826393)*x[0]+(-0.458646022327)*x[1]**o+(-0.650857410927)*x[1]
        arg[3,3,1,0]=(0.413477874719)*x[0]**o+(-0.903116096879)*x[0]+(-0.505961511722)*x[1]**o+(0.649554735459)*x[1]
        arg[3,3,1,1]=(-0.58459645458)*x[0]**o+(0.861318565628)*x[0]+(0.378511063123)*x[1]**o+(-0.694820011654)*x[1]
        arg[3,3,2,0]=(0.82844733258)*x[0]**o+(-0.151929131446)*x[0]+(0.322848411081)*x[1]**o+(0.436062233263)*x[1]
        arg[3,3,2,1]=(0.269627291792)*x[0]**o+(0.618140998132)*x[0]+(0.851057431407)*x[1]**o+(0.716088832502)*x[1]
        arg[3,4,0,0]=(0.0557605432613)*x[0]**o+(0.0689569775656)*x[0]+(0.489781121774)*x[1]**o+(0.214459998733)*x[1]
        arg[3,4,0,1]=(0.429111134289)*x[0]**o+(0.648324951051)*x[0]+(0.00596799425797)*x[1]**o+(-0.752637714965)*x[1]
        arg[3,4,1,0]=(-0.0394648097179)*x[0]**o+(0.842965554058)*x[0]+(-0.124493160479)*x[1]**o+(0.81914053244)*x[1]
        arg[3,4,1,1]=(0.309465325874)*x[0]**o+(0.631381272937)*x[0]+(-0.164415508845)*x[1]**o+(0.316908911599)*x[1]
        arg[3,4,2,0]=(-0.615108314792)*x[0]**o+(0.300936829452)*x[0]+(-0.289514252333)*x[1]**o+(0.844949030767)*x[1]
        arg[3,4,2,1]=(0.80117516909)*x[0]**o+(-0.602632221639)*x[0]+(-0.68193581457)*x[1]**o+(0.203319693067)*x[1]
        ref[0,0,0,0]=(0.0737428272134)*x_ref[0]**o+(0.841780036645)*x_ref[0]+(-0.110824257531)*x_ref[1]**o+(-0.957726444468)*x_ref[1]
        ref[0,0,0,1]=(-0.934840410671)*x_ref[0]**o+(-0.83301064321)*x_ref[0]+(-0.352287650266)*x_ref[1]**o+(-0.961111564077)*x_ref[1]
        ref[0,0,1,0]=(-0.829500940226)*x_ref[0]**o+(0.644171710931)*x_ref[0]+(0.232545739722)*x_ref[1]**o+(0.584336012862)*x_ref[1]
        ref[0,0,1,1]=(0.426298830976)*x_ref[0]**o+(-0.615281197268)*x_ref[0]+(0.206334026192)*x_ref[1]**o+(-0.443698947044)*x_ref[1]
        ref[0,0,2,0]=(-0.236495891617)*x_ref[0]**o+(0.0883376851179)*x_ref[0]+(-0.176541434356)*x_ref[1]**o+(0.297175301617)*x_ref[1]
        ref[0,0,2,1]=(-0.879955324373)*x_ref[0]**o+(-0.244778913771)*x_ref[0]+(0.244131188594)*x_ref[1]**o+(-0.334739112198)*x_ref[1]
        ref[0,1,0,0]=(-0.64887128006)*x_ref[0]**o+(-0.489951454377)*x_ref[0]+(-0.328402433668)*x_ref[1]**o+(0.229814237117)*x_ref[1]
        ref[0,1,0,1]=(-0.339451656097)*x_ref[0]**o+(-0.409320682304)*x_ref[0]+(0.216482338677)*x_ref[1]**o+(-0.10594373055)*x_ref[1]
        ref[0,1,1,0]=(-0.350538479425)*x_ref[0]**o+(0.466450046985)*x_ref[0]+(-0.779124795292)*x_ref[1]**o+(-0.654562596704)*x_ref[1]
        ref[0,1,1,1]=(0.00350092879868)*x_ref[0]**o+(-0.214643088647)*x_ref[0]+(0.786934081411)*x_ref[1]**o+(-0.296853380043)*x_ref[1]
        ref[0,1,2,0]=(-0.640183955648)*x_ref[0]**o+(0.568647968237)*x_ref[0]+(0.575533476901)*x_ref[1]**o+(0.854540594085)*x_ref[1]
        ref[0,1,2,1]=(0.633674899518)*x_ref[0]**o+(0.829255679443)*x_ref[0]+(0.537888424303)*x_ref[1]**o+(-0.495082587617)*x_ref[1]
        ref[0,2,0,0]=(-0.700436798342)*x_ref[0]**o+(-0.732581859026)*x_ref[0]+(-0.508905353053)*x_ref[1]**o+(0.664601151676)*x_ref[1]
        ref[0,2,0,1]=(0.720335764302)*x_ref[0]**o+(0.941159917988)*x_ref[0]+(0.2817440547)*x_ref[1]**o+(0.553712160058)*x_ref[1]
        ref[0,2,1,0]=(-0.937114312262)*x_ref[0]**o+(-0.115517396201)*x_ref[0]+(-0.672697125905)*x_ref[1]**o+(-0.221320095369)*x_ref[1]
        ref[0,2,1,1]=(-0.52059341062)*x_ref[0]**o+(-0.653834647475)*x_ref[0]+(-0.0851100861135)*x_ref[1]**o+(-0.668300210067)*x_ref[1]
        ref[0,2,2,0]=(-0.943613035509)*x_ref[0]**o+(-0.294843145822)*x_ref[0]+(0.496278748831)*x_ref[1]**o+(0.901154419425)*x_ref[1]
        ref[0,2,2,1]=(-0.421698818203)*x_ref[0]**o+(-0.721918177544)*x_ref[0]+(-0.197855702215)*x_ref[1]**o+(0.50882827715)*x_ref[1]
        ref[0,3,0,0]=(0.931990780298)*x_ref[0]**o+(-0.550332606951)*x_ref[0]+(0.401584654642)*x_ref[1]**o+(0.0965259747601)*x_ref[1]
        ref[0,3,0,1]=(0.881357708445)*x_ref[0]**o+(0.564624034197)*x_ref[0]+(-0.458191897924)*x_ref[1]**o+(-0.427701112807)*x_ref[1]
        ref[0,3,1,0]=(0.624777088081)*x_ref[0]**o+(0.862654721208)*x_ref[0]+(-0.707618093268)*x_ref[1]**o+(0.743304953313)*x_ref[1]
        ref[0,3,1,1]=(0.4510836974)*x_ref[0]**o+(-0.357023031453)*x_ref[0]+(-0.00503924010273)*x_ref[1]**o+(0.545024336011)*x_ref[1]
        ref[0,3,2,0]=(0.109003851094)*x_ref[0]**o+(0.168659376881)*x_ref[0]+(-0.0578953780655)*x_ref[1]**o+(-0.481212462712)*x_ref[1]
        ref[0,3,2,1]=(-0.191183501636)*x_ref[0]**o+(-0.316145516377)*x_ref[0]+(0.402460103016)*x_ref[1]**o+(-0.163612464438)*x_ref[1]
        ref[0,4,0,0]=(0.489815345033)*x_ref[0]**o+(0.999138280765)*x_ref[0]+(-0.626627766812)*x_ref[1]**o+(-0.339653992568)*x_ref[1]
        ref[0,4,0,1]=(0.616350228791)*x_ref[0]**o+(0.933446438765)*x_ref[0]+(0.944244035512)*x_ref[1]**o+(0.829407789353)*x_ref[1]
        ref[0,4,1,0]=(0.619918547863)*x_ref[0]**o+(0.995488719729)*x_ref[0]+(-0.78847429581)*x_ref[1]**o+(0.542645674462)*x_ref[1]
        ref[0,4,1,1]=(-0.149244819992)*x_ref[0]**o+(0.921766299925)*x_ref[0]+(0.435255339567)*x_ref[1]**o+(-0.0596561300741)*x_ref[1]
        ref[0,4,2,0]=(-0.816364336598)*x_ref[0]**o+(0.701160246658)*x_ref[0]+(-0.616385505652)*x_ref[1]**o+(-0.558625482105)*x_ref[1]
        ref[0,4,2,1]=(-0.36944364693)*x_ref[0]**o+(0.243426977623)*x_ref[0]+(-0.61578435391)*x_ref[1]**o+(0.932570326139)*x_ref[1]
        ref[1,0,0,0]=(-0.36839962807)*x_ref[0]**o+(0.590534439534)*x_ref[0]+(-0.615158899289)*x_ref[1]**o+(0.973649158296)*x_ref[1]
        ref[1,0,0,1]=(0.761515772771)*x_ref[0]**o+(0.706406307012)*x_ref[0]+(0.328712319043)*x_ref[1]**o+(0.438600140517)*x_ref[1]
        ref[1,0,1,0]=(0.895080579207)*x_ref[0]**o+(-0.166789328225)*x_ref[0]+(0.807540897371)*x_ref[1]**o+(0.0345522692723)*x_ref[1]
        ref[1,0,1,1]=(-0.460062661172)*x_ref[0]**o+(0.505167337562)*x_ref[0]+(0.25155527144)*x_ref[1]**o+(0.601093887451)*x_ref[1]
        ref[1,0,2,0]=(0.382129818527)*x_ref[0]**o+(0.694676536542)*x_ref[0]+(0.702794119083)*x_ref[1]**o+(0.704094013056)*x_ref[1]
        ref[1,0,2,1]=(0.265476982913)*x_ref[0]**o+(-0.134338744881)*x_ref[0]+(-0.0938910335356)*x_ref[1]**o+(-0.560073316097)*x_ref[1]
        ref[1,1,0,0]=(0.977368328792)*x_ref[0]**o+(0.084482890741)*x_ref[0]+(0.57068389287)*x_ref[1]**o+(-0.242087071526)*x_ref[1]
        ref[1,1,0,1]=(-0.36067775897)*x_ref[0]**o+(0.768187470286)*x_ref[0]+(-0.881150642619)*x_ref[1]**o+(-0.15839498619)*x_ref[1]
        ref[1,1,1,0]=(-0.808263913645)*x_ref[0]**o+(0.788403321664)*x_ref[0]+(0.976206012343)*x_ref[1]**o+(-0.636933353579)*x_ref[1]
        ref[1,1,1,1]=(0.699185445546)*x_ref[0]**o+(0.251811206876)*x_ref[0]+(-0.146320762812)*x_ref[1]**o+(-0.248033048423)*x_ref[1]
        ref[1,1,2,0]=(-0.782940880812)*x_ref[0]**o+(0.789401686528)*x_ref[0]+(-0.806883904562)*x_ref[1]**o+(0.83689696772)*x_ref[1]
        ref[1,1,2,1]=(0.652404993342)*x_ref[0]**o+(-0.135318598212)*x_ref[0]+(0.138928760301)*x_ref[1]**o+(0.675593554805)*x_ref[1]
        ref[1,2,0,0]=(0.471746379585)*x_ref[0]**o+(-0.989783415847)*x_ref[0]+(-0.0935052296418)*x_ref[1]**o+(-0.532245102898)*x_ref[1]
        ref[1,2,0,1]=(-0.44541059979)*x_ref[0]**o+(0.593734266824)*x_ref[0]+(-0.616647011751)*x_ref[1]**o+(0.461140297024)*x_ref[1]
        ref[1,2,1,0]=(-0.9144210459)*x_ref[0]**o+(0.116806032197)*x_ref[0]+(0.935610395306)*x_ref[1]**o+(0.550884822213)*x_ref[1]
        ref[1,2,1,1]=(-0.962444455944)*x_ref[0]**o+(-0.654328013652)*x_ref[0]+(-0.545330994236)*x_ref[1]**o+(-0.919705456162)*x_ref[1]
        ref[1,2,2,0]=(-0.724491374462)*x_ref[0]**o+(-0.820402375498)*x_ref[0]+(-0.953956431336)*x_ref[1]**o+(-0.201450523182)*x_ref[1]
        ref[1,2,2,1]=(0.723637497881)*x_ref[0]**o+(-0.144483238523)*x_ref[0]+(0.00379745988102)*x_ref[1]**o+(-0.838616087543)*x_ref[1]
        ref[1,3,0,0]=(-0.962256046042)*x_ref[0]**o+(-0.453580322779)*x_ref[0]+(0.887446007424)*x_ref[1]**o+(-0.404454768131)*x_ref[1]
        ref[1,3,0,1]=(0.829189872775)*x_ref[0]**o+(0.327985850399)*x_ref[0]+(0.113600094307)*x_ref[1]**o+(-0.338775377551)*x_ref[1]
        ref[1,3,1,0]=(0.912320027833)*x_ref[0]**o+(0.0658152563951)*x_ref[0]+(-0.843491517568)*x_ref[1]**o+(0.858344434434)*x_ref[1]
        ref[1,3,1,1]=(-0.133918678317)*x_ref[0]**o+(0.644272729149)*x_ref[0]+(0.140663927965)*x_ref[1]**o+(0.853385355163)*x_ref[1]
        ref[1,3,2,0]=(0.37733357322)*x_ref[0]**o+(0.255105934063)*x_ref[0]+(0.88881892183)*x_ref[1]**o+(-0.722661948279)*x_ref[1]
        ref[1,3,2,1]=(-0.2827466345)*x_ref[0]**o+(-0.518815029512)*x_ref[0]+(-0.995043824929)*x_ref[1]**o+(0.973473354481)*x_ref[1]
        ref[1,4,0,0]=(0.044904732821)*x_ref[0]**o+(-0.0373604593083)*x_ref[0]+(0.184254626748)*x_ref[1]**o+(0.031832590233)*x_ref[1]
        ref[1,4,0,1]=(0.388699053911)*x_ref[0]**o+(-0.0619606394414)*x_ref[0]+(0.477947819745)*x_ref[1]**o+(0.982909216221)*x_ref[1]
        ref[1,4,1,0]=(0.854348679973)*x_ref[0]**o+(-0.822052612379)*x_ref[0]+(-0.950305089968)*x_ref[1]**o+(0.0100277256384)*x_ref[1]
        ref[1,4,1,1]=(0.265075277179)*x_ref[0]**o+(0.489846993407)*x_ref[0]+(-0.961111499722)*x_ref[1]**o+(-0.0457909055602)*x_ref[1]
        ref[1,4,2,0]=(0.701951191951)*x_ref[0]**o+(0.675232620259)*x_ref[0]+(-0.141848599227)*x_ref[1]**o+(0.737150257564)*x_ref[1]
        ref[1,4,2,1]=(-0.450684564415)*x_ref[0]**o+(0.854849922021)*x_ref[0]+(0.729670746723)*x_ref[1]**o+(0.530372628879)*x_ref[1]
        ref[2,0,0,0]=(-0.131153262675)*x_ref[0]**o+(-0.477683407979)*x_ref[0]+(0.231416752516)*x_ref[1]**o+(0.790897064402)*x_ref[1]
        ref[2,0,0,1]=(0.0956557004382)*x_ref[0]**o+(0.209084913471)*x_ref[0]+(-0.328544695733)*x_ref[1]**o+(0.701245770328)*x_ref[1]
        ref[2,0,1,0]=(-0.476403718361)*x_ref[0]**o+(0.689023533158)*x_ref[0]+(0.108732189071)*x_ref[1]**o+(0.243379977453)*x_ref[1]
        ref[2,0,1,1]=(0.414917801157)*x_ref[0]**o+(-0.843544596785)*x_ref[0]+(0.772236792935)*x_ref[1]**o+(0.836028831232)*x_ref[1]
        ref[2,0,2,0]=(0.754970600617)*x_ref[0]**o+(-0.422318891122)*x_ref[0]+(0.770767911534)*x_ref[1]**o+(-0.446737894095)*x_ref[1]
        ref[2,0,2,1]=(0.0139989331471)*x_ref[0]**o+(0.262286042291)*x_ref[0]+(-0.440109329114)*x_ref[1]**o+(-0.844965114722)*x_ref[1]
        ref[2,1,0,0]=(-0.758977153009)*x_ref[0]**o+(-0.991192011039)*x_ref[0]+(0.642267239475)*x_ref[1]**o+(-0.701360491163)*x_ref[1]
        ref[2,1,0,1]=(-0.123019007509)*x_ref[0]**o+(-0.622323252979)*x_ref[0]+(0.441364031664)*x_ref[1]**o+(0.311433222821)*x_ref[1]
        ref[2,1,1,0]=(-0.0446644182928)*x_ref[0]**o+(0.614952704493)*x_ref[0]+(-0.228284494629)*x_ref[1]**o+(0.077603612991)*x_ref[1]
        ref[2,1,1,1]=(0.712924108225)*x_ref[0]**o+(-0.153966300537)*x_ref[0]+(-0.584450014127)*x_ref[1]**o+(0.476329149833)*x_ref[1]
        ref[2,1,2,0]=(0.554026516037)*x_ref[0]**o+(0.730575821847)*x_ref[0]+(-0.247529915107)*x_ref[1]**o+(0.210060454223)*x_ref[1]
        ref[2,1,2,1]=(0.877017208752)*x_ref[0]**o+(-0.800916200997)*x_ref[0]+(0.0996710064295)*x_ref[1]**o+(-0.0890375735808)*x_ref[1]
        ref[2,2,0,0]=(-0.198026186062)*x_ref[0]**o+(-0.416160224681)*x_ref[0]+(-0.626491103997)*x_ref[1]**o+(-0.950177752043)*x_ref[1]
        ref[2,2,0,1]=(-0.401319234479)*x_ref[0]**o+(0.948120359627)*x_ref[0]+(0.420351950404)*x_ref[1]**o+(-0.242850312594)*x_ref[1]
        ref[2,2,1,0]=(-0.239652069063)*x_ref[0]**o+(-0.894811076801)*x_ref[0]+(0.979721413509)*x_ref[1]**o+(-0.712493893566)*x_ref[1]
        ref[2,2,1,1]=(0.239416494209)*x_ref[0]**o+(0.449047649538)*x_ref[0]+(0.0503751073583)*x_ref[1]**o+(0.739949331521)*x_ref[1]
        ref[2,2,2,0]=(-0.318067701034)*x_ref[0]**o+(-0.00208209268356)*x_ref[0]+(-0.513821624843)*x_ref[1]**o+(-0.202688815501)*x_ref[1]
        ref[2,2,2,1]=(-0.194111524544)*x_ref[0]**o+(-0.843213241929)*x_ref[0]+(0.124223747087)*x_ref[1]**o+(-0.988305059412)*x_ref[1]
        ref[2,3,0,0]=(0.450894655873)*x_ref[0]**o+(-0.809351657482)*x_ref[0]+(-0.80555128185)*x_ref[1]**o+(-0.922123806508)*x_ref[1]
        ref[2,3,0,1]=(-0.462822602909)*x_ref[0]**o+(0.504388766027)*x_ref[0]+(-0.152725064915)*x_ref[1]**o+(0.988745142638)*x_ref[1]
        ref[2,3,1,0]=(-0.717873944493)*x_ref[0]**o+(0.987011714613)*x_ref[0]+(0.908500737929)*x_ref[1]**o+(0.612497707425)*x_ref[1]
        ref[2,3,1,1]=(0.221645363399)*x_ref[0]**o+(0.219147687681)*x_ref[0]+(-0.195831118343)*x_ref[1]**o+(-0.59541920919)*x_ref[1]
        ref[2,3,2,0]=(-0.289487342206)*x_ref[0]**o+(-0.96904981397)*x_ref[0]+(-0.844880792969)*x_ref[1]**o+(0.803145388038)*x_ref[1]
        ref[2,3,2,1]=(-0.0234295826132)*x_ref[0]**o+(-0.839069887016)*x_ref[0]+(0.591231341444)*x_ref[1]**o+(-0.678605318787)*x_ref[1]
        ref[2,4,0,0]=(0.717943320641)*x_ref[0]**o+(0.625728011485)*x_ref[0]+(0.456836257053)*x_ref[1]**o+(-0.00523685585287)*x_ref[1]
        ref[2,4,0,1]=(-0.549526374185)*x_ref[0]**o+(0.57441917876)*x_ref[0]+(-0.913459106781)*x_ref[1]**o+(-0.610369375195)*x_ref[1]
        ref[2,4,1,0]=(-0.27166563459)*x_ref[0]**o+(0.146710581194)*x_ref[0]+(0.297934124718)*x_ref[1]**o+(0.989561209048)*x_ref[1]
        ref[2,4,1,1]=(-0.0973858709778)*x_ref[0]**o+(-0.909133394381)*x_ref[0]+(0.634182571287)*x_ref[1]**o+(0.477163052831)*x_ref[1]
        ref[2,4,2,0]=(0.0878998038301)*x_ref[0]**o+(-0.90568487021)*x_ref[0]+(0.798978355456)*x_ref[1]**o+(0.977402478037)*x_ref[1]
        ref[2,4,2,1]=(-0.456868538667)*x_ref[0]**o+(0.0723272702119)*x_ref[0]+(0.143461337462)*x_ref[1]**o+(-0.923713568343)*x_ref[1]
        ref[3,0,0,0]=(-0.154410689788)*x_ref[0]**o+(0.634925838303)*x_ref[0]+(-0.0508765762892)*x_ref[1]**o+(-0.889731180834)*x_ref[1]
        ref[3,0,0,1]=(-0.611336323518)*x_ref[0]**o+(0.748058372943)*x_ref[0]+(-0.395060568704)*x_ref[1]**o+(-0.413772262434)*x_ref[1]
        ref[3,0,1,0]=(-0.76924573047)*x_ref[0]**o+(0.61286303551)*x_ref[0]+(0.810883211884)*x_ref[1]**o+(-0.908861022244)*x_ref[1]
        ref[3,0,1,1]=(0.269999661235)*x_ref[0]**o+(0.0330954575695)*x_ref[0]+(0.560272546054)*x_ref[1]**o+(0.877730763728)*x_ref[1]
        ref[3,0,2,0]=(-0.617658215856)*x_ref[0]**o+(0.758414649277)*x_ref[0]+(-0.223497113863)*x_ref[1]**o+(-0.0706589108357)*x_ref[1]
        ref[3,0,2,1]=(0.47956571804)*x_ref[0]**o+(0.938375191345)*x_ref[0]+(-0.193997511219)*x_ref[1]**o+(0.500063249143)*x_ref[1]
        ref[3,1,0,0]=(0.994033773649)*x_ref[0]**o+(0.203022945646)*x_ref[0]+(0.439296640709)*x_ref[1]**o+(0.283679657893)*x_ref[1]
        ref[3,1,0,1]=(0.32495149734)*x_ref[0]**o+(-0.871627347694)*x_ref[0]+(-0.438159387872)*x_ref[1]**o+(0.169112665588)*x_ref[1]
        ref[3,1,1,0]=(-0.0979322945554)*x_ref[0]**o+(-0.936928842822)*x_ref[0]+(0.0402738011011)*x_ref[1]**o+(-0.445031618539)*x_ref[1]
        ref[3,1,1,1]=(-0.511218949907)*x_ref[0]**o+(0.255146779315)*x_ref[0]+(-0.802012163396)*x_ref[1]**o+(0.361819705056)*x_ref[1]
        ref[3,1,2,0]=(-0.902499203011)*x_ref[0]**o+(0.76589952141)*x_ref[0]+(0.52498929732)*x_ref[1]**o+(0.493894292321)*x_ref[1]
        ref[3,1,2,1]=(-0.475117609018)*x_ref[0]**o+(0.682205309303)*x_ref[0]+(-0.661824086649)*x_ref[1]**o+(0.985521553774)*x_ref[1]
        ref[3,2,0,0]=(0.937321686473)*x_ref[0]**o+(-0.518305775925)*x_ref[0]+(0.171689574942)*x_ref[1]**o+(-0.751579960819)*x_ref[1]
        ref[3,2,0,1]=(-0.607004597985)*x_ref[0]**o+(0.342330208344)*x_ref[0]+(0.692094243326)*x_ref[1]**o+(-0.457756057357)*x_ref[1]
        ref[3,2,1,0]=(0.126171621354)*x_ref[0]**o+(0.0499938423847)*x_ref[0]+(0.572006209576)*x_ref[1]**o+(0.121601182464)*x_ref[1]
        ref[3,2,1,1]=(0.337936456181)*x_ref[0]**o+(-0.940186907514)*x_ref[0]+(-0.965354931124)*x_ref[1]**o+(-0.0503295532873)*x_ref[1]
        ref[3,2,2,0]=(0.439590816231)*x_ref[0]**o+(-0.145314466072)*x_ref[0]+(-0.847175364201)*x_ref[1]**o+(-0.570000251495)*x_ref[1]
        ref[3,2,2,1]=(0.401069335123)*x_ref[0]**o+(-0.117139167508)*x_ref[0]+(0.192907527853)*x_ref[1]**o+(0.573283391581)*x_ref[1]
        ref[3,3,0,0]=(-0.434029411666)*x_ref[0]**o+(-0.689115317009)*x_ref[0]+(0.63450218888)*x_ref[1]**o+(0.492858023519)*x_ref[1]
        ref[3,3,0,1]=(-0.678722283369)*x_ref[0]**o+(-0.62031826393)*x_ref[0]+(-0.458646022327)*x_ref[1]**o+(-0.650857410927)*x_ref[1]
        ref[3,3,1,0]=(0.413477874719)*x_ref[0]**o+(-0.903116096879)*x_ref[0]+(-0.505961511722)*x_ref[1]**o+(0.649554735459)*x_ref[1]
        ref[3,3,1,1]=(-0.58459645458)*x_ref[0]**o+(0.861318565628)*x_ref[0]+(0.378511063123)*x_ref[1]**o+(-0.694820011654)*x_ref[1]
        ref[3,3,2,0]=(0.82844733258)*x_ref[0]**o+(-0.151929131446)*x_ref[0]+(0.322848411081)*x_ref[1]**o+(0.436062233263)*x_ref[1]
        ref[3,3,2,1]=(0.269627291792)*x_ref[0]**o+(0.618140998132)*x_ref[0]+(0.851057431407)*x_ref[1]**o+(0.716088832502)*x_ref[1]
        ref[3,4,0,0]=(0.0557605432613)*x_ref[0]**o+(0.0689569775656)*x_ref[0]+(0.489781121774)*x_ref[1]**o+(0.214459998733)*x_ref[1]
        ref[3,4,0,1]=(0.429111134289)*x_ref[0]**o+(0.648324951051)*x_ref[0]+(0.00596799425797)*x_ref[1]**o+(-0.752637714965)*x_ref[1]
        ref[3,4,1,0]=(-0.0394648097179)*x_ref[0]**o+(0.842965554058)*x_ref[0]+(-0.124493160479)*x_ref[1]**o+(0.81914053244)*x_ref[1]
        ref[3,4,1,1]=(0.309465325874)*x_ref[0]**o+(0.631381272937)*x_ref[0]+(-0.164415508845)*x_ref[1]**o+(0.316908911599)*x_ref[1]
        ref[3,4,2,0]=(-0.615108314792)*x_ref[0]**o+(0.300936829452)*x_ref[0]+(-0.289514252333)*x_ref[1]**o+(0.844949030767)*x_ref[1]
        ref[3,4,2,1]=(0.80117516909)*x_ref[0]**o+(-0.602632221639)*x_ref[0]+(-0.68193581457)*x_ref[1]**o+(0.203319693067)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.349437323395)*x[0]**o+(0.817821891642)*x[0]+(-0.357998520848)*x[1]**o+(0.349311985429)*x[1]+(0.470351907942)*x[2]**o+(0.264007843212)*x[2]
        arg[0,0,0,1]=(-0.747824148035)*x[0]**o+(0.551817215468)*x[0]+(0.980476067966)*x[1]**o+(0.521120126285)*x[1]+(-0.472357720779)*x[2]**o+(-0.379502066444)*x[2]
        arg[0,0,1,0]=(0.302984899868)*x[0]**o+(0.556601505796)*x[0]+(-0.698948476338)*x[1]**o+(-0.587379572024)*x[1]+(-0.625760359909)*x[2]**o+(0.231956543432)*x[2]
        arg[0,0,1,1]=(-0.335943560463)*x[0]**o+(-0.497643152155)*x[0]+(-0.849829024888)*x[1]**o+(-0.388751135318)*x[1]+(-0.517696637743)*x[2]**o+(-0.966210901)*x[2]
        arg[0,0,2,0]=(-0.0953058405321)*x[0]**o+(-0.864580918175)*x[0]+(-0.732086619685)*x[1]**o+(0.872044654816)*x[1]+(-0.225850391375)*x[2]**o+(-0.502203654371)*x[2]
        arg[0,0,2,1]=(0.228408069101)*x[0]**o+(0.267590362458)*x[0]+(0.127397041031)*x[1]**o+(0.0142393720782)*x[1]+(-0.257563412478)*x[2]**o+(0.0211877877507)*x[2]
        arg[0,1,0,0]=(0.921125039994)*x[0]**o+(0.372548304521)*x[0]+(-0.66574364308)*x[1]**o+(0.0895600219155)*x[1]+(-0.656870562037)*x[2]**o+(-0.608819638268)*x[2]
        arg[0,1,0,1]=(-0.396625252579)*x[0]**o+(0.655542320616)*x[0]+(-0.542629502059)*x[1]**o+(0.587317299384)*x[1]+(0.96380848967)*x[2]**o+(0.121341827766)*x[2]
        arg[0,1,1,0]=(0.0908208346003)*x[0]**o+(-0.0398227321951)*x[0]+(0.907607555288)*x[1]**o+(0.527230589013)*x[1]+(0.519287416653)*x[2]**o+(0.103180358249)*x[2]
        arg[0,1,1,1]=(0.0615195625837)*x[0]**o+(0.695720276474)*x[0]+(0.0862101684049)*x[1]**o+(0.517150419955)*x[1]+(0.511207669575)*x[2]**o+(-0.9556472362)*x[2]
        arg[0,1,2,0]=(0.70040395652)*x[0]**o+(-0.576266027865)*x[0]+(-0.133947982034)*x[1]**o+(-0.764976368456)*x[1]+(-0.860780506782)*x[2]**o+(-0.137946340663)*x[2]
        arg[0,1,2,1]=(0.356017596832)*x[0]**o+(0.94912783604)*x[0]+(0.325951760294)*x[1]**o+(0.0623695610721)*x[1]+(-0.106031051895)*x[2]**o+(-0.339314970581)*x[2]
        arg[0,2,0,0]=(-0.124954395016)*x[0]**o+(0.558233271782)*x[0]+(-0.956846820394)*x[1]**o+(-0.599332648704)*x[1]+(-0.322444201477)*x[2]**o+(0.0833108732532)*x[2]
        arg[0,2,0,1]=(-0.713065259589)*x[0]**o+(-0.597746726477)*x[0]+(-0.976654446591)*x[1]**o+(0.977779698697)*x[1]+(0.694427679422)*x[2]**o+(-0.364463986712)*x[2]
        arg[0,2,1,0]=(-0.415861468352)*x[0]**o+(-0.159802964908)*x[0]+(-0.132331280778)*x[1]**o+(0.120829594848)*x[1]+(-0.114181951611)*x[2]**o+(-0.810187313564)*x[2]
        arg[0,2,1,1]=(-0.869925695451)*x[0]**o+(-0.535444611026)*x[0]+(0.62908837594)*x[1]**o+(-0.488583607314)*x[1]+(0.8461615049)*x[2]**o+(-0.558010294727)*x[2]
        arg[0,2,2,0]=(-0.0822232217466)*x[0]**o+(-0.673631382141)*x[0]+(0.803717792107)*x[1]**o+(0.339934715615)*x[1]+(-0.886365924996)*x[2]**o+(0.0206517750753)*x[2]
        arg[0,2,2,1]=(0.367656249869)*x[0]**o+(0.000987963028539)*x[0]+(0.399182215451)*x[1]**o+(0.923259635138)*x[1]+(0.8573338122)*x[2]**o+(0.679082213415)*x[2]
        arg[0,3,0,0]=(-0.722680496644)*x[0]**o+(0.199427912651)*x[0]+(-0.351799245099)*x[1]**o+(0.240285416231)*x[1]+(-0.701704868242)*x[2]**o+(0.265610436018)*x[2]
        arg[0,3,0,1]=(-0.14479043328)*x[0]**o+(-0.192670959504)*x[0]+(0.327412080296)*x[1]**o+(-0.390894637576)*x[1]+(-0.401862813677)*x[2]**o+(-0.87150071561)*x[2]
        arg[0,3,1,0]=(0.430039400239)*x[0]**o+(-0.0691443536291)*x[0]+(0.205492039071)*x[1]**o+(-0.391580258314)*x[1]+(0.791748125642)*x[2]**o+(-0.0896722225431)*x[2]
        arg[0,3,1,1]=(-0.177554766913)*x[0]**o+(-0.399389195454)*x[0]+(-0.304228953477)*x[1]**o+(0.00462595697167)*x[1]+(0.271814579884)*x[2]**o+(-0.17305639368)*x[2]
        arg[0,3,2,0]=(-0.105869589157)*x[0]**o+(0.803488342924)*x[0]+(-0.65117861655)*x[1]**o+(-0.27529059731)*x[1]+(-0.777107387582)*x[2]**o+(-0.690980328369)*x[2]
        arg[0,3,2,1]=(-0.13792230295)*x[0]**o+(0.649721420509)*x[0]+(0.401441756822)*x[1]**o+(0.468949007973)*x[1]+(-0.762015699412)*x[2]**o+(0.389781367434)*x[2]
        arg[0,4,0,0]=(0.873653819021)*x[0]**o+(0.539406798913)*x[0]+(0.802177801924)*x[1]**o+(-0.861314989922)*x[1]+(0.847509960118)*x[2]**o+(-0.41381965291)*x[2]
        arg[0,4,0,1]=(0.0416858647633)*x[0]**o+(0.866825467268)*x[0]+(0.89776163679)*x[1]**o+(-0.978326218858)*x[1]+(-0.409795413277)*x[2]**o+(0.982449327977)*x[2]
        arg[0,4,1,0]=(-0.114530966075)*x[0]**o+(-0.182414280034)*x[0]+(0.154385814218)*x[1]**o+(0.869624332271)*x[1]+(0.477213291128)*x[2]**o+(-0.137123375612)*x[2]
        arg[0,4,1,1]=(-0.52963019184)*x[0]**o+(0.0217955878025)*x[0]+(-0.174531798693)*x[1]**o+(-0.738947178141)*x[1]+(0.198130667502)*x[2]**o+(-0.449952166994)*x[2]
        arg[0,4,2,0]=(0.998209708224)*x[0]**o+(0.734906841374)*x[0]+(0.225596363429)*x[1]**o+(0.862772101878)*x[1]+(-0.133563132638)*x[2]**o+(-0.78059821331)*x[2]
        arg[0,4,2,1]=(0.384580893901)*x[0]**o+(-0.881918368669)*x[0]+(-0.950768236608)*x[1]**o+(-0.452275072093)*x[1]+(0.592779023744)*x[2]**o+(0.798735695084)*x[2]
        arg[1,0,0,0]=(0.786827894179)*x[0]**o+(-0.233241290567)*x[0]+(0.946717698992)*x[1]**o+(0.179368479607)*x[1]+(-0.10409968587)*x[2]**o+(-0.770115022756)*x[2]
        arg[1,0,0,1]=(-0.355389699793)*x[0]**o+(0.356407960494)*x[0]+(-0.371831767713)*x[1]**o+(0.139120259823)*x[1]+(0.910761795881)*x[2]**o+(0.725966050775)*x[2]
        arg[1,0,1,0]=(0.316372320948)*x[0]**o+(0.432472869369)*x[0]+(0.733187356771)*x[1]**o+(0.133915829584)*x[1]+(0.380713482112)*x[2]**o+(-0.874827460114)*x[2]
        arg[1,0,1,1]=(-0.0625786527083)*x[0]**o+(-0.469518973019)*x[0]+(-0.318610570327)*x[1]**o+(-0.304412247656)*x[1]+(0.844057224107)*x[2]**o+(0.508926494536)*x[2]
        arg[1,0,2,0]=(-0.759697281694)*x[0]**o+(0.701456398418)*x[0]+(-0.671612899124)*x[1]**o+(0.292474197717)*x[1]+(-0.445886723952)*x[2]**o+(-0.319953406782)*x[2]
        arg[1,0,2,1]=(-0.865703619781)*x[0]**o+(0.115639820869)*x[0]+(0.493634056223)*x[1]**o+(0.0343245670418)*x[1]+(0.748200872819)*x[2]**o+(-0.208438395054)*x[2]
        arg[1,1,0,0]=(0.767661930656)*x[0]**o+(-0.578385615919)*x[0]+(0.241313268557)*x[1]**o+(0.74276605359)*x[1]+(0.373638986644)*x[2]**o+(0.475479466784)*x[2]
        arg[1,1,0,1]=(0.723228882826)*x[0]**o+(-0.988524686322)*x[0]+(0.333664096597)*x[1]**o+(-0.756435323592)*x[1]+(0.771173510221)*x[2]**o+(-0.879337556561)*x[2]
        arg[1,1,1,0]=(0.939682105701)*x[0]**o+(-0.172311578817)*x[0]+(-0.407241492831)*x[1]**o+(0.446660064834)*x[1]+(0.0371203367053)*x[2]**o+(-0.302853430561)*x[2]
        arg[1,1,1,1]=(0.1093213068)*x[0]**o+(0.31232933407)*x[0]+(0.47474581408)*x[1]**o+(-0.602763201317)*x[1]+(0.803198523872)*x[2]**o+(-0.178088233667)*x[2]
        arg[1,1,2,0]=(-0.217141686114)*x[0]**o+(0.287363815362)*x[0]+(-0.352552947551)*x[1]**o+(0.106445580983)*x[1]+(0.340814562712)*x[2]**o+(-0.963691249733)*x[2]
        arg[1,1,2,1]=(0.612302241126)*x[0]**o+(-0.771765095612)*x[0]+(-0.18662763339)*x[1]**o+(0.10259882275)*x[1]+(-0.29535785913)*x[2]**o+(-0.577876877465)*x[2]
        arg[1,2,0,0]=(0.497131459554)*x[0]**o+(0.912133134267)*x[0]+(0.944369633113)*x[1]**o+(0.163408823815)*x[1]+(0.85636766551)*x[2]**o+(-0.00282458078743)*x[2]
        arg[1,2,0,1]=(0.362490806338)*x[0]**o+(0.22490192994)*x[0]+(-0.913665309243)*x[1]**o+(-0.88010395132)*x[1]+(-0.926205930269)*x[2]**o+(-0.231830082111)*x[2]
        arg[1,2,1,0]=(-0.974792784505)*x[0]**o+(-0.765605618923)*x[0]+(0.241470054253)*x[1]**o+(0.445863114014)*x[1]+(0.561062749944)*x[2]**o+(-0.215674406206)*x[2]
        arg[1,2,1,1]=(0.827013456823)*x[0]**o+(-0.191864800977)*x[0]+(-0.158099224794)*x[1]**o+(-0.921959562711)*x[1]+(0.930693535629)*x[2]**o+(-0.352967183792)*x[2]
        arg[1,2,2,0]=(0.258970603305)*x[0]**o+(0.586779837275)*x[0]+(0.82707087137)*x[1]**o+(-0.215210535461)*x[1]+(0.36890299705)*x[2]**o+(-0.724385890088)*x[2]
        arg[1,2,2,1]=(-0.125969269863)*x[0]**o+(-0.505216539273)*x[0]+(0.43242008053)*x[1]**o+(0.212079268112)*x[1]+(0.345759190477)*x[2]**o+(-0.98505468199)*x[2]
        arg[1,3,0,0]=(-0.647118215098)*x[0]**o+(0.318566502446)*x[0]+(-0.151406735047)*x[1]**o+(0.253202745125)*x[1]+(-0.404823111329)*x[2]**o+(-0.520797366557)*x[2]
        arg[1,3,0,1]=(-0.946886311645)*x[0]**o+(0.733699524424)*x[0]+(-0.615470980807)*x[1]**o+(-0.103976314414)*x[1]+(-0.932104431883)*x[2]**o+(0.157563806559)*x[2]
        arg[1,3,1,0]=(-0.289471503181)*x[0]**o+(0.749124853135)*x[0]+(-0.578861403092)*x[1]**o+(-0.136231378985)*x[1]+(0.532719246315)*x[2]**o+(0.542455831657)*x[2]
        arg[1,3,1,1]=(-0.343842490684)*x[0]**o+(0.110039182926)*x[0]+(0.901234650534)*x[1]**o+(-0.513362876241)*x[1]+(0.371952189126)*x[2]**o+(-0.169400349862)*x[2]
        arg[1,3,2,0]=(-0.766791944503)*x[0]**o+(0.0267491122166)*x[0]+(-0.699293509597)*x[1]**o+(-0.697363669971)*x[1]+(0.567854170308)*x[2]**o+(-0.516301389629)*x[2]
        arg[1,3,2,1]=(-0.288508579358)*x[0]**o+(0.239321880772)*x[0]+(0.677249635795)*x[1]**o+(0.48369084768)*x[1]+(0.681460633595)*x[2]**o+(-0.80218734437)*x[2]
        arg[1,4,0,0]=(0.510273388936)*x[0]**o+(0.35600020965)*x[0]+(0.271077500616)*x[1]**o+(-0.00479335810576)*x[1]+(-0.348457583464)*x[2]**o+(-0.965895122152)*x[2]
        arg[1,4,0,1]=(0.0134532748813)*x[0]**o+(-0.396954380686)*x[0]+(-0.176682002946)*x[1]**o+(0.971224261138)*x[1]+(-0.359602180261)*x[2]**o+(0.373748703363)*x[2]
        arg[1,4,1,0]=(-0.519413160925)*x[0]**o+(0.856346255965)*x[0]+(0.159167243935)*x[1]**o+(0.92068779838)*x[1]+(0.415122182947)*x[2]**o+(-0.595702311496)*x[2]
        arg[1,4,1,1]=(-0.608969598184)*x[0]**o+(-0.434750765676)*x[0]+(0.74471475701)*x[1]**o+(-0.229333246615)*x[1]+(-0.678138032736)*x[2]**o+(-0.514112926798)*x[2]
        arg[1,4,2,0]=(0.0633453561735)*x[0]**o+(-0.278569473082)*x[0]+(-0.842227144473)*x[1]**o+(-0.00289799815652)*x[1]+(0.153392291113)*x[2]**o+(-0.61448940993)*x[2]
        arg[1,4,2,1]=(0.0897863401674)*x[0]**o+(-0.487876787706)*x[0]+(-0.464008634014)*x[1]**o+(0.576466896344)*x[1]+(0.962345361537)*x[2]**o+(0.580878819156)*x[2]
        arg[2,0,0,0]=(0.316582275434)*x[0]**o+(0.0607872538221)*x[0]+(0.642492366848)*x[1]**o+(0.200300167843)*x[1]+(-0.402802396184)*x[2]**o+(-0.834254866233)*x[2]
        arg[2,0,0,1]=(0.644614184272)*x[0]**o+(-0.086367798694)*x[0]+(-0.428671462705)*x[1]**o+(-0.517827373575)*x[1]+(0.528197945165)*x[2]**o+(-0.523092888005)*x[2]
        arg[2,0,1,0]=(0.41376145869)*x[0]**o+(-0.27425322418)*x[0]+(0.816942168928)*x[1]**o+(-0.882784454814)*x[1]+(-0.270205535673)*x[2]**o+(0.642372102899)*x[2]
        arg[2,0,1,1]=(0.584931736844)*x[0]**o+(0.622511792702)*x[0]+(0.76174625736)*x[1]**o+(0.0561486448557)*x[1]+(-0.588861269552)*x[2]**o+(-0.341896323569)*x[2]
        arg[2,0,2,0]=(0.635215108566)*x[0]**o+(0.299769133671)*x[0]+(0.141053572388)*x[1]**o+(0.102844255879)*x[1]+(-0.995154777534)*x[2]**o+(0.501319388994)*x[2]
        arg[2,0,2,1]=(0.750089643836)*x[0]**o+(0.57570708561)*x[0]+(-0.449935823319)*x[1]**o+(-0.902982109411)*x[1]+(0.561185740222)*x[2]**o+(0.422654056847)*x[2]
        arg[2,1,0,0]=(0.953986415316)*x[0]**o+(-0.7883326473)*x[0]+(-0.723690234059)*x[1]**o+(-0.537242781471)*x[1]+(-0.503466717519)*x[2]**o+(-0.777865169241)*x[2]
        arg[2,1,0,1]=(0.317357554441)*x[0]**o+(-0.326828957855)*x[0]+(-0.609992651237)*x[1]**o+(0.811066934743)*x[1]+(-0.90668441447)*x[2]**o+(0.92581926687)*x[2]
        arg[2,1,1,0]=(-0.0801236189368)*x[0]**o+(0.49151264813)*x[0]+(-0.0770654446151)*x[1]**o+(-0.314981217276)*x[1]+(-0.743180837478)*x[2]**o+(-0.554491413146)*x[2]
        arg[2,1,1,1]=(0.831017344937)*x[0]**o+(-0.61064793935)*x[0]+(0.652976737027)*x[1]**o+(0.242104127969)*x[1]+(0.593330958678)*x[2]**o+(-0.35899584298)*x[2]
        arg[2,1,2,0]=(-0.580618798556)*x[0]**o+(-0.936176474082)*x[0]+(-0.768716477288)*x[1]**o+(-0.353687473159)*x[1]+(-0.261509314019)*x[2]**o+(-0.442139199257)*x[2]
        arg[2,1,2,1]=(0.945760793889)*x[0]**o+(-0.646958670891)*x[0]+(-0.440341726063)*x[1]**o+(0.223283751014)*x[1]+(0.799798944743)*x[2]**o+(0.591902067633)*x[2]
        arg[2,2,0,0]=(-0.267699746405)*x[0]**o+(0.271554172399)*x[0]+(0.747278903642)*x[1]**o+(-0.832877459834)*x[1]+(-0.640373164716)*x[2]**o+(-0.666487164763)*x[2]
        arg[2,2,0,1]=(0.795630715924)*x[0]**o+(0.682973866186)*x[0]+(-0.750818874926)*x[1]**o+(-0.904782408217)*x[1]+(-0.578138718379)*x[2]**o+(-0.369846516992)*x[2]
        arg[2,2,1,0]=(0.165654497062)*x[0]**o+(-0.329827871165)*x[0]+(-0.808981530672)*x[1]**o+(0.350106763785)*x[1]+(0.106638087689)*x[2]**o+(0.425516820398)*x[2]
        arg[2,2,1,1]=(0.017419119421)*x[0]**o+(0.500359047996)*x[0]+(-0.888509940049)*x[1]**o+(-0.375469124891)*x[1]+(0.863718970618)*x[2]**o+(-0.807667604763)*x[2]
        arg[2,2,2,0]=(0.860662652165)*x[0]**o+(0.162530730499)*x[0]+(-0.657510966961)*x[1]**o+(0.215905840286)*x[1]+(0.97380371264)*x[2]**o+(-0.431027671897)*x[2]
        arg[2,2,2,1]=(-0.649151503991)*x[0]**o+(0.484589779517)*x[0]+(0.905080495553)*x[1]**o+(0.997504119204)*x[1]+(-0.4268802512)*x[2]**o+(-0.687236308612)*x[2]
        arg[2,3,0,0]=(-0.60485902589)*x[0]**o+(0.49500777286)*x[0]+(0.33977532104)*x[1]**o+(-0.219209845338)*x[1]+(-0.197311627683)*x[2]**o+(0.285859591044)*x[2]
        arg[2,3,0,1]=(0.117438340483)*x[0]**o+(-0.562954894877)*x[0]+(-0.302982842797)*x[1]**o+(0.977923453436)*x[1]+(0.300127749784)*x[2]**o+(0.45784201373)*x[2]
        arg[2,3,1,0]=(-0.910061442266)*x[0]**o+(0.723114553174)*x[0]+(-0.435693478652)*x[1]**o+(-0.584326584996)*x[1]+(-0.51000911177)*x[2]**o+(0.506615483889)*x[2]
        arg[2,3,1,1]=(-0.50179741763)*x[0]**o+(-0.954551290001)*x[0]+(0.375839266192)*x[1]**o+(0.122120198059)*x[1]+(-0.593050674641)*x[2]**o+(0.549989594312)*x[2]
        arg[2,3,2,0]=(-0.889519666705)*x[0]**o+(-0.815572507674)*x[0]+(0.453677651057)*x[1]**o+(0.450910828217)*x[1]+(-0.00465373008682)*x[2]**o+(0.731899900165)*x[2]
        arg[2,3,2,1]=(-0.969971151589)*x[0]**o+(-0.156948104522)*x[0]+(0.492885029353)*x[1]**o+(-0.0310677847833)*x[1]+(0.327614233837)*x[2]**o+(-0.274974040707)*x[2]
        arg[2,4,0,0]=(0.844667900826)*x[0]**o+(0.863374870522)*x[0]+(0.251456923241)*x[1]**o+(0.434916005183)*x[1]+(-0.139023370677)*x[2]**o+(0.0655176616053)*x[2]
        arg[2,4,0,1]=(0.831755059318)*x[0]**o+(0.0897642674136)*x[0]+(0.190414878812)*x[1]**o+(0.976565327031)*x[1]+(0.413321767371)*x[2]**o+(0.755395415341)*x[2]
        arg[2,4,1,0]=(-0.790364876377)*x[0]**o+(0.829442094248)*x[0]+(0.556251403891)*x[1]**o+(-0.953487711658)*x[1]+(-0.610781905401)*x[2]**o+(-0.320764072326)*x[2]
        arg[2,4,1,1]=(-0.757101565136)*x[0]**o+(0.789340839395)*x[0]+(-0.615189514438)*x[1]**o+(-0.910992676978)*x[1]+(0.693736813401)*x[2]**o+(0.0538707588266)*x[2]
        arg[2,4,2,0]=(0.000335625616611)*x[0]**o+(-0.195171446476)*x[0]+(0.0281149133705)*x[1]**o+(-0.362294159243)*x[1]+(0.830520288743)*x[2]**o+(-0.431967694247)*x[2]
        arg[2,4,2,1]=(-0.232668364843)*x[0]**o+(0.655302452302)*x[0]+(-0.0125806408337)*x[1]**o+(-0.103979202109)*x[1]+(-0.277450383033)*x[2]**o+(0.708914971885)*x[2]
        arg[3,0,0,0]=(-0.813447875186)*x[0]**o+(0.133491078101)*x[0]+(-0.675510725453)*x[1]**o+(-0.161489202717)*x[1]+(0.0474778506607)*x[2]**o+(0.513729029652)*x[2]
        arg[3,0,0,1]=(0.45388386743)*x[0]**o+(-0.909074944715)*x[0]+(-0.540576382376)*x[1]**o+(-0.849753706068)*x[1]+(-0.496320191697)*x[2]**o+(-0.930725196363)*x[2]
        arg[3,0,1,0]=(-0.145526830518)*x[0]**o+(-0.34717641097)*x[0]+(-0.818886792022)*x[1]**o+(0.574315644784)*x[1]+(-0.780431329222)*x[2]**o+(0.892840747651)*x[2]
        arg[3,0,1,1]=(-0.415342962745)*x[0]**o+(-0.899768945518)*x[0]+(0.798573910731)*x[1]**o+(-0.0652893933479)*x[1]+(-0.444459519602)*x[2]**o+(-0.0100131702565)*x[2]
        arg[3,0,2,0]=(-0.997936676013)*x[0]**o+(0.822000858069)*x[0]+(-0.553794408949)*x[1]**o+(0.829037287423)*x[1]+(-0.700288437806)*x[2]**o+(-0.449584600891)*x[2]
        arg[3,0,2,1]=(-0.70979868598)*x[0]**o+(0.689868687655)*x[0]+(0.151257146397)*x[1]**o+(0.32895215144)*x[1]+(0.286043871383)*x[2]**o+(-0.604421567781)*x[2]
        arg[3,1,0,0]=(0.60431849542)*x[0]**o+(0.518618291627)*x[0]+(-0.471868097013)*x[1]**o+(-0.639110899151)*x[1]+(-0.0904184061601)*x[2]**o+(0.113762528304)*x[2]
        arg[3,1,0,1]=(0.307624103157)*x[0]**o+(-0.388281302162)*x[0]+(0.772429578102)*x[1]**o+(-0.197799629585)*x[1]+(-0.149120667036)*x[2]**o+(0.830154780264)*x[2]
        arg[3,1,1,0]=(0.957127413133)*x[0]**o+(0.730080945389)*x[0]+(0.497191948774)*x[1]**o+(-0.586439221663)*x[1]+(-0.56223071093)*x[2]**o+(0.162899395804)*x[2]
        arg[3,1,1,1]=(-0.285740818451)*x[0]**o+(-0.573919970411)*x[0]+(-0.400833211669)*x[1]**o+(0.84414534634)*x[1]+(-0.124352096731)*x[2]**o+(0.907013800834)*x[2]
        arg[3,1,2,0]=(0.186947347989)*x[0]**o+(0.499582449213)*x[0]+(0.525473636695)*x[1]**o+(-0.926500953998)*x[1]+(-0.0392615366415)*x[2]**o+(0.909002911192)*x[2]
        arg[3,1,2,1]=(0.634350440683)*x[0]**o+(0.901699291665)*x[0]+(-0.683899667575)*x[1]**o+(0.411642037744)*x[1]+(-0.72791611585)*x[2]**o+(-0.013293689839)*x[2]
        arg[3,2,0,0]=(0.185164429507)*x[0]**o+(-0.671011652257)*x[0]+(0.973361040829)*x[1]**o+(0.989738565899)*x[1]+(-0.501245887091)*x[2]**o+(-0.405445444386)*x[2]
        arg[3,2,0,1]=(-0.311554020884)*x[0]**o+(-0.929020990333)*x[0]+(0.366016375797)*x[1]**o+(0.838234804773)*x[1]+(0.939602632409)*x[2]**o+(-0.885524260372)*x[2]
        arg[3,2,1,0]=(-0.0101604655589)*x[0]**o+(0.670274078493)*x[0]+(0.198135855354)*x[1]**o+(-0.000275422963644)*x[1]+(0.136674233404)*x[2]**o+(-0.442510118497)*x[2]
        arg[3,2,1,1]=(-0.677948769785)*x[0]**o+(-0.165403610813)*x[0]+(-0.219353502981)*x[1]**o+(0.664297154009)*x[1]+(0.158536302892)*x[2]**o+(0.293940884925)*x[2]
        arg[3,2,2,0]=(0.666669534708)*x[0]**o+(0.609459298502)*x[0]+(0.297058070582)*x[1]**o+(-0.0809555825821)*x[1]+(0.139430791042)*x[2]**o+(-0.253514861624)*x[2]
        arg[3,2,2,1]=(-0.577932413274)*x[0]**o+(-0.583353108895)*x[0]+(-0.839262034394)*x[1]**o+(0.795642236106)*x[1]+(0.0882627381543)*x[2]**o+(-0.382970623122)*x[2]
        arg[3,3,0,0]=(0.187941627429)*x[0]**o+(0.552040688531)*x[0]+(0.837458832716)*x[1]**o+(0.296793121897)*x[1]+(-0.665699511216)*x[2]**o+(-0.318565991052)*x[2]
        arg[3,3,0,1]=(-0.78770176804)*x[0]**o+(0.807509260396)*x[0]+(0.46241086085)*x[1]**o+(0.668506908751)*x[1]+(-0.988954457343)*x[2]**o+(0.0896365679915)*x[2]
        arg[3,3,1,0]=(0.727665746433)*x[0]**o+(-0.819391676739)*x[0]+(-0.118648934607)*x[1]**o+(-0.433862939031)*x[1]+(0.827788711313)*x[2]**o+(-0.615751446784)*x[2]
        arg[3,3,1,1]=(-0.00427901841482)*x[0]**o+(-0.38545713036)*x[0]+(0.570160083404)*x[1]**o+(-0.235505162795)*x[1]+(0.514578304371)*x[2]**o+(-0.152692113279)*x[2]
        arg[3,3,2,0]=(-0.396900613831)*x[0]**o+(0.161875314543)*x[0]+(0.28471295984)*x[1]**o+(-0.623276596728)*x[1]+(-0.748772373219)*x[2]**o+(0.584246898704)*x[2]
        arg[3,3,2,1]=(0.210026027009)*x[0]**o+(-0.530437530387)*x[0]+(0.437155825716)*x[1]**o+(0.984102407204)*x[1]+(0.125030773605)*x[2]**o+(-0.821764907318)*x[2]
        arg[3,4,0,0]=(-0.819129946783)*x[0]**o+(-0.729517565451)*x[0]+(-0.126013244602)*x[1]**o+(0.592832125254)*x[1]+(-0.850426487488)*x[2]**o+(0.271550912174)*x[2]
        arg[3,4,0,1]=(-0.573554440921)*x[0]**o+(0.232106829584)*x[0]+(0.810201614042)*x[1]**o+(-0.85518164211)*x[1]+(0.249535563949)*x[2]**o+(-0.69728963017)*x[2]
        arg[3,4,1,0]=(0.785661643628)*x[0]**o+(0.936424600424)*x[0]+(0.321363811055)*x[1]**o+(0.362341458592)*x[1]+(-0.593576579363)*x[2]**o+(0.0601278435924)*x[2]
        arg[3,4,1,1]=(-0.854420686885)*x[0]**o+(0.993352097501)*x[0]+(-0.572116112224)*x[1]**o+(0.308008845182)*x[1]+(0.0443339578639)*x[2]**o+(-0.1469857285)*x[2]
        arg[3,4,2,0]=(0.696373642151)*x[0]**o+(0.82425126832)*x[0]+(-0.0891735358455)*x[1]**o+(-0.22218993613)*x[1]+(-0.238393149017)*x[2]**o+(-0.631352718194)*x[2]
        arg[3,4,2,1]=(-0.0932690431928)*x[0]**o+(-0.391900483919)*x[0]+(0.236843565361)*x[1]**o+(-0.164197648532)*x[1]+(0.235502596353)*x[2]**o+(0.0110131867917)*x[2]
        ref[0,0,0,0]=(0.349437323395)*x_ref[0]**o+(0.817821891642)*x_ref[0]+(-0.357998520848)*x_ref[1]**o+(0.349311985429)*x_ref[1]+(0.470351907942)*x_ref[2]**o+(0.264007843212)*x_ref[2]
        ref[0,0,0,1]=(-0.747824148035)*x_ref[0]**o+(0.551817215468)*x_ref[0]+(0.980476067966)*x_ref[1]**o+(0.521120126285)*x_ref[1]+(-0.472357720779)*x_ref[2]**o+(-0.379502066444)*x_ref[2]
        ref[0,0,1,0]=(0.302984899868)*x_ref[0]**o+(0.556601505796)*x_ref[0]+(-0.698948476338)*x_ref[1]**o+(-0.587379572024)*x_ref[1]+(-0.625760359909)*x_ref[2]**o+(0.231956543432)*x_ref[2]
        ref[0,0,1,1]=(-0.335943560463)*x_ref[0]**o+(-0.497643152155)*x_ref[0]+(-0.849829024888)*x_ref[1]**o+(-0.388751135318)*x_ref[1]+(-0.517696637743)*x_ref[2]**o+(-0.966210901)*x_ref[2]
        ref[0,0,2,0]=(-0.0953058405321)*x_ref[0]**o+(-0.864580918175)*x_ref[0]+(-0.732086619685)*x_ref[1]**o+(0.872044654816)*x_ref[1]+(-0.225850391375)*x_ref[2]**o+(-0.502203654371)*x_ref[2]
        ref[0,0,2,1]=(0.228408069101)*x_ref[0]**o+(0.267590362458)*x_ref[0]+(0.127397041031)*x_ref[1]**o+(0.0142393720782)*x_ref[1]+(-0.257563412478)*x_ref[2]**o+(0.0211877877507)*x_ref[2]
        ref[0,1,0,0]=(0.921125039994)*x_ref[0]**o+(0.372548304521)*x_ref[0]+(-0.66574364308)*x_ref[1]**o+(0.0895600219155)*x_ref[1]+(-0.656870562037)*x_ref[2]**o+(-0.608819638268)*x_ref[2]
        ref[0,1,0,1]=(-0.396625252579)*x_ref[0]**o+(0.655542320616)*x_ref[0]+(-0.542629502059)*x_ref[1]**o+(0.587317299384)*x_ref[1]+(0.96380848967)*x_ref[2]**o+(0.121341827766)*x_ref[2]
        ref[0,1,1,0]=(0.0908208346003)*x_ref[0]**o+(-0.0398227321951)*x_ref[0]+(0.907607555288)*x_ref[1]**o+(0.527230589013)*x_ref[1]+(0.519287416653)*x_ref[2]**o+(0.103180358249)*x_ref[2]
        ref[0,1,1,1]=(0.0615195625837)*x_ref[0]**o+(0.695720276474)*x_ref[0]+(0.0862101684049)*x_ref[1]**o+(0.517150419955)*x_ref[1]+(0.511207669575)*x_ref[2]**o+(-0.9556472362)*x_ref[2]
        ref[0,1,2,0]=(0.70040395652)*x_ref[0]**o+(-0.576266027865)*x_ref[0]+(-0.133947982034)*x_ref[1]**o+(-0.764976368456)*x_ref[1]+(-0.860780506782)*x_ref[2]**o+(-0.137946340663)*x_ref[2]
        ref[0,1,2,1]=(0.356017596832)*x_ref[0]**o+(0.94912783604)*x_ref[0]+(0.325951760294)*x_ref[1]**o+(0.0623695610721)*x_ref[1]+(-0.106031051895)*x_ref[2]**o+(-0.339314970581)*x_ref[2]
        ref[0,2,0,0]=(-0.124954395016)*x_ref[0]**o+(0.558233271782)*x_ref[0]+(-0.956846820394)*x_ref[1]**o+(-0.599332648704)*x_ref[1]+(-0.322444201477)*x_ref[2]**o+(0.0833108732532)*x_ref[2]
        ref[0,2,0,1]=(-0.713065259589)*x_ref[0]**o+(-0.597746726477)*x_ref[0]+(-0.976654446591)*x_ref[1]**o+(0.977779698697)*x_ref[1]+(0.694427679422)*x_ref[2]**o+(-0.364463986712)*x_ref[2]
        ref[0,2,1,0]=(-0.415861468352)*x_ref[0]**o+(-0.159802964908)*x_ref[0]+(-0.132331280778)*x_ref[1]**o+(0.120829594848)*x_ref[1]+(-0.114181951611)*x_ref[2]**o+(-0.810187313564)*x_ref[2]
        ref[0,2,1,1]=(-0.869925695451)*x_ref[0]**o+(-0.535444611026)*x_ref[0]+(0.62908837594)*x_ref[1]**o+(-0.488583607314)*x_ref[1]+(0.8461615049)*x_ref[2]**o+(-0.558010294727)*x_ref[2]
        ref[0,2,2,0]=(-0.0822232217466)*x_ref[0]**o+(-0.673631382141)*x_ref[0]+(0.803717792107)*x_ref[1]**o+(0.339934715615)*x_ref[1]+(-0.886365924996)*x_ref[2]**o+(0.0206517750753)*x_ref[2]
        ref[0,2,2,1]=(0.367656249869)*x_ref[0]**o+(0.000987963028539)*x_ref[0]+(0.399182215451)*x_ref[1]**o+(0.923259635138)*x_ref[1]+(0.8573338122)*x_ref[2]**o+(0.679082213415)*x_ref[2]
        ref[0,3,0,0]=(-0.722680496644)*x_ref[0]**o+(0.199427912651)*x_ref[0]+(-0.351799245099)*x_ref[1]**o+(0.240285416231)*x_ref[1]+(-0.701704868242)*x_ref[2]**o+(0.265610436018)*x_ref[2]
        ref[0,3,0,1]=(-0.14479043328)*x_ref[0]**o+(-0.192670959504)*x_ref[0]+(0.327412080296)*x_ref[1]**o+(-0.390894637576)*x_ref[1]+(-0.401862813677)*x_ref[2]**o+(-0.87150071561)*x_ref[2]
        ref[0,3,1,0]=(0.430039400239)*x_ref[0]**o+(-0.0691443536291)*x_ref[0]+(0.205492039071)*x_ref[1]**o+(-0.391580258314)*x_ref[1]+(0.791748125642)*x_ref[2]**o+(-0.0896722225431)*x_ref[2]
        ref[0,3,1,1]=(-0.177554766913)*x_ref[0]**o+(-0.399389195454)*x_ref[0]+(-0.304228953477)*x_ref[1]**o+(0.00462595697167)*x_ref[1]+(0.271814579884)*x_ref[2]**o+(-0.17305639368)*x_ref[2]
        ref[0,3,2,0]=(-0.105869589157)*x_ref[0]**o+(0.803488342924)*x_ref[0]+(-0.65117861655)*x_ref[1]**o+(-0.27529059731)*x_ref[1]+(-0.777107387582)*x_ref[2]**o+(-0.690980328369)*x_ref[2]
        ref[0,3,2,1]=(-0.13792230295)*x_ref[0]**o+(0.649721420509)*x_ref[0]+(0.401441756822)*x_ref[1]**o+(0.468949007973)*x_ref[1]+(-0.762015699412)*x_ref[2]**o+(0.389781367434)*x_ref[2]
        ref[0,4,0,0]=(0.873653819021)*x_ref[0]**o+(0.539406798913)*x_ref[0]+(0.802177801924)*x_ref[1]**o+(-0.861314989922)*x_ref[1]+(0.847509960118)*x_ref[2]**o+(-0.41381965291)*x_ref[2]
        ref[0,4,0,1]=(0.0416858647633)*x_ref[0]**o+(0.866825467268)*x_ref[0]+(0.89776163679)*x_ref[1]**o+(-0.978326218858)*x_ref[1]+(-0.409795413277)*x_ref[2]**o+(0.982449327977)*x_ref[2]
        ref[0,4,1,0]=(-0.114530966075)*x_ref[0]**o+(-0.182414280034)*x_ref[0]+(0.154385814218)*x_ref[1]**o+(0.869624332271)*x_ref[1]+(0.477213291128)*x_ref[2]**o+(-0.137123375612)*x_ref[2]
        ref[0,4,1,1]=(-0.52963019184)*x_ref[0]**o+(0.0217955878025)*x_ref[0]+(-0.174531798693)*x_ref[1]**o+(-0.738947178141)*x_ref[1]+(0.198130667502)*x_ref[2]**o+(-0.449952166994)*x_ref[2]
        ref[0,4,2,0]=(0.998209708224)*x_ref[0]**o+(0.734906841374)*x_ref[0]+(0.225596363429)*x_ref[1]**o+(0.862772101878)*x_ref[1]+(-0.133563132638)*x_ref[2]**o+(-0.78059821331)*x_ref[2]
        ref[0,4,2,1]=(0.384580893901)*x_ref[0]**o+(-0.881918368669)*x_ref[0]+(-0.950768236608)*x_ref[1]**o+(-0.452275072093)*x_ref[1]+(0.592779023744)*x_ref[2]**o+(0.798735695084)*x_ref[2]
        ref[1,0,0,0]=(0.786827894179)*x_ref[0]**o+(-0.233241290567)*x_ref[0]+(0.946717698992)*x_ref[1]**o+(0.179368479607)*x_ref[1]+(-0.10409968587)*x_ref[2]**o+(-0.770115022756)*x_ref[2]
        ref[1,0,0,1]=(-0.355389699793)*x_ref[0]**o+(0.356407960494)*x_ref[0]+(-0.371831767713)*x_ref[1]**o+(0.139120259823)*x_ref[1]+(0.910761795881)*x_ref[2]**o+(0.725966050775)*x_ref[2]
        ref[1,0,1,0]=(0.316372320948)*x_ref[0]**o+(0.432472869369)*x_ref[0]+(0.733187356771)*x_ref[1]**o+(0.133915829584)*x_ref[1]+(0.380713482112)*x_ref[2]**o+(-0.874827460114)*x_ref[2]
        ref[1,0,1,1]=(-0.0625786527083)*x_ref[0]**o+(-0.469518973019)*x_ref[0]+(-0.318610570327)*x_ref[1]**o+(-0.304412247656)*x_ref[1]+(0.844057224107)*x_ref[2]**o+(0.508926494536)*x_ref[2]
        ref[1,0,2,0]=(-0.759697281694)*x_ref[0]**o+(0.701456398418)*x_ref[0]+(-0.671612899124)*x_ref[1]**o+(0.292474197717)*x_ref[1]+(-0.445886723952)*x_ref[2]**o+(-0.319953406782)*x_ref[2]
        ref[1,0,2,1]=(-0.865703619781)*x_ref[0]**o+(0.115639820869)*x_ref[0]+(0.493634056223)*x_ref[1]**o+(0.0343245670418)*x_ref[1]+(0.748200872819)*x_ref[2]**o+(-0.208438395054)*x_ref[2]
        ref[1,1,0,0]=(0.767661930656)*x_ref[0]**o+(-0.578385615919)*x_ref[0]+(0.241313268557)*x_ref[1]**o+(0.74276605359)*x_ref[1]+(0.373638986644)*x_ref[2]**o+(0.475479466784)*x_ref[2]
        ref[1,1,0,1]=(0.723228882826)*x_ref[0]**o+(-0.988524686322)*x_ref[0]+(0.333664096597)*x_ref[1]**o+(-0.756435323592)*x_ref[1]+(0.771173510221)*x_ref[2]**o+(-0.879337556561)*x_ref[2]
        ref[1,1,1,0]=(0.939682105701)*x_ref[0]**o+(-0.172311578817)*x_ref[0]+(-0.407241492831)*x_ref[1]**o+(0.446660064834)*x_ref[1]+(0.0371203367053)*x_ref[2]**o+(-0.302853430561)*x_ref[2]
        ref[1,1,1,1]=(0.1093213068)*x_ref[0]**o+(0.31232933407)*x_ref[0]+(0.47474581408)*x_ref[1]**o+(-0.602763201317)*x_ref[1]+(0.803198523872)*x_ref[2]**o+(-0.178088233667)*x_ref[2]
        ref[1,1,2,0]=(-0.217141686114)*x_ref[0]**o+(0.287363815362)*x_ref[0]+(-0.352552947551)*x_ref[1]**o+(0.106445580983)*x_ref[1]+(0.340814562712)*x_ref[2]**o+(-0.963691249733)*x_ref[2]
        ref[1,1,2,1]=(0.612302241126)*x_ref[0]**o+(-0.771765095612)*x_ref[0]+(-0.18662763339)*x_ref[1]**o+(0.10259882275)*x_ref[1]+(-0.29535785913)*x_ref[2]**o+(-0.577876877465)*x_ref[2]
        ref[1,2,0,0]=(0.497131459554)*x_ref[0]**o+(0.912133134267)*x_ref[0]+(0.944369633113)*x_ref[1]**o+(0.163408823815)*x_ref[1]+(0.85636766551)*x_ref[2]**o+(-0.00282458078743)*x_ref[2]
        ref[1,2,0,1]=(0.362490806338)*x_ref[0]**o+(0.22490192994)*x_ref[0]+(-0.913665309243)*x_ref[1]**o+(-0.88010395132)*x_ref[1]+(-0.926205930269)*x_ref[2]**o+(-0.231830082111)*x_ref[2]
        ref[1,2,1,0]=(-0.974792784505)*x_ref[0]**o+(-0.765605618923)*x_ref[0]+(0.241470054253)*x_ref[1]**o+(0.445863114014)*x_ref[1]+(0.561062749944)*x_ref[2]**o+(-0.215674406206)*x_ref[2]
        ref[1,2,1,1]=(0.827013456823)*x_ref[0]**o+(-0.191864800977)*x_ref[0]+(-0.158099224794)*x_ref[1]**o+(-0.921959562711)*x_ref[1]+(0.930693535629)*x_ref[2]**o+(-0.352967183792)*x_ref[2]
        ref[1,2,2,0]=(0.258970603305)*x_ref[0]**o+(0.586779837275)*x_ref[0]+(0.82707087137)*x_ref[1]**o+(-0.215210535461)*x_ref[1]+(0.36890299705)*x_ref[2]**o+(-0.724385890088)*x_ref[2]
        ref[1,2,2,1]=(-0.125969269863)*x_ref[0]**o+(-0.505216539273)*x_ref[0]+(0.43242008053)*x_ref[1]**o+(0.212079268112)*x_ref[1]+(0.345759190477)*x_ref[2]**o+(-0.98505468199)*x_ref[2]
        ref[1,3,0,0]=(-0.647118215098)*x_ref[0]**o+(0.318566502446)*x_ref[0]+(-0.151406735047)*x_ref[1]**o+(0.253202745125)*x_ref[1]+(-0.404823111329)*x_ref[2]**o+(-0.520797366557)*x_ref[2]
        ref[1,3,0,1]=(-0.946886311645)*x_ref[0]**o+(0.733699524424)*x_ref[0]+(-0.615470980807)*x_ref[1]**o+(-0.103976314414)*x_ref[1]+(-0.932104431883)*x_ref[2]**o+(0.157563806559)*x_ref[2]
        ref[1,3,1,0]=(-0.289471503181)*x_ref[0]**o+(0.749124853135)*x_ref[0]+(-0.578861403092)*x_ref[1]**o+(-0.136231378985)*x_ref[1]+(0.532719246315)*x_ref[2]**o+(0.542455831657)*x_ref[2]
        ref[1,3,1,1]=(-0.343842490684)*x_ref[0]**o+(0.110039182926)*x_ref[0]+(0.901234650534)*x_ref[1]**o+(-0.513362876241)*x_ref[1]+(0.371952189126)*x_ref[2]**o+(-0.169400349862)*x_ref[2]
        ref[1,3,2,0]=(-0.766791944503)*x_ref[0]**o+(0.0267491122166)*x_ref[0]+(-0.699293509597)*x_ref[1]**o+(-0.697363669971)*x_ref[1]+(0.567854170308)*x_ref[2]**o+(-0.516301389629)*x_ref[2]
        ref[1,3,2,1]=(-0.288508579358)*x_ref[0]**o+(0.239321880772)*x_ref[0]+(0.677249635795)*x_ref[1]**o+(0.48369084768)*x_ref[1]+(0.681460633595)*x_ref[2]**o+(-0.80218734437)*x_ref[2]
        ref[1,4,0,0]=(0.510273388936)*x_ref[0]**o+(0.35600020965)*x_ref[0]+(0.271077500616)*x_ref[1]**o+(-0.00479335810576)*x_ref[1]+(-0.348457583464)*x_ref[2]**o+(-0.965895122152)*x_ref[2]
        ref[1,4,0,1]=(0.0134532748813)*x_ref[0]**o+(-0.396954380686)*x_ref[0]+(-0.176682002946)*x_ref[1]**o+(0.971224261138)*x_ref[1]+(-0.359602180261)*x_ref[2]**o+(0.373748703363)*x_ref[2]
        ref[1,4,1,0]=(-0.519413160925)*x_ref[0]**o+(0.856346255965)*x_ref[0]+(0.159167243935)*x_ref[1]**o+(0.92068779838)*x_ref[1]+(0.415122182947)*x_ref[2]**o+(-0.595702311496)*x_ref[2]
        ref[1,4,1,1]=(-0.608969598184)*x_ref[0]**o+(-0.434750765676)*x_ref[0]+(0.74471475701)*x_ref[1]**o+(-0.229333246615)*x_ref[1]+(-0.678138032736)*x_ref[2]**o+(-0.514112926798)*x_ref[2]
        ref[1,4,2,0]=(0.0633453561735)*x_ref[0]**o+(-0.278569473082)*x_ref[0]+(-0.842227144473)*x_ref[1]**o+(-0.00289799815652)*x_ref[1]+(0.153392291113)*x_ref[2]**o+(-0.61448940993)*x_ref[2]
        ref[1,4,2,1]=(0.0897863401674)*x_ref[0]**o+(-0.487876787706)*x_ref[0]+(-0.464008634014)*x_ref[1]**o+(0.576466896344)*x_ref[1]+(0.962345361537)*x_ref[2]**o+(0.580878819156)*x_ref[2]
        ref[2,0,0,0]=(0.316582275434)*x_ref[0]**o+(0.0607872538221)*x_ref[0]+(0.642492366848)*x_ref[1]**o+(0.200300167843)*x_ref[1]+(-0.402802396184)*x_ref[2]**o+(-0.834254866233)*x_ref[2]
        ref[2,0,0,1]=(0.644614184272)*x_ref[0]**o+(-0.086367798694)*x_ref[0]+(-0.428671462705)*x_ref[1]**o+(-0.517827373575)*x_ref[1]+(0.528197945165)*x_ref[2]**o+(-0.523092888005)*x_ref[2]
        ref[2,0,1,0]=(0.41376145869)*x_ref[0]**o+(-0.27425322418)*x_ref[0]+(0.816942168928)*x_ref[1]**o+(-0.882784454814)*x_ref[1]+(-0.270205535673)*x_ref[2]**o+(0.642372102899)*x_ref[2]
        ref[2,0,1,1]=(0.584931736844)*x_ref[0]**o+(0.622511792702)*x_ref[0]+(0.76174625736)*x_ref[1]**o+(0.0561486448557)*x_ref[1]+(-0.588861269552)*x_ref[2]**o+(-0.341896323569)*x_ref[2]
        ref[2,0,2,0]=(0.635215108566)*x_ref[0]**o+(0.299769133671)*x_ref[0]+(0.141053572388)*x_ref[1]**o+(0.102844255879)*x_ref[1]+(-0.995154777534)*x_ref[2]**o+(0.501319388994)*x_ref[2]
        ref[2,0,2,1]=(0.750089643836)*x_ref[0]**o+(0.57570708561)*x_ref[0]+(-0.449935823319)*x_ref[1]**o+(-0.902982109411)*x_ref[1]+(0.561185740222)*x_ref[2]**o+(0.422654056847)*x_ref[2]
        ref[2,1,0,0]=(0.953986415316)*x_ref[0]**o+(-0.7883326473)*x_ref[0]+(-0.723690234059)*x_ref[1]**o+(-0.537242781471)*x_ref[1]+(-0.503466717519)*x_ref[2]**o+(-0.777865169241)*x_ref[2]
        ref[2,1,0,1]=(0.317357554441)*x_ref[0]**o+(-0.326828957855)*x_ref[0]+(-0.609992651237)*x_ref[1]**o+(0.811066934743)*x_ref[1]+(-0.90668441447)*x_ref[2]**o+(0.92581926687)*x_ref[2]
        ref[2,1,1,0]=(-0.0801236189368)*x_ref[0]**o+(0.49151264813)*x_ref[0]+(-0.0770654446151)*x_ref[1]**o+(-0.314981217276)*x_ref[1]+(-0.743180837478)*x_ref[2]**o+(-0.554491413146)*x_ref[2]
        ref[2,1,1,1]=(0.831017344937)*x_ref[0]**o+(-0.61064793935)*x_ref[0]+(0.652976737027)*x_ref[1]**o+(0.242104127969)*x_ref[1]+(0.593330958678)*x_ref[2]**o+(-0.35899584298)*x_ref[2]
        ref[2,1,2,0]=(-0.580618798556)*x_ref[0]**o+(-0.936176474082)*x_ref[0]+(-0.768716477288)*x_ref[1]**o+(-0.353687473159)*x_ref[1]+(-0.261509314019)*x_ref[2]**o+(-0.442139199257)*x_ref[2]
        ref[2,1,2,1]=(0.945760793889)*x_ref[0]**o+(-0.646958670891)*x_ref[0]+(-0.440341726063)*x_ref[1]**o+(0.223283751014)*x_ref[1]+(0.799798944743)*x_ref[2]**o+(0.591902067633)*x_ref[2]
        ref[2,2,0,0]=(-0.267699746405)*x_ref[0]**o+(0.271554172399)*x_ref[0]+(0.747278903642)*x_ref[1]**o+(-0.832877459834)*x_ref[1]+(-0.640373164716)*x_ref[2]**o+(-0.666487164763)*x_ref[2]
        ref[2,2,0,1]=(0.795630715924)*x_ref[0]**o+(0.682973866186)*x_ref[0]+(-0.750818874926)*x_ref[1]**o+(-0.904782408217)*x_ref[1]+(-0.578138718379)*x_ref[2]**o+(-0.369846516992)*x_ref[2]
        ref[2,2,1,0]=(0.165654497062)*x_ref[0]**o+(-0.329827871165)*x_ref[0]+(-0.808981530672)*x_ref[1]**o+(0.350106763785)*x_ref[1]+(0.106638087689)*x_ref[2]**o+(0.425516820398)*x_ref[2]
        ref[2,2,1,1]=(0.017419119421)*x_ref[0]**o+(0.500359047996)*x_ref[0]+(-0.888509940049)*x_ref[1]**o+(-0.375469124891)*x_ref[1]+(0.863718970618)*x_ref[2]**o+(-0.807667604763)*x_ref[2]
        ref[2,2,2,0]=(0.860662652165)*x_ref[0]**o+(0.162530730499)*x_ref[0]+(-0.657510966961)*x_ref[1]**o+(0.215905840286)*x_ref[1]+(0.97380371264)*x_ref[2]**o+(-0.431027671897)*x_ref[2]
        ref[2,2,2,1]=(-0.649151503991)*x_ref[0]**o+(0.484589779517)*x_ref[0]+(0.905080495553)*x_ref[1]**o+(0.997504119204)*x_ref[1]+(-0.4268802512)*x_ref[2]**o+(-0.687236308612)*x_ref[2]
        ref[2,3,0,0]=(-0.60485902589)*x_ref[0]**o+(0.49500777286)*x_ref[0]+(0.33977532104)*x_ref[1]**o+(-0.219209845338)*x_ref[1]+(-0.197311627683)*x_ref[2]**o+(0.285859591044)*x_ref[2]
        ref[2,3,0,1]=(0.117438340483)*x_ref[0]**o+(-0.562954894877)*x_ref[0]+(-0.302982842797)*x_ref[1]**o+(0.977923453436)*x_ref[1]+(0.300127749784)*x_ref[2]**o+(0.45784201373)*x_ref[2]
        ref[2,3,1,0]=(-0.910061442266)*x_ref[0]**o+(0.723114553174)*x_ref[0]+(-0.435693478652)*x_ref[1]**o+(-0.584326584996)*x_ref[1]+(-0.51000911177)*x_ref[2]**o+(0.506615483889)*x_ref[2]
        ref[2,3,1,1]=(-0.50179741763)*x_ref[0]**o+(-0.954551290001)*x_ref[0]+(0.375839266192)*x_ref[1]**o+(0.122120198059)*x_ref[1]+(-0.593050674641)*x_ref[2]**o+(0.549989594312)*x_ref[2]
        ref[2,3,2,0]=(-0.889519666705)*x_ref[0]**o+(-0.815572507674)*x_ref[0]+(0.453677651057)*x_ref[1]**o+(0.450910828217)*x_ref[1]+(-0.00465373008682)*x_ref[2]**o+(0.731899900165)*x_ref[2]
        ref[2,3,2,1]=(-0.969971151589)*x_ref[0]**o+(-0.156948104522)*x_ref[0]+(0.492885029353)*x_ref[1]**o+(-0.0310677847833)*x_ref[1]+(0.327614233837)*x_ref[2]**o+(-0.274974040707)*x_ref[2]
        ref[2,4,0,0]=(0.844667900826)*x_ref[0]**o+(0.863374870522)*x_ref[0]+(0.251456923241)*x_ref[1]**o+(0.434916005183)*x_ref[1]+(-0.139023370677)*x_ref[2]**o+(0.0655176616053)*x_ref[2]
        ref[2,4,0,1]=(0.831755059318)*x_ref[0]**o+(0.0897642674136)*x_ref[0]+(0.190414878812)*x_ref[1]**o+(0.976565327031)*x_ref[1]+(0.413321767371)*x_ref[2]**o+(0.755395415341)*x_ref[2]
        ref[2,4,1,0]=(-0.790364876377)*x_ref[0]**o+(0.829442094248)*x_ref[0]+(0.556251403891)*x_ref[1]**o+(-0.953487711658)*x_ref[1]+(-0.610781905401)*x_ref[2]**o+(-0.320764072326)*x_ref[2]
        ref[2,4,1,1]=(-0.757101565136)*x_ref[0]**o+(0.789340839395)*x_ref[0]+(-0.615189514438)*x_ref[1]**o+(-0.910992676978)*x_ref[1]+(0.693736813401)*x_ref[2]**o+(0.0538707588266)*x_ref[2]
        ref[2,4,2,0]=(0.000335625616611)*x_ref[0]**o+(-0.195171446476)*x_ref[0]+(0.0281149133705)*x_ref[1]**o+(-0.362294159243)*x_ref[1]+(0.830520288743)*x_ref[2]**o+(-0.431967694247)*x_ref[2]
        ref[2,4,2,1]=(-0.232668364843)*x_ref[0]**o+(0.655302452302)*x_ref[0]+(-0.0125806408337)*x_ref[1]**o+(-0.103979202109)*x_ref[1]+(-0.277450383033)*x_ref[2]**o+(0.708914971885)*x_ref[2]
        ref[3,0,0,0]=(-0.813447875186)*x_ref[0]**o+(0.133491078101)*x_ref[0]+(-0.675510725453)*x_ref[1]**o+(-0.161489202717)*x_ref[1]+(0.0474778506607)*x_ref[2]**o+(0.513729029652)*x_ref[2]
        ref[3,0,0,1]=(0.45388386743)*x_ref[0]**o+(-0.909074944715)*x_ref[0]+(-0.540576382376)*x_ref[1]**o+(-0.849753706068)*x_ref[1]+(-0.496320191697)*x_ref[2]**o+(-0.930725196363)*x_ref[2]
        ref[3,0,1,0]=(-0.145526830518)*x_ref[0]**o+(-0.34717641097)*x_ref[0]+(-0.818886792022)*x_ref[1]**o+(0.574315644784)*x_ref[1]+(-0.780431329222)*x_ref[2]**o+(0.892840747651)*x_ref[2]
        ref[3,0,1,1]=(-0.415342962745)*x_ref[0]**o+(-0.899768945518)*x_ref[0]+(0.798573910731)*x_ref[1]**o+(-0.0652893933479)*x_ref[1]+(-0.444459519602)*x_ref[2]**o+(-0.0100131702565)*x_ref[2]
        ref[3,0,2,0]=(-0.997936676013)*x_ref[0]**o+(0.822000858069)*x_ref[0]+(-0.553794408949)*x_ref[1]**o+(0.829037287423)*x_ref[1]+(-0.700288437806)*x_ref[2]**o+(-0.449584600891)*x_ref[2]
        ref[3,0,2,1]=(-0.70979868598)*x_ref[0]**o+(0.689868687655)*x_ref[0]+(0.151257146397)*x_ref[1]**o+(0.32895215144)*x_ref[1]+(0.286043871383)*x_ref[2]**o+(-0.604421567781)*x_ref[2]
        ref[3,1,0,0]=(0.60431849542)*x_ref[0]**o+(0.518618291627)*x_ref[0]+(-0.471868097013)*x_ref[1]**o+(-0.639110899151)*x_ref[1]+(-0.0904184061601)*x_ref[2]**o+(0.113762528304)*x_ref[2]
        ref[3,1,0,1]=(0.307624103157)*x_ref[0]**o+(-0.388281302162)*x_ref[0]+(0.772429578102)*x_ref[1]**o+(-0.197799629585)*x_ref[1]+(-0.149120667036)*x_ref[2]**o+(0.830154780264)*x_ref[2]
        ref[3,1,1,0]=(0.957127413133)*x_ref[0]**o+(0.730080945389)*x_ref[0]+(0.497191948774)*x_ref[1]**o+(-0.586439221663)*x_ref[1]+(-0.56223071093)*x_ref[2]**o+(0.162899395804)*x_ref[2]
        ref[3,1,1,1]=(-0.285740818451)*x_ref[0]**o+(-0.573919970411)*x_ref[0]+(-0.400833211669)*x_ref[1]**o+(0.84414534634)*x_ref[1]+(-0.124352096731)*x_ref[2]**o+(0.907013800834)*x_ref[2]
        ref[3,1,2,0]=(0.186947347989)*x_ref[0]**o+(0.499582449213)*x_ref[0]+(0.525473636695)*x_ref[1]**o+(-0.926500953998)*x_ref[1]+(-0.0392615366415)*x_ref[2]**o+(0.909002911192)*x_ref[2]
        ref[3,1,2,1]=(0.634350440683)*x_ref[0]**o+(0.901699291665)*x_ref[0]+(-0.683899667575)*x_ref[1]**o+(0.411642037744)*x_ref[1]+(-0.72791611585)*x_ref[2]**o+(-0.013293689839)*x_ref[2]
        ref[3,2,0,0]=(0.185164429507)*x_ref[0]**o+(-0.671011652257)*x_ref[0]+(0.973361040829)*x_ref[1]**o+(0.989738565899)*x_ref[1]+(-0.501245887091)*x_ref[2]**o+(-0.405445444386)*x_ref[2]
        ref[3,2,0,1]=(-0.311554020884)*x_ref[0]**o+(-0.929020990333)*x_ref[0]+(0.366016375797)*x_ref[1]**o+(0.838234804773)*x_ref[1]+(0.939602632409)*x_ref[2]**o+(-0.885524260372)*x_ref[2]
        ref[3,2,1,0]=(-0.0101604655589)*x_ref[0]**o+(0.670274078493)*x_ref[0]+(0.198135855354)*x_ref[1]**o+(-0.000275422963644)*x_ref[1]+(0.136674233404)*x_ref[2]**o+(-0.442510118497)*x_ref[2]
        ref[3,2,1,1]=(-0.677948769785)*x_ref[0]**o+(-0.165403610813)*x_ref[0]+(-0.219353502981)*x_ref[1]**o+(0.664297154009)*x_ref[1]+(0.158536302892)*x_ref[2]**o+(0.293940884925)*x_ref[2]
        ref[3,2,2,0]=(0.666669534708)*x_ref[0]**o+(0.609459298502)*x_ref[0]+(0.297058070582)*x_ref[1]**o+(-0.0809555825821)*x_ref[1]+(0.139430791042)*x_ref[2]**o+(-0.253514861624)*x_ref[2]
        ref[3,2,2,1]=(-0.577932413274)*x_ref[0]**o+(-0.583353108895)*x_ref[0]+(-0.839262034394)*x_ref[1]**o+(0.795642236106)*x_ref[1]+(0.0882627381543)*x_ref[2]**o+(-0.382970623122)*x_ref[2]
        ref[3,3,0,0]=(0.187941627429)*x_ref[0]**o+(0.552040688531)*x_ref[0]+(0.837458832716)*x_ref[1]**o+(0.296793121897)*x_ref[1]+(-0.665699511216)*x_ref[2]**o+(-0.318565991052)*x_ref[2]
        ref[3,3,0,1]=(-0.78770176804)*x_ref[0]**o+(0.807509260396)*x_ref[0]+(0.46241086085)*x_ref[1]**o+(0.668506908751)*x_ref[1]+(-0.988954457343)*x_ref[2]**o+(0.0896365679915)*x_ref[2]
        ref[3,3,1,0]=(0.727665746433)*x_ref[0]**o+(-0.819391676739)*x_ref[0]+(-0.118648934607)*x_ref[1]**o+(-0.433862939031)*x_ref[1]+(0.827788711313)*x_ref[2]**o+(-0.615751446784)*x_ref[2]
        ref[3,3,1,1]=(-0.00427901841482)*x_ref[0]**o+(-0.38545713036)*x_ref[0]+(0.570160083404)*x_ref[1]**o+(-0.235505162795)*x_ref[1]+(0.514578304371)*x_ref[2]**o+(-0.152692113279)*x_ref[2]
        ref[3,3,2,0]=(-0.396900613831)*x_ref[0]**o+(0.161875314543)*x_ref[0]+(0.28471295984)*x_ref[1]**o+(-0.623276596728)*x_ref[1]+(-0.748772373219)*x_ref[2]**o+(0.584246898704)*x_ref[2]
        ref[3,3,2,1]=(0.210026027009)*x_ref[0]**o+(-0.530437530387)*x_ref[0]+(0.437155825716)*x_ref[1]**o+(0.984102407204)*x_ref[1]+(0.125030773605)*x_ref[2]**o+(-0.821764907318)*x_ref[2]
        ref[3,4,0,0]=(-0.819129946783)*x_ref[0]**o+(-0.729517565451)*x_ref[0]+(-0.126013244602)*x_ref[1]**o+(0.592832125254)*x_ref[1]+(-0.850426487488)*x_ref[2]**o+(0.271550912174)*x_ref[2]
        ref[3,4,0,1]=(-0.573554440921)*x_ref[0]**o+(0.232106829584)*x_ref[0]+(0.810201614042)*x_ref[1]**o+(-0.85518164211)*x_ref[1]+(0.249535563949)*x_ref[2]**o+(-0.69728963017)*x_ref[2]
        ref[3,4,1,0]=(0.785661643628)*x_ref[0]**o+(0.936424600424)*x_ref[0]+(0.321363811055)*x_ref[1]**o+(0.362341458592)*x_ref[1]+(-0.593576579363)*x_ref[2]**o+(0.0601278435924)*x_ref[2]
        ref[3,4,1,1]=(-0.854420686885)*x_ref[0]**o+(0.993352097501)*x_ref[0]+(-0.572116112224)*x_ref[1]**o+(0.308008845182)*x_ref[1]+(0.0443339578639)*x_ref[2]**o+(-0.1469857285)*x_ref[2]
        ref[3,4,2,0]=(0.696373642151)*x_ref[0]**o+(0.82425126832)*x_ref[0]+(-0.0891735358455)*x_ref[1]**o+(-0.22218993613)*x_ref[1]+(-0.238393149017)*x_ref[2]**o+(-0.631352718194)*x_ref[2]
        ref[3,4,2,1]=(-0.0932690431928)*x_ref[0]**o+(-0.391900483919)*x_ref[0]+(0.236843565361)*x_ref[1]**o+(-0.164197648532)*x_ref[1]+(0.235502596353)*x_ref[2]**o+(0.0110131867917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.521673635599)*x[0]**o+(-0.288664980886)*x[0]+(0.57565949318)*x[1]**o+(-0.555613874185)*x[1]
        ref=(-0.521673635599)*x_ref[0]**o+(-0.288664980886)*x_ref[0]+(0.57565949318)*x_ref[1]**o+(-0.555613874185)*x_ref[1]
      else:
        arg=(0.520429804914)*x[0]**o+(-0.000683147004841)*x[0]+(0.676879975826)*x[1]**o+(0.760707578765)*x[1]+(0.479225734268)*x[2]**o+(0.132006876687)*x[2]
        ref=(0.520429804914)*x_ref[0]**o+(-0.000683147004841)*x_ref[0]+(0.676879975826)*x_ref[1]**o+(0.760707578765)*x_ref[1]+(0.479225734268)*x_ref[2]**o+(0.132006876687)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.264333591663)*x[0]**o+(-0.884594789709)*x[0]+(0.809768917911)*x[1]**o+(-0.999064121889)*x[1]
        arg[1]=(-0.476663257532)*x[0]**o+(0.703121978294)*x[0]+(0.768222344055)*x[1]**o+(0.0126615637065)*x[1]
        ref[0]=(0.264333591663)*x_ref[0]**o+(-0.884594789709)*x_ref[0]+(0.809768917911)*x_ref[1]**o+(-0.999064121889)*x_ref[1]
        ref[1]=(-0.476663257532)*x_ref[0]**o+(0.703121978294)*x_ref[0]+(0.768222344055)*x_ref[1]**o+(0.0126615637065)*x_ref[1]
      else:
        arg[0]=(-0.814801897766)*x[0]**o+(0.417782597744)*x[0]+(0.354691012333)*x[1]**o+(-0.0529647248706)*x[1]+(-0.455085903639)*x[2]**o+(-0.501751102496)*x[2]
        arg[1]=(0.165354246971)*x[0]**o+(-0.238754573504)*x[0]+(-0.293225479626)*x[1]**o+(-0.42192384305)*x[1]+(-0.109509840118)*x[2]**o+(0.511666964461)*x[2]
        ref[0]=(-0.814801897766)*x_ref[0]**o+(0.417782597744)*x_ref[0]+(0.354691012333)*x_ref[1]**o+(-0.0529647248706)*x_ref[1]+(-0.455085903639)*x_ref[2]**o+(-0.501751102496)*x_ref[2]
        ref[1]=(0.165354246971)*x_ref[0]**o+(-0.238754573504)*x_ref[0]+(-0.293225479626)*x_ref[1]**o+(-0.42192384305)*x_ref[1]+(-0.109509840118)*x_ref[2]**o+(0.511666964461)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.469225177227)*x[0]**o+(-0.312308556752)*x[0]+(0.0207638302484)*x[1]**o+(-0.42312789395)*x[1]
        arg[0,1]=(0.435465800442)*x[0]**o+(0.66981866069)*x[0]+(-0.563664236008)*x[1]**o+(-0.979640187021)*x[1]
        arg[0,2]=(0.761422556363)*x[0]**o+(-0.374824039734)*x[0]+(-0.258990622087)*x[1]**o+(-0.553555723801)*x[1]
        arg[0,3]=(-0.346737956865)*x[0]**o+(-0.509820084012)*x[0]+(-0.270279814718)*x[1]**o+(-0.77787682055)*x[1]
        arg[0,4]=(-0.607151504575)*x[0]**o+(0.0370061792286)*x[0]+(-0.597194755504)*x[1]**o+(0.463748874183)*x[1]
        arg[1,0]=(-0.614871053235)*x[0]**o+(0.984151090349)*x[0]+(0.176558255935)*x[1]**o+(-0.00143697546002)*x[1]
        arg[1,1]=(-0.678550535699)*x[0]**o+(-0.40068020768)*x[0]+(-0.73609233692)*x[1]**o+(0.204641737633)*x[1]
        arg[1,2]=(0.644566611818)*x[0]**o+(0.676838663571)*x[0]+(0.592803079261)*x[1]**o+(-0.257851556005)*x[1]
        arg[1,3]=(-0.811008507599)*x[0]**o+(-0.810627035248)*x[0]+(0.470993697001)*x[1]**o+(0.524157551045)*x[1]
        arg[1,4]=(0.507989473752)*x[0]**o+(0.474629400564)*x[0]+(0.972372960952)*x[1]**o+(-0.713136826767)*x[1]
        arg[2,0]=(-0.18215096122)*x[0]**o+(0.0926167766546)*x[0]+(-0.526818906453)*x[1]**o+(-0.169071065789)*x[1]
        arg[2,1]=(-0.402265903759)*x[0]**o+(-0.460219787922)*x[0]+(0.160493468523)*x[1]**o+(-0.816631415356)*x[1]
        arg[2,2]=(-0.826968582935)*x[0]**o+(-0.477060281375)*x[0]+(-0.753204398647)*x[1]**o+(0.340944419869)*x[1]
        arg[2,3]=(-0.289758403392)*x[0]**o+(-0.216496476621)*x[0]+(-0.0729120551729)*x[1]**o+(0.0206397202194)*x[1]
        arg[2,4]=(0.900190015504)*x[0]**o+(0.554334407718)*x[0]+(-0.188091500011)*x[1]**o+(-0.239344021421)*x[1]
        arg[3,0]=(-0.846253042969)*x[0]**o+(0.212019924467)*x[0]+(-0.209309311993)*x[1]**o+(0.47889463943)*x[1]
        arg[3,1]=(-0.796067196362)*x[0]**o+(0.570882191233)*x[0]+(0.789930258)*x[1]**o+(-0.687939656508)*x[1]
        arg[3,2]=(0.0603011963958)*x[0]**o+(-0.291794860225)*x[0]+(-0.572598612637)*x[1]**o+(0.308585554728)*x[1]
        arg[3,3]=(0.883976313141)*x[0]**o+(0.857027876956)*x[0]+(-0.991113135222)*x[1]**o+(-0.0649848727938)*x[1]
        arg[3,4]=(-0.420604320052)*x[0]**o+(-0.147041915772)*x[0]+(0.696294525095)*x[1]**o+(-0.45791291995)*x[1]
        ref[0,0]=(-0.469225177227)*x_ref[0]**o+(-0.312308556752)*x_ref[0]+(0.0207638302484)*x_ref[1]**o+(-0.42312789395)*x_ref[1]
        ref[0,1]=(0.435465800442)*x_ref[0]**o+(0.66981866069)*x_ref[0]+(-0.563664236008)*x_ref[1]**o+(-0.979640187021)*x_ref[1]
        ref[0,2]=(0.761422556363)*x_ref[0]**o+(-0.374824039734)*x_ref[0]+(-0.258990622087)*x_ref[1]**o+(-0.553555723801)*x_ref[1]
        ref[0,3]=(-0.346737956865)*x_ref[0]**o+(-0.509820084012)*x_ref[0]+(-0.270279814718)*x_ref[1]**o+(-0.77787682055)*x_ref[1]
        ref[0,4]=(-0.607151504575)*x_ref[0]**o+(0.0370061792286)*x_ref[0]+(-0.597194755504)*x_ref[1]**o+(0.463748874183)*x_ref[1]
        ref[1,0]=(-0.614871053235)*x_ref[0]**o+(0.984151090349)*x_ref[0]+(0.176558255935)*x_ref[1]**o+(-0.00143697546002)*x_ref[1]
        ref[1,1]=(-0.678550535699)*x_ref[0]**o+(-0.40068020768)*x_ref[0]+(-0.73609233692)*x_ref[1]**o+(0.204641737633)*x_ref[1]
        ref[1,2]=(0.644566611818)*x_ref[0]**o+(0.676838663571)*x_ref[0]+(0.592803079261)*x_ref[1]**o+(-0.257851556005)*x_ref[1]
        ref[1,3]=(-0.811008507599)*x_ref[0]**o+(-0.810627035248)*x_ref[0]+(0.470993697001)*x_ref[1]**o+(0.524157551045)*x_ref[1]
        ref[1,4]=(0.507989473752)*x_ref[0]**o+(0.474629400564)*x_ref[0]+(0.972372960952)*x_ref[1]**o+(-0.713136826767)*x_ref[1]
        ref[2,0]=(-0.18215096122)*x_ref[0]**o+(0.0926167766546)*x_ref[0]+(-0.526818906453)*x_ref[1]**o+(-0.169071065789)*x_ref[1]
        ref[2,1]=(-0.402265903759)*x_ref[0]**o+(-0.460219787922)*x_ref[0]+(0.160493468523)*x_ref[1]**o+(-0.816631415356)*x_ref[1]
        ref[2,2]=(-0.826968582935)*x_ref[0]**o+(-0.477060281375)*x_ref[0]+(-0.753204398647)*x_ref[1]**o+(0.340944419869)*x_ref[1]
        ref[2,3]=(-0.289758403392)*x_ref[0]**o+(-0.216496476621)*x_ref[0]+(-0.0729120551729)*x_ref[1]**o+(0.0206397202194)*x_ref[1]
        ref[2,4]=(0.900190015504)*x_ref[0]**o+(0.554334407718)*x_ref[0]+(-0.188091500011)*x_ref[1]**o+(-0.239344021421)*x_ref[1]
        ref[3,0]=(-0.846253042969)*x_ref[0]**o+(0.212019924467)*x_ref[0]+(-0.209309311993)*x_ref[1]**o+(0.47889463943)*x_ref[1]
        ref[3,1]=(-0.796067196362)*x_ref[0]**o+(0.570882191233)*x_ref[0]+(0.789930258)*x_ref[1]**o+(-0.687939656508)*x_ref[1]
        ref[3,2]=(0.0603011963958)*x_ref[0]**o+(-0.291794860225)*x_ref[0]+(-0.572598612637)*x_ref[1]**o+(0.308585554728)*x_ref[1]
        ref[3,3]=(0.883976313141)*x_ref[0]**o+(0.857027876956)*x_ref[0]+(-0.991113135222)*x_ref[1]**o+(-0.0649848727938)*x_ref[1]
        ref[3,4]=(-0.420604320052)*x_ref[0]**o+(-0.147041915772)*x_ref[0]+(0.696294525095)*x_ref[1]**o+(-0.45791291995)*x_ref[1]
      else:
        arg[0,0]=(0.650989082369)*x[0]**o+(-0.291482513935)*x[0]+(-0.791203000903)*x[1]**o+(0.297385821221)*x[1]+(-0.169814832171)*x[2]**o+(-0.125360282254)*x[2]
        arg[0,1]=(0.166614460758)*x[0]**o+(0.0840046062434)*x[0]+(-0.114646748006)*x[1]**o+(-0.872705089326)*x[1]+(-0.687707782796)*x[2]**o+(-0.230067474846)*x[2]
        arg[0,2]=(-0.219927466535)*x[0]**o+(-0.466007112543)*x[0]+(-0.255341401047)*x[1]**o+(0.615864283673)*x[1]+(-0.983984041665)*x[2]**o+(0.305409617185)*x[2]
        arg[0,3]=(-0.0879672823913)*x[0]**o+(-0.956382814278)*x[0]+(-0.3091322614)*x[1]**o+(0.147177613425)*x[1]+(-0.98911137256)*x[2]**o+(-0.750791857003)*x[2]
        arg[0,4]=(0.715277956634)*x[0]**o+(-0.73358535985)*x[0]+(0.14587372226)*x[1]**o+(0.411467948058)*x[1]+(0.925673970492)*x[2]**o+(-0.0678728528684)*x[2]
        arg[1,0]=(0.150067062674)*x[0]**o+(-0.522799232667)*x[0]+(-0.256261929148)*x[1]**o+(-0.316346866901)*x[1]+(-0.849523865324)*x[2]**o+(-0.585421363115)*x[2]
        arg[1,1]=(-0.49684130662)*x[0]**o+(0.868986326423)*x[0]+(0.650641361203)*x[1]**o+(0.558696373611)*x[1]+(0.945790183046)*x[2]**o+(0.379250336542)*x[2]
        arg[1,2]=(-0.67349806443)*x[0]**o+(-0.337586770393)*x[0]+(0.719579305241)*x[1]**o+(-0.300711356473)*x[1]+(-0.350130081791)*x[2]**o+(0.604354519926)*x[2]
        arg[1,3]=(-0.242995155153)*x[0]**o+(0.668842006698)*x[0]+(-0.981902121891)*x[1]**o+(-0.358779020262)*x[1]+(0.298801927955)*x[2]**o+(-0.859807798836)*x[2]
        arg[1,4]=(-0.248474519407)*x[0]**o+(-0.309793754496)*x[0]+(-0.967889559474)*x[1]**o+(0.541761608865)*x[1]+(0.573726565934)*x[2]**o+(-0.876387334383)*x[2]
        arg[2,0]=(0.879662395476)*x[0]**o+(0.172656747739)*x[0]+(0.74129137412)*x[1]**o+(0.236748254611)*x[1]+(0.764805785803)*x[2]**o+(-0.257470519211)*x[2]
        arg[2,1]=(0.44326363254)*x[0]**o+(0.121548286191)*x[0]+(0.605969252858)*x[1]**o+(0.261621973968)*x[1]+(-0.694403080782)*x[2]**o+(0.15039482013)*x[2]
        arg[2,2]=(-0.372758098923)*x[0]**o+(0.577481130716)*x[0]+(-0.120284392247)*x[1]**o+(-0.630815000386)*x[1]+(-0.601595664522)*x[2]**o+(0.806894830983)*x[2]
        arg[2,3]=(0.548413470095)*x[0]**o+(0.804535150265)*x[0]+(0.469766192947)*x[1]**o+(-0.796431483578)*x[1]+(0.434732585165)*x[2]**o+(-0.20040808975)*x[2]
        arg[2,4]=(-0.119065000638)*x[0]**o+(-0.651948510575)*x[0]+(0.868418571224)*x[1]**o+(-0.987758488584)*x[1]+(0.495264099949)*x[2]**o+(-0.928684424484)*x[2]
        arg[3,0]=(-0.712239839816)*x[0]**o+(0.90734601523)*x[0]+(0.901855037856)*x[1]**o+(0.773630983477)*x[1]+(0.599873877074)*x[2]**o+(-0.511864534352)*x[2]
        arg[3,1]=(0.0499812109271)*x[0]**o+(0.959791265887)*x[0]+(0.861696669737)*x[1]**o+(-0.887520952646)*x[1]+(-0.609317719697)*x[2]**o+(-0.314327480839)*x[2]
        arg[3,2]=(0.717930038261)*x[0]**o+(-0.480998377737)*x[0]+(0.890930499387)*x[1]**o+(-0.880493180277)*x[1]+(0.540792831261)*x[2]**o+(0.343985055993)*x[2]
        arg[3,3]=(0.20531032325)*x[0]**o+(0.0400856366105)*x[0]+(0.829583965232)*x[1]**o+(0.701555952938)*x[1]+(0.676177356543)*x[2]**o+(0.934250384088)*x[2]
        arg[3,4]=(-0.426083844591)*x[0]**o+(0.0732912754379)*x[0]+(0.555586413705)*x[1]**o+(0.379137165516)*x[1]+(-0.481638055369)*x[2]**o+(-0.156787271652)*x[2]
        ref[0,0]=(0.650989082369)*x_ref[0]**o+(-0.291482513935)*x_ref[0]+(-0.791203000903)*x_ref[1]**o+(0.297385821221)*x_ref[1]+(-0.169814832171)*x_ref[2]**o+(-0.125360282254)*x_ref[2]
        ref[0,1]=(0.166614460758)*x_ref[0]**o+(0.0840046062434)*x_ref[0]+(-0.114646748006)*x_ref[1]**o+(-0.872705089326)*x_ref[1]+(-0.687707782796)*x_ref[2]**o+(-0.230067474846)*x_ref[2]
        ref[0,2]=(-0.219927466535)*x_ref[0]**o+(-0.466007112543)*x_ref[0]+(-0.255341401047)*x_ref[1]**o+(0.615864283673)*x_ref[1]+(-0.983984041665)*x_ref[2]**o+(0.305409617185)*x_ref[2]
        ref[0,3]=(-0.0879672823913)*x_ref[0]**o+(-0.956382814278)*x_ref[0]+(-0.3091322614)*x_ref[1]**o+(0.147177613425)*x_ref[1]+(-0.98911137256)*x_ref[2]**o+(-0.750791857003)*x_ref[2]
        ref[0,4]=(0.715277956634)*x_ref[0]**o+(-0.73358535985)*x_ref[0]+(0.14587372226)*x_ref[1]**o+(0.411467948058)*x_ref[1]+(0.925673970492)*x_ref[2]**o+(-0.0678728528684)*x_ref[2]
        ref[1,0]=(0.150067062674)*x_ref[0]**o+(-0.522799232667)*x_ref[0]+(-0.256261929148)*x_ref[1]**o+(-0.316346866901)*x_ref[1]+(-0.849523865324)*x_ref[2]**o+(-0.585421363115)*x_ref[2]
        ref[1,1]=(-0.49684130662)*x_ref[0]**o+(0.868986326423)*x_ref[0]+(0.650641361203)*x_ref[1]**o+(0.558696373611)*x_ref[1]+(0.945790183046)*x_ref[2]**o+(0.379250336542)*x_ref[2]
        ref[1,2]=(-0.67349806443)*x_ref[0]**o+(-0.337586770393)*x_ref[0]+(0.719579305241)*x_ref[1]**o+(-0.300711356473)*x_ref[1]+(-0.350130081791)*x_ref[2]**o+(0.604354519926)*x_ref[2]
        ref[1,3]=(-0.242995155153)*x_ref[0]**o+(0.668842006698)*x_ref[0]+(-0.981902121891)*x_ref[1]**o+(-0.358779020262)*x_ref[1]+(0.298801927955)*x_ref[2]**o+(-0.859807798836)*x_ref[2]
        ref[1,4]=(-0.248474519407)*x_ref[0]**o+(-0.309793754496)*x_ref[0]+(-0.967889559474)*x_ref[1]**o+(0.541761608865)*x_ref[1]+(0.573726565934)*x_ref[2]**o+(-0.876387334383)*x_ref[2]
        ref[2,0]=(0.879662395476)*x_ref[0]**o+(0.172656747739)*x_ref[0]+(0.74129137412)*x_ref[1]**o+(0.236748254611)*x_ref[1]+(0.764805785803)*x_ref[2]**o+(-0.257470519211)*x_ref[2]
        ref[2,1]=(0.44326363254)*x_ref[0]**o+(0.121548286191)*x_ref[0]+(0.605969252858)*x_ref[1]**o+(0.261621973968)*x_ref[1]+(-0.694403080782)*x_ref[2]**o+(0.15039482013)*x_ref[2]
        ref[2,2]=(-0.372758098923)*x_ref[0]**o+(0.577481130716)*x_ref[0]+(-0.120284392247)*x_ref[1]**o+(-0.630815000386)*x_ref[1]+(-0.601595664522)*x_ref[2]**o+(0.806894830983)*x_ref[2]
        ref[2,3]=(0.548413470095)*x_ref[0]**o+(0.804535150265)*x_ref[0]+(0.469766192947)*x_ref[1]**o+(-0.796431483578)*x_ref[1]+(0.434732585165)*x_ref[2]**o+(-0.20040808975)*x_ref[2]
        ref[2,4]=(-0.119065000638)*x_ref[0]**o+(-0.651948510575)*x_ref[0]+(0.868418571224)*x_ref[1]**o+(-0.987758488584)*x_ref[1]+(0.495264099949)*x_ref[2]**o+(-0.928684424484)*x_ref[2]
        ref[3,0]=(-0.712239839816)*x_ref[0]**o+(0.90734601523)*x_ref[0]+(0.901855037856)*x_ref[1]**o+(0.773630983477)*x_ref[1]+(0.599873877074)*x_ref[2]**o+(-0.511864534352)*x_ref[2]
        ref[3,1]=(0.0499812109271)*x_ref[0]**o+(0.959791265887)*x_ref[0]+(0.861696669737)*x_ref[1]**o+(-0.887520952646)*x_ref[1]+(-0.609317719697)*x_ref[2]**o+(-0.314327480839)*x_ref[2]
        ref[3,2]=(0.717930038261)*x_ref[0]**o+(-0.480998377737)*x_ref[0]+(0.890930499387)*x_ref[1]**o+(-0.880493180277)*x_ref[1]+(0.540792831261)*x_ref[2]**o+(0.343985055993)*x_ref[2]
        ref[3,3]=(0.20531032325)*x_ref[0]**o+(0.0400856366105)*x_ref[0]+(0.829583965232)*x_ref[1]**o+(0.701555952938)*x_ref[1]+(0.676177356543)*x_ref[2]**o+(0.934250384088)*x_ref[2]
        ref[3,4]=(-0.426083844591)*x_ref[0]**o+(0.0732912754379)*x_ref[0]+(0.555586413705)*x_ref[1]**o+(0.379137165516)*x_ref[1]+(-0.481638055369)*x_ref[2]**o+(-0.156787271652)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.0451039988637)*x[0]**o+(0.159426346984)*x[0]+(0.693271264689)*x[1]**o+(-0.129057826038)*x[1]
        arg[0,0,1]=(0.165096887432)*x[0]**o+(0.421267288474)*x[0]+(0.259248346979)*x[1]**o+(-0.216110688389)*x[1]
        arg[0,1,0]=(-0.419500329289)*x[0]**o+(-0.516326782996)*x[0]+(-0.726209291778)*x[1]**o+(0.80434579811)*x[1]
        arg[0,1,1]=(-0.680429758627)*x[0]**o+(-0.539365634112)*x[0]+(-0.20379891956)*x[1]**o+(-0.760436888151)*x[1]
        arg[1,0,0]=(-0.472638888322)*x[0]**o+(0.982469790589)*x[0]+(-0.383590346736)*x[1]**o+(-0.69130786846)*x[1]
        arg[1,0,1]=(0.334498522873)*x[0]**o+(0.190404359107)*x[0]+(0.210624993429)*x[1]**o+(-0.744921567562)*x[1]
        arg[1,1,0]=(0.632591775382)*x[0]**o+(-0.405215541727)*x[0]+(-0.880471123936)*x[1]**o+(0.0942729925938)*x[1]
        arg[1,1,1]=(-0.837885933046)*x[0]**o+(-0.798882152889)*x[0]+(0.648322104814)*x[1]**o+(-0.357620132856)*x[1]
        arg[2,0,0]=(-0.490592533882)*x[0]**o+(0.540431305386)*x[0]+(-0.777299904284)*x[1]**o+(-0.308682146808)*x[1]
        arg[2,0,1]=(0.971265898941)*x[0]**o+(0.546875456819)*x[0]+(0.627171799906)*x[1]**o+(0.0732789392684)*x[1]
        arg[2,1,0]=(0.178076638358)*x[0]**o+(0.587800697151)*x[0]+(0.748336208169)*x[1]**o+(-0.0341258196613)*x[1]
        arg[2,1,1]=(0.36345594982)*x[0]**o+(-0.899402201831)*x[0]+(0.0206286665096)*x[1]**o+(0.303557722727)*x[1]
        arg[3,0,0]=(-0.544119125777)*x[0]**o+(-0.805891689123)*x[0]+(-0.687902925271)*x[1]**o+(0.346236636736)*x[1]
        arg[3,0,1]=(0.725902084271)*x[0]**o+(0.605461404125)*x[0]+(0.452767036006)*x[1]**o+(0.266584718883)*x[1]
        arg[3,1,0]=(0.322052654392)*x[0]**o+(-0.91949636775)*x[0]+(0.948097272559)*x[1]**o+(-0.463715765101)*x[1]
        arg[3,1,1]=(-0.369192142232)*x[0]**o+(0.616001201865)*x[0]+(-0.257989673651)*x[1]**o+(0.507550625249)*x[1]
        arg[4,0,0]=(0.270343836597)*x[0]**o+(0.851686984957)*x[0]+(0.838756000051)*x[1]**o+(-0.329625536762)*x[1]
        arg[4,0,1]=(0.00026238459417)*x[0]**o+(0.0274581707442)*x[0]+(0.770120998708)*x[1]**o+(0.574619799143)*x[1]
        arg[4,1,0]=(-0.709788768284)*x[0]**o+(0.0856634157603)*x[0]+(-0.294662391976)*x[1]**o+(-0.90639308515)*x[1]
        arg[4,1,1]=(0.223381811001)*x[0]**o+(0.941451590223)*x[0]+(0.917865807171)*x[1]**o+(0.912972271206)*x[1]
        arg[5,0,0]=(-0.731163054411)*x[0]**o+(0.185548632151)*x[0]+(0.907527980776)*x[1]**o+(0.816017248525)*x[1]
        arg[5,0,1]=(0.719177781335)*x[0]**o+(-0.788006096642)*x[0]+(-0.434802402615)*x[1]**o+(0.0249465251663)*x[1]
        arg[5,1,0]=(0.982421863977)*x[0]**o+(0.127223012805)*x[0]+(-0.858491508817)*x[1]**o+(-0.152253965468)*x[1]
        arg[5,1,1]=(-0.779596616421)*x[0]**o+(0.049815217296)*x[0]+(0.892121656728)*x[1]**o+(0.258587664554)*x[1]
        ref[0,0,0]=(0.0451039988637)*x_ref[0]**o+(0.159426346984)*x_ref[0]+(0.693271264689)*x_ref[1]**o+(-0.129057826038)*x_ref[1]
        ref[0,0,1]=(0.165096887432)*x_ref[0]**o+(0.421267288474)*x_ref[0]+(0.259248346979)*x_ref[1]**o+(-0.216110688389)*x_ref[1]
        ref[0,1,0]=(-0.419500329289)*x_ref[0]**o+(-0.516326782996)*x_ref[0]+(-0.726209291778)*x_ref[1]**o+(0.80434579811)*x_ref[1]
        ref[0,1,1]=(-0.680429758627)*x_ref[0]**o+(-0.539365634112)*x_ref[0]+(-0.20379891956)*x_ref[1]**o+(-0.760436888151)*x_ref[1]
        ref[1,0,0]=(-0.472638888322)*x_ref[0]**o+(0.982469790589)*x_ref[0]+(-0.383590346736)*x_ref[1]**o+(-0.69130786846)*x_ref[1]
        ref[1,0,1]=(0.334498522873)*x_ref[0]**o+(0.190404359107)*x_ref[0]+(0.210624993429)*x_ref[1]**o+(-0.744921567562)*x_ref[1]
        ref[1,1,0]=(0.632591775382)*x_ref[0]**o+(-0.405215541727)*x_ref[0]+(-0.880471123936)*x_ref[1]**o+(0.0942729925938)*x_ref[1]
        ref[1,1,1]=(-0.837885933046)*x_ref[0]**o+(-0.798882152889)*x_ref[0]+(0.648322104814)*x_ref[1]**o+(-0.357620132856)*x_ref[1]
        ref[2,0,0]=(-0.490592533882)*x_ref[0]**o+(0.540431305386)*x_ref[0]+(-0.777299904284)*x_ref[1]**o+(-0.308682146808)*x_ref[1]
        ref[2,0,1]=(0.971265898941)*x_ref[0]**o+(0.546875456819)*x_ref[0]+(0.627171799906)*x_ref[1]**o+(0.0732789392684)*x_ref[1]
        ref[2,1,0]=(0.178076638358)*x_ref[0]**o+(0.587800697151)*x_ref[0]+(0.748336208169)*x_ref[1]**o+(-0.0341258196613)*x_ref[1]
        ref[2,1,1]=(0.36345594982)*x_ref[0]**o+(-0.899402201831)*x_ref[0]+(0.0206286665096)*x_ref[1]**o+(0.303557722727)*x_ref[1]
        ref[3,0,0]=(-0.544119125777)*x_ref[0]**o+(-0.805891689123)*x_ref[0]+(-0.687902925271)*x_ref[1]**o+(0.346236636736)*x_ref[1]
        ref[3,0,1]=(0.725902084271)*x_ref[0]**o+(0.605461404125)*x_ref[0]+(0.452767036006)*x_ref[1]**o+(0.266584718883)*x_ref[1]
        ref[3,1,0]=(0.322052654392)*x_ref[0]**o+(-0.91949636775)*x_ref[0]+(0.948097272559)*x_ref[1]**o+(-0.463715765101)*x_ref[1]
        ref[3,1,1]=(-0.369192142232)*x_ref[0]**o+(0.616001201865)*x_ref[0]+(-0.257989673651)*x_ref[1]**o+(0.507550625249)*x_ref[1]
        ref[4,0,0]=(0.270343836597)*x_ref[0]**o+(0.851686984957)*x_ref[0]+(0.838756000051)*x_ref[1]**o+(-0.329625536762)*x_ref[1]
        ref[4,0,1]=(0.00026238459417)*x_ref[0]**o+(0.0274581707442)*x_ref[0]+(0.770120998708)*x_ref[1]**o+(0.574619799143)*x_ref[1]
        ref[4,1,0]=(-0.709788768284)*x_ref[0]**o+(0.0856634157603)*x_ref[0]+(-0.294662391976)*x_ref[1]**o+(-0.90639308515)*x_ref[1]
        ref[4,1,1]=(0.223381811001)*x_ref[0]**o+(0.941451590223)*x_ref[0]+(0.917865807171)*x_ref[1]**o+(0.912972271206)*x_ref[1]
        ref[5,0,0]=(-0.731163054411)*x_ref[0]**o+(0.185548632151)*x_ref[0]+(0.907527980776)*x_ref[1]**o+(0.816017248525)*x_ref[1]
        ref[5,0,1]=(0.719177781335)*x_ref[0]**o+(-0.788006096642)*x_ref[0]+(-0.434802402615)*x_ref[1]**o+(0.0249465251663)*x_ref[1]
        ref[5,1,0]=(0.982421863977)*x_ref[0]**o+(0.127223012805)*x_ref[0]+(-0.858491508817)*x_ref[1]**o+(-0.152253965468)*x_ref[1]
        ref[5,1,1]=(-0.779596616421)*x_ref[0]**o+(0.049815217296)*x_ref[0]+(0.892121656728)*x_ref[1]**o+(0.258587664554)*x_ref[1]
      else:
        arg[0,0,0]=(0.804406156662)*x[0]**o+(0.304298949556)*x[0]+(-9.23949223481e-05)*x[1]**o+(0.521245437597)*x[1]+(0.822885620595)*x[2]**o+(0.44472437921)*x[2]
        arg[0,0,1]=(0.208119461011)*x[0]**o+(-0.967847500694)*x[0]+(-0.182961335156)*x[1]**o+(-0.0510538508635)*x[1]+(0.932285255322)*x[2]**o+(-0.463525963095)*x[2]
        arg[0,1,0]=(-0.0514207812926)*x[0]**o+(-0.730723672364)*x[0]+(-0.471144684451)*x[1]**o+(0.0359308791998)*x[1]+(-0.935079157697)*x[2]**o+(0.841965819816)*x[2]
        arg[0,1,1]=(0.607427052267)*x[0]**o+(-0.344216102016)*x[0]+(0.967271353509)*x[1]**o+(-0.412288382331)*x[1]+(-0.0716617724983)*x[2]**o+(0.56895080823)*x[2]
        arg[1,0,0]=(-0.574314769591)*x[0]**o+(-0.819434106932)*x[0]+(0.143229095273)*x[1]**o+(0.505343958082)*x[1]+(0.497764227469)*x[2]**o+(0.592609757968)*x[2]
        arg[1,0,1]=(0.414984900558)*x[0]**o+(0.624574053857)*x[0]+(-0.367330776954)*x[1]**o+(-0.30093924896)*x[1]+(-0.99416380749)*x[2]**o+(0.492443599924)*x[2]
        arg[1,1,0]=(0.171673119954)*x[0]**o+(0.612986140068)*x[0]+(-0.671355907363)*x[1]**o+(-0.292080539973)*x[1]+(-0.450890819125)*x[2]**o+(-0.64742047997)*x[2]
        arg[1,1,1]=(0.452494887295)*x[0]**o+(0.187308285528)*x[0]+(-0.995316825981)*x[1]**o+(-0.825194018202)*x[1]+(-0.00651224511881)*x[2]**o+(-0.715747547803)*x[2]
        arg[2,0,0]=(0.546297169466)*x[0]**o+(-0.424626884041)*x[0]+(0.327647062351)*x[1]**o+(-0.050651105527)*x[1]+(-0.667117310429)*x[2]**o+(-0.37659304263)*x[2]
        arg[2,0,1]=(-0.323595958622)*x[0]**o+(0.541718593289)*x[0]+(0.6285998056)*x[1]**o+(0.488539567593)*x[1]+(0.983086232412)*x[2]**o+(0.692821558469)*x[2]
        arg[2,1,0]=(0.16009856161)*x[0]**o+(-0.776069988482)*x[0]+(0.750717144319)*x[1]**o+(-0.244287895989)*x[1]+(0.345242766297)*x[2]**o+(0.572063651895)*x[2]
        arg[2,1,1]=(-0.34626843638)*x[0]**o+(0.197125033681)*x[0]+(-0.570015603866)*x[1]**o+(-0.633161422872)*x[1]+(-0.602998606808)*x[2]**o+(-0.674829268895)*x[2]
        arg[3,0,0]=(-0.659696747189)*x[0]**o+(-0.571128578512)*x[0]+(-0.704414117225)*x[1]**o+(-0.334924171135)*x[1]+(-0.0200665796175)*x[2]**o+(-0.660688341854)*x[2]
        arg[3,0,1]=(-0.497019087713)*x[0]**o+(0.25850813993)*x[0]+(-0.713426904693)*x[1]**o+(-0.994061006282)*x[1]+(-0.669422282701)*x[2]**o+(-0.553981025574)*x[2]
        arg[3,1,0]=(-0.151249618276)*x[0]**o+(0.29432574729)*x[0]+(0.173047919293)*x[1]**o+(0.844456199843)*x[1]+(0.845402272715)*x[2]**o+(-0.115698924822)*x[2]
        arg[3,1,1]=(0.840440627264)*x[0]**o+(0.593400475213)*x[0]+(0.974677794562)*x[1]**o+(0.938845603167)*x[1]+(-0.944730436891)*x[2]**o+(-0.657543633218)*x[2]
        arg[4,0,0]=(0.968290161629)*x[0]**o+(0.578471458995)*x[0]+(0.946558149029)*x[1]**o+(-0.232692317459)*x[1]+(-0.97762469848)*x[2]**o+(0.660519364613)*x[2]
        arg[4,0,1]=(-0.187317070761)*x[0]**o+(-0.0363170179961)*x[0]+(0.278499613269)*x[1]**o+(-0.0685831047696)*x[1]+(-0.540282212359)*x[2]**o+(0.564656560066)*x[2]
        arg[4,1,0]=(-0.211825530749)*x[0]**o+(0.225546568724)*x[0]+(-0.83241974026)*x[1]**o+(-0.503828393954)*x[1]+(-0.635937384108)*x[2]**o+(-0.513183062177)*x[2]
        arg[4,1,1]=(-0.257123983215)*x[0]**o+(-0.0700924567638)*x[0]+(0.550859393713)*x[1]**o+(0.132070212001)*x[1]+(-0.478549332853)*x[2]**o+(-0.560087260102)*x[2]
        arg[5,0,0]=(0.677259763344)*x[0]**o+(-0.213422444013)*x[0]+(-0.294112138735)*x[1]**o+(0.083135232348)*x[1]+(-0.899953896237)*x[2]**o+(-0.117143433064)*x[2]
        arg[5,0,1]=(-0.489822592306)*x[0]**o+(0.368217759904)*x[0]+(-0.927837261915)*x[1]**o+(-0.708243327427)*x[1]+(-0.132256551163)*x[2]**o+(0.316204052014)*x[2]
        arg[5,1,0]=(0.314897334392)*x[0]**o+(-0.848168680639)*x[0]+(0.92421318346)*x[1]**o+(0.995505186786)*x[1]+(0.372381949158)*x[2]**o+(-0.563415970977)*x[2]
        arg[5,1,1]=(0.745011483885)*x[0]**o+(-0.6423283108)*x[0]+(-0.119592965986)*x[1]**o+(0.241068250891)*x[1]+(0.477172198315)*x[2]**o+(-0.468580220406)*x[2]
        ref[0,0,0]=(0.804406156662)*x_ref[0]**o+(0.304298949556)*x_ref[0]+(-9.23949223481e-05)*x_ref[1]**o+(0.521245437597)*x_ref[1]+(0.822885620595)*x_ref[2]**o+(0.44472437921)*x_ref[2]
        ref[0,0,1]=(0.208119461011)*x_ref[0]**o+(-0.967847500694)*x_ref[0]+(-0.182961335156)*x_ref[1]**o+(-0.0510538508635)*x_ref[1]+(0.932285255322)*x_ref[2]**o+(-0.463525963095)*x_ref[2]
        ref[0,1,0]=(-0.0514207812926)*x_ref[0]**o+(-0.730723672364)*x_ref[0]+(-0.471144684451)*x_ref[1]**o+(0.0359308791998)*x_ref[1]+(-0.935079157697)*x_ref[2]**o+(0.841965819816)*x_ref[2]
        ref[0,1,1]=(0.607427052267)*x_ref[0]**o+(-0.344216102016)*x_ref[0]+(0.967271353509)*x_ref[1]**o+(-0.412288382331)*x_ref[1]+(-0.0716617724983)*x_ref[2]**o+(0.56895080823)*x_ref[2]
        ref[1,0,0]=(-0.574314769591)*x_ref[0]**o+(-0.819434106932)*x_ref[0]+(0.143229095273)*x_ref[1]**o+(0.505343958082)*x_ref[1]+(0.497764227469)*x_ref[2]**o+(0.592609757968)*x_ref[2]
        ref[1,0,1]=(0.414984900558)*x_ref[0]**o+(0.624574053857)*x_ref[0]+(-0.367330776954)*x_ref[1]**o+(-0.30093924896)*x_ref[1]+(-0.99416380749)*x_ref[2]**o+(0.492443599924)*x_ref[2]
        ref[1,1,0]=(0.171673119954)*x_ref[0]**o+(0.612986140068)*x_ref[0]+(-0.671355907363)*x_ref[1]**o+(-0.292080539973)*x_ref[1]+(-0.450890819125)*x_ref[2]**o+(-0.64742047997)*x_ref[2]
        ref[1,1,1]=(0.452494887295)*x_ref[0]**o+(0.187308285528)*x_ref[0]+(-0.995316825981)*x_ref[1]**o+(-0.825194018202)*x_ref[1]+(-0.00651224511881)*x_ref[2]**o+(-0.715747547803)*x_ref[2]
        ref[2,0,0]=(0.546297169466)*x_ref[0]**o+(-0.424626884041)*x_ref[0]+(0.327647062351)*x_ref[1]**o+(-0.050651105527)*x_ref[1]+(-0.667117310429)*x_ref[2]**o+(-0.37659304263)*x_ref[2]
        ref[2,0,1]=(-0.323595958622)*x_ref[0]**o+(0.541718593289)*x_ref[0]+(0.6285998056)*x_ref[1]**o+(0.488539567593)*x_ref[1]+(0.983086232412)*x_ref[2]**o+(0.692821558469)*x_ref[2]
        ref[2,1,0]=(0.16009856161)*x_ref[0]**o+(-0.776069988482)*x_ref[0]+(0.750717144319)*x_ref[1]**o+(-0.244287895989)*x_ref[1]+(0.345242766297)*x_ref[2]**o+(0.572063651895)*x_ref[2]
        ref[2,1,1]=(-0.34626843638)*x_ref[0]**o+(0.197125033681)*x_ref[0]+(-0.570015603866)*x_ref[1]**o+(-0.633161422872)*x_ref[1]+(-0.602998606808)*x_ref[2]**o+(-0.674829268895)*x_ref[2]
        ref[3,0,0]=(-0.659696747189)*x_ref[0]**o+(-0.571128578512)*x_ref[0]+(-0.704414117225)*x_ref[1]**o+(-0.334924171135)*x_ref[1]+(-0.0200665796175)*x_ref[2]**o+(-0.660688341854)*x_ref[2]
        ref[3,0,1]=(-0.497019087713)*x_ref[0]**o+(0.25850813993)*x_ref[0]+(-0.713426904693)*x_ref[1]**o+(-0.994061006282)*x_ref[1]+(-0.669422282701)*x_ref[2]**o+(-0.553981025574)*x_ref[2]
        ref[3,1,0]=(-0.151249618276)*x_ref[0]**o+(0.29432574729)*x_ref[0]+(0.173047919293)*x_ref[1]**o+(0.844456199843)*x_ref[1]+(0.845402272715)*x_ref[2]**o+(-0.115698924822)*x_ref[2]
        ref[3,1,1]=(0.840440627264)*x_ref[0]**o+(0.593400475213)*x_ref[0]+(0.974677794562)*x_ref[1]**o+(0.938845603167)*x_ref[1]+(-0.944730436891)*x_ref[2]**o+(-0.657543633218)*x_ref[2]
        ref[4,0,0]=(0.968290161629)*x_ref[0]**o+(0.578471458995)*x_ref[0]+(0.946558149029)*x_ref[1]**o+(-0.232692317459)*x_ref[1]+(-0.97762469848)*x_ref[2]**o+(0.660519364613)*x_ref[2]
        ref[4,0,1]=(-0.187317070761)*x_ref[0]**o+(-0.0363170179961)*x_ref[0]+(0.278499613269)*x_ref[1]**o+(-0.0685831047696)*x_ref[1]+(-0.540282212359)*x_ref[2]**o+(0.564656560066)*x_ref[2]
        ref[4,1,0]=(-0.211825530749)*x_ref[0]**o+(0.225546568724)*x_ref[0]+(-0.83241974026)*x_ref[1]**o+(-0.503828393954)*x_ref[1]+(-0.635937384108)*x_ref[2]**o+(-0.513183062177)*x_ref[2]
        ref[4,1,1]=(-0.257123983215)*x_ref[0]**o+(-0.0700924567638)*x_ref[0]+(0.550859393713)*x_ref[1]**o+(0.132070212001)*x_ref[1]+(-0.478549332853)*x_ref[2]**o+(-0.560087260102)*x_ref[2]
        ref[5,0,0]=(0.677259763344)*x_ref[0]**o+(-0.213422444013)*x_ref[0]+(-0.294112138735)*x_ref[1]**o+(0.083135232348)*x_ref[1]+(-0.899953896237)*x_ref[2]**o+(-0.117143433064)*x_ref[2]
        ref[5,0,1]=(-0.489822592306)*x_ref[0]**o+(0.368217759904)*x_ref[0]+(-0.927837261915)*x_ref[1]**o+(-0.708243327427)*x_ref[1]+(-0.132256551163)*x_ref[2]**o+(0.316204052014)*x_ref[2]
        ref[5,1,0]=(0.314897334392)*x_ref[0]**o+(-0.848168680639)*x_ref[0]+(0.92421318346)*x_ref[1]**o+(0.995505186786)*x_ref[1]+(0.372381949158)*x_ref[2]**o+(-0.563415970977)*x_ref[2]
        ref[5,1,1]=(0.745011483885)*x_ref[0]**o+(-0.6423283108)*x_ref[0]+(-0.119592965986)*x_ref[1]**o+(0.241068250891)*x_ref[1]+(0.477172198315)*x_ref[2]**o+(-0.468580220406)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.633576516648)*x[0]**o+(0.77399256211)*x[0]+(-0.396696417195)*x[1]**o+(-0.604572540725)*x[1]
        arg[0,0,0,1]=(0.612672991978)*x[0]**o+(-0.331012302611)*x[0]+(0.885533578235)*x[1]**o+(0.334686818058)*x[1]
        arg[0,0,1,0]=(-0.982701069025)*x[0]**o+(0.467849033642)*x[0]+(0.525392773375)*x[1]**o+(0.355622324425)*x[1]
        arg[0,0,1,1]=(-0.0243671469666)*x[0]**o+(0.790044842383)*x[0]+(0.413384172033)*x[1]**o+(0.740805135968)*x[1]
        arg[0,0,2,0]=(0.0384789957691)*x[0]**o+(-0.317009899417)*x[0]+(0.0389710369615)*x[1]**o+(0.784549852598)*x[1]
        arg[0,0,2,1]=(0.361832259252)*x[0]**o+(0.410815792944)*x[0]+(0.39496145831)*x[1]**o+(-0.820784143964)*x[1]
        arg[0,1,0,0]=(0.224261088976)*x[0]**o+(-0.824694527256)*x[0]+(0.318674472135)*x[1]**o+(0.827696354338)*x[1]
        arg[0,1,0,1]=(-0.442154901359)*x[0]**o+(0.141775320862)*x[0]+(0.688711841268)*x[1]**o+(-0.27047542755)*x[1]
        arg[0,1,1,0]=(-0.570125802019)*x[0]**o+(0.050303762895)*x[0]+(-0.972329837417)*x[1]**o+(0.143700303765)*x[1]
        arg[0,1,1,1]=(0.852888550499)*x[0]**o+(-0.51216090329)*x[0]+(-0.895663922335)*x[1]**o+(-0.183466225547)*x[1]
        arg[0,1,2,0]=(0.863840920739)*x[0]**o+(-0.247789353524)*x[0]+(0.63056436157)*x[1]**o+(0.932522589942)*x[1]
        arg[0,1,2,1]=(-0.403832121794)*x[0]**o+(-0.769625259789)*x[0]+(0.507724303191)*x[1]**o+(0.314178775526)*x[1]
        arg[0,2,0,0]=(0.543867153515)*x[0]**o+(-0.414107817028)*x[0]+(0.423922720842)*x[1]**o+(0.166632407916)*x[1]
        arg[0,2,0,1]=(0.731968209929)*x[0]**o+(-0.114742736683)*x[0]+(0.112562415021)*x[1]**o+(0.166924480325)*x[1]
        arg[0,2,1,0]=(0.919472066242)*x[0]**o+(0.560565315205)*x[0]+(0.4967650028)*x[1]**o+(0.708376671625)*x[1]
        arg[0,2,1,1]=(0.147500239739)*x[0]**o+(-0.632670389919)*x[0]+(-0.992479063437)*x[1]**o+(0.258941122962)*x[1]
        arg[0,2,2,0]=(0.474481785329)*x[0]**o+(0.592902604094)*x[0]+(0.64058685211)*x[1]**o+(-0.477843307901)*x[1]
        arg[0,2,2,1]=(0.769898355146)*x[0]**o+(0.563458279487)*x[0]+(0.876276425449)*x[1]**o+(-0.0625727749869)*x[1]
        arg[0,3,0,0]=(-0.330055037059)*x[0]**o+(-0.727656105163)*x[0]+(-0.559901684392)*x[1]**o+(-0.433646878263)*x[1]
        arg[0,3,0,1]=(-0.725314328666)*x[0]**o+(0.638326609105)*x[0]+(-0.126619368417)*x[1]**o+(-0.232194758676)*x[1]
        arg[0,3,1,0]=(0.318096005545)*x[0]**o+(0.325247494289)*x[0]+(-0.514305253643)*x[1]**o+(0.108523350086)*x[1]
        arg[0,3,1,1]=(0.298034894956)*x[0]**o+(0.825714393651)*x[0]+(-0.67162850079)*x[1]**o+(0.681809083004)*x[1]
        arg[0,3,2,0]=(-0.00715890582648)*x[0]**o+(0.893900731227)*x[0]+(-0.773710274545)*x[1]**o+(0.415913728569)*x[1]
        arg[0,3,2,1]=(0.0853551940958)*x[0]**o+(-0.228876011127)*x[0]+(0.87699456762)*x[1]**o+(-0.744605229253)*x[1]
        arg[0,4,0,0]=(0.942646937302)*x[0]**o+(0.92717236759)*x[0]+(0.794806821569)*x[1]**o+(0.848766303526)*x[1]
        arg[0,4,0,1]=(0.20140672561)*x[0]**o+(0.974490372414)*x[0]+(-0.450332419587)*x[1]**o+(-0.859102184782)*x[1]
        arg[0,4,1,0]=(0.632153053193)*x[0]**o+(0.702336756277)*x[0]+(-0.571496076236)*x[1]**o+(-0.882472852219)*x[1]
        arg[0,4,1,1]=(0.366452829499)*x[0]**o+(-0.514660164741)*x[0]+(-0.152794832253)*x[1]**o+(-0.658688117796)*x[1]
        arg[0,4,2,0]=(0.406218394702)*x[0]**o+(0.782359698198)*x[0]+(-0.0582499587562)*x[1]**o+(0.664506852595)*x[1]
        arg[0,4,2,1]=(0.768240870343)*x[0]**o+(0.188470592904)*x[0]+(0.855933509171)*x[1]**o+(0.639174087541)*x[1]
        arg[1,0,0,0]=(-0.946841398046)*x[0]**o+(0.892729256183)*x[0]+(-0.15644094115)*x[1]**o+(-0.859371368233)*x[1]
        arg[1,0,0,1]=(0.238113647666)*x[0]**o+(0.173695470467)*x[0]+(-0.753529252313)*x[1]**o+(-0.243203902463)*x[1]
        arg[1,0,1,0]=(0.267937523443)*x[0]**o+(0.600200492374)*x[0]+(0.442653529257)*x[1]**o+(-0.394984210541)*x[1]
        arg[1,0,1,1]=(-0.495022907622)*x[0]**o+(-0.593273808093)*x[0]+(0.888764746608)*x[1]**o+(-0.745127825763)*x[1]
        arg[1,0,2,0]=(0.352827238496)*x[0]**o+(0.293200928839)*x[0]+(-0.00369812109034)*x[1]**o+(0.478875774193)*x[1]
        arg[1,0,2,1]=(-0.403358878981)*x[0]**o+(0.756947364667)*x[0]+(0.649335795039)*x[1]**o+(-0.339962796859)*x[1]
        arg[1,1,0,0]=(-0.607505349015)*x[0]**o+(-0.392535399582)*x[0]+(0.799782389634)*x[1]**o+(0.186339165863)*x[1]
        arg[1,1,0,1]=(0.997492304627)*x[0]**o+(-0.281456622581)*x[0]+(-0.959294693613)*x[1]**o+(-0.454812606674)*x[1]
        arg[1,1,1,0]=(0.454682037549)*x[0]**o+(0.797129935647)*x[0]+(-0.949539573053)*x[1]**o+(-0.918499382199)*x[1]
        arg[1,1,1,1]=(0.903016449758)*x[0]**o+(-0.575039046836)*x[0]+(0.957158277074)*x[1]**o+(-0.715797460906)*x[1]
        arg[1,1,2,0]=(0.92005547899)*x[0]**o+(-0.716975919302)*x[0]+(-0.900904311708)*x[1]**o+(0.989458809643)*x[1]
        arg[1,1,2,1]=(-0.633214991087)*x[0]**o+(0.680658578627)*x[0]+(0.552194054556)*x[1]**o+(0.343557072769)*x[1]
        arg[1,2,0,0]=(-0.0182903530463)*x[0]**o+(-0.0445569497267)*x[0]+(-0.259920069365)*x[1]**o+(0.118765787279)*x[1]
        arg[1,2,0,1]=(-0.667382174733)*x[0]**o+(-0.79898800024)*x[0]+(-0.217638556567)*x[1]**o+(-0.0288504127936)*x[1]
        arg[1,2,1,0]=(-0.437982445862)*x[0]**o+(-0.104054454808)*x[0]+(0.711429718088)*x[1]**o+(-0.607226296329)*x[1]
        arg[1,2,1,1]=(0.252861218037)*x[0]**o+(-0.635701456504)*x[0]+(-0.555222304592)*x[1]**o+(0.723744010092)*x[1]
        arg[1,2,2,0]=(-0.73532710885)*x[0]**o+(0.600599078741)*x[0]+(0.0465795173194)*x[1]**o+(0.524165544645)*x[1]
        arg[1,2,2,1]=(-0.025246019744)*x[0]**o+(-0.376526675619)*x[0]+(0.213011788698)*x[1]**o+(-0.477018225777)*x[1]
        arg[1,3,0,0]=(-0.52682690957)*x[0]**o+(0.0567472858499)*x[0]+(0.574116256102)*x[1]**o+(-0.130012791898)*x[1]
        arg[1,3,0,1]=(0.113502182936)*x[0]**o+(0.7643861313)*x[0]+(0.48414948847)*x[1]**o+(0.863168250318)*x[1]
        arg[1,3,1,0]=(-0.208102162378)*x[0]**o+(0.327770728981)*x[0]+(0.843778895278)*x[1]**o+(0.474843504177)*x[1]
        arg[1,3,1,1]=(-0.549818192184)*x[0]**o+(-0.863333816685)*x[0]+(-0.365631583957)*x[1]**o+(-0.491320441182)*x[1]
        arg[1,3,2,0]=(-0.265833845759)*x[0]**o+(-0.94416746081)*x[0]+(0.48465442593)*x[1]**o+(0.258284621408)*x[1]
        arg[1,3,2,1]=(-0.625716082268)*x[0]**o+(-0.234697130359)*x[0]+(-0.973704406131)*x[1]**o+(0.00141746117847)*x[1]
        arg[1,4,0,0]=(0.565044361964)*x[0]**o+(-0.0672474991748)*x[0]+(-0.125901930524)*x[1]**o+(-0.753103899732)*x[1]
        arg[1,4,0,1]=(-0.979743467289)*x[0]**o+(-0.25225085214)*x[0]+(-0.141635164635)*x[1]**o+(0.378153168667)*x[1]
        arg[1,4,1,0]=(0.97728399206)*x[0]**o+(-0.202043160532)*x[0]+(0.997445424954)*x[1]**o+(-0.100049830154)*x[1]
        arg[1,4,1,1]=(0.364840105486)*x[0]**o+(-0.682234327437)*x[0]+(0.252584592259)*x[1]**o+(-0.0785430242425)*x[1]
        arg[1,4,2,0]=(-0.0168764722616)*x[0]**o+(-0.997460139026)*x[0]+(0.560548548912)*x[1]**o+(0.220532224022)*x[1]
        arg[1,4,2,1]=(0.307927943715)*x[0]**o+(0.496390872708)*x[0]+(0.982750813274)*x[1]**o+(0.392968327778)*x[1]
        arg[2,0,0,0]=(0.645991188948)*x[0]**o+(0.104035822886)*x[0]+(0.18285829097)*x[1]**o+(0.51186286294)*x[1]
        arg[2,0,0,1]=(-0.996971717318)*x[0]**o+(0.126488047141)*x[0]+(0.123585312562)*x[1]**o+(-0.775854123529)*x[1]
        arg[2,0,1,0]=(-0.0258590124655)*x[0]**o+(0.0079454735671)*x[0]+(-0.42819024422)*x[1]**o+(0.12377360967)*x[1]
        arg[2,0,1,1]=(0.990441351225)*x[0]**o+(-0.453806376472)*x[0]+(0.134316618645)*x[1]**o+(-0.902666953282)*x[1]
        arg[2,0,2,0]=(0.149304385108)*x[0]**o+(-0.226132046999)*x[0]+(0.647186106203)*x[1]**o+(-0.269247029225)*x[1]
        arg[2,0,2,1]=(-0.159068164422)*x[0]**o+(-0.392275577366)*x[0]+(0.131423093404)*x[1]**o+(-0.353311282658)*x[1]
        arg[2,1,0,0]=(-0.2862918681)*x[0]**o+(-0.392264098939)*x[0]+(-0.344033237285)*x[1]**o+(0.644261713963)*x[1]
        arg[2,1,0,1]=(0.398022911658)*x[0]**o+(-0.84552546032)*x[0]+(-0.823861260943)*x[1]**o+(0.739994988755)*x[1]
        arg[2,1,1,0]=(-0.696854341234)*x[0]**o+(-0.810851183298)*x[0]+(-0.712182042896)*x[1]**o+(0.594887691783)*x[1]
        arg[2,1,1,1]=(-0.182583840117)*x[0]**o+(0.962594377088)*x[0]+(-0.0157575661453)*x[1]**o+(-0.432167995007)*x[1]
        arg[2,1,2,0]=(0.293029196216)*x[0]**o+(0.46839918212)*x[0]+(0.262375931577)*x[1]**o+(0.0348536956306)*x[1]
        arg[2,1,2,1]=(0.775632135095)*x[0]**o+(-0.874598534107)*x[0]+(0.711602976736)*x[1]**o+(0.37065543293)*x[1]
        arg[2,2,0,0]=(-0.103625214158)*x[0]**o+(0.553558117877)*x[0]+(-0.0522465477589)*x[1]**o+(-0.747139195529)*x[1]
        arg[2,2,0,1]=(0.507427663615)*x[0]**o+(-0.536361186134)*x[0]+(0.809054726193)*x[1]**o+(-0.979546347794)*x[1]
        arg[2,2,1,0]=(0.786210623833)*x[0]**o+(0.244510127091)*x[0]+(0.159799334111)*x[1]**o+(0.0499263839273)*x[1]
        arg[2,2,1,1]=(0.454205244371)*x[0]**o+(-0.650927822092)*x[0]+(-0.939252656087)*x[1]**o+(-0.894014755314)*x[1]
        arg[2,2,2,0]=(-0.829617920904)*x[0]**o+(0.764070900994)*x[0]+(0.972419443556)*x[1]**o+(-0.824913543157)*x[1]
        arg[2,2,2,1]=(0.694310748478)*x[0]**o+(-0.191413364091)*x[0]+(0.588258464472)*x[1]**o+(0.743973223755)*x[1]
        arg[2,3,0,0]=(-0.588397365184)*x[0]**o+(-0.634207034861)*x[0]+(0.0853498555546)*x[1]**o+(0.559933556865)*x[1]
        arg[2,3,0,1]=(0.626829165689)*x[0]**o+(0.942185296616)*x[0]+(0.0360770405565)*x[1]**o+(0.635966218547)*x[1]
        arg[2,3,1,0]=(-0.0103526873852)*x[0]**o+(0.546330663496)*x[0]+(0.799224103454)*x[1]**o+(0.791922324897)*x[1]
        arg[2,3,1,1]=(0.921564474495)*x[0]**o+(-0.420410835144)*x[0]+(-0.20482503843)*x[1]**o+(-0.483331643037)*x[1]
        arg[2,3,2,0]=(-0.903600870314)*x[0]**o+(0.901956716274)*x[0]+(-0.113926342023)*x[1]**o+(-0.061596184768)*x[1]
        arg[2,3,2,1]=(0.653250267262)*x[0]**o+(-0.13770933726)*x[0]+(-0.0791563636079)*x[1]**o+(-0.578313150893)*x[1]
        arg[2,4,0,0]=(0.993903405725)*x[0]**o+(0.694324226249)*x[0]+(0.203872883541)*x[1]**o+(-0.526880767264)*x[1]
        arg[2,4,0,1]=(0.0120058513647)*x[0]**o+(0.00907603337269)*x[0]+(0.0247958601677)*x[1]**o+(-0.996916120875)*x[1]
        arg[2,4,1,0]=(0.281518049478)*x[0]**o+(0.0773799323777)*x[0]+(-0.968565098773)*x[1]**o+(-0.275558076229)*x[1]
        arg[2,4,1,1]=(0.565275744757)*x[0]**o+(0.978636675869)*x[0]+(-0.352951657751)*x[1]**o+(-0.54170487914)*x[1]
        arg[2,4,2,0]=(-0.571207550678)*x[0]**o+(0.196754283318)*x[0]+(0.665558917127)*x[1]**o+(0.361915269355)*x[1]
        arg[2,4,2,1]=(0.54579191885)*x[0]**o+(0.798319371798)*x[0]+(0.961669996133)*x[1]**o+(-0.549305529342)*x[1]
        arg[3,0,0,0]=(0.264064946121)*x[0]**o+(-0.314396185593)*x[0]+(-0.017492206135)*x[1]**o+(0.190513393091)*x[1]
        arg[3,0,0,1]=(0.812213666919)*x[0]**o+(-0.757919357489)*x[0]+(0.294829087285)*x[1]**o+(0.916181101963)*x[1]
        arg[3,0,1,0]=(-0.529906091574)*x[0]**o+(0.553682716765)*x[0]+(-0.693177625995)*x[1]**o+(0.377145561683)*x[1]
        arg[3,0,1,1]=(-0.298942248169)*x[0]**o+(0.876878593705)*x[0]+(-0.609807982832)*x[1]**o+(-0.574954919539)*x[1]
        arg[3,0,2,0]=(-0.729324018245)*x[0]**o+(-0.495194565254)*x[0]+(-0.616162751356)*x[1]**o+(0.878549207376)*x[1]
        arg[3,0,2,1]=(0.625373988248)*x[0]**o+(0.106736108396)*x[0]+(-0.871225833944)*x[1]**o+(0.840614493429)*x[1]
        arg[3,1,0,0]=(-0.90880839589)*x[0]**o+(0.390298288126)*x[0]+(0.446872042362)*x[1]**o+(-0.0703294219415)*x[1]
        arg[3,1,0,1]=(-0.0706115592617)*x[0]**o+(-0.158386152093)*x[0]+(-0.687459181052)*x[1]**o+(-0.251170026598)*x[1]
        arg[3,1,1,0]=(-0.286907297081)*x[0]**o+(0.185480985579)*x[0]+(-0.49763263834)*x[1]**o+(-0.773056048642)*x[1]
        arg[3,1,1,1]=(-0.849572703562)*x[0]**o+(0.481739532516)*x[0]+(-0.227573063955)*x[1]**o+(-0.282367611381)*x[1]
        arg[3,1,2,0]=(-0.524975224603)*x[0]**o+(-0.795450756917)*x[0]+(-0.656865716441)*x[1]**o+(-0.0166902340102)*x[1]
        arg[3,1,2,1]=(-0.297441818145)*x[0]**o+(-0.726759852457)*x[0]+(0.146859888138)*x[1]**o+(-0.73690394554)*x[1]
        arg[3,2,0,0]=(-0.278794538596)*x[0]**o+(-0.463428732496)*x[0]+(-0.478500378607)*x[1]**o+(0.806643392698)*x[1]
        arg[3,2,0,1]=(0.573853299578)*x[0]**o+(-0.324830388013)*x[0]+(-0.185528736416)*x[1]**o+(0.444228024864)*x[1]
        arg[3,2,1,0]=(0.686054084195)*x[0]**o+(-0.914380567743)*x[0]+(0.751959744445)*x[1]**o+(0.531759675659)*x[1]
        arg[3,2,1,1]=(-0.876330263589)*x[0]**o+(0.566155144894)*x[0]+(-0.0233988901429)*x[1]**o+(0.42480184987)*x[1]
        arg[3,2,2,0]=(0.957053098936)*x[0]**o+(0.152292007141)*x[0]+(0.332521504492)*x[1]**o+(-0.289068173241)*x[1]
        arg[3,2,2,1]=(0.923170924493)*x[0]**o+(0.191154526504)*x[0]+(0.647848811717)*x[1]**o+(0.72509033544)*x[1]
        arg[3,3,0,0]=(0.388829993841)*x[0]**o+(0.243404004988)*x[0]+(-0.0313000323506)*x[1]**o+(0.558135056894)*x[1]
        arg[3,3,0,1]=(-0.784568790167)*x[0]**o+(0.742074534256)*x[0]+(-0.624869261724)*x[1]**o+(-0.412100194843)*x[1]
        arg[3,3,1,0]=(0.924173486785)*x[0]**o+(-0.966635363007)*x[0]+(0.395550623492)*x[1]**o+(0.469204722173)*x[1]
        arg[3,3,1,1]=(-0.842963105645)*x[0]**o+(0.685039360106)*x[0]+(0.279605123662)*x[1]**o+(-0.953092961091)*x[1]
        arg[3,3,2,0]=(0.598860382403)*x[0]**o+(-0.226413712435)*x[0]+(0.722250410625)*x[1]**o+(-0.944980469407)*x[1]
        arg[3,3,2,1]=(0.787289357927)*x[0]**o+(-0.270737247579)*x[0]+(-0.134351386793)*x[1]**o+(-0.0437387143766)*x[1]
        arg[3,4,0,0]=(-0.0601939617429)*x[0]**o+(-0.212346121247)*x[0]+(0.0137510326612)*x[1]**o+(0.749376479477)*x[1]
        arg[3,4,0,1]=(-0.515391847891)*x[0]**o+(-0.191867157458)*x[0]+(0.0883438399597)*x[1]**o+(-0.155095894625)*x[1]
        arg[3,4,1,0]=(0.172678225476)*x[0]**o+(-0.917390756059)*x[0]+(-0.498696112089)*x[1]**o+(-0.267655707135)*x[1]
        arg[3,4,1,1]=(-0.498294994476)*x[0]**o+(-0.301078252047)*x[0]+(0.100215523329)*x[1]**o+(-0.424039854141)*x[1]
        arg[3,4,2,0]=(0.904127003142)*x[0]**o+(-0.70704194874)*x[0]+(0.309638449466)*x[1]**o+(-0.290894515899)*x[1]
        arg[3,4,2,1]=(-0.924994398191)*x[0]**o+(0.624996423224)*x[0]+(0.461586866277)*x[1]**o+(-0.499709859121)*x[1]
        ref[0,0,0,0]=(-0.633576516648)*x_ref[0]**o+(0.77399256211)*x_ref[0]+(-0.396696417195)*x_ref[1]**o+(-0.604572540725)*x_ref[1]
        ref[0,0,0,1]=(0.612672991978)*x_ref[0]**o+(-0.331012302611)*x_ref[0]+(0.885533578235)*x_ref[1]**o+(0.334686818058)*x_ref[1]
        ref[0,0,1,0]=(-0.982701069025)*x_ref[0]**o+(0.467849033642)*x_ref[0]+(0.525392773375)*x_ref[1]**o+(0.355622324425)*x_ref[1]
        ref[0,0,1,1]=(-0.0243671469666)*x_ref[0]**o+(0.790044842383)*x_ref[0]+(0.413384172033)*x_ref[1]**o+(0.740805135968)*x_ref[1]
        ref[0,0,2,0]=(0.0384789957691)*x_ref[0]**o+(-0.317009899417)*x_ref[0]+(0.0389710369615)*x_ref[1]**o+(0.784549852598)*x_ref[1]
        ref[0,0,2,1]=(0.361832259252)*x_ref[0]**o+(0.410815792944)*x_ref[0]+(0.39496145831)*x_ref[1]**o+(-0.820784143964)*x_ref[1]
        ref[0,1,0,0]=(0.224261088976)*x_ref[0]**o+(-0.824694527256)*x_ref[0]+(0.318674472135)*x_ref[1]**o+(0.827696354338)*x_ref[1]
        ref[0,1,0,1]=(-0.442154901359)*x_ref[0]**o+(0.141775320862)*x_ref[0]+(0.688711841268)*x_ref[1]**o+(-0.27047542755)*x_ref[1]
        ref[0,1,1,0]=(-0.570125802019)*x_ref[0]**o+(0.050303762895)*x_ref[0]+(-0.972329837417)*x_ref[1]**o+(0.143700303765)*x_ref[1]
        ref[0,1,1,1]=(0.852888550499)*x_ref[0]**o+(-0.51216090329)*x_ref[0]+(-0.895663922335)*x_ref[1]**o+(-0.183466225547)*x_ref[1]
        ref[0,1,2,0]=(0.863840920739)*x_ref[0]**o+(-0.247789353524)*x_ref[0]+(0.63056436157)*x_ref[1]**o+(0.932522589942)*x_ref[1]
        ref[0,1,2,1]=(-0.403832121794)*x_ref[0]**o+(-0.769625259789)*x_ref[0]+(0.507724303191)*x_ref[1]**o+(0.314178775526)*x_ref[1]
        ref[0,2,0,0]=(0.543867153515)*x_ref[0]**o+(-0.414107817028)*x_ref[0]+(0.423922720842)*x_ref[1]**o+(0.166632407916)*x_ref[1]
        ref[0,2,0,1]=(0.731968209929)*x_ref[0]**o+(-0.114742736683)*x_ref[0]+(0.112562415021)*x_ref[1]**o+(0.166924480325)*x_ref[1]
        ref[0,2,1,0]=(0.919472066242)*x_ref[0]**o+(0.560565315205)*x_ref[0]+(0.4967650028)*x_ref[1]**o+(0.708376671625)*x_ref[1]
        ref[0,2,1,1]=(0.147500239739)*x_ref[0]**o+(-0.632670389919)*x_ref[0]+(-0.992479063437)*x_ref[1]**o+(0.258941122962)*x_ref[1]
        ref[0,2,2,0]=(0.474481785329)*x_ref[0]**o+(0.592902604094)*x_ref[0]+(0.64058685211)*x_ref[1]**o+(-0.477843307901)*x_ref[1]
        ref[0,2,2,1]=(0.769898355146)*x_ref[0]**o+(0.563458279487)*x_ref[0]+(0.876276425449)*x_ref[1]**o+(-0.0625727749869)*x_ref[1]
        ref[0,3,0,0]=(-0.330055037059)*x_ref[0]**o+(-0.727656105163)*x_ref[0]+(-0.559901684392)*x_ref[1]**o+(-0.433646878263)*x_ref[1]
        ref[0,3,0,1]=(-0.725314328666)*x_ref[0]**o+(0.638326609105)*x_ref[0]+(-0.126619368417)*x_ref[1]**o+(-0.232194758676)*x_ref[1]
        ref[0,3,1,0]=(0.318096005545)*x_ref[0]**o+(0.325247494289)*x_ref[0]+(-0.514305253643)*x_ref[1]**o+(0.108523350086)*x_ref[1]
        ref[0,3,1,1]=(0.298034894956)*x_ref[0]**o+(0.825714393651)*x_ref[0]+(-0.67162850079)*x_ref[1]**o+(0.681809083004)*x_ref[1]
        ref[0,3,2,0]=(-0.00715890582648)*x_ref[0]**o+(0.893900731227)*x_ref[0]+(-0.773710274545)*x_ref[1]**o+(0.415913728569)*x_ref[1]
        ref[0,3,2,1]=(0.0853551940958)*x_ref[0]**o+(-0.228876011127)*x_ref[0]+(0.87699456762)*x_ref[1]**o+(-0.744605229253)*x_ref[1]
        ref[0,4,0,0]=(0.942646937302)*x_ref[0]**o+(0.92717236759)*x_ref[0]+(0.794806821569)*x_ref[1]**o+(0.848766303526)*x_ref[1]
        ref[0,4,0,1]=(0.20140672561)*x_ref[0]**o+(0.974490372414)*x_ref[0]+(-0.450332419587)*x_ref[1]**o+(-0.859102184782)*x_ref[1]
        ref[0,4,1,0]=(0.632153053193)*x_ref[0]**o+(0.702336756277)*x_ref[0]+(-0.571496076236)*x_ref[1]**o+(-0.882472852219)*x_ref[1]
        ref[0,4,1,1]=(0.366452829499)*x_ref[0]**o+(-0.514660164741)*x_ref[0]+(-0.152794832253)*x_ref[1]**o+(-0.658688117796)*x_ref[1]
        ref[0,4,2,0]=(0.406218394702)*x_ref[0]**o+(0.782359698198)*x_ref[0]+(-0.0582499587562)*x_ref[1]**o+(0.664506852595)*x_ref[1]
        ref[0,4,2,1]=(0.768240870343)*x_ref[0]**o+(0.188470592904)*x_ref[0]+(0.855933509171)*x_ref[1]**o+(0.639174087541)*x_ref[1]
        ref[1,0,0,0]=(-0.946841398046)*x_ref[0]**o+(0.892729256183)*x_ref[0]+(-0.15644094115)*x_ref[1]**o+(-0.859371368233)*x_ref[1]
        ref[1,0,0,1]=(0.238113647666)*x_ref[0]**o+(0.173695470467)*x_ref[0]+(-0.753529252313)*x_ref[1]**o+(-0.243203902463)*x_ref[1]
        ref[1,0,1,0]=(0.267937523443)*x_ref[0]**o+(0.600200492374)*x_ref[0]+(0.442653529257)*x_ref[1]**o+(-0.394984210541)*x_ref[1]
        ref[1,0,1,1]=(-0.495022907622)*x_ref[0]**o+(-0.593273808093)*x_ref[0]+(0.888764746608)*x_ref[1]**o+(-0.745127825763)*x_ref[1]
        ref[1,0,2,0]=(0.352827238496)*x_ref[0]**o+(0.293200928839)*x_ref[0]+(-0.00369812109034)*x_ref[1]**o+(0.478875774193)*x_ref[1]
        ref[1,0,2,1]=(-0.403358878981)*x_ref[0]**o+(0.756947364667)*x_ref[0]+(0.649335795039)*x_ref[1]**o+(-0.339962796859)*x_ref[1]
        ref[1,1,0,0]=(-0.607505349015)*x_ref[0]**o+(-0.392535399582)*x_ref[0]+(0.799782389634)*x_ref[1]**o+(0.186339165863)*x_ref[1]
        ref[1,1,0,1]=(0.997492304627)*x_ref[0]**o+(-0.281456622581)*x_ref[0]+(-0.959294693613)*x_ref[1]**o+(-0.454812606674)*x_ref[1]
        ref[1,1,1,0]=(0.454682037549)*x_ref[0]**o+(0.797129935647)*x_ref[0]+(-0.949539573053)*x_ref[1]**o+(-0.918499382199)*x_ref[1]
        ref[1,1,1,1]=(0.903016449758)*x_ref[0]**o+(-0.575039046836)*x_ref[0]+(0.957158277074)*x_ref[1]**o+(-0.715797460906)*x_ref[1]
        ref[1,1,2,0]=(0.92005547899)*x_ref[0]**o+(-0.716975919302)*x_ref[0]+(-0.900904311708)*x_ref[1]**o+(0.989458809643)*x_ref[1]
        ref[1,1,2,1]=(-0.633214991087)*x_ref[0]**o+(0.680658578627)*x_ref[0]+(0.552194054556)*x_ref[1]**o+(0.343557072769)*x_ref[1]
        ref[1,2,0,0]=(-0.0182903530463)*x_ref[0]**o+(-0.0445569497267)*x_ref[0]+(-0.259920069365)*x_ref[1]**o+(0.118765787279)*x_ref[1]
        ref[1,2,0,1]=(-0.667382174733)*x_ref[0]**o+(-0.79898800024)*x_ref[0]+(-0.217638556567)*x_ref[1]**o+(-0.0288504127936)*x_ref[1]
        ref[1,2,1,0]=(-0.437982445862)*x_ref[0]**o+(-0.104054454808)*x_ref[0]+(0.711429718088)*x_ref[1]**o+(-0.607226296329)*x_ref[1]
        ref[1,2,1,1]=(0.252861218037)*x_ref[0]**o+(-0.635701456504)*x_ref[0]+(-0.555222304592)*x_ref[1]**o+(0.723744010092)*x_ref[1]
        ref[1,2,2,0]=(-0.73532710885)*x_ref[0]**o+(0.600599078741)*x_ref[0]+(0.0465795173194)*x_ref[1]**o+(0.524165544645)*x_ref[1]
        ref[1,2,2,1]=(-0.025246019744)*x_ref[0]**o+(-0.376526675619)*x_ref[0]+(0.213011788698)*x_ref[1]**o+(-0.477018225777)*x_ref[1]
        ref[1,3,0,0]=(-0.52682690957)*x_ref[0]**o+(0.0567472858499)*x_ref[0]+(0.574116256102)*x_ref[1]**o+(-0.130012791898)*x_ref[1]
        ref[1,3,0,1]=(0.113502182936)*x_ref[0]**o+(0.7643861313)*x_ref[0]+(0.48414948847)*x_ref[1]**o+(0.863168250318)*x_ref[1]
        ref[1,3,1,0]=(-0.208102162378)*x_ref[0]**o+(0.327770728981)*x_ref[0]+(0.843778895278)*x_ref[1]**o+(0.474843504177)*x_ref[1]
        ref[1,3,1,1]=(-0.549818192184)*x_ref[0]**o+(-0.863333816685)*x_ref[0]+(-0.365631583957)*x_ref[1]**o+(-0.491320441182)*x_ref[1]
        ref[1,3,2,0]=(-0.265833845759)*x_ref[0]**o+(-0.94416746081)*x_ref[0]+(0.48465442593)*x_ref[1]**o+(0.258284621408)*x_ref[1]
        ref[1,3,2,1]=(-0.625716082268)*x_ref[0]**o+(-0.234697130359)*x_ref[0]+(-0.973704406131)*x_ref[1]**o+(0.00141746117847)*x_ref[1]
        ref[1,4,0,0]=(0.565044361964)*x_ref[0]**o+(-0.0672474991748)*x_ref[0]+(-0.125901930524)*x_ref[1]**o+(-0.753103899732)*x_ref[1]
        ref[1,4,0,1]=(-0.979743467289)*x_ref[0]**o+(-0.25225085214)*x_ref[0]+(-0.141635164635)*x_ref[1]**o+(0.378153168667)*x_ref[1]
        ref[1,4,1,0]=(0.97728399206)*x_ref[0]**o+(-0.202043160532)*x_ref[0]+(0.997445424954)*x_ref[1]**o+(-0.100049830154)*x_ref[1]
        ref[1,4,1,1]=(0.364840105486)*x_ref[0]**o+(-0.682234327437)*x_ref[0]+(0.252584592259)*x_ref[1]**o+(-0.0785430242425)*x_ref[1]
        ref[1,4,2,0]=(-0.0168764722616)*x_ref[0]**o+(-0.997460139026)*x_ref[0]+(0.560548548912)*x_ref[1]**o+(0.220532224022)*x_ref[1]
        ref[1,4,2,1]=(0.307927943715)*x_ref[0]**o+(0.496390872708)*x_ref[0]+(0.982750813274)*x_ref[1]**o+(0.392968327778)*x_ref[1]
        ref[2,0,0,0]=(0.645991188948)*x_ref[0]**o+(0.104035822886)*x_ref[0]+(0.18285829097)*x_ref[1]**o+(0.51186286294)*x_ref[1]
        ref[2,0,0,1]=(-0.996971717318)*x_ref[0]**o+(0.126488047141)*x_ref[0]+(0.123585312562)*x_ref[1]**o+(-0.775854123529)*x_ref[1]
        ref[2,0,1,0]=(-0.0258590124655)*x_ref[0]**o+(0.0079454735671)*x_ref[0]+(-0.42819024422)*x_ref[1]**o+(0.12377360967)*x_ref[1]
        ref[2,0,1,1]=(0.990441351225)*x_ref[0]**o+(-0.453806376472)*x_ref[0]+(0.134316618645)*x_ref[1]**o+(-0.902666953282)*x_ref[1]
        ref[2,0,2,0]=(0.149304385108)*x_ref[0]**o+(-0.226132046999)*x_ref[0]+(0.647186106203)*x_ref[1]**o+(-0.269247029225)*x_ref[1]
        ref[2,0,2,1]=(-0.159068164422)*x_ref[0]**o+(-0.392275577366)*x_ref[0]+(0.131423093404)*x_ref[1]**o+(-0.353311282658)*x_ref[1]
        ref[2,1,0,0]=(-0.2862918681)*x_ref[0]**o+(-0.392264098939)*x_ref[0]+(-0.344033237285)*x_ref[1]**o+(0.644261713963)*x_ref[1]
        ref[2,1,0,1]=(0.398022911658)*x_ref[0]**o+(-0.84552546032)*x_ref[0]+(-0.823861260943)*x_ref[1]**o+(0.739994988755)*x_ref[1]
        ref[2,1,1,0]=(-0.696854341234)*x_ref[0]**o+(-0.810851183298)*x_ref[0]+(-0.712182042896)*x_ref[1]**o+(0.594887691783)*x_ref[1]
        ref[2,1,1,1]=(-0.182583840117)*x_ref[0]**o+(0.962594377088)*x_ref[0]+(-0.0157575661453)*x_ref[1]**o+(-0.432167995007)*x_ref[1]
        ref[2,1,2,0]=(0.293029196216)*x_ref[0]**o+(0.46839918212)*x_ref[0]+(0.262375931577)*x_ref[1]**o+(0.0348536956306)*x_ref[1]
        ref[2,1,2,1]=(0.775632135095)*x_ref[0]**o+(-0.874598534107)*x_ref[0]+(0.711602976736)*x_ref[1]**o+(0.37065543293)*x_ref[1]
        ref[2,2,0,0]=(-0.103625214158)*x_ref[0]**o+(0.553558117877)*x_ref[0]+(-0.0522465477589)*x_ref[1]**o+(-0.747139195529)*x_ref[1]
        ref[2,2,0,1]=(0.507427663615)*x_ref[0]**o+(-0.536361186134)*x_ref[0]+(0.809054726193)*x_ref[1]**o+(-0.979546347794)*x_ref[1]
        ref[2,2,1,0]=(0.786210623833)*x_ref[0]**o+(0.244510127091)*x_ref[0]+(0.159799334111)*x_ref[1]**o+(0.0499263839273)*x_ref[1]
        ref[2,2,1,1]=(0.454205244371)*x_ref[0]**o+(-0.650927822092)*x_ref[0]+(-0.939252656087)*x_ref[1]**o+(-0.894014755314)*x_ref[1]
        ref[2,2,2,0]=(-0.829617920904)*x_ref[0]**o+(0.764070900994)*x_ref[0]+(0.972419443556)*x_ref[1]**o+(-0.824913543157)*x_ref[1]
        ref[2,2,2,1]=(0.694310748478)*x_ref[0]**o+(-0.191413364091)*x_ref[0]+(0.588258464472)*x_ref[1]**o+(0.743973223755)*x_ref[1]
        ref[2,3,0,0]=(-0.588397365184)*x_ref[0]**o+(-0.634207034861)*x_ref[0]+(0.0853498555546)*x_ref[1]**o+(0.559933556865)*x_ref[1]
        ref[2,3,0,1]=(0.626829165689)*x_ref[0]**o+(0.942185296616)*x_ref[0]+(0.0360770405565)*x_ref[1]**o+(0.635966218547)*x_ref[1]
        ref[2,3,1,0]=(-0.0103526873852)*x_ref[0]**o+(0.546330663496)*x_ref[0]+(0.799224103454)*x_ref[1]**o+(0.791922324897)*x_ref[1]
        ref[2,3,1,1]=(0.921564474495)*x_ref[0]**o+(-0.420410835144)*x_ref[0]+(-0.20482503843)*x_ref[1]**o+(-0.483331643037)*x_ref[1]
        ref[2,3,2,0]=(-0.903600870314)*x_ref[0]**o+(0.901956716274)*x_ref[0]+(-0.113926342023)*x_ref[1]**o+(-0.061596184768)*x_ref[1]
        ref[2,3,2,1]=(0.653250267262)*x_ref[0]**o+(-0.13770933726)*x_ref[0]+(-0.0791563636079)*x_ref[1]**o+(-0.578313150893)*x_ref[1]
        ref[2,4,0,0]=(0.993903405725)*x_ref[0]**o+(0.694324226249)*x_ref[0]+(0.203872883541)*x_ref[1]**o+(-0.526880767264)*x_ref[1]
        ref[2,4,0,1]=(0.0120058513647)*x_ref[0]**o+(0.00907603337269)*x_ref[0]+(0.0247958601677)*x_ref[1]**o+(-0.996916120875)*x_ref[1]
        ref[2,4,1,0]=(0.281518049478)*x_ref[0]**o+(0.0773799323777)*x_ref[0]+(-0.968565098773)*x_ref[1]**o+(-0.275558076229)*x_ref[1]
        ref[2,4,1,1]=(0.565275744757)*x_ref[0]**o+(0.978636675869)*x_ref[0]+(-0.352951657751)*x_ref[1]**o+(-0.54170487914)*x_ref[1]
        ref[2,4,2,0]=(-0.571207550678)*x_ref[0]**o+(0.196754283318)*x_ref[0]+(0.665558917127)*x_ref[1]**o+(0.361915269355)*x_ref[1]
        ref[2,4,2,1]=(0.54579191885)*x_ref[0]**o+(0.798319371798)*x_ref[0]+(0.961669996133)*x_ref[1]**o+(-0.549305529342)*x_ref[1]
        ref[3,0,0,0]=(0.264064946121)*x_ref[0]**o+(-0.314396185593)*x_ref[0]+(-0.017492206135)*x_ref[1]**o+(0.190513393091)*x_ref[1]
        ref[3,0,0,1]=(0.812213666919)*x_ref[0]**o+(-0.757919357489)*x_ref[0]+(0.294829087285)*x_ref[1]**o+(0.916181101963)*x_ref[1]
        ref[3,0,1,0]=(-0.529906091574)*x_ref[0]**o+(0.553682716765)*x_ref[0]+(-0.693177625995)*x_ref[1]**o+(0.377145561683)*x_ref[1]
        ref[3,0,1,1]=(-0.298942248169)*x_ref[0]**o+(0.876878593705)*x_ref[0]+(-0.609807982832)*x_ref[1]**o+(-0.574954919539)*x_ref[1]
        ref[3,0,2,0]=(-0.729324018245)*x_ref[0]**o+(-0.495194565254)*x_ref[0]+(-0.616162751356)*x_ref[1]**o+(0.878549207376)*x_ref[1]
        ref[3,0,2,1]=(0.625373988248)*x_ref[0]**o+(0.106736108396)*x_ref[0]+(-0.871225833944)*x_ref[1]**o+(0.840614493429)*x_ref[1]
        ref[3,1,0,0]=(-0.90880839589)*x_ref[0]**o+(0.390298288126)*x_ref[0]+(0.446872042362)*x_ref[1]**o+(-0.0703294219415)*x_ref[1]
        ref[3,1,0,1]=(-0.0706115592617)*x_ref[0]**o+(-0.158386152093)*x_ref[0]+(-0.687459181052)*x_ref[1]**o+(-0.251170026598)*x_ref[1]
        ref[3,1,1,0]=(-0.286907297081)*x_ref[0]**o+(0.185480985579)*x_ref[0]+(-0.49763263834)*x_ref[1]**o+(-0.773056048642)*x_ref[1]
        ref[3,1,1,1]=(-0.849572703562)*x_ref[0]**o+(0.481739532516)*x_ref[0]+(-0.227573063955)*x_ref[1]**o+(-0.282367611381)*x_ref[1]
        ref[3,1,2,0]=(-0.524975224603)*x_ref[0]**o+(-0.795450756917)*x_ref[0]+(-0.656865716441)*x_ref[1]**o+(-0.0166902340102)*x_ref[1]
        ref[3,1,2,1]=(-0.297441818145)*x_ref[0]**o+(-0.726759852457)*x_ref[0]+(0.146859888138)*x_ref[1]**o+(-0.73690394554)*x_ref[1]
        ref[3,2,0,0]=(-0.278794538596)*x_ref[0]**o+(-0.463428732496)*x_ref[0]+(-0.478500378607)*x_ref[1]**o+(0.806643392698)*x_ref[1]
        ref[3,2,0,1]=(0.573853299578)*x_ref[0]**o+(-0.324830388013)*x_ref[0]+(-0.185528736416)*x_ref[1]**o+(0.444228024864)*x_ref[1]
        ref[3,2,1,0]=(0.686054084195)*x_ref[0]**o+(-0.914380567743)*x_ref[0]+(0.751959744445)*x_ref[1]**o+(0.531759675659)*x_ref[1]
        ref[3,2,1,1]=(-0.876330263589)*x_ref[0]**o+(0.566155144894)*x_ref[0]+(-0.0233988901429)*x_ref[1]**o+(0.42480184987)*x_ref[1]
        ref[3,2,2,0]=(0.957053098936)*x_ref[0]**o+(0.152292007141)*x_ref[0]+(0.332521504492)*x_ref[1]**o+(-0.289068173241)*x_ref[1]
        ref[3,2,2,1]=(0.923170924493)*x_ref[0]**o+(0.191154526504)*x_ref[0]+(0.647848811717)*x_ref[1]**o+(0.72509033544)*x_ref[1]
        ref[3,3,0,0]=(0.388829993841)*x_ref[0]**o+(0.243404004988)*x_ref[0]+(-0.0313000323506)*x_ref[1]**o+(0.558135056894)*x_ref[1]
        ref[3,3,0,1]=(-0.784568790167)*x_ref[0]**o+(0.742074534256)*x_ref[0]+(-0.624869261724)*x_ref[1]**o+(-0.412100194843)*x_ref[1]
        ref[3,3,1,0]=(0.924173486785)*x_ref[0]**o+(-0.966635363007)*x_ref[0]+(0.395550623492)*x_ref[1]**o+(0.469204722173)*x_ref[1]
        ref[3,3,1,1]=(-0.842963105645)*x_ref[0]**o+(0.685039360106)*x_ref[0]+(0.279605123662)*x_ref[1]**o+(-0.953092961091)*x_ref[1]
        ref[3,3,2,0]=(0.598860382403)*x_ref[0]**o+(-0.226413712435)*x_ref[0]+(0.722250410625)*x_ref[1]**o+(-0.944980469407)*x_ref[1]
        ref[3,3,2,1]=(0.787289357927)*x_ref[0]**o+(-0.270737247579)*x_ref[0]+(-0.134351386793)*x_ref[1]**o+(-0.0437387143766)*x_ref[1]
        ref[3,4,0,0]=(-0.0601939617429)*x_ref[0]**o+(-0.212346121247)*x_ref[0]+(0.0137510326612)*x_ref[1]**o+(0.749376479477)*x_ref[1]
        ref[3,4,0,1]=(-0.515391847891)*x_ref[0]**o+(-0.191867157458)*x_ref[0]+(0.0883438399597)*x_ref[1]**o+(-0.155095894625)*x_ref[1]
        ref[3,4,1,0]=(0.172678225476)*x_ref[0]**o+(-0.917390756059)*x_ref[0]+(-0.498696112089)*x_ref[1]**o+(-0.267655707135)*x_ref[1]
        ref[3,4,1,1]=(-0.498294994476)*x_ref[0]**o+(-0.301078252047)*x_ref[0]+(0.100215523329)*x_ref[1]**o+(-0.424039854141)*x_ref[1]
        ref[3,4,2,0]=(0.904127003142)*x_ref[0]**o+(-0.70704194874)*x_ref[0]+(0.309638449466)*x_ref[1]**o+(-0.290894515899)*x_ref[1]
        ref[3,4,2,1]=(-0.924994398191)*x_ref[0]**o+(0.624996423224)*x_ref[0]+(0.461586866277)*x_ref[1]**o+(-0.499709859121)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.0998996137438)*x[0]**o+(-0.648495171588)*x[0]+(-0.836177837758)*x[1]**o+(-0.9597381047)*x[1]+(-0.958069967907)*x[2]**o+(0.399097251753)*x[2]
        arg[0,0,0,1]=(0.625409406207)*x[0]**o+(-0.522476054218)*x[0]+(0.794724584797)*x[1]**o+(-0.450875661781)*x[1]+(-0.189505729613)*x[2]**o+(0.0910808218379)*x[2]
        arg[0,0,1,0]=(-0.36073198549)*x[0]**o+(-0.0286144078107)*x[0]+(-0.53827851785)*x[1]**o+(-0.563600303848)*x[1]+(0.970211345168)*x[2]**o+(-0.405253675856)*x[2]
        arg[0,0,1,1]=(-0.0204302558323)*x[0]**o+(0.927570053009)*x[0]+(-0.358571532503)*x[1]**o+(0.282513525404)*x[1]+(0.105530612734)*x[2]**o+(-0.156753729378)*x[2]
        arg[0,0,2,0]=(0.25049329858)*x[0]**o+(0.412299654624)*x[0]+(0.827683104949)*x[1]**o+(0.0195885473957)*x[1]+(0.849521623993)*x[2]**o+(-0.107922097347)*x[2]
        arg[0,0,2,1]=(0.824054630326)*x[0]**o+(0.315612619612)*x[0]+(-0.33528975625)*x[1]**o+(-0.848543364468)*x[1]+(0.67462279419)*x[2]**o+(-0.0387606720463)*x[2]
        arg[0,1,0,0]=(0.251613993994)*x[0]**o+(-0.871217704935)*x[0]+(-0.648182240752)*x[1]**o+(0.324536900202)*x[1]+(-0.884871901147)*x[2]**o+(0.353160185772)*x[2]
        arg[0,1,0,1]=(-0.629894976165)*x[0]**o+(0.99092287348)*x[0]+(0.87471936536)*x[1]**o+(0.732884844801)*x[1]+(0.423531105096)*x[2]**o+(0.648959408663)*x[2]
        arg[0,1,1,0]=(0.123366820062)*x[0]**o+(-0.856900246151)*x[0]+(-0.887819649211)*x[1]**o+(-0.901367472665)*x[1]+(0.222535500196)*x[2]**o+(0.42830671462)*x[2]
        arg[0,1,1,1]=(0.343789367141)*x[0]**o+(0.614284333526)*x[0]+(0.976225499577)*x[1]**o+(0.955115562801)*x[1]+(0.0700409126681)*x[2]**o+(-0.709602275548)*x[2]
        arg[0,1,2,0]=(-0.863502900367)*x[0]**o+(-0.452731447667)*x[0]+(0.495519161334)*x[1]**o+(0.0254261045314)*x[1]+(-0.484479110985)*x[2]**o+(0.615421138564)*x[2]
        arg[0,1,2,1]=(-0.861966549825)*x[0]**o+(0.957385177208)*x[0]+(0.710449966969)*x[1]**o+(-0.367149885665)*x[1]+(-0.100800220614)*x[2]**o+(-0.266781849874)*x[2]
        arg[0,2,0,0]=(-0.781309947495)*x[0]**o+(-0.709036446117)*x[0]+(0.86136976691)*x[1]**o+(-0.745743057084)*x[1]+(0.519795761984)*x[2]**o+(0.298838092751)*x[2]
        arg[0,2,0,1]=(-0.514037154361)*x[0]**o+(-0.155867632386)*x[0]+(-0.125008826668)*x[1]**o+(-0.138197902584)*x[1]+(0.360698647284)*x[2]**o+(0.432385729464)*x[2]
        arg[0,2,1,0]=(-0.19580514344)*x[0]**o+(-0.430510395896)*x[0]+(0.676498502854)*x[1]**o+(-0.472411864377)*x[1]+(-0.323468590594)*x[2]**o+(0.837971311859)*x[2]
        arg[0,2,1,1]=(-0.358714577548)*x[0]**o+(-0.583080123482)*x[0]+(-0.225887451598)*x[1]**o+(0.744101609163)*x[1]+(-0.0300539574094)*x[2]**o+(-0.985181847505)*x[2]
        arg[0,2,2,0]=(0.820610118447)*x[0]**o+(-0.560334915671)*x[0]+(0.478672329569)*x[1]**o+(-0.956892480643)*x[1]+(0.302100338619)*x[2]**o+(0.597584594886)*x[2]
        arg[0,2,2,1]=(-0.0658925658941)*x[0]**o+(-0.136738027019)*x[0]+(0.617259524472)*x[1]**o+(-0.486181324584)*x[1]+(0.225924222893)*x[2]**o+(-0.304893108343)*x[2]
        arg[0,3,0,0]=(0.0824288958004)*x[0]**o+(-0.67546530428)*x[0]+(-0.726843419073)*x[1]**o+(0.150753236521)*x[1]+(0.200621841908)*x[2]**o+(0.869623957921)*x[2]
        arg[0,3,0,1]=(0.66227806631)*x[0]**o+(-0.389229756136)*x[0]+(0.739548522805)*x[1]**o+(0.131026218939)*x[1]+(0.743271925877)*x[2]**o+(-0.910220233981)*x[2]
        arg[0,3,1,0]=(-0.259647259175)*x[0]**o+(-0.208585364012)*x[0]+(-0.323867955964)*x[1]**o+(0.778380345151)*x[1]+(-0.555780852614)*x[2]**o+(-0.16366483426)*x[2]
        arg[0,3,1,1]=(0.248968056109)*x[0]**o+(0.84999990087)*x[0]+(0.180937025847)*x[1]**o+(0.332075828156)*x[1]+(-0.00577042146131)*x[2]**o+(0.708188905312)*x[2]
        arg[0,3,2,0]=(-0.80327596428)*x[0]**o+(-0.131004563471)*x[0]+(0.443094741792)*x[1]**o+(0.515867966266)*x[1]+(-0.964003373109)*x[2]**o+(-0.580721642935)*x[2]
        arg[0,3,2,1]=(-0.107361389717)*x[0]**o+(-0.543617395264)*x[0]+(0.840867379108)*x[1]**o+(-0.459941493027)*x[1]+(-0.300033656162)*x[2]**o+(0.440763554498)*x[2]
        arg[0,4,0,0]=(0.272916362055)*x[0]**o+(-0.554210888243)*x[0]+(-0.458102512273)*x[1]**o+(-0.54594406835)*x[1]+(0.570670715852)*x[2]**o+(0.792731049718)*x[2]
        arg[0,4,0,1]=(0.896330380602)*x[0]**o+(-0.929757950797)*x[0]+(0.229049814758)*x[1]**o+(0.1245974562)*x[1]+(0.853250033203)*x[2]**o+(-0.877052530974)*x[2]
        arg[0,4,1,0]=(0.76027467594)*x[0]**o+(-0.510095378376)*x[0]+(0.697258966361)*x[1]**o+(-0.44013714543)*x[1]+(0.617675705023)*x[2]**o+(-0.00118327085347)*x[2]
        arg[0,4,1,1]=(0.92472377319)*x[0]**o+(0.27783405618)*x[0]+(-0.24317632451)*x[1]**o+(-0.0264727232824)*x[1]+(0.428751701408)*x[2]**o+(0.184142273303)*x[2]
        arg[0,4,2,0]=(0.619110983031)*x[0]**o+(-0.375037512361)*x[0]+(0.298576843317)*x[1]**o+(-0.218802335776)*x[1]+(-0.394889850336)*x[2]**o+(-0.842359750924)*x[2]
        arg[0,4,2,1]=(0.899714346356)*x[0]**o+(0.0630133184109)*x[0]+(0.0796796488932)*x[1]**o+(-0.693371400327)*x[1]+(-0.412090402419)*x[2]**o+(0.681871832039)*x[2]
        arg[1,0,0,0]=(-0.0367677770358)*x[0]**o+(-0.75352301678)*x[0]+(0.0705016755404)*x[1]**o+(0.870115630516)*x[1]+(0.136757694474)*x[2]**o+(-0.557722157484)*x[2]
        arg[1,0,0,1]=(0.550371937413)*x[0]**o+(-0.590052751632)*x[0]+(0.38665744963)*x[1]**o+(-0.896537565188)*x[1]+(0.550869514185)*x[2]**o+(0.693828237691)*x[2]
        arg[1,0,1,0]=(-0.150321579562)*x[0]**o+(0.111135028893)*x[0]+(0.582595079816)*x[1]**o+(-0.250920804447)*x[1]+(0.392998664266)*x[2]**o+(-0.226735735878)*x[2]
        arg[1,0,1,1]=(0.435196573865)*x[0]**o+(-0.31161203165)*x[0]+(-0.117580582163)*x[1]**o+(0.99583282896)*x[1]+(-0.164872214322)*x[2]**o+(0.243837126961)*x[2]
        arg[1,0,2,0]=(-0.320135099791)*x[0]**o+(0.339871290643)*x[0]+(0.593784119675)*x[1]**o+(0.695166367768)*x[1]+(0.407551427099)*x[2]**o+(-0.768445418933)*x[2]
        arg[1,0,2,1]=(-0.390718323619)*x[0]**o+(-0.0515766643887)*x[0]+(0.394112545837)*x[1]**o+(-0.379592335881)*x[1]+(-0.29680237927)*x[2]**o+(0.500583589142)*x[2]
        arg[1,1,0,0]=(-0.958930666062)*x[0]**o+(0.516874100468)*x[0]+(0.0859034666349)*x[1]**o+(-0.370561856015)*x[1]+(-0.175293417132)*x[2]**o+(0.672554615126)*x[2]
        arg[1,1,0,1]=(-0.38129381629)*x[0]**o+(-0.615636530444)*x[0]+(0.447284140424)*x[1]**o+(0.507583882952)*x[1]+(-0.942413761354)*x[2]**o+(-0.607928807162)*x[2]
        arg[1,1,1,0]=(-0.456462200294)*x[0]**o+(0.778557108965)*x[0]+(-0.939624730616)*x[1]**o+(0.965857713341)*x[1]+(-0.913772635427)*x[2]**o+(-0.459963514384)*x[2]
        arg[1,1,1,1]=(0.138896120228)*x[0]**o+(0.941491395046)*x[0]+(0.546538931051)*x[1]**o+(-0.526098363992)*x[1]+(0.32139556996)*x[2]**o+(0.899346403792)*x[2]
        arg[1,1,2,0]=(0.591597103411)*x[0]**o+(0.926278363416)*x[0]+(0.565926211046)*x[1]**o+(0.930516925784)*x[1]+(-0.498131779353)*x[2]**o+(0.358157310681)*x[2]
        arg[1,1,2,1]=(0.758395199797)*x[0]**o+(0.769578766255)*x[0]+(0.86819976051)*x[1]**o+(-0.380436310842)*x[1]+(-0.513012037048)*x[2]**o+(-0.984361866072)*x[2]
        arg[1,2,0,0]=(0.0845846076213)*x[0]**o+(0.987145343601)*x[0]+(-0.447680898266)*x[1]**o+(-0.30278197108)*x[1]+(0.271667503636)*x[2]**o+(0.698967822656)*x[2]
        arg[1,2,0,1]=(-0.314076111079)*x[0]**o+(0.507407943673)*x[0]+(-0.38467148586)*x[1]**o+(0.802764957621)*x[1]+(0.692949890402)*x[2]**o+(-0.965766183849)*x[2]
        arg[1,2,1,0]=(-0.855806343042)*x[0]**o+(0.753362713203)*x[0]+(-0.9877924012)*x[1]**o+(-0.54047626663)*x[1]+(-0.488859780266)*x[2]**o+(0.444279382381)*x[2]
        arg[1,2,1,1]=(-0.313257382113)*x[0]**o+(-0.926058887186)*x[0]+(0.509510780366)*x[1]**o+(0.738049597453)*x[1]+(0.191676625722)*x[2]**o+(-0.354148482258)*x[2]
        arg[1,2,2,0]=(-0.955204517346)*x[0]**o+(0.234220317948)*x[0]+(0.44435792198)*x[1]**o+(0.96099571541)*x[1]+(-0.422462745872)*x[2]**o+(0.808817733192)*x[2]
        arg[1,2,2,1]=(0.932454071169)*x[0]**o+(-0.449841123911)*x[0]+(-0.371587228288)*x[1]**o+(-0.552185751938)*x[1]+(-0.222070148528)*x[2]**o+(-0.911768306425)*x[2]
        arg[1,3,0,0]=(-0.355162862133)*x[0]**o+(-0.00562439573347)*x[0]+(0.982216168139)*x[1]**o+(0.0493986220989)*x[1]+(0.0242698649494)*x[2]**o+(0.593411693569)*x[2]
        arg[1,3,0,1]=(-0.982164819772)*x[0]**o+(-0.417475530445)*x[0]+(-0.421752719264)*x[1]**o+(-0.102522772133)*x[1]+(-0.384947583437)*x[2]**o+(-0.675090951882)*x[2]
        arg[1,3,1,0]=(0.862354670281)*x[0]**o+(0.0332806244503)*x[0]+(0.248152892705)*x[1]**o+(-0.380941068103)*x[1]+(0.499668747376)*x[2]**o+(-0.732344330916)*x[2]
        arg[1,3,1,1]=(-0.367669055003)*x[0]**o+(-0.28761578782)*x[0]+(0.10838982068)*x[1]**o+(-0.495999837684)*x[1]+(0.599492939122)*x[2]**o+(0.886609670648)*x[2]
        arg[1,3,2,0]=(0.707026122148)*x[0]**o+(-0.625314188271)*x[0]+(-0.734283031657)*x[1]**o+(0.779848908302)*x[1]+(-0.965798269237)*x[2]**o+(0.35160103487)*x[2]
        arg[1,3,2,1]=(0.198008431503)*x[0]**o+(0.145470568903)*x[0]+(-0.342584113223)*x[1]**o+(-0.976390968602)*x[1]+(-0.371678081685)*x[2]**o+(0.569699640624)*x[2]
        arg[1,4,0,0]=(-0.816642540092)*x[0]**o+(0.466127552066)*x[0]+(-0.491109738021)*x[1]**o+(-0.177222784312)*x[1]+(0.723624314914)*x[2]**o+(-0.103929106752)*x[2]
        arg[1,4,0,1]=(-0.461770336596)*x[0]**o+(-0.416473447722)*x[0]+(-0.844639456093)*x[1]**o+(-0.021509624606)*x[1]+(-0.507857805076)*x[2]**o+(-0.0132005113079)*x[2]
        arg[1,4,1,0]=(0.279614653607)*x[0]**o+(-0.547332204747)*x[0]+(0.329751635637)*x[1]**o+(0.813149533168)*x[1]+(0.0361758823053)*x[2]**o+(0.0635753923527)*x[2]
        arg[1,4,1,1]=(0.124940246141)*x[0]**o+(-0.470715329674)*x[0]+(0.401922174627)*x[1]**o+(-0.292371975658)*x[1]+(0.302438467505)*x[2]**o+(0.131045792461)*x[2]
        arg[1,4,2,0]=(0.817170444398)*x[0]**o+(0.153182033447)*x[0]+(-0.98908973731)*x[1]**o+(0.359609127275)*x[1]+(-0.891501663456)*x[2]**o+(0.34700998982)*x[2]
        arg[1,4,2,1]=(0.619050545861)*x[0]**o+(-0.784466574079)*x[0]+(-0.15316511355)*x[1]**o+(0.587255937366)*x[1]+(0.740644692907)*x[2]**o+(0.0651236072508)*x[2]
        arg[2,0,0,0]=(0.67904774175)*x[0]**o+(-0.24998331377)*x[0]+(-0.263635906274)*x[1]**o+(0.157008949549)*x[1]+(-0.905158896842)*x[2]**o+(-0.975062080366)*x[2]
        arg[2,0,0,1]=(0.724305963876)*x[0]**o+(0.666891161317)*x[0]+(0.769784501206)*x[1]**o+(-0.84353242369)*x[1]+(0.631055556571)*x[2]**o+(-0.433528873211)*x[2]
        arg[2,0,1,0]=(-0.545702871321)*x[0]**o+(-0.153233364062)*x[0]+(-0.892642075879)*x[1]**o+(-0.303965445232)*x[1]+(0.675938211305)*x[2]**o+(0.134016007164)*x[2]
        arg[2,0,1,1]=(0.612528479764)*x[0]**o+(0.444039154606)*x[0]+(-0.185281638413)*x[1]**o+(0.643315773689)*x[1]+(-0.5562387833)*x[2]**o+(-0.960249468718)*x[2]
        arg[2,0,2,0]=(-0.330135152971)*x[0]**o+(0.359951176686)*x[0]+(-0.797088515842)*x[1]**o+(-0.257909258598)*x[1]+(-0.927156080256)*x[2]**o+(-0.843405252125)*x[2]
        arg[2,0,2,1]=(0.447041226497)*x[0]**o+(0.691225329527)*x[0]+(0.462798475523)*x[1]**o+(0.157497098383)*x[1]+(-0.77785224695)*x[2]**o+(0.886430058902)*x[2]
        arg[2,1,0,0]=(0.397605326971)*x[0]**o+(0.66227843262)*x[0]+(-0.634922556623)*x[1]**o+(-0.263342020187)*x[1]+(-0.81486179124)*x[2]**o+(-0.216716259698)*x[2]
        arg[2,1,0,1]=(-0.868246547394)*x[0]**o+(-0.194983059368)*x[0]+(-0.667277750808)*x[1]**o+(0.284011270958)*x[1]+(-0.276295620425)*x[2]**o+(0.215159516398)*x[2]
        arg[2,1,1,0]=(0.990906320323)*x[0]**o+(0.526754062323)*x[0]+(-0.963398315893)*x[1]**o+(-0.72561275544)*x[1]+(0.539779065056)*x[2]**o+(0.37738462043)*x[2]
        arg[2,1,1,1]=(-0.668520028258)*x[0]**o+(-0.292883094355)*x[0]+(0.260949538172)*x[1]**o+(-0.952706524522)*x[1]+(-0.715386804875)*x[2]**o+(-0.683320136224)*x[2]
        arg[2,1,2,0]=(-0.646186301281)*x[0]**o+(-0.62747987465)*x[0]+(-0.858714171368)*x[1]**o+(0.46910094021)*x[1]+(0.468975652807)*x[2]**o+(0.242115907129)*x[2]
        arg[2,1,2,1]=(0.262205767991)*x[0]**o+(-0.460376423099)*x[0]+(0.88572453638)*x[1]**o+(0.611761462581)*x[1]+(0.763860971141)*x[2]**o+(0.108470198525)*x[2]
        arg[2,2,0,0]=(-0.0218691308052)*x[0]**o+(0.942271392012)*x[0]+(-0.778469966054)*x[1]**o+(0.0873305333765)*x[1]+(-0.300479183111)*x[2]**o+(-0.862547713455)*x[2]
        arg[2,2,0,1]=(-0.736808990311)*x[0]**o+(-0.200478457878)*x[0]+(0.272608362607)*x[1]**o+(-0.0651056524306)*x[1]+(0.616628495568)*x[2]**o+(-0.8623490351)*x[2]
        arg[2,2,1,0]=(0.931530668297)*x[0]**o+(0.39701649536)*x[0]+(0.552135074711)*x[1]**o+(-0.489731326344)*x[1]+(-0.864724191354)*x[2]**o+(0.369557062583)*x[2]
        arg[2,2,1,1]=(0.418780453395)*x[0]**o+(-0.4935990418)*x[0]+(0.34625989303)*x[1]**o+(0.979829617988)*x[1]+(0.853169987441)*x[2]**o+(-0.945958722805)*x[2]
        arg[2,2,2,0]=(0.973689940887)*x[0]**o+(-0.0585784323994)*x[0]+(-0.472419744635)*x[1]**o+(0.851814193674)*x[1]+(-0.393393340405)*x[2]**o+(-0.755376551633)*x[2]
        arg[2,2,2,1]=(-0.217049913648)*x[0]**o+(-0.262150796255)*x[0]+(-0.407805273745)*x[1]**o+(0.819646964669)*x[1]+(-0.0130316627236)*x[2]**o+(0.179034130621)*x[2]
        arg[2,3,0,0]=(-0.377708814803)*x[0]**o+(-0.943893058411)*x[0]+(-0.354341838264)*x[1]**o+(0.0696491295689)*x[1]+(-0.192838410156)*x[2]**o+(-0.22208932914)*x[2]
        arg[2,3,0,1]=(0.252429489766)*x[0]**o+(0.808052283089)*x[0]+(0.282941493712)*x[1]**o+(0.89816007357)*x[1]+(0.000979423783521)*x[2]**o+(0.645599553602)*x[2]
        arg[2,3,1,0]=(0.788089987784)*x[0]**o+(0.414254044423)*x[0]+(0.105204158639)*x[1]**o+(0.97527041289)*x[1]+(0.447463569594)*x[2]**o+(0.91465488394)*x[2]
        arg[2,3,1,1]=(0.882323493958)*x[0]**o+(-0.98135461385)*x[0]+(0.641881834427)*x[1]**o+(-0.888068999894)*x[1]+(0.705728558352)*x[2]**o+(0.295815435834)*x[2]
        arg[2,3,2,0]=(-0.570666876652)*x[0]**o+(0.52308582534)*x[0]+(-0.604213776925)*x[1]**o+(-0.568684610916)*x[1]+(-0.0378463784885)*x[2]**o+(0.557994858276)*x[2]
        arg[2,3,2,1]=(-0.497489149519)*x[0]**o+(0.739341493998)*x[0]+(-0.375869065986)*x[1]**o+(-0.895553158302)*x[1]+(0.916829505813)*x[2]**o+(-0.650522069672)*x[2]
        arg[2,4,0,0]=(0.627540843427)*x[0]**o+(-0.572125230014)*x[0]+(-0.404781893475)*x[1]**o+(-0.114896790942)*x[1]+(-0.108914764977)*x[2]**o+(0.739758413931)*x[2]
        arg[2,4,0,1]=(0.243205825969)*x[0]**o+(-0.989602234642)*x[0]+(-0.925730664677)*x[1]**o+(0.454801040316)*x[1]+(0.51758036389)*x[2]**o+(0.325160660047)*x[2]
        arg[2,4,1,0]=(0.559764013478)*x[0]**o+(2.69034311577e-05)*x[0]+(-0.691119937581)*x[1]**o+(-0.347457112767)*x[1]+(0.714683489268)*x[2]**o+(-0.589270137254)*x[2]
        arg[2,4,1,1]=(-0.657837524953)*x[0]**o+(0.071158671662)*x[0]+(-0.998677178035)*x[1]**o+(-0.544136374784)*x[1]+(0.6200901771)*x[2]**o+(-0.280453395396)*x[2]
        arg[2,4,2,0]=(0.692403743003)*x[0]**o+(-0.570086669428)*x[0]+(0.455036498939)*x[1]**o+(0.291433179453)*x[1]+(-0.203004457401)*x[2]**o+(0.0192489593981)*x[2]
        arg[2,4,2,1]=(-0.656783033826)*x[0]**o+(0.865748949755)*x[0]+(-0.297731124456)*x[1]**o+(-0.826786841513)*x[1]+(0.891374127144)*x[2]**o+(-0.736789456498)*x[2]
        arg[3,0,0,0]=(-0.963120586854)*x[0]**o+(0.466089565408)*x[0]+(-0.785137906849)*x[1]**o+(0.194880437113)*x[1]+(0.652057839639)*x[2]**o+(-0.63674893761)*x[2]
        arg[3,0,0,1]=(-0.923234625819)*x[0]**o+(-0.303847080864)*x[0]+(-0.374245465515)*x[1]**o+(0.131093877726)*x[1]+(-0.13469395748)*x[2]**o+(0.42721676585)*x[2]
        arg[3,0,1,0]=(-0.613510153481)*x[0]**o+(0.986696685318)*x[0]+(0.188583200151)*x[1]**o+(-0.895175969251)*x[1]+(-0.699510846377)*x[2]**o+(0.953157679279)*x[2]
        arg[3,0,1,1]=(-0.805809890231)*x[0]**o+(-0.786467529592)*x[0]+(-0.287240865489)*x[1]**o+(-0.78714682888)*x[1]+(-0.301034554724)*x[2]**o+(0.698100092246)*x[2]
        arg[3,0,2,0]=(-0.29320618172)*x[0]**o+(0.482436772935)*x[0]+(-0.126052929261)*x[1]**o+(-0.0638862140753)*x[1]+(0.42769411972)*x[2]**o+(0.741118745829)*x[2]
        arg[3,0,2,1]=(-0.380041427783)*x[0]**o+(-0.0347904596008)*x[0]+(-0.321599184546)*x[1]**o+(-0.061333518544)*x[1]+(-0.885193271364)*x[2]**o+(0.216506075886)*x[2]
        arg[3,1,0,0]=(-0.872457576682)*x[0]**o+(0.1368749584)*x[0]+(0.883209741333)*x[1]**o+(0.946644762296)*x[1]+(-0.941665219846)*x[2]**o+(0.683787986795)*x[2]
        arg[3,1,0,1]=(0.564843030077)*x[0]**o+(0.987198898299)*x[0]+(0.808116156661)*x[1]**o+(-0.712172259479)*x[1]+(0.942066890171)*x[2]**o+(-0.265957158917)*x[2]
        arg[3,1,1,0]=(-0.171018264631)*x[0]**o+(0.277202848962)*x[0]+(0.0726306773528)*x[1]**o+(-0.0265640725409)*x[1]+(-0.871118922929)*x[2]**o+(0.0562314512261)*x[2]
        arg[3,1,1,1]=(0.0861984079085)*x[0]**o+(-0.954818837051)*x[0]+(0.817362109646)*x[1]**o+(-0.513086270222)*x[1]+(-0.487472170741)*x[2]**o+(0.702514192773)*x[2]
        arg[3,1,2,0]=(0.272419057202)*x[0]**o+(-0.0454292509397)*x[0]+(-0.117471126603)*x[1]**o+(0.638237510553)*x[1]+(0.156858948691)*x[2]**o+(0.499459702141)*x[2]
        arg[3,1,2,1]=(-0.384525883334)*x[0]**o+(-0.795638828759)*x[0]+(-0.105795059118)*x[1]**o+(0.539873566967)*x[1]+(-0.402558868714)*x[2]**o+(-0.690578475243)*x[2]
        arg[3,2,0,0]=(0.776889517136)*x[0]**o+(0.710931840891)*x[0]+(0.335992677388)*x[1]**o+(-0.689182863932)*x[1]+(0.294149805572)*x[2]**o+(-0.144223257292)*x[2]
        arg[3,2,0,1]=(0.738703695733)*x[0]**o+(-0.523207061995)*x[0]+(0.707078352596)*x[1]**o+(-0.594373978096)*x[1]+(-0.833468457639)*x[2]**o+(-0.60181674446)*x[2]
        arg[3,2,1,0]=(-0.70243682985)*x[0]**o+(-0.557914174271)*x[0]+(-0.97153723053)*x[1]**o+(0.791692594024)*x[1]+(-0.442273808305)*x[2]**o+(-0.74069605321)*x[2]
        arg[3,2,1,1]=(0.136307039789)*x[0]**o+(0.820661597623)*x[0]+(0.379752060994)*x[1]**o+(-0.739083116742)*x[1]+(-0.964857055764)*x[2]**o+(0.523159928671)*x[2]
        arg[3,2,2,0]=(0.530883295074)*x[0]**o+(-0.859538509888)*x[0]+(-0.33917158254)*x[1]**o+(0.476510741055)*x[1]+(0.851975683024)*x[2]**o+(0.146277015139)*x[2]
        arg[3,2,2,1]=(-0.532984574212)*x[0]**o+(-0.413169979754)*x[0]+(0.0113284200501)*x[1]**o+(0.406768750567)*x[1]+(0.730968378807)*x[2]**o+(-0.74004422909)*x[2]
        arg[3,3,0,0]=(0.784793271465)*x[0]**o+(0.879569037436)*x[0]+(0.687804194006)*x[1]**o+(0.153617608524)*x[1]+(-0.410247334526)*x[2]**o+(0.147941373268)*x[2]
        arg[3,3,0,1]=(-0.896748874838)*x[0]**o+(-0.408404228864)*x[0]+(0.831967697989)*x[1]**o+(-0.290875629476)*x[1]+(-0.244101719785)*x[2]**o+(0.240453425519)*x[2]
        arg[3,3,1,0]=(-0.352293918285)*x[0]**o+(0.0341233590345)*x[0]+(-0.0165872699681)*x[1]**o+(-0.0668507705942)*x[1]+(-0.133733710061)*x[2]**o+(-0.00620237772745)*x[2]
        arg[3,3,1,1]=(0.173638606531)*x[0]**o+(0.619705705811)*x[0]+(0.279972531576)*x[1]**o+(-0.0811181576037)*x[1]+(-0.645673968637)*x[2]**o+(0.913753411758)*x[2]
        arg[3,3,2,0]=(0.0623921286142)*x[0]**o+(0.393414889204)*x[0]+(-0.300034851075)*x[1]**o+(-0.519515206457)*x[1]+(-0.0074981731879)*x[2]**o+(0.202996242895)*x[2]
        arg[3,3,2,1]=(0.583279597505)*x[0]**o+(-0.940603487209)*x[0]+(0.027966157338)*x[1]**o+(0.251466344692)*x[1]+(-0.97011033154)*x[2]**o+(0.46527056203)*x[2]
        arg[3,4,0,0]=(0.865170740218)*x[0]**o+(-0.466853211289)*x[0]+(0.0935492537278)*x[1]**o+(-0.632374632657)*x[1]+(0.673466346463)*x[2]**o+(-0.47934739873)*x[2]
        arg[3,4,0,1]=(0.0417451136027)*x[0]**o+(-0.781570733286)*x[0]+(0.547452389454)*x[1]**o+(-0.0286368211528)*x[1]+(-0.98907358808)*x[2]**o+(0.62749495057)*x[2]
        arg[3,4,1,0]=(-0.632067479664)*x[0]**o+(-0.872835386336)*x[0]+(-0.228785631078)*x[1]**o+(-0.224206127306)*x[1]+(0.769729051217)*x[2]**o+(-0.406750034352)*x[2]
        arg[3,4,1,1]=(0.216883071299)*x[0]**o+(-0.225632200121)*x[0]+(0.827186402743)*x[1]**o+(0.161488636886)*x[1]+(0.773801288836)*x[2]**o+(0.977006832166)*x[2]
        arg[3,4,2,0]=(-0.867098845884)*x[0]**o+(0.14989544838)*x[0]+(0.756720043262)*x[1]**o+(-0.496074302626)*x[1]+(-0.662787072451)*x[2]**o+(-0.142473973368)*x[2]
        arg[3,4,2,1]=(-0.294128396829)*x[0]**o+(-0.931379392091)*x[0]+(0.906552131161)*x[1]**o+(-0.422008485494)*x[1]+(0.806742455679)*x[2]**o+(0.940488803006)*x[2]
        ref[0,0,0,0]=(-0.0998996137438)*x_ref[0]**o+(-0.648495171588)*x_ref[0]+(-0.836177837758)*x_ref[1]**o+(-0.9597381047)*x_ref[1]+(-0.958069967907)*x_ref[2]**o+(0.399097251753)*x_ref[2]
        ref[0,0,0,1]=(0.625409406207)*x_ref[0]**o+(-0.522476054218)*x_ref[0]+(0.794724584797)*x_ref[1]**o+(-0.450875661781)*x_ref[1]+(-0.189505729613)*x_ref[2]**o+(0.0910808218379)*x_ref[2]
        ref[0,0,1,0]=(-0.36073198549)*x_ref[0]**o+(-0.0286144078107)*x_ref[0]+(-0.53827851785)*x_ref[1]**o+(-0.563600303848)*x_ref[1]+(0.970211345168)*x_ref[2]**o+(-0.405253675856)*x_ref[2]
        ref[0,0,1,1]=(-0.0204302558323)*x_ref[0]**o+(0.927570053009)*x_ref[0]+(-0.358571532503)*x_ref[1]**o+(0.282513525404)*x_ref[1]+(0.105530612734)*x_ref[2]**o+(-0.156753729378)*x_ref[2]
        ref[0,0,2,0]=(0.25049329858)*x_ref[0]**o+(0.412299654624)*x_ref[0]+(0.827683104949)*x_ref[1]**o+(0.0195885473957)*x_ref[1]+(0.849521623993)*x_ref[2]**o+(-0.107922097347)*x_ref[2]
        ref[0,0,2,1]=(0.824054630326)*x_ref[0]**o+(0.315612619612)*x_ref[0]+(-0.33528975625)*x_ref[1]**o+(-0.848543364468)*x_ref[1]+(0.67462279419)*x_ref[2]**o+(-0.0387606720463)*x_ref[2]
        ref[0,1,0,0]=(0.251613993994)*x_ref[0]**o+(-0.871217704935)*x_ref[0]+(-0.648182240752)*x_ref[1]**o+(0.324536900202)*x_ref[1]+(-0.884871901147)*x_ref[2]**o+(0.353160185772)*x_ref[2]
        ref[0,1,0,1]=(-0.629894976165)*x_ref[0]**o+(0.99092287348)*x_ref[0]+(0.87471936536)*x_ref[1]**o+(0.732884844801)*x_ref[1]+(0.423531105096)*x_ref[2]**o+(0.648959408663)*x_ref[2]
        ref[0,1,1,0]=(0.123366820062)*x_ref[0]**o+(-0.856900246151)*x_ref[0]+(-0.887819649211)*x_ref[1]**o+(-0.901367472665)*x_ref[1]+(0.222535500196)*x_ref[2]**o+(0.42830671462)*x_ref[2]
        ref[0,1,1,1]=(0.343789367141)*x_ref[0]**o+(0.614284333526)*x_ref[0]+(0.976225499577)*x_ref[1]**o+(0.955115562801)*x_ref[1]+(0.0700409126681)*x_ref[2]**o+(-0.709602275548)*x_ref[2]
        ref[0,1,2,0]=(-0.863502900367)*x_ref[0]**o+(-0.452731447667)*x_ref[0]+(0.495519161334)*x_ref[1]**o+(0.0254261045314)*x_ref[1]+(-0.484479110985)*x_ref[2]**o+(0.615421138564)*x_ref[2]
        ref[0,1,2,1]=(-0.861966549825)*x_ref[0]**o+(0.957385177208)*x_ref[0]+(0.710449966969)*x_ref[1]**o+(-0.367149885665)*x_ref[1]+(-0.100800220614)*x_ref[2]**o+(-0.266781849874)*x_ref[2]
        ref[0,2,0,0]=(-0.781309947495)*x_ref[0]**o+(-0.709036446117)*x_ref[0]+(0.86136976691)*x_ref[1]**o+(-0.745743057084)*x_ref[1]+(0.519795761984)*x_ref[2]**o+(0.298838092751)*x_ref[2]
        ref[0,2,0,1]=(-0.514037154361)*x_ref[0]**o+(-0.155867632386)*x_ref[0]+(-0.125008826668)*x_ref[1]**o+(-0.138197902584)*x_ref[1]+(0.360698647284)*x_ref[2]**o+(0.432385729464)*x_ref[2]
        ref[0,2,1,0]=(-0.19580514344)*x_ref[0]**o+(-0.430510395896)*x_ref[0]+(0.676498502854)*x_ref[1]**o+(-0.472411864377)*x_ref[1]+(-0.323468590594)*x_ref[2]**o+(0.837971311859)*x_ref[2]
        ref[0,2,1,1]=(-0.358714577548)*x_ref[0]**o+(-0.583080123482)*x_ref[0]+(-0.225887451598)*x_ref[1]**o+(0.744101609163)*x_ref[1]+(-0.0300539574094)*x_ref[2]**o+(-0.985181847505)*x_ref[2]
        ref[0,2,2,0]=(0.820610118447)*x_ref[0]**o+(-0.560334915671)*x_ref[0]+(0.478672329569)*x_ref[1]**o+(-0.956892480643)*x_ref[1]+(0.302100338619)*x_ref[2]**o+(0.597584594886)*x_ref[2]
        ref[0,2,2,1]=(-0.0658925658941)*x_ref[0]**o+(-0.136738027019)*x_ref[0]+(0.617259524472)*x_ref[1]**o+(-0.486181324584)*x_ref[1]+(0.225924222893)*x_ref[2]**o+(-0.304893108343)*x_ref[2]
        ref[0,3,0,0]=(0.0824288958004)*x_ref[0]**o+(-0.67546530428)*x_ref[0]+(-0.726843419073)*x_ref[1]**o+(0.150753236521)*x_ref[1]+(0.200621841908)*x_ref[2]**o+(0.869623957921)*x_ref[2]
        ref[0,3,0,1]=(0.66227806631)*x_ref[0]**o+(-0.389229756136)*x_ref[0]+(0.739548522805)*x_ref[1]**o+(0.131026218939)*x_ref[1]+(0.743271925877)*x_ref[2]**o+(-0.910220233981)*x_ref[2]
        ref[0,3,1,0]=(-0.259647259175)*x_ref[0]**o+(-0.208585364012)*x_ref[0]+(-0.323867955964)*x_ref[1]**o+(0.778380345151)*x_ref[1]+(-0.555780852614)*x_ref[2]**o+(-0.16366483426)*x_ref[2]
        ref[0,3,1,1]=(0.248968056109)*x_ref[0]**o+(0.84999990087)*x_ref[0]+(0.180937025847)*x_ref[1]**o+(0.332075828156)*x_ref[1]+(-0.00577042146131)*x_ref[2]**o+(0.708188905312)*x_ref[2]
        ref[0,3,2,0]=(-0.80327596428)*x_ref[0]**o+(-0.131004563471)*x_ref[0]+(0.443094741792)*x_ref[1]**o+(0.515867966266)*x_ref[1]+(-0.964003373109)*x_ref[2]**o+(-0.580721642935)*x_ref[2]
        ref[0,3,2,1]=(-0.107361389717)*x_ref[0]**o+(-0.543617395264)*x_ref[0]+(0.840867379108)*x_ref[1]**o+(-0.459941493027)*x_ref[1]+(-0.300033656162)*x_ref[2]**o+(0.440763554498)*x_ref[2]
        ref[0,4,0,0]=(0.272916362055)*x_ref[0]**o+(-0.554210888243)*x_ref[0]+(-0.458102512273)*x_ref[1]**o+(-0.54594406835)*x_ref[1]+(0.570670715852)*x_ref[2]**o+(0.792731049718)*x_ref[2]
        ref[0,4,0,1]=(0.896330380602)*x_ref[0]**o+(-0.929757950797)*x_ref[0]+(0.229049814758)*x_ref[1]**o+(0.1245974562)*x_ref[1]+(0.853250033203)*x_ref[2]**o+(-0.877052530974)*x_ref[2]
        ref[0,4,1,0]=(0.76027467594)*x_ref[0]**o+(-0.510095378376)*x_ref[0]+(0.697258966361)*x_ref[1]**o+(-0.44013714543)*x_ref[1]+(0.617675705023)*x_ref[2]**o+(-0.00118327085347)*x_ref[2]
        ref[0,4,1,1]=(0.92472377319)*x_ref[0]**o+(0.27783405618)*x_ref[0]+(-0.24317632451)*x_ref[1]**o+(-0.0264727232824)*x_ref[1]+(0.428751701408)*x_ref[2]**o+(0.184142273303)*x_ref[2]
        ref[0,4,2,0]=(0.619110983031)*x_ref[0]**o+(-0.375037512361)*x_ref[0]+(0.298576843317)*x_ref[1]**o+(-0.218802335776)*x_ref[1]+(-0.394889850336)*x_ref[2]**o+(-0.842359750924)*x_ref[2]
        ref[0,4,2,1]=(0.899714346356)*x_ref[0]**o+(0.0630133184109)*x_ref[0]+(0.0796796488932)*x_ref[1]**o+(-0.693371400327)*x_ref[1]+(-0.412090402419)*x_ref[2]**o+(0.681871832039)*x_ref[2]
        ref[1,0,0,0]=(-0.0367677770358)*x_ref[0]**o+(-0.75352301678)*x_ref[0]+(0.0705016755404)*x_ref[1]**o+(0.870115630516)*x_ref[1]+(0.136757694474)*x_ref[2]**o+(-0.557722157484)*x_ref[2]
        ref[1,0,0,1]=(0.550371937413)*x_ref[0]**o+(-0.590052751632)*x_ref[0]+(0.38665744963)*x_ref[1]**o+(-0.896537565188)*x_ref[1]+(0.550869514185)*x_ref[2]**o+(0.693828237691)*x_ref[2]
        ref[1,0,1,0]=(-0.150321579562)*x_ref[0]**o+(0.111135028893)*x_ref[0]+(0.582595079816)*x_ref[1]**o+(-0.250920804447)*x_ref[1]+(0.392998664266)*x_ref[2]**o+(-0.226735735878)*x_ref[2]
        ref[1,0,1,1]=(0.435196573865)*x_ref[0]**o+(-0.31161203165)*x_ref[0]+(-0.117580582163)*x_ref[1]**o+(0.99583282896)*x_ref[1]+(-0.164872214322)*x_ref[2]**o+(0.243837126961)*x_ref[2]
        ref[1,0,2,0]=(-0.320135099791)*x_ref[0]**o+(0.339871290643)*x_ref[0]+(0.593784119675)*x_ref[1]**o+(0.695166367768)*x_ref[1]+(0.407551427099)*x_ref[2]**o+(-0.768445418933)*x_ref[2]
        ref[1,0,2,1]=(-0.390718323619)*x_ref[0]**o+(-0.0515766643887)*x_ref[0]+(0.394112545837)*x_ref[1]**o+(-0.379592335881)*x_ref[1]+(-0.29680237927)*x_ref[2]**o+(0.500583589142)*x_ref[2]
        ref[1,1,0,0]=(-0.958930666062)*x_ref[0]**o+(0.516874100468)*x_ref[0]+(0.0859034666349)*x_ref[1]**o+(-0.370561856015)*x_ref[1]+(-0.175293417132)*x_ref[2]**o+(0.672554615126)*x_ref[2]
        ref[1,1,0,1]=(-0.38129381629)*x_ref[0]**o+(-0.615636530444)*x_ref[0]+(0.447284140424)*x_ref[1]**o+(0.507583882952)*x_ref[1]+(-0.942413761354)*x_ref[2]**o+(-0.607928807162)*x_ref[2]
        ref[1,1,1,0]=(-0.456462200294)*x_ref[0]**o+(0.778557108965)*x_ref[0]+(-0.939624730616)*x_ref[1]**o+(0.965857713341)*x_ref[1]+(-0.913772635427)*x_ref[2]**o+(-0.459963514384)*x_ref[2]
        ref[1,1,1,1]=(0.138896120228)*x_ref[0]**o+(0.941491395046)*x_ref[0]+(0.546538931051)*x_ref[1]**o+(-0.526098363992)*x_ref[1]+(0.32139556996)*x_ref[2]**o+(0.899346403792)*x_ref[2]
        ref[1,1,2,0]=(0.591597103411)*x_ref[0]**o+(0.926278363416)*x_ref[0]+(0.565926211046)*x_ref[1]**o+(0.930516925784)*x_ref[1]+(-0.498131779353)*x_ref[2]**o+(0.358157310681)*x_ref[2]
        ref[1,1,2,1]=(0.758395199797)*x_ref[0]**o+(0.769578766255)*x_ref[0]+(0.86819976051)*x_ref[1]**o+(-0.380436310842)*x_ref[1]+(-0.513012037048)*x_ref[2]**o+(-0.984361866072)*x_ref[2]
        ref[1,2,0,0]=(0.0845846076213)*x_ref[0]**o+(0.987145343601)*x_ref[0]+(-0.447680898266)*x_ref[1]**o+(-0.30278197108)*x_ref[1]+(0.271667503636)*x_ref[2]**o+(0.698967822656)*x_ref[2]
        ref[1,2,0,1]=(-0.314076111079)*x_ref[0]**o+(0.507407943673)*x_ref[0]+(-0.38467148586)*x_ref[1]**o+(0.802764957621)*x_ref[1]+(0.692949890402)*x_ref[2]**o+(-0.965766183849)*x_ref[2]
        ref[1,2,1,0]=(-0.855806343042)*x_ref[0]**o+(0.753362713203)*x_ref[0]+(-0.9877924012)*x_ref[1]**o+(-0.54047626663)*x_ref[1]+(-0.488859780266)*x_ref[2]**o+(0.444279382381)*x_ref[2]
        ref[1,2,1,1]=(-0.313257382113)*x_ref[0]**o+(-0.926058887186)*x_ref[0]+(0.509510780366)*x_ref[1]**o+(0.738049597453)*x_ref[1]+(0.191676625722)*x_ref[2]**o+(-0.354148482258)*x_ref[2]
        ref[1,2,2,0]=(-0.955204517346)*x_ref[0]**o+(0.234220317948)*x_ref[0]+(0.44435792198)*x_ref[1]**o+(0.96099571541)*x_ref[1]+(-0.422462745872)*x_ref[2]**o+(0.808817733192)*x_ref[2]
        ref[1,2,2,1]=(0.932454071169)*x_ref[0]**o+(-0.449841123911)*x_ref[0]+(-0.371587228288)*x_ref[1]**o+(-0.552185751938)*x_ref[1]+(-0.222070148528)*x_ref[2]**o+(-0.911768306425)*x_ref[2]
        ref[1,3,0,0]=(-0.355162862133)*x_ref[0]**o+(-0.00562439573347)*x_ref[0]+(0.982216168139)*x_ref[1]**o+(0.0493986220989)*x_ref[1]+(0.0242698649494)*x_ref[2]**o+(0.593411693569)*x_ref[2]
        ref[1,3,0,1]=(-0.982164819772)*x_ref[0]**o+(-0.417475530445)*x_ref[0]+(-0.421752719264)*x_ref[1]**o+(-0.102522772133)*x_ref[1]+(-0.384947583437)*x_ref[2]**o+(-0.675090951882)*x_ref[2]
        ref[1,3,1,0]=(0.862354670281)*x_ref[0]**o+(0.0332806244503)*x_ref[0]+(0.248152892705)*x_ref[1]**o+(-0.380941068103)*x_ref[1]+(0.499668747376)*x_ref[2]**o+(-0.732344330916)*x_ref[2]
        ref[1,3,1,1]=(-0.367669055003)*x_ref[0]**o+(-0.28761578782)*x_ref[0]+(0.10838982068)*x_ref[1]**o+(-0.495999837684)*x_ref[1]+(0.599492939122)*x_ref[2]**o+(0.886609670648)*x_ref[2]
        ref[1,3,2,0]=(0.707026122148)*x_ref[0]**o+(-0.625314188271)*x_ref[0]+(-0.734283031657)*x_ref[1]**o+(0.779848908302)*x_ref[1]+(-0.965798269237)*x_ref[2]**o+(0.35160103487)*x_ref[2]
        ref[1,3,2,1]=(0.198008431503)*x_ref[0]**o+(0.145470568903)*x_ref[0]+(-0.342584113223)*x_ref[1]**o+(-0.976390968602)*x_ref[1]+(-0.371678081685)*x_ref[2]**o+(0.569699640624)*x_ref[2]
        ref[1,4,0,0]=(-0.816642540092)*x_ref[0]**o+(0.466127552066)*x_ref[0]+(-0.491109738021)*x_ref[1]**o+(-0.177222784312)*x_ref[1]+(0.723624314914)*x_ref[2]**o+(-0.103929106752)*x_ref[2]
        ref[1,4,0,1]=(-0.461770336596)*x_ref[0]**o+(-0.416473447722)*x_ref[0]+(-0.844639456093)*x_ref[1]**o+(-0.021509624606)*x_ref[1]+(-0.507857805076)*x_ref[2]**o+(-0.0132005113079)*x_ref[2]
        ref[1,4,1,0]=(0.279614653607)*x_ref[0]**o+(-0.547332204747)*x_ref[0]+(0.329751635637)*x_ref[1]**o+(0.813149533168)*x_ref[1]+(0.0361758823053)*x_ref[2]**o+(0.0635753923527)*x_ref[2]
        ref[1,4,1,1]=(0.124940246141)*x_ref[0]**o+(-0.470715329674)*x_ref[0]+(0.401922174627)*x_ref[1]**o+(-0.292371975658)*x_ref[1]+(0.302438467505)*x_ref[2]**o+(0.131045792461)*x_ref[2]
        ref[1,4,2,0]=(0.817170444398)*x_ref[0]**o+(0.153182033447)*x_ref[0]+(-0.98908973731)*x_ref[1]**o+(0.359609127275)*x_ref[1]+(-0.891501663456)*x_ref[2]**o+(0.34700998982)*x_ref[2]
        ref[1,4,2,1]=(0.619050545861)*x_ref[0]**o+(-0.784466574079)*x_ref[0]+(-0.15316511355)*x_ref[1]**o+(0.587255937366)*x_ref[1]+(0.740644692907)*x_ref[2]**o+(0.0651236072508)*x_ref[2]
        ref[2,0,0,0]=(0.67904774175)*x_ref[0]**o+(-0.24998331377)*x_ref[0]+(-0.263635906274)*x_ref[1]**o+(0.157008949549)*x_ref[1]+(-0.905158896842)*x_ref[2]**o+(-0.975062080366)*x_ref[2]
        ref[2,0,0,1]=(0.724305963876)*x_ref[0]**o+(0.666891161317)*x_ref[0]+(0.769784501206)*x_ref[1]**o+(-0.84353242369)*x_ref[1]+(0.631055556571)*x_ref[2]**o+(-0.433528873211)*x_ref[2]
        ref[2,0,1,0]=(-0.545702871321)*x_ref[0]**o+(-0.153233364062)*x_ref[0]+(-0.892642075879)*x_ref[1]**o+(-0.303965445232)*x_ref[1]+(0.675938211305)*x_ref[2]**o+(0.134016007164)*x_ref[2]
        ref[2,0,1,1]=(0.612528479764)*x_ref[0]**o+(0.444039154606)*x_ref[0]+(-0.185281638413)*x_ref[1]**o+(0.643315773689)*x_ref[1]+(-0.5562387833)*x_ref[2]**o+(-0.960249468718)*x_ref[2]
        ref[2,0,2,0]=(-0.330135152971)*x_ref[0]**o+(0.359951176686)*x_ref[0]+(-0.797088515842)*x_ref[1]**o+(-0.257909258598)*x_ref[1]+(-0.927156080256)*x_ref[2]**o+(-0.843405252125)*x_ref[2]
        ref[2,0,2,1]=(0.447041226497)*x_ref[0]**o+(0.691225329527)*x_ref[0]+(0.462798475523)*x_ref[1]**o+(0.157497098383)*x_ref[1]+(-0.77785224695)*x_ref[2]**o+(0.886430058902)*x_ref[2]
        ref[2,1,0,0]=(0.397605326971)*x_ref[0]**o+(0.66227843262)*x_ref[0]+(-0.634922556623)*x_ref[1]**o+(-0.263342020187)*x_ref[1]+(-0.81486179124)*x_ref[2]**o+(-0.216716259698)*x_ref[2]
        ref[2,1,0,1]=(-0.868246547394)*x_ref[0]**o+(-0.194983059368)*x_ref[0]+(-0.667277750808)*x_ref[1]**o+(0.284011270958)*x_ref[1]+(-0.276295620425)*x_ref[2]**o+(0.215159516398)*x_ref[2]
        ref[2,1,1,0]=(0.990906320323)*x_ref[0]**o+(0.526754062323)*x_ref[0]+(-0.963398315893)*x_ref[1]**o+(-0.72561275544)*x_ref[1]+(0.539779065056)*x_ref[2]**o+(0.37738462043)*x_ref[2]
        ref[2,1,1,1]=(-0.668520028258)*x_ref[0]**o+(-0.292883094355)*x_ref[0]+(0.260949538172)*x_ref[1]**o+(-0.952706524522)*x_ref[1]+(-0.715386804875)*x_ref[2]**o+(-0.683320136224)*x_ref[2]
        ref[2,1,2,0]=(-0.646186301281)*x_ref[0]**o+(-0.62747987465)*x_ref[0]+(-0.858714171368)*x_ref[1]**o+(0.46910094021)*x_ref[1]+(0.468975652807)*x_ref[2]**o+(0.242115907129)*x_ref[2]
        ref[2,1,2,1]=(0.262205767991)*x_ref[0]**o+(-0.460376423099)*x_ref[0]+(0.88572453638)*x_ref[1]**o+(0.611761462581)*x_ref[1]+(0.763860971141)*x_ref[2]**o+(0.108470198525)*x_ref[2]
        ref[2,2,0,0]=(-0.0218691308052)*x_ref[0]**o+(0.942271392012)*x_ref[0]+(-0.778469966054)*x_ref[1]**o+(0.0873305333765)*x_ref[1]+(-0.300479183111)*x_ref[2]**o+(-0.862547713455)*x_ref[2]
        ref[2,2,0,1]=(-0.736808990311)*x_ref[0]**o+(-0.200478457878)*x_ref[0]+(0.272608362607)*x_ref[1]**o+(-0.0651056524306)*x_ref[1]+(0.616628495568)*x_ref[2]**o+(-0.8623490351)*x_ref[2]
        ref[2,2,1,0]=(0.931530668297)*x_ref[0]**o+(0.39701649536)*x_ref[0]+(0.552135074711)*x_ref[1]**o+(-0.489731326344)*x_ref[1]+(-0.864724191354)*x_ref[2]**o+(0.369557062583)*x_ref[2]
        ref[2,2,1,1]=(0.418780453395)*x_ref[0]**o+(-0.4935990418)*x_ref[0]+(0.34625989303)*x_ref[1]**o+(0.979829617988)*x_ref[1]+(0.853169987441)*x_ref[2]**o+(-0.945958722805)*x_ref[2]
        ref[2,2,2,0]=(0.973689940887)*x_ref[0]**o+(-0.0585784323994)*x_ref[0]+(-0.472419744635)*x_ref[1]**o+(0.851814193674)*x_ref[1]+(-0.393393340405)*x_ref[2]**o+(-0.755376551633)*x_ref[2]
        ref[2,2,2,1]=(-0.217049913648)*x_ref[0]**o+(-0.262150796255)*x_ref[0]+(-0.407805273745)*x_ref[1]**o+(0.819646964669)*x_ref[1]+(-0.0130316627236)*x_ref[2]**o+(0.179034130621)*x_ref[2]
        ref[2,3,0,0]=(-0.377708814803)*x_ref[0]**o+(-0.943893058411)*x_ref[0]+(-0.354341838264)*x_ref[1]**o+(0.0696491295689)*x_ref[1]+(-0.192838410156)*x_ref[2]**o+(-0.22208932914)*x_ref[2]
        ref[2,3,0,1]=(0.252429489766)*x_ref[0]**o+(0.808052283089)*x_ref[0]+(0.282941493712)*x_ref[1]**o+(0.89816007357)*x_ref[1]+(0.000979423783521)*x_ref[2]**o+(0.645599553602)*x_ref[2]
        ref[2,3,1,0]=(0.788089987784)*x_ref[0]**o+(0.414254044423)*x_ref[0]+(0.105204158639)*x_ref[1]**o+(0.97527041289)*x_ref[1]+(0.447463569594)*x_ref[2]**o+(0.91465488394)*x_ref[2]
        ref[2,3,1,1]=(0.882323493958)*x_ref[0]**o+(-0.98135461385)*x_ref[0]+(0.641881834427)*x_ref[1]**o+(-0.888068999894)*x_ref[1]+(0.705728558352)*x_ref[2]**o+(0.295815435834)*x_ref[2]
        ref[2,3,2,0]=(-0.570666876652)*x_ref[0]**o+(0.52308582534)*x_ref[0]+(-0.604213776925)*x_ref[1]**o+(-0.568684610916)*x_ref[1]+(-0.0378463784885)*x_ref[2]**o+(0.557994858276)*x_ref[2]
        ref[2,3,2,1]=(-0.497489149519)*x_ref[0]**o+(0.739341493998)*x_ref[0]+(-0.375869065986)*x_ref[1]**o+(-0.895553158302)*x_ref[1]+(0.916829505813)*x_ref[2]**o+(-0.650522069672)*x_ref[2]
        ref[2,4,0,0]=(0.627540843427)*x_ref[0]**o+(-0.572125230014)*x_ref[0]+(-0.404781893475)*x_ref[1]**o+(-0.114896790942)*x_ref[1]+(-0.108914764977)*x_ref[2]**o+(0.739758413931)*x_ref[2]
        ref[2,4,0,1]=(0.243205825969)*x_ref[0]**o+(-0.989602234642)*x_ref[0]+(-0.925730664677)*x_ref[1]**o+(0.454801040316)*x_ref[1]+(0.51758036389)*x_ref[2]**o+(0.325160660047)*x_ref[2]
        ref[2,4,1,0]=(0.559764013478)*x_ref[0]**o+(2.69034311577e-05)*x_ref[0]+(-0.691119937581)*x_ref[1]**o+(-0.347457112767)*x_ref[1]+(0.714683489268)*x_ref[2]**o+(-0.589270137254)*x_ref[2]
        ref[2,4,1,1]=(-0.657837524953)*x_ref[0]**o+(0.071158671662)*x_ref[0]+(-0.998677178035)*x_ref[1]**o+(-0.544136374784)*x_ref[1]+(0.6200901771)*x_ref[2]**o+(-0.280453395396)*x_ref[2]
        ref[2,4,2,0]=(0.692403743003)*x_ref[0]**o+(-0.570086669428)*x_ref[0]+(0.455036498939)*x_ref[1]**o+(0.291433179453)*x_ref[1]+(-0.203004457401)*x_ref[2]**o+(0.0192489593981)*x_ref[2]
        ref[2,4,2,1]=(-0.656783033826)*x_ref[0]**o+(0.865748949755)*x_ref[0]+(-0.297731124456)*x_ref[1]**o+(-0.826786841513)*x_ref[1]+(0.891374127144)*x_ref[2]**o+(-0.736789456498)*x_ref[2]
        ref[3,0,0,0]=(-0.963120586854)*x_ref[0]**o+(0.466089565408)*x_ref[0]+(-0.785137906849)*x_ref[1]**o+(0.194880437113)*x_ref[1]+(0.652057839639)*x_ref[2]**o+(-0.63674893761)*x_ref[2]
        ref[3,0,0,1]=(-0.923234625819)*x_ref[0]**o+(-0.303847080864)*x_ref[0]+(-0.374245465515)*x_ref[1]**o+(0.131093877726)*x_ref[1]+(-0.13469395748)*x_ref[2]**o+(0.42721676585)*x_ref[2]
        ref[3,0,1,0]=(-0.613510153481)*x_ref[0]**o+(0.986696685318)*x_ref[0]+(0.188583200151)*x_ref[1]**o+(-0.895175969251)*x_ref[1]+(-0.699510846377)*x_ref[2]**o+(0.953157679279)*x_ref[2]
        ref[3,0,1,1]=(-0.805809890231)*x_ref[0]**o+(-0.786467529592)*x_ref[0]+(-0.287240865489)*x_ref[1]**o+(-0.78714682888)*x_ref[1]+(-0.301034554724)*x_ref[2]**o+(0.698100092246)*x_ref[2]
        ref[3,0,2,0]=(-0.29320618172)*x_ref[0]**o+(0.482436772935)*x_ref[0]+(-0.126052929261)*x_ref[1]**o+(-0.0638862140753)*x_ref[1]+(0.42769411972)*x_ref[2]**o+(0.741118745829)*x_ref[2]
        ref[3,0,2,1]=(-0.380041427783)*x_ref[0]**o+(-0.0347904596008)*x_ref[0]+(-0.321599184546)*x_ref[1]**o+(-0.061333518544)*x_ref[1]+(-0.885193271364)*x_ref[2]**o+(0.216506075886)*x_ref[2]
        ref[3,1,0,0]=(-0.872457576682)*x_ref[0]**o+(0.1368749584)*x_ref[0]+(0.883209741333)*x_ref[1]**o+(0.946644762296)*x_ref[1]+(-0.941665219846)*x_ref[2]**o+(0.683787986795)*x_ref[2]
        ref[3,1,0,1]=(0.564843030077)*x_ref[0]**o+(0.987198898299)*x_ref[0]+(0.808116156661)*x_ref[1]**o+(-0.712172259479)*x_ref[1]+(0.942066890171)*x_ref[2]**o+(-0.265957158917)*x_ref[2]
        ref[3,1,1,0]=(-0.171018264631)*x_ref[0]**o+(0.277202848962)*x_ref[0]+(0.0726306773528)*x_ref[1]**o+(-0.0265640725409)*x_ref[1]+(-0.871118922929)*x_ref[2]**o+(0.0562314512261)*x_ref[2]
        ref[3,1,1,1]=(0.0861984079085)*x_ref[0]**o+(-0.954818837051)*x_ref[0]+(0.817362109646)*x_ref[1]**o+(-0.513086270222)*x_ref[1]+(-0.487472170741)*x_ref[2]**o+(0.702514192773)*x_ref[2]
        ref[3,1,2,0]=(0.272419057202)*x_ref[0]**o+(-0.0454292509397)*x_ref[0]+(-0.117471126603)*x_ref[1]**o+(0.638237510553)*x_ref[1]+(0.156858948691)*x_ref[2]**o+(0.499459702141)*x_ref[2]
        ref[3,1,2,1]=(-0.384525883334)*x_ref[0]**o+(-0.795638828759)*x_ref[0]+(-0.105795059118)*x_ref[1]**o+(0.539873566967)*x_ref[1]+(-0.402558868714)*x_ref[2]**o+(-0.690578475243)*x_ref[2]
        ref[3,2,0,0]=(0.776889517136)*x_ref[0]**o+(0.710931840891)*x_ref[0]+(0.335992677388)*x_ref[1]**o+(-0.689182863932)*x_ref[1]+(0.294149805572)*x_ref[2]**o+(-0.144223257292)*x_ref[2]
        ref[3,2,0,1]=(0.738703695733)*x_ref[0]**o+(-0.523207061995)*x_ref[0]+(0.707078352596)*x_ref[1]**o+(-0.594373978096)*x_ref[1]+(-0.833468457639)*x_ref[2]**o+(-0.60181674446)*x_ref[2]
        ref[3,2,1,0]=(-0.70243682985)*x_ref[0]**o+(-0.557914174271)*x_ref[0]+(-0.97153723053)*x_ref[1]**o+(0.791692594024)*x_ref[1]+(-0.442273808305)*x_ref[2]**o+(-0.74069605321)*x_ref[2]
        ref[3,2,1,1]=(0.136307039789)*x_ref[0]**o+(0.820661597623)*x_ref[0]+(0.379752060994)*x_ref[1]**o+(-0.739083116742)*x_ref[1]+(-0.964857055764)*x_ref[2]**o+(0.523159928671)*x_ref[2]
        ref[3,2,2,0]=(0.530883295074)*x_ref[0]**o+(-0.859538509888)*x_ref[0]+(-0.33917158254)*x_ref[1]**o+(0.476510741055)*x_ref[1]+(0.851975683024)*x_ref[2]**o+(0.146277015139)*x_ref[2]
        ref[3,2,2,1]=(-0.532984574212)*x_ref[0]**o+(-0.413169979754)*x_ref[0]+(0.0113284200501)*x_ref[1]**o+(0.406768750567)*x_ref[1]+(0.730968378807)*x_ref[2]**o+(-0.74004422909)*x_ref[2]
        ref[3,3,0,0]=(0.784793271465)*x_ref[0]**o+(0.879569037436)*x_ref[0]+(0.687804194006)*x_ref[1]**o+(0.153617608524)*x_ref[1]+(-0.410247334526)*x_ref[2]**o+(0.147941373268)*x_ref[2]
        ref[3,3,0,1]=(-0.896748874838)*x_ref[0]**o+(-0.408404228864)*x_ref[0]+(0.831967697989)*x_ref[1]**o+(-0.290875629476)*x_ref[1]+(-0.244101719785)*x_ref[2]**o+(0.240453425519)*x_ref[2]
        ref[3,3,1,0]=(-0.352293918285)*x_ref[0]**o+(0.0341233590345)*x_ref[0]+(-0.0165872699681)*x_ref[1]**o+(-0.0668507705942)*x_ref[1]+(-0.133733710061)*x_ref[2]**o+(-0.00620237772745)*x_ref[2]
        ref[3,3,1,1]=(0.173638606531)*x_ref[0]**o+(0.619705705811)*x_ref[0]+(0.279972531576)*x_ref[1]**o+(-0.0811181576037)*x_ref[1]+(-0.645673968637)*x_ref[2]**o+(0.913753411758)*x_ref[2]
        ref[3,3,2,0]=(0.0623921286142)*x_ref[0]**o+(0.393414889204)*x_ref[0]+(-0.300034851075)*x_ref[1]**o+(-0.519515206457)*x_ref[1]+(-0.0074981731879)*x_ref[2]**o+(0.202996242895)*x_ref[2]
        ref[3,3,2,1]=(0.583279597505)*x_ref[0]**o+(-0.940603487209)*x_ref[0]+(0.027966157338)*x_ref[1]**o+(0.251466344692)*x_ref[1]+(-0.97011033154)*x_ref[2]**o+(0.46527056203)*x_ref[2]
        ref[3,4,0,0]=(0.865170740218)*x_ref[0]**o+(-0.466853211289)*x_ref[0]+(0.0935492537278)*x_ref[1]**o+(-0.632374632657)*x_ref[1]+(0.673466346463)*x_ref[2]**o+(-0.47934739873)*x_ref[2]
        ref[3,4,0,1]=(0.0417451136027)*x_ref[0]**o+(-0.781570733286)*x_ref[0]+(0.547452389454)*x_ref[1]**o+(-0.0286368211528)*x_ref[1]+(-0.98907358808)*x_ref[2]**o+(0.62749495057)*x_ref[2]
        ref[3,4,1,0]=(-0.632067479664)*x_ref[0]**o+(-0.872835386336)*x_ref[0]+(-0.228785631078)*x_ref[1]**o+(-0.224206127306)*x_ref[1]+(0.769729051217)*x_ref[2]**o+(-0.406750034352)*x_ref[2]
        ref[3,4,1,1]=(0.216883071299)*x_ref[0]**o+(-0.225632200121)*x_ref[0]+(0.827186402743)*x_ref[1]**o+(0.161488636886)*x_ref[1]+(0.773801288836)*x_ref[2]**o+(0.977006832166)*x_ref[2]
        ref[3,4,2,0]=(-0.867098845884)*x_ref[0]**o+(0.14989544838)*x_ref[0]+(0.756720043262)*x_ref[1]**o+(-0.496074302626)*x_ref[1]+(-0.662787072451)*x_ref[2]**o+(-0.142473973368)*x_ref[2]
        ref[3,4,2,1]=(-0.294128396829)*x_ref[0]**o+(-0.931379392091)*x_ref[0]+(0.906552131161)*x_ref[1]**o+(-0.422008485494)*x_ref[1]+(0.806742455679)*x_ref[2]**o+(0.940488803006)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.387919945961)*x[0]**o+(-0.65988832446)*x[0]+(-0.271357984194)*x[1]**o+(0.911983947389)*x[1]
        ref=(-0.387919945961)*x_ref[0]**o+(-0.65988832446)*x_ref[0]+(-0.271357984194)*x_ref[1]**o+(0.911983947389)*x_ref[1]
      else:
        arg=(-0.0447075351093)*x[0]**o+(0.253461091432)*x[0]+(-0.218567784216)*x[1]**o+(-0.79030708754)*x[1]+(0.977690873872)*x[2]**o+(0.832406975763)*x[2]
        ref=(-0.0447075351093)*x_ref[0]**o+(0.253461091432)*x_ref[0]+(-0.218567784216)*x_ref[1]**o+(-0.79030708754)*x_ref[1]+(0.977690873872)*x_ref[2]**o+(0.832406975763)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.0351141106851)*x[0]**o+(-0.330587401)*x[0]+(0.244307738971)*x[1]**o+(0.623126601064)*x[1]
        arg[1]=(-0.828515624351)*x[0]**o+(-0.128313696693)*x[0]+(-0.419830633587)*x[1]**o+(0.207694344977)*x[1]
        ref[0]=(-0.0351141106851)*x_ref[0]**o+(-0.330587401)*x_ref[0]+(0.244307738971)*x_ref[1]**o+(0.623126601064)*x_ref[1]
        ref[1]=(-0.828515624351)*x_ref[0]**o+(-0.128313696693)*x_ref[0]+(-0.419830633587)*x_ref[1]**o+(0.207694344977)*x_ref[1]
      else:
        arg[0]=(-0.704794455532)*x[0]**o+(0.913653448977)*x[0]+(-0.159574360706)*x[1]**o+(-0.766199806615)*x[1]+(-0.571643737767)*x[2]**o+(-0.818256320426)*x[2]
        arg[1]=(-0.951104850067)*x[0]**o+(-0.661009825365)*x[0]+(0.415252265672)*x[1]**o+(-0.373834229809)*x[1]+(-0.727526942776)*x[2]**o+(-0.832051613933)*x[2]
        ref[0]=(-0.704794455532)*x_ref[0]**o+(0.913653448977)*x_ref[0]+(-0.159574360706)*x_ref[1]**o+(-0.766199806615)*x_ref[1]+(-0.571643737767)*x_ref[2]**o+(-0.818256320426)*x_ref[2]
        ref[1]=(-0.951104850067)*x_ref[0]**o+(-0.661009825365)*x_ref[0]+(0.415252265672)*x_ref[1]**o+(-0.373834229809)*x_ref[1]+(-0.727526942776)*x_ref[2]**o+(-0.832051613933)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.597205522265)*x[0]**o+(-0.990224489001)*x[0]+(0.409554155812)*x[1]**o+(0.0483989008223)*x[1]
        arg[0,1]=(0.397936023817)*x[0]**o+(-0.930167709289)*x[0]+(0.92611965278)*x[1]**o+(-0.964519543905)*x[1]
        arg[0,2]=(-0.311484417828)*x[0]**o+(-0.602916971577)*x[0]+(-0.420016358622)*x[1]**o+(0.837176635137)*x[1]
        arg[0,3]=(0.819115285277)*x[0]**o+(0.34838887804)*x[0]+(-0.690550904634)*x[1]**o+(0.373665882427)*x[1]
        arg[0,4]=(-0.7631912026)*x[0]**o+(-0.945320401033)*x[0]+(0.694292397122)*x[1]**o+(-0.420347487671)*x[1]
        arg[1,0]=(-0.572828652444)*x[0]**o+(0.354646387926)*x[0]+(-0.0507519017004)*x[1]**o+(-0.613889773716)*x[1]
        arg[1,1]=(-0.816957267715)*x[0]**o+(0.733537081373)*x[0]+(-0.660041553847)*x[1]**o+(0.441096986506)*x[1]
        arg[1,2]=(-0.803370878245)*x[0]**o+(-0.876568267674)*x[0]+(0.530143828975)*x[1]**o+(0.208130191749)*x[1]
        arg[1,3]=(-0.471900498331)*x[0]**o+(0.537812384385)*x[0]+(-0.882977296775)*x[1]**o+(-0.0668303027948)*x[1]
        arg[1,4]=(0.145525825808)*x[0]**o+(-0.743588928172)*x[0]+(-0.571360483951)*x[1]**o+(-0.779639689697)*x[1]
        arg[2,0]=(0.664624516473)*x[0]**o+(-0.903029393285)*x[0]+(-0.602274374813)*x[1]**o+(-0.0349025848298)*x[1]
        arg[2,1]=(-0.826235832601)*x[0]**o+(-0.56956830806)*x[0]+(-0.720529973236)*x[1]**o+(0.844549160654)*x[1]
        arg[2,2]=(0.865808902895)*x[0]**o+(0.806173922192)*x[0]+(-0.430500586315)*x[1]**o+(-0.903014287464)*x[1]
        arg[2,3]=(0.567909856768)*x[0]**o+(0.702374819746)*x[0]+(-0.931635013223)*x[1]**o+(0.35484582218)*x[1]
        arg[2,4]=(0.98750760384)*x[0]**o+(0.759340896388)*x[0]+(0.54913653446)*x[1]**o+(-0.620836882788)*x[1]
        arg[3,0]=(-0.886288358587)*x[0]**o+(0.393932009441)*x[0]+(0.370738890674)*x[1]**o+(0.540588748949)*x[1]
        arg[3,1]=(0.281435147449)*x[0]**o+(0.455838903759)*x[0]+(0.115176811183)*x[1]**o+(0.214335969407)*x[1]
        arg[3,2]=(-0.939824922583)*x[0]**o+(-0.725457358605)*x[0]+(0.0326844083908)*x[1]**o+(-0.743922869733)*x[1]
        arg[3,3]=(-0.189824771744)*x[0]**o+(-0.26508334021)*x[0]+(-0.323185410564)*x[1]**o+(0.242480018197)*x[1]
        arg[3,4]=(-0.675784178633)*x[0]**o+(0.0995857467734)*x[0]+(0.564599922701)*x[1]**o+(-0.716819470891)*x[1]
        ref[0,0]=(0.597205522265)*x_ref[0]**o+(-0.990224489001)*x_ref[0]+(0.409554155812)*x_ref[1]**o+(0.0483989008223)*x_ref[1]
        ref[0,1]=(0.397936023817)*x_ref[0]**o+(-0.930167709289)*x_ref[0]+(0.92611965278)*x_ref[1]**o+(-0.964519543905)*x_ref[1]
        ref[0,2]=(-0.311484417828)*x_ref[0]**o+(-0.602916971577)*x_ref[0]+(-0.420016358622)*x_ref[1]**o+(0.837176635137)*x_ref[1]
        ref[0,3]=(0.819115285277)*x_ref[0]**o+(0.34838887804)*x_ref[0]+(-0.690550904634)*x_ref[1]**o+(0.373665882427)*x_ref[1]
        ref[0,4]=(-0.7631912026)*x_ref[0]**o+(-0.945320401033)*x_ref[0]+(0.694292397122)*x_ref[1]**o+(-0.420347487671)*x_ref[1]
        ref[1,0]=(-0.572828652444)*x_ref[0]**o+(0.354646387926)*x_ref[0]+(-0.0507519017004)*x_ref[1]**o+(-0.613889773716)*x_ref[1]
        ref[1,1]=(-0.816957267715)*x_ref[0]**o+(0.733537081373)*x_ref[0]+(-0.660041553847)*x_ref[1]**o+(0.441096986506)*x_ref[1]
        ref[1,2]=(-0.803370878245)*x_ref[0]**o+(-0.876568267674)*x_ref[0]+(0.530143828975)*x_ref[1]**o+(0.208130191749)*x_ref[1]
        ref[1,3]=(-0.471900498331)*x_ref[0]**o+(0.537812384385)*x_ref[0]+(-0.882977296775)*x_ref[1]**o+(-0.0668303027948)*x_ref[1]
        ref[1,4]=(0.145525825808)*x_ref[0]**o+(-0.743588928172)*x_ref[0]+(-0.571360483951)*x_ref[1]**o+(-0.779639689697)*x_ref[1]
        ref[2,0]=(0.664624516473)*x_ref[0]**o+(-0.903029393285)*x_ref[0]+(-0.602274374813)*x_ref[1]**o+(-0.0349025848298)*x_ref[1]
        ref[2,1]=(-0.826235832601)*x_ref[0]**o+(-0.56956830806)*x_ref[0]+(-0.720529973236)*x_ref[1]**o+(0.844549160654)*x_ref[1]
        ref[2,2]=(0.865808902895)*x_ref[0]**o+(0.806173922192)*x_ref[0]+(-0.430500586315)*x_ref[1]**o+(-0.903014287464)*x_ref[1]
        ref[2,3]=(0.567909856768)*x_ref[0]**o+(0.702374819746)*x_ref[0]+(-0.931635013223)*x_ref[1]**o+(0.35484582218)*x_ref[1]
        ref[2,4]=(0.98750760384)*x_ref[0]**o+(0.759340896388)*x_ref[0]+(0.54913653446)*x_ref[1]**o+(-0.620836882788)*x_ref[1]
        ref[3,0]=(-0.886288358587)*x_ref[0]**o+(0.393932009441)*x_ref[0]+(0.370738890674)*x_ref[1]**o+(0.540588748949)*x_ref[1]
        ref[3,1]=(0.281435147449)*x_ref[0]**o+(0.455838903759)*x_ref[0]+(0.115176811183)*x_ref[1]**o+(0.214335969407)*x_ref[1]
        ref[3,2]=(-0.939824922583)*x_ref[0]**o+(-0.725457358605)*x_ref[0]+(0.0326844083908)*x_ref[1]**o+(-0.743922869733)*x_ref[1]
        ref[3,3]=(-0.189824771744)*x_ref[0]**o+(-0.26508334021)*x_ref[0]+(-0.323185410564)*x_ref[1]**o+(0.242480018197)*x_ref[1]
        ref[3,4]=(-0.675784178633)*x_ref[0]**o+(0.0995857467734)*x_ref[0]+(0.564599922701)*x_ref[1]**o+(-0.716819470891)*x_ref[1]
      else:
        arg[0,0]=(0.0429272016915)*x[0]**o+(-0.277438145968)*x[0]+(-0.663736923945)*x[1]**o+(0.754568662274)*x[1]+(-0.702929906882)*x[2]**o+(-0.744050838266)*x[2]
        arg[0,1]=(-0.245995343753)*x[0]**o+(-0.838699721329)*x[0]+(-0.266398695342)*x[1]**o+(0.116104359849)*x[1]+(-0.62406648751)*x[2]**o+(-0.85767565383)*x[2]
        arg[0,2]=(-0.969332932419)*x[0]**o+(-0.815839911463)*x[0]+(0.0532903209734)*x[1]**o+(-0.00108699727615)*x[1]+(0.142541999046)*x[2]**o+(0.971482543817)*x[2]
        arg[0,3]=(-0.73941517118)*x[0]**o+(0.728994326071)*x[0]+(0.962711932241)*x[1]**o+(-0.418607572448)*x[1]+(-0.374703265781)*x[2]**o+(-0.278661032739)*x[2]
        arg[0,4]=(-0.258020436163)*x[0]**o+(-0.0252121036831)*x[0]+(-0.756480314139)*x[1]**o+(-0.705045476776)*x[1]+(0.222309499093)*x[2]**o+(0.999237507123)*x[2]
        arg[1,0]=(-0.00753673411562)*x[0]**o+(-0.712947643092)*x[0]+(-0.946860154348)*x[1]**o+(-0.887293091702)*x[1]+(-0.57071975934)*x[2]**o+(-0.000995212472675)*x[2]
        arg[1,1]=(-0.547724214274)*x[0]**o+(-0.338492998124)*x[0]+(-0.191652366102)*x[1]**o+(-0.260315919511)*x[1]+(-0.103992070026)*x[2]**o+(-0.836638261621)*x[2]
        arg[1,2]=(-0.354711259362)*x[0]**o+(-0.916549411392)*x[0]+(0.744081321252)*x[1]**o+(-0.375805257126)*x[1]+(-0.78900115815)*x[2]**o+(-0.431116699853)*x[2]
        arg[1,3]=(0.543469030875)*x[0]**o+(-0.154316090927)*x[0]+(0.794234357287)*x[1]**o+(-0.894714087173)*x[1]+(-0.523490310245)*x[2]**o+(-0.734864830975)*x[2]
        arg[1,4]=(0.856478460898)*x[0]**o+(0.550846305146)*x[0]+(0.883099588119)*x[1]**o+(-0.928432890741)*x[1]+(0.953422014409)*x[2]**o+(-0.0295234113856)*x[2]
        arg[2,0]=(0.391420547389)*x[0]**o+(-0.0618665973255)*x[0]+(-0.536627285311)*x[1]**o+(-0.701891441965)*x[1]+(0.433098401229)*x[2]**o+(-0.821674981019)*x[2]
        arg[2,1]=(0.102455732687)*x[0]**o+(0.168078832413)*x[0]+(0.911099070324)*x[1]**o+(-0.776659572448)*x[1]+(0.942192938907)*x[2]**o+(-0.96284944292)*x[2]
        arg[2,2]=(0.677666801373)*x[0]**o+(0.579429978893)*x[0]+(-0.642666485576)*x[1]**o+(-0.0709929161913)*x[1]+(-0.686526389023)*x[2]**o+(0.0941806754766)*x[2]
        arg[2,3]=(-0.942137271102)*x[0]**o+(-0.448762007046)*x[0]+(0.403668316046)*x[1]**o+(-0.885790833279)*x[1]+(0.0448846125948)*x[2]**o+(0.123128346219)*x[2]
        arg[2,4]=(0.64615774988)*x[0]**o+(0.641971975809)*x[0]+(-0.0486200548159)*x[1]**o+(0.106284669825)*x[1]+(0.968867785284)*x[2]**o+(-0.693546581787)*x[2]
        arg[3,0]=(-0.0926306498682)*x[0]**o+(0.759541313455)*x[0]+(-0.569638025121)*x[1]**o+(0.236957551756)*x[1]+(-0.137429816873)*x[2]**o+(0.285836556676)*x[2]
        arg[3,1]=(0.0382484871512)*x[0]**o+(-0.481489628305)*x[0]+(0.809404670659)*x[1]**o+(0.124155422674)*x[1]+(-0.239216545446)*x[2]**o+(0.436110807769)*x[2]
        arg[3,2]=(-0.154044250435)*x[0]**o+(0.730285603192)*x[0]+(-0.986531989452)*x[1]**o+(0.0424454222102)*x[1]+(0.9651845632)*x[2]**o+(-0.173872544124)*x[2]
        arg[3,3]=(-0.528075843446)*x[0]**o+(0.226347788534)*x[0]+(-0.314941509835)*x[1]**o+(0.508180139398)*x[1]+(0.681251000904)*x[2]**o+(-0.442663397606)*x[2]
        arg[3,4]=(-0.335444251365)*x[0]**o+(-0.7788088839)*x[0]+(0.602538296871)*x[1]**o+(-0.440564659745)*x[1]+(-0.204809090144)*x[2]**o+(-0.449409088071)*x[2]
        ref[0,0]=(0.0429272016915)*x_ref[0]**o+(-0.277438145968)*x_ref[0]+(-0.663736923945)*x_ref[1]**o+(0.754568662274)*x_ref[1]+(-0.702929906882)*x_ref[2]**o+(-0.744050838266)*x_ref[2]
        ref[0,1]=(-0.245995343753)*x_ref[0]**o+(-0.838699721329)*x_ref[0]+(-0.266398695342)*x_ref[1]**o+(0.116104359849)*x_ref[1]+(-0.62406648751)*x_ref[2]**o+(-0.85767565383)*x_ref[2]
        ref[0,2]=(-0.969332932419)*x_ref[0]**o+(-0.815839911463)*x_ref[0]+(0.0532903209734)*x_ref[1]**o+(-0.00108699727615)*x_ref[1]+(0.142541999046)*x_ref[2]**o+(0.971482543817)*x_ref[2]
        ref[0,3]=(-0.73941517118)*x_ref[0]**o+(0.728994326071)*x_ref[0]+(0.962711932241)*x_ref[1]**o+(-0.418607572448)*x_ref[1]+(-0.374703265781)*x_ref[2]**o+(-0.278661032739)*x_ref[2]
        ref[0,4]=(-0.258020436163)*x_ref[0]**o+(-0.0252121036831)*x_ref[0]+(-0.756480314139)*x_ref[1]**o+(-0.705045476776)*x_ref[1]+(0.222309499093)*x_ref[2]**o+(0.999237507123)*x_ref[2]
        ref[1,0]=(-0.00753673411562)*x_ref[0]**o+(-0.712947643092)*x_ref[0]+(-0.946860154348)*x_ref[1]**o+(-0.887293091702)*x_ref[1]+(-0.57071975934)*x_ref[2]**o+(-0.000995212472675)*x_ref[2]
        ref[1,1]=(-0.547724214274)*x_ref[0]**o+(-0.338492998124)*x_ref[0]+(-0.191652366102)*x_ref[1]**o+(-0.260315919511)*x_ref[1]+(-0.103992070026)*x_ref[2]**o+(-0.836638261621)*x_ref[2]
        ref[1,2]=(-0.354711259362)*x_ref[0]**o+(-0.916549411392)*x_ref[0]+(0.744081321252)*x_ref[1]**o+(-0.375805257126)*x_ref[1]+(-0.78900115815)*x_ref[2]**o+(-0.431116699853)*x_ref[2]
        ref[1,3]=(0.543469030875)*x_ref[0]**o+(-0.154316090927)*x_ref[0]+(0.794234357287)*x_ref[1]**o+(-0.894714087173)*x_ref[1]+(-0.523490310245)*x_ref[2]**o+(-0.734864830975)*x_ref[2]
        ref[1,4]=(0.856478460898)*x_ref[0]**o+(0.550846305146)*x_ref[0]+(0.883099588119)*x_ref[1]**o+(-0.928432890741)*x_ref[1]+(0.953422014409)*x_ref[2]**o+(-0.0295234113856)*x_ref[2]
        ref[2,0]=(0.391420547389)*x_ref[0]**o+(-0.0618665973255)*x_ref[0]+(-0.536627285311)*x_ref[1]**o+(-0.701891441965)*x_ref[1]+(0.433098401229)*x_ref[2]**o+(-0.821674981019)*x_ref[2]
        ref[2,1]=(0.102455732687)*x_ref[0]**o+(0.168078832413)*x_ref[0]+(0.911099070324)*x_ref[1]**o+(-0.776659572448)*x_ref[1]+(0.942192938907)*x_ref[2]**o+(-0.96284944292)*x_ref[2]
        ref[2,2]=(0.677666801373)*x_ref[0]**o+(0.579429978893)*x_ref[0]+(-0.642666485576)*x_ref[1]**o+(-0.0709929161913)*x_ref[1]+(-0.686526389023)*x_ref[2]**o+(0.0941806754766)*x_ref[2]
        ref[2,3]=(-0.942137271102)*x_ref[0]**o+(-0.448762007046)*x_ref[0]+(0.403668316046)*x_ref[1]**o+(-0.885790833279)*x_ref[1]+(0.0448846125948)*x_ref[2]**o+(0.123128346219)*x_ref[2]
        ref[2,4]=(0.64615774988)*x_ref[0]**o+(0.641971975809)*x_ref[0]+(-0.0486200548159)*x_ref[1]**o+(0.106284669825)*x_ref[1]+(0.968867785284)*x_ref[2]**o+(-0.693546581787)*x_ref[2]
        ref[3,0]=(-0.0926306498682)*x_ref[0]**o+(0.759541313455)*x_ref[0]+(-0.569638025121)*x_ref[1]**o+(0.236957551756)*x_ref[1]+(-0.137429816873)*x_ref[2]**o+(0.285836556676)*x_ref[2]
        ref[3,1]=(0.0382484871512)*x_ref[0]**o+(-0.481489628305)*x_ref[0]+(0.809404670659)*x_ref[1]**o+(0.124155422674)*x_ref[1]+(-0.239216545446)*x_ref[2]**o+(0.436110807769)*x_ref[2]
        ref[3,2]=(-0.154044250435)*x_ref[0]**o+(0.730285603192)*x_ref[0]+(-0.986531989452)*x_ref[1]**o+(0.0424454222102)*x_ref[1]+(0.9651845632)*x_ref[2]**o+(-0.173872544124)*x_ref[2]
        ref[3,3]=(-0.528075843446)*x_ref[0]**o+(0.226347788534)*x_ref[0]+(-0.314941509835)*x_ref[1]**o+(0.508180139398)*x_ref[1]+(0.681251000904)*x_ref[2]**o+(-0.442663397606)*x_ref[2]
        ref[3,4]=(-0.335444251365)*x_ref[0]**o+(-0.7788088839)*x_ref[0]+(0.602538296871)*x_ref[1]**o+(-0.440564659745)*x_ref[1]+(-0.204809090144)*x_ref[2]**o+(-0.449409088071)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.88752640419)*x[0]**o+(-0.432514405615)*x[0]+(0.714823891809)*x[1]**o+(0.697307667986)*x[1]
        arg[0,0,1]=(0.738741506554)*x[0]**o+(0.188618518993)*x[0]+(-0.482898810269)*x[1]**o+(-0.914123176195)*x[1]
        arg[0,1,0]=(0.863872888691)*x[0]**o+(0.74507667202)*x[0]+(0.363690250942)*x[1]**o+(0.578470064401)*x[1]
        arg[0,1,1]=(-0.764594250812)*x[0]**o+(0.692105995894)*x[0]+(-0.950074761486)*x[1]**o+(-0.229643756622)*x[1]
        arg[1,0,0]=(-0.514123807377)*x[0]**o+(-0.770729480038)*x[0]+(-0.0503641588682)*x[1]**o+(-0.334340604627)*x[1]
        arg[1,0,1]=(-0.753427043851)*x[0]**o+(0.482153344254)*x[0]+(0.00162169870171)*x[1]**o+(-0.877830413948)*x[1]
        arg[1,1,0]=(0.603963041861)*x[0]**o+(-0.545048115282)*x[0]+(0.464500827027)*x[1]**o+(-0.527556750287)*x[1]
        arg[1,1,1]=(0.753201247064)*x[0]**o+(-0.99039225613)*x[0]+(-0.787959315932)*x[1]**o+(-0.918878072132)*x[1]
        arg[2,0,0]=(0.254002467345)*x[0]**o+(-0.918196533244)*x[0]+(-0.178408279918)*x[1]**o+(0.14954396814)*x[1]
        arg[2,0,1]=(0.449025666157)*x[0]**o+(-0.854430018688)*x[0]+(0.0393762923755)*x[1]**o+(0.576680446192)*x[1]
        arg[2,1,0]=(-0.252429553053)*x[0]**o+(-0.300398117312)*x[0]+(-0.0197102583755)*x[1]**o+(-0.189780116752)*x[1]
        arg[2,1,1]=(0.406291235871)*x[0]**o+(-0.763190166909)*x[0]+(-0.219957648446)*x[1]**o+(-0.928990138166)*x[1]
        arg[3,0,0]=(0.947321725555)*x[0]**o+(0.35310915009)*x[0]+(0.436067723051)*x[1]**o+(-0.86339979781)*x[1]
        arg[3,0,1]=(0.349458398624)*x[0]**o+(-0.139435138128)*x[0]+(-0.239809689531)*x[1]**o+(-0.0600998666403)*x[1]
        arg[3,1,0]=(-0.630539588121)*x[0]**o+(-0.700765237169)*x[0]+(-0.333988786061)*x[1]**o+(-0.125401172016)*x[1]
        arg[3,1,1]=(0.726389530806)*x[0]**o+(-0.691681620304)*x[0]+(0.241060086313)*x[1]**o+(-0.0970053499082)*x[1]
        arg[4,0,0]=(-0.798548958727)*x[0]**o+(0.429037250139)*x[0]+(0.598115431903)*x[1]**o+(-0.688464532796)*x[1]
        arg[4,0,1]=(0.698302852699)*x[0]**o+(0.36852807703)*x[0]+(-0.570086427187)*x[1]**o+(0.110387807246)*x[1]
        arg[4,1,0]=(0.295729624355)*x[0]**o+(0.247060092718)*x[0]+(-0.874774487091)*x[1]**o+(-0.40111840264)*x[1]
        arg[4,1,1]=(-0.773206505432)*x[0]**o+(-0.238827976714)*x[0]+(0.994288595399)*x[1]**o+(-0.6429945796)*x[1]
        arg[5,0,0]=(-0.398445578383)*x[0]**o+(-0.893518494161)*x[0]+(-0.855543638814)*x[1]**o+(0.93304149836)*x[1]
        arg[5,0,1]=(-0.811797150327)*x[0]**o+(0.0702614194127)*x[0]+(-0.0906132870911)*x[1]**o+(-0.944409144857)*x[1]
        arg[5,1,0]=(0.710407361984)*x[0]**o+(-0.854493576437)*x[0]+(-0.354683789897)*x[1]**o+(-0.896605671415)*x[1]
        arg[5,1,1]=(0.0668604398168)*x[0]**o+(0.464597275769)*x[0]+(-0.683756009627)*x[1]**o+(-0.368806656344)*x[1]
        ref[0,0,0]=(0.88752640419)*x_ref[0]**o+(-0.432514405615)*x_ref[0]+(0.714823891809)*x_ref[1]**o+(0.697307667986)*x_ref[1]
        ref[0,0,1]=(0.738741506554)*x_ref[0]**o+(0.188618518993)*x_ref[0]+(-0.482898810269)*x_ref[1]**o+(-0.914123176195)*x_ref[1]
        ref[0,1,0]=(0.863872888691)*x_ref[0]**o+(0.74507667202)*x_ref[0]+(0.363690250942)*x_ref[1]**o+(0.578470064401)*x_ref[1]
        ref[0,1,1]=(-0.764594250812)*x_ref[0]**o+(0.692105995894)*x_ref[0]+(-0.950074761486)*x_ref[1]**o+(-0.229643756622)*x_ref[1]
        ref[1,0,0]=(-0.514123807377)*x_ref[0]**o+(-0.770729480038)*x_ref[0]+(-0.0503641588682)*x_ref[1]**o+(-0.334340604627)*x_ref[1]
        ref[1,0,1]=(-0.753427043851)*x_ref[0]**o+(0.482153344254)*x_ref[0]+(0.00162169870171)*x_ref[1]**o+(-0.877830413948)*x_ref[1]
        ref[1,1,0]=(0.603963041861)*x_ref[0]**o+(-0.545048115282)*x_ref[0]+(0.464500827027)*x_ref[1]**o+(-0.527556750287)*x_ref[1]
        ref[1,1,1]=(0.753201247064)*x_ref[0]**o+(-0.99039225613)*x_ref[0]+(-0.787959315932)*x_ref[1]**o+(-0.918878072132)*x_ref[1]
        ref[2,0,0]=(0.254002467345)*x_ref[0]**o+(-0.918196533244)*x_ref[0]+(-0.178408279918)*x_ref[1]**o+(0.14954396814)*x_ref[1]
        ref[2,0,1]=(0.449025666157)*x_ref[0]**o+(-0.854430018688)*x_ref[0]+(0.0393762923755)*x_ref[1]**o+(0.576680446192)*x_ref[1]
        ref[2,1,0]=(-0.252429553053)*x_ref[0]**o+(-0.300398117312)*x_ref[0]+(-0.0197102583755)*x_ref[1]**o+(-0.189780116752)*x_ref[1]
        ref[2,1,1]=(0.406291235871)*x_ref[0]**o+(-0.763190166909)*x_ref[0]+(-0.219957648446)*x_ref[1]**o+(-0.928990138166)*x_ref[1]
        ref[3,0,0]=(0.947321725555)*x_ref[0]**o+(0.35310915009)*x_ref[0]+(0.436067723051)*x_ref[1]**o+(-0.86339979781)*x_ref[1]
        ref[3,0,1]=(0.349458398624)*x_ref[0]**o+(-0.139435138128)*x_ref[0]+(-0.239809689531)*x_ref[1]**o+(-0.0600998666403)*x_ref[1]
        ref[3,1,0]=(-0.630539588121)*x_ref[0]**o+(-0.700765237169)*x_ref[0]+(-0.333988786061)*x_ref[1]**o+(-0.125401172016)*x_ref[1]
        ref[3,1,1]=(0.726389530806)*x_ref[0]**o+(-0.691681620304)*x_ref[0]+(0.241060086313)*x_ref[1]**o+(-0.0970053499082)*x_ref[1]
        ref[4,0,0]=(-0.798548958727)*x_ref[0]**o+(0.429037250139)*x_ref[0]+(0.598115431903)*x_ref[1]**o+(-0.688464532796)*x_ref[1]
        ref[4,0,1]=(0.698302852699)*x_ref[0]**o+(0.36852807703)*x_ref[0]+(-0.570086427187)*x_ref[1]**o+(0.110387807246)*x_ref[1]
        ref[4,1,0]=(0.295729624355)*x_ref[0]**o+(0.247060092718)*x_ref[0]+(-0.874774487091)*x_ref[1]**o+(-0.40111840264)*x_ref[1]
        ref[4,1,1]=(-0.773206505432)*x_ref[0]**o+(-0.238827976714)*x_ref[0]+(0.994288595399)*x_ref[1]**o+(-0.6429945796)*x_ref[1]
        ref[5,0,0]=(-0.398445578383)*x_ref[0]**o+(-0.893518494161)*x_ref[0]+(-0.855543638814)*x_ref[1]**o+(0.93304149836)*x_ref[1]
        ref[5,0,1]=(-0.811797150327)*x_ref[0]**o+(0.0702614194127)*x_ref[0]+(-0.0906132870911)*x_ref[1]**o+(-0.944409144857)*x_ref[1]
        ref[5,1,0]=(0.710407361984)*x_ref[0]**o+(-0.854493576437)*x_ref[0]+(-0.354683789897)*x_ref[1]**o+(-0.896605671415)*x_ref[1]
        ref[5,1,1]=(0.0668604398168)*x_ref[0]**o+(0.464597275769)*x_ref[0]+(-0.683756009627)*x_ref[1]**o+(-0.368806656344)*x_ref[1]
      else:
        arg[0,0,0]=(-0.0873708254613)*x[0]**o+(-0.843989447866)*x[0]+(0.44751871924)*x[1]**o+(0.56511771844)*x[1]+(0.367414341304)*x[2]**o+(-0.147729223167)*x[2]
        arg[0,0,1]=(0.00752707306425)*x[0]**o+(0.819213438452)*x[0]+(0.14826826627)*x[1]**o+(-0.169362794728)*x[1]+(0.51736237082)*x[2]**o+(0.832746819884)*x[2]
        arg[0,1,0]=(0.607746431545)*x[0]**o+(0.253170100132)*x[0]+(-0.992132084299)*x[1]**o+(0.353865695431)*x[1]+(-0.787229685473)*x[2]**o+(0.86154695401)*x[2]
        arg[0,1,1]=(0.563636829468)*x[0]**o+(0.459627185677)*x[0]+(-0.511957702356)*x[1]**o+(0.309573033639)*x[1]+(0.806861185555)*x[2]**o+(0.182177897377)*x[2]
        arg[1,0,0]=(0.600850271814)*x[0]**o+(0.440922206975)*x[0]+(0.0307377466119)*x[1]**o+(-0.419096719454)*x[1]+(0.979284456332)*x[2]**o+(-0.00240732351601)*x[2]
        arg[1,0,1]=(0.51273517393)*x[0]**o+(-0.947643736857)*x[0]+(0.272403256871)*x[1]**o+(0.0119584891405)*x[1]+(-0.290604573286)*x[2]**o+(0.822291254588)*x[2]
        arg[1,1,0]=(-0.201909330605)*x[0]**o+(-0.560221198146)*x[0]+(0.564306911695)*x[1]**o+(-0.936933580593)*x[1]+(-0.41391834592)*x[2]**o+(0.133244793725)*x[2]
        arg[1,1,1]=(0.13318221571)*x[0]**o+(-0.565278280388)*x[0]+(0.677613822728)*x[1]**o+(-0.386612668638)*x[1]+(0.207570026134)*x[2]**o+(0.798362457061)*x[2]
        arg[2,0,0]=(-0.3777906845)*x[0]**o+(-0.285455928514)*x[0]+(-0.0901885959293)*x[1]**o+(-0.385922214051)*x[1]+(0.322566508003)*x[2]**o+(0.171191301702)*x[2]
        arg[2,0,1]=(-0.113954650916)*x[0]**o+(-0.373563461907)*x[0]+(0.73321196284)*x[1]**o+(0.535150793249)*x[1]+(-0.736944901672)*x[2]**o+(0.86133553525)*x[2]
        arg[2,1,0]=(-0.724228484035)*x[0]**o+(0.310971627294)*x[0]+(-0.497468717028)*x[1]**o+(0.467315341957)*x[1]+(-0.598941984951)*x[2]**o+(0.990367400534)*x[2]
        arg[2,1,1]=(0.594502107684)*x[0]**o+(0.261477996581)*x[0]+(0.738413631986)*x[1]**o+(0.363468523272)*x[1]+(-0.945845846052)*x[2]**o+(0.281144164925)*x[2]
        arg[3,0,0]=(0.804174935673)*x[0]**o+(-0.633954348085)*x[0]+(-0.953193492156)*x[1]**o+(0.632681376482)*x[1]+(-0.409900341362)*x[2]**o+(0.951988363964)*x[2]
        arg[3,0,1]=(-0.469895421121)*x[0]**o+(0.666115304582)*x[0]+(0.676351738247)*x[1]**o+(-0.0738256360284)*x[1]+(0.34976675523)*x[2]**o+(-0.561058490613)*x[2]
        arg[3,1,0]=(-0.999916478236)*x[0]**o+(-0.362871459365)*x[0]+(0.645604661917)*x[1]**o+(0.215878330855)*x[1]+(-0.0561173996823)*x[2]**o+(0.272143302204)*x[2]
        arg[3,1,1]=(0.233461828376)*x[0]**o+(-0.110673705352)*x[0]+(0.699710673197)*x[1]**o+(-0.628805264645)*x[1]+(0.677467640352)*x[2]**o+(-0.156224522726)*x[2]
        arg[4,0,0]=(0.180468931719)*x[0]**o+(0.299451524372)*x[0]+(-0.677131137371)*x[1]**o+(-0.422802649252)*x[1]+(0.237349200367)*x[2]**o+(0.739758623757)*x[2]
        arg[4,0,1]=(0.22209981155)*x[0]**o+(-0.226751058136)*x[0]+(-0.0345697170216)*x[1]**o+(0.265795819906)*x[1]+(0.542547943322)*x[2]**o+(0.785610668157)*x[2]
        arg[4,1,0]=(-0.520898604151)*x[0]**o+(-0.84307105084)*x[0]+(0.818061645146)*x[1]**o+(-0.173710357379)*x[1]+(0.344023171063)*x[2]**o+(0.194662166094)*x[2]
        arg[4,1,1]=(-0.288466404824)*x[0]**o+(-0.891856587908)*x[0]+(0.378329571068)*x[1]**o+(0.437919935526)*x[1]+(-0.630733422962)*x[2]**o+(0.518444630882)*x[2]
        arg[5,0,0]=(0.00978685202942)*x[0]**o+(-0.474443494209)*x[0]+(-0.509964215112)*x[1]**o+(0.341199671453)*x[1]+(-0.459333887251)*x[2]**o+(0.491961590667)*x[2]
        arg[5,0,1]=(-0.899850325159)*x[0]**o+(-0.191369137667)*x[0]+(0.18882430985)*x[1]**o+(-0.185214597442)*x[1]+(-0.967187797553)*x[2]**o+(0.366127839986)*x[2]
        arg[5,1,0]=(0.428007441668)*x[0]**o+(0.920828656579)*x[0]+(-0.298863178441)*x[1]**o+(-0.226020983335)*x[1]+(-0.01907141259)*x[2]**o+(-0.359698040557)*x[2]
        arg[5,1,1]=(0.980673633441)*x[0]**o+(0.362357489872)*x[0]+(0.825233899658)*x[1]**o+(0.846433827949)*x[1]+(-0.870197174533)*x[2]**o+(0.621884204716)*x[2]
        ref[0,0,0]=(-0.0873708254613)*x_ref[0]**o+(-0.843989447866)*x_ref[0]+(0.44751871924)*x_ref[1]**o+(0.56511771844)*x_ref[1]+(0.367414341304)*x_ref[2]**o+(-0.147729223167)*x_ref[2]
        ref[0,0,1]=(0.00752707306425)*x_ref[0]**o+(0.819213438452)*x_ref[0]+(0.14826826627)*x_ref[1]**o+(-0.169362794728)*x_ref[1]+(0.51736237082)*x_ref[2]**o+(0.832746819884)*x_ref[2]
        ref[0,1,0]=(0.607746431545)*x_ref[0]**o+(0.253170100132)*x_ref[0]+(-0.992132084299)*x_ref[1]**o+(0.353865695431)*x_ref[1]+(-0.787229685473)*x_ref[2]**o+(0.86154695401)*x_ref[2]
        ref[0,1,1]=(0.563636829468)*x_ref[0]**o+(0.459627185677)*x_ref[0]+(-0.511957702356)*x_ref[1]**o+(0.309573033639)*x_ref[1]+(0.806861185555)*x_ref[2]**o+(0.182177897377)*x_ref[2]
        ref[1,0,0]=(0.600850271814)*x_ref[0]**o+(0.440922206975)*x_ref[0]+(0.0307377466119)*x_ref[1]**o+(-0.419096719454)*x_ref[1]+(0.979284456332)*x_ref[2]**o+(-0.00240732351601)*x_ref[2]
        ref[1,0,1]=(0.51273517393)*x_ref[0]**o+(-0.947643736857)*x_ref[0]+(0.272403256871)*x_ref[1]**o+(0.0119584891405)*x_ref[1]+(-0.290604573286)*x_ref[2]**o+(0.822291254588)*x_ref[2]
        ref[1,1,0]=(-0.201909330605)*x_ref[0]**o+(-0.560221198146)*x_ref[0]+(0.564306911695)*x_ref[1]**o+(-0.936933580593)*x_ref[1]+(-0.41391834592)*x_ref[2]**o+(0.133244793725)*x_ref[2]
        ref[1,1,1]=(0.13318221571)*x_ref[0]**o+(-0.565278280388)*x_ref[0]+(0.677613822728)*x_ref[1]**o+(-0.386612668638)*x_ref[1]+(0.207570026134)*x_ref[2]**o+(0.798362457061)*x_ref[2]
        ref[2,0,0]=(-0.3777906845)*x_ref[0]**o+(-0.285455928514)*x_ref[0]+(-0.0901885959293)*x_ref[1]**o+(-0.385922214051)*x_ref[1]+(0.322566508003)*x_ref[2]**o+(0.171191301702)*x_ref[2]
        ref[2,0,1]=(-0.113954650916)*x_ref[0]**o+(-0.373563461907)*x_ref[0]+(0.73321196284)*x_ref[1]**o+(0.535150793249)*x_ref[1]+(-0.736944901672)*x_ref[2]**o+(0.86133553525)*x_ref[2]
        ref[2,1,0]=(-0.724228484035)*x_ref[0]**o+(0.310971627294)*x_ref[0]+(-0.497468717028)*x_ref[1]**o+(0.467315341957)*x_ref[1]+(-0.598941984951)*x_ref[2]**o+(0.990367400534)*x_ref[2]
        ref[2,1,1]=(0.594502107684)*x_ref[0]**o+(0.261477996581)*x_ref[0]+(0.738413631986)*x_ref[1]**o+(0.363468523272)*x_ref[1]+(-0.945845846052)*x_ref[2]**o+(0.281144164925)*x_ref[2]
        ref[3,0,0]=(0.804174935673)*x_ref[0]**o+(-0.633954348085)*x_ref[0]+(-0.953193492156)*x_ref[1]**o+(0.632681376482)*x_ref[1]+(-0.409900341362)*x_ref[2]**o+(0.951988363964)*x_ref[2]
        ref[3,0,1]=(-0.469895421121)*x_ref[0]**o+(0.666115304582)*x_ref[0]+(0.676351738247)*x_ref[1]**o+(-0.0738256360284)*x_ref[1]+(0.34976675523)*x_ref[2]**o+(-0.561058490613)*x_ref[2]
        ref[3,1,0]=(-0.999916478236)*x_ref[0]**o+(-0.362871459365)*x_ref[0]+(0.645604661917)*x_ref[1]**o+(0.215878330855)*x_ref[1]+(-0.0561173996823)*x_ref[2]**o+(0.272143302204)*x_ref[2]
        ref[3,1,1]=(0.233461828376)*x_ref[0]**o+(-0.110673705352)*x_ref[0]+(0.699710673197)*x_ref[1]**o+(-0.628805264645)*x_ref[1]+(0.677467640352)*x_ref[2]**o+(-0.156224522726)*x_ref[2]
        ref[4,0,0]=(0.180468931719)*x_ref[0]**o+(0.299451524372)*x_ref[0]+(-0.677131137371)*x_ref[1]**o+(-0.422802649252)*x_ref[1]+(0.237349200367)*x_ref[2]**o+(0.739758623757)*x_ref[2]
        ref[4,0,1]=(0.22209981155)*x_ref[0]**o+(-0.226751058136)*x_ref[0]+(-0.0345697170216)*x_ref[1]**o+(0.265795819906)*x_ref[1]+(0.542547943322)*x_ref[2]**o+(0.785610668157)*x_ref[2]
        ref[4,1,0]=(-0.520898604151)*x_ref[0]**o+(-0.84307105084)*x_ref[0]+(0.818061645146)*x_ref[1]**o+(-0.173710357379)*x_ref[1]+(0.344023171063)*x_ref[2]**o+(0.194662166094)*x_ref[2]
        ref[4,1,1]=(-0.288466404824)*x_ref[0]**o+(-0.891856587908)*x_ref[0]+(0.378329571068)*x_ref[1]**o+(0.437919935526)*x_ref[1]+(-0.630733422962)*x_ref[2]**o+(0.518444630882)*x_ref[2]
        ref[5,0,0]=(0.00978685202942)*x_ref[0]**o+(-0.474443494209)*x_ref[0]+(-0.509964215112)*x_ref[1]**o+(0.341199671453)*x_ref[1]+(-0.459333887251)*x_ref[2]**o+(0.491961590667)*x_ref[2]
        ref[5,0,1]=(-0.899850325159)*x_ref[0]**o+(-0.191369137667)*x_ref[0]+(0.18882430985)*x_ref[1]**o+(-0.185214597442)*x_ref[1]+(-0.967187797553)*x_ref[2]**o+(0.366127839986)*x_ref[2]
        ref[5,1,0]=(0.428007441668)*x_ref[0]**o+(0.920828656579)*x_ref[0]+(-0.298863178441)*x_ref[1]**o+(-0.226020983335)*x_ref[1]+(-0.01907141259)*x_ref[2]**o+(-0.359698040557)*x_ref[2]
        ref[5,1,1]=(0.980673633441)*x_ref[0]**o+(0.362357489872)*x_ref[0]+(0.825233899658)*x_ref[1]**o+(0.846433827949)*x_ref[1]+(-0.870197174533)*x_ref[2]**o+(0.621884204716)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.93572957134)*x[0]**o+(-0.878081515023)*x[0]+(0.821839395118)*x[1]**o+(-0.088026044081)*x[1]
        arg[0,0,0,1]=(-0.00393438000839)*x[0]**o+(-0.180693225971)*x[0]+(0.192781914487)*x[1]**o+(-0.561366977374)*x[1]
        arg[0,0,1,0]=(0.116028558326)*x[0]**o+(-0.47739282219)*x[0]+(0.97806148353)*x[1]**o+(0.660972051388)*x[1]
        arg[0,0,1,1]=(0.092683552469)*x[0]**o+(0.22083599745)*x[0]+(-0.423073290439)*x[1]**o+(-0.0789522599045)*x[1]
        arg[0,0,2,0]=(0.140023403084)*x[0]**o+(-0.927034796519)*x[0]+(0.0987957525564)*x[1]**o+(0.0703695028562)*x[1]
        arg[0,0,2,1]=(-0.00485147064392)*x[0]**o+(-0.649679595932)*x[0]+(0.503435626816)*x[1]**o+(0.393293792881)*x[1]
        arg[0,1,0,0]=(-0.753383897772)*x[0]**o+(-0.988499439357)*x[0]+(-0.651037417175)*x[1]**o+(0.45184892046)*x[1]
        arg[0,1,0,1]=(0.0120892569605)*x[0]**o+(0.52287896209)*x[0]+(0.103256695817)*x[1]**o+(-0.0244835681092)*x[1]
        arg[0,1,1,0]=(-0.0949239075135)*x[0]**o+(0.835062318231)*x[0]+(-0.0522981103769)*x[1]**o+(0.60020994838)*x[1]
        arg[0,1,1,1]=(0.139896941482)*x[0]**o+(0.409042945999)*x[0]+(0.10019215394)*x[1]**o+(-0.312224123017)*x[1]
        arg[0,1,2,0]=(-0.892437964342)*x[0]**o+(0.0828807279746)*x[0]+(0.777811646235)*x[1]**o+(-0.96509471481)*x[1]
        arg[0,1,2,1]=(-0.397341226148)*x[0]**o+(0.102987581336)*x[0]+(-0.69165111598)*x[1]**o+(0.743298211061)*x[1]
        arg[0,2,0,0]=(-0.300129864031)*x[0]**o+(-0.418981381232)*x[0]+(0.624854922555)*x[1]**o+(0.892651746561)*x[1]
        arg[0,2,0,1]=(-0.511463603881)*x[0]**o+(-0.808468332238)*x[0]+(-0.342755248919)*x[1]**o+(0.357985875808)*x[1]
        arg[0,2,1,0]=(-0.269520452932)*x[0]**o+(0.869553937081)*x[0]+(-0.183222122614)*x[1]**o+(-0.135280721586)*x[1]
        arg[0,2,1,1]=(0.0396712123969)*x[0]**o+(0.344314072793)*x[0]+(0.644676507168)*x[1]**o+(-0.0799551435053)*x[1]
        arg[0,2,2,0]=(0.302628743334)*x[0]**o+(-0.59557047376)*x[0]+(-0.277654790212)*x[1]**o+(0.0299035042863)*x[1]
        arg[0,2,2,1]=(0.805296568248)*x[0]**o+(0.726404711154)*x[0]+(0.0728356962577)*x[1]**o+(0.199997237962)*x[1]
        arg[0,3,0,0]=(-0.292658800952)*x[0]**o+(-0.40080106651)*x[0]+(0.844109590553)*x[1]**o+(-0.339364790362)*x[1]
        arg[0,3,0,1]=(-0.621465058894)*x[0]**o+(-0.376804350545)*x[0]+(0.745038406103)*x[1]**o+(-0.281482724757)*x[1]
        arg[0,3,1,0]=(0.00625816347702)*x[0]**o+(0.392750189511)*x[0]+(-0.232258708605)*x[1]**o+(-0.634552067635)*x[1]
        arg[0,3,1,1]=(-0.0893109545391)*x[0]**o+(-0.735678058768)*x[0]+(-0.603248350212)*x[1]**o+(-0.664366286468)*x[1]
        arg[0,3,2,0]=(-0.0539114688775)*x[0]**o+(-0.110100592496)*x[0]+(-0.700496078261)*x[1]**o+(0.203311460549)*x[1]
        arg[0,3,2,1]=(-0.554691350426)*x[0]**o+(-0.255001039659)*x[0]+(0.567977217289)*x[1]**o+(0.61993007447)*x[1]
        arg[0,4,0,0]=(-0.12193490862)*x[0]**o+(0.731139340661)*x[0]+(-0.321971724057)*x[1]**o+(0.741215588897)*x[1]
        arg[0,4,0,1]=(-0.394743164913)*x[0]**o+(-0.602249917929)*x[0]+(0.27675734517)*x[1]**o+(-0.968806169573)*x[1]
        arg[0,4,1,0]=(-0.629331232169)*x[0]**o+(0.687483125805)*x[0]+(0.633698693186)*x[1]**o+(0.837564332901)*x[1]
        arg[0,4,1,1]=(0.0470601891829)*x[0]**o+(-0.365490813497)*x[0]+(-0.162532786979)*x[1]**o+(0.504146804122)*x[1]
        arg[0,4,2,0]=(0.193790444091)*x[0]**o+(0.742132840869)*x[0]+(-0.871677496589)*x[1]**o+(0.492817750307)*x[1]
        arg[0,4,2,1]=(0.741218732478)*x[0]**o+(-0.746973604492)*x[0]+(-0.548073707926)*x[1]**o+(-0.42258185128)*x[1]
        arg[1,0,0,0]=(0.580989627459)*x[0]**o+(-0.498423525574)*x[0]+(-0.349651580594)*x[1]**o+(0.0193060779482)*x[1]
        arg[1,0,0,1]=(0.398917011903)*x[0]**o+(-0.122064497238)*x[0]+(0.924654557299)*x[1]**o+(0.485284066068)*x[1]
        arg[1,0,1,0]=(-0.985949382826)*x[0]**o+(-0.0188384874977)*x[0]+(0.177995429697)*x[1]**o+(0.967727833143)*x[1]
        arg[1,0,1,1]=(0.920799980503)*x[0]**o+(-0.749281119074)*x[0]+(-0.628411322607)*x[1]**o+(-0.446396361734)*x[1]
        arg[1,0,2,0]=(0.752863123667)*x[0]**o+(0.112427175329)*x[0]+(0.975568251269)*x[1]**o+(0.316938392669)*x[1]
        arg[1,0,2,1]=(0.798409861933)*x[0]**o+(-0.145715624761)*x[0]+(-0.847813831553)*x[1]**o+(0.302999756348)*x[1]
        arg[1,1,0,0]=(0.897029838502)*x[0]**o+(-0.540852276529)*x[0]+(-0.541450963121)*x[1]**o+(-0.118819043917)*x[1]
        arg[1,1,0,1]=(-0.616797542123)*x[0]**o+(-0.139394144428)*x[0]+(-0.480348413917)*x[1]**o+(0.500244829288)*x[1]
        arg[1,1,1,0]=(-0.960388730531)*x[0]**o+(0.87391147728)*x[0]+(0.0128673975991)*x[1]**o+(0.444341296582)*x[1]
        arg[1,1,1,1]=(0.136810869097)*x[0]**o+(-0.544756866447)*x[0]+(-0.198052789907)*x[1]**o+(-0.536732583929)*x[1]
        arg[1,1,2,0]=(-0.808981135979)*x[0]**o+(-0.122088197438)*x[0]+(-0.363958821187)*x[1]**o+(-0.521614066151)*x[1]
        arg[1,1,2,1]=(0.299382605072)*x[0]**o+(0.143397060894)*x[0]+(-0.223044874711)*x[1]**o+(-0.04723866032)*x[1]
        arg[1,2,0,0]=(-0.568967551455)*x[0]**o+(0.750777908696)*x[0]+(0.203117458853)*x[1]**o+(-0.0797168995572)*x[1]
        arg[1,2,0,1]=(0.0439147552635)*x[0]**o+(-0.0746749889653)*x[0]+(0.192885940929)*x[1]**o+(0.794090802345)*x[1]
        arg[1,2,1,0]=(0.137177876023)*x[0]**o+(0.625847668003)*x[0]+(0.252761858157)*x[1]**o+(0.496512978117)*x[1]
        arg[1,2,1,1]=(0.475218888846)*x[0]**o+(-0.286926184073)*x[0]+(0.551427803487)*x[1]**o+(0.782373661286)*x[1]
        arg[1,2,2,0]=(0.903004502545)*x[0]**o+(-0.262952251654)*x[0]+(0.110351860812)*x[1]**o+(-0.685355934696)*x[1]
        arg[1,2,2,1]=(-0.157457317404)*x[0]**o+(-0.243850623177)*x[0]+(0.28853283432)*x[1]**o+(0.612508908614)*x[1]
        arg[1,3,0,0]=(-0.741890002973)*x[0]**o+(-0.118262424825)*x[0]+(0.567100734646)*x[1]**o+(0.301557125811)*x[1]
        arg[1,3,0,1]=(-0.00745396583928)*x[0]**o+(-0.00558491344276)*x[0]+(-0.967132370197)*x[1]**o+(-0.61737594402)*x[1]
        arg[1,3,1,0]=(0.0767961523966)*x[0]**o+(0.341952000996)*x[0]+(0.762617832019)*x[1]**o+(-0.38337284713)*x[1]
        arg[1,3,1,1]=(0.32265018902)*x[0]**o+(-0.352521736206)*x[0]+(-0.0119143186529)*x[1]**o+(-0.64727561829)*x[1]
        arg[1,3,2,0]=(0.627690584587)*x[0]**o+(0.0993819822193)*x[0]+(-0.189363448841)*x[1]**o+(0.49004001194)*x[1]
        arg[1,3,2,1]=(-0.657836476156)*x[0]**o+(0.708970470136)*x[0]+(0.252190020576)*x[1]**o+(-0.481214196298)*x[1]
        arg[1,4,0,0]=(0.0837563164179)*x[0]**o+(0.741536525079)*x[0]+(0.645543094286)*x[1]**o+(0.62358090815)*x[1]
        arg[1,4,0,1]=(0.262043398371)*x[0]**o+(-0.47794750936)*x[0]+(0.93086244965)*x[1]**o+(-0.670399062677)*x[1]
        arg[1,4,1,0]=(-0.76634902923)*x[0]**o+(-0.809846143928)*x[0]+(0.639099549305)*x[1]**o+(0.793524639236)*x[1]
        arg[1,4,1,1]=(0.814737503957)*x[0]**o+(0.775593374042)*x[0]+(0.240976904111)*x[1]**o+(0.2352421356)*x[1]
        arg[1,4,2,0]=(0.519210584047)*x[0]**o+(-0.702579589691)*x[0]+(-0.450156317853)*x[1]**o+(0.588864692581)*x[1]
        arg[1,4,2,1]=(-0.927751546037)*x[0]**o+(-0.239737576019)*x[0]+(-0.527658440447)*x[1]**o+(0.79623440559)*x[1]
        arg[2,0,0,0]=(0.719144336419)*x[0]**o+(0.916780231277)*x[0]+(0.279073450277)*x[1]**o+(-0.0130671321167)*x[1]
        arg[2,0,0,1]=(-0.556604998962)*x[0]**o+(-0.661580370137)*x[0]+(0.980704664791)*x[1]**o+(0.935843251131)*x[1]
        arg[2,0,1,0]=(0.85501679753)*x[0]**o+(-0.908342467889)*x[0]+(0.969030151294)*x[1]**o+(0.534192961922)*x[1]
        arg[2,0,1,1]=(0.527672385529)*x[0]**o+(0.997699955667)*x[0]+(-0.388249629327)*x[1]**o+(-0.714956937512)*x[1]
        arg[2,0,2,0]=(-0.629149810667)*x[0]**o+(-0.0515018376692)*x[0]+(-0.80355991702)*x[1]**o+(0.235999988886)*x[1]
        arg[2,0,2,1]=(0.960715934156)*x[0]**o+(-0.782787305497)*x[0]+(-0.309348231959)*x[1]**o+(0.854911294717)*x[1]
        arg[2,1,0,0]=(-0.560749125074)*x[0]**o+(-0.120515615757)*x[0]+(0.055090634431)*x[1]**o+(-0.539205949358)*x[1]
        arg[2,1,0,1]=(-0.762113417295)*x[0]**o+(-0.207196980376)*x[0]+(0.904541245494)*x[1]**o+(0.320359459939)*x[1]
        arg[2,1,1,0]=(-0.272627586486)*x[0]**o+(-0.885987982522)*x[0]+(0.641737808216)*x[1]**o+(0.358390367373)*x[1]
        arg[2,1,1,1]=(-0.847502563761)*x[0]**o+(-0.430298277492)*x[0]+(-0.869043300185)*x[1]**o+(0.584700145915)*x[1]
        arg[2,1,2,0]=(0.497688368946)*x[0]**o+(-0.14413770704)*x[0]+(-0.82002941888)*x[1]**o+(0.573569419279)*x[1]
        arg[2,1,2,1]=(0.959549623603)*x[0]**o+(0.881196513588)*x[0]+(-0.202951096266)*x[1]**o+(-9.87983182765e-05)*x[1]
        arg[2,2,0,0]=(0.0811393810501)*x[0]**o+(-0.853555165896)*x[0]+(0.0259257246335)*x[1]**o+(0.638575306959)*x[1]
        arg[2,2,0,1]=(-0.525990031379)*x[0]**o+(0.11235160765)*x[0]+(-0.502853879455)*x[1]**o+(0.414066683507)*x[1]
        arg[2,2,1,0]=(-0.533919716834)*x[0]**o+(-0.773411100656)*x[0]+(0.311424028491)*x[1]**o+(0.332733301377)*x[1]
        arg[2,2,1,1]=(0.280548593462)*x[0]**o+(-0.0747640762104)*x[0]+(-0.7110232877)*x[1]**o+(-0.178672817441)*x[1]
        arg[2,2,2,0]=(0.445025926846)*x[0]**o+(-0.669782903005)*x[0]+(0.236610411793)*x[1]**o+(0.437836936476)*x[1]
        arg[2,2,2,1]=(-0.503761305098)*x[0]**o+(-0.889922659865)*x[0]+(-0.0767256016402)*x[1]**o+(-0.614874217835)*x[1]
        arg[2,3,0,0]=(-0.041960855672)*x[0]**o+(-0.593486242764)*x[0]+(0.344751892525)*x[1]**o+(-0.131268873786)*x[1]
        arg[2,3,0,1]=(-0.192704843083)*x[0]**o+(0.325356381696)*x[0]+(-0.235690651937)*x[1]**o+(0.59113535617)*x[1]
        arg[2,3,1,0]=(-0.0794745628024)*x[0]**o+(0.7996737586)*x[0]+(-0.324970505375)*x[1]**o+(-0.14920403468)*x[1]
        arg[2,3,1,1]=(0.31997432673)*x[0]**o+(-0.571739807799)*x[0]+(0.517683480263)*x[1]**o+(-0.250680741704)*x[1]
        arg[2,3,2,0]=(-0.379114277883)*x[0]**o+(0.953285017352)*x[0]+(-0.701111847548)*x[1]**o+(-0.354199942528)*x[1]
        arg[2,3,2,1]=(0.492245629486)*x[0]**o+(-0.299380102293)*x[0]+(-0.126371826833)*x[1]**o+(0.933664060647)*x[1]
        arg[2,4,0,0]=(0.714822159533)*x[0]**o+(0.359430380478)*x[0]+(-0.691610768506)*x[1]**o+(0.414061853203)*x[1]
        arg[2,4,0,1]=(0.412799253275)*x[0]**o+(0.862905533148)*x[0]+(-0.0507440946561)*x[1]**o+(0.779012369976)*x[1]
        arg[2,4,1,0]=(-0.981179053034)*x[0]**o+(0.984777005573)*x[0]+(0.715584788452)*x[1]**o+(0.920666888724)*x[1]
        arg[2,4,1,1]=(-0.0287775685032)*x[0]**o+(0.617592345344)*x[0]+(-0.767451297794)*x[1]**o+(0.95828976091)*x[1]
        arg[2,4,2,0]=(0.905037473705)*x[0]**o+(-0.706987158021)*x[0]+(0.535436652493)*x[1]**o+(-0.891712664249)*x[1]
        arg[2,4,2,1]=(-0.818709474925)*x[0]**o+(-0.0223479273553)*x[0]+(-0.370644568884)*x[1]**o+(0.286468739105)*x[1]
        arg[3,0,0,0]=(0.310461949876)*x[0]**o+(0.770193446516)*x[0]+(-0.0858335164699)*x[1]**o+(0.12572608122)*x[1]
        arg[3,0,0,1]=(-0.781332422249)*x[0]**o+(-0.238368527872)*x[0]+(-0.449272256582)*x[1]**o+(-0.645638738212)*x[1]
        arg[3,0,1,0]=(-0.91143447678)*x[0]**o+(-0.885131081008)*x[0]+(-0.253833463581)*x[1]**o+(0.051439649459)*x[1]
        arg[3,0,1,1]=(0.447730169203)*x[0]**o+(0.980354452325)*x[0]+(-0.0484566141047)*x[1]**o+(0.443407604362)*x[1]
        arg[3,0,2,0]=(-0.223975544426)*x[0]**o+(0.0141796274312)*x[0]+(0.664080206442)*x[1]**o+(0.193318634886)*x[1]
        arg[3,0,2,1]=(-0.388901068755)*x[0]**o+(-0.632210208163)*x[0]+(0.567264328931)*x[1]**o+(-0.308710733281)*x[1]
        arg[3,1,0,0]=(0.259562910069)*x[0]**o+(-0.315409720689)*x[0]+(0.999281841483)*x[1]**o+(0.695981779357)*x[1]
        arg[3,1,0,1]=(0.627508253716)*x[0]**o+(-0.540950954632)*x[0]+(0.626209667445)*x[1]**o+(-0.920403595109)*x[1]
        arg[3,1,1,0]=(0.750166528891)*x[0]**o+(-0.145494783124)*x[0]+(0.71048043416)*x[1]**o+(-0.0920505950401)*x[1]
        arg[3,1,1,1]=(-0.755570630755)*x[0]**o+(-0.323456877629)*x[0]+(-0.0318418288689)*x[1]**o+(0.551823313307)*x[1]
        arg[3,1,2,0]=(-0.301380276426)*x[0]**o+(-0.102899117196)*x[0]+(0.816857028902)*x[1]**o+(-0.571555621895)*x[1]
        arg[3,1,2,1]=(-0.791148729922)*x[0]**o+(0.400286696027)*x[0]+(-0.17362002139)*x[1]**o+(0.431696064424)*x[1]
        arg[3,2,0,0]=(0.241944451237)*x[0]**o+(-0.629401496412)*x[0]+(-0.00333636607647)*x[1]**o+(0.167334133677)*x[1]
        arg[3,2,0,1]=(-0.300944831932)*x[0]**o+(0.358282773504)*x[0]+(0.759495661081)*x[1]**o+(-0.225250488484)*x[1]
        arg[3,2,1,0]=(0.755950588768)*x[0]**o+(0.718423180013)*x[0]+(0.705150216505)*x[1]**o+(0.269283969196)*x[1]
        arg[3,2,1,1]=(0.811064150458)*x[0]**o+(-0.67847842696)*x[0]+(-0.478091805711)*x[1]**o+(0.292774429011)*x[1]
        arg[3,2,2,0]=(-0.307877253758)*x[0]**o+(-0.563961673764)*x[0]+(0.428322493356)*x[1]**o+(0.564230126489)*x[1]
        arg[3,2,2,1]=(-0.623877056979)*x[0]**o+(-0.148337578412)*x[0]+(-0.910662605581)*x[1]**o+(-0.38603493572)*x[1]
        arg[3,3,0,0]=(-0.826474270729)*x[0]**o+(0.977614689932)*x[0]+(-0.369824837717)*x[1]**o+(-0.985478011812)*x[1]
        arg[3,3,0,1]=(-0.251252769561)*x[0]**o+(-0.296732765057)*x[0]+(-0.888899418995)*x[1]**o+(0.640313055734)*x[1]
        arg[3,3,1,0]=(-0.189106218841)*x[0]**o+(-0.0441656339777)*x[0]+(0.368716219801)*x[1]**o+(-0.924144919367)*x[1]
        arg[3,3,1,1]=(-0.525205981948)*x[0]**o+(-0.148352446593)*x[0]+(0.869212821652)*x[1]**o+(-0.575463559793)*x[1]
        arg[3,3,2,0]=(0.668772663439)*x[0]**o+(0.4330630456)*x[0]+(-0.321158907274)*x[1]**o+(0.570853139693)*x[1]
        arg[3,3,2,1]=(0.0236674031998)*x[0]**o+(-0.390106024651)*x[0]+(0.541717893494)*x[1]**o+(-0.144655705849)*x[1]
        arg[3,4,0,0]=(0.971861188602)*x[0]**o+(-0.834463063474)*x[0]+(-0.419545662218)*x[1]**o+(-0.0387341522049)*x[1]
        arg[3,4,0,1]=(0.226357589896)*x[0]**o+(0.305148895256)*x[0]+(0.00306527642431)*x[1]**o+(0.4715619746)*x[1]
        arg[3,4,1,0]=(0.529602340964)*x[0]**o+(-0.482050910795)*x[0]+(0.964983911769)*x[1]**o+(0.524018220532)*x[1]
        arg[3,4,1,1]=(0.940676158872)*x[0]**o+(0.547078925647)*x[0]+(-0.636907424453)*x[1]**o+(0.938262396281)*x[1]
        arg[3,4,2,0]=(0.282512784255)*x[0]**o+(-0.556345478931)*x[0]+(0.651866614458)*x[1]**o+(0.8895132214)*x[1]
        arg[3,4,2,1]=(0.905077434472)*x[0]**o+(0.595878819874)*x[0]+(0.416350660905)*x[1]**o+(0.100262997545)*x[1]
        ref[0,0,0,0]=(-0.93572957134)*x_ref[0]**o+(-0.878081515023)*x_ref[0]+(0.821839395118)*x_ref[1]**o+(-0.088026044081)*x_ref[1]
        ref[0,0,0,1]=(-0.00393438000839)*x_ref[0]**o+(-0.180693225971)*x_ref[0]+(0.192781914487)*x_ref[1]**o+(-0.561366977374)*x_ref[1]
        ref[0,0,1,0]=(0.116028558326)*x_ref[0]**o+(-0.47739282219)*x_ref[0]+(0.97806148353)*x_ref[1]**o+(0.660972051388)*x_ref[1]
        ref[0,0,1,1]=(0.092683552469)*x_ref[0]**o+(0.22083599745)*x_ref[0]+(-0.423073290439)*x_ref[1]**o+(-0.0789522599045)*x_ref[1]
        ref[0,0,2,0]=(0.140023403084)*x_ref[0]**o+(-0.927034796519)*x_ref[0]+(0.0987957525564)*x_ref[1]**o+(0.0703695028562)*x_ref[1]
        ref[0,0,2,1]=(-0.00485147064392)*x_ref[0]**o+(-0.649679595932)*x_ref[0]+(0.503435626816)*x_ref[1]**o+(0.393293792881)*x_ref[1]
        ref[0,1,0,0]=(-0.753383897772)*x_ref[0]**o+(-0.988499439357)*x_ref[0]+(-0.651037417175)*x_ref[1]**o+(0.45184892046)*x_ref[1]
        ref[0,1,0,1]=(0.0120892569605)*x_ref[0]**o+(0.52287896209)*x_ref[0]+(0.103256695817)*x_ref[1]**o+(-0.0244835681092)*x_ref[1]
        ref[0,1,1,0]=(-0.0949239075135)*x_ref[0]**o+(0.835062318231)*x_ref[0]+(-0.0522981103769)*x_ref[1]**o+(0.60020994838)*x_ref[1]
        ref[0,1,1,1]=(0.139896941482)*x_ref[0]**o+(0.409042945999)*x_ref[0]+(0.10019215394)*x_ref[1]**o+(-0.312224123017)*x_ref[1]
        ref[0,1,2,0]=(-0.892437964342)*x_ref[0]**o+(0.0828807279746)*x_ref[0]+(0.777811646235)*x_ref[1]**o+(-0.96509471481)*x_ref[1]
        ref[0,1,2,1]=(-0.397341226148)*x_ref[0]**o+(0.102987581336)*x_ref[0]+(-0.69165111598)*x_ref[1]**o+(0.743298211061)*x_ref[1]
        ref[0,2,0,0]=(-0.300129864031)*x_ref[0]**o+(-0.418981381232)*x_ref[0]+(0.624854922555)*x_ref[1]**o+(0.892651746561)*x_ref[1]
        ref[0,2,0,1]=(-0.511463603881)*x_ref[0]**o+(-0.808468332238)*x_ref[0]+(-0.342755248919)*x_ref[1]**o+(0.357985875808)*x_ref[1]
        ref[0,2,1,0]=(-0.269520452932)*x_ref[0]**o+(0.869553937081)*x_ref[0]+(-0.183222122614)*x_ref[1]**o+(-0.135280721586)*x_ref[1]
        ref[0,2,1,1]=(0.0396712123969)*x_ref[0]**o+(0.344314072793)*x_ref[0]+(0.644676507168)*x_ref[1]**o+(-0.0799551435053)*x_ref[1]
        ref[0,2,2,0]=(0.302628743334)*x_ref[0]**o+(-0.59557047376)*x_ref[0]+(-0.277654790212)*x_ref[1]**o+(0.0299035042863)*x_ref[1]
        ref[0,2,2,1]=(0.805296568248)*x_ref[0]**o+(0.726404711154)*x_ref[0]+(0.0728356962577)*x_ref[1]**o+(0.199997237962)*x_ref[1]
        ref[0,3,0,0]=(-0.292658800952)*x_ref[0]**o+(-0.40080106651)*x_ref[0]+(0.844109590553)*x_ref[1]**o+(-0.339364790362)*x_ref[1]
        ref[0,3,0,1]=(-0.621465058894)*x_ref[0]**o+(-0.376804350545)*x_ref[0]+(0.745038406103)*x_ref[1]**o+(-0.281482724757)*x_ref[1]
        ref[0,3,1,0]=(0.00625816347702)*x_ref[0]**o+(0.392750189511)*x_ref[0]+(-0.232258708605)*x_ref[1]**o+(-0.634552067635)*x_ref[1]
        ref[0,3,1,1]=(-0.0893109545391)*x_ref[0]**o+(-0.735678058768)*x_ref[0]+(-0.603248350212)*x_ref[1]**o+(-0.664366286468)*x_ref[1]
        ref[0,3,2,0]=(-0.0539114688775)*x_ref[0]**o+(-0.110100592496)*x_ref[0]+(-0.700496078261)*x_ref[1]**o+(0.203311460549)*x_ref[1]
        ref[0,3,2,1]=(-0.554691350426)*x_ref[0]**o+(-0.255001039659)*x_ref[0]+(0.567977217289)*x_ref[1]**o+(0.61993007447)*x_ref[1]
        ref[0,4,0,0]=(-0.12193490862)*x_ref[0]**o+(0.731139340661)*x_ref[0]+(-0.321971724057)*x_ref[1]**o+(0.741215588897)*x_ref[1]
        ref[0,4,0,1]=(-0.394743164913)*x_ref[0]**o+(-0.602249917929)*x_ref[0]+(0.27675734517)*x_ref[1]**o+(-0.968806169573)*x_ref[1]
        ref[0,4,1,0]=(-0.629331232169)*x_ref[0]**o+(0.687483125805)*x_ref[0]+(0.633698693186)*x_ref[1]**o+(0.837564332901)*x_ref[1]
        ref[0,4,1,1]=(0.0470601891829)*x_ref[0]**o+(-0.365490813497)*x_ref[0]+(-0.162532786979)*x_ref[1]**o+(0.504146804122)*x_ref[1]
        ref[0,4,2,0]=(0.193790444091)*x_ref[0]**o+(0.742132840869)*x_ref[0]+(-0.871677496589)*x_ref[1]**o+(0.492817750307)*x_ref[1]
        ref[0,4,2,1]=(0.741218732478)*x_ref[0]**o+(-0.746973604492)*x_ref[0]+(-0.548073707926)*x_ref[1]**o+(-0.42258185128)*x_ref[1]
        ref[1,0,0,0]=(0.580989627459)*x_ref[0]**o+(-0.498423525574)*x_ref[0]+(-0.349651580594)*x_ref[1]**o+(0.0193060779482)*x_ref[1]
        ref[1,0,0,1]=(0.398917011903)*x_ref[0]**o+(-0.122064497238)*x_ref[0]+(0.924654557299)*x_ref[1]**o+(0.485284066068)*x_ref[1]
        ref[1,0,1,0]=(-0.985949382826)*x_ref[0]**o+(-0.0188384874977)*x_ref[0]+(0.177995429697)*x_ref[1]**o+(0.967727833143)*x_ref[1]
        ref[1,0,1,1]=(0.920799980503)*x_ref[0]**o+(-0.749281119074)*x_ref[0]+(-0.628411322607)*x_ref[1]**o+(-0.446396361734)*x_ref[1]
        ref[1,0,2,0]=(0.752863123667)*x_ref[0]**o+(0.112427175329)*x_ref[0]+(0.975568251269)*x_ref[1]**o+(0.316938392669)*x_ref[1]
        ref[1,0,2,1]=(0.798409861933)*x_ref[0]**o+(-0.145715624761)*x_ref[0]+(-0.847813831553)*x_ref[1]**o+(0.302999756348)*x_ref[1]
        ref[1,1,0,0]=(0.897029838502)*x_ref[0]**o+(-0.540852276529)*x_ref[0]+(-0.541450963121)*x_ref[1]**o+(-0.118819043917)*x_ref[1]
        ref[1,1,0,1]=(-0.616797542123)*x_ref[0]**o+(-0.139394144428)*x_ref[0]+(-0.480348413917)*x_ref[1]**o+(0.500244829288)*x_ref[1]
        ref[1,1,1,0]=(-0.960388730531)*x_ref[0]**o+(0.87391147728)*x_ref[0]+(0.0128673975991)*x_ref[1]**o+(0.444341296582)*x_ref[1]
        ref[1,1,1,1]=(0.136810869097)*x_ref[0]**o+(-0.544756866447)*x_ref[0]+(-0.198052789907)*x_ref[1]**o+(-0.536732583929)*x_ref[1]
        ref[1,1,2,0]=(-0.808981135979)*x_ref[0]**o+(-0.122088197438)*x_ref[0]+(-0.363958821187)*x_ref[1]**o+(-0.521614066151)*x_ref[1]
        ref[1,1,2,1]=(0.299382605072)*x_ref[0]**o+(0.143397060894)*x_ref[0]+(-0.223044874711)*x_ref[1]**o+(-0.04723866032)*x_ref[1]
        ref[1,2,0,0]=(-0.568967551455)*x_ref[0]**o+(0.750777908696)*x_ref[0]+(0.203117458853)*x_ref[1]**o+(-0.0797168995572)*x_ref[1]
        ref[1,2,0,1]=(0.0439147552635)*x_ref[0]**o+(-0.0746749889653)*x_ref[0]+(0.192885940929)*x_ref[1]**o+(0.794090802345)*x_ref[1]
        ref[1,2,1,0]=(0.137177876023)*x_ref[0]**o+(0.625847668003)*x_ref[0]+(0.252761858157)*x_ref[1]**o+(0.496512978117)*x_ref[1]
        ref[1,2,1,1]=(0.475218888846)*x_ref[0]**o+(-0.286926184073)*x_ref[0]+(0.551427803487)*x_ref[1]**o+(0.782373661286)*x_ref[1]
        ref[1,2,2,0]=(0.903004502545)*x_ref[0]**o+(-0.262952251654)*x_ref[0]+(0.110351860812)*x_ref[1]**o+(-0.685355934696)*x_ref[1]
        ref[1,2,2,1]=(-0.157457317404)*x_ref[0]**o+(-0.243850623177)*x_ref[0]+(0.28853283432)*x_ref[1]**o+(0.612508908614)*x_ref[1]
        ref[1,3,0,0]=(-0.741890002973)*x_ref[0]**o+(-0.118262424825)*x_ref[0]+(0.567100734646)*x_ref[1]**o+(0.301557125811)*x_ref[1]
        ref[1,3,0,1]=(-0.00745396583928)*x_ref[0]**o+(-0.00558491344276)*x_ref[0]+(-0.967132370197)*x_ref[1]**o+(-0.61737594402)*x_ref[1]
        ref[1,3,1,0]=(0.0767961523966)*x_ref[0]**o+(0.341952000996)*x_ref[0]+(0.762617832019)*x_ref[1]**o+(-0.38337284713)*x_ref[1]
        ref[1,3,1,1]=(0.32265018902)*x_ref[0]**o+(-0.352521736206)*x_ref[0]+(-0.0119143186529)*x_ref[1]**o+(-0.64727561829)*x_ref[1]
        ref[1,3,2,0]=(0.627690584587)*x_ref[0]**o+(0.0993819822193)*x_ref[0]+(-0.189363448841)*x_ref[1]**o+(0.49004001194)*x_ref[1]
        ref[1,3,2,1]=(-0.657836476156)*x_ref[0]**o+(0.708970470136)*x_ref[0]+(0.252190020576)*x_ref[1]**o+(-0.481214196298)*x_ref[1]
        ref[1,4,0,0]=(0.0837563164179)*x_ref[0]**o+(0.741536525079)*x_ref[0]+(0.645543094286)*x_ref[1]**o+(0.62358090815)*x_ref[1]
        ref[1,4,0,1]=(0.262043398371)*x_ref[0]**o+(-0.47794750936)*x_ref[0]+(0.93086244965)*x_ref[1]**o+(-0.670399062677)*x_ref[1]
        ref[1,4,1,0]=(-0.76634902923)*x_ref[0]**o+(-0.809846143928)*x_ref[0]+(0.639099549305)*x_ref[1]**o+(0.793524639236)*x_ref[1]
        ref[1,4,1,1]=(0.814737503957)*x_ref[0]**o+(0.775593374042)*x_ref[0]+(0.240976904111)*x_ref[1]**o+(0.2352421356)*x_ref[1]
        ref[1,4,2,0]=(0.519210584047)*x_ref[0]**o+(-0.702579589691)*x_ref[0]+(-0.450156317853)*x_ref[1]**o+(0.588864692581)*x_ref[1]
        ref[1,4,2,1]=(-0.927751546037)*x_ref[0]**o+(-0.239737576019)*x_ref[0]+(-0.527658440447)*x_ref[1]**o+(0.79623440559)*x_ref[1]
        ref[2,0,0,0]=(0.719144336419)*x_ref[0]**o+(0.916780231277)*x_ref[0]+(0.279073450277)*x_ref[1]**o+(-0.0130671321167)*x_ref[1]
        ref[2,0,0,1]=(-0.556604998962)*x_ref[0]**o+(-0.661580370137)*x_ref[0]+(0.980704664791)*x_ref[1]**o+(0.935843251131)*x_ref[1]
        ref[2,0,1,0]=(0.85501679753)*x_ref[0]**o+(-0.908342467889)*x_ref[0]+(0.969030151294)*x_ref[1]**o+(0.534192961922)*x_ref[1]
        ref[2,0,1,1]=(0.527672385529)*x_ref[0]**o+(0.997699955667)*x_ref[0]+(-0.388249629327)*x_ref[1]**o+(-0.714956937512)*x_ref[1]
        ref[2,0,2,0]=(-0.629149810667)*x_ref[0]**o+(-0.0515018376692)*x_ref[0]+(-0.80355991702)*x_ref[1]**o+(0.235999988886)*x_ref[1]
        ref[2,0,2,1]=(0.960715934156)*x_ref[0]**o+(-0.782787305497)*x_ref[0]+(-0.309348231959)*x_ref[1]**o+(0.854911294717)*x_ref[1]
        ref[2,1,0,0]=(-0.560749125074)*x_ref[0]**o+(-0.120515615757)*x_ref[0]+(0.055090634431)*x_ref[1]**o+(-0.539205949358)*x_ref[1]
        ref[2,1,0,1]=(-0.762113417295)*x_ref[0]**o+(-0.207196980376)*x_ref[0]+(0.904541245494)*x_ref[1]**o+(0.320359459939)*x_ref[1]
        ref[2,1,1,0]=(-0.272627586486)*x_ref[0]**o+(-0.885987982522)*x_ref[0]+(0.641737808216)*x_ref[1]**o+(0.358390367373)*x_ref[1]
        ref[2,1,1,1]=(-0.847502563761)*x_ref[0]**o+(-0.430298277492)*x_ref[0]+(-0.869043300185)*x_ref[1]**o+(0.584700145915)*x_ref[1]
        ref[2,1,2,0]=(0.497688368946)*x_ref[0]**o+(-0.14413770704)*x_ref[0]+(-0.82002941888)*x_ref[1]**o+(0.573569419279)*x_ref[1]
        ref[2,1,2,1]=(0.959549623603)*x_ref[0]**o+(0.881196513588)*x_ref[0]+(-0.202951096266)*x_ref[1]**o+(-9.87983182765e-05)*x_ref[1]
        ref[2,2,0,0]=(0.0811393810501)*x_ref[0]**o+(-0.853555165896)*x_ref[0]+(0.0259257246335)*x_ref[1]**o+(0.638575306959)*x_ref[1]
        ref[2,2,0,1]=(-0.525990031379)*x_ref[0]**o+(0.11235160765)*x_ref[0]+(-0.502853879455)*x_ref[1]**o+(0.414066683507)*x_ref[1]
        ref[2,2,1,0]=(-0.533919716834)*x_ref[0]**o+(-0.773411100656)*x_ref[0]+(0.311424028491)*x_ref[1]**o+(0.332733301377)*x_ref[1]
        ref[2,2,1,1]=(0.280548593462)*x_ref[0]**o+(-0.0747640762104)*x_ref[0]+(-0.7110232877)*x_ref[1]**o+(-0.178672817441)*x_ref[1]
        ref[2,2,2,0]=(0.445025926846)*x_ref[0]**o+(-0.669782903005)*x_ref[0]+(0.236610411793)*x_ref[1]**o+(0.437836936476)*x_ref[1]
        ref[2,2,2,1]=(-0.503761305098)*x_ref[0]**o+(-0.889922659865)*x_ref[0]+(-0.0767256016402)*x_ref[1]**o+(-0.614874217835)*x_ref[1]
        ref[2,3,0,0]=(-0.041960855672)*x_ref[0]**o+(-0.593486242764)*x_ref[0]+(0.344751892525)*x_ref[1]**o+(-0.131268873786)*x_ref[1]
        ref[2,3,0,1]=(-0.192704843083)*x_ref[0]**o+(0.325356381696)*x_ref[0]+(-0.235690651937)*x_ref[1]**o+(0.59113535617)*x_ref[1]
        ref[2,3,1,0]=(-0.0794745628024)*x_ref[0]**o+(0.7996737586)*x_ref[0]+(-0.324970505375)*x_ref[1]**o+(-0.14920403468)*x_ref[1]
        ref[2,3,1,1]=(0.31997432673)*x_ref[0]**o+(-0.571739807799)*x_ref[0]+(0.517683480263)*x_ref[1]**o+(-0.250680741704)*x_ref[1]
        ref[2,3,2,0]=(-0.379114277883)*x_ref[0]**o+(0.953285017352)*x_ref[0]+(-0.701111847548)*x_ref[1]**o+(-0.354199942528)*x_ref[1]
        ref[2,3,2,1]=(0.492245629486)*x_ref[0]**o+(-0.299380102293)*x_ref[0]+(-0.126371826833)*x_ref[1]**o+(0.933664060647)*x_ref[1]
        ref[2,4,0,0]=(0.714822159533)*x_ref[0]**o+(0.359430380478)*x_ref[0]+(-0.691610768506)*x_ref[1]**o+(0.414061853203)*x_ref[1]
        ref[2,4,0,1]=(0.412799253275)*x_ref[0]**o+(0.862905533148)*x_ref[0]+(-0.0507440946561)*x_ref[1]**o+(0.779012369976)*x_ref[1]
        ref[2,4,1,0]=(-0.981179053034)*x_ref[0]**o+(0.984777005573)*x_ref[0]+(0.715584788452)*x_ref[1]**o+(0.920666888724)*x_ref[1]
        ref[2,4,1,1]=(-0.0287775685032)*x_ref[0]**o+(0.617592345344)*x_ref[0]+(-0.767451297794)*x_ref[1]**o+(0.95828976091)*x_ref[1]
        ref[2,4,2,0]=(0.905037473705)*x_ref[0]**o+(-0.706987158021)*x_ref[0]+(0.535436652493)*x_ref[1]**o+(-0.891712664249)*x_ref[1]
        ref[2,4,2,1]=(-0.818709474925)*x_ref[0]**o+(-0.0223479273553)*x_ref[0]+(-0.370644568884)*x_ref[1]**o+(0.286468739105)*x_ref[1]
        ref[3,0,0,0]=(0.310461949876)*x_ref[0]**o+(0.770193446516)*x_ref[0]+(-0.0858335164699)*x_ref[1]**o+(0.12572608122)*x_ref[1]
        ref[3,0,0,1]=(-0.781332422249)*x_ref[0]**o+(-0.238368527872)*x_ref[0]+(-0.449272256582)*x_ref[1]**o+(-0.645638738212)*x_ref[1]
        ref[3,0,1,0]=(-0.91143447678)*x_ref[0]**o+(-0.885131081008)*x_ref[0]+(-0.253833463581)*x_ref[1]**o+(0.051439649459)*x_ref[1]
        ref[3,0,1,1]=(0.447730169203)*x_ref[0]**o+(0.980354452325)*x_ref[0]+(-0.0484566141047)*x_ref[1]**o+(0.443407604362)*x_ref[1]
        ref[3,0,2,0]=(-0.223975544426)*x_ref[0]**o+(0.0141796274312)*x_ref[0]+(0.664080206442)*x_ref[1]**o+(0.193318634886)*x_ref[1]
        ref[3,0,2,1]=(-0.388901068755)*x_ref[0]**o+(-0.632210208163)*x_ref[0]+(0.567264328931)*x_ref[1]**o+(-0.308710733281)*x_ref[1]
        ref[3,1,0,0]=(0.259562910069)*x_ref[0]**o+(-0.315409720689)*x_ref[0]+(0.999281841483)*x_ref[1]**o+(0.695981779357)*x_ref[1]
        ref[3,1,0,1]=(0.627508253716)*x_ref[0]**o+(-0.540950954632)*x_ref[0]+(0.626209667445)*x_ref[1]**o+(-0.920403595109)*x_ref[1]
        ref[3,1,1,0]=(0.750166528891)*x_ref[0]**o+(-0.145494783124)*x_ref[0]+(0.71048043416)*x_ref[1]**o+(-0.0920505950401)*x_ref[1]
        ref[3,1,1,1]=(-0.755570630755)*x_ref[0]**o+(-0.323456877629)*x_ref[0]+(-0.0318418288689)*x_ref[1]**o+(0.551823313307)*x_ref[1]
        ref[3,1,2,0]=(-0.301380276426)*x_ref[0]**o+(-0.102899117196)*x_ref[0]+(0.816857028902)*x_ref[1]**o+(-0.571555621895)*x_ref[1]
        ref[3,1,2,1]=(-0.791148729922)*x_ref[0]**o+(0.400286696027)*x_ref[0]+(-0.17362002139)*x_ref[1]**o+(0.431696064424)*x_ref[1]
        ref[3,2,0,0]=(0.241944451237)*x_ref[0]**o+(-0.629401496412)*x_ref[0]+(-0.00333636607647)*x_ref[1]**o+(0.167334133677)*x_ref[1]
        ref[3,2,0,1]=(-0.300944831932)*x_ref[0]**o+(0.358282773504)*x_ref[0]+(0.759495661081)*x_ref[1]**o+(-0.225250488484)*x_ref[1]
        ref[3,2,1,0]=(0.755950588768)*x_ref[0]**o+(0.718423180013)*x_ref[0]+(0.705150216505)*x_ref[1]**o+(0.269283969196)*x_ref[1]
        ref[3,2,1,1]=(0.811064150458)*x_ref[0]**o+(-0.67847842696)*x_ref[0]+(-0.478091805711)*x_ref[1]**o+(0.292774429011)*x_ref[1]
        ref[3,2,2,0]=(-0.307877253758)*x_ref[0]**o+(-0.563961673764)*x_ref[0]+(0.428322493356)*x_ref[1]**o+(0.564230126489)*x_ref[1]
        ref[3,2,2,1]=(-0.623877056979)*x_ref[0]**o+(-0.148337578412)*x_ref[0]+(-0.910662605581)*x_ref[1]**o+(-0.38603493572)*x_ref[1]
        ref[3,3,0,0]=(-0.826474270729)*x_ref[0]**o+(0.977614689932)*x_ref[0]+(-0.369824837717)*x_ref[1]**o+(-0.985478011812)*x_ref[1]
        ref[3,3,0,1]=(-0.251252769561)*x_ref[0]**o+(-0.296732765057)*x_ref[0]+(-0.888899418995)*x_ref[1]**o+(0.640313055734)*x_ref[1]
        ref[3,3,1,0]=(-0.189106218841)*x_ref[0]**o+(-0.0441656339777)*x_ref[0]+(0.368716219801)*x_ref[1]**o+(-0.924144919367)*x_ref[1]
        ref[3,3,1,1]=(-0.525205981948)*x_ref[0]**o+(-0.148352446593)*x_ref[0]+(0.869212821652)*x_ref[1]**o+(-0.575463559793)*x_ref[1]
        ref[3,3,2,0]=(0.668772663439)*x_ref[0]**o+(0.4330630456)*x_ref[0]+(-0.321158907274)*x_ref[1]**o+(0.570853139693)*x_ref[1]
        ref[3,3,2,1]=(0.0236674031998)*x_ref[0]**o+(-0.390106024651)*x_ref[0]+(0.541717893494)*x_ref[1]**o+(-0.144655705849)*x_ref[1]
        ref[3,4,0,0]=(0.971861188602)*x_ref[0]**o+(-0.834463063474)*x_ref[0]+(-0.419545662218)*x_ref[1]**o+(-0.0387341522049)*x_ref[1]
        ref[3,4,0,1]=(0.226357589896)*x_ref[0]**o+(0.305148895256)*x_ref[0]+(0.00306527642431)*x_ref[1]**o+(0.4715619746)*x_ref[1]
        ref[3,4,1,0]=(0.529602340964)*x_ref[0]**o+(-0.482050910795)*x_ref[0]+(0.964983911769)*x_ref[1]**o+(0.524018220532)*x_ref[1]
        ref[3,4,1,1]=(0.940676158872)*x_ref[0]**o+(0.547078925647)*x_ref[0]+(-0.636907424453)*x_ref[1]**o+(0.938262396281)*x_ref[1]
        ref[3,4,2,0]=(0.282512784255)*x_ref[0]**o+(-0.556345478931)*x_ref[0]+(0.651866614458)*x_ref[1]**o+(0.8895132214)*x_ref[1]
        ref[3,4,2,1]=(0.905077434472)*x_ref[0]**o+(0.595878819874)*x_ref[0]+(0.416350660905)*x_ref[1]**o+(0.100262997545)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.836302395465)*x[0]**o+(-0.842568675604)*x[0]+(-0.780133441207)*x[1]**o+(0.54483947113)*x[1]+(-0.130253019696)*x[2]**o+(0.358303389828)*x[2]
        arg[0,0,0,1]=(-0.722847500652)*x[0]**o+(0.42887368878)*x[0]+(0.979187056933)*x[1]**o+(0.624875714982)*x[1]+(-0.672529765247)*x[2]**o+(0.822363583702)*x[2]
        arg[0,0,1,0]=(-0.28989650691)*x[0]**o+(-0.847647954773)*x[0]+(-0.11119862113)*x[1]**o+(0.0809009676818)*x[1]+(-0.672626533587)*x[2]**o+(0.580561232267)*x[2]
        arg[0,0,1,1]=(-0.64783363248)*x[0]**o+(0.212828552873)*x[0]+(-0.845297180936)*x[1]**o+(-0.714760411186)*x[1]+(-0.385173605122)*x[2]**o+(-0.760750317556)*x[2]
        arg[0,0,2,0]=(-0.328757006046)*x[0]**o+(-0.0935907722247)*x[0]+(-0.895866135795)*x[1]**o+(-0.701241976745)*x[1]+(0.893684139941)*x[2]**o+(-0.265339318222)*x[2]
        arg[0,0,2,1]=(0.134066662346)*x[0]**o+(0.0383231508856)*x[0]+(-0.347744261995)*x[1]**o+(-0.86897907336)*x[1]+(-0.287792154521)*x[2]**o+(0.859351846891)*x[2]
        arg[0,1,0,0]=(-0.635128124936)*x[0]**o+(-0.659462189733)*x[0]+(-0.0964365747936)*x[1]**o+(-0.385702168747)*x[1]+(0.179539914057)*x[2]**o+(-0.0176420847859)*x[2]
        arg[0,1,0,1]=(-0.580750807831)*x[0]**o+(0.756493147583)*x[0]+(-0.436842251289)*x[1]**o+(-0.705624646095)*x[1]+(0.983064096262)*x[2]**o+(-0.0197153825423)*x[2]
        arg[0,1,1,0]=(-0.567748582365)*x[0]**o+(0.579324420356)*x[0]+(0.0434168493419)*x[1]**o+(-0.839692920334)*x[1]+(-0.119239847745)*x[2]**o+(0.527047942834)*x[2]
        arg[0,1,1,1]=(0.52870403655)*x[0]**o+(0.700090892471)*x[0]+(0.730224313921)*x[1]**o+(-0.815451751207)*x[1]+(-0.943945382413)*x[2]**o+(-0.571714141916)*x[2]
        arg[0,1,2,0]=(-0.803841094887)*x[0]**o+(0.0595126006237)*x[0]+(-0.612341797645)*x[1]**o+(0.944552805211)*x[1]+(-0.00253009731786)*x[2]**o+(0.0174867439912)*x[2]
        arg[0,1,2,1]=(-0.145623212549)*x[0]**o+(0.546039957008)*x[0]+(-0.221425237579)*x[1]**o+(0.229367671159)*x[1]+(-0.784903869759)*x[2]**o+(0.898045208871)*x[2]
        arg[0,2,0,0]=(0.417924494005)*x[0]**o+(-0.192820051987)*x[0]+(0.878916271492)*x[1]**o+(-0.35056320984)*x[1]+(0.982897575604)*x[2]**o+(-0.225793730616)*x[2]
        arg[0,2,0,1]=(-0.493279498849)*x[0]**o+(-0.154177411087)*x[0]+(-0.923592860492)*x[1]**o+(0.872010949252)*x[1]+(-0.936252147064)*x[2]**o+(0.947136819308)*x[2]
        arg[0,2,1,0]=(0.109357630607)*x[0]**o+(-0.518825477709)*x[0]+(-0.375510667498)*x[1]**o+(0.476537999644)*x[1]+(0.676017421029)*x[2]**o+(-0.633301158984)*x[2]
        arg[0,2,1,1]=(-0.0662265225539)*x[0]**o+(-0.276634748263)*x[0]+(-0.0655431981102)*x[1]**o+(0.0578482920991)*x[1]+(-0.264370611185)*x[2]**o+(-0.674612985594)*x[2]
        arg[0,2,2,0]=(-0.924186592743)*x[0]**o+(0.197603043149)*x[0]+(-0.676882805046)*x[1]**o+(0.564142845865)*x[1]+(0.590369345306)*x[2]**o+(-0.872019490978)*x[2]
        arg[0,2,2,1]=(0.480683776968)*x[0]**o+(-0.412055284173)*x[0]+(0.829996313473)*x[1]**o+(-0.235665473048)*x[1]+(-0.429191305284)*x[2]**o+(-0.0867804318457)*x[2]
        arg[0,3,0,0]=(-0.0903660006043)*x[0]**o+(-0.507571901766)*x[0]+(-0.0305305861809)*x[1]**o+(-0.915027969989)*x[1]+(0.955644774347)*x[2]**o+(0.739244618775)*x[2]
        arg[0,3,0,1]=(0.600619122998)*x[0]**o+(0.0128699484527)*x[0]+(-0.360431921281)*x[1]**o+(0.304773619811)*x[1]+(-0.302423314054)*x[2]**o+(0.680149401572)*x[2]
        arg[0,3,1,0]=(-0.638540513323)*x[0]**o+(0.766655754067)*x[0]+(0.276879854223)*x[1]**o+(-0.628688275247)*x[1]+(0.724584107114)*x[2]**o+(0.682193551967)*x[2]
        arg[0,3,1,1]=(-0.516656801349)*x[0]**o+(-0.513901319383)*x[0]+(-0.227957010777)*x[1]**o+(-0.164363207703)*x[1]+(-0.919936042047)*x[2]**o+(0.682400184997)*x[2]
        arg[0,3,2,0]=(-0.819629653095)*x[0]**o+(0.183937049274)*x[0]+(0.0411075800579)*x[1]**o+(0.0813088270699)*x[1]+(0.762416334575)*x[2]**o+(0.236553726705)*x[2]
        arg[0,3,2,1]=(0.513141183759)*x[0]**o+(-0.452475512734)*x[0]+(0.0461424221984)*x[1]**o+(-0.673452844636)*x[1]+(-0.194493328878)*x[2]**o+(0.812459956422)*x[2]
        arg[0,4,0,0]=(0.672167415195)*x[0]**o+(-0.641004730405)*x[0]+(-0.383031420851)*x[1]**o+(0.392366431186)*x[1]+(0.66479124224)*x[2]**o+(0.845503225104)*x[2]
        arg[0,4,0,1]=(0.908241966695)*x[0]**o+(-0.816989438726)*x[0]+(-0.144902267583)*x[1]**o+(-0.654901540598)*x[1]+(-0.9840470269)*x[2]**o+(0.510970695154)*x[2]
        arg[0,4,1,0]=(0.788858704366)*x[0]**o+(-0.50266306504)*x[0]+(-0.681248796723)*x[1]**o+(0.0317663112509)*x[1]+(0.662428035619)*x[2]**o+(0.377116420912)*x[2]
        arg[0,4,1,1]=(-0.0734522829037)*x[0]**o+(0.0822741484655)*x[0]+(-0.509279578815)*x[1]**o+(-0.95372420098)*x[1]+(0.682868362508)*x[2]**o+(-0.947156883638)*x[2]
        arg[0,4,2,0]=(-0.843236607337)*x[0]**o+(-0.925326238513)*x[0]+(-0.153688519983)*x[1]**o+(-0.68456413487)*x[1]+(0.920442617368)*x[2]**o+(0.852945557893)*x[2]
        arg[0,4,2,1]=(-0.283772512528)*x[0]**o+(0.202730738521)*x[0]+(-0.287375172356)*x[1]**o+(-0.176699132816)*x[1]+(0.446391113264)*x[2]**o+(-0.783266847348)*x[2]
        arg[1,0,0,0]=(0.81284364562)*x[0]**o+(-0.454432332505)*x[0]+(-0.684692596262)*x[1]**o+(-0.345923888237)*x[1]+(0.1201728631)*x[2]**o+(0.808246135028)*x[2]
        arg[1,0,0,1]=(0.941104328706)*x[0]**o+(-0.802008654964)*x[0]+(0.902047483677)*x[1]**o+(0.777049224612)*x[1]+(0.120184332854)*x[2]**o+(0.983216899102)*x[2]
        arg[1,0,1,0]=(-0.420996925767)*x[0]**o+(0.0941750615502)*x[0]+(-0.935342933578)*x[1]**o+(0.925891822058)*x[1]+(-0.153652985582)*x[2]**o+(-0.336548986153)*x[2]
        arg[1,0,1,1]=(-0.692886221795)*x[0]**o+(-0.920419410139)*x[0]+(-0.586575583263)*x[1]**o+(-0.269347362009)*x[1]+(-0.872169736045)*x[2]**o+(-0.724727186336)*x[2]
        arg[1,0,2,0]=(-0.0171222282058)*x[0]**o+(0.338180602976)*x[0]+(-0.0345958505395)*x[1]**o+(0.631728421505)*x[1]+(0.90596443453)*x[2]**o+(-0.144785193015)*x[2]
        arg[1,0,2,1]=(-0.91636583885)*x[0]**o+(-0.807496789574)*x[0]+(0.63472495957)*x[1]**o+(0.966464343993)*x[1]+(-0.471311384621)*x[2]**o+(-0.449420138238)*x[2]
        arg[1,1,0,0]=(-0.239930503474)*x[0]**o+(-0.476439715994)*x[0]+(0.421600511888)*x[1]**o+(0.477823144792)*x[1]+(0.251409914807)*x[2]**o+(0.636818001419)*x[2]
        arg[1,1,0,1]=(-0.427510410263)*x[0]**o+(0.494810857516)*x[0]+(-0.129415181352)*x[1]**o+(0.754147770071)*x[1]+(0.0218776365985)*x[2]**o+(0.706183216616)*x[2]
        arg[1,1,1,0]=(-0.469150375165)*x[0]**o+(-0.0651101638073)*x[0]+(-0.697853003861)*x[1]**o+(0.0389330102026)*x[1]+(-0.256557942045)*x[2]**o+(0.409565340425)*x[2]
        arg[1,1,1,1]=(0.14957005608)*x[0]**o+(-0.209482968126)*x[0]+(0.368899650309)*x[1]**o+(-0.0899321000137)*x[1]+(0.282389903357)*x[2]**o+(-0.166633883916)*x[2]
        arg[1,1,2,0]=(0.0374938231714)*x[0]**o+(-0.687972899893)*x[0]+(-0.0686478107779)*x[1]**o+(0.318977937888)*x[1]+(-0.601269126788)*x[2]**o+(0.453188855786)*x[2]
        arg[1,1,2,1]=(-0.853894223222)*x[0]**o+(0.441830151478)*x[0]+(0.973114898368)*x[1]**o+(-0.264454889071)*x[1]+(-0.828332687531)*x[2]**o+(0.302493748565)*x[2]
        arg[1,2,0,0]=(0.210365824111)*x[0]**o+(0.4674117857)*x[0]+(-0.129546969044)*x[1]**o+(0.566357125352)*x[1]+(-0.721733448514)*x[2]**o+(0.154253207106)*x[2]
        arg[1,2,0,1]=(0.805325508181)*x[0]**o+(-0.929113563452)*x[0]+(-0.487301761321)*x[1]**o+(-0.822405526519)*x[1]+(0.254754490326)*x[2]**o+(-0.196480710709)*x[2]
        arg[1,2,1,0]=(-0.520777275475)*x[0]**o+(-0.912569745644)*x[0]+(-0.784504791087)*x[1]**o+(-0.903473217653)*x[1]+(0.583297107374)*x[2]**o+(0.664811898195)*x[2]
        arg[1,2,1,1]=(0.129152152958)*x[0]**o+(0.497832200867)*x[0]+(0.121217248234)*x[1]**o+(-0.183368066714)*x[1]+(-0.776145485536)*x[2]**o+(-0.223165724715)*x[2]
        arg[1,2,2,0]=(-0.698981324488)*x[0]**o+(0.00951277777924)*x[0]+(0.997410976507)*x[1]**o+(0.534670303453)*x[1]+(0.251309223218)*x[2]**o+(-0.921148646979)*x[2]
        arg[1,2,2,1]=(-0.731509934487)*x[0]**o+(0.87731926834)*x[0]+(0.0865507713913)*x[1]**o+(0.708835987019)*x[1]+(0.294182269171)*x[2]**o+(-0.316660750729)*x[2]
        arg[1,3,0,0]=(-0.282903962918)*x[0]**o+(-0.793457978844)*x[0]+(0.287425807914)*x[1]**o+(-0.0252722364835)*x[1]+(-0.956202022515)*x[2]**o+(0.650271579135)*x[2]
        arg[1,3,0,1]=(-0.946723111202)*x[0]**o+(0.669431653693)*x[0]+(-0.791125976188)*x[1]**o+(-0.466484914794)*x[1]+(-0.217225801872)*x[2]**o+(0.224719473996)*x[2]
        arg[1,3,1,0]=(0.516479301043)*x[0]**o+(-0.42420332366)*x[0]+(-0.0649227247176)*x[1]**o+(0.208849752675)*x[1]+(-0.939001099693)*x[2]**o+(-0.0537827140932)*x[2]
        arg[1,3,1,1]=(-0.292077069459)*x[0]**o+(-0.444848119744)*x[0]+(0.268592719199)*x[1]**o+(-0.349401060323)*x[1]+(-0.540803784489)*x[2]**o+(-0.487170123526)*x[2]
        arg[1,3,2,0]=(-0.626222042661)*x[0]**o+(0.245937409624)*x[0]+(-0.24692162006)*x[1]**o+(0.108230029208)*x[1]+(-0.760463241928)*x[2]**o+(-0.587078263201)*x[2]
        arg[1,3,2,1]=(-0.923891147756)*x[0]**o+(0.0961285842162)*x[0]+(0.0146853466585)*x[1]**o+(0.00951339122359)*x[1]+(-0.261383704723)*x[2]**o+(-0.0262515691324)*x[2]
        arg[1,4,0,0]=(-0.0370003524292)*x[0]**o+(-0.0463848489957)*x[0]+(0.940061458694)*x[1]**o+(0.592787058925)*x[1]+(-0.217284379638)*x[2]**o+(-0.11746834109)*x[2]
        arg[1,4,0,1]=(0.830404283093)*x[0]**o+(0.268135490403)*x[0]+(-0.894500398195)*x[1]**o+(0.751042648917)*x[1]+(0.687583292754)*x[2]**o+(-0.884596509742)*x[2]
        arg[1,4,1,0]=(0.414764592469)*x[0]**o+(-0.529613347919)*x[0]+(-0.727435115594)*x[1]**o+(-0.546579777725)*x[1]+(0.378011893971)*x[2]**o+(-0.578081183238)*x[2]
        arg[1,4,1,1]=(0.406513376747)*x[0]**o+(-0.350132869579)*x[0]+(-0.557954631299)*x[1]**o+(0.868176175754)*x[1]+(0.793021832632)*x[2]**o+(0.280529771851)*x[2]
        arg[1,4,2,0]=(0.56741953575)*x[0]**o+(0.0115947684302)*x[0]+(0.872812725363)*x[1]**o+(-0.215183541745)*x[1]+(0.64853692203)*x[2]**o+(-0.139561471922)*x[2]
        arg[1,4,2,1]=(0.650916904666)*x[0]**o+(0.878765271839)*x[0]+(0.0141264348052)*x[1]**o+(-0.56314426774)*x[1]+(-0.267854377704)*x[2]**o+(0.852271380667)*x[2]
        arg[2,0,0,0]=(0.948359141577)*x[0]**o+(-0.0881747280974)*x[0]+(0.804871563368)*x[1]**o+(-0.0298585763446)*x[1]+(0.506923333569)*x[2]**o+(-0.920947093648)*x[2]
        arg[2,0,0,1]=(-0.0144221258584)*x[0]**o+(-0.679723072824)*x[0]+(0.00829336956687)*x[1]**o+(0.675680809303)*x[1]+(-0.803195231536)*x[2]**o+(0.70240358571)*x[2]
        arg[2,0,1,0]=(-0.159174987706)*x[0]**o+(0.922309271483)*x[0]+(0.344354343035)*x[1]**o+(-0.00954038311323)*x[1]+(-0.50225452804)*x[2]**o+(0.355899734308)*x[2]
        arg[2,0,1,1]=(0.0857611780404)*x[0]**o+(0.876479431983)*x[0]+(0.105049158502)*x[1]**o+(-0.530308461426)*x[1]+(-0.553136649343)*x[2]**o+(0.385458516481)*x[2]
        arg[2,0,2,0]=(-0.260086070842)*x[0]**o+(0.417766241499)*x[0]+(-0.292370383673)*x[1]**o+(-0.321315349553)*x[1]+(-0.437301444082)*x[2]**o+(0.730845669715)*x[2]
        arg[2,0,2,1]=(0.132547674276)*x[0]**o+(0.783602886333)*x[0]+(0.822361450413)*x[1]**o+(-0.810402783)*x[1]+(0.759701191023)*x[2]**o+(0.644375788608)*x[2]
        arg[2,1,0,0]=(-0.841458111227)*x[0]**o+(0.779917641736)*x[0]+(-0.0745587217256)*x[1]**o+(0.938545280143)*x[1]+(0.454600163373)*x[2]**o+(-0.490348377697)*x[2]
        arg[2,1,0,1]=(-0.876477136174)*x[0]**o+(-0.248200369939)*x[0]+(-0.973654243816)*x[1]**o+(0.482402123794)*x[1]+(-0.776970490827)*x[2]**o+(-0.571349467703)*x[2]
        arg[2,1,1,0]=(-0.433528193102)*x[0]**o+(-0.57712915767)*x[0]+(-0.328671198058)*x[1]**o+(-0.926873019075)*x[1]+(0.314232739356)*x[2]**o+(-0.199625466412)*x[2]
        arg[2,1,1,1]=(0.892707770392)*x[0]**o+(-0.680081898234)*x[0]+(-0.754181058483)*x[1]**o+(-0.253728945575)*x[1]+(-0.906373406684)*x[2]**o+(-0.0266435080833)*x[2]
        arg[2,1,2,0]=(-0.121605378472)*x[0]**o+(-0.503955015634)*x[0]+(-0.0960767196114)*x[1]**o+(-0.937914890361)*x[1]+(0.212042931853)*x[2]**o+(0.930703155408)*x[2]
        arg[2,1,2,1]=(0.0564159441047)*x[0]**o+(0.842332721105)*x[0]+(0.954741657608)*x[1]**o+(0.73503176)*x[1]+(-0.390679938582)*x[2]**o+(-0.440014747741)*x[2]
        arg[2,2,0,0]=(0.0100637970821)*x[0]**o+(-0.0283353154145)*x[0]+(0.865360758262)*x[1]**o+(0.780975027988)*x[1]+(0.16115842919)*x[2]**o+(0.247580155656)*x[2]
        arg[2,2,0,1]=(0.61102059426)*x[0]**o+(-0.198005744101)*x[0]+(0.503568797262)*x[1]**o+(0.0288809726497)*x[1]+(-0.765127098183)*x[2]**o+(0.409803857723)*x[2]
        arg[2,2,1,0]=(-0.712272018295)*x[0]**o+(-0.62562142063)*x[0]+(-0.579501949719)*x[1]**o+(0.628539275236)*x[1]+(0.0200945124964)*x[2]**o+(0.175966024769)*x[2]
        arg[2,2,1,1]=(-0.798244445519)*x[0]**o+(-0.376440962535)*x[0]+(-0.134467401497)*x[1]**o+(0.877538294477)*x[1]+(-0.484870741192)*x[2]**o+(-0.224064469935)*x[2]
        arg[2,2,2,0]=(0.200196926368)*x[0]**o+(-0.528000867086)*x[0]+(-0.708942342624)*x[1]**o+(-0.591994969141)*x[1]+(-0.990641592656)*x[2]**o+(-0.875897347303)*x[2]
        arg[2,2,2,1]=(0.949113190797)*x[0]**o+(0.867855496786)*x[0]+(-0.0314953438065)*x[1]**o+(-0.371928428005)*x[1]+(-0.526403191063)*x[2]**o+(0.653049122326)*x[2]
        arg[2,3,0,0]=(-0.353418327127)*x[0]**o+(-0.418588306429)*x[0]+(0.400872043269)*x[1]**o+(0.141907924042)*x[1]+(-0.999869579462)*x[2]**o+(-0.410423954563)*x[2]
        arg[2,3,0,1]=(-0.240390877307)*x[0]**o+(0.0901879916115)*x[0]+(-0.0734683209657)*x[1]**o+(0.274836005886)*x[1]+(-0.336375498181)*x[2]**o+(-0.754776455221)*x[2]
        arg[2,3,1,0]=(0.259928460822)*x[0]**o+(-0.227630454708)*x[0]+(0.725968066278)*x[1]**o+(-0.545440695763)*x[1]+(-0.0354926607815)*x[2]**o+(0.365010532181)*x[2]
        arg[2,3,1,1]=(-0.969799373398)*x[0]**o+(0.514498233214)*x[0]+(0.907450925318)*x[1]**o+(-0.234993706406)*x[1]+(0.206323910806)*x[2]**o+(0.209364250369)*x[2]
        arg[2,3,2,0]=(0.712905565941)*x[0]**o+(-0.353322758263)*x[0]+(0.816861181494)*x[1]**o+(-0.362388243619)*x[1]+(0.302543517607)*x[2]**o+(-0.831280418852)*x[2]
        arg[2,3,2,1]=(-0.386373711182)*x[0]**o+(0.946914567792)*x[0]+(-0.52795085887)*x[1]**o+(0.86635245964)*x[1]+(0.654261054178)*x[2]**o+(0.452334882498)*x[2]
        arg[2,4,0,0]=(0.374747033547)*x[0]**o+(-0.3471578669)*x[0]+(0.322217002639)*x[1]**o+(0.92988068982)*x[1]+(-0.642423472369)*x[2]**o+(-0.759623225653)*x[2]
        arg[2,4,0,1]=(-0.322255157841)*x[0]**o+(-0.253806844569)*x[0]+(-0.0688746199637)*x[1]**o+(0.232840317719)*x[1]+(-0.815623721026)*x[2]**o+(0.43055485562)*x[2]
        arg[2,4,1,0]=(-0.452320341493)*x[0]**o+(0.755240168228)*x[0]+(-0.892552611066)*x[1]**o+(0.438275161458)*x[1]+(0.441378369388)*x[2]**o+(-0.870501114753)*x[2]
        arg[2,4,1,1]=(0.496731258286)*x[0]**o+(-0.499918335322)*x[0]+(0.000476061942027)*x[1]**o+(-0.857762621856)*x[1]+(0.957433798133)*x[2]**o+(-0.766450629361)*x[2]
        arg[2,4,2,0]=(0.860379598562)*x[0]**o+(-0.0816478389485)*x[0]+(-0.438181864676)*x[1]**o+(0.78937427676)*x[1]+(-0.97940010041)*x[2]**o+(0.583914605259)*x[2]
        arg[2,4,2,1]=(0.23729037277)*x[0]**o+(-0.365380141474)*x[0]+(-0.471505632575)*x[1]**o+(-0.168838288)*x[1]+(-0.519847065791)*x[2]**o+(-0.816100518668)*x[2]
        arg[3,0,0,0]=(0.729921808661)*x[0]**o+(0.732868255388)*x[0]+(0.305434623795)*x[1]**o+(-0.45402804883)*x[1]+(0.604368334171)*x[2]**o+(-0.179233711008)*x[2]
        arg[3,0,0,1]=(-0.208916687176)*x[0]**o+(-0.297916049539)*x[0]+(0.854941564423)*x[1]**o+(0.775355543915)*x[1]+(-0.373382357688)*x[2]**o+(0.792557314074)*x[2]
        arg[3,0,1,0]=(0.606041025674)*x[0]**o+(0.115814644681)*x[0]+(-0.414470914985)*x[1]**o+(-0.136577390174)*x[1]+(0.672303216396)*x[2]**o+(-0.479680752765)*x[2]
        arg[3,0,1,1]=(-0.535633649471)*x[0]**o+(0.133761944775)*x[0]+(-0.980107507665)*x[1]**o+(0.211561747114)*x[1]+(-0.530648770004)*x[2]**o+(0.417977629512)*x[2]
        arg[3,0,2,0]=(-0.242797826282)*x[0]**o+(-0.828691638823)*x[0]+(-0.187401612749)*x[1]**o+(-0.453821620107)*x[1]+(-0.475718408605)*x[2]**o+(0.683502702497)*x[2]
        arg[3,0,2,1]=(-0.66640812951)*x[0]**o+(-0.259184719546)*x[0]+(0.463743732654)*x[1]**o+(0.6746349156)*x[1]+(-0.0532918236783)*x[2]**o+(-0.714197225876)*x[2]
        arg[3,1,0,0]=(-0.97693616823)*x[0]**o+(-0.825560555757)*x[0]+(0.118452690664)*x[1]**o+(-0.125279874388)*x[1]+(0.733803136335)*x[2]**o+(-0.884902209956)*x[2]
        arg[3,1,0,1]=(-0.652466190334)*x[0]**o+(0.370106640243)*x[0]+(0.0850861818306)*x[1]**o+(-0.49744492012)*x[1]+(-0.843498619618)*x[2]**o+(-0.0480070808698)*x[2]
        arg[3,1,1,0]=(-0.450526704276)*x[0]**o+(0.430774951776)*x[0]+(-0.956799540696)*x[1]**o+(0.820090546747)*x[1]+(0.831582198997)*x[2]**o+(0.491042906394)*x[2]
        arg[3,1,1,1]=(0.419717807665)*x[0]**o+(0.320904534666)*x[0]+(-0.483247938999)*x[1]**o+(0.82174341564)*x[1]+(-0.441719998974)*x[2]**o+(0.691713322239)*x[2]
        arg[3,1,2,0]=(0.255464093123)*x[0]**o+(-0.256334953799)*x[0]+(0.677105284128)*x[1]**o+(-0.584438025212)*x[1]+(0.505820591597)*x[2]**o+(0.247943562937)*x[2]
        arg[3,1,2,1]=(-0.495204107745)*x[0]**o+(0.234426349601)*x[0]+(0.736969108189)*x[1]**o+(-0.369684409765)*x[1]+(0.164303379282)*x[2]**o+(-0.595757590519)*x[2]
        arg[3,2,0,0]=(0.238639684591)*x[0]**o+(0.568638143194)*x[0]+(0.972646679042)*x[1]**o+(-0.652648194795)*x[1]+(-0.0530713046414)*x[2]**o+(0.356043665396)*x[2]
        arg[3,2,0,1]=(0.458539145106)*x[0]**o+(0.111417037398)*x[0]+(-0.135374868193)*x[1]**o+(-0.535451685997)*x[1]+(-0.820797762877)*x[2]**o+(0.122239787933)*x[2]
        arg[3,2,1,0]=(-0.323251425039)*x[0]**o+(0.106998797206)*x[0]+(-0.0671655760766)*x[1]**o+(0.836342794315)*x[1]+(0.34628715493)*x[2]**o+(-0.11058327806)*x[2]
        arg[3,2,1,1]=(0.181994203911)*x[0]**o+(-0.710678453639)*x[0]+(0.625606409786)*x[1]**o+(-0.363170328543)*x[1]+(-0.932998745422)*x[2]**o+(-0.0411581284314)*x[2]
        arg[3,2,2,0]=(0.710847104903)*x[0]**o+(0.483971915285)*x[0]+(0.42996482533)*x[1]**o+(0.456817348253)*x[1]+(0.844562058604)*x[2]**o+(-0.469586014721)*x[2]
        arg[3,2,2,1]=(-0.845771280837)*x[0]**o+(-0.081384696494)*x[0]+(0.71823377714)*x[1]**o+(-0.846672143168)*x[1]+(-0.139540918461)*x[2]**o+(0.837646131987)*x[2]
        arg[3,3,0,0]=(-0.958726792109)*x[0]**o+(0.53196455268)*x[0]+(-0.621810423658)*x[1]**o+(-0.0352106197658)*x[1]+(-0.636059207474)*x[2]**o+(-0.0491118207063)*x[2]
        arg[3,3,0,1]=(0.987836799279)*x[0]**o+(0.41633576224)*x[0]+(0.312021799176)*x[1]**o+(-0.168836582938)*x[1]+(-0.391725003735)*x[2]**o+(-0.158000182269)*x[2]
        arg[3,3,1,0]=(-0.280868116529)*x[0]**o+(-0.687924000092)*x[0]+(0.865840229722)*x[1]**o+(-0.0823203022874)*x[1]+(-0.857309899784)*x[2]**o+(-0.0657473493059)*x[2]
        arg[3,3,1,1]=(0.680496030567)*x[0]**o+(0.159575720724)*x[0]+(0.0531895122046)*x[1]**o+(0.359273896084)*x[1]+(0.35084774502)*x[2]**o+(0.0982220259531)*x[2]
        arg[3,3,2,0]=(0.140872311111)*x[0]**o+(-0.00845309046907)*x[0]+(-0.845527142503)*x[1]**o+(-0.67713103521)*x[1]+(0.972054187828)*x[2]**o+(-0.800664963509)*x[2]
        arg[3,3,2,1]=(-0.490866666782)*x[0]**o+(-0.0988939523952)*x[0]+(0.236210969508)*x[1]**o+(-0.672617589992)*x[1]+(-0.640558724403)*x[2]**o+(-0.203125267713)*x[2]
        arg[3,4,0,0]=(0.318438839683)*x[0]**o+(0.0864322377812)*x[0]+(-0.371419701638)*x[1]**o+(-0.311583139142)*x[1]+(-0.0533111244011)*x[2]**o+(0.841240973423)*x[2]
        arg[3,4,0,1]=(-0.573110771334)*x[0]**o+(0.703852077914)*x[0]+(0.613309217217)*x[1]**o+(0.247223193846)*x[1]+(0.618952485205)*x[2]**o+(-0.395998308184)*x[2]
        arg[3,4,1,0]=(0.512896938934)*x[0]**o+(0.848653171505)*x[0]+(-0.944903687062)*x[1]**o+(0.179782177607)*x[1]+(-0.0287288996538)*x[2]**o+(-0.385519118293)*x[2]
        arg[3,4,1,1]=(0.234190647718)*x[0]**o+(0.445478221419)*x[0]+(0.588931676434)*x[1]**o+(0.140788139998)*x[1]+(-0.761952359657)*x[2]**o+(0.92377426359)*x[2]
        arg[3,4,2,0]=(-0.836038966171)*x[0]**o+(-0.772990045933)*x[0]+(0.0622128209877)*x[1]**o+(-0.73733496141)*x[1]+(0.0234446925161)*x[2]**o+(-0.682234213315)*x[2]
        arg[3,4,2,1]=(0.250498084214)*x[0]**o+(0.539088356159)*x[0]+(-0.691842680176)*x[1]**o+(0.710040802953)*x[1]+(-0.43691551501)*x[2]**o+(0.710668600989)*x[2]
        ref[0,0,0,0]=(-0.836302395465)*x_ref[0]**o+(-0.842568675604)*x_ref[0]+(-0.780133441207)*x_ref[1]**o+(0.54483947113)*x_ref[1]+(-0.130253019696)*x_ref[2]**o+(0.358303389828)*x_ref[2]
        ref[0,0,0,1]=(-0.722847500652)*x_ref[0]**o+(0.42887368878)*x_ref[0]+(0.979187056933)*x_ref[1]**o+(0.624875714982)*x_ref[1]+(-0.672529765247)*x_ref[2]**o+(0.822363583702)*x_ref[2]
        ref[0,0,1,0]=(-0.28989650691)*x_ref[0]**o+(-0.847647954773)*x_ref[0]+(-0.11119862113)*x_ref[1]**o+(0.0809009676818)*x_ref[1]+(-0.672626533587)*x_ref[2]**o+(0.580561232267)*x_ref[2]
        ref[0,0,1,1]=(-0.64783363248)*x_ref[0]**o+(0.212828552873)*x_ref[0]+(-0.845297180936)*x_ref[1]**o+(-0.714760411186)*x_ref[1]+(-0.385173605122)*x_ref[2]**o+(-0.760750317556)*x_ref[2]
        ref[0,0,2,0]=(-0.328757006046)*x_ref[0]**o+(-0.0935907722247)*x_ref[0]+(-0.895866135795)*x_ref[1]**o+(-0.701241976745)*x_ref[1]+(0.893684139941)*x_ref[2]**o+(-0.265339318222)*x_ref[2]
        ref[0,0,2,1]=(0.134066662346)*x_ref[0]**o+(0.0383231508856)*x_ref[0]+(-0.347744261995)*x_ref[1]**o+(-0.86897907336)*x_ref[1]+(-0.287792154521)*x_ref[2]**o+(0.859351846891)*x_ref[2]
        ref[0,1,0,0]=(-0.635128124936)*x_ref[0]**o+(-0.659462189733)*x_ref[0]+(-0.0964365747936)*x_ref[1]**o+(-0.385702168747)*x_ref[1]+(0.179539914057)*x_ref[2]**o+(-0.0176420847859)*x_ref[2]
        ref[0,1,0,1]=(-0.580750807831)*x_ref[0]**o+(0.756493147583)*x_ref[0]+(-0.436842251289)*x_ref[1]**o+(-0.705624646095)*x_ref[1]+(0.983064096262)*x_ref[2]**o+(-0.0197153825423)*x_ref[2]
        ref[0,1,1,0]=(-0.567748582365)*x_ref[0]**o+(0.579324420356)*x_ref[0]+(0.0434168493419)*x_ref[1]**o+(-0.839692920334)*x_ref[1]+(-0.119239847745)*x_ref[2]**o+(0.527047942834)*x_ref[2]
        ref[0,1,1,1]=(0.52870403655)*x_ref[0]**o+(0.700090892471)*x_ref[0]+(0.730224313921)*x_ref[1]**o+(-0.815451751207)*x_ref[1]+(-0.943945382413)*x_ref[2]**o+(-0.571714141916)*x_ref[2]
        ref[0,1,2,0]=(-0.803841094887)*x_ref[0]**o+(0.0595126006237)*x_ref[0]+(-0.612341797645)*x_ref[1]**o+(0.944552805211)*x_ref[1]+(-0.00253009731786)*x_ref[2]**o+(0.0174867439912)*x_ref[2]
        ref[0,1,2,1]=(-0.145623212549)*x_ref[0]**o+(0.546039957008)*x_ref[0]+(-0.221425237579)*x_ref[1]**o+(0.229367671159)*x_ref[1]+(-0.784903869759)*x_ref[2]**o+(0.898045208871)*x_ref[2]
        ref[0,2,0,0]=(0.417924494005)*x_ref[0]**o+(-0.192820051987)*x_ref[0]+(0.878916271492)*x_ref[1]**o+(-0.35056320984)*x_ref[1]+(0.982897575604)*x_ref[2]**o+(-0.225793730616)*x_ref[2]
        ref[0,2,0,1]=(-0.493279498849)*x_ref[0]**o+(-0.154177411087)*x_ref[0]+(-0.923592860492)*x_ref[1]**o+(0.872010949252)*x_ref[1]+(-0.936252147064)*x_ref[2]**o+(0.947136819308)*x_ref[2]
        ref[0,2,1,0]=(0.109357630607)*x_ref[0]**o+(-0.518825477709)*x_ref[0]+(-0.375510667498)*x_ref[1]**o+(0.476537999644)*x_ref[1]+(0.676017421029)*x_ref[2]**o+(-0.633301158984)*x_ref[2]
        ref[0,2,1,1]=(-0.0662265225539)*x_ref[0]**o+(-0.276634748263)*x_ref[0]+(-0.0655431981102)*x_ref[1]**o+(0.0578482920991)*x_ref[1]+(-0.264370611185)*x_ref[2]**o+(-0.674612985594)*x_ref[2]
        ref[0,2,2,0]=(-0.924186592743)*x_ref[0]**o+(0.197603043149)*x_ref[0]+(-0.676882805046)*x_ref[1]**o+(0.564142845865)*x_ref[1]+(0.590369345306)*x_ref[2]**o+(-0.872019490978)*x_ref[2]
        ref[0,2,2,1]=(0.480683776968)*x_ref[0]**o+(-0.412055284173)*x_ref[0]+(0.829996313473)*x_ref[1]**o+(-0.235665473048)*x_ref[1]+(-0.429191305284)*x_ref[2]**o+(-0.0867804318457)*x_ref[2]
        ref[0,3,0,0]=(-0.0903660006043)*x_ref[0]**o+(-0.507571901766)*x_ref[0]+(-0.0305305861809)*x_ref[1]**o+(-0.915027969989)*x_ref[1]+(0.955644774347)*x_ref[2]**o+(0.739244618775)*x_ref[2]
        ref[0,3,0,1]=(0.600619122998)*x_ref[0]**o+(0.0128699484527)*x_ref[0]+(-0.360431921281)*x_ref[1]**o+(0.304773619811)*x_ref[1]+(-0.302423314054)*x_ref[2]**o+(0.680149401572)*x_ref[2]
        ref[0,3,1,0]=(-0.638540513323)*x_ref[0]**o+(0.766655754067)*x_ref[0]+(0.276879854223)*x_ref[1]**o+(-0.628688275247)*x_ref[1]+(0.724584107114)*x_ref[2]**o+(0.682193551967)*x_ref[2]
        ref[0,3,1,1]=(-0.516656801349)*x_ref[0]**o+(-0.513901319383)*x_ref[0]+(-0.227957010777)*x_ref[1]**o+(-0.164363207703)*x_ref[1]+(-0.919936042047)*x_ref[2]**o+(0.682400184997)*x_ref[2]
        ref[0,3,2,0]=(-0.819629653095)*x_ref[0]**o+(0.183937049274)*x_ref[0]+(0.0411075800579)*x_ref[1]**o+(0.0813088270699)*x_ref[1]+(0.762416334575)*x_ref[2]**o+(0.236553726705)*x_ref[2]
        ref[0,3,2,1]=(0.513141183759)*x_ref[0]**o+(-0.452475512734)*x_ref[0]+(0.0461424221984)*x_ref[1]**o+(-0.673452844636)*x_ref[1]+(-0.194493328878)*x_ref[2]**o+(0.812459956422)*x_ref[2]
        ref[0,4,0,0]=(0.672167415195)*x_ref[0]**o+(-0.641004730405)*x_ref[0]+(-0.383031420851)*x_ref[1]**o+(0.392366431186)*x_ref[1]+(0.66479124224)*x_ref[2]**o+(0.845503225104)*x_ref[2]
        ref[0,4,0,1]=(0.908241966695)*x_ref[0]**o+(-0.816989438726)*x_ref[0]+(-0.144902267583)*x_ref[1]**o+(-0.654901540598)*x_ref[1]+(-0.9840470269)*x_ref[2]**o+(0.510970695154)*x_ref[2]
        ref[0,4,1,0]=(0.788858704366)*x_ref[0]**o+(-0.50266306504)*x_ref[0]+(-0.681248796723)*x_ref[1]**o+(0.0317663112509)*x_ref[1]+(0.662428035619)*x_ref[2]**o+(0.377116420912)*x_ref[2]
        ref[0,4,1,1]=(-0.0734522829037)*x_ref[0]**o+(0.0822741484655)*x_ref[0]+(-0.509279578815)*x_ref[1]**o+(-0.95372420098)*x_ref[1]+(0.682868362508)*x_ref[2]**o+(-0.947156883638)*x_ref[2]
        ref[0,4,2,0]=(-0.843236607337)*x_ref[0]**o+(-0.925326238513)*x_ref[0]+(-0.153688519983)*x_ref[1]**o+(-0.68456413487)*x_ref[1]+(0.920442617368)*x_ref[2]**o+(0.852945557893)*x_ref[2]
        ref[0,4,2,1]=(-0.283772512528)*x_ref[0]**o+(0.202730738521)*x_ref[0]+(-0.287375172356)*x_ref[1]**o+(-0.176699132816)*x_ref[1]+(0.446391113264)*x_ref[2]**o+(-0.783266847348)*x_ref[2]
        ref[1,0,0,0]=(0.81284364562)*x_ref[0]**o+(-0.454432332505)*x_ref[0]+(-0.684692596262)*x_ref[1]**o+(-0.345923888237)*x_ref[1]+(0.1201728631)*x_ref[2]**o+(0.808246135028)*x_ref[2]
        ref[1,0,0,1]=(0.941104328706)*x_ref[0]**o+(-0.802008654964)*x_ref[0]+(0.902047483677)*x_ref[1]**o+(0.777049224612)*x_ref[1]+(0.120184332854)*x_ref[2]**o+(0.983216899102)*x_ref[2]
        ref[1,0,1,0]=(-0.420996925767)*x_ref[0]**o+(0.0941750615502)*x_ref[0]+(-0.935342933578)*x_ref[1]**o+(0.925891822058)*x_ref[1]+(-0.153652985582)*x_ref[2]**o+(-0.336548986153)*x_ref[2]
        ref[1,0,1,1]=(-0.692886221795)*x_ref[0]**o+(-0.920419410139)*x_ref[0]+(-0.586575583263)*x_ref[1]**o+(-0.269347362009)*x_ref[1]+(-0.872169736045)*x_ref[2]**o+(-0.724727186336)*x_ref[2]
        ref[1,0,2,0]=(-0.0171222282058)*x_ref[0]**o+(0.338180602976)*x_ref[0]+(-0.0345958505395)*x_ref[1]**o+(0.631728421505)*x_ref[1]+(0.90596443453)*x_ref[2]**o+(-0.144785193015)*x_ref[2]
        ref[1,0,2,1]=(-0.91636583885)*x_ref[0]**o+(-0.807496789574)*x_ref[0]+(0.63472495957)*x_ref[1]**o+(0.966464343993)*x_ref[1]+(-0.471311384621)*x_ref[2]**o+(-0.449420138238)*x_ref[2]
        ref[1,1,0,0]=(-0.239930503474)*x_ref[0]**o+(-0.476439715994)*x_ref[0]+(0.421600511888)*x_ref[1]**o+(0.477823144792)*x_ref[1]+(0.251409914807)*x_ref[2]**o+(0.636818001419)*x_ref[2]
        ref[1,1,0,1]=(-0.427510410263)*x_ref[0]**o+(0.494810857516)*x_ref[0]+(-0.129415181352)*x_ref[1]**o+(0.754147770071)*x_ref[1]+(0.0218776365985)*x_ref[2]**o+(0.706183216616)*x_ref[2]
        ref[1,1,1,0]=(-0.469150375165)*x_ref[0]**o+(-0.0651101638073)*x_ref[0]+(-0.697853003861)*x_ref[1]**o+(0.0389330102026)*x_ref[1]+(-0.256557942045)*x_ref[2]**o+(0.409565340425)*x_ref[2]
        ref[1,1,1,1]=(0.14957005608)*x_ref[0]**o+(-0.209482968126)*x_ref[0]+(0.368899650309)*x_ref[1]**o+(-0.0899321000137)*x_ref[1]+(0.282389903357)*x_ref[2]**o+(-0.166633883916)*x_ref[2]
        ref[1,1,2,0]=(0.0374938231714)*x_ref[0]**o+(-0.687972899893)*x_ref[0]+(-0.0686478107779)*x_ref[1]**o+(0.318977937888)*x_ref[1]+(-0.601269126788)*x_ref[2]**o+(0.453188855786)*x_ref[2]
        ref[1,1,2,1]=(-0.853894223222)*x_ref[0]**o+(0.441830151478)*x_ref[0]+(0.973114898368)*x_ref[1]**o+(-0.264454889071)*x_ref[1]+(-0.828332687531)*x_ref[2]**o+(0.302493748565)*x_ref[2]
        ref[1,2,0,0]=(0.210365824111)*x_ref[0]**o+(0.4674117857)*x_ref[0]+(-0.129546969044)*x_ref[1]**o+(0.566357125352)*x_ref[1]+(-0.721733448514)*x_ref[2]**o+(0.154253207106)*x_ref[2]
        ref[1,2,0,1]=(0.805325508181)*x_ref[0]**o+(-0.929113563452)*x_ref[0]+(-0.487301761321)*x_ref[1]**o+(-0.822405526519)*x_ref[1]+(0.254754490326)*x_ref[2]**o+(-0.196480710709)*x_ref[2]
        ref[1,2,1,0]=(-0.520777275475)*x_ref[0]**o+(-0.912569745644)*x_ref[0]+(-0.784504791087)*x_ref[1]**o+(-0.903473217653)*x_ref[1]+(0.583297107374)*x_ref[2]**o+(0.664811898195)*x_ref[2]
        ref[1,2,1,1]=(0.129152152958)*x_ref[0]**o+(0.497832200867)*x_ref[0]+(0.121217248234)*x_ref[1]**o+(-0.183368066714)*x_ref[1]+(-0.776145485536)*x_ref[2]**o+(-0.223165724715)*x_ref[2]
        ref[1,2,2,0]=(-0.698981324488)*x_ref[0]**o+(0.00951277777924)*x_ref[0]+(0.997410976507)*x_ref[1]**o+(0.534670303453)*x_ref[1]+(0.251309223218)*x_ref[2]**o+(-0.921148646979)*x_ref[2]
        ref[1,2,2,1]=(-0.731509934487)*x_ref[0]**o+(0.87731926834)*x_ref[0]+(0.0865507713913)*x_ref[1]**o+(0.708835987019)*x_ref[1]+(0.294182269171)*x_ref[2]**o+(-0.316660750729)*x_ref[2]
        ref[1,3,0,0]=(-0.282903962918)*x_ref[0]**o+(-0.793457978844)*x_ref[0]+(0.287425807914)*x_ref[1]**o+(-0.0252722364835)*x_ref[1]+(-0.956202022515)*x_ref[2]**o+(0.650271579135)*x_ref[2]
        ref[1,3,0,1]=(-0.946723111202)*x_ref[0]**o+(0.669431653693)*x_ref[0]+(-0.791125976188)*x_ref[1]**o+(-0.466484914794)*x_ref[1]+(-0.217225801872)*x_ref[2]**o+(0.224719473996)*x_ref[2]
        ref[1,3,1,0]=(0.516479301043)*x_ref[0]**o+(-0.42420332366)*x_ref[0]+(-0.0649227247176)*x_ref[1]**o+(0.208849752675)*x_ref[1]+(-0.939001099693)*x_ref[2]**o+(-0.0537827140932)*x_ref[2]
        ref[1,3,1,1]=(-0.292077069459)*x_ref[0]**o+(-0.444848119744)*x_ref[0]+(0.268592719199)*x_ref[1]**o+(-0.349401060323)*x_ref[1]+(-0.540803784489)*x_ref[2]**o+(-0.487170123526)*x_ref[2]
        ref[1,3,2,0]=(-0.626222042661)*x_ref[0]**o+(0.245937409624)*x_ref[0]+(-0.24692162006)*x_ref[1]**o+(0.108230029208)*x_ref[1]+(-0.760463241928)*x_ref[2]**o+(-0.587078263201)*x_ref[2]
        ref[1,3,2,1]=(-0.923891147756)*x_ref[0]**o+(0.0961285842162)*x_ref[0]+(0.0146853466585)*x_ref[1]**o+(0.00951339122359)*x_ref[1]+(-0.261383704723)*x_ref[2]**o+(-0.0262515691324)*x_ref[2]
        ref[1,4,0,0]=(-0.0370003524292)*x_ref[0]**o+(-0.0463848489957)*x_ref[0]+(0.940061458694)*x_ref[1]**o+(0.592787058925)*x_ref[1]+(-0.217284379638)*x_ref[2]**o+(-0.11746834109)*x_ref[2]
        ref[1,4,0,1]=(0.830404283093)*x_ref[0]**o+(0.268135490403)*x_ref[0]+(-0.894500398195)*x_ref[1]**o+(0.751042648917)*x_ref[1]+(0.687583292754)*x_ref[2]**o+(-0.884596509742)*x_ref[2]
        ref[1,4,1,0]=(0.414764592469)*x_ref[0]**o+(-0.529613347919)*x_ref[0]+(-0.727435115594)*x_ref[1]**o+(-0.546579777725)*x_ref[1]+(0.378011893971)*x_ref[2]**o+(-0.578081183238)*x_ref[2]
        ref[1,4,1,1]=(0.406513376747)*x_ref[0]**o+(-0.350132869579)*x_ref[0]+(-0.557954631299)*x_ref[1]**o+(0.868176175754)*x_ref[1]+(0.793021832632)*x_ref[2]**o+(0.280529771851)*x_ref[2]
        ref[1,4,2,0]=(0.56741953575)*x_ref[0]**o+(0.0115947684302)*x_ref[0]+(0.872812725363)*x_ref[1]**o+(-0.215183541745)*x_ref[1]+(0.64853692203)*x_ref[2]**o+(-0.139561471922)*x_ref[2]
        ref[1,4,2,1]=(0.650916904666)*x_ref[0]**o+(0.878765271839)*x_ref[0]+(0.0141264348052)*x_ref[1]**o+(-0.56314426774)*x_ref[1]+(-0.267854377704)*x_ref[2]**o+(0.852271380667)*x_ref[2]
        ref[2,0,0,0]=(0.948359141577)*x_ref[0]**o+(-0.0881747280974)*x_ref[0]+(0.804871563368)*x_ref[1]**o+(-0.0298585763446)*x_ref[1]+(0.506923333569)*x_ref[2]**o+(-0.920947093648)*x_ref[2]
        ref[2,0,0,1]=(-0.0144221258584)*x_ref[0]**o+(-0.679723072824)*x_ref[0]+(0.00829336956687)*x_ref[1]**o+(0.675680809303)*x_ref[1]+(-0.803195231536)*x_ref[2]**o+(0.70240358571)*x_ref[2]
        ref[2,0,1,0]=(-0.159174987706)*x_ref[0]**o+(0.922309271483)*x_ref[0]+(0.344354343035)*x_ref[1]**o+(-0.00954038311323)*x_ref[1]+(-0.50225452804)*x_ref[2]**o+(0.355899734308)*x_ref[2]
        ref[2,0,1,1]=(0.0857611780404)*x_ref[0]**o+(0.876479431983)*x_ref[0]+(0.105049158502)*x_ref[1]**o+(-0.530308461426)*x_ref[1]+(-0.553136649343)*x_ref[2]**o+(0.385458516481)*x_ref[2]
        ref[2,0,2,0]=(-0.260086070842)*x_ref[0]**o+(0.417766241499)*x_ref[0]+(-0.292370383673)*x_ref[1]**o+(-0.321315349553)*x_ref[1]+(-0.437301444082)*x_ref[2]**o+(0.730845669715)*x_ref[2]
        ref[2,0,2,1]=(0.132547674276)*x_ref[0]**o+(0.783602886333)*x_ref[0]+(0.822361450413)*x_ref[1]**o+(-0.810402783)*x_ref[1]+(0.759701191023)*x_ref[2]**o+(0.644375788608)*x_ref[2]
        ref[2,1,0,0]=(-0.841458111227)*x_ref[0]**o+(0.779917641736)*x_ref[0]+(-0.0745587217256)*x_ref[1]**o+(0.938545280143)*x_ref[1]+(0.454600163373)*x_ref[2]**o+(-0.490348377697)*x_ref[2]
        ref[2,1,0,1]=(-0.876477136174)*x_ref[0]**o+(-0.248200369939)*x_ref[0]+(-0.973654243816)*x_ref[1]**o+(0.482402123794)*x_ref[1]+(-0.776970490827)*x_ref[2]**o+(-0.571349467703)*x_ref[2]
        ref[2,1,1,0]=(-0.433528193102)*x_ref[0]**o+(-0.57712915767)*x_ref[0]+(-0.328671198058)*x_ref[1]**o+(-0.926873019075)*x_ref[1]+(0.314232739356)*x_ref[2]**o+(-0.199625466412)*x_ref[2]
        ref[2,1,1,1]=(0.892707770392)*x_ref[0]**o+(-0.680081898234)*x_ref[0]+(-0.754181058483)*x_ref[1]**o+(-0.253728945575)*x_ref[1]+(-0.906373406684)*x_ref[2]**o+(-0.0266435080833)*x_ref[2]
        ref[2,1,2,0]=(-0.121605378472)*x_ref[0]**o+(-0.503955015634)*x_ref[0]+(-0.0960767196114)*x_ref[1]**o+(-0.937914890361)*x_ref[1]+(0.212042931853)*x_ref[2]**o+(0.930703155408)*x_ref[2]
        ref[2,1,2,1]=(0.0564159441047)*x_ref[0]**o+(0.842332721105)*x_ref[0]+(0.954741657608)*x_ref[1]**o+(0.73503176)*x_ref[1]+(-0.390679938582)*x_ref[2]**o+(-0.440014747741)*x_ref[2]
        ref[2,2,0,0]=(0.0100637970821)*x_ref[0]**o+(-0.0283353154145)*x_ref[0]+(0.865360758262)*x_ref[1]**o+(0.780975027988)*x_ref[1]+(0.16115842919)*x_ref[2]**o+(0.247580155656)*x_ref[2]
        ref[2,2,0,1]=(0.61102059426)*x_ref[0]**o+(-0.198005744101)*x_ref[0]+(0.503568797262)*x_ref[1]**o+(0.0288809726497)*x_ref[1]+(-0.765127098183)*x_ref[2]**o+(0.409803857723)*x_ref[2]
        ref[2,2,1,0]=(-0.712272018295)*x_ref[0]**o+(-0.62562142063)*x_ref[0]+(-0.579501949719)*x_ref[1]**o+(0.628539275236)*x_ref[1]+(0.0200945124964)*x_ref[2]**o+(0.175966024769)*x_ref[2]
        ref[2,2,1,1]=(-0.798244445519)*x_ref[0]**o+(-0.376440962535)*x_ref[0]+(-0.134467401497)*x_ref[1]**o+(0.877538294477)*x_ref[1]+(-0.484870741192)*x_ref[2]**o+(-0.224064469935)*x_ref[2]
        ref[2,2,2,0]=(0.200196926368)*x_ref[0]**o+(-0.528000867086)*x_ref[0]+(-0.708942342624)*x_ref[1]**o+(-0.591994969141)*x_ref[1]+(-0.990641592656)*x_ref[2]**o+(-0.875897347303)*x_ref[2]
        ref[2,2,2,1]=(0.949113190797)*x_ref[0]**o+(0.867855496786)*x_ref[0]+(-0.0314953438065)*x_ref[1]**o+(-0.371928428005)*x_ref[1]+(-0.526403191063)*x_ref[2]**o+(0.653049122326)*x_ref[2]
        ref[2,3,0,0]=(-0.353418327127)*x_ref[0]**o+(-0.418588306429)*x_ref[0]+(0.400872043269)*x_ref[1]**o+(0.141907924042)*x_ref[1]+(-0.999869579462)*x_ref[2]**o+(-0.410423954563)*x_ref[2]
        ref[2,3,0,1]=(-0.240390877307)*x_ref[0]**o+(0.0901879916115)*x_ref[0]+(-0.0734683209657)*x_ref[1]**o+(0.274836005886)*x_ref[1]+(-0.336375498181)*x_ref[2]**o+(-0.754776455221)*x_ref[2]
        ref[2,3,1,0]=(0.259928460822)*x_ref[0]**o+(-0.227630454708)*x_ref[0]+(0.725968066278)*x_ref[1]**o+(-0.545440695763)*x_ref[1]+(-0.0354926607815)*x_ref[2]**o+(0.365010532181)*x_ref[2]
        ref[2,3,1,1]=(-0.969799373398)*x_ref[0]**o+(0.514498233214)*x_ref[0]+(0.907450925318)*x_ref[1]**o+(-0.234993706406)*x_ref[1]+(0.206323910806)*x_ref[2]**o+(0.209364250369)*x_ref[2]
        ref[2,3,2,0]=(0.712905565941)*x_ref[0]**o+(-0.353322758263)*x_ref[0]+(0.816861181494)*x_ref[1]**o+(-0.362388243619)*x_ref[1]+(0.302543517607)*x_ref[2]**o+(-0.831280418852)*x_ref[2]
        ref[2,3,2,1]=(-0.386373711182)*x_ref[0]**o+(0.946914567792)*x_ref[0]+(-0.52795085887)*x_ref[1]**o+(0.86635245964)*x_ref[1]+(0.654261054178)*x_ref[2]**o+(0.452334882498)*x_ref[2]
        ref[2,4,0,0]=(0.374747033547)*x_ref[0]**o+(-0.3471578669)*x_ref[0]+(0.322217002639)*x_ref[1]**o+(0.92988068982)*x_ref[1]+(-0.642423472369)*x_ref[2]**o+(-0.759623225653)*x_ref[2]
        ref[2,4,0,1]=(-0.322255157841)*x_ref[0]**o+(-0.253806844569)*x_ref[0]+(-0.0688746199637)*x_ref[1]**o+(0.232840317719)*x_ref[1]+(-0.815623721026)*x_ref[2]**o+(0.43055485562)*x_ref[2]
        ref[2,4,1,0]=(-0.452320341493)*x_ref[0]**o+(0.755240168228)*x_ref[0]+(-0.892552611066)*x_ref[1]**o+(0.438275161458)*x_ref[1]+(0.441378369388)*x_ref[2]**o+(-0.870501114753)*x_ref[2]
        ref[2,4,1,1]=(0.496731258286)*x_ref[0]**o+(-0.499918335322)*x_ref[0]+(0.000476061942027)*x_ref[1]**o+(-0.857762621856)*x_ref[1]+(0.957433798133)*x_ref[2]**o+(-0.766450629361)*x_ref[2]
        ref[2,4,2,0]=(0.860379598562)*x_ref[0]**o+(-0.0816478389485)*x_ref[0]+(-0.438181864676)*x_ref[1]**o+(0.78937427676)*x_ref[1]+(-0.97940010041)*x_ref[2]**o+(0.583914605259)*x_ref[2]
        ref[2,4,2,1]=(0.23729037277)*x_ref[0]**o+(-0.365380141474)*x_ref[0]+(-0.471505632575)*x_ref[1]**o+(-0.168838288)*x_ref[1]+(-0.519847065791)*x_ref[2]**o+(-0.816100518668)*x_ref[2]
        ref[3,0,0,0]=(0.729921808661)*x_ref[0]**o+(0.732868255388)*x_ref[0]+(0.305434623795)*x_ref[1]**o+(-0.45402804883)*x_ref[1]+(0.604368334171)*x_ref[2]**o+(-0.179233711008)*x_ref[2]
        ref[3,0,0,1]=(-0.208916687176)*x_ref[0]**o+(-0.297916049539)*x_ref[0]+(0.854941564423)*x_ref[1]**o+(0.775355543915)*x_ref[1]+(-0.373382357688)*x_ref[2]**o+(0.792557314074)*x_ref[2]
        ref[3,0,1,0]=(0.606041025674)*x_ref[0]**o+(0.115814644681)*x_ref[0]+(-0.414470914985)*x_ref[1]**o+(-0.136577390174)*x_ref[1]+(0.672303216396)*x_ref[2]**o+(-0.479680752765)*x_ref[2]
        ref[3,0,1,1]=(-0.535633649471)*x_ref[0]**o+(0.133761944775)*x_ref[0]+(-0.980107507665)*x_ref[1]**o+(0.211561747114)*x_ref[1]+(-0.530648770004)*x_ref[2]**o+(0.417977629512)*x_ref[2]
        ref[3,0,2,0]=(-0.242797826282)*x_ref[0]**o+(-0.828691638823)*x_ref[0]+(-0.187401612749)*x_ref[1]**o+(-0.453821620107)*x_ref[1]+(-0.475718408605)*x_ref[2]**o+(0.683502702497)*x_ref[2]
        ref[3,0,2,1]=(-0.66640812951)*x_ref[0]**o+(-0.259184719546)*x_ref[0]+(0.463743732654)*x_ref[1]**o+(0.6746349156)*x_ref[1]+(-0.0532918236783)*x_ref[2]**o+(-0.714197225876)*x_ref[2]
        ref[3,1,0,0]=(-0.97693616823)*x_ref[0]**o+(-0.825560555757)*x_ref[0]+(0.118452690664)*x_ref[1]**o+(-0.125279874388)*x_ref[1]+(0.733803136335)*x_ref[2]**o+(-0.884902209956)*x_ref[2]
        ref[3,1,0,1]=(-0.652466190334)*x_ref[0]**o+(0.370106640243)*x_ref[0]+(0.0850861818306)*x_ref[1]**o+(-0.49744492012)*x_ref[1]+(-0.843498619618)*x_ref[2]**o+(-0.0480070808698)*x_ref[2]
        ref[3,1,1,0]=(-0.450526704276)*x_ref[0]**o+(0.430774951776)*x_ref[0]+(-0.956799540696)*x_ref[1]**o+(0.820090546747)*x_ref[1]+(0.831582198997)*x_ref[2]**o+(0.491042906394)*x_ref[2]
        ref[3,1,1,1]=(0.419717807665)*x_ref[0]**o+(0.320904534666)*x_ref[0]+(-0.483247938999)*x_ref[1]**o+(0.82174341564)*x_ref[1]+(-0.441719998974)*x_ref[2]**o+(0.691713322239)*x_ref[2]
        ref[3,1,2,0]=(0.255464093123)*x_ref[0]**o+(-0.256334953799)*x_ref[0]+(0.677105284128)*x_ref[1]**o+(-0.584438025212)*x_ref[1]+(0.505820591597)*x_ref[2]**o+(0.247943562937)*x_ref[2]
        ref[3,1,2,1]=(-0.495204107745)*x_ref[0]**o+(0.234426349601)*x_ref[0]+(0.736969108189)*x_ref[1]**o+(-0.369684409765)*x_ref[1]+(0.164303379282)*x_ref[2]**o+(-0.595757590519)*x_ref[2]
        ref[3,2,0,0]=(0.238639684591)*x_ref[0]**o+(0.568638143194)*x_ref[0]+(0.972646679042)*x_ref[1]**o+(-0.652648194795)*x_ref[1]+(-0.0530713046414)*x_ref[2]**o+(0.356043665396)*x_ref[2]
        ref[3,2,0,1]=(0.458539145106)*x_ref[0]**o+(0.111417037398)*x_ref[0]+(-0.135374868193)*x_ref[1]**o+(-0.535451685997)*x_ref[1]+(-0.820797762877)*x_ref[2]**o+(0.122239787933)*x_ref[2]
        ref[3,2,1,0]=(-0.323251425039)*x_ref[0]**o+(0.106998797206)*x_ref[0]+(-0.0671655760766)*x_ref[1]**o+(0.836342794315)*x_ref[1]+(0.34628715493)*x_ref[2]**o+(-0.11058327806)*x_ref[2]
        ref[3,2,1,1]=(0.181994203911)*x_ref[0]**o+(-0.710678453639)*x_ref[0]+(0.625606409786)*x_ref[1]**o+(-0.363170328543)*x_ref[1]+(-0.932998745422)*x_ref[2]**o+(-0.0411581284314)*x_ref[2]
        ref[3,2,2,0]=(0.710847104903)*x_ref[0]**o+(0.483971915285)*x_ref[0]+(0.42996482533)*x_ref[1]**o+(0.456817348253)*x_ref[1]+(0.844562058604)*x_ref[2]**o+(-0.469586014721)*x_ref[2]
        ref[3,2,2,1]=(-0.845771280837)*x_ref[0]**o+(-0.081384696494)*x_ref[0]+(0.71823377714)*x_ref[1]**o+(-0.846672143168)*x_ref[1]+(-0.139540918461)*x_ref[2]**o+(0.837646131987)*x_ref[2]
        ref[3,3,0,0]=(-0.958726792109)*x_ref[0]**o+(0.53196455268)*x_ref[0]+(-0.621810423658)*x_ref[1]**o+(-0.0352106197658)*x_ref[1]+(-0.636059207474)*x_ref[2]**o+(-0.0491118207063)*x_ref[2]
        ref[3,3,0,1]=(0.987836799279)*x_ref[0]**o+(0.41633576224)*x_ref[0]+(0.312021799176)*x_ref[1]**o+(-0.168836582938)*x_ref[1]+(-0.391725003735)*x_ref[2]**o+(-0.158000182269)*x_ref[2]
        ref[3,3,1,0]=(-0.280868116529)*x_ref[0]**o+(-0.687924000092)*x_ref[0]+(0.865840229722)*x_ref[1]**o+(-0.0823203022874)*x_ref[1]+(-0.857309899784)*x_ref[2]**o+(-0.0657473493059)*x_ref[2]
        ref[3,3,1,1]=(0.680496030567)*x_ref[0]**o+(0.159575720724)*x_ref[0]+(0.0531895122046)*x_ref[1]**o+(0.359273896084)*x_ref[1]+(0.35084774502)*x_ref[2]**o+(0.0982220259531)*x_ref[2]
        ref[3,3,2,0]=(0.140872311111)*x_ref[0]**o+(-0.00845309046907)*x_ref[0]+(-0.845527142503)*x_ref[1]**o+(-0.67713103521)*x_ref[1]+(0.972054187828)*x_ref[2]**o+(-0.800664963509)*x_ref[2]
        ref[3,3,2,1]=(-0.490866666782)*x_ref[0]**o+(-0.0988939523952)*x_ref[0]+(0.236210969508)*x_ref[1]**o+(-0.672617589992)*x_ref[1]+(-0.640558724403)*x_ref[2]**o+(-0.203125267713)*x_ref[2]
        ref[3,4,0,0]=(0.318438839683)*x_ref[0]**o+(0.0864322377812)*x_ref[0]+(-0.371419701638)*x_ref[1]**o+(-0.311583139142)*x_ref[1]+(-0.0533111244011)*x_ref[2]**o+(0.841240973423)*x_ref[2]
        ref[3,4,0,1]=(-0.573110771334)*x_ref[0]**o+(0.703852077914)*x_ref[0]+(0.613309217217)*x_ref[1]**o+(0.247223193846)*x_ref[1]+(0.618952485205)*x_ref[2]**o+(-0.395998308184)*x_ref[2]
        ref[3,4,1,0]=(0.512896938934)*x_ref[0]**o+(0.848653171505)*x_ref[0]+(-0.944903687062)*x_ref[1]**o+(0.179782177607)*x_ref[1]+(-0.0287288996538)*x_ref[2]**o+(-0.385519118293)*x_ref[2]
        ref[3,4,1,1]=(0.234190647718)*x_ref[0]**o+(0.445478221419)*x_ref[0]+(0.588931676434)*x_ref[1]**o+(0.140788139998)*x_ref[1]+(-0.761952359657)*x_ref[2]**o+(0.92377426359)*x_ref[2]
        ref[3,4,2,0]=(-0.836038966171)*x_ref[0]**o+(-0.772990045933)*x_ref[0]+(0.0622128209877)*x_ref[1]**o+(-0.73733496141)*x_ref[1]+(0.0234446925161)*x_ref[2]**o+(-0.682234213315)*x_ref[2]
        ref[3,4,2,1]=(0.250498084214)*x_ref[0]**o+(0.539088356159)*x_ref[0]+(-0.691842680176)*x_ref[1]**o+(0.710040802953)*x_ref[1]+(-0.43691551501)*x_ref[2]**o+(0.710668600989)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.70941034293)*x[0]**o+(-0.988234004949)*x[0]+(0.440342659022)*x[1]**o+(0.221444369839)*x[1]
        ref=(0.70941034293)*x_ref[0]**o+(-0.988234004949)*x_ref[0]+(0.440342659022)*x_ref[1]**o+(0.221444369839)*x_ref[1]
      else:
        arg=(0.720798141776)*x[0]**o+(0.841475017012)*x[0]+(-0.76285313232)*x[1]**o+(-0.214354625268)*x[1]+(-0.887218533751)*x[2]**o+(-0.066368425118)*x[2]
        ref=(0.720798141776)*x_ref[0]**o+(0.841475017012)*x_ref[0]+(-0.76285313232)*x_ref[1]**o+(-0.214354625268)*x_ref[1]+(-0.887218533751)*x_ref[2]**o+(-0.066368425118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.940260123916)*x[0]**o+(-0.977192570293)*x[0]+(-0.0910686020625)*x[1]**o+(0.464205016172)*x[1]
        arg[1]=(-0.634146043992)*x[0]**o+(0.00438663094072)*x[0]+(0.905996907036)*x[1]**o+(0.175291641923)*x[1]
        ref[0]=(-0.940260123916)*x_ref[0]**o+(-0.977192570293)*x_ref[0]+(-0.0910686020625)*x_ref[1]**o+(0.464205016172)*x_ref[1]
        ref[1]=(-0.634146043992)*x_ref[0]**o+(0.00438663094072)*x_ref[0]+(0.905996907036)*x_ref[1]**o+(0.175291641923)*x_ref[1]
      else:
        arg[0]=(0.0746302635373)*x[0]**o+(0.611568121627)*x[0]+(0.0392401043636)*x[1]**o+(0.846339953866)*x[1]+(0.775935477701)*x[2]**o+(0.309530355508)*x[2]
        arg[1]=(0.790143430408)*x[0]**o+(0.987665779965)*x[0]+(0.569313447766)*x[1]**o+(0.170690891282)*x[1]+(0.224015411231)*x[2]**o+(-0.670946134012)*x[2]
        ref[0]=(0.0746302635373)*x_ref[0]**o+(0.611568121627)*x_ref[0]+(0.0392401043636)*x_ref[1]**o+(0.846339953866)*x_ref[1]+(0.775935477701)*x_ref[2]**o+(0.309530355508)*x_ref[2]
        ref[1]=(0.790143430408)*x_ref[0]**o+(0.987665779965)*x_ref[0]+(0.569313447766)*x_ref[1]**o+(0.170690891282)*x_ref[1]+(0.224015411231)*x_ref[2]**o+(-0.670946134012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.339464191741)*x[0]**o+(0.898634031554)*x[0]+(-0.67217412133)*x[1]**o+(-0.732882671617)*x[1]
        arg[0,1]=(0.475031035618)*x[0]**o+(-0.712271710437)*x[0]+(-0.346485473024)*x[1]**o+(0.936873073538)*x[1]
        arg[0,2]=(0.139217050193)*x[0]**o+(0.532762047045)*x[0]+(-0.411674842781)*x[1]**o+(-0.423096144938)*x[1]
        arg[0,3]=(-0.146193322436)*x[0]**o+(-0.227693594412)*x[0]+(0.74715898621)*x[1]**o+(0.538402714773)*x[1]
        arg[0,4]=(0.233316268084)*x[0]**o+(-0.0155557244918)*x[0]+(0.588017790807)*x[1]**o+(-0.58604454457)*x[1]
        arg[1,0]=(0.30473687856)*x[0]**o+(0.933154007362)*x[0]+(-0.0797692178609)*x[1]**o+(0.270021035922)*x[1]
        arg[1,1]=(-0.495613786111)*x[0]**o+(0.0857849462525)*x[0]+(0.330048664047)*x[1]**o+(0.0442768259753)*x[1]
        arg[1,2]=(0.203830377049)*x[0]**o+(-0.891094217974)*x[0]+(-0.643947141174)*x[1]**o+(-0.564315919126)*x[1]
        arg[1,3]=(-0.783514920199)*x[0]**o+(0.356882300706)*x[0]+(-0.307634481801)*x[1]**o+(0.521370775315)*x[1]
        arg[1,4]=(0.418545793821)*x[0]**o+(0.574742804708)*x[0]+(-0.340992300811)*x[1]**o+(-0.892118088815)*x[1]
        arg[2,0]=(-0.361915184269)*x[0]**o+(-0.834492406999)*x[0]+(0.750344593844)*x[1]**o+(-0.516981733186)*x[1]
        arg[2,1]=(-0.0123811923638)*x[0]**o+(-0.9808055138)*x[0]+(-0.176978785352)*x[1]**o+(0.519804799089)*x[1]
        arg[2,2]=(-0.901764228726)*x[0]**o+(-0.610547400097)*x[0]+(0.384870544403)*x[1]**o+(0.0308096944571)*x[1]
        arg[2,3]=(-0.163626872732)*x[0]**o+(0.278523098981)*x[0]+(-0.0544994735633)*x[1]**o+(-0.319354318824)*x[1]
        arg[2,4]=(0.558959525812)*x[0]**o+(0.435157723929)*x[0]+(0.758922109245)*x[1]**o+(-0.702427033542)*x[1]
        arg[3,0]=(0.832800430277)*x[0]**o+(-0.709285644484)*x[0]+(-0.201572915312)*x[1]**o+(0.398294738552)*x[1]
        arg[3,1]=(-0.836258924261)*x[0]**o+(-0.461452427039)*x[0]+(0.314485542425)*x[1]**o+(0.860807479051)*x[1]
        arg[3,2]=(-0.569661610862)*x[0]**o+(-0.46125190469)*x[0]+(0.807311492846)*x[1]**o+(0.72312107176)*x[1]
        arg[3,3]=(0.842126813179)*x[0]**o+(-0.714991954637)*x[0]+(-0.853100178283)*x[1]**o+(-0.644363753291)*x[1]
        arg[3,4]=(-0.0480369646571)*x[0]**o+(-0.330580215997)*x[0]+(0.609757645513)*x[1]**o+(0.915764017541)*x[1]
        ref[0,0]=(0.339464191741)*x_ref[0]**o+(0.898634031554)*x_ref[0]+(-0.67217412133)*x_ref[1]**o+(-0.732882671617)*x_ref[1]
        ref[0,1]=(0.475031035618)*x_ref[0]**o+(-0.712271710437)*x_ref[0]+(-0.346485473024)*x_ref[1]**o+(0.936873073538)*x_ref[1]
        ref[0,2]=(0.139217050193)*x_ref[0]**o+(0.532762047045)*x_ref[0]+(-0.411674842781)*x_ref[1]**o+(-0.423096144938)*x_ref[1]
        ref[0,3]=(-0.146193322436)*x_ref[0]**o+(-0.227693594412)*x_ref[0]+(0.74715898621)*x_ref[1]**o+(0.538402714773)*x_ref[1]
        ref[0,4]=(0.233316268084)*x_ref[0]**o+(-0.0155557244918)*x_ref[0]+(0.588017790807)*x_ref[1]**o+(-0.58604454457)*x_ref[1]
        ref[1,0]=(0.30473687856)*x_ref[0]**o+(0.933154007362)*x_ref[0]+(-0.0797692178609)*x_ref[1]**o+(0.270021035922)*x_ref[1]
        ref[1,1]=(-0.495613786111)*x_ref[0]**o+(0.0857849462525)*x_ref[0]+(0.330048664047)*x_ref[1]**o+(0.0442768259753)*x_ref[1]
        ref[1,2]=(0.203830377049)*x_ref[0]**o+(-0.891094217974)*x_ref[0]+(-0.643947141174)*x_ref[1]**o+(-0.564315919126)*x_ref[1]
        ref[1,3]=(-0.783514920199)*x_ref[0]**o+(0.356882300706)*x_ref[0]+(-0.307634481801)*x_ref[1]**o+(0.521370775315)*x_ref[1]
        ref[1,4]=(0.418545793821)*x_ref[0]**o+(0.574742804708)*x_ref[0]+(-0.340992300811)*x_ref[1]**o+(-0.892118088815)*x_ref[1]
        ref[2,0]=(-0.361915184269)*x_ref[0]**o+(-0.834492406999)*x_ref[0]+(0.750344593844)*x_ref[1]**o+(-0.516981733186)*x_ref[1]
        ref[2,1]=(-0.0123811923638)*x_ref[0]**o+(-0.9808055138)*x_ref[0]+(-0.176978785352)*x_ref[1]**o+(0.519804799089)*x_ref[1]
        ref[2,2]=(-0.901764228726)*x_ref[0]**o+(-0.610547400097)*x_ref[0]+(0.384870544403)*x_ref[1]**o+(0.0308096944571)*x_ref[1]
        ref[2,3]=(-0.163626872732)*x_ref[0]**o+(0.278523098981)*x_ref[0]+(-0.0544994735633)*x_ref[1]**o+(-0.319354318824)*x_ref[1]
        ref[2,4]=(0.558959525812)*x_ref[0]**o+(0.435157723929)*x_ref[0]+(0.758922109245)*x_ref[1]**o+(-0.702427033542)*x_ref[1]
        ref[3,0]=(0.832800430277)*x_ref[0]**o+(-0.709285644484)*x_ref[0]+(-0.201572915312)*x_ref[1]**o+(0.398294738552)*x_ref[1]
        ref[3,1]=(-0.836258924261)*x_ref[0]**o+(-0.461452427039)*x_ref[0]+(0.314485542425)*x_ref[1]**o+(0.860807479051)*x_ref[1]
        ref[3,2]=(-0.569661610862)*x_ref[0]**o+(-0.46125190469)*x_ref[0]+(0.807311492846)*x_ref[1]**o+(0.72312107176)*x_ref[1]
        ref[3,3]=(0.842126813179)*x_ref[0]**o+(-0.714991954637)*x_ref[0]+(-0.853100178283)*x_ref[1]**o+(-0.644363753291)*x_ref[1]
        ref[3,4]=(-0.0480369646571)*x_ref[0]**o+(-0.330580215997)*x_ref[0]+(0.609757645513)*x_ref[1]**o+(0.915764017541)*x_ref[1]
      else:
        arg[0,0]=(-0.256088293513)*x[0]**o+(0.980138068946)*x[0]+(0.375373123738)*x[1]**o+(0.118268751779)*x[1]+(-0.848800006681)*x[2]**o+(0.861808952535)*x[2]
        arg[0,1]=(-0.0658163161725)*x[0]**o+(-0.21582867866)*x[0]+(0.99787757796)*x[1]**o+(0.0175357022646)*x[1]+(-0.305438242469)*x[2]**o+(0.114385992332)*x[2]
        arg[0,2]=(-0.296948302927)*x[0]**o+(0.146852900649)*x[0]+(0.386910961544)*x[1]**o+(0.353651457394)*x[1]+(0.420964895802)*x[2]**o+(0.71145244268)*x[2]
        arg[0,3]=(0.61102494979)*x[0]**o+(-0.316055853885)*x[0]+(-0.117200390837)*x[1]**o+(-0.808630441349)*x[1]+(0.0914828526025)*x[2]**o+(0.305837943839)*x[2]
        arg[0,4]=(-0.22863340097)*x[0]**o+(0.398083421617)*x[0]+(-0.39971917463)*x[1]**o+(-0.0307312416932)*x[1]+(0.510443863383)*x[2]**o+(-0.129495846726)*x[2]
        arg[1,0]=(0.423968874298)*x[0]**o+(0.371480448801)*x[0]+(-0.934565508125)*x[1]**o+(0.890972208341)*x[1]+(-0.169067233411)*x[2]**o+(-0.757264272484)*x[2]
        arg[1,1]=(0.999152573666)*x[0]**o+(-0.659676649566)*x[0]+(0.993153643567)*x[1]**o+(0.350141892848)*x[1]+(-0.493784056536)*x[2]**o+(-0.966494205893)*x[2]
        arg[1,2]=(-0.141337889298)*x[0]**o+(0.689935513838)*x[0]+(0.335444607712)*x[1]**o+(-0.733351283662)*x[1]+(0.323775887172)*x[2]**o+(0.314099156544)*x[2]
        arg[1,3]=(-0.157132412118)*x[0]**o+(0.602468153313)*x[0]+(-0.618134886591)*x[1]**o+(-0.404425578444)*x[1]+(-0.5069585924)*x[2]**o+(0.616829810596)*x[2]
        arg[1,4]=(-0.107902610028)*x[0]**o+(0.664787826515)*x[0]+(-0.19940280262)*x[1]**o+(0.461794954077)*x[1]+(-0.0516671817578)*x[2]**o+(0.426165289622)*x[2]
        arg[2,0]=(-0.568061237616)*x[0]**o+(-0.744340542576)*x[0]+(-0.0125924903194)*x[1]**o+(0.710153866172)*x[1]+(0.979871553596)*x[2]**o+(0.945862455036)*x[2]
        arg[2,1]=(0.201599120538)*x[0]**o+(-0.11057289344)*x[0]+(0.522031147141)*x[1]**o+(0.194694594101)*x[1]+(-0.771997994079)*x[2]**o+(-0.513574957413)*x[2]
        arg[2,2]=(0.401979025382)*x[0]**o+(-0.743641669014)*x[0]+(-0.796071557874)*x[1]**o+(-0.352651095657)*x[1]+(-0.564117536877)*x[2]**o+(0.952729715875)*x[2]
        arg[2,3]=(-0.526162893019)*x[0]**o+(0.691019471759)*x[0]+(-0.68646084088)*x[1]**o+(0.839630438544)*x[1]+(-0.491997512287)*x[2]**o+(-0.405069600537)*x[2]
        arg[2,4]=(0.117664099703)*x[0]**o+(-0.807496546596)*x[0]+(0.607810703702)*x[1]**o+(-0.696118154009)*x[1]+(-0.156011683803)*x[2]**o+(-0.26100420077)*x[2]
        arg[3,0]=(0.280066174527)*x[0]**o+(-0.225940135983)*x[0]+(-0.338716163489)*x[1]**o+(0.628405669096)*x[1]+(-0.890543405425)*x[2]**o+(-0.142082524497)*x[2]
        arg[3,1]=(-0.659241351226)*x[0]**o+(0.275483386126)*x[0]+(-0.318352239058)*x[1]**o+(0.168536648793)*x[1]+(0.539555610218)*x[2]**o+(-0.95280560435)*x[2]
        arg[3,2]=(0.673861565841)*x[0]**o+(-0.0303873544723)*x[0]+(-0.638241094445)*x[1]**o+(0.664488908579)*x[1]+(-0.789484082135)*x[2]**o+(-0.982437408359)*x[2]
        arg[3,3]=(-0.902346309293)*x[0]**o+(0.818106160371)*x[0]+(-0.410074078135)*x[1]**o+(0.977554592344)*x[1]+(-0.00195907247124)*x[2]**o+(0.94762151647)*x[2]
        arg[3,4]=(0.27272813538)*x[0]**o+(0.0563423990113)*x[0]+(0.254612551508)*x[1]**o+(-0.901948547799)*x[1]+(0.263613075865)*x[2]**o+(0.879691314116)*x[2]
        ref[0,0]=(-0.256088293513)*x_ref[0]**o+(0.980138068946)*x_ref[0]+(0.375373123738)*x_ref[1]**o+(0.118268751779)*x_ref[1]+(-0.848800006681)*x_ref[2]**o+(0.861808952535)*x_ref[2]
        ref[0,1]=(-0.0658163161725)*x_ref[0]**o+(-0.21582867866)*x_ref[0]+(0.99787757796)*x_ref[1]**o+(0.0175357022646)*x_ref[1]+(-0.305438242469)*x_ref[2]**o+(0.114385992332)*x_ref[2]
        ref[0,2]=(-0.296948302927)*x_ref[0]**o+(0.146852900649)*x_ref[0]+(0.386910961544)*x_ref[1]**o+(0.353651457394)*x_ref[1]+(0.420964895802)*x_ref[2]**o+(0.71145244268)*x_ref[2]
        ref[0,3]=(0.61102494979)*x_ref[0]**o+(-0.316055853885)*x_ref[0]+(-0.117200390837)*x_ref[1]**o+(-0.808630441349)*x_ref[1]+(0.0914828526025)*x_ref[2]**o+(0.305837943839)*x_ref[2]
        ref[0,4]=(-0.22863340097)*x_ref[0]**o+(0.398083421617)*x_ref[0]+(-0.39971917463)*x_ref[1]**o+(-0.0307312416932)*x_ref[1]+(0.510443863383)*x_ref[2]**o+(-0.129495846726)*x_ref[2]
        ref[1,0]=(0.423968874298)*x_ref[0]**o+(0.371480448801)*x_ref[0]+(-0.934565508125)*x_ref[1]**o+(0.890972208341)*x_ref[1]+(-0.169067233411)*x_ref[2]**o+(-0.757264272484)*x_ref[2]
        ref[1,1]=(0.999152573666)*x_ref[0]**o+(-0.659676649566)*x_ref[0]+(0.993153643567)*x_ref[1]**o+(0.350141892848)*x_ref[1]+(-0.493784056536)*x_ref[2]**o+(-0.966494205893)*x_ref[2]
        ref[1,2]=(-0.141337889298)*x_ref[0]**o+(0.689935513838)*x_ref[0]+(0.335444607712)*x_ref[1]**o+(-0.733351283662)*x_ref[1]+(0.323775887172)*x_ref[2]**o+(0.314099156544)*x_ref[2]
        ref[1,3]=(-0.157132412118)*x_ref[0]**o+(0.602468153313)*x_ref[0]+(-0.618134886591)*x_ref[1]**o+(-0.404425578444)*x_ref[1]+(-0.5069585924)*x_ref[2]**o+(0.616829810596)*x_ref[2]
        ref[1,4]=(-0.107902610028)*x_ref[0]**o+(0.664787826515)*x_ref[0]+(-0.19940280262)*x_ref[1]**o+(0.461794954077)*x_ref[1]+(-0.0516671817578)*x_ref[2]**o+(0.426165289622)*x_ref[2]
        ref[2,0]=(-0.568061237616)*x_ref[0]**o+(-0.744340542576)*x_ref[0]+(-0.0125924903194)*x_ref[1]**o+(0.710153866172)*x_ref[1]+(0.979871553596)*x_ref[2]**o+(0.945862455036)*x_ref[2]
        ref[2,1]=(0.201599120538)*x_ref[0]**o+(-0.11057289344)*x_ref[0]+(0.522031147141)*x_ref[1]**o+(0.194694594101)*x_ref[1]+(-0.771997994079)*x_ref[2]**o+(-0.513574957413)*x_ref[2]
        ref[2,2]=(0.401979025382)*x_ref[0]**o+(-0.743641669014)*x_ref[0]+(-0.796071557874)*x_ref[1]**o+(-0.352651095657)*x_ref[1]+(-0.564117536877)*x_ref[2]**o+(0.952729715875)*x_ref[2]
        ref[2,3]=(-0.526162893019)*x_ref[0]**o+(0.691019471759)*x_ref[0]+(-0.68646084088)*x_ref[1]**o+(0.839630438544)*x_ref[1]+(-0.491997512287)*x_ref[2]**o+(-0.405069600537)*x_ref[2]
        ref[2,4]=(0.117664099703)*x_ref[0]**o+(-0.807496546596)*x_ref[0]+(0.607810703702)*x_ref[1]**o+(-0.696118154009)*x_ref[1]+(-0.156011683803)*x_ref[2]**o+(-0.26100420077)*x_ref[2]
        ref[3,0]=(0.280066174527)*x_ref[0]**o+(-0.225940135983)*x_ref[0]+(-0.338716163489)*x_ref[1]**o+(0.628405669096)*x_ref[1]+(-0.890543405425)*x_ref[2]**o+(-0.142082524497)*x_ref[2]
        ref[3,1]=(-0.659241351226)*x_ref[0]**o+(0.275483386126)*x_ref[0]+(-0.318352239058)*x_ref[1]**o+(0.168536648793)*x_ref[1]+(0.539555610218)*x_ref[2]**o+(-0.95280560435)*x_ref[2]
        ref[3,2]=(0.673861565841)*x_ref[0]**o+(-0.0303873544723)*x_ref[0]+(-0.638241094445)*x_ref[1]**o+(0.664488908579)*x_ref[1]+(-0.789484082135)*x_ref[2]**o+(-0.982437408359)*x_ref[2]
        ref[3,3]=(-0.902346309293)*x_ref[0]**o+(0.818106160371)*x_ref[0]+(-0.410074078135)*x_ref[1]**o+(0.977554592344)*x_ref[1]+(-0.00195907247124)*x_ref[2]**o+(0.94762151647)*x_ref[2]
        ref[3,4]=(0.27272813538)*x_ref[0]**o+(0.0563423990113)*x_ref[0]+(0.254612551508)*x_ref[1]**o+(-0.901948547799)*x_ref[1]+(0.263613075865)*x_ref[2]**o+(0.879691314116)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.023497804886)*x[0]**o+(0.934780781287)*x[0]+(0.488935915319)*x[1]**o+(-0.843152487959)*x[1]
        arg[0,0,1]=(0.074870516034)*x[0]**o+(-0.515115482624)*x[0]+(-0.582499912837)*x[1]**o+(-0.148911376041)*x[1]
        arg[0,1,0]=(0.0364775990427)*x[0]**o+(-0.258379408121)*x[0]+(0.109704222415)*x[1]**o+(-0.0149203380717)*x[1]
        arg[0,1,1]=(0.749804547117)*x[0]**o+(-0.65162531688)*x[0]+(-0.529952602569)*x[1]**o+(0.901735743956)*x[1]
        arg[1,0,0]=(0.74941277085)*x[0]**o+(-0.878903916612)*x[0]+(0.866605680273)*x[1]**o+(0.182572622995)*x[1]
        arg[1,0,1]=(-0.787476915117)*x[0]**o+(0.945314823283)*x[0]+(0.165599597226)*x[1]**o+(0.733712603726)*x[1]
        arg[1,1,0]=(0.696240085313)*x[0]**o+(0.775501454572)*x[0]+(0.225131553571)*x[1]**o+(-0.75288780057)*x[1]
        arg[1,1,1]=(-0.352416785284)*x[0]**o+(-0.681447081047)*x[0]+(0.164273429957)*x[1]**o+(-0.0977235074729)*x[1]
        arg[2,0,0]=(-0.484651384894)*x[0]**o+(0.162048679786)*x[0]+(0.433731200543)*x[1]**o+(-0.2255104779)*x[1]
        arg[2,0,1]=(0.230113349848)*x[0]**o+(0.674218493869)*x[0]+(0.124852942645)*x[1]**o+(0.361278270173)*x[1]
        arg[2,1,0]=(-0.604246197352)*x[0]**o+(0.26763755318)*x[0]+(-0.821482774631)*x[1]**o+(-0.821813801109)*x[1]
        arg[2,1,1]=(0.71669276429)*x[0]**o+(0.861782230871)*x[0]+(-0.837503788825)*x[1]**o+(0.0437253153979)*x[1]
        arg[3,0,0]=(-0.0941343724431)*x[0]**o+(0.376175369776)*x[0]+(-0.455752787585)*x[1]**o+(0.777480094583)*x[1]
        arg[3,0,1]=(-0.730897141543)*x[0]**o+(0.750046665085)*x[0]+(0.855742434657)*x[1]**o+(0.941826664412)*x[1]
        arg[3,1,0]=(-0.788757907886)*x[0]**o+(-0.0255067018415)*x[0]+(-0.875292088318)*x[1]**o+(0.401677848256)*x[1]
        arg[3,1,1]=(-0.980570910972)*x[0]**o+(0.936108649845)*x[0]+(-0.0463362409126)*x[1]**o+(0.630408734656)*x[1]
        arg[4,0,0]=(-0.72642075391)*x[0]**o+(0.283788406364)*x[0]+(-0.119645741265)*x[1]**o+(0.66948892422)*x[1]
        arg[4,0,1]=(0.178235706017)*x[0]**o+(0.693466107843)*x[0]+(-0.585602683437)*x[1]**o+(-0.493386713436)*x[1]
        arg[4,1,0]=(0.348039277129)*x[0]**o+(-0.806776084421)*x[0]+(-0.73591780841)*x[1]**o+(-0.310246367505)*x[1]
        arg[4,1,1]=(0.993294877418)*x[0]**o+(-0.683114302069)*x[0]+(0.0679022755308)*x[1]**o+(0.341981988936)*x[1]
        arg[5,0,0]=(-0.807380701959)*x[0]**o+(-0.78936779476)*x[0]+(0.142089012614)*x[1]**o+(-0.129335524031)*x[1]
        arg[5,0,1]=(-0.855714388031)*x[0]**o+(0.00825196217877)*x[0]+(0.310039590415)*x[1]**o+(-0.242861862397)*x[1]
        arg[5,1,0]=(-0.0908183346277)*x[0]**o+(-0.941876543235)*x[0]+(0.981459557493)*x[1]**o+(0.559563798858)*x[1]
        arg[5,1,1]=(-0.852223567371)*x[0]**o+(-0.272541886199)*x[0]+(-0.873908247032)*x[1]**o+(-0.387571769068)*x[1]
        ref[0,0,0]=(-0.023497804886)*x_ref[0]**o+(0.934780781287)*x_ref[0]+(0.488935915319)*x_ref[1]**o+(-0.843152487959)*x_ref[1]
        ref[0,0,1]=(0.074870516034)*x_ref[0]**o+(-0.515115482624)*x_ref[0]+(-0.582499912837)*x_ref[1]**o+(-0.148911376041)*x_ref[1]
        ref[0,1,0]=(0.0364775990427)*x_ref[0]**o+(-0.258379408121)*x_ref[0]+(0.109704222415)*x_ref[1]**o+(-0.0149203380717)*x_ref[1]
        ref[0,1,1]=(0.749804547117)*x_ref[0]**o+(-0.65162531688)*x_ref[0]+(-0.529952602569)*x_ref[1]**o+(0.901735743956)*x_ref[1]
        ref[1,0,0]=(0.74941277085)*x_ref[0]**o+(-0.878903916612)*x_ref[0]+(0.866605680273)*x_ref[1]**o+(0.182572622995)*x_ref[1]
        ref[1,0,1]=(-0.787476915117)*x_ref[0]**o+(0.945314823283)*x_ref[0]+(0.165599597226)*x_ref[1]**o+(0.733712603726)*x_ref[1]
        ref[1,1,0]=(0.696240085313)*x_ref[0]**o+(0.775501454572)*x_ref[0]+(0.225131553571)*x_ref[1]**o+(-0.75288780057)*x_ref[1]
        ref[1,1,1]=(-0.352416785284)*x_ref[0]**o+(-0.681447081047)*x_ref[0]+(0.164273429957)*x_ref[1]**o+(-0.0977235074729)*x_ref[1]
        ref[2,0,0]=(-0.484651384894)*x_ref[0]**o+(0.162048679786)*x_ref[0]+(0.433731200543)*x_ref[1]**o+(-0.2255104779)*x_ref[1]
        ref[2,0,1]=(0.230113349848)*x_ref[0]**o+(0.674218493869)*x_ref[0]+(0.124852942645)*x_ref[1]**o+(0.361278270173)*x_ref[1]
        ref[2,1,0]=(-0.604246197352)*x_ref[0]**o+(0.26763755318)*x_ref[0]+(-0.821482774631)*x_ref[1]**o+(-0.821813801109)*x_ref[1]
        ref[2,1,1]=(0.71669276429)*x_ref[0]**o+(0.861782230871)*x_ref[0]+(-0.837503788825)*x_ref[1]**o+(0.0437253153979)*x_ref[1]
        ref[3,0,0]=(-0.0941343724431)*x_ref[0]**o+(0.376175369776)*x_ref[0]+(-0.455752787585)*x_ref[1]**o+(0.777480094583)*x_ref[1]
        ref[3,0,1]=(-0.730897141543)*x_ref[0]**o+(0.750046665085)*x_ref[0]+(0.855742434657)*x_ref[1]**o+(0.941826664412)*x_ref[1]
        ref[3,1,0]=(-0.788757907886)*x_ref[0]**o+(-0.0255067018415)*x_ref[0]+(-0.875292088318)*x_ref[1]**o+(0.401677848256)*x_ref[1]
        ref[3,1,1]=(-0.980570910972)*x_ref[0]**o+(0.936108649845)*x_ref[0]+(-0.0463362409126)*x_ref[1]**o+(0.630408734656)*x_ref[1]
        ref[4,0,0]=(-0.72642075391)*x_ref[0]**o+(0.283788406364)*x_ref[0]+(-0.119645741265)*x_ref[1]**o+(0.66948892422)*x_ref[1]
        ref[4,0,1]=(0.178235706017)*x_ref[0]**o+(0.693466107843)*x_ref[0]+(-0.585602683437)*x_ref[1]**o+(-0.493386713436)*x_ref[1]
        ref[4,1,0]=(0.348039277129)*x_ref[0]**o+(-0.806776084421)*x_ref[0]+(-0.73591780841)*x_ref[1]**o+(-0.310246367505)*x_ref[1]
        ref[4,1,1]=(0.993294877418)*x_ref[0]**o+(-0.683114302069)*x_ref[0]+(0.0679022755308)*x_ref[1]**o+(0.341981988936)*x_ref[1]
        ref[5,0,0]=(-0.807380701959)*x_ref[0]**o+(-0.78936779476)*x_ref[0]+(0.142089012614)*x_ref[1]**o+(-0.129335524031)*x_ref[1]
        ref[5,0,1]=(-0.855714388031)*x_ref[0]**o+(0.00825196217877)*x_ref[0]+(0.310039590415)*x_ref[1]**o+(-0.242861862397)*x_ref[1]
        ref[5,1,0]=(-0.0908183346277)*x_ref[0]**o+(-0.941876543235)*x_ref[0]+(0.981459557493)*x_ref[1]**o+(0.559563798858)*x_ref[1]
        ref[5,1,1]=(-0.852223567371)*x_ref[0]**o+(-0.272541886199)*x_ref[0]+(-0.873908247032)*x_ref[1]**o+(-0.387571769068)*x_ref[1]
      else:
        arg[0,0,0]=(0.418981395334)*x[0]**o+(-0.720305875042)*x[0]+(0.876880439526)*x[1]**o+(-0.17785803746)*x[1]+(0.779797986429)*x[2]**o+(0.363827072746)*x[2]
        arg[0,0,1]=(0.159567966427)*x[0]**o+(-0.614151008569)*x[0]+(0.249982224851)*x[1]**o+(-0.119598774468)*x[1]+(-0.66603578755)*x[2]**o+(0.202190479552)*x[2]
        arg[0,1,0]=(0.337868020939)*x[0]**o+(-0.0828148711399)*x[0]+(-0.0631659329293)*x[1]**o+(-0.725100695213)*x[1]+(0.852068398846)*x[2]**o+(-0.951116521211)*x[2]
        arg[0,1,1]=(0.710866728186)*x[0]**o+(-0.517133891599)*x[0]+(0.320028321875)*x[1]**o+(0.0331233805016)*x[1]+(-0.0945382588101)*x[2]**o+(-0.795725457058)*x[2]
        arg[1,0,0]=(0.261503329866)*x[0]**o+(0.697062723662)*x[0]+(0.926550326146)*x[1]**o+(-0.318704885103)*x[1]+(-0.76904956799)*x[2]**o+(0.0750521033639)*x[2]
        arg[1,0,1]=(-0.594460213676)*x[0]**o+(0.356692399969)*x[0]+(-0.611558953982)*x[1]**o+(0.289469152887)*x[1]+(-0.73222413782)*x[2]**o+(-0.576883793051)*x[2]
        arg[1,1,0]=(-0.399324648161)*x[0]**o+(0.975428000926)*x[0]+(-0.889762441468)*x[1]**o+(-0.137573068244)*x[1]+(0.252276361744)*x[2]**o+(-0.554435968553)*x[2]
        arg[1,1,1]=(0.677933864695)*x[0]**o+(-0.19825134391)*x[0]+(0.556783296976)*x[1]**o+(0.0349255641106)*x[1]+(-0.387378575139)*x[2]**o+(0.889628545207)*x[2]
        arg[2,0,0]=(-0.839356850436)*x[0]**o+(0.0186947034157)*x[0]+(-0.264793716814)*x[1]**o+(-0.767468006197)*x[1]+(0.91661695419)*x[2]**o+(-0.374036622497)*x[2]
        arg[2,0,1]=(-0.030357536408)*x[0]**o+(0.57424762933)*x[0]+(-0.590115256932)*x[1]**o+(-0.176169776757)*x[1]+(-0.502845936823)*x[2]**o+(0.0399243413515)*x[2]
        arg[2,1,0]=(0.257324735469)*x[0]**o+(-0.686083724325)*x[0]+(-0.166609564696)*x[1]**o+(-0.132181292726)*x[1]+(0.10419146703)*x[2]**o+(-0.827571838114)*x[2]
        arg[2,1,1]=(0.339465982319)*x[0]**o+(0.766586466841)*x[0]+(0.277538823819)*x[1]**o+(0.222607512697)*x[1]+(-0.909105282607)*x[2]**o+(0.463729842225)*x[2]
        arg[3,0,0]=(-0.00132043337846)*x[0]**o+(0.670451981959)*x[0]+(-0.0122245380346)*x[1]**o+(-0.922535904926)*x[1]+(0.49486742)*x[2]**o+(0.0785274524546)*x[2]
        arg[3,0,1]=(0.0360876610851)*x[0]**o+(0.69319604661)*x[0]+(0.980826095079)*x[1]**o+(0.167689443923)*x[1]+(0.561634492527)*x[2]**o+(0.996948790413)*x[2]
        arg[3,1,0]=(-0.238705628819)*x[0]**o+(-0.915859436874)*x[0]+(0.942122040692)*x[1]**o+(0.971025407162)*x[1]+(-0.965945363654)*x[2]**o+(0.136537412741)*x[2]
        arg[3,1,1]=(-0.672314439537)*x[0]**o+(0.792647645097)*x[0]+(-0.432226224113)*x[1]**o+(0.424246857841)*x[1]+(-0.639842531517)*x[2]**o+(0.0570057507731)*x[2]
        arg[4,0,0]=(0.450831267081)*x[0]**o+(0.75992870254)*x[0]+(0.768779936214)*x[1]**o+(-0.0135947286596)*x[1]+(0.126345191973)*x[2]**o+(-0.619763308655)*x[2]
        arg[4,0,1]=(-0.325005948888)*x[0]**o+(-0.0655421631802)*x[0]+(0.607182269636)*x[1]**o+(-0.631044141835)*x[1]+(-0.356580309429)*x[2]**o+(-0.593198006905)*x[2]
        arg[4,1,0]=(0.304932527518)*x[0]**o+(-0.763995484499)*x[0]+(0.884855995183)*x[1]**o+(-0.42180322102)*x[1]+(0.605742059302)*x[2]**o+(0.448553789891)*x[2]
        arg[4,1,1]=(-0.787145437063)*x[0]**o+(-0.465220934558)*x[0]+(-0.370336760606)*x[1]**o+(-0.854850692759)*x[1]+(0.482896516171)*x[2]**o+(-0.551567519291)*x[2]
        arg[5,0,0]=(-0.537738873639)*x[0]**o+(-0.213268842545)*x[0]+(0.492208051415)*x[1]**o+(-0.609304634017)*x[1]+(-0.487753207718)*x[2]**o+(-0.70650655894)*x[2]
        arg[5,0,1]=(0.600359837292)*x[0]**o+(-0.604421599924)*x[0]+(-0.051356466466)*x[1]**o+(-0.168592544661)*x[1]+(0.789340570036)*x[2]**o+(0.910249026599)*x[2]
        arg[5,1,0]=(0.264506197816)*x[0]**o+(-0.711775430363)*x[0]+(0.384666624979)*x[1]**o+(0.0926714847986)*x[1]+(-0.551232557648)*x[2]**o+(0.758927448717)*x[2]
        arg[5,1,1]=(0.894998835449)*x[0]**o+(-0.466338210072)*x[0]+(-0.522475632898)*x[1]**o+(0.766949173671)*x[1]+(0.915063619607)*x[2]**o+(-0.270696846692)*x[2]
        ref[0,0,0]=(0.418981395334)*x_ref[0]**o+(-0.720305875042)*x_ref[0]+(0.876880439526)*x_ref[1]**o+(-0.17785803746)*x_ref[1]+(0.779797986429)*x_ref[2]**o+(0.363827072746)*x_ref[2]
        ref[0,0,1]=(0.159567966427)*x_ref[0]**o+(-0.614151008569)*x_ref[0]+(0.249982224851)*x_ref[1]**o+(-0.119598774468)*x_ref[1]+(-0.66603578755)*x_ref[2]**o+(0.202190479552)*x_ref[2]
        ref[0,1,0]=(0.337868020939)*x_ref[0]**o+(-0.0828148711399)*x_ref[0]+(-0.0631659329293)*x_ref[1]**o+(-0.725100695213)*x_ref[1]+(0.852068398846)*x_ref[2]**o+(-0.951116521211)*x_ref[2]
        ref[0,1,1]=(0.710866728186)*x_ref[0]**o+(-0.517133891599)*x_ref[0]+(0.320028321875)*x_ref[1]**o+(0.0331233805016)*x_ref[1]+(-0.0945382588101)*x_ref[2]**o+(-0.795725457058)*x_ref[2]
        ref[1,0,0]=(0.261503329866)*x_ref[0]**o+(0.697062723662)*x_ref[0]+(0.926550326146)*x_ref[1]**o+(-0.318704885103)*x_ref[1]+(-0.76904956799)*x_ref[2]**o+(0.0750521033639)*x_ref[2]
        ref[1,0,1]=(-0.594460213676)*x_ref[0]**o+(0.356692399969)*x_ref[0]+(-0.611558953982)*x_ref[1]**o+(0.289469152887)*x_ref[1]+(-0.73222413782)*x_ref[2]**o+(-0.576883793051)*x_ref[2]
        ref[1,1,0]=(-0.399324648161)*x_ref[0]**o+(0.975428000926)*x_ref[0]+(-0.889762441468)*x_ref[1]**o+(-0.137573068244)*x_ref[1]+(0.252276361744)*x_ref[2]**o+(-0.554435968553)*x_ref[2]
        ref[1,1,1]=(0.677933864695)*x_ref[0]**o+(-0.19825134391)*x_ref[0]+(0.556783296976)*x_ref[1]**o+(0.0349255641106)*x_ref[1]+(-0.387378575139)*x_ref[2]**o+(0.889628545207)*x_ref[2]
        ref[2,0,0]=(-0.839356850436)*x_ref[0]**o+(0.0186947034157)*x_ref[0]+(-0.264793716814)*x_ref[1]**o+(-0.767468006197)*x_ref[1]+(0.91661695419)*x_ref[2]**o+(-0.374036622497)*x_ref[2]
        ref[2,0,1]=(-0.030357536408)*x_ref[0]**o+(0.57424762933)*x_ref[0]+(-0.590115256932)*x_ref[1]**o+(-0.176169776757)*x_ref[1]+(-0.502845936823)*x_ref[2]**o+(0.0399243413515)*x_ref[2]
        ref[2,1,0]=(0.257324735469)*x_ref[0]**o+(-0.686083724325)*x_ref[0]+(-0.166609564696)*x_ref[1]**o+(-0.132181292726)*x_ref[1]+(0.10419146703)*x_ref[2]**o+(-0.827571838114)*x_ref[2]
        ref[2,1,1]=(0.339465982319)*x_ref[0]**o+(0.766586466841)*x_ref[0]+(0.277538823819)*x_ref[1]**o+(0.222607512697)*x_ref[1]+(-0.909105282607)*x_ref[2]**o+(0.463729842225)*x_ref[2]
        ref[3,0,0]=(-0.00132043337846)*x_ref[0]**o+(0.670451981959)*x_ref[0]+(-0.0122245380346)*x_ref[1]**o+(-0.922535904926)*x_ref[1]+(0.49486742)*x_ref[2]**o+(0.0785274524546)*x_ref[2]
        ref[3,0,1]=(0.0360876610851)*x_ref[0]**o+(0.69319604661)*x_ref[0]+(0.980826095079)*x_ref[1]**o+(0.167689443923)*x_ref[1]+(0.561634492527)*x_ref[2]**o+(0.996948790413)*x_ref[2]
        ref[3,1,0]=(-0.238705628819)*x_ref[0]**o+(-0.915859436874)*x_ref[0]+(0.942122040692)*x_ref[1]**o+(0.971025407162)*x_ref[1]+(-0.965945363654)*x_ref[2]**o+(0.136537412741)*x_ref[2]
        ref[3,1,1]=(-0.672314439537)*x_ref[0]**o+(0.792647645097)*x_ref[0]+(-0.432226224113)*x_ref[1]**o+(0.424246857841)*x_ref[1]+(-0.639842531517)*x_ref[2]**o+(0.0570057507731)*x_ref[2]
        ref[4,0,0]=(0.450831267081)*x_ref[0]**o+(0.75992870254)*x_ref[0]+(0.768779936214)*x_ref[1]**o+(-0.0135947286596)*x_ref[1]+(0.126345191973)*x_ref[2]**o+(-0.619763308655)*x_ref[2]
        ref[4,0,1]=(-0.325005948888)*x_ref[0]**o+(-0.0655421631802)*x_ref[0]+(0.607182269636)*x_ref[1]**o+(-0.631044141835)*x_ref[1]+(-0.356580309429)*x_ref[2]**o+(-0.593198006905)*x_ref[2]
        ref[4,1,0]=(0.304932527518)*x_ref[0]**o+(-0.763995484499)*x_ref[0]+(0.884855995183)*x_ref[1]**o+(-0.42180322102)*x_ref[1]+(0.605742059302)*x_ref[2]**o+(0.448553789891)*x_ref[2]
        ref[4,1,1]=(-0.787145437063)*x_ref[0]**o+(-0.465220934558)*x_ref[0]+(-0.370336760606)*x_ref[1]**o+(-0.854850692759)*x_ref[1]+(0.482896516171)*x_ref[2]**o+(-0.551567519291)*x_ref[2]
        ref[5,0,0]=(-0.537738873639)*x_ref[0]**o+(-0.213268842545)*x_ref[0]+(0.492208051415)*x_ref[1]**o+(-0.609304634017)*x_ref[1]+(-0.487753207718)*x_ref[2]**o+(-0.70650655894)*x_ref[2]
        ref[5,0,1]=(0.600359837292)*x_ref[0]**o+(-0.604421599924)*x_ref[0]+(-0.051356466466)*x_ref[1]**o+(-0.168592544661)*x_ref[1]+(0.789340570036)*x_ref[2]**o+(0.910249026599)*x_ref[2]
        ref[5,1,0]=(0.264506197816)*x_ref[0]**o+(-0.711775430363)*x_ref[0]+(0.384666624979)*x_ref[1]**o+(0.0926714847986)*x_ref[1]+(-0.551232557648)*x_ref[2]**o+(0.758927448717)*x_ref[2]
        ref[5,1,1]=(0.894998835449)*x_ref[0]**o+(-0.466338210072)*x_ref[0]+(-0.522475632898)*x_ref[1]**o+(0.766949173671)*x_ref[1]+(0.915063619607)*x_ref[2]**o+(-0.270696846692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.240887319278)*x[0]**o+(-0.192166587545)*x[0]+(0.0176974287409)*x[1]**o+(0.100761748824)*x[1]
        arg[0,0,0,1]=(-0.549649845775)*x[0]**o+(-0.293969245646)*x[0]+(-0.605885325133)*x[1]**o+(0.799558451173)*x[1]
        arg[0,0,1,0]=(-0.899658769582)*x[0]**o+(0.484524193987)*x[0]+(0.437485870351)*x[1]**o+(-0.220549034927)*x[1]
        arg[0,0,1,1]=(-0.146539961793)*x[0]**o+(-0.278343486929)*x[0]+(0.0913912456082)*x[1]**o+(-0.751028975123)*x[1]
        arg[0,0,2,0]=(0.806344701194)*x[0]**o+(0.0537872824561)*x[0]+(-0.70407828379)*x[1]**o+(-0.922963486601)*x[1]
        arg[0,0,2,1]=(0.307202902495)*x[0]**o+(0.301603506731)*x[0]+(0.232983743616)*x[1]**o+(-0.744267542759)*x[1]
        arg[0,1,0,0]=(-0.550616060824)*x[0]**o+(-0.661367750254)*x[0]+(-0.695905895864)*x[1]**o+(-0.788045922398)*x[1]
        arg[0,1,0,1]=(-0.934169134035)*x[0]**o+(-0.843861105521)*x[0]+(-0.237877254439)*x[1]**o+(-0.364956821348)*x[1]
        arg[0,1,1,0]=(0.0950675750026)*x[0]**o+(-0.815510640594)*x[0]+(0.72526035968)*x[1]**o+(0.898675190106)*x[1]
        arg[0,1,1,1]=(0.715609066776)*x[0]**o+(0.371314845206)*x[0]+(0.326878867244)*x[1]**o+(-0.468798950199)*x[1]
        arg[0,1,2,0]=(0.392175939547)*x[0]**o+(0.105332682674)*x[0]+(0.997967455477)*x[1]**o+(0.0322491822496)*x[1]
        arg[0,1,2,1]=(-0.804498399187)*x[0]**o+(0.819653744739)*x[0]+(0.609610573703)*x[1]**o+(-0.623299047794)*x[1]
        arg[0,2,0,0]=(0.993877917386)*x[0]**o+(0.815952897899)*x[0]+(-0.258432174486)*x[1]**o+(0.209324387308)*x[1]
        arg[0,2,0,1]=(0.395946889364)*x[0]**o+(-0.859500539908)*x[0]+(-0.68495734785)*x[1]**o+(0.791813946302)*x[1]
        arg[0,2,1,0]=(0.154347661108)*x[0]**o+(0.0873140744322)*x[0]+(-0.423935109956)*x[1]**o+(-0.754297239226)*x[1]
        arg[0,2,1,1]=(-0.623998419484)*x[0]**o+(0.487300130941)*x[0]+(-0.323613219784)*x[1]**o+(0.744877983729)*x[1]
        arg[0,2,2,0]=(0.515686374371)*x[0]**o+(0.512572153346)*x[0]+(-0.641245783272)*x[1]**o+(-0.390660158759)*x[1]
        arg[0,2,2,1]=(0.24398238139)*x[0]**o+(-0.172020271678)*x[0]+(0.980197812701)*x[1]**o+(0.336941934503)*x[1]
        arg[0,3,0,0]=(0.0154584447615)*x[0]**o+(-0.547048191238)*x[0]+(-0.555319223005)*x[1]**o+(0.0972794598106)*x[1]
        arg[0,3,0,1]=(0.44744389335)*x[0]**o+(0.475046337874)*x[0]+(0.876228982066)*x[1]**o+(-0.0636405241175)*x[1]
        arg[0,3,1,0]=(0.392948465298)*x[0]**o+(-0.341342654182)*x[0]+(0.747185683968)*x[1]**o+(-0.99981172166)*x[1]
        arg[0,3,1,1]=(-0.172355321621)*x[0]**o+(-0.531391484795)*x[0]+(0.451654274019)*x[1]**o+(0.0795298667915)*x[1]
        arg[0,3,2,0]=(0.0841642474674)*x[0]**o+(0.308494435164)*x[0]+(0.53181918181)*x[1]**o+(0.480258067959)*x[1]
        arg[0,3,2,1]=(-0.222903054864)*x[0]**o+(-0.0330655869754)*x[0]+(0.227482525015)*x[1]**o+(0.536136212347)*x[1]
        arg[0,4,0,0]=(0.574798381284)*x[0]**o+(0.181354466484)*x[0]+(0.103049514171)*x[1]**o+(0.156282818835)*x[1]
        arg[0,4,0,1]=(0.478314255289)*x[0]**o+(0.224797222592)*x[0]+(0.307668606272)*x[1]**o+(-0.785262182937)*x[1]
        arg[0,4,1,0]=(0.519356961182)*x[0]**o+(0.646200458344)*x[0]+(0.0522618783181)*x[1]**o+(0.702335307324)*x[1]
        arg[0,4,1,1]=(-0.168756029541)*x[0]**o+(-0.00618364750666)*x[0]+(0.472704616213)*x[1]**o+(0.974257451871)*x[1]
        arg[0,4,2,0]=(-0.279196869573)*x[0]**o+(-0.333083174091)*x[0]+(-0.980089574428)*x[1]**o+(0.405273709405)*x[1]
        arg[0,4,2,1]=(-0.199431380673)*x[0]**o+(0.823294899866)*x[0]+(-0.174125117685)*x[1]**o+(-0.730500493887)*x[1]
        arg[1,0,0,0]=(-0.577281489635)*x[0]**o+(0.231906139898)*x[0]+(0.537128708963)*x[1]**o+(0.363444540898)*x[1]
        arg[1,0,0,1]=(0.204108541882)*x[0]**o+(-0.217661714888)*x[0]+(0.963384293759)*x[1]**o+(-0.0817519801351)*x[1]
        arg[1,0,1,0]=(0.540681325439)*x[0]**o+(-0.496036960808)*x[0]+(0.421037689661)*x[1]**o+(0.19399086996)*x[1]
        arg[1,0,1,1]=(0.482938542126)*x[0]**o+(-0.0725027675209)*x[0]+(-0.441517962054)*x[1]**o+(0.787445860714)*x[1]
        arg[1,0,2,0]=(0.0378928111257)*x[0]**o+(0.451368848019)*x[0]+(-0.228668284485)*x[1]**o+(0.505927229432)*x[1]
        arg[1,0,2,1]=(-0.59833518872)*x[0]**o+(0.956068253354)*x[0]+(0.667905995469)*x[1]**o+(0.963273292214)*x[1]
        arg[1,1,0,0]=(0.901637865236)*x[0]**o+(-0.642678971748)*x[0]+(-0.603915019524)*x[1]**o+(-0.208004993919)*x[1]
        arg[1,1,0,1]=(-0.227162661052)*x[0]**o+(-0.5816434303)*x[0]+(0.274961293801)*x[1]**o+(-0.614001328621)*x[1]
        arg[1,1,1,0]=(0.078438078548)*x[0]**o+(0.570191242883)*x[0]+(0.0952878212484)*x[1]**o+(0.844460205497)*x[1]
        arg[1,1,1,1]=(-0.901445811786)*x[0]**o+(-0.996376775344)*x[0]+(0.707470471613)*x[1]**o+(-0.187517936674)*x[1]
        arg[1,1,2,0]=(-0.66891443067)*x[0]**o+(0.0462057239324)*x[0]+(-0.825820539128)*x[1]**o+(-0.766932836843)*x[1]
        arg[1,1,2,1]=(-0.851428828811)*x[0]**o+(0.504484422346)*x[0]+(0.121378315545)*x[1]**o+(0.903947012757)*x[1]
        arg[1,2,0,0]=(-0.393827449768)*x[0]**o+(-0.251827904608)*x[0]+(0.8320040882)*x[1]**o+(0.57595149855)*x[1]
        arg[1,2,0,1]=(0.638738110712)*x[0]**o+(-0.439962152539)*x[0]+(-0.977211078995)*x[1]**o+(0.0528475212451)*x[1]
        arg[1,2,1,0]=(0.824313588941)*x[0]**o+(-0.589968996823)*x[0]+(-0.512156794152)*x[1]**o+(0.142809794443)*x[1]
        arg[1,2,1,1]=(-0.31789333147)*x[0]**o+(0.598839896949)*x[0]+(0.638319398651)*x[1]**o+(0.353625999317)*x[1]
        arg[1,2,2,0]=(-0.720472365012)*x[0]**o+(0.896538030051)*x[0]+(-0.224639169972)*x[1]**o+(-0.898299360738)*x[1]
        arg[1,2,2,1]=(0.123077707167)*x[0]**o+(0.15922851038)*x[0]+(0.678984783843)*x[1]**o+(-0.778188269852)*x[1]
        arg[1,3,0,0]=(0.344256301209)*x[0]**o+(0.884367738289)*x[0]+(0.00556636219289)*x[1]**o+(0.684361331195)*x[1]
        arg[1,3,0,1]=(-0.812245131065)*x[0]**o+(0.971365671228)*x[0]+(0.0696053239569)*x[1]**o+(-0.300266235138)*x[1]
        arg[1,3,1,0]=(0.470254588058)*x[0]**o+(0.0946825509913)*x[0]+(-0.415619255025)*x[1]**o+(-0.27381652858)*x[1]
        arg[1,3,1,1]=(-0.84555138241)*x[0]**o+(0.873510832091)*x[0]+(-0.138010338152)*x[1]**o+(-0.982274515316)*x[1]
        arg[1,3,2,0]=(0.695614136148)*x[0]**o+(-0.192588082771)*x[0]+(-0.21579127791)*x[1]**o+(-0.0335939736388)*x[1]
        arg[1,3,2,1]=(-0.633363463896)*x[0]**o+(-0.958320549993)*x[0]+(0.197331067481)*x[1]**o+(0.315089888409)*x[1]
        arg[1,4,0,0]=(-0.211531219945)*x[0]**o+(-0.399700721799)*x[0]+(0.578927638839)*x[1]**o+(-0.389034976195)*x[1]
        arg[1,4,0,1]=(-0.846399571893)*x[0]**o+(-0.952904755903)*x[0]+(-0.858399802693)*x[1]**o+(-0.23264354907)*x[1]
        arg[1,4,1,0]=(0.255822441586)*x[0]**o+(0.896237710393)*x[0]+(-0.90248008271)*x[1]**o+(0.654804848244)*x[1]
        arg[1,4,1,1]=(-0.112475508163)*x[0]**o+(-0.861697204453)*x[0]+(-0.782581931187)*x[1]**o+(-0.129345341581)*x[1]
        arg[1,4,2,0]=(0.690095687033)*x[0]**o+(0.729317334019)*x[0]+(-0.00961483612288)*x[1]**o+(-0.288320551243)*x[1]
        arg[1,4,2,1]=(-0.648146433517)*x[0]**o+(-0.301246601263)*x[0]+(-0.490119612994)*x[1]**o+(0.231186845615)*x[1]
        arg[2,0,0,0]=(-0.645560891595)*x[0]**o+(-0.889067544988)*x[0]+(-0.193353414221)*x[1]**o+(-0.726988479344)*x[1]
        arg[2,0,0,1]=(-0.083614738484)*x[0]**o+(-0.822530927241)*x[0]+(0.660338934241)*x[1]**o+(0.0383821036508)*x[1]
        arg[2,0,1,0]=(-0.132250736664)*x[0]**o+(-0.571474370931)*x[0]+(-0.755931100342)*x[1]**o+(-0.145196865115)*x[1]
        arg[2,0,1,1]=(0.545219196727)*x[0]**o+(-0.846368787913)*x[0]+(-0.845706160842)*x[1]**o+(0.550703610989)*x[1]
        arg[2,0,2,0]=(-0.766303732195)*x[0]**o+(0.156822893641)*x[0]+(0.00954334254282)*x[1]**o+(0.201506144561)*x[1]
        arg[2,0,2,1]=(-0.274476449632)*x[0]**o+(-0.361690468185)*x[0]+(-0.816535747268)*x[1]**o+(-0.473913811701)*x[1]
        arg[2,1,0,0]=(-0.927592771482)*x[0]**o+(0.534615269814)*x[0]+(-0.571217133452)*x[1]**o+(0.57448299216)*x[1]
        arg[2,1,0,1]=(-0.298153953211)*x[0]**o+(-0.143636909462)*x[0]+(0.990660288864)*x[1]**o+(-0.495864284763)*x[1]
        arg[2,1,1,0]=(-0.110625727208)*x[0]**o+(-0.221057774317)*x[0]+(0.557195078358)*x[1]**o+(0.803162691808)*x[1]
        arg[2,1,1,1]=(-0.115589176775)*x[0]**o+(0.08253351782)*x[0]+(-0.595996958977)*x[1]**o+(0.501435717501)*x[1]
        arg[2,1,2,0]=(0.78764976658)*x[0]**o+(0.966729831711)*x[0]+(-0.585879575667)*x[1]**o+(0.341148176793)*x[1]
        arg[2,1,2,1]=(-0.551697945691)*x[0]**o+(0.545857249323)*x[0]+(-0.919321724495)*x[1]**o+(-0.286043327415)*x[1]
        arg[2,2,0,0]=(-0.669933890655)*x[0]**o+(-0.344862118495)*x[0]+(0.878331282597)*x[1]**o+(0.881893760718)*x[1]
        arg[2,2,0,1]=(-0.217876618414)*x[0]**o+(-0.289118007199)*x[0]+(0.513464171525)*x[1]**o+(0.310055899102)*x[1]
        arg[2,2,1,0]=(0.0472051379154)*x[0]**o+(0.42455388451)*x[0]+(-0.920571834891)*x[1]**o+(-0.779095076537)*x[1]
        arg[2,2,1,1]=(-0.299619304537)*x[0]**o+(-0.289777287294)*x[0]+(0.920505062988)*x[1]**o+(-0.243924179095)*x[1]
        arg[2,2,2,0]=(-0.819343530877)*x[0]**o+(0.975953690875)*x[0]+(-0.787239459241)*x[1]**o+(-0.199900734691)*x[1]
        arg[2,2,2,1]=(0.0980085780878)*x[0]**o+(0.642946515451)*x[0]+(0.951790535152)*x[1]**o+(-0.253914817366)*x[1]
        arg[2,3,0,0]=(-0.882075522824)*x[0]**o+(0.200934502682)*x[0]+(-0.83200464496)*x[1]**o+(0.941479345375)*x[1]
        arg[2,3,0,1]=(-0.615941782209)*x[0]**o+(-0.693242388092)*x[0]+(-0.640417278871)*x[1]**o+(-0.346459094803)*x[1]
        arg[2,3,1,0]=(-0.447919427952)*x[0]**o+(0.205055689141)*x[0]+(-0.453290032411)*x[1]**o+(0.879062836521)*x[1]
        arg[2,3,1,1]=(0.721133924016)*x[0]**o+(0.71294558459)*x[0]+(-0.250087045805)*x[1]**o+(0.605146721129)*x[1]
        arg[2,3,2,0]=(0.247782322921)*x[0]**o+(-0.956367425119)*x[0]+(-0.054045082005)*x[1]**o+(-0.128151832132)*x[1]
        arg[2,3,2,1]=(-0.427039259088)*x[0]**o+(-0.359849242839)*x[0]+(-0.244990934481)*x[1]**o+(-0.038020035629)*x[1]
        arg[2,4,0,0]=(-0.216409150834)*x[0]**o+(0.492547989921)*x[0]+(0.878317438469)*x[1]**o+(-0.112718719119)*x[1]
        arg[2,4,0,1]=(0.487403443102)*x[0]**o+(0.228155519468)*x[0]+(0.923150527516)*x[1]**o+(0.642654867908)*x[1]
        arg[2,4,1,0]=(0.916418483417)*x[0]**o+(-0.177198935173)*x[0]+(-0.790548940412)*x[1]**o+(0.87540167052)*x[1]
        arg[2,4,1,1]=(0.718502075538)*x[0]**o+(0.691254010569)*x[0]+(-0.0134430122729)*x[1]**o+(-0.739888770749)*x[1]
        arg[2,4,2,0]=(0.287825086861)*x[0]**o+(0.596621809347)*x[0]+(0.729784205391)*x[1]**o+(-0.25984257955)*x[1]
        arg[2,4,2,1]=(-0.515102480025)*x[0]**o+(-0.609951840015)*x[0]+(0.229834340075)*x[1]**o+(-0.727157816241)*x[1]
        arg[3,0,0,0]=(-0.349406091363)*x[0]**o+(0.802348451103)*x[0]+(-0.174421346276)*x[1]**o+(0.120711198036)*x[1]
        arg[3,0,0,1]=(0.367085169423)*x[0]**o+(-0.319683236434)*x[0]+(-0.590820542566)*x[1]**o+(0.607625953303)*x[1]
        arg[3,0,1,0]=(0.231259976433)*x[0]**o+(-0.493297408825)*x[0]+(0.919363640504)*x[1]**o+(-0.839841270429)*x[1]
        arg[3,0,1,1]=(0.57672135684)*x[0]**o+(0.882062879704)*x[0]+(-0.600979706444)*x[1]**o+(-0.704008344509)*x[1]
        arg[3,0,2,0]=(-0.208162114527)*x[0]**o+(0.593682243052)*x[0]+(-0.428436832487)*x[1]**o+(0.51007021604)*x[1]
        arg[3,0,2,1]=(0.315264672653)*x[0]**o+(-0.0835897376534)*x[0]+(-0.135315173106)*x[1]**o+(-0.646372309498)*x[1]
        arg[3,1,0,0]=(0.815505953752)*x[0]**o+(0.0625493643292)*x[0]+(0.383118433815)*x[1]**o+(0.17372720435)*x[1]
        arg[3,1,0,1]=(0.557295786969)*x[0]**o+(-0.568823508165)*x[0]+(-0.00874733267675)*x[1]**o+(-0.27454140598)*x[1]
        arg[3,1,1,0]=(0.649021026591)*x[0]**o+(0.232255030448)*x[0]+(0.845939296425)*x[1]**o+(0.116387378294)*x[1]
        arg[3,1,1,1]=(0.431124940343)*x[0]**o+(-0.0865937271969)*x[0]+(0.937911741288)*x[1]**o+(0.362850039234)*x[1]
        arg[3,1,2,0]=(0.155965414439)*x[0]**o+(0.110850421841)*x[0]+(0.244563080191)*x[1]**o+(0.266325498006)*x[1]
        arg[3,1,2,1]=(0.766813300935)*x[0]**o+(0.77492422889)*x[0]+(-0.0141436554714)*x[1]**o+(0.780444449009)*x[1]
        arg[3,2,0,0]=(0.619086019019)*x[0]**o+(-0.574130567601)*x[0]+(-0.417708159593)*x[1]**o+(0.707727284254)*x[1]
        arg[3,2,0,1]=(-0.845657409353)*x[0]**o+(0.705282507748)*x[0]+(0.124557418112)*x[1]**o+(-0.337493494733)*x[1]
        arg[3,2,1,0]=(0.114875146211)*x[0]**o+(-0.724480426892)*x[0]+(-0.0763476904405)*x[1]**o+(0.12064907114)*x[1]
        arg[3,2,1,1]=(-0.74994413376)*x[0]**o+(-0.916476035217)*x[0]+(-0.360412756599)*x[1]**o+(-0.917457363098)*x[1]
        arg[3,2,2,0]=(-0.446839470984)*x[0]**o+(-0.097704838316)*x[0]+(-0.653584364909)*x[1]**o+(0.561014757758)*x[1]
        arg[3,2,2,1]=(-0.178975301038)*x[0]**o+(-0.843669982667)*x[0]+(-0.566610094242)*x[1]**o+(0.213047649003)*x[1]
        arg[3,3,0,0]=(0.144115764355)*x[0]**o+(0.199435411954)*x[0]+(0.173334385199)*x[1]**o+(-0.942957035808)*x[1]
        arg[3,3,0,1]=(-0.417589926342)*x[0]**o+(-0.819188022852)*x[0]+(-0.68299160003)*x[1]**o+(-0.314851447894)*x[1]
        arg[3,3,1,0]=(0.737898219164)*x[0]**o+(0.612906277393)*x[0]+(-0.768349396729)*x[1]**o+(-0.192998755095)*x[1]
        arg[3,3,1,1]=(0.35433362007)*x[0]**o+(0.894174225208)*x[0]+(-0.450333030837)*x[1]**o+(0.819450589483)*x[1]
        arg[3,3,2,0]=(-0.455377131302)*x[0]**o+(-0.748183144687)*x[0]+(-0.482739450231)*x[1]**o+(0.300625929022)*x[1]
        arg[3,3,2,1]=(-0.602766693698)*x[0]**o+(-0.416288436985)*x[0]+(0.837501983536)*x[1]**o+(-0.745935308188)*x[1]
        arg[3,4,0,0]=(-0.706774010405)*x[0]**o+(0.407766818701)*x[0]+(-0.248964401443)*x[1]**o+(0.675466487229)*x[1]
        arg[3,4,0,1]=(0.512302883876)*x[0]**o+(0.0809544867105)*x[0]+(-0.204555203536)*x[1]**o+(-0.628574073749)*x[1]
        arg[3,4,1,0]=(0.126376796421)*x[0]**o+(0.174607814129)*x[0]+(0.68867194136)*x[1]**o+(0.940025934568)*x[1]
        arg[3,4,1,1]=(0.583042444739)*x[0]**o+(-0.786981763836)*x[0]+(0.319397032405)*x[1]**o+(-0.324993351432)*x[1]
        arg[3,4,2,0]=(-0.941997068836)*x[0]**o+(-0.892825520378)*x[0]+(0.275995419027)*x[1]**o+(0.841260061022)*x[1]
        arg[3,4,2,1]=(0.520110030788)*x[0]**o+(0.519262786324)*x[0]+(0.861221788948)*x[1]**o+(-0.376714934937)*x[1]
        ref[0,0,0,0]=(0.240887319278)*x_ref[0]**o+(-0.192166587545)*x_ref[0]+(0.0176974287409)*x_ref[1]**o+(0.100761748824)*x_ref[1]
        ref[0,0,0,1]=(-0.549649845775)*x_ref[0]**o+(-0.293969245646)*x_ref[0]+(-0.605885325133)*x_ref[1]**o+(0.799558451173)*x_ref[1]
        ref[0,0,1,0]=(-0.899658769582)*x_ref[0]**o+(0.484524193987)*x_ref[0]+(0.437485870351)*x_ref[1]**o+(-0.220549034927)*x_ref[1]
        ref[0,0,1,1]=(-0.146539961793)*x_ref[0]**o+(-0.278343486929)*x_ref[0]+(0.0913912456082)*x_ref[1]**o+(-0.751028975123)*x_ref[1]
        ref[0,0,2,0]=(0.806344701194)*x_ref[0]**o+(0.0537872824561)*x_ref[0]+(-0.70407828379)*x_ref[1]**o+(-0.922963486601)*x_ref[1]
        ref[0,0,2,1]=(0.307202902495)*x_ref[0]**o+(0.301603506731)*x_ref[0]+(0.232983743616)*x_ref[1]**o+(-0.744267542759)*x_ref[1]
        ref[0,1,0,0]=(-0.550616060824)*x_ref[0]**o+(-0.661367750254)*x_ref[0]+(-0.695905895864)*x_ref[1]**o+(-0.788045922398)*x_ref[1]
        ref[0,1,0,1]=(-0.934169134035)*x_ref[0]**o+(-0.843861105521)*x_ref[0]+(-0.237877254439)*x_ref[1]**o+(-0.364956821348)*x_ref[1]
        ref[0,1,1,0]=(0.0950675750026)*x_ref[0]**o+(-0.815510640594)*x_ref[0]+(0.72526035968)*x_ref[1]**o+(0.898675190106)*x_ref[1]
        ref[0,1,1,1]=(0.715609066776)*x_ref[0]**o+(0.371314845206)*x_ref[0]+(0.326878867244)*x_ref[1]**o+(-0.468798950199)*x_ref[1]
        ref[0,1,2,0]=(0.392175939547)*x_ref[0]**o+(0.105332682674)*x_ref[0]+(0.997967455477)*x_ref[1]**o+(0.0322491822496)*x_ref[1]
        ref[0,1,2,1]=(-0.804498399187)*x_ref[0]**o+(0.819653744739)*x_ref[0]+(0.609610573703)*x_ref[1]**o+(-0.623299047794)*x_ref[1]
        ref[0,2,0,0]=(0.993877917386)*x_ref[0]**o+(0.815952897899)*x_ref[0]+(-0.258432174486)*x_ref[1]**o+(0.209324387308)*x_ref[1]
        ref[0,2,0,1]=(0.395946889364)*x_ref[0]**o+(-0.859500539908)*x_ref[0]+(-0.68495734785)*x_ref[1]**o+(0.791813946302)*x_ref[1]
        ref[0,2,1,0]=(0.154347661108)*x_ref[0]**o+(0.0873140744322)*x_ref[0]+(-0.423935109956)*x_ref[1]**o+(-0.754297239226)*x_ref[1]
        ref[0,2,1,1]=(-0.623998419484)*x_ref[0]**o+(0.487300130941)*x_ref[0]+(-0.323613219784)*x_ref[1]**o+(0.744877983729)*x_ref[1]
        ref[0,2,2,0]=(0.515686374371)*x_ref[0]**o+(0.512572153346)*x_ref[0]+(-0.641245783272)*x_ref[1]**o+(-0.390660158759)*x_ref[1]
        ref[0,2,2,1]=(0.24398238139)*x_ref[0]**o+(-0.172020271678)*x_ref[0]+(0.980197812701)*x_ref[1]**o+(0.336941934503)*x_ref[1]
        ref[0,3,0,0]=(0.0154584447615)*x_ref[0]**o+(-0.547048191238)*x_ref[0]+(-0.555319223005)*x_ref[1]**o+(0.0972794598106)*x_ref[1]
        ref[0,3,0,1]=(0.44744389335)*x_ref[0]**o+(0.475046337874)*x_ref[0]+(0.876228982066)*x_ref[1]**o+(-0.0636405241175)*x_ref[1]
        ref[0,3,1,0]=(0.392948465298)*x_ref[0]**o+(-0.341342654182)*x_ref[0]+(0.747185683968)*x_ref[1]**o+(-0.99981172166)*x_ref[1]
        ref[0,3,1,1]=(-0.172355321621)*x_ref[0]**o+(-0.531391484795)*x_ref[0]+(0.451654274019)*x_ref[1]**o+(0.0795298667915)*x_ref[1]
        ref[0,3,2,0]=(0.0841642474674)*x_ref[0]**o+(0.308494435164)*x_ref[0]+(0.53181918181)*x_ref[1]**o+(0.480258067959)*x_ref[1]
        ref[0,3,2,1]=(-0.222903054864)*x_ref[0]**o+(-0.0330655869754)*x_ref[0]+(0.227482525015)*x_ref[1]**o+(0.536136212347)*x_ref[1]
        ref[0,4,0,0]=(0.574798381284)*x_ref[0]**o+(0.181354466484)*x_ref[0]+(0.103049514171)*x_ref[1]**o+(0.156282818835)*x_ref[1]
        ref[0,4,0,1]=(0.478314255289)*x_ref[0]**o+(0.224797222592)*x_ref[0]+(0.307668606272)*x_ref[1]**o+(-0.785262182937)*x_ref[1]
        ref[0,4,1,0]=(0.519356961182)*x_ref[0]**o+(0.646200458344)*x_ref[0]+(0.0522618783181)*x_ref[1]**o+(0.702335307324)*x_ref[1]
        ref[0,4,1,1]=(-0.168756029541)*x_ref[0]**o+(-0.00618364750666)*x_ref[0]+(0.472704616213)*x_ref[1]**o+(0.974257451871)*x_ref[1]
        ref[0,4,2,0]=(-0.279196869573)*x_ref[0]**o+(-0.333083174091)*x_ref[0]+(-0.980089574428)*x_ref[1]**o+(0.405273709405)*x_ref[1]
        ref[0,4,2,1]=(-0.199431380673)*x_ref[0]**o+(0.823294899866)*x_ref[0]+(-0.174125117685)*x_ref[1]**o+(-0.730500493887)*x_ref[1]
        ref[1,0,0,0]=(-0.577281489635)*x_ref[0]**o+(0.231906139898)*x_ref[0]+(0.537128708963)*x_ref[1]**o+(0.363444540898)*x_ref[1]
        ref[1,0,0,1]=(0.204108541882)*x_ref[0]**o+(-0.217661714888)*x_ref[0]+(0.963384293759)*x_ref[1]**o+(-0.0817519801351)*x_ref[1]
        ref[1,0,1,0]=(0.540681325439)*x_ref[0]**o+(-0.496036960808)*x_ref[0]+(0.421037689661)*x_ref[1]**o+(0.19399086996)*x_ref[1]
        ref[1,0,1,1]=(0.482938542126)*x_ref[0]**o+(-0.0725027675209)*x_ref[0]+(-0.441517962054)*x_ref[1]**o+(0.787445860714)*x_ref[1]
        ref[1,0,2,0]=(0.0378928111257)*x_ref[0]**o+(0.451368848019)*x_ref[0]+(-0.228668284485)*x_ref[1]**o+(0.505927229432)*x_ref[1]
        ref[1,0,2,1]=(-0.59833518872)*x_ref[0]**o+(0.956068253354)*x_ref[0]+(0.667905995469)*x_ref[1]**o+(0.963273292214)*x_ref[1]
        ref[1,1,0,0]=(0.901637865236)*x_ref[0]**o+(-0.642678971748)*x_ref[0]+(-0.603915019524)*x_ref[1]**o+(-0.208004993919)*x_ref[1]
        ref[1,1,0,1]=(-0.227162661052)*x_ref[0]**o+(-0.5816434303)*x_ref[0]+(0.274961293801)*x_ref[1]**o+(-0.614001328621)*x_ref[1]
        ref[1,1,1,0]=(0.078438078548)*x_ref[0]**o+(0.570191242883)*x_ref[0]+(0.0952878212484)*x_ref[1]**o+(0.844460205497)*x_ref[1]
        ref[1,1,1,1]=(-0.901445811786)*x_ref[0]**o+(-0.996376775344)*x_ref[0]+(0.707470471613)*x_ref[1]**o+(-0.187517936674)*x_ref[1]
        ref[1,1,2,0]=(-0.66891443067)*x_ref[0]**o+(0.0462057239324)*x_ref[0]+(-0.825820539128)*x_ref[1]**o+(-0.766932836843)*x_ref[1]
        ref[1,1,2,1]=(-0.851428828811)*x_ref[0]**o+(0.504484422346)*x_ref[0]+(0.121378315545)*x_ref[1]**o+(0.903947012757)*x_ref[1]
        ref[1,2,0,0]=(-0.393827449768)*x_ref[0]**o+(-0.251827904608)*x_ref[0]+(0.8320040882)*x_ref[1]**o+(0.57595149855)*x_ref[1]
        ref[1,2,0,1]=(0.638738110712)*x_ref[0]**o+(-0.439962152539)*x_ref[0]+(-0.977211078995)*x_ref[1]**o+(0.0528475212451)*x_ref[1]
        ref[1,2,1,0]=(0.824313588941)*x_ref[0]**o+(-0.589968996823)*x_ref[0]+(-0.512156794152)*x_ref[1]**o+(0.142809794443)*x_ref[1]
        ref[1,2,1,1]=(-0.31789333147)*x_ref[0]**o+(0.598839896949)*x_ref[0]+(0.638319398651)*x_ref[1]**o+(0.353625999317)*x_ref[1]
        ref[1,2,2,0]=(-0.720472365012)*x_ref[0]**o+(0.896538030051)*x_ref[0]+(-0.224639169972)*x_ref[1]**o+(-0.898299360738)*x_ref[1]
        ref[1,2,2,1]=(0.123077707167)*x_ref[0]**o+(0.15922851038)*x_ref[0]+(0.678984783843)*x_ref[1]**o+(-0.778188269852)*x_ref[1]
        ref[1,3,0,0]=(0.344256301209)*x_ref[0]**o+(0.884367738289)*x_ref[0]+(0.00556636219289)*x_ref[1]**o+(0.684361331195)*x_ref[1]
        ref[1,3,0,1]=(-0.812245131065)*x_ref[0]**o+(0.971365671228)*x_ref[0]+(0.0696053239569)*x_ref[1]**o+(-0.300266235138)*x_ref[1]
        ref[1,3,1,0]=(0.470254588058)*x_ref[0]**o+(0.0946825509913)*x_ref[0]+(-0.415619255025)*x_ref[1]**o+(-0.27381652858)*x_ref[1]
        ref[1,3,1,1]=(-0.84555138241)*x_ref[0]**o+(0.873510832091)*x_ref[0]+(-0.138010338152)*x_ref[1]**o+(-0.982274515316)*x_ref[1]
        ref[1,3,2,0]=(0.695614136148)*x_ref[0]**o+(-0.192588082771)*x_ref[0]+(-0.21579127791)*x_ref[1]**o+(-0.0335939736388)*x_ref[1]
        ref[1,3,2,1]=(-0.633363463896)*x_ref[0]**o+(-0.958320549993)*x_ref[0]+(0.197331067481)*x_ref[1]**o+(0.315089888409)*x_ref[1]
        ref[1,4,0,0]=(-0.211531219945)*x_ref[0]**o+(-0.399700721799)*x_ref[0]+(0.578927638839)*x_ref[1]**o+(-0.389034976195)*x_ref[1]
        ref[1,4,0,1]=(-0.846399571893)*x_ref[0]**o+(-0.952904755903)*x_ref[0]+(-0.858399802693)*x_ref[1]**o+(-0.23264354907)*x_ref[1]
        ref[1,4,1,0]=(0.255822441586)*x_ref[0]**o+(0.896237710393)*x_ref[0]+(-0.90248008271)*x_ref[1]**o+(0.654804848244)*x_ref[1]
        ref[1,4,1,1]=(-0.112475508163)*x_ref[0]**o+(-0.861697204453)*x_ref[0]+(-0.782581931187)*x_ref[1]**o+(-0.129345341581)*x_ref[1]
        ref[1,4,2,0]=(0.690095687033)*x_ref[0]**o+(0.729317334019)*x_ref[0]+(-0.00961483612288)*x_ref[1]**o+(-0.288320551243)*x_ref[1]
        ref[1,4,2,1]=(-0.648146433517)*x_ref[0]**o+(-0.301246601263)*x_ref[0]+(-0.490119612994)*x_ref[1]**o+(0.231186845615)*x_ref[1]
        ref[2,0,0,0]=(-0.645560891595)*x_ref[0]**o+(-0.889067544988)*x_ref[0]+(-0.193353414221)*x_ref[1]**o+(-0.726988479344)*x_ref[1]
        ref[2,0,0,1]=(-0.083614738484)*x_ref[0]**o+(-0.822530927241)*x_ref[0]+(0.660338934241)*x_ref[1]**o+(0.0383821036508)*x_ref[1]
        ref[2,0,1,0]=(-0.132250736664)*x_ref[0]**o+(-0.571474370931)*x_ref[0]+(-0.755931100342)*x_ref[1]**o+(-0.145196865115)*x_ref[1]
        ref[2,0,1,1]=(0.545219196727)*x_ref[0]**o+(-0.846368787913)*x_ref[0]+(-0.845706160842)*x_ref[1]**o+(0.550703610989)*x_ref[1]
        ref[2,0,2,0]=(-0.766303732195)*x_ref[0]**o+(0.156822893641)*x_ref[0]+(0.00954334254282)*x_ref[1]**o+(0.201506144561)*x_ref[1]
        ref[2,0,2,1]=(-0.274476449632)*x_ref[0]**o+(-0.361690468185)*x_ref[0]+(-0.816535747268)*x_ref[1]**o+(-0.473913811701)*x_ref[1]
        ref[2,1,0,0]=(-0.927592771482)*x_ref[0]**o+(0.534615269814)*x_ref[0]+(-0.571217133452)*x_ref[1]**o+(0.57448299216)*x_ref[1]
        ref[2,1,0,1]=(-0.298153953211)*x_ref[0]**o+(-0.143636909462)*x_ref[0]+(0.990660288864)*x_ref[1]**o+(-0.495864284763)*x_ref[1]
        ref[2,1,1,0]=(-0.110625727208)*x_ref[0]**o+(-0.221057774317)*x_ref[0]+(0.557195078358)*x_ref[1]**o+(0.803162691808)*x_ref[1]
        ref[2,1,1,1]=(-0.115589176775)*x_ref[0]**o+(0.08253351782)*x_ref[0]+(-0.595996958977)*x_ref[1]**o+(0.501435717501)*x_ref[1]
        ref[2,1,2,0]=(0.78764976658)*x_ref[0]**o+(0.966729831711)*x_ref[0]+(-0.585879575667)*x_ref[1]**o+(0.341148176793)*x_ref[1]
        ref[2,1,2,1]=(-0.551697945691)*x_ref[0]**o+(0.545857249323)*x_ref[0]+(-0.919321724495)*x_ref[1]**o+(-0.286043327415)*x_ref[1]
        ref[2,2,0,0]=(-0.669933890655)*x_ref[0]**o+(-0.344862118495)*x_ref[0]+(0.878331282597)*x_ref[1]**o+(0.881893760718)*x_ref[1]
        ref[2,2,0,1]=(-0.217876618414)*x_ref[0]**o+(-0.289118007199)*x_ref[0]+(0.513464171525)*x_ref[1]**o+(0.310055899102)*x_ref[1]
        ref[2,2,1,0]=(0.0472051379154)*x_ref[0]**o+(0.42455388451)*x_ref[0]+(-0.920571834891)*x_ref[1]**o+(-0.779095076537)*x_ref[1]
        ref[2,2,1,1]=(-0.299619304537)*x_ref[0]**o+(-0.289777287294)*x_ref[0]+(0.920505062988)*x_ref[1]**o+(-0.243924179095)*x_ref[1]
        ref[2,2,2,0]=(-0.819343530877)*x_ref[0]**o+(0.975953690875)*x_ref[0]+(-0.787239459241)*x_ref[1]**o+(-0.199900734691)*x_ref[1]
        ref[2,2,2,1]=(0.0980085780878)*x_ref[0]**o+(0.642946515451)*x_ref[0]+(0.951790535152)*x_ref[1]**o+(-0.253914817366)*x_ref[1]
        ref[2,3,0,0]=(-0.882075522824)*x_ref[0]**o+(0.200934502682)*x_ref[0]+(-0.83200464496)*x_ref[1]**o+(0.941479345375)*x_ref[1]
        ref[2,3,0,1]=(-0.615941782209)*x_ref[0]**o+(-0.693242388092)*x_ref[0]+(-0.640417278871)*x_ref[1]**o+(-0.346459094803)*x_ref[1]
        ref[2,3,1,0]=(-0.447919427952)*x_ref[0]**o+(0.205055689141)*x_ref[0]+(-0.453290032411)*x_ref[1]**o+(0.879062836521)*x_ref[1]
        ref[2,3,1,1]=(0.721133924016)*x_ref[0]**o+(0.71294558459)*x_ref[0]+(-0.250087045805)*x_ref[1]**o+(0.605146721129)*x_ref[1]
        ref[2,3,2,0]=(0.247782322921)*x_ref[0]**o+(-0.956367425119)*x_ref[0]+(-0.054045082005)*x_ref[1]**o+(-0.128151832132)*x_ref[1]
        ref[2,3,2,1]=(-0.427039259088)*x_ref[0]**o+(-0.359849242839)*x_ref[0]+(-0.244990934481)*x_ref[1]**o+(-0.038020035629)*x_ref[1]
        ref[2,4,0,0]=(-0.216409150834)*x_ref[0]**o+(0.492547989921)*x_ref[0]+(0.878317438469)*x_ref[1]**o+(-0.112718719119)*x_ref[1]
        ref[2,4,0,1]=(0.487403443102)*x_ref[0]**o+(0.228155519468)*x_ref[0]+(0.923150527516)*x_ref[1]**o+(0.642654867908)*x_ref[1]
        ref[2,4,1,0]=(0.916418483417)*x_ref[0]**o+(-0.177198935173)*x_ref[0]+(-0.790548940412)*x_ref[1]**o+(0.87540167052)*x_ref[1]
        ref[2,4,1,1]=(0.718502075538)*x_ref[0]**o+(0.691254010569)*x_ref[0]+(-0.0134430122729)*x_ref[1]**o+(-0.739888770749)*x_ref[1]
        ref[2,4,2,0]=(0.287825086861)*x_ref[0]**o+(0.596621809347)*x_ref[0]+(0.729784205391)*x_ref[1]**o+(-0.25984257955)*x_ref[1]
        ref[2,4,2,1]=(-0.515102480025)*x_ref[0]**o+(-0.609951840015)*x_ref[0]+(0.229834340075)*x_ref[1]**o+(-0.727157816241)*x_ref[1]
        ref[3,0,0,0]=(-0.349406091363)*x_ref[0]**o+(0.802348451103)*x_ref[0]+(-0.174421346276)*x_ref[1]**o+(0.120711198036)*x_ref[1]
        ref[3,0,0,1]=(0.367085169423)*x_ref[0]**o+(-0.319683236434)*x_ref[0]+(-0.590820542566)*x_ref[1]**o+(0.607625953303)*x_ref[1]
        ref[3,0,1,0]=(0.231259976433)*x_ref[0]**o+(-0.493297408825)*x_ref[0]+(0.919363640504)*x_ref[1]**o+(-0.839841270429)*x_ref[1]
        ref[3,0,1,1]=(0.57672135684)*x_ref[0]**o+(0.882062879704)*x_ref[0]+(-0.600979706444)*x_ref[1]**o+(-0.704008344509)*x_ref[1]
        ref[3,0,2,0]=(-0.208162114527)*x_ref[0]**o+(0.593682243052)*x_ref[0]+(-0.428436832487)*x_ref[1]**o+(0.51007021604)*x_ref[1]
        ref[3,0,2,1]=(0.315264672653)*x_ref[0]**o+(-0.0835897376534)*x_ref[0]+(-0.135315173106)*x_ref[1]**o+(-0.646372309498)*x_ref[1]
        ref[3,1,0,0]=(0.815505953752)*x_ref[0]**o+(0.0625493643292)*x_ref[0]+(0.383118433815)*x_ref[1]**o+(0.17372720435)*x_ref[1]
        ref[3,1,0,1]=(0.557295786969)*x_ref[0]**o+(-0.568823508165)*x_ref[0]+(-0.00874733267675)*x_ref[1]**o+(-0.27454140598)*x_ref[1]
        ref[3,1,1,0]=(0.649021026591)*x_ref[0]**o+(0.232255030448)*x_ref[0]+(0.845939296425)*x_ref[1]**o+(0.116387378294)*x_ref[1]
        ref[3,1,1,1]=(0.431124940343)*x_ref[0]**o+(-0.0865937271969)*x_ref[0]+(0.937911741288)*x_ref[1]**o+(0.362850039234)*x_ref[1]
        ref[3,1,2,0]=(0.155965414439)*x_ref[0]**o+(0.110850421841)*x_ref[0]+(0.244563080191)*x_ref[1]**o+(0.266325498006)*x_ref[1]
        ref[3,1,2,1]=(0.766813300935)*x_ref[0]**o+(0.77492422889)*x_ref[0]+(-0.0141436554714)*x_ref[1]**o+(0.780444449009)*x_ref[1]
        ref[3,2,0,0]=(0.619086019019)*x_ref[0]**o+(-0.574130567601)*x_ref[0]+(-0.417708159593)*x_ref[1]**o+(0.707727284254)*x_ref[1]
        ref[3,2,0,1]=(-0.845657409353)*x_ref[0]**o+(0.705282507748)*x_ref[0]+(0.124557418112)*x_ref[1]**o+(-0.337493494733)*x_ref[1]
        ref[3,2,1,0]=(0.114875146211)*x_ref[0]**o+(-0.724480426892)*x_ref[0]+(-0.0763476904405)*x_ref[1]**o+(0.12064907114)*x_ref[1]
        ref[3,2,1,1]=(-0.74994413376)*x_ref[0]**o+(-0.916476035217)*x_ref[0]+(-0.360412756599)*x_ref[1]**o+(-0.917457363098)*x_ref[1]
        ref[3,2,2,0]=(-0.446839470984)*x_ref[0]**o+(-0.097704838316)*x_ref[0]+(-0.653584364909)*x_ref[1]**o+(0.561014757758)*x_ref[1]
        ref[3,2,2,1]=(-0.178975301038)*x_ref[0]**o+(-0.843669982667)*x_ref[0]+(-0.566610094242)*x_ref[1]**o+(0.213047649003)*x_ref[1]
        ref[3,3,0,0]=(0.144115764355)*x_ref[0]**o+(0.199435411954)*x_ref[0]+(0.173334385199)*x_ref[1]**o+(-0.942957035808)*x_ref[1]
        ref[3,3,0,1]=(-0.417589926342)*x_ref[0]**o+(-0.819188022852)*x_ref[0]+(-0.68299160003)*x_ref[1]**o+(-0.314851447894)*x_ref[1]
        ref[3,3,1,0]=(0.737898219164)*x_ref[0]**o+(0.612906277393)*x_ref[0]+(-0.768349396729)*x_ref[1]**o+(-0.192998755095)*x_ref[1]
        ref[3,3,1,1]=(0.35433362007)*x_ref[0]**o+(0.894174225208)*x_ref[0]+(-0.450333030837)*x_ref[1]**o+(0.819450589483)*x_ref[1]
        ref[3,3,2,0]=(-0.455377131302)*x_ref[0]**o+(-0.748183144687)*x_ref[0]+(-0.482739450231)*x_ref[1]**o+(0.300625929022)*x_ref[1]
        ref[3,3,2,1]=(-0.602766693698)*x_ref[0]**o+(-0.416288436985)*x_ref[0]+(0.837501983536)*x_ref[1]**o+(-0.745935308188)*x_ref[1]
        ref[3,4,0,0]=(-0.706774010405)*x_ref[0]**o+(0.407766818701)*x_ref[0]+(-0.248964401443)*x_ref[1]**o+(0.675466487229)*x_ref[1]
        ref[3,4,0,1]=(0.512302883876)*x_ref[0]**o+(0.0809544867105)*x_ref[0]+(-0.204555203536)*x_ref[1]**o+(-0.628574073749)*x_ref[1]
        ref[3,4,1,0]=(0.126376796421)*x_ref[0]**o+(0.174607814129)*x_ref[0]+(0.68867194136)*x_ref[1]**o+(0.940025934568)*x_ref[1]
        ref[3,4,1,1]=(0.583042444739)*x_ref[0]**o+(-0.786981763836)*x_ref[0]+(0.319397032405)*x_ref[1]**o+(-0.324993351432)*x_ref[1]
        ref[3,4,2,0]=(-0.941997068836)*x_ref[0]**o+(-0.892825520378)*x_ref[0]+(0.275995419027)*x_ref[1]**o+(0.841260061022)*x_ref[1]
        ref[3,4,2,1]=(0.520110030788)*x_ref[0]**o+(0.519262786324)*x_ref[0]+(0.861221788948)*x_ref[1]**o+(-0.376714934937)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.615899511086)*x[0]**o+(0.823637401824)*x[0]+(-0.986967740198)*x[1]**o+(-0.782088977641)*x[1]+(-0.576799974679)*x[2]**o+(-0.910942885711)*x[2]
        arg[0,0,0,1]=(-0.790830593449)*x[0]**o+(-0.897724279499)*x[0]+(-0.881691805807)*x[1]**o+(-0.133793238901)*x[1]+(-0.0903391684489)*x[2]**o+(0.659893469877)*x[2]
        arg[0,0,1,0]=(-0.0567117900114)*x[0]**o+(-0.858031985686)*x[0]+(-0.397443445702)*x[1]**o+(0.549982838823)*x[1]+(-0.636905143813)*x[2]**o+(0.8488142632)*x[2]
        arg[0,0,1,1]=(-0.644557373082)*x[0]**o+(-0.323654898522)*x[0]+(0.210458705923)*x[1]**o+(0.990436827296)*x[1]+(-0.597872757151)*x[2]**o+(-0.83865548618)*x[2]
        arg[0,0,2,0]=(-0.29830667084)*x[0]**o+(0.327035546909)*x[0]+(0.132690049791)*x[1]**o+(-0.55625443282)*x[1]+(0.0972781385283)*x[2]**o+(0.151885837608)*x[2]
        arg[0,0,2,1]=(-0.435983465578)*x[0]**o+(-0.0852522953446)*x[0]+(0.257654807094)*x[1]**o+(-0.336603197583)*x[1]+(0.637913421592)*x[2]**o+(0.8151042055)*x[2]
        arg[0,1,0,0]=(-0.0373980191125)*x[0]**o+(-0.45937285505)*x[0]+(-0.857933684889)*x[1]**o+(-0.843826686256)*x[1]+(-0.727077320504)*x[2]**o+(-0.65222863395)*x[2]
        arg[0,1,0,1]=(0.63426017384)*x[0]**o+(0.845708641601)*x[0]+(0.246160302486)*x[1]**o+(-0.660855572094)*x[1]+(0.67420981328)*x[2]**o+(-0.280803968781)*x[2]
        arg[0,1,1,0]=(0.881193264597)*x[0]**o+(-0.881936969515)*x[0]+(-0.883246837152)*x[1]**o+(0.182222471927)*x[1]+(0.405380496272)*x[2]**o+(0.175721746883)*x[2]
        arg[0,1,1,1]=(-0.44447667237)*x[0]**o+(0.956169274606)*x[0]+(0.0370532940652)*x[1]**o+(-0.181239635583)*x[1]+(-0.678412161285)*x[2]**o+(-0.653172157401)*x[2]
        arg[0,1,2,0]=(-0.50038634718)*x[0]**o+(0.00278321765167)*x[0]+(-0.831539460218)*x[1]**o+(-0.555243320385)*x[1]+(-0.626789452003)*x[2]**o+(-0.495085243933)*x[2]
        arg[0,1,2,1]=(-0.32886359569)*x[0]**o+(0.406099965475)*x[0]+(0.764756177266)*x[1]**o+(-0.272070344181)*x[1]+(-0.41812596516)*x[2]**o+(0.349063696376)*x[2]
        arg[0,2,0,0]=(-0.580544978365)*x[0]**o+(0.53441241334)*x[0]+(-0.808749795876)*x[1]**o+(0.172110801632)*x[1]+(-0.939262606098)*x[2]**o+(-0.724859325104)*x[2]
        arg[0,2,0,1]=(0.256335221944)*x[0]**o+(0.463522371018)*x[0]+(0.990516478508)*x[1]**o+(0.423931177054)*x[1]+(0.184277693683)*x[2]**o+(-0.272967447689)*x[2]
        arg[0,2,1,0]=(-0.126943811008)*x[0]**o+(0.645821387926)*x[0]+(0.150370174987)*x[1]**o+(-0.150954538301)*x[1]+(-0.378720078559)*x[2]**o+(0.0975963807109)*x[2]
        arg[0,2,1,1]=(-0.451434544339)*x[0]**o+(0.557314109088)*x[0]+(-0.0402970749558)*x[1]**o+(0.306984557206)*x[1]+(-0.297018139186)*x[2]**o+(0.593800698785)*x[2]
        arg[0,2,2,0]=(-0.376199144911)*x[0]**o+(-0.230454116373)*x[0]+(0.992738297651)*x[1]**o+(0.183510534927)*x[1]+(0.409381711992)*x[2]**o+(0.171945265727)*x[2]
        arg[0,2,2,1]=(0.363233095971)*x[0]**o+(-0.130589148249)*x[0]+(0.809611042018)*x[1]**o+(-0.354546307961)*x[1]+(0.0631410169203)*x[2]**o+(0.630306161781)*x[2]
        arg[0,3,0,0]=(0.49655016161)*x[0]**o+(0.878957258614)*x[0]+(-0.397408696784)*x[1]**o+(-0.199445321288)*x[1]+(0.929857086663)*x[2]**o+(-0.648770887161)*x[2]
        arg[0,3,0,1]=(-0.85578334594)*x[0]**o+(-0.294617109229)*x[0]+(-0.0790295769768)*x[1]**o+(0.82151942311)*x[1]+(0.541628117444)*x[2]**o+(0.805418174827)*x[2]
        arg[0,3,1,0]=(0.800775454804)*x[0]**o+(0.599550462983)*x[0]+(-0.384788634082)*x[1]**o+(-0.356933972363)*x[1]+(-0.423116265473)*x[2]**o+(-0.125492982003)*x[2]
        arg[0,3,1,1]=(0.929652994661)*x[0]**o+(-0.57396682406)*x[0]+(0.313584116036)*x[1]**o+(-0.190575043496)*x[1]+(-0.29110884935)*x[2]**o+(0.967530639925)*x[2]
        arg[0,3,2,0]=(0.761136478868)*x[0]**o+(0.565721170014)*x[0]+(0.239619122131)*x[1]**o+(-0.0821651979956)*x[1]+(-0.699218775545)*x[2]**o+(0.19973415585)*x[2]
        arg[0,3,2,1]=(0.209543244648)*x[0]**o+(0.93957796874)*x[0]+(0.354405887426)*x[1]**o+(0.504177262864)*x[1]+(0.446866674894)*x[2]**o+(-0.522936485464)*x[2]
        arg[0,4,0,0]=(0.427541399652)*x[0]**o+(-0.114806067206)*x[0]+(0.236780877643)*x[1]**o+(0.422226381272)*x[1]+(-0.506450961214)*x[2]**o+(0.616091720212)*x[2]
        arg[0,4,0,1]=(0.698773571323)*x[0]**o+(-0.357878713085)*x[0]+(-0.453730359129)*x[1]**o+(-0.328701041355)*x[1]+(-0.28646346679)*x[2]**o+(-0.0449784265203)*x[2]
        arg[0,4,1,0]=(-0.419416306509)*x[0]**o+(-0.1970394955)*x[0]+(-0.814008932379)*x[1]**o+(-0.805127589929)*x[1]+(0.244059789875)*x[2]**o+(-0.983027843847)*x[2]
        arg[0,4,1,1]=(0.84837346004)*x[0]**o+(0.501808510624)*x[0]+(-0.105800351358)*x[1]**o+(0.903999231718)*x[1]+(-0.0473996978124)*x[2]**o+(0.683647253381)*x[2]
        arg[0,4,2,0]=(0.658722899682)*x[0]**o+(-0.401846635016)*x[0]+(-0.967322202581)*x[1]**o+(-0.948306922217)*x[1]+(0.496409308219)*x[2]**o+(-0.184768777052)*x[2]
        arg[0,4,2,1]=(-0.454468390888)*x[0]**o+(-0.220640468215)*x[0]+(0.995269925943)*x[1]**o+(0.778951329211)*x[1]+(0.0439534466302)*x[2]**o+(0.806197038383)*x[2]
        arg[1,0,0,0]=(-0.547461802236)*x[0]**o+(0.370768827789)*x[0]+(-0.619991757552)*x[1]**o+(0.391727195645)*x[1]+(-0.649274922953)*x[2]**o+(-0.478788248429)*x[2]
        arg[1,0,0,1]=(-0.899230505882)*x[0]**o+(0.603064321735)*x[0]+(-0.737475386349)*x[1]**o+(0.594197953001)*x[1]+(0.341324049064)*x[2]**o+(0.285202633917)*x[2]
        arg[1,0,1,0]=(-0.107063589618)*x[0]**o+(-0.260491562358)*x[0]+(-0.667059815618)*x[1]**o+(0.473227354537)*x[1]+(-0.587170510558)*x[2]**o+(-0.740299218393)*x[2]
        arg[1,0,1,1]=(0.87705715098)*x[0]**o+(0.312753471827)*x[0]+(-0.307530283315)*x[1]**o+(0.195564644822)*x[1]+(-0.64384485166)*x[2]**o+(-0.776362951729)*x[2]
        arg[1,0,2,0]=(0.0949408564739)*x[0]**o+(0.765922179855)*x[0]+(0.441215568253)*x[1]**o+(0.339112969437)*x[1]+(-0.0774615292449)*x[2]**o+(0.673258055783)*x[2]
        arg[1,0,2,1]=(0.114828212302)*x[0]**o+(0.0831258766277)*x[0]+(0.674540402248)*x[1]**o+(-0.304921039094)*x[1]+(-0.725514770578)*x[2]**o+(-0.627584773746)*x[2]
        arg[1,1,0,0]=(0.291561714675)*x[0]**o+(0.898587624783)*x[0]+(0.476696264545)*x[1]**o+(-0.454648604978)*x[1]+(0.677949150743)*x[2]**o+(-0.59438289851)*x[2]
        arg[1,1,0,1]=(0.356687258009)*x[0]**o+(0.806236251629)*x[0]+(0.843636690354)*x[1]**o+(-0.649776947181)*x[1]+(0.208542122388)*x[2]**o+(-0.0726045573458)*x[2]
        arg[1,1,1,0]=(0.712800450558)*x[0]**o+(0.748271875755)*x[0]+(0.403880649356)*x[1]**o+(0.956549336754)*x[1]+(0.0420128495043)*x[2]**o+(-0.83177870499)*x[2]
        arg[1,1,1,1]=(-0.886862904174)*x[0]**o+(-0.03822986594)*x[0]+(0.587151517496)*x[1]**o+(-0.120272060898)*x[1]+(-0.573366391933)*x[2]**o+(-0.59772763241)*x[2]
        arg[1,1,2,0]=(0.311960197804)*x[0]**o+(0.486854949444)*x[0]+(-0.647995651954)*x[1]**o+(0.30104277964)*x[1]+(-0.174718703748)*x[2]**o+(-0.506993216295)*x[2]
        arg[1,1,2,1]=(0.979056206991)*x[0]**o+(0.892104464812)*x[0]+(0.0507432703245)*x[1]**o+(0.103630858454)*x[1]+(-0.0254965952266)*x[2]**o+(0.13034027239)*x[2]
        arg[1,2,0,0]=(0.259797966548)*x[0]**o+(0.53018999371)*x[0]+(0.533745722992)*x[1]**o+(0.53029084173)*x[1]+(-0.595126791061)*x[2]**o+(-0.51668295558)*x[2]
        arg[1,2,0,1]=(0.926395590808)*x[0]**o+(-0.0700141207179)*x[0]+(-0.973392454012)*x[1]**o+(-0.46204067097)*x[1]+(0.814964497207)*x[2]**o+(-0.183546091376)*x[2]
        arg[1,2,1,0]=(-0.0872635441658)*x[0]**o+(-0.0307630839035)*x[0]+(0.539798653095)*x[1]**o+(0.966367734848)*x[1]+(-0.817495053612)*x[2]**o+(-0.268397867848)*x[2]
        arg[1,2,1,1]=(-0.933074613921)*x[0]**o+(0.48896533938)*x[0]+(0.453696133488)*x[1]**o+(-0.238730406749)*x[1]+(-0.696785951086)*x[2]**o+(-0.359858385642)*x[2]
        arg[1,2,2,0]=(-0.844406286809)*x[0]**o+(-0.506332243946)*x[0]+(0.976436731658)*x[1]**o+(-0.201489161857)*x[1]+(0.851236173166)*x[2]**o+(-0.998803612947)*x[2]
        arg[1,2,2,1]=(0.259707321449)*x[0]**o+(-0.888149478655)*x[0]+(-0.273502842869)*x[1]**o+(0.291193582158)*x[1]+(-0.897542178027)*x[2]**o+(-0.345994824429)*x[2]
        arg[1,3,0,0]=(0.961298984697)*x[0]**o+(0.248141115092)*x[0]+(-0.623053025178)*x[1]**o+(-0.775095452078)*x[1]+(0.955362707049)*x[2]**o+(-0.517547333705)*x[2]
        arg[1,3,0,1]=(0.858098596227)*x[0]**o+(0.309307483873)*x[0]+(-0.209217034755)*x[1]**o+(0.783803593379)*x[1]+(0.138164128092)*x[2]**o+(-0.922854003461)*x[2]
        arg[1,3,1,0]=(-0.139017942141)*x[0]**o+(-0.338223039119)*x[0]+(0.737882941498)*x[1]**o+(0.948854903087)*x[1]+(-0.0792461633266)*x[2]**o+(0.848875660203)*x[2]
        arg[1,3,1,1]=(0.693284332212)*x[0]**o+(-0.392272875348)*x[0]+(-0.475709045479)*x[1]**o+(0.182948077236)*x[1]+(0.0277569761538)*x[2]**o+(-0.550294947596)*x[2]
        arg[1,3,2,0]=(-0.105381649505)*x[0]**o+(-0.403051690238)*x[0]+(-0.722219735818)*x[1]**o+(-0.380114388869)*x[1]+(0.321452572491)*x[2]**o+(0.858583245253)*x[2]
        arg[1,3,2,1]=(-0.405370062024)*x[0]**o+(0.0743704840047)*x[0]+(0.998947017815)*x[1]**o+(0.983611079522)*x[1]+(0.854202982674)*x[2]**o+(0.288478746498)*x[2]
        arg[1,4,0,0]=(0.0584782489998)*x[0]**o+(0.999286927604)*x[0]+(-0.458166020555)*x[1]**o+(0.529008844128)*x[1]+(0.424592641922)*x[2]**o+(-0.489692340187)*x[2]
        arg[1,4,0,1]=(-0.150555966138)*x[0]**o+(0.362589320838)*x[0]+(0.515820915391)*x[1]**o+(0.742403661312)*x[1]+(0.230278843183)*x[2]**o+(-0.213977724366)*x[2]
        arg[1,4,1,0]=(-0.361161802362)*x[0]**o+(-0.49718332222)*x[0]+(0.143956046197)*x[1]**o+(-0.044550912758)*x[1]+(-0.299892874708)*x[2]**o+(0.370248709792)*x[2]
        arg[1,4,1,1]=(0.259365238173)*x[0]**o+(-0.474694548557)*x[0]+(0.732983396669)*x[1]**o+(-0.295151546774)*x[1]+(-0.337144628526)*x[2]**o+(0.664201775714)*x[2]
        arg[1,4,2,0]=(-0.318016754284)*x[0]**o+(0.555291443852)*x[0]+(0.538150727998)*x[1]**o+(-0.800118353132)*x[1]+(0.652262386596)*x[2]**o+(0.303156645857)*x[2]
        arg[1,4,2,1]=(0.0827142921796)*x[0]**o+(-0.88498648372)*x[0]+(-0.941249844349)*x[1]**o+(-0.162576659039)*x[1]+(0.867621252192)*x[2]**o+(0.0834583795227)*x[2]
        arg[2,0,0,0]=(0.0670088199264)*x[0]**o+(-0.0425961894346)*x[0]+(0.329605789302)*x[1]**o+(0.860453156363)*x[1]+(-0.37410469727)*x[2]**o+(0.55476432384)*x[2]
        arg[2,0,0,1]=(0.801738221879)*x[0]**o+(-0.836563530923)*x[0]+(-0.449651673142)*x[1]**o+(-0.19423133918)*x[1]+(0.947648200044)*x[2]**o+(0.0854067601423)*x[2]
        arg[2,0,1,0]=(0.986595046855)*x[0]**o+(-0.358319352175)*x[0]+(0.493917713513)*x[1]**o+(-0.92052282445)*x[1]+(-0.692726467259)*x[2]**o+(-0.0636943737807)*x[2]
        arg[2,0,1,1]=(0.38491150492)*x[0]**o+(0.766893245938)*x[0]+(0.420188917945)*x[1]**o+(0.43760538854)*x[1]+(-0.770749451981)*x[2]**o+(0.124882164378)*x[2]
        arg[2,0,2,0]=(0.775715718797)*x[0]**o+(0.89516703208)*x[0]+(-0.829021524898)*x[1]**o+(-0.650262454311)*x[1]+(-0.812556336598)*x[2]**o+(-0.997075743997)*x[2]
        arg[2,0,2,1]=(0.3503761008)*x[0]**o+(-0.822009897455)*x[0]+(-0.657657365354)*x[1]**o+(0.476295612156)*x[1]+(0.315554008711)*x[2]**o+(0.691631078049)*x[2]
        arg[2,1,0,0]=(0.680177375433)*x[0]**o+(-0.922726813311)*x[0]+(-0.581116746321)*x[1]**o+(0.628138099343)*x[1]+(-0.568262800293)*x[2]**o+(-0.228201382395)*x[2]
        arg[2,1,0,1]=(0.210094833324)*x[0]**o+(0.281549673876)*x[0]+(0.23785231073)*x[1]**o+(-0.103812455993)*x[1]+(0.530445971191)*x[2]**o+(-0.0830520449939)*x[2]
        arg[2,1,1,0]=(0.367133297563)*x[0]**o+(0.168289954585)*x[0]+(0.614117392787)*x[1]**o+(0.790467766841)*x[1]+(-0.958002150805)*x[2]**o+(0.328916248824)*x[2]
        arg[2,1,1,1]=(0.089535347231)*x[0]**o+(-0.93155993701)*x[0]+(-0.674670760794)*x[1]**o+(-0.274148261773)*x[1]+(-0.807020126056)*x[2]**o+(-0.323334769496)*x[2]
        arg[2,1,2,0]=(-0.948679835219)*x[0]**o+(-0.577235742425)*x[0]+(-0.732584834146)*x[1]**o+(-0.475242504288)*x[1]+(0.213663223262)*x[2]**o+(-0.151401354381)*x[2]
        arg[2,1,2,1]=(0.264513246206)*x[0]**o+(-0.605806437391)*x[0]+(-0.058730926693)*x[1]**o+(-0.661696304412)*x[1]+(-0.361277352457)*x[2]**o+(-0.804408258253)*x[2]
        arg[2,2,0,0]=(-0.144562358054)*x[0]**o+(-0.0960550555725)*x[0]+(-0.905096867803)*x[1]**o+(0.188861784103)*x[1]+(-0.616751005568)*x[2]**o+(0.604711974248)*x[2]
        arg[2,2,0,1]=(-0.0582524307858)*x[0]**o+(-0.503648879785)*x[0]+(-0.682755209956)*x[1]**o+(-0.808213164425)*x[1]+(0.539987276327)*x[2]**o+(0.650557703689)*x[2]
        arg[2,2,1,0]=(0.0607729567733)*x[0]**o+(-0.498819638428)*x[0]+(0.434442595334)*x[1]**o+(-0.491063977603)*x[1]+(0.499436212229)*x[2]**o+(-0.932071991206)*x[2]
        arg[2,2,1,1]=(-0.461295295294)*x[0]**o+(-0.153600767916)*x[0]+(-0.86452392231)*x[1]**o+(0.136760034153)*x[1]+(-0.471642416082)*x[2]**o+(0.453820403438)*x[2]
        arg[2,2,2,0]=(-0.970051888584)*x[0]**o+(0.415728825057)*x[0]+(0.971643909569)*x[1]**o+(0.923719420524)*x[1]+(-0.837313513804)*x[2]**o+(-0.738057676093)*x[2]
        arg[2,2,2,1]=(-0.941355918089)*x[0]**o+(-0.9629784524)*x[0]+(-0.931184348072)*x[1]**o+(0.720006097658)*x[1]+(-0.140750567978)*x[2]**o+(0.55381939119)*x[2]
        arg[2,3,0,0]=(0.753455933669)*x[0]**o+(0.352384922037)*x[0]+(0.737644348002)*x[1]**o+(-0.565144188248)*x[1]+(0.257417624014)*x[2]**o+(-0.660006641683)*x[2]
        arg[2,3,0,1]=(0.886996770925)*x[0]**o+(-0.876595418252)*x[0]+(-0.189778073871)*x[1]**o+(0.560867442336)*x[1]+(-0.262338156346)*x[2]**o+(0.614749098901)*x[2]
        arg[2,3,1,0]=(-0.722912374016)*x[0]**o+(0.447451310273)*x[0]+(0.397510140568)*x[1]**o+(0.932405917932)*x[1]+(-0.206969252336)*x[2]**o+(-0.897366208888)*x[2]
        arg[2,3,1,1]=(0.106324074547)*x[0]**o+(0.822109425884)*x[0]+(0.62699223277)*x[1]**o+(0.238371557839)*x[1]+(-0.422051841243)*x[2]**o+(0.868932130329)*x[2]
        arg[2,3,2,0]=(0.0155790371581)*x[0]**o+(-0.696029577918)*x[0]+(0.259909585217)*x[1]**o+(0.260209121862)*x[1]+(0.780977081459)*x[2]**o+(-0.819957534538)*x[2]
        arg[2,3,2,1]=(0.736223721904)*x[0]**o+(-0.433132598174)*x[0]+(-0.880913649812)*x[1]**o+(-0.375988777461)*x[1]+(-0.351763406307)*x[2]**o+(0.948468252467)*x[2]
        arg[2,4,0,0]=(0.943349180556)*x[0]**o+(-0.720044514103)*x[0]+(0.671900161807)*x[1]**o+(0.358571142942)*x[1]+(-0.132755998648)*x[2]**o+(0.0749815044365)*x[2]
        arg[2,4,0,1]=(0.639070752347)*x[0]**o+(0.551441363406)*x[0]+(0.507547844432)*x[1]**o+(-0.800043772307)*x[1]+(-0.506359503394)*x[2]**o+(0.952191432651)*x[2]
        arg[2,4,1,0]=(-0.655864732254)*x[0]**o+(0.142198390642)*x[0]+(0.075292486702)*x[1]**o+(0.191013189029)*x[1]+(0.106768453247)*x[2]**o+(0.974526515449)*x[2]
        arg[2,4,1,1]=(-0.354750205648)*x[0]**o+(0.141506849368)*x[0]+(-0.170074898141)*x[1]**o+(-0.0882776529335)*x[1]+(0.845254802876)*x[2]**o+(-0.435414692125)*x[2]
        arg[2,4,2,0]=(-0.412836027163)*x[0]**o+(0.830777132582)*x[0]+(-0.87554952236)*x[1]**o+(0.970647516313)*x[1]+(0.989864827286)*x[2]**o+(-0.67790755421)*x[2]
        arg[2,4,2,1]=(0.762907045031)*x[0]**o+(0.140771931118)*x[0]+(0.716859940142)*x[1]**o+(0.151769314189)*x[1]+(-0.812408853675)*x[2]**o+(0.755727078666)*x[2]
        arg[3,0,0,0]=(-0.13358237438)*x[0]**o+(-0.594735856789)*x[0]+(-0.367418402298)*x[1]**o+(0.5349162872)*x[1]+(0.238285320118)*x[2]**o+(0.999627906459)*x[2]
        arg[3,0,0,1]=(-0.537433369218)*x[0]**o+(0.133034652667)*x[0]+(-0.190302464596)*x[1]**o+(0.869919012762)*x[1]+(0.320294306008)*x[2]**o+(0.865908856629)*x[2]
        arg[3,0,1,0]=(0.480851043043)*x[0]**o+(0.833358259022)*x[0]+(0.893051476621)*x[1]**o+(-0.505255581199)*x[1]+(-0.15316145317)*x[2]**o+(-0.937103563776)*x[2]
        arg[3,0,1,1]=(0.765360137605)*x[0]**o+(0.502399552842)*x[0]+(-0.215707512276)*x[1]**o+(-0.402720816967)*x[1]+(0.231360098578)*x[2]**o+(0.732903327067)*x[2]
        arg[3,0,2,0]=(0.120836669045)*x[0]**o+(0.198346794548)*x[0]+(-0.28708447536)*x[1]**o+(0.289672075353)*x[1]+(-0.0305777738322)*x[2]**o+(-0.0686651569373)*x[2]
        arg[3,0,2,1]=(0.0574598844614)*x[0]**o+(0.031092492399)*x[0]+(-0.406805507341)*x[1]**o+(-0.623370509947)*x[1]+(-0.476094160966)*x[2]**o+(0.0840271758875)*x[2]
        arg[3,1,0,0]=(-0.569415249126)*x[0]**o+(-0.339675540399)*x[0]+(0.182536980707)*x[1]**o+(0.727480200453)*x[1]+(0.851672551504)*x[2]**o+(0.817312313075)*x[2]
        arg[3,1,0,1]=(0.921422642904)*x[0]**o+(0.697656143906)*x[0]+(0.969442847752)*x[1]**o+(-0.00957994710064)*x[1]+(-0.315230556399)*x[2]**o+(-0.870759158384)*x[2]
        arg[3,1,1,0]=(-0.813310465463)*x[0]**o+(-0.809362802903)*x[0]+(-0.0685101056428)*x[1]**o+(0.708195651847)*x[1]+(-0.876449456312)*x[2]**o+(0.882503536479)*x[2]
        arg[3,1,1,1]=(-0.277504922905)*x[0]**o+(-0.819428706347)*x[0]+(0.463203258731)*x[1]**o+(0.597493096401)*x[1]+(0.630522187524)*x[2]**o+(0.92480079466)*x[2]
        arg[3,1,2,0]=(-0.703167803764)*x[0]**o+(0.730459205382)*x[0]+(-0.507036921123)*x[1]**o+(0.419358545765)*x[1]+(-0.542758923679)*x[2]**o+(0.831844312917)*x[2]
        arg[3,1,2,1]=(0.135200299704)*x[0]**o+(-0.160734675188)*x[0]+(-0.893851689192)*x[1]**o+(0.748734906185)*x[1]+(0.970464447167)*x[2]**o+(0.277911376757)*x[2]
        arg[3,2,0,0]=(0.750104328784)*x[0]**o+(-0.818350593893)*x[0]+(-0.749900854281)*x[1]**o+(0.289117653816)*x[1]+(-0.219904822162)*x[2]**o+(0.805208050452)*x[2]
        arg[3,2,0,1]=(-0.129614389749)*x[0]**o+(-0.64067443816)*x[0]+(0.513013650233)*x[1]**o+(0.868869035341)*x[1]+(-0.040145121639)*x[2]**o+(-0.282350136119)*x[2]
        arg[3,2,1,0]=(-0.172814929104)*x[0]**o+(-0.296930471915)*x[0]+(-0.796114111585)*x[1]**o+(-0.210181080249)*x[1]+(-0.606107841173)*x[2]**o+(-0.954329084175)*x[2]
        arg[3,2,1,1]=(0.136149261959)*x[0]**o+(-0.613980696755)*x[0]+(0.968430714778)*x[1]**o+(0.683553952099)*x[1]+(-0.321383789121)*x[2]**o+(-0.139570984234)*x[2]
        arg[3,2,2,0]=(-0.273321720654)*x[0]**o+(0.963651110162)*x[0]+(-0.847402611499)*x[1]**o+(0.237131262905)*x[1]+(-0.617296821558)*x[2]**o+(-0.681856394629)*x[2]
        arg[3,2,2,1]=(-0.760071928624)*x[0]**o+(0.0345053779755)*x[0]+(-0.583680069732)*x[1]**o+(0.150497658077)*x[1]+(0.0257096621429)*x[2]**o+(-0.159406203385)*x[2]
        arg[3,3,0,0]=(-0.0668501746947)*x[0]**o+(-0.183645899611)*x[0]+(0.123667845765)*x[1]**o+(0.0928142420352)*x[1]+(-0.933250766317)*x[2]**o+(-0.629986949639)*x[2]
        arg[3,3,0,1]=(-0.0212790928206)*x[0]**o+(0.444036360325)*x[0]+(0.09182358127)*x[1]**o+(0.473088563672)*x[1]+(0.0852925382953)*x[2]**o+(-0.238517929786)*x[2]
        arg[3,3,1,0]=(0.0647235093728)*x[0]**o+(-0.266285898716)*x[0]+(0.944999560866)*x[1]**o+(-0.757696716486)*x[1]+(-0.439165278732)*x[2]**o+(0.960091479774)*x[2]
        arg[3,3,1,1]=(-0.792249238894)*x[0]**o+(-0.568585439587)*x[0]+(-0.0426655335607)*x[1]**o+(0.980004032206)*x[1]+(0.635227812889)*x[2]**o+(0.6687165384)*x[2]
        arg[3,3,2,0]=(0.340468442769)*x[0]**o+(0.791071154735)*x[0]+(-0.817919975547)*x[1]**o+(0.0328383066982)*x[1]+(-0.323093904705)*x[2]**o+(-0.302641620002)*x[2]
        arg[3,3,2,1]=(-0.476501733213)*x[0]**o+(-0.429845943537)*x[0]+(0.518551215268)*x[1]**o+(-0.551690143654)*x[1]+(0.591983049997)*x[2]**o+(0.586292683497)*x[2]
        arg[3,4,0,0]=(0.119036526248)*x[0]**o+(-0.156404044048)*x[0]+(0.158504048681)*x[1]**o+(-0.3779600709)*x[1]+(-0.362107671534)*x[2]**o+(-0.314508628804)*x[2]
        arg[3,4,0,1]=(0.382011103106)*x[0]**o+(-0.123711676964)*x[0]+(-0.939014457264)*x[1]**o+(-0.5986155807)*x[1]+(-0.852092491566)*x[2]**o+(-0.410438924788)*x[2]
        arg[3,4,1,0]=(0.83612586535)*x[0]**o+(-0.606098645092)*x[0]+(0.643469717148)*x[1]**o+(0.845106525654)*x[1]+(0.347455649691)*x[2]**o+(-0.166796580458)*x[2]
        arg[3,4,1,1]=(0.693431758156)*x[0]**o+(-0.313983401147)*x[0]+(-0.862690834893)*x[1]**o+(0.417233988055)*x[1]+(-0.361769997589)*x[2]**o+(-0.22749888397)*x[2]
        arg[3,4,2,0]=(0.680358405387)*x[0]**o+(-0.786041842215)*x[0]+(0.288823579415)*x[1]**o+(-0.961099851196)*x[1]+(0.669431785695)*x[2]**o+(0.956564160268)*x[2]
        arg[3,4,2,1]=(-0.21416500005)*x[0]**o+(0.48495191486)*x[0]+(-0.173465452645)*x[1]**o+(0.662530321557)*x[1]+(0.984573283824)*x[2]**o+(-0.568211383785)*x[2]
        ref[0,0,0,0]=(-0.615899511086)*x_ref[0]**o+(0.823637401824)*x_ref[0]+(-0.986967740198)*x_ref[1]**o+(-0.782088977641)*x_ref[1]+(-0.576799974679)*x_ref[2]**o+(-0.910942885711)*x_ref[2]
        ref[0,0,0,1]=(-0.790830593449)*x_ref[0]**o+(-0.897724279499)*x_ref[0]+(-0.881691805807)*x_ref[1]**o+(-0.133793238901)*x_ref[1]+(-0.0903391684489)*x_ref[2]**o+(0.659893469877)*x_ref[2]
        ref[0,0,1,0]=(-0.0567117900114)*x_ref[0]**o+(-0.858031985686)*x_ref[0]+(-0.397443445702)*x_ref[1]**o+(0.549982838823)*x_ref[1]+(-0.636905143813)*x_ref[2]**o+(0.8488142632)*x_ref[2]
        ref[0,0,1,1]=(-0.644557373082)*x_ref[0]**o+(-0.323654898522)*x_ref[0]+(0.210458705923)*x_ref[1]**o+(0.990436827296)*x_ref[1]+(-0.597872757151)*x_ref[2]**o+(-0.83865548618)*x_ref[2]
        ref[0,0,2,0]=(-0.29830667084)*x_ref[0]**o+(0.327035546909)*x_ref[0]+(0.132690049791)*x_ref[1]**o+(-0.55625443282)*x_ref[1]+(0.0972781385283)*x_ref[2]**o+(0.151885837608)*x_ref[2]
        ref[0,0,2,1]=(-0.435983465578)*x_ref[0]**o+(-0.0852522953446)*x_ref[0]+(0.257654807094)*x_ref[1]**o+(-0.336603197583)*x_ref[1]+(0.637913421592)*x_ref[2]**o+(0.8151042055)*x_ref[2]
        ref[0,1,0,0]=(-0.0373980191125)*x_ref[0]**o+(-0.45937285505)*x_ref[0]+(-0.857933684889)*x_ref[1]**o+(-0.843826686256)*x_ref[1]+(-0.727077320504)*x_ref[2]**o+(-0.65222863395)*x_ref[2]
        ref[0,1,0,1]=(0.63426017384)*x_ref[0]**o+(0.845708641601)*x_ref[0]+(0.246160302486)*x_ref[1]**o+(-0.660855572094)*x_ref[1]+(0.67420981328)*x_ref[2]**o+(-0.280803968781)*x_ref[2]
        ref[0,1,1,0]=(0.881193264597)*x_ref[0]**o+(-0.881936969515)*x_ref[0]+(-0.883246837152)*x_ref[1]**o+(0.182222471927)*x_ref[1]+(0.405380496272)*x_ref[2]**o+(0.175721746883)*x_ref[2]
        ref[0,1,1,1]=(-0.44447667237)*x_ref[0]**o+(0.956169274606)*x_ref[0]+(0.0370532940652)*x_ref[1]**o+(-0.181239635583)*x_ref[1]+(-0.678412161285)*x_ref[2]**o+(-0.653172157401)*x_ref[2]
        ref[0,1,2,0]=(-0.50038634718)*x_ref[0]**o+(0.00278321765167)*x_ref[0]+(-0.831539460218)*x_ref[1]**o+(-0.555243320385)*x_ref[1]+(-0.626789452003)*x_ref[2]**o+(-0.495085243933)*x_ref[2]
        ref[0,1,2,1]=(-0.32886359569)*x_ref[0]**o+(0.406099965475)*x_ref[0]+(0.764756177266)*x_ref[1]**o+(-0.272070344181)*x_ref[1]+(-0.41812596516)*x_ref[2]**o+(0.349063696376)*x_ref[2]
        ref[0,2,0,0]=(-0.580544978365)*x_ref[0]**o+(0.53441241334)*x_ref[0]+(-0.808749795876)*x_ref[1]**o+(0.172110801632)*x_ref[1]+(-0.939262606098)*x_ref[2]**o+(-0.724859325104)*x_ref[2]
        ref[0,2,0,1]=(0.256335221944)*x_ref[0]**o+(0.463522371018)*x_ref[0]+(0.990516478508)*x_ref[1]**o+(0.423931177054)*x_ref[1]+(0.184277693683)*x_ref[2]**o+(-0.272967447689)*x_ref[2]
        ref[0,2,1,0]=(-0.126943811008)*x_ref[0]**o+(0.645821387926)*x_ref[0]+(0.150370174987)*x_ref[1]**o+(-0.150954538301)*x_ref[1]+(-0.378720078559)*x_ref[2]**o+(0.0975963807109)*x_ref[2]
        ref[0,2,1,1]=(-0.451434544339)*x_ref[0]**o+(0.557314109088)*x_ref[0]+(-0.0402970749558)*x_ref[1]**o+(0.306984557206)*x_ref[1]+(-0.297018139186)*x_ref[2]**o+(0.593800698785)*x_ref[2]
        ref[0,2,2,0]=(-0.376199144911)*x_ref[0]**o+(-0.230454116373)*x_ref[0]+(0.992738297651)*x_ref[1]**o+(0.183510534927)*x_ref[1]+(0.409381711992)*x_ref[2]**o+(0.171945265727)*x_ref[2]
        ref[0,2,2,1]=(0.363233095971)*x_ref[0]**o+(-0.130589148249)*x_ref[0]+(0.809611042018)*x_ref[1]**o+(-0.354546307961)*x_ref[1]+(0.0631410169203)*x_ref[2]**o+(0.630306161781)*x_ref[2]
        ref[0,3,0,0]=(0.49655016161)*x_ref[0]**o+(0.878957258614)*x_ref[0]+(-0.397408696784)*x_ref[1]**o+(-0.199445321288)*x_ref[1]+(0.929857086663)*x_ref[2]**o+(-0.648770887161)*x_ref[2]
        ref[0,3,0,1]=(-0.85578334594)*x_ref[0]**o+(-0.294617109229)*x_ref[0]+(-0.0790295769768)*x_ref[1]**o+(0.82151942311)*x_ref[1]+(0.541628117444)*x_ref[2]**o+(0.805418174827)*x_ref[2]
        ref[0,3,1,0]=(0.800775454804)*x_ref[0]**o+(0.599550462983)*x_ref[0]+(-0.384788634082)*x_ref[1]**o+(-0.356933972363)*x_ref[1]+(-0.423116265473)*x_ref[2]**o+(-0.125492982003)*x_ref[2]
        ref[0,3,1,1]=(0.929652994661)*x_ref[0]**o+(-0.57396682406)*x_ref[0]+(0.313584116036)*x_ref[1]**o+(-0.190575043496)*x_ref[1]+(-0.29110884935)*x_ref[2]**o+(0.967530639925)*x_ref[2]
        ref[0,3,2,0]=(0.761136478868)*x_ref[0]**o+(0.565721170014)*x_ref[0]+(0.239619122131)*x_ref[1]**o+(-0.0821651979956)*x_ref[1]+(-0.699218775545)*x_ref[2]**o+(0.19973415585)*x_ref[2]
        ref[0,3,2,1]=(0.209543244648)*x_ref[0]**o+(0.93957796874)*x_ref[0]+(0.354405887426)*x_ref[1]**o+(0.504177262864)*x_ref[1]+(0.446866674894)*x_ref[2]**o+(-0.522936485464)*x_ref[2]
        ref[0,4,0,0]=(0.427541399652)*x_ref[0]**o+(-0.114806067206)*x_ref[0]+(0.236780877643)*x_ref[1]**o+(0.422226381272)*x_ref[1]+(-0.506450961214)*x_ref[2]**o+(0.616091720212)*x_ref[2]
        ref[0,4,0,1]=(0.698773571323)*x_ref[0]**o+(-0.357878713085)*x_ref[0]+(-0.453730359129)*x_ref[1]**o+(-0.328701041355)*x_ref[1]+(-0.28646346679)*x_ref[2]**o+(-0.0449784265203)*x_ref[2]
        ref[0,4,1,0]=(-0.419416306509)*x_ref[0]**o+(-0.1970394955)*x_ref[0]+(-0.814008932379)*x_ref[1]**o+(-0.805127589929)*x_ref[1]+(0.244059789875)*x_ref[2]**o+(-0.983027843847)*x_ref[2]
        ref[0,4,1,1]=(0.84837346004)*x_ref[0]**o+(0.501808510624)*x_ref[0]+(-0.105800351358)*x_ref[1]**o+(0.903999231718)*x_ref[1]+(-0.0473996978124)*x_ref[2]**o+(0.683647253381)*x_ref[2]
        ref[0,4,2,0]=(0.658722899682)*x_ref[0]**o+(-0.401846635016)*x_ref[0]+(-0.967322202581)*x_ref[1]**o+(-0.948306922217)*x_ref[1]+(0.496409308219)*x_ref[2]**o+(-0.184768777052)*x_ref[2]
        ref[0,4,2,1]=(-0.454468390888)*x_ref[0]**o+(-0.220640468215)*x_ref[0]+(0.995269925943)*x_ref[1]**o+(0.778951329211)*x_ref[1]+(0.0439534466302)*x_ref[2]**o+(0.806197038383)*x_ref[2]
        ref[1,0,0,0]=(-0.547461802236)*x_ref[0]**o+(0.370768827789)*x_ref[0]+(-0.619991757552)*x_ref[1]**o+(0.391727195645)*x_ref[1]+(-0.649274922953)*x_ref[2]**o+(-0.478788248429)*x_ref[2]
        ref[1,0,0,1]=(-0.899230505882)*x_ref[0]**o+(0.603064321735)*x_ref[0]+(-0.737475386349)*x_ref[1]**o+(0.594197953001)*x_ref[1]+(0.341324049064)*x_ref[2]**o+(0.285202633917)*x_ref[2]
        ref[1,0,1,0]=(-0.107063589618)*x_ref[0]**o+(-0.260491562358)*x_ref[0]+(-0.667059815618)*x_ref[1]**o+(0.473227354537)*x_ref[1]+(-0.587170510558)*x_ref[2]**o+(-0.740299218393)*x_ref[2]
        ref[1,0,1,1]=(0.87705715098)*x_ref[0]**o+(0.312753471827)*x_ref[0]+(-0.307530283315)*x_ref[1]**o+(0.195564644822)*x_ref[1]+(-0.64384485166)*x_ref[2]**o+(-0.776362951729)*x_ref[2]
        ref[1,0,2,0]=(0.0949408564739)*x_ref[0]**o+(0.765922179855)*x_ref[0]+(0.441215568253)*x_ref[1]**o+(0.339112969437)*x_ref[1]+(-0.0774615292449)*x_ref[2]**o+(0.673258055783)*x_ref[2]
        ref[1,0,2,1]=(0.114828212302)*x_ref[0]**o+(0.0831258766277)*x_ref[0]+(0.674540402248)*x_ref[1]**o+(-0.304921039094)*x_ref[1]+(-0.725514770578)*x_ref[2]**o+(-0.627584773746)*x_ref[2]
        ref[1,1,0,0]=(0.291561714675)*x_ref[0]**o+(0.898587624783)*x_ref[0]+(0.476696264545)*x_ref[1]**o+(-0.454648604978)*x_ref[1]+(0.677949150743)*x_ref[2]**o+(-0.59438289851)*x_ref[2]
        ref[1,1,0,1]=(0.356687258009)*x_ref[0]**o+(0.806236251629)*x_ref[0]+(0.843636690354)*x_ref[1]**o+(-0.649776947181)*x_ref[1]+(0.208542122388)*x_ref[2]**o+(-0.0726045573458)*x_ref[2]
        ref[1,1,1,0]=(0.712800450558)*x_ref[0]**o+(0.748271875755)*x_ref[0]+(0.403880649356)*x_ref[1]**o+(0.956549336754)*x_ref[1]+(0.0420128495043)*x_ref[2]**o+(-0.83177870499)*x_ref[2]
        ref[1,1,1,1]=(-0.886862904174)*x_ref[0]**o+(-0.03822986594)*x_ref[0]+(0.587151517496)*x_ref[1]**o+(-0.120272060898)*x_ref[1]+(-0.573366391933)*x_ref[2]**o+(-0.59772763241)*x_ref[2]
        ref[1,1,2,0]=(0.311960197804)*x_ref[0]**o+(0.486854949444)*x_ref[0]+(-0.647995651954)*x_ref[1]**o+(0.30104277964)*x_ref[1]+(-0.174718703748)*x_ref[2]**o+(-0.506993216295)*x_ref[2]
        ref[1,1,2,1]=(0.979056206991)*x_ref[0]**o+(0.892104464812)*x_ref[0]+(0.0507432703245)*x_ref[1]**o+(0.103630858454)*x_ref[1]+(-0.0254965952266)*x_ref[2]**o+(0.13034027239)*x_ref[2]
        ref[1,2,0,0]=(0.259797966548)*x_ref[0]**o+(0.53018999371)*x_ref[0]+(0.533745722992)*x_ref[1]**o+(0.53029084173)*x_ref[1]+(-0.595126791061)*x_ref[2]**o+(-0.51668295558)*x_ref[2]
        ref[1,2,0,1]=(0.926395590808)*x_ref[0]**o+(-0.0700141207179)*x_ref[0]+(-0.973392454012)*x_ref[1]**o+(-0.46204067097)*x_ref[1]+(0.814964497207)*x_ref[2]**o+(-0.183546091376)*x_ref[2]
        ref[1,2,1,0]=(-0.0872635441658)*x_ref[0]**o+(-0.0307630839035)*x_ref[0]+(0.539798653095)*x_ref[1]**o+(0.966367734848)*x_ref[1]+(-0.817495053612)*x_ref[2]**o+(-0.268397867848)*x_ref[2]
        ref[1,2,1,1]=(-0.933074613921)*x_ref[0]**o+(0.48896533938)*x_ref[0]+(0.453696133488)*x_ref[1]**o+(-0.238730406749)*x_ref[1]+(-0.696785951086)*x_ref[2]**o+(-0.359858385642)*x_ref[2]
        ref[1,2,2,0]=(-0.844406286809)*x_ref[0]**o+(-0.506332243946)*x_ref[0]+(0.976436731658)*x_ref[1]**o+(-0.201489161857)*x_ref[1]+(0.851236173166)*x_ref[2]**o+(-0.998803612947)*x_ref[2]
        ref[1,2,2,1]=(0.259707321449)*x_ref[0]**o+(-0.888149478655)*x_ref[0]+(-0.273502842869)*x_ref[1]**o+(0.291193582158)*x_ref[1]+(-0.897542178027)*x_ref[2]**o+(-0.345994824429)*x_ref[2]
        ref[1,3,0,0]=(0.961298984697)*x_ref[0]**o+(0.248141115092)*x_ref[0]+(-0.623053025178)*x_ref[1]**o+(-0.775095452078)*x_ref[1]+(0.955362707049)*x_ref[2]**o+(-0.517547333705)*x_ref[2]
        ref[1,3,0,1]=(0.858098596227)*x_ref[0]**o+(0.309307483873)*x_ref[0]+(-0.209217034755)*x_ref[1]**o+(0.783803593379)*x_ref[1]+(0.138164128092)*x_ref[2]**o+(-0.922854003461)*x_ref[2]
        ref[1,3,1,0]=(-0.139017942141)*x_ref[0]**o+(-0.338223039119)*x_ref[0]+(0.737882941498)*x_ref[1]**o+(0.948854903087)*x_ref[1]+(-0.0792461633266)*x_ref[2]**o+(0.848875660203)*x_ref[2]
        ref[1,3,1,1]=(0.693284332212)*x_ref[0]**o+(-0.392272875348)*x_ref[0]+(-0.475709045479)*x_ref[1]**o+(0.182948077236)*x_ref[1]+(0.0277569761538)*x_ref[2]**o+(-0.550294947596)*x_ref[2]
        ref[1,3,2,0]=(-0.105381649505)*x_ref[0]**o+(-0.403051690238)*x_ref[0]+(-0.722219735818)*x_ref[1]**o+(-0.380114388869)*x_ref[1]+(0.321452572491)*x_ref[2]**o+(0.858583245253)*x_ref[2]
        ref[1,3,2,1]=(-0.405370062024)*x_ref[0]**o+(0.0743704840047)*x_ref[0]+(0.998947017815)*x_ref[1]**o+(0.983611079522)*x_ref[1]+(0.854202982674)*x_ref[2]**o+(0.288478746498)*x_ref[2]
        ref[1,4,0,0]=(0.0584782489998)*x_ref[0]**o+(0.999286927604)*x_ref[0]+(-0.458166020555)*x_ref[1]**o+(0.529008844128)*x_ref[1]+(0.424592641922)*x_ref[2]**o+(-0.489692340187)*x_ref[2]
        ref[1,4,0,1]=(-0.150555966138)*x_ref[0]**o+(0.362589320838)*x_ref[0]+(0.515820915391)*x_ref[1]**o+(0.742403661312)*x_ref[1]+(0.230278843183)*x_ref[2]**o+(-0.213977724366)*x_ref[2]
        ref[1,4,1,0]=(-0.361161802362)*x_ref[0]**o+(-0.49718332222)*x_ref[0]+(0.143956046197)*x_ref[1]**o+(-0.044550912758)*x_ref[1]+(-0.299892874708)*x_ref[2]**o+(0.370248709792)*x_ref[2]
        ref[1,4,1,1]=(0.259365238173)*x_ref[0]**o+(-0.474694548557)*x_ref[0]+(0.732983396669)*x_ref[1]**o+(-0.295151546774)*x_ref[1]+(-0.337144628526)*x_ref[2]**o+(0.664201775714)*x_ref[2]
        ref[1,4,2,0]=(-0.318016754284)*x_ref[0]**o+(0.555291443852)*x_ref[0]+(0.538150727998)*x_ref[1]**o+(-0.800118353132)*x_ref[1]+(0.652262386596)*x_ref[2]**o+(0.303156645857)*x_ref[2]
        ref[1,4,2,1]=(0.0827142921796)*x_ref[0]**o+(-0.88498648372)*x_ref[0]+(-0.941249844349)*x_ref[1]**o+(-0.162576659039)*x_ref[1]+(0.867621252192)*x_ref[2]**o+(0.0834583795227)*x_ref[2]
        ref[2,0,0,0]=(0.0670088199264)*x_ref[0]**o+(-0.0425961894346)*x_ref[0]+(0.329605789302)*x_ref[1]**o+(0.860453156363)*x_ref[1]+(-0.37410469727)*x_ref[2]**o+(0.55476432384)*x_ref[2]
        ref[2,0,0,1]=(0.801738221879)*x_ref[0]**o+(-0.836563530923)*x_ref[0]+(-0.449651673142)*x_ref[1]**o+(-0.19423133918)*x_ref[1]+(0.947648200044)*x_ref[2]**o+(0.0854067601423)*x_ref[2]
        ref[2,0,1,0]=(0.986595046855)*x_ref[0]**o+(-0.358319352175)*x_ref[0]+(0.493917713513)*x_ref[1]**o+(-0.92052282445)*x_ref[1]+(-0.692726467259)*x_ref[2]**o+(-0.0636943737807)*x_ref[2]
        ref[2,0,1,1]=(0.38491150492)*x_ref[0]**o+(0.766893245938)*x_ref[0]+(0.420188917945)*x_ref[1]**o+(0.43760538854)*x_ref[1]+(-0.770749451981)*x_ref[2]**o+(0.124882164378)*x_ref[2]
        ref[2,0,2,0]=(0.775715718797)*x_ref[0]**o+(0.89516703208)*x_ref[0]+(-0.829021524898)*x_ref[1]**o+(-0.650262454311)*x_ref[1]+(-0.812556336598)*x_ref[2]**o+(-0.997075743997)*x_ref[2]
        ref[2,0,2,1]=(0.3503761008)*x_ref[0]**o+(-0.822009897455)*x_ref[0]+(-0.657657365354)*x_ref[1]**o+(0.476295612156)*x_ref[1]+(0.315554008711)*x_ref[2]**o+(0.691631078049)*x_ref[2]
        ref[2,1,0,0]=(0.680177375433)*x_ref[0]**o+(-0.922726813311)*x_ref[0]+(-0.581116746321)*x_ref[1]**o+(0.628138099343)*x_ref[1]+(-0.568262800293)*x_ref[2]**o+(-0.228201382395)*x_ref[2]
        ref[2,1,0,1]=(0.210094833324)*x_ref[0]**o+(0.281549673876)*x_ref[0]+(0.23785231073)*x_ref[1]**o+(-0.103812455993)*x_ref[1]+(0.530445971191)*x_ref[2]**o+(-0.0830520449939)*x_ref[2]
        ref[2,1,1,0]=(0.367133297563)*x_ref[0]**o+(0.168289954585)*x_ref[0]+(0.614117392787)*x_ref[1]**o+(0.790467766841)*x_ref[1]+(-0.958002150805)*x_ref[2]**o+(0.328916248824)*x_ref[2]
        ref[2,1,1,1]=(0.089535347231)*x_ref[0]**o+(-0.93155993701)*x_ref[0]+(-0.674670760794)*x_ref[1]**o+(-0.274148261773)*x_ref[1]+(-0.807020126056)*x_ref[2]**o+(-0.323334769496)*x_ref[2]
        ref[2,1,2,0]=(-0.948679835219)*x_ref[0]**o+(-0.577235742425)*x_ref[0]+(-0.732584834146)*x_ref[1]**o+(-0.475242504288)*x_ref[1]+(0.213663223262)*x_ref[2]**o+(-0.151401354381)*x_ref[2]
        ref[2,1,2,1]=(0.264513246206)*x_ref[0]**o+(-0.605806437391)*x_ref[0]+(-0.058730926693)*x_ref[1]**o+(-0.661696304412)*x_ref[1]+(-0.361277352457)*x_ref[2]**o+(-0.804408258253)*x_ref[2]
        ref[2,2,0,0]=(-0.144562358054)*x_ref[0]**o+(-0.0960550555725)*x_ref[0]+(-0.905096867803)*x_ref[1]**o+(0.188861784103)*x_ref[1]+(-0.616751005568)*x_ref[2]**o+(0.604711974248)*x_ref[2]
        ref[2,2,0,1]=(-0.0582524307858)*x_ref[0]**o+(-0.503648879785)*x_ref[0]+(-0.682755209956)*x_ref[1]**o+(-0.808213164425)*x_ref[1]+(0.539987276327)*x_ref[2]**o+(0.650557703689)*x_ref[2]
        ref[2,2,1,0]=(0.0607729567733)*x_ref[0]**o+(-0.498819638428)*x_ref[0]+(0.434442595334)*x_ref[1]**o+(-0.491063977603)*x_ref[1]+(0.499436212229)*x_ref[2]**o+(-0.932071991206)*x_ref[2]
        ref[2,2,1,1]=(-0.461295295294)*x_ref[0]**o+(-0.153600767916)*x_ref[0]+(-0.86452392231)*x_ref[1]**o+(0.136760034153)*x_ref[1]+(-0.471642416082)*x_ref[2]**o+(0.453820403438)*x_ref[2]
        ref[2,2,2,0]=(-0.970051888584)*x_ref[0]**o+(0.415728825057)*x_ref[0]+(0.971643909569)*x_ref[1]**o+(0.923719420524)*x_ref[1]+(-0.837313513804)*x_ref[2]**o+(-0.738057676093)*x_ref[2]
        ref[2,2,2,1]=(-0.941355918089)*x_ref[0]**o+(-0.9629784524)*x_ref[0]+(-0.931184348072)*x_ref[1]**o+(0.720006097658)*x_ref[1]+(-0.140750567978)*x_ref[2]**o+(0.55381939119)*x_ref[2]
        ref[2,3,0,0]=(0.753455933669)*x_ref[0]**o+(0.352384922037)*x_ref[0]+(0.737644348002)*x_ref[1]**o+(-0.565144188248)*x_ref[1]+(0.257417624014)*x_ref[2]**o+(-0.660006641683)*x_ref[2]
        ref[2,3,0,1]=(0.886996770925)*x_ref[0]**o+(-0.876595418252)*x_ref[0]+(-0.189778073871)*x_ref[1]**o+(0.560867442336)*x_ref[1]+(-0.262338156346)*x_ref[2]**o+(0.614749098901)*x_ref[2]
        ref[2,3,1,0]=(-0.722912374016)*x_ref[0]**o+(0.447451310273)*x_ref[0]+(0.397510140568)*x_ref[1]**o+(0.932405917932)*x_ref[1]+(-0.206969252336)*x_ref[2]**o+(-0.897366208888)*x_ref[2]
        ref[2,3,1,1]=(0.106324074547)*x_ref[0]**o+(0.822109425884)*x_ref[0]+(0.62699223277)*x_ref[1]**o+(0.238371557839)*x_ref[1]+(-0.422051841243)*x_ref[2]**o+(0.868932130329)*x_ref[2]
        ref[2,3,2,0]=(0.0155790371581)*x_ref[0]**o+(-0.696029577918)*x_ref[0]+(0.259909585217)*x_ref[1]**o+(0.260209121862)*x_ref[1]+(0.780977081459)*x_ref[2]**o+(-0.819957534538)*x_ref[2]
        ref[2,3,2,1]=(0.736223721904)*x_ref[0]**o+(-0.433132598174)*x_ref[0]+(-0.880913649812)*x_ref[1]**o+(-0.375988777461)*x_ref[1]+(-0.351763406307)*x_ref[2]**o+(0.948468252467)*x_ref[2]
        ref[2,4,0,0]=(0.943349180556)*x_ref[0]**o+(-0.720044514103)*x_ref[0]+(0.671900161807)*x_ref[1]**o+(0.358571142942)*x_ref[1]+(-0.132755998648)*x_ref[2]**o+(0.0749815044365)*x_ref[2]
        ref[2,4,0,1]=(0.639070752347)*x_ref[0]**o+(0.551441363406)*x_ref[0]+(0.507547844432)*x_ref[1]**o+(-0.800043772307)*x_ref[1]+(-0.506359503394)*x_ref[2]**o+(0.952191432651)*x_ref[2]
        ref[2,4,1,0]=(-0.655864732254)*x_ref[0]**o+(0.142198390642)*x_ref[0]+(0.075292486702)*x_ref[1]**o+(0.191013189029)*x_ref[1]+(0.106768453247)*x_ref[2]**o+(0.974526515449)*x_ref[2]
        ref[2,4,1,1]=(-0.354750205648)*x_ref[0]**o+(0.141506849368)*x_ref[0]+(-0.170074898141)*x_ref[1]**o+(-0.0882776529335)*x_ref[1]+(0.845254802876)*x_ref[2]**o+(-0.435414692125)*x_ref[2]
        ref[2,4,2,0]=(-0.412836027163)*x_ref[0]**o+(0.830777132582)*x_ref[0]+(-0.87554952236)*x_ref[1]**o+(0.970647516313)*x_ref[1]+(0.989864827286)*x_ref[2]**o+(-0.67790755421)*x_ref[2]
        ref[2,4,2,1]=(0.762907045031)*x_ref[0]**o+(0.140771931118)*x_ref[0]+(0.716859940142)*x_ref[1]**o+(0.151769314189)*x_ref[1]+(-0.812408853675)*x_ref[2]**o+(0.755727078666)*x_ref[2]
        ref[3,0,0,0]=(-0.13358237438)*x_ref[0]**o+(-0.594735856789)*x_ref[0]+(-0.367418402298)*x_ref[1]**o+(0.5349162872)*x_ref[1]+(0.238285320118)*x_ref[2]**o+(0.999627906459)*x_ref[2]
        ref[3,0,0,1]=(-0.537433369218)*x_ref[0]**o+(0.133034652667)*x_ref[0]+(-0.190302464596)*x_ref[1]**o+(0.869919012762)*x_ref[1]+(0.320294306008)*x_ref[2]**o+(0.865908856629)*x_ref[2]
        ref[3,0,1,0]=(0.480851043043)*x_ref[0]**o+(0.833358259022)*x_ref[0]+(0.893051476621)*x_ref[1]**o+(-0.505255581199)*x_ref[1]+(-0.15316145317)*x_ref[2]**o+(-0.937103563776)*x_ref[2]
        ref[3,0,1,1]=(0.765360137605)*x_ref[0]**o+(0.502399552842)*x_ref[0]+(-0.215707512276)*x_ref[1]**o+(-0.402720816967)*x_ref[1]+(0.231360098578)*x_ref[2]**o+(0.732903327067)*x_ref[2]
        ref[3,0,2,0]=(0.120836669045)*x_ref[0]**o+(0.198346794548)*x_ref[0]+(-0.28708447536)*x_ref[1]**o+(0.289672075353)*x_ref[1]+(-0.0305777738322)*x_ref[2]**o+(-0.0686651569373)*x_ref[2]
        ref[3,0,2,1]=(0.0574598844614)*x_ref[0]**o+(0.031092492399)*x_ref[0]+(-0.406805507341)*x_ref[1]**o+(-0.623370509947)*x_ref[1]+(-0.476094160966)*x_ref[2]**o+(0.0840271758875)*x_ref[2]
        ref[3,1,0,0]=(-0.569415249126)*x_ref[0]**o+(-0.339675540399)*x_ref[0]+(0.182536980707)*x_ref[1]**o+(0.727480200453)*x_ref[1]+(0.851672551504)*x_ref[2]**o+(0.817312313075)*x_ref[2]
        ref[3,1,0,1]=(0.921422642904)*x_ref[0]**o+(0.697656143906)*x_ref[0]+(0.969442847752)*x_ref[1]**o+(-0.00957994710064)*x_ref[1]+(-0.315230556399)*x_ref[2]**o+(-0.870759158384)*x_ref[2]
        ref[3,1,1,0]=(-0.813310465463)*x_ref[0]**o+(-0.809362802903)*x_ref[0]+(-0.0685101056428)*x_ref[1]**o+(0.708195651847)*x_ref[1]+(-0.876449456312)*x_ref[2]**o+(0.882503536479)*x_ref[2]
        ref[3,1,1,1]=(-0.277504922905)*x_ref[0]**o+(-0.819428706347)*x_ref[0]+(0.463203258731)*x_ref[1]**o+(0.597493096401)*x_ref[1]+(0.630522187524)*x_ref[2]**o+(0.92480079466)*x_ref[2]
        ref[3,1,2,0]=(-0.703167803764)*x_ref[0]**o+(0.730459205382)*x_ref[0]+(-0.507036921123)*x_ref[1]**o+(0.419358545765)*x_ref[1]+(-0.542758923679)*x_ref[2]**o+(0.831844312917)*x_ref[2]
        ref[3,1,2,1]=(0.135200299704)*x_ref[0]**o+(-0.160734675188)*x_ref[0]+(-0.893851689192)*x_ref[1]**o+(0.748734906185)*x_ref[1]+(0.970464447167)*x_ref[2]**o+(0.277911376757)*x_ref[2]
        ref[3,2,0,0]=(0.750104328784)*x_ref[0]**o+(-0.818350593893)*x_ref[0]+(-0.749900854281)*x_ref[1]**o+(0.289117653816)*x_ref[1]+(-0.219904822162)*x_ref[2]**o+(0.805208050452)*x_ref[2]
        ref[3,2,0,1]=(-0.129614389749)*x_ref[0]**o+(-0.64067443816)*x_ref[0]+(0.513013650233)*x_ref[1]**o+(0.868869035341)*x_ref[1]+(-0.040145121639)*x_ref[2]**o+(-0.282350136119)*x_ref[2]
        ref[3,2,1,0]=(-0.172814929104)*x_ref[0]**o+(-0.296930471915)*x_ref[0]+(-0.796114111585)*x_ref[1]**o+(-0.210181080249)*x_ref[1]+(-0.606107841173)*x_ref[2]**o+(-0.954329084175)*x_ref[2]
        ref[3,2,1,1]=(0.136149261959)*x_ref[0]**o+(-0.613980696755)*x_ref[0]+(0.968430714778)*x_ref[1]**o+(0.683553952099)*x_ref[1]+(-0.321383789121)*x_ref[2]**o+(-0.139570984234)*x_ref[2]
        ref[3,2,2,0]=(-0.273321720654)*x_ref[0]**o+(0.963651110162)*x_ref[0]+(-0.847402611499)*x_ref[1]**o+(0.237131262905)*x_ref[1]+(-0.617296821558)*x_ref[2]**o+(-0.681856394629)*x_ref[2]
        ref[3,2,2,1]=(-0.760071928624)*x_ref[0]**o+(0.0345053779755)*x_ref[0]+(-0.583680069732)*x_ref[1]**o+(0.150497658077)*x_ref[1]+(0.0257096621429)*x_ref[2]**o+(-0.159406203385)*x_ref[2]
        ref[3,3,0,0]=(-0.0668501746947)*x_ref[0]**o+(-0.183645899611)*x_ref[0]+(0.123667845765)*x_ref[1]**o+(0.0928142420352)*x_ref[1]+(-0.933250766317)*x_ref[2]**o+(-0.629986949639)*x_ref[2]
        ref[3,3,0,1]=(-0.0212790928206)*x_ref[0]**o+(0.444036360325)*x_ref[0]+(0.09182358127)*x_ref[1]**o+(0.473088563672)*x_ref[1]+(0.0852925382953)*x_ref[2]**o+(-0.238517929786)*x_ref[2]
        ref[3,3,1,0]=(0.0647235093728)*x_ref[0]**o+(-0.266285898716)*x_ref[0]+(0.944999560866)*x_ref[1]**o+(-0.757696716486)*x_ref[1]+(-0.439165278732)*x_ref[2]**o+(0.960091479774)*x_ref[2]
        ref[3,3,1,1]=(-0.792249238894)*x_ref[0]**o+(-0.568585439587)*x_ref[0]+(-0.0426655335607)*x_ref[1]**o+(0.980004032206)*x_ref[1]+(0.635227812889)*x_ref[2]**o+(0.6687165384)*x_ref[2]
        ref[3,3,2,0]=(0.340468442769)*x_ref[0]**o+(0.791071154735)*x_ref[0]+(-0.817919975547)*x_ref[1]**o+(0.0328383066982)*x_ref[1]+(-0.323093904705)*x_ref[2]**o+(-0.302641620002)*x_ref[2]
        ref[3,3,2,1]=(-0.476501733213)*x_ref[0]**o+(-0.429845943537)*x_ref[0]+(0.518551215268)*x_ref[1]**o+(-0.551690143654)*x_ref[1]+(0.591983049997)*x_ref[2]**o+(0.586292683497)*x_ref[2]
        ref[3,4,0,0]=(0.119036526248)*x_ref[0]**o+(-0.156404044048)*x_ref[0]+(0.158504048681)*x_ref[1]**o+(-0.3779600709)*x_ref[1]+(-0.362107671534)*x_ref[2]**o+(-0.314508628804)*x_ref[2]
        ref[3,4,0,1]=(0.382011103106)*x_ref[0]**o+(-0.123711676964)*x_ref[0]+(-0.939014457264)*x_ref[1]**o+(-0.5986155807)*x_ref[1]+(-0.852092491566)*x_ref[2]**o+(-0.410438924788)*x_ref[2]
        ref[3,4,1,0]=(0.83612586535)*x_ref[0]**o+(-0.606098645092)*x_ref[0]+(0.643469717148)*x_ref[1]**o+(0.845106525654)*x_ref[1]+(0.347455649691)*x_ref[2]**o+(-0.166796580458)*x_ref[2]
        ref[3,4,1,1]=(0.693431758156)*x_ref[0]**o+(-0.313983401147)*x_ref[0]+(-0.862690834893)*x_ref[1]**o+(0.417233988055)*x_ref[1]+(-0.361769997589)*x_ref[2]**o+(-0.22749888397)*x_ref[2]
        ref[3,4,2,0]=(0.680358405387)*x_ref[0]**o+(-0.786041842215)*x_ref[0]+(0.288823579415)*x_ref[1]**o+(-0.961099851196)*x_ref[1]+(0.669431785695)*x_ref[2]**o+(0.956564160268)*x_ref[2]
        ref[3,4,2,1]=(-0.21416500005)*x_ref[0]**o+(0.48495191486)*x_ref[0]+(-0.173465452645)*x_ref[1]**o+(0.662530321557)*x_ref[1]+(0.984573283824)*x_ref[2]**o+(-0.568211383785)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.264333591663)*x[0]**o+(-0.884594789709)*x[0]+(0.809768917911)*x[1]**o+(-0.999064121889)*x[1]
        arg[1]=(-0.476663257532)*x[0]**o+(0.703121978294)*x[0]+(0.768222344055)*x[1]**o+(0.0126615637065)*x[1]
        ref[0]=(0.264333591663)*x_ref[0]**o+(-0.884594789709)*x_ref[0]+(0.809768917911)*x_ref[1]**o+(-0.999064121889)*x_ref[1]
        ref[1]=(-0.476663257532)*x_ref[0]**o+(0.703121978294)*x_ref[0]+(0.768222344055)*x_ref[1]**o+(0.0126615637065)*x_ref[1]
      else:
        arg[0]=(-0.814801897766)*x[0]**o+(0.417782597744)*x[0]+(0.354691012333)*x[1]**o+(-0.0529647248706)*x[1]+(-0.455085903639)*x[2]**o+(-0.501751102496)*x[2]
        arg[1]=(0.165354246971)*x[0]**o+(-0.238754573504)*x[0]+(-0.293225479626)*x[1]**o+(-0.42192384305)*x[1]+(-0.109509840118)*x[2]**o+(0.511666964461)*x[2]
        ref[0]=(-0.814801897766)*x_ref[0]**o+(0.417782597744)*x_ref[0]+(0.354691012333)*x_ref[1]**o+(-0.0529647248706)*x_ref[1]+(-0.455085903639)*x_ref[2]**o+(-0.501751102496)*x_ref[2]
        ref[1]=(0.165354246971)*x_ref[0]**o+(-0.238754573504)*x_ref[0]+(-0.293225479626)*x_ref[1]**o+(-0.42192384305)*x_ref[1]+(-0.109509840118)*x_ref[2]**o+(0.511666964461)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.469225177227)*x[0]**o+(-0.312308556752)*x[0]+(0.0207638302484)*x[1]**o+(-0.42312789395)*x[1]
        arg[0,1]=(0.435465800442)*x[0]**o+(0.66981866069)*x[0]+(-0.563664236008)*x[1]**o+(-0.979640187021)*x[1]
        arg[0,2]=(0.761422556363)*x[0]**o+(-0.374824039734)*x[0]+(-0.258990622087)*x[1]**o+(-0.553555723801)*x[1]
        arg[0,3]=(-0.346737956865)*x[0]**o+(-0.509820084012)*x[0]+(-0.270279814718)*x[1]**o+(-0.77787682055)*x[1]
        arg[0,4]=(-0.607151504575)*x[0]**o+(0.0370061792286)*x[0]+(-0.597194755504)*x[1]**o+(0.463748874183)*x[1]
        arg[1,0]=(-0.614871053235)*x[0]**o+(0.984151090349)*x[0]+(0.176558255935)*x[1]**o+(-0.00143697546002)*x[1]
        arg[1,1]=(-0.678550535699)*x[0]**o+(-0.40068020768)*x[0]+(-0.73609233692)*x[1]**o+(0.204641737633)*x[1]
        arg[1,2]=(0.644566611818)*x[0]**o+(0.676838663571)*x[0]+(0.592803079261)*x[1]**o+(-0.257851556005)*x[1]
        arg[1,3]=(-0.811008507599)*x[0]**o+(-0.810627035248)*x[0]+(0.470993697001)*x[1]**o+(0.524157551045)*x[1]
        arg[1,4]=(0.507989473752)*x[0]**o+(0.474629400564)*x[0]+(0.972372960952)*x[1]**o+(-0.713136826767)*x[1]
        arg[2,0]=(-0.18215096122)*x[0]**o+(0.0926167766546)*x[0]+(-0.526818906453)*x[1]**o+(-0.169071065789)*x[1]
        arg[2,1]=(-0.402265903759)*x[0]**o+(-0.460219787922)*x[0]+(0.160493468523)*x[1]**o+(-0.816631415356)*x[1]
        arg[2,2]=(-0.826968582935)*x[0]**o+(-0.477060281375)*x[0]+(-0.753204398647)*x[1]**o+(0.340944419869)*x[1]
        arg[2,3]=(-0.289758403392)*x[0]**o+(-0.216496476621)*x[0]+(-0.0729120551729)*x[1]**o+(0.0206397202194)*x[1]
        arg[2,4]=(0.900190015504)*x[0]**o+(0.554334407718)*x[0]+(-0.188091500011)*x[1]**o+(-0.239344021421)*x[1]
        arg[3,0]=(-0.846253042969)*x[0]**o+(0.212019924467)*x[0]+(-0.209309311993)*x[1]**o+(0.47889463943)*x[1]
        arg[3,1]=(-0.796067196362)*x[0]**o+(0.570882191233)*x[0]+(0.789930258)*x[1]**o+(-0.687939656508)*x[1]
        arg[3,2]=(0.0603011963958)*x[0]**o+(-0.291794860225)*x[0]+(-0.572598612637)*x[1]**o+(0.308585554728)*x[1]
        arg[3,3]=(0.883976313141)*x[0]**o+(0.857027876956)*x[0]+(-0.991113135222)*x[1]**o+(-0.0649848727938)*x[1]
        arg[3,4]=(-0.420604320052)*x[0]**o+(-0.147041915772)*x[0]+(0.696294525095)*x[1]**o+(-0.45791291995)*x[1]
        ref[0,0]=(-0.469225177227)*x_ref[0]**o+(-0.312308556752)*x_ref[0]+(0.0207638302484)*x_ref[1]**o+(-0.42312789395)*x_ref[1]
        ref[0,1]=(0.435465800442)*x_ref[0]**o+(0.66981866069)*x_ref[0]+(-0.563664236008)*x_ref[1]**o+(-0.979640187021)*x_ref[1]
        ref[0,2]=(0.761422556363)*x_ref[0]**o+(-0.374824039734)*x_ref[0]+(-0.258990622087)*x_ref[1]**o+(-0.553555723801)*x_ref[1]
        ref[0,3]=(-0.346737956865)*x_ref[0]**o+(-0.509820084012)*x_ref[0]+(-0.270279814718)*x_ref[1]**o+(-0.77787682055)*x_ref[1]
        ref[0,4]=(-0.607151504575)*x_ref[0]**o+(0.0370061792286)*x_ref[0]+(-0.597194755504)*x_ref[1]**o+(0.463748874183)*x_ref[1]
        ref[1,0]=(-0.614871053235)*x_ref[0]**o+(0.984151090349)*x_ref[0]+(0.176558255935)*x_ref[1]**o+(-0.00143697546002)*x_ref[1]
        ref[1,1]=(-0.678550535699)*x_ref[0]**o+(-0.40068020768)*x_ref[0]+(-0.73609233692)*x_ref[1]**o+(0.204641737633)*x_ref[1]
        ref[1,2]=(0.644566611818)*x_ref[0]**o+(0.676838663571)*x_ref[0]+(0.592803079261)*x_ref[1]**o+(-0.257851556005)*x_ref[1]
        ref[1,3]=(-0.811008507599)*x_ref[0]**o+(-0.810627035248)*x_ref[0]+(0.470993697001)*x_ref[1]**o+(0.524157551045)*x_ref[1]
        ref[1,4]=(0.507989473752)*x_ref[0]**o+(0.474629400564)*x_ref[0]+(0.972372960952)*x_ref[1]**o+(-0.713136826767)*x_ref[1]
        ref[2,0]=(-0.18215096122)*x_ref[0]**o+(0.0926167766546)*x_ref[0]+(-0.526818906453)*x_ref[1]**o+(-0.169071065789)*x_ref[1]
        ref[2,1]=(-0.402265903759)*x_ref[0]**o+(-0.460219787922)*x_ref[0]+(0.160493468523)*x_ref[1]**o+(-0.816631415356)*x_ref[1]
        ref[2,2]=(-0.826968582935)*x_ref[0]**o+(-0.477060281375)*x_ref[0]+(-0.753204398647)*x_ref[1]**o+(0.340944419869)*x_ref[1]
        ref[2,3]=(-0.289758403392)*x_ref[0]**o+(-0.216496476621)*x_ref[0]+(-0.0729120551729)*x_ref[1]**o+(0.0206397202194)*x_ref[1]
        ref[2,4]=(0.900190015504)*x_ref[0]**o+(0.554334407718)*x_ref[0]+(-0.188091500011)*x_ref[1]**o+(-0.239344021421)*x_ref[1]
        ref[3,0]=(-0.846253042969)*x_ref[0]**o+(0.212019924467)*x_ref[0]+(-0.209309311993)*x_ref[1]**o+(0.47889463943)*x_ref[1]
        ref[3,1]=(-0.796067196362)*x_ref[0]**o+(0.570882191233)*x_ref[0]+(0.789930258)*x_ref[1]**o+(-0.687939656508)*x_ref[1]
        ref[3,2]=(0.0603011963958)*x_ref[0]**o+(-0.291794860225)*x_ref[0]+(-0.572598612637)*x_ref[1]**o+(0.308585554728)*x_ref[1]
        ref[3,3]=(0.883976313141)*x_ref[0]**o+(0.857027876956)*x_ref[0]+(-0.991113135222)*x_ref[1]**o+(-0.0649848727938)*x_ref[1]
        ref[3,4]=(-0.420604320052)*x_ref[0]**o+(-0.147041915772)*x_ref[0]+(0.696294525095)*x_ref[1]**o+(-0.45791291995)*x_ref[1]
      else:
        arg[0,0]=(0.650989082369)*x[0]**o+(-0.291482513935)*x[0]+(-0.791203000903)*x[1]**o+(0.297385821221)*x[1]+(-0.169814832171)*x[2]**o+(-0.125360282254)*x[2]
        arg[0,1]=(0.166614460758)*x[0]**o+(0.0840046062434)*x[0]+(-0.114646748006)*x[1]**o+(-0.872705089326)*x[1]+(-0.687707782796)*x[2]**o+(-0.230067474846)*x[2]
        arg[0,2]=(-0.219927466535)*x[0]**o+(-0.466007112543)*x[0]+(-0.255341401047)*x[1]**o+(0.615864283673)*x[1]+(-0.983984041665)*x[2]**o+(0.305409617185)*x[2]
        arg[0,3]=(-0.0879672823913)*x[0]**o+(-0.956382814278)*x[0]+(-0.3091322614)*x[1]**o+(0.147177613425)*x[1]+(-0.98911137256)*x[2]**o+(-0.750791857003)*x[2]
        arg[0,4]=(0.715277956634)*x[0]**o+(-0.73358535985)*x[0]+(0.14587372226)*x[1]**o+(0.411467948058)*x[1]+(0.925673970492)*x[2]**o+(-0.0678728528684)*x[2]
        arg[1,0]=(0.150067062674)*x[0]**o+(-0.522799232667)*x[0]+(-0.256261929148)*x[1]**o+(-0.316346866901)*x[1]+(-0.849523865324)*x[2]**o+(-0.585421363115)*x[2]
        arg[1,1]=(-0.49684130662)*x[0]**o+(0.868986326423)*x[0]+(0.650641361203)*x[1]**o+(0.558696373611)*x[1]+(0.945790183046)*x[2]**o+(0.379250336542)*x[2]
        arg[1,2]=(-0.67349806443)*x[0]**o+(-0.337586770393)*x[0]+(0.719579305241)*x[1]**o+(-0.300711356473)*x[1]+(-0.350130081791)*x[2]**o+(0.604354519926)*x[2]
        arg[1,3]=(-0.242995155153)*x[0]**o+(0.668842006698)*x[0]+(-0.981902121891)*x[1]**o+(-0.358779020262)*x[1]+(0.298801927955)*x[2]**o+(-0.859807798836)*x[2]
        arg[1,4]=(-0.248474519407)*x[0]**o+(-0.309793754496)*x[0]+(-0.967889559474)*x[1]**o+(0.541761608865)*x[1]+(0.573726565934)*x[2]**o+(-0.876387334383)*x[2]
        arg[2,0]=(0.879662395476)*x[0]**o+(0.172656747739)*x[0]+(0.74129137412)*x[1]**o+(0.236748254611)*x[1]+(0.764805785803)*x[2]**o+(-0.257470519211)*x[2]
        arg[2,1]=(0.44326363254)*x[0]**o+(0.121548286191)*x[0]+(0.605969252858)*x[1]**o+(0.261621973968)*x[1]+(-0.694403080782)*x[2]**o+(0.15039482013)*x[2]
        arg[2,2]=(-0.372758098923)*x[0]**o+(0.577481130716)*x[0]+(-0.120284392247)*x[1]**o+(-0.630815000386)*x[1]+(-0.601595664522)*x[2]**o+(0.806894830983)*x[2]
        arg[2,3]=(0.548413470095)*x[0]**o+(0.804535150265)*x[0]+(0.469766192947)*x[1]**o+(-0.796431483578)*x[1]+(0.434732585165)*x[2]**o+(-0.20040808975)*x[2]
        arg[2,4]=(-0.119065000638)*x[0]**o+(-0.651948510575)*x[0]+(0.868418571224)*x[1]**o+(-0.987758488584)*x[1]+(0.495264099949)*x[2]**o+(-0.928684424484)*x[2]
        arg[3,0]=(-0.712239839816)*x[0]**o+(0.90734601523)*x[0]+(0.901855037856)*x[1]**o+(0.773630983477)*x[1]+(0.599873877074)*x[2]**o+(-0.511864534352)*x[2]
        arg[3,1]=(0.0499812109271)*x[0]**o+(0.959791265887)*x[0]+(0.861696669737)*x[1]**o+(-0.887520952646)*x[1]+(-0.609317719697)*x[2]**o+(-0.314327480839)*x[2]
        arg[3,2]=(0.717930038261)*x[0]**o+(-0.480998377737)*x[0]+(0.890930499387)*x[1]**o+(-0.880493180277)*x[1]+(0.540792831261)*x[2]**o+(0.343985055993)*x[2]
        arg[3,3]=(0.20531032325)*x[0]**o+(0.0400856366105)*x[0]+(0.829583965232)*x[1]**o+(0.701555952938)*x[1]+(0.676177356543)*x[2]**o+(0.934250384088)*x[2]
        arg[3,4]=(-0.426083844591)*x[0]**o+(0.0732912754379)*x[0]+(0.555586413705)*x[1]**o+(0.379137165516)*x[1]+(-0.481638055369)*x[2]**o+(-0.156787271652)*x[2]
        ref[0,0]=(0.650989082369)*x_ref[0]**o+(-0.291482513935)*x_ref[0]+(-0.791203000903)*x_ref[1]**o+(0.297385821221)*x_ref[1]+(-0.169814832171)*x_ref[2]**o+(-0.125360282254)*x_ref[2]
        ref[0,1]=(0.166614460758)*x_ref[0]**o+(0.0840046062434)*x_ref[0]+(-0.114646748006)*x_ref[1]**o+(-0.872705089326)*x_ref[1]+(-0.687707782796)*x_ref[2]**o+(-0.230067474846)*x_ref[2]
        ref[0,2]=(-0.219927466535)*x_ref[0]**o+(-0.466007112543)*x_ref[0]+(-0.255341401047)*x_ref[1]**o+(0.615864283673)*x_ref[1]+(-0.983984041665)*x_ref[2]**o+(0.305409617185)*x_ref[2]
        ref[0,3]=(-0.0879672823913)*x_ref[0]**o+(-0.956382814278)*x_ref[0]+(-0.3091322614)*x_ref[1]**o+(0.147177613425)*x_ref[1]+(-0.98911137256)*x_ref[2]**o+(-0.750791857003)*x_ref[2]
        ref[0,4]=(0.715277956634)*x_ref[0]**o+(-0.73358535985)*x_ref[0]+(0.14587372226)*x_ref[1]**o+(0.411467948058)*x_ref[1]+(0.925673970492)*x_ref[2]**o+(-0.0678728528684)*x_ref[2]
        ref[1,0]=(0.150067062674)*x_ref[0]**o+(-0.522799232667)*x_ref[0]+(-0.256261929148)*x_ref[1]**o+(-0.316346866901)*x_ref[1]+(-0.849523865324)*x_ref[2]**o+(-0.585421363115)*x_ref[2]
        ref[1,1]=(-0.49684130662)*x_ref[0]**o+(0.868986326423)*x_ref[0]+(0.650641361203)*x_ref[1]**o+(0.558696373611)*x_ref[1]+(0.945790183046)*x_ref[2]**o+(0.379250336542)*x_ref[2]
        ref[1,2]=(-0.67349806443)*x_ref[0]**o+(-0.337586770393)*x_ref[0]+(0.719579305241)*x_ref[1]**o+(-0.300711356473)*x_ref[1]+(-0.350130081791)*x_ref[2]**o+(0.604354519926)*x_ref[2]
        ref[1,3]=(-0.242995155153)*x_ref[0]**o+(0.668842006698)*x_ref[0]+(-0.981902121891)*x_ref[1]**o+(-0.358779020262)*x_ref[1]+(0.298801927955)*x_ref[2]**o+(-0.859807798836)*x_ref[2]
        ref[1,4]=(-0.248474519407)*x_ref[0]**o+(-0.309793754496)*x_ref[0]+(-0.967889559474)*x_ref[1]**o+(0.541761608865)*x_ref[1]+(0.573726565934)*x_ref[2]**o+(-0.876387334383)*x_ref[2]
        ref[2,0]=(0.879662395476)*x_ref[0]**o+(0.172656747739)*x_ref[0]+(0.74129137412)*x_ref[1]**o+(0.236748254611)*x_ref[1]+(0.764805785803)*x_ref[2]**o+(-0.257470519211)*x_ref[2]
        ref[2,1]=(0.44326363254)*x_ref[0]**o+(0.121548286191)*x_ref[0]+(0.605969252858)*x_ref[1]**o+(0.261621973968)*x_ref[1]+(-0.694403080782)*x_ref[2]**o+(0.15039482013)*x_ref[2]
        ref[2,2]=(-0.372758098923)*x_ref[0]**o+(0.577481130716)*x_ref[0]+(-0.120284392247)*x_ref[1]**o+(-0.630815000386)*x_ref[1]+(-0.601595664522)*x_ref[2]**o+(0.806894830983)*x_ref[2]
        ref[2,3]=(0.548413470095)*x_ref[0]**o+(0.804535150265)*x_ref[0]+(0.469766192947)*x_ref[1]**o+(-0.796431483578)*x_ref[1]+(0.434732585165)*x_ref[2]**o+(-0.20040808975)*x_ref[2]
        ref[2,4]=(-0.119065000638)*x_ref[0]**o+(-0.651948510575)*x_ref[0]+(0.868418571224)*x_ref[1]**o+(-0.987758488584)*x_ref[1]+(0.495264099949)*x_ref[2]**o+(-0.928684424484)*x_ref[2]
        ref[3,0]=(-0.712239839816)*x_ref[0]**o+(0.90734601523)*x_ref[0]+(0.901855037856)*x_ref[1]**o+(0.773630983477)*x_ref[1]+(0.599873877074)*x_ref[2]**o+(-0.511864534352)*x_ref[2]
        ref[3,1]=(0.0499812109271)*x_ref[0]**o+(0.959791265887)*x_ref[0]+(0.861696669737)*x_ref[1]**o+(-0.887520952646)*x_ref[1]+(-0.609317719697)*x_ref[2]**o+(-0.314327480839)*x_ref[2]
        ref[3,2]=(0.717930038261)*x_ref[0]**o+(-0.480998377737)*x_ref[0]+(0.890930499387)*x_ref[1]**o+(-0.880493180277)*x_ref[1]+(0.540792831261)*x_ref[2]**o+(0.343985055993)*x_ref[2]
        ref[3,3]=(0.20531032325)*x_ref[0]**o+(0.0400856366105)*x_ref[0]+(0.829583965232)*x_ref[1]**o+(0.701555952938)*x_ref[1]+(0.676177356543)*x_ref[2]**o+(0.934250384088)*x_ref[2]
        ref[3,4]=(-0.426083844591)*x_ref[0]**o+(0.0732912754379)*x_ref[0]+(0.555586413705)*x_ref[1]**o+(0.379137165516)*x_ref[1]+(-0.481638055369)*x_ref[2]**o+(-0.156787271652)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.0451039988637)*x[0]**o+(0.159426346984)*x[0]+(0.693271264689)*x[1]**o+(-0.129057826038)*x[1]
        arg[0,0,1]=(0.165096887432)*x[0]**o+(0.421267288474)*x[0]+(0.259248346979)*x[1]**o+(-0.216110688389)*x[1]
        arg[0,1,0]=(-0.419500329289)*x[0]**o+(-0.516326782996)*x[0]+(-0.726209291778)*x[1]**o+(0.80434579811)*x[1]
        arg[0,1,1]=(-0.680429758627)*x[0]**o+(-0.539365634112)*x[0]+(-0.20379891956)*x[1]**o+(-0.760436888151)*x[1]
        arg[1,0,0]=(-0.472638888322)*x[0]**o+(0.982469790589)*x[0]+(-0.383590346736)*x[1]**o+(-0.69130786846)*x[1]
        arg[1,0,1]=(0.334498522873)*x[0]**o+(0.190404359107)*x[0]+(0.210624993429)*x[1]**o+(-0.744921567562)*x[1]
        arg[1,1,0]=(0.632591775382)*x[0]**o+(-0.405215541727)*x[0]+(-0.880471123936)*x[1]**o+(0.0942729925938)*x[1]
        arg[1,1,1]=(-0.837885933046)*x[0]**o+(-0.798882152889)*x[0]+(0.648322104814)*x[1]**o+(-0.357620132856)*x[1]
        arg[2,0,0]=(-0.490592533882)*x[0]**o+(0.540431305386)*x[0]+(-0.777299904284)*x[1]**o+(-0.308682146808)*x[1]
        arg[2,0,1]=(0.971265898941)*x[0]**o+(0.546875456819)*x[0]+(0.627171799906)*x[1]**o+(0.0732789392684)*x[1]
        arg[2,1,0]=(0.178076638358)*x[0]**o+(0.587800697151)*x[0]+(0.748336208169)*x[1]**o+(-0.0341258196613)*x[1]
        arg[2,1,1]=(0.36345594982)*x[0]**o+(-0.899402201831)*x[0]+(0.0206286665096)*x[1]**o+(0.303557722727)*x[1]
        arg[3,0,0]=(-0.544119125777)*x[0]**o+(-0.805891689123)*x[0]+(-0.687902925271)*x[1]**o+(0.346236636736)*x[1]
        arg[3,0,1]=(0.725902084271)*x[0]**o+(0.605461404125)*x[0]+(0.452767036006)*x[1]**o+(0.266584718883)*x[1]
        arg[3,1,0]=(0.322052654392)*x[0]**o+(-0.91949636775)*x[0]+(0.948097272559)*x[1]**o+(-0.463715765101)*x[1]
        arg[3,1,1]=(-0.369192142232)*x[0]**o+(0.616001201865)*x[0]+(-0.257989673651)*x[1]**o+(0.507550625249)*x[1]
        arg[4,0,0]=(0.270343836597)*x[0]**o+(0.851686984957)*x[0]+(0.838756000051)*x[1]**o+(-0.329625536762)*x[1]
        arg[4,0,1]=(0.00026238459417)*x[0]**o+(0.0274581707442)*x[0]+(0.770120998708)*x[1]**o+(0.574619799143)*x[1]
        arg[4,1,0]=(-0.709788768284)*x[0]**o+(0.0856634157603)*x[0]+(-0.294662391976)*x[1]**o+(-0.90639308515)*x[1]
        arg[4,1,1]=(0.223381811001)*x[0]**o+(0.941451590223)*x[0]+(0.917865807171)*x[1]**o+(0.912972271206)*x[1]
        arg[5,0,0]=(-0.731163054411)*x[0]**o+(0.185548632151)*x[0]+(0.907527980776)*x[1]**o+(0.816017248525)*x[1]
        arg[5,0,1]=(0.719177781335)*x[0]**o+(-0.788006096642)*x[0]+(-0.434802402615)*x[1]**o+(0.0249465251663)*x[1]
        arg[5,1,0]=(0.982421863977)*x[0]**o+(0.127223012805)*x[0]+(-0.858491508817)*x[1]**o+(-0.152253965468)*x[1]
        arg[5,1,1]=(-0.779596616421)*x[0]**o+(0.049815217296)*x[0]+(0.892121656728)*x[1]**o+(0.258587664554)*x[1]
        ref[0,0,0]=(0.0451039988637)*x_ref[0]**o+(0.159426346984)*x_ref[0]+(0.693271264689)*x_ref[1]**o+(-0.129057826038)*x_ref[1]
        ref[0,0,1]=(0.165096887432)*x_ref[0]**o+(0.421267288474)*x_ref[0]+(0.259248346979)*x_ref[1]**o+(-0.216110688389)*x_ref[1]
        ref[0,1,0]=(-0.419500329289)*x_ref[0]**o+(-0.516326782996)*x_ref[0]+(-0.726209291778)*x_ref[1]**o+(0.80434579811)*x_ref[1]
        ref[0,1,1]=(-0.680429758627)*x_ref[0]**o+(-0.539365634112)*x_ref[0]+(-0.20379891956)*x_ref[1]**o+(-0.760436888151)*x_ref[1]
        ref[1,0,0]=(-0.472638888322)*x_ref[0]**o+(0.982469790589)*x_ref[0]+(-0.383590346736)*x_ref[1]**o+(-0.69130786846)*x_ref[1]
        ref[1,0,1]=(0.334498522873)*x_ref[0]**o+(0.190404359107)*x_ref[0]+(0.210624993429)*x_ref[1]**o+(-0.744921567562)*x_ref[1]
        ref[1,1,0]=(0.632591775382)*x_ref[0]**o+(-0.405215541727)*x_ref[0]+(-0.880471123936)*x_ref[1]**o+(0.0942729925938)*x_ref[1]
        ref[1,1,1]=(-0.837885933046)*x_ref[0]**o+(-0.798882152889)*x_ref[0]+(0.648322104814)*x_ref[1]**o+(-0.357620132856)*x_ref[1]
        ref[2,0,0]=(-0.490592533882)*x_ref[0]**o+(0.540431305386)*x_ref[0]+(-0.777299904284)*x_ref[1]**o+(-0.308682146808)*x_ref[1]
        ref[2,0,1]=(0.971265898941)*x_ref[0]**o+(0.546875456819)*x_ref[0]+(0.627171799906)*x_ref[1]**o+(0.0732789392684)*x_ref[1]
        ref[2,1,0]=(0.178076638358)*x_ref[0]**o+(0.587800697151)*x_ref[0]+(0.748336208169)*x_ref[1]**o+(-0.0341258196613)*x_ref[1]
        ref[2,1,1]=(0.36345594982)*x_ref[0]**o+(-0.899402201831)*x_ref[0]+(0.0206286665096)*x_ref[1]**o+(0.303557722727)*x_ref[1]
        ref[3,0,0]=(-0.544119125777)*x_ref[0]**o+(-0.805891689123)*x_ref[0]+(-0.687902925271)*x_ref[1]**o+(0.346236636736)*x_ref[1]
        ref[3,0,1]=(0.725902084271)*x_ref[0]**o+(0.605461404125)*x_ref[0]+(0.452767036006)*x_ref[1]**o+(0.266584718883)*x_ref[1]
        ref[3,1,0]=(0.322052654392)*x_ref[0]**o+(-0.91949636775)*x_ref[0]+(0.948097272559)*x_ref[1]**o+(-0.463715765101)*x_ref[1]
        ref[3,1,1]=(-0.369192142232)*x_ref[0]**o+(0.616001201865)*x_ref[0]+(-0.257989673651)*x_ref[1]**o+(0.507550625249)*x_ref[1]
        ref[4,0,0]=(0.270343836597)*x_ref[0]**o+(0.851686984957)*x_ref[0]+(0.838756000051)*x_ref[1]**o+(-0.329625536762)*x_ref[1]
        ref[4,0,1]=(0.00026238459417)*x_ref[0]**o+(0.0274581707442)*x_ref[0]+(0.770120998708)*x_ref[1]**o+(0.574619799143)*x_ref[1]
        ref[4,1,0]=(-0.709788768284)*x_ref[0]**o+(0.0856634157603)*x_ref[0]+(-0.294662391976)*x_ref[1]**o+(-0.90639308515)*x_ref[1]
        ref[4,1,1]=(0.223381811001)*x_ref[0]**o+(0.941451590223)*x_ref[0]+(0.917865807171)*x_ref[1]**o+(0.912972271206)*x_ref[1]
        ref[5,0,0]=(-0.731163054411)*x_ref[0]**o+(0.185548632151)*x_ref[0]+(0.907527980776)*x_ref[1]**o+(0.816017248525)*x_ref[1]
        ref[5,0,1]=(0.719177781335)*x_ref[0]**o+(-0.788006096642)*x_ref[0]+(-0.434802402615)*x_ref[1]**o+(0.0249465251663)*x_ref[1]
        ref[5,1,0]=(0.982421863977)*x_ref[0]**o+(0.127223012805)*x_ref[0]+(-0.858491508817)*x_ref[1]**o+(-0.152253965468)*x_ref[1]
        ref[5,1,1]=(-0.779596616421)*x_ref[0]**o+(0.049815217296)*x_ref[0]+(0.892121656728)*x_ref[1]**o+(0.258587664554)*x_ref[1]
      else:
        arg[0,0,0]=(0.804406156662)*x[0]**o+(0.304298949556)*x[0]+(-9.23949223481e-05)*x[1]**o+(0.521245437597)*x[1]+(0.822885620595)*x[2]**o+(0.44472437921)*x[2]
        arg[0,0,1]=(0.208119461011)*x[0]**o+(-0.967847500694)*x[0]+(-0.182961335156)*x[1]**o+(-0.0510538508635)*x[1]+(0.932285255322)*x[2]**o+(-0.463525963095)*x[2]
        arg[0,1,0]=(-0.0514207812926)*x[0]**o+(-0.730723672364)*x[0]+(-0.471144684451)*x[1]**o+(0.0359308791998)*x[1]+(-0.935079157697)*x[2]**o+(0.841965819816)*x[2]
        arg[0,1,1]=(0.607427052267)*x[0]**o+(-0.344216102016)*x[0]+(0.967271353509)*x[1]**o+(-0.412288382331)*x[1]+(-0.0716617724983)*x[2]**o+(0.56895080823)*x[2]
        arg[1,0,0]=(-0.574314769591)*x[0]**o+(-0.819434106932)*x[0]+(0.143229095273)*x[1]**o+(0.505343958082)*x[1]+(0.497764227469)*x[2]**o+(0.592609757968)*x[2]
        arg[1,0,1]=(0.414984900558)*x[0]**o+(0.624574053857)*x[0]+(-0.367330776954)*x[1]**o+(-0.30093924896)*x[1]+(-0.99416380749)*x[2]**o+(0.492443599924)*x[2]
        arg[1,1,0]=(0.171673119954)*x[0]**o+(0.612986140068)*x[0]+(-0.671355907363)*x[1]**o+(-0.292080539973)*x[1]+(-0.450890819125)*x[2]**o+(-0.64742047997)*x[2]
        arg[1,1,1]=(0.452494887295)*x[0]**o+(0.187308285528)*x[0]+(-0.995316825981)*x[1]**o+(-0.825194018202)*x[1]+(-0.00651224511881)*x[2]**o+(-0.715747547803)*x[2]
        arg[2,0,0]=(0.546297169466)*x[0]**o+(-0.424626884041)*x[0]+(0.327647062351)*x[1]**o+(-0.050651105527)*x[1]+(-0.667117310429)*x[2]**o+(-0.37659304263)*x[2]
        arg[2,0,1]=(-0.323595958622)*x[0]**o+(0.541718593289)*x[0]+(0.6285998056)*x[1]**o+(0.488539567593)*x[1]+(0.983086232412)*x[2]**o+(0.692821558469)*x[2]
        arg[2,1,0]=(0.16009856161)*x[0]**o+(-0.776069988482)*x[0]+(0.750717144319)*x[1]**o+(-0.244287895989)*x[1]+(0.345242766297)*x[2]**o+(0.572063651895)*x[2]
        arg[2,1,1]=(-0.34626843638)*x[0]**o+(0.197125033681)*x[0]+(-0.570015603866)*x[1]**o+(-0.633161422872)*x[1]+(-0.602998606808)*x[2]**o+(-0.674829268895)*x[2]
        arg[3,0,0]=(-0.659696747189)*x[0]**o+(-0.571128578512)*x[0]+(-0.704414117225)*x[1]**o+(-0.334924171135)*x[1]+(-0.0200665796175)*x[2]**o+(-0.660688341854)*x[2]
        arg[3,0,1]=(-0.497019087713)*x[0]**o+(0.25850813993)*x[0]+(-0.713426904693)*x[1]**o+(-0.994061006282)*x[1]+(-0.669422282701)*x[2]**o+(-0.553981025574)*x[2]
        arg[3,1,0]=(-0.151249618276)*x[0]**o+(0.29432574729)*x[0]+(0.173047919293)*x[1]**o+(0.844456199843)*x[1]+(0.845402272715)*x[2]**o+(-0.115698924822)*x[2]
        arg[3,1,1]=(0.840440627264)*x[0]**o+(0.593400475213)*x[0]+(0.974677794562)*x[1]**o+(0.938845603167)*x[1]+(-0.944730436891)*x[2]**o+(-0.657543633218)*x[2]
        arg[4,0,0]=(0.968290161629)*x[0]**o+(0.578471458995)*x[0]+(0.946558149029)*x[1]**o+(-0.232692317459)*x[1]+(-0.97762469848)*x[2]**o+(0.660519364613)*x[2]
        arg[4,0,1]=(-0.187317070761)*x[0]**o+(-0.0363170179961)*x[0]+(0.278499613269)*x[1]**o+(-0.0685831047696)*x[1]+(-0.540282212359)*x[2]**o+(0.564656560066)*x[2]
        arg[4,1,0]=(-0.211825530749)*x[0]**o+(0.225546568724)*x[0]+(-0.83241974026)*x[1]**o+(-0.503828393954)*x[1]+(-0.635937384108)*x[2]**o+(-0.513183062177)*x[2]
        arg[4,1,1]=(-0.257123983215)*x[0]**o+(-0.0700924567638)*x[0]+(0.550859393713)*x[1]**o+(0.132070212001)*x[1]+(-0.478549332853)*x[2]**o+(-0.560087260102)*x[2]
        arg[5,0,0]=(0.677259763344)*x[0]**o+(-0.213422444013)*x[0]+(-0.294112138735)*x[1]**o+(0.083135232348)*x[1]+(-0.899953896237)*x[2]**o+(-0.117143433064)*x[2]
        arg[5,0,1]=(-0.489822592306)*x[0]**o+(0.368217759904)*x[0]+(-0.927837261915)*x[1]**o+(-0.708243327427)*x[1]+(-0.132256551163)*x[2]**o+(0.316204052014)*x[2]
        arg[5,1,0]=(0.314897334392)*x[0]**o+(-0.848168680639)*x[0]+(0.92421318346)*x[1]**o+(0.995505186786)*x[1]+(0.372381949158)*x[2]**o+(-0.563415970977)*x[2]
        arg[5,1,1]=(0.745011483885)*x[0]**o+(-0.6423283108)*x[0]+(-0.119592965986)*x[1]**o+(0.241068250891)*x[1]+(0.477172198315)*x[2]**o+(-0.468580220406)*x[2]
        ref[0,0,0]=(0.804406156662)*x_ref[0]**o+(0.304298949556)*x_ref[0]+(-9.23949223481e-05)*x_ref[1]**o+(0.521245437597)*x_ref[1]+(0.822885620595)*x_ref[2]**o+(0.44472437921)*x_ref[2]
        ref[0,0,1]=(0.208119461011)*x_ref[0]**o+(-0.967847500694)*x_ref[0]+(-0.182961335156)*x_ref[1]**o+(-0.0510538508635)*x_ref[1]+(0.932285255322)*x_ref[2]**o+(-0.463525963095)*x_ref[2]
        ref[0,1,0]=(-0.0514207812926)*x_ref[0]**o+(-0.730723672364)*x_ref[0]+(-0.471144684451)*x_ref[1]**o+(0.0359308791998)*x_ref[1]+(-0.935079157697)*x_ref[2]**o+(0.841965819816)*x_ref[2]
        ref[0,1,1]=(0.607427052267)*x_ref[0]**o+(-0.344216102016)*x_ref[0]+(0.967271353509)*x_ref[1]**o+(-0.412288382331)*x_ref[1]+(-0.0716617724983)*x_ref[2]**o+(0.56895080823)*x_ref[2]
        ref[1,0,0]=(-0.574314769591)*x_ref[0]**o+(-0.819434106932)*x_ref[0]+(0.143229095273)*x_ref[1]**o+(0.505343958082)*x_ref[1]+(0.497764227469)*x_ref[2]**o+(0.592609757968)*x_ref[2]
        ref[1,0,1]=(0.414984900558)*x_ref[0]**o+(0.624574053857)*x_ref[0]+(-0.367330776954)*x_ref[1]**o+(-0.30093924896)*x_ref[1]+(-0.99416380749)*x_ref[2]**o+(0.492443599924)*x_ref[2]
        ref[1,1,0]=(0.171673119954)*x_ref[0]**o+(0.612986140068)*x_ref[0]+(-0.671355907363)*x_ref[1]**o+(-0.292080539973)*x_ref[1]+(-0.450890819125)*x_ref[2]**o+(-0.64742047997)*x_ref[2]
        ref[1,1,1]=(0.452494887295)*x_ref[0]**o+(0.187308285528)*x_ref[0]+(-0.995316825981)*x_ref[1]**o+(-0.825194018202)*x_ref[1]+(-0.00651224511881)*x_ref[2]**o+(-0.715747547803)*x_ref[2]
        ref[2,0,0]=(0.546297169466)*x_ref[0]**o+(-0.424626884041)*x_ref[0]+(0.327647062351)*x_ref[1]**o+(-0.050651105527)*x_ref[1]+(-0.667117310429)*x_ref[2]**o+(-0.37659304263)*x_ref[2]
        ref[2,0,1]=(-0.323595958622)*x_ref[0]**o+(0.541718593289)*x_ref[0]+(0.6285998056)*x_ref[1]**o+(0.488539567593)*x_ref[1]+(0.983086232412)*x_ref[2]**o+(0.692821558469)*x_ref[2]
        ref[2,1,0]=(0.16009856161)*x_ref[0]**o+(-0.776069988482)*x_ref[0]+(0.750717144319)*x_ref[1]**o+(-0.244287895989)*x_ref[1]+(0.345242766297)*x_ref[2]**o+(0.572063651895)*x_ref[2]
        ref[2,1,1]=(-0.34626843638)*x_ref[0]**o+(0.197125033681)*x_ref[0]+(-0.570015603866)*x_ref[1]**o+(-0.633161422872)*x_ref[1]+(-0.602998606808)*x_ref[2]**o+(-0.674829268895)*x_ref[2]
        ref[3,0,0]=(-0.659696747189)*x_ref[0]**o+(-0.571128578512)*x_ref[0]+(-0.704414117225)*x_ref[1]**o+(-0.334924171135)*x_ref[1]+(-0.0200665796175)*x_ref[2]**o+(-0.660688341854)*x_ref[2]
        ref[3,0,1]=(-0.497019087713)*x_ref[0]**o+(0.25850813993)*x_ref[0]+(-0.713426904693)*x_ref[1]**o+(-0.994061006282)*x_ref[1]+(-0.669422282701)*x_ref[2]**o+(-0.553981025574)*x_ref[2]
        ref[3,1,0]=(-0.151249618276)*x_ref[0]**o+(0.29432574729)*x_ref[0]+(0.173047919293)*x_ref[1]**o+(0.844456199843)*x_ref[1]+(0.845402272715)*x_ref[2]**o+(-0.115698924822)*x_ref[2]
        ref[3,1,1]=(0.840440627264)*x_ref[0]**o+(0.593400475213)*x_ref[0]+(0.974677794562)*x_ref[1]**o+(0.938845603167)*x_ref[1]+(-0.944730436891)*x_ref[2]**o+(-0.657543633218)*x_ref[2]
        ref[4,0,0]=(0.968290161629)*x_ref[0]**o+(0.578471458995)*x_ref[0]+(0.946558149029)*x_ref[1]**o+(-0.232692317459)*x_ref[1]+(-0.97762469848)*x_ref[2]**o+(0.660519364613)*x_ref[2]
        ref[4,0,1]=(-0.187317070761)*x_ref[0]**o+(-0.0363170179961)*x_ref[0]+(0.278499613269)*x_ref[1]**o+(-0.0685831047696)*x_ref[1]+(-0.540282212359)*x_ref[2]**o+(0.564656560066)*x_ref[2]
        ref[4,1,0]=(-0.211825530749)*x_ref[0]**o+(0.225546568724)*x_ref[0]+(-0.83241974026)*x_ref[1]**o+(-0.503828393954)*x_ref[1]+(-0.635937384108)*x_ref[2]**o+(-0.513183062177)*x_ref[2]
        ref[4,1,1]=(-0.257123983215)*x_ref[0]**o+(-0.0700924567638)*x_ref[0]+(0.550859393713)*x_ref[1]**o+(0.132070212001)*x_ref[1]+(-0.478549332853)*x_ref[2]**o+(-0.560087260102)*x_ref[2]
        ref[5,0,0]=(0.677259763344)*x_ref[0]**o+(-0.213422444013)*x_ref[0]+(-0.294112138735)*x_ref[1]**o+(0.083135232348)*x_ref[1]+(-0.899953896237)*x_ref[2]**o+(-0.117143433064)*x_ref[2]
        ref[5,0,1]=(-0.489822592306)*x_ref[0]**o+(0.368217759904)*x_ref[0]+(-0.927837261915)*x_ref[1]**o+(-0.708243327427)*x_ref[1]+(-0.132256551163)*x_ref[2]**o+(0.316204052014)*x_ref[2]
        ref[5,1,0]=(0.314897334392)*x_ref[0]**o+(-0.848168680639)*x_ref[0]+(0.92421318346)*x_ref[1]**o+(0.995505186786)*x_ref[1]+(0.372381949158)*x_ref[2]**o+(-0.563415970977)*x_ref[2]
        ref[5,1,1]=(0.745011483885)*x_ref[0]**o+(-0.6423283108)*x_ref[0]+(-0.119592965986)*x_ref[1]**o+(0.241068250891)*x_ref[1]+(0.477172198315)*x_ref[2]**o+(-0.468580220406)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.633576516648)*x[0]**o+(0.77399256211)*x[0]+(-0.396696417195)*x[1]**o+(-0.604572540725)*x[1]
        arg[0,0,0,1]=(0.612672991978)*x[0]**o+(-0.331012302611)*x[0]+(0.885533578235)*x[1]**o+(0.334686818058)*x[1]
        arg[0,0,1,0]=(-0.982701069025)*x[0]**o+(0.467849033642)*x[0]+(0.525392773375)*x[1]**o+(0.355622324425)*x[1]
        arg[0,0,1,1]=(-0.0243671469666)*x[0]**o+(0.790044842383)*x[0]+(0.413384172033)*x[1]**o+(0.740805135968)*x[1]
        arg[0,0,2,0]=(0.0384789957691)*x[0]**o+(-0.317009899417)*x[0]+(0.0389710369615)*x[1]**o+(0.784549852598)*x[1]
        arg[0,0,2,1]=(0.361832259252)*x[0]**o+(0.410815792944)*x[0]+(0.39496145831)*x[1]**o+(-0.820784143964)*x[1]
        arg[0,1,0,0]=(0.224261088976)*x[0]**o+(-0.824694527256)*x[0]+(0.318674472135)*x[1]**o+(0.827696354338)*x[1]
        arg[0,1,0,1]=(-0.442154901359)*x[0]**o+(0.141775320862)*x[0]+(0.688711841268)*x[1]**o+(-0.27047542755)*x[1]
        arg[0,1,1,0]=(-0.570125802019)*x[0]**o+(0.050303762895)*x[0]+(-0.972329837417)*x[1]**o+(0.143700303765)*x[1]
        arg[0,1,1,1]=(0.852888550499)*x[0]**o+(-0.51216090329)*x[0]+(-0.895663922335)*x[1]**o+(-0.183466225547)*x[1]
        arg[0,1,2,0]=(0.863840920739)*x[0]**o+(-0.247789353524)*x[0]+(0.63056436157)*x[1]**o+(0.932522589942)*x[1]
        arg[0,1,2,1]=(-0.403832121794)*x[0]**o+(-0.769625259789)*x[0]+(0.507724303191)*x[1]**o+(0.314178775526)*x[1]
        arg[0,2,0,0]=(0.543867153515)*x[0]**o+(-0.414107817028)*x[0]+(0.423922720842)*x[1]**o+(0.166632407916)*x[1]
        arg[0,2,0,1]=(0.731968209929)*x[0]**o+(-0.114742736683)*x[0]+(0.112562415021)*x[1]**o+(0.166924480325)*x[1]
        arg[0,2,1,0]=(0.919472066242)*x[0]**o+(0.560565315205)*x[0]+(0.4967650028)*x[1]**o+(0.708376671625)*x[1]
        arg[0,2,1,1]=(0.147500239739)*x[0]**o+(-0.632670389919)*x[0]+(-0.992479063437)*x[1]**o+(0.258941122962)*x[1]
        arg[0,2,2,0]=(0.474481785329)*x[0]**o+(0.592902604094)*x[0]+(0.64058685211)*x[1]**o+(-0.477843307901)*x[1]
        arg[0,2,2,1]=(0.769898355146)*x[0]**o+(0.563458279487)*x[0]+(0.876276425449)*x[1]**o+(-0.0625727749869)*x[1]
        arg[0,3,0,0]=(-0.330055037059)*x[0]**o+(-0.727656105163)*x[0]+(-0.559901684392)*x[1]**o+(-0.433646878263)*x[1]
        arg[0,3,0,1]=(-0.725314328666)*x[0]**o+(0.638326609105)*x[0]+(-0.126619368417)*x[1]**o+(-0.232194758676)*x[1]
        arg[0,3,1,0]=(0.318096005545)*x[0]**o+(0.325247494289)*x[0]+(-0.514305253643)*x[1]**o+(0.108523350086)*x[1]
        arg[0,3,1,1]=(0.298034894956)*x[0]**o+(0.825714393651)*x[0]+(-0.67162850079)*x[1]**o+(0.681809083004)*x[1]
        arg[0,3,2,0]=(-0.00715890582648)*x[0]**o+(0.893900731227)*x[0]+(-0.773710274545)*x[1]**o+(0.415913728569)*x[1]
        arg[0,3,2,1]=(0.0853551940958)*x[0]**o+(-0.228876011127)*x[0]+(0.87699456762)*x[1]**o+(-0.744605229253)*x[1]
        arg[0,4,0,0]=(0.942646937302)*x[0]**o+(0.92717236759)*x[0]+(0.794806821569)*x[1]**o+(0.848766303526)*x[1]
        arg[0,4,0,1]=(0.20140672561)*x[0]**o+(0.974490372414)*x[0]+(-0.450332419587)*x[1]**o+(-0.859102184782)*x[1]
        arg[0,4,1,0]=(0.632153053193)*x[0]**o+(0.702336756277)*x[0]+(-0.571496076236)*x[1]**o+(-0.882472852219)*x[1]
        arg[0,4,1,1]=(0.366452829499)*x[0]**o+(-0.514660164741)*x[0]+(-0.152794832253)*x[1]**o+(-0.658688117796)*x[1]
        arg[0,4,2,0]=(0.406218394702)*x[0]**o+(0.782359698198)*x[0]+(-0.0582499587562)*x[1]**o+(0.664506852595)*x[1]
        arg[0,4,2,1]=(0.768240870343)*x[0]**o+(0.188470592904)*x[0]+(0.855933509171)*x[1]**o+(0.639174087541)*x[1]
        arg[1,0,0,0]=(-0.946841398046)*x[0]**o+(0.892729256183)*x[0]+(-0.15644094115)*x[1]**o+(-0.859371368233)*x[1]
        arg[1,0,0,1]=(0.238113647666)*x[0]**o+(0.173695470467)*x[0]+(-0.753529252313)*x[1]**o+(-0.243203902463)*x[1]
        arg[1,0,1,0]=(0.267937523443)*x[0]**o+(0.600200492374)*x[0]+(0.442653529257)*x[1]**o+(-0.394984210541)*x[1]
        arg[1,0,1,1]=(-0.495022907622)*x[0]**o+(-0.593273808093)*x[0]+(0.888764746608)*x[1]**o+(-0.745127825763)*x[1]
        arg[1,0,2,0]=(0.352827238496)*x[0]**o+(0.293200928839)*x[0]+(-0.00369812109034)*x[1]**o+(0.478875774193)*x[1]
        arg[1,0,2,1]=(-0.403358878981)*x[0]**o+(0.756947364667)*x[0]+(0.649335795039)*x[1]**o+(-0.339962796859)*x[1]
        arg[1,1,0,0]=(-0.607505349015)*x[0]**o+(-0.392535399582)*x[0]+(0.799782389634)*x[1]**o+(0.186339165863)*x[1]
        arg[1,1,0,1]=(0.997492304627)*x[0]**o+(-0.281456622581)*x[0]+(-0.959294693613)*x[1]**o+(-0.454812606674)*x[1]
        arg[1,1,1,0]=(0.454682037549)*x[0]**o+(0.797129935647)*x[0]+(-0.949539573053)*x[1]**o+(-0.918499382199)*x[1]
        arg[1,1,1,1]=(0.903016449758)*x[0]**o+(-0.575039046836)*x[0]+(0.957158277074)*x[1]**o+(-0.715797460906)*x[1]
        arg[1,1,2,0]=(0.92005547899)*x[0]**o+(-0.716975919302)*x[0]+(-0.900904311708)*x[1]**o+(0.989458809643)*x[1]
        arg[1,1,2,1]=(-0.633214991087)*x[0]**o+(0.680658578627)*x[0]+(0.552194054556)*x[1]**o+(0.343557072769)*x[1]
        arg[1,2,0,0]=(-0.0182903530463)*x[0]**o+(-0.0445569497267)*x[0]+(-0.259920069365)*x[1]**o+(0.118765787279)*x[1]
        arg[1,2,0,1]=(-0.667382174733)*x[0]**o+(-0.79898800024)*x[0]+(-0.217638556567)*x[1]**o+(-0.0288504127936)*x[1]
        arg[1,2,1,0]=(-0.437982445862)*x[0]**o+(-0.104054454808)*x[0]+(0.711429718088)*x[1]**o+(-0.607226296329)*x[1]
        arg[1,2,1,1]=(0.252861218037)*x[0]**o+(-0.635701456504)*x[0]+(-0.555222304592)*x[1]**o+(0.723744010092)*x[1]
        arg[1,2,2,0]=(-0.73532710885)*x[0]**o+(0.600599078741)*x[0]+(0.0465795173194)*x[1]**o+(0.524165544645)*x[1]
        arg[1,2,2,1]=(-0.025246019744)*x[0]**o+(-0.376526675619)*x[0]+(0.213011788698)*x[1]**o+(-0.477018225777)*x[1]
        arg[1,3,0,0]=(-0.52682690957)*x[0]**o+(0.0567472858499)*x[0]+(0.574116256102)*x[1]**o+(-0.130012791898)*x[1]
        arg[1,3,0,1]=(0.113502182936)*x[0]**o+(0.7643861313)*x[0]+(0.48414948847)*x[1]**o+(0.863168250318)*x[1]
        arg[1,3,1,0]=(-0.208102162378)*x[0]**o+(0.327770728981)*x[0]+(0.843778895278)*x[1]**o+(0.474843504177)*x[1]
        arg[1,3,1,1]=(-0.549818192184)*x[0]**o+(-0.863333816685)*x[0]+(-0.365631583957)*x[1]**o+(-0.491320441182)*x[1]
        arg[1,3,2,0]=(-0.265833845759)*x[0]**o+(-0.94416746081)*x[0]+(0.48465442593)*x[1]**o+(0.258284621408)*x[1]
        arg[1,3,2,1]=(-0.625716082268)*x[0]**o+(-0.234697130359)*x[0]+(-0.973704406131)*x[1]**o+(0.00141746117847)*x[1]
        arg[1,4,0,0]=(0.565044361964)*x[0]**o+(-0.0672474991748)*x[0]+(-0.125901930524)*x[1]**o+(-0.753103899732)*x[1]
        arg[1,4,0,1]=(-0.979743467289)*x[0]**o+(-0.25225085214)*x[0]+(-0.141635164635)*x[1]**o+(0.378153168667)*x[1]
        arg[1,4,1,0]=(0.97728399206)*x[0]**o+(-0.202043160532)*x[0]+(0.997445424954)*x[1]**o+(-0.100049830154)*x[1]
        arg[1,4,1,1]=(0.364840105486)*x[0]**o+(-0.682234327437)*x[0]+(0.252584592259)*x[1]**o+(-0.0785430242425)*x[1]
        arg[1,4,2,0]=(-0.0168764722616)*x[0]**o+(-0.997460139026)*x[0]+(0.560548548912)*x[1]**o+(0.220532224022)*x[1]
        arg[1,4,2,1]=(0.307927943715)*x[0]**o+(0.496390872708)*x[0]+(0.982750813274)*x[1]**o+(0.392968327778)*x[1]
        arg[2,0,0,0]=(0.645991188948)*x[0]**o+(0.104035822886)*x[0]+(0.18285829097)*x[1]**o+(0.51186286294)*x[1]
        arg[2,0,0,1]=(-0.996971717318)*x[0]**o+(0.126488047141)*x[0]+(0.123585312562)*x[1]**o+(-0.775854123529)*x[1]
        arg[2,0,1,0]=(-0.0258590124655)*x[0]**o+(0.0079454735671)*x[0]+(-0.42819024422)*x[1]**o+(0.12377360967)*x[1]
        arg[2,0,1,1]=(0.990441351225)*x[0]**o+(-0.453806376472)*x[0]+(0.134316618645)*x[1]**o+(-0.902666953282)*x[1]
        arg[2,0,2,0]=(0.149304385108)*x[0]**o+(-0.226132046999)*x[0]+(0.647186106203)*x[1]**o+(-0.269247029225)*x[1]
        arg[2,0,2,1]=(-0.159068164422)*x[0]**o+(-0.392275577366)*x[0]+(0.131423093404)*x[1]**o+(-0.353311282658)*x[1]
        arg[2,1,0,0]=(-0.2862918681)*x[0]**o+(-0.392264098939)*x[0]+(-0.344033237285)*x[1]**o+(0.644261713963)*x[1]
        arg[2,1,0,1]=(0.398022911658)*x[0]**o+(-0.84552546032)*x[0]+(-0.823861260943)*x[1]**o+(0.739994988755)*x[1]
        arg[2,1,1,0]=(-0.696854341234)*x[0]**o+(-0.810851183298)*x[0]+(-0.712182042896)*x[1]**o+(0.594887691783)*x[1]
        arg[2,1,1,1]=(-0.182583840117)*x[0]**o+(0.962594377088)*x[0]+(-0.0157575661453)*x[1]**o+(-0.432167995007)*x[1]
        arg[2,1,2,0]=(0.293029196216)*x[0]**o+(0.46839918212)*x[0]+(0.262375931577)*x[1]**o+(0.0348536956306)*x[1]
        arg[2,1,2,1]=(0.775632135095)*x[0]**o+(-0.874598534107)*x[0]+(0.711602976736)*x[1]**o+(0.37065543293)*x[1]
        arg[2,2,0,0]=(-0.103625214158)*x[0]**o+(0.553558117877)*x[0]+(-0.0522465477589)*x[1]**o+(-0.747139195529)*x[1]
        arg[2,2,0,1]=(0.507427663615)*x[0]**o+(-0.536361186134)*x[0]+(0.809054726193)*x[1]**o+(-0.979546347794)*x[1]
        arg[2,2,1,0]=(0.786210623833)*x[0]**o+(0.244510127091)*x[0]+(0.159799334111)*x[1]**o+(0.0499263839273)*x[1]
        arg[2,2,1,1]=(0.454205244371)*x[0]**o+(-0.650927822092)*x[0]+(-0.939252656087)*x[1]**o+(-0.894014755314)*x[1]
        arg[2,2,2,0]=(-0.829617920904)*x[0]**o+(0.764070900994)*x[0]+(0.972419443556)*x[1]**o+(-0.824913543157)*x[1]
        arg[2,2,2,1]=(0.694310748478)*x[0]**o+(-0.191413364091)*x[0]+(0.588258464472)*x[1]**o+(0.743973223755)*x[1]
        arg[2,3,0,0]=(-0.588397365184)*x[0]**o+(-0.634207034861)*x[0]+(0.0853498555546)*x[1]**o+(0.559933556865)*x[1]
        arg[2,3,0,1]=(0.626829165689)*x[0]**o+(0.942185296616)*x[0]+(0.0360770405565)*x[1]**o+(0.635966218547)*x[1]
        arg[2,3,1,0]=(-0.0103526873852)*x[0]**o+(0.546330663496)*x[0]+(0.799224103454)*x[1]**o+(0.791922324897)*x[1]
        arg[2,3,1,1]=(0.921564474495)*x[0]**o+(-0.420410835144)*x[0]+(-0.20482503843)*x[1]**o+(-0.483331643037)*x[1]
        arg[2,3,2,0]=(-0.903600870314)*x[0]**o+(0.901956716274)*x[0]+(-0.113926342023)*x[1]**o+(-0.061596184768)*x[1]
        arg[2,3,2,1]=(0.653250267262)*x[0]**o+(-0.13770933726)*x[0]+(-0.0791563636079)*x[1]**o+(-0.578313150893)*x[1]
        arg[2,4,0,0]=(0.993903405725)*x[0]**o+(0.694324226249)*x[0]+(0.203872883541)*x[1]**o+(-0.526880767264)*x[1]
        arg[2,4,0,1]=(0.0120058513647)*x[0]**o+(0.00907603337269)*x[0]+(0.0247958601677)*x[1]**o+(-0.996916120875)*x[1]
        arg[2,4,1,0]=(0.281518049478)*x[0]**o+(0.0773799323777)*x[0]+(-0.968565098773)*x[1]**o+(-0.275558076229)*x[1]
        arg[2,4,1,1]=(0.565275744757)*x[0]**o+(0.978636675869)*x[0]+(-0.352951657751)*x[1]**o+(-0.54170487914)*x[1]
        arg[2,4,2,0]=(-0.571207550678)*x[0]**o+(0.196754283318)*x[0]+(0.665558917127)*x[1]**o+(0.361915269355)*x[1]
        arg[2,4,2,1]=(0.54579191885)*x[0]**o+(0.798319371798)*x[0]+(0.961669996133)*x[1]**o+(-0.549305529342)*x[1]
        arg[3,0,0,0]=(0.264064946121)*x[0]**o+(-0.314396185593)*x[0]+(-0.017492206135)*x[1]**o+(0.190513393091)*x[1]
        arg[3,0,0,1]=(0.812213666919)*x[0]**o+(-0.757919357489)*x[0]+(0.294829087285)*x[1]**o+(0.916181101963)*x[1]
        arg[3,0,1,0]=(-0.529906091574)*x[0]**o+(0.553682716765)*x[0]+(-0.693177625995)*x[1]**o+(0.377145561683)*x[1]
        arg[3,0,1,1]=(-0.298942248169)*x[0]**o+(0.876878593705)*x[0]+(-0.609807982832)*x[1]**o+(-0.574954919539)*x[1]
        arg[3,0,2,0]=(-0.729324018245)*x[0]**o+(-0.495194565254)*x[0]+(-0.616162751356)*x[1]**o+(0.878549207376)*x[1]
        arg[3,0,2,1]=(0.625373988248)*x[0]**o+(0.106736108396)*x[0]+(-0.871225833944)*x[1]**o+(0.840614493429)*x[1]
        arg[3,1,0,0]=(-0.90880839589)*x[0]**o+(0.390298288126)*x[0]+(0.446872042362)*x[1]**o+(-0.0703294219415)*x[1]
        arg[3,1,0,1]=(-0.0706115592617)*x[0]**o+(-0.158386152093)*x[0]+(-0.687459181052)*x[1]**o+(-0.251170026598)*x[1]
        arg[3,1,1,0]=(-0.286907297081)*x[0]**o+(0.185480985579)*x[0]+(-0.49763263834)*x[1]**o+(-0.773056048642)*x[1]
        arg[3,1,1,1]=(-0.849572703562)*x[0]**o+(0.481739532516)*x[0]+(-0.227573063955)*x[1]**o+(-0.282367611381)*x[1]
        arg[3,1,2,0]=(-0.524975224603)*x[0]**o+(-0.795450756917)*x[0]+(-0.656865716441)*x[1]**o+(-0.0166902340102)*x[1]
        arg[3,1,2,1]=(-0.297441818145)*x[0]**o+(-0.726759852457)*x[0]+(0.146859888138)*x[1]**o+(-0.73690394554)*x[1]
        arg[3,2,0,0]=(-0.278794538596)*x[0]**o+(-0.463428732496)*x[0]+(-0.478500378607)*x[1]**o+(0.806643392698)*x[1]
        arg[3,2,0,1]=(0.573853299578)*x[0]**o+(-0.324830388013)*x[0]+(-0.185528736416)*x[1]**o+(0.444228024864)*x[1]
        arg[3,2,1,0]=(0.686054084195)*x[0]**o+(-0.914380567743)*x[0]+(0.751959744445)*x[1]**o+(0.531759675659)*x[1]
        arg[3,2,1,1]=(-0.876330263589)*x[0]**o+(0.566155144894)*x[0]+(-0.0233988901429)*x[1]**o+(0.42480184987)*x[1]
        arg[3,2,2,0]=(0.957053098936)*x[0]**o+(0.152292007141)*x[0]+(0.332521504492)*x[1]**o+(-0.289068173241)*x[1]
        arg[3,2,2,1]=(0.923170924493)*x[0]**o+(0.191154526504)*x[0]+(0.647848811717)*x[1]**o+(0.72509033544)*x[1]
        arg[3,3,0,0]=(0.388829993841)*x[0]**o+(0.243404004988)*x[0]+(-0.0313000323506)*x[1]**o+(0.558135056894)*x[1]
        arg[3,3,0,1]=(-0.784568790167)*x[0]**o+(0.742074534256)*x[0]+(-0.624869261724)*x[1]**o+(-0.412100194843)*x[1]
        arg[3,3,1,0]=(0.924173486785)*x[0]**o+(-0.966635363007)*x[0]+(0.395550623492)*x[1]**o+(0.469204722173)*x[1]
        arg[3,3,1,1]=(-0.842963105645)*x[0]**o+(0.685039360106)*x[0]+(0.279605123662)*x[1]**o+(-0.953092961091)*x[1]
        arg[3,3,2,0]=(0.598860382403)*x[0]**o+(-0.226413712435)*x[0]+(0.722250410625)*x[1]**o+(-0.944980469407)*x[1]
        arg[3,3,2,1]=(0.787289357927)*x[0]**o+(-0.270737247579)*x[0]+(-0.134351386793)*x[1]**o+(-0.0437387143766)*x[1]
        arg[3,4,0,0]=(-0.0601939617429)*x[0]**o+(-0.212346121247)*x[0]+(0.0137510326612)*x[1]**o+(0.749376479477)*x[1]
        arg[3,4,0,1]=(-0.515391847891)*x[0]**o+(-0.191867157458)*x[0]+(0.0883438399597)*x[1]**o+(-0.155095894625)*x[1]
        arg[3,4,1,0]=(0.172678225476)*x[0]**o+(-0.917390756059)*x[0]+(-0.498696112089)*x[1]**o+(-0.267655707135)*x[1]
        arg[3,4,1,1]=(-0.498294994476)*x[0]**o+(-0.301078252047)*x[0]+(0.100215523329)*x[1]**o+(-0.424039854141)*x[1]
        arg[3,4,2,0]=(0.904127003142)*x[0]**o+(-0.70704194874)*x[0]+(0.309638449466)*x[1]**o+(-0.290894515899)*x[1]
        arg[3,4,2,1]=(-0.924994398191)*x[0]**o+(0.624996423224)*x[0]+(0.461586866277)*x[1]**o+(-0.499709859121)*x[1]
        ref[0,0,0,0]=(-0.633576516648)*x_ref[0]**o+(0.77399256211)*x_ref[0]+(-0.396696417195)*x_ref[1]**o+(-0.604572540725)*x_ref[1]
        ref[0,0,0,1]=(0.612672991978)*x_ref[0]**o+(-0.331012302611)*x_ref[0]+(0.885533578235)*x_ref[1]**o+(0.334686818058)*x_ref[1]
        ref[0,0,1,0]=(-0.982701069025)*x_ref[0]**o+(0.467849033642)*x_ref[0]+(0.525392773375)*x_ref[1]**o+(0.355622324425)*x_ref[1]
        ref[0,0,1,1]=(-0.0243671469666)*x_ref[0]**o+(0.790044842383)*x_ref[0]+(0.413384172033)*x_ref[1]**o+(0.740805135968)*x_ref[1]
        ref[0,0,2,0]=(0.0384789957691)*x_ref[0]**o+(-0.317009899417)*x_ref[0]+(0.0389710369615)*x_ref[1]**o+(0.784549852598)*x_ref[1]
        ref[0,0,2,1]=(0.361832259252)*x_ref[0]**o+(0.410815792944)*x_ref[0]+(0.39496145831)*x_ref[1]**o+(-0.820784143964)*x_ref[1]
        ref[0,1,0,0]=(0.224261088976)*x_ref[0]**o+(-0.824694527256)*x_ref[0]+(0.318674472135)*x_ref[1]**o+(0.827696354338)*x_ref[1]
        ref[0,1,0,1]=(-0.442154901359)*x_ref[0]**o+(0.141775320862)*x_ref[0]+(0.688711841268)*x_ref[1]**o+(-0.27047542755)*x_ref[1]
        ref[0,1,1,0]=(-0.570125802019)*x_ref[0]**o+(0.050303762895)*x_ref[0]+(-0.972329837417)*x_ref[1]**o+(0.143700303765)*x_ref[1]
        ref[0,1,1,1]=(0.852888550499)*x_ref[0]**o+(-0.51216090329)*x_ref[0]+(-0.895663922335)*x_ref[1]**o+(-0.183466225547)*x_ref[1]
        ref[0,1,2,0]=(0.863840920739)*x_ref[0]**o+(-0.247789353524)*x_ref[0]+(0.63056436157)*x_ref[1]**o+(0.932522589942)*x_ref[1]
        ref[0,1,2,1]=(-0.403832121794)*x_ref[0]**o+(-0.769625259789)*x_ref[0]+(0.507724303191)*x_ref[1]**o+(0.314178775526)*x_ref[1]
        ref[0,2,0,0]=(0.543867153515)*x_ref[0]**o+(-0.414107817028)*x_ref[0]+(0.423922720842)*x_ref[1]**o+(0.166632407916)*x_ref[1]
        ref[0,2,0,1]=(0.731968209929)*x_ref[0]**o+(-0.114742736683)*x_ref[0]+(0.112562415021)*x_ref[1]**o+(0.166924480325)*x_ref[1]
        ref[0,2,1,0]=(0.919472066242)*x_ref[0]**o+(0.560565315205)*x_ref[0]+(0.4967650028)*x_ref[1]**o+(0.708376671625)*x_ref[1]
        ref[0,2,1,1]=(0.147500239739)*x_ref[0]**o+(-0.632670389919)*x_ref[0]+(-0.992479063437)*x_ref[1]**o+(0.258941122962)*x_ref[1]
        ref[0,2,2,0]=(0.474481785329)*x_ref[0]**o+(0.592902604094)*x_ref[0]+(0.64058685211)*x_ref[1]**o+(-0.477843307901)*x_ref[1]
        ref[0,2,2,1]=(0.769898355146)*x_ref[0]**o+(0.563458279487)*x_ref[0]+(0.876276425449)*x_ref[1]**o+(-0.0625727749869)*x_ref[1]
        ref[0,3,0,0]=(-0.330055037059)*x_ref[0]**o+(-0.727656105163)*x_ref[0]+(-0.559901684392)*x_ref[1]**o+(-0.433646878263)*x_ref[1]
        ref[0,3,0,1]=(-0.725314328666)*x_ref[0]**o+(0.638326609105)*x_ref[0]+(-0.126619368417)*x_ref[1]**o+(-0.232194758676)*x_ref[1]
        ref[0,3,1,0]=(0.318096005545)*x_ref[0]**o+(0.325247494289)*x_ref[0]+(-0.514305253643)*x_ref[1]**o+(0.108523350086)*x_ref[1]
        ref[0,3,1,1]=(0.298034894956)*x_ref[0]**o+(0.825714393651)*x_ref[0]+(-0.67162850079)*x_ref[1]**o+(0.681809083004)*x_ref[1]
        ref[0,3,2,0]=(-0.00715890582648)*x_ref[0]**o+(0.893900731227)*x_ref[0]+(-0.773710274545)*x_ref[1]**o+(0.415913728569)*x_ref[1]
        ref[0,3,2,1]=(0.0853551940958)*x_ref[0]**o+(-0.228876011127)*x_ref[0]+(0.87699456762)*x_ref[1]**o+(-0.744605229253)*x_ref[1]
        ref[0,4,0,0]=(0.942646937302)*x_ref[0]**o+(0.92717236759)*x_ref[0]+(0.794806821569)*x_ref[1]**o+(0.848766303526)*x_ref[1]
        ref[0,4,0,1]=(0.20140672561)*x_ref[0]**o+(0.974490372414)*x_ref[0]+(-0.450332419587)*x_ref[1]**o+(-0.859102184782)*x_ref[1]
        ref[0,4,1,0]=(0.632153053193)*x_ref[0]**o+(0.702336756277)*x_ref[0]+(-0.571496076236)*x_ref[1]**o+(-0.882472852219)*x_ref[1]
        ref[0,4,1,1]=(0.366452829499)*x_ref[0]**o+(-0.514660164741)*x_ref[0]+(-0.152794832253)*x_ref[1]**o+(-0.658688117796)*x_ref[1]
        ref[0,4,2,0]=(0.406218394702)*x_ref[0]**o+(0.782359698198)*x_ref[0]+(-0.0582499587562)*x_ref[1]**o+(0.664506852595)*x_ref[1]
        ref[0,4,2,1]=(0.768240870343)*x_ref[0]**o+(0.188470592904)*x_ref[0]+(0.855933509171)*x_ref[1]**o+(0.639174087541)*x_ref[1]
        ref[1,0,0,0]=(-0.946841398046)*x_ref[0]**o+(0.892729256183)*x_ref[0]+(-0.15644094115)*x_ref[1]**o+(-0.859371368233)*x_ref[1]
        ref[1,0,0,1]=(0.238113647666)*x_ref[0]**o+(0.173695470467)*x_ref[0]+(-0.753529252313)*x_ref[1]**o+(-0.243203902463)*x_ref[1]
        ref[1,0,1,0]=(0.267937523443)*x_ref[0]**o+(0.600200492374)*x_ref[0]+(0.442653529257)*x_ref[1]**o+(-0.394984210541)*x_ref[1]
        ref[1,0,1,1]=(-0.495022907622)*x_ref[0]**o+(-0.593273808093)*x_ref[0]+(0.888764746608)*x_ref[1]**o+(-0.745127825763)*x_ref[1]
        ref[1,0,2,0]=(0.352827238496)*x_ref[0]**o+(0.293200928839)*x_ref[0]+(-0.00369812109034)*x_ref[1]**o+(0.478875774193)*x_ref[1]
        ref[1,0,2,1]=(-0.403358878981)*x_ref[0]**o+(0.756947364667)*x_ref[0]+(0.649335795039)*x_ref[1]**o+(-0.339962796859)*x_ref[1]
        ref[1,1,0,0]=(-0.607505349015)*x_ref[0]**o+(-0.392535399582)*x_ref[0]+(0.799782389634)*x_ref[1]**o+(0.186339165863)*x_ref[1]
        ref[1,1,0,1]=(0.997492304627)*x_ref[0]**o+(-0.281456622581)*x_ref[0]+(-0.959294693613)*x_ref[1]**o+(-0.454812606674)*x_ref[1]
        ref[1,1,1,0]=(0.454682037549)*x_ref[0]**o+(0.797129935647)*x_ref[0]+(-0.949539573053)*x_ref[1]**o+(-0.918499382199)*x_ref[1]
        ref[1,1,1,1]=(0.903016449758)*x_ref[0]**o+(-0.575039046836)*x_ref[0]+(0.957158277074)*x_ref[1]**o+(-0.715797460906)*x_ref[1]
        ref[1,1,2,0]=(0.92005547899)*x_ref[0]**o+(-0.716975919302)*x_ref[0]+(-0.900904311708)*x_ref[1]**o+(0.989458809643)*x_ref[1]
        ref[1,1,2,1]=(-0.633214991087)*x_ref[0]**o+(0.680658578627)*x_ref[0]+(0.552194054556)*x_ref[1]**o+(0.343557072769)*x_ref[1]
        ref[1,2,0,0]=(-0.0182903530463)*x_ref[0]**o+(-0.0445569497267)*x_ref[0]+(-0.259920069365)*x_ref[1]**o+(0.118765787279)*x_ref[1]
        ref[1,2,0,1]=(-0.667382174733)*x_ref[0]**o+(-0.79898800024)*x_ref[0]+(-0.217638556567)*x_ref[1]**o+(-0.0288504127936)*x_ref[1]
        ref[1,2,1,0]=(-0.437982445862)*x_ref[0]**o+(-0.104054454808)*x_ref[0]+(0.711429718088)*x_ref[1]**o+(-0.607226296329)*x_ref[1]
        ref[1,2,1,1]=(0.252861218037)*x_ref[0]**o+(-0.635701456504)*x_ref[0]+(-0.555222304592)*x_ref[1]**o+(0.723744010092)*x_ref[1]
        ref[1,2,2,0]=(-0.73532710885)*x_ref[0]**o+(0.600599078741)*x_ref[0]+(0.0465795173194)*x_ref[1]**o+(0.524165544645)*x_ref[1]
        ref[1,2,2,1]=(-0.025246019744)*x_ref[0]**o+(-0.376526675619)*x_ref[0]+(0.213011788698)*x_ref[1]**o+(-0.477018225777)*x_ref[1]
        ref[1,3,0,0]=(-0.52682690957)*x_ref[0]**o+(0.0567472858499)*x_ref[0]+(0.574116256102)*x_ref[1]**o+(-0.130012791898)*x_ref[1]
        ref[1,3,0,1]=(0.113502182936)*x_ref[0]**o+(0.7643861313)*x_ref[0]+(0.48414948847)*x_ref[1]**o+(0.863168250318)*x_ref[1]
        ref[1,3,1,0]=(-0.208102162378)*x_ref[0]**o+(0.327770728981)*x_ref[0]+(0.843778895278)*x_ref[1]**o+(0.474843504177)*x_ref[1]
        ref[1,3,1,1]=(-0.549818192184)*x_ref[0]**o+(-0.863333816685)*x_ref[0]+(-0.365631583957)*x_ref[1]**o+(-0.491320441182)*x_ref[1]
        ref[1,3,2,0]=(-0.265833845759)*x_ref[0]**o+(-0.94416746081)*x_ref[0]+(0.48465442593)*x_ref[1]**o+(0.258284621408)*x_ref[1]
        ref[1,3,2,1]=(-0.625716082268)*x_ref[0]**o+(-0.234697130359)*x_ref[0]+(-0.973704406131)*x_ref[1]**o+(0.00141746117847)*x_ref[1]
        ref[1,4,0,0]=(0.565044361964)*x_ref[0]**o+(-0.0672474991748)*x_ref[0]+(-0.125901930524)*x_ref[1]**o+(-0.753103899732)*x_ref[1]
        ref[1,4,0,1]=(-0.979743467289)*x_ref[0]**o+(-0.25225085214)*x_ref[0]+(-0.141635164635)*x_ref[1]**o+(0.378153168667)*x_ref[1]
        ref[1,4,1,0]=(0.97728399206)*x_ref[0]**o+(-0.202043160532)*x_ref[0]+(0.997445424954)*x_ref[1]**o+(-0.100049830154)*x_ref[1]
        ref[1,4,1,1]=(0.364840105486)*x_ref[0]**o+(-0.682234327437)*x_ref[0]+(0.252584592259)*x_ref[1]**o+(-0.0785430242425)*x_ref[1]
        ref[1,4,2,0]=(-0.0168764722616)*x_ref[0]**o+(-0.997460139026)*x_ref[0]+(0.560548548912)*x_ref[1]**o+(0.220532224022)*x_ref[1]
        ref[1,4,2,1]=(0.307927943715)*x_ref[0]**o+(0.496390872708)*x_ref[0]+(0.982750813274)*x_ref[1]**o+(0.392968327778)*x_ref[1]
        ref[2,0,0,0]=(0.645991188948)*x_ref[0]**o+(0.104035822886)*x_ref[0]+(0.18285829097)*x_ref[1]**o+(0.51186286294)*x_ref[1]
        ref[2,0,0,1]=(-0.996971717318)*x_ref[0]**o+(0.126488047141)*x_ref[0]+(0.123585312562)*x_ref[1]**o+(-0.775854123529)*x_ref[1]
        ref[2,0,1,0]=(-0.0258590124655)*x_ref[0]**o+(0.0079454735671)*x_ref[0]+(-0.42819024422)*x_ref[1]**o+(0.12377360967)*x_ref[1]
        ref[2,0,1,1]=(0.990441351225)*x_ref[0]**o+(-0.453806376472)*x_ref[0]+(0.134316618645)*x_ref[1]**o+(-0.902666953282)*x_ref[1]
        ref[2,0,2,0]=(0.149304385108)*x_ref[0]**o+(-0.226132046999)*x_ref[0]+(0.647186106203)*x_ref[1]**o+(-0.269247029225)*x_ref[1]
        ref[2,0,2,1]=(-0.159068164422)*x_ref[0]**o+(-0.392275577366)*x_ref[0]+(0.131423093404)*x_ref[1]**o+(-0.353311282658)*x_ref[1]
        ref[2,1,0,0]=(-0.2862918681)*x_ref[0]**o+(-0.392264098939)*x_ref[0]+(-0.344033237285)*x_ref[1]**o+(0.644261713963)*x_ref[1]
        ref[2,1,0,1]=(0.398022911658)*x_ref[0]**o+(-0.84552546032)*x_ref[0]+(-0.823861260943)*x_ref[1]**o+(0.739994988755)*x_ref[1]
        ref[2,1,1,0]=(-0.696854341234)*x_ref[0]**o+(-0.810851183298)*x_ref[0]+(-0.712182042896)*x_ref[1]**o+(0.594887691783)*x_ref[1]
        ref[2,1,1,1]=(-0.182583840117)*x_ref[0]**o+(0.962594377088)*x_ref[0]+(-0.0157575661453)*x_ref[1]**o+(-0.432167995007)*x_ref[1]
        ref[2,1,2,0]=(0.293029196216)*x_ref[0]**o+(0.46839918212)*x_ref[0]+(0.262375931577)*x_ref[1]**o+(0.0348536956306)*x_ref[1]
        ref[2,1,2,1]=(0.775632135095)*x_ref[0]**o+(-0.874598534107)*x_ref[0]+(0.711602976736)*x_ref[1]**o+(0.37065543293)*x_ref[1]
        ref[2,2,0,0]=(-0.103625214158)*x_ref[0]**o+(0.553558117877)*x_ref[0]+(-0.0522465477589)*x_ref[1]**o+(-0.747139195529)*x_ref[1]
        ref[2,2,0,1]=(0.507427663615)*x_ref[0]**o+(-0.536361186134)*x_ref[0]+(0.809054726193)*x_ref[1]**o+(-0.979546347794)*x_ref[1]
        ref[2,2,1,0]=(0.786210623833)*x_ref[0]**o+(0.244510127091)*x_ref[0]+(0.159799334111)*x_ref[1]**o+(0.0499263839273)*x_ref[1]
        ref[2,2,1,1]=(0.454205244371)*x_ref[0]**o+(-0.650927822092)*x_ref[0]+(-0.939252656087)*x_ref[1]**o+(-0.894014755314)*x_ref[1]
        ref[2,2,2,0]=(-0.829617920904)*x_ref[0]**o+(0.764070900994)*x_ref[0]+(0.972419443556)*x_ref[1]**o+(-0.824913543157)*x_ref[1]
        ref[2,2,2,1]=(0.694310748478)*x_ref[0]**o+(-0.191413364091)*x_ref[0]+(0.588258464472)*x_ref[1]**o+(0.743973223755)*x_ref[1]
        ref[2,3,0,0]=(-0.588397365184)*x_ref[0]**o+(-0.634207034861)*x_ref[0]+(0.0853498555546)*x_ref[1]**o+(0.559933556865)*x_ref[1]
        ref[2,3,0,1]=(0.626829165689)*x_ref[0]**o+(0.942185296616)*x_ref[0]+(0.0360770405565)*x_ref[1]**o+(0.635966218547)*x_ref[1]
        ref[2,3,1,0]=(-0.0103526873852)*x_ref[0]**o+(0.546330663496)*x_ref[0]+(0.799224103454)*x_ref[1]**o+(0.791922324897)*x_ref[1]
        ref[2,3,1,1]=(0.921564474495)*x_ref[0]**o+(-0.420410835144)*x_ref[0]+(-0.20482503843)*x_ref[1]**o+(-0.483331643037)*x_ref[1]
        ref[2,3,2,0]=(-0.903600870314)*x_ref[0]**o+(0.901956716274)*x_ref[0]+(-0.113926342023)*x_ref[1]**o+(-0.061596184768)*x_ref[1]
        ref[2,3,2,1]=(0.653250267262)*x_ref[0]**o+(-0.13770933726)*x_ref[0]+(-0.0791563636079)*x_ref[1]**o+(-0.578313150893)*x_ref[1]
        ref[2,4,0,0]=(0.993903405725)*x_ref[0]**o+(0.694324226249)*x_ref[0]+(0.203872883541)*x_ref[1]**o+(-0.526880767264)*x_ref[1]
        ref[2,4,0,1]=(0.0120058513647)*x_ref[0]**o+(0.00907603337269)*x_ref[0]+(0.0247958601677)*x_ref[1]**o+(-0.996916120875)*x_ref[1]
        ref[2,4,1,0]=(0.281518049478)*x_ref[0]**o+(0.0773799323777)*x_ref[0]+(-0.968565098773)*x_ref[1]**o+(-0.275558076229)*x_ref[1]
        ref[2,4,1,1]=(0.565275744757)*x_ref[0]**o+(0.978636675869)*x_ref[0]+(-0.352951657751)*x_ref[1]**o+(-0.54170487914)*x_ref[1]
        ref[2,4,2,0]=(-0.571207550678)*x_ref[0]**o+(0.196754283318)*x_ref[0]+(0.665558917127)*x_ref[1]**o+(0.361915269355)*x_ref[1]
        ref[2,4,2,1]=(0.54579191885)*x_ref[0]**o+(0.798319371798)*x_ref[0]+(0.961669996133)*x_ref[1]**o+(-0.549305529342)*x_ref[1]
        ref[3,0,0,0]=(0.264064946121)*x_ref[0]**o+(-0.314396185593)*x_ref[0]+(-0.017492206135)*x_ref[1]**o+(0.190513393091)*x_ref[1]
        ref[3,0,0,1]=(0.812213666919)*x_ref[0]**o+(-0.757919357489)*x_ref[0]+(0.294829087285)*x_ref[1]**o+(0.916181101963)*x_ref[1]
        ref[3,0,1,0]=(-0.529906091574)*x_ref[0]**o+(0.553682716765)*x_ref[0]+(-0.693177625995)*x_ref[1]**o+(0.377145561683)*x_ref[1]
        ref[3,0,1,1]=(-0.298942248169)*x_ref[0]**o+(0.876878593705)*x_ref[0]+(-0.609807982832)*x_ref[1]**o+(-0.574954919539)*x_ref[1]
        ref[3,0,2,0]=(-0.729324018245)*x_ref[0]**o+(-0.495194565254)*x_ref[0]+(-0.616162751356)*x_ref[1]**o+(0.878549207376)*x_ref[1]
        ref[3,0,2,1]=(0.625373988248)*x_ref[0]**o+(0.106736108396)*x_ref[0]+(-0.871225833944)*x_ref[1]**o+(0.840614493429)*x_ref[1]
        ref[3,1,0,0]=(-0.90880839589)*x_ref[0]**o+(0.390298288126)*x_ref[0]+(0.446872042362)*x_ref[1]**o+(-0.0703294219415)*x_ref[1]
        ref[3,1,0,1]=(-0.0706115592617)*x_ref[0]**o+(-0.158386152093)*x_ref[0]+(-0.687459181052)*x_ref[1]**o+(-0.251170026598)*x_ref[1]
        ref[3,1,1,0]=(-0.286907297081)*x_ref[0]**o+(0.185480985579)*x_ref[0]+(-0.49763263834)*x_ref[1]**o+(-0.773056048642)*x_ref[1]
        ref[3,1,1,1]=(-0.849572703562)*x_ref[0]**o+(0.481739532516)*x_ref[0]+(-0.227573063955)*x_ref[1]**o+(-0.282367611381)*x_ref[1]
        ref[3,1,2,0]=(-0.524975224603)*x_ref[0]**o+(-0.795450756917)*x_ref[0]+(-0.656865716441)*x_ref[1]**o+(-0.0166902340102)*x_ref[1]
        ref[3,1,2,1]=(-0.297441818145)*x_ref[0]**o+(-0.726759852457)*x_ref[0]+(0.146859888138)*x_ref[1]**o+(-0.73690394554)*x_ref[1]
        ref[3,2,0,0]=(-0.278794538596)*x_ref[0]**o+(-0.463428732496)*x_ref[0]+(-0.478500378607)*x_ref[1]**o+(0.806643392698)*x_ref[1]
        ref[3,2,0,1]=(0.573853299578)*x_ref[0]**o+(-0.324830388013)*x_ref[0]+(-0.185528736416)*x_ref[1]**o+(0.444228024864)*x_ref[1]
        ref[3,2,1,0]=(0.686054084195)*x_ref[0]**o+(-0.914380567743)*x_ref[0]+(0.751959744445)*x_ref[1]**o+(0.531759675659)*x_ref[1]
        ref[3,2,1,1]=(-0.876330263589)*x_ref[0]**o+(0.566155144894)*x_ref[0]+(-0.0233988901429)*x_ref[1]**o+(0.42480184987)*x_ref[1]
        ref[3,2,2,0]=(0.957053098936)*x_ref[0]**o+(0.152292007141)*x_ref[0]+(0.332521504492)*x_ref[1]**o+(-0.289068173241)*x_ref[1]
        ref[3,2,2,1]=(0.923170924493)*x_ref[0]**o+(0.191154526504)*x_ref[0]+(0.647848811717)*x_ref[1]**o+(0.72509033544)*x_ref[1]
        ref[3,3,0,0]=(0.388829993841)*x_ref[0]**o+(0.243404004988)*x_ref[0]+(-0.0313000323506)*x_ref[1]**o+(0.558135056894)*x_ref[1]
        ref[3,3,0,1]=(-0.784568790167)*x_ref[0]**o+(0.742074534256)*x_ref[0]+(-0.624869261724)*x_ref[1]**o+(-0.412100194843)*x_ref[1]
        ref[3,3,1,0]=(0.924173486785)*x_ref[0]**o+(-0.966635363007)*x_ref[0]+(0.395550623492)*x_ref[1]**o+(0.469204722173)*x_ref[1]
        ref[3,3,1,1]=(-0.842963105645)*x_ref[0]**o+(0.685039360106)*x_ref[0]+(0.279605123662)*x_ref[1]**o+(-0.953092961091)*x_ref[1]
        ref[3,3,2,0]=(0.598860382403)*x_ref[0]**o+(-0.226413712435)*x_ref[0]+(0.722250410625)*x_ref[1]**o+(-0.944980469407)*x_ref[1]
        ref[3,3,2,1]=(0.787289357927)*x_ref[0]**o+(-0.270737247579)*x_ref[0]+(-0.134351386793)*x_ref[1]**o+(-0.0437387143766)*x_ref[1]
        ref[3,4,0,0]=(-0.0601939617429)*x_ref[0]**o+(-0.212346121247)*x_ref[0]+(0.0137510326612)*x_ref[1]**o+(0.749376479477)*x_ref[1]
        ref[3,4,0,1]=(-0.515391847891)*x_ref[0]**o+(-0.191867157458)*x_ref[0]+(0.0883438399597)*x_ref[1]**o+(-0.155095894625)*x_ref[1]
        ref[3,4,1,0]=(0.172678225476)*x_ref[0]**o+(-0.917390756059)*x_ref[0]+(-0.498696112089)*x_ref[1]**o+(-0.267655707135)*x_ref[1]
        ref[3,4,1,1]=(-0.498294994476)*x_ref[0]**o+(-0.301078252047)*x_ref[0]+(0.100215523329)*x_ref[1]**o+(-0.424039854141)*x_ref[1]
        ref[3,4,2,0]=(0.904127003142)*x_ref[0]**o+(-0.70704194874)*x_ref[0]+(0.309638449466)*x_ref[1]**o+(-0.290894515899)*x_ref[1]
        ref[3,4,2,1]=(-0.924994398191)*x_ref[0]**o+(0.624996423224)*x_ref[0]+(0.461586866277)*x_ref[1]**o+(-0.499709859121)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.0998996137438)*x[0]**o+(-0.648495171588)*x[0]+(-0.836177837758)*x[1]**o+(-0.9597381047)*x[1]+(-0.958069967907)*x[2]**o+(0.399097251753)*x[2]
        arg[0,0,0,1]=(0.625409406207)*x[0]**o+(-0.522476054218)*x[0]+(0.794724584797)*x[1]**o+(-0.450875661781)*x[1]+(-0.189505729613)*x[2]**o+(0.0910808218379)*x[2]
        arg[0,0,1,0]=(-0.36073198549)*x[0]**o+(-0.0286144078107)*x[0]+(-0.53827851785)*x[1]**o+(-0.563600303848)*x[1]+(0.970211345168)*x[2]**o+(-0.405253675856)*x[2]
        arg[0,0,1,1]=(-0.0204302558323)*x[0]**o+(0.927570053009)*x[0]+(-0.358571532503)*x[1]**o+(0.282513525404)*x[1]+(0.105530612734)*x[2]**o+(-0.156753729378)*x[2]
        arg[0,0,2,0]=(0.25049329858)*x[0]**o+(0.412299654624)*x[0]+(0.827683104949)*x[1]**o+(0.0195885473957)*x[1]+(0.849521623993)*x[2]**o+(-0.107922097347)*x[2]
        arg[0,0,2,1]=(0.824054630326)*x[0]**o+(0.315612619612)*x[0]+(-0.33528975625)*x[1]**o+(-0.848543364468)*x[1]+(0.67462279419)*x[2]**o+(-0.0387606720463)*x[2]
        arg[0,1,0,0]=(0.251613993994)*x[0]**o+(-0.871217704935)*x[0]+(-0.648182240752)*x[1]**o+(0.324536900202)*x[1]+(-0.884871901147)*x[2]**o+(0.353160185772)*x[2]
        arg[0,1,0,1]=(-0.629894976165)*x[0]**o+(0.99092287348)*x[0]+(0.87471936536)*x[1]**o+(0.732884844801)*x[1]+(0.423531105096)*x[2]**o+(0.648959408663)*x[2]
        arg[0,1,1,0]=(0.123366820062)*x[0]**o+(-0.856900246151)*x[0]+(-0.887819649211)*x[1]**o+(-0.901367472665)*x[1]+(0.222535500196)*x[2]**o+(0.42830671462)*x[2]
        arg[0,1,1,1]=(0.343789367141)*x[0]**o+(0.614284333526)*x[0]+(0.976225499577)*x[1]**o+(0.955115562801)*x[1]+(0.0700409126681)*x[2]**o+(-0.709602275548)*x[2]
        arg[0,1,2,0]=(-0.863502900367)*x[0]**o+(-0.452731447667)*x[0]+(0.495519161334)*x[1]**o+(0.0254261045314)*x[1]+(-0.484479110985)*x[2]**o+(0.615421138564)*x[2]
        arg[0,1,2,1]=(-0.861966549825)*x[0]**o+(0.957385177208)*x[0]+(0.710449966969)*x[1]**o+(-0.367149885665)*x[1]+(-0.100800220614)*x[2]**o+(-0.266781849874)*x[2]
        arg[0,2,0,0]=(-0.781309947495)*x[0]**o+(-0.709036446117)*x[0]+(0.86136976691)*x[1]**o+(-0.745743057084)*x[1]+(0.519795761984)*x[2]**o+(0.298838092751)*x[2]
        arg[0,2,0,1]=(-0.514037154361)*x[0]**o+(-0.155867632386)*x[0]+(-0.125008826668)*x[1]**o+(-0.138197902584)*x[1]+(0.360698647284)*x[2]**o+(0.432385729464)*x[2]
        arg[0,2,1,0]=(-0.19580514344)*x[0]**o+(-0.430510395896)*x[0]+(0.676498502854)*x[1]**o+(-0.472411864377)*x[1]+(-0.323468590594)*x[2]**o+(0.837971311859)*x[2]
        arg[0,2,1,1]=(-0.358714577548)*x[0]**o+(-0.583080123482)*x[0]+(-0.225887451598)*x[1]**o+(0.744101609163)*x[1]+(-0.0300539574094)*x[2]**o+(-0.985181847505)*x[2]
        arg[0,2,2,0]=(0.820610118447)*x[0]**o+(-0.560334915671)*x[0]+(0.478672329569)*x[1]**o+(-0.956892480643)*x[1]+(0.302100338619)*x[2]**o+(0.597584594886)*x[2]
        arg[0,2,2,1]=(-0.0658925658941)*x[0]**o+(-0.136738027019)*x[0]+(0.617259524472)*x[1]**o+(-0.486181324584)*x[1]+(0.225924222893)*x[2]**o+(-0.304893108343)*x[2]
        arg[0,3,0,0]=(0.0824288958004)*x[0]**o+(-0.67546530428)*x[0]+(-0.726843419073)*x[1]**o+(0.150753236521)*x[1]+(0.200621841908)*x[2]**o+(0.869623957921)*x[2]
        arg[0,3,0,1]=(0.66227806631)*x[0]**o+(-0.389229756136)*x[0]+(0.739548522805)*x[1]**o+(0.131026218939)*x[1]+(0.743271925877)*x[2]**o+(-0.910220233981)*x[2]
        arg[0,3,1,0]=(-0.259647259175)*x[0]**o+(-0.208585364012)*x[0]+(-0.323867955964)*x[1]**o+(0.778380345151)*x[1]+(-0.555780852614)*x[2]**o+(-0.16366483426)*x[2]
        arg[0,3,1,1]=(0.248968056109)*x[0]**o+(0.84999990087)*x[0]+(0.180937025847)*x[1]**o+(0.332075828156)*x[1]+(-0.00577042146131)*x[2]**o+(0.708188905312)*x[2]
        arg[0,3,2,0]=(-0.80327596428)*x[0]**o+(-0.131004563471)*x[0]+(0.443094741792)*x[1]**o+(0.515867966266)*x[1]+(-0.964003373109)*x[2]**o+(-0.580721642935)*x[2]
        arg[0,3,2,1]=(-0.107361389717)*x[0]**o+(-0.543617395264)*x[0]+(0.840867379108)*x[1]**o+(-0.459941493027)*x[1]+(-0.300033656162)*x[2]**o+(0.440763554498)*x[2]
        arg[0,4,0,0]=(0.272916362055)*x[0]**o+(-0.554210888243)*x[0]+(-0.458102512273)*x[1]**o+(-0.54594406835)*x[1]+(0.570670715852)*x[2]**o+(0.792731049718)*x[2]
        arg[0,4,0,1]=(0.896330380602)*x[0]**o+(-0.929757950797)*x[0]+(0.229049814758)*x[1]**o+(0.1245974562)*x[1]+(0.853250033203)*x[2]**o+(-0.877052530974)*x[2]
        arg[0,4,1,0]=(0.76027467594)*x[0]**o+(-0.510095378376)*x[0]+(0.697258966361)*x[1]**o+(-0.44013714543)*x[1]+(0.617675705023)*x[2]**o+(-0.00118327085347)*x[2]
        arg[0,4,1,1]=(0.92472377319)*x[0]**o+(0.27783405618)*x[0]+(-0.24317632451)*x[1]**o+(-0.0264727232824)*x[1]+(0.428751701408)*x[2]**o+(0.184142273303)*x[2]
        arg[0,4,2,0]=(0.619110983031)*x[0]**o+(-0.375037512361)*x[0]+(0.298576843317)*x[1]**o+(-0.218802335776)*x[1]+(-0.394889850336)*x[2]**o+(-0.842359750924)*x[2]
        arg[0,4,2,1]=(0.899714346356)*x[0]**o+(0.0630133184109)*x[0]+(0.0796796488932)*x[1]**o+(-0.693371400327)*x[1]+(-0.412090402419)*x[2]**o+(0.681871832039)*x[2]
        arg[1,0,0,0]=(-0.0367677770358)*x[0]**o+(-0.75352301678)*x[0]+(0.0705016755404)*x[1]**o+(0.870115630516)*x[1]+(0.136757694474)*x[2]**o+(-0.557722157484)*x[2]
        arg[1,0,0,1]=(0.550371937413)*x[0]**o+(-0.590052751632)*x[0]+(0.38665744963)*x[1]**o+(-0.896537565188)*x[1]+(0.550869514185)*x[2]**o+(0.693828237691)*x[2]
        arg[1,0,1,0]=(-0.150321579562)*x[0]**o+(0.111135028893)*x[0]+(0.582595079816)*x[1]**o+(-0.250920804447)*x[1]+(0.392998664266)*x[2]**o+(-0.226735735878)*x[2]
        arg[1,0,1,1]=(0.435196573865)*x[0]**o+(-0.31161203165)*x[0]+(-0.117580582163)*x[1]**o+(0.99583282896)*x[1]+(-0.164872214322)*x[2]**o+(0.243837126961)*x[2]
        arg[1,0,2,0]=(-0.320135099791)*x[0]**o+(0.339871290643)*x[0]+(0.593784119675)*x[1]**o+(0.695166367768)*x[1]+(0.407551427099)*x[2]**o+(-0.768445418933)*x[2]
        arg[1,0,2,1]=(-0.390718323619)*x[0]**o+(-0.0515766643887)*x[0]+(0.394112545837)*x[1]**o+(-0.379592335881)*x[1]+(-0.29680237927)*x[2]**o+(0.500583589142)*x[2]
        arg[1,1,0,0]=(-0.958930666062)*x[0]**o+(0.516874100468)*x[0]+(0.0859034666349)*x[1]**o+(-0.370561856015)*x[1]+(-0.175293417132)*x[2]**o+(0.672554615126)*x[2]
        arg[1,1,0,1]=(-0.38129381629)*x[0]**o+(-0.615636530444)*x[0]+(0.447284140424)*x[1]**o+(0.507583882952)*x[1]+(-0.942413761354)*x[2]**o+(-0.607928807162)*x[2]
        arg[1,1,1,0]=(-0.456462200294)*x[0]**o+(0.778557108965)*x[0]+(-0.939624730616)*x[1]**o+(0.965857713341)*x[1]+(-0.913772635427)*x[2]**o+(-0.459963514384)*x[2]
        arg[1,1,1,1]=(0.138896120228)*x[0]**o+(0.941491395046)*x[0]+(0.546538931051)*x[1]**o+(-0.526098363992)*x[1]+(0.32139556996)*x[2]**o+(0.899346403792)*x[2]
        arg[1,1,2,0]=(0.591597103411)*x[0]**o+(0.926278363416)*x[0]+(0.565926211046)*x[1]**o+(0.930516925784)*x[1]+(-0.498131779353)*x[2]**o+(0.358157310681)*x[2]
        arg[1,1,2,1]=(0.758395199797)*x[0]**o+(0.769578766255)*x[0]+(0.86819976051)*x[1]**o+(-0.380436310842)*x[1]+(-0.513012037048)*x[2]**o+(-0.984361866072)*x[2]
        arg[1,2,0,0]=(0.0845846076213)*x[0]**o+(0.987145343601)*x[0]+(-0.447680898266)*x[1]**o+(-0.30278197108)*x[1]+(0.271667503636)*x[2]**o+(0.698967822656)*x[2]
        arg[1,2,0,1]=(-0.314076111079)*x[0]**o+(0.507407943673)*x[0]+(-0.38467148586)*x[1]**o+(0.802764957621)*x[1]+(0.692949890402)*x[2]**o+(-0.965766183849)*x[2]
        arg[1,2,1,0]=(-0.855806343042)*x[0]**o+(0.753362713203)*x[0]+(-0.9877924012)*x[1]**o+(-0.54047626663)*x[1]+(-0.488859780266)*x[2]**o+(0.444279382381)*x[2]
        arg[1,2,1,1]=(-0.313257382113)*x[0]**o+(-0.926058887186)*x[0]+(0.509510780366)*x[1]**o+(0.738049597453)*x[1]+(0.191676625722)*x[2]**o+(-0.354148482258)*x[2]
        arg[1,2,2,0]=(-0.955204517346)*x[0]**o+(0.234220317948)*x[0]+(0.44435792198)*x[1]**o+(0.96099571541)*x[1]+(-0.422462745872)*x[2]**o+(0.808817733192)*x[2]
        arg[1,2,2,1]=(0.932454071169)*x[0]**o+(-0.449841123911)*x[0]+(-0.371587228288)*x[1]**o+(-0.552185751938)*x[1]+(-0.222070148528)*x[2]**o+(-0.911768306425)*x[2]
        arg[1,3,0,0]=(-0.355162862133)*x[0]**o+(-0.00562439573347)*x[0]+(0.982216168139)*x[1]**o+(0.0493986220989)*x[1]+(0.0242698649494)*x[2]**o+(0.593411693569)*x[2]
        arg[1,3,0,1]=(-0.982164819772)*x[0]**o+(-0.417475530445)*x[0]+(-0.421752719264)*x[1]**o+(-0.102522772133)*x[1]+(-0.384947583437)*x[2]**o+(-0.675090951882)*x[2]
        arg[1,3,1,0]=(0.862354670281)*x[0]**o+(0.0332806244503)*x[0]+(0.248152892705)*x[1]**o+(-0.380941068103)*x[1]+(0.499668747376)*x[2]**o+(-0.732344330916)*x[2]
        arg[1,3,1,1]=(-0.367669055003)*x[0]**o+(-0.28761578782)*x[0]+(0.10838982068)*x[1]**o+(-0.495999837684)*x[1]+(0.599492939122)*x[2]**o+(0.886609670648)*x[2]
        arg[1,3,2,0]=(0.707026122148)*x[0]**o+(-0.625314188271)*x[0]+(-0.734283031657)*x[1]**o+(0.779848908302)*x[1]+(-0.965798269237)*x[2]**o+(0.35160103487)*x[2]
        arg[1,3,2,1]=(0.198008431503)*x[0]**o+(0.145470568903)*x[0]+(-0.342584113223)*x[1]**o+(-0.976390968602)*x[1]+(-0.371678081685)*x[2]**o+(0.569699640624)*x[2]
        arg[1,4,0,0]=(-0.816642540092)*x[0]**o+(0.466127552066)*x[0]+(-0.491109738021)*x[1]**o+(-0.177222784312)*x[1]+(0.723624314914)*x[2]**o+(-0.103929106752)*x[2]
        arg[1,4,0,1]=(-0.461770336596)*x[0]**o+(-0.416473447722)*x[0]+(-0.844639456093)*x[1]**o+(-0.021509624606)*x[1]+(-0.507857805076)*x[2]**o+(-0.0132005113079)*x[2]
        arg[1,4,1,0]=(0.279614653607)*x[0]**o+(-0.547332204747)*x[0]+(0.329751635637)*x[1]**o+(0.813149533168)*x[1]+(0.0361758823053)*x[2]**o+(0.0635753923527)*x[2]
        arg[1,4,1,1]=(0.124940246141)*x[0]**o+(-0.470715329674)*x[0]+(0.401922174627)*x[1]**o+(-0.292371975658)*x[1]+(0.302438467505)*x[2]**o+(0.131045792461)*x[2]
        arg[1,4,2,0]=(0.817170444398)*x[0]**o+(0.153182033447)*x[0]+(-0.98908973731)*x[1]**o+(0.359609127275)*x[1]+(-0.891501663456)*x[2]**o+(0.34700998982)*x[2]
        arg[1,4,2,1]=(0.619050545861)*x[0]**o+(-0.784466574079)*x[0]+(-0.15316511355)*x[1]**o+(0.587255937366)*x[1]+(0.740644692907)*x[2]**o+(0.0651236072508)*x[2]
        arg[2,0,0,0]=(0.67904774175)*x[0]**o+(-0.24998331377)*x[0]+(-0.263635906274)*x[1]**o+(0.157008949549)*x[1]+(-0.905158896842)*x[2]**o+(-0.975062080366)*x[2]
        arg[2,0,0,1]=(0.724305963876)*x[0]**o+(0.666891161317)*x[0]+(0.769784501206)*x[1]**o+(-0.84353242369)*x[1]+(0.631055556571)*x[2]**o+(-0.433528873211)*x[2]
        arg[2,0,1,0]=(-0.545702871321)*x[0]**o+(-0.153233364062)*x[0]+(-0.892642075879)*x[1]**o+(-0.303965445232)*x[1]+(0.675938211305)*x[2]**o+(0.134016007164)*x[2]
        arg[2,0,1,1]=(0.612528479764)*x[0]**o+(0.444039154606)*x[0]+(-0.185281638413)*x[1]**o+(0.643315773689)*x[1]+(-0.5562387833)*x[2]**o+(-0.960249468718)*x[2]
        arg[2,0,2,0]=(-0.330135152971)*x[0]**o+(0.359951176686)*x[0]+(-0.797088515842)*x[1]**o+(-0.257909258598)*x[1]+(-0.927156080256)*x[2]**o+(-0.843405252125)*x[2]
        arg[2,0,2,1]=(0.447041226497)*x[0]**o+(0.691225329527)*x[0]+(0.462798475523)*x[1]**o+(0.157497098383)*x[1]+(-0.77785224695)*x[2]**o+(0.886430058902)*x[2]
        arg[2,1,0,0]=(0.397605326971)*x[0]**o+(0.66227843262)*x[0]+(-0.634922556623)*x[1]**o+(-0.263342020187)*x[1]+(-0.81486179124)*x[2]**o+(-0.216716259698)*x[2]
        arg[2,1,0,1]=(-0.868246547394)*x[0]**o+(-0.194983059368)*x[0]+(-0.667277750808)*x[1]**o+(0.284011270958)*x[1]+(-0.276295620425)*x[2]**o+(0.215159516398)*x[2]
        arg[2,1,1,0]=(0.990906320323)*x[0]**o+(0.526754062323)*x[0]+(-0.963398315893)*x[1]**o+(-0.72561275544)*x[1]+(0.539779065056)*x[2]**o+(0.37738462043)*x[2]
        arg[2,1,1,1]=(-0.668520028258)*x[0]**o+(-0.292883094355)*x[0]+(0.260949538172)*x[1]**o+(-0.952706524522)*x[1]+(-0.715386804875)*x[2]**o+(-0.683320136224)*x[2]
        arg[2,1,2,0]=(-0.646186301281)*x[0]**o+(-0.62747987465)*x[0]+(-0.858714171368)*x[1]**o+(0.46910094021)*x[1]+(0.468975652807)*x[2]**o+(0.242115907129)*x[2]
        arg[2,1,2,1]=(0.262205767991)*x[0]**o+(-0.460376423099)*x[0]+(0.88572453638)*x[1]**o+(0.611761462581)*x[1]+(0.763860971141)*x[2]**o+(0.108470198525)*x[2]
        arg[2,2,0,0]=(-0.0218691308052)*x[0]**o+(0.942271392012)*x[0]+(-0.778469966054)*x[1]**o+(0.0873305333765)*x[1]+(-0.300479183111)*x[2]**o+(-0.862547713455)*x[2]
        arg[2,2,0,1]=(-0.736808990311)*x[0]**o+(-0.200478457878)*x[0]+(0.272608362607)*x[1]**o+(-0.0651056524306)*x[1]+(0.616628495568)*x[2]**o+(-0.8623490351)*x[2]
        arg[2,2,1,0]=(0.931530668297)*x[0]**o+(0.39701649536)*x[0]+(0.552135074711)*x[1]**o+(-0.489731326344)*x[1]+(-0.864724191354)*x[2]**o+(0.369557062583)*x[2]
        arg[2,2,1,1]=(0.418780453395)*x[0]**o+(-0.4935990418)*x[0]+(0.34625989303)*x[1]**o+(0.979829617988)*x[1]+(0.853169987441)*x[2]**o+(-0.945958722805)*x[2]
        arg[2,2,2,0]=(0.973689940887)*x[0]**o+(-0.0585784323994)*x[0]+(-0.472419744635)*x[1]**o+(0.851814193674)*x[1]+(-0.393393340405)*x[2]**o+(-0.755376551633)*x[2]
        arg[2,2,2,1]=(-0.217049913648)*x[0]**o+(-0.262150796255)*x[0]+(-0.407805273745)*x[1]**o+(0.819646964669)*x[1]+(-0.0130316627236)*x[2]**o+(0.179034130621)*x[2]
        arg[2,3,0,0]=(-0.377708814803)*x[0]**o+(-0.943893058411)*x[0]+(-0.354341838264)*x[1]**o+(0.0696491295689)*x[1]+(-0.192838410156)*x[2]**o+(-0.22208932914)*x[2]
        arg[2,3,0,1]=(0.252429489766)*x[0]**o+(0.808052283089)*x[0]+(0.282941493712)*x[1]**o+(0.89816007357)*x[1]+(0.000979423783521)*x[2]**o+(0.645599553602)*x[2]
        arg[2,3,1,0]=(0.788089987784)*x[0]**o+(0.414254044423)*x[0]+(0.105204158639)*x[1]**o+(0.97527041289)*x[1]+(0.447463569594)*x[2]**o+(0.91465488394)*x[2]
        arg[2,3,1,1]=(0.882323493958)*x[0]**o+(-0.98135461385)*x[0]+(0.641881834427)*x[1]**o+(-0.888068999894)*x[1]+(0.705728558352)*x[2]**o+(0.295815435834)*x[2]
        arg[2,3,2,0]=(-0.570666876652)*x[0]**o+(0.52308582534)*x[0]+(-0.604213776925)*x[1]**o+(-0.568684610916)*x[1]+(-0.0378463784885)*x[2]**o+(0.557994858276)*x[2]
        arg[2,3,2,1]=(-0.497489149519)*x[0]**o+(0.739341493998)*x[0]+(-0.375869065986)*x[1]**o+(-0.895553158302)*x[1]+(0.916829505813)*x[2]**o+(-0.650522069672)*x[2]
        arg[2,4,0,0]=(0.627540843427)*x[0]**o+(-0.572125230014)*x[0]+(-0.404781893475)*x[1]**o+(-0.114896790942)*x[1]+(-0.108914764977)*x[2]**o+(0.739758413931)*x[2]
        arg[2,4,0,1]=(0.243205825969)*x[0]**o+(-0.989602234642)*x[0]+(-0.925730664677)*x[1]**o+(0.454801040316)*x[1]+(0.51758036389)*x[2]**o+(0.325160660047)*x[2]
        arg[2,4,1,0]=(0.559764013478)*x[0]**o+(2.69034311577e-05)*x[0]+(-0.691119937581)*x[1]**o+(-0.347457112767)*x[1]+(0.714683489268)*x[2]**o+(-0.589270137254)*x[2]
        arg[2,4,1,1]=(-0.657837524953)*x[0]**o+(0.071158671662)*x[0]+(-0.998677178035)*x[1]**o+(-0.544136374784)*x[1]+(0.6200901771)*x[2]**o+(-0.280453395396)*x[2]
        arg[2,4,2,0]=(0.692403743003)*x[0]**o+(-0.570086669428)*x[0]+(0.455036498939)*x[1]**o+(0.291433179453)*x[1]+(-0.203004457401)*x[2]**o+(0.0192489593981)*x[2]
        arg[2,4,2,1]=(-0.656783033826)*x[0]**o+(0.865748949755)*x[0]+(-0.297731124456)*x[1]**o+(-0.826786841513)*x[1]+(0.891374127144)*x[2]**o+(-0.736789456498)*x[2]
        arg[3,0,0,0]=(-0.963120586854)*x[0]**o+(0.466089565408)*x[0]+(-0.785137906849)*x[1]**o+(0.194880437113)*x[1]+(0.652057839639)*x[2]**o+(-0.63674893761)*x[2]
        arg[3,0,0,1]=(-0.923234625819)*x[0]**o+(-0.303847080864)*x[0]+(-0.374245465515)*x[1]**o+(0.131093877726)*x[1]+(-0.13469395748)*x[2]**o+(0.42721676585)*x[2]
        arg[3,0,1,0]=(-0.613510153481)*x[0]**o+(0.986696685318)*x[0]+(0.188583200151)*x[1]**o+(-0.895175969251)*x[1]+(-0.699510846377)*x[2]**o+(0.953157679279)*x[2]
        arg[3,0,1,1]=(-0.805809890231)*x[0]**o+(-0.786467529592)*x[0]+(-0.287240865489)*x[1]**o+(-0.78714682888)*x[1]+(-0.301034554724)*x[2]**o+(0.698100092246)*x[2]
        arg[3,0,2,0]=(-0.29320618172)*x[0]**o+(0.482436772935)*x[0]+(-0.126052929261)*x[1]**o+(-0.0638862140753)*x[1]+(0.42769411972)*x[2]**o+(0.741118745829)*x[2]
        arg[3,0,2,1]=(-0.380041427783)*x[0]**o+(-0.0347904596008)*x[0]+(-0.321599184546)*x[1]**o+(-0.061333518544)*x[1]+(-0.885193271364)*x[2]**o+(0.216506075886)*x[2]
        arg[3,1,0,0]=(-0.872457576682)*x[0]**o+(0.1368749584)*x[0]+(0.883209741333)*x[1]**o+(0.946644762296)*x[1]+(-0.941665219846)*x[2]**o+(0.683787986795)*x[2]
        arg[3,1,0,1]=(0.564843030077)*x[0]**o+(0.987198898299)*x[0]+(0.808116156661)*x[1]**o+(-0.712172259479)*x[1]+(0.942066890171)*x[2]**o+(-0.265957158917)*x[2]
        arg[3,1,1,0]=(-0.171018264631)*x[0]**o+(0.277202848962)*x[0]+(0.0726306773528)*x[1]**o+(-0.0265640725409)*x[1]+(-0.871118922929)*x[2]**o+(0.0562314512261)*x[2]
        arg[3,1,1,1]=(0.0861984079085)*x[0]**o+(-0.954818837051)*x[0]+(0.817362109646)*x[1]**o+(-0.513086270222)*x[1]+(-0.487472170741)*x[2]**o+(0.702514192773)*x[2]
        arg[3,1,2,0]=(0.272419057202)*x[0]**o+(-0.0454292509397)*x[0]+(-0.117471126603)*x[1]**o+(0.638237510553)*x[1]+(0.156858948691)*x[2]**o+(0.499459702141)*x[2]
        arg[3,1,2,1]=(-0.384525883334)*x[0]**o+(-0.795638828759)*x[0]+(-0.105795059118)*x[1]**o+(0.539873566967)*x[1]+(-0.402558868714)*x[2]**o+(-0.690578475243)*x[2]
        arg[3,2,0,0]=(0.776889517136)*x[0]**o+(0.710931840891)*x[0]+(0.335992677388)*x[1]**o+(-0.689182863932)*x[1]+(0.294149805572)*x[2]**o+(-0.144223257292)*x[2]
        arg[3,2,0,1]=(0.738703695733)*x[0]**o+(-0.523207061995)*x[0]+(0.707078352596)*x[1]**o+(-0.594373978096)*x[1]+(-0.833468457639)*x[2]**o+(-0.60181674446)*x[2]
        arg[3,2,1,0]=(-0.70243682985)*x[0]**o+(-0.557914174271)*x[0]+(-0.97153723053)*x[1]**o+(0.791692594024)*x[1]+(-0.442273808305)*x[2]**o+(-0.74069605321)*x[2]
        arg[3,2,1,1]=(0.136307039789)*x[0]**o+(0.820661597623)*x[0]+(0.379752060994)*x[1]**o+(-0.739083116742)*x[1]+(-0.964857055764)*x[2]**o+(0.523159928671)*x[2]
        arg[3,2,2,0]=(0.530883295074)*x[0]**o+(-0.859538509888)*x[0]+(-0.33917158254)*x[1]**o+(0.476510741055)*x[1]+(0.851975683024)*x[2]**o+(0.146277015139)*x[2]
        arg[3,2,2,1]=(-0.532984574212)*x[0]**o+(-0.413169979754)*x[0]+(0.0113284200501)*x[1]**o+(0.406768750567)*x[1]+(0.730968378807)*x[2]**o+(-0.74004422909)*x[2]
        arg[3,3,0,0]=(0.784793271465)*x[0]**o+(0.879569037436)*x[0]+(0.687804194006)*x[1]**o+(0.153617608524)*x[1]+(-0.410247334526)*x[2]**o+(0.147941373268)*x[2]
        arg[3,3,0,1]=(-0.896748874838)*x[0]**o+(-0.408404228864)*x[0]+(0.831967697989)*x[1]**o+(-0.290875629476)*x[1]+(-0.244101719785)*x[2]**o+(0.240453425519)*x[2]
        arg[3,3,1,0]=(-0.352293918285)*x[0]**o+(0.0341233590345)*x[0]+(-0.0165872699681)*x[1]**o+(-0.0668507705942)*x[1]+(-0.133733710061)*x[2]**o+(-0.00620237772745)*x[2]
        arg[3,3,1,1]=(0.173638606531)*x[0]**o+(0.619705705811)*x[0]+(0.279972531576)*x[1]**o+(-0.0811181576037)*x[1]+(-0.645673968637)*x[2]**o+(0.913753411758)*x[2]
        arg[3,3,2,0]=(0.0623921286142)*x[0]**o+(0.393414889204)*x[0]+(-0.300034851075)*x[1]**o+(-0.519515206457)*x[1]+(-0.0074981731879)*x[2]**o+(0.202996242895)*x[2]
        arg[3,3,2,1]=(0.583279597505)*x[0]**o+(-0.940603487209)*x[0]+(0.027966157338)*x[1]**o+(0.251466344692)*x[1]+(-0.97011033154)*x[2]**o+(0.46527056203)*x[2]
        arg[3,4,0,0]=(0.865170740218)*x[0]**o+(-0.466853211289)*x[0]+(0.0935492537278)*x[1]**o+(-0.632374632657)*x[1]+(0.673466346463)*x[2]**o+(-0.47934739873)*x[2]
        arg[3,4,0,1]=(0.0417451136027)*x[0]**o+(-0.781570733286)*x[0]+(0.547452389454)*x[1]**o+(-0.0286368211528)*x[1]+(-0.98907358808)*x[2]**o+(0.62749495057)*x[2]
        arg[3,4,1,0]=(-0.632067479664)*x[0]**o+(-0.872835386336)*x[0]+(-0.228785631078)*x[1]**o+(-0.224206127306)*x[1]+(0.769729051217)*x[2]**o+(-0.406750034352)*x[2]
        arg[3,4,1,1]=(0.216883071299)*x[0]**o+(-0.225632200121)*x[0]+(0.827186402743)*x[1]**o+(0.161488636886)*x[1]+(0.773801288836)*x[2]**o+(0.977006832166)*x[2]
        arg[3,4,2,0]=(-0.867098845884)*x[0]**o+(0.14989544838)*x[0]+(0.756720043262)*x[1]**o+(-0.496074302626)*x[1]+(-0.662787072451)*x[2]**o+(-0.142473973368)*x[2]
        arg[3,4,2,1]=(-0.294128396829)*x[0]**o+(-0.931379392091)*x[0]+(0.906552131161)*x[1]**o+(-0.422008485494)*x[1]+(0.806742455679)*x[2]**o+(0.940488803006)*x[2]
        ref[0,0,0,0]=(-0.0998996137438)*x_ref[0]**o+(-0.648495171588)*x_ref[0]+(-0.836177837758)*x_ref[1]**o+(-0.9597381047)*x_ref[1]+(-0.958069967907)*x_ref[2]**o+(0.399097251753)*x_ref[2]
        ref[0,0,0,1]=(0.625409406207)*x_ref[0]**o+(-0.522476054218)*x_ref[0]+(0.794724584797)*x_ref[1]**o+(-0.450875661781)*x_ref[1]+(-0.189505729613)*x_ref[2]**o+(0.0910808218379)*x_ref[2]
        ref[0,0,1,0]=(-0.36073198549)*x_ref[0]**o+(-0.0286144078107)*x_ref[0]+(-0.53827851785)*x_ref[1]**o+(-0.563600303848)*x_ref[1]+(0.970211345168)*x_ref[2]**o+(-0.405253675856)*x_ref[2]
        ref[0,0,1,1]=(-0.0204302558323)*x_ref[0]**o+(0.927570053009)*x_ref[0]+(-0.358571532503)*x_ref[1]**o+(0.282513525404)*x_ref[1]+(0.105530612734)*x_ref[2]**o+(-0.156753729378)*x_ref[2]
        ref[0,0,2,0]=(0.25049329858)*x_ref[0]**o+(0.412299654624)*x_ref[0]+(0.827683104949)*x_ref[1]**o+(0.0195885473957)*x_ref[1]+(0.849521623993)*x_ref[2]**o+(-0.107922097347)*x_ref[2]
        ref[0,0,2,1]=(0.824054630326)*x_ref[0]**o+(0.315612619612)*x_ref[0]+(-0.33528975625)*x_ref[1]**o+(-0.848543364468)*x_ref[1]+(0.67462279419)*x_ref[2]**o+(-0.0387606720463)*x_ref[2]
        ref[0,1,0,0]=(0.251613993994)*x_ref[0]**o+(-0.871217704935)*x_ref[0]+(-0.648182240752)*x_ref[1]**o+(0.324536900202)*x_ref[1]+(-0.884871901147)*x_ref[2]**o+(0.353160185772)*x_ref[2]
        ref[0,1,0,1]=(-0.629894976165)*x_ref[0]**o+(0.99092287348)*x_ref[0]+(0.87471936536)*x_ref[1]**o+(0.732884844801)*x_ref[1]+(0.423531105096)*x_ref[2]**o+(0.648959408663)*x_ref[2]
        ref[0,1,1,0]=(0.123366820062)*x_ref[0]**o+(-0.856900246151)*x_ref[0]+(-0.887819649211)*x_ref[1]**o+(-0.901367472665)*x_ref[1]+(0.222535500196)*x_ref[2]**o+(0.42830671462)*x_ref[2]
        ref[0,1,1,1]=(0.343789367141)*x_ref[0]**o+(0.614284333526)*x_ref[0]+(0.976225499577)*x_ref[1]**o+(0.955115562801)*x_ref[1]+(0.0700409126681)*x_ref[2]**o+(-0.709602275548)*x_ref[2]
        ref[0,1,2,0]=(-0.863502900367)*x_ref[0]**o+(-0.452731447667)*x_ref[0]+(0.495519161334)*x_ref[1]**o+(0.0254261045314)*x_ref[1]+(-0.484479110985)*x_ref[2]**o+(0.615421138564)*x_ref[2]
        ref[0,1,2,1]=(-0.861966549825)*x_ref[0]**o+(0.957385177208)*x_ref[0]+(0.710449966969)*x_ref[1]**o+(-0.367149885665)*x_ref[1]+(-0.100800220614)*x_ref[2]**o+(-0.266781849874)*x_ref[2]
        ref[0,2,0,0]=(-0.781309947495)*x_ref[0]**o+(-0.709036446117)*x_ref[0]+(0.86136976691)*x_ref[1]**o+(-0.745743057084)*x_ref[1]+(0.519795761984)*x_ref[2]**o+(0.298838092751)*x_ref[2]
        ref[0,2,0,1]=(-0.514037154361)*x_ref[0]**o+(-0.155867632386)*x_ref[0]+(-0.125008826668)*x_ref[1]**o+(-0.138197902584)*x_ref[1]+(0.360698647284)*x_ref[2]**o+(0.432385729464)*x_ref[2]
        ref[0,2,1,0]=(-0.19580514344)*x_ref[0]**o+(-0.430510395896)*x_ref[0]+(0.676498502854)*x_ref[1]**o+(-0.472411864377)*x_ref[1]+(-0.323468590594)*x_ref[2]**o+(0.837971311859)*x_ref[2]
        ref[0,2,1,1]=(-0.358714577548)*x_ref[0]**o+(-0.583080123482)*x_ref[0]+(-0.225887451598)*x_ref[1]**o+(0.744101609163)*x_ref[1]+(-0.0300539574094)*x_ref[2]**o+(-0.985181847505)*x_ref[2]
        ref[0,2,2,0]=(0.820610118447)*x_ref[0]**o+(-0.560334915671)*x_ref[0]+(0.478672329569)*x_ref[1]**o+(-0.956892480643)*x_ref[1]+(0.302100338619)*x_ref[2]**o+(0.597584594886)*x_ref[2]
        ref[0,2,2,1]=(-0.0658925658941)*x_ref[0]**o+(-0.136738027019)*x_ref[0]+(0.617259524472)*x_ref[1]**o+(-0.486181324584)*x_ref[1]+(0.225924222893)*x_ref[2]**o+(-0.304893108343)*x_ref[2]
        ref[0,3,0,0]=(0.0824288958004)*x_ref[0]**o+(-0.67546530428)*x_ref[0]+(-0.726843419073)*x_ref[1]**o+(0.150753236521)*x_ref[1]+(0.200621841908)*x_ref[2]**o+(0.869623957921)*x_ref[2]
        ref[0,3,0,1]=(0.66227806631)*x_ref[0]**o+(-0.389229756136)*x_ref[0]+(0.739548522805)*x_ref[1]**o+(0.131026218939)*x_ref[1]+(0.743271925877)*x_ref[2]**o+(-0.910220233981)*x_ref[2]
        ref[0,3,1,0]=(-0.259647259175)*x_ref[0]**o+(-0.208585364012)*x_ref[0]+(-0.323867955964)*x_ref[1]**o+(0.778380345151)*x_ref[1]+(-0.555780852614)*x_ref[2]**o+(-0.16366483426)*x_ref[2]
        ref[0,3,1,1]=(0.248968056109)*x_ref[0]**o+(0.84999990087)*x_ref[0]+(0.180937025847)*x_ref[1]**o+(0.332075828156)*x_ref[1]+(-0.00577042146131)*x_ref[2]**o+(0.708188905312)*x_ref[2]
        ref[0,3,2,0]=(-0.80327596428)*x_ref[0]**o+(-0.131004563471)*x_ref[0]+(0.443094741792)*x_ref[1]**o+(0.515867966266)*x_ref[1]+(-0.964003373109)*x_ref[2]**o+(-0.580721642935)*x_ref[2]
        ref[0,3,2,1]=(-0.107361389717)*x_ref[0]**o+(-0.543617395264)*x_ref[0]+(0.840867379108)*x_ref[1]**o+(-0.459941493027)*x_ref[1]+(-0.300033656162)*x_ref[2]**o+(0.440763554498)*x_ref[2]
        ref[0,4,0,0]=(0.272916362055)*x_ref[0]**o+(-0.554210888243)*x_ref[0]+(-0.458102512273)*x_ref[1]**o+(-0.54594406835)*x_ref[1]+(0.570670715852)*x_ref[2]**o+(0.792731049718)*x_ref[2]
        ref[0,4,0,1]=(0.896330380602)*x_ref[0]**o+(-0.929757950797)*x_ref[0]+(0.229049814758)*x_ref[1]**o+(0.1245974562)*x_ref[1]+(0.853250033203)*x_ref[2]**o+(-0.877052530974)*x_ref[2]
        ref[0,4,1,0]=(0.76027467594)*x_ref[0]**o+(-0.510095378376)*x_ref[0]+(0.697258966361)*x_ref[1]**o+(-0.44013714543)*x_ref[1]+(0.617675705023)*x_ref[2]**o+(-0.00118327085347)*x_ref[2]
        ref[0,4,1,1]=(0.92472377319)*x_ref[0]**o+(0.27783405618)*x_ref[0]+(-0.24317632451)*x_ref[1]**o+(-0.0264727232824)*x_ref[1]+(0.428751701408)*x_ref[2]**o+(0.184142273303)*x_ref[2]
        ref[0,4,2,0]=(0.619110983031)*x_ref[0]**o+(-0.375037512361)*x_ref[0]+(0.298576843317)*x_ref[1]**o+(-0.218802335776)*x_ref[1]+(-0.394889850336)*x_ref[2]**o+(-0.842359750924)*x_ref[2]
        ref[0,4,2,1]=(0.899714346356)*x_ref[0]**o+(0.0630133184109)*x_ref[0]+(0.0796796488932)*x_ref[1]**o+(-0.693371400327)*x_ref[1]+(-0.412090402419)*x_ref[2]**o+(0.681871832039)*x_ref[2]
        ref[1,0,0,0]=(-0.0367677770358)*x_ref[0]**o+(-0.75352301678)*x_ref[0]+(0.0705016755404)*x_ref[1]**o+(0.870115630516)*x_ref[1]+(0.136757694474)*x_ref[2]**o+(-0.557722157484)*x_ref[2]
        ref[1,0,0,1]=(0.550371937413)*x_ref[0]**o+(-0.590052751632)*x_ref[0]+(0.38665744963)*x_ref[1]**o+(-0.896537565188)*x_ref[1]+(0.550869514185)*x_ref[2]**o+(0.693828237691)*x_ref[2]
        ref[1,0,1,0]=(-0.150321579562)*x_ref[0]**o+(0.111135028893)*x_ref[0]+(0.582595079816)*x_ref[1]**o+(-0.250920804447)*x_ref[1]+(0.392998664266)*x_ref[2]**o+(-0.226735735878)*x_ref[2]
        ref[1,0,1,1]=(0.435196573865)*x_ref[0]**o+(-0.31161203165)*x_ref[0]+(-0.117580582163)*x_ref[1]**o+(0.99583282896)*x_ref[1]+(-0.164872214322)*x_ref[2]**o+(0.243837126961)*x_ref[2]
        ref[1,0,2,0]=(-0.320135099791)*x_ref[0]**o+(0.339871290643)*x_ref[0]+(0.593784119675)*x_ref[1]**o+(0.695166367768)*x_ref[1]+(0.407551427099)*x_ref[2]**o+(-0.768445418933)*x_ref[2]
        ref[1,0,2,1]=(-0.390718323619)*x_ref[0]**o+(-0.0515766643887)*x_ref[0]+(0.394112545837)*x_ref[1]**o+(-0.379592335881)*x_ref[1]+(-0.29680237927)*x_ref[2]**o+(0.500583589142)*x_ref[2]
        ref[1,1,0,0]=(-0.958930666062)*x_ref[0]**o+(0.516874100468)*x_ref[0]+(0.0859034666349)*x_ref[1]**o+(-0.370561856015)*x_ref[1]+(-0.175293417132)*x_ref[2]**o+(0.672554615126)*x_ref[2]
        ref[1,1,0,1]=(-0.38129381629)*x_ref[0]**o+(-0.615636530444)*x_ref[0]+(0.447284140424)*x_ref[1]**o+(0.507583882952)*x_ref[1]+(-0.942413761354)*x_ref[2]**o+(-0.607928807162)*x_ref[2]
        ref[1,1,1,0]=(-0.456462200294)*x_ref[0]**o+(0.778557108965)*x_ref[0]+(-0.939624730616)*x_ref[1]**o+(0.965857713341)*x_ref[1]+(-0.913772635427)*x_ref[2]**o+(-0.459963514384)*x_ref[2]
        ref[1,1,1,1]=(0.138896120228)*x_ref[0]**o+(0.941491395046)*x_ref[0]+(0.546538931051)*x_ref[1]**o+(-0.526098363992)*x_ref[1]+(0.32139556996)*x_ref[2]**o+(0.899346403792)*x_ref[2]
        ref[1,1,2,0]=(0.591597103411)*x_ref[0]**o+(0.926278363416)*x_ref[0]+(0.565926211046)*x_ref[1]**o+(0.930516925784)*x_ref[1]+(-0.498131779353)*x_ref[2]**o+(0.358157310681)*x_ref[2]
        ref[1,1,2,1]=(0.758395199797)*x_ref[0]**o+(0.769578766255)*x_ref[0]+(0.86819976051)*x_ref[1]**o+(-0.380436310842)*x_ref[1]+(-0.513012037048)*x_ref[2]**o+(-0.984361866072)*x_ref[2]
        ref[1,2,0,0]=(0.0845846076213)*x_ref[0]**o+(0.987145343601)*x_ref[0]+(-0.447680898266)*x_ref[1]**o+(-0.30278197108)*x_ref[1]+(0.271667503636)*x_ref[2]**o+(0.698967822656)*x_ref[2]
        ref[1,2,0,1]=(-0.314076111079)*x_ref[0]**o+(0.507407943673)*x_ref[0]+(-0.38467148586)*x_ref[1]**o+(0.802764957621)*x_ref[1]+(0.692949890402)*x_ref[2]**o+(-0.965766183849)*x_ref[2]
        ref[1,2,1,0]=(-0.855806343042)*x_ref[0]**o+(0.753362713203)*x_ref[0]+(-0.9877924012)*x_ref[1]**o+(-0.54047626663)*x_ref[1]+(-0.488859780266)*x_ref[2]**o+(0.444279382381)*x_ref[2]
        ref[1,2,1,1]=(-0.313257382113)*x_ref[0]**o+(-0.926058887186)*x_ref[0]+(0.509510780366)*x_ref[1]**o+(0.738049597453)*x_ref[1]+(0.191676625722)*x_ref[2]**o+(-0.354148482258)*x_ref[2]
        ref[1,2,2,0]=(-0.955204517346)*x_ref[0]**o+(0.234220317948)*x_ref[0]+(0.44435792198)*x_ref[1]**o+(0.96099571541)*x_ref[1]+(-0.422462745872)*x_ref[2]**o+(0.808817733192)*x_ref[2]
        ref[1,2,2,1]=(0.932454071169)*x_ref[0]**o+(-0.449841123911)*x_ref[0]+(-0.371587228288)*x_ref[1]**o+(-0.552185751938)*x_ref[1]+(-0.222070148528)*x_ref[2]**o+(-0.911768306425)*x_ref[2]
        ref[1,3,0,0]=(-0.355162862133)*x_ref[0]**o+(-0.00562439573347)*x_ref[0]+(0.982216168139)*x_ref[1]**o+(0.0493986220989)*x_ref[1]+(0.0242698649494)*x_ref[2]**o+(0.593411693569)*x_ref[2]
        ref[1,3,0,1]=(-0.982164819772)*x_ref[0]**o+(-0.417475530445)*x_ref[0]+(-0.421752719264)*x_ref[1]**o+(-0.102522772133)*x_ref[1]+(-0.384947583437)*x_ref[2]**o+(-0.675090951882)*x_ref[2]
        ref[1,3,1,0]=(0.862354670281)*x_ref[0]**o+(0.0332806244503)*x_ref[0]+(0.248152892705)*x_ref[1]**o+(-0.380941068103)*x_ref[1]+(0.499668747376)*x_ref[2]**o+(-0.732344330916)*x_ref[2]
        ref[1,3,1,1]=(-0.367669055003)*x_ref[0]**o+(-0.28761578782)*x_ref[0]+(0.10838982068)*x_ref[1]**o+(-0.495999837684)*x_ref[1]+(0.599492939122)*x_ref[2]**o+(0.886609670648)*x_ref[2]
        ref[1,3,2,0]=(0.707026122148)*x_ref[0]**o+(-0.625314188271)*x_ref[0]+(-0.734283031657)*x_ref[1]**o+(0.779848908302)*x_ref[1]+(-0.965798269237)*x_ref[2]**o+(0.35160103487)*x_ref[2]
        ref[1,3,2,1]=(0.198008431503)*x_ref[0]**o+(0.145470568903)*x_ref[0]+(-0.342584113223)*x_ref[1]**o+(-0.976390968602)*x_ref[1]+(-0.371678081685)*x_ref[2]**o+(0.569699640624)*x_ref[2]
        ref[1,4,0,0]=(-0.816642540092)*x_ref[0]**o+(0.466127552066)*x_ref[0]+(-0.491109738021)*x_ref[1]**o+(-0.177222784312)*x_ref[1]+(0.723624314914)*x_ref[2]**o+(-0.103929106752)*x_ref[2]
        ref[1,4,0,1]=(-0.461770336596)*x_ref[0]**o+(-0.416473447722)*x_ref[0]+(-0.844639456093)*x_ref[1]**o+(-0.021509624606)*x_ref[1]+(-0.507857805076)*x_ref[2]**o+(-0.0132005113079)*x_ref[2]
        ref[1,4,1,0]=(0.279614653607)*x_ref[0]**o+(-0.547332204747)*x_ref[0]+(0.329751635637)*x_ref[1]**o+(0.813149533168)*x_ref[1]+(0.0361758823053)*x_ref[2]**o+(0.0635753923527)*x_ref[2]
        ref[1,4,1,1]=(0.124940246141)*x_ref[0]**o+(-0.470715329674)*x_ref[0]+(0.401922174627)*x_ref[1]**o+(-0.292371975658)*x_ref[1]+(0.302438467505)*x_ref[2]**o+(0.131045792461)*x_ref[2]
        ref[1,4,2,0]=(0.817170444398)*x_ref[0]**o+(0.153182033447)*x_ref[0]+(-0.98908973731)*x_ref[1]**o+(0.359609127275)*x_ref[1]+(-0.891501663456)*x_ref[2]**o+(0.34700998982)*x_ref[2]
        ref[1,4,2,1]=(0.619050545861)*x_ref[0]**o+(-0.784466574079)*x_ref[0]+(-0.15316511355)*x_ref[1]**o+(0.587255937366)*x_ref[1]+(0.740644692907)*x_ref[2]**o+(0.0651236072508)*x_ref[2]
        ref[2,0,0,0]=(0.67904774175)*x_ref[0]**o+(-0.24998331377)*x_ref[0]+(-0.263635906274)*x_ref[1]**o+(0.157008949549)*x_ref[1]+(-0.905158896842)*x_ref[2]**o+(-0.975062080366)*x_ref[2]
        ref[2,0,0,1]=(0.724305963876)*x_ref[0]**o+(0.666891161317)*x_ref[0]+(0.769784501206)*x_ref[1]**o+(-0.84353242369)*x_ref[1]+(0.631055556571)*x_ref[2]**o+(-0.433528873211)*x_ref[2]
        ref[2,0,1,0]=(-0.545702871321)*x_ref[0]**o+(-0.153233364062)*x_ref[0]+(-0.892642075879)*x_ref[1]**o+(-0.303965445232)*x_ref[1]+(0.675938211305)*x_ref[2]**o+(0.134016007164)*x_ref[2]
        ref[2,0,1,1]=(0.612528479764)*x_ref[0]**o+(0.444039154606)*x_ref[0]+(-0.185281638413)*x_ref[1]**o+(0.643315773689)*x_ref[1]+(-0.5562387833)*x_ref[2]**o+(-0.960249468718)*x_ref[2]
        ref[2,0,2,0]=(-0.330135152971)*x_ref[0]**o+(0.359951176686)*x_ref[0]+(-0.797088515842)*x_ref[1]**o+(-0.257909258598)*x_ref[1]+(-0.927156080256)*x_ref[2]**o+(-0.843405252125)*x_ref[2]
        ref[2,0,2,1]=(0.447041226497)*x_ref[0]**o+(0.691225329527)*x_ref[0]+(0.462798475523)*x_ref[1]**o+(0.157497098383)*x_ref[1]+(-0.77785224695)*x_ref[2]**o+(0.886430058902)*x_ref[2]
        ref[2,1,0,0]=(0.397605326971)*x_ref[0]**o+(0.66227843262)*x_ref[0]+(-0.634922556623)*x_ref[1]**o+(-0.263342020187)*x_ref[1]+(-0.81486179124)*x_ref[2]**o+(-0.216716259698)*x_ref[2]
        ref[2,1,0,1]=(-0.868246547394)*x_ref[0]**o+(-0.194983059368)*x_ref[0]+(-0.667277750808)*x_ref[1]**o+(0.284011270958)*x_ref[1]+(-0.276295620425)*x_ref[2]**o+(0.215159516398)*x_ref[2]
        ref[2,1,1,0]=(0.990906320323)*x_ref[0]**o+(0.526754062323)*x_ref[0]+(-0.963398315893)*x_ref[1]**o+(-0.72561275544)*x_ref[1]+(0.539779065056)*x_ref[2]**o+(0.37738462043)*x_ref[2]
        ref[2,1,1,1]=(-0.668520028258)*x_ref[0]**o+(-0.292883094355)*x_ref[0]+(0.260949538172)*x_ref[1]**o+(-0.952706524522)*x_ref[1]+(-0.715386804875)*x_ref[2]**o+(-0.683320136224)*x_ref[2]
        ref[2,1,2,0]=(-0.646186301281)*x_ref[0]**o+(-0.62747987465)*x_ref[0]+(-0.858714171368)*x_ref[1]**o+(0.46910094021)*x_ref[1]+(0.468975652807)*x_ref[2]**o+(0.242115907129)*x_ref[2]
        ref[2,1,2,1]=(0.262205767991)*x_ref[0]**o+(-0.460376423099)*x_ref[0]+(0.88572453638)*x_ref[1]**o+(0.611761462581)*x_ref[1]+(0.763860971141)*x_ref[2]**o+(0.108470198525)*x_ref[2]
        ref[2,2,0,0]=(-0.0218691308052)*x_ref[0]**o+(0.942271392012)*x_ref[0]+(-0.778469966054)*x_ref[1]**o+(0.0873305333765)*x_ref[1]+(-0.300479183111)*x_ref[2]**o+(-0.862547713455)*x_ref[2]
        ref[2,2,0,1]=(-0.736808990311)*x_ref[0]**o+(-0.200478457878)*x_ref[0]+(0.272608362607)*x_ref[1]**o+(-0.0651056524306)*x_ref[1]+(0.616628495568)*x_ref[2]**o+(-0.8623490351)*x_ref[2]
        ref[2,2,1,0]=(0.931530668297)*x_ref[0]**o+(0.39701649536)*x_ref[0]+(0.552135074711)*x_ref[1]**o+(-0.489731326344)*x_ref[1]+(-0.864724191354)*x_ref[2]**o+(0.369557062583)*x_ref[2]
        ref[2,2,1,1]=(0.418780453395)*x_ref[0]**o+(-0.4935990418)*x_ref[0]+(0.34625989303)*x_ref[1]**o+(0.979829617988)*x_ref[1]+(0.853169987441)*x_ref[2]**o+(-0.945958722805)*x_ref[2]
        ref[2,2,2,0]=(0.973689940887)*x_ref[0]**o+(-0.0585784323994)*x_ref[0]+(-0.472419744635)*x_ref[1]**o+(0.851814193674)*x_ref[1]+(-0.393393340405)*x_ref[2]**o+(-0.755376551633)*x_ref[2]
        ref[2,2,2,1]=(-0.217049913648)*x_ref[0]**o+(-0.262150796255)*x_ref[0]+(-0.407805273745)*x_ref[1]**o+(0.819646964669)*x_ref[1]+(-0.0130316627236)*x_ref[2]**o+(0.179034130621)*x_ref[2]
        ref[2,3,0,0]=(-0.377708814803)*x_ref[0]**o+(-0.943893058411)*x_ref[0]+(-0.354341838264)*x_ref[1]**o+(0.0696491295689)*x_ref[1]+(-0.192838410156)*x_ref[2]**o+(-0.22208932914)*x_ref[2]
        ref[2,3,0,1]=(0.252429489766)*x_ref[0]**o+(0.808052283089)*x_ref[0]+(0.282941493712)*x_ref[1]**o+(0.89816007357)*x_ref[1]+(0.000979423783521)*x_ref[2]**o+(0.645599553602)*x_ref[2]
        ref[2,3,1,0]=(0.788089987784)*x_ref[0]**o+(0.414254044423)*x_ref[0]+(0.105204158639)*x_ref[1]**o+(0.97527041289)*x_ref[1]+(0.447463569594)*x_ref[2]**o+(0.91465488394)*x_ref[2]
        ref[2,3,1,1]=(0.882323493958)*x_ref[0]**o+(-0.98135461385)*x_ref[0]+(0.641881834427)*x_ref[1]**o+(-0.888068999894)*x_ref[1]+(0.705728558352)*x_ref[2]**o+(0.295815435834)*x_ref[2]
        ref[2,3,2,0]=(-0.570666876652)*x_ref[0]**o+(0.52308582534)*x_ref[0]+(-0.604213776925)*x_ref[1]**o+(-0.568684610916)*x_ref[1]+(-0.0378463784885)*x_ref[2]**o+(0.557994858276)*x_ref[2]
        ref[2,3,2,1]=(-0.497489149519)*x_ref[0]**o+(0.739341493998)*x_ref[0]+(-0.375869065986)*x_ref[1]**o+(-0.895553158302)*x_ref[1]+(0.916829505813)*x_ref[2]**o+(-0.650522069672)*x_ref[2]
        ref[2,4,0,0]=(0.627540843427)*x_ref[0]**o+(-0.572125230014)*x_ref[0]+(-0.404781893475)*x_ref[1]**o+(-0.114896790942)*x_ref[1]+(-0.108914764977)*x_ref[2]**o+(0.739758413931)*x_ref[2]
        ref[2,4,0,1]=(0.243205825969)*x_ref[0]**o+(-0.989602234642)*x_ref[0]+(-0.925730664677)*x_ref[1]**o+(0.454801040316)*x_ref[1]+(0.51758036389)*x_ref[2]**o+(0.325160660047)*x_ref[2]
        ref[2,4,1,0]=(0.559764013478)*x_ref[0]**o+(2.69034311577e-05)*x_ref[0]+(-0.691119937581)*x_ref[1]**o+(-0.347457112767)*x_ref[1]+(0.714683489268)*x_ref[2]**o+(-0.589270137254)*x_ref[2]
        ref[2,4,1,1]=(-0.657837524953)*x_ref[0]**o+(0.071158671662)*x_ref[0]+(-0.998677178035)*x_ref[1]**o+(-0.544136374784)*x_ref[1]+(0.6200901771)*x_ref[2]**o+(-0.280453395396)*x_ref[2]
        ref[2,4,2,0]=(0.692403743003)*x_ref[0]**o+(-0.570086669428)*x_ref[0]+(0.455036498939)*x_ref[1]**o+(0.291433179453)*x_ref[1]+(-0.203004457401)*x_ref[2]**o+(0.0192489593981)*x_ref[2]
        ref[2,4,2,1]=(-0.656783033826)*x_ref[0]**o+(0.865748949755)*x_ref[0]+(-0.297731124456)*x_ref[1]**o+(-0.826786841513)*x_ref[1]+(0.891374127144)*x_ref[2]**o+(-0.736789456498)*x_ref[2]
        ref[3,0,0,0]=(-0.963120586854)*x_ref[0]**o+(0.466089565408)*x_ref[0]+(-0.785137906849)*x_ref[1]**o+(0.194880437113)*x_ref[1]+(0.652057839639)*x_ref[2]**o+(-0.63674893761)*x_ref[2]
        ref[3,0,0,1]=(-0.923234625819)*x_ref[0]**o+(-0.303847080864)*x_ref[0]+(-0.374245465515)*x_ref[1]**o+(0.131093877726)*x_ref[1]+(-0.13469395748)*x_ref[2]**o+(0.42721676585)*x_ref[2]
        ref[3,0,1,0]=(-0.613510153481)*x_ref[0]**o+(0.986696685318)*x_ref[0]+(0.188583200151)*x_ref[1]**o+(-0.895175969251)*x_ref[1]+(-0.699510846377)*x_ref[2]**o+(0.953157679279)*x_ref[2]
        ref[3,0,1,1]=(-0.805809890231)*x_ref[0]**o+(-0.786467529592)*x_ref[0]+(-0.287240865489)*x_ref[1]**o+(-0.78714682888)*x_ref[1]+(-0.301034554724)*x_ref[2]**o+(0.698100092246)*x_ref[2]
        ref[3,0,2,0]=(-0.29320618172)*x_ref[0]**o+(0.482436772935)*x_ref[0]+(-0.126052929261)*x_ref[1]**o+(-0.0638862140753)*x_ref[1]+(0.42769411972)*x_ref[2]**o+(0.741118745829)*x_ref[2]
        ref[3,0,2,1]=(-0.380041427783)*x_ref[0]**o+(-0.0347904596008)*x_ref[0]+(-0.321599184546)*x_ref[1]**o+(-0.061333518544)*x_ref[1]+(-0.885193271364)*x_ref[2]**o+(0.216506075886)*x_ref[2]
        ref[3,1,0,0]=(-0.872457576682)*x_ref[0]**o+(0.1368749584)*x_ref[0]+(0.883209741333)*x_ref[1]**o+(0.946644762296)*x_ref[1]+(-0.941665219846)*x_ref[2]**o+(0.683787986795)*x_ref[2]
        ref[3,1,0,1]=(0.564843030077)*x_ref[0]**o+(0.987198898299)*x_ref[0]+(0.808116156661)*x_ref[1]**o+(-0.712172259479)*x_ref[1]+(0.942066890171)*x_ref[2]**o+(-0.265957158917)*x_ref[2]
        ref[3,1,1,0]=(-0.171018264631)*x_ref[0]**o+(0.277202848962)*x_ref[0]+(0.0726306773528)*x_ref[1]**o+(-0.0265640725409)*x_ref[1]+(-0.871118922929)*x_ref[2]**o+(0.0562314512261)*x_ref[2]
        ref[3,1,1,1]=(0.0861984079085)*x_ref[0]**o+(-0.954818837051)*x_ref[0]+(0.817362109646)*x_ref[1]**o+(-0.513086270222)*x_ref[1]+(-0.487472170741)*x_ref[2]**o+(0.702514192773)*x_ref[2]
        ref[3,1,2,0]=(0.272419057202)*x_ref[0]**o+(-0.0454292509397)*x_ref[0]+(-0.117471126603)*x_ref[1]**o+(0.638237510553)*x_ref[1]+(0.156858948691)*x_ref[2]**o+(0.499459702141)*x_ref[2]
        ref[3,1,2,1]=(-0.384525883334)*x_ref[0]**o+(-0.795638828759)*x_ref[0]+(-0.105795059118)*x_ref[1]**o+(0.539873566967)*x_ref[1]+(-0.402558868714)*x_ref[2]**o+(-0.690578475243)*x_ref[2]
        ref[3,2,0,0]=(0.776889517136)*x_ref[0]**o+(0.710931840891)*x_ref[0]+(0.335992677388)*x_ref[1]**o+(-0.689182863932)*x_ref[1]+(0.294149805572)*x_ref[2]**o+(-0.144223257292)*x_ref[2]
        ref[3,2,0,1]=(0.738703695733)*x_ref[0]**o+(-0.523207061995)*x_ref[0]+(0.707078352596)*x_ref[1]**o+(-0.594373978096)*x_ref[1]+(-0.833468457639)*x_ref[2]**o+(-0.60181674446)*x_ref[2]
        ref[3,2,1,0]=(-0.70243682985)*x_ref[0]**o+(-0.557914174271)*x_ref[0]+(-0.97153723053)*x_ref[1]**o+(0.791692594024)*x_ref[1]+(-0.442273808305)*x_ref[2]**o+(-0.74069605321)*x_ref[2]
        ref[3,2,1,1]=(0.136307039789)*x_ref[0]**o+(0.820661597623)*x_ref[0]+(0.379752060994)*x_ref[1]**o+(-0.739083116742)*x_ref[1]+(-0.964857055764)*x_ref[2]**o+(0.523159928671)*x_ref[2]
        ref[3,2,2,0]=(0.530883295074)*x_ref[0]**o+(-0.859538509888)*x_ref[0]+(-0.33917158254)*x_ref[1]**o+(0.476510741055)*x_ref[1]+(0.851975683024)*x_ref[2]**o+(0.146277015139)*x_ref[2]
        ref[3,2,2,1]=(-0.532984574212)*x_ref[0]**o+(-0.413169979754)*x_ref[0]+(0.0113284200501)*x_ref[1]**o+(0.406768750567)*x_ref[1]+(0.730968378807)*x_ref[2]**o+(-0.74004422909)*x_ref[2]
        ref[3,3,0,0]=(0.784793271465)*x_ref[0]**o+(0.879569037436)*x_ref[0]+(0.687804194006)*x_ref[1]**o+(0.153617608524)*x_ref[1]+(-0.410247334526)*x_ref[2]**o+(0.147941373268)*x_ref[2]
        ref[3,3,0,1]=(-0.896748874838)*x_ref[0]**o+(-0.408404228864)*x_ref[0]+(0.831967697989)*x_ref[1]**o+(-0.290875629476)*x_ref[1]+(-0.244101719785)*x_ref[2]**o+(0.240453425519)*x_ref[2]
        ref[3,3,1,0]=(-0.352293918285)*x_ref[0]**o+(0.0341233590345)*x_ref[0]+(-0.0165872699681)*x_ref[1]**o+(-0.0668507705942)*x_ref[1]+(-0.133733710061)*x_ref[2]**o+(-0.00620237772745)*x_ref[2]
        ref[3,3,1,1]=(0.173638606531)*x_ref[0]**o+(0.619705705811)*x_ref[0]+(0.279972531576)*x_ref[1]**o+(-0.0811181576037)*x_ref[1]+(-0.645673968637)*x_ref[2]**o+(0.913753411758)*x_ref[2]
        ref[3,3,2,0]=(0.0623921286142)*x_ref[0]**o+(0.393414889204)*x_ref[0]+(-0.300034851075)*x_ref[1]**o+(-0.519515206457)*x_ref[1]+(-0.0074981731879)*x_ref[2]**o+(0.202996242895)*x_ref[2]
        ref[3,3,2,1]=(0.583279597505)*x_ref[0]**o+(-0.940603487209)*x_ref[0]+(0.027966157338)*x_ref[1]**o+(0.251466344692)*x_ref[1]+(-0.97011033154)*x_ref[2]**o+(0.46527056203)*x_ref[2]
        ref[3,4,0,0]=(0.865170740218)*x_ref[0]**o+(-0.466853211289)*x_ref[0]+(0.0935492537278)*x_ref[1]**o+(-0.632374632657)*x_ref[1]+(0.673466346463)*x_ref[2]**o+(-0.47934739873)*x_ref[2]
        ref[3,4,0,1]=(0.0417451136027)*x_ref[0]**o+(-0.781570733286)*x_ref[0]+(0.547452389454)*x_ref[1]**o+(-0.0286368211528)*x_ref[1]+(-0.98907358808)*x_ref[2]**o+(0.62749495057)*x_ref[2]
        ref[3,4,1,0]=(-0.632067479664)*x_ref[0]**o+(-0.872835386336)*x_ref[0]+(-0.228785631078)*x_ref[1]**o+(-0.224206127306)*x_ref[1]+(0.769729051217)*x_ref[2]**o+(-0.406750034352)*x_ref[2]
        ref[3,4,1,1]=(0.216883071299)*x_ref[0]**o+(-0.225632200121)*x_ref[0]+(0.827186402743)*x_ref[1]**o+(0.161488636886)*x_ref[1]+(0.773801288836)*x_ref[2]**o+(0.977006832166)*x_ref[2]
        ref[3,4,2,0]=(-0.867098845884)*x_ref[0]**o+(0.14989544838)*x_ref[0]+(0.756720043262)*x_ref[1]**o+(-0.496074302626)*x_ref[1]+(-0.662787072451)*x_ref[2]**o+(-0.142473973368)*x_ref[2]
        ref[3,4,2,1]=(-0.294128396829)*x_ref[0]**o+(-0.931379392091)*x_ref[0]+(0.906552131161)*x_ref[1]**o+(-0.422008485494)*x_ref[1]+(0.806742455679)*x_ref[2]**o+(0.940488803006)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.387919945961)*x[0]**o+(-0.65988832446)*x[0]+(-0.271357984194)*x[1]**o+(0.911983947389)*x[1]
        ref=(-0.387919945961)*x_ref[0]**o+(-0.65988832446)*x_ref[0]+(-0.271357984194)*x_ref[1]**o+(0.911983947389)*x_ref[1]
      else:
        arg=(-0.0447075351093)*x[0]**o+(0.253461091432)*x[0]+(-0.218567784216)*x[1]**o+(-0.79030708754)*x[1]+(0.977690873872)*x[2]**o+(0.832406975763)*x[2]
        ref=(-0.0447075351093)*x_ref[0]**o+(0.253461091432)*x_ref[0]+(-0.218567784216)*x_ref[1]**o+(-0.79030708754)*x_ref[1]+(0.977690873872)*x_ref[2]**o+(0.832406975763)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.0351141106851)*x[0]**o+(-0.330587401)*x[0]+(0.244307738971)*x[1]**o+(0.623126601064)*x[1]
        arg[1]=(-0.828515624351)*x[0]**o+(-0.128313696693)*x[0]+(-0.419830633587)*x[1]**o+(0.207694344977)*x[1]
        ref[0]=(-0.0351141106851)*x_ref[0]**o+(-0.330587401)*x_ref[0]+(0.244307738971)*x_ref[1]**o+(0.623126601064)*x_ref[1]
        ref[1]=(-0.828515624351)*x_ref[0]**o+(-0.128313696693)*x_ref[0]+(-0.419830633587)*x_ref[1]**o+(0.207694344977)*x_ref[1]
      else:
        arg[0]=(-0.704794455532)*x[0]**o+(0.913653448977)*x[0]+(-0.159574360706)*x[1]**o+(-0.766199806615)*x[1]+(-0.571643737767)*x[2]**o+(-0.818256320426)*x[2]
        arg[1]=(-0.951104850067)*x[0]**o+(-0.661009825365)*x[0]+(0.415252265672)*x[1]**o+(-0.373834229809)*x[1]+(-0.727526942776)*x[2]**o+(-0.832051613933)*x[2]
        ref[0]=(-0.704794455532)*x_ref[0]**o+(0.913653448977)*x_ref[0]+(-0.159574360706)*x_ref[1]**o+(-0.766199806615)*x_ref[1]+(-0.571643737767)*x_ref[2]**o+(-0.818256320426)*x_ref[2]
        ref[1]=(-0.951104850067)*x_ref[0]**o+(-0.661009825365)*x_ref[0]+(0.415252265672)*x_ref[1]**o+(-0.373834229809)*x_ref[1]+(-0.727526942776)*x_ref[2]**o+(-0.832051613933)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.597205522265)*x[0]**o+(-0.990224489001)*x[0]+(0.409554155812)*x[1]**o+(0.0483989008223)*x[1]
        arg[0,1]=(0.397936023817)*x[0]**o+(-0.930167709289)*x[0]+(0.92611965278)*x[1]**o+(-0.964519543905)*x[1]
        arg[0,2]=(-0.311484417828)*x[0]**o+(-0.602916971577)*x[0]+(-0.420016358622)*x[1]**o+(0.837176635137)*x[1]
        arg[0,3]=(0.819115285277)*x[0]**o+(0.34838887804)*x[0]+(-0.690550904634)*x[1]**o+(0.373665882427)*x[1]
        arg[0,4]=(-0.7631912026)*x[0]**o+(-0.945320401033)*x[0]+(0.694292397122)*x[1]**o+(-0.420347487671)*x[1]
        arg[1,0]=(-0.572828652444)*x[0]**o+(0.354646387926)*x[0]+(-0.0507519017004)*x[1]**o+(-0.613889773716)*x[1]
        arg[1,1]=(-0.816957267715)*x[0]**o+(0.733537081373)*x[0]+(-0.660041553847)*x[1]**o+(0.441096986506)*x[1]
        arg[1,2]=(-0.803370878245)*x[0]**o+(-0.876568267674)*x[0]+(0.530143828975)*x[1]**o+(0.208130191749)*x[1]
        arg[1,3]=(-0.471900498331)*x[0]**o+(0.537812384385)*x[0]+(-0.882977296775)*x[1]**o+(-0.0668303027948)*x[1]
        arg[1,4]=(0.145525825808)*x[0]**o+(-0.743588928172)*x[0]+(-0.571360483951)*x[1]**o+(-0.779639689697)*x[1]
        arg[2,0]=(0.664624516473)*x[0]**o+(-0.903029393285)*x[0]+(-0.602274374813)*x[1]**o+(-0.0349025848298)*x[1]
        arg[2,1]=(-0.826235832601)*x[0]**o+(-0.56956830806)*x[0]+(-0.720529973236)*x[1]**o+(0.844549160654)*x[1]
        arg[2,2]=(0.865808902895)*x[0]**o+(0.806173922192)*x[0]+(-0.430500586315)*x[1]**o+(-0.903014287464)*x[1]
        arg[2,3]=(0.567909856768)*x[0]**o+(0.702374819746)*x[0]+(-0.931635013223)*x[1]**o+(0.35484582218)*x[1]
        arg[2,4]=(0.98750760384)*x[0]**o+(0.759340896388)*x[0]+(0.54913653446)*x[1]**o+(-0.620836882788)*x[1]
        arg[3,0]=(-0.886288358587)*x[0]**o+(0.393932009441)*x[0]+(0.370738890674)*x[1]**o+(0.540588748949)*x[1]
        arg[3,1]=(0.281435147449)*x[0]**o+(0.455838903759)*x[0]+(0.115176811183)*x[1]**o+(0.214335969407)*x[1]
        arg[3,2]=(-0.939824922583)*x[0]**o+(-0.725457358605)*x[0]+(0.0326844083908)*x[1]**o+(-0.743922869733)*x[1]
        arg[3,3]=(-0.189824771744)*x[0]**o+(-0.26508334021)*x[0]+(-0.323185410564)*x[1]**o+(0.242480018197)*x[1]
        arg[3,4]=(-0.675784178633)*x[0]**o+(0.0995857467734)*x[0]+(0.564599922701)*x[1]**o+(-0.716819470891)*x[1]
        ref[0,0]=(0.597205522265)*x_ref[0]**o+(-0.990224489001)*x_ref[0]+(0.409554155812)*x_ref[1]**o+(0.0483989008223)*x_ref[1]
        ref[0,1]=(0.397936023817)*x_ref[0]**o+(-0.930167709289)*x_ref[0]+(0.92611965278)*x_ref[1]**o+(-0.964519543905)*x_ref[1]
        ref[0,2]=(-0.311484417828)*x_ref[0]**o+(-0.602916971577)*x_ref[0]+(-0.420016358622)*x_ref[1]**o+(0.837176635137)*x_ref[1]
        ref[0,3]=(0.819115285277)*x_ref[0]**o+(0.34838887804)*x_ref[0]+(-0.690550904634)*x_ref[1]**o+(0.373665882427)*x_ref[1]
        ref[0,4]=(-0.7631912026)*x_ref[0]**o+(-0.945320401033)*x_ref[0]+(0.694292397122)*x_ref[1]**o+(-0.420347487671)*x_ref[1]
        ref[1,0]=(-0.572828652444)*x_ref[0]**o+(0.354646387926)*x_ref[0]+(-0.0507519017004)*x_ref[1]**o+(-0.613889773716)*x_ref[1]
        ref[1,1]=(-0.816957267715)*x_ref[0]**o+(0.733537081373)*x_ref[0]+(-0.660041553847)*x_ref[1]**o+(0.441096986506)*x_ref[1]
        ref[1,2]=(-0.803370878245)*x_ref[0]**o+(-0.876568267674)*x_ref[0]+(0.530143828975)*x_ref[1]**o+(0.208130191749)*x_ref[1]
        ref[1,3]=(-0.471900498331)*x_ref[0]**o+(0.537812384385)*x_ref[0]+(-0.882977296775)*x_ref[1]**o+(-0.0668303027948)*x_ref[1]
        ref[1,4]=(0.145525825808)*x_ref[0]**o+(-0.743588928172)*x_ref[0]+(-0.571360483951)*x_ref[1]**o+(-0.779639689697)*x_ref[1]
        ref[2,0]=(0.664624516473)*x_ref[0]**o+(-0.903029393285)*x_ref[0]+(-0.602274374813)*x_ref[1]**o+(-0.0349025848298)*x_ref[1]
        ref[2,1]=(-0.826235832601)*x_ref[0]**o+(-0.56956830806)*x_ref[0]+(-0.720529973236)*x_ref[1]**o+(0.844549160654)*x_ref[1]
        ref[2,2]=(0.865808902895)*x_ref[0]**o+(0.806173922192)*x_ref[0]+(-0.430500586315)*x_ref[1]**o+(-0.903014287464)*x_ref[1]
        ref[2,3]=(0.567909856768)*x_ref[0]**o+(0.702374819746)*x_ref[0]+(-0.931635013223)*x_ref[1]**o+(0.35484582218)*x_ref[1]
        ref[2,4]=(0.98750760384)*x_ref[0]**o+(0.759340896388)*x_ref[0]+(0.54913653446)*x_ref[1]**o+(-0.620836882788)*x_ref[1]
        ref[3,0]=(-0.886288358587)*x_ref[0]**o+(0.393932009441)*x_ref[0]+(0.370738890674)*x_ref[1]**o+(0.540588748949)*x_ref[1]
        ref[3,1]=(0.281435147449)*x_ref[0]**o+(0.455838903759)*x_ref[0]+(0.115176811183)*x_ref[1]**o+(0.214335969407)*x_ref[1]
        ref[3,2]=(-0.939824922583)*x_ref[0]**o+(-0.725457358605)*x_ref[0]+(0.0326844083908)*x_ref[1]**o+(-0.743922869733)*x_ref[1]
        ref[3,3]=(-0.189824771744)*x_ref[0]**o+(-0.26508334021)*x_ref[0]+(-0.323185410564)*x_ref[1]**o+(0.242480018197)*x_ref[1]
        ref[3,4]=(-0.675784178633)*x_ref[0]**o+(0.0995857467734)*x_ref[0]+(0.564599922701)*x_ref[1]**o+(-0.716819470891)*x_ref[1]
      else:
        arg[0,0]=(0.0429272016915)*x[0]**o+(-0.277438145968)*x[0]+(-0.663736923945)*x[1]**o+(0.754568662274)*x[1]+(-0.702929906882)*x[2]**o+(-0.744050838266)*x[2]
        arg[0,1]=(-0.245995343753)*x[0]**o+(-0.838699721329)*x[0]+(-0.266398695342)*x[1]**o+(0.116104359849)*x[1]+(-0.62406648751)*x[2]**o+(-0.85767565383)*x[2]
        arg[0,2]=(-0.969332932419)*x[0]**o+(-0.815839911463)*x[0]+(0.0532903209734)*x[1]**o+(-0.00108699727615)*x[1]+(0.142541999046)*x[2]**o+(0.971482543817)*x[2]
        arg[0,3]=(-0.73941517118)*x[0]**o+(0.728994326071)*x[0]+(0.962711932241)*x[1]**o+(-0.418607572448)*x[1]+(-0.374703265781)*x[2]**o+(-0.278661032739)*x[2]
        arg[0,4]=(-0.258020436163)*x[0]**o+(-0.0252121036831)*x[0]+(-0.756480314139)*x[1]**o+(-0.705045476776)*x[1]+(0.222309499093)*x[2]**o+(0.999237507123)*x[2]
        arg[1,0]=(-0.00753673411562)*x[0]**o+(-0.712947643092)*x[0]+(-0.946860154348)*x[1]**o+(-0.887293091702)*x[1]+(-0.57071975934)*x[2]**o+(-0.000995212472675)*x[2]
        arg[1,1]=(-0.547724214274)*x[0]**o+(-0.338492998124)*x[0]+(-0.191652366102)*x[1]**o+(-0.260315919511)*x[1]+(-0.103992070026)*x[2]**o+(-0.836638261621)*x[2]
        arg[1,2]=(-0.354711259362)*x[0]**o+(-0.916549411392)*x[0]+(0.744081321252)*x[1]**o+(-0.375805257126)*x[1]+(-0.78900115815)*x[2]**o+(-0.431116699853)*x[2]
        arg[1,3]=(0.543469030875)*x[0]**o+(-0.154316090927)*x[0]+(0.794234357287)*x[1]**o+(-0.894714087173)*x[1]+(-0.523490310245)*x[2]**o+(-0.734864830975)*x[2]
        arg[1,4]=(0.856478460898)*x[0]**o+(0.550846305146)*x[0]+(0.883099588119)*x[1]**o+(-0.928432890741)*x[1]+(0.953422014409)*x[2]**o+(-0.0295234113856)*x[2]
        arg[2,0]=(0.391420547389)*x[0]**o+(-0.0618665973255)*x[0]+(-0.536627285311)*x[1]**o+(-0.701891441965)*x[1]+(0.433098401229)*x[2]**o+(-0.821674981019)*x[2]
        arg[2,1]=(0.102455732687)*x[0]**o+(0.168078832413)*x[0]+(0.911099070324)*x[1]**o+(-0.776659572448)*x[1]+(0.942192938907)*x[2]**o+(-0.96284944292)*x[2]
        arg[2,2]=(0.677666801373)*x[0]**o+(0.579429978893)*x[0]+(-0.642666485576)*x[1]**o+(-0.0709929161913)*x[1]+(-0.686526389023)*x[2]**o+(0.0941806754766)*x[2]
        arg[2,3]=(-0.942137271102)*x[0]**o+(-0.448762007046)*x[0]+(0.403668316046)*x[1]**o+(-0.885790833279)*x[1]+(0.0448846125948)*x[2]**o+(0.123128346219)*x[2]
        arg[2,4]=(0.64615774988)*x[0]**o+(0.641971975809)*x[0]+(-0.0486200548159)*x[1]**o+(0.106284669825)*x[1]+(0.968867785284)*x[2]**o+(-0.693546581787)*x[2]
        arg[3,0]=(-0.0926306498682)*x[0]**o+(0.759541313455)*x[0]+(-0.569638025121)*x[1]**o+(0.236957551756)*x[1]+(-0.137429816873)*x[2]**o+(0.285836556676)*x[2]
        arg[3,1]=(0.0382484871512)*x[0]**o+(-0.481489628305)*x[0]+(0.809404670659)*x[1]**o+(0.124155422674)*x[1]+(-0.239216545446)*x[2]**o+(0.436110807769)*x[2]
        arg[3,2]=(-0.154044250435)*x[0]**o+(0.730285603192)*x[0]+(-0.986531989452)*x[1]**o+(0.0424454222102)*x[1]+(0.9651845632)*x[2]**o+(-0.173872544124)*x[2]
        arg[3,3]=(-0.528075843446)*x[0]**o+(0.226347788534)*x[0]+(-0.314941509835)*x[1]**o+(0.508180139398)*x[1]+(0.681251000904)*x[2]**o+(-0.442663397606)*x[2]
        arg[3,4]=(-0.335444251365)*x[0]**o+(-0.7788088839)*x[0]+(0.602538296871)*x[1]**o+(-0.440564659745)*x[1]+(-0.204809090144)*x[2]**o+(-0.449409088071)*x[2]
        ref[0,0]=(0.0429272016915)*x_ref[0]**o+(-0.277438145968)*x_ref[0]+(-0.663736923945)*x_ref[1]**o+(0.754568662274)*x_ref[1]+(-0.702929906882)*x_ref[2]**o+(-0.744050838266)*x_ref[2]
        ref[0,1]=(-0.245995343753)*x_ref[0]**o+(-0.838699721329)*x_ref[0]+(-0.266398695342)*x_ref[1]**o+(0.116104359849)*x_ref[1]+(-0.62406648751)*x_ref[2]**o+(-0.85767565383)*x_ref[2]
        ref[0,2]=(-0.969332932419)*x_ref[0]**o+(-0.815839911463)*x_ref[0]+(0.0532903209734)*x_ref[1]**o+(-0.00108699727615)*x_ref[1]+(0.142541999046)*x_ref[2]**o+(0.971482543817)*x_ref[2]
        ref[0,3]=(-0.73941517118)*x_ref[0]**o+(0.728994326071)*x_ref[0]+(0.962711932241)*x_ref[1]**o+(-0.418607572448)*x_ref[1]+(-0.374703265781)*x_ref[2]**o+(-0.278661032739)*x_ref[2]
        ref[0,4]=(-0.258020436163)*x_ref[0]**o+(-0.0252121036831)*x_ref[0]+(-0.756480314139)*x_ref[1]**o+(-0.705045476776)*x_ref[1]+(0.222309499093)*x_ref[2]**o+(0.999237507123)*x_ref[2]
        ref[1,0]=(-0.00753673411562)*x_ref[0]**o+(-0.712947643092)*x_ref[0]+(-0.946860154348)*x_ref[1]**o+(-0.887293091702)*x_ref[1]+(-0.57071975934)*x_ref[2]**o+(-0.000995212472675)*x_ref[2]
        ref[1,1]=(-0.547724214274)*x_ref[0]**o+(-0.338492998124)*x_ref[0]+(-0.191652366102)*x_ref[1]**o+(-0.260315919511)*x_ref[1]+(-0.103992070026)*x_ref[2]**o+(-0.836638261621)*x_ref[2]
        ref[1,2]=(-0.354711259362)*x_ref[0]**o+(-0.916549411392)*x_ref[0]+(0.744081321252)*x_ref[1]**o+(-0.375805257126)*x_ref[1]+(-0.78900115815)*x_ref[2]**o+(-0.431116699853)*x_ref[2]
        ref[1,3]=(0.543469030875)*x_ref[0]**o+(-0.154316090927)*x_ref[0]+(0.794234357287)*x_ref[1]**o+(-0.894714087173)*x_ref[1]+(-0.523490310245)*x_ref[2]**o+(-0.734864830975)*x_ref[2]
        ref[1,4]=(0.856478460898)*x_ref[0]**o+(0.550846305146)*x_ref[0]+(0.883099588119)*x_ref[1]**o+(-0.928432890741)*x_ref[1]+(0.953422014409)*x_ref[2]**o+(-0.0295234113856)*x_ref[2]
        ref[2,0]=(0.391420547389)*x_ref[0]**o+(-0.0618665973255)*x_ref[0]+(-0.536627285311)*x_ref[1]**o+(-0.701891441965)*x_ref[1]+(0.433098401229)*x_ref[2]**o+(-0.821674981019)*x_ref[2]
        ref[2,1]=(0.102455732687)*x_ref[0]**o+(0.168078832413)*x_ref[0]+(0.911099070324)*x_ref[1]**o+(-0.776659572448)*x_ref[1]+(0.942192938907)*x_ref[2]**o+(-0.96284944292)*x_ref[2]
        ref[2,2]=(0.677666801373)*x_ref[0]**o+(0.579429978893)*x_ref[0]+(-0.642666485576)*x_ref[1]**o+(-0.0709929161913)*x_ref[1]+(-0.686526389023)*x_ref[2]**o+(0.0941806754766)*x_ref[2]
        ref[2,3]=(-0.942137271102)*x_ref[0]**o+(-0.448762007046)*x_ref[0]+(0.403668316046)*x_ref[1]**o+(-0.885790833279)*x_ref[1]+(0.0448846125948)*x_ref[2]**o+(0.123128346219)*x_ref[2]
        ref[2,4]=(0.64615774988)*x_ref[0]**o+(0.641971975809)*x_ref[0]+(-0.0486200548159)*x_ref[1]**o+(0.106284669825)*x_ref[1]+(0.968867785284)*x_ref[2]**o+(-0.693546581787)*x_ref[2]
        ref[3,0]=(-0.0926306498682)*x_ref[0]**o+(0.759541313455)*x_ref[0]+(-0.569638025121)*x_ref[1]**o+(0.236957551756)*x_ref[1]+(-0.137429816873)*x_ref[2]**o+(0.285836556676)*x_ref[2]
        ref[3,1]=(0.0382484871512)*x_ref[0]**o+(-0.481489628305)*x_ref[0]+(0.809404670659)*x_ref[1]**o+(0.124155422674)*x_ref[1]+(-0.239216545446)*x_ref[2]**o+(0.436110807769)*x_ref[2]
        ref[3,2]=(-0.154044250435)*x_ref[0]**o+(0.730285603192)*x_ref[0]+(-0.986531989452)*x_ref[1]**o+(0.0424454222102)*x_ref[1]+(0.9651845632)*x_ref[2]**o+(-0.173872544124)*x_ref[2]
        ref[3,3]=(-0.528075843446)*x_ref[0]**o+(0.226347788534)*x_ref[0]+(-0.314941509835)*x_ref[1]**o+(0.508180139398)*x_ref[1]+(0.681251000904)*x_ref[2]**o+(-0.442663397606)*x_ref[2]
        ref[3,4]=(-0.335444251365)*x_ref[0]**o+(-0.7788088839)*x_ref[0]+(0.602538296871)*x_ref[1]**o+(-0.440564659745)*x_ref[1]+(-0.204809090144)*x_ref[2]**o+(-0.449409088071)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.88752640419)*x[0]**o+(-0.432514405615)*x[0]+(0.714823891809)*x[1]**o+(0.697307667986)*x[1]
        arg[0,0,1]=(0.738741506554)*x[0]**o+(0.188618518993)*x[0]+(-0.482898810269)*x[1]**o+(-0.914123176195)*x[1]
        arg[0,1,0]=(0.863872888691)*x[0]**o+(0.74507667202)*x[0]+(0.363690250942)*x[1]**o+(0.578470064401)*x[1]
        arg[0,1,1]=(-0.764594250812)*x[0]**o+(0.692105995894)*x[0]+(-0.950074761486)*x[1]**o+(-0.229643756622)*x[1]
        arg[1,0,0]=(-0.514123807377)*x[0]**o+(-0.770729480038)*x[0]+(-0.0503641588682)*x[1]**o+(-0.334340604627)*x[1]
        arg[1,0,1]=(-0.753427043851)*x[0]**o+(0.482153344254)*x[0]+(0.00162169870171)*x[1]**o+(-0.877830413948)*x[1]
        arg[1,1,0]=(0.603963041861)*x[0]**o+(-0.545048115282)*x[0]+(0.464500827027)*x[1]**o+(-0.527556750287)*x[1]
        arg[1,1,1]=(0.753201247064)*x[0]**o+(-0.99039225613)*x[0]+(-0.787959315932)*x[1]**o+(-0.918878072132)*x[1]
        arg[2,0,0]=(0.254002467345)*x[0]**o+(-0.918196533244)*x[0]+(-0.178408279918)*x[1]**o+(0.14954396814)*x[1]
        arg[2,0,1]=(0.449025666157)*x[0]**o+(-0.854430018688)*x[0]+(0.0393762923755)*x[1]**o+(0.576680446192)*x[1]
        arg[2,1,0]=(-0.252429553053)*x[0]**o+(-0.300398117312)*x[0]+(-0.0197102583755)*x[1]**o+(-0.189780116752)*x[1]
        arg[2,1,1]=(0.406291235871)*x[0]**o+(-0.763190166909)*x[0]+(-0.219957648446)*x[1]**o+(-0.928990138166)*x[1]
        arg[3,0,0]=(0.947321725555)*x[0]**o+(0.35310915009)*x[0]+(0.436067723051)*x[1]**o+(-0.86339979781)*x[1]
        arg[3,0,1]=(0.349458398624)*x[0]**o+(-0.139435138128)*x[0]+(-0.239809689531)*x[1]**o+(-0.0600998666403)*x[1]
        arg[3,1,0]=(-0.630539588121)*x[0]**o+(-0.700765237169)*x[0]+(-0.333988786061)*x[1]**o+(-0.125401172016)*x[1]
        arg[3,1,1]=(0.726389530806)*x[0]**o+(-0.691681620304)*x[0]+(0.241060086313)*x[1]**o+(-0.0970053499082)*x[1]
        arg[4,0,0]=(-0.798548958727)*x[0]**o+(0.429037250139)*x[0]+(0.598115431903)*x[1]**o+(-0.688464532796)*x[1]
        arg[4,0,1]=(0.698302852699)*x[0]**o+(0.36852807703)*x[0]+(-0.570086427187)*x[1]**o+(0.110387807246)*x[1]
        arg[4,1,0]=(0.295729624355)*x[0]**o+(0.247060092718)*x[0]+(-0.874774487091)*x[1]**o+(-0.40111840264)*x[1]
        arg[4,1,1]=(-0.773206505432)*x[0]**o+(-0.238827976714)*x[0]+(0.994288595399)*x[1]**o+(-0.6429945796)*x[1]
        arg[5,0,0]=(-0.398445578383)*x[0]**o+(-0.893518494161)*x[0]+(-0.855543638814)*x[1]**o+(0.93304149836)*x[1]
        arg[5,0,1]=(-0.811797150327)*x[0]**o+(0.0702614194127)*x[0]+(-0.0906132870911)*x[1]**o+(-0.944409144857)*x[1]
        arg[5,1,0]=(0.710407361984)*x[0]**o+(-0.854493576437)*x[0]+(-0.354683789897)*x[1]**o+(-0.896605671415)*x[1]
        arg[5,1,1]=(0.0668604398168)*x[0]**o+(0.464597275769)*x[0]+(-0.683756009627)*x[1]**o+(-0.368806656344)*x[1]
        ref[0,0,0]=(0.88752640419)*x_ref[0]**o+(-0.432514405615)*x_ref[0]+(0.714823891809)*x_ref[1]**o+(0.697307667986)*x_ref[1]
        ref[0,0,1]=(0.738741506554)*x_ref[0]**o+(0.188618518993)*x_ref[0]+(-0.482898810269)*x_ref[1]**o+(-0.914123176195)*x_ref[1]
        ref[0,1,0]=(0.863872888691)*x_ref[0]**o+(0.74507667202)*x_ref[0]+(0.363690250942)*x_ref[1]**o+(0.578470064401)*x_ref[1]
        ref[0,1,1]=(-0.764594250812)*x_ref[0]**o+(0.692105995894)*x_ref[0]+(-0.950074761486)*x_ref[1]**o+(-0.229643756622)*x_ref[1]
        ref[1,0,0]=(-0.514123807377)*x_ref[0]**o+(-0.770729480038)*x_ref[0]+(-0.0503641588682)*x_ref[1]**o+(-0.334340604627)*x_ref[1]
        ref[1,0,1]=(-0.753427043851)*x_ref[0]**o+(0.482153344254)*x_ref[0]+(0.00162169870171)*x_ref[1]**o+(-0.877830413948)*x_ref[1]
        ref[1,1,0]=(0.603963041861)*x_ref[0]**o+(-0.545048115282)*x_ref[0]+(0.464500827027)*x_ref[1]**o+(-0.527556750287)*x_ref[1]
        ref[1,1,1]=(0.753201247064)*x_ref[0]**o+(-0.99039225613)*x_ref[0]+(-0.787959315932)*x_ref[1]**o+(-0.918878072132)*x_ref[1]
        ref[2,0,0]=(0.254002467345)*x_ref[0]**o+(-0.918196533244)*x_ref[0]+(-0.178408279918)*x_ref[1]**o+(0.14954396814)*x_ref[1]
        ref[2,0,1]=(0.449025666157)*x_ref[0]**o+(-0.854430018688)*x_ref[0]+(0.0393762923755)*x_ref[1]**o+(0.576680446192)*x_ref[1]
        ref[2,1,0]=(-0.252429553053)*x_ref[0]**o+(-0.300398117312)*x_ref[0]+(-0.0197102583755)*x_ref[1]**o+(-0.189780116752)*x_ref[1]
        ref[2,1,1]=(0.406291235871)*x_ref[0]**o+(-0.763190166909)*x_ref[0]+(-0.219957648446)*x_ref[1]**o+(-0.928990138166)*x_ref[1]
        ref[3,0,0]=(0.947321725555)*x_ref[0]**o+(0.35310915009)*x_ref[0]+(0.436067723051)*x_ref[1]**o+(-0.86339979781)*x_ref[1]
        ref[3,0,1]=(0.349458398624)*x_ref[0]**o+(-0.139435138128)*x_ref[0]+(-0.239809689531)*x_ref[1]**o+(-0.0600998666403)*x_ref[1]
        ref[3,1,0]=(-0.630539588121)*x_ref[0]**o+(-0.700765237169)*x_ref[0]+(-0.333988786061)*x_ref[1]**o+(-0.125401172016)*x_ref[1]
        ref[3,1,1]=(0.726389530806)*x_ref[0]**o+(-0.691681620304)*x_ref[0]+(0.241060086313)*x_ref[1]**o+(-0.0970053499082)*x_ref[1]
        ref[4,0,0]=(-0.798548958727)*x_ref[0]**o+(0.429037250139)*x_ref[0]+(0.598115431903)*x_ref[1]**o+(-0.688464532796)*x_ref[1]
        ref[4,0,1]=(0.698302852699)*x_ref[0]**o+(0.36852807703)*x_ref[0]+(-0.570086427187)*x_ref[1]**o+(0.110387807246)*x_ref[1]
        ref[4,1,0]=(0.295729624355)*x_ref[0]**o+(0.247060092718)*x_ref[0]+(-0.874774487091)*x_ref[1]**o+(-0.40111840264)*x_ref[1]
        ref[4,1,1]=(-0.773206505432)*x_ref[0]**o+(-0.238827976714)*x_ref[0]+(0.994288595399)*x_ref[1]**o+(-0.6429945796)*x_ref[1]
        ref[5,0,0]=(-0.398445578383)*x_ref[0]**o+(-0.893518494161)*x_ref[0]+(-0.855543638814)*x_ref[1]**o+(0.93304149836)*x_ref[1]
        ref[5,0,1]=(-0.811797150327)*x_ref[0]**o+(0.0702614194127)*x_ref[0]+(-0.0906132870911)*x_ref[1]**o+(-0.944409144857)*x_ref[1]
        ref[5,1,0]=(0.710407361984)*x_ref[0]**o+(-0.854493576437)*x_ref[0]+(-0.354683789897)*x_ref[1]**o+(-0.896605671415)*x_ref[1]
        ref[5,1,1]=(0.0668604398168)*x_ref[0]**o+(0.464597275769)*x_ref[0]+(-0.683756009627)*x_ref[1]**o+(-0.368806656344)*x_ref[1]
      else:
        arg[0,0,0]=(-0.0873708254613)*x[0]**o+(-0.843989447866)*x[0]+(0.44751871924)*x[1]**o+(0.56511771844)*x[1]+(0.367414341304)*x[2]**o+(-0.147729223167)*x[2]
        arg[0,0,1]=(0.00752707306425)*x[0]**o+(0.819213438452)*x[0]+(0.14826826627)*x[1]**o+(-0.169362794728)*x[1]+(0.51736237082)*x[2]**o+(0.832746819884)*x[2]
        arg[0,1,0]=(0.607746431545)*x[0]**o+(0.253170100132)*x[0]+(-0.992132084299)*x[1]**o+(0.353865695431)*x[1]+(-0.787229685473)*x[2]**o+(0.86154695401)*x[2]
        arg[0,1,1]=(0.563636829468)*x[0]**o+(0.459627185677)*x[0]+(-0.511957702356)*x[1]**o+(0.309573033639)*x[1]+(0.806861185555)*x[2]**o+(0.182177897377)*x[2]
        arg[1,0,0]=(0.600850271814)*x[0]**o+(0.440922206975)*x[0]+(0.0307377466119)*x[1]**o+(-0.419096719454)*x[1]+(0.979284456332)*x[2]**o+(-0.00240732351601)*x[2]
        arg[1,0,1]=(0.51273517393)*x[0]**o+(-0.947643736857)*x[0]+(0.272403256871)*x[1]**o+(0.0119584891405)*x[1]+(-0.290604573286)*x[2]**o+(0.822291254588)*x[2]
        arg[1,1,0]=(-0.201909330605)*x[0]**o+(-0.560221198146)*x[0]+(0.564306911695)*x[1]**o+(-0.936933580593)*x[1]+(-0.41391834592)*x[2]**o+(0.133244793725)*x[2]
        arg[1,1,1]=(0.13318221571)*x[0]**o+(-0.565278280388)*x[0]+(0.677613822728)*x[1]**o+(-0.386612668638)*x[1]+(0.207570026134)*x[2]**o+(0.798362457061)*x[2]
        arg[2,0,0]=(-0.3777906845)*x[0]**o+(-0.285455928514)*x[0]+(-0.0901885959293)*x[1]**o+(-0.385922214051)*x[1]+(0.322566508003)*x[2]**o+(0.171191301702)*x[2]
        arg[2,0,1]=(-0.113954650916)*x[0]**o+(-0.373563461907)*x[0]+(0.73321196284)*x[1]**o+(0.535150793249)*x[1]+(-0.736944901672)*x[2]**o+(0.86133553525)*x[2]
        arg[2,1,0]=(-0.724228484035)*x[0]**o+(0.310971627294)*x[0]+(-0.497468717028)*x[1]**o+(0.467315341957)*x[1]+(-0.598941984951)*x[2]**o+(0.990367400534)*x[2]
        arg[2,1,1]=(0.594502107684)*x[0]**o+(0.261477996581)*x[0]+(0.738413631986)*x[1]**o+(0.363468523272)*x[1]+(-0.945845846052)*x[2]**o+(0.281144164925)*x[2]
        arg[3,0,0]=(0.804174935673)*x[0]**o+(-0.633954348085)*x[0]+(-0.953193492156)*x[1]**o+(0.632681376482)*x[1]+(-0.409900341362)*x[2]**o+(0.951988363964)*x[2]
        arg[3,0,1]=(-0.469895421121)*x[0]**o+(0.666115304582)*x[0]+(0.676351738247)*x[1]**o+(-0.0738256360284)*x[1]+(0.34976675523)*x[2]**o+(-0.561058490613)*x[2]
        arg[3,1,0]=(-0.999916478236)*x[0]**o+(-0.362871459365)*x[0]+(0.645604661917)*x[1]**o+(0.215878330855)*x[1]+(-0.0561173996823)*x[2]**o+(0.272143302204)*x[2]
        arg[3,1,1]=(0.233461828376)*x[0]**o+(-0.110673705352)*x[0]+(0.699710673197)*x[1]**o+(-0.628805264645)*x[1]+(0.677467640352)*x[2]**o+(-0.156224522726)*x[2]
        arg[4,0,0]=(0.180468931719)*x[0]**o+(0.299451524372)*x[0]+(-0.677131137371)*x[1]**o+(-0.422802649252)*x[1]+(0.237349200367)*x[2]**o+(0.739758623757)*x[2]
        arg[4,0,1]=(0.22209981155)*x[0]**o+(-0.226751058136)*x[0]+(-0.0345697170216)*x[1]**o+(0.265795819906)*x[1]+(0.542547943322)*x[2]**o+(0.785610668157)*x[2]
        arg[4,1,0]=(-0.520898604151)*x[0]**o+(-0.84307105084)*x[0]+(0.818061645146)*x[1]**o+(-0.173710357379)*x[1]+(0.344023171063)*x[2]**o+(0.194662166094)*x[2]
        arg[4,1,1]=(-0.288466404824)*x[0]**o+(-0.891856587908)*x[0]+(0.378329571068)*x[1]**o+(0.437919935526)*x[1]+(-0.630733422962)*x[2]**o+(0.518444630882)*x[2]
        arg[5,0,0]=(0.00978685202942)*x[0]**o+(-0.474443494209)*x[0]+(-0.509964215112)*x[1]**o+(0.341199671453)*x[1]+(-0.459333887251)*x[2]**o+(0.491961590667)*x[2]
        arg[5,0,1]=(-0.899850325159)*x[0]**o+(-0.191369137667)*x[0]+(0.18882430985)*x[1]**o+(-0.185214597442)*x[1]+(-0.967187797553)*x[2]**o+(0.366127839986)*x[2]
        arg[5,1,0]=(0.428007441668)*x[0]**o+(0.920828656579)*x[0]+(-0.298863178441)*x[1]**o+(-0.226020983335)*x[1]+(-0.01907141259)*x[2]**o+(-0.359698040557)*x[2]
        arg[5,1,1]=(0.980673633441)*x[0]**o+(0.362357489872)*x[0]+(0.825233899658)*x[1]**o+(0.846433827949)*x[1]+(-0.870197174533)*x[2]**o+(0.621884204716)*x[2]
        ref[0,0,0]=(-0.0873708254613)*x_ref[0]**o+(-0.843989447866)*x_ref[0]+(0.44751871924)*x_ref[1]**o+(0.56511771844)*x_ref[1]+(0.367414341304)*x_ref[2]**o+(-0.147729223167)*x_ref[2]
        ref[0,0,1]=(0.00752707306425)*x_ref[0]**o+(0.819213438452)*x_ref[0]+(0.14826826627)*x_ref[1]**o+(-0.169362794728)*x_ref[1]+(0.51736237082)*x_ref[2]**o+(0.832746819884)*x_ref[2]
        ref[0,1,0]=(0.607746431545)*x_ref[0]**o+(0.253170100132)*x_ref[0]+(-0.992132084299)*x_ref[1]**o+(0.353865695431)*x_ref[1]+(-0.787229685473)*x_ref[2]**o+(0.86154695401)*x_ref[2]
        ref[0,1,1]=(0.563636829468)*x_ref[0]**o+(0.459627185677)*x_ref[0]+(-0.511957702356)*x_ref[1]**o+(0.309573033639)*x_ref[1]+(0.806861185555)*x_ref[2]**o+(0.182177897377)*x_ref[2]
        ref[1,0,0]=(0.600850271814)*x_ref[0]**o+(0.440922206975)*x_ref[0]+(0.0307377466119)*x_ref[1]**o+(-0.419096719454)*x_ref[1]+(0.979284456332)*x_ref[2]**o+(-0.00240732351601)*x_ref[2]
        ref[1,0,1]=(0.51273517393)*x_ref[0]**o+(-0.947643736857)*x_ref[0]+(0.272403256871)*x_ref[1]**o+(0.0119584891405)*x_ref[1]+(-0.290604573286)*x_ref[2]**o+(0.822291254588)*x_ref[2]
        ref[1,1,0]=(-0.201909330605)*x_ref[0]**o+(-0.560221198146)*x_ref[0]+(0.564306911695)*x_ref[1]**o+(-0.936933580593)*x_ref[1]+(-0.41391834592)*x_ref[2]**o+(0.133244793725)*x_ref[2]
        ref[1,1,1]=(0.13318221571)*x_ref[0]**o+(-0.565278280388)*x_ref[0]+(0.677613822728)*x_ref[1]**o+(-0.386612668638)*x_ref[1]+(0.207570026134)*x_ref[2]**o+(0.798362457061)*x_ref[2]
        ref[2,0,0]=(-0.3777906845)*x_ref[0]**o+(-0.285455928514)*x_ref[0]+(-0.0901885959293)*x_ref[1]**o+(-0.385922214051)*x_ref[1]+(0.322566508003)*x_ref[2]**o+(0.171191301702)*x_ref[2]
        ref[2,0,1]=(-0.113954650916)*x_ref[0]**o+(-0.373563461907)*x_ref[0]+(0.73321196284)*x_ref[1]**o+(0.535150793249)*x_ref[1]+(-0.736944901672)*x_ref[2]**o+(0.86133553525)*x_ref[2]
        ref[2,1,0]=(-0.724228484035)*x_ref[0]**o+(0.310971627294)*x_ref[0]+(-0.497468717028)*x_ref[1]**o+(0.467315341957)*x_ref[1]+(-0.598941984951)*x_ref[2]**o+(0.990367400534)*x_ref[2]
        ref[2,1,1]=(0.594502107684)*x_ref[0]**o+(0.261477996581)*x_ref[0]+(0.738413631986)*x_ref[1]**o+(0.363468523272)*x_ref[1]+(-0.945845846052)*x_ref[2]**o+(0.281144164925)*x_ref[2]
        ref[3,0,0]=(0.804174935673)*x_ref[0]**o+(-0.633954348085)*x_ref[0]+(-0.953193492156)*x_ref[1]**o+(0.632681376482)*x_ref[1]+(-0.409900341362)*x_ref[2]**o+(0.951988363964)*x_ref[2]
        ref[3,0,1]=(-0.469895421121)*x_ref[0]**o+(0.666115304582)*x_ref[0]+(0.676351738247)*x_ref[1]**o+(-0.0738256360284)*x_ref[1]+(0.34976675523)*x_ref[2]**o+(-0.561058490613)*x_ref[2]
        ref[3,1,0]=(-0.999916478236)*x_ref[0]**o+(-0.362871459365)*x_ref[0]+(0.645604661917)*x_ref[1]**o+(0.215878330855)*x_ref[1]+(-0.0561173996823)*x_ref[2]**o+(0.272143302204)*x_ref[2]
        ref[3,1,1]=(0.233461828376)*x_ref[0]**o+(-0.110673705352)*x_ref[0]+(0.699710673197)*x_ref[1]**o+(-0.628805264645)*x_ref[1]+(0.677467640352)*x_ref[2]**o+(-0.156224522726)*x_ref[2]
        ref[4,0,0]=(0.180468931719)*x_ref[0]**o+(0.299451524372)*x_ref[0]+(-0.677131137371)*x_ref[1]**o+(-0.422802649252)*x_ref[1]+(0.237349200367)*x_ref[2]**o+(0.739758623757)*x_ref[2]
        ref[4,0,1]=(0.22209981155)*x_ref[0]**o+(-0.226751058136)*x_ref[0]+(-0.0345697170216)*x_ref[1]**o+(0.265795819906)*x_ref[1]+(0.542547943322)*x_ref[2]**o+(0.785610668157)*x_ref[2]
        ref[4,1,0]=(-0.520898604151)*x_ref[0]**o+(-0.84307105084)*x_ref[0]+(0.818061645146)*x_ref[1]**o+(-0.173710357379)*x_ref[1]+(0.344023171063)*x_ref[2]**o+(0.194662166094)*x_ref[2]
        ref[4,1,1]=(-0.288466404824)*x_ref[0]**o+(-0.891856587908)*x_ref[0]+(0.378329571068)*x_ref[1]**o+(0.437919935526)*x_ref[1]+(-0.630733422962)*x_ref[2]**o+(0.518444630882)*x_ref[2]
        ref[5,0,0]=(0.00978685202942)*x_ref[0]**o+(-0.474443494209)*x_ref[0]+(-0.509964215112)*x_ref[1]**o+(0.341199671453)*x_ref[1]+(-0.459333887251)*x_ref[2]**o+(0.491961590667)*x_ref[2]
        ref[5,0,1]=(-0.899850325159)*x_ref[0]**o+(-0.191369137667)*x_ref[0]+(0.18882430985)*x_ref[1]**o+(-0.185214597442)*x_ref[1]+(-0.967187797553)*x_ref[2]**o+(0.366127839986)*x_ref[2]
        ref[5,1,0]=(0.428007441668)*x_ref[0]**o+(0.920828656579)*x_ref[0]+(-0.298863178441)*x_ref[1]**o+(-0.226020983335)*x_ref[1]+(-0.01907141259)*x_ref[2]**o+(-0.359698040557)*x_ref[2]
        ref[5,1,1]=(0.980673633441)*x_ref[0]**o+(0.362357489872)*x_ref[0]+(0.825233899658)*x_ref[1]**o+(0.846433827949)*x_ref[1]+(-0.870197174533)*x_ref[2]**o+(0.621884204716)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.93572957134)*x[0]**o+(-0.878081515023)*x[0]+(0.821839395118)*x[1]**o+(-0.088026044081)*x[1]
        arg[0,0,0,1]=(-0.00393438000839)*x[0]**o+(-0.180693225971)*x[0]+(0.192781914487)*x[1]**o+(-0.561366977374)*x[1]
        arg[0,0,1,0]=(0.116028558326)*x[0]**o+(-0.47739282219)*x[0]+(0.97806148353)*x[1]**o+(0.660972051388)*x[1]
        arg[0,0,1,1]=(0.092683552469)*x[0]**o+(0.22083599745)*x[0]+(-0.423073290439)*x[1]**o+(-0.0789522599045)*x[1]
        arg[0,0,2,0]=(0.140023403084)*x[0]**o+(-0.927034796519)*x[0]+(0.0987957525564)*x[1]**o+(0.0703695028562)*x[1]
        arg[0,0,2,1]=(-0.00485147064392)*x[0]**o+(-0.649679595932)*x[0]+(0.503435626816)*x[1]**o+(0.393293792881)*x[1]
        arg[0,1,0,0]=(-0.753383897772)*x[0]**o+(-0.988499439357)*x[0]+(-0.651037417175)*x[1]**o+(0.45184892046)*x[1]
        arg[0,1,0,1]=(0.0120892569605)*x[0]**o+(0.52287896209)*x[0]+(0.103256695817)*x[1]**o+(-0.0244835681092)*x[1]
        arg[0,1,1,0]=(-0.0949239075135)*x[0]**o+(0.835062318231)*x[0]+(-0.0522981103769)*x[1]**o+(0.60020994838)*x[1]
        arg[0,1,1,1]=(0.139896941482)*x[0]**o+(0.409042945999)*x[0]+(0.10019215394)*x[1]**o+(-0.312224123017)*x[1]
        arg[0,1,2,0]=(-0.892437964342)*x[0]**o+(0.0828807279746)*x[0]+(0.777811646235)*x[1]**o+(-0.96509471481)*x[1]
        arg[0,1,2,1]=(-0.397341226148)*x[0]**o+(0.102987581336)*x[0]+(-0.69165111598)*x[1]**o+(0.743298211061)*x[1]
        arg[0,2,0,0]=(-0.300129864031)*x[0]**o+(-0.418981381232)*x[0]+(0.624854922555)*x[1]**o+(0.892651746561)*x[1]
        arg[0,2,0,1]=(-0.511463603881)*x[0]**o+(-0.808468332238)*x[0]+(-0.342755248919)*x[1]**o+(0.357985875808)*x[1]
        arg[0,2,1,0]=(-0.269520452932)*x[0]**o+(0.869553937081)*x[0]+(-0.183222122614)*x[1]**o+(-0.135280721586)*x[1]
        arg[0,2,1,1]=(0.0396712123969)*x[0]**o+(0.344314072793)*x[0]+(0.644676507168)*x[1]**o+(-0.0799551435053)*x[1]
        arg[0,2,2,0]=(0.302628743334)*x[0]**o+(-0.59557047376)*x[0]+(-0.277654790212)*x[1]**o+(0.0299035042863)*x[1]
        arg[0,2,2,1]=(0.805296568248)*x[0]**o+(0.726404711154)*x[0]+(0.0728356962577)*x[1]**o+(0.199997237962)*x[1]
        arg[0,3,0,0]=(-0.292658800952)*x[0]**o+(-0.40080106651)*x[0]+(0.844109590553)*x[1]**o+(-0.339364790362)*x[1]
        arg[0,3,0,1]=(-0.621465058894)*x[0]**o+(-0.376804350545)*x[0]+(0.745038406103)*x[1]**o+(-0.281482724757)*x[1]
        arg[0,3,1,0]=(0.00625816347702)*x[0]**o+(0.392750189511)*x[0]+(-0.232258708605)*x[1]**o+(-0.634552067635)*x[1]
        arg[0,3,1,1]=(-0.0893109545391)*x[0]**o+(-0.735678058768)*x[0]+(-0.603248350212)*x[1]**o+(-0.664366286468)*x[1]
        arg[0,3,2,0]=(-0.0539114688775)*x[0]**o+(-0.110100592496)*x[0]+(-0.700496078261)*x[1]**o+(0.203311460549)*x[1]
        arg[0,3,2,1]=(-0.554691350426)*x[0]**o+(-0.255001039659)*x[0]+(0.567977217289)*x[1]**o+(0.61993007447)*x[1]
        arg[0,4,0,0]=(-0.12193490862)*x[0]**o+(0.731139340661)*x[0]+(-0.321971724057)*x[1]**o+(0.741215588897)*x[1]
        arg[0,4,0,1]=(-0.394743164913)*x[0]**o+(-0.602249917929)*x[0]+(0.27675734517)*x[1]**o+(-0.968806169573)*x[1]
        arg[0,4,1,0]=(-0.629331232169)*x[0]**o+(0.687483125805)*x[0]+(0.633698693186)*x[1]**o+(0.837564332901)*x[1]
        arg[0,4,1,1]=(0.0470601891829)*x[0]**o+(-0.365490813497)*x[0]+(-0.162532786979)*x[1]**o+(0.504146804122)*x[1]
        arg[0,4,2,0]=(0.193790444091)*x[0]**o+(0.742132840869)*x[0]+(-0.871677496589)*x[1]**o+(0.492817750307)*x[1]
        arg[0,4,2,1]=(0.741218732478)*x[0]**o+(-0.746973604492)*x[0]+(-0.548073707926)*x[1]**o+(-0.42258185128)*x[1]
        arg[1,0,0,0]=(0.580989627459)*x[0]**o+(-0.498423525574)*x[0]+(-0.349651580594)*x[1]**o+(0.0193060779482)*x[1]
        arg[1,0,0,1]=(0.398917011903)*x[0]**o+(-0.122064497238)*x[0]+(0.924654557299)*x[1]**o+(0.485284066068)*x[1]
        arg[1,0,1,0]=(-0.985949382826)*x[0]**o+(-0.0188384874977)*x[0]+(0.177995429697)*x[1]**o+(0.967727833143)*x[1]
        arg[1,0,1,1]=(0.920799980503)*x[0]**o+(-0.749281119074)*x[0]+(-0.628411322607)*x[1]**o+(-0.446396361734)*x[1]
        arg[1,0,2,0]=(0.752863123667)*x[0]**o+(0.112427175329)*x[0]+(0.975568251269)*x[1]**o+(0.316938392669)*x[1]
        arg[1,0,2,1]=(0.798409861933)*x[0]**o+(-0.145715624761)*x[0]+(-0.847813831553)*x[1]**o+(0.302999756348)*x[1]
        arg[1,1,0,0]=(0.897029838502)*x[0]**o+(-0.540852276529)*x[0]+(-0.541450963121)*x[1]**o+(-0.118819043917)*x[1]
        arg[1,1,0,1]=(-0.616797542123)*x[0]**o+(-0.139394144428)*x[0]+(-0.480348413917)*x[1]**o+(0.500244829288)*x[1]
        arg[1,1,1,0]=(-0.960388730531)*x[0]**o+(0.87391147728)*x[0]+(0.0128673975991)*x[1]**o+(0.444341296582)*x[1]
        arg[1,1,1,1]=(0.136810869097)*x[0]**o+(-0.544756866447)*x[0]+(-0.198052789907)*x[1]**o+(-0.536732583929)*x[1]
        arg[1,1,2,0]=(-0.808981135979)*x[0]**o+(-0.122088197438)*x[0]+(-0.363958821187)*x[1]**o+(-0.521614066151)*x[1]
        arg[1,1,2,1]=(0.299382605072)*x[0]**o+(0.143397060894)*x[0]+(-0.223044874711)*x[1]**o+(-0.04723866032)*x[1]
        arg[1,2,0,0]=(-0.568967551455)*x[0]**o+(0.750777908696)*x[0]+(0.203117458853)*x[1]**o+(-0.0797168995572)*x[1]
        arg[1,2,0,1]=(0.0439147552635)*x[0]**o+(-0.0746749889653)*x[0]+(0.192885940929)*x[1]**o+(0.794090802345)*x[1]
        arg[1,2,1,0]=(0.137177876023)*x[0]**o+(0.625847668003)*x[0]+(0.252761858157)*x[1]**o+(0.496512978117)*x[1]
        arg[1,2,1,1]=(0.475218888846)*x[0]**o+(-0.286926184073)*x[0]+(0.551427803487)*x[1]**o+(0.782373661286)*x[1]
        arg[1,2,2,0]=(0.903004502545)*x[0]**o+(-0.262952251654)*x[0]+(0.110351860812)*x[1]**o+(-0.685355934696)*x[1]
        arg[1,2,2,1]=(-0.157457317404)*x[0]**o+(-0.243850623177)*x[0]+(0.28853283432)*x[1]**o+(0.612508908614)*x[1]
        arg[1,3,0,0]=(-0.741890002973)*x[0]**o+(-0.118262424825)*x[0]+(0.567100734646)*x[1]**o+(0.301557125811)*x[1]
        arg[1,3,0,1]=(-0.00745396583928)*x[0]**o+(-0.00558491344276)*x[0]+(-0.967132370197)*x[1]**o+(-0.61737594402)*x[1]
        arg[1,3,1,0]=(0.0767961523966)*x[0]**o+(0.341952000996)*x[0]+(0.762617832019)*x[1]**o+(-0.38337284713)*x[1]
        arg[1,3,1,1]=(0.32265018902)*x[0]**o+(-0.352521736206)*x[0]+(-0.0119143186529)*x[1]**o+(-0.64727561829)*x[1]
        arg[1,3,2,0]=(0.627690584587)*x[0]**o+(0.0993819822193)*x[0]+(-0.189363448841)*x[1]**o+(0.49004001194)*x[1]
        arg[1,3,2,1]=(-0.657836476156)*x[0]**o+(0.708970470136)*x[0]+(0.252190020576)*x[1]**o+(-0.481214196298)*x[1]
        arg[1,4,0,0]=(0.0837563164179)*x[0]**o+(0.741536525079)*x[0]+(0.645543094286)*x[1]**o+(0.62358090815)*x[1]
        arg[1,4,0,1]=(0.262043398371)*x[0]**o+(-0.47794750936)*x[0]+(0.93086244965)*x[1]**o+(-0.670399062677)*x[1]
        arg[1,4,1,0]=(-0.76634902923)*x[0]**o+(-0.809846143928)*x[0]+(0.639099549305)*x[1]**o+(0.793524639236)*x[1]
        arg[1,4,1,1]=(0.814737503957)*x[0]**o+(0.775593374042)*x[0]+(0.240976904111)*x[1]**o+(0.2352421356)*x[1]
        arg[1,4,2,0]=(0.519210584047)*x[0]**o+(-0.702579589691)*x[0]+(-0.450156317853)*x[1]**o+(0.588864692581)*x[1]
        arg[1,4,2,1]=(-0.927751546037)*x[0]**o+(-0.239737576019)*x[0]+(-0.527658440447)*x[1]**o+(0.79623440559)*x[1]
        arg[2,0,0,0]=(0.719144336419)*x[0]**o+(0.916780231277)*x[0]+(0.279073450277)*x[1]**o+(-0.0130671321167)*x[1]
        arg[2,0,0,1]=(-0.556604998962)*x[0]**o+(-0.661580370137)*x[0]+(0.980704664791)*x[1]**o+(0.935843251131)*x[1]
        arg[2,0,1,0]=(0.85501679753)*x[0]**o+(-0.908342467889)*x[0]+(0.969030151294)*x[1]**o+(0.534192961922)*x[1]
        arg[2,0,1,1]=(0.527672385529)*x[0]**o+(0.997699955667)*x[0]+(-0.388249629327)*x[1]**o+(-0.714956937512)*x[1]
        arg[2,0,2,0]=(-0.629149810667)*x[0]**o+(-0.0515018376692)*x[0]+(-0.80355991702)*x[1]**o+(0.235999988886)*x[1]
        arg[2,0,2,1]=(0.960715934156)*x[0]**o+(-0.782787305497)*x[0]+(-0.309348231959)*x[1]**o+(0.854911294717)*x[1]
        arg[2,1,0,0]=(-0.560749125074)*x[0]**o+(-0.120515615757)*x[0]+(0.055090634431)*x[1]**o+(-0.539205949358)*x[1]
        arg[2,1,0,1]=(-0.762113417295)*x[0]**o+(-0.207196980376)*x[0]+(0.904541245494)*x[1]**o+(0.320359459939)*x[1]
        arg[2,1,1,0]=(-0.272627586486)*x[0]**o+(-0.885987982522)*x[0]+(0.641737808216)*x[1]**o+(0.358390367373)*x[1]
        arg[2,1,1,1]=(-0.847502563761)*x[0]**o+(-0.430298277492)*x[0]+(-0.869043300185)*x[1]**o+(0.584700145915)*x[1]
        arg[2,1,2,0]=(0.497688368946)*x[0]**o+(-0.14413770704)*x[0]+(-0.82002941888)*x[1]**o+(0.573569419279)*x[1]
        arg[2,1,2,1]=(0.959549623603)*x[0]**o+(0.881196513588)*x[0]+(-0.202951096266)*x[1]**o+(-9.87983182765e-05)*x[1]
        arg[2,2,0,0]=(0.0811393810501)*x[0]**o+(-0.853555165896)*x[0]+(0.0259257246335)*x[1]**o+(0.638575306959)*x[1]
        arg[2,2,0,1]=(-0.525990031379)*x[0]**o+(0.11235160765)*x[0]+(-0.502853879455)*x[1]**o+(0.414066683507)*x[1]
        arg[2,2,1,0]=(-0.533919716834)*x[0]**o+(-0.773411100656)*x[0]+(0.311424028491)*x[1]**o+(0.332733301377)*x[1]
        arg[2,2,1,1]=(0.280548593462)*x[0]**o+(-0.0747640762104)*x[0]+(-0.7110232877)*x[1]**o+(-0.178672817441)*x[1]
        arg[2,2,2,0]=(0.445025926846)*x[0]**o+(-0.669782903005)*x[0]+(0.236610411793)*x[1]**o+(0.437836936476)*x[1]
        arg[2,2,2,1]=(-0.503761305098)*x[0]**o+(-0.889922659865)*x[0]+(-0.0767256016402)*x[1]**o+(-0.614874217835)*x[1]
        arg[2,3,0,0]=(-0.041960855672)*x[0]**o+(-0.593486242764)*x[0]+(0.344751892525)*x[1]**o+(-0.131268873786)*x[1]
        arg[2,3,0,1]=(-0.192704843083)*x[0]**o+(0.325356381696)*x[0]+(-0.235690651937)*x[1]**o+(0.59113535617)*x[1]
        arg[2,3,1,0]=(-0.0794745628024)*x[0]**o+(0.7996737586)*x[0]+(-0.324970505375)*x[1]**o+(-0.14920403468)*x[1]
        arg[2,3,1,1]=(0.31997432673)*x[0]**o+(-0.571739807799)*x[0]+(0.517683480263)*x[1]**o+(-0.250680741704)*x[1]
        arg[2,3,2,0]=(-0.379114277883)*x[0]**o+(0.953285017352)*x[0]+(-0.701111847548)*x[1]**o+(-0.354199942528)*x[1]
        arg[2,3,2,1]=(0.492245629486)*x[0]**o+(-0.299380102293)*x[0]+(-0.126371826833)*x[1]**o+(0.933664060647)*x[1]
        arg[2,4,0,0]=(0.714822159533)*x[0]**o+(0.359430380478)*x[0]+(-0.691610768506)*x[1]**o+(0.414061853203)*x[1]
        arg[2,4,0,1]=(0.412799253275)*x[0]**o+(0.862905533148)*x[0]+(-0.0507440946561)*x[1]**o+(0.779012369976)*x[1]
        arg[2,4,1,0]=(-0.981179053034)*x[0]**o+(0.984777005573)*x[0]+(0.715584788452)*x[1]**o+(0.920666888724)*x[1]
        arg[2,4,1,1]=(-0.0287775685032)*x[0]**o+(0.617592345344)*x[0]+(-0.767451297794)*x[1]**o+(0.95828976091)*x[1]
        arg[2,4,2,0]=(0.905037473705)*x[0]**o+(-0.706987158021)*x[0]+(0.535436652493)*x[1]**o+(-0.891712664249)*x[1]
        arg[2,4,2,1]=(-0.818709474925)*x[0]**o+(-0.0223479273553)*x[0]+(-0.370644568884)*x[1]**o+(0.286468739105)*x[1]
        arg[3,0,0,0]=(0.310461949876)*x[0]**o+(0.770193446516)*x[0]+(-0.0858335164699)*x[1]**o+(0.12572608122)*x[1]
        arg[3,0,0,1]=(-0.781332422249)*x[0]**o+(-0.238368527872)*x[0]+(-0.449272256582)*x[1]**o+(-0.645638738212)*x[1]
        arg[3,0,1,0]=(-0.91143447678)*x[0]**o+(-0.885131081008)*x[0]+(-0.253833463581)*x[1]**o+(0.051439649459)*x[1]
        arg[3,0,1,1]=(0.447730169203)*x[0]**o+(0.980354452325)*x[0]+(-0.0484566141047)*x[1]**o+(0.443407604362)*x[1]
        arg[3,0,2,0]=(-0.223975544426)*x[0]**o+(0.0141796274312)*x[0]+(0.664080206442)*x[1]**o+(0.193318634886)*x[1]
        arg[3,0,2,1]=(-0.388901068755)*x[0]**o+(-0.632210208163)*x[0]+(0.567264328931)*x[1]**o+(-0.308710733281)*x[1]
        arg[3,1,0,0]=(0.259562910069)*x[0]**o+(-0.315409720689)*x[0]+(0.999281841483)*x[1]**o+(0.695981779357)*x[1]
        arg[3,1,0,1]=(0.627508253716)*x[0]**o+(-0.540950954632)*x[0]+(0.626209667445)*x[1]**o+(-0.920403595109)*x[1]
        arg[3,1,1,0]=(0.750166528891)*x[0]**o+(-0.145494783124)*x[0]+(0.71048043416)*x[1]**o+(-0.0920505950401)*x[1]
        arg[3,1,1,1]=(-0.755570630755)*x[0]**o+(-0.323456877629)*x[0]+(-0.0318418288689)*x[1]**o+(0.551823313307)*x[1]
        arg[3,1,2,0]=(-0.301380276426)*x[0]**o+(-0.102899117196)*x[0]+(0.816857028902)*x[1]**o+(-0.571555621895)*x[1]
        arg[3,1,2,1]=(-0.791148729922)*x[0]**o+(0.400286696027)*x[0]+(-0.17362002139)*x[1]**o+(0.431696064424)*x[1]
        arg[3,2,0,0]=(0.241944451237)*x[0]**o+(-0.629401496412)*x[0]+(-0.00333636607647)*x[1]**o+(0.167334133677)*x[1]
        arg[3,2,0,1]=(-0.300944831932)*x[0]**o+(0.358282773504)*x[0]+(0.759495661081)*x[1]**o+(-0.225250488484)*x[1]
        arg[3,2,1,0]=(0.755950588768)*x[0]**o+(0.718423180013)*x[0]+(0.705150216505)*x[1]**o+(0.269283969196)*x[1]
        arg[3,2,1,1]=(0.811064150458)*x[0]**o+(-0.67847842696)*x[0]+(-0.478091805711)*x[1]**o+(0.292774429011)*x[1]
        arg[3,2,2,0]=(-0.307877253758)*x[0]**o+(-0.563961673764)*x[0]+(0.428322493356)*x[1]**o+(0.564230126489)*x[1]
        arg[3,2,2,1]=(-0.623877056979)*x[0]**o+(-0.148337578412)*x[0]+(-0.910662605581)*x[1]**o+(-0.38603493572)*x[1]
        arg[3,3,0,0]=(-0.826474270729)*x[0]**o+(0.977614689932)*x[0]+(-0.369824837717)*x[1]**o+(-0.985478011812)*x[1]
        arg[3,3,0,1]=(-0.251252769561)*x[0]**o+(-0.296732765057)*x[0]+(-0.888899418995)*x[1]**o+(0.640313055734)*x[1]
        arg[3,3,1,0]=(-0.189106218841)*x[0]**o+(-0.0441656339777)*x[0]+(0.368716219801)*x[1]**o+(-0.924144919367)*x[1]
        arg[3,3,1,1]=(-0.525205981948)*x[0]**o+(-0.148352446593)*x[0]+(0.869212821652)*x[1]**o+(-0.575463559793)*x[1]
        arg[3,3,2,0]=(0.668772663439)*x[0]**o+(0.4330630456)*x[0]+(-0.321158907274)*x[1]**o+(0.570853139693)*x[1]
        arg[3,3,2,1]=(0.0236674031998)*x[0]**o+(-0.390106024651)*x[0]+(0.541717893494)*x[1]**o+(-0.144655705849)*x[1]
        arg[3,4,0,0]=(0.971861188602)*x[0]**o+(-0.834463063474)*x[0]+(-0.419545662218)*x[1]**o+(-0.0387341522049)*x[1]
        arg[3,4,0,1]=(0.226357589896)*x[0]**o+(0.305148895256)*x[0]+(0.00306527642431)*x[1]**o+(0.4715619746)*x[1]
        arg[3,4,1,0]=(0.529602340964)*x[0]**o+(-0.482050910795)*x[0]+(0.964983911769)*x[1]**o+(0.524018220532)*x[1]
        arg[3,4,1,1]=(0.940676158872)*x[0]**o+(0.547078925647)*x[0]+(-0.636907424453)*x[1]**o+(0.938262396281)*x[1]
        arg[3,4,2,0]=(0.282512784255)*x[0]**o+(-0.556345478931)*x[0]+(0.651866614458)*x[1]**o+(0.8895132214)*x[1]
        arg[3,4,2,1]=(0.905077434472)*x[0]**o+(0.595878819874)*x[0]+(0.416350660905)*x[1]**o+(0.100262997545)*x[1]
        ref[0,0,0,0]=(-0.93572957134)*x_ref[0]**o+(-0.878081515023)*x_ref[0]+(0.821839395118)*x_ref[1]**o+(-0.088026044081)*x_ref[1]
        ref[0,0,0,1]=(-0.00393438000839)*x_ref[0]**o+(-0.180693225971)*x_ref[0]+(0.192781914487)*x_ref[1]**o+(-0.561366977374)*x_ref[1]
        ref[0,0,1,0]=(0.116028558326)*x_ref[0]**o+(-0.47739282219)*x_ref[0]+(0.97806148353)*x_ref[1]**o+(0.660972051388)*x_ref[1]
        ref[0,0,1,1]=(0.092683552469)*x_ref[0]**o+(0.22083599745)*x_ref[0]+(-0.423073290439)*x_ref[1]**o+(-0.0789522599045)*x_ref[1]
        ref[0,0,2,0]=(0.140023403084)*x_ref[0]**o+(-0.927034796519)*x_ref[0]+(0.0987957525564)*x_ref[1]**o+(0.0703695028562)*x_ref[1]
        ref[0,0,2,1]=(-0.00485147064392)*x_ref[0]**o+(-0.649679595932)*x_ref[0]+(0.503435626816)*x_ref[1]**o+(0.393293792881)*x_ref[1]
        ref[0,1,0,0]=(-0.753383897772)*x_ref[0]**o+(-0.988499439357)*x_ref[0]+(-0.651037417175)*x_ref[1]**o+(0.45184892046)*x_ref[1]
        ref[0,1,0,1]=(0.0120892569605)*x_ref[0]**o+(0.52287896209)*x_ref[0]+(0.103256695817)*x_ref[1]**o+(-0.0244835681092)*x_ref[1]
        ref[0,1,1,0]=(-0.0949239075135)*x_ref[0]**o+(0.835062318231)*x_ref[0]+(-0.0522981103769)*x_ref[1]**o+(0.60020994838)*x_ref[1]
        ref[0,1,1,1]=(0.139896941482)*x_ref[0]**o+(0.409042945999)*x_ref[0]+(0.10019215394)*x_ref[1]**o+(-0.312224123017)*x_ref[1]
        ref[0,1,2,0]=(-0.892437964342)*x_ref[0]**o+(0.0828807279746)*x_ref[0]+(0.777811646235)*x_ref[1]**o+(-0.96509471481)*x_ref[1]
        ref[0,1,2,1]=(-0.397341226148)*x_ref[0]**o+(0.102987581336)*x_ref[0]+(-0.69165111598)*x_ref[1]**o+(0.743298211061)*x_ref[1]
        ref[0,2,0,0]=(-0.300129864031)*x_ref[0]**o+(-0.418981381232)*x_ref[0]+(0.624854922555)*x_ref[1]**o+(0.892651746561)*x_ref[1]
        ref[0,2,0,1]=(-0.511463603881)*x_ref[0]**o+(-0.808468332238)*x_ref[0]+(-0.342755248919)*x_ref[1]**o+(0.357985875808)*x_ref[1]
        ref[0,2,1,0]=(-0.269520452932)*x_ref[0]**o+(0.869553937081)*x_ref[0]+(-0.183222122614)*x_ref[1]**o+(-0.135280721586)*x_ref[1]
        ref[0,2,1,1]=(0.0396712123969)*x_ref[0]**o+(0.344314072793)*x_ref[0]+(0.644676507168)*x_ref[1]**o+(-0.0799551435053)*x_ref[1]
        ref[0,2,2,0]=(0.302628743334)*x_ref[0]**o+(-0.59557047376)*x_ref[0]+(-0.277654790212)*x_ref[1]**o+(0.0299035042863)*x_ref[1]
        ref[0,2,2,1]=(0.805296568248)*x_ref[0]**o+(0.726404711154)*x_ref[0]+(0.0728356962577)*x_ref[1]**o+(0.199997237962)*x_ref[1]
        ref[0,3,0,0]=(-0.292658800952)*x_ref[0]**o+(-0.40080106651)*x_ref[0]+(0.844109590553)*x_ref[1]**o+(-0.339364790362)*x_ref[1]
        ref[0,3,0,1]=(-0.621465058894)*x_ref[0]**o+(-0.376804350545)*x_ref[0]+(0.745038406103)*x_ref[1]**o+(-0.281482724757)*x_ref[1]
        ref[0,3,1,0]=(0.00625816347702)*x_ref[0]**o+(0.392750189511)*x_ref[0]+(-0.232258708605)*x_ref[1]**o+(-0.634552067635)*x_ref[1]
        ref[0,3,1,1]=(-0.0893109545391)*x_ref[0]**o+(-0.735678058768)*x_ref[0]+(-0.603248350212)*x_ref[1]**o+(-0.664366286468)*x_ref[1]
        ref[0,3,2,0]=(-0.0539114688775)*x_ref[0]**o+(-0.110100592496)*x_ref[0]+(-0.700496078261)*x_ref[1]**o+(0.203311460549)*x_ref[1]
        ref[0,3,2,1]=(-0.554691350426)*x_ref[0]**o+(-0.255001039659)*x_ref[0]+(0.567977217289)*x_ref[1]**o+(0.61993007447)*x_ref[1]
        ref[0,4,0,0]=(-0.12193490862)*x_ref[0]**o+(0.731139340661)*x_ref[0]+(-0.321971724057)*x_ref[1]**o+(0.741215588897)*x_ref[1]
        ref[0,4,0,1]=(-0.394743164913)*x_ref[0]**o+(-0.602249917929)*x_ref[0]+(0.27675734517)*x_ref[1]**o+(-0.968806169573)*x_ref[1]
        ref[0,4,1,0]=(-0.629331232169)*x_ref[0]**o+(0.687483125805)*x_ref[0]+(0.633698693186)*x_ref[1]**o+(0.837564332901)*x_ref[1]
        ref[0,4,1,1]=(0.0470601891829)*x_ref[0]**o+(-0.365490813497)*x_ref[0]+(-0.162532786979)*x_ref[1]**o+(0.504146804122)*x_ref[1]
        ref[0,4,2,0]=(0.193790444091)*x_ref[0]**o+(0.742132840869)*x_ref[0]+(-0.871677496589)*x_ref[1]**o+(0.492817750307)*x_ref[1]
        ref[0,4,2,1]=(0.741218732478)*x_ref[0]**o+(-0.746973604492)*x_ref[0]+(-0.548073707926)*x_ref[1]**o+(-0.42258185128)*x_ref[1]
        ref[1,0,0,0]=(0.580989627459)*x_ref[0]**o+(-0.498423525574)*x_ref[0]+(-0.349651580594)*x_ref[1]**o+(0.0193060779482)*x_ref[1]
        ref[1,0,0,1]=(0.398917011903)*x_ref[0]**o+(-0.122064497238)*x_ref[0]+(0.924654557299)*x_ref[1]**o+(0.485284066068)*x_ref[1]
        ref[1,0,1,0]=(-0.985949382826)*x_ref[0]**o+(-0.0188384874977)*x_ref[0]+(0.177995429697)*x_ref[1]**o+(0.967727833143)*x_ref[1]
        ref[1,0,1,1]=(0.920799980503)*x_ref[0]**o+(-0.749281119074)*x_ref[0]+(-0.628411322607)*x_ref[1]**o+(-0.446396361734)*x_ref[1]
        ref[1,0,2,0]=(0.752863123667)*x_ref[0]**o+(0.112427175329)*x_ref[0]+(0.975568251269)*x_ref[1]**o+(0.316938392669)*x_ref[1]
        ref[1,0,2,1]=(0.798409861933)*x_ref[0]**o+(-0.145715624761)*x_ref[0]+(-0.847813831553)*x_ref[1]**o+(0.302999756348)*x_ref[1]
        ref[1,1,0,0]=(0.897029838502)*x_ref[0]**o+(-0.540852276529)*x_ref[0]+(-0.541450963121)*x_ref[1]**o+(-0.118819043917)*x_ref[1]
        ref[1,1,0,1]=(-0.616797542123)*x_ref[0]**o+(-0.139394144428)*x_ref[0]+(-0.480348413917)*x_ref[1]**o+(0.500244829288)*x_ref[1]
        ref[1,1,1,0]=(-0.960388730531)*x_ref[0]**o+(0.87391147728)*x_ref[0]+(0.0128673975991)*x_ref[1]**o+(0.444341296582)*x_ref[1]
        ref[1,1,1,1]=(0.136810869097)*x_ref[0]**o+(-0.544756866447)*x_ref[0]+(-0.198052789907)*x_ref[1]**o+(-0.536732583929)*x_ref[1]
        ref[1,1,2,0]=(-0.808981135979)*x_ref[0]**o+(-0.122088197438)*x_ref[0]+(-0.363958821187)*x_ref[1]**o+(-0.521614066151)*x_ref[1]
        ref[1,1,2,1]=(0.299382605072)*x_ref[0]**o+(0.143397060894)*x_ref[0]+(-0.223044874711)*x_ref[1]**o+(-0.04723866032)*x_ref[1]
        ref[1,2,0,0]=(-0.568967551455)*x_ref[0]**o+(0.750777908696)*x_ref[0]+(0.203117458853)*x_ref[1]**o+(-0.0797168995572)*x_ref[1]
        ref[1,2,0,1]=(0.0439147552635)*x_ref[0]**o+(-0.0746749889653)*x_ref[0]+(0.192885940929)*x_ref[1]**o+(0.794090802345)*x_ref[1]
        ref[1,2,1,0]=(0.137177876023)*x_ref[0]**o+(0.625847668003)*x_ref[0]+(0.252761858157)*x_ref[1]**o+(0.496512978117)*x_ref[1]
        ref[1,2,1,1]=(0.475218888846)*x_ref[0]**o+(-0.286926184073)*x_ref[0]+(0.551427803487)*x_ref[1]**o+(0.782373661286)*x_ref[1]
        ref[1,2,2,0]=(0.903004502545)*x_ref[0]**o+(-0.262952251654)*x_ref[0]+(0.110351860812)*x_ref[1]**o+(-0.685355934696)*x_ref[1]
        ref[1,2,2,1]=(-0.157457317404)*x_ref[0]**o+(-0.243850623177)*x_ref[0]+(0.28853283432)*x_ref[1]**o+(0.612508908614)*x_ref[1]
        ref[1,3,0,0]=(-0.741890002973)*x_ref[0]**o+(-0.118262424825)*x_ref[0]+(0.567100734646)*x_ref[1]**o+(0.301557125811)*x_ref[1]
        ref[1,3,0,1]=(-0.00745396583928)*x_ref[0]**o+(-0.00558491344276)*x_ref[0]+(-0.967132370197)*x_ref[1]**o+(-0.61737594402)*x_ref[1]
        ref[1,3,1,0]=(0.0767961523966)*x_ref[0]**o+(0.341952000996)*x_ref[0]+(0.762617832019)*x_ref[1]**o+(-0.38337284713)*x_ref[1]
        ref[1,3,1,1]=(0.32265018902)*x_ref[0]**o+(-0.352521736206)*x_ref[0]+(-0.0119143186529)*x_ref[1]**o+(-0.64727561829)*x_ref[1]
        ref[1,3,2,0]=(0.627690584587)*x_ref[0]**o+(0.0993819822193)*x_ref[0]+(-0.189363448841)*x_ref[1]**o+(0.49004001194)*x_ref[1]
        ref[1,3,2,1]=(-0.657836476156)*x_ref[0]**o+(0.708970470136)*x_ref[0]+(0.252190020576)*x_ref[1]**o+(-0.481214196298)*x_ref[1]
        ref[1,4,0,0]=(0.0837563164179)*x_ref[0]**o+(0.741536525079)*x_ref[0]+(0.645543094286)*x_ref[1]**o+(0.62358090815)*x_ref[1]
        ref[1,4,0,1]=(0.262043398371)*x_ref[0]**o+(-0.47794750936)*x_ref[0]+(0.93086244965)*x_ref[1]**o+(-0.670399062677)*x_ref[1]
        ref[1,4,1,0]=(-0.76634902923)*x_ref[0]**o+(-0.809846143928)*x_ref[0]+(0.639099549305)*x_ref[1]**o+(0.793524639236)*x_ref[1]
        ref[1,4,1,1]=(0.814737503957)*x_ref[0]**o+(0.775593374042)*x_ref[0]+(0.240976904111)*x_ref[1]**o+(0.2352421356)*x_ref[1]
        ref[1,4,2,0]=(0.519210584047)*x_ref[0]**o+(-0.702579589691)*x_ref[0]+(-0.450156317853)*x_ref[1]**o+(0.588864692581)*x_ref[1]
        ref[1,4,2,1]=(-0.927751546037)*x_ref[0]**o+(-0.239737576019)*x_ref[0]+(-0.527658440447)*x_ref[1]**o+(0.79623440559)*x_ref[1]
        ref[2,0,0,0]=(0.719144336419)*x_ref[0]**o+(0.916780231277)*x_ref[0]+(0.279073450277)*x_ref[1]**o+(-0.0130671321167)*x_ref[1]
        ref[2,0,0,1]=(-0.556604998962)*x_ref[0]**o+(-0.661580370137)*x_ref[0]+(0.980704664791)*x_ref[1]**o+(0.935843251131)*x_ref[1]
        ref[2,0,1,0]=(0.85501679753)*x_ref[0]**o+(-0.908342467889)*x_ref[0]+(0.969030151294)*x_ref[1]**o+(0.534192961922)*x_ref[1]
        ref[2,0,1,1]=(0.527672385529)*x_ref[0]**o+(0.997699955667)*x_ref[0]+(-0.388249629327)*x_ref[1]**o+(-0.714956937512)*x_ref[1]
        ref[2,0,2,0]=(-0.629149810667)*x_ref[0]**o+(-0.0515018376692)*x_ref[0]+(-0.80355991702)*x_ref[1]**o+(0.235999988886)*x_ref[1]
        ref[2,0,2,1]=(0.960715934156)*x_ref[0]**o+(-0.782787305497)*x_ref[0]+(-0.309348231959)*x_ref[1]**o+(0.854911294717)*x_ref[1]
        ref[2,1,0,0]=(-0.560749125074)*x_ref[0]**o+(-0.120515615757)*x_ref[0]+(0.055090634431)*x_ref[1]**o+(-0.539205949358)*x_ref[1]
        ref[2,1,0,1]=(-0.762113417295)*x_ref[0]**o+(-0.207196980376)*x_ref[0]+(0.904541245494)*x_ref[1]**o+(0.320359459939)*x_ref[1]
        ref[2,1,1,0]=(-0.272627586486)*x_ref[0]**o+(-0.885987982522)*x_ref[0]+(0.641737808216)*x_ref[1]**o+(0.358390367373)*x_ref[1]
        ref[2,1,1,1]=(-0.847502563761)*x_ref[0]**o+(-0.430298277492)*x_ref[0]+(-0.869043300185)*x_ref[1]**o+(0.584700145915)*x_ref[1]
        ref[2,1,2,0]=(0.497688368946)*x_ref[0]**o+(-0.14413770704)*x_ref[0]+(-0.82002941888)*x_ref[1]**o+(0.573569419279)*x_ref[1]
        ref[2,1,2,1]=(0.959549623603)*x_ref[0]**o+(0.881196513588)*x_ref[0]+(-0.202951096266)*x_ref[1]**o+(-9.87983182765e-05)*x_ref[1]
        ref[2,2,0,0]=(0.0811393810501)*x_ref[0]**o+(-0.853555165896)*x_ref[0]+(0.0259257246335)*x_ref[1]**o+(0.638575306959)*x_ref[1]
        ref[2,2,0,1]=(-0.525990031379)*x_ref[0]**o+(0.11235160765)*x_ref[0]+(-0.502853879455)*x_ref[1]**o+(0.414066683507)*x_ref[1]
        ref[2,2,1,0]=(-0.533919716834)*x_ref[0]**o+(-0.773411100656)*x_ref[0]+(0.311424028491)*x_ref[1]**o+(0.332733301377)*x_ref[1]
        ref[2,2,1,1]=(0.280548593462)*x_ref[0]**o+(-0.0747640762104)*x_ref[0]+(-0.7110232877)*x_ref[1]**o+(-0.178672817441)*x_ref[1]
        ref[2,2,2,0]=(0.445025926846)*x_ref[0]**o+(-0.669782903005)*x_ref[0]+(0.236610411793)*x_ref[1]**o+(0.437836936476)*x_ref[1]
        ref[2,2,2,1]=(-0.503761305098)*x_ref[0]**o+(-0.889922659865)*x_ref[0]+(-0.0767256016402)*x_ref[1]**o+(-0.614874217835)*x_ref[1]
        ref[2,3,0,0]=(-0.041960855672)*x_ref[0]**o+(-0.593486242764)*x_ref[0]+(0.344751892525)*x_ref[1]**o+(-0.131268873786)*x_ref[1]
        ref[2,3,0,1]=(-0.192704843083)*x_ref[0]**o+(0.325356381696)*x_ref[0]+(-0.235690651937)*x_ref[1]**o+(0.59113535617)*x_ref[1]
        ref[2,3,1,0]=(-0.0794745628024)*x_ref[0]**o+(0.7996737586)*x_ref[0]+(-0.324970505375)*x_ref[1]**o+(-0.14920403468)*x_ref[1]
        ref[2,3,1,1]=(0.31997432673)*x_ref[0]**o+(-0.571739807799)*x_ref[0]+(0.517683480263)*x_ref[1]**o+(-0.250680741704)*x_ref[1]
        ref[2,3,2,0]=(-0.379114277883)*x_ref[0]**o+(0.953285017352)*x_ref[0]+(-0.701111847548)*x_ref[1]**o+(-0.354199942528)*x_ref[1]
        ref[2,3,2,1]=(0.492245629486)*x_ref[0]**o+(-0.299380102293)*x_ref[0]+(-0.126371826833)*x_ref[1]**o+(0.933664060647)*x_ref[1]
        ref[2,4,0,0]=(0.714822159533)*x_ref[0]**o+(0.359430380478)*x_ref[0]+(-0.691610768506)*x_ref[1]**o+(0.414061853203)*x_ref[1]
        ref[2,4,0,1]=(0.412799253275)*x_ref[0]**o+(0.862905533148)*x_ref[0]+(-0.0507440946561)*x_ref[1]**o+(0.779012369976)*x_ref[1]
        ref[2,4,1,0]=(-0.981179053034)*x_ref[0]**o+(0.984777005573)*x_ref[0]+(0.715584788452)*x_ref[1]**o+(0.920666888724)*x_ref[1]
        ref[2,4,1,1]=(-0.0287775685032)*x_ref[0]**o+(0.617592345344)*x_ref[0]+(-0.767451297794)*x_ref[1]**o+(0.95828976091)*x_ref[1]
        ref[2,4,2,0]=(0.905037473705)*x_ref[0]**o+(-0.706987158021)*x_ref[0]+(0.535436652493)*x_ref[1]**o+(-0.891712664249)*x_ref[1]
        ref[2,4,2,1]=(-0.818709474925)*x_ref[0]**o+(-0.0223479273553)*x_ref[0]+(-0.370644568884)*x_ref[1]**o+(0.286468739105)*x_ref[1]
        ref[3,0,0,0]=(0.310461949876)*x_ref[0]**o+(0.770193446516)*x_ref[0]+(-0.0858335164699)*x_ref[1]**o+(0.12572608122)*x_ref[1]
        ref[3,0,0,1]=(-0.781332422249)*x_ref[0]**o+(-0.238368527872)*x_ref[0]+(-0.449272256582)*x_ref[1]**o+(-0.645638738212)*x_ref[1]
        ref[3,0,1,0]=(-0.91143447678)*x_ref[0]**o+(-0.885131081008)*x_ref[0]+(-0.253833463581)*x_ref[1]**o+(0.051439649459)*x_ref[1]
        ref[3,0,1,1]=(0.447730169203)*x_ref[0]**o+(0.980354452325)*x_ref[0]+(-0.0484566141047)*x_ref[1]**o+(0.443407604362)*x_ref[1]
        ref[3,0,2,0]=(-0.223975544426)*x_ref[0]**o+(0.0141796274312)*x_ref[0]+(0.664080206442)*x_ref[1]**o+(0.193318634886)*x_ref[1]
        ref[3,0,2,1]=(-0.388901068755)*x_ref[0]**o+(-0.632210208163)*x_ref[0]+(0.567264328931)*x_ref[1]**o+(-0.308710733281)*x_ref[1]
        ref[3,1,0,0]=(0.259562910069)*x_ref[0]**o+(-0.315409720689)*x_ref[0]+(0.999281841483)*x_ref[1]**o+(0.695981779357)*x_ref[1]
        ref[3,1,0,1]=(0.627508253716)*x_ref[0]**o+(-0.540950954632)*x_ref[0]+(0.626209667445)*x_ref[1]**o+(-0.920403595109)*x_ref[1]
        ref[3,1,1,0]=(0.750166528891)*x_ref[0]**o+(-0.145494783124)*x_ref[0]+(0.71048043416)*x_ref[1]**o+(-0.0920505950401)*x_ref[1]
        ref[3,1,1,1]=(-0.755570630755)*x_ref[0]**o+(-0.323456877629)*x_ref[0]+(-0.0318418288689)*x_ref[1]**o+(0.551823313307)*x_ref[1]
        ref[3,1,2,0]=(-0.301380276426)*x_ref[0]**o+(-0.102899117196)*x_ref[0]+(0.816857028902)*x_ref[1]**o+(-0.571555621895)*x_ref[1]
        ref[3,1,2,1]=(-0.791148729922)*x_ref[0]**o+(0.400286696027)*x_ref[0]+(-0.17362002139)*x_ref[1]**o+(0.431696064424)*x_ref[1]
        ref[3,2,0,0]=(0.241944451237)*x_ref[0]**o+(-0.629401496412)*x_ref[0]+(-0.00333636607647)*x_ref[1]**o+(0.167334133677)*x_ref[1]
        ref[3,2,0,1]=(-0.300944831932)*x_ref[0]**o+(0.358282773504)*x_ref[0]+(0.759495661081)*x_ref[1]**o+(-0.225250488484)*x_ref[1]
        ref[3,2,1,0]=(0.755950588768)*x_ref[0]**o+(0.718423180013)*x_ref[0]+(0.705150216505)*x_ref[1]**o+(0.269283969196)*x_ref[1]
        ref[3,2,1,1]=(0.811064150458)*x_ref[0]**o+(-0.67847842696)*x_ref[0]+(-0.478091805711)*x_ref[1]**o+(0.292774429011)*x_ref[1]
        ref[3,2,2,0]=(-0.307877253758)*x_ref[0]**o+(-0.563961673764)*x_ref[0]+(0.428322493356)*x_ref[1]**o+(0.564230126489)*x_ref[1]
        ref[3,2,2,1]=(-0.623877056979)*x_ref[0]**o+(-0.148337578412)*x_ref[0]+(-0.910662605581)*x_ref[1]**o+(-0.38603493572)*x_ref[1]
        ref[3,3,0,0]=(-0.826474270729)*x_ref[0]**o+(0.977614689932)*x_ref[0]+(-0.369824837717)*x_ref[1]**o+(-0.985478011812)*x_ref[1]
        ref[3,3,0,1]=(-0.251252769561)*x_ref[0]**o+(-0.296732765057)*x_ref[0]+(-0.888899418995)*x_ref[1]**o+(0.640313055734)*x_ref[1]
        ref[3,3,1,0]=(-0.189106218841)*x_ref[0]**o+(-0.0441656339777)*x_ref[0]+(0.368716219801)*x_ref[1]**o+(-0.924144919367)*x_ref[1]
        ref[3,3,1,1]=(-0.525205981948)*x_ref[0]**o+(-0.148352446593)*x_ref[0]+(0.869212821652)*x_ref[1]**o+(-0.575463559793)*x_ref[1]
        ref[3,3,2,0]=(0.668772663439)*x_ref[0]**o+(0.4330630456)*x_ref[0]+(-0.321158907274)*x_ref[1]**o+(0.570853139693)*x_ref[1]
        ref[3,3,2,1]=(0.0236674031998)*x_ref[0]**o+(-0.390106024651)*x_ref[0]+(0.541717893494)*x_ref[1]**o+(-0.144655705849)*x_ref[1]
        ref[3,4,0,0]=(0.971861188602)*x_ref[0]**o+(-0.834463063474)*x_ref[0]+(-0.419545662218)*x_ref[1]**o+(-0.0387341522049)*x_ref[1]
        ref[3,4,0,1]=(0.226357589896)*x_ref[0]**o+(0.305148895256)*x_ref[0]+(0.00306527642431)*x_ref[1]**o+(0.4715619746)*x_ref[1]
        ref[3,4,1,0]=(0.529602340964)*x_ref[0]**o+(-0.482050910795)*x_ref[0]+(0.964983911769)*x_ref[1]**o+(0.524018220532)*x_ref[1]
        ref[3,4,1,1]=(0.940676158872)*x_ref[0]**o+(0.547078925647)*x_ref[0]+(-0.636907424453)*x_ref[1]**o+(0.938262396281)*x_ref[1]
        ref[3,4,2,0]=(0.282512784255)*x_ref[0]**o+(-0.556345478931)*x_ref[0]+(0.651866614458)*x_ref[1]**o+(0.8895132214)*x_ref[1]
        ref[3,4,2,1]=(0.905077434472)*x_ref[0]**o+(0.595878819874)*x_ref[0]+(0.416350660905)*x_ref[1]**o+(0.100262997545)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.836302395465)*x[0]**o+(-0.842568675604)*x[0]+(-0.780133441207)*x[1]**o+(0.54483947113)*x[1]+(-0.130253019696)*x[2]**o+(0.358303389828)*x[2]
        arg[0,0,0,1]=(-0.722847500652)*x[0]**o+(0.42887368878)*x[0]+(0.979187056933)*x[1]**o+(0.624875714982)*x[1]+(-0.672529765247)*x[2]**o+(0.822363583702)*x[2]
        arg[0,0,1,0]=(-0.28989650691)*x[0]**o+(-0.847647954773)*x[0]+(-0.11119862113)*x[1]**o+(0.0809009676818)*x[1]+(-0.672626533587)*x[2]**o+(0.580561232267)*x[2]
        arg[0,0,1,1]=(-0.64783363248)*x[0]**o+(0.212828552873)*x[0]+(-0.845297180936)*x[1]**o+(-0.714760411186)*x[1]+(-0.385173605122)*x[2]**o+(-0.760750317556)*x[2]
        arg[0,0,2,0]=(-0.328757006046)*x[0]**o+(-0.0935907722247)*x[0]+(-0.895866135795)*x[1]**o+(-0.701241976745)*x[1]+(0.893684139941)*x[2]**o+(-0.265339318222)*x[2]
        arg[0,0,2,1]=(0.134066662346)*x[0]**o+(0.0383231508856)*x[0]+(-0.347744261995)*x[1]**o+(-0.86897907336)*x[1]+(-0.287792154521)*x[2]**o+(0.859351846891)*x[2]
        arg[0,1,0,0]=(-0.635128124936)*x[0]**o+(-0.659462189733)*x[0]+(-0.0964365747936)*x[1]**o+(-0.385702168747)*x[1]+(0.179539914057)*x[2]**o+(-0.0176420847859)*x[2]
        arg[0,1,0,1]=(-0.580750807831)*x[0]**o+(0.756493147583)*x[0]+(-0.436842251289)*x[1]**o+(-0.705624646095)*x[1]+(0.983064096262)*x[2]**o+(-0.0197153825423)*x[2]
        arg[0,1,1,0]=(-0.567748582365)*x[0]**o+(0.579324420356)*x[0]+(0.0434168493419)*x[1]**o+(-0.839692920334)*x[1]+(-0.119239847745)*x[2]**o+(0.527047942834)*x[2]
        arg[0,1,1,1]=(0.52870403655)*x[0]**o+(0.700090892471)*x[0]+(0.730224313921)*x[1]**o+(-0.815451751207)*x[1]+(-0.943945382413)*x[2]**o+(-0.571714141916)*x[2]
        arg[0,1,2,0]=(-0.803841094887)*x[0]**o+(0.0595126006237)*x[0]+(-0.612341797645)*x[1]**o+(0.944552805211)*x[1]+(-0.00253009731786)*x[2]**o+(0.0174867439912)*x[2]
        arg[0,1,2,1]=(-0.145623212549)*x[0]**o+(0.546039957008)*x[0]+(-0.221425237579)*x[1]**o+(0.229367671159)*x[1]+(-0.784903869759)*x[2]**o+(0.898045208871)*x[2]
        arg[0,2,0,0]=(0.417924494005)*x[0]**o+(-0.192820051987)*x[0]+(0.878916271492)*x[1]**o+(-0.35056320984)*x[1]+(0.982897575604)*x[2]**o+(-0.225793730616)*x[2]
        arg[0,2,0,1]=(-0.493279498849)*x[0]**o+(-0.154177411087)*x[0]+(-0.923592860492)*x[1]**o+(0.872010949252)*x[1]+(-0.936252147064)*x[2]**o+(0.947136819308)*x[2]
        arg[0,2,1,0]=(0.109357630607)*x[0]**o+(-0.518825477709)*x[0]+(-0.375510667498)*x[1]**o+(0.476537999644)*x[1]+(0.676017421029)*x[2]**o+(-0.633301158984)*x[2]
        arg[0,2,1,1]=(-0.0662265225539)*x[0]**o+(-0.276634748263)*x[0]+(-0.0655431981102)*x[1]**o+(0.0578482920991)*x[1]+(-0.264370611185)*x[2]**o+(-0.674612985594)*x[2]
        arg[0,2,2,0]=(-0.924186592743)*x[0]**o+(0.197603043149)*x[0]+(-0.676882805046)*x[1]**o+(0.564142845865)*x[1]+(0.590369345306)*x[2]**o+(-0.872019490978)*x[2]
        arg[0,2,2,1]=(0.480683776968)*x[0]**o+(-0.412055284173)*x[0]+(0.829996313473)*x[1]**o+(-0.235665473048)*x[1]+(-0.429191305284)*x[2]**o+(-0.0867804318457)*x[2]
        arg[0,3,0,0]=(-0.0903660006043)*x[0]**o+(-0.507571901766)*x[0]+(-0.0305305861809)*x[1]**o+(-0.915027969989)*x[1]+(0.955644774347)*x[2]**o+(0.739244618775)*x[2]
        arg[0,3,0,1]=(0.600619122998)*x[0]**o+(0.0128699484527)*x[0]+(-0.360431921281)*x[1]**o+(0.304773619811)*x[1]+(-0.302423314054)*x[2]**o+(0.680149401572)*x[2]
        arg[0,3,1,0]=(-0.638540513323)*x[0]**o+(0.766655754067)*x[0]+(0.276879854223)*x[1]**o+(-0.628688275247)*x[1]+(0.724584107114)*x[2]**o+(0.682193551967)*x[2]
        arg[0,3,1,1]=(-0.516656801349)*x[0]**o+(-0.513901319383)*x[0]+(-0.227957010777)*x[1]**o+(-0.164363207703)*x[1]+(-0.919936042047)*x[2]**o+(0.682400184997)*x[2]
        arg[0,3,2,0]=(-0.819629653095)*x[0]**o+(0.183937049274)*x[0]+(0.0411075800579)*x[1]**o+(0.0813088270699)*x[1]+(0.762416334575)*x[2]**o+(0.236553726705)*x[2]
        arg[0,3,2,1]=(0.513141183759)*x[0]**o+(-0.452475512734)*x[0]+(0.0461424221984)*x[1]**o+(-0.673452844636)*x[1]+(-0.194493328878)*x[2]**o+(0.812459956422)*x[2]
        arg[0,4,0,0]=(0.672167415195)*x[0]**o+(-0.641004730405)*x[0]+(-0.383031420851)*x[1]**o+(0.392366431186)*x[1]+(0.66479124224)*x[2]**o+(0.845503225104)*x[2]
        arg[0,4,0,1]=(0.908241966695)*x[0]**o+(-0.816989438726)*x[0]+(-0.144902267583)*x[1]**o+(-0.654901540598)*x[1]+(-0.9840470269)*x[2]**o+(0.510970695154)*x[2]
        arg[0,4,1,0]=(0.788858704366)*x[0]**o+(-0.50266306504)*x[0]+(-0.681248796723)*x[1]**o+(0.0317663112509)*x[1]+(0.662428035619)*x[2]**o+(0.377116420912)*x[2]
        arg[0,4,1,1]=(-0.0734522829037)*x[0]**o+(0.0822741484655)*x[0]+(-0.509279578815)*x[1]**o+(-0.95372420098)*x[1]+(0.682868362508)*x[2]**o+(-0.947156883638)*x[2]
        arg[0,4,2,0]=(-0.843236607337)*x[0]**o+(-0.925326238513)*x[0]+(-0.153688519983)*x[1]**o+(-0.68456413487)*x[1]+(0.920442617368)*x[2]**o+(0.852945557893)*x[2]
        arg[0,4,2,1]=(-0.283772512528)*x[0]**o+(0.202730738521)*x[0]+(-0.287375172356)*x[1]**o+(-0.176699132816)*x[1]+(0.446391113264)*x[2]**o+(-0.783266847348)*x[2]
        arg[1,0,0,0]=(0.81284364562)*x[0]**o+(-0.454432332505)*x[0]+(-0.684692596262)*x[1]**o+(-0.345923888237)*x[1]+(0.1201728631)*x[2]**o+(0.808246135028)*x[2]
        arg[1,0,0,1]=(0.941104328706)*x[0]**o+(-0.802008654964)*x[0]+(0.902047483677)*x[1]**o+(0.777049224612)*x[1]+(0.120184332854)*x[2]**o+(0.983216899102)*x[2]
        arg[1,0,1,0]=(-0.420996925767)*x[0]**o+(0.0941750615502)*x[0]+(-0.935342933578)*x[1]**o+(0.925891822058)*x[1]+(-0.153652985582)*x[2]**o+(-0.336548986153)*x[2]
        arg[1,0,1,1]=(-0.692886221795)*x[0]**o+(-0.920419410139)*x[0]+(-0.586575583263)*x[1]**o+(-0.269347362009)*x[1]+(-0.872169736045)*x[2]**o+(-0.724727186336)*x[2]
        arg[1,0,2,0]=(-0.0171222282058)*x[0]**o+(0.338180602976)*x[0]+(-0.0345958505395)*x[1]**o+(0.631728421505)*x[1]+(0.90596443453)*x[2]**o+(-0.144785193015)*x[2]
        arg[1,0,2,1]=(-0.91636583885)*x[0]**o+(-0.807496789574)*x[0]+(0.63472495957)*x[1]**o+(0.966464343993)*x[1]+(-0.471311384621)*x[2]**o+(-0.449420138238)*x[2]
        arg[1,1,0,0]=(-0.239930503474)*x[0]**o+(-0.476439715994)*x[0]+(0.421600511888)*x[1]**o+(0.477823144792)*x[1]+(0.251409914807)*x[2]**o+(0.636818001419)*x[2]
        arg[1,1,0,1]=(-0.427510410263)*x[0]**o+(0.494810857516)*x[0]+(-0.129415181352)*x[1]**o+(0.754147770071)*x[1]+(0.0218776365985)*x[2]**o+(0.706183216616)*x[2]
        arg[1,1,1,0]=(-0.469150375165)*x[0]**o+(-0.0651101638073)*x[0]+(-0.697853003861)*x[1]**o+(0.0389330102026)*x[1]+(-0.256557942045)*x[2]**o+(0.409565340425)*x[2]
        arg[1,1,1,1]=(0.14957005608)*x[0]**o+(-0.209482968126)*x[0]+(0.368899650309)*x[1]**o+(-0.0899321000137)*x[1]+(0.282389903357)*x[2]**o+(-0.166633883916)*x[2]
        arg[1,1,2,0]=(0.0374938231714)*x[0]**o+(-0.687972899893)*x[0]+(-0.0686478107779)*x[1]**o+(0.318977937888)*x[1]+(-0.601269126788)*x[2]**o+(0.453188855786)*x[2]
        arg[1,1,2,1]=(-0.853894223222)*x[0]**o+(0.441830151478)*x[0]+(0.973114898368)*x[1]**o+(-0.264454889071)*x[1]+(-0.828332687531)*x[2]**o+(0.302493748565)*x[2]
        arg[1,2,0,0]=(0.210365824111)*x[0]**o+(0.4674117857)*x[0]+(-0.129546969044)*x[1]**o+(0.566357125352)*x[1]+(-0.721733448514)*x[2]**o+(0.154253207106)*x[2]
        arg[1,2,0,1]=(0.805325508181)*x[0]**o+(-0.929113563452)*x[0]+(-0.487301761321)*x[1]**o+(-0.822405526519)*x[1]+(0.254754490326)*x[2]**o+(-0.196480710709)*x[2]
        arg[1,2,1,0]=(-0.520777275475)*x[0]**o+(-0.912569745644)*x[0]+(-0.784504791087)*x[1]**o+(-0.903473217653)*x[1]+(0.583297107374)*x[2]**o+(0.664811898195)*x[2]
        arg[1,2,1,1]=(0.129152152958)*x[0]**o+(0.497832200867)*x[0]+(0.121217248234)*x[1]**o+(-0.183368066714)*x[1]+(-0.776145485536)*x[2]**o+(-0.223165724715)*x[2]
        arg[1,2,2,0]=(-0.698981324488)*x[0]**o+(0.00951277777924)*x[0]+(0.997410976507)*x[1]**o+(0.534670303453)*x[1]+(0.251309223218)*x[2]**o+(-0.921148646979)*x[2]
        arg[1,2,2,1]=(-0.731509934487)*x[0]**o+(0.87731926834)*x[0]+(0.0865507713913)*x[1]**o+(0.708835987019)*x[1]+(0.294182269171)*x[2]**o+(-0.316660750729)*x[2]
        arg[1,3,0,0]=(-0.282903962918)*x[0]**o+(-0.793457978844)*x[0]+(0.287425807914)*x[1]**o+(-0.0252722364835)*x[1]+(-0.956202022515)*x[2]**o+(0.650271579135)*x[2]
        arg[1,3,0,1]=(-0.946723111202)*x[0]**o+(0.669431653693)*x[0]+(-0.791125976188)*x[1]**o+(-0.466484914794)*x[1]+(-0.217225801872)*x[2]**o+(0.224719473996)*x[2]
        arg[1,3,1,0]=(0.516479301043)*x[0]**o+(-0.42420332366)*x[0]+(-0.0649227247176)*x[1]**o+(0.208849752675)*x[1]+(-0.939001099693)*x[2]**o+(-0.0537827140932)*x[2]
        arg[1,3,1,1]=(-0.292077069459)*x[0]**o+(-0.444848119744)*x[0]+(0.268592719199)*x[1]**o+(-0.349401060323)*x[1]+(-0.540803784489)*x[2]**o+(-0.487170123526)*x[2]
        arg[1,3,2,0]=(-0.626222042661)*x[0]**o+(0.245937409624)*x[0]+(-0.24692162006)*x[1]**o+(0.108230029208)*x[1]+(-0.760463241928)*x[2]**o+(-0.587078263201)*x[2]
        arg[1,3,2,1]=(-0.923891147756)*x[0]**o+(0.0961285842162)*x[0]+(0.0146853466585)*x[1]**o+(0.00951339122359)*x[1]+(-0.261383704723)*x[2]**o+(-0.0262515691324)*x[2]
        arg[1,4,0,0]=(-0.0370003524292)*x[0]**o+(-0.0463848489957)*x[0]+(0.940061458694)*x[1]**o+(0.592787058925)*x[1]+(-0.217284379638)*x[2]**o+(-0.11746834109)*x[2]
        arg[1,4,0,1]=(0.830404283093)*x[0]**o+(0.268135490403)*x[0]+(-0.894500398195)*x[1]**o+(0.751042648917)*x[1]+(0.687583292754)*x[2]**o+(-0.884596509742)*x[2]
        arg[1,4,1,0]=(0.414764592469)*x[0]**o+(-0.529613347919)*x[0]+(-0.727435115594)*x[1]**o+(-0.546579777725)*x[1]+(0.378011893971)*x[2]**o+(-0.578081183238)*x[2]
        arg[1,4,1,1]=(0.406513376747)*x[0]**o+(-0.350132869579)*x[0]+(-0.557954631299)*x[1]**o+(0.868176175754)*x[1]+(0.793021832632)*x[2]**o+(0.280529771851)*x[2]
        arg[1,4,2,0]=(0.56741953575)*x[0]**o+(0.0115947684302)*x[0]+(0.872812725363)*x[1]**o+(-0.215183541745)*x[1]+(0.64853692203)*x[2]**o+(-0.139561471922)*x[2]
        arg[1,4,2,1]=(0.650916904666)*x[0]**o+(0.878765271839)*x[0]+(0.0141264348052)*x[1]**o+(-0.56314426774)*x[1]+(-0.267854377704)*x[2]**o+(0.852271380667)*x[2]
        arg[2,0,0,0]=(0.948359141577)*x[0]**o+(-0.0881747280974)*x[0]+(0.804871563368)*x[1]**o+(-0.0298585763446)*x[1]+(0.506923333569)*x[2]**o+(-0.920947093648)*x[2]
        arg[2,0,0,1]=(-0.0144221258584)*x[0]**o+(-0.679723072824)*x[0]+(0.00829336956687)*x[1]**o+(0.675680809303)*x[1]+(-0.803195231536)*x[2]**o+(0.70240358571)*x[2]
        arg[2,0,1,0]=(-0.159174987706)*x[0]**o+(0.922309271483)*x[0]+(0.344354343035)*x[1]**o+(-0.00954038311323)*x[1]+(-0.50225452804)*x[2]**o+(0.355899734308)*x[2]
        arg[2,0,1,1]=(0.0857611780404)*x[0]**o+(0.876479431983)*x[0]+(0.105049158502)*x[1]**o+(-0.530308461426)*x[1]+(-0.553136649343)*x[2]**o+(0.385458516481)*x[2]
        arg[2,0,2,0]=(-0.260086070842)*x[0]**o+(0.417766241499)*x[0]+(-0.292370383673)*x[1]**o+(-0.321315349553)*x[1]+(-0.437301444082)*x[2]**o+(0.730845669715)*x[2]
        arg[2,0,2,1]=(0.132547674276)*x[0]**o+(0.783602886333)*x[0]+(0.822361450413)*x[1]**o+(-0.810402783)*x[1]+(0.759701191023)*x[2]**o+(0.644375788608)*x[2]
        arg[2,1,0,0]=(-0.841458111227)*x[0]**o+(0.779917641736)*x[0]+(-0.0745587217256)*x[1]**o+(0.938545280143)*x[1]+(0.454600163373)*x[2]**o+(-0.490348377697)*x[2]
        arg[2,1,0,1]=(-0.876477136174)*x[0]**o+(-0.248200369939)*x[0]+(-0.973654243816)*x[1]**o+(0.482402123794)*x[1]+(-0.776970490827)*x[2]**o+(-0.571349467703)*x[2]
        arg[2,1,1,0]=(-0.433528193102)*x[0]**o+(-0.57712915767)*x[0]+(-0.328671198058)*x[1]**o+(-0.926873019075)*x[1]+(0.314232739356)*x[2]**o+(-0.199625466412)*x[2]
        arg[2,1,1,1]=(0.892707770392)*x[0]**o+(-0.680081898234)*x[0]+(-0.754181058483)*x[1]**o+(-0.253728945575)*x[1]+(-0.906373406684)*x[2]**o+(-0.0266435080833)*x[2]
        arg[2,1,2,0]=(-0.121605378472)*x[0]**o+(-0.503955015634)*x[0]+(-0.0960767196114)*x[1]**o+(-0.937914890361)*x[1]+(0.212042931853)*x[2]**o+(0.930703155408)*x[2]
        arg[2,1,2,1]=(0.0564159441047)*x[0]**o+(0.842332721105)*x[0]+(0.954741657608)*x[1]**o+(0.73503176)*x[1]+(-0.390679938582)*x[2]**o+(-0.440014747741)*x[2]
        arg[2,2,0,0]=(0.0100637970821)*x[0]**o+(-0.0283353154145)*x[0]+(0.865360758262)*x[1]**o+(0.780975027988)*x[1]+(0.16115842919)*x[2]**o+(0.247580155656)*x[2]
        arg[2,2,0,1]=(0.61102059426)*x[0]**o+(-0.198005744101)*x[0]+(0.503568797262)*x[1]**o+(0.0288809726497)*x[1]+(-0.765127098183)*x[2]**o+(0.409803857723)*x[2]
        arg[2,2,1,0]=(-0.712272018295)*x[0]**o+(-0.62562142063)*x[0]+(-0.579501949719)*x[1]**o+(0.628539275236)*x[1]+(0.0200945124964)*x[2]**o+(0.175966024769)*x[2]
        arg[2,2,1,1]=(-0.798244445519)*x[0]**o+(-0.376440962535)*x[0]+(-0.134467401497)*x[1]**o+(0.877538294477)*x[1]+(-0.484870741192)*x[2]**o+(-0.224064469935)*x[2]
        arg[2,2,2,0]=(0.200196926368)*x[0]**o+(-0.528000867086)*x[0]+(-0.708942342624)*x[1]**o+(-0.591994969141)*x[1]+(-0.990641592656)*x[2]**o+(-0.875897347303)*x[2]
        arg[2,2,2,1]=(0.949113190797)*x[0]**o+(0.867855496786)*x[0]+(-0.0314953438065)*x[1]**o+(-0.371928428005)*x[1]+(-0.526403191063)*x[2]**o+(0.653049122326)*x[2]
        arg[2,3,0,0]=(-0.353418327127)*x[0]**o+(-0.418588306429)*x[0]+(0.400872043269)*x[1]**o+(0.141907924042)*x[1]+(-0.999869579462)*x[2]**o+(-0.410423954563)*x[2]
        arg[2,3,0,1]=(-0.240390877307)*x[0]**o+(0.0901879916115)*x[0]+(-0.0734683209657)*x[1]**o+(0.274836005886)*x[1]+(-0.336375498181)*x[2]**o+(-0.754776455221)*x[2]
        arg[2,3,1,0]=(0.259928460822)*x[0]**o+(-0.227630454708)*x[0]+(0.725968066278)*x[1]**o+(-0.545440695763)*x[1]+(-0.0354926607815)*x[2]**o+(0.365010532181)*x[2]
        arg[2,3,1,1]=(-0.969799373398)*x[0]**o+(0.514498233214)*x[0]+(0.907450925318)*x[1]**o+(-0.234993706406)*x[1]+(0.206323910806)*x[2]**o+(0.209364250369)*x[2]
        arg[2,3,2,0]=(0.712905565941)*x[0]**o+(-0.353322758263)*x[0]+(0.816861181494)*x[1]**o+(-0.362388243619)*x[1]+(0.302543517607)*x[2]**o+(-0.831280418852)*x[2]
        arg[2,3,2,1]=(-0.386373711182)*x[0]**o+(0.946914567792)*x[0]+(-0.52795085887)*x[1]**o+(0.86635245964)*x[1]+(0.654261054178)*x[2]**o+(0.452334882498)*x[2]
        arg[2,4,0,0]=(0.374747033547)*x[0]**o+(-0.3471578669)*x[0]+(0.322217002639)*x[1]**o+(0.92988068982)*x[1]+(-0.642423472369)*x[2]**o+(-0.759623225653)*x[2]
        arg[2,4,0,1]=(-0.322255157841)*x[0]**o+(-0.253806844569)*x[0]+(-0.0688746199637)*x[1]**o+(0.232840317719)*x[1]+(-0.815623721026)*x[2]**o+(0.43055485562)*x[2]
        arg[2,4,1,0]=(-0.452320341493)*x[0]**o+(0.755240168228)*x[0]+(-0.892552611066)*x[1]**o+(0.438275161458)*x[1]+(0.441378369388)*x[2]**o+(-0.870501114753)*x[2]
        arg[2,4,1,1]=(0.496731258286)*x[0]**o+(-0.499918335322)*x[0]+(0.000476061942027)*x[1]**o+(-0.857762621856)*x[1]+(0.957433798133)*x[2]**o+(-0.766450629361)*x[2]
        arg[2,4,2,0]=(0.860379598562)*x[0]**o+(-0.0816478389485)*x[0]+(-0.438181864676)*x[1]**o+(0.78937427676)*x[1]+(-0.97940010041)*x[2]**o+(0.583914605259)*x[2]
        arg[2,4,2,1]=(0.23729037277)*x[0]**o+(-0.365380141474)*x[0]+(-0.471505632575)*x[1]**o+(-0.168838288)*x[1]+(-0.519847065791)*x[2]**o+(-0.816100518668)*x[2]
        arg[3,0,0,0]=(0.729921808661)*x[0]**o+(0.732868255388)*x[0]+(0.305434623795)*x[1]**o+(-0.45402804883)*x[1]+(0.604368334171)*x[2]**o+(-0.179233711008)*x[2]
        arg[3,0,0,1]=(-0.208916687176)*x[0]**o+(-0.297916049539)*x[0]+(0.854941564423)*x[1]**o+(0.775355543915)*x[1]+(-0.373382357688)*x[2]**o+(0.792557314074)*x[2]
        arg[3,0,1,0]=(0.606041025674)*x[0]**o+(0.115814644681)*x[0]+(-0.414470914985)*x[1]**o+(-0.136577390174)*x[1]+(0.672303216396)*x[2]**o+(-0.479680752765)*x[2]
        arg[3,0,1,1]=(-0.535633649471)*x[0]**o+(0.133761944775)*x[0]+(-0.980107507665)*x[1]**o+(0.211561747114)*x[1]+(-0.530648770004)*x[2]**o+(0.417977629512)*x[2]
        arg[3,0,2,0]=(-0.242797826282)*x[0]**o+(-0.828691638823)*x[0]+(-0.187401612749)*x[1]**o+(-0.453821620107)*x[1]+(-0.475718408605)*x[2]**o+(0.683502702497)*x[2]
        arg[3,0,2,1]=(-0.66640812951)*x[0]**o+(-0.259184719546)*x[0]+(0.463743732654)*x[1]**o+(0.6746349156)*x[1]+(-0.0532918236783)*x[2]**o+(-0.714197225876)*x[2]
        arg[3,1,0,0]=(-0.97693616823)*x[0]**o+(-0.825560555757)*x[0]+(0.118452690664)*x[1]**o+(-0.125279874388)*x[1]+(0.733803136335)*x[2]**o+(-0.884902209956)*x[2]
        arg[3,1,0,1]=(-0.652466190334)*x[0]**o+(0.370106640243)*x[0]+(0.0850861818306)*x[1]**o+(-0.49744492012)*x[1]+(-0.843498619618)*x[2]**o+(-0.0480070808698)*x[2]
        arg[3,1,1,0]=(-0.450526704276)*x[0]**o+(0.430774951776)*x[0]+(-0.956799540696)*x[1]**o+(0.820090546747)*x[1]+(0.831582198997)*x[2]**o+(0.491042906394)*x[2]
        arg[3,1,1,1]=(0.419717807665)*x[0]**o+(0.320904534666)*x[0]+(-0.483247938999)*x[1]**o+(0.82174341564)*x[1]+(-0.441719998974)*x[2]**o+(0.691713322239)*x[2]
        arg[3,1,2,0]=(0.255464093123)*x[0]**o+(-0.256334953799)*x[0]+(0.677105284128)*x[1]**o+(-0.584438025212)*x[1]+(0.505820591597)*x[2]**o+(0.247943562937)*x[2]
        arg[3,1,2,1]=(-0.495204107745)*x[0]**o+(0.234426349601)*x[0]+(0.736969108189)*x[1]**o+(-0.369684409765)*x[1]+(0.164303379282)*x[2]**o+(-0.595757590519)*x[2]
        arg[3,2,0,0]=(0.238639684591)*x[0]**o+(0.568638143194)*x[0]+(0.972646679042)*x[1]**o+(-0.652648194795)*x[1]+(-0.0530713046414)*x[2]**o+(0.356043665396)*x[2]
        arg[3,2,0,1]=(0.458539145106)*x[0]**o+(0.111417037398)*x[0]+(-0.135374868193)*x[1]**o+(-0.535451685997)*x[1]+(-0.820797762877)*x[2]**o+(0.122239787933)*x[2]
        arg[3,2,1,0]=(-0.323251425039)*x[0]**o+(0.106998797206)*x[0]+(-0.0671655760766)*x[1]**o+(0.836342794315)*x[1]+(0.34628715493)*x[2]**o+(-0.11058327806)*x[2]
        arg[3,2,1,1]=(0.181994203911)*x[0]**o+(-0.710678453639)*x[0]+(0.625606409786)*x[1]**o+(-0.363170328543)*x[1]+(-0.932998745422)*x[2]**o+(-0.0411581284314)*x[2]
        arg[3,2,2,0]=(0.710847104903)*x[0]**o+(0.483971915285)*x[0]+(0.42996482533)*x[1]**o+(0.456817348253)*x[1]+(0.844562058604)*x[2]**o+(-0.469586014721)*x[2]
        arg[3,2,2,1]=(-0.845771280837)*x[0]**o+(-0.081384696494)*x[0]+(0.71823377714)*x[1]**o+(-0.846672143168)*x[1]+(-0.139540918461)*x[2]**o+(0.837646131987)*x[2]
        arg[3,3,0,0]=(-0.958726792109)*x[0]**o+(0.53196455268)*x[0]+(-0.621810423658)*x[1]**o+(-0.0352106197658)*x[1]+(-0.636059207474)*x[2]**o+(-0.0491118207063)*x[2]
        arg[3,3,0,1]=(0.987836799279)*x[0]**o+(0.41633576224)*x[0]+(0.312021799176)*x[1]**o+(-0.168836582938)*x[1]+(-0.391725003735)*x[2]**o+(-0.158000182269)*x[2]
        arg[3,3,1,0]=(-0.280868116529)*x[0]**o+(-0.687924000092)*x[0]+(0.865840229722)*x[1]**o+(-0.0823203022874)*x[1]+(-0.857309899784)*x[2]**o+(-0.0657473493059)*x[2]
        arg[3,3,1,1]=(0.680496030567)*x[0]**o+(0.159575720724)*x[0]+(0.0531895122046)*x[1]**o+(0.359273896084)*x[1]+(0.35084774502)*x[2]**o+(0.0982220259531)*x[2]
        arg[3,3,2,0]=(0.140872311111)*x[0]**o+(-0.00845309046907)*x[0]+(-0.845527142503)*x[1]**o+(-0.67713103521)*x[1]+(0.972054187828)*x[2]**o+(-0.800664963509)*x[2]
        arg[3,3,2,1]=(-0.490866666782)*x[0]**o+(-0.0988939523952)*x[0]+(0.236210969508)*x[1]**o+(-0.672617589992)*x[1]+(-0.640558724403)*x[2]**o+(-0.203125267713)*x[2]
        arg[3,4,0,0]=(0.318438839683)*x[0]**o+(0.0864322377812)*x[0]+(-0.371419701638)*x[1]**o+(-0.311583139142)*x[1]+(-0.0533111244011)*x[2]**o+(0.841240973423)*x[2]
        arg[3,4,0,1]=(-0.573110771334)*x[0]**o+(0.703852077914)*x[0]+(0.613309217217)*x[1]**o+(0.247223193846)*x[1]+(0.618952485205)*x[2]**o+(-0.395998308184)*x[2]
        arg[3,4,1,0]=(0.512896938934)*x[0]**o+(0.848653171505)*x[0]+(-0.944903687062)*x[1]**o+(0.179782177607)*x[1]+(-0.0287288996538)*x[2]**o+(-0.385519118293)*x[2]
        arg[3,4,1,1]=(0.234190647718)*x[0]**o+(0.445478221419)*x[0]+(0.588931676434)*x[1]**o+(0.140788139998)*x[1]+(-0.761952359657)*x[2]**o+(0.92377426359)*x[2]
        arg[3,4,2,0]=(-0.836038966171)*x[0]**o+(-0.772990045933)*x[0]+(0.0622128209877)*x[1]**o+(-0.73733496141)*x[1]+(0.0234446925161)*x[2]**o+(-0.682234213315)*x[2]
        arg[3,4,2,1]=(0.250498084214)*x[0]**o+(0.539088356159)*x[0]+(-0.691842680176)*x[1]**o+(0.710040802953)*x[1]+(-0.43691551501)*x[2]**o+(0.710668600989)*x[2]
        ref[0,0,0,0]=(-0.836302395465)*x_ref[0]**o+(-0.842568675604)*x_ref[0]+(-0.780133441207)*x_ref[1]**o+(0.54483947113)*x_ref[1]+(-0.130253019696)*x_ref[2]**o+(0.358303389828)*x_ref[2]
        ref[0,0,0,1]=(-0.722847500652)*x_ref[0]**o+(0.42887368878)*x_ref[0]+(0.979187056933)*x_ref[1]**o+(0.624875714982)*x_ref[1]+(-0.672529765247)*x_ref[2]**o+(0.822363583702)*x_ref[2]
        ref[0,0,1,0]=(-0.28989650691)*x_ref[0]**o+(-0.847647954773)*x_ref[0]+(-0.11119862113)*x_ref[1]**o+(0.0809009676818)*x_ref[1]+(-0.672626533587)*x_ref[2]**o+(0.580561232267)*x_ref[2]
        ref[0,0,1,1]=(-0.64783363248)*x_ref[0]**o+(0.212828552873)*x_ref[0]+(-0.845297180936)*x_ref[1]**o+(-0.714760411186)*x_ref[1]+(-0.385173605122)*x_ref[2]**o+(-0.760750317556)*x_ref[2]
        ref[0,0,2,0]=(-0.328757006046)*x_ref[0]**o+(-0.0935907722247)*x_ref[0]+(-0.895866135795)*x_ref[1]**o+(-0.701241976745)*x_ref[1]+(0.893684139941)*x_ref[2]**o+(-0.265339318222)*x_ref[2]
        ref[0,0,2,1]=(0.134066662346)*x_ref[0]**o+(0.0383231508856)*x_ref[0]+(-0.347744261995)*x_ref[1]**o+(-0.86897907336)*x_ref[1]+(-0.287792154521)*x_ref[2]**o+(0.859351846891)*x_ref[2]
        ref[0,1,0,0]=(-0.635128124936)*x_ref[0]**o+(-0.659462189733)*x_ref[0]+(-0.0964365747936)*x_ref[1]**o+(-0.385702168747)*x_ref[1]+(0.179539914057)*x_ref[2]**o+(-0.0176420847859)*x_ref[2]
        ref[0,1,0,1]=(-0.580750807831)*x_ref[0]**o+(0.756493147583)*x_ref[0]+(-0.436842251289)*x_ref[1]**o+(-0.705624646095)*x_ref[1]+(0.983064096262)*x_ref[2]**o+(-0.0197153825423)*x_ref[2]
        ref[0,1,1,0]=(-0.567748582365)*x_ref[0]**o+(0.579324420356)*x_ref[0]+(0.0434168493419)*x_ref[1]**o+(-0.839692920334)*x_ref[1]+(-0.119239847745)*x_ref[2]**o+(0.527047942834)*x_ref[2]
        ref[0,1,1,1]=(0.52870403655)*x_ref[0]**o+(0.700090892471)*x_ref[0]+(0.730224313921)*x_ref[1]**o+(-0.815451751207)*x_ref[1]+(-0.943945382413)*x_ref[2]**o+(-0.571714141916)*x_ref[2]
        ref[0,1,2,0]=(-0.803841094887)*x_ref[0]**o+(0.0595126006237)*x_ref[0]+(-0.612341797645)*x_ref[1]**o+(0.944552805211)*x_ref[1]+(-0.00253009731786)*x_ref[2]**o+(0.0174867439912)*x_ref[2]
        ref[0,1,2,1]=(-0.145623212549)*x_ref[0]**o+(0.546039957008)*x_ref[0]+(-0.221425237579)*x_ref[1]**o+(0.229367671159)*x_ref[1]+(-0.784903869759)*x_ref[2]**o+(0.898045208871)*x_ref[2]
        ref[0,2,0,0]=(0.417924494005)*x_ref[0]**o+(-0.192820051987)*x_ref[0]+(0.878916271492)*x_ref[1]**o+(-0.35056320984)*x_ref[1]+(0.982897575604)*x_ref[2]**o+(-0.225793730616)*x_ref[2]
        ref[0,2,0,1]=(-0.493279498849)*x_ref[0]**o+(-0.154177411087)*x_ref[0]+(-0.923592860492)*x_ref[1]**o+(0.872010949252)*x_ref[1]+(-0.936252147064)*x_ref[2]**o+(0.947136819308)*x_ref[2]
        ref[0,2,1,0]=(0.109357630607)*x_ref[0]**o+(-0.518825477709)*x_ref[0]+(-0.375510667498)*x_ref[1]**o+(0.476537999644)*x_ref[1]+(0.676017421029)*x_ref[2]**o+(-0.633301158984)*x_ref[2]
        ref[0,2,1,1]=(-0.0662265225539)*x_ref[0]**o+(-0.276634748263)*x_ref[0]+(-0.0655431981102)*x_ref[1]**o+(0.0578482920991)*x_ref[1]+(-0.264370611185)*x_ref[2]**o+(-0.674612985594)*x_ref[2]
        ref[0,2,2,0]=(-0.924186592743)*x_ref[0]**o+(0.197603043149)*x_ref[0]+(-0.676882805046)*x_ref[1]**o+(0.564142845865)*x_ref[1]+(0.590369345306)*x_ref[2]**o+(-0.872019490978)*x_ref[2]
        ref[0,2,2,1]=(0.480683776968)*x_ref[0]**o+(-0.412055284173)*x_ref[0]+(0.829996313473)*x_ref[1]**o+(-0.235665473048)*x_ref[1]+(-0.429191305284)*x_ref[2]**o+(-0.0867804318457)*x_ref[2]
        ref[0,3,0,0]=(-0.0903660006043)*x_ref[0]**o+(-0.507571901766)*x_ref[0]+(-0.0305305861809)*x_ref[1]**o+(-0.915027969989)*x_ref[1]+(0.955644774347)*x_ref[2]**o+(0.739244618775)*x_ref[2]
        ref[0,3,0,1]=(0.600619122998)*x_ref[0]**o+(0.0128699484527)*x_ref[0]+(-0.360431921281)*x_ref[1]**o+(0.304773619811)*x_ref[1]+(-0.302423314054)*x_ref[2]**o+(0.680149401572)*x_ref[2]
        ref[0,3,1,0]=(-0.638540513323)*x_ref[0]**o+(0.766655754067)*x_ref[0]+(0.276879854223)*x_ref[1]**o+(-0.628688275247)*x_ref[1]+(0.724584107114)*x_ref[2]**o+(0.682193551967)*x_ref[2]
        ref[0,3,1,1]=(-0.516656801349)*x_ref[0]**o+(-0.513901319383)*x_ref[0]+(-0.227957010777)*x_ref[1]**o+(-0.164363207703)*x_ref[1]+(-0.919936042047)*x_ref[2]**o+(0.682400184997)*x_ref[2]
        ref[0,3,2,0]=(-0.819629653095)*x_ref[0]**o+(0.183937049274)*x_ref[0]+(0.0411075800579)*x_ref[1]**o+(0.0813088270699)*x_ref[1]+(0.762416334575)*x_ref[2]**o+(0.236553726705)*x_ref[2]
        ref[0,3,2,1]=(0.513141183759)*x_ref[0]**o+(-0.452475512734)*x_ref[0]+(0.0461424221984)*x_ref[1]**o+(-0.673452844636)*x_ref[1]+(-0.194493328878)*x_ref[2]**o+(0.812459956422)*x_ref[2]
        ref[0,4,0,0]=(0.672167415195)*x_ref[0]**o+(-0.641004730405)*x_ref[0]+(-0.383031420851)*x_ref[1]**o+(0.392366431186)*x_ref[1]+(0.66479124224)*x_ref[2]**o+(0.845503225104)*x_ref[2]
        ref[0,4,0,1]=(0.908241966695)*x_ref[0]**o+(-0.816989438726)*x_ref[0]+(-0.144902267583)*x_ref[1]**o+(-0.654901540598)*x_ref[1]+(-0.9840470269)*x_ref[2]**o+(0.510970695154)*x_ref[2]
        ref[0,4,1,0]=(0.788858704366)*x_ref[0]**o+(-0.50266306504)*x_ref[0]+(-0.681248796723)*x_ref[1]**o+(0.0317663112509)*x_ref[1]+(0.662428035619)*x_ref[2]**o+(0.377116420912)*x_ref[2]
        ref[0,4,1,1]=(-0.0734522829037)*x_ref[0]**o+(0.0822741484655)*x_ref[0]+(-0.509279578815)*x_ref[1]**o+(-0.95372420098)*x_ref[1]+(0.682868362508)*x_ref[2]**o+(-0.947156883638)*x_ref[2]
        ref[0,4,2,0]=(-0.843236607337)*x_ref[0]**o+(-0.925326238513)*x_ref[0]+(-0.153688519983)*x_ref[1]**o+(-0.68456413487)*x_ref[1]+(0.920442617368)*x_ref[2]**o+(0.852945557893)*x_ref[2]
        ref[0,4,2,1]=(-0.283772512528)*x_ref[0]**o+(0.202730738521)*x_ref[0]+(-0.287375172356)*x_ref[1]**o+(-0.176699132816)*x_ref[1]+(0.446391113264)*x_ref[2]**o+(-0.783266847348)*x_ref[2]
        ref[1,0,0,0]=(0.81284364562)*x_ref[0]**o+(-0.454432332505)*x_ref[0]+(-0.684692596262)*x_ref[1]**o+(-0.345923888237)*x_ref[1]+(0.1201728631)*x_ref[2]**o+(0.808246135028)*x_ref[2]
        ref[1,0,0,1]=(0.941104328706)*x_ref[0]**o+(-0.802008654964)*x_ref[0]+(0.902047483677)*x_ref[1]**o+(0.777049224612)*x_ref[1]+(0.120184332854)*x_ref[2]**o+(0.983216899102)*x_ref[2]
        ref[1,0,1,0]=(-0.420996925767)*x_ref[0]**o+(0.0941750615502)*x_ref[0]+(-0.935342933578)*x_ref[1]**o+(0.925891822058)*x_ref[1]+(-0.153652985582)*x_ref[2]**o+(-0.336548986153)*x_ref[2]
        ref[1,0,1,1]=(-0.692886221795)*x_ref[0]**o+(-0.920419410139)*x_ref[0]+(-0.586575583263)*x_ref[1]**o+(-0.269347362009)*x_ref[1]+(-0.872169736045)*x_ref[2]**o+(-0.724727186336)*x_ref[2]
        ref[1,0,2,0]=(-0.0171222282058)*x_ref[0]**o+(0.338180602976)*x_ref[0]+(-0.0345958505395)*x_ref[1]**o+(0.631728421505)*x_ref[1]+(0.90596443453)*x_ref[2]**o+(-0.144785193015)*x_ref[2]
        ref[1,0,2,1]=(-0.91636583885)*x_ref[0]**o+(-0.807496789574)*x_ref[0]+(0.63472495957)*x_ref[1]**o+(0.966464343993)*x_ref[1]+(-0.471311384621)*x_ref[2]**o+(-0.449420138238)*x_ref[2]
        ref[1,1,0,0]=(-0.239930503474)*x_ref[0]**o+(-0.476439715994)*x_ref[0]+(0.421600511888)*x_ref[1]**o+(0.477823144792)*x_ref[1]+(0.251409914807)*x_ref[2]**o+(0.636818001419)*x_ref[2]
        ref[1,1,0,1]=(-0.427510410263)*x_ref[0]**o+(0.494810857516)*x_ref[0]+(-0.129415181352)*x_ref[1]**o+(0.754147770071)*x_ref[1]+(0.0218776365985)*x_ref[2]**o+(0.706183216616)*x_ref[2]
        ref[1,1,1,0]=(-0.469150375165)*x_ref[0]**o+(-0.0651101638073)*x_ref[0]+(-0.697853003861)*x_ref[1]**o+(0.0389330102026)*x_ref[1]+(-0.256557942045)*x_ref[2]**o+(0.409565340425)*x_ref[2]
        ref[1,1,1,1]=(0.14957005608)*x_ref[0]**o+(-0.209482968126)*x_ref[0]+(0.368899650309)*x_ref[1]**o+(-0.0899321000137)*x_ref[1]+(0.282389903357)*x_ref[2]**o+(-0.166633883916)*x_ref[2]
        ref[1,1,2,0]=(0.0374938231714)*x_ref[0]**o+(-0.687972899893)*x_ref[0]+(-0.0686478107779)*x_ref[1]**o+(0.318977937888)*x_ref[1]+(-0.601269126788)*x_ref[2]**o+(0.453188855786)*x_ref[2]
        ref[1,1,2,1]=(-0.853894223222)*x_ref[0]**o+(0.441830151478)*x_ref[0]+(0.973114898368)*x_ref[1]**o+(-0.264454889071)*x_ref[1]+(-0.828332687531)*x_ref[2]**o+(0.302493748565)*x_ref[2]
        ref[1,2,0,0]=(0.210365824111)*x_ref[0]**o+(0.4674117857)*x_ref[0]+(-0.129546969044)*x_ref[1]**o+(0.566357125352)*x_ref[1]+(-0.721733448514)*x_ref[2]**o+(0.154253207106)*x_ref[2]
        ref[1,2,0,1]=(0.805325508181)*x_ref[0]**o+(-0.929113563452)*x_ref[0]+(-0.487301761321)*x_ref[1]**o+(-0.822405526519)*x_ref[1]+(0.254754490326)*x_ref[2]**o+(-0.196480710709)*x_ref[2]
        ref[1,2,1,0]=(-0.520777275475)*x_ref[0]**o+(-0.912569745644)*x_ref[0]+(-0.784504791087)*x_ref[1]**o+(-0.903473217653)*x_ref[1]+(0.583297107374)*x_ref[2]**o+(0.664811898195)*x_ref[2]
        ref[1,2,1,1]=(0.129152152958)*x_ref[0]**o+(0.497832200867)*x_ref[0]+(0.121217248234)*x_ref[1]**o+(-0.183368066714)*x_ref[1]+(-0.776145485536)*x_ref[2]**o+(-0.223165724715)*x_ref[2]
        ref[1,2,2,0]=(-0.698981324488)*x_ref[0]**o+(0.00951277777924)*x_ref[0]+(0.997410976507)*x_ref[1]**o+(0.534670303453)*x_ref[1]+(0.251309223218)*x_ref[2]**o+(-0.921148646979)*x_ref[2]
        ref[1,2,2,1]=(-0.731509934487)*x_ref[0]**o+(0.87731926834)*x_ref[0]+(0.0865507713913)*x_ref[1]**o+(0.708835987019)*x_ref[1]+(0.294182269171)*x_ref[2]**o+(-0.316660750729)*x_ref[2]
        ref[1,3,0,0]=(-0.282903962918)*x_ref[0]**o+(-0.793457978844)*x_ref[0]+(0.287425807914)*x_ref[1]**o+(-0.0252722364835)*x_ref[1]+(-0.956202022515)*x_ref[2]**o+(0.650271579135)*x_ref[2]
        ref[1,3,0,1]=(-0.946723111202)*x_ref[0]**o+(0.669431653693)*x_ref[0]+(-0.791125976188)*x_ref[1]**o+(-0.466484914794)*x_ref[1]+(-0.217225801872)*x_ref[2]**o+(0.224719473996)*x_ref[2]
        ref[1,3,1,0]=(0.516479301043)*x_ref[0]**o+(-0.42420332366)*x_ref[0]+(-0.0649227247176)*x_ref[1]**o+(0.208849752675)*x_ref[1]+(-0.939001099693)*x_ref[2]**o+(-0.0537827140932)*x_ref[2]
        ref[1,3,1,1]=(-0.292077069459)*x_ref[0]**o+(-0.444848119744)*x_ref[0]+(0.268592719199)*x_ref[1]**o+(-0.349401060323)*x_ref[1]+(-0.540803784489)*x_ref[2]**o+(-0.487170123526)*x_ref[2]
        ref[1,3,2,0]=(-0.626222042661)*x_ref[0]**o+(0.245937409624)*x_ref[0]+(-0.24692162006)*x_ref[1]**o+(0.108230029208)*x_ref[1]+(-0.760463241928)*x_ref[2]**o+(-0.587078263201)*x_ref[2]
        ref[1,3,2,1]=(-0.923891147756)*x_ref[0]**o+(0.0961285842162)*x_ref[0]+(0.0146853466585)*x_ref[1]**o+(0.00951339122359)*x_ref[1]+(-0.261383704723)*x_ref[2]**o+(-0.0262515691324)*x_ref[2]
        ref[1,4,0,0]=(-0.0370003524292)*x_ref[0]**o+(-0.0463848489957)*x_ref[0]+(0.940061458694)*x_ref[1]**o+(0.592787058925)*x_ref[1]+(-0.217284379638)*x_ref[2]**o+(-0.11746834109)*x_ref[2]
        ref[1,4,0,1]=(0.830404283093)*x_ref[0]**o+(0.268135490403)*x_ref[0]+(-0.894500398195)*x_ref[1]**o+(0.751042648917)*x_ref[1]+(0.687583292754)*x_ref[2]**o+(-0.884596509742)*x_ref[2]
        ref[1,4,1,0]=(0.414764592469)*x_ref[0]**o+(-0.529613347919)*x_ref[0]+(-0.727435115594)*x_ref[1]**o+(-0.546579777725)*x_ref[1]+(0.378011893971)*x_ref[2]**o+(-0.578081183238)*x_ref[2]
        ref[1,4,1,1]=(0.406513376747)*x_ref[0]**o+(-0.350132869579)*x_ref[0]+(-0.557954631299)*x_ref[1]**o+(0.868176175754)*x_ref[1]+(0.793021832632)*x_ref[2]**o+(0.280529771851)*x_ref[2]
        ref[1,4,2,0]=(0.56741953575)*x_ref[0]**o+(0.0115947684302)*x_ref[0]+(0.872812725363)*x_ref[1]**o+(-0.215183541745)*x_ref[1]+(0.64853692203)*x_ref[2]**o+(-0.139561471922)*x_ref[2]
        ref[1,4,2,1]=(0.650916904666)*x_ref[0]**o+(0.878765271839)*x_ref[0]+(0.0141264348052)*x_ref[1]**o+(-0.56314426774)*x_ref[1]+(-0.267854377704)*x_ref[2]**o+(0.852271380667)*x_ref[2]
        ref[2,0,0,0]=(0.948359141577)*x_ref[0]**o+(-0.0881747280974)*x_ref[0]+(0.804871563368)*x_ref[1]**o+(-0.0298585763446)*x_ref[1]+(0.506923333569)*x_ref[2]**o+(-0.920947093648)*x_ref[2]
        ref[2,0,0,1]=(-0.0144221258584)*x_ref[0]**o+(-0.679723072824)*x_ref[0]+(0.00829336956687)*x_ref[1]**o+(0.675680809303)*x_ref[1]+(-0.803195231536)*x_ref[2]**o+(0.70240358571)*x_ref[2]
        ref[2,0,1,0]=(-0.159174987706)*x_ref[0]**o+(0.922309271483)*x_ref[0]+(0.344354343035)*x_ref[1]**o+(-0.00954038311323)*x_ref[1]+(-0.50225452804)*x_ref[2]**o+(0.355899734308)*x_ref[2]
        ref[2,0,1,1]=(0.0857611780404)*x_ref[0]**o+(0.876479431983)*x_ref[0]+(0.105049158502)*x_ref[1]**o+(-0.530308461426)*x_ref[1]+(-0.553136649343)*x_ref[2]**o+(0.385458516481)*x_ref[2]
        ref[2,0,2,0]=(-0.260086070842)*x_ref[0]**o+(0.417766241499)*x_ref[0]+(-0.292370383673)*x_ref[1]**o+(-0.321315349553)*x_ref[1]+(-0.437301444082)*x_ref[2]**o+(0.730845669715)*x_ref[2]
        ref[2,0,2,1]=(0.132547674276)*x_ref[0]**o+(0.783602886333)*x_ref[0]+(0.822361450413)*x_ref[1]**o+(-0.810402783)*x_ref[1]+(0.759701191023)*x_ref[2]**o+(0.644375788608)*x_ref[2]
        ref[2,1,0,0]=(-0.841458111227)*x_ref[0]**o+(0.779917641736)*x_ref[0]+(-0.0745587217256)*x_ref[1]**o+(0.938545280143)*x_ref[1]+(0.454600163373)*x_ref[2]**o+(-0.490348377697)*x_ref[2]
        ref[2,1,0,1]=(-0.876477136174)*x_ref[0]**o+(-0.248200369939)*x_ref[0]+(-0.973654243816)*x_ref[1]**o+(0.482402123794)*x_ref[1]+(-0.776970490827)*x_ref[2]**o+(-0.571349467703)*x_ref[2]
        ref[2,1,1,0]=(-0.433528193102)*x_ref[0]**o+(-0.57712915767)*x_ref[0]+(-0.328671198058)*x_ref[1]**o+(-0.926873019075)*x_ref[1]+(0.314232739356)*x_ref[2]**o+(-0.199625466412)*x_ref[2]
        ref[2,1,1,1]=(0.892707770392)*x_ref[0]**o+(-0.680081898234)*x_ref[0]+(-0.754181058483)*x_ref[1]**o+(-0.253728945575)*x_ref[1]+(-0.906373406684)*x_ref[2]**o+(-0.0266435080833)*x_ref[2]
        ref[2,1,2,0]=(-0.121605378472)*x_ref[0]**o+(-0.503955015634)*x_ref[0]+(-0.0960767196114)*x_ref[1]**o+(-0.937914890361)*x_ref[1]+(0.212042931853)*x_ref[2]**o+(0.930703155408)*x_ref[2]
        ref[2,1,2,1]=(0.0564159441047)*x_ref[0]**o+(0.842332721105)*x_ref[0]+(0.954741657608)*x_ref[1]**o+(0.73503176)*x_ref[1]+(-0.390679938582)*x_ref[2]**o+(-0.440014747741)*x_ref[2]
        ref[2,2,0,0]=(0.0100637970821)*x_ref[0]**o+(-0.0283353154145)*x_ref[0]+(0.865360758262)*x_ref[1]**o+(0.780975027988)*x_ref[1]+(0.16115842919)*x_ref[2]**o+(0.247580155656)*x_ref[2]
        ref[2,2,0,1]=(0.61102059426)*x_ref[0]**o+(-0.198005744101)*x_ref[0]+(0.503568797262)*x_ref[1]**o+(0.0288809726497)*x_ref[1]+(-0.765127098183)*x_ref[2]**o+(0.409803857723)*x_ref[2]
        ref[2,2,1,0]=(-0.712272018295)*x_ref[0]**o+(-0.62562142063)*x_ref[0]+(-0.579501949719)*x_ref[1]**o+(0.628539275236)*x_ref[1]+(0.0200945124964)*x_ref[2]**o+(0.175966024769)*x_ref[2]
        ref[2,2,1,1]=(-0.798244445519)*x_ref[0]**o+(-0.376440962535)*x_ref[0]+(-0.134467401497)*x_ref[1]**o+(0.877538294477)*x_ref[1]+(-0.484870741192)*x_ref[2]**o+(-0.224064469935)*x_ref[2]
        ref[2,2,2,0]=(0.200196926368)*x_ref[0]**o+(-0.528000867086)*x_ref[0]+(-0.708942342624)*x_ref[1]**o+(-0.591994969141)*x_ref[1]+(-0.990641592656)*x_ref[2]**o+(-0.875897347303)*x_ref[2]
        ref[2,2,2,1]=(0.949113190797)*x_ref[0]**o+(0.867855496786)*x_ref[0]+(-0.0314953438065)*x_ref[1]**o+(-0.371928428005)*x_ref[1]+(-0.526403191063)*x_ref[2]**o+(0.653049122326)*x_ref[2]
        ref[2,3,0,0]=(-0.353418327127)*x_ref[0]**o+(-0.418588306429)*x_ref[0]+(0.400872043269)*x_ref[1]**o+(0.141907924042)*x_ref[1]+(-0.999869579462)*x_ref[2]**o+(-0.410423954563)*x_ref[2]
        ref[2,3,0,1]=(-0.240390877307)*x_ref[0]**o+(0.0901879916115)*x_ref[0]+(-0.0734683209657)*x_ref[1]**o+(0.274836005886)*x_ref[1]+(-0.336375498181)*x_ref[2]**o+(-0.754776455221)*x_ref[2]
        ref[2,3,1,0]=(0.259928460822)*x_ref[0]**o+(-0.227630454708)*x_ref[0]+(0.725968066278)*x_ref[1]**o+(-0.545440695763)*x_ref[1]+(-0.0354926607815)*x_ref[2]**o+(0.365010532181)*x_ref[2]
        ref[2,3,1,1]=(-0.969799373398)*x_ref[0]**o+(0.514498233214)*x_ref[0]+(0.907450925318)*x_ref[1]**o+(-0.234993706406)*x_ref[1]+(0.206323910806)*x_ref[2]**o+(0.209364250369)*x_ref[2]
        ref[2,3,2,0]=(0.712905565941)*x_ref[0]**o+(-0.353322758263)*x_ref[0]+(0.816861181494)*x_ref[1]**o+(-0.362388243619)*x_ref[1]+(0.302543517607)*x_ref[2]**o+(-0.831280418852)*x_ref[2]
        ref[2,3,2,1]=(-0.386373711182)*x_ref[0]**o+(0.946914567792)*x_ref[0]+(-0.52795085887)*x_ref[1]**o+(0.86635245964)*x_ref[1]+(0.654261054178)*x_ref[2]**o+(0.452334882498)*x_ref[2]
        ref[2,4,0,0]=(0.374747033547)*x_ref[0]**o+(-0.3471578669)*x_ref[0]+(0.322217002639)*x_ref[1]**o+(0.92988068982)*x_ref[1]+(-0.642423472369)*x_ref[2]**o+(-0.759623225653)*x_ref[2]
        ref[2,4,0,1]=(-0.322255157841)*x_ref[0]**o+(-0.253806844569)*x_ref[0]+(-0.0688746199637)*x_ref[1]**o+(0.232840317719)*x_ref[1]+(-0.815623721026)*x_ref[2]**o+(0.43055485562)*x_ref[2]
        ref[2,4,1,0]=(-0.452320341493)*x_ref[0]**o+(0.755240168228)*x_ref[0]+(-0.892552611066)*x_ref[1]**o+(0.438275161458)*x_ref[1]+(0.441378369388)*x_ref[2]**o+(-0.870501114753)*x_ref[2]
        ref[2,4,1,1]=(0.496731258286)*x_ref[0]**o+(-0.499918335322)*x_ref[0]+(0.000476061942027)*x_ref[1]**o+(-0.857762621856)*x_ref[1]+(0.957433798133)*x_ref[2]**o+(-0.766450629361)*x_ref[2]
        ref[2,4,2,0]=(0.860379598562)*x_ref[0]**o+(-0.0816478389485)*x_ref[0]+(-0.438181864676)*x_ref[1]**o+(0.78937427676)*x_ref[1]+(-0.97940010041)*x_ref[2]**o+(0.583914605259)*x_ref[2]
        ref[2,4,2,1]=(0.23729037277)*x_ref[0]**o+(-0.365380141474)*x_ref[0]+(-0.471505632575)*x_ref[1]**o+(-0.168838288)*x_ref[1]+(-0.519847065791)*x_ref[2]**o+(-0.816100518668)*x_ref[2]
        ref[3,0,0,0]=(0.729921808661)*x_ref[0]**o+(0.732868255388)*x_ref[0]+(0.305434623795)*x_ref[1]**o+(-0.45402804883)*x_ref[1]+(0.604368334171)*x_ref[2]**o+(-0.179233711008)*x_ref[2]
        ref[3,0,0,1]=(-0.208916687176)*x_ref[0]**o+(-0.297916049539)*x_ref[0]+(0.854941564423)*x_ref[1]**o+(0.775355543915)*x_ref[1]+(-0.373382357688)*x_ref[2]**o+(0.792557314074)*x_ref[2]
        ref[3,0,1,0]=(0.606041025674)*x_ref[0]**o+(0.115814644681)*x_ref[0]+(-0.414470914985)*x_ref[1]**o+(-0.136577390174)*x_ref[1]+(0.672303216396)*x_ref[2]**o+(-0.479680752765)*x_ref[2]
        ref[3,0,1,1]=(-0.535633649471)*x_ref[0]**o+(0.133761944775)*x_ref[0]+(-0.980107507665)*x_ref[1]**o+(0.211561747114)*x_ref[1]+(-0.530648770004)*x_ref[2]**o+(0.417977629512)*x_ref[2]
        ref[3,0,2,0]=(-0.242797826282)*x_ref[0]**o+(-0.828691638823)*x_ref[0]+(-0.187401612749)*x_ref[1]**o+(-0.453821620107)*x_ref[1]+(-0.475718408605)*x_ref[2]**o+(0.683502702497)*x_ref[2]
        ref[3,0,2,1]=(-0.66640812951)*x_ref[0]**o+(-0.259184719546)*x_ref[0]+(0.463743732654)*x_ref[1]**o+(0.6746349156)*x_ref[1]+(-0.0532918236783)*x_ref[2]**o+(-0.714197225876)*x_ref[2]
        ref[3,1,0,0]=(-0.97693616823)*x_ref[0]**o+(-0.825560555757)*x_ref[0]+(0.118452690664)*x_ref[1]**o+(-0.125279874388)*x_ref[1]+(0.733803136335)*x_ref[2]**o+(-0.884902209956)*x_ref[2]
        ref[3,1,0,1]=(-0.652466190334)*x_ref[0]**o+(0.370106640243)*x_ref[0]+(0.0850861818306)*x_ref[1]**o+(-0.49744492012)*x_ref[1]+(-0.843498619618)*x_ref[2]**o+(-0.0480070808698)*x_ref[2]
        ref[3,1,1,0]=(-0.450526704276)*x_ref[0]**o+(0.430774951776)*x_ref[0]+(-0.956799540696)*x_ref[1]**o+(0.820090546747)*x_ref[1]+(0.831582198997)*x_ref[2]**o+(0.491042906394)*x_ref[2]
        ref[3,1,1,1]=(0.419717807665)*x_ref[0]**o+(0.320904534666)*x_ref[0]+(-0.483247938999)*x_ref[1]**o+(0.82174341564)*x_ref[1]+(-0.441719998974)*x_ref[2]**o+(0.691713322239)*x_ref[2]
        ref[3,1,2,0]=(0.255464093123)*x_ref[0]**o+(-0.256334953799)*x_ref[0]+(0.677105284128)*x_ref[1]**o+(-0.584438025212)*x_ref[1]+(0.505820591597)*x_ref[2]**o+(0.247943562937)*x_ref[2]
        ref[3,1,2,1]=(-0.495204107745)*x_ref[0]**o+(0.234426349601)*x_ref[0]+(0.736969108189)*x_ref[1]**o+(-0.369684409765)*x_ref[1]+(0.164303379282)*x_ref[2]**o+(-0.595757590519)*x_ref[2]
        ref[3,2,0,0]=(0.238639684591)*x_ref[0]**o+(0.568638143194)*x_ref[0]+(0.972646679042)*x_ref[1]**o+(-0.652648194795)*x_ref[1]+(-0.0530713046414)*x_ref[2]**o+(0.356043665396)*x_ref[2]
        ref[3,2,0,1]=(0.458539145106)*x_ref[0]**o+(0.111417037398)*x_ref[0]+(-0.135374868193)*x_ref[1]**o+(-0.535451685997)*x_ref[1]+(-0.820797762877)*x_ref[2]**o+(0.122239787933)*x_ref[2]
        ref[3,2,1,0]=(-0.323251425039)*x_ref[0]**o+(0.106998797206)*x_ref[0]+(-0.0671655760766)*x_ref[1]**o+(0.836342794315)*x_ref[1]+(0.34628715493)*x_ref[2]**o+(-0.11058327806)*x_ref[2]
        ref[3,2,1,1]=(0.181994203911)*x_ref[0]**o+(-0.710678453639)*x_ref[0]+(0.625606409786)*x_ref[1]**o+(-0.363170328543)*x_ref[1]+(-0.932998745422)*x_ref[2]**o+(-0.0411581284314)*x_ref[2]
        ref[3,2,2,0]=(0.710847104903)*x_ref[0]**o+(0.483971915285)*x_ref[0]+(0.42996482533)*x_ref[1]**o+(0.456817348253)*x_ref[1]+(0.844562058604)*x_ref[2]**o+(-0.469586014721)*x_ref[2]
        ref[3,2,2,1]=(-0.845771280837)*x_ref[0]**o+(-0.081384696494)*x_ref[0]+(0.71823377714)*x_ref[1]**o+(-0.846672143168)*x_ref[1]+(-0.139540918461)*x_ref[2]**o+(0.837646131987)*x_ref[2]
        ref[3,3,0,0]=(-0.958726792109)*x_ref[0]**o+(0.53196455268)*x_ref[0]+(-0.621810423658)*x_ref[1]**o+(-0.0352106197658)*x_ref[1]+(-0.636059207474)*x_ref[2]**o+(-0.0491118207063)*x_ref[2]
        ref[3,3,0,1]=(0.987836799279)*x_ref[0]**o+(0.41633576224)*x_ref[0]+(0.312021799176)*x_ref[1]**o+(-0.168836582938)*x_ref[1]+(-0.391725003735)*x_ref[2]**o+(-0.158000182269)*x_ref[2]
        ref[3,3,1,0]=(-0.280868116529)*x_ref[0]**o+(-0.687924000092)*x_ref[0]+(0.865840229722)*x_ref[1]**o+(-0.0823203022874)*x_ref[1]+(-0.857309899784)*x_ref[2]**o+(-0.0657473493059)*x_ref[2]
        ref[3,3,1,1]=(0.680496030567)*x_ref[0]**o+(0.159575720724)*x_ref[0]+(0.0531895122046)*x_ref[1]**o+(0.359273896084)*x_ref[1]+(0.35084774502)*x_ref[2]**o+(0.0982220259531)*x_ref[2]
        ref[3,3,2,0]=(0.140872311111)*x_ref[0]**o+(-0.00845309046907)*x_ref[0]+(-0.845527142503)*x_ref[1]**o+(-0.67713103521)*x_ref[1]+(0.972054187828)*x_ref[2]**o+(-0.800664963509)*x_ref[2]
        ref[3,3,2,1]=(-0.490866666782)*x_ref[0]**o+(-0.0988939523952)*x_ref[0]+(0.236210969508)*x_ref[1]**o+(-0.672617589992)*x_ref[1]+(-0.640558724403)*x_ref[2]**o+(-0.203125267713)*x_ref[2]
        ref[3,4,0,0]=(0.318438839683)*x_ref[0]**o+(0.0864322377812)*x_ref[0]+(-0.371419701638)*x_ref[1]**o+(-0.311583139142)*x_ref[1]+(-0.0533111244011)*x_ref[2]**o+(0.841240973423)*x_ref[2]
        ref[3,4,0,1]=(-0.573110771334)*x_ref[0]**o+(0.703852077914)*x_ref[0]+(0.613309217217)*x_ref[1]**o+(0.247223193846)*x_ref[1]+(0.618952485205)*x_ref[2]**o+(-0.395998308184)*x_ref[2]
        ref[3,4,1,0]=(0.512896938934)*x_ref[0]**o+(0.848653171505)*x_ref[0]+(-0.944903687062)*x_ref[1]**o+(0.179782177607)*x_ref[1]+(-0.0287288996538)*x_ref[2]**o+(-0.385519118293)*x_ref[2]
        ref[3,4,1,1]=(0.234190647718)*x_ref[0]**o+(0.445478221419)*x_ref[0]+(0.588931676434)*x_ref[1]**o+(0.140788139998)*x_ref[1]+(-0.761952359657)*x_ref[2]**o+(0.92377426359)*x_ref[2]
        ref[3,4,2,0]=(-0.836038966171)*x_ref[0]**o+(-0.772990045933)*x_ref[0]+(0.0622128209877)*x_ref[1]**o+(-0.73733496141)*x_ref[1]+(0.0234446925161)*x_ref[2]**o+(-0.682234213315)*x_ref[2]
        ref[3,4,2,1]=(0.250498084214)*x_ref[0]**o+(0.539088356159)*x_ref[0]+(-0.691842680176)*x_ref[1]**o+(0.710040802953)*x_ref[1]+(-0.43691551501)*x_ref[2]**o+(0.710668600989)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.70941034293)*x[0]**o+(-0.988234004949)*x[0]+(0.440342659022)*x[1]**o+(0.221444369839)*x[1]
        ref=(0.70941034293)*x_ref[0]**o+(-0.988234004949)*x_ref[0]+(0.440342659022)*x_ref[1]**o+(0.221444369839)*x_ref[1]
      else:
        arg=(0.720798141776)*x[0]**o+(0.841475017012)*x[0]+(-0.76285313232)*x[1]**o+(-0.214354625268)*x[1]+(-0.887218533751)*x[2]**o+(-0.066368425118)*x[2]
        ref=(0.720798141776)*x_ref[0]**o+(0.841475017012)*x_ref[0]+(-0.76285313232)*x_ref[1]**o+(-0.214354625268)*x_ref[1]+(-0.887218533751)*x_ref[2]**o+(-0.066368425118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.940260123916)*x[0]**o+(-0.977192570293)*x[0]+(-0.0910686020625)*x[1]**o+(0.464205016172)*x[1]
        arg[1]=(-0.634146043992)*x[0]**o+(0.00438663094072)*x[0]+(0.905996907036)*x[1]**o+(0.175291641923)*x[1]
        ref[0]=(-0.940260123916)*x_ref[0]**o+(-0.977192570293)*x_ref[0]+(-0.0910686020625)*x_ref[1]**o+(0.464205016172)*x_ref[1]
        ref[1]=(-0.634146043992)*x_ref[0]**o+(0.00438663094072)*x_ref[0]+(0.905996907036)*x_ref[1]**o+(0.175291641923)*x_ref[1]
      else:
        arg[0]=(0.0746302635373)*x[0]**o+(0.611568121627)*x[0]+(0.0392401043636)*x[1]**o+(0.846339953866)*x[1]+(0.775935477701)*x[2]**o+(0.309530355508)*x[2]
        arg[1]=(0.790143430408)*x[0]**o+(0.987665779965)*x[0]+(0.569313447766)*x[1]**o+(0.170690891282)*x[1]+(0.224015411231)*x[2]**o+(-0.670946134012)*x[2]
        ref[0]=(0.0746302635373)*x_ref[0]**o+(0.611568121627)*x_ref[0]+(0.0392401043636)*x_ref[1]**o+(0.846339953866)*x_ref[1]+(0.775935477701)*x_ref[2]**o+(0.309530355508)*x_ref[2]
        ref[1]=(0.790143430408)*x_ref[0]**o+(0.987665779965)*x_ref[0]+(0.569313447766)*x_ref[1]**o+(0.170690891282)*x_ref[1]+(0.224015411231)*x_ref[2]**o+(-0.670946134012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.339464191741)*x[0]**o+(0.898634031554)*x[0]+(-0.67217412133)*x[1]**o+(-0.732882671617)*x[1]
        arg[0,1]=(0.475031035618)*x[0]**o+(-0.712271710437)*x[0]+(-0.346485473024)*x[1]**o+(0.936873073538)*x[1]
        arg[0,2]=(0.139217050193)*x[0]**o+(0.532762047045)*x[0]+(-0.411674842781)*x[1]**o+(-0.423096144938)*x[1]
        arg[0,3]=(-0.146193322436)*x[0]**o+(-0.227693594412)*x[0]+(0.74715898621)*x[1]**o+(0.538402714773)*x[1]
        arg[0,4]=(0.233316268084)*x[0]**o+(-0.0155557244918)*x[0]+(0.588017790807)*x[1]**o+(-0.58604454457)*x[1]
        arg[1,0]=(0.30473687856)*x[0]**o+(0.933154007362)*x[0]+(-0.0797692178609)*x[1]**o+(0.270021035922)*x[1]
        arg[1,1]=(-0.495613786111)*x[0]**o+(0.0857849462525)*x[0]+(0.330048664047)*x[1]**o+(0.0442768259753)*x[1]
        arg[1,2]=(0.203830377049)*x[0]**o+(-0.891094217974)*x[0]+(-0.643947141174)*x[1]**o+(-0.564315919126)*x[1]
        arg[1,3]=(-0.783514920199)*x[0]**o+(0.356882300706)*x[0]+(-0.307634481801)*x[1]**o+(0.521370775315)*x[1]
        arg[1,4]=(0.418545793821)*x[0]**o+(0.574742804708)*x[0]+(-0.340992300811)*x[1]**o+(-0.892118088815)*x[1]
        arg[2,0]=(-0.361915184269)*x[0]**o+(-0.834492406999)*x[0]+(0.750344593844)*x[1]**o+(-0.516981733186)*x[1]
        arg[2,1]=(-0.0123811923638)*x[0]**o+(-0.9808055138)*x[0]+(-0.176978785352)*x[1]**o+(0.519804799089)*x[1]
        arg[2,2]=(-0.901764228726)*x[0]**o+(-0.610547400097)*x[0]+(0.384870544403)*x[1]**o+(0.0308096944571)*x[1]
        arg[2,3]=(-0.163626872732)*x[0]**o+(0.278523098981)*x[0]+(-0.0544994735633)*x[1]**o+(-0.319354318824)*x[1]
        arg[2,4]=(0.558959525812)*x[0]**o+(0.435157723929)*x[0]+(0.758922109245)*x[1]**o+(-0.702427033542)*x[1]
        arg[3,0]=(0.832800430277)*x[0]**o+(-0.709285644484)*x[0]+(-0.201572915312)*x[1]**o+(0.398294738552)*x[1]
        arg[3,1]=(-0.836258924261)*x[0]**o+(-0.461452427039)*x[0]+(0.314485542425)*x[1]**o+(0.860807479051)*x[1]
        arg[3,2]=(-0.569661610862)*x[0]**o+(-0.46125190469)*x[0]+(0.807311492846)*x[1]**o+(0.72312107176)*x[1]
        arg[3,3]=(0.842126813179)*x[0]**o+(-0.714991954637)*x[0]+(-0.853100178283)*x[1]**o+(-0.644363753291)*x[1]
        arg[3,4]=(-0.0480369646571)*x[0]**o+(-0.330580215997)*x[0]+(0.609757645513)*x[1]**o+(0.915764017541)*x[1]
        ref[0,0]=(0.339464191741)*x_ref[0]**o+(0.898634031554)*x_ref[0]+(-0.67217412133)*x_ref[1]**o+(-0.732882671617)*x_ref[1]
        ref[0,1]=(0.475031035618)*x_ref[0]**o+(-0.712271710437)*x_ref[0]+(-0.346485473024)*x_ref[1]**o+(0.936873073538)*x_ref[1]
        ref[0,2]=(0.139217050193)*x_ref[0]**o+(0.532762047045)*x_ref[0]+(-0.411674842781)*x_ref[1]**o+(-0.423096144938)*x_ref[1]
        ref[0,3]=(-0.146193322436)*x_ref[0]**o+(-0.227693594412)*x_ref[0]+(0.74715898621)*x_ref[1]**o+(0.538402714773)*x_ref[1]
        ref[0,4]=(0.233316268084)*x_ref[0]**o+(-0.0155557244918)*x_ref[0]+(0.588017790807)*x_ref[1]**o+(-0.58604454457)*x_ref[1]
        ref[1,0]=(0.30473687856)*x_ref[0]**o+(0.933154007362)*x_ref[0]+(-0.0797692178609)*x_ref[1]**o+(0.270021035922)*x_ref[1]
        ref[1,1]=(-0.495613786111)*x_ref[0]**o+(0.0857849462525)*x_ref[0]+(0.330048664047)*x_ref[1]**o+(0.0442768259753)*x_ref[1]
        ref[1,2]=(0.203830377049)*x_ref[0]**o+(-0.891094217974)*x_ref[0]+(-0.643947141174)*x_ref[1]**o+(-0.564315919126)*x_ref[1]
        ref[1,3]=(-0.783514920199)*x_ref[0]**o+(0.356882300706)*x_ref[0]+(-0.307634481801)*x_ref[1]**o+(0.521370775315)*x_ref[1]
        ref[1,4]=(0.418545793821)*x_ref[0]**o+(0.574742804708)*x_ref[0]+(-0.340992300811)*x_ref[1]**o+(-0.892118088815)*x_ref[1]
        ref[2,0]=(-0.361915184269)*x_ref[0]**o+(-0.834492406999)*x_ref[0]+(0.750344593844)*x_ref[1]**o+(-0.516981733186)*x_ref[1]
        ref[2,1]=(-0.0123811923638)*x_ref[0]**o+(-0.9808055138)*x_ref[0]+(-0.176978785352)*x_ref[1]**o+(0.519804799089)*x_ref[1]
        ref[2,2]=(-0.901764228726)*x_ref[0]**o+(-0.610547400097)*x_ref[0]+(0.384870544403)*x_ref[1]**o+(0.0308096944571)*x_ref[1]
        ref[2,3]=(-0.163626872732)*x_ref[0]**o+(0.278523098981)*x_ref[0]+(-0.0544994735633)*x_ref[1]**o+(-0.319354318824)*x_ref[1]
        ref[2,4]=(0.558959525812)*x_ref[0]**o+(0.435157723929)*x_ref[0]+(0.758922109245)*x_ref[1]**o+(-0.702427033542)*x_ref[1]
        ref[3,0]=(0.832800430277)*x_ref[0]**o+(-0.709285644484)*x_ref[0]+(-0.201572915312)*x_ref[1]**o+(0.398294738552)*x_ref[1]
        ref[3,1]=(-0.836258924261)*x_ref[0]**o+(-0.461452427039)*x_ref[0]+(0.314485542425)*x_ref[1]**o+(0.860807479051)*x_ref[1]
        ref[3,2]=(-0.569661610862)*x_ref[0]**o+(-0.46125190469)*x_ref[0]+(0.807311492846)*x_ref[1]**o+(0.72312107176)*x_ref[1]
        ref[3,3]=(0.842126813179)*x_ref[0]**o+(-0.714991954637)*x_ref[0]+(-0.853100178283)*x_ref[1]**o+(-0.644363753291)*x_ref[1]
        ref[3,4]=(-0.0480369646571)*x_ref[0]**o+(-0.330580215997)*x_ref[0]+(0.609757645513)*x_ref[1]**o+(0.915764017541)*x_ref[1]
      else:
        arg[0,0]=(-0.256088293513)*x[0]**o+(0.980138068946)*x[0]+(0.375373123738)*x[1]**o+(0.118268751779)*x[1]+(-0.848800006681)*x[2]**o+(0.861808952535)*x[2]
        arg[0,1]=(-0.0658163161725)*x[0]**o+(-0.21582867866)*x[0]+(0.99787757796)*x[1]**o+(0.0175357022646)*x[1]+(-0.305438242469)*x[2]**o+(0.114385992332)*x[2]
        arg[0,2]=(-0.296948302927)*x[0]**o+(0.146852900649)*x[0]+(0.386910961544)*x[1]**o+(0.353651457394)*x[1]+(0.420964895802)*x[2]**o+(0.71145244268)*x[2]
        arg[0,3]=(0.61102494979)*x[0]**o+(-0.316055853885)*x[0]+(-0.117200390837)*x[1]**o+(-0.808630441349)*x[1]+(0.0914828526025)*x[2]**o+(0.305837943839)*x[2]
        arg[0,4]=(-0.22863340097)*x[0]**o+(0.398083421617)*x[0]+(-0.39971917463)*x[1]**o+(-0.0307312416932)*x[1]+(0.510443863383)*x[2]**o+(-0.129495846726)*x[2]
        arg[1,0]=(0.423968874298)*x[0]**o+(0.371480448801)*x[0]+(-0.934565508125)*x[1]**o+(0.890972208341)*x[1]+(-0.169067233411)*x[2]**o+(-0.757264272484)*x[2]
        arg[1,1]=(0.999152573666)*x[0]**o+(-0.659676649566)*x[0]+(0.993153643567)*x[1]**o+(0.350141892848)*x[1]+(-0.493784056536)*x[2]**o+(-0.966494205893)*x[2]
        arg[1,2]=(-0.141337889298)*x[0]**o+(0.689935513838)*x[0]+(0.335444607712)*x[1]**o+(-0.733351283662)*x[1]+(0.323775887172)*x[2]**o+(0.314099156544)*x[2]
        arg[1,3]=(-0.157132412118)*x[0]**o+(0.602468153313)*x[0]+(-0.618134886591)*x[1]**o+(-0.404425578444)*x[1]+(-0.5069585924)*x[2]**o+(0.616829810596)*x[2]
        arg[1,4]=(-0.107902610028)*x[0]**o+(0.664787826515)*x[0]+(-0.19940280262)*x[1]**o+(0.461794954077)*x[1]+(-0.0516671817578)*x[2]**o+(0.426165289622)*x[2]
        arg[2,0]=(-0.568061237616)*x[0]**o+(-0.744340542576)*x[0]+(-0.0125924903194)*x[1]**o+(0.710153866172)*x[1]+(0.979871553596)*x[2]**o+(0.945862455036)*x[2]
        arg[2,1]=(0.201599120538)*x[0]**o+(-0.11057289344)*x[0]+(0.522031147141)*x[1]**o+(0.194694594101)*x[1]+(-0.771997994079)*x[2]**o+(-0.513574957413)*x[2]
        arg[2,2]=(0.401979025382)*x[0]**o+(-0.743641669014)*x[0]+(-0.796071557874)*x[1]**o+(-0.352651095657)*x[1]+(-0.564117536877)*x[2]**o+(0.952729715875)*x[2]
        arg[2,3]=(-0.526162893019)*x[0]**o+(0.691019471759)*x[0]+(-0.68646084088)*x[1]**o+(0.839630438544)*x[1]+(-0.491997512287)*x[2]**o+(-0.405069600537)*x[2]
        arg[2,4]=(0.117664099703)*x[0]**o+(-0.807496546596)*x[0]+(0.607810703702)*x[1]**o+(-0.696118154009)*x[1]+(-0.156011683803)*x[2]**o+(-0.26100420077)*x[2]
        arg[3,0]=(0.280066174527)*x[0]**o+(-0.225940135983)*x[0]+(-0.338716163489)*x[1]**o+(0.628405669096)*x[1]+(-0.890543405425)*x[2]**o+(-0.142082524497)*x[2]
        arg[3,1]=(-0.659241351226)*x[0]**o+(0.275483386126)*x[0]+(-0.318352239058)*x[1]**o+(0.168536648793)*x[1]+(0.539555610218)*x[2]**o+(-0.95280560435)*x[2]
        arg[3,2]=(0.673861565841)*x[0]**o+(-0.0303873544723)*x[0]+(-0.638241094445)*x[1]**o+(0.664488908579)*x[1]+(-0.789484082135)*x[2]**o+(-0.982437408359)*x[2]
        arg[3,3]=(-0.902346309293)*x[0]**o+(0.818106160371)*x[0]+(-0.410074078135)*x[1]**o+(0.977554592344)*x[1]+(-0.00195907247124)*x[2]**o+(0.94762151647)*x[2]
        arg[3,4]=(0.27272813538)*x[0]**o+(0.0563423990113)*x[0]+(0.254612551508)*x[1]**o+(-0.901948547799)*x[1]+(0.263613075865)*x[2]**o+(0.879691314116)*x[2]
        ref[0,0]=(-0.256088293513)*x_ref[0]**o+(0.980138068946)*x_ref[0]+(0.375373123738)*x_ref[1]**o+(0.118268751779)*x_ref[1]+(-0.848800006681)*x_ref[2]**o+(0.861808952535)*x_ref[2]
        ref[0,1]=(-0.0658163161725)*x_ref[0]**o+(-0.21582867866)*x_ref[0]+(0.99787757796)*x_ref[1]**o+(0.0175357022646)*x_ref[1]+(-0.305438242469)*x_ref[2]**o+(0.114385992332)*x_ref[2]
        ref[0,2]=(-0.296948302927)*x_ref[0]**o+(0.146852900649)*x_ref[0]+(0.386910961544)*x_ref[1]**o+(0.353651457394)*x_ref[1]+(0.420964895802)*x_ref[2]**o+(0.71145244268)*x_ref[2]
        ref[0,3]=(0.61102494979)*x_ref[0]**o+(-0.316055853885)*x_ref[0]+(-0.117200390837)*x_ref[1]**o+(-0.808630441349)*x_ref[1]+(0.0914828526025)*x_ref[2]**o+(0.305837943839)*x_ref[2]
        ref[0,4]=(-0.22863340097)*x_ref[0]**o+(0.398083421617)*x_ref[0]+(-0.39971917463)*x_ref[1]**o+(-0.0307312416932)*x_ref[1]+(0.510443863383)*x_ref[2]**o+(-0.129495846726)*x_ref[2]
        ref[1,0]=(0.423968874298)*x_ref[0]**o+(0.371480448801)*x_ref[0]+(-0.934565508125)*x_ref[1]**o+(0.890972208341)*x_ref[1]+(-0.169067233411)*x_ref[2]**o+(-0.757264272484)*x_ref[2]
        ref[1,1]=(0.999152573666)*x_ref[0]**o+(-0.659676649566)*x_ref[0]+(0.993153643567)*x_ref[1]**o+(0.350141892848)*x_ref[1]+(-0.493784056536)*x_ref[2]**o+(-0.966494205893)*x_ref[2]
        ref[1,2]=(-0.141337889298)*x_ref[0]**o+(0.689935513838)*x_ref[0]+(0.335444607712)*x_ref[1]**o+(-0.733351283662)*x_ref[1]+(0.323775887172)*x_ref[2]**o+(0.314099156544)*x_ref[2]
        ref[1,3]=(-0.157132412118)*x_ref[0]**o+(0.602468153313)*x_ref[0]+(-0.618134886591)*x_ref[1]**o+(-0.404425578444)*x_ref[1]+(-0.5069585924)*x_ref[2]**o+(0.616829810596)*x_ref[2]
        ref[1,4]=(-0.107902610028)*x_ref[0]**o+(0.664787826515)*x_ref[0]+(-0.19940280262)*x_ref[1]**o+(0.461794954077)*x_ref[1]+(-0.0516671817578)*x_ref[2]**o+(0.426165289622)*x_ref[2]
        ref[2,0]=(-0.568061237616)*x_ref[0]**o+(-0.744340542576)*x_ref[0]+(-0.0125924903194)*x_ref[1]**o+(0.710153866172)*x_ref[1]+(0.979871553596)*x_ref[2]**o+(0.945862455036)*x_ref[2]
        ref[2,1]=(0.201599120538)*x_ref[0]**o+(-0.11057289344)*x_ref[0]+(0.522031147141)*x_ref[1]**o+(0.194694594101)*x_ref[1]+(-0.771997994079)*x_ref[2]**o+(-0.513574957413)*x_ref[2]
        ref[2,2]=(0.401979025382)*x_ref[0]**o+(-0.743641669014)*x_ref[0]+(-0.796071557874)*x_ref[1]**o+(-0.352651095657)*x_ref[1]+(-0.564117536877)*x_ref[2]**o+(0.952729715875)*x_ref[2]
        ref[2,3]=(-0.526162893019)*x_ref[0]**o+(0.691019471759)*x_ref[0]+(-0.68646084088)*x_ref[1]**o+(0.839630438544)*x_ref[1]+(-0.491997512287)*x_ref[2]**o+(-0.405069600537)*x_ref[2]
        ref[2,4]=(0.117664099703)*x_ref[0]**o+(-0.807496546596)*x_ref[0]+(0.607810703702)*x_ref[1]**o+(-0.696118154009)*x_ref[1]+(-0.156011683803)*x_ref[2]**o+(-0.26100420077)*x_ref[2]
        ref[3,0]=(0.280066174527)*x_ref[0]**o+(-0.225940135983)*x_ref[0]+(-0.338716163489)*x_ref[1]**o+(0.628405669096)*x_ref[1]+(-0.890543405425)*x_ref[2]**o+(-0.142082524497)*x_ref[2]
        ref[3,1]=(-0.659241351226)*x_ref[0]**o+(0.275483386126)*x_ref[0]+(-0.318352239058)*x_ref[1]**o+(0.168536648793)*x_ref[1]+(0.539555610218)*x_ref[2]**o+(-0.95280560435)*x_ref[2]
        ref[3,2]=(0.673861565841)*x_ref[0]**o+(-0.0303873544723)*x_ref[0]+(-0.638241094445)*x_ref[1]**o+(0.664488908579)*x_ref[1]+(-0.789484082135)*x_ref[2]**o+(-0.982437408359)*x_ref[2]
        ref[3,3]=(-0.902346309293)*x_ref[0]**o+(0.818106160371)*x_ref[0]+(-0.410074078135)*x_ref[1]**o+(0.977554592344)*x_ref[1]+(-0.00195907247124)*x_ref[2]**o+(0.94762151647)*x_ref[2]
        ref[3,4]=(0.27272813538)*x_ref[0]**o+(0.0563423990113)*x_ref[0]+(0.254612551508)*x_ref[1]**o+(-0.901948547799)*x_ref[1]+(0.263613075865)*x_ref[2]**o+(0.879691314116)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.023497804886)*x[0]**o+(0.934780781287)*x[0]+(0.488935915319)*x[1]**o+(-0.843152487959)*x[1]
        arg[0,0,1]=(0.074870516034)*x[0]**o+(-0.515115482624)*x[0]+(-0.582499912837)*x[1]**o+(-0.148911376041)*x[1]
        arg[0,1,0]=(0.0364775990427)*x[0]**o+(-0.258379408121)*x[0]+(0.109704222415)*x[1]**o+(-0.0149203380717)*x[1]
        arg[0,1,1]=(0.749804547117)*x[0]**o+(-0.65162531688)*x[0]+(-0.529952602569)*x[1]**o+(0.901735743956)*x[1]
        arg[1,0,0]=(0.74941277085)*x[0]**o+(-0.878903916612)*x[0]+(0.866605680273)*x[1]**o+(0.182572622995)*x[1]
        arg[1,0,1]=(-0.787476915117)*x[0]**o+(0.945314823283)*x[0]+(0.165599597226)*x[1]**o+(0.733712603726)*x[1]
        arg[1,1,0]=(0.696240085313)*x[0]**o+(0.775501454572)*x[0]+(0.225131553571)*x[1]**o+(-0.75288780057)*x[1]
        arg[1,1,1]=(-0.352416785284)*x[0]**o+(-0.681447081047)*x[0]+(0.164273429957)*x[1]**o+(-0.0977235074729)*x[1]
        arg[2,0,0]=(-0.484651384894)*x[0]**o+(0.162048679786)*x[0]+(0.433731200543)*x[1]**o+(-0.2255104779)*x[1]
        arg[2,0,1]=(0.230113349848)*x[0]**o+(0.674218493869)*x[0]+(0.124852942645)*x[1]**o+(0.361278270173)*x[1]
        arg[2,1,0]=(-0.604246197352)*x[0]**o+(0.26763755318)*x[0]+(-0.821482774631)*x[1]**o+(-0.821813801109)*x[1]
        arg[2,1,1]=(0.71669276429)*x[0]**o+(0.861782230871)*x[0]+(-0.837503788825)*x[1]**o+(0.0437253153979)*x[1]
        arg[3,0,0]=(-0.0941343724431)*x[0]**o+(0.376175369776)*x[0]+(-0.455752787585)*x[1]**o+(0.777480094583)*x[1]
        arg[3,0,1]=(-0.730897141543)*x[0]**o+(0.750046665085)*x[0]+(0.855742434657)*x[1]**o+(0.941826664412)*x[1]
        arg[3,1,0]=(-0.788757907886)*x[0]**o+(-0.0255067018415)*x[0]+(-0.875292088318)*x[1]**o+(0.401677848256)*x[1]
        arg[3,1,1]=(-0.980570910972)*x[0]**o+(0.936108649845)*x[0]+(-0.0463362409126)*x[1]**o+(0.630408734656)*x[1]
        arg[4,0,0]=(-0.72642075391)*x[0]**o+(0.283788406364)*x[0]+(-0.119645741265)*x[1]**o+(0.66948892422)*x[1]
        arg[4,0,1]=(0.178235706017)*x[0]**o+(0.693466107843)*x[0]+(-0.585602683437)*x[1]**o+(-0.493386713436)*x[1]
        arg[4,1,0]=(0.348039277129)*x[0]**o+(-0.806776084421)*x[0]+(-0.73591780841)*x[1]**o+(-0.310246367505)*x[1]
        arg[4,1,1]=(0.993294877418)*x[0]**o+(-0.683114302069)*x[0]+(0.0679022755308)*x[1]**o+(0.341981988936)*x[1]
        arg[5,0,0]=(-0.807380701959)*x[0]**o+(-0.78936779476)*x[0]+(0.142089012614)*x[1]**o+(-0.129335524031)*x[1]
        arg[5,0,1]=(-0.855714388031)*x[0]**o+(0.00825196217877)*x[0]+(0.310039590415)*x[1]**o+(-0.242861862397)*x[1]
        arg[5,1,0]=(-0.0908183346277)*x[0]**o+(-0.941876543235)*x[0]+(0.981459557493)*x[1]**o+(0.559563798858)*x[1]
        arg[5,1,1]=(-0.852223567371)*x[0]**o+(-0.272541886199)*x[0]+(-0.873908247032)*x[1]**o+(-0.387571769068)*x[1]
        ref[0,0,0]=(-0.023497804886)*x_ref[0]**o+(0.934780781287)*x_ref[0]+(0.488935915319)*x_ref[1]**o+(-0.843152487959)*x_ref[1]
        ref[0,0,1]=(0.074870516034)*x_ref[0]**o+(-0.515115482624)*x_ref[0]+(-0.582499912837)*x_ref[1]**o+(-0.148911376041)*x_ref[1]
        ref[0,1,0]=(0.0364775990427)*x_ref[0]**o+(-0.258379408121)*x_ref[0]+(0.109704222415)*x_ref[1]**o+(-0.0149203380717)*x_ref[1]
        ref[0,1,1]=(0.749804547117)*x_ref[0]**o+(-0.65162531688)*x_ref[0]+(-0.529952602569)*x_ref[1]**o+(0.901735743956)*x_ref[1]
        ref[1,0,0]=(0.74941277085)*x_ref[0]**o+(-0.878903916612)*x_ref[0]+(0.866605680273)*x_ref[1]**o+(0.182572622995)*x_ref[1]
        ref[1,0,1]=(-0.787476915117)*x_ref[0]**o+(0.945314823283)*x_ref[0]+(0.165599597226)*x_ref[1]**o+(0.733712603726)*x_ref[1]
        ref[1,1,0]=(0.696240085313)*x_ref[0]**o+(0.775501454572)*x_ref[0]+(0.225131553571)*x_ref[1]**o+(-0.75288780057)*x_ref[1]
        ref[1,1,1]=(-0.352416785284)*x_ref[0]**o+(-0.681447081047)*x_ref[0]+(0.164273429957)*x_ref[1]**o+(-0.0977235074729)*x_ref[1]
        ref[2,0,0]=(-0.484651384894)*x_ref[0]**o+(0.162048679786)*x_ref[0]+(0.433731200543)*x_ref[1]**o+(-0.2255104779)*x_ref[1]
        ref[2,0,1]=(0.230113349848)*x_ref[0]**o+(0.674218493869)*x_ref[0]+(0.124852942645)*x_ref[1]**o+(0.361278270173)*x_ref[1]
        ref[2,1,0]=(-0.604246197352)*x_ref[0]**o+(0.26763755318)*x_ref[0]+(-0.821482774631)*x_ref[1]**o+(-0.821813801109)*x_ref[1]
        ref[2,1,1]=(0.71669276429)*x_ref[0]**o+(0.861782230871)*x_ref[0]+(-0.837503788825)*x_ref[1]**o+(0.0437253153979)*x_ref[1]
        ref[3,0,0]=(-0.0941343724431)*x_ref[0]**o+(0.376175369776)*x_ref[0]+(-0.455752787585)*x_ref[1]**o+(0.777480094583)*x_ref[1]
        ref[3,0,1]=(-0.730897141543)*x_ref[0]**o+(0.750046665085)*x_ref[0]+(0.855742434657)*x_ref[1]**o+(0.941826664412)*x_ref[1]
        ref[3,1,0]=(-0.788757907886)*x_ref[0]**o+(-0.0255067018415)*x_ref[0]+(-0.875292088318)*x_ref[1]**o+(0.401677848256)*x_ref[1]
        ref[3,1,1]=(-0.980570910972)*x_ref[0]**o+(0.936108649845)*x_ref[0]+(-0.0463362409126)*x_ref[1]**o+(0.630408734656)*x_ref[1]
        ref[4,0,0]=(-0.72642075391)*x_ref[0]**o+(0.283788406364)*x_ref[0]+(-0.119645741265)*x_ref[1]**o+(0.66948892422)*x_ref[1]
        ref[4,0,1]=(0.178235706017)*x_ref[0]**o+(0.693466107843)*x_ref[0]+(-0.585602683437)*x_ref[1]**o+(-0.493386713436)*x_ref[1]
        ref[4,1,0]=(0.348039277129)*x_ref[0]**o+(-0.806776084421)*x_ref[0]+(-0.73591780841)*x_ref[1]**o+(-0.310246367505)*x_ref[1]
        ref[4,1,1]=(0.993294877418)*x_ref[0]**o+(-0.683114302069)*x_ref[0]+(0.0679022755308)*x_ref[1]**o+(0.341981988936)*x_ref[1]
        ref[5,0,0]=(-0.807380701959)*x_ref[0]**o+(-0.78936779476)*x_ref[0]+(0.142089012614)*x_ref[1]**o+(-0.129335524031)*x_ref[1]
        ref[5,0,1]=(-0.855714388031)*x_ref[0]**o+(0.00825196217877)*x_ref[0]+(0.310039590415)*x_ref[1]**o+(-0.242861862397)*x_ref[1]
        ref[5,1,0]=(-0.0908183346277)*x_ref[0]**o+(-0.941876543235)*x_ref[0]+(0.981459557493)*x_ref[1]**o+(0.559563798858)*x_ref[1]
        ref[5,1,1]=(-0.852223567371)*x_ref[0]**o+(-0.272541886199)*x_ref[0]+(-0.873908247032)*x_ref[1]**o+(-0.387571769068)*x_ref[1]
      else:
        arg[0,0,0]=(0.418981395334)*x[0]**o+(-0.720305875042)*x[0]+(0.876880439526)*x[1]**o+(-0.17785803746)*x[1]+(0.779797986429)*x[2]**o+(0.363827072746)*x[2]
        arg[0,0,1]=(0.159567966427)*x[0]**o+(-0.614151008569)*x[0]+(0.249982224851)*x[1]**o+(-0.119598774468)*x[1]+(-0.66603578755)*x[2]**o+(0.202190479552)*x[2]
        arg[0,1,0]=(0.337868020939)*x[0]**o+(-0.0828148711399)*x[0]+(-0.0631659329293)*x[1]**o+(-0.725100695213)*x[1]+(0.852068398846)*x[2]**o+(-0.951116521211)*x[2]
        arg[0,1,1]=(0.710866728186)*x[0]**o+(-0.517133891599)*x[0]+(0.320028321875)*x[1]**o+(0.0331233805016)*x[1]+(-0.0945382588101)*x[2]**o+(-0.795725457058)*x[2]
        arg[1,0,0]=(0.261503329866)*x[0]**o+(0.697062723662)*x[0]+(0.926550326146)*x[1]**o+(-0.318704885103)*x[1]+(-0.76904956799)*x[2]**o+(0.0750521033639)*x[2]
        arg[1,0,1]=(-0.594460213676)*x[0]**o+(0.356692399969)*x[0]+(-0.611558953982)*x[1]**o+(0.289469152887)*x[1]+(-0.73222413782)*x[2]**o+(-0.576883793051)*x[2]
        arg[1,1,0]=(-0.399324648161)*x[0]**o+(0.975428000926)*x[0]+(-0.889762441468)*x[1]**o+(-0.137573068244)*x[1]+(0.252276361744)*x[2]**o+(-0.554435968553)*x[2]
        arg[1,1,1]=(0.677933864695)*x[0]**o+(-0.19825134391)*x[0]+(0.556783296976)*x[1]**o+(0.0349255641106)*x[1]+(-0.387378575139)*x[2]**o+(0.889628545207)*x[2]
        arg[2,0,0]=(-0.839356850436)*x[0]**o+(0.0186947034157)*x[0]+(-0.264793716814)*x[1]**o+(-0.767468006197)*x[1]+(0.91661695419)*x[2]**o+(-0.374036622497)*x[2]
        arg[2,0,1]=(-0.030357536408)*x[0]**o+(0.57424762933)*x[0]+(-0.590115256932)*x[1]**o+(-0.176169776757)*x[1]+(-0.502845936823)*x[2]**o+(0.0399243413515)*x[2]
        arg[2,1,0]=(0.257324735469)*x[0]**o+(-0.686083724325)*x[0]+(-0.166609564696)*x[1]**o+(-0.132181292726)*x[1]+(0.10419146703)*x[2]**o+(-0.827571838114)*x[2]
        arg[2,1,1]=(0.339465982319)*x[0]**o+(0.766586466841)*x[0]+(0.277538823819)*x[1]**o+(0.222607512697)*x[1]+(-0.909105282607)*x[2]**o+(0.463729842225)*x[2]
        arg[3,0,0]=(-0.00132043337846)*x[0]**o+(0.670451981959)*x[0]+(-0.0122245380346)*x[1]**o+(-0.922535904926)*x[1]+(0.49486742)*x[2]**o+(0.0785274524546)*x[2]
        arg[3,0,1]=(0.0360876610851)*x[0]**o+(0.69319604661)*x[0]+(0.980826095079)*x[1]**o+(0.167689443923)*x[1]+(0.561634492527)*x[2]**o+(0.996948790413)*x[2]
        arg[3,1,0]=(-0.238705628819)*x[0]**o+(-0.915859436874)*x[0]+(0.942122040692)*x[1]**o+(0.971025407162)*x[1]+(-0.965945363654)*x[2]**o+(0.136537412741)*x[2]
        arg[3,1,1]=(-0.672314439537)*x[0]**o+(0.792647645097)*x[0]+(-0.432226224113)*x[1]**o+(0.424246857841)*x[1]+(-0.639842531517)*x[2]**o+(0.0570057507731)*x[2]
        arg[4,0,0]=(0.450831267081)*x[0]**o+(0.75992870254)*x[0]+(0.768779936214)*x[1]**o+(-0.0135947286596)*x[1]+(0.126345191973)*x[2]**o+(-0.619763308655)*x[2]
        arg[4,0,1]=(-0.325005948888)*x[0]**o+(-0.0655421631802)*x[0]+(0.607182269636)*x[1]**o+(-0.631044141835)*x[1]+(-0.356580309429)*x[2]**o+(-0.593198006905)*x[2]
        arg[4,1,0]=(0.304932527518)*x[0]**o+(-0.763995484499)*x[0]+(0.884855995183)*x[1]**o+(-0.42180322102)*x[1]+(0.605742059302)*x[2]**o+(0.448553789891)*x[2]
        arg[4,1,1]=(-0.787145437063)*x[0]**o+(-0.465220934558)*x[0]+(-0.370336760606)*x[1]**o+(-0.854850692759)*x[1]+(0.482896516171)*x[2]**o+(-0.551567519291)*x[2]
        arg[5,0,0]=(-0.537738873639)*x[0]**o+(-0.213268842545)*x[0]+(0.492208051415)*x[1]**o+(-0.609304634017)*x[1]+(-0.487753207718)*x[2]**o+(-0.70650655894)*x[2]
        arg[5,0,1]=(0.600359837292)*x[0]**o+(-0.604421599924)*x[0]+(-0.051356466466)*x[1]**o+(-0.168592544661)*x[1]+(0.789340570036)*x[2]**o+(0.910249026599)*x[2]
        arg[5,1,0]=(0.264506197816)*x[0]**o+(-0.711775430363)*x[0]+(0.384666624979)*x[1]**o+(0.0926714847986)*x[1]+(-0.551232557648)*x[2]**o+(0.758927448717)*x[2]
        arg[5,1,1]=(0.894998835449)*x[0]**o+(-0.466338210072)*x[0]+(-0.522475632898)*x[1]**o+(0.766949173671)*x[1]+(0.915063619607)*x[2]**o+(-0.270696846692)*x[2]
        ref[0,0,0]=(0.418981395334)*x_ref[0]**o+(-0.720305875042)*x_ref[0]+(0.876880439526)*x_ref[1]**o+(-0.17785803746)*x_ref[1]+(0.779797986429)*x_ref[2]**o+(0.363827072746)*x_ref[2]
        ref[0,0,1]=(0.159567966427)*x_ref[0]**o+(-0.614151008569)*x_ref[0]+(0.249982224851)*x_ref[1]**o+(-0.119598774468)*x_ref[1]+(-0.66603578755)*x_ref[2]**o+(0.202190479552)*x_ref[2]
        ref[0,1,0]=(0.337868020939)*x_ref[0]**o+(-0.0828148711399)*x_ref[0]+(-0.0631659329293)*x_ref[1]**o+(-0.725100695213)*x_ref[1]+(0.852068398846)*x_ref[2]**o+(-0.951116521211)*x_ref[2]
        ref[0,1,1]=(0.710866728186)*x_ref[0]**o+(-0.517133891599)*x_ref[0]+(0.320028321875)*x_ref[1]**o+(0.0331233805016)*x_ref[1]+(-0.0945382588101)*x_ref[2]**o+(-0.795725457058)*x_ref[2]
        ref[1,0,0]=(0.261503329866)*x_ref[0]**o+(0.697062723662)*x_ref[0]+(0.926550326146)*x_ref[1]**o+(-0.318704885103)*x_ref[1]+(-0.76904956799)*x_ref[2]**o+(0.0750521033639)*x_ref[2]
        ref[1,0,1]=(-0.594460213676)*x_ref[0]**o+(0.356692399969)*x_ref[0]+(-0.611558953982)*x_ref[1]**o+(0.289469152887)*x_ref[1]+(-0.73222413782)*x_ref[2]**o+(-0.576883793051)*x_ref[2]
        ref[1,1,0]=(-0.399324648161)*x_ref[0]**o+(0.975428000926)*x_ref[0]+(-0.889762441468)*x_ref[1]**o+(-0.137573068244)*x_ref[1]+(0.252276361744)*x_ref[2]**o+(-0.554435968553)*x_ref[2]
        ref[1,1,1]=(0.677933864695)*x_ref[0]**o+(-0.19825134391)*x_ref[0]+(0.556783296976)*x_ref[1]**o+(0.0349255641106)*x_ref[1]+(-0.387378575139)*x_ref[2]**o+(0.889628545207)*x_ref[2]
        ref[2,0,0]=(-0.839356850436)*x_ref[0]**o+(0.0186947034157)*x_ref[0]+(-0.264793716814)*x_ref[1]**o+(-0.767468006197)*x_ref[1]+(0.91661695419)*x_ref[2]**o+(-0.374036622497)*x_ref[2]
        ref[2,0,1]=(-0.030357536408)*x_ref[0]**o+(0.57424762933)*x_ref[0]+(-0.590115256932)*x_ref[1]**o+(-0.176169776757)*x_ref[1]+(-0.502845936823)*x_ref[2]**o+(0.0399243413515)*x_ref[2]
        ref[2,1,0]=(0.257324735469)*x_ref[0]**o+(-0.686083724325)*x_ref[0]+(-0.166609564696)*x_ref[1]**o+(-0.132181292726)*x_ref[1]+(0.10419146703)*x_ref[2]**o+(-0.827571838114)*x_ref[2]
        ref[2,1,1]=(0.339465982319)*x_ref[0]**o+(0.766586466841)*x_ref[0]+(0.277538823819)*x_ref[1]**o+(0.222607512697)*x_ref[1]+(-0.909105282607)*x_ref[2]**o+(0.463729842225)*x_ref[2]
        ref[3,0,0]=(-0.00132043337846)*x_ref[0]**o+(0.670451981959)*x_ref[0]+(-0.0122245380346)*x_ref[1]**o+(-0.922535904926)*x_ref[1]+(0.49486742)*x_ref[2]**o+(0.0785274524546)*x_ref[2]
        ref[3,0,1]=(0.0360876610851)*x_ref[0]**o+(0.69319604661)*x_ref[0]+(0.980826095079)*x_ref[1]**o+(0.167689443923)*x_ref[1]+(0.561634492527)*x_ref[2]**o+(0.996948790413)*x_ref[2]
        ref[3,1,0]=(-0.238705628819)*x_ref[0]**o+(-0.915859436874)*x_ref[0]+(0.942122040692)*x_ref[1]**o+(0.971025407162)*x_ref[1]+(-0.965945363654)*x_ref[2]**o+(0.136537412741)*x_ref[2]
        ref[3,1,1]=(-0.672314439537)*x_ref[0]**o+(0.792647645097)*x_ref[0]+(-0.432226224113)*x_ref[1]**o+(0.424246857841)*x_ref[1]+(-0.639842531517)*x_ref[2]**o+(0.0570057507731)*x_ref[2]
        ref[4,0,0]=(0.450831267081)*x_ref[0]**o+(0.75992870254)*x_ref[0]+(0.768779936214)*x_ref[1]**o+(-0.0135947286596)*x_ref[1]+(0.126345191973)*x_ref[2]**o+(-0.619763308655)*x_ref[2]
        ref[4,0,1]=(-0.325005948888)*x_ref[0]**o+(-0.0655421631802)*x_ref[0]+(0.607182269636)*x_ref[1]**o+(-0.631044141835)*x_ref[1]+(-0.356580309429)*x_ref[2]**o+(-0.593198006905)*x_ref[2]
        ref[4,1,0]=(0.304932527518)*x_ref[0]**o+(-0.763995484499)*x_ref[0]+(0.884855995183)*x_ref[1]**o+(-0.42180322102)*x_ref[1]+(0.605742059302)*x_ref[2]**o+(0.448553789891)*x_ref[2]
        ref[4,1,1]=(-0.787145437063)*x_ref[0]**o+(-0.465220934558)*x_ref[0]+(-0.370336760606)*x_ref[1]**o+(-0.854850692759)*x_ref[1]+(0.482896516171)*x_ref[2]**o+(-0.551567519291)*x_ref[2]
        ref[5,0,0]=(-0.537738873639)*x_ref[0]**o+(-0.213268842545)*x_ref[0]+(0.492208051415)*x_ref[1]**o+(-0.609304634017)*x_ref[1]+(-0.487753207718)*x_ref[2]**o+(-0.70650655894)*x_ref[2]
        ref[5,0,1]=(0.600359837292)*x_ref[0]**o+(-0.604421599924)*x_ref[0]+(-0.051356466466)*x_ref[1]**o+(-0.168592544661)*x_ref[1]+(0.789340570036)*x_ref[2]**o+(0.910249026599)*x_ref[2]
        ref[5,1,0]=(0.264506197816)*x_ref[0]**o+(-0.711775430363)*x_ref[0]+(0.384666624979)*x_ref[1]**o+(0.0926714847986)*x_ref[1]+(-0.551232557648)*x_ref[2]**o+(0.758927448717)*x_ref[2]
        ref[5,1,1]=(0.894998835449)*x_ref[0]**o+(-0.466338210072)*x_ref[0]+(-0.522475632898)*x_ref[1]**o+(0.766949173671)*x_ref[1]+(0.915063619607)*x_ref[2]**o+(-0.270696846692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.240887319278)*x[0]**o+(-0.192166587545)*x[0]+(0.0176974287409)*x[1]**o+(0.100761748824)*x[1]
        arg[0,0,0,1]=(-0.549649845775)*x[0]**o+(-0.293969245646)*x[0]+(-0.605885325133)*x[1]**o+(0.799558451173)*x[1]
        arg[0,0,1,0]=(-0.899658769582)*x[0]**o+(0.484524193987)*x[0]+(0.437485870351)*x[1]**o+(-0.220549034927)*x[1]
        arg[0,0,1,1]=(-0.146539961793)*x[0]**o+(-0.278343486929)*x[0]+(0.0913912456082)*x[1]**o+(-0.751028975123)*x[1]
        arg[0,0,2,0]=(0.806344701194)*x[0]**o+(0.0537872824561)*x[0]+(-0.70407828379)*x[1]**o+(-0.922963486601)*x[1]
        arg[0,0,2,1]=(0.307202902495)*x[0]**o+(0.301603506731)*x[0]+(0.232983743616)*x[1]**o+(-0.744267542759)*x[1]
        arg[0,1,0,0]=(-0.550616060824)*x[0]**o+(-0.661367750254)*x[0]+(-0.695905895864)*x[1]**o+(-0.788045922398)*x[1]
        arg[0,1,0,1]=(-0.934169134035)*x[0]**o+(-0.843861105521)*x[0]+(-0.237877254439)*x[1]**o+(-0.364956821348)*x[1]
        arg[0,1,1,0]=(0.0950675750026)*x[0]**o+(-0.815510640594)*x[0]+(0.72526035968)*x[1]**o+(0.898675190106)*x[1]
        arg[0,1,1,1]=(0.715609066776)*x[0]**o+(0.371314845206)*x[0]+(0.326878867244)*x[1]**o+(-0.468798950199)*x[1]
        arg[0,1,2,0]=(0.392175939547)*x[0]**o+(0.105332682674)*x[0]+(0.997967455477)*x[1]**o+(0.0322491822496)*x[1]
        arg[0,1,2,1]=(-0.804498399187)*x[0]**o+(0.819653744739)*x[0]+(0.609610573703)*x[1]**o+(-0.623299047794)*x[1]
        arg[0,2,0,0]=(0.993877917386)*x[0]**o+(0.815952897899)*x[0]+(-0.258432174486)*x[1]**o+(0.209324387308)*x[1]
        arg[0,2,0,1]=(0.395946889364)*x[0]**o+(-0.859500539908)*x[0]+(-0.68495734785)*x[1]**o+(0.791813946302)*x[1]
        arg[0,2,1,0]=(0.154347661108)*x[0]**o+(0.0873140744322)*x[0]+(-0.423935109956)*x[1]**o+(-0.754297239226)*x[1]
        arg[0,2,1,1]=(-0.623998419484)*x[0]**o+(0.487300130941)*x[0]+(-0.323613219784)*x[1]**o+(0.744877983729)*x[1]
        arg[0,2,2,0]=(0.515686374371)*x[0]**o+(0.512572153346)*x[0]+(-0.641245783272)*x[1]**o+(-0.390660158759)*x[1]
        arg[0,2,2,1]=(0.24398238139)*x[0]**o+(-0.172020271678)*x[0]+(0.980197812701)*x[1]**o+(0.336941934503)*x[1]
        arg[0,3,0,0]=(0.0154584447615)*x[0]**o+(-0.547048191238)*x[0]+(-0.555319223005)*x[1]**o+(0.0972794598106)*x[1]
        arg[0,3,0,1]=(0.44744389335)*x[0]**o+(0.475046337874)*x[0]+(0.876228982066)*x[1]**o+(-0.0636405241175)*x[1]
        arg[0,3,1,0]=(0.392948465298)*x[0]**o+(-0.341342654182)*x[0]+(0.747185683968)*x[1]**o+(-0.99981172166)*x[1]
        arg[0,3,1,1]=(-0.172355321621)*x[0]**o+(-0.531391484795)*x[0]+(0.451654274019)*x[1]**o+(0.0795298667915)*x[1]
        arg[0,3,2,0]=(0.0841642474674)*x[0]**o+(0.308494435164)*x[0]+(0.53181918181)*x[1]**o+(0.480258067959)*x[1]
        arg[0,3,2,1]=(-0.222903054864)*x[0]**o+(-0.0330655869754)*x[0]+(0.227482525015)*x[1]**o+(0.536136212347)*x[1]
        arg[0,4,0,0]=(0.574798381284)*x[0]**o+(0.181354466484)*x[0]+(0.103049514171)*x[1]**o+(0.156282818835)*x[1]
        arg[0,4,0,1]=(0.478314255289)*x[0]**o+(0.224797222592)*x[0]+(0.307668606272)*x[1]**o+(-0.785262182937)*x[1]
        arg[0,4,1,0]=(0.519356961182)*x[0]**o+(0.646200458344)*x[0]+(0.0522618783181)*x[1]**o+(0.702335307324)*x[1]
        arg[0,4,1,1]=(-0.168756029541)*x[0]**o+(-0.00618364750666)*x[0]+(0.472704616213)*x[1]**o+(0.974257451871)*x[1]
        arg[0,4,2,0]=(-0.279196869573)*x[0]**o+(-0.333083174091)*x[0]+(-0.980089574428)*x[1]**o+(0.405273709405)*x[1]
        arg[0,4,2,1]=(-0.199431380673)*x[0]**o+(0.823294899866)*x[0]+(-0.174125117685)*x[1]**o+(-0.730500493887)*x[1]
        arg[1,0,0,0]=(-0.577281489635)*x[0]**o+(0.231906139898)*x[0]+(0.537128708963)*x[1]**o+(0.363444540898)*x[1]
        arg[1,0,0,1]=(0.204108541882)*x[0]**o+(-0.217661714888)*x[0]+(0.963384293759)*x[1]**o+(-0.0817519801351)*x[1]
        arg[1,0,1,0]=(0.540681325439)*x[0]**o+(-0.496036960808)*x[0]+(0.421037689661)*x[1]**o+(0.19399086996)*x[1]
        arg[1,0,1,1]=(0.482938542126)*x[0]**o+(-0.0725027675209)*x[0]+(-0.441517962054)*x[1]**o+(0.787445860714)*x[1]
        arg[1,0,2,0]=(0.0378928111257)*x[0]**o+(0.451368848019)*x[0]+(-0.228668284485)*x[1]**o+(0.505927229432)*x[1]
        arg[1,0,2,1]=(-0.59833518872)*x[0]**o+(0.956068253354)*x[0]+(0.667905995469)*x[1]**o+(0.963273292214)*x[1]
        arg[1,1,0,0]=(0.901637865236)*x[0]**o+(-0.642678971748)*x[0]+(-0.603915019524)*x[1]**o+(-0.208004993919)*x[1]
        arg[1,1,0,1]=(-0.227162661052)*x[0]**o+(-0.5816434303)*x[0]+(0.274961293801)*x[1]**o+(-0.614001328621)*x[1]
        arg[1,1,1,0]=(0.078438078548)*x[0]**o+(0.570191242883)*x[0]+(0.0952878212484)*x[1]**o+(0.844460205497)*x[1]
        arg[1,1,1,1]=(-0.901445811786)*x[0]**o+(-0.996376775344)*x[0]+(0.707470471613)*x[1]**o+(-0.187517936674)*x[1]
        arg[1,1,2,0]=(-0.66891443067)*x[0]**o+(0.0462057239324)*x[0]+(-0.825820539128)*x[1]**o+(-0.766932836843)*x[1]
        arg[1,1,2,1]=(-0.851428828811)*x[0]**o+(0.504484422346)*x[0]+(0.121378315545)*x[1]**o+(0.903947012757)*x[1]
        arg[1,2,0,0]=(-0.393827449768)*x[0]**o+(-0.251827904608)*x[0]+(0.8320040882)*x[1]**o+(0.57595149855)*x[1]
        arg[1,2,0,1]=(0.638738110712)*x[0]**o+(-0.439962152539)*x[0]+(-0.977211078995)*x[1]**o+(0.0528475212451)*x[1]
        arg[1,2,1,0]=(0.824313588941)*x[0]**o+(-0.589968996823)*x[0]+(-0.512156794152)*x[1]**o+(0.142809794443)*x[1]
        arg[1,2,1,1]=(-0.31789333147)*x[0]**o+(0.598839896949)*x[0]+(0.638319398651)*x[1]**o+(0.353625999317)*x[1]
        arg[1,2,2,0]=(-0.720472365012)*x[0]**o+(0.896538030051)*x[0]+(-0.224639169972)*x[1]**o+(-0.898299360738)*x[1]
        arg[1,2,2,1]=(0.123077707167)*x[0]**o+(0.15922851038)*x[0]+(0.678984783843)*x[1]**o+(-0.778188269852)*x[1]
        arg[1,3,0,0]=(0.344256301209)*x[0]**o+(0.884367738289)*x[0]+(0.00556636219289)*x[1]**o+(0.684361331195)*x[1]
        arg[1,3,0,1]=(-0.812245131065)*x[0]**o+(0.971365671228)*x[0]+(0.0696053239569)*x[1]**o+(-0.300266235138)*x[1]
        arg[1,3,1,0]=(0.470254588058)*x[0]**o+(0.0946825509913)*x[0]+(-0.415619255025)*x[1]**o+(-0.27381652858)*x[1]
        arg[1,3,1,1]=(-0.84555138241)*x[0]**o+(0.873510832091)*x[0]+(-0.138010338152)*x[1]**o+(-0.982274515316)*x[1]
        arg[1,3,2,0]=(0.695614136148)*x[0]**o+(-0.192588082771)*x[0]+(-0.21579127791)*x[1]**o+(-0.0335939736388)*x[1]
        arg[1,3,2,1]=(-0.633363463896)*x[0]**o+(-0.958320549993)*x[0]+(0.197331067481)*x[1]**o+(0.315089888409)*x[1]
        arg[1,4,0,0]=(-0.211531219945)*x[0]**o+(-0.399700721799)*x[0]+(0.578927638839)*x[1]**o+(-0.389034976195)*x[1]
        arg[1,4,0,1]=(-0.846399571893)*x[0]**o+(-0.952904755903)*x[0]+(-0.858399802693)*x[1]**o+(-0.23264354907)*x[1]
        arg[1,4,1,0]=(0.255822441586)*x[0]**o+(0.896237710393)*x[0]+(-0.90248008271)*x[1]**o+(0.654804848244)*x[1]
        arg[1,4,1,1]=(-0.112475508163)*x[0]**o+(-0.861697204453)*x[0]+(-0.782581931187)*x[1]**o+(-0.129345341581)*x[1]
        arg[1,4,2,0]=(0.690095687033)*x[0]**o+(0.729317334019)*x[0]+(-0.00961483612288)*x[1]**o+(-0.288320551243)*x[1]
        arg[1,4,2,1]=(-0.648146433517)*x[0]**o+(-0.301246601263)*x[0]+(-0.490119612994)*x[1]**o+(0.231186845615)*x[1]
        arg[2,0,0,0]=(-0.645560891595)*x[0]**o+(-0.889067544988)*x[0]+(-0.193353414221)*x[1]**o+(-0.726988479344)*x[1]
        arg[2,0,0,1]=(-0.083614738484)*x[0]**o+(-0.822530927241)*x[0]+(0.660338934241)*x[1]**o+(0.0383821036508)*x[1]
        arg[2,0,1,0]=(-0.132250736664)*x[0]**o+(-0.571474370931)*x[0]+(-0.755931100342)*x[1]**o+(-0.145196865115)*x[1]
        arg[2,0,1,1]=(0.545219196727)*x[0]**o+(-0.846368787913)*x[0]+(-0.845706160842)*x[1]**o+(0.550703610989)*x[1]
        arg[2,0,2,0]=(-0.766303732195)*x[0]**o+(0.156822893641)*x[0]+(0.00954334254282)*x[1]**o+(0.201506144561)*x[1]
        arg[2,0,2,1]=(-0.274476449632)*x[0]**o+(-0.361690468185)*x[0]+(-0.816535747268)*x[1]**o+(-0.473913811701)*x[1]
        arg[2,1,0,0]=(-0.927592771482)*x[0]**o+(0.534615269814)*x[0]+(-0.571217133452)*x[1]**o+(0.57448299216)*x[1]
        arg[2,1,0,1]=(-0.298153953211)*x[0]**o+(-0.143636909462)*x[0]+(0.990660288864)*x[1]**o+(-0.495864284763)*x[1]
        arg[2,1,1,0]=(-0.110625727208)*x[0]**o+(-0.221057774317)*x[0]+(0.557195078358)*x[1]**o+(0.803162691808)*x[1]
        arg[2,1,1,1]=(-0.115589176775)*x[0]**o+(0.08253351782)*x[0]+(-0.595996958977)*x[1]**o+(0.501435717501)*x[1]
        arg[2,1,2,0]=(0.78764976658)*x[0]**o+(0.966729831711)*x[0]+(-0.585879575667)*x[1]**o+(0.341148176793)*x[1]
        arg[2,1,2,1]=(-0.551697945691)*x[0]**o+(0.545857249323)*x[0]+(-0.919321724495)*x[1]**o+(-0.286043327415)*x[1]
        arg[2,2,0,0]=(-0.669933890655)*x[0]**o+(-0.344862118495)*x[0]+(0.878331282597)*x[1]**o+(0.881893760718)*x[1]
        arg[2,2,0,1]=(-0.217876618414)*x[0]**o+(-0.289118007199)*x[0]+(0.513464171525)*x[1]**o+(0.310055899102)*x[1]
        arg[2,2,1,0]=(0.0472051379154)*x[0]**o+(0.42455388451)*x[0]+(-0.920571834891)*x[1]**o+(-0.779095076537)*x[1]
        arg[2,2,1,1]=(-0.299619304537)*x[0]**o+(-0.289777287294)*x[0]+(0.920505062988)*x[1]**o+(-0.243924179095)*x[1]
        arg[2,2,2,0]=(-0.819343530877)*x[0]**o+(0.975953690875)*x[0]+(-0.787239459241)*x[1]**o+(-0.199900734691)*x[1]
        arg[2,2,2,1]=(0.0980085780878)*x[0]**o+(0.642946515451)*x[0]+(0.951790535152)*x[1]**o+(-0.253914817366)*x[1]
        arg[2,3,0,0]=(-0.882075522824)*x[0]**o+(0.200934502682)*x[0]+(-0.83200464496)*x[1]**o+(0.941479345375)*x[1]
        arg[2,3,0,1]=(-0.615941782209)*x[0]**o+(-0.693242388092)*x[0]+(-0.640417278871)*x[1]**o+(-0.346459094803)*x[1]
        arg[2,3,1,0]=(-0.447919427952)*x[0]**o+(0.205055689141)*x[0]+(-0.453290032411)*x[1]**o+(0.879062836521)*x[1]
        arg[2,3,1,1]=(0.721133924016)*x[0]**o+(0.71294558459)*x[0]+(-0.250087045805)*x[1]**o+(0.605146721129)*x[1]
        arg[2,3,2,0]=(0.247782322921)*x[0]**o+(-0.956367425119)*x[0]+(-0.054045082005)*x[1]**o+(-0.128151832132)*x[1]
        arg[2,3,2,1]=(-0.427039259088)*x[0]**o+(-0.359849242839)*x[0]+(-0.244990934481)*x[1]**o+(-0.038020035629)*x[1]
        arg[2,4,0,0]=(-0.216409150834)*x[0]**o+(0.492547989921)*x[0]+(0.878317438469)*x[1]**o+(-0.112718719119)*x[1]
        arg[2,4,0,1]=(0.487403443102)*x[0]**o+(0.228155519468)*x[0]+(0.923150527516)*x[1]**o+(0.642654867908)*x[1]
        arg[2,4,1,0]=(0.916418483417)*x[0]**o+(-0.177198935173)*x[0]+(-0.790548940412)*x[1]**o+(0.87540167052)*x[1]
        arg[2,4,1,1]=(0.718502075538)*x[0]**o+(0.691254010569)*x[0]+(-0.0134430122729)*x[1]**o+(-0.739888770749)*x[1]
        arg[2,4,2,0]=(0.287825086861)*x[0]**o+(0.596621809347)*x[0]+(0.729784205391)*x[1]**o+(-0.25984257955)*x[1]
        arg[2,4,2,1]=(-0.515102480025)*x[0]**o+(-0.609951840015)*x[0]+(0.229834340075)*x[1]**o+(-0.727157816241)*x[1]
        arg[3,0,0,0]=(-0.349406091363)*x[0]**o+(0.802348451103)*x[0]+(-0.174421346276)*x[1]**o+(0.120711198036)*x[1]
        arg[3,0,0,1]=(0.367085169423)*x[0]**o+(-0.319683236434)*x[0]+(-0.590820542566)*x[1]**o+(0.607625953303)*x[1]
        arg[3,0,1,0]=(0.231259976433)*x[0]**o+(-0.493297408825)*x[0]+(0.919363640504)*x[1]**o+(-0.839841270429)*x[1]
        arg[3,0,1,1]=(0.57672135684)*x[0]**o+(0.882062879704)*x[0]+(-0.600979706444)*x[1]**o+(-0.704008344509)*x[1]
        arg[3,0,2,0]=(-0.208162114527)*x[0]**o+(0.593682243052)*x[0]+(-0.428436832487)*x[1]**o+(0.51007021604)*x[1]
        arg[3,0,2,1]=(0.315264672653)*x[0]**o+(-0.0835897376534)*x[0]+(-0.135315173106)*x[1]**o+(-0.646372309498)*x[1]
        arg[3,1,0,0]=(0.815505953752)*x[0]**o+(0.0625493643292)*x[0]+(0.383118433815)*x[1]**o+(0.17372720435)*x[1]
        arg[3,1,0,1]=(0.557295786969)*x[0]**o+(-0.568823508165)*x[0]+(-0.00874733267675)*x[1]**o+(-0.27454140598)*x[1]
        arg[3,1,1,0]=(0.649021026591)*x[0]**o+(0.232255030448)*x[0]+(0.845939296425)*x[1]**o+(0.116387378294)*x[1]
        arg[3,1,1,1]=(0.431124940343)*x[0]**o+(-0.0865937271969)*x[0]+(0.937911741288)*x[1]**o+(0.362850039234)*x[1]
        arg[3,1,2,0]=(0.155965414439)*x[0]**o+(0.110850421841)*x[0]+(0.244563080191)*x[1]**o+(0.266325498006)*x[1]
        arg[3,1,2,1]=(0.766813300935)*x[0]**o+(0.77492422889)*x[0]+(-0.0141436554714)*x[1]**o+(0.780444449009)*x[1]
        arg[3,2,0,0]=(0.619086019019)*x[0]**o+(-0.574130567601)*x[0]+(-0.417708159593)*x[1]**o+(0.707727284254)*x[1]
        arg[3,2,0,1]=(-0.845657409353)*x[0]**o+(0.705282507748)*x[0]+(0.124557418112)*x[1]**o+(-0.337493494733)*x[1]
        arg[3,2,1,0]=(0.114875146211)*x[0]**o+(-0.724480426892)*x[0]+(-0.0763476904405)*x[1]**o+(0.12064907114)*x[1]
        arg[3,2,1,1]=(-0.74994413376)*x[0]**o+(-0.916476035217)*x[0]+(-0.360412756599)*x[1]**o+(-0.917457363098)*x[1]
        arg[3,2,2,0]=(-0.446839470984)*x[0]**o+(-0.097704838316)*x[0]+(-0.653584364909)*x[1]**o+(0.561014757758)*x[1]
        arg[3,2,2,1]=(-0.178975301038)*x[0]**o+(-0.843669982667)*x[0]+(-0.566610094242)*x[1]**o+(0.213047649003)*x[1]
        arg[3,3,0,0]=(0.144115764355)*x[0]**o+(0.199435411954)*x[0]+(0.173334385199)*x[1]**o+(-0.942957035808)*x[1]
        arg[3,3,0,1]=(-0.417589926342)*x[0]**o+(-0.819188022852)*x[0]+(-0.68299160003)*x[1]**o+(-0.314851447894)*x[1]
        arg[3,3,1,0]=(0.737898219164)*x[0]**o+(0.612906277393)*x[0]+(-0.768349396729)*x[1]**o+(-0.192998755095)*x[1]
        arg[3,3,1,1]=(0.35433362007)*x[0]**o+(0.894174225208)*x[0]+(-0.450333030837)*x[1]**o+(0.819450589483)*x[1]
        arg[3,3,2,0]=(-0.455377131302)*x[0]**o+(-0.748183144687)*x[0]+(-0.482739450231)*x[1]**o+(0.300625929022)*x[1]
        arg[3,3,2,1]=(-0.602766693698)*x[0]**o+(-0.416288436985)*x[0]+(0.837501983536)*x[1]**o+(-0.745935308188)*x[1]
        arg[3,4,0,0]=(-0.706774010405)*x[0]**o+(0.407766818701)*x[0]+(-0.248964401443)*x[1]**o+(0.675466487229)*x[1]
        arg[3,4,0,1]=(0.512302883876)*x[0]**o+(0.0809544867105)*x[0]+(-0.204555203536)*x[1]**o+(-0.628574073749)*x[1]
        arg[3,4,1,0]=(0.126376796421)*x[0]**o+(0.174607814129)*x[0]+(0.68867194136)*x[1]**o+(0.940025934568)*x[1]
        arg[3,4,1,1]=(0.583042444739)*x[0]**o+(-0.786981763836)*x[0]+(0.319397032405)*x[1]**o+(-0.324993351432)*x[1]
        arg[3,4,2,0]=(-0.941997068836)*x[0]**o+(-0.892825520378)*x[0]+(0.275995419027)*x[1]**o+(0.841260061022)*x[1]
        arg[3,4,2,1]=(0.520110030788)*x[0]**o+(0.519262786324)*x[0]+(0.861221788948)*x[1]**o+(-0.376714934937)*x[1]
        ref[0,0,0,0]=(0.240887319278)*x_ref[0]**o+(-0.192166587545)*x_ref[0]+(0.0176974287409)*x_ref[1]**o+(0.100761748824)*x_ref[1]
        ref[0,0,0,1]=(-0.549649845775)*x_ref[0]**o+(-0.293969245646)*x_ref[0]+(-0.605885325133)*x_ref[1]**o+(0.799558451173)*x_ref[1]
        ref[0,0,1,0]=(-0.899658769582)*x_ref[0]**o+(0.484524193987)*x_ref[0]+(0.437485870351)*x_ref[1]**o+(-0.220549034927)*x_ref[1]
        ref[0,0,1,1]=(-0.146539961793)*x_ref[0]**o+(-0.278343486929)*x_ref[0]+(0.0913912456082)*x_ref[1]**o+(-0.751028975123)*x_ref[1]
        ref[0,0,2,0]=(0.806344701194)*x_ref[0]**o+(0.0537872824561)*x_ref[0]+(-0.70407828379)*x_ref[1]**o+(-0.922963486601)*x_ref[1]
        ref[0,0,2,1]=(0.307202902495)*x_ref[0]**o+(0.301603506731)*x_ref[0]+(0.232983743616)*x_ref[1]**o+(-0.744267542759)*x_ref[1]
        ref[0,1,0,0]=(-0.550616060824)*x_ref[0]**o+(-0.661367750254)*x_ref[0]+(-0.695905895864)*x_ref[1]**o+(-0.788045922398)*x_ref[1]
        ref[0,1,0,1]=(-0.934169134035)*x_ref[0]**o+(-0.843861105521)*x_ref[0]+(-0.237877254439)*x_ref[1]**o+(-0.364956821348)*x_ref[1]
        ref[0,1,1,0]=(0.0950675750026)*x_ref[0]**o+(-0.815510640594)*x_ref[0]+(0.72526035968)*x_ref[1]**o+(0.898675190106)*x_ref[1]
        ref[0,1,1,1]=(0.715609066776)*x_ref[0]**o+(0.371314845206)*x_ref[0]+(0.326878867244)*x_ref[1]**o+(-0.468798950199)*x_ref[1]
        ref[0,1,2,0]=(0.392175939547)*x_ref[0]**o+(0.105332682674)*x_ref[0]+(0.997967455477)*x_ref[1]**o+(0.0322491822496)*x_ref[1]
        ref[0,1,2,1]=(-0.804498399187)*x_ref[0]**o+(0.819653744739)*x_ref[0]+(0.609610573703)*x_ref[1]**o+(-0.623299047794)*x_ref[1]
        ref[0,2,0,0]=(0.993877917386)*x_ref[0]**o+(0.815952897899)*x_ref[0]+(-0.258432174486)*x_ref[1]**o+(0.209324387308)*x_ref[1]
        ref[0,2,0,1]=(0.395946889364)*x_ref[0]**o+(-0.859500539908)*x_ref[0]+(-0.68495734785)*x_ref[1]**o+(0.791813946302)*x_ref[1]
        ref[0,2,1,0]=(0.154347661108)*x_ref[0]**o+(0.0873140744322)*x_ref[0]+(-0.423935109956)*x_ref[1]**o+(-0.754297239226)*x_ref[1]
        ref[0,2,1,1]=(-0.623998419484)*x_ref[0]**o+(0.487300130941)*x_ref[0]+(-0.323613219784)*x_ref[1]**o+(0.744877983729)*x_ref[1]
        ref[0,2,2,0]=(0.515686374371)*x_ref[0]**o+(0.512572153346)*x_ref[0]+(-0.641245783272)*x_ref[1]**o+(-0.390660158759)*x_ref[1]
        ref[0,2,2,1]=(0.24398238139)*x_ref[0]**o+(-0.172020271678)*x_ref[0]+(0.980197812701)*x_ref[1]**o+(0.336941934503)*x_ref[1]
        ref[0,3,0,0]=(0.0154584447615)*x_ref[0]**o+(-0.547048191238)*x_ref[0]+(-0.555319223005)*x_ref[1]**o+(0.0972794598106)*x_ref[1]
        ref[0,3,0,1]=(0.44744389335)*x_ref[0]**o+(0.475046337874)*x_ref[0]+(0.876228982066)*x_ref[1]**o+(-0.0636405241175)*x_ref[1]
        ref[0,3,1,0]=(0.392948465298)*x_ref[0]**o+(-0.341342654182)*x_ref[0]+(0.747185683968)*x_ref[1]**o+(-0.99981172166)*x_ref[1]
        ref[0,3,1,1]=(-0.172355321621)*x_ref[0]**o+(-0.531391484795)*x_ref[0]+(0.451654274019)*x_ref[1]**o+(0.0795298667915)*x_ref[1]
        ref[0,3,2,0]=(0.0841642474674)*x_ref[0]**o+(0.308494435164)*x_ref[0]+(0.53181918181)*x_ref[1]**o+(0.480258067959)*x_ref[1]
        ref[0,3,2,1]=(-0.222903054864)*x_ref[0]**o+(-0.0330655869754)*x_ref[0]+(0.227482525015)*x_ref[1]**o+(0.536136212347)*x_ref[1]
        ref[0,4,0,0]=(0.574798381284)*x_ref[0]**o+(0.181354466484)*x_ref[0]+(0.103049514171)*x_ref[1]**o+(0.156282818835)*x_ref[1]
        ref[0,4,0,1]=(0.478314255289)*x_ref[0]**o+(0.224797222592)*x_ref[0]+(0.307668606272)*x_ref[1]**o+(-0.785262182937)*x_ref[1]
        ref[0,4,1,0]=(0.519356961182)*x_ref[0]**o+(0.646200458344)*x_ref[0]+(0.0522618783181)*x_ref[1]**o+(0.702335307324)*x_ref[1]
        ref[0,4,1,1]=(-0.168756029541)*x_ref[0]**o+(-0.00618364750666)*x_ref[0]+(0.472704616213)*x_ref[1]**o+(0.974257451871)*x_ref[1]
        ref[0,4,2,0]=(-0.279196869573)*x_ref[0]**o+(-0.333083174091)*x_ref[0]+(-0.980089574428)*x_ref[1]**o+(0.405273709405)*x_ref[1]
        ref[0,4,2,1]=(-0.199431380673)*x_ref[0]**o+(0.823294899866)*x_ref[0]+(-0.174125117685)*x_ref[1]**o+(-0.730500493887)*x_ref[1]
        ref[1,0,0,0]=(-0.577281489635)*x_ref[0]**o+(0.231906139898)*x_ref[0]+(0.537128708963)*x_ref[1]**o+(0.363444540898)*x_ref[1]
        ref[1,0,0,1]=(0.204108541882)*x_ref[0]**o+(-0.217661714888)*x_ref[0]+(0.963384293759)*x_ref[1]**o+(-0.0817519801351)*x_ref[1]
        ref[1,0,1,0]=(0.540681325439)*x_ref[0]**o+(-0.496036960808)*x_ref[0]+(0.421037689661)*x_ref[1]**o+(0.19399086996)*x_ref[1]
        ref[1,0,1,1]=(0.482938542126)*x_ref[0]**o+(-0.0725027675209)*x_ref[0]+(-0.441517962054)*x_ref[1]**o+(0.787445860714)*x_ref[1]
        ref[1,0,2,0]=(0.0378928111257)*x_ref[0]**o+(0.451368848019)*x_ref[0]+(-0.228668284485)*x_ref[1]**o+(0.505927229432)*x_ref[1]
        ref[1,0,2,1]=(-0.59833518872)*x_ref[0]**o+(0.956068253354)*x_ref[0]+(0.667905995469)*x_ref[1]**o+(0.963273292214)*x_ref[1]
        ref[1,1,0,0]=(0.901637865236)*x_ref[0]**o+(-0.642678971748)*x_ref[0]+(-0.603915019524)*x_ref[1]**o+(-0.208004993919)*x_ref[1]
        ref[1,1,0,1]=(-0.227162661052)*x_ref[0]**o+(-0.5816434303)*x_ref[0]+(0.274961293801)*x_ref[1]**o+(-0.614001328621)*x_ref[1]
        ref[1,1,1,0]=(0.078438078548)*x_ref[0]**o+(0.570191242883)*x_ref[0]+(0.0952878212484)*x_ref[1]**o+(0.844460205497)*x_ref[1]
        ref[1,1,1,1]=(-0.901445811786)*x_ref[0]**o+(-0.996376775344)*x_ref[0]+(0.707470471613)*x_ref[1]**o+(-0.187517936674)*x_ref[1]
        ref[1,1,2,0]=(-0.66891443067)*x_ref[0]**o+(0.0462057239324)*x_ref[0]+(-0.825820539128)*x_ref[1]**o+(-0.766932836843)*x_ref[1]
        ref[1,1,2,1]=(-0.851428828811)*x_ref[0]**o+(0.504484422346)*x_ref[0]+(0.121378315545)*x_ref[1]**o+(0.903947012757)*x_ref[1]
        ref[1,2,0,0]=(-0.393827449768)*x_ref[0]**o+(-0.251827904608)*x_ref[0]+(0.8320040882)*x_ref[1]**o+(0.57595149855)*x_ref[1]
        ref[1,2,0,1]=(0.638738110712)*x_ref[0]**o+(-0.439962152539)*x_ref[0]+(-0.977211078995)*x_ref[1]**o+(0.0528475212451)*x_ref[1]
        ref[1,2,1,0]=(0.824313588941)*x_ref[0]**o+(-0.589968996823)*x_ref[0]+(-0.512156794152)*x_ref[1]**o+(0.142809794443)*x_ref[1]
        ref[1,2,1,1]=(-0.31789333147)*x_ref[0]**o+(0.598839896949)*x_ref[0]+(0.638319398651)*x_ref[1]**o+(0.353625999317)*x_ref[1]
        ref[1,2,2,0]=(-0.720472365012)*x_ref[0]**o+(0.896538030051)*x_ref[0]+(-0.224639169972)*x_ref[1]**o+(-0.898299360738)*x_ref[1]
        ref[1,2,2,1]=(0.123077707167)*x_ref[0]**o+(0.15922851038)*x_ref[0]+(0.678984783843)*x_ref[1]**o+(-0.778188269852)*x_ref[1]
        ref[1,3,0,0]=(0.344256301209)*x_ref[0]**o+(0.884367738289)*x_ref[0]+(0.00556636219289)*x_ref[1]**o+(0.684361331195)*x_ref[1]
        ref[1,3,0,1]=(-0.812245131065)*x_ref[0]**o+(0.971365671228)*x_ref[0]+(0.0696053239569)*x_ref[1]**o+(-0.300266235138)*x_ref[1]
        ref[1,3,1,0]=(0.470254588058)*x_ref[0]**o+(0.0946825509913)*x_ref[0]+(-0.415619255025)*x_ref[1]**o+(-0.27381652858)*x_ref[1]
        ref[1,3,1,1]=(-0.84555138241)*x_ref[0]**o+(0.873510832091)*x_ref[0]+(-0.138010338152)*x_ref[1]**o+(-0.982274515316)*x_ref[1]
        ref[1,3,2,0]=(0.695614136148)*x_ref[0]**o+(-0.192588082771)*x_ref[0]+(-0.21579127791)*x_ref[1]**o+(-0.0335939736388)*x_ref[1]
        ref[1,3,2,1]=(-0.633363463896)*x_ref[0]**o+(-0.958320549993)*x_ref[0]+(0.197331067481)*x_ref[1]**o+(0.315089888409)*x_ref[1]
        ref[1,4,0,0]=(-0.211531219945)*x_ref[0]**o+(-0.399700721799)*x_ref[0]+(0.578927638839)*x_ref[1]**o+(-0.389034976195)*x_ref[1]
        ref[1,4,0,1]=(-0.846399571893)*x_ref[0]**o+(-0.952904755903)*x_ref[0]+(-0.858399802693)*x_ref[1]**o+(-0.23264354907)*x_ref[1]
        ref[1,4,1,0]=(0.255822441586)*x_ref[0]**o+(0.896237710393)*x_ref[0]+(-0.90248008271)*x_ref[1]**o+(0.654804848244)*x_ref[1]
        ref[1,4,1,1]=(-0.112475508163)*x_ref[0]**o+(-0.861697204453)*x_ref[0]+(-0.782581931187)*x_ref[1]**o+(-0.129345341581)*x_ref[1]
        ref[1,4,2,0]=(0.690095687033)*x_ref[0]**o+(0.729317334019)*x_ref[0]+(-0.00961483612288)*x_ref[1]**o+(-0.288320551243)*x_ref[1]
        ref[1,4,2,1]=(-0.648146433517)*x_ref[0]**o+(-0.301246601263)*x_ref[0]+(-0.490119612994)*x_ref[1]**o+(0.231186845615)*x_ref[1]
        ref[2,0,0,0]=(-0.645560891595)*x_ref[0]**o+(-0.889067544988)*x_ref[0]+(-0.193353414221)*x_ref[1]**o+(-0.726988479344)*x_ref[1]
        ref[2,0,0,1]=(-0.083614738484)*x_ref[0]**o+(-0.822530927241)*x_ref[0]+(0.660338934241)*x_ref[1]**o+(0.0383821036508)*x_ref[1]
        ref[2,0,1,0]=(-0.132250736664)*x_ref[0]**o+(-0.571474370931)*x_ref[0]+(-0.755931100342)*x_ref[1]**o+(-0.145196865115)*x_ref[1]
        ref[2,0,1,1]=(0.545219196727)*x_ref[0]**o+(-0.846368787913)*x_ref[0]+(-0.845706160842)*x_ref[1]**o+(0.550703610989)*x_ref[1]
        ref[2,0,2,0]=(-0.766303732195)*x_ref[0]**o+(0.156822893641)*x_ref[0]+(0.00954334254282)*x_ref[1]**o+(0.201506144561)*x_ref[1]
        ref[2,0,2,1]=(-0.274476449632)*x_ref[0]**o+(-0.361690468185)*x_ref[0]+(-0.816535747268)*x_ref[1]**o+(-0.473913811701)*x_ref[1]
        ref[2,1,0,0]=(-0.927592771482)*x_ref[0]**o+(0.534615269814)*x_ref[0]+(-0.571217133452)*x_ref[1]**o+(0.57448299216)*x_ref[1]
        ref[2,1,0,1]=(-0.298153953211)*x_ref[0]**o+(-0.143636909462)*x_ref[0]+(0.990660288864)*x_ref[1]**o+(-0.495864284763)*x_ref[1]
        ref[2,1,1,0]=(-0.110625727208)*x_ref[0]**o+(-0.221057774317)*x_ref[0]+(0.557195078358)*x_ref[1]**o+(0.803162691808)*x_ref[1]
        ref[2,1,1,1]=(-0.115589176775)*x_ref[0]**o+(0.08253351782)*x_ref[0]+(-0.595996958977)*x_ref[1]**o+(0.501435717501)*x_ref[1]
        ref[2,1,2,0]=(0.78764976658)*x_ref[0]**o+(0.966729831711)*x_ref[0]+(-0.585879575667)*x_ref[1]**o+(0.341148176793)*x_ref[1]
        ref[2,1,2,1]=(-0.551697945691)*x_ref[0]**o+(0.545857249323)*x_ref[0]+(-0.919321724495)*x_ref[1]**o+(-0.286043327415)*x_ref[1]
        ref[2,2,0,0]=(-0.669933890655)*x_ref[0]**o+(-0.344862118495)*x_ref[0]+(0.878331282597)*x_ref[1]**o+(0.881893760718)*x_ref[1]
        ref[2,2,0,1]=(-0.217876618414)*x_ref[0]**o+(-0.289118007199)*x_ref[0]+(0.513464171525)*x_ref[1]**o+(0.310055899102)*x_ref[1]
        ref[2,2,1,0]=(0.0472051379154)*x_ref[0]**o+(0.42455388451)*x_ref[0]+(-0.920571834891)*x_ref[1]**o+(-0.779095076537)*x_ref[1]
        ref[2,2,1,1]=(-0.299619304537)*x_ref[0]**o+(-0.289777287294)*x_ref[0]+(0.920505062988)*x_ref[1]**o+(-0.243924179095)*x_ref[1]
        ref[2,2,2,0]=(-0.819343530877)*x_ref[0]**o+(0.975953690875)*x_ref[0]+(-0.787239459241)*x_ref[1]**o+(-0.199900734691)*x_ref[1]
        ref[2,2,2,1]=(0.0980085780878)*x_ref[0]**o+(0.642946515451)*x_ref[0]+(0.951790535152)*x_ref[1]**o+(-0.253914817366)*x_ref[1]
        ref[2,3,0,0]=(-0.882075522824)*x_ref[0]**o+(0.200934502682)*x_ref[0]+(-0.83200464496)*x_ref[1]**o+(0.941479345375)*x_ref[1]
        ref[2,3,0,1]=(-0.615941782209)*x_ref[0]**o+(-0.693242388092)*x_ref[0]+(-0.640417278871)*x_ref[1]**o+(-0.346459094803)*x_ref[1]
        ref[2,3,1,0]=(-0.447919427952)*x_ref[0]**o+(0.205055689141)*x_ref[0]+(-0.453290032411)*x_ref[1]**o+(0.879062836521)*x_ref[1]
        ref[2,3,1,1]=(0.721133924016)*x_ref[0]**o+(0.71294558459)*x_ref[0]+(-0.250087045805)*x_ref[1]**o+(0.605146721129)*x_ref[1]
        ref[2,3,2,0]=(0.247782322921)*x_ref[0]**o+(-0.956367425119)*x_ref[0]+(-0.054045082005)*x_ref[1]**o+(-0.128151832132)*x_ref[1]
        ref[2,3,2,1]=(-0.427039259088)*x_ref[0]**o+(-0.359849242839)*x_ref[0]+(-0.244990934481)*x_ref[1]**o+(-0.038020035629)*x_ref[1]
        ref[2,4,0,0]=(-0.216409150834)*x_ref[0]**o+(0.492547989921)*x_ref[0]+(0.878317438469)*x_ref[1]**o+(-0.112718719119)*x_ref[1]
        ref[2,4,0,1]=(0.487403443102)*x_ref[0]**o+(0.228155519468)*x_ref[0]+(0.923150527516)*x_ref[1]**o+(0.642654867908)*x_ref[1]
        ref[2,4,1,0]=(0.916418483417)*x_ref[0]**o+(-0.177198935173)*x_ref[0]+(-0.790548940412)*x_ref[1]**o+(0.87540167052)*x_ref[1]
        ref[2,4,1,1]=(0.718502075538)*x_ref[0]**o+(0.691254010569)*x_ref[0]+(-0.0134430122729)*x_ref[1]**o+(-0.739888770749)*x_ref[1]
        ref[2,4,2,0]=(0.287825086861)*x_ref[0]**o+(0.596621809347)*x_ref[0]+(0.729784205391)*x_ref[1]**o+(-0.25984257955)*x_ref[1]
        ref[2,4,2,1]=(-0.515102480025)*x_ref[0]**o+(-0.609951840015)*x_ref[0]+(0.229834340075)*x_ref[1]**o+(-0.727157816241)*x_ref[1]
        ref[3,0,0,0]=(-0.349406091363)*x_ref[0]**o+(0.802348451103)*x_ref[0]+(-0.174421346276)*x_ref[1]**o+(0.120711198036)*x_ref[1]
        ref[3,0,0,1]=(0.367085169423)*x_ref[0]**o+(-0.319683236434)*x_ref[0]+(-0.590820542566)*x_ref[1]**o+(0.607625953303)*x_ref[1]
        ref[3,0,1,0]=(0.231259976433)*x_ref[0]**o+(-0.493297408825)*x_ref[0]+(0.919363640504)*x_ref[1]**o+(-0.839841270429)*x_ref[1]
        ref[3,0,1,1]=(0.57672135684)*x_ref[0]**o+(0.882062879704)*x_ref[0]+(-0.600979706444)*x_ref[1]**o+(-0.704008344509)*x_ref[1]
        ref[3,0,2,0]=(-0.208162114527)*x_ref[0]**o+(0.593682243052)*x_ref[0]+(-0.428436832487)*x_ref[1]**o+(0.51007021604)*x_ref[1]
        ref[3,0,2,1]=(0.315264672653)*x_ref[0]**o+(-0.0835897376534)*x_ref[0]+(-0.135315173106)*x_ref[1]**o+(-0.646372309498)*x_ref[1]
        ref[3,1,0,0]=(0.815505953752)*x_ref[0]**o+(0.0625493643292)*x_ref[0]+(0.383118433815)*x_ref[1]**o+(0.17372720435)*x_ref[1]
        ref[3,1,0,1]=(0.557295786969)*x_ref[0]**o+(-0.568823508165)*x_ref[0]+(-0.00874733267675)*x_ref[1]**o+(-0.27454140598)*x_ref[1]
        ref[3,1,1,0]=(0.649021026591)*x_ref[0]**o+(0.232255030448)*x_ref[0]+(0.845939296425)*x_ref[1]**o+(0.116387378294)*x_ref[1]
        ref[3,1,1,1]=(0.431124940343)*x_ref[0]**o+(-0.0865937271969)*x_ref[0]+(0.937911741288)*x_ref[1]**o+(0.362850039234)*x_ref[1]
        ref[3,1,2,0]=(0.155965414439)*x_ref[0]**o+(0.110850421841)*x_ref[0]+(0.244563080191)*x_ref[1]**o+(0.266325498006)*x_ref[1]
        ref[3,1,2,1]=(0.766813300935)*x_ref[0]**o+(0.77492422889)*x_ref[0]+(-0.0141436554714)*x_ref[1]**o+(0.780444449009)*x_ref[1]
        ref[3,2,0,0]=(0.619086019019)*x_ref[0]**o+(-0.574130567601)*x_ref[0]+(-0.417708159593)*x_ref[1]**o+(0.707727284254)*x_ref[1]
        ref[3,2,0,1]=(-0.845657409353)*x_ref[0]**o+(0.705282507748)*x_ref[0]+(0.124557418112)*x_ref[1]**o+(-0.337493494733)*x_ref[1]
        ref[3,2,1,0]=(0.114875146211)*x_ref[0]**o+(-0.724480426892)*x_ref[0]+(-0.0763476904405)*x_ref[1]**o+(0.12064907114)*x_ref[1]
        ref[3,2,1,1]=(-0.74994413376)*x_ref[0]**o+(-0.916476035217)*x_ref[0]+(-0.360412756599)*x_ref[1]**o+(-0.917457363098)*x_ref[1]
        ref[3,2,2,0]=(-0.446839470984)*x_ref[0]**o+(-0.097704838316)*x_ref[0]+(-0.653584364909)*x_ref[1]**o+(0.561014757758)*x_ref[1]
        ref[3,2,2,1]=(-0.178975301038)*x_ref[0]**o+(-0.843669982667)*x_ref[0]+(-0.566610094242)*x_ref[1]**o+(0.213047649003)*x_ref[1]
        ref[3,3,0,0]=(0.144115764355)*x_ref[0]**o+(0.199435411954)*x_ref[0]+(0.173334385199)*x_ref[1]**o+(-0.942957035808)*x_ref[1]
        ref[3,3,0,1]=(-0.417589926342)*x_ref[0]**o+(-0.819188022852)*x_ref[0]+(-0.68299160003)*x_ref[1]**o+(-0.314851447894)*x_ref[1]
        ref[3,3,1,0]=(0.737898219164)*x_ref[0]**o+(0.612906277393)*x_ref[0]+(-0.768349396729)*x_ref[1]**o+(-0.192998755095)*x_ref[1]
        ref[3,3,1,1]=(0.35433362007)*x_ref[0]**o+(0.894174225208)*x_ref[0]+(-0.450333030837)*x_ref[1]**o+(0.819450589483)*x_ref[1]
        ref[3,3,2,0]=(-0.455377131302)*x_ref[0]**o+(-0.748183144687)*x_ref[0]+(-0.482739450231)*x_ref[1]**o+(0.300625929022)*x_ref[1]
        ref[3,3,2,1]=(-0.602766693698)*x_ref[0]**o+(-0.416288436985)*x_ref[0]+(0.837501983536)*x_ref[1]**o+(-0.745935308188)*x_ref[1]
        ref[3,4,0,0]=(-0.706774010405)*x_ref[0]**o+(0.407766818701)*x_ref[0]+(-0.248964401443)*x_ref[1]**o+(0.675466487229)*x_ref[1]
        ref[3,4,0,1]=(0.512302883876)*x_ref[0]**o+(0.0809544867105)*x_ref[0]+(-0.204555203536)*x_ref[1]**o+(-0.628574073749)*x_ref[1]
        ref[3,4,1,0]=(0.126376796421)*x_ref[0]**o+(0.174607814129)*x_ref[0]+(0.68867194136)*x_ref[1]**o+(0.940025934568)*x_ref[1]
        ref[3,4,1,1]=(0.583042444739)*x_ref[0]**o+(-0.786981763836)*x_ref[0]+(0.319397032405)*x_ref[1]**o+(-0.324993351432)*x_ref[1]
        ref[3,4,2,0]=(-0.941997068836)*x_ref[0]**o+(-0.892825520378)*x_ref[0]+(0.275995419027)*x_ref[1]**o+(0.841260061022)*x_ref[1]
        ref[3,4,2,1]=(0.520110030788)*x_ref[0]**o+(0.519262786324)*x_ref[0]+(0.861221788948)*x_ref[1]**o+(-0.376714934937)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.615899511086)*x[0]**o+(0.823637401824)*x[0]+(-0.986967740198)*x[1]**o+(-0.782088977641)*x[1]+(-0.576799974679)*x[2]**o+(-0.910942885711)*x[2]
        arg[0,0,0,1]=(-0.790830593449)*x[0]**o+(-0.897724279499)*x[0]+(-0.881691805807)*x[1]**o+(-0.133793238901)*x[1]+(-0.0903391684489)*x[2]**o+(0.659893469877)*x[2]
        arg[0,0,1,0]=(-0.0567117900114)*x[0]**o+(-0.858031985686)*x[0]+(-0.397443445702)*x[1]**o+(0.549982838823)*x[1]+(-0.636905143813)*x[2]**o+(0.8488142632)*x[2]
        arg[0,0,1,1]=(-0.644557373082)*x[0]**o+(-0.323654898522)*x[0]+(0.210458705923)*x[1]**o+(0.990436827296)*x[1]+(-0.597872757151)*x[2]**o+(-0.83865548618)*x[2]
        arg[0,0,2,0]=(-0.29830667084)*x[0]**o+(0.327035546909)*x[0]+(0.132690049791)*x[1]**o+(-0.55625443282)*x[1]+(0.0972781385283)*x[2]**o+(0.151885837608)*x[2]
        arg[0,0,2,1]=(-0.435983465578)*x[0]**o+(-0.0852522953446)*x[0]+(0.257654807094)*x[1]**o+(-0.336603197583)*x[1]+(0.637913421592)*x[2]**o+(0.8151042055)*x[2]
        arg[0,1,0,0]=(-0.0373980191125)*x[0]**o+(-0.45937285505)*x[0]+(-0.857933684889)*x[1]**o+(-0.843826686256)*x[1]+(-0.727077320504)*x[2]**o+(-0.65222863395)*x[2]
        arg[0,1,0,1]=(0.63426017384)*x[0]**o+(0.845708641601)*x[0]+(0.246160302486)*x[1]**o+(-0.660855572094)*x[1]+(0.67420981328)*x[2]**o+(-0.280803968781)*x[2]
        arg[0,1,1,0]=(0.881193264597)*x[0]**o+(-0.881936969515)*x[0]+(-0.883246837152)*x[1]**o+(0.182222471927)*x[1]+(0.405380496272)*x[2]**o+(0.175721746883)*x[2]
        arg[0,1,1,1]=(-0.44447667237)*x[0]**o+(0.956169274606)*x[0]+(0.0370532940652)*x[1]**o+(-0.181239635583)*x[1]+(-0.678412161285)*x[2]**o+(-0.653172157401)*x[2]
        arg[0,1,2,0]=(-0.50038634718)*x[0]**o+(0.00278321765167)*x[0]+(-0.831539460218)*x[1]**o+(-0.555243320385)*x[1]+(-0.626789452003)*x[2]**o+(-0.495085243933)*x[2]
        arg[0,1,2,1]=(-0.32886359569)*x[0]**o+(0.406099965475)*x[0]+(0.764756177266)*x[1]**o+(-0.272070344181)*x[1]+(-0.41812596516)*x[2]**o+(0.349063696376)*x[2]
        arg[0,2,0,0]=(-0.580544978365)*x[0]**o+(0.53441241334)*x[0]+(-0.808749795876)*x[1]**o+(0.172110801632)*x[1]+(-0.939262606098)*x[2]**o+(-0.724859325104)*x[2]
        arg[0,2,0,1]=(0.256335221944)*x[0]**o+(0.463522371018)*x[0]+(0.990516478508)*x[1]**o+(0.423931177054)*x[1]+(0.184277693683)*x[2]**o+(-0.272967447689)*x[2]
        arg[0,2,1,0]=(-0.126943811008)*x[0]**o+(0.645821387926)*x[0]+(0.150370174987)*x[1]**o+(-0.150954538301)*x[1]+(-0.378720078559)*x[2]**o+(0.0975963807109)*x[2]
        arg[0,2,1,1]=(-0.451434544339)*x[0]**o+(0.557314109088)*x[0]+(-0.0402970749558)*x[1]**o+(0.306984557206)*x[1]+(-0.297018139186)*x[2]**o+(0.593800698785)*x[2]
        arg[0,2,2,0]=(-0.376199144911)*x[0]**o+(-0.230454116373)*x[0]+(0.992738297651)*x[1]**o+(0.183510534927)*x[1]+(0.409381711992)*x[2]**o+(0.171945265727)*x[2]
        arg[0,2,2,1]=(0.363233095971)*x[0]**o+(-0.130589148249)*x[0]+(0.809611042018)*x[1]**o+(-0.354546307961)*x[1]+(0.0631410169203)*x[2]**o+(0.630306161781)*x[2]
        arg[0,3,0,0]=(0.49655016161)*x[0]**o+(0.878957258614)*x[0]+(-0.397408696784)*x[1]**o+(-0.199445321288)*x[1]+(0.929857086663)*x[2]**o+(-0.648770887161)*x[2]
        arg[0,3,0,1]=(-0.85578334594)*x[0]**o+(-0.294617109229)*x[0]+(-0.0790295769768)*x[1]**o+(0.82151942311)*x[1]+(0.541628117444)*x[2]**o+(0.805418174827)*x[2]
        arg[0,3,1,0]=(0.800775454804)*x[0]**o+(0.599550462983)*x[0]+(-0.384788634082)*x[1]**o+(-0.356933972363)*x[1]+(-0.423116265473)*x[2]**o+(-0.125492982003)*x[2]
        arg[0,3,1,1]=(0.929652994661)*x[0]**o+(-0.57396682406)*x[0]+(0.313584116036)*x[1]**o+(-0.190575043496)*x[1]+(-0.29110884935)*x[2]**o+(0.967530639925)*x[2]
        arg[0,3,2,0]=(0.761136478868)*x[0]**o+(0.565721170014)*x[0]+(0.239619122131)*x[1]**o+(-0.0821651979956)*x[1]+(-0.699218775545)*x[2]**o+(0.19973415585)*x[2]
        arg[0,3,2,1]=(0.209543244648)*x[0]**o+(0.93957796874)*x[0]+(0.354405887426)*x[1]**o+(0.504177262864)*x[1]+(0.446866674894)*x[2]**o+(-0.522936485464)*x[2]
        arg[0,4,0,0]=(0.427541399652)*x[0]**o+(-0.114806067206)*x[0]+(0.236780877643)*x[1]**o+(0.422226381272)*x[1]+(-0.506450961214)*x[2]**o+(0.616091720212)*x[2]
        arg[0,4,0,1]=(0.698773571323)*x[0]**o+(-0.357878713085)*x[0]+(-0.453730359129)*x[1]**o+(-0.328701041355)*x[1]+(-0.28646346679)*x[2]**o+(-0.0449784265203)*x[2]
        arg[0,4,1,0]=(-0.419416306509)*x[0]**o+(-0.1970394955)*x[0]+(-0.814008932379)*x[1]**o+(-0.805127589929)*x[1]+(0.244059789875)*x[2]**o+(-0.983027843847)*x[2]
        arg[0,4,1,1]=(0.84837346004)*x[0]**o+(0.501808510624)*x[0]+(-0.105800351358)*x[1]**o+(0.903999231718)*x[1]+(-0.0473996978124)*x[2]**o+(0.683647253381)*x[2]
        arg[0,4,2,0]=(0.658722899682)*x[0]**o+(-0.401846635016)*x[0]+(-0.967322202581)*x[1]**o+(-0.948306922217)*x[1]+(0.496409308219)*x[2]**o+(-0.184768777052)*x[2]
        arg[0,4,2,1]=(-0.454468390888)*x[0]**o+(-0.220640468215)*x[0]+(0.995269925943)*x[1]**o+(0.778951329211)*x[1]+(0.0439534466302)*x[2]**o+(0.806197038383)*x[2]
        arg[1,0,0,0]=(-0.547461802236)*x[0]**o+(0.370768827789)*x[0]+(-0.619991757552)*x[1]**o+(0.391727195645)*x[1]+(-0.649274922953)*x[2]**o+(-0.478788248429)*x[2]
        arg[1,0,0,1]=(-0.899230505882)*x[0]**o+(0.603064321735)*x[0]+(-0.737475386349)*x[1]**o+(0.594197953001)*x[1]+(0.341324049064)*x[2]**o+(0.285202633917)*x[2]
        arg[1,0,1,0]=(-0.107063589618)*x[0]**o+(-0.260491562358)*x[0]+(-0.667059815618)*x[1]**o+(0.473227354537)*x[1]+(-0.587170510558)*x[2]**o+(-0.740299218393)*x[2]
        arg[1,0,1,1]=(0.87705715098)*x[0]**o+(0.312753471827)*x[0]+(-0.307530283315)*x[1]**o+(0.195564644822)*x[1]+(-0.64384485166)*x[2]**o+(-0.776362951729)*x[2]
        arg[1,0,2,0]=(0.0949408564739)*x[0]**o+(0.765922179855)*x[0]+(0.441215568253)*x[1]**o+(0.339112969437)*x[1]+(-0.0774615292449)*x[2]**o+(0.673258055783)*x[2]
        arg[1,0,2,1]=(0.114828212302)*x[0]**o+(0.0831258766277)*x[0]+(0.674540402248)*x[1]**o+(-0.304921039094)*x[1]+(-0.725514770578)*x[2]**o+(-0.627584773746)*x[2]
        arg[1,1,0,0]=(0.291561714675)*x[0]**o+(0.898587624783)*x[0]+(0.476696264545)*x[1]**o+(-0.454648604978)*x[1]+(0.677949150743)*x[2]**o+(-0.59438289851)*x[2]
        arg[1,1,0,1]=(0.356687258009)*x[0]**o+(0.806236251629)*x[0]+(0.843636690354)*x[1]**o+(-0.649776947181)*x[1]+(0.208542122388)*x[2]**o+(-0.0726045573458)*x[2]
        arg[1,1,1,0]=(0.712800450558)*x[0]**o+(0.748271875755)*x[0]+(0.403880649356)*x[1]**o+(0.956549336754)*x[1]+(0.0420128495043)*x[2]**o+(-0.83177870499)*x[2]
        arg[1,1,1,1]=(-0.886862904174)*x[0]**o+(-0.03822986594)*x[0]+(0.587151517496)*x[1]**o+(-0.120272060898)*x[1]+(-0.573366391933)*x[2]**o+(-0.59772763241)*x[2]
        arg[1,1,2,0]=(0.311960197804)*x[0]**o+(0.486854949444)*x[0]+(-0.647995651954)*x[1]**o+(0.30104277964)*x[1]+(-0.174718703748)*x[2]**o+(-0.506993216295)*x[2]
        arg[1,1,2,1]=(0.979056206991)*x[0]**o+(0.892104464812)*x[0]+(0.0507432703245)*x[1]**o+(0.103630858454)*x[1]+(-0.0254965952266)*x[2]**o+(0.13034027239)*x[2]
        arg[1,2,0,0]=(0.259797966548)*x[0]**o+(0.53018999371)*x[0]+(0.533745722992)*x[1]**o+(0.53029084173)*x[1]+(-0.595126791061)*x[2]**o+(-0.51668295558)*x[2]
        arg[1,2,0,1]=(0.926395590808)*x[0]**o+(-0.0700141207179)*x[0]+(-0.973392454012)*x[1]**o+(-0.46204067097)*x[1]+(0.814964497207)*x[2]**o+(-0.183546091376)*x[2]
        arg[1,2,1,0]=(-0.0872635441658)*x[0]**o+(-0.0307630839035)*x[0]+(0.539798653095)*x[1]**o+(0.966367734848)*x[1]+(-0.817495053612)*x[2]**o+(-0.268397867848)*x[2]
        arg[1,2,1,1]=(-0.933074613921)*x[0]**o+(0.48896533938)*x[0]+(0.453696133488)*x[1]**o+(-0.238730406749)*x[1]+(-0.696785951086)*x[2]**o+(-0.359858385642)*x[2]
        arg[1,2,2,0]=(-0.844406286809)*x[0]**o+(-0.506332243946)*x[0]+(0.976436731658)*x[1]**o+(-0.201489161857)*x[1]+(0.851236173166)*x[2]**o+(-0.998803612947)*x[2]
        arg[1,2,2,1]=(0.259707321449)*x[0]**o+(-0.888149478655)*x[0]+(-0.273502842869)*x[1]**o+(0.291193582158)*x[1]+(-0.897542178027)*x[2]**o+(-0.345994824429)*x[2]
        arg[1,3,0,0]=(0.961298984697)*x[0]**o+(0.248141115092)*x[0]+(-0.623053025178)*x[1]**o+(-0.775095452078)*x[1]+(0.955362707049)*x[2]**o+(-0.517547333705)*x[2]
        arg[1,3,0,1]=(0.858098596227)*x[0]**o+(0.309307483873)*x[0]+(-0.209217034755)*x[1]**o+(0.783803593379)*x[1]+(0.138164128092)*x[2]**o+(-0.922854003461)*x[2]
        arg[1,3,1,0]=(-0.139017942141)*x[0]**o+(-0.338223039119)*x[0]+(0.737882941498)*x[1]**o+(0.948854903087)*x[1]+(-0.0792461633266)*x[2]**o+(0.848875660203)*x[2]
        arg[1,3,1,1]=(0.693284332212)*x[0]**o+(-0.392272875348)*x[0]+(-0.475709045479)*x[1]**o+(0.182948077236)*x[1]+(0.0277569761538)*x[2]**o+(-0.550294947596)*x[2]
        arg[1,3,2,0]=(-0.105381649505)*x[0]**o+(-0.403051690238)*x[0]+(-0.722219735818)*x[1]**o+(-0.380114388869)*x[1]+(0.321452572491)*x[2]**o+(0.858583245253)*x[2]
        arg[1,3,2,1]=(-0.405370062024)*x[0]**o+(0.0743704840047)*x[0]+(0.998947017815)*x[1]**o+(0.983611079522)*x[1]+(0.854202982674)*x[2]**o+(0.288478746498)*x[2]
        arg[1,4,0,0]=(0.0584782489998)*x[0]**o+(0.999286927604)*x[0]+(-0.458166020555)*x[1]**o+(0.529008844128)*x[1]+(0.424592641922)*x[2]**o+(-0.489692340187)*x[2]
        arg[1,4,0,1]=(-0.150555966138)*x[0]**o+(0.362589320838)*x[0]+(0.515820915391)*x[1]**o+(0.742403661312)*x[1]+(0.230278843183)*x[2]**o+(-0.213977724366)*x[2]
        arg[1,4,1,0]=(-0.361161802362)*x[0]**o+(-0.49718332222)*x[0]+(0.143956046197)*x[1]**o+(-0.044550912758)*x[1]+(-0.299892874708)*x[2]**o+(0.370248709792)*x[2]
        arg[1,4,1,1]=(0.259365238173)*x[0]**o+(-0.474694548557)*x[0]+(0.732983396669)*x[1]**o+(-0.295151546774)*x[1]+(-0.337144628526)*x[2]**o+(0.664201775714)*x[2]
        arg[1,4,2,0]=(-0.318016754284)*x[0]**o+(0.555291443852)*x[0]+(0.538150727998)*x[1]**o+(-0.800118353132)*x[1]+(0.652262386596)*x[2]**o+(0.303156645857)*x[2]
        arg[1,4,2,1]=(0.0827142921796)*x[0]**o+(-0.88498648372)*x[0]+(-0.941249844349)*x[1]**o+(-0.162576659039)*x[1]+(0.867621252192)*x[2]**o+(0.0834583795227)*x[2]
        arg[2,0,0,0]=(0.0670088199264)*x[0]**o+(-0.0425961894346)*x[0]+(0.329605789302)*x[1]**o+(0.860453156363)*x[1]+(-0.37410469727)*x[2]**o+(0.55476432384)*x[2]
        arg[2,0,0,1]=(0.801738221879)*x[0]**o+(-0.836563530923)*x[0]+(-0.449651673142)*x[1]**o+(-0.19423133918)*x[1]+(0.947648200044)*x[2]**o+(0.0854067601423)*x[2]
        arg[2,0,1,0]=(0.986595046855)*x[0]**o+(-0.358319352175)*x[0]+(0.493917713513)*x[1]**o+(-0.92052282445)*x[1]+(-0.692726467259)*x[2]**o+(-0.0636943737807)*x[2]
        arg[2,0,1,1]=(0.38491150492)*x[0]**o+(0.766893245938)*x[0]+(0.420188917945)*x[1]**o+(0.43760538854)*x[1]+(-0.770749451981)*x[2]**o+(0.124882164378)*x[2]
        arg[2,0,2,0]=(0.775715718797)*x[0]**o+(0.89516703208)*x[0]+(-0.829021524898)*x[1]**o+(-0.650262454311)*x[1]+(-0.812556336598)*x[2]**o+(-0.997075743997)*x[2]
        arg[2,0,2,1]=(0.3503761008)*x[0]**o+(-0.822009897455)*x[0]+(-0.657657365354)*x[1]**o+(0.476295612156)*x[1]+(0.315554008711)*x[2]**o+(0.691631078049)*x[2]
        arg[2,1,0,0]=(0.680177375433)*x[0]**o+(-0.922726813311)*x[0]+(-0.581116746321)*x[1]**o+(0.628138099343)*x[1]+(-0.568262800293)*x[2]**o+(-0.228201382395)*x[2]
        arg[2,1,0,1]=(0.210094833324)*x[0]**o+(0.281549673876)*x[0]+(0.23785231073)*x[1]**o+(-0.103812455993)*x[1]+(0.530445971191)*x[2]**o+(-0.0830520449939)*x[2]
        arg[2,1,1,0]=(0.367133297563)*x[0]**o+(0.168289954585)*x[0]+(0.614117392787)*x[1]**o+(0.790467766841)*x[1]+(-0.958002150805)*x[2]**o+(0.328916248824)*x[2]
        arg[2,1,1,1]=(0.089535347231)*x[0]**o+(-0.93155993701)*x[0]+(-0.674670760794)*x[1]**o+(-0.274148261773)*x[1]+(-0.807020126056)*x[2]**o+(-0.323334769496)*x[2]
        arg[2,1,2,0]=(-0.948679835219)*x[0]**o+(-0.577235742425)*x[0]+(-0.732584834146)*x[1]**o+(-0.475242504288)*x[1]+(0.213663223262)*x[2]**o+(-0.151401354381)*x[2]
        arg[2,1,2,1]=(0.264513246206)*x[0]**o+(-0.605806437391)*x[0]+(-0.058730926693)*x[1]**o+(-0.661696304412)*x[1]+(-0.361277352457)*x[2]**o+(-0.804408258253)*x[2]
        arg[2,2,0,0]=(-0.144562358054)*x[0]**o+(-0.0960550555725)*x[0]+(-0.905096867803)*x[1]**o+(0.188861784103)*x[1]+(-0.616751005568)*x[2]**o+(0.604711974248)*x[2]
        arg[2,2,0,1]=(-0.0582524307858)*x[0]**o+(-0.503648879785)*x[0]+(-0.682755209956)*x[1]**o+(-0.808213164425)*x[1]+(0.539987276327)*x[2]**o+(0.650557703689)*x[2]
        arg[2,2,1,0]=(0.0607729567733)*x[0]**o+(-0.498819638428)*x[0]+(0.434442595334)*x[1]**o+(-0.491063977603)*x[1]+(0.499436212229)*x[2]**o+(-0.932071991206)*x[2]
        arg[2,2,1,1]=(-0.461295295294)*x[0]**o+(-0.153600767916)*x[0]+(-0.86452392231)*x[1]**o+(0.136760034153)*x[1]+(-0.471642416082)*x[2]**o+(0.453820403438)*x[2]
        arg[2,2,2,0]=(-0.970051888584)*x[0]**o+(0.415728825057)*x[0]+(0.971643909569)*x[1]**o+(0.923719420524)*x[1]+(-0.837313513804)*x[2]**o+(-0.738057676093)*x[2]
        arg[2,2,2,1]=(-0.941355918089)*x[0]**o+(-0.9629784524)*x[0]+(-0.931184348072)*x[1]**o+(0.720006097658)*x[1]+(-0.140750567978)*x[2]**o+(0.55381939119)*x[2]
        arg[2,3,0,0]=(0.753455933669)*x[0]**o+(0.352384922037)*x[0]+(0.737644348002)*x[1]**o+(-0.565144188248)*x[1]+(0.257417624014)*x[2]**o+(-0.660006641683)*x[2]
        arg[2,3,0,1]=(0.886996770925)*x[0]**o+(-0.876595418252)*x[0]+(-0.189778073871)*x[1]**o+(0.560867442336)*x[1]+(-0.262338156346)*x[2]**o+(0.614749098901)*x[2]
        arg[2,3,1,0]=(-0.722912374016)*x[0]**o+(0.447451310273)*x[0]+(0.397510140568)*x[1]**o+(0.932405917932)*x[1]+(-0.206969252336)*x[2]**o+(-0.897366208888)*x[2]
        arg[2,3,1,1]=(0.106324074547)*x[0]**o+(0.822109425884)*x[0]+(0.62699223277)*x[1]**o+(0.238371557839)*x[1]+(-0.422051841243)*x[2]**o+(0.868932130329)*x[2]
        arg[2,3,2,0]=(0.0155790371581)*x[0]**o+(-0.696029577918)*x[0]+(0.259909585217)*x[1]**o+(0.260209121862)*x[1]+(0.780977081459)*x[2]**o+(-0.819957534538)*x[2]
        arg[2,3,2,1]=(0.736223721904)*x[0]**o+(-0.433132598174)*x[0]+(-0.880913649812)*x[1]**o+(-0.375988777461)*x[1]+(-0.351763406307)*x[2]**o+(0.948468252467)*x[2]
        arg[2,4,0,0]=(0.943349180556)*x[0]**o+(-0.720044514103)*x[0]+(0.671900161807)*x[1]**o+(0.358571142942)*x[1]+(-0.132755998648)*x[2]**o+(0.0749815044365)*x[2]
        arg[2,4,0,1]=(0.639070752347)*x[0]**o+(0.551441363406)*x[0]+(0.507547844432)*x[1]**o+(-0.800043772307)*x[1]+(-0.506359503394)*x[2]**o+(0.952191432651)*x[2]
        arg[2,4,1,0]=(-0.655864732254)*x[0]**o+(0.142198390642)*x[0]+(0.075292486702)*x[1]**o+(0.191013189029)*x[1]+(0.106768453247)*x[2]**o+(0.974526515449)*x[2]
        arg[2,4,1,1]=(-0.354750205648)*x[0]**o+(0.141506849368)*x[0]+(-0.170074898141)*x[1]**o+(-0.0882776529335)*x[1]+(0.845254802876)*x[2]**o+(-0.435414692125)*x[2]
        arg[2,4,2,0]=(-0.412836027163)*x[0]**o+(0.830777132582)*x[0]+(-0.87554952236)*x[1]**o+(0.970647516313)*x[1]+(0.989864827286)*x[2]**o+(-0.67790755421)*x[2]
        arg[2,4,2,1]=(0.762907045031)*x[0]**o+(0.140771931118)*x[0]+(0.716859940142)*x[1]**o+(0.151769314189)*x[1]+(-0.812408853675)*x[2]**o+(0.755727078666)*x[2]
        arg[3,0,0,0]=(-0.13358237438)*x[0]**o+(-0.594735856789)*x[0]+(-0.367418402298)*x[1]**o+(0.5349162872)*x[1]+(0.238285320118)*x[2]**o+(0.999627906459)*x[2]
        arg[3,0,0,1]=(-0.537433369218)*x[0]**o+(0.133034652667)*x[0]+(-0.190302464596)*x[1]**o+(0.869919012762)*x[1]+(0.320294306008)*x[2]**o+(0.865908856629)*x[2]
        arg[3,0,1,0]=(0.480851043043)*x[0]**o+(0.833358259022)*x[0]+(0.893051476621)*x[1]**o+(-0.505255581199)*x[1]+(-0.15316145317)*x[2]**o+(-0.937103563776)*x[2]
        arg[3,0,1,1]=(0.765360137605)*x[0]**o+(0.502399552842)*x[0]+(-0.215707512276)*x[1]**o+(-0.402720816967)*x[1]+(0.231360098578)*x[2]**o+(0.732903327067)*x[2]
        arg[3,0,2,0]=(0.120836669045)*x[0]**o+(0.198346794548)*x[0]+(-0.28708447536)*x[1]**o+(0.289672075353)*x[1]+(-0.0305777738322)*x[2]**o+(-0.0686651569373)*x[2]
        arg[3,0,2,1]=(0.0574598844614)*x[0]**o+(0.031092492399)*x[0]+(-0.406805507341)*x[1]**o+(-0.623370509947)*x[1]+(-0.476094160966)*x[2]**o+(0.0840271758875)*x[2]
        arg[3,1,0,0]=(-0.569415249126)*x[0]**o+(-0.339675540399)*x[0]+(0.182536980707)*x[1]**o+(0.727480200453)*x[1]+(0.851672551504)*x[2]**o+(0.817312313075)*x[2]
        arg[3,1,0,1]=(0.921422642904)*x[0]**o+(0.697656143906)*x[0]+(0.969442847752)*x[1]**o+(-0.00957994710064)*x[1]+(-0.315230556399)*x[2]**o+(-0.870759158384)*x[2]
        arg[3,1,1,0]=(-0.813310465463)*x[0]**o+(-0.809362802903)*x[0]+(-0.0685101056428)*x[1]**o+(0.708195651847)*x[1]+(-0.876449456312)*x[2]**o+(0.882503536479)*x[2]
        arg[3,1,1,1]=(-0.277504922905)*x[0]**o+(-0.819428706347)*x[0]+(0.463203258731)*x[1]**o+(0.597493096401)*x[1]+(0.630522187524)*x[2]**o+(0.92480079466)*x[2]
        arg[3,1,2,0]=(-0.703167803764)*x[0]**o+(0.730459205382)*x[0]+(-0.507036921123)*x[1]**o+(0.419358545765)*x[1]+(-0.542758923679)*x[2]**o+(0.831844312917)*x[2]
        arg[3,1,2,1]=(0.135200299704)*x[0]**o+(-0.160734675188)*x[0]+(-0.893851689192)*x[1]**o+(0.748734906185)*x[1]+(0.970464447167)*x[2]**o+(0.277911376757)*x[2]
        arg[3,2,0,0]=(0.750104328784)*x[0]**o+(-0.818350593893)*x[0]+(-0.749900854281)*x[1]**o+(0.289117653816)*x[1]+(-0.219904822162)*x[2]**o+(0.805208050452)*x[2]
        arg[3,2,0,1]=(-0.129614389749)*x[0]**o+(-0.64067443816)*x[0]+(0.513013650233)*x[1]**o+(0.868869035341)*x[1]+(-0.040145121639)*x[2]**o+(-0.282350136119)*x[2]
        arg[3,2,1,0]=(-0.172814929104)*x[0]**o+(-0.296930471915)*x[0]+(-0.796114111585)*x[1]**o+(-0.210181080249)*x[1]+(-0.606107841173)*x[2]**o+(-0.954329084175)*x[2]
        arg[3,2,1,1]=(0.136149261959)*x[0]**o+(-0.613980696755)*x[0]+(0.968430714778)*x[1]**o+(0.683553952099)*x[1]+(-0.321383789121)*x[2]**o+(-0.139570984234)*x[2]
        arg[3,2,2,0]=(-0.273321720654)*x[0]**o+(0.963651110162)*x[0]+(-0.847402611499)*x[1]**o+(0.237131262905)*x[1]+(-0.617296821558)*x[2]**o+(-0.681856394629)*x[2]
        arg[3,2,2,1]=(-0.760071928624)*x[0]**o+(0.0345053779755)*x[0]+(-0.583680069732)*x[1]**o+(0.150497658077)*x[1]+(0.0257096621429)*x[2]**o+(-0.159406203385)*x[2]
        arg[3,3,0,0]=(-0.0668501746947)*x[0]**o+(-0.183645899611)*x[0]+(0.123667845765)*x[1]**o+(0.0928142420352)*x[1]+(-0.933250766317)*x[2]**o+(-0.629986949639)*x[2]
        arg[3,3,0,1]=(-0.0212790928206)*x[0]**o+(0.444036360325)*x[0]+(0.09182358127)*x[1]**o+(0.473088563672)*x[1]+(0.0852925382953)*x[2]**o+(-0.238517929786)*x[2]
        arg[3,3,1,0]=(0.0647235093728)*x[0]**o+(-0.266285898716)*x[0]+(0.944999560866)*x[1]**o+(-0.757696716486)*x[1]+(-0.439165278732)*x[2]**o+(0.960091479774)*x[2]
        arg[3,3,1,1]=(-0.792249238894)*x[0]**o+(-0.568585439587)*x[0]+(-0.0426655335607)*x[1]**o+(0.980004032206)*x[1]+(0.635227812889)*x[2]**o+(0.6687165384)*x[2]
        arg[3,3,2,0]=(0.340468442769)*x[0]**o+(0.791071154735)*x[0]+(-0.817919975547)*x[1]**o+(0.0328383066982)*x[1]+(-0.323093904705)*x[2]**o+(-0.302641620002)*x[2]
        arg[3,3,2,1]=(-0.476501733213)*x[0]**o+(-0.429845943537)*x[0]+(0.518551215268)*x[1]**o+(-0.551690143654)*x[1]+(0.591983049997)*x[2]**o+(0.586292683497)*x[2]
        arg[3,4,0,0]=(0.119036526248)*x[0]**o+(-0.156404044048)*x[0]+(0.158504048681)*x[1]**o+(-0.3779600709)*x[1]+(-0.362107671534)*x[2]**o+(-0.314508628804)*x[2]
        arg[3,4,0,1]=(0.382011103106)*x[0]**o+(-0.123711676964)*x[0]+(-0.939014457264)*x[1]**o+(-0.5986155807)*x[1]+(-0.852092491566)*x[2]**o+(-0.410438924788)*x[2]
        arg[3,4,1,0]=(0.83612586535)*x[0]**o+(-0.606098645092)*x[0]+(0.643469717148)*x[1]**o+(0.845106525654)*x[1]+(0.347455649691)*x[2]**o+(-0.166796580458)*x[2]
        arg[3,4,1,1]=(0.693431758156)*x[0]**o+(-0.313983401147)*x[0]+(-0.862690834893)*x[1]**o+(0.417233988055)*x[1]+(-0.361769997589)*x[2]**o+(-0.22749888397)*x[2]
        arg[3,4,2,0]=(0.680358405387)*x[0]**o+(-0.786041842215)*x[0]+(0.288823579415)*x[1]**o+(-0.961099851196)*x[1]+(0.669431785695)*x[2]**o+(0.956564160268)*x[2]
        arg[3,4,2,1]=(-0.21416500005)*x[0]**o+(0.48495191486)*x[0]+(-0.173465452645)*x[1]**o+(0.662530321557)*x[1]+(0.984573283824)*x[2]**o+(-0.568211383785)*x[2]
        ref[0,0,0,0]=(-0.615899511086)*x_ref[0]**o+(0.823637401824)*x_ref[0]+(-0.986967740198)*x_ref[1]**o+(-0.782088977641)*x_ref[1]+(-0.576799974679)*x_ref[2]**o+(-0.910942885711)*x_ref[2]
        ref[0,0,0,1]=(-0.790830593449)*x_ref[0]**o+(-0.897724279499)*x_ref[0]+(-0.881691805807)*x_ref[1]**o+(-0.133793238901)*x_ref[1]+(-0.0903391684489)*x_ref[2]**o+(0.659893469877)*x_ref[2]
        ref[0,0,1,0]=(-0.0567117900114)*x_ref[0]**o+(-0.858031985686)*x_ref[0]+(-0.397443445702)*x_ref[1]**o+(0.549982838823)*x_ref[1]+(-0.636905143813)*x_ref[2]**o+(0.8488142632)*x_ref[2]
        ref[0,0,1,1]=(-0.644557373082)*x_ref[0]**o+(-0.323654898522)*x_ref[0]+(0.210458705923)*x_ref[1]**o+(0.990436827296)*x_ref[1]+(-0.597872757151)*x_ref[2]**o+(-0.83865548618)*x_ref[2]
        ref[0,0,2,0]=(-0.29830667084)*x_ref[0]**o+(0.327035546909)*x_ref[0]+(0.132690049791)*x_ref[1]**o+(-0.55625443282)*x_ref[1]+(0.0972781385283)*x_ref[2]**o+(0.151885837608)*x_ref[2]
        ref[0,0,2,1]=(-0.435983465578)*x_ref[0]**o+(-0.0852522953446)*x_ref[0]+(0.257654807094)*x_ref[1]**o+(-0.336603197583)*x_ref[1]+(0.637913421592)*x_ref[2]**o+(0.8151042055)*x_ref[2]
        ref[0,1,0,0]=(-0.0373980191125)*x_ref[0]**o+(-0.45937285505)*x_ref[0]+(-0.857933684889)*x_ref[1]**o+(-0.843826686256)*x_ref[1]+(-0.727077320504)*x_ref[2]**o+(-0.65222863395)*x_ref[2]
        ref[0,1,0,1]=(0.63426017384)*x_ref[0]**o+(0.845708641601)*x_ref[0]+(0.246160302486)*x_ref[1]**o+(-0.660855572094)*x_ref[1]+(0.67420981328)*x_ref[2]**o+(-0.280803968781)*x_ref[2]
        ref[0,1,1,0]=(0.881193264597)*x_ref[0]**o+(-0.881936969515)*x_ref[0]+(-0.883246837152)*x_ref[1]**o+(0.182222471927)*x_ref[1]+(0.405380496272)*x_ref[2]**o+(0.175721746883)*x_ref[2]
        ref[0,1,1,1]=(-0.44447667237)*x_ref[0]**o+(0.956169274606)*x_ref[0]+(0.0370532940652)*x_ref[1]**o+(-0.181239635583)*x_ref[1]+(-0.678412161285)*x_ref[2]**o+(-0.653172157401)*x_ref[2]
        ref[0,1,2,0]=(-0.50038634718)*x_ref[0]**o+(0.00278321765167)*x_ref[0]+(-0.831539460218)*x_ref[1]**o+(-0.555243320385)*x_ref[1]+(-0.626789452003)*x_ref[2]**o+(-0.495085243933)*x_ref[2]
        ref[0,1,2,1]=(-0.32886359569)*x_ref[0]**o+(0.406099965475)*x_ref[0]+(0.764756177266)*x_ref[1]**o+(-0.272070344181)*x_ref[1]+(-0.41812596516)*x_ref[2]**o+(0.349063696376)*x_ref[2]
        ref[0,2,0,0]=(-0.580544978365)*x_ref[0]**o+(0.53441241334)*x_ref[0]+(-0.808749795876)*x_ref[1]**o+(0.172110801632)*x_ref[1]+(-0.939262606098)*x_ref[2]**o+(-0.724859325104)*x_ref[2]
        ref[0,2,0,1]=(0.256335221944)*x_ref[0]**o+(0.463522371018)*x_ref[0]+(0.990516478508)*x_ref[1]**o+(0.423931177054)*x_ref[1]+(0.184277693683)*x_ref[2]**o+(-0.272967447689)*x_ref[2]
        ref[0,2,1,0]=(-0.126943811008)*x_ref[0]**o+(0.645821387926)*x_ref[0]+(0.150370174987)*x_ref[1]**o+(-0.150954538301)*x_ref[1]+(-0.378720078559)*x_ref[2]**o+(0.0975963807109)*x_ref[2]
        ref[0,2,1,1]=(-0.451434544339)*x_ref[0]**o+(0.557314109088)*x_ref[0]+(-0.0402970749558)*x_ref[1]**o+(0.306984557206)*x_ref[1]+(-0.297018139186)*x_ref[2]**o+(0.593800698785)*x_ref[2]
        ref[0,2,2,0]=(-0.376199144911)*x_ref[0]**o+(-0.230454116373)*x_ref[0]+(0.992738297651)*x_ref[1]**o+(0.183510534927)*x_ref[1]+(0.409381711992)*x_ref[2]**o+(0.171945265727)*x_ref[2]
        ref[0,2,2,1]=(0.363233095971)*x_ref[0]**o+(-0.130589148249)*x_ref[0]+(0.809611042018)*x_ref[1]**o+(-0.354546307961)*x_ref[1]+(0.0631410169203)*x_ref[2]**o+(0.630306161781)*x_ref[2]
        ref[0,3,0,0]=(0.49655016161)*x_ref[0]**o+(0.878957258614)*x_ref[0]+(-0.397408696784)*x_ref[1]**o+(-0.199445321288)*x_ref[1]+(0.929857086663)*x_ref[2]**o+(-0.648770887161)*x_ref[2]
        ref[0,3,0,1]=(-0.85578334594)*x_ref[0]**o+(-0.294617109229)*x_ref[0]+(-0.0790295769768)*x_ref[1]**o+(0.82151942311)*x_ref[1]+(0.541628117444)*x_ref[2]**o+(0.805418174827)*x_ref[2]
        ref[0,3,1,0]=(0.800775454804)*x_ref[0]**o+(0.599550462983)*x_ref[0]+(-0.384788634082)*x_ref[1]**o+(-0.356933972363)*x_ref[1]+(-0.423116265473)*x_ref[2]**o+(-0.125492982003)*x_ref[2]
        ref[0,3,1,1]=(0.929652994661)*x_ref[0]**o+(-0.57396682406)*x_ref[0]+(0.313584116036)*x_ref[1]**o+(-0.190575043496)*x_ref[1]+(-0.29110884935)*x_ref[2]**o+(0.967530639925)*x_ref[2]
        ref[0,3,2,0]=(0.761136478868)*x_ref[0]**o+(0.565721170014)*x_ref[0]+(0.239619122131)*x_ref[1]**o+(-0.0821651979956)*x_ref[1]+(-0.699218775545)*x_ref[2]**o+(0.19973415585)*x_ref[2]
        ref[0,3,2,1]=(0.209543244648)*x_ref[0]**o+(0.93957796874)*x_ref[0]+(0.354405887426)*x_ref[1]**o+(0.504177262864)*x_ref[1]+(0.446866674894)*x_ref[2]**o+(-0.522936485464)*x_ref[2]
        ref[0,4,0,0]=(0.427541399652)*x_ref[0]**o+(-0.114806067206)*x_ref[0]+(0.236780877643)*x_ref[1]**o+(0.422226381272)*x_ref[1]+(-0.506450961214)*x_ref[2]**o+(0.616091720212)*x_ref[2]
        ref[0,4,0,1]=(0.698773571323)*x_ref[0]**o+(-0.357878713085)*x_ref[0]+(-0.453730359129)*x_ref[1]**o+(-0.328701041355)*x_ref[1]+(-0.28646346679)*x_ref[2]**o+(-0.0449784265203)*x_ref[2]
        ref[0,4,1,0]=(-0.419416306509)*x_ref[0]**o+(-0.1970394955)*x_ref[0]+(-0.814008932379)*x_ref[1]**o+(-0.805127589929)*x_ref[1]+(0.244059789875)*x_ref[2]**o+(-0.983027843847)*x_ref[2]
        ref[0,4,1,1]=(0.84837346004)*x_ref[0]**o+(0.501808510624)*x_ref[0]+(-0.105800351358)*x_ref[1]**o+(0.903999231718)*x_ref[1]+(-0.0473996978124)*x_ref[2]**o+(0.683647253381)*x_ref[2]
        ref[0,4,2,0]=(0.658722899682)*x_ref[0]**o+(-0.401846635016)*x_ref[0]+(-0.967322202581)*x_ref[1]**o+(-0.948306922217)*x_ref[1]+(0.496409308219)*x_ref[2]**o+(-0.184768777052)*x_ref[2]
        ref[0,4,2,1]=(-0.454468390888)*x_ref[0]**o+(-0.220640468215)*x_ref[0]+(0.995269925943)*x_ref[1]**o+(0.778951329211)*x_ref[1]+(0.0439534466302)*x_ref[2]**o+(0.806197038383)*x_ref[2]
        ref[1,0,0,0]=(-0.547461802236)*x_ref[0]**o+(0.370768827789)*x_ref[0]+(-0.619991757552)*x_ref[1]**o+(0.391727195645)*x_ref[1]+(-0.649274922953)*x_ref[2]**o+(-0.478788248429)*x_ref[2]
        ref[1,0,0,1]=(-0.899230505882)*x_ref[0]**o+(0.603064321735)*x_ref[0]+(-0.737475386349)*x_ref[1]**o+(0.594197953001)*x_ref[1]+(0.341324049064)*x_ref[2]**o+(0.285202633917)*x_ref[2]
        ref[1,0,1,0]=(-0.107063589618)*x_ref[0]**o+(-0.260491562358)*x_ref[0]+(-0.667059815618)*x_ref[1]**o+(0.473227354537)*x_ref[1]+(-0.587170510558)*x_ref[2]**o+(-0.740299218393)*x_ref[2]
        ref[1,0,1,1]=(0.87705715098)*x_ref[0]**o+(0.312753471827)*x_ref[0]+(-0.307530283315)*x_ref[1]**o+(0.195564644822)*x_ref[1]+(-0.64384485166)*x_ref[2]**o+(-0.776362951729)*x_ref[2]
        ref[1,0,2,0]=(0.0949408564739)*x_ref[0]**o+(0.765922179855)*x_ref[0]+(0.441215568253)*x_ref[1]**o+(0.339112969437)*x_ref[1]+(-0.0774615292449)*x_ref[2]**o+(0.673258055783)*x_ref[2]
        ref[1,0,2,1]=(0.114828212302)*x_ref[0]**o+(0.0831258766277)*x_ref[0]+(0.674540402248)*x_ref[1]**o+(-0.304921039094)*x_ref[1]+(-0.725514770578)*x_ref[2]**o+(-0.627584773746)*x_ref[2]
        ref[1,1,0,0]=(0.291561714675)*x_ref[0]**o+(0.898587624783)*x_ref[0]+(0.476696264545)*x_ref[1]**o+(-0.454648604978)*x_ref[1]+(0.677949150743)*x_ref[2]**o+(-0.59438289851)*x_ref[2]
        ref[1,1,0,1]=(0.356687258009)*x_ref[0]**o+(0.806236251629)*x_ref[0]+(0.843636690354)*x_ref[1]**o+(-0.649776947181)*x_ref[1]+(0.208542122388)*x_ref[2]**o+(-0.0726045573458)*x_ref[2]
        ref[1,1,1,0]=(0.712800450558)*x_ref[0]**o+(0.748271875755)*x_ref[0]+(0.403880649356)*x_ref[1]**o+(0.956549336754)*x_ref[1]+(0.0420128495043)*x_ref[2]**o+(-0.83177870499)*x_ref[2]
        ref[1,1,1,1]=(-0.886862904174)*x_ref[0]**o+(-0.03822986594)*x_ref[0]+(0.587151517496)*x_ref[1]**o+(-0.120272060898)*x_ref[1]+(-0.573366391933)*x_ref[2]**o+(-0.59772763241)*x_ref[2]
        ref[1,1,2,0]=(0.311960197804)*x_ref[0]**o+(0.486854949444)*x_ref[0]+(-0.647995651954)*x_ref[1]**o+(0.30104277964)*x_ref[1]+(-0.174718703748)*x_ref[2]**o+(-0.506993216295)*x_ref[2]
        ref[1,1,2,1]=(0.979056206991)*x_ref[0]**o+(0.892104464812)*x_ref[0]+(0.0507432703245)*x_ref[1]**o+(0.103630858454)*x_ref[1]+(-0.0254965952266)*x_ref[2]**o+(0.13034027239)*x_ref[2]
        ref[1,2,0,0]=(0.259797966548)*x_ref[0]**o+(0.53018999371)*x_ref[0]+(0.533745722992)*x_ref[1]**o+(0.53029084173)*x_ref[1]+(-0.595126791061)*x_ref[2]**o+(-0.51668295558)*x_ref[2]
        ref[1,2,0,1]=(0.926395590808)*x_ref[0]**o+(-0.0700141207179)*x_ref[0]+(-0.973392454012)*x_ref[1]**o+(-0.46204067097)*x_ref[1]+(0.814964497207)*x_ref[2]**o+(-0.183546091376)*x_ref[2]
        ref[1,2,1,0]=(-0.0872635441658)*x_ref[0]**o+(-0.0307630839035)*x_ref[0]+(0.539798653095)*x_ref[1]**o+(0.966367734848)*x_ref[1]+(-0.817495053612)*x_ref[2]**o+(-0.268397867848)*x_ref[2]
        ref[1,2,1,1]=(-0.933074613921)*x_ref[0]**o+(0.48896533938)*x_ref[0]+(0.453696133488)*x_ref[1]**o+(-0.238730406749)*x_ref[1]+(-0.696785951086)*x_ref[2]**o+(-0.359858385642)*x_ref[2]
        ref[1,2,2,0]=(-0.844406286809)*x_ref[0]**o+(-0.506332243946)*x_ref[0]+(0.976436731658)*x_ref[1]**o+(-0.201489161857)*x_ref[1]+(0.851236173166)*x_ref[2]**o+(-0.998803612947)*x_ref[2]
        ref[1,2,2,1]=(0.259707321449)*x_ref[0]**o+(-0.888149478655)*x_ref[0]+(-0.273502842869)*x_ref[1]**o+(0.291193582158)*x_ref[1]+(-0.897542178027)*x_ref[2]**o+(-0.345994824429)*x_ref[2]
        ref[1,3,0,0]=(0.961298984697)*x_ref[0]**o+(0.248141115092)*x_ref[0]+(-0.623053025178)*x_ref[1]**o+(-0.775095452078)*x_ref[1]+(0.955362707049)*x_ref[2]**o+(-0.517547333705)*x_ref[2]
        ref[1,3,0,1]=(0.858098596227)*x_ref[0]**o+(0.309307483873)*x_ref[0]+(-0.209217034755)*x_ref[1]**o+(0.783803593379)*x_ref[1]+(0.138164128092)*x_ref[2]**o+(-0.922854003461)*x_ref[2]
        ref[1,3,1,0]=(-0.139017942141)*x_ref[0]**o+(-0.338223039119)*x_ref[0]+(0.737882941498)*x_ref[1]**o+(0.948854903087)*x_ref[1]+(-0.0792461633266)*x_ref[2]**o+(0.848875660203)*x_ref[2]
        ref[1,3,1,1]=(0.693284332212)*x_ref[0]**o+(-0.392272875348)*x_ref[0]+(-0.475709045479)*x_ref[1]**o+(0.182948077236)*x_ref[1]+(0.0277569761538)*x_ref[2]**o+(-0.550294947596)*x_ref[2]
        ref[1,3,2,0]=(-0.105381649505)*x_ref[0]**o+(-0.403051690238)*x_ref[0]+(-0.722219735818)*x_ref[1]**o+(-0.380114388869)*x_ref[1]+(0.321452572491)*x_ref[2]**o+(0.858583245253)*x_ref[2]
        ref[1,3,2,1]=(-0.405370062024)*x_ref[0]**o+(0.0743704840047)*x_ref[0]+(0.998947017815)*x_ref[1]**o+(0.983611079522)*x_ref[1]+(0.854202982674)*x_ref[2]**o+(0.288478746498)*x_ref[2]
        ref[1,4,0,0]=(0.0584782489998)*x_ref[0]**o+(0.999286927604)*x_ref[0]+(-0.458166020555)*x_ref[1]**o+(0.529008844128)*x_ref[1]+(0.424592641922)*x_ref[2]**o+(-0.489692340187)*x_ref[2]
        ref[1,4,0,1]=(-0.150555966138)*x_ref[0]**o+(0.362589320838)*x_ref[0]+(0.515820915391)*x_ref[1]**o+(0.742403661312)*x_ref[1]+(0.230278843183)*x_ref[2]**o+(-0.213977724366)*x_ref[2]
        ref[1,4,1,0]=(-0.361161802362)*x_ref[0]**o+(-0.49718332222)*x_ref[0]+(0.143956046197)*x_ref[1]**o+(-0.044550912758)*x_ref[1]+(-0.299892874708)*x_ref[2]**o+(0.370248709792)*x_ref[2]
        ref[1,4,1,1]=(0.259365238173)*x_ref[0]**o+(-0.474694548557)*x_ref[0]+(0.732983396669)*x_ref[1]**o+(-0.295151546774)*x_ref[1]+(-0.337144628526)*x_ref[2]**o+(0.664201775714)*x_ref[2]
        ref[1,4,2,0]=(-0.318016754284)*x_ref[0]**o+(0.555291443852)*x_ref[0]+(0.538150727998)*x_ref[1]**o+(-0.800118353132)*x_ref[1]+(0.652262386596)*x_ref[2]**o+(0.303156645857)*x_ref[2]
        ref[1,4,2,1]=(0.0827142921796)*x_ref[0]**o+(-0.88498648372)*x_ref[0]+(-0.941249844349)*x_ref[1]**o+(-0.162576659039)*x_ref[1]+(0.867621252192)*x_ref[2]**o+(0.0834583795227)*x_ref[2]
        ref[2,0,0,0]=(0.0670088199264)*x_ref[0]**o+(-0.0425961894346)*x_ref[0]+(0.329605789302)*x_ref[1]**o+(0.860453156363)*x_ref[1]+(-0.37410469727)*x_ref[2]**o+(0.55476432384)*x_ref[2]
        ref[2,0,0,1]=(0.801738221879)*x_ref[0]**o+(-0.836563530923)*x_ref[0]+(-0.449651673142)*x_ref[1]**o+(-0.19423133918)*x_ref[1]+(0.947648200044)*x_ref[2]**o+(0.0854067601423)*x_ref[2]
        ref[2,0,1,0]=(0.986595046855)*x_ref[0]**o+(-0.358319352175)*x_ref[0]+(0.493917713513)*x_ref[1]**o+(-0.92052282445)*x_ref[1]+(-0.692726467259)*x_ref[2]**o+(-0.0636943737807)*x_ref[2]
        ref[2,0,1,1]=(0.38491150492)*x_ref[0]**o+(0.766893245938)*x_ref[0]+(0.420188917945)*x_ref[1]**o+(0.43760538854)*x_ref[1]+(-0.770749451981)*x_ref[2]**o+(0.124882164378)*x_ref[2]
        ref[2,0,2,0]=(0.775715718797)*x_ref[0]**o+(0.89516703208)*x_ref[0]+(-0.829021524898)*x_ref[1]**o+(-0.650262454311)*x_ref[1]+(-0.812556336598)*x_ref[2]**o+(-0.997075743997)*x_ref[2]
        ref[2,0,2,1]=(0.3503761008)*x_ref[0]**o+(-0.822009897455)*x_ref[0]+(-0.657657365354)*x_ref[1]**o+(0.476295612156)*x_ref[1]+(0.315554008711)*x_ref[2]**o+(0.691631078049)*x_ref[2]
        ref[2,1,0,0]=(0.680177375433)*x_ref[0]**o+(-0.922726813311)*x_ref[0]+(-0.581116746321)*x_ref[1]**o+(0.628138099343)*x_ref[1]+(-0.568262800293)*x_ref[2]**o+(-0.228201382395)*x_ref[2]
        ref[2,1,0,1]=(0.210094833324)*x_ref[0]**o+(0.281549673876)*x_ref[0]+(0.23785231073)*x_ref[1]**o+(-0.103812455993)*x_ref[1]+(0.530445971191)*x_ref[2]**o+(-0.0830520449939)*x_ref[2]
        ref[2,1,1,0]=(0.367133297563)*x_ref[0]**o+(0.168289954585)*x_ref[0]+(0.614117392787)*x_ref[1]**o+(0.790467766841)*x_ref[1]+(-0.958002150805)*x_ref[2]**o+(0.328916248824)*x_ref[2]
        ref[2,1,1,1]=(0.089535347231)*x_ref[0]**o+(-0.93155993701)*x_ref[0]+(-0.674670760794)*x_ref[1]**o+(-0.274148261773)*x_ref[1]+(-0.807020126056)*x_ref[2]**o+(-0.323334769496)*x_ref[2]
        ref[2,1,2,0]=(-0.948679835219)*x_ref[0]**o+(-0.577235742425)*x_ref[0]+(-0.732584834146)*x_ref[1]**o+(-0.475242504288)*x_ref[1]+(0.213663223262)*x_ref[2]**o+(-0.151401354381)*x_ref[2]
        ref[2,1,2,1]=(0.264513246206)*x_ref[0]**o+(-0.605806437391)*x_ref[0]+(-0.058730926693)*x_ref[1]**o+(-0.661696304412)*x_ref[1]+(-0.361277352457)*x_ref[2]**o+(-0.804408258253)*x_ref[2]
        ref[2,2,0,0]=(-0.144562358054)*x_ref[0]**o+(-0.0960550555725)*x_ref[0]+(-0.905096867803)*x_ref[1]**o+(0.188861784103)*x_ref[1]+(-0.616751005568)*x_ref[2]**o+(0.604711974248)*x_ref[2]
        ref[2,2,0,1]=(-0.0582524307858)*x_ref[0]**o+(-0.503648879785)*x_ref[0]+(-0.682755209956)*x_ref[1]**o+(-0.808213164425)*x_ref[1]+(0.539987276327)*x_ref[2]**o+(0.650557703689)*x_ref[2]
        ref[2,2,1,0]=(0.0607729567733)*x_ref[0]**o+(-0.498819638428)*x_ref[0]+(0.434442595334)*x_ref[1]**o+(-0.491063977603)*x_ref[1]+(0.499436212229)*x_ref[2]**o+(-0.932071991206)*x_ref[2]
        ref[2,2,1,1]=(-0.461295295294)*x_ref[0]**o+(-0.153600767916)*x_ref[0]+(-0.86452392231)*x_ref[1]**o+(0.136760034153)*x_ref[1]+(-0.471642416082)*x_ref[2]**o+(0.453820403438)*x_ref[2]
        ref[2,2,2,0]=(-0.970051888584)*x_ref[0]**o+(0.415728825057)*x_ref[0]+(0.971643909569)*x_ref[1]**o+(0.923719420524)*x_ref[1]+(-0.837313513804)*x_ref[2]**o+(-0.738057676093)*x_ref[2]
        ref[2,2,2,1]=(-0.941355918089)*x_ref[0]**o+(-0.9629784524)*x_ref[0]+(-0.931184348072)*x_ref[1]**o+(0.720006097658)*x_ref[1]+(-0.140750567978)*x_ref[2]**o+(0.55381939119)*x_ref[2]
        ref[2,3,0,0]=(0.753455933669)*x_ref[0]**o+(0.352384922037)*x_ref[0]+(0.737644348002)*x_ref[1]**o+(-0.565144188248)*x_ref[1]+(0.257417624014)*x_ref[2]**o+(-0.660006641683)*x_ref[2]
        ref[2,3,0,1]=(0.886996770925)*x_ref[0]**o+(-0.876595418252)*x_ref[0]+(-0.189778073871)*x_ref[1]**o+(0.560867442336)*x_ref[1]+(-0.262338156346)*x_ref[2]**o+(0.614749098901)*x_ref[2]
        ref[2,3,1,0]=(-0.722912374016)*x_ref[0]**o+(0.447451310273)*x_ref[0]+(0.397510140568)*x_ref[1]**o+(0.932405917932)*x_ref[1]+(-0.206969252336)*x_ref[2]**o+(-0.897366208888)*x_ref[2]
        ref[2,3,1,1]=(0.106324074547)*x_ref[0]**o+(0.822109425884)*x_ref[0]+(0.62699223277)*x_ref[1]**o+(0.238371557839)*x_ref[1]+(-0.422051841243)*x_ref[2]**o+(0.868932130329)*x_ref[2]
        ref[2,3,2,0]=(0.0155790371581)*x_ref[0]**o+(-0.696029577918)*x_ref[0]+(0.259909585217)*x_ref[1]**o+(0.260209121862)*x_ref[1]+(0.780977081459)*x_ref[2]**o+(-0.819957534538)*x_ref[2]
        ref[2,3,2,1]=(0.736223721904)*x_ref[0]**o+(-0.433132598174)*x_ref[0]+(-0.880913649812)*x_ref[1]**o+(-0.375988777461)*x_ref[1]+(-0.351763406307)*x_ref[2]**o+(0.948468252467)*x_ref[2]
        ref[2,4,0,0]=(0.943349180556)*x_ref[0]**o+(-0.720044514103)*x_ref[0]+(0.671900161807)*x_ref[1]**o+(0.358571142942)*x_ref[1]+(-0.132755998648)*x_ref[2]**o+(0.0749815044365)*x_ref[2]
        ref[2,4,0,1]=(0.639070752347)*x_ref[0]**o+(0.551441363406)*x_ref[0]+(0.507547844432)*x_ref[1]**o+(-0.800043772307)*x_ref[1]+(-0.506359503394)*x_ref[2]**o+(0.952191432651)*x_ref[2]
        ref[2,4,1,0]=(-0.655864732254)*x_ref[0]**o+(0.142198390642)*x_ref[0]+(0.075292486702)*x_ref[1]**o+(0.191013189029)*x_ref[1]+(0.106768453247)*x_ref[2]**o+(0.974526515449)*x_ref[2]
        ref[2,4,1,1]=(-0.354750205648)*x_ref[0]**o+(0.141506849368)*x_ref[0]+(-0.170074898141)*x_ref[1]**o+(-0.0882776529335)*x_ref[1]+(0.845254802876)*x_ref[2]**o+(-0.435414692125)*x_ref[2]
        ref[2,4,2,0]=(-0.412836027163)*x_ref[0]**o+(0.830777132582)*x_ref[0]+(-0.87554952236)*x_ref[1]**o+(0.970647516313)*x_ref[1]+(0.989864827286)*x_ref[2]**o+(-0.67790755421)*x_ref[2]
        ref[2,4,2,1]=(0.762907045031)*x_ref[0]**o+(0.140771931118)*x_ref[0]+(0.716859940142)*x_ref[1]**o+(0.151769314189)*x_ref[1]+(-0.812408853675)*x_ref[2]**o+(0.755727078666)*x_ref[2]
        ref[3,0,0,0]=(-0.13358237438)*x_ref[0]**o+(-0.594735856789)*x_ref[0]+(-0.367418402298)*x_ref[1]**o+(0.5349162872)*x_ref[1]+(0.238285320118)*x_ref[2]**o+(0.999627906459)*x_ref[2]
        ref[3,0,0,1]=(-0.537433369218)*x_ref[0]**o+(0.133034652667)*x_ref[0]+(-0.190302464596)*x_ref[1]**o+(0.869919012762)*x_ref[1]+(0.320294306008)*x_ref[2]**o+(0.865908856629)*x_ref[2]
        ref[3,0,1,0]=(0.480851043043)*x_ref[0]**o+(0.833358259022)*x_ref[0]+(0.893051476621)*x_ref[1]**o+(-0.505255581199)*x_ref[1]+(-0.15316145317)*x_ref[2]**o+(-0.937103563776)*x_ref[2]
        ref[3,0,1,1]=(0.765360137605)*x_ref[0]**o+(0.502399552842)*x_ref[0]+(-0.215707512276)*x_ref[1]**o+(-0.402720816967)*x_ref[1]+(0.231360098578)*x_ref[2]**o+(0.732903327067)*x_ref[2]
        ref[3,0,2,0]=(0.120836669045)*x_ref[0]**o+(0.198346794548)*x_ref[0]+(-0.28708447536)*x_ref[1]**o+(0.289672075353)*x_ref[1]+(-0.0305777738322)*x_ref[2]**o+(-0.0686651569373)*x_ref[2]
        ref[3,0,2,1]=(0.0574598844614)*x_ref[0]**o+(0.031092492399)*x_ref[0]+(-0.406805507341)*x_ref[1]**o+(-0.623370509947)*x_ref[1]+(-0.476094160966)*x_ref[2]**o+(0.0840271758875)*x_ref[2]
        ref[3,1,0,0]=(-0.569415249126)*x_ref[0]**o+(-0.339675540399)*x_ref[0]+(0.182536980707)*x_ref[1]**o+(0.727480200453)*x_ref[1]+(0.851672551504)*x_ref[2]**o+(0.817312313075)*x_ref[2]
        ref[3,1,0,1]=(0.921422642904)*x_ref[0]**o+(0.697656143906)*x_ref[0]+(0.969442847752)*x_ref[1]**o+(-0.00957994710064)*x_ref[1]+(-0.315230556399)*x_ref[2]**o+(-0.870759158384)*x_ref[2]
        ref[3,1,1,0]=(-0.813310465463)*x_ref[0]**o+(-0.809362802903)*x_ref[0]+(-0.0685101056428)*x_ref[1]**o+(0.708195651847)*x_ref[1]+(-0.876449456312)*x_ref[2]**o+(0.882503536479)*x_ref[2]
        ref[3,1,1,1]=(-0.277504922905)*x_ref[0]**o+(-0.819428706347)*x_ref[0]+(0.463203258731)*x_ref[1]**o+(0.597493096401)*x_ref[1]+(0.630522187524)*x_ref[2]**o+(0.92480079466)*x_ref[2]
        ref[3,1,2,0]=(-0.703167803764)*x_ref[0]**o+(0.730459205382)*x_ref[0]+(-0.507036921123)*x_ref[1]**o+(0.419358545765)*x_ref[1]+(-0.542758923679)*x_ref[2]**o+(0.831844312917)*x_ref[2]
        ref[3,1,2,1]=(0.135200299704)*x_ref[0]**o+(-0.160734675188)*x_ref[0]+(-0.893851689192)*x_ref[1]**o+(0.748734906185)*x_ref[1]+(0.970464447167)*x_ref[2]**o+(0.277911376757)*x_ref[2]
        ref[3,2,0,0]=(0.750104328784)*x_ref[0]**o+(-0.818350593893)*x_ref[0]+(-0.749900854281)*x_ref[1]**o+(0.289117653816)*x_ref[1]+(-0.219904822162)*x_ref[2]**o+(0.805208050452)*x_ref[2]
        ref[3,2,0,1]=(-0.129614389749)*x_ref[0]**o+(-0.64067443816)*x_ref[0]+(0.513013650233)*x_ref[1]**o+(0.868869035341)*x_ref[1]+(-0.040145121639)*x_ref[2]**o+(-0.282350136119)*x_ref[2]
        ref[3,2,1,0]=(-0.172814929104)*x_ref[0]**o+(-0.296930471915)*x_ref[0]+(-0.796114111585)*x_ref[1]**o+(-0.210181080249)*x_ref[1]+(-0.606107841173)*x_ref[2]**o+(-0.954329084175)*x_ref[2]
        ref[3,2,1,1]=(0.136149261959)*x_ref[0]**o+(-0.613980696755)*x_ref[0]+(0.968430714778)*x_ref[1]**o+(0.683553952099)*x_ref[1]+(-0.321383789121)*x_ref[2]**o+(-0.139570984234)*x_ref[2]
        ref[3,2,2,0]=(-0.273321720654)*x_ref[0]**o+(0.963651110162)*x_ref[0]+(-0.847402611499)*x_ref[1]**o+(0.237131262905)*x_ref[1]+(-0.617296821558)*x_ref[2]**o+(-0.681856394629)*x_ref[2]
        ref[3,2,2,1]=(-0.760071928624)*x_ref[0]**o+(0.0345053779755)*x_ref[0]+(-0.583680069732)*x_ref[1]**o+(0.150497658077)*x_ref[1]+(0.0257096621429)*x_ref[2]**o+(-0.159406203385)*x_ref[2]
        ref[3,3,0,0]=(-0.0668501746947)*x_ref[0]**o+(-0.183645899611)*x_ref[0]+(0.123667845765)*x_ref[1]**o+(0.0928142420352)*x_ref[1]+(-0.933250766317)*x_ref[2]**o+(-0.629986949639)*x_ref[2]
        ref[3,3,0,1]=(-0.0212790928206)*x_ref[0]**o+(0.444036360325)*x_ref[0]+(0.09182358127)*x_ref[1]**o+(0.473088563672)*x_ref[1]+(0.0852925382953)*x_ref[2]**o+(-0.238517929786)*x_ref[2]
        ref[3,3,1,0]=(0.0647235093728)*x_ref[0]**o+(-0.266285898716)*x_ref[0]+(0.944999560866)*x_ref[1]**o+(-0.757696716486)*x_ref[1]+(-0.439165278732)*x_ref[2]**o+(0.960091479774)*x_ref[2]
        ref[3,3,1,1]=(-0.792249238894)*x_ref[0]**o+(-0.568585439587)*x_ref[0]+(-0.0426655335607)*x_ref[1]**o+(0.980004032206)*x_ref[1]+(0.635227812889)*x_ref[2]**o+(0.6687165384)*x_ref[2]
        ref[3,3,2,0]=(0.340468442769)*x_ref[0]**o+(0.791071154735)*x_ref[0]+(-0.817919975547)*x_ref[1]**o+(0.0328383066982)*x_ref[1]+(-0.323093904705)*x_ref[2]**o+(-0.302641620002)*x_ref[2]
        ref[3,3,2,1]=(-0.476501733213)*x_ref[0]**o+(-0.429845943537)*x_ref[0]+(0.518551215268)*x_ref[1]**o+(-0.551690143654)*x_ref[1]+(0.591983049997)*x_ref[2]**o+(0.586292683497)*x_ref[2]
        ref[3,4,0,0]=(0.119036526248)*x_ref[0]**o+(-0.156404044048)*x_ref[0]+(0.158504048681)*x_ref[1]**o+(-0.3779600709)*x_ref[1]+(-0.362107671534)*x_ref[2]**o+(-0.314508628804)*x_ref[2]
        ref[3,4,0,1]=(0.382011103106)*x_ref[0]**o+(-0.123711676964)*x_ref[0]+(-0.939014457264)*x_ref[1]**o+(-0.5986155807)*x_ref[1]+(-0.852092491566)*x_ref[2]**o+(-0.410438924788)*x_ref[2]
        ref[3,4,1,0]=(0.83612586535)*x_ref[0]**o+(-0.606098645092)*x_ref[0]+(0.643469717148)*x_ref[1]**o+(0.845106525654)*x_ref[1]+(0.347455649691)*x_ref[2]**o+(-0.166796580458)*x_ref[2]
        ref[3,4,1,1]=(0.693431758156)*x_ref[0]**o+(-0.313983401147)*x_ref[0]+(-0.862690834893)*x_ref[1]**o+(0.417233988055)*x_ref[1]+(-0.361769997589)*x_ref[2]**o+(-0.22749888397)*x_ref[2]
        ref[3,4,2,0]=(0.680358405387)*x_ref[0]**o+(-0.786041842215)*x_ref[0]+(0.288823579415)*x_ref[1]**o+(-0.961099851196)*x_ref[1]+(0.669431785695)*x_ref[2]**o+(0.956564160268)*x_ref[2]
        ref[3,4,2,1]=(-0.21416500005)*x_ref[0]**o+(0.48495191486)*x_ref[0]+(-0.173465452645)*x_ref[1]**o+(0.662530321557)*x_ref[1]+(0.984573283824)*x_ref[2]**o+(-0.568211383785)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.901681414005)*x[0]**o+(-0.888451321413)*x[0]+(-0.22985947361)*x[1]**o+(0.492641717896)*x[1]
        ref=(-0.901681414005)*x_ref[0]**o+(-0.888451321413)*x_ref[0]+(-0.22985947361)*x_ref[1]**o+(0.492641717896)*x_ref[1]
      else:
        arg=(0.428268592739)*x[0]**o+(-0.996603298562)*x[0]+(0.677084172419)*x[1]**o+(0.0826120162779)*x[1]+(0.0206562071525)*x[2]**o+(-0.79500921274)*x[2]
        ref=(0.428268592739)*x_ref[0]**o+(-0.996603298562)*x_ref[0]+(0.677084172419)*x_ref[1]**o+(0.0826120162779)*x_ref[1]+(0.0206562071525)*x_ref[2]**o+(-0.79500921274)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.756853446555)*x[0]**o+(-0.767585641318)*x[0]+(-0.426982757541)*x[1]**o+(0.399806070854)*x[1]
        arg[1]=(-0.821677396516)*x[0]**o+(-0.591723513375)*x[0]+(-0.355824690337)*x[1]**o+(0.1636370609)*x[1]
        ref[0]=(0.756853446555)*x_ref[0]**o+(-0.767585641318)*x_ref[0]+(-0.426982757541)*x_ref[1]**o+(0.399806070854)*x_ref[1]
        ref[1]=(-0.821677396516)*x_ref[0]**o+(-0.591723513375)*x_ref[0]+(-0.355824690337)*x_ref[1]**o+(0.1636370609)*x_ref[1]
      else:
        arg[0]=(0.938054074343)*x[0]**o+(0.539056499967)*x[0]+(-0.617107808198)*x[1]**o+(-0.00223894362929)*x[1]+(-0.789674052883)*x[2]**o+(-0.61592268025)*x[2]
        arg[1]=(-0.820479870378)*x[0]**o+(0.34390095545)*x[0]+(-0.501478928282)*x[1]**o+(0.174440566415)*x[1]+(-0.964557050796)*x[2]**o+(0.116014110585)*x[2]
        ref[0]=(0.938054074343)*x_ref[0]**o+(0.539056499967)*x_ref[0]+(-0.617107808198)*x_ref[1]**o+(-0.00223894362929)*x_ref[1]+(-0.789674052883)*x_ref[2]**o+(-0.61592268025)*x_ref[2]
        ref[1]=(-0.820479870378)*x_ref[0]**o+(0.34390095545)*x_ref[0]+(-0.501478928282)*x_ref[1]**o+(0.174440566415)*x_ref[1]+(-0.964557050796)*x_ref[2]**o+(0.116014110585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.717924477933)*x[0]**o+(-0.629682284247)*x[0]+(0.0982832832089)*x[1]**o+(-0.762193509689)*x[1]
        arg[0,1]=(-0.997822794773)*x[0]**o+(-0.561696934643)*x[0]+(-0.145516013467)*x[1]**o+(0.630378080139)*x[1]
        arg[0,2]=(-0.751323417918)*x[0]**o+(0.479440101564)*x[0]+(-0.00432956740012)*x[1]**o+(0.767912131968)*x[1]
        arg[0,3]=(0.507272228985)*x[0]**o+(-0.389291668683)*x[0]+(-0.883016581216)*x[1]**o+(0.0437908939974)*x[1]
        arg[0,4]=(-0.137343257739)*x[0]**o+(0.743817161998)*x[0]+(0.507769921129)*x[1]**o+(-0.300283389883)*x[1]
        arg[1,0]=(0.257422912353)*x[0]**o+(0.681717676022)*x[0]+(0.79106312381)*x[1]**o+(-0.671634660385)*x[1]
        arg[1,1]=(0.136314663047)*x[0]**o+(-0.782099147193)*x[0]+(0.885974280892)*x[1]**o+(-0.798425202013)*x[1]
        arg[1,2]=(0.247486687708)*x[0]**o+(0.795328036132)*x[0]+(-0.875686910387)*x[1]**o+(0.0568336903007)*x[1]
        arg[1,3]=(0.080829127843)*x[0]**o+(-0.386433809573)*x[0]+(0.314034401629)*x[1]**o+(-0.872662795541)*x[1]
        arg[1,4]=(-0.807283287271)*x[0]**o+(0.725492503192)*x[0]+(-0.584629907108)*x[1]**o+(-0.268021081298)*x[1]
        arg[2,0]=(0.581563251557)*x[0]**o+(-0.886151193174)*x[0]+(0.193339118534)*x[1]**o+(0.978747305418)*x[1]
        arg[2,1]=(0.141736956675)*x[0]**o+(0.668515974495)*x[0]+(-0.0791701432021)*x[1]**o+(0.410197138729)*x[1]
        arg[2,2]=(-0.287456411619)*x[0]**o+(-0.772391150719)*x[0]+(-0.748682403636)*x[1]**o+(0.538840924967)*x[1]
        arg[2,3]=(0.637529128236)*x[0]**o+(0.471131728149)*x[0]+(-0.433070242121)*x[1]**o+(-0.583240138198)*x[1]
        arg[2,4]=(-0.624702895751)*x[0]**o+(-0.118770841979)*x[0]+(0.587084874695)*x[1]**o+(0.46959556491)*x[1]
        arg[3,0]=(0.209166082888)*x[0]**o+(0.870738524234)*x[0]+(0.00990651539088)*x[1]**o+(-0.423308444734)*x[1]
        arg[3,1]=(0.00446369533847)*x[0]**o+(-0.445436215897)*x[0]+(-0.939826442104)*x[1]**o+(0.486172187596)*x[1]
        arg[3,2]=(-0.697054010365)*x[0]**o+(0.543343700538)*x[0]+(0.491284777488)*x[1]**o+(-0.021410495956)*x[1]
        arg[3,3]=(-0.00457600194633)*x[0]**o+(-0.234648384385)*x[0]+(0.200988754362)*x[1]**o+(0.74772788001)*x[1]
        arg[3,4]=(0.721800958056)*x[0]**o+(0.930824787753)*x[0]+(0.374340208263)*x[1]**o+(0.474293362128)*x[1]
        ref[0,0]=(-0.717924477933)*x_ref[0]**o+(-0.629682284247)*x_ref[0]+(0.0982832832089)*x_ref[1]**o+(-0.762193509689)*x_ref[1]
        ref[0,1]=(-0.997822794773)*x_ref[0]**o+(-0.561696934643)*x_ref[0]+(-0.145516013467)*x_ref[1]**o+(0.630378080139)*x_ref[1]
        ref[0,2]=(-0.751323417918)*x_ref[0]**o+(0.479440101564)*x_ref[0]+(-0.00432956740012)*x_ref[1]**o+(0.767912131968)*x_ref[1]
        ref[0,3]=(0.507272228985)*x_ref[0]**o+(-0.389291668683)*x_ref[0]+(-0.883016581216)*x_ref[1]**o+(0.0437908939974)*x_ref[1]
        ref[0,4]=(-0.137343257739)*x_ref[0]**o+(0.743817161998)*x_ref[0]+(0.507769921129)*x_ref[1]**o+(-0.300283389883)*x_ref[1]
        ref[1,0]=(0.257422912353)*x_ref[0]**o+(0.681717676022)*x_ref[0]+(0.79106312381)*x_ref[1]**o+(-0.671634660385)*x_ref[1]
        ref[1,1]=(0.136314663047)*x_ref[0]**o+(-0.782099147193)*x_ref[0]+(0.885974280892)*x_ref[1]**o+(-0.798425202013)*x_ref[1]
        ref[1,2]=(0.247486687708)*x_ref[0]**o+(0.795328036132)*x_ref[0]+(-0.875686910387)*x_ref[1]**o+(0.0568336903007)*x_ref[1]
        ref[1,3]=(0.080829127843)*x_ref[0]**o+(-0.386433809573)*x_ref[0]+(0.314034401629)*x_ref[1]**o+(-0.872662795541)*x_ref[1]
        ref[1,4]=(-0.807283287271)*x_ref[0]**o+(0.725492503192)*x_ref[0]+(-0.584629907108)*x_ref[1]**o+(-0.268021081298)*x_ref[1]
        ref[2,0]=(0.581563251557)*x_ref[0]**o+(-0.886151193174)*x_ref[0]+(0.193339118534)*x_ref[1]**o+(0.978747305418)*x_ref[1]
        ref[2,1]=(0.141736956675)*x_ref[0]**o+(0.668515974495)*x_ref[0]+(-0.0791701432021)*x_ref[1]**o+(0.410197138729)*x_ref[1]
        ref[2,2]=(-0.287456411619)*x_ref[0]**o+(-0.772391150719)*x_ref[0]+(-0.748682403636)*x_ref[1]**o+(0.538840924967)*x_ref[1]
        ref[2,3]=(0.637529128236)*x_ref[0]**o+(0.471131728149)*x_ref[0]+(-0.433070242121)*x_ref[1]**o+(-0.583240138198)*x_ref[1]
        ref[2,4]=(-0.624702895751)*x_ref[0]**o+(-0.118770841979)*x_ref[0]+(0.587084874695)*x_ref[1]**o+(0.46959556491)*x_ref[1]
        ref[3,0]=(0.209166082888)*x_ref[0]**o+(0.870738524234)*x_ref[0]+(0.00990651539088)*x_ref[1]**o+(-0.423308444734)*x_ref[1]
        ref[3,1]=(0.00446369533847)*x_ref[0]**o+(-0.445436215897)*x_ref[0]+(-0.939826442104)*x_ref[1]**o+(0.486172187596)*x_ref[1]
        ref[3,2]=(-0.697054010365)*x_ref[0]**o+(0.543343700538)*x_ref[0]+(0.491284777488)*x_ref[1]**o+(-0.021410495956)*x_ref[1]
        ref[3,3]=(-0.00457600194633)*x_ref[0]**o+(-0.234648384385)*x_ref[0]+(0.200988754362)*x_ref[1]**o+(0.74772788001)*x_ref[1]
        ref[3,4]=(0.721800958056)*x_ref[0]**o+(0.930824787753)*x_ref[0]+(0.374340208263)*x_ref[1]**o+(0.474293362128)*x_ref[1]
      else:
        arg[0,0]=(-0.740386815771)*x[0]**o+(-0.0523021332821)*x[0]+(0.0398159725211)*x[1]**o+(-0.426888207497)*x[1]+(0.0567170299031)*x[2]**o+(-0.183938925645)*x[2]
        arg[0,1]=(-0.201789590886)*x[0]**o+(-0.367737630124)*x[0]+(0.109364462083)*x[1]**o+(-0.28601743506)*x[1]+(-0.399129253776)*x[2]**o+(0.583952239438)*x[2]
        arg[0,2]=(0.542532512804)*x[0]**o+(0.494398493909)*x[0]+(0.292582889596)*x[1]**o+(-0.954001260258)*x[1]+(0.943184269693)*x[2]**o+(-0.52768765864)*x[2]
        arg[0,3]=(0.804183352494)*x[0]**o+(-0.975349076994)*x[0]+(-0.827375030117)*x[1]**o+(-0.669898333686)*x[1]+(0.748094239981)*x[2]**o+(0.979045988454)*x[2]
        arg[0,4]=(0.425028347472)*x[0]**o+(0.875274586351)*x[0]+(0.854098942537)*x[1]**o+(0.579588487828)*x[1]+(0.082708589378)*x[2]**o+(0.476235952573)*x[2]
        arg[1,0]=(-0.624794007494)*x[0]**o+(-0.138866900862)*x[0]+(0.970509187706)*x[1]**o+(0.0253783907447)*x[1]+(0.468865716232)*x[2]**o+(0.676697194907)*x[2]
        arg[1,1]=(-0.153766978847)*x[0]**o+(-0.426626120688)*x[0]+(-0.773847907631)*x[1]**o+(0.245927709658)*x[1]+(0.255639913539)*x[2]**o+(0.808676567134)*x[2]
        arg[1,2]=(0.232094367521)*x[0]**o+(0.297706438396)*x[0]+(0.446407126908)*x[1]**o+(0.740223236781)*x[1]+(0.39141114872)*x[2]**o+(0.773847600596)*x[2]
        arg[1,3]=(0.255568261799)*x[0]**o+(-0.978846043266)*x[0]+(0.354385282456)*x[1]**o+(0.105066752677)*x[1]+(0.942760820578)*x[2]**o+(0.780092891693)*x[2]
        arg[1,4]=(0.529084170195)*x[0]**o+(0.747311333653)*x[0]+(-0.545359901284)*x[1]**o+(-0.281724707651)*x[1]+(0.313967713269)*x[2]**o+(-0.215058831724)*x[2]
        arg[2,0]=(-0.831081561199)*x[0]**o+(-0.773735958748)*x[0]+(0.368966594678)*x[1]**o+(0.666937704315)*x[1]+(-0.974922863204)*x[2]**o+(0.715578049273)*x[2]
        arg[2,1]=(0.578362294564)*x[0]**o+(0.889684843543)*x[0]+(0.588806147025)*x[1]**o+(0.0617034215889)*x[1]+(-0.16631520681)*x[2]**o+(0.730086241349)*x[2]
        arg[2,2]=(0.448838826514)*x[0]**o+(-0.0709459785553)*x[0]+(-0.876698569979)*x[1]**o+(-0.506314680523)*x[1]+(0.688738500291)*x[2]**o+(0.618955116168)*x[2]
        arg[2,3]=(0.290663008974)*x[0]**o+(0.73644965206)*x[0]+(-0.14313225421)*x[1]**o+(0.83074119153)*x[1]+(-0.247893391581)*x[2]**o+(-0.541433435359)*x[2]
        arg[2,4]=(-0.215556795169)*x[0]**o+(-0.340257823449)*x[0]+(0.326604718415)*x[1]**o+(-0.939889251927)*x[1]+(0.425604617369)*x[2]**o+(0.430639358886)*x[2]
        arg[3,0]=(0.0317232868586)*x[0]**o+(0.820693319131)*x[0]+(0.973940331646)*x[1]**o+(0.626976945683)*x[1]+(0.788447937745)*x[2]**o+(-0.718491682844)*x[2]
        arg[3,1]=(0.292924833121)*x[0]**o+(-0.648128515891)*x[0]+(-0.300791515646)*x[1]**o+(-0.26123312127)*x[1]+(-0.263541248274)*x[2]**o+(0.890699912174)*x[2]
        arg[3,2]=(-0.741612826841)*x[0]**o+(-0.095466430566)*x[0]+(-0.987682602605)*x[1]**o+(-0.0258256251685)*x[1]+(0.544387724417)*x[2]**o+(0.917676395583)*x[2]
        arg[3,3]=(0.148313573908)*x[0]**o+(-0.438648673746)*x[0]+(-0.151998255103)*x[1]**o+(-0.500903665338)*x[1]+(0.61275976182)*x[2]**o+(-0.0509716866391)*x[2]
        arg[3,4]=(0.292189003653)*x[0]**o+(0.564355242514)*x[0]+(-0.585718580333)*x[1]**o+(0.635365614994)*x[1]+(-0.00185257696349)*x[2]**o+(0.813357014726)*x[2]
        ref[0,0]=(-0.740386815771)*x_ref[0]**o+(-0.0523021332821)*x_ref[0]+(0.0398159725211)*x_ref[1]**o+(-0.426888207497)*x_ref[1]+(0.0567170299031)*x_ref[2]**o+(-0.183938925645)*x_ref[2]
        ref[0,1]=(-0.201789590886)*x_ref[0]**o+(-0.367737630124)*x_ref[0]+(0.109364462083)*x_ref[1]**o+(-0.28601743506)*x_ref[1]+(-0.399129253776)*x_ref[2]**o+(0.583952239438)*x_ref[2]
        ref[0,2]=(0.542532512804)*x_ref[0]**o+(0.494398493909)*x_ref[0]+(0.292582889596)*x_ref[1]**o+(-0.954001260258)*x_ref[1]+(0.943184269693)*x_ref[2]**o+(-0.52768765864)*x_ref[2]
        ref[0,3]=(0.804183352494)*x_ref[0]**o+(-0.975349076994)*x_ref[0]+(-0.827375030117)*x_ref[1]**o+(-0.669898333686)*x_ref[1]+(0.748094239981)*x_ref[2]**o+(0.979045988454)*x_ref[2]
        ref[0,4]=(0.425028347472)*x_ref[0]**o+(0.875274586351)*x_ref[0]+(0.854098942537)*x_ref[1]**o+(0.579588487828)*x_ref[1]+(0.082708589378)*x_ref[2]**o+(0.476235952573)*x_ref[2]
        ref[1,0]=(-0.624794007494)*x_ref[0]**o+(-0.138866900862)*x_ref[0]+(0.970509187706)*x_ref[1]**o+(0.0253783907447)*x_ref[1]+(0.468865716232)*x_ref[2]**o+(0.676697194907)*x_ref[2]
        ref[1,1]=(-0.153766978847)*x_ref[0]**o+(-0.426626120688)*x_ref[0]+(-0.773847907631)*x_ref[1]**o+(0.245927709658)*x_ref[1]+(0.255639913539)*x_ref[2]**o+(0.808676567134)*x_ref[2]
        ref[1,2]=(0.232094367521)*x_ref[0]**o+(0.297706438396)*x_ref[0]+(0.446407126908)*x_ref[1]**o+(0.740223236781)*x_ref[1]+(0.39141114872)*x_ref[2]**o+(0.773847600596)*x_ref[2]
        ref[1,3]=(0.255568261799)*x_ref[0]**o+(-0.978846043266)*x_ref[0]+(0.354385282456)*x_ref[1]**o+(0.105066752677)*x_ref[1]+(0.942760820578)*x_ref[2]**o+(0.780092891693)*x_ref[2]
        ref[1,4]=(0.529084170195)*x_ref[0]**o+(0.747311333653)*x_ref[0]+(-0.545359901284)*x_ref[1]**o+(-0.281724707651)*x_ref[1]+(0.313967713269)*x_ref[2]**o+(-0.215058831724)*x_ref[2]
        ref[2,0]=(-0.831081561199)*x_ref[0]**o+(-0.773735958748)*x_ref[0]+(0.368966594678)*x_ref[1]**o+(0.666937704315)*x_ref[1]+(-0.974922863204)*x_ref[2]**o+(0.715578049273)*x_ref[2]
        ref[2,1]=(0.578362294564)*x_ref[0]**o+(0.889684843543)*x_ref[0]+(0.588806147025)*x_ref[1]**o+(0.0617034215889)*x_ref[1]+(-0.16631520681)*x_ref[2]**o+(0.730086241349)*x_ref[2]
        ref[2,2]=(0.448838826514)*x_ref[0]**o+(-0.0709459785553)*x_ref[0]+(-0.876698569979)*x_ref[1]**o+(-0.506314680523)*x_ref[1]+(0.688738500291)*x_ref[2]**o+(0.618955116168)*x_ref[2]
        ref[2,3]=(0.290663008974)*x_ref[0]**o+(0.73644965206)*x_ref[0]+(-0.14313225421)*x_ref[1]**o+(0.83074119153)*x_ref[1]+(-0.247893391581)*x_ref[2]**o+(-0.541433435359)*x_ref[2]
        ref[2,4]=(-0.215556795169)*x_ref[0]**o+(-0.340257823449)*x_ref[0]+(0.326604718415)*x_ref[1]**o+(-0.939889251927)*x_ref[1]+(0.425604617369)*x_ref[2]**o+(0.430639358886)*x_ref[2]
        ref[3,0]=(0.0317232868586)*x_ref[0]**o+(0.820693319131)*x_ref[0]+(0.973940331646)*x_ref[1]**o+(0.626976945683)*x_ref[1]+(0.788447937745)*x_ref[2]**o+(-0.718491682844)*x_ref[2]
        ref[3,1]=(0.292924833121)*x_ref[0]**o+(-0.648128515891)*x_ref[0]+(-0.300791515646)*x_ref[1]**o+(-0.26123312127)*x_ref[1]+(-0.263541248274)*x_ref[2]**o+(0.890699912174)*x_ref[2]
        ref[3,2]=(-0.741612826841)*x_ref[0]**o+(-0.095466430566)*x_ref[0]+(-0.987682602605)*x_ref[1]**o+(-0.0258256251685)*x_ref[1]+(0.544387724417)*x_ref[2]**o+(0.917676395583)*x_ref[2]
        ref[3,3]=(0.148313573908)*x_ref[0]**o+(-0.438648673746)*x_ref[0]+(-0.151998255103)*x_ref[1]**o+(-0.500903665338)*x_ref[1]+(0.61275976182)*x_ref[2]**o+(-0.0509716866391)*x_ref[2]
        ref[3,4]=(0.292189003653)*x_ref[0]**o+(0.564355242514)*x_ref[0]+(-0.585718580333)*x_ref[1]**o+(0.635365614994)*x_ref[1]+(-0.00185257696349)*x_ref[2]**o+(0.813357014726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0128414848839)*x[0]**o+(0.465050006132)*x[0]+(0.965707055318)*x[1]**o+(-0.0245081370356)*x[1]
        arg[0,0,1]=(-0.200472047017)*x[0]**o+(-0.663028900371)*x[0]+(-0.943040910702)*x[1]**o+(-0.120563561139)*x[1]
        arg[0,1,0]=(-0.582794137412)*x[0]**o+(-0.17277488426)*x[0]+(-0.328856513193)*x[1]**o+(-0.995542154323)*x[1]
        arg[0,1,1]=(0.628900216244)*x[0]**o+(0.42314911968)*x[0]+(0.720137185717)*x[1]**o+(-0.360847007396)*x[1]
        arg[1,0,0]=(-0.320599577005)*x[0]**o+(0.551919914795)*x[0]+(-0.0459652088094)*x[1]**o+(-0.559395421304)*x[1]
        arg[1,0,1]=(0.148179561705)*x[0]**o+(0.434674232208)*x[0]+(-0.475375603732)*x[1]**o+(-0.58939903081)*x[1]
        arg[1,1,0]=(0.617416045561)*x[0]**o+(0.937293925453)*x[0]+(0.348092333305)*x[1]**o+(0.503125307597)*x[1]
        arg[1,1,1]=(-0.569221008948)*x[0]**o+(-0.00405502766105)*x[0]+(-0.389443082681)*x[1]**o+(-0.149118473431)*x[1]
        arg[2,0,0]=(0.118769588477)*x[0]**o+(-0.96708351582)*x[0]+(-0.0145855500917)*x[1]**o+(0.370997674141)*x[1]
        arg[2,0,1]=(-0.3129615573)*x[0]**o+(0.206299148764)*x[0]+(-0.721288027396)*x[1]**o+(-0.884411444055)*x[1]
        arg[2,1,0]=(-0.780163860911)*x[0]**o+(-0.568714964563)*x[0]+(-0.0381936992532)*x[1]**o+(0.830243364454)*x[1]
        arg[2,1,1]=(0.440674933834)*x[0]**o+(0.659074990645)*x[0]+(0.373520241513)*x[1]**o+(0.609301500641)*x[1]
        arg[3,0,0]=(0.377299608317)*x[0]**o+(0.258769326825)*x[0]+(0.108755954544)*x[1]**o+(0.416736701504)*x[1]
        arg[3,0,1]=(-0.368375671218)*x[0]**o+(-0.194408664196)*x[0]+(0.22066484944)*x[1]**o+(-0.105513517239)*x[1]
        arg[3,1,0]=(-0.45311889531)*x[0]**o+(0.758137223649)*x[0]+(0.642250313677)*x[1]**o+(0.67780041241)*x[1]
        arg[3,1,1]=(-0.584467518038)*x[0]**o+(0.873450785862)*x[0]+(0.949121750467)*x[1]**o+(-0.774601385823)*x[1]
        arg[4,0,0]=(0.36378291068)*x[0]**o+(-0.489477884746)*x[0]+(-0.665910069342)*x[1]**o+(0.9839391739)*x[1]
        arg[4,0,1]=(-0.539599763616)*x[0]**o+(0.651501126417)*x[0]+(-0.291030902108)*x[1]**o+(0.719794225744)*x[1]
        arg[4,1,0]=(0.144776367586)*x[0]**o+(-0.154981280527)*x[0]+(0.170526162601)*x[1]**o+(-0.821596149241)*x[1]
        arg[4,1,1]=(-0.460621351209)*x[0]**o+(-0.75486897537)*x[0]+(0.615982502995)*x[1]**o+(-0.756893047476)*x[1]
        arg[5,0,0]=(0.209926220994)*x[0]**o+(0.412132938536)*x[0]+(0.28119780747)*x[1]**o+(-0.829072062922)*x[1]
        arg[5,0,1]=(0.347615632334)*x[0]**o+(-0.101954769092)*x[0]+(-0.535842225724)*x[1]**o+(-0.0790159065772)*x[1]
        arg[5,1,0]=(-0.951105825038)*x[0]**o+(-0.167113372047)*x[0]+(-0.368722457931)*x[1]**o+(0.00398630826759)*x[1]
        arg[5,1,1]=(-0.280966220945)*x[0]**o+(0.749467953656)*x[0]+(-0.862358852162)*x[1]**o+(-0.947469365289)*x[1]
        ref[0,0,0]=(-0.0128414848839)*x_ref[0]**o+(0.465050006132)*x_ref[0]+(0.965707055318)*x_ref[1]**o+(-0.0245081370356)*x_ref[1]
        ref[0,0,1]=(-0.200472047017)*x_ref[0]**o+(-0.663028900371)*x_ref[0]+(-0.943040910702)*x_ref[1]**o+(-0.120563561139)*x_ref[1]
        ref[0,1,0]=(-0.582794137412)*x_ref[0]**o+(-0.17277488426)*x_ref[0]+(-0.328856513193)*x_ref[1]**o+(-0.995542154323)*x_ref[1]
        ref[0,1,1]=(0.628900216244)*x_ref[0]**o+(0.42314911968)*x_ref[0]+(0.720137185717)*x_ref[1]**o+(-0.360847007396)*x_ref[1]
        ref[1,0,0]=(-0.320599577005)*x_ref[0]**o+(0.551919914795)*x_ref[0]+(-0.0459652088094)*x_ref[1]**o+(-0.559395421304)*x_ref[1]
        ref[1,0,1]=(0.148179561705)*x_ref[0]**o+(0.434674232208)*x_ref[0]+(-0.475375603732)*x_ref[1]**o+(-0.58939903081)*x_ref[1]
        ref[1,1,0]=(0.617416045561)*x_ref[0]**o+(0.937293925453)*x_ref[0]+(0.348092333305)*x_ref[1]**o+(0.503125307597)*x_ref[1]
        ref[1,1,1]=(-0.569221008948)*x_ref[0]**o+(-0.00405502766105)*x_ref[0]+(-0.389443082681)*x_ref[1]**o+(-0.149118473431)*x_ref[1]
        ref[2,0,0]=(0.118769588477)*x_ref[0]**o+(-0.96708351582)*x_ref[0]+(-0.0145855500917)*x_ref[1]**o+(0.370997674141)*x_ref[1]
        ref[2,0,1]=(-0.3129615573)*x_ref[0]**o+(0.206299148764)*x_ref[0]+(-0.721288027396)*x_ref[1]**o+(-0.884411444055)*x_ref[1]
        ref[2,1,0]=(-0.780163860911)*x_ref[0]**o+(-0.568714964563)*x_ref[0]+(-0.0381936992532)*x_ref[1]**o+(0.830243364454)*x_ref[1]
        ref[2,1,1]=(0.440674933834)*x_ref[0]**o+(0.659074990645)*x_ref[0]+(0.373520241513)*x_ref[1]**o+(0.609301500641)*x_ref[1]
        ref[3,0,0]=(0.377299608317)*x_ref[0]**o+(0.258769326825)*x_ref[0]+(0.108755954544)*x_ref[1]**o+(0.416736701504)*x_ref[1]
        ref[3,0,1]=(-0.368375671218)*x_ref[0]**o+(-0.194408664196)*x_ref[0]+(0.22066484944)*x_ref[1]**o+(-0.105513517239)*x_ref[1]
        ref[3,1,0]=(-0.45311889531)*x_ref[0]**o+(0.758137223649)*x_ref[0]+(0.642250313677)*x_ref[1]**o+(0.67780041241)*x_ref[1]
        ref[3,1,1]=(-0.584467518038)*x_ref[0]**o+(0.873450785862)*x_ref[0]+(0.949121750467)*x_ref[1]**o+(-0.774601385823)*x_ref[1]
        ref[4,0,0]=(0.36378291068)*x_ref[0]**o+(-0.489477884746)*x_ref[0]+(-0.665910069342)*x_ref[1]**o+(0.9839391739)*x_ref[1]
        ref[4,0,1]=(-0.539599763616)*x_ref[0]**o+(0.651501126417)*x_ref[0]+(-0.291030902108)*x_ref[1]**o+(0.719794225744)*x_ref[1]
        ref[4,1,0]=(0.144776367586)*x_ref[0]**o+(-0.154981280527)*x_ref[0]+(0.170526162601)*x_ref[1]**o+(-0.821596149241)*x_ref[1]
        ref[4,1,1]=(-0.460621351209)*x_ref[0]**o+(-0.75486897537)*x_ref[0]+(0.615982502995)*x_ref[1]**o+(-0.756893047476)*x_ref[1]
        ref[5,0,0]=(0.209926220994)*x_ref[0]**o+(0.412132938536)*x_ref[0]+(0.28119780747)*x_ref[1]**o+(-0.829072062922)*x_ref[1]
        ref[5,0,1]=(0.347615632334)*x_ref[0]**o+(-0.101954769092)*x_ref[0]+(-0.535842225724)*x_ref[1]**o+(-0.0790159065772)*x_ref[1]
        ref[5,1,0]=(-0.951105825038)*x_ref[0]**o+(-0.167113372047)*x_ref[0]+(-0.368722457931)*x_ref[1]**o+(0.00398630826759)*x_ref[1]
        ref[5,1,1]=(-0.280966220945)*x_ref[0]**o+(0.749467953656)*x_ref[0]+(-0.862358852162)*x_ref[1]**o+(-0.947469365289)*x_ref[1]
      else:
        arg[0,0,0]=(0.0415833284333)*x[0]**o+(0.194911622005)*x[0]+(0.50312866507)*x[1]**o+(0.0264724613449)*x[1]+(0.8791018499)*x[2]**o+(0.55217332601)*x[2]
        arg[0,0,1]=(0.272403975118)*x[0]**o+(0.747717699355)*x[0]+(0.630267108269)*x[1]**o+(-0.532392610953)*x[1]+(-0.335800539303)*x[2]**o+(-0.45202877276)*x[2]
        arg[0,1,0]=(-0.345230079718)*x[0]**o+(-0.463428154207)*x[0]+(-0.235612061263)*x[1]**o+(0.25205454387)*x[1]+(0.00383801658594)*x[2]**o+(0.94886179389)*x[2]
        arg[0,1,1]=(0.83425074828)*x[0]**o+(0.0678634336979)*x[0]+(0.58303694875)*x[1]**o+(-0.390423615317)*x[1]+(-0.198506665254)*x[2]**o+(-0.153926424072)*x[2]
        arg[1,0,0]=(0.748583887807)*x[0]**o+(-0.619758023753)*x[0]+(-0.637192600624)*x[1]**o+(-0.77944796466)*x[1]+(-0.745635449844)*x[2]**o+(0.385052490848)*x[2]
        arg[1,0,1]=(0.178249452349)*x[0]**o+(-0.758000510749)*x[0]+(0.173656449057)*x[1]**o+(-0.496282271341)*x[1]+(-0.75929155861)*x[2]**o+(-0.0916648747906)*x[2]
        arg[1,1,0]=(-0.106375867266)*x[0]**o+(-0.985973054508)*x[0]+(-0.828346807601)*x[1]**o+(-0.780869332747)*x[1]+(0.956166673219)*x[2]**o+(0.0115486522377)*x[2]
        arg[1,1,1]=(-0.347056651581)*x[0]**o+(0.159310474468)*x[0]+(0.169356329598)*x[1]**o+(0.0379131764296)*x[1]+(0.547302672178)*x[2]**o+(-0.192168337756)*x[2]
        arg[2,0,0]=(-0.264646719961)*x[0]**o+(-0.32990817159)*x[0]+(0.360599653268)*x[1]**o+(0.53317997944)*x[1]+(-0.288609084526)*x[2]**o+(0.117445043254)*x[2]
        arg[2,0,1]=(-0.717210086649)*x[0]**o+(-0.913859607432)*x[0]+(0.494198228468)*x[1]**o+(-0.52463221583)*x[1]+(0.711597567219)*x[2]**o+(0.585196869172)*x[2]
        arg[2,1,0]=(-0.398182449924)*x[0]**o+(-0.224237207359)*x[0]+(0.245111044519)*x[1]**o+(0.556757042489)*x[1]+(-0.747490064606)*x[2]**o+(0.974848330189)*x[2]
        arg[2,1,1]=(-0.525098204339)*x[0]**o+(-0.787485492675)*x[0]+(-0.233307380294)*x[1]**o+(-0.609570071623)*x[1]+(-0.709676638973)*x[2]**o+(0.428630406545)*x[2]
        arg[3,0,0]=(-0.604801204556)*x[0]**o+(0.483355489609)*x[0]+(0.602251424119)*x[1]**o+(-0.512483059948)*x[1]+(0.530298952371)*x[2]**o+(0.460582484929)*x[2]
        arg[3,0,1]=(-0.223026657674)*x[0]**o+(-0.843908334265)*x[0]+(0.630773219249)*x[1]**o+(-0.962463994312)*x[1]+(-0.086366368162)*x[2]**o+(0.441208496119)*x[2]
        arg[3,1,0]=(-0.861051836152)*x[0]**o+(0.282995621585)*x[0]+(-0.971513512619)*x[1]**o+(0.691494798209)*x[1]+(-0.264269824543)*x[2]**o+(0.430075879498)*x[2]
        arg[3,1,1]=(0.195995685886)*x[0]**o+(-0.441184708365)*x[0]+(-0.768193293663)*x[1]**o+(-0.308275906926)*x[1]+(0.849067046286)*x[2]**o+(-0.822000111049)*x[2]
        arg[4,0,0]=(-0.776936350216)*x[0]**o+(0.987846537986)*x[0]+(0.027119101665)*x[1]**o+(-0.0179353982234)*x[1]+(-0.453287099644)*x[2]**o+(0.970207770741)*x[2]
        arg[4,0,1]=(0.781708265017)*x[0]**o+(-0.625176653149)*x[0]+(0.0637837953361)*x[1]**o+(0.381401415932)*x[1]+(0.575142751522)*x[2]**o+(0.482009069423)*x[2]
        arg[4,1,0]=(-0.864838932985)*x[0]**o+(0.116211445658)*x[0]+(-0.103985339842)*x[1]**o+(0.616435777123)*x[1]+(0.0343826165658)*x[2]**o+(0.67732900142)*x[2]
        arg[4,1,1]=(0.637545268689)*x[0]**o+(0.495935489884)*x[0]+(-0.949001441035)*x[1]**o+(-0.600568665738)*x[1]+(0.339278287549)*x[2]**o+(0.979389021599)*x[2]
        arg[5,0,0]=(-0.486412877541)*x[0]**o+(-0.220311090368)*x[0]+(-0.941323254025)*x[1]**o+(0.122065018067)*x[1]+(0.537799827381)*x[2]**o+(0.942570757444)*x[2]
        arg[5,0,1]=(0.713697533198)*x[0]**o+(-0.382404769437)*x[0]+(0.629644547983)*x[1]**o+(-0.618351494049)*x[1]+(0.257178470315)*x[2]**o+(-0.488029666124)*x[2]
        arg[5,1,0]=(-0.531207567018)*x[0]**o+(-0.873678479376)*x[0]+(0.344856295667)*x[1]**o+(-0.401836367886)*x[1]+(0.397967958128)*x[2]**o+(-0.518169984782)*x[2]
        arg[5,1,1]=(-0.0270647126884)*x[0]**o+(-0.493698732948)*x[0]+(-0.523416157841)*x[1]**o+(-0.710897753488)*x[1]+(-0.597569566681)*x[2]**o+(0.240935331308)*x[2]
        ref[0,0,0]=(0.0415833284333)*x_ref[0]**o+(0.194911622005)*x_ref[0]+(0.50312866507)*x_ref[1]**o+(0.0264724613449)*x_ref[1]+(0.8791018499)*x_ref[2]**o+(0.55217332601)*x_ref[2]
        ref[0,0,1]=(0.272403975118)*x_ref[0]**o+(0.747717699355)*x_ref[0]+(0.630267108269)*x_ref[1]**o+(-0.532392610953)*x_ref[1]+(-0.335800539303)*x_ref[2]**o+(-0.45202877276)*x_ref[2]
        ref[0,1,0]=(-0.345230079718)*x_ref[0]**o+(-0.463428154207)*x_ref[0]+(-0.235612061263)*x_ref[1]**o+(0.25205454387)*x_ref[1]+(0.00383801658594)*x_ref[2]**o+(0.94886179389)*x_ref[2]
        ref[0,1,1]=(0.83425074828)*x_ref[0]**o+(0.0678634336979)*x_ref[0]+(0.58303694875)*x_ref[1]**o+(-0.390423615317)*x_ref[1]+(-0.198506665254)*x_ref[2]**o+(-0.153926424072)*x_ref[2]
        ref[1,0,0]=(0.748583887807)*x_ref[0]**o+(-0.619758023753)*x_ref[0]+(-0.637192600624)*x_ref[1]**o+(-0.77944796466)*x_ref[1]+(-0.745635449844)*x_ref[2]**o+(0.385052490848)*x_ref[2]
        ref[1,0,1]=(0.178249452349)*x_ref[0]**o+(-0.758000510749)*x_ref[0]+(0.173656449057)*x_ref[1]**o+(-0.496282271341)*x_ref[1]+(-0.75929155861)*x_ref[2]**o+(-0.0916648747906)*x_ref[2]
        ref[1,1,0]=(-0.106375867266)*x_ref[0]**o+(-0.985973054508)*x_ref[0]+(-0.828346807601)*x_ref[1]**o+(-0.780869332747)*x_ref[1]+(0.956166673219)*x_ref[2]**o+(0.0115486522377)*x_ref[2]
        ref[1,1,1]=(-0.347056651581)*x_ref[0]**o+(0.159310474468)*x_ref[0]+(0.169356329598)*x_ref[1]**o+(0.0379131764296)*x_ref[1]+(0.547302672178)*x_ref[2]**o+(-0.192168337756)*x_ref[2]
        ref[2,0,0]=(-0.264646719961)*x_ref[0]**o+(-0.32990817159)*x_ref[0]+(0.360599653268)*x_ref[1]**o+(0.53317997944)*x_ref[1]+(-0.288609084526)*x_ref[2]**o+(0.117445043254)*x_ref[2]
        ref[2,0,1]=(-0.717210086649)*x_ref[0]**o+(-0.913859607432)*x_ref[0]+(0.494198228468)*x_ref[1]**o+(-0.52463221583)*x_ref[1]+(0.711597567219)*x_ref[2]**o+(0.585196869172)*x_ref[2]
        ref[2,1,0]=(-0.398182449924)*x_ref[0]**o+(-0.224237207359)*x_ref[0]+(0.245111044519)*x_ref[1]**o+(0.556757042489)*x_ref[1]+(-0.747490064606)*x_ref[2]**o+(0.974848330189)*x_ref[2]
        ref[2,1,1]=(-0.525098204339)*x_ref[0]**o+(-0.787485492675)*x_ref[0]+(-0.233307380294)*x_ref[1]**o+(-0.609570071623)*x_ref[1]+(-0.709676638973)*x_ref[2]**o+(0.428630406545)*x_ref[2]
        ref[3,0,0]=(-0.604801204556)*x_ref[0]**o+(0.483355489609)*x_ref[0]+(0.602251424119)*x_ref[1]**o+(-0.512483059948)*x_ref[1]+(0.530298952371)*x_ref[2]**o+(0.460582484929)*x_ref[2]
        ref[3,0,1]=(-0.223026657674)*x_ref[0]**o+(-0.843908334265)*x_ref[0]+(0.630773219249)*x_ref[1]**o+(-0.962463994312)*x_ref[1]+(-0.086366368162)*x_ref[2]**o+(0.441208496119)*x_ref[2]
        ref[3,1,0]=(-0.861051836152)*x_ref[0]**o+(0.282995621585)*x_ref[0]+(-0.971513512619)*x_ref[1]**o+(0.691494798209)*x_ref[1]+(-0.264269824543)*x_ref[2]**o+(0.430075879498)*x_ref[2]
        ref[3,1,1]=(0.195995685886)*x_ref[0]**o+(-0.441184708365)*x_ref[0]+(-0.768193293663)*x_ref[1]**o+(-0.308275906926)*x_ref[1]+(0.849067046286)*x_ref[2]**o+(-0.822000111049)*x_ref[2]
        ref[4,0,0]=(-0.776936350216)*x_ref[0]**o+(0.987846537986)*x_ref[0]+(0.027119101665)*x_ref[1]**o+(-0.0179353982234)*x_ref[1]+(-0.453287099644)*x_ref[2]**o+(0.970207770741)*x_ref[2]
        ref[4,0,1]=(0.781708265017)*x_ref[0]**o+(-0.625176653149)*x_ref[0]+(0.0637837953361)*x_ref[1]**o+(0.381401415932)*x_ref[1]+(0.575142751522)*x_ref[2]**o+(0.482009069423)*x_ref[2]
        ref[4,1,0]=(-0.864838932985)*x_ref[0]**o+(0.116211445658)*x_ref[0]+(-0.103985339842)*x_ref[1]**o+(0.616435777123)*x_ref[1]+(0.0343826165658)*x_ref[2]**o+(0.67732900142)*x_ref[2]
        ref[4,1,1]=(0.637545268689)*x_ref[0]**o+(0.495935489884)*x_ref[0]+(-0.949001441035)*x_ref[1]**o+(-0.600568665738)*x_ref[1]+(0.339278287549)*x_ref[2]**o+(0.979389021599)*x_ref[2]
        ref[5,0,0]=(-0.486412877541)*x_ref[0]**o+(-0.220311090368)*x_ref[0]+(-0.941323254025)*x_ref[1]**o+(0.122065018067)*x_ref[1]+(0.537799827381)*x_ref[2]**o+(0.942570757444)*x_ref[2]
        ref[5,0,1]=(0.713697533198)*x_ref[0]**o+(-0.382404769437)*x_ref[0]+(0.629644547983)*x_ref[1]**o+(-0.618351494049)*x_ref[1]+(0.257178470315)*x_ref[2]**o+(-0.488029666124)*x_ref[2]
        ref[5,1,0]=(-0.531207567018)*x_ref[0]**o+(-0.873678479376)*x_ref[0]+(0.344856295667)*x_ref[1]**o+(-0.401836367886)*x_ref[1]+(0.397967958128)*x_ref[2]**o+(-0.518169984782)*x_ref[2]
        ref[5,1,1]=(-0.0270647126884)*x_ref[0]**o+(-0.493698732948)*x_ref[0]+(-0.523416157841)*x_ref[1]**o+(-0.710897753488)*x_ref[1]+(-0.597569566681)*x_ref[2]**o+(0.240935331308)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.70041600073)*x[0]**o+(0.0368444393254)*x[0]+(-0.939011570256)*x[1]**o+(0.762796835248)*x[1]
        arg[0,0,0,1]=(-0.383188465228)*x[0]**o+(-0.956202930503)*x[0]+(-0.0433250191873)*x[1]**o+(-0.263494367743)*x[1]
        arg[0,0,1,0]=(0.494889208666)*x[0]**o+(0.721800653649)*x[0]+(0.275748820606)*x[1]**o+(0.190701062508)*x[1]
        arg[0,0,1,1]=(0.317345288325)*x[0]**o+(-0.0713887739659)*x[0]+(-0.0901699882517)*x[1]**o+(0.350861486991)*x[1]
        arg[0,0,2,0]=(-0.0399664417282)*x[0]**o+(-0.73007860251)*x[0]+(-0.123301660907)*x[1]**o+(0.935774110316)*x[1]
        arg[0,0,2,1]=(-0.774484178038)*x[0]**o+(-0.427643344733)*x[0]+(0.00466934570081)*x[1]**o+(0.644433150981)*x[1]
        arg[0,1,0,0]=(0.532951038821)*x[0]**o+(-0.267404179599)*x[0]+(-0.488651484135)*x[1]**o+(-0.354416641555)*x[1]
        arg[0,1,0,1]=(0.816219473407)*x[0]**o+(0.095041670769)*x[0]+(0.130824542151)*x[1]**o+(-0.91799466789)*x[1]
        arg[0,1,1,0]=(-0.926659688215)*x[0]**o+(-0.99257560771)*x[0]+(0.76472846278)*x[1]**o+(-0.77301872817)*x[1]
        arg[0,1,1,1]=(-0.291848070507)*x[0]**o+(0.988869135358)*x[0]+(-0.355431590669)*x[1]**o+(-0.397109808154)*x[1]
        arg[0,1,2,0]=(-0.630380994644)*x[0]**o+(0.121188015244)*x[0]+(-0.527521198931)*x[1]**o+(0.743015163843)*x[1]
        arg[0,1,2,1]=(0.624350853711)*x[0]**o+(-0.269065524076)*x[0]+(0.381181149379)*x[1]**o+(0.964176747301)*x[1]
        arg[0,2,0,0]=(0.386078831223)*x[0]**o+(-0.143893625755)*x[0]+(-0.957649154953)*x[1]**o+(-0.728625629809)*x[1]
        arg[0,2,0,1]=(-0.752542988554)*x[0]**o+(-0.334338454849)*x[0]+(-0.888583790871)*x[1]**o+(-0.49515709924)*x[1]
        arg[0,2,1,0]=(0.67213418036)*x[0]**o+(0.63778853358)*x[0]+(-0.305139731169)*x[1]**o+(-0.349307973811)*x[1]
        arg[0,2,1,1]=(0.226718551492)*x[0]**o+(0.0927018122394)*x[0]+(-0.742387419573)*x[1]**o+(0.528087866126)*x[1]
        arg[0,2,2,0]=(-0.724349079038)*x[0]**o+(-0.758167121924)*x[0]+(-0.687742086176)*x[1]**o+(-0.415119729664)*x[1]
        arg[0,2,2,1]=(0.0399247837941)*x[0]**o+(-0.728504254588)*x[0]+(0.382011768693)*x[1]**o+(-0.743726886782)*x[1]
        arg[0,3,0,0]=(-0.994002993417)*x[0]**o+(0.293311671699)*x[0]+(-0.9731492726)*x[1]**o+(0.0673759518597)*x[1]
        arg[0,3,0,1]=(-0.638426251977)*x[0]**o+(-0.537134730931)*x[0]+(0.993741859056)*x[1]**o+(-0.909049905282)*x[1]
        arg[0,3,1,0]=(-0.994595072064)*x[0]**o+(0.0252993150106)*x[0]+(-0.739129165128)*x[1]**o+(-0.132376880599)*x[1]
        arg[0,3,1,1]=(-0.539219874104)*x[0]**o+(0.796849273908)*x[0]+(0.682514446887)*x[1]**o+(-0.704901251482)*x[1]
        arg[0,3,2,0]=(0.0401802421987)*x[0]**o+(-0.0354713202277)*x[0]+(-0.660064980444)*x[1]**o+(0.0477264910037)*x[1]
        arg[0,3,2,1]=(0.832824897985)*x[0]**o+(0.877654166375)*x[0]+(-0.909219867154)*x[1]**o+(-0.796258795798)*x[1]
        arg[0,4,0,0]=(0.0310731814412)*x[0]**o+(0.39714436047)*x[0]+(0.61669106719)*x[1]**o+(0.789287424452)*x[1]
        arg[0,4,0,1]=(-0.8919943646)*x[0]**o+(0.55041132884)*x[0]+(-0.0397475694211)*x[1]**o+(0.993245641481)*x[1]
        arg[0,4,1,0]=(0.550134389567)*x[0]**o+(-0.657732937381)*x[0]+(0.204770242863)*x[1]**o+(-0.0296391840031)*x[1]
        arg[0,4,1,1]=(0.753185285169)*x[0]**o+(0.74420982803)*x[0]+(0.108834676972)*x[1]**o+(0.183891847557)*x[1]
        arg[0,4,2,0]=(-0.259653246466)*x[0]**o+(-0.120308818888)*x[0]+(-0.827119024697)*x[1]**o+(-0.672357922311)*x[1]
        arg[0,4,2,1]=(-0.807666107771)*x[0]**o+(0.873748263895)*x[0]+(-0.160494957388)*x[1]**o+(0.7551018226)*x[1]
        arg[1,0,0,0]=(0.780978036182)*x[0]**o+(-0.260409851047)*x[0]+(-0.597455360543)*x[1]**o+(0.0689482125215)*x[1]
        arg[1,0,0,1]=(0.811121169417)*x[0]**o+(-0.614791318516)*x[0]+(0.663679164632)*x[1]**o+(-0.853931508608)*x[1]
        arg[1,0,1,0]=(-0.601473156596)*x[0]**o+(-0.1232863608)*x[0]+(0.200905517492)*x[1]**o+(0.337784236126)*x[1]
        arg[1,0,1,1]=(-0.924015099396)*x[0]**o+(0.197096316191)*x[0]+(-0.580606258161)*x[1]**o+(0.517410299916)*x[1]
        arg[1,0,2,0]=(0.32384939376)*x[0]**o+(0.871622229492)*x[0]+(0.403013167399)*x[1]**o+(-0.50285362219)*x[1]
        arg[1,0,2,1]=(-0.875266612786)*x[0]**o+(0.773236652169)*x[0]+(0.93949088522)*x[1]**o+(-0.607791659163)*x[1]
        arg[1,1,0,0]=(-0.972124688294)*x[0]**o+(0.895545224323)*x[0]+(-0.904953764575)*x[1]**o+(0.388629839442)*x[1]
        arg[1,1,0,1]=(0.792386948636)*x[0]**o+(-0.467370445542)*x[0]+(0.16664753332)*x[1]**o+(-0.202002832911)*x[1]
        arg[1,1,1,0]=(0.748979883957)*x[0]**o+(-0.272594191544)*x[0]+(-0.724533609431)*x[1]**o+(0.0670080576995)*x[1]
        arg[1,1,1,1]=(0.155216275716)*x[0]**o+(-0.394137469825)*x[0]+(-0.527098097796)*x[1]**o+(0.504014563188)*x[1]
        arg[1,1,2,0]=(-0.806704090439)*x[0]**o+(-0.791435275759)*x[0]+(0.866962482)*x[1]**o+(0.60302217776)*x[1]
        arg[1,1,2,1]=(-0.728053309956)*x[0]**o+(0.0559583022356)*x[0]+(0.0926294355801)*x[1]**o+(-0.835984838669)*x[1]
        arg[1,2,0,0]=(-0.0661747052735)*x[0]**o+(0.0537098973346)*x[0]+(0.310702398279)*x[1]**o+(-0.489393478975)*x[1]
        arg[1,2,0,1]=(-0.0910654471721)*x[0]**o+(0.327801831182)*x[0]+(-0.42833439187)*x[1]**o+(0.267067714743)*x[1]
        arg[1,2,1,0]=(-0.693916583922)*x[0]**o+(-0.100272651463)*x[0]+(-0.195971225883)*x[1]**o+(0.660851591786)*x[1]
        arg[1,2,1,1]=(-0.158662866194)*x[0]**o+(0.756412009601)*x[0]+(-0.923974707134)*x[1]**o+(0.61664700863)*x[1]
        arg[1,2,2,0]=(-0.449981632451)*x[0]**o+(0.272083805383)*x[0]+(-0.207156704608)*x[1]**o+(0.413596665304)*x[1]
        arg[1,2,2,1]=(0.67039491127)*x[0]**o+(-0.606081942118)*x[0]+(-0.358862791857)*x[1]**o+(-0.191412169743)*x[1]
        arg[1,3,0,0]=(0.214562206436)*x[0]**o+(0.475358432768)*x[0]+(0.639237285541)*x[1]**o+(-0.221799470417)*x[1]
        arg[1,3,0,1]=(0.468332920849)*x[0]**o+(0.359173095186)*x[0]+(0.513170435198)*x[1]**o+(0.472621770345)*x[1]
        arg[1,3,1,0]=(-0.722500106774)*x[0]**o+(0.781087784307)*x[0]+(0.633141730145)*x[1]**o+(0.958777327006)*x[1]
        arg[1,3,1,1]=(0.488269631028)*x[0]**o+(0.148575659631)*x[0]+(-0.229858507326)*x[1]**o+(-0.857118507195)*x[1]
        arg[1,3,2,0]=(0.99463505025)*x[0]**o+(-0.970275234551)*x[0]+(-0.431529655117)*x[1]**o+(-0.208112330684)*x[1]
        arg[1,3,2,1]=(-0.715411408332)*x[0]**o+(0.554401881121)*x[0]+(-0.980883542633)*x[1]**o+(0.316356494975)*x[1]
        arg[1,4,0,0]=(0.747620638111)*x[0]**o+(-0.505093883319)*x[0]+(-0.217289877652)*x[1]**o+(-0.122961101484)*x[1]
        arg[1,4,0,1]=(-0.329916740819)*x[0]**o+(0.778840575985)*x[0]+(-0.696382782242)*x[1]**o+(-0.096185679235)*x[1]
        arg[1,4,1,0]=(0.574218122656)*x[0]**o+(0.0949666147475)*x[0]+(-0.789015428301)*x[1]**o+(-0.400376747438)*x[1]
        arg[1,4,1,1]=(0.569036824328)*x[0]**o+(0.132679527923)*x[0]+(-0.229208784312)*x[1]**o+(0.405134795549)*x[1]
        arg[1,4,2,0]=(0.0123943827354)*x[0]**o+(-0.0446024014977)*x[0]+(-0.221944819571)*x[1]**o+(-0.596890233794)*x[1]
        arg[1,4,2,1]=(0.499247385143)*x[0]**o+(0.891843495388)*x[0]+(-0.348380026198)*x[1]**o+(0.146344723761)*x[1]
        arg[2,0,0,0]=(-0.929032869237)*x[0]**o+(-0.438035650633)*x[0]+(0.117030848904)*x[1]**o+(-0.489094945626)*x[1]
        arg[2,0,0,1]=(0.661960656348)*x[0]**o+(0.437346307001)*x[0]+(0.15713464744)*x[1]**o+(0.578309634731)*x[1]
        arg[2,0,1,0]=(-0.368410149772)*x[0]**o+(-0.135484081321)*x[0]+(-0.51575875197)*x[1]**o+(0.26997748887)*x[1]
        arg[2,0,1,1]=(0.784159013964)*x[0]**o+(-0.558040177257)*x[0]+(-0.100963880851)*x[1]**o+(-0.319503517522)*x[1]
        arg[2,0,2,0]=(-0.0260076717983)*x[0]**o+(-0.252140406297)*x[0]+(-0.335458650519)*x[1]**o+(-0.0739605444814)*x[1]
        arg[2,0,2,1]=(0.484818876559)*x[0]**o+(-0.59157296366)*x[0]+(-0.83491407929)*x[1]**o+(0.143217129439)*x[1]
        arg[2,1,0,0]=(0.672216209481)*x[0]**o+(0.225722770162)*x[0]+(-0.46608814621)*x[1]**o+(-0.779035303765)*x[1]
        arg[2,1,0,1]=(-0.738780643364)*x[0]**o+(0.0741487570855)*x[0]+(-0.362480267529)*x[1]**o+(0.474314990314)*x[1]
        arg[2,1,1,0]=(0.942380641613)*x[0]**o+(0.834468091419)*x[0]+(-0.413376836135)*x[1]**o+(0.996367350352)*x[1]
        arg[2,1,1,1]=(0.544360198404)*x[0]**o+(-0.327857763859)*x[0]+(-0.00306338907092)*x[1]**o+(0.448394961269)*x[1]
        arg[2,1,2,0]=(-0.345632293594)*x[0]**o+(0.505451030738)*x[0]+(-0.644095213138)*x[1]**o+(0.603588130789)*x[1]
        arg[2,1,2,1]=(-0.300809177049)*x[0]**o+(-0.531168939571)*x[0]+(0.513243419239)*x[1]**o+(0.62918148667)*x[1]
        arg[2,2,0,0]=(0.610906338615)*x[0]**o+(-0.656455478681)*x[0]+(-0.386484840223)*x[1]**o+(0.720074853711)*x[1]
        arg[2,2,0,1]=(-0.403600090425)*x[0]**o+(0.454789490943)*x[0]+(-0.399073455918)*x[1]**o+(-0.0552326255369)*x[1]
        arg[2,2,1,0]=(-0.308592576779)*x[0]**o+(0.253298307559)*x[0]+(-0.862040156595)*x[1]**o+(0.195088197599)*x[1]
        arg[2,2,1,1]=(0.143353860538)*x[0]**o+(-0.590315683091)*x[0]+(-0.955465510619)*x[1]**o+(0.372352027984)*x[1]
        arg[2,2,2,0]=(-0.13314652456)*x[0]**o+(0.580744198064)*x[0]+(-0.903035808104)*x[1]**o+(-0.193570555816)*x[1]
        arg[2,2,2,1]=(-0.647218355992)*x[0]**o+(0.00521075379725)*x[0]+(0.877022602564)*x[1]**o+(-0.104569127278)*x[1]
        arg[2,3,0,0]=(-0.532184390508)*x[0]**o+(0.739226859366)*x[0]+(0.595122128361)*x[1]**o+(0.0135063212516)*x[1]
        arg[2,3,0,1]=(-0.121073574394)*x[0]**o+(-0.175990877109)*x[0]+(-0.00319744367579)*x[1]**o+(0.140264888668)*x[1]
        arg[2,3,1,0]=(-0.0778296806357)*x[0]**o+(0.760344573573)*x[0]+(-0.250209333031)*x[1]**o+(-0.413073766011)*x[1]
        arg[2,3,1,1]=(-0.809655592364)*x[0]**o+(-0.147000481704)*x[0]+(-0.760732613787)*x[1]**o+(0.399494900513)*x[1]
        arg[2,3,2,0]=(-0.131633255327)*x[0]**o+(-0.0268295912955)*x[0]+(-0.934414490057)*x[1]**o+(-0.339746795066)*x[1]
        arg[2,3,2,1]=(0.503687469605)*x[0]**o+(-0.415050204706)*x[0]+(0.152421197486)*x[1]**o+(0.529931932615)*x[1]
        arg[2,4,0,0]=(0.930580833077)*x[0]**o+(-0.377748298816)*x[0]+(-0.159931528324)*x[1]**o+(0.374143651635)*x[1]
        arg[2,4,0,1]=(0.254339441506)*x[0]**o+(0.695672173367)*x[0]+(-0.499989877641)*x[1]**o+(-0.745148857716)*x[1]
        arg[2,4,1,0]=(0.205325187322)*x[0]**o+(0.317887475767)*x[0]+(0.755338840391)*x[1]**o+(-0.106253370416)*x[1]
        arg[2,4,1,1]=(0.175711808412)*x[0]**o+(-0.437006733188)*x[0]+(-0.449737979076)*x[1]**o+(-0.0421697213234)*x[1]
        arg[2,4,2,0]=(-0.543184540283)*x[0]**o+(0.616024774255)*x[0]+(-0.429403317932)*x[1]**o+(-0.557076956301)*x[1]
        arg[2,4,2,1]=(0.224704163386)*x[0]**o+(-0.945537746026)*x[0]+(0.947303520688)*x[1]**o+(0.909695468251)*x[1]
        arg[3,0,0,0]=(-0.671167109353)*x[0]**o+(-0.513526771737)*x[0]+(-0.511672331777)*x[1]**o+(-0.301926771817)*x[1]
        arg[3,0,0,1]=(0.39343937992)*x[0]**o+(-0.194833946828)*x[0]+(0.882702377998)*x[1]**o+(0.191544038874)*x[1]
        arg[3,0,1,0]=(0.521271797889)*x[0]**o+(0.924897421226)*x[0]+(-0.597268916449)*x[1]**o+(-0.897621830518)*x[1]
        arg[3,0,1,1]=(-0.446311941735)*x[0]**o+(-0.223997368297)*x[0]+(-0.348064954666)*x[1]**o+(0.661704488202)*x[1]
        arg[3,0,2,0]=(-0.785706283015)*x[0]**o+(0.434487336824)*x[0]+(-0.0447968030358)*x[1]**o+(0.0231287166506)*x[1]
        arg[3,0,2,1]=(0.347155629717)*x[0]**o+(-0.836660082409)*x[0]+(0.391492038747)*x[1]**o+(0.883578619998)*x[1]
        arg[3,1,0,0]=(0.571669585809)*x[0]**o+(0.966740526379)*x[0]+(-0.657144215536)*x[1]**o+(0.357025591929)*x[1]
        arg[3,1,0,1]=(-0.269900462792)*x[0]**o+(0.770680682031)*x[0]+(0.744696440247)*x[1]**o+(0.755227887237)*x[1]
        arg[3,1,1,0]=(-0.667568521409)*x[0]**o+(-0.483641971546)*x[0]+(0.687266187401)*x[1]**o+(-0.597241584922)*x[1]
        arg[3,1,1,1]=(-0.947316636771)*x[0]**o+(0.0797873425176)*x[0]+(0.254921442278)*x[1]**o+(-0.0500536165406)*x[1]
        arg[3,1,2,0]=(0.966192971983)*x[0]**o+(-0.597929938773)*x[0]+(-0.174360538)*x[1]**o+(0.46405478391)*x[1]
        arg[3,1,2,1]=(-0.675528960994)*x[0]**o+(0.51539766317)*x[0]+(-0.99673299807)*x[1]**o+(-0.488746102106)*x[1]
        arg[3,2,0,0]=(-0.604138916979)*x[0]**o+(-0.0431165687312)*x[0]+(-0.346697341122)*x[1]**o+(-0.966811623167)*x[1]
        arg[3,2,0,1]=(0.963965750281)*x[0]**o+(-0.059166688536)*x[0]+(0.166515828506)*x[1]**o+(0.832637512663)*x[1]
        arg[3,2,1,0]=(-0.808155600589)*x[0]**o+(-0.178526483687)*x[0]+(-0.012382108719)*x[1]**o+(0.194810458281)*x[1]
        arg[3,2,1,1]=(0.905922872414)*x[0]**o+(0.580755461959)*x[0]+(0.84253193847)*x[1]**o+(0.300676716124)*x[1]
        arg[3,2,2,0]=(-0.560928128044)*x[0]**o+(0.215498997249)*x[0]+(0.630359499164)*x[1]**o+(0.0955049343157)*x[1]
        arg[3,2,2,1]=(-0.977307154392)*x[0]**o+(0.543314643158)*x[0]+(-0.617962918616)*x[1]**o+(0.222318165123)*x[1]
        arg[3,3,0,0]=(0.910351367774)*x[0]**o+(-0.349015106312)*x[0]+(-0.367900620181)*x[1]**o+(-0.371131872755)*x[1]
        arg[3,3,0,1]=(0.663510744888)*x[0]**o+(-0.16380815063)*x[0]+(0.770895156906)*x[1]**o+(0.84395768389)*x[1]
        arg[3,3,1,0]=(0.553097521137)*x[0]**o+(-0.139012828014)*x[0]+(-0.986291992888)*x[1]**o+(-0.780024041157)*x[1]
        arg[3,3,1,1]=(-0.894441791323)*x[0]**o+(0.91199469535)*x[0]+(-0.349154843905)*x[1]**o+(0.538946349908)*x[1]
        arg[3,3,2,0]=(-0.464541735353)*x[0]**o+(-0.44894927329)*x[0]+(-0.488621463959)*x[1]**o+(-0.987756549629)*x[1]
        arg[3,3,2,1]=(0.136069699682)*x[0]**o+(-0.794904844715)*x[0]+(0.300038435268)*x[1]**o+(-0.599562792523)*x[1]
        arg[3,4,0,0]=(0.391227355996)*x[0]**o+(-0.634937772688)*x[0]+(0.0618326566602)*x[1]**o+(0.396379987618)*x[1]
        arg[3,4,0,1]=(0.485984857628)*x[0]**o+(-0.884694203334)*x[0]+(-0.791392877516)*x[1]**o+(0.832758707832)*x[1]
        arg[3,4,1,0]=(0.983115009543)*x[0]**o+(0.830522788331)*x[0]+(-0.947942131374)*x[1]**o+(0.890681434885)*x[1]
        arg[3,4,1,1]=(-0.0596041988514)*x[0]**o+(0.203878482428)*x[0]+(-0.739288417295)*x[1]**o+(0.150351531263)*x[1]
        arg[3,4,2,0]=(-0.821412631737)*x[0]**o+(0.366097632474)*x[0]+(-0.978999319755)*x[1]**o+(-0.556584079904)*x[1]
        arg[3,4,2,1]=(0.677346101991)*x[0]**o+(0.617303832579)*x[0]+(0.963322024959)*x[1]**o+(0.0765656809305)*x[1]
        ref[0,0,0,0]=(0.70041600073)*x_ref[0]**o+(0.0368444393254)*x_ref[0]+(-0.939011570256)*x_ref[1]**o+(0.762796835248)*x_ref[1]
        ref[0,0,0,1]=(-0.383188465228)*x_ref[0]**o+(-0.956202930503)*x_ref[0]+(-0.0433250191873)*x_ref[1]**o+(-0.263494367743)*x_ref[1]
        ref[0,0,1,0]=(0.494889208666)*x_ref[0]**o+(0.721800653649)*x_ref[0]+(0.275748820606)*x_ref[1]**o+(0.190701062508)*x_ref[1]
        ref[0,0,1,1]=(0.317345288325)*x_ref[0]**o+(-0.0713887739659)*x_ref[0]+(-0.0901699882517)*x_ref[1]**o+(0.350861486991)*x_ref[1]
        ref[0,0,2,0]=(-0.0399664417282)*x_ref[0]**o+(-0.73007860251)*x_ref[0]+(-0.123301660907)*x_ref[1]**o+(0.935774110316)*x_ref[1]
        ref[0,0,2,1]=(-0.774484178038)*x_ref[0]**o+(-0.427643344733)*x_ref[0]+(0.00466934570081)*x_ref[1]**o+(0.644433150981)*x_ref[1]
        ref[0,1,0,0]=(0.532951038821)*x_ref[0]**o+(-0.267404179599)*x_ref[0]+(-0.488651484135)*x_ref[1]**o+(-0.354416641555)*x_ref[1]
        ref[0,1,0,1]=(0.816219473407)*x_ref[0]**o+(0.095041670769)*x_ref[0]+(0.130824542151)*x_ref[1]**o+(-0.91799466789)*x_ref[1]
        ref[0,1,1,0]=(-0.926659688215)*x_ref[0]**o+(-0.99257560771)*x_ref[0]+(0.76472846278)*x_ref[1]**o+(-0.77301872817)*x_ref[1]
        ref[0,1,1,1]=(-0.291848070507)*x_ref[0]**o+(0.988869135358)*x_ref[0]+(-0.355431590669)*x_ref[1]**o+(-0.397109808154)*x_ref[1]
        ref[0,1,2,0]=(-0.630380994644)*x_ref[0]**o+(0.121188015244)*x_ref[0]+(-0.527521198931)*x_ref[1]**o+(0.743015163843)*x_ref[1]
        ref[0,1,2,1]=(0.624350853711)*x_ref[0]**o+(-0.269065524076)*x_ref[0]+(0.381181149379)*x_ref[1]**o+(0.964176747301)*x_ref[1]
        ref[0,2,0,0]=(0.386078831223)*x_ref[0]**o+(-0.143893625755)*x_ref[0]+(-0.957649154953)*x_ref[1]**o+(-0.728625629809)*x_ref[1]
        ref[0,2,0,1]=(-0.752542988554)*x_ref[0]**o+(-0.334338454849)*x_ref[0]+(-0.888583790871)*x_ref[1]**o+(-0.49515709924)*x_ref[1]
        ref[0,2,1,0]=(0.67213418036)*x_ref[0]**o+(0.63778853358)*x_ref[0]+(-0.305139731169)*x_ref[1]**o+(-0.349307973811)*x_ref[1]
        ref[0,2,1,1]=(0.226718551492)*x_ref[0]**o+(0.0927018122394)*x_ref[0]+(-0.742387419573)*x_ref[1]**o+(0.528087866126)*x_ref[1]
        ref[0,2,2,0]=(-0.724349079038)*x_ref[0]**o+(-0.758167121924)*x_ref[0]+(-0.687742086176)*x_ref[1]**o+(-0.415119729664)*x_ref[1]
        ref[0,2,2,1]=(0.0399247837941)*x_ref[0]**o+(-0.728504254588)*x_ref[0]+(0.382011768693)*x_ref[1]**o+(-0.743726886782)*x_ref[1]
        ref[0,3,0,0]=(-0.994002993417)*x_ref[0]**o+(0.293311671699)*x_ref[0]+(-0.9731492726)*x_ref[1]**o+(0.0673759518597)*x_ref[1]
        ref[0,3,0,1]=(-0.638426251977)*x_ref[0]**o+(-0.537134730931)*x_ref[0]+(0.993741859056)*x_ref[1]**o+(-0.909049905282)*x_ref[1]
        ref[0,3,1,0]=(-0.994595072064)*x_ref[0]**o+(0.0252993150106)*x_ref[0]+(-0.739129165128)*x_ref[1]**o+(-0.132376880599)*x_ref[1]
        ref[0,3,1,1]=(-0.539219874104)*x_ref[0]**o+(0.796849273908)*x_ref[0]+(0.682514446887)*x_ref[1]**o+(-0.704901251482)*x_ref[1]
        ref[0,3,2,0]=(0.0401802421987)*x_ref[0]**o+(-0.0354713202277)*x_ref[0]+(-0.660064980444)*x_ref[1]**o+(0.0477264910037)*x_ref[1]
        ref[0,3,2,1]=(0.832824897985)*x_ref[0]**o+(0.877654166375)*x_ref[0]+(-0.909219867154)*x_ref[1]**o+(-0.796258795798)*x_ref[1]
        ref[0,4,0,0]=(0.0310731814412)*x_ref[0]**o+(0.39714436047)*x_ref[0]+(0.61669106719)*x_ref[1]**o+(0.789287424452)*x_ref[1]
        ref[0,4,0,1]=(-0.8919943646)*x_ref[0]**o+(0.55041132884)*x_ref[0]+(-0.0397475694211)*x_ref[1]**o+(0.993245641481)*x_ref[1]
        ref[0,4,1,0]=(0.550134389567)*x_ref[0]**o+(-0.657732937381)*x_ref[0]+(0.204770242863)*x_ref[1]**o+(-0.0296391840031)*x_ref[1]
        ref[0,4,1,1]=(0.753185285169)*x_ref[0]**o+(0.74420982803)*x_ref[0]+(0.108834676972)*x_ref[1]**o+(0.183891847557)*x_ref[1]
        ref[0,4,2,0]=(-0.259653246466)*x_ref[0]**o+(-0.120308818888)*x_ref[0]+(-0.827119024697)*x_ref[1]**o+(-0.672357922311)*x_ref[1]
        ref[0,4,2,1]=(-0.807666107771)*x_ref[0]**o+(0.873748263895)*x_ref[0]+(-0.160494957388)*x_ref[1]**o+(0.7551018226)*x_ref[1]
        ref[1,0,0,0]=(0.780978036182)*x_ref[0]**o+(-0.260409851047)*x_ref[0]+(-0.597455360543)*x_ref[1]**o+(0.0689482125215)*x_ref[1]
        ref[1,0,0,1]=(0.811121169417)*x_ref[0]**o+(-0.614791318516)*x_ref[0]+(0.663679164632)*x_ref[1]**o+(-0.853931508608)*x_ref[1]
        ref[1,0,1,0]=(-0.601473156596)*x_ref[0]**o+(-0.1232863608)*x_ref[0]+(0.200905517492)*x_ref[1]**o+(0.337784236126)*x_ref[1]
        ref[1,0,1,1]=(-0.924015099396)*x_ref[0]**o+(0.197096316191)*x_ref[0]+(-0.580606258161)*x_ref[1]**o+(0.517410299916)*x_ref[1]
        ref[1,0,2,0]=(0.32384939376)*x_ref[0]**o+(0.871622229492)*x_ref[0]+(0.403013167399)*x_ref[1]**o+(-0.50285362219)*x_ref[1]
        ref[1,0,2,1]=(-0.875266612786)*x_ref[0]**o+(0.773236652169)*x_ref[0]+(0.93949088522)*x_ref[1]**o+(-0.607791659163)*x_ref[1]
        ref[1,1,0,0]=(-0.972124688294)*x_ref[0]**o+(0.895545224323)*x_ref[0]+(-0.904953764575)*x_ref[1]**o+(0.388629839442)*x_ref[1]
        ref[1,1,0,1]=(0.792386948636)*x_ref[0]**o+(-0.467370445542)*x_ref[0]+(0.16664753332)*x_ref[1]**o+(-0.202002832911)*x_ref[1]
        ref[1,1,1,0]=(0.748979883957)*x_ref[0]**o+(-0.272594191544)*x_ref[0]+(-0.724533609431)*x_ref[1]**o+(0.0670080576995)*x_ref[1]
        ref[1,1,1,1]=(0.155216275716)*x_ref[0]**o+(-0.394137469825)*x_ref[0]+(-0.527098097796)*x_ref[1]**o+(0.504014563188)*x_ref[1]
        ref[1,1,2,0]=(-0.806704090439)*x_ref[0]**o+(-0.791435275759)*x_ref[0]+(0.866962482)*x_ref[1]**o+(0.60302217776)*x_ref[1]
        ref[1,1,2,1]=(-0.728053309956)*x_ref[0]**o+(0.0559583022356)*x_ref[0]+(0.0926294355801)*x_ref[1]**o+(-0.835984838669)*x_ref[1]
        ref[1,2,0,0]=(-0.0661747052735)*x_ref[0]**o+(0.0537098973346)*x_ref[0]+(0.310702398279)*x_ref[1]**o+(-0.489393478975)*x_ref[1]
        ref[1,2,0,1]=(-0.0910654471721)*x_ref[0]**o+(0.327801831182)*x_ref[0]+(-0.42833439187)*x_ref[1]**o+(0.267067714743)*x_ref[1]
        ref[1,2,1,0]=(-0.693916583922)*x_ref[0]**o+(-0.100272651463)*x_ref[0]+(-0.195971225883)*x_ref[1]**o+(0.660851591786)*x_ref[1]
        ref[1,2,1,1]=(-0.158662866194)*x_ref[0]**o+(0.756412009601)*x_ref[0]+(-0.923974707134)*x_ref[1]**o+(0.61664700863)*x_ref[1]
        ref[1,2,2,0]=(-0.449981632451)*x_ref[0]**o+(0.272083805383)*x_ref[0]+(-0.207156704608)*x_ref[1]**o+(0.413596665304)*x_ref[1]
        ref[1,2,2,1]=(0.67039491127)*x_ref[0]**o+(-0.606081942118)*x_ref[0]+(-0.358862791857)*x_ref[1]**o+(-0.191412169743)*x_ref[1]
        ref[1,3,0,0]=(0.214562206436)*x_ref[0]**o+(0.475358432768)*x_ref[0]+(0.639237285541)*x_ref[1]**o+(-0.221799470417)*x_ref[1]
        ref[1,3,0,1]=(0.468332920849)*x_ref[0]**o+(0.359173095186)*x_ref[0]+(0.513170435198)*x_ref[1]**o+(0.472621770345)*x_ref[1]
        ref[1,3,1,0]=(-0.722500106774)*x_ref[0]**o+(0.781087784307)*x_ref[0]+(0.633141730145)*x_ref[1]**o+(0.958777327006)*x_ref[1]
        ref[1,3,1,1]=(0.488269631028)*x_ref[0]**o+(0.148575659631)*x_ref[0]+(-0.229858507326)*x_ref[1]**o+(-0.857118507195)*x_ref[1]
        ref[1,3,2,0]=(0.99463505025)*x_ref[0]**o+(-0.970275234551)*x_ref[0]+(-0.431529655117)*x_ref[1]**o+(-0.208112330684)*x_ref[1]
        ref[1,3,2,1]=(-0.715411408332)*x_ref[0]**o+(0.554401881121)*x_ref[0]+(-0.980883542633)*x_ref[1]**o+(0.316356494975)*x_ref[1]
        ref[1,4,0,0]=(0.747620638111)*x_ref[0]**o+(-0.505093883319)*x_ref[0]+(-0.217289877652)*x_ref[1]**o+(-0.122961101484)*x_ref[1]
        ref[1,4,0,1]=(-0.329916740819)*x_ref[0]**o+(0.778840575985)*x_ref[0]+(-0.696382782242)*x_ref[1]**o+(-0.096185679235)*x_ref[1]
        ref[1,4,1,0]=(0.574218122656)*x_ref[0]**o+(0.0949666147475)*x_ref[0]+(-0.789015428301)*x_ref[1]**o+(-0.400376747438)*x_ref[1]
        ref[1,4,1,1]=(0.569036824328)*x_ref[0]**o+(0.132679527923)*x_ref[0]+(-0.229208784312)*x_ref[1]**o+(0.405134795549)*x_ref[1]
        ref[1,4,2,0]=(0.0123943827354)*x_ref[0]**o+(-0.0446024014977)*x_ref[0]+(-0.221944819571)*x_ref[1]**o+(-0.596890233794)*x_ref[1]
        ref[1,4,2,1]=(0.499247385143)*x_ref[0]**o+(0.891843495388)*x_ref[0]+(-0.348380026198)*x_ref[1]**o+(0.146344723761)*x_ref[1]
        ref[2,0,0,0]=(-0.929032869237)*x_ref[0]**o+(-0.438035650633)*x_ref[0]+(0.117030848904)*x_ref[1]**o+(-0.489094945626)*x_ref[1]
        ref[2,0,0,1]=(0.661960656348)*x_ref[0]**o+(0.437346307001)*x_ref[0]+(0.15713464744)*x_ref[1]**o+(0.578309634731)*x_ref[1]
        ref[2,0,1,0]=(-0.368410149772)*x_ref[0]**o+(-0.135484081321)*x_ref[0]+(-0.51575875197)*x_ref[1]**o+(0.26997748887)*x_ref[1]
        ref[2,0,1,1]=(0.784159013964)*x_ref[0]**o+(-0.558040177257)*x_ref[0]+(-0.100963880851)*x_ref[1]**o+(-0.319503517522)*x_ref[1]
        ref[2,0,2,0]=(-0.0260076717983)*x_ref[0]**o+(-0.252140406297)*x_ref[0]+(-0.335458650519)*x_ref[1]**o+(-0.0739605444814)*x_ref[1]
        ref[2,0,2,1]=(0.484818876559)*x_ref[0]**o+(-0.59157296366)*x_ref[0]+(-0.83491407929)*x_ref[1]**o+(0.143217129439)*x_ref[1]
        ref[2,1,0,0]=(0.672216209481)*x_ref[0]**o+(0.225722770162)*x_ref[0]+(-0.46608814621)*x_ref[1]**o+(-0.779035303765)*x_ref[1]
        ref[2,1,0,1]=(-0.738780643364)*x_ref[0]**o+(0.0741487570855)*x_ref[0]+(-0.362480267529)*x_ref[1]**o+(0.474314990314)*x_ref[1]
        ref[2,1,1,0]=(0.942380641613)*x_ref[0]**o+(0.834468091419)*x_ref[0]+(-0.413376836135)*x_ref[1]**o+(0.996367350352)*x_ref[1]
        ref[2,1,1,1]=(0.544360198404)*x_ref[0]**o+(-0.327857763859)*x_ref[0]+(-0.00306338907092)*x_ref[1]**o+(0.448394961269)*x_ref[1]
        ref[2,1,2,0]=(-0.345632293594)*x_ref[0]**o+(0.505451030738)*x_ref[0]+(-0.644095213138)*x_ref[1]**o+(0.603588130789)*x_ref[1]
        ref[2,1,2,1]=(-0.300809177049)*x_ref[0]**o+(-0.531168939571)*x_ref[0]+(0.513243419239)*x_ref[1]**o+(0.62918148667)*x_ref[1]
        ref[2,2,0,0]=(0.610906338615)*x_ref[0]**o+(-0.656455478681)*x_ref[0]+(-0.386484840223)*x_ref[1]**o+(0.720074853711)*x_ref[1]
        ref[2,2,0,1]=(-0.403600090425)*x_ref[0]**o+(0.454789490943)*x_ref[0]+(-0.399073455918)*x_ref[1]**o+(-0.0552326255369)*x_ref[1]
        ref[2,2,1,0]=(-0.308592576779)*x_ref[0]**o+(0.253298307559)*x_ref[0]+(-0.862040156595)*x_ref[1]**o+(0.195088197599)*x_ref[1]
        ref[2,2,1,1]=(0.143353860538)*x_ref[0]**o+(-0.590315683091)*x_ref[0]+(-0.955465510619)*x_ref[1]**o+(0.372352027984)*x_ref[1]
        ref[2,2,2,0]=(-0.13314652456)*x_ref[0]**o+(0.580744198064)*x_ref[0]+(-0.903035808104)*x_ref[1]**o+(-0.193570555816)*x_ref[1]
        ref[2,2,2,1]=(-0.647218355992)*x_ref[0]**o+(0.00521075379725)*x_ref[0]+(0.877022602564)*x_ref[1]**o+(-0.104569127278)*x_ref[1]
        ref[2,3,0,0]=(-0.532184390508)*x_ref[0]**o+(0.739226859366)*x_ref[0]+(0.595122128361)*x_ref[1]**o+(0.0135063212516)*x_ref[1]
        ref[2,3,0,1]=(-0.121073574394)*x_ref[0]**o+(-0.175990877109)*x_ref[0]+(-0.00319744367579)*x_ref[1]**o+(0.140264888668)*x_ref[1]
        ref[2,3,1,0]=(-0.0778296806357)*x_ref[0]**o+(0.760344573573)*x_ref[0]+(-0.250209333031)*x_ref[1]**o+(-0.413073766011)*x_ref[1]
        ref[2,3,1,1]=(-0.809655592364)*x_ref[0]**o+(-0.147000481704)*x_ref[0]+(-0.760732613787)*x_ref[1]**o+(0.399494900513)*x_ref[1]
        ref[2,3,2,0]=(-0.131633255327)*x_ref[0]**o+(-0.0268295912955)*x_ref[0]+(-0.934414490057)*x_ref[1]**o+(-0.339746795066)*x_ref[1]
        ref[2,3,2,1]=(0.503687469605)*x_ref[0]**o+(-0.415050204706)*x_ref[0]+(0.152421197486)*x_ref[1]**o+(0.529931932615)*x_ref[1]
        ref[2,4,0,0]=(0.930580833077)*x_ref[0]**o+(-0.377748298816)*x_ref[0]+(-0.159931528324)*x_ref[1]**o+(0.374143651635)*x_ref[1]
        ref[2,4,0,1]=(0.254339441506)*x_ref[0]**o+(0.695672173367)*x_ref[0]+(-0.499989877641)*x_ref[1]**o+(-0.745148857716)*x_ref[1]
        ref[2,4,1,0]=(0.205325187322)*x_ref[0]**o+(0.317887475767)*x_ref[0]+(0.755338840391)*x_ref[1]**o+(-0.106253370416)*x_ref[1]
        ref[2,4,1,1]=(0.175711808412)*x_ref[0]**o+(-0.437006733188)*x_ref[0]+(-0.449737979076)*x_ref[1]**o+(-0.0421697213234)*x_ref[1]
        ref[2,4,2,0]=(-0.543184540283)*x_ref[0]**o+(0.616024774255)*x_ref[0]+(-0.429403317932)*x_ref[1]**o+(-0.557076956301)*x_ref[1]
        ref[2,4,2,1]=(0.224704163386)*x_ref[0]**o+(-0.945537746026)*x_ref[0]+(0.947303520688)*x_ref[1]**o+(0.909695468251)*x_ref[1]
        ref[3,0,0,0]=(-0.671167109353)*x_ref[0]**o+(-0.513526771737)*x_ref[0]+(-0.511672331777)*x_ref[1]**o+(-0.301926771817)*x_ref[1]
        ref[3,0,0,1]=(0.39343937992)*x_ref[0]**o+(-0.194833946828)*x_ref[0]+(0.882702377998)*x_ref[1]**o+(0.191544038874)*x_ref[1]
        ref[3,0,1,0]=(0.521271797889)*x_ref[0]**o+(0.924897421226)*x_ref[0]+(-0.597268916449)*x_ref[1]**o+(-0.897621830518)*x_ref[1]
        ref[3,0,1,1]=(-0.446311941735)*x_ref[0]**o+(-0.223997368297)*x_ref[0]+(-0.348064954666)*x_ref[1]**o+(0.661704488202)*x_ref[1]
        ref[3,0,2,0]=(-0.785706283015)*x_ref[0]**o+(0.434487336824)*x_ref[0]+(-0.0447968030358)*x_ref[1]**o+(0.0231287166506)*x_ref[1]
        ref[3,0,2,1]=(0.347155629717)*x_ref[0]**o+(-0.836660082409)*x_ref[0]+(0.391492038747)*x_ref[1]**o+(0.883578619998)*x_ref[1]
        ref[3,1,0,0]=(0.571669585809)*x_ref[0]**o+(0.966740526379)*x_ref[0]+(-0.657144215536)*x_ref[1]**o+(0.357025591929)*x_ref[1]
        ref[3,1,0,1]=(-0.269900462792)*x_ref[0]**o+(0.770680682031)*x_ref[0]+(0.744696440247)*x_ref[1]**o+(0.755227887237)*x_ref[1]
        ref[3,1,1,0]=(-0.667568521409)*x_ref[0]**o+(-0.483641971546)*x_ref[0]+(0.687266187401)*x_ref[1]**o+(-0.597241584922)*x_ref[1]
        ref[3,1,1,1]=(-0.947316636771)*x_ref[0]**o+(0.0797873425176)*x_ref[0]+(0.254921442278)*x_ref[1]**o+(-0.0500536165406)*x_ref[1]
        ref[3,1,2,0]=(0.966192971983)*x_ref[0]**o+(-0.597929938773)*x_ref[0]+(-0.174360538)*x_ref[1]**o+(0.46405478391)*x_ref[1]
        ref[3,1,2,1]=(-0.675528960994)*x_ref[0]**o+(0.51539766317)*x_ref[0]+(-0.99673299807)*x_ref[1]**o+(-0.488746102106)*x_ref[1]
        ref[3,2,0,0]=(-0.604138916979)*x_ref[0]**o+(-0.0431165687312)*x_ref[0]+(-0.346697341122)*x_ref[1]**o+(-0.966811623167)*x_ref[1]
        ref[3,2,0,1]=(0.963965750281)*x_ref[0]**o+(-0.059166688536)*x_ref[0]+(0.166515828506)*x_ref[1]**o+(0.832637512663)*x_ref[1]
        ref[3,2,1,0]=(-0.808155600589)*x_ref[0]**o+(-0.178526483687)*x_ref[0]+(-0.012382108719)*x_ref[1]**o+(0.194810458281)*x_ref[1]
        ref[3,2,1,1]=(0.905922872414)*x_ref[0]**o+(0.580755461959)*x_ref[0]+(0.84253193847)*x_ref[1]**o+(0.300676716124)*x_ref[1]
        ref[3,2,2,0]=(-0.560928128044)*x_ref[0]**o+(0.215498997249)*x_ref[0]+(0.630359499164)*x_ref[1]**o+(0.0955049343157)*x_ref[1]
        ref[3,2,2,1]=(-0.977307154392)*x_ref[0]**o+(0.543314643158)*x_ref[0]+(-0.617962918616)*x_ref[1]**o+(0.222318165123)*x_ref[1]
        ref[3,3,0,0]=(0.910351367774)*x_ref[0]**o+(-0.349015106312)*x_ref[0]+(-0.367900620181)*x_ref[1]**o+(-0.371131872755)*x_ref[1]
        ref[3,3,0,1]=(0.663510744888)*x_ref[0]**o+(-0.16380815063)*x_ref[0]+(0.770895156906)*x_ref[1]**o+(0.84395768389)*x_ref[1]
        ref[3,3,1,0]=(0.553097521137)*x_ref[0]**o+(-0.139012828014)*x_ref[0]+(-0.986291992888)*x_ref[1]**o+(-0.780024041157)*x_ref[1]
        ref[3,3,1,1]=(-0.894441791323)*x_ref[0]**o+(0.91199469535)*x_ref[0]+(-0.349154843905)*x_ref[1]**o+(0.538946349908)*x_ref[1]
        ref[3,3,2,0]=(-0.464541735353)*x_ref[0]**o+(-0.44894927329)*x_ref[0]+(-0.488621463959)*x_ref[1]**o+(-0.987756549629)*x_ref[1]
        ref[3,3,2,1]=(0.136069699682)*x_ref[0]**o+(-0.794904844715)*x_ref[0]+(0.300038435268)*x_ref[1]**o+(-0.599562792523)*x_ref[1]
        ref[3,4,0,0]=(0.391227355996)*x_ref[0]**o+(-0.634937772688)*x_ref[0]+(0.0618326566602)*x_ref[1]**o+(0.396379987618)*x_ref[1]
        ref[3,4,0,1]=(0.485984857628)*x_ref[0]**o+(-0.884694203334)*x_ref[0]+(-0.791392877516)*x_ref[1]**o+(0.832758707832)*x_ref[1]
        ref[3,4,1,0]=(0.983115009543)*x_ref[0]**o+(0.830522788331)*x_ref[0]+(-0.947942131374)*x_ref[1]**o+(0.890681434885)*x_ref[1]
        ref[3,4,1,1]=(-0.0596041988514)*x_ref[0]**o+(0.203878482428)*x_ref[0]+(-0.739288417295)*x_ref[1]**o+(0.150351531263)*x_ref[1]
        ref[3,4,2,0]=(-0.821412631737)*x_ref[0]**o+(0.366097632474)*x_ref[0]+(-0.978999319755)*x_ref[1]**o+(-0.556584079904)*x_ref[1]
        ref[3,4,2,1]=(0.677346101991)*x_ref[0]**o+(0.617303832579)*x_ref[0]+(0.963322024959)*x_ref[1]**o+(0.0765656809305)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.412268765993)*x[0]**o+(-0.31214585911)*x[0]+(0.153520487818)*x[1]**o+(-0.3104168266)*x[1]+(0.909798069186)*x[2]**o+(0.742916044571)*x[2]
        arg[0,0,0,1]=(0.0735781063807)*x[0]**o+(0.725527279895)*x[0]+(-0.884760504783)*x[1]**o+(0.751678022946)*x[1]+(-0.169364499242)*x[2]**o+(-0.946699946704)*x[2]
        arg[0,0,1,0]=(-0.308928255723)*x[0]**o+(-0.601764425359)*x[0]+(-0.176466074749)*x[1]**o+(-0.14838705699)*x[1]+(0.897444985593)*x[2]**o+(-0.0552696055317)*x[2]
        arg[0,0,1,1]=(0.278269305614)*x[0]**o+(-0.419731200355)*x[0]+(0.450955076279)*x[1]**o+(-0.899203207333)*x[1]+(-0.458812838787)*x[2]**o+(0.507871397895)*x[2]
        arg[0,0,2,0]=(0.887873653495)*x[0]**o+(0.130110213077)*x[0]+(0.203761425496)*x[1]**o+(-0.0563898018944)*x[1]+(0.0125548052261)*x[2]**o+(-0.594578723877)*x[2]
        arg[0,0,2,1]=(-0.104210554142)*x[0]**o+(-0.792596213906)*x[0]+(0.570869430862)*x[1]**o+(-0.798160727294)*x[1]+(0.241269856254)*x[2]**o+(0.194885978057)*x[2]
        arg[0,1,0,0]=(0.819550431682)*x[0]**o+(-0.239970258367)*x[0]+(-0.960362264523)*x[1]**o+(0.872093886273)*x[1]+(-0.857325000718)*x[2]**o+(-0.907620072404)*x[2]
        arg[0,1,0,1]=(-0.989011816788)*x[0]**o+(0.853577413178)*x[0]+(-0.160335863166)*x[1]**o+(-0.751266166108)*x[1]+(0.631146799273)*x[2]**o+(0.739741622363)*x[2]
        arg[0,1,1,0]=(0.162828987854)*x[0]**o+(-0.200652438446)*x[0]+(-0.890837153528)*x[1]**o+(-0.972029967593)*x[1]+(0.850395155009)*x[2]**o+(-0.0868272213807)*x[2]
        arg[0,1,1,1]=(-0.770711232994)*x[0]**o+(0.986518762896)*x[0]+(0.0970713645745)*x[1]**o+(0.996843418865)*x[1]+(0.208134397469)*x[2]**o+(0.946600905711)*x[2]
        arg[0,1,2,0]=(-0.543213476288)*x[0]**o+(0.0335009416664)*x[0]+(0.165751788147)*x[1]**o+(0.660276877927)*x[1]+(-0.485925597349)*x[2]**o+(-0.632302900376)*x[2]
        arg[0,1,2,1]=(0.175798920586)*x[0]**o+(-0.42873239224)*x[0]+(0.164896129395)*x[1]**o+(0.662142729145)*x[1]+(-0.425944275652)*x[2]**o+(0.306978282183)*x[2]
        arg[0,2,0,0]=(0.790303828628)*x[0]**o+(0.169346034559)*x[0]+(-0.901026918205)*x[1]**o+(0.661856377517)*x[1]+(0.194850846261)*x[2]**o+(0.155282302183)*x[2]
        arg[0,2,0,1]=(-0.43947970094)*x[0]**o+(-0.535569903832)*x[0]+(-0.534035397281)*x[1]**o+(-0.169724793112)*x[1]+(-0.717226228021)*x[2]**o+(0.771292292609)*x[2]
        arg[0,2,1,0]=(-0.424991151717)*x[0]**o+(-0.0837800409811)*x[0]+(-0.747139428246)*x[1]**o+(0.739025612454)*x[1]+(0.464486611081)*x[2]**o+(0.824926522179)*x[2]
        arg[0,2,1,1]=(-0.4666399509)*x[0]**o+(-0.479130555457)*x[0]+(-0.676718441278)*x[1]**o+(0.352685580535)*x[1]+(0.537789487962)*x[2]**o+(-0.774383681009)*x[2]
        arg[0,2,2,0]=(-0.30983510311)*x[0]**o+(0.00141603592636)*x[0]+(0.546192108985)*x[1]**o+(0.229628329192)*x[1]+(0.788466929747)*x[2]**o+(-0.169587087096)*x[2]
        arg[0,2,2,1]=(0.810282347322)*x[0]**o+(0.961056346191)*x[0]+(0.204622339898)*x[1]**o+(0.288786527052)*x[1]+(0.87096784221)*x[2]**o+(0.967565754848)*x[2]
        arg[0,3,0,0]=(-0.154669358868)*x[0]**o+(-0.449349763143)*x[0]+(-0.692997905123)*x[1]**o+(0.666308436198)*x[1]+(0.0638963436229)*x[2]**o+(-0.318233405188)*x[2]
        arg[0,3,0,1]=(0.951324381413)*x[0]**o+(0.10400502165)*x[0]+(-0.910416795399)*x[1]**o+(-0.931501370526)*x[1]+(0.455893141186)*x[2]**o+(0.536556088937)*x[2]
        arg[0,3,1,0]=(-0.315779909544)*x[0]**o+(0.189718508813)*x[0]+(0.0995068185023)*x[1]**o+(-0.863775500494)*x[1]+(0.99696081153)*x[2]**o+(0.0941784151393)*x[2]
        arg[0,3,1,1]=(0.122766381329)*x[0]**o+(-0.953220464103)*x[0]+(-0.853780873931)*x[1]**o+(-0.719623724002)*x[1]+(-0.58783153111)*x[2]**o+(-0.919075056141)*x[2]
        arg[0,3,2,0]=(-0.740349380942)*x[0]**o+(-0.0161629946027)*x[0]+(0.790593892223)*x[1]**o+(0.81563688519)*x[1]+(0.515002236312)*x[2]**o+(0.906399338047)*x[2]
        arg[0,3,2,1]=(-0.21876846821)*x[0]**o+(0.099226679818)*x[0]+(-0.501220667305)*x[1]**o+(-0.645539150063)*x[1]+(0.705501523136)*x[2]**o+(-0.640426266054)*x[2]
        arg[0,4,0,0]=(0.382430876525)*x[0]**o+(0.728752519135)*x[0]+(-0.696198853186)*x[1]**o+(0.181187001577)*x[1]+(0.818931936956)*x[2]**o+(0.267607197602)*x[2]
        arg[0,4,0,1]=(0.667658893159)*x[0]**o+(0.129176508255)*x[0]+(0.579266506883)*x[1]**o+(-0.27290178784)*x[1]+(0.804307104985)*x[2]**o+(-0.384179099369)*x[2]
        arg[0,4,1,0]=(0.472611036284)*x[0]**o+(0.532026319531)*x[0]+(-0.11900311657)*x[1]**o+(0.818635083182)*x[1]+(-0.127618053257)*x[2]**o+(-0.8464024968)*x[2]
        arg[0,4,1,1]=(-0.52572498077)*x[0]**o+(-0.222527963121)*x[0]+(-0.733344395469)*x[1]**o+(-0.519230573211)*x[1]+(0.550298161034)*x[2]**o+(-0.091253041563)*x[2]
        arg[0,4,2,0]=(-0.995556194807)*x[0]**o+(-0.157265619525)*x[0]+(0.646987111139)*x[1]**o+(0.114269194694)*x[1]+(0.787559342851)*x[2]**o+(0.655874189666)*x[2]
        arg[0,4,2,1]=(-0.534706379451)*x[0]**o+(0.89989539183)*x[0]+(0.584227019406)*x[1]**o+(-0.261650143902)*x[1]+(-0.0741247974759)*x[2]**o+(-0.646603374886)*x[2]
        arg[1,0,0,0]=(-0.16346294005)*x[0]**o+(0.515625993676)*x[0]+(-0.0294710658489)*x[1]**o+(-0.886489878353)*x[1]+(0.184376222466)*x[2]**o+(0.872455440027)*x[2]
        arg[1,0,0,1]=(-0.239569565644)*x[0]**o+(-0.0196989227301)*x[0]+(-0.333999804595)*x[1]**o+(0.79703425655)*x[1]+(0.332925813867)*x[2]**o+(0.384766118619)*x[2]
        arg[1,0,1,0]=(0.337775521215)*x[0]**o+(-0.214105347244)*x[0]+(-0.055476365922)*x[1]**o+(-0.511469291523)*x[1]+(0.390455425321)*x[2]**o+(0.346749928211)*x[2]
        arg[1,0,1,1]=(0.255034684548)*x[0]**o+(-0.975969854422)*x[0]+(-0.588154151576)*x[1]**o+(0.981825996023)*x[1]+(-0.16323013592)*x[2]**o+(-0.165779157778)*x[2]
        arg[1,0,2,0]=(-0.421647673573)*x[0]**o+(0.0892056009306)*x[0]+(-0.659566191892)*x[1]**o+(-0.640769435519)*x[1]+(-0.5474352085)*x[2]**o+(0.619855556292)*x[2]
        arg[1,0,2,1]=(0.678408254014)*x[0]**o+(0.114013125353)*x[0]+(0.222310135576)*x[1]**o+(-0.297909583371)*x[1]+(0.885716237809)*x[2]**o+(0.151576219626)*x[2]
        arg[1,1,0,0]=(-0.724825546043)*x[0]**o+(-0.796280833482)*x[0]+(0.894034526494)*x[1]**o+(-0.496334573913)*x[1]+(0.161130154754)*x[2]**o+(0.0137999166704)*x[2]
        arg[1,1,0,1]=(-0.959049520871)*x[0]**o+(0.0727868983195)*x[0]+(-0.0475524824482)*x[1]**o+(0.8098016511)*x[1]+(0.154268238454)*x[2]**o+(-0.0583213692933)*x[2]
        arg[1,1,1,0]=(-0.684320939672)*x[0]**o+(-0.841165491324)*x[0]+(0.307147508965)*x[1]**o+(0.0204906256209)*x[1]+(0.658432728833)*x[2]**o+(0.398570233826)*x[2]
        arg[1,1,1,1]=(0.921894102193)*x[0]**o+(0.33017900913)*x[0]+(0.566856960125)*x[1]**o+(0.826695387569)*x[1]+(-0.6253390282)*x[2]**o+(-0.527610353599)*x[2]
        arg[1,1,2,0]=(0.0640251272456)*x[0]**o+(0.789847887603)*x[0]+(-0.00457885872972)*x[1]**o+(-0.24681292381)*x[1]+(-0.0832674042894)*x[2]**o+(-0.254280656434)*x[2]
        arg[1,1,2,1]=(0.786208197486)*x[0]**o+(0.519575725184)*x[0]+(-0.452657209679)*x[1]**o+(0.36570806994)*x[1]+(0.875049922923)*x[2]**o+(0.881026792786)*x[2]
        arg[1,2,0,0]=(-0.415919514711)*x[0]**o+(-0.933890170641)*x[0]+(0.717990882823)*x[1]**o+(0.355555237378)*x[1]+(0.458971778535)*x[2]**o+(0.888647676837)*x[2]
        arg[1,2,0,1]=(-0.453925437723)*x[0]**o+(0.153501471971)*x[0]+(0.815830410181)*x[1]**o+(-0.519500430324)*x[1]+(-0.791530261324)*x[2]**o+(0.849179201427)*x[2]
        arg[1,2,1,0]=(0.299736697301)*x[0]**o+(0.645276033981)*x[0]+(-0.731636401343)*x[1]**o+(0.127189134157)*x[1]+(0.891427235216)*x[2]**o+(0.750052326597)*x[2]
        arg[1,2,1,1]=(-0.987392548561)*x[0]**o+(0.129053618821)*x[0]+(0.220541501123)*x[1]**o+(0.276873448998)*x[1]+(0.645231056419)*x[2]**o+(0.776634462468)*x[2]
        arg[1,2,2,0]=(-0.879738240565)*x[0]**o+(-0.0352288276658)*x[0]+(-0.247550105519)*x[1]**o+(0.782133317388)*x[1]+(-0.948666864992)*x[2]**o+(0.310357478071)*x[2]
        arg[1,2,2,1]=(-0.606778611022)*x[0]**o+(0.964356172073)*x[0]+(-0.520291905065)*x[1]**o+(-0.219411824153)*x[1]+(-0.405356968701)*x[2]**o+(0.969492630653)*x[2]
        arg[1,3,0,0]=(-0.991624055852)*x[0]**o+(-0.762240034313)*x[0]+(0.423262307551)*x[1]**o+(-0.05789904696)*x[1]+(-0.696181866276)*x[2]**o+(-0.327406198937)*x[2]
        arg[1,3,0,1]=(-0.614181964006)*x[0]**o+(0.927675019363)*x[0]+(-0.750120185372)*x[1]**o+(0.440548313639)*x[1]+(-0.116776982636)*x[2]**o+(-0.799181582827)*x[2]
        arg[1,3,1,0]=(0.286032474963)*x[0]**o+(0.391407042942)*x[0]+(-0.268363949002)*x[1]**o+(-0.421319501252)*x[1]+(-0.318595882607)*x[2]**o+(0.273120761382)*x[2]
        arg[1,3,1,1]=(0.918425098597)*x[0]**o+(-0.211260796786)*x[0]+(-0.793780530123)*x[1]**o+(-0.839335876894)*x[1]+(-0.595786792228)*x[2]**o+(-0.592690829684)*x[2]
        arg[1,3,2,0]=(-0.959108168319)*x[0]**o+(-0.94492160223)*x[0]+(0.0778030192626)*x[1]**o+(-0.169988420265)*x[1]+(0.042653831433)*x[2]**o+(-0.595426371148)*x[2]
        arg[1,3,2,1]=(0.897478414413)*x[0]**o+(-0.222864711011)*x[0]+(0.900965727203)*x[1]**o+(0.501243022513)*x[1]+(0.75496626209)*x[2]**o+(0.312225549338)*x[2]
        arg[1,4,0,0]=(0.168321984115)*x[0]**o+(-0.495995847038)*x[0]+(0.595773350289)*x[1]**o+(-0.914906345927)*x[1]+(0.224831726299)*x[2]**o+(-0.0743642212453)*x[2]
        arg[1,4,0,1]=(0.352717252696)*x[0]**o+(-0.873077111264)*x[0]+(-0.296366541486)*x[1]**o+(-0.00338438754009)*x[1]+(0.0294923296914)*x[2]**o+(-0.067479335313)*x[2]
        arg[1,4,1,0]=(-0.518757232583)*x[0]**o+(0.463948954744)*x[0]+(0.712607983792)*x[1]**o+(0.717184663437)*x[1]+(0.323391383016)*x[2]**o+(-0.225868711026)*x[2]
        arg[1,4,1,1]=(-0.24381531137)*x[0]**o+(0.630251169076)*x[0]+(0.70462966259)*x[1]**o+(-0.125357204688)*x[1]+(-0.258633435955)*x[2]**o+(0.902171387163)*x[2]
        arg[1,4,2,0]=(-0.632179111447)*x[0]**o+(-0.59290738268)*x[0]+(-0.825795231706)*x[1]**o+(0.399660693305)*x[1]+(0.0620015248102)*x[2]**o+(-0.451556558781)*x[2]
        arg[1,4,2,1]=(-0.858145688916)*x[0]**o+(-0.787993091462)*x[0]+(0.803723603768)*x[1]**o+(-0.835043986384)*x[1]+(0.467577922972)*x[2]**o+(0.819139387663)*x[2]
        arg[2,0,0,0]=(0.274886550898)*x[0]**o+(-0.0742978734335)*x[0]+(-0.773015479116)*x[1]**o+(-0.138971598539)*x[1]+(-0.702175483988)*x[2]**o+(0.9147072714)*x[2]
        arg[2,0,0,1]=(-0.598714805644)*x[0]**o+(0.11878664701)*x[0]+(-0.554608391808)*x[1]**o+(-0.734151257491)*x[1]+(0.234629053469)*x[2]**o+(0.911072319646)*x[2]
        arg[2,0,1,0]=(-0.83850906403)*x[0]**o+(0.799369711531)*x[0]+(0.436211860763)*x[1]**o+(-0.855173818525)*x[1]+(0.256241482884)*x[2]**o+(0.657883842259)*x[2]
        arg[2,0,1,1]=(-0.724328143383)*x[0]**o+(-0.781110608395)*x[0]+(-0.987464957106)*x[1]**o+(0.936448332957)*x[1]+(0.711870454531)*x[2]**o+(0.64958966746)*x[2]
        arg[2,0,2,0]=(0.485490298364)*x[0]**o+(-0.648402414524)*x[0]+(0.0493575279485)*x[1]**o+(-0.168695660031)*x[1]+(0.236500977054)*x[2]**o+(-0.699035972752)*x[2]
        arg[2,0,2,1]=(-0.508561382238)*x[0]**o+(0.0587152225749)*x[0]+(-0.92072872998)*x[1]**o+(-0.210536816562)*x[1]+(0.735465423941)*x[2]**o+(0.557002016885)*x[2]
        arg[2,1,0,0]=(-0.301683530613)*x[0]**o+(-0.501004347082)*x[0]+(-0.415125116853)*x[1]**o+(-0.975209043974)*x[1]+(0.0952958379219)*x[2]**o+(-0.882009529251)*x[2]
        arg[2,1,0,1]=(0.569047454657)*x[0]**o+(0.87107862461)*x[0]+(0.977655371485)*x[1]**o+(-0.870823933133)*x[1]+(-0.12817645442)*x[2]**o+(-0.412450404053)*x[2]
        arg[2,1,1,0]=(0.532638832559)*x[0]**o+(-0.669778224526)*x[0]+(0.516969768896)*x[1]**o+(0.725495018612)*x[1]+(0.604820640277)*x[2]**o+(-0.537483520132)*x[2]
        arg[2,1,1,1]=(-0.413776169533)*x[0]**o+(0.19037961067)*x[0]+(-0.109551737415)*x[1]**o+(-0.195599236834)*x[1]+(-0.183237365211)*x[2]**o+(0.0279625528407)*x[2]
        arg[2,1,2,0]=(-0.46089173487)*x[0]**o+(-0.151158992117)*x[0]+(0.359508089079)*x[1]**o+(-0.0808350778984)*x[1]+(0.845464278946)*x[2]**o+(0.177891463715)*x[2]
        arg[2,1,2,1]=(0.416828200435)*x[0]**o+(0.763110094817)*x[0]+(0.855595717948)*x[1]**o+(0.379358815658)*x[1]+(0.954688280021)*x[2]**o+(-0.808098938468)*x[2]
        arg[2,2,0,0]=(-0.697342667428)*x[0]**o+(-0.855608259014)*x[0]+(0.866084171318)*x[1]**o+(0.632470977708)*x[1]+(-0.399703926338)*x[2]**o+(0.211173996927)*x[2]
        arg[2,2,0,1]=(-0.940742486357)*x[0]**o+(0.249267367828)*x[0]+(0.197336418576)*x[1]**o+(-0.0573784352507)*x[1]+(-0.328108062127)*x[2]**o+(0.570204614968)*x[2]
        arg[2,2,1,0]=(0.134960606335)*x[0]**o+(0.490123643194)*x[0]+(-0.102294210304)*x[1]**o+(0.427727810449)*x[1]+(-0.158572569775)*x[2]**o+(-0.165504764483)*x[2]
        arg[2,2,1,1]=(-0.427912549906)*x[0]**o+(-0.311524081823)*x[0]+(0.220211421031)*x[1]**o+(0.123679381711)*x[1]+(0.593149361283)*x[2]**o+(-0.283233437917)*x[2]
        arg[2,2,2,0]=(0.901179489164)*x[0]**o+(0.292780095159)*x[0]+(-0.295149620179)*x[1]**o+(0.860483795735)*x[1]+(0.848872993178)*x[2]**o+(-0.612494025994)*x[2]
        arg[2,2,2,1]=(0.904689835334)*x[0]**o+(0.33689603833)*x[0]+(0.0765595244364)*x[1]**o+(0.215174063849)*x[1]+(0.973122153293)*x[2]**o+(-0.49840689696)*x[2]
        arg[2,3,0,0]=(0.973713543725)*x[0]**o+(0.0727815483094)*x[0]+(-0.0723336244169)*x[1]**o+(-0.102547958784)*x[1]+(0.573980704603)*x[2]**o+(0.784756916348)*x[2]
        arg[2,3,0,1]=(0.871042616423)*x[0]**o+(-0.966021391258)*x[0]+(0.9402258232)*x[1]**o+(0.163522806825)*x[1]+(-0.260871238896)*x[2]**o+(0.68201099906)*x[2]
        arg[2,3,1,0]=(0.916349358422)*x[0]**o+(-0.923720463388)*x[0]+(0.814116732807)*x[1]**o+(-0.521079005717)*x[1]+(-0.996963157941)*x[2]**o+(-0.251248775427)*x[2]
        arg[2,3,1,1]=(0.752568485119)*x[0]**o+(-0.134517175046)*x[0]+(0.691477362575)*x[1]**o+(-0.864047765263)*x[1]+(0.0632525427291)*x[2]**o+(0.27617797274)*x[2]
        arg[2,3,2,0]=(0.028335700298)*x[0]**o+(-0.853592964385)*x[0]+(0.554336703119)*x[1]**o+(0.0895176258833)*x[1]+(-0.932756504547)*x[2]**o+(-0.18860048352)*x[2]
        arg[2,3,2,1]=(-0.959044064582)*x[0]**o+(-0.287170878547)*x[0]+(-0.897642111161)*x[1]**o+(0.831990727144)*x[1]+(-0.92281664542)*x[2]**o+(0.96636082362)*x[2]
        arg[2,4,0,0]=(-0.626564609232)*x[0]**o+(0.20773355223)*x[0]+(0.356017212402)*x[1]**o+(0.308760707002)*x[1]+(0.206460527236)*x[2]**o+(-0.343529244774)*x[2]
        arg[2,4,0,1]=(0.176611466931)*x[0]**o+(-0.838329948167)*x[0]+(0.354707816246)*x[1]**o+(0.649522955877)*x[1]+(0.545930184768)*x[2]**o+(-0.859928910201)*x[2]
        arg[2,4,1,0]=(-0.487262434448)*x[0]**o+(0.162363232036)*x[0]+(0.0818357859487)*x[1]**o+(-0.0114240923675)*x[1]+(0.391159686245)*x[2]**o+(0.591883722488)*x[2]
        arg[2,4,1,1]=(0.89780703396)*x[0]**o+(-0.672932240376)*x[0]+(0.166200721661)*x[1]**o+(0.556829138872)*x[1]+(0.420925668076)*x[2]**o+(0.543645319787)*x[2]
        arg[2,4,2,0]=(-0.25384993317)*x[0]**o+(0.347964473446)*x[0]+(-0.579025655422)*x[1]**o+(-0.790871375452)*x[1]+(0.937422382916)*x[2]**o+(0.34952117072)*x[2]
        arg[2,4,2,1]=(-0.979549805807)*x[0]**o+(-0.514338851166)*x[0]+(0.590481348069)*x[1]**o+(0.70090898505)*x[1]+(0.0211126831373)*x[2]**o+(0.112132700392)*x[2]
        arg[3,0,0,0]=(0.712615071708)*x[0]**o+(-0.395279000567)*x[0]+(0.572303465332)*x[1]**o+(-0.18990757808)*x[1]+(0.552173988053)*x[2]**o+(0.66384838566)*x[2]
        arg[3,0,0,1]=(-0.144778596155)*x[0]**o+(0.587515130999)*x[0]+(-0.329480229216)*x[1]**o+(-0.67629797205)*x[1]+(0.132217152221)*x[2]**o+(0.688653172908)*x[2]
        arg[3,0,1,0]=(0.540370436223)*x[0]**o+(0.34677963828)*x[0]+(0.554331354626)*x[1]**o+(0.936508271818)*x[1]+(-0.375652281911)*x[2]**o+(-0.599300575121)*x[2]
        arg[3,0,1,1]=(0.534919542204)*x[0]**o+(-0.153133704085)*x[0]+(-0.121018513777)*x[1]**o+(0.850253887541)*x[1]+(-0.356227750213)*x[2]**o+(0.610895662256)*x[2]
        arg[3,0,2,0]=(-0.29877753893)*x[0]**o+(0.0284310686839)*x[0]+(0.647432584397)*x[1]**o+(-0.215364318325)*x[1]+(-0.580839810547)*x[2]**o+(0.811531717828)*x[2]
        arg[3,0,2,1]=(-0.166467908391)*x[0]**o+(0.75238635157)*x[0]+(0.387810278118)*x[1]**o+(0.516762295584)*x[1]+(-0.451298232627)*x[2]**o+(-0.877792995315)*x[2]
        arg[3,1,0,0]=(0.732332840481)*x[0]**o+(0.0567319375815)*x[0]+(-0.405681445532)*x[1]**o+(0.866920327583)*x[1]+(0.744087567632)*x[2]**o+(-0.257414527262)*x[2]
        arg[3,1,0,1]=(0.641165454863)*x[0]**o+(-0.880320701499)*x[0]+(-0.488201727677)*x[1]**o+(-0.604286576258)*x[1]+(-0.396469224601)*x[2]**o+(0.919924897798)*x[2]
        arg[3,1,1,0]=(0.276953637864)*x[0]**o+(0.825294595252)*x[0]+(-0.00681547733671)*x[1]**o+(0.610898921858)*x[1]+(-0.113640919359)*x[2]**o+(-0.41745293894)*x[2]
        arg[3,1,1,1]=(-0.0556423502541)*x[0]**o+(0.0992203569857)*x[0]+(0.26999259247)*x[1]**o+(0.687729401581)*x[1]+(-0.42626708563)*x[2]**o+(0.137838345424)*x[2]
        arg[3,1,2,0]=(-0.82862652423)*x[0]**o+(0.384688571211)*x[0]+(-0.485754978842)*x[1]**o+(0.289363174139)*x[1]+(0.809823491656)*x[2]**o+(-0.468263092294)*x[2]
        arg[3,1,2,1]=(0.330774389497)*x[0]**o+(0.29069658199)*x[0]+(-0.0852835960508)*x[1]**o+(-0.496327335153)*x[1]+(-0.701004642133)*x[2]**o+(-0.850968582466)*x[2]
        arg[3,2,0,0]=(-0.211872121822)*x[0]**o+(-0.544218825487)*x[0]+(-0.554174609604)*x[1]**o+(0.28186606841)*x[1]+(0.272511831479)*x[2]**o+(0.130059249976)*x[2]
        arg[3,2,0,1]=(0.496592232013)*x[0]**o+(-0.541037353713)*x[0]+(0.848137836475)*x[1]**o+(-0.508526127555)*x[1]+(0.468412010792)*x[2]**o+(0.0139501620456)*x[2]
        arg[3,2,1,0]=(0.987830151765)*x[0]**o+(0.356732413851)*x[0]+(0.958334312356)*x[1]**o+(0.244871130977)*x[1]+(0.200379593754)*x[2]**o+(0.105621740392)*x[2]
        arg[3,2,1,1]=(-0.751728177957)*x[0]**o+(-0.71132733669)*x[0]+(-0.792801223542)*x[1]**o+(0.932243495719)*x[1]+(0.278721571387)*x[2]**o+(0.16204818826)*x[2]
        arg[3,2,2,0]=(-0.844470031769)*x[0]**o+(0.233237514806)*x[0]+(0.567935434819)*x[1]**o+(0.315504620349)*x[1]+(-0.886427966592)*x[2]**o+(-0.514415287917)*x[2]
        arg[3,2,2,1]=(-0.113390105251)*x[0]**o+(-0.227293501546)*x[0]+(-0.501842049116)*x[1]**o+(-0.2393770551)*x[1]+(-0.636288744693)*x[2]**o+(-0.364637404491)*x[2]
        arg[3,3,0,0]=(-0.454817772137)*x[0]**o+(0.0991199341563)*x[0]+(-0.891035204309)*x[1]**o+(-0.894694305011)*x[1]+(0.476722687543)*x[2]**o+(-0.616323189052)*x[2]
        arg[3,3,0,1]=(-0.641204554575)*x[0]**o+(0.418692269871)*x[0]+(-0.497353963751)*x[1]**o+(0.443236835952)*x[1]+(0.771380772745)*x[2]**o+(0.913143060887)*x[2]
        arg[3,3,1,0]=(0.340731926682)*x[0]**o+(0.231807962221)*x[0]+(-0.35677786751)*x[1]**o+(-0.171036937234)*x[1]+(-0.697055158669)*x[2]**o+(0.58599029377)*x[2]
        arg[3,3,1,1]=(0.754473421244)*x[0]**o+(0.658054374737)*x[0]+(0.696211053347)*x[1]**o+(0.401778988971)*x[1]+(0.121272999311)*x[2]**o+(0.509498775106)*x[2]
        arg[3,3,2,0]=(-0.0930185763819)*x[0]**o+(-0.0798587560353)*x[0]+(-0.0370412597984)*x[1]**o+(-0.440515510516)*x[1]+(-0.855771461636)*x[2]**o+(0.216680759655)*x[2]
        arg[3,3,2,1]=(0.763831414202)*x[0]**o+(0.207864762366)*x[0]+(-0.815789136019)*x[1]**o+(-0.808073090924)*x[1]+(0.97581837589)*x[2]**o+(-0.560842077761)*x[2]
        arg[3,4,0,0]=(-0.241036031728)*x[0]**o+(-0.634789191932)*x[0]+(-0.321678486205)*x[1]**o+(-0.754677817943)*x[1]+(-0.078216578971)*x[2]**o+(-0.924028965218)*x[2]
        arg[3,4,0,1]=(-0.851189221011)*x[0]**o+(0.425561499671)*x[0]+(-0.293305262525)*x[1]**o+(-0.191023566093)*x[1]+(-0.810379587547)*x[2]**o+(0.593883389908)*x[2]
        arg[3,4,1,0]=(0.905139594202)*x[0]**o+(-0.904771548418)*x[0]+(0.228410726617)*x[1]**o+(0.327253374618)*x[1]+(-0.214649399974)*x[2]**o+(-0.766320611635)*x[2]
        arg[3,4,1,1]=(0.494131404603)*x[0]**o+(-0.902996362946)*x[0]+(-0.88445944326)*x[1]**o+(-0.458372750995)*x[1]+(-0.306694339897)*x[2]**o+(0.0486088169447)*x[2]
        arg[3,4,2,0]=(0.856321787222)*x[0]**o+(0.314315880734)*x[0]+(-0.996526923819)*x[1]**o+(-0.905235331678)*x[1]+(-0.775589521164)*x[2]**o+(0.812013030327)*x[2]
        arg[3,4,2,1]=(0.271281347275)*x[0]**o+(-0.600074638655)*x[0]+(-0.16332213744)*x[1]**o+(-0.483539594381)*x[1]+(0.820910011338)*x[2]**o+(0.14947223917)*x[2]
        ref[0,0,0,0]=(0.412268765993)*x_ref[0]**o+(-0.31214585911)*x_ref[0]+(0.153520487818)*x_ref[1]**o+(-0.3104168266)*x_ref[1]+(0.909798069186)*x_ref[2]**o+(0.742916044571)*x_ref[2]
        ref[0,0,0,1]=(0.0735781063807)*x_ref[0]**o+(0.725527279895)*x_ref[0]+(-0.884760504783)*x_ref[1]**o+(0.751678022946)*x_ref[1]+(-0.169364499242)*x_ref[2]**o+(-0.946699946704)*x_ref[2]
        ref[0,0,1,0]=(-0.308928255723)*x_ref[0]**o+(-0.601764425359)*x_ref[0]+(-0.176466074749)*x_ref[1]**o+(-0.14838705699)*x_ref[1]+(0.897444985593)*x_ref[2]**o+(-0.0552696055317)*x_ref[2]
        ref[0,0,1,1]=(0.278269305614)*x_ref[0]**o+(-0.419731200355)*x_ref[0]+(0.450955076279)*x_ref[1]**o+(-0.899203207333)*x_ref[1]+(-0.458812838787)*x_ref[2]**o+(0.507871397895)*x_ref[2]
        ref[0,0,2,0]=(0.887873653495)*x_ref[0]**o+(0.130110213077)*x_ref[0]+(0.203761425496)*x_ref[1]**o+(-0.0563898018944)*x_ref[1]+(0.0125548052261)*x_ref[2]**o+(-0.594578723877)*x_ref[2]
        ref[0,0,2,1]=(-0.104210554142)*x_ref[0]**o+(-0.792596213906)*x_ref[0]+(0.570869430862)*x_ref[1]**o+(-0.798160727294)*x_ref[1]+(0.241269856254)*x_ref[2]**o+(0.194885978057)*x_ref[2]
        ref[0,1,0,0]=(0.819550431682)*x_ref[0]**o+(-0.239970258367)*x_ref[0]+(-0.960362264523)*x_ref[1]**o+(0.872093886273)*x_ref[1]+(-0.857325000718)*x_ref[2]**o+(-0.907620072404)*x_ref[2]
        ref[0,1,0,1]=(-0.989011816788)*x_ref[0]**o+(0.853577413178)*x_ref[0]+(-0.160335863166)*x_ref[1]**o+(-0.751266166108)*x_ref[1]+(0.631146799273)*x_ref[2]**o+(0.739741622363)*x_ref[2]
        ref[0,1,1,0]=(0.162828987854)*x_ref[0]**o+(-0.200652438446)*x_ref[0]+(-0.890837153528)*x_ref[1]**o+(-0.972029967593)*x_ref[1]+(0.850395155009)*x_ref[2]**o+(-0.0868272213807)*x_ref[2]
        ref[0,1,1,1]=(-0.770711232994)*x_ref[0]**o+(0.986518762896)*x_ref[0]+(0.0970713645745)*x_ref[1]**o+(0.996843418865)*x_ref[1]+(0.208134397469)*x_ref[2]**o+(0.946600905711)*x_ref[2]
        ref[0,1,2,0]=(-0.543213476288)*x_ref[0]**o+(0.0335009416664)*x_ref[0]+(0.165751788147)*x_ref[1]**o+(0.660276877927)*x_ref[1]+(-0.485925597349)*x_ref[2]**o+(-0.632302900376)*x_ref[2]
        ref[0,1,2,1]=(0.175798920586)*x_ref[0]**o+(-0.42873239224)*x_ref[0]+(0.164896129395)*x_ref[1]**o+(0.662142729145)*x_ref[1]+(-0.425944275652)*x_ref[2]**o+(0.306978282183)*x_ref[2]
        ref[0,2,0,0]=(0.790303828628)*x_ref[0]**o+(0.169346034559)*x_ref[0]+(-0.901026918205)*x_ref[1]**o+(0.661856377517)*x_ref[1]+(0.194850846261)*x_ref[2]**o+(0.155282302183)*x_ref[2]
        ref[0,2,0,1]=(-0.43947970094)*x_ref[0]**o+(-0.535569903832)*x_ref[0]+(-0.534035397281)*x_ref[1]**o+(-0.169724793112)*x_ref[1]+(-0.717226228021)*x_ref[2]**o+(0.771292292609)*x_ref[2]
        ref[0,2,1,0]=(-0.424991151717)*x_ref[0]**o+(-0.0837800409811)*x_ref[0]+(-0.747139428246)*x_ref[1]**o+(0.739025612454)*x_ref[1]+(0.464486611081)*x_ref[2]**o+(0.824926522179)*x_ref[2]
        ref[0,2,1,1]=(-0.4666399509)*x_ref[0]**o+(-0.479130555457)*x_ref[0]+(-0.676718441278)*x_ref[1]**o+(0.352685580535)*x_ref[1]+(0.537789487962)*x_ref[2]**o+(-0.774383681009)*x_ref[2]
        ref[0,2,2,0]=(-0.30983510311)*x_ref[0]**o+(0.00141603592636)*x_ref[0]+(0.546192108985)*x_ref[1]**o+(0.229628329192)*x_ref[1]+(0.788466929747)*x_ref[2]**o+(-0.169587087096)*x_ref[2]
        ref[0,2,2,1]=(0.810282347322)*x_ref[0]**o+(0.961056346191)*x_ref[0]+(0.204622339898)*x_ref[1]**o+(0.288786527052)*x_ref[1]+(0.87096784221)*x_ref[2]**o+(0.967565754848)*x_ref[2]
        ref[0,3,0,0]=(-0.154669358868)*x_ref[0]**o+(-0.449349763143)*x_ref[0]+(-0.692997905123)*x_ref[1]**o+(0.666308436198)*x_ref[1]+(0.0638963436229)*x_ref[2]**o+(-0.318233405188)*x_ref[2]
        ref[0,3,0,1]=(0.951324381413)*x_ref[0]**o+(0.10400502165)*x_ref[0]+(-0.910416795399)*x_ref[1]**o+(-0.931501370526)*x_ref[1]+(0.455893141186)*x_ref[2]**o+(0.536556088937)*x_ref[2]
        ref[0,3,1,0]=(-0.315779909544)*x_ref[0]**o+(0.189718508813)*x_ref[0]+(0.0995068185023)*x_ref[1]**o+(-0.863775500494)*x_ref[1]+(0.99696081153)*x_ref[2]**o+(0.0941784151393)*x_ref[2]
        ref[0,3,1,1]=(0.122766381329)*x_ref[0]**o+(-0.953220464103)*x_ref[0]+(-0.853780873931)*x_ref[1]**o+(-0.719623724002)*x_ref[1]+(-0.58783153111)*x_ref[2]**o+(-0.919075056141)*x_ref[2]
        ref[0,3,2,0]=(-0.740349380942)*x_ref[0]**o+(-0.0161629946027)*x_ref[0]+(0.790593892223)*x_ref[1]**o+(0.81563688519)*x_ref[1]+(0.515002236312)*x_ref[2]**o+(0.906399338047)*x_ref[2]
        ref[0,3,2,1]=(-0.21876846821)*x_ref[0]**o+(0.099226679818)*x_ref[0]+(-0.501220667305)*x_ref[1]**o+(-0.645539150063)*x_ref[1]+(0.705501523136)*x_ref[2]**o+(-0.640426266054)*x_ref[2]
        ref[0,4,0,0]=(0.382430876525)*x_ref[0]**o+(0.728752519135)*x_ref[0]+(-0.696198853186)*x_ref[1]**o+(0.181187001577)*x_ref[1]+(0.818931936956)*x_ref[2]**o+(0.267607197602)*x_ref[2]
        ref[0,4,0,1]=(0.667658893159)*x_ref[0]**o+(0.129176508255)*x_ref[0]+(0.579266506883)*x_ref[1]**o+(-0.27290178784)*x_ref[1]+(0.804307104985)*x_ref[2]**o+(-0.384179099369)*x_ref[2]
        ref[0,4,1,0]=(0.472611036284)*x_ref[0]**o+(0.532026319531)*x_ref[0]+(-0.11900311657)*x_ref[1]**o+(0.818635083182)*x_ref[1]+(-0.127618053257)*x_ref[2]**o+(-0.8464024968)*x_ref[2]
        ref[0,4,1,1]=(-0.52572498077)*x_ref[0]**o+(-0.222527963121)*x_ref[0]+(-0.733344395469)*x_ref[1]**o+(-0.519230573211)*x_ref[1]+(0.550298161034)*x_ref[2]**o+(-0.091253041563)*x_ref[2]
        ref[0,4,2,0]=(-0.995556194807)*x_ref[0]**o+(-0.157265619525)*x_ref[0]+(0.646987111139)*x_ref[1]**o+(0.114269194694)*x_ref[1]+(0.787559342851)*x_ref[2]**o+(0.655874189666)*x_ref[2]
        ref[0,4,2,1]=(-0.534706379451)*x_ref[0]**o+(0.89989539183)*x_ref[0]+(0.584227019406)*x_ref[1]**o+(-0.261650143902)*x_ref[1]+(-0.0741247974759)*x_ref[2]**o+(-0.646603374886)*x_ref[2]
        ref[1,0,0,0]=(-0.16346294005)*x_ref[0]**o+(0.515625993676)*x_ref[0]+(-0.0294710658489)*x_ref[1]**o+(-0.886489878353)*x_ref[1]+(0.184376222466)*x_ref[2]**o+(0.872455440027)*x_ref[2]
        ref[1,0,0,1]=(-0.239569565644)*x_ref[0]**o+(-0.0196989227301)*x_ref[0]+(-0.333999804595)*x_ref[1]**o+(0.79703425655)*x_ref[1]+(0.332925813867)*x_ref[2]**o+(0.384766118619)*x_ref[2]
        ref[1,0,1,0]=(0.337775521215)*x_ref[0]**o+(-0.214105347244)*x_ref[0]+(-0.055476365922)*x_ref[1]**o+(-0.511469291523)*x_ref[1]+(0.390455425321)*x_ref[2]**o+(0.346749928211)*x_ref[2]
        ref[1,0,1,1]=(0.255034684548)*x_ref[0]**o+(-0.975969854422)*x_ref[0]+(-0.588154151576)*x_ref[1]**o+(0.981825996023)*x_ref[1]+(-0.16323013592)*x_ref[2]**o+(-0.165779157778)*x_ref[2]
        ref[1,0,2,0]=(-0.421647673573)*x_ref[0]**o+(0.0892056009306)*x_ref[0]+(-0.659566191892)*x_ref[1]**o+(-0.640769435519)*x_ref[1]+(-0.5474352085)*x_ref[2]**o+(0.619855556292)*x_ref[2]
        ref[1,0,2,1]=(0.678408254014)*x_ref[0]**o+(0.114013125353)*x_ref[0]+(0.222310135576)*x_ref[1]**o+(-0.297909583371)*x_ref[1]+(0.885716237809)*x_ref[2]**o+(0.151576219626)*x_ref[2]
        ref[1,1,0,0]=(-0.724825546043)*x_ref[0]**o+(-0.796280833482)*x_ref[0]+(0.894034526494)*x_ref[1]**o+(-0.496334573913)*x_ref[1]+(0.161130154754)*x_ref[2]**o+(0.0137999166704)*x_ref[2]
        ref[1,1,0,1]=(-0.959049520871)*x_ref[0]**o+(0.0727868983195)*x_ref[0]+(-0.0475524824482)*x_ref[1]**o+(0.8098016511)*x_ref[1]+(0.154268238454)*x_ref[2]**o+(-0.0583213692933)*x_ref[2]
        ref[1,1,1,0]=(-0.684320939672)*x_ref[0]**o+(-0.841165491324)*x_ref[0]+(0.307147508965)*x_ref[1]**o+(0.0204906256209)*x_ref[1]+(0.658432728833)*x_ref[2]**o+(0.398570233826)*x_ref[2]
        ref[1,1,1,1]=(0.921894102193)*x_ref[0]**o+(0.33017900913)*x_ref[0]+(0.566856960125)*x_ref[1]**o+(0.826695387569)*x_ref[1]+(-0.6253390282)*x_ref[2]**o+(-0.527610353599)*x_ref[2]
        ref[1,1,2,0]=(0.0640251272456)*x_ref[0]**o+(0.789847887603)*x_ref[0]+(-0.00457885872972)*x_ref[1]**o+(-0.24681292381)*x_ref[1]+(-0.0832674042894)*x_ref[2]**o+(-0.254280656434)*x_ref[2]
        ref[1,1,2,1]=(0.786208197486)*x_ref[0]**o+(0.519575725184)*x_ref[0]+(-0.452657209679)*x_ref[1]**o+(0.36570806994)*x_ref[1]+(0.875049922923)*x_ref[2]**o+(0.881026792786)*x_ref[2]
        ref[1,2,0,0]=(-0.415919514711)*x_ref[0]**o+(-0.933890170641)*x_ref[0]+(0.717990882823)*x_ref[1]**o+(0.355555237378)*x_ref[1]+(0.458971778535)*x_ref[2]**o+(0.888647676837)*x_ref[2]
        ref[1,2,0,1]=(-0.453925437723)*x_ref[0]**o+(0.153501471971)*x_ref[0]+(0.815830410181)*x_ref[1]**o+(-0.519500430324)*x_ref[1]+(-0.791530261324)*x_ref[2]**o+(0.849179201427)*x_ref[2]
        ref[1,2,1,0]=(0.299736697301)*x_ref[0]**o+(0.645276033981)*x_ref[0]+(-0.731636401343)*x_ref[1]**o+(0.127189134157)*x_ref[1]+(0.891427235216)*x_ref[2]**o+(0.750052326597)*x_ref[2]
        ref[1,2,1,1]=(-0.987392548561)*x_ref[0]**o+(0.129053618821)*x_ref[0]+(0.220541501123)*x_ref[1]**o+(0.276873448998)*x_ref[1]+(0.645231056419)*x_ref[2]**o+(0.776634462468)*x_ref[2]
        ref[1,2,2,0]=(-0.879738240565)*x_ref[0]**o+(-0.0352288276658)*x_ref[0]+(-0.247550105519)*x_ref[1]**o+(0.782133317388)*x_ref[1]+(-0.948666864992)*x_ref[2]**o+(0.310357478071)*x_ref[2]
        ref[1,2,2,1]=(-0.606778611022)*x_ref[0]**o+(0.964356172073)*x_ref[0]+(-0.520291905065)*x_ref[1]**o+(-0.219411824153)*x_ref[1]+(-0.405356968701)*x_ref[2]**o+(0.969492630653)*x_ref[2]
        ref[1,3,0,0]=(-0.991624055852)*x_ref[0]**o+(-0.762240034313)*x_ref[0]+(0.423262307551)*x_ref[1]**o+(-0.05789904696)*x_ref[1]+(-0.696181866276)*x_ref[2]**o+(-0.327406198937)*x_ref[2]
        ref[1,3,0,1]=(-0.614181964006)*x_ref[0]**o+(0.927675019363)*x_ref[0]+(-0.750120185372)*x_ref[1]**o+(0.440548313639)*x_ref[1]+(-0.116776982636)*x_ref[2]**o+(-0.799181582827)*x_ref[2]
        ref[1,3,1,0]=(0.286032474963)*x_ref[0]**o+(0.391407042942)*x_ref[0]+(-0.268363949002)*x_ref[1]**o+(-0.421319501252)*x_ref[1]+(-0.318595882607)*x_ref[2]**o+(0.273120761382)*x_ref[2]
        ref[1,3,1,1]=(0.918425098597)*x_ref[0]**o+(-0.211260796786)*x_ref[0]+(-0.793780530123)*x_ref[1]**o+(-0.839335876894)*x_ref[1]+(-0.595786792228)*x_ref[2]**o+(-0.592690829684)*x_ref[2]
        ref[1,3,2,0]=(-0.959108168319)*x_ref[0]**o+(-0.94492160223)*x_ref[0]+(0.0778030192626)*x_ref[1]**o+(-0.169988420265)*x_ref[1]+(0.042653831433)*x_ref[2]**o+(-0.595426371148)*x_ref[2]
        ref[1,3,2,1]=(0.897478414413)*x_ref[0]**o+(-0.222864711011)*x_ref[0]+(0.900965727203)*x_ref[1]**o+(0.501243022513)*x_ref[1]+(0.75496626209)*x_ref[2]**o+(0.312225549338)*x_ref[2]
        ref[1,4,0,0]=(0.168321984115)*x_ref[0]**o+(-0.495995847038)*x_ref[0]+(0.595773350289)*x_ref[1]**o+(-0.914906345927)*x_ref[1]+(0.224831726299)*x_ref[2]**o+(-0.0743642212453)*x_ref[2]
        ref[1,4,0,1]=(0.352717252696)*x_ref[0]**o+(-0.873077111264)*x_ref[0]+(-0.296366541486)*x_ref[1]**o+(-0.00338438754009)*x_ref[1]+(0.0294923296914)*x_ref[2]**o+(-0.067479335313)*x_ref[2]
        ref[1,4,1,0]=(-0.518757232583)*x_ref[0]**o+(0.463948954744)*x_ref[0]+(0.712607983792)*x_ref[1]**o+(0.717184663437)*x_ref[1]+(0.323391383016)*x_ref[2]**o+(-0.225868711026)*x_ref[2]
        ref[1,4,1,1]=(-0.24381531137)*x_ref[0]**o+(0.630251169076)*x_ref[0]+(0.70462966259)*x_ref[1]**o+(-0.125357204688)*x_ref[1]+(-0.258633435955)*x_ref[2]**o+(0.902171387163)*x_ref[2]
        ref[1,4,2,0]=(-0.632179111447)*x_ref[0]**o+(-0.59290738268)*x_ref[0]+(-0.825795231706)*x_ref[1]**o+(0.399660693305)*x_ref[1]+(0.0620015248102)*x_ref[2]**o+(-0.451556558781)*x_ref[2]
        ref[1,4,2,1]=(-0.858145688916)*x_ref[0]**o+(-0.787993091462)*x_ref[0]+(0.803723603768)*x_ref[1]**o+(-0.835043986384)*x_ref[1]+(0.467577922972)*x_ref[2]**o+(0.819139387663)*x_ref[2]
        ref[2,0,0,0]=(0.274886550898)*x_ref[0]**o+(-0.0742978734335)*x_ref[0]+(-0.773015479116)*x_ref[1]**o+(-0.138971598539)*x_ref[1]+(-0.702175483988)*x_ref[2]**o+(0.9147072714)*x_ref[2]
        ref[2,0,0,1]=(-0.598714805644)*x_ref[0]**o+(0.11878664701)*x_ref[0]+(-0.554608391808)*x_ref[1]**o+(-0.734151257491)*x_ref[1]+(0.234629053469)*x_ref[2]**o+(0.911072319646)*x_ref[2]
        ref[2,0,1,0]=(-0.83850906403)*x_ref[0]**o+(0.799369711531)*x_ref[0]+(0.436211860763)*x_ref[1]**o+(-0.855173818525)*x_ref[1]+(0.256241482884)*x_ref[2]**o+(0.657883842259)*x_ref[2]
        ref[2,0,1,1]=(-0.724328143383)*x_ref[0]**o+(-0.781110608395)*x_ref[0]+(-0.987464957106)*x_ref[1]**o+(0.936448332957)*x_ref[1]+(0.711870454531)*x_ref[2]**o+(0.64958966746)*x_ref[2]
        ref[2,0,2,0]=(0.485490298364)*x_ref[0]**o+(-0.648402414524)*x_ref[0]+(0.0493575279485)*x_ref[1]**o+(-0.168695660031)*x_ref[1]+(0.236500977054)*x_ref[2]**o+(-0.699035972752)*x_ref[2]
        ref[2,0,2,1]=(-0.508561382238)*x_ref[0]**o+(0.0587152225749)*x_ref[0]+(-0.92072872998)*x_ref[1]**o+(-0.210536816562)*x_ref[1]+(0.735465423941)*x_ref[2]**o+(0.557002016885)*x_ref[2]
        ref[2,1,0,0]=(-0.301683530613)*x_ref[0]**o+(-0.501004347082)*x_ref[0]+(-0.415125116853)*x_ref[1]**o+(-0.975209043974)*x_ref[1]+(0.0952958379219)*x_ref[2]**o+(-0.882009529251)*x_ref[2]
        ref[2,1,0,1]=(0.569047454657)*x_ref[0]**o+(0.87107862461)*x_ref[0]+(0.977655371485)*x_ref[1]**o+(-0.870823933133)*x_ref[1]+(-0.12817645442)*x_ref[2]**o+(-0.412450404053)*x_ref[2]
        ref[2,1,1,0]=(0.532638832559)*x_ref[0]**o+(-0.669778224526)*x_ref[0]+(0.516969768896)*x_ref[1]**o+(0.725495018612)*x_ref[1]+(0.604820640277)*x_ref[2]**o+(-0.537483520132)*x_ref[2]
        ref[2,1,1,1]=(-0.413776169533)*x_ref[0]**o+(0.19037961067)*x_ref[0]+(-0.109551737415)*x_ref[1]**o+(-0.195599236834)*x_ref[1]+(-0.183237365211)*x_ref[2]**o+(0.0279625528407)*x_ref[2]
        ref[2,1,2,0]=(-0.46089173487)*x_ref[0]**o+(-0.151158992117)*x_ref[0]+(0.359508089079)*x_ref[1]**o+(-0.0808350778984)*x_ref[1]+(0.845464278946)*x_ref[2]**o+(0.177891463715)*x_ref[2]
        ref[2,1,2,1]=(0.416828200435)*x_ref[0]**o+(0.763110094817)*x_ref[0]+(0.855595717948)*x_ref[1]**o+(0.379358815658)*x_ref[1]+(0.954688280021)*x_ref[2]**o+(-0.808098938468)*x_ref[2]
        ref[2,2,0,0]=(-0.697342667428)*x_ref[0]**o+(-0.855608259014)*x_ref[0]+(0.866084171318)*x_ref[1]**o+(0.632470977708)*x_ref[1]+(-0.399703926338)*x_ref[2]**o+(0.211173996927)*x_ref[2]
        ref[2,2,0,1]=(-0.940742486357)*x_ref[0]**o+(0.249267367828)*x_ref[0]+(0.197336418576)*x_ref[1]**o+(-0.0573784352507)*x_ref[1]+(-0.328108062127)*x_ref[2]**o+(0.570204614968)*x_ref[2]
        ref[2,2,1,0]=(0.134960606335)*x_ref[0]**o+(0.490123643194)*x_ref[0]+(-0.102294210304)*x_ref[1]**o+(0.427727810449)*x_ref[1]+(-0.158572569775)*x_ref[2]**o+(-0.165504764483)*x_ref[2]
        ref[2,2,1,1]=(-0.427912549906)*x_ref[0]**o+(-0.311524081823)*x_ref[0]+(0.220211421031)*x_ref[1]**o+(0.123679381711)*x_ref[1]+(0.593149361283)*x_ref[2]**o+(-0.283233437917)*x_ref[2]
        ref[2,2,2,0]=(0.901179489164)*x_ref[0]**o+(0.292780095159)*x_ref[0]+(-0.295149620179)*x_ref[1]**o+(0.860483795735)*x_ref[1]+(0.848872993178)*x_ref[2]**o+(-0.612494025994)*x_ref[2]
        ref[2,2,2,1]=(0.904689835334)*x_ref[0]**o+(0.33689603833)*x_ref[0]+(0.0765595244364)*x_ref[1]**o+(0.215174063849)*x_ref[1]+(0.973122153293)*x_ref[2]**o+(-0.49840689696)*x_ref[2]
        ref[2,3,0,0]=(0.973713543725)*x_ref[0]**o+(0.0727815483094)*x_ref[0]+(-0.0723336244169)*x_ref[1]**o+(-0.102547958784)*x_ref[1]+(0.573980704603)*x_ref[2]**o+(0.784756916348)*x_ref[2]
        ref[2,3,0,1]=(0.871042616423)*x_ref[0]**o+(-0.966021391258)*x_ref[0]+(0.9402258232)*x_ref[1]**o+(0.163522806825)*x_ref[1]+(-0.260871238896)*x_ref[2]**o+(0.68201099906)*x_ref[2]
        ref[2,3,1,0]=(0.916349358422)*x_ref[0]**o+(-0.923720463388)*x_ref[0]+(0.814116732807)*x_ref[1]**o+(-0.521079005717)*x_ref[1]+(-0.996963157941)*x_ref[2]**o+(-0.251248775427)*x_ref[2]
        ref[2,3,1,1]=(0.752568485119)*x_ref[0]**o+(-0.134517175046)*x_ref[0]+(0.691477362575)*x_ref[1]**o+(-0.864047765263)*x_ref[1]+(0.0632525427291)*x_ref[2]**o+(0.27617797274)*x_ref[2]
        ref[2,3,2,0]=(0.028335700298)*x_ref[0]**o+(-0.853592964385)*x_ref[0]+(0.554336703119)*x_ref[1]**o+(0.0895176258833)*x_ref[1]+(-0.932756504547)*x_ref[2]**o+(-0.18860048352)*x_ref[2]
        ref[2,3,2,1]=(-0.959044064582)*x_ref[0]**o+(-0.287170878547)*x_ref[0]+(-0.897642111161)*x_ref[1]**o+(0.831990727144)*x_ref[1]+(-0.92281664542)*x_ref[2]**o+(0.96636082362)*x_ref[2]
        ref[2,4,0,0]=(-0.626564609232)*x_ref[0]**o+(0.20773355223)*x_ref[0]+(0.356017212402)*x_ref[1]**o+(0.308760707002)*x_ref[1]+(0.206460527236)*x_ref[2]**o+(-0.343529244774)*x_ref[2]
        ref[2,4,0,1]=(0.176611466931)*x_ref[0]**o+(-0.838329948167)*x_ref[0]+(0.354707816246)*x_ref[1]**o+(0.649522955877)*x_ref[1]+(0.545930184768)*x_ref[2]**o+(-0.859928910201)*x_ref[2]
        ref[2,4,1,0]=(-0.487262434448)*x_ref[0]**o+(0.162363232036)*x_ref[0]+(0.0818357859487)*x_ref[1]**o+(-0.0114240923675)*x_ref[1]+(0.391159686245)*x_ref[2]**o+(0.591883722488)*x_ref[2]
        ref[2,4,1,1]=(0.89780703396)*x_ref[0]**o+(-0.672932240376)*x_ref[0]+(0.166200721661)*x_ref[1]**o+(0.556829138872)*x_ref[1]+(0.420925668076)*x_ref[2]**o+(0.543645319787)*x_ref[2]
        ref[2,4,2,0]=(-0.25384993317)*x_ref[0]**o+(0.347964473446)*x_ref[0]+(-0.579025655422)*x_ref[1]**o+(-0.790871375452)*x_ref[1]+(0.937422382916)*x_ref[2]**o+(0.34952117072)*x_ref[2]
        ref[2,4,2,1]=(-0.979549805807)*x_ref[0]**o+(-0.514338851166)*x_ref[0]+(0.590481348069)*x_ref[1]**o+(0.70090898505)*x_ref[1]+(0.0211126831373)*x_ref[2]**o+(0.112132700392)*x_ref[2]
        ref[3,0,0,0]=(0.712615071708)*x_ref[0]**o+(-0.395279000567)*x_ref[0]+(0.572303465332)*x_ref[1]**o+(-0.18990757808)*x_ref[1]+(0.552173988053)*x_ref[2]**o+(0.66384838566)*x_ref[2]
        ref[3,0,0,1]=(-0.144778596155)*x_ref[0]**o+(0.587515130999)*x_ref[0]+(-0.329480229216)*x_ref[1]**o+(-0.67629797205)*x_ref[1]+(0.132217152221)*x_ref[2]**o+(0.688653172908)*x_ref[2]
        ref[3,0,1,0]=(0.540370436223)*x_ref[0]**o+(0.34677963828)*x_ref[0]+(0.554331354626)*x_ref[1]**o+(0.936508271818)*x_ref[1]+(-0.375652281911)*x_ref[2]**o+(-0.599300575121)*x_ref[2]
        ref[3,0,1,1]=(0.534919542204)*x_ref[0]**o+(-0.153133704085)*x_ref[0]+(-0.121018513777)*x_ref[1]**o+(0.850253887541)*x_ref[1]+(-0.356227750213)*x_ref[2]**o+(0.610895662256)*x_ref[2]
        ref[3,0,2,0]=(-0.29877753893)*x_ref[0]**o+(0.0284310686839)*x_ref[0]+(0.647432584397)*x_ref[1]**o+(-0.215364318325)*x_ref[1]+(-0.580839810547)*x_ref[2]**o+(0.811531717828)*x_ref[2]
        ref[3,0,2,1]=(-0.166467908391)*x_ref[0]**o+(0.75238635157)*x_ref[0]+(0.387810278118)*x_ref[1]**o+(0.516762295584)*x_ref[1]+(-0.451298232627)*x_ref[2]**o+(-0.877792995315)*x_ref[2]
        ref[3,1,0,0]=(0.732332840481)*x_ref[0]**o+(0.0567319375815)*x_ref[0]+(-0.405681445532)*x_ref[1]**o+(0.866920327583)*x_ref[1]+(0.744087567632)*x_ref[2]**o+(-0.257414527262)*x_ref[2]
        ref[3,1,0,1]=(0.641165454863)*x_ref[0]**o+(-0.880320701499)*x_ref[0]+(-0.488201727677)*x_ref[1]**o+(-0.604286576258)*x_ref[1]+(-0.396469224601)*x_ref[2]**o+(0.919924897798)*x_ref[2]
        ref[3,1,1,0]=(0.276953637864)*x_ref[0]**o+(0.825294595252)*x_ref[0]+(-0.00681547733671)*x_ref[1]**o+(0.610898921858)*x_ref[1]+(-0.113640919359)*x_ref[2]**o+(-0.41745293894)*x_ref[2]
        ref[3,1,1,1]=(-0.0556423502541)*x_ref[0]**o+(0.0992203569857)*x_ref[0]+(0.26999259247)*x_ref[1]**o+(0.687729401581)*x_ref[1]+(-0.42626708563)*x_ref[2]**o+(0.137838345424)*x_ref[2]
        ref[3,1,2,0]=(-0.82862652423)*x_ref[0]**o+(0.384688571211)*x_ref[0]+(-0.485754978842)*x_ref[1]**o+(0.289363174139)*x_ref[1]+(0.809823491656)*x_ref[2]**o+(-0.468263092294)*x_ref[2]
        ref[3,1,2,1]=(0.330774389497)*x_ref[0]**o+(0.29069658199)*x_ref[0]+(-0.0852835960508)*x_ref[1]**o+(-0.496327335153)*x_ref[1]+(-0.701004642133)*x_ref[2]**o+(-0.850968582466)*x_ref[2]
        ref[3,2,0,0]=(-0.211872121822)*x_ref[0]**o+(-0.544218825487)*x_ref[0]+(-0.554174609604)*x_ref[1]**o+(0.28186606841)*x_ref[1]+(0.272511831479)*x_ref[2]**o+(0.130059249976)*x_ref[2]
        ref[3,2,0,1]=(0.496592232013)*x_ref[0]**o+(-0.541037353713)*x_ref[0]+(0.848137836475)*x_ref[1]**o+(-0.508526127555)*x_ref[1]+(0.468412010792)*x_ref[2]**o+(0.0139501620456)*x_ref[2]
        ref[3,2,1,0]=(0.987830151765)*x_ref[0]**o+(0.356732413851)*x_ref[0]+(0.958334312356)*x_ref[1]**o+(0.244871130977)*x_ref[1]+(0.200379593754)*x_ref[2]**o+(0.105621740392)*x_ref[2]
        ref[3,2,1,1]=(-0.751728177957)*x_ref[0]**o+(-0.71132733669)*x_ref[0]+(-0.792801223542)*x_ref[1]**o+(0.932243495719)*x_ref[1]+(0.278721571387)*x_ref[2]**o+(0.16204818826)*x_ref[2]
        ref[3,2,2,0]=(-0.844470031769)*x_ref[0]**o+(0.233237514806)*x_ref[0]+(0.567935434819)*x_ref[1]**o+(0.315504620349)*x_ref[1]+(-0.886427966592)*x_ref[2]**o+(-0.514415287917)*x_ref[2]
        ref[3,2,2,1]=(-0.113390105251)*x_ref[0]**o+(-0.227293501546)*x_ref[0]+(-0.501842049116)*x_ref[1]**o+(-0.2393770551)*x_ref[1]+(-0.636288744693)*x_ref[2]**o+(-0.364637404491)*x_ref[2]
        ref[3,3,0,0]=(-0.454817772137)*x_ref[0]**o+(0.0991199341563)*x_ref[0]+(-0.891035204309)*x_ref[1]**o+(-0.894694305011)*x_ref[1]+(0.476722687543)*x_ref[2]**o+(-0.616323189052)*x_ref[2]
        ref[3,3,0,1]=(-0.641204554575)*x_ref[0]**o+(0.418692269871)*x_ref[0]+(-0.497353963751)*x_ref[1]**o+(0.443236835952)*x_ref[1]+(0.771380772745)*x_ref[2]**o+(0.913143060887)*x_ref[2]
        ref[3,3,1,0]=(0.340731926682)*x_ref[0]**o+(0.231807962221)*x_ref[0]+(-0.35677786751)*x_ref[1]**o+(-0.171036937234)*x_ref[1]+(-0.697055158669)*x_ref[2]**o+(0.58599029377)*x_ref[2]
        ref[3,3,1,1]=(0.754473421244)*x_ref[0]**o+(0.658054374737)*x_ref[0]+(0.696211053347)*x_ref[1]**o+(0.401778988971)*x_ref[1]+(0.121272999311)*x_ref[2]**o+(0.509498775106)*x_ref[2]
        ref[3,3,2,0]=(-0.0930185763819)*x_ref[0]**o+(-0.0798587560353)*x_ref[0]+(-0.0370412597984)*x_ref[1]**o+(-0.440515510516)*x_ref[1]+(-0.855771461636)*x_ref[2]**o+(0.216680759655)*x_ref[2]
        ref[3,3,2,1]=(0.763831414202)*x_ref[0]**o+(0.207864762366)*x_ref[0]+(-0.815789136019)*x_ref[1]**o+(-0.808073090924)*x_ref[1]+(0.97581837589)*x_ref[2]**o+(-0.560842077761)*x_ref[2]
        ref[3,4,0,0]=(-0.241036031728)*x_ref[0]**o+(-0.634789191932)*x_ref[0]+(-0.321678486205)*x_ref[1]**o+(-0.754677817943)*x_ref[1]+(-0.078216578971)*x_ref[2]**o+(-0.924028965218)*x_ref[2]
        ref[3,4,0,1]=(-0.851189221011)*x_ref[0]**o+(0.425561499671)*x_ref[0]+(-0.293305262525)*x_ref[1]**o+(-0.191023566093)*x_ref[1]+(-0.810379587547)*x_ref[2]**o+(0.593883389908)*x_ref[2]
        ref[3,4,1,0]=(0.905139594202)*x_ref[0]**o+(-0.904771548418)*x_ref[0]+(0.228410726617)*x_ref[1]**o+(0.327253374618)*x_ref[1]+(-0.214649399974)*x_ref[2]**o+(-0.766320611635)*x_ref[2]
        ref[3,4,1,1]=(0.494131404603)*x_ref[0]**o+(-0.902996362946)*x_ref[0]+(-0.88445944326)*x_ref[1]**o+(-0.458372750995)*x_ref[1]+(-0.306694339897)*x_ref[2]**o+(0.0486088169447)*x_ref[2]
        ref[3,4,2,0]=(0.856321787222)*x_ref[0]**o+(0.314315880734)*x_ref[0]+(-0.996526923819)*x_ref[1]**o+(-0.905235331678)*x_ref[1]+(-0.775589521164)*x_ref[2]**o+(0.812013030327)*x_ref[2]
        ref[3,4,2,1]=(0.271281347275)*x_ref[0]**o+(-0.600074638655)*x_ref[0]+(-0.16332213744)*x_ref[1]**o+(-0.483539594381)*x_ref[1]+(0.820910011338)*x_ref[2]**o+(0.14947223917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0368926203259)*x[0]**o+(-0.256519664931)*x[0]+(0.168576661932)*x[1]**o+(-0.231158401549)*x[1]
        ref=(0.0368926203259)*x_ref[0]**o+(-0.256519664931)*x_ref[0]+(0.168576661932)*x_ref[1]**o+(-0.231158401549)*x_ref[1]
      else:
        arg=(0.597124147524)*x[0]**o+(-0.404660180764)*x[0]+(-0.923138194509)*x[1]**o+(-0.263963516316)*x[1]+(-0.897643246613)*x[2]**o+(0.0180371552704)*x[2]
        ref=(0.597124147524)*x_ref[0]**o+(-0.404660180764)*x_ref[0]+(-0.923138194509)*x_ref[1]**o+(-0.263963516316)*x_ref[1]+(-0.897643246613)*x_ref[2]**o+(0.0180371552704)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.639330723902)*x[0]**o+(0.404929658475)*x[0]+(0.000241871115531)*x[1]**o+(0.770530713516)*x[1]
        arg[1]=(0.245348516629)*x[0]**o+(-0.00142757994203)*x[0]+(0.297525985266)*x[1]**o+(0.683033847851)*x[1]
        ref[0]=(0.639330723902)*x_ref[0]**o+(0.404929658475)*x_ref[0]+(0.000241871115531)*x_ref[1]**o+(0.770530713516)*x_ref[1]
        ref[1]=(0.245348516629)*x_ref[0]**o+(-0.00142757994203)*x_ref[0]+(0.297525985266)*x_ref[1]**o+(0.683033847851)*x_ref[1]
      else:
        arg[0]=(-0.45360978548)*x[0]**o+(-0.069702953017)*x[0]+(0.745603275442)*x[1]**o+(0.591869498191)*x[1]+(-0.90835287387)*x[2]**o+(-0.0405228474208)*x[2]
        arg[1]=(-0.659979134044)*x[0]**o+(0.393634998529)*x[0]+(-0.685193662459)*x[1]**o+(0.773497371593)*x[1]+(0.139726737595)*x[2]**o+(-0.549174223126)*x[2]
        ref[0]=(-0.45360978548)*x_ref[0]**o+(-0.069702953017)*x_ref[0]+(0.745603275442)*x_ref[1]**o+(0.591869498191)*x_ref[1]+(-0.90835287387)*x_ref[2]**o+(-0.0405228474208)*x_ref[2]
        ref[1]=(-0.659979134044)*x_ref[0]**o+(0.393634998529)*x_ref[0]+(-0.685193662459)*x_ref[1]**o+(0.773497371593)*x_ref[1]+(0.139726737595)*x_ref[2]**o+(-0.549174223126)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0734861043324)*x[0]**o+(0.698132844974)*x[0]+(-0.0711338889844)*x[1]**o+(0.667313854486)*x[1]
        arg[0,1]=(0.846132724611)*x[0]**o+(-0.215521619475)*x[0]+(-0.920637540655)*x[1]**o+(0.666739063184)*x[1]
        arg[0,2]=(-0.963337851184)*x[0]**o+(-0.961935972833)*x[0]+(-0.0689798144115)*x[1]**o+(0.0106557770119)*x[1]
        arg[0,3]=(0.875063015269)*x[0]**o+(-0.735727431445)*x[0]+(-0.599251007104)*x[1]**o+(-0.274659381946)*x[1]
        arg[0,4]=(-0.378214565543)*x[0]**o+(-0.476230056857)*x[0]+(0.617722084788)*x[1]**o+(-0.808732634287)*x[1]
        arg[1,0]=(-0.324652264865)*x[0]**o+(-0.314957074449)*x[0]+(0.467828791079)*x[1]**o+(-0.105499299461)*x[1]
        arg[1,1]=(-0.0815291997441)*x[0]**o+(0.1473072227)*x[0]+(-0.866265670368)*x[1]**o+(-0.66256092454)*x[1]
        arg[1,2]=(-0.83962092333)*x[0]**o+(0.681365396039)*x[0]+(-0.607267339395)*x[1]**o+(0.278637239291)*x[1]
        arg[1,3]=(-0.300965517299)*x[0]**o+(-0.0884159654351)*x[0]+(-0.547465656655)*x[1]**o+(0.708740949324)*x[1]
        arg[1,4]=(-0.0250728833518)*x[0]**o+(0.306828709412)*x[0]+(-0.972269802286)*x[1]**o+(0.619649412235)*x[1]
        arg[2,0]=(-0.873228641325)*x[0]**o+(-0.184974115411)*x[0]+(-0.500142861985)*x[1]**o+(0.708256240395)*x[1]
        arg[2,1]=(-0.588540228523)*x[0]**o+(-0.747378078994)*x[0]+(0.560791728487)*x[1]**o+(-0.619937635414)*x[1]
        arg[2,2]=(-0.738023386417)*x[0]**o+(-0.35311314517)*x[0]+(0.0350349516849)*x[1]**o+(-0.256133328724)*x[1]
        arg[2,3]=(0.836816159205)*x[0]**o+(-0.15869161496)*x[0]+(-0.115472743499)*x[1]**o+(-0.664972307042)*x[1]
        arg[2,4]=(0.257379090396)*x[0]**o+(0.970107472837)*x[0]+(0.323300957751)*x[1]**o+(0.028118478034)*x[1]
        arg[3,0]=(0.198639991638)*x[0]**o+(-0.386442010103)*x[0]+(0.566123916917)*x[1]**o+(0.442943885392)*x[1]
        arg[3,1]=(-0.108796148814)*x[0]**o+(0.896004033429)*x[0]+(-0.860364745295)*x[1]**o+(-0.362632925024)*x[1]
        arg[3,2]=(0.247204027814)*x[0]**o+(0.469777186671)*x[0]+(0.545991233155)*x[1]**o+(-0.320776951704)*x[1]
        arg[3,3]=(-0.98116086999)*x[0]**o+(-0.860611730622)*x[0]+(-0.333595175081)*x[1]**o+(0.437722032729)*x[1]
        arg[3,4]=(0.678428238286)*x[0]**o+(-0.827472935935)*x[0]+(0.355143303916)*x[1]**o+(-0.613985799175)*x[1]
        ref[0,0]=(0.0734861043324)*x_ref[0]**o+(0.698132844974)*x_ref[0]+(-0.0711338889844)*x_ref[1]**o+(0.667313854486)*x_ref[1]
        ref[0,1]=(0.846132724611)*x_ref[0]**o+(-0.215521619475)*x_ref[0]+(-0.920637540655)*x_ref[1]**o+(0.666739063184)*x_ref[1]
        ref[0,2]=(-0.963337851184)*x_ref[0]**o+(-0.961935972833)*x_ref[0]+(-0.0689798144115)*x_ref[1]**o+(0.0106557770119)*x_ref[1]
        ref[0,3]=(0.875063015269)*x_ref[0]**o+(-0.735727431445)*x_ref[0]+(-0.599251007104)*x_ref[1]**o+(-0.274659381946)*x_ref[1]
        ref[0,4]=(-0.378214565543)*x_ref[0]**o+(-0.476230056857)*x_ref[0]+(0.617722084788)*x_ref[1]**o+(-0.808732634287)*x_ref[1]
        ref[1,0]=(-0.324652264865)*x_ref[0]**o+(-0.314957074449)*x_ref[0]+(0.467828791079)*x_ref[1]**o+(-0.105499299461)*x_ref[1]
        ref[1,1]=(-0.0815291997441)*x_ref[0]**o+(0.1473072227)*x_ref[0]+(-0.866265670368)*x_ref[1]**o+(-0.66256092454)*x_ref[1]
        ref[1,2]=(-0.83962092333)*x_ref[0]**o+(0.681365396039)*x_ref[0]+(-0.607267339395)*x_ref[1]**o+(0.278637239291)*x_ref[1]
        ref[1,3]=(-0.300965517299)*x_ref[0]**o+(-0.0884159654351)*x_ref[0]+(-0.547465656655)*x_ref[1]**o+(0.708740949324)*x_ref[1]
        ref[1,4]=(-0.0250728833518)*x_ref[0]**o+(0.306828709412)*x_ref[0]+(-0.972269802286)*x_ref[1]**o+(0.619649412235)*x_ref[1]
        ref[2,0]=(-0.873228641325)*x_ref[0]**o+(-0.184974115411)*x_ref[0]+(-0.500142861985)*x_ref[1]**o+(0.708256240395)*x_ref[1]
        ref[2,1]=(-0.588540228523)*x_ref[0]**o+(-0.747378078994)*x_ref[0]+(0.560791728487)*x_ref[1]**o+(-0.619937635414)*x_ref[1]
        ref[2,2]=(-0.738023386417)*x_ref[0]**o+(-0.35311314517)*x_ref[0]+(0.0350349516849)*x_ref[1]**o+(-0.256133328724)*x_ref[1]
        ref[2,3]=(0.836816159205)*x_ref[0]**o+(-0.15869161496)*x_ref[0]+(-0.115472743499)*x_ref[1]**o+(-0.664972307042)*x_ref[1]
        ref[2,4]=(0.257379090396)*x_ref[0]**o+(0.970107472837)*x_ref[0]+(0.323300957751)*x_ref[1]**o+(0.028118478034)*x_ref[1]
        ref[3,0]=(0.198639991638)*x_ref[0]**o+(-0.386442010103)*x_ref[0]+(0.566123916917)*x_ref[1]**o+(0.442943885392)*x_ref[1]
        ref[3,1]=(-0.108796148814)*x_ref[0]**o+(0.896004033429)*x_ref[0]+(-0.860364745295)*x_ref[1]**o+(-0.362632925024)*x_ref[1]
        ref[3,2]=(0.247204027814)*x_ref[0]**o+(0.469777186671)*x_ref[0]+(0.545991233155)*x_ref[1]**o+(-0.320776951704)*x_ref[1]
        ref[3,3]=(-0.98116086999)*x_ref[0]**o+(-0.860611730622)*x_ref[0]+(-0.333595175081)*x_ref[1]**o+(0.437722032729)*x_ref[1]
        ref[3,4]=(0.678428238286)*x_ref[0]**o+(-0.827472935935)*x_ref[0]+(0.355143303916)*x_ref[1]**o+(-0.613985799175)*x_ref[1]
      else:
        arg[0,0]=(-0.34607777608)*x[0]**o+(0.932253964366)*x[0]+(0.698472742413)*x[1]**o+(-0.661983421645)*x[1]+(-0.965529107787)*x[2]**o+(-0.478528818669)*x[2]
        arg[0,1]=(0.664589900347)*x[0]**o+(-0.0719631882089)*x[0]+(-0.273316024389)*x[1]**o+(-0.887470310768)*x[1]+(-0.362253378888)*x[2]**o+(-0.604337291362)*x[2]
        arg[0,2]=(-0.0505574996951)*x[0]**o+(0.249062378634)*x[0]+(-0.760304066926)*x[1]**o+(-0.314569115107)*x[1]+(-0.679403446488)*x[2]**o+(-0.788567095894)*x[2]
        arg[0,3]=(0.737488025877)*x[0]**o+(0.489265711974)*x[0]+(0.192836157119)*x[1]**o+(-0.599394457037)*x[1]+(-0.972415976641)*x[2]**o+(0.620281802505)*x[2]
        arg[0,4]=(0.830347930734)*x[0]**o+(0.65690057858)*x[0]+(0.756058142792)*x[1]**o+(-0.886073273277)*x[1]+(-0.817193788749)*x[2]**o+(-0.906989660372)*x[2]
        arg[1,0]=(-0.961741874676)*x[0]**o+(0.231117463615)*x[0]+(-0.911422849525)*x[1]**o+(0.616612023918)*x[1]+(-0.563171272809)*x[2]**o+(0.00995177904006)*x[2]
        arg[1,1]=(-0.431814251982)*x[0]**o+(-0.244589543993)*x[0]+(-0.169319798745)*x[1]**o+(0.818064351799)*x[1]+(-0.899145214151)*x[2]**o+(-0.164644770082)*x[2]
        arg[1,2]=(0.149443250669)*x[0]**o+(-0.574248417945)*x[0]+(0.660724984377)*x[1]**o+(0.180990753976)*x[1]+(-0.36059400132)*x[2]**o+(-0.864734906872)*x[2]
        arg[1,3]=(-0.450956708576)*x[0]**o+(0.80934912947)*x[0]+(-0.921267395954)*x[1]**o+(0.449214650071)*x[1]+(-0.833237094175)*x[2]**o+(0.579827382459)*x[2]
        arg[1,4]=(-0.078106995473)*x[0]**o+(0.379085563974)*x[0]+(0.227204387882)*x[1]**o+(0.968199115202)*x[1]+(-0.272353745148)*x[2]**o+(0.380201872298)*x[2]
        arg[2,0]=(-0.852858598591)*x[0]**o+(0.720596990849)*x[0]+(0.281332272808)*x[1]**o+(-0.692404010367)*x[1]+(0.00869602393889)*x[2]**o+(0.0870674708905)*x[2]
        arg[2,1]=(0.466694162196)*x[0]**o+(-0.985083301887)*x[0]+(-0.208640915086)*x[1]**o+(-0.515440367775)*x[1]+(-0.76882243815)*x[2]**o+(0.345968814999)*x[2]
        arg[2,2]=(-0.0459664460618)*x[0]**o+(0.0919391841846)*x[0]+(0.687996833826)*x[1]**o+(0.94270298674)*x[1]+(0.0757885229342)*x[2]**o+(0.60760259541)*x[2]
        arg[2,3]=(0.968401710017)*x[0]**o+(0.0562744847066)*x[0]+(-0.345586665712)*x[1]**o+(-0.190986464506)*x[1]+(-0.202797239715)*x[2]**o+(-0.817657264098)*x[2]
        arg[2,4]=(-0.0288083775852)*x[0]**o+(0.164133034209)*x[0]+(-0.597572578553)*x[1]**o+(-0.669747923644)*x[1]+(-0.77713249687)*x[2]**o+(-0.0955335853718)*x[2]
        arg[3,0]=(-0.173723012835)*x[0]**o+(-0.306216351741)*x[0]+(-0.253810127374)*x[1]**o+(0.145737581929)*x[1]+(-0.950430852252)*x[2]**o+(-0.425608679334)*x[2]
        arg[3,1]=(0.0107577437502)*x[0]**o+(0.0213302772169)*x[0]+(0.403278323078)*x[1]**o+(0.390227936687)*x[1]+(-0.398743418731)*x[2]**o+(0.39090803029)*x[2]
        arg[3,2]=(0.151330768449)*x[0]**o+(0.290347114058)*x[0]+(-0.709672288966)*x[1]**o+(0.379791524303)*x[1]+(-0.403890944944)*x[2]**o+(0.327620813352)*x[2]
        arg[3,3]=(0.377649942493)*x[0]**o+(-0.140974638681)*x[0]+(-0.150965798704)*x[1]**o+(0.203654011715)*x[1]+(-0.70422609343)*x[2]**o+(-0.238707376032)*x[2]
        arg[3,4]=(0.454156920234)*x[0]**o+(-0.0803177888347)*x[0]+(0.325504272926)*x[1]**o+(0.285029937842)*x[1]+(-0.0684685710649)*x[2]**o+(0.471348026862)*x[2]
        ref[0,0]=(-0.34607777608)*x_ref[0]**o+(0.932253964366)*x_ref[0]+(0.698472742413)*x_ref[1]**o+(-0.661983421645)*x_ref[1]+(-0.965529107787)*x_ref[2]**o+(-0.478528818669)*x_ref[2]
        ref[0,1]=(0.664589900347)*x_ref[0]**o+(-0.0719631882089)*x_ref[0]+(-0.273316024389)*x_ref[1]**o+(-0.887470310768)*x_ref[1]+(-0.362253378888)*x_ref[2]**o+(-0.604337291362)*x_ref[2]
        ref[0,2]=(-0.0505574996951)*x_ref[0]**o+(0.249062378634)*x_ref[0]+(-0.760304066926)*x_ref[1]**o+(-0.314569115107)*x_ref[1]+(-0.679403446488)*x_ref[2]**o+(-0.788567095894)*x_ref[2]
        ref[0,3]=(0.737488025877)*x_ref[0]**o+(0.489265711974)*x_ref[0]+(0.192836157119)*x_ref[1]**o+(-0.599394457037)*x_ref[1]+(-0.972415976641)*x_ref[2]**o+(0.620281802505)*x_ref[2]
        ref[0,4]=(0.830347930734)*x_ref[0]**o+(0.65690057858)*x_ref[0]+(0.756058142792)*x_ref[1]**o+(-0.886073273277)*x_ref[1]+(-0.817193788749)*x_ref[2]**o+(-0.906989660372)*x_ref[2]
        ref[1,0]=(-0.961741874676)*x_ref[0]**o+(0.231117463615)*x_ref[0]+(-0.911422849525)*x_ref[1]**o+(0.616612023918)*x_ref[1]+(-0.563171272809)*x_ref[2]**o+(0.00995177904006)*x_ref[2]
        ref[1,1]=(-0.431814251982)*x_ref[0]**o+(-0.244589543993)*x_ref[0]+(-0.169319798745)*x_ref[1]**o+(0.818064351799)*x_ref[1]+(-0.899145214151)*x_ref[2]**o+(-0.164644770082)*x_ref[2]
        ref[1,2]=(0.149443250669)*x_ref[0]**o+(-0.574248417945)*x_ref[0]+(0.660724984377)*x_ref[1]**o+(0.180990753976)*x_ref[1]+(-0.36059400132)*x_ref[2]**o+(-0.864734906872)*x_ref[2]
        ref[1,3]=(-0.450956708576)*x_ref[0]**o+(0.80934912947)*x_ref[0]+(-0.921267395954)*x_ref[1]**o+(0.449214650071)*x_ref[1]+(-0.833237094175)*x_ref[2]**o+(0.579827382459)*x_ref[2]
        ref[1,4]=(-0.078106995473)*x_ref[0]**o+(0.379085563974)*x_ref[0]+(0.227204387882)*x_ref[1]**o+(0.968199115202)*x_ref[1]+(-0.272353745148)*x_ref[2]**o+(0.380201872298)*x_ref[2]
        ref[2,0]=(-0.852858598591)*x_ref[0]**o+(0.720596990849)*x_ref[0]+(0.281332272808)*x_ref[1]**o+(-0.692404010367)*x_ref[1]+(0.00869602393889)*x_ref[2]**o+(0.0870674708905)*x_ref[2]
        ref[2,1]=(0.466694162196)*x_ref[0]**o+(-0.985083301887)*x_ref[0]+(-0.208640915086)*x_ref[1]**o+(-0.515440367775)*x_ref[1]+(-0.76882243815)*x_ref[2]**o+(0.345968814999)*x_ref[2]
        ref[2,2]=(-0.0459664460618)*x_ref[0]**o+(0.0919391841846)*x_ref[0]+(0.687996833826)*x_ref[1]**o+(0.94270298674)*x_ref[1]+(0.0757885229342)*x_ref[2]**o+(0.60760259541)*x_ref[2]
        ref[2,3]=(0.968401710017)*x_ref[0]**o+(0.0562744847066)*x_ref[0]+(-0.345586665712)*x_ref[1]**o+(-0.190986464506)*x_ref[1]+(-0.202797239715)*x_ref[2]**o+(-0.817657264098)*x_ref[2]
        ref[2,4]=(-0.0288083775852)*x_ref[0]**o+(0.164133034209)*x_ref[0]+(-0.597572578553)*x_ref[1]**o+(-0.669747923644)*x_ref[1]+(-0.77713249687)*x_ref[2]**o+(-0.0955335853718)*x_ref[2]
        ref[3,0]=(-0.173723012835)*x_ref[0]**o+(-0.306216351741)*x_ref[0]+(-0.253810127374)*x_ref[1]**o+(0.145737581929)*x_ref[1]+(-0.950430852252)*x_ref[2]**o+(-0.425608679334)*x_ref[2]
        ref[3,1]=(0.0107577437502)*x_ref[0]**o+(0.0213302772169)*x_ref[0]+(0.403278323078)*x_ref[1]**o+(0.390227936687)*x_ref[1]+(-0.398743418731)*x_ref[2]**o+(0.39090803029)*x_ref[2]
        ref[3,2]=(0.151330768449)*x_ref[0]**o+(0.290347114058)*x_ref[0]+(-0.709672288966)*x_ref[1]**o+(0.379791524303)*x_ref[1]+(-0.403890944944)*x_ref[2]**o+(0.327620813352)*x_ref[2]
        ref[3,3]=(0.377649942493)*x_ref[0]**o+(-0.140974638681)*x_ref[0]+(-0.150965798704)*x_ref[1]**o+(0.203654011715)*x_ref[1]+(-0.70422609343)*x_ref[2]**o+(-0.238707376032)*x_ref[2]
        ref[3,4]=(0.454156920234)*x_ref[0]**o+(-0.0803177888347)*x_ref[0]+(0.325504272926)*x_ref[1]**o+(0.285029937842)*x_ref[1]+(-0.0684685710649)*x_ref[2]**o+(0.471348026862)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.754456420402)*x[0]**o+(0.379017389471)*x[0]+(-0.131681017079)*x[1]**o+(0.0844844232338)*x[1]
        arg[0,0,1]=(0.358583374641)*x[0]**o+(-0.237291325422)*x[0]+(-0.636434065455)*x[1]**o+(0.797375166528)*x[1]
        arg[0,1,0]=(0.715400241041)*x[0]**o+(-0.953885532065)*x[0]+(0.505634627788)*x[1]**o+(-0.0943694996143)*x[1]
        arg[0,1,1]=(-0.748388212428)*x[0]**o+(0.645715603392)*x[0]+(0.225587797337)*x[1]**o+(-0.274312832296)*x[1]
        arg[1,0,0]=(-0.696048542115)*x[0]**o+(0.728824156388)*x[0]+(0.475125287255)*x[1]**o+(0.231272864783)*x[1]
        arg[1,0,1]=(-0.171961266224)*x[0]**o+(0.751228523936)*x[0]+(0.0204128165955)*x[1]**o+(0.391751936939)*x[1]
        arg[1,1,0]=(0.702487008399)*x[0]**o+(0.408391530323)*x[0]+(-0.523719800121)*x[1]**o+(-0.142757224099)*x[1]
        arg[1,1,1]=(0.356477993939)*x[0]**o+(-0.552570052325)*x[0]+(-0.728412167408)*x[1]**o+(-0.138741983568)*x[1]
        arg[2,0,0]=(-0.972746877268)*x[0]**o+(-0.563197198314)*x[0]+(0.660003706244)*x[1]**o+(0.0312816821703)*x[1]
        arg[2,0,1]=(0.181440001149)*x[0]**o+(-0.551943857642)*x[0]+(0.550354945981)*x[1]**o+(0.222234118435)*x[1]
        arg[2,1,0]=(-0.843578289392)*x[0]**o+(-0.926317212728)*x[0]+(0.473372460648)*x[1]**o+(0.271910634427)*x[1]
        arg[2,1,1]=(-0.388589229676)*x[0]**o+(-0.117517308339)*x[0]+(0.0211660415594)*x[1]**o+(0.620489966427)*x[1]
        arg[3,0,0]=(-0.259455891555)*x[0]**o+(0.602520729775)*x[0]+(0.755086640963)*x[1]**o+(0.420352546305)*x[1]
        arg[3,0,1]=(0.901413641143)*x[0]**o+(-0.229939015002)*x[0]+(-0.0225626235937)*x[1]**o+(-0.319692416473)*x[1]
        arg[3,1,0]=(-0.398546041269)*x[0]**o+(-0.715449072888)*x[0]+(-0.079576973955)*x[1]**o+(0.132797262272)*x[1]
        arg[3,1,1]=(-0.0790389300192)*x[0]**o+(0.681411537598)*x[0]+(0.311941366364)*x[1]**o+(-0.350457682161)*x[1]
        arg[4,0,0]=(0.046674838542)*x[0]**o+(-0.765876197237)*x[0]+(-0.791075696221)*x[1]**o+(0.0338652019786)*x[1]
        arg[4,0,1]=(-0.00225965167336)*x[0]**o+(-0.773773082204)*x[0]+(0.696918916543)*x[1]**o+(-0.215060949558)*x[1]
        arg[4,1,0]=(-0.38907085938)*x[0]**o+(-0.907941787301)*x[0]+(-0.447045949233)*x[1]**o+(0.780078047474)*x[1]
        arg[4,1,1]=(0.198269028442)*x[0]**o+(0.991680341644)*x[0]+(-0.117854134367)*x[1]**o+(0.818191822875)*x[1]
        arg[5,0,0]=(-0.205861859192)*x[0]**o+(-0.230700784928)*x[0]+(-0.91205826553)*x[1]**o+(-0.563882902438)*x[1]
        arg[5,0,1]=(-0.667447962115)*x[0]**o+(0.0539650266455)*x[0]+(0.917333817791)*x[1]**o+(0.0571269304606)*x[1]
        arg[5,1,0]=(0.69461644672)*x[0]**o+(-0.901397746901)*x[0]+(0.388360725574)*x[1]**o+(-0.833240932275)*x[1]
        arg[5,1,1]=(0.236741248684)*x[0]**o+(0.637532818313)*x[0]+(-0.774742465882)*x[1]**o+(0.895299029069)*x[1]
        ref[0,0,0]=(0.754456420402)*x_ref[0]**o+(0.379017389471)*x_ref[0]+(-0.131681017079)*x_ref[1]**o+(0.0844844232338)*x_ref[1]
        ref[0,0,1]=(0.358583374641)*x_ref[0]**o+(-0.237291325422)*x_ref[0]+(-0.636434065455)*x_ref[1]**o+(0.797375166528)*x_ref[1]
        ref[0,1,0]=(0.715400241041)*x_ref[0]**o+(-0.953885532065)*x_ref[0]+(0.505634627788)*x_ref[1]**o+(-0.0943694996143)*x_ref[1]
        ref[0,1,1]=(-0.748388212428)*x_ref[0]**o+(0.645715603392)*x_ref[0]+(0.225587797337)*x_ref[1]**o+(-0.274312832296)*x_ref[1]
        ref[1,0,0]=(-0.696048542115)*x_ref[0]**o+(0.728824156388)*x_ref[0]+(0.475125287255)*x_ref[1]**o+(0.231272864783)*x_ref[1]
        ref[1,0,1]=(-0.171961266224)*x_ref[0]**o+(0.751228523936)*x_ref[0]+(0.0204128165955)*x_ref[1]**o+(0.391751936939)*x_ref[1]
        ref[1,1,0]=(0.702487008399)*x_ref[0]**o+(0.408391530323)*x_ref[0]+(-0.523719800121)*x_ref[1]**o+(-0.142757224099)*x_ref[1]
        ref[1,1,1]=(0.356477993939)*x_ref[0]**o+(-0.552570052325)*x_ref[0]+(-0.728412167408)*x_ref[1]**o+(-0.138741983568)*x_ref[1]
        ref[2,0,0]=(-0.972746877268)*x_ref[0]**o+(-0.563197198314)*x_ref[0]+(0.660003706244)*x_ref[1]**o+(0.0312816821703)*x_ref[1]
        ref[2,0,1]=(0.181440001149)*x_ref[0]**o+(-0.551943857642)*x_ref[0]+(0.550354945981)*x_ref[1]**o+(0.222234118435)*x_ref[1]
        ref[2,1,0]=(-0.843578289392)*x_ref[0]**o+(-0.926317212728)*x_ref[0]+(0.473372460648)*x_ref[1]**o+(0.271910634427)*x_ref[1]
        ref[2,1,1]=(-0.388589229676)*x_ref[0]**o+(-0.117517308339)*x_ref[0]+(0.0211660415594)*x_ref[1]**o+(0.620489966427)*x_ref[1]
        ref[3,0,0]=(-0.259455891555)*x_ref[0]**o+(0.602520729775)*x_ref[0]+(0.755086640963)*x_ref[1]**o+(0.420352546305)*x_ref[1]
        ref[3,0,1]=(0.901413641143)*x_ref[0]**o+(-0.229939015002)*x_ref[0]+(-0.0225626235937)*x_ref[1]**o+(-0.319692416473)*x_ref[1]
        ref[3,1,0]=(-0.398546041269)*x_ref[0]**o+(-0.715449072888)*x_ref[0]+(-0.079576973955)*x_ref[1]**o+(0.132797262272)*x_ref[1]
        ref[3,1,1]=(-0.0790389300192)*x_ref[0]**o+(0.681411537598)*x_ref[0]+(0.311941366364)*x_ref[1]**o+(-0.350457682161)*x_ref[1]
        ref[4,0,0]=(0.046674838542)*x_ref[0]**o+(-0.765876197237)*x_ref[0]+(-0.791075696221)*x_ref[1]**o+(0.0338652019786)*x_ref[1]
        ref[4,0,1]=(-0.00225965167336)*x_ref[0]**o+(-0.773773082204)*x_ref[0]+(0.696918916543)*x_ref[1]**o+(-0.215060949558)*x_ref[1]
        ref[4,1,0]=(-0.38907085938)*x_ref[0]**o+(-0.907941787301)*x_ref[0]+(-0.447045949233)*x_ref[1]**o+(0.780078047474)*x_ref[1]
        ref[4,1,1]=(0.198269028442)*x_ref[0]**o+(0.991680341644)*x_ref[0]+(-0.117854134367)*x_ref[1]**o+(0.818191822875)*x_ref[1]
        ref[5,0,0]=(-0.205861859192)*x_ref[0]**o+(-0.230700784928)*x_ref[0]+(-0.91205826553)*x_ref[1]**o+(-0.563882902438)*x_ref[1]
        ref[5,0,1]=(-0.667447962115)*x_ref[0]**o+(0.0539650266455)*x_ref[0]+(0.917333817791)*x_ref[1]**o+(0.0571269304606)*x_ref[1]
        ref[5,1,0]=(0.69461644672)*x_ref[0]**o+(-0.901397746901)*x_ref[0]+(0.388360725574)*x_ref[1]**o+(-0.833240932275)*x_ref[1]
        ref[5,1,1]=(0.236741248684)*x_ref[0]**o+(0.637532818313)*x_ref[0]+(-0.774742465882)*x_ref[1]**o+(0.895299029069)*x_ref[1]
      else:
        arg[0,0,0]=(-0.349847065801)*x[0]**o+(-0.373197072565)*x[0]+(-0.744002096082)*x[1]**o+(-0.32885837005)*x[1]+(-0.123733527788)*x[2]**o+(0.0357511439931)*x[2]
        arg[0,0,1]=(0.92390791115)*x[0]**o+(0.359314162684)*x[0]+(-0.344639305985)*x[1]**o+(0.87224607823)*x[1]+(0.603450217818)*x[2]**o+(0.767736049618)*x[2]
        arg[0,1,0]=(0.619626616187)*x[0]**o+(0.0777214591672)*x[0]+(-0.454938621339)*x[1]**o+(0.397372303072)*x[1]+(0.64005891461)*x[2]**o+(0.384837098241)*x[2]
        arg[0,1,1]=(0.694681223586)*x[0]**o+(0.746056110128)*x[0]+(-0.299128956618)*x[1]**o+(-0.411577781871)*x[1]+(0.0769629748738)*x[2]**o+(0.185345614784)*x[2]
        arg[1,0,0]=(-0.686338455095)*x[0]**o+(0.516443141718)*x[0]+(0.0963836215517)*x[1]**o+(-0.142756276631)*x[1]+(0.783984002173)*x[2]**o+(0.453983016774)*x[2]
        arg[1,0,1]=(-0.261515294422)*x[0]**o+(0.193353051852)*x[0]+(0.536103258837)*x[1]**o+(0.739075865062)*x[1]+(0.0854181253451)*x[2]**o+(0.993960851485)*x[2]
        arg[1,1,0]=(-0.041143202263)*x[0]**o+(-0.0483966496614)*x[0]+(0.100238616242)*x[1]**o+(0.229736237164)*x[1]+(-0.427145221147)*x[2]**o+(-0.143755555255)*x[2]
        arg[1,1,1]=(-0.489208229053)*x[0]**o+(-0.339638394047)*x[0]+(-0.0197108849498)*x[1]**o+(-0.64396285961)*x[1]+(0.644185397988)*x[2]**o+(0.952646865856)*x[2]
        arg[2,0,0]=(0.524059731144)*x[0]**o+(0.184311280754)*x[0]+(0.119889599722)*x[1]**o+(0.953938302354)*x[1]+(0.0331959462044)*x[2]**o+(-0.695477785189)*x[2]
        arg[2,0,1]=(0.925889695911)*x[0]**o+(0.612691423308)*x[0]+(0.492933561499)*x[1]**o+(-0.320060666384)*x[1]+(-0.308466393813)*x[2]**o+(0.907998151313)*x[2]
        arg[2,1,0]=(-0.874135284498)*x[0]**o+(0.0151718942733)*x[0]+(0.861045338983)*x[1]**o+(-0.228228955072)*x[1]+(-0.350972290538)*x[2]**o+(0.961075321834)*x[2]
        arg[2,1,1]=(-0.889805119627)*x[0]**o+(0.561790945688)*x[0]+(-0.500273719691)*x[1]**o+(-0.633623450372)*x[1]+(-0.0153487772927)*x[2]**o+(-0.0798904596833)*x[2]
        arg[3,0,0]=(-0.873621421496)*x[0]**o+(0.513411619207)*x[0]+(-0.961906901835)*x[1]**o+(0.321006445811)*x[1]+(-0.592020038506)*x[2]**o+(0.555697967812)*x[2]
        arg[3,0,1]=(0.850390230672)*x[0]**o+(-0.391435099771)*x[0]+(0.700909972778)*x[1]**o+(0.794987613709)*x[1]+(-0.775188800108)*x[2]**o+(-0.132226119354)*x[2]
        arg[3,1,0]=(0.597260092409)*x[0]**o+(0.393550227561)*x[0]+(-0.554683881247)*x[1]**o+(-0.371724214354)*x[1]+(-0.53562488411)*x[2]**o+(-0.555005739246)*x[2]
        arg[3,1,1]=(-0.14964970997)*x[0]**o+(-0.524377521321)*x[0]+(-0.738788452632)*x[1]**o+(-0.832575942733)*x[1]+(-0.766690459508)*x[2]**o+(-0.788273000486)*x[2]
        arg[4,0,0]=(-0.772922450679)*x[0]**o+(-0.0329959778533)*x[0]+(-0.504822565972)*x[1]**o+(-0.0328377918948)*x[1]+(0.598087910099)*x[2]**o+(-0.598910302754)*x[2]
        arg[4,0,1]=(-0.516823078367)*x[0]**o+(-0.783263163828)*x[0]+(0.793230730885)*x[1]**o+(-0.760393659268)*x[1]+(-0.39824557053)*x[2]**o+(0.448151384682)*x[2]
        arg[4,1,0]=(0.95067110965)*x[0]**o+(0.107038151555)*x[0]+(0.986100034409)*x[1]**o+(-0.603396251285)*x[1]+(0.463709896646)*x[2]**o+(-0.0752001209014)*x[2]
        arg[4,1,1]=(0.704778063889)*x[0]**o+(-0.00723361955023)*x[0]+(-0.229254019288)*x[1]**o+(0.972066929322)*x[1]+(-0.792377966898)*x[2]**o+(0.00519477641991)*x[2]
        arg[5,0,0]=(-0.37253642897)*x[0]**o+(-0.283784448234)*x[0]+(0.361753608877)*x[1]**o+(-0.119118907688)*x[1]+(0.403505165985)*x[2]**o+(0.223093000071)*x[2]
        arg[5,0,1]=(-0.78071834268)*x[0]**o+(0.0321159500156)*x[0]+(0.829040628045)*x[1]**o+(0.864378422992)*x[1]+(0.343294472729)*x[2]**o+(-0.502881977207)*x[2]
        arg[5,1,0]=(0.445553133352)*x[0]**o+(0.167144149279)*x[0]+(-0.773477294809)*x[1]**o+(0.400540684182)*x[1]+(-0.806904730971)*x[2]**o+(-0.789278588588)*x[2]
        arg[5,1,1]=(0.290545952984)*x[0]**o+(0.692814030915)*x[0]+(0.176313232933)*x[1]**o+(-0.88461128488)*x[1]+(-0.17903700494)*x[2]**o+(0.448842621603)*x[2]
        ref[0,0,0]=(-0.349847065801)*x_ref[0]**o+(-0.373197072565)*x_ref[0]+(-0.744002096082)*x_ref[1]**o+(-0.32885837005)*x_ref[1]+(-0.123733527788)*x_ref[2]**o+(0.0357511439931)*x_ref[2]
        ref[0,0,1]=(0.92390791115)*x_ref[0]**o+(0.359314162684)*x_ref[0]+(-0.344639305985)*x_ref[1]**o+(0.87224607823)*x_ref[1]+(0.603450217818)*x_ref[2]**o+(0.767736049618)*x_ref[2]
        ref[0,1,0]=(0.619626616187)*x_ref[0]**o+(0.0777214591672)*x_ref[0]+(-0.454938621339)*x_ref[1]**o+(0.397372303072)*x_ref[1]+(0.64005891461)*x_ref[2]**o+(0.384837098241)*x_ref[2]
        ref[0,1,1]=(0.694681223586)*x_ref[0]**o+(0.746056110128)*x_ref[0]+(-0.299128956618)*x_ref[1]**o+(-0.411577781871)*x_ref[1]+(0.0769629748738)*x_ref[2]**o+(0.185345614784)*x_ref[2]
        ref[1,0,0]=(-0.686338455095)*x_ref[0]**o+(0.516443141718)*x_ref[0]+(0.0963836215517)*x_ref[1]**o+(-0.142756276631)*x_ref[1]+(0.783984002173)*x_ref[2]**o+(0.453983016774)*x_ref[2]
        ref[1,0,1]=(-0.261515294422)*x_ref[0]**o+(0.193353051852)*x_ref[0]+(0.536103258837)*x_ref[1]**o+(0.739075865062)*x_ref[1]+(0.0854181253451)*x_ref[2]**o+(0.993960851485)*x_ref[2]
        ref[1,1,0]=(-0.041143202263)*x_ref[0]**o+(-0.0483966496614)*x_ref[0]+(0.100238616242)*x_ref[1]**o+(0.229736237164)*x_ref[1]+(-0.427145221147)*x_ref[2]**o+(-0.143755555255)*x_ref[2]
        ref[1,1,1]=(-0.489208229053)*x_ref[0]**o+(-0.339638394047)*x_ref[0]+(-0.0197108849498)*x_ref[1]**o+(-0.64396285961)*x_ref[1]+(0.644185397988)*x_ref[2]**o+(0.952646865856)*x_ref[2]
        ref[2,0,0]=(0.524059731144)*x_ref[0]**o+(0.184311280754)*x_ref[0]+(0.119889599722)*x_ref[1]**o+(0.953938302354)*x_ref[1]+(0.0331959462044)*x_ref[2]**o+(-0.695477785189)*x_ref[2]
        ref[2,0,1]=(0.925889695911)*x_ref[0]**o+(0.612691423308)*x_ref[0]+(0.492933561499)*x_ref[1]**o+(-0.320060666384)*x_ref[1]+(-0.308466393813)*x_ref[2]**o+(0.907998151313)*x_ref[2]
        ref[2,1,0]=(-0.874135284498)*x_ref[0]**o+(0.0151718942733)*x_ref[0]+(0.861045338983)*x_ref[1]**o+(-0.228228955072)*x_ref[1]+(-0.350972290538)*x_ref[2]**o+(0.961075321834)*x_ref[2]
        ref[2,1,1]=(-0.889805119627)*x_ref[0]**o+(0.561790945688)*x_ref[0]+(-0.500273719691)*x_ref[1]**o+(-0.633623450372)*x_ref[1]+(-0.0153487772927)*x_ref[2]**o+(-0.0798904596833)*x_ref[2]
        ref[3,0,0]=(-0.873621421496)*x_ref[0]**o+(0.513411619207)*x_ref[0]+(-0.961906901835)*x_ref[1]**o+(0.321006445811)*x_ref[1]+(-0.592020038506)*x_ref[2]**o+(0.555697967812)*x_ref[2]
        ref[3,0,1]=(0.850390230672)*x_ref[0]**o+(-0.391435099771)*x_ref[0]+(0.700909972778)*x_ref[1]**o+(0.794987613709)*x_ref[1]+(-0.775188800108)*x_ref[2]**o+(-0.132226119354)*x_ref[2]
        ref[3,1,0]=(0.597260092409)*x_ref[0]**o+(0.393550227561)*x_ref[0]+(-0.554683881247)*x_ref[1]**o+(-0.371724214354)*x_ref[1]+(-0.53562488411)*x_ref[2]**o+(-0.555005739246)*x_ref[2]
        ref[3,1,1]=(-0.14964970997)*x_ref[0]**o+(-0.524377521321)*x_ref[0]+(-0.738788452632)*x_ref[1]**o+(-0.832575942733)*x_ref[1]+(-0.766690459508)*x_ref[2]**o+(-0.788273000486)*x_ref[2]
        ref[4,0,0]=(-0.772922450679)*x_ref[0]**o+(-0.0329959778533)*x_ref[0]+(-0.504822565972)*x_ref[1]**o+(-0.0328377918948)*x_ref[1]+(0.598087910099)*x_ref[2]**o+(-0.598910302754)*x_ref[2]
        ref[4,0,1]=(-0.516823078367)*x_ref[0]**o+(-0.783263163828)*x_ref[0]+(0.793230730885)*x_ref[1]**o+(-0.760393659268)*x_ref[1]+(-0.39824557053)*x_ref[2]**o+(0.448151384682)*x_ref[2]
        ref[4,1,0]=(0.95067110965)*x_ref[0]**o+(0.107038151555)*x_ref[0]+(0.986100034409)*x_ref[1]**o+(-0.603396251285)*x_ref[1]+(0.463709896646)*x_ref[2]**o+(-0.0752001209014)*x_ref[2]
        ref[4,1,1]=(0.704778063889)*x_ref[0]**o+(-0.00723361955023)*x_ref[0]+(-0.229254019288)*x_ref[1]**o+(0.972066929322)*x_ref[1]+(-0.792377966898)*x_ref[2]**o+(0.00519477641991)*x_ref[2]
        ref[5,0,0]=(-0.37253642897)*x_ref[0]**o+(-0.283784448234)*x_ref[0]+(0.361753608877)*x_ref[1]**o+(-0.119118907688)*x_ref[1]+(0.403505165985)*x_ref[2]**o+(0.223093000071)*x_ref[2]
        ref[5,0,1]=(-0.78071834268)*x_ref[0]**o+(0.0321159500156)*x_ref[0]+(0.829040628045)*x_ref[1]**o+(0.864378422992)*x_ref[1]+(0.343294472729)*x_ref[2]**o+(-0.502881977207)*x_ref[2]
        ref[5,1,0]=(0.445553133352)*x_ref[0]**o+(0.167144149279)*x_ref[0]+(-0.773477294809)*x_ref[1]**o+(0.400540684182)*x_ref[1]+(-0.806904730971)*x_ref[2]**o+(-0.789278588588)*x_ref[2]
        ref[5,1,1]=(0.290545952984)*x_ref[0]**o+(0.692814030915)*x_ref[0]+(0.176313232933)*x_ref[1]**o+(-0.88461128488)*x_ref[1]+(-0.17903700494)*x_ref[2]**o+(0.448842621603)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.355090869915)*x[0]**o+(0.62333594265)*x[0]+(0.127393746399)*x[1]**o+(0.938815550559)*x[1]
        arg[0,0,0,1]=(0.361108442937)*x[0]**o+(0.848759016468)*x[0]+(-0.100173964534)*x[1]**o+(-0.798463736192)*x[1]
        arg[0,0,1,0]=(0.85312225866)*x[0]**o+(0.862709543579)*x[0]+(0.0794424365586)*x[1]**o+(-0.798377510326)*x[1]
        arg[0,0,1,1]=(-0.370853792692)*x[0]**o+(-0.967610789809)*x[0]+(0.277035447492)*x[1]**o+(-0.880157165176)*x[1]
        arg[0,0,2,0]=(-0.717702490059)*x[0]**o+(0.701614319774)*x[0]+(0.870301462626)*x[1]**o+(0.97842528383)*x[1]
        arg[0,0,2,1]=(0.398938846238)*x[0]**o+(0.460600049445)*x[0]+(-0.719014936029)*x[1]**o+(0.909351103488)*x[1]
        arg[0,1,0,0]=(-0.607340194261)*x[0]**o+(-0.851772216829)*x[0]+(0.7593824952)*x[1]**o+(-0.621669238017)*x[1]
        arg[0,1,0,1]=(-0.769916079359)*x[0]**o+(0.628665644279)*x[0]+(0.198700163204)*x[1]**o+(-0.878872706777)*x[1]
        arg[0,1,1,0]=(-0.133970652198)*x[0]**o+(0.817475343341)*x[0]+(-0.553628012005)*x[1]**o+(-0.780184324119)*x[1]
        arg[0,1,1,1]=(-0.91240774502)*x[0]**o+(-0.438683032351)*x[0]+(-0.651970251415)*x[1]**o+(-0.985718981791)*x[1]
        arg[0,1,2,0]=(-0.0894695484803)*x[0]**o+(0.230969119002)*x[0]+(-0.181318345335)*x[1]**o+(-0.684304266075)*x[1]
        arg[0,1,2,1]=(0.116582610294)*x[0]**o+(0.508506980325)*x[0]+(-0.227569256068)*x[1]**o+(-0.743396388621)*x[1]
        arg[0,2,0,0]=(0.163058551784)*x[0]**o+(-0.819992657695)*x[0]+(-0.0400315416665)*x[1]**o+(-0.368673546217)*x[1]
        arg[0,2,0,1]=(-0.760195093607)*x[0]**o+(-0.186788923915)*x[0]+(0.0439683490286)*x[1]**o+(-0.329021652328)*x[1]
        arg[0,2,1,0]=(-0.740086957456)*x[0]**o+(0.67648478989)*x[0]+(0.233575048863)*x[1]**o+(-0.353142296901)*x[1]
        arg[0,2,1,1]=(0.0374034623006)*x[0]**o+(-0.912304735553)*x[0]+(-0.982504180489)*x[1]**o+(0.388144019319)*x[1]
        arg[0,2,2,0]=(-0.812061018965)*x[0]**o+(0.641603736452)*x[0]+(0.260352725676)*x[1]**o+(-0.361534248663)*x[1]
        arg[0,2,2,1]=(0.0783398648132)*x[0]**o+(0.276408362801)*x[0]+(-0.107410956411)*x[1]**o+(-0.0733118712953)*x[1]
        arg[0,3,0,0]=(-0.085393595974)*x[0]**o+(0.545214990205)*x[0]+(0.451103178758)*x[1]**o+(0.741889771181)*x[1]
        arg[0,3,0,1]=(-0.377253694991)*x[0]**o+(0.243160465116)*x[0]+(-0.330892504638)*x[1]**o+(0.0688453403018)*x[1]
        arg[0,3,1,0]=(0.569018176673)*x[0]**o+(-0.519384243562)*x[0]+(0.807469891684)*x[1]**o+(-0.511601914558)*x[1]
        arg[0,3,1,1]=(-0.624467771068)*x[0]**o+(0.722555599188)*x[0]+(0.165243550597)*x[1]**o+(0.124887773794)*x[1]
        arg[0,3,2,0]=(0.846969233067)*x[0]**o+(0.143910732534)*x[0]+(-0.933748958347)*x[1]**o+(-0.468985899594)*x[1]
        arg[0,3,2,1]=(-0.00559860682228)*x[0]**o+(0.6265077144)*x[0]+(-0.597502391816)*x[1]**o+(-0.0269950883749)*x[1]
        arg[0,4,0,0]=(-0.929830228823)*x[0]**o+(0.533173796675)*x[0]+(-0.513355534671)*x[1]**o+(0.739452877653)*x[1]
        arg[0,4,0,1]=(0.573157694447)*x[0]**o+(0.533950659465)*x[0]+(-0.470668156002)*x[1]**o+(-0.267083337629)*x[1]
        arg[0,4,1,0]=(-0.786758281623)*x[0]**o+(-0.559211099907)*x[0]+(-0.981412883007)*x[1]**o+(-0.760247332612)*x[1]
        arg[0,4,1,1]=(0.811476615714)*x[0]**o+(-0.483217733654)*x[0]+(-0.473107949588)*x[1]**o+(-0.388655922158)*x[1]
        arg[0,4,2,0]=(0.743903538516)*x[0]**o+(-0.0914243557971)*x[0]+(0.523767499227)*x[1]**o+(0.105519926963)*x[1]
        arg[0,4,2,1]=(-0.425629408466)*x[0]**o+(0.294957108568)*x[0]+(-0.248527445843)*x[1]**o+(0.343947931379)*x[1]
        arg[1,0,0,0]=(0.413841637504)*x[0]**o+(0.609974192806)*x[0]+(-0.0785287734637)*x[1]**o+(0.538510227782)*x[1]
        arg[1,0,0,1]=(-0.905065001974)*x[0]**o+(-0.440318385611)*x[0]+(-0.0108293746138)*x[1]**o+(0.132125994317)*x[1]
        arg[1,0,1,0]=(0.212448400234)*x[0]**o+(0.117680077385)*x[0]+(0.501566152733)*x[1]**o+(-0.0914837162066)*x[1]
        arg[1,0,1,1]=(-0.763152636104)*x[0]**o+(0.318595673875)*x[0]+(-0.166759749724)*x[1]**o+(0.742420099653)*x[1]
        arg[1,0,2,0]=(-0.711420698498)*x[0]**o+(-0.0777910925778)*x[0]+(-0.885593755559)*x[1]**o+(0.302761673106)*x[1]
        arg[1,0,2,1]=(0.980964252534)*x[0]**o+(-0.890328597184)*x[0]+(-0.230421339981)*x[1]**o+(-0.656523995481)*x[1]
        arg[1,1,0,0]=(0.376070672474)*x[0]**o+(0.940362151504)*x[0]+(-0.297285488825)*x[1]**o+(0.451414529855)*x[1]
        arg[1,1,0,1]=(0.867987607318)*x[0]**o+(0.0477620815134)*x[0]+(0.257575148838)*x[1]**o+(-0.54187477723)*x[1]
        arg[1,1,1,0]=(-0.0643073021305)*x[0]**o+(0.20805724465)*x[0]+(-0.16501072904)*x[1]**o+(-0.341858346811)*x[1]
        arg[1,1,1,1]=(0.317830146671)*x[0]**o+(-0.307344984878)*x[0]+(0.170352713722)*x[1]**o+(-0.219561938951)*x[1]
        arg[1,1,2,0]=(0.89815928322)*x[0]**o+(0.51678942721)*x[0]+(0.897095493092)*x[1]**o+(-0.732217042123)*x[1]
        arg[1,1,2,1]=(0.0216578477569)*x[0]**o+(-0.942894641381)*x[0]+(-0.475731377321)*x[1]**o+(-0.753173025139)*x[1]
        arg[1,2,0,0]=(-0.897860515192)*x[0]**o+(0.833870034804)*x[0]+(0.554274053224)*x[1]**o+(-0.785991934149)*x[1]
        arg[1,2,0,1]=(-0.0648893111877)*x[0]**o+(0.173757424631)*x[0]+(0.860427552735)*x[1]**o+(0.139897107451)*x[1]
        arg[1,2,1,0]=(-0.146915585006)*x[0]**o+(-0.290881625114)*x[0]+(0.98541624122)*x[1]**o+(0.868617044417)*x[1]
        arg[1,2,1,1]=(0.17817541898)*x[0]**o+(-0.0112388444881)*x[0]+(0.646731986079)*x[1]**o+(0.653380181018)*x[1]
        arg[1,2,2,0]=(0.208250797377)*x[0]**o+(0.362989066599)*x[0]+(-0.384302232322)*x[1]**o+(0.376460044532)*x[1]
        arg[1,2,2,1]=(0.0437567097332)*x[0]**o+(0.739696851008)*x[0]+(0.0308848949312)*x[1]**o+(-0.445528203988)*x[1]
        arg[1,3,0,0]=(-0.257780793962)*x[0]**o+(0.840582446797)*x[0]+(0.599894892775)*x[1]**o+(-0.744691829118)*x[1]
        arg[1,3,0,1]=(0.913814269186)*x[0]**o+(0.295261304054)*x[0]+(0.0934778065196)*x[1]**o+(0.567115895371)*x[1]
        arg[1,3,1,0]=(-0.399785878199)*x[0]**o+(0.0336841328611)*x[0]+(0.236101490273)*x[1]**o+(0.915159045152)*x[1]
        arg[1,3,1,1]=(-0.955438743542)*x[0]**o+(-0.457060494715)*x[0]+(0.582140402321)*x[1]**o+(-0.999341752789)*x[1]
        arg[1,3,2,0]=(0.0206746776356)*x[0]**o+(-0.307327438556)*x[0]+(-0.646238381067)*x[1]**o+(0.248313182827)*x[1]
        arg[1,3,2,1]=(-0.10971097551)*x[0]**o+(-0.825303025298)*x[0]+(0.876884308015)*x[1]**o+(0.53621927318)*x[1]
        arg[1,4,0,0]=(0.833490742222)*x[0]**o+(0.865604127412)*x[0]+(0.627418353815)*x[1]**o+(0.523133814878)*x[1]
        arg[1,4,0,1]=(0.57616288545)*x[0]**o+(0.438552559891)*x[0]+(0.0256681709722)*x[1]**o+(-0.738314027898)*x[1]
        arg[1,4,1,0]=(0.319863593844)*x[0]**o+(0.177342228601)*x[0]+(0.0865784775196)*x[1]**o+(0.642550298518)*x[1]
        arg[1,4,1,1]=(-0.0141646689097)*x[0]**o+(0.317108687762)*x[0]+(0.188163493751)*x[1]**o+(0.990958509748)*x[1]
        arg[1,4,2,0]=(0.199211769035)*x[0]**o+(-0.0280132568124)*x[0]+(0.144960116505)*x[1]**o+(0.226821898255)*x[1]
        arg[1,4,2,1]=(-0.892464680507)*x[0]**o+(-0.60166056141)*x[0]+(-0.492226295363)*x[1]**o+(0.189924737916)*x[1]
        arg[2,0,0,0]=(0.329199148463)*x[0]**o+(-0.345343145094)*x[0]+(-0.22723822677)*x[1]**o+(-0.214232512114)*x[1]
        arg[2,0,0,1]=(-0.195525660216)*x[0]**o+(0.323518517888)*x[0]+(-0.116808991213)*x[1]**o+(0.22923091239)*x[1]
        arg[2,0,1,0]=(-0.826978585914)*x[0]**o+(0.641345887454)*x[0]+(-0.335583694811)*x[1]**o+(0.168685506673)*x[1]
        arg[2,0,1,1]=(-0.756460980453)*x[0]**o+(-0.9181609332)*x[0]+(0.0735111723828)*x[1]**o+(0.312524164436)*x[1]
        arg[2,0,2,0]=(0.297314747638)*x[0]**o+(0.522844740561)*x[0]+(0.890404335261)*x[1]**o+(-0.158861326888)*x[1]
        arg[2,0,2,1]=(-0.280311121771)*x[0]**o+(0.561223136791)*x[0]+(-0.952737235492)*x[1]**o+(0.581165512883)*x[1]
        arg[2,1,0,0]=(-0.202509823332)*x[0]**o+(0.298626146584)*x[0]+(0.533843705865)*x[1]**o+(0.652127174748)*x[1]
        arg[2,1,0,1]=(-0.349734740031)*x[0]**o+(-0.0589736478602)*x[0]+(0.741348559428)*x[1]**o+(0.993586735262)*x[1]
        arg[2,1,1,0]=(-0.350986159862)*x[0]**o+(-0.814566621733)*x[0]+(0.0556573702074)*x[1]**o+(0.363924401853)*x[1]
        arg[2,1,1,1]=(-0.562086543404)*x[0]**o+(0.126278013645)*x[0]+(-0.488684138283)*x[1]**o+(-0.990773306296)*x[1]
        arg[2,1,2,0]=(-0.873707667752)*x[0]**o+(-0.00610530100216)*x[0]+(-0.927847882023)*x[1]**o+(0.255850655097)*x[1]
        arg[2,1,2,1]=(0.683462220662)*x[0]**o+(-0.515973835798)*x[0]+(-0.676735030671)*x[1]**o+(0.704690967291)*x[1]
        arg[2,2,0,0]=(-0.0891987051347)*x[0]**o+(0.783509807114)*x[0]+(0.583865611337)*x[1]**o+(-0.373096806492)*x[1]
        arg[2,2,0,1]=(0.606443213957)*x[0]**o+(0.887055818748)*x[0]+(-0.281844491159)*x[1]**o+(0.420965776696)*x[1]
        arg[2,2,1,0]=(0.205051077925)*x[0]**o+(0.267690802839)*x[0]+(-0.821781634855)*x[1]**o+(-0.903592743179)*x[1]
        arg[2,2,1,1]=(-0.0438536252522)*x[0]**o+(-0.870635324748)*x[0]+(-0.238830174768)*x[1]**o+(-0.230782152037)*x[1]
        arg[2,2,2,0]=(-0.94437182489)*x[0]**o+(0.0666311523262)*x[0]+(0.481913247572)*x[1]**o+(-0.646415708448)*x[1]
        arg[2,2,2,1]=(0.443901323099)*x[0]**o+(0.349352493291)*x[0]+(-0.158789593075)*x[1]**o+(-0.603521202861)*x[1]
        arg[2,3,0,0]=(0.716829455981)*x[0]**o+(0.0186823101139)*x[0]+(-0.209053562063)*x[1]**o+(0.847906310296)*x[1]
        arg[2,3,0,1]=(0.305837642219)*x[0]**o+(0.138301301592)*x[0]+(0.0645236431399)*x[1]**o+(0.909641719567)*x[1]
        arg[2,3,1,0]=(-0.0177501904561)*x[0]**o+(0.177278703549)*x[0]+(-0.445915965222)*x[1]**o+(-0.894038892768)*x[1]
        arg[2,3,1,1]=(0.00187039478357)*x[0]**o+(-0.598690809424)*x[0]+(0.299307927538)*x[1]**o+(-0.882235917926)*x[1]
        arg[2,3,2,0]=(0.730056097248)*x[0]**o+(-0.307654523095)*x[0]+(0.246914489171)*x[1]**o+(-0.60035541725)*x[1]
        arg[2,3,2,1]=(-0.608344323894)*x[0]**o+(-0.081794637624)*x[0]+(0.099524567015)*x[1]**o+(-0.0654861171366)*x[1]
        arg[2,4,0,0]=(-0.225654564715)*x[0]**o+(-0.529959598356)*x[0]+(-0.890574962317)*x[1]**o+(0.994618025898)*x[1]
        arg[2,4,0,1]=(-0.231311865123)*x[0]**o+(0.479613722979)*x[0]+(-0.614977654518)*x[1]**o+(0.804794516306)*x[1]
        arg[2,4,1,0]=(0.365865384972)*x[0]**o+(-0.889202382492)*x[0]+(-0.300736557219)*x[1]**o+(0.838201123405)*x[1]
        arg[2,4,1,1]=(-0.809876332779)*x[0]**o+(0.1890391825)*x[0]+(-0.607869144071)*x[1]**o+(0.0448966585205)*x[1]
        arg[2,4,2,0]=(-0.83080984753)*x[0]**o+(0.305865817658)*x[0]+(-0.359639965091)*x[1]**o+(-0.288164510937)*x[1]
        arg[2,4,2,1]=(0.616022940224)*x[0]**o+(-0.636356653421)*x[0]+(0.693033289027)*x[1]**o+(-0.780662522342)*x[1]
        arg[3,0,0,0]=(0.985497135193)*x[0]**o+(-0.441259417159)*x[0]+(0.0550399586931)*x[1]**o+(-0.401551085867)*x[1]
        arg[3,0,0,1]=(0.66919872999)*x[0]**o+(-0.21372913425)*x[0]+(-0.904643614761)*x[1]**o+(0.254555219564)*x[1]
        arg[3,0,1,0]=(0.342630495616)*x[0]**o+(-0.688520932208)*x[0]+(-0.795149453533)*x[1]**o+(-0.532353008569)*x[1]
        arg[3,0,1,1]=(-0.157354065395)*x[0]**o+(-0.574302953359)*x[0]+(-0.0505170558232)*x[1]**o+(-0.916081229428)*x[1]
        arg[3,0,2,0]=(-0.276092076819)*x[0]**o+(0.755203166655)*x[0]+(0.71481609597)*x[1]**o+(-0.623929271337)*x[1]
        arg[3,0,2,1]=(0.901066640027)*x[0]**o+(0.615459930577)*x[0]+(-0.65213904597)*x[1]**o+(0.52496253783)*x[1]
        arg[3,1,0,0]=(-0.90847876416)*x[0]**o+(0.787519284245)*x[0]+(0.260252163627)*x[1]**o+(0.24849374459)*x[1]
        arg[3,1,0,1]=(-0.519950434741)*x[0]**o+(-0.146327164967)*x[0]+(0.199399508881)*x[1]**o+(0.583497630868)*x[1]
        arg[3,1,1,0]=(-0.956798390834)*x[0]**o+(0.145681684376)*x[0]+(0.308458821145)*x[1]**o+(-0.759589804755)*x[1]
        arg[3,1,1,1]=(-0.195569939954)*x[0]**o+(-0.544313649532)*x[0]+(0.387017502787)*x[1]**o+(0.174663629204)*x[1]
        arg[3,1,2,0]=(-0.3697024469)*x[0]**o+(0.356841678749)*x[0]+(0.922691790563)*x[1]**o+(-0.906755290703)*x[1]
        arg[3,1,2,1]=(-0.343599294558)*x[0]**o+(-0.213227630549)*x[0]+(-0.822516525353)*x[1]**o+(-0.0350701846979)*x[1]
        arg[3,2,0,0]=(-0.671120249859)*x[0]**o+(-0.409571854187)*x[0]+(-0.948171208395)*x[1]**o+(0.00128117529499)*x[1]
        arg[3,2,0,1]=(0.573086687235)*x[0]**o+(0.272624522478)*x[0]+(0.404610375833)*x[1]**o+(0.177742792198)*x[1]
        arg[3,2,1,0]=(-0.0481585168757)*x[0]**o+(0.108306965224)*x[0]+(0.962229835277)*x[1]**o+(0.457705229155)*x[1]
        arg[3,2,1,1]=(0.843866939471)*x[0]**o+(-0.209955643474)*x[0]+(0.861076954064)*x[1]**o+(-0.0920496448252)*x[1]
        arg[3,2,2,0]=(-0.482466555781)*x[0]**o+(0.905888895109)*x[0]+(0.969984046224)*x[1]**o+(0.779155706115)*x[1]
        arg[3,2,2,1]=(-0.119059605916)*x[0]**o+(-0.413165838565)*x[0]+(-0.458952085266)*x[1]**o+(0.667357851674)*x[1]
        arg[3,3,0,0]=(-0.139020214893)*x[0]**o+(-0.430716093157)*x[0]+(0.587151976237)*x[1]**o+(0.538779583614)*x[1]
        arg[3,3,0,1]=(-0.196698073225)*x[0]**o+(0.28357504894)*x[0]+(0.628834405238)*x[1]**o+(0.991210138395)*x[1]
        arg[3,3,1,0]=(-0.375644375731)*x[0]**o+(0.889016565145)*x[0]+(-0.495700828743)*x[1]**o+(-0.498859003938)*x[1]
        arg[3,3,1,1]=(0.783773071811)*x[0]**o+(-0.0875026023418)*x[0]+(-0.421213063239)*x[1]**o+(0.750220508147)*x[1]
        arg[3,3,2,0]=(0.211305046835)*x[0]**o+(-0.828436860631)*x[0]+(0.225275220049)*x[1]**o+(0.264723427485)*x[1]
        arg[3,3,2,1]=(0.582818095069)*x[0]**o+(0.497614150982)*x[0]+(-0.445084955033)*x[1]**o+(-0.600656630963)*x[1]
        arg[3,4,0,0]=(-0.886999530258)*x[0]**o+(0.386008089061)*x[0]+(-0.292729716725)*x[1]**o+(-0.159269788338)*x[1]
        arg[3,4,0,1]=(0.260932511148)*x[0]**o+(-0.533553803234)*x[0]+(0.174946866584)*x[1]**o+(-0.17701796222)*x[1]
        arg[3,4,1,0]=(0.37985010505)*x[0]**o+(0.934809225579)*x[0]+(-0.76271501884)*x[1]**o+(0.171072659)*x[1]
        arg[3,4,1,1]=(-0.369797459679)*x[0]**o+(-0.714539720717)*x[0]+(-0.122698805149)*x[1]**o+(0.874232715421)*x[1]
        arg[3,4,2,0]=(0.912569854856)*x[0]**o+(-0.773508960793)*x[0]+(-0.447748330632)*x[1]**o+(-0.0117517938995)*x[1]
        arg[3,4,2,1]=(-0.467442835658)*x[0]**o+(-0.792740521854)*x[0]+(-0.33441666815)*x[1]**o+(-0.821650769855)*x[1]
        ref[0,0,0,0]=(0.355090869915)*x_ref[0]**o+(0.62333594265)*x_ref[0]+(0.127393746399)*x_ref[1]**o+(0.938815550559)*x_ref[1]
        ref[0,0,0,1]=(0.361108442937)*x_ref[0]**o+(0.848759016468)*x_ref[0]+(-0.100173964534)*x_ref[1]**o+(-0.798463736192)*x_ref[1]
        ref[0,0,1,0]=(0.85312225866)*x_ref[0]**o+(0.862709543579)*x_ref[0]+(0.0794424365586)*x_ref[1]**o+(-0.798377510326)*x_ref[1]
        ref[0,0,1,1]=(-0.370853792692)*x_ref[0]**o+(-0.967610789809)*x_ref[0]+(0.277035447492)*x_ref[1]**o+(-0.880157165176)*x_ref[1]
        ref[0,0,2,0]=(-0.717702490059)*x_ref[0]**o+(0.701614319774)*x_ref[0]+(0.870301462626)*x_ref[1]**o+(0.97842528383)*x_ref[1]
        ref[0,0,2,1]=(0.398938846238)*x_ref[0]**o+(0.460600049445)*x_ref[0]+(-0.719014936029)*x_ref[1]**o+(0.909351103488)*x_ref[1]
        ref[0,1,0,0]=(-0.607340194261)*x_ref[0]**o+(-0.851772216829)*x_ref[0]+(0.7593824952)*x_ref[1]**o+(-0.621669238017)*x_ref[1]
        ref[0,1,0,1]=(-0.769916079359)*x_ref[0]**o+(0.628665644279)*x_ref[0]+(0.198700163204)*x_ref[1]**o+(-0.878872706777)*x_ref[1]
        ref[0,1,1,0]=(-0.133970652198)*x_ref[0]**o+(0.817475343341)*x_ref[0]+(-0.553628012005)*x_ref[1]**o+(-0.780184324119)*x_ref[1]
        ref[0,1,1,1]=(-0.91240774502)*x_ref[0]**o+(-0.438683032351)*x_ref[0]+(-0.651970251415)*x_ref[1]**o+(-0.985718981791)*x_ref[1]
        ref[0,1,2,0]=(-0.0894695484803)*x_ref[0]**o+(0.230969119002)*x_ref[0]+(-0.181318345335)*x_ref[1]**o+(-0.684304266075)*x_ref[1]
        ref[0,1,2,1]=(0.116582610294)*x_ref[0]**o+(0.508506980325)*x_ref[0]+(-0.227569256068)*x_ref[1]**o+(-0.743396388621)*x_ref[1]
        ref[0,2,0,0]=(0.163058551784)*x_ref[0]**o+(-0.819992657695)*x_ref[0]+(-0.0400315416665)*x_ref[1]**o+(-0.368673546217)*x_ref[1]
        ref[0,2,0,1]=(-0.760195093607)*x_ref[0]**o+(-0.186788923915)*x_ref[0]+(0.0439683490286)*x_ref[1]**o+(-0.329021652328)*x_ref[1]
        ref[0,2,1,0]=(-0.740086957456)*x_ref[0]**o+(0.67648478989)*x_ref[0]+(0.233575048863)*x_ref[1]**o+(-0.353142296901)*x_ref[1]
        ref[0,2,1,1]=(0.0374034623006)*x_ref[0]**o+(-0.912304735553)*x_ref[0]+(-0.982504180489)*x_ref[1]**o+(0.388144019319)*x_ref[1]
        ref[0,2,2,0]=(-0.812061018965)*x_ref[0]**o+(0.641603736452)*x_ref[0]+(0.260352725676)*x_ref[1]**o+(-0.361534248663)*x_ref[1]
        ref[0,2,2,1]=(0.0783398648132)*x_ref[0]**o+(0.276408362801)*x_ref[0]+(-0.107410956411)*x_ref[1]**o+(-0.0733118712953)*x_ref[1]
        ref[0,3,0,0]=(-0.085393595974)*x_ref[0]**o+(0.545214990205)*x_ref[0]+(0.451103178758)*x_ref[1]**o+(0.741889771181)*x_ref[1]
        ref[0,3,0,1]=(-0.377253694991)*x_ref[0]**o+(0.243160465116)*x_ref[0]+(-0.330892504638)*x_ref[1]**o+(0.0688453403018)*x_ref[1]
        ref[0,3,1,0]=(0.569018176673)*x_ref[0]**o+(-0.519384243562)*x_ref[0]+(0.807469891684)*x_ref[1]**o+(-0.511601914558)*x_ref[1]
        ref[0,3,1,1]=(-0.624467771068)*x_ref[0]**o+(0.722555599188)*x_ref[0]+(0.165243550597)*x_ref[1]**o+(0.124887773794)*x_ref[1]
        ref[0,3,2,0]=(0.846969233067)*x_ref[0]**o+(0.143910732534)*x_ref[0]+(-0.933748958347)*x_ref[1]**o+(-0.468985899594)*x_ref[1]
        ref[0,3,2,1]=(-0.00559860682228)*x_ref[0]**o+(0.6265077144)*x_ref[0]+(-0.597502391816)*x_ref[1]**o+(-0.0269950883749)*x_ref[1]
        ref[0,4,0,0]=(-0.929830228823)*x_ref[0]**o+(0.533173796675)*x_ref[0]+(-0.513355534671)*x_ref[1]**o+(0.739452877653)*x_ref[1]
        ref[0,4,0,1]=(0.573157694447)*x_ref[0]**o+(0.533950659465)*x_ref[0]+(-0.470668156002)*x_ref[1]**o+(-0.267083337629)*x_ref[1]
        ref[0,4,1,0]=(-0.786758281623)*x_ref[0]**o+(-0.559211099907)*x_ref[0]+(-0.981412883007)*x_ref[1]**o+(-0.760247332612)*x_ref[1]
        ref[0,4,1,1]=(0.811476615714)*x_ref[0]**o+(-0.483217733654)*x_ref[0]+(-0.473107949588)*x_ref[1]**o+(-0.388655922158)*x_ref[1]
        ref[0,4,2,0]=(0.743903538516)*x_ref[0]**o+(-0.0914243557971)*x_ref[0]+(0.523767499227)*x_ref[1]**o+(0.105519926963)*x_ref[1]
        ref[0,4,2,1]=(-0.425629408466)*x_ref[0]**o+(0.294957108568)*x_ref[0]+(-0.248527445843)*x_ref[1]**o+(0.343947931379)*x_ref[1]
        ref[1,0,0,0]=(0.413841637504)*x_ref[0]**o+(0.609974192806)*x_ref[0]+(-0.0785287734637)*x_ref[1]**o+(0.538510227782)*x_ref[1]
        ref[1,0,0,1]=(-0.905065001974)*x_ref[0]**o+(-0.440318385611)*x_ref[0]+(-0.0108293746138)*x_ref[1]**o+(0.132125994317)*x_ref[1]
        ref[1,0,1,0]=(0.212448400234)*x_ref[0]**o+(0.117680077385)*x_ref[0]+(0.501566152733)*x_ref[1]**o+(-0.0914837162066)*x_ref[1]
        ref[1,0,1,1]=(-0.763152636104)*x_ref[0]**o+(0.318595673875)*x_ref[0]+(-0.166759749724)*x_ref[1]**o+(0.742420099653)*x_ref[1]
        ref[1,0,2,0]=(-0.711420698498)*x_ref[0]**o+(-0.0777910925778)*x_ref[0]+(-0.885593755559)*x_ref[1]**o+(0.302761673106)*x_ref[1]
        ref[1,0,2,1]=(0.980964252534)*x_ref[0]**o+(-0.890328597184)*x_ref[0]+(-0.230421339981)*x_ref[1]**o+(-0.656523995481)*x_ref[1]
        ref[1,1,0,0]=(0.376070672474)*x_ref[0]**o+(0.940362151504)*x_ref[0]+(-0.297285488825)*x_ref[1]**o+(0.451414529855)*x_ref[1]
        ref[1,1,0,1]=(0.867987607318)*x_ref[0]**o+(0.0477620815134)*x_ref[0]+(0.257575148838)*x_ref[1]**o+(-0.54187477723)*x_ref[1]
        ref[1,1,1,0]=(-0.0643073021305)*x_ref[0]**o+(0.20805724465)*x_ref[0]+(-0.16501072904)*x_ref[1]**o+(-0.341858346811)*x_ref[1]
        ref[1,1,1,1]=(0.317830146671)*x_ref[0]**o+(-0.307344984878)*x_ref[0]+(0.170352713722)*x_ref[1]**o+(-0.219561938951)*x_ref[1]
        ref[1,1,2,0]=(0.89815928322)*x_ref[0]**o+(0.51678942721)*x_ref[0]+(0.897095493092)*x_ref[1]**o+(-0.732217042123)*x_ref[1]
        ref[1,1,2,1]=(0.0216578477569)*x_ref[0]**o+(-0.942894641381)*x_ref[0]+(-0.475731377321)*x_ref[1]**o+(-0.753173025139)*x_ref[1]
        ref[1,2,0,0]=(-0.897860515192)*x_ref[0]**o+(0.833870034804)*x_ref[0]+(0.554274053224)*x_ref[1]**o+(-0.785991934149)*x_ref[1]
        ref[1,2,0,1]=(-0.0648893111877)*x_ref[0]**o+(0.173757424631)*x_ref[0]+(0.860427552735)*x_ref[1]**o+(0.139897107451)*x_ref[1]
        ref[1,2,1,0]=(-0.146915585006)*x_ref[0]**o+(-0.290881625114)*x_ref[0]+(0.98541624122)*x_ref[1]**o+(0.868617044417)*x_ref[1]
        ref[1,2,1,1]=(0.17817541898)*x_ref[0]**o+(-0.0112388444881)*x_ref[0]+(0.646731986079)*x_ref[1]**o+(0.653380181018)*x_ref[1]
        ref[1,2,2,0]=(0.208250797377)*x_ref[0]**o+(0.362989066599)*x_ref[0]+(-0.384302232322)*x_ref[1]**o+(0.376460044532)*x_ref[1]
        ref[1,2,2,1]=(0.0437567097332)*x_ref[0]**o+(0.739696851008)*x_ref[0]+(0.0308848949312)*x_ref[1]**o+(-0.445528203988)*x_ref[1]
        ref[1,3,0,0]=(-0.257780793962)*x_ref[0]**o+(0.840582446797)*x_ref[0]+(0.599894892775)*x_ref[1]**o+(-0.744691829118)*x_ref[1]
        ref[1,3,0,1]=(0.913814269186)*x_ref[0]**o+(0.295261304054)*x_ref[0]+(0.0934778065196)*x_ref[1]**o+(0.567115895371)*x_ref[1]
        ref[1,3,1,0]=(-0.399785878199)*x_ref[0]**o+(0.0336841328611)*x_ref[0]+(0.236101490273)*x_ref[1]**o+(0.915159045152)*x_ref[1]
        ref[1,3,1,1]=(-0.955438743542)*x_ref[0]**o+(-0.457060494715)*x_ref[0]+(0.582140402321)*x_ref[1]**o+(-0.999341752789)*x_ref[1]
        ref[1,3,2,0]=(0.0206746776356)*x_ref[0]**o+(-0.307327438556)*x_ref[0]+(-0.646238381067)*x_ref[1]**o+(0.248313182827)*x_ref[1]
        ref[1,3,2,1]=(-0.10971097551)*x_ref[0]**o+(-0.825303025298)*x_ref[0]+(0.876884308015)*x_ref[1]**o+(0.53621927318)*x_ref[1]
        ref[1,4,0,0]=(0.833490742222)*x_ref[0]**o+(0.865604127412)*x_ref[0]+(0.627418353815)*x_ref[1]**o+(0.523133814878)*x_ref[1]
        ref[1,4,0,1]=(0.57616288545)*x_ref[0]**o+(0.438552559891)*x_ref[0]+(0.0256681709722)*x_ref[1]**o+(-0.738314027898)*x_ref[1]
        ref[1,4,1,0]=(0.319863593844)*x_ref[0]**o+(0.177342228601)*x_ref[0]+(0.0865784775196)*x_ref[1]**o+(0.642550298518)*x_ref[1]
        ref[1,4,1,1]=(-0.0141646689097)*x_ref[0]**o+(0.317108687762)*x_ref[0]+(0.188163493751)*x_ref[1]**o+(0.990958509748)*x_ref[1]
        ref[1,4,2,0]=(0.199211769035)*x_ref[0]**o+(-0.0280132568124)*x_ref[0]+(0.144960116505)*x_ref[1]**o+(0.226821898255)*x_ref[1]
        ref[1,4,2,1]=(-0.892464680507)*x_ref[0]**o+(-0.60166056141)*x_ref[0]+(-0.492226295363)*x_ref[1]**o+(0.189924737916)*x_ref[1]
        ref[2,0,0,0]=(0.329199148463)*x_ref[0]**o+(-0.345343145094)*x_ref[0]+(-0.22723822677)*x_ref[1]**o+(-0.214232512114)*x_ref[1]
        ref[2,0,0,1]=(-0.195525660216)*x_ref[0]**o+(0.323518517888)*x_ref[0]+(-0.116808991213)*x_ref[1]**o+(0.22923091239)*x_ref[1]
        ref[2,0,1,0]=(-0.826978585914)*x_ref[0]**o+(0.641345887454)*x_ref[0]+(-0.335583694811)*x_ref[1]**o+(0.168685506673)*x_ref[1]
        ref[2,0,1,1]=(-0.756460980453)*x_ref[0]**o+(-0.9181609332)*x_ref[0]+(0.0735111723828)*x_ref[1]**o+(0.312524164436)*x_ref[1]
        ref[2,0,2,0]=(0.297314747638)*x_ref[0]**o+(0.522844740561)*x_ref[0]+(0.890404335261)*x_ref[1]**o+(-0.158861326888)*x_ref[1]
        ref[2,0,2,1]=(-0.280311121771)*x_ref[0]**o+(0.561223136791)*x_ref[0]+(-0.952737235492)*x_ref[1]**o+(0.581165512883)*x_ref[1]
        ref[2,1,0,0]=(-0.202509823332)*x_ref[0]**o+(0.298626146584)*x_ref[0]+(0.533843705865)*x_ref[1]**o+(0.652127174748)*x_ref[1]
        ref[2,1,0,1]=(-0.349734740031)*x_ref[0]**o+(-0.0589736478602)*x_ref[0]+(0.741348559428)*x_ref[1]**o+(0.993586735262)*x_ref[1]
        ref[2,1,1,0]=(-0.350986159862)*x_ref[0]**o+(-0.814566621733)*x_ref[0]+(0.0556573702074)*x_ref[1]**o+(0.363924401853)*x_ref[1]
        ref[2,1,1,1]=(-0.562086543404)*x_ref[0]**o+(0.126278013645)*x_ref[0]+(-0.488684138283)*x_ref[1]**o+(-0.990773306296)*x_ref[1]
        ref[2,1,2,0]=(-0.873707667752)*x_ref[0]**o+(-0.00610530100216)*x_ref[0]+(-0.927847882023)*x_ref[1]**o+(0.255850655097)*x_ref[1]
        ref[2,1,2,1]=(0.683462220662)*x_ref[0]**o+(-0.515973835798)*x_ref[0]+(-0.676735030671)*x_ref[1]**o+(0.704690967291)*x_ref[1]
        ref[2,2,0,0]=(-0.0891987051347)*x_ref[0]**o+(0.783509807114)*x_ref[0]+(0.583865611337)*x_ref[1]**o+(-0.373096806492)*x_ref[1]
        ref[2,2,0,1]=(0.606443213957)*x_ref[0]**o+(0.887055818748)*x_ref[0]+(-0.281844491159)*x_ref[1]**o+(0.420965776696)*x_ref[1]
        ref[2,2,1,0]=(0.205051077925)*x_ref[0]**o+(0.267690802839)*x_ref[0]+(-0.821781634855)*x_ref[1]**o+(-0.903592743179)*x_ref[1]
        ref[2,2,1,1]=(-0.0438536252522)*x_ref[0]**o+(-0.870635324748)*x_ref[0]+(-0.238830174768)*x_ref[1]**o+(-0.230782152037)*x_ref[1]
        ref[2,2,2,0]=(-0.94437182489)*x_ref[0]**o+(0.0666311523262)*x_ref[0]+(0.481913247572)*x_ref[1]**o+(-0.646415708448)*x_ref[1]
        ref[2,2,2,1]=(0.443901323099)*x_ref[0]**o+(0.349352493291)*x_ref[0]+(-0.158789593075)*x_ref[1]**o+(-0.603521202861)*x_ref[1]
        ref[2,3,0,0]=(0.716829455981)*x_ref[0]**o+(0.0186823101139)*x_ref[0]+(-0.209053562063)*x_ref[1]**o+(0.847906310296)*x_ref[1]
        ref[2,3,0,1]=(0.305837642219)*x_ref[0]**o+(0.138301301592)*x_ref[0]+(0.0645236431399)*x_ref[1]**o+(0.909641719567)*x_ref[1]
        ref[2,3,1,0]=(-0.0177501904561)*x_ref[0]**o+(0.177278703549)*x_ref[0]+(-0.445915965222)*x_ref[1]**o+(-0.894038892768)*x_ref[1]
        ref[2,3,1,1]=(0.00187039478357)*x_ref[0]**o+(-0.598690809424)*x_ref[0]+(0.299307927538)*x_ref[1]**o+(-0.882235917926)*x_ref[1]
        ref[2,3,2,0]=(0.730056097248)*x_ref[0]**o+(-0.307654523095)*x_ref[0]+(0.246914489171)*x_ref[1]**o+(-0.60035541725)*x_ref[1]
        ref[2,3,2,1]=(-0.608344323894)*x_ref[0]**o+(-0.081794637624)*x_ref[0]+(0.099524567015)*x_ref[1]**o+(-0.0654861171366)*x_ref[1]
        ref[2,4,0,0]=(-0.225654564715)*x_ref[0]**o+(-0.529959598356)*x_ref[0]+(-0.890574962317)*x_ref[1]**o+(0.994618025898)*x_ref[1]
        ref[2,4,0,1]=(-0.231311865123)*x_ref[0]**o+(0.479613722979)*x_ref[0]+(-0.614977654518)*x_ref[1]**o+(0.804794516306)*x_ref[1]
        ref[2,4,1,0]=(0.365865384972)*x_ref[0]**o+(-0.889202382492)*x_ref[0]+(-0.300736557219)*x_ref[1]**o+(0.838201123405)*x_ref[1]
        ref[2,4,1,1]=(-0.809876332779)*x_ref[0]**o+(0.1890391825)*x_ref[0]+(-0.607869144071)*x_ref[1]**o+(0.0448966585205)*x_ref[1]
        ref[2,4,2,0]=(-0.83080984753)*x_ref[0]**o+(0.305865817658)*x_ref[0]+(-0.359639965091)*x_ref[1]**o+(-0.288164510937)*x_ref[1]
        ref[2,4,2,1]=(0.616022940224)*x_ref[0]**o+(-0.636356653421)*x_ref[0]+(0.693033289027)*x_ref[1]**o+(-0.780662522342)*x_ref[1]
        ref[3,0,0,0]=(0.985497135193)*x_ref[0]**o+(-0.441259417159)*x_ref[0]+(0.0550399586931)*x_ref[1]**o+(-0.401551085867)*x_ref[1]
        ref[3,0,0,1]=(0.66919872999)*x_ref[0]**o+(-0.21372913425)*x_ref[0]+(-0.904643614761)*x_ref[1]**o+(0.254555219564)*x_ref[1]
        ref[3,0,1,0]=(0.342630495616)*x_ref[0]**o+(-0.688520932208)*x_ref[0]+(-0.795149453533)*x_ref[1]**o+(-0.532353008569)*x_ref[1]
        ref[3,0,1,1]=(-0.157354065395)*x_ref[0]**o+(-0.574302953359)*x_ref[0]+(-0.0505170558232)*x_ref[1]**o+(-0.916081229428)*x_ref[1]
        ref[3,0,2,0]=(-0.276092076819)*x_ref[0]**o+(0.755203166655)*x_ref[0]+(0.71481609597)*x_ref[1]**o+(-0.623929271337)*x_ref[1]
        ref[3,0,2,1]=(0.901066640027)*x_ref[0]**o+(0.615459930577)*x_ref[0]+(-0.65213904597)*x_ref[1]**o+(0.52496253783)*x_ref[1]
        ref[3,1,0,0]=(-0.90847876416)*x_ref[0]**o+(0.787519284245)*x_ref[0]+(0.260252163627)*x_ref[1]**o+(0.24849374459)*x_ref[1]
        ref[3,1,0,1]=(-0.519950434741)*x_ref[0]**o+(-0.146327164967)*x_ref[0]+(0.199399508881)*x_ref[1]**o+(0.583497630868)*x_ref[1]
        ref[3,1,1,0]=(-0.956798390834)*x_ref[0]**o+(0.145681684376)*x_ref[0]+(0.308458821145)*x_ref[1]**o+(-0.759589804755)*x_ref[1]
        ref[3,1,1,1]=(-0.195569939954)*x_ref[0]**o+(-0.544313649532)*x_ref[0]+(0.387017502787)*x_ref[1]**o+(0.174663629204)*x_ref[1]
        ref[3,1,2,0]=(-0.3697024469)*x_ref[0]**o+(0.356841678749)*x_ref[0]+(0.922691790563)*x_ref[1]**o+(-0.906755290703)*x_ref[1]
        ref[3,1,2,1]=(-0.343599294558)*x_ref[0]**o+(-0.213227630549)*x_ref[0]+(-0.822516525353)*x_ref[1]**o+(-0.0350701846979)*x_ref[1]
        ref[3,2,0,0]=(-0.671120249859)*x_ref[0]**o+(-0.409571854187)*x_ref[0]+(-0.948171208395)*x_ref[1]**o+(0.00128117529499)*x_ref[1]
        ref[3,2,0,1]=(0.573086687235)*x_ref[0]**o+(0.272624522478)*x_ref[0]+(0.404610375833)*x_ref[1]**o+(0.177742792198)*x_ref[1]
        ref[3,2,1,0]=(-0.0481585168757)*x_ref[0]**o+(0.108306965224)*x_ref[0]+(0.962229835277)*x_ref[1]**o+(0.457705229155)*x_ref[1]
        ref[3,2,1,1]=(0.843866939471)*x_ref[0]**o+(-0.209955643474)*x_ref[0]+(0.861076954064)*x_ref[1]**o+(-0.0920496448252)*x_ref[1]
        ref[3,2,2,0]=(-0.482466555781)*x_ref[0]**o+(0.905888895109)*x_ref[0]+(0.969984046224)*x_ref[1]**o+(0.779155706115)*x_ref[1]
        ref[3,2,2,1]=(-0.119059605916)*x_ref[0]**o+(-0.413165838565)*x_ref[0]+(-0.458952085266)*x_ref[1]**o+(0.667357851674)*x_ref[1]
        ref[3,3,0,0]=(-0.139020214893)*x_ref[0]**o+(-0.430716093157)*x_ref[0]+(0.587151976237)*x_ref[1]**o+(0.538779583614)*x_ref[1]
        ref[3,3,0,1]=(-0.196698073225)*x_ref[0]**o+(0.28357504894)*x_ref[0]+(0.628834405238)*x_ref[1]**o+(0.991210138395)*x_ref[1]
        ref[3,3,1,0]=(-0.375644375731)*x_ref[0]**o+(0.889016565145)*x_ref[0]+(-0.495700828743)*x_ref[1]**o+(-0.498859003938)*x_ref[1]
        ref[3,3,1,1]=(0.783773071811)*x_ref[0]**o+(-0.0875026023418)*x_ref[0]+(-0.421213063239)*x_ref[1]**o+(0.750220508147)*x_ref[1]
        ref[3,3,2,0]=(0.211305046835)*x_ref[0]**o+(-0.828436860631)*x_ref[0]+(0.225275220049)*x_ref[1]**o+(0.264723427485)*x_ref[1]
        ref[3,3,2,1]=(0.582818095069)*x_ref[0]**o+(0.497614150982)*x_ref[0]+(-0.445084955033)*x_ref[1]**o+(-0.600656630963)*x_ref[1]
        ref[3,4,0,0]=(-0.886999530258)*x_ref[0]**o+(0.386008089061)*x_ref[0]+(-0.292729716725)*x_ref[1]**o+(-0.159269788338)*x_ref[1]
        ref[3,4,0,1]=(0.260932511148)*x_ref[0]**o+(-0.533553803234)*x_ref[0]+(0.174946866584)*x_ref[1]**o+(-0.17701796222)*x_ref[1]
        ref[3,4,1,0]=(0.37985010505)*x_ref[0]**o+(0.934809225579)*x_ref[0]+(-0.76271501884)*x_ref[1]**o+(0.171072659)*x_ref[1]
        ref[3,4,1,1]=(-0.369797459679)*x_ref[0]**o+(-0.714539720717)*x_ref[0]+(-0.122698805149)*x_ref[1]**o+(0.874232715421)*x_ref[1]
        ref[3,4,2,0]=(0.912569854856)*x_ref[0]**o+(-0.773508960793)*x_ref[0]+(-0.447748330632)*x_ref[1]**o+(-0.0117517938995)*x_ref[1]
        ref[3,4,2,1]=(-0.467442835658)*x_ref[0]**o+(-0.792740521854)*x_ref[0]+(-0.33441666815)*x_ref[1]**o+(-0.821650769855)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.171866645738)*x[0]**o+(0.561850203514)*x[0]+(0.416451445795)*x[1]**o+(-0.235582248693)*x[1]+(-0.196259994197)*x[2]**o+(0.940566952036)*x[2]
        arg[0,0,0,1]=(-0.00971758894531)*x[0]**o+(-0.590140440451)*x[0]+(-0.719992277571)*x[1]**o+(-0.0411344370809)*x[1]+(-0.301518484556)*x[2]**o+(-0.624210342075)*x[2]
        arg[0,0,1,0]=(0.142493885612)*x[0]**o+(0.473915143223)*x[0]+(0.236217877001)*x[1]**o+(-0.26535451945)*x[1]+(0.197580058644)*x[2]**o+(0.440093349171)*x[2]
        arg[0,0,1,1]=(0.532930185145)*x[0]**o+(0.206573108032)*x[0]+(-0.903235976279)*x[1]**o+(-0.192856647484)*x[1]+(0.247399954122)*x[2]**o+(-0.396060116726)*x[2]
        arg[0,0,2,0]=(0.106261668354)*x[0]**o+(-0.169092441923)*x[0]+(0.495324512568)*x[1]**o+(0.612873784609)*x[1]+(-0.181887252227)*x[2]**o+(-0.349265048779)*x[2]
        arg[0,0,2,1]=(0.0407741271369)*x[0]**o+(-0.286131832952)*x[0]+(0.00801790328545)*x[1]**o+(-0.930580202402)*x[1]+(-0.977595004998)*x[2]**o+(-0.176312831325)*x[2]
        arg[0,1,0,0]=(0.0660805950534)*x[0]**o+(0.27279659698)*x[0]+(0.383210617595)*x[1]**o+(-0.453071384847)*x[1]+(-0.497520020161)*x[2]**o+(-0.0622858051806)*x[2]
        arg[0,1,0,1]=(0.343928280381)*x[0]**o+(0.612026453223)*x[0]+(-0.278958038519)*x[1]**o+(-0.287367815759)*x[1]+(-0.975555454735)*x[2]**o+(-0.0381461641238)*x[2]
        arg[0,1,1,0]=(-0.651824952239)*x[0]**o+(0.504032353063)*x[0]+(-0.505340720597)*x[1]**o+(0.804430789885)*x[1]+(0.626321487603)*x[2]**o+(0.771921294247)*x[2]
        arg[0,1,1,1]=(-0.677075301947)*x[0]**o+(0.977678280199)*x[0]+(-0.0282503291518)*x[1]**o+(0.383805851778)*x[1]+(-0.942267069472)*x[2]**o+(-0.168439203764)*x[2]
        arg[0,1,2,0]=(-0.218809773899)*x[0]**o+(-0.923486201344)*x[0]+(-0.771580374186)*x[1]**o+(-0.802318678915)*x[1]+(-0.817711294844)*x[2]**o+(-0.524488450611)*x[2]
        arg[0,1,2,1]=(-0.258192021864)*x[0]**o+(-0.30201835948)*x[0]+(0.908539532951)*x[1]**o+(0.53218981056)*x[1]+(0.0337395431484)*x[2]**o+(0.358676387738)*x[2]
        arg[0,2,0,0]=(0.181349802806)*x[0]**o+(-0.67447696451)*x[0]+(-0.905389539679)*x[1]**o+(-0.0192162208351)*x[1]+(0.475994769738)*x[2]**o+(0.258880626391)*x[2]
        arg[0,2,0,1]=(0.683461009428)*x[0]**o+(0.101633874876)*x[0]+(-0.666120146454)*x[1]**o+(-0.79937847801)*x[1]+(-0.0519440794346)*x[2]**o+(0.412543335301)*x[2]
        arg[0,2,1,0]=(-0.705756906195)*x[0]**o+(-0.914133519297)*x[0]+(0.203329174425)*x[1]**o+(-0.125897560162)*x[1]+(-0.845555978817)*x[2]**o+(-0.376117448037)*x[2]
        arg[0,2,1,1]=(0.129675218633)*x[0]**o+(0.230278463142)*x[0]+(-0.0465924540599)*x[1]**o+(0.618850924426)*x[1]+(-0.981020601489)*x[2]**o+(-0.446786649897)*x[2]
        arg[0,2,2,0]=(0.709293140521)*x[0]**o+(-0.117595235921)*x[0]+(0.00527289231769)*x[1]**o+(0.407867521466)*x[1]+(-0.0513211295402)*x[2]**o+(-0.775396743388)*x[2]
        arg[0,2,2,1]=(0.456146533185)*x[0]**o+(0.357343068604)*x[0]+(-0.596648639307)*x[1]**o+(0.0400180634504)*x[1]+(-0.773508416513)*x[2]**o+(0.873275482135)*x[2]
        arg[0,3,0,0]=(-0.553923806627)*x[0]**o+(0.416008424454)*x[0]+(-0.18857445771)*x[1]**o+(0.668616517137)*x[1]+(0.850445273446)*x[2]**o+(0.0324670478739)*x[2]
        arg[0,3,0,1]=(-0.556129092409)*x[0]**o+(-0.207129568151)*x[0]+(-0.616200225149)*x[1]**o+(0.173694871558)*x[1]+(-0.389943310499)*x[2]**o+(-0.0171456312507)*x[2]
        arg[0,3,1,0]=(0.029375229156)*x[0]**o+(-0.710817311305)*x[0]+(0.487826213322)*x[1]**o+(-0.353110144226)*x[1]+(-0.326240277953)*x[2]**o+(0.435853126161)*x[2]
        arg[0,3,1,1]=(-0.973283235545)*x[0]**o+(0.119293099332)*x[0]+(-0.987231590871)*x[1]**o+(-0.778410371876)*x[1]+(-0.123668803712)*x[2]**o+(0.836409669329)*x[2]
        arg[0,3,2,0]=(-0.953711717094)*x[0]**o+(0.0211462821501)*x[0]+(0.989232030291)*x[1]**o+(0.684525652819)*x[1]+(-0.934499312995)*x[2]**o+(0.0759185573729)*x[2]
        arg[0,3,2,1]=(0.962555034668)*x[0]**o+(-0.344084097742)*x[0]+(0.627919365161)*x[1]**o+(-0.0520511084603)*x[1]+(0.548149991599)*x[2]**o+(0.885011282903)*x[2]
        arg[0,4,0,0]=(-0.604624034071)*x[0]**o+(0.194224357892)*x[0]+(-0.171628912052)*x[1]**o+(0.0361170699761)*x[1]+(0.68259161986)*x[2]**o+(-0.47807189026)*x[2]
        arg[0,4,0,1]=(-0.441358360042)*x[0]**o+(0.854885170771)*x[0]+(0.739245976992)*x[1]**o+(-0.712495821809)*x[1]+(0.786863605702)*x[2]**o+(0.244289741577)*x[2]
        arg[0,4,1,0]=(-0.93437768143)*x[0]**o+(0.909235101604)*x[0]+(0.398343776489)*x[1]**o+(-0.8015457882)*x[1]+(-0.045056420534)*x[2]**o+(-0.0696087139589)*x[2]
        arg[0,4,1,1]=(0.997282980634)*x[0]**o+(-0.523146965095)*x[0]+(0.888300277406)*x[1]**o+(0.0692584552317)*x[1]+(-0.361183524085)*x[2]**o+(0.298003855572)*x[2]
        arg[0,4,2,0]=(0.510704730088)*x[0]**o+(-0.288621287507)*x[0]+(0.694375966407)*x[1]**o+(0.438558764462)*x[1]+(0.999793245458)*x[2]**o+(0.146242565823)*x[2]
        arg[0,4,2,1]=(0.461717367889)*x[0]**o+(0.585419576536)*x[0]+(0.172514336678)*x[1]**o+(-0.372717466388)*x[1]+(0.354162503692)*x[2]**o+(0.465497372897)*x[2]
        arg[1,0,0,0]=(0.536758620313)*x[0]**o+(0.428277991055)*x[0]+(-0.60617935999)*x[1]**o+(-0.889897682367)*x[1]+(-0.326378737219)*x[2]**o+(0.735260248159)*x[2]
        arg[1,0,0,1]=(-0.629157417304)*x[0]**o+(-0.00171372949641)*x[0]+(0.110550936359)*x[1]**o+(-0.418314227619)*x[1]+(0.968147035552)*x[2]**o+(-0.58304546387)*x[2]
        arg[1,0,1,0]=(0.528767817229)*x[0]**o+(0.111986940482)*x[0]+(0.853287140505)*x[1]**o+(0.551761795603)*x[1]+(-0.848026685332)*x[2]**o+(0.799725876993)*x[2]
        arg[1,0,1,1]=(0.154974399407)*x[0]**o+(-0.662212037776)*x[0]+(0.628908352518)*x[1]**o+(-0.0175576929699)*x[1]+(-0.0501253203152)*x[2]**o+(0.147651837545)*x[2]
        arg[1,0,2,0]=(0.53576644491)*x[0]**o+(-0.935865707952)*x[0]+(-0.256921357737)*x[1]**o+(0.452311436169)*x[1]+(-0.449646854354)*x[2]**o+(0.0137632023964)*x[2]
        arg[1,0,2,1]=(-0.103323182353)*x[0]**o+(-0.0636296047221)*x[0]+(-0.681883331812)*x[1]**o+(-0.284589673307)*x[1]+(-0.525524911635)*x[2]**o+(0.525757772842)*x[2]
        arg[1,1,0,0]=(0.106186299307)*x[0]**o+(0.193833690397)*x[0]+(0.755686652252)*x[1]**o+(-0.954503081175)*x[1]+(-0.946058471992)*x[2]**o+(0.258850283735)*x[2]
        arg[1,1,0,1]=(-0.352425152661)*x[0]**o+(-0.955207432287)*x[0]+(0.946989898911)*x[1]**o+(0.426365715943)*x[1]+(-0.225898340088)*x[2]**o+(-0.108613105442)*x[2]
        arg[1,1,1,0]=(0.505488482593)*x[0]**o+(-0.0193114896733)*x[0]+(-0.809537000564)*x[1]**o+(0.39999431472)*x[1]+(-0.968315699948)*x[2]**o+(0.672744543525)*x[2]
        arg[1,1,1,1]=(0.0967454491182)*x[0]**o+(-0.608840080579)*x[0]+(0.287989672484)*x[1]**o+(-0.989435924681)*x[1]+(0.46833865882)*x[2]**o+(0.0211042567481)*x[2]
        arg[1,1,2,0]=(-0.461777513493)*x[0]**o+(-0.539619783995)*x[0]+(-0.719051252797)*x[1]**o+(-0.980491094424)*x[1]+(0.526872369892)*x[2]**o+(0.534649090672)*x[2]
        arg[1,1,2,1]=(-0.498068667635)*x[0]**o+(-0.977138256259)*x[0]+(0.388405877405)*x[1]**o+(-0.652220953523)*x[1]+(0.214733833456)*x[2]**o+(0.925064130776)*x[2]
        arg[1,2,0,0]=(-0.728915088302)*x[0]**o+(-0.867653377354)*x[0]+(-0.0240751909884)*x[1]**o+(-0.617932013511)*x[1]+(0.342417783703)*x[2]**o+(0.269029568448)*x[2]
        arg[1,2,0,1]=(0.659030580895)*x[0]**o+(0.460057036148)*x[0]+(0.0556749040721)*x[1]**o+(-0.0943860780082)*x[1]+(0.940068201493)*x[2]**o+(-0.156531042318)*x[2]
        arg[1,2,1,0]=(-0.0373627034239)*x[0]**o+(-0.631862915794)*x[0]+(0.384889489295)*x[1]**o+(-0.85645836005)*x[1]+(-0.448074134473)*x[2]**o+(-0.103825620602)*x[2]
        arg[1,2,1,1]=(-0.19142928199)*x[0]**o+(-0.993719455322)*x[0]+(0.204439624689)*x[1]**o+(-0.0616134769914)*x[1]+(0.581621603387)*x[2]**o+(-0.554871585864)*x[2]
        arg[1,2,2,0]=(0.684658549351)*x[0]**o+(-0.0520805658251)*x[0]+(-0.549405651115)*x[1]**o+(-0.0890223507953)*x[1]+(0.538847108562)*x[2]**o+(0.555575374976)*x[2]
        arg[1,2,2,1]=(-0.197138807828)*x[0]**o+(0.691528695453)*x[0]+(0.337770131451)*x[1]**o+(-0.0662330700506)*x[1]+(-0.556170962811)*x[2]**o+(-0.415321654558)*x[2]
        arg[1,3,0,0]=(-0.255542847866)*x[0]**o+(-0.731103056475)*x[0]+(-0.764119472108)*x[1]**o+(0.264775911336)*x[1]+(0.976604892324)*x[2]**o+(-0.929820558374)*x[2]
        arg[1,3,0,1]=(-0.506143192998)*x[0]**o+(-0.139848041055)*x[0]+(-0.074839899541)*x[1]**o+(-0.35100597507)*x[1]+(-0.127574366194)*x[2]**o+(-0.0084655916902)*x[2]
        arg[1,3,1,0]=(0.099997170805)*x[0]**o+(-0.149089017918)*x[0]+(0.419790611192)*x[1]**o+(-0.148698964676)*x[1]+(0.490423198386)*x[2]**o+(-0.210076971679)*x[2]
        arg[1,3,1,1]=(-0.139371350292)*x[0]**o+(-0.320648977772)*x[0]+(0.836558622998)*x[1]**o+(0.394491153689)*x[1]+(0.0857621395676)*x[2]**o+(-0.709138669156)*x[2]
        arg[1,3,2,0]=(0.743901534332)*x[0]**o+(-0.249486095838)*x[0]+(0.281087339807)*x[1]**o+(0.617669579771)*x[1]+(-0.383135240681)*x[2]**o+(-0.647267353645)*x[2]
        arg[1,3,2,1]=(-0.465587857279)*x[0]**o+(0.578094609537)*x[0]+(0.400668301655)*x[1]**o+(-0.999984469723)*x[1]+(-0.460584024789)*x[2]**o+(0.943466859654)*x[2]
        arg[1,4,0,0]=(-0.246915327144)*x[0]**o+(-0.203163427715)*x[0]+(0.138589009969)*x[1]**o+(-0.705709971036)*x[1]+(0.273085758612)*x[2]**o+(0.901084781824)*x[2]
        arg[1,4,0,1]=(-0.919682459969)*x[0]**o+(-0.476395399467)*x[0]+(0.00638460254264)*x[1]**o+(-0.474710529098)*x[1]+(-0.935681998683)*x[2]**o+(0.461344674976)*x[2]
        arg[1,4,1,0]=(0.8225038116)*x[0]**o+(0.0855444334027)*x[0]+(-0.685983943046)*x[1]**o+(-0.46700703117)*x[1]+(0.760260028729)*x[2]**o+(-0.190951535155)*x[2]
        arg[1,4,1,1]=(0.512940503439)*x[0]**o+(0.998803686398)*x[0]+(0.200914807938)*x[1]**o+(0.793533216819)*x[1]+(0.980475771536)*x[2]**o+(-0.00307515759413)*x[2]
        arg[1,4,2,0]=(0.278761841603)*x[0]**o+(-0.368232992504)*x[0]+(0.939116142167)*x[1]**o+(0.13661977072)*x[1]+(-0.490378973691)*x[2]**o+(-0.409117913136)*x[2]
        arg[1,4,2,1]=(-0.456090186141)*x[0]**o+(-0.881413614887)*x[0]+(-0.951099957833)*x[1]**o+(0.859690794224)*x[1]+(0.0723671154601)*x[2]**o+(0.468823687196)*x[2]
        arg[2,0,0,0]=(-0.415483671814)*x[0]**o+(0.415160471625)*x[0]+(-0.730798143253)*x[1]**o+(-0.90883538519)*x[1]+(-0.241870631235)*x[2]**o+(-0.549577536182)*x[2]
        arg[2,0,0,1]=(0.86985374761)*x[0]**o+(0.378073808491)*x[0]+(0.0923146623212)*x[1]**o+(-0.912604209727)*x[1]+(0.529162434789)*x[2]**o+(-0.135836425669)*x[2]
        arg[2,0,1,0]=(0.702609345242)*x[0]**o+(-0.63859466548)*x[0]+(-0.2668268067)*x[1]**o+(0.107377268995)*x[1]+(-0.575458650368)*x[2]**o+(-0.359972696269)*x[2]
        arg[2,0,1,1]=(-0.169295117134)*x[0]**o+(-0.0780580709275)*x[0]+(0.9803920116)*x[1]**o+(0.794090800797)*x[1]+(0.71271688197)*x[2]**o+(0.265355025589)*x[2]
        arg[2,0,2,0]=(0.700475435131)*x[0]**o+(0.317025249264)*x[0]+(0.885140001575)*x[1]**o+(-0.733992012528)*x[1]+(-0.937990577058)*x[2]**o+(-0.566197708973)*x[2]
        arg[2,0,2,1]=(0.369801749993)*x[0]**o+(0.292267254334)*x[0]+(0.24962184892)*x[1]**o+(-0.912670113872)*x[1]+(0.590235505871)*x[2]**o+(0.199279333126)*x[2]
        arg[2,1,0,0]=(-0.165657934182)*x[0]**o+(0.837030192613)*x[0]+(-0.615301249273)*x[1]**o+(-0.947633937776)*x[1]+(0.916677465451)*x[2]**o+(0.633727682324)*x[2]
        arg[2,1,0,1]=(0.260787013495)*x[0]**o+(0.367146934614)*x[0]+(0.0271903085878)*x[1]**o+(0.853398837558)*x[1]+(-0.381379969861)*x[2]**o+(0.140001581749)*x[2]
        arg[2,1,1,0]=(0.121628758983)*x[0]**o+(-0.212350487769)*x[0]+(-0.198031244861)*x[1]**o+(-0.695620829484)*x[1]+(-0.0533521134282)*x[2]**o+(0.956103910898)*x[2]
        arg[2,1,1,1]=(-0.321102504727)*x[0]**o+(-0.920842546991)*x[0]+(0.082643684589)*x[1]**o+(-0.0960346141926)*x[1]+(0.475269198456)*x[2]**o+(-0.184473007427)*x[2]
        arg[2,1,2,0]=(0.412066628472)*x[0]**o+(-0.667420990332)*x[0]+(0.754028295628)*x[1]**o+(0.967928426879)*x[1]+(-0.129320768473)*x[2]**o+(-0.496085464716)*x[2]
        arg[2,1,2,1]=(-0.379495871082)*x[0]**o+(0.768674450671)*x[0]+(0.786595353344)*x[1]**o+(-0.191866280074)*x[1]+(0.314572026852)*x[2]**o+(-0.0245751436484)*x[2]
        arg[2,2,0,0]=(-0.291171443165)*x[0]**o+(0.857660522727)*x[0]+(-0.0611031323308)*x[1]**o+(0.365691706254)*x[1]+(0.906199720505)*x[2]**o+(0.621747944156)*x[2]
        arg[2,2,0,1]=(-0.110402534173)*x[0]**o+(-0.465943589112)*x[0]+(0.896364485404)*x[1]**o+(0.16234913823)*x[1]+(0.443564946892)*x[2]**o+(0.259358795418)*x[2]
        arg[2,2,1,0]=(-0.480613805459)*x[0]**o+(-0.837751646963)*x[0]+(0.803669049568)*x[1]**o+(-0.859501917481)*x[1]+(0.633523536226)*x[2]**o+(0.351742190573)*x[2]
        arg[2,2,1,1]=(0.133665578883)*x[0]**o+(-0.283052337652)*x[0]+(-0.382761111917)*x[1]**o+(-0.197039443436)*x[1]+(0.081520912443)*x[2]**o+(-0.167181778234)*x[2]
        arg[2,2,2,0]=(0.207787204976)*x[0]**o+(0.245207436631)*x[0]+(0.340092449827)*x[1]**o+(0.541457158709)*x[1]+(-0.606955763027)*x[2]**o+(-0.197506280875)*x[2]
        arg[2,2,2,1]=(0.399727427867)*x[0]**o+(-0.0452717156729)*x[0]+(0.85643441)*x[1]**o+(-0.938966108133)*x[1]+(0.832098573693)*x[2]**o+(-0.375366657117)*x[2]
        arg[2,3,0,0]=(-0.661411409657)*x[0]**o+(0.702058724486)*x[0]+(0.851820224497)*x[1]**o+(-0.984572528809)*x[1]+(0.700617112775)*x[2]**o+(0.990437560702)*x[2]
        arg[2,3,0,1]=(-0.937637737824)*x[0]**o+(0.970774222765)*x[0]+(0.0263154279607)*x[1]**o+(-0.453145473796)*x[1]+(-0.740101764089)*x[2]**o+(0.248604522836)*x[2]
        arg[2,3,1,0]=(0.624217308865)*x[0]**o+(0.682749411369)*x[0]+(0.394444314015)*x[1]**o+(0.288021257523)*x[1]+(-0.943712429774)*x[2]**o+(-0.299272610192)*x[2]
        arg[2,3,1,1]=(0.00512058360895)*x[0]**o+(-0.0263977566592)*x[0]+(-0.769491595627)*x[1]**o+(-0.403091037334)*x[1]+(-0.940178247778)*x[2]**o+(-0.664590180141)*x[2]
        arg[2,3,2,0]=(-0.606361369179)*x[0]**o+(0.0252279171688)*x[0]+(0.0125424811264)*x[1]**o+(-0.249080159522)*x[1]+(0.969599361427)*x[2]**o+(0.369428527325)*x[2]
        arg[2,3,2,1]=(-0.899501409225)*x[0]**o+(0.834437734927)*x[0]+(-0.992484449733)*x[1]**o+(0.00567086116985)*x[1]+(0.112432151044)*x[2]**o+(0.725013775434)*x[2]
        arg[2,4,0,0]=(-0.864861209468)*x[0]**o+(0.475518351325)*x[0]+(-0.575172287508)*x[1]**o+(-0.695754751991)*x[1]+(0.935933655733)*x[2]**o+(-0.156489388279)*x[2]
        arg[2,4,0,1]=(-0.072905918035)*x[0]**o+(0.912617659416)*x[0]+(0.76767638187)*x[1]**o+(0.517738398356)*x[1]+(-0.116771137305)*x[2]**o+(-0.887027853796)*x[2]
        arg[2,4,1,0]=(-0.169351806151)*x[0]**o+(-0.667994142736)*x[0]+(-0.742400183746)*x[1]**o+(-0.345483488309)*x[1]+(0.534459208989)*x[2]**o+(-0.314632556106)*x[2]
        arg[2,4,1,1]=(0.571889164138)*x[0]**o+(-0.603799705213)*x[0]+(0.574246013631)*x[1]**o+(-0.156121300765)*x[1]+(-0.300131083217)*x[2]**o+(0.781527389701)*x[2]
        arg[2,4,2,0]=(-0.433340783774)*x[0]**o+(-0.337577532056)*x[0]+(-0.991017325155)*x[1]**o+(0.169866142612)*x[1]+(0.531407605197)*x[2]**o+(0.91525228665)*x[2]
        arg[2,4,2,1]=(0.837883720242)*x[0]**o+(-0.647682324772)*x[0]+(0.125057981769)*x[1]**o+(-0.734369688409)*x[1]+(0.5864622868)*x[2]**o+(-0.378030111621)*x[2]
        arg[3,0,0,0]=(-0.497406654549)*x[0]**o+(0.773396244829)*x[0]+(0.587080735986)*x[1]**o+(0.80571073543)*x[1]+(-0.621583700283)*x[2]**o+(0.214586226671)*x[2]
        arg[3,0,0,1]=(0.602557120764)*x[0]**o+(0.362884272628)*x[0]+(-0.469709604155)*x[1]**o+(0.570294839284)*x[1]+(-0.977082842755)*x[2]**o+(-0.871580274035)*x[2]
        arg[3,0,1,0]=(0.801845637209)*x[0]**o+(0.65646927893)*x[0]+(0.181490001743)*x[1]**o+(-0.579155840484)*x[1]+(-0.690604456924)*x[2]**o+(0.999798826875)*x[2]
        arg[3,0,1,1]=(-0.0835741146147)*x[0]**o+(0.251697832709)*x[0]+(0.915417127133)*x[1]**o+(0.859412289228)*x[1]+(0.529025246336)*x[2]**o+(0.0290945910046)*x[2]
        arg[3,0,2,0]=(0.43632860403)*x[0]**o+(-0.71378809102)*x[0]+(0.996769149715)*x[1]**o+(-0.694095684973)*x[1]+(0.24606095312)*x[2]**o+(0.0974139948716)*x[2]
        arg[3,0,2,1]=(-0.615800860737)*x[0]**o+(-0.0194774462371)*x[0]+(-0.323266780598)*x[1]**o+(0.70039319915)*x[1]+(0.0990606501268)*x[2]**o+(-0.265400117193)*x[2]
        arg[3,1,0,0]=(0.637788917841)*x[0]**o+(-0.898134292781)*x[0]+(0.397058421405)*x[1]**o+(0.424875003774)*x[1]+(0.493154896437)*x[2]**o+(0.426966521522)*x[2]
        arg[3,1,0,1]=(0.307016370261)*x[0]**o+(0.583037237501)*x[0]+(-0.715510433757)*x[1]**o+(-0.778084356162)*x[1]+(0.624410868665)*x[2]**o+(0.534413229876)*x[2]
        arg[3,1,1,0]=(0.53494197621)*x[0]**o+(0.0347747301727)*x[0]+(0.387683566308)*x[1]**o+(0.79111981305)*x[1]+(0.499346741486)*x[2]**o+(-0.89736311061)*x[2]
        arg[3,1,1,1]=(-0.414505914492)*x[0]**o+(0.738343334487)*x[0]+(0.0839982715753)*x[1]**o+(0.306059781071)*x[1]+(0.205292188103)*x[2]**o+(0.478479664712)*x[2]
        arg[3,1,2,0]=(0.649444916094)*x[0]**o+(-0.616363283285)*x[0]+(-0.890710011556)*x[1]**o+(-0.55638070962)*x[1]+(0.63397005463)*x[2]**o+(-0.296632593945)*x[2]
        arg[3,1,2,1]=(0.867500426375)*x[0]**o+(0.087056846585)*x[0]+(-0.347167872287)*x[1]**o+(-0.593410409382)*x[1]+(-0.0853161439992)*x[2]**o+(0.32855927156)*x[2]
        arg[3,2,0,0]=(0.176880322402)*x[0]**o+(-0.342939990014)*x[0]+(0.927975205318)*x[1]**o+(0.395166493409)*x[1]+(-0.114428155908)*x[2]**o+(0.518356015938)*x[2]
        arg[3,2,0,1]=(-0.560834784775)*x[0]**o+(0.989745819909)*x[0]+(-0.201732235802)*x[1]**o+(-0.953838293418)*x[1]+(-0.797223863765)*x[2]**o+(-0.44533752286)*x[2]
        arg[3,2,1,0]=(-0.83129861256)*x[0]**o+(-0.768067351258)*x[0]+(-0.564704574946)*x[1]**o+(0.419961478425)*x[1]+(0.561243126484)*x[2]**o+(0.536547545961)*x[2]
        arg[3,2,1,1]=(-0.245876224002)*x[0]**o+(-0.0399839529678)*x[0]+(-0.728071866725)*x[1]**o+(0.68478402008)*x[1]+(0.907895423543)*x[2]**o+(0.454698589619)*x[2]
        arg[3,2,2,0]=(-0.412012705012)*x[0]**o+(-0.575272496028)*x[0]+(0.877861906104)*x[1]**o+(-0.586319475679)*x[1]+(0.668997061563)*x[2]**o+(-0.418342969428)*x[2]
        arg[3,2,2,1]=(0.270232215934)*x[0]**o+(0.236642564222)*x[0]+(-0.0693779443521)*x[1]**o+(0.77329883812)*x[1]+(-0.743424502879)*x[2]**o+(0.296073856497)*x[2]
        arg[3,3,0,0]=(-0.0659865092333)*x[0]**o+(-0.645286369943)*x[0]+(0.746214220465)*x[1]**o+(-0.549592848453)*x[1]+(-0.194552740752)*x[2]**o+(0.136734604935)*x[2]
        arg[3,3,0,1]=(0.0407640382533)*x[0]**o+(-0.720002890774)*x[0]+(-0.902500650998)*x[1]**o+(-0.0763091975748)*x[1]+(-0.70279308164)*x[2]**o+(-0.412384977965)*x[2]
        arg[3,3,1,0]=(0.460226151062)*x[0]**o+(-0.359911871464)*x[0]+(0.654909639152)*x[1]**o+(-0.574154331236)*x[1]+(-0.226246538275)*x[2]**o+(-0.168145308987)*x[2]
        arg[3,3,1,1]=(-0.966169813904)*x[0]**o+(-0.784593122487)*x[0]+(-0.857455334787)*x[1]**o+(-0.116963343366)*x[1]+(0.231055675695)*x[2]**o+(-0.953091037377)*x[2]
        arg[3,3,2,0]=(-0.796170783237)*x[0]**o+(-0.605030579381)*x[0]+(0.388310055252)*x[1]**o+(-0.504349251868)*x[1]+(0.83840677955)*x[2]**o+(-0.586099851479)*x[2]
        arg[3,3,2,1]=(0.195092011781)*x[0]**o+(-0.358423372419)*x[0]+(-0.643669425281)*x[1]**o+(0.721564003575)*x[1]+(-0.588431193874)*x[2]**o+(-0.378625221371)*x[2]
        arg[3,4,0,0]=(0.585756187041)*x[0]**o+(-0.753400165107)*x[0]+(0.790439809615)*x[1]**o+(0.0473109698221)*x[1]+(0.951768026456)*x[2]**o+(0.423960993524)*x[2]
        arg[3,4,0,1]=(0.32857275647)*x[0]**o+(0.566198295813)*x[0]+(-0.372998380277)*x[1]**o+(0.671551057967)*x[1]+(-0.990512903698)*x[2]**o+(-0.598820986842)*x[2]
        arg[3,4,1,0]=(0.512362923169)*x[0]**o+(0.161976775973)*x[0]+(-0.823370116651)*x[1]**o+(-0.539694552746)*x[1]+(-0.252135888305)*x[2]**o+(-0.323536133539)*x[2]
        arg[3,4,1,1]=(0.342762199745)*x[0]**o+(0.569702637139)*x[0]+(0.514189130968)*x[1]**o+(-0.219203654565)*x[1]+(-0.839606678057)*x[2]**o+(0.374386070142)*x[2]
        arg[3,4,2,0]=(0.0568713695577)*x[0]**o+(-0.494349350234)*x[0]+(0.829657653367)*x[1]**o+(0.541489358189)*x[1]+(-0.9469529858)*x[2]**o+(0.230257255487)*x[2]
        arg[3,4,2,1]=(0.874227414064)*x[0]**o+(0.76572262681)*x[0]+(0.305716770957)*x[1]**o+(0.225904980409)*x[1]+(-0.10063057638)*x[2]**o+(-0.87920628616)*x[2]
        ref[0,0,0,0]=(-0.171866645738)*x_ref[0]**o+(0.561850203514)*x_ref[0]+(0.416451445795)*x_ref[1]**o+(-0.235582248693)*x_ref[1]+(-0.196259994197)*x_ref[2]**o+(0.940566952036)*x_ref[2]
        ref[0,0,0,1]=(-0.00971758894531)*x_ref[0]**o+(-0.590140440451)*x_ref[0]+(-0.719992277571)*x_ref[1]**o+(-0.0411344370809)*x_ref[1]+(-0.301518484556)*x_ref[2]**o+(-0.624210342075)*x_ref[2]
        ref[0,0,1,0]=(0.142493885612)*x_ref[0]**o+(0.473915143223)*x_ref[0]+(0.236217877001)*x_ref[1]**o+(-0.26535451945)*x_ref[1]+(0.197580058644)*x_ref[2]**o+(0.440093349171)*x_ref[2]
        ref[0,0,1,1]=(0.532930185145)*x_ref[0]**o+(0.206573108032)*x_ref[0]+(-0.903235976279)*x_ref[1]**o+(-0.192856647484)*x_ref[1]+(0.247399954122)*x_ref[2]**o+(-0.396060116726)*x_ref[2]
        ref[0,0,2,0]=(0.106261668354)*x_ref[0]**o+(-0.169092441923)*x_ref[0]+(0.495324512568)*x_ref[1]**o+(0.612873784609)*x_ref[1]+(-0.181887252227)*x_ref[2]**o+(-0.349265048779)*x_ref[2]
        ref[0,0,2,1]=(0.0407741271369)*x_ref[0]**o+(-0.286131832952)*x_ref[0]+(0.00801790328545)*x_ref[1]**o+(-0.930580202402)*x_ref[1]+(-0.977595004998)*x_ref[2]**o+(-0.176312831325)*x_ref[2]
        ref[0,1,0,0]=(0.0660805950534)*x_ref[0]**o+(0.27279659698)*x_ref[0]+(0.383210617595)*x_ref[1]**o+(-0.453071384847)*x_ref[1]+(-0.497520020161)*x_ref[2]**o+(-0.0622858051806)*x_ref[2]
        ref[0,1,0,1]=(0.343928280381)*x_ref[0]**o+(0.612026453223)*x_ref[0]+(-0.278958038519)*x_ref[1]**o+(-0.287367815759)*x_ref[1]+(-0.975555454735)*x_ref[2]**o+(-0.0381461641238)*x_ref[2]
        ref[0,1,1,0]=(-0.651824952239)*x_ref[0]**o+(0.504032353063)*x_ref[0]+(-0.505340720597)*x_ref[1]**o+(0.804430789885)*x_ref[1]+(0.626321487603)*x_ref[2]**o+(0.771921294247)*x_ref[2]
        ref[0,1,1,1]=(-0.677075301947)*x_ref[0]**o+(0.977678280199)*x_ref[0]+(-0.0282503291518)*x_ref[1]**o+(0.383805851778)*x_ref[1]+(-0.942267069472)*x_ref[2]**o+(-0.168439203764)*x_ref[2]
        ref[0,1,2,0]=(-0.218809773899)*x_ref[0]**o+(-0.923486201344)*x_ref[0]+(-0.771580374186)*x_ref[1]**o+(-0.802318678915)*x_ref[1]+(-0.817711294844)*x_ref[2]**o+(-0.524488450611)*x_ref[2]
        ref[0,1,2,1]=(-0.258192021864)*x_ref[0]**o+(-0.30201835948)*x_ref[0]+(0.908539532951)*x_ref[1]**o+(0.53218981056)*x_ref[1]+(0.0337395431484)*x_ref[2]**o+(0.358676387738)*x_ref[2]
        ref[0,2,0,0]=(0.181349802806)*x_ref[0]**o+(-0.67447696451)*x_ref[0]+(-0.905389539679)*x_ref[1]**o+(-0.0192162208351)*x_ref[1]+(0.475994769738)*x_ref[2]**o+(0.258880626391)*x_ref[2]
        ref[0,2,0,1]=(0.683461009428)*x_ref[0]**o+(0.101633874876)*x_ref[0]+(-0.666120146454)*x_ref[1]**o+(-0.79937847801)*x_ref[1]+(-0.0519440794346)*x_ref[2]**o+(0.412543335301)*x_ref[2]
        ref[0,2,1,0]=(-0.705756906195)*x_ref[0]**o+(-0.914133519297)*x_ref[0]+(0.203329174425)*x_ref[1]**o+(-0.125897560162)*x_ref[1]+(-0.845555978817)*x_ref[2]**o+(-0.376117448037)*x_ref[2]
        ref[0,2,1,1]=(0.129675218633)*x_ref[0]**o+(0.230278463142)*x_ref[0]+(-0.0465924540599)*x_ref[1]**o+(0.618850924426)*x_ref[1]+(-0.981020601489)*x_ref[2]**o+(-0.446786649897)*x_ref[2]
        ref[0,2,2,0]=(0.709293140521)*x_ref[0]**o+(-0.117595235921)*x_ref[0]+(0.00527289231769)*x_ref[1]**o+(0.407867521466)*x_ref[1]+(-0.0513211295402)*x_ref[2]**o+(-0.775396743388)*x_ref[2]
        ref[0,2,2,1]=(0.456146533185)*x_ref[0]**o+(0.357343068604)*x_ref[0]+(-0.596648639307)*x_ref[1]**o+(0.0400180634504)*x_ref[1]+(-0.773508416513)*x_ref[2]**o+(0.873275482135)*x_ref[2]
        ref[0,3,0,0]=(-0.553923806627)*x_ref[0]**o+(0.416008424454)*x_ref[0]+(-0.18857445771)*x_ref[1]**o+(0.668616517137)*x_ref[1]+(0.850445273446)*x_ref[2]**o+(0.0324670478739)*x_ref[2]
        ref[0,3,0,1]=(-0.556129092409)*x_ref[0]**o+(-0.207129568151)*x_ref[0]+(-0.616200225149)*x_ref[1]**o+(0.173694871558)*x_ref[1]+(-0.389943310499)*x_ref[2]**o+(-0.0171456312507)*x_ref[2]
        ref[0,3,1,0]=(0.029375229156)*x_ref[0]**o+(-0.710817311305)*x_ref[0]+(0.487826213322)*x_ref[1]**o+(-0.353110144226)*x_ref[1]+(-0.326240277953)*x_ref[2]**o+(0.435853126161)*x_ref[2]
        ref[0,3,1,1]=(-0.973283235545)*x_ref[0]**o+(0.119293099332)*x_ref[0]+(-0.987231590871)*x_ref[1]**o+(-0.778410371876)*x_ref[1]+(-0.123668803712)*x_ref[2]**o+(0.836409669329)*x_ref[2]
        ref[0,3,2,0]=(-0.953711717094)*x_ref[0]**o+(0.0211462821501)*x_ref[0]+(0.989232030291)*x_ref[1]**o+(0.684525652819)*x_ref[1]+(-0.934499312995)*x_ref[2]**o+(0.0759185573729)*x_ref[2]
        ref[0,3,2,1]=(0.962555034668)*x_ref[0]**o+(-0.344084097742)*x_ref[0]+(0.627919365161)*x_ref[1]**o+(-0.0520511084603)*x_ref[1]+(0.548149991599)*x_ref[2]**o+(0.885011282903)*x_ref[2]
        ref[0,4,0,0]=(-0.604624034071)*x_ref[0]**o+(0.194224357892)*x_ref[0]+(-0.171628912052)*x_ref[1]**o+(0.0361170699761)*x_ref[1]+(0.68259161986)*x_ref[2]**o+(-0.47807189026)*x_ref[2]
        ref[0,4,0,1]=(-0.441358360042)*x_ref[0]**o+(0.854885170771)*x_ref[0]+(0.739245976992)*x_ref[1]**o+(-0.712495821809)*x_ref[1]+(0.786863605702)*x_ref[2]**o+(0.244289741577)*x_ref[2]
        ref[0,4,1,0]=(-0.93437768143)*x_ref[0]**o+(0.909235101604)*x_ref[0]+(0.398343776489)*x_ref[1]**o+(-0.8015457882)*x_ref[1]+(-0.045056420534)*x_ref[2]**o+(-0.0696087139589)*x_ref[2]
        ref[0,4,1,1]=(0.997282980634)*x_ref[0]**o+(-0.523146965095)*x_ref[0]+(0.888300277406)*x_ref[1]**o+(0.0692584552317)*x_ref[1]+(-0.361183524085)*x_ref[2]**o+(0.298003855572)*x_ref[2]
        ref[0,4,2,0]=(0.510704730088)*x_ref[0]**o+(-0.288621287507)*x_ref[0]+(0.694375966407)*x_ref[1]**o+(0.438558764462)*x_ref[1]+(0.999793245458)*x_ref[2]**o+(0.146242565823)*x_ref[2]
        ref[0,4,2,1]=(0.461717367889)*x_ref[0]**o+(0.585419576536)*x_ref[0]+(0.172514336678)*x_ref[1]**o+(-0.372717466388)*x_ref[1]+(0.354162503692)*x_ref[2]**o+(0.465497372897)*x_ref[2]
        ref[1,0,0,0]=(0.536758620313)*x_ref[0]**o+(0.428277991055)*x_ref[0]+(-0.60617935999)*x_ref[1]**o+(-0.889897682367)*x_ref[1]+(-0.326378737219)*x_ref[2]**o+(0.735260248159)*x_ref[2]
        ref[1,0,0,1]=(-0.629157417304)*x_ref[0]**o+(-0.00171372949641)*x_ref[0]+(0.110550936359)*x_ref[1]**o+(-0.418314227619)*x_ref[1]+(0.968147035552)*x_ref[2]**o+(-0.58304546387)*x_ref[2]
        ref[1,0,1,0]=(0.528767817229)*x_ref[0]**o+(0.111986940482)*x_ref[0]+(0.853287140505)*x_ref[1]**o+(0.551761795603)*x_ref[1]+(-0.848026685332)*x_ref[2]**o+(0.799725876993)*x_ref[2]
        ref[1,0,1,1]=(0.154974399407)*x_ref[0]**o+(-0.662212037776)*x_ref[0]+(0.628908352518)*x_ref[1]**o+(-0.0175576929699)*x_ref[1]+(-0.0501253203152)*x_ref[2]**o+(0.147651837545)*x_ref[2]
        ref[1,0,2,0]=(0.53576644491)*x_ref[0]**o+(-0.935865707952)*x_ref[0]+(-0.256921357737)*x_ref[1]**o+(0.452311436169)*x_ref[1]+(-0.449646854354)*x_ref[2]**o+(0.0137632023964)*x_ref[2]
        ref[1,0,2,1]=(-0.103323182353)*x_ref[0]**o+(-0.0636296047221)*x_ref[0]+(-0.681883331812)*x_ref[1]**o+(-0.284589673307)*x_ref[1]+(-0.525524911635)*x_ref[2]**o+(0.525757772842)*x_ref[2]
        ref[1,1,0,0]=(0.106186299307)*x_ref[0]**o+(0.193833690397)*x_ref[0]+(0.755686652252)*x_ref[1]**o+(-0.954503081175)*x_ref[1]+(-0.946058471992)*x_ref[2]**o+(0.258850283735)*x_ref[2]
        ref[1,1,0,1]=(-0.352425152661)*x_ref[0]**o+(-0.955207432287)*x_ref[0]+(0.946989898911)*x_ref[1]**o+(0.426365715943)*x_ref[1]+(-0.225898340088)*x_ref[2]**o+(-0.108613105442)*x_ref[2]
        ref[1,1,1,0]=(0.505488482593)*x_ref[0]**o+(-0.0193114896733)*x_ref[0]+(-0.809537000564)*x_ref[1]**o+(0.39999431472)*x_ref[1]+(-0.968315699948)*x_ref[2]**o+(0.672744543525)*x_ref[2]
        ref[1,1,1,1]=(0.0967454491182)*x_ref[0]**o+(-0.608840080579)*x_ref[0]+(0.287989672484)*x_ref[1]**o+(-0.989435924681)*x_ref[1]+(0.46833865882)*x_ref[2]**o+(0.0211042567481)*x_ref[2]
        ref[1,1,2,0]=(-0.461777513493)*x_ref[0]**o+(-0.539619783995)*x_ref[0]+(-0.719051252797)*x_ref[1]**o+(-0.980491094424)*x_ref[1]+(0.526872369892)*x_ref[2]**o+(0.534649090672)*x_ref[2]
        ref[1,1,2,1]=(-0.498068667635)*x_ref[0]**o+(-0.977138256259)*x_ref[0]+(0.388405877405)*x_ref[1]**o+(-0.652220953523)*x_ref[1]+(0.214733833456)*x_ref[2]**o+(0.925064130776)*x_ref[2]
        ref[1,2,0,0]=(-0.728915088302)*x_ref[0]**o+(-0.867653377354)*x_ref[0]+(-0.0240751909884)*x_ref[1]**o+(-0.617932013511)*x_ref[1]+(0.342417783703)*x_ref[2]**o+(0.269029568448)*x_ref[2]
        ref[1,2,0,1]=(0.659030580895)*x_ref[0]**o+(0.460057036148)*x_ref[0]+(0.0556749040721)*x_ref[1]**o+(-0.0943860780082)*x_ref[1]+(0.940068201493)*x_ref[2]**o+(-0.156531042318)*x_ref[2]
        ref[1,2,1,0]=(-0.0373627034239)*x_ref[0]**o+(-0.631862915794)*x_ref[0]+(0.384889489295)*x_ref[1]**o+(-0.85645836005)*x_ref[1]+(-0.448074134473)*x_ref[2]**o+(-0.103825620602)*x_ref[2]
        ref[1,2,1,1]=(-0.19142928199)*x_ref[0]**o+(-0.993719455322)*x_ref[0]+(0.204439624689)*x_ref[1]**o+(-0.0616134769914)*x_ref[1]+(0.581621603387)*x_ref[2]**o+(-0.554871585864)*x_ref[2]
        ref[1,2,2,0]=(0.684658549351)*x_ref[0]**o+(-0.0520805658251)*x_ref[0]+(-0.549405651115)*x_ref[1]**o+(-0.0890223507953)*x_ref[1]+(0.538847108562)*x_ref[2]**o+(0.555575374976)*x_ref[2]
        ref[1,2,2,1]=(-0.197138807828)*x_ref[0]**o+(0.691528695453)*x_ref[0]+(0.337770131451)*x_ref[1]**o+(-0.0662330700506)*x_ref[1]+(-0.556170962811)*x_ref[2]**o+(-0.415321654558)*x_ref[2]
        ref[1,3,0,0]=(-0.255542847866)*x_ref[0]**o+(-0.731103056475)*x_ref[0]+(-0.764119472108)*x_ref[1]**o+(0.264775911336)*x_ref[1]+(0.976604892324)*x_ref[2]**o+(-0.929820558374)*x_ref[2]
        ref[1,3,0,1]=(-0.506143192998)*x_ref[0]**o+(-0.139848041055)*x_ref[0]+(-0.074839899541)*x_ref[1]**o+(-0.35100597507)*x_ref[1]+(-0.127574366194)*x_ref[2]**o+(-0.0084655916902)*x_ref[2]
        ref[1,3,1,0]=(0.099997170805)*x_ref[0]**o+(-0.149089017918)*x_ref[0]+(0.419790611192)*x_ref[1]**o+(-0.148698964676)*x_ref[1]+(0.490423198386)*x_ref[2]**o+(-0.210076971679)*x_ref[2]
        ref[1,3,1,1]=(-0.139371350292)*x_ref[0]**o+(-0.320648977772)*x_ref[0]+(0.836558622998)*x_ref[1]**o+(0.394491153689)*x_ref[1]+(0.0857621395676)*x_ref[2]**o+(-0.709138669156)*x_ref[2]
        ref[1,3,2,0]=(0.743901534332)*x_ref[0]**o+(-0.249486095838)*x_ref[0]+(0.281087339807)*x_ref[1]**o+(0.617669579771)*x_ref[1]+(-0.383135240681)*x_ref[2]**o+(-0.647267353645)*x_ref[2]
        ref[1,3,2,1]=(-0.465587857279)*x_ref[0]**o+(0.578094609537)*x_ref[0]+(0.400668301655)*x_ref[1]**o+(-0.999984469723)*x_ref[1]+(-0.460584024789)*x_ref[2]**o+(0.943466859654)*x_ref[2]
        ref[1,4,0,0]=(-0.246915327144)*x_ref[0]**o+(-0.203163427715)*x_ref[0]+(0.138589009969)*x_ref[1]**o+(-0.705709971036)*x_ref[1]+(0.273085758612)*x_ref[2]**o+(0.901084781824)*x_ref[2]
        ref[1,4,0,1]=(-0.919682459969)*x_ref[0]**o+(-0.476395399467)*x_ref[0]+(0.00638460254264)*x_ref[1]**o+(-0.474710529098)*x_ref[1]+(-0.935681998683)*x_ref[2]**o+(0.461344674976)*x_ref[2]
        ref[1,4,1,0]=(0.8225038116)*x_ref[0]**o+(0.0855444334027)*x_ref[0]+(-0.685983943046)*x_ref[1]**o+(-0.46700703117)*x_ref[1]+(0.760260028729)*x_ref[2]**o+(-0.190951535155)*x_ref[2]
        ref[1,4,1,1]=(0.512940503439)*x_ref[0]**o+(0.998803686398)*x_ref[0]+(0.200914807938)*x_ref[1]**o+(0.793533216819)*x_ref[1]+(0.980475771536)*x_ref[2]**o+(-0.00307515759413)*x_ref[2]
        ref[1,4,2,0]=(0.278761841603)*x_ref[0]**o+(-0.368232992504)*x_ref[0]+(0.939116142167)*x_ref[1]**o+(0.13661977072)*x_ref[1]+(-0.490378973691)*x_ref[2]**o+(-0.409117913136)*x_ref[2]
        ref[1,4,2,1]=(-0.456090186141)*x_ref[0]**o+(-0.881413614887)*x_ref[0]+(-0.951099957833)*x_ref[1]**o+(0.859690794224)*x_ref[1]+(0.0723671154601)*x_ref[2]**o+(0.468823687196)*x_ref[2]
        ref[2,0,0,0]=(-0.415483671814)*x_ref[0]**o+(0.415160471625)*x_ref[0]+(-0.730798143253)*x_ref[1]**o+(-0.90883538519)*x_ref[1]+(-0.241870631235)*x_ref[2]**o+(-0.549577536182)*x_ref[2]
        ref[2,0,0,1]=(0.86985374761)*x_ref[0]**o+(0.378073808491)*x_ref[0]+(0.0923146623212)*x_ref[1]**o+(-0.912604209727)*x_ref[1]+(0.529162434789)*x_ref[2]**o+(-0.135836425669)*x_ref[2]
        ref[2,0,1,0]=(0.702609345242)*x_ref[0]**o+(-0.63859466548)*x_ref[0]+(-0.2668268067)*x_ref[1]**o+(0.107377268995)*x_ref[1]+(-0.575458650368)*x_ref[2]**o+(-0.359972696269)*x_ref[2]
        ref[2,0,1,1]=(-0.169295117134)*x_ref[0]**o+(-0.0780580709275)*x_ref[0]+(0.9803920116)*x_ref[1]**o+(0.794090800797)*x_ref[1]+(0.71271688197)*x_ref[2]**o+(0.265355025589)*x_ref[2]
        ref[2,0,2,0]=(0.700475435131)*x_ref[0]**o+(0.317025249264)*x_ref[0]+(0.885140001575)*x_ref[1]**o+(-0.733992012528)*x_ref[1]+(-0.937990577058)*x_ref[2]**o+(-0.566197708973)*x_ref[2]
        ref[2,0,2,1]=(0.369801749993)*x_ref[0]**o+(0.292267254334)*x_ref[0]+(0.24962184892)*x_ref[1]**o+(-0.912670113872)*x_ref[1]+(0.590235505871)*x_ref[2]**o+(0.199279333126)*x_ref[2]
        ref[2,1,0,0]=(-0.165657934182)*x_ref[0]**o+(0.837030192613)*x_ref[0]+(-0.615301249273)*x_ref[1]**o+(-0.947633937776)*x_ref[1]+(0.916677465451)*x_ref[2]**o+(0.633727682324)*x_ref[2]
        ref[2,1,0,1]=(0.260787013495)*x_ref[0]**o+(0.367146934614)*x_ref[0]+(0.0271903085878)*x_ref[1]**o+(0.853398837558)*x_ref[1]+(-0.381379969861)*x_ref[2]**o+(0.140001581749)*x_ref[2]
        ref[2,1,1,0]=(0.121628758983)*x_ref[0]**o+(-0.212350487769)*x_ref[0]+(-0.198031244861)*x_ref[1]**o+(-0.695620829484)*x_ref[1]+(-0.0533521134282)*x_ref[2]**o+(0.956103910898)*x_ref[2]
        ref[2,1,1,1]=(-0.321102504727)*x_ref[0]**o+(-0.920842546991)*x_ref[0]+(0.082643684589)*x_ref[1]**o+(-0.0960346141926)*x_ref[1]+(0.475269198456)*x_ref[2]**o+(-0.184473007427)*x_ref[2]
        ref[2,1,2,0]=(0.412066628472)*x_ref[0]**o+(-0.667420990332)*x_ref[0]+(0.754028295628)*x_ref[1]**o+(0.967928426879)*x_ref[1]+(-0.129320768473)*x_ref[2]**o+(-0.496085464716)*x_ref[2]
        ref[2,1,2,1]=(-0.379495871082)*x_ref[0]**o+(0.768674450671)*x_ref[0]+(0.786595353344)*x_ref[1]**o+(-0.191866280074)*x_ref[1]+(0.314572026852)*x_ref[2]**o+(-0.0245751436484)*x_ref[2]
        ref[2,2,0,0]=(-0.291171443165)*x_ref[0]**o+(0.857660522727)*x_ref[0]+(-0.0611031323308)*x_ref[1]**o+(0.365691706254)*x_ref[1]+(0.906199720505)*x_ref[2]**o+(0.621747944156)*x_ref[2]
        ref[2,2,0,1]=(-0.110402534173)*x_ref[0]**o+(-0.465943589112)*x_ref[0]+(0.896364485404)*x_ref[1]**o+(0.16234913823)*x_ref[1]+(0.443564946892)*x_ref[2]**o+(0.259358795418)*x_ref[2]
        ref[2,2,1,0]=(-0.480613805459)*x_ref[0]**o+(-0.837751646963)*x_ref[0]+(0.803669049568)*x_ref[1]**o+(-0.859501917481)*x_ref[1]+(0.633523536226)*x_ref[2]**o+(0.351742190573)*x_ref[2]
        ref[2,2,1,1]=(0.133665578883)*x_ref[0]**o+(-0.283052337652)*x_ref[0]+(-0.382761111917)*x_ref[1]**o+(-0.197039443436)*x_ref[1]+(0.081520912443)*x_ref[2]**o+(-0.167181778234)*x_ref[2]
        ref[2,2,2,0]=(0.207787204976)*x_ref[0]**o+(0.245207436631)*x_ref[0]+(0.340092449827)*x_ref[1]**o+(0.541457158709)*x_ref[1]+(-0.606955763027)*x_ref[2]**o+(-0.197506280875)*x_ref[2]
        ref[2,2,2,1]=(0.399727427867)*x_ref[0]**o+(-0.0452717156729)*x_ref[0]+(0.85643441)*x_ref[1]**o+(-0.938966108133)*x_ref[1]+(0.832098573693)*x_ref[2]**o+(-0.375366657117)*x_ref[2]
        ref[2,3,0,0]=(-0.661411409657)*x_ref[0]**o+(0.702058724486)*x_ref[0]+(0.851820224497)*x_ref[1]**o+(-0.984572528809)*x_ref[1]+(0.700617112775)*x_ref[2]**o+(0.990437560702)*x_ref[2]
        ref[2,3,0,1]=(-0.937637737824)*x_ref[0]**o+(0.970774222765)*x_ref[0]+(0.0263154279607)*x_ref[1]**o+(-0.453145473796)*x_ref[1]+(-0.740101764089)*x_ref[2]**o+(0.248604522836)*x_ref[2]
        ref[2,3,1,0]=(0.624217308865)*x_ref[0]**o+(0.682749411369)*x_ref[0]+(0.394444314015)*x_ref[1]**o+(0.288021257523)*x_ref[1]+(-0.943712429774)*x_ref[2]**o+(-0.299272610192)*x_ref[2]
        ref[2,3,1,1]=(0.00512058360895)*x_ref[0]**o+(-0.0263977566592)*x_ref[0]+(-0.769491595627)*x_ref[1]**o+(-0.403091037334)*x_ref[1]+(-0.940178247778)*x_ref[2]**o+(-0.664590180141)*x_ref[2]
        ref[2,3,2,0]=(-0.606361369179)*x_ref[0]**o+(0.0252279171688)*x_ref[0]+(0.0125424811264)*x_ref[1]**o+(-0.249080159522)*x_ref[1]+(0.969599361427)*x_ref[2]**o+(0.369428527325)*x_ref[2]
        ref[2,3,2,1]=(-0.899501409225)*x_ref[0]**o+(0.834437734927)*x_ref[0]+(-0.992484449733)*x_ref[1]**o+(0.00567086116985)*x_ref[1]+(0.112432151044)*x_ref[2]**o+(0.725013775434)*x_ref[2]
        ref[2,4,0,0]=(-0.864861209468)*x_ref[0]**o+(0.475518351325)*x_ref[0]+(-0.575172287508)*x_ref[1]**o+(-0.695754751991)*x_ref[1]+(0.935933655733)*x_ref[2]**o+(-0.156489388279)*x_ref[2]
        ref[2,4,0,1]=(-0.072905918035)*x_ref[0]**o+(0.912617659416)*x_ref[0]+(0.76767638187)*x_ref[1]**o+(0.517738398356)*x_ref[1]+(-0.116771137305)*x_ref[2]**o+(-0.887027853796)*x_ref[2]
        ref[2,4,1,0]=(-0.169351806151)*x_ref[0]**o+(-0.667994142736)*x_ref[0]+(-0.742400183746)*x_ref[1]**o+(-0.345483488309)*x_ref[1]+(0.534459208989)*x_ref[2]**o+(-0.314632556106)*x_ref[2]
        ref[2,4,1,1]=(0.571889164138)*x_ref[0]**o+(-0.603799705213)*x_ref[0]+(0.574246013631)*x_ref[1]**o+(-0.156121300765)*x_ref[1]+(-0.300131083217)*x_ref[2]**o+(0.781527389701)*x_ref[2]
        ref[2,4,2,0]=(-0.433340783774)*x_ref[0]**o+(-0.337577532056)*x_ref[0]+(-0.991017325155)*x_ref[1]**o+(0.169866142612)*x_ref[1]+(0.531407605197)*x_ref[2]**o+(0.91525228665)*x_ref[2]
        ref[2,4,2,1]=(0.837883720242)*x_ref[0]**o+(-0.647682324772)*x_ref[0]+(0.125057981769)*x_ref[1]**o+(-0.734369688409)*x_ref[1]+(0.5864622868)*x_ref[2]**o+(-0.378030111621)*x_ref[2]
        ref[3,0,0,0]=(-0.497406654549)*x_ref[0]**o+(0.773396244829)*x_ref[0]+(0.587080735986)*x_ref[1]**o+(0.80571073543)*x_ref[1]+(-0.621583700283)*x_ref[2]**o+(0.214586226671)*x_ref[2]
        ref[3,0,0,1]=(0.602557120764)*x_ref[0]**o+(0.362884272628)*x_ref[0]+(-0.469709604155)*x_ref[1]**o+(0.570294839284)*x_ref[1]+(-0.977082842755)*x_ref[2]**o+(-0.871580274035)*x_ref[2]
        ref[3,0,1,0]=(0.801845637209)*x_ref[0]**o+(0.65646927893)*x_ref[0]+(0.181490001743)*x_ref[1]**o+(-0.579155840484)*x_ref[1]+(-0.690604456924)*x_ref[2]**o+(0.999798826875)*x_ref[2]
        ref[3,0,1,1]=(-0.0835741146147)*x_ref[0]**o+(0.251697832709)*x_ref[0]+(0.915417127133)*x_ref[1]**o+(0.859412289228)*x_ref[1]+(0.529025246336)*x_ref[2]**o+(0.0290945910046)*x_ref[2]
        ref[3,0,2,0]=(0.43632860403)*x_ref[0]**o+(-0.71378809102)*x_ref[0]+(0.996769149715)*x_ref[1]**o+(-0.694095684973)*x_ref[1]+(0.24606095312)*x_ref[2]**o+(0.0974139948716)*x_ref[2]
        ref[3,0,2,1]=(-0.615800860737)*x_ref[0]**o+(-0.0194774462371)*x_ref[0]+(-0.323266780598)*x_ref[1]**o+(0.70039319915)*x_ref[1]+(0.0990606501268)*x_ref[2]**o+(-0.265400117193)*x_ref[2]
        ref[3,1,0,0]=(0.637788917841)*x_ref[0]**o+(-0.898134292781)*x_ref[0]+(0.397058421405)*x_ref[1]**o+(0.424875003774)*x_ref[1]+(0.493154896437)*x_ref[2]**o+(0.426966521522)*x_ref[2]
        ref[3,1,0,1]=(0.307016370261)*x_ref[0]**o+(0.583037237501)*x_ref[0]+(-0.715510433757)*x_ref[1]**o+(-0.778084356162)*x_ref[1]+(0.624410868665)*x_ref[2]**o+(0.534413229876)*x_ref[2]
        ref[3,1,1,0]=(0.53494197621)*x_ref[0]**o+(0.0347747301727)*x_ref[0]+(0.387683566308)*x_ref[1]**o+(0.79111981305)*x_ref[1]+(0.499346741486)*x_ref[2]**o+(-0.89736311061)*x_ref[2]
        ref[3,1,1,1]=(-0.414505914492)*x_ref[0]**o+(0.738343334487)*x_ref[0]+(0.0839982715753)*x_ref[1]**o+(0.306059781071)*x_ref[1]+(0.205292188103)*x_ref[2]**o+(0.478479664712)*x_ref[2]
        ref[3,1,2,0]=(0.649444916094)*x_ref[0]**o+(-0.616363283285)*x_ref[0]+(-0.890710011556)*x_ref[1]**o+(-0.55638070962)*x_ref[1]+(0.63397005463)*x_ref[2]**o+(-0.296632593945)*x_ref[2]
        ref[3,1,2,1]=(0.867500426375)*x_ref[0]**o+(0.087056846585)*x_ref[0]+(-0.347167872287)*x_ref[1]**o+(-0.593410409382)*x_ref[1]+(-0.0853161439992)*x_ref[2]**o+(0.32855927156)*x_ref[2]
        ref[3,2,0,0]=(0.176880322402)*x_ref[0]**o+(-0.342939990014)*x_ref[0]+(0.927975205318)*x_ref[1]**o+(0.395166493409)*x_ref[1]+(-0.114428155908)*x_ref[2]**o+(0.518356015938)*x_ref[2]
        ref[3,2,0,1]=(-0.560834784775)*x_ref[0]**o+(0.989745819909)*x_ref[0]+(-0.201732235802)*x_ref[1]**o+(-0.953838293418)*x_ref[1]+(-0.797223863765)*x_ref[2]**o+(-0.44533752286)*x_ref[2]
        ref[3,2,1,0]=(-0.83129861256)*x_ref[0]**o+(-0.768067351258)*x_ref[0]+(-0.564704574946)*x_ref[1]**o+(0.419961478425)*x_ref[1]+(0.561243126484)*x_ref[2]**o+(0.536547545961)*x_ref[2]
        ref[3,2,1,1]=(-0.245876224002)*x_ref[0]**o+(-0.0399839529678)*x_ref[0]+(-0.728071866725)*x_ref[1]**o+(0.68478402008)*x_ref[1]+(0.907895423543)*x_ref[2]**o+(0.454698589619)*x_ref[2]
        ref[3,2,2,0]=(-0.412012705012)*x_ref[0]**o+(-0.575272496028)*x_ref[0]+(0.877861906104)*x_ref[1]**o+(-0.586319475679)*x_ref[1]+(0.668997061563)*x_ref[2]**o+(-0.418342969428)*x_ref[2]
        ref[3,2,2,1]=(0.270232215934)*x_ref[0]**o+(0.236642564222)*x_ref[0]+(-0.0693779443521)*x_ref[1]**o+(0.77329883812)*x_ref[1]+(-0.743424502879)*x_ref[2]**o+(0.296073856497)*x_ref[2]
        ref[3,3,0,0]=(-0.0659865092333)*x_ref[0]**o+(-0.645286369943)*x_ref[0]+(0.746214220465)*x_ref[1]**o+(-0.549592848453)*x_ref[1]+(-0.194552740752)*x_ref[2]**o+(0.136734604935)*x_ref[2]
        ref[3,3,0,1]=(0.0407640382533)*x_ref[0]**o+(-0.720002890774)*x_ref[0]+(-0.902500650998)*x_ref[1]**o+(-0.0763091975748)*x_ref[1]+(-0.70279308164)*x_ref[2]**o+(-0.412384977965)*x_ref[2]
        ref[3,3,1,0]=(0.460226151062)*x_ref[0]**o+(-0.359911871464)*x_ref[0]+(0.654909639152)*x_ref[1]**o+(-0.574154331236)*x_ref[1]+(-0.226246538275)*x_ref[2]**o+(-0.168145308987)*x_ref[2]
        ref[3,3,1,1]=(-0.966169813904)*x_ref[0]**o+(-0.784593122487)*x_ref[0]+(-0.857455334787)*x_ref[1]**o+(-0.116963343366)*x_ref[1]+(0.231055675695)*x_ref[2]**o+(-0.953091037377)*x_ref[2]
        ref[3,3,2,0]=(-0.796170783237)*x_ref[0]**o+(-0.605030579381)*x_ref[0]+(0.388310055252)*x_ref[1]**o+(-0.504349251868)*x_ref[1]+(0.83840677955)*x_ref[2]**o+(-0.586099851479)*x_ref[2]
        ref[3,3,2,1]=(0.195092011781)*x_ref[0]**o+(-0.358423372419)*x_ref[0]+(-0.643669425281)*x_ref[1]**o+(0.721564003575)*x_ref[1]+(-0.588431193874)*x_ref[2]**o+(-0.378625221371)*x_ref[2]
        ref[3,4,0,0]=(0.585756187041)*x_ref[0]**o+(-0.753400165107)*x_ref[0]+(0.790439809615)*x_ref[1]**o+(0.0473109698221)*x_ref[1]+(0.951768026456)*x_ref[2]**o+(0.423960993524)*x_ref[2]
        ref[3,4,0,1]=(0.32857275647)*x_ref[0]**o+(0.566198295813)*x_ref[0]+(-0.372998380277)*x_ref[1]**o+(0.671551057967)*x_ref[1]+(-0.990512903698)*x_ref[2]**o+(-0.598820986842)*x_ref[2]
        ref[3,4,1,0]=(0.512362923169)*x_ref[0]**o+(0.161976775973)*x_ref[0]+(-0.823370116651)*x_ref[1]**o+(-0.539694552746)*x_ref[1]+(-0.252135888305)*x_ref[2]**o+(-0.323536133539)*x_ref[2]
        ref[3,4,1,1]=(0.342762199745)*x_ref[0]**o+(0.569702637139)*x_ref[0]+(0.514189130968)*x_ref[1]**o+(-0.219203654565)*x_ref[1]+(-0.839606678057)*x_ref[2]**o+(0.374386070142)*x_ref[2]
        ref[3,4,2,0]=(0.0568713695577)*x_ref[0]**o+(-0.494349350234)*x_ref[0]+(0.829657653367)*x_ref[1]**o+(0.541489358189)*x_ref[1]+(-0.9469529858)*x_ref[2]**o+(0.230257255487)*x_ref[2]
        ref[3,4,2,1]=(0.874227414064)*x_ref[0]**o+(0.76572262681)*x_ref[0]+(0.305716770957)*x_ref[1]**o+(0.225904980409)*x_ref[1]+(-0.10063057638)*x_ref[2]**o+(-0.87920628616)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-1.02854286859)*x[0]+(0.105922853194)*x[1]
        ref=(-1.02854286859)*x_ref[0]+(0.105922853194)*x_ref[1]
      else:
        arg=(1.36620376771)*x[0]+(-0.336798536224)*x[1]+(0.534538499605)*x[2]
        ref=(1.36620376771)*x_ref[0]+(-0.336798536224)*x_ref[1]+(0.534538499605)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        arg[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
        ref[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        ref[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
      else:
        arg[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        arg[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
        ref[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        ref[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.452635897444)*x[0]+(-0.210513646405)*x[1]
        arg[0,1]=(-0.590704436627)*x[0]+(1.78000273884)*x[1]
        arg[0,2]=(1.01331349402)*x[0]+(1.76279765125)*x[1]
        arg[0,3]=(1.5588852701)*x[0]+(-0.550895238707)*x[1]
        arg[0,4]=(-0.352453804527)*x[0]+(-0.338832613421)*x[1]
        arg[1,0]=(-0.809837877163)*x[0]+(-0.531009602702)*x[1]
        arg[1,1]=(0.513641503665)*x[0]+(0.232114989082)*x[1]
        arg[1,2]=(-0.424635655997)*x[0]+(0.0796757576821)*x[1]
        arg[1,3]=(-1.5899938122)*x[0]+(-0.178949548482)*x[1]
        arg[1,4]=(0.989317971495)*x[0]+(0.969711267468)*x[1]
        arg[2,0]=(0.384490947785)*x[0]+(1.42230193156)*x[1]
        arg[2,1]=(-0.0832227929255)*x[0]+(0.859020891429)*x[1]
        arg[2,2]=(-1.36129563124)*x[0]+(-0.240985031888)*x[1]
        arg[2,3]=(0.319706901063)*x[0]+(1.03419924566)*x[1]
        arg[2,4]=(0.986965249719)*x[0]+(0.228684031472)*x[1]
        arg[3,0]=(0.588520127337)*x[0]+(0.016799594822)*x[1]
        arg[3,1]=(1.46193438322)*x[0]+(0.431838900383)*x[1]
        arg[3,2]=(-0.0235512206818)*x[0]+(0.00227038037474)*x[1]
        arg[3,3]=(0.977383733076)*x[0]+(0.383979958766)*x[1]
        arg[3,4]=(0.447782383836)*x[0]+(-0.504218231709)*x[1]
        ref[0,0]=(-0.452635897444)*x_ref[0]+(-0.210513646405)*x_ref[1]
        ref[0,1]=(-0.590704436627)*x_ref[0]+(1.78000273884)*x_ref[1]
        ref[0,2]=(1.01331349402)*x_ref[0]+(1.76279765125)*x_ref[1]
        ref[0,3]=(1.5588852701)*x_ref[0]+(-0.550895238707)*x_ref[1]
        ref[0,4]=(-0.352453804527)*x_ref[0]+(-0.338832613421)*x_ref[1]
        ref[1,0]=(-0.809837877163)*x_ref[0]+(-0.531009602702)*x_ref[1]
        ref[1,1]=(0.513641503665)*x_ref[0]+(0.232114989082)*x_ref[1]
        ref[1,2]=(-0.424635655997)*x_ref[0]+(0.0796757576821)*x_ref[1]
        ref[1,3]=(-1.5899938122)*x_ref[0]+(-0.178949548482)*x_ref[1]
        ref[1,4]=(0.989317971495)*x_ref[0]+(0.969711267468)*x_ref[1]
        ref[2,0]=(0.384490947785)*x_ref[0]+(1.42230193156)*x_ref[1]
        ref[2,1]=(-0.0832227929255)*x_ref[0]+(0.859020891429)*x_ref[1]
        ref[2,2]=(-1.36129563124)*x_ref[0]+(-0.240985031888)*x_ref[1]
        ref[2,3]=(0.319706901063)*x_ref[0]+(1.03419924566)*x_ref[1]
        ref[2,4]=(0.986965249719)*x_ref[0]+(0.228684031472)*x_ref[1]
        ref[3,0]=(0.588520127337)*x_ref[0]+(0.016799594822)*x_ref[1]
        ref[3,1]=(1.46193438322)*x_ref[0]+(0.431838900383)*x_ref[1]
        ref[3,2]=(-0.0235512206818)*x_ref[0]+(0.00227038037474)*x_ref[1]
        ref[3,3]=(0.977383733076)*x_ref[0]+(0.383979958766)*x_ref[1]
        ref[3,4]=(0.447782383836)*x_ref[0]+(-0.504218231709)*x_ref[1]
      else:
        arg[0,0]=(-0.566459971675)*x[0]+(0.240246648893)*x[1]+(1.84693181923)*x[2]
        arg[0,1]=(1.09715015899)*x[0]+(1.26531924413)*x[1]+(0.627813347679)*x[2]
        arg[0,2]=(-1.17535073843)*x[0]+(-0.215601498941)*x[1]+(1.36346500888)*x[2]
        arg[0,3]=(-0.637207120161)*x[0]+(1.08542566952)*x[1]+(-1.13757514425)*x[2]
        arg[0,4]=(1.18247424069)*x[0]+(-1.48951614815)*x[1]+(0.615535039719)*x[2]
        arg[1,0]=(0.0937432100426)*x[0]+(-0.510326894156)*x[1]+(-0.456208996457)*x[2]
        arg[1,1]=(-0.866120342442)*x[0]+(0.927484984985)*x[1]+(0.918081524156)*x[2]
        arg[1,2]=(0.025238374235)*x[0]+(-0.182938599269)*x[1]+(-1.50048281322)*x[2]
        arg[1,3]=(-0.302334971294)*x[0]+(-0.925062799423)*x[1]+(-1.4675138953)*x[2]
        arg[1,4]=(-0.497738566447)*x[0]+(-0.324727071875)*x[1]+(-1.43341212055)*x[2]
        arg[2,0]=(-1.4605237958)*x[0]+(-0.332826007088)*x[1]+(-0.230483401111)*x[2]
        arg[2,1]=(0.801769248581)*x[0]+(-0.930588036503)*x[1]+(-1.06804808381)*x[2]
        arg[2,2]=(-1.07972165017)*x[0]+(1.09295362882)*x[1]+(0.048550098477)*x[2]
        arg[2,3]=(0.028290230337)*x[0]+(0.179461246635)*x[1]+(-1.09165257072)*x[2]
        arg[2,4]=(0.238850986174)*x[0]+(-0.837307206184)*x[1]+(-0.617254874594)*x[2]
        arg[3,0]=(-0.125921204582)*x[0]+(0.234758657879)*x[1]+(0.0847162247843)*x[2]
        arg[3,1]=(-0.213352414907)*x[0]+(0.206777352455)*x[1]+(-0.135357638131)*x[2]
        arg[3,2]=(-0.41897394434)*x[0]+(0.152706100016)*x[1]+(1.52390356012)*x[2]
        arg[3,3]=(0.1831705248)*x[0]+(-1.13556841364)*x[1]+(-1.7661652693)*x[2]
        arg[3,4]=(0.43473708388)*x[0]+(1.04496896213)*x[1]+(-0.816894739325)*x[2]
        ref[0,0]=(-0.566459971675)*x_ref[0]+(0.240246648893)*x_ref[1]+(1.84693181923)*x_ref[2]
        ref[0,1]=(1.09715015899)*x_ref[0]+(1.26531924413)*x_ref[1]+(0.627813347679)*x_ref[2]
        ref[0,2]=(-1.17535073843)*x_ref[0]+(-0.215601498941)*x_ref[1]+(1.36346500888)*x_ref[2]
        ref[0,3]=(-0.637207120161)*x_ref[0]+(1.08542566952)*x_ref[1]+(-1.13757514425)*x_ref[2]
        ref[0,4]=(1.18247424069)*x_ref[0]+(-1.48951614815)*x_ref[1]+(0.615535039719)*x_ref[2]
        ref[1,0]=(0.0937432100426)*x_ref[0]+(-0.510326894156)*x_ref[1]+(-0.456208996457)*x_ref[2]
        ref[1,1]=(-0.866120342442)*x_ref[0]+(0.927484984985)*x_ref[1]+(0.918081524156)*x_ref[2]
        ref[1,2]=(0.025238374235)*x_ref[0]+(-0.182938599269)*x_ref[1]+(-1.50048281322)*x_ref[2]
        ref[1,3]=(-0.302334971294)*x_ref[0]+(-0.925062799423)*x_ref[1]+(-1.4675138953)*x_ref[2]
        ref[1,4]=(-0.497738566447)*x_ref[0]+(-0.324727071875)*x_ref[1]+(-1.43341212055)*x_ref[2]
        ref[2,0]=(-1.4605237958)*x_ref[0]+(-0.332826007088)*x_ref[1]+(-0.230483401111)*x_ref[2]
        ref[2,1]=(0.801769248581)*x_ref[0]+(-0.930588036503)*x_ref[1]+(-1.06804808381)*x_ref[2]
        ref[2,2]=(-1.07972165017)*x_ref[0]+(1.09295362882)*x_ref[1]+(0.048550098477)*x_ref[2]
        ref[2,3]=(0.028290230337)*x_ref[0]+(0.179461246635)*x_ref[1]+(-1.09165257072)*x_ref[2]
        ref[2,4]=(0.238850986174)*x_ref[0]+(-0.837307206184)*x_ref[1]+(-0.617254874594)*x_ref[2]
        ref[3,0]=(-0.125921204582)*x_ref[0]+(0.234758657879)*x_ref[1]+(0.0847162247843)*x_ref[2]
        ref[3,1]=(-0.213352414907)*x_ref[0]+(0.206777352455)*x_ref[1]+(-0.135357638131)*x_ref[2]
        ref[3,2]=(-0.41897394434)*x_ref[0]+(0.152706100016)*x_ref[1]+(1.52390356012)*x_ref[2]
        ref[3,3]=(0.1831705248)*x_ref[0]+(-1.13556841364)*x_ref[1]+(-1.7661652693)*x_ref[2]
        ref[3,4]=(0.43473708388)*x_ref[0]+(1.04496896213)*x_ref[1]+(-0.816894739325)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.324237136046)*x[0]+(0.67172278334)*x[1]
        arg[0,0,1]=(-1.07360258996)*x[0]+(-0.646208771542)*x[1]
        arg[0,1,0]=(-0.040139403311)*x[0]+(-0.281626553874)*x[1]
        arg[0,1,1]=(0.963343050385)*x[0]+(-0.724838290038)*x[1]
        arg[1,0,0]=(-0.909697586367)*x[0]+(0.48041903189)*x[1]
        arg[1,0,1]=(0.414758676032)*x[0]+(-0.0921084575218)*x[1]
        arg[1,1,0]=(1.06852211241)*x[0]+(-1.01026709886)*x[1]
        arg[1,1,1]=(-0.675258695211)*x[0]+(0.51361078905)*x[1]
        arg[2,0,0]=(-0.791246198786)*x[0]+(-0.0283034425042)*x[1]
        arg[2,0,1]=(-0.87756010848)*x[0]+(-0.00868464377539)*x[1]
        arg[2,1,0]=(0.44209070344)*x[0]+(0.0846344294002)*x[1]
        arg[2,1,1]=(0.777287404475)*x[0]+(0.72614634535)*x[1]
        arg[3,0,0]=(-0.363354104693)*x[0]+(-1.22959837449)*x[1]
        arg[3,0,1]=(-0.532220757425)*x[0]+(-1.54990170304)*x[1]
        arg[3,1,0]=(-0.491099325731)*x[0]+(-0.412402332731)*x[1]
        arg[3,1,1]=(0.0750673227925)*x[0]+(-0.725108664028)*x[1]
        arg[4,0,0]=(-0.419689899608)*x[0]+(-1.32423898043)*x[1]
        arg[4,0,1]=(0.532682204197)*x[0]+(0.0703779631797)*x[1]
        arg[4,1,0]=(1.68643150913)*x[0]+(0.105443733595)*x[1]
        arg[4,1,1]=(0.936329701936)*x[0]+(1.04067808474)*x[1]
        arg[5,0,0]=(1.37285560133)*x[0]+(-0.680844454631)*x[1]
        arg[5,0,1]=(0.843677131436)*x[0]+(0.272365035967)*x[1]
        arg[5,1,0]=(0.00586472074043)*x[0]+(-1.09785622721)*x[1]
        arg[5,1,1]=(1.1410854961)*x[0]+(-0.591468439837)*x[1]
        ref[0,0,0]=(0.324237136046)*x_ref[0]+(0.67172278334)*x_ref[1]
        ref[0,0,1]=(-1.07360258996)*x_ref[0]+(-0.646208771542)*x_ref[1]
        ref[0,1,0]=(-0.040139403311)*x_ref[0]+(-0.281626553874)*x_ref[1]
        ref[0,1,1]=(0.963343050385)*x_ref[0]+(-0.724838290038)*x_ref[1]
        ref[1,0,0]=(-0.909697586367)*x_ref[0]+(0.48041903189)*x_ref[1]
        ref[1,0,1]=(0.414758676032)*x_ref[0]+(-0.0921084575218)*x_ref[1]
        ref[1,1,0]=(1.06852211241)*x_ref[0]+(-1.01026709886)*x_ref[1]
        ref[1,1,1]=(-0.675258695211)*x_ref[0]+(0.51361078905)*x_ref[1]
        ref[2,0,0]=(-0.791246198786)*x_ref[0]+(-0.0283034425042)*x_ref[1]
        ref[2,0,1]=(-0.87756010848)*x_ref[0]+(-0.00868464377539)*x_ref[1]
        ref[2,1,0]=(0.44209070344)*x_ref[0]+(0.0846344294002)*x_ref[1]
        ref[2,1,1]=(0.777287404475)*x_ref[0]+(0.72614634535)*x_ref[1]
        ref[3,0,0]=(-0.363354104693)*x_ref[0]+(-1.22959837449)*x_ref[1]
        ref[3,0,1]=(-0.532220757425)*x_ref[0]+(-1.54990170304)*x_ref[1]
        ref[3,1,0]=(-0.491099325731)*x_ref[0]+(-0.412402332731)*x_ref[1]
        ref[3,1,1]=(0.0750673227925)*x_ref[0]+(-0.725108664028)*x_ref[1]
        ref[4,0,0]=(-0.419689899608)*x_ref[0]+(-1.32423898043)*x_ref[1]
        ref[4,0,1]=(0.532682204197)*x_ref[0]+(0.0703779631797)*x_ref[1]
        ref[4,1,0]=(1.68643150913)*x_ref[0]+(0.105443733595)*x_ref[1]
        ref[4,1,1]=(0.936329701936)*x_ref[0]+(1.04067808474)*x_ref[1]
        ref[5,0,0]=(1.37285560133)*x_ref[0]+(-0.680844454631)*x_ref[1]
        ref[5,0,1]=(0.843677131436)*x_ref[0]+(0.272365035967)*x_ref[1]
        ref[5,1,0]=(0.00586472074043)*x_ref[0]+(-1.09785622721)*x_ref[1]
        ref[5,1,1]=(1.1410854961)*x_ref[0]+(-0.591468439837)*x_ref[1]
      else:
        arg[0,0,0]=(-0.289427308366)*x[0]+(1.17077011703)*x[1]+(0.763196501047)*x[2]
        arg[0,0,1]=(0.642730500265)*x[0]+(1.1719001632)*x[1]+(-0.121378040894)*x[2]
        arg[0,1,0]=(0.0879005621769)*x[0]+(-0.627582537594)*x[1]+(0.220072369447)*x[2]
        arg[0,1,1]=(-0.518762071566)*x[0]+(-0.700754417623)*x[1]+(-0.340623426752)*x[2]
        arg[1,0,0]=(-0.162635708202)*x[0]+(-0.996922517356)*x[1]+(0.48000598379)*x[2]
        arg[1,0,1]=(-0.49910792923)*x[0]+(1.11334748765)*x[1]+(0.0050657496159)*x[2]
        arg[1,1,0]=(-1.55972927502)*x[0]+(0.233936234238)*x[1]+(-0.546906209213)*x[2]
        arg[1,1,1]=(-0.6012330303)*x[0]+(-0.267951301884)*x[1]+(-0.193021060351)*x[2]
        arg[2,0,0]=(1.32515959423)*x[0]+(1.2070735022)*x[1]+(1.11474945521)*x[2]
        arg[2,0,1]=(-1.05500093406)*x[0]+(-0.937935240268)*x[1]+(0.180289797658)*x[2]
        arg[2,1,0]=(-0.450015559749)*x[0]+(-1.17715272891)*x[1]+(0.197160834934)*x[2]
        arg[2,1,1]=(-1.24969258786)*x[0]+(-0.300915973735)*x[1]+(-0.195058423793)*x[2]
        arg[3,0,0]=(0.704171975525)*x[0]+(1.0327384159)*x[1]+(-1.23664355084)*x[2]
        arg[3,0,1]=(0.916288412006)*x[0]+(1.6929893349)*x[1]+(-1.16682244827)*x[2]
        arg[3,1,0]=(-0.756147901025)*x[0]+(0.0232005287342)*x[1]+(0.213796289692)*x[2]
        arg[3,1,1]=(0.92799443461)*x[0]+(-0.228093116826)*x[1]+(-1.53373185211)*x[2]
        arg[4,0,0]=(-0.272514671459)*x[0]+(0.310026631492)*x[1]+(0.735127519475)*x[2]
        arg[4,0,1]=(1.23963120106)*x[0]+(-0.0769223754537)*x[1]+(0.767002455909)*x[2]
        arg[4,1,0]=(-1.1121433063)*x[0]+(0.748482838233)*x[1]+(-0.0985898502473)*x[2]
        arg[4,1,1]=(-1.08169600038)*x[0]+(-1.39575461699)*x[1]+(0.468455733883)*x[2]
        arg[5,0,0]=(-0.737100332594)*x[0]+(1.0950941569)*x[1]+(0.530144605873)*x[2]
        arg[5,0,1]=(1.94353080429)*x[0]+(1.61187736568)*x[1]+(0.887255997784)*x[2]
        arg[5,1,0]=(-0.492812782764)*x[0]+(-1.17859119673)*x[1]+(-0.627254028293)*x[2]
        arg[5,1,1]=(0.932565304399)*x[0]+(0.760125636669)*x[1]+(-0.687485104891)*x[2]
        ref[0,0,0]=(-0.289427308366)*x_ref[0]+(1.17077011703)*x_ref[1]+(0.763196501047)*x_ref[2]
        ref[0,0,1]=(0.642730500265)*x_ref[0]+(1.1719001632)*x_ref[1]+(-0.121378040894)*x_ref[2]
        ref[0,1,0]=(0.0879005621769)*x_ref[0]+(-0.627582537594)*x_ref[1]+(0.220072369447)*x_ref[2]
        ref[0,1,1]=(-0.518762071566)*x_ref[0]+(-0.700754417623)*x_ref[1]+(-0.340623426752)*x_ref[2]
        ref[1,0,0]=(-0.162635708202)*x_ref[0]+(-0.996922517356)*x_ref[1]+(0.48000598379)*x_ref[2]
        ref[1,0,1]=(-0.49910792923)*x_ref[0]+(1.11334748765)*x_ref[1]+(0.0050657496159)*x_ref[2]
        ref[1,1,0]=(-1.55972927502)*x_ref[0]+(0.233936234238)*x_ref[1]+(-0.546906209213)*x_ref[2]
        ref[1,1,1]=(-0.6012330303)*x_ref[0]+(-0.267951301884)*x_ref[1]+(-0.193021060351)*x_ref[2]
        ref[2,0,0]=(1.32515959423)*x_ref[0]+(1.2070735022)*x_ref[1]+(1.11474945521)*x_ref[2]
        ref[2,0,1]=(-1.05500093406)*x_ref[0]+(-0.937935240268)*x_ref[1]+(0.180289797658)*x_ref[2]
        ref[2,1,0]=(-0.450015559749)*x_ref[0]+(-1.17715272891)*x_ref[1]+(0.197160834934)*x_ref[2]
        ref[2,1,1]=(-1.24969258786)*x_ref[0]+(-0.300915973735)*x_ref[1]+(-0.195058423793)*x_ref[2]
        ref[3,0,0]=(0.704171975525)*x_ref[0]+(1.0327384159)*x_ref[1]+(-1.23664355084)*x_ref[2]
        ref[3,0,1]=(0.916288412006)*x_ref[0]+(1.6929893349)*x_ref[1]+(-1.16682244827)*x_ref[2]
        ref[3,1,0]=(-0.756147901025)*x_ref[0]+(0.0232005287342)*x_ref[1]+(0.213796289692)*x_ref[2]
        ref[3,1,1]=(0.92799443461)*x_ref[0]+(-0.228093116826)*x_ref[1]+(-1.53373185211)*x_ref[2]
        ref[4,0,0]=(-0.272514671459)*x_ref[0]+(0.310026631492)*x_ref[1]+(0.735127519475)*x_ref[2]
        ref[4,0,1]=(1.23963120106)*x_ref[0]+(-0.0769223754537)*x_ref[1]+(0.767002455909)*x_ref[2]
        ref[4,1,0]=(-1.1121433063)*x_ref[0]+(0.748482838233)*x_ref[1]+(-0.0985898502473)*x_ref[2]
        ref[4,1,1]=(-1.08169600038)*x_ref[0]+(-1.39575461699)*x_ref[1]+(0.468455733883)*x_ref[2]
        ref[5,0,0]=(-0.737100332594)*x_ref[0]+(1.0950941569)*x_ref[1]+(0.530144605873)*x_ref[2]
        ref[5,0,1]=(1.94353080429)*x_ref[0]+(1.61187736568)*x_ref[1]+(0.887255997784)*x_ref[2]
        ref[5,1,0]=(-0.492812782764)*x_ref[0]+(-1.17859119673)*x_ref[1]+(-0.627254028293)*x_ref[2]
        ref[5,1,1]=(0.932565304399)*x_ref[0]+(0.760125636669)*x_ref[1]+(-0.687485104891)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.261713255264)*x[0]+(-0.650225996431)*x[1]
        arg[0,0,0,1]=(0.177056727607)*x[0]+(-0.486487540393)*x[1]
        arg[0,0,1,0]=(1.38050234107)*x[0]+(0.273417897264)*x[1]
        arg[0,0,1,1]=(-1.059824929)*x[0]+(0.432893707387)*x[1]
        arg[0,0,2,0]=(1.11328992978)*x[0]+(1.14949237037)*x[1]
        arg[0,0,2,1]=(-0.176459646227)*x[0]+(-0.835980248375)*x[1]
        arg[0,1,0,0]=(0.992757237392)*x[0]+(0.113511598452)*x[1]
        arg[0,1,0,1]=(-0.211425610821)*x[0]+(-0.472765285276)*x[1]
        arg[0,1,1,0]=(-1.52498305232)*x[0]+(-0.869737083114)*x[1]
        arg[0,1,1,1]=(-0.474203117317)*x[0]+(-0.284532442353)*x[1]
        arg[0,1,2,0]=(-0.431542961546)*x[0]+(-0.0427397090453)*x[1]
        arg[0,1,2,1]=(-1.0562625269)*x[0]+(-0.74486767124)*x[1]
        arg[0,2,0,0]=(1.46553867064)*x[0]+(-0.650182195657)*x[1]
        arg[0,2,0,1]=(-0.717482663647)*x[0]+(-0.716320000777)*x[1]
        arg[0,2,1,0]=(-1.21701970743)*x[0]+(0.636901443761)*x[1]
        arg[0,2,1,1]=(0.438592928507)*x[0]+(0.812362899723)*x[1]
        arg[0,2,2,0]=(0.104885477141)*x[0]+(-0.127968978832)*x[1]
        arg[0,2,2,1]=(0.285116712849)*x[0]+(0.169740783305)*x[1]
        arg[0,3,0,0]=(-1.22825142912)*x[0]+(-0.562911687838)*x[1]
        arg[0,3,0,1]=(-0.569159534025)*x[0]+(1.30898732707)*x[1]
        arg[0,3,1,0]=(0.164709387569)*x[0]+(-1.62733582919)*x[1]
        arg[0,3,1,1]=(-0.477197320128)*x[0]+(-0.115207237191)*x[1]
        arg[0,3,2,0]=(-0.046871003317)*x[0]+(0.0371149345252)*x[1]
        arg[0,3,2,1]=(0.839624830979)*x[0]+(0.0785937909827)*x[1]
        arg[0,4,0,0]=(-0.0250725769598)*x[0]+(-0.535781603767)*x[1]
        arg[0,4,0,1]=(-0.761849740443)*x[0]+(-0.300714501152)*x[1]
        arg[0,4,1,0]=(0.0212637838833)*x[0]+(-1.56914428088)*x[1]
        arg[0,4,1,1]=(-0.254524677425)*x[0]+(0.699288623343)*x[1]
        arg[0,4,2,0]=(1.14678964272)*x[0]+(1.24738649379)*x[1]
        arg[0,4,2,1]=(1.80545916472)*x[0]+(1.22689087643)*x[1]
        arg[1,0,0,0]=(-0.116129927341)*x[0]+(-0.875165834648)*x[1]
        arg[1,0,0,1]=(-0.49905851236)*x[0]+(-1.10711214732)*x[1]
        arg[1,0,1,0]=(-1.44706167224)*x[0]+(-1.80259802262)*x[1]
        arg[1,0,1,1]=(0.555288273554)*x[0]+(0.153437302525)*x[1]
        arg[1,0,2,0]=(0.00774838569142)*x[0]+(-1.35683757658)*x[1]
        arg[1,0,2,1]=(-0.625156986286)*x[0]+(-0.201091038824)*x[1]
        arg[1,1,0,0]=(-0.765068445971)*x[0]+(1.32890117415)*x[1]
        arg[1,1,0,1]=(-0.629548919312)*x[0]+(-0.0512345513979)*x[1]
        arg[1,1,1,0]=(-0.361717423775)*x[0]+(0.286169858965)*x[1]
        arg[1,1,1,1]=(-0.533065922681)*x[0]+(0.103362397133)*x[1]
        arg[1,1,2,0]=(0.978041972715)*x[0]+(0.405949452109)*x[1]
        arg[1,1,2,1]=(-1.08418078618)*x[0]+(0.027879770152)*x[1]
        arg[1,2,0,0]=(1.23545256768)*x[0]+(0.574400116282)*x[1]
        arg[1,2,0,1]=(0.245362478488)*x[0]+(0.241577755816)*x[1]
        arg[1,2,1,0]=(0.645309443244)*x[0]+(0.0437773509036)*x[1]
        arg[1,2,1,1]=(0.00930197364105)*x[0]+(0.472462788946)*x[1]
        arg[1,2,2,0]=(1.44241534167)*x[0]+(-0.901332994844)*x[1]
        arg[1,2,2,1]=(0.893335284585)*x[0]+(0.109130146628)*x[1]
        arg[1,3,0,0]=(0.601907323675)*x[0]+(-1.54057502586)*x[1]
        arg[1,3,0,1]=(1.38086759376)*x[0]+(0.356522488945)*x[1]
        arg[1,3,1,0]=(1.16786130652)*x[0]+(-0.788756049484)*x[1]
        arg[1,3,1,1]=(-0.22580270698)*x[0]+(0.857686971693)*x[1]
        arg[1,3,2,0]=(0.709549396131)*x[0]+(-0.644330864877)*x[1]
        arg[1,3,2,1]=(1.04901965211)*x[0]+(0.0178384468958)*x[1]
        arg[1,4,0,0]=(-0.525050731412)*x[0]+(-0.253639122312)*x[1]
        arg[1,4,0,1]=(-1.63200911423)*x[0]+(0.519020443372)*x[1]
        arg[1,4,1,0]=(0.23956666878)*x[0]+(-1.76973024011)*x[1]
        arg[1,4,1,1]=(-0.412317050346)*x[0]+(-0.391711749094)*x[1]
        arg[1,4,2,0]=(1.63293176518)*x[0]+(0.583921572851)*x[1]
        arg[1,4,2,1]=(0.434893659373)*x[0]+(-0.324706930043)*x[1]
        arg[2,0,0,0]=(0.368478865639)*x[0]+(1.87637158068)*x[1]
        arg[2,0,0,1]=(1.32519020398)*x[0]+(0.306016568948)*x[1]
        arg[2,0,1,0]=(-0.284264003593)*x[0]+(-0.899022593367)*x[1]
        arg[2,0,1,1]=(-0.0439761413236)*x[0]+(-0.696386615358)*x[1]
        arg[2,0,2,0]=(0.407393812349)*x[0]+(-1.04815464317)*x[1]
        arg[2,0,2,1]=(-0.956661282071)*x[0]+(0.243281223782)*x[1]
        arg[2,1,0,0]=(-1.29201911939)*x[0]+(-0.693049706038)*x[1]
        arg[2,1,0,1]=(-0.711059848853)*x[0]+(-1.14204907272)*x[1]
        arg[2,1,1,0]=(-0.347693333057)*x[0]+(0.818879269672)*x[1]
        arg[2,1,1,1]=(0.00742807299875)*x[0]+(0.138845309984)*x[1]
        arg[2,1,2,0]=(1.13412244048)*x[0]+(-0.481422214102)*x[1]
        arg[2,1,2,1]=(0.64440302826)*x[0]+(-0.955231893089)*x[1]
        arg[2,2,0,0]=(0.13123141544)*x[0]+(-0.602266537541)*x[1]
        arg[2,2,0,1]=(-0.204818071763)*x[0]+(-1.13480248293)*x[1]
        arg[2,2,1,0]=(0.781058287579)*x[0]+(-1.31936224435)*x[1]
        arg[2,2,1,1]=(-0.591275351662)*x[0]+(0.691816969517)*x[1]
        arg[2,2,2,0]=(-0.921964826241)*x[0]+(1.28687075541)*x[1]
        arg[2,2,2,1]=(0.246260462337)*x[0]+(-1.04699282194)*x[1]
        arg[2,3,0,0]=(0.264209126513)*x[0]+(-0.697094400566)*x[1]
        arg[2,3,0,1]=(0.456158298484)*x[0]+(0.0288036445565)*x[1]
        arg[2,3,1,0]=(-0.720337925392)*x[0]+(0.0965182642186)*x[1]
        arg[2,3,1,1]=(0.397480419547)*x[0]+(-1.55232798344)*x[1]
        arg[2,3,2,0]=(-0.337139999283)*x[0]+(0.918809751807)*x[1]
        arg[2,3,2,1]=(1.34009447198)*x[0]+(-1.41390962018)*x[1]
        arg[2,4,0,0]=(1.67781838006)*x[0]+(0.521276193814)*x[1]
        arg[2,4,0,1]=(-0.114014321939)*x[0]+(0.529316778773)*x[1]
        arg[2,4,1,0]=(-0.19938131807)*x[0]+(-0.0712417614843)*x[1]
        arg[2,4,1,1]=(-0.724676753871)*x[0]+(0.731341170171)*x[1]
        arg[2,4,2,0]=(0.988755659656)*x[0]+(-1.0327352189)*x[1]
        arg[2,4,2,1]=(-0.526286732859)*x[0]+(-1.18295671598)*x[1]
        arg[3,0,0,0]=(-0.685223304723)*x[0]+(0.508431586068)*x[1]
        arg[3,0,0,1]=(0.529240274944)*x[0]+(0.77519128969)*x[1]
        arg[3,0,1,0]=(-0.757098431928)*x[0]+(1.41375812725)*x[1]
        arg[3,0,1,1]=(1.20805599335)*x[0]+(-0.184660334182)*x[1]
        arg[3,0,2,0]=(0.718413765561)*x[0]+(0.971149168706)*x[1]
        arg[3,0,2,1]=(-0.632670534518)*x[0]+(0.0389618967363)*x[1]
        arg[3,1,0,0]=(-0.525814673755)*x[0]+(0.574122587598)*x[1]
        arg[3,1,0,1]=(-1.04117288975)*x[0]+(-1.00738089823)*x[1]
        arg[3,1,1,0]=(0.0312806064182)*x[0]+(-0.30773511847)*x[1]
        arg[3,1,1,1]=(0.889930889355)*x[0]+(0.833689153489)*x[1]
        arg[3,1,2,0]=(-1.21586293381)*x[0]+(0.512182349574)*x[1]
        arg[3,1,2,1]=(-0.00437508755415)*x[0]+(-1.13228664497)*x[1]
        arg[3,2,0,0]=(1.64174174554)*x[0]+(-0.696500062562)*x[1]
        arg[3,2,0,1]=(0.0265343210604)*x[0]+(-0.707378716636)*x[1]
        arg[3,2,1,0]=(0.952922122967)*x[0]+(0.0946904561805)*x[1]
        arg[3,2,1,1]=(-1.06905266456)*x[0]+(-0.155122789807)*x[1]
        arg[3,2,2,0]=(-1.14410789976)*x[0]+(-0.379020164446)*x[1]
        arg[3,2,2,1]=(-0.0922611198553)*x[0]+(1.36680699099)*x[1]
        arg[3,3,0,0]=(0.849028205422)*x[0]+(-0.968698293811)*x[1]
        arg[3,3,0,1]=(0.649894718675)*x[0]+(0.238249259292)*x[1]
        arg[3,3,1,0]=(1.28170508471)*x[0]+(-0.736544947537)*x[1]
        arg[3,3,1,1]=(-0.188411685414)*x[0]+(1.14145010691)*x[1]
        arg[3,3,2,0]=(-1.13973037849)*x[0]+(-0.256565914821)*x[1]
        arg[3,3,2,1]=(0.955478585358)*x[0]+(-0.83417326515)*x[1]
        arg[3,4,0,0]=(0.155710789468)*x[0]+(0.204956811648)*x[1]
        arg[3,4,0,1]=(0.567866952974)*x[0]+(-0.168398320099)*x[1]
        arg[3,4,1,0]=(-1.36997367)*x[0]+(0.141984040172)*x[1]
        arg[3,4,1,1]=(-0.809532285206)*x[0]+(0.70611485405)*x[1]
        arg[3,4,2,0]=(-0.453260232051)*x[0]+(0.0665127865858)*x[1]
        arg[3,4,2,1]=(0.00142021115277)*x[0]+(1.02307505343)*x[1]
        ref[0,0,0,0]=(-0.261713255264)*x_ref[0]+(-0.650225996431)*x_ref[1]
        ref[0,0,0,1]=(0.177056727607)*x_ref[0]+(-0.486487540393)*x_ref[1]
        ref[0,0,1,0]=(1.38050234107)*x_ref[0]+(0.273417897264)*x_ref[1]
        ref[0,0,1,1]=(-1.059824929)*x_ref[0]+(0.432893707387)*x_ref[1]
        ref[0,0,2,0]=(1.11328992978)*x_ref[0]+(1.14949237037)*x_ref[1]
        ref[0,0,2,1]=(-0.176459646227)*x_ref[0]+(-0.835980248375)*x_ref[1]
        ref[0,1,0,0]=(0.992757237392)*x_ref[0]+(0.113511598452)*x_ref[1]
        ref[0,1,0,1]=(-0.211425610821)*x_ref[0]+(-0.472765285276)*x_ref[1]
        ref[0,1,1,0]=(-1.52498305232)*x_ref[0]+(-0.869737083114)*x_ref[1]
        ref[0,1,1,1]=(-0.474203117317)*x_ref[0]+(-0.284532442353)*x_ref[1]
        ref[0,1,2,0]=(-0.431542961546)*x_ref[0]+(-0.0427397090453)*x_ref[1]
        ref[0,1,2,1]=(-1.0562625269)*x_ref[0]+(-0.74486767124)*x_ref[1]
        ref[0,2,0,0]=(1.46553867064)*x_ref[0]+(-0.650182195657)*x_ref[1]
        ref[0,2,0,1]=(-0.717482663647)*x_ref[0]+(-0.716320000777)*x_ref[1]
        ref[0,2,1,0]=(-1.21701970743)*x_ref[0]+(0.636901443761)*x_ref[1]
        ref[0,2,1,1]=(0.438592928507)*x_ref[0]+(0.812362899723)*x_ref[1]
        ref[0,2,2,0]=(0.104885477141)*x_ref[0]+(-0.127968978832)*x_ref[1]
        ref[0,2,2,1]=(0.285116712849)*x_ref[0]+(0.169740783305)*x_ref[1]
        ref[0,3,0,0]=(-1.22825142912)*x_ref[0]+(-0.562911687838)*x_ref[1]
        ref[0,3,0,1]=(-0.569159534025)*x_ref[0]+(1.30898732707)*x_ref[1]
        ref[0,3,1,0]=(0.164709387569)*x_ref[0]+(-1.62733582919)*x_ref[1]
        ref[0,3,1,1]=(-0.477197320128)*x_ref[0]+(-0.115207237191)*x_ref[1]
        ref[0,3,2,0]=(-0.046871003317)*x_ref[0]+(0.0371149345252)*x_ref[1]
        ref[0,3,2,1]=(0.839624830979)*x_ref[0]+(0.0785937909827)*x_ref[1]
        ref[0,4,0,0]=(-0.0250725769598)*x_ref[0]+(-0.535781603767)*x_ref[1]
        ref[0,4,0,1]=(-0.761849740443)*x_ref[0]+(-0.300714501152)*x_ref[1]
        ref[0,4,1,0]=(0.0212637838833)*x_ref[0]+(-1.56914428088)*x_ref[1]
        ref[0,4,1,1]=(-0.254524677425)*x_ref[0]+(0.699288623343)*x_ref[1]
        ref[0,4,2,0]=(1.14678964272)*x_ref[0]+(1.24738649379)*x_ref[1]
        ref[0,4,2,1]=(1.80545916472)*x_ref[0]+(1.22689087643)*x_ref[1]
        ref[1,0,0,0]=(-0.116129927341)*x_ref[0]+(-0.875165834648)*x_ref[1]
        ref[1,0,0,1]=(-0.49905851236)*x_ref[0]+(-1.10711214732)*x_ref[1]
        ref[1,0,1,0]=(-1.44706167224)*x_ref[0]+(-1.80259802262)*x_ref[1]
        ref[1,0,1,1]=(0.555288273554)*x_ref[0]+(0.153437302525)*x_ref[1]
        ref[1,0,2,0]=(0.00774838569142)*x_ref[0]+(-1.35683757658)*x_ref[1]
        ref[1,0,2,1]=(-0.625156986286)*x_ref[0]+(-0.201091038824)*x_ref[1]
        ref[1,1,0,0]=(-0.765068445971)*x_ref[0]+(1.32890117415)*x_ref[1]
        ref[1,1,0,1]=(-0.629548919312)*x_ref[0]+(-0.0512345513979)*x_ref[1]
        ref[1,1,1,0]=(-0.361717423775)*x_ref[0]+(0.286169858965)*x_ref[1]
        ref[1,1,1,1]=(-0.533065922681)*x_ref[0]+(0.103362397133)*x_ref[1]
        ref[1,1,2,0]=(0.978041972715)*x_ref[0]+(0.405949452109)*x_ref[1]
        ref[1,1,2,1]=(-1.08418078618)*x_ref[0]+(0.027879770152)*x_ref[1]
        ref[1,2,0,0]=(1.23545256768)*x_ref[0]+(0.574400116282)*x_ref[1]
        ref[1,2,0,1]=(0.245362478488)*x_ref[0]+(0.241577755816)*x_ref[1]
        ref[1,2,1,0]=(0.645309443244)*x_ref[0]+(0.0437773509036)*x_ref[1]
        ref[1,2,1,1]=(0.00930197364105)*x_ref[0]+(0.472462788946)*x_ref[1]
        ref[1,2,2,0]=(1.44241534167)*x_ref[0]+(-0.901332994844)*x_ref[1]
        ref[1,2,2,1]=(0.893335284585)*x_ref[0]+(0.109130146628)*x_ref[1]
        ref[1,3,0,0]=(0.601907323675)*x_ref[0]+(-1.54057502586)*x_ref[1]
        ref[1,3,0,1]=(1.38086759376)*x_ref[0]+(0.356522488945)*x_ref[1]
        ref[1,3,1,0]=(1.16786130652)*x_ref[0]+(-0.788756049484)*x_ref[1]
        ref[1,3,1,1]=(-0.22580270698)*x_ref[0]+(0.857686971693)*x_ref[1]
        ref[1,3,2,0]=(0.709549396131)*x_ref[0]+(-0.644330864877)*x_ref[1]
        ref[1,3,2,1]=(1.04901965211)*x_ref[0]+(0.0178384468958)*x_ref[1]
        ref[1,4,0,0]=(-0.525050731412)*x_ref[0]+(-0.253639122312)*x_ref[1]
        ref[1,4,0,1]=(-1.63200911423)*x_ref[0]+(0.519020443372)*x_ref[1]
        ref[1,4,1,0]=(0.23956666878)*x_ref[0]+(-1.76973024011)*x_ref[1]
        ref[1,4,1,1]=(-0.412317050346)*x_ref[0]+(-0.391711749094)*x_ref[1]
        ref[1,4,2,0]=(1.63293176518)*x_ref[0]+(0.583921572851)*x_ref[1]
        ref[1,4,2,1]=(0.434893659373)*x_ref[0]+(-0.324706930043)*x_ref[1]
        ref[2,0,0,0]=(0.368478865639)*x_ref[0]+(1.87637158068)*x_ref[1]
        ref[2,0,0,1]=(1.32519020398)*x_ref[0]+(0.306016568948)*x_ref[1]
        ref[2,0,1,0]=(-0.284264003593)*x_ref[0]+(-0.899022593367)*x_ref[1]
        ref[2,0,1,1]=(-0.0439761413236)*x_ref[0]+(-0.696386615358)*x_ref[1]
        ref[2,0,2,0]=(0.407393812349)*x_ref[0]+(-1.04815464317)*x_ref[1]
        ref[2,0,2,1]=(-0.956661282071)*x_ref[0]+(0.243281223782)*x_ref[1]
        ref[2,1,0,0]=(-1.29201911939)*x_ref[0]+(-0.693049706038)*x_ref[1]
        ref[2,1,0,1]=(-0.711059848853)*x_ref[0]+(-1.14204907272)*x_ref[1]
        ref[2,1,1,0]=(-0.347693333057)*x_ref[0]+(0.818879269672)*x_ref[1]
        ref[2,1,1,1]=(0.00742807299875)*x_ref[0]+(0.138845309984)*x_ref[1]
        ref[2,1,2,0]=(1.13412244048)*x_ref[0]+(-0.481422214102)*x_ref[1]
        ref[2,1,2,1]=(0.64440302826)*x_ref[0]+(-0.955231893089)*x_ref[1]
        ref[2,2,0,0]=(0.13123141544)*x_ref[0]+(-0.602266537541)*x_ref[1]
        ref[2,2,0,1]=(-0.204818071763)*x_ref[0]+(-1.13480248293)*x_ref[1]
        ref[2,2,1,0]=(0.781058287579)*x_ref[0]+(-1.31936224435)*x_ref[1]
        ref[2,2,1,1]=(-0.591275351662)*x_ref[0]+(0.691816969517)*x_ref[1]
        ref[2,2,2,0]=(-0.921964826241)*x_ref[0]+(1.28687075541)*x_ref[1]
        ref[2,2,2,1]=(0.246260462337)*x_ref[0]+(-1.04699282194)*x_ref[1]
        ref[2,3,0,0]=(0.264209126513)*x_ref[0]+(-0.697094400566)*x_ref[1]
        ref[2,3,0,1]=(0.456158298484)*x_ref[0]+(0.0288036445565)*x_ref[1]
        ref[2,3,1,0]=(-0.720337925392)*x_ref[0]+(0.0965182642186)*x_ref[1]
        ref[2,3,1,1]=(0.397480419547)*x_ref[0]+(-1.55232798344)*x_ref[1]
        ref[2,3,2,0]=(-0.337139999283)*x_ref[0]+(0.918809751807)*x_ref[1]
        ref[2,3,2,1]=(1.34009447198)*x_ref[0]+(-1.41390962018)*x_ref[1]
        ref[2,4,0,0]=(1.67781838006)*x_ref[0]+(0.521276193814)*x_ref[1]
        ref[2,4,0,1]=(-0.114014321939)*x_ref[0]+(0.529316778773)*x_ref[1]
        ref[2,4,1,0]=(-0.19938131807)*x_ref[0]+(-0.0712417614843)*x_ref[1]
        ref[2,4,1,1]=(-0.724676753871)*x_ref[0]+(0.731341170171)*x_ref[1]
        ref[2,4,2,0]=(0.988755659656)*x_ref[0]+(-1.0327352189)*x_ref[1]
        ref[2,4,2,1]=(-0.526286732859)*x_ref[0]+(-1.18295671598)*x_ref[1]
        ref[3,0,0,0]=(-0.685223304723)*x_ref[0]+(0.508431586068)*x_ref[1]
        ref[3,0,0,1]=(0.529240274944)*x_ref[0]+(0.77519128969)*x_ref[1]
        ref[3,0,1,0]=(-0.757098431928)*x_ref[0]+(1.41375812725)*x_ref[1]
        ref[3,0,1,1]=(1.20805599335)*x_ref[0]+(-0.184660334182)*x_ref[1]
        ref[3,0,2,0]=(0.718413765561)*x_ref[0]+(0.971149168706)*x_ref[1]
        ref[3,0,2,1]=(-0.632670534518)*x_ref[0]+(0.0389618967363)*x_ref[1]
        ref[3,1,0,0]=(-0.525814673755)*x_ref[0]+(0.574122587598)*x_ref[1]
        ref[3,1,0,1]=(-1.04117288975)*x_ref[0]+(-1.00738089823)*x_ref[1]
        ref[3,1,1,0]=(0.0312806064182)*x_ref[0]+(-0.30773511847)*x_ref[1]
        ref[3,1,1,1]=(0.889930889355)*x_ref[0]+(0.833689153489)*x_ref[1]
        ref[3,1,2,0]=(-1.21586293381)*x_ref[0]+(0.512182349574)*x_ref[1]
        ref[3,1,2,1]=(-0.00437508755415)*x_ref[0]+(-1.13228664497)*x_ref[1]
        ref[3,2,0,0]=(1.64174174554)*x_ref[0]+(-0.696500062562)*x_ref[1]
        ref[3,2,0,1]=(0.0265343210604)*x_ref[0]+(-0.707378716636)*x_ref[1]
        ref[3,2,1,0]=(0.952922122967)*x_ref[0]+(0.0946904561805)*x_ref[1]
        ref[3,2,1,1]=(-1.06905266456)*x_ref[0]+(-0.155122789807)*x_ref[1]
        ref[3,2,2,0]=(-1.14410789976)*x_ref[0]+(-0.379020164446)*x_ref[1]
        ref[3,2,2,1]=(-0.0922611198553)*x_ref[0]+(1.36680699099)*x_ref[1]
        ref[3,3,0,0]=(0.849028205422)*x_ref[0]+(-0.968698293811)*x_ref[1]
        ref[3,3,0,1]=(0.649894718675)*x_ref[0]+(0.238249259292)*x_ref[1]
        ref[3,3,1,0]=(1.28170508471)*x_ref[0]+(-0.736544947537)*x_ref[1]
        ref[3,3,1,1]=(-0.188411685414)*x_ref[0]+(1.14145010691)*x_ref[1]
        ref[3,3,2,0]=(-1.13973037849)*x_ref[0]+(-0.256565914821)*x_ref[1]
        ref[3,3,2,1]=(0.955478585358)*x_ref[0]+(-0.83417326515)*x_ref[1]
        ref[3,4,0,0]=(0.155710789468)*x_ref[0]+(0.204956811648)*x_ref[1]
        ref[3,4,0,1]=(0.567866952974)*x_ref[0]+(-0.168398320099)*x_ref[1]
        ref[3,4,1,0]=(-1.36997367)*x_ref[0]+(0.141984040172)*x_ref[1]
        ref[3,4,1,1]=(-0.809532285206)*x_ref[0]+(0.70611485405)*x_ref[1]
        ref[3,4,2,0]=(-0.453260232051)*x_ref[0]+(0.0665127865858)*x_ref[1]
        ref[3,4,2,1]=(0.00142021115277)*x_ref[0]+(1.02307505343)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.59037121238)*x[0]+(-0.432259357387)*x[1]+(1.06313464506)*x[2]
        arg[0,0,0,1]=(-0.29317617186)*x[0]+(-0.477878024138)*x[1]+(-0.943279870083)*x[2]
        arg[0,0,1,0]=(-0.0140970955833)*x[0]+(0.141752937281)*x[1]+(0.46748311988)*x[2]
        arg[0,0,1,1]=(1.06645552756)*x[0]+(-1.4360320823)*x[1]+(-1.05643251873)*x[2]
        arg[0,0,2,0]=(-0.527854671619)*x[0]+(0.68918711423)*x[1]+(0.225750468582)*x[2]
        arg[0,0,2,1]=(0.765285169635)*x[0]+(-0.292098568154)*x[1]+(-0.611702446821)*x[2]
        arg[0,1,0,0]=(0.472029729041)*x[0]+(-0.391383508709)*x[1]+(-0.0882199174494)*x[2]
        arg[0,1,0,1]=(1.76010861737)*x[0]+(1.40668629263)*x[1]+(-0.171024087043)*x[2]
        arg[0,1,1,0]=(0.251929705017)*x[0]+(-0.326420063624)*x[1]+(-0.337849942832)*x[2]
        arg[0,1,1,1]=(0.049289902392)*x[0]+(1.61775987596)*x[1]+(0.545741506198)*x[2]
        arg[0,1,2,0]=(-0.500710333189)*x[0]+(-0.551222136922)*x[1]+(0.665856619864)*x[2]
        arg[0,1,2,1]=(0.271547935249)*x[0]+(0.46154733209)*x[1]+(0.0776993672589)*x[2]
        arg[0,2,0,0]=(0.46744023541)*x[0]+(0.281754164985)*x[1]+(0.11721465855)*x[2]
        arg[0,2,0,1]=(0.58865724402)*x[0]+(-0.280911783535)*x[1]+(0.793461201474)*x[2]
        arg[0,2,1,0]=(-0.348215254122)*x[0]+(-0.274357607001)*x[1]+(-0.975813503666)*x[2]
        arg[0,2,1,1]=(-0.546269224094)*x[0]+(0.54132986357)*x[1]+(-0.346797273682)*x[2]
        arg[0,2,2,0]=(0.956395362583)*x[0]+(-0.609333787208)*x[1]+(-0.86408067145)*x[2]
        arg[0,2,2,1]=(0.570075324955)*x[0]+(-0.251638346903)*x[1]+(0.292841304398)*x[2]
        arg[0,3,0,0]=(-0.240717117869)*x[0]+(0.571307481772)*x[1]+(0.66826910816)*x[2]
        arg[0,3,0,1]=(0.113871023723)*x[0]+(0.022498355636)*x[1]+(0.95331234379)*x[2]
        arg[0,3,1,0]=(-0.264946899001)*x[0]+(1.08412373438)*x[1]+(-1.10548347411)*x[2]
        arg[0,3,1,1]=(-0.744510132854)*x[0]+(-0.716638983187)*x[1]+(0.64330838226)*x[2]
        arg[0,3,2,0]=(1.15237343787)*x[0]+(-1.42565901229)*x[1]+(0.727585401396)*x[2]
        arg[0,3,2,1]=(-0.251464863627)*x[0]+(1.4067624912)*x[1]+(0.674418739118)*x[2]
        arg[0,4,0,0]=(-0.00551820316191)*x[0]+(0.193862853101)*x[1]+(1.8004924522)*x[2]
        arg[0,4,0,1]=(-0.985879358916)*x[0]+(0.476965815991)*x[1]+(0.96363449822)*x[2]
        arg[0,4,1,0]=(-0.657967361584)*x[0]+(0.833452284341)*x[1]+(-1.40128003822)*x[2]
        arg[0,4,1,1]=(0.301379967988)*x[0]+(0.0474545542353)*x[1]+(-0.633087920569)*x[2]
        arg[0,4,2,0]=(-1.94409755409)*x[0]+(-0.468638363928)*x[1]+(-0.713880237452)*x[2]
        arg[0,4,2,1]=(-1.13536316065)*x[0]+(-0.250761280127)*x[1]+(0.328722833182)*x[2]
        arg[1,0,0,0]=(0.556886266554)*x[0]+(-0.70523094638)*x[1]+(-0.333621268224)*x[2]
        arg[1,0,0,1]=(-0.995565683628)*x[0]+(-0.158120855239)*x[1]+(0.0738870470169)*x[2]
        arg[1,0,1,0]=(-0.209556712574)*x[0]+(0.54393408711)*x[1]+(-1.1617498966)*x[2]
        arg[1,0,1,1]=(0.666809649924)*x[0]+(1.18555984779)*x[1]+(-0.442831598796)*x[2]
        arg[1,0,2,0]=(-0.21354395591)*x[0]+(-0.634626698269)*x[1]+(1.25643757008)*x[2]
        arg[1,0,2,1]=(-0.676259187073)*x[0]+(-0.458532774397)*x[1]+(0.271156919902)*x[2]
        arg[1,1,0,0]=(-1.45529139262)*x[0]+(0.311382249732)*x[1]+(-0.134620909882)*x[2]
        arg[1,1,0,1]=(-1.47708829198)*x[0]+(0.968291536601)*x[1]+(-0.566606887608)*x[2]
        arg[1,1,1,0]=(1.18904329045)*x[0]+(-0.209072247061)*x[1]+(-0.26261457093)*x[2]
        arg[1,1,1,1]=(0.313723817215)*x[0]+(-0.169177309423)*x[1]+(-0.00119161234247)*x[2]
        arg[1,1,2,0]=(0.431298197457)*x[0]+(0.264113678484)*x[1]+(0.192619025313)*x[2]
        arg[1,1,2,1]=(0.0469221259627)*x[0]+(1.53477696238)*x[1]+(0.00794662724457)*x[2]
        arg[1,2,0,0]=(0.102179220279)*x[0]+(-0.391365988191)*x[1]+(0.302910557263)*x[2]
        arg[1,2,0,1]=(0.841803002697)*x[0]+(0.581539244301)*x[1]+(0.00409108885792)*x[2]
        arg[1,2,1,0]=(0.0840378161144)*x[0]+(-0.431329807551)*x[1]+(-1.0774029159)*x[2]
        arg[1,2,1,1]=(-0.270803943532)*x[0]+(0.0124514531652)*x[1]+(0.197735227059)*x[2]
        arg[1,2,2,0]=(0.479336250918)*x[0]+(0.0435331360815)*x[1]+(0.458048853417)*x[2]
        arg[1,2,2,1]=(-0.59916419564)*x[0]+(0.415546807791)*x[1]+(-1.12105410423)*x[2]
        arg[1,3,0,0]=(0.378277354507)*x[0]+(1.00734343779)*x[1]+(-0.282811389334)*x[2]
        arg[1,3,0,1]=(0.717420487474)*x[0]+(-0.730395427407)*x[1]+(1.63850706843)*x[2]
        arg[1,3,1,0]=(-0.0126954991872)*x[0]+(-1.35916858377)*x[1]+(-0.955731750212)*x[2]
        arg[1,3,1,1]=(-0.560951559862)*x[0]+(0.271085100351)*x[1]+(0.60767567276)*x[2]
        arg[1,3,2,0]=(-0.73286080289)*x[0]+(-1.12018839274)*x[1]+(0.365586479937)*x[2]
        arg[1,3,2,1]=(-0.495391543232)*x[0]+(-0.404605421036)*x[1]+(1.21261493791)*x[2]
        arg[1,4,0,0]=(-0.322147956238)*x[0]+(0.278257549185)*x[1]+(-0.927648624956)*x[2]
        arg[1,4,0,1]=(1.23243511737)*x[0]+(1.38355111729)*x[1]+(0.0813739945129)*x[2]
        arg[1,4,1,0]=(0.668384291635)*x[0]+(0.995328961231)*x[1]+(0.803009543307)*x[2]
        arg[1,4,1,1]=(1.42643018603)*x[0]+(0.940855692156)*x[1]+(0.277556451163)*x[2]
        arg[1,4,2,0]=(0.489588226757)*x[0]+(-1.63608809124)*x[1]+(-0.262918068522)*x[2]
        arg[1,4,2,1]=(-1.06922340601)*x[0]+(-0.0241938384008)*x[1]+(-0.244107290038)*x[2]
        arg[2,0,0,0]=(-0.556338154459)*x[0]+(0.12071126785)*x[1]+(0.11215804852)*x[2]
        arg[2,0,0,1]=(-0.479017339774)*x[0]+(-0.530313718209)*x[1]+(0.490083551372)*x[2]
        arg[2,0,1,0]=(0.731269264444)*x[0]+(0.245793038193)*x[1]+(0.0479833912422)*x[2]
        arg[2,0,1,1]=(-0.286189382681)*x[0]+(0.601756275554)*x[1]+(-0.559334203137)*x[2]
        arg[2,0,2,0]=(-0.876495720673)*x[0]+(0.592513223261)*x[1]+(0.844551261652)*x[2]
        arg[2,0,2,1]=(1.22915142187)*x[0]+(0.136223623577)*x[1]+(0.88656874029)*x[2]
        arg[2,1,0,0]=(0.542222650916)*x[0]+(-0.709719163926)*x[1]+(0.262207188498)*x[2]
        arg[2,1,0,1]=(-0.208896980258)*x[0]+(1.33968890395)*x[1]+(-0.599621090138)*x[2]
        arg[2,1,1,0]=(-0.171147406604)*x[0]+(0.332227749427)*x[1]+(1.15043521586)*x[2]
        arg[2,1,1,1]=(0.571607567027)*x[0]+(0.1367434941)*x[1]+(1.47797159814)*x[2]
        arg[2,1,2,0]=(0.839845574634)*x[0]+(0.732437808234)*x[1]+(-0.0311071654701)*x[2]
        arg[2,1,2,1]=(1.45193213653)*x[0]+(-0.0192312612414)*x[1]+(0.308521161282)*x[2]
        arg[2,2,0,0]=(-0.122755384028)*x[0]+(0.133510038844)*x[1]+(0.837213365425)*x[2]
        arg[2,2,0,1]=(-0.411967923958)*x[0]+(-0.955489021071)*x[1]+(0.269005602186)*x[2]
        arg[2,2,1,0]=(-1.19993089382)*x[0]+(0.276541296246)*x[1]+(0.523388479396)*x[2]
        arg[2,2,1,1]=(0.0586514653869)*x[0]+(-0.624975232742)*x[1]+(-1.55738876751)*x[2]
        arg[2,2,2,0]=(-0.892065938887)*x[0]+(-0.4848241954)*x[1]+(0.315463883511)*x[2]
        arg[2,2,2,1]=(0.0272963181118)*x[0]+(0.0187713454669)*x[1]+(-1.06524661854)*x[2]
        arg[2,3,0,0]=(-1.44099287886)*x[0]+(0.495718421075)*x[1]+(0.100401960223)*x[2]
        arg[2,3,0,1]=(0.171175119412)*x[0]+(-1.34795558607)*x[1]+(-0.361766653264)*x[2]
        arg[2,3,1,0]=(-1.80591882588)*x[0]+(-0.511459455136)*x[1]+(-0.491310264055)*x[2]
        arg[2,3,1,1]=(-0.419508934622)*x[0]+(-1.12084042844)*x[1]+(-0.086947672548)*x[2]
        arg[2,3,2,0]=(0.492388303934)*x[0]+(-0.623505828597)*x[1]+(1.6965873745)*x[2]
        arg[2,3,2,1]=(1.59203810356)*x[0]+(0.36918142231)*x[1]+(-1.90067115052)*x[2]
        arg[2,4,0,0]=(1.00875649298)*x[0]+(-1.43246384934)*x[1]+(-0.51811906593)*x[2]
        arg[2,4,0,1]=(0.260118918034)*x[0]+(-1.57606061865)*x[1]+(-0.79257165655)*x[2]
        arg[2,4,1,0]=(0.954414074467)*x[0]+(-0.230132876613)*x[1]+(0.0956378957424)*x[2]
        arg[2,4,1,1]=(0.958843806364)*x[0]+(-0.155804429995)*x[1]+(-1.02904877039)*x[2]
        arg[2,4,2,0]=(0.252668700811)*x[0]+(0.164829885749)*x[1]+(0.747032722421)*x[2]
        arg[2,4,2,1]=(1.96393052323)*x[0]+(-0.821363958807)*x[1]+(-1.01663555063)*x[2]
        arg[3,0,0,0]=(1.58086037053)*x[0]+(-1.66495295611)*x[1]+(-0.510526191383)*x[2]
        arg[3,0,0,1]=(-0.931871976029)*x[0]+(1.33102205125)*x[1]+(-0.95170580468)*x[2]
        arg[3,0,1,0]=(0.99825404115)*x[0]+(0.915274413552)*x[1]+(1.5057227674)*x[2]
        arg[3,0,1,1]=(0.0485551924989)*x[0]+(-1.0877405758)*x[1]+(-0.887342573702)*x[2]
        arg[3,0,2,0]=(-0.745146140677)*x[0]+(-0.0108141266015)*x[1]+(0.194597576308)*x[2]
        arg[3,0,2,1]=(-0.935029218338)*x[0]+(0.726322809025)*x[1]+(0.57508467103)*x[2]
        arg[3,1,0,0]=(0.115668237156)*x[0]+(0.0970255991906)*x[1]+(-0.227291123547)*x[2]
        arg[3,1,0,1]=(0.167722335863)*x[0]+(-0.552168424291)*x[1]+(1.36293027463)*x[2]
        arg[3,1,1,0]=(-1.05781878028)*x[0]+(-0.0429389047972)*x[1]+(-0.257166337509)*x[2]
        arg[3,1,1,1]=(0.129148486118)*x[0]+(-0.0209434899369)*x[1]+(0.253153005253)*x[2]
        arg[3,1,2,0]=(1.3236615365)*x[0]+(-1.18164469224)*x[1]+(0.698012697426)*x[2]
        arg[3,1,2,1]=(1.10438661886)*x[0]+(-0.852903948395)*x[1]+(1.40365162024)*x[2]
        arg[3,2,0,0]=(-0.949308009957)*x[0]+(1.01993206161)*x[1]+(-0.821721697476)*x[2]
        arg[3,2,0,1]=(0.326450662787)*x[0]+(0.395096088107)*x[1]+(-0.368904800584)*x[2]
        arg[3,2,1,0]=(0.13107780566)*x[0]+(-0.342904134623)*x[1]+(0.00835087701185)*x[2]
        arg[3,2,1,1]=(-0.500694550647)*x[0]+(1.38322390003)*x[1]+(0.0576453183412)*x[2]
        arg[3,2,2,0]=(-1.32014489811)*x[0]+(-0.444079317864)*x[1]+(1.1135946018)*x[2]
        arg[3,2,2,1]=(-0.19593105007)*x[0]+(-0.00224390969377)*x[1]+(-0.73869338942)*x[2]
        arg[3,3,0,0]=(0.0151201622255)*x[0]+(-1.10873788153)*x[1]+(0.189116254832)*x[2]
        arg[3,3,0,1]=(1.05996205566)*x[0]+(0.398465639867)*x[1]+(-0.319773104448)*x[2]
        arg[3,3,1,0]=(0.687612827243)*x[0]+(1.10970823257)*x[1]+(1.02543288975)*x[2]
        arg[3,3,1,1]=(1.27258509981)*x[0]+(-1.7060826391)*x[1]+(-0.84948812227)*x[2]
        arg[3,3,2,0]=(1.2343036886)*x[0]+(0.521090253023)*x[1]+(-1.13377412954)*x[2]
        arg[3,3,2,1]=(-0.65568738385)*x[0]+(1.00354617712)*x[1]+(0.836719015812)*x[2]
        arg[3,4,0,0]=(-0.545324078594)*x[0]+(0.452747782208)*x[1]+(-1.59718954237)*x[2]
        arg[3,4,0,1]=(1.71651153317)*x[0]+(-0.161984538787)*x[1]+(-0.377244827904)*x[2]
        arg[3,4,1,0]=(-0.284033324558)*x[0]+(-0.405941433062)*x[1]+(0.824573264594)*x[2]
        arg[3,4,1,1]=(-0.112542702803)*x[0]+(-0.470814805786)*x[1]+(-0.827427436576)*x[2]
        arg[3,4,2,0]=(-0.763655132981)*x[0]+(-0.442818775864)*x[1]+(0.809814510041)*x[2]
        arg[3,4,2,1]=(-0.33794306298)*x[0]+(1.91169591332)*x[1]+(-0.487268088196)*x[2]
        ref[0,0,0,0]=(-0.59037121238)*x_ref[0]+(-0.432259357387)*x_ref[1]+(1.06313464506)*x_ref[2]
        ref[0,0,0,1]=(-0.29317617186)*x_ref[0]+(-0.477878024138)*x_ref[1]+(-0.943279870083)*x_ref[2]
        ref[0,0,1,0]=(-0.0140970955833)*x_ref[0]+(0.141752937281)*x_ref[1]+(0.46748311988)*x_ref[2]
        ref[0,0,1,1]=(1.06645552756)*x_ref[0]+(-1.4360320823)*x_ref[1]+(-1.05643251873)*x_ref[2]
        ref[0,0,2,0]=(-0.527854671619)*x_ref[0]+(0.68918711423)*x_ref[1]+(0.225750468582)*x_ref[2]
        ref[0,0,2,1]=(0.765285169635)*x_ref[0]+(-0.292098568154)*x_ref[1]+(-0.611702446821)*x_ref[2]
        ref[0,1,0,0]=(0.472029729041)*x_ref[0]+(-0.391383508709)*x_ref[1]+(-0.0882199174494)*x_ref[2]
        ref[0,1,0,1]=(1.76010861737)*x_ref[0]+(1.40668629263)*x_ref[1]+(-0.171024087043)*x_ref[2]
        ref[0,1,1,0]=(0.251929705017)*x_ref[0]+(-0.326420063624)*x_ref[1]+(-0.337849942832)*x_ref[2]
        ref[0,1,1,1]=(0.049289902392)*x_ref[0]+(1.61775987596)*x_ref[1]+(0.545741506198)*x_ref[2]
        ref[0,1,2,0]=(-0.500710333189)*x_ref[0]+(-0.551222136922)*x_ref[1]+(0.665856619864)*x_ref[2]
        ref[0,1,2,1]=(0.271547935249)*x_ref[0]+(0.46154733209)*x_ref[1]+(0.0776993672589)*x_ref[2]
        ref[0,2,0,0]=(0.46744023541)*x_ref[0]+(0.281754164985)*x_ref[1]+(0.11721465855)*x_ref[2]
        ref[0,2,0,1]=(0.58865724402)*x_ref[0]+(-0.280911783535)*x_ref[1]+(0.793461201474)*x_ref[2]
        ref[0,2,1,0]=(-0.348215254122)*x_ref[0]+(-0.274357607001)*x_ref[1]+(-0.975813503666)*x_ref[2]
        ref[0,2,1,1]=(-0.546269224094)*x_ref[0]+(0.54132986357)*x_ref[1]+(-0.346797273682)*x_ref[2]
        ref[0,2,2,0]=(0.956395362583)*x_ref[0]+(-0.609333787208)*x_ref[1]+(-0.86408067145)*x_ref[2]
        ref[0,2,2,1]=(0.570075324955)*x_ref[0]+(-0.251638346903)*x_ref[1]+(0.292841304398)*x_ref[2]
        ref[0,3,0,0]=(-0.240717117869)*x_ref[0]+(0.571307481772)*x_ref[1]+(0.66826910816)*x_ref[2]
        ref[0,3,0,1]=(0.113871023723)*x_ref[0]+(0.022498355636)*x_ref[1]+(0.95331234379)*x_ref[2]
        ref[0,3,1,0]=(-0.264946899001)*x_ref[0]+(1.08412373438)*x_ref[1]+(-1.10548347411)*x_ref[2]
        ref[0,3,1,1]=(-0.744510132854)*x_ref[0]+(-0.716638983187)*x_ref[1]+(0.64330838226)*x_ref[2]
        ref[0,3,2,0]=(1.15237343787)*x_ref[0]+(-1.42565901229)*x_ref[1]+(0.727585401396)*x_ref[2]
        ref[0,3,2,1]=(-0.251464863627)*x_ref[0]+(1.4067624912)*x_ref[1]+(0.674418739118)*x_ref[2]
        ref[0,4,0,0]=(-0.00551820316191)*x_ref[0]+(0.193862853101)*x_ref[1]+(1.8004924522)*x_ref[2]
        ref[0,4,0,1]=(-0.985879358916)*x_ref[0]+(0.476965815991)*x_ref[1]+(0.96363449822)*x_ref[2]
        ref[0,4,1,0]=(-0.657967361584)*x_ref[0]+(0.833452284341)*x_ref[1]+(-1.40128003822)*x_ref[2]
        ref[0,4,1,1]=(0.301379967988)*x_ref[0]+(0.0474545542353)*x_ref[1]+(-0.633087920569)*x_ref[2]
        ref[0,4,2,0]=(-1.94409755409)*x_ref[0]+(-0.468638363928)*x_ref[1]+(-0.713880237452)*x_ref[2]
        ref[0,4,2,1]=(-1.13536316065)*x_ref[0]+(-0.250761280127)*x_ref[1]+(0.328722833182)*x_ref[2]
        ref[1,0,0,0]=(0.556886266554)*x_ref[0]+(-0.70523094638)*x_ref[1]+(-0.333621268224)*x_ref[2]
        ref[1,0,0,1]=(-0.995565683628)*x_ref[0]+(-0.158120855239)*x_ref[1]+(0.0738870470169)*x_ref[2]
        ref[1,0,1,0]=(-0.209556712574)*x_ref[0]+(0.54393408711)*x_ref[1]+(-1.1617498966)*x_ref[2]
        ref[1,0,1,1]=(0.666809649924)*x_ref[0]+(1.18555984779)*x_ref[1]+(-0.442831598796)*x_ref[2]
        ref[1,0,2,0]=(-0.21354395591)*x_ref[0]+(-0.634626698269)*x_ref[1]+(1.25643757008)*x_ref[2]
        ref[1,0,2,1]=(-0.676259187073)*x_ref[0]+(-0.458532774397)*x_ref[1]+(0.271156919902)*x_ref[2]
        ref[1,1,0,0]=(-1.45529139262)*x_ref[0]+(0.311382249732)*x_ref[1]+(-0.134620909882)*x_ref[2]
        ref[1,1,0,1]=(-1.47708829198)*x_ref[0]+(0.968291536601)*x_ref[1]+(-0.566606887608)*x_ref[2]
        ref[1,1,1,0]=(1.18904329045)*x_ref[0]+(-0.209072247061)*x_ref[1]+(-0.26261457093)*x_ref[2]
        ref[1,1,1,1]=(0.313723817215)*x_ref[0]+(-0.169177309423)*x_ref[1]+(-0.00119161234247)*x_ref[2]
        ref[1,1,2,0]=(0.431298197457)*x_ref[0]+(0.264113678484)*x_ref[1]+(0.192619025313)*x_ref[2]
        ref[1,1,2,1]=(0.0469221259627)*x_ref[0]+(1.53477696238)*x_ref[1]+(0.00794662724457)*x_ref[2]
        ref[1,2,0,0]=(0.102179220279)*x_ref[0]+(-0.391365988191)*x_ref[1]+(0.302910557263)*x_ref[2]
        ref[1,2,0,1]=(0.841803002697)*x_ref[0]+(0.581539244301)*x_ref[1]+(0.00409108885792)*x_ref[2]
        ref[1,2,1,0]=(0.0840378161144)*x_ref[0]+(-0.431329807551)*x_ref[1]+(-1.0774029159)*x_ref[2]
        ref[1,2,1,1]=(-0.270803943532)*x_ref[0]+(0.0124514531652)*x_ref[1]+(0.197735227059)*x_ref[2]
        ref[1,2,2,0]=(0.479336250918)*x_ref[0]+(0.0435331360815)*x_ref[1]+(0.458048853417)*x_ref[2]
        ref[1,2,2,1]=(-0.59916419564)*x_ref[0]+(0.415546807791)*x_ref[1]+(-1.12105410423)*x_ref[2]
        ref[1,3,0,0]=(0.378277354507)*x_ref[0]+(1.00734343779)*x_ref[1]+(-0.282811389334)*x_ref[2]
        ref[1,3,0,1]=(0.717420487474)*x_ref[0]+(-0.730395427407)*x_ref[1]+(1.63850706843)*x_ref[2]
        ref[1,3,1,0]=(-0.0126954991872)*x_ref[0]+(-1.35916858377)*x_ref[1]+(-0.955731750212)*x_ref[2]
        ref[1,3,1,1]=(-0.560951559862)*x_ref[0]+(0.271085100351)*x_ref[1]+(0.60767567276)*x_ref[2]
        ref[1,3,2,0]=(-0.73286080289)*x_ref[0]+(-1.12018839274)*x_ref[1]+(0.365586479937)*x_ref[2]
        ref[1,3,2,1]=(-0.495391543232)*x_ref[0]+(-0.404605421036)*x_ref[1]+(1.21261493791)*x_ref[2]
        ref[1,4,0,0]=(-0.322147956238)*x_ref[0]+(0.278257549185)*x_ref[1]+(-0.927648624956)*x_ref[2]
        ref[1,4,0,1]=(1.23243511737)*x_ref[0]+(1.38355111729)*x_ref[1]+(0.0813739945129)*x_ref[2]
        ref[1,4,1,0]=(0.668384291635)*x_ref[0]+(0.995328961231)*x_ref[1]+(0.803009543307)*x_ref[2]
        ref[1,4,1,1]=(1.42643018603)*x_ref[0]+(0.940855692156)*x_ref[1]+(0.277556451163)*x_ref[2]
        ref[1,4,2,0]=(0.489588226757)*x_ref[0]+(-1.63608809124)*x_ref[1]+(-0.262918068522)*x_ref[2]
        ref[1,4,2,1]=(-1.06922340601)*x_ref[0]+(-0.0241938384008)*x_ref[1]+(-0.244107290038)*x_ref[2]
        ref[2,0,0,0]=(-0.556338154459)*x_ref[0]+(0.12071126785)*x_ref[1]+(0.11215804852)*x_ref[2]
        ref[2,0,0,1]=(-0.479017339774)*x_ref[0]+(-0.530313718209)*x_ref[1]+(0.490083551372)*x_ref[2]
        ref[2,0,1,0]=(0.731269264444)*x_ref[0]+(0.245793038193)*x_ref[1]+(0.0479833912422)*x_ref[2]
        ref[2,0,1,1]=(-0.286189382681)*x_ref[0]+(0.601756275554)*x_ref[1]+(-0.559334203137)*x_ref[2]
        ref[2,0,2,0]=(-0.876495720673)*x_ref[0]+(0.592513223261)*x_ref[1]+(0.844551261652)*x_ref[2]
        ref[2,0,2,1]=(1.22915142187)*x_ref[0]+(0.136223623577)*x_ref[1]+(0.88656874029)*x_ref[2]
        ref[2,1,0,0]=(0.542222650916)*x_ref[0]+(-0.709719163926)*x_ref[1]+(0.262207188498)*x_ref[2]
        ref[2,1,0,1]=(-0.208896980258)*x_ref[0]+(1.33968890395)*x_ref[1]+(-0.599621090138)*x_ref[2]
        ref[2,1,1,0]=(-0.171147406604)*x_ref[0]+(0.332227749427)*x_ref[1]+(1.15043521586)*x_ref[2]
        ref[2,1,1,1]=(0.571607567027)*x_ref[0]+(0.1367434941)*x_ref[1]+(1.47797159814)*x_ref[2]
        ref[2,1,2,0]=(0.839845574634)*x_ref[0]+(0.732437808234)*x_ref[1]+(-0.0311071654701)*x_ref[2]
        ref[2,1,2,1]=(1.45193213653)*x_ref[0]+(-0.0192312612414)*x_ref[1]+(0.308521161282)*x_ref[2]
        ref[2,2,0,0]=(-0.122755384028)*x_ref[0]+(0.133510038844)*x_ref[1]+(0.837213365425)*x_ref[2]
        ref[2,2,0,1]=(-0.411967923958)*x_ref[0]+(-0.955489021071)*x_ref[1]+(0.269005602186)*x_ref[2]
        ref[2,2,1,0]=(-1.19993089382)*x_ref[0]+(0.276541296246)*x_ref[1]+(0.523388479396)*x_ref[2]
        ref[2,2,1,1]=(0.0586514653869)*x_ref[0]+(-0.624975232742)*x_ref[1]+(-1.55738876751)*x_ref[2]
        ref[2,2,2,0]=(-0.892065938887)*x_ref[0]+(-0.4848241954)*x_ref[1]+(0.315463883511)*x_ref[2]
        ref[2,2,2,1]=(0.0272963181118)*x_ref[0]+(0.0187713454669)*x_ref[1]+(-1.06524661854)*x_ref[2]
        ref[2,3,0,0]=(-1.44099287886)*x_ref[0]+(0.495718421075)*x_ref[1]+(0.100401960223)*x_ref[2]
        ref[2,3,0,1]=(0.171175119412)*x_ref[0]+(-1.34795558607)*x_ref[1]+(-0.361766653264)*x_ref[2]
        ref[2,3,1,0]=(-1.80591882588)*x_ref[0]+(-0.511459455136)*x_ref[1]+(-0.491310264055)*x_ref[2]
        ref[2,3,1,1]=(-0.419508934622)*x_ref[0]+(-1.12084042844)*x_ref[1]+(-0.086947672548)*x_ref[2]
        ref[2,3,2,0]=(0.492388303934)*x_ref[0]+(-0.623505828597)*x_ref[1]+(1.6965873745)*x_ref[2]
        ref[2,3,2,1]=(1.59203810356)*x_ref[0]+(0.36918142231)*x_ref[1]+(-1.90067115052)*x_ref[2]
        ref[2,4,0,0]=(1.00875649298)*x_ref[0]+(-1.43246384934)*x_ref[1]+(-0.51811906593)*x_ref[2]
        ref[2,4,0,1]=(0.260118918034)*x_ref[0]+(-1.57606061865)*x_ref[1]+(-0.79257165655)*x_ref[2]
        ref[2,4,1,0]=(0.954414074467)*x_ref[0]+(-0.230132876613)*x_ref[1]+(0.0956378957424)*x_ref[2]
        ref[2,4,1,1]=(0.958843806364)*x_ref[0]+(-0.155804429995)*x_ref[1]+(-1.02904877039)*x_ref[2]
        ref[2,4,2,0]=(0.252668700811)*x_ref[0]+(0.164829885749)*x_ref[1]+(0.747032722421)*x_ref[2]
        ref[2,4,2,1]=(1.96393052323)*x_ref[0]+(-0.821363958807)*x_ref[1]+(-1.01663555063)*x_ref[2]
        ref[3,0,0,0]=(1.58086037053)*x_ref[0]+(-1.66495295611)*x_ref[1]+(-0.510526191383)*x_ref[2]
        ref[3,0,0,1]=(-0.931871976029)*x_ref[0]+(1.33102205125)*x_ref[1]+(-0.95170580468)*x_ref[2]
        ref[3,0,1,0]=(0.99825404115)*x_ref[0]+(0.915274413552)*x_ref[1]+(1.5057227674)*x_ref[2]
        ref[3,0,1,1]=(0.0485551924989)*x_ref[0]+(-1.0877405758)*x_ref[1]+(-0.887342573702)*x_ref[2]
        ref[3,0,2,0]=(-0.745146140677)*x_ref[0]+(-0.0108141266015)*x_ref[1]+(0.194597576308)*x_ref[2]
        ref[3,0,2,1]=(-0.935029218338)*x_ref[0]+(0.726322809025)*x_ref[1]+(0.57508467103)*x_ref[2]
        ref[3,1,0,0]=(0.115668237156)*x_ref[0]+(0.0970255991906)*x_ref[1]+(-0.227291123547)*x_ref[2]
        ref[3,1,0,1]=(0.167722335863)*x_ref[0]+(-0.552168424291)*x_ref[1]+(1.36293027463)*x_ref[2]
        ref[3,1,1,0]=(-1.05781878028)*x_ref[0]+(-0.0429389047972)*x_ref[1]+(-0.257166337509)*x_ref[2]
        ref[3,1,1,1]=(0.129148486118)*x_ref[0]+(-0.0209434899369)*x_ref[1]+(0.253153005253)*x_ref[2]
        ref[3,1,2,0]=(1.3236615365)*x_ref[0]+(-1.18164469224)*x_ref[1]+(0.698012697426)*x_ref[2]
        ref[3,1,2,1]=(1.10438661886)*x_ref[0]+(-0.852903948395)*x_ref[1]+(1.40365162024)*x_ref[2]
        ref[3,2,0,0]=(-0.949308009957)*x_ref[0]+(1.01993206161)*x_ref[1]+(-0.821721697476)*x_ref[2]
        ref[3,2,0,1]=(0.326450662787)*x_ref[0]+(0.395096088107)*x_ref[1]+(-0.368904800584)*x_ref[2]
        ref[3,2,1,0]=(0.13107780566)*x_ref[0]+(-0.342904134623)*x_ref[1]+(0.00835087701185)*x_ref[2]
        ref[3,2,1,1]=(-0.500694550647)*x_ref[0]+(1.38322390003)*x_ref[1]+(0.0576453183412)*x_ref[2]
        ref[3,2,2,0]=(-1.32014489811)*x_ref[0]+(-0.444079317864)*x_ref[1]+(1.1135946018)*x_ref[2]
        ref[3,2,2,1]=(-0.19593105007)*x_ref[0]+(-0.00224390969377)*x_ref[1]+(-0.73869338942)*x_ref[2]
        ref[3,3,0,0]=(0.0151201622255)*x_ref[0]+(-1.10873788153)*x_ref[1]+(0.189116254832)*x_ref[2]
        ref[3,3,0,1]=(1.05996205566)*x_ref[0]+(0.398465639867)*x_ref[1]+(-0.319773104448)*x_ref[2]
        ref[3,3,1,0]=(0.687612827243)*x_ref[0]+(1.10970823257)*x_ref[1]+(1.02543288975)*x_ref[2]
        ref[3,3,1,1]=(1.27258509981)*x_ref[0]+(-1.7060826391)*x_ref[1]+(-0.84948812227)*x_ref[2]
        ref[3,3,2,0]=(1.2343036886)*x_ref[0]+(0.521090253023)*x_ref[1]+(-1.13377412954)*x_ref[2]
        ref[3,3,2,1]=(-0.65568738385)*x_ref[0]+(1.00354617712)*x_ref[1]+(0.836719015812)*x_ref[2]
        ref[3,4,0,0]=(-0.545324078594)*x_ref[0]+(0.452747782208)*x_ref[1]+(-1.59718954237)*x_ref[2]
        ref[3,4,0,1]=(1.71651153317)*x_ref[0]+(-0.161984538787)*x_ref[1]+(-0.377244827904)*x_ref[2]
        ref[3,4,1,0]=(-0.284033324558)*x_ref[0]+(-0.405941433062)*x_ref[1]+(0.824573264594)*x_ref[2]
        ref[3,4,1,1]=(-0.112542702803)*x_ref[0]+(-0.470814805786)*x_ref[1]+(-0.827427436576)*x_ref[2]
        ref[3,4,2,0]=(-0.763655132981)*x_ref[0]+(-0.442818775864)*x_ref[1]+(0.809814510041)*x_ref[2]
        ref[3,4,2,1]=(-0.33794306298)*x_ref[0]+(1.91169591332)*x_ref[1]+(-0.487268088196)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.898871557679)*x[0]+(0.0218492481542)*x[1]
        ref=(0.898871557679)*x_ref[0]+(0.0218492481542)*x_ref[1]
      else:
        arg=(-0.490730396919)*x[0]+(0.840683566031)*x[1]+(-0.265457839434)*x[2]
        ref=(-0.490730396919)*x_ref[0]+(0.840683566031)*x_ref[1]+(-0.265457839434)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        arg[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
        ref[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        ref[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
      else:
        arg[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        arg[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
        ref[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        ref[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.609515733352)*x[0]+(0.477437125804)*x[1]
        arg[0,1]=(0.225027969847)*x[0]+(-1.2178256661)*x[1]
        arg[0,2]=(-0.414641287889)*x[0]+(-0.536814389632)*x[1]
        arg[0,3]=(0.993370642676)*x[0]+(0.430000528217)*x[1]
        arg[0,4]=(0.577804610335)*x[0]+(0.0950391061127)*x[1]
        arg[1,0]=(0.437140855856)*x[0]+(0.764946114688)*x[1]
        arg[1,1]=(-0.491195658809)*x[0]+(-1.04729131261)*x[1]
        arg[1,2]=(-0.179287249145)*x[0]+(-1.84699873991)*x[1]
        arg[1,3]=(0.650408136022)*x[0]+(0.437145116534)*x[1]
        arg[1,4]=(0.43853490169)*x[0]+(-0.633716126064)*x[1]
        arg[2,0]=(0.6275418675)*x[0]+(0.812839621482)*x[1]
        arg[2,1]=(0.151298084104)*x[0]+(0.191677853013)*x[1]
        arg[2,2]=(1.35800571802)*x[0]+(1.08489593249)*x[1]
        arg[2,3]=(0.343386100925)*x[0]+(-0.108323271881)*x[1]
        arg[2,4]=(1.5968773444)*x[0]+(0.421328566279)*x[1]
        arg[3,0]=(1.0607367804)*x[0]+(-0.347352081886)*x[1]
        arg[3,1]=(-0.307183028406)*x[0]+(0.171667489459)*x[1]
        arg[3,2]=(0.255609148103)*x[0]+(0.21702578756)*x[1]
        arg[3,3]=(0.204487068999)*x[0]+(0.380168463726)*x[1]
        arg[3,4]=(-0.727714186561)*x[0]+(-0.129616906837)*x[1]
        ref[0,0]=(-0.609515733352)*x_ref[0]+(0.477437125804)*x_ref[1]
        ref[0,1]=(0.225027969847)*x_ref[0]+(-1.2178256661)*x_ref[1]
        ref[0,2]=(-0.414641287889)*x_ref[0]+(-0.536814389632)*x_ref[1]
        ref[0,3]=(0.993370642676)*x_ref[0]+(0.430000528217)*x_ref[1]
        ref[0,4]=(0.577804610335)*x_ref[0]+(0.0950391061127)*x_ref[1]
        ref[1,0]=(0.437140855856)*x_ref[0]+(0.764946114688)*x_ref[1]
        ref[1,1]=(-0.491195658809)*x_ref[0]+(-1.04729131261)*x_ref[1]
        ref[1,2]=(-0.179287249145)*x_ref[0]+(-1.84699873991)*x_ref[1]
        ref[1,3]=(0.650408136022)*x_ref[0]+(0.437145116534)*x_ref[1]
        ref[1,4]=(0.43853490169)*x_ref[0]+(-0.633716126064)*x_ref[1]
        ref[2,0]=(0.6275418675)*x_ref[0]+(0.812839621482)*x_ref[1]
        ref[2,1]=(0.151298084104)*x_ref[0]+(0.191677853013)*x_ref[1]
        ref[2,2]=(1.35800571802)*x_ref[0]+(1.08489593249)*x_ref[1]
        ref[2,3]=(0.343386100925)*x_ref[0]+(-0.108323271881)*x_ref[1]
        ref[2,4]=(1.5968773444)*x_ref[0]+(0.421328566279)*x_ref[1]
        ref[3,0]=(1.0607367804)*x_ref[0]+(-0.347352081886)*x_ref[1]
        ref[3,1]=(-0.307183028406)*x_ref[0]+(0.171667489459)*x_ref[1]
        ref[3,2]=(0.255609148103)*x_ref[0]+(0.21702578756)*x_ref[1]
        ref[3,3]=(0.204487068999)*x_ref[0]+(0.380168463726)*x_ref[1]
        ref[3,4]=(-0.727714186561)*x_ref[0]+(-0.129616906837)*x_ref[1]
      else:
        arg[0,0]=(-0.282058347281)*x[0]+(-0.804878634484)*x[1]+(1.59321960831)*x[2]
        arg[0,1]=(0.046711660417)*x[0]+(-0.353452776565)*x[1]+(-1.22459734449)*x[2]
        arg[0,2]=(0.846231855307)*x[0]+(1.08618070898)*x[1]+(-1.50365393652)*x[2]
        arg[0,3]=(-0.358546890162)*x[0]+(1.53476318049)*x[1]+(-1.19382694871)*x[2]
        arg[0,4]=(-1.1943310015)*x[0]+(0.260842673421)*x[1]+(-0.337737972605)*x[2]
        arg[1,0]=(1.87166717118)*x[0]+(0.853452519127)*x[1]+(-0.59081660847)*x[2]
        arg[1,1]=(0.23310002871)*x[0]+(0.306981228888)*x[1]+(0.36040454683)*x[2]
        arg[1,2]=(-1.619288679)*x[0]+(-0.459254234211)*x[1]+(0.722479242965)*x[2]
        arg[1,3]=(-1.07724802769)*x[0]+(0.836465436044)*x[1]+(0.255446930493)*x[2]
        arg[1,4]=(1.07857092921)*x[0]+(1.25642859092)*x[1]+(-0.922799422574)*x[2]
        arg[2,0]=(-0.832179820568)*x[0]+(-0.511243247139)*x[1]+(0.589457094147)*x[2]
        arg[2,1]=(-0.22616483359)*x[0]+(1.75842052561)*x[1]+(-1.19438769374)*x[2]
        arg[2,2]=(-0.690678634371)*x[0]+(0.963335056891)*x[1]+(0.53653322138)*x[2]
        arg[2,3]=(-0.906140718249)*x[0]+(0.766237211441)*x[1]+(1.19445635284)*x[2]
        arg[2,4]=(-0.16966820901)*x[0]+(0.185706811112)*x[1]+(0.975546783315)*x[2]
        arg[3,0]=(1.85053644213)*x[0]+(1.000328518)*x[1]+(0.332466610087)*x[2]
        arg[3,1]=(-0.199393988242)*x[0]+(0.637601295557)*x[1]+(1.29246614993)*x[2]
        arg[3,2]=(-0.275167528438)*x[0]+(1.00557873837)*x[1]+(1.17143679274)*x[2]
        arg[3,3]=(-0.306797319235)*x[0]+(0.984827835253)*x[1]+(0.0324710900497)*x[2]
        arg[3,4]=(-1.07994829511)*x[0]+(0.341016609147)*x[1]+(0.115829923585)*x[2]
        ref[0,0]=(-0.282058347281)*x_ref[0]+(-0.804878634484)*x_ref[1]+(1.59321960831)*x_ref[2]
        ref[0,1]=(0.046711660417)*x_ref[0]+(-0.353452776565)*x_ref[1]+(-1.22459734449)*x_ref[2]
        ref[0,2]=(0.846231855307)*x_ref[0]+(1.08618070898)*x_ref[1]+(-1.50365393652)*x_ref[2]
        ref[0,3]=(-0.358546890162)*x_ref[0]+(1.53476318049)*x_ref[1]+(-1.19382694871)*x_ref[2]
        ref[0,4]=(-1.1943310015)*x_ref[0]+(0.260842673421)*x_ref[1]+(-0.337737972605)*x_ref[2]
        ref[1,0]=(1.87166717118)*x_ref[0]+(0.853452519127)*x_ref[1]+(-0.59081660847)*x_ref[2]
        ref[1,1]=(0.23310002871)*x_ref[0]+(0.306981228888)*x_ref[1]+(0.36040454683)*x_ref[2]
        ref[1,2]=(-1.619288679)*x_ref[0]+(-0.459254234211)*x_ref[1]+(0.722479242965)*x_ref[2]
        ref[1,3]=(-1.07724802769)*x_ref[0]+(0.836465436044)*x_ref[1]+(0.255446930493)*x_ref[2]
        ref[1,4]=(1.07857092921)*x_ref[0]+(1.25642859092)*x_ref[1]+(-0.922799422574)*x_ref[2]
        ref[2,0]=(-0.832179820568)*x_ref[0]+(-0.511243247139)*x_ref[1]+(0.589457094147)*x_ref[2]
        ref[2,1]=(-0.22616483359)*x_ref[0]+(1.75842052561)*x_ref[1]+(-1.19438769374)*x_ref[2]
        ref[2,2]=(-0.690678634371)*x_ref[0]+(0.963335056891)*x_ref[1]+(0.53653322138)*x_ref[2]
        ref[2,3]=(-0.906140718249)*x_ref[0]+(0.766237211441)*x_ref[1]+(1.19445635284)*x_ref[2]
        ref[2,4]=(-0.16966820901)*x_ref[0]+(0.185706811112)*x_ref[1]+(0.975546783315)*x_ref[2]
        ref[3,0]=(1.85053644213)*x_ref[0]+(1.000328518)*x_ref[1]+(0.332466610087)*x_ref[2]
        ref[3,1]=(-0.199393988242)*x_ref[0]+(0.637601295557)*x_ref[1]+(1.29246614993)*x_ref[2]
        ref[3,2]=(-0.275167528438)*x_ref[0]+(1.00557873837)*x_ref[1]+(1.17143679274)*x_ref[2]
        ref[3,3]=(-0.306797319235)*x_ref[0]+(0.984827835253)*x_ref[1]+(0.0324710900497)*x_ref[2]
        ref[3,4]=(-1.07994829511)*x_ref[0]+(0.341016609147)*x_ref[1]+(0.115829923585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.618861298758)*x[0]+(-0.764616734859)*x[1]
        arg[0,0,1]=(-0.220720229024)*x[0]+(-0.670415412641)*x[1]
        arg[0,1,0]=(0.847082172076)*x[0]+(0.104515154796)*x[1]
        arg[0,1,1]=(-1.39487869037)*x[0]+(1.37639032205)*x[1]
        arg[1,0,0]=(-0.12944065767)*x[0]+(-0.91214412252)*x[1]
        arg[1,0,1]=(0.0421119211697)*x[0]+(1.19918915102)*x[1]
        arg[1,1,0]=(0.657099038454)*x[0]+(0.102886690913)*x[1]
        arg[1,1,1]=(0.114311284584)*x[0]+(-0.827908524893)*x[1]
        arg[2,0,0]=(-0.905281334495)*x[0]+(1.02228309099)*x[1]
        arg[2,0,1]=(-0.560619228879)*x[0]+(-0.924304525905)*x[1]
        arg[2,1,0]=(0.082868967617)*x[0]+(-0.129063346527)*x[1]
        arg[2,1,1]=(0.386687881433)*x[0]+(-0.874995344656)*x[1]
        arg[3,0,0]=(0.460976605295)*x[0]+(0.145631537258)*x[1]
        arg[3,0,1]=(0.439977672901)*x[0]+(0.714321229059)*x[1]
        arg[3,1,0]=(-0.0459363656635)*x[0]+(-0.0754715887416)*x[1]
        arg[3,1,1]=(-0.789910257931)*x[0]+(-0.806078036909)*x[1]
        arg[4,0,0]=(-1.09951071165)*x[0]+(-1.33461983123)*x[1]
        arg[4,0,1]=(0.857184302111)*x[0]+(0.823036666525)*x[1]
        arg[4,1,0]=(-1.0682486872)*x[0]+(-0.149861981932)*x[1]
        arg[4,1,1]=(-0.0758367136425)*x[0]+(0.0460556685916)*x[1]
        arg[5,0,0]=(0.856207295965)*x[0]+(-0.885450852663)*x[1]
        arg[5,0,1]=(-0.113939086685)*x[0]+(0.742808810968)*x[1]
        arg[5,1,0]=(-0.174573302919)*x[0]+(0.486803189896)*x[1]
        arg[5,1,1]=(1.1726771004)*x[0]+(-0.329765988186)*x[1]
        ref[0,0,0]=(0.618861298758)*x_ref[0]+(-0.764616734859)*x_ref[1]
        ref[0,0,1]=(-0.220720229024)*x_ref[0]+(-0.670415412641)*x_ref[1]
        ref[0,1,0]=(0.847082172076)*x_ref[0]+(0.104515154796)*x_ref[1]
        ref[0,1,1]=(-1.39487869037)*x_ref[0]+(1.37639032205)*x_ref[1]
        ref[1,0,0]=(-0.12944065767)*x_ref[0]+(-0.91214412252)*x_ref[1]
        ref[1,0,1]=(0.0421119211697)*x_ref[0]+(1.19918915102)*x_ref[1]
        ref[1,1,0]=(0.657099038454)*x_ref[0]+(0.102886690913)*x_ref[1]
        ref[1,1,1]=(0.114311284584)*x_ref[0]+(-0.827908524893)*x_ref[1]
        ref[2,0,0]=(-0.905281334495)*x_ref[0]+(1.02228309099)*x_ref[1]
        ref[2,0,1]=(-0.560619228879)*x_ref[0]+(-0.924304525905)*x_ref[1]
        ref[2,1,0]=(0.082868967617)*x_ref[0]+(-0.129063346527)*x_ref[1]
        ref[2,1,1]=(0.386687881433)*x_ref[0]+(-0.874995344656)*x_ref[1]
        ref[3,0,0]=(0.460976605295)*x_ref[0]+(0.145631537258)*x_ref[1]
        ref[3,0,1]=(0.439977672901)*x_ref[0]+(0.714321229059)*x_ref[1]
        ref[3,1,0]=(-0.0459363656635)*x_ref[0]+(-0.0754715887416)*x_ref[1]
        ref[3,1,1]=(-0.789910257931)*x_ref[0]+(-0.806078036909)*x_ref[1]
        ref[4,0,0]=(-1.09951071165)*x_ref[0]+(-1.33461983123)*x_ref[1]
        ref[4,0,1]=(0.857184302111)*x_ref[0]+(0.823036666525)*x_ref[1]
        ref[4,1,0]=(-1.0682486872)*x_ref[0]+(-0.149861981932)*x_ref[1]
        ref[4,1,1]=(-0.0758367136425)*x_ref[0]+(0.0460556685916)*x_ref[1]
        ref[5,0,0]=(0.856207295965)*x_ref[0]+(-0.885450852663)*x_ref[1]
        ref[5,0,1]=(-0.113939086685)*x_ref[0]+(0.742808810968)*x_ref[1]
        ref[5,1,0]=(-0.174573302919)*x_ref[0]+(0.486803189896)*x_ref[1]
        ref[5,1,1]=(1.1726771004)*x_ref[0]+(-0.329765988186)*x_ref[1]
      else:
        arg[0,0,0]=(-0.640205798541)*x[0]+(0.271670719935)*x[1]+(-0.163000598261)*x[2]
        arg[0,0,1]=(0.735605297981)*x[0]+(-1.05892412676)*x[1]+(-1.1083007441)*x[2]
        arg[0,1,0]=(0.0349869408333)*x[0]+(0.699147831788)*x[1]+(-0.454390826725)*x[2]
        arg[0,1,1]=(-0.703957679243)*x[0]+(-1.63654420176)*x[1]+(-1.13045075795)*x[2]
        arg[1,0,0]=(1.01167296102)*x[0]+(-0.362443738659)*x[1]+(-0.0481138040854)*x[2]
        arg[1,0,1]=(0.110033706339)*x[0]+(1.47809236549)*x[1]+(0.240737599484)*x[2]
        arg[1,1,0]=(-0.0429437939148)*x[0]+(-1.16945967196)*x[1]+(0.924577053186)*x[2]
        arg[1,1,1]=(-1.57856509244)*x[0]+(-0.460441809594)*x[1]+(-0.792590928973)*x[2]
        arg[2,0,0]=(-1.93020920065)*x[0]+(0.391612451745)*x[1]+(-0.0130238634938)*x[2]
        arg[2,0,1]=(0.0678908982414)*x[0]+(-0.929813889361)*x[1]+(0.727462587818)*x[2]
        arg[2,1,0]=(-0.297494494419)*x[0]+(0.167234033247)*x[1]+(-1.01372084798)*x[2]
        arg[2,1,1]=(-0.608810623174)*x[0]+(1.04629900384)*x[1]+(1.09464953456)*x[2]
        arg[3,0,0]=(1.2996449893)*x[0]+(-0.458197425323)*x[1]+(1.31523663461)*x[2]
        arg[3,0,1]=(-0.848124038387)*x[0]+(0.0136487458675)*x[1]+(1.13573336016)*x[2]
        arg[3,1,0]=(0.377245000633)*x[0]+(-0.766336675857)*x[1]+(0.380031677083)*x[2]
        arg[3,1,1]=(0.158289533163)*x[0]+(-0.149454925108)*x[1]+(-0.59348598427)*x[2]
        arg[4,0,0]=(-1.13101661269)*x[0]+(-0.694062133171)*x[1]+(0.00925221725394)*x[2]
        arg[4,0,1]=(-0.977317473287)*x[0]+(-0.121712752812)*x[1]+(-0.480470720169)*x[2]
        arg[4,1,0]=(1.14884269473)*x[0]+(-0.204958154665)*x[1]+(1.24328207187)*x[2]
        arg[4,1,1]=(0.0791969146134)*x[0]+(0.589873409499)*x[1]+(1.90521993698)*x[2]
        arg[5,0,0]=(1.75625788908)*x[0]+(0.0134922175773)*x[1]+(-0.699884009939)*x[2]
        arg[5,0,1]=(-0.73456900252)*x[0]+(0.923248295457)*x[1]+(1.05161845199)*x[2]
        arg[5,1,0]=(-0.0335528775888)*x[0]+(0.728742831853)*x[1]+(-0.5212350627)*x[2]
        arg[5,1,1]=(0.556540394981)*x[0]+(-0.128428520433)*x[1]+(1.06358498081)*x[2]
        ref[0,0,0]=(-0.640205798541)*x_ref[0]+(0.271670719935)*x_ref[1]+(-0.163000598261)*x_ref[2]
        ref[0,0,1]=(0.735605297981)*x_ref[0]+(-1.05892412676)*x_ref[1]+(-1.1083007441)*x_ref[2]
        ref[0,1,0]=(0.0349869408333)*x_ref[0]+(0.699147831788)*x_ref[1]+(-0.454390826725)*x_ref[2]
        ref[0,1,1]=(-0.703957679243)*x_ref[0]+(-1.63654420176)*x_ref[1]+(-1.13045075795)*x_ref[2]
        ref[1,0,0]=(1.01167296102)*x_ref[0]+(-0.362443738659)*x_ref[1]+(-0.0481138040854)*x_ref[2]
        ref[1,0,1]=(0.110033706339)*x_ref[0]+(1.47809236549)*x_ref[1]+(0.240737599484)*x_ref[2]
        ref[1,1,0]=(-0.0429437939148)*x_ref[0]+(-1.16945967196)*x_ref[1]+(0.924577053186)*x_ref[2]
        ref[1,1,1]=(-1.57856509244)*x_ref[0]+(-0.460441809594)*x_ref[1]+(-0.792590928973)*x_ref[2]
        ref[2,0,0]=(-1.93020920065)*x_ref[0]+(0.391612451745)*x_ref[1]+(-0.0130238634938)*x_ref[2]
        ref[2,0,1]=(0.0678908982414)*x_ref[0]+(-0.929813889361)*x_ref[1]+(0.727462587818)*x_ref[2]
        ref[2,1,0]=(-0.297494494419)*x_ref[0]+(0.167234033247)*x_ref[1]+(-1.01372084798)*x_ref[2]
        ref[2,1,1]=(-0.608810623174)*x_ref[0]+(1.04629900384)*x_ref[1]+(1.09464953456)*x_ref[2]
        ref[3,0,0]=(1.2996449893)*x_ref[0]+(-0.458197425323)*x_ref[1]+(1.31523663461)*x_ref[2]
        ref[3,0,1]=(-0.848124038387)*x_ref[0]+(0.0136487458675)*x_ref[1]+(1.13573336016)*x_ref[2]
        ref[3,1,0]=(0.377245000633)*x_ref[0]+(-0.766336675857)*x_ref[1]+(0.380031677083)*x_ref[2]
        ref[3,1,1]=(0.158289533163)*x_ref[0]+(-0.149454925108)*x_ref[1]+(-0.59348598427)*x_ref[2]
        ref[4,0,0]=(-1.13101661269)*x_ref[0]+(-0.694062133171)*x_ref[1]+(0.00925221725394)*x_ref[2]
        ref[4,0,1]=(-0.977317473287)*x_ref[0]+(-0.121712752812)*x_ref[1]+(-0.480470720169)*x_ref[2]
        ref[4,1,0]=(1.14884269473)*x_ref[0]+(-0.204958154665)*x_ref[1]+(1.24328207187)*x_ref[2]
        ref[4,1,1]=(0.0791969146134)*x_ref[0]+(0.589873409499)*x_ref[1]+(1.90521993698)*x_ref[2]
        ref[5,0,0]=(1.75625788908)*x_ref[0]+(0.0134922175773)*x_ref[1]+(-0.699884009939)*x_ref[2]
        ref[5,0,1]=(-0.73456900252)*x_ref[0]+(0.923248295457)*x_ref[1]+(1.05161845199)*x_ref[2]
        ref[5,1,0]=(-0.0335528775888)*x_ref[0]+(0.728742831853)*x_ref[1]+(-0.5212350627)*x_ref[2]
        ref[5,1,1]=(0.556540394981)*x_ref[0]+(-0.128428520433)*x_ref[1]+(1.06358498081)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.507324804235)*x[0]+(0.725831700089)*x[1]
        arg[0,0,0,1]=(0.847901454936)*x[0]+(0.84787024099)*x[1]
        arg[0,0,1,0]=(0.221488465784)*x[0]+(0.0449286988708)*x[1]
        arg[0,0,1,1]=(0.00897176578501)*x[0]+(-0.799118487206)*x[1]
        arg[0,0,2,0]=(0.171491682599)*x[0]+(1.0259065872)*x[1]
        arg[0,0,2,1]=(-0.213244683798)*x[0]+(-0.706816385516)*x[1]
        arg[0,1,0,0]=(1.54199875186)*x[0]+(-1.09479402082)*x[1]
        arg[0,1,0,1]=(-0.755278051349)*x[0]+(-0.371785089226)*x[1]
        arg[0,1,1,0]=(-0.798931370137)*x[0]+(-0.716047155117)*x[1]
        arg[0,1,1,1]=(-0.79104072521)*x[0]+(-1.45842710768)*x[1]
        arg[0,1,2,0]=(-0.489161301965)*x[0]+(0.115368605994)*x[1]
        arg[0,1,2,1]=(0.00167228618616)*x[0]+(-0.636896504873)*x[1]
        arg[0,2,0,0]=(-0.239182729045)*x[0]+(0.470786327067)*x[1]
        arg[0,2,0,1]=(0.182105943188)*x[0]+(0.328443662342)*x[1]
        arg[0,2,1,0]=(0.000476540430199)*x[0]+(-0.498066512126)*x[1]
        arg[0,2,1,1]=(-0.822985659305)*x[0]+(-0.118645880185)*x[1]
        arg[0,2,2,0]=(1.18375382554)*x[0]+(1.51417307474)*x[1]
        arg[0,2,2,1]=(-0.99755868797)*x[0]+(-1.4187068658)*x[1]
        arg[0,3,0,0]=(-0.0908782314954)*x[0]+(-0.25901340471)*x[1]
        arg[0,3,0,1]=(-1.77626335863)*x[0]+(-0.862132832663)*x[1]
        arg[0,3,1,0]=(0.532147327491)*x[0]+(0.454949026506)*x[1]
        arg[0,3,1,1]=(-0.882307958628)*x[0]+(0.00780544746858)*x[1]
        arg[0,3,2,0]=(0.700104420574)*x[0]+(0.720508036709)*x[1]
        arg[0,3,2,1]=(-1.27167118811)*x[0]+(-0.372999410908)*x[1]
        arg[0,4,0,0]=(0.396222903345)*x[0]+(1.70664799562)*x[1]
        arg[0,4,0,1]=(0.229159055372)*x[0]+(0.0162677758439)*x[1]
        arg[0,4,1,0]=(1.69470499818)*x[0]+(0.0971804891468)*x[1]
        arg[0,4,1,1]=(0.180710573838)*x[0]+(1.05311557127)*x[1]
        arg[0,4,2,0]=(-0.072661496917)*x[0]+(0.466256643644)*x[1]
        arg[0,4,2,1]=(0.25755631084)*x[0]+(0.306990925808)*x[1]
        arg[1,0,0,0]=(0.835353191317)*x[0]+(-0.619237060145)*x[1]
        arg[1,0,0,1]=(1.02966866371)*x[0]+(0.327724003031)*x[1]
        arg[1,0,1,0]=(-0.59214788986)*x[0]+(-0.927738485428)*x[1]
        arg[1,0,1,1]=(-0.805929278446)*x[0]+(-0.366113927406)*x[1]
        arg[1,0,2,0]=(0.952876155721)*x[0]+(0.490531201467)*x[1]
        arg[1,0,2,1]=(1.44341651657)*x[0]+(-0.269899260772)*x[1]
        arg[1,1,0,0]=(0.302587966785)*x[0]+(-0.0373883991695)*x[1]
        arg[1,1,0,1]=(0.77498995648)*x[0]+(1.81052574555)*x[1]
        arg[1,1,1,0]=(1.47950560134)*x[0]+(-0.714878219566)*x[1]
        arg[1,1,1,1]=(0.0713827629606)*x[0]+(-1.22018291576)*x[1]
        arg[1,1,2,0]=(1.30715631604)*x[0]+(-0.191666253513)*x[1]
        arg[1,1,2,1]=(0.330921962628)*x[0]+(-1.57160261912)*x[1]
        arg[1,2,0,0]=(0.593477126597)*x[0]+(0.728196634765)*x[1]
        arg[1,2,0,1]=(0.853860342143)*x[0]+(0.142625101433)*x[1]
        arg[1,2,1,0]=(-0.743590115343)*x[0]+(0.108893736366)*x[1]
        arg[1,2,1,1]=(-0.651514863459)*x[0]+(0.365517180762)*x[1]
        arg[1,2,2,0]=(1.491516719)*x[0]+(0.28900806658)*x[1]
        arg[1,2,2,1]=(1.23733736339)*x[0]+(-0.58438294882)*x[1]
        arg[1,3,0,0]=(0.185696018505)*x[0]+(-1.29119084907)*x[1]
        arg[1,3,0,1]=(1.73584656085)*x[0]+(0.720989518003)*x[1]
        arg[1,3,1,0]=(-1.16044192371)*x[0]+(-0.131128994204)*x[1]
        arg[1,3,1,1]=(-0.00913995119508)*x[0]+(0.506516682557)*x[1]
        arg[1,3,2,0]=(-1.4970171408)*x[0]+(0.692262679219)*x[1]
        arg[1,3,2,1]=(0.433442352395)*x[0]+(-0.118782513258)*x[1]
        arg[1,4,0,0]=(0.186272280804)*x[0]+(-0.136627137097)*x[1]
        arg[1,4,0,1]=(1.25477168515)*x[0]+(-0.0871831369474)*x[1]
        arg[1,4,1,0]=(1.22746022677)*x[0]+(0.799329883559)*x[1]
        arg[1,4,1,1]=(-0.756298280123)*x[0]+(0.317500967634)*x[1]
        arg[1,4,2,0]=(0.201184703435)*x[0]+(-0.116537935236)*x[1]
        arg[1,4,2,1]=(0.851291897607)*x[0]+(0.404063606516)*x[1]
        arg[2,0,0,0]=(-0.257605370026)*x[0]+(-0.202424701671)*x[1]
        arg[2,0,0,1]=(0.0219332720675)*x[0]+(0.518222103227)*x[1]
        arg[2,0,1,0]=(1.03726036144)*x[0]+(-1.01482867596)*x[1]
        arg[2,0,1,1]=(0.00677025785109)*x[0]+(0.607932484906)*x[1]
        arg[2,0,2,0]=(-0.753000974637)*x[0]+(0.112884693676)*x[1]
        arg[2,0,2,1]=(-0.585627190582)*x[0]+(0.548136242722)*x[1]
        arg[2,1,0,0]=(0.382466685883)*x[0]+(-1.15520454298)*x[1]
        arg[2,1,0,1]=(-0.722705123323)*x[0]+(-0.599069291154)*x[1]
        arg[2,1,1,0]=(-0.845885208256)*x[0]+(-1.56066350212)*x[1]
        arg[2,1,1,1]=(-1.31472469553)*x[0]+(0.18973548452)*x[1]
        arg[2,1,2,0]=(-0.419142599733)*x[0]+(1.43873372527)*x[1]
        arg[2,1,2,1]=(0.22685530286)*x[0]+(-0.53636714393)*x[1]
        arg[2,2,0,0]=(1.32884214935)*x[0]+(-0.377575726395)*x[1]
        arg[2,2,0,1]=(0.0845542736847)*x[0]+(0.0888339000429)*x[1]
        arg[2,2,1,0]=(-1.05813285521)*x[0]+(-0.980382395463)*x[1]
        arg[2,2,1,1]=(0.306785463268)*x[0]+(0.622558337967)*x[1]
        arg[2,2,2,0]=(-0.765475308614)*x[0]+(1.40320055807)*x[1]
        arg[2,2,2,1]=(-1.23199081441)*x[0]+(0.245517841221)*x[1]
        arg[2,3,0,0]=(0.655775411743)*x[0]+(-1.09496249546)*x[1]
        arg[2,3,0,1]=(-1.57906394006)*x[0]+(-0.223321547826)*x[1]
        arg[2,3,1,0]=(-0.329934697034)*x[0]+(-1.36492062074)*x[1]
        arg[2,3,1,1]=(1.2584701256)*x[0]+(0.172958559745)*x[1]
        arg[2,3,2,0]=(0.265769150885)*x[0]+(-0.273148332255)*x[1]
        arg[2,3,2,1]=(-1.18294356647)*x[0]+(-0.269813147795)*x[1]
        arg[2,4,0,0]=(-1.80352122024)*x[0]+(-0.417693287507)*x[1]
        arg[2,4,0,1]=(0.829979623343)*x[0]+(0.0526239642579)*x[1]
        arg[2,4,1,0]=(0.967999103063)*x[0]+(0.0184170526665)*x[1]
        arg[2,4,1,1]=(-0.326938715134)*x[0]+(-0.575040647639)*x[1]
        arg[2,4,2,0]=(1.44727400099)*x[0]+(0.710269500685)*x[1]
        arg[2,4,2,1]=(-0.248166665809)*x[0]+(0.247820501661)*x[1]
        arg[3,0,0,0]=(0.0463415093786)*x[0]+(0.650085996364)*x[1]
        arg[3,0,0,1]=(-0.440939509697)*x[0]+(-0.323307669473)*x[1]
        arg[3,0,1,0]=(0.273356160699)*x[0]+(-0.66721600197)*x[1]
        arg[3,0,1,1]=(1.29465477714)*x[0]+(-0.37486215111)*x[1]
        arg[3,0,2,0]=(-0.809896216106)*x[0]+(-1.51076103892)*x[1]
        arg[3,0,2,1]=(0.686428827303)*x[0]+(-0.538164742287)*x[1]
        arg[3,1,0,0]=(-0.863651610678)*x[0]+(0.673803453645)*x[1]
        arg[3,1,0,1]=(0.526827519816)*x[0]+(-0.0379595314201)*x[1]
        arg[3,1,1,0]=(-1.04759557033)*x[0]+(0.114198098457)*x[1]
        arg[3,1,1,1]=(0.552205171704)*x[0]+(0.116364312378)*x[1]
        arg[3,1,2,0]=(-0.926549911354)*x[0]+(0.296927704155)*x[1]
        arg[3,1,2,1]=(0.46702104365)*x[0]+(0.355015968814)*x[1]
        arg[3,2,0,0]=(0.603177046544)*x[0]+(-0.600600764625)*x[1]
        arg[3,2,0,1]=(0.656827398088)*x[0]+(1.19922812106)*x[1]
        arg[3,2,1,0]=(-0.0185352951675)*x[0]+(-0.575006610999)*x[1]
        arg[3,2,1,1]=(0.954361462974)*x[0]+(-0.239751892125)*x[1]
        arg[3,2,2,0]=(0.880568376655)*x[0]+(1.46879509876)*x[1]
        arg[3,2,2,1]=(-1.19051913627)*x[0]+(-0.0387119361514)*x[1]
        arg[3,3,0,0]=(1.19520855834)*x[0]+(1.3515728248)*x[1]
        arg[3,3,0,1]=(0.695487159903)*x[0]+(0.783460154378)*x[1]
        arg[3,3,1,0]=(0.482352542106)*x[0]+(-1.321707654)*x[1]
        arg[3,3,1,1]=(0.0637384915311)*x[0]+(-0.5903067787)*x[1]
        arg[3,3,2,0]=(0.571974795979)*x[0]+(1.33251727342)*x[1]
        arg[3,3,2,1]=(0.579723470137)*x[0]+(-0.260885275522)*x[1]
        arg[3,4,0,0]=(0.0717612359082)*x[0]+(-0.576982745638)*x[1]
        arg[3,4,0,1]=(-0.227660227523)*x[0]+(0.232974094548)*x[1]
        arg[3,4,1,0]=(0.767797269581)*x[0]+(0.98857043831)*x[1]
        arg[3,4,1,1]=(0.31200408415)*x[0]+(-0.8317658048)*x[1]
        arg[3,4,2,0]=(0.381688388928)*x[0]+(-0.335620162467)*x[1]
        arg[3,4,2,1]=(0.781228510591)*x[0]+(0.960966953643)*x[1]
        ref[0,0,0,0]=(0.507324804235)*x_ref[0]+(0.725831700089)*x_ref[1]
        ref[0,0,0,1]=(0.847901454936)*x_ref[0]+(0.84787024099)*x_ref[1]
        ref[0,0,1,0]=(0.221488465784)*x_ref[0]+(0.0449286988708)*x_ref[1]
        ref[0,0,1,1]=(0.00897176578501)*x_ref[0]+(-0.799118487206)*x_ref[1]
        ref[0,0,2,0]=(0.171491682599)*x_ref[0]+(1.0259065872)*x_ref[1]
        ref[0,0,2,1]=(-0.213244683798)*x_ref[0]+(-0.706816385516)*x_ref[1]
        ref[0,1,0,0]=(1.54199875186)*x_ref[0]+(-1.09479402082)*x_ref[1]
        ref[0,1,0,1]=(-0.755278051349)*x_ref[0]+(-0.371785089226)*x_ref[1]
        ref[0,1,1,0]=(-0.798931370137)*x_ref[0]+(-0.716047155117)*x_ref[1]
        ref[0,1,1,1]=(-0.79104072521)*x_ref[0]+(-1.45842710768)*x_ref[1]
        ref[0,1,2,0]=(-0.489161301965)*x_ref[0]+(0.115368605994)*x_ref[1]
        ref[0,1,2,1]=(0.00167228618616)*x_ref[0]+(-0.636896504873)*x_ref[1]
        ref[0,2,0,0]=(-0.239182729045)*x_ref[0]+(0.470786327067)*x_ref[1]
        ref[0,2,0,1]=(0.182105943188)*x_ref[0]+(0.328443662342)*x_ref[1]
        ref[0,2,1,0]=(0.000476540430199)*x_ref[0]+(-0.498066512126)*x_ref[1]
        ref[0,2,1,1]=(-0.822985659305)*x_ref[0]+(-0.118645880185)*x_ref[1]
        ref[0,2,2,0]=(1.18375382554)*x_ref[0]+(1.51417307474)*x_ref[1]
        ref[0,2,2,1]=(-0.99755868797)*x_ref[0]+(-1.4187068658)*x_ref[1]
        ref[0,3,0,0]=(-0.0908782314954)*x_ref[0]+(-0.25901340471)*x_ref[1]
        ref[0,3,0,1]=(-1.77626335863)*x_ref[0]+(-0.862132832663)*x_ref[1]
        ref[0,3,1,0]=(0.532147327491)*x_ref[0]+(0.454949026506)*x_ref[1]
        ref[0,3,1,1]=(-0.882307958628)*x_ref[0]+(0.00780544746858)*x_ref[1]
        ref[0,3,2,0]=(0.700104420574)*x_ref[0]+(0.720508036709)*x_ref[1]
        ref[0,3,2,1]=(-1.27167118811)*x_ref[0]+(-0.372999410908)*x_ref[1]
        ref[0,4,0,0]=(0.396222903345)*x_ref[0]+(1.70664799562)*x_ref[1]
        ref[0,4,0,1]=(0.229159055372)*x_ref[0]+(0.0162677758439)*x_ref[1]
        ref[0,4,1,0]=(1.69470499818)*x_ref[0]+(0.0971804891468)*x_ref[1]
        ref[0,4,1,1]=(0.180710573838)*x_ref[0]+(1.05311557127)*x_ref[1]
        ref[0,4,2,0]=(-0.072661496917)*x_ref[0]+(0.466256643644)*x_ref[1]
        ref[0,4,2,1]=(0.25755631084)*x_ref[0]+(0.306990925808)*x_ref[1]
        ref[1,0,0,0]=(0.835353191317)*x_ref[0]+(-0.619237060145)*x_ref[1]
        ref[1,0,0,1]=(1.02966866371)*x_ref[0]+(0.327724003031)*x_ref[1]
        ref[1,0,1,0]=(-0.59214788986)*x_ref[0]+(-0.927738485428)*x_ref[1]
        ref[1,0,1,1]=(-0.805929278446)*x_ref[0]+(-0.366113927406)*x_ref[1]
        ref[1,0,2,0]=(0.952876155721)*x_ref[0]+(0.490531201467)*x_ref[1]
        ref[1,0,2,1]=(1.44341651657)*x_ref[0]+(-0.269899260772)*x_ref[1]
        ref[1,1,0,0]=(0.302587966785)*x_ref[0]+(-0.0373883991695)*x_ref[1]
        ref[1,1,0,1]=(0.77498995648)*x_ref[0]+(1.81052574555)*x_ref[1]
        ref[1,1,1,0]=(1.47950560134)*x_ref[0]+(-0.714878219566)*x_ref[1]
        ref[1,1,1,1]=(0.0713827629606)*x_ref[0]+(-1.22018291576)*x_ref[1]
        ref[1,1,2,0]=(1.30715631604)*x_ref[0]+(-0.191666253513)*x_ref[1]
        ref[1,1,2,1]=(0.330921962628)*x_ref[0]+(-1.57160261912)*x_ref[1]
        ref[1,2,0,0]=(0.593477126597)*x_ref[0]+(0.728196634765)*x_ref[1]
        ref[1,2,0,1]=(0.853860342143)*x_ref[0]+(0.142625101433)*x_ref[1]
        ref[1,2,1,0]=(-0.743590115343)*x_ref[0]+(0.108893736366)*x_ref[1]
        ref[1,2,1,1]=(-0.651514863459)*x_ref[0]+(0.365517180762)*x_ref[1]
        ref[1,2,2,0]=(1.491516719)*x_ref[0]+(0.28900806658)*x_ref[1]
        ref[1,2,2,1]=(1.23733736339)*x_ref[0]+(-0.58438294882)*x_ref[1]
        ref[1,3,0,0]=(0.185696018505)*x_ref[0]+(-1.29119084907)*x_ref[1]
        ref[1,3,0,1]=(1.73584656085)*x_ref[0]+(0.720989518003)*x_ref[1]
        ref[1,3,1,0]=(-1.16044192371)*x_ref[0]+(-0.131128994204)*x_ref[1]
        ref[1,3,1,1]=(-0.00913995119508)*x_ref[0]+(0.506516682557)*x_ref[1]
        ref[1,3,2,0]=(-1.4970171408)*x_ref[0]+(0.692262679219)*x_ref[1]
        ref[1,3,2,1]=(0.433442352395)*x_ref[0]+(-0.118782513258)*x_ref[1]
        ref[1,4,0,0]=(0.186272280804)*x_ref[0]+(-0.136627137097)*x_ref[1]
        ref[1,4,0,1]=(1.25477168515)*x_ref[0]+(-0.0871831369474)*x_ref[1]
        ref[1,4,1,0]=(1.22746022677)*x_ref[0]+(0.799329883559)*x_ref[1]
        ref[1,4,1,1]=(-0.756298280123)*x_ref[0]+(0.317500967634)*x_ref[1]
        ref[1,4,2,0]=(0.201184703435)*x_ref[0]+(-0.116537935236)*x_ref[1]
        ref[1,4,2,1]=(0.851291897607)*x_ref[0]+(0.404063606516)*x_ref[1]
        ref[2,0,0,0]=(-0.257605370026)*x_ref[0]+(-0.202424701671)*x_ref[1]
        ref[2,0,0,1]=(0.0219332720675)*x_ref[0]+(0.518222103227)*x_ref[1]
        ref[2,0,1,0]=(1.03726036144)*x_ref[0]+(-1.01482867596)*x_ref[1]
        ref[2,0,1,1]=(0.00677025785109)*x_ref[0]+(0.607932484906)*x_ref[1]
        ref[2,0,2,0]=(-0.753000974637)*x_ref[0]+(0.112884693676)*x_ref[1]
        ref[2,0,2,1]=(-0.585627190582)*x_ref[0]+(0.548136242722)*x_ref[1]
        ref[2,1,0,0]=(0.382466685883)*x_ref[0]+(-1.15520454298)*x_ref[1]
        ref[2,1,0,1]=(-0.722705123323)*x_ref[0]+(-0.599069291154)*x_ref[1]
        ref[2,1,1,0]=(-0.845885208256)*x_ref[0]+(-1.56066350212)*x_ref[1]
        ref[2,1,1,1]=(-1.31472469553)*x_ref[0]+(0.18973548452)*x_ref[1]
        ref[2,1,2,0]=(-0.419142599733)*x_ref[0]+(1.43873372527)*x_ref[1]
        ref[2,1,2,1]=(0.22685530286)*x_ref[0]+(-0.53636714393)*x_ref[1]
        ref[2,2,0,0]=(1.32884214935)*x_ref[0]+(-0.377575726395)*x_ref[1]
        ref[2,2,0,1]=(0.0845542736847)*x_ref[0]+(0.0888339000429)*x_ref[1]
        ref[2,2,1,0]=(-1.05813285521)*x_ref[0]+(-0.980382395463)*x_ref[1]
        ref[2,2,1,1]=(0.306785463268)*x_ref[0]+(0.622558337967)*x_ref[1]
        ref[2,2,2,0]=(-0.765475308614)*x_ref[0]+(1.40320055807)*x_ref[1]
        ref[2,2,2,1]=(-1.23199081441)*x_ref[0]+(0.245517841221)*x_ref[1]
        ref[2,3,0,0]=(0.655775411743)*x_ref[0]+(-1.09496249546)*x_ref[1]
        ref[2,3,0,1]=(-1.57906394006)*x_ref[0]+(-0.223321547826)*x_ref[1]
        ref[2,3,1,0]=(-0.329934697034)*x_ref[0]+(-1.36492062074)*x_ref[1]
        ref[2,3,1,1]=(1.2584701256)*x_ref[0]+(0.172958559745)*x_ref[1]
        ref[2,3,2,0]=(0.265769150885)*x_ref[0]+(-0.273148332255)*x_ref[1]
        ref[2,3,2,1]=(-1.18294356647)*x_ref[0]+(-0.269813147795)*x_ref[1]
        ref[2,4,0,0]=(-1.80352122024)*x_ref[0]+(-0.417693287507)*x_ref[1]
        ref[2,4,0,1]=(0.829979623343)*x_ref[0]+(0.0526239642579)*x_ref[1]
        ref[2,4,1,0]=(0.967999103063)*x_ref[0]+(0.0184170526665)*x_ref[1]
        ref[2,4,1,1]=(-0.326938715134)*x_ref[0]+(-0.575040647639)*x_ref[1]
        ref[2,4,2,0]=(1.44727400099)*x_ref[0]+(0.710269500685)*x_ref[1]
        ref[2,4,2,1]=(-0.248166665809)*x_ref[0]+(0.247820501661)*x_ref[1]
        ref[3,0,0,0]=(0.0463415093786)*x_ref[0]+(0.650085996364)*x_ref[1]
        ref[3,0,0,1]=(-0.440939509697)*x_ref[0]+(-0.323307669473)*x_ref[1]
        ref[3,0,1,0]=(0.273356160699)*x_ref[0]+(-0.66721600197)*x_ref[1]
        ref[3,0,1,1]=(1.29465477714)*x_ref[0]+(-0.37486215111)*x_ref[1]
        ref[3,0,2,0]=(-0.809896216106)*x_ref[0]+(-1.51076103892)*x_ref[1]
        ref[3,0,2,1]=(0.686428827303)*x_ref[0]+(-0.538164742287)*x_ref[1]
        ref[3,1,0,0]=(-0.863651610678)*x_ref[0]+(0.673803453645)*x_ref[1]
        ref[3,1,0,1]=(0.526827519816)*x_ref[0]+(-0.0379595314201)*x_ref[1]
        ref[3,1,1,0]=(-1.04759557033)*x_ref[0]+(0.114198098457)*x_ref[1]
        ref[3,1,1,1]=(0.552205171704)*x_ref[0]+(0.116364312378)*x_ref[1]
        ref[3,1,2,0]=(-0.926549911354)*x_ref[0]+(0.296927704155)*x_ref[1]
        ref[3,1,2,1]=(0.46702104365)*x_ref[0]+(0.355015968814)*x_ref[1]
        ref[3,2,0,0]=(0.603177046544)*x_ref[0]+(-0.600600764625)*x_ref[1]
        ref[3,2,0,1]=(0.656827398088)*x_ref[0]+(1.19922812106)*x_ref[1]
        ref[3,2,1,0]=(-0.0185352951675)*x_ref[0]+(-0.575006610999)*x_ref[1]
        ref[3,2,1,1]=(0.954361462974)*x_ref[0]+(-0.239751892125)*x_ref[1]
        ref[3,2,2,0]=(0.880568376655)*x_ref[0]+(1.46879509876)*x_ref[1]
        ref[3,2,2,1]=(-1.19051913627)*x_ref[0]+(-0.0387119361514)*x_ref[1]
        ref[3,3,0,0]=(1.19520855834)*x_ref[0]+(1.3515728248)*x_ref[1]
        ref[3,3,0,1]=(0.695487159903)*x_ref[0]+(0.783460154378)*x_ref[1]
        ref[3,3,1,0]=(0.482352542106)*x_ref[0]+(-1.321707654)*x_ref[1]
        ref[3,3,1,1]=(0.0637384915311)*x_ref[0]+(-0.5903067787)*x_ref[1]
        ref[3,3,2,0]=(0.571974795979)*x_ref[0]+(1.33251727342)*x_ref[1]
        ref[3,3,2,1]=(0.579723470137)*x_ref[0]+(-0.260885275522)*x_ref[1]
        ref[3,4,0,0]=(0.0717612359082)*x_ref[0]+(-0.576982745638)*x_ref[1]
        ref[3,4,0,1]=(-0.227660227523)*x_ref[0]+(0.232974094548)*x_ref[1]
        ref[3,4,1,0]=(0.767797269581)*x_ref[0]+(0.98857043831)*x_ref[1]
        ref[3,4,1,1]=(0.31200408415)*x_ref[0]+(-0.8317658048)*x_ref[1]
        ref[3,4,2,0]=(0.381688388928)*x_ref[0]+(-0.335620162467)*x_ref[1]
        ref[3,4,2,1]=(0.781228510591)*x_ref[0]+(0.960966953643)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.875410081875)*x[0]+(0.859961622438)*x[1]+(0.716805919947)*x[2]
        arg[0,0,0,1]=(0.343337453662)*x[0]+(0.802028072985)*x[1]+(0.588370750348)*x[2]
        arg[0,0,1,0]=(0.467615640242)*x[0]+(0.0457804028001)*x[1]+(-0.163677176662)*x[2]
        arg[0,0,1,1]=(-0.46378923457)*x[0]+(1.29379360824)*x[1]+(0.183839231748)*x[2]
        arg[0,0,2,0]=(1.65864647053)*x[0]+(1.23523675105)*x[1]+(-0.556155025527)*x[2]
        arg[0,0,2,1]=(-0.41581445233)*x[0]+(-0.0709880820438)*x[1]+(0.45785589934)*x[2]
        arg[0,1,0,0]=(1.83790068156)*x[0]+(-0.1690836925)*x[1]+(1.25356286366)*x[2]
        arg[0,1,0,1]=(-0.22584314958)*x[0]+(0.684595424346)*x[1]+(1.13919885217)*x[2]
        arg[0,1,1,0]=(1.02792150528)*x[0]+(-0.312817626145)*x[1]+(-0.121465758957)*x[2]
        arg[0,1,1,1]=(0.533491564413)*x[0]+(-0.62208812136)*x[1]+(-0.717947781071)*x[2]
        arg[0,1,2,0]=(-0.864505701555)*x[0]+(-0.296047724934)*x[1]+(-0.122917018588)*x[2]
        arg[0,1,2,1]=(1.05520768182)*x[0]+(0.662053590424)*x[1]+(0.0251841065476)*x[2]
        arg[0,2,0,0]=(-0.408305818721)*x[0]+(-0.355992587618)*x[1]+(-0.075231462678)*x[2]
        arg[0,2,0,1]=(0.926563898549)*x[0]+(0.203358704647)*x[1]+(-0.365269987808)*x[2]
        arg[0,2,1,0]=(-0.403860257745)*x[0]+(-0.0312835722524)*x[1]+(0.241379582617)*x[2]
        arg[0,2,1,1]=(-0.287297815371)*x[0]+(-0.381813267253)*x[1]+(0.46047051318)*x[2]
        arg[0,2,2,0]=(-0.168164506906)*x[0]+(-1.24152426841)*x[1]+(-0.629704769515)*x[2]
        arg[0,2,2,1]=(-0.937219354291)*x[0]+(-0.736950161299)*x[1]+(0.284057170092)*x[2]
        arg[0,3,0,0]=(-0.292205445759)*x[0]+(1.21842674086)*x[1]+(0.608979796402)*x[2]
        arg[0,3,0,1]=(0.53232240924)*x[0]+(0.226439017439)*x[1]+(-0.42603939944)*x[2]
        arg[0,3,1,0]=(1.67410358342)*x[0]+(-0.00813619771347)*x[1]+(-0.463361889732)*x[2]
        arg[0,3,1,1]=(1.3991502461)*x[0]+(0.239999857694)*x[1]+(-0.369907067208)*x[2]
        arg[0,3,2,0]=(-0.248767913403)*x[0]+(0.893584158229)*x[1]+(1.97656682548)*x[2]
        arg[0,3,2,1]=(-0.835046144341)*x[0]+(-0.00979850832247)*x[1]+(-0.281439904932)*x[2]
        arg[0,4,0,0]=(0.974179494371)*x[0]+(-0.0432121220464)*x[1]+(-0.475018088554)*x[2]
        arg[0,4,0,1]=(0.245490151034)*x[0]+(-0.367572047585)*x[1]+(-0.926123425123)*x[2]
        arg[0,4,1,0]=(-0.239057491277)*x[0]+(0.339734302574)*x[1]+(0.760546363123)*x[2]
        arg[0,4,1,1]=(-1.0425426898)*x[0]+(0.774210702874)*x[1]+(-1.15990543484)*x[2]
        arg[0,4,2,0]=(1.02994063274)*x[0]+(1.69313195574)*x[1]+(-1.02468862318)*x[2]
        arg[0,4,2,1]=(0.221805673731)*x[0]+(-1.06588014758)*x[1]+(-0.546973870537)*x[2]
        arg[1,0,0,0]=(-0.140765393552)*x[0]+(-0.309057115268)*x[1]+(0.07814948965)*x[2]
        arg[1,0,0,1]=(-0.29622101259)*x[0]+(-1.0654452087)*x[1]+(1.77500323701)*x[2]
        arg[1,0,1,0]=(1.13202947738)*x[0]+(-0.515454892388)*x[1]+(-1.34230346013)*x[2]
        arg[1,0,1,1]=(1.50210238409)*x[0]+(-0.75673088573)*x[1]+(-0.792078632258)*x[2]
        arg[1,0,2,0]=(0.249111383954)*x[0]+(-0.972418007776)*x[1]+(0.766060429469)*x[2]
        arg[1,0,2,1]=(-0.0710717292409)*x[0]+(0.164949706666)*x[1]+(1.46045331387)*x[2]
        arg[1,1,0,0]=(-0.587668209291)*x[0]+(0.303249169818)*x[1]+(0.352197808782)*x[2]
        arg[1,1,0,1]=(-0.145609904248)*x[0]+(0.91895760024)*x[1]+(-1.25063091478)*x[2]
        arg[1,1,1,0]=(0.450869946651)*x[0]+(0.752704827905)*x[1]+(0.132042943533)*x[2]
        arg[1,1,1,1]=(-0.87296460294)*x[0]+(-0.459840609019)*x[1]+(0.422574948491)*x[2]
        arg[1,1,2,0]=(0.828459269131)*x[0]+(0.45869153936)*x[1]+(0.854640684176)*x[2]
        arg[1,1,2,1]=(1.84720399473)*x[0]+(0.0416642546234)*x[1]+(-0.309583524852)*x[2]
        arg[1,2,0,0]=(-1.12563293246)*x[0]+(0.0142184473972)*x[1]+(-1.18150362584)*x[2]
        arg[1,2,0,1]=(-0.0750965768147)*x[0]+(-0.419276687751)*x[1]+(0.521012213863)*x[2]
        arg[1,2,1,0]=(-1.06905807284)*x[0]+(0.692258030743)*x[1]+(0.262254909068)*x[2]
        arg[1,2,1,1]=(0.0552441016888)*x[0]+(-0.131040560322)*x[1]+(-0.321517725963)*x[2]
        arg[1,2,2,0]=(-0.740410514693)*x[0]+(0.219020143899)*x[1]+(1.35104682965)*x[2]
        arg[1,2,2,1]=(-1.38198583939)*x[0]+(-0.19342887842)*x[1]+(-0.780008129494)*x[2]
        arg[1,3,0,0]=(0.378358305605)*x[0]+(1.32138618734)*x[1]+(0.627209784997)*x[2]
        arg[1,3,0,1]=(-0.0383657917537)*x[0]+(-1.41770189628)*x[1]+(-1.21625280681)*x[2]
        arg[1,3,1,0]=(0.819666413875)*x[0]+(0.160369878383)*x[1]+(-0.229388558682)*x[2]
        arg[1,3,1,1]=(0.393409555876)*x[0]+(-0.525985341296)*x[1]+(-1.11892438309)*x[2]
        arg[1,3,2,0]=(0.145077010777)*x[0]+(0.229800008438)*x[1]+(-0.202652862622)*x[2]
        arg[1,3,2,1]=(-1.48698378177)*x[0]+(0.553698672636)*x[1]+(-0.427263736796)*x[2]
        arg[1,4,0,0]=(-0.129527725249)*x[0]+(-0.639824801211)*x[1]+(0.759081980717)*x[2]
        arg[1,4,0,1]=(-0.117772473784)*x[0]+(0.387474104611)*x[1]+(0.507306747135)*x[2]
        arg[1,4,1,0]=(-0.82518622198)*x[0]+(-0.754474320343)*x[1]+(-0.978049152511)*x[2]
        arg[1,4,1,1]=(-0.370385633483)*x[0]+(-0.348429423043)*x[1]+(-0.254278257899)*x[2]
        arg[1,4,2,0]=(1.42596167697)*x[0]+(-0.871785904514)*x[1]+(-0.5343959325)*x[2]
        arg[1,4,2,1]=(0.817877672506)*x[0]+(-0.222622791737)*x[1]+(-0.677237468595)*x[2]
        arg[2,0,0,0]=(0.877660484813)*x[0]+(0.818663846913)*x[1]+(-1.26520874625)*x[2]
        arg[2,0,0,1]=(-0.822344385073)*x[0]+(-0.274691696266)*x[1]+(-0.35079248697)*x[2]
        arg[2,0,1,0]=(-0.319040285312)*x[0]+(-0.750406131559)*x[1]+(-0.920332967032)*x[2]
        arg[2,0,1,1]=(1.48758278445)*x[0]+(1.30587557873)*x[1]+(0.230692557888)*x[2]
        arg[2,0,2,0]=(1.44269354043)*x[0]+(-0.63270356894)*x[1]+(0.229323632455)*x[2]
        arg[2,0,2,1]=(-0.645065877536)*x[0]+(-0.393083499676)*x[1]+(-0.305117645453)*x[2]
        arg[2,1,0,0]=(-0.63304134143)*x[0]+(0.0664768401917)*x[1]+(0.87587743066)*x[2]
        arg[2,1,0,1]=(-0.531856958285)*x[0]+(0.747953146648)*x[1]+(1.23560779546)*x[2]
        arg[2,1,1,0]=(0.0564479101025)*x[0]+(0.546307803332)*x[1]+(0.655044890368)*x[2]
        arg[2,1,1,1]=(-1.12793057228)*x[0]+(-0.736676194013)*x[1]+(0.0375913829411)*x[2]
        arg[2,1,2,0]=(-0.531384188067)*x[0]+(0.249416904081)*x[1]+(0.711430712986)*x[2]
        arg[2,1,2,1]=(0.884244664065)*x[0]+(0.456066941968)*x[1]+(0.181630861785)*x[2]
        arg[2,2,0,0]=(0.0844543345062)*x[0]+(-0.435288041256)*x[1]+(-0.434816763062)*x[2]
        arg[2,2,0,1]=(0.373388105049)*x[0]+(0.239833846429)*x[1]+(-0.437771573282)*x[2]
        arg[2,2,1,0]=(0.534401521907)*x[0]+(0.393077203448)*x[1]+(-0.42440387717)*x[2]
        arg[2,2,1,1]=(0.318683369617)*x[0]+(-0.538864643793)*x[1]+(1.83466945806)*x[2]
        arg[2,2,2,0]=(-1.79048122074)*x[0]+(-1.05441213478)*x[1]+(-0.573963806281)*x[2]
        arg[2,2,2,1]=(0.944017405182)*x[0]+(0.772756791538)*x[1]+(0.0269673237336)*x[2]
        arg[2,3,0,0]=(0.989703689142)*x[0]+(1.21302325945)*x[1]+(-0.92209419073)*x[2]
        arg[2,3,0,1]=(-0.0957244757576)*x[0]+(1.50132638315)*x[1]+(-1.38716099756)*x[2]
        arg[2,3,1,0]=(1.6445441776)*x[0]+(-0.490207070694)*x[1]+(0.626469347781)*x[2]
        arg[2,3,1,1]=(0.0185568939147)*x[0]+(0.412024752941)*x[1]+(-0.22626102813)*x[2]
        arg[2,3,2,0]=(0.0143609822966)*x[0]+(0.115739025893)*x[1]+(-0.544468631925)*x[2]
        arg[2,3,2,1]=(-1.80522341493)*x[0]+(0.33352152695)*x[1]+(-0.841286464932)*x[2]
        arg[2,4,0,0]=(0.0618444122042)*x[0]+(1.91232484622)*x[1]+(-0.406317200023)*x[2]
        arg[2,4,0,1]=(0.279131062922)*x[0]+(0.0194232637634)*x[1]+(-0.00179069113634)*x[2]
        arg[2,4,1,0]=(-0.701344579135)*x[0]+(-0.27410935473)*x[1]+(0.763362100733)*x[2]
        arg[2,4,1,1]=(0.831198601185)*x[0]+(0.90336781771)*x[1]+(1.21856633382)*x[2]
        arg[2,4,2,0]=(0.526278615033)*x[0]+(-1.08673855801)*x[1]+(0.0784355900913)*x[2]
        arg[2,4,2,1]=(0.853226856851)*x[0]+(-1.15919062627)*x[1]+(0.185686031732)*x[2]
        arg[3,0,0,0]=(0.272127498117)*x[0]+(-0.457140162438)*x[1]+(0.318358670541)*x[2]
        arg[3,0,0,1]=(1.71056378307)*x[0]+(0.200052730234)*x[1]+(-0.127733325792)*x[2]
        arg[3,0,1,0]=(-0.285340963261)*x[0]+(-0.524544624385)*x[1]+(0.916366474944)*x[2]
        arg[3,0,1,1]=(-0.625506557557)*x[0]+(1.0858476479)*x[1]+(-1.24478983268)*x[2]
        arg[3,0,2,0]=(-1.37299619916)*x[0]+(-0.482246528202)*x[1]+(-1.36857669075)*x[2]
        arg[3,0,2,1]=(-0.924421844916)*x[0]+(-1.50402098115)*x[1]+(1.00045130673)*x[2]
        arg[3,1,0,0]=(0.0417304700827)*x[0]+(0.554527451004)*x[1]+(0.299758163628)*x[2]
        arg[3,1,0,1]=(0.639611639279)*x[0]+(-0.444675638162)*x[1]+(1.43150046757)*x[2]
        arg[3,1,1,0]=(0.288697220898)*x[0]+(1.32660433773)*x[1]+(0.0889412116759)*x[2]
        arg[3,1,1,1]=(-0.110516770201)*x[0]+(-0.0214248733565)*x[1]+(-0.758638794785)*x[2]
        arg[3,1,2,0]=(1.34357138654)*x[0]+(-0.13759272972)*x[1]+(-0.00183845032784)*x[2]
        arg[3,1,2,1]=(1.44174769627)*x[0]+(0.49190176388)*x[1]+(0.477331360141)*x[2]
        arg[3,2,0,0]=(-0.10284777121)*x[0]+(-0.63294157419)*x[1]+(0.246064460451)*x[2]
        arg[3,2,0,1]=(0.974270980699)*x[0]+(1.03835886161)*x[1]+(0.420000854838)*x[2]
        arg[3,2,1,0]=(-0.00617831615923)*x[0]+(-0.354063634073)*x[1]+(0.764285884619)*x[2]
        arg[3,2,1,1]=(0.210815790746)*x[0]+(0.76609515744)*x[1]+(-0.564401558401)*x[2]
        arg[3,2,2,0]=(1.30471234011)*x[0]+(1.54864119332)*x[1]+(-0.493583650519)*x[2]
        arg[3,2,2,1]=(0.42287513668)*x[0]+(-0.603093868171)*x[1]+(-0.811892974203)*x[2]
        arg[3,3,0,0]=(-1.38358134693)*x[0]+(-1.53706018098)*x[1]+(-1.29701227501)*x[2]
        arg[3,3,0,1]=(0.598581779902)*x[0]+(1.02424010922)*x[1]+(-1.1814313009)*x[2]
        arg[3,3,1,0]=(1.30938529116)*x[0]+(-0.504468477244)*x[1]+(1.63655848854)*x[2]
        arg[3,3,1,1]=(-1.47006424113)*x[0]+(0.225951520203)*x[1]+(-0.412273003169)*x[2]
        arg[3,3,2,0]=(-0.116902234482)*x[0]+(0.488144536044)*x[1]+(1.01732461694)*x[2]
        arg[3,3,2,1]=(1.0935814877)*x[0]+(1.06389385916)*x[1]+(-0.17981400591)*x[2]
        arg[3,4,0,0]=(-0.833265062101)*x[0]+(-0.754461148648)*x[1]+(1.26702012815)*x[2]
        arg[3,4,0,1]=(-0.793744278171)*x[0]+(0.112704521653)*x[1]+(-0.448684922187)*x[2]
        arg[3,4,1,0]=(-0.396153523625)*x[0]+(-0.0370870723568)*x[1]+(0.554347618511)*x[2]
        arg[3,4,1,1]=(-0.323276139711)*x[0]+(-0.397389928392)*x[1]+(-1.1707028324)*x[2]
        arg[3,4,2,0]=(0.823350512513)*x[0]+(1.47105759864)*x[1]+(-0.977456877855)*x[2]
        arg[3,4,2,1]=(-0.719791170741)*x[0]+(-1.00363340645)*x[1]+(0.814612770572)*x[2]
        ref[0,0,0,0]=(-0.875410081875)*x_ref[0]+(0.859961622438)*x_ref[1]+(0.716805919947)*x_ref[2]
        ref[0,0,0,1]=(0.343337453662)*x_ref[0]+(0.802028072985)*x_ref[1]+(0.588370750348)*x_ref[2]
        ref[0,0,1,0]=(0.467615640242)*x_ref[0]+(0.0457804028001)*x_ref[1]+(-0.163677176662)*x_ref[2]
        ref[0,0,1,1]=(-0.46378923457)*x_ref[0]+(1.29379360824)*x_ref[1]+(0.183839231748)*x_ref[2]
        ref[0,0,2,0]=(1.65864647053)*x_ref[0]+(1.23523675105)*x_ref[1]+(-0.556155025527)*x_ref[2]
        ref[0,0,2,1]=(-0.41581445233)*x_ref[0]+(-0.0709880820438)*x_ref[1]+(0.45785589934)*x_ref[2]
        ref[0,1,0,0]=(1.83790068156)*x_ref[0]+(-0.1690836925)*x_ref[1]+(1.25356286366)*x_ref[2]
        ref[0,1,0,1]=(-0.22584314958)*x_ref[0]+(0.684595424346)*x_ref[1]+(1.13919885217)*x_ref[2]
        ref[0,1,1,0]=(1.02792150528)*x_ref[0]+(-0.312817626145)*x_ref[1]+(-0.121465758957)*x_ref[2]
        ref[0,1,1,1]=(0.533491564413)*x_ref[0]+(-0.62208812136)*x_ref[1]+(-0.717947781071)*x_ref[2]
        ref[0,1,2,0]=(-0.864505701555)*x_ref[0]+(-0.296047724934)*x_ref[1]+(-0.122917018588)*x_ref[2]
        ref[0,1,2,1]=(1.05520768182)*x_ref[0]+(0.662053590424)*x_ref[1]+(0.0251841065476)*x_ref[2]
        ref[0,2,0,0]=(-0.408305818721)*x_ref[0]+(-0.355992587618)*x_ref[1]+(-0.075231462678)*x_ref[2]
        ref[0,2,0,1]=(0.926563898549)*x_ref[0]+(0.203358704647)*x_ref[1]+(-0.365269987808)*x_ref[2]
        ref[0,2,1,0]=(-0.403860257745)*x_ref[0]+(-0.0312835722524)*x_ref[1]+(0.241379582617)*x_ref[2]
        ref[0,2,1,1]=(-0.287297815371)*x_ref[0]+(-0.381813267253)*x_ref[1]+(0.46047051318)*x_ref[2]
        ref[0,2,2,0]=(-0.168164506906)*x_ref[0]+(-1.24152426841)*x_ref[1]+(-0.629704769515)*x_ref[2]
        ref[0,2,2,1]=(-0.937219354291)*x_ref[0]+(-0.736950161299)*x_ref[1]+(0.284057170092)*x_ref[2]
        ref[0,3,0,0]=(-0.292205445759)*x_ref[0]+(1.21842674086)*x_ref[1]+(0.608979796402)*x_ref[2]
        ref[0,3,0,1]=(0.53232240924)*x_ref[0]+(0.226439017439)*x_ref[1]+(-0.42603939944)*x_ref[2]
        ref[0,3,1,0]=(1.67410358342)*x_ref[0]+(-0.00813619771347)*x_ref[1]+(-0.463361889732)*x_ref[2]
        ref[0,3,1,1]=(1.3991502461)*x_ref[0]+(0.239999857694)*x_ref[1]+(-0.369907067208)*x_ref[2]
        ref[0,3,2,0]=(-0.248767913403)*x_ref[0]+(0.893584158229)*x_ref[1]+(1.97656682548)*x_ref[2]
        ref[0,3,2,1]=(-0.835046144341)*x_ref[0]+(-0.00979850832247)*x_ref[1]+(-0.281439904932)*x_ref[2]
        ref[0,4,0,0]=(0.974179494371)*x_ref[0]+(-0.0432121220464)*x_ref[1]+(-0.475018088554)*x_ref[2]
        ref[0,4,0,1]=(0.245490151034)*x_ref[0]+(-0.367572047585)*x_ref[1]+(-0.926123425123)*x_ref[2]
        ref[0,4,1,0]=(-0.239057491277)*x_ref[0]+(0.339734302574)*x_ref[1]+(0.760546363123)*x_ref[2]
        ref[0,4,1,1]=(-1.0425426898)*x_ref[0]+(0.774210702874)*x_ref[1]+(-1.15990543484)*x_ref[2]
        ref[0,4,2,0]=(1.02994063274)*x_ref[0]+(1.69313195574)*x_ref[1]+(-1.02468862318)*x_ref[2]
        ref[0,4,2,1]=(0.221805673731)*x_ref[0]+(-1.06588014758)*x_ref[1]+(-0.546973870537)*x_ref[2]
        ref[1,0,0,0]=(-0.140765393552)*x_ref[0]+(-0.309057115268)*x_ref[1]+(0.07814948965)*x_ref[2]
        ref[1,0,0,1]=(-0.29622101259)*x_ref[0]+(-1.0654452087)*x_ref[1]+(1.77500323701)*x_ref[2]
        ref[1,0,1,0]=(1.13202947738)*x_ref[0]+(-0.515454892388)*x_ref[1]+(-1.34230346013)*x_ref[2]
        ref[1,0,1,1]=(1.50210238409)*x_ref[0]+(-0.75673088573)*x_ref[1]+(-0.792078632258)*x_ref[2]
        ref[1,0,2,0]=(0.249111383954)*x_ref[0]+(-0.972418007776)*x_ref[1]+(0.766060429469)*x_ref[2]
        ref[1,0,2,1]=(-0.0710717292409)*x_ref[0]+(0.164949706666)*x_ref[1]+(1.46045331387)*x_ref[2]
        ref[1,1,0,0]=(-0.587668209291)*x_ref[0]+(0.303249169818)*x_ref[1]+(0.352197808782)*x_ref[2]
        ref[1,1,0,1]=(-0.145609904248)*x_ref[0]+(0.91895760024)*x_ref[1]+(-1.25063091478)*x_ref[2]
        ref[1,1,1,0]=(0.450869946651)*x_ref[0]+(0.752704827905)*x_ref[1]+(0.132042943533)*x_ref[2]
        ref[1,1,1,1]=(-0.87296460294)*x_ref[0]+(-0.459840609019)*x_ref[1]+(0.422574948491)*x_ref[2]
        ref[1,1,2,0]=(0.828459269131)*x_ref[0]+(0.45869153936)*x_ref[1]+(0.854640684176)*x_ref[2]
        ref[1,1,2,1]=(1.84720399473)*x_ref[0]+(0.0416642546234)*x_ref[1]+(-0.309583524852)*x_ref[2]
        ref[1,2,0,0]=(-1.12563293246)*x_ref[0]+(0.0142184473972)*x_ref[1]+(-1.18150362584)*x_ref[2]
        ref[1,2,0,1]=(-0.0750965768147)*x_ref[0]+(-0.419276687751)*x_ref[1]+(0.521012213863)*x_ref[2]
        ref[1,2,1,0]=(-1.06905807284)*x_ref[0]+(0.692258030743)*x_ref[1]+(0.262254909068)*x_ref[2]
        ref[1,2,1,1]=(0.0552441016888)*x_ref[0]+(-0.131040560322)*x_ref[1]+(-0.321517725963)*x_ref[2]
        ref[1,2,2,0]=(-0.740410514693)*x_ref[0]+(0.219020143899)*x_ref[1]+(1.35104682965)*x_ref[2]
        ref[1,2,2,1]=(-1.38198583939)*x_ref[0]+(-0.19342887842)*x_ref[1]+(-0.780008129494)*x_ref[2]
        ref[1,3,0,0]=(0.378358305605)*x_ref[0]+(1.32138618734)*x_ref[1]+(0.627209784997)*x_ref[2]
        ref[1,3,0,1]=(-0.0383657917537)*x_ref[0]+(-1.41770189628)*x_ref[1]+(-1.21625280681)*x_ref[2]
        ref[1,3,1,0]=(0.819666413875)*x_ref[0]+(0.160369878383)*x_ref[1]+(-0.229388558682)*x_ref[2]
        ref[1,3,1,1]=(0.393409555876)*x_ref[0]+(-0.525985341296)*x_ref[1]+(-1.11892438309)*x_ref[2]
        ref[1,3,2,0]=(0.145077010777)*x_ref[0]+(0.229800008438)*x_ref[1]+(-0.202652862622)*x_ref[2]
        ref[1,3,2,1]=(-1.48698378177)*x_ref[0]+(0.553698672636)*x_ref[1]+(-0.427263736796)*x_ref[2]
        ref[1,4,0,0]=(-0.129527725249)*x_ref[0]+(-0.639824801211)*x_ref[1]+(0.759081980717)*x_ref[2]
        ref[1,4,0,1]=(-0.117772473784)*x_ref[0]+(0.387474104611)*x_ref[1]+(0.507306747135)*x_ref[2]
        ref[1,4,1,0]=(-0.82518622198)*x_ref[0]+(-0.754474320343)*x_ref[1]+(-0.978049152511)*x_ref[2]
        ref[1,4,1,1]=(-0.370385633483)*x_ref[0]+(-0.348429423043)*x_ref[1]+(-0.254278257899)*x_ref[2]
        ref[1,4,2,0]=(1.42596167697)*x_ref[0]+(-0.871785904514)*x_ref[1]+(-0.5343959325)*x_ref[2]
        ref[1,4,2,1]=(0.817877672506)*x_ref[0]+(-0.222622791737)*x_ref[1]+(-0.677237468595)*x_ref[2]
        ref[2,0,0,0]=(0.877660484813)*x_ref[0]+(0.818663846913)*x_ref[1]+(-1.26520874625)*x_ref[2]
        ref[2,0,0,1]=(-0.822344385073)*x_ref[0]+(-0.274691696266)*x_ref[1]+(-0.35079248697)*x_ref[2]
        ref[2,0,1,0]=(-0.319040285312)*x_ref[0]+(-0.750406131559)*x_ref[1]+(-0.920332967032)*x_ref[2]
        ref[2,0,1,1]=(1.48758278445)*x_ref[0]+(1.30587557873)*x_ref[1]+(0.230692557888)*x_ref[2]
        ref[2,0,2,0]=(1.44269354043)*x_ref[0]+(-0.63270356894)*x_ref[1]+(0.229323632455)*x_ref[2]
        ref[2,0,2,1]=(-0.645065877536)*x_ref[0]+(-0.393083499676)*x_ref[1]+(-0.305117645453)*x_ref[2]
        ref[2,1,0,0]=(-0.63304134143)*x_ref[0]+(0.0664768401917)*x_ref[1]+(0.87587743066)*x_ref[2]
        ref[2,1,0,1]=(-0.531856958285)*x_ref[0]+(0.747953146648)*x_ref[1]+(1.23560779546)*x_ref[2]
        ref[2,1,1,0]=(0.0564479101025)*x_ref[0]+(0.546307803332)*x_ref[1]+(0.655044890368)*x_ref[2]
        ref[2,1,1,1]=(-1.12793057228)*x_ref[0]+(-0.736676194013)*x_ref[1]+(0.0375913829411)*x_ref[2]
        ref[2,1,2,0]=(-0.531384188067)*x_ref[0]+(0.249416904081)*x_ref[1]+(0.711430712986)*x_ref[2]
        ref[2,1,2,1]=(0.884244664065)*x_ref[0]+(0.456066941968)*x_ref[1]+(0.181630861785)*x_ref[2]
        ref[2,2,0,0]=(0.0844543345062)*x_ref[0]+(-0.435288041256)*x_ref[1]+(-0.434816763062)*x_ref[2]
        ref[2,2,0,1]=(0.373388105049)*x_ref[0]+(0.239833846429)*x_ref[1]+(-0.437771573282)*x_ref[2]
        ref[2,2,1,0]=(0.534401521907)*x_ref[0]+(0.393077203448)*x_ref[1]+(-0.42440387717)*x_ref[2]
        ref[2,2,1,1]=(0.318683369617)*x_ref[0]+(-0.538864643793)*x_ref[1]+(1.83466945806)*x_ref[2]
        ref[2,2,2,0]=(-1.79048122074)*x_ref[0]+(-1.05441213478)*x_ref[1]+(-0.573963806281)*x_ref[2]
        ref[2,2,2,1]=(0.944017405182)*x_ref[0]+(0.772756791538)*x_ref[1]+(0.0269673237336)*x_ref[2]
        ref[2,3,0,0]=(0.989703689142)*x_ref[0]+(1.21302325945)*x_ref[1]+(-0.92209419073)*x_ref[2]
        ref[2,3,0,1]=(-0.0957244757576)*x_ref[0]+(1.50132638315)*x_ref[1]+(-1.38716099756)*x_ref[2]
        ref[2,3,1,0]=(1.6445441776)*x_ref[0]+(-0.490207070694)*x_ref[1]+(0.626469347781)*x_ref[2]
        ref[2,3,1,1]=(0.0185568939147)*x_ref[0]+(0.412024752941)*x_ref[1]+(-0.22626102813)*x_ref[2]
        ref[2,3,2,0]=(0.0143609822966)*x_ref[0]+(0.115739025893)*x_ref[1]+(-0.544468631925)*x_ref[2]
        ref[2,3,2,1]=(-1.80522341493)*x_ref[0]+(0.33352152695)*x_ref[1]+(-0.841286464932)*x_ref[2]
        ref[2,4,0,0]=(0.0618444122042)*x_ref[0]+(1.91232484622)*x_ref[1]+(-0.406317200023)*x_ref[2]
        ref[2,4,0,1]=(0.279131062922)*x_ref[0]+(0.0194232637634)*x_ref[1]+(-0.00179069113634)*x_ref[2]
        ref[2,4,1,0]=(-0.701344579135)*x_ref[0]+(-0.27410935473)*x_ref[1]+(0.763362100733)*x_ref[2]
        ref[2,4,1,1]=(0.831198601185)*x_ref[0]+(0.90336781771)*x_ref[1]+(1.21856633382)*x_ref[2]
        ref[2,4,2,0]=(0.526278615033)*x_ref[0]+(-1.08673855801)*x_ref[1]+(0.0784355900913)*x_ref[2]
        ref[2,4,2,1]=(0.853226856851)*x_ref[0]+(-1.15919062627)*x_ref[1]+(0.185686031732)*x_ref[2]
        ref[3,0,0,0]=(0.272127498117)*x_ref[0]+(-0.457140162438)*x_ref[1]+(0.318358670541)*x_ref[2]
        ref[3,0,0,1]=(1.71056378307)*x_ref[0]+(0.200052730234)*x_ref[1]+(-0.127733325792)*x_ref[2]
        ref[3,0,1,0]=(-0.285340963261)*x_ref[0]+(-0.524544624385)*x_ref[1]+(0.916366474944)*x_ref[2]
        ref[3,0,1,1]=(-0.625506557557)*x_ref[0]+(1.0858476479)*x_ref[1]+(-1.24478983268)*x_ref[2]
        ref[3,0,2,0]=(-1.37299619916)*x_ref[0]+(-0.482246528202)*x_ref[1]+(-1.36857669075)*x_ref[2]
        ref[3,0,2,1]=(-0.924421844916)*x_ref[0]+(-1.50402098115)*x_ref[1]+(1.00045130673)*x_ref[2]
        ref[3,1,0,0]=(0.0417304700827)*x_ref[0]+(0.554527451004)*x_ref[1]+(0.299758163628)*x_ref[2]
        ref[3,1,0,1]=(0.639611639279)*x_ref[0]+(-0.444675638162)*x_ref[1]+(1.43150046757)*x_ref[2]
        ref[3,1,1,0]=(0.288697220898)*x_ref[0]+(1.32660433773)*x_ref[1]+(0.0889412116759)*x_ref[2]
        ref[3,1,1,1]=(-0.110516770201)*x_ref[0]+(-0.0214248733565)*x_ref[1]+(-0.758638794785)*x_ref[2]
        ref[3,1,2,0]=(1.34357138654)*x_ref[0]+(-0.13759272972)*x_ref[1]+(-0.00183845032784)*x_ref[2]
        ref[3,1,2,1]=(1.44174769627)*x_ref[0]+(0.49190176388)*x_ref[1]+(0.477331360141)*x_ref[2]
        ref[3,2,0,0]=(-0.10284777121)*x_ref[0]+(-0.63294157419)*x_ref[1]+(0.246064460451)*x_ref[2]
        ref[3,2,0,1]=(0.974270980699)*x_ref[0]+(1.03835886161)*x_ref[1]+(0.420000854838)*x_ref[2]
        ref[3,2,1,0]=(-0.00617831615923)*x_ref[0]+(-0.354063634073)*x_ref[1]+(0.764285884619)*x_ref[2]
        ref[3,2,1,1]=(0.210815790746)*x_ref[0]+(0.76609515744)*x_ref[1]+(-0.564401558401)*x_ref[2]
        ref[3,2,2,0]=(1.30471234011)*x_ref[0]+(1.54864119332)*x_ref[1]+(-0.493583650519)*x_ref[2]
        ref[3,2,2,1]=(0.42287513668)*x_ref[0]+(-0.603093868171)*x_ref[1]+(-0.811892974203)*x_ref[2]
        ref[3,3,0,0]=(-1.38358134693)*x_ref[0]+(-1.53706018098)*x_ref[1]+(-1.29701227501)*x_ref[2]
        ref[3,3,0,1]=(0.598581779902)*x_ref[0]+(1.02424010922)*x_ref[1]+(-1.1814313009)*x_ref[2]
        ref[3,3,1,0]=(1.30938529116)*x_ref[0]+(-0.504468477244)*x_ref[1]+(1.63655848854)*x_ref[2]
        ref[3,3,1,1]=(-1.47006424113)*x_ref[0]+(0.225951520203)*x_ref[1]+(-0.412273003169)*x_ref[2]
        ref[3,3,2,0]=(-0.116902234482)*x_ref[0]+(0.488144536044)*x_ref[1]+(1.01732461694)*x_ref[2]
        ref[3,3,2,1]=(1.0935814877)*x_ref[0]+(1.06389385916)*x_ref[1]+(-0.17981400591)*x_ref[2]
        ref[3,4,0,0]=(-0.833265062101)*x_ref[0]+(-0.754461148648)*x_ref[1]+(1.26702012815)*x_ref[2]
        ref[3,4,0,1]=(-0.793744278171)*x_ref[0]+(0.112704521653)*x_ref[1]+(-0.448684922187)*x_ref[2]
        ref[3,4,1,0]=(-0.396153523625)*x_ref[0]+(-0.0370870723568)*x_ref[1]+(0.554347618511)*x_ref[2]
        ref[3,4,1,1]=(-0.323276139711)*x_ref[0]+(-0.397389928392)*x_ref[1]+(-1.1707028324)*x_ref[2]
        ref[3,4,2,0]=(0.823350512513)*x_ref[0]+(1.47105759864)*x_ref[1]+(-0.977456877855)*x_ref[2]
        ref[3,4,2,1]=(-0.719791170741)*x_ref[0]+(-1.00363340645)*x_ref[1]+(0.814612770572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedContinuousFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-1.02854286859)*x[0]+(0.105922853194)*x[1]
        ref=(-1.02854286859)*x_ref[0]+(0.105922853194)*x_ref[1]
      else:
        arg=(1.36620376771)*x[0]+(-0.336798536224)*x[1]+(0.534538499605)*x[2]
        ref=(1.36620376771)*x_ref[0]+(-0.336798536224)*x_ref[1]+(0.534538499605)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        arg[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
        ref[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        ref[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
      else:
        arg[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        arg[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
        ref[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        ref[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.452635897444)*x[0]+(-0.210513646405)*x[1]
        arg[0,1]=(-0.590704436627)*x[0]+(1.78000273884)*x[1]
        arg[0,2]=(1.01331349402)*x[0]+(1.76279765125)*x[1]
        arg[0,3]=(1.5588852701)*x[0]+(-0.550895238707)*x[1]
        arg[0,4]=(-0.352453804527)*x[0]+(-0.338832613421)*x[1]
        arg[1,0]=(-0.809837877163)*x[0]+(-0.531009602702)*x[1]
        arg[1,1]=(0.513641503665)*x[0]+(0.232114989082)*x[1]
        arg[1,2]=(-0.424635655997)*x[0]+(0.0796757576821)*x[1]
        arg[1,3]=(-1.5899938122)*x[0]+(-0.178949548482)*x[1]
        arg[1,4]=(0.989317971495)*x[0]+(0.969711267468)*x[1]
        arg[2,0]=(0.384490947785)*x[0]+(1.42230193156)*x[1]
        arg[2,1]=(-0.0832227929255)*x[0]+(0.859020891429)*x[1]
        arg[2,2]=(-1.36129563124)*x[0]+(-0.240985031888)*x[1]
        arg[2,3]=(0.319706901063)*x[0]+(1.03419924566)*x[1]
        arg[2,4]=(0.986965249719)*x[0]+(0.228684031472)*x[1]
        arg[3,0]=(0.588520127337)*x[0]+(0.016799594822)*x[1]
        arg[3,1]=(1.46193438322)*x[0]+(0.431838900383)*x[1]
        arg[3,2]=(-0.0235512206818)*x[0]+(0.00227038037474)*x[1]
        arg[3,3]=(0.977383733076)*x[0]+(0.383979958766)*x[1]
        arg[3,4]=(0.447782383836)*x[0]+(-0.504218231709)*x[1]
        ref[0,0]=(-0.452635897444)*x_ref[0]+(-0.210513646405)*x_ref[1]
        ref[0,1]=(-0.590704436627)*x_ref[0]+(1.78000273884)*x_ref[1]
        ref[0,2]=(1.01331349402)*x_ref[0]+(1.76279765125)*x_ref[1]
        ref[0,3]=(1.5588852701)*x_ref[0]+(-0.550895238707)*x_ref[1]
        ref[0,4]=(-0.352453804527)*x_ref[0]+(-0.338832613421)*x_ref[1]
        ref[1,0]=(-0.809837877163)*x_ref[0]+(-0.531009602702)*x_ref[1]
        ref[1,1]=(0.513641503665)*x_ref[0]+(0.232114989082)*x_ref[1]
        ref[1,2]=(-0.424635655997)*x_ref[0]+(0.0796757576821)*x_ref[1]
        ref[1,3]=(-1.5899938122)*x_ref[0]+(-0.178949548482)*x_ref[1]
        ref[1,4]=(0.989317971495)*x_ref[0]+(0.969711267468)*x_ref[1]
        ref[2,0]=(0.384490947785)*x_ref[0]+(1.42230193156)*x_ref[1]
        ref[2,1]=(-0.0832227929255)*x_ref[0]+(0.859020891429)*x_ref[1]
        ref[2,2]=(-1.36129563124)*x_ref[0]+(-0.240985031888)*x_ref[1]
        ref[2,3]=(0.319706901063)*x_ref[0]+(1.03419924566)*x_ref[1]
        ref[2,4]=(0.986965249719)*x_ref[0]+(0.228684031472)*x_ref[1]
        ref[3,0]=(0.588520127337)*x_ref[0]+(0.016799594822)*x_ref[1]
        ref[3,1]=(1.46193438322)*x_ref[0]+(0.431838900383)*x_ref[1]
        ref[3,2]=(-0.0235512206818)*x_ref[0]+(0.00227038037474)*x_ref[1]
        ref[3,3]=(0.977383733076)*x_ref[0]+(0.383979958766)*x_ref[1]
        ref[3,4]=(0.447782383836)*x_ref[0]+(-0.504218231709)*x_ref[1]
      else:
        arg[0,0]=(-0.566459971675)*x[0]+(0.240246648893)*x[1]+(1.84693181923)*x[2]
        arg[0,1]=(1.09715015899)*x[0]+(1.26531924413)*x[1]+(0.627813347679)*x[2]
        arg[0,2]=(-1.17535073843)*x[0]+(-0.215601498941)*x[1]+(1.36346500888)*x[2]
        arg[0,3]=(-0.637207120161)*x[0]+(1.08542566952)*x[1]+(-1.13757514425)*x[2]
        arg[0,4]=(1.18247424069)*x[0]+(-1.48951614815)*x[1]+(0.615535039719)*x[2]
        arg[1,0]=(0.0937432100426)*x[0]+(-0.510326894156)*x[1]+(-0.456208996457)*x[2]
        arg[1,1]=(-0.866120342442)*x[0]+(0.927484984985)*x[1]+(0.918081524156)*x[2]
        arg[1,2]=(0.025238374235)*x[0]+(-0.182938599269)*x[1]+(-1.50048281322)*x[2]
        arg[1,3]=(-0.302334971294)*x[0]+(-0.925062799423)*x[1]+(-1.4675138953)*x[2]
        arg[1,4]=(-0.497738566447)*x[0]+(-0.324727071875)*x[1]+(-1.43341212055)*x[2]
        arg[2,0]=(-1.4605237958)*x[0]+(-0.332826007088)*x[1]+(-0.230483401111)*x[2]
        arg[2,1]=(0.801769248581)*x[0]+(-0.930588036503)*x[1]+(-1.06804808381)*x[2]
        arg[2,2]=(-1.07972165017)*x[0]+(1.09295362882)*x[1]+(0.048550098477)*x[2]
        arg[2,3]=(0.028290230337)*x[0]+(0.179461246635)*x[1]+(-1.09165257072)*x[2]
        arg[2,4]=(0.238850986174)*x[0]+(-0.837307206184)*x[1]+(-0.617254874594)*x[2]
        arg[3,0]=(-0.125921204582)*x[0]+(0.234758657879)*x[1]+(0.0847162247843)*x[2]
        arg[3,1]=(-0.213352414907)*x[0]+(0.206777352455)*x[1]+(-0.135357638131)*x[2]
        arg[3,2]=(-0.41897394434)*x[0]+(0.152706100016)*x[1]+(1.52390356012)*x[2]
        arg[3,3]=(0.1831705248)*x[0]+(-1.13556841364)*x[1]+(-1.7661652693)*x[2]
        arg[3,4]=(0.43473708388)*x[0]+(1.04496896213)*x[1]+(-0.816894739325)*x[2]
        ref[0,0]=(-0.566459971675)*x_ref[0]+(0.240246648893)*x_ref[1]+(1.84693181923)*x_ref[2]
        ref[0,1]=(1.09715015899)*x_ref[0]+(1.26531924413)*x_ref[1]+(0.627813347679)*x_ref[2]
        ref[0,2]=(-1.17535073843)*x_ref[0]+(-0.215601498941)*x_ref[1]+(1.36346500888)*x_ref[2]
        ref[0,3]=(-0.637207120161)*x_ref[0]+(1.08542566952)*x_ref[1]+(-1.13757514425)*x_ref[2]
        ref[0,4]=(1.18247424069)*x_ref[0]+(-1.48951614815)*x_ref[1]+(0.615535039719)*x_ref[2]
        ref[1,0]=(0.0937432100426)*x_ref[0]+(-0.510326894156)*x_ref[1]+(-0.456208996457)*x_ref[2]
        ref[1,1]=(-0.866120342442)*x_ref[0]+(0.927484984985)*x_ref[1]+(0.918081524156)*x_ref[2]
        ref[1,2]=(0.025238374235)*x_ref[0]+(-0.182938599269)*x_ref[1]+(-1.50048281322)*x_ref[2]
        ref[1,3]=(-0.302334971294)*x_ref[0]+(-0.925062799423)*x_ref[1]+(-1.4675138953)*x_ref[2]
        ref[1,4]=(-0.497738566447)*x_ref[0]+(-0.324727071875)*x_ref[1]+(-1.43341212055)*x_ref[2]
        ref[2,0]=(-1.4605237958)*x_ref[0]+(-0.332826007088)*x_ref[1]+(-0.230483401111)*x_ref[2]
        ref[2,1]=(0.801769248581)*x_ref[0]+(-0.930588036503)*x_ref[1]+(-1.06804808381)*x_ref[2]
        ref[2,2]=(-1.07972165017)*x_ref[0]+(1.09295362882)*x_ref[1]+(0.048550098477)*x_ref[2]
        ref[2,3]=(0.028290230337)*x_ref[0]+(0.179461246635)*x_ref[1]+(-1.09165257072)*x_ref[2]
        ref[2,4]=(0.238850986174)*x_ref[0]+(-0.837307206184)*x_ref[1]+(-0.617254874594)*x_ref[2]
        ref[3,0]=(-0.125921204582)*x_ref[0]+(0.234758657879)*x_ref[1]+(0.0847162247843)*x_ref[2]
        ref[3,1]=(-0.213352414907)*x_ref[0]+(0.206777352455)*x_ref[1]+(-0.135357638131)*x_ref[2]
        ref[3,2]=(-0.41897394434)*x_ref[0]+(0.152706100016)*x_ref[1]+(1.52390356012)*x_ref[2]
        ref[3,3]=(0.1831705248)*x_ref[0]+(-1.13556841364)*x_ref[1]+(-1.7661652693)*x_ref[2]
        ref[3,4]=(0.43473708388)*x_ref[0]+(1.04496896213)*x_ref[1]+(-0.816894739325)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.324237136046)*x[0]+(0.67172278334)*x[1]
        arg[0,0,1]=(-1.07360258996)*x[0]+(-0.646208771542)*x[1]
        arg[0,1,0]=(-0.040139403311)*x[0]+(-0.281626553874)*x[1]
        arg[0,1,1]=(0.963343050385)*x[0]+(-0.724838290038)*x[1]
        arg[1,0,0]=(-0.909697586367)*x[0]+(0.48041903189)*x[1]
        arg[1,0,1]=(0.414758676032)*x[0]+(-0.0921084575218)*x[1]
        arg[1,1,0]=(1.06852211241)*x[0]+(-1.01026709886)*x[1]
        arg[1,1,1]=(-0.675258695211)*x[0]+(0.51361078905)*x[1]
        arg[2,0,0]=(-0.791246198786)*x[0]+(-0.0283034425042)*x[1]
        arg[2,0,1]=(-0.87756010848)*x[0]+(-0.00868464377539)*x[1]
        arg[2,1,0]=(0.44209070344)*x[0]+(0.0846344294002)*x[1]
        arg[2,1,1]=(0.777287404475)*x[0]+(0.72614634535)*x[1]
        arg[3,0,0]=(-0.363354104693)*x[0]+(-1.22959837449)*x[1]
        arg[3,0,1]=(-0.532220757425)*x[0]+(-1.54990170304)*x[1]
        arg[3,1,0]=(-0.491099325731)*x[0]+(-0.412402332731)*x[1]
        arg[3,1,1]=(0.0750673227925)*x[0]+(-0.725108664028)*x[1]
        arg[4,0,0]=(-0.419689899608)*x[0]+(-1.32423898043)*x[1]
        arg[4,0,1]=(0.532682204197)*x[0]+(0.0703779631797)*x[1]
        arg[4,1,0]=(1.68643150913)*x[0]+(0.105443733595)*x[1]
        arg[4,1,1]=(0.936329701936)*x[0]+(1.04067808474)*x[1]
        arg[5,0,0]=(1.37285560133)*x[0]+(-0.680844454631)*x[1]
        arg[5,0,1]=(0.843677131436)*x[0]+(0.272365035967)*x[1]
        arg[5,1,0]=(0.00586472074043)*x[0]+(-1.09785622721)*x[1]
        arg[5,1,1]=(1.1410854961)*x[0]+(-0.591468439837)*x[1]
        ref[0,0,0]=(0.324237136046)*x_ref[0]+(0.67172278334)*x_ref[1]
        ref[0,0,1]=(-1.07360258996)*x_ref[0]+(-0.646208771542)*x_ref[1]
        ref[0,1,0]=(-0.040139403311)*x_ref[0]+(-0.281626553874)*x_ref[1]
        ref[0,1,1]=(0.963343050385)*x_ref[0]+(-0.724838290038)*x_ref[1]
        ref[1,0,0]=(-0.909697586367)*x_ref[0]+(0.48041903189)*x_ref[1]
        ref[1,0,1]=(0.414758676032)*x_ref[0]+(-0.0921084575218)*x_ref[1]
        ref[1,1,0]=(1.06852211241)*x_ref[0]+(-1.01026709886)*x_ref[1]
        ref[1,1,1]=(-0.675258695211)*x_ref[0]+(0.51361078905)*x_ref[1]
        ref[2,0,0]=(-0.791246198786)*x_ref[0]+(-0.0283034425042)*x_ref[1]
        ref[2,0,1]=(-0.87756010848)*x_ref[0]+(-0.00868464377539)*x_ref[1]
        ref[2,1,0]=(0.44209070344)*x_ref[0]+(0.0846344294002)*x_ref[1]
        ref[2,1,1]=(0.777287404475)*x_ref[0]+(0.72614634535)*x_ref[1]
        ref[3,0,0]=(-0.363354104693)*x_ref[0]+(-1.22959837449)*x_ref[1]
        ref[3,0,1]=(-0.532220757425)*x_ref[0]+(-1.54990170304)*x_ref[1]
        ref[3,1,0]=(-0.491099325731)*x_ref[0]+(-0.412402332731)*x_ref[1]
        ref[3,1,1]=(0.0750673227925)*x_ref[0]+(-0.725108664028)*x_ref[1]
        ref[4,0,0]=(-0.419689899608)*x_ref[0]+(-1.32423898043)*x_ref[1]
        ref[4,0,1]=(0.532682204197)*x_ref[0]+(0.0703779631797)*x_ref[1]
        ref[4,1,0]=(1.68643150913)*x_ref[0]+(0.105443733595)*x_ref[1]
        ref[4,1,1]=(0.936329701936)*x_ref[0]+(1.04067808474)*x_ref[1]
        ref[5,0,0]=(1.37285560133)*x_ref[0]+(-0.680844454631)*x_ref[1]
        ref[5,0,1]=(0.843677131436)*x_ref[0]+(0.272365035967)*x_ref[1]
        ref[5,1,0]=(0.00586472074043)*x_ref[0]+(-1.09785622721)*x_ref[1]
        ref[5,1,1]=(1.1410854961)*x_ref[0]+(-0.591468439837)*x_ref[1]
      else:
        arg[0,0,0]=(-0.289427308366)*x[0]+(1.17077011703)*x[1]+(0.763196501047)*x[2]
        arg[0,0,1]=(0.642730500265)*x[0]+(1.1719001632)*x[1]+(-0.121378040894)*x[2]
        arg[0,1,0]=(0.0879005621769)*x[0]+(-0.627582537594)*x[1]+(0.220072369447)*x[2]
        arg[0,1,1]=(-0.518762071566)*x[0]+(-0.700754417623)*x[1]+(-0.340623426752)*x[2]
        arg[1,0,0]=(-0.162635708202)*x[0]+(-0.996922517356)*x[1]+(0.48000598379)*x[2]
        arg[1,0,1]=(-0.49910792923)*x[0]+(1.11334748765)*x[1]+(0.0050657496159)*x[2]
        arg[1,1,0]=(-1.55972927502)*x[0]+(0.233936234238)*x[1]+(-0.546906209213)*x[2]
        arg[1,1,1]=(-0.6012330303)*x[0]+(-0.267951301884)*x[1]+(-0.193021060351)*x[2]
        arg[2,0,0]=(1.32515959423)*x[0]+(1.2070735022)*x[1]+(1.11474945521)*x[2]
        arg[2,0,1]=(-1.05500093406)*x[0]+(-0.937935240268)*x[1]+(0.180289797658)*x[2]
        arg[2,1,0]=(-0.450015559749)*x[0]+(-1.17715272891)*x[1]+(0.197160834934)*x[2]
        arg[2,1,1]=(-1.24969258786)*x[0]+(-0.300915973735)*x[1]+(-0.195058423793)*x[2]
        arg[3,0,0]=(0.704171975525)*x[0]+(1.0327384159)*x[1]+(-1.23664355084)*x[2]
        arg[3,0,1]=(0.916288412006)*x[0]+(1.6929893349)*x[1]+(-1.16682244827)*x[2]
        arg[3,1,0]=(-0.756147901025)*x[0]+(0.0232005287342)*x[1]+(0.213796289692)*x[2]
        arg[3,1,1]=(0.92799443461)*x[0]+(-0.228093116826)*x[1]+(-1.53373185211)*x[2]
        arg[4,0,0]=(-0.272514671459)*x[0]+(0.310026631492)*x[1]+(0.735127519475)*x[2]
        arg[4,0,1]=(1.23963120106)*x[0]+(-0.0769223754537)*x[1]+(0.767002455909)*x[2]
        arg[4,1,0]=(-1.1121433063)*x[0]+(0.748482838233)*x[1]+(-0.0985898502473)*x[2]
        arg[4,1,1]=(-1.08169600038)*x[0]+(-1.39575461699)*x[1]+(0.468455733883)*x[2]
        arg[5,0,0]=(-0.737100332594)*x[0]+(1.0950941569)*x[1]+(0.530144605873)*x[2]
        arg[5,0,1]=(1.94353080429)*x[0]+(1.61187736568)*x[1]+(0.887255997784)*x[2]
        arg[5,1,0]=(-0.492812782764)*x[0]+(-1.17859119673)*x[1]+(-0.627254028293)*x[2]
        arg[5,1,1]=(0.932565304399)*x[0]+(0.760125636669)*x[1]+(-0.687485104891)*x[2]
        ref[0,0,0]=(-0.289427308366)*x_ref[0]+(1.17077011703)*x_ref[1]+(0.763196501047)*x_ref[2]
        ref[0,0,1]=(0.642730500265)*x_ref[0]+(1.1719001632)*x_ref[1]+(-0.121378040894)*x_ref[2]
        ref[0,1,0]=(0.0879005621769)*x_ref[0]+(-0.627582537594)*x_ref[1]+(0.220072369447)*x_ref[2]
        ref[0,1,1]=(-0.518762071566)*x_ref[0]+(-0.700754417623)*x_ref[1]+(-0.340623426752)*x_ref[2]
        ref[1,0,0]=(-0.162635708202)*x_ref[0]+(-0.996922517356)*x_ref[1]+(0.48000598379)*x_ref[2]
        ref[1,0,1]=(-0.49910792923)*x_ref[0]+(1.11334748765)*x_ref[1]+(0.0050657496159)*x_ref[2]
        ref[1,1,0]=(-1.55972927502)*x_ref[0]+(0.233936234238)*x_ref[1]+(-0.546906209213)*x_ref[2]
        ref[1,1,1]=(-0.6012330303)*x_ref[0]+(-0.267951301884)*x_ref[1]+(-0.193021060351)*x_ref[2]
        ref[2,0,0]=(1.32515959423)*x_ref[0]+(1.2070735022)*x_ref[1]+(1.11474945521)*x_ref[2]
        ref[2,0,1]=(-1.05500093406)*x_ref[0]+(-0.937935240268)*x_ref[1]+(0.180289797658)*x_ref[2]
        ref[2,1,0]=(-0.450015559749)*x_ref[0]+(-1.17715272891)*x_ref[1]+(0.197160834934)*x_ref[2]
        ref[2,1,1]=(-1.24969258786)*x_ref[0]+(-0.300915973735)*x_ref[1]+(-0.195058423793)*x_ref[2]
        ref[3,0,0]=(0.704171975525)*x_ref[0]+(1.0327384159)*x_ref[1]+(-1.23664355084)*x_ref[2]
        ref[3,0,1]=(0.916288412006)*x_ref[0]+(1.6929893349)*x_ref[1]+(-1.16682244827)*x_ref[2]
        ref[3,1,0]=(-0.756147901025)*x_ref[0]+(0.0232005287342)*x_ref[1]+(0.213796289692)*x_ref[2]
        ref[3,1,1]=(0.92799443461)*x_ref[0]+(-0.228093116826)*x_ref[1]+(-1.53373185211)*x_ref[2]
        ref[4,0,0]=(-0.272514671459)*x_ref[0]+(0.310026631492)*x_ref[1]+(0.735127519475)*x_ref[2]
        ref[4,0,1]=(1.23963120106)*x_ref[0]+(-0.0769223754537)*x_ref[1]+(0.767002455909)*x_ref[2]
        ref[4,1,0]=(-1.1121433063)*x_ref[0]+(0.748482838233)*x_ref[1]+(-0.0985898502473)*x_ref[2]
        ref[4,1,1]=(-1.08169600038)*x_ref[0]+(-1.39575461699)*x_ref[1]+(0.468455733883)*x_ref[2]
        ref[5,0,0]=(-0.737100332594)*x_ref[0]+(1.0950941569)*x_ref[1]+(0.530144605873)*x_ref[2]
        ref[5,0,1]=(1.94353080429)*x_ref[0]+(1.61187736568)*x_ref[1]+(0.887255997784)*x_ref[2]
        ref[5,1,0]=(-0.492812782764)*x_ref[0]+(-1.17859119673)*x_ref[1]+(-0.627254028293)*x_ref[2]
        ref[5,1,1]=(0.932565304399)*x_ref[0]+(0.760125636669)*x_ref[1]+(-0.687485104891)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.261713255264)*x[0]+(-0.650225996431)*x[1]
        arg[0,0,0,1]=(0.177056727607)*x[0]+(-0.486487540393)*x[1]
        arg[0,0,1,0]=(1.38050234107)*x[0]+(0.273417897264)*x[1]
        arg[0,0,1,1]=(-1.059824929)*x[0]+(0.432893707387)*x[1]
        arg[0,0,2,0]=(1.11328992978)*x[0]+(1.14949237037)*x[1]
        arg[0,0,2,1]=(-0.176459646227)*x[0]+(-0.835980248375)*x[1]
        arg[0,1,0,0]=(0.992757237392)*x[0]+(0.113511598452)*x[1]
        arg[0,1,0,1]=(-0.211425610821)*x[0]+(-0.472765285276)*x[1]
        arg[0,1,1,0]=(-1.52498305232)*x[0]+(-0.869737083114)*x[1]
        arg[0,1,1,1]=(-0.474203117317)*x[0]+(-0.284532442353)*x[1]
        arg[0,1,2,0]=(-0.431542961546)*x[0]+(-0.0427397090453)*x[1]
        arg[0,1,2,1]=(-1.0562625269)*x[0]+(-0.74486767124)*x[1]
        arg[0,2,0,0]=(1.46553867064)*x[0]+(-0.650182195657)*x[1]
        arg[0,2,0,1]=(-0.717482663647)*x[0]+(-0.716320000777)*x[1]
        arg[0,2,1,0]=(-1.21701970743)*x[0]+(0.636901443761)*x[1]
        arg[0,2,1,1]=(0.438592928507)*x[0]+(0.812362899723)*x[1]
        arg[0,2,2,0]=(0.104885477141)*x[0]+(-0.127968978832)*x[1]
        arg[0,2,2,1]=(0.285116712849)*x[0]+(0.169740783305)*x[1]
        arg[0,3,0,0]=(-1.22825142912)*x[0]+(-0.562911687838)*x[1]
        arg[0,3,0,1]=(-0.569159534025)*x[0]+(1.30898732707)*x[1]
        arg[0,3,1,0]=(0.164709387569)*x[0]+(-1.62733582919)*x[1]
        arg[0,3,1,1]=(-0.477197320128)*x[0]+(-0.115207237191)*x[1]
        arg[0,3,2,0]=(-0.046871003317)*x[0]+(0.0371149345252)*x[1]
        arg[0,3,2,1]=(0.839624830979)*x[0]+(0.0785937909827)*x[1]
        arg[0,4,0,0]=(-0.0250725769598)*x[0]+(-0.535781603767)*x[1]
        arg[0,4,0,1]=(-0.761849740443)*x[0]+(-0.300714501152)*x[1]
        arg[0,4,1,0]=(0.0212637838833)*x[0]+(-1.56914428088)*x[1]
        arg[0,4,1,1]=(-0.254524677425)*x[0]+(0.699288623343)*x[1]
        arg[0,4,2,0]=(1.14678964272)*x[0]+(1.24738649379)*x[1]
        arg[0,4,2,1]=(1.80545916472)*x[0]+(1.22689087643)*x[1]
        arg[1,0,0,0]=(-0.116129927341)*x[0]+(-0.875165834648)*x[1]
        arg[1,0,0,1]=(-0.49905851236)*x[0]+(-1.10711214732)*x[1]
        arg[1,0,1,0]=(-1.44706167224)*x[0]+(-1.80259802262)*x[1]
        arg[1,0,1,1]=(0.555288273554)*x[0]+(0.153437302525)*x[1]
        arg[1,0,2,0]=(0.00774838569142)*x[0]+(-1.35683757658)*x[1]
        arg[1,0,2,1]=(-0.625156986286)*x[0]+(-0.201091038824)*x[1]
        arg[1,1,0,0]=(-0.765068445971)*x[0]+(1.32890117415)*x[1]
        arg[1,1,0,1]=(-0.629548919312)*x[0]+(-0.0512345513979)*x[1]
        arg[1,1,1,0]=(-0.361717423775)*x[0]+(0.286169858965)*x[1]
        arg[1,1,1,1]=(-0.533065922681)*x[0]+(0.103362397133)*x[1]
        arg[1,1,2,0]=(0.978041972715)*x[0]+(0.405949452109)*x[1]
        arg[1,1,2,1]=(-1.08418078618)*x[0]+(0.027879770152)*x[1]
        arg[1,2,0,0]=(1.23545256768)*x[0]+(0.574400116282)*x[1]
        arg[1,2,0,1]=(0.245362478488)*x[0]+(0.241577755816)*x[1]
        arg[1,2,1,0]=(0.645309443244)*x[0]+(0.0437773509036)*x[1]
        arg[1,2,1,1]=(0.00930197364105)*x[0]+(0.472462788946)*x[1]
        arg[1,2,2,0]=(1.44241534167)*x[0]+(-0.901332994844)*x[1]
        arg[1,2,2,1]=(0.893335284585)*x[0]+(0.109130146628)*x[1]
        arg[1,3,0,0]=(0.601907323675)*x[0]+(-1.54057502586)*x[1]
        arg[1,3,0,1]=(1.38086759376)*x[0]+(0.356522488945)*x[1]
        arg[1,3,1,0]=(1.16786130652)*x[0]+(-0.788756049484)*x[1]
        arg[1,3,1,1]=(-0.22580270698)*x[0]+(0.857686971693)*x[1]
        arg[1,3,2,0]=(0.709549396131)*x[0]+(-0.644330864877)*x[1]
        arg[1,3,2,1]=(1.04901965211)*x[0]+(0.0178384468958)*x[1]
        arg[1,4,0,0]=(-0.525050731412)*x[0]+(-0.253639122312)*x[1]
        arg[1,4,0,1]=(-1.63200911423)*x[0]+(0.519020443372)*x[1]
        arg[1,4,1,0]=(0.23956666878)*x[0]+(-1.76973024011)*x[1]
        arg[1,4,1,1]=(-0.412317050346)*x[0]+(-0.391711749094)*x[1]
        arg[1,4,2,0]=(1.63293176518)*x[0]+(0.583921572851)*x[1]
        arg[1,4,2,1]=(0.434893659373)*x[0]+(-0.324706930043)*x[1]
        arg[2,0,0,0]=(0.368478865639)*x[0]+(1.87637158068)*x[1]
        arg[2,0,0,1]=(1.32519020398)*x[0]+(0.306016568948)*x[1]
        arg[2,0,1,0]=(-0.284264003593)*x[0]+(-0.899022593367)*x[1]
        arg[2,0,1,1]=(-0.0439761413236)*x[0]+(-0.696386615358)*x[1]
        arg[2,0,2,0]=(0.407393812349)*x[0]+(-1.04815464317)*x[1]
        arg[2,0,2,1]=(-0.956661282071)*x[0]+(0.243281223782)*x[1]
        arg[2,1,0,0]=(-1.29201911939)*x[0]+(-0.693049706038)*x[1]
        arg[2,1,0,1]=(-0.711059848853)*x[0]+(-1.14204907272)*x[1]
        arg[2,1,1,0]=(-0.347693333057)*x[0]+(0.818879269672)*x[1]
        arg[2,1,1,1]=(0.00742807299875)*x[0]+(0.138845309984)*x[1]
        arg[2,1,2,0]=(1.13412244048)*x[0]+(-0.481422214102)*x[1]
        arg[2,1,2,1]=(0.64440302826)*x[0]+(-0.955231893089)*x[1]
        arg[2,2,0,0]=(0.13123141544)*x[0]+(-0.602266537541)*x[1]
        arg[2,2,0,1]=(-0.204818071763)*x[0]+(-1.13480248293)*x[1]
        arg[2,2,1,0]=(0.781058287579)*x[0]+(-1.31936224435)*x[1]
        arg[2,2,1,1]=(-0.591275351662)*x[0]+(0.691816969517)*x[1]
        arg[2,2,2,0]=(-0.921964826241)*x[0]+(1.28687075541)*x[1]
        arg[2,2,2,1]=(0.246260462337)*x[0]+(-1.04699282194)*x[1]
        arg[2,3,0,0]=(0.264209126513)*x[0]+(-0.697094400566)*x[1]
        arg[2,3,0,1]=(0.456158298484)*x[0]+(0.0288036445565)*x[1]
        arg[2,3,1,0]=(-0.720337925392)*x[0]+(0.0965182642186)*x[1]
        arg[2,3,1,1]=(0.397480419547)*x[0]+(-1.55232798344)*x[1]
        arg[2,3,2,0]=(-0.337139999283)*x[0]+(0.918809751807)*x[1]
        arg[2,3,2,1]=(1.34009447198)*x[0]+(-1.41390962018)*x[1]
        arg[2,4,0,0]=(1.67781838006)*x[0]+(0.521276193814)*x[1]
        arg[2,4,0,1]=(-0.114014321939)*x[0]+(0.529316778773)*x[1]
        arg[2,4,1,0]=(-0.19938131807)*x[0]+(-0.0712417614843)*x[1]
        arg[2,4,1,1]=(-0.724676753871)*x[0]+(0.731341170171)*x[1]
        arg[2,4,2,0]=(0.988755659656)*x[0]+(-1.0327352189)*x[1]
        arg[2,4,2,1]=(-0.526286732859)*x[0]+(-1.18295671598)*x[1]
        arg[3,0,0,0]=(-0.685223304723)*x[0]+(0.508431586068)*x[1]
        arg[3,0,0,1]=(0.529240274944)*x[0]+(0.77519128969)*x[1]
        arg[3,0,1,0]=(-0.757098431928)*x[0]+(1.41375812725)*x[1]
        arg[3,0,1,1]=(1.20805599335)*x[0]+(-0.184660334182)*x[1]
        arg[3,0,2,0]=(0.718413765561)*x[0]+(0.971149168706)*x[1]
        arg[3,0,2,1]=(-0.632670534518)*x[0]+(0.0389618967363)*x[1]
        arg[3,1,0,0]=(-0.525814673755)*x[0]+(0.574122587598)*x[1]
        arg[3,1,0,1]=(-1.04117288975)*x[0]+(-1.00738089823)*x[1]
        arg[3,1,1,0]=(0.0312806064182)*x[0]+(-0.30773511847)*x[1]
        arg[3,1,1,1]=(0.889930889355)*x[0]+(0.833689153489)*x[1]
        arg[3,1,2,0]=(-1.21586293381)*x[0]+(0.512182349574)*x[1]
        arg[3,1,2,1]=(-0.00437508755415)*x[0]+(-1.13228664497)*x[1]
        arg[3,2,0,0]=(1.64174174554)*x[0]+(-0.696500062562)*x[1]
        arg[3,2,0,1]=(0.0265343210604)*x[0]+(-0.707378716636)*x[1]
        arg[3,2,1,0]=(0.952922122967)*x[0]+(0.0946904561805)*x[1]
        arg[3,2,1,1]=(-1.06905266456)*x[0]+(-0.155122789807)*x[1]
        arg[3,2,2,0]=(-1.14410789976)*x[0]+(-0.379020164446)*x[1]
        arg[3,2,2,1]=(-0.0922611198553)*x[0]+(1.36680699099)*x[1]
        arg[3,3,0,0]=(0.849028205422)*x[0]+(-0.968698293811)*x[1]
        arg[3,3,0,1]=(0.649894718675)*x[0]+(0.238249259292)*x[1]
        arg[3,3,1,0]=(1.28170508471)*x[0]+(-0.736544947537)*x[1]
        arg[3,3,1,1]=(-0.188411685414)*x[0]+(1.14145010691)*x[1]
        arg[3,3,2,0]=(-1.13973037849)*x[0]+(-0.256565914821)*x[1]
        arg[3,3,2,1]=(0.955478585358)*x[0]+(-0.83417326515)*x[1]
        arg[3,4,0,0]=(0.155710789468)*x[0]+(0.204956811648)*x[1]
        arg[3,4,0,1]=(0.567866952974)*x[0]+(-0.168398320099)*x[1]
        arg[3,4,1,0]=(-1.36997367)*x[0]+(0.141984040172)*x[1]
        arg[3,4,1,1]=(-0.809532285206)*x[0]+(0.70611485405)*x[1]
        arg[3,4,2,0]=(-0.453260232051)*x[0]+(0.0665127865858)*x[1]
        arg[3,4,2,1]=(0.00142021115277)*x[0]+(1.02307505343)*x[1]
        ref[0,0,0,0]=(-0.261713255264)*x_ref[0]+(-0.650225996431)*x_ref[1]
        ref[0,0,0,1]=(0.177056727607)*x_ref[0]+(-0.486487540393)*x_ref[1]
        ref[0,0,1,0]=(1.38050234107)*x_ref[0]+(0.273417897264)*x_ref[1]
        ref[0,0,1,1]=(-1.059824929)*x_ref[0]+(0.432893707387)*x_ref[1]
        ref[0,0,2,0]=(1.11328992978)*x_ref[0]+(1.14949237037)*x_ref[1]
        ref[0,0,2,1]=(-0.176459646227)*x_ref[0]+(-0.835980248375)*x_ref[1]
        ref[0,1,0,0]=(0.992757237392)*x_ref[0]+(0.113511598452)*x_ref[1]
        ref[0,1,0,1]=(-0.211425610821)*x_ref[0]+(-0.472765285276)*x_ref[1]
        ref[0,1,1,0]=(-1.52498305232)*x_ref[0]+(-0.869737083114)*x_ref[1]
        ref[0,1,1,1]=(-0.474203117317)*x_ref[0]+(-0.284532442353)*x_ref[1]
        ref[0,1,2,0]=(-0.431542961546)*x_ref[0]+(-0.0427397090453)*x_ref[1]
        ref[0,1,2,1]=(-1.0562625269)*x_ref[0]+(-0.74486767124)*x_ref[1]
        ref[0,2,0,0]=(1.46553867064)*x_ref[0]+(-0.650182195657)*x_ref[1]
        ref[0,2,0,1]=(-0.717482663647)*x_ref[0]+(-0.716320000777)*x_ref[1]
        ref[0,2,1,0]=(-1.21701970743)*x_ref[0]+(0.636901443761)*x_ref[1]
        ref[0,2,1,1]=(0.438592928507)*x_ref[0]+(0.812362899723)*x_ref[1]
        ref[0,2,2,0]=(0.104885477141)*x_ref[0]+(-0.127968978832)*x_ref[1]
        ref[0,2,2,1]=(0.285116712849)*x_ref[0]+(0.169740783305)*x_ref[1]
        ref[0,3,0,0]=(-1.22825142912)*x_ref[0]+(-0.562911687838)*x_ref[1]
        ref[0,3,0,1]=(-0.569159534025)*x_ref[0]+(1.30898732707)*x_ref[1]
        ref[0,3,1,0]=(0.164709387569)*x_ref[0]+(-1.62733582919)*x_ref[1]
        ref[0,3,1,1]=(-0.477197320128)*x_ref[0]+(-0.115207237191)*x_ref[1]
        ref[0,3,2,0]=(-0.046871003317)*x_ref[0]+(0.0371149345252)*x_ref[1]
        ref[0,3,2,1]=(0.839624830979)*x_ref[0]+(0.0785937909827)*x_ref[1]
        ref[0,4,0,0]=(-0.0250725769598)*x_ref[0]+(-0.535781603767)*x_ref[1]
        ref[0,4,0,1]=(-0.761849740443)*x_ref[0]+(-0.300714501152)*x_ref[1]
        ref[0,4,1,0]=(0.0212637838833)*x_ref[0]+(-1.56914428088)*x_ref[1]
        ref[0,4,1,1]=(-0.254524677425)*x_ref[0]+(0.699288623343)*x_ref[1]
        ref[0,4,2,0]=(1.14678964272)*x_ref[0]+(1.24738649379)*x_ref[1]
        ref[0,4,2,1]=(1.80545916472)*x_ref[0]+(1.22689087643)*x_ref[1]
        ref[1,0,0,0]=(-0.116129927341)*x_ref[0]+(-0.875165834648)*x_ref[1]
        ref[1,0,0,1]=(-0.49905851236)*x_ref[0]+(-1.10711214732)*x_ref[1]
        ref[1,0,1,0]=(-1.44706167224)*x_ref[0]+(-1.80259802262)*x_ref[1]
        ref[1,0,1,1]=(0.555288273554)*x_ref[0]+(0.153437302525)*x_ref[1]
        ref[1,0,2,0]=(0.00774838569142)*x_ref[0]+(-1.35683757658)*x_ref[1]
        ref[1,0,2,1]=(-0.625156986286)*x_ref[0]+(-0.201091038824)*x_ref[1]
        ref[1,1,0,0]=(-0.765068445971)*x_ref[0]+(1.32890117415)*x_ref[1]
        ref[1,1,0,1]=(-0.629548919312)*x_ref[0]+(-0.0512345513979)*x_ref[1]
        ref[1,1,1,0]=(-0.361717423775)*x_ref[0]+(0.286169858965)*x_ref[1]
        ref[1,1,1,1]=(-0.533065922681)*x_ref[0]+(0.103362397133)*x_ref[1]
        ref[1,1,2,0]=(0.978041972715)*x_ref[0]+(0.405949452109)*x_ref[1]
        ref[1,1,2,1]=(-1.08418078618)*x_ref[0]+(0.027879770152)*x_ref[1]
        ref[1,2,0,0]=(1.23545256768)*x_ref[0]+(0.574400116282)*x_ref[1]
        ref[1,2,0,1]=(0.245362478488)*x_ref[0]+(0.241577755816)*x_ref[1]
        ref[1,2,1,0]=(0.645309443244)*x_ref[0]+(0.0437773509036)*x_ref[1]
        ref[1,2,1,1]=(0.00930197364105)*x_ref[0]+(0.472462788946)*x_ref[1]
        ref[1,2,2,0]=(1.44241534167)*x_ref[0]+(-0.901332994844)*x_ref[1]
        ref[1,2,2,1]=(0.893335284585)*x_ref[0]+(0.109130146628)*x_ref[1]
        ref[1,3,0,0]=(0.601907323675)*x_ref[0]+(-1.54057502586)*x_ref[1]
        ref[1,3,0,1]=(1.38086759376)*x_ref[0]+(0.356522488945)*x_ref[1]
        ref[1,3,1,0]=(1.16786130652)*x_ref[0]+(-0.788756049484)*x_ref[1]
        ref[1,3,1,1]=(-0.22580270698)*x_ref[0]+(0.857686971693)*x_ref[1]
        ref[1,3,2,0]=(0.709549396131)*x_ref[0]+(-0.644330864877)*x_ref[1]
        ref[1,3,2,1]=(1.04901965211)*x_ref[0]+(0.0178384468958)*x_ref[1]
        ref[1,4,0,0]=(-0.525050731412)*x_ref[0]+(-0.253639122312)*x_ref[1]
        ref[1,4,0,1]=(-1.63200911423)*x_ref[0]+(0.519020443372)*x_ref[1]
        ref[1,4,1,0]=(0.23956666878)*x_ref[0]+(-1.76973024011)*x_ref[1]
        ref[1,4,1,1]=(-0.412317050346)*x_ref[0]+(-0.391711749094)*x_ref[1]
        ref[1,4,2,0]=(1.63293176518)*x_ref[0]+(0.583921572851)*x_ref[1]
        ref[1,4,2,1]=(0.434893659373)*x_ref[0]+(-0.324706930043)*x_ref[1]
        ref[2,0,0,0]=(0.368478865639)*x_ref[0]+(1.87637158068)*x_ref[1]
        ref[2,0,0,1]=(1.32519020398)*x_ref[0]+(0.306016568948)*x_ref[1]
        ref[2,0,1,0]=(-0.284264003593)*x_ref[0]+(-0.899022593367)*x_ref[1]
        ref[2,0,1,1]=(-0.0439761413236)*x_ref[0]+(-0.696386615358)*x_ref[1]
        ref[2,0,2,0]=(0.407393812349)*x_ref[0]+(-1.04815464317)*x_ref[1]
        ref[2,0,2,1]=(-0.956661282071)*x_ref[0]+(0.243281223782)*x_ref[1]
        ref[2,1,0,0]=(-1.29201911939)*x_ref[0]+(-0.693049706038)*x_ref[1]
        ref[2,1,0,1]=(-0.711059848853)*x_ref[0]+(-1.14204907272)*x_ref[1]
        ref[2,1,1,0]=(-0.347693333057)*x_ref[0]+(0.818879269672)*x_ref[1]
        ref[2,1,1,1]=(0.00742807299875)*x_ref[0]+(0.138845309984)*x_ref[1]
        ref[2,1,2,0]=(1.13412244048)*x_ref[0]+(-0.481422214102)*x_ref[1]
        ref[2,1,2,1]=(0.64440302826)*x_ref[0]+(-0.955231893089)*x_ref[1]
        ref[2,2,0,0]=(0.13123141544)*x_ref[0]+(-0.602266537541)*x_ref[1]
        ref[2,2,0,1]=(-0.204818071763)*x_ref[0]+(-1.13480248293)*x_ref[1]
        ref[2,2,1,0]=(0.781058287579)*x_ref[0]+(-1.31936224435)*x_ref[1]
        ref[2,2,1,1]=(-0.591275351662)*x_ref[0]+(0.691816969517)*x_ref[1]
        ref[2,2,2,0]=(-0.921964826241)*x_ref[0]+(1.28687075541)*x_ref[1]
        ref[2,2,2,1]=(0.246260462337)*x_ref[0]+(-1.04699282194)*x_ref[1]
        ref[2,3,0,0]=(0.264209126513)*x_ref[0]+(-0.697094400566)*x_ref[1]
        ref[2,3,0,1]=(0.456158298484)*x_ref[0]+(0.0288036445565)*x_ref[1]
        ref[2,3,1,0]=(-0.720337925392)*x_ref[0]+(0.0965182642186)*x_ref[1]
        ref[2,3,1,1]=(0.397480419547)*x_ref[0]+(-1.55232798344)*x_ref[1]
        ref[2,3,2,0]=(-0.337139999283)*x_ref[0]+(0.918809751807)*x_ref[1]
        ref[2,3,2,1]=(1.34009447198)*x_ref[0]+(-1.41390962018)*x_ref[1]
        ref[2,4,0,0]=(1.67781838006)*x_ref[0]+(0.521276193814)*x_ref[1]
        ref[2,4,0,1]=(-0.114014321939)*x_ref[0]+(0.529316778773)*x_ref[1]
        ref[2,4,1,0]=(-0.19938131807)*x_ref[0]+(-0.0712417614843)*x_ref[1]
        ref[2,4,1,1]=(-0.724676753871)*x_ref[0]+(0.731341170171)*x_ref[1]
        ref[2,4,2,0]=(0.988755659656)*x_ref[0]+(-1.0327352189)*x_ref[1]
        ref[2,4,2,1]=(-0.526286732859)*x_ref[0]+(-1.18295671598)*x_ref[1]
        ref[3,0,0,0]=(-0.685223304723)*x_ref[0]+(0.508431586068)*x_ref[1]
        ref[3,0,0,1]=(0.529240274944)*x_ref[0]+(0.77519128969)*x_ref[1]
        ref[3,0,1,0]=(-0.757098431928)*x_ref[0]+(1.41375812725)*x_ref[1]
        ref[3,0,1,1]=(1.20805599335)*x_ref[0]+(-0.184660334182)*x_ref[1]
        ref[3,0,2,0]=(0.718413765561)*x_ref[0]+(0.971149168706)*x_ref[1]
        ref[3,0,2,1]=(-0.632670534518)*x_ref[0]+(0.0389618967363)*x_ref[1]
        ref[3,1,0,0]=(-0.525814673755)*x_ref[0]+(0.574122587598)*x_ref[1]
        ref[3,1,0,1]=(-1.04117288975)*x_ref[0]+(-1.00738089823)*x_ref[1]
        ref[3,1,1,0]=(0.0312806064182)*x_ref[0]+(-0.30773511847)*x_ref[1]
        ref[3,1,1,1]=(0.889930889355)*x_ref[0]+(0.833689153489)*x_ref[1]
        ref[3,1,2,0]=(-1.21586293381)*x_ref[0]+(0.512182349574)*x_ref[1]
        ref[3,1,2,1]=(-0.00437508755415)*x_ref[0]+(-1.13228664497)*x_ref[1]
        ref[3,2,0,0]=(1.64174174554)*x_ref[0]+(-0.696500062562)*x_ref[1]
        ref[3,2,0,1]=(0.0265343210604)*x_ref[0]+(-0.707378716636)*x_ref[1]
        ref[3,2,1,0]=(0.952922122967)*x_ref[0]+(0.0946904561805)*x_ref[1]
        ref[3,2,1,1]=(-1.06905266456)*x_ref[0]+(-0.155122789807)*x_ref[1]
        ref[3,2,2,0]=(-1.14410789976)*x_ref[0]+(-0.379020164446)*x_ref[1]
        ref[3,2,2,1]=(-0.0922611198553)*x_ref[0]+(1.36680699099)*x_ref[1]
        ref[3,3,0,0]=(0.849028205422)*x_ref[0]+(-0.968698293811)*x_ref[1]
        ref[3,3,0,1]=(0.649894718675)*x_ref[0]+(0.238249259292)*x_ref[1]
        ref[3,3,1,0]=(1.28170508471)*x_ref[0]+(-0.736544947537)*x_ref[1]
        ref[3,3,1,1]=(-0.188411685414)*x_ref[0]+(1.14145010691)*x_ref[1]
        ref[3,3,2,0]=(-1.13973037849)*x_ref[0]+(-0.256565914821)*x_ref[1]
        ref[3,3,2,1]=(0.955478585358)*x_ref[0]+(-0.83417326515)*x_ref[1]
        ref[3,4,0,0]=(0.155710789468)*x_ref[0]+(0.204956811648)*x_ref[1]
        ref[3,4,0,1]=(0.567866952974)*x_ref[0]+(-0.168398320099)*x_ref[1]
        ref[3,4,1,0]=(-1.36997367)*x_ref[0]+(0.141984040172)*x_ref[1]
        ref[3,4,1,1]=(-0.809532285206)*x_ref[0]+(0.70611485405)*x_ref[1]
        ref[3,4,2,0]=(-0.453260232051)*x_ref[0]+(0.0665127865858)*x_ref[1]
        ref[3,4,2,1]=(0.00142021115277)*x_ref[0]+(1.02307505343)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.59037121238)*x[0]+(-0.432259357387)*x[1]+(1.06313464506)*x[2]
        arg[0,0,0,1]=(-0.29317617186)*x[0]+(-0.477878024138)*x[1]+(-0.943279870083)*x[2]
        arg[0,0,1,0]=(-0.0140970955833)*x[0]+(0.141752937281)*x[1]+(0.46748311988)*x[2]
        arg[0,0,1,1]=(1.06645552756)*x[0]+(-1.4360320823)*x[1]+(-1.05643251873)*x[2]
        arg[0,0,2,0]=(-0.527854671619)*x[0]+(0.68918711423)*x[1]+(0.225750468582)*x[2]
        arg[0,0,2,1]=(0.765285169635)*x[0]+(-0.292098568154)*x[1]+(-0.611702446821)*x[2]
        arg[0,1,0,0]=(0.472029729041)*x[0]+(-0.391383508709)*x[1]+(-0.0882199174494)*x[2]
        arg[0,1,0,1]=(1.76010861737)*x[0]+(1.40668629263)*x[1]+(-0.171024087043)*x[2]
        arg[0,1,1,0]=(0.251929705017)*x[0]+(-0.326420063624)*x[1]+(-0.337849942832)*x[2]
        arg[0,1,1,1]=(0.049289902392)*x[0]+(1.61775987596)*x[1]+(0.545741506198)*x[2]
        arg[0,1,2,0]=(-0.500710333189)*x[0]+(-0.551222136922)*x[1]+(0.665856619864)*x[2]
        arg[0,1,2,1]=(0.271547935249)*x[0]+(0.46154733209)*x[1]+(0.0776993672589)*x[2]
        arg[0,2,0,0]=(0.46744023541)*x[0]+(0.281754164985)*x[1]+(0.11721465855)*x[2]
        arg[0,2,0,1]=(0.58865724402)*x[0]+(-0.280911783535)*x[1]+(0.793461201474)*x[2]
        arg[0,2,1,0]=(-0.348215254122)*x[0]+(-0.274357607001)*x[1]+(-0.975813503666)*x[2]
        arg[0,2,1,1]=(-0.546269224094)*x[0]+(0.54132986357)*x[1]+(-0.346797273682)*x[2]
        arg[0,2,2,0]=(0.956395362583)*x[0]+(-0.609333787208)*x[1]+(-0.86408067145)*x[2]
        arg[0,2,2,1]=(0.570075324955)*x[0]+(-0.251638346903)*x[1]+(0.292841304398)*x[2]
        arg[0,3,0,0]=(-0.240717117869)*x[0]+(0.571307481772)*x[1]+(0.66826910816)*x[2]
        arg[0,3,0,1]=(0.113871023723)*x[0]+(0.022498355636)*x[1]+(0.95331234379)*x[2]
        arg[0,3,1,0]=(-0.264946899001)*x[0]+(1.08412373438)*x[1]+(-1.10548347411)*x[2]
        arg[0,3,1,1]=(-0.744510132854)*x[0]+(-0.716638983187)*x[1]+(0.64330838226)*x[2]
        arg[0,3,2,0]=(1.15237343787)*x[0]+(-1.42565901229)*x[1]+(0.727585401396)*x[2]
        arg[0,3,2,1]=(-0.251464863627)*x[0]+(1.4067624912)*x[1]+(0.674418739118)*x[2]
        arg[0,4,0,0]=(-0.00551820316191)*x[0]+(0.193862853101)*x[1]+(1.8004924522)*x[2]
        arg[0,4,0,1]=(-0.985879358916)*x[0]+(0.476965815991)*x[1]+(0.96363449822)*x[2]
        arg[0,4,1,0]=(-0.657967361584)*x[0]+(0.833452284341)*x[1]+(-1.40128003822)*x[2]
        arg[0,4,1,1]=(0.301379967988)*x[0]+(0.0474545542353)*x[1]+(-0.633087920569)*x[2]
        arg[0,4,2,0]=(-1.94409755409)*x[0]+(-0.468638363928)*x[1]+(-0.713880237452)*x[2]
        arg[0,4,2,1]=(-1.13536316065)*x[0]+(-0.250761280127)*x[1]+(0.328722833182)*x[2]
        arg[1,0,0,0]=(0.556886266554)*x[0]+(-0.70523094638)*x[1]+(-0.333621268224)*x[2]
        arg[1,0,0,1]=(-0.995565683628)*x[0]+(-0.158120855239)*x[1]+(0.0738870470169)*x[2]
        arg[1,0,1,0]=(-0.209556712574)*x[0]+(0.54393408711)*x[1]+(-1.1617498966)*x[2]
        arg[1,0,1,1]=(0.666809649924)*x[0]+(1.18555984779)*x[1]+(-0.442831598796)*x[2]
        arg[1,0,2,0]=(-0.21354395591)*x[0]+(-0.634626698269)*x[1]+(1.25643757008)*x[2]
        arg[1,0,2,1]=(-0.676259187073)*x[0]+(-0.458532774397)*x[1]+(0.271156919902)*x[2]
        arg[1,1,0,0]=(-1.45529139262)*x[0]+(0.311382249732)*x[1]+(-0.134620909882)*x[2]
        arg[1,1,0,1]=(-1.47708829198)*x[0]+(0.968291536601)*x[1]+(-0.566606887608)*x[2]
        arg[1,1,1,0]=(1.18904329045)*x[0]+(-0.209072247061)*x[1]+(-0.26261457093)*x[2]
        arg[1,1,1,1]=(0.313723817215)*x[0]+(-0.169177309423)*x[1]+(-0.00119161234247)*x[2]
        arg[1,1,2,0]=(0.431298197457)*x[0]+(0.264113678484)*x[1]+(0.192619025313)*x[2]
        arg[1,1,2,1]=(0.0469221259627)*x[0]+(1.53477696238)*x[1]+(0.00794662724457)*x[2]
        arg[1,2,0,0]=(0.102179220279)*x[0]+(-0.391365988191)*x[1]+(0.302910557263)*x[2]
        arg[1,2,0,1]=(0.841803002697)*x[0]+(0.581539244301)*x[1]+(0.00409108885792)*x[2]
        arg[1,2,1,0]=(0.0840378161144)*x[0]+(-0.431329807551)*x[1]+(-1.0774029159)*x[2]
        arg[1,2,1,1]=(-0.270803943532)*x[0]+(0.0124514531652)*x[1]+(0.197735227059)*x[2]
        arg[1,2,2,0]=(0.479336250918)*x[0]+(0.0435331360815)*x[1]+(0.458048853417)*x[2]
        arg[1,2,2,1]=(-0.59916419564)*x[0]+(0.415546807791)*x[1]+(-1.12105410423)*x[2]
        arg[1,3,0,0]=(0.378277354507)*x[0]+(1.00734343779)*x[1]+(-0.282811389334)*x[2]
        arg[1,3,0,1]=(0.717420487474)*x[0]+(-0.730395427407)*x[1]+(1.63850706843)*x[2]
        arg[1,3,1,0]=(-0.0126954991872)*x[0]+(-1.35916858377)*x[1]+(-0.955731750212)*x[2]
        arg[1,3,1,1]=(-0.560951559862)*x[0]+(0.271085100351)*x[1]+(0.60767567276)*x[2]
        arg[1,3,2,0]=(-0.73286080289)*x[0]+(-1.12018839274)*x[1]+(0.365586479937)*x[2]
        arg[1,3,2,1]=(-0.495391543232)*x[0]+(-0.404605421036)*x[1]+(1.21261493791)*x[2]
        arg[1,4,0,0]=(-0.322147956238)*x[0]+(0.278257549185)*x[1]+(-0.927648624956)*x[2]
        arg[1,4,0,1]=(1.23243511737)*x[0]+(1.38355111729)*x[1]+(0.0813739945129)*x[2]
        arg[1,4,1,0]=(0.668384291635)*x[0]+(0.995328961231)*x[1]+(0.803009543307)*x[2]
        arg[1,4,1,1]=(1.42643018603)*x[0]+(0.940855692156)*x[1]+(0.277556451163)*x[2]
        arg[1,4,2,0]=(0.489588226757)*x[0]+(-1.63608809124)*x[1]+(-0.262918068522)*x[2]
        arg[1,4,2,1]=(-1.06922340601)*x[0]+(-0.0241938384008)*x[1]+(-0.244107290038)*x[2]
        arg[2,0,0,0]=(-0.556338154459)*x[0]+(0.12071126785)*x[1]+(0.11215804852)*x[2]
        arg[2,0,0,1]=(-0.479017339774)*x[0]+(-0.530313718209)*x[1]+(0.490083551372)*x[2]
        arg[2,0,1,0]=(0.731269264444)*x[0]+(0.245793038193)*x[1]+(0.0479833912422)*x[2]
        arg[2,0,1,1]=(-0.286189382681)*x[0]+(0.601756275554)*x[1]+(-0.559334203137)*x[2]
        arg[2,0,2,0]=(-0.876495720673)*x[0]+(0.592513223261)*x[1]+(0.844551261652)*x[2]
        arg[2,0,2,1]=(1.22915142187)*x[0]+(0.136223623577)*x[1]+(0.88656874029)*x[2]
        arg[2,1,0,0]=(0.542222650916)*x[0]+(-0.709719163926)*x[1]+(0.262207188498)*x[2]
        arg[2,1,0,1]=(-0.208896980258)*x[0]+(1.33968890395)*x[1]+(-0.599621090138)*x[2]
        arg[2,1,1,0]=(-0.171147406604)*x[0]+(0.332227749427)*x[1]+(1.15043521586)*x[2]
        arg[2,1,1,1]=(0.571607567027)*x[0]+(0.1367434941)*x[1]+(1.47797159814)*x[2]
        arg[2,1,2,0]=(0.839845574634)*x[0]+(0.732437808234)*x[1]+(-0.0311071654701)*x[2]
        arg[2,1,2,1]=(1.45193213653)*x[0]+(-0.0192312612414)*x[1]+(0.308521161282)*x[2]
        arg[2,2,0,0]=(-0.122755384028)*x[0]+(0.133510038844)*x[1]+(0.837213365425)*x[2]
        arg[2,2,0,1]=(-0.411967923958)*x[0]+(-0.955489021071)*x[1]+(0.269005602186)*x[2]
        arg[2,2,1,0]=(-1.19993089382)*x[0]+(0.276541296246)*x[1]+(0.523388479396)*x[2]
        arg[2,2,1,1]=(0.0586514653869)*x[0]+(-0.624975232742)*x[1]+(-1.55738876751)*x[2]
        arg[2,2,2,0]=(-0.892065938887)*x[0]+(-0.4848241954)*x[1]+(0.315463883511)*x[2]
        arg[2,2,2,1]=(0.0272963181118)*x[0]+(0.0187713454669)*x[1]+(-1.06524661854)*x[2]
        arg[2,3,0,0]=(-1.44099287886)*x[0]+(0.495718421075)*x[1]+(0.100401960223)*x[2]
        arg[2,3,0,1]=(0.171175119412)*x[0]+(-1.34795558607)*x[1]+(-0.361766653264)*x[2]
        arg[2,3,1,0]=(-1.80591882588)*x[0]+(-0.511459455136)*x[1]+(-0.491310264055)*x[2]
        arg[2,3,1,1]=(-0.419508934622)*x[0]+(-1.12084042844)*x[1]+(-0.086947672548)*x[2]
        arg[2,3,2,0]=(0.492388303934)*x[0]+(-0.623505828597)*x[1]+(1.6965873745)*x[2]
        arg[2,3,2,1]=(1.59203810356)*x[0]+(0.36918142231)*x[1]+(-1.90067115052)*x[2]
        arg[2,4,0,0]=(1.00875649298)*x[0]+(-1.43246384934)*x[1]+(-0.51811906593)*x[2]
        arg[2,4,0,1]=(0.260118918034)*x[0]+(-1.57606061865)*x[1]+(-0.79257165655)*x[2]
        arg[2,4,1,0]=(0.954414074467)*x[0]+(-0.230132876613)*x[1]+(0.0956378957424)*x[2]
        arg[2,4,1,1]=(0.958843806364)*x[0]+(-0.155804429995)*x[1]+(-1.02904877039)*x[2]
        arg[2,4,2,0]=(0.252668700811)*x[0]+(0.164829885749)*x[1]+(0.747032722421)*x[2]
        arg[2,4,2,1]=(1.96393052323)*x[0]+(-0.821363958807)*x[1]+(-1.01663555063)*x[2]
        arg[3,0,0,0]=(1.58086037053)*x[0]+(-1.66495295611)*x[1]+(-0.510526191383)*x[2]
        arg[3,0,0,1]=(-0.931871976029)*x[0]+(1.33102205125)*x[1]+(-0.95170580468)*x[2]
        arg[3,0,1,0]=(0.99825404115)*x[0]+(0.915274413552)*x[1]+(1.5057227674)*x[2]
        arg[3,0,1,1]=(0.0485551924989)*x[0]+(-1.0877405758)*x[1]+(-0.887342573702)*x[2]
        arg[3,0,2,0]=(-0.745146140677)*x[0]+(-0.0108141266015)*x[1]+(0.194597576308)*x[2]
        arg[3,0,2,1]=(-0.935029218338)*x[0]+(0.726322809025)*x[1]+(0.57508467103)*x[2]
        arg[3,1,0,0]=(0.115668237156)*x[0]+(0.0970255991906)*x[1]+(-0.227291123547)*x[2]
        arg[3,1,0,1]=(0.167722335863)*x[0]+(-0.552168424291)*x[1]+(1.36293027463)*x[2]
        arg[3,1,1,0]=(-1.05781878028)*x[0]+(-0.0429389047972)*x[1]+(-0.257166337509)*x[2]
        arg[3,1,1,1]=(0.129148486118)*x[0]+(-0.0209434899369)*x[1]+(0.253153005253)*x[2]
        arg[3,1,2,0]=(1.3236615365)*x[0]+(-1.18164469224)*x[1]+(0.698012697426)*x[2]
        arg[3,1,2,1]=(1.10438661886)*x[0]+(-0.852903948395)*x[1]+(1.40365162024)*x[2]
        arg[3,2,0,0]=(-0.949308009957)*x[0]+(1.01993206161)*x[1]+(-0.821721697476)*x[2]
        arg[3,2,0,1]=(0.326450662787)*x[0]+(0.395096088107)*x[1]+(-0.368904800584)*x[2]
        arg[3,2,1,0]=(0.13107780566)*x[0]+(-0.342904134623)*x[1]+(0.00835087701185)*x[2]
        arg[3,2,1,1]=(-0.500694550647)*x[0]+(1.38322390003)*x[1]+(0.0576453183412)*x[2]
        arg[3,2,2,0]=(-1.32014489811)*x[0]+(-0.444079317864)*x[1]+(1.1135946018)*x[2]
        arg[3,2,2,1]=(-0.19593105007)*x[0]+(-0.00224390969377)*x[1]+(-0.73869338942)*x[2]
        arg[3,3,0,0]=(0.0151201622255)*x[0]+(-1.10873788153)*x[1]+(0.189116254832)*x[2]
        arg[3,3,0,1]=(1.05996205566)*x[0]+(0.398465639867)*x[1]+(-0.319773104448)*x[2]
        arg[3,3,1,0]=(0.687612827243)*x[0]+(1.10970823257)*x[1]+(1.02543288975)*x[2]
        arg[3,3,1,1]=(1.27258509981)*x[0]+(-1.7060826391)*x[1]+(-0.84948812227)*x[2]
        arg[3,3,2,0]=(1.2343036886)*x[0]+(0.521090253023)*x[1]+(-1.13377412954)*x[2]
        arg[3,3,2,1]=(-0.65568738385)*x[0]+(1.00354617712)*x[1]+(0.836719015812)*x[2]
        arg[3,4,0,0]=(-0.545324078594)*x[0]+(0.452747782208)*x[1]+(-1.59718954237)*x[2]
        arg[3,4,0,1]=(1.71651153317)*x[0]+(-0.161984538787)*x[1]+(-0.377244827904)*x[2]
        arg[3,4,1,0]=(-0.284033324558)*x[0]+(-0.405941433062)*x[1]+(0.824573264594)*x[2]
        arg[3,4,1,1]=(-0.112542702803)*x[0]+(-0.470814805786)*x[1]+(-0.827427436576)*x[2]
        arg[3,4,2,0]=(-0.763655132981)*x[0]+(-0.442818775864)*x[1]+(0.809814510041)*x[2]
        arg[3,4,2,1]=(-0.33794306298)*x[0]+(1.91169591332)*x[1]+(-0.487268088196)*x[2]
        ref[0,0,0,0]=(-0.59037121238)*x_ref[0]+(-0.432259357387)*x_ref[1]+(1.06313464506)*x_ref[2]
        ref[0,0,0,1]=(-0.29317617186)*x_ref[0]+(-0.477878024138)*x_ref[1]+(-0.943279870083)*x_ref[2]
        ref[0,0,1,0]=(-0.0140970955833)*x_ref[0]+(0.141752937281)*x_ref[1]+(0.46748311988)*x_ref[2]
        ref[0,0,1,1]=(1.06645552756)*x_ref[0]+(-1.4360320823)*x_ref[1]+(-1.05643251873)*x_ref[2]
        ref[0,0,2,0]=(-0.527854671619)*x_ref[0]+(0.68918711423)*x_ref[1]+(0.225750468582)*x_ref[2]
        ref[0,0,2,1]=(0.765285169635)*x_ref[0]+(-0.292098568154)*x_ref[1]+(-0.611702446821)*x_ref[2]
        ref[0,1,0,0]=(0.472029729041)*x_ref[0]+(-0.391383508709)*x_ref[1]+(-0.0882199174494)*x_ref[2]
        ref[0,1,0,1]=(1.76010861737)*x_ref[0]+(1.40668629263)*x_ref[1]+(-0.171024087043)*x_ref[2]
        ref[0,1,1,0]=(0.251929705017)*x_ref[0]+(-0.326420063624)*x_ref[1]+(-0.337849942832)*x_ref[2]
        ref[0,1,1,1]=(0.049289902392)*x_ref[0]+(1.61775987596)*x_ref[1]+(0.545741506198)*x_ref[2]
        ref[0,1,2,0]=(-0.500710333189)*x_ref[0]+(-0.551222136922)*x_ref[1]+(0.665856619864)*x_ref[2]
        ref[0,1,2,1]=(0.271547935249)*x_ref[0]+(0.46154733209)*x_ref[1]+(0.0776993672589)*x_ref[2]
        ref[0,2,0,0]=(0.46744023541)*x_ref[0]+(0.281754164985)*x_ref[1]+(0.11721465855)*x_ref[2]
        ref[0,2,0,1]=(0.58865724402)*x_ref[0]+(-0.280911783535)*x_ref[1]+(0.793461201474)*x_ref[2]
        ref[0,2,1,0]=(-0.348215254122)*x_ref[0]+(-0.274357607001)*x_ref[1]+(-0.975813503666)*x_ref[2]
        ref[0,2,1,1]=(-0.546269224094)*x_ref[0]+(0.54132986357)*x_ref[1]+(-0.346797273682)*x_ref[2]
        ref[0,2,2,0]=(0.956395362583)*x_ref[0]+(-0.609333787208)*x_ref[1]+(-0.86408067145)*x_ref[2]
        ref[0,2,2,1]=(0.570075324955)*x_ref[0]+(-0.251638346903)*x_ref[1]+(0.292841304398)*x_ref[2]
        ref[0,3,0,0]=(-0.240717117869)*x_ref[0]+(0.571307481772)*x_ref[1]+(0.66826910816)*x_ref[2]
        ref[0,3,0,1]=(0.113871023723)*x_ref[0]+(0.022498355636)*x_ref[1]+(0.95331234379)*x_ref[2]
        ref[0,3,1,0]=(-0.264946899001)*x_ref[0]+(1.08412373438)*x_ref[1]+(-1.10548347411)*x_ref[2]
        ref[0,3,1,1]=(-0.744510132854)*x_ref[0]+(-0.716638983187)*x_ref[1]+(0.64330838226)*x_ref[2]
        ref[0,3,2,0]=(1.15237343787)*x_ref[0]+(-1.42565901229)*x_ref[1]+(0.727585401396)*x_ref[2]
        ref[0,3,2,1]=(-0.251464863627)*x_ref[0]+(1.4067624912)*x_ref[1]+(0.674418739118)*x_ref[2]
        ref[0,4,0,0]=(-0.00551820316191)*x_ref[0]+(0.193862853101)*x_ref[1]+(1.8004924522)*x_ref[2]
        ref[0,4,0,1]=(-0.985879358916)*x_ref[0]+(0.476965815991)*x_ref[1]+(0.96363449822)*x_ref[2]
        ref[0,4,1,0]=(-0.657967361584)*x_ref[0]+(0.833452284341)*x_ref[1]+(-1.40128003822)*x_ref[2]
        ref[0,4,1,1]=(0.301379967988)*x_ref[0]+(0.0474545542353)*x_ref[1]+(-0.633087920569)*x_ref[2]
        ref[0,4,2,0]=(-1.94409755409)*x_ref[0]+(-0.468638363928)*x_ref[1]+(-0.713880237452)*x_ref[2]
        ref[0,4,2,1]=(-1.13536316065)*x_ref[0]+(-0.250761280127)*x_ref[1]+(0.328722833182)*x_ref[2]
        ref[1,0,0,0]=(0.556886266554)*x_ref[0]+(-0.70523094638)*x_ref[1]+(-0.333621268224)*x_ref[2]
        ref[1,0,0,1]=(-0.995565683628)*x_ref[0]+(-0.158120855239)*x_ref[1]+(0.0738870470169)*x_ref[2]
        ref[1,0,1,0]=(-0.209556712574)*x_ref[0]+(0.54393408711)*x_ref[1]+(-1.1617498966)*x_ref[2]
        ref[1,0,1,1]=(0.666809649924)*x_ref[0]+(1.18555984779)*x_ref[1]+(-0.442831598796)*x_ref[2]
        ref[1,0,2,0]=(-0.21354395591)*x_ref[0]+(-0.634626698269)*x_ref[1]+(1.25643757008)*x_ref[2]
        ref[1,0,2,1]=(-0.676259187073)*x_ref[0]+(-0.458532774397)*x_ref[1]+(0.271156919902)*x_ref[2]
        ref[1,1,0,0]=(-1.45529139262)*x_ref[0]+(0.311382249732)*x_ref[1]+(-0.134620909882)*x_ref[2]
        ref[1,1,0,1]=(-1.47708829198)*x_ref[0]+(0.968291536601)*x_ref[1]+(-0.566606887608)*x_ref[2]
        ref[1,1,1,0]=(1.18904329045)*x_ref[0]+(-0.209072247061)*x_ref[1]+(-0.26261457093)*x_ref[2]
        ref[1,1,1,1]=(0.313723817215)*x_ref[0]+(-0.169177309423)*x_ref[1]+(-0.00119161234247)*x_ref[2]
        ref[1,1,2,0]=(0.431298197457)*x_ref[0]+(0.264113678484)*x_ref[1]+(0.192619025313)*x_ref[2]
        ref[1,1,2,1]=(0.0469221259627)*x_ref[0]+(1.53477696238)*x_ref[1]+(0.00794662724457)*x_ref[2]
        ref[1,2,0,0]=(0.102179220279)*x_ref[0]+(-0.391365988191)*x_ref[1]+(0.302910557263)*x_ref[2]
        ref[1,2,0,1]=(0.841803002697)*x_ref[0]+(0.581539244301)*x_ref[1]+(0.00409108885792)*x_ref[2]
        ref[1,2,1,0]=(0.0840378161144)*x_ref[0]+(-0.431329807551)*x_ref[1]+(-1.0774029159)*x_ref[2]
        ref[1,2,1,1]=(-0.270803943532)*x_ref[0]+(0.0124514531652)*x_ref[1]+(0.197735227059)*x_ref[2]
        ref[1,2,2,0]=(0.479336250918)*x_ref[0]+(0.0435331360815)*x_ref[1]+(0.458048853417)*x_ref[2]
        ref[1,2,2,1]=(-0.59916419564)*x_ref[0]+(0.415546807791)*x_ref[1]+(-1.12105410423)*x_ref[2]
        ref[1,3,0,0]=(0.378277354507)*x_ref[0]+(1.00734343779)*x_ref[1]+(-0.282811389334)*x_ref[2]
        ref[1,3,0,1]=(0.717420487474)*x_ref[0]+(-0.730395427407)*x_ref[1]+(1.63850706843)*x_ref[2]
        ref[1,3,1,0]=(-0.0126954991872)*x_ref[0]+(-1.35916858377)*x_ref[1]+(-0.955731750212)*x_ref[2]
        ref[1,3,1,1]=(-0.560951559862)*x_ref[0]+(0.271085100351)*x_ref[1]+(0.60767567276)*x_ref[2]
        ref[1,3,2,0]=(-0.73286080289)*x_ref[0]+(-1.12018839274)*x_ref[1]+(0.365586479937)*x_ref[2]
        ref[1,3,2,1]=(-0.495391543232)*x_ref[0]+(-0.404605421036)*x_ref[1]+(1.21261493791)*x_ref[2]
        ref[1,4,0,0]=(-0.322147956238)*x_ref[0]+(0.278257549185)*x_ref[1]+(-0.927648624956)*x_ref[2]
        ref[1,4,0,1]=(1.23243511737)*x_ref[0]+(1.38355111729)*x_ref[1]+(0.0813739945129)*x_ref[2]
        ref[1,4,1,0]=(0.668384291635)*x_ref[0]+(0.995328961231)*x_ref[1]+(0.803009543307)*x_ref[2]
        ref[1,4,1,1]=(1.42643018603)*x_ref[0]+(0.940855692156)*x_ref[1]+(0.277556451163)*x_ref[2]
        ref[1,4,2,0]=(0.489588226757)*x_ref[0]+(-1.63608809124)*x_ref[1]+(-0.262918068522)*x_ref[2]
        ref[1,4,2,1]=(-1.06922340601)*x_ref[0]+(-0.0241938384008)*x_ref[1]+(-0.244107290038)*x_ref[2]
        ref[2,0,0,0]=(-0.556338154459)*x_ref[0]+(0.12071126785)*x_ref[1]+(0.11215804852)*x_ref[2]
        ref[2,0,0,1]=(-0.479017339774)*x_ref[0]+(-0.530313718209)*x_ref[1]+(0.490083551372)*x_ref[2]
        ref[2,0,1,0]=(0.731269264444)*x_ref[0]+(0.245793038193)*x_ref[1]+(0.0479833912422)*x_ref[2]
        ref[2,0,1,1]=(-0.286189382681)*x_ref[0]+(0.601756275554)*x_ref[1]+(-0.559334203137)*x_ref[2]
        ref[2,0,2,0]=(-0.876495720673)*x_ref[0]+(0.592513223261)*x_ref[1]+(0.844551261652)*x_ref[2]
        ref[2,0,2,1]=(1.22915142187)*x_ref[0]+(0.136223623577)*x_ref[1]+(0.88656874029)*x_ref[2]
        ref[2,1,0,0]=(0.542222650916)*x_ref[0]+(-0.709719163926)*x_ref[1]+(0.262207188498)*x_ref[2]
        ref[2,1,0,1]=(-0.208896980258)*x_ref[0]+(1.33968890395)*x_ref[1]+(-0.599621090138)*x_ref[2]
        ref[2,1,1,0]=(-0.171147406604)*x_ref[0]+(0.332227749427)*x_ref[1]+(1.15043521586)*x_ref[2]
        ref[2,1,1,1]=(0.571607567027)*x_ref[0]+(0.1367434941)*x_ref[1]+(1.47797159814)*x_ref[2]
        ref[2,1,2,0]=(0.839845574634)*x_ref[0]+(0.732437808234)*x_ref[1]+(-0.0311071654701)*x_ref[2]
        ref[2,1,2,1]=(1.45193213653)*x_ref[0]+(-0.0192312612414)*x_ref[1]+(0.308521161282)*x_ref[2]
        ref[2,2,0,0]=(-0.122755384028)*x_ref[0]+(0.133510038844)*x_ref[1]+(0.837213365425)*x_ref[2]
        ref[2,2,0,1]=(-0.411967923958)*x_ref[0]+(-0.955489021071)*x_ref[1]+(0.269005602186)*x_ref[2]
        ref[2,2,1,0]=(-1.19993089382)*x_ref[0]+(0.276541296246)*x_ref[1]+(0.523388479396)*x_ref[2]
        ref[2,2,1,1]=(0.0586514653869)*x_ref[0]+(-0.624975232742)*x_ref[1]+(-1.55738876751)*x_ref[2]
        ref[2,2,2,0]=(-0.892065938887)*x_ref[0]+(-0.4848241954)*x_ref[1]+(0.315463883511)*x_ref[2]
        ref[2,2,2,1]=(0.0272963181118)*x_ref[0]+(0.0187713454669)*x_ref[1]+(-1.06524661854)*x_ref[2]
        ref[2,3,0,0]=(-1.44099287886)*x_ref[0]+(0.495718421075)*x_ref[1]+(0.100401960223)*x_ref[2]
        ref[2,3,0,1]=(0.171175119412)*x_ref[0]+(-1.34795558607)*x_ref[1]+(-0.361766653264)*x_ref[2]
        ref[2,3,1,0]=(-1.80591882588)*x_ref[0]+(-0.511459455136)*x_ref[1]+(-0.491310264055)*x_ref[2]
        ref[2,3,1,1]=(-0.419508934622)*x_ref[0]+(-1.12084042844)*x_ref[1]+(-0.086947672548)*x_ref[2]
        ref[2,3,2,0]=(0.492388303934)*x_ref[0]+(-0.623505828597)*x_ref[1]+(1.6965873745)*x_ref[2]
        ref[2,3,2,1]=(1.59203810356)*x_ref[0]+(0.36918142231)*x_ref[1]+(-1.90067115052)*x_ref[2]
        ref[2,4,0,0]=(1.00875649298)*x_ref[0]+(-1.43246384934)*x_ref[1]+(-0.51811906593)*x_ref[2]
        ref[2,4,0,1]=(0.260118918034)*x_ref[0]+(-1.57606061865)*x_ref[1]+(-0.79257165655)*x_ref[2]
        ref[2,4,1,0]=(0.954414074467)*x_ref[0]+(-0.230132876613)*x_ref[1]+(0.0956378957424)*x_ref[2]
        ref[2,4,1,1]=(0.958843806364)*x_ref[0]+(-0.155804429995)*x_ref[1]+(-1.02904877039)*x_ref[2]
        ref[2,4,2,0]=(0.252668700811)*x_ref[0]+(0.164829885749)*x_ref[1]+(0.747032722421)*x_ref[2]
        ref[2,4,2,1]=(1.96393052323)*x_ref[0]+(-0.821363958807)*x_ref[1]+(-1.01663555063)*x_ref[2]
        ref[3,0,0,0]=(1.58086037053)*x_ref[0]+(-1.66495295611)*x_ref[1]+(-0.510526191383)*x_ref[2]
        ref[3,0,0,1]=(-0.931871976029)*x_ref[0]+(1.33102205125)*x_ref[1]+(-0.95170580468)*x_ref[2]
        ref[3,0,1,0]=(0.99825404115)*x_ref[0]+(0.915274413552)*x_ref[1]+(1.5057227674)*x_ref[2]
        ref[3,0,1,1]=(0.0485551924989)*x_ref[0]+(-1.0877405758)*x_ref[1]+(-0.887342573702)*x_ref[2]
        ref[3,0,2,0]=(-0.745146140677)*x_ref[0]+(-0.0108141266015)*x_ref[1]+(0.194597576308)*x_ref[2]
        ref[3,0,2,1]=(-0.935029218338)*x_ref[0]+(0.726322809025)*x_ref[1]+(0.57508467103)*x_ref[2]
        ref[3,1,0,0]=(0.115668237156)*x_ref[0]+(0.0970255991906)*x_ref[1]+(-0.227291123547)*x_ref[2]
        ref[3,1,0,1]=(0.167722335863)*x_ref[0]+(-0.552168424291)*x_ref[1]+(1.36293027463)*x_ref[2]
        ref[3,1,1,0]=(-1.05781878028)*x_ref[0]+(-0.0429389047972)*x_ref[1]+(-0.257166337509)*x_ref[2]
        ref[3,1,1,1]=(0.129148486118)*x_ref[0]+(-0.0209434899369)*x_ref[1]+(0.253153005253)*x_ref[2]
        ref[3,1,2,0]=(1.3236615365)*x_ref[0]+(-1.18164469224)*x_ref[1]+(0.698012697426)*x_ref[2]
        ref[3,1,2,1]=(1.10438661886)*x_ref[0]+(-0.852903948395)*x_ref[1]+(1.40365162024)*x_ref[2]
        ref[3,2,0,0]=(-0.949308009957)*x_ref[0]+(1.01993206161)*x_ref[1]+(-0.821721697476)*x_ref[2]
        ref[3,2,0,1]=(0.326450662787)*x_ref[0]+(0.395096088107)*x_ref[1]+(-0.368904800584)*x_ref[2]
        ref[3,2,1,0]=(0.13107780566)*x_ref[0]+(-0.342904134623)*x_ref[1]+(0.00835087701185)*x_ref[2]
        ref[3,2,1,1]=(-0.500694550647)*x_ref[0]+(1.38322390003)*x_ref[1]+(0.0576453183412)*x_ref[2]
        ref[3,2,2,0]=(-1.32014489811)*x_ref[0]+(-0.444079317864)*x_ref[1]+(1.1135946018)*x_ref[2]
        ref[3,2,2,1]=(-0.19593105007)*x_ref[0]+(-0.00224390969377)*x_ref[1]+(-0.73869338942)*x_ref[2]
        ref[3,3,0,0]=(0.0151201622255)*x_ref[0]+(-1.10873788153)*x_ref[1]+(0.189116254832)*x_ref[2]
        ref[3,3,0,1]=(1.05996205566)*x_ref[0]+(0.398465639867)*x_ref[1]+(-0.319773104448)*x_ref[2]
        ref[3,3,1,0]=(0.687612827243)*x_ref[0]+(1.10970823257)*x_ref[1]+(1.02543288975)*x_ref[2]
        ref[3,3,1,1]=(1.27258509981)*x_ref[0]+(-1.7060826391)*x_ref[1]+(-0.84948812227)*x_ref[2]
        ref[3,3,2,0]=(1.2343036886)*x_ref[0]+(0.521090253023)*x_ref[1]+(-1.13377412954)*x_ref[2]
        ref[3,3,2,1]=(-0.65568738385)*x_ref[0]+(1.00354617712)*x_ref[1]+(0.836719015812)*x_ref[2]
        ref[3,4,0,0]=(-0.545324078594)*x_ref[0]+(0.452747782208)*x_ref[1]+(-1.59718954237)*x_ref[2]
        ref[3,4,0,1]=(1.71651153317)*x_ref[0]+(-0.161984538787)*x_ref[1]+(-0.377244827904)*x_ref[2]
        ref[3,4,1,0]=(-0.284033324558)*x_ref[0]+(-0.405941433062)*x_ref[1]+(0.824573264594)*x_ref[2]
        ref[3,4,1,1]=(-0.112542702803)*x_ref[0]+(-0.470814805786)*x_ref[1]+(-0.827427436576)*x_ref[2]
        ref[3,4,2,0]=(-0.763655132981)*x_ref[0]+(-0.442818775864)*x_ref[1]+(0.809814510041)*x_ref[2]
        ref[3,4,2,1]=(-0.33794306298)*x_ref[0]+(1.91169591332)*x_ref[1]+(-0.487268088196)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.898871557679)*x[0]+(0.0218492481542)*x[1]
        ref=(0.898871557679)*x_ref[0]+(0.0218492481542)*x_ref[1]
      else:
        arg=(-0.490730396919)*x[0]+(0.840683566031)*x[1]+(-0.265457839434)*x[2]
        ref=(-0.490730396919)*x_ref[0]+(0.840683566031)*x_ref[1]+(-0.265457839434)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        arg[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
        ref[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        ref[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
      else:
        arg[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        arg[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
        ref[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        ref[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.609515733352)*x[0]+(0.477437125804)*x[1]
        arg[0,1]=(0.225027969847)*x[0]+(-1.2178256661)*x[1]
        arg[0,2]=(-0.414641287889)*x[0]+(-0.536814389632)*x[1]
        arg[0,3]=(0.993370642676)*x[0]+(0.430000528217)*x[1]
        arg[0,4]=(0.577804610335)*x[0]+(0.0950391061127)*x[1]
        arg[1,0]=(0.437140855856)*x[0]+(0.764946114688)*x[1]
        arg[1,1]=(-0.491195658809)*x[0]+(-1.04729131261)*x[1]
        arg[1,2]=(-0.179287249145)*x[0]+(-1.84699873991)*x[1]
        arg[1,3]=(0.650408136022)*x[0]+(0.437145116534)*x[1]
        arg[1,4]=(0.43853490169)*x[0]+(-0.633716126064)*x[1]
        arg[2,0]=(0.6275418675)*x[0]+(0.812839621482)*x[1]
        arg[2,1]=(0.151298084104)*x[0]+(0.191677853013)*x[1]
        arg[2,2]=(1.35800571802)*x[0]+(1.08489593249)*x[1]
        arg[2,3]=(0.343386100925)*x[0]+(-0.108323271881)*x[1]
        arg[2,4]=(1.5968773444)*x[0]+(0.421328566279)*x[1]
        arg[3,0]=(1.0607367804)*x[0]+(-0.347352081886)*x[1]
        arg[3,1]=(-0.307183028406)*x[0]+(0.171667489459)*x[1]
        arg[3,2]=(0.255609148103)*x[0]+(0.21702578756)*x[1]
        arg[3,3]=(0.204487068999)*x[0]+(0.380168463726)*x[1]
        arg[3,4]=(-0.727714186561)*x[0]+(-0.129616906837)*x[1]
        ref[0,0]=(-0.609515733352)*x_ref[0]+(0.477437125804)*x_ref[1]
        ref[0,1]=(0.225027969847)*x_ref[0]+(-1.2178256661)*x_ref[1]
        ref[0,2]=(-0.414641287889)*x_ref[0]+(-0.536814389632)*x_ref[1]
        ref[0,3]=(0.993370642676)*x_ref[0]+(0.430000528217)*x_ref[1]
        ref[0,4]=(0.577804610335)*x_ref[0]+(0.0950391061127)*x_ref[1]
        ref[1,0]=(0.437140855856)*x_ref[0]+(0.764946114688)*x_ref[1]
        ref[1,1]=(-0.491195658809)*x_ref[0]+(-1.04729131261)*x_ref[1]
        ref[1,2]=(-0.179287249145)*x_ref[0]+(-1.84699873991)*x_ref[1]
        ref[1,3]=(0.650408136022)*x_ref[0]+(0.437145116534)*x_ref[1]
        ref[1,4]=(0.43853490169)*x_ref[0]+(-0.633716126064)*x_ref[1]
        ref[2,0]=(0.6275418675)*x_ref[0]+(0.812839621482)*x_ref[1]
        ref[2,1]=(0.151298084104)*x_ref[0]+(0.191677853013)*x_ref[1]
        ref[2,2]=(1.35800571802)*x_ref[0]+(1.08489593249)*x_ref[1]
        ref[2,3]=(0.343386100925)*x_ref[0]+(-0.108323271881)*x_ref[1]
        ref[2,4]=(1.5968773444)*x_ref[0]+(0.421328566279)*x_ref[1]
        ref[3,0]=(1.0607367804)*x_ref[0]+(-0.347352081886)*x_ref[1]
        ref[3,1]=(-0.307183028406)*x_ref[0]+(0.171667489459)*x_ref[1]
        ref[3,2]=(0.255609148103)*x_ref[0]+(0.21702578756)*x_ref[1]
        ref[3,3]=(0.204487068999)*x_ref[0]+(0.380168463726)*x_ref[1]
        ref[3,4]=(-0.727714186561)*x_ref[0]+(-0.129616906837)*x_ref[1]
      else:
        arg[0,0]=(-0.282058347281)*x[0]+(-0.804878634484)*x[1]+(1.59321960831)*x[2]
        arg[0,1]=(0.046711660417)*x[0]+(-0.353452776565)*x[1]+(-1.22459734449)*x[2]
        arg[0,2]=(0.846231855307)*x[0]+(1.08618070898)*x[1]+(-1.50365393652)*x[2]
        arg[0,3]=(-0.358546890162)*x[0]+(1.53476318049)*x[1]+(-1.19382694871)*x[2]
        arg[0,4]=(-1.1943310015)*x[0]+(0.260842673421)*x[1]+(-0.337737972605)*x[2]
        arg[1,0]=(1.87166717118)*x[0]+(0.853452519127)*x[1]+(-0.59081660847)*x[2]
        arg[1,1]=(0.23310002871)*x[0]+(0.306981228888)*x[1]+(0.36040454683)*x[2]
        arg[1,2]=(-1.619288679)*x[0]+(-0.459254234211)*x[1]+(0.722479242965)*x[2]
        arg[1,3]=(-1.07724802769)*x[0]+(0.836465436044)*x[1]+(0.255446930493)*x[2]
        arg[1,4]=(1.07857092921)*x[0]+(1.25642859092)*x[1]+(-0.922799422574)*x[2]
        arg[2,0]=(-0.832179820568)*x[0]+(-0.511243247139)*x[1]+(0.589457094147)*x[2]
        arg[2,1]=(-0.22616483359)*x[0]+(1.75842052561)*x[1]+(-1.19438769374)*x[2]
        arg[2,2]=(-0.690678634371)*x[0]+(0.963335056891)*x[1]+(0.53653322138)*x[2]
        arg[2,3]=(-0.906140718249)*x[0]+(0.766237211441)*x[1]+(1.19445635284)*x[2]
        arg[2,4]=(-0.16966820901)*x[0]+(0.185706811112)*x[1]+(0.975546783315)*x[2]
        arg[3,0]=(1.85053644213)*x[0]+(1.000328518)*x[1]+(0.332466610087)*x[2]
        arg[3,1]=(-0.199393988242)*x[0]+(0.637601295557)*x[1]+(1.29246614993)*x[2]
        arg[3,2]=(-0.275167528438)*x[0]+(1.00557873837)*x[1]+(1.17143679274)*x[2]
        arg[3,3]=(-0.306797319235)*x[0]+(0.984827835253)*x[1]+(0.0324710900497)*x[2]
        arg[3,4]=(-1.07994829511)*x[0]+(0.341016609147)*x[1]+(0.115829923585)*x[2]
        ref[0,0]=(-0.282058347281)*x_ref[0]+(-0.804878634484)*x_ref[1]+(1.59321960831)*x_ref[2]
        ref[0,1]=(0.046711660417)*x_ref[0]+(-0.353452776565)*x_ref[1]+(-1.22459734449)*x_ref[2]
        ref[0,2]=(0.846231855307)*x_ref[0]+(1.08618070898)*x_ref[1]+(-1.50365393652)*x_ref[2]
        ref[0,3]=(-0.358546890162)*x_ref[0]+(1.53476318049)*x_ref[1]+(-1.19382694871)*x_ref[2]
        ref[0,4]=(-1.1943310015)*x_ref[0]+(0.260842673421)*x_ref[1]+(-0.337737972605)*x_ref[2]
        ref[1,0]=(1.87166717118)*x_ref[0]+(0.853452519127)*x_ref[1]+(-0.59081660847)*x_ref[2]
        ref[1,1]=(0.23310002871)*x_ref[0]+(0.306981228888)*x_ref[1]+(0.36040454683)*x_ref[2]
        ref[1,2]=(-1.619288679)*x_ref[0]+(-0.459254234211)*x_ref[1]+(0.722479242965)*x_ref[2]
        ref[1,3]=(-1.07724802769)*x_ref[0]+(0.836465436044)*x_ref[1]+(0.255446930493)*x_ref[2]
        ref[1,4]=(1.07857092921)*x_ref[0]+(1.25642859092)*x_ref[1]+(-0.922799422574)*x_ref[2]
        ref[2,0]=(-0.832179820568)*x_ref[0]+(-0.511243247139)*x_ref[1]+(0.589457094147)*x_ref[2]
        ref[2,1]=(-0.22616483359)*x_ref[0]+(1.75842052561)*x_ref[1]+(-1.19438769374)*x_ref[2]
        ref[2,2]=(-0.690678634371)*x_ref[0]+(0.963335056891)*x_ref[1]+(0.53653322138)*x_ref[2]
        ref[2,3]=(-0.906140718249)*x_ref[0]+(0.766237211441)*x_ref[1]+(1.19445635284)*x_ref[2]
        ref[2,4]=(-0.16966820901)*x_ref[0]+(0.185706811112)*x_ref[1]+(0.975546783315)*x_ref[2]
        ref[3,0]=(1.85053644213)*x_ref[0]+(1.000328518)*x_ref[1]+(0.332466610087)*x_ref[2]
        ref[3,1]=(-0.199393988242)*x_ref[0]+(0.637601295557)*x_ref[1]+(1.29246614993)*x_ref[2]
        ref[3,2]=(-0.275167528438)*x_ref[0]+(1.00557873837)*x_ref[1]+(1.17143679274)*x_ref[2]
        ref[3,3]=(-0.306797319235)*x_ref[0]+(0.984827835253)*x_ref[1]+(0.0324710900497)*x_ref[2]
        ref[3,4]=(-1.07994829511)*x_ref[0]+(0.341016609147)*x_ref[1]+(0.115829923585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.618861298758)*x[0]+(-0.764616734859)*x[1]
        arg[0,0,1]=(-0.220720229024)*x[0]+(-0.670415412641)*x[1]
        arg[0,1,0]=(0.847082172076)*x[0]+(0.104515154796)*x[1]
        arg[0,1,1]=(-1.39487869037)*x[0]+(1.37639032205)*x[1]
        arg[1,0,0]=(-0.12944065767)*x[0]+(-0.91214412252)*x[1]
        arg[1,0,1]=(0.0421119211697)*x[0]+(1.19918915102)*x[1]
        arg[1,1,0]=(0.657099038454)*x[0]+(0.102886690913)*x[1]
        arg[1,1,1]=(0.114311284584)*x[0]+(-0.827908524893)*x[1]
        arg[2,0,0]=(-0.905281334495)*x[0]+(1.02228309099)*x[1]
        arg[2,0,1]=(-0.560619228879)*x[0]+(-0.924304525905)*x[1]
        arg[2,1,0]=(0.082868967617)*x[0]+(-0.129063346527)*x[1]
        arg[2,1,1]=(0.386687881433)*x[0]+(-0.874995344656)*x[1]
        arg[3,0,0]=(0.460976605295)*x[0]+(0.145631537258)*x[1]
        arg[3,0,1]=(0.439977672901)*x[0]+(0.714321229059)*x[1]
        arg[3,1,0]=(-0.0459363656635)*x[0]+(-0.0754715887416)*x[1]
        arg[3,1,1]=(-0.789910257931)*x[0]+(-0.806078036909)*x[1]
        arg[4,0,0]=(-1.09951071165)*x[0]+(-1.33461983123)*x[1]
        arg[4,0,1]=(0.857184302111)*x[0]+(0.823036666525)*x[1]
        arg[4,1,0]=(-1.0682486872)*x[0]+(-0.149861981932)*x[1]
        arg[4,1,1]=(-0.0758367136425)*x[0]+(0.0460556685916)*x[1]
        arg[5,0,0]=(0.856207295965)*x[0]+(-0.885450852663)*x[1]
        arg[5,0,1]=(-0.113939086685)*x[0]+(0.742808810968)*x[1]
        arg[5,1,0]=(-0.174573302919)*x[0]+(0.486803189896)*x[1]
        arg[5,1,1]=(1.1726771004)*x[0]+(-0.329765988186)*x[1]
        ref[0,0,0]=(0.618861298758)*x_ref[0]+(-0.764616734859)*x_ref[1]
        ref[0,0,1]=(-0.220720229024)*x_ref[0]+(-0.670415412641)*x_ref[1]
        ref[0,1,0]=(0.847082172076)*x_ref[0]+(0.104515154796)*x_ref[1]
        ref[0,1,1]=(-1.39487869037)*x_ref[0]+(1.37639032205)*x_ref[1]
        ref[1,0,0]=(-0.12944065767)*x_ref[0]+(-0.91214412252)*x_ref[1]
        ref[1,0,1]=(0.0421119211697)*x_ref[0]+(1.19918915102)*x_ref[1]
        ref[1,1,0]=(0.657099038454)*x_ref[0]+(0.102886690913)*x_ref[1]
        ref[1,1,1]=(0.114311284584)*x_ref[0]+(-0.827908524893)*x_ref[1]
        ref[2,0,0]=(-0.905281334495)*x_ref[0]+(1.02228309099)*x_ref[1]
        ref[2,0,1]=(-0.560619228879)*x_ref[0]+(-0.924304525905)*x_ref[1]
        ref[2,1,0]=(0.082868967617)*x_ref[0]+(-0.129063346527)*x_ref[1]
        ref[2,1,1]=(0.386687881433)*x_ref[0]+(-0.874995344656)*x_ref[1]
        ref[3,0,0]=(0.460976605295)*x_ref[0]+(0.145631537258)*x_ref[1]
        ref[3,0,1]=(0.439977672901)*x_ref[0]+(0.714321229059)*x_ref[1]
        ref[3,1,0]=(-0.0459363656635)*x_ref[0]+(-0.0754715887416)*x_ref[1]
        ref[3,1,1]=(-0.789910257931)*x_ref[0]+(-0.806078036909)*x_ref[1]
        ref[4,0,0]=(-1.09951071165)*x_ref[0]+(-1.33461983123)*x_ref[1]
        ref[4,0,1]=(0.857184302111)*x_ref[0]+(0.823036666525)*x_ref[1]
        ref[4,1,0]=(-1.0682486872)*x_ref[0]+(-0.149861981932)*x_ref[1]
        ref[4,1,1]=(-0.0758367136425)*x_ref[0]+(0.0460556685916)*x_ref[1]
        ref[5,0,0]=(0.856207295965)*x_ref[0]+(-0.885450852663)*x_ref[1]
        ref[5,0,1]=(-0.113939086685)*x_ref[0]+(0.742808810968)*x_ref[1]
        ref[5,1,0]=(-0.174573302919)*x_ref[0]+(0.486803189896)*x_ref[1]
        ref[5,1,1]=(1.1726771004)*x_ref[0]+(-0.329765988186)*x_ref[1]
      else:
        arg[0,0,0]=(-0.640205798541)*x[0]+(0.271670719935)*x[1]+(-0.163000598261)*x[2]
        arg[0,0,1]=(0.735605297981)*x[0]+(-1.05892412676)*x[1]+(-1.1083007441)*x[2]
        arg[0,1,0]=(0.0349869408333)*x[0]+(0.699147831788)*x[1]+(-0.454390826725)*x[2]
        arg[0,1,1]=(-0.703957679243)*x[0]+(-1.63654420176)*x[1]+(-1.13045075795)*x[2]
        arg[1,0,0]=(1.01167296102)*x[0]+(-0.362443738659)*x[1]+(-0.0481138040854)*x[2]
        arg[1,0,1]=(0.110033706339)*x[0]+(1.47809236549)*x[1]+(0.240737599484)*x[2]
        arg[1,1,0]=(-0.0429437939148)*x[0]+(-1.16945967196)*x[1]+(0.924577053186)*x[2]
        arg[1,1,1]=(-1.57856509244)*x[0]+(-0.460441809594)*x[1]+(-0.792590928973)*x[2]
        arg[2,0,0]=(-1.93020920065)*x[0]+(0.391612451745)*x[1]+(-0.0130238634938)*x[2]
        arg[2,0,1]=(0.0678908982414)*x[0]+(-0.929813889361)*x[1]+(0.727462587818)*x[2]
        arg[2,1,0]=(-0.297494494419)*x[0]+(0.167234033247)*x[1]+(-1.01372084798)*x[2]
        arg[2,1,1]=(-0.608810623174)*x[0]+(1.04629900384)*x[1]+(1.09464953456)*x[2]
        arg[3,0,0]=(1.2996449893)*x[0]+(-0.458197425323)*x[1]+(1.31523663461)*x[2]
        arg[3,0,1]=(-0.848124038387)*x[0]+(0.0136487458675)*x[1]+(1.13573336016)*x[2]
        arg[3,1,0]=(0.377245000633)*x[0]+(-0.766336675857)*x[1]+(0.380031677083)*x[2]
        arg[3,1,1]=(0.158289533163)*x[0]+(-0.149454925108)*x[1]+(-0.59348598427)*x[2]
        arg[4,0,0]=(-1.13101661269)*x[0]+(-0.694062133171)*x[1]+(0.00925221725394)*x[2]
        arg[4,0,1]=(-0.977317473287)*x[0]+(-0.121712752812)*x[1]+(-0.480470720169)*x[2]
        arg[4,1,0]=(1.14884269473)*x[0]+(-0.204958154665)*x[1]+(1.24328207187)*x[2]
        arg[4,1,1]=(0.0791969146134)*x[0]+(0.589873409499)*x[1]+(1.90521993698)*x[2]
        arg[5,0,0]=(1.75625788908)*x[0]+(0.0134922175773)*x[1]+(-0.699884009939)*x[2]
        arg[5,0,1]=(-0.73456900252)*x[0]+(0.923248295457)*x[1]+(1.05161845199)*x[2]
        arg[5,1,0]=(-0.0335528775888)*x[0]+(0.728742831853)*x[1]+(-0.5212350627)*x[2]
        arg[5,1,1]=(0.556540394981)*x[0]+(-0.128428520433)*x[1]+(1.06358498081)*x[2]
        ref[0,0,0]=(-0.640205798541)*x_ref[0]+(0.271670719935)*x_ref[1]+(-0.163000598261)*x_ref[2]
        ref[0,0,1]=(0.735605297981)*x_ref[0]+(-1.05892412676)*x_ref[1]+(-1.1083007441)*x_ref[2]
        ref[0,1,0]=(0.0349869408333)*x_ref[0]+(0.699147831788)*x_ref[1]+(-0.454390826725)*x_ref[2]
        ref[0,1,1]=(-0.703957679243)*x_ref[0]+(-1.63654420176)*x_ref[1]+(-1.13045075795)*x_ref[2]
        ref[1,0,0]=(1.01167296102)*x_ref[0]+(-0.362443738659)*x_ref[1]+(-0.0481138040854)*x_ref[2]
        ref[1,0,1]=(0.110033706339)*x_ref[0]+(1.47809236549)*x_ref[1]+(0.240737599484)*x_ref[2]
        ref[1,1,0]=(-0.0429437939148)*x_ref[0]+(-1.16945967196)*x_ref[1]+(0.924577053186)*x_ref[2]
        ref[1,1,1]=(-1.57856509244)*x_ref[0]+(-0.460441809594)*x_ref[1]+(-0.792590928973)*x_ref[2]
        ref[2,0,0]=(-1.93020920065)*x_ref[0]+(0.391612451745)*x_ref[1]+(-0.0130238634938)*x_ref[2]
        ref[2,0,1]=(0.0678908982414)*x_ref[0]+(-0.929813889361)*x_ref[1]+(0.727462587818)*x_ref[2]
        ref[2,1,0]=(-0.297494494419)*x_ref[0]+(0.167234033247)*x_ref[1]+(-1.01372084798)*x_ref[2]
        ref[2,1,1]=(-0.608810623174)*x_ref[0]+(1.04629900384)*x_ref[1]+(1.09464953456)*x_ref[2]
        ref[3,0,0]=(1.2996449893)*x_ref[0]+(-0.458197425323)*x_ref[1]+(1.31523663461)*x_ref[2]
        ref[3,0,1]=(-0.848124038387)*x_ref[0]+(0.0136487458675)*x_ref[1]+(1.13573336016)*x_ref[2]
        ref[3,1,0]=(0.377245000633)*x_ref[0]+(-0.766336675857)*x_ref[1]+(0.380031677083)*x_ref[2]
        ref[3,1,1]=(0.158289533163)*x_ref[0]+(-0.149454925108)*x_ref[1]+(-0.59348598427)*x_ref[2]
        ref[4,0,0]=(-1.13101661269)*x_ref[0]+(-0.694062133171)*x_ref[1]+(0.00925221725394)*x_ref[2]
        ref[4,0,1]=(-0.977317473287)*x_ref[0]+(-0.121712752812)*x_ref[1]+(-0.480470720169)*x_ref[2]
        ref[4,1,0]=(1.14884269473)*x_ref[0]+(-0.204958154665)*x_ref[1]+(1.24328207187)*x_ref[2]
        ref[4,1,1]=(0.0791969146134)*x_ref[0]+(0.589873409499)*x_ref[1]+(1.90521993698)*x_ref[2]
        ref[5,0,0]=(1.75625788908)*x_ref[0]+(0.0134922175773)*x_ref[1]+(-0.699884009939)*x_ref[2]
        ref[5,0,1]=(-0.73456900252)*x_ref[0]+(0.923248295457)*x_ref[1]+(1.05161845199)*x_ref[2]
        ref[5,1,0]=(-0.0335528775888)*x_ref[0]+(0.728742831853)*x_ref[1]+(-0.5212350627)*x_ref[2]
        ref[5,1,1]=(0.556540394981)*x_ref[0]+(-0.128428520433)*x_ref[1]+(1.06358498081)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.507324804235)*x[0]+(0.725831700089)*x[1]
        arg[0,0,0,1]=(0.847901454936)*x[0]+(0.84787024099)*x[1]
        arg[0,0,1,0]=(0.221488465784)*x[0]+(0.0449286988708)*x[1]
        arg[0,0,1,1]=(0.00897176578501)*x[0]+(-0.799118487206)*x[1]
        arg[0,0,2,0]=(0.171491682599)*x[0]+(1.0259065872)*x[1]
        arg[0,0,2,1]=(-0.213244683798)*x[0]+(-0.706816385516)*x[1]
        arg[0,1,0,0]=(1.54199875186)*x[0]+(-1.09479402082)*x[1]
        arg[0,1,0,1]=(-0.755278051349)*x[0]+(-0.371785089226)*x[1]
        arg[0,1,1,0]=(-0.798931370137)*x[0]+(-0.716047155117)*x[1]
        arg[0,1,1,1]=(-0.79104072521)*x[0]+(-1.45842710768)*x[1]
        arg[0,1,2,0]=(-0.489161301965)*x[0]+(0.115368605994)*x[1]
        arg[0,1,2,1]=(0.00167228618616)*x[0]+(-0.636896504873)*x[1]
        arg[0,2,0,0]=(-0.239182729045)*x[0]+(0.470786327067)*x[1]
        arg[0,2,0,1]=(0.182105943188)*x[0]+(0.328443662342)*x[1]
        arg[0,2,1,0]=(0.000476540430199)*x[0]+(-0.498066512126)*x[1]
        arg[0,2,1,1]=(-0.822985659305)*x[0]+(-0.118645880185)*x[1]
        arg[0,2,2,0]=(1.18375382554)*x[0]+(1.51417307474)*x[1]
        arg[0,2,2,1]=(-0.99755868797)*x[0]+(-1.4187068658)*x[1]
        arg[0,3,0,0]=(-0.0908782314954)*x[0]+(-0.25901340471)*x[1]
        arg[0,3,0,1]=(-1.77626335863)*x[0]+(-0.862132832663)*x[1]
        arg[0,3,1,0]=(0.532147327491)*x[0]+(0.454949026506)*x[1]
        arg[0,3,1,1]=(-0.882307958628)*x[0]+(0.00780544746858)*x[1]
        arg[0,3,2,0]=(0.700104420574)*x[0]+(0.720508036709)*x[1]
        arg[0,3,2,1]=(-1.27167118811)*x[0]+(-0.372999410908)*x[1]
        arg[0,4,0,0]=(0.396222903345)*x[0]+(1.70664799562)*x[1]
        arg[0,4,0,1]=(0.229159055372)*x[0]+(0.0162677758439)*x[1]
        arg[0,4,1,0]=(1.69470499818)*x[0]+(0.0971804891468)*x[1]
        arg[0,4,1,1]=(0.180710573838)*x[0]+(1.05311557127)*x[1]
        arg[0,4,2,0]=(-0.072661496917)*x[0]+(0.466256643644)*x[1]
        arg[0,4,2,1]=(0.25755631084)*x[0]+(0.306990925808)*x[1]
        arg[1,0,0,0]=(0.835353191317)*x[0]+(-0.619237060145)*x[1]
        arg[1,0,0,1]=(1.02966866371)*x[0]+(0.327724003031)*x[1]
        arg[1,0,1,0]=(-0.59214788986)*x[0]+(-0.927738485428)*x[1]
        arg[1,0,1,1]=(-0.805929278446)*x[0]+(-0.366113927406)*x[1]
        arg[1,0,2,0]=(0.952876155721)*x[0]+(0.490531201467)*x[1]
        arg[1,0,2,1]=(1.44341651657)*x[0]+(-0.269899260772)*x[1]
        arg[1,1,0,0]=(0.302587966785)*x[0]+(-0.0373883991695)*x[1]
        arg[1,1,0,1]=(0.77498995648)*x[0]+(1.81052574555)*x[1]
        arg[1,1,1,0]=(1.47950560134)*x[0]+(-0.714878219566)*x[1]
        arg[1,1,1,1]=(0.0713827629606)*x[0]+(-1.22018291576)*x[1]
        arg[1,1,2,0]=(1.30715631604)*x[0]+(-0.191666253513)*x[1]
        arg[1,1,2,1]=(0.330921962628)*x[0]+(-1.57160261912)*x[1]
        arg[1,2,0,0]=(0.593477126597)*x[0]+(0.728196634765)*x[1]
        arg[1,2,0,1]=(0.853860342143)*x[0]+(0.142625101433)*x[1]
        arg[1,2,1,0]=(-0.743590115343)*x[0]+(0.108893736366)*x[1]
        arg[1,2,1,1]=(-0.651514863459)*x[0]+(0.365517180762)*x[1]
        arg[1,2,2,0]=(1.491516719)*x[0]+(0.28900806658)*x[1]
        arg[1,2,2,1]=(1.23733736339)*x[0]+(-0.58438294882)*x[1]
        arg[1,3,0,0]=(0.185696018505)*x[0]+(-1.29119084907)*x[1]
        arg[1,3,0,1]=(1.73584656085)*x[0]+(0.720989518003)*x[1]
        arg[1,3,1,0]=(-1.16044192371)*x[0]+(-0.131128994204)*x[1]
        arg[1,3,1,1]=(-0.00913995119508)*x[0]+(0.506516682557)*x[1]
        arg[1,3,2,0]=(-1.4970171408)*x[0]+(0.692262679219)*x[1]
        arg[1,3,2,1]=(0.433442352395)*x[0]+(-0.118782513258)*x[1]
        arg[1,4,0,0]=(0.186272280804)*x[0]+(-0.136627137097)*x[1]
        arg[1,4,0,1]=(1.25477168515)*x[0]+(-0.0871831369474)*x[1]
        arg[1,4,1,0]=(1.22746022677)*x[0]+(0.799329883559)*x[1]
        arg[1,4,1,1]=(-0.756298280123)*x[0]+(0.317500967634)*x[1]
        arg[1,4,2,0]=(0.201184703435)*x[0]+(-0.116537935236)*x[1]
        arg[1,4,2,1]=(0.851291897607)*x[0]+(0.404063606516)*x[1]
        arg[2,0,0,0]=(-0.257605370026)*x[0]+(-0.202424701671)*x[1]
        arg[2,0,0,1]=(0.0219332720675)*x[0]+(0.518222103227)*x[1]
        arg[2,0,1,0]=(1.03726036144)*x[0]+(-1.01482867596)*x[1]
        arg[2,0,1,1]=(0.00677025785109)*x[0]+(0.607932484906)*x[1]
        arg[2,0,2,0]=(-0.753000974637)*x[0]+(0.112884693676)*x[1]
        arg[2,0,2,1]=(-0.585627190582)*x[0]+(0.548136242722)*x[1]
        arg[2,1,0,0]=(0.382466685883)*x[0]+(-1.15520454298)*x[1]
        arg[2,1,0,1]=(-0.722705123323)*x[0]+(-0.599069291154)*x[1]
        arg[2,1,1,0]=(-0.845885208256)*x[0]+(-1.56066350212)*x[1]
        arg[2,1,1,1]=(-1.31472469553)*x[0]+(0.18973548452)*x[1]
        arg[2,1,2,0]=(-0.419142599733)*x[0]+(1.43873372527)*x[1]
        arg[2,1,2,1]=(0.22685530286)*x[0]+(-0.53636714393)*x[1]
        arg[2,2,0,0]=(1.32884214935)*x[0]+(-0.377575726395)*x[1]
        arg[2,2,0,1]=(0.0845542736847)*x[0]+(0.0888339000429)*x[1]
        arg[2,2,1,0]=(-1.05813285521)*x[0]+(-0.980382395463)*x[1]
        arg[2,2,1,1]=(0.306785463268)*x[0]+(0.622558337967)*x[1]
        arg[2,2,2,0]=(-0.765475308614)*x[0]+(1.40320055807)*x[1]
        arg[2,2,2,1]=(-1.23199081441)*x[0]+(0.245517841221)*x[1]
        arg[2,3,0,0]=(0.655775411743)*x[0]+(-1.09496249546)*x[1]
        arg[2,3,0,1]=(-1.57906394006)*x[0]+(-0.223321547826)*x[1]
        arg[2,3,1,0]=(-0.329934697034)*x[0]+(-1.36492062074)*x[1]
        arg[2,3,1,1]=(1.2584701256)*x[0]+(0.172958559745)*x[1]
        arg[2,3,2,0]=(0.265769150885)*x[0]+(-0.273148332255)*x[1]
        arg[2,3,2,1]=(-1.18294356647)*x[0]+(-0.269813147795)*x[1]
        arg[2,4,0,0]=(-1.80352122024)*x[0]+(-0.417693287507)*x[1]
        arg[2,4,0,1]=(0.829979623343)*x[0]+(0.0526239642579)*x[1]
        arg[2,4,1,0]=(0.967999103063)*x[0]+(0.0184170526665)*x[1]
        arg[2,4,1,1]=(-0.326938715134)*x[0]+(-0.575040647639)*x[1]
        arg[2,4,2,0]=(1.44727400099)*x[0]+(0.710269500685)*x[1]
        arg[2,4,2,1]=(-0.248166665809)*x[0]+(0.247820501661)*x[1]
        arg[3,0,0,0]=(0.0463415093786)*x[0]+(0.650085996364)*x[1]
        arg[3,0,0,1]=(-0.440939509697)*x[0]+(-0.323307669473)*x[1]
        arg[3,0,1,0]=(0.273356160699)*x[0]+(-0.66721600197)*x[1]
        arg[3,0,1,1]=(1.29465477714)*x[0]+(-0.37486215111)*x[1]
        arg[3,0,2,0]=(-0.809896216106)*x[0]+(-1.51076103892)*x[1]
        arg[3,0,2,1]=(0.686428827303)*x[0]+(-0.538164742287)*x[1]
        arg[3,1,0,0]=(-0.863651610678)*x[0]+(0.673803453645)*x[1]
        arg[3,1,0,1]=(0.526827519816)*x[0]+(-0.0379595314201)*x[1]
        arg[3,1,1,0]=(-1.04759557033)*x[0]+(0.114198098457)*x[1]
        arg[3,1,1,1]=(0.552205171704)*x[0]+(0.116364312378)*x[1]
        arg[3,1,2,0]=(-0.926549911354)*x[0]+(0.296927704155)*x[1]
        arg[3,1,2,1]=(0.46702104365)*x[0]+(0.355015968814)*x[1]
        arg[3,2,0,0]=(0.603177046544)*x[0]+(-0.600600764625)*x[1]
        arg[3,2,0,1]=(0.656827398088)*x[0]+(1.19922812106)*x[1]
        arg[3,2,1,0]=(-0.0185352951675)*x[0]+(-0.575006610999)*x[1]
        arg[3,2,1,1]=(0.954361462974)*x[0]+(-0.239751892125)*x[1]
        arg[3,2,2,0]=(0.880568376655)*x[0]+(1.46879509876)*x[1]
        arg[3,2,2,1]=(-1.19051913627)*x[0]+(-0.0387119361514)*x[1]
        arg[3,3,0,0]=(1.19520855834)*x[0]+(1.3515728248)*x[1]
        arg[3,3,0,1]=(0.695487159903)*x[0]+(0.783460154378)*x[1]
        arg[3,3,1,0]=(0.482352542106)*x[0]+(-1.321707654)*x[1]
        arg[3,3,1,1]=(0.0637384915311)*x[0]+(-0.5903067787)*x[1]
        arg[3,3,2,0]=(0.571974795979)*x[0]+(1.33251727342)*x[1]
        arg[3,3,2,1]=(0.579723470137)*x[0]+(-0.260885275522)*x[1]
        arg[3,4,0,0]=(0.0717612359082)*x[0]+(-0.576982745638)*x[1]
        arg[3,4,0,1]=(-0.227660227523)*x[0]+(0.232974094548)*x[1]
        arg[3,4,1,0]=(0.767797269581)*x[0]+(0.98857043831)*x[1]
        arg[3,4,1,1]=(0.31200408415)*x[0]+(-0.8317658048)*x[1]
        arg[3,4,2,0]=(0.381688388928)*x[0]+(-0.335620162467)*x[1]
        arg[3,4,2,1]=(0.781228510591)*x[0]+(0.960966953643)*x[1]
        ref[0,0,0,0]=(0.507324804235)*x_ref[0]+(0.725831700089)*x_ref[1]
        ref[0,0,0,1]=(0.847901454936)*x_ref[0]+(0.84787024099)*x_ref[1]
        ref[0,0,1,0]=(0.221488465784)*x_ref[0]+(0.0449286988708)*x_ref[1]
        ref[0,0,1,1]=(0.00897176578501)*x_ref[0]+(-0.799118487206)*x_ref[1]
        ref[0,0,2,0]=(0.171491682599)*x_ref[0]+(1.0259065872)*x_ref[1]
        ref[0,0,2,1]=(-0.213244683798)*x_ref[0]+(-0.706816385516)*x_ref[1]
        ref[0,1,0,0]=(1.54199875186)*x_ref[0]+(-1.09479402082)*x_ref[1]
        ref[0,1,0,1]=(-0.755278051349)*x_ref[0]+(-0.371785089226)*x_ref[1]
        ref[0,1,1,0]=(-0.798931370137)*x_ref[0]+(-0.716047155117)*x_ref[1]
        ref[0,1,1,1]=(-0.79104072521)*x_ref[0]+(-1.45842710768)*x_ref[1]
        ref[0,1,2,0]=(-0.489161301965)*x_ref[0]+(0.115368605994)*x_ref[1]
        ref[0,1,2,1]=(0.00167228618616)*x_ref[0]+(-0.636896504873)*x_ref[1]
        ref[0,2,0,0]=(-0.239182729045)*x_ref[0]+(0.470786327067)*x_ref[1]
        ref[0,2,0,1]=(0.182105943188)*x_ref[0]+(0.328443662342)*x_ref[1]
        ref[0,2,1,0]=(0.000476540430199)*x_ref[0]+(-0.498066512126)*x_ref[1]
        ref[0,2,1,1]=(-0.822985659305)*x_ref[0]+(-0.118645880185)*x_ref[1]
        ref[0,2,2,0]=(1.18375382554)*x_ref[0]+(1.51417307474)*x_ref[1]
        ref[0,2,2,1]=(-0.99755868797)*x_ref[0]+(-1.4187068658)*x_ref[1]
        ref[0,3,0,0]=(-0.0908782314954)*x_ref[0]+(-0.25901340471)*x_ref[1]
        ref[0,3,0,1]=(-1.77626335863)*x_ref[0]+(-0.862132832663)*x_ref[1]
        ref[0,3,1,0]=(0.532147327491)*x_ref[0]+(0.454949026506)*x_ref[1]
        ref[0,3,1,1]=(-0.882307958628)*x_ref[0]+(0.00780544746858)*x_ref[1]
        ref[0,3,2,0]=(0.700104420574)*x_ref[0]+(0.720508036709)*x_ref[1]
        ref[0,3,2,1]=(-1.27167118811)*x_ref[0]+(-0.372999410908)*x_ref[1]
        ref[0,4,0,0]=(0.396222903345)*x_ref[0]+(1.70664799562)*x_ref[1]
        ref[0,4,0,1]=(0.229159055372)*x_ref[0]+(0.0162677758439)*x_ref[1]
        ref[0,4,1,0]=(1.69470499818)*x_ref[0]+(0.0971804891468)*x_ref[1]
        ref[0,4,1,1]=(0.180710573838)*x_ref[0]+(1.05311557127)*x_ref[1]
        ref[0,4,2,0]=(-0.072661496917)*x_ref[0]+(0.466256643644)*x_ref[1]
        ref[0,4,2,1]=(0.25755631084)*x_ref[0]+(0.306990925808)*x_ref[1]
        ref[1,0,0,0]=(0.835353191317)*x_ref[0]+(-0.619237060145)*x_ref[1]
        ref[1,0,0,1]=(1.02966866371)*x_ref[0]+(0.327724003031)*x_ref[1]
        ref[1,0,1,0]=(-0.59214788986)*x_ref[0]+(-0.927738485428)*x_ref[1]
        ref[1,0,1,1]=(-0.805929278446)*x_ref[0]+(-0.366113927406)*x_ref[1]
        ref[1,0,2,0]=(0.952876155721)*x_ref[0]+(0.490531201467)*x_ref[1]
        ref[1,0,2,1]=(1.44341651657)*x_ref[0]+(-0.269899260772)*x_ref[1]
        ref[1,1,0,0]=(0.302587966785)*x_ref[0]+(-0.0373883991695)*x_ref[1]
        ref[1,1,0,1]=(0.77498995648)*x_ref[0]+(1.81052574555)*x_ref[1]
        ref[1,1,1,0]=(1.47950560134)*x_ref[0]+(-0.714878219566)*x_ref[1]
        ref[1,1,1,1]=(0.0713827629606)*x_ref[0]+(-1.22018291576)*x_ref[1]
        ref[1,1,2,0]=(1.30715631604)*x_ref[0]+(-0.191666253513)*x_ref[1]
        ref[1,1,2,1]=(0.330921962628)*x_ref[0]+(-1.57160261912)*x_ref[1]
        ref[1,2,0,0]=(0.593477126597)*x_ref[0]+(0.728196634765)*x_ref[1]
        ref[1,2,0,1]=(0.853860342143)*x_ref[0]+(0.142625101433)*x_ref[1]
        ref[1,2,1,0]=(-0.743590115343)*x_ref[0]+(0.108893736366)*x_ref[1]
        ref[1,2,1,1]=(-0.651514863459)*x_ref[0]+(0.365517180762)*x_ref[1]
        ref[1,2,2,0]=(1.491516719)*x_ref[0]+(0.28900806658)*x_ref[1]
        ref[1,2,2,1]=(1.23733736339)*x_ref[0]+(-0.58438294882)*x_ref[1]
        ref[1,3,0,0]=(0.185696018505)*x_ref[0]+(-1.29119084907)*x_ref[1]
        ref[1,3,0,1]=(1.73584656085)*x_ref[0]+(0.720989518003)*x_ref[1]
        ref[1,3,1,0]=(-1.16044192371)*x_ref[0]+(-0.131128994204)*x_ref[1]
        ref[1,3,1,1]=(-0.00913995119508)*x_ref[0]+(0.506516682557)*x_ref[1]
        ref[1,3,2,0]=(-1.4970171408)*x_ref[0]+(0.692262679219)*x_ref[1]
        ref[1,3,2,1]=(0.433442352395)*x_ref[0]+(-0.118782513258)*x_ref[1]
        ref[1,4,0,0]=(0.186272280804)*x_ref[0]+(-0.136627137097)*x_ref[1]
        ref[1,4,0,1]=(1.25477168515)*x_ref[0]+(-0.0871831369474)*x_ref[1]
        ref[1,4,1,0]=(1.22746022677)*x_ref[0]+(0.799329883559)*x_ref[1]
        ref[1,4,1,1]=(-0.756298280123)*x_ref[0]+(0.317500967634)*x_ref[1]
        ref[1,4,2,0]=(0.201184703435)*x_ref[0]+(-0.116537935236)*x_ref[1]
        ref[1,4,2,1]=(0.851291897607)*x_ref[0]+(0.404063606516)*x_ref[1]
        ref[2,0,0,0]=(-0.257605370026)*x_ref[0]+(-0.202424701671)*x_ref[1]
        ref[2,0,0,1]=(0.0219332720675)*x_ref[0]+(0.518222103227)*x_ref[1]
        ref[2,0,1,0]=(1.03726036144)*x_ref[0]+(-1.01482867596)*x_ref[1]
        ref[2,0,1,1]=(0.00677025785109)*x_ref[0]+(0.607932484906)*x_ref[1]
        ref[2,0,2,0]=(-0.753000974637)*x_ref[0]+(0.112884693676)*x_ref[1]
        ref[2,0,2,1]=(-0.585627190582)*x_ref[0]+(0.548136242722)*x_ref[1]
        ref[2,1,0,0]=(0.382466685883)*x_ref[0]+(-1.15520454298)*x_ref[1]
        ref[2,1,0,1]=(-0.722705123323)*x_ref[0]+(-0.599069291154)*x_ref[1]
        ref[2,1,1,0]=(-0.845885208256)*x_ref[0]+(-1.56066350212)*x_ref[1]
        ref[2,1,1,1]=(-1.31472469553)*x_ref[0]+(0.18973548452)*x_ref[1]
        ref[2,1,2,0]=(-0.419142599733)*x_ref[0]+(1.43873372527)*x_ref[1]
        ref[2,1,2,1]=(0.22685530286)*x_ref[0]+(-0.53636714393)*x_ref[1]
        ref[2,2,0,0]=(1.32884214935)*x_ref[0]+(-0.377575726395)*x_ref[1]
        ref[2,2,0,1]=(0.0845542736847)*x_ref[0]+(0.0888339000429)*x_ref[1]
        ref[2,2,1,0]=(-1.05813285521)*x_ref[0]+(-0.980382395463)*x_ref[1]
        ref[2,2,1,1]=(0.306785463268)*x_ref[0]+(0.622558337967)*x_ref[1]
        ref[2,2,2,0]=(-0.765475308614)*x_ref[0]+(1.40320055807)*x_ref[1]
        ref[2,2,2,1]=(-1.23199081441)*x_ref[0]+(0.245517841221)*x_ref[1]
        ref[2,3,0,0]=(0.655775411743)*x_ref[0]+(-1.09496249546)*x_ref[1]
        ref[2,3,0,1]=(-1.57906394006)*x_ref[0]+(-0.223321547826)*x_ref[1]
        ref[2,3,1,0]=(-0.329934697034)*x_ref[0]+(-1.36492062074)*x_ref[1]
        ref[2,3,1,1]=(1.2584701256)*x_ref[0]+(0.172958559745)*x_ref[1]
        ref[2,3,2,0]=(0.265769150885)*x_ref[0]+(-0.273148332255)*x_ref[1]
        ref[2,3,2,1]=(-1.18294356647)*x_ref[0]+(-0.269813147795)*x_ref[1]
        ref[2,4,0,0]=(-1.80352122024)*x_ref[0]+(-0.417693287507)*x_ref[1]
        ref[2,4,0,1]=(0.829979623343)*x_ref[0]+(0.0526239642579)*x_ref[1]
        ref[2,4,1,0]=(0.967999103063)*x_ref[0]+(0.0184170526665)*x_ref[1]
        ref[2,4,1,1]=(-0.326938715134)*x_ref[0]+(-0.575040647639)*x_ref[1]
        ref[2,4,2,0]=(1.44727400099)*x_ref[0]+(0.710269500685)*x_ref[1]
        ref[2,4,2,1]=(-0.248166665809)*x_ref[0]+(0.247820501661)*x_ref[1]
        ref[3,0,0,0]=(0.0463415093786)*x_ref[0]+(0.650085996364)*x_ref[1]
        ref[3,0,0,1]=(-0.440939509697)*x_ref[0]+(-0.323307669473)*x_ref[1]
        ref[3,0,1,0]=(0.273356160699)*x_ref[0]+(-0.66721600197)*x_ref[1]
        ref[3,0,1,1]=(1.29465477714)*x_ref[0]+(-0.37486215111)*x_ref[1]
        ref[3,0,2,0]=(-0.809896216106)*x_ref[0]+(-1.51076103892)*x_ref[1]
        ref[3,0,2,1]=(0.686428827303)*x_ref[0]+(-0.538164742287)*x_ref[1]
        ref[3,1,0,0]=(-0.863651610678)*x_ref[0]+(0.673803453645)*x_ref[1]
        ref[3,1,0,1]=(0.526827519816)*x_ref[0]+(-0.0379595314201)*x_ref[1]
        ref[3,1,1,0]=(-1.04759557033)*x_ref[0]+(0.114198098457)*x_ref[1]
        ref[3,1,1,1]=(0.552205171704)*x_ref[0]+(0.116364312378)*x_ref[1]
        ref[3,1,2,0]=(-0.926549911354)*x_ref[0]+(0.296927704155)*x_ref[1]
        ref[3,1,2,1]=(0.46702104365)*x_ref[0]+(0.355015968814)*x_ref[1]
        ref[3,2,0,0]=(0.603177046544)*x_ref[0]+(-0.600600764625)*x_ref[1]
        ref[3,2,0,1]=(0.656827398088)*x_ref[0]+(1.19922812106)*x_ref[1]
        ref[3,2,1,0]=(-0.0185352951675)*x_ref[0]+(-0.575006610999)*x_ref[1]
        ref[3,2,1,1]=(0.954361462974)*x_ref[0]+(-0.239751892125)*x_ref[1]
        ref[3,2,2,0]=(0.880568376655)*x_ref[0]+(1.46879509876)*x_ref[1]
        ref[3,2,2,1]=(-1.19051913627)*x_ref[0]+(-0.0387119361514)*x_ref[1]
        ref[3,3,0,0]=(1.19520855834)*x_ref[0]+(1.3515728248)*x_ref[1]
        ref[3,3,0,1]=(0.695487159903)*x_ref[0]+(0.783460154378)*x_ref[1]
        ref[3,3,1,0]=(0.482352542106)*x_ref[0]+(-1.321707654)*x_ref[1]
        ref[3,3,1,1]=(0.0637384915311)*x_ref[0]+(-0.5903067787)*x_ref[1]
        ref[3,3,2,0]=(0.571974795979)*x_ref[0]+(1.33251727342)*x_ref[1]
        ref[3,3,2,1]=(0.579723470137)*x_ref[0]+(-0.260885275522)*x_ref[1]
        ref[3,4,0,0]=(0.0717612359082)*x_ref[0]+(-0.576982745638)*x_ref[1]
        ref[3,4,0,1]=(-0.227660227523)*x_ref[0]+(0.232974094548)*x_ref[1]
        ref[3,4,1,0]=(0.767797269581)*x_ref[0]+(0.98857043831)*x_ref[1]
        ref[3,4,1,1]=(0.31200408415)*x_ref[0]+(-0.8317658048)*x_ref[1]
        ref[3,4,2,0]=(0.381688388928)*x_ref[0]+(-0.335620162467)*x_ref[1]
        ref[3,4,2,1]=(0.781228510591)*x_ref[0]+(0.960966953643)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.875410081875)*x[0]+(0.859961622438)*x[1]+(0.716805919947)*x[2]
        arg[0,0,0,1]=(0.343337453662)*x[0]+(0.802028072985)*x[1]+(0.588370750348)*x[2]
        arg[0,0,1,0]=(0.467615640242)*x[0]+(0.0457804028001)*x[1]+(-0.163677176662)*x[2]
        arg[0,0,1,1]=(-0.46378923457)*x[0]+(1.29379360824)*x[1]+(0.183839231748)*x[2]
        arg[0,0,2,0]=(1.65864647053)*x[0]+(1.23523675105)*x[1]+(-0.556155025527)*x[2]
        arg[0,0,2,1]=(-0.41581445233)*x[0]+(-0.0709880820438)*x[1]+(0.45785589934)*x[2]
        arg[0,1,0,0]=(1.83790068156)*x[0]+(-0.1690836925)*x[1]+(1.25356286366)*x[2]
        arg[0,1,0,1]=(-0.22584314958)*x[0]+(0.684595424346)*x[1]+(1.13919885217)*x[2]
        arg[0,1,1,0]=(1.02792150528)*x[0]+(-0.312817626145)*x[1]+(-0.121465758957)*x[2]
        arg[0,1,1,1]=(0.533491564413)*x[0]+(-0.62208812136)*x[1]+(-0.717947781071)*x[2]
        arg[0,1,2,0]=(-0.864505701555)*x[0]+(-0.296047724934)*x[1]+(-0.122917018588)*x[2]
        arg[0,1,2,1]=(1.05520768182)*x[0]+(0.662053590424)*x[1]+(0.0251841065476)*x[2]
        arg[0,2,0,0]=(-0.408305818721)*x[0]+(-0.355992587618)*x[1]+(-0.075231462678)*x[2]
        arg[0,2,0,1]=(0.926563898549)*x[0]+(0.203358704647)*x[1]+(-0.365269987808)*x[2]
        arg[0,2,1,0]=(-0.403860257745)*x[0]+(-0.0312835722524)*x[1]+(0.241379582617)*x[2]
        arg[0,2,1,1]=(-0.287297815371)*x[0]+(-0.381813267253)*x[1]+(0.46047051318)*x[2]
        arg[0,2,2,0]=(-0.168164506906)*x[0]+(-1.24152426841)*x[1]+(-0.629704769515)*x[2]
        arg[0,2,2,1]=(-0.937219354291)*x[0]+(-0.736950161299)*x[1]+(0.284057170092)*x[2]
        arg[0,3,0,0]=(-0.292205445759)*x[0]+(1.21842674086)*x[1]+(0.608979796402)*x[2]
        arg[0,3,0,1]=(0.53232240924)*x[0]+(0.226439017439)*x[1]+(-0.42603939944)*x[2]
        arg[0,3,1,0]=(1.67410358342)*x[0]+(-0.00813619771347)*x[1]+(-0.463361889732)*x[2]
        arg[0,3,1,1]=(1.3991502461)*x[0]+(0.239999857694)*x[1]+(-0.369907067208)*x[2]
        arg[0,3,2,0]=(-0.248767913403)*x[0]+(0.893584158229)*x[1]+(1.97656682548)*x[2]
        arg[0,3,2,1]=(-0.835046144341)*x[0]+(-0.00979850832247)*x[1]+(-0.281439904932)*x[2]
        arg[0,4,0,0]=(0.974179494371)*x[0]+(-0.0432121220464)*x[1]+(-0.475018088554)*x[2]
        arg[0,4,0,1]=(0.245490151034)*x[0]+(-0.367572047585)*x[1]+(-0.926123425123)*x[2]
        arg[0,4,1,0]=(-0.239057491277)*x[0]+(0.339734302574)*x[1]+(0.760546363123)*x[2]
        arg[0,4,1,1]=(-1.0425426898)*x[0]+(0.774210702874)*x[1]+(-1.15990543484)*x[2]
        arg[0,4,2,0]=(1.02994063274)*x[0]+(1.69313195574)*x[1]+(-1.02468862318)*x[2]
        arg[0,4,2,1]=(0.221805673731)*x[0]+(-1.06588014758)*x[1]+(-0.546973870537)*x[2]
        arg[1,0,0,0]=(-0.140765393552)*x[0]+(-0.309057115268)*x[1]+(0.07814948965)*x[2]
        arg[1,0,0,1]=(-0.29622101259)*x[0]+(-1.0654452087)*x[1]+(1.77500323701)*x[2]
        arg[1,0,1,0]=(1.13202947738)*x[0]+(-0.515454892388)*x[1]+(-1.34230346013)*x[2]
        arg[1,0,1,1]=(1.50210238409)*x[0]+(-0.75673088573)*x[1]+(-0.792078632258)*x[2]
        arg[1,0,2,0]=(0.249111383954)*x[0]+(-0.972418007776)*x[1]+(0.766060429469)*x[2]
        arg[1,0,2,1]=(-0.0710717292409)*x[0]+(0.164949706666)*x[1]+(1.46045331387)*x[2]
        arg[1,1,0,0]=(-0.587668209291)*x[0]+(0.303249169818)*x[1]+(0.352197808782)*x[2]
        arg[1,1,0,1]=(-0.145609904248)*x[0]+(0.91895760024)*x[1]+(-1.25063091478)*x[2]
        arg[1,1,1,0]=(0.450869946651)*x[0]+(0.752704827905)*x[1]+(0.132042943533)*x[2]
        arg[1,1,1,1]=(-0.87296460294)*x[0]+(-0.459840609019)*x[1]+(0.422574948491)*x[2]
        arg[1,1,2,0]=(0.828459269131)*x[0]+(0.45869153936)*x[1]+(0.854640684176)*x[2]
        arg[1,1,2,1]=(1.84720399473)*x[0]+(0.0416642546234)*x[1]+(-0.309583524852)*x[2]
        arg[1,2,0,0]=(-1.12563293246)*x[0]+(0.0142184473972)*x[1]+(-1.18150362584)*x[2]
        arg[1,2,0,1]=(-0.0750965768147)*x[0]+(-0.419276687751)*x[1]+(0.521012213863)*x[2]
        arg[1,2,1,0]=(-1.06905807284)*x[0]+(0.692258030743)*x[1]+(0.262254909068)*x[2]
        arg[1,2,1,1]=(0.0552441016888)*x[0]+(-0.131040560322)*x[1]+(-0.321517725963)*x[2]
        arg[1,2,2,0]=(-0.740410514693)*x[0]+(0.219020143899)*x[1]+(1.35104682965)*x[2]
        arg[1,2,2,1]=(-1.38198583939)*x[0]+(-0.19342887842)*x[1]+(-0.780008129494)*x[2]
        arg[1,3,0,0]=(0.378358305605)*x[0]+(1.32138618734)*x[1]+(0.627209784997)*x[2]
        arg[1,3,0,1]=(-0.0383657917537)*x[0]+(-1.41770189628)*x[1]+(-1.21625280681)*x[2]
        arg[1,3,1,0]=(0.819666413875)*x[0]+(0.160369878383)*x[1]+(-0.229388558682)*x[2]
        arg[1,3,1,1]=(0.393409555876)*x[0]+(-0.525985341296)*x[1]+(-1.11892438309)*x[2]
        arg[1,3,2,0]=(0.145077010777)*x[0]+(0.229800008438)*x[1]+(-0.202652862622)*x[2]
        arg[1,3,2,1]=(-1.48698378177)*x[0]+(0.553698672636)*x[1]+(-0.427263736796)*x[2]
        arg[1,4,0,0]=(-0.129527725249)*x[0]+(-0.639824801211)*x[1]+(0.759081980717)*x[2]
        arg[1,4,0,1]=(-0.117772473784)*x[0]+(0.387474104611)*x[1]+(0.507306747135)*x[2]
        arg[1,4,1,0]=(-0.82518622198)*x[0]+(-0.754474320343)*x[1]+(-0.978049152511)*x[2]
        arg[1,4,1,1]=(-0.370385633483)*x[0]+(-0.348429423043)*x[1]+(-0.254278257899)*x[2]
        arg[1,4,2,0]=(1.42596167697)*x[0]+(-0.871785904514)*x[1]+(-0.5343959325)*x[2]
        arg[1,4,2,1]=(0.817877672506)*x[0]+(-0.222622791737)*x[1]+(-0.677237468595)*x[2]
        arg[2,0,0,0]=(0.877660484813)*x[0]+(0.818663846913)*x[1]+(-1.26520874625)*x[2]
        arg[2,0,0,1]=(-0.822344385073)*x[0]+(-0.274691696266)*x[1]+(-0.35079248697)*x[2]
        arg[2,0,1,0]=(-0.319040285312)*x[0]+(-0.750406131559)*x[1]+(-0.920332967032)*x[2]
        arg[2,0,1,1]=(1.48758278445)*x[0]+(1.30587557873)*x[1]+(0.230692557888)*x[2]
        arg[2,0,2,0]=(1.44269354043)*x[0]+(-0.63270356894)*x[1]+(0.229323632455)*x[2]
        arg[2,0,2,1]=(-0.645065877536)*x[0]+(-0.393083499676)*x[1]+(-0.305117645453)*x[2]
        arg[2,1,0,0]=(-0.63304134143)*x[0]+(0.0664768401917)*x[1]+(0.87587743066)*x[2]
        arg[2,1,0,1]=(-0.531856958285)*x[0]+(0.747953146648)*x[1]+(1.23560779546)*x[2]
        arg[2,1,1,0]=(0.0564479101025)*x[0]+(0.546307803332)*x[1]+(0.655044890368)*x[2]
        arg[2,1,1,1]=(-1.12793057228)*x[0]+(-0.736676194013)*x[1]+(0.0375913829411)*x[2]
        arg[2,1,2,0]=(-0.531384188067)*x[0]+(0.249416904081)*x[1]+(0.711430712986)*x[2]
        arg[2,1,2,1]=(0.884244664065)*x[0]+(0.456066941968)*x[1]+(0.181630861785)*x[2]
        arg[2,2,0,0]=(0.0844543345062)*x[0]+(-0.435288041256)*x[1]+(-0.434816763062)*x[2]
        arg[2,2,0,1]=(0.373388105049)*x[0]+(0.239833846429)*x[1]+(-0.437771573282)*x[2]
        arg[2,2,1,0]=(0.534401521907)*x[0]+(0.393077203448)*x[1]+(-0.42440387717)*x[2]
        arg[2,2,1,1]=(0.318683369617)*x[0]+(-0.538864643793)*x[1]+(1.83466945806)*x[2]
        arg[2,2,2,0]=(-1.79048122074)*x[0]+(-1.05441213478)*x[1]+(-0.573963806281)*x[2]
        arg[2,2,2,1]=(0.944017405182)*x[0]+(0.772756791538)*x[1]+(0.0269673237336)*x[2]
        arg[2,3,0,0]=(0.989703689142)*x[0]+(1.21302325945)*x[1]+(-0.92209419073)*x[2]
        arg[2,3,0,1]=(-0.0957244757576)*x[0]+(1.50132638315)*x[1]+(-1.38716099756)*x[2]
        arg[2,3,1,0]=(1.6445441776)*x[0]+(-0.490207070694)*x[1]+(0.626469347781)*x[2]
        arg[2,3,1,1]=(0.0185568939147)*x[0]+(0.412024752941)*x[1]+(-0.22626102813)*x[2]
        arg[2,3,2,0]=(0.0143609822966)*x[0]+(0.115739025893)*x[1]+(-0.544468631925)*x[2]
        arg[2,3,2,1]=(-1.80522341493)*x[0]+(0.33352152695)*x[1]+(-0.841286464932)*x[2]
        arg[2,4,0,0]=(0.0618444122042)*x[0]+(1.91232484622)*x[1]+(-0.406317200023)*x[2]
        arg[2,4,0,1]=(0.279131062922)*x[0]+(0.0194232637634)*x[1]+(-0.00179069113634)*x[2]
        arg[2,4,1,0]=(-0.701344579135)*x[0]+(-0.27410935473)*x[1]+(0.763362100733)*x[2]
        arg[2,4,1,1]=(0.831198601185)*x[0]+(0.90336781771)*x[1]+(1.21856633382)*x[2]
        arg[2,4,2,0]=(0.526278615033)*x[0]+(-1.08673855801)*x[1]+(0.0784355900913)*x[2]
        arg[2,4,2,1]=(0.853226856851)*x[0]+(-1.15919062627)*x[1]+(0.185686031732)*x[2]
        arg[3,0,0,0]=(0.272127498117)*x[0]+(-0.457140162438)*x[1]+(0.318358670541)*x[2]
        arg[3,0,0,1]=(1.71056378307)*x[0]+(0.200052730234)*x[1]+(-0.127733325792)*x[2]
        arg[3,0,1,0]=(-0.285340963261)*x[0]+(-0.524544624385)*x[1]+(0.916366474944)*x[2]
        arg[3,0,1,1]=(-0.625506557557)*x[0]+(1.0858476479)*x[1]+(-1.24478983268)*x[2]
        arg[3,0,2,0]=(-1.37299619916)*x[0]+(-0.482246528202)*x[1]+(-1.36857669075)*x[2]
        arg[3,0,2,1]=(-0.924421844916)*x[0]+(-1.50402098115)*x[1]+(1.00045130673)*x[2]
        arg[3,1,0,0]=(0.0417304700827)*x[0]+(0.554527451004)*x[1]+(0.299758163628)*x[2]
        arg[3,1,0,1]=(0.639611639279)*x[0]+(-0.444675638162)*x[1]+(1.43150046757)*x[2]
        arg[3,1,1,0]=(0.288697220898)*x[0]+(1.32660433773)*x[1]+(0.0889412116759)*x[2]
        arg[3,1,1,1]=(-0.110516770201)*x[0]+(-0.0214248733565)*x[1]+(-0.758638794785)*x[2]
        arg[3,1,2,0]=(1.34357138654)*x[0]+(-0.13759272972)*x[1]+(-0.00183845032784)*x[2]
        arg[3,1,2,1]=(1.44174769627)*x[0]+(0.49190176388)*x[1]+(0.477331360141)*x[2]
        arg[3,2,0,0]=(-0.10284777121)*x[0]+(-0.63294157419)*x[1]+(0.246064460451)*x[2]
        arg[3,2,0,1]=(0.974270980699)*x[0]+(1.03835886161)*x[1]+(0.420000854838)*x[2]
        arg[3,2,1,0]=(-0.00617831615923)*x[0]+(-0.354063634073)*x[1]+(0.764285884619)*x[2]
        arg[3,2,1,1]=(0.210815790746)*x[0]+(0.76609515744)*x[1]+(-0.564401558401)*x[2]
        arg[3,2,2,0]=(1.30471234011)*x[0]+(1.54864119332)*x[1]+(-0.493583650519)*x[2]
        arg[3,2,2,1]=(0.42287513668)*x[0]+(-0.603093868171)*x[1]+(-0.811892974203)*x[2]
        arg[3,3,0,0]=(-1.38358134693)*x[0]+(-1.53706018098)*x[1]+(-1.29701227501)*x[2]
        arg[3,3,0,1]=(0.598581779902)*x[0]+(1.02424010922)*x[1]+(-1.1814313009)*x[2]
        arg[3,3,1,0]=(1.30938529116)*x[0]+(-0.504468477244)*x[1]+(1.63655848854)*x[2]
        arg[3,3,1,1]=(-1.47006424113)*x[0]+(0.225951520203)*x[1]+(-0.412273003169)*x[2]
        arg[3,3,2,0]=(-0.116902234482)*x[0]+(0.488144536044)*x[1]+(1.01732461694)*x[2]
        arg[3,3,2,1]=(1.0935814877)*x[0]+(1.06389385916)*x[1]+(-0.17981400591)*x[2]
        arg[3,4,0,0]=(-0.833265062101)*x[0]+(-0.754461148648)*x[1]+(1.26702012815)*x[2]
        arg[3,4,0,1]=(-0.793744278171)*x[0]+(0.112704521653)*x[1]+(-0.448684922187)*x[2]
        arg[3,4,1,0]=(-0.396153523625)*x[0]+(-0.0370870723568)*x[1]+(0.554347618511)*x[2]
        arg[3,4,1,1]=(-0.323276139711)*x[0]+(-0.397389928392)*x[1]+(-1.1707028324)*x[2]
        arg[3,4,2,0]=(0.823350512513)*x[0]+(1.47105759864)*x[1]+(-0.977456877855)*x[2]
        arg[3,4,2,1]=(-0.719791170741)*x[0]+(-1.00363340645)*x[1]+(0.814612770572)*x[2]
        ref[0,0,0,0]=(-0.875410081875)*x_ref[0]+(0.859961622438)*x_ref[1]+(0.716805919947)*x_ref[2]
        ref[0,0,0,1]=(0.343337453662)*x_ref[0]+(0.802028072985)*x_ref[1]+(0.588370750348)*x_ref[2]
        ref[0,0,1,0]=(0.467615640242)*x_ref[0]+(0.0457804028001)*x_ref[1]+(-0.163677176662)*x_ref[2]
        ref[0,0,1,1]=(-0.46378923457)*x_ref[0]+(1.29379360824)*x_ref[1]+(0.183839231748)*x_ref[2]
        ref[0,0,2,0]=(1.65864647053)*x_ref[0]+(1.23523675105)*x_ref[1]+(-0.556155025527)*x_ref[2]
        ref[0,0,2,1]=(-0.41581445233)*x_ref[0]+(-0.0709880820438)*x_ref[1]+(0.45785589934)*x_ref[2]
        ref[0,1,0,0]=(1.83790068156)*x_ref[0]+(-0.1690836925)*x_ref[1]+(1.25356286366)*x_ref[2]
        ref[0,1,0,1]=(-0.22584314958)*x_ref[0]+(0.684595424346)*x_ref[1]+(1.13919885217)*x_ref[2]
        ref[0,1,1,0]=(1.02792150528)*x_ref[0]+(-0.312817626145)*x_ref[1]+(-0.121465758957)*x_ref[2]
        ref[0,1,1,1]=(0.533491564413)*x_ref[0]+(-0.62208812136)*x_ref[1]+(-0.717947781071)*x_ref[2]
        ref[0,1,2,0]=(-0.864505701555)*x_ref[0]+(-0.296047724934)*x_ref[1]+(-0.122917018588)*x_ref[2]
        ref[0,1,2,1]=(1.05520768182)*x_ref[0]+(0.662053590424)*x_ref[1]+(0.0251841065476)*x_ref[2]
        ref[0,2,0,0]=(-0.408305818721)*x_ref[0]+(-0.355992587618)*x_ref[1]+(-0.075231462678)*x_ref[2]
        ref[0,2,0,1]=(0.926563898549)*x_ref[0]+(0.203358704647)*x_ref[1]+(-0.365269987808)*x_ref[2]
        ref[0,2,1,0]=(-0.403860257745)*x_ref[0]+(-0.0312835722524)*x_ref[1]+(0.241379582617)*x_ref[2]
        ref[0,2,1,1]=(-0.287297815371)*x_ref[0]+(-0.381813267253)*x_ref[1]+(0.46047051318)*x_ref[2]
        ref[0,2,2,0]=(-0.168164506906)*x_ref[0]+(-1.24152426841)*x_ref[1]+(-0.629704769515)*x_ref[2]
        ref[0,2,2,1]=(-0.937219354291)*x_ref[0]+(-0.736950161299)*x_ref[1]+(0.284057170092)*x_ref[2]
        ref[0,3,0,0]=(-0.292205445759)*x_ref[0]+(1.21842674086)*x_ref[1]+(0.608979796402)*x_ref[2]
        ref[0,3,0,1]=(0.53232240924)*x_ref[0]+(0.226439017439)*x_ref[1]+(-0.42603939944)*x_ref[2]
        ref[0,3,1,0]=(1.67410358342)*x_ref[0]+(-0.00813619771347)*x_ref[1]+(-0.463361889732)*x_ref[2]
        ref[0,3,1,1]=(1.3991502461)*x_ref[0]+(0.239999857694)*x_ref[1]+(-0.369907067208)*x_ref[2]
        ref[0,3,2,0]=(-0.248767913403)*x_ref[0]+(0.893584158229)*x_ref[1]+(1.97656682548)*x_ref[2]
        ref[0,3,2,1]=(-0.835046144341)*x_ref[0]+(-0.00979850832247)*x_ref[1]+(-0.281439904932)*x_ref[2]
        ref[0,4,0,0]=(0.974179494371)*x_ref[0]+(-0.0432121220464)*x_ref[1]+(-0.475018088554)*x_ref[2]
        ref[0,4,0,1]=(0.245490151034)*x_ref[0]+(-0.367572047585)*x_ref[1]+(-0.926123425123)*x_ref[2]
        ref[0,4,1,0]=(-0.239057491277)*x_ref[0]+(0.339734302574)*x_ref[1]+(0.760546363123)*x_ref[2]
        ref[0,4,1,1]=(-1.0425426898)*x_ref[0]+(0.774210702874)*x_ref[1]+(-1.15990543484)*x_ref[2]
        ref[0,4,2,0]=(1.02994063274)*x_ref[0]+(1.69313195574)*x_ref[1]+(-1.02468862318)*x_ref[2]
        ref[0,4,2,1]=(0.221805673731)*x_ref[0]+(-1.06588014758)*x_ref[1]+(-0.546973870537)*x_ref[2]
        ref[1,0,0,0]=(-0.140765393552)*x_ref[0]+(-0.309057115268)*x_ref[1]+(0.07814948965)*x_ref[2]
        ref[1,0,0,1]=(-0.29622101259)*x_ref[0]+(-1.0654452087)*x_ref[1]+(1.77500323701)*x_ref[2]
        ref[1,0,1,0]=(1.13202947738)*x_ref[0]+(-0.515454892388)*x_ref[1]+(-1.34230346013)*x_ref[2]
        ref[1,0,1,1]=(1.50210238409)*x_ref[0]+(-0.75673088573)*x_ref[1]+(-0.792078632258)*x_ref[2]
        ref[1,0,2,0]=(0.249111383954)*x_ref[0]+(-0.972418007776)*x_ref[1]+(0.766060429469)*x_ref[2]
        ref[1,0,2,1]=(-0.0710717292409)*x_ref[0]+(0.164949706666)*x_ref[1]+(1.46045331387)*x_ref[2]
        ref[1,1,0,0]=(-0.587668209291)*x_ref[0]+(0.303249169818)*x_ref[1]+(0.352197808782)*x_ref[2]
        ref[1,1,0,1]=(-0.145609904248)*x_ref[0]+(0.91895760024)*x_ref[1]+(-1.25063091478)*x_ref[2]
        ref[1,1,1,0]=(0.450869946651)*x_ref[0]+(0.752704827905)*x_ref[1]+(0.132042943533)*x_ref[2]
        ref[1,1,1,1]=(-0.87296460294)*x_ref[0]+(-0.459840609019)*x_ref[1]+(0.422574948491)*x_ref[2]
        ref[1,1,2,0]=(0.828459269131)*x_ref[0]+(0.45869153936)*x_ref[1]+(0.854640684176)*x_ref[2]
        ref[1,1,2,1]=(1.84720399473)*x_ref[0]+(0.0416642546234)*x_ref[1]+(-0.309583524852)*x_ref[2]
        ref[1,2,0,0]=(-1.12563293246)*x_ref[0]+(0.0142184473972)*x_ref[1]+(-1.18150362584)*x_ref[2]
        ref[1,2,0,1]=(-0.0750965768147)*x_ref[0]+(-0.419276687751)*x_ref[1]+(0.521012213863)*x_ref[2]
        ref[1,2,1,0]=(-1.06905807284)*x_ref[0]+(0.692258030743)*x_ref[1]+(0.262254909068)*x_ref[2]
        ref[1,2,1,1]=(0.0552441016888)*x_ref[0]+(-0.131040560322)*x_ref[1]+(-0.321517725963)*x_ref[2]
        ref[1,2,2,0]=(-0.740410514693)*x_ref[0]+(0.219020143899)*x_ref[1]+(1.35104682965)*x_ref[2]
        ref[1,2,2,1]=(-1.38198583939)*x_ref[0]+(-0.19342887842)*x_ref[1]+(-0.780008129494)*x_ref[2]
        ref[1,3,0,0]=(0.378358305605)*x_ref[0]+(1.32138618734)*x_ref[1]+(0.627209784997)*x_ref[2]
        ref[1,3,0,1]=(-0.0383657917537)*x_ref[0]+(-1.41770189628)*x_ref[1]+(-1.21625280681)*x_ref[2]
        ref[1,3,1,0]=(0.819666413875)*x_ref[0]+(0.160369878383)*x_ref[1]+(-0.229388558682)*x_ref[2]
        ref[1,3,1,1]=(0.393409555876)*x_ref[0]+(-0.525985341296)*x_ref[1]+(-1.11892438309)*x_ref[2]
        ref[1,3,2,0]=(0.145077010777)*x_ref[0]+(0.229800008438)*x_ref[1]+(-0.202652862622)*x_ref[2]
        ref[1,3,2,1]=(-1.48698378177)*x_ref[0]+(0.553698672636)*x_ref[1]+(-0.427263736796)*x_ref[2]
        ref[1,4,0,0]=(-0.129527725249)*x_ref[0]+(-0.639824801211)*x_ref[1]+(0.759081980717)*x_ref[2]
        ref[1,4,0,1]=(-0.117772473784)*x_ref[0]+(0.387474104611)*x_ref[1]+(0.507306747135)*x_ref[2]
        ref[1,4,1,0]=(-0.82518622198)*x_ref[0]+(-0.754474320343)*x_ref[1]+(-0.978049152511)*x_ref[2]
        ref[1,4,1,1]=(-0.370385633483)*x_ref[0]+(-0.348429423043)*x_ref[1]+(-0.254278257899)*x_ref[2]
        ref[1,4,2,0]=(1.42596167697)*x_ref[0]+(-0.871785904514)*x_ref[1]+(-0.5343959325)*x_ref[2]
        ref[1,4,2,1]=(0.817877672506)*x_ref[0]+(-0.222622791737)*x_ref[1]+(-0.677237468595)*x_ref[2]
        ref[2,0,0,0]=(0.877660484813)*x_ref[0]+(0.818663846913)*x_ref[1]+(-1.26520874625)*x_ref[2]
        ref[2,0,0,1]=(-0.822344385073)*x_ref[0]+(-0.274691696266)*x_ref[1]+(-0.35079248697)*x_ref[2]
        ref[2,0,1,0]=(-0.319040285312)*x_ref[0]+(-0.750406131559)*x_ref[1]+(-0.920332967032)*x_ref[2]
        ref[2,0,1,1]=(1.48758278445)*x_ref[0]+(1.30587557873)*x_ref[1]+(0.230692557888)*x_ref[2]
        ref[2,0,2,0]=(1.44269354043)*x_ref[0]+(-0.63270356894)*x_ref[1]+(0.229323632455)*x_ref[2]
        ref[2,0,2,1]=(-0.645065877536)*x_ref[0]+(-0.393083499676)*x_ref[1]+(-0.305117645453)*x_ref[2]
        ref[2,1,0,0]=(-0.63304134143)*x_ref[0]+(0.0664768401917)*x_ref[1]+(0.87587743066)*x_ref[2]
        ref[2,1,0,1]=(-0.531856958285)*x_ref[0]+(0.747953146648)*x_ref[1]+(1.23560779546)*x_ref[2]
        ref[2,1,1,0]=(0.0564479101025)*x_ref[0]+(0.546307803332)*x_ref[1]+(0.655044890368)*x_ref[2]
        ref[2,1,1,1]=(-1.12793057228)*x_ref[0]+(-0.736676194013)*x_ref[1]+(0.0375913829411)*x_ref[2]
        ref[2,1,2,0]=(-0.531384188067)*x_ref[0]+(0.249416904081)*x_ref[1]+(0.711430712986)*x_ref[2]
        ref[2,1,2,1]=(0.884244664065)*x_ref[0]+(0.456066941968)*x_ref[1]+(0.181630861785)*x_ref[2]
        ref[2,2,0,0]=(0.0844543345062)*x_ref[0]+(-0.435288041256)*x_ref[1]+(-0.434816763062)*x_ref[2]
        ref[2,2,0,1]=(0.373388105049)*x_ref[0]+(0.239833846429)*x_ref[1]+(-0.437771573282)*x_ref[2]
        ref[2,2,1,0]=(0.534401521907)*x_ref[0]+(0.393077203448)*x_ref[1]+(-0.42440387717)*x_ref[2]
        ref[2,2,1,1]=(0.318683369617)*x_ref[0]+(-0.538864643793)*x_ref[1]+(1.83466945806)*x_ref[2]
        ref[2,2,2,0]=(-1.79048122074)*x_ref[0]+(-1.05441213478)*x_ref[1]+(-0.573963806281)*x_ref[2]
        ref[2,2,2,1]=(0.944017405182)*x_ref[0]+(0.772756791538)*x_ref[1]+(0.0269673237336)*x_ref[2]
        ref[2,3,0,0]=(0.989703689142)*x_ref[0]+(1.21302325945)*x_ref[1]+(-0.92209419073)*x_ref[2]
        ref[2,3,0,1]=(-0.0957244757576)*x_ref[0]+(1.50132638315)*x_ref[1]+(-1.38716099756)*x_ref[2]
        ref[2,3,1,0]=(1.6445441776)*x_ref[0]+(-0.490207070694)*x_ref[1]+(0.626469347781)*x_ref[2]
        ref[2,3,1,1]=(0.0185568939147)*x_ref[0]+(0.412024752941)*x_ref[1]+(-0.22626102813)*x_ref[2]
        ref[2,3,2,0]=(0.0143609822966)*x_ref[0]+(0.115739025893)*x_ref[1]+(-0.544468631925)*x_ref[2]
        ref[2,3,2,1]=(-1.80522341493)*x_ref[0]+(0.33352152695)*x_ref[1]+(-0.841286464932)*x_ref[2]
        ref[2,4,0,0]=(0.0618444122042)*x_ref[0]+(1.91232484622)*x_ref[1]+(-0.406317200023)*x_ref[2]
        ref[2,4,0,1]=(0.279131062922)*x_ref[0]+(0.0194232637634)*x_ref[1]+(-0.00179069113634)*x_ref[2]
        ref[2,4,1,0]=(-0.701344579135)*x_ref[0]+(-0.27410935473)*x_ref[1]+(0.763362100733)*x_ref[2]
        ref[2,4,1,1]=(0.831198601185)*x_ref[0]+(0.90336781771)*x_ref[1]+(1.21856633382)*x_ref[2]
        ref[2,4,2,0]=(0.526278615033)*x_ref[0]+(-1.08673855801)*x_ref[1]+(0.0784355900913)*x_ref[2]
        ref[2,4,2,1]=(0.853226856851)*x_ref[0]+(-1.15919062627)*x_ref[1]+(0.185686031732)*x_ref[2]
        ref[3,0,0,0]=(0.272127498117)*x_ref[0]+(-0.457140162438)*x_ref[1]+(0.318358670541)*x_ref[2]
        ref[3,0,0,1]=(1.71056378307)*x_ref[0]+(0.200052730234)*x_ref[1]+(-0.127733325792)*x_ref[2]
        ref[3,0,1,0]=(-0.285340963261)*x_ref[0]+(-0.524544624385)*x_ref[1]+(0.916366474944)*x_ref[2]
        ref[3,0,1,1]=(-0.625506557557)*x_ref[0]+(1.0858476479)*x_ref[1]+(-1.24478983268)*x_ref[2]
        ref[3,0,2,0]=(-1.37299619916)*x_ref[0]+(-0.482246528202)*x_ref[1]+(-1.36857669075)*x_ref[2]
        ref[3,0,2,1]=(-0.924421844916)*x_ref[0]+(-1.50402098115)*x_ref[1]+(1.00045130673)*x_ref[2]
        ref[3,1,0,0]=(0.0417304700827)*x_ref[0]+(0.554527451004)*x_ref[1]+(0.299758163628)*x_ref[2]
        ref[3,1,0,1]=(0.639611639279)*x_ref[0]+(-0.444675638162)*x_ref[1]+(1.43150046757)*x_ref[2]
        ref[3,1,1,0]=(0.288697220898)*x_ref[0]+(1.32660433773)*x_ref[1]+(0.0889412116759)*x_ref[2]
        ref[3,1,1,1]=(-0.110516770201)*x_ref[0]+(-0.0214248733565)*x_ref[1]+(-0.758638794785)*x_ref[2]
        ref[3,1,2,0]=(1.34357138654)*x_ref[0]+(-0.13759272972)*x_ref[1]+(-0.00183845032784)*x_ref[2]
        ref[3,1,2,1]=(1.44174769627)*x_ref[0]+(0.49190176388)*x_ref[1]+(0.477331360141)*x_ref[2]
        ref[3,2,0,0]=(-0.10284777121)*x_ref[0]+(-0.63294157419)*x_ref[1]+(0.246064460451)*x_ref[2]
        ref[3,2,0,1]=(0.974270980699)*x_ref[0]+(1.03835886161)*x_ref[1]+(0.420000854838)*x_ref[2]
        ref[3,2,1,0]=(-0.00617831615923)*x_ref[0]+(-0.354063634073)*x_ref[1]+(0.764285884619)*x_ref[2]
        ref[3,2,1,1]=(0.210815790746)*x_ref[0]+(0.76609515744)*x_ref[1]+(-0.564401558401)*x_ref[2]
        ref[3,2,2,0]=(1.30471234011)*x_ref[0]+(1.54864119332)*x_ref[1]+(-0.493583650519)*x_ref[2]
        ref[3,2,2,1]=(0.42287513668)*x_ref[0]+(-0.603093868171)*x_ref[1]+(-0.811892974203)*x_ref[2]
        ref[3,3,0,0]=(-1.38358134693)*x_ref[0]+(-1.53706018098)*x_ref[1]+(-1.29701227501)*x_ref[2]
        ref[3,3,0,1]=(0.598581779902)*x_ref[0]+(1.02424010922)*x_ref[1]+(-1.1814313009)*x_ref[2]
        ref[3,3,1,0]=(1.30938529116)*x_ref[0]+(-0.504468477244)*x_ref[1]+(1.63655848854)*x_ref[2]
        ref[3,3,1,1]=(-1.47006424113)*x_ref[0]+(0.225951520203)*x_ref[1]+(-0.412273003169)*x_ref[2]
        ref[3,3,2,0]=(-0.116902234482)*x_ref[0]+(0.488144536044)*x_ref[1]+(1.01732461694)*x_ref[2]
        ref[3,3,2,1]=(1.0935814877)*x_ref[0]+(1.06389385916)*x_ref[1]+(-0.17981400591)*x_ref[2]
        ref[3,4,0,0]=(-0.833265062101)*x_ref[0]+(-0.754461148648)*x_ref[1]+(1.26702012815)*x_ref[2]
        ref[3,4,0,1]=(-0.793744278171)*x_ref[0]+(0.112704521653)*x_ref[1]+(-0.448684922187)*x_ref[2]
        ref[3,4,1,0]=(-0.396153523625)*x_ref[0]+(-0.0370870723568)*x_ref[1]+(0.554347618511)*x_ref[2]
        ref[3,4,1,1]=(-0.323276139711)*x_ref[0]+(-0.397389928392)*x_ref[1]+(-1.1707028324)*x_ref[2]
        ref[3,4,2,0]=(0.823350512513)*x_ref[0]+(1.47105759864)*x_ref[1]+(-0.977456877855)*x_ref[2]
        ref[3,4,2,1]=(-0.719791170741)*x_ref[0]+(-1.00363340645)*x_ref[1]+(0.814612770572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedContinuousFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ContinuousFunction(self):
      """
      tests divergence of Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.627139584179)*x[0]**o+(0.0871027334774)*x[0]+(-0.540784323698)*x[1]**o+(0.951143032871)*x[1]
        arg[1]=(-0.119239347115)*x[0]**o+(0.277154010744)*x[0]+(-0.0745557686973)*x[1]**o+(-0.15832879535)*x[1]
        ref=o*(-0.627139584179)*x_ref[0]**(o-1)+o*(-0.0745557686973)*x_ref[1]**(o-1)+(-0.0712260618722)
      else:
        arg[0]=(0.687070465546)*x[0]**o+(0.789237250884)*x[0]+(-0.00415360035876)*x[1]**o+(-0.809802614239)*x[1]+(0.418087966616)*x[2]**o+(0.730203229934)*x[2]
        arg[1]=(-0.0962217374268)*x[0]**o+(-0.180692492986)*x[0]+(-0.59921793179)*x[1]**o+(-0.799342788982)*x[1]+(0.421364766686)*x[2]**o+(-0.586211831795)*x[2]
        arg[2]=(0.774915504102)*x[0]**o+(-0.968436517935)*x[0]+(0.325821385033)*x[1]**o+(-0.159885723341)*x[1]+(0.411134413872)*x[2]**o+(0.083388357951)*x[2]
        ref=o*(0.687070465546)*x_ref[0]**(o-1)+o*(-0.59921793179)*x_ref[1]**(o-1)+o*(0.411134413872)*x_ref[2]**(o-1)+(0.0732828198535)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_Solution(self):
      """
      tests divergence of Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.037254966228)*x[0]**o+(-0.175442318226)*x[0]+(-0.723089592066)*x[1]**o+(-0.487393332634)*x[1]
        arg[1]=(0.617589532887)*x[0]**o+(0.392126578405)*x[0]+(0.612110931251)*x[1]**o+(-0.932236829843)*x[1]
        ref=o*(-0.037254966228)*x_ref[0]**(o-1)+o*(0.612110931251)*x_ref[1]**(o-1)+(-1.10767914807)
      else:
        arg[0]=(-0.855344637098)*x[0]**o+(0.152992485241)*x[0]+(0.519809077415)*x[1]**o+(-0.0514985311411)*x[1]+(-0.473095642241)*x[2]**o+(-0.743711945025)*x[2]
        arg[1]=(0.629819444749)*x[0]**o+(-0.649277860811)*x[0]+(-0.382231371814)*x[1]**o+(-0.354475847319)*x[1]+(0.523961319618)*x[2]**o+(-0.0301027430993)*x[2]
        arg[2]=(0.489214433445)*x[0]**o+(-0.340824606014)*x[0]+(-0.78445257629)*x[1]**o+(-0.821498813523)*x[1]+(-0.147119976738)*x[2]**o+(0.464427291378)*x[2]
        ref=o*(-0.855344637098)*x_ref[0]**(o-1)+o*(-0.382231371814)*x_ref[1]**(o-1)+o*(-0.147119976738)*x_ref[2]**(o-1)+(0.2629439293)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ReducedSolution(self):
      """
      tests divergence of Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ReducedContinuousFunction(self):
      """
      tests divergence of Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ContinuousFunction(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.627139584179)*x[0]**o+(0.0871027334774)*x[0]+(-0.540784323698)*x[1]**o+(0.951143032871)*x[1]
        arg[1]=(-0.119239347115)*x[0]**o+(0.277154010744)*x[0]+(-0.0745557686973)*x[1]**o+(-0.15832879535)*x[1]
        ref=o*(-0.627139584179)*x_ref[0]**(o-1)+o*(-0.0745557686973)*x_ref[1]**(o-1)+(-0.0712260618722)
      else:
        arg[0]=(0.687070465546)*x[0]**o+(0.789237250884)*x[0]+(-0.00415360035876)*x[1]**o+(-0.809802614239)*x[1]+(0.418087966616)*x[2]**o+(0.730203229934)*x[2]
        arg[1]=(-0.0962217374268)*x[0]**o+(-0.180692492986)*x[0]+(-0.59921793179)*x[1]**o+(-0.799342788982)*x[1]+(0.421364766686)*x[2]**o+(-0.586211831795)*x[2]
        arg[2]=(0.774915504102)*x[0]**o+(-0.968436517935)*x[0]+(0.325821385033)*x[1]**o+(-0.159885723341)*x[1]+(0.411134413872)*x[2]**o+(0.083388357951)*x[2]
        ref=o*(0.687070465546)*x_ref[0]**(o-1)+o*(-0.59921793179)*x_ref[1]**(o-1)+o*(0.411134413872)*x_ref[2]**(o-1)+(0.0732828198535)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_Solution(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.037254966228)*x[0]**o+(-0.175442318226)*x[0]+(-0.723089592066)*x[1]**o+(-0.487393332634)*x[1]
        arg[1]=(0.617589532887)*x[0]**o+(0.392126578405)*x[0]+(0.612110931251)*x[1]**o+(-0.932236829843)*x[1]
        ref=o*(-0.037254966228)*x_ref[0]**(o-1)+o*(0.612110931251)*x_ref[1]**(o-1)+(-1.10767914807)
      else:
        arg[0]=(-0.855344637098)*x[0]**o+(0.152992485241)*x[0]+(0.519809077415)*x[1]**o+(-0.0514985311411)*x[1]+(-0.473095642241)*x[2]**o+(-0.743711945025)*x[2]
        arg[1]=(0.629819444749)*x[0]**o+(-0.649277860811)*x[0]+(-0.382231371814)*x[1]**o+(-0.354475847319)*x[1]+(0.523961319618)*x[2]**o+(-0.0301027430993)*x[2]
        arg[2]=(0.489214433445)*x[0]**o+(-0.340824606014)*x[0]+(-0.78445257629)*x[1]**o+(-0.821498813523)*x[1]+(-0.147119976738)*x[2]**o+(0.464427291378)*x[2]
        ref=o*(-0.855344637098)*x_ref[0]**(o-1)+o*(-0.382231371814)*x_ref[1]**(o-1)+o*(-0.147119976738)*x_ref[2]**(o-1)+(0.2629439293)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ReducedSolution(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ReducedContinuousFunction(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertLess(Lsup(res-ref), self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank0(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(0.608797336225)*x[0]
        ref=sqrt((0.123544732198))

      else:
        arg=Data(0,(),w)
        arg=(0.136031275673)*x[0]
        ref=sqrt((0.00616816932037))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank1(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(-0.212143919436)*x[0]
        arg[1]=(-0.256194155686)*x[1]
        ref=sqrt((0.0368801626538))

      else:
        arg=Data(0,(3,),w)
        arg[0]=(0.0452831341416)*x[0]
        arg[1]=(-0.278640180656)*x[1]
        arg[2]=(-0.607035001062)*x[2]
        ref=sqrt((0.149394135009))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank2(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.239448813076)*x[0]
        arg[0,1]=(-0.529349708753)*x[1]
        arg[1,0]=(-0.381557161859)*x[0]
        arg[1,1]=(0.731658534249)*x[1]
        arg[2,0]=(-0.813679062342)*x[0]
        arg[2,1]=(0.528100089704)*x[1]
        arg[3,0]=(-0.480867528161)*x[0]
        arg[3,1]=(-0.167862206972)*x[1]
        ref=sqrt((0.739610516051))

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(0.951209543612)*x[0]
        arg[0,1]=(0.735178735637)*x[1]
        arg[0,2]=(0.13074673272)*x[2]
        arg[1,0]=(0.412295676715)*x[0]
        arg[1,1]=(-0.657695950153)*x[1]
        arg[1,2]=(-0.900044734695)*x[2]
        arg[2,0]=(0.741773926224)*x[0]
        arg[2,1]=(0.0521828807406)*x[1]
        arg[2,2]=(0.797728501985)*x[2]
        arg[3,0]=(-0.61235554051)*x[0]
        arg[3,1]=(0.456652747412)*x[1]
        arg[3,2]=(-0.734303857319)*x[2]
        ref=sqrt((1.72901661926))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank3(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.449174971953)*x[0]
        arg[0,0,1]=(-0.0109398763289)*x[1]
        arg[0,1,0]=(-0.202497187709)*x[0]
        arg[0,1,1]=(-0.12970879334)*x[1]
        arg[1,0,0]=(-0.138092481719)*x[0]
        arg[1,0,1]=(-0.528752200917)*x[1]
        arg[1,1,0]=(-0.605919441662)*x[0]
        arg[1,1,1]=(0.215615032334)*x[1]
        arg[2,0,0]=(-0.998734541972)*x[0]
        arg[2,0,1]=(0.725811901251)*x[1]
        arg[2,1,0]=(-0.966536503228)*x[0]
        arg[2,1,1]=(-0.528692217355)*x[1]
        arg[3,0,0]=(0.757633851466)*x[0]
        arg[3,0,1]=(-0.524660157377)*x[1]
        arg[3,1,0]=(0.983733431677)*x[0]
        arg[3,1,1]=(0.061279109546)*x[1]
        arg[4,0,0]=(0.85914215305)*x[0]
        arg[4,0,1]=(0.941714045112)*x[1]
        arg[4,1,0]=(0.172235529555)*x[0]
        arg[4,1,1]=(-0.108381454437)*x[1]
        arg[5,0,0]=(-0.736373697727)*x[0]
        arg[5,0,1]=(-0.599337929679)*x[1]
        arg[5,1,0]=(0.661072686392)*x[0]
        arg[5,1,1]=(-0.55107327409)*x[1]
        ref=sqrt((2.94641432714))

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.69227064904)*x[0]
        arg[0,0,1]=(-0.968336177418)*x[1]
        arg[0,0,2]=(-0.634883146685)*x[2]
        arg[0,1,0]=(-0.12640661422)*x[0]
        arg[0,1,1]=(-0.637386589888)*x[1]
        arg[0,1,2]=(0.26060859356)*x[2]
        arg[1,0,0]=(-0.986864633297)*x[0]
        arg[1,0,1]=(-0.441589142379)*x[1]
        arg[1,0,2]=(-0.587865539582)*x[2]
        arg[1,1,0]=(0.596052465031)*x[0]
        arg[1,1,1]=(0.312732336652)*x[1]
        arg[1,1,2]=(-0.514423945092)*x[2]
        arg[2,0,0]=(-0.892391254794)*x[0]
        arg[2,0,1]=(0.377920185756)*x[1]
        arg[2,0,2]=(-0.120174597181)*x[2]
        arg[2,1,0]=(-0.469951576468)*x[0]
        arg[2,1,1]=(-0.788362249555)*x[1]
        arg[2,1,2]=(0.745625354986)*x[2]
        arg[3,0,0]=(0.542802498569)*x[0]
        arg[3,0,1]=(-0.814541028706)*x[1]
        arg[3,0,2]=(0.298410992196)*x[2]
        arg[3,1,0]=(0.981190341206)*x[0]
        arg[3,1,1]=(0.666421298608)*x[1]
        arg[3,1,2]=(-0.369751722626)*x[2]
        arg[4,0,0]=(-0.75379530597)*x[0]
        arg[4,0,1]=(0.283357267139)*x[1]
        arg[4,0,2]=(0.247787072861)*x[2]
        arg[4,1,0]=(0.301766692533)*x[0]
        arg[4,1,1]=(0.828183439224)*x[1]
        arg[4,1,2]=(-0.580824060547)*x[2]
        arg[5,0,0]=(0.637345610764)*x[0]
        arg[5,0,1]=(-0.234409115997)*x[1]
        arg[5,0,2]=(-0.192639300316)*x[2]
        arg[5,1,0]=(-0.62609237162)*x[0]
        arg[5,1,1]=(0.463404958552)*x[1]
        arg[5,1,2]=(-0.547814448738)*x[2]
        ref=sqrt((4.2381131862))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank4(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.232618585183)*x[0]
        arg[0,0,0,1]=(0.39796117869)*x[1]
        arg[0,0,1,0]=(-0.997336958262)*x[0]
        arg[0,0,1,1]=(-0.351780915076)*x[1]
        arg[0,0,2,0]=(-0.876764070136)*x[0]
        arg[0,0,2,1]=(0.808730805817)*x[1]
        arg[0,1,0,0]=(-0.197154744966)*x[0]
        arg[0,1,0,1]=(0.416246096086)*x[1]
        arg[0,1,1,0]=(0.708038457121)*x[0]
        arg[0,1,1,1]=(-0.00954021503188)*x[1]
        arg[0,1,2,0]=(-0.62630809425)*x[0]
        arg[0,1,2,1]=(0.430228727912)*x[1]
        arg[0,2,0,0]=(0.0306704747648)*x[0]
        arg[0,2,0,1]=(-0.913877199453)*x[1]
        arg[0,2,1,0]=(-0.697612800829)*x[0]
        arg[0,2,1,1]=(-0.17996376822)*x[1]
        arg[0,2,2,0]=(-0.304509578871)*x[0]
        arg[0,2,2,1]=(-0.610556755811)*x[1]
        arg[0,3,0,0]=(-0.452355972234)*x[0]
        arg[0,3,0,1]=(-0.368921242518)*x[1]
        arg[0,3,1,0]=(-0.478275554932)*x[0]
        arg[0,3,1,1]=(0.257178549127)*x[1]
        arg[0,3,2,0]=(0.530736487177)*x[0]
        arg[0,3,2,1]=(-0.567126272463)*x[1]
        arg[0,4,0,0]=(0.801519165938)*x[0]
        arg[0,4,0,1]=(-0.509816703951)*x[1]
        arg[0,4,1,0]=(-0.255412646934)*x[0]
        arg[0,4,1,1]=(0.437540101896)*x[1]
        arg[0,4,2,0]=(-0.815574969538)*x[0]
        arg[0,4,2,1]=(-0.94691547137)*x[1]
        arg[1,0,0,0]=(-0.732550722593)*x[0]
        arg[1,0,0,1]=(0.515752381704)*x[1]
        arg[1,0,1,0]=(-0.343590210899)*x[0]
        arg[1,0,1,1]=(-0.0601907964915)*x[1]
        arg[1,0,2,0]=(0.0199916154421)*x[0]
        arg[1,0,2,1]=(-0.136927227821)*x[1]
        arg[1,1,0,0]=(0.397994441702)*x[0]
        arg[1,1,0,1]=(0.953873148948)*x[1]
        arg[1,1,1,0]=(0.419416235967)*x[0]
        arg[1,1,1,1]=(0.700998577193)*x[1]
        arg[1,1,2,0]=(-0.497358799271)*x[0]
        arg[1,1,2,1]=(0.0851768858379)*x[1]
        arg[1,2,0,0]=(0.0936678875202)*x[0]
        arg[1,2,0,1]=(0.869883786896)*x[1]
        arg[1,2,1,0]=(0.582700123485)*x[0]
        arg[1,2,1,1]=(-0.433381106794)*x[1]
        arg[1,2,2,0]=(-0.527031777974)*x[0]
        arg[1,2,2,1]=(0.105105137652)*x[1]
        arg[1,3,0,0]=(-0.716750829134)*x[0]
        arg[1,3,0,1]=(0.774519209008)*x[1]
        arg[1,3,1,0]=(-0.568743372716)*x[0]
        arg[1,3,1,1]=(0.794732483944)*x[1]
        arg[1,3,2,0]=(0.246606002015)*x[0]
        arg[1,3,2,1]=(-0.988869494994)*x[1]
        arg[1,4,0,0]=(0.482379298083)*x[0]
        arg[1,4,0,1]=(-0.386268387903)*x[1]
        arg[1,4,1,0]=(0.137184889675)*x[0]
        arg[1,4,1,1]=(-0.140520035321)*x[1]
        arg[1,4,2,0]=(0.822755050415)*x[0]
        arg[1,4,2,1]=(-0.815562139522)*x[1]
        arg[2,0,0,0]=(-0.462891511962)*x[0]
        arg[2,0,0,1]=(-0.122643411631)*x[1]
        arg[2,0,1,0]=(-0.520861119962)*x[0]
        arg[2,0,1,1]=(-0.881189618018)*x[1]
        arg[2,0,2,0]=(-0.776157842774)*x[0]
        arg[2,0,2,1]=(-0.12354053207)*x[1]
        arg[2,1,0,0]=(0.395495230826)*x[0]
        arg[2,1,0,1]=(-0.388106659423)*x[1]
        arg[2,1,1,0]=(0.354250242834)*x[0]
        arg[2,1,1,1]=(-0.666514210192)*x[1]
        arg[2,1,2,0]=(0.951294655083)*x[0]
        arg[2,1,2,1]=(0.074024416386)*x[1]
        arg[2,2,0,0]=(0.335448485459)*x[0]
        arg[2,2,0,1]=(-0.40988282528)*x[1]
        arg[2,2,1,0]=(-0.805725968875)*x[0]
        arg[2,2,1,1]=(-0.949883082118)*x[1]
        arg[2,2,2,0]=(0.531549210683)*x[0]
        arg[2,2,2,1]=(-0.398401016682)*x[1]
        arg[2,3,0,0]=(-0.953963433205)*x[0]
        arg[2,3,0,1]=(0.643431126406)*x[1]
        arg[2,3,1,0]=(-0.167611998738)*x[0]
        arg[2,3,1,1]=(0.226130056552)*x[1]
        arg[2,3,2,0]=(0.0752687641131)*x[0]
        arg[2,3,2,1]=(-0.115742756362)*x[1]
        arg[2,4,0,0]=(0.579694491028)*x[0]
        arg[2,4,0,1]=(-0.112005738299)*x[1]
        arg[2,4,1,0]=(0.657291764224)*x[0]
        arg[2,4,1,1]=(0.62671154177)*x[1]
        arg[2,4,2,0]=(0.103695027944)*x[0]
        arg[2,4,2,1]=(0.462828491544)*x[1]
        arg[3,0,0,0]=(0.697692979998)*x[0]
        arg[3,0,0,1]=(-0.123481859619)*x[1]
        arg[3,0,1,0]=(-0.749745629459)*x[0]
        arg[3,0,1,1]=(-0.541969524069)*x[1]
        arg[3,0,2,0]=(0.819484470759)*x[0]
        arg[3,0,2,1]=(-0.860592326469)*x[1]
        arg[3,1,0,0]=(-0.716566084771)*x[0]
        arg[3,1,0,1]=(-0.949235434827)*x[1]
        arg[3,1,1,0]=(-0.826699498174)*x[0]
        arg[3,1,1,1]=(-0.138511521583)*x[1]
        arg[3,1,2,0]=(-0.951682890904)*x[0]
        arg[3,1,2,1]=(0.413293316925)*x[1]
        arg[3,2,0,0]=(0.909516836775)*x[0]
        arg[3,2,0,1]=(-0.919989721277)*x[1]
        arg[3,2,1,0]=(0.0994860369337)*x[0]
        arg[3,2,1,1]=(-0.933647246623)*x[1]
        arg[3,2,2,0]=(-0.759215183015)*x[0]
        arg[3,2,2,1]=(0.0975793309286)*x[1]
        arg[3,3,0,0]=(-0.130256739381)*x[0]
        arg[3,3,0,1]=(-0.582280862311)*x[1]
        arg[3,3,1,0]=(0.206970526192)*x[0]
        arg[3,3,1,1]=(-0.8678322258)*x[1]
        arg[3,3,2,0]=(0.133004501279)*x[0]
        arg[3,3,2,1]=(0.802921710935)*x[1]
        arg[3,4,0,0]=(-0.255737792764)*x[0]
        arg[3,4,0,1]=(-0.34168114937)*x[1]
        arg[3,4,1,0]=(-0.859309090399)*x[0]
        arg[3,4,1,1]=(0.245043986435)*x[1]
        arg[3,4,2,0]=(0.893062018695)*x[0]
        arg[3,4,2,1]=(0.709422742588)*x[1]
        ref=sqrt((13.7289280362))

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.0312828390439)*x[0]
        arg[0,0,0,1]=(-0.524970416212)*x[1]
        arg[0,0,0,2]=(0.561865217554)*x[2]
        arg[0,0,1,0]=(0.692457187384)*x[0]
        arg[0,0,1,1]=(0.946967182157)*x[1]
        arg[0,0,1,2]=(-0.863842279464)*x[2]
        arg[0,0,2,0]=(0.993922921598)*x[0]
        arg[0,0,2,1]=(0.322812768679)*x[1]
        arg[0,0,2,2]=(0.901876132204)*x[2]
        arg[0,1,0,0]=(0.967569979365)*x[0]
        arg[0,1,0,1]=(0.840979131355)*x[1]
        arg[0,1,0,2]=(0.0494811460856)*x[2]
        arg[0,1,1,0]=(0.315178456102)*x[0]
        arg[0,1,1,1]=(0.449848313024)*x[1]
        arg[0,1,1,2]=(0.765887852886)*x[2]
        arg[0,1,2,0]=(0.975541574352)*x[0]
        arg[0,1,2,1]=(-0.797851290751)*x[1]
        arg[0,1,2,2]=(0.628918775319)*x[2]
        arg[0,2,0,0]=(0.685635794312)*x[0]
        arg[0,2,0,1]=(0.10341799962)*x[1]
        arg[0,2,0,2]=(-0.964822756043)*x[2]
        arg[0,2,1,0]=(-0.56160368212)*x[0]
        arg[0,2,1,1]=(0.676344298102)*x[1]
        arg[0,2,1,2]=(-0.713924121843)*x[2]
        arg[0,2,2,0]=(-0.276655136263)*x[0]
        arg[0,2,2,1]=(0.336046973788)*x[1]
        arg[0,2,2,2]=(-0.68789392396)*x[2]
        arg[0,3,0,0]=(0.0172861311571)*x[0]
        arg[0,3,0,1]=(-0.301075956456)*x[1]
        arg[0,3,0,2]=(0.779442985415)*x[2]
        arg[0,3,1,0]=(-0.517629576558)*x[0]
        arg[0,3,1,1]=(0.584779586639)*x[1]
        arg[0,3,1,2]=(-0.53266435436)*x[2]
        arg[0,3,2,0]=(0.841533567102)*x[0]
        arg[0,3,2,1]=(0.0458746415489)*x[1]
        arg[0,3,2,2]=(0.921237870758)*x[2]
        arg[0,4,0,0]=(0.0548343238805)*x[0]
        arg[0,4,0,1]=(0.687022707412)*x[1]
        arg[0,4,0,2]=(-0.319803609795)*x[2]
        arg[0,4,1,0]=(0.409763007811)*x[0]
        arg[0,4,1,1]=(0.165501957435)*x[1]
        arg[0,4,1,2]=(0.116001692781)*x[2]
        arg[0,4,2,0]=(-0.515571394238)*x[0]
        arg[0,4,2,1]=(0.209467945147)*x[1]
        arg[0,4,2,2]=(-0.344827191247)*x[2]
        arg[1,0,0,0]=(0.57193838014)*x[0]
        arg[1,0,0,1]=(-0.0880683799076)*x[1]
        arg[1,0,0,2]=(0.956899617441)*x[2]
        arg[1,0,1,0]=(-0.783689636357)*x[0]
        arg[1,0,1,1]=(-0.25177506885)*x[1]
        arg[1,0,1,2]=(-0.97074584634)*x[2]
        arg[1,0,2,0]=(0.432543519806)*x[0]
        arg[1,0,2,1]=(0.481003021954)*x[1]
        arg[1,0,2,2]=(-0.0630751518268)*x[2]
        arg[1,1,0,0]=(-0.65152446796)*x[0]
        arg[1,1,0,1]=(-0.0323685084425)*x[1]
        arg[1,1,0,2]=(-0.508674033909)*x[2]
        arg[1,1,1,0]=(-0.533367818916)*x[0]
        arg[1,1,1,1]=(0.310738340288)*x[1]
        arg[1,1,1,2]=(0.694612234326)*x[2]
        arg[1,1,2,0]=(-0.622052473032)*x[0]
        arg[1,1,2,1]=(0.0498443793671)*x[1]
        arg[1,1,2,2]=(0.61023707512)*x[2]
        arg[1,2,0,0]=(0.0730267406859)*x[0]
        arg[1,2,0,1]=(0.146909334607)*x[1]
        arg[1,2,0,2]=(-0.641860284448)*x[2]
        arg[1,2,1,0]=(0.917976589737)*x[0]
        arg[1,2,1,1]=(0.50219672122)*x[1]
        arg[1,2,1,2]=(0.634559579812)*x[2]
        arg[1,2,2,0]=(0.0578772734534)*x[0]
        arg[1,2,2,1]=(0.288730973517)*x[1]
        arg[1,2,2,2]=(-0.0525978796154)*x[2]
        arg[1,3,0,0]=(-0.926152433388)*x[0]
        arg[1,3,0,1]=(0.0616647680855)*x[1]
        arg[1,3,0,2]=(-0.875889217846)*x[2]
        arg[1,3,1,0]=(-0.638931542845)*x[0]
        arg[1,3,1,1]=(0.708848122964)*x[1]
        arg[1,3,1,2]=(0.119066979792)*x[2]
        arg[1,3,2,0]=(0.853716218591)*x[0]
        arg[1,3,2,1]=(-0.92754322201)*x[1]
        arg[1,3,2,2]=(-0.671530626265)*x[2]
        arg[1,4,0,0]=(0.337424536231)*x[0]
        arg[1,4,0,1]=(0.335704451719)*x[1]
        arg[1,4,0,2]=(-0.484565969466)*x[2]
        arg[1,4,1,0]=(-0.855476192012)*x[0]
        arg[1,4,1,1]=(0.405674615553)*x[1]
        arg[1,4,1,2]=(0.728310771323)*x[2]
        arg[1,4,2,0]=(0.363651308265)*x[0]
        arg[1,4,2,1]=(0.174460594531)*x[1]
        arg[1,4,2,2]=(-0.0418244838617)*x[2]
        arg[2,0,0,0]=(-0.531341992511)*x[0]
        arg[2,0,0,1]=(0.584996796272)*x[1]
        arg[2,0,0,2]=(-0.752430968716)*x[2]
        arg[2,0,1,0]=(-0.341989849747)*x[0]
        arg[2,0,1,1]=(0.153572646953)*x[1]
        arg[2,0,1,2]=(-0.197130051737)*x[2]
        arg[2,0,2,0]=(-0.338082424082)*x[0]
        arg[2,0,2,1]=(0.000173657394772)*x[1]
        arg[2,0,2,2]=(0.365272907692)*x[2]
        arg[2,1,0,0]=(0.904304126564)*x[0]
        arg[2,1,0,1]=(0.161252368484)*x[1]
        arg[2,1,0,2]=(0.246854092422)*x[2]
        arg[2,1,1,0]=(-0.299880647529)*x[0]
        arg[2,1,1,1]=(-0.566917528608)*x[1]
        arg[2,1,1,2]=(0.243183337285)*x[2]
        arg[2,1,2,0]=(0.437406011474)*x[0]
        arg[2,1,2,1]=(0.727447394053)*x[1]
        arg[2,1,2,2]=(0.380752950664)*x[2]
        arg[2,2,0,0]=(0.172292846911)*x[0]
        arg[2,2,0,1]=(0.334201791643)*x[1]
        arg[2,2,0,2]=(0.739989926962)*x[2]
        arg[2,2,1,0]=(-0.0669843715042)*x[0]
        arg[2,2,1,1]=(-0.540497281635)*x[1]
        arg[2,2,1,2]=(-0.744217027088)*x[2]
        arg[2,2,2,0]=(-0.287295952259)*x[0]
        arg[2,2,2,1]=(-0.512411849183)*x[1]
        arg[2,2,2,2]=(0.953107417666)*x[2]
        arg[2,3,0,0]=(0.998168116695)*x[0]
        arg[2,3,0,1]=(0.960065646359)*x[1]
        arg[2,3,0,2]=(0.110048258832)*x[2]
        arg[2,3,1,0]=(-0.477271134724)*x[0]
        arg[2,3,1,1]=(0.707182612251)*x[1]
        arg[2,3,1,2]=(0.285500891755)*x[2]
        arg[2,3,2,0]=(-0.863497506661)*x[0]
        arg[2,3,2,1]=(-0.293917669879)*x[1]
        arg[2,3,2,2]=(-0.403384244295)*x[2]
        arg[2,4,0,0]=(0.848455277702)*x[0]
        arg[2,4,0,1]=(-0.530101455578)*x[1]
        arg[2,4,0,2]=(0.33887313048)*x[2]
        arg[2,4,1,0]=(-0.195313538124)*x[0]
        arg[2,4,1,1]=(-0.62754572008)*x[1]
        arg[2,4,1,2]=(-0.385132960582)*x[2]
        arg[2,4,2,0]=(0.240048012886)*x[0]
        arg[2,4,2,1]=(0.900766252969)*x[1]
        arg[2,4,2,2]=(0.669620533505)*x[2]
        arg[3,0,0,0]=(0.375766827301)*x[0]
        arg[3,0,0,1]=(0.705484960308)*x[1]
        arg[3,0,0,2]=(0.440931516034)*x[2]
        arg[3,0,1,0]=(-0.44724403177)*x[0]
        arg[3,0,1,1]=(-0.31558249626)*x[1]
        arg[3,0,1,2]=(-0.00419436365172)*x[2]
        arg[3,0,2,0]=(0.750599752032)*x[0]
        arg[3,0,2,1]=(0.367649951795)*x[1]
        arg[3,0,2,2]=(0.0488013073654)*x[2]
        arg[3,1,0,0]=(-0.992890068274)*x[0]
        arg[3,1,0,1]=(0.671447745511)*x[1]
        arg[3,1,0,2]=(0.85613331404)*x[2]
        arg[3,1,1,0]=(-0.46064764242)*x[0]
        arg[3,1,1,1]=(0.48138877715)*x[1]
        arg[3,1,1,2]=(0.396741761803)*x[2]
        arg[3,1,2,0]=(-0.879391967543)*x[0]
        arg[3,1,2,1]=(-0.44039462138)*x[1]
        arg[3,1,2,2]=(0.0330511573872)*x[2]
        arg[3,2,0,0]=(-0.367413701648)*x[0]
        arg[3,2,0,1]=(0.0359818324891)*x[1]
        arg[3,2,0,2]=(-0.307532667032)*x[2]
        arg[3,2,1,0]=(0.334663597166)*x[0]
        arg[3,2,1,1]=(0.541941978066)*x[1]
        arg[3,2,1,2]=(-0.609184079318)*x[2]
        arg[3,2,2,0]=(0.359349239826)*x[0]
        arg[3,2,2,1]=(0.0419272305685)*x[1]
        arg[3,2,2,2]=(0.557189794296)*x[2]
        arg[3,3,0,0]=(-0.85864165554)*x[0]
        arg[3,3,0,1]=(-0.185411404213)*x[1]
        arg[3,3,0,2]=(0.254294865253)*x[2]
        arg[3,3,1,0]=(0.870362177541)*x[0]
        arg[3,3,1,1]=(-0.439688612864)*x[1]
        arg[3,3,1,2]=(0.26006729357)*x[2]
        arg[3,3,2,0]=(-0.0724034754175)*x[0]
        arg[3,3,2,1]=(0.444871564246)*x[1]
        arg[3,3,2,2]=(0.485634530531)*x[2]
        arg[3,4,0,0]=(-0.744756961758)*x[0]
        arg[3,4,0,1]=(0.429761406102)*x[1]
        arg[3,4,0,2]=(-0.584963735834)*x[2]
        arg[3,4,1,0]=(0.684578379159)*x[0]
        arg[3,4,1,1]=(0.949460132601)*x[1]
        arg[3,4,1,2]=(-0.592179909559)*x[2]
        arg[3,4,2,0]=(0.707154437797)*x[0]
        arg[3,4,2,1]=(0.619200407063)*x[1]
        arg[3,4,2,2]=(-0.338547165)*x[2]
        ref=sqrt((19.2170638478))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=1.*sqrt(x[0])
        ref=sqrt(0.5)

      else:
        arg=Data(0,(),w)
        arg=1.*sqrt(x[0])
        ref=sqrt(0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=1.*sqrt(x[0])
        arg[1]=2.*sqrt(x[1])
        ref=sqrt(2.5)

      else:
        arg=Data(0,(3,),w)
        arg[0]=1.*sqrt(x[0])
        arg[1]=2.*sqrt(x[1])
        arg[2]=3.*sqrt(x[2])
        ref=sqrt(7.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=11.*sqrt(x[0])
        arg[0,1]=1.*sqrt(x[1])
        arg[1,0]=10.*sqrt(x[0])
        arg[1,1]=11.*sqrt(x[1])
        arg[2,0]=20.*sqrt(x[0])
        arg[2,1]=21.*sqrt(x[1])
        arg[3,0]=30.*sqrt(x[0])
        arg[3,1]=31.*sqrt(x[1])
        ref=sqrt(1522.5)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=11.*sqrt(x[0])
        arg[0,1]=1.*sqrt(x[1])
        arg[0,2]=2.*sqrt(x[2])
        arg[1,0]=10.*sqrt(x[0])
        arg[1,1]=11.*sqrt(x[1])
        arg[1,2]=12.*sqrt(x[2])
        arg[2,0]=20.*sqrt(x[0])
        arg[2,1]=21.*sqrt(x[1])
        arg[2,2]=22.*sqrt(x[2])
        arg[3,0]=30.*sqrt(x[0])
        arg[3,1]=31.*sqrt(x[1])
        arg[3,2]=32.*sqrt(x[2])
        ref=sqrt(2350.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.449174971953)*sqrt(x[0])
        arg[0,0,1]=(-0.0109398763289)*sqrt(x[1])
        arg[0,1,0]=(-0.202497187709)*sqrt(x[0])
        arg[0,1,1]=(-0.12970879334)*sqrt(x[1])
        arg[1,0,0]=(-0.138092481719)*sqrt(x[0])
        arg[1,0,1]=(-0.528752200917)*sqrt(x[1])
        arg[1,1,0]=(-0.605919441662)*sqrt(x[0])
        arg[1,1,1]=(0.215615032334)*sqrt(x[1])
        arg[2,0,0]=(-0.998734541972)*sqrt(x[0])
        arg[2,0,1]=(0.725811901251)*sqrt(x[1])
        arg[2,1,0]=(-0.966536503228)*sqrt(x[0])
        arg[2,1,1]=(-0.528692217355)*sqrt(x[1])
        arg[3,0,0]=(0.757633851466)*sqrt(x[0])
        arg[3,0,1]=(-0.524660157377)*sqrt(x[1])
        arg[3,1,0]=(0.983733431677)*sqrt(x[0])
        arg[3,1,1]=(0.061279109546)*sqrt(x[1])
        arg[4,0,0]=(0.85914215305)*sqrt(x[0])
        arg[4,0,1]=(0.941714045112)*sqrt(x[1])
        arg[4,1,0]=(0.172235529555)*sqrt(x[0])
        arg[4,1,1]=(-0.108381454437)*sqrt(x[1])
        arg[5,0,0]=(-0.736373697727)*sqrt(x[0])
        arg[5,0,1]=(-0.599337929679)*sqrt(x[1])
        arg[5,1,0]=(0.661072686392)*sqrt(x[0])
        arg[5,1,1]=(-0.55107327409)*sqrt(x[1])
        ref=sqrt(4.4196214907099591)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.69227064904)*sqrt(x[0])
        arg[0,0,1]=(-0.968336177418)*sqrt(x[1])
        arg[0,0,2]=(-0.634883146685)*sqrt(x[2])
        arg[0,1,0]=(-0.12640661422)*sqrt(x[0])
        arg[0,1,1]=(-0.637386589888)*sqrt(x[1])
        arg[0,1,2]=(0.26060859356)*sqrt(x[2])
        arg[1,0,0]=(-0.986864633297)*sqrt(x[0])
        arg[1,0,1]=(-0.441589142379)*sqrt(x[1])
        arg[1,0,2]=(-0.587865539582)*sqrt(x[2])
        arg[1,1,0]=(0.596052465031)*sqrt(x[0])
        arg[1,1,1]=(0.312732336652)*sqrt(x[1])
        arg[1,1,2]=(-0.514423945092)*sqrt(x[2])
        arg[2,0,0]=(-0.892391254794)*sqrt(x[0])
        arg[2,0,1]=(0.377920185756)*sqrt(x[1])
        arg[2,0,2]=(-0.120174597181)*sqrt(x[2])
        arg[2,1,0]=(-0.469951576468)*sqrt(x[0])
        arg[2,1,1]=(-0.788362249555)*sqrt(x[1])
        arg[2,1,2]=(0.745625354986)*sqrt(x[2])
        arg[3,0,0]=(0.542802498569)*sqrt(x[0])
        arg[3,0,1]=(-0.814541028706)*sqrt(x[1])
        arg[3,0,2]=(0.298410992196)*sqrt(x[2])
        arg[3,1,0]=(0.981190341206)*sqrt(x[0])
        arg[3,1,1]=(0.666421298608)*sqrt(x[1])
        arg[3,1,2]=(-0.369751722626)*sqrt(x[2])
        arg[4,0,0]=(-0.75379530597)*sqrt(x[0])
        arg[4,0,1]=(0.283357267139)*sqrt(x[1])
        arg[4,0,2]=(0.247787072861)*sqrt(x[2])
        arg[4,1,0]=(0.301766692533)*sqrt(x[0])
        arg[4,1,1]=(0.828183439224)*sqrt(x[1])
        arg[4,1,2]=(-0.580824060547)*sqrt(x[2])
        arg[5,0,0]=(0.637345610764)*sqrt(x[0])
        arg[5,0,1]=(-0.234409115997)*sqrt(x[1])
        arg[5,0,2]=(-0.192639300316)*sqrt(x[2])
        arg[5,1,0]=(-0.62609237162)*sqrt(x[0])
        arg[5,1,1]=(0.463404958552)*sqrt(x[1])
        arg[5,1,2]=(-0.547814448738)*sqrt(x[2])
        ref=sqrt(6.3571697792950923)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.232618585183)*sqrt(x[0])
        arg[0,0,0,1]=(0.39796117869)*sqrt(x[1])
        arg[0,0,1,0]=(-0.997336958262)*sqrt(x[0])
        arg[0,0,1,1]=(-0.351780915076)*sqrt(x[1])
        arg[0,0,2,0]=(-0.876764070136)*sqrt(x[0])
        arg[0,0,2,1]=(0.808730805817)*sqrt(x[1])
        arg[0,1,0,0]=(-0.197154744966)*sqrt(x[0])
        arg[0,1,0,1]=(0.416246096086)*sqrt(x[1])
        arg[0,1,1,0]=(0.708038457121)*sqrt(x[0])
        arg[0,1,1,1]=(-0.00954021503188)*sqrt(x[1])
        arg[0,1,2,0]=(-0.62630809425)*sqrt(x[0])
        arg[0,1,2,1]=(0.430228727912)*sqrt(x[1])
        arg[0,2,0,0]=(0.0306704747648)*sqrt(x[0])
        arg[0,2,0,1]=(-0.913877199453)*sqrt(x[1])
        arg[0,2,1,0]=(-0.697612800829)*sqrt(x[0])
        arg[0,2,1,1]=(-0.17996376822)*sqrt(x[1])
        arg[0,2,2,0]=(-0.304509578871)*sqrt(x[0])
        arg[0,2,2,1]=(-0.610556755811)*sqrt(x[1])
        arg[0,3,0,0]=(-0.452355972234)*sqrt(x[0])
        arg[0,3,0,1]=(-0.368921242518)*sqrt(x[1])
        arg[0,3,1,0]=(-0.478275554932)*sqrt(x[0])
        arg[0,3,1,1]=(0.257178549127)*sqrt(x[1])
        arg[0,3,2,0]=(0.530736487177)*sqrt(x[0])
        arg[0,3,2,1]=(-0.567126272463)*sqrt(x[1])
        arg[0,4,0,0]=(0.801519165938)*sqrt(x[0])
        arg[0,4,0,1]=(-0.509816703951)*sqrt(x[1])
        arg[0,4,1,0]=(-0.255412646934)*sqrt(x[0])
        arg[0,4,1,1]=(0.437540101896)*sqrt(x[1])
        arg[0,4,2,0]=(-0.815574969538)*sqrt(x[0])
        arg[0,4,2,1]=(-0.94691547137)*sqrt(x[1])
        arg[1,0,0,0]=(-0.732550722593)*sqrt(x[0])
        arg[1,0,0,1]=(0.515752381704)*sqrt(x[1])
        arg[1,0,1,0]=(-0.343590210899)*sqrt(x[0])
        arg[1,0,1,1]=(-0.0601907964915)*sqrt(x[1])
        arg[1,0,2,0]=(0.0199916154421)*sqrt(x[0])
        arg[1,0,2,1]=(-0.136927227821)*sqrt(x[1])
        arg[1,1,0,0]=(0.397994441702)*sqrt(x[0])
        arg[1,1,0,1]=(0.953873148948)*sqrt(x[1])
        arg[1,1,1,0]=(0.419416235967)*sqrt(x[0])
        arg[1,1,1,1]=(0.700998577193)*sqrt(x[1])
        arg[1,1,2,0]=(-0.497358799271)*sqrt(x[0])
        arg[1,1,2,1]=(0.0851768858379)*sqrt(x[1])
        arg[1,2,0,0]=(0.0936678875202)*sqrt(x[0])
        arg[1,2,0,1]=(0.869883786896)*sqrt(x[1])
        arg[1,2,1,0]=(0.582700123485)*sqrt(x[0])
        arg[1,2,1,1]=(-0.433381106794)*sqrt(x[1])
        arg[1,2,2,0]=(-0.527031777974)*sqrt(x[0])
        arg[1,2,2,1]=(0.105105137652)*sqrt(x[1])
        arg[1,3,0,0]=(-0.716750829134)*sqrt(x[0])
        arg[1,3,0,1]=(0.774519209008)*sqrt(x[1])
        arg[1,3,1,0]=(-0.568743372716)*sqrt(x[0])
        arg[1,3,1,1]=(0.794732483944)*sqrt(x[1])
        arg[1,3,2,0]=(0.246606002015)*sqrt(x[0])
        arg[1,3,2,1]=(-0.988869494994)*sqrt(x[1])
        arg[1,4,0,0]=(0.482379298083)*sqrt(x[0])
        arg[1,4,0,1]=(-0.386268387903)*sqrt(x[1])
        arg[1,4,1,0]=(0.137184889675)*sqrt(x[0])
        arg[1,4,1,1]=(-0.140520035321)*sqrt(x[1])
        arg[1,4,2,0]=(0.822755050415)*sqrt(x[0])
        arg[1,4,2,1]=(-0.815562139522)*sqrt(x[1])
        arg[2,0,0,0]=(-0.462891511962)*sqrt(x[0])
        arg[2,0,0,1]=(-0.122643411631)*sqrt(x[1])
        arg[2,0,1,0]=(-0.520861119962)*sqrt(x[0])
        arg[2,0,1,1]=(-0.881189618018)*sqrt(x[1])
        arg[2,0,2,0]=(-0.776157842774)*sqrt(x[0])
        arg[2,0,2,1]=(-0.12354053207)*sqrt(x[1])
        arg[2,1,0,0]=(0.395495230826)*sqrt(x[0])
        arg[2,1,0,1]=(-0.388106659423)*sqrt(x[1])
        arg[2,1,1,0]=(0.354250242834)*sqrt(x[0])
        arg[2,1,1,1]=(-0.666514210192)*sqrt(x[1])
        arg[2,1,2,0]=(0.951294655083)*sqrt(x[0])
        arg[2,1,2,1]=(0.074024416386)*sqrt(x[1])
        arg[2,2,0,0]=(0.335448485459)*sqrt(x[0])
        arg[2,2,0,1]=(-0.40988282528)*sqrt(x[1])
        arg[2,2,1,0]=(-0.805725968875)*sqrt(x[0])
        arg[2,2,1,1]=(-0.949883082118)*sqrt(x[1])
        arg[2,2,2,0]=(0.531549210683)*sqrt(x[0])
        arg[2,2,2,1]=(-0.398401016682)*sqrt(x[1])
        arg[2,3,0,0]=(-0.953963433205)*sqrt(x[0])
        arg[2,3,0,1]=(0.643431126406)*sqrt(x[1])
        arg[2,3,1,0]=(-0.167611998738)*sqrt(x[0])
        arg[2,3,1,1]=(0.226130056552)*sqrt(x[1])
        arg[2,3,2,0]=(0.0752687641131)*sqrt(x[0])
        arg[2,3,2,1]=(-0.115742756362)*sqrt(x[1])
        arg[2,4,0,0]=(0.579694491028)*sqrt(x[0])
        arg[2,4,0,1]=(-0.112005738299)*sqrt(x[1])
        arg[2,4,1,0]=(0.657291764224)*sqrt(x[0])
        arg[2,4,1,1]=(0.62671154177)*sqrt(x[1])
        arg[2,4,2,0]=(0.103695027944)*sqrt(x[0])
        arg[2,4,2,1]=(0.462828491544)*sqrt(x[1])
        arg[3,0,0,0]=(0.697692979998)*sqrt(x[0])
        arg[3,0,0,1]=(-0.123481859619)*sqrt(x[1])
        arg[3,0,1,0]=(-0.749745629459)*sqrt(x[0])
        arg[3,0,1,1]=(-0.541969524069)*sqrt(x[1])
        arg[3,0,2,0]=(0.819484470759)*sqrt(x[0])
        arg[3,0,2,1]=(-0.860592326469)*sqrt(x[1])
        arg[3,1,0,0]=(-0.716566084771)*sqrt(x[0])
        arg[3,1,0,1]=(-0.949235434827)*sqrt(x[1])
        arg[3,1,1,0]=(-0.826699498174)*sqrt(x[0])
        arg[3,1,1,1]=(-0.138511521583)*sqrt(x[1])
        arg[3,1,2,0]=(-0.951682890904)*sqrt(x[0])
        arg[3,1,2,1]=(0.413293316925)*sqrt(x[1])
        arg[3,2,0,0]=(0.909516836775)*sqrt(x[0])
        arg[3,2,0,1]=(-0.919989721277)*sqrt(x[1])
        arg[3,2,1,0]=(0.0994860369337)*sqrt(x[0])
        arg[3,2,1,1]=(-0.933647246623)*sqrt(x[1])
        arg[3,2,2,0]=(-0.759215183015)*sqrt(x[0])
        arg[3,2,2,1]=(0.0975793309286)*sqrt(x[1])
        arg[3,3,0,0]=(-0.130256739381)*sqrt(x[0])
        arg[3,3,0,1]=(-0.582280862311)*sqrt(x[1])
        arg[3,3,1,0]=(0.206970526192)*sqrt(x[0])
        arg[3,3,1,1]=(-0.8678322258)*sqrt(x[1])
        arg[3,3,2,0]=(0.133004501279)*sqrt(x[0])
        arg[3,3,2,1]=(0.802921710935)*sqrt(x[1])
        arg[3,4,0,0]=(-0.255737792764)*sqrt(x[0])
        arg[3,4,0,1]=(-0.34168114937)*sqrt(x[1])
        arg[3,4,1,0]=(-0.859309090399)*sqrt(x[0])
        arg[3,4,1,1]=(0.245043986435)*sqrt(x[1])
        arg[3,4,2,0]=(0.893062018695)*sqrt(x[0])
        arg[3,4,2,1]=(0.709422742588)*sqrt(x[1])
        ref=sqrt(20.5933920543)
      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.0312828390439)*sqrt(x[0])
        arg[0,0,0,1]=(-0.524970416212)*sqrt(x[1])
        arg[0,0,0,2]=(0.561865217554)*sqrt(x[2])
        arg[0,0,1,0]=(0.692457187384)*sqrt(x[0])
        arg[0,0,1,1]=(0.946967182157)*sqrt(x[1])
        arg[0,0,1,2]=(-0.863842279464)*sqrt(x[2])
        arg[0,0,2,0]=(0.993922921598)*sqrt(x[0])
        arg[0,0,2,1]=(0.322812768679)*sqrt(x[1])
        arg[0,0,2,2]=(0.901876132204)*sqrt(x[2])
        arg[0,1,0,0]=(0.967569979365)*sqrt(x[0])
        arg[0,1,0,1]=(0.840979131355)*sqrt(x[1])
        arg[0,1,0,2]=(0.0494811460856)*sqrt(x[2])
        arg[0,1,1,0]=(0.315178456102)*sqrt(x[0])
        arg[0,1,1,1]=(0.449848313024)*sqrt(x[1])
        arg[0,1,1,2]=(0.765887852886)*sqrt(x[2])
        arg[0,1,2,0]=(0.975541574352)*sqrt(x[0])
        arg[0,1,2,1]=(-0.797851290751)*sqrt(x[1])
        arg[0,1,2,2]=(0.628918775319)*sqrt(x[2])
        arg[0,2,0,0]=(0.685635794312)*sqrt(x[0])
        arg[0,2,0,1]=(0.10341799962)*sqrt(x[1])
        arg[0,2,0,2]=(-0.964822756043)*sqrt(x[2])
        arg[0,2,1,0]=(-0.56160368212)*sqrt(x[0])
        arg[0,2,1,1]=(0.676344298102)*sqrt(x[1])
        arg[0,2,1,2]=(-0.713924121843)*sqrt(x[2])
        arg[0,2,2,0]=(-0.276655136263)*sqrt(x[0])
        arg[0,2,2,1]=(0.336046973788)*sqrt(x[1])
        arg[0,2,2,2]=(-0.68789392396)*sqrt(x[2])
        arg[0,3,0,0]=(0.0172861311571)*sqrt(x[0])
        arg[0,3,0,1]=(-0.301075956456)*sqrt(x[1])
        arg[0,3,0,2]=(0.779442985415)*sqrt(x[2])
        arg[0,3,1,0]=(-0.517629576558)*sqrt(x[0])
        arg[0,3,1,1]=(0.584779586639)*sqrt(x[1])
        arg[0,3,1,2]=(-0.53266435436)*sqrt(x[2])
        arg[0,3,2,0]=(0.841533567102)*sqrt(x[0])
        arg[0,3,2,1]=(0.0458746415489)*sqrt(x[1])
        arg[0,3,2,2]=(0.921237870758)*sqrt(x[2])
        arg[0,4,0,0]=(0.0548343238805)*sqrt(x[0])
        arg[0,4,0,1]=(0.687022707412)*sqrt(x[1])
        arg[0,4,0,2]=(-0.319803609795)*sqrt(x[2])
        arg[0,4,1,0]=(0.409763007811)*sqrt(x[0])
        arg[0,4,1,1]=(0.165501957435)*sqrt(x[1])
        arg[0,4,1,2]=(0.116001692781)*sqrt(x[2])
        arg[0,4,2,0]=(-0.515571394238)*sqrt(x[0])
        arg[0,4,2,1]=(0.209467945147)*sqrt(x[1])
        arg[0,4,2,2]=(-0.344827191247)*sqrt(x[2])
        arg[1,0,0,0]=(0.57193838014)*sqrt(x[0])
        arg[1,0,0,1]=(-0.0880683799076)*sqrt(x[1])
        arg[1,0,0,2]=(0.956899617441)*sqrt(x[2])
        arg[1,0,1,0]=(-0.783689636357)*sqrt(x[0])
        arg[1,0,1,1]=(-0.25177506885)*sqrt(x[1])
        arg[1,0,1,2]=(-0.97074584634)*sqrt(x[2])
        arg[1,0,2,0]=(0.432543519806)*sqrt(x[0])
        arg[1,0,2,1]=(0.481003021954)*sqrt(x[1])
        arg[1,0,2,2]=(-0.0630751518268)*sqrt(x[2])
        arg[1,1,0,0]=(-0.65152446796)*sqrt(x[0])
        arg[1,1,0,1]=(-0.0323685084425)*sqrt(x[1])
        arg[1,1,0,2]=(-0.508674033909)*sqrt(x[2])
        arg[1,1,1,0]=(-0.533367818916)*sqrt(x[0])
        arg[1,1,1,1]=(0.310738340288)*sqrt(x[1])
        arg[1,1,1,2]=(0.694612234326)*sqrt(x[2])
        arg[1,1,2,0]=(-0.622052473032)*sqrt(x[0])
        arg[1,1,2,1]=(0.0498443793671)*sqrt(x[1])
        arg[1,1,2,2]=(0.61023707512)*sqrt(x[2])
        arg[1,2,0,0]=(0.0730267406859)*sqrt(x[0])
        arg[1,2,0,1]=(0.146909334607)*sqrt(x[1])
        arg[1,2,0,2]=(-0.641860284448)*sqrt(x[2])
        arg[1,2,1,0]=(0.917976589737)*sqrt(x[0])
        arg[1,2,1,1]=(0.50219672122)*sqrt(x[1])
        arg[1,2,1,2]=(0.634559579812)*sqrt(x[2])
        arg[1,2,2,0]=(0.0578772734534)*sqrt(x[0])
        arg[1,2,2,1]=(0.288730973517)*sqrt(x[1])
        arg[1,2,2,2]=(-0.0525978796154)*sqrt(x[2])
        arg[1,3,0,0]=(-0.926152433388)*sqrt(x[0])
        arg[1,3,0,1]=(0.0616647680855)*sqrt(x[1])
        arg[1,3,0,2]=(-0.875889217846)*sqrt(x[2])
        arg[1,3,1,0]=(-0.638931542845)*sqrt(x[0])
        arg[1,3,1,1]=(0.708848122964)*sqrt(x[1])
        arg[1,3,1,2]=(0.119066979792)*sqrt(x[2])
        arg[1,3,2,0]=(0.853716218591)*sqrt(x[0])
        arg[1,3,2,1]=(-0.92754322201)*sqrt(x[1])
        arg[1,3,2,2]=(-0.671530626265)*sqrt(x[2])
        arg[1,4,0,0]=(0.337424536231)*sqrt(x[0])
        arg[1,4,0,1]=(0.335704451719)*sqrt(x[1])
        arg[1,4,0,2]=(-0.484565969466)*sqrt(x[2])
        arg[1,4,1,0]=(-0.855476192012)*sqrt(x[0])
        arg[1,4,1,1]=(0.405674615553)*sqrt(x[1])
        arg[1,4,1,2]=(0.728310771323)*sqrt(x[2])
        arg[1,4,2,0]=(0.363651308265)*sqrt(x[0])
        arg[1,4,2,1]=(0.174460594531)*sqrt(x[1])
        arg[1,4,2,2]=(-0.0418244838617)*sqrt(x[2])
        arg[2,0,0,0]=(-0.531341992511)*sqrt(x[0])
        arg[2,0,0,1]=(0.584996796272)*sqrt(x[1])
        arg[2,0,0,2]=(-0.752430968716)*sqrt(x[2])
        arg[2,0,1,0]=(-0.341989849747)*sqrt(x[0])
        arg[2,0,1,1]=(0.153572646953)*sqrt(x[1])
        arg[2,0,1,2]=(-0.197130051737)*sqrt(x[2])
        arg[2,0,2,0]=(-0.338082424082)*sqrt(x[0])
        arg[2,0,2,1]=(0.000173657394772)*sqrt(x[1])
        arg[2,0,2,2]=(0.365272907692)*sqrt(x[2])
        arg[2,1,0,0]=(0.904304126564)*sqrt(x[0])
        arg[2,1,0,1]=(0.161252368484)*sqrt(x[1])
        arg[2,1,0,2]=(0.246854092422)*sqrt(x[2])
        arg[2,1,1,0]=(-0.299880647529)*sqrt(x[0])
        arg[2,1,1,1]=(-0.566917528608)*sqrt(x[1])
        arg[2,1,1,2]=(0.243183337285)*sqrt(x[2])
        arg[2,1,2,0]=(0.437406011474)*sqrt(x[0])
        arg[2,1,2,1]=(0.727447394053)*sqrt(x[1])
        arg[2,1,2,2]=(0.380752950664)*sqrt(x[2])
        arg[2,2,0,0]=(0.172292846911)*sqrt(x[0])
        arg[2,2,0,1]=(0.334201791643)*sqrt(x[1])
        arg[2,2,0,2]=(0.739989926962)*sqrt(x[2])
        arg[2,2,1,0]=(-0.0669843715042)*sqrt(x[0])
        arg[2,2,1,1]=(-0.540497281635)*sqrt(x[1])
        arg[2,2,1,2]=(-0.744217027088)*sqrt(x[2])
        arg[2,2,2,0]=(-0.287295952259)*sqrt(x[0])
        arg[2,2,2,1]=(-0.512411849183)*sqrt(x[1])
        arg[2,2,2,2]=(0.953107417666)*sqrt(x[2])
        arg[2,3,0,0]=(0.998168116695)*sqrt(x[0])
        arg[2,3,0,1]=(0.960065646359)*sqrt(x[1])
        arg[2,3,0,2]=(0.110048258832)*sqrt(x[2])
        arg[2,3,1,0]=(-0.477271134724)*sqrt(x[0])
        arg[2,3,1,1]=(0.707182612251)*sqrt(x[1])
        arg[2,3,1,2]=(0.285500891755)*sqrt(x[2])
        arg[2,3,2,0]=(-0.863497506661)*sqrt(x[0])
        arg[2,3,2,1]=(-0.293917669879)*sqrt(x[1])
        arg[2,3,2,2]=(-0.403384244295)*sqrt(x[2])
        arg[2,4,0,0]=(0.848455277702)*sqrt(x[0])
        arg[2,4,0,1]=(-0.530101455578)*sqrt(x[1])
        arg[2,4,0,2]=(0.33887313048)*sqrt(x[2])
        arg[2,4,1,0]=(-0.195313538124)*sqrt(x[0])
        arg[2,4,1,1]=(-0.62754572008)*sqrt(x[1])
        arg[2,4,1,2]=(-0.385132960582)*sqrt(x[2])
        arg[2,4,2,0]=(0.240048012886)*sqrt(x[0])
        arg[2,4,2,1]=(0.900766252969)*sqrt(x[1])
        arg[2,4,2,2]=(0.669620533505)*sqrt(x[2])
        arg[3,0,0,0]=(0.375766827301)*sqrt(x[0])
        arg[3,0,0,1]=(0.705484960308)*sqrt(x[1])
        arg[3,0,0,2]=(0.440931516034)*sqrt(x[2])
        arg[3,0,1,0]=(-0.44724403177)*sqrt(x[0])
        arg[3,0,1,1]=(-0.31558249626)*sqrt(x[1])
        arg[3,0,1,2]=(-0.00419436365172)*sqrt(x[2])
        arg[3,0,2,0]=(0.750599752032)*sqrt(x[0])
        arg[3,0,2,1]=(0.367649951795)*sqrt(x[1])
        arg[3,0,2,2]=(0.0488013073654)*sqrt(x[2])
        arg[3,1,0,0]=(-0.992890068274)*sqrt(x[0])
        arg[3,1,0,1]=(0.671447745511)*sqrt(x[1])
        arg[3,1,0,2]=(0.85613331404)*sqrt(x[2])
        arg[3,1,1,0]=(-0.46064764242)*sqrt(x[0])
        arg[3,1,1,1]=(0.48138877715)*sqrt(x[1])
        arg[3,1,1,2]=(0.396741761803)*sqrt(x[2])
        arg[3,1,2,0]=(-0.879391967543)*sqrt(x[0])
        arg[3,1,2,1]=(-0.44039462138)*sqrt(x[1])
        arg[3,1,2,2]=(0.0330511573872)*sqrt(x[2])
        arg[3,2,0,0]=(-0.367413701648)*sqrt(x[0])
        arg[3,2,0,1]=(0.0359818324891)*sqrt(x[1])
        arg[3,2,0,2]=(-0.307532667032)*sqrt(x[2])
        arg[3,2,1,0]=(0.334663597166)*sqrt(x[0])
        arg[3,2,1,1]=(0.541941978066)*sqrt(x[1])
        arg[3,2,1,2]=(-0.609184079318)*sqrt(x[2])
        arg[3,2,2,0]=(0.359349239826)*sqrt(x[0])
        arg[3,2,2,1]=(0.0419272305685)*sqrt(x[1])
        arg[3,2,2,2]=(0.557189794296)*sqrt(x[2])
        arg[3,3,0,0]=(-0.85864165554)*sqrt(x[0])
        arg[3,3,0,1]=(-0.185411404213)*sqrt(x[1])
        arg[3,3,0,2]=(0.254294865253)*sqrt(x[2])
        arg[3,3,1,0]=(0.870362177541)*sqrt(x[0])
        arg[3,3,1,1]=(-0.439688612864)*sqrt(x[1])
        arg[3,3,1,2]=(0.26006729357)*sqrt(x[2])
        arg[3,3,2,0]=(-0.0724034754175)*sqrt(x[0])
        arg[3,3,2,1]=(0.444871564246)*sqrt(x[1])
        arg[3,3,2,2]=(0.485634530531)*sqrt(x[2])
        arg[3,4,0,0]=(-0.744756961758)*sqrt(x[0])
        arg[3,4,0,1]=(0.429761406102)*sqrt(x[1])
        arg[3,4,0,2]=(-0.584963735834)*sqrt(x[2])
        arg[3,4,1,0]=(0.684578379159)*sqrt(x[0])
        arg[3,4,1,1]=(0.949460132601)*sqrt(x[1])
        arg[3,4,1,2]=(-0.592179909559)*sqrt(x[2])
        arg[3,4,2,0]=(0.707154437797)*sqrt(x[0])
        arg[3,4,2,1]=(0.619200407063)*sqrt(x[1])
        arg[3,4,2,2]=(-0.338547165)*sqrt(x[2])
        ref=sqrt(28.8255957718)
      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank0(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.245574919477)*x[0]
        ref=sqrt((0.0603070410759)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(),w)
        arg=(0.757324521515)*x[0]
        ref=sqrt((0.573540430888)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank1(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.723421565407)*x[0]
        arg[1]=(-0.460477393103)*x[1]
        ref=sqrt((0.735378190855)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.88528497163)*x[0]
        arg[1]=(-0.65510214636)*x[1]
        arg[2]=(0.399538866363)*x[2]
        ref=sqrt((1.37251960889)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank2(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.955620993904)*x[0]
        arg[0,1]=(-0.0987865813703)*x[1]
        arg[1,0]=(0.0288267231531)*x[0]
        arg[1,1]=(0.655440599879)*x[1]
        arg[2,0]=(0.685627284533)*x[0]
        arg[2,1]=(-0.989832824892)*x[1]
        arg[3,0]=(0.292184093194)*x[0]
        arg[3,1]=(0.149553857773)*x[1]
        ref=sqrt((2.91099532781)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.325908541533)*x[0]
        arg[0,1]=(-0.992480479749)*x[1]
        arg[0,2]=(0.660360271799)*x[2]
        arg[1,0]=(0.173485908581)*x[0]
        arg[1,1]=(-0.328755199781)*x[1]
        arg[1,2]=(-0.943354674948)*x[2]
        arg[2,0]=(0.680713222646)*x[0]
        arg[2,1]=(-0.765971835693)*x[1]
        arg[2,2]=(0.0413284847528)*x[2]
        arg[3,0]=(0.990074004708)*x[0]
        arg[3,1]=(0.941801786766)*x[1]
        arg[3,2]=(0.886926192201)*x[2]
        ref=sqrt((6.26107155228)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank3(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.0781611551598)*x[0]
        arg[0,0,1]=(0.291016249575)*x[1]
        arg[0,1,0]=(-0.107555233086)*x[0]
        arg[0,1,1]=(-0.559067108546)*x[1]
        arg[1,0,0]=(-0.0818406701266)*x[0]
        arg[1,0,1]=(-0.594866806483)*x[1]
        arg[1,1,0]=(-0.725814803863)*x[0]
        arg[1,1,1]=(0.59128101992)*x[1]
        arg[2,0,0]=(-0.15381555291)*x[0]
        arg[2,0,1]=(-0.679882948503)*x[1]
        arg[2,1,0]=(-0.58437193917)*x[0]
        arg[2,1,1]=(0.136304615849)*x[1]
        arg[3,0,0]=(0.0671365410096)*x[0]
        arg[3,0,1]=(-0.645687212187)*x[1]
        arg[3,1,0]=(-0.642492412392)*x[0]
        arg[3,1,1]=(-0.125760054735)*x[1]
        arg[4,0,0]=(0.731110824794)*x[0]
        arg[4,0,1]=(0.491668422979)*x[1]
        arg[4,1,0]=(-0.775841478292)*x[0]
        arg[4,1,1]=(0.728265567974)*x[1]
        arg[5,0,0]=(0.84511832373)*x[0]
        arg[5,0,1]=(-0.513796801068)*x[1]
        arg[5,1,0]=(0.113072243554)*x[0]
        arg[5,1,1]=(0.246630838744)*x[1]
        ref=sqrt((6.30829536252)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.369748116859)*x[0]
        arg[0,0,1]=(-0.758056560031)*x[1]
        arg[0,0,2]=(-0.873984709951)*x[2]
        arg[0,1,0]=(0.311680165784)*x[0]
        arg[0,1,1]=(0.374400673651)*x[1]
        arg[0,1,2]=(0.712484217076)*x[2]
        arg[1,0,0]=(0.829379714484)*x[0]
        arg[1,0,1]=(-0.0551589596149)*x[1]
        arg[1,0,2]=(0.965672208426)*x[2]
        arg[1,1,0]=(-0.205044281547)*x[0]
        arg[1,1,1]=(0.238197452756)*x[1]
        arg[1,1,2]=(-0.33456139292)*x[2]
        arg[2,0,0]=(0.649928288926)*x[0]
        arg[2,0,1]=(-0.661384953389)*x[1]
        arg[2,0,2]=(-0.253241222975)*x[2]
        arg[2,1,0]=(-0.491716575992)*x[0]
        arg[2,1,1]=(-0.970872527468)*x[1]
        arg[2,1,2]=(0.222410198921)*x[2]
        arg[3,0,0]=(0.205752630262)*x[0]
        arg[3,0,1]=(0.864804362697)*x[1]
        arg[3,0,2]=(-0.417975564033)*x[2]
        arg[3,1,0]=(0.586425694033)*x[0]
        arg[3,1,1]=(0.952661122184)*x[1]
        arg[3,1,2]=(0.608680080453)*x[2]
        arg[4,0,0]=(0.625968903369)*x[0]
        arg[4,0,1]=(-0.573909405003)*x[1]
        arg[4,0,2]=(-0.762256394595)*x[2]
        arg[4,1,0]=(0.0710742394418)*x[0]
        arg[4,1,1]=(0.583378040574)*x[1]
        arg[4,1,2]=(0.719032893115)*x[2]
        arg[5,0,0]=(0.032173368884)*x[0]
        arg[5,0,1]=(-0.434042549492)*x[1]
        arg[5,0,2]=(0.363504765447)*x[2]
        arg[5,1,0]=(0.598817469198)*x[0]
        arg[5,1,1]=(-0.163967008775)*x[1]
        arg[5,1,2]=(0.546778730604)*x[2]
        ref=sqrt((11.9696343123)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank4(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.273446520069)*x[0]
        arg[0,0,0,1]=(-0.913305910831)*x[1]
        arg[0,0,1,0]=(0.0745566747537)*x[0]
        arg[0,0,1,1]=(0.98803601919)*x[1]
        arg[0,0,2,0]=(-0.244120818875)*x[0]
        arg[0,0,2,1]=(0.247509644998)*x[1]
        arg[0,1,0,0]=(-0.548756417777)*x[0]
        arg[0,1,0,1]=(-0.354587911923)*x[1]
        arg[0,1,1,0]=(0.104268198867)*x[0]
        arg[0,1,1,1]=(-0.541700877072)*x[1]
        arg[0,1,2,0]=(-0.25900060689)*x[0]
        arg[0,1,2,1]=(-0.859660175231)*x[1]
        arg[0,2,0,0]=(0.195235560321)*x[0]
        arg[0,2,0,1]=(0.175518738589)*x[1]
        arg[0,2,1,0]=(-0.0638854232272)*x[0]
        arg[0,2,1,1]=(-0.586161016541)*x[1]
        arg[0,2,2,0]=(0.580258892247)*x[0]
        arg[0,2,2,1]=(-0.931927145435)*x[1]
        arg[0,3,0,0]=(-0.408209600298)*x[0]
        arg[0,3,0,1]=(-0.0344882667014)*x[1]
        arg[0,3,1,0]=(-0.131763534163)*x[0]
        arg[0,3,1,1]=(-0.787653739965)*x[1]
        arg[0,3,2,0]=(0.0910808104711)*x[0]
        arg[0,3,2,1]=(-0.280409023611)*x[1]
        arg[0,4,0,0]=(0.97745754012)*x[0]
        arg[0,4,0,1]=(-0.59829020936)*x[1]
        arg[0,4,1,0]=(0.890520260543)*x[0]
        arg[0,4,1,1]=(-0.600760090231)*x[1]
        arg[0,4,2,0]=(-0.897992297974)*x[0]
        arg[0,4,2,1]=(-0.841169898923)*x[1]
        arg[1,0,0,0]=(-0.249868816409)*x[0]
        arg[1,0,0,1]=(0.620375082228)*x[1]
        arg[1,0,1,0]=(-0.660480789306)*x[0]
        arg[1,0,1,1]=(-0.73638571806)*x[1]
        arg[1,0,2,0]=(0.339987316643)*x[0]
        arg[1,0,2,1]=(0.541112529894)*x[1]
        arg[1,1,0,0]=(-0.468808186705)*x[0]
        arg[1,1,0,1]=(-0.32919679792)*x[1]
        arg[1,1,1,0]=(0.917292803419)*x[0]
        arg[1,1,1,1]=(-0.834265058005)*x[1]
        arg[1,1,2,0]=(-0.247536849264)*x[0]
        arg[1,1,2,1]=(-0.197503469238)*x[1]
        arg[1,2,0,0]=(0.897591919909)*x[0]
        arg[1,2,0,1]=(-0.807446231234)*x[1]
        arg[1,2,1,0]=(-0.369878499382)*x[0]
        arg[1,2,1,1]=(0.985678692179)*x[1]
        arg[1,2,2,0]=(-0.709976427525)*x[0]
        arg[1,2,2,1]=(-0.368744647016)*x[1]
        arg[1,3,0,0]=(0.299630726462)*x[0]
        arg[1,3,0,1]=(-0.445295757899)*x[1]
        arg[1,3,1,0]=(-0.922386577254)*x[0]
        arg[1,3,1,1]=(0.234794853697)*x[1]
        arg[1,3,2,0]=(0.953377720197)*x[0]
        arg[1,3,2,1]=(0.409778183998)*x[1]
        arg[1,4,0,0]=(0.271967945488)*x[0]
        arg[1,4,0,1]=(-0.578629001202)*x[1]
        arg[1,4,1,0]=(0.210755226769)*x[0]
        arg[1,4,1,1]=(-0.0902751419945)*x[1]
        arg[1,4,2,0]=(0.70033387381)*x[0]
        arg[1,4,2,1]=(0.305733565661)*x[1]
        arg[2,0,0,0]=(-0.662483167298)*x[0]
        arg[2,0,0,1]=(0.585252048652)*x[1]
        arg[2,0,1,0]=(-0.398813785959)*x[0]
        arg[2,0,1,1]=(-0.797438697186)*x[1]
        arg[2,0,2,0]=(-0.508308971009)*x[0]
        arg[2,0,2,1]=(0.302249407524)*x[1]
        arg[2,1,0,0]=(0.208644491879)*x[0]
        arg[2,1,0,1]=(-0.604749055374)*x[1]
        arg[2,1,1,0]=(0.641654284594)*x[0]
        arg[2,1,1,1]=(0.456898593356)*x[1]
        arg[2,1,2,0]=(-0.398043867778)*x[0]
        arg[2,1,2,1]=(-0.0712344657587)*x[1]
        arg[2,2,0,0]=(0.0967860954865)*x[0]
        arg[2,2,0,1]=(0.520449905952)*x[1]
        arg[2,2,1,0]=(0.770198029595)*x[0]
        arg[2,2,1,1]=(-0.594004671621)*x[1]
        arg[2,2,2,0]=(-0.744571452885)*x[0]
        arg[2,2,2,1]=(0.544447367825)*x[1]
        arg[2,3,0,0]=(-0.137087966968)*x[0]
        arg[2,3,0,1]=(0.120672667497)*x[1]
        arg[2,3,1,0]=(0.204800088057)*x[0]
        arg[2,3,1,1]=(0.626526346076)*x[1]
        arg[2,3,2,0]=(-0.696480393227)*x[0]
        arg[2,3,2,1]=(0.188533741996)*x[1]
        arg[2,4,0,0]=(-0.403523821067)*x[0]
        arg[2,4,0,1]=(-0.428048989483)*x[1]
        arg[2,4,1,0]=(-0.244186366584)*x[0]
        arg[2,4,1,1]=(0.00866444909003)*x[1]
        arg[2,4,2,0]=(-0.445991308853)*x[0]
        arg[2,4,2,1]=(-0.899951068935)*x[1]
        arg[3,0,0,0]=(0.609340085418)*x[0]
        arg[3,0,0,1]=(0.878750391425)*x[1]
        arg[3,0,1,0]=(0.258064654464)*x[0]
        arg[3,0,1,1]=(-0.482402612985)*x[1]
        arg[3,0,2,0]=(0.943732283389)*x[0]
        arg[3,0,2,1]=(0.65514211843)*x[1]
        arg[3,1,0,0]=(-0.894551979619)*x[0]
        arg[3,1,0,1]=(0.220116541042)*x[1]
        arg[3,1,1,0]=(0.386887699577)*x[0]
        arg[3,1,1,1]=(-0.422560075108)*x[1]
        arg[3,1,2,0]=(0.00387273783493)*x[0]
        arg[3,1,2,1]=(0.465673505613)*x[1]
        arg[3,2,0,0]=(0.987383428982)*x[0]
        arg[3,2,0,1]=(0.376320055964)*x[1]
        arg[3,2,1,0]=(-0.463778689128)*x[0]
        arg[3,2,1,1]=(0.179816566227)*x[1]
        arg[3,2,2,0]=(0.961522856801)*x[0]
        arg[3,2,2,1]=(-0.257779627946)*x[1]
        arg[3,3,0,0]=(0.748886531458)*x[0]
        arg[3,3,0,1]=(-0.257342282566)*x[1]
        arg[3,3,1,0]=(0.377494024401)*x[0]
        arg[3,3,1,1]=(-0.334588346017)*x[1]
        arg[3,3,2,0]=(0.502495149189)*x[0]
        arg[3,3,2,1]=(0.534612429702)*x[1]
        arg[3,4,0,0]=(-0.308551337355)*x[0]
        arg[3,4,0,1]=(-0.471825826745)*x[1]
        arg[3,4,1,0]=(-0.262606531584)*x[0]
        arg[3,4,1,1]=(0.766089616367)*x[1]
        arg[3,4,2,0]=(-0.136526755642)*x[0]
        arg[3,4,2,1]=(0.675111459363)*x[1]
        ref=sqrt((37.3453550914)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.532946231146)*x[0]
        arg[0,0,0,1]=(0.269364089513)*x[1]
        arg[0,0,0,2]=(-0.207412457081)*x[2]
        arg[0,0,1,0]=(-0.843104704858)*x[0]
        arg[0,0,1,1]=(0.0416216508473)*x[1]
        arg[0,0,1,2]=(-0.836074693662)*x[2]
        arg[0,0,2,0]=(0.943609268731)*x[0]
        arg[0,0,2,1]=(0.0154543737816)*x[1]
        arg[0,0,2,2]=(0.0726268788381)*x[2]
        arg[0,1,0,0]=(0.108422740078)*x[0]
        arg[0,1,0,1]=(-0.296667916638)*x[1]
        arg[0,1,0,2]=(-0.769732600535)*x[2]
        arg[0,1,1,0]=(-0.428575493834)*x[0]
        arg[0,1,1,1]=(0.421245456722)*x[1]
        arg[0,1,1,2]=(-0.588277652277)*x[2]
        arg[0,1,2,0]=(0.145294576795)*x[0]
        arg[0,1,2,1]=(0.323206623794)*x[1]
        arg[0,1,2,2]=(0.788115602892)*x[2]
        arg[0,2,0,0]=(-0.227877282292)*x[0]
        arg[0,2,0,1]=(-0.630647460719)*x[1]
        arg[0,2,0,2]=(0.58754882135)*x[2]
        arg[0,2,1,0]=(0.347191113403)*x[0]
        arg[0,2,1,1]=(0.464093634725)*x[1]
        arg[0,2,1,2]=(-0.0412800774497)*x[2]
        arg[0,2,2,0]=(0.223364317185)*x[0]
        arg[0,2,2,1]=(0.257201130157)*x[1]
        arg[0,2,2,2]=(0.063203467463)*x[2]
        arg[0,3,0,0]=(-0.723240451643)*x[0]
        arg[0,3,0,1]=(-0.862468295097)*x[1]
        arg[0,3,0,2]=(-0.149283247587)*x[2]
        arg[0,3,1,0]=(0.15680097839)*x[0]
        arg[0,3,1,1]=(0.421563637547)*x[1]
        arg[0,3,1,2]=(0.111549188549)*x[2]
        arg[0,3,2,0]=(-0.272783329363)*x[0]
        arg[0,3,2,1]=(-0.420352789853)*x[1]
        arg[0,3,2,2]=(0.570865117722)*x[2]
        arg[0,4,0,0]=(-0.321910078414)*x[0]
        arg[0,4,0,1]=(0.988695599439)*x[1]
        arg[0,4,0,2]=(0.920200893398)*x[2]
        arg[0,4,1,0]=(0.0260910072651)*x[0]
        arg[0,4,1,1]=(0.460012578184)*x[1]
        arg[0,4,1,2]=(0.848099524112)*x[2]
        arg[0,4,2,0]=(0.242157803251)*x[0]
        arg[0,4,2,1]=(0.394528777004)*x[1]
        arg[0,4,2,2]=(0.562996837311)*x[2]
        arg[1,0,0,0]=(0.459886225958)*x[0]
        arg[1,0,0,1]=(-0.721868942003)*x[1]
        arg[1,0,0,2]=(0.432203082994)*x[2]
        arg[1,0,1,0]=(0.409831045482)*x[0]
        arg[1,0,1,1]=(-0.481677513473)*x[1]
        arg[1,0,1,2]=(0.439387853437)*x[2]
        arg[1,0,2,0]=(0.261583198434)*x[0]
        arg[1,0,2,1]=(0.290993423577)*x[1]
        arg[1,0,2,2]=(0.477993114134)*x[2]
        arg[1,1,0,0]=(0.586344598248)*x[0]
        arg[1,1,0,1]=(-0.105390792831)*x[1]
        arg[1,1,0,2]=(0.335990751314)*x[2]
        arg[1,1,1,0]=(-0.191500562856)*x[0]
        arg[1,1,1,1]=(0.244514598216)*x[1]
        arg[1,1,1,2]=(-0.804402720669)*x[2]
        arg[1,1,2,0]=(-0.455225710648)*x[0]
        arg[1,1,2,1]=(-0.505052700585)*x[1]
        arg[1,1,2,2]=(-0.0240295199362)*x[2]
        arg[1,2,0,0]=(-0.718487964893)*x[0]
        arg[1,2,0,1]=(-0.0899522570462)*x[1]
        arg[1,2,0,2]=(-0.293353754696)*x[2]
        arg[1,2,1,0]=(-0.180013826342)*x[0]
        arg[1,2,1,1]=(0.793689231922)*x[1]
        arg[1,2,1,2]=(0.673066555571)*x[2]
        arg[1,2,2,0]=(0.705362155032)*x[0]
        arg[1,2,2,1]=(0.54476742883)*x[1]
        arg[1,2,2,2]=(-0.331195064878)*x[2]
        arg[1,3,0,0]=(-0.360927441647)*x[0]
        arg[1,3,0,1]=(0.230772030282)*x[1]
        arg[1,3,0,2]=(0.912342489431)*x[2]
        arg[1,3,1,0]=(-0.817510690014)*x[0]
        arg[1,3,1,1]=(0.397583721353)*x[1]
        arg[1,3,1,2]=(-0.982551067917)*x[2]
        arg[1,3,2,0]=(0.86380240427)*x[0]
        arg[1,3,2,1]=(-0.415018976841)*x[1]
        arg[1,3,2,2]=(0.271582572267)*x[2]
        arg[1,4,0,0]=(0.252845347406)*x[0]
        arg[1,4,0,1]=(0.687786802906)*x[1]
        arg[1,4,0,2]=(0.465501171342)*x[2]
        arg[1,4,1,0]=(-0.613703721675)*x[0]
        arg[1,4,1,1]=(-0.110297640533)*x[1]
        arg[1,4,1,2]=(-0.836768056501)*x[2]
        arg[1,4,2,0]=(-0.0400898232224)*x[0]
        arg[1,4,2,1]=(0.0358172759009)*x[1]
        arg[1,4,2,2]=(-0.335751455408)*x[2]
        arg[2,0,0,0]=(-0.309992915015)*x[0]
        arg[2,0,0,1]=(-0.721404217867)*x[1]
        arg[2,0,0,2]=(-0.548000635629)*x[2]
        arg[2,0,1,0]=(0.651175831531)*x[0]
        arg[2,0,1,1]=(0.158960783491)*x[1]
        arg[2,0,1,2]=(-0.310676926155)*x[2]
        arg[2,0,2,0]=(-0.122289734411)*x[0]
        arg[2,0,2,1]=(-0.252405938421)*x[1]
        arg[2,0,2,2]=(-0.938280244213)*x[2]
        arg[2,1,0,0]=(0.559495801686)*x[0]
        arg[2,1,0,1]=(-0.547182622716)*x[1]
        arg[2,1,0,2]=(0.397441517898)*x[2]
        arg[2,1,1,0]=(-0.406112472071)*x[0]
        arg[2,1,1,1]=(0.355063810677)*x[1]
        arg[2,1,1,2]=(0.760400203215)*x[2]
        arg[2,1,2,0]=(0.992201320481)*x[0]
        arg[2,1,2,1]=(0.0580660882576)*x[1]
        arg[2,1,2,2]=(-0.643170879939)*x[2]
        arg[2,2,0,0]=(-0.280644461832)*x[0]
        arg[2,2,0,1]=(-0.0467430285531)*x[1]
        arg[2,2,0,2]=(0.314050593255)*x[2]
        arg[2,2,1,0]=(-0.230032618609)*x[0]
        arg[2,2,1,1]=(0.0996058698273)*x[1]
        arg[2,2,1,2]=(-0.0270266073208)*x[2]
        arg[2,2,2,0]=(0.767914132956)*x[0]
        arg[2,2,2,1]=(0.496930363612)*x[1]
        arg[2,2,2,2]=(-0.599525033616)*x[2]
        arg[2,3,0,0]=(-0.326433376073)*x[0]
        arg[2,3,0,1]=(-0.0366374501025)*x[1]
        arg[2,3,0,2]=(0.22555705749)*x[2]
        arg[2,3,1,0]=(-0.162548813895)*x[0]
        arg[2,3,1,1]=(-0.110074212194)*x[1]
        arg[2,3,1,2]=(-0.143600895553)*x[2]
        arg[2,3,2,0]=(0.771148880174)*x[0]
        arg[2,3,2,1]=(0.112528116552)*x[1]
        arg[2,3,2,2]=(-0.955735294341)*x[2]
        arg[2,4,0,0]=(-0.968392951034)*x[0]
        arg[2,4,0,1]=(-0.36901708507)*x[1]
        arg[2,4,0,2]=(0.283692515492)*x[2]
        arg[2,4,1,0]=(0.997238032837)*x[0]
        arg[2,4,1,1]=(-0.625794124653)*x[1]
        arg[2,4,1,2]=(0.533386027556)*x[2]
        arg[2,4,2,0]=(0.977311695557)*x[0]
        arg[2,4,2,1]=(0.693009976689)*x[1]
        arg[2,4,2,2]=(0.711179347652)*x[2]
        arg[3,0,0,0]=(-0.155585788931)*x[0]
        arg[3,0,0,1]=(0.0228078851234)*x[1]
        arg[3,0,0,2]=(0.510104938032)*x[2]
        arg[3,0,1,0]=(0.74865995369)*x[0]
        arg[3,0,1,1]=(0.672153736284)*x[1]
        arg[3,0,1,2]=(0.588012355098)*x[2]
        arg[3,0,2,0]=(-0.924508475715)*x[0]
        arg[3,0,2,1]=(-0.392784674758)*x[1]
        arg[3,0,2,2]=(-0.36371454642)*x[2]
        arg[3,1,0,0]=(-0.709783490337)*x[0]
        arg[3,1,0,1]=(0.844136172222)*x[1]
        arg[3,1,0,2]=(0.621011730043)*x[2]
        arg[3,1,1,0]=(0.428807337181)*x[0]
        arg[3,1,1,1]=(0.126300214574)*x[1]
        arg[3,1,1,2]=(0.795972806221)*x[2]
        arg[3,1,2,0]=(-0.252334324004)*x[0]
        arg[3,1,2,1]=(-0.722829467938)*x[1]
        arg[3,1,2,2]=(-0.551540062366)*x[2]
        arg[3,2,0,0]=(-0.134668475963)*x[0]
        arg[3,2,0,1]=(-0.598747540536)*x[1]
        arg[3,2,0,2]=(0.426422436624)*x[2]
        arg[3,2,1,0]=(-0.363050323762)*x[0]
        arg[3,2,1,1]=(0.980891457977)*x[1]
        arg[3,2,1,2]=(0.162831912555)*x[2]
        arg[3,2,2,0]=(-0.126505493475)*x[0]
        arg[3,2,2,1]=(-0.578567864811)*x[1]
        arg[3,2,2,2]=(-0.509843129095)*x[2]
        arg[3,3,0,0]=(-0.446171262265)*x[0]
        arg[3,3,0,1]=(-0.715175197494)*x[1]
        arg[3,3,0,2]=(-0.881016888806)*x[2]
        arg[3,3,1,0]=(-0.942020866327)*x[0]
        arg[3,3,1,1]=(0.156434646828)*x[1]
        arg[3,3,1,2]=(0.523624761583)*x[2]
        arg[3,3,2,0]=(-0.683550923926)*x[0]
        arg[3,3,2,1]=(0.857075218033)*x[1]
        arg[3,3,2,2]=(0.297672594023)*x[2]
        arg[3,4,0,0]=(0.74317121113)*x[0]
        arg[3,4,0,1]=(0.076464540756)*x[1]
        arg[3,4,0,2]=(0.781965468281)*x[2]
        arg[3,4,1,0]=(0.417750169098)*x[0]
        arg[3,4,1,1]=(0.82275428729)*x[1]
        arg[3,4,1,2]=(0.919072321093)*x[2]
        arg[3,4,2,0]=(-0.0246706472217)*x[0]
        arg[3,4,2,1]=(0.179863245513)*x[1]
        arg[3,4,2,2]=(0.539115287766)*x[2]
        ref=sqrt((52.492676775)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.245574919477)*sqrt(x[0])
        ref=sqrt((0.0603070410759)*dim)

      else:
        arg=Data(0,(),w)
        arg=(0.757324521515)*sqrt(x[0])
        ref=sqrt((0.573540430888)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.723421565407)*sqrt(x[0])
        arg[1]=(-0.460477393103)*sqrt(x[1])
        ref=sqrt((0.735378190855)*dim)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.88528497163)*sqrt(x[0])
        arg[1]=(-0.65510214636)*sqrt(x[1])
        arg[2]=(0.399538866363)*sqrt(x[2])
        ref=sqrt((1.37251960889)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.955620993904)*sqrt(x[0])
        arg[0,1]=(-0.0987865813703)*sqrt(x[1])
        arg[1,0]=(0.0288267231531)*sqrt(x[0])
        arg[1,1]=(0.655440599879)*sqrt(x[1])
        arg[2,0]=(0.685627284533)*sqrt(x[0])
        arg[2,1]=(-0.989832824892)*sqrt(x[1])
        arg[3,0]=(0.292184093194)*sqrt(x[0])
        arg[3,1]=(0.149553857773)*sqrt(x[1])
        ref=sqrt((2.91099532781)*dim)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.325908541533)*sqrt(x[0])
        arg[0,1]=(-0.992480479749)*sqrt(x[1])
        arg[0,2]=(0.660360271799)*sqrt(x[2])
        arg[1,0]=(0.173485908581)*sqrt(x[0])
        arg[1,1]=(-0.328755199781)*sqrt(x[1])
        arg[1,2]=(-0.943354674948)*sqrt(x[2])
        arg[2,0]=(0.680713222646)*sqrt(x[0])
        arg[2,1]=(-0.765971835693)*sqrt(x[1])
        arg[2,2]=(0.0413284847528)*sqrt(x[2])
        arg[3,0]=(0.990074004708)*sqrt(x[0])
        arg[3,1]=(0.941801786766)*sqrt(x[1])
        arg[3,2]=(0.886926192201)*sqrt(x[2])
        ref=sqrt((6.26107155228)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.0781611551598)*sqrt(x[0])
        arg[0,0,1]=(0.291016249575)*sqrt(x[1])
        arg[0,1,0]=(-0.107555233086)*sqrt(x[0])
        arg[0,1,1]=(-0.559067108546)*sqrt(x[1])
        arg[1,0,0]=(-0.0818406701266)*sqrt(x[0])
        arg[1,0,1]=(-0.594866806483)*sqrt(x[1])
        arg[1,1,0]=(-0.725814803863)*sqrt(x[0])
        arg[1,1,1]=(0.59128101992)*sqrt(x[1])
        arg[2,0,0]=(-0.15381555291)*sqrt(x[0])
        arg[2,0,1]=(-0.679882948503)*sqrt(x[1])
        arg[2,1,0]=(-0.58437193917)*sqrt(x[0])
        arg[2,1,1]=(0.136304615849)*sqrt(x[1])
        arg[3,0,0]=(0.0671365410096)*sqrt(x[0])
        arg[3,0,1]=(-0.645687212187)*sqrt(x[1])
        arg[3,1,0]=(-0.642492412392)*sqrt(x[0])
        arg[3,1,1]=(-0.125760054735)*sqrt(x[1])
        arg[4,0,0]=(0.731110824794)*sqrt(x[0])
        arg[4,0,1]=(0.491668422979)*sqrt(x[1])
        arg[4,1,0]=(-0.775841478292)*sqrt(x[0])
        arg[4,1,1]=(0.728265567974)*sqrt(x[1])
        arg[5,0,0]=(0.84511832373)*sqrt(x[0])
        arg[5,0,1]=(-0.513796801068)*sqrt(x[1])
        arg[5,1,0]=(0.113072243554)*sqrt(x[0])
        arg[5,1,1]=(0.246630838744)*sqrt(x[1])
        ref=sqrt((6.30829536252)*dim)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.369748116859)*sqrt(x[0])
        arg[0,0,1]=(-0.758056560031)*sqrt(x[1])
        arg[0,0,2]=(-0.873984709951)*sqrt(x[2])
        arg[0,1,0]=(0.311680165784)*sqrt(x[0])
        arg[0,1,1]=(0.374400673651)*sqrt(x[1])
        arg[0,1,2]=(0.712484217076)*sqrt(x[2])
        arg[1,0,0]=(0.829379714484)*sqrt(x[0])
        arg[1,0,1]=(-0.0551589596149)*sqrt(x[1])
        arg[1,0,2]=(0.965672208426)*sqrt(x[2])
        arg[1,1,0]=(-0.205044281547)*sqrt(x[0])
        arg[1,1,1]=(0.238197452756)*sqrt(x[1])
        arg[1,1,2]=(-0.33456139292)*sqrt(x[2])
        arg[2,0,0]=(0.649928288926)*sqrt(x[0])
        arg[2,0,1]=(-0.661384953389)*sqrt(x[1])
        arg[2,0,2]=(-0.253241222975)*sqrt(x[2])
        arg[2,1,0]=(-0.491716575992)*sqrt(x[0])
        arg[2,1,1]=(-0.970872527468)*sqrt(x[1])
        arg[2,1,2]=(0.222410198921)*sqrt(x[2])
        arg[3,0,0]=(0.205752630262)*sqrt(x[0])
        arg[3,0,1]=(0.864804362697)*sqrt(x[1])
        arg[3,0,2]=(-0.417975564033)*sqrt(x[2])
        arg[3,1,0]=(0.586425694033)*sqrt(x[0])
        arg[3,1,1]=(0.952661122184)*sqrt(x[1])
        arg[3,1,2]=(0.608680080453)*sqrt(x[2])
        arg[4,0,0]=(0.625968903369)*sqrt(x[0])
        arg[4,0,1]=(-0.573909405003)*sqrt(x[1])
        arg[4,0,2]=(-0.762256394595)*sqrt(x[2])
        arg[4,1,0]=(0.0710742394418)*sqrt(x[0])
        arg[4,1,1]=(0.583378040574)*sqrt(x[1])
        arg[4,1,2]=(0.719032893115)*sqrt(x[2])
        arg[5,0,0]=(0.032173368884)*sqrt(x[0])
        arg[5,0,1]=(-0.434042549492)*sqrt(x[1])
        arg[5,0,2]=(0.363504765447)*sqrt(x[2])
        arg[5,1,0]=(0.598817469198)*sqrt(x[0])
        arg[5,1,1]=(-0.163967008775)*sqrt(x[1])
        arg[5,1,2]=(0.546778730604)*sqrt(x[2])
        ref=sqrt((11.9696343123)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.273446520069)*sqrt(x[0])
        arg[0,0,0,1]=(-0.913305910831)*sqrt(x[1])
        arg[0,0,1,0]=(0.0745566747537)*sqrt(x[0])
        arg[0,0,1,1]=(0.98803601919)*sqrt(x[1])
        arg[0,0,2,0]=(-0.244120818875)*sqrt(x[0])
        arg[0,0,2,1]=(0.247509644998)*sqrt(x[1])
        arg[0,1,0,0]=(-0.548756417777)*sqrt(x[0])
        arg[0,1,0,1]=(-0.354587911923)*sqrt(x[1])
        arg[0,1,1,0]=(0.104268198867)*sqrt(x[0])
        arg[0,1,1,1]=(-0.541700877072)*sqrt(x[1])
        arg[0,1,2,0]=(-0.25900060689)*sqrt(x[0])
        arg[0,1,2,1]=(-0.859660175231)*sqrt(x[1])
        arg[0,2,0,0]=(0.195235560321)*sqrt(x[0])
        arg[0,2,0,1]=(0.175518738589)*sqrt(x[1])
        arg[0,2,1,0]=(-0.0638854232272)*sqrt(x[0])
        arg[0,2,1,1]=(-0.586161016541)*sqrt(x[1])
        arg[0,2,2,0]=(0.580258892247)*sqrt(x[0])
        arg[0,2,2,1]=(-0.931927145435)*sqrt(x[1])
        arg[0,3,0,0]=(-0.408209600298)*sqrt(x[0])
        arg[0,3,0,1]=(-0.0344882667014)*sqrt(x[1])
        arg[0,3,1,0]=(-0.131763534163)*sqrt(x[0])
        arg[0,3,1,1]=(-0.787653739965)*sqrt(x[1])
        arg[0,3,2,0]=(0.0910808104711)*sqrt(x[0])
        arg[0,3,2,1]=(-0.280409023611)*sqrt(x[1])
        arg[0,4,0,0]=(0.97745754012)*sqrt(x[0])
        arg[0,4,0,1]=(-0.59829020936)*sqrt(x[1])
        arg[0,4,1,0]=(0.890520260543)*sqrt(x[0])
        arg[0,4,1,1]=(-0.600760090231)*sqrt(x[1])
        arg[0,4,2,0]=(-0.897992297974)*sqrt(x[0])
        arg[0,4,2,1]=(-0.841169898923)*sqrt(x[1])
        arg[1,0,0,0]=(-0.249868816409)*sqrt(x[0])
        arg[1,0,0,1]=(0.620375082228)*sqrt(x[1])
        arg[1,0,1,0]=(-0.660480789306)*sqrt(x[0])
        arg[1,0,1,1]=(-0.73638571806)*sqrt(x[1])
        arg[1,0,2,0]=(0.339987316643)*sqrt(x[0])
        arg[1,0,2,1]=(0.541112529894)*sqrt(x[1])
        arg[1,1,0,0]=(-0.468808186705)*sqrt(x[0])
        arg[1,1,0,1]=(-0.32919679792)*sqrt(x[1])
        arg[1,1,1,0]=(0.917292803419)*sqrt(x[0])
        arg[1,1,1,1]=(-0.834265058005)*sqrt(x[1])
        arg[1,1,2,0]=(-0.247536849264)*sqrt(x[0])
        arg[1,1,2,1]=(-0.197503469238)*sqrt(x[1])
        arg[1,2,0,0]=(0.897591919909)*sqrt(x[0])
        arg[1,2,0,1]=(-0.807446231234)*sqrt(x[1])
        arg[1,2,1,0]=(-0.369878499382)*sqrt(x[0])
        arg[1,2,1,1]=(0.985678692179)*sqrt(x[1])
        arg[1,2,2,0]=(-0.709976427525)*sqrt(x[0])
        arg[1,2,2,1]=(-0.368744647016)*sqrt(x[1])
        arg[1,3,0,0]=(0.299630726462)*sqrt(x[0])
        arg[1,3,0,1]=(-0.445295757899)*sqrt(x[1])
        arg[1,3,1,0]=(-0.922386577254)*sqrt(x[0])
        arg[1,3,1,1]=(0.234794853697)*sqrt(x[1])
        arg[1,3,2,0]=(0.953377720197)*sqrt(x[0])
        arg[1,3,2,1]=(0.409778183998)*sqrt(x[1])
        arg[1,4,0,0]=(0.271967945488)*sqrt(x[0])
        arg[1,4,0,1]=(-0.578629001202)*sqrt(x[1])
        arg[1,4,1,0]=(0.210755226769)*sqrt(x[0])
        arg[1,4,1,1]=(-0.0902751419945)*sqrt(x[1])
        arg[1,4,2,0]=(0.70033387381)*sqrt(x[0])
        arg[1,4,2,1]=(0.305733565661)*sqrt(x[1])
        arg[2,0,0,0]=(-0.662483167298)*sqrt(x[0])
        arg[2,0,0,1]=(0.585252048652)*sqrt(x[1])
        arg[2,0,1,0]=(-0.398813785959)*sqrt(x[0])
        arg[2,0,1,1]=(-0.797438697186)*sqrt(x[1])
        arg[2,0,2,0]=(-0.508308971009)*sqrt(x[0])
        arg[2,0,2,1]=(0.302249407524)*sqrt(x[1])
        arg[2,1,0,0]=(0.208644491879)*sqrt(x[0])
        arg[2,1,0,1]=(-0.604749055374)*sqrt(x[1])
        arg[2,1,1,0]=(0.641654284594)*sqrt(x[0])
        arg[2,1,1,1]=(0.456898593356)*sqrt(x[1])
        arg[2,1,2,0]=(-0.398043867778)*sqrt(x[0])
        arg[2,1,2,1]=(-0.0712344657587)*sqrt(x[1])
        arg[2,2,0,0]=(0.0967860954865)*sqrt(x[0])
        arg[2,2,0,1]=(0.520449905952)*sqrt(x[1])
        arg[2,2,1,0]=(0.770198029595)*sqrt(x[0])
        arg[2,2,1,1]=(-0.594004671621)*sqrt(x[1])
        arg[2,2,2,0]=(-0.744571452885)*sqrt(x[0])
        arg[2,2,2,1]=(0.544447367825)*sqrt(x[1])
        arg[2,3,0,0]=(-0.137087966968)*sqrt(x[0])
        arg[2,3,0,1]=(0.120672667497)*sqrt(x[1])
        arg[2,3,1,0]=(0.204800088057)*sqrt(x[0])
        arg[2,3,1,1]=(0.626526346076)*sqrt(x[1])
        arg[2,3,2,0]=(-0.696480393227)*sqrt(x[0])
        arg[2,3,2,1]=(0.188533741996)*sqrt(x[1])
        arg[2,4,0,0]=(-0.403523821067)*sqrt(x[0])
        arg[2,4,0,1]=(-0.428048989483)*sqrt(x[1])
        arg[2,4,1,0]=(-0.244186366584)*sqrt(x[0])
        arg[2,4,1,1]=(0.00866444909003)*sqrt(x[1])
        arg[2,4,2,0]=(-0.445991308853)*sqrt(x[0])
        arg[2,4,2,1]=(-0.899951068935)*sqrt(x[1])
        arg[3,0,0,0]=(0.609340085418)*sqrt(x[0])
        arg[3,0,0,1]=(0.878750391425)*sqrt(x[1])
        arg[3,0,1,0]=(0.258064654464)*sqrt(x[0])
        arg[3,0,1,1]=(-0.482402612985)*sqrt(x[1])
        arg[3,0,2,0]=(0.943732283389)*sqrt(x[0])
        arg[3,0,2,1]=(0.65514211843)*sqrt(x[1])
        arg[3,1,0,0]=(-0.894551979619)*sqrt(x[0])
        arg[3,1,0,1]=(0.220116541042)*sqrt(x[1])
        arg[3,1,1,0]=(0.386887699577)*sqrt(x[0])
        arg[3,1,1,1]=(-0.422560075108)*sqrt(x[1])
        arg[3,1,2,0]=(0.00387273783493)*sqrt(x[0])
        arg[3,1,2,1]=(0.465673505613)*sqrt(x[1])
        arg[3,2,0,0]=(0.987383428982)*sqrt(x[0])
        arg[3,2,0,1]=(0.376320055964)*sqrt(x[1])
        arg[3,2,1,0]=(-0.463778689128)*sqrt(x[0])
        arg[3,2,1,1]=(0.179816566227)*sqrt(x[1])
        arg[3,2,2,0]=(0.961522856801)*sqrt(x[0])
        arg[3,2,2,1]=(-0.257779627946)*sqrt(x[1])
        arg[3,3,0,0]=(0.748886531458)*sqrt(x[0])
        arg[3,3,0,1]=(-0.257342282566)*sqrt(x[1])
        arg[3,3,1,0]=(0.377494024401)*sqrt(x[0])
        arg[3,3,1,1]=(-0.334588346017)*sqrt(x[1])
        arg[3,3,2,0]=(0.502495149189)*sqrt(x[0])
        arg[3,3,2,1]=(0.534612429702)*sqrt(x[1])
        arg[3,4,0,0]=(-0.308551337355)*sqrt(x[0])
        arg[3,4,0,1]=(-0.471825826745)*sqrt(x[1])
        arg[3,4,1,0]=(-0.262606531584)*sqrt(x[0])
        arg[3,4,1,1]=(0.766089616367)*sqrt(x[1])
        arg[3,4,2,0]=(-0.136526755642)*sqrt(x[0])
        arg[3,4,2,1]=(0.675111459363)*sqrt(x[1])
        ref=sqrt((37.3453550914)*dim)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.532946231146)*sqrt(x[0])
        arg[0,0,0,1]=(0.269364089513)*sqrt(x[1])
        arg[0,0,0,2]=(-0.207412457081)*sqrt(x[2])
        arg[0,0,1,0]=(-0.843104704858)*sqrt(x[0])
        arg[0,0,1,1]=(0.0416216508473)*sqrt(x[1])
        arg[0,0,1,2]=(-0.836074693662)*sqrt(x[2])
        arg[0,0,2,0]=(0.943609268731)*sqrt(x[0])
        arg[0,0,2,1]=(0.0154543737816)*sqrt(x[1])
        arg[0,0,2,2]=(0.0726268788381)*sqrt(x[2])
        arg[0,1,0,0]=(0.108422740078)*sqrt(x[0])
        arg[0,1,0,1]=(-0.296667916638)*sqrt(x[1])
        arg[0,1,0,2]=(-0.769732600535)*sqrt(x[2])
        arg[0,1,1,0]=(-0.428575493834)*sqrt(x[0])
        arg[0,1,1,1]=(0.421245456722)*sqrt(x[1])
        arg[0,1,1,2]=(-0.588277652277)*sqrt(x[2])
        arg[0,1,2,0]=(0.145294576795)*sqrt(x[0])
        arg[0,1,2,1]=(0.323206623794)*sqrt(x[1])
        arg[0,1,2,2]=(0.788115602892)*sqrt(x[2])
        arg[0,2,0,0]=(-0.227877282292)*sqrt(x[0])
        arg[0,2,0,1]=(-0.630647460719)*sqrt(x[1])
        arg[0,2,0,2]=(0.58754882135)*sqrt(x[2])
        arg[0,2,1,0]=(0.347191113403)*sqrt(x[0])
        arg[0,2,1,1]=(0.464093634725)*sqrt(x[1])
        arg[0,2,1,2]=(-0.0412800774497)*sqrt(x[2])
        arg[0,2,2,0]=(0.223364317185)*sqrt(x[0])
        arg[0,2,2,1]=(0.257201130157)*sqrt(x[1])
        arg[0,2,2,2]=(0.063203467463)*sqrt(x[2])
        arg[0,3,0,0]=(-0.723240451643)*sqrt(x[0])
        arg[0,3,0,1]=(-0.862468295097)*sqrt(x[1])
        arg[0,3,0,2]=(-0.149283247587)*sqrt(x[2])
        arg[0,3,1,0]=(0.15680097839)*sqrt(x[0])
        arg[0,3,1,1]=(0.421563637547)*sqrt(x[1])
        arg[0,3,1,2]=(0.111549188549)*sqrt(x[2])
        arg[0,3,2,0]=(-0.272783329363)*sqrt(x[0])
        arg[0,3,2,1]=(-0.420352789853)*sqrt(x[1])
        arg[0,3,2,2]=(0.570865117722)*sqrt(x[2])
        arg[0,4,0,0]=(-0.321910078414)*sqrt(x[0])
        arg[0,4,0,1]=(0.988695599439)*sqrt(x[1])
        arg[0,4,0,2]=(0.920200893398)*sqrt(x[2])
        arg[0,4,1,0]=(0.0260910072651)*sqrt(x[0])
        arg[0,4,1,1]=(0.460012578184)*sqrt(x[1])
        arg[0,4,1,2]=(0.848099524112)*sqrt(x[2])
        arg[0,4,2,0]=(0.242157803251)*sqrt(x[0])
        arg[0,4,2,1]=(0.394528777004)*sqrt(x[1])
        arg[0,4,2,2]=(0.562996837311)*sqrt(x[2])
        arg[1,0,0,0]=(0.459886225958)*sqrt(x[0])
        arg[1,0,0,1]=(-0.721868942003)*sqrt(x[1])
        arg[1,0,0,2]=(0.432203082994)*sqrt(x[2])
        arg[1,0,1,0]=(0.409831045482)*sqrt(x[0])
        arg[1,0,1,1]=(-0.481677513473)*sqrt(x[1])
        arg[1,0,1,2]=(0.439387853437)*sqrt(x[2])
        arg[1,0,2,0]=(0.261583198434)*sqrt(x[0])
        arg[1,0,2,1]=(0.290993423577)*sqrt(x[1])
        arg[1,0,2,2]=(0.477993114134)*sqrt(x[2])
        arg[1,1,0,0]=(0.586344598248)*sqrt(x[0])
        arg[1,1,0,1]=(-0.105390792831)*sqrt(x[1])
        arg[1,1,0,2]=(0.335990751314)*sqrt(x[2])
        arg[1,1,1,0]=(-0.191500562856)*sqrt(x[0])
        arg[1,1,1,1]=(0.244514598216)*sqrt(x[1])
        arg[1,1,1,2]=(-0.804402720669)*sqrt(x[2])
        arg[1,1,2,0]=(-0.455225710648)*sqrt(x[0])
        arg[1,1,2,1]=(-0.505052700585)*sqrt(x[1])
        arg[1,1,2,2]=(-0.0240295199362)*sqrt(x[2])
        arg[1,2,0,0]=(-0.718487964893)*sqrt(x[0])
        arg[1,2,0,1]=(-0.0899522570462)*sqrt(x[1])
        arg[1,2,0,2]=(-0.293353754696)*sqrt(x[2])
        arg[1,2,1,0]=(-0.180013826342)*sqrt(x[0])
        arg[1,2,1,1]=(0.793689231922)*sqrt(x[1])
        arg[1,2,1,2]=(0.673066555571)*sqrt(x[2])
        arg[1,2,2,0]=(0.705362155032)*sqrt(x[0])
        arg[1,2,2,1]=(0.54476742883)*sqrt(x[1])
        arg[1,2,2,2]=(-0.331195064878)*sqrt(x[2])
        arg[1,3,0,0]=(-0.360927441647)*sqrt(x[0])
        arg[1,3,0,1]=(0.230772030282)*sqrt(x[1])
        arg[1,3,0,2]=(0.912342489431)*sqrt(x[2])
        arg[1,3,1,0]=(-0.817510690014)*sqrt(x[0])
        arg[1,3,1,1]=(0.397583721353)*sqrt(x[1])
        arg[1,3,1,2]=(-0.982551067917)*sqrt(x[2])
        arg[1,3,2,0]=(0.86380240427)*sqrt(x[0])
        arg[1,3,2,1]=(-0.415018976841)*sqrt(x[1])
        arg[1,3,2,2]=(0.271582572267)*sqrt(x[2])
        arg[1,4,0,0]=(0.252845347406)*sqrt(x[0])
        arg[1,4,0,1]=(0.687786802906)*sqrt(x[1])
        arg[1,4,0,2]=(0.465501171342)*sqrt(x[2])
        arg[1,4,1,0]=(-0.613703721675)*sqrt(x[0])
        arg[1,4,1,1]=(-0.110297640533)*sqrt(x[1])
        arg[1,4,1,2]=(-0.836768056501)*sqrt(x[2])
        arg[1,4,2,0]=(-0.0400898232224)*sqrt(x[0])
        arg[1,4,2,1]=(0.0358172759009)*sqrt(x[1])
        arg[1,4,2,2]=(-0.335751455408)*sqrt(x[2])
        arg[2,0,0,0]=(-0.309992915015)*sqrt(x[0])
        arg[2,0,0,1]=(-0.721404217867)*sqrt(x[1])
        arg[2,0,0,2]=(-0.548000635629)*sqrt(x[2])
        arg[2,0,1,0]=(0.651175831531)*sqrt(x[0])
        arg[2,0,1,1]=(0.158960783491)*sqrt(x[1])
        arg[2,0,1,2]=(-0.310676926155)*sqrt(x[2])
        arg[2,0,2,0]=(-0.122289734411)*sqrt(x[0])
        arg[2,0,2,1]=(-0.252405938421)*sqrt(x[1])
        arg[2,0,2,2]=(-0.938280244213)*sqrt(x[2])
        arg[2,1,0,0]=(0.559495801686)*sqrt(x[0])
        arg[2,1,0,1]=(-0.547182622716)*sqrt(x[1])
        arg[2,1,0,2]=(0.397441517898)*sqrt(x[2])
        arg[2,1,1,0]=(-0.406112472071)*sqrt(x[0])
        arg[2,1,1,1]=(0.355063810677)*sqrt(x[1])
        arg[2,1,1,2]=(0.760400203215)*sqrt(x[2])
        arg[2,1,2,0]=(0.992201320481)*sqrt(x[0])
        arg[2,1,2,1]=(0.0580660882576)*sqrt(x[1])
        arg[2,1,2,2]=(-0.643170879939)*sqrt(x[2])
        arg[2,2,0,0]=(-0.280644461832)*sqrt(x[0])
        arg[2,2,0,1]=(-0.0467430285531)*sqrt(x[1])
        arg[2,2,0,2]=(0.314050593255)*sqrt(x[2])
        arg[2,2,1,0]=(-0.230032618609)*sqrt(x[0])
        arg[2,2,1,1]=(0.0996058698273)*sqrt(x[1])
        arg[2,2,1,2]=(-0.0270266073208)*sqrt(x[2])
        arg[2,2,2,0]=(0.767914132956)*sqrt(x[0])
        arg[2,2,2,1]=(0.496930363612)*sqrt(x[1])
        arg[2,2,2,2]=(-0.599525033616)*sqrt(x[2])
        arg[2,3,0,0]=(-0.326433376073)*sqrt(x[0])
        arg[2,3,0,1]=(-0.0366374501025)*sqrt(x[1])
        arg[2,3,0,2]=(0.22555705749)*sqrt(x[2])
        arg[2,3,1,0]=(-0.162548813895)*sqrt(x[0])
        arg[2,3,1,1]=(-0.110074212194)*sqrt(x[1])
        arg[2,3,1,2]=(-0.143600895553)*sqrt(x[2])
        arg[2,3,2,0]=(0.771148880174)*sqrt(x[0])
        arg[2,3,2,1]=(0.112528116552)*sqrt(x[1])
        arg[2,3,2,2]=(-0.955735294341)*sqrt(x[2])
        arg[2,4,0,0]=(-0.968392951034)*sqrt(x[0])
        arg[2,4,0,1]=(-0.36901708507)*sqrt(x[1])
        arg[2,4,0,2]=(0.283692515492)*sqrt(x[2])
        arg[2,4,1,0]=(0.997238032837)*sqrt(x[0])
        arg[2,4,1,1]=(-0.625794124653)*sqrt(x[1])
        arg[2,4,1,2]=(0.533386027556)*sqrt(x[2])
        arg[2,4,2,0]=(0.977311695557)*sqrt(x[0])
        arg[2,4,2,1]=(0.693009976689)*sqrt(x[1])
        arg[2,4,2,2]=(0.711179347652)*sqrt(x[2])
        arg[3,0,0,0]=(-0.155585788931)*sqrt(x[0])
        arg[3,0,0,1]=(0.0228078851234)*sqrt(x[1])
        arg[3,0,0,2]=(0.510104938032)*sqrt(x[2])
        arg[3,0,1,0]=(0.74865995369)*sqrt(x[0])
        arg[3,0,1,1]=(0.672153736284)*sqrt(x[1])
        arg[3,0,1,2]=(0.588012355098)*sqrt(x[2])
        arg[3,0,2,0]=(-0.924508475715)*sqrt(x[0])
        arg[3,0,2,1]=(-0.392784674758)*sqrt(x[1])
        arg[3,0,2,2]=(-0.36371454642)*sqrt(x[2])
        arg[3,1,0,0]=(-0.709783490337)*sqrt(x[0])
        arg[3,1,0,1]=(0.844136172222)*sqrt(x[1])
        arg[3,1,0,2]=(0.621011730043)*sqrt(x[2])
        arg[3,1,1,0]=(0.428807337181)*sqrt(x[0])
        arg[3,1,1,1]=(0.126300214574)*sqrt(x[1])
        arg[3,1,1,2]=(0.795972806221)*sqrt(x[2])
        arg[3,1,2,0]=(-0.252334324004)*sqrt(x[0])
        arg[3,1,2,1]=(-0.722829467938)*sqrt(x[1])
        arg[3,1,2,2]=(-0.551540062366)*sqrt(x[2])
        arg[3,2,0,0]=(-0.134668475963)*sqrt(x[0])
        arg[3,2,0,1]=(-0.598747540536)*sqrt(x[1])
        arg[3,2,0,2]=(0.426422436624)*sqrt(x[2])
        arg[3,2,1,0]=(-0.363050323762)*sqrt(x[0])
        arg[3,2,1,1]=(0.980891457977)*sqrt(x[1])
        arg[3,2,1,2]=(0.162831912555)*sqrt(x[2])
        arg[3,2,2,0]=(-0.126505493475)*sqrt(x[0])
        arg[3,2,2,1]=(-0.578567864811)*sqrt(x[1])
        arg[3,2,2,2]=(-0.509843129095)*sqrt(x[2])
        arg[3,3,0,0]=(-0.446171262265)*sqrt(x[0])
        arg[3,3,0,1]=(-0.715175197494)*sqrt(x[1])
        arg[3,3,0,2]=(-0.881016888806)*sqrt(x[2])
        arg[3,3,1,0]=(-0.942020866327)*sqrt(x[0])
        arg[3,3,1,1]=(0.156434646828)*sqrt(x[1])
        arg[3,3,1,2]=(0.523624761583)*sqrt(x[2])
        arg[3,3,2,0]=(-0.683550923926)*sqrt(x[0])
        arg[3,3,2,1]=(0.857075218033)*sqrt(x[1])
        arg[3,3,2,2]=(0.297672594023)*sqrt(x[2])
        arg[3,4,0,0]=(0.74317121113)*sqrt(x[0])
        arg[3,4,0,1]=(0.076464540756)*sqrt(x[1])
        arg[3,4,0,2]=(0.781965468281)*sqrt(x[2])
        arg[3,4,1,0]=(0.417750169098)*sqrt(x[0])
        arg[3,4,1,1]=(0.82275428729)*sqrt(x[1])
        arg[3,4,1,2]=(0.919072321093)*sqrt(x[2])
        arg[3,4,2,0]=(-0.0246706472217)*sqrt(x[0])
        arg[3,4,2,1]=(0.179863245513)*sqrt(x[1])
        arg[3,4,2,2]=(0.539115287766)*sqrt(x[2])
        ref=sqrt((52.492676775)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")

