
##############################################################################
#
# Copyright (c) 2003-2016 by The University of Queensland
# http://www.uq.edu.au
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# Development until 2012 by Earth Systems Science Computational Center (ESSCC)
# Development 2012-2013 by School of Earth Sciences
# Development from 2014 by Centre for Geoscience Computing (GeoComp)
#
##############################################################################

from __future__ import print_function, division

__copyright__="""Copyright (c) 2003-2016 by The University of Queensland
http://www.uq.edu.au
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://launchpad.net/escript-finley"

"""
basic tests for functions in util.py effecting the spatial distribution 

it is assumed that the domain is the usint square/cube 

not all these test will run for all domains. check the doc string for the assumptions of a particular test

:var __author__: name of author
:var __copyright__: copyrights
:var __license__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""
__author__="Lutz Gross, l.gross@uq.edu.au"

import esys.escriptcore.utestselect as unittest
from esys.escript import *
from numpy import array
import numpy

from test_util_spatial_functions1 import Test_Util_SpatialFunctions_noGradOnBoundary_noContact

#begin

class Test_Util_SpatialFunctions_noGradOnBoundary(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
   RES_TOL=1.e-8

   def test_normal_onFunctionOnContactZero(self):
     """
     test getNormal() on contact side 0

     assumptions: FunctionOnContactZero(self.domain) exists
     """
     dim=self.domain.getDim()
     f=FunctionOnContactZero(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[1,0,0]
     else:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[1,0]

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(length(res)-1)<=self.RES_TOL,"wrong length")
     self.assertTrue(Lsup(abs(inner(ref,res))-1)<=self.RES_TOL,"wrong direction")
   def test_normal_onReducedFunctionOnContactZero(self):
     """
     test getNormal() on contact side 0

     assumptions: FunctionOnContactZero(self.domain) exists
     """
     dim=self.domain.getDim()
     f=ReducedFunctionOnContactZero(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[1,0,0]
     else:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[1,0]

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(length(res)-1)<=self.RES_TOL,"wrong length")
     self.assertTrue(Lsup(abs(inner(ref,res))-1)<=self.RES_TOL,"wrong direction")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_normal_onFunctionOnContactOne(self):
     """
     test getNormal() on contact side 1

     assumptions: FunctionOnContactOne(self.domain) exists
     """
     dim=self.domain.getDim()
     f=FunctionOnContactOne(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[-1,0,0]
     else:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[-1,0]

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(length(res)-1)<=self.RES_TOL,"wrong length")
     self.assertTrue(Lsup(abs(inner(ref,res))-1)<=self.RES_TOL,"wrong direction")
   def test_normal_onReducedFunctionOnContactOne(self):
     """
     test getNormal() on contact side 1

     assumptions: FunctionOnContactOne(self.domain) exists
     """
     dim=self.domain.getDim()
     f=ReducedFunctionOnContactOne(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[-1,0,0]
     else:
         ref+=whereZero(x[0]-0.5,tol=self.RES_TOL)*[-1,0]

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(length(res)-1)<=self.RES_TOL,"wrong length")
     self.assertTrue(Lsup(abs(inner(ref,res))-1)<=self.RES_TOL,"wrong direction")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.311774053155)*x[0]**o+(0.52292841627)*x[0]+(-0.518467126196)*x[1]**o+(-0.11329290159)*x[1]
        ref=(-0.518467126196)/(o+1.)+(0.20481775734)+(0.311774053155)*0.5**o
      else:
        arg=(0.233765319091)*x[0]**o+(-0.453551209762)*x[0]+(0.920140231713)*x[1]**o+(0.741695479084)*x[1]+(-0.255834461065)*x[2]**o+(-0.102580899053)*x[2]
        ref=(0.664305770648)/(o+1.)+(0.0927816851348)+(0.233765319091)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.651678780872)*x[0]**o+(-0.749709447394)*x[0]+(0.311088209082)*x[1]**o+(0.553081401193)*x[1]
        arg[1]=(0.657887728987)*x[0]**o+(0.052270143604)*x[0]+(-0.15507994277)*x[1]**o+(-0.30302150082)*x[1]
        ref[0]=(0.311088209082)/(o+1.)+(-0.0983140231004)+(-0.651678780872)*0.5**o
        ref[1]=(-0.15507994277)/(o+1.)+(-0.125375678608)+(0.657887728987)*0.5**o
      else:
        arg[0]=(0.844909347644)*x[0]**o+(-0.342697697483)*x[0]+(0.126242548186)*x[1]**o+(-0.854010732849)*x[1]+(-0.307627587461)*x[2]**o+(-0.998566290754)*x[2]
        arg[1]=(0.900073726701)*x[0]**o+(-0.703841701859)*x[0]+(-0.651690896499)*x[1]**o+(0.828362322181)*x[1]+(0.284410420895)*x[2]**o+(-0.394606348974)*x[2]
        ref[0]=(-0.181385039275)/(o+1.)+(-1.09763736054)+(0.844909347644)*0.5**o
        ref[1]=(-0.367280475604)/(o+1.)+(-0.135042864326)+(0.900073726701)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.263642681371)*x[0]**o+(0.668992571787)*x[0]+(0.873939846194)*x[1]**o+(0.784820355998)*x[1]
        arg[0,1]=(-0.389306088628)*x[0]**o+(-0.0549482389907)*x[0]+(-0.787101008457)*x[1]**o+(0.860345123812)*x[1]
        arg[0,2]=(-0.995721191961)*x[0]**o+(-0.587928136153)*x[0]+(-0.378312614832)*x[1]**o+(0.199765674281)*x[1]
        arg[0,3]=(-0.374679425145)*x[0]**o+(-0.503606183146)*x[0]+(0.23995738992)*x[1]**o+(-0.525359181744)*x[1]
        arg[0,4]=(-0.69923255665)*x[0]**o+(0.547634979941)*x[0]+(0.728792702739)*x[1]**o+(0.949888533828)*x[1]
        arg[1,0]=(-0.61856415528)*x[0]**o+(-0.0516239651093)*x[0]+(0.158516649979)*x[1]**o+(-0.0522288881011)*x[1]
        arg[1,1]=(0.0285524133857)*x[0]**o+(0.741357758307)*x[0]+(-0.0714860903847)*x[1]**o+(-0.539632750385)*x[1]
        arg[1,2]=(-0.291101401382)*x[0]**o+(-0.114900246437)*x[0]+(-0.982305031665)*x[1]**o+(0.0823478844055)*x[1]
        arg[1,3]=(-0.931305732525)*x[0]**o+(0.077332436417)*x[0]+(-0.880626154921)*x[1]**o+(-0.321222234507)*x[1]
        arg[1,4]=(-0.00778033208815)*x[0]**o+(-0.536626847957)*x[0]+(-0.599742714546)*x[1]**o+(0.454108299336)*x[1]
        arg[2,0]=(-0.575243090445)*x[0]**o+(-0.920247228564)*x[0]+(0.4569784462)*x[1]**o+(-0.105175798804)*x[1]
        arg[2,1]=(0.333985911748)*x[0]**o+(-0.017876213076)*x[0]+(0.475164778165)*x[1]**o+(0.561072178015)*x[1]
        arg[2,2]=(-0.559587786773)*x[0]**o+(-0.917134128672)*x[0]+(0.316654889327)*x[1]**o+(-0.640569595027)*x[1]
        arg[2,3]=(-0.140461324817)*x[0]**o+(0.259052969883)*x[0]+(0.397898509159)*x[1]**o+(0.932233331785)*x[1]
        arg[2,4]=(0.368626304985)*x[0]**o+(0.616310502384)*x[0]+(-0.756845293566)*x[1]**o+(-0.920088968132)*x[1]
        arg[3,0]=(-0.824398741087)*x[0]**o+(0.0498115894252)*x[0]+(-0.5256621655)*x[1]**o+(0.114806237108)*x[1]
        arg[3,1]=(-0.77671057316)*x[0]**o+(0.569249443162)*x[0]+(0.416628465801)*x[1]**o+(0.0603627487913)*x[1]
        arg[3,2]=(-0.247875220913)*x[0]**o+(0.19548828878)*x[0]+(0.643226101838)*x[1]**o+(-0.154558826499)*x[1]
        arg[3,3]=(-0.918937118963)*x[0]**o+(0.193777174417)*x[0]+(-0.906042347107)*x[1]**o+(0.872033230033)*x[1]
        arg[3,4]=(-0.628918336086)*x[0]**o+(0.823242191714)*x[0]+(-0.593691166712)*x[1]**o+(-0.661715840814)*x[1]
        ref[0,0]=(0.873939846194)/(o+1.)+(0.726906463892)+(-0.263642681371)*0.5**o
        ref[0,1]=(-0.787101008457)/(o+1.)+(0.402698442411)+(-0.389306088628)*0.5**o
        ref[0,2]=(-0.378312614832)/(o+1.)+(-0.194081230936)+(-0.995721191961)*0.5**o
        ref[0,3]=(0.23995738992)/(o+1.)+(-0.514482682445)+(-0.374679425145)*0.5**o
        ref[0,4]=(0.728792702739)/(o+1.)+(0.748761756885)+(-0.69923255665)*0.5**o
        ref[1,0]=(0.158516649979)/(o+1.)+(-0.0519264266052)+(-0.61856415528)*0.5**o
        ref[1,1]=(-0.0714860903847)/(o+1.)+(0.100862503961)+(0.0285524133857)*0.5**o
        ref[1,2]=(-0.982305031665)/(o+1.)+(-0.0162761810159)+(-0.291101401382)*0.5**o
        ref[1,3]=(-0.880626154921)/(o+1.)+(-0.121944899045)+(-0.931305732525)*0.5**o
        ref[1,4]=(-0.599742714546)/(o+1.)+(-0.0412592743102)+(-0.00778033208815)*0.5**o
        ref[2,0]=(0.4569784462)/(o+1.)+(-0.512711513684)+(-0.575243090445)*0.5**o
        ref[2,1]=(0.475164778165)/(o+1.)+(0.27159798247)+(0.333985911748)*0.5**o
        ref[2,2]=(0.316654889327)/(o+1.)+(-0.778851861849)+(-0.559587786773)*0.5**o
        ref[2,3]=(0.397898509159)/(o+1.)+(0.595643150834)+(-0.140461324817)*0.5**o
        ref[2,4]=(-0.756845293566)/(o+1.)+(-0.151889232874)+(0.368626304985)*0.5**o
        ref[3,0]=(-0.5256621655)/(o+1.)+(0.0823089132668)+(-0.824398741087)*0.5**o
        ref[3,1]=(0.416628465801)/(o+1.)+(0.314806095977)+(-0.77671057316)*0.5**o
        ref[3,2]=(0.643226101838)/(o+1.)+(0.0204647311405)+(-0.247875220913)*0.5**o
        ref[3,3]=(-0.906042347107)/(o+1.)+(0.532905202225)+(-0.918937118963)*0.5**o
        ref[3,4]=(-0.593691166712)/(o+1.)+(0.0807631754499)+(-0.628918336086)*0.5**o
      else:
        arg[0,0]=(0.748857269399)*x[0]**o+(-0.21982973469)*x[0]+(-0.765987921295)*x[1]**o+(0.556577019909)*x[1]+(-0.0438548079278)*x[2]**o+(0.543610874817)*x[2]
        arg[0,1]=(-0.784526004748)*x[0]**o+(-0.949423172035)*x[0]+(-0.29888188317)*x[1]**o+(-0.850638766848)*x[1]+(-0.853027558387)*x[2]**o+(-0.846378486588)*x[2]
        arg[0,2]=(0.53445882525)*x[0]**o+(-0.653655652989)*x[0]+(-0.968129875689)*x[1]**o+(-0.603598808323)*x[1]+(0.545841014272)*x[2]**o+(0.126131861082)*x[2]
        arg[0,3]=(0.181662498503)*x[0]**o+(-0.769571080865)*x[0]+(0.771109181078)*x[1]**o+(0.980677373303)*x[1]+(-0.625057940268)*x[2]**o+(0.0585672357346)*x[2]
        arg[0,4]=(-0.0320907540755)*x[0]**o+(-0.519517369917)*x[0]+(-0.365190231996)*x[1]**o+(-0.192392652661)*x[1]+(-0.693267159516)*x[2]**o+(0.106458640425)*x[2]
        arg[1,0]=(0.414286358161)*x[0]**o+(0.0223282279507)*x[0]+(0.215881279469)*x[1]**o+(0.68990977849)*x[1]+(-0.532075885581)*x[2]**o+(-0.269164904232)*x[2]
        arg[1,1]=(-0.0445879250589)*x[0]**o+(0.889985400599)*x[0]+(0.842050582947)*x[1]**o+(-0.0275307884642)*x[1]+(0.774650107671)*x[2]**o+(0.19446514971)*x[2]
        arg[1,2]=(0.913032880034)*x[0]**o+(-0.286537470008)*x[0]+(0.0902778267032)*x[1]**o+(0.548596933564)*x[1]+(-0.996100405074)*x[2]**o+(-0.509554019549)*x[2]
        arg[1,3]=(0.44308812765)*x[0]**o+(-0.0641227726924)*x[0]+(0.484249008075)*x[1]**o+(0.581496384894)*x[1]+(0.592538533183)*x[2]**o+(-0.821638541512)*x[2]
        arg[1,4]=(-0.898304735153)*x[0]**o+(0.333217967121)*x[0]+(-0.788218899371)*x[1]**o+(0.579723459039)*x[1]+(0.467006473159)*x[2]**o+(-0.277113860535)*x[2]
        arg[2,0]=(0.515308716564)*x[0]**o+(0.255947750577)*x[0]+(0.232578337945)*x[1]**o+(0.900611887592)*x[1]+(-0.247710965611)*x[2]**o+(-0.985523228221)*x[2]
        arg[2,1]=(0.611196756618)*x[0]**o+(0.146822999187)*x[0]+(-0.318793147278)*x[1]**o+(0.790772432028)*x[1]+(0.138036152668)*x[2]**o+(-0.800234207543)*x[2]
        arg[2,2]=(-0.293576196546)*x[0]**o+(-0.521059592737)*x[0]+(0.301178293957)*x[1]**o+(-0.141430900198)*x[1]+(-0.782381874467)*x[2]**o+(0.92221179729)*x[2]
        arg[2,3]=(0.673951778483)*x[0]**o+(-0.335068772752)*x[0]+(0.408793200723)*x[1]**o+(0.36906212176)*x[1]+(0.246292093855)*x[2]**o+(0.356147233675)*x[2]
        arg[2,4]=(-0.218899302236)*x[0]**o+(0.271285962886)*x[0]+(-0.16364707283)*x[1]**o+(0.177195607727)*x[1]+(-0.287331396699)*x[2]**o+(-0.465391460497)*x[2]
        arg[3,0]=(-0.369727856735)*x[0]**o+(-0.44351931666)*x[0]+(0.0375097686939)*x[1]**o+(0.276779121402)*x[1]+(0.662505672881)*x[2]**o+(-0.642709866174)*x[2]
        arg[3,1]=(0.444940917816)*x[0]**o+(-0.178926126454)*x[0]+(0.824660645994)*x[1]**o+(0.825996820355)*x[1]+(0.798260651848)*x[2]**o+(0.331744022239)*x[2]
        arg[3,2]=(-0.82489261483)*x[0]**o+(-0.340862079766)*x[0]+(-0.998375587383)*x[1]**o+(0.880222106116)*x[1]+(0.656096580473)*x[2]**o+(0.530854486127)*x[2]
        arg[3,3]=(0.732863411435)*x[0]**o+(-0.324252004147)*x[0]+(0.789472654884)*x[1]**o+(-0.207794796011)*x[1]+(0.50903180567)*x[2]**o+(0.724319112504)*x[2]
        arg[3,4]=(-0.980748139471)*x[0]**o+(0.0974331161906)*x[0]+(0.196275528452)*x[1]**o+(0.305231910003)*x[1]+(-0.0417825350947)*x[2]**o+(-0.2916834261)*x[2]
        ref[0,0]=(-0.809842729223)/(o+1.)+(0.440179080018)+(0.748857269399)*0.5**o
        ref[0,1]=(-1.15190944156)/(o+1.)+(-1.32322021274)+(-0.784526004748)*0.5**o
        ref[0,2]=(-0.422288861417)/(o+1.)+(-0.565561300115)+(0.53445882525)*0.5**o
        ref[0,3]=(0.146051240811)/(o+1.)+(0.134836764087)+(0.181662498503)*0.5**o
        ref[0,4]=(-1.05845739151)/(o+1.)+(-0.302725691077)+(-0.0320907540755)*0.5**o
        ref[1,0]=(-0.316194606113)/(o+1.)+(0.221536551104)+(0.414286358161)*0.5**o
        ref[1,1]=(1.61670069062)/(o+1.)+(0.528459880922)+(-0.0445879250589)*0.5**o
        ref[1,2]=(-0.905822578371)/(o+1.)+(-0.123747277996)+(0.913032880034)*0.5**o
        ref[1,3]=(1.07678754126)/(o+1.)+(-0.152132464655)+(0.44308812765)*0.5**o
        ref[1,4]=(-0.321212426212)/(o+1.)+(0.317913782812)+(-0.898304735153)*0.5**o
        ref[2,0]=(-0.015132627666)/(o+1.)+(0.085518204974)+(0.515308716564)*0.5**o
        ref[2,1]=(-0.18075699461)/(o+1.)+(0.0686806118359)+(0.611196756618)*0.5**o
        ref[2,2]=(-0.481203580509)/(o+1.)+(0.129860652178)+(-0.293576196546)*0.5**o
        ref[2,3]=(0.655085294578)/(o+1.)+(0.195070291341)+(0.673951778483)*0.5**o
        ref[2,4]=(-0.450978469529)/(o+1.)+(-0.00845494494169)+(-0.218899302236)*0.5**o
        ref[3,0]=(0.700015441575)/(o+1.)+(-0.404725030716)+(-0.369727856735)*0.5**o
        ref[3,1]=(1.62292129784)/(o+1.)+(0.48940735807)+(0.444940917816)*0.5**o
        ref[3,2]=(-0.342279006911)/(o+1.)+(0.535107256239)+(-0.82489261483)*0.5**o
        ref[3,3]=(1.29850446055)/(o+1.)+(0.0961361561734)+(0.732863411435)*0.5**o
        ref[3,4]=(0.154492993358)/(o+1.)+(0.055490800047)+(-0.980748139471)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.33788666753)*x[0]**o+(0.0147671449643)*x[0]+(0.14066054662)*x[1]**o+(-0.675231899756)*x[1]
        arg[0,0,1]=(0.0100961441611)*x[0]**o+(0.401304171984)*x[0]+(-0.594786282119)*x[1]**o+(-0.561543495067)*x[1]
        arg[0,1,0]=(-0.118311175162)*x[0]**o+(0.62454992688)*x[0]+(-0.8163725079)*x[1]**o+(-0.756355655277)*x[1]
        arg[0,1,1]=(0.586991709411)*x[0]**o+(-0.600757941586)*x[0]+(0.0953383109011)*x[1]**o+(0.317827629801)*x[1]
        arg[1,0,0]=(-0.251125236294)*x[0]**o+(0.166603616954)*x[0]+(0.978704943828)*x[1]**o+(0.589215389731)*x[1]
        arg[1,0,1]=(0.0543182786738)*x[0]**o+(-0.334756305272)*x[0]+(-0.919571042987)*x[1]**o+(-0.0425734727213)*x[1]
        arg[1,1,0]=(-0.867177621518)*x[0]**o+(-0.0750659593232)*x[0]+(-0.227277029921)*x[1]**o+(0.361120763296)*x[1]
        arg[1,1,1]=(0.85129514889)*x[0]**o+(-0.441962000795)*x[0]+(0.425335668395)*x[1]**o+(0.204091109081)*x[1]
        arg[2,0,0]=(-0.537930401585)*x[0]**o+(-0.206510941636)*x[0]+(-0.247374421196)*x[1]**o+(0.67043841725)*x[1]
        arg[2,0,1]=(-0.101273864474)*x[0]**o+(0.283769727506)*x[0]+(-0.0850059833654)*x[1]**o+(-0.71291661379)*x[1]
        arg[2,1,0]=(0.684430088582)*x[0]**o+(0.107733116086)*x[0]+(0.40398578341)*x[1]**o+(-0.525004991854)*x[1]
        arg[2,1,1]=(-0.00801759014554)*x[0]**o+(-0.214858014851)*x[0]+(-0.427822011518)*x[1]**o+(0.0168962506945)*x[1]
        arg[3,0,0]=(0.528563052182)*x[0]**o+(-0.630485796029)*x[0]+(0.0525765317703)*x[1]**o+(-0.469639006097)*x[1]
        arg[3,0,1]=(0.178622695427)*x[0]**o+(0.0952996500383)*x[0]+(-0.785851578496)*x[1]**o+(0.856665398338)*x[1]
        arg[3,1,0]=(0.547391926072)*x[0]**o+(-0.0179195091852)*x[0]+(-0.831368961862)*x[1]**o+(-0.437229107453)*x[1]
        arg[3,1,1]=(-0.657011566453)*x[0]**o+(0.530176142378)*x[0]+(0.649145894746)*x[1]**o+(-0.755185337598)*x[1]
        arg[4,0,0]=(-0.188305872652)*x[0]**o+(0.0207245930982)*x[0]+(-0.590757998421)*x[1]**o+(-0.793662033952)*x[1]
        arg[4,0,1]=(0.417712824349)*x[0]**o+(-0.138327612527)*x[0]+(-0.381595921766)*x[1]**o+(0.0287368364725)*x[1]
        arg[4,1,0]=(-0.83750540403)*x[0]**o+(0.247261273008)*x[0]+(-0.0769629479034)*x[1]**o+(0.0693501875686)*x[1]
        arg[4,1,1]=(0.697434044074)*x[0]**o+(0.396582403702)*x[0]+(0.736132107034)*x[1]**o+(0.482245183242)*x[1]
        arg[5,0,0]=(0.769936158184)*x[0]**o+(0.808871896459)*x[0]+(-0.559205314785)*x[1]**o+(-0.258014827146)*x[1]
        arg[5,0,1]=(0.667132553969)*x[0]**o+(0.618184492083)*x[0]+(0.475852440157)*x[1]**o+(0.084499293443)*x[1]
        arg[5,1,0]=(0.910882712823)*x[0]**o+(0.153897139268)*x[0]+(-0.61197280039)*x[1]**o+(0.441331985295)*x[1]
        arg[5,1,1]=(-0.290188623857)*x[0]**o+(0.318266320131)*x[0]+(0.9587856247)*x[1]**o+(0.311981486508)*x[1]
        ref[0,0,0]=(0.14066054662)/(o+1.)+(-0.330232377396)+(-0.33788666753)*0.5**o
        ref[0,0,1]=(-0.594786282119)/(o+1.)+(-0.0801196615418)+(0.0100961441611)*0.5**o
        ref[0,1,0]=(-0.8163725079)/(o+1.)+(-0.0659028641985)+(-0.118311175162)*0.5**o
        ref[0,1,1]=(0.0953383109011)/(o+1.)+(-0.141465155893)+(0.586991709411)*0.5**o
        ref[1,0,0]=(0.978704943828)/(o+1.)+(0.377909503343)+(-0.251125236294)*0.5**o
        ref[1,0,1]=(-0.919571042987)/(o+1.)+(-0.188664888997)+(0.0543182786738)*0.5**o
        ref[1,1,0]=(-0.227277029921)/(o+1.)+(0.143027401987)+(-0.867177621518)*0.5**o
        ref[1,1,1]=(0.425335668395)/(o+1.)+(-0.118935445857)+(0.85129514889)*0.5**o
        ref[2,0,0]=(-0.247374421196)/(o+1.)+(0.231963737807)+(-0.537930401585)*0.5**o
        ref[2,0,1]=(-0.0850059833654)/(o+1.)+(-0.214573443142)+(-0.101273864474)*0.5**o
        ref[2,1,0]=(0.40398578341)/(o+1.)+(-0.208635937884)+(0.684430088582)*0.5**o
        ref[2,1,1]=(-0.427822011518)/(o+1.)+(-0.0989808820785)+(-0.00801759014554)*0.5**o
        ref[3,0,0]=(0.0525765317703)/(o+1.)+(-0.550062401063)+(0.528563052182)*0.5**o
        ref[3,0,1]=(-0.785851578496)/(o+1.)+(0.475982524188)+(0.178622695427)*0.5**o
        ref[3,1,0]=(-0.831368961862)/(o+1.)+(-0.227574308319)+(0.547391926072)*0.5**o
        ref[3,1,1]=(0.649145894746)/(o+1.)+(-0.11250459761)+(-0.657011566453)*0.5**o
        ref[4,0,0]=(-0.590757998421)/(o+1.)+(-0.386468720427)+(-0.188305872652)*0.5**o
        ref[4,0,1]=(-0.381595921766)/(o+1.)+(-0.0547953880271)+(0.417712824349)*0.5**o
        ref[4,1,0]=(-0.0769629479034)/(o+1.)+(0.158305730288)+(-0.83750540403)*0.5**o
        ref[4,1,1]=(0.736132107034)/(o+1.)+(0.439413793472)+(0.697434044074)*0.5**o
        ref[5,0,0]=(-0.559205314785)/(o+1.)+(0.275428534657)+(0.769936158184)*0.5**o
        ref[5,0,1]=(0.475852440157)/(o+1.)+(0.351341892763)+(0.667132553969)*0.5**o
        ref[5,1,0]=(-0.61197280039)/(o+1.)+(0.297614562282)+(0.910882712823)*0.5**o
        ref[5,1,1]=(0.9587856247)/(o+1.)+(0.315123903319)+(-0.290188623857)*0.5**o
      else:
        arg[0,0,0]=(0.292346424086)*x[0]**o+(0.105058465241)*x[0]+(-0.281467283372)*x[1]**o+(0.828345267021)*x[1]+(0.741242558638)*x[2]**o+(0.0858780765977)*x[2]
        arg[0,0,1]=(-0.460556250835)*x[0]**o+(-0.845199338719)*x[0]+(-0.547000154821)*x[1]**o+(-0.972236450512)*x[1]+(0.823031420765)*x[2]**o+(0.278285294464)*x[2]
        arg[0,1,0]=(0.380027894487)*x[0]**o+(-0.579359210385)*x[0]+(-0.760849798908)*x[1]**o+(0.0065870459398)*x[1]+(-0.869735580799)*x[2]**o+(-0.937378284335)*x[2]
        arg[0,1,1]=(0.948682597897)*x[0]**o+(-0.882044080711)*x[0]+(-0.694398833899)*x[1]**o+(0.883635235205)*x[1]+(-0.473761903609)*x[2]**o+(0.961453193379)*x[2]
        arg[1,0,0]=(0.357218603696)*x[0]**o+(-0.694504294144)*x[0]+(-0.187119222895)*x[1]**o+(0.727463485088)*x[1]+(0.150410254875)*x[2]**o+(-0.373430202907)*x[2]
        arg[1,0,1]=(-0.776893233002)*x[0]**o+(0.565238637388)*x[0]+(-0.155674628662)*x[1]**o+(-0.938226259914)*x[1]+(0.575183633137)*x[2]**o+(0.319871583783)*x[2]
        arg[1,1,0]=(0.593512911569)*x[0]**o+(-0.0578762085927)*x[0]+(-0.913450847289)*x[1]**o+(0.7074653217)*x[1]+(0.991175636143)*x[2]**o+(-0.535632452281)*x[2]
        arg[1,1,1]=(-0.154542349616)*x[0]**o+(-0.841642050346)*x[0]+(0.878833704431)*x[1]**o+(0.85568524402)*x[1]+(-0.710374336281)*x[2]**o+(0.346540187252)*x[2]
        arg[2,0,0]=(-0.765160864305)*x[0]**o+(0.608017010487)*x[0]+(0.841265469171)*x[1]**o+(-0.409854579584)*x[1]+(0.728931324538)*x[2]**o+(0.121315209213)*x[2]
        arg[2,0,1]=(0.0397067301375)*x[0]**o+(-0.0451925117802)*x[0]+(-0.412724345414)*x[1]**o+(-0.447675002107)*x[1]+(0.141728105561)*x[2]**o+(0.923970189305)*x[2]
        arg[2,1,0]=(-0.237044096923)*x[0]**o+(-0.826168294669)*x[0]+(0.19833520121)*x[1]**o+(-0.796076984382)*x[1]+(0.367399642334)*x[2]**o+(0.71172848801)*x[2]
        arg[2,1,1]=(-0.3881830464)*x[0]**o+(0.988354722657)*x[0]+(0.762756805728)*x[1]**o+(0.551778683619)*x[1]+(0.975932105591)*x[2]**o+(-0.173192098292)*x[2]
        arg[3,0,0]=(-0.226163777371)*x[0]**o+(-0.301094659695)*x[0]+(0.0682567814743)*x[1]**o+(-0.154914902824)*x[1]+(0.984669238436)*x[2]**o+(-0.431011897738)*x[2]
        arg[3,0,1]=(-0.47230847571)*x[0]**o+(-0.579315893476)*x[0]+(-0.989289806396)*x[1]**o+(-0.673951334134)*x[1]+(0.98143767614)*x[2]**o+(-0.311074531383)*x[2]
        arg[3,1,0]=(-0.276309910109)*x[0]**o+(-0.70244304164)*x[0]+(-0.963669800079)*x[1]**o+(-0.947723474952)*x[1]+(0.513451083144)*x[2]**o+(-0.0651803692263)*x[2]
        arg[3,1,1]=(0.65175527234)*x[0]**o+(-0.294569414615)*x[0]+(-0.32846386125)*x[1]**o+(-0.868822318527)*x[1]+(0.268414480892)*x[2]**o+(-0.866835042217)*x[2]
        arg[4,0,0]=(0.304002845434)*x[0]**o+(0.938467246002)*x[0]+(-0.106648262772)*x[1]**o+(0.227863374571)*x[1]+(-0.789305204045)*x[2]**o+(-0.171526142133)*x[2]
        arg[4,0,1]=(0.70982305774)*x[0]**o+(0.404228697442)*x[0]+(0.190029230022)*x[1]**o+(-0.825276083634)*x[1]+(-0.739774636921)*x[2]**o+(-0.726529211281)*x[2]
        arg[4,1,0]=(-0.238487290076)*x[0]**o+(-0.4901234628)*x[0]+(-0.836670757945)*x[1]**o+(0.0841493712661)*x[1]+(-0.685144435198)*x[2]**o+(-0.304978244786)*x[2]
        arg[4,1,1]=(0.264176905985)*x[0]**o+(-0.410582158319)*x[0]+(-0.271915471538)*x[1]**o+(-0.996082427599)*x[1]+(-0.466654446945)*x[2]**o+(-0.723243781429)*x[2]
        arg[5,0,0]=(0.379390860481)*x[0]**o+(-0.78556086955)*x[0]+(0.189126836375)*x[1]**o+(0.235947881164)*x[1]+(0.149142076852)*x[2]**o+(-0.944419890441)*x[2]
        arg[5,0,1]=(0.768867280189)*x[0]**o+(-0.728364777249)*x[0]+(0.60414049451)*x[1]**o+(-0.538876171991)*x[1]+(-0.0947560131508)*x[2]**o+(0.474986163327)*x[2]
        arg[5,1,0]=(-0.228069986734)*x[0]**o+(-0.680841588894)*x[0]+(-0.663806510587)*x[1]**o+(-0.852775021416)*x[1]+(-0.59539582999)*x[2]**o+(0.52369159839)*x[2]
        arg[5,1,1]=(0.177872029425)*x[0]**o+(0.93423742256)*x[0]+(-0.0369475988192)*x[1]**o+(0.242202072343)*x[1]+(0.526144455402)*x[2]**o+(0.0217547142691)*x[2]
        ref[0,0,0]=(0.459775275267)/(o+1.)+(0.50964090443)+(0.292346424086)*0.5**o
        ref[0,0,1]=(0.276031265944)/(o+1.)+(-0.769575247383)+(-0.460556250835)*0.5**o
        ref[0,1,0]=(-1.63058537971)/(o+1.)+(-0.75507522439)+(0.380027894487)*0.5**o
        ref[0,1,1]=(-1.16816073751)/(o+1.)+(0.481522173937)+(0.948682597897)*0.5**o
        ref[1,0,0]=(-0.0367089680209)/(o+1.)+(-0.170235505982)+(0.357218603696)*0.5**o
        ref[1,0,1]=(0.419509004475)/(o+1.)+(-0.0265580193719)+(-0.776893233002)*0.5**o
        ref[1,1,0]=(0.0777247888533)/(o+1.)+(0.0569783304129)+(0.593512911569)*0.5**o
        ref[1,1,1]=(0.168459368149)/(o+1.)+(0.180291690463)+(-0.154542349616)*0.5**o
        ref[2,0,0]=(1.57019679371)/(o+1.)+(0.159738820058)+(-0.765160864305)*0.5**o
        ref[2,0,1]=(-0.270996239853)/(o+1.)+(0.215551337709)+(0.0397067301375)*0.5**o
        ref[2,1,0]=(0.565734843545)/(o+1.)+(-0.455258395521)+(-0.237044096923)*0.5**o
        ref[2,1,1]=(1.73868891132)/(o+1.)+(0.683470653992)+(-0.3881830464)*0.5**o
        ref[3,0,0]=(1.05292601991)/(o+1.)+(-0.443510730129)+(-0.226163777371)*0.5**o
        ref[3,0,1]=(-0.00785213025503)/(o+1.)+(-0.782170879496)+(-0.47230847571)*0.5**o
        ref[3,1,0]=(-0.450218716935)/(o+1.)+(-0.857673442909)+(-0.276309910109)*0.5**o
        ref[3,1,1]=(-0.0600493803575)/(o+1.)+(-1.01511338768)+(0.65175527234)*0.5**o
        ref[4,0,0]=(-0.895953466817)/(o+1.)+(0.49740223922)+(0.304002845434)*0.5**o
        ref[4,0,1]=(-0.549745406899)/(o+1.)+(-0.573788298736)+(0.70982305774)*0.5**o
        ref[4,1,0]=(-1.52181519314)/(o+1.)+(-0.35547616816)+(-0.238487290076)*0.5**o
        ref[4,1,1]=(-0.738569918483)/(o+1.)+(-1.06495418367)+(0.264176905985)*0.5**o
        ref[5,0,0]=(0.338268913228)/(o+1.)+(-0.747016439414)+(0.379390860481)*0.5**o
        ref[5,0,1]=(0.50938448136)/(o+1.)+(-0.396127392956)+(0.768867280189)*0.5**o
        ref[5,1,0]=(-1.25920234058)/(o+1.)+(-0.50496250596)+(-0.228069986734)*0.5**o
        ref[5,1,1]=(0.489196856583)/(o+1.)+(0.599097104586)+(0.177872029425)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.423100205741)*x[0]**o+(-0.681307394941)*x[0]+(-0.848175950825)*x[1]**o+(0.559338235783)*x[1]
        arg[0,0,0,1]=(-0.739687474039)*x[0]**o+(-0.202202358237)*x[0]+(-0.838397517739)*x[1]**o+(-0.377580621185)*x[1]
        arg[0,0,1,0]=(-0.423462958438)*x[0]**o+(-0.335929684088)*x[0]+(-0.616976579504)*x[1]**o+(-0.77803694581)*x[1]
        arg[0,0,1,1]=(-0.432533121428)*x[0]**o+(0.0723865130937)*x[0]+(-0.953014661461)*x[1]**o+(0.92761707252)*x[1]
        arg[0,0,2,0]=(-0.891613775753)*x[0]**o+(0.780592229933)*x[0]+(-0.156191993934)*x[1]**o+(0.985464289737)*x[1]
        arg[0,0,2,1]=(0.909573147888)*x[0]**o+(-0.225603851475)*x[0]+(-0.68287286477)*x[1]**o+(0.330105905794)*x[1]
        arg[0,1,0,0]=(-0.39808363424)*x[0]**o+(-0.290872560992)*x[0]+(-0.631624427826)*x[1]**o+(-0.389239289826)*x[1]
        arg[0,1,0,1]=(-0.748658247224)*x[0]**o+(0.22314629362)*x[0]+(-0.0750972377036)*x[1]**o+(0.908914722147)*x[1]
        arg[0,1,1,0]=(0.877627912148)*x[0]**o+(-0.988434656301)*x[0]+(-0.163646238293)*x[1]**o+(-0.169410711361)*x[1]
        arg[0,1,1,1]=(-0.592580581615)*x[0]**o+(0.997989912544)*x[0]+(0.0667564973731)*x[1]**o+(-0.121580259121)*x[1]
        arg[0,1,2,0]=(-0.505569292541)*x[0]**o+(-0.534428841308)*x[0]+(-0.899276958155)*x[1]**o+(-0.833972415825)*x[1]
        arg[0,1,2,1]=(-0.215752366701)*x[0]**o+(-0.249993262232)*x[0]+(0.912452712698)*x[1]**o+(0.0601524115958)*x[1]
        arg[0,2,0,0]=(0.684112319204)*x[0]**o+(0.343018234122)*x[0]+(0.267409202785)*x[1]**o+(0.7781133929)*x[1]
        arg[0,2,0,1]=(0.552438501301)*x[0]**o+(-0.101558113559)*x[0]+(-0.152472725029)*x[1]**o+(-0.8771654517)*x[1]
        arg[0,2,1,0]=(-0.31213422297)*x[0]**o+(0.597042319279)*x[0]+(-0.873134889407)*x[1]**o+(-0.68920786345)*x[1]
        arg[0,2,1,1]=(0.883048532408)*x[0]**o+(-0.653997880306)*x[0]+(-0.761310589424)*x[1]**o+(-0.698342623962)*x[1]
        arg[0,2,2,0]=(-0.324489237508)*x[0]**o+(-0.717724692716)*x[0]+(-0.936381308418)*x[1]**o+(0.712416971077)*x[1]
        arg[0,2,2,1]=(-0.971735193469)*x[0]**o+(0.0943629094751)*x[0]+(0.419811305063)*x[1]**o+(0.125623316569)*x[1]
        arg[0,3,0,0]=(-0.305553454905)*x[0]**o+(-0.936112994503)*x[0]+(0.707279059071)*x[1]**o+(-0.794754809337)*x[1]
        arg[0,3,0,1]=(0.803934620926)*x[0]**o+(-0.632020033455)*x[0]+(0.819264129099)*x[1]**o+(0.770744069672)*x[1]
        arg[0,3,1,0]=(0.996632326055)*x[0]**o+(-0.537850178404)*x[0]+(-0.60262522048)*x[1]**o+(-0.431667704232)*x[1]
        arg[0,3,1,1]=(0.402799911553)*x[0]**o+(0.155798565063)*x[0]+(0.722837842981)*x[1]**o+(0.462463978227)*x[1]
        arg[0,3,2,0]=(0.937966588808)*x[0]**o+(-0.797151899551)*x[0]+(0.62363567076)*x[1]**o+(-0.60458214868)*x[1]
        arg[0,3,2,1]=(-0.0389713032301)*x[0]**o+(0.497820326557)*x[0]+(-0.950660964237)*x[1]**o+(0.0247372614337)*x[1]
        arg[0,4,0,0]=(-0.589529640529)*x[0]**o+(-0.71761078076)*x[0]+(-0.872148906918)*x[1]**o+(-0.666618835503)*x[1]
        arg[0,4,0,1]=(0.78615039967)*x[0]**o+(0.99287672137)*x[0]+(0.329194160571)*x[1]**o+(0.309370990805)*x[1]
        arg[0,4,1,0]=(-0.801187683466)*x[0]**o+(0.476957978386)*x[0]+(0.107007555422)*x[1]**o+(-0.211909676249)*x[1]
        arg[0,4,1,1]=(0.817551079419)*x[0]**o+(0.958544826763)*x[0]+(-0.720479146591)*x[1]**o+(0.0769824586505)*x[1]
        arg[0,4,2,0]=(-0.245391293539)*x[0]**o+(0.240588876322)*x[0]+(0.766214937295)*x[1]**o+(0.625282150348)*x[1]
        arg[0,4,2,1]=(-0.780163545255)*x[0]**o+(-0.84424180556)*x[0]+(-0.993826355778)*x[1]**o+(-0.708432857746)*x[1]
        arg[1,0,0,0]=(0.0998104654851)*x[0]**o+(-0.933870759023)*x[0]+(-0.00667060999489)*x[1]**o+(-0.304328472241)*x[1]
        arg[1,0,0,1]=(0.716858223574)*x[0]**o+(0.0907083966301)*x[0]+(0.118004968786)*x[1]**o+(-0.989434356769)*x[1]
        arg[1,0,1,0]=(0.433505460082)*x[0]**o+(0.927792393402)*x[0]+(0.10117572965)*x[1]**o+(-0.719175235748)*x[1]
        arg[1,0,1,1]=(0.93394463477)*x[0]**o+(0.30554551436)*x[0]+(-0.221528653572)*x[1]**o+(-0.965153365935)*x[1]
        arg[1,0,2,0]=(0.217926136552)*x[0]**o+(0.75089182057)*x[0]+(0.798438119547)*x[1]**o+(-0.131885152617)*x[1]
        arg[1,0,2,1]=(0.489885370386)*x[0]**o+(-0.958785443386)*x[0]+(-0.934023631551)*x[1]**o+(-0.120727978484)*x[1]
        arg[1,1,0,0]=(-0.104659311537)*x[0]**o+(0.323817119133)*x[0]+(0.207178629335)*x[1]**o+(0.132677138305)*x[1]
        arg[1,1,0,1]=(0.855593354128)*x[0]**o+(-0.319099268106)*x[0]+(-0.75522659926)*x[1]**o+(0.398648791668)*x[1]
        arg[1,1,1,0]=(0.165499959353)*x[0]**o+(0.025416591231)*x[0]+(-0.454711374202)*x[1]**o+(-0.149643847094)*x[1]
        arg[1,1,1,1]=(0.668195333828)*x[0]**o+(-0.954315351102)*x[0]+(-0.15549992542)*x[1]**o+(0.960203842246)*x[1]
        arg[1,1,2,0]=(-0.0830862028828)*x[0]**o+(0.104413343709)*x[0]+(0.0666836881524)*x[1]**o+(-0.909918510114)*x[1]
        arg[1,1,2,1]=(-0.614061685098)*x[0]**o+(-0.696536572273)*x[0]+(0.674867602585)*x[1]**o+(0.742233774001)*x[1]
        arg[1,2,0,0]=(-0.271025278576)*x[0]**o+(0.194146585691)*x[0]+(0.369545867263)*x[1]**o+(-0.444209773003)*x[1]
        arg[1,2,0,1]=(-0.646570016511)*x[0]**o+(-0.1181935343)*x[0]+(-0.473468176784)*x[1]**o+(0.980639666617)*x[1]
        arg[1,2,1,0]=(-0.274558546949)*x[0]**o+(-0.994564294804)*x[0]+(-0.710677848448)*x[1]**o+(-0.471981673909)*x[1]
        arg[1,2,1,1]=(0.920326662139)*x[0]**o+(-0.503477706392)*x[0]+(-0.488075258363)*x[1]**o+(-0.390701293427)*x[1]
        arg[1,2,2,0]=(-0.734779609501)*x[0]**o+(-0.211522964885)*x[0]+(-0.209704158218)*x[1]**o+(0.602143419316)*x[1]
        arg[1,2,2,1]=(-0.71042927509)*x[0]**o+(-0.320801535296)*x[0]+(-0.307645214006)*x[1]**o+(0.902857741355)*x[1]
        arg[1,3,0,0]=(0.430044544461)*x[0]**o+(-0.321809854969)*x[0]+(-0.202705286059)*x[1]**o+(-0.136245903236)*x[1]
        arg[1,3,0,1]=(-0.0215232394148)*x[0]**o+(0.560515199075)*x[0]+(0.183422356152)*x[1]**o+(0.0500950744933)*x[1]
        arg[1,3,1,0]=(0.956072319541)*x[0]**o+(-0.910407504398)*x[0]+(0.710351406673)*x[1]**o+(-0.673149941902)*x[1]
        arg[1,3,1,1]=(0.512105745894)*x[0]**o+(0.617656019287)*x[0]+(0.178878380134)*x[1]**o+(-0.304972517759)*x[1]
        arg[1,3,2,0]=(0.505194196756)*x[0]**o+(-0.588278604691)*x[0]+(0.657312673212)*x[1]**o+(-0.198666173753)*x[1]
        arg[1,3,2,1]=(0.562199962955)*x[0]**o+(-0.855080919048)*x[0]+(0.798106457622)*x[1]**o+(-0.462368594285)*x[1]
        arg[1,4,0,0]=(-0.95103819151)*x[0]**o+(0.996179304819)*x[0]+(0.720809353558)*x[1]**o+(-0.357882371835)*x[1]
        arg[1,4,0,1]=(-0.337212448334)*x[0]**o+(-0.647939092937)*x[0]+(0.432152553594)*x[1]**o+(0.284393585377)*x[1]
        arg[1,4,1,0]=(0.286170290079)*x[0]**o+(0.813168170391)*x[0]+(-0.178814496497)*x[1]**o+(0.414687546663)*x[1]
        arg[1,4,1,1]=(0.00961378883254)*x[0]**o+(-0.946718643529)*x[0]+(0.828780485661)*x[1]**o+(0.54461217692)*x[1]
        arg[1,4,2,0]=(0.784403348421)*x[0]**o+(0.430078584466)*x[0]+(-0.542741131881)*x[1]**o+(0.0290828265624)*x[1]
        arg[1,4,2,1]=(-0.246343667616)*x[0]**o+(-0.27406067089)*x[0]+(-0.938704878102)*x[1]**o+(0.907921504301)*x[1]
        arg[2,0,0,0]=(-0.551056874357)*x[0]**o+(-0.697006980356)*x[0]+(-0.733447125487)*x[1]**o+(-0.241902075189)*x[1]
        arg[2,0,0,1]=(-0.75556701159)*x[0]**o+(0.155303331948)*x[0]+(0.480546881289)*x[1]**o+(-0.420954472057)*x[1]
        arg[2,0,1,0]=(0.12990139506)*x[0]**o+(-0.0990388974263)*x[0]+(-0.913089275462)*x[1]**o+(-0.766316334455)*x[1]
        arg[2,0,1,1]=(-0.96760564527)*x[0]**o+(0.94596820894)*x[0]+(-0.208507053422)*x[1]**o+(0.82642580152)*x[1]
        arg[2,0,2,0]=(-0.801660576339)*x[0]**o+(0.724091208253)*x[0]+(-0.416451086696)*x[1]**o+(-0.255959329575)*x[1]
        arg[2,0,2,1]=(0.380233654687)*x[0]**o+(-0.27467061721)*x[0]+(-0.626861863133)*x[1]**o+(0.498725464891)*x[1]
        arg[2,1,0,0]=(0.733227197081)*x[0]**o+(-0.111622822078)*x[0]+(-0.199363221943)*x[1]**o+(-0.449664634046)*x[1]
        arg[2,1,0,1]=(0.381774304841)*x[0]**o+(0.914208421879)*x[0]+(-0.545958908798)*x[1]**o+(-0.344721917598)*x[1]
        arg[2,1,1,0]=(-0.861412661903)*x[0]**o+(-0.202264578976)*x[0]+(-0.225404313047)*x[1]**o+(0.716189374265)*x[1]
        arg[2,1,1,1]=(-0.180088205315)*x[0]**o+(0.133136657032)*x[0]+(-0.341650807958)*x[1]**o+(-0.570348715256)*x[1]
        arg[2,1,2,0]=(-0.626863150737)*x[0]**o+(-0.236854925766)*x[0]+(-0.271186155613)*x[1]**o+(-0.750500826476)*x[1]
        arg[2,1,2,1]=(-0.138580929054)*x[0]**o+(-0.925057205334)*x[0]+(0.404961686669)*x[1]**o+(0.542298342705)*x[1]
        arg[2,2,0,0]=(0.453186536783)*x[0]**o+(0.368065177204)*x[0]+(0.545211472015)*x[1]**o+(0.334551805933)*x[1]
        arg[2,2,0,1]=(0.0639699038803)*x[0]**o+(0.154192875522)*x[0]+(0.101347900441)*x[1]**o+(-0.668313576443)*x[1]
        arg[2,2,1,0]=(0.465173445732)*x[0]**o+(-0.475009875122)*x[0]+(-0.748444748361)*x[1]**o+(0.620845514745)*x[1]
        arg[2,2,1,1]=(0.720975965615)*x[0]**o+(-0.266631032721)*x[0]+(0.0996389916136)*x[1]**o+(-0.96886977517)*x[1]
        arg[2,2,2,0]=(-0.396391719953)*x[0]**o+(-0.240091626001)*x[0]+(-0.42411685734)*x[1]**o+(-0.458446010754)*x[1]
        arg[2,2,2,1]=(-0.662584035264)*x[0]**o+(0.23813614536)*x[0]+(0.810893160816)*x[1]**o+(-0.130143663026)*x[1]
        arg[2,3,0,0]=(-0.651550480594)*x[0]**o+(0.370844612999)*x[0]+(-0.808002027275)*x[1]**o+(0.531083477622)*x[1]
        arg[2,3,0,1]=(-0.287184182382)*x[0]**o+(0.928877646221)*x[0]+(-0.271563784371)*x[1]**o+(-0.46908309398)*x[1]
        arg[2,3,1,0]=(0.342080138259)*x[0]**o+(0.458994887529)*x[0]+(-0.544475387926)*x[1]**o+(-0.203224249914)*x[1]
        arg[2,3,1,1]=(-0.0791009979206)*x[0]**o+(0.536480501773)*x[0]+(-0.624008015173)*x[1]**o+(-0.380504135174)*x[1]
        arg[2,3,2,0]=(-0.707089360149)*x[0]**o+(0.217716900812)*x[0]+(0.020441009853)*x[1]**o+(0.680214150153)*x[1]
        arg[2,3,2,1]=(0.0368635610611)*x[0]**o+(0.646697307691)*x[0]+(0.36118713723)*x[1]**o+(-0.535295583008)*x[1]
        arg[2,4,0,0]=(0.806938680905)*x[0]**o+(0.449096825134)*x[0]+(-0.43507826254)*x[1]**o+(-0.300956286252)*x[1]
        arg[2,4,0,1]=(-0.268005914326)*x[0]**o+(0.583132947291)*x[0]+(-0.344067495236)*x[1]**o+(-0.526011968282)*x[1]
        arg[2,4,1,0]=(-0.0953663458348)*x[0]**o+(0.834765967023)*x[0]+(-0.0410345097335)*x[1]**o+(0.493198054713)*x[1]
        arg[2,4,1,1]=(0.964777812988)*x[0]**o+(-0.791397009733)*x[0]+(-0.962108127047)*x[1]**o+(0.378083818799)*x[1]
        arg[2,4,2,0]=(0.0977715857241)*x[0]**o+(-0.118265910205)*x[0]+(-0.0181736456148)*x[1]**o+(-0.480499109439)*x[1]
        arg[2,4,2,1]=(0.396936262802)*x[0]**o+(0.979807961416)*x[0]+(0.383251286227)*x[1]**o+(-0.921599329372)*x[1]
        arg[3,0,0,0]=(-0.193869568904)*x[0]**o+(0.335533150532)*x[0]+(-0.542087879921)*x[1]**o+(0.916531699685)*x[1]
        arg[3,0,0,1]=(0.732192438113)*x[0]**o+(0.610866210623)*x[0]+(0.232455584069)*x[1]**o+(0.109577342902)*x[1]
        arg[3,0,1,0]=(0.0487640958479)*x[0]**o+(-0.250229114846)*x[0]+(-0.539343546695)*x[1]**o+(-0.456485153733)*x[1]
        arg[3,0,1,1]=(0.951844805407)*x[0]**o+(-0.811442383281)*x[0]+(-0.537581471096)*x[1]**o+(0.350217063609)*x[1]
        arg[3,0,2,0]=(0.495490307938)*x[0]**o+(0.529667417659)*x[0]+(-0.853447043865)*x[1]**o+(0.97002519322)*x[1]
        arg[3,0,2,1]=(0.38188303469)*x[0]**o+(-0.75911672537)*x[0]+(-0.67563531103)*x[1]**o+(0.384002769031)*x[1]
        arg[3,1,0,0]=(0.44471228922)*x[0]**o+(-0.958995287488)*x[0]+(-0.356730659314)*x[1]**o+(0.853749081247)*x[1]
        arg[3,1,0,1]=(-0.674677862576)*x[0]**o+(-0.651945231545)*x[0]+(-0.869871613483)*x[1]**o+(0.518493292422)*x[1]
        arg[3,1,1,0]=(-0.803923033456)*x[0]**o+(0.916638584514)*x[0]+(-0.410056071233)*x[1]**o+(-0.313811798732)*x[1]
        arg[3,1,1,1]=(-0.677586574312)*x[0]**o+(0.101317771211)*x[0]+(-0.848629733526)*x[1]**o+(0.956935407827)*x[1]
        arg[3,1,2,0]=(0.723956969878)*x[0]**o+(-0.677929268987)*x[0]+(0.168694427833)*x[1]**o+(0.675664564011)*x[1]
        arg[3,1,2,1]=(-0.0961257584415)*x[0]**o+(0.870273004118)*x[0]+(0.0979731186867)*x[1]**o+(-0.561786099655)*x[1]
        arg[3,2,0,0]=(-0.506857925404)*x[0]**o+(0.950429186893)*x[0]+(-0.765686972231)*x[1]**o+(0.300787357163)*x[1]
        arg[3,2,0,1]=(-0.479464705052)*x[0]**o+(-0.814881390688)*x[0]+(-0.549033177495)*x[1]**o+(0.875164713771)*x[1]
        arg[3,2,1,0]=(0.984579775952)*x[0]**o+(0.822464384935)*x[0]+(-0.352937841183)*x[1]**o+(-0.389516730371)*x[1]
        arg[3,2,1,1]=(-0.75230207503)*x[0]**o+(0.921759892886)*x[0]+(0.683120790039)*x[1]**o+(0.470250085193)*x[1]
        arg[3,2,2,0]=(-0.257935272452)*x[0]**o+(-0.437370511818)*x[0]+(-0.350363036315)*x[1]**o+(-0.232415372202)*x[1]
        arg[3,2,2,1]=(0.0784702766009)*x[0]**o+(0.460959683819)*x[0]+(0.758201028024)*x[1]**o+(0.101597092257)*x[1]
        arg[3,3,0,0]=(0.292597181224)*x[0]**o+(-0.322992033991)*x[0]+(-0.536413055541)*x[1]**o+(-0.95609054732)*x[1]
        arg[3,3,0,1]=(0.174134715952)*x[0]**o+(0.811729326606)*x[0]+(0.89040777319)*x[1]**o+(0.893572419966)*x[1]
        arg[3,3,1,0]=(0.768234215847)*x[0]**o+(0.36391719033)*x[0]+(0.598889528719)*x[1]**o+(0.152035960251)*x[1]
        arg[3,3,1,1]=(-0.77938361663)*x[0]**o+(-0.448945619013)*x[0]+(0.471242869055)*x[1]**o+(0.112888214554)*x[1]
        arg[3,3,2,0]=(-0.00617753872016)*x[0]**o+(0.0155218100192)*x[0]+(0.674100983537)*x[1]**o+(0.211406831421)*x[1]
        arg[3,3,2,1]=(0.124470359441)*x[0]**o+(0.117393617733)*x[0]+(-0.297791856337)*x[1]**o+(-0.210405798241)*x[1]
        arg[3,4,0,0]=(0.139920132608)*x[0]**o+(-0.0073381380034)*x[0]+(-0.77267392403)*x[1]**o+(0.792299687107)*x[1]
        arg[3,4,0,1]=(0.328592839797)*x[0]**o+(-0.758899359217)*x[0]+(0.730856047416)*x[1]**o+(-0.389548922357)*x[1]
        arg[3,4,1,0]=(-0.239362896774)*x[0]**o+(0.513205289336)*x[0]+(-0.393781550386)*x[1]**o+(0.229968513732)*x[1]
        arg[3,4,1,1]=(0.777985150139)*x[0]**o+(0.273278793895)*x[0]+(0.17782134987)*x[1]**o+(0.491127331923)*x[1]
        arg[3,4,2,0]=(0.348482781008)*x[0]**o+(0.458696185739)*x[0]+(-0.77882422986)*x[1]**o+(0.455739446004)*x[1]
        arg[3,4,2,1]=(-0.396751338134)*x[0]**o+(0.449026376364)*x[0]+(0.308553967581)*x[1]**o+(-0.777653415416)*x[1]
        ref[0,0,0,0]=(-0.848175950825)/(o+1.)+(-0.060984579579)+(0.423100205741)*0.5**o
        ref[0,0,0,1]=(-0.838397517739)/(o+1.)+(-0.289891489711)+(-0.739687474039)*0.5**o
        ref[0,0,1,0]=(-0.616976579504)/(o+1.)+(-0.556983314949)+(-0.423462958438)*0.5**o
        ref[0,0,1,1]=(-0.953014661461)/(o+1.)+(0.500001792807)+(-0.432533121428)*0.5**o
        ref[0,0,2,0]=(-0.156191993934)/(o+1.)+(0.883028259835)+(-0.891613775753)*0.5**o
        ref[0,0,2,1]=(-0.68287286477)/(o+1.)+(0.0522510271596)+(0.909573147888)*0.5**o
        ref[0,1,0,0]=(-0.631624427826)/(o+1.)+(-0.340055925409)+(-0.39808363424)*0.5**o
        ref[0,1,0,1]=(-0.0750972377036)/(o+1.)+(0.566030507884)+(-0.748658247224)*0.5**o
        ref[0,1,1,0]=(-0.163646238293)/(o+1.)+(-0.578922683831)+(0.877627912148)*0.5**o
        ref[0,1,1,1]=(0.0667564973731)/(o+1.)+(0.438204826712)+(-0.592580581615)*0.5**o
        ref[0,1,2,0]=(-0.899276958155)/(o+1.)+(-0.684200628566)+(-0.505569292541)*0.5**o
        ref[0,1,2,1]=(0.912452712698)/(o+1.)+(-0.0949204253183)+(-0.215752366701)*0.5**o
        ref[0,2,0,0]=(0.267409202785)/(o+1.)+(0.560565813511)+(0.684112319204)*0.5**o
        ref[0,2,0,1]=(-0.152472725029)/(o+1.)+(-0.48936178263)+(0.552438501301)*0.5**o
        ref[0,2,1,0]=(-0.873134889407)/(o+1.)+(-0.0460827720855)+(-0.31213422297)*0.5**o
        ref[0,2,1,1]=(-0.761310589424)/(o+1.)+(-0.676170252134)+(0.883048532408)*0.5**o
        ref[0,2,2,0]=(-0.936381308418)/(o+1.)+(-0.00265386081956)+(-0.324489237508)*0.5**o
        ref[0,2,2,1]=(0.419811305063)/(o+1.)+(0.109993113022)+(-0.971735193469)*0.5**o
        ref[0,3,0,0]=(0.707279059071)/(o+1.)+(-0.86543390192)+(-0.305553454905)*0.5**o
        ref[0,3,0,1]=(0.819264129099)/(o+1.)+(0.0693620181083)+(0.803934620926)*0.5**o
        ref[0,3,1,0]=(-0.60262522048)/(o+1.)+(-0.484758941318)+(0.996632326055)*0.5**o
        ref[0,3,1,1]=(0.722837842981)/(o+1.)+(0.309131271645)+(0.402799911553)*0.5**o
        ref[0,3,2,0]=(0.62363567076)/(o+1.)+(-0.700867024115)+(0.937966588808)*0.5**o
        ref[0,3,2,1]=(-0.950660964237)/(o+1.)+(0.261278793995)+(-0.0389713032301)*0.5**o
        ref[0,4,0,0]=(-0.872148906918)/(o+1.)+(-0.692114808131)+(-0.589529640529)*0.5**o
        ref[0,4,0,1]=(0.329194160571)/(o+1.)+(0.651123856087)+(0.78615039967)*0.5**o
        ref[0,4,1,0]=(0.107007555422)/(o+1.)+(0.132524151069)+(-0.801187683466)*0.5**o
        ref[0,4,1,1]=(-0.720479146591)/(o+1.)+(0.517763642707)+(0.817551079419)*0.5**o
        ref[0,4,2,0]=(0.766214937295)/(o+1.)+(0.432935513335)+(-0.245391293539)*0.5**o
        ref[0,4,2,1]=(-0.993826355778)/(o+1.)+(-0.776337331653)+(-0.780163545255)*0.5**o
        ref[1,0,0,0]=(-0.00667060999489)/(o+1.)+(-0.619099615632)+(0.0998104654851)*0.5**o
        ref[1,0,0,1]=(0.118004968786)/(o+1.)+(-0.449362980069)+(0.716858223574)*0.5**o
        ref[1,0,1,0]=(0.10117572965)/(o+1.)+(0.104308578827)+(0.433505460082)*0.5**o
        ref[1,0,1,1]=(-0.221528653572)/(o+1.)+(-0.329803925788)+(0.93394463477)*0.5**o
        ref[1,0,2,0]=(0.798438119547)/(o+1.)+(0.309503333977)+(0.217926136552)*0.5**o
        ref[1,0,2,1]=(-0.934023631551)/(o+1.)+(-0.539756710935)+(0.489885370386)*0.5**o
        ref[1,1,0,0]=(0.207178629335)/(o+1.)+(0.228247128719)+(-0.104659311537)*0.5**o
        ref[1,1,0,1]=(-0.75522659926)/(o+1.)+(0.0397747617809)+(0.855593354128)*0.5**o
        ref[1,1,1,0]=(-0.454711374202)/(o+1.)+(-0.0621136279316)+(0.165499959353)*0.5**o
        ref[1,1,1,1]=(-0.15549992542)/(o+1.)+(0.00294424557217)+(0.668195333828)*0.5**o
        ref[1,1,2,0]=(0.0666836881524)/(o+1.)+(-0.402752583203)+(-0.0830862028828)*0.5**o
        ref[1,1,2,1]=(0.674867602585)/(o+1.)+(0.0228486008642)+(-0.614061685098)*0.5**o
        ref[1,2,0,0]=(0.369545867263)/(o+1.)+(-0.125031593656)+(-0.271025278576)*0.5**o
        ref[1,2,0,1]=(-0.473468176784)/(o+1.)+(0.431223066159)+(-0.646570016511)*0.5**o
        ref[1,2,1,0]=(-0.710677848448)/(o+1.)+(-0.733272984356)+(-0.274558546949)*0.5**o
        ref[1,2,1,1]=(-0.488075258363)/(o+1.)+(-0.44708949991)+(0.920326662139)*0.5**o
        ref[1,2,2,0]=(-0.209704158218)/(o+1.)+(0.195310227215)+(-0.734779609501)*0.5**o
        ref[1,2,2,1]=(-0.307645214006)/(o+1.)+(0.29102810303)+(-0.71042927509)*0.5**o
        ref[1,3,0,0]=(-0.202705286059)/(o+1.)+(-0.229027879102)+(0.430044544461)*0.5**o
        ref[1,3,0,1]=(0.183422356152)/(o+1.)+(0.305305136784)+(-0.0215232394148)*0.5**o
        ref[1,3,1,0]=(0.710351406673)/(o+1.)+(-0.79177872315)+(0.956072319541)*0.5**o
        ref[1,3,1,1]=(0.178878380134)/(o+1.)+(0.156341750764)+(0.512105745894)*0.5**o
        ref[1,3,2,0]=(0.657312673212)/(o+1.)+(-0.393472389222)+(0.505194196756)*0.5**o
        ref[1,3,2,1]=(0.798106457622)/(o+1.)+(-0.658724756667)+(0.562199962955)*0.5**o
        ref[1,4,0,0]=(0.720809353558)/(o+1.)+(0.319148466492)+(-0.95103819151)*0.5**o
        ref[1,4,0,1]=(0.432152553594)/(o+1.)+(-0.18177275378)+(-0.337212448334)*0.5**o
        ref[1,4,1,0]=(-0.178814496497)/(o+1.)+(0.613927858527)+(0.286170290079)*0.5**o
        ref[1,4,1,1]=(0.828780485661)/(o+1.)+(-0.201053233304)+(0.00961378883254)*0.5**o
        ref[1,4,2,0]=(-0.542741131881)/(o+1.)+(0.229580705514)+(0.784403348421)*0.5**o
        ref[1,4,2,1]=(-0.938704878102)/(o+1.)+(0.316930416706)+(-0.246343667616)*0.5**o
        ref[2,0,0,0]=(-0.733447125487)/(o+1.)+(-0.469454527772)+(-0.551056874357)*0.5**o
        ref[2,0,0,1]=(0.480546881289)/(o+1.)+(-0.132825570055)+(-0.75556701159)*0.5**o
        ref[2,0,1,0]=(-0.913089275462)/(o+1.)+(-0.432677615941)+(0.12990139506)*0.5**o
        ref[2,0,1,1]=(-0.208507053422)/(o+1.)+(0.88619700523)+(-0.96760564527)*0.5**o
        ref[2,0,2,0]=(-0.416451086696)/(o+1.)+(0.234065939339)+(-0.801660576339)*0.5**o
        ref[2,0,2,1]=(-0.626861863133)/(o+1.)+(0.112027423841)+(0.380233654687)*0.5**o
        ref[2,1,0,0]=(-0.199363221943)/(o+1.)+(-0.280643728062)+(0.733227197081)*0.5**o
        ref[2,1,0,1]=(-0.545958908798)/(o+1.)+(0.284743252141)+(0.381774304841)*0.5**o
        ref[2,1,1,0]=(-0.225404313047)/(o+1.)+(0.256962397644)+(-0.861412661903)*0.5**o
        ref[2,1,1,1]=(-0.341650807958)/(o+1.)+(-0.218606029112)+(-0.180088205315)*0.5**o
        ref[2,1,2,0]=(-0.271186155613)/(o+1.)+(-0.493677876121)+(-0.626863150737)*0.5**o
        ref[2,1,2,1]=(0.404961686669)/(o+1.)+(-0.191379431315)+(-0.138580929054)*0.5**o
        ref[2,2,0,0]=(0.545211472015)/(o+1.)+(0.351308491569)+(0.453186536783)*0.5**o
        ref[2,2,0,1]=(0.101347900441)/(o+1.)+(-0.257060350461)+(0.0639699038803)*0.5**o
        ref[2,2,1,0]=(-0.748444748361)/(o+1.)+(0.0729178198114)+(0.465173445732)*0.5**o
        ref[2,2,1,1]=(0.0996389916136)/(o+1.)+(-0.617750403946)+(0.720975965615)*0.5**o
        ref[2,2,2,0]=(-0.42411685734)/(o+1.)+(-0.349268818378)+(-0.396391719953)*0.5**o
        ref[2,2,2,1]=(0.810893160816)/(o+1.)+(0.053996241167)+(-0.662584035264)*0.5**o
        ref[2,3,0,0]=(-0.808002027275)/(o+1.)+(0.45096404531)+(-0.651550480594)*0.5**o
        ref[2,3,0,1]=(-0.271563784371)/(o+1.)+(0.22989727612)+(-0.287184182382)*0.5**o
        ref[2,3,1,0]=(-0.544475387926)/(o+1.)+(0.127885318807)+(0.342080138259)*0.5**o
        ref[2,3,1,1]=(-0.624008015173)/(o+1.)+(0.0779881832997)+(-0.0791009979206)*0.5**o
        ref[2,3,2,0]=(0.020441009853)/(o+1.)+(0.448965525483)+(-0.707089360149)*0.5**o
        ref[2,3,2,1]=(0.36118713723)/(o+1.)+(0.0557008623413)+(0.0368635610611)*0.5**o
        ref[2,4,0,0]=(-0.43507826254)/(o+1.)+(0.0740702694408)+(0.806938680905)*0.5**o
        ref[2,4,0,1]=(-0.344067495236)/(o+1.)+(0.0285604895043)+(-0.268005914326)*0.5**o
        ref[2,4,1,0]=(-0.0410345097335)/(o+1.)+(0.663982010868)+(-0.0953663458348)*0.5**o
        ref[2,4,1,1]=(-0.962108127047)/(o+1.)+(-0.206656595467)+(0.964777812988)*0.5**o
        ref[2,4,2,0]=(-0.0181736456148)/(o+1.)+(-0.299382509822)+(0.0977715857241)*0.5**o
        ref[2,4,2,1]=(0.383251286227)/(o+1.)+(0.0291043160221)+(0.396936262802)*0.5**o
        ref[3,0,0,0]=(-0.542087879921)/(o+1.)+(0.626032425109)+(-0.193869568904)*0.5**o
        ref[3,0,0,1]=(0.232455584069)/(o+1.)+(0.360221776762)+(0.732192438113)*0.5**o
        ref[3,0,1,0]=(-0.539343546695)/(o+1.)+(-0.353357134289)+(0.0487640958479)*0.5**o
        ref[3,0,1,1]=(-0.537581471096)/(o+1.)+(-0.230612659836)+(0.951844805407)*0.5**o
        ref[3,0,2,0]=(-0.853447043865)/(o+1.)+(0.74984630544)+(0.495490307938)*0.5**o
        ref[3,0,2,1]=(-0.67563531103)/(o+1.)+(-0.187556978169)+(0.38188303469)*0.5**o
        ref[3,1,0,0]=(-0.356730659314)/(o+1.)+(-0.0526231031206)+(0.44471228922)*0.5**o
        ref[3,1,0,1]=(-0.869871613483)/(o+1.)+(-0.0667259695618)+(-0.674677862576)*0.5**o
        ref[3,1,1,0]=(-0.410056071233)/(o+1.)+(0.301413392891)+(-0.803923033456)*0.5**o
        ref[3,1,1,1]=(-0.848629733526)/(o+1.)+(0.529126589519)+(-0.677586574312)*0.5**o
        ref[3,1,2,0]=(0.168694427833)/(o+1.)+(-0.00113235248793)+(0.723956969878)*0.5**o
        ref[3,1,2,1]=(0.0979731186867)/(o+1.)+(0.154243452232)+(-0.0961257584415)*0.5**o
        ref[3,2,0,0]=(-0.765686972231)/(o+1.)+(0.625608272028)+(-0.506857925404)*0.5**o
        ref[3,2,0,1]=(-0.549033177495)/(o+1.)+(0.0301416615418)+(-0.479464705052)*0.5**o
        ref[3,2,1,0]=(-0.352937841183)/(o+1.)+(0.216473827282)+(0.984579775952)*0.5**o
        ref[3,2,1,1]=(0.683120790039)/(o+1.)+(0.696004989039)+(-0.75230207503)*0.5**o
        ref[3,2,2,0]=(-0.350363036315)/(o+1.)+(-0.33489294201)+(-0.257935272452)*0.5**o
        ref[3,2,2,1]=(0.758201028024)/(o+1.)+(0.281278388038)+(0.0784702766009)*0.5**o
        ref[3,3,0,0]=(-0.536413055541)/(o+1.)+(-0.639541290655)+(0.292597181224)*0.5**o
        ref[3,3,0,1]=(0.89040777319)/(o+1.)+(0.852650873286)+(0.174134715952)*0.5**o
        ref[3,3,1,0]=(0.598889528719)/(o+1.)+(0.257976575291)+(0.768234215847)*0.5**o
        ref[3,3,1,1]=(0.471242869055)/(o+1.)+(-0.168028702229)+(-0.77938361663)*0.5**o
        ref[3,3,2,0]=(0.674100983537)/(o+1.)+(0.11346432072)+(-0.00617753872016)*0.5**o
        ref[3,3,2,1]=(-0.297791856337)/(o+1.)+(-0.0465060902539)+(0.124470359441)*0.5**o
        ref[3,4,0,0]=(-0.77267392403)/(o+1.)+(0.392480774552)+(0.139920132608)*0.5**o
        ref[3,4,0,1]=(0.730856047416)/(o+1.)+(-0.574224140787)+(0.328592839797)*0.5**o
        ref[3,4,1,0]=(-0.393781550386)/(o+1.)+(0.371586901534)+(-0.239362896774)*0.5**o
        ref[3,4,1,1]=(0.17782134987)/(o+1.)+(0.382203062909)+(0.777985150139)*0.5**o
        ref[3,4,2,0]=(-0.77882422986)/(o+1.)+(0.457217815872)+(0.348482781008)*0.5**o
        ref[3,4,2,1]=(0.308553967581)/(o+1.)+(-0.164313519526)+(-0.396751338134)*0.5**o
      else:
        arg[0,0,0,0]=(0.663998508168)*x[0]**o+(0.797340164931)*x[0]+(0.285182001599)*x[1]**o+(0.99720606542)*x[1]+(0.127610260108)*x[2]**o+(0.0141819362674)*x[2]
        arg[0,0,0,1]=(-0.010130464476)*x[0]**o+(-0.694140348803)*x[0]+(-0.554528445238)*x[1]**o+(0.192639775817)*x[1]+(-0.520164891934)*x[2]**o+(0.316830081989)*x[2]
        arg[0,0,1,0]=(-0.176542289385)*x[0]**o+(-0.312451025451)*x[0]+(0.33951668081)*x[1]**o+(-0.66674551066)*x[1]+(-0.0515929546862)*x[2]**o+(0.688180995863)*x[2]
        arg[0,0,1,1]=(-0.98152572229)*x[0]**o+(0.156411605196)*x[0]+(0.588065240125)*x[1]**o+(0.665830732172)*x[1]+(-0.834625744106)*x[2]**o+(-0.138276534671)*x[2]
        arg[0,0,2,0]=(0.555897422173)*x[0]**o+(0.0269431977874)*x[0]+(0.671161746561)*x[1]**o+(0.0857445285374)*x[1]+(0.578027911522)*x[2]**o+(0.217805007393)*x[2]
        arg[0,0,2,1]=(0.23127040923)*x[0]**o+(0.624353504617)*x[0]+(-0.433915231286)*x[1]**o+(-0.860173971996)*x[1]+(-0.801472434932)*x[2]**o+(-0.779575445432)*x[2]
        arg[0,1,0,0]=(0.0202445177842)*x[0]**o+(-0.0654289514039)*x[0]+(-0.877581892208)*x[1]**o+(-0.231833216639)*x[1]+(0.609752401432)*x[2]**o+(-0.802673983603)*x[2]
        arg[0,1,0,1]=(0.99935098551)*x[0]**o+(0.665639246142)*x[0]+(0.593510187831)*x[1]**o+(0.213336090684)*x[1]+(0.232667772738)*x[2]**o+(-0.651386262633)*x[2]
        arg[0,1,1,0]=(0.151799117419)*x[0]**o+(-0.34631043288)*x[0]+(-0.264898731671)*x[1]**o+(0.697527693137)*x[1]+(0.389030412719)*x[2]**o+(0.613338381577)*x[2]
        arg[0,1,1,1]=(-0.51422532626)*x[0]**o+(0.228040429268)*x[0]+(-0.276760984751)*x[1]**o+(0.379484010198)*x[1]+(0.896837565192)*x[2]**o+(0.195866552401)*x[2]
        arg[0,1,2,0]=(0.908244888839)*x[0]**o+(0.217098236162)*x[0]+(0.798725721675)*x[1]**o+(0.216635630826)*x[1]+(0.960227216319)*x[2]**o+(-0.885381182879)*x[2]
        arg[0,1,2,1]=(-0.872307766634)*x[0]**o+(0.843197737136)*x[0]+(-0.895220198791)*x[1]**o+(0.866433237518)*x[1]+(0.560344834226)*x[2]**o+(0.895797915384)*x[2]
        arg[0,2,0,0]=(0.00173619592035)*x[0]**o+(-0.209651858019)*x[0]+(0.323249367733)*x[1]**o+(-0.848445788048)*x[1]+(-0.747443572813)*x[2]**o+(0.342929929066)*x[2]
        arg[0,2,0,1]=(0.998184336552)*x[0]**o+(-0.669882578313)*x[0]+(-0.632794054696)*x[1]**o+(-0.779832061924)*x[1]+(0.0588815237456)*x[2]**o+(0.419538634163)*x[2]
        arg[0,2,1,0]=(-0.921234331968)*x[0]**o+(0.319724478943)*x[0]+(0.867192523265)*x[1]**o+(-0.183047755109)*x[1]+(-0.753047661507)*x[2]**o+(-0.797179040607)*x[2]
        arg[0,2,1,1]=(0.153578015941)*x[0]**o+(0.887820807812)*x[0]+(-0.607944647299)*x[1]**o+(-0.0487413965527)*x[1]+(-0.812369692315)*x[2]**o+(0.323102876844)*x[2]
        arg[0,2,2,0]=(0.29601674794)*x[0]**o+(-0.108738759628)*x[0]+(0.242837345895)*x[1]**o+(-0.324487554664)*x[1]+(0.584399369463)*x[2]**o+(0.683877139789)*x[2]
        arg[0,2,2,1]=(0.0580007647099)*x[0]**o+(0.214488016057)*x[0]+(0.669087265419)*x[1]**o+(-0.311999860513)*x[1]+(0.678940598195)*x[2]**o+(0.658847203441)*x[2]
        arg[0,3,0,0]=(-0.337999890505)*x[0]**o+(-0.842479530409)*x[0]+(0.786376708623)*x[1]**o+(0.597298234359)*x[1]+(0.495442668997)*x[2]**o+(-0.183828960943)*x[2]
        arg[0,3,0,1]=(-0.825813329629)*x[0]**o+(0.197040670392)*x[0]+(0.997784261354)*x[1]**o+(-0.355290279278)*x[1]+(0.657909516664)*x[2]**o+(-0.568281670001)*x[2]
        arg[0,3,1,0]=(0.699313512079)*x[0]**o+(-0.916350147519)*x[0]+(-0.311605808179)*x[1]**o+(0.222618888987)*x[1]+(0.872150798976)*x[2]**o+(0.139773150125)*x[2]
        arg[0,3,1,1]=(0.140764124286)*x[0]**o+(0.994013777309)*x[0]+(-0.0792107151558)*x[1]**o+(-0.285357856751)*x[1]+(-0.800285107223)*x[2]**o+(0.413029089978)*x[2]
        arg[0,3,2,0]=(-0.68177752688)*x[0]**o+(0.0339522871939)*x[0]+(-0.641399841911)*x[1]**o+(0.263279621719)*x[1]+(-0.406115766497)*x[2]**o+(-0.774639092638)*x[2]
        arg[0,3,2,1]=(0.655685968811)*x[0]**o+(0.20085918281)*x[0]+(-0.753491228829)*x[1]**o+(0.393669658887)*x[1]+(0.339469131428)*x[2]**o+(0.424803724709)*x[2]
        arg[0,4,0,0]=(-0.18027010068)*x[0]**o+(-0.773007683353)*x[0]+(0.767133020895)*x[1]**o+(-0.810449138606)*x[1]+(0.295421955604)*x[2]**o+(-0.955926880008)*x[2]
        arg[0,4,0,1]=(-0.914768182433)*x[0]**o+(-0.67024319498)*x[0]+(0.533244666277)*x[1]**o+(-0.705719628704)*x[1]+(-0.963633923897)*x[2]**o+(-0.0818895784023)*x[2]
        arg[0,4,1,0]=(0.274727423835)*x[0]**o+(0.838218676826)*x[0]+(0.269989900726)*x[1]**o+(0.650600179293)*x[1]+(-0.855210211935)*x[2]**o+(-0.800261683904)*x[2]
        arg[0,4,1,1]=(0.257574026756)*x[0]**o+(-0.787643878696)*x[0]+(-0.619136916008)*x[1]**o+(-0.242580269843)*x[1]+(0.0229959873555)*x[2]**o+(-0.685725009107)*x[2]
        arg[0,4,2,0]=(0.703185665611)*x[0]**o+(0.37686548395)*x[0]+(-0.414901304236)*x[1]**o+(0.310335568741)*x[1]+(-0.740764649572)*x[2]**o+(-0.389356987349)*x[2]
        arg[0,4,2,1]=(-0.153792811673)*x[0]**o+(-0.179777301012)*x[0]+(0.694402537357)*x[1]**o+(-0.928918851291)*x[1]+(-0.439307255129)*x[2]**o+(0.148671231214)*x[2]
        arg[1,0,0,0]=(0.298044833507)*x[0]**o+(0.538000859279)*x[0]+(-0.405432041063)*x[1]**o+(-0.8912548734)*x[1]+(0.929554154255)*x[2]**o+(-0.515905504512)*x[2]
        arg[1,0,0,1]=(-0.575485929806)*x[0]**o+(0.757951009672)*x[0]+(-0.815024122535)*x[1]**o+(0.846519632117)*x[1]+(0.594901198403)*x[2]**o+(-0.672034020351)*x[2]
        arg[1,0,1,0]=(0.998358005707)*x[0]**o+(-0.144729504804)*x[0]+(0.463445062889)*x[1]**o+(0.428918148475)*x[1]+(-0.172684103233)*x[2]**o+(-0.394742387769)*x[2]
        arg[1,0,1,1]=(0.783593367143)*x[0]**o+(-0.269882783638)*x[0]+(-0.880336283806)*x[1]**o+(0.337087191943)*x[1]+(-0.819512478233)*x[2]**o+(0.582924902797)*x[2]
        arg[1,0,2,0]=(-0.42345692756)*x[0]**o+(-0.251910598711)*x[0]+(-0.882193473121)*x[1]**o+(0.535405977525)*x[1]+(-0.117109433242)*x[2]**o+(0.99565775159)*x[2]
        arg[1,0,2,1]=(-0.790917746681)*x[0]**o+(-0.291814540513)*x[0]+(-0.0828604085821)*x[1]**o+(-0.0715382319561)*x[1]+(-0.188453557262)*x[2]**o+(0.297596604241)*x[2]
        arg[1,1,0,0]=(0.944343622558)*x[0]**o+(0.537893843235)*x[0]+(-0.370697218599)*x[1]**o+(-0.043709074747)*x[1]+(-0.803318845377)*x[2]**o+(0.814392946479)*x[2]
        arg[1,1,0,1]=(-0.246711023019)*x[0]**o+(-0.847902406307)*x[0]+(0.062135604814)*x[1]**o+(0.581341495124)*x[1]+(0.561484848132)*x[2]**o+(0.06820017559)*x[2]
        arg[1,1,1,0]=(-0.530782063061)*x[0]**o+(-0.951246029994)*x[0]+(-0.809647094351)*x[1]**o+(0.198165508669)*x[1]+(0.591475451485)*x[2]**o+(-0.0567469904903)*x[2]
        arg[1,1,1,1]=(-0.710844761085)*x[0]**o+(0.742146961382)*x[0]+(0.482950602277)*x[1]**o+(-0.701576649521)*x[1]+(0.752305355358)*x[2]**o+(0.658478074845)*x[2]
        arg[1,1,2,0]=(0.574657505707)*x[0]**o+(-0.268094098242)*x[0]+(-0.79745549318)*x[1]**o+(0.994732385826)*x[1]+(0.36752110393)*x[2]**o+(0.0733521344978)*x[2]
        arg[1,1,2,1]=(-0.365373199615)*x[0]**o+(0.295532428151)*x[0]+(-0.916049089517)*x[1]**o+(0.562876895658)*x[1]+(0.698818392781)*x[2]**o+(0.413166153315)*x[2]
        arg[1,2,0,0]=(0.544776615353)*x[0]**o+(-0.647889580869)*x[0]+(-0.0569263298125)*x[1]**o+(0.145083519829)*x[1]+(-0.0456368749942)*x[2]**o+(0.125915481669)*x[2]
        arg[1,2,0,1]=(0.11684721915)*x[0]**o+(-0.793592672443)*x[0]+(-0.428585283805)*x[1]**o+(0.280358391066)*x[1]+(0.72731741847)*x[2]**o+(-0.320459136831)*x[2]
        arg[1,2,1,0]=(0.561736948232)*x[0]**o+(0.833411538285)*x[0]+(0.361519222525)*x[1]**o+(-0.785948355356)*x[1]+(-0.595938528895)*x[2]**o+(0.720247127573)*x[2]
        arg[1,2,1,1]=(0.578630452142)*x[0]**o+(0.218238639275)*x[0]+(0.792496779641)*x[1]**o+(-0.818431209813)*x[1]+(-0.252522807953)*x[2]**o+(0.648505893606)*x[2]
        arg[1,2,2,0]=(-0.0269203575755)*x[0]**o+(-0.586454813205)*x[0]+(-0.239885637578)*x[1]**o+(0.186263805313)*x[1]+(0.585909451795)*x[2]**o+(-0.630731036615)*x[2]
        arg[1,2,2,1]=(-0.534642633404)*x[0]**o+(0.921034930282)*x[0]+(-0.326672365556)*x[1]**o+(0.690509146367)*x[1]+(-0.802567575176)*x[2]**o+(-0.235903818924)*x[2]
        arg[1,3,0,0]=(0.892302802156)*x[0]**o+(-0.528836075071)*x[0]+(-0.806520835766)*x[1]**o+(0.771321197577)*x[1]+(-0.545545197249)*x[2]**o+(-0.633185919719)*x[2]
        arg[1,3,0,1]=(0.883702447985)*x[0]**o+(0.698176356191)*x[0]+(-0.348189466787)*x[1]**o+(-0.390058291315)*x[1]+(0.805736008481)*x[2]**o+(0.861787745203)*x[2]
        arg[1,3,1,0]=(0.492527250981)*x[0]**o+(0.635712008949)*x[0]+(-0.464397346193)*x[1]**o+(-0.820952575379)*x[1]+(0.235015001164)*x[2]**o+(-0.846160845231)*x[2]
        arg[1,3,1,1]=(0.372499842901)*x[0]**o+(-0.524824414377)*x[0]+(0.553450359486)*x[1]**o+(-0.187372498756)*x[1]+(-0.389983195918)*x[2]**o+(-0.310421627605)*x[2]
        arg[1,3,2,0]=(0.533098362924)*x[0]**o+(0.917820078197)*x[0]+(-0.557683070597)*x[1]**o+(-0.388498248383)*x[1]+(-0.551866333066)*x[2]**o+(0.696873468511)*x[2]
        arg[1,3,2,1]=(-0.909038123273)*x[0]**o+(0.19323884412)*x[0]+(-0.00134861813725)*x[1]**o+(-0.858193032021)*x[1]+(0.0436856504635)*x[2]**o+(-0.646994265986)*x[2]
        arg[1,4,0,0]=(-0.974474210185)*x[0]**o+(-0.39786371112)*x[0]+(-0.829024534519)*x[1]**o+(0.836151642189)*x[1]+(0.679654655991)*x[2]**o+(0.750995377779)*x[2]
        arg[1,4,0,1]=(0.964745405431)*x[0]**o+(0.269189443297)*x[0]+(0.531636623506)*x[1]**o+(-0.522687387383)*x[1]+(0.463500385567)*x[2]**o+(0.160556224359)*x[2]
        arg[1,4,1,0]=(-0.714725894011)*x[0]**o+(0.0478807291181)*x[0]+(0.591494552904)*x[1]**o+(-0.240293119817)*x[1]+(-0.0463360360003)*x[2]**o+(-0.932835908096)*x[2]
        arg[1,4,1,1]=(-0.809072545597)*x[0]**o+(-0.963335650229)*x[0]+(-0.781885522838)*x[1]**o+(0.822395174213)*x[1]+(-0.42590899181)*x[2]**o+(-0.237357227214)*x[2]
        arg[1,4,2,0]=(-0.21281876112)*x[0]**o+(-0.131640536998)*x[0]+(-0.648680371617)*x[1]**o+(0.555759774031)*x[1]+(-0.911030244394)*x[2]**o+(0.962859403252)*x[2]
        arg[1,4,2,1]=(-0.247930702516)*x[0]**o+(0.509829884687)*x[0]+(0.0387816417333)*x[1]**o+(-0.605945140743)*x[1]+(0.187880415018)*x[2]**o+(-0.813242364046)*x[2]
        arg[2,0,0,0]=(0.616352869937)*x[0]**o+(-0.930708209966)*x[0]+(-0.874778661809)*x[1]**o+(-0.330225154904)*x[1]+(0.632020336392)*x[2]**o+(0.205631598228)*x[2]
        arg[2,0,0,1]=(-0.85704374519)*x[0]**o+(0.99958496465)*x[0]+(0.807095767473)*x[1]**o+(0.63097388144)*x[1]+(0.000665614692828)*x[2]**o+(0.418070858864)*x[2]
        arg[2,0,1,0]=(0.326870102902)*x[0]**o+(0.565855791221)*x[0]+(-0.39418644908)*x[1]**o+(-0.96081735968)*x[1]+(0.937255293725)*x[2]**o+(0.533015538451)*x[2]
        arg[2,0,1,1]=(-0.699939842176)*x[0]**o+(0.808055402204)*x[0]+(0.448948790025)*x[1]**o+(-0.720122241439)*x[1]+(0.0532200858319)*x[2]**o+(-0.204851171261)*x[2]
        arg[2,0,2,0]=(-0.985996827188)*x[0]**o+(-0.604602591167)*x[0]+(0.445903925662)*x[1]**o+(-0.163135114049)*x[1]+(-0.264281316158)*x[2]**o+(-0.381814212863)*x[2]
        arg[2,0,2,1]=(-0.195357773436)*x[0]**o+(0.73285122682)*x[0]+(-0.914846292757)*x[1]**o+(-0.190524885095)*x[1]+(-0.103440307801)*x[2]**o+(-0.0309647945217)*x[2]
        arg[2,1,0,0]=(0.601652939329)*x[0]**o+(-0.757171386065)*x[0]+(-0.91147090556)*x[1]**o+(-0.665258414941)*x[1]+(-0.0504245123191)*x[2]**o+(-0.274928427711)*x[2]
        arg[2,1,0,1]=(0.534038022604)*x[0]**o+(-0.977365399298)*x[0]+(0.884656682614)*x[1]**o+(0.0866421048629)*x[1]+(-0.70687555879)*x[2]**o+(-0.99928465036)*x[2]
        arg[2,1,1,0]=(-0.361196978257)*x[0]**o+(0.192559020092)*x[0]+(-0.0767715201942)*x[1]**o+(0.222254953069)*x[1]+(-0.128362555436)*x[2]**o+(-0.667485829139)*x[2]
        arg[2,1,1,1]=(0.00930987921229)*x[0]**o+(0.269207633273)*x[0]+(0.943798095916)*x[1]**o+(0.483123564952)*x[1]+(-0.231669365037)*x[2]**o+(-0.877005245489)*x[2]
        arg[2,1,2,0]=(-0.181479246679)*x[0]**o+(0.619551810132)*x[0]+(0.499400298722)*x[1]**o+(-0.307864349855)*x[1]+(0.19303040809)*x[2]**o+(0.523640816928)*x[2]
        arg[2,1,2,1]=(-0.99665050175)*x[0]**o+(-0.647775674514)*x[0]+(-0.0015481673821)*x[1]**o+(-0.749198451256)*x[1]+(0.652528193579)*x[2]**o+(-0.0194127861341)*x[2]
        arg[2,2,0,0]=(0.870069083779)*x[0]**o+(0.179747121592)*x[0]+(0.387523112512)*x[1]**o+(0.625711925024)*x[1]+(0.840141938096)*x[2]**o+(0.629709976556)*x[2]
        arg[2,2,0,1]=(-0.841259995642)*x[0]**o+(0.220858862445)*x[0]+(0.397334296484)*x[1]**o+(-0.036654179636)*x[1]+(0.235704322138)*x[2]**o+(-0.6299228864)*x[2]
        arg[2,2,1,0]=(-0.561184923307)*x[0]**o+(-0.643644162376)*x[0]+(0.728784994685)*x[1]**o+(-0.697418070844)*x[1]+(-0.56166463788)*x[2]**o+(-0.668760084534)*x[2]
        arg[2,2,1,1]=(0.92243000401)*x[0]**o+(0.96165833182)*x[0]+(0.540379002453)*x[1]**o+(0.630245078647)*x[1]+(-0.375873656541)*x[2]**o+(0.71424898756)*x[2]
        arg[2,2,2,0]=(0.331131296272)*x[0]**o+(-0.136167378345)*x[0]+(-0.375407637025)*x[1]**o+(-0.0890691004072)*x[1]+(-0.787306129385)*x[2]**o+(-0.447474749986)*x[2]
        arg[2,2,2,1]=(-0.47892171854)*x[0]**o+(0.341448171328)*x[0]+(0.662403724393)*x[1]**o+(0.581128263913)*x[1]+(0.00987835902932)*x[2]**o+(-0.884847129103)*x[2]
        arg[2,3,0,0]=(0.0219406022052)*x[0]**o+(-0.370001982912)*x[0]+(0.948162807268)*x[1]**o+(-0.870857624404)*x[1]+(0.580053532448)*x[2]**o+(-0.488657040041)*x[2]
        arg[2,3,0,1]=(-0.114799511617)*x[0]**o+(0.547853919356)*x[0]+(-0.224367707174)*x[1]**o+(0.551491902831)*x[1]+(-0.101845907889)*x[2]**o+(-0.793527438395)*x[2]
        arg[2,3,1,0]=(-0.103518223905)*x[0]**o+(-0.319840011093)*x[0]+(-0.856612861717)*x[1]**o+(0.0208578084364)*x[1]+(0.132248308449)*x[2]**o+(-0.580663554499)*x[2]
        arg[2,3,1,1]=(-0.958142224242)*x[0]**o+(-0.401299403423)*x[0]+(-0.289034019185)*x[1]**o+(-0.0913958572584)*x[1]+(-0.818515945817)*x[2]**o+(-0.918366425737)*x[2]
        arg[2,3,2,0]=(0.730554594734)*x[0]**o+(0.209964841318)*x[0]+(-0.4157212696)*x[1]**o+(-0.673375403124)*x[1]+(-0.409899056244)*x[2]**o+(0.124418003637)*x[2]
        arg[2,3,2,1]=(-0.930974589897)*x[0]**o+(0.256452400333)*x[0]+(0.778606258675)*x[1]**o+(0.23173182074)*x[1]+(-0.846439682798)*x[2]**o+(-0.408860675089)*x[2]
        arg[2,4,0,0]=(-0.246046780287)*x[0]**o+(-0.653766430943)*x[0]+(0.725379461155)*x[1]**o+(-0.305490823497)*x[1]+(0.0837637468427)*x[2]**o+(-0.458770631133)*x[2]
        arg[2,4,0,1]=(-0.612751147889)*x[0]**o+(0.504418327803)*x[0]+(-0.835051783805)*x[1]**o+(-0.86410559166)*x[1]+(0.928919612861)*x[2]**o+(0.303164594248)*x[2]
        arg[2,4,1,0]=(-0.26092972064)*x[0]**o+(-0.990592356395)*x[0]+(-0.48716443965)*x[1]**o+(0.659447608304)*x[1]+(0.429865593656)*x[2]**o+(-0.661470437053)*x[2]
        arg[2,4,1,1]=(0.325840618456)*x[0]**o+(-0.261974904995)*x[0]+(0.476982730267)*x[1]**o+(-0.738119182885)*x[1]+(0.903618858571)*x[2]**o+(-0.287442022624)*x[2]
        arg[2,4,2,0]=(0.655590135983)*x[0]**o+(-0.135505327297)*x[0]+(-0.38079131346)*x[1]**o+(0.663850751769)*x[1]+(-0.658568109867)*x[2]**o+(-0.267875066063)*x[2]
        arg[2,4,2,1]=(0.570712162768)*x[0]**o+(0.390833274898)*x[0]+(-0.97326103687)*x[1]**o+(-0.419300688467)*x[1]+(0.55695479018)*x[2]**o+(0.547271407555)*x[2]
        arg[3,0,0,0]=(-0.486023722572)*x[0]**o+(-0.790544923201)*x[0]+(0.101237181961)*x[1]**o+(0.23267779263)*x[1]+(-0.458580701132)*x[2]**o+(-0.479449733139)*x[2]
        arg[3,0,0,1]=(-0.705291625776)*x[0]**o+(0.144166702405)*x[0]+(-0.210596631253)*x[1]**o+(0.843721258425)*x[1]+(-0.716420627484)*x[2]**o+(0.145427688875)*x[2]
        arg[3,0,1,0]=(-0.838974120992)*x[0]**o+(0.824423674933)*x[0]+(-0.211643366873)*x[1]**o+(0.603255736693)*x[1]+(-0.119484817371)*x[2]**o+(-0.357272543813)*x[2]
        arg[3,0,1,1]=(0.140027541071)*x[0]**o+(0.373598174368)*x[0]+(-0.661831753147)*x[1]**o+(-0.973724961653)*x[1]+(-0.0738899100301)*x[2]**o+(0.0128579666074)*x[2]
        arg[3,0,2,0]=(-0.39083471939)*x[0]**o+(-0.00821070494132)*x[0]+(-0.315936839906)*x[1]**o+(-0.967556350778)*x[1]+(0.136991762476)*x[2]**o+(0.374041520875)*x[2]
        arg[3,0,2,1]=(0.0734743009776)*x[0]**o+(-0.103186084491)*x[0]+(0.426933353985)*x[1]**o+(-0.631917740007)*x[1]+(0.755296770294)*x[2]**o+(0.492171872216)*x[2]
        arg[3,1,0,0]=(-0.414502599055)*x[0]**o+(0.512031410722)*x[0]+(0.610677049483)*x[1]**o+(-0.075275839509)*x[1]+(0.853407100545)*x[2]**o+(-0.569731924126)*x[2]
        arg[3,1,0,1]=(0.886115095396)*x[0]**o+(0.811807969433)*x[0]+(0.550623354147)*x[1]**o+(-0.703728894875)*x[1]+(0.784318903356)*x[2]**o+(0.45891829089)*x[2]
        arg[3,1,1,0]=(0.312577129879)*x[0]**o+(0.0127110992422)*x[0]+(-0.174614340667)*x[1]**o+(0.744535847679)*x[1]+(0.991125452344)*x[2]**o+(-0.405595301834)*x[2]
        arg[3,1,1,1]=(0.784369258288)*x[0]**o+(-0.709503832235)*x[0]+(0.032963558933)*x[1]**o+(-0.471036543959)*x[1]+(-0.138022967476)*x[2]**o+(-0.38452297541)*x[2]
        arg[3,1,2,0]=(-0.337332295082)*x[0]**o+(-0.460457867002)*x[0]+(-0.102806204559)*x[1]**o+(-0.502889108364)*x[1]+(-0.641488011668)*x[2]**o+(0.359501036527)*x[2]
        arg[3,1,2,1]=(0.791575892417)*x[0]**o+(0.126853712619)*x[0]+(-0.408911942637)*x[1]**o+(0.557039825987)*x[1]+(-0.260719710621)*x[2]**o+(-0.648326395935)*x[2]
        arg[3,2,0,0]=(-0.0909007497318)*x[0]**o+(-0.337085519884)*x[0]+(-0.651987922715)*x[1]**o+(-0.130225558042)*x[1]+(0.926567188664)*x[2]**o+(0.764369275478)*x[2]
        arg[3,2,0,1]=(-0.764137934896)*x[0]**o+(0.443892891832)*x[0]+(0.688197384487)*x[1]**o+(0.485210058592)*x[1]+(0.267932741224)*x[2]**o+(0.0366064885582)*x[2]
        arg[3,2,1,0]=(0.49435832562)*x[0]**o+(0.860649583676)*x[0]+(0.667415091305)*x[1]**o+(0.144069385904)*x[1]+(-0.933847080974)*x[2]**o+(-0.378123304101)*x[2]
        arg[3,2,1,1]=(-0.460752644946)*x[0]**o+(0.491079753273)*x[0]+(-0.536055633506)*x[1]**o+(0.786378615881)*x[1]+(-0.190647453518)*x[2]**o+(-0.192868019034)*x[2]
        arg[3,2,2,0]=(0.0946687985697)*x[0]**o+(0.978197346579)*x[0]+(0.972747948856)*x[1]**o+(-0.575670550665)*x[1]+(-0.096229657916)*x[2]**o+(-0.772567430449)*x[2]
        arg[3,2,2,1]=(0.177521713734)*x[0]**o+(0.0683685252148)*x[0]+(0.423995112682)*x[1]**o+(-0.489247981968)*x[1]+(-0.537130434099)*x[2]**o+(-0.0228879131449)*x[2]
        arg[3,3,0,0]=(0.592201472175)*x[0]**o+(0.674304504132)*x[0]+(0.428414872879)*x[1]**o+(0.956414209094)*x[1]+(-0.272853242491)*x[2]**o+(-0.639858910901)*x[2]
        arg[3,3,0,1]=(0.572972211447)*x[0]**o+(0.943096858489)*x[0]+(-0.990948797302)*x[1]**o+(-0.475383526007)*x[1]+(0.542076985363)*x[2]**o+(-0.884809421862)*x[2]
        arg[3,3,1,0]=(0.775163727164)*x[0]**o+(0.101808074039)*x[0]+(0.575490330857)*x[1]**o+(0.199751068512)*x[1]+(0.538742155685)*x[2]**o+(-0.725822956214)*x[2]
        arg[3,3,1,1]=(-0.244586606882)*x[0]**o+(0.983712254807)*x[0]+(0.187370538052)*x[1]**o+(0.332611385073)*x[1]+(0.536377667373)*x[2]**o+(0.255272558489)*x[2]
        arg[3,3,2,0]=(-0.676040282048)*x[0]**o+(-0.172059284021)*x[0]+(0.25100451419)*x[1]**o+(0.303226061816)*x[1]+(0.138976177556)*x[2]**o+(-0.330104849626)*x[2]
        arg[3,3,2,1]=(-0.136437682299)*x[0]**o+(0.938786085213)*x[0]+(0.853018944962)*x[1]**o+(-0.628078903025)*x[1]+(-0.903567849919)*x[2]**o+(0.837621994591)*x[2]
        arg[3,4,0,0]=(0.28632587806)*x[0]**o+(-0.480626263522)*x[0]+(-0.0537965100699)*x[1]**o+(0.423252432505)*x[1]+(0.200777153039)*x[2]**o+(0.629713098071)*x[2]
        arg[3,4,0,1]=(0.326448964148)*x[0]**o+(0.809991245673)*x[0]+(-0.527362539073)*x[1]**o+(-0.0727890632651)*x[1]+(0.210806278508)*x[2]**o+(0.736535063786)*x[2]
        arg[3,4,1,0]=(0.760474271531)*x[0]**o+(0.378083396435)*x[0]+(-0.991296179693)*x[1]**o+(0.431965391269)*x[1]+(-0.12597704605)*x[2]**o+(-0.128017966575)*x[2]
        arg[3,4,1,1]=(-0.21508452109)*x[0]**o+(-0.583147117492)*x[0]+(0.779535489322)*x[1]**o+(0.601144982341)*x[1]+(-0.870362610456)*x[2]**o+(-0.318535391517)*x[2]
        arg[3,4,2,0]=(-0.195181632154)*x[0]**o+(0.237570596161)*x[0]+(-0.794725125363)*x[1]**o+(0.862030507166)*x[1]+(-0.612206533195)*x[2]**o+(0.494078406766)*x[2]
        arg[3,4,2,1]=(0.270278273932)*x[0]**o+(-0.115833233113)*x[0]+(0.754464364285)*x[1]**o+(-0.186069521903)*x[1]+(-0.864426716229)*x[2]**o+(-0.425054438064)*x[2]
        ref[0,0,0,0]=(0.412792261707)/(o+1.)+(0.904364083309)+(0.663998508168)*0.5**o
        ref[0,0,0,1]=(-1.07469333717)/(o+1.)+(-0.0923352454984)+(-0.010130464476)*0.5**o
        ref[0,0,1,0]=(0.287923726124)/(o+1.)+(-0.145507770124)+(-0.176542289385)*0.5**o
        ref[0,0,1,1]=(-0.24656050398)/(o+1.)+(0.341982901349)+(-0.98152572229)*0.5**o
        ref[0,0,2,0]=(1.24918965808)/(o+1.)+(0.165246366859)+(0.555897422173)*0.5**o
        ref[0,0,2,1]=(-1.23538766622)/(o+1.)+(-0.507697956406)+(0.23127040923)*0.5**o
        ref[0,1,0,0]=(-0.267829490776)/(o+1.)+(-0.549968075823)+(0.0202445177842)*0.5**o
        ref[0,1,0,1]=(0.826177960569)/(o+1.)+(0.113794537096)+(0.99935098551)*0.5**o
        ref[0,1,1,0]=(0.124131681048)/(o+1.)+(0.482277820918)+(0.151799117419)*0.5**o
        ref[0,1,1,1]=(0.620076580442)/(o+1.)+(0.401695495934)+(-0.51422532626)*0.5**o
        ref[0,1,2,0]=(1.75895293799)/(o+1.)+(-0.225823657945)+(0.908244888839)*0.5**o
        ref[0,1,2,1]=(-0.334875364565)/(o+1.)+(1.30271444502)+(-0.872307766634)*0.5**o
        ref[0,2,0,0]=(-0.42419420508)/(o+1.)+(-0.3575838585)+(0.00173619592035)*0.5**o
        ref[0,2,0,1]=(-0.57391253095)/(o+1.)+(-0.515088003037)+(0.998184336552)*0.5**o
        ref[0,2,1,0]=(0.114144861758)/(o+1.)+(-0.330251158386)+(-0.921234331968)*0.5**o
        ref[0,2,1,1]=(-1.42031433961)/(o+1.)+(0.581091144051)+(0.153578015941)*0.5**o
        ref[0,2,2,0]=(0.827236715358)/(o+1.)+(0.125325412749)+(0.29601674794)*0.5**o
        ref[0,2,2,1]=(1.34802786361)/(o+1.)+(0.280667679493)+(0.0580007647099)*0.5**o
        ref[0,3,0,0]=(1.28181937762)/(o+1.)+(-0.214505128496)+(-0.337999890505)*0.5**o
        ref[0,3,0,1]=(1.65569377802)/(o+1.)+(-0.363265639443)+(-0.825813329629)*0.5**o
        ref[0,3,1,0]=(0.560544990797)/(o+1.)+(-0.276979054204)+(0.699313512079)*0.5**o
        ref[0,3,1,1]=(-0.879495822379)/(o+1.)+(0.560842505268)+(0.140764124286)*0.5**o
        ref[0,3,2,0]=(-1.04751560841)/(o+1.)+(-0.238703591863)+(-0.68177752688)*0.5**o
        ref[0,3,2,1]=(-0.414022097401)/(o+1.)+(0.509666283203)+(0.655685968811)*0.5**o
        ref[0,4,0,0]=(1.0625549765)/(o+1.)+(-1.26969185098)+(-0.18027010068)*0.5**o
        ref[0,4,0,1]=(-0.430389257621)/(o+1.)+(-0.728926201043)+(-0.914768182433)*0.5**o
        ref[0,4,1,0]=(-0.585220311209)/(o+1.)+(0.344278586108)+(0.274727423835)*0.5**o
        ref[0,4,1,1]=(-0.596140928652)/(o+1.)+(-0.857974578823)+(0.257574026756)*0.5**o
        ref[0,4,2,0]=(-1.15566595381)/(o+1.)+(0.148922032671)+(0.703185665611)*0.5**o
        ref[0,4,2,1]=(0.255095282229)/(o+1.)+(-0.480012460544)+(-0.153792811673)*0.5**o
        ref[1,0,0,0]=(0.524122113192)/(o+1.)+(-0.434579759316)+(0.298044833507)*0.5**o
        ref[1,0,0,1]=(-0.220122924131)/(o+1.)+(0.466218310719)+(-0.575485929806)*0.5**o
        ref[1,0,1,0]=(0.290760959656)/(o+1.)+(-0.0552768720487)+(0.998358005707)*0.5**o
        ref[1,0,1,1]=(-1.69984876204)/(o+1.)+(0.325064655551)+(0.783593367143)*0.5**o
        ref[1,0,2,0]=(-0.999302906363)/(o+1.)+(0.639576565202)+(-0.42345692756)*0.5**o
        ref[1,0,2,1]=(-0.271313965844)/(o+1.)+(-0.0328780841139)+(-0.790917746681)*0.5**o
        ref[1,1,0,0]=(-1.17401606398)/(o+1.)+(0.654288857484)+(0.944343622558)*0.5**o
        ref[1,1,0,1]=(0.623620452946)/(o+1.)+(-0.0991803677964)+(-0.246711023019)*0.5**o
        ref[1,1,1,0]=(-0.218171642866)/(o+1.)+(-0.404913755908)+(-0.530782063061)*0.5**o
        ref[1,1,1,1]=(1.23525595763)/(o+1.)+(0.349524193353)+(-0.710844761085)*0.5**o
        ref[1,1,2,0]=(-0.42993438925)/(o+1.)+(0.399995211041)+(0.574657505707)*0.5**o
        ref[1,1,2,1]=(-0.217230696736)/(o+1.)+(0.635787738562)+(-0.365373199615)*0.5**o
        ref[1,2,0,0]=(-0.102563204807)/(o+1.)+(-0.188445289686)+(0.544776615353)*0.5**o
        ref[1,2,0,1]=(0.298732134665)/(o+1.)+(-0.416846709104)+(0.11684721915)*0.5**o
        ref[1,2,1,0]=(-0.234419306371)/(o+1.)+(0.383855155251)+(0.561736948232)*0.5**o
        ref[1,2,1,1]=(0.539973971687)/(o+1.)+(0.0241566615339)+(0.578630452142)*0.5**o
        ref[1,2,2,0]=(0.346023814217)/(o+1.)+(-0.515461022254)+(-0.0269203575755)*0.5**o
        ref[1,2,2,1]=(-1.12923994073)/(o+1.)+(0.687820128862)+(-0.534642633404)*0.5**o
        ref[1,3,0,0]=(-1.35206603301)/(o+1.)+(-0.195350398606)+(0.892302802156)*0.5**o
        ref[1,3,0,1]=(0.457546541694)/(o+1.)+(0.584952905039)+(0.883702447985)*0.5**o
        ref[1,3,1,0]=(-0.229382345029)/(o+1.)+(-0.51570070583)+(0.492527250981)*0.5**o
        ref[1,3,1,1]=(0.163467163568)/(o+1.)+(-0.511309270369)+(0.372499842901)*0.5**o
        ref[1,3,2,0]=(-1.10954940366)/(o+1.)+(0.613097649163)+(0.533098362924)*0.5**o
        ref[1,3,2,1]=(0.0423370323263)/(o+1.)+(-0.655974226943)+(-0.909038123273)*0.5**o
        ref[1,4,0,0]=(-0.149369878527)/(o+1.)+(0.594641654424)+(-0.974474210185)*0.5**o
        ref[1,4,0,1]=(0.995137009073)/(o+1.)+(-0.0464708598635)+(0.964745405431)*0.5**o
        ref[1,4,1,0]=(0.545158516903)/(o+1.)+(-0.562624149397)+(-0.714725894011)*0.5**o
        ref[1,4,1,1]=(-1.20779451465)/(o+1.)+(-0.189148851615)+(-0.809072545597)*0.5**o
        ref[1,4,2,0]=(-1.55971061601)/(o+1.)+(0.693489320143)+(-0.21281876112)*0.5**o
        ref[1,4,2,1]=(0.226662056752)/(o+1.)+(-0.454678810051)+(-0.247930702516)*0.5**o
        ref[2,0,0,0]=(-0.242758325417)/(o+1.)+(-0.527650883321)+(0.616352869937)*0.5**o
        ref[2,0,0,1]=(0.807761382166)/(o+1.)+(1.02431485248)+(-0.85704374519)*0.5**o
        ref[2,0,1,0]=(0.543068844645)/(o+1.)+(0.0690269849966)+(0.326870102902)*0.5**o
        ref[2,0,1,1]=(0.502168875857)/(o+1.)+(-0.0584590052478)+(-0.699939842176)*0.5**o
        ref[2,0,2,0]=(0.181622609504)/(o+1.)+(-0.57477595904)+(-0.985996827188)*0.5**o
        ref[2,0,2,1]=(-1.01828660056)/(o+1.)+(0.255680773601)+(-0.195357773436)*0.5**o
        ref[2,1,0,0]=(-0.961895417879)/(o+1.)+(-0.848679114358)+(0.601652939329)*0.5**o
        ref[2,1,0,1]=(0.177781123824)/(o+1.)+(-0.945003972397)+(0.534038022604)*0.5**o
        ref[2,1,1,0]=(-0.20513407563)/(o+1.)+(-0.126335927989)+(-0.361196978257)*0.5**o
        ref[2,1,1,1]=(0.712128730879)/(o+1.)+(-0.0623370236322)+(0.00930987921229)*0.5**o
        ref[2,1,2,0]=(0.692430706812)/(o+1.)+(0.417664138602)+(-0.181479246679)*0.5**o
        ref[2,1,2,1]=(0.650980026197)/(o+1.)+(-0.708193455952)+(-0.99665050175)*0.5**o
        ref[2,2,0,0]=(1.22766505061)/(o+1.)+(0.717584511586)+(0.870069083779)*0.5**o
        ref[2,2,0,1]=(0.633038618622)/(o+1.)+(-0.222859101796)+(-0.841259995642)*0.5**o
        ref[2,2,1,0]=(0.167120356805)/(o+1.)+(-1.00491115888)+(-0.561184923307)*0.5**o
        ref[2,2,1,1]=(0.164505345911)/(o+1.)+(1.15307619901)+(0.92243000401)*0.5**o
        ref[2,2,2,0]=(-1.16271376641)/(o+1.)+(-0.336355614369)+(0.331131296272)*0.5**o
        ref[2,2,2,1]=(0.672282083422)/(o+1.)+(0.018864653069)+(-0.47892171854)*0.5**o
        ref[2,3,0,0]=(1.52821633972)/(o+1.)+(-0.864758323678)+(0.0219406022052)*0.5**o
        ref[2,3,0,1]=(-0.326213615062)/(o+1.)+(0.152909191896)+(-0.114799511617)*0.5**o
        ref[2,3,1,0]=(-0.724364553267)/(o+1.)+(-0.439822878578)+(-0.103518223905)*0.5**o
        ref[2,3,1,1]=(-1.107549965)/(o+1.)+(-0.705530843209)+(-0.958142224242)*0.5**o
        ref[2,3,2,0]=(-0.825620325844)/(o+1.)+(-0.169496279085)+(0.730554594734)*0.5**o
        ref[2,3,2,1]=(-0.0678334241233)/(o+1.)+(0.0396617729923)+(-0.930974589897)*0.5**o
        ref[2,4,0,0]=(0.809143207998)/(o+1.)+(-0.709013942786)+(-0.246046780287)*0.5**o
        ref[2,4,0,1]=(0.0938678290563)/(o+1.)+(-0.0282613348041)+(-0.612751147889)*0.5**o
        ref[2,4,1,0]=(-0.0572988459938)/(o+1.)+(-0.496307592572)+(-0.26092972064)*0.5**o
        ref[2,4,1,1]=(1.38060158884)/(o+1.)+(-0.643768055252)+(0.325840618456)*0.5**o
        ref[2,4,2,0]=(-1.03935942333)/(o+1.)+(0.130235179204)+(0.655590135983)*0.5**o
        ref[2,4,2,1]=(-0.41630624669)/(o+1.)+(0.259401996992)+(0.570712162768)*0.5**o
        ref[3,0,0,0]=(-0.357343519171)/(o+1.)+(-0.518658431855)+(-0.486023722572)*0.5**o
        ref[3,0,0,1]=(-0.927017258737)/(o+1.)+(0.566657824853)+(-0.705291625776)*0.5**o
        ref[3,0,1,0]=(-0.331128184244)/(o+1.)+(0.535203433907)+(-0.838974120992)*0.5**o
        ref[3,0,1,1]=(-0.735721663177)/(o+1.)+(-0.293634410339)+(0.140027541071)*0.5**o
        ref[3,0,2,0]=(-0.178945077431)/(o+1.)+(-0.300862767422)+(-0.39083471939)*0.5**o
        ref[3,0,2,1]=(1.18223012428)/(o+1.)+(-0.121465976141)+(0.0734743009776)*0.5**o
        ref[3,1,0,0]=(1.46408415003)/(o+1.)+(-0.0664881764563)+(-0.414502599055)*0.5**o
        ref[3,1,0,1]=(1.3349422575)/(o+1.)+(0.283498682724)+(0.886115095396)*0.5**o
        ref[3,1,1,0]=(0.816511111677)/(o+1.)+(0.175825822544)+(0.312577129879)*0.5**o
        ref[3,1,1,1]=(-0.105059408543)/(o+1.)+(-0.782531675802)+(0.784369258288)*0.5**o
        ref[3,1,2,0]=(-0.744294216228)/(o+1.)+(-0.301922969419)+(-0.337332295082)*0.5**o
        ref[3,1,2,1]=(-0.669631653258)/(o+1.)+(0.0177835713351)+(0.791575892417)*0.5**o
        ref[3,2,0,0]=(0.274579265949)/(o+1.)+(0.148529098776)+(-0.0909007497318)*0.5**o
        ref[3,2,0,1]=(0.956130125711)/(o+1.)+(0.482854719491)+(-0.764137934896)*0.5**o
        ref[3,2,1,0]=(-0.266431989669)/(o+1.)+(0.313297832739)+(0.49435832562)*0.5**o
        ref[3,2,1,1]=(-0.726703087024)/(o+1.)+(0.542295175061)+(-0.460752644946)*0.5**o
        ref[3,2,2,0]=(0.87651829094)/(o+1.)+(-0.185020317268)+(0.0946687985697)*0.5**o
        ref[3,2,2,1]=(-0.113135321417)/(o+1.)+(-0.221883684949)+(0.177521713734)*0.5**o
        ref[3,3,0,0]=(0.155561630388)/(o+1.)+(0.495429901163)+(0.592201472175)*0.5**o
        ref[3,3,0,1]=(-0.44887181194)/(o+1.)+(-0.20854804469)+(0.572972211447)*0.5**o
        ref[3,3,1,0]=(1.11423248654)/(o+1.)+(-0.212131906831)+(0.775163727164)*0.5**o
        ref[3,3,1,1]=(0.723748205425)/(o+1.)+(0.785798099184)+(-0.244586606882)*0.5**o
        ref[3,3,2,0]=(0.389980691746)/(o+1.)+(-0.0994690359155)+(-0.676040282048)*0.5**o
        ref[3,3,2,1]=(-0.0505489049572)/(o+1.)+(0.57416458839)+(-0.136437682299)*0.5**o
        ref[3,4,0,0]=(0.14698064297)/(o+1.)+(0.286169633527)+(0.28632587806)*0.5**o
        ref[3,4,0,1]=(-0.316556260565)/(o+1.)+(0.736868623097)+(0.326448964148)*0.5**o
        ref[3,4,1,0]=(-1.11727322574)/(o+1.)+(0.341015410564)+(0.760474271531)*0.5**o
        ref[3,4,1,1]=(-0.0908271211344)/(o+1.)+(-0.150268763334)+(-0.21508452109)*0.5**o
        ref[3,4,2,0]=(-1.40693165856)/(o+1.)+(0.796839755046)+(-0.195181632154)*0.5**o
        ref[3,4,2,1]=(-0.109962351944)/(o+1.)+(-0.36347859654)+(0.270278273932)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.0433006726989)*x[0]**o+(0.819109331838)*x[0]+(0.150292245587)*x[1]**o+(0.949397165008)*x[1]
        ref=(0.150292245587)/(o+1.)+(0.884253248423)+(0.0433006726989)*0.5**o
      else:
        arg=(-0.765858486678)*x[0]**o+(-0.829061249978)*x[0]+(-0.0255273602735)*x[1]**o+(-0.93097352763)*x[1]+(-0.775974817954)*x[2]**o+(0.0414580763278)*x[2]
        ref=(-0.801502178228)/(o+1.)+(-0.85928835064)+(-0.765858486678)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.708418284738)*x[0]**o+(-0.265416249865)*x[0]+(0.0412873279676)*x[1]**o+(0.18827690539)*x[1]
        arg[1]=(-0.0639578348425)*x[0]**o+(-0.435408788697)*x[0]+(0.429134354603)*x[1]**o+(0.0391558318157)*x[1]
        ref[0]=(0.0412873279676)/(o+1.)+(-0.0385696722374)+(0.708418284738)*0.5**o
        ref[1]=(0.429134354603)/(o+1.)+(-0.198126478441)+(-0.0639578348425)*0.5**o
      else:
        arg[0]=(-0.516066856372)*x[0]**o+(-0.509677909948)*x[0]+(0.189594948337)*x[1]**o+(0.743432553688)*x[1]+(0.687864109046)*x[2]**o+(0.257280277379)*x[2]
        arg[1]=(0.948762785699)*x[0]**o+(0.372861763607)*x[0]+(-0.166460043702)*x[1]**o+(0.0474018901531)*x[1]+(-0.485380626595)*x[2]**o+(-0.341701937267)*x[2]
        ref[0]=(0.877459057383)/(o+1.)+(0.245517460559)+(-0.516066856372)*0.5**o
        ref[1]=(-0.651840670298)/(o+1.)+(0.0392808582465)+(0.948762785699)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.0540280486222)*x[0]**o+(-0.936629447756)*x[0]+(0.530867101627)*x[1]**o+(-0.404522355122)*x[1]
        arg[0,1]=(0.585051034599)*x[0]**o+(-0.899683342239)*x[0]+(0.61865830472)*x[1]**o+(0.491768251743)*x[1]
        arg[0,2]=(0.924131379905)*x[0]**o+(-0.394367734246)*x[0]+(0.432122525065)*x[1]**o+(0.79923634141)*x[1]
        arg[0,3]=(-0.41435196904)*x[0]**o+(-0.22039395628)*x[0]+(-0.866551833459)*x[1]**o+(-0.0966534238435)*x[1]
        arg[0,4]=(0.855785701017)*x[0]**o+(0.0879754354535)*x[0]+(-0.0727782016239)*x[1]**o+(0.746878225435)*x[1]
        arg[1,0]=(-0.868450306659)*x[0]**o+(-0.00845792225901)*x[0]+(0.227399876353)*x[1]**o+(-0.522196483799)*x[1]
        arg[1,1]=(0.53076931297)*x[0]**o+(0.165744140226)*x[0]+(0.14586398181)*x[1]**o+(-0.694169551043)*x[1]
        arg[1,2]=(-0.936146144862)*x[0]**o+(-0.198367580291)*x[0]+(0.848822682163)*x[1]**o+(0.360521711263)*x[1]
        arg[1,3]=(-0.395889451325)*x[0]**o+(0.510713345487)*x[0]+(-0.62440033386)*x[1]**o+(0.29695714683)*x[1]
        arg[1,4]=(0.847685553395)*x[0]**o+(-0.291266469606)*x[0]+(-0.517114839541)*x[1]**o+(-0.279948585383)*x[1]
        arg[2,0]=(-0.622481093435)*x[0]**o+(0.0911922519622)*x[0]+(0.216648336495)*x[1]**o+(0.663562516802)*x[1]
        arg[2,1]=(0.567833843918)*x[0]**o+(0.18116887359)*x[0]+(-0.0433012798352)*x[1]**o+(0.998332142285)*x[1]
        arg[2,2]=(-0.81729786652)*x[0]**o+(-0.634680104287)*x[0]+(0.991513806139)*x[1]**o+(-0.0721700196003)*x[1]
        arg[2,3]=(0.121181286827)*x[0]**o+(-0.677803464822)*x[0]+(0.870139639102)*x[1]**o+(-0.134662002755)*x[1]
        arg[2,4]=(-0.167376807346)*x[0]**o+(-0.500556007898)*x[0]+(-0.839256017197)*x[1]**o+(-0.82147890878)*x[1]
        arg[3,0]=(0.576521064132)*x[0]**o+(-0.510276372588)*x[0]+(0.92888227487)*x[1]**o+(-0.88293512036)*x[1]
        arg[3,1]=(-0.159048471862)*x[0]**o+(-0.0479500210678)*x[0]+(-0.0216438779517)*x[1]**o+(-0.108252189548)*x[1]
        arg[3,2]=(-0.483779576334)*x[0]**o+(-0.871829147209)*x[0]+(-0.22556218283)*x[1]**o+(0.451007116911)*x[1]
        arg[3,3]=(-0.358888369902)*x[0]**o+(0.436554685191)*x[0]+(-0.0359187908139)*x[1]**o+(-0.0148930474244)*x[1]
        arg[3,4]=(-0.995411478882)*x[0]**o+(-0.457206328785)*x[0]+(-0.833823947273)*x[1]**o+(-0.0517068976093)*x[1]
        ref[0,0]=(0.530867101627)/(o+1.)+(-0.670575901439)+(-0.0540280486222)*0.5**o
        ref[0,1]=(0.61865830472)/(o+1.)+(-0.203957545248)+(0.585051034599)*0.5**o
        ref[0,2]=(0.432122525065)/(o+1.)+(0.202434303582)+(0.924131379905)*0.5**o
        ref[0,3]=(-0.866551833459)/(o+1.)+(-0.158523690062)+(-0.41435196904)*0.5**o
        ref[0,4]=(-0.0727782016239)/(o+1.)+(0.417426830444)+(0.855785701017)*0.5**o
        ref[1,0]=(0.227399876353)/(o+1.)+(-0.265327203029)+(-0.868450306659)*0.5**o
        ref[1,1]=(0.14586398181)/(o+1.)+(-0.264212705408)+(0.53076931297)*0.5**o
        ref[1,2]=(0.848822682163)/(o+1.)+(0.0810770654864)+(-0.936146144862)*0.5**o
        ref[1,3]=(-0.62440033386)/(o+1.)+(0.403835246159)+(-0.395889451325)*0.5**o
        ref[1,4]=(-0.517114839541)/(o+1.)+(-0.285607527495)+(0.847685553395)*0.5**o
        ref[2,0]=(0.216648336495)/(o+1.)+(0.377377384382)+(-0.622481093435)*0.5**o
        ref[2,1]=(-0.0433012798352)/(o+1.)+(0.589750507938)+(0.567833843918)*0.5**o
        ref[2,2]=(0.991513806139)/(o+1.)+(-0.353425061944)+(-0.81729786652)*0.5**o
        ref[2,3]=(0.870139639102)/(o+1.)+(-0.406232733789)+(0.121181286827)*0.5**o
        ref[2,4]=(-0.839256017197)/(o+1.)+(-0.661017458339)+(-0.167376807346)*0.5**o
        ref[3,0]=(0.92888227487)/(o+1.)+(-0.696605746474)+(0.576521064132)*0.5**o
        ref[3,1]=(-0.0216438779517)/(o+1.)+(-0.078101105308)+(-0.159048471862)*0.5**o
        ref[3,2]=(-0.22556218283)/(o+1.)+(-0.210411015149)+(-0.483779576334)*0.5**o
        ref[3,3]=(-0.0359187908139)/(o+1.)+(0.210830818883)+(-0.358888369902)*0.5**o
        ref[3,4]=(-0.833823947273)/(o+1.)+(-0.254456613197)+(-0.995411478882)*0.5**o
      else:
        arg[0,0]=(-0.937858877179)*x[0]**o+(-0.41590845461)*x[0]+(0.317385444442)*x[1]**o+(-0.166284606371)*x[1]+(-0.846106994172)*x[2]**o+(0.227876218054)*x[2]
        arg[0,1]=(0.497785867084)*x[0]**o+(0.883557362705)*x[0]+(-0.846602726567)*x[1]**o+(-0.859091910548)*x[1]+(-0.295742503292)*x[2]**o+(-0.853931233266)*x[2]
        arg[0,2]=(0.744501879094)*x[0]**o+(-0.0181115125382)*x[0]+(-0.856099357899)*x[1]**o+(0.767449074734)*x[1]+(-0.858552187616)*x[2]**o+(-0.580142057156)*x[2]
        arg[0,3]=(0.210031467359)*x[0]**o+(-0.994426053833)*x[0]+(0.648522478175)*x[1]**o+(0.0355805603417)*x[1]+(-0.274347447112)*x[2]**o+(-0.280165537012)*x[2]
        arg[0,4]=(-0.323650660168)*x[0]**o+(-0.279592941595)*x[0]+(-0.478811425713)*x[1]**o+(0.19753271141)*x[1]+(0.476202875551)*x[2]**o+(0.657358100659)*x[2]
        arg[1,0]=(0.890659868437)*x[0]**o+(0.100235405989)*x[0]+(0.325177871418)*x[1]**o+(0.675379114656)*x[1]+(0.458473828724)*x[2]**o+(-0.449822467434)*x[2]
        arg[1,1]=(-0.467895039242)*x[0]**o+(-0.551300000422)*x[0]+(0.976111922801)*x[1]**o+(0.809389941274)*x[1]+(-0.693066235685)*x[2]**o+(0.595719115505)*x[2]
        arg[1,2]=(-0.727304617172)*x[0]**o+(-0.270682330445)*x[0]+(-0.32926621635)*x[1]**o+(0.0389068111165)*x[1]+(0.300096425532)*x[2]**o+(-0.964910767632)*x[2]
        arg[1,3]=(0.195012106644)*x[0]**o+(0.18137839607)*x[0]+(-0.823129279257)*x[1]**o+(-0.0440106889304)*x[1]+(-0.298898659898)*x[2]**o+(0.801623925974)*x[2]
        arg[1,4]=(0.590200326369)*x[0]**o+(-0.800085209049)*x[0]+(-0.323836133834)*x[1]**o+(-0.953413527643)*x[1]+(0.0257128844735)*x[2]**o+(0.652469516082)*x[2]
        arg[2,0]=(-0.467537281458)*x[0]**o+(-0.05673369912)*x[0]+(-0.813542619659)*x[1]**o+(-0.778596374878)*x[1]+(0.752073310476)*x[2]**o+(0.187824163254)*x[2]
        arg[2,1]=(-0.75708112983)*x[0]**o+(0.866111788737)*x[0]+(0.715424327432)*x[1]**o+(-0.297967739435)*x[1]+(0.420241611178)*x[2]**o+(-0.616867390634)*x[2]
        arg[2,2]=(0.00628628803839)*x[0]**o+(0.700962301988)*x[0]+(0.712979263919)*x[1]**o+(-0.414401132518)*x[1]+(-0.975498928756)*x[2]**o+(0.858116391039)*x[2]
        arg[2,3]=(0.948164399923)*x[0]**o+(-0.88227825142)*x[0]+(0.30703295864)*x[1]**o+(0.510607674984)*x[1]+(0.362361719392)*x[2]**o+(-0.208696380629)*x[2]
        arg[2,4]=(0.661545494115)*x[0]**o+(-0.122960355184)*x[0]+(-0.801367604574)*x[1]**o+(0.395708459987)*x[1]+(0.0197225340457)*x[2]**o+(-0.377005073637)*x[2]
        arg[3,0]=(0.733597350529)*x[0]**o+(0.339318978843)*x[0]+(0.317608635437)*x[1]**o+(0.147746051469)*x[1]+(-0.33413935488)*x[2]**o+(-0.793337674933)*x[2]
        arg[3,1]=(0.0918904281646)*x[0]**o+(-0.530786152204)*x[0]+(-0.587542750401)*x[1]**o+(-0.810331029621)*x[1]+(-0.40221065987)*x[2]**o+(-0.250472092093)*x[2]
        arg[3,2]=(0.966442952042)*x[0]**o+(-0.756649866552)*x[0]+(-0.36466223222)*x[1]**o+(-0.48639005055)*x[1]+(-0.831374462742)*x[2]**o+(0.503451143644)*x[2]
        arg[3,3]=(-0.322275644664)*x[0]**o+(-0.254067149982)*x[0]+(-0.440264794128)*x[1]**o+(0.226546045529)*x[1]+(-0.761870916789)*x[2]**o+(0.70024066379)*x[2]
        arg[3,4]=(0.595459342613)*x[0]**o+(-0.439436200258)*x[0]+(0.0932538180176)*x[1]**o+(-0.975450806707)*x[1]+(0.0461423323623)*x[2]**o+(0.0294352472154)*x[2]
        ref[0,0]=(-0.52872154973)/(o+1.)+(-0.177158421464)+(-0.937858877179)*0.5**o
        ref[0,1]=(-1.14234522986)/(o+1.)+(-0.414732890554)+(0.497785867084)*0.5**o
        ref[0,2]=(-1.71465154551)/(o+1.)+(0.0845977525197)+(0.744501879094)*0.5**o
        ref[0,3]=(0.374175031062)/(o+1.)+(-0.619505515252)+(0.210031467359)*0.5**o
        ref[0,4]=(-0.00260855016132)/(o+1.)+(0.287648935237)+(-0.323650660168)*0.5**o
        ref[1,0]=(0.783651700142)/(o+1.)+(0.162896026605)+(0.890659868437)*0.5**o
        ref[1,1]=(0.283045687117)/(o+1.)+(0.426904528178)+(-0.467895039242)*0.5**o
        ref[1,2]=(-0.0291697908188)/(o+1.)+(-0.59834314348)+(-0.727304617172)*0.5**o
        ref[1,3]=(-1.12202793916)/(o+1.)+(0.469495816557)+(0.195012106644)*0.5**o
        ref[1,4]=(-0.29812324936)/(o+1.)+(-0.550514610305)+(0.590200326369)*0.5**o
        ref[2,0]=(-0.0614693091826)/(o+1.)+(-0.323752955372)+(-0.467537281458)*0.5**o
        ref[2,1]=(1.13566593861)/(o+1.)+(-0.024361670666)+(-0.75708112983)*0.5**o
        ref[2,2]=(-0.262519664837)/(o+1.)+(0.572338780255)+(0.00628628803839)*0.5**o
        ref[2,3]=(0.669394678032)/(o+1.)+(-0.290183478532)+(0.948164399923)*0.5**o
        ref[2,4]=(-0.781645070528)/(o+1.)+(-0.0521284844165)+(0.661545494115)*0.5**o
        ref[3,0]=(-0.0165307194439)/(o+1.)+(-0.15313632231)+(0.733597350529)*0.5**o
        ref[3,1]=(-0.989753410271)/(o+1.)+(-0.795794636959)+(0.0918904281646)*0.5**o
        ref[3,2]=(-1.19603669496)/(o+1.)+(-0.369794386729)+(0.966442952042)*0.5**o
        ref[3,3]=(-1.20213571092)/(o+1.)+(0.336359779668)+(-0.322275644664)*0.5**o
        ref[3,4]=(0.13939615038)/(o+1.)+(-0.692725879875)+(0.595459342613)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.12237588236)*x[0]**o+(-0.671341785528)*x[0]+(0.998910161099)*x[1]**o+(-0.552290246966)*x[1]
        arg[0,0,1]=(-0.90821257221)*x[0]**o+(-0.413391927764)*x[0]+(-0.428904099491)*x[1]**o+(-0.226766222189)*x[1]
        arg[0,1,0]=(-0.333819313127)*x[0]**o+(-0.458569899029)*x[0]+(-0.83776177332)*x[1]**o+(0.22951184631)*x[1]
        arg[0,1,1]=(-0.0571879315489)*x[0]**o+(-0.38056227895)*x[0]+(0.0784628125977)*x[1]**o+(-0.659353613638)*x[1]
        arg[1,0,0]=(0.645871468353)*x[0]**o+(-0.986005542006)*x[0]+(0.395704509276)*x[1]**o+(0.980666918726)*x[1]
        arg[1,0,1]=(0.337293036597)*x[0]**o+(-0.520809490378)*x[0]+(0.543657498632)*x[1]**o+(0.632777244118)*x[1]
        arg[1,1,0]=(-0.432349538121)*x[0]**o+(-0.579685247099)*x[0]+(0.331943072726)*x[1]**o+(0.25042188644)*x[1]
        arg[1,1,1]=(0.329861800715)*x[0]**o+(-0.48725090895)*x[0]+(-0.149357998414)*x[1]**o+(-0.580650134153)*x[1]
        arg[2,0,0]=(-0.75424658357)*x[0]**o+(-0.696352517414)*x[0]+(-0.00272313166917)*x[1]**o+(-0.488581343546)*x[1]
        arg[2,0,1]=(-0.136489901339)*x[0]**o+(-0.752587775699)*x[0]+(0.986413392791)*x[1]**o+(0.759344728162)*x[1]
        arg[2,1,0]=(-0.0832674877963)*x[0]**o+(0.765419402778)*x[0]+(0.958533611564)*x[1]**o+(-0.315461136294)*x[1]
        arg[2,1,1]=(0.362764663642)*x[0]**o+(-0.351519282456)*x[0]+(-0.115554433923)*x[1]**o+(-0.789365036398)*x[1]
        arg[3,0,0]=(0.0977752933224)*x[0]**o+(-0.838881001576)*x[0]+(0.275578737811)*x[1]**o+(-0.971208155806)*x[1]
        arg[3,0,1]=(0.0588243034654)*x[0]**o+(-0.141302234145)*x[0]+(-0.460044396206)*x[1]**o+(-0.776793718354)*x[1]
        arg[3,1,0]=(0.786922930047)*x[0]**o+(-0.187797924361)*x[0]+(0.971516347812)*x[1]**o+(-0.250547377602)*x[1]
        arg[3,1,1]=(-0.69240453335)*x[0]**o+(0.439425578473)*x[0]+(0.00864940781276)*x[1]**o+(-0.458898387231)*x[1]
        arg[4,0,0]=(0.0700104829681)*x[0]**o+(0.905031121846)*x[0]+(0.37676750477)*x[1]**o+(0.411190149331)*x[1]
        arg[4,0,1]=(-0.615737714754)*x[0]**o+(0.1945204765)*x[0]+(0.0889014902938)*x[1]**o+(0.650020022597)*x[1]
        arg[4,1,0]=(-0.883448519911)*x[0]**o+(0.00124188371434)*x[0]+(-0.501662684156)*x[1]**o+(0.290235036398)*x[1]
        arg[4,1,1]=(-0.723885968554)*x[0]**o+(-0.0112308623755)*x[0]+(0.679499427805)*x[1]**o+(-0.00261690404287)*x[1]
        arg[5,0,0]=(0.0834731365594)*x[0]**o+(-0.873857991293)*x[0]+(0.831158842487)*x[1]**o+(-0.585808323324)*x[1]
        arg[5,0,1]=(0.250830078938)*x[0]**o+(0.310347394372)*x[0]+(-0.412138455827)*x[1]**o+(0.974823718029)*x[1]
        arg[5,1,0]=(-0.865155065381)*x[0]**o+(0.72190541012)*x[0]+(0.72020941875)*x[1]**o+(-0.604997858133)*x[1]
        arg[5,1,1]=(0.57252234131)*x[0]**o+(-0.366230226771)*x[0]+(0.115400105646)*x[1]**o+(0.0485229323141)*x[1]
        ref[0,0,0]=(0.998910161099)/(o+1.)+(-0.611816016247)+(-0.12237588236)*0.5**o
        ref[0,0,1]=(-0.428904099491)/(o+1.)+(-0.320079074977)+(-0.90821257221)*0.5**o
        ref[0,1,0]=(-0.83776177332)/(o+1.)+(-0.11452902636)+(-0.333819313127)*0.5**o
        ref[0,1,1]=(0.0784628125977)/(o+1.)+(-0.519957946294)+(-0.0571879315489)*0.5**o
        ref[1,0,0]=(0.395704509276)/(o+1.)+(-0.00266931163989)+(0.645871468353)*0.5**o
        ref[1,0,1]=(0.543657498632)/(o+1.)+(0.0559838768697)+(0.337293036597)*0.5**o
        ref[1,1,0]=(0.331943072726)/(o+1.)+(-0.16463168033)+(-0.432349538121)*0.5**o
        ref[1,1,1]=(-0.149357998414)/(o+1.)+(-0.533950521552)+(0.329861800715)*0.5**o
        ref[2,0,0]=(-0.00272313166917)/(o+1.)+(-0.59246693048)+(-0.75424658357)*0.5**o
        ref[2,0,1]=(0.986413392791)/(o+1.)+(0.00337847623164)+(-0.136489901339)*0.5**o
        ref[2,1,0]=(0.958533611564)/(o+1.)+(0.224979133242)+(-0.0832674877963)*0.5**o
        ref[2,1,1]=(-0.115554433923)/(o+1.)+(-0.570442159427)+(0.362764663642)*0.5**o
        ref[3,0,0]=(0.275578737811)/(o+1.)+(-0.905044578691)+(0.0977752933224)*0.5**o
        ref[3,0,1]=(-0.460044396206)/(o+1.)+(-0.45904797625)+(0.0588243034654)*0.5**o
        ref[3,1,0]=(0.971516347812)/(o+1.)+(-0.219172650981)+(0.786922930047)*0.5**o
        ref[3,1,1]=(0.00864940781276)/(o+1.)+(-0.00973640437899)+(-0.69240453335)*0.5**o
        ref[4,0,0]=(0.37676750477)/(o+1.)+(0.658110635589)+(0.0700104829681)*0.5**o
        ref[4,0,1]=(0.0889014902938)/(o+1.)+(0.422270249549)+(-0.615737714754)*0.5**o
        ref[4,1,0]=(-0.501662684156)/(o+1.)+(0.145738460056)+(-0.883448519911)*0.5**o
        ref[4,1,1]=(0.679499427805)/(o+1.)+(-0.00692388320921)+(-0.723885968554)*0.5**o
        ref[5,0,0]=(0.831158842487)/(o+1.)+(-0.729833157309)+(0.0834731365594)*0.5**o
        ref[5,0,1]=(-0.412138455827)/(o+1.)+(0.642585556201)+(0.250830078938)*0.5**o
        ref[5,1,0]=(0.72020941875)/(o+1.)+(0.0584537759935)+(-0.865155065381)*0.5**o
        ref[5,1,1]=(0.115400105646)/(o+1.)+(-0.158853647228)+(0.57252234131)*0.5**o
      else:
        arg[0,0,0]=(0.962792187023)*x[0]**o+(0.799861778652)*x[0]+(-0.811367630134)*x[1]**o+(-0.553496861095)*x[1]+(-0.810089307714)*x[2]**o+(0.135449765364)*x[2]
        arg[0,0,1]=(0.644573004654)*x[0]**o+(-0.565823623798)*x[0]+(0.337026645127)*x[1]**o+(0.413612857799)*x[1]+(0.604366347987)*x[2]**o+(0.0867373815054)*x[2]
        arg[0,1,0]=(0.829914553754)*x[0]**o+(-0.462419132852)*x[0]+(0.889464473044)*x[1]**o+(-0.425594549042)*x[1]+(0.679251542389)*x[2]**o+(-0.501090535615)*x[2]
        arg[0,1,1]=(0.874166303558)*x[0]**o+(0.437595762221)*x[0]+(-0.836229998654)*x[1]**o+(-0.704020633222)*x[1]+(-0.279878325513)*x[2]**o+(0.697289463565)*x[2]
        arg[1,0,0]=(-0.963425479949)*x[0]**o+(0.399090631945)*x[0]+(-0.80371987894)*x[1]**o+(0.614102223184)*x[1]+(-0.821741166107)*x[2]**o+(-0.245209306654)*x[2]
        arg[1,0,1]=(-0.288791346674)*x[0]**o+(0.0125896799209)*x[0]+(0.0536290891541)*x[1]**o+(-0.127968165901)*x[1]+(-0.635591402779)*x[2]**o+(-0.723850350215)*x[2]
        arg[1,1,0]=(-0.554196270608)*x[0]**o+(-0.409293351346)*x[0]+(-0.759821124187)*x[1]**o+(-0.828913354317)*x[1]+(0.66900958614)*x[2]**o+(0.412403847239)*x[2]
        arg[1,1,1]=(-0.862735330269)*x[0]**o+(-0.406644646366)*x[0]+(-0.375399872184)*x[1]**o+(-0.10762837777)*x[1]+(0.540886844249)*x[2]**o+(0.97487281528)*x[2]
        arg[2,0,0]=(0.388899900771)*x[0]**o+(-0.692092360652)*x[0]+(0.794415524913)*x[1]**o+(0.369239170999)*x[1]+(0.482269641926)*x[2]**o+(0.58717167121)*x[2]
        arg[2,0,1]=(0.324795219208)*x[0]**o+(0.508631698719)*x[0]+(0.128456415737)*x[1]**o+(0.73678163559)*x[1]+(0.0527520660005)*x[2]**o+(-0.180289299372)*x[2]
        arg[2,1,0]=(-0.901972287405)*x[0]**o+(-0.635701120264)*x[0]+(0.298343193246)*x[1]**o+(-0.870902858159)*x[1]+(-0.979619200422)*x[2]**o+(0.708567755329)*x[2]
        arg[2,1,1]=(0.391663724321)*x[0]**o+(0.69891810611)*x[0]+(0.717069638885)*x[1]**o+(0.493487853359)*x[1]+(0.270536570073)*x[2]**o+(0.679115502666)*x[2]
        arg[3,0,0]=(0.830590943609)*x[0]**o+(-0.000320739222143)*x[0]+(-0.659146599349)*x[1]**o+(-0.858428982679)*x[1]+(-0.158070494634)*x[2]**o+(-0.364394913197)*x[2]
        arg[3,0,1]=(0.903615175916)*x[0]**o+(0.186562406505)*x[0]+(0.463860110418)*x[1]**o+(0.0382550628307)*x[1]+(0.835148180357)*x[2]**o+(-0.117637969912)*x[2]
        arg[3,1,0]=(-0.901874408714)*x[0]**o+(0.00493918499294)*x[0]+(0.180179475909)*x[1]**o+(-0.549428619271)*x[1]+(-0.392536708051)*x[2]**o+(0.68054148024)*x[2]
        arg[3,1,1]=(0.807693382033)*x[0]**o+(-0.47532617097)*x[0]+(-0.52499678021)*x[1]**o+(-0.0365395713918)*x[1]+(0.654881606829)*x[2]**o+(-0.63350736651)*x[2]
        arg[4,0,0]=(0.440830144989)*x[0]**o+(-0.454694325046)*x[0]+(0.403509667018)*x[1]**o+(0.360080283937)*x[1]+(0.164208775222)*x[2]**o+(0.203945755083)*x[2]
        arg[4,0,1]=(-0.836804887763)*x[0]**o+(-0.910979763109)*x[0]+(-0.629411293881)*x[1]**o+(0.0546020113997)*x[1]+(0.63435754491)*x[2]**o+(-0.323224122475)*x[2]
        arg[4,1,0]=(0.204216251019)*x[0]**o+(-0.372349067194)*x[0]+(-0.320768886225)*x[1]**o+(-0.139715856072)*x[1]+(0.62329130659)*x[2]**o+(0.626784461235)*x[2]
        arg[4,1,1]=(0.275074685897)*x[0]**o+(0.0664691742991)*x[0]+(0.949175504221)*x[1]**o+(-0.438858093561)*x[1]+(0.330489174462)*x[2]**o+(0.361586647545)*x[2]
        arg[5,0,0]=(0.370194889673)*x[0]**o+(-0.0549428893157)*x[0]+(0.0219932528442)*x[1]**o+(-0.699553487775)*x[1]+(0.0161212665659)*x[2]**o+(-0.268426682266)*x[2]
        arg[5,0,1]=(-0.527801267225)*x[0]**o+(0.151774276637)*x[0]+(-0.0623666836223)*x[1]**o+(-0.953572390482)*x[1]+(0.778912307756)*x[2]**o+(-0.851760269136)*x[2]
        arg[5,1,0]=(0.11418535486)*x[0]**o+(-0.299719975234)*x[0]+(-0.508303365848)*x[1]**o+(-0.856367302992)*x[1]+(0.311609370032)*x[2]**o+(0.4731881155)*x[2]
        arg[5,1,1]=(0.455947680647)*x[0]**o+(0.856811965126)*x[0]+(-0.176215162097)*x[1]**o+(0.581325693162)*x[1]+(-0.306514780798)*x[2]**o+(-0.005241787834)*x[2]
        ref[0,0,0]=(-1.62145693785)/(o+1.)+(0.19090734146)+(0.962792187023)*0.5**o
        ref[0,0,1]=(0.941392993115)/(o+1.)+(-0.0327366922471)+(0.644573004654)*0.5**o
        ref[0,1,0]=(1.56871601543)/(o+1.)+(-0.694552108754)+(0.829914553754)*0.5**o
        ref[0,1,1]=(-1.11610832417)/(o+1.)+(0.215432296282)+(0.874166303558)*0.5**o
        ref[1,0,0]=(-1.62546104505)/(o+1.)+(0.383991774237)+(-0.963425479949)*0.5**o
        ref[1,0,1]=(-0.581962313625)/(o+1.)+(-0.419614418097)+(-0.288791346674)*0.5**o
        ref[1,1,0]=(-0.0908115380472)/(o+1.)+(-0.412901429212)+(-0.554196270608)*0.5**o
        ref[1,1,1]=(0.165486972066)/(o+1.)+(0.230299895572)+(-0.862735330269)*0.5**o
        ref[2,0,0]=(1.27668516684)/(o+1.)+(0.132159240779)+(0.388899900771)*0.5**o
        ref[2,0,1]=(0.181208481738)/(o+1.)+(0.532562017469)+(0.324795219208)*0.5**o
        ref[2,1,0]=(-0.681276007176)/(o+1.)+(-0.399018111547)+(-0.901972287405)*0.5**o
        ref[2,1,1]=(0.987606208958)/(o+1.)+(0.935760731067)+(0.391663724321)*0.5**o
        ref[3,0,0]=(-0.817217093983)/(o+1.)+(-0.611572317549)+(0.830590943609)*0.5**o
        ref[3,0,1]=(1.29900829077)/(o+1.)+(0.0535897497116)+(0.903615175916)*0.5**o
        ref[3,1,0]=(-0.212357232141)/(o+1.)+(0.0680260229807)+(-0.901874408714)*0.5**o
        ref[3,1,1]=(0.12988482662)/(o+1.)+(-0.572686554436)+(0.807693382033)*0.5**o
        ref[4,0,0]=(0.567718442239)/(o+1.)+(0.0546658569873)+(0.440830144989)*0.5**o
        ref[4,0,1]=(0.00494625102884)/(o+1.)+(-0.589800937092)+(-0.836804887763)*0.5**o
        ref[4,1,0]=(0.302522420366)/(o+1.)+(0.0573597689848)+(0.204216251019)*0.5**o
        ref[4,1,1]=(1.27966467868)/(o+1.)+(-0.0054011358584)+(0.275074685897)*0.5**o
        ref[5,0,0]=(0.0381145194101)/(o+1.)+(-0.511461529678)+(0.370194889673)*0.5**o
        ref[5,0,1]=(0.716545624134)/(o+1.)+(-0.826779191491)+(-0.527801267225)*0.5**o
        ref[5,1,0]=(-0.196693995816)/(o+1.)+(-0.341449581363)+(0.11418535486)*0.5**o
        ref[5,1,1]=(-0.482729942895)/(o+1.)+(0.716447935227)+(0.455947680647)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.415003066171)*x[0]**o+(-0.47441250281)*x[0]+(0.765391656872)*x[1]**o+(-0.170257448926)*x[1]
        arg[0,0,0,1]=(-0.928539602254)*x[0]**o+(0.792240012243)*x[0]+(-0.152730990885)*x[1]**o+(-0.667500454847)*x[1]
        arg[0,0,1,0]=(-0.937831078575)*x[0]**o+(0.963649346531)*x[0]+(0.505558859525)*x[1]**o+(-0.921354612243)*x[1]
        arg[0,0,1,1]=(0.909415074451)*x[0]**o+(0.199360871726)*x[0]+(0.194087349442)*x[1]**o+(0.875065381425)*x[1]
        arg[0,0,2,0]=(0.935190448625)*x[0]**o+(-0.109984767478)*x[0]+(0.518460161897)*x[1]**o+(-0.597684157349)*x[1]
        arg[0,0,2,1]=(-0.94062714999)*x[0]**o+(-0.308960574286)*x[0]+(0.578391007321)*x[1]**o+(0.864501040998)*x[1]
        arg[0,1,0,0]=(-0.197219678026)*x[0]**o+(-0.982899373778)*x[0]+(-0.603956105825)*x[1]**o+(-0.242717148913)*x[1]
        arg[0,1,0,1]=(-0.590715606611)*x[0]**o+(0.461185487788)*x[0]+(-0.270612249495)*x[1]**o+(-0.644742980654)*x[1]
        arg[0,1,1,0]=(-0.365516638755)*x[0]**o+(-0.255307471195)*x[0]+(0.0875989587033)*x[1]**o+(0.923189427818)*x[1]
        arg[0,1,1,1]=(0.315218558548)*x[0]**o+(-0.373070897551)*x[0]+(0.625272530385)*x[1]**o+(-0.174054543645)*x[1]
        arg[0,1,2,0]=(-0.502497884169)*x[0]**o+(-0.933785405889)*x[0]+(0.115500534905)*x[1]**o+(-0.864454092408)*x[1]
        arg[0,1,2,1]=(0.615837581258)*x[0]**o+(0.832538142367)*x[0]+(-0.0715650027129)*x[1]**o+(-0.775173801322)*x[1]
        arg[0,2,0,0]=(-0.898904577889)*x[0]**o+(-0.430072579352)*x[0]+(-0.532213103734)*x[1]**o+(0.970885652167)*x[1]
        arg[0,2,0,1]=(-0.0056483770747)*x[0]**o+(-0.566550554106)*x[0]+(0.751663691868)*x[1]**o+(0.677878661708)*x[1]
        arg[0,2,1,0]=(0.271196923039)*x[0]**o+(-0.291608730386)*x[0]+(-0.932579716748)*x[1]**o+(-0.437823650176)*x[1]
        arg[0,2,1,1]=(-0.788206773125)*x[0]**o+(0.18675055747)*x[0]+(0.747625244884)*x[1]**o+(0.654933611758)*x[1]
        arg[0,2,2,0]=(-0.787603701957)*x[0]**o+(-0.789784850865)*x[0]+(0.197684766271)*x[1]**o+(-0.988404958072)*x[1]
        arg[0,2,2,1]=(0.417538087399)*x[0]**o+(0.745832250585)*x[0]+(-0.525306115268)*x[1]**o+(0.134086803871)*x[1]
        arg[0,3,0,0]=(-0.230658480436)*x[0]**o+(-0.072983085881)*x[0]+(-0.149386157207)*x[1]**o+(-0.506870604002)*x[1]
        arg[0,3,0,1]=(-0.815859549425)*x[0]**o+(-0.962632495033)*x[0]+(-0.542257139282)*x[1]**o+(0.93598141472)*x[1]
        arg[0,3,1,0]=(0.0508974791413)*x[0]**o+(-0.722669867743)*x[0]+(0.714019527678)*x[1]**o+(0.710015283752)*x[1]
        arg[0,3,1,1]=(0.119419200166)*x[0]**o+(0.365663074152)*x[0]+(0.352455072404)*x[1]**o+(-0.0788953701144)*x[1]
        arg[0,3,2,0]=(-0.82075904747)*x[0]**o+(-0.224534176808)*x[0]+(-0.585094684368)*x[1]**o+(0.242958171035)*x[1]
        arg[0,3,2,1]=(-0.748061395205)*x[0]**o+(0.676995891547)*x[0]+(0.686644239408)*x[1]**o+(0.260350484065)*x[1]
        arg[0,4,0,0]=(0.128293903485)*x[0]**o+(0.226748407809)*x[0]+(0.129373873322)*x[1]**o+(0.988775321025)*x[1]
        arg[0,4,0,1]=(0.639638672694)*x[0]**o+(-0.817791454136)*x[0]+(-0.170677596319)*x[1]**o+(0.404161643274)*x[1]
        arg[0,4,1,0]=(0.43743097727)*x[0]**o+(-0.408230274495)*x[0]+(0.282544628333)*x[1]**o+(-0.542352697967)*x[1]
        arg[0,4,1,1]=(-0.995599965283)*x[0]**o+(-0.0512373236039)*x[0]+(0.218383977926)*x[1]**o+(0.929355335423)*x[1]
        arg[0,4,2,0]=(0.922609534479)*x[0]**o+(-0.166291574079)*x[0]+(-0.449331655188)*x[1]**o+(-0.0279347569009)*x[1]
        arg[0,4,2,1]=(0.950186933361)*x[0]**o+(-0.426615341822)*x[0]+(-0.167639663907)*x[1]**o+(0.219020502174)*x[1]
        arg[1,0,0,0]=(0.0573632292788)*x[0]**o+(0.543681035865)*x[0]+(-0.622510875807)*x[1]**o+(-0.171567467655)*x[1]
        arg[1,0,0,1]=(0.368975600444)*x[0]**o+(-0.754217532725)*x[0]+(0.469744047929)*x[1]**o+(-0.922445394777)*x[1]
        arg[1,0,1,0]=(-0.810194964715)*x[0]**o+(0.0861796117934)*x[0]+(0.625367069433)*x[1]**o+(-0.950906871025)*x[1]
        arg[1,0,1,1]=(0.744923275143)*x[0]**o+(0.465744045178)*x[0]+(0.359324627118)*x[1]**o+(0.273934205663)*x[1]
        arg[1,0,2,0]=(0.891959004941)*x[0]**o+(-0.498241249484)*x[0]+(0.318198109895)*x[1]**o+(0.0674564169104)*x[1]
        arg[1,0,2,1]=(0.842125439023)*x[0]**o+(0.731884782944)*x[0]+(-0.674805144805)*x[1]**o+(0.431782566319)*x[1]
        arg[1,1,0,0]=(-0.786719774805)*x[0]**o+(-0.55929144961)*x[0]+(-0.740263211102)*x[1]**o+(0.777914467998)*x[1]
        arg[1,1,0,1]=(0.831245977993)*x[0]**o+(-0.758991188073)*x[0]+(0.714032957488)*x[1]**o+(-0.879935556766)*x[1]
        arg[1,1,1,0]=(0.862796574462)*x[0]**o+(0.746875042565)*x[0]+(0.455949140326)*x[1]**o+(0.585491946591)*x[1]
        arg[1,1,1,1]=(-0.500409398892)*x[0]**o+(0.292753163538)*x[0]+(-0.62474682088)*x[1]**o+(0.583368504524)*x[1]
        arg[1,1,2,0]=(-0.402886961189)*x[0]**o+(-0.288733078165)*x[0]+(-0.770315341866)*x[1]**o+(-0.462764773299)*x[1]
        arg[1,1,2,1]=(0.358161520677)*x[0]**o+(-0.524052652951)*x[0]+(0.654162491323)*x[1]**o+(0.902382016998)*x[1]
        arg[1,2,0,0]=(-0.15539934232)*x[0]**o+(0.693162750731)*x[0]+(-0.0538737948754)*x[1]**o+(-0.025534990068)*x[1]
        arg[1,2,0,1]=(0.371159212282)*x[0]**o+(-0.751292369171)*x[0]+(-0.377617770006)*x[1]**o+(0.707647483842)*x[1]
        arg[1,2,1,0]=(0.117790556271)*x[0]**o+(0.848017682445)*x[0]+(-0.360936107893)*x[1]**o+(-0.564060873236)*x[1]
        arg[1,2,1,1]=(-0.524828128971)*x[0]**o+(0.737876348154)*x[0]+(0.183316445987)*x[1]**o+(-0.821207746237)*x[1]
        arg[1,2,2,0]=(0.300516177535)*x[0]**o+(-0.204187510383)*x[0]+(0.126601242737)*x[1]**o+(0.0794834423875)*x[1]
        arg[1,2,2,1]=(-0.29046926624)*x[0]**o+(-0.310057692431)*x[0]+(-0.810227948574)*x[1]**o+(-0.464128160277)*x[1]
        arg[1,3,0,0]=(-0.134275834453)*x[0]**o+(0.448577162559)*x[0]+(0.757723452446)*x[1]**o+(0.139202834725)*x[1]
        arg[1,3,0,1]=(0.882915728842)*x[0]**o+(-0.0750591947315)*x[0]+(0.503482692135)*x[1]**o+(-0.176350465386)*x[1]
        arg[1,3,1,0]=(-0.954792910945)*x[0]**o+(0.772215353622)*x[0]+(-0.861747803918)*x[1]**o+(-0.932727438775)*x[1]
        arg[1,3,1,1]=(0.747255393187)*x[0]**o+(0.0392661297355)*x[0]+(-0.322274864209)*x[1]**o+(-0.491462177972)*x[1]
        arg[1,3,2,0]=(0.139287057836)*x[0]**o+(0.312258185231)*x[0]+(-0.488288412308)*x[1]**o+(-0.185131700928)*x[1]
        arg[1,3,2,1]=(-0.861625516304)*x[0]**o+(0.681454010645)*x[0]+(0.206644380777)*x[1]**o+(-0.0344138929268)*x[1]
        arg[1,4,0,0]=(-0.714198122614)*x[0]**o+(-0.413113909739)*x[0]+(-0.971680855409)*x[1]**o+(0.602598611545)*x[1]
        arg[1,4,0,1]=(0.293578109913)*x[0]**o+(-0.495635685351)*x[0]+(0.347408878778)*x[1]**o+(-0.923814185786)*x[1]
        arg[1,4,1,0]=(0.843760126513)*x[0]**o+(-0.671075074679)*x[0]+(-0.284086603407)*x[1]**o+(0.437492233496)*x[1]
        arg[1,4,1,1]=(-0.342911357374)*x[0]**o+(-0.440232224635)*x[0]+(0.537701123084)*x[1]**o+(-0.525203729082)*x[1]
        arg[1,4,2,0]=(-0.71201360314)*x[0]**o+(-0.706332525171)*x[0]+(0.831438049852)*x[1]**o+(0.819141670055)*x[1]
        arg[1,4,2,1]=(-0.616403541928)*x[0]**o+(0.635973739636)*x[0]+(-0.0369972206543)*x[1]**o+(-0.08595692743)*x[1]
        arg[2,0,0,0]=(-0.959754423491)*x[0]**o+(-0.332473557912)*x[0]+(-0.464334304695)*x[1]**o+(0.597214020843)*x[1]
        arg[2,0,0,1]=(0.231051530608)*x[0]**o+(0.368524293555)*x[0]+(0.173044245392)*x[1]**o+(0.505867390994)*x[1]
        arg[2,0,1,0]=(0.376577055194)*x[0]**o+(-0.739990982509)*x[0]+(-0.816201949845)*x[1]**o+(-0.71233193641)*x[1]
        arg[2,0,1,1]=(-0.4891837531)*x[0]**o+(-0.121562226208)*x[0]+(-0.876294548464)*x[1]**o+(0.377763924093)*x[1]
        arg[2,0,2,0]=(-0.759548081841)*x[0]**o+(-0.118749434442)*x[0]+(0.602375475355)*x[1]**o+(-0.877213577079)*x[1]
        arg[2,0,2,1]=(0.114018259802)*x[0]**o+(-0.627149204501)*x[0]+(0.813648191424)*x[1]**o+(-0.612548616323)*x[1]
        arg[2,1,0,0]=(-0.572699800518)*x[0]**o+(0.507715829662)*x[0]+(-0.12122456959)*x[1]**o+(0.75522878228)*x[1]
        arg[2,1,0,1]=(0.0655268452799)*x[0]**o+(0.211191572823)*x[0]+(0.539760655497)*x[1]**o+(0.657397992057)*x[1]
        arg[2,1,1,0]=(0.358281119336)*x[0]**o+(0.119364032507)*x[0]+(-0.0242249415295)*x[1]**o+(-0.645986730609)*x[1]
        arg[2,1,1,1]=(0.523751704889)*x[0]**o+(-0.237670103222)*x[0]+(-0.995787126907)*x[1]**o+(-0.504129088442)*x[1]
        arg[2,1,2,0]=(-0.185380508646)*x[0]**o+(-0.218341933583)*x[0]+(0.149400154478)*x[1]**o+(0.00252616040715)*x[1]
        arg[2,1,2,1]=(0.81086932122)*x[0]**o+(0.387271527481)*x[0]+(0.908525646396)*x[1]**o+(-0.75383394204)*x[1]
        arg[2,2,0,0]=(-0.463416100256)*x[0]**o+(-0.891654021409)*x[0]+(-0.0495088159461)*x[1]**o+(0.398915277911)*x[1]
        arg[2,2,0,1]=(-0.200698149193)*x[0]**o+(0.127899540753)*x[0]+(-0.895280577411)*x[1]**o+(-0.197025422242)*x[1]
        arg[2,2,1,0]=(0.888683772867)*x[0]**o+(-0.84436563509)*x[0]+(-0.375709672588)*x[1]**o+(0.591336312818)*x[1]
        arg[2,2,1,1]=(-0.0167949020901)*x[0]**o+(0.145017501865)*x[0]+(0.868194479789)*x[1]**o+(-0.62439676328)*x[1]
        arg[2,2,2,0]=(0.441981698578)*x[0]**o+(-0.481156166596)*x[0]+(0.267937553127)*x[1]**o+(-0.37032246733)*x[1]
        arg[2,2,2,1]=(0.876701424579)*x[0]**o+(0.473798212378)*x[0]+(0.697645775876)*x[1]**o+(-0.798143376838)*x[1]
        arg[2,3,0,0]=(0.111231509217)*x[0]**o+(0.188121806171)*x[0]+(-0.0706155476337)*x[1]**o+(0.779423336221)*x[1]
        arg[2,3,0,1]=(-0.858913402812)*x[0]**o+(0.655459361097)*x[0]+(0.326282512878)*x[1]**o+(0.863386952614)*x[1]
        arg[2,3,1,0]=(-0.524551221953)*x[0]**o+(-0.807242415277)*x[0]+(0.575866251831)*x[1]**o+(-0.74503615846)*x[1]
        arg[2,3,1,1]=(0.391514075129)*x[0]**o+(0.477257809825)*x[0]+(-0.488263373008)*x[1]**o+(-0.29101442481)*x[1]
        arg[2,3,2,0]=(0.26532129161)*x[0]**o+(0.22794872765)*x[0]+(0.85094548724)*x[1]**o+(-0.369690624817)*x[1]
        arg[2,3,2,1]=(0.2339994493)*x[0]**o+(-0.901432708956)*x[0]+(0.309349741876)*x[1]**o+(0.132010344317)*x[1]
        arg[2,4,0,0]=(-0.713553783036)*x[0]**o+(0.791484993301)*x[0]+(0.339247480459)*x[1]**o+(0.158842307171)*x[1]
        arg[2,4,0,1]=(-0.733490022564)*x[0]**o+(0.631284830844)*x[0]+(0.340009298488)*x[1]**o+(-0.483162310574)*x[1]
        arg[2,4,1,0]=(0.709413437683)*x[0]**o+(0.896590124659)*x[0]+(0.680862759763)*x[1]**o+(-0.23579845663)*x[1]
        arg[2,4,1,1]=(-0.130761551034)*x[0]**o+(0.269095282394)*x[0]+(0.687201138134)*x[1]**o+(0.905183436376)*x[1]
        arg[2,4,2,0]=(0.256894911335)*x[0]**o+(0.706744915172)*x[0]+(0.910504833397)*x[1]**o+(-0.285088297895)*x[1]
        arg[2,4,2,1]=(0.95137247728)*x[0]**o+(-0.924669436646)*x[0]+(-0.146356272875)*x[1]**o+(-0.673326194162)*x[1]
        arg[3,0,0,0]=(0.663180444534)*x[0]**o+(0.894728781279)*x[0]+(0.894383747285)*x[1]**o+(0.969779862148)*x[1]
        arg[3,0,0,1]=(0.424556283891)*x[0]**o+(-0.97849198275)*x[0]+(0.367908872519)*x[1]**o+(0.624098355956)*x[1]
        arg[3,0,1,0]=(0.908698580251)*x[0]**o+(-0.899102004477)*x[0]+(0.509310001763)*x[1]**o+(-0.8399697646)*x[1]
        arg[3,0,1,1]=(0.523444225814)*x[0]**o+(0.410650511976)*x[0]+(-0.986699126405)*x[1]**o+(0.081494682537)*x[1]
        arg[3,0,2,0]=(-0.163604985314)*x[0]**o+(0.959338619715)*x[0]+(-0.843900271478)*x[1]**o+(-0.0218386544947)*x[1]
        arg[3,0,2,1]=(-0.490804271432)*x[0]**o+(0.123332672938)*x[0]+(0.902072835174)*x[1]**o+(-0.202340709957)*x[1]
        arg[3,1,0,0]=(-0.949731382685)*x[0]**o+(0.533695608969)*x[0]+(0.430512336347)*x[1]**o+(0.977434372776)*x[1]
        arg[3,1,0,1]=(0.945102298088)*x[0]**o+(-0.117807763531)*x[0]+(-0.203921504027)*x[1]**o+(-0.261266079346)*x[1]
        arg[3,1,1,0]=(0.762442770795)*x[0]**o+(-0.318456087863)*x[0]+(-0.365021977395)*x[1]**o+(0.0516932816311)*x[1]
        arg[3,1,1,1]=(-0.829061607399)*x[0]**o+(0.854900781365)*x[0]+(-0.397139314214)*x[1]**o+(-0.549679351555)*x[1]
        arg[3,1,2,0]=(-0.780040060687)*x[0]**o+(0.719393460042)*x[0]+(-0.435377419129)*x[1]**o+(0.655962063649)*x[1]
        arg[3,1,2,1]=(-0.853144390228)*x[0]**o+(0.0940222442927)*x[0]+(-0.642258121264)*x[1]**o+(-0.615386923437)*x[1]
        arg[3,2,0,0]=(0.377161814597)*x[0]**o+(0.349238572634)*x[0]+(0.361300611554)*x[1]**o+(0.554382639639)*x[1]
        arg[3,2,0,1]=(-0.51174671413)*x[0]**o+(-0.0586248632463)*x[0]+(0.083224329707)*x[1]**o+(0.506614699443)*x[1]
        arg[3,2,1,0]=(0.612695079311)*x[0]**o+(0.82053876378)*x[0]+(-0.132608876197)*x[1]**o+(0.701720563108)*x[1]
        arg[3,2,1,1]=(0.0928734170244)*x[0]**o+(-0.127564289866)*x[0]+(0.68468196965)*x[1]**o+(0.0168958305852)*x[1]
        arg[3,2,2,0]=(-0.863542893122)*x[0]**o+(0.0533007397457)*x[0]+(-0.0602194449612)*x[1]**o+(0.391122261927)*x[1]
        arg[3,2,2,1]=(-0.0942537498633)*x[0]**o+(-0.421741120475)*x[0]+(-0.0746938765947)*x[1]**o+(0.988111802567)*x[1]
        arg[3,3,0,0]=(-0.0588913579205)*x[0]**o+(0.439330145374)*x[0]+(-0.541714868844)*x[1]**o+(0.845344272279)*x[1]
        arg[3,3,0,1]=(0.221784155922)*x[0]**o+(0.983539923382)*x[0]+(-0.32332836636)*x[1]**o+(-0.892579743718)*x[1]
        arg[3,3,1,0]=(0.804097639203)*x[0]**o+(-0.342072604164)*x[0]+(0.755681096012)*x[1]**o+(0.432897246925)*x[1]
        arg[3,3,1,1]=(0.825696602707)*x[0]**o+(0.536999112965)*x[0]+(0.745033643035)*x[1]**o+(0.156806358776)*x[1]
        arg[3,3,2,0]=(-0.372114391352)*x[0]**o+(0.670417556114)*x[0]+(-0.569395510546)*x[1]**o+(-0.523784408534)*x[1]
        arg[3,3,2,1]=(0.439723359721)*x[0]**o+(0.809733189174)*x[0]+(0.781753556415)*x[1]**o+(0.375991547044)*x[1]
        arg[3,4,0,0]=(-0.429896412297)*x[0]**o+(-0.908820042346)*x[0]+(0.328787484151)*x[1]**o+(-0.610488251755)*x[1]
        arg[3,4,0,1]=(0.803567111353)*x[0]**o+(-0.487219379817)*x[0]+(-0.541538468561)*x[1]**o+(-0.692964506777)*x[1]
        arg[3,4,1,0]=(-0.963266835039)*x[0]**o+(0.667510333677)*x[0]+(0.308312799035)*x[1]**o+(0.686605377458)*x[1]
        arg[3,4,1,1]=(0.651198604373)*x[0]**o+(0.306465142929)*x[0]+(-0.923691546947)*x[1]**o+(-0.376240035765)*x[1]
        arg[3,4,2,0]=(-0.339406539705)*x[0]**o+(0.524709759622)*x[0]+(-0.0326564953763)*x[1]**o+(-0.658296780229)*x[1]
        arg[3,4,2,1]=(0.471355209305)*x[0]**o+(-0.473756698531)*x[0]+(0.193890658409)*x[1]**o+(-0.266632530307)*x[1]
        ref[0,0,0,0]=(0.765391656872)/(o+1.)+(-0.322334975868)+(0.415003066171)*0.5**o
        ref[0,0,0,1]=(-0.152730990885)/(o+1.)+(0.0623697786981)+(-0.928539602254)*0.5**o
        ref[0,0,1,0]=(0.505558859525)/(o+1.)+(0.0211473671443)+(-0.937831078575)*0.5**o
        ref[0,0,1,1]=(0.194087349442)/(o+1.)+(0.537213126576)+(0.909415074451)*0.5**o
        ref[0,0,2,0]=(0.518460161897)/(o+1.)+(-0.353834462413)+(0.935190448625)*0.5**o
        ref[0,0,2,1]=(0.578391007321)/(o+1.)+(0.277770233356)+(-0.94062714999)*0.5**o
        ref[0,1,0,0]=(-0.603956105825)/(o+1.)+(-0.612808261345)+(-0.197219678026)*0.5**o
        ref[0,1,0,1]=(-0.270612249495)/(o+1.)+(-0.0917787464331)+(-0.590715606611)*0.5**o
        ref[0,1,1,0]=(0.0875989587033)/(o+1.)+(0.333940978312)+(-0.365516638755)*0.5**o
        ref[0,1,1,1]=(0.625272530385)/(o+1.)+(-0.273562720598)+(0.315218558548)*0.5**o
        ref[0,1,2,0]=(0.115500534905)/(o+1.)+(-0.899119749149)+(-0.502497884169)*0.5**o
        ref[0,1,2,1]=(-0.0715650027129)/(o+1.)+(0.0286821705227)+(0.615837581258)*0.5**o
        ref[0,2,0,0]=(-0.532213103734)/(o+1.)+(0.270406536408)+(-0.898904577889)*0.5**o
        ref[0,2,0,1]=(0.751663691868)/(o+1.)+(0.0556640538013)+(-0.0056483770747)*0.5**o
        ref[0,2,1,0]=(-0.932579716748)/(o+1.)+(-0.364716190281)+(0.271196923039)*0.5**o
        ref[0,2,1,1]=(0.747625244884)/(o+1.)+(0.420842084614)+(-0.788206773125)*0.5**o
        ref[0,2,2,0]=(0.197684766271)/(o+1.)+(-0.889094904468)+(-0.787603701957)*0.5**o
        ref[0,2,2,1]=(-0.525306115268)/(o+1.)+(0.439959527228)+(0.417538087399)*0.5**o
        ref[0,3,0,0]=(-0.149386157207)/(o+1.)+(-0.289926844941)+(-0.230658480436)*0.5**o
        ref[0,3,0,1]=(-0.542257139282)/(o+1.)+(-0.0133255401565)+(-0.815859549425)*0.5**o
        ref[0,3,1,0]=(0.714019527678)/(o+1.)+(-0.00632729199566)+(0.0508974791413)*0.5**o
        ref[0,3,1,1]=(0.352455072404)/(o+1.)+(0.143383852019)+(0.119419200166)*0.5**o
        ref[0,3,2,0]=(-0.585094684368)/(o+1.)+(0.00921199711359)+(-0.82075904747)*0.5**o
        ref[0,3,2,1]=(0.686644239408)/(o+1.)+(0.468673187806)+(-0.748061395205)*0.5**o
        ref[0,4,0,0]=(0.129373873322)/(o+1.)+(0.607761864417)+(0.128293903485)*0.5**o
        ref[0,4,0,1]=(-0.170677596319)/(o+1.)+(-0.206814905431)+(0.639638672694)*0.5**o
        ref[0,4,1,0]=(0.282544628333)/(o+1.)+(-0.475291486231)+(0.43743097727)*0.5**o
        ref[0,4,1,1]=(0.218383977926)/(o+1.)+(0.43905900591)+(-0.995599965283)*0.5**o
        ref[0,4,2,0]=(-0.449331655188)/(o+1.)+(-0.0971131654902)+(0.922609534479)*0.5**o
        ref[0,4,2,1]=(-0.167639663907)/(o+1.)+(-0.103797419824)+(0.950186933361)*0.5**o
        ref[1,0,0,0]=(-0.622510875807)/(o+1.)+(0.186056784105)+(0.0573632292788)*0.5**o
        ref[1,0,0,1]=(0.469744047929)/(o+1.)+(-0.838331463751)+(0.368975600444)*0.5**o
        ref[1,0,1,0]=(0.625367069433)/(o+1.)+(-0.432363629616)+(-0.810194964715)*0.5**o
        ref[1,0,1,1]=(0.359324627118)/(o+1.)+(0.36983912542)+(0.744923275143)*0.5**o
        ref[1,0,2,0]=(0.318198109895)/(o+1.)+(-0.215392416287)+(0.891959004941)*0.5**o
        ref[1,0,2,1]=(-0.674805144805)/(o+1.)+(0.581833674632)+(0.842125439023)*0.5**o
        ref[1,1,0,0]=(-0.740263211102)/(o+1.)+(0.109311509194)+(-0.786719774805)*0.5**o
        ref[1,1,0,1]=(0.714032957488)/(o+1.)+(-0.81946337242)+(0.831245977993)*0.5**o
        ref[1,1,1,0]=(0.455949140326)/(o+1.)+(0.666183494578)+(0.862796574462)*0.5**o
        ref[1,1,1,1]=(-0.62474682088)/(o+1.)+(0.438060834031)+(-0.500409398892)*0.5**o
        ref[1,1,2,0]=(-0.770315341866)/(o+1.)+(-0.375748925732)+(-0.402886961189)*0.5**o
        ref[1,1,2,1]=(0.654162491323)/(o+1.)+(0.189164682023)+(0.358161520677)*0.5**o
        ref[1,2,0,0]=(-0.0538737948754)/(o+1.)+(0.333813880332)+(-0.15539934232)*0.5**o
        ref[1,2,0,1]=(-0.377617770006)/(o+1.)+(-0.0218224426646)+(0.371159212282)*0.5**o
        ref[1,2,1,0]=(-0.360936107893)/(o+1.)+(0.141978404604)+(0.117790556271)*0.5**o
        ref[1,2,1,1]=(0.183316445987)/(o+1.)+(-0.0416656990415)+(-0.524828128971)*0.5**o
        ref[1,2,2,0]=(0.126601242737)/(o+1.)+(-0.0623520339976)+(0.300516177535)*0.5**o
        ref[1,2,2,1]=(-0.810227948574)/(o+1.)+(-0.387092926354)+(-0.29046926624)*0.5**o
        ref[1,3,0,0]=(0.757723452446)/(o+1.)+(0.293889998642)+(-0.134275834453)*0.5**o
        ref[1,3,0,1]=(0.503482692135)/(o+1.)+(-0.125704830059)+(0.882915728842)*0.5**o
        ref[1,3,1,0]=(-0.861747803918)/(o+1.)+(-0.0802560425765)+(-0.954792910945)*0.5**o
        ref[1,3,1,1]=(-0.322274864209)/(o+1.)+(-0.226098024118)+(0.747255393187)*0.5**o
        ref[1,3,2,0]=(-0.488288412308)/(o+1.)+(0.0635632421511)+(0.139287057836)*0.5**o
        ref[1,3,2,1]=(0.206644380777)/(o+1.)+(0.323520058859)+(-0.861625516304)*0.5**o
        ref[1,4,0,0]=(-0.971680855409)/(o+1.)+(0.0947423509032)+(-0.714198122614)*0.5**o
        ref[1,4,0,1]=(0.347408878778)/(o+1.)+(-0.709724935569)+(0.293578109913)*0.5**o
        ref[1,4,1,0]=(-0.284086603407)/(o+1.)+(-0.116791420591)+(0.843760126513)*0.5**o
        ref[1,4,1,1]=(0.537701123084)/(o+1.)+(-0.482717976859)+(-0.342911357374)*0.5**o
        ref[1,4,2,0]=(0.831438049852)/(o+1.)+(0.056404572442)+(-0.71201360314)*0.5**o
        ref[1,4,2,1]=(-0.0369972206543)/(o+1.)+(0.275008406103)+(-0.616403541928)*0.5**o
        ref[2,0,0,0]=(-0.464334304695)/(o+1.)+(0.132370231466)+(-0.959754423491)*0.5**o
        ref[2,0,0,1]=(0.173044245392)/(o+1.)+(0.437195842274)+(0.231051530608)*0.5**o
        ref[2,0,1,0]=(-0.816201949845)/(o+1.)+(-0.72616145946)+(0.376577055194)*0.5**o
        ref[2,0,1,1]=(-0.876294548464)/(o+1.)+(0.128100848943)+(-0.4891837531)*0.5**o
        ref[2,0,2,0]=(0.602375475355)/(o+1.)+(-0.497981505761)+(-0.759548081841)*0.5**o
        ref[2,0,2,1]=(0.813648191424)/(o+1.)+(-0.619848910412)+(0.114018259802)*0.5**o
        ref[2,1,0,0]=(-0.12122456959)/(o+1.)+(0.631472305971)+(-0.572699800518)*0.5**o
        ref[2,1,0,1]=(0.539760655497)/(o+1.)+(0.43429478244)+(0.0655268452799)*0.5**o
        ref[2,1,1,0]=(-0.0242249415295)/(o+1.)+(-0.263311349051)+(0.358281119336)*0.5**o
        ref[2,1,1,1]=(-0.995787126907)/(o+1.)+(-0.370899595832)+(0.523751704889)*0.5**o
        ref[2,1,2,0]=(0.149400154478)/(o+1.)+(-0.107907886588)+(-0.185380508646)*0.5**o
        ref[2,1,2,1]=(0.908525646396)/(o+1.)+(-0.18328120728)+(0.81086932122)*0.5**o
        ref[2,2,0,0]=(-0.0495088159461)/(o+1.)+(-0.246369371749)+(-0.463416100256)*0.5**o
        ref[2,2,0,1]=(-0.895280577411)/(o+1.)+(-0.0345629407446)+(-0.200698149193)*0.5**o
        ref[2,2,1,0]=(-0.375709672588)/(o+1.)+(-0.126514661136)+(0.888683772867)*0.5**o
        ref[2,2,1,1]=(0.868194479789)/(o+1.)+(-0.239689630708)+(-0.0167949020901)*0.5**o
        ref[2,2,2,0]=(0.267937553127)/(o+1.)+(-0.425739316963)+(0.441981698578)*0.5**o
        ref[2,2,2,1]=(0.697645775876)/(o+1.)+(-0.16217258223)+(0.876701424579)*0.5**o
        ref[2,3,0,0]=(-0.0706155476337)/(o+1.)+(0.483772571196)+(0.111231509217)*0.5**o
        ref[2,3,0,1]=(0.326282512878)/(o+1.)+(0.759423156855)+(-0.858913402812)*0.5**o
        ref[2,3,1,0]=(0.575866251831)/(o+1.)+(-0.776139286869)+(-0.524551221953)*0.5**o
        ref[2,3,1,1]=(-0.488263373008)/(o+1.)+(0.0931216925076)+(0.391514075129)*0.5**o
        ref[2,3,2,0]=(0.85094548724)/(o+1.)+(-0.0708709485834)+(0.26532129161)*0.5**o
        ref[2,3,2,1]=(0.309349741876)/(o+1.)+(-0.384711182319)+(0.2339994493)*0.5**o
        ref[2,4,0,0]=(0.339247480459)/(o+1.)+(0.475163650236)+(-0.713553783036)*0.5**o
        ref[2,4,0,1]=(0.340009298488)/(o+1.)+(0.0740612601349)+(-0.733490022564)*0.5**o
        ref[2,4,1,0]=(0.680862759763)/(o+1.)+(0.330395834014)+(0.709413437683)*0.5**o
        ref[2,4,1,1]=(0.687201138134)/(o+1.)+(0.587139359385)+(-0.130761551034)*0.5**o
        ref[2,4,2,0]=(0.910504833397)/(o+1.)+(0.210828308638)+(0.256894911335)*0.5**o
        ref[2,4,2,1]=(-0.146356272875)/(o+1.)+(-0.798997815404)+(0.95137247728)*0.5**o
        ref[3,0,0,0]=(0.894383747285)/(o+1.)+(0.932254321713)+(0.663180444534)*0.5**o
        ref[3,0,0,1]=(0.367908872519)/(o+1.)+(-0.177196813397)+(0.424556283891)*0.5**o
        ref[3,0,1,0]=(0.509310001763)/(o+1.)+(-0.869535884539)+(0.908698580251)*0.5**o
        ref[3,0,1,1]=(-0.986699126405)/(o+1.)+(0.246072597257)+(0.523444225814)*0.5**o
        ref[3,0,2,0]=(-0.843900271478)/(o+1.)+(0.46874998261)+(-0.163604985314)*0.5**o
        ref[3,0,2,1]=(0.902072835174)/(o+1.)+(-0.0395040185096)+(-0.490804271432)*0.5**o
        ref[3,1,0,0]=(0.430512336347)/(o+1.)+(0.755564990872)+(-0.949731382685)*0.5**o
        ref[3,1,0,1]=(-0.203921504027)/(o+1.)+(-0.189536921439)+(0.945102298088)*0.5**o
        ref[3,1,1,0]=(-0.365021977395)/(o+1.)+(-0.133381403116)+(0.762442770795)*0.5**o
        ref[3,1,1,1]=(-0.397139314214)/(o+1.)+(0.152610714905)+(-0.829061607399)*0.5**o
        ref[3,1,2,0]=(-0.435377419129)/(o+1.)+(0.687677761846)+(-0.780040060687)*0.5**o
        ref[3,1,2,1]=(-0.642258121264)/(o+1.)+(-0.260682339572)+(-0.853144390228)*0.5**o
        ref[3,2,0,0]=(0.361300611554)/(o+1.)+(0.451810606136)+(0.377161814597)*0.5**o
        ref[3,2,0,1]=(0.083224329707)/(o+1.)+(0.223994918098)+(-0.51174671413)*0.5**o
        ref[3,2,1,0]=(-0.132608876197)/(o+1.)+(0.761129663444)+(0.612695079311)*0.5**o
        ref[3,2,1,1]=(0.68468196965)/(o+1.)+(-0.0553342296404)+(0.0928734170244)*0.5**o
        ref[3,2,2,0]=(-0.0602194449612)/(o+1.)+(0.222211500837)+(-0.863542893122)*0.5**o
        ref[3,2,2,1]=(-0.0746938765947)/(o+1.)+(0.283185341046)+(-0.0942537498633)*0.5**o
        ref[3,3,0,0]=(-0.541714868844)/(o+1.)+(0.642337208826)+(-0.0588913579205)*0.5**o
        ref[3,3,0,1]=(-0.32332836636)/(o+1.)+(0.0454800898319)+(0.221784155922)*0.5**o
        ref[3,3,1,0]=(0.755681096012)/(o+1.)+(0.0454123213802)+(0.804097639203)*0.5**o
        ref[3,3,1,1]=(0.745033643035)/(o+1.)+(0.34690273587)+(0.825696602707)*0.5**o
        ref[3,3,2,0]=(-0.569395510546)/(o+1.)+(0.0733165737898)+(-0.372114391352)*0.5**o
        ref[3,3,2,1]=(0.781753556415)/(o+1.)+(0.592862368109)+(0.439723359721)*0.5**o
        ref[3,4,0,0]=(0.328787484151)/(o+1.)+(-0.75965414705)+(-0.429896412297)*0.5**o
        ref[3,4,0,1]=(-0.541538468561)/(o+1.)+(-0.590091943297)+(0.803567111353)*0.5**o
        ref[3,4,1,0]=(0.308312799035)/(o+1.)+(0.677057855567)+(-0.963266835039)*0.5**o
        ref[3,4,1,1]=(-0.923691546947)/(o+1.)+(-0.0348874464182)+(0.651198604373)*0.5**o
        ref[3,4,2,0]=(-0.0326564953763)/(o+1.)+(-0.0667935103036)+(-0.339406539705)*0.5**o
        ref[3,4,2,1]=(0.193890658409)/(o+1.)+(-0.370194614419)+(0.471355209305)*0.5**o
      else:
        arg[0,0,0,0]=(0.0737627367843)*x[0]**o+(-0.897508234376)*x[0]+(0.499361502035)*x[1]**o+(-0.830793706142)*x[1]+(0.967672023686)*x[2]**o+(-0.94209897356)*x[2]
        arg[0,0,0,1]=(0.353968533247)*x[0]**o+(-0.828575104081)*x[0]+(0.111197799977)*x[1]**o+(0.475627052455)*x[1]+(0.10233739786)*x[2]**o+(-0.297090205985)*x[2]
        arg[0,0,1,0]=(0.315704665816)*x[0]**o+(0.450577419574)*x[0]+(-0.0225047225996)*x[1]**o+(0.221436098936)*x[1]+(-0.210708820992)*x[2]**o+(0.22813686037)*x[2]
        arg[0,0,1,1]=(-0.453912049362)*x[0]**o+(0.443266972459)*x[0]+(-0.166003491301)*x[1]**o+(0.325157034345)*x[1]+(0.975435329063)*x[2]**o+(0.516477372955)*x[2]
        arg[0,0,2,0]=(0.705225432895)*x[0]**o+(0.406870556161)*x[0]+(0.967131048725)*x[1]**o+(-0.31111693058)*x[1]+(0.915125425876)*x[2]**o+(-0.417407852906)*x[2]
        arg[0,0,2,1]=(-0.397162258761)*x[0]**o+(-0.49079016751)*x[0]+(-0.444905720899)*x[1]**o+(0.258998080457)*x[1]+(-0.406172174923)*x[2]**o+(0.166620074269)*x[2]
        arg[0,1,0,0]=(0.00444300957662)*x[0]**o+(0.59448685984)*x[0]+(-0.755596203939)*x[1]**o+(-0.587866578467)*x[1]+(-0.994620629569)*x[2]**o+(0.558539124849)*x[2]
        arg[0,1,0,1]=(0.726535641072)*x[0]**o+(-0.515254556008)*x[0]+(-0.737095044797)*x[1]**o+(0.396218854768)*x[1]+(-0.873340093434)*x[2]**o+(0.538194586759)*x[2]
        arg[0,1,1,0]=(-0.554579794515)*x[0]**o+(-0.754117891877)*x[0]+(0.242579027102)*x[1]**o+(0.368307663104)*x[1]+(0.308372932825)*x[2]**o+(-0.521440130826)*x[2]
        arg[0,1,1,1]=(0.294744137189)*x[0]**o+(0.44996935134)*x[0]+(0.118263237453)*x[1]**o+(-0.143566244715)*x[1]+(-0.927385432821)*x[2]**o+(0.912735237305)*x[2]
        arg[0,1,2,0]=(-0.30261020867)*x[0]**o+(-0.994883662031)*x[0]+(0.514662053738)*x[1]**o+(0.327770859964)*x[1]+(0.862452368458)*x[2]**o+(-0.739026842429)*x[2]
        arg[0,1,2,1]=(0.494855964972)*x[0]**o+(-0.358461686787)*x[0]+(0.406489619959)*x[1]**o+(0.417722618206)*x[1]+(0.679808195501)*x[2]**o+(-0.00502769522311)*x[2]
        arg[0,2,0,0]=(-0.914585215232)*x[0]**o+(0.305890296017)*x[0]+(-0.867826544668)*x[1]**o+(0.304580882671)*x[1]+(-0.401401956053)*x[2]**o+(0.842276423411)*x[2]
        arg[0,2,0,1]=(-0.260590397572)*x[0]**o+(0.541867335986)*x[0]+(0.0279475511153)*x[1]**o+(0.0255320376447)*x[1]+(-0.584584718877)*x[2]**o+(-0.640388032783)*x[2]
        arg[0,2,1,0]=(0.0455797513404)*x[0]**o+(-0.764984318339)*x[0]+(-0.257174640639)*x[1]**o+(0.745888292962)*x[1]+(0.964524202534)*x[2]**o+(0.435680191282)*x[2]
        arg[0,2,1,1]=(-0.300900136798)*x[0]**o+(-0.899164082043)*x[0]+(-0.63273853331)*x[1]**o+(0.56522562781)*x[1]+(-0.846482371222)*x[2]**o+(-0.909985752269)*x[2]
        arg[0,2,2,0]=(0.745711773093)*x[0]**o+(0.0627291150681)*x[0]+(-0.090340386706)*x[1]**o+(-0.0912293084958)*x[1]+(0.108613205226)*x[2]**o+(-0.28985444593)*x[2]
        arg[0,2,2,1]=(0.291340077334)*x[0]**o+(0.592856577974)*x[0]+(0.141857446535)*x[1]**o+(0.830526360843)*x[1]+(0.956840406828)*x[2]**o+(-0.717465405888)*x[2]
        arg[0,3,0,0]=(0.89242099254)*x[0]**o+(0.529348654364)*x[0]+(-0.074382350183)*x[1]**o+(0.874670648345)*x[1]+(-0.728104560168)*x[2]**o+(0.0781153963606)*x[2]
        arg[0,3,0,1]=(0.095390666498)*x[0]**o+(0.0124610436496)*x[0]+(0.278014506594)*x[1]**o+(0.923928359216)*x[1]+(-0.0905667357972)*x[2]**o+(-0.894031439498)*x[2]
        arg[0,3,1,0]=(-0.317750910722)*x[0]**o+(0.677406387747)*x[0]+(0.550094997541)*x[1]**o+(-0.30098732949)*x[1]+(0.501753868155)*x[2]**o+(-0.42859127752)*x[2]
        arg[0,3,1,1]=(-0.204455448392)*x[0]**o+(0.445237113438)*x[0]+(0.853596204887)*x[1]**o+(0.315701645178)*x[1]+(0.307192864939)*x[2]**o+(-0.708325809173)*x[2]
        arg[0,3,2,0]=(0.38581656579)*x[0]**o+(-0.79289714243)*x[0]+(-0.412226202311)*x[1]**o+(0.655534112291)*x[1]+(0.31352728014)*x[2]**o+(0.502089385396)*x[2]
        arg[0,3,2,1]=(-0.225571343976)*x[0]**o+(0.687765563016)*x[0]+(0.0421378888728)*x[1]**o+(0.54466301835)*x[1]+(-0.430288255786)*x[2]**o+(-0.117571440224)*x[2]
        arg[0,4,0,0]=(0.834083057621)*x[0]**o+(0.0292441566815)*x[0]+(-0.947154369416)*x[1]**o+(0.597153808995)*x[1]+(0.323706009194)*x[2]**o+(0.664093182644)*x[2]
        arg[0,4,0,1]=(-0.613152578007)*x[0]**o+(0.253155875339)*x[0]+(0.297138228345)*x[1]**o+(-0.604393247773)*x[1]+(-0.936718722281)*x[2]**o+(-0.410140848312)*x[2]
        arg[0,4,1,0]=(-0.373668614153)*x[0]**o+(-0.401240526097)*x[0]+(-0.417458139032)*x[1]**o+(0.456033864594)*x[1]+(-0.158434012225)*x[2]**o+(-0.33938136039)*x[2]
        arg[0,4,1,1]=(0.00943802199128)*x[0]**o+(-0.800756064115)*x[0]+(0.899199671858)*x[1]**o+(0.376830611387)*x[1]+(-0.930822199046)*x[2]**o+(0.192235807799)*x[2]
        arg[0,4,2,0]=(0.157099024354)*x[0]**o+(0.195942080381)*x[0]+(-0.554519054021)*x[1]**o+(0.871823312361)*x[1]+(-0.34803033039)*x[2]**o+(-0.646933331859)*x[2]
        arg[0,4,2,1]=(0.933762543076)*x[0]**o+(-0.169843985744)*x[0]+(0.971089176971)*x[1]**o+(-0.605176382286)*x[1]+(0.254369209427)*x[2]**o+(0.0314820745541)*x[2]
        arg[1,0,0,0]=(0.0417476479889)*x[0]**o+(-0.144293810851)*x[0]+(-0.677039420744)*x[1]**o+(-0.137097132377)*x[1]+(0.721082639859)*x[2]**o+(-0.32510577691)*x[2]
        arg[1,0,0,1]=(0.450986977446)*x[0]**o+(-0.921371106219)*x[0]+(0.536898979198)*x[1]**o+(-0.814951108186)*x[1]+(-0.733774506956)*x[2]**o+(0.938941747764)*x[2]
        arg[1,0,1,0]=(-0.23712795738)*x[0]**o+(0.450883240541)*x[0]+(0.101102533189)*x[1]**o+(-0.998253664232)*x[1]+(-0.843977488601)*x[2]**o+(-0.807025450448)*x[2]
        arg[1,0,1,1]=(-0.308721170682)*x[0]**o+(-0.546268734305)*x[0]+(0.643628083343)*x[1]**o+(0.206934468523)*x[1]+(-0.257667927436)*x[2]**o+(0.468860780891)*x[2]
        arg[1,0,2,0]=(-0.267232319593)*x[0]**o+(0.221737693042)*x[0]+(0.202218228468)*x[1]**o+(-0.560941556406)*x[1]+(0.910244160614)*x[2]**o+(0.361584784536)*x[2]
        arg[1,0,2,1]=(0.405254374432)*x[0]**o+(-0.102574362257)*x[0]+(0.527874164745)*x[1]**o+(0.0318395137256)*x[1]+(-0.448050293943)*x[2]**o+(-0.6401160036)*x[2]
        arg[1,1,0,0]=(-0.877224972777)*x[0]**o+(-0.544175806369)*x[0]+(0.528037913494)*x[1]**o+(-0.477854554671)*x[1]+(-0.394875438861)*x[2]**o+(-0.704656012962)*x[2]
        arg[1,1,0,1]=(-0.873928172379)*x[0]**o+(0.273455449944)*x[0]+(0.681090480547)*x[1]**o+(0.883243766093)*x[1]+(0.557357114631)*x[2]**o+(0.586148966774)*x[2]
        arg[1,1,1,0]=(0.552826305913)*x[0]**o+(-0.277269605193)*x[0]+(-0.681872908424)*x[1]**o+(-0.84578699459)*x[1]+(0.6543624385)*x[2]**o+(-0.355051513153)*x[2]
        arg[1,1,1,1]=(-0.71233607947)*x[0]**o+(-0.779861555179)*x[0]+(0.11496674995)*x[1]**o+(0.634901982779)*x[1]+(0.833044137166)*x[2]**o+(0.0443514408009)*x[2]
        arg[1,1,2,0]=(-0.411667274338)*x[0]**o+(0.218709061898)*x[0]+(0.132573928273)*x[1]**o+(-0.450677430962)*x[1]+(0.88420873514)*x[2]**o+(-0.618513305411)*x[2]
        arg[1,1,2,1]=(0.533005373575)*x[0]**o+(0.846343252936)*x[0]+(0.697076801609)*x[1]**o+(0.342260840742)*x[1]+(0.923380136498)*x[2]**o+(-0.727450493778)*x[2]
        arg[1,2,0,0]=(-0.637366803484)*x[0]**o+(-0.654649539562)*x[0]+(-0.443035194144)*x[1]**o+(0.4928304386)*x[1]+(0.889290653473)*x[2]**o+(0.208513517091)*x[2]
        arg[1,2,0,1]=(-0.404870452835)*x[0]**o+(-0.556431486625)*x[0]+(-0.428885935216)*x[1]**o+(0.208771213187)*x[1]+(0.706109349268)*x[2]**o+(0.726168991522)*x[2]
        arg[1,2,1,0]=(0.85574176569)*x[0]**o+(-0.371989246254)*x[0]+(0.0376283091444)*x[1]**o+(-0.164811219077)*x[1]+(0.842513532253)*x[2]**o+(-0.963454790343)*x[2]
        arg[1,2,1,1]=(0.576825533647)*x[0]**o+(0.866570909412)*x[0]+(-0.286228055595)*x[1]**o+(-0.509258889664)*x[1]+(0.358087406384)*x[2]**o+(0.2558870556)*x[2]
        arg[1,2,2,0]=(0.741411326194)*x[0]**o+(0.887444210472)*x[0]+(0.176659455323)*x[1]**o+(-0.292731470152)*x[1]+(0.179836617557)*x[2]**o+(-0.10894424483)*x[2]
        arg[1,2,2,1]=(0.386515069261)*x[0]**o+(-0.423389396097)*x[0]+(0.300384913804)*x[1]**o+(-0.569308908813)*x[1]+(0.845849686169)*x[2]**o+(-0.802483942058)*x[2]
        arg[1,3,0,0]=(0.848952965744)*x[0]**o+(0.113555873929)*x[0]+(0.132168122705)*x[1]**o+(0.464116031047)*x[1]+(-0.928746352905)*x[2]**o+(-0.176374657603)*x[2]
        arg[1,3,0,1]=(0.357870838328)*x[0]**o+(-0.294290586373)*x[0]+(0.395859596543)*x[1]**o+(-0.265635871788)*x[1]+(0.283887401402)*x[2]**o+(0.500378090749)*x[2]
        arg[1,3,1,0]=(-0.509333127149)*x[0]**o+(0.328856435768)*x[0]+(-0.377387784476)*x[1]**o+(-0.878505533565)*x[1]+(0.448760304667)*x[2]**o+(0.996692047048)*x[2]
        arg[1,3,1,1]=(-0.837885550386)*x[0]**o+(0.0608579436757)*x[0]+(-0.627132023609)*x[1]**o+(0.902538475491)*x[1]+(-0.449596037062)*x[2]**o+(-0.158786724549)*x[2]
        arg[1,3,2,0]=(0.205889276236)*x[0]**o+(0.576561726109)*x[0]+(-0.00281354319689)*x[1]**o+(0.214968600999)*x[1]+(0.811758703832)*x[2]**o+(0.260164242506)*x[2]
        arg[1,3,2,1]=(-0.331660060588)*x[0]**o+(0.713879758554)*x[0]+(0.804549392751)*x[1]**o+(-0.346397636822)*x[1]+(-0.364073503513)*x[2]**o+(0.338612348994)*x[2]
        arg[1,4,0,0]=(0.394201404247)*x[0]**o+(-0.416853395415)*x[0]+(-0.0581465874206)*x[1]**o+(0.105267182588)*x[1]+(0.791799718009)*x[2]**o+(0.0840172853234)*x[2]
        arg[1,4,0,1]=(-0.346144268453)*x[0]**o+(-0.661228785728)*x[0]+(0.326588395606)*x[1]**o+(0.766137046896)*x[1]+(0.566377202267)*x[2]**o+(0.613565461425)*x[2]
        arg[1,4,1,0]=(-0.907566140749)*x[0]**o+(-0.391612881392)*x[0]+(0.263322351365)*x[1]**o+(-0.4373853914)*x[1]+(0.248855925063)*x[2]**o+(0.973515668505)*x[2]
        arg[1,4,1,1]=(0.726194147182)*x[0]**o+(0.497593423927)*x[0]+(-0.951344923018)*x[1]**o+(0.55327321528)*x[1]+(-0.266195110749)*x[2]**o+(-0.310893832826)*x[2]
        arg[1,4,2,0]=(-0.940824546553)*x[0]**o+(-0.47838524024)*x[0]+(0.408728941942)*x[1]**o+(0.627300786998)*x[1]+(-0.0547545650101)*x[2]**o+(0.433290388525)*x[2]
        arg[1,4,2,1]=(0.728499572603)*x[0]**o+(-0.232088507059)*x[0]+(0.0550057703801)*x[1]**o+(0.159356057226)*x[1]+(0.111191390097)*x[2]**o+(-0.64498018454)*x[2]
        arg[2,0,0,0]=(0.467025642879)*x[0]**o+(-0.695379603222)*x[0]+(0.591732172092)*x[1]**o+(0.84325648142)*x[1]+(0.965696472916)*x[2]**o+(0.626196181241)*x[2]
        arg[2,0,0,1]=(-0.211129222294)*x[0]**o+(-0.501208172412)*x[0]+(-0.0162146409175)*x[1]**o+(-0.672364205306)*x[1]+(-0.851073479662)*x[2]**o+(-0.817568385627)*x[2]
        arg[2,0,1,0]=(-0.668065295797)*x[0]**o+(-0.90359857855)*x[0]+(0.040519772083)*x[1]**o+(-0.295766307198)*x[1]+(0.0491234828524)*x[2]**o+(-0.589557178985)*x[2]
        arg[2,0,1,1]=(0.928751918298)*x[0]**o+(-0.773787731767)*x[0]+(-0.744325160449)*x[1]**o+(-0.426987905682)*x[1]+(0.503339956851)*x[2]**o+(0.386756648382)*x[2]
        arg[2,0,2,0]=(-0.531818105049)*x[0]**o+(-0.586811490681)*x[0]+(0.780379352762)*x[1]**o+(-0.5575328003)*x[1]+(0.0354867101462)*x[2]**o+(0.581226069105)*x[2]
        arg[2,0,2,1]=(0.0150988200159)*x[0]**o+(0.56415563975)*x[0]+(0.540723773613)*x[1]**o+(0.0653377679637)*x[1]+(-0.52633696371)*x[2]**o+(-0.122811009405)*x[2]
        arg[2,1,0,0]=(-0.0829154151121)*x[0]**o+(-0.416192198737)*x[0]+(0.944232067474)*x[1]**o+(-0.159852455491)*x[1]+(-0.198397797188)*x[2]**o+(-0.684211447305)*x[2]
        arg[2,1,0,1]=(-0.834502772339)*x[0]**o+(-0.856817328109)*x[0]+(0.213303574315)*x[1]**o+(-0.283140791695)*x[1]+(-0.727061494352)*x[2]**o+(0.12871372765)*x[2]
        arg[2,1,1,0]=(0.247439924718)*x[0]**o+(-0.912464825903)*x[0]+(0.791177764168)*x[1]**o+(0.160047921728)*x[1]+(0.591086305304)*x[2]**o+(-0.774025941232)*x[2]
        arg[2,1,1,1]=(-0.120169534728)*x[0]**o+(-0.747254680906)*x[0]+(-0.394003570933)*x[1]**o+(-0.938198960695)*x[1]+(-0.0142776288517)*x[2]**o+(0.0627959740849)*x[2]
        arg[2,1,2,0]=(0.703841647274)*x[0]**o+(-0.946039767174)*x[0]+(-0.836717609846)*x[1]**o+(0.776573515772)*x[1]+(0.250502751027)*x[2]**o+(0.403896552902)*x[2]
        arg[2,1,2,1]=(-0.19760048131)*x[0]**o+(-0.250991189554)*x[0]+(0.278810924634)*x[1]**o+(-0.580148097047)*x[1]+(-0.197518781697)*x[2]**o+(0.116835806412)*x[2]
        arg[2,2,0,0]=(-0.507872937196)*x[0]**o+(-0.771841217512)*x[0]+(0.852666314762)*x[1]**o+(-0.482947444136)*x[1]+(0.805824662389)*x[2]**o+(-0.299255374349)*x[2]
        arg[2,2,0,1]=(-0.481403986338)*x[0]**o+(0.904322169038)*x[0]+(0.937553651477)*x[1]**o+(0.727763542606)*x[1]+(-0.69485539356)*x[2]**o+(0.547755372933)*x[2]
        arg[2,2,1,0]=(0.241675645258)*x[0]**o+(0.957681326492)*x[0]+(-0.305747218513)*x[1]**o+(-0.0393749799535)*x[1]+(-0.448892234759)*x[2]**o+(0.707429367761)*x[2]
        arg[2,2,1,1]=(0.114772627763)*x[0]**o+(-0.411147710203)*x[0]+(0.13389320737)*x[1]**o+(0.806309117051)*x[1]+(-0.572620769665)*x[2]**o+(0.0389925355751)*x[2]
        arg[2,2,2,0]=(0.887023209688)*x[0]**o+(0.158261816385)*x[0]+(0.687399231822)*x[1]**o+(-0.858491519094)*x[1]+(-0.00348150506922)*x[2]**o+(-0.863496570874)*x[2]
        arg[2,2,2,1]=(0.836731324893)*x[0]**o+(0.975472534225)*x[0]+(0.407275572226)*x[1]**o+(0.565132026147)*x[1]+(0.79110912675)*x[2]**o+(-0.0816802575667)*x[2]
        arg[2,3,0,0]=(0.196161020545)*x[0]**o+(0.254301467318)*x[0]+(0.759363127106)*x[1]**o+(0.536320544038)*x[1]+(0.522705219849)*x[2]**o+(0.376531450811)*x[2]
        arg[2,3,0,1]=(-0.862326385376)*x[0]**o+(0.766714365146)*x[0]+(0.279327620722)*x[1]**o+(-0.44416461578)*x[1]+(0.115982232563)*x[2]**o+(0.824064889151)*x[2]
        arg[2,3,1,0]=(0.571263839862)*x[0]**o+(-0.679998793927)*x[0]+(0.63548515376)*x[1]**o+(0.464171242163)*x[1]+(0.712124752567)*x[2]**o+(0.913131755622)*x[2]
        arg[2,3,1,1]=(0.272559456412)*x[0]**o+(0.0577748172306)*x[0]+(-0.610464803438)*x[1]**o+(0.233987628538)*x[1]+(0.904147555826)*x[2]**o+(-0.131245442847)*x[2]
        arg[2,3,2,0]=(0.3409363126)*x[0]**o+(0.40060221588)*x[0]+(-0.603551487822)*x[1]**o+(-0.697430353793)*x[1]+(-0.867009193249)*x[2]**o+(-0.178991284979)*x[2]
        arg[2,3,2,1]=(0.949170673623)*x[0]**o+(-0.0217541208215)*x[0]+(-0.775307373317)*x[1]**o+(0.421565612874)*x[1]+(0.419254478365)*x[2]**o+(-0.732203964657)*x[2]
        arg[2,4,0,0]=(0.876774885889)*x[0]**o+(0.441155536177)*x[0]+(-0.292113312365)*x[1]**o+(0.880590008772)*x[1]+(0.0732363738743)*x[2]**o+(-0.590855067487)*x[2]
        arg[2,4,0,1]=(0.347441329383)*x[0]**o+(-0.96566344519)*x[0]+(-0.0870932691958)*x[1]**o+(-0.793898415692)*x[1]+(0.911757004621)*x[2]**o+(0.623070769788)*x[2]
        arg[2,4,1,0]=(0.732123350436)*x[0]**o+(-0.707615856771)*x[0]+(-0.57969408421)*x[1]**o+(0.0643464469617)*x[1]+(-0.648380091597)*x[2]**o+(0.567100447405)*x[2]
        arg[2,4,1,1]=(-0.128872747083)*x[0]**o+(-0.463336399393)*x[0]+(-0.1679086419)*x[1]**o+(0.602218708561)*x[1]+(0.62788576637)*x[2]**o+(0.223548351914)*x[2]
        arg[2,4,2,0]=(0.619034818105)*x[0]**o+(0.167994824318)*x[0]+(-0.244892750219)*x[1]**o+(0.471338499791)*x[1]+(0.526950200066)*x[2]**o+(0.345611623572)*x[2]
        arg[2,4,2,1]=(0.645979236759)*x[0]**o+(0.79896884484)*x[0]+(-0.587486435644)*x[1]**o+(-0.467838428827)*x[1]+(0.672707985612)*x[2]**o+(-0.169680966884)*x[2]
        arg[3,0,0,0]=(-0.265836015694)*x[0]**o+(-0.383762854285)*x[0]+(0.420375720992)*x[1]**o+(0.739662661302)*x[1]+(-0.593559713248)*x[2]**o+(0.0743208592009)*x[2]
        arg[3,0,0,1]=(0.306246512933)*x[0]**o+(0.0357065903179)*x[0]+(0.213209244461)*x[1]**o+(-0.069836605971)*x[1]+(0.668535341179)*x[2]**o+(0.59870885392)*x[2]
        arg[3,0,1,0]=(0.509420593773)*x[0]**o+(0.995859415429)*x[0]+(-0.07734119959)*x[1]**o+(0.269006992126)*x[1]+(-0.465576902019)*x[2]**o+(0.318766487761)*x[2]
        arg[3,0,1,1]=(-0.609626065482)*x[0]**o+(0.970976299205)*x[0]+(-0.314734168176)*x[1]**o+(-0.536438958228)*x[1]+(0.900096595598)*x[2]**o+(-0.893816431641)*x[2]
        arg[3,0,2,0]=(0.269212290711)*x[0]**o+(-0.119897130167)*x[0]+(-0.937350065698)*x[1]**o+(0.913313769348)*x[1]+(0.311695335026)*x[2]**o+(0.58822688023)*x[2]
        arg[3,0,2,1]=(-0.445116541697)*x[0]**o+(-0.64630078598)*x[0]+(0.811173945666)*x[1]**o+(-0.910503175962)*x[1]+(0.160819266078)*x[2]**o+(-0.807050548571)*x[2]
        arg[3,1,0,0]=(0.25585673506)*x[0]**o+(-0.174293490388)*x[0]+(-0.510750951271)*x[1]**o+(-0.215780338673)*x[1]+(-0.531739311594)*x[2]**o+(0.479569840128)*x[2]
        arg[3,1,0,1]=(-0.159301665223)*x[0]**o+(0.88558738006)*x[0]+(-0.933255168276)*x[1]**o+(0.739621529924)*x[1]+(-0.371501556668)*x[2]**o+(-0.224043643673)*x[2]
        arg[3,1,1,0]=(0.673375237931)*x[0]**o+(0.29315421593)*x[0]+(0.378983835599)*x[1]**o+(-0.187220470447)*x[1]+(-0.122366431969)*x[2]**o+(-0.178182488821)*x[2]
        arg[3,1,1,1]=(0.273992606955)*x[0]**o+(-0.42302685869)*x[0]+(0.490650718142)*x[1]**o+(0.067822229282)*x[1]+(0.758676637186)*x[2]**o+(-0.840062442278)*x[2]
        arg[3,1,2,0]=(-0.533129977534)*x[0]**o+(0.743362501638)*x[0]+(0.0397932337285)*x[1]**o+(0.134261172942)*x[1]+(0.076467475088)*x[2]**o+(0.750124646154)*x[2]
        arg[3,1,2,1]=(0.977824314677)*x[0]**o+(0.322104772509)*x[0]+(0.478366489119)*x[1]**o+(0.139966919282)*x[1]+(0.823631645543)*x[2]**o+(0.29259439301)*x[2]
        arg[3,2,0,0]=(0.969349250234)*x[0]**o+(0.158604629065)*x[0]+(0.517280925739)*x[1]**o+(-0.32151445591)*x[1]+(0.516829269337)*x[2]**o+(0.935139450142)*x[2]
        arg[3,2,0,1]=(-0.673721580725)*x[0]**o+(0.1932869215)*x[0]+(0.47375679317)*x[1]**o+(0.377941367578)*x[1]+(-0.359306451576)*x[2]**o+(0.00498822119998)*x[2]
        arg[3,2,1,0]=(-0.432873699676)*x[0]**o+(0.191782167044)*x[0]+(0.191395777927)*x[1]**o+(-0.540088847941)*x[1]+(0.608451872364)*x[2]**o+(0.151589512705)*x[2]
        arg[3,2,1,1]=(0.633726614581)*x[0]**o+(-0.824133266602)*x[0]+(-0.380289061608)*x[1]**o+(-0.722136850408)*x[1]+(-0.807853184729)*x[2]**o+(0.152552828168)*x[2]
        arg[3,2,2,0]=(-0.182700878265)*x[0]**o+(0.105776705469)*x[0]+(0.837527799957)*x[1]**o+(0.120438395343)*x[1]+(-0.211142277471)*x[2]**o+(-0.0499206664041)*x[2]
        arg[3,2,2,1]=(0.999790796581)*x[0]**o+(0.523238189614)*x[0]+(0.108118557801)*x[1]**o+(-0.899694790145)*x[1]+(0.165089392648)*x[2]**o+(0.38599870475)*x[2]
        arg[3,3,0,0]=(-0.959143897162)*x[0]**o+(-0.968197835338)*x[0]+(-0.521866033091)*x[1]**o+(-0.560799919169)*x[1]+(0.604562365858)*x[2]**o+(0.195593261393)*x[2]
        arg[3,3,0,1]=(0.385513188376)*x[0]**o+(0.918485762628)*x[0]+(0.581126961809)*x[1]**o+(-0.365882492483)*x[1]+(-0.260563773993)*x[2]**o+(0.997016165521)*x[2]
        arg[3,3,1,0]=(0.809269330969)*x[0]**o+(0.472424328796)*x[0]+(-0.920680159926)*x[1]**o+(0.756205962059)*x[1]+(-0.886275742364)*x[2]**o+(-0.800132407145)*x[2]
        arg[3,3,1,1]=(-0.983474390608)*x[0]**o+(-0.436192301851)*x[0]+(0.653210945793)*x[1]**o+(-0.465357783291)*x[1]+(-0.799261076035)*x[2]**o+(-0.349990963152)*x[2]
        arg[3,3,2,0]=(0.710820041642)*x[0]**o+(-0.973768204693)*x[0]+(0.0408403957292)*x[1]**o+(0.61546181161)*x[1]+(0.915939389579)*x[2]**o+(0.634659938301)*x[2]
        arg[3,3,2,1]=(0.633364771689)*x[0]**o+(0.887553239682)*x[0]+(-0.64913373177)*x[1]**o+(0.557465560353)*x[1]+(-0.744072153243)*x[2]**o+(-0.301528624747)*x[2]
        arg[3,4,0,0]=(0.44039719913)*x[0]**o+(-0.0339525806656)*x[0]+(0.574166395893)*x[1]**o+(-0.810441051899)*x[1]+(0.254523432152)*x[2]**o+(-0.798403721953)*x[2]
        arg[3,4,0,1]=(0.253149910502)*x[0]**o+(-0.909846068796)*x[0]+(0.855107984234)*x[1]**o+(-0.973196168972)*x[1]+(0.646832394105)*x[2]**o+(0.811892995874)*x[2]
        arg[3,4,1,0]=(0.13027869782)*x[0]**o+(-0.424094152568)*x[0]+(-0.197473699021)*x[1]**o+(0.0795845331856)*x[1]+(-0.511301398362)*x[2]**o+(-0.331330168767)*x[2]
        arg[3,4,1,1]=(0.476554240918)*x[0]**o+(0.18800431479)*x[0]+(0.259031626216)*x[1]**o+(-0.0532533771273)*x[1]+(0.914847196458)*x[2]**o+(0.333851370508)*x[2]
        arg[3,4,2,0]=(0.800058524963)*x[0]**o+(-0.402028928896)*x[0]+(0.913881556051)*x[1]**o+(0.895516442873)*x[1]+(-0.0227700921994)*x[2]**o+(-0.0502319834692)*x[2]
        arg[3,4,2,1]=(-0.533834741079)*x[0]**o+(0.557954226232)*x[0]+(-0.121676666641)*x[1]**o+(0.781782385789)*x[1]+(0.0682870158781)*x[2]**o+(-0.236432552686)*x[2]
        ref[0,0,0,0]=(1.46703352572)/(o+1.)+(-1.33520045704)+(0.0737627367843)*0.5**o
        ref[0,0,0,1]=(0.213535197837)/(o+1.)+(-0.325019128805)+(0.353968533247)*0.5**o
        ref[0,0,1,0]=(-0.233213543591)/(o+1.)+(0.450075189439)+(0.315704665816)*0.5**o
        ref[0,0,1,1]=(0.809431837761)/(o+1.)+(0.642450689879)+(-0.453912049362)*0.5**o
        ref[0,0,2,0]=(1.8822564746)/(o+1.)+(-0.160827113662)+(0.705225432895)*0.5**o
        ref[0,0,2,1]=(-0.851077895823)/(o+1.)+(-0.032586006392)+(-0.397162258761)*0.5**o
        ref[0,1,0,0]=(-1.75021683351)/(o+1.)+(0.282579703111)+(0.00444300957662)*0.5**o
        ref[0,1,0,1]=(-1.61043513823)/(o+1.)+(0.209579442759)+(0.726535641072)*0.5**o
        ref[0,1,1,0]=(0.550951959927)/(o+1.)+(-0.453625179799)+(-0.554579794515)*0.5**o
        ref[0,1,1,1]=(-0.809122195368)/(o+1.)+(0.609569171964)+(0.294744137189)*0.5**o
        ref[0,1,2,0]=(1.3771144222)/(o+1.)+(-0.703069822248)+(-0.30261020867)*0.5**o
        ref[0,1,2,1]=(1.08629781546)/(o+1.)+(0.0271166180983)+(0.494855964972)*0.5**o
        ref[0,2,0,0]=(-1.26922850072)/(o+1.)+(0.72637380105)+(-0.914585215232)*0.5**o
        ref[0,2,0,1]=(-0.556637167762)/(o+1.)+(-0.036494329576)+(-0.260590397572)*0.5**o
        ref[0,2,1,0]=(0.707349561895)/(o+1.)+(0.208292082952)+(0.0455797513404)*0.5**o
        ref[0,2,1,1]=(-1.47922090453)/(o+1.)+(-0.621962103251)+(-0.300900136798)*0.5**o
        ref[0,2,2,0]=(0.01827281852)/(o+1.)+(-0.159177319679)+(0.745711773093)*0.5**o
        ref[0,2,2,1]=(1.09869785336)/(o+1.)+(0.352958766464)+(0.291340077334)*0.5**o
        ref[0,3,0,0]=(-0.802486910351)/(o+1.)+(0.741067349535)+(0.89242099254)*0.5**o
        ref[0,3,0,1]=(0.187447770797)/(o+1.)+(0.0211789816838)+(0.095390666498)*0.5**o
        ref[0,3,1,0]=(1.0518488657)/(o+1.)+(-0.0260861096315)+(-0.317750910722)*0.5**o
        ref[0,3,1,1]=(1.16078906983)/(o+1.)+(0.0263064747213)+(-0.204455448392)*0.5**o
        ref[0,3,2,0]=(-0.0986989221708)/(o+1.)+(0.182363177629)+(0.38581656579)*0.5**o
        ref[0,3,2,1]=(-0.388150366913)/(o+1.)+(0.557428570571)+(-0.225571343976)*0.5**o
        ref[0,4,0,0]=(-0.623448360222)/(o+1.)+(0.64524557416)+(0.834083057621)*0.5**o
        ref[0,4,0,1]=(-0.639580493935)/(o+1.)+(-0.380689110373)+(-0.613152578007)*0.5**o
        ref[0,4,1,0]=(-0.575892151257)/(o+1.)+(-0.142294010946)+(-0.373668614153)*0.5**o
        ref[0,4,1,1]=(-0.0316225271882)/(o+1.)+(-0.115844822465)+(0.00943802199128)*0.5**o
        ref[0,4,2,0]=(-0.902549384411)/(o+1.)+(0.210416030442)+(0.157099024354)*0.5**o
        ref[0,4,2,1]=(1.2254583864)/(o+1.)+(-0.371769146738)+(0.933762543076)*0.5**o
        ref[1,0,0,0]=(0.0440432191149)/(o+1.)+(-0.303248360069)+(0.0417476479889)*0.5**o
        ref[1,0,0,1]=(-0.196875527758)/(o+1.)+(-0.39869023332)+(0.450986977446)*0.5**o
        ref[1,0,1,0]=(-0.742874955412)/(o+1.)+(-0.677197937069)+(-0.23712795738)*0.5**o
        ref[1,0,1,1]=(0.385960155907)/(o+1.)+(0.0647632575546)+(-0.308721170682)*0.5**o
        ref[1,0,2,0]=(1.11246238908)/(o+1.)+(0.0111904605859)+(-0.267232319593)*0.5**o
        ref[1,0,2,1]=(0.0798238708014)/(o+1.)+(-0.355425426066)+(0.405254374432)*0.5**o
        ref[1,1,0,0]=(0.133162474633)/(o+1.)+(-0.863343187001)+(-0.877224972777)*0.5**o
        ref[1,1,0,1]=(1.23844759518)/(o+1.)+(0.871424091405)+(-0.873928172379)*0.5**o
        ref[1,1,1,0]=(-0.0275104699246)/(o+1.)+(-0.739054056468)+(0.552826305913)*0.5**o
        ref[1,1,1,1]=(0.948010887116)/(o+1.)+(-0.0503040657995)+(-0.71233607947)*0.5**o
        ref[1,1,2,0]=(1.01678266341)/(o+1.)+(-0.425240837237)+(-0.411667274338)*0.5**o
        ref[1,1,2,1]=(1.62045693811)/(o+1.)+(0.23057679995)+(0.533005373575)*0.5**o
        ref[1,2,0,0]=(0.44625545933)/(o+1.)+(0.0233472080645)+(-0.637366803484)*0.5**o
        ref[1,2,0,1]=(0.277223414052)/(o+1.)+(0.189254359042)+(-0.404870452835)*0.5**o
        ref[1,2,1,0]=(0.880141841398)/(o+1.)+(-0.750127627837)+(0.85574176569)*0.5**o
        ref[1,2,1,1]=(0.0718593507895)/(o+1.)+(0.306599537674)+(0.576825533647)*0.5**o
        ref[1,2,2,0]=(0.35649607288)/(o+1.)+(0.242884247745)+(0.741411326194)*0.5**o
        ref[1,2,2,1]=(1.14623459997)/(o+1.)+(-0.897591123484)+(0.386515069261)*0.5**o
        ref[1,3,0,0]=(-0.796578230199)/(o+1.)+(0.200648623687)+(0.848952965744)*0.5**o
        ref[1,3,0,1]=(0.679746997945)/(o+1.)+(-0.0297741837061)+(0.357870838328)*0.5**o
        ref[1,3,1,0]=(0.0713725201916)/(o+1.)+(0.223521474626)+(-0.509333127149)*0.5**o
        ref[1,3,1,1]=(-1.07672806067)/(o+1.)+(0.402304847309)+(-0.837885550386)*0.5**o
        ref[1,3,2,0]=(0.808945160636)/(o+1.)+(0.525847284807)+(0.205889276236)*0.5**o
        ref[1,3,2,1]=(0.440475889238)/(o+1.)+(0.353047235363)+(-0.331660060588)*0.5**o
        ref[1,4,0,0]=(0.733653130588)/(o+1.)+(-0.113784463751)+(0.394201404247)*0.5**o
        ref[1,4,0,1]=(0.892965597874)/(o+1.)+(0.359236861297)+(-0.346144268453)*0.5**o
        ref[1,4,1,0]=(0.512178276429)/(o+1.)+(0.0722586978563)+(-0.907566140749)*0.5**o
        ref[1,4,1,1]=(-1.21754003377)/(o+1.)+(0.369986403191)+(0.726194147182)*0.5**o
        ref[1,4,2,0]=(0.353974376932)/(o+1.)+(0.291102967642)+(-0.940824546553)*0.5**o
        ref[1,4,2,1]=(0.166197160477)/(o+1.)+(-0.358856317187)+(0.728499572603)*0.5**o
        ref[2,0,0,0]=(1.55742864501)/(o+1.)+(0.38703652972)+(0.467025642879)*0.5**o
        ref[2,0,0,1]=(-0.867288120579)/(o+1.)+(-0.995570381672)+(-0.211129222294)*0.5**o
        ref[2,0,1,0]=(0.0896432549354)/(o+1.)+(-0.894461032367)+(-0.668065295797)*0.5**o
        ref[2,0,1,1]=(-0.240985203598)/(o+1.)+(-0.407009494534)+(0.928751918298)*0.5**o
        ref[2,0,2,0]=(0.815866062908)/(o+1.)+(-0.281559110938)+(-0.531818105049)*0.5**o
        ref[2,0,2,1]=(0.0143868099026)/(o+1.)+(0.253341199154)+(0.0150988200159)*0.5**o
        ref[2,1,0,0]=(0.745834270286)/(o+1.)+(-0.630128050767)+(-0.0829154151121)*0.5**o
        ref[2,1,0,1]=(-0.513757920038)/(o+1.)+(-0.505622196077)+(-0.834502772339)*0.5**o
        ref[2,1,1,0]=(1.38226406947)/(o+1.)+(-0.763221422703)+(0.247439924718)*0.5**o
        ref[2,1,1,1]=(-0.408281199784)/(o+1.)+(-0.811328833758)+(-0.120169534728)*0.5**o
        ref[2,1,2,0]=(-0.586214858819)/(o+1.)+(0.11721515075)+(0.703841647274)*0.5**o
        ref[2,1,2,1]=(0.0812921429368)/(o+1.)+(-0.357151740094)+(-0.19760048131)*0.5**o
        ref[2,2,0,0]=(1.65849097715)/(o+1.)+(-0.777022017999)+(-0.507872937196)*0.5**o
        ref[2,2,0,1]=(0.242698257917)/(o+1.)+(1.08992054229)+(-0.481403986338)*0.5**o
        ref[2,2,1,0]=(-0.754639453272)/(o+1.)+(0.81286785715)+(0.241675645258)*0.5**o
        ref[2,2,1,1]=(-0.438727562295)/(o+1.)+(0.217076971211)+(0.114772627763)*0.5**o
        ref[2,2,2,0]=(0.683917726753)/(o+1.)+(-0.781863136791)+(0.887023209688)*0.5**o
        ref[2,2,2,1]=(1.19838469898)/(o+1.)+(0.729462151403)+(0.836731324893)*0.5**o
        ref[2,3,0,0]=(1.28206834695)/(o+1.)+(0.583576731084)+(0.196161020545)*0.5**o
        ref[2,3,0,1]=(0.395309853285)/(o+1.)+(0.573307319258)+(-0.862326385376)*0.5**o
        ref[2,3,1,0]=(1.34760990633)/(o+1.)+(0.348652101929)+(0.571263839862)*0.5**o
        ref[2,3,1,1]=(0.293682752388)/(o+1.)+(0.080258501461)+(0.272559456412)*0.5**o
        ref[2,3,2,0]=(-1.47056068107)/(o+1.)+(-0.237909711446)+(0.3409363126)*0.5**o
        ref[2,3,2,1]=(-0.356052894952)/(o+1.)+(-0.166196236302)+(0.949170673623)*0.5**o
        ref[2,4,0,0]=(-0.21887693849)/(o+1.)+(0.365445238731)+(0.876774885889)*0.5**o
        ref[2,4,0,1]=(0.824663735425)/(o+1.)+(-0.568245545547)+(0.347441329383)*0.5**o
        ref[2,4,1,0]=(-1.22807417581)/(o+1.)+(-0.0380844812022)+(0.732123350436)*0.5**o
        ref[2,4,1,1]=(0.45997712447)/(o+1.)+(0.181215330541)+(-0.128872747083)*0.5**o
        ref[2,4,2,0]=(0.282057449847)/(o+1.)+(0.49247247384)+(0.619034818105)*0.5**o
        ref[2,4,2,1]=(0.0852215499679)/(o+1.)+(0.0807247245641)+(0.645979236759)*0.5**o
        ref[3,0,0,0]=(-0.173183992256)/(o+1.)+(0.215110333109)+(-0.265836015694)*0.5**o
        ref[3,0,0,1]=(0.88174458564)/(o+1.)+(0.282289419133)+(0.306246512933)*0.5**o
        ref[3,0,1,0]=(-0.542918101609)/(o+1.)+(0.791816447658)+(0.509420593773)*0.5**o
        ref[3,0,1,1]=(0.585362427423)/(o+1.)+(-0.229639545331)+(-0.609626065482)*0.5**o
        ref[3,0,2,0]=(-0.625654730672)/(o+1.)+(0.690821759706)+(0.269212290711)*0.5**o
        ref[3,0,2,1]=(0.971993211744)/(o+1.)+(-1.18192725526)+(-0.445116541697)*0.5**o
        ref[3,1,0,0]=(-1.04249026287)/(o+1.)+(0.0447480055334)+(0.25585673506)*0.5**o
        ref[3,1,0,1]=(-1.30475672494)/(o+1.)+(0.700582633156)+(-0.159301665223)*0.5**o
        ref[3,1,1,0]=(0.25661740363)/(o+1.)+(-0.036124371669)+(0.673375237931)*0.5**o
        ref[3,1,1,1]=(1.24932735533)/(o+1.)+(-0.597633535843)+(0.273992606955)*0.5**o
        ref[3,1,2,0]=(0.116260708817)/(o+1.)+(0.813874160367)+(-0.533129977534)*0.5**o
        ref[3,1,2,1]=(1.30199813466)/(o+1.)+(0.377333042401)+(0.977824314677)*0.5**o
        ref[3,2,0,0]=(1.03411019508)/(o+1.)+(0.386114811649)+(0.969349250234)*0.5**o
        ref[3,2,0,1]=(0.114450341594)/(o+1.)+(0.288108255139)+(-0.673721580725)*0.5**o
        ref[3,2,1,0]=(0.799847650292)/(o+1.)+(-0.0983585840958)+(-0.432873699676)*0.5**o
        ref[3,2,1,1]=(-1.18814224634)/(o+1.)+(-0.696858644421)+(0.633726614581)*0.5**o
        ref[3,2,2,0]=(0.626385522485)/(o+1.)+(0.0881472172038)+(-0.182700878265)*0.5**o
        ref[3,2,2,1]=(0.27320795045)/(o+1.)+(0.0047710521094)+(0.999790796581)*0.5**o
        ref[3,3,0,0]=(0.0826963327674)/(o+1.)+(-0.666702246557)+(-0.959143897162)*0.5**o
        ref[3,3,0,1]=(0.320563187816)/(o+1.)+(0.774809717833)+(0.385513188376)*0.5**o
        ref[3,3,1,0]=(-1.80695590229)/(o+1.)+(0.214248941855)+(0.809269330969)*0.5**o
        ref[3,3,1,1]=(-0.146050130242)/(o+1.)+(-0.625770524147)+(-0.983474390608)*0.5**o
        ref[3,3,2,0]=(0.956779785308)/(o+1.)+(0.138176772609)+(0.710820041642)*0.5**o
        ref[3,3,2,1]=(-1.39320588501)/(o+1.)+(0.571745087644)+(0.633364771689)*0.5**o
        ref[3,4,0,0]=(0.828689828045)/(o+1.)+(-0.821398677259)+(0.44039719913)*0.5**o
        ref[3,4,0,1]=(1.50194037834)/(o+1.)+(-0.535574620947)+(0.253149910502)*0.5**o
        ref[3,4,1,0]=(-0.708775097383)/(o+1.)+(-0.337919894074)+(0.13027869782)*0.5**o
        ref[3,4,1,1]=(1.17387882267)/(o+1.)+(0.234301154085)+(0.476554240918)*0.5**o
        ref[3,4,2,0]=(0.891111463851)/(o+1.)+(0.221627765254)+(0.800058524963)*0.5**o
        ref[3,4,2,1]=(-0.0533896507631)/(o+1.)+(0.551652029667)+(-0.533834741079)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.570707549273)*x[0]+(-0.726264032186)*x[1]
        ref=(-0.0423765208873)/(o+1.)+(-0.791542656673)+(0.328490252775)*0.5**o
      else:
        arg=(1.28222892931)*x[0]+(-0.798170675576)*x[1]+(1.5701208405)*x[2]
        ref=(0.721975372868)/(o+1.)+(0.344720905832)+(0.642761909705)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.575166906798)*x[0]+(-1.63698773418)*x[1]
        arg[1]=(-0.0535946584636)*x[0]+(0.105969101775)*x[1]
        ref[0]=(-0.751926426344)/(o+1.)+(-0.395800138888)+(0.481705876734)*0.5**o
        ref[1]=(0.698496252242)/(o+1.)+(-0.265927161866)+(-0.114267485199)*0.5**o
      else:
        arg[0]=(-0.17879254138)*x[0]+(-1.12752123792)*x[1]+(0.921297311605)*x[2]
        arg[1]=(1.33131824618)*x[0]+(-0.392175059176)*x[1]+(0.545754281465)*x[2]
        ref[0]=(-0.494870303061)/(o+1.)+(0.0697636237654)+(-0.0296734121631)*0.5**o
        ref[1]=(-0.394300983704)/(o+1.)+(0.722261312754)+(0.434675826665)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.338658191734)*x[0]+(1.46127609683)*x[1]
        arg[0,1]=(-0.629116867491)*x[0]+(1.40141569066)*x[1]
        arg[0,2]=(0.105569779347)*x[0]+(0.239936324967)*x[1]
        arg[0,3]=(-0.281718331718)*x[0]+(0.280323404866)*x[1]
        arg[0,4]=(0.467341113661)*x[0]+(-0.50452062075)*x[1]
        arg[1,0]=(-0.841295219186)*x[0]+(0.524038163723)*x[1]
        arg[1,1]=(1.40143704236)*x[0]+(-1.50467961637)*x[1]
        arg[1,2]=(1.21078790505)*x[0]+(-0.731453470292)*x[1]
        arg[1,3]=(0.469556922076)*x[0]+(0.0686293204391)*x[1]
        arg[1,4]=(-1.26394637585)*x[0]+(-0.0361395276585)*x[1]
        arg[2,0]=(-1.12850883148)*x[0]+(0.284296673326)*x[1]
        arg[2,1]=(-0.626362828813)*x[0]+(1.19459827573)*x[1]
        arg[2,2]=(-0.545457293083)*x[0]+(1.547337019)*x[1]
        arg[2,3]=(0.664461986629)*x[0]+(-1.69082198266)*x[1]
        arg[2,4]=(-0.100004763305)*x[0]+(0.152159044423)*x[1]
        arg[3,0]=(0.307574843061)*x[0]+(1.20209181301)*x[1]
        arg[3,1]=(-0.0663203356627)*x[0]+(-1.26669863527)*x[1]
        arg[3,2]=(-1.20436798894)*x[0]+(-0.0375693379661)*x[1]
        arg[3,3]=(0.753649087747)*x[0]+(0.69376223486)*x[1]
        arg[3,4]=(-0.0512732264834)*x[0]+(-0.658097606436)*x[1]
        ref[0,0]=(0.996821439952)/(o+1.)+(0.015088160779)+(0.0956201435879)*0.5**o
        ref[0,1]=(0.643092963726)/(o+1.)+(0.239390724844)+(-0.349575590244)*0.5**o
        ref[0,2]=(-0.198468008977)/(o+1.)+(0.00411201743342)+(0.535750078425)*0.5**o
        ref[0,3]=(0.642491201137)/(o+1.)+(-0.623413199652)+(0.602940271315)*0.5**o
        ref[0,4]=(-0.105706120549)/(o+1.)+(-0.353560054939)+(0.775646723338)*0.5**o
        ref[1,0]=(0.144540669221)/(o+1.)+(0.0228268215803)+(-0.507451367845)*0.5**o
        ref[1,1]=(-0.820379339897)/(o+1.)+(0.0951825678509)+(0.526771630185)*0.5**o
        ref[1,2]=(-0.21214585824)/(o+1.)+(-0.0889183062985)+(0.869316905591)*0.5**o
        ref[1,3]=(0.291228940674)/(o+1.)+(0.192623928703)+(-0.138290555565)*0.5**o
        ref[1,4]=(-0.0663718956434)/(o+1.)+(-0.309484335614)+(-0.614745336642)*0.5**o
        ref[2,0]=(0.484892962201)/(o+1.)+(-0.367129859793)+(-0.594845400763)*0.5**o
        ref[2,1]=(0.768457634672)/(o+1.)+(0.154447657562)+(-0.509117502876)*0.5**o
        ref[2,2]=(0.707320338363)/(o+1.)+(-0.0592149232183)+(0.412989233992)*0.5**o
        ref[2,3]=(-0.983684339171)/(o+1.)+(-0.495377645598)+(0.948079634337)*0.5**o
        ref[2,4]=(-0.0315947220465)/(o+1.)+(0.506833986759)+(-0.929918970353)*0.5**o
        ref[3,0]=(0.54141067098)/(o+1.)+(0.489049131493)+(-0.00984227789859)*0.5**o
        ref[3,1]=(-0.552335053051)/(o+1.)+(-0.383325638803)+(-0.0140326402734)*0.5**o
        ref[3,2]=(0.0821472412575)/(o+1.)+(-0.335827719718)+(-0.652429128733)*0.5**o
        ref[3,3]=(0.112917208924)/(o+1.)+(0.667772807338)+(-0.00105150099312)*0.5**o
        ref[3,4]=(-0.19331508157)/(o+1.)+(-0.102296385708)+(-0.311462979932)*0.5**o
      else:
        arg[0,0]=(0.069245963142)*x[0]+(-0.185727379494)*x[1]+(-0.887410889699)*x[2]
        arg[0,1]=(-0.816745668949)*x[0]+(-0.250435680662)*x[1]+(0.701607830582)*x[2]
        arg[0,2]=(-0.862433954308)*x[0]+(0.731874372527)*x[1]+(-0.115384982779)*x[2]
        arg[0,3]=(-0.265247631604)*x[0]+(-0.157059741958)*x[1]+(-0.368419335078)*x[2]
        arg[0,4]=(1.66720203567)*x[0]+(-0.759393708719)*x[1]+(1.36360850704)*x[2]
        arg[1,0]=(-0.192822813795)*x[0]+(0.0358776152527)*x[1]+(-0.722864520788)*x[2]
        arg[1,1]=(-0.480606215471)*x[0]+(1.13682774573)*x[1]+(0.17900467623)*x[2]
        arg[1,2]=(0.346760852584)*x[0]+(0.994833030281)*x[1]+(0.507055933913)*x[2]
        arg[1,3]=(-0.509961853478)*x[0]+(-0.852997575505)*x[1]+(0.586527030253)*x[2]
        arg[1,4]=(0.93957975846)*x[0]+(0.28878335152)*x[1]+(0.128070931738)*x[2]
        arg[2,0]=(0.338283306313)*x[0]+(0.0244500761608)*x[1]+(-0.606875554888)*x[2]
        arg[2,1]=(0.850782624552)*x[0]+(-0.836382038189)*x[1]+(-1.41808459947)*x[2]
        arg[2,2]=(-1.06176229613)*x[0]+(-0.259247046061)*x[1]+(-0.583455484089)*x[2]
        arg[2,3]=(1.17437923282)*x[0]+(1.31485601495)*x[1]+(1.3139743231)*x[2]
        arg[2,4]=(0.362851636275)*x[0]+(0.804049783156)*x[1]+(1.09205000214)*x[2]
        arg[3,0]=(0.343603837195)*x[0]+(-1.49082257981)*x[1]+(-0.628668674545)*x[2]
        arg[3,1]=(-0.676847711765)*x[0]+(1.32350560453)*x[1]+(1.07275185686)*x[2]
        arg[3,2]=(0.570784985469)*x[0]+(-1.01212360602)*x[1]+(-0.0808776667606)*x[2]
        arg[3,3]=(0.506392444236)*x[0]+(-1.00848318041)*x[1]+(0.558777901541)*x[2]
        arg[3,4]=(-0.60925738291)*x[0]+(-0.386755957809)*x[1]+(0.60930115434)*x[2]
        ref[0,0]=(-0.695230136419)/(o+1.)+(-0.388400729083)+(0.468139288533)*0.5**o
        ref[0,1]=(0.462098132187)/(o+1.)+(-0.215041893287)+(-0.397587864642)*0.5**o
        ref[0,2]=(-0.262411385363)/(o+1.)+(0.275729226874)+(-0.534991632946)*0.5**o
        ref[0,3]=(0.117822633378)/(o+1.)+(0.0286827897592)+(-0.965914921536)*0.5**o
        ref[0,4]=(0.293575681039)/(o+1.)+(0.584039692827)+(0.809761767299)*0.5**o
        ref[1,0]=(0.627951922463)/(o+1.)+(-0.994889162557)+(0.482016683321)*0.5**o
        ref[1,1]=(1.83482238662)/(o+1.)+(-0.205719558739)+(-0.588157062647)*0.5**o
        ref[1,2]=(1.13612020398)/(o+1.)+(0.0262824938134)+(0.659964625174)*0.5**o
        ref[1,3]=(0.325264487741)/(o+1.)+(-0.0752154813413)+(-0.951265923787)*0.5**o
        ref[1,4]=(-0.149919283302)/(o+1.)+(0.77701502585)+(-0.0476767266789)*0.5**o
        ref[2,0]=(0.491504686171)/(o+1.)+(-0.562534706926)+(0.389422555265)*0.5**o
        ref[2,1]=(-1.52704217417)/(o+1.)+(0.120525042098)+(-0.117691923134)*0.5**o
        ref[2,2]=(-0.779796772543)/(o+1.)+(-0.318864997267)+(-0.486938059201)*0.5**o
        ref[2,3]=(1.53343318682)/(o+1.)+(0.731149467308)+(0.807477449443)*0.5**o
        ref[2,4]=(1.15919207694)/(o+1.)+(0.113921448815)+(0.871916446996)*0.5**o
        ref[3,0]=(-1.75312920944)/(o+1.)+(-0.153798982308)+(0.284839756895)*0.5**o
        ref[3,1]=(1.30985814569)/(o+1.)+(0.438430874237)+(-0.467310144534)*0.5**o
        ref[3,2]=(-0.737480320972)/(o+1.)+(0.0604403094397)+(0.094383414777)*0.5**o
        ref[3,3]=(-0.688723266524)/(o+1.)+(0.43486027018)+(-0.124310108473)*0.5**o
        ref[3,4]=(0.128399262674)/(o+1.)+(-0.372445844122)+(0.229780239192)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.993339371954)*x[0]+(-1.35308128985)*x[1]
        arg[0,0,1]=(-1.07886149134)*x[0]+(-1.38999283206)*x[1]
        arg[0,1,0]=(0.246468026972)*x[0]+(-1.37217464882)*x[1]
        arg[0,1,1]=(0.645672648599)*x[0]+(1.8070624103)*x[1]
        arg[1,0,0]=(-1.02278989388)*x[0]+(0.151634023103)*x[1]
        arg[1,0,1]=(-0.667332860541)*x[0]+(0.312742149193)*x[1]
        arg[1,1,0]=(-0.848647385832)*x[0]+(0.466438038516)*x[1]
        arg[1,1,1]=(-0.185495911372)*x[0]+(-0.137118786891)*x[1]
        arg[2,0,0]=(-0.0250555870731)*x[0]+(0.905553573283)*x[1]
        arg[2,0,1]=(-1.03940852939)*x[0]+(-1.57326016394)*x[1]
        arg[2,1,0]=(1.48834718811)*x[0]+(-0.340889253246)*x[1]
        arg[2,1,1]=(0.875721514739)*x[0]+(-0.310857412212)*x[1]
        arg[3,0,0]=(1.25129508831)*x[0]+(-1.69312043525)*x[1]
        arg[3,0,1]=(1.50124675312)*x[0]+(0.244195748654)*x[1]
        arg[3,1,0]=(0.292060478909)*x[0]+(-0.660005422696)*x[1]
        arg[3,1,1]=(0.791335231308)*x[0]+(0.466481573591)*x[1]
        arg[4,0,0]=(-0.361241947236)*x[0]+(-1.43591648587)*x[1]
        arg[4,0,1]=(-1.10385456695)*x[0]+(-0.325314195825)*x[1]
        arg[4,1,0]=(1.01004173033)*x[0]+(0.426824552857)*x[1]
        arg[4,1,1]=(-0.631382479612)*x[0]+(0.0603792190966)*x[1]
        arg[5,0,0]=(-0.30404368786)*x[0]+(-0.0102498522932)*x[1]
        arg[5,0,1]=(0.109188848746)*x[0]+(-1.16846160105)*x[1]
        arg[5,1,0]=(-0.351916879547)*x[0]+(-0.320714795448)*x[1]
        arg[5,1,1]=(-0.534138562652)*x[0]+(-0.256013589844)*x[1]
        ref[0,0,0]=(-0.506159375456)/(o+1.)+(-0.522895228783)+(-0.794470828783)*0.5**o
        ref[0,0,1]=(-0.930984367068)/(o+1.)+(-0.636851258685)+(-0.264167438968)*0.5**o
        ref[0,1,0]=(-0.674222792172)/(o+1.)+(-0.52131379048)+(0.591143751282)*0.5**o
        ref[0,1,1]=(0.953370751283)/(o+1.)+(0.377773224645)+(0.743817858326)*0.5**o
        ref[1,0,0]=(0.700508215428)/(o+1.)+(-0.303343530867)+(-0.964977024471)*0.5**o
        ref[1,0,1]=(-0.408185116452)/(o+1.)+(0.0668983569743)+(-0.0802023088445)*0.5**o
        ref[1,1,0]=(0.80742345527)/(o+1.)+(-0.405780705683)+(-0.37807139122)*0.5**o
        ref[1,1,1]=(0.0184615568095)/(o+1.)+(0.221921562266)+(-0.784919379605)*0.5**o
        ref[2,0,0]=(0.571906256883)/(o+1.)+(-0.0415588382133)+(0.391709405753)*0.5**o
        ref[2,0,1]=(-0.778176009451)/(o+1.)+(-0.761680990566)+(-0.311130702752)*0.5**o
        ref[2,1,0]=(0.498214686321)/(o+1.)+(-0.146173705483)+(0.941590659509)*0.5**o
        ref[2,1,1]=(-0.273747071818)/(o+1.)+(-0.0319184747153)+(0.902448123776)*0.5**o
        ref[3,0,0]=(-0.730029471178)/(o+1.)+(-0.288672747933)+(0.865549620109)*0.5**o
        ref[3,0,1]=(-0.371409186475)/(o+1.)+(0.623169375401)+(0.870512937445)*0.5**o
        ref[3,1,0]=(0.0611454391867)/(o+1.)+(-0.627723874214)+(0.826357365454)*0.5**o
        ref[3,1,1]=(0.785553447935)/(o+1.)+(0.246940213405)+(-0.021617069845)*0.5**o
        ref[4,0,0]=(-0.936201955907)/(o+1.)+(-0.484823720021)+(0.108690962842)*0.5**o
        ref[4,0,1]=(-0.277830390815)/(o+1.)+(-0.151786547071)+(-0.847765277819)*0.5**o
        ref[4,1,0]=(0.12293099533)/(o+1.)+(0.295620382054)+(0.722694523749)*0.5**o
        ref[4,1,1]=(-0.5517019185)/(o+1.)+(0.0850748937088)+(-0.189451129433)*0.5**o
        ref[5,0,0]=(-0.942533148805)/(o+1.)+(0.425680504872)+(-0.223121401092)*0.5**o
        ref[5,0,1]=(-0.377134228332)/(o+1.)+(-0.721019409569)+(0.759900295162)*0.5**o
        ref[5,1,0]=(-0.643828391942)/(o+1.)+(0.399974456311)+(-0.828752195674)*0.5**o
        ref[5,1,1]=(-0.118293731439)/(o+1.)+(-0.355568544431)+(0.0392786678054)*0.5**o
      else:
        arg[0,0,0]=(0.205446621454)*x[0]+(0.10254878667)*x[1]+(1.27673728979)*x[2]
        arg[0,0,1]=(0.758846313193)*x[0]+(0.63320336713)*x[1]+(-0.42638505807)*x[2]
        arg[0,1,0]=(-0.0127116561603)*x[0]+(-0.360359420027)*x[1]+(0.540508591493)*x[2]
        arg[0,1,1]=(-0.278918629538)*x[0]+(-0.0291534835772)*x[1]+(0.708655731565)*x[2]
        arg[1,0,0]=(-0.612965210483)*x[0]+(0.787787556912)*x[1]+(0.222384448462)*x[2]
        arg[1,0,1]=(1.07409083742)*x[0]+(1.19429173706)*x[1]+(-0.484979976779)*x[2]
        arg[1,1,0]=(-1.25714999157)*x[0]+(-0.321346019622)*x[1]+(0.278329320837)*x[2]
        arg[1,1,1]=(-0.512262689707)*x[0]+(1.6891992053)*x[1]+(-1.02351830068)*x[2]
        arg[2,0,0]=(-1.57911907773)*x[0]+(0.096124258122)*x[1]+(0.336619253521)*x[2]
        arg[2,0,1]=(0.0901144254199)*x[0]+(-1.36818186255)*x[1]+(1.2969609239)*x[2]
        arg[2,1,0]=(0.158329511499)*x[0]+(0.696997185217)*x[1]+(-0.400535235927)*x[2]
        arg[2,1,1]=(-1.44570853201)*x[0]+(0.237359587195)*x[1]+(-1.17323024801)*x[2]
        arg[3,0,0]=(-1.4325812824)*x[0]+(-0.733085265902)*x[1]+(0.125883186484)*x[2]
        arg[3,0,1]=(-0.0529889021435)*x[0]+(0.335154600942)*x[1]+(0.839337375209)*x[2]
        arg[3,1,0]=(-0.423552182963)*x[0]+(1.31813289315)*x[1]+(0.13235854915)*x[2]
        arg[3,1,1]=(0.53278572217)*x[0]+(0.159434073071)*x[1]+(-0.313621199136)*x[2]
        arg[4,0,0]=(0.747779887673)*x[0]+(0.30977483911)*x[1]+(-0.0760295896739)*x[2]
        arg[4,0,1]=(0.00295118909194)*x[0]+(-0.687914867842)*x[1]+(0.0386268422915)*x[2]
        arg[4,1,0]=(-0.276363532754)*x[0]+(0.52102466785)*x[1]+(-0.684840191917)*x[2]
        arg[4,1,1]=(0.0283138898538)*x[0]+(0.116042458294)*x[1]+(0.485578401528)*x[2]
        arg[5,0,0]=(-0.118402060192)*x[0]+(-1.18409809944)*x[1]+(-0.304048852342)*x[2]
        arg[5,0,1]=(0.174393790186)*x[0]+(-0.524912689263)*x[1]+(-0.648708392355)*x[2]
        arg[5,1,0]=(-0.254958756344)*x[0]+(-0.772071348728)*x[1]+(-0.0572494064027)*x[2]
        arg[5,1,1]=(-0.74777072018)*x[0]+(0.19109939364)*x[1]+(0.498973987814)*x[2]
        ref[0,0,0]=(0.986457444152)/(o+1.)+(0.258994989412)+(0.0802852749346)*0.5**o
        ref[0,0,1]=(-0.0286717143415)/(o+1.)+(0.554867138611)+(-0.115397940627)*0.5**o
        ref[0,1,0]=(-0.886482131138)/(o+1.)+(0.464293129749)+(0.125333386946)*0.5**o
        ref[0,1,1]=(0.199214164457)/(o+1.)+(0.336882900218)+(-0.472396346443)*0.5**o
        ref[1,0,0]=(0.36021680954)/(o+1.)+(0.408903924703)+(-0.780817864055)*0.5**o
        ref[1,0,1]=(0.44795581074)/(o+1.)+(0.472159267742)+(0.391128251477)*0.5**o
        ref[1,1,0]=(1.02665734893)/(o+1.)+(-0.71418625728)+(-0.898451524719)*0.5**o
        ref[1,1,1]=(0.0322676607777)/(o+1.)+(0.105378958385)+(-0.0896073626306)*0.5**o
        ref[2,0,0]=(1.15723247637)/(o+1.)+(-0.719604538702)+(-0.86439896505)*0.5**o
        ref[2,0,1]=(-0.254100181576)/(o+1.)+(-0.0387997401585)+(0.35059314866)*0.5**o
        ref[2,1,0]=(-0.636439167982)/(o+1.)+(0.563390681981)+(-0.0355507351918)*0.5**o
        ref[2,1,1]=(-0.265265225142)/(o+1.)+(-0.635620798949)+(-0.845072369786)*0.5**o
        ref[3,0,0]=(-0.062162800164)/(o+1.)+(-0.554499928282)+(-0.868620705087)*0.5**o
        ref[3,0,1]=(0.784681961813)/(o+1.)+(0.355202655131)+(-0.373584198067)*0.5**o
        ref[3,1,0]=(1.20272043429)/(o+1.)+(0.352677694103)+(-0.881136563162)*0.5**o
        ref[3,1,1]=(-0.311152797156)/(o+1.)+(0.313736563953)+(0.0622782653549)*0.5**o
        ref[4,0,0]=(-0.365517620238)/(o+1.)+(0.733502174253)+(-0.119961591158)*0.5**o
        ref[4,0,1]=(0.0155845896619)/(o+1.)+(-0.654963543839)+(0.648005661558)*0.5**o
        ref[4,1,0]=(0.405697975915)/(o+1.)+(-0.637805133954)+(0.429733235172)*0.5**o
        ref[4,1,1]=(-0.0469738326054)/(o+1.)+(0.0590721074184)+(0.558764367445)*0.5**o
        ref[5,0,0]=(-0.333825007954)/(o+1.)+(-0.46448520882)+(-0.343753586377)*0.5**o
        ref[5,0,1]=(-0.469125954795)/(o+1.)+(-0.147192309255)+(-0.235716718127)*0.5**o
        ref[5,1,0]=(-0.573275891777)/(o+1.)+(-0.478328163121)+(0.445652706545)*0.5**o
        ref[5,1,1]=(0.0545515234726)/(o+1.)+(-0.153919681346)+(0.195590500493)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.226142137728)*x[0]+(0.85681895183)*x[1]
        arg[0,0,0,1]=(-0.824407181762)*x[0]+(-0.290023144892)*x[1]
        arg[0,0,1,0]=(-1.25987816017)*x[0]+(0.824340990595)*x[1]
        arg[0,0,1,1]=(-0.0917811606765)*x[0]+(0.660211932964)*x[1]
        arg[0,0,2,0]=(-1.31073149178)*x[0]+(0.482324984918)*x[1]
        arg[0,0,2,1]=(0.256995458227)*x[0]+(0.79244254852)*x[1]
        arg[0,1,0,0]=(-0.0682147475407)*x[0]+(0.243620052061)*x[1]
        arg[0,1,0,1]=(0.071099982687)*x[0]+(-0.62986754897)*x[1]
        arg[0,1,1,0]=(-0.0716793733131)*x[0]+(-1.15235525814)*x[1]
        arg[0,1,1,1]=(0.193107575216)*x[0]+(-0.596149957543)*x[1]
        arg[0,1,2,0]=(-0.160142629886)*x[0]+(1.36581707067)*x[1]
        arg[0,1,2,1]=(-1.21477648376)*x[0]+(-0.732000298369)*x[1]
        arg[0,2,0,0]=(-1.02468971059)*x[0]+(0.330721595217)*x[1]
        arg[0,2,0,1]=(1.11414907391)*x[0]+(1.29386127812)*x[1]
        arg[0,2,1,0]=(1.1485850154)*x[0]+(0.387585977495)*x[1]
        arg[0,2,1,1]=(0.237382583072)*x[0]+(1.56248221279)*x[1]
        arg[0,2,2,0]=(-0.949129049229)*x[0]+(0.578823379063)*x[1]
        arg[0,2,2,1]=(-1.0667182537)*x[0]+(0.462227485959)*x[1]
        arg[0,3,0,0]=(1.37954377554)*x[0]+(0.541856880139)*x[1]
        arg[0,3,0,1]=(0.401584700319)*x[0]+(0.577102327147)*x[1]
        arg[0,3,1,0]=(-0.289796408108)*x[0]+(0.807853785325)*x[1]
        arg[0,3,1,1]=(0.0038467861822)*x[0]+(-0.931311510771)*x[1]
        arg[0,3,2,0]=(0.859036254821)*x[0]+(0.198884218516)*x[1]
        arg[0,3,2,1]=(-0.00310256070968)*x[0]+(0.511838475968)*x[1]
        arg[0,4,0,0]=(0.539686886802)*x[0]+(-0.324612924267)*x[1]
        arg[0,4,0,1]=(-0.0458655898565)*x[0]+(0.211810380966)*x[1]
        arg[0,4,1,0]=(0.302333533003)*x[0]+(0.0663033316831)*x[1]
        arg[0,4,1,1]=(-0.981519161639)*x[0]+(-0.00207756909914)*x[1]
        arg[0,4,2,0]=(0.283659447004)*x[0]+(-0.231526592979)*x[1]
        arg[0,4,2,1]=(-1.17679167875)*x[0]+(0.285698985756)*x[1]
        arg[1,0,0,0]=(-0.162900059582)*x[0]+(-0.877164340783)*x[1]
        arg[1,0,0,1]=(-0.50468605065)*x[0]+(-0.160440073709)*x[1]
        arg[1,0,1,0]=(-0.100901676644)*x[0]+(-0.296111526029)*x[1]
        arg[1,0,1,1]=(0.176107421869)*x[0]+(1.74393708745)*x[1]
        arg[1,0,2,0]=(-0.979807721703)*x[0]+(-0.229400841067)*x[1]
        arg[1,0,2,1]=(0.217472795645)*x[0]+(0.0881503243323)*x[1]
        arg[1,1,0,0]=(-0.127980123625)*x[0]+(0.0932095456541)*x[1]
        arg[1,1,0,1]=(0.829674410931)*x[0]+(-0.796196041601)*x[1]
        arg[1,1,1,0]=(1.28656110233)*x[0]+(-1.27850471392)*x[1]
        arg[1,1,1,1]=(1.159269884)*x[0]+(-0.973135433631)*x[1]
        arg[1,1,2,0]=(1.46274837856)*x[0]+(-1.81554270863)*x[1]
        arg[1,1,2,1]=(0.0973472511688)*x[0]+(0.551451261495)*x[1]
        arg[1,2,0,0]=(0.510667671293)*x[0]+(-0.51075982099)*x[1]
        arg[1,2,0,1]=(-1.41612093451)*x[0]+(0.614417669629)*x[1]
        arg[1,2,1,0]=(1.46595125796)*x[0]+(0.60374279122)*x[1]
        arg[1,2,1,1]=(-0.317185218187)*x[0]+(-0.0785532882696)*x[1]
        arg[1,2,2,0]=(-0.388504041993)*x[0]+(-0.838003398775)*x[1]
        arg[1,2,2,1]=(0.0497207129012)*x[0]+(-0.622155692247)*x[1]
        arg[1,3,0,0]=(-0.0471400179563)*x[0]+(1.3378948207)*x[1]
        arg[1,3,0,1]=(-1.22250822195)*x[0]+(0.629714783138)*x[1]
        arg[1,3,1,0]=(-1.28801080408)*x[0]+(-1.54079710038)*x[1]
        arg[1,3,1,1]=(-0.426412448469)*x[0]+(-1.09410474021)*x[1]
        arg[1,3,2,0]=(0.022834167351)*x[0]+(-0.497423389788)*x[1]
        arg[1,3,2,1]=(-0.787492124301)*x[0]+(-0.245621168419)*x[1]
        arg[1,4,0,0]=(-1.37814880919)*x[0]+(-0.704805848451)*x[1]
        arg[1,4,0,1]=(-1.12494770055)*x[0]+(1.44920024305)*x[1]
        arg[1,4,1,0]=(-0.809532572761)*x[0]+(1.16638587657)*x[1]
        arg[1,4,1,1]=(0.931784486816)*x[0]+(1.34629889453)*x[1]
        arg[1,4,2,0]=(-0.172430942793)*x[0]+(-0.584736998321)*x[1]
        arg[1,4,2,1]=(0.00542948736389)*x[0]+(0.43917597284)*x[1]
        arg[2,0,0,0]=(-0.946122337406)*x[0]+(-0.161940168137)*x[1]
        arg[2,0,0,1]=(-0.7856653779)*x[0]+(-0.75287701788)*x[1]
        arg[2,0,1,0]=(-1.08858367592)*x[0]+(-0.530380818694)*x[1]
        arg[2,0,1,1]=(-1.0573555095)*x[0]+(-0.854148765785)*x[1]
        arg[2,0,2,0]=(0.723412825328)*x[0]+(0.326761950437)*x[1]
        arg[2,0,2,1]=(-0.0143142161225)*x[0]+(-0.752724311876)*x[1]
        arg[2,1,0,0]=(-0.942548092355)*x[0]+(0.700861094897)*x[1]
        arg[2,1,0,1]=(-0.712286567957)*x[0]+(0.230406545738)*x[1]
        arg[2,1,1,0]=(1.27414839774)*x[0]+(0.724784649037)*x[1]
        arg[2,1,1,1]=(0.637494163002)*x[0]+(0.442489599192)*x[1]
        arg[2,1,2,0]=(1.81617631726)*x[0]+(1.56870218809)*x[1]
        arg[2,1,2,1]=(1.16419900666)*x[0]+(-0.044314729141)*x[1]
        arg[2,2,0,0]=(0.880267125196)*x[0]+(0.77124406259)*x[1]
        arg[2,2,0,1]=(0.482402867542)*x[0]+(0.0746924305017)*x[1]
        arg[2,2,1,0]=(0.105861259696)*x[0]+(0.484292207267)*x[1]
        arg[2,2,1,1]=(-0.107785838196)*x[0]+(0.701132243209)*x[1]
        arg[2,2,2,0]=(0.559717405844)*x[0]+(-1.5585942717)*x[1]
        arg[2,2,2,1]=(-0.989130712605)*x[0]+(0.221361137556)*x[1]
        arg[2,3,0,0]=(-0.045471818095)*x[0]+(-0.437941959792)*x[1]
        arg[2,3,0,1]=(1.25840468379)*x[0]+(-1.18872011057)*x[1]
        arg[2,3,1,0]=(1.70665829408)*x[0]+(0.705273921003)*x[1]
        arg[2,3,1,1]=(-0.0928709253845)*x[0]+(0.785115449756)*x[1]
        arg[2,3,2,0]=(-0.887243418838)*x[0]+(1.06790660642)*x[1]
        arg[2,3,2,1]=(-0.572644686872)*x[0]+(-0.672831948394)*x[1]
        arg[2,4,0,0]=(1.28282566801)*x[0]+(0.651985199447)*x[1]
        arg[2,4,0,1]=(0.0965663229335)*x[0]+(1.33819998052)*x[1]
        arg[2,4,1,0]=(0.205569041098)*x[0]+(-0.873095358733)*x[1]
        arg[2,4,1,1]=(-1.04452025745)*x[0]+(-0.204917033907)*x[1]
        arg[2,4,2,0]=(-0.181163938972)*x[0]+(0.0341993409272)*x[1]
        arg[2,4,2,1]=(1.68327027059)*x[0]+(-0.289436355642)*x[1]
        arg[3,0,0,0]=(1.05712810731)*x[0]+(0.915711819735)*x[1]
        arg[3,0,0,1]=(-0.171072039482)*x[0]+(-0.160656943623)*x[1]
        arg[3,0,1,0]=(-1.27063799869)*x[0]+(-0.605131642813)*x[1]
        arg[3,0,1,1]=(0.940512990835)*x[0]+(-0.116675577531)*x[1]
        arg[3,0,2,0]=(-0.194213869952)*x[0]+(0.364066752025)*x[1]
        arg[3,0,2,1]=(-0.328780465397)*x[0]+(-0.114689413673)*x[1]
        arg[3,1,0,0]=(-0.825914305895)*x[0]+(-0.167010056731)*x[1]
        arg[3,1,0,1]=(0.657779056207)*x[0]+(0.0666252451638)*x[1]
        arg[3,1,1,0]=(-0.464004426036)*x[0]+(-0.340932918409)*x[1]
        arg[3,1,1,1]=(-1.6698652667)*x[0]+(0.731870373141)*x[1]
        arg[3,1,2,0]=(1.39626250779)*x[0]+(-0.879152855557)*x[1]
        arg[3,1,2,1]=(1.34973919508)*x[0]+(-0.129753332761)*x[1]
        arg[3,2,0,0]=(1.03443450706)*x[0]+(0.807333477811)*x[1]
        arg[3,2,0,1]=(0.366011302596)*x[0]+(0.543804800926)*x[1]
        arg[3,2,1,0]=(0.262446846036)*x[0]+(-0.247786850991)*x[1]
        arg[3,2,1,1]=(-1.04657091196)*x[0]+(-0.485443426155)*x[1]
        arg[3,2,2,0]=(0.342434327538)*x[0]+(-0.148503585588)*x[1]
        arg[3,2,2,1]=(-0.189833429716)*x[0]+(0.424391874819)*x[1]
        arg[3,3,0,0]=(-0.118348838118)*x[0]+(0.397628264341)*x[1]
        arg[3,3,0,1]=(0.481440510084)*x[0]+(-0.393576221424)*x[1]
        arg[3,3,1,0]=(0.907533552208)*x[0]+(-0.668646549376)*x[1]
        arg[3,3,1,1]=(-1.49731714213)*x[0]+(-1.18710678264)*x[1]
        arg[3,3,2,0]=(-0.460660079207)*x[0]+(1.48078266444)*x[1]
        arg[3,3,2,1]=(-0.161073556988)*x[0]+(-0.486827042754)*x[1]
        arg[3,4,0,0]=(-0.668755455623)*x[0]+(0.345804059434)*x[1]
        arg[3,4,0,1]=(1.00528889193)*x[0]+(0.978507928191)*x[1]
        arg[3,4,1,0]=(0.98955856331)*x[0]+(1.0421852144)*x[1]
        arg[3,4,1,1]=(0.711688260873)*x[0]+(1.39962222322)*x[1]
        arg[3,4,2,0]=(1.08512429209)*x[0]+(-0.555249097743)*x[1]
        arg[3,4,2,1]=(-1.10443784007)*x[0]+(-0.782274429169)*x[1]
        ref[0,0,0,0]=(0.151153165564)/(o+1.)+(0.112609656739)+(0.254304335059)*0.5**o
        ref[0,0,0,1]=(-0.530602583511)/(o+1.)+(0.1817855378)+(-0.947398818743)*0.5**o
        ref[0,0,1,0]=(0.308747437738)/(o+1.)+(-0.0338129255643)+(-0.676658756184)*0.5**o
        ref[0,0,1,1]=(0.321026553427)/(o+1.)+(0.0692705367618)+(0.108863145337)*0.5**o
        ref[0,0,2,0]=(-0.0908034666655)/(o+1.)+(-0.1100205155)+(-0.517562009199)*0.5**o
        ref[0,0,2,1]=(0.303456040903)/(o+1.)+(0.500090875336)+(-0.254199784829)*0.5**o
        ref[0,1,0,0]=(-0.440641221831)/(o+1.)+(0.456218946873)+(-0.296391367396)*0.5**o
        ref[0,1,0,1]=(0.222628561762)/(o+1.)+(-0.162064412854)+(-0.457267302337)*0.5**o
        ref[0,1,1,0]=(-0.913361887408)/(o+1.)+(-0.356168674239)+(0.401664604431)*0.5**o
        ref[0,1,1,1]=(-0.543116639248)/(o+1.)+(-0.120902439422)+(0.381879135767)*0.5**o
        ref[0,1,2,0]=(0.993649262831)/(o+1.)+(0.384695858206)+(-0.55736653846)*0.5**o
        ref[0,1,2,1]=(-0.968077797451)/(o+1.)+(-0.187765942637)+(-0.603167099406)*0.5**o
        ref[0,2,0,0]=(0.873662019918)/(o+1.)+(-0.710635360033)+(-0.146359415223)*0.5**o
        ref[0,2,0,1]=(0.466408230693)/(o+1.)+(0.77626565285)+(0.389070815631)*0.5**o
        ref[0,2,1,0]=(-0.475010813579)/(o+1.)+(0.525435880006)+(0.960310046461)*0.5**o
        ref[0,2,1,1]=(0.697127520484)/(o+1.)+(0.357204276552)+(0.388328722275)*0.5**o
        ref[0,2,2,0]=(-0.108434738487)/(o+1.)+(0.356443300466)+(-0.97475753261)*0.5**o
        ref[0,2,2,1]=(0.952274964304)/(o+1.)+(-0.497379508976)+(-0.562006714089)*0.5**o
        ref[0,3,0,0]=(-0.0368012761671)/(o+1.)+(0.594000625518)+(0.770200680808)*0.5**o
        ref[0,3,0,1]=(0.450361861836)/(o+1.)+(0.182031052848)+(0.164263059934)*0.5**o
        ref[0,3,1,0]=(0.500714741956)/(o+1.)+(0.0479711343168)+(-0.0785996333735)*0.5**o
        ref[0,3,1,1]=(-0.949280938645)/(o+1.)+(0.452441874504)+(-0.88306753495)*0.5**o
        ref[0,3,2,0]=(0.186931712837)/(o+1.)+(0.222335175903)+(0.426318408694)*0.5**o
        ref[0,3,2,1]=(-0.288015428918)/(o+1.)+(0.80948657963)+(-0.822221815085)*0.5**o
        ref[0,4,0,0]=(0.670151184695)/(o+1.)+(-0.415883847013)+(0.376690471866)*0.5**o
        ref[0,4,0,1]=(-0.107106843992)/(o+1.)+(0.555375298435)+(-0.837698961768)*0.5**o
        ref[0,4,1,0]=(-0.299999096802)/(o+1.)+(0.668996782774)+(-0.669357604061)*0.5**o
        ref[0,4,1,1]=(0.778006183401)/(o+1.)+(-0.614473529032)+(-0.532655856076)*0.5**o
        ref[0,4,2,0]=(-0.250068389818)/(o+1.)+(0.156274314197)+(-0.0103473845504)*0.5**o
        ref[0,4,2,1]=(0.902125322216)/(o+1.)+(-0.616424567683)+(-0.560368879846)*0.5**o
        ref[1,0,0,0]=(-0.133690258295)/(o+1.)+(-0.715949360249)+(0.525524578427)*0.5**o
        ref[1,0,0,1]=(-0.788097689216)/(o+1.)+(0.0998405659256)+(-0.076709566994)*0.5**o
        ref[1,0,1,0]=(0.144280886187)/(o+1.)+(0.0775377290314)+(-0.696369546924)*0.5**o
        ref[1,0,1,1]=(0.824172196605)/(o+1.)+(0.561804029444)+(-0.0277357461773)*0.5**o
        ref[1,0,2,0]=(0.422564421802)/(o+1.)+(-0.60719307099)+(-0.417386842591)*0.5**o
        ref[1,0,2,1]=(0.778908672762)/(o+1.)+(0.0163446340765)+(-0.505974820938)*0.5**o
        ref[1,1,0,0]=(0.0574071047387)/(o+1.)+(-0.350514214742)+(0.608850746775)*0.5**o
        ref[1,1,0,1]=(-0.991660961792)/(o+1.)+(0.322084279455)+(0.380970772212)*0.5**o
        ref[1,1,1,0]=(-0.823761120268)/(o+1.)+(-0.0495419092898)+(0.930901327253)*0.5**o
        ref[1,1,1,1]=(-0.384729980569)/(o+1.)+(-0.0392583789163)+(0.649381188769)*0.5**o
        ref[1,1,2,0]=(-0.840365304069)/(o+1.)+(-0.0150512662777)+(0.517673506558)*0.5**o
        ref[1,1,2,1]=(0.555976660846)/(o+1.)+(-0.307265886898)+(0.707353625614)*0.5**o
        ref[1,2,0,0]=(-0.423056705491)/(o+1.)+(0.0559404393793)+(0.311083677036)*0.5**o
        ref[1,2,0,1]=(0.15409797664)/(o+1.)+(-0.0906637067519)+(-0.774473828018)*0.5**o
        ref[1,2,1,0]=(0.796718644088)/(o+1.)+(0.248012077098)+(0.776951250898)*0.5**o
        ref[1,2,1,1]=(-0.342830889263)/(o+1.)+(-0.282162737841)+(0.511417858487)*0.5**o
        ref[1,2,2,0]=(-0.722403016671)/(o+1.)+(-0.367059457954)+(0.230014491812)*0.5**o
        ref[1,2,2,1]=(-0.773685058881)/(o+1.)+(0.412807900451)+(-0.624365721366)*0.5**o
        ref[1,3,0,0]=(0.527481434758)/(o+1.)+(0.36602961117)+(0.031214145649)*0.5**o
        ref[1,3,0,1]=(0.0630927531585)/(o+1.)+(-0.164989701995)+(-0.325906787975)*0.5**o
        ref[1,3,1,0]=(-0.612390590593)/(o+1.)+(-0.78855996274)+(-0.63929738839)*0.5**o
        ref[1,3,1,1]=(-0.962656044265)/(o+1.)+(-0.00576535037253)+(-0.546330443673)*0.5**o
        ref[1,3,2,0]=(-0.947280922008)/(o+1.)+(-0.137449579996)+(0.747590859563)*0.5**o
        ref[1,3,2,1]=(-0.946623026202)/(o+1.)+(0.0787536550213)+(-0.24399757656)*0.5**o
        ref[1,4,0,0]=(-0.174471016943)/(o+1.)+(-0.636500556039)+(-0.635482528623)*0.5**o
        ref[1,4,0,1]=(0.806379391595)/(o+1.)+(-0.00304106577093)+(-0.476044717545)*0.5**o
        ref[1,4,1,0]=(0.729649125134)/(o+1.)+(0.0151477051616)+(-0.403091231646)*0.5**o
        ref[1,4,1,1]=(0.559976454745)/(o+1.)+(0.479295896421)+(0.75951513376)*0.5**o
        ref[1,4,2,0]=(0.272199270818)/(o+1.)+(-0.0929072274332)+(-0.843552757065)*0.5**o
        ref[1,4,2,1]=(-0.552663678377)/(o+1.)+(0.613784745778)+(-0.230300352975)*0.5**o
        ref[2,0,0,0]=(-0.174247908098)/(o+1.)+(-0.108873783282)+(-0.716067030882)*0.5**o
        ref[2,0,0,1]=(-0.300584841532)/(o+1.)+(-0.485363665493)+(-0.267230223262)*0.5**o
        ref[2,0,1,0]=(0.316261805154)/(o+1.)+(-0.523474106451)+(-0.888278086869)*0.5**o
        ref[2,0,1,1]=(-0.699622562861)/(o+1.)+(-0.151089538248)+(-0.909702635929)*0.5**o
        ref[2,0,2,0]=(0.728545639829)/(o+1.)+(0.192083148877)+(-0.0625371618174)*0.5**o
        ref[2,0,2,1]=(-0.22042267239)/(o+1.)+(-0.611974599324)+(0.677333343041)*0.5**o
        ref[2,1,0,0]=(-0.0734883811113)/(o+1.)+(0.36100831022)+(-0.890215236787)*0.5**o
        ref[2,1,0,1]=(0.121656829057)/(o+1.)+(-0.130801635799)+(-0.341933579678)*0.5**o
        ref[2,1,1,0]=(0.221353168629)/(o+1.)+(0.618635673483)+(0.540308531179)*0.5**o
        ref[2,1,1,1]=(0.0152580531273)/(o+1.)+(0.252173415542)+(0.560378877984)*0.5**o
        ref[2,1,2,0]=(0.580319777731)/(o+1.)+(0.90566612137)+(0.993226484879)*0.5**o
        ref[2,1,2,1]=(-0.710250320949)/(o+1.)+(0.742177021871)+(0.345780554724)*0.5**o
        ref[2,2,0,0]=(0.770701563507)/(o+1.)+(0.240823471881)+(0.399162680517)*0.5**o
        ref[2,2,0,1]=(0.76121761761)/(o+1.)+(0.128382656738)+(-0.460887633042)*0.5**o
        ref[2,2,1,0]=(-0.269854962057)/(o+1.)+(0.690854898294)+(-0.521701367567)*0.5**o
        ref[2,2,1,1]=(0.719518595712)/(o+1.)+(0.029596430244)+(-0.185365051187)*0.5**o
        ref[2,2,2,0]=(-0.671157761165)/(o+1.)+(-0.142688285012)+(-0.0423425346644)*0.5**o
        ref[2,2,2,1]=(-0.541156097112)/(o+1.)+(0.0765180918929)+(-0.379649661722)*0.5**o
        ref[2,3,0,0]=(-0.0604347109212)/(o+1.)+(-0.19403702073)+(-0.034905025507)*0.5**o
        ref[2,3,0,1]=(-0.793398507071)/(o+1.)+(0.255681280626)+(0.351720519037)*0.5**o
        ref[2,3,1,0]=(-0.267058413731)/(o+1.)+(0.852211426049)+(0.974567776717)*0.5**o
        ref[2,3,1,1]=(0.228396641207)/(o+1.)+(0.322922896735)+(-0.181997910305)*0.5**o
        ref[2,3,2,0]=(0.480987773937)/(o+1.)+(-0.197709028569)+(0.0950934707835)*0.5**o
        ref[2,3,2,1]=(-0.648934475027)/(o+1.)+(-0.418383297875)+(0.240224435511)*0.5**o
        ref[2,4,0,0]=(0.430020261528)/(o+1.)+(0.53421018733)+(0.436370231269)*0.5**o
        ref[2,4,0,1]=(0.906390660223)/(o+1.)+(-0.056057545856)+(0.640490734948)*0.5**o
        ref[2,4,1,0]=(-0.732099770943)/(o+1.)+(-0.0485027617732)+(0.161578976854)*0.5**o
        ref[2,4,1,1]=(-0.827495680848)/(o+1.)+(0.285044728936)+(-0.992031068384)*0.5**o
        ref[2,4,2,0]=(-0.919905471787)/(o+1.)+(0.656971386264)+(-0.541001898785)*0.5**o
        ref[2,4,2,1]=(0.161975875302)/(o+1.)+(0.272665206557)+(0.686527626533)*0.5**o
        ref[3,0,0,0]=(0.584120963259)/(o+1.)+(0.207930729804)+(0.972857504177)*0.5**o
        ref[3,0,0,1]=(0.315823141491)/(o+1.)+(-0.217561208386)+(-0.212429707823)*0.5**o
        ref[3,0,1,0]=(-0.732294996139)/(o+1.)+(-0.278653143596)+(-0.586168358175)*0.5**o
        ref[3,0,1,1]=(0.481321847225)/(o+1.)+(-0.070080648792)+(0.482676863662)*0.5**o
        ref[3,0,2,0]=(0.616679859076)/(o+1.)+(-0.418784546677)+(0.390742116352)*0.5**o
        ref[3,0,2,1]=(0.866778417448)/(o+1.)+(-0.686585656643)+(0.0629230167692)*0.5**o
        ref[3,1,0,0]=(0.531890054167)/(o+1.)+(-0.516187607633)+(-0.492439201527)*0.5**o
        ref[3,1,0,1]=(-0.512361639413)/(o+1.)+(0.499557182545)+(0.237651575694)*0.5**o
        ref[3,1,1,0]=(-0.67542965405)/(o+1.)+(0.274414249559)+(-0.678336189514)*0.5**o
        ref[3,1,1,1]=(0.371936242507)/(o+1.)+(-0.1636482466)+(-0.982634642862)*0.5**o
        ref[3,1,2,0]=(-0.799408831501)/(o+1.)+(0.458633399549)+(0.399251684639)*0.5**o
        ref[3,1,2,1]=(-0.0145253903686)/(o+1.)+(0.238472345905)+(0.757566560882)*0.5**o
        ref[3,2,0,0]=(-0.130519473112)/(o+1.)+(0.695277604837)+(0.581732248306)*0.5**o
        ref[3,2,0,1]=(0.632010536438)/(o+1.)+(0.123066205529)+(0.0316731560259)*0.5**o
        ref[3,2,1,0]=(-0.709926210356)/(o+1.)+(0.0926699972302)+(0.53924621094)*0.5**o
        ref[3,2,1,1]=(0.428094002247)/(o+1.)+(-0.841997926131)+(-0.276112488096)*0.5**o
        ref[3,2,2,0]=(0.708481148086)/(o+1.)+(-0.50498161534)+(0.495412824545)*0.5**o
        ref[3,2,2,1]=(0.782824622703)/(o+1.)+(-0.151509698687)+(-0.245246780227)*0.5**o
        ref[3,3,0,0]=(0.848460868418)/(o+1.)+(-0.104639248461)+(-0.359902945273)*0.5**o
        ref[3,3,0,1]=(0.0569151307298)/(o+1.)+(-0.0674224535469)+(0.165794065024)*0.5**o
        ref[3,3,1,0]=(-0.165683739908)/(o+1.)+(-0.183388699236)+(0.771348141212)*0.5**o
        ref[3,3,1,1]=(-0.932550058069)/(o+1.)+(-0.47828109001)+(-0.795311686688)*0.5**o
        ref[3,3,2,0]=(0.821542940923)/(o+1.)+(0.166853194487)+(-0.13512674466)*0.5**o
        ref[3,3,2,1]=(-0.46375871628)/(o+1.)+(0.132701267732)+(-0.449544418926)*0.5**o
        ref[3,4,0,0]=(-0.074530542168)/(o+1.)+(-0.0567877963797)+(-0.134845261261)*0.5**o
        ref[3,4,0,1]=(0.238671520479)/(o+1.)+(0.50689253572)+(0.731340228204)*0.5**o
        ref[3,4,1,0]=(0.981601513335)/(o+1.)+(0.400717209822)+(0.248707844726)*0.5**o
        ref[3,4,1,1]=(0.592304390107)/(o+1.)+(0.342408637328)+(0.834188819333)*0.5**o
        ref[3,4,2,0]=(-0.339991330909)/(o+1.)+(-0.0145922376156)+(0.899051000485)*0.5**o
        ref[3,4,2,1]=(-0.626610688164)/(o+1.)+(-0.417008165115)+(-0.426085250846)*0.5**o
      else:
        arg[0,0,0,0]=(0.697666818785)*x[0]+(1.00202295791)*x[1]+(0.617700157184)*x[2]
        arg[0,0,0,1]=(-0.465577819457)*x[0]+(-1.40054693355)*x[1]+(1.16912653295)*x[2]
        arg[0,0,1,0]=(-1.26696500634)*x[0]+(-0.0469756169975)*x[1]+(-0.445591738991)*x[2]
        arg[0,0,1,1]=(-1.61654321709)*x[0]+(-1.69522962467)*x[1]+(-1.06089094851)*x[2]
        arg[0,0,2,0]=(0.229468378506)*x[0]+(0.716813596991)*x[1]+(-0.237703937029)*x[2]
        arg[0,0,2,1]=(-1.12052051905)*x[0]+(0.835000004618)*x[1]+(-0.853004812404)*x[2]
        arg[0,1,0,0]=(-1.75500745705)*x[0]+(0.655055836882)*x[1]+(1.05207750735)*x[2]
        arg[0,1,0,1]=(-0.0746024824253)*x[0]+(-0.421596816958)*x[1]+(-0.375715855989)*x[2]
        arg[0,1,1,0]=(-0.963269684396)*x[0]+(0.431993248836)*x[1]+(-1.63061928649)*x[2]
        arg[0,1,1,1]=(-0.418757265018)*x[0]+(-0.665626168583)*x[1]+(-0.335734432507)*x[2]
        arg[0,1,2,0]=(0.980346404307)*x[0]+(1.45442049802)*x[1]+(-0.833420308612)*x[2]
        arg[0,1,2,1]=(-1.52536017847)*x[0]+(0.99484028681)*x[1]+(0.109754321663)*x[2]
        arg[0,2,0,0]=(1.697179262)*x[0]+(-0.928732802147)*x[1]+(-0.653379513626)*x[2]
        arg[0,2,0,1]=(0.682698902948)*x[0]+(0.829307959583)*x[1]+(-1.0300112276)*x[2]
        arg[0,2,1,0]=(0.026084323266)*x[0]+(1.28865784608)*x[1]+(0.830446580645)*x[2]
        arg[0,2,1,1]=(0.486291506032)*x[0]+(-0.519637742913)*x[1]+(0.81671330629)*x[2]
        arg[0,2,2,0]=(-0.761591624634)*x[0]+(0.370951131011)*x[1]+(-0.256748479312)*x[2]
        arg[0,2,2,1]=(-0.249974326513)*x[0]+(0.267323484635)*x[1]+(-1.09642130905)*x[2]
        arg[0,3,0,0]=(0.752057563255)*x[0]+(-1.2772262175)*x[1]+(-0.264569668983)*x[2]
        arg[0,3,0,1]=(-0.173167796447)*x[0]+(0.31873208356)*x[1]+(-0.599780701112)*x[2]
        arg[0,3,1,0]=(-0.912464014941)*x[0]+(-1.56452517166)*x[1]+(1.37651660499)*x[2]
        arg[0,3,1,1]=(0.302009667854)*x[0]+(0.524213064769)*x[1]+(0.316301522683)*x[2]
        arg[0,3,2,0]=(-1.2940154673)*x[0]+(-0.900061396992)*x[1]+(0.3860029847)*x[2]
        arg[0,3,2,1]=(-0.97354131341)*x[0]+(1.08221883964)*x[1]+(-0.739323076416)*x[2]
        arg[0,4,0,0]=(1.86535317192)*x[0]+(-0.478359589546)*x[1]+(-0.798014592302)*x[2]
        arg[0,4,0,1]=(-1.08518693851)*x[0]+(-0.186048654536)*x[1]+(1.04915573267)*x[2]
        arg[0,4,1,0]=(-1.09692811055)*x[0]+(1.01042920935)*x[1]+(-0.73978089555)*x[2]
        arg[0,4,1,1]=(0.654656964663)*x[0]+(0.279204596126)*x[1]+(0.886217781361)*x[2]
        arg[0,4,2,0]=(-0.54009886291)*x[0]+(0.378037248739)*x[1]+(1.09785568886)*x[2]
        arg[0,4,2,1]=(-1.43995958913)*x[0]+(-0.720157090994)*x[1]+(0.0676362028136)*x[2]
        arg[1,0,0,0]=(1.47735412608)*x[0]+(0.342355546228)*x[1]+(-0.0134502415508)*x[2]
        arg[1,0,0,1]=(-0.419536590369)*x[0]+(-1.61108470511)*x[1]+(1.51281979355)*x[2]
        arg[1,0,1,0]=(1.31765698245)*x[0]+(-0.7499566294)*x[1]+(0.934537503478)*x[2]
        arg[1,0,1,1]=(0.688467914883)*x[0]+(-1.32681280673)*x[1]+(-0.387405705058)*x[2]
        arg[1,0,2,0]=(1.37069587556)*x[0]+(0.540830111773)*x[1]+(0.477311068263)*x[2]
        arg[1,0,2,1]=(-0.572066213761)*x[0]+(-1.11950925494)*x[1]+(1.07939217872)*x[2]
        arg[1,1,0,0]=(0.564997105326)*x[0]+(-0.73130546389)*x[1]+(0.435910937304)*x[2]
        arg[1,1,0,1]=(0.513931739674)*x[0]+(-0.944043404826)*x[1]+(0.014114227468)*x[2]
        arg[1,1,1,0]=(-0.77850597256)*x[0]+(-0.445835136484)*x[1]+(0.157478357479)*x[2]
        arg[1,1,1,1]=(0.752431176866)*x[0]+(0.659759664199)*x[1]+(-0.859515433521)*x[2]
        arg[1,1,2,0]=(0.0693158730911)*x[0]+(-0.820997827204)*x[1]+(-0.714694436829)*x[2]
        arg[1,1,2,1]=(0.604674152847)*x[0]+(0.647261183132)*x[1]+(-1.21981352028)*x[2]
        arg[1,2,0,0]=(0.0997504695984)*x[0]+(-0.271549839083)*x[1]+(-0.573467835942)*x[2]
        arg[1,2,0,1]=(0.01861845581)*x[0]+(0.157410502406)*x[1]+(0.939257868776)*x[2]
        arg[1,2,1,0]=(1.44909894631)*x[0]+(0.116038517547)*x[1]+(-0.851940489233)*x[2]
        arg[1,2,1,1]=(1.34598841007)*x[0]+(1.79163125841)*x[1]+(0.719171044245)*x[2]
        arg[1,2,2,0]=(-0.214737791448)*x[0]+(-1.85810839095)*x[1]+(-0.240776599927)*x[2]
        arg[1,2,2,1]=(-0.574707645928)*x[0]+(-0.664173831956)*x[1]+(0.310262106696)*x[2]
        arg[1,3,0,0]=(0.296213453592)*x[0]+(-0.831112482555)*x[1]+(0.130075739128)*x[2]
        arg[1,3,0,1]=(-0.369510604457)*x[0]+(0.0373909697998)*x[1]+(1.6689129783)*x[2]
        arg[1,3,1,0]=(0.0922435881419)*x[0]+(-0.280519827004)*x[1]+(0.285933839232)*x[2]
        arg[1,3,1,1]=(-0.649253282272)*x[0]+(-0.460065044123)*x[1]+(-0.0711164948063)*x[2]
        arg[1,3,2,0]=(0.793313307581)*x[0]+(-0.0587956351149)*x[1]+(0.810487259739)*x[2]
        arg[1,3,2,1]=(-1.38668597055)*x[0]+(1.51396934198)*x[1]+(-1.13918550255)*x[2]
        arg[1,4,0,0]=(1.40078167737)*x[0]+(-1.61915436345)*x[1]+(0.421710091214)*x[2]
        arg[1,4,0,1]=(-1.37982015877)*x[0]+(1.30607371738)*x[1]+(-0.608888848419)*x[2]
        arg[1,4,1,0]=(-0.558087802805)*x[0]+(-0.214169499512)*x[1]+(-0.766242546503)*x[2]
        arg[1,4,1,1]=(-1.75503436584)*x[0]+(0.183604705451)*x[1]+(-0.148071484576)*x[2]
        arg[1,4,2,0]=(0.537942069142)*x[0]+(-0.00640749928252)*x[1]+(-0.277544629657)*x[2]
        arg[1,4,2,1]=(-0.652645833371)*x[0]+(-1.33439077026)*x[1]+(-0.226308197201)*x[2]
        arg[2,0,0,0]=(1.45646289279)*x[0]+(1.40842127675)*x[1]+(1.75078281495)*x[2]
        arg[2,0,0,1]=(0.867988651995)*x[0]+(1.35061812241)*x[1]+(-1.56520776323)*x[2]
        arg[2,0,1,0]=(0.549686523522)*x[0]+(-0.528085465813)*x[1]+(-0.994978820267)*x[2]
        arg[2,0,1,1]=(0.383595996552)*x[0]+(0.612989859087)*x[1]+(-0.560401247231)*x[2]
        arg[2,0,2,0]=(-0.341259919499)*x[0]+(-0.125491675042)*x[1]+(0.545572340757)*x[2]
        arg[2,0,2,1]=(-0.414269395434)*x[0]+(-0.433969881898)*x[1]+(-1.01820199328)*x[2]
        arg[2,1,0,0]=(-0.113575280389)*x[0]+(0.326916039942)*x[1]+(-0.576072890249)*x[2]
        arg[2,1,0,1]=(-0.296473969661)*x[0]+(0.195675099431)*x[1]+(-1.72378150831)*x[2]
        arg[2,1,1,0]=(-1.31575932758)*x[0]+(-0.122557423522)*x[1]+(0.53435027695)*x[2]
        arg[2,1,1,1]=(0.299435754608)*x[0]+(0.727140413706)*x[1]+(0.848931358475)*x[2]
        arg[2,1,2,0]=(0.2618965198)*x[0]+(-0.461672573339)*x[1]+(0.923592803172)*x[2]
        arg[2,1,2,1]=(-0.326817258523)*x[0]+(-0.0363300805442)*x[1]+(1.82049875918)*x[2]
        arg[2,2,0,0]=(-0.435113320888)*x[0]+(0.0257984353786)*x[1]+(0.287846577657)*x[2]
        arg[2,2,0,1]=(1.39178068279)*x[0]+(0.136781967539)*x[1]+(-1.38161590351)*x[2]
        arg[2,2,1,0]=(-0.887874199356)*x[0]+(-0.476769153654)*x[1]+(-0.618760485829)*x[2]
        arg[2,2,1,1]=(1.58883355098)*x[0]+(-0.889469047555)*x[1]+(-0.678486755843)*x[2]
        arg[2,2,2,0]=(0.237147029365)*x[0]+(-0.509883462401)*x[1]+(0.533163151551)*x[2]
        arg[2,2,2,1]=(0.66513913749)*x[0]+(1.30317030657)*x[1]+(0.804083365617)*x[2]
        arg[2,3,0,0]=(1.01531518403)*x[0]+(1.3753053143)*x[1]+(0.000877284235964)*x[2]
        arg[2,3,0,1]=(-0.40853642543)*x[0]+(0.154794268783)*x[1]+(-1.34306671117)*x[2]
        arg[2,3,1,0]=(0.43065284359)*x[0]+(-0.28965146009)*x[1]+(0.660248828789)*x[2]
        arg[2,3,1,1]=(-1.4334922681)*x[0]+(-0.0348802858609)*x[1]+(0.502051304974)*x[2]
        arg[2,3,2,0]=(-1.17399537487)*x[0]+(0.865787043946)*x[1]+(-0.0105949677478)*x[2]
        arg[2,3,2,1]=(0.63104158794)*x[0]+(0.436196079951)*x[1]+(0.498606623061)*x[2]
        arg[2,4,0,0]=(-1.44410177354)*x[0]+(0.513406326011)*x[1]+(0.699127199447)*x[2]
        arg[2,4,0,1]=(-0.981727023077)*x[0]+(0.484182120695)*x[1]+(0.667175832045)*x[2]
        arg[2,4,1,0]=(-1.3352602588)*x[0]+(0.626269023364)*x[1]+(0.813537646695)*x[2]
        arg[2,4,1,1]=(-0.932486930114)*x[0]+(0.299448807849)*x[1]+(0.195862014797)*x[2]
        arg[2,4,2,0]=(-0.0621489644806)*x[0]+(0.131979664158)*x[1]+(0.442029269778)*x[2]
        arg[2,4,2,1]=(-0.507508608633)*x[0]+(-1.28613226533)*x[1]+(-0.697920444023)*x[2]
        arg[3,0,0,0]=(-0.984942677585)*x[0]+(-1.58913299616)*x[1]+(-0.373026784822)*x[2]
        arg[3,0,0,1]=(0.616666427113)*x[0]+(-0.625162800471)*x[1]+(-0.796307290839)*x[2]
        arg[3,0,1,0]=(-0.48469203683)*x[0]+(-0.0572105166553)*x[1]+(0.809572383581)*x[2]
        arg[3,0,1,1]=(-0.487160309892)*x[0]+(0.77646454341)*x[1]+(-0.41947544573)*x[2]
        arg[3,0,2,0]=(-1.17902121554)*x[0]+(-0.244939532033)*x[1]+(-0.282994359476)*x[2]
        arg[3,0,2,1]=(0.33515909881)*x[0]+(0.146866561705)*x[1]+(1.81486098364)*x[2]
        arg[3,1,0,0]=(0.379977901341)*x[0]+(0.834531701517)*x[1]+(-0.0242027062166)*x[2]
        arg[3,1,0,1]=(0.184318292766)*x[0]+(1.08367911805)*x[1]+(-0.164531741594)*x[2]
        arg[3,1,1,0]=(-1.00967115996)*x[0]+(0.542699640149)*x[1]+(0.707616701259)*x[2]
        arg[3,1,1,1]=(0.856948233811)*x[0]+(-1.2186550297)*x[1]+(-0.034268798159)*x[2]
        arg[3,1,2,0]=(0.259470684477)*x[0]+(-1.03586983732)*x[1]+(-0.157275979341)*x[2]
        arg[3,1,2,1]=(-0.161954890658)*x[0]+(1.65853209885)*x[1]+(0.381113456719)*x[2]
        arg[3,2,0,0]=(-0.437413004681)*x[0]+(-0.730340273929)*x[1]+(-0.687689272494)*x[2]
        arg[3,2,0,1]=(-0.336618164118)*x[0]+(-1.24388666895)*x[1]+(-0.2198689254)*x[2]
        arg[3,2,1,0]=(0.0872793087887)*x[0]+(-0.336554573215)*x[1]+(-0.0597803523262)*x[2]
        arg[3,2,1,1]=(0.192805432018)*x[0]+(0.0540495582064)*x[1]+(0.846658042071)*x[2]
        arg[3,2,2,0]=(0.00794192604414)*x[0]+(-0.33504559053)*x[1]+(0.615701149787)*x[2]
        arg[3,2,2,1]=(0.210870961503)*x[0]+(1.14164095846)*x[1]+(-0.192785742791)*x[2]
        arg[3,3,0,0]=(-0.25895162041)*x[0]+(-0.225800436488)*x[1]+(-0.198007079838)*x[2]
        arg[3,3,0,1]=(-0.150363703231)*x[0]+(1.23870429535)*x[1]+(0.28635881247)*x[2]
        arg[3,3,1,0]=(-0.865980718216)*x[0]+(-0.0849853244335)*x[1]+(-0.292164217189)*x[2]
        arg[3,3,1,1]=(-0.350465968722)*x[0]+(0.09853736936)*x[1]+(-0.476711524665)*x[2]
        arg[3,3,2,0]=(0.311396300263)*x[0]+(-0.92437612034)*x[1]+(-0.757139962928)*x[2]
        arg[3,3,2,1]=(-0.912311919837)*x[0]+(0.267793631099)*x[1]+(0.4917345401)*x[2]
        arg[3,4,0,0]=(0.613662541753)*x[0]+(0.782757341361)*x[1]+(1.04209808158)*x[2]
        arg[3,4,0,1]=(0.492828613997)*x[0]+(1.11979095664)*x[1]+(-0.843545072354)*x[2]
        arg[3,4,1,0]=(-0.51939120556)*x[0]+(0.476004484286)*x[1]+(-0.516311785432)*x[2]
        arg[3,4,1,1]=(1.00423313567)*x[0]+(0.90776890388)*x[1]+(1.34423309497)*x[2]
        arg[3,4,2,0]=(-1.44069600826)*x[0]+(-1.39515290909)*x[1]+(1.19851590247)*x[2]
        arg[3,4,2,1]=(0.391851619877)*x[0]+(0.265529120553)*x[1]+(0.996610410913)*x[2]
        ref[0,0,0,0]=(1.19330263935)/(o+1.)+(0.55394275251)+(0.0162017895041)*0.5**o
        ref[0,0,0,1]=(-0.672351783684)/(o+1.)+(-0.187889622039)+(0.351132807711)*0.5**o
        ref[0,0,1,0]=(-1.27021996264)/(o+1.)+(0.236286530101)+(-0.961885459894)*0.5**o
        ref[0,0,1,1]=(-1.43442517107)/(o+1.)+(-1.00113374859)+(-0.935971122022)*0.5**o
        ref[0,0,2,0]=(-0.434813695127)/(o+1.)+(0.606673209583)+(-0.0699546855712)*0.5**o
        ref[0,0,2,1]=(-0.0434901247948)/(o+1.)+(-0.180911877983)+(-0.733211446073)*0.5**o
        ref[0,1,0,0]=(1.41514302182)/(o+1.)+(-0.247977413325)+(-0.967062307989)*0.5**o
        ref[0,1,0,1]=(0.69128100142)/(o+1.)+(-0.621609882923)+(-0.319976390946)*0.5**o
        ref[0,1,1,0]=(-0.881870429191)/(o+1.)+(-0.654075774082)+(0.0281262553057)*0.5**o
        ref[0,1,1,1]=(-1.27239116086)/(o+1.)+(-0.0854066429492)+(0.0230865806541)*0.5**o
        ref[0,1,2,0]=(-0.222663945351)/(o+1.)+(0.685312598747)+(0.453385341573)*0.5**o
        ref[0,1,2,1]=(0.101409038706)/(o+1.)+(0.151367946744)+(-0.824910502191)*0.5**o
        ref[0,2,0,0]=(-1.19509443892)/(o+1.)+(0.225437856866)+(0.85928567141)*0.5**o
        ref[0,2,0,1]=(0.285764347754)/(o+1.)+(-0.383432127133)+(0.963095541446)*0.5**o
        ref[0,2,1,0]=(0.728010227911)/(o+1.)+(0.983790919896)+(-0.550403317717)*0.5**o
        ref[0,2,1,1]=(0.563748259036)/(o+1.)+(0.165164845927)+(-0.110710881481)*0.5**o
        ref[0,2,2,0]=(-0.361502606429)/(o+1.)+(0.0276145482611)+(-0.341115463028)*0.5**o
        ref[0,2,2,1]=(-0.293320207933)/(o+1.)+(0.0336757304385)+(-0.853103403872)*0.5**o
        ref[0,3,0,0]=(-0.410233956527)/(o+1.)+(-0.474281126245)+(0.569057885785)*0.5**o
        ref[0,3,0,1]=(-0.835861970115)/(o+1.)+(0.0173715483175)+(0.346902459481)*0.5**o
        ref[0,3,1,0]=(0.0830466456187)/(o+1.)+(-0.288204681495)+(-0.607109864241)*0.5**o
        ref[0,3,1,1]=(1.07074130823)/(o+1.)+(-0.0437983012867)+(0.159379549652)*0.5**o
        ref[0,3,2,0]=(0.172301049609)/(o+1.)+(-0.491724874152)+(-0.996925180895)*0.5**o
        ref[0,3,2,1]=(-0.24891941221)/(o+1.)+(-0.03935856269)+(-0.3030090126)*0.5**o
        ref[0,4,0,0]=(-0.778575122429)/(o+1.)+(0.213085222078)+(0.941383668349)*0.5**o
        ref[0,4,0,1]=(0.880801627174)/(o+1.)+(-0.454451723401)+(-0.193978040747)*0.5**o
        ref[0,4,1,0]=(-0.835604724882)/(o+1.)+(0.127240834083)+(-0.245156740036)*0.5**o
        ref[0,4,1,1]=(1.62336716324)/(o+1.)+(-0.101085918464)+(0.398884015839)*0.5**o
        ref[0,4,2,0]=(1.16805827702)/(o+1.)+(0.211232591688)+(-0.654729385702)*0.5**o
        ref[0,4,2,1]=(-0.305424138957)/(o+1.)+(-0.431076572693)+(-0.924903192969)*0.5**o
        ref[1,0,0,0]=(-0.481596247529)/(o+1.)+(0.769902632559)+(0.748050413171)*0.5**o
        ref[1,0,0,1]=(0.177388661252)/(o+1.)+(-0.118766442613)+(-0.457657277952)*0.5**o
        ref[1,0,1,0]=(0.709890860372)/(o+1.)+(0.0668010230727)+(0.658744950012)*0.5**o
        ref[1,0,1,1]=(-1.45685781381)/(o+1.)+(-0.178674426089)+(0.788456069079)*0.5**o
        ref[1,0,2,0]=(0.194388185612)/(o+1.)+(0.649177543063)+(0.896093783855)*0.5**o
        ref[1,0,2,1]=(0.181561279507)/(o+1.)+(-0.200618291902)+(-0.392507985678)*0.5**o
        ref[1,1,0,0]=(-0.884376814027)/(o+1.)+(0.280382736159)+(0.593213920451)*0.5**o
        ref[1,1,0,1]=(-0.250709602218)/(o+1.)+(-0.492104211211)+(0.818920586956)*0.5**o
        ref[1,1,1,0]=(0.0529723694625)/(o+1.)+(-0.502047650638)+(-0.115739819752)*0.5**o
        ref[1,1,1,1]=(0.215812202957)/(o+1.)+(-0.308796450292)+(0.954456105171)*0.5**o
        ref[1,1,2,0]=(-0.991481050902)/(o+1.)+(0.162626252838)+(-0.800147845717)*0.5**o
        ref[1,1,2,1]=(-0.433989756951)/(o+1.)+(0.362776336712)+(-0.259441100777)*0.5**o
        ref[1,2,0,0]=(-0.930287052624)/(o+1.)+(-0.347752763108)+(0.880525373413)*0.5**o
        ref[1,2,0,1]=(0.698674852861)/(o+1.)+(-0.15119613196)+(0.719004238051)*0.5**o
        ref[1,2,1,0]=(-0.777216821365)/(o+1.)+(0.436336076297)+(0.617741643399)*0.5**o
        ref[1,2,1,1]=(1.09595223135)/(o+1.)+(0.978251130983)+(0.804336219412)*0.5**o
        ref[1,2,2,0]=(-0.651012630849)/(o+1.)+(-1.04894606664)+(0.43528198181)*0.5**o
        ref[1,2,2,1]=(-1.5227353883)/(o+1.)+(0.670019307557)+(-0.745922598001)*0.5**o
        ref[1,3,0,0]=(-0.750195637997)/(o+1.)+(0.231864245348)+(-0.118356142533)*0.5**o
        ref[1,3,0,1]=(1.17136226596)/(o+1.)+(-0.114976421389)+(0.395383920468)*0.5**o
        ref[1,3,1,0]=(-0.108070636769)/(o+1.)+(0.279145429234)+(-0.352562621327)*0.5**o
        ref[1,3,1,1]=(-0.144223481714)/(o+1.)+(-0.472954352493)+(-0.0903026345013)*0.5**o
        ref[1,3,2,0]=(-0.487564307786)/(o+1.)+(0.636973351449)+(0.758622537094)*0.5**o
        ref[1,3,2,1]=(0.390860965954)/(o+1.)+(-0.36647678165)+(-0.66980953377)*0.5**o
        ref[1,4,0,0]=(-1.04824984837)/(o+1.)+(0.213932231728)+(0.823722790054)*0.5**o
        ref[1,4,0,1]=(-0.0984653601336)/(o+1.)+(0.07867877705)+(-0.741527483773)*0.5**o
        ref[1,4,1,0]=(-0.146287867383)/(o+1.)+(-0.908951868158)+(0.425691754878)*0.5**o
        ref[1,4,1,1]=(-0.94160416554)/(o+1.)+(0.0459877010016)+(-0.869872381431)*0.5**o
        ref[1,4,2,0]=(0.0115742858952)/(o+1.)+(-0.222488399041)+(0.687392452389)*0.5**o
        ref[1,4,2,1]=(-0.956561936575)/(o+1.)+(-0.59770879042)+(-0.0613652834159)*0.5**o
        ref[2,0,0,0]=(1.83987381729)/(o+1.)+(0.956579317963)+(0.862634531274)*0.5**o
        ref[2,0,0,1]=(0.105446127421)/(o+1.)+(0.180661881191)+(0.186629121375)*0.5**o
        ref[2,0,1,0]=(0.0212167440766)/(o+1.)+(-0.337825588888)+(-0.318943328859)*0.5**o
        ref[2,0,1,1]=(0.744834715631)/(o+1.)+(-0.350058416663)+(0.391466726104)*0.5**o
        ref[2,0,2,0]=(-0.788776613856)/(o+1.)+(0.233776806014)+(0.400043748043)*0.5**o
        ref[2,0,2,1]=(-0.95346976307)/(o+1.)+(0.0274998652324)+(-0.967971238003)*0.5**o
        ref[2,1,0,0]=(-0.536051408942)/(o+1.)+(-0.208498016994)+(0.590315312233)*0.5**o
        ref[2,1,0,1]=(-1.04158508788)/(o+1.)+(-0.397885793101)+(0.0127762955396)*0.5**o
        ref[2,1,1,0]=(0.529914207253)/(o+1.)+(-0.547503443987)+(-0.338873793431)*0.5**o
        ref[2,1,1,1]=(0.62298670274)/(o+1.)+(0.449860048786)+(0.352800726478)*0.5**o
        ref[2,1,2,0]=(0.965292650595)/(o+1.)+(0.00980498639)+(-0.261085873742)*0.5**o
        ref[2,1,2,1]=(0.403285360803)/(o+1.)+(0.369160592269)+(0.315744874774)*0.5**o
        ref[2,2,0,0]=(-0.457764884724)/(o+1.)+(0.563657776706)+(-0.791018976541)*0.5**o
        ref[2,2,0,1]=(0.188151274044)/(o+1.)+(-0.476470545911)+(0.911736564588)*0.5**o
        ref[2,2,1,0]=(-0.558515085739)/(o+1.)+(-0.445583796553)+(-0.533721159994)*0.5**o
        ref[2,2,1,1]=(-0.431992337265)/(o+1.)+(-0.0957726221961)+(0.64441532924)*0.5**o
        ref[2,2,2,0]=(-0.159056461947)/(o+1.)+(-0.0326864536156)+(0.484856087693)*0.5**o
        ref[2,2,2,1]=(0.533116456037)/(o+1.)+(1.10197257725)+(0.0353311991352)*0.5**o
        ref[2,3,0,0]=(0.0570625147023)/(o+1.)+(0.74170779854)+(0.851019670777)*0.5**o
        ref[2,3,0,1]=(-0.495059904832)/(o+1.)+(-0.169954912181)+(-0.761839138623)*0.5**o
        ref[2,3,1,0]=(0.482375173314)/(o+1.)+(-0.19374838651)+(0.706371811996)*0.5**o
        ref[2,3,1,1]=(0.12928100721)/(o+1.)+(-0.317327862062)+(-0.460946532077)*0.5**o
        ref[2,3,2,0]=(-0.234362084763)/(o+1.)+(0.423486279181)+(-0.93141377227)*0.5**o
        ref[2,3,2,1]=(1.25586693002)/(o+1.)+(-0.127551329993)+(0.56508002092)*0.5**o
        ref[2,4,0,0]=(0.45235297749)/(o+1.)+(-0.0217537364788)+(-0.640413752615)*0.5**o
        ref[2,4,0,1]=(0.936809207384)/(o+1.)+(-0.151519610695)+(-0.464139056331)*0.5**o
        ref[2,4,1,0]=(0.867693208863)/(o+1.)+(0.105683574314)+(-0.974513946233)*0.5**o
        ref[2,4,1,1]=(-1.27821575101)/(o+1.)+(0.463163847311)+(-0.0852880510814)*0.5**o
        ref[2,4,2,0]=(-0.846224501366)/(o+1.)+(0.432619461409)+(0.492845548002)*0.5**o
        ref[2,4,2,1]=(-1.48923229995)/(o+1.)+(-0.505747396375)+(0.00916577472118)*0.5**o
        ref[3,0,0,0]=(-0.959215731121)/(o+1.)+(-0.904609108587)+(-0.178668510268)*0.5**o
        ref[3,0,0,1]=(-1.53181231865)/(o+1.)+(-0.0758543610408)+(0.878717376536)*0.5**o
        ref[3,0,1,0]=(1.18643238831)/(o+1.)+(-0.288825159097)+(-0.341112240015)*0.5**o
        ref[3,0,1,1]=(-0.0675694917031)/(o+1.)+(-0.168955703616)+(0.275309686724)*0.5**o
        ref[3,0,2,0]=(-0.663269860133)/(o+1.)+(-0.235102421885)+(-0.573480403148)*0.5**o
        ref[3,0,2,1]=(1.0969289778)/(o+1.)+(0.46603172821)+(0.267894209932)*0.5**o
        ref[3,1,0,0]=(-0.716870513559)/(o+1.)+(0.712039409283)+(0.483098591635)*0.5**o
        ref[3,1,0,1]=(0.0328531999411)/(o+1.)+(0.379319450988)+(0.311973567308)*0.5**o
        ref[3,1,1,0]=(1.08648208064)/(o+1.)+(-0.196831959338)+(-0.452172980519)*0.5**o
        ref[3,1,1,1]=(0.0813181502612)/(o+1.)+(-0.227364064527)+(-0.0225656152548)*0.5**o
        ref[3,1,2,0]=(-0.271000454408)/(o+1.)+(-0.765711687155)+(0.868748696531)*0.5**o
        ref[3,1,2,1]=(1.60569646253)/(o+1.)+(0.137437739736)+(-0.00288127709197)*0.5**o
        ref[3,2,0,0]=(-1.25887089059)/(o+1.)+(-0.525842290945)+(0.455112921376)*0.5**o
        ref[3,2,0,1]=(-1.63787880725)/(o+1.)+(0.00214706725405)+(-0.166789085722)*0.5**o
        ref[3,2,1,0]=(-1.82111307894)/(o+1.)+(0.561806270366)+(0.388444921454)*0.5**o
        ref[3,2,1,1]=(-0.489259906414)/(o+1.)+(0.582548398381)+(0.417676141947)*0.5**o
        ref[3,2,2,0]=(-0.0674375340391)/(o+1.)+(-0.00345549227349)+(0.362946003887)*0.5**o
        ref[3,2,2,1]=(0.662780682001)/(o+1.)+(-0.15453848218)+(0.806022459534)*0.5**o
        ref[3,3,0,0]=(-1.15795406515)/(o+1.)+(0.0926725746467)+(0.289849779118)*0.5**o
        ref[3,3,0,1]=(1.27520269547)/(o+1.)+(-0.286529818361)+(0.672556345847)*0.5**o
        ref[3,3,1,0]=(-0.343290651469)/(o+1.)+(-0.0269962100904)+(-0.845847188189)*0.5**o
        ref[3,3,1,1]=(0.497715053371)/(o+1.)+(-0.158577177766)+(-0.909200821866)*0.5**o
        ref[3,3,2,0]=(-0.952431538662)/(o+1.)+(0.0816482063045)+(-0.580984656952)*0.5**o
        ref[3,3,2,1]=(-0.176227469799)/(o+1.)+(0.0456157853688)+(-0.0677878495767)*0.5**o
        ref[3,4,0,0]=(0.976460483802)/(o+1.)+(0.347595506011)+(0.766866468871)*0.5**o
        ref[3,4,0,1]=(0.0358170327662)/(o+1.)+(-0.0991456259778)+(0.931548717473)*0.5**o
        ref[3,4,1,0]=(0.114544318514)/(o+1.)+(-0.201014865792)+(-0.272213093635)*0.5**o
        ref[3,4,1,1]=(0.554509925385)/(o+1.)+(1.23219332557)+(0.237338557994)*0.5**o
        ref[3,4,2,0]=(0.0803465444931)/(o+1.)+(-0.518257569997)+(-0.681164419371)*0.5**o
        ref[3,4,2,1]=(-0.0443609570522)/(o+1.)+(0.572813368096)+(0.552725372202)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.570707549273)*x[0]+(-0.726264032186)*x[1]
        ref=(-0.0423765208873)/(o+1.)+(-0.791542656673)+(0.328490252775)*0.5**o
      else:
        arg=(1.28222892931)*x[0]+(-0.798170675576)*x[1]+(1.5701208405)*x[2]
        ref=(0.721975372868)/(o+1.)+(0.344720905832)+(0.642761909705)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.575166906798)*x[0]+(-1.63698773418)*x[1]
        arg[1]=(-0.0535946584636)*x[0]+(0.105969101775)*x[1]
        ref[0]=(-0.751926426344)/(o+1.)+(-0.395800138888)+(0.481705876734)*0.5**o
        ref[1]=(0.698496252242)/(o+1.)+(-0.265927161866)+(-0.114267485199)*0.5**o
      else:
        arg[0]=(-0.17879254138)*x[0]+(-1.12752123792)*x[1]+(0.921297311605)*x[2]
        arg[1]=(1.33131824618)*x[0]+(-0.392175059176)*x[1]+(0.545754281465)*x[2]
        ref[0]=(-0.494870303061)/(o+1.)+(0.0697636237654)+(-0.0296734121631)*0.5**o
        ref[1]=(-0.394300983704)/(o+1.)+(0.722261312754)+(0.434675826665)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.338658191734)*x[0]+(1.46127609683)*x[1]
        arg[0,1]=(-0.629116867491)*x[0]+(1.40141569066)*x[1]
        arg[0,2]=(0.105569779347)*x[0]+(0.239936324967)*x[1]
        arg[0,3]=(-0.281718331718)*x[0]+(0.280323404866)*x[1]
        arg[0,4]=(0.467341113661)*x[0]+(-0.50452062075)*x[1]
        arg[1,0]=(-0.841295219186)*x[0]+(0.524038163723)*x[1]
        arg[1,1]=(1.40143704236)*x[0]+(-1.50467961637)*x[1]
        arg[1,2]=(1.21078790505)*x[0]+(-0.731453470292)*x[1]
        arg[1,3]=(0.469556922076)*x[0]+(0.0686293204391)*x[1]
        arg[1,4]=(-1.26394637585)*x[0]+(-0.0361395276585)*x[1]
        arg[2,0]=(-1.12850883148)*x[0]+(0.284296673326)*x[1]
        arg[2,1]=(-0.626362828813)*x[0]+(1.19459827573)*x[1]
        arg[2,2]=(-0.545457293083)*x[0]+(1.547337019)*x[1]
        arg[2,3]=(0.664461986629)*x[0]+(-1.69082198266)*x[1]
        arg[2,4]=(-0.100004763305)*x[0]+(0.152159044423)*x[1]
        arg[3,0]=(0.307574843061)*x[0]+(1.20209181301)*x[1]
        arg[3,1]=(-0.0663203356627)*x[0]+(-1.26669863527)*x[1]
        arg[3,2]=(-1.20436798894)*x[0]+(-0.0375693379661)*x[1]
        arg[3,3]=(0.753649087747)*x[0]+(0.69376223486)*x[1]
        arg[3,4]=(-0.0512732264834)*x[0]+(-0.658097606436)*x[1]
        ref[0,0]=(0.996821439952)/(o+1.)+(0.015088160779)+(0.0956201435879)*0.5**o
        ref[0,1]=(0.643092963726)/(o+1.)+(0.239390724844)+(-0.349575590244)*0.5**o
        ref[0,2]=(-0.198468008977)/(o+1.)+(0.00411201743342)+(0.535750078425)*0.5**o
        ref[0,3]=(0.642491201137)/(o+1.)+(-0.623413199652)+(0.602940271315)*0.5**o
        ref[0,4]=(-0.105706120549)/(o+1.)+(-0.353560054939)+(0.775646723338)*0.5**o
        ref[1,0]=(0.144540669221)/(o+1.)+(0.0228268215803)+(-0.507451367845)*0.5**o
        ref[1,1]=(-0.820379339897)/(o+1.)+(0.0951825678509)+(0.526771630185)*0.5**o
        ref[1,2]=(-0.21214585824)/(o+1.)+(-0.0889183062985)+(0.869316905591)*0.5**o
        ref[1,3]=(0.291228940674)/(o+1.)+(0.192623928703)+(-0.138290555565)*0.5**o
        ref[1,4]=(-0.0663718956434)/(o+1.)+(-0.309484335614)+(-0.614745336642)*0.5**o
        ref[2,0]=(0.484892962201)/(o+1.)+(-0.367129859793)+(-0.594845400763)*0.5**o
        ref[2,1]=(0.768457634672)/(o+1.)+(0.154447657562)+(-0.509117502876)*0.5**o
        ref[2,2]=(0.707320338363)/(o+1.)+(-0.0592149232183)+(0.412989233992)*0.5**o
        ref[2,3]=(-0.983684339171)/(o+1.)+(-0.495377645598)+(0.948079634337)*0.5**o
        ref[2,4]=(-0.0315947220465)/(o+1.)+(0.506833986759)+(-0.929918970353)*0.5**o
        ref[3,0]=(0.54141067098)/(o+1.)+(0.489049131493)+(-0.00984227789859)*0.5**o
        ref[3,1]=(-0.552335053051)/(o+1.)+(-0.383325638803)+(-0.0140326402734)*0.5**o
        ref[3,2]=(0.0821472412575)/(o+1.)+(-0.335827719718)+(-0.652429128733)*0.5**o
        ref[3,3]=(0.112917208924)/(o+1.)+(0.667772807338)+(-0.00105150099312)*0.5**o
        ref[3,4]=(-0.19331508157)/(o+1.)+(-0.102296385708)+(-0.311462979932)*0.5**o
      else:
        arg[0,0]=(0.069245963142)*x[0]+(-0.185727379494)*x[1]+(-0.887410889699)*x[2]
        arg[0,1]=(-0.816745668949)*x[0]+(-0.250435680662)*x[1]+(0.701607830582)*x[2]
        arg[0,2]=(-0.862433954308)*x[0]+(0.731874372527)*x[1]+(-0.115384982779)*x[2]
        arg[0,3]=(-0.265247631604)*x[0]+(-0.157059741958)*x[1]+(-0.368419335078)*x[2]
        arg[0,4]=(1.66720203567)*x[0]+(-0.759393708719)*x[1]+(1.36360850704)*x[2]
        arg[1,0]=(-0.192822813795)*x[0]+(0.0358776152527)*x[1]+(-0.722864520788)*x[2]
        arg[1,1]=(-0.480606215471)*x[0]+(1.13682774573)*x[1]+(0.17900467623)*x[2]
        arg[1,2]=(0.346760852584)*x[0]+(0.994833030281)*x[1]+(0.507055933913)*x[2]
        arg[1,3]=(-0.509961853478)*x[0]+(-0.852997575505)*x[1]+(0.586527030253)*x[2]
        arg[1,4]=(0.93957975846)*x[0]+(0.28878335152)*x[1]+(0.128070931738)*x[2]
        arg[2,0]=(0.338283306313)*x[0]+(0.0244500761608)*x[1]+(-0.606875554888)*x[2]
        arg[2,1]=(0.850782624552)*x[0]+(-0.836382038189)*x[1]+(-1.41808459947)*x[2]
        arg[2,2]=(-1.06176229613)*x[0]+(-0.259247046061)*x[1]+(-0.583455484089)*x[2]
        arg[2,3]=(1.17437923282)*x[0]+(1.31485601495)*x[1]+(1.3139743231)*x[2]
        arg[2,4]=(0.362851636275)*x[0]+(0.804049783156)*x[1]+(1.09205000214)*x[2]
        arg[3,0]=(0.343603837195)*x[0]+(-1.49082257981)*x[1]+(-0.628668674545)*x[2]
        arg[3,1]=(-0.676847711765)*x[0]+(1.32350560453)*x[1]+(1.07275185686)*x[2]
        arg[3,2]=(0.570784985469)*x[0]+(-1.01212360602)*x[1]+(-0.0808776667606)*x[2]
        arg[3,3]=(0.506392444236)*x[0]+(-1.00848318041)*x[1]+(0.558777901541)*x[2]
        arg[3,4]=(-0.60925738291)*x[0]+(-0.386755957809)*x[1]+(0.60930115434)*x[2]
        ref[0,0]=(-0.695230136419)/(o+1.)+(-0.388400729083)+(0.468139288533)*0.5**o
        ref[0,1]=(0.462098132187)/(o+1.)+(-0.215041893287)+(-0.397587864642)*0.5**o
        ref[0,2]=(-0.262411385363)/(o+1.)+(0.275729226874)+(-0.534991632946)*0.5**o
        ref[0,3]=(0.117822633378)/(o+1.)+(0.0286827897592)+(-0.965914921536)*0.5**o
        ref[0,4]=(0.293575681039)/(o+1.)+(0.584039692827)+(0.809761767299)*0.5**o
        ref[1,0]=(0.627951922463)/(o+1.)+(-0.994889162557)+(0.482016683321)*0.5**o
        ref[1,1]=(1.83482238662)/(o+1.)+(-0.205719558739)+(-0.588157062647)*0.5**o
        ref[1,2]=(1.13612020398)/(o+1.)+(0.0262824938134)+(0.659964625174)*0.5**o
        ref[1,3]=(0.325264487741)/(o+1.)+(-0.0752154813413)+(-0.951265923787)*0.5**o
        ref[1,4]=(-0.149919283302)/(o+1.)+(0.77701502585)+(-0.0476767266789)*0.5**o
        ref[2,0]=(0.491504686171)/(o+1.)+(-0.562534706926)+(0.389422555265)*0.5**o
        ref[2,1]=(-1.52704217417)/(o+1.)+(0.120525042098)+(-0.117691923134)*0.5**o
        ref[2,2]=(-0.779796772543)/(o+1.)+(-0.318864997267)+(-0.486938059201)*0.5**o
        ref[2,3]=(1.53343318682)/(o+1.)+(0.731149467308)+(0.807477449443)*0.5**o
        ref[2,4]=(1.15919207694)/(o+1.)+(0.113921448815)+(0.871916446996)*0.5**o
        ref[3,0]=(-1.75312920944)/(o+1.)+(-0.153798982308)+(0.284839756895)*0.5**o
        ref[3,1]=(1.30985814569)/(o+1.)+(0.438430874237)+(-0.467310144534)*0.5**o
        ref[3,2]=(-0.737480320972)/(o+1.)+(0.0604403094397)+(0.094383414777)*0.5**o
        ref[3,3]=(-0.688723266524)/(o+1.)+(0.43486027018)+(-0.124310108473)*0.5**o
        ref[3,4]=(0.128399262674)/(o+1.)+(-0.372445844122)+(0.229780239192)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.993339371954)*x[0]+(-1.35308128985)*x[1]
        arg[0,0,1]=(-1.07886149134)*x[0]+(-1.38999283206)*x[1]
        arg[0,1,0]=(0.246468026972)*x[0]+(-1.37217464882)*x[1]
        arg[0,1,1]=(0.645672648599)*x[0]+(1.8070624103)*x[1]
        arg[1,0,0]=(-1.02278989388)*x[0]+(0.151634023103)*x[1]
        arg[1,0,1]=(-0.667332860541)*x[0]+(0.312742149193)*x[1]
        arg[1,1,0]=(-0.848647385832)*x[0]+(0.466438038516)*x[1]
        arg[1,1,1]=(-0.185495911372)*x[0]+(-0.137118786891)*x[1]
        arg[2,0,0]=(-0.0250555870731)*x[0]+(0.905553573283)*x[1]
        arg[2,0,1]=(-1.03940852939)*x[0]+(-1.57326016394)*x[1]
        arg[2,1,0]=(1.48834718811)*x[0]+(-0.340889253246)*x[1]
        arg[2,1,1]=(0.875721514739)*x[0]+(-0.310857412212)*x[1]
        arg[3,0,0]=(1.25129508831)*x[0]+(-1.69312043525)*x[1]
        arg[3,0,1]=(1.50124675312)*x[0]+(0.244195748654)*x[1]
        arg[3,1,0]=(0.292060478909)*x[0]+(-0.660005422696)*x[1]
        arg[3,1,1]=(0.791335231308)*x[0]+(0.466481573591)*x[1]
        arg[4,0,0]=(-0.361241947236)*x[0]+(-1.43591648587)*x[1]
        arg[4,0,1]=(-1.10385456695)*x[0]+(-0.325314195825)*x[1]
        arg[4,1,0]=(1.01004173033)*x[0]+(0.426824552857)*x[1]
        arg[4,1,1]=(-0.631382479612)*x[0]+(0.0603792190966)*x[1]
        arg[5,0,0]=(-0.30404368786)*x[0]+(-0.0102498522932)*x[1]
        arg[5,0,1]=(0.109188848746)*x[0]+(-1.16846160105)*x[1]
        arg[5,1,0]=(-0.351916879547)*x[0]+(-0.320714795448)*x[1]
        arg[5,1,1]=(-0.534138562652)*x[0]+(-0.256013589844)*x[1]
        ref[0,0,0]=(-0.506159375456)/(o+1.)+(-0.522895228783)+(-0.794470828783)*0.5**o
        ref[0,0,1]=(-0.930984367068)/(o+1.)+(-0.636851258685)+(-0.264167438968)*0.5**o
        ref[0,1,0]=(-0.674222792172)/(o+1.)+(-0.52131379048)+(0.591143751282)*0.5**o
        ref[0,1,1]=(0.953370751283)/(o+1.)+(0.377773224645)+(0.743817858326)*0.5**o
        ref[1,0,0]=(0.700508215428)/(o+1.)+(-0.303343530867)+(-0.964977024471)*0.5**o
        ref[1,0,1]=(-0.408185116452)/(o+1.)+(0.0668983569743)+(-0.0802023088445)*0.5**o
        ref[1,1,0]=(0.80742345527)/(o+1.)+(-0.405780705683)+(-0.37807139122)*0.5**o
        ref[1,1,1]=(0.0184615568095)/(o+1.)+(0.221921562266)+(-0.784919379605)*0.5**o
        ref[2,0,0]=(0.571906256883)/(o+1.)+(-0.0415588382133)+(0.391709405753)*0.5**o
        ref[2,0,1]=(-0.778176009451)/(o+1.)+(-0.761680990566)+(-0.311130702752)*0.5**o
        ref[2,1,0]=(0.498214686321)/(o+1.)+(-0.146173705483)+(0.941590659509)*0.5**o
        ref[2,1,1]=(-0.273747071818)/(o+1.)+(-0.0319184747153)+(0.902448123776)*0.5**o
        ref[3,0,0]=(-0.730029471178)/(o+1.)+(-0.288672747933)+(0.865549620109)*0.5**o
        ref[3,0,1]=(-0.371409186475)/(o+1.)+(0.623169375401)+(0.870512937445)*0.5**o
        ref[3,1,0]=(0.0611454391867)/(o+1.)+(-0.627723874214)+(0.826357365454)*0.5**o
        ref[3,1,1]=(0.785553447935)/(o+1.)+(0.246940213405)+(-0.021617069845)*0.5**o
        ref[4,0,0]=(-0.936201955907)/(o+1.)+(-0.484823720021)+(0.108690962842)*0.5**o
        ref[4,0,1]=(-0.277830390815)/(o+1.)+(-0.151786547071)+(-0.847765277819)*0.5**o
        ref[4,1,0]=(0.12293099533)/(o+1.)+(0.295620382054)+(0.722694523749)*0.5**o
        ref[4,1,1]=(-0.5517019185)/(o+1.)+(0.0850748937088)+(-0.189451129433)*0.5**o
        ref[5,0,0]=(-0.942533148805)/(o+1.)+(0.425680504872)+(-0.223121401092)*0.5**o
        ref[5,0,1]=(-0.377134228332)/(o+1.)+(-0.721019409569)+(0.759900295162)*0.5**o
        ref[5,1,0]=(-0.643828391942)/(o+1.)+(0.399974456311)+(-0.828752195674)*0.5**o
        ref[5,1,1]=(-0.118293731439)/(o+1.)+(-0.355568544431)+(0.0392786678054)*0.5**o
      else:
        arg[0,0,0]=(0.205446621454)*x[0]+(0.10254878667)*x[1]+(1.27673728979)*x[2]
        arg[0,0,1]=(0.758846313193)*x[0]+(0.63320336713)*x[1]+(-0.42638505807)*x[2]
        arg[0,1,0]=(-0.0127116561603)*x[0]+(-0.360359420027)*x[1]+(0.540508591493)*x[2]
        arg[0,1,1]=(-0.278918629538)*x[0]+(-0.0291534835772)*x[1]+(0.708655731565)*x[2]
        arg[1,0,0]=(-0.612965210483)*x[0]+(0.787787556912)*x[1]+(0.222384448462)*x[2]
        arg[1,0,1]=(1.07409083742)*x[0]+(1.19429173706)*x[1]+(-0.484979976779)*x[2]
        arg[1,1,0]=(-1.25714999157)*x[0]+(-0.321346019622)*x[1]+(0.278329320837)*x[2]
        arg[1,1,1]=(-0.512262689707)*x[0]+(1.6891992053)*x[1]+(-1.02351830068)*x[2]
        arg[2,0,0]=(-1.57911907773)*x[0]+(0.096124258122)*x[1]+(0.336619253521)*x[2]
        arg[2,0,1]=(0.0901144254199)*x[0]+(-1.36818186255)*x[1]+(1.2969609239)*x[2]
        arg[2,1,0]=(0.158329511499)*x[0]+(0.696997185217)*x[1]+(-0.400535235927)*x[2]
        arg[2,1,1]=(-1.44570853201)*x[0]+(0.237359587195)*x[1]+(-1.17323024801)*x[2]
        arg[3,0,0]=(-1.4325812824)*x[0]+(-0.733085265902)*x[1]+(0.125883186484)*x[2]
        arg[3,0,1]=(-0.0529889021435)*x[0]+(0.335154600942)*x[1]+(0.839337375209)*x[2]
        arg[3,1,0]=(-0.423552182963)*x[0]+(1.31813289315)*x[1]+(0.13235854915)*x[2]
        arg[3,1,1]=(0.53278572217)*x[0]+(0.159434073071)*x[1]+(-0.313621199136)*x[2]
        arg[4,0,0]=(0.747779887673)*x[0]+(0.30977483911)*x[1]+(-0.0760295896739)*x[2]
        arg[4,0,1]=(0.00295118909194)*x[0]+(-0.687914867842)*x[1]+(0.0386268422915)*x[2]
        arg[4,1,0]=(-0.276363532754)*x[0]+(0.52102466785)*x[1]+(-0.684840191917)*x[2]
        arg[4,1,1]=(0.0283138898538)*x[0]+(0.116042458294)*x[1]+(0.485578401528)*x[2]
        arg[5,0,0]=(-0.118402060192)*x[0]+(-1.18409809944)*x[1]+(-0.304048852342)*x[2]
        arg[5,0,1]=(0.174393790186)*x[0]+(-0.524912689263)*x[1]+(-0.648708392355)*x[2]
        arg[5,1,0]=(-0.254958756344)*x[0]+(-0.772071348728)*x[1]+(-0.0572494064027)*x[2]
        arg[5,1,1]=(-0.74777072018)*x[0]+(0.19109939364)*x[1]+(0.498973987814)*x[2]
        ref[0,0,0]=(0.986457444152)/(o+1.)+(0.258994989412)+(0.0802852749346)*0.5**o
        ref[0,0,1]=(-0.0286717143415)/(o+1.)+(0.554867138611)+(-0.115397940627)*0.5**o
        ref[0,1,0]=(-0.886482131138)/(o+1.)+(0.464293129749)+(0.125333386946)*0.5**o
        ref[0,1,1]=(0.199214164457)/(o+1.)+(0.336882900218)+(-0.472396346443)*0.5**o
        ref[1,0,0]=(0.36021680954)/(o+1.)+(0.408903924703)+(-0.780817864055)*0.5**o
        ref[1,0,1]=(0.44795581074)/(o+1.)+(0.472159267742)+(0.391128251477)*0.5**o
        ref[1,1,0]=(1.02665734893)/(o+1.)+(-0.71418625728)+(-0.898451524719)*0.5**o
        ref[1,1,1]=(0.0322676607777)/(o+1.)+(0.105378958385)+(-0.0896073626306)*0.5**o
        ref[2,0,0]=(1.15723247637)/(o+1.)+(-0.719604538702)+(-0.86439896505)*0.5**o
        ref[2,0,1]=(-0.254100181576)/(o+1.)+(-0.0387997401585)+(0.35059314866)*0.5**o
        ref[2,1,0]=(-0.636439167982)/(o+1.)+(0.563390681981)+(-0.0355507351918)*0.5**o
        ref[2,1,1]=(-0.265265225142)/(o+1.)+(-0.635620798949)+(-0.845072369786)*0.5**o
        ref[3,0,0]=(-0.062162800164)/(o+1.)+(-0.554499928282)+(-0.868620705087)*0.5**o
        ref[3,0,1]=(0.784681961813)/(o+1.)+(0.355202655131)+(-0.373584198067)*0.5**o
        ref[3,1,0]=(1.20272043429)/(o+1.)+(0.352677694103)+(-0.881136563162)*0.5**o
        ref[3,1,1]=(-0.311152797156)/(o+1.)+(0.313736563953)+(0.0622782653549)*0.5**o
        ref[4,0,0]=(-0.365517620238)/(o+1.)+(0.733502174253)+(-0.119961591158)*0.5**o
        ref[4,0,1]=(0.0155845896619)/(o+1.)+(-0.654963543839)+(0.648005661558)*0.5**o
        ref[4,1,0]=(0.405697975915)/(o+1.)+(-0.637805133954)+(0.429733235172)*0.5**o
        ref[4,1,1]=(-0.0469738326054)/(o+1.)+(0.0590721074184)+(0.558764367445)*0.5**o
        ref[5,0,0]=(-0.333825007954)/(o+1.)+(-0.46448520882)+(-0.343753586377)*0.5**o
        ref[5,0,1]=(-0.469125954795)/(o+1.)+(-0.147192309255)+(-0.235716718127)*0.5**o
        ref[5,1,0]=(-0.573275891777)/(o+1.)+(-0.478328163121)+(0.445652706545)*0.5**o
        ref[5,1,1]=(0.0545515234726)/(o+1.)+(-0.153919681346)+(0.195590500493)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.226142137728)*x[0]+(0.85681895183)*x[1]
        arg[0,0,0,1]=(-0.824407181762)*x[0]+(-0.290023144892)*x[1]
        arg[0,0,1,0]=(-1.25987816017)*x[0]+(0.824340990595)*x[1]
        arg[0,0,1,1]=(-0.0917811606765)*x[0]+(0.660211932964)*x[1]
        arg[0,0,2,0]=(-1.31073149178)*x[0]+(0.482324984918)*x[1]
        arg[0,0,2,1]=(0.256995458227)*x[0]+(0.79244254852)*x[1]
        arg[0,1,0,0]=(-0.0682147475407)*x[0]+(0.243620052061)*x[1]
        arg[0,1,0,1]=(0.071099982687)*x[0]+(-0.62986754897)*x[1]
        arg[0,1,1,0]=(-0.0716793733131)*x[0]+(-1.15235525814)*x[1]
        arg[0,1,1,1]=(0.193107575216)*x[0]+(-0.596149957543)*x[1]
        arg[0,1,2,0]=(-0.160142629886)*x[0]+(1.36581707067)*x[1]
        arg[0,1,2,1]=(-1.21477648376)*x[0]+(-0.732000298369)*x[1]
        arg[0,2,0,0]=(-1.02468971059)*x[0]+(0.330721595217)*x[1]
        arg[0,2,0,1]=(1.11414907391)*x[0]+(1.29386127812)*x[1]
        arg[0,2,1,0]=(1.1485850154)*x[0]+(0.387585977495)*x[1]
        arg[0,2,1,1]=(0.237382583072)*x[0]+(1.56248221279)*x[1]
        arg[0,2,2,0]=(-0.949129049229)*x[0]+(0.578823379063)*x[1]
        arg[0,2,2,1]=(-1.0667182537)*x[0]+(0.462227485959)*x[1]
        arg[0,3,0,0]=(1.37954377554)*x[0]+(0.541856880139)*x[1]
        arg[0,3,0,1]=(0.401584700319)*x[0]+(0.577102327147)*x[1]
        arg[0,3,1,0]=(-0.289796408108)*x[0]+(0.807853785325)*x[1]
        arg[0,3,1,1]=(0.0038467861822)*x[0]+(-0.931311510771)*x[1]
        arg[0,3,2,0]=(0.859036254821)*x[0]+(0.198884218516)*x[1]
        arg[0,3,2,1]=(-0.00310256070968)*x[0]+(0.511838475968)*x[1]
        arg[0,4,0,0]=(0.539686886802)*x[0]+(-0.324612924267)*x[1]
        arg[0,4,0,1]=(-0.0458655898565)*x[0]+(0.211810380966)*x[1]
        arg[0,4,1,0]=(0.302333533003)*x[0]+(0.0663033316831)*x[1]
        arg[0,4,1,1]=(-0.981519161639)*x[0]+(-0.00207756909914)*x[1]
        arg[0,4,2,0]=(0.283659447004)*x[0]+(-0.231526592979)*x[1]
        arg[0,4,2,1]=(-1.17679167875)*x[0]+(0.285698985756)*x[1]
        arg[1,0,0,0]=(-0.162900059582)*x[0]+(-0.877164340783)*x[1]
        arg[1,0,0,1]=(-0.50468605065)*x[0]+(-0.160440073709)*x[1]
        arg[1,0,1,0]=(-0.100901676644)*x[0]+(-0.296111526029)*x[1]
        arg[1,0,1,1]=(0.176107421869)*x[0]+(1.74393708745)*x[1]
        arg[1,0,2,0]=(-0.979807721703)*x[0]+(-0.229400841067)*x[1]
        arg[1,0,2,1]=(0.217472795645)*x[0]+(0.0881503243323)*x[1]
        arg[1,1,0,0]=(-0.127980123625)*x[0]+(0.0932095456541)*x[1]
        arg[1,1,0,1]=(0.829674410931)*x[0]+(-0.796196041601)*x[1]
        arg[1,1,1,0]=(1.28656110233)*x[0]+(-1.27850471392)*x[1]
        arg[1,1,1,1]=(1.159269884)*x[0]+(-0.973135433631)*x[1]
        arg[1,1,2,0]=(1.46274837856)*x[0]+(-1.81554270863)*x[1]
        arg[1,1,2,1]=(0.0973472511688)*x[0]+(0.551451261495)*x[1]
        arg[1,2,0,0]=(0.510667671293)*x[0]+(-0.51075982099)*x[1]
        arg[1,2,0,1]=(-1.41612093451)*x[0]+(0.614417669629)*x[1]
        arg[1,2,1,0]=(1.46595125796)*x[0]+(0.60374279122)*x[1]
        arg[1,2,1,1]=(-0.317185218187)*x[0]+(-0.0785532882696)*x[1]
        arg[1,2,2,0]=(-0.388504041993)*x[0]+(-0.838003398775)*x[1]
        arg[1,2,2,1]=(0.0497207129012)*x[0]+(-0.622155692247)*x[1]
        arg[1,3,0,0]=(-0.0471400179563)*x[0]+(1.3378948207)*x[1]
        arg[1,3,0,1]=(-1.22250822195)*x[0]+(0.629714783138)*x[1]
        arg[1,3,1,0]=(-1.28801080408)*x[0]+(-1.54079710038)*x[1]
        arg[1,3,1,1]=(-0.426412448469)*x[0]+(-1.09410474021)*x[1]
        arg[1,3,2,0]=(0.022834167351)*x[0]+(-0.497423389788)*x[1]
        arg[1,3,2,1]=(-0.787492124301)*x[0]+(-0.245621168419)*x[1]
        arg[1,4,0,0]=(-1.37814880919)*x[0]+(-0.704805848451)*x[1]
        arg[1,4,0,1]=(-1.12494770055)*x[0]+(1.44920024305)*x[1]
        arg[1,4,1,0]=(-0.809532572761)*x[0]+(1.16638587657)*x[1]
        arg[1,4,1,1]=(0.931784486816)*x[0]+(1.34629889453)*x[1]
        arg[1,4,2,0]=(-0.172430942793)*x[0]+(-0.584736998321)*x[1]
        arg[1,4,2,1]=(0.00542948736389)*x[0]+(0.43917597284)*x[1]
        arg[2,0,0,0]=(-0.946122337406)*x[0]+(-0.161940168137)*x[1]
        arg[2,0,0,1]=(-0.7856653779)*x[0]+(-0.75287701788)*x[1]
        arg[2,0,1,0]=(-1.08858367592)*x[0]+(-0.530380818694)*x[1]
        arg[2,0,1,1]=(-1.0573555095)*x[0]+(-0.854148765785)*x[1]
        arg[2,0,2,0]=(0.723412825328)*x[0]+(0.326761950437)*x[1]
        arg[2,0,2,1]=(-0.0143142161225)*x[0]+(-0.752724311876)*x[1]
        arg[2,1,0,0]=(-0.942548092355)*x[0]+(0.700861094897)*x[1]
        arg[2,1,0,1]=(-0.712286567957)*x[0]+(0.230406545738)*x[1]
        arg[2,1,1,0]=(1.27414839774)*x[0]+(0.724784649037)*x[1]
        arg[2,1,1,1]=(0.637494163002)*x[0]+(0.442489599192)*x[1]
        arg[2,1,2,0]=(1.81617631726)*x[0]+(1.56870218809)*x[1]
        arg[2,1,2,1]=(1.16419900666)*x[0]+(-0.044314729141)*x[1]
        arg[2,2,0,0]=(0.880267125196)*x[0]+(0.77124406259)*x[1]
        arg[2,2,0,1]=(0.482402867542)*x[0]+(0.0746924305017)*x[1]
        arg[2,2,1,0]=(0.105861259696)*x[0]+(0.484292207267)*x[1]
        arg[2,2,1,1]=(-0.107785838196)*x[0]+(0.701132243209)*x[1]
        arg[2,2,2,0]=(0.559717405844)*x[0]+(-1.5585942717)*x[1]
        arg[2,2,2,1]=(-0.989130712605)*x[0]+(0.221361137556)*x[1]
        arg[2,3,0,0]=(-0.045471818095)*x[0]+(-0.437941959792)*x[1]
        arg[2,3,0,1]=(1.25840468379)*x[0]+(-1.18872011057)*x[1]
        arg[2,3,1,0]=(1.70665829408)*x[0]+(0.705273921003)*x[1]
        arg[2,3,1,1]=(-0.0928709253845)*x[0]+(0.785115449756)*x[1]
        arg[2,3,2,0]=(-0.887243418838)*x[0]+(1.06790660642)*x[1]
        arg[2,3,2,1]=(-0.572644686872)*x[0]+(-0.672831948394)*x[1]
        arg[2,4,0,0]=(1.28282566801)*x[0]+(0.651985199447)*x[1]
        arg[2,4,0,1]=(0.0965663229335)*x[0]+(1.33819998052)*x[1]
        arg[2,4,1,0]=(0.205569041098)*x[0]+(-0.873095358733)*x[1]
        arg[2,4,1,1]=(-1.04452025745)*x[0]+(-0.204917033907)*x[1]
        arg[2,4,2,0]=(-0.181163938972)*x[0]+(0.0341993409272)*x[1]
        arg[2,4,2,1]=(1.68327027059)*x[0]+(-0.289436355642)*x[1]
        arg[3,0,0,0]=(1.05712810731)*x[0]+(0.915711819735)*x[1]
        arg[3,0,0,1]=(-0.171072039482)*x[0]+(-0.160656943623)*x[1]
        arg[3,0,1,0]=(-1.27063799869)*x[0]+(-0.605131642813)*x[1]
        arg[3,0,1,1]=(0.940512990835)*x[0]+(-0.116675577531)*x[1]
        arg[3,0,2,0]=(-0.194213869952)*x[0]+(0.364066752025)*x[1]
        arg[3,0,2,1]=(-0.328780465397)*x[0]+(-0.114689413673)*x[1]
        arg[3,1,0,0]=(-0.825914305895)*x[0]+(-0.167010056731)*x[1]
        arg[3,1,0,1]=(0.657779056207)*x[0]+(0.0666252451638)*x[1]
        arg[3,1,1,0]=(-0.464004426036)*x[0]+(-0.340932918409)*x[1]
        arg[3,1,1,1]=(-1.6698652667)*x[0]+(0.731870373141)*x[1]
        arg[3,1,2,0]=(1.39626250779)*x[0]+(-0.879152855557)*x[1]
        arg[3,1,2,1]=(1.34973919508)*x[0]+(-0.129753332761)*x[1]
        arg[3,2,0,0]=(1.03443450706)*x[0]+(0.807333477811)*x[1]
        arg[3,2,0,1]=(0.366011302596)*x[0]+(0.543804800926)*x[1]
        arg[3,2,1,0]=(0.262446846036)*x[0]+(-0.247786850991)*x[1]
        arg[3,2,1,1]=(-1.04657091196)*x[0]+(-0.485443426155)*x[1]
        arg[3,2,2,0]=(0.342434327538)*x[0]+(-0.148503585588)*x[1]
        arg[3,2,2,1]=(-0.189833429716)*x[0]+(0.424391874819)*x[1]
        arg[3,3,0,0]=(-0.118348838118)*x[0]+(0.397628264341)*x[1]
        arg[3,3,0,1]=(0.481440510084)*x[0]+(-0.393576221424)*x[1]
        arg[3,3,1,0]=(0.907533552208)*x[0]+(-0.668646549376)*x[1]
        arg[3,3,1,1]=(-1.49731714213)*x[0]+(-1.18710678264)*x[1]
        arg[3,3,2,0]=(-0.460660079207)*x[0]+(1.48078266444)*x[1]
        arg[3,3,2,1]=(-0.161073556988)*x[0]+(-0.486827042754)*x[1]
        arg[3,4,0,0]=(-0.668755455623)*x[0]+(0.345804059434)*x[1]
        arg[3,4,0,1]=(1.00528889193)*x[0]+(0.978507928191)*x[1]
        arg[3,4,1,0]=(0.98955856331)*x[0]+(1.0421852144)*x[1]
        arg[3,4,1,1]=(0.711688260873)*x[0]+(1.39962222322)*x[1]
        arg[3,4,2,0]=(1.08512429209)*x[0]+(-0.555249097743)*x[1]
        arg[3,4,2,1]=(-1.10443784007)*x[0]+(-0.782274429169)*x[1]
        ref[0,0,0,0]=(0.151153165564)/(o+1.)+(0.112609656739)+(0.254304335059)*0.5**o
        ref[0,0,0,1]=(-0.530602583511)/(o+1.)+(0.1817855378)+(-0.947398818743)*0.5**o
        ref[0,0,1,0]=(0.308747437738)/(o+1.)+(-0.0338129255643)+(-0.676658756184)*0.5**o
        ref[0,0,1,1]=(0.321026553427)/(o+1.)+(0.0692705367618)+(0.108863145337)*0.5**o
        ref[0,0,2,0]=(-0.0908034666655)/(o+1.)+(-0.1100205155)+(-0.517562009199)*0.5**o
        ref[0,0,2,1]=(0.303456040903)/(o+1.)+(0.500090875336)+(-0.254199784829)*0.5**o
        ref[0,1,0,0]=(-0.440641221831)/(o+1.)+(0.456218946873)+(-0.296391367396)*0.5**o
        ref[0,1,0,1]=(0.222628561762)/(o+1.)+(-0.162064412854)+(-0.457267302337)*0.5**o
        ref[0,1,1,0]=(-0.913361887408)/(o+1.)+(-0.356168674239)+(0.401664604431)*0.5**o
        ref[0,1,1,1]=(-0.543116639248)/(o+1.)+(-0.120902439422)+(0.381879135767)*0.5**o
        ref[0,1,2,0]=(0.993649262831)/(o+1.)+(0.384695858206)+(-0.55736653846)*0.5**o
        ref[0,1,2,1]=(-0.968077797451)/(o+1.)+(-0.187765942637)+(-0.603167099406)*0.5**o
        ref[0,2,0,0]=(0.873662019918)/(o+1.)+(-0.710635360033)+(-0.146359415223)*0.5**o
        ref[0,2,0,1]=(0.466408230693)/(o+1.)+(0.77626565285)+(0.389070815631)*0.5**o
        ref[0,2,1,0]=(-0.475010813579)/(o+1.)+(0.525435880006)+(0.960310046461)*0.5**o
        ref[0,2,1,1]=(0.697127520484)/(o+1.)+(0.357204276552)+(0.388328722275)*0.5**o
        ref[0,2,2,0]=(-0.108434738487)/(o+1.)+(0.356443300466)+(-0.97475753261)*0.5**o
        ref[0,2,2,1]=(0.952274964304)/(o+1.)+(-0.497379508976)+(-0.562006714089)*0.5**o
        ref[0,3,0,0]=(-0.0368012761671)/(o+1.)+(0.594000625518)+(0.770200680808)*0.5**o
        ref[0,3,0,1]=(0.450361861836)/(o+1.)+(0.182031052848)+(0.164263059934)*0.5**o
        ref[0,3,1,0]=(0.500714741956)/(o+1.)+(0.0479711343168)+(-0.0785996333735)*0.5**o
        ref[0,3,1,1]=(-0.949280938645)/(o+1.)+(0.452441874504)+(-0.88306753495)*0.5**o
        ref[0,3,2,0]=(0.186931712837)/(o+1.)+(0.222335175903)+(0.426318408694)*0.5**o
        ref[0,3,2,1]=(-0.288015428918)/(o+1.)+(0.80948657963)+(-0.822221815085)*0.5**o
        ref[0,4,0,0]=(0.670151184695)/(o+1.)+(-0.415883847013)+(0.376690471866)*0.5**o
        ref[0,4,0,1]=(-0.107106843992)/(o+1.)+(0.555375298435)+(-0.837698961768)*0.5**o
        ref[0,4,1,0]=(-0.299999096802)/(o+1.)+(0.668996782774)+(-0.669357604061)*0.5**o
        ref[0,4,1,1]=(0.778006183401)/(o+1.)+(-0.614473529032)+(-0.532655856076)*0.5**o
        ref[0,4,2,0]=(-0.250068389818)/(o+1.)+(0.156274314197)+(-0.0103473845504)*0.5**o
        ref[0,4,2,1]=(0.902125322216)/(o+1.)+(-0.616424567683)+(-0.560368879846)*0.5**o
        ref[1,0,0,0]=(-0.133690258295)/(o+1.)+(-0.715949360249)+(0.525524578427)*0.5**o
        ref[1,0,0,1]=(-0.788097689216)/(o+1.)+(0.0998405659256)+(-0.076709566994)*0.5**o
        ref[1,0,1,0]=(0.144280886187)/(o+1.)+(0.0775377290314)+(-0.696369546924)*0.5**o
        ref[1,0,1,1]=(0.824172196605)/(o+1.)+(0.561804029444)+(-0.0277357461773)*0.5**o
        ref[1,0,2,0]=(0.422564421802)/(o+1.)+(-0.60719307099)+(-0.417386842591)*0.5**o
        ref[1,0,2,1]=(0.778908672762)/(o+1.)+(0.0163446340765)+(-0.505974820938)*0.5**o
        ref[1,1,0,0]=(0.0574071047387)/(o+1.)+(-0.350514214742)+(0.608850746775)*0.5**o
        ref[1,1,0,1]=(-0.991660961792)/(o+1.)+(0.322084279455)+(0.380970772212)*0.5**o
        ref[1,1,1,0]=(-0.823761120268)/(o+1.)+(-0.0495419092898)+(0.930901327253)*0.5**o
        ref[1,1,1,1]=(-0.384729980569)/(o+1.)+(-0.0392583789163)+(0.649381188769)*0.5**o
        ref[1,1,2,0]=(-0.840365304069)/(o+1.)+(-0.0150512662777)+(0.517673506558)*0.5**o
        ref[1,1,2,1]=(0.555976660846)/(o+1.)+(-0.307265886898)+(0.707353625614)*0.5**o
        ref[1,2,0,0]=(-0.423056705491)/(o+1.)+(0.0559404393793)+(0.311083677036)*0.5**o
        ref[1,2,0,1]=(0.15409797664)/(o+1.)+(-0.0906637067519)+(-0.774473828018)*0.5**o
        ref[1,2,1,0]=(0.796718644088)/(o+1.)+(0.248012077098)+(0.776951250898)*0.5**o
        ref[1,2,1,1]=(-0.342830889263)/(o+1.)+(-0.282162737841)+(0.511417858487)*0.5**o
        ref[1,2,2,0]=(-0.722403016671)/(o+1.)+(-0.367059457954)+(0.230014491812)*0.5**o
        ref[1,2,2,1]=(-0.773685058881)/(o+1.)+(0.412807900451)+(-0.624365721366)*0.5**o
        ref[1,3,0,0]=(0.527481434758)/(o+1.)+(0.36602961117)+(0.031214145649)*0.5**o
        ref[1,3,0,1]=(0.0630927531585)/(o+1.)+(-0.164989701995)+(-0.325906787975)*0.5**o
        ref[1,3,1,0]=(-0.612390590593)/(o+1.)+(-0.78855996274)+(-0.63929738839)*0.5**o
        ref[1,3,1,1]=(-0.962656044265)/(o+1.)+(-0.00576535037253)+(-0.546330443673)*0.5**o
        ref[1,3,2,0]=(-0.947280922008)/(o+1.)+(-0.137449579996)+(0.747590859563)*0.5**o
        ref[1,3,2,1]=(-0.946623026202)/(o+1.)+(0.0787536550213)+(-0.24399757656)*0.5**o
        ref[1,4,0,0]=(-0.174471016943)/(o+1.)+(-0.636500556039)+(-0.635482528623)*0.5**o
        ref[1,4,0,1]=(0.806379391595)/(o+1.)+(-0.00304106577093)+(-0.476044717545)*0.5**o
        ref[1,4,1,0]=(0.729649125134)/(o+1.)+(0.0151477051616)+(-0.403091231646)*0.5**o
        ref[1,4,1,1]=(0.559976454745)/(o+1.)+(0.479295896421)+(0.75951513376)*0.5**o
        ref[1,4,2,0]=(0.272199270818)/(o+1.)+(-0.0929072274332)+(-0.843552757065)*0.5**o
        ref[1,4,2,1]=(-0.552663678377)/(o+1.)+(0.613784745778)+(-0.230300352975)*0.5**o
        ref[2,0,0,0]=(-0.174247908098)/(o+1.)+(-0.108873783282)+(-0.716067030882)*0.5**o
        ref[2,0,0,1]=(-0.300584841532)/(o+1.)+(-0.485363665493)+(-0.267230223262)*0.5**o
        ref[2,0,1,0]=(0.316261805154)/(o+1.)+(-0.523474106451)+(-0.888278086869)*0.5**o
        ref[2,0,1,1]=(-0.699622562861)/(o+1.)+(-0.151089538248)+(-0.909702635929)*0.5**o
        ref[2,0,2,0]=(0.728545639829)/(o+1.)+(0.192083148877)+(-0.0625371618174)*0.5**o
        ref[2,0,2,1]=(-0.22042267239)/(o+1.)+(-0.611974599324)+(0.677333343041)*0.5**o
        ref[2,1,0,0]=(-0.0734883811113)/(o+1.)+(0.36100831022)+(-0.890215236787)*0.5**o
        ref[2,1,0,1]=(0.121656829057)/(o+1.)+(-0.130801635799)+(-0.341933579678)*0.5**o
        ref[2,1,1,0]=(0.221353168629)/(o+1.)+(0.618635673483)+(0.540308531179)*0.5**o
        ref[2,1,1,1]=(0.0152580531273)/(o+1.)+(0.252173415542)+(0.560378877984)*0.5**o
        ref[2,1,2,0]=(0.580319777731)/(o+1.)+(0.90566612137)+(0.993226484879)*0.5**o
        ref[2,1,2,1]=(-0.710250320949)/(o+1.)+(0.742177021871)+(0.345780554724)*0.5**o
        ref[2,2,0,0]=(0.770701563507)/(o+1.)+(0.240823471881)+(0.399162680517)*0.5**o
        ref[2,2,0,1]=(0.76121761761)/(o+1.)+(0.128382656738)+(-0.460887633042)*0.5**o
        ref[2,2,1,0]=(-0.269854962057)/(o+1.)+(0.690854898294)+(-0.521701367567)*0.5**o
        ref[2,2,1,1]=(0.719518595712)/(o+1.)+(0.029596430244)+(-0.185365051187)*0.5**o
        ref[2,2,2,0]=(-0.671157761165)/(o+1.)+(-0.142688285012)+(-0.0423425346644)*0.5**o
        ref[2,2,2,1]=(-0.541156097112)/(o+1.)+(0.0765180918929)+(-0.379649661722)*0.5**o
        ref[2,3,0,0]=(-0.0604347109212)/(o+1.)+(-0.19403702073)+(-0.034905025507)*0.5**o
        ref[2,3,0,1]=(-0.793398507071)/(o+1.)+(0.255681280626)+(0.351720519037)*0.5**o
        ref[2,3,1,0]=(-0.267058413731)/(o+1.)+(0.852211426049)+(0.974567776717)*0.5**o
        ref[2,3,1,1]=(0.228396641207)/(o+1.)+(0.322922896735)+(-0.181997910305)*0.5**o
        ref[2,3,2,0]=(0.480987773937)/(o+1.)+(-0.197709028569)+(0.0950934707835)*0.5**o
        ref[2,3,2,1]=(-0.648934475027)/(o+1.)+(-0.418383297875)+(0.240224435511)*0.5**o
        ref[2,4,0,0]=(0.430020261528)/(o+1.)+(0.53421018733)+(0.436370231269)*0.5**o
        ref[2,4,0,1]=(0.906390660223)/(o+1.)+(-0.056057545856)+(0.640490734948)*0.5**o
        ref[2,4,1,0]=(-0.732099770943)/(o+1.)+(-0.0485027617732)+(0.161578976854)*0.5**o
        ref[2,4,1,1]=(-0.827495680848)/(o+1.)+(0.285044728936)+(-0.992031068384)*0.5**o
        ref[2,4,2,0]=(-0.919905471787)/(o+1.)+(0.656971386264)+(-0.541001898785)*0.5**o
        ref[2,4,2,1]=(0.161975875302)/(o+1.)+(0.272665206557)+(0.686527626533)*0.5**o
        ref[3,0,0,0]=(0.584120963259)/(o+1.)+(0.207930729804)+(0.972857504177)*0.5**o
        ref[3,0,0,1]=(0.315823141491)/(o+1.)+(-0.217561208386)+(-0.212429707823)*0.5**o
        ref[3,0,1,0]=(-0.732294996139)/(o+1.)+(-0.278653143596)+(-0.586168358175)*0.5**o
        ref[3,0,1,1]=(0.481321847225)/(o+1.)+(-0.070080648792)+(0.482676863662)*0.5**o
        ref[3,0,2,0]=(0.616679859076)/(o+1.)+(-0.418784546677)+(0.390742116352)*0.5**o
        ref[3,0,2,1]=(0.866778417448)/(o+1.)+(-0.686585656643)+(0.0629230167692)*0.5**o
        ref[3,1,0,0]=(0.531890054167)/(o+1.)+(-0.516187607633)+(-0.492439201527)*0.5**o
        ref[3,1,0,1]=(-0.512361639413)/(o+1.)+(0.499557182545)+(0.237651575694)*0.5**o
        ref[3,1,1,0]=(-0.67542965405)/(o+1.)+(0.274414249559)+(-0.678336189514)*0.5**o
        ref[3,1,1,1]=(0.371936242507)/(o+1.)+(-0.1636482466)+(-0.982634642862)*0.5**o
        ref[3,1,2,0]=(-0.799408831501)/(o+1.)+(0.458633399549)+(0.399251684639)*0.5**o
        ref[3,1,2,1]=(-0.0145253903686)/(o+1.)+(0.238472345905)+(0.757566560882)*0.5**o
        ref[3,2,0,0]=(-0.130519473112)/(o+1.)+(0.695277604837)+(0.581732248306)*0.5**o
        ref[3,2,0,1]=(0.632010536438)/(o+1.)+(0.123066205529)+(0.0316731560259)*0.5**o
        ref[3,2,1,0]=(-0.709926210356)/(o+1.)+(0.0926699972302)+(0.53924621094)*0.5**o
        ref[3,2,1,1]=(0.428094002247)/(o+1.)+(-0.841997926131)+(-0.276112488096)*0.5**o
        ref[3,2,2,0]=(0.708481148086)/(o+1.)+(-0.50498161534)+(0.495412824545)*0.5**o
        ref[3,2,2,1]=(0.782824622703)/(o+1.)+(-0.151509698687)+(-0.245246780227)*0.5**o
        ref[3,3,0,0]=(0.848460868418)/(o+1.)+(-0.104639248461)+(-0.359902945273)*0.5**o
        ref[3,3,0,1]=(0.0569151307298)/(o+1.)+(-0.0674224535469)+(0.165794065024)*0.5**o
        ref[3,3,1,0]=(-0.165683739908)/(o+1.)+(-0.183388699236)+(0.771348141212)*0.5**o
        ref[3,3,1,1]=(-0.932550058069)/(o+1.)+(-0.47828109001)+(-0.795311686688)*0.5**o
        ref[3,3,2,0]=(0.821542940923)/(o+1.)+(0.166853194487)+(-0.13512674466)*0.5**o
        ref[3,3,2,1]=(-0.46375871628)/(o+1.)+(0.132701267732)+(-0.449544418926)*0.5**o
        ref[3,4,0,0]=(-0.074530542168)/(o+1.)+(-0.0567877963797)+(-0.134845261261)*0.5**o
        ref[3,4,0,1]=(0.238671520479)/(o+1.)+(0.50689253572)+(0.731340228204)*0.5**o
        ref[3,4,1,0]=(0.981601513335)/(o+1.)+(0.400717209822)+(0.248707844726)*0.5**o
        ref[3,4,1,1]=(0.592304390107)/(o+1.)+(0.342408637328)+(0.834188819333)*0.5**o
        ref[3,4,2,0]=(-0.339991330909)/(o+1.)+(-0.0145922376156)+(0.899051000485)*0.5**o
        ref[3,4,2,1]=(-0.626610688164)/(o+1.)+(-0.417008165115)+(-0.426085250846)*0.5**o
      else:
        arg[0,0,0,0]=(0.697666818785)*x[0]+(1.00202295791)*x[1]+(0.617700157184)*x[2]
        arg[0,0,0,1]=(-0.465577819457)*x[0]+(-1.40054693355)*x[1]+(1.16912653295)*x[2]
        arg[0,0,1,0]=(-1.26696500634)*x[0]+(-0.0469756169975)*x[1]+(-0.445591738991)*x[2]
        arg[0,0,1,1]=(-1.61654321709)*x[0]+(-1.69522962467)*x[1]+(-1.06089094851)*x[2]
        arg[0,0,2,0]=(0.229468378506)*x[0]+(0.716813596991)*x[1]+(-0.237703937029)*x[2]
        arg[0,0,2,1]=(-1.12052051905)*x[0]+(0.835000004618)*x[1]+(-0.853004812404)*x[2]
        arg[0,1,0,0]=(-1.75500745705)*x[0]+(0.655055836882)*x[1]+(1.05207750735)*x[2]
        arg[0,1,0,1]=(-0.0746024824253)*x[0]+(-0.421596816958)*x[1]+(-0.375715855989)*x[2]
        arg[0,1,1,0]=(-0.963269684396)*x[0]+(0.431993248836)*x[1]+(-1.63061928649)*x[2]
        arg[0,1,1,1]=(-0.418757265018)*x[0]+(-0.665626168583)*x[1]+(-0.335734432507)*x[2]
        arg[0,1,2,0]=(0.980346404307)*x[0]+(1.45442049802)*x[1]+(-0.833420308612)*x[2]
        arg[0,1,2,1]=(-1.52536017847)*x[0]+(0.99484028681)*x[1]+(0.109754321663)*x[2]
        arg[0,2,0,0]=(1.697179262)*x[0]+(-0.928732802147)*x[1]+(-0.653379513626)*x[2]
        arg[0,2,0,1]=(0.682698902948)*x[0]+(0.829307959583)*x[1]+(-1.0300112276)*x[2]
        arg[0,2,1,0]=(0.026084323266)*x[0]+(1.28865784608)*x[1]+(0.830446580645)*x[2]
        arg[0,2,1,1]=(0.486291506032)*x[0]+(-0.519637742913)*x[1]+(0.81671330629)*x[2]
        arg[0,2,2,0]=(-0.761591624634)*x[0]+(0.370951131011)*x[1]+(-0.256748479312)*x[2]
        arg[0,2,2,1]=(-0.249974326513)*x[0]+(0.267323484635)*x[1]+(-1.09642130905)*x[2]
        arg[0,3,0,0]=(0.752057563255)*x[0]+(-1.2772262175)*x[1]+(-0.264569668983)*x[2]
        arg[0,3,0,1]=(-0.173167796447)*x[0]+(0.31873208356)*x[1]+(-0.599780701112)*x[2]
        arg[0,3,1,0]=(-0.912464014941)*x[0]+(-1.56452517166)*x[1]+(1.37651660499)*x[2]
        arg[0,3,1,1]=(0.302009667854)*x[0]+(0.524213064769)*x[1]+(0.316301522683)*x[2]
        arg[0,3,2,0]=(-1.2940154673)*x[0]+(-0.900061396992)*x[1]+(0.3860029847)*x[2]
        arg[0,3,2,1]=(-0.97354131341)*x[0]+(1.08221883964)*x[1]+(-0.739323076416)*x[2]
        arg[0,4,0,0]=(1.86535317192)*x[0]+(-0.478359589546)*x[1]+(-0.798014592302)*x[2]
        arg[0,4,0,1]=(-1.08518693851)*x[0]+(-0.186048654536)*x[1]+(1.04915573267)*x[2]
        arg[0,4,1,0]=(-1.09692811055)*x[0]+(1.01042920935)*x[1]+(-0.73978089555)*x[2]
        arg[0,4,1,1]=(0.654656964663)*x[0]+(0.279204596126)*x[1]+(0.886217781361)*x[2]
        arg[0,4,2,0]=(-0.54009886291)*x[0]+(0.378037248739)*x[1]+(1.09785568886)*x[2]
        arg[0,4,2,1]=(-1.43995958913)*x[0]+(-0.720157090994)*x[1]+(0.0676362028136)*x[2]
        arg[1,0,0,0]=(1.47735412608)*x[0]+(0.342355546228)*x[1]+(-0.0134502415508)*x[2]
        arg[1,0,0,1]=(-0.419536590369)*x[0]+(-1.61108470511)*x[1]+(1.51281979355)*x[2]
        arg[1,0,1,0]=(1.31765698245)*x[0]+(-0.7499566294)*x[1]+(0.934537503478)*x[2]
        arg[1,0,1,1]=(0.688467914883)*x[0]+(-1.32681280673)*x[1]+(-0.387405705058)*x[2]
        arg[1,0,2,0]=(1.37069587556)*x[0]+(0.540830111773)*x[1]+(0.477311068263)*x[2]
        arg[1,0,2,1]=(-0.572066213761)*x[0]+(-1.11950925494)*x[1]+(1.07939217872)*x[2]
        arg[1,1,0,0]=(0.564997105326)*x[0]+(-0.73130546389)*x[1]+(0.435910937304)*x[2]
        arg[1,1,0,1]=(0.513931739674)*x[0]+(-0.944043404826)*x[1]+(0.014114227468)*x[2]
        arg[1,1,1,0]=(-0.77850597256)*x[0]+(-0.445835136484)*x[1]+(0.157478357479)*x[2]
        arg[1,1,1,1]=(0.752431176866)*x[0]+(0.659759664199)*x[1]+(-0.859515433521)*x[2]
        arg[1,1,2,0]=(0.0693158730911)*x[0]+(-0.820997827204)*x[1]+(-0.714694436829)*x[2]
        arg[1,1,2,1]=(0.604674152847)*x[0]+(0.647261183132)*x[1]+(-1.21981352028)*x[2]
        arg[1,2,0,0]=(0.0997504695984)*x[0]+(-0.271549839083)*x[1]+(-0.573467835942)*x[2]
        arg[1,2,0,1]=(0.01861845581)*x[0]+(0.157410502406)*x[1]+(0.939257868776)*x[2]
        arg[1,2,1,0]=(1.44909894631)*x[0]+(0.116038517547)*x[1]+(-0.851940489233)*x[2]
        arg[1,2,1,1]=(1.34598841007)*x[0]+(1.79163125841)*x[1]+(0.719171044245)*x[2]
        arg[1,2,2,0]=(-0.214737791448)*x[0]+(-1.85810839095)*x[1]+(-0.240776599927)*x[2]
        arg[1,2,2,1]=(-0.574707645928)*x[0]+(-0.664173831956)*x[1]+(0.310262106696)*x[2]
        arg[1,3,0,0]=(0.296213453592)*x[0]+(-0.831112482555)*x[1]+(0.130075739128)*x[2]
        arg[1,3,0,1]=(-0.369510604457)*x[0]+(0.0373909697998)*x[1]+(1.6689129783)*x[2]
        arg[1,3,1,0]=(0.0922435881419)*x[0]+(-0.280519827004)*x[1]+(0.285933839232)*x[2]
        arg[1,3,1,1]=(-0.649253282272)*x[0]+(-0.460065044123)*x[1]+(-0.0711164948063)*x[2]
        arg[1,3,2,0]=(0.793313307581)*x[0]+(-0.0587956351149)*x[1]+(0.810487259739)*x[2]
        arg[1,3,2,1]=(-1.38668597055)*x[0]+(1.51396934198)*x[1]+(-1.13918550255)*x[2]
        arg[1,4,0,0]=(1.40078167737)*x[0]+(-1.61915436345)*x[1]+(0.421710091214)*x[2]
        arg[1,4,0,1]=(-1.37982015877)*x[0]+(1.30607371738)*x[1]+(-0.608888848419)*x[2]
        arg[1,4,1,0]=(-0.558087802805)*x[0]+(-0.214169499512)*x[1]+(-0.766242546503)*x[2]
        arg[1,4,1,1]=(-1.75503436584)*x[0]+(0.183604705451)*x[1]+(-0.148071484576)*x[2]
        arg[1,4,2,0]=(0.537942069142)*x[0]+(-0.00640749928252)*x[1]+(-0.277544629657)*x[2]
        arg[1,4,2,1]=(-0.652645833371)*x[0]+(-1.33439077026)*x[1]+(-0.226308197201)*x[2]
        arg[2,0,0,0]=(1.45646289279)*x[0]+(1.40842127675)*x[1]+(1.75078281495)*x[2]
        arg[2,0,0,1]=(0.867988651995)*x[0]+(1.35061812241)*x[1]+(-1.56520776323)*x[2]
        arg[2,0,1,0]=(0.549686523522)*x[0]+(-0.528085465813)*x[1]+(-0.994978820267)*x[2]
        arg[2,0,1,1]=(0.383595996552)*x[0]+(0.612989859087)*x[1]+(-0.560401247231)*x[2]
        arg[2,0,2,0]=(-0.341259919499)*x[0]+(-0.125491675042)*x[1]+(0.545572340757)*x[2]
        arg[2,0,2,1]=(-0.414269395434)*x[0]+(-0.433969881898)*x[1]+(-1.01820199328)*x[2]
        arg[2,1,0,0]=(-0.113575280389)*x[0]+(0.326916039942)*x[1]+(-0.576072890249)*x[2]
        arg[2,1,0,1]=(-0.296473969661)*x[0]+(0.195675099431)*x[1]+(-1.72378150831)*x[2]
        arg[2,1,1,0]=(-1.31575932758)*x[0]+(-0.122557423522)*x[1]+(0.53435027695)*x[2]
        arg[2,1,1,1]=(0.299435754608)*x[0]+(0.727140413706)*x[1]+(0.848931358475)*x[2]
        arg[2,1,2,0]=(0.2618965198)*x[0]+(-0.461672573339)*x[1]+(0.923592803172)*x[2]
        arg[2,1,2,1]=(-0.326817258523)*x[0]+(-0.0363300805442)*x[1]+(1.82049875918)*x[2]
        arg[2,2,0,0]=(-0.435113320888)*x[0]+(0.0257984353786)*x[1]+(0.287846577657)*x[2]
        arg[2,2,0,1]=(1.39178068279)*x[0]+(0.136781967539)*x[1]+(-1.38161590351)*x[2]
        arg[2,2,1,0]=(-0.887874199356)*x[0]+(-0.476769153654)*x[1]+(-0.618760485829)*x[2]
        arg[2,2,1,1]=(1.58883355098)*x[0]+(-0.889469047555)*x[1]+(-0.678486755843)*x[2]
        arg[2,2,2,0]=(0.237147029365)*x[0]+(-0.509883462401)*x[1]+(0.533163151551)*x[2]
        arg[2,2,2,1]=(0.66513913749)*x[0]+(1.30317030657)*x[1]+(0.804083365617)*x[2]
        arg[2,3,0,0]=(1.01531518403)*x[0]+(1.3753053143)*x[1]+(0.000877284235964)*x[2]
        arg[2,3,0,1]=(-0.40853642543)*x[0]+(0.154794268783)*x[1]+(-1.34306671117)*x[2]
        arg[2,3,1,0]=(0.43065284359)*x[0]+(-0.28965146009)*x[1]+(0.660248828789)*x[2]
        arg[2,3,1,1]=(-1.4334922681)*x[0]+(-0.0348802858609)*x[1]+(0.502051304974)*x[2]
        arg[2,3,2,0]=(-1.17399537487)*x[0]+(0.865787043946)*x[1]+(-0.0105949677478)*x[2]
        arg[2,3,2,1]=(0.63104158794)*x[0]+(0.436196079951)*x[1]+(0.498606623061)*x[2]
        arg[2,4,0,0]=(-1.44410177354)*x[0]+(0.513406326011)*x[1]+(0.699127199447)*x[2]
        arg[2,4,0,1]=(-0.981727023077)*x[0]+(0.484182120695)*x[1]+(0.667175832045)*x[2]
        arg[2,4,1,0]=(-1.3352602588)*x[0]+(0.626269023364)*x[1]+(0.813537646695)*x[2]
        arg[2,4,1,1]=(-0.932486930114)*x[0]+(0.299448807849)*x[1]+(0.195862014797)*x[2]
        arg[2,4,2,0]=(-0.0621489644806)*x[0]+(0.131979664158)*x[1]+(0.442029269778)*x[2]
        arg[2,4,2,1]=(-0.507508608633)*x[0]+(-1.28613226533)*x[1]+(-0.697920444023)*x[2]
        arg[3,0,0,0]=(-0.984942677585)*x[0]+(-1.58913299616)*x[1]+(-0.373026784822)*x[2]
        arg[3,0,0,1]=(0.616666427113)*x[0]+(-0.625162800471)*x[1]+(-0.796307290839)*x[2]
        arg[3,0,1,0]=(-0.48469203683)*x[0]+(-0.0572105166553)*x[1]+(0.809572383581)*x[2]
        arg[3,0,1,1]=(-0.487160309892)*x[0]+(0.77646454341)*x[1]+(-0.41947544573)*x[2]
        arg[3,0,2,0]=(-1.17902121554)*x[0]+(-0.244939532033)*x[1]+(-0.282994359476)*x[2]
        arg[3,0,2,1]=(0.33515909881)*x[0]+(0.146866561705)*x[1]+(1.81486098364)*x[2]
        arg[3,1,0,0]=(0.379977901341)*x[0]+(0.834531701517)*x[1]+(-0.0242027062166)*x[2]
        arg[3,1,0,1]=(0.184318292766)*x[0]+(1.08367911805)*x[1]+(-0.164531741594)*x[2]
        arg[3,1,1,0]=(-1.00967115996)*x[0]+(0.542699640149)*x[1]+(0.707616701259)*x[2]
        arg[3,1,1,1]=(0.856948233811)*x[0]+(-1.2186550297)*x[1]+(-0.034268798159)*x[2]
        arg[3,1,2,0]=(0.259470684477)*x[0]+(-1.03586983732)*x[1]+(-0.157275979341)*x[2]
        arg[3,1,2,1]=(-0.161954890658)*x[0]+(1.65853209885)*x[1]+(0.381113456719)*x[2]
        arg[3,2,0,0]=(-0.437413004681)*x[0]+(-0.730340273929)*x[1]+(-0.687689272494)*x[2]
        arg[3,2,0,1]=(-0.336618164118)*x[0]+(-1.24388666895)*x[1]+(-0.2198689254)*x[2]
        arg[3,2,1,0]=(0.0872793087887)*x[0]+(-0.336554573215)*x[1]+(-0.0597803523262)*x[2]
        arg[3,2,1,1]=(0.192805432018)*x[0]+(0.0540495582064)*x[1]+(0.846658042071)*x[2]
        arg[3,2,2,0]=(0.00794192604414)*x[0]+(-0.33504559053)*x[1]+(0.615701149787)*x[2]
        arg[3,2,2,1]=(0.210870961503)*x[0]+(1.14164095846)*x[1]+(-0.192785742791)*x[2]
        arg[3,3,0,0]=(-0.25895162041)*x[0]+(-0.225800436488)*x[1]+(-0.198007079838)*x[2]
        arg[3,3,0,1]=(-0.150363703231)*x[0]+(1.23870429535)*x[1]+(0.28635881247)*x[2]
        arg[3,3,1,0]=(-0.865980718216)*x[0]+(-0.0849853244335)*x[1]+(-0.292164217189)*x[2]
        arg[3,3,1,1]=(-0.350465968722)*x[0]+(0.09853736936)*x[1]+(-0.476711524665)*x[2]
        arg[3,3,2,0]=(0.311396300263)*x[0]+(-0.92437612034)*x[1]+(-0.757139962928)*x[2]
        arg[3,3,2,1]=(-0.912311919837)*x[0]+(0.267793631099)*x[1]+(0.4917345401)*x[2]
        arg[3,4,0,0]=(0.613662541753)*x[0]+(0.782757341361)*x[1]+(1.04209808158)*x[2]
        arg[3,4,0,1]=(0.492828613997)*x[0]+(1.11979095664)*x[1]+(-0.843545072354)*x[2]
        arg[3,4,1,0]=(-0.51939120556)*x[0]+(0.476004484286)*x[1]+(-0.516311785432)*x[2]
        arg[3,4,1,1]=(1.00423313567)*x[0]+(0.90776890388)*x[1]+(1.34423309497)*x[2]
        arg[3,4,2,0]=(-1.44069600826)*x[0]+(-1.39515290909)*x[1]+(1.19851590247)*x[2]
        arg[3,4,2,1]=(0.391851619877)*x[0]+(0.265529120553)*x[1]+(0.996610410913)*x[2]
        ref[0,0,0,0]=(1.19330263935)/(o+1.)+(0.55394275251)+(0.0162017895041)*0.5**o
        ref[0,0,0,1]=(-0.672351783684)/(o+1.)+(-0.187889622039)+(0.351132807711)*0.5**o
        ref[0,0,1,0]=(-1.27021996264)/(o+1.)+(0.236286530101)+(-0.961885459894)*0.5**o
        ref[0,0,1,1]=(-1.43442517107)/(o+1.)+(-1.00113374859)+(-0.935971122022)*0.5**o
        ref[0,0,2,0]=(-0.434813695127)/(o+1.)+(0.606673209583)+(-0.0699546855712)*0.5**o
        ref[0,0,2,1]=(-0.0434901247948)/(o+1.)+(-0.180911877983)+(-0.733211446073)*0.5**o
        ref[0,1,0,0]=(1.41514302182)/(o+1.)+(-0.247977413325)+(-0.967062307989)*0.5**o
        ref[0,1,0,1]=(0.69128100142)/(o+1.)+(-0.621609882923)+(-0.319976390946)*0.5**o
        ref[0,1,1,0]=(-0.881870429191)/(o+1.)+(-0.654075774082)+(0.0281262553057)*0.5**o
        ref[0,1,1,1]=(-1.27239116086)/(o+1.)+(-0.0854066429492)+(0.0230865806541)*0.5**o
        ref[0,1,2,0]=(-0.222663945351)/(o+1.)+(0.685312598747)+(0.453385341573)*0.5**o
        ref[0,1,2,1]=(0.101409038706)/(o+1.)+(0.151367946744)+(-0.824910502191)*0.5**o
        ref[0,2,0,0]=(-1.19509443892)/(o+1.)+(0.225437856866)+(0.85928567141)*0.5**o
        ref[0,2,0,1]=(0.285764347754)/(o+1.)+(-0.383432127133)+(0.963095541446)*0.5**o
        ref[0,2,1,0]=(0.728010227911)/(o+1.)+(0.983790919896)+(-0.550403317717)*0.5**o
        ref[0,2,1,1]=(0.563748259036)/(o+1.)+(0.165164845927)+(-0.110710881481)*0.5**o
        ref[0,2,2,0]=(-0.361502606429)/(o+1.)+(0.0276145482611)+(-0.341115463028)*0.5**o
        ref[0,2,2,1]=(-0.293320207933)/(o+1.)+(0.0336757304385)+(-0.853103403872)*0.5**o
        ref[0,3,0,0]=(-0.410233956527)/(o+1.)+(-0.474281126245)+(0.569057885785)*0.5**o
        ref[0,3,0,1]=(-0.835861970115)/(o+1.)+(0.0173715483175)+(0.346902459481)*0.5**o
        ref[0,3,1,0]=(0.0830466456187)/(o+1.)+(-0.288204681495)+(-0.607109864241)*0.5**o
        ref[0,3,1,1]=(1.07074130823)/(o+1.)+(-0.0437983012867)+(0.159379549652)*0.5**o
        ref[0,3,2,0]=(0.172301049609)/(o+1.)+(-0.491724874152)+(-0.996925180895)*0.5**o
        ref[0,3,2,1]=(-0.24891941221)/(o+1.)+(-0.03935856269)+(-0.3030090126)*0.5**o
        ref[0,4,0,0]=(-0.778575122429)/(o+1.)+(0.213085222078)+(0.941383668349)*0.5**o
        ref[0,4,0,1]=(0.880801627174)/(o+1.)+(-0.454451723401)+(-0.193978040747)*0.5**o
        ref[0,4,1,0]=(-0.835604724882)/(o+1.)+(0.127240834083)+(-0.245156740036)*0.5**o
        ref[0,4,1,1]=(1.62336716324)/(o+1.)+(-0.101085918464)+(0.398884015839)*0.5**o
        ref[0,4,2,0]=(1.16805827702)/(o+1.)+(0.211232591688)+(-0.654729385702)*0.5**o
        ref[0,4,2,1]=(-0.305424138957)/(o+1.)+(-0.431076572693)+(-0.924903192969)*0.5**o
        ref[1,0,0,0]=(-0.481596247529)/(o+1.)+(0.769902632559)+(0.748050413171)*0.5**o
        ref[1,0,0,1]=(0.177388661252)/(o+1.)+(-0.118766442613)+(-0.457657277952)*0.5**o
        ref[1,0,1,0]=(0.709890860372)/(o+1.)+(0.0668010230727)+(0.658744950012)*0.5**o
        ref[1,0,1,1]=(-1.45685781381)/(o+1.)+(-0.178674426089)+(0.788456069079)*0.5**o
        ref[1,0,2,0]=(0.194388185612)/(o+1.)+(0.649177543063)+(0.896093783855)*0.5**o
        ref[1,0,2,1]=(0.181561279507)/(o+1.)+(-0.200618291902)+(-0.392507985678)*0.5**o
        ref[1,1,0,0]=(-0.884376814027)/(o+1.)+(0.280382736159)+(0.593213920451)*0.5**o
        ref[1,1,0,1]=(-0.250709602218)/(o+1.)+(-0.492104211211)+(0.818920586956)*0.5**o
        ref[1,1,1,0]=(0.0529723694625)/(o+1.)+(-0.502047650638)+(-0.115739819752)*0.5**o
        ref[1,1,1,1]=(0.215812202957)/(o+1.)+(-0.308796450292)+(0.954456105171)*0.5**o
        ref[1,1,2,0]=(-0.991481050902)/(o+1.)+(0.162626252838)+(-0.800147845717)*0.5**o
        ref[1,1,2,1]=(-0.433989756951)/(o+1.)+(0.362776336712)+(-0.259441100777)*0.5**o
        ref[1,2,0,0]=(-0.930287052624)/(o+1.)+(-0.347752763108)+(0.880525373413)*0.5**o
        ref[1,2,0,1]=(0.698674852861)/(o+1.)+(-0.15119613196)+(0.719004238051)*0.5**o
        ref[1,2,1,0]=(-0.777216821365)/(o+1.)+(0.436336076297)+(0.617741643399)*0.5**o
        ref[1,2,1,1]=(1.09595223135)/(o+1.)+(0.978251130983)+(0.804336219412)*0.5**o
        ref[1,2,2,0]=(-0.651012630849)/(o+1.)+(-1.04894606664)+(0.43528198181)*0.5**o
        ref[1,2,2,1]=(-1.5227353883)/(o+1.)+(0.670019307557)+(-0.745922598001)*0.5**o
        ref[1,3,0,0]=(-0.750195637997)/(o+1.)+(0.231864245348)+(-0.118356142533)*0.5**o
        ref[1,3,0,1]=(1.17136226596)/(o+1.)+(-0.114976421389)+(0.395383920468)*0.5**o
        ref[1,3,1,0]=(-0.108070636769)/(o+1.)+(0.279145429234)+(-0.352562621327)*0.5**o
        ref[1,3,1,1]=(-0.144223481714)/(o+1.)+(-0.472954352493)+(-0.0903026345013)*0.5**o
        ref[1,3,2,0]=(-0.487564307786)/(o+1.)+(0.636973351449)+(0.758622537094)*0.5**o
        ref[1,3,2,1]=(0.390860965954)/(o+1.)+(-0.36647678165)+(-0.66980953377)*0.5**o
        ref[1,4,0,0]=(-1.04824984837)/(o+1.)+(0.213932231728)+(0.823722790054)*0.5**o
        ref[1,4,0,1]=(-0.0984653601336)/(o+1.)+(0.07867877705)+(-0.741527483773)*0.5**o
        ref[1,4,1,0]=(-0.146287867383)/(o+1.)+(-0.908951868158)+(0.425691754878)*0.5**o
        ref[1,4,1,1]=(-0.94160416554)/(o+1.)+(0.0459877010016)+(-0.869872381431)*0.5**o
        ref[1,4,2,0]=(0.0115742858952)/(o+1.)+(-0.222488399041)+(0.687392452389)*0.5**o
        ref[1,4,2,1]=(-0.956561936575)/(o+1.)+(-0.59770879042)+(-0.0613652834159)*0.5**o
        ref[2,0,0,0]=(1.83987381729)/(o+1.)+(0.956579317963)+(0.862634531274)*0.5**o
        ref[2,0,0,1]=(0.105446127421)/(o+1.)+(0.180661881191)+(0.186629121375)*0.5**o
        ref[2,0,1,0]=(0.0212167440766)/(o+1.)+(-0.337825588888)+(-0.318943328859)*0.5**o
        ref[2,0,1,1]=(0.744834715631)/(o+1.)+(-0.350058416663)+(0.391466726104)*0.5**o
        ref[2,0,2,0]=(-0.788776613856)/(o+1.)+(0.233776806014)+(0.400043748043)*0.5**o
        ref[2,0,2,1]=(-0.95346976307)/(o+1.)+(0.0274998652324)+(-0.967971238003)*0.5**o
        ref[2,1,0,0]=(-0.536051408942)/(o+1.)+(-0.208498016994)+(0.590315312233)*0.5**o
        ref[2,1,0,1]=(-1.04158508788)/(o+1.)+(-0.397885793101)+(0.0127762955396)*0.5**o
        ref[2,1,1,0]=(0.529914207253)/(o+1.)+(-0.547503443987)+(-0.338873793431)*0.5**o
        ref[2,1,1,1]=(0.62298670274)/(o+1.)+(0.449860048786)+(0.352800726478)*0.5**o
        ref[2,1,2,0]=(0.965292650595)/(o+1.)+(0.00980498639)+(-0.261085873742)*0.5**o
        ref[2,1,2,1]=(0.403285360803)/(o+1.)+(0.369160592269)+(0.315744874774)*0.5**o
        ref[2,2,0,0]=(-0.457764884724)/(o+1.)+(0.563657776706)+(-0.791018976541)*0.5**o
        ref[2,2,0,1]=(0.188151274044)/(o+1.)+(-0.476470545911)+(0.911736564588)*0.5**o
        ref[2,2,1,0]=(-0.558515085739)/(o+1.)+(-0.445583796553)+(-0.533721159994)*0.5**o
        ref[2,2,1,1]=(-0.431992337265)/(o+1.)+(-0.0957726221961)+(0.64441532924)*0.5**o
        ref[2,2,2,0]=(-0.159056461947)/(o+1.)+(-0.0326864536156)+(0.484856087693)*0.5**o
        ref[2,2,2,1]=(0.533116456037)/(o+1.)+(1.10197257725)+(0.0353311991352)*0.5**o
        ref[2,3,0,0]=(0.0570625147023)/(o+1.)+(0.74170779854)+(0.851019670777)*0.5**o
        ref[2,3,0,1]=(-0.495059904832)/(o+1.)+(-0.169954912181)+(-0.761839138623)*0.5**o
        ref[2,3,1,0]=(0.482375173314)/(o+1.)+(-0.19374838651)+(0.706371811996)*0.5**o
        ref[2,3,1,1]=(0.12928100721)/(o+1.)+(-0.317327862062)+(-0.460946532077)*0.5**o
        ref[2,3,2,0]=(-0.234362084763)/(o+1.)+(0.423486279181)+(-0.93141377227)*0.5**o
        ref[2,3,2,1]=(1.25586693002)/(o+1.)+(-0.127551329993)+(0.56508002092)*0.5**o
        ref[2,4,0,0]=(0.45235297749)/(o+1.)+(-0.0217537364788)+(-0.640413752615)*0.5**o
        ref[2,4,0,1]=(0.936809207384)/(o+1.)+(-0.151519610695)+(-0.464139056331)*0.5**o
        ref[2,4,1,0]=(0.867693208863)/(o+1.)+(0.105683574314)+(-0.974513946233)*0.5**o
        ref[2,4,1,1]=(-1.27821575101)/(o+1.)+(0.463163847311)+(-0.0852880510814)*0.5**o
        ref[2,4,2,0]=(-0.846224501366)/(o+1.)+(0.432619461409)+(0.492845548002)*0.5**o
        ref[2,4,2,1]=(-1.48923229995)/(o+1.)+(-0.505747396375)+(0.00916577472118)*0.5**o
        ref[3,0,0,0]=(-0.959215731121)/(o+1.)+(-0.904609108587)+(-0.178668510268)*0.5**o
        ref[3,0,0,1]=(-1.53181231865)/(o+1.)+(-0.0758543610408)+(0.878717376536)*0.5**o
        ref[3,0,1,0]=(1.18643238831)/(o+1.)+(-0.288825159097)+(-0.341112240015)*0.5**o
        ref[3,0,1,1]=(-0.0675694917031)/(o+1.)+(-0.168955703616)+(0.275309686724)*0.5**o
        ref[3,0,2,0]=(-0.663269860133)/(o+1.)+(-0.235102421885)+(-0.573480403148)*0.5**o
        ref[3,0,2,1]=(1.0969289778)/(o+1.)+(0.46603172821)+(0.267894209932)*0.5**o
        ref[3,1,0,0]=(-0.716870513559)/(o+1.)+(0.712039409283)+(0.483098591635)*0.5**o
        ref[3,1,0,1]=(0.0328531999411)/(o+1.)+(0.379319450988)+(0.311973567308)*0.5**o
        ref[3,1,1,0]=(1.08648208064)/(o+1.)+(-0.196831959338)+(-0.452172980519)*0.5**o
        ref[3,1,1,1]=(0.0813181502612)/(o+1.)+(-0.227364064527)+(-0.0225656152548)*0.5**o
        ref[3,1,2,0]=(-0.271000454408)/(o+1.)+(-0.765711687155)+(0.868748696531)*0.5**o
        ref[3,1,2,1]=(1.60569646253)/(o+1.)+(0.137437739736)+(-0.00288127709197)*0.5**o
        ref[3,2,0,0]=(-1.25887089059)/(o+1.)+(-0.525842290945)+(0.455112921376)*0.5**o
        ref[3,2,0,1]=(-1.63787880725)/(o+1.)+(0.00214706725405)+(-0.166789085722)*0.5**o
        ref[3,2,1,0]=(-1.82111307894)/(o+1.)+(0.561806270366)+(0.388444921454)*0.5**o
        ref[3,2,1,1]=(-0.489259906414)/(o+1.)+(0.582548398381)+(0.417676141947)*0.5**o
        ref[3,2,2,0]=(-0.0674375340391)/(o+1.)+(-0.00345549227349)+(0.362946003887)*0.5**o
        ref[3,2,2,1]=(0.662780682001)/(o+1.)+(-0.15453848218)+(0.806022459534)*0.5**o
        ref[3,3,0,0]=(-1.15795406515)/(o+1.)+(0.0926725746467)+(0.289849779118)*0.5**o
        ref[3,3,0,1]=(1.27520269547)/(o+1.)+(-0.286529818361)+(0.672556345847)*0.5**o
        ref[3,3,1,0]=(-0.343290651469)/(o+1.)+(-0.0269962100904)+(-0.845847188189)*0.5**o
        ref[3,3,1,1]=(0.497715053371)/(o+1.)+(-0.158577177766)+(-0.909200821866)*0.5**o
        ref[3,3,2,0]=(-0.952431538662)/(o+1.)+(0.0816482063045)+(-0.580984656952)*0.5**o
        ref[3,3,2,1]=(-0.176227469799)/(o+1.)+(0.0456157853688)+(-0.0677878495767)*0.5**o
        ref[3,4,0,0]=(0.976460483802)/(o+1.)+(0.347595506011)+(0.766866468871)*0.5**o
        ref[3,4,0,1]=(0.0358170327662)/(o+1.)+(-0.0991456259778)+(0.931548717473)*0.5**o
        ref[3,4,1,0]=(0.114544318514)/(o+1.)+(-0.201014865792)+(-0.272213093635)*0.5**o
        ref[3,4,1,1]=(0.554509925385)/(o+1.)+(1.23219332557)+(0.237338557994)*0.5**o
        ref[3,4,2,0]=(0.0803465444931)/(o+1.)+(-0.518257569997)+(-0.681164419371)*0.5**o
        ref[3,4,2,1]=(-0.0443609570522)/(o+1.)+(0.572813368096)+(0.552725372202)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_FunctionOnContactZero_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactZero

      assumptions: FunctionOnContactZero(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.114918917602)*x[0]**o+(0.304834778607)*x[0]+(0.704057176139)*x[1]**o+(-0.0664624375434)*x[1]
        ref=(0.704057176139)/(o+1.)+(0.119186170532)+(-0.114918917602)*0.5**o
      else:
        arg=(-0.134961573162)*x[0]**o+(0.585433861814)*x[0]+(0.209596676649)*x[1]**o+(0.870867416311)*x[1]+(-0.902854033973)*x[2]**o+(-0.812135783993)*x[2]
        ref=(-0.693257357324)/(o+1.)+(0.322082747066)+(-0.134961573162)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_FunctionOnContactZero_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactZero

      assumptions: FunctionOnContactZero(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.441677901994)*x[0]**o+(0.0234076994082)*x[0]+(-0.520465368981)*x[1]**o+(-0.668717137452)*x[1]
        arg[1]=(0.0637451453664)*x[0]**o+(0.610543665309)*x[0]+(0.639008465664)*x[1]**o+(-0.460500756968)*x[1]
        ref[0]=(-0.520465368981)/(o+1.)+(-0.322654719022)+(0.441677901994)*0.5**o
        ref[1]=(0.639008465664)/(o+1.)+(0.0750214541708)+(0.0637451453664)*0.5**o
      else:
        arg[0]=(0.207042421638)*x[0]**o+(0.586067682859)*x[0]+(-0.731189752251)*x[1]**o+(0.439755444977)*x[1]+(-0.215027146515)*x[2]**o+(-0.22109699198)*x[2]
        arg[1]=(-0.482897746923)*x[0]**o+(-0.262848758384)*x[0]+(-0.130231246709)*x[1]**o+(0.494143899685)*x[1]+(0.871226664446)*x[2]**o+(-0.45136577521)*x[2]
        ref[0]=(-0.946216898767)/(o+1.)+(0.402363067928)+(0.207042421638)*0.5**o
        ref[1]=(0.740995417737)/(o+1.)+(-0.110035316955)+(-0.482897746923)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_FunctionOnContactZero_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactZero

      assumptions: FunctionOnContactZero(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.418217706031)*x[0]**o+(0.817101274679)*x[0]+(0.359227724634)*x[1]**o+(0.106952446872)*x[1]
        arg[0,1]=(0.735477258623)*x[0]**o+(-0.181692404492)*x[0]+(0.118248218763)*x[1]**o+(0.910033225874)*x[1]
        arg[0,2]=(-0.370301174151)*x[0]**o+(0.421589410528)*x[0]+(-0.91974297037)*x[1]**o+(0.676441845707)*x[1]
        arg[0,3]=(0.170535646217)*x[0]**o+(0.650458506701)*x[0]+(-0.431898202846)*x[1]**o+(-0.89133664224)*x[1]
        arg[0,4]=(-0.288793192482)*x[0]**o+(-0.752066662422)*x[0]+(0.279565184651)*x[1]**o+(-0.20569858505)*x[1]
        arg[1,0]=(-0.831478902729)*x[0]**o+(-0.545347589281)*x[0]+(-0.315738709444)*x[1]**o+(-0.407156023524)*x[1]
        arg[1,1]=(0.119396512736)*x[0]**o+(0.56655186098)*x[0]+(0.798206239231)*x[1]**o+(0.0342467197178)*x[1]
        arg[1,2]=(-0.264721761214)*x[0]**o+(-0.7583989175)*x[0]+(-0.350411882147)*x[1]**o+(-0.556879396863)*x[1]
        arg[1,3]=(0.976029513102)*x[0]**o+(-0.117830794452)*x[0]+(0.318067419468)*x[1]**o+(-0.496129688948)*x[1]
        arg[1,4]=(0.462797018845)*x[0]**o+(-0.214502973992)*x[0]+(0.155052684044)*x[1]**o+(0.264794267973)*x[1]
        arg[2,0]=(0.193892132778)*x[0]**o+(-0.440406597181)*x[0]+(0.527330868842)*x[1]**o+(0.856602716155)*x[1]
        arg[2,1]=(0.0418968976051)*x[0]**o+(0.944283628708)*x[0]+(0.923548919008)*x[1]**o+(0.958011594347)*x[1]
        arg[2,2]=(0.744026460715)*x[0]**o+(0.238896690247)*x[0]+(0.968891312128)*x[1]**o+(0.0636831438113)*x[1]
        arg[2,3]=(-0.289048983574)*x[0]**o+(0.897469203226)*x[0]+(0.948131139828)*x[1]**o+(0.821749060686)*x[1]
        arg[2,4]=(-0.903766144138)*x[0]**o+(-0.453139974489)*x[0]+(-0.715818671181)*x[1]**o+(-0.712367492971)*x[1]
        arg[3,0]=(0.315277263598)*x[0]**o+(-0.676815102073)*x[0]+(0.428704415911)*x[1]**o+(0.54703485039)*x[1]
        arg[3,1]=(-0.301882901745)*x[0]**o+(0.888822293672)*x[0]+(-0.496204235393)*x[1]**o+(0.210598259748)*x[1]
        arg[3,2]=(0.568071178492)*x[0]**o+(-0.552863936415)*x[0]+(0.758619958752)*x[1]**o+(0.520428540448)*x[1]
        arg[3,3]=(0.477425351692)*x[0]**o+(-0.427405588029)*x[0]+(0.197959279245)*x[1]**o+(0.328047998649)*x[1]
        arg[3,4]=(-0.869704203328)*x[0]**o+(-0.0825164966999)*x[0]+(0.779325592334)*x[1]**o+(0.419916607608)*x[1]
        ref[0,0]=(0.359227724634)/(o+1.)+(0.462026860776)+(0.418217706031)*0.5**o
        ref[0,1]=(0.118248218763)/(o+1.)+(0.364170410691)+(0.735477258623)*0.5**o
        ref[0,2]=(-0.91974297037)/(o+1.)+(0.549015628118)+(-0.370301174151)*0.5**o
        ref[0,3]=(-0.431898202846)/(o+1.)+(-0.120439067769)+(0.170535646217)*0.5**o
        ref[0,4]=(0.279565184651)/(o+1.)+(-0.478882623736)+(-0.288793192482)*0.5**o
        ref[1,0]=(-0.315738709444)/(o+1.)+(-0.476251806402)+(-0.831478902729)*0.5**o
        ref[1,1]=(0.798206239231)/(o+1.)+(0.300399290349)+(0.119396512736)*0.5**o
        ref[1,2]=(-0.350411882147)/(o+1.)+(-0.657639157182)+(-0.264721761214)*0.5**o
        ref[1,3]=(0.318067419468)/(o+1.)+(-0.3069802417)+(0.976029513102)*0.5**o
        ref[1,4]=(0.155052684044)/(o+1.)+(0.02514564699)+(0.462797018845)*0.5**o
        ref[2,0]=(0.527330868842)/(o+1.)+(0.208098059487)+(0.193892132778)*0.5**o
        ref[2,1]=(0.923548919008)/(o+1.)+(0.951147611528)+(0.0418968976051)*0.5**o
        ref[2,2]=(0.968891312128)/(o+1.)+(0.151289917029)+(0.744026460715)*0.5**o
        ref[2,3]=(0.948131139828)/(o+1.)+(0.859609131956)+(-0.289048983574)*0.5**o
        ref[2,4]=(-0.715818671181)/(o+1.)+(-0.58275373373)+(-0.903766144138)*0.5**o
        ref[3,0]=(0.428704415911)/(o+1.)+(-0.0648901258416)+(0.315277263598)*0.5**o
        ref[3,1]=(-0.496204235393)/(o+1.)+(0.54971027671)+(-0.301882901745)*0.5**o
        ref[3,2]=(0.758619958752)/(o+1.)+(-0.0162176979831)+(0.568071178492)*0.5**o
        ref[3,3]=(0.197959279245)/(o+1.)+(-0.0496787946897)+(0.477425351692)*0.5**o
        ref[3,4]=(0.779325592334)/(o+1.)+(0.168700055454)+(-0.869704203328)*0.5**o
      else:
        arg[0,0]=(0.204452789147)*x[0]**o+(0.411843111144)*x[0]+(-0.0502866367412)*x[1]**o+(0.957356261522)*x[1]+(0.173522462943)*x[2]**o+(-0.0194908382525)*x[2]
        arg[0,1]=(-0.49641528149)*x[0]**o+(0.158625922243)*x[0]+(0.670932733725)*x[1]**o+(-0.611829064862)*x[1]+(0.758373881782)*x[2]**o+(0.151236368375)*x[2]
        arg[0,2]=(0.577504765611)*x[0]**o+(0.872996499874)*x[0]+(-0.734000014722)*x[1]**o+(-0.890824538872)*x[1]+(-0.796786080271)*x[2]**o+(-0.389005562901)*x[2]
        arg[0,3]=(-0.933166619447)*x[0]**o+(-0.528627639194)*x[0]+(-0.828111326322)*x[1]**o+(0.528345547301)*x[1]+(0.99894455165)*x[2]**o+(0.472982583662)*x[2]
        arg[0,4]=(-0.495503045054)*x[0]**o+(0.629909217292)*x[0]+(0.901515784751)*x[1]**o+(-0.863668288904)*x[1]+(0.76892679355)*x[2]**o+(0.34730747389)*x[2]
        arg[1,0]=(0.303189803506)*x[0]**o+(-0.17543401741)*x[0]+(-0.283157585975)*x[1]**o+(0.950171529024)*x[1]+(0.0483247572497)*x[2]**o+(-0.604332298954)*x[2]
        arg[1,1]=(-0.225480407643)*x[0]**o+(-0.0521067696425)*x[0]+(-0.10259234894)*x[1]**o+(0.373551917005)*x[1]+(0.89047195372)*x[2]**o+(-0.292740890417)*x[2]
        arg[1,2]=(-0.476041156832)*x[0]**o+(0.916078681628)*x[0]+(-0.393887610676)*x[1]**o+(-0.651852504525)*x[1]+(0.611404175606)*x[2]**o+(0.336594647903)*x[2]
        arg[1,3]=(-0.415178048882)*x[0]**o+(0.145394549559)*x[0]+(0.680115054186)*x[1]**o+(0.261517386893)*x[1]+(0.781566117396)*x[2]**o+(-0.659634539391)*x[2]
        arg[1,4]=(0.874821341857)*x[0]**o+(0.317418381126)*x[0]+(0.051455923305)*x[1]**o+(0.670170143071)*x[1]+(-0.211902435336)*x[2]**o+(-0.175299130233)*x[2]
        arg[2,0]=(-0.372961546517)*x[0]**o+(0.199493150648)*x[0]+(-0.630239499285)*x[1]**o+(0.657412283442)*x[1]+(-0.82211915324)*x[2]**o+(-0.468612977158)*x[2]
        arg[2,1]=(-0.503168527887)*x[0]**o+(0.904406483201)*x[0]+(-0.206179628633)*x[1]**o+(0.0787026584997)*x[1]+(0.899590466467)*x[2]**o+(-0.893401768413)*x[2]
        arg[2,2]=(-0.860556408954)*x[0]**o+(0.39021140432)*x[0]+(-0.66425383352)*x[1]**o+(0.287711780417)*x[1]+(0.776884760126)*x[2]**o+(-0.115905192334)*x[2]
        arg[2,3]=(0.461359695993)*x[0]**o+(0.997877835019)*x[0]+(0.0513514112685)*x[1]**o+(-0.904812988908)*x[1]+(-0.903222331838)*x[2]**o+(-0.470185679041)*x[2]
        arg[2,4]=(-0.646560730528)*x[0]**o+(0.10038324067)*x[0]+(0.426753245517)*x[1]**o+(-0.109889624344)*x[1]+(-0.20158193753)*x[2]**o+(0.014258531224)*x[2]
        arg[3,0]=(0.364407704271)*x[0]**o+(0.840519040103)*x[0]+(0.528721642544)*x[1]**o+(-0.853679853614)*x[1]+(0.079836122575)*x[2]**o+(-0.439039178308)*x[2]
        arg[3,1]=(-0.0726491927955)*x[0]**o+(0.217804431494)*x[0]+(-0.454354285545)*x[1]**o+(0.390539936003)*x[1]+(-0.989667912688)*x[2]**o+(-0.0674079606774)*x[2]
        arg[3,2]=(0.62239422825)*x[0]**o+(0.998733716844)*x[0]+(0.796258161536)*x[1]**o+(0.618545466026)*x[1]+(-0.243188896911)*x[2]**o+(-0.677687584224)*x[2]
        arg[3,3]=(0.444073180258)*x[0]**o+(-0.764729326792)*x[0]+(-0.823491453567)*x[1]**o+(0.985393963131)*x[1]+(0.623800470977)*x[2]**o+(0.947328422513)*x[2]
        arg[3,4]=(0.686438920325)*x[0]**o+(0.789760134176)*x[0]+(-0.26011082782)*x[1]**o+(0.0692377043339)*x[1]+(-0.3712894236)*x[2]**o+(-0.542974191103)*x[2]
        ref[0,0]=(0.123235826202)/(o+1.)+(0.674854267207)+(0.204452789147)*0.5**o
        ref[0,1]=(1.42930661551)/(o+1.)+(-0.150983387122)+(-0.49641528149)*0.5**o
        ref[0,2]=(-1.53078609499)/(o+1.)+(-0.203416800949)+(0.577504765611)*0.5**o
        ref[0,3]=(0.170833225328)/(o+1.)+(0.236350245885)+(-0.933166619447)*0.5**o
        ref[0,4]=(1.6704425783)/(o+1.)+(0.0567742011389)+(-0.495503045054)*0.5**o
        ref[1,0]=(-0.234832828725)/(o+1.)+(0.0852026063299)+(0.303189803506)*0.5**o
        ref[1,1]=(0.787879604779)/(o+1.)+(0.0143521284725)+(-0.225480407643)*0.5**o
        ref[1,2]=(0.21751656493)/(o+1.)+(0.300410412503)+(-0.476041156832)*0.5**o
        ref[1,3]=(1.46168117158)/(o+1.)+(-0.126361301469)+(-0.415178048882)*0.5**o
        ref[1,4]=(-0.160446512031)/(o+1.)+(0.406144696982)+(0.874821341857)*0.5**o
        ref[2,0]=(-1.45235865253)/(o+1.)+(0.194146228466)+(-0.372961546517)*0.5**o
        ref[2,1]=(0.693410837835)/(o+1.)+(0.0448536866438)+(-0.503168527887)*0.5**o
        ref[2,2]=(0.112630926606)/(o+1.)+(0.281008996202)+(-0.860556408954)*0.5**o
        ref[2,3]=(-0.85187092057)/(o+1.)+(-0.188560416464)+(0.461359695993)*0.5**o
        ref[2,4]=(0.225171307987)/(o+1.)+(0.00237607377473)+(-0.646560730528)*0.5**o
        ref[3,0]=(0.608557765119)/(o+1.)+(-0.22609999591)+(0.364407704271)*0.5**o
        ref[3,1]=(-1.44402219823)/(o+1.)+(0.27046820341)+(-0.0726491927955)*0.5**o
        ref[3,2]=(0.553069264625)/(o+1.)+(0.469795799323)+(0.62239422825)*0.5**o
        ref[3,3]=(-0.19969098259)/(o+1.)+(0.583996529426)+(0.444073180258)*0.5**o
        ref[3,4]=(-0.631400251421)/(o+1.)+(0.158011823704)+(0.686438920325)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_FunctionOnContactZero_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactZero

      assumptions: FunctionOnContactZero(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.963014866001)*x[0]**o+(-0.592091539519)*x[0]+(0.397897510985)*x[1]**o+(0.485612885034)*x[1]
        arg[0,0,1]=(-0.764849027936)*x[0]**o+(0.936217923087)*x[0]+(-0.6256860991)*x[1]**o+(0.162359755936)*x[1]
        arg[0,1,0]=(-0.269607090458)*x[0]**o+(0.197100067821)*x[0]+(-0.216513043667)*x[1]**o+(0.143038940436)*x[1]
        arg[0,1,1]=(-0.432206914977)*x[0]**o+(0.177851208963)*x[0]+(0.0477652910843)*x[1]**o+(0.278333389583)*x[1]
        arg[1,0,0]=(-0.123940629628)*x[0]**o+(0.818553665062)*x[0]+(-0.168322144522)*x[1]**o+(0.485324994762)*x[1]
        arg[1,0,1]=(0.593609516922)*x[0]**o+(0.530508390905)*x[0]+(-0.938421636915)*x[1]**o+(-0.686459217917)*x[1]
        arg[1,1,0]=(-0.874022523162)*x[0]**o+(0.694500749056)*x[0]+(-0.650951516798)*x[1]**o+(0.45826621855)*x[1]
        arg[1,1,1]=(-0.259058106634)*x[0]**o+(0.0837282545694)*x[0]+(0.465734884341)*x[1]**o+(0.185558060279)*x[1]
        arg[2,0,0]=(0.346292866419)*x[0]**o+(0.982902010078)*x[0]+(0.0297606445334)*x[1]**o+(0.339043518987)*x[1]
        arg[2,0,1]=(0.749684718966)*x[0]**o+(-0.855799436867)*x[0]+(-0.186333913876)*x[1]**o+(-0.548968316245)*x[1]
        arg[2,1,0]=(-0.255759466052)*x[0]**o+(0.441882703647)*x[0]+(-0.0239807362322)*x[1]**o+(-0.236417762222)*x[1]
        arg[2,1,1]=(-0.0586694299212)*x[0]**o+(-0.912208422332)*x[0]+(-0.11610871626)*x[1]**o+(-0.512805786403)*x[1]
        arg[3,0,0]=(0.826027179564)*x[0]**o+(-0.589876767386)*x[0]+(-0.654858828463)*x[1]**o+(0.601656609023)*x[1]
        arg[3,0,1]=(-0.148714999203)*x[0]**o+(-0.409138795316)*x[0]+(0.290726299307)*x[1]**o+(-0.94009130345)*x[1]
        arg[3,1,0]=(0.63431415979)*x[0]**o+(-0.184553980231)*x[0]+(0.4087395755)*x[1]**o+(-0.694376406449)*x[1]
        arg[3,1,1]=(-0.000928514830108)*x[0]**o+(-0.782814501236)*x[0]+(0.177390244103)*x[1]**o+(0.698562703841)*x[1]
        arg[4,0,0]=(-0.939264907756)*x[0]**o+(0.191360117367)*x[0]+(-0.318750748728)*x[1]**o+(0.910713477162)*x[1]
        arg[4,0,1]=(-0.105186863713)*x[0]**o+(0.0319652211881)*x[0]+(0.118896230389)*x[1]**o+(-0.212677289177)*x[1]
        arg[4,1,0]=(-0.702838100428)*x[0]**o+(0.881695294345)*x[0]+(0.728231539571)*x[1]**o+(-0.192311604244)*x[1]
        arg[4,1,1]=(0.25197450886)*x[0]**o+(0.464857988125)*x[0]+(0.851996888259)*x[1]**o+(-0.408525083481)*x[1]
        arg[5,0,0]=(0.665627182446)*x[0]**o+(0.5367740452)*x[0]+(0.295800616406)*x[1]**o+(0.0620589956995)*x[1]
        arg[5,0,1]=(-0.37320168697)*x[0]**o+(0.072855481498)*x[0]+(0.539806161326)*x[1]**o+(-0.999091813027)*x[1]
        arg[5,1,0]=(-0.060845772709)*x[0]**o+(-0.931869215952)*x[0]+(-0.393477853475)*x[1]**o+(0.968288081996)*x[1]
        arg[5,1,1]=(-0.0768045781739)*x[0]**o+(-0.257318730887)*x[0]+(0.781473167319)*x[1]**o+(-0.458870794544)*x[1]
        ref[0,0,0]=(0.397897510985)/(o+1.)+(-0.0532393272425)+(-0.963014866001)*0.5**o
        ref[0,0,1]=(-0.6256860991)/(o+1.)+(0.549288839511)+(-0.764849027936)*0.5**o
        ref[0,1,0]=(-0.216513043667)/(o+1.)+(0.170069504128)+(-0.269607090458)*0.5**o
        ref[0,1,1]=(0.0477652910843)/(o+1.)+(0.228092299273)+(-0.432206914977)*0.5**o
        ref[1,0,0]=(-0.168322144522)/(o+1.)+(0.651939329912)+(-0.123940629628)*0.5**o
        ref[1,0,1]=(-0.938421636915)/(o+1.)+(-0.0779754135063)+(0.593609516922)*0.5**o
        ref[1,1,0]=(-0.650951516798)/(o+1.)+(0.576383483803)+(-0.874022523162)*0.5**o
        ref[1,1,1]=(0.465734884341)/(o+1.)+(0.134643157424)+(-0.259058106634)*0.5**o
        ref[2,0,0]=(0.0297606445334)/(o+1.)+(0.660972764533)+(0.346292866419)*0.5**o
        ref[2,0,1]=(-0.186333913876)/(o+1.)+(-0.702383876556)+(0.749684718966)*0.5**o
        ref[2,1,0]=(-0.0239807362322)/(o+1.)+(0.102732470713)+(-0.255759466052)*0.5**o
        ref[2,1,1]=(-0.11610871626)/(o+1.)+(-0.712507104367)+(-0.0586694299212)*0.5**o
        ref[3,0,0]=(-0.654858828463)/(o+1.)+(0.0058899208184)+(0.826027179564)*0.5**o
        ref[3,0,1]=(0.290726299307)/(o+1.)+(-0.674615049383)+(-0.148714999203)*0.5**o
        ref[3,1,0]=(0.4087395755)/(o+1.)+(-0.43946519334)+(0.63431415979)*0.5**o
        ref[3,1,1]=(0.177390244103)/(o+1.)+(-0.0421258986973)+(-0.000928514830108)*0.5**o
        ref[4,0,0]=(-0.318750748728)/(o+1.)+(0.551036797264)+(-0.939264907756)*0.5**o
        ref[4,0,1]=(0.118896230389)/(o+1.)+(-0.0903560339946)+(-0.105186863713)*0.5**o
        ref[4,1,0]=(0.728231539571)/(o+1.)+(0.344691845051)+(-0.702838100428)*0.5**o
        ref[4,1,1]=(0.851996888259)/(o+1.)+(0.0281664523221)+(0.25197450886)*0.5**o
        ref[5,0,0]=(0.295800616406)/(o+1.)+(0.29941652045)+(0.665627182446)*0.5**o
        ref[5,0,1]=(0.539806161326)/(o+1.)+(-0.463118165764)+(-0.37320168697)*0.5**o
        ref[5,1,0]=(-0.393477853475)/(o+1.)+(0.0182094330217)+(-0.060845772709)*0.5**o
        ref[5,1,1]=(0.781473167319)/(o+1.)+(-0.358094762716)+(-0.0768045781739)*0.5**o
      else:
        arg[0,0,0]=(0.638389367885)*x[0]**o+(-0.77703387527)*x[0]+(0.951496212227)*x[1]**o+(-0.484305156823)*x[1]+(-0.582437996275)*x[2]**o+(-0.460130065965)*x[2]
        arg[0,0,1]=(0.859011160063)*x[0]**o+(0.999372374168)*x[0]+(-0.946049724069)*x[1]**o+(0.522184535496)*x[1]+(0.00820405492191)*x[2]**o+(0.769305042211)*x[2]
        arg[0,1,0]=(-0.998681640062)*x[0]**o+(0.916063974594)*x[0]+(0.0801858148899)*x[1]**o+(0.475476401258)*x[1]+(-0.805247247843)*x[2]**o+(-0.809331278141)*x[2]
        arg[0,1,1]=(0.1340261052)*x[0]**o+(0.843204972218)*x[0]+(-0.809433597166)*x[1]**o+(-0.947798215446)*x[1]+(0.268796347717)*x[2]**o+(-0.896620567629)*x[2]
        arg[1,0,0]=(0.251201824008)*x[0]**o+(0.762438451805)*x[0]+(0.892722429031)*x[1]**o+(0.146944927124)*x[1]+(0.00953973207152)*x[2]**o+(0.423262196414)*x[2]
        arg[1,0,1]=(-0.258843593921)*x[0]**o+(-0.950574260099)*x[0]+(-0.49096256134)*x[1]**o+(0.446238902975)*x[1]+(-0.880706204083)*x[2]**o+(-0.400075002454)*x[2]
        arg[1,1,0]=(-0.854529145157)*x[0]**o+(0.118165533114)*x[0]+(-0.913218271084)*x[1]**o+(0.991977570414)*x[1]+(0.626241160161)*x[2]**o+(0.667843914102)*x[2]
        arg[1,1,1]=(0.636818650335)*x[0]**o+(-0.174996100829)*x[0]+(0.77317621688)*x[1]**o+(0.369280208456)*x[1]+(-0.0130124322187)*x[2]**o+(0.857098751007)*x[2]
        arg[2,0,0]=(-0.415847479141)*x[0]**o+(0.675986357495)*x[0]+(0.704926614111)*x[1]**o+(-0.166544295134)*x[1]+(-0.00917226654826)*x[2]**o+(-0.896432190739)*x[2]
        arg[2,0,1]=(-0.882383192316)*x[0]**o+(0.11122730635)*x[0]+(-0.693587089789)*x[1]**o+(-0.0905457091132)*x[1]+(-0.610526000157)*x[2]**o+(-0.992542379681)*x[2]
        arg[2,1,0]=(-0.476987072283)*x[0]**o+(-0.450296923812)*x[0]+(0.905270434608)*x[1]**o+(0.976270820516)*x[1]+(-0.233422717373)*x[2]**o+(-0.641253735386)*x[2]
        arg[2,1,1]=(0.873832514958)*x[0]**o+(-0.478059419217)*x[0]+(-0.941650455606)*x[1]**o+(0.307250152442)*x[1]+(0.865160620819)*x[2]**o+(0.961340749451)*x[2]
        arg[3,0,0]=(-0.443837945036)*x[0]**o+(-0.688773058425)*x[0]+(-0.654988618952)*x[1]**o+(-0.259511611372)*x[1]+(-0.931332326494)*x[2]**o+(0.939379631125)*x[2]
        arg[3,0,1]=(-0.181860312567)*x[0]**o+(-0.229066064127)*x[0]+(-0.359408642857)*x[1]**o+(-0.47945907585)*x[1]+(0.312345402782)*x[2]**o+(0.559298232285)*x[2]
        arg[3,1,0]=(-0.372379819015)*x[0]**o+(0.308179383085)*x[0]+(0.123613160006)*x[1]**o+(0.191944822984)*x[1]+(0.950869988808)*x[2]**o+(0.229360419984)*x[2]
        arg[3,1,1]=(-0.803126271269)*x[0]**o+(0.833324260042)*x[0]+(-0.917001746509)*x[1]**o+(-0.649776906898)*x[1]+(0.49624769362)*x[2]**o+(0.414080802594)*x[2]
        arg[4,0,0]=(-0.951747333891)*x[0]**o+(0.5919442178)*x[0]+(0.121310623341)*x[1]**o+(-0.321939845724)*x[1]+(0.358051737083)*x[2]**o+(-0.222801347833)*x[2]
        arg[4,0,1]=(-0.59462534105)*x[0]**o+(0.943860617698)*x[0]+(0.787822231831)*x[1]**o+(-0.0913435772425)*x[1]+(0.831651946729)*x[2]**o+(-0.903604465879)*x[2]
        arg[4,1,0]=(0.68655463207)*x[0]**o+(-0.98918689213)*x[0]+(0.251300089098)*x[1]**o+(-0.317828131349)*x[1]+(-0.105934216764)*x[2]**o+(-0.874506361945)*x[2]
        arg[4,1,1]=(-0.80544616293)*x[0]**o+(-0.321523684723)*x[0]+(0.874455533992)*x[1]**o+(0.200912811442)*x[1]+(-0.646776981862)*x[2]**o+(-0.988728118812)*x[2]
        arg[5,0,0]=(0.413387141328)*x[0]**o+(0.349210032211)*x[0]+(0.80870689028)*x[1]**o+(-0.633296311752)*x[1]+(0.24621029112)*x[2]**o+(0.853398479355)*x[2]
        arg[5,0,1]=(0.398787817372)*x[0]**o+(0.662875239272)*x[0]+(-0.719962467558)*x[1]**o+(0.49562860294)*x[1]+(-0.694859272211)*x[2]**o+(-0.792616661463)*x[2]
        arg[5,1,0]=(-0.298838035128)*x[0]**o+(0.121578680468)*x[0]+(-0.931062588633)*x[1]**o+(-0.0575832984853)*x[1]+(-0.696140085453)*x[2]**o+(0.836738246058)*x[2]
        arg[5,1,1]=(-0.949626960081)*x[0]**o+(-0.449843145228)*x[0]+(-0.304260073311)*x[1]**o+(-0.258880519466)*x[1]+(0.880466679751)*x[2]**o+(0.97505579364)*x[2]
        ref[0,0,0]=(0.369058215952)/(o+1.)+(-0.860734549029)+(0.638389367885)*0.5**o
        ref[0,0,1]=(-0.937845669147)/(o+1.)+(1.14543097594)+(0.859011160063)*0.5**o
        ref[0,1,0]=(-0.725061432954)/(o+1.)+(0.291104548855)+(-0.998681640062)*0.5**o
        ref[0,1,1]=(-0.540637249448)/(o+1.)+(-0.500606905428)+(0.1340261052)*0.5**o
        ref[1,0,0]=(0.902262161102)/(o+1.)+(0.666322787672)+(0.251201824008)*0.5**o
        ref[1,0,1]=(-1.37166876542)/(o+1.)+(-0.452205179789)+(-0.258843593921)*0.5**o
        ref[1,1,0]=(-0.286977110923)/(o+1.)+(0.888993508815)+(-0.854529145157)*0.5**o
        ref[1,1,1]=(0.760163784662)/(o+1.)+(0.525691429317)+(0.636818650335)*0.5**o
        ref[2,0,0]=(0.695754347563)/(o+1.)+(-0.193495064189)+(-0.415847479141)*0.5**o
        ref[2,0,1]=(-1.30411308995)/(o+1.)+(-0.485930391222)+(-0.882383192316)*0.5**o
        ref[2,1,0]=(0.671847717235)/(o+1.)+(-0.0576399193408)+(-0.476987072283)*0.5**o
        ref[2,1,1]=(-0.0764898347868)/(o+1.)+(0.395265741338)+(0.873832514958)*0.5**o
        ref[3,0,0]=(-1.58632094545)/(o+1.)+(-0.00445251933618)+(-0.443837945036)*0.5**o
        ref[3,0,1]=(-0.0470632400747)/(o+1.)+(-0.0746134538458)+(-0.181860312567)*0.5**o
        ref[3,1,0]=(1.07448314881)/(o+1.)+(0.364742313027)+(-0.372379819015)*0.5**o
        ref[3,1,1]=(-0.420754052889)/(o+1.)+(0.298814077869)+(-0.803126271269)*0.5**o
        ref[4,0,0]=(0.479362360424)/(o+1.)+(0.0236015121217)+(-0.951747333891)*0.5**o
        ref[4,0,1]=(1.61947417856)/(o+1.)+(-0.0255437127116)+(-0.59462534105)*0.5**o
        ref[4,1,0]=(0.145365872334)/(o+1.)+(-1.09076069271)+(0.68655463207)*0.5**o
        ref[4,1,1]=(0.22767855213)/(o+1.)+(-0.554669496046)+(-0.80544616293)*0.5**o
        ref[5,0,0]=(1.0549171814)/(o+1.)+(0.284656099907)+(0.413387141328)*0.5**o
        ref[5,0,1]=(-1.41482173977)/(o+1.)+(0.182943590374)+(0.398787817372)*0.5**o
        ref[5,1,0]=(-1.62720267409)/(o+1.)+(0.450366814021)+(-0.298838035128)*0.5**o
        ref[5,1,1]=(0.576206606439)/(o+1.)+(0.133166064473)+(-0.949626960081)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactZero_fromData_FunctionOnContactZero_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactZero

      assumptions: FunctionOnContactZero(self.domain) exists
                   self.domain supports integral on FunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactZero(self.domain)
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.392628958003)*x[0]**o+(0.186084293395)*x[0]+(0.965300679335)*x[1]**o+(0.0986282244747)*x[1]
        arg[0,0,0,1]=(-0.503312425492)*x[0]**o+(-0.205791466802)*x[0]+(0.384782163488)*x[1]**o+(-0.608371097114)*x[1]
        arg[0,0,1,0]=(0.660162511699)*x[0]**o+(-0.263141686561)*x[0]+(-0.765235012795)*x[1]**o+(-0.371233144611)*x[1]
        arg[0,0,1,1]=(-0.986935412732)*x[0]**o+(0.645433351743)*x[0]+(-0.451665346257)*x[1]**o+(-0.713815759872)*x[1]
        arg[0,0,2,0]=(0.632875178502)*x[0]**o+(-0.257542208821)*x[0]+(-0.695553902133)*x[1]**o+(-0.677920876645)*x[1]
        arg[0,0,2,1]=(0.102625648731)*x[0]**o+(0.511896902175)*x[0]+(0.345909088853)*x[1]**o+(-0.0609413722414)*x[1]
        arg[0,1,0,0]=(-0.0108106931425)*x[0]**o+(-0.595497517075)*x[0]+(0.332480141972)*x[1]**o+(-0.18131109924)*x[1]
        arg[0,1,0,1]=(0.220826593491)*x[0]**o+(-0.202422451222)*x[0]+(0.328480565578)*x[1]**o+(0.851432750557)*x[1]
        arg[0,1,1,0]=(0.22735402466)*x[0]**o+(0.51437836535)*x[0]+(0.292927451477)*x[1]**o+(-0.148121856023)*x[1]
        arg[0,1,1,1]=(0.495079875966)*x[0]**o+(0.582755117761)*x[0]+(-0.322061131008)*x[1]**o+(0.953437449986)*x[1]
        arg[0,1,2,0]=(0.968583889554)*x[0]**o+(-0.593253483731)*x[0]+(0.286037351205)*x[1]**o+(0.997314411263)*x[1]
        arg[0,1,2,1]=(0.361498191707)*x[0]**o+(0.633275777697)*x[0]+(0.891305263219)*x[1]**o+(-0.194468605638)*x[1]
        arg[0,2,0,0]=(0.185647566516)*x[0]**o+(0.101448861247)*x[0]+(0.0870608077911)*x[1]**o+(-0.86477908948)*x[1]
        arg[0,2,0,1]=(0.227262327826)*x[0]**o+(0.812586702875)*x[0]+(-0.0229924234496)*x[1]**o+(-0.323581477982)*x[1]
        arg[0,2,1,0]=(0.0405753322301)*x[0]**o+(0.891289727227)*x[0]+(0.957377329048)*x[1]**o+(-0.982119724379)*x[1]
        arg[0,2,1,1]=(-0.546046301701)*x[0]**o+(0.628511534929)*x[0]+(-0.628262263897)*x[1]**o+(-0.454315708999)*x[1]
        arg[0,2,2,0]=(-0.828455793175)*x[0]**o+(0.190147126828)*x[0]+(0.499595322279)*x[1]**o+(-0.0882297087937)*x[1]
        arg[0,2,2,1]=(0.926138085222)*x[0]**o+(0.402659001583)*x[0]+(-0.678432418122)*x[1]**o+(0.236054704896)*x[1]
        arg[0,3,0,0]=(-0.473066058658)*x[0]**o+(0.57622727181)*x[0]+(0.62665109546)*x[1]**o+(0.948409659533)*x[1]
        arg[0,3,0,1]=(0.776790978673)*x[0]**o+(-0.0244752180934)*x[0]+(-0.459252843816)*x[1]**o+(0.616203849528)*x[1]
        arg[0,3,1,0]=(0.34783214571)*x[0]**o+(0.806528120142)*x[0]+(0.445864469556)*x[1]**o+(-0.221961922099)*x[1]
        arg[0,3,1,1]=(0.0717966037463)*x[0]**o+(-0.434117388366)*x[0]+(0.577708844669)*x[1]**o+(-0.548519157145)*x[1]
        arg[0,3,2,0]=(-0.548573971799)*x[0]**o+(-0.926704595492)*x[0]+(-0.33936340707)*x[1]**o+(0.0687290934591)*x[1]
        arg[0,3,2,1]=(-0.28309240461)*x[0]**o+(0.12914675768)*x[0]+(-0.833323497891)*x[1]**o+(-0.577727985559)*x[1]
        arg[0,4,0,0]=(-0.105416428922)*x[0]**o+(0.41564239907)*x[0]+(0.146657173319)*x[1]**o+(-0.0141417271845)*x[1]
        arg[0,4,0,1]=(-0.876111094069)*x[0]**o+(0.491957669628)*x[0]+(0.534941970665)*x[1]**o+(-0.995507289274)*x[1]
        arg[0,4,1,0]=(-0.678996394303)*x[0]**o+(0.183159947613)*x[0]+(-0.739368558349)*x[1]**o+(0.274319401135)*x[1]
        arg[0,4,1,1]=(0.0472145581152)*x[0]**o+(-0.669737608207)*x[0]+(-0.651647023142)*x[1]**o+(0.351208329111)*x[1]
        arg[0,4,2,0]=(0.114936857967)*x[0]**o+(0.564688790898)*x[0]+(-0.524234284672)*x[1]**o+(-0.127615425724)*x[1]
        arg[0,4,2,1]=(0.568834117773)*x[0]**o+(-0.936513147599)*x[0]+(0.133122220811)*x[1]**o+(0.31510090512)*x[1]
        arg[1,0,0,0]=(-0.821241886235)*x[0]**o+(0.779734235239)*x[0]+(-0.132449009841)*x[1]**o+(-0.0179700618848)*x[1]
        arg[1,0,0,1]=(0.624372422235)*x[0]**o+(-0.752831611955)*x[0]+(-0.573597121344)*x[1]**o+(0.376463665377)*x[1]
        arg[1,0,1,0]=(-0.568749808355)*x[0]**o+(-0.697521100346)*x[0]+(-0.249786553024)*x[1]**o+(-0.624593568423)*x[1]
        arg[1,0,1,1]=(0.0765557360835)*x[0]**o+(-0.744005823966)*x[0]+(0.59246455477)*x[1]**o+(0.231310687774)*x[1]
        arg[1,0,2,0]=(0.909614864187)*x[0]**o+(0.350543076144)*x[0]+(0.0689432766231)*x[1]**o+(-0.88465698758)*x[1]
        arg[1,0,2,1]=(0.0610206203468)*x[0]**o+(0.620401328094)*x[0]+(0.378273922837)*x[1]**o+(0.91988744934)*x[1]
        arg[1,1,0,0]=(0.917249169051)*x[0]**o+(0.756656473206)*x[0]+(0.728786835113)*x[1]**o+(0.116994159781)*x[1]
        arg[1,1,0,1]=(0.331129708654)*x[0]**o+(-0.14342302905)*x[0]+(-0.418402258918)*x[1]**o+(-0.808676850609)*x[1]
        arg[1,1,1,0]=(-0.527408405279)*x[0]**o+(-0.671440239997)*x[0]+(0.989603613532)*x[1]**o+(0.364280392731)*x[1]
        arg[1,1,1,1]=(-0.189543145217)*x[0]**o+(0.14762392216)*x[0]+(0.11459047938)*x[1]**o+(-0.760477267267)*x[1]
        arg[1,1,2,0]=(0.385274935809)*x[0]**o+(0.527744200995)*x[0]+(-0.862654310338)*x[1]**o+(0.904005467598)*x[1]
        arg[1,1,2,1]=(-0.380421555378)*x[0]**o+(0.598300323394)*x[0]+(-0.541819816686)*x[1]**o+(0.446934572528)*x[1]
        arg[1,2,0,0]=(0.446781075767)*x[0]**o+(0.906830344178)*x[0]+(0.158734254777)*x[1]**o+(-0.722982586936)*x[1]
        arg[1,2,0,1]=(0.927811104869)*x[0]**o+(-0.446795135388)*x[0]+(-0.90301726295)*x[1]**o+(0.289530807979)*x[1]
        arg[1,2,1,0]=(0.741527109588)*x[0]**o+(-0.370029728344)*x[0]+(0.157695458386)*x[1]**o+(-0.620804465293)*x[1]
        arg[1,2,1,1]=(0.193242816889)*x[0]**o+(0.455789913444)*x[0]+(0.0491001091271)*x[1]**o+(-0.942127264151)*x[1]
        arg[1,2,2,0]=(0.220709813709)*x[0]**o+(-0.713472965425)*x[0]+(-0.736262418061)*x[1]**o+(0.341347605023)*x[1]
        arg[1,2,2,1]=(0.0034506429721)*x[0]**o+(0.397289753453)*x[0]+(-0.167013881353)*x[1]**o+(0.390364445974)*x[1]
        arg[1,3,0,0]=(0.321471168172)*x[0]**o+(-0.237896433618)*x[0]+(-0.199843421009)*x[1]**o+(0.811319091471)*x[1]
        arg[1,3,0,1]=(-0.0799622908037)*x[0]**o+(-0.66155883346)*x[0]+(0.0303112997369)*x[1]**o+(-0.199256497975)*x[1]
        arg[1,3,1,0]=(-0.12051943489)*x[0]**o+(0.674374603779)*x[0]+(0.335933608833)*x[1]**o+(0.366671959553)*x[1]
        arg[1,3,1,1]=(-0.587242845986)*x[0]**o+(-0.495652656487)*x[0]+(0.10609614346)*x[1]**o+(0.177210588935)*x[1]
        arg[1,3,2,0]=(0.947782336638)*x[0]**o+(-0.327283099142)*x[0]+(-0.0280648836948)*x[1]**o+(0.133334543825)*x[1]
        arg[1,3,2,1]=(-0.0887414134508)*x[0]**o+(0.640694455701)*x[0]+(0.266916366237)*x[1]**o+(0.345886606493)*x[1]
        arg[1,4,0,0]=(-0.138184872938)*x[0]**o+(0.658635897645)*x[0]+(-0.00526240442626)*x[1]**o+(0.800443113371)*x[1]
        arg[1,4,0,1]=(-0.549831917928)*x[0]**o+(-0.666733703837)*x[0]+(0.400982699808)*x[1]**o+(0.739099662817)*x[1]
        arg[1,4,1,0]=(-0.137046722563)*x[0]**o+(-0.818450480933)*x[0]+(-0.827665726293)*x[1]**o+(-0.286585607624)*x[1]
        arg[1,4,1,1]=(-0.474156202572)*x[0]**o+(-0.855381707793)*x[0]+(0.113318564035)*x[1]**o+(0.218410511108)*x[1]
        arg[1,4,2,0]=(0.18476112439)*x[0]**o+(-0.964533442641)*x[0]+(-0.968611931598)*x[1]**o+(0.0973026144203)*x[1]
        arg[1,4,2,1]=(-0.372995418552)*x[0]**o+(0.998322146553)*x[0]+(-0.384672070744)*x[1]**o+(0.47536977268)*x[1]
        arg[2,0,0,0]=(0.882400857222)*x[0]**o+(0.455324640128)*x[0]+(0.764297272234)*x[1]**o+(0.498669000791)*x[1]
        arg[2,0,0,1]=(0.894017590794)*x[0]**o+(0.217984119634)*x[0]+(-0.0934000967188)*x[1]**o+(-0.886960376026)*x[1]
        arg[2,0,1,0]=(-0.382158057826)*x[0]**o+(-0.897639065996)*x[0]+(0.819002513335)*x[1]**o+(-0.0739374877518)*x[1]
        arg[2,0,1,1]=(-0.259771272131)*x[0]**o+(-0.533540086977)*x[0]+(0.366917477923)*x[1]**o+(-0.751904525722)*x[1]
        arg[2,0,2,0]=(-0.33946164111)*x[0]**o+(-0.317441178182)*x[0]+(-0.857625945147)*x[1]**o+(-0.84647892719)*x[1]
        arg[2,0,2,1]=(0.153353093148)*x[0]**o+(0.832883645984)*x[0]+(-0.359476461654)*x[1]**o+(0.332390899654)*x[1]
        arg[2,1,0,0]=(0.780980225775)*x[0]**o+(0.619330950566)*x[0]+(0.194312746335)*x[1]**o+(0.595311976655)*x[1]
        arg[2,1,0,1]=(-0.334149973283)*x[0]**o+(-0.670178805858)*x[0]+(-0.313038325156)*x[1]**o+(0.793149618246)*x[1]
        arg[2,1,1,0]=(-0.832852528577)*x[0]**o+(-0.606634692926)*x[0]+(-0.0643816133251)*x[1]**o+(-0.663765987285)*x[1]
        arg[2,1,1,1]=(0.300975732963)*x[0]**o+(0.965431406497)*x[0]+(-0.618456119184)*x[1]**o+(0.459530421885)*x[1]
        arg[2,1,2,0]=(0.122326105776)*x[0]**o+(0.0940802696378)*x[0]+(-0.986852501828)*x[1]**o+(0.370428043236)*x[1]
        arg[2,1,2,1]=(0.447916985436)*x[0]**o+(-0.161926311978)*x[0]+(-0.0933563424166)*x[1]**o+(0.836290263421)*x[1]
        arg[2,2,0,0]=(0.032076347277)*x[0]**o+(-0.447517143643)*x[0]+(-0.053769157756)*x[1]**o+(-0.0242981207801)*x[1]
        arg[2,2,0,1]=(-0.403935703163)*x[0]**o+(-0.960750093184)*x[0]+(-0.287956277998)*x[1]**o+(-0.168614701082)*x[1]
        arg[2,2,1,0]=(-0.205077009492)*x[0]**o+(-0.409536289697)*x[0]+(0.359071620541)*x[1]**o+(-0.466455759025)*x[1]
        arg[2,2,1,1]=(0.443658057141)*x[0]**o+(-0.748681939527)*x[0]+(0.2153167916)*x[1]**o+(-0.475899799738)*x[1]
        arg[2,2,2,0]=(-0.463879681251)*x[0]**o+(0.758317222041)*x[0]+(0.608221916958)*x[1]**o+(-0.237798399085)*x[1]
        arg[2,2,2,1]=(0.533981754168)*x[0]**o+(0.590125912931)*x[0]+(0.8957267808)*x[1]**o+(0.294612478767)*x[1]
        arg[2,3,0,0]=(0.499387701403)*x[0]**o+(0.0379132293822)*x[0]+(0.717384215206)*x[1]**o+(0.754497698503)*x[1]
        arg[2,3,0,1]=(0.866500398882)*x[0]**o+(0.23678788012)*x[0]+(0.395132533716)*x[1]**o+(-0.451619930593)*x[1]
        arg[2,3,1,0]=(-0.879663354176)*x[0]**o+(-0.328937922978)*x[0]+(0.688771917581)*x[1]**o+(-0.87392051489)*x[1]
        arg[2,3,1,1]=(0.291214284356)*x[0]**o+(-0.783159133488)*x[0]+(0.240960084233)*x[1]**o+(0.339043900275)*x[1]
        arg[2,3,2,0]=(0.47064755192)*x[0]**o+(-0.42618815064)*x[0]+(-0.505342655776)*x[1]**o+(0.408218280069)*x[1]
        arg[2,3,2,1]=(0.775731405735)*x[0]**o+(-0.768750178591)*x[0]+(-0.896333275063)*x[1]**o+(0.286558848957)*x[1]
        arg[2,4,0,0]=(-0.7221248688)*x[0]**o+(0.612509823967)*x[0]+(0.436744528143)*x[1]**o+(-0.761502758764)*x[1]
        arg[2,4,0,1]=(-0.742652222385)*x[0]**o+(0.214811997988)*x[0]+(-0.031320962337)*x[1]**o+(-0.320134664789)*x[1]
        arg[2,4,1,0]=(-0.529575527815)*x[0]**o+(0.623770482091)*x[0]+(-0.50790148262)*x[1]**o+(0.629874274271)*x[1]
        arg[2,4,1,1]=(0.535631957883)*x[0]**o+(0.439578618993)*x[0]+(-0.770747718564)*x[1]**o+(0.0798696116058)*x[1]
        arg[2,4,2,0]=(0.81688202721)*x[0]**o+(-0.0301054197201)*x[0]+(0.613495031968)*x[1]**o+(-0.153619455295)*x[1]
        arg[2,4,2,1]=(0.493471708612)*x[0]**o+(0.423626757568)*x[0]+(0.093174141001)*x[1]**o+(0.307452108722)*x[1]
        arg[3,0,0,0]=(-0.0668630243284)*x[0]**o+(-0.577709692979)*x[0]+(0.984307685082)*x[1]**o+(-0.171487030152)*x[1]
        arg[3,0,0,1]=(-0.210452484293)*x[0]**o+(-0.130476372055)*x[0]+(-0.693596086089)*x[1]**o+(-0.692143724329)*x[1]
        arg[3,0,1,0]=(-0.34853285523)*x[0]**o+(-0.369649709966)*x[0]+(0.824648362105)*x[1]**o+(0.733046240399)*x[1]
        arg[3,0,1,1]=(0.861903045965)*x[0]**o+(0.609535692452)*x[0]+(0.681514895836)*x[1]**o+(0.66773275418)*x[1]
        arg[3,0,2,0]=(0.0197670342844)*x[0]**o+(-0.251645539228)*x[0]+(-0.0390293468405)*x[1]**o+(-0.00955605638373)*x[1]
        arg[3,0,2,1]=(-0.499093940395)*x[0]**o+(0.660356623112)*x[0]+(0.969057635866)*x[1]**o+(-0.821377317197)*x[1]
        arg[3,1,0,0]=(-0.955568800095)*x[0]**o+(0.557877698917)*x[0]+(0.829599262959)*x[1]**o+(0.933736802562)*x[1]
        arg[3,1,0,1]=(-0.910307638058)*x[0]**o+(-0.3287646946)*x[0]+(-0.777548579252)*x[1]**o+(-0.971744353969)*x[1]
        arg[3,1,1,0]=(0.22353820834)*x[0]**o+(0.904479243377)*x[0]+(-0.108768166204)*x[1]**o+(0.823767232517)*x[1]
        arg[3,1,1,1]=(0.371995768061)*x[0]**o+(0.747630377273)*x[0]+(0.931152588498)*x[1]**o+(-0.0154330268188)*x[1]
        arg[3,1,2,0]=(0.4567417369)*x[0]**o+(0.424951293892)*x[0]+(-0.14032259215)*x[1]**o+(0.603329614924)*x[1]
        arg[3,1,2,1]=(-0.855100213731)*x[0]**o+(-0.667832882495)*x[0]+(0.942530922378)*x[1]**o+(0.55861156259)*x[1]
        arg[3,2,0,0]=(0.17190936572)*x[0]**o+(-0.904098615762)*x[0]+(0.991497679624)*x[1]**o+(0.373020358381)*x[1]
        arg[3,2,0,1]=(-0.683587206983)*x[0]**o+(0.718257814718)*x[0]+(-0.730243933765)*x[1]**o+(-0.583892189979)*x[1]
        arg[3,2,1,0]=(0.051672736629)*x[0]**o+(-0.991172563807)*x[0]+(0.36605487466)*x[1]**o+(0.680719877794)*x[1]
        arg[3,2,1,1]=(0.438014662973)*x[0]**o+(0.172969450125)*x[0]+(-0.424515284177)*x[1]**o+(0.615315132547)*x[1]
        arg[3,2,2,0]=(-0.709499532674)*x[0]**o+(0.985097212621)*x[0]+(0.326610298255)*x[1]**o+(0.994701669258)*x[1]
        arg[3,2,2,1]=(-0.657647707844)*x[0]**o+(-0.683954483768)*x[0]+(-0.846657945161)*x[1]**o+(0.852308904063)*x[1]
        arg[3,3,0,0]=(0.21346320453)*x[0]**o+(-0.181665119768)*x[0]+(0.549866100727)*x[1]**o+(-0.0858045316871)*x[1]
        arg[3,3,0,1]=(0.249957470685)*x[0]**o+(0.537464773638)*x[0]+(0.0224330511636)*x[1]**o+(0.114623595638)*x[1]
        arg[3,3,1,0]=(0.166507029533)*x[0]**o+(0.232150195873)*x[0]+(0.119453667263)*x[1]**o+(-0.261110960642)*x[1]
        arg[3,3,1,1]=(0.358248720077)*x[0]**o+(-0.871911261059)*x[0]+(0.57743453064)*x[1]**o+(0.950322066335)*x[1]
        arg[3,3,2,0]=(-0.896899724447)*x[0]**o+(-0.898884767716)*x[0]+(0.63832012742)*x[1]**o+(-0.833800375634)*x[1]
        arg[3,3,2,1]=(0.425194254583)*x[0]**o+(0.161328467295)*x[0]+(0.496219405916)*x[1]**o+(0.380935896093)*x[1]
        arg[3,4,0,0]=(0.793902581592)*x[0]**o+(-0.33428511164)*x[0]+(-0.283510438468)*x[1]**o+(0.228913046074)*x[1]
        arg[3,4,0,1]=(0.372028980539)*x[0]**o+(-0.388909011671)*x[0]+(-0.259885631206)*x[1]**o+(-0.950910172897)*x[1]
        arg[3,4,1,0]=(-0.80489673786)*x[0]**o+(-0.828175420038)*x[0]+(0.895070489116)*x[1]**o+(-0.742272132103)*x[1]
        arg[3,4,1,1]=(0.272535126101)*x[0]**o+(0.902668974891)*x[0]+(0.399816667015)*x[1]**o+(-0.953186908789)*x[1]
        arg[3,4,2,0]=(0.0341775288582)*x[0]**o+(0.962241492639)*x[0]+(0.906020184876)*x[1]**o+(-0.400536255726)*x[1]
        arg[3,4,2,1]=(-0.745016464485)*x[0]**o+(-0.85241829747)*x[0]+(-0.648936544722)*x[1]**o+(-0.997030074218)*x[1]
        ref[0,0,0,0]=(0.965300679335)/(o+1.)+(0.142356258935)+(0.392628958003)*0.5**o
        ref[0,0,0,1]=(0.384782163488)/(o+1.)+(-0.407081281958)+(-0.503312425492)*0.5**o
        ref[0,0,1,0]=(-0.765235012795)/(o+1.)+(-0.317187415586)+(0.660162511699)*0.5**o
        ref[0,0,1,1]=(-0.451665346257)/(o+1.)+(-0.0341912040643)+(-0.986935412732)*0.5**o
        ref[0,0,2,0]=(-0.695553902133)/(o+1.)+(-0.467731542733)+(0.632875178502)*0.5**o
        ref[0,0,2,1]=(0.345909088853)/(o+1.)+(0.225477764967)+(0.102625648731)*0.5**o
        ref[0,1,0,0]=(0.332480141972)/(o+1.)+(-0.388404308158)+(-0.0108106931425)*0.5**o
        ref[0,1,0,1]=(0.328480565578)/(o+1.)+(0.324505149668)+(0.220826593491)*0.5**o
        ref[0,1,1,0]=(0.292927451477)/(o+1.)+(0.183128254664)+(0.22735402466)*0.5**o
        ref[0,1,1,1]=(-0.322061131008)/(o+1.)+(0.768096283874)+(0.495079875966)*0.5**o
        ref[0,1,2,0]=(0.286037351205)/(o+1.)+(0.202030463766)+(0.968583889554)*0.5**o
        ref[0,1,2,1]=(0.891305263219)/(o+1.)+(0.219403586029)+(0.361498191707)*0.5**o
        ref[0,2,0,0]=(0.0870608077911)/(o+1.)+(-0.381665114117)+(0.185647566516)*0.5**o
        ref[0,2,0,1]=(-0.0229924234496)/(o+1.)+(0.244502612446)+(0.227262327826)*0.5**o
        ref[0,2,1,0]=(0.957377329048)/(o+1.)+(-0.0454149985758)+(0.0405753322301)*0.5**o
        ref[0,2,1,1]=(-0.628262263897)/(o+1.)+(0.0870979129647)+(-0.546046301701)*0.5**o
        ref[0,2,2,0]=(0.499595322279)/(o+1.)+(0.0509587090171)+(-0.828455793175)*0.5**o
        ref[0,2,2,1]=(-0.678432418122)/(o+1.)+(0.319356853239)+(0.926138085222)*0.5**o
        ref[0,3,0,0]=(0.62665109546)/(o+1.)+(0.762318465671)+(-0.473066058658)*0.5**o
        ref[0,3,0,1]=(-0.459252843816)/(o+1.)+(0.295864315717)+(0.776790978673)*0.5**o
        ref[0,3,1,0]=(0.445864469556)/(o+1.)+(0.292283099022)+(0.34783214571)*0.5**o
        ref[0,3,1,1]=(0.577708844669)/(o+1.)+(-0.491318272755)+(0.0717966037463)*0.5**o
        ref[0,3,2,0]=(-0.33936340707)/(o+1.)+(-0.428987751016)+(-0.548573971799)*0.5**o
        ref[0,3,2,1]=(-0.833323497891)/(o+1.)+(-0.224290613939)+(-0.28309240461)*0.5**o
        ref[0,4,0,0]=(0.146657173319)/(o+1.)+(0.200750335943)+(-0.105416428922)*0.5**o
        ref[0,4,0,1]=(0.534941970665)/(o+1.)+(-0.251774809823)+(-0.876111094069)*0.5**o
        ref[0,4,1,0]=(-0.739368558349)/(o+1.)+(0.228739674374)+(-0.678996394303)*0.5**o
        ref[0,4,1,1]=(-0.651647023142)/(o+1.)+(-0.159264639548)+(0.0472145581152)*0.5**o
        ref[0,4,2,0]=(-0.524234284672)/(o+1.)+(0.218536682587)+(0.114936857967)*0.5**o
        ref[0,4,2,1]=(0.133122220811)/(o+1.)+(-0.31070612124)+(0.568834117773)*0.5**o
        ref[1,0,0,0]=(-0.132449009841)/(o+1.)+(0.380882086677)+(-0.821241886235)*0.5**o
        ref[1,0,0,1]=(-0.573597121344)/(o+1.)+(-0.188183973289)+(0.624372422235)*0.5**o
        ref[1,0,1,0]=(-0.249786553024)/(o+1.)+(-0.661057334385)+(-0.568749808355)*0.5**o
        ref[1,0,1,1]=(0.59246455477)/(o+1.)+(-0.256347568096)+(0.0765557360835)*0.5**o
        ref[1,0,2,0]=(0.0689432766231)/(o+1.)+(-0.267056955718)+(0.909614864187)*0.5**o
        ref[1,0,2,1]=(0.378273922837)/(o+1.)+(0.770144388717)+(0.0610206203468)*0.5**o
        ref[1,1,0,0]=(0.728786835113)/(o+1.)+(0.436825316493)+(0.917249169051)*0.5**o
        ref[1,1,0,1]=(-0.418402258918)/(o+1.)+(-0.47604993983)+(0.331129708654)*0.5**o
        ref[1,1,1,0]=(0.989603613532)/(o+1.)+(-0.153579923633)+(-0.527408405279)*0.5**o
        ref[1,1,1,1]=(0.11459047938)/(o+1.)+(-0.306426672553)+(-0.189543145217)*0.5**o
        ref[1,1,2,0]=(-0.862654310338)/(o+1.)+(0.715874834296)+(0.385274935809)*0.5**o
        ref[1,1,2,1]=(-0.541819816686)/(o+1.)+(0.522617447961)+(-0.380421555378)*0.5**o
        ref[1,2,0,0]=(0.158734254777)/(o+1.)+(0.0919238786215)+(0.446781075767)*0.5**o
        ref[1,2,0,1]=(-0.90301726295)/(o+1.)+(-0.0786321637044)+(0.927811104869)*0.5**o
        ref[1,2,1,0]=(0.157695458386)/(o+1.)+(-0.495417096819)+(0.741527109588)*0.5**o
        ref[1,2,1,1]=(0.0491001091271)/(o+1.)+(-0.243168675353)+(0.193242816889)*0.5**o
        ref[1,2,2,0]=(-0.736262418061)/(o+1.)+(-0.186062680201)+(0.220709813709)*0.5**o
        ref[1,2,2,1]=(-0.167013881353)/(o+1.)+(0.393827099713)+(0.0034506429721)*0.5**o
        ref[1,3,0,0]=(-0.199843421009)/(o+1.)+(0.286711328927)+(0.321471168172)*0.5**o
        ref[1,3,0,1]=(0.0303112997369)/(o+1.)+(-0.430407665717)+(-0.0799622908037)*0.5**o
        ref[1,3,1,0]=(0.335933608833)/(o+1.)+(0.520523281666)+(-0.12051943489)*0.5**o
        ref[1,3,1,1]=(0.10609614346)/(o+1.)+(-0.159221033776)+(-0.587242845986)*0.5**o
        ref[1,3,2,0]=(-0.0280648836948)/(o+1.)+(-0.0969742776586)+(0.947782336638)*0.5**o
        ref[1,3,2,1]=(0.266916366237)/(o+1.)+(0.493290531097)+(-0.0887414134508)*0.5**o
        ref[1,4,0,0]=(-0.00526240442626)/(o+1.)+(0.729539505508)+(-0.138184872938)*0.5**o
        ref[1,4,0,1]=(0.400982699808)/(o+1.)+(0.0361829794896)+(-0.549831917928)*0.5**o
        ref[1,4,1,0]=(-0.827665726293)/(o+1.)+(-0.552518044279)+(-0.137046722563)*0.5**o
        ref[1,4,1,1]=(0.113318564035)/(o+1.)+(-0.318485598343)+(-0.474156202572)*0.5**o
        ref[1,4,2,0]=(-0.968611931598)/(o+1.)+(-0.433615414111)+(0.18476112439)*0.5**o
        ref[1,4,2,1]=(-0.384672070744)/(o+1.)+(0.736845959617)+(-0.372995418552)*0.5**o
        ref[2,0,0,0]=(0.764297272234)/(o+1.)+(0.476996820459)+(0.882400857222)*0.5**o
        ref[2,0,0,1]=(-0.0934000967188)/(o+1.)+(-0.334488128196)+(0.894017590794)*0.5**o
        ref[2,0,1,0]=(0.819002513335)/(o+1.)+(-0.485788276874)+(-0.382158057826)*0.5**o
        ref[2,0,1,1]=(0.366917477923)/(o+1.)+(-0.64272230635)+(-0.259771272131)*0.5**o
        ref[2,0,2,0]=(-0.857625945147)/(o+1.)+(-0.581960052686)+(-0.33946164111)*0.5**o
        ref[2,0,2,1]=(-0.359476461654)/(o+1.)+(0.582637272819)+(0.153353093148)*0.5**o
        ref[2,1,0,0]=(0.194312746335)/(o+1.)+(0.607321463611)+(0.780980225775)*0.5**o
        ref[2,1,0,1]=(-0.313038325156)/(o+1.)+(0.0614854061938)+(-0.334149973283)*0.5**o
        ref[2,1,1,0]=(-0.0643816133251)/(o+1.)+(-0.635200340106)+(-0.832852528577)*0.5**o
        ref[2,1,1,1]=(-0.618456119184)/(o+1.)+(0.712480914191)+(0.300975732963)*0.5**o
        ref[2,1,2,0]=(-0.986852501828)/(o+1.)+(0.232254156437)+(0.122326105776)*0.5**o
        ref[2,1,2,1]=(-0.0933563424166)/(o+1.)+(0.337181975722)+(0.447916985436)*0.5**o
        ref[2,2,0,0]=(-0.053769157756)/(o+1.)+(-0.235907632211)+(0.032076347277)*0.5**o
        ref[2,2,0,1]=(-0.287956277998)/(o+1.)+(-0.564682397133)+(-0.403935703163)*0.5**o
        ref[2,2,1,0]=(0.359071620541)/(o+1.)+(-0.437996024361)+(-0.205077009492)*0.5**o
        ref[2,2,1,1]=(0.2153167916)/(o+1.)+(-0.612290869632)+(0.443658057141)*0.5**o
        ref[2,2,2,0]=(0.608221916958)/(o+1.)+(0.260259411478)+(-0.463879681251)*0.5**o
        ref[2,2,2,1]=(0.8957267808)/(o+1.)+(0.442369195849)+(0.533981754168)*0.5**o
        ref[2,3,0,0]=(0.717384215206)/(o+1.)+(0.396205463942)+(0.499387701403)*0.5**o
        ref[2,3,0,1]=(0.395132533716)/(o+1.)+(-0.107416025237)+(0.866500398882)*0.5**o
        ref[2,3,1,0]=(0.688771917581)/(o+1.)+(-0.601429218934)+(-0.879663354176)*0.5**o
        ref[2,3,1,1]=(0.240960084233)/(o+1.)+(-0.222057616607)+(0.291214284356)*0.5**o
        ref[2,3,2,0]=(-0.505342655776)/(o+1.)+(-0.00898493528542)+(0.47064755192)*0.5**o
        ref[2,3,2,1]=(-0.896333275063)/(o+1.)+(-0.241095664817)+(0.775731405735)*0.5**o
        ref[2,4,0,0]=(0.436744528143)/(o+1.)+(-0.0744964673986)+(-0.7221248688)*0.5**o
        ref[2,4,0,1]=(-0.031320962337)/(o+1.)+(-0.0526613334003)+(-0.742652222385)*0.5**o
        ref[2,4,1,0]=(-0.50790148262)/(o+1.)+(0.626822378181)+(-0.529575527815)*0.5**o
        ref[2,4,1,1]=(-0.770747718564)/(o+1.)+(0.2597241153)+(0.535631957883)*0.5**o
        ref[2,4,2,0]=(0.613495031968)/(o+1.)+(-0.0918624375077)+(0.81688202721)*0.5**o
        ref[2,4,2,1]=(0.093174141001)/(o+1.)+(0.365539433145)+(0.493471708612)*0.5**o
        ref[3,0,0,0]=(0.984307685082)/(o+1.)+(-0.374598361566)+(-0.0668630243284)*0.5**o
        ref[3,0,0,1]=(-0.693596086089)/(o+1.)+(-0.411310048192)+(-0.210452484293)*0.5**o
        ref[3,0,1,0]=(0.824648362105)/(o+1.)+(0.181698265217)+(-0.34853285523)*0.5**o
        ref[3,0,1,1]=(0.681514895836)/(o+1.)+(0.638634223316)+(0.861903045965)*0.5**o
        ref[3,0,2,0]=(-0.0390293468405)/(o+1.)+(-0.130600797806)+(0.0197670342844)*0.5**o
        ref[3,0,2,1]=(0.969057635866)/(o+1.)+(-0.0805103470423)+(-0.499093940395)*0.5**o
        ref[3,1,0,0]=(0.829599262959)/(o+1.)+(0.745807250739)+(-0.955568800095)*0.5**o
        ref[3,1,0,1]=(-0.777548579252)/(o+1.)+(-0.650254524284)+(-0.910307638058)*0.5**o
        ref[3,1,1,0]=(-0.108768166204)/(o+1.)+(0.864123237947)+(0.22353820834)*0.5**o
        ref[3,1,1,1]=(0.931152588498)/(o+1.)+(0.366098675227)+(0.371995768061)*0.5**o
        ref[3,1,2,0]=(-0.14032259215)/(o+1.)+(0.514140454408)+(0.4567417369)*0.5**o
        ref[3,1,2,1]=(0.942530922378)/(o+1.)+(-0.0546106599524)+(-0.855100213731)*0.5**o
        ref[3,2,0,0]=(0.991497679624)/(o+1.)+(-0.26553912869)+(0.17190936572)*0.5**o
        ref[3,2,0,1]=(-0.730243933765)/(o+1.)+(0.0671828123695)+(-0.683587206983)*0.5**o
        ref[3,2,1,0]=(0.36605487466)/(o+1.)+(-0.155226343006)+(0.051672736629)*0.5**o
        ref[3,2,1,1]=(-0.424515284177)/(o+1.)+(0.394142291336)+(0.438014662973)*0.5**o
        ref[3,2,2,0]=(0.326610298255)/(o+1.)+(0.98989944094)+(-0.709499532674)*0.5**o
        ref[3,2,2,1]=(-0.846657945161)/(o+1.)+(0.0841772101477)+(-0.657647707844)*0.5**o
        ref[3,3,0,0]=(0.549866100727)/(o+1.)+(-0.133734825728)+(0.21346320453)*0.5**o
        ref[3,3,0,1]=(0.0224330511636)/(o+1.)+(0.326044184638)+(0.249957470685)*0.5**o
        ref[3,3,1,0]=(0.119453667263)/(o+1.)+(-0.0144803823844)+(0.166507029533)*0.5**o
        ref[3,3,1,1]=(0.57743453064)/(o+1.)+(0.0392054026384)+(0.358248720077)*0.5**o
        ref[3,3,2,0]=(0.63832012742)/(o+1.)+(-0.866342571675)+(-0.896899724447)*0.5**o
        ref[3,3,2,1]=(0.496219405916)/(o+1.)+(0.271132181694)+(0.425194254583)*0.5**o
        ref[3,4,0,0]=(-0.283510438468)/(o+1.)+(-0.0526860327828)+(0.793902581592)*0.5**o
        ref[3,4,0,1]=(-0.259885631206)/(o+1.)+(-0.669909592284)+(0.372028980539)*0.5**o
        ref[3,4,1,0]=(0.895070489116)/(o+1.)+(-0.78522377607)+(-0.80489673786)*0.5**o
        ref[3,4,1,1]=(0.399816667015)/(o+1.)+(-0.0252589669489)+(0.272535126101)*0.5**o
        ref[3,4,2,0]=(0.906020184876)/(o+1.)+(0.280852618457)+(0.0341775288582)*0.5**o
        ref[3,4,2,1]=(-0.648936544722)/(o+1.)+(-0.924724185844)+(-0.745016464485)*0.5**o
      else:
        arg[0,0,0,0]=(-0.891975745754)*x[0]**o+(0.678331209591)*x[0]+(-0.450022305538)*x[1]**o+(0.521272586024)*x[1]+(-0.28868708769)*x[2]**o+(-0.631617544339)*x[2]
        arg[0,0,0,1]=(0.428698930349)*x[0]**o+(-0.856251438077)*x[0]+(0.915766367729)*x[1]**o+(-0.35019463199)*x[1]+(-0.0029340113102)*x[2]**o+(-0.394458505835)*x[2]
        arg[0,0,1,0]=(0.974069118653)*x[0]**o+(-0.674028781962)*x[0]+(0.489718068973)*x[1]**o+(-0.471323892144)*x[1]+(0.11469271678)*x[2]**o+(-0.264584383354)*x[2]
        arg[0,0,1,1]=(0.813412823502)*x[0]**o+(0.739276532652)*x[0]+(-0.863216628203)*x[1]**o+(0.486539683681)*x[1]+(-0.771448803712)*x[2]**o+(-0.621526889056)*x[2]
        arg[0,0,2,0]=(-0.874502916672)*x[0]**o+(0.0708047915752)*x[0]+(0.917576817333)*x[1]**o+(-0.257626031306)*x[1]+(-0.161822453369)*x[2]**o+(0.488866912786)*x[2]
        arg[0,0,2,1]=(-0.258524075807)*x[0]**o+(0.206589504672)*x[0]+(0.232971263871)*x[1]**o+(-0.917991566808)*x[1]+(0.725943193794)*x[2]**o+(-0.170341108992)*x[2]
        arg[0,1,0,0]=(0.489505903018)*x[0]**o+(0.386376068728)*x[0]+(0.510797276159)*x[1]**o+(0.45356046467)*x[1]+(-0.664963144251)*x[2]**o+(0.128579947316)*x[2]
        arg[0,1,0,1]=(0.441567332096)*x[0]**o+(0.658686083904)*x[0]+(0.46601056348)*x[1]**o+(-0.130759928108)*x[1]+(0.324576589)*x[2]**o+(0.844700852997)*x[2]
        arg[0,1,1,0]=(0.388576628492)*x[0]**o+(0.41731399682)*x[0]+(-0.37037879729)*x[1]**o+(-0.00306376860454)*x[1]+(-0.464420131682)*x[2]**o+(0.570470055678)*x[2]
        arg[0,1,1,1]=(-0.0537780037583)*x[0]**o+(0.641854502644)*x[0]+(-0.475355608842)*x[1]**o+(-0.155501886306)*x[1]+(0.159833591997)*x[2]**o+(0.889951997355)*x[2]
        arg[0,1,2,0]=(0.58262140787)*x[0]**o+(0.487380368384)*x[0]+(-0.783228845976)*x[1]**o+(-0.683221249176)*x[1]+(0.98628901689)*x[2]**o+(0.591884616904)*x[2]
        arg[0,1,2,1]=(0.438472233911)*x[0]**o+(0.989088416794)*x[0]+(0.221688966923)*x[1]**o+(-0.119266918098)*x[1]+(-0.887421255415)*x[2]**o+(0.728082459156)*x[2]
        arg[0,2,0,0]=(-0.334110476968)*x[0]**o+(-0.877094545216)*x[0]+(-0.293135209996)*x[1]**o+(0.578110480224)*x[1]+(-0.898686340104)*x[2]**o+(0.528072783107)*x[2]
        arg[0,2,0,1]=(0.33041258354)*x[0]**o+(-0.70349630629)*x[0]+(0.349551161792)*x[1]**o+(-0.296687014146)*x[1]+(-0.272377440466)*x[2]**o+(-0.42221458274)*x[2]
        arg[0,2,1,0]=(-0.80021736912)*x[0]**o+(0.271727608951)*x[0]+(0.689351606243)*x[1]**o+(-0.95063104158)*x[1]+(0.447293221661)*x[2]**o+(0.962435109303)*x[2]
        arg[0,2,1,1]=(-0.773331228218)*x[0]**o+(0.569225340051)*x[0]+(-0.0330042729183)*x[1]**o+(-0.732867386826)*x[1]+(-0.399631128084)*x[2]**o+(0.235907094874)*x[2]
        arg[0,2,2,0]=(-0.943072868965)*x[0]**o+(-0.959364749946)*x[0]+(0.922934300445)*x[1]**o+(0.27257203243)*x[1]+(0.450453864748)*x[2]**o+(0.521284310775)*x[2]
        arg[0,2,2,1]=(-0.821201079286)*x[0]**o+(-0.801011107627)*x[0]+(0.554575277011)*x[1]**o+(-0.411361733513)*x[1]+(-0.858195570474)*x[2]**o+(0.420660103079)*x[2]
        arg[0,3,0,0]=(-0.0564164524098)*x[0]**o+(-0.865516143298)*x[0]+(-0.85754022189)*x[1]**o+(-0.0159800361584)*x[1]+(-0.891158469429)*x[2]**o+(-0.809909518115)*x[2]
        arg[0,3,0,1]=(-0.553858251068)*x[0]**o+(0.827558838148)*x[0]+(0.690081127613)*x[1]**o+(0.3475090336)*x[1]+(-0.190426828891)*x[2]**o+(-0.99589316535)*x[2]
        arg[0,3,1,0]=(-0.992317453163)*x[0]**o+(-0.992812531491)*x[0]+(-0.859683310869)*x[1]**o+(0.441779587739)*x[1]+(0.582457020078)*x[2]**o+(-0.959111461424)*x[2]
        arg[0,3,1,1]=(-0.713753922213)*x[0]**o+(0.504885148776)*x[0]+(-0.076988545634)*x[1]**o+(0.530828889101)*x[1]+(-0.335616242708)*x[2]**o+(-0.0421293420747)*x[2]
        arg[0,3,2,0]=(0.536024118289)*x[0]**o+(-0.527038052847)*x[0]+(-0.558392588656)*x[1]**o+(0.928618057856)*x[1]+(0.427507510033)*x[2]**o+(0.223570914141)*x[2]
        arg[0,3,2,1]=(-0.197968756117)*x[0]**o+(-0.58157606173)*x[0]+(-0.878630012825)*x[1]**o+(-0.380635552726)*x[1]+(-0.477570149125)*x[2]**o+(-0.642904015434)*x[2]
        arg[0,4,0,0]=(0.688763655127)*x[0]**o+(0.286011523933)*x[0]+(0.0933014250929)*x[1]**o+(0.355550563233)*x[1]+(0.121445291305)*x[2]**o+(-0.814538082974)*x[2]
        arg[0,4,0,1]=(-0.402717742839)*x[0]**o+(0.403946649835)*x[0]+(0.286978800579)*x[1]**o+(-0.363358056216)*x[1]+(-0.185233594102)*x[2]**o+(-0.902660250807)*x[2]
        arg[0,4,1,0]=(0.39352838659)*x[0]**o+(-0.430755011337)*x[0]+(-0.912518631162)*x[1]**o+(-0.634118795867)*x[1]+(-0.802445843416)*x[2]**o+(0.643395158048)*x[2]
        arg[0,4,1,1]=(-0.375507250175)*x[0]**o+(-0.609576942939)*x[0]+(0.35060243693)*x[1]**o+(0.384630130603)*x[1]+(0.165576820334)*x[2]**o+(-0.319012357341)*x[2]
        arg[0,4,2,0]=(0.0603590263818)*x[0]**o+(-0.435954775883)*x[0]+(-0.895329092913)*x[1]**o+(0.364809539103)*x[1]+(0.912864788565)*x[2]**o+(0.582044419963)*x[2]
        arg[0,4,2,1]=(-0.111089454685)*x[0]**o+(-0.247527778755)*x[0]+(0.897263901657)*x[1]**o+(0.168097655548)*x[1]+(-0.532273075067)*x[2]**o+(-0.158529268933)*x[2]
        arg[1,0,0,0]=(-0.940865155622)*x[0]**o+(0.639652431549)*x[0]+(0.204433172026)*x[1]**o+(-0.333435100662)*x[1]+(0.922286591304)*x[2]**o+(0.245091036189)*x[2]
        arg[1,0,0,1]=(-0.843207880309)*x[0]**o+(0.496567621809)*x[0]+(-0.747925907677)*x[1]**o+(-0.666195656311)*x[1]+(-0.836848401917)*x[2]**o+(-0.77301298659)*x[2]
        arg[1,0,1,0]=(-0.484987528212)*x[0]**o+(-0.208798022576)*x[0]+(-0.516295195217)*x[1]**o+(0.273733676249)*x[1]+(0.642865886081)*x[2]**o+(0.94556566928)*x[2]
        arg[1,0,1,1]=(0.285548625366)*x[0]**o+(-0.316870037241)*x[0]+(0.166055429845)*x[1]**o+(-0.907240066063)*x[1]+(-0.973602611383)*x[2]**o+(-0.232655535454)*x[2]
        arg[1,0,2,0]=(0.96797757731)*x[0]**o+(-0.0920831342923)*x[0]+(0.0367163548581)*x[1]**o+(0.372503137376)*x[1]+(0.114902128151)*x[2]**o+(0.147390259838)*x[2]
        arg[1,0,2,1]=(-0.486190283922)*x[0]**o+(-0.713417720041)*x[0]+(0.0212905465522)*x[1]**o+(-0.358636168795)*x[1]+(0.719370965272)*x[2]**o+(-0.506502687359)*x[2]
        arg[1,1,0,0]=(-0.216544472879)*x[0]**o+(-0.690540499698)*x[0]+(0.190155427044)*x[1]**o+(0.832206912182)*x[1]+(-0.108713936252)*x[2]**o+(-0.221373087337)*x[2]
        arg[1,1,0,1]=(0.982934604971)*x[0]**o+(0.604369826835)*x[0]+(-0.34365978148)*x[1]**o+(-0.167486779981)*x[1]+(-0.753609370582)*x[2]**o+(-0.220160888977)*x[2]
        arg[1,1,1,0]=(0.516177509446)*x[0]**o+(-0.474445723273)*x[0]+(0.813647087566)*x[1]**o+(-0.281201093802)*x[1]+(0.941465944108)*x[2]**o+(0.352728360059)*x[2]
        arg[1,1,1,1]=(-0.914732890633)*x[0]**o+(0.214962090901)*x[0]+(-0.710485058274)*x[1]**o+(-0.070756754056)*x[1]+(-0.449571590211)*x[2]**o+(0.353655711451)*x[2]
        arg[1,1,2,0]=(-0.96655555619)*x[0]**o+(-0.47460534732)*x[0]+(0.0610695053327)*x[1]**o+(0.815390116709)*x[1]+(-0.844916163552)*x[2]**o+(-0.433015304307)*x[2]
        arg[1,1,2,1]=(-0.611879453709)*x[0]**o+(0.849092848509)*x[0]+(0.0181322361566)*x[1]**o+(0.264173017827)*x[1]+(0.0452154806937)*x[2]**o+(0.0395300648465)*x[2]
        arg[1,2,0,0]=(0.558820946357)*x[0]**o+(-0.471519705598)*x[0]+(-0.200404864145)*x[1]**o+(0.452044160691)*x[1]+(-0.507559485963)*x[2]**o+(-0.993437841726)*x[2]
        arg[1,2,0,1]=(-0.944738664192)*x[0]**o+(-0.497536772117)*x[0]+(0.111399525675)*x[1]**o+(-0.93450294271)*x[1]+(-0.463702615737)*x[2]**o+(-0.330112672884)*x[2]
        arg[1,2,1,0]=(0.238225961058)*x[0]**o+(-0.471520173419)*x[0]+(0.483767084892)*x[1]**o+(-0.0359407284537)*x[1]+(-0.767379409753)*x[2]**o+(-0.283596371915)*x[2]
        arg[1,2,1,1]=(0.188699181238)*x[0]**o+(0.829593121989)*x[0]+(0.603377767593)*x[1]**o+(0.163334775468)*x[1]+(-0.695141594131)*x[2]**o+(-0.335208640181)*x[2]
        arg[1,2,2,0]=(-0.0953160757881)*x[0]**o+(0.179594592176)*x[0]+(0.95983902191)*x[1]**o+(-0.689622683527)*x[1]+(0.418442085936)*x[2]**o+(-0.271399861129)*x[2]
        arg[1,2,2,1]=(0.598691886555)*x[0]**o+(0.334120340646)*x[0]+(0.580763872443)*x[1]**o+(0.769016736836)*x[1]+(-0.442121483951)*x[2]**o+(0.0685264980499)*x[2]
        arg[1,3,0,0]=(-0.482944276372)*x[0]**o+(-0.0518057205519)*x[0]+(-0.878100108766)*x[1]**o+(-0.130957239599)*x[1]+(0.71270341547)*x[2]**o+(-0.313822116756)*x[2]
        arg[1,3,0,1]=(0.565128725562)*x[0]**o+(0.706149004661)*x[0]+(0.837747652328)*x[1]**o+(-0.429417376176)*x[1]+(0.179437627997)*x[2]**o+(-0.841831183522)*x[2]
        arg[1,3,1,0]=(0.608648887224)*x[0]**o+(0.118220851451)*x[0]+(0.492890967377)*x[1]**o+(-0.162344084552)*x[1]+(0.141203210671)*x[2]**o+(0.883457779687)*x[2]
        arg[1,3,1,1]=(0.0322521291544)*x[0]**o+(0.839748977403)*x[0]+(0.75094786544)*x[1]**o+(0.954833145672)*x[1]+(0.461457599013)*x[2]**o+(-0.761973385747)*x[2]
        arg[1,3,2,0]=(-0.757022609619)*x[0]**o+(-0.250893790099)*x[0]+(-0.973631035961)*x[1]**o+(0.506574565217)*x[1]+(0.103557430896)*x[2]**o+(0.881026337134)*x[2]
        arg[1,3,2,1]=(-0.395945302352)*x[0]**o+(-0.125428842065)*x[0]+(-0.339663004568)*x[1]**o+(-0.0272044592179)*x[1]+(0.75200506467)*x[2]**o+(-0.632922582454)*x[2]
        arg[1,4,0,0]=(0.0462127928202)*x[0]**o+(-0.745613488751)*x[0]+(0.357338297422)*x[1]**o+(0.739034094804)*x[1]+(-0.451247272253)*x[2]**o+(0.418030110847)*x[2]
        arg[1,4,0,1]=(0.914131238526)*x[0]**o+(-0.0280436066624)*x[0]+(0.861048515118)*x[1]**o+(0.148206807819)*x[1]+(-0.990856990354)*x[2]**o+(0.903615279172)*x[2]
        arg[1,4,1,0]=(0.189809824038)*x[0]**o+(-0.749477832258)*x[0]+(-0.802713794942)*x[1]**o+(0.0610727255572)*x[1]+(0.292940815327)*x[2]**o+(0.213443422205)*x[2]
        arg[1,4,1,1]=(0.792313292544)*x[0]**o+(0.321344675677)*x[0]+(-0.417516596097)*x[1]**o+(-0.63797359819)*x[1]+(-0.426112319872)*x[2]**o+(0.69281310288)*x[2]
        arg[1,4,2,0]=(-0.11017764073)*x[0]**o+(0.703037562344)*x[0]+(0.739591950873)*x[1]**o+(0.445612268331)*x[1]+(-0.071251825024)*x[2]**o+(0.00577350654185)*x[2]
        arg[1,4,2,1]=(-0.793584849832)*x[0]**o+(-0.546439256231)*x[0]+(-0.947194030803)*x[1]**o+(-0.0455700937677)*x[1]+(0.535368301918)*x[2]**o+(0.814435535521)*x[2]
        arg[2,0,0,0]=(0.249661949438)*x[0]**o+(-0.581533619092)*x[0]+(-0.674373505747)*x[1]**o+(0.464175602196)*x[1]+(-0.528086662834)*x[2]**o+(-0.212416176351)*x[2]
        arg[2,0,0,1]=(-0.588282108798)*x[0]**o+(0.761293399494)*x[0]+(0.99918874903)*x[1]**o+(0.542499992928)*x[1]+(0.433925746483)*x[2]**o+(0.972864491889)*x[2]
        arg[2,0,1,0]=(0.0572335623442)*x[0]**o+(0.444906443533)*x[0]+(-0.305538151235)*x[1]**o+(0.244530068532)*x[1]+(-0.20198975858)*x[2]**o+(-0.00490928477351)*x[2]
        arg[2,0,1,1]=(0.832496085695)*x[0]**o+(-0.797246004595)*x[0]+(0.873816520001)*x[1]**o+(0.628293733929)*x[1]+(-0.110922710624)*x[2]**o+(-0.0718221914062)*x[2]
        arg[2,0,2,0]=(0.140970192196)*x[0]**o+(0.582230323368)*x[0]+(-0.757764751267)*x[1]**o+(0.206061262231)*x[1]+(0.657237474788)*x[2]**o+(0.598632779121)*x[2]
        arg[2,0,2,1]=(-0.686218790734)*x[0]**o+(-0.157012470162)*x[0]+(0.448565721848)*x[1]**o+(0.0539087954076)*x[1]+(0.619821805007)*x[2]**o+(0.289144379283)*x[2]
        arg[2,1,0,0]=(0.787423884494)*x[0]**o+(-0.985959420171)*x[0]+(0.155584360983)*x[1]**o+(0.223292293778)*x[1]+(-0.437349719833)*x[2]**o+(0.989077189275)*x[2]
        arg[2,1,0,1]=(-0.597301631508)*x[0]**o+(-0.0414813766671)*x[0]+(-0.64883048443)*x[1]**o+(-0.18731615522)*x[1]+(0.735189823623)*x[2]**o+(-0.746633725819)*x[2]
        arg[2,1,1,0]=(-0.393486075452)*x[0]**o+(-0.480110060647)*x[0]+(0.559811870081)*x[1]**o+(0.0814282056993)*x[1]+(0.244899906416)*x[2]**o+(-0.884478480652)*x[2]
        arg[2,1,1,1]=(0.547747813175)*x[0]**o+(0.317890585903)*x[0]+(-0.70925099068)*x[1]**o+(-0.543227445702)*x[1]+(0.749693240223)*x[2]**o+(0.79984501038)*x[2]
        arg[2,1,2,0]=(-0.289583636666)*x[0]**o+(0.513926622265)*x[0]+(-0.00283544918865)*x[1]**o+(0.530724065811)*x[1]+(0.900479623243)*x[2]**o+(-0.146332301083)*x[2]
        arg[2,1,2,1]=(-0.487223003118)*x[0]**o+(0.161546483594)*x[0]+(-0.685211654772)*x[1]**o+(0.4803179816)*x[1]+(0.676120396917)*x[2]**o+(-0.174953085068)*x[2]
        arg[2,2,0,0]=(0.31746495834)*x[0]**o+(-0.513225182059)*x[0]+(0.33220677322)*x[1]**o+(-0.302371334854)*x[1]+(0.0964988251907)*x[2]**o+(-0.259532745762)*x[2]
        arg[2,2,0,1]=(0.313462846078)*x[0]**o+(0.079720902627)*x[0]+(-0.487932457563)*x[1]**o+(-0.786522786773)*x[1]+(0.357868214554)*x[2]**o+(0.251856050362)*x[2]
        arg[2,2,1,0]=(-0.180472378556)*x[0]**o+(0.345084802362)*x[0]+(0.641108557953)*x[1]**o+(0.198717742483)*x[1]+(0.0374114682326)*x[2]**o+(-0.915249907773)*x[2]
        arg[2,2,1,1]=(0.303787742688)*x[0]**o+(-0.941822487745)*x[0]+(-0.262567025842)*x[1]**o+(-0.197051445307)*x[1]+(0.381561965565)*x[2]**o+(-0.198549777232)*x[2]
        arg[2,2,2,0]=(-0.850284190461)*x[0]**o+(0.985224868544)*x[0]+(0.633868153422)*x[1]**o+(-0.57559927346)*x[1]+(0.68753726951)*x[2]**o+(-0.923696997656)*x[2]
        arg[2,2,2,1]=(0.559527478503)*x[0]**o+(-0.619764636475)*x[0]+(-0.0352163693945)*x[1]**o+(-0.37351931995)*x[1]+(0.0772333960938)*x[2]**o+(-0.353057755509)*x[2]
        arg[2,3,0,0]=(-0.274260374984)*x[0]**o+(0.365180349506)*x[0]+(-0.879564869545)*x[1]**o+(-0.0908162765131)*x[1]+(0.813491263836)*x[2]**o+(-0.396739585204)*x[2]
        arg[2,3,0,1]=(-0.793218219979)*x[0]**o+(0.227474218201)*x[0]+(0.97716610661)*x[1]**o+(0.317074032722)*x[1]+(0.076776446926)*x[2]**o+(0.069448617786)*x[2]
        arg[2,3,1,0]=(0.411400569812)*x[0]**o+(0.517743454926)*x[0]+(0.646697523033)*x[1]**o+(-0.973369313024)*x[1]+(-0.599741793926)*x[2]**o+(-0.743687020283)*x[2]
        arg[2,3,1,1]=(-0.491445354697)*x[0]**o+(-0.549840110396)*x[0]+(-0.260012915614)*x[1]**o+(0.671054626537)*x[1]+(0.0394966796969)*x[2]**o+(-0.701243931411)*x[2]
        arg[2,3,2,0]=(-0.806859723517)*x[0]**o+(-0.753940066715)*x[0]+(-0.664229274865)*x[1]**o+(0.0595484609931)*x[1]+(-0.921892262667)*x[2]**o+(-0.701770672395)*x[2]
        arg[2,3,2,1]=(-0.313979805844)*x[0]**o+(-0.542583864222)*x[0]+(-0.985990886553)*x[1]**o+(0.664926443979)*x[1]+(-0.99778334615)*x[2]**o+(0.268046810979)*x[2]
        arg[2,4,0,0]=(-0.41348240167)*x[0]**o+(0.473517720534)*x[0]+(-0.290880345559)*x[1]**o+(-0.607611624107)*x[1]+(-0.25537697471)*x[2]**o+(-0.472402477362)*x[2]
        arg[2,4,0,1]=(0.0222534371058)*x[0]**o+(-0.0739354104963)*x[0]+(-0.960057765718)*x[1]**o+(-0.772161403037)*x[1]+(-0.0356127352146)*x[2]**o+(-0.623605628308)*x[2]
        arg[2,4,1,0]=(-0.750995516053)*x[0]**o+(-0.289601799215)*x[0]+(0.294370180152)*x[1]**o+(0.956012147239)*x[1]+(-0.27958662106)*x[2]**o+(-0.370609593537)*x[2]
        arg[2,4,1,1]=(0.321767371741)*x[0]**o+(0.999223434136)*x[0]+(0.665865428241)*x[1]**o+(0.427942082269)*x[1]+(-0.135020552387)*x[2]**o+(0.692025479514)*x[2]
        arg[2,4,2,0]=(-0.519508249649)*x[0]**o+(0.0779688946471)*x[0]+(0.493489242872)*x[1]**o+(-0.0961136530093)*x[1]+(-0.0750243559261)*x[2]**o+(0.860970740104)*x[2]
        arg[2,4,2,1]=(0.807909049346)*x[0]**o+(0.388180877227)*x[0]+(-0.9044578433)*x[1]**o+(0.86765389355)*x[1]+(-0.196105779343)*x[2]**o+(0.448661151455)*x[2]
        arg[3,0,0,0]=(-0.780895567041)*x[0]**o+(-0.571253266808)*x[0]+(-0.904183399932)*x[1]**o+(-0.883296586722)*x[1]+(0.855945559487)*x[2]**o+(0.618533553098)*x[2]
        arg[3,0,0,1]=(-0.831076704037)*x[0]**o+(0.898347446798)*x[0]+(-0.716442620396)*x[1]**o+(0.761028369421)*x[1]+(-0.486505604444)*x[2]**o+(0.25474220145)*x[2]
        arg[3,0,1,0]=(-0.0127541044953)*x[0]**o+(0.759482315308)*x[0]+(-0.725655006866)*x[1]**o+(0.667310781285)*x[1]+(0.266075927439)*x[2]**o+(0.423300182252)*x[2]
        arg[3,0,1,1]=(-0.594972266882)*x[0]**o+(-0.842501149639)*x[0]+(-0.347686790781)*x[1]**o+(0.686403640474)*x[1]+(0.3452971966)*x[2]**o+(0.722258585706)*x[2]
        arg[3,0,2,0]=(-0.715711146058)*x[0]**o+(-0.835739963371)*x[0]+(0.966177361301)*x[1]**o+(-0.909111754995)*x[1]+(-0.175184576557)*x[2]**o+(0.695058239921)*x[2]
        arg[3,0,2,1]=(0.306251544915)*x[0]**o+(0.944775715803)*x[0]+(-0.740601754617)*x[1]**o+(0.434394890125)*x[1]+(-0.876527474971)*x[2]**o+(0.209321129689)*x[2]
        arg[3,1,0,0]=(-0.86788221871)*x[0]**o+(-0.833030657309)*x[0]+(0.84576621947)*x[1]**o+(-0.469186887277)*x[1]+(-0.527611740694)*x[2]**o+(-0.312054431001)*x[2]
        arg[3,1,0,1]=(-0.353678394652)*x[0]**o+(0.591287444611)*x[0]+(-0.213012375931)*x[1]**o+(-0.917259099251)*x[1]+(0.739016360622)*x[2]**o+(0.718754459378)*x[2]
        arg[3,1,1,0]=(-0.92646825221)*x[0]**o+(-0.870564973175)*x[0]+(-0.781805219404)*x[1]**o+(0.078530434642)*x[1]+(0.386018501087)*x[2]**o+(-0.120490492288)*x[2]
        arg[3,1,1,1]=(-0.824273737615)*x[0]**o+(0.401211543421)*x[0]+(0.360077381624)*x[1]**o+(-0.0599234883152)*x[1]+(0.246820480884)*x[2]**o+(-0.175313542134)*x[2]
        arg[3,1,2,0]=(0.807756928769)*x[0]**o+(0.889534807508)*x[0]+(0.633296189954)*x[1]**o+(-0.471821951252)*x[1]+(0.396011249874)*x[2]**o+(0.386800809575)*x[2]
        arg[3,1,2,1]=(-0.494145142543)*x[0]**o+(-0.514793985721)*x[0]+(0.825379848403)*x[1]**o+(-0.457860259786)*x[1]+(-0.599557868275)*x[2]**o+(-0.952704784672)*x[2]
        arg[3,2,0,0]=(0.970345295013)*x[0]**o+(-0.838468036663)*x[0]+(0.265634348738)*x[1]**o+(-0.873847124748)*x[1]+(0.416242648913)*x[2]**o+(0.59908438222)*x[2]
        arg[3,2,0,1]=(0.692377432429)*x[0]**o+(-0.338749041227)*x[0]+(-0.449383750398)*x[1]**o+(0.841669527559)*x[1]+(0.902482059741)*x[2]**o+(-0.655719647704)*x[2]
        arg[3,2,1,0]=(0.152417917522)*x[0]**o+(-0.78860144)*x[0]+(-0.881276170743)*x[1]**o+(0.692901550061)*x[1]+(-0.985377819322)*x[2]**o+(0.362747973601)*x[2]
        arg[3,2,1,1]=(0.238845792119)*x[0]**o+(-0.65838373309)*x[0]+(-0.659676514296)*x[1]**o+(0.870742876778)*x[1]+(-0.367345415947)*x[2]**o+(0.322353666541)*x[2]
        arg[3,2,2,0]=(0.456386394533)*x[0]**o+(-0.135575574979)*x[0]+(-0.731714323922)*x[1]**o+(0.936807859479)*x[1]+(0.511595594967)*x[2]**o+(0.570282464669)*x[2]
        arg[3,2,2,1]=(0.53299212624)*x[0]**o+(0.800998166434)*x[0]+(-0.431740087552)*x[1]**o+(-0.880092532661)*x[1]+(-0.418657988674)*x[2]**o+(-0.181008149338)*x[2]
        arg[3,3,0,0]=(0.60015312936)*x[0]**o+(-0.163125583081)*x[0]+(0.299668981953)*x[1]**o+(-0.761494422747)*x[1]+(0.404839953231)*x[2]**o+(-0.114039836538)*x[2]
        arg[3,3,0,1]=(-0.597481072133)*x[0]**o+(-0.506468582896)*x[0]+(0.406822348574)*x[1]**o+(0.247431348859)*x[1]+(-0.137941563908)*x[2]**o+(0.0503979259383)*x[2]
        arg[3,3,1,0]=(0.0921069092744)*x[0]**o+(-0.113960490794)*x[0]+(0.750059773909)*x[1]**o+(0.894945048248)*x[1]+(-0.154873449786)*x[2]**o+(-0.520229467775)*x[2]
        arg[3,3,1,1]=(0.349078696154)*x[0]**o+(0.141298430827)*x[0]+(0.249161007832)*x[1]**o+(0.0240310278591)*x[1]+(-0.18294527838)*x[2]**o+(-0.666698832632)*x[2]
        arg[3,3,2,0]=(-0.476684936147)*x[0]**o+(0.334093402615)*x[0]+(-0.326434684009)*x[1]**o+(0.91336431975)*x[1]+(0.483080537399)*x[2]**o+(0.439096732335)*x[2]
        arg[3,3,2,1]=(0.214713426286)*x[0]**o+(-0.973239849984)*x[0]+(0.984290637567)*x[1]**o+(-0.644668525299)*x[1]+(0.00985524319943)*x[2]**o+(0.68777403533)*x[2]
        arg[3,4,0,0]=(-0.88715423933)*x[0]**o+(0.642412222188)*x[0]+(-0.119667464132)*x[1]**o+(-0.201583116481)*x[1]+(-0.870921042381)*x[2]**o+(-0.00379039429884)*x[2]
        arg[3,4,0,1]=(0.511114986801)*x[0]**o+(-0.497617928997)*x[0]+(-0.189435233543)*x[1]**o+(-0.29288141421)*x[1]+(-0.476297664539)*x[2]**o+(-0.202915859137)*x[2]
        arg[3,4,1,0]=(0.0739878219028)*x[0]**o+(0.675050675951)*x[0]+(-0.45950101521)*x[1]**o+(0.306538598103)*x[1]+(-0.726069135031)*x[2]**o+(0.193290422951)*x[2]
        arg[3,4,1,1]=(-0.141324267565)*x[0]**o+(0.377343882109)*x[0]+(-0.16075604872)*x[1]**o+(0.263332274591)*x[1]+(0.886416182992)*x[2]**o+(0.931905996631)*x[2]
        arg[3,4,2,0]=(-0.76912732001)*x[0]**o+(-0.255011103121)*x[0]+(-0.372810348364)*x[1]**o+(0.108137766687)*x[1]+(-0.0761382180171)*x[2]**o+(0.46825331282)*x[2]
        arg[3,4,2,1]=(-0.515945940217)*x[0]**o+(-0.630017609554)*x[0]+(-0.815236764508)*x[1]**o+(0.636572471005)*x[1]+(0.244112465789)*x[2]**o+(-0.471661017589)*x[2]
        ref[0,0,0,0]=(-0.738709393227)/(o+1.)+(0.283993125638)+(-0.891975745754)*0.5**o
        ref[0,0,0,1]=(0.912832356419)/(o+1.)+(-0.800452287951)+(0.428698930349)*0.5**o
        ref[0,0,1,0]=(0.604410785754)/(o+1.)+(-0.70496852873)+(0.974069118653)*0.5**o
        ref[0,0,1,1]=(-1.63466543191)/(o+1.)+(0.302144663639)+(0.813412823502)*0.5**o
        ref[0,0,2,0]=(0.755754363963)/(o+1.)+(0.151022836528)+(-0.874502916672)*0.5**o
        ref[0,0,2,1]=(0.958914457666)/(o+1.)+(-0.440871585564)+(-0.258524075807)*0.5**o
        ref[0,1,0,0]=(-0.154165868092)/(o+1.)+(0.484258240357)+(0.489505903018)*0.5**o
        ref[0,1,0,1]=(0.79058715248)/(o+1.)+(0.686313504396)+(0.441567332096)*0.5**o
        ref[0,1,1,0]=(-0.834798928973)/(o+1.)+(0.492360141946)+(0.388576628492)*0.5**o
        ref[0,1,1,1]=(-0.315522016845)/(o+1.)+(0.688152306846)+(-0.0537780037583)*0.5**o
        ref[0,1,2,0]=(0.203060170914)/(o+1.)+(0.198021868056)+(0.58262140787)*0.5**o
        ref[0,1,2,1]=(-0.665732288492)/(o+1.)+(0.798951978927)+(0.438472233911)*0.5**o
        ref[0,2,0,0]=(-1.1918215501)/(o+1.)+(0.114544359058)+(-0.334110476968)*0.5**o
        ref[0,2,0,1]=(0.0771737213265)/(o+1.)+(-0.711198951588)+(0.33041258354)*0.5**o
        ref[0,2,1,0]=(1.1366448279)/(o+1.)+(0.141765838337)+(-0.80021736912)*0.5**o
        ref[0,2,1,1]=(-0.432635401002)/(o+1.)+(0.0361325240497)+(-0.773331228218)*0.5**o
        ref[0,2,2,0]=(1.37338816519)/(o+1.)+(-0.0827542033711)+(-0.943072868965)*0.5**o
        ref[0,2,2,1]=(-0.303620293463)/(o+1.)+(-0.39585636903)+(-0.821201079286)*0.5**o
        ref[0,3,0,0]=(-1.74869869132)/(o+1.)+(-0.845702848785)+(-0.0564164524098)*0.5**o
        ref[0,3,0,1]=(0.499654298722)/(o+1.)+(0.0895873531986)+(-0.553858251068)*0.5**o
        ref[0,3,1,0]=(-0.277226290791)/(o+1.)+(-0.755072202588)+(-0.992317453163)*0.5**o
        ref[0,3,1,1]=(-0.412604788342)/(o+1.)+(0.496792347901)+(-0.713753922213)*0.5**o
        ref[0,3,2,0]=(-0.130885078624)/(o+1.)+(0.312575459575)+(0.536024118289)*0.5**o
        ref[0,3,2,1]=(-1.35620016195)/(o+1.)+(-0.802557814945)+(-0.197968756117)*0.5**o
        ref[0,4,0,0]=(0.214746716397)/(o+1.)+(-0.0864879979042)+(0.688763655127)*0.5**o
        ref[0,4,0,1]=(0.101745206477)/(o+1.)+(-0.431035828594)+(-0.402717742839)*0.5**o
        ref[0,4,1,0]=(-1.71496447458)/(o+1.)+(-0.210739324578)+(0.39352838659)*0.5**o
        ref[0,4,1,1]=(0.516179257264)/(o+1.)+(-0.271979584839)+(-0.375507250175)*0.5**o
        ref[0,4,2,0]=(0.0175356956523)/(o+1.)+(0.255449591592)+(0.0603590263818)*0.5**o
        ref[0,4,2,1]=(0.36499082659)/(o+1.)+(-0.11897969607)+(-0.111089454685)*0.5**o
        ref[1,0,0,0]=(1.12671976333)/(o+1.)+(0.275654183538)+(-0.940865155622)*0.5**o
        ref[1,0,0,1]=(-1.58477430959)/(o+1.)+(-0.471320510546)+(-0.843207880309)*0.5**o
        ref[1,0,1,0]=(0.126570690864)/(o+1.)+(0.505250661477)+(-0.484987528212)*0.5**o
        ref[1,0,1,1]=(-0.807547181539)/(o+1.)+(-0.728382819379)+(0.285548625366)*0.5**o
        ref[1,0,2,0]=(0.151618483009)/(o+1.)+(0.213905131461)+(0.96797757731)*0.5**o
        ref[1,0,2,1]=(0.740661511824)/(o+1.)+(-0.789278288097)+(-0.486190283922)*0.5**o
        ref[1,1,0,0]=(0.0814414907926)/(o+1.)+(-0.0398533374263)+(-0.216544472879)*0.5**o
        ref[1,1,0,1]=(-1.09726915206)/(o+1.)+(0.108361078938)+(0.982934604971)*0.5**o
        ref[1,1,1,0]=(1.75511303167)/(o+1.)+(-0.201459228508)+(0.516177509446)*0.5**o
        ref[1,1,1,1]=(-1.16005664849)/(o+1.)+(0.248930524148)+(-0.914732890633)*0.5**o
        ref[1,1,2,0]=(-0.78384665822)/(o+1.)+(-0.0461152674588)+(-0.96655555619)*0.5**o
        ref[1,1,2,1]=(0.0633477168503)/(o+1.)+(0.576397965591)+(-0.611879453709)*0.5**o
        ref[1,2,0,0]=(-0.707964350108)/(o+1.)+(-0.506456693317)+(0.558820946357)*0.5**o
        ref[1,2,0,1]=(-0.352303090063)/(o+1.)+(-0.881076193856)+(-0.944738664192)*0.5**o
        ref[1,2,1,0]=(-0.283612324861)/(o+1.)+(-0.395528636894)+(0.238225961058)*0.5**o
        ref[1,2,1,1]=(-0.0917638265383)/(o+1.)+(0.328859628638)+(0.188699181238)*0.5**o
        ref[1,2,2,0]=(1.37828110785)/(o+1.)+(-0.39071397624)+(-0.0953160757881)*0.5**o
        ref[1,2,2,1]=(0.138642388492)/(o+1.)+(0.585831787766)+(0.598691886555)*0.5**o
        ref[1,3,0,0]=(-0.165396693296)/(o+1.)+(-0.248292538454)+(-0.482944276372)*0.5**o
        ref[1,3,0,1]=(1.01718528032)/(o+1.)+(-0.282549777519)+(0.565128725562)*0.5**o
        ref[1,3,1,0]=(0.634094178049)/(o+1.)+(0.419667273293)+(0.608648887224)*0.5**o
        ref[1,3,1,1]=(1.21240546445)/(o+1.)+(0.516304368664)+(0.0322521291544)*0.5**o
        ref[1,3,2,0]=(-0.870073605065)/(o+1.)+(0.568353556126)+(-0.757022609619)*0.5**o
        ref[1,3,2,1]=(0.412342060102)/(o+1.)+(-0.392777941868)+(-0.395945302352)*0.5**o
        ref[1,4,0,0]=(-0.0939089748304)/(o+1.)+(0.20572535845)+(0.0462127928202)*0.5**o
        ref[1,4,0,1]=(-0.129808475236)/(o+1.)+(0.511889240164)+(0.914131238526)*0.5**o
        ref[1,4,1,0]=(-0.509772979615)/(o+1.)+(-0.237480842248)+(0.189809824038)*0.5**o
        ref[1,4,1,1]=(-0.843628915969)/(o+1.)+(0.188092090183)+(0.792313292544)*0.5**o
        ref[1,4,2,0]=(0.668340125849)/(o+1.)+(0.577211668608)+(-0.11017764073)*0.5**o
        ref[1,4,2,1]=(-0.411825728886)/(o+1.)+(0.111213092761)+(-0.793584849832)*0.5**o
        ref[2,0,0,0]=(-1.20246016858)/(o+1.)+(-0.164887096623)+(0.249661949438)*0.5**o
        ref[2,0,0,1]=(1.43311449551)/(o+1.)+(1.13832894216)+(-0.588282108798)*0.5**o
        ref[2,0,1,0]=(-0.507527909815)/(o+1.)+(0.342263613646)+(0.0572335623442)*0.5**o
        ref[2,0,1,1]=(0.762893809377)/(o+1.)+(-0.120387231036)+(0.832496085695)*0.5**o
        ref[2,0,2,0]=(-0.100527276479)/(o+1.)+(0.69346218236)+(0.140970192196)*0.5**o
        ref[2,0,2,1]=(1.06838752686)/(o+1.)+(0.0930203522642)+(-0.686218790734)*0.5**o
        ref[2,1,0,0]=(-0.28176535885)/(o+1.)+(0.113205031441)+(0.787423884494)*0.5**o
        ref[2,1,0,1]=(0.0863593391922)/(o+1.)+(-0.487715628853)+(-0.597301631508)*0.5**o
        ref[2,1,1,0]=(0.804711776497)/(o+1.)+(-0.6415801678)+(-0.393486075452)*0.5**o
        ref[2,1,1,1]=(0.0404422495423)/(o+1.)+(0.28725407529)+(0.547747813175)*0.5**o
        ref[2,1,2,0]=(0.897644174054)/(o+1.)+(0.449159193497)+(-0.289583636666)*0.5**o
        ref[2,1,2,1]=(-0.00909125785473)/(o+1.)+(0.233455690063)+(-0.487223003118)*0.5**o
        ref[2,2,0,0]=(0.428705598411)/(o+1.)+(-0.537564631337)+(0.31746495834)*0.5**o
        ref[2,2,0,1]=(-0.13006424301)/(o+1.)+(-0.227472916892)+(0.313462846078)*0.5**o
        ref[2,2,1,0]=(0.678520026185)/(o+1.)+(-0.185723681464)+(-0.180472378556)*0.5**o
        ref[2,2,1,1]=(0.118994939723)/(o+1.)+(-0.668711855142)+(0.303787742688)*0.5**o
        ref[2,2,2,0]=(1.32140542293)/(o+1.)+(-0.257035701286)+(-0.850284190461)*0.5**o
        ref[2,2,2,1]=(0.0420170266993)/(o+1.)+(-0.673170855967)+(0.559527478503)*0.5**o
        ref[2,3,0,0]=(-0.0660736057092)/(o+1.)+(-0.0611877561056)+(-0.274260374984)*0.5**o
        ref[2,3,0,1]=(1.05394255354)/(o+1.)+(0.306998434354)+(-0.793218219979)*0.5**o
        ref[2,3,1,0]=(0.046955729107)/(o+1.)+(-0.599656439191)+(0.411400569812)*0.5**o
        ref[2,3,1,1]=(-0.220516235917)/(o+1.)+(-0.290014707635)+(-0.491445354697)*0.5**o
        ref[2,3,2,0]=(-1.58612153753)/(o+1.)+(-0.698081139059)+(-0.806859723517)*0.5**o
        ref[2,3,2,1]=(-1.9837742327)/(o+1.)+(0.195194695368)+(-0.313979805844)*0.5**o
        ref[2,4,0,0]=(-0.546257320269)/(o+1.)+(-0.303248190467)+(-0.41348240167)*0.5**o
        ref[2,4,0,1]=(-0.995670500932)/(o+1.)+(-0.734851220921)+(0.0222534371058)*0.5**o
        ref[2,4,1,0]=(0.0147835590923)/(o+1.)+(0.147900377243)+(-0.750995516053)*0.5**o
        ref[2,4,1,1]=(0.530844875854)/(o+1.)+(1.05959549796)+(0.321767371741)*0.5**o
        ref[2,4,2,0]=(0.418464886946)/(o+1.)+(0.421412990871)+(-0.519508249649)*0.5**o
        ref[2,4,2,1]=(-1.10056362264)/(o+1.)+(0.852247961116)+(0.807909049346)*0.5**o
        ref[3,0,0,0]=(-0.0482378404454)/(o+1.)+(-0.418008150216)+(-0.780895567041)*0.5**o
        ref[3,0,0,1]=(-1.20294822484)/(o+1.)+(0.957059008835)+(-0.831076704037)*0.5**o
        ref[3,0,1,0]=(-0.459579079427)/(o+1.)+(0.925046639422)+(-0.0127541044953)*0.5**o
        ref[3,0,1,1]=(-0.00238959418047)/(o+1.)+(0.28308053827)+(-0.594972266882)*0.5**o
        ref[3,0,2,0]=(0.790992784744)/(o+1.)+(-0.524896739222)+(-0.715711146058)*0.5**o
        ref[3,0,2,1]=(-1.61712922959)/(o+1.)+(0.794245867809)+(0.306251544915)*0.5**o
        ref[3,1,0,0]=(0.318154478777)/(o+1.)+(-0.807135987793)+(-0.86788221871)*0.5**o
        ref[3,1,0,1]=(0.526003984691)/(o+1.)+(0.196391402369)+(-0.353678394652)*0.5**o
        ref[3,1,1,0]=(-0.395786718317)/(o+1.)+(-0.456262515411)+(-0.92646825221)*0.5**o
        ref[3,1,1,1]=(0.606897862508)/(o+1.)+(0.082987256486)+(-0.824273737615)*0.5**o
        ref[3,1,2,0]=(1.02930743983)/(o+1.)+(0.402256832916)+(0.807756928769)*0.5**o
        ref[3,1,2,1]=(0.225821980128)/(o+1.)+(-0.96267951509)+(-0.494145142543)*0.5**o
        ref[3,2,0,0]=(0.681876997651)/(o+1.)+(-0.556615389596)+(0.970345295013)*0.5**o
        ref[3,2,0,1]=(0.453098309344)/(o+1.)+(-0.0763995806862)+(0.692377432429)*0.5**o
        ref[3,2,1,0]=(-1.86665399006)/(o+1.)+(0.133524041831)+(0.152417917522)*0.5**o
        ref[3,2,1,1]=(-1.02702193024)/(o+1.)+(0.267356405114)+(0.238845792119)*0.5**o
        ref[3,2,2,0]=(-0.220118728954)/(o+1.)+(0.685757374584)+(0.456386394533)*0.5**o
        ref[3,2,2,1]=(-0.850398076225)/(o+1.)+(-0.130051257782)+(0.53299212624)*0.5**o
        ref[3,3,0,0]=(0.704508935183)/(o+1.)+(-0.519329921183)+(0.60015312936)*0.5**o
        ref[3,3,0,1]=(0.268880784666)/(o+1.)+(-0.10431965405)+(-0.597481072133)*0.5**o
        ref[3,3,1,0]=(0.595186324123)/(o+1.)+(0.13037754484)+(0.0921069092744)*0.5**o
        ref[3,3,1,1]=(0.0662157294518)/(o+1.)+(-0.250684686973)+(0.349078696154)*0.5**o
        ref[3,3,2,0]=(0.156645853389)/(o+1.)+(0.84327722735)+(-0.476684936147)*0.5**o
        ref[3,3,2,1]=(0.994145880767)/(o+1.)+(-0.465067169976)+(0.214713426286)*0.5**o
        ref[3,4,0,0]=(-0.990588506513)/(o+1.)+(0.218519355704)+(-0.88715423933)*0.5**o
        ref[3,4,0,1]=(-0.665732898082)/(o+1.)+(-0.496707601172)+(0.511114986801)*0.5**o
        ref[3,4,1,0]=(-1.18557015024)/(o+1.)+(0.587439848502)+(0.0739878219028)*0.5**o
        ref[3,4,1,1]=(0.725660134271)/(o+1.)+(0.786291076666)+(-0.141324267565)*0.5**o
        ref[3,4,2,0]=(-0.448948566381)/(o+1.)+(0.160689988193)+(-0.76912732001)*0.5**o
        ref[3,4,2,1]=(-0.571124298719)/(o+1.)+(-0.232553078069)+(-0.515945940217)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.311774053155)*x[0]**o+(0.52292841627)*x[0]+(-0.518467126196)*x[1]**o+(-0.11329290159)*x[1]
        ref=(-0.518467126196)/(o+1.)+(0.20481775734)+(0.311774053155)*0.5**o
      else:
        arg=(0.233765319091)*x[0]**o+(-0.453551209762)*x[0]+(0.920140231713)*x[1]**o+(0.741695479084)*x[1]+(-0.255834461065)*x[2]**o+(-0.102580899053)*x[2]
        ref=(0.664305770648)/(o+1.)+(0.0927816851348)+(0.233765319091)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.651678780872)*x[0]**o+(-0.749709447394)*x[0]+(0.311088209082)*x[1]**o+(0.553081401193)*x[1]
        arg[1]=(0.657887728987)*x[0]**o+(0.052270143604)*x[0]+(-0.15507994277)*x[1]**o+(-0.30302150082)*x[1]
        ref[0]=(0.311088209082)/(o+1.)+(-0.0983140231004)+(-0.651678780872)*0.5**o
        ref[1]=(-0.15507994277)/(o+1.)+(-0.125375678608)+(0.657887728987)*0.5**o
      else:
        arg[0]=(0.844909347644)*x[0]**o+(-0.342697697483)*x[0]+(0.126242548186)*x[1]**o+(-0.854010732849)*x[1]+(-0.307627587461)*x[2]**o+(-0.998566290754)*x[2]
        arg[1]=(0.900073726701)*x[0]**o+(-0.703841701859)*x[0]+(-0.651690896499)*x[1]**o+(0.828362322181)*x[1]+(0.284410420895)*x[2]**o+(-0.394606348974)*x[2]
        ref[0]=(-0.181385039275)/(o+1.)+(-1.09763736054)+(0.844909347644)*0.5**o
        ref[1]=(-0.367280475604)/(o+1.)+(-0.135042864326)+(0.900073726701)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.263642681371)*x[0]**o+(0.668992571787)*x[0]+(0.873939846194)*x[1]**o+(0.784820355998)*x[1]
        arg[0,1]=(-0.389306088628)*x[0]**o+(-0.0549482389907)*x[0]+(-0.787101008457)*x[1]**o+(0.860345123812)*x[1]
        arg[0,2]=(-0.995721191961)*x[0]**o+(-0.587928136153)*x[0]+(-0.378312614832)*x[1]**o+(0.199765674281)*x[1]
        arg[0,3]=(-0.374679425145)*x[0]**o+(-0.503606183146)*x[0]+(0.23995738992)*x[1]**o+(-0.525359181744)*x[1]
        arg[0,4]=(-0.69923255665)*x[0]**o+(0.547634979941)*x[0]+(0.728792702739)*x[1]**o+(0.949888533828)*x[1]
        arg[1,0]=(-0.61856415528)*x[0]**o+(-0.0516239651093)*x[0]+(0.158516649979)*x[1]**o+(-0.0522288881011)*x[1]
        arg[1,1]=(0.0285524133857)*x[0]**o+(0.741357758307)*x[0]+(-0.0714860903847)*x[1]**o+(-0.539632750385)*x[1]
        arg[1,2]=(-0.291101401382)*x[0]**o+(-0.114900246437)*x[0]+(-0.982305031665)*x[1]**o+(0.0823478844055)*x[1]
        arg[1,3]=(-0.931305732525)*x[0]**o+(0.077332436417)*x[0]+(-0.880626154921)*x[1]**o+(-0.321222234507)*x[1]
        arg[1,4]=(-0.00778033208815)*x[0]**o+(-0.536626847957)*x[0]+(-0.599742714546)*x[1]**o+(0.454108299336)*x[1]
        arg[2,0]=(-0.575243090445)*x[0]**o+(-0.920247228564)*x[0]+(0.4569784462)*x[1]**o+(-0.105175798804)*x[1]
        arg[2,1]=(0.333985911748)*x[0]**o+(-0.017876213076)*x[0]+(0.475164778165)*x[1]**o+(0.561072178015)*x[1]
        arg[2,2]=(-0.559587786773)*x[0]**o+(-0.917134128672)*x[0]+(0.316654889327)*x[1]**o+(-0.640569595027)*x[1]
        arg[2,3]=(-0.140461324817)*x[0]**o+(0.259052969883)*x[0]+(0.397898509159)*x[1]**o+(0.932233331785)*x[1]
        arg[2,4]=(0.368626304985)*x[0]**o+(0.616310502384)*x[0]+(-0.756845293566)*x[1]**o+(-0.920088968132)*x[1]
        arg[3,0]=(-0.824398741087)*x[0]**o+(0.0498115894252)*x[0]+(-0.5256621655)*x[1]**o+(0.114806237108)*x[1]
        arg[3,1]=(-0.77671057316)*x[0]**o+(0.569249443162)*x[0]+(0.416628465801)*x[1]**o+(0.0603627487913)*x[1]
        arg[3,2]=(-0.247875220913)*x[0]**o+(0.19548828878)*x[0]+(0.643226101838)*x[1]**o+(-0.154558826499)*x[1]
        arg[3,3]=(-0.918937118963)*x[0]**o+(0.193777174417)*x[0]+(-0.906042347107)*x[1]**o+(0.872033230033)*x[1]
        arg[3,4]=(-0.628918336086)*x[0]**o+(0.823242191714)*x[0]+(-0.593691166712)*x[1]**o+(-0.661715840814)*x[1]
        ref[0,0]=(0.873939846194)/(o+1.)+(0.726906463892)+(-0.263642681371)*0.5**o
        ref[0,1]=(-0.787101008457)/(o+1.)+(0.402698442411)+(-0.389306088628)*0.5**o
        ref[0,2]=(-0.378312614832)/(o+1.)+(-0.194081230936)+(-0.995721191961)*0.5**o
        ref[0,3]=(0.23995738992)/(o+1.)+(-0.514482682445)+(-0.374679425145)*0.5**o
        ref[0,4]=(0.728792702739)/(o+1.)+(0.748761756885)+(-0.69923255665)*0.5**o
        ref[1,0]=(0.158516649979)/(o+1.)+(-0.0519264266052)+(-0.61856415528)*0.5**o
        ref[1,1]=(-0.0714860903847)/(o+1.)+(0.100862503961)+(0.0285524133857)*0.5**o
        ref[1,2]=(-0.982305031665)/(o+1.)+(-0.0162761810159)+(-0.291101401382)*0.5**o
        ref[1,3]=(-0.880626154921)/(o+1.)+(-0.121944899045)+(-0.931305732525)*0.5**o
        ref[1,4]=(-0.599742714546)/(o+1.)+(-0.0412592743102)+(-0.00778033208815)*0.5**o
        ref[2,0]=(0.4569784462)/(o+1.)+(-0.512711513684)+(-0.575243090445)*0.5**o
        ref[2,1]=(0.475164778165)/(o+1.)+(0.27159798247)+(0.333985911748)*0.5**o
        ref[2,2]=(0.316654889327)/(o+1.)+(-0.778851861849)+(-0.559587786773)*0.5**o
        ref[2,3]=(0.397898509159)/(o+1.)+(0.595643150834)+(-0.140461324817)*0.5**o
        ref[2,4]=(-0.756845293566)/(o+1.)+(-0.151889232874)+(0.368626304985)*0.5**o
        ref[3,0]=(-0.5256621655)/(o+1.)+(0.0823089132668)+(-0.824398741087)*0.5**o
        ref[3,1]=(0.416628465801)/(o+1.)+(0.314806095977)+(-0.77671057316)*0.5**o
        ref[3,2]=(0.643226101838)/(o+1.)+(0.0204647311405)+(-0.247875220913)*0.5**o
        ref[3,3]=(-0.906042347107)/(o+1.)+(0.532905202225)+(-0.918937118963)*0.5**o
        ref[3,4]=(-0.593691166712)/(o+1.)+(0.0807631754499)+(-0.628918336086)*0.5**o
      else:
        arg[0,0]=(0.748857269399)*x[0]**o+(-0.21982973469)*x[0]+(-0.765987921295)*x[1]**o+(0.556577019909)*x[1]+(-0.0438548079278)*x[2]**o+(0.543610874817)*x[2]
        arg[0,1]=(-0.784526004748)*x[0]**o+(-0.949423172035)*x[0]+(-0.29888188317)*x[1]**o+(-0.850638766848)*x[1]+(-0.853027558387)*x[2]**o+(-0.846378486588)*x[2]
        arg[0,2]=(0.53445882525)*x[0]**o+(-0.653655652989)*x[0]+(-0.968129875689)*x[1]**o+(-0.603598808323)*x[1]+(0.545841014272)*x[2]**o+(0.126131861082)*x[2]
        arg[0,3]=(0.181662498503)*x[0]**o+(-0.769571080865)*x[0]+(0.771109181078)*x[1]**o+(0.980677373303)*x[1]+(-0.625057940268)*x[2]**o+(0.0585672357346)*x[2]
        arg[0,4]=(-0.0320907540755)*x[0]**o+(-0.519517369917)*x[0]+(-0.365190231996)*x[1]**o+(-0.192392652661)*x[1]+(-0.693267159516)*x[2]**o+(0.106458640425)*x[2]
        arg[1,0]=(0.414286358161)*x[0]**o+(0.0223282279507)*x[0]+(0.215881279469)*x[1]**o+(0.68990977849)*x[1]+(-0.532075885581)*x[2]**o+(-0.269164904232)*x[2]
        arg[1,1]=(-0.0445879250589)*x[0]**o+(0.889985400599)*x[0]+(0.842050582947)*x[1]**o+(-0.0275307884642)*x[1]+(0.774650107671)*x[2]**o+(0.19446514971)*x[2]
        arg[1,2]=(0.913032880034)*x[0]**o+(-0.286537470008)*x[0]+(0.0902778267032)*x[1]**o+(0.548596933564)*x[1]+(-0.996100405074)*x[2]**o+(-0.509554019549)*x[2]
        arg[1,3]=(0.44308812765)*x[0]**o+(-0.0641227726924)*x[0]+(0.484249008075)*x[1]**o+(0.581496384894)*x[1]+(0.592538533183)*x[2]**o+(-0.821638541512)*x[2]
        arg[1,4]=(-0.898304735153)*x[0]**o+(0.333217967121)*x[0]+(-0.788218899371)*x[1]**o+(0.579723459039)*x[1]+(0.467006473159)*x[2]**o+(-0.277113860535)*x[2]
        arg[2,0]=(0.515308716564)*x[0]**o+(0.255947750577)*x[0]+(0.232578337945)*x[1]**o+(0.900611887592)*x[1]+(-0.247710965611)*x[2]**o+(-0.985523228221)*x[2]
        arg[2,1]=(0.611196756618)*x[0]**o+(0.146822999187)*x[0]+(-0.318793147278)*x[1]**o+(0.790772432028)*x[1]+(0.138036152668)*x[2]**o+(-0.800234207543)*x[2]
        arg[2,2]=(-0.293576196546)*x[0]**o+(-0.521059592737)*x[0]+(0.301178293957)*x[1]**o+(-0.141430900198)*x[1]+(-0.782381874467)*x[2]**o+(0.92221179729)*x[2]
        arg[2,3]=(0.673951778483)*x[0]**o+(-0.335068772752)*x[0]+(0.408793200723)*x[1]**o+(0.36906212176)*x[1]+(0.246292093855)*x[2]**o+(0.356147233675)*x[2]
        arg[2,4]=(-0.218899302236)*x[0]**o+(0.271285962886)*x[0]+(-0.16364707283)*x[1]**o+(0.177195607727)*x[1]+(-0.287331396699)*x[2]**o+(-0.465391460497)*x[2]
        arg[3,0]=(-0.369727856735)*x[0]**o+(-0.44351931666)*x[0]+(0.0375097686939)*x[1]**o+(0.276779121402)*x[1]+(0.662505672881)*x[2]**o+(-0.642709866174)*x[2]
        arg[3,1]=(0.444940917816)*x[0]**o+(-0.178926126454)*x[0]+(0.824660645994)*x[1]**o+(0.825996820355)*x[1]+(0.798260651848)*x[2]**o+(0.331744022239)*x[2]
        arg[3,2]=(-0.82489261483)*x[0]**o+(-0.340862079766)*x[0]+(-0.998375587383)*x[1]**o+(0.880222106116)*x[1]+(0.656096580473)*x[2]**o+(0.530854486127)*x[2]
        arg[3,3]=(0.732863411435)*x[0]**o+(-0.324252004147)*x[0]+(0.789472654884)*x[1]**o+(-0.207794796011)*x[1]+(0.50903180567)*x[2]**o+(0.724319112504)*x[2]
        arg[3,4]=(-0.980748139471)*x[0]**o+(0.0974331161906)*x[0]+(0.196275528452)*x[1]**o+(0.305231910003)*x[1]+(-0.0417825350947)*x[2]**o+(-0.2916834261)*x[2]
        ref[0,0]=(-0.809842729223)/(o+1.)+(0.440179080018)+(0.748857269399)*0.5**o
        ref[0,1]=(-1.15190944156)/(o+1.)+(-1.32322021274)+(-0.784526004748)*0.5**o
        ref[0,2]=(-0.422288861417)/(o+1.)+(-0.565561300115)+(0.53445882525)*0.5**o
        ref[0,3]=(0.146051240811)/(o+1.)+(0.134836764087)+(0.181662498503)*0.5**o
        ref[0,4]=(-1.05845739151)/(o+1.)+(-0.302725691077)+(-0.0320907540755)*0.5**o
        ref[1,0]=(-0.316194606113)/(o+1.)+(0.221536551104)+(0.414286358161)*0.5**o
        ref[1,1]=(1.61670069062)/(o+1.)+(0.528459880922)+(-0.0445879250589)*0.5**o
        ref[1,2]=(-0.905822578371)/(o+1.)+(-0.123747277996)+(0.913032880034)*0.5**o
        ref[1,3]=(1.07678754126)/(o+1.)+(-0.152132464655)+(0.44308812765)*0.5**o
        ref[1,4]=(-0.321212426212)/(o+1.)+(0.317913782812)+(-0.898304735153)*0.5**o
        ref[2,0]=(-0.015132627666)/(o+1.)+(0.085518204974)+(0.515308716564)*0.5**o
        ref[2,1]=(-0.18075699461)/(o+1.)+(0.0686806118359)+(0.611196756618)*0.5**o
        ref[2,2]=(-0.481203580509)/(o+1.)+(0.129860652178)+(-0.293576196546)*0.5**o
        ref[2,3]=(0.655085294578)/(o+1.)+(0.195070291341)+(0.673951778483)*0.5**o
        ref[2,4]=(-0.450978469529)/(o+1.)+(-0.00845494494169)+(-0.218899302236)*0.5**o
        ref[3,0]=(0.700015441575)/(o+1.)+(-0.404725030716)+(-0.369727856735)*0.5**o
        ref[3,1]=(1.62292129784)/(o+1.)+(0.48940735807)+(0.444940917816)*0.5**o
        ref[3,2]=(-0.342279006911)/(o+1.)+(0.535107256239)+(-0.82489261483)*0.5**o
        ref[3,3]=(1.29850446055)/(o+1.)+(0.0961361561734)+(0.732863411435)*0.5**o
        ref[3,4]=(0.154492993358)/(o+1.)+(0.055490800047)+(-0.980748139471)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.33788666753)*x[0]**o+(0.0147671449643)*x[0]+(0.14066054662)*x[1]**o+(-0.675231899756)*x[1]
        arg[0,0,1]=(0.0100961441611)*x[0]**o+(0.401304171984)*x[0]+(-0.594786282119)*x[1]**o+(-0.561543495067)*x[1]
        arg[0,1,0]=(-0.118311175162)*x[0]**o+(0.62454992688)*x[0]+(-0.8163725079)*x[1]**o+(-0.756355655277)*x[1]
        arg[0,1,1]=(0.586991709411)*x[0]**o+(-0.600757941586)*x[0]+(0.0953383109011)*x[1]**o+(0.317827629801)*x[1]
        arg[1,0,0]=(-0.251125236294)*x[0]**o+(0.166603616954)*x[0]+(0.978704943828)*x[1]**o+(0.589215389731)*x[1]
        arg[1,0,1]=(0.0543182786738)*x[0]**o+(-0.334756305272)*x[0]+(-0.919571042987)*x[1]**o+(-0.0425734727213)*x[1]
        arg[1,1,0]=(-0.867177621518)*x[0]**o+(-0.0750659593232)*x[0]+(-0.227277029921)*x[1]**o+(0.361120763296)*x[1]
        arg[1,1,1]=(0.85129514889)*x[0]**o+(-0.441962000795)*x[0]+(0.425335668395)*x[1]**o+(0.204091109081)*x[1]
        arg[2,0,0]=(-0.537930401585)*x[0]**o+(-0.206510941636)*x[0]+(-0.247374421196)*x[1]**o+(0.67043841725)*x[1]
        arg[2,0,1]=(-0.101273864474)*x[0]**o+(0.283769727506)*x[0]+(-0.0850059833654)*x[1]**o+(-0.71291661379)*x[1]
        arg[2,1,0]=(0.684430088582)*x[0]**o+(0.107733116086)*x[0]+(0.40398578341)*x[1]**o+(-0.525004991854)*x[1]
        arg[2,1,1]=(-0.00801759014554)*x[0]**o+(-0.214858014851)*x[0]+(-0.427822011518)*x[1]**o+(0.0168962506945)*x[1]
        arg[3,0,0]=(0.528563052182)*x[0]**o+(-0.630485796029)*x[0]+(0.0525765317703)*x[1]**o+(-0.469639006097)*x[1]
        arg[3,0,1]=(0.178622695427)*x[0]**o+(0.0952996500383)*x[0]+(-0.785851578496)*x[1]**o+(0.856665398338)*x[1]
        arg[3,1,0]=(0.547391926072)*x[0]**o+(-0.0179195091852)*x[0]+(-0.831368961862)*x[1]**o+(-0.437229107453)*x[1]
        arg[3,1,1]=(-0.657011566453)*x[0]**o+(0.530176142378)*x[0]+(0.649145894746)*x[1]**o+(-0.755185337598)*x[1]
        arg[4,0,0]=(-0.188305872652)*x[0]**o+(0.0207245930982)*x[0]+(-0.590757998421)*x[1]**o+(-0.793662033952)*x[1]
        arg[4,0,1]=(0.417712824349)*x[0]**o+(-0.138327612527)*x[0]+(-0.381595921766)*x[1]**o+(0.0287368364725)*x[1]
        arg[4,1,0]=(-0.83750540403)*x[0]**o+(0.247261273008)*x[0]+(-0.0769629479034)*x[1]**o+(0.0693501875686)*x[1]
        arg[4,1,1]=(0.697434044074)*x[0]**o+(0.396582403702)*x[0]+(0.736132107034)*x[1]**o+(0.482245183242)*x[1]
        arg[5,0,0]=(0.769936158184)*x[0]**o+(0.808871896459)*x[0]+(-0.559205314785)*x[1]**o+(-0.258014827146)*x[1]
        arg[5,0,1]=(0.667132553969)*x[0]**o+(0.618184492083)*x[0]+(0.475852440157)*x[1]**o+(0.084499293443)*x[1]
        arg[5,1,0]=(0.910882712823)*x[0]**o+(0.153897139268)*x[0]+(-0.61197280039)*x[1]**o+(0.441331985295)*x[1]
        arg[5,1,1]=(-0.290188623857)*x[0]**o+(0.318266320131)*x[0]+(0.9587856247)*x[1]**o+(0.311981486508)*x[1]
        ref[0,0,0]=(0.14066054662)/(o+1.)+(-0.330232377396)+(-0.33788666753)*0.5**o
        ref[0,0,1]=(-0.594786282119)/(o+1.)+(-0.0801196615418)+(0.0100961441611)*0.5**o
        ref[0,1,0]=(-0.8163725079)/(o+1.)+(-0.0659028641985)+(-0.118311175162)*0.5**o
        ref[0,1,1]=(0.0953383109011)/(o+1.)+(-0.141465155893)+(0.586991709411)*0.5**o
        ref[1,0,0]=(0.978704943828)/(o+1.)+(0.377909503343)+(-0.251125236294)*0.5**o
        ref[1,0,1]=(-0.919571042987)/(o+1.)+(-0.188664888997)+(0.0543182786738)*0.5**o
        ref[1,1,0]=(-0.227277029921)/(o+1.)+(0.143027401987)+(-0.867177621518)*0.5**o
        ref[1,1,1]=(0.425335668395)/(o+1.)+(-0.118935445857)+(0.85129514889)*0.5**o
        ref[2,0,0]=(-0.247374421196)/(o+1.)+(0.231963737807)+(-0.537930401585)*0.5**o
        ref[2,0,1]=(-0.0850059833654)/(o+1.)+(-0.214573443142)+(-0.101273864474)*0.5**o
        ref[2,1,0]=(0.40398578341)/(o+1.)+(-0.208635937884)+(0.684430088582)*0.5**o
        ref[2,1,1]=(-0.427822011518)/(o+1.)+(-0.0989808820785)+(-0.00801759014554)*0.5**o
        ref[3,0,0]=(0.0525765317703)/(o+1.)+(-0.550062401063)+(0.528563052182)*0.5**o
        ref[3,0,1]=(-0.785851578496)/(o+1.)+(0.475982524188)+(0.178622695427)*0.5**o
        ref[3,1,0]=(-0.831368961862)/(o+1.)+(-0.227574308319)+(0.547391926072)*0.5**o
        ref[3,1,1]=(0.649145894746)/(o+1.)+(-0.11250459761)+(-0.657011566453)*0.5**o
        ref[4,0,0]=(-0.590757998421)/(o+1.)+(-0.386468720427)+(-0.188305872652)*0.5**o
        ref[4,0,1]=(-0.381595921766)/(o+1.)+(-0.0547953880271)+(0.417712824349)*0.5**o
        ref[4,1,0]=(-0.0769629479034)/(o+1.)+(0.158305730288)+(-0.83750540403)*0.5**o
        ref[4,1,1]=(0.736132107034)/(o+1.)+(0.439413793472)+(0.697434044074)*0.5**o
        ref[5,0,0]=(-0.559205314785)/(o+1.)+(0.275428534657)+(0.769936158184)*0.5**o
        ref[5,0,1]=(0.475852440157)/(o+1.)+(0.351341892763)+(0.667132553969)*0.5**o
        ref[5,1,0]=(-0.61197280039)/(o+1.)+(0.297614562282)+(0.910882712823)*0.5**o
        ref[5,1,1]=(0.9587856247)/(o+1.)+(0.315123903319)+(-0.290188623857)*0.5**o
      else:
        arg[0,0,0]=(0.292346424086)*x[0]**o+(0.105058465241)*x[0]+(-0.281467283372)*x[1]**o+(0.828345267021)*x[1]+(0.741242558638)*x[2]**o+(0.0858780765977)*x[2]
        arg[0,0,1]=(-0.460556250835)*x[0]**o+(-0.845199338719)*x[0]+(-0.547000154821)*x[1]**o+(-0.972236450512)*x[1]+(0.823031420765)*x[2]**o+(0.278285294464)*x[2]
        arg[0,1,0]=(0.380027894487)*x[0]**o+(-0.579359210385)*x[0]+(-0.760849798908)*x[1]**o+(0.0065870459398)*x[1]+(-0.869735580799)*x[2]**o+(-0.937378284335)*x[2]
        arg[0,1,1]=(0.948682597897)*x[0]**o+(-0.882044080711)*x[0]+(-0.694398833899)*x[1]**o+(0.883635235205)*x[1]+(-0.473761903609)*x[2]**o+(0.961453193379)*x[2]
        arg[1,0,0]=(0.357218603696)*x[0]**o+(-0.694504294144)*x[0]+(-0.187119222895)*x[1]**o+(0.727463485088)*x[1]+(0.150410254875)*x[2]**o+(-0.373430202907)*x[2]
        arg[1,0,1]=(-0.776893233002)*x[0]**o+(0.565238637388)*x[0]+(-0.155674628662)*x[1]**o+(-0.938226259914)*x[1]+(0.575183633137)*x[2]**o+(0.319871583783)*x[2]
        arg[1,1,0]=(0.593512911569)*x[0]**o+(-0.0578762085927)*x[0]+(-0.913450847289)*x[1]**o+(0.7074653217)*x[1]+(0.991175636143)*x[2]**o+(-0.535632452281)*x[2]
        arg[1,1,1]=(-0.154542349616)*x[0]**o+(-0.841642050346)*x[0]+(0.878833704431)*x[1]**o+(0.85568524402)*x[1]+(-0.710374336281)*x[2]**o+(0.346540187252)*x[2]
        arg[2,0,0]=(-0.765160864305)*x[0]**o+(0.608017010487)*x[0]+(0.841265469171)*x[1]**o+(-0.409854579584)*x[1]+(0.728931324538)*x[2]**o+(0.121315209213)*x[2]
        arg[2,0,1]=(0.0397067301375)*x[0]**o+(-0.0451925117802)*x[0]+(-0.412724345414)*x[1]**o+(-0.447675002107)*x[1]+(0.141728105561)*x[2]**o+(0.923970189305)*x[2]
        arg[2,1,0]=(-0.237044096923)*x[0]**o+(-0.826168294669)*x[0]+(0.19833520121)*x[1]**o+(-0.796076984382)*x[1]+(0.367399642334)*x[2]**o+(0.71172848801)*x[2]
        arg[2,1,1]=(-0.3881830464)*x[0]**o+(0.988354722657)*x[0]+(0.762756805728)*x[1]**o+(0.551778683619)*x[1]+(0.975932105591)*x[2]**o+(-0.173192098292)*x[2]
        arg[3,0,0]=(-0.226163777371)*x[0]**o+(-0.301094659695)*x[0]+(0.0682567814743)*x[1]**o+(-0.154914902824)*x[1]+(0.984669238436)*x[2]**o+(-0.431011897738)*x[2]
        arg[3,0,1]=(-0.47230847571)*x[0]**o+(-0.579315893476)*x[0]+(-0.989289806396)*x[1]**o+(-0.673951334134)*x[1]+(0.98143767614)*x[2]**o+(-0.311074531383)*x[2]
        arg[3,1,0]=(-0.276309910109)*x[0]**o+(-0.70244304164)*x[0]+(-0.963669800079)*x[1]**o+(-0.947723474952)*x[1]+(0.513451083144)*x[2]**o+(-0.0651803692263)*x[2]
        arg[3,1,1]=(0.65175527234)*x[0]**o+(-0.294569414615)*x[0]+(-0.32846386125)*x[1]**o+(-0.868822318527)*x[1]+(0.268414480892)*x[2]**o+(-0.866835042217)*x[2]
        arg[4,0,0]=(0.304002845434)*x[0]**o+(0.938467246002)*x[0]+(-0.106648262772)*x[1]**o+(0.227863374571)*x[1]+(-0.789305204045)*x[2]**o+(-0.171526142133)*x[2]
        arg[4,0,1]=(0.70982305774)*x[0]**o+(0.404228697442)*x[0]+(0.190029230022)*x[1]**o+(-0.825276083634)*x[1]+(-0.739774636921)*x[2]**o+(-0.726529211281)*x[2]
        arg[4,1,0]=(-0.238487290076)*x[0]**o+(-0.4901234628)*x[0]+(-0.836670757945)*x[1]**o+(0.0841493712661)*x[1]+(-0.685144435198)*x[2]**o+(-0.304978244786)*x[2]
        arg[4,1,1]=(0.264176905985)*x[0]**o+(-0.410582158319)*x[0]+(-0.271915471538)*x[1]**o+(-0.996082427599)*x[1]+(-0.466654446945)*x[2]**o+(-0.723243781429)*x[2]
        arg[5,0,0]=(0.379390860481)*x[0]**o+(-0.78556086955)*x[0]+(0.189126836375)*x[1]**o+(0.235947881164)*x[1]+(0.149142076852)*x[2]**o+(-0.944419890441)*x[2]
        arg[5,0,1]=(0.768867280189)*x[0]**o+(-0.728364777249)*x[0]+(0.60414049451)*x[1]**o+(-0.538876171991)*x[1]+(-0.0947560131508)*x[2]**o+(0.474986163327)*x[2]
        arg[5,1,0]=(-0.228069986734)*x[0]**o+(-0.680841588894)*x[0]+(-0.663806510587)*x[1]**o+(-0.852775021416)*x[1]+(-0.59539582999)*x[2]**o+(0.52369159839)*x[2]
        arg[5,1,1]=(0.177872029425)*x[0]**o+(0.93423742256)*x[0]+(-0.0369475988192)*x[1]**o+(0.242202072343)*x[1]+(0.526144455402)*x[2]**o+(0.0217547142691)*x[2]
        ref[0,0,0]=(0.459775275267)/(o+1.)+(0.50964090443)+(0.292346424086)*0.5**o
        ref[0,0,1]=(0.276031265944)/(o+1.)+(-0.769575247383)+(-0.460556250835)*0.5**o
        ref[0,1,0]=(-1.63058537971)/(o+1.)+(-0.75507522439)+(0.380027894487)*0.5**o
        ref[0,1,1]=(-1.16816073751)/(o+1.)+(0.481522173937)+(0.948682597897)*0.5**o
        ref[1,0,0]=(-0.0367089680209)/(o+1.)+(-0.170235505982)+(0.357218603696)*0.5**o
        ref[1,0,1]=(0.419509004475)/(o+1.)+(-0.0265580193719)+(-0.776893233002)*0.5**o
        ref[1,1,0]=(0.0777247888533)/(o+1.)+(0.0569783304129)+(0.593512911569)*0.5**o
        ref[1,1,1]=(0.168459368149)/(o+1.)+(0.180291690463)+(-0.154542349616)*0.5**o
        ref[2,0,0]=(1.57019679371)/(o+1.)+(0.159738820058)+(-0.765160864305)*0.5**o
        ref[2,0,1]=(-0.270996239853)/(o+1.)+(0.215551337709)+(0.0397067301375)*0.5**o
        ref[2,1,0]=(0.565734843545)/(o+1.)+(-0.455258395521)+(-0.237044096923)*0.5**o
        ref[2,1,1]=(1.73868891132)/(o+1.)+(0.683470653992)+(-0.3881830464)*0.5**o
        ref[3,0,0]=(1.05292601991)/(o+1.)+(-0.443510730129)+(-0.226163777371)*0.5**o
        ref[3,0,1]=(-0.00785213025503)/(o+1.)+(-0.782170879496)+(-0.47230847571)*0.5**o
        ref[3,1,0]=(-0.450218716935)/(o+1.)+(-0.857673442909)+(-0.276309910109)*0.5**o
        ref[3,1,1]=(-0.0600493803575)/(o+1.)+(-1.01511338768)+(0.65175527234)*0.5**o
        ref[4,0,0]=(-0.895953466817)/(o+1.)+(0.49740223922)+(0.304002845434)*0.5**o
        ref[4,0,1]=(-0.549745406899)/(o+1.)+(-0.573788298736)+(0.70982305774)*0.5**o
        ref[4,1,0]=(-1.52181519314)/(o+1.)+(-0.35547616816)+(-0.238487290076)*0.5**o
        ref[4,1,1]=(-0.738569918483)/(o+1.)+(-1.06495418367)+(0.264176905985)*0.5**o
        ref[5,0,0]=(0.338268913228)/(o+1.)+(-0.747016439414)+(0.379390860481)*0.5**o
        ref[5,0,1]=(0.50938448136)/(o+1.)+(-0.396127392956)+(0.768867280189)*0.5**o
        ref[5,1,0]=(-1.25920234058)/(o+1.)+(-0.50496250596)+(-0.228069986734)*0.5**o
        ref[5,1,1]=(0.489196856583)/(o+1.)+(0.599097104586)+(0.177872029425)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactZero

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.423100205741)*x[0]**o+(-0.681307394941)*x[0]+(-0.848175950825)*x[1]**o+(0.559338235783)*x[1]
        arg[0,0,0,1]=(-0.739687474039)*x[0]**o+(-0.202202358237)*x[0]+(-0.838397517739)*x[1]**o+(-0.377580621185)*x[1]
        arg[0,0,1,0]=(-0.423462958438)*x[0]**o+(-0.335929684088)*x[0]+(-0.616976579504)*x[1]**o+(-0.77803694581)*x[1]
        arg[0,0,1,1]=(-0.432533121428)*x[0]**o+(0.0723865130937)*x[0]+(-0.953014661461)*x[1]**o+(0.92761707252)*x[1]
        arg[0,0,2,0]=(-0.891613775753)*x[0]**o+(0.780592229933)*x[0]+(-0.156191993934)*x[1]**o+(0.985464289737)*x[1]
        arg[0,0,2,1]=(0.909573147888)*x[0]**o+(-0.225603851475)*x[0]+(-0.68287286477)*x[1]**o+(0.330105905794)*x[1]
        arg[0,1,0,0]=(-0.39808363424)*x[0]**o+(-0.290872560992)*x[0]+(-0.631624427826)*x[1]**o+(-0.389239289826)*x[1]
        arg[0,1,0,1]=(-0.748658247224)*x[0]**o+(0.22314629362)*x[0]+(-0.0750972377036)*x[1]**o+(0.908914722147)*x[1]
        arg[0,1,1,0]=(0.877627912148)*x[0]**o+(-0.988434656301)*x[0]+(-0.163646238293)*x[1]**o+(-0.169410711361)*x[1]
        arg[0,1,1,1]=(-0.592580581615)*x[0]**o+(0.997989912544)*x[0]+(0.0667564973731)*x[1]**o+(-0.121580259121)*x[1]
        arg[0,1,2,0]=(-0.505569292541)*x[0]**o+(-0.534428841308)*x[0]+(-0.899276958155)*x[1]**o+(-0.833972415825)*x[1]
        arg[0,1,2,1]=(-0.215752366701)*x[0]**o+(-0.249993262232)*x[0]+(0.912452712698)*x[1]**o+(0.0601524115958)*x[1]
        arg[0,2,0,0]=(0.684112319204)*x[0]**o+(0.343018234122)*x[0]+(0.267409202785)*x[1]**o+(0.7781133929)*x[1]
        arg[0,2,0,1]=(0.552438501301)*x[0]**o+(-0.101558113559)*x[0]+(-0.152472725029)*x[1]**o+(-0.8771654517)*x[1]
        arg[0,2,1,0]=(-0.31213422297)*x[0]**o+(0.597042319279)*x[0]+(-0.873134889407)*x[1]**o+(-0.68920786345)*x[1]
        arg[0,2,1,1]=(0.883048532408)*x[0]**o+(-0.653997880306)*x[0]+(-0.761310589424)*x[1]**o+(-0.698342623962)*x[1]
        arg[0,2,2,0]=(-0.324489237508)*x[0]**o+(-0.717724692716)*x[0]+(-0.936381308418)*x[1]**o+(0.712416971077)*x[1]
        arg[0,2,2,1]=(-0.971735193469)*x[0]**o+(0.0943629094751)*x[0]+(0.419811305063)*x[1]**o+(0.125623316569)*x[1]
        arg[0,3,0,0]=(-0.305553454905)*x[0]**o+(-0.936112994503)*x[0]+(0.707279059071)*x[1]**o+(-0.794754809337)*x[1]
        arg[0,3,0,1]=(0.803934620926)*x[0]**o+(-0.632020033455)*x[0]+(0.819264129099)*x[1]**o+(0.770744069672)*x[1]
        arg[0,3,1,0]=(0.996632326055)*x[0]**o+(-0.537850178404)*x[0]+(-0.60262522048)*x[1]**o+(-0.431667704232)*x[1]
        arg[0,3,1,1]=(0.402799911553)*x[0]**o+(0.155798565063)*x[0]+(0.722837842981)*x[1]**o+(0.462463978227)*x[1]
        arg[0,3,2,0]=(0.937966588808)*x[0]**o+(-0.797151899551)*x[0]+(0.62363567076)*x[1]**o+(-0.60458214868)*x[1]
        arg[0,3,2,1]=(-0.0389713032301)*x[0]**o+(0.497820326557)*x[0]+(-0.950660964237)*x[1]**o+(0.0247372614337)*x[1]
        arg[0,4,0,0]=(-0.589529640529)*x[0]**o+(-0.71761078076)*x[0]+(-0.872148906918)*x[1]**o+(-0.666618835503)*x[1]
        arg[0,4,0,1]=(0.78615039967)*x[0]**o+(0.99287672137)*x[0]+(0.329194160571)*x[1]**o+(0.309370990805)*x[1]
        arg[0,4,1,0]=(-0.801187683466)*x[0]**o+(0.476957978386)*x[0]+(0.107007555422)*x[1]**o+(-0.211909676249)*x[1]
        arg[0,4,1,1]=(0.817551079419)*x[0]**o+(0.958544826763)*x[0]+(-0.720479146591)*x[1]**o+(0.0769824586505)*x[1]
        arg[0,4,2,0]=(-0.245391293539)*x[0]**o+(0.240588876322)*x[0]+(0.766214937295)*x[1]**o+(0.625282150348)*x[1]
        arg[0,4,2,1]=(-0.780163545255)*x[0]**o+(-0.84424180556)*x[0]+(-0.993826355778)*x[1]**o+(-0.708432857746)*x[1]
        arg[1,0,0,0]=(0.0998104654851)*x[0]**o+(-0.933870759023)*x[0]+(-0.00667060999489)*x[1]**o+(-0.304328472241)*x[1]
        arg[1,0,0,1]=(0.716858223574)*x[0]**o+(0.0907083966301)*x[0]+(0.118004968786)*x[1]**o+(-0.989434356769)*x[1]
        arg[1,0,1,0]=(0.433505460082)*x[0]**o+(0.927792393402)*x[0]+(0.10117572965)*x[1]**o+(-0.719175235748)*x[1]
        arg[1,0,1,1]=(0.93394463477)*x[0]**o+(0.30554551436)*x[0]+(-0.221528653572)*x[1]**o+(-0.965153365935)*x[1]
        arg[1,0,2,0]=(0.217926136552)*x[0]**o+(0.75089182057)*x[0]+(0.798438119547)*x[1]**o+(-0.131885152617)*x[1]
        arg[1,0,2,1]=(0.489885370386)*x[0]**o+(-0.958785443386)*x[0]+(-0.934023631551)*x[1]**o+(-0.120727978484)*x[1]
        arg[1,1,0,0]=(-0.104659311537)*x[0]**o+(0.323817119133)*x[0]+(0.207178629335)*x[1]**o+(0.132677138305)*x[1]
        arg[1,1,0,1]=(0.855593354128)*x[0]**o+(-0.319099268106)*x[0]+(-0.75522659926)*x[1]**o+(0.398648791668)*x[1]
        arg[1,1,1,0]=(0.165499959353)*x[0]**o+(0.025416591231)*x[0]+(-0.454711374202)*x[1]**o+(-0.149643847094)*x[1]
        arg[1,1,1,1]=(0.668195333828)*x[0]**o+(-0.954315351102)*x[0]+(-0.15549992542)*x[1]**o+(0.960203842246)*x[1]
        arg[1,1,2,0]=(-0.0830862028828)*x[0]**o+(0.104413343709)*x[0]+(0.0666836881524)*x[1]**o+(-0.909918510114)*x[1]
        arg[1,1,2,1]=(-0.614061685098)*x[0]**o+(-0.696536572273)*x[0]+(0.674867602585)*x[1]**o+(0.742233774001)*x[1]
        arg[1,2,0,0]=(-0.271025278576)*x[0]**o+(0.194146585691)*x[0]+(0.369545867263)*x[1]**o+(-0.444209773003)*x[1]
        arg[1,2,0,1]=(-0.646570016511)*x[0]**o+(-0.1181935343)*x[0]+(-0.473468176784)*x[1]**o+(0.980639666617)*x[1]
        arg[1,2,1,0]=(-0.274558546949)*x[0]**o+(-0.994564294804)*x[0]+(-0.710677848448)*x[1]**o+(-0.471981673909)*x[1]
        arg[1,2,1,1]=(0.920326662139)*x[0]**o+(-0.503477706392)*x[0]+(-0.488075258363)*x[1]**o+(-0.390701293427)*x[1]
        arg[1,2,2,0]=(-0.734779609501)*x[0]**o+(-0.211522964885)*x[0]+(-0.209704158218)*x[1]**o+(0.602143419316)*x[1]
        arg[1,2,2,1]=(-0.71042927509)*x[0]**o+(-0.320801535296)*x[0]+(-0.307645214006)*x[1]**o+(0.902857741355)*x[1]
        arg[1,3,0,0]=(0.430044544461)*x[0]**o+(-0.321809854969)*x[0]+(-0.202705286059)*x[1]**o+(-0.136245903236)*x[1]
        arg[1,3,0,1]=(-0.0215232394148)*x[0]**o+(0.560515199075)*x[0]+(0.183422356152)*x[1]**o+(0.0500950744933)*x[1]
        arg[1,3,1,0]=(0.956072319541)*x[0]**o+(-0.910407504398)*x[0]+(0.710351406673)*x[1]**o+(-0.673149941902)*x[1]
        arg[1,3,1,1]=(0.512105745894)*x[0]**o+(0.617656019287)*x[0]+(0.178878380134)*x[1]**o+(-0.304972517759)*x[1]
        arg[1,3,2,0]=(0.505194196756)*x[0]**o+(-0.588278604691)*x[0]+(0.657312673212)*x[1]**o+(-0.198666173753)*x[1]
        arg[1,3,2,1]=(0.562199962955)*x[0]**o+(-0.855080919048)*x[0]+(0.798106457622)*x[1]**o+(-0.462368594285)*x[1]
        arg[1,4,0,0]=(-0.95103819151)*x[0]**o+(0.996179304819)*x[0]+(0.720809353558)*x[1]**o+(-0.357882371835)*x[1]
        arg[1,4,0,1]=(-0.337212448334)*x[0]**o+(-0.647939092937)*x[0]+(0.432152553594)*x[1]**o+(0.284393585377)*x[1]
        arg[1,4,1,0]=(0.286170290079)*x[0]**o+(0.813168170391)*x[0]+(-0.178814496497)*x[1]**o+(0.414687546663)*x[1]
        arg[1,4,1,1]=(0.00961378883254)*x[0]**o+(-0.946718643529)*x[0]+(0.828780485661)*x[1]**o+(0.54461217692)*x[1]
        arg[1,4,2,0]=(0.784403348421)*x[0]**o+(0.430078584466)*x[0]+(-0.542741131881)*x[1]**o+(0.0290828265624)*x[1]
        arg[1,4,2,1]=(-0.246343667616)*x[0]**o+(-0.27406067089)*x[0]+(-0.938704878102)*x[1]**o+(0.907921504301)*x[1]
        arg[2,0,0,0]=(-0.551056874357)*x[0]**o+(-0.697006980356)*x[0]+(-0.733447125487)*x[1]**o+(-0.241902075189)*x[1]
        arg[2,0,0,1]=(-0.75556701159)*x[0]**o+(0.155303331948)*x[0]+(0.480546881289)*x[1]**o+(-0.420954472057)*x[1]
        arg[2,0,1,0]=(0.12990139506)*x[0]**o+(-0.0990388974263)*x[0]+(-0.913089275462)*x[1]**o+(-0.766316334455)*x[1]
        arg[2,0,1,1]=(-0.96760564527)*x[0]**o+(0.94596820894)*x[0]+(-0.208507053422)*x[1]**o+(0.82642580152)*x[1]
        arg[2,0,2,0]=(-0.801660576339)*x[0]**o+(0.724091208253)*x[0]+(-0.416451086696)*x[1]**o+(-0.255959329575)*x[1]
        arg[2,0,2,1]=(0.380233654687)*x[0]**o+(-0.27467061721)*x[0]+(-0.626861863133)*x[1]**o+(0.498725464891)*x[1]
        arg[2,1,0,0]=(0.733227197081)*x[0]**o+(-0.111622822078)*x[0]+(-0.199363221943)*x[1]**o+(-0.449664634046)*x[1]
        arg[2,1,0,1]=(0.381774304841)*x[0]**o+(0.914208421879)*x[0]+(-0.545958908798)*x[1]**o+(-0.344721917598)*x[1]
        arg[2,1,1,0]=(-0.861412661903)*x[0]**o+(-0.202264578976)*x[0]+(-0.225404313047)*x[1]**o+(0.716189374265)*x[1]
        arg[2,1,1,1]=(-0.180088205315)*x[0]**o+(0.133136657032)*x[0]+(-0.341650807958)*x[1]**o+(-0.570348715256)*x[1]
        arg[2,1,2,0]=(-0.626863150737)*x[0]**o+(-0.236854925766)*x[0]+(-0.271186155613)*x[1]**o+(-0.750500826476)*x[1]
        arg[2,1,2,1]=(-0.138580929054)*x[0]**o+(-0.925057205334)*x[0]+(0.404961686669)*x[1]**o+(0.542298342705)*x[1]
        arg[2,2,0,0]=(0.453186536783)*x[0]**o+(0.368065177204)*x[0]+(0.545211472015)*x[1]**o+(0.334551805933)*x[1]
        arg[2,2,0,1]=(0.0639699038803)*x[0]**o+(0.154192875522)*x[0]+(0.101347900441)*x[1]**o+(-0.668313576443)*x[1]
        arg[2,2,1,0]=(0.465173445732)*x[0]**o+(-0.475009875122)*x[0]+(-0.748444748361)*x[1]**o+(0.620845514745)*x[1]
        arg[2,2,1,1]=(0.720975965615)*x[0]**o+(-0.266631032721)*x[0]+(0.0996389916136)*x[1]**o+(-0.96886977517)*x[1]
        arg[2,2,2,0]=(-0.396391719953)*x[0]**o+(-0.240091626001)*x[0]+(-0.42411685734)*x[1]**o+(-0.458446010754)*x[1]
        arg[2,2,2,1]=(-0.662584035264)*x[0]**o+(0.23813614536)*x[0]+(0.810893160816)*x[1]**o+(-0.130143663026)*x[1]
        arg[2,3,0,0]=(-0.651550480594)*x[0]**o+(0.370844612999)*x[0]+(-0.808002027275)*x[1]**o+(0.531083477622)*x[1]
        arg[2,3,0,1]=(-0.287184182382)*x[0]**o+(0.928877646221)*x[0]+(-0.271563784371)*x[1]**o+(-0.46908309398)*x[1]
        arg[2,3,1,0]=(0.342080138259)*x[0]**o+(0.458994887529)*x[0]+(-0.544475387926)*x[1]**o+(-0.203224249914)*x[1]
        arg[2,3,1,1]=(-0.0791009979206)*x[0]**o+(0.536480501773)*x[0]+(-0.624008015173)*x[1]**o+(-0.380504135174)*x[1]
        arg[2,3,2,0]=(-0.707089360149)*x[0]**o+(0.217716900812)*x[0]+(0.020441009853)*x[1]**o+(0.680214150153)*x[1]
        arg[2,3,2,1]=(0.0368635610611)*x[0]**o+(0.646697307691)*x[0]+(0.36118713723)*x[1]**o+(-0.535295583008)*x[1]
        arg[2,4,0,0]=(0.806938680905)*x[0]**o+(0.449096825134)*x[0]+(-0.43507826254)*x[1]**o+(-0.300956286252)*x[1]
        arg[2,4,0,1]=(-0.268005914326)*x[0]**o+(0.583132947291)*x[0]+(-0.344067495236)*x[1]**o+(-0.526011968282)*x[1]
        arg[2,4,1,0]=(-0.0953663458348)*x[0]**o+(0.834765967023)*x[0]+(-0.0410345097335)*x[1]**o+(0.493198054713)*x[1]
        arg[2,4,1,1]=(0.964777812988)*x[0]**o+(-0.791397009733)*x[0]+(-0.962108127047)*x[1]**o+(0.378083818799)*x[1]
        arg[2,4,2,0]=(0.0977715857241)*x[0]**o+(-0.118265910205)*x[0]+(-0.0181736456148)*x[1]**o+(-0.480499109439)*x[1]
        arg[2,4,2,1]=(0.396936262802)*x[0]**o+(0.979807961416)*x[0]+(0.383251286227)*x[1]**o+(-0.921599329372)*x[1]
        arg[3,0,0,0]=(-0.193869568904)*x[0]**o+(0.335533150532)*x[0]+(-0.542087879921)*x[1]**o+(0.916531699685)*x[1]
        arg[3,0,0,1]=(0.732192438113)*x[0]**o+(0.610866210623)*x[0]+(0.232455584069)*x[1]**o+(0.109577342902)*x[1]
        arg[3,0,1,0]=(0.0487640958479)*x[0]**o+(-0.250229114846)*x[0]+(-0.539343546695)*x[1]**o+(-0.456485153733)*x[1]
        arg[3,0,1,1]=(0.951844805407)*x[0]**o+(-0.811442383281)*x[0]+(-0.537581471096)*x[1]**o+(0.350217063609)*x[1]
        arg[3,0,2,0]=(0.495490307938)*x[0]**o+(0.529667417659)*x[0]+(-0.853447043865)*x[1]**o+(0.97002519322)*x[1]
        arg[3,0,2,1]=(0.38188303469)*x[0]**o+(-0.75911672537)*x[0]+(-0.67563531103)*x[1]**o+(0.384002769031)*x[1]
        arg[3,1,0,0]=(0.44471228922)*x[0]**o+(-0.958995287488)*x[0]+(-0.356730659314)*x[1]**o+(0.853749081247)*x[1]
        arg[3,1,0,1]=(-0.674677862576)*x[0]**o+(-0.651945231545)*x[0]+(-0.869871613483)*x[1]**o+(0.518493292422)*x[1]
        arg[3,1,1,0]=(-0.803923033456)*x[0]**o+(0.916638584514)*x[0]+(-0.410056071233)*x[1]**o+(-0.313811798732)*x[1]
        arg[3,1,1,1]=(-0.677586574312)*x[0]**o+(0.101317771211)*x[0]+(-0.848629733526)*x[1]**o+(0.956935407827)*x[1]
        arg[3,1,2,0]=(0.723956969878)*x[0]**o+(-0.677929268987)*x[0]+(0.168694427833)*x[1]**o+(0.675664564011)*x[1]
        arg[3,1,2,1]=(-0.0961257584415)*x[0]**o+(0.870273004118)*x[0]+(0.0979731186867)*x[1]**o+(-0.561786099655)*x[1]
        arg[3,2,0,0]=(-0.506857925404)*x[0]**o+(0.950429186893)*x[0]+(-0.765686972231)*x[1]**o+(0.300787357163)*x[1]
        arg[3,2,0,1]=(-0.479464705052)*x[0]**o+(-0.814881390688)*x[0]+(-0.549033177495)*x[1]**o+(0.875164713771)*x[1]
        arg[3,2,1,0]=(0.984579775952)*x[0]**o+(0.822464384935)*x[0]+(-0.352937841183)*x[1]**o+(-0.389516730371)*x[1]
        arg[3,2,1,1]=(-0.75230207503)*x[0]**o+(0.921759892886)*x[0]+(0.683120790039)*x[1]**o+(0.470250085193)*x[1]
        arg[3,2,2,0]=(-0.257935272452)*x[0]**o+(-0.437370511818)*x[0]+(-0.350363036315)*x[1]**o+(-0.232415372202)*x[1]
        arg[3,2,2,1]=(0.0784702766009)*x[0]**o+(0.460959683819)*x[0]+(0.758201028024)*x[1]**o+(0.101597092257)*x[1]
        arg[3,3,0,0]=(0.292597181224)*x[0]**o+(-0.322992033991)*x[0]+(-0.536413055541)*x[1]**o+(-0.95609054732)*x[1]
        arg[3,3,0,1]=(0.174134715952)*x[0]**o+(0.811729326606)*x[0]+(0.89040777319)*x[1]**o+(0.893572419966)*x[1]
        arg[3,3,1,0]=(0.768234215847)*x[0]**o+(0.36391719033)*x[0]+(0.598889528719)*x[1]**o+(0.152035960251)*x[1]
        arg[3,3,1,1]=(-0.77938361663)*x[0]**o+(-0.448945619013)*x[0]+(0.471242869055)*x[1]**o+(0.112888214554)*x[1]
        arg[3,3,2,0]=(-0.00617753872016)*x[0]**o+(0.0155218100192)*x[0]+(0.674100983537)*x[1]**o+(0.211406831421)*x[1]
        arg[3,3,2,1]=(0.124470359441)*x[0]**o+(0.117393617733)*x[0]+(-0.297791856337)*x[1]**o+(-0.210405798241)*x[1]
        arg[3,4,0,0]=(0.139920132608)*x[0]**o+(-0.0073381380034)*x[0]+(-0.77267392403)*x[1]**o+(0.792299687107)*x[1]
        arg[3,4,0,1]=(0.328592839797)*x[0]**o+(-0.758899359217)*x[0]+(0.730856047416)*x[1]**o+(-0.389548922357)*x[1]
        arg[3,4,1,0]=(-0.239362896774)*x[0]**o+(0.513205289336)*x[0]+(-0.393781550386)*x[1]**o+(0.229968513732)*x[1]
        arg[3,4,1,1]=(0.777985150139)*x[0]**o+(0.273278793895)*x[0]+(0.17782134987)*x[1]**o+(0.491127331923)*x[1]
        arg[3,4,2,0]=(0.348482781008)*x[0]**o+(0.458696185739)*x[0]+(-0.77882422986)*x[1]**o+(0.455739446004)*x[1]
        arg[3,4,2,1]=(-0.396751338134)*x[0]**o+(0.449026376364)*x[0]+(0.308553967581)*x[1]**o+(-0.777653415416)*x[1]
        ref[0,0,0,0]=(-0.848175950825)/(o+1.)+(-0.060984579579)+(0.423100205741)*0.5**o
        ref[0,0,0,1]=(-0.838397517739)/(o+1.)+(-0.289891489711)+(-0.739687474039)*0.5**o
        ref[0,0,1,0]=(-0.616976579504)/(o+1.)+(-0.556983314949)+(-0.423462958438)*0.5**o
        ref[0,0,1,1]=(-0.953014661461)/(o+1.)+(0.500001792807)+(-0.432533121428)*0.5**o
        ref[0,0,2,0]=(-0.156191993934)/(o+1.)+(0.883028259835)+(-0.891613775753)*0.5**o
        ref[0,0,2,1]=(-0.68287286477)/(o+1.)+(0.0522510271596)+(0.909573147888)*0.5**o
        ref[0,1,0,0]=(-0.631624427826)/(o+1.)+(-0.340055925409)+(-0.39808363424)*0.5**o
        ref[0,1,0,1]=(-0.0750972377036)/(o+1.)+(0.566030507884)+(-0.748658247224)*0.5**o
        ref[0,1,1,0]=(-0.163646238293)/(o+1.)+(-0.578922683831)+(0.877627912148)*0.5**o
        ref[0,1,1,1]=(0.0667564973731)/(o+1.)+(0.438204826712)+(-0.592580581615)*0.5**o
        ref[0,1,2,0]=(-0.899276958155)/(o+1.)+(-0.684200628566)+(-0.505569292541)*0.5**o
        ref[0,1,2,1]=(0.912452712698)/(o+1.)+(-0.0949204253183)+(-0.215752366701)*0.5**o
        ref[0,2,0,0]=(0.267409202785)/(o+1.)+(0.560565813511)+(0.684112319204)*0.5**o
        ref[0,2,0,1]=(-0.152472725029)/(o+1.)+(-0.48936178263)+(0.552438501301)*0.5**o
        ref[0,2,1,0]=(-0.873134889407)/(o+1.)+(-0.0460827720855)+(-0.31213422297)*0.5**o
        ref[0,2,1,1]=(-0.761310589424)/(o+1.)+(-0.676170252134)+(0.883048532408)*0.5**o
        ref[0,2,2,0]=(-0.936381308418)/(o+1.)+(-0.00265386081956)+(-0.324489237508)*0.5**o
        ref[0,2,2,1]=(0.419811305063)/(o+1.)+(0.109993113022)+(-0.971735193469)*0.5**o
        ref[0,3,0,0]=(0.707279059071)/(o+1.)+(-0.86543390192)+(-0.305553454905)*0.5**o
        ref[0,3,0,1]=(0.819264129099)/(o+1.)+(0.0693620181083)+(0.803934620926)*0.5**o
        ref[0,3,1,0]=(-0.60262522048)/(o+1.)+(-0.484758941318)+(0.996632326055)*0.5**o
        ref[0,3,1,1]=(0.722837842981)/(o+1.)+(0.309131271645)+(0.402799911553)*0.5**o
        ref[0,3,2,0]=(0.62363567076)/(o+1.)+(-0.700867024115)+(0.937966588808)*0.5**o
        ref[0,3,2,1]=(-0.950660964237)/(o+1.)+(0.261278793995)+(-0.0389713032301)*0.5**o
        ref[0,4,0,0]=(-0.872148906918)/(o+1.)+(-0.692114808131)+(-0.589529640529)*0.5**o
        ref[0,4,0,1]=(0.329194160571)/(o+1.)+(0.651123856087)+(0.78615039967)*0.5**o
        ref[0,4,1,0]=(0.107007555422)/(o+1.)+(0.132524151069)+(-0.801187683466)*0.5**o
        ref[0,4,1,1]=(-0.720479146591)/(o+1.)+(0.517763642707)+(0.817551079419)*0.5**o
        ref[0,4,2,0]=(0.766214937295)/(o+1.)+(0.432935513335)+(-0.245391293539)*0.5**o
        ref[0,4,2,1]=(-0.993826355778)/(o+1.)+(-0.776337331653)+(-0.780163545255)*0.5**o
        ref[1,0,0,0]=(-0.00667060999489)/(o+1.)+(-0.619099615632)+(0.0998104654851)*0.5**o
        ref[1,0,0,1]=(0.118004968786)/(o+1.)+(-0.449362980069)+(0.716858223574)*0.5**o
        ref[1,0,1,0]=(0.10117572965)/(o+1.)+(0.104308578827)+(0.433505460082)*0.5**o
        ref[1,0,1,1]=(-0.221528653572)/(o+1.)+(-0.329803925788)+(0.93394463477)*0.5**o
        ref[1,0,2,0]=(0.798438119547)/(o+1.)+(0.309503333977)+(0.217926136552)*0.5**o
        ref[1,0,2,1]=(-0.934023631551)/(o+1.)+(-0.539756710935)+(0.489885370386)*0.5**o
        ref[1,1,0,0]=(0.207178629335)/(o+1.)+(0.228247128719)+(-0.104659311537)*0.5**o
        ref[1,1,0,1]=(-0.75522659926)/(o+1.)+(0.0397747617809)+(0.855593354128)*0.5**o
        ref[1,1,1,0]=(-0.454711374202)/(o+1.)+(-0.0621136279316)+(0.165499959353)*0.5**o
        ref[1,1,1,1]=(-0.15549992542)/(o+1.)+(0.00294424557217)+(0.668195333828)*0.5**o
        ref[1,1,2,0]=(0.0666836881524)/(o+1.)+(-0.402752583203)+(-0.0830862028828)*0.5**o
        ref[1,1,2,1]=(0.674867602585)/(o+1.)+(0.0228486008642)+(-0.614061685098)*0.5**o
        ref[1,2,0,0]=(0.369545867263)/(o+1.)+(-0.125031593656)+(-0.271025278576)*0.5**o
        ref[1,2,0,1]=(-0.473468176784)/(o+1.)+(0.431223066159)+(-0.646570016511)*0.5**o
        ref[1,2,1,0]=(-0.710677848448)/(o+1.)+(-0.733272984356)+(-0.274558546949)*0.5**o
        ref[1,2,1,1]=(-0.488075258363)/(o+1.)+(-0.44708949991)+(0.920326662139)*0.5**o
        ref[1,2,2,0]=(-0.209704158218)/(o+1.)+(0.195310227215)+(-0.734779609501)*0.5**o
        ref[1,2,2,1]=(-0.307645214006)/(o+1.)+(0.29102810303)+(-0.71042927509)*0.5**o
        ref[1,3,0,0]=(-0.202705286059)/(o+1.)+(-0.229027879102)+(0.430044544461)*0.5**o
        ref[1,3,0,1]=(0.183422356152)/(o+1.)+(0.305305136784)+(-0.0215232394148)*0.5**o
        ref[1,3,1,0]=(0.710351406673)/(o+1.)+(-0.79177872315)+(0.956072319541)*0.5**o
        ref[1,3,1,1]=(0.178878380134)/(o+1.)+(0.156341750764)+(0.512105745894)*0.5**o
        ref[1,3,2,0]=(0.657312673212)/(o+1.)+(-0.393472389222)+(0.505194196756)*0.5**o
        ref[1,3,2,1]=(0.798106457622)/(o+1.)+(-0.658724756667)+(0.562199962955)*0.5**o
        ref[1,4,0,0]=(0.720809353558)/(o+1.)+(0.319148466492)+(-0.95103819151)*0.5**o
        ref[1,4,0,1]=(0.432152553594)/(o+1.)+(-0.18177275378)+(-0.337212448334)*0.5**o
        ref[1,4,1,0]=(-0.178814496497)/(o+1.)+(0.613927858527)+(0.286170290079)*0.5**o
        ref[1,4,1,1]=(0.828780485661)/(o+1.)+(-0.201053233304)+(0.00961378883254)*0.5**o
        ref[1,4,2,0]=(-0.542741131881)/(o+1.)+(0.229580705514)+(0.784403348421)*0.5**o
        ref[1,4,2,1]=(-0.938704878102)/(o+1.)+(0.316930416706)+(-0.246343667616)*0.5**o
        ref[2,0,0,0]=(-0.733447125487)/(o+1.)+(-0.469454527772)+(-0.551056874357)*0.5**o
        ref[2,0,0,1]=(0.480546881289)/(o+1.)+(-0.132825570055)+(-0.75556701159)*0.5**o
        ref[2,0,1,0]=(-0.913089275462)/(o+1.)+(-0.432677615941)+(0.12990139506)*0.5**o
        ref[2,0,1,1]=(-0.208507053422)/(o+1.)+(0.88619700523)+(-0.96760564527)*0.5**o
        ref[2,0,2,0]=(-0.416451086696)/(o+1.)+(0.234065939339)+(-0.801660576339)*0.5**o
        ref[2,0,2,1]=(-0.626861863133)/(o+1.)+(0.112027423841)+(0.380233654687)*0.5**o
        ref[2,1,0,0]=(-0.199363221943)/(o+1.)+(-0.280643728062)+(0.733227197081)*0.5**o
        ref[2,1,0,1]=(-0.545958908798)/(o+1.)+(0.284743252141)+(0.381774304841)*0.5**o
        ref[2,1,1,0]=(-0.225404313047)/(o+1.)+(0.256962397644)+(-0.861412661903)*0.5**o
        ref[2,1,1,1]=(-0.341650807958)/(o+1.)+(-0.218606029112)+(-0.180088205315)*0.5**o
        ref[2,1,2,0]=(-0.271186155613)/(o+1.)+(-0.493677876121)+(-0.626863150737)*0.5**o
        ref[2,1,2,1]=(0.404961686669)/(o+1.)+(-0.191379431315)+(-0.138580929054)*0.5**o
        ref[2,2,0,0]=(0.545211472015)/(o+1.)+(0.351308491569)+(0.453186536783)*0.5**o
        ref[2,2,0,1]=(0.101347900441)/(o+1.)+(-0.257060350461)+(0.0639699038803)*0.5**o
        ref[2,2,1,0]=(-0.748444748361)/(o+1.)+(0.0729178198114)+(0.465173445732)*0.5**o
        ref[2,2,1,1]=(0.0996389916136)/(o+1.)+(-0.617750403946)+(0.720975965615)*0.5**o
        ref[2,2,2,0]=(-0.42411685734)/(o+1.)+(-0.349268818378)+(-0.396391719953)*0.5**o
        ref[2,2,2,1]=(0.810893160816)/(o+1.)+(0.053996241167)+(-0.662584035264)*0.5**o
        ref[2,3,0,0]=(-0.808002027275)/(o+1.)+(0.45096404531)+(-0.651550480594)*0.5**o
        ref[2,3,0,1]=(-0.271563784371)/(o+1.)+(0.22989727612)+(-0.287184182382)*0.5**o
        ref[2,3,1,0]=(-0.544475387926)/(o+1.)+(0.127885318807)+(0.342080138259)*0.5**o
        ref[2,3,1,1]=(-0.624008015173)/(o+1.)+(0.0779881832997)+(-0.0791009979206)*0.5**o
        ref[2,3,2,0]=(0.020441009853)/(o+1.)+(0.448965525483)+(-0.707089360149)*0.5**o
        ref[2,3,2,1]=(0.36118713723)/(o+1.)+(0.0557008623413)+(0.0368635610611)*0.5**o
        ref[2,4,0,0]=(-0.43507826254)/(o+1.)+(0.0740702694408)+(0.806938680905)*0.5**o
        ref[2,4,0,1]=(-0.344067495236)/(o+1.)+(0.0285604895043)+(-0.268005914326)*0.5**o
        ref[2,4,1,0]=(-0.0410345097335)/(o+1.)+(0.663982010868)+(-0.0953663458348)*0.5**o
        ref[2,4,1,1]=(-0.962108127047)/(o+1.)+(-0.206656595467)+(0.964777812988)*0.5**o
        ref[2,4,2,0]=(-0.0181736456148)/(o+1.)+(-0.299382509822)+(0.0977715857241)*0.5**o
        ref[2,4,2,1]=(0.383251286227)/(o+1.)+(0.0291043160221)+(0.396936262802)*0.5**o
        ref[3,0,0,0]=(-0.542087879921)/(o+1.)+(0.626032425109)+(-0.193869568904)*0.5**o
        ref[3,0,0,1]=(0.232455584069)/(o+1.)+(0.360221776762)+(0.732192438113)*0.5**o
        ref[3,0,1,0]=(-0.539343546695)/(o+1.)+(-0.353357134289)+(0.0487640958479)*0.5**o
        ref[3,0,1,1]=(-0.537581471096)/(o+1.)+(-0.230612659836)+(0.951844805407)*0.5**o
        ref[3,0,2,0]=(-0.853447043865)/(o+1.)+(0.74984630544)+(0.495490307938)*0.5**o
        ref[3,0,2,1]=(-0.67563531103)/(o+1.)+(-0.187556978169)+(0.38188303469)*0.5**o
        ref[3,1,0,0]=(-0.356730659314)/(o+1.)+(-0.0526231031206)+(0.44471228922)*0.5**o
        ref[3,1,0,1]=(-0.869871613483)/(o+1.)+(-0.0667259695618)+(-0.674677862576)*0.5**o
        ref[3,1,1,0]=(-0.410056071233)/(o+1.)+(0.301413392891)+(-0.803923033456)*0.5**o
        ref[3,1,1,1]=(-0.848629733526)/(o+1.)+(0.529126589519)+(-0.677586574312)*0.5**o
        ref[3,1,2,0]=(0.168694427833)/(o+1.)+(-0.00113235248793)+(0.723956969878)*0.5**o
        ref[3,1,2,1]=(0.0979731186867)/(o+1.)+(0.154243452232)+(-0.0961257584415)*0.5**o
        ref[3,2,0,0]=(-0.765686972231)/(o+1.)+(0.625608272028)+(-0.506857925404)*0.5**o
        ref[3,2,0,1]=(-0.549033177495)/(o+1.)+(0.0301416615418)+(-0.479464705052)*0.5**o
        ref[3,2,1,0]=(-0.352937841183)/(o+1.)+(0.216473827282)+(0.984579775952)*0.5**o
        ref[3,2,1,1]=(0.683120790039)/(o+1.)+(0.696004989039)+(-0.75230207503)*0.5**o
        ref[3,2,2,0]=(-0.350363036315)/(o+1.)+(-0.33489294201)+(-0.257935272452)*0.5**o
        ref[3,2,2,1]=(0.758201028024)/(o+1.)+(0.281278388038)+(0.0784702766009)*0.5**o
        ref[3,3,0,0]=(-0.536413055541)/(o+1.)+(-0.639541290655)+(0.292597181224)*0.5**o
        ref[3,3,0,1]=(0.89040777319)/(o+1.)+(0.852650873286)+(0.174134715952)*0.5**o
        ref[3,3,1,0]=(0.598889528719)/(o+1.)+(0.257976575291)+(0.768234215847)*0.5**o
        ref[3,3,1,1]=(0.471242869055)/(o+1.)+(-0.168028702229)+(-0.77938361663)*0.5**o
        ref[3,3,2,0]=(0.674100983537)/(o+1.)+(0.11346432072)+(-0.00617753872016)*0.5**o
        ref[3,3,2,1]=(-0.297791856337)/(o+1.)+(-0.0465060902539)+(0.124470359441)*0.5**o
        ref[3,4,0,0]=(-0.77267392403)/(o+1.)+(0.392480774552)+(0.139920132608)*0.5**o
        ref[3,4,0,1]=(0.730856047416)/(o+1.)+(-0.574224140787)+(0.328592839797)*0.5**o
        ref[3,4,1,0]=(-0.393781550386)/(o+1.)+(0.371586901534)+(-0.239362896774)*0.5**o
        ref[3,4,1,1]=(0.17782134987)/(o+1.)+(0.382203062909)+(0.777985150139)*0.5**o
        ref[3,4,2,0]=(-0.77882422986)/(o+1.)+(0.457217815872)+(0.348482781008)*0.5**o
        ref[3,4,2,1]=(0.308553967581)/(o+1.)+(-0.164313519526)+(-0.396751338134)*0.5**o
      else:
        arg[0,0,0,0]=(0.663998508168)*x[0]**o+(0.797340164931)*x[0]+(0.285182001599)*x[1]**o+(0.99720606542)*x[1]+(0.127610260108)*x[2]**o+(0.0141819362674)*x[2]
        arg[0,0,0,1]=(-0.010130464476)*x[0]**o+(-0.694140348803)*x[0]+(-0.554528445238)*x[1]**o+(0.192639775817)*x[1]+(-0.520164891934)*x[2]**o+(0.316830081989)*x[2]
        arg[0,0,1,0]=(-0.176542289385)*x[0]**o+(-0.312451025451)*x[0]+(0.33951668081)*x[1]**o+(-0.66674551066)*x[1]+(-0.0515929546862)*x[2]**o+(0.688180995863)*x[2]
        arg[0,0,1,1]=(-0.98152572229)*x[0]**o+(0.156411605196)*x[0]+(0.588065240125)*x[1]**o+(0.665830732172)*x[1]+(-0.834625744106)*x[2]**o+(-0.138276534671)*x[2]
        arg[0,0,2,0]=(0.555897422173)*x[0]**o+(0.0269431977874)*x[0]+(0.671161746561)*x[1]**o+(0.0857445285374)*x[1]+(0.578027911522)*x[2]**o+(0.217805007393)*x[2]
        arg[0,0,2,1]=(0.23127040923)*x[0]**o+(0.624353504617)*x[0]+(-0.433915231286)*x[1]**o+(-0.860173971996)*x[1]+(-0.801472434932)*x[2]**o+(-0.779575445432)*x[2]
        arg[0,1,0,0]=(0.0202445177842)*x[0]**o+(-0.0654289514039)*x[0]+(-0.877581892208)*x[1]**o+(-0.231833216639)*x[1]+(0.609752401432)*x[2]**o+(-0.802673983603)*x[2]
        arg[0,1,0,1]=(0.99935098551)*x[0]**o+(0.665639246142)*x[0]+(0.593510187831)*x[1]**o+(0.213336090684)*x[1]+(0.232667772738)*x[2]**o+(-0.651386262633)*x[2]
        arg[0,1,1,0]=(0.151799117419)*x[0]**o+(-0.34631043288)*x[0]+(-0.264898731671)*x[1]**o+(0.697527693137)*x[1]+(0.389030412719)*x[2]**o+(0.613338381577)*x[2]
        arg[0,1,1,1]=(-0.51422532626)*x[0]**o+(0.228040429268)*x[0]+(-0.276760984751)*x[1]**o+(0.379484010198)*x[1]+(0.896837565192)*x[2]**o+(0.195866552401)*x[2]
        arg[0,1,2,0]=(0.908244888839)*x[0]**o+(0.217098236162)*x[0]+(0.798725721675)*x[1]**o+(0.216635630826)*x[1]+(0.960227216319)*x[2]**o+(-0.885381182879)*x[2]
        arg[0,1,2,1]=(-0.872307766634)*x[0]**o+(0.843197737136)*x[0]+(-0.895220198791)*x[1]**o+(0.866433237518)*x[1]+(0.560344834226)*x[2]**o+(0.895797915384)*x[2]
        arg[0,2,0,0]=(0.00173619592035)*x[0]**o+(-0.209651858019)*x[0]+(0.323249367733)*x[1]**o+(-0.848445788048)*x[1]+(-0.747443572813)*x[2]**o+(0.342929929066)*x[2]
        arg[0,2,0,1]=(0.998184336552)*x[0]**o+(-0.669882578313)*x[0]+(-0.632794054696)*x[1]**o+(-0.779832061924)*x[1]+(0.0588815237456)*x[2]**o+(0.419538634163)*x[2]
        arg[0,2,1,0]=(-0.921234331968)*x[0]**o+(0.319724478943)*x[0]+(0.867192523265)*x[1]**o+(-0.183047755109)*x[1]+(-0.753047661507)*x[2]**o+(-0.797179040607)*x[2]
        arg[0,2,1,1]=(0.153578015941)*x[0]**o+(0.887820807812)*x[0]+(-0.607944647299)*x[1]**o+(-0.0487413965527)*x[1]+(-0.812369692315)*x[2]**o+(0.323102876844)*x[2]
        arg[0,2,2,0]=(0.29601674794)*x[0]**o+(-0.108738759628)*x[0]+(0.242837345895)*x[1]**o+(-0.324487554664)*x[1]+(0.584399369463)*x[2]**o+(0.683877139789)*x[2]
        arg[0,2,2,1]=(0.0580007647099)*x[0]**o+(0.214488016057)*x[0]+(0.669087265419)*x[1]**o+(-0.311999860513)*x[1]+(0.678940598195)*x[2]**o+(0.658847203441)*x[2]
        arg[0,3,0,0]=(-0.337999890505)*x[0]**o+(-0.842479530409)*x[0]+(0.786376708623)*x[1]**o+(0.597298234359)*x[1]+(0.495442668997)*x[2]**o+(-0.183828960943)*x[2]
        arg[0,3,0,1]=(-0.825813329629)*x[0]**o+(0.197040670392)*x[0]+(0.997784261354)*x[1]**o+(-0.355290279278)*x[1]+(0.657909516664)*x[2]**o+(-0.568281670001)*x[2]
        arg[0,3,1,0]=(0.699313512079)*x[0]**o+(-0.916350147519)*x[0]+(-0.311605808179)*x[1]**o+(0.222618888987)*x[1]+(0.872150798976)*x[2]**o+(0.139773150125)*x[2]
        arg[0,3,1,1]=(0.140764124286)*x[0]**o+(0.994013777309)*x[0]+(-0.0792107151558)*x[1]**o+(-0.285357856751)*x[1]+(-0.800285107223)*x[2]**o+(0.413029089978)*x[2]
        arg[0,3,2,0]=(-0.68177752688)*x[0]**o+(0.0339522871939)*x[0]+(-0.641399841911)*x[1]**o+(0.263279621719)*x[1]+(-0.406115766497)*x[2]**o+(-0.774639092638)*x[2]
        arg[0,3,2,1]=(0.655685968811)*x[0]**o+(0.20085918281)*x[0]+(-0.753491228829)*x[1]**o+(0.393669658887)*x[1]+(0.339469131428)*x[2]**o+(0.424803724709)*x[2]
        arg[0,4,0,0]=(-0.18027010068)*x[0]**o+(-0.773007683353)*x[0]+(0.767133020895)*x[1]**o+(-0.810449138606)*x[1]+(0.295421955604)*x[2]**o+(-0.955926880008)*x[2]
        arg[0,4,0,1]=(-0.914768182433)*x[0]**o+(-0.67024319498)*x[0]+(0.533244666277)*x[1]**o+(-0.705719628704)*x[1]+(-0.963633923897)*x[2]**o+(-0.0818895784023)*x[2]
        arg[0,4,1,0]=(0.274727423835)*x[0]**o+(0.838218676826)*x[0]+(0.269989900726)*x[1]**o+(0.650600179293)*x[1]+(-0.855210211935)*x[2]**o+(-0.800261683904)*x[2]
        arg[0,4,1,1]=(0.257574026756)*x[0]**o+(-0.787643878696)*x[0]+(-0.619136916008)*x[1]**o+(-0.242580269843)*x[1]+(0.0229959873555)*x[2]**o+(-0.685725009107)*x[2]
        arg[0,4,2,0]=(0.703185665611)*x[0]**o+(0.37686548395)*x[0]+(-0.414901304236)*x[1]**o+(0.310335568741)*x[1]+(-0.740764649572)*x[2]**o+(-0.389356987349)*x[2]
        arg[0,4,2,1]=(-0.153792811673)*x[0]**o+(-0.179777301012)*x[0]+(0.694402537357)*x[1]**o+(-0.928918851291)*x[1]+(-0.439307255129)*x[2]**o+(0.148671231214)*x[2]
        arg[1,0,0,0]=(0.298044833507)*x[0]**o+(0.538000859279)*x[0]+(-0.405432041063)*x[1]**o+(-0.8912548734)*x[1]+(0.929554154255)*x[2]**o+(-0.515905504512)*x[2]
        arg[1,0,0,1]=(-0.575485929806)*x[0]**o+(0.757951009672)*x[0]+(-0.815024122535)*x[1]**o+(0.846519632117)*x[1]+(0.594901198403)*x[2]**o+(-0.672034020351)*x[2]
        arg[1,0,1,0]=(0.998358005707)*x[0]**o+(-0.144729504804)*x[0]+(0.463445062889)*x[1]**o+(0.428918148475)*x[1]+(-0.172684103233)*x[2]**o+(-0.394742387769)*x[2]
        arg[1,0,1,1]=(0.783593367143)*x[0]**o+(-0.269882783638)*x[0]+(-0.880336283806)*x[1]**o+(0.337087191943)*x[1]+(-0.819512478233)*x[2]**o+(0.582924902797)*x[2]
        arg[1,0,2,0]=(-0.42345692756)*x[0]**o+(-0.251910598711)*x[0]+(-0.882193473121)*x[1]**o+(0.535405977525)*x[1]+(-0.117109433242)*x[2]**o+(0.99565775159)*x[2]
        arg[1,0,2,1]=(-0.790917746681)*x[0]**o+(-0.291814540513)*x[0]+(-0.0828604085821)*x[1]**o+(-0.0715382319561)*x[1]+(-0.188453557262)*x[2]**o+(0.297596604241)*x[2]
        arg[1,1,0,0]=(0.944343622558)*x[0]**o+(0.537893843235)*x[0]+(-0.370697218599)*x[1]**o+(-0.043709074747)*x[1]+(-0.803318845377)*x[2]**o+(0.814392946479)*x[2]
        arg[1,1,0,1]=(-0.246711023019)*x[0]**o+(-0.847902406307)*x[0]+(0.062135604814)*x[1]**o+(0.581341495124)*x[1]+(0.561484848132)*x[2]**o+(0.06820017559)*x[2]
        arg[1,1,1,0]=(-0.530782063061)*x[0]**o+(-0.951246029994)*x[0]+(-0.809647094351)*x[1]**o+(0.198165508669)*x[1]+(0.591475451485)*x[2]**o+(-0.0567469904903)*x[2]
        arg[1,1,1,1]=(-0.710844761085)*x[0]**o+(0.742146961382)*x[0]+(0.482950602277)*x[1]**o+(-0.701576649521)*x[1]+(0.752305355358)*x[2]**o+(0.658478074845)*x[2]
        arg[1,1,2,0]=(0.574657505707)*x[0]**o+(-0.268094098242)*x[0]+(-0.79745549318)*x[1]**o+(0.994732385826)*x[1]+(0.36752110393)*x[2]**o+(0.0733521344978)*x[2]
        arg[1,1,2,1]=(-0.365373199615)*x[0]**o+(0.295532428151)*x[0]+(-0.916049089517)*x[1]**o+(0.562876895658)*x[1]+(0.698818392781)*x[2]**o+(0.413166153315)*x[2]
        arg[1,2,0,0]=(0.544776615353)*x[0]**o+(-0.647889580869)*x[0]+(-0.0569263298125)*x[1]**o+(0.145083519829)*x[1]+(-0.0456368749942)*x[2]**o+(0.125915481669)*x[2]
        arg[1,2,0,1]=(0.11684721915)*x[0]**o+(-0.793592672443)*x[0]+(-0.428585283805)*x[1]**o+(0.280358391066)*x[1]+(0.72731741847)*x[2]**o+(-0.320459136831)*x[2]
        arg[1,2,1,0]=(0.561736948232)*x[0]**o+(0.833411538285)*x[0]+(0.361519222525)*x[1]**o+(-0.785948355356)*x[1]+(-0.595938528895)*x[2]**o+(0.720247127573)*x[2]
        arg[1,2,1,1]=(0.578630452142)*x[0]**o+(0.218238639275)*x[0]+(0.792496779641)*x[1]**o+(-0.818431209813)*x[1]+(-0.252522807953)*x[2]**o+(0.648505893606)*x[2]
        arg[1,2,2,0]=(-0.0269203575755)*x[0]**o+(-0.586454813205)*x[0]+(-0.239885637578)*x[1]**o+(0.186263805313)*x[1]+(0.585909451795)*x[2]**o+(-0.630731036615)*x[2]
        arg[1,2,2,1]=(-0.534642633404)*x[0]**o+(0.921034930282)*x[0]+(-0.326672365556)*x[1]**o+(0.690509146367)*x[1]+(-0.802567575176)*x[2]**o+(-0.235903818924)*x[2]
        arg[1,3,0,0]=(0.892302802156)*x[0]**o+(-0.528836075071)*x[0]+(-0.806520835766)*x[1]**o+(0.771321197577)*x[1]+(-0.545545197249)*x[2]**o+(-0.633185919719)*x[2]
        arg[1,3,0,1]=(0.883702447985)*x[0]**o+(0.698176356191)*x[0]+(-0.348189466787)*x[1]**o+(-0.390058291315)*x[1]+(0.805736008481)*x[2]**o+(0.861787745203)*x[2]
        arg[1,3,1,0]=(0.492527250981)*x[0]**o+(0.635712008949)*x[0]+(-0.464397346193)*x[1]**o+(-0.820952575379)*x[1]+(0.235015001164)*x[2]**o+(-0.846160845231)*x[2]
        arg[1,3,1,1]=(0.372499842901)*x[0]**o+(-0.524824414377)*x[0]+(0.553450359486)*x[1]**o+(-0.187372498756)*x[1]+(-0.389983195918)*x[2]**o+(-0.310421627605)*x[2]
        arg[1,3,2,0]=(0.533098362924)*x[0]**o+(0.917820078197)*x[0]+(-0.557683070597)*x[1]**o+(-0.388498248383)*x[1]+(-0.551866333066)*x[2]**o+(0.696873468511)*x[2]
        arg[1,3,2,1]=(-0.909038123273)*x[0]**o+(0.19323884412)*x[0]+(-0.00134861813725)*x[1]**o+(-0.858193032021)*x[1]+(0.0436856504635)*x[2]**o+(-0.646994265986)*x[2]
        arg[1,4,0,0]=(-0.974474210185)*x[0]**o+(-0.39786371112)*x[0]+(-0.829024534519)*x[1]**o+(0.836151642189)*x[1]+(0.679654655991)*x[2]**o+(0.750995377779)*x[2]
        arg[1,4,0,1]=(0.964745405431)*x[0]**o+(0.269189443297)*x[0]+(0.531636623506)*x[1]**o+(-0.522687387383)*x[1]+(0.463500385567)*x[2]**o+(0.160556224359)*x[2]
        arg[1,4,1,0]=(-0.714725894011)*x[0]**o+(0.0478807291181)*x[0]+(0.591494552904)*x[1]**o+(-0.240293119817)*x[1]+(-0.0463360360003)*x[2]**o+(-0.932835908096)*x[2]
        arg[1,4,1,1]=(-0.809072545597)*x[0]**o+(-0.963335650229)*x[0]+(-0.781885522838)*x[1]**o+(0.822395174213)*x[1]+(-0.42590899181)*x[2]**o+(-0.237357227214)*x[2]
        arg[1,4,2,0]=(-0.21281876112)*x[0]**o+(-0.131640536998)*x[0]+(-0.648680371617)*x[1]**o+(0.555759774031)*x[1]+(-0.911030244394)*x[2]**o+(0.962859403252)*x[2]
        arg[1,4,2,1]=(-0.247930702516)*x[0]**o+(0.509829884687)*x[0]+(0.0387816417333)*x[1]**o+(-0.605945140743)*x[1]+(0.187880415018)*x[2]**o+(-0.813242364046)*x[2]
        arg[2,0,0,0]=(0.616352869937)*x[0]**o+(-0.930708209966)*x[0]+(-0.874778661809)*x[1]**o+(-0.330225154904)*x[1]+(0.632020336392)*x[2]**o+(0.205631598228)*x[2]
        arg[2,0,0,1]=(-0.85704374519)*x[0]**o+(0.99958496465)*x[0]+(0.807095767473)*x[1]**o+(0.63097388144)*x[1]+(0.000665614692828)*x[2]**o+(0.418070858864)*x[2]
        arg[2,0,1,0]=(0.326870102902)*x[0]**o+(0.565855791221)*x[0]+(-0.39418644908)*x[1]**o+(-0.96081735968)*x[1]+(0.937255293725)*x[2]**o+(0.533015538451)*x[2]
        arg[2,0,1,1]=(-0.699939842176)*x[0]**o+(0.808055402204)*x[0]+(0.448948790025)*x[1]**o+(-0.720122241439)*x[1]+(0.0532200858319)*x[2]**o+(-0.204851171261)*x[2]
        arg[2,0,2,0]=(-0.985996827188)*x[0]**o+(-0.604602591167)*x[0]+(0.445903925662)*x[1]**o+(-0.163135114049)*x[1]+(-0.264281316158)*x[2]**o+(-0.381814212863)*x[2]
        arg[2,0,2,1]=(-0.195357773436)*x[0]**o+(0.73285122682)*x[0]+(-0.914846292757)*x[1]**o+(-0.190524885095)*x[1]+(-0.103440307801)*x[2]**o+(-0.0309647945217)*x[2]
        arg[2,1,0,0]=(0.601652939329)*x[0]**o+(-0.757171386065)*x[0]+(-0.91147090556)*x[1]**o+(-0.665258414941)*x[1]+(-0.0504245123191)*x[2]**o+(-0.274928427711)*x[2]
        arg[2,1,0,1]=(0.534038022604)*x[0]**o+(-0.977365399298)*x[0]+(0.884656682614)*x[1]**o+(0.0866421048629)*x[1]+(-0.70687555879)*x[2]**o+(-0.99928465036)*x[2]
        arg[2,1,1,0]=(-0.361196978257)*x[0]**o+(0.192559020092)*x[0]+(-0.0767715201942)*x[1]**o+(0.222254953069)*x[1]+(-0.128362555436)*x[2]**o+(-0.667485829139)*x[2]
        arg[2,1,1,1]=(0.00930987921229)*x[0]**o+(0.269207633273)*x[0]+(0.943798095916)*x[1]**o+(0.483123564952)*x[1]+(-0.231669365037)*x[2]**o+(-0.877005245489)*x[2]
        arg[2,1,2,0]=(-0.181479246679)*x[0]**o+(0.619551810132)*x[0]+(0.499400298722)*x[1]**o+(-0.307864349855)*x[1]+(0.19303040809)*x[2]**o+(0.523640816928)*x[2]
        arg[2,1,2,1]=(-0.99665050175)*x[0]**o+(-0.647775674514)*x[0]+(-0.0015481673821)*x[1]**o+(-0.749198451256)*x[1]+(0.652528193579)*x[2]**o+(-0.0194127861341)*x[2]
        arg[2,2,0,0]=(0.870069083779)*x[0]**o+(0.179747121592)*x[0]+(0.387523112512)*x[1]**o+(0.625711925024)*x[1]+(0.840141938096)*x[2]**o+(0.629709976556)*x[2]
        arg[2,2,0,1]=(-0.841259995642)*x[0]**o+(0.220858862445)*x[0]+(0.397334296484)*x[1]**o+(-0.036654179636)*x[1]+(0.235704322138)*x[2]**o+(-0.6299228864)*x[2]
        arg[2,2,1,0]=(-0.561184923307)*x[0]**o+(-0.643644162376)*x[0]+(0.728784994685)*x[1]**o+(-0.697418070844)*x[1]+(-0.56166463788)*x[2]**o+(-0.668760084534)*x[2]
        arg[2,2,1,1]=(0.92243000401)*x[0]**o+(0.96165833182)*x[0]+(0.540379002453)*x[1]**o+(0.630245078647)*x[1]+(-0.375873656541)*x[2]**o+(0.71424898756)*x[2]
        arg[2,2,2,0]=(0.331131296272)*x[0]**o+(-0.136167378345)*x[0]+(-0.375407637025)*x[1]**o+(-0.0890691004072)*x[1]+(-0.787306129385)*x[2]**o+(-0.447474749986)*x[2]
        arg[2,2,2,1]=(-0.47892171854)*x[0]**o+(0.341448171328)*x[0]+(0.662403724393)*x[1]**o+(0.581128263913)*x[1]+(0.00987835902932)*x[2]**o+(-0.884847129103)*x[2]
        arg[2,3,0,0]=(0.0219406022052)*x[0]**o+(-0.370001982912)*x[0]+(0.948162807268)*x[1]**o+(-0.870857624404)*x[1]+(0.580053532448)*x[2]**o+(-0.488657040041)*x[2]
        arg[2,3,0,1]=(-0.114799511617)*x[0]**o+(0.547853919356)*x[0]+(-0.224367707174)*x[1]**o+(0.551491902831)*x[1]+(-0.101845907889)*x[2]**o+(-0.793527438395)*x[2]
        arg[2,3,1,0]=(-0.103518223905)*x[0]**o+(-0.319840011093)*x[0]+(-0.856612861717)*x[1]**o+(0.0208578084364)*x[1]+(0.132248308449)*x[2]**o+(-0.580663554499)*x[2]
        arg[2,3,1,1]=(-0.958142224242)*x[0]**o+(-0.401299403423)*x[0]+(-0.289034019185)*x[1]**o+(-0.0913958572584)*x[1]+(-0.818515945817)*x[2]**o+(-0.918366425737)*x[2]
        arg[2,3,2,0]=(0.730554594734)*x[0]**o+(0.209964841318)*x[0]+(-0.4157212696)*x[1]**o+(-0.673375403124)*x[1]+(-0.409899056244)*x[2]**o+(0.124418003637)*x[2]
        arg[2,3,2,1]=(-0.930974589897)*x[0]**o+(0.256452400333)*x[0]+(0.778606258675)*x[1]**o+(0.23173182074)*x[1]+(-0.846439682798)*x[2]**o+(-0.408860675089)*x[2]
        arg[2,4,0,0]=(-0.246046780287)*x[0]**o+(-0.653766430943)*x[0]+(0.725379461155)*x[1]**o+(-0.305490823497)*x[1]+(0.0837637468427)*x[2]**o+(-0.458770631133)*x[2]
        arg[2,4,0,1]=(-0.612751147889)*x[0]**o+(0.504418327803)*x[0]+(-0.835051783805)*x[1]**o+(-0.86410559166)*x[1]+(0.928919612861)*x[2]**o+(0.303164594248)*x[2]
        arg[2,4,1,0]=(-0.26092972064)*x[0]**o+(-0.990592356395)*x[0]+(-0.48716443965)*x[1]**o+(0.659447608304)*x[1]+(0.429865593656)*x[2]**o+(-0.661470437053)*x[2]
        arg[2,4,1,1]=(0.325840618456)*x[0]**o+(-0.261974904995)*x[0]+(0.476982730267)*x[1]**o+(-0.738119182885)*x[1]+(0.903618858571)*x[2]**o+(-0.287442022624)*x[2]
        arg[2,4,2,0]=(0.655590135983)*x[0]**o+(-0.135505327297)*x[0]+(-0.38079131346)*x[1]**o+(0.663850751769)*x[1]+(-0.658568109867)*x[2]**o+(-0.267875066063)*x[2]
        arg[2,4,2,1]=(0.570712162768)*x[0]**o+(0.390833274898)*x[0]+(-0.97326103687)*x[1]**o+(-0.419300688467)*x[1]+(0.55695479018)*x[2]**o+(0.547271407555)*x[2]
        arg[3,0,0,0]=(-0.486023722572)*x[0]**o+(-0.790544923201)*x[0]+(0.101237181961)*x[1]**o+(0.23267779263)*x[1]+(-0.458580701132)*x[2]**o+(-0.479449733139)*x[2]
        arg[3,0,0,1]=(-0.705291625776)*x[0]**o+(0.144166702405)*x[0]+(-0.210596631253)*x[1]**o+(0.843721258425)*x[1]+(-0.716420627484)*x[2]**o+(0.145427688875)*x[2]
        arg[3,0,1,0]=(-0.838974120992)*x[0]**o+(0.824423674933)*x[0]+(-0.211643366873)*x[1]**o+(0.603255736693)*x[1]+(-0.119484817371)*x[2]**o+(-0.357272543813)*x[2]
        arg[3,0,1,1]=(0.140027541071)*x[0]**o+(0.373598174368)*x[0]+(-0.661831753147)*x[1]**o+(-0.973724961653)*x[1]+(-0.0738899100301)*x[2]**o+(0.0128579666074)*x[2]
        arg[3,0,2,0]=(-0.39083471939)*x[0]**o+(-0.00821070494132)*x[0]+(-0.315936839906)*x[1]**o+(-0.967556350778)*x[1]+(0.136991762476)*x[2]**o+(0.374041520875)*x[2]
        arg[3,0,2,1]=(0.0734743009776)*x[0]**o+(-0.103186084491)*x[0]+(0.426933353985)*x[1]**o+(-0.631917740007)*x[1]+(0.755296770294)*x[2]**o+(0.492171872216)*x[2]
        arg[3,1,0,0]=(-0.414502599055)*x[0]**o+(0.512031410722)*x[0]+(0.610677049483)*x[1]**o+(-0.075275839509)*x[1]+(0.853407100545)*x[2]**o+(-0.569731924126)*x[2]
        arg[3,1,0,1]=(0.886115095396)*x[0]**o+(0.811807969433)*x[0]+(0.550623354147)*x[1]**o+(-0.703728894875)*x[1]+(0.784318903356)*x[2]**o+(0.45891829089)*x[2]
        arg[3,1,1,0]=(0.312577129879)*x[0]**o+(0.0127110992422)*x[0]+(-0.174614340667)*x[1]**o+(0.744535847679)*x[1]+(0.991125452344)*x[2]**o+(-0.405595301834)*x[2]
        arg[3,1,1,1]=(0.784369258288)*x[0]**o+(-0.709503832235)*x[0]+(0.032963558933)*x[1]**o+(-0.471036543959)*x[1]+(-0.138022967476)*x[2]**o+(-0.38452297541)*x[2]
        arg[3,1,2,0]=(-0.337332295082)*x[0]**o+(-0.460457867002)*x[0]+(-0.102806204559)*x[1]**o+(-0.502889108364)*x[1]+(-0.641488011668)*x[2]**o+(0.359501036527)*x[2]
        arg[3,1,2,1]=(0.791575892417)*x[0]**o+(0.126853712619)*x[0]+(-0.408911942637)*x[1]**o+(0.557039825987)*x[1]+(-0.260719710621)*x[2]**o+(-0.648326395935)*x[2]
        arg[3,2,0,0]=(-0.0909007497318)*x[0]**o+(-0.337085519884)*x[0]+(-0.651987922715)*x[1]**o+(-0.130225558042)*x[1]+(0.926567188664)*x[2]**o+(0.764369275478)*x[2]
        arg[3,2,0,1]=(-0.764137934896)*x[0]**o+(0.443892891832)*x[0]+(0.688197384487)*x[1]**o+(0.485210058592)*x[1]+(0.267932741224)*x[2]**o+(0.0366064885582)*x[2]
        arg[3,2,1,0]=(0.49435832562)*x[0]**o+(0.860649583676)*x[0]+(0.667415091305)*x[1]**o+(0.144069385904)*x[1]+(-0.933847080974)*x[2]**o+(-0.378123304101)*x[2]
        arg[3,2,1,1]=(-0.460752644946)*x[0]**o+(0.491079753273)*x[0]+(-0.536055633506)*x[1]**o+(0.786378615881)*x[1]+(-0.190647453518)*x[2]**o+(-0.192868019034)*x[2]
        arg[3,2,2,0]=(0.0946687985697)*x[0]**o+(0.978197346579)*x[0]+(0.972747948856)*x[1]**o+(-0.575670550665)*x[1]+(-0.096229657916)*x[2]**o+(-0.772567430449)*x[2]
        arg[3,2,2,1]=(0.177521713734)*x[0]**o+(0.0683685252148)*x[0]+(0.423995112682)*x[1]**o+(-0.489247981968)*x[1]+(-0.537130434099)*x[2]**o+(-0.0228879131449)*x[2]
        arg[3,3,0,0]=(0.592201472175)*x[0]**o+(0.674304504132)*x[0]+(0.428414872879)*x[1]**o+(0.956414209094)*x[1]+(-0.272853242491)*x[2]**o+(-0.639858910901)*x[2]
        arg[3,3,0,1]=(0.572972211447)*x[0]**o+(0.943096858489)*x[0]+(-0.990948797302)*x[1]**o+(-0.475383526007)*x[1]+(0.542076985363)*x[2]**o+(-0.884809421862)*x[2]
        arg[3,3,1,0]=(0.775163727164)*x[0]**o+(0.101808074039)*x[0]+(0.575490330857)*x[1]**o+(0.199751068512)*x[1]+(0.538742155685)*x[2]**o+(-0.725822956214)*x[2]
        arg[3,3,1,1]=(-0.244586606882)*x[0]**o+(0.983712254807)*x[0]+(0.187370538052)*x[1]**o+(0.332611385073)*x[1]+(0.536377667373)*x[2]**o+(0.255272558489)*x[2]
        arg[3,3,2,0]=(-0.676040282048)*x[0]**o+(-0.172059284021)*x[0]+(0.25100451419)*x[1]**o+(0.303226061816)*x[1]+(0.138976177556)*x[2]**o+(-0.330104849626)*x[2]
        arg[3,3,2,1]=(-0.136437682299)*x[0]**o+(0.938786085213)*x[0]+(0.853018944962)*x[1]**o+(-0.628078903025)*x[1]+(-0.903567849919)*x[2]**o+(0.837621994591)*x[2]
        arg[3,4,0,0]=(0.28632587806)*x[0]**o+(-0.480626263522)*x[0]+(-0.0537965100699)*x[1]**o+(0.423252432505)*x[1]+(0.200777153039)*x[2]**o+(0.629713098071)*x[2]
        arg[3,4,0,1]=(0.326448964148)*x[0]**o+(0.809991245673)*x[0]+(-0.527362539073)*x[1]**o+(-0.0727890632651)*x[1]+(0.210806278508)*x[2]**o+(0.736535063786)*x[2]
        arg[3,4,1,0]=(0.760474271531)*x[0]**o+(0.378083396435)*x[0]+(-0.991296179693)*x[1]**o+(0.431965391269)*x[1]+(-0.12597704605)*x[2]**o+(-0.128017966575)*x[2]
        arg[3,4,1,1]=(-0.21508452109)*x[0]**o+(-0.583147117492)*x[0]+(0.779535489322)*x[1]**o+(0.601144982341)*x[1]+(-0.870362610456)*x[2]**o+(-0.318535391517)*x[2]
        arg[3,4,2,0]=(-0.195181632154)*x[0]**o+(0.237570596161)*x[0]+(-0.794725125363)*x[1]**o+(0.862030507166)*x[1]+(-0.612206533195)*x[2]**o+(0.494078406766)*x[2]
        arg[3,4,2,1]=(0.270278273932)*x[0]**o+(-0.115833233113)*x[0]+(0.754464364285)*x[1]**o+(-0.186069521903)*x[1]+(-0.864426716229)*x[2]**o+(-0.425054438064)*x[2]
        ref[0,0,0,0]=(0.412792261707)/(o+1.)+(0.904364083309)+(0.663998508168)*0.5**o
        ref[0,0,0,1]=(-1.07469333717)/(o+1.)+(-0.0923352454984)+(-0.010130464476)*0.5**o
        ref[0,0,1,0]=(0.287923726124)/(o+1.)+(-0.145507770124)+(-0.176542289385)*0.5**o
        ref[0,0,1,1]=(-0.24656050398)/(o+1.)+(0.341982901349)+(-0.98152572229)*0.5**o
        ref[0,0,2,0]=(1.24918965808)/(o+1.)+(0.165246366859)+(0.555897422173)*0.5**o
        ref[0,0,2,1]=(-1.23538766622)/(o+1.)+(-0.507697956406)+(0.23127040923)*0.5**o
        ref[0,1,0,0]=(-0.267829490776)/(o+1.)+(-0.549968075823)+(0.0202445177842)*0.5**o
        ref[0,1,0,1]=(0.826177960569)/(o+1.)+(0.113794537096)+(0.99935098551)*0.5**o
        ref[0,1,1,0]=(0.124131681048)/(o+1.)+(0.482277820918)+(0.151799117419)*0.5**o
        ref[0,1,1,1]=(0.620076580442)/(o+1.)+(0.401695495934)+(-0.51422532626)*0.5**o
        ref[0,1,2,0]=(1.75895293799)/(o+1.)+(-0.225823657945)+(0.908244888839)*0.5**o
        ref[0,1,2,1]=(-0.334875364565)/(o+1.)+(1.30271444502)+(-0.872307766634)*0.5**o
        ref[0,2,0,0]=(-0.42419420508)/(o+1.)+(-0.3575838585)+(0.00173619592035)*0.5**o
        ref[0,2,0,1]=(-0.57391253095)/(o+1.)+(-0.515088003037)+(0.998184336552)*0.5**o
        ref[0,2,1,0]=(0.114144861758)/(o+1.)+(-0.330251158386)+(-0.921234331968)*0.5**o
        ref[0,2,1,1]=(-1.42031433961)/(o+1.)+(0.581091144051)+(0.153578015941)*0.5**o
        ref[0,2,2,0]=(0.827236715358)/(o+1.)+(0.125325412749)+(0.29601674794)*0.5**o
        ref[0,2,2,1]=(1.34802786361)/(o+1.)+(0.280667679493)+(0.0580007647099)*0.5**o
        ref[0,3,0,0]=(1.28181937762)/(o+1.)+(-0.214505128496)+(-0.337999890505)*0.5**o
        ref[0,3,0,1]=(1.65569377802)/(o+1.)+(-0.363265639443)+(-0.825813329629)*0.5**o
        ref[0,3,1,0]=(0.560544990797)/(o+1.)+(-0.276979054204)+(0.699313512079)*0.5**o
        ref[0,3,1,1]=(-0.879495822379)/(o+1.)+(0.560842505268)+(0.140764124286)*0.5**o
        ref[0,3,2,0]=(-1.04751560841)/(o+1.)+(-0.238703591863)+(-0.68177752688)*0.5**o
        ref[0,3,2,1]=(-0.414022097401)/(o+1.)+(0.509666283203)+(0.655685968811)*0.5**o
        ref[0,4,0,0]=(1.0625549765)/(o+1.)+(-1.26969185098)+(-0.18027010068)*0.5**o
        ref[0,4,0,1]=(-0.430389257621)/(o+1.)+(-0.728926201043)+(-0.914768182433)*0.5**o
        ref[0,4,1,0]=(-0.585220311209)/(o+1.)+(0.344278586108)+(0.274727423835)*0.5**o
        ref[0,4,1,1]=(-0.596140928652)/(o+1.)+(-0.857974578823)+(0.257574026756)*0.5**o
        ref[0,4,2,0]=(-1.15566595381)/(o+1.)+(0.148922032671)+(0.703185665611)*0.5**o
        ref[0,4,2,1]=(0.255095282229)/(o+1.)+(-0.480012460544)+(-0.153792811673)*0.5**o
        ref[1,0,0,0]=(0.524122113192)/(o+1.)+(-0.434579759316)+(0.298044833507)*0.5**o
        ref[1,0,0,1]=(-0.220122924131)/(o+1.)+(0.466218310719)+(-0.575485929806)*0.5**o
        ref[1,0,1,0]=(0.290760959656)/(o+1.)+(-0.0552768720487)+(0.998358005707)*0.5**o
        ref[1,0,1,1]=(-1.69984876204)/(o+1.)+(0.325064655551)+(0.783593367143)*0.5**o
        ref[1,0,2,0]=(-0.999302906363)/(o+1.)+(0.639576565202)+(-0.42345692756)*0.5**o
        ref[1,0,2,1]=(-0.271313965844)/(o+1.)+(-0.0328780841139)+(-0.790917746681)*0.5**o
        ref[1,1,0,0]=(-1.17401606398)/(o+1.)+(0.654288857484)+(0.944343622558)*0.5**o
        ref[1,1,0,1]=(0.623620452946)/(o+1.)+(-0.0991803677964)+(-0.246711023019)*0.5**o
        ref[1,1,1,0]=(-0.218171642866)/(o+1.)+(-0.404913755908)+(-0.530782063061)*0.5**o
        ref[1,1,1,1]=(1.23525595763)/(o+1.)+(0.349524193353)+(-0.710844761085)*0.5**o
        ref[1,1,2,0]=(-0.42993438925)/(o+1.)+(0.399995211041)+(0.574657505707)*0.5**o
        ref[1,1,2,1]=(-0.217230696736)/(o+1.)+(0.635787738562)+(-0.365373199615)*0.5**o
        ref[1,2,0,0]=(-0.102563204807)/(o+1.)+(-0.188445289686)+(0.544776615353)*0.5**o
        ref[1,2,0,1]=(0.298732134665)/(o+1.)+(-0.416846709104)+(0.11684721915)*0.5**o
        ref[1,2,1,0]=(-0.234419306371)/(o+1.)+(0.383855155251)+(0.561736948232)*0.5**o
        ref[1,2,1,1]=(0.539973971687)/(o+1.)+(0.0241566615339)+(0.578630452142)*0.5**o
        ref[1,2,2,0]=(0.346023814217)/(o+1.)+(-0.515461022254)+(-0.0269203575755)*0.5**o
        ref[1,2,2,1]=(-1.12923994073)/(o+1.)+(0.687820128862)+(-0.534642633404)*0.5**o
        ref[1,3,0,0]=(-1.35206603301)/(o+1.)+(-0.195350398606)+(0.892302802156)*0.5**o
        ref[1,3,0,1]=(0.457546541694)/(o+1.)+(0.584952905039)+(0.883702447985)*0.5**o
        ref[1,3,1,0]=(-0.229382345029)/(o+1.)+(-0.51570070583)+(0.492527250981)*0.5**o
        ref[1,3,1,1]=(0.163467163568)/(o+1.)+(-0.511309270369)+(0.372499842901)*0.5**o
        ref[1,3,2,0]=(-1.10954940366)/(o+1.)+(0.613097649163)+(0.533098362924)*0.5**o
        ref[1,3,2,1]=(0.0423370323263)/(o+1.)+(-0.655974226943)+(-0.909038123273)*0.5**o
        ref[1,4,0,0]=(-0.149369878527)/(o+1.)+(0.594641654424)+(-0.974474210185)*0.5**o
        ref[1,4,0,1]=(0.995137009073)/(o+1.)+(-0.0464708598635)+(0.964745405431)*0.5**o
        ref[1,4,1,0]=(0.545158516903)/(o+1.)+(-0.562624149397)+(-0.714725894011)*0.5**o
        ref[1,4,1,1]=(-1.20779451465)/(o+1.)+(-0.189148851615)+(-0.809072545597)*0.5**o
        ref[1,4,2,0]=(-1.55971061601)/(o+1.)+(0.693489320143)+(-0.21281876112)*0.5**o
        ref[1,4,2,1]=(0.226662056752)/(o+1.)+(-0.454678810051)+(-0.247930702516)*0.5**o
        ref[2,0,0,0]=(-0.242758325417)/(o+1.)+(-0.527650883321)+(0.616352869937)*0.5**o
        ref[2,0,0,1]=(0.807761382166)/(o+1.)+(1.02431485248)+(-0.85704374519)*0.5**o
        ref[2,0,1,0]=(0.543068844645)/(o+1.)+(0.0690269849966)+(0.326870102902)*0.5**o
        ref[2,0,1,1]=(0.502168875857)/(o+1.)+(-0.0584590052478)+(-0.699939842176)*0.5**o
        ref[2,0,2,0]=(0.181622609504)/(o+1.)+(-0.57477595904)+(-0.985996827188)*0.5**o
        ref[2,0,2,1]=(-1.01828660056)/(o+1.)+(0.255680773601)+(-0.195357773436)*0.5**o
        ref[2,1,0,0]=(-0.961895417879)/(o+1.)+(-0.848679114358)+(0.601652939329)*0.5**o
        ref[2,1,0,1]=(0.177781123824)/(o+1.)+(-0.945003972397)+(0.534038022604)*0.5**o
        ref[2,1,1,0]=(-0.20513407563)/(o+1.)+(-0.126335927989)+(-0.361196978257)*0.5**o
        ref[2,1,1,1]=(0.712128730879)/(o+1.)+(-0.0623370236322)+(0.00930987921229)*0.5**o
        ref[2,1,2,0]=(0.692430706812)/(o+1.)+(0.417664138602)+(-0.181479246679)*0.5**o
        ref[2,1,2,1]=(0.650980026197)/(o+1.)+(-0.708193455952)+(-0.99665050175)*0.5**o
        ref[2,2,0,0]=(1.22766505061)/(o+1.)+(0.717584511586)+(0.870069083779)*0.5**o
        ref[2,2,0,1]=(0.633038618622)/(o+1.)+(-0.222859101796)+(-0.841259995642)*0.5**o
        ref[2,2,1,0]=(0.167120356805)/(o+1.)+(-1.00491115888)+(-0.561184923307)*0.5**o
        ref[2,2,1,1]=(0.164505345911)/(o+1.)+(1.15307619901)+(0.92243000401)*0.5**o
        ref[2,2,2,0]=(-1.16271376641)/(o+1.)+(-0.336355614369)+(0.331131296272)*0.5**o
        ref[2,2,2,1]=(0.672282083422)/(o+1.)+(0.018864653069)+(-0.47892171854)*0.5**o
        ref[2,3,0,0]=(1.52821633972)/(o+1.)+(-0.864758323678)+(0.0219406022052)*0.5**o
        ref[2,3,0,1]=(-0.326213615062)/(o+1.)+(0.152909191896)+(-0.114799511617)*0.5**o
        ref[2,3,1,0]=(-0.724364553267)/(o+1.)+(-0.439822878578)+(-0.103518223905)*0.5**o
        ref[2,3,1,1]=(-1.107549965)/(o+1.)+(-0.705530843209)+(-0.958142224242)*0.5**o
        ref[2,3,2,0]=(-0.825620325844)/(o+1.)+(-0.169496279085)+(0.730554594734)*0.5**o
        ref[2,3,2,1]=(-0.0678334241233)/(o+1.)+(0.0396617729923)+(-0.930974589897)*0.5**o
        ref[2,4,0,0]=(0.809143207998)/(o+1.)+(-0.709013942786)+(-0.246046780287)*0.5**o
        ref[2,4,0,1]=(0.0938678290563)/(o+1.)+(-0.0282613348041)+(-0.612751147889)*0.5**o
        ref[2,4,1,0]=(-0.0572988459938)/(o+1.)+(-0.496307592572)+(-0.26092972064)*0.5**o
        ref[2,4,1,1]=(1.38060158884)/(o+1.)+(-0.643768055252)+(0.325840618456)*0.5**o
        ref[2,4,2,0]=(-1.03935942333)/(o+1.)+(0.130235179204)+(0.655590135983)*0.5**o
        ref[2,4,2,1]=(-0.41630624669)/(o+1.)+(0.259401996992)+(0.570712162768)*0.5**o
        ref[3,0,0,0]=(-0.357343519171)/(o+1.)+(-0.518658431855)+(-0.486023722572)*0.5**o
        ref[3,0,0,1]=(-0.927017258737)/(o+1.)+(0.566657824853)+(-0.705291625776)*0.5**o
        ref[3,0,1,0]=(-0.331128184244)/(o+1.)+(0.535203433907)+(-0.838974120992)*0.5**o
        ref[3,0,1,1]=(-0.735721663177)/(o+1.)+(-0.293634410339)+(0.140027541071)*0.5**o
        ref[3,0,2,0]=(-0.178945077431)/(o+1.)+(-0.300862767422)+(-0.39083471939)*0.5**o
        ref[3,0,2,1]=(1.18223012428)/(o+1.)+(-0.121465976141)+(0.0734743009776)*0.5**o
        ref[3,1,0,0]=(1.46408415003)/(o+1.)+(-0.0664881764563)+(-0.414502599055)*0.5**o
        ref[3,1,0,1]=(1.3349422575)/(o+1.)+(0.283498682724)+(0.886115095396)*0.5**o
        ref[3,1,1,0]=(0.816511111677)/(o+1.)+(0.175825822544)+(0.312577129879)*0.5**o
        ref[3,1,1,1]=(-0.105059408543)/(o+1.)+(-0.782531675802)+(0.784369258288)*0.5**o
        ref[3,1,2,0]=(-0.744294216228)/(o+1.)+(-0.301922969419)+(-0.337332295082)*0.5**o
        ref[3,1,2,1]=(-0.669631653258)/(o+1.)+(0.0177835713351)+(0.791575892417)*0.5**o
        ref[3,2,0,0]=(0.274579265949)/(o+1.)+(0.148529098776)+(-0.0909007497318)*0.5**o
        ref[3,2,0,1]=(0.956130125711)/(o+1.)+(0.482854719491)+(-0.764137934896)*0.5**o
        ref[3,2,1,0]=(-0.266431989669)/(o+1.)+(0.313297832739)+(0.49435832562)*0.5**o
        ref[3,2,1,1]=(-0.726703087024)/(o+1.)+(0.542295175061)+(-0.460752644946)*0.5**o
        ref[3,2,2,0]=(0.87651829094)/(o+1.)+(-0.185020317268)+(0.0946687985697)*0.5**o
        ref[3,2,2,1]=(-0.113135321417)/(o+1.)+(-0.221883684949)+(0.177521713734)*0.5**o
        ref[3,3,0,0]=(0.155561630388)/(o+1.)+(0.495429901163)+(0.592201472175)*0.5**o
        ref[3,3,0,1]=(-0.44887181194)/(o+1.)+(-0.20854804469)+(0.572972211447)*0.5**o
        ref[3,3,1,0]=(1.11423248654)/(o+1.)+(-0.212131906831)+(0.775163727164)*0.5**o
        ref[3,3,1,1]=(0.723748205425)/(o+1.)+(0.785798099184)+(-0.244586606882)*0.5**o
        ref[3,3,2,0]=(0.389980691746)/(o+1.)+(-0.0994690359155)+(-0.676040282048)*0.5**o
        ref[3,3,2,1]=(-0.0505489049572)/(o+1.)+(0.57416458839)+(-0.136437682299)*0.5**o
        ref[3,4,0,0]=(0.14698064297)/(o+1.)+(0.286169633527)+(0.28632587806)*0.5**o
        ref[3,4,0,1]=(-0.316556260565)/(o+1.)+(0.736868623097)+(0.326448964148)*0.5**o
        ref[3,4,1,0]=(-1.11727322574)/(o+1.)+(0.341015410564)+(0.760474271531)*0.5**o
        ref[3,4,1,1]=(-0.0908271211344)/(o+1.)+(-0.150268763334)+(-0.21508452109)*0.5**o
        ref[3,4,2,0]=(-1.40693165856)/(o+1.)+(0.796839755046)+(-0.195181632154)*0.5**o
        ref[3,4,2,1]=(-0.109962351944)/(o+1.)+(-0.36347859654)+(0.270278273932)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.0433006726989)*x[0]**o+(0.819109331838)*x[0]+(0.150292245587)*x[1]**o+(0.949397165008)*x[1]
        ref=(0.150292245587)/(o+1.)+(0.884253248423)+(0.0433006726989)*0.5**o
      else:
        arg=(-0.765858486678)*x[0]**o+(-0.829061249978)*x[0]+(-0.0255273602735)*x[1]**o+(-0.93097352763)*x[1]+(-0.775974817954)*x[2]**o+(0.0414580763278)*x[2]
        ref=(-0.801502178228)/(o+1.)+(-0.85928835064)+(-0.765858486678)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.708418284738)*x[0]**o+(-0.265416249865)*x[0]+(0.0412873279676)*x[1]**o+(0.18827690539)*x[1]
        arg[1]=(-0.0639578348425)*x[0]**o+(-0.435408788697)*x[0]+(0.429134354603)*x[1]**o+(0.0391558318157)*x[1]
        ref[0]=(0.0412873279676)/(o+1.)+(-0.0385696722374)+(0.708418284738)*0.5**o
        ref[1]=(0.429134354603)/(o+1.)+(-0.198126478441)+(-0.0639578348425)*0.5**o
      else:
        arg[0]=(-0.516066856372)*x[0]**o+(-0.509677909948)*x[0]+(0.189594948337)*x[1]**o+(0.743432553688)*x[1]+(0.687864109046)*x[2]**o+(0.257280277379)*x[2]
        arg[1]=(0.948762785699)*x[0]**o+(0.372861763607)*x[0]+(-0.166460043702)*x[1]**o+(0.0474018901531)*x[1]+(-0.485380626595)*x[2]**o+(-0.341701937267)*x[2]
        ref[0]=(0.877459057383)/(o+1.)+(0.245517460559)+(-0.516066856372)*0.5**o
        ref[1]=(-0.651840670298)/(o+1.)+(0.0392808582465)+(0.948762785699)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.0540280486222)*x[0]**o+(-0.936629447756)*x[0]+(0.530867101627)*x[1]**o+(-0.404522355122)*x[1]
        arg[0,1]=(0.585051034599)*x[0]**o+(-0.899683342239)*x[0]+(0.61865830472)*x[1]**o+(0.491768251743)*x[1]
        arg[0,2]=(0.924131379905)*x[0]**o+(-0.394367734246)*x[0]+(0.432122525065)*x[1]**o+(0.79923634141)*x[1]
        arg[0,3]=(-0.41435196904)*x[0]**o+(-0.22039395628)*x[0]+(-0.866551833459)*x[1]**o+(-0.0966534238435)*x[1]
        arg[0,4]=(0.855785701017)*x[0]**o+(0.0879754354535)*x[0]+(-0.0727782016239)*x[1]**o+(0.746878225435)*x[1]
        arg[1,0]=(-0.868450306659)*x[0]**o+(-0.00845792225901)*x[0]+(0.227399876353)*x[1]**o+(-0.522196483799)*x[1]
        arg[1,1]=(0.53076931297)*x[0]**o+(0.165744140226)*x[0]+(0.14586398181)*x[1]**o+(-0.694169551043)*x[1]
        arg[1,2]=(-0.936146144862)*x[0]**o+(-0.198367580291)*x[0]+(0.848822682163)*x[1]**o+(0.360521711263)*x[1]
        arg[1,3]=(-0.395889451325)*x[0]**o+(0.510713345487)*x[0]+(-0.62440033386)*x[1]**o+(0.29695714683)*x[1]
        arg[1,4]=(0.847685553395)*x[0]**o+(-0.291266469606)*x[0]+(-0.517114839541)*x[1]**o+(-0.279948585383)*x[1]
        arg[2,0]=(-0.622481093435)*x[0]**o+(0.0911922519622)*x[0]+(0.216648336495)*x[1]**o+(0.663562516802)*x[1]
        arg[2,1]=(0.567833843918)*x[0]**o+(0.18116887359)*x[0]+(-0.0433012798352)*x[1]**o+(0.998332142285)*x[1]
        arg[2,2]=(-0.81729786652)*x[0]**o+(-0.634680104287)*x[0]+(0.991513806139)*x[1]**o+(-0.0721700196003)*x[1]
        arg[2,3]=(0.121181286827)*x[0]**o+(-0.677803464822)*x[0]+(0.870139639102)*x[1]**o+(-0.134662002755)*x[1]
        arg[2,4]=(-0.167376807346)*x[0]**o+(-0.500556007898)*x[0]+(-0.839256017197)*x[1]**o+(-0.82147890878)*x[1]
        arg[3,0]=(0.576521064132)*x[0]**o+(-0.510276372588)*x[0]+(0.92888227487)*x[1]**o+(-0.88293512036)*x[1]
        arg[3,1]=(-0.159048471862)*x[0]**o+(-0.0479500210678)*x[0]+(-0.0216438779517)*x[1]**o+(-0.108252189548)*x[1]
        arg[3,2]=(-0.483779576334)*x[0]**o+(-0.871829147209)*x[0]+(-0.22556218283)*x[1]**o+(0.451007116911)*x[1]
        arg[3,3]=(-0.358888369902)*x[0]**o+(0.436554685191)*x[0]+(-0.0359187908139)*x[1]**o+(-0.0148930474244)*x[1]
        arg[3,4]=(-0.995411478882)*x[0]**o+(-0.457206328785)*x[0]+(-0.833823947273)*x[1]**o+(-0.0517068976093)*x[1]
        ref[0,0]=(0.530867101627)/(o+1.)+(-0.670575901439)+(-0.0540280486222)*0.5**o
        ref[0,1]=(0.61865830472)/(o+1.)+(-0.203957545248)+(0.585051034599)*0.5**o
        ref[0,2]=(0.432122525065)/(o+1.)+(0.202434303582)+(0.924131379905)*0.5**o
        ref[0,3]=(-0.866551833459)/(o+1.)+(-0.158523690062)+(-0.41435196904)*0.5**o
        ref[0,4]=(-0.0727782016239)/(o+1.)+(0.417426830444)+(0.855785701017)*0.5**o
        ref[1,0]=(0.227399876353)/(o+1.)+(-0.265327203029)+(-0.868450306659)*0.5**o
        ref[1,1]=(0.14586398181)/(o+1.)+(-0.264212705408)+(0.53076931297)*0.5**o
        ref[1,2]=(0.848822682163)/(o+1.)+(0.0810770654864)+(-0.936146144862)*0.5**o
        ref[1,3]=(-0.62440033386)/(o+1.)+(0.403835246159)+(-0.395889451325)*0.5**o
        ref[1,4]=(-0.517114839541)/(o+1.)+(-0.285607527495)+(0.847685553395)*0.5**o
        ref[2,0]=(0.216648336495)/(o+1.)+(0.377377384382)+(-0.622481093435)*0.5**o
        ref[2,1]=(-0.0433012798352)/(o+1.)+(0.589750507938)+(0.567833843918)*0.5**o
        ref[2,2]=(0.991513806139)/(o+1.)+(-0.353425061944)+(-0.81729786652)*0.5**o
        ref[2,3]=(0.870139639102)/(o+1.)+(-0.406232733789)+(0.121181286827)*0.5**o
        ref[2,4]=(-0.839256017197)/(o+1.)+(-0.661017458339)+(-0.167376807346)*0.5**o
        ref[3,0]=(0.92888227487)/(o+1.)+(-0.696605746474)+(0.576521064132)*0.5**o
        ref[3,1]=(-0.0216438779517)/(o+1.)+(-0.078101105308)+(-0.159048471862)*0.5**o
        ref[3,2]=(-0.22556218283)/(o+1.)+(-0.210411015149)+(-0.483779576334)*0.5**o
        ref[3,3]=(-0.0359187908139)/(o+1.)+(0.210830818883)+(-0.358888369902)*0.5**o
        ref[3,4]=(-0.833823947273)/(o+1.)+(-0.254456613197)+(-0.995411478882)*0.5**o
      else:
        arg[0,0]=(-0.937858877179)*x[0]**o+(-0.41590845461)*x[0]+(0.317385444442)*x[1]**o+(-0.166284606371)*x[1]+(-0.846106994172)*x[2]**o+(0.227876218054)*x[2]
        arg[0,1]=(0.497785867084)*x[0]**o+(0.883557362705)*x[0]+(-0.846602726567)*x[1]**o+(-0.859091910548)*x[1]+(-0.295742503292)*x[2]**o+(-0.853931233266)*x[2]
        arg[0,2]=(0.744501879094)*x[0]**o+(-0.0181115125382)*x[0]+(-0.856099357899)*x[1]**o+(0.767449074734)*x[1]+(-0.858552187616)*x[2]**o+(-0.580142057156)*x[2]
        arg[0,3]=(0.210031467359)*x[0]**o+(-0.994426053833)*x[0]+(0.648522478175)*x[1]**o+(0.0355805603417)*x[1]+(-0.274347447112)*x[2]**o+(-0.280165537012)*x[2]
        arg[0,4]=(-0.323650660168)*x[0]**o+(-0.279592941595)*x[0]+(-0.478811425713)*x[1]**o+(0.19753271141)*x[1]+(0.476202875551)*x[2]**o+(0.657358100659)*x[2]
        arg[1,0]=(0.890659868437)*x[0]**o+(0.100235405989)*x[0]+(0.325177871418)*x[1]**o+(0.675379114656)*x[1]+(0.458473828724)*x[2]**o+(-0.449822467434)*x[2]
        arg[1,1]=(-0.467895039242)*x[0]**o+(-0.551300000422)*x[0]+(0.976111922801)*x[1]**o+(0.809389941274)*x[1]+(-0.693066235685)*x[2]**o+(0.595719115505)*x[2]
        arg[1,2]=(-0.727304617172)*x[0]**o+(-0.270682330445)*x[0]+(-0.32926621635)*x[1]**o+(0.0389068111165)*x[1]+(0.300096425532)*x[2]**o+(-0.964910767632)*x[2]
        arg[1,3]=(0.195012106644)*x[0]**o+(0.18137839607)*x[0]+(-0.823129279257)*x[1]**o+(-0.0440106889304)*x[1]+(-0.298898659898)*x[2]**o+(0.801623925974)*x[2]
        arg[1,4]=(0.590200326369)*x[0]**o+(-0.800085209049)*x[0]+(-0.323836133834)*x[1]**o+(-0.953413527643)*x[1]+(0.0257128844735)*x[2]**o+(0.652469516082)*x[2]
        arg[2,0]=(-0.467537281458)*x[0]**o+(-0.05673369912)*x[0]+(-0.813542619659)*x[1]**o+(-0.778596374878)*x[1]+(0.752073310476)*x[2]**o+(0.187824163254)*x[2]
        arg[2,1]=(-0.75708112983)*x[0]**o+(0.866111788737)*x[0]+(0.715424327432)*x[1]**o+(-0.297967739435)*x[1]+(0.420241611178)*x[2]**o+(-0.616867390634)*x[2]
        arg[2,2]=(0.00628628803839)*x[0]**o+(0.700962301988)*x[0]+(0.712979263919)*x[1]**o+(-0.414401132518)*x[1]+(-0.975498928756)*x[2]**o+(0.858116391039)*x[2]
        arg[2,3]=(0.948164399923)*x[0]**o+(-0.88227825142)*x[0]+(0.30703295864)*x[1]**o+(0.510607674984)*x[1]+(0.362361719392)*x[2]**o+(-0.208696380629)*x[2]
        arg[2,4]=(0.661545494115)*x[0]**o+(-0.122960355184)*x[0]+(-0.801367604574)*x[1]**o+(0.395708459987)*x[1]+(0.0197225340457)*x[2]**o+(-0.377005073637)*x[2]
        arg[3,0]=(0.733597350529)*x[0]**o+(0.339318978843)*x[0]+(0.317608635437)*x[1]**o+(0.147746051469)*x[1]+(-0.33413935488)*x[2]**o+(-0.793337674933)*x[2]
        arg[3,1]=(0.0918904281646)*x[0]**o+(-0.530786152204)*x[0]+(-0.587542750401)*x[1]**o+(-0.810331029621)*x[1]+(-0.40221065987)*x[2]**o+(-0.250472092093)*x[2]
        arg[3,2]=(0.966442952042)*x[0]**o+(-0.756649866552)*x[0]+(-0.36466223222)*x[1]**o+(-0.48639005055)*x[1]+(-0.831374462742)*x[2]**o+(0.503451143644)*x[2]
        arg[3,3]=(-0.322275644664)*x[0]**o+(-0.254067149982)*x[0]+(-0.440264794128)*x[1]**o+(0.226546045529)*x[1]+(-0.761870916789)*x[2]**o+(0.70024066379)*x[2]
        arg[3,4]=(0.595459342613)*x[0]**o+(-0.439436200258)*x[0]+(0.0932538180176)*x[1]**o+(-0.975450806707)*x[1]+(0.0461423323623)*x[2]**o+(0.0294352472154)*x[2]
        ref[0,0]=(-0.52872154973)/(o+1.)+(-0.177158421464)+(-0.937858877179)*0.5**o
        ref[0,1]=(-1.14234522986)/(o+1.)+(-0.414732890554)+(0.497785867084)*0.5**o
        ref[0,2]=(-1.71465154551)/(o+1.)+(0.0845977525197)+(0.744501879094)*0.5**o
        ref[0,3]=(0.374175031062)/(o+1.)+(-0.619505515252)+(0.210031467359)*0.5**o
        ref[0,4]=(-0.00260855016132)/(o+1.)+(0.287648935237)+(-0.323650660168)*0.5**o
        ref[1,0]=(0.783651700142)/(o+1.)+(0.162896026605)+(0.890659868437)*0.5**o
        ref[1,1]=(0.283045687117)/(o+1.)+(0.426904528178)+(-0.467895039242)*0.5**o
        ref[1,2]=(-0.0291697908188)/(o+1.)+(-0.59834314348)+(-0.727304617172)*0.5**o
        ref[1,3]=(-1.12202793916)/(o+1.)+(0.469495816557)+(0.195012106644)*0.5**o
        ref[1,4]=(-0.29812324936)/(o+1.)+(-0.550514610305)+(0.590200326369)*0.5**o
        ref[2,0]=(-0.0614693091826)/(o+1.)+(-0.323752955372)+(-0.467537281458)*0.5**o
        ref[2,1]=(1.13566593861)/(o+1.)+(-0.024361670666)+(-0.75708112983)*0.5**o
        ref[2,2]=(-0.262519664837)/(o+1.)+(0.572338780255)+(0.00628628803839)*0.5**o
        ref[2,3]=(0.669394678032)/(o+1.)+(-0.290183478532)+(0.948164399923)*0.5**o
        ref[2,4]=(-0.781645070528)/(o+1.)+(-0.0521284844165)+(0.661545494115)*0.5**o
        ref[3,0]=(-0.0165307194439)/(o+1.)+(-0.15313632231)+(0.733597350529)*0.5**o
        ref[3,1]=(-0.989753410271)/(o+1.)+(-0.795794636959)+(0.0918904281646)*0.5**o
        ref[3,2]=(-1.19603669496)/(o+1.)+(-0.369794386729)+(0.966442952042)*0.5**o
        ref[3,3]=(-1.20213571092)/(o+1.)+(0.336359779668)+(-0.322275644664)*0.5**o
        ref[3,4]=(0.13939615038)/(o+1.)+(-0.692725879875)+(0.595459342613)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.12237588236)*x[0]**o+(-0.671341785528)*x[0]+(0.998910161099)*x[1]**o+(-0.552290246966)*x[1]
        arg[0,0,1]=(-0.90821257221)*x[0]**o+(-0.413391927764)*x[0]+(-0.428904099491)*x[1]**o+(-0.226766222189)*x[1]
        arg[0,1,0]=(-0.333819313127)*x[0]**o+(-0.458569899029)*x[0]+(-0.83776177332)*x[1]**o+(0.22951184631)*x[1]
        arg[0,1,1]=(-0.0571879315489)*x[0]**o+(-0.38056227895)*x[0]+(0.0784628125977)*x[1]**o+(-0.659353613638)*x[1]
        arg[1,0,0]=(0.645871468353)*x[0]**o+(-0.986005542006)*x[0]+(0.395704509276)*x[1]**o+(0.980666918726)*x[1]
        arg[1,0,1]=(0.337293036597)*x[0]**o+(-0.520809490378)*x[0]+(0.543657498632)*x[1]**o+(0.632777244118)*x[1]
        arg[1,1,0]=(-0.432349538121)*x[0]**o+(-0.579685247099)*x[0]+(0.331943072726)*x[1]**o+(0.25042188644)*x[1]
        arg[1,1,1]=(0.329861800715)*x[0]**o+(-0.48725090895)*x[0]+(-0.149357998414)*x[1]**o+(-0.580650134153)*x[1]
        arg[2,0,0]=(-0.75424658357)*x[0]**o+(-0.696352517414)*x[0]+(-0.00272313166917)*x[1]**o+(-0.488581343546)*x[1]
        arg[2,0,1]=(-0.136489901339)*x[0]**o+(-0.752587775699)*x[0]+(0.986413392791)*x[1]**o+(0.759344728162)*x[1]
        arg[2,1,0]=(-0.0832674877963)*x[0]**o+(0.765419402778)*x[0]+(0.958533611564)*x[1]**o+(-0.315461136294)*x[1]
        arg[2,1,1]=(0.362764663642)*x[0]**o+(-0.351519282456)*x[0]+(-0.115554433923)*x[1]**o+(-0.789365036398)*x[1]
        arg[3,0,0]=(0.0977752933224)*x[0]**o+(-0.838881001576)*x[0]+(0.275578737811)*x[1]**o+(-0.971208155806)*x[1]
        arg[3,0,1]=(0.0588243034654)*x[0]**o+(-0.141302234145)*x[0]+(-0.460044396206)*x[1]**o+(-0.776793718354)*x[1]
        arg[3,1,0]=(0.786922930047)*x[0]**o+(-0.187797924361)*x[0]+(0.971516347812)*x[1]**o+(-0.250547377602)*x[1]
        arg[3,1,1]=(-0.69240453335)*x[0]**o+(0.439425578473)*x[0]+(0.00864940781276)*x[1]**o+(-0.458898387231)*x[1]
        arg[4,0,0]=(0.0700104829681)*x[0]**o+(0.905031121846)*x[0]+(0.37676750477)*x[1]**o+(0.411190149331)*x[1]
        arg[4,0,1]=(-0.615737714754)*x[0]**o+(0.1945204765)*x[0]+(0.0889014902938)*x[1]**o+(0.650020022597)*x[1]
        arg[4,1,0]=(-0.883448519911)*x[0]**o+(0.00124188371434)*x[0]+(-0.501662684156)*x[1]**o+(0.290235036398)*x[1]
        arg[4,1,1]=(-0.723885968554)*x[0]**o+(-0.0112308623755)*x[0]+(0.679499427805)*x[1]**o+(-0.00261690404287)*x[1]
        arg[5,0,0]=(0.0834731365594)*x[0]**o+(-0.873857991293)*x[0]+(0.831158842487)*x[1]**o+(-0.585808323324)*x[1]
        arg[5,0,1]=(0.250830078938)*x[0]**o+(0.310347394372)*x[0]+(-0.412138455827)*x[1]**o+(0.974823718029)*x[1]
        arg[5,1,0]=(-0.865155065381)*x[0]**o+(0.72190541012)*x[0]+(0.72020941875)*x[1]**o+(-0.604997858133)*x[1]
        arg[5,1,1]=(0.57252234131)*x[0]**o+(-0.366230226771)*x[0]+(0.115400105646)*x[1]**o+(0.0485229323141)*x[1]
        ref[0,0,0]=(0.998910161099)/(o+1.)+(-0.611816016247)+(-0.12237588236)*0.5**o
        ref[0,0,1]=(-0.428904099491)/(o+1.)+(-0.320079074977)+(-0.90821257221)*0.5**o
        ref[0,1,0]=(-0.83776177332)/(o+1.)+(-0.11452902636)+(-0.333819313127)*0.5**o
        ref[0,1,1]=(0.0784628125977)/(o+1.)+(-0.519957946294)+(-0.0571879315489)*0.5**o
        ref[1,0,0]=(0.395704509276)/(o+1.)+(-0.00266931163989)+(0.645871468353)*0.5**o
        ref[1,0,1]=(0.543657498632)/(o+1.)+(0.0559838768697)+(0.337293036597)*0.5**o
        ref[1,1,0]=(0.331943072726)/(o+1.)+(-0.16463168033)+(-0.432349538121)*0.5**o
        ref[1,1,1]=(-0.149357998414)/(o+1.)+(-0.533950521552)+(0.329861800715)*0.5**o
        ref[2,0,0]=(-0.00272313166917)/(o+1.)+(-0.59246693048)+(-0.75424658357)*0.5**o
        ref[2,0,1]=(0.986413392791)/(o+1.)+(0.00337847623164)+(-0.136489901339)*0.5**o
        ref[2,1,0]=(0.958533611564)/(o+1.)+(0.224979133242)+(-0.0832674877963)*0.5**o
        ref[2,1,1]=(-0.115554433923)/(o+1.)+(-0.570442159427)+(0.362764663642)*0.5**o
        ref[3,0,0]=(0.275578737811)/(o+1.)+(-0.905044578691)+(0.0977752933224)*0.5**o
        ref[3,0,1]=(-0.460044396206)/(o+1.)+(-0.45904797625)+(0.0588243034654)*0.5**o
        ref[3,1,0]=(0.971516347812)/(o+1.)+(-0.219172650981)+(0.786922930047)*0.5**o
        ref[3,1,1]=(0.00864940781276)/(o+1.)+(-0.00973640437899)+(-0.69240453335)*0.5**o
        ref[4,0,0]=(0.37676750477)/(o+1.)+(0.658110635589)+(0.0700104829681)*0.5**o
        ref[4,0,1]=(0.0889014902938)/(o+1.)+(0.422270249549)+(-0.615737714754)*0.5**o
        ref[4,1,0]=(-0.501662684156)/(o+1.)+(0.145738460056)+(-0.883448519911)*0.5**o
        ref[4,1,1]=(0.679499427805)/(o+1.)+(-0.00692388320921)+(-0.723885968554)*0.5**o
        ref[5,0,0]=(0.831158842487)/(o+1.)+(-0.729833157309)+(0.0834731365594)*0.5**o
        ref[5,0,1]=(-0.412138455827)/(o+1.)+(0.642585556201)+(0.250830078938)*0.5**o
        ref[5,1,0]=(0.72020941875)/(o+1.)+(0.0584537759935)+(-0.865155065381)*0.5**o
        ref[5,1,1]=(0.115400105646)/(o+1.)+(-0.158853647228)+(0.57252234131)*0.5**o
      else:
        arg[0,0,0]=(0.962792187023)*x[0]**o+(0.799861778652)*x[0]+(-0.811367630134)*x[1]**o+(-0.553496861095)*x[1]+(-0.810089307714)*x[2]**o+(0.135449765364)*x[2]
        arg[0,0,1]=(0.644573004654)*x[0]**o+(-0.565823623798)*x[0]+(0.337026645127)*x[1]**o+(0.413612857799)*x[1]+(0.604366347987)*x[2]**o+(0.0867373815054)*x[2]
        arg[0,1,0]=(0.829914553754)*x[0]**o+(-0.462419132852)*x[0]+(0.889464473044)*x[1]**o+(-0.425594549042)*x[1]+(0.679251542389)*x[2]**o+(-0.501090535615)*x[2]
        arg[0,1,1]=(0.874166303558)*x[0]**o+(0.437595762221)*x[0]+(-0.836229998654)*x[1]**o+(-0.704020633222)*x[1]+(-0.279878325513)*x[2]**o+(0.697289463565)*x[2]
        arg[1,0,0]=(-0.963425479949)*x[0]**o+(0.399090631945)*x[0]+(-0.80371987894)*x[1]**o+(0.614102223184)*x[1]+(-0.821741166107)*x[2]**o+(-0.245209306654)*x[2]
        arg[1,0,1]=(-0.288791346674)*x[0]**o+(0.0125896799209)*x[0]+(0.0536290891541)*x[1]**o+(-0.127968165901)*x[1]+(-0.635591402779)*x[2]**o+(-0.723850350215)*x[2]
        arg[1,1,0]=(-0.554196270608)*x[0]**o+(-0.409293351346)*x[0]+(-0.759821124187)*x[1]**o+(-0.828913354317)*x[1]+(0.66900958614)*x[2]**o+(0.412403847239)*x[2]
        arg[1,1,1]=(-0.862735330269)*x[0]**o+(-0.406644646366)*x[0]+(-0.375399872184)*x[1]**o+(-0.10762837777)*x[1]+(0.540886844249)*x[2]**o+(0.97487281528)*x[2]
        arg[2,0,0]=(0.388899900771)*x[0]**o+(-0.692092360652)*x[0]+(0.794415524913)*x[1]**o+(0.369239170999)*x[1]+(0.482269641926)*x[2]**o+(0.58717167121)*x[2]
        arg[2,0,1]=(0.324795219208)*x[0]**o+(0.508631698719)*x[0]+(0.128456415737)*x[1]**o+(0.73678163559)*x[1]+(0.0527520660005)*x[2]**o+(-0.180289299372)*x[2]
        arg[2,1,0]=(-0.901972287405)*x[0]**o+(-0.635701120264)*x[0]+(0.298343193246)*x[1]**o+(-0.870902858159)*x[1]+(-0.979619200422)*x[2]**o+(0.708567755329)*x[2]
        arg[2,1,1]=(0.391663724321)*x[0]**o+(0.69891810611)*x[0]+(0.717069638885)*x[1]**o+(0.493487853359)*x[1]+(0.270536570073)*x[2]**o+(0.679115502666)*x[2]
        arg[3,0,0]=(0.830590943609)*x[0]**o+(-0.000320739222143)*x[0]+(-0.659146599349)*x[1]**o+(-0.858428982679)*x[1]+(-0.158070494634)*x[2]**o+(-0.364394913197)*x[2]
        arg[3,0,1]=(0.903615175916)*x[0]**o+(0.186562406505)*x[0]+(0.463860110418)*x[1]**o+(0.0382550628307)*x[1]+(0.835148180357)*x[2]**o+(-0.117637969912)*x[2]
        arg[3,1,0]=(-0.901874408714)*x[0]**o+(0.00493918499294)*x[0]+(0.180179475909)*x[1]**o+(-0.549428619271)*x[1]+(-0.392536708051)*x[2]**o+(0.68054148024)*x[2]
        arg[3,1,1]=(0.807693382033)*x[0]**o+(-0.47532617097)*x[0]+(-0.52499678021)*x[1]**o+(-0.0365395713918)*x[1]+(0.654881606829)*x[2]**o+(-0.63350736651)*x[2]
        arg[4,0,0]=(0.440830144989)*x[0]**o+(-0.454694325046)*x[0]+(0.403509667018)*x[1]**o+(0.360080283937)*x[1]+(0.164208775222)*x[2]**o+(0.203945755083)*x[2]
        arg[4,0,1]=(-0.836804887763)*x[0]**o+(-0.910979763109)*x[0]+(-0.629411293881)*x[1]**o+(0.0546020113997)*x[1]+(0.63435754491)*x[2]**o+(-0.323224122475)*x[2]
        arg[4,1,0]=(0.204216251019)*x[0]**o+(-0.372349067194)*x[0]+(-0.320768886225)*x[1]**o+(-0.139715856072)*x[1]+(0.62329130659)*x[2]**o+(0.626784461235)*x[2]
        arg[4,1,1]=(0.275074685897)*x[0]**o+(0.0664691742991)*x[0]+(0.949175504221)*x[1]**o+(-0.438858093561)*x[1]+(0.330489174462)*x[2]**o+(0.361586647545)*x[2]
        arg[5,0,0]=(0.370194889673)*x[0]**o+(-0.0549428893157)*x[0]+(0.0219932528442)*x[1]**o+(-0.699553487775)*x[1]+(0.0161212665659)*x[2]**o+(-0.268426682266)*x[2]
        arg[5,0,1]=(-0.527801267225)*x[0]**o+(0.151774276637)*x[0]+(-0.0623666836223)*x[1]**o+(-0.953572390482)*x[1]+(0.778912307756)*x[2]**o+(-0.851760269136)*x[2]
        arg[5,1,0]=(0.11418535486)*x[0]**o+(-0.299719975234)*x[0]+(-0.508303365848)*x[1]**o+(-0.856367302992)*x[1]+(0.311609370032)*x[2]**o+(0.4731881155)*x[2]
        arg[5,1,1]=(0.455947680647)*x[0]**o+(0.856811965126)*x[0]+(-0.176215162097)*x[1]**o+(0.581325693162)*x[1]+(-0.306514780798)*x[2]**o+(-0.005241787834)*x[2]
        ref[0,0,0]=(-1.62145693785)/(o+1.)+(0.19090734146)+(0.962792187023)*0.5**o
        ref[0,0,1]=(0.941392993115)/(o+1.)+(-0.0327366922471)+(0.644573004654)*0.5**o
        ref[0,1,0]=(1.56871601543)/(o+1.)+(-0.694552108754)+(0.829914553754)*0.5**o
        ref[0,1,1]=(-1.11610832417)/(o+1.)+(0.215432296282)+(0.874166303558)*0.5**o
        ref[1,0,0]=(-1.62546104505)/(o+1.)+(0.383991774237)+(-0.963425479949)*0.5**o
        ref[1,0,1]=(-0.581962313625)/(o+1.)+(-0.419614418097)+(-0.288791346674)*0.5**o
        ref[1,1,0]=(-0.0908115380472)/(o+1.)+(-0.412901429212)+(-0.554196270608)*0.5**o
        ref[1,1,1]=(0.165486972066)/(o+1.)+(0.230299895572)+(-0.862735330269)*0.5**o
        ref[2,0,0]=(1.27668516684)/(o+1.)+(0.132159240779)+(0.388899900771)*0.5**o
        ref[2,0,1]=(0.181208481738)/(o+1.)+(0.532562017469)+(0.324795219208)*0.5**o
        ref[2,1,0]=(-0.681276007176)/(o+1.)+(-0.399018111547)+(-0.901972287405)*0.5**o
        ref[2,1,1]=(0.987606208958)/(o+1.)+(0.935760731067)+(0.391663724321)*0.5**o
        ref[3,0,0]=(-0.817217093983)/(o+1.)+(-0.611572317549)+(0.830590943609)*0.5**o
        ref[3,0,1]=(1.29900829077)/(o+1.)+(0.0535897497116)+(0.903615175916)*0.5**o
        ref[3,1,0]=(-0.212357232141)/(o+1.)+(0.0680260229807)+(-0.901874408714)*0.5**o
        ref[3,1,1]=(0.12988482662)/(o+1.)+(-0.572686554436)+(0.807693382033)*0.5**o
        ref[4,0,0]=(0.567718442239)/(o+1.)+(0.0546658569873)+(0.440830144989)*0.5**o
        ref[4,0,1]=(0.00494625102884)/(o+1.)+(-0.589800937092)+(-0.836804887763)*0.5**o
        ref[4,1,0]=(0.302522420366)/(o+1.)+(0.0573597689848)+(0.204216251019)*0.5**o
        ref[4,1,1]=(1.27966467868)/(o+1.)+(-0.0054011358584)+(0.275074685897)*0.5**o
        ref[5,0,0]=(0.0381145194101)/(o+1.)+(-0.511461529678)+(0.370194889673)*0.5**o
        ref[5,0,1]=(0.716545624134)/(o+1.)+(-0.826779191491)+(-0.527801267225)*0.5**o
        ref[5,1,0]=(-0.196693995816)/(o+1.)+(-0.341449581363)+(0.11418535486)*0.5**o
        ref[5,1,1]=(-0.482729942895)/(o+1.)+(0.716447935227)+(0.455947680647)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactZero

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.415003066171)*x[0]**o+(-0.47441250281)*x[0]+(0.765391656872)*x[1]**o+(-0.170257448926)*x[1]
        arg[0,0,0,1]=(-0.928539602254)*x[0]**o+(0.792240012243)*x[0]+(-0.152730990885)*x[1]**o+(-0.667500454847)*x[1]
        arg[0,0,1,0]=(-0.937831078575)*x[0]**o+(0.963649346531)*x[0]+(0.505558859525)*x[1]**o+(-0.921354612243)*x[1]
        arg[0,0,1,1]=(0.909415074451)*x[0]**o+(0.199360871726)*x[0]+(0.194087349442)*x[1]**o+(0.875065381425)*x[1]
        arg[0,0,2,0]=(0.935190448625)*x[0]**o+(-0.109984767478)*x[0]+(0.518460161897)*x[1]**o+(-0.597684157349)*x[1]
        arg[0,0,2,1]=(-0.94062714999)*x[0]**o+(-0.308960574286)*x[0]+(0.578391007321)*x[1]**o+(0.864501040998)*x[1]
        arg[0,1,0,0]=(-0.197219678026)*x[0]**o+(-0.982899373778)*x[0]+(-0.603956105825)*x[1]**o+(-0.242717148913)*x[1]
        arg[0,1,0,1]=(-0.590715606611)*x[0]**o+(0.461185487788)*x[0]+(-0.270612249495)*x[1]**o+(-0.644742980654)*x[1]
        arg[0,1,1,0]=(-0.365516638755)*x[0]**o+(-0.255307471195)*x[0]+(0.0875989587033)*x[1]**o+(0.923189427818)*x[1]
        arg[0,1,1,1]=(0.315218558548)*x[0]**o+(-0.373070897551)*x[0]+(0.625272530385)*x[1]**o+(-0.174054543645)*x[1]
        arg[0,1,2,0]=(-0.502497884169)*x[0]**o+(-0.933785405889)*x[0]+(0.115500534905)*x[1]**o+(-0.864454092408)*x[1]
        arg[0,1,2,1]=(0.615837581258)*x[0]**o+(0.832538142367)*x[0]+(-0.0715650027129)*x[1]**o+(-0.775173801322)*x[1]
        arg[0,2,0,0]=(-0.898904577889)*x[0]**o+(-0.430072579352)*x[0]+(-0.532213103734)*x[1]**o+(0.970885652167)*x[1]
        arg[0,2,0,1]=(-0.0056483770747)*x[0]**o+(-0.566550554106)*x[0]+(0.751663691868)*x[1]**o+(0.677878661708)*x[1]
        arg[0,2,1,0]=(0.271196923039)*x[0]**o+(-0.291608730386)*x[0]+(-0.932579716748)*x[1]**o+(-0.437823650176)*x[1]
        arg[0,2,1,1]=(-0.788206773125)*x[0]**o+(0.18675055747)*x[0]+(0.747625244884)*x[1]**o+(0.654933611758)*x[1]
        arg[0,2,2,0]=(-0.787603701957)*x[0]**o+(-0.789784850865)*x[0]+(0.197684766271)*x[1]**o+(-0.988404958072)*x[1]
        arg[0,2,2,1]=(0.417538087399)*x[0]**o+(0.745832250585)*x[0]+(-0.525306115268)*x[1]**o+(0.134086803871)*x[1]
        arg[0,3,0,0]=(-0.230658480436)*x[0]**o+(-0.072983085881)*x[0]+(-0.149386157207)*x[1]**o+(-0.506870604002)*x[1]
        arg[0,3,0,1]=(-0.815859549425)*x[0]**o+(-0.962632495033)*x[0]+(-0.542257139282)*x[1]**o+(0.93598141472)*x[1]
        arg[0,3,1,0]=(0.0508974791413)*x[0]**o+(-0.722669867743)*x[0]+(0.714019527678)*x[1]**o+(0.710015283752)*x[1]
        arg[0,3,1,1]=(0.119419200166)*x[0]**o+(0.365663074152)*x[0]+(0.352455072404)*x[1]**o+(-0.0788953701144)*x[1]
        arg[0,3,2,0]=(-0.82075904747)*x[0]**o+(-0.224534176808)*x[0]+(-0.585094684368)*x[1]**o+(0.242958171035)*x[1]
        arg[0,3,2,1]=(-0.748061395205)*x[0]**o+(0.676995891547)*x[0]+(0.686644239408)*x[1]**o+(0.260350484065)*x[1]
        arg[0,4,0,0]=(0.128293903485)*x[0]**o+(0.226748407809)*x[0]+(0.129373873322)*x[1]**o+(0.988775321025)*x[1]
        arg[0,4,0,1]=(0.639638672694)*x[0]**o+(-0.817791454136)*x[0]+(-0.170677596319)*x[1]**o+(0.404161643274)*x[1]
        arg[0,4,1,0]=(0.43743097727)*x[0]**o+(-0.408230274495)*x[0]+(0.282544628333)*x[1]**o+(-0.542352697967)*x[1]
        arg[0,4,1,1]=(-0.995599965283)*x[0]**o+(-0.0512373236039)*x[0]+(0.218383977926)*x[1]**o+(0.929355335423)*x[1]
        arg[0,4,2,0]=(0.922609534479)*x[0]**o+(-0.166291574079)*x[0]+(-0.449331655188)*x[1]**o+(-0.0279347569009)*x[1]
        arg[0,4,2,1]=(0.950186933361)*x[0]**o+(-0.426615341822)*x[0]+(-0.167639663907)*x[1]**o+(0.219020502174)*x[1]
        arg[1,0,0,0]=(0.0573632292788)*x[0]**o+(0.543681035865)*x[0]+(-0.622510875807)*x[1]**o+(-0.171567467655)*x[1]
        arg[1,0,0,1]=(0.368975600444)*x[0]**o+(-0.754217532725)*x[0]+(0.469744047929)*x[1]**o+(-0.922445394777)*x[1]
        arg[1,0,1,0]=(-0.810194964715)*x[0]**o+(0.0861796117934)*x[0]+(0.625367069433)*x[1]**o+(-0.950906871025)*x[1]
        arg[1,0,1,1]=(0.744923275143)*x[0]**o+(0.465744045178)*x[0]+(0.359324627118)*x[1]**o+(0.273934205663)*x[1]
        arg[1,0,2,0]=(0.891959004941)*x[0]**o+(-0.498241249484)*x[0]+(0.318198109895)*x[1]**o+(0.0674564169104)*x[1]
        arg[1,0,2,1]=(0.842125439023)*x[0]**o+(0.731884782944)*x[0]+(-0.674805144805)*x[1]**o+(0.431782566319)*x[1]
        arg[1,1,0,0]=(-0.786719774805)*x[0]**o+(-0.55929144961)*x[0]+(-0.740263211102)*x[1]**o+(0.777914467998)*x[1]
        arg[1,1,0,1]=(0.831245977993)*x[0]**o+(-0.758991188073)*x[0]+(0.714032957488)*x[1]**o+(-0.879935556766)*x[1]
        arg[1,1,1,0]=(0.862796574462)*x[0]**o+(0.746875042565)*x[0]+(0.455949140326)*x[1]**o+(0.585491946591)*x[1]
        arg[1,1,1,1]=(-0.500409398892)*x[0]**o+(0.292753163538)*x[0]+(-0.62474682088)*x[1]**o+(0.583368504524)*x[1]
        arg[1,1,2,0]=(-0.402886961189)*x[0]**o+(-0.288733078165)*x[0]+(-0.770315341866)*x[1]**o+(-0.462764773299)*x[1]
        arg[1,1,2,1]=(0.358161520677)*x[0]**o+(-0.524052652951)*x[0]+(0.654162491323)*x[1]**o+(0.902382016998)*x[1]
        arg[1,2,0,0]=(-0.15539934232)*x[0]**o+(0.693162750731)*x[0]+(-0.0538737948754)*x[1]**o+(-0.025534990068)*x[1]
        arg[1,2,0,1]=(0.371159212282)*x[0]**o+(-0.751292369171)*x[0]+(-0.377617770006)*x[1]**o+(0.707647483842)*x[1]
        arg[1,2,1,0]=(0.117790556271)*x[0]**o+(0.848017682445)*x[0]+(-0.360936107893)*x[1]**o+(-0.564060873236)*x[1]
        arg[1,2,1,1]=(-0.524828128971)*x[0]**o+(0.737876348154)*x[0]+(0.183316445987)*x[1]**o+(-0.821207746237)*x[1]
        arg[1,2,2,0]=(0.300516177535)*x[0]**o+(-0.204187510383)*x[0]+(0.126601242737)*x[1]**o+(0.0794834423875)*x[1]
        arg[1,2,2,1]=(-0.29046926624)*x[0]**o+(-0.310057692431)*x[0]+(-0.810227948574)*x[1]**o+(-0.464128160277)*x[1]
        arg[1,3,0,0]=(-0.134275834453)*x[0]**o+(0.448577162559)*x[0]+(0.757723452446)*x[1]**o+(0.139202834725)*x[1]
        arg[1,3,0,1]=(0.882915728842)*x[0]**o+(-0.0750591947315)*x[0]+(0.503482692135)*x[1]**o+(-0.176350465386)*x[1]
        arg[1,3,1,0]=(-0.954792910945)*x[0]**o+(0.772215353622)*x[0]+(-0.861747803918)*x[1]**o+(-0.932727438775)*x[1]
        arg[1,3,1,1]=(0.747255393187)*x[0]**o+(0.0392661297355)*x[0]+(-0.322274864209)*x[1]**o+(-0.491462177972)*x[1]
        arg[1,3,2,0]=(0.139287057836)*x[0]**o+(0.312258185231)*x[0]+(-0.488288412308)*x[1]**o+(-0.185131700928)*x[1]
        arg[1,3,2,1]=(-0.861625516304)*x[0]**o+(0.681454010645)*x[0]+(0.206644380777)*x[1]**o+(-0.0344138929268)*x[1]
        arg[1,4,0,0]=(-0.714198122614)*x[0]**o+(-0.413113909739)*x[0]+(-0.971680855409)*x[1]**o+(0.602598611545)*x[1]
        arg[1,4,0,1]=(0.293578109913)*x[0]**o+(-0.495635685351)*x[0]+(0.347408878778)*x[1]**o+(-0.923814185786)*x[1]
        arg[1,4,1,0]=(0.843760126513)*x[0]**o+(-0.671075074679)*x[0]+(-0.284086603407)*x[1]**o+(0.437492233496)*x[1]
        arg[1,4,1,1]=(-0.342911357374)*x[0]**o+(-0.440232224635)*x[0]+(0.537701123084)*x[1]**o+(-0.525203729082)*x[1]
        arg[1,4,2,0]=(-0.71201360314)*x[0]**o+(-0.706332525171)*x[0]+(0.831438049852)*x[1]**o+(0.819141670055)*x[1]
        arg[1,4,2,1]=(-0.616403541928)*x[0]**o+(0.635973739636)*x[0]+(-0.0369972206543)*x[1]**o+(-0.08595692743)*x[1]
        arg[2,0,0,0]=(-0.959754423491)*x[0]**o+(-0.332473557912)*x[0]+(-0.464334304695)*x[1]**o+(0.597214020843)*x[1]
        arg[2,0,0,1]=(0.231051530608)*x[0]**o+(0.368524293555)*x[0]+(0.173044245392)*x[1]**o+(0.505867390994)*x[1]
        arg[2,0,1,0]=(0.376577055194)*x[0]**o+(-0.739990982509)*x[0]+(-0.816201949845)*x[1]**o+(-0.71233193641)*x[1]
        arg[2,0,1,1]=(-0.4891837531)*x[0]**o+(-0.121562226208)*x[0]+(-0.876294548464)*x[1]**o+(0.377763924093)*x[1]
        arg[2,0,2,0]=(-0.759548081841)*x[0]**o+(-0.118749434442)*x[0]+(0.602375475355)*x[1]**o+(-0.877213577079)*x[1]
        arg[2,0,2,1]=(0.114018259802)*x[0]**o+(-0.627149204501)*x[0]+(0.813648191424)*x[1]**o+(-0.612548616323)*x[1]
        arg[2,1,0,0]=(-0.572699800518)*x[0]**o+(0.507715829662)*x[0]+(-0.12122456959)*x[1]**o+(0.75522878228)*x[1]
        arg[2,1,0,1]=(0.0655268452799)*x[0]**o+(0.211191572823)*x[0]+(0.539760655497)*x[1]**o+(0.657397992057)*x[1]
        arg[2,1,1,0]=(0.358281119336)*x[0]**o+(0.119364032507)*x[0]+(-0.0242249415295)*x[1]**o+(-0.645986730609)*x[1]
        arg[2,1,1,1]=(0.523751704889)*x[0]**o+(-0.237670103222)*x[0]+(-0.995787126907)*x[1]**o+(-0.504129088442)*x[1]
        arg[2,1,2,0]=(-0.185380508646)*x[0]**o+(-0.218341933583)*x[0]+(0.149400154478)*x[1]**o+(0.00252616040715)*x[1]
        arg[2,1,2,1]=(0.81086932122)*x[0]**o+(0.387271527481)*x[0]+(0.908525646396)*x[1]**o+(-0.75383394204)*x[1]
        arg[2,2,0,0]=(-0.463416100256)*x[0]**o+(-0.891654021409)*x[0]+(-0.0495088159461)*x[1]**o+(0.398915277911)*x[1]
        arg[2,2,0,1]=(-0.200698149193)*x[0]**o+(0.127899540753)*x[0]+(-0.895280577411)*x[1]**o+(-0.197025422242)*x[1]
        arg[2,2,1,0]=(0.888683772867)*x[0]**o+(-0.84436563509)*x[0]+(-0.375709672588)*x[1]**o+(0.591336312818)*x[1]
        arg[2,2,1,1]=(-0.0167949020901)*x[0]**o+(0.145017501865)*x[0]+(0.868194479789)*x[1]**o+(-0.62439676328)*x[1]
        arg[2,2,2,0]=(0.441981698578)*x[0]**o+(-0.481156166596)*x[0]+(0.267937553127)*x[1]**o+(-0.37032246733)*x[1]
        arg[2,2,2,1]=(0.876701424579)*x[0]**o+(0.473798212378)*x[0]+(0.697645775876)*x[1]**o+(-0.798143376838)*x[1]
        arg[2,3,0,0]=(0.111231509217)*x[0]**o+(0.188121806171)*x[0]+(-0.0706155476337)*x[1]**o+(0.779423336221)*x[1]
        arg[2,3,0,1]=(-0.858913402812)*x[0]**o+(0.655459361097)*x[0]+(0.326282512878)*x[1]**o+(0.863386952614)*x[1]
        arg[2,3,1,0]=(-0.524551221953)*x[0]**o+(-0.807242415277)*x[0]+(0.575866251831)*x[1]**o+(-0.74503615846)*x[1]
        arg[2,3,1,1]=(0.391514075129)*x[0]**o+(0.477257809825)*x[0]+(-0.488263373008)*x[1]**o+(-0.29101442481)*x[1]
        arg[2,3,2,0]=(0.26532129161)*x[0]**o+(0.22794872765)*x[0]+(0.85094548724)*x[1]**o+(-0.369690624817)*x[1]
        arg[2,3,2,1]=(0.2339994493)*x[0]**o+(-0.901432708956)*x[0]+(0.309349741876)*x[1]**o+(0.132010344317)*x[1]
        arg[2,4,0,0]=(-0.713553783036)*x[0]**o+(0.791484993301)*x[0]+(0.339247480459)*x[1]**o+(0.158842307171)*x[1]
        arg[2,4,0,1]=(-0.733490022564)*x[0]**o+(0.631284830844)*x[0]+(0.340009298488)*x[1]**o+(-0.483162310574)*x[1]
        arg[2,4,1,0]=(0.709413437683)*x[0]**o+(0.896590124659)*x[0]+(0.680862759763)*x[1]**o+(-0.23579845663)*x[1]
        arg[2,4,1,1]=(-0.130761551034)*x[0]**o+(0.269095282394)*x[0]+(0.687201138134)*x[1]**o+(0.905183436376)*x[1]
        arg[2,4,2,0]=(0.256894911335)*x[0]**o+(0.706744915172)*x[0]+(0.910504833397)*x[1]**o+(-0.285088297895)*x[1]
        arg[2,4,2,1]=(0.95137247728)*x[0]**o+(-0.924669436646)*x[0]+(-0.146356272875)*x[1]**o+(-0.673326194162)*x[1]
        arg[3,0,0,0]=(0.663180444534)*x[0]**o+(0.894728781279)*x[0]+(0.894383747285)*x[1]**o+(0.969779862148)*x[1]
        arg[3,0,0,1]=(0.424556283891)*x[0]**o+(-0.97849198275)*x[0]+(0.367908872519)*x[1]**o+(0.624098355956)*x[1]
        arg[3,0,1,0]=(0.908698580251)*x[0]**o+(-0.899102004477)*x[0]+(0.509310001763)*x[1]**o+(-0.8399697646)*x[1]
        arg[3,0,1,1]=(0.523444225814)*x[0]**o+(0.410650511976)*x[0]+(-0.986699126405)*x[1]**o+(0.081494682537)*x[1]
        arg[3,0,2,0]=(-0.163604985314)*x[0]**o+(0.959338619715)*x[0]+(-0.843900271478)*x[1]**o+(-0.0218386544947)*x[1]
        arg[3,0,2,1]=(-0.490804271432)*x[0]**o+(0.123332672938)*x[0]+(0.902072835174)*x[1]**o+(-0.202340709957)*x[1]
        arg[3,1,0,0]=(-0.949731382685)*x[0]**o+(0.533695608969)*x[0]+(0.430512336347)*x[1]**o+(0.977434372776)*x[1]
        arg[3,1,0,1]=(0.945102298088)*x[0]**o+(-0.117807763531)*x[0]+(-0.203921504027)*x[1]**o+(-0.261266079346)*x[1]
        arg[3,1,1,0]=(0.762442770795)*x[0]**o+(-0.318456087863)*x[0]+(-0.365021977395)*x[1]**o+(0.0516932816311)*x[1]
        arg[3,1,1,1]=(-0.829061607399)*x[0]**o+(0.854900781365)*x[0]+(-0.397139314214)*x[1]**o+(-0.549679351555)*x[1]
        arg[3,1,2,0]=(-0.780040060687)*x[0]**o+(0.719393460042)*x[0]+(-0.435377419129)*x[1]**o+(0.655962063649)*x[1]
        arg[3,1,2,1]=(-0.853144390228)*x[0]**o+(0.0940222442927)*x[0]+(-0.642258121264)*x[1]**o+(-0.615386923437)*x[1]
        arg[3,2,0,0]=(0.377161814597)*x[0]**o+(0.349238572634)*x[0]+(0.361300611554)*x[1]**o+(0.554382639639)*x[1]
        arg[3,2,0,1]=(-0.51174671413)*x[0]**o+(-0.0586248632463)*x[0]+(0.083224329707)*x[1]**o+(0.506614699443)*x[1]
        arg[3,2,1,0]=(0.612695079311)*x[0]**o+(0.82053876378)*x[0]+(-0.132608876197)*x[1]**o+(0.701720563108)*x[1]
        arg[3,2,1,1]=(0.0928734170244)*x[0]**o+(-0.127564289866)*x[0]+(0.68468196965)*x[1]**o+(0.0168958305852)*x[1]
        arg[3,2,2,0]=(-0.863542893122)*x[0]**o+(0.0533007397457)*x[0]+(-0.0602194449612)*x[1]**o+(0.391122261927)*x[1]
        arg[3,2,2,1]=(-0.0942537498633)*x[0]**o+(-0.421741120475)*x[0]+(-0.0746938765947)*x[1]**o+(0.988111802567)*x[1]
        arg[3,3,0,0]=(-0.0588913579205)*x[0]**o+(0.439330145374)*x[0]+(-0.541714868844)*x[1]**o+(0.845344272279)*x[1]
        arg[3,3,0,1]=(0.221784155922)*x[0]**o+(0.983539923382)*x[0]+(-0.32332836636)*x[1]**o+(-0.892579743718)*x[1]
        arg[3,3,1,0]=(0.804097639203)*x[0]**o+(-0.342072604164)*x[0]+(0.755681096012)*x[1]**o+(0.432897246925)*x[1]
        arg[3,3,1,1]=(0.825696602707)*x[0]**o+(0.536999112965)*x[0]+(0.745033643035)*x[1]**o+(0.156806358776)*x[1]
        arg[3,3,2,0]=(-0.372114391352)*x[0]**o+(0.670417556114)*x[0]+(-0.569395510546)*x[1]**o+(-0.523784408534)*x[1]
        arg[3,3,2,1]=(0.439723359721)*x[0]**o+(0.809733189174)*x[0]+(0.781753556415)*x[1]**o+(0.375991547044)*x[1]
        arg[3,4,0,0]=(-0.429896412297)*x[0]**o+(-0.908820042346)*x[0]+(0.328787484151)*x[1]**o+(-0.610488251755)*x[1]
        arg[3,4,0,1]=(0.803567111353)*x[0]**o+(-0.487219379817)*x[0]+(-0.541538468561)*x[1]**o+(-0.692964506777)*x[1]
        arg[3,4,1,0]=(-0.963266835039)*x[0]**o+(0.667510333677)*x[0]+(0.308312799035)*x[1]**o+(0.686605377458)*x[1]
        arg[3,4,1,1]=(0.651198604373)*x[0]**o+(0.306465142929)*x[0]+(-0.923691546947)*x[1]**o+(-0.376240035765)*x[1]
        arg[3,4,2,0]=(-0.339406539705)*x[0]**o+(0.524709759622)*x[0]+(-0.0326564953763)*x[1]**o+(-0.658296780229)*x[1]
        arg[3,4,2,1]=(0.471355209305)*x[0]**o+(-0.473756698531)*x[0]+(0.193890658409)*x[1]**o+(-0.266632530307)*x[1]
        ref[0,0,0,0]=(0.765391656872)/(o+1.)+(-0.322334975868)+(0.415003066171)*0.5**o
        ref[0,0,0,1]=(-0.152730990885)/(o+1.)+(0.0623697786981)+(-0.928539602254)*0.5**o
        ref[0,0,1,0]=(0.505558859525)/(o+1.)+(0.0211473671443)+(-0.937831078575)*0.5**o
        ref[0,0,1,1]=(0.194087349442)/(o+1.)+(0.537213126576)+(0.909415074451)*0.5**o
        ref[0,0,2,0]=(0.518460161897)/(o+1.)+(-0.353834462413)+(0.935190448625)*0.5**o
        ref[0,0,2,1]=(0.578391007321)/(o+1.)+(0.277770233356)+(-0.94062714999)*0.5**o
        ref[0,1,0,0]=(-0.603956105825)/(o+1.)+(-0.612808261345)+(-0.197219678026)*0.5**o
        ref[0,1,0,1]=(-0.270612249495)/(o+1.)+(-0.0917787464331)+(-0.590715606611)*0.5**o
        ref[0,1,1,0]=(0.0875989587033)/(o+1.)+(0.333940978312)+(-0.365516638755)*0.5**o
        ref[0,1,1,1]=(0.625272530385)/(o+1.)+(-0.273562720598)+(0.315218558548)*0.5**o
        ref[0,1,2,0]=(0.115500534905)/(o+1.)+(-0.899119749149)+(-0.502497884169)*0.5**o
        ref[0,1,2,1]=(-0.0715650027129)/(o+1.)+(0.0286821705227)+(0.615837581258)*0.5**o
        ref[0,2,0,0]=(-0.532213103734)/(o+1.)+(0.270406536408)+(-0.898904577889)*0.5**o
        ref[0,2,0,1]=(0.751663691868)/(o+1.)+(0.0556640538013)+(-0.0056483770747)*0.5**o
        ref[0,2,1,0]=(-0.932579716748)/(o+1.)+(-0.364716190281)+(0.271196923039)*0.5**o
        ref[0,2,1,1]=(0.747625244884)/(o+1.)+(0.420842084614)+(-0.788206773125)*0.5**o
        ref[0,2,2,0]=(0.197684766271)/(o+1.)+(-0.889094904468)+(-0.787603701957)*0.5**o
        ref[0,2,2,1]=(-0.525306115268)/(o+1.)+(0.439959527228)+(0.417538087399)*0.5**o
        ref[0,3,0,0]=(-0.149386157207)/(o+1.)+(-0.289926844941)+(-0.230658480436)*0.5**o
        ref[0,3,0,1]=(-0.542257139282)/(o+1.)+(-0.0133255401565)+(-0.815859549425)*0.5**o
        ref[0,3,1,0]=(0.714019527678)/(o+1.)+(-0.00632729199566)+(0.0508974791413)*0.5**o
        ref[0,3,1,1]=(0.352455072404)/(o+1.)+(0.143383852019)+(0.119419200166)*0.5**o
        ref[0,3,2,0]=(-0.585094684368)/(o+1.)+(0.00921199711359)+(-0.82075904747)*0.5**o
        ref[0,3,2,1]=(0.686644239408)/(o+1.)+(0.468673187806)+(-0.748061395205)*0.5**o
        ref[0,4,0,0]=(0.129373873322)/(o+1.)+(0.607761864417)+(0.128293903485)*0.5**o
        ref[0,4,0,1]=(-0.170677596319)/(o+1.)+(-0.206814905431)+(0.639638672694)*0.5**o
        ref[0,4,1,0]=(0.282544628333)/(o+1.)+(-0.475291486231)+(0.43743097727)*0.5**o
        ref[0,4,1,1]=(0.218383977926)/(o+1.)+(0.43905900591)+(-0.995599965283)*0.5**o
        ref[0,4,2,0]=(-0.449331655188)/(o+1.)+(-0.0971131654902)+(0.922609534479)*0.5**o
        ref[0,4,2,1]=(-0.167639663907)/(o+1.)+(-0.103797419824)+(0.950186933361)*0.5**o
        ref[1,0,0,0]=(-0.622510875807)/(o+1.)+(0.186056784105)+(0.0573632292788)*0.5**o
        ref[1,0,0,1]=(0.469744047929)/(o+1.)+(-0.838331463751)+(0.368975600444)*0.5**o
        ref[1,0,1,0]=(0.625367069433)/(o+1.)+(-0.432363629616)+(-0.810194964715)*0.5**o
        ref[1,0,1,1]=(0.359324627118)/(o+1.)+(0.36983912542)+(0.744923275143)*0.5**o
        ref[1,0,2,0]=(0.318198109895)/(o+1.)+(-0.215392416287)+(0.891959004941)*0.5**o
        ref[1,0,2,1]=(-0.674805144805)/(o+1.)+(0.581833674632)+(0.842125439023)*0.5**o
        ref[1,1,0,0]=(-0.740263211102)/(o+1.)+(0.109311509194)+(-0.786719774805)*0.5**o
        ref[1,1,0,1]=(0.714032957488)/(o+1.)+(-0.81946337242)+(0.831245977993)*0.5**o
        ref[1,1,1,0]=(0.455949140326)/(o+1.)+(0.666183494578)+(0.862796574462)*0.5**o
        ref[1,1,1,1]=(-0.62474682088)/(o+1.)+(0.438060834031)+(-0.500409398892)*0.5**o
        ref[1,1,2,0]=(-0.770315341866)/(o+1.)+(-0.375748925732)+(-0.402886961189)*0.5**o
        ref[1,1,2,1]=(0.654162491323)/(o+1.)+(0.189164682023)+(0.358161520677)*0.5**o
        ref[1,2,0,0]=(-0.0538737948754)/(o+1.)+(0.333813880332)+(-0.15539934232)*0.5**o
        ref[1,2,0,1]=(-0.377617770006)/(o+1.)+(-0.0218224426646)+(0.371159212282)*0.5**o
        ref[1,2,1,0]=(-0.360936107893)/(o+1.)+(0.141978404604)+(0.117790556271)*0.5**o
        ref[1,2,1,1]=(0.183316445987)/(o+1.)+(-0.0416656990415)+(-0.524828128971)*0.5**o
        ref[1,2,2,0]=(0.126601242737)/(o+1.)+(-0.0623520339976)+(0.300516177535)*0.5**o
        ref[1,2,2,1]=(-0.810227948574)/(o+1.)+(-0.387092926354)+(-0.29046926624)*0.5**o
        ref[1,3,0,0]=(0.757723452446)/(o+1.)+(0.293889998642)+(-0.134275834453)*0.5**o
        ref[1,3,0,1]=(0.503482692135)/(o+1.)+(-0.125704830059)+(0.882915728842)*0.5**o
        ref[1,3,1,0]=(-0.861747803918)/(o+1.)+(-0.0802560425765)+(-0.954792910945)*0.5**o
        ref[1,3,1,1]=(-0.322274864209)/(o+1.)+(-0.226098024118)+(0.747255393187)*0.5**o
        ref[1,3,2,0]=(-0.488288412308)/(o+1.)+(0.0635632421511)+(0.139287057836)*0.5**o
        ref[1,3,2,1]=(0.206644380777)/(o+1.)+(0.323520058859)+(-0.861625516304)*0.5**o
        ref[1,4,0,0]=(-0.971680855409)/(o+1.)+(0.0947423509032)+(-0.714198122614)*0.5**o
        ref[1,4,0,1]=(0.347408878778)/(o+1.)+(-0.709724935569)+(0.293578109913)*0.5**o
        ref[1,4,1,0]=(-0.284086603407)/(o+1.)+(-0.116791420591)+(0.843760126513)*0.5**o
        ref[1,4,1,1]=(0.537701123084)/(o+1.)+(-0.482717976859)+(-0.342911357374)*0.5**o
        ref[1,4,2,0]=(0.831438049852)/(o+1.)+(0.056404572442)+(-0.71201360314)*0.5**o
        ref[1,4,2,1]=(-0.0369972206543)/(o+1.)+(0.275008406103)+(-0.616403541928)*0.5**o
        ref[2,0,0,0]=(-0.464334304695)/(o+1.)+(0.132370231466)+(-0.959754423491)*0.5**o
        ref[2,0,0,1]=(0.173044245392)/(o+1.)+(0.437195842274)+(0.231051530608)*0.5**o
        ref[2,0,1,0]=(-0.816201949845)/(o+1.)+(-0.72616145946)+(0.376577055194)*0.5**o
        ref[2,0,1,1]=(-0.876294548464)/(o+1.)+(0.128100848943)+(-0.4891837531)*0.5**o
        ref[2,0,2,0]=(0.602375475355)/(o+1.)+(-0.497981505761)+(-0.759548081841)*0.5**o
        ref[2,0,2,1]=(0.813648191424)/(o+1.)+(-0.619848910412)+(0.114018259802)*0.5**o
        ref[2,1,0,0]=(-0.12122456959)/(o+1.)+(0.631472305971)+(-0.572699800518)*0.5**o
        ref[2,1,0,1]=(0.539760655497)/(o+1.)+(0.43429478244)+(0.0655268452799)*0.5**o
        ref[2,1,1,0]=(-0.0242249415295)/(o+1.)+(-0.263311349051)+(0.358281119336)*0.5**o
        ref[2,1,1,1]=(-0.995787126907)/(o+1.)+(-0.370899595832)+(0.523751704889)*0.5**o
        ref[2,1,2,0]=(0.149400154478)/(o+1.)+(-0.107907886588)+(-0.185380508646)*0.5**o
        ref[2,1,2,1]=(0.908525646396)/(o+1.)+(-0.18328120728)+(0.81086932122)*0.5**o
        ref[2,2,0,0]=(-0.0495088159461)/(o+1.)+(-0.246369371749)+(-0.463416100256)*0.5**o
        ref[2,2,0,1]=(-0.895280577411)/(o+1.)+(-0.0345629407446)+(-0.200698149193)*0.5**o
        ref[2,2,1,0]=(-0.375709672588)/(o+1.)+(-0.126514661136)+(0.888683772867)*0.5**o
        ref[2,2,1,1]=(0.868194479789)/(o+1.)+(-0.239689630708)+(-0.0167949020901)*0.5**o
        ref[2,2,2,0]=(0.267937553127)/(o+1.)+(-0.425739316963)+(0.441981698578)*0.5**o
        ref[2,2,2,1]=(0.697645775876)/(o+1.)+(-0.16217258223)+(0.876701424579)*0.5**o
        ref[2,3,0,0]=(-0.0706155476337)/(o+1.)+(0.483772571196)+(0.111231509217)*0.5**o
        ref[2,3,0,1]=(0.326282512878)/(o+1.)+(0.759423156855)+(-0.858913402812)*0.5**o
        ref[2,3,1,0]=(0.575866251831)/(o+1.)+(-0.776139286869)+(-0.524551221953)*0.5**o
        ref[2,3,1,1]=(-0.488263373008)/(o+1.)+(0.0931216925076)+(0.391514075129)*0.5**o
        ref[2,3,2,0]=(0.85094548724)/(o+1.)+(-0.0708709485834)+(0.26532129161)*0.5**o
        ref[2,3,2,1]=(0.309349741876)/(o+1.)+(-0.384711182319)+(0.2339994493)*0.5**o
        ref[2,4,0,0]=(0.339247480459)/(o+1.)+(0.475163650236)+(-0.713553783036)*0.5**o
        ref[2,4,0,1]=(0.340009298488)/(o+1.)+(0.0740612601349)+(-0.733490022564)*0.5**o
        ref[2,4,1,0]=(0.680862759763)/(o+1.)+(0.330395834014)+(0.709413437683)*0.5**o
        ref[2,4,1,1]=(0.687201138134)/(o+1.)+(0.587139359385)+(-0.130761551034)*0.5**o
        ref[2,4,2,0]=(0.910504833397)/(o+1.)+(0.210828308638)+(0.256894911335)*0.5**o
        ref[2,4,2,1]=(-0.146356272875)/(o+1.)+(-0.798997815404)+(0.95137247728)*0.5**o
        ref[3,0,0,0]=(0.894383747285)/(o+1.)+(0.932254321713)+(0.663180444534)*0.5**o
        ref[3,0,0,1]=(0.367908872519)/(o+1.)+(-0.177196813397)+(0.424556283891)*0.5**o
        ref[3,0,1,0]=(0.509310001763)/(o+1.)+(-0.869535884539)+(0.908698580251)*0.5**o
        ref[3,0,1,1]=(-0.986699126405)/(o+1.)+(0.246072597257)+(0.523444225814)*0.5**o
        ref[3,0,2,0]=(-0.843900271478)/(o+1.)+(0.46874998261)+(-0.163604985314)*0.5**o
        ref[3,0,2,1]=(0.902072835174)/(o+1.)+(-0.0395040185096)+(-0.490804271432)*0.5**o
        ref[3,1,0,0]=(0.430512336347)/(o+1.)+(0.755564990872)+(-0.949731382685)*0.5**o
        ref[3,1,0,1]=(-0.203921504027)/(o+1.)+(-0.189536921439)+(0.945102298088)*0.5**o
        ref[3,1,1,0]=(-0.365021977395)/(o+1.)+(-0.133381403116)+(0.762442770795)*0.5**o
        ref[3,1,1,1]=(-0.397139314214)/(o+1.)+(0.152610714905)+(-0.829061607399)*0.5**o
        ref[3,1,2,0]=(-0.435377419129)/(o+1.)+(0.687677761846)+(-0.780040060687)*0.5**o
        ref[3,1,2,1]=(-0.642258121264)/(o+1.)+(-0.260682339572)+(-0.853144390228)*0.5**o
        ref[3,2,0,0]=(0.361300611554)/(o+1.)+(0.451810606136)+(0.377161814597)*0.5**o
        ref[3,2,0,1]=(0.083224329707)/(o+1.)+(0.223994918098)+(-0.51174671413)*0.5**o
        ref[3,2,1,0]=(-0.132608876197)/(o+1.)+(0.761129663444)+(0.612695079311)*0.5**o
        ref[3,2,1,1]=(0.68468196965)/(o+1.)+(-0.0553342296404)+(0.0928734170244)*0.5**o
        ref[3,2,2,0]=(-0.0602194449612)/(o+1.)+(0.222211500837)+(-0.863542893122)*0.5**o
        ref[3,2,2,1]=(-0.0746938765947)/(o+1.)+(0.283185341046)+(-0.0942537498633)*0.5**o
        ref[3,3,0,0]=(-0.541714868844)/(o+1.)+(0.642337208826)+(-0.0588913579205)*0.5**o
        ref[3,3,0,1]=(-0.32332836636)/(o+1.)+(0.0454800898319)+(0.221784155922)*0.5**o
        ref[3,3,1,0]=(0.755681096012)/(o+1.)+(0.0454123213802)+(0.804097639203)*0.5**o
        ref[3,3,1,1]=(0.745033643035)/(o+1.)+(0.34690273587)+(0.825696602707)*0.5**o
        ref[3,3,2,0]=(-0.569395510546)/(o+1.)+(0.0733165737898)+(-0.372114391352)*0.5**o
        ref[3,3,2,1]=(0.781753556415)/(o+1.)+(0.592862368109)+(0.439723359721)*0.5**o
        ref[3,4,0,0]=(0.328787484151)/(o+1.)+(-0.75965414705)+(-0.429896412297)*0.5**o
        ref[3,4,0,1]=(-0.541538468561)/(o+1.)+(-0.590091943297)+(0.803567111353)*0.5**o
        ref[3,4,1,0]=(0.308312799035)/(o+1.)+(0.677057855567)+(-0.963266835039)*0.5**o
        ref[3,4,1,1]=(-0.923691546947)/(o+1.)+(-0.0348874464182)+(0.651198604373)*0.5**o
        ref[3,4,2,0]=(-0.0326564953763)/(o+1.)+(-0.0667935103036)+(-0.339406539705)*0.5**o
        ref[3,4,2,1]=(0.193890658409)/(o+1.)+(-0.370194614419)+(0.471355209305)*0.5**o
      else:
        arg[0,0,0,0]=(0.0737627367843)*x[0]**o+(-0.897508234376)*x[0]+(0.499361502035)*x[1]**o+(-0.830793706142)*x[1]+(0.967672023686)*x[2]**o+(-0.94209897356)*x[2]
        arg[0,0,0,1]=(0.353968533247)*x[0]**o+(-0.828575104081)*x[0]+(0.111197799977)*x[1]**o+(0.475627052455)*x[1]+(0.10233739786)*x[2]**o+(-0.297090205985)*x[2]
        arg[0,0,1,0]=(0.315704665816)*x[0]**o+(0.450577419574)*x[0]+(-0.0225047225996)*x[1]**o+(0.221436098936)*x[1]+(-0.210708820992)*x[2]**o+(0.22813686037)*x[2]
        arg[0,0,1,1]=(-0.453912049362)*x[0]**o+(0.443266972459)*x[0]+(-0.166003491301)*x[1]**o+(0.325157034345)*x[1]+(0.975435329063)*x[2]**o+(0.516477372955)*x[2]
        arg[0,0,2,0]=(0.705225432895)*x[0]**o+(0.406870556161)*x[0]+(0.967131048725)*x[1]**o+(-0.31111693058)*x[1]+(0.915125425876)*x[2]**o+(-0.417407852906)*x[2]
        arg[0,0,2,1]=(-0.397162258761)*x[0]**o+(-0.49079016751)*x[0]+(-0.444905720899)*x[1]**o+(0.258998080457)*x[1]+(-0.406172174923)*x[2]**o+(0.166620074269)*x[2]
        arg[0,1,0,0]=(0.00444300957662)*x[0]**o+(0.59448685984)*x[0]+(-0.755596203939)*x[1]**o+(-0.587866578467)*x[1]+(-0.994620629569)*x[2]**o+(0.558539124849)*x[2]
        arg[0,1,0,1]=(0.726535641072)*x[0]**o+(-0.515254556008)*x[0]+(-0.737095044797)*x[1]**o+(0.396218854768)*x[1]+(-0.873340093434)*x[2]**o+(0.538194586759)*x[2]
        arg[0,1,1,0]=(-0.554579794515)*x[0]**o+(-0.754117891877)*x[0]+(0.242579027102)*x[1]**o+(0.368307663104)*x[1]+(0.308372932825)*x[2]**o+(-0.521440130826)*x[2]
        arg[0,1,1,1]=(0.294744137189)*x[0]**o+(0.44996935134)*x[0]+(0.118263237453)*x[1]**o+(-0.143566244715)*x[1]+(-0.927385432821)*x[2]**o+(0.912735237305)*x[2]
        arg[0,1,2,0]=(-0.30261020867)*x[0]**o+(-0.994883662031)*x[0]+(0.514662053738)*x[1]**o+(0.327770859964)*x[1]+(0.862452368458)*x[2]**o+(-0.739026842429)*x[2]
        arg[0,1,2,1]=(0.494855964972)*x[0]**o+(-0.358461686787)*x[0]+(0.406489619959)*x[1]**o+(0.417722618206)*x[1]+(0.679808195501)*x[2]**o+(-0.00502769522311)*x[2]
        arg[0,2,0,0]=(-0.914585215232)*x[0]**o+(0.305890296017)*x[0]+(-0.867826544668)*x[1]**o+(0.304580882671)*x[1]+(-0.401401956053)*x[2]**o+(0.842276423411)*x[2]
        arg[0,2,0,1]=(-0.260590397572)*x[0]**o+(0.541867335986)*x[0]+(0.0279475511153)*x[1]**o+(0.0255320376447)*x[1]+(-0.584584718877)*x[2]**o+(-0.640388032783)*x[2]
        arg[0,2,1,0]=(0.0455797513404)*x[0]**o+(-0.764984318339)*x[0]+(-0.257174640639)*x[1]**o+(0.745888292962)*x[1]+(0.964524202534)*x[2]**o+(0.435680191282)*x[2]
        arg[0,2,1,1]=(-0.300900136798)*x[0]**o+(-0.899164082043)*x[0]+(-0.63273853331)*x[1]**o+(0.56522562781)*x[1]+(-0.846482371222)*x[2]**o+(-0.909985752269)*x[2]
        arg[0,2,2,0]=(0.745711773093)*x[0]**o+(0.0627291150681)*x[0]+(-0.090340386706)*x[1]**o+(-0.0912293084958)*x[1]+(0.108613205226)*x[2]**o+(-0.28985444593)*x[2]
        arg[0,2,2,1]=(0.291340077334)*x[0]**o+(0.592856577974)*x[0]+(0.141857446535)*x[1]**o+(0.830526360843)*x[1]+(0.956840406828)*x[2]**o+(-0.717465405888)*x[2]
        arg[0,3,0,0]=(0.89242099254)*x[0]**o+(0.529348654364)*x[0]+(-0.074382350183)*x[1]**o+(0.874670648345)*x[1]+(-0.728104560168)*x[2]**o+(0.0781153963606)*x[2]
        arg[0,3,0,1]=(0.095390666498)*x[0]**o+(0.0124610436496)*x[0]+(0.278014506594)*x[1]**o+(0.923928359216)*x[1]+(-0.0905667357972)*x[2]**o+(-0.894031439498)*x[2]
        arg[0,3,1,0]=(-0.317750910722)*x[0]**o+(0.677406387747)*x[0]+(0.550094997541)*x[1]**o+(-0.30098732949)*x[1]+(0.501753868155)*x[2]**o+(-0.42859127752)*x[2]
        arg[0,3,1,1]=(-0.204455448392)*x[0]**o+(0.445237113438)*x[0]+(0.853596204887)*x[1]**o+(0.315701645178)*x[1]+(0.307192864939)*x[2]**o+(-0.708325809173)*x[2]
        arg[0,3,2,0]=(0.38581656579)*x[0]**o+(-0.79289714243)*x[0]+(-0.412226202311)*x[1]**o+(0.655534112291)*x[1]+(0.31352728014)*x[2]**o+(0.502089385396)*x[2]
        arg[0,3,2,1]=(-0.225571343976)*x[0]**o+(0.687765563016)*x[0]+(0.0421378888728)*x[1]**o+(0.54466301835)*x[1]+(-0.430288255786)*x[2]**o+(-0.117571440224)*x[2]
        arg[0,4,0,0]=(0.834083057621)*x[0]**o+(0.0292441566815)*x[0]+(-0.947154369416)*x[1]**o+(0.597153808995)*x[1]+(0.323706009194)*x[2]**o+(0.664093182644)*x[2]
        arg[0,4,0,1]=(-0.613152578007)*x[0]**o+(0.253155875339)*x[0]+(0.297138228345)*x[1]**o+(-0.604393247773)*x[1]+(-0.936718722281)*x[2]**o+(-0.410140848312)*x[2]
        arg[0,4,1,0]=(-0.373668614153)*x[0]**o+(-0.401240526097)*x[0]+(-0.417458139032)*x[1]**o+(0.456033864594)*x[1]+(-0.158434012225)*x[2]**o+(-0.33938136039)*x[2]
        arg[0,4,1,1]=(0.00943802199128)*x[0]**o+(-0.800756064115)*x[0]+(0.899199671858)*x[1]**o+(0.376830611387)*x[1]+(-0.930822199046)*x[2]**o+(0.192235807799)*x[2]
        arg[0,4,2,0]=(0.157099024354)*x[0]**o+(0.195942080381)*x[0]+(-0.554519054021)*x[1]**o+(0.871823312361)*x[1]+(-0.34803033039)*x[2]**o+(-0.646933331859)*x[2]
        arg[0,4,2,1]=(0.933762543076)*x[0]**o+(-0.169843985744)*x[0]+(0.971089176971)*x[1]**o+(-0.605176382286)*x[1]+(0.254369209427)*x[2]**o+(0.0314820745541)*x[2]
        arg[1,0,0,0]=(0.0417476479889)*x[0]**o+(-0.144293810851)*x[0]+(-0.677039420744)*x[1]**o+(-0.137097132377)*x[1]+(0.721082639859)*x[2]**o+(-0.32510577691)*x[2]
        arg[1,0,0,1]=(0.450986977446)*x[0]**o+(-0.921371106219)*x[0]+(0.536898979198)*x[1]**o+(-0.814951108186)*x[1]+(-0.733774506956)*x[2]**o+(0.938941747764)*x[2]
        arg[1,0,1,0]=(-0.23712795738)*x[0]**o+(0.450883240541)*x[0]+(0.101102533189)*x[1]**o+(-0.998253664232)*x[1]+(-0.843977488601)*x[2]**o+(-0.807025450448)*x[2]
        arg[1,0,1,1]=(-0.308721170682)*x[0]**o+(-0.546268734305)*x[0]+(0.643628083343)*x[1]**o+(0.206934468523)*x[1]+(-0.257667927436)*x[2]**o+(0.468860780891)*x[2]
        arg[1,0,2,0]=(-0.267232319593)*x[0]**o+(0.221737693042)*x[0]+(0.202218228468)*x[1]**o+(-0.560941556406)*x[1]+(0.910244160614)*x[2]**o+(0.361584784536)*x[2]
        arg[1,0,2,1]=(0.405254374432)*x[0]**o+(-0.102574362257)*x[0]+(0.527874164745)*x[1]**o+(0.0318395137256)*x[1]+(-0.448050293943)*x[2]**o+(-0.6401160036)*x[2]
        arg[1,1,0,0]=(-0.877224972777)*x[0]**o+(-0.544175806369)*x[0]+(0.528037913494)*x[1]**o+(-0.477854554671)*x[1]+(-0.394875438861)*x[2]**o+(-0.704656012962)*x[2]
        arg[1,1,0,1]=(-0.873928172379)*x[0]**o+(0.273455449944)*x[0]+(0.681090480547)*x[1]**o+(0.883243766093)*x[1]+(0.557357114631)*x[2]**o+(0.586148966774)*x[2]
        arg[1,1,1,0]=(0.552826305913)*x[0]**o+(-0.277269605193)*x[0]+(-0.681872908424)*x[1]**o+(-0.84578699459)*x[1]+(0.6543624385)*x[2]**o+(-0.355051513153)*x[2]
        arg[1,1,1,1]=(-0.71233607947)*x[0]**o+(-0.779861555179)*x[0]+(0.11496674995)*x[1]**o+(0.634901982779)*x[1]+(0.833044137166)*x[2]**o+(0.0443514408009)*x[2]
        arg[1,1,2,0]=(-0.411667274338)*x[0]**o+(0.218709061898)*x[0]+(0.132573928273)*x[1]**o+(-0.450677430962)*x[1]+(0.88420873514)*x[2]**o+(-0.618513305411)*x[2]
        arg[1,1,2,1]=(0.533005373575)*x[0]**o+(0.846343252936)*x[0]+(0.697076801609)*x[1]**o+(0.342260840742)*x[1]+(0.923380136498)*x[2]**o+(-0.727450493778)*x[2]
        arg[1,2,0,0]=(-0.637366803484)*x[0]**o+(-0.654649539562)*x[0]+(-0.443035194144)*x[1]**o+(0.4928304386)*x[1]+(0.889290653473)*x[2]**o+(0.208513517091)*x[2]
        arg[1,2,0,1]=(-0.404870452835)*x[0]**o+(-0.556431486625)*x[0]+(-0.428885935216)*x[1]**o+(0.208771213187)*x[1]+(0.706109349268)*x[2]**o+(0.726168991522)*x[2]
        arg[1,2,1,0]=(0.85574176569)*x[0]**o+(-0.371989246254)*x[0]+(0.0376283091444)*x[1]**o+(-0.164811219077)*x[1]+(0.842513532253)*x[2]**o+(-0.963454790343)*x[2]
        arg[1,2,1,1]=(0.576825533647)*x[0]**o+(0.866570909412)*x[0]+(-0.286228055595)*x[1]**o+(-0.509258889664)*x[1]+(0.358087406384)*x[2]**o+(0.2558870556)*x[2]
        arg[1,2,2,0]=(0.741411326194)*x[0]**o+(0.887444210472)*x[0]+(0.176659455323)*x[1]**o+(-0.292731470152)*x[1]+(0.179836617557)*x[2]**o+(-0.10894424483)*x[2]
        arg[1,2,2,1]=(0.386515069261)*x[0]**o+(-0.423389396097)*x[0]+(0.300384913804)*x[1]**o+(-0.569308908813)*x[1]+(0.845849686169)*x[2]**o+(-0.802483942058)*x[2]
        arg[1,3,0,0]=(0.848952965744)*x[0]**o+(0.113555873929)*x[0]+(0.132168122705)*x[1]**o+(0.464116031047)*x[1]+(-0.928746352905)*x[2]**o+(-0.176374657603)*x[2]
        arg[1,3,0,1]=(0.357870838328)*x[0]**o+(-0.294290586373)*x[0]+(0.395859596543)*x[1]**o+(-0.265635871788)*x[1]+(0.283887401402)*x[2]**o+(0.500378090749)*x[2]
        arg[1,3,1,0]=(-0.509333127149)*x[0]**o+(0.328856435768)*x[0]+(-0.377387784476)*x[1]**o+(-0.878505533565)*x[1]+(0.448760304667)*x[2]**o+(0.996692047048)*x[2]
        arg[1,3,1,1]=(-0.837885550386)*x[0]**o+(0.0608579436757)*x[0]+(-0.627132023609)*x[1]**o+(0.902538475491)*x[1]+(-0.449596037062)*x[2]**o+(-0.158786724549)*x[2]
        arg[1,3,2,0]=(0.205889276236)*x[0]**o+(0.576561726109)*x[0]+(-0.00281354319689)*x[1]**o+(0.214968600999)*x[1]+(0.811758703832)*x[2]**o+(0.260164242506)*x[2]
        arg[1,3,2,1]=(-0.331660060588)*x[0]**o+(0.713879758554)*x[0]+(0.804549392751)*x[1]**o+(-0.346397636822)*x[1]+(-0.364073503513)*x[2]**o+(0.338612348994)*x[2]
        arg[1,4,0,0]=(0.394201404247)*x[0]**o+(-0.416853395415)*x[0]+(-0.0581465874206)*x[1]**o+(0.105267182588)*x[1]+(0.791799718009)*x[2]**o+(0.0840172853234)*x[2]
        arg[1,4,0,1]=(-0.346144268453)*x[0]**o+(-0.661228785728)*x[0]+(0.326588395606)*x[1]**o+(0.766137046896)*x[1]+(0.566377202267)*x[2]**o+(0.613565461425)*x[2]
        arg[1,4,1,0]=(-0.907566140749)*x[0]**o+(-0.391612881392)*x[0]+(0.263322351365)*x[1]**o+(-0.4373853914)*x[1]+(0.248855925063)*x[2]**o+(0.973515668505)*x[2]
        arg[1,4,1,1]=(0.726194147182)*x[0]**o+(0.497593423927)*x[0]+(-0.951344923018)*x[1]**o+(0.55327321528)*x[1]+(-0.266195110749)*x[2]**o+(-0.310893832826)*x[2]
        arg[1,4,2,0]=(-0.940824546553)*x[0]**o+(-0.47838524024)*x[0]+(0.408728941942)*x[1]**o+(0.627300786998)*x[1]+(-0.0547545650101)*x[2]**o+(0.433290388525)*x[2]
        arg[1,4,2,1]=(0.728499572603)*x[0]**o+(-0.232088507059)*x[0]+(0.0550057703801)*x[1]**o+(0.159356057226)*x[1]+(0.111191390097)*x[2]**o+(-0.64498018454)*x[2]
        arg[2,0,0,0]=(0.467025642879)*x[0]**o+(-0.695379603222)*x[0]+(0.591732172092)*x[1]**o+(0.84325648142)*x[1]+(0.965696472916)*x[2]**o+(0.626196181241)*x[2]
        arg[2,0,0,1]=(-0.211129222294)*x[0]**o+(-0.501208172412)*x[0]+(-0.0162146409175)*x[1]**o+(-0.672364205306)*x[1]+(-0.851073479662)*x[2]**o+(-0.817568385627)*x[2]
        arg[2,0,1,0]=(-0.668065295797)*x[0]**o+(-0.90359857855)*x[0]+(0.040519772083)*x[1]**o+(-0.295766307198)*x[1]+(0.0491234828524)*x[2]**o+(-0.589557178985)*x[2]
        arg[2,0,1,1]=(0.928751918298)*x[0]**o+(-0.773787731767)*x[0]+(-0.744325160449)*x[1]**o+(-0.426987905682)*x[1]+(0.503339956851)*x[2]**o+(0.386756648382)*x[2]
        arg[2,0,2,0]=(-0.531818105049)*x[0]**o+(-0.586811490681)*x[0]+(0.780379352762)*x[1]**o+(-0.5575328003)*x[1]+(0.0354867101462)*x[2]**o+(0.581226069105)*x[2]
        arg[2,0,2,1]=(0.0150988200159)*x[0]**o+(0.56415563975)*x[0]+(0.540723773613)*x[1]**o+(0.0653377679637)*x[1]+(-0.52633696371)*x[2]**o+(-0.122811009405)*x[2]
        arg[2,1,0,0]=(-0.0829154151121)*x[0]**o+(-0.416192198737)*x[0]+(0.944232067474)*x[1]**o+(-0.159852455491)*x[1]+(-0.198397797188)*x[2]**o+(-0.684211447305)*x[2]
        arg[2,1,0,1]=(-0.834502772339)*x[0]**o+(-0.856817328109)*x[0]+(0.213303574315)*x[1]**o+(-0.283140791695)*x[1]+(-0.727061494352)*x[2]**o+(0.12871372765)*x[2]
        arg[2,1,1,0]=(0.247439924718)*x[0]**o+(-0.912464825903)*x[0]+(0.791177764168)*x[1]**o+(0.160047921728)*x[1]+(0.591086305304)*x[2]**o+(-0.774025941232)*x[2]
        arg[2,1,1,1]=(-0.120169534728)*x[0]**o+(-0.747254680906)*x[0]+(-0.394003570933)*x[1]**o+(-0.938198960695)*x[1]+(-0.0142776288517)*x[2]**o+(0.0627959740849)*x[2]
        arg[2,1,2,0]=(0.703841647274)*x[0]**o+(-0.946039767174)*x[0]+(-0.836717609846)*x[1]**o+(0.776573515772)*x[1]+(0.250502751027)*x[2]**o+(0.403896552902)*x[2]
        arg[2,1,2,1]=(-0.19760048131)*x[0]**o+(-0.250991189554)*x[0]+(0.278810924634)*x[1]**o+(-0.580148097047)*x[1]+(-0.197518781697)*x[2]**o+(0.116835806412)*x[2]
        arg[2,2,0,0]=(-0.507872937196)*x[0]**o+(-0.771841217512)*x[0]+(0.852666314762)*x[1]**o+(-0.482947444136)*x[1]+(0.805824662389)*x[2]**o+(-0.299255374349)*x[2]
        arg[2,2,0,1]=(-0.481403986338)*x[0]**o+(0.904322169038)*x[0]+(0.937553651477)*x[1]**o+(0.727763542606)*x[1]+(-0.69485539356)*x[2]**o+(0.547755372933)*x[2]
        arg[2,2,1,0]=(0.241675645258)*x[0]**o+(0.957681326492)*x[0]+(-0.305747218513)*x[1]**o+(-0.0393749799535)*x[1]+(-0.448892234759)*x[2]**o+(0.707429367761)*x[2]
        arg[2,2,1,1]=(0.114772627763)*x[0]**o+(-0.411147710203)*x[0]+(0.13389320737)*x[1]**o+(0.806309117051)*x[1]+(-0.572620769665)*x[2]**o+(0.0389925355751)*x[2]
        arg[2,2,2,0]=(0.887023209688)*x[0]**o+(0.158261816385)*x[0]+(0.687399231822)*x[1]**o+(-0.858491519094)*x[1]+(-0.00348150506922)*x[2]**o+(-0.863496570874)*x[2]
        arg[2,2,2,1]=(0.836731324893)*x[0]**o+(0.975472534225)*x[0]+(0.407275572226)*x[1]**o+(0.565132026147)*x[1]+(0.79110912675)*x[2]**o+(-0.0816802575667)*x[2]
        arg[2,3,0,0]=(0.196161020545)*x[0]**o+(0.254301467318)*x[0]+(0.759363127106)*x[1]**o+(0.536320544038)*x[1]+(0.522705219849)*x[2]**o+(0.376531450811)*x[2]
        arg[2,3,0,1]=(-0.862326385376)*x[0]**o+(0.766714365146)*x[0]+(0.279327620722)*x[1]**o+(-0.44416461578)*x[1]+(0.115982232563)*x[2]**o+(0.824064889151)*x[2]
        arg[2,3,1,0]=(0.571263839862)*x[0]**o+(-0.679998793927)*x[0]+(0.63548515376)*x[1]**o+(0.464171242163)*x[1]+(0.712124752567)*x[2]**o+(0.913131755622)*x[2]
        arg[2,3,1,1]=(0.272559456412)*x[0]**o+(0.0577748172306)*x[0]+(-0.610464803438)*x[1]**o+(0.233987628538)*x[1]+(0.904147555826)*x[2]**o+(-0.131245442847)*x[2]
        arg[2,3,2,0]=(0.3409363126)*x[0]**o+(0.40060221588)*x[0]+(-0.603551487822)*x[1]**o+(-0.697430353793)*x[1]+(-0.867009193249)*x[2]**o+(-0.178991284979)*x[2]
        arg[2,3,2,1]=(0.949170673623)*x[0]**o+(-0.0217541208215)*x[0]+(-0.775307373317)*x[1]**o+(0.421565612874)*x[1]+(0.419254478365)*x[2]**o+(-0.732203964657)*x[2]
        arg[2,4,0,0]=(0.876774885889)*x[0]**o+(0.441155536177)*x[0]+(-0.292113312365)*x[1]**o+(0.880590008772)*x[1]+(0.0732363738743)*x[2]**o+(-0.590855067487)*x[2]
        arg[2,4,0,1]=(0.347441329383)*x[0]**o+(-0.96566344519)*x[0]+(-0.0870932691958)*x[1]**o+(-0.793898415692)*x[1]+(0.911757004621)*x[2]**o+(0.623070769788)*x[2]
        arg[2,4,1,0]=(0.732123350436)*x[0]**o+(-0.707615856771)*x[0]+(-0.57969408421)*x[1]**o+(0.0643464469617)*x[1]+(-0.648380091597)*x[2]**o+(0.567100447405)*x[2]
        arg[2,4,1,1]=(-0.128872747083)*x[0]**o+(-0.463336399393)*x[0]+(-0.1679086419)*x[1]**o+(0.602218708561)*x[1]+(0.62788576637)*x[2]**o+(0.223548351914)*x[2]
        arg[2,4,2,0]=(0.619034818105)*x[0]**o+(0.167994824318)*x[0]+(-0.244892750219)*x[1]**o+(0.471338499791)*x[1]+(0.526950200066)*x[2]**o+(0.345611623572)*x[2]
        arg[2,4,2,1]=(0.645979236759)*x[0]**o+(0.79896884484)*x[0]+(-0.587486435644)*x[1]**o+(-0.467838428827)*x[1]+(0.672707985612)*x[2]**o+(-0.169680966884)*x[2]
        arg[3,0,0,0]=(-0.265836015694)*x[0]**o+(-0.383762854285)*x[0]+(0.420375720992)*x[1]**o+(0.739662661302)*x[1]+(-0.593559713248)*x[2]**o+(0.0743208592009)*x[2]
        arg[3,0,0,1]=(0.306246512933)*x[0]**o+(0.0357065903179)*x[0]+(0.213209244461)*x[1]**o+(-0.069836605971)*x[1]+(0.668535341179)*x[2]**o+(0.59870885392)*x[2]
        arg[3,0,1,0]=(0.509420593773)*x[0]**o+(0.995859415429)*x[0]+(-0.07734119959)*x[1]**o+(0.269006992126)*x[1]+(-0.465576902019)*x[2]**o+(0.318766487761)*x[2]
        arg[3,0,1,1]=(-0.609626065482)*x[0]**o+(0.970976299205)*x[0]+(-0.314734168176)*x[1]**o+(-0.536438958228)*x[1]+(0.900096595598)*x[2]**o+(-0.893816431641)*x[2]
        arg[3,0,2,0]=(0.269212290711)*x[0]**o+(-0.119897130167)*x[0]+(-0.937350065698)*x[1]**o+(0.913313769348)*x[1]+(0.311695335026)*x[2]**o+(0.58822688023)*x[2]
        arg[3,0,2,1]=(-0.445116541697)*x[0]**o+(-0.64630078598)*x[0]+(0.811173945666)*x[1]**o+(-0.910503175962)*x[1]+(0.160819266078)*x[2]**o+(-0.807050548571)*x[2]
        arg[3,1,0,0]=(0.25585673506)*x[0]**o+(-0.174293490388)*x[0]+(-0.510750951271)*x[1]**o+(-0.215780338673)*x[1]+(-0.531739311594)*x[2]**o+(0.479569840128)*x[2]
        arg[3,1,0,1]=(-0.159301665223)*x[0]**o+(0.88558738006)*x[0]+(-0.933255168276)*x[1]**o+(0.739621529924)*x[1]+(-0.371501556668)*x[2]**o+(-0.224043643673)*x[2]
        arg[3,1,1,0]=(0.673375237931)*x[0]**o+(0.29315421593)*x[0]+(0.378983835599)*x[1]**o+(-0.187220470447)*x[1]+(-0.122366431969)*x[2]**o+(-0.178182488821)*x[2]
        arg[3,1,1,1]=(0.273992606955)*x[0]**o+(-0.42302685869)*x[0]+(0.490650718142)*x[1]**o+(0.067822229282)*x[1]+(0.758676637186)*x[2]**o+(-0.840062442278)*x[2]
        arg[3,1,2,0]=(-0.533129977534)*x[0]**o+(0.743362501638)*x[0]+(0.0397932337285)*x[1]**o+(0.134261172942)*x[1]+(0.076467475088)*x[2]**o+(0.750124646154)*x[2]
        arg[3,1,2,1]=(0.977824314677)*x[0]**o+(0.322104772509)*x[0]+(0.478366489119)*x[1]**o+(0.139966919282)*x[1]+(0.823631645543)*x[2]**o+(0.29259439301)*x[2]
        arg[3,2,0,0]=(0.969349250234)*x[0]**o+(0.158604629065)*x[0]+(0.517280925739)*x[1]**o+(-0.32151445591)*x[1]+(0.516829269337)*x[2]**o+(0.935139450142)*x[2]
        arg[3,2,0,1]=(-0.673721580725)*x[0]**o+(0.1932869215)*x[0]+(0.47375679317)*x[1]**o+(0.377941367578)*x[1]+(-0.359306451576)*x[2]**o+(0.00498822119998)*x[2]
        arg[3,2,1,0]=(-0.432873699676)*x[0]**o+(0.191782167044)*x[0]+(0.191395777927)*x[1]**o+(-0.540088847941)*x[1]+(0.608451872364)*x[2]**o+(0.151589512705)*x[2]
        arg[3,2,1,1]=(0.633726614581)*x[0]**o+(-0.824133266602)*x[0]+(-0.380289061608)*x[1]**o+(-0.722136850408)*x[1]+(-0.807853184729)*x[2]**o+(0.152552828168)*x[2]
        arg[3,2,2,0]=(-0.182700878265)*x[0]**o+(0.105776705469)*x[0]+(0.837527799957)*x[1]**o+(0.120438395343)*x[1]+(-0.211142277471)*x[2]**o+(-0.0499206664041)*x[2]
        arg[3,2,2,1]=(0.999790796581)*x[0]**o+(0.523238189614)*x[0]+(0.108118557801)*x[1]**o+(-0.899694790145)*x[1]+(0.165089392648)*x[2]**o+(0.38599870475)*x[2]
        arg[3,3,0,0]=(-0.959143897162)*x[0]**o+(-0.968197835338)*x[0]+(-0.521866033091)*x[1]**o+(-0.560799919169)*x[1]+(0.604562365858)*x[2]**o+(0.195593261393)*x[2]
        arg[3,3,0,1]=(0.385513188376)*x[0]**o+(0.918485762628)*x[0]+(0.581126961809)*x[1]**o+(-0.365882492483)*x[1]+(-0.260563773993)*x[2]**o+(0.997016165521)*x[2]
        arg[3,3,1,0]=(0.809269330969)*x[0]**o+(0.472424328796)*x[0]+(-0.920680159926)*x[1]**o+(0.756205962059)*x[1]+(-0.886275742364)*x[2]**o+(-0.800132407145)*x[2]
        arg[3,3,1,1]=(-0.983474390608)*x[0]**o+(-0.436192301851)*x[0]+(0.653210945793)*x[1]**o+(-0.465357783291)*x[1]+(-0.799261076035)*x[2]**o+(-0.349990963152)*x[2]
        arg[3,3,2,0]=(0.710820041642)*x[0]**o+(-0.973768204693)*x[0]+(0.0408403957292)*x[1]**o+(0.61546181161)*x[1]+(0.915939389579)*x[2]**o+(0.634659938301)*x[2]
        arg[3,3,2,1]=(0.633364771689)*x[0]**o+(0.887553239682)*x[0]+(-0.64913373177)*x[1]**o+(0.557465560353)*x[1]+(-0.744072153243)*x[2]**o+(-0.301528624747)*x[2]
        arg[3,4,0,0]=(0.44039719913)*x[0]**o+(-0.0339525806656)*x[0]+(0.574166395893)*x[1]**o+(-0.810441051899)*x[1]+(0.254523432152)*x[2]**o+(-0.798403721953)*x[2]
        arg[3,4,0,1]=(0.253149910502)*x[0]**o+(-0.909846068796)*x[0]+(0.855107984234)*x[1]**o+(-0.973196168972)*x[1]+(0.646832394105)*x[2]**o+(0.811892995874)*x[2]
        arg[3,4,1,0]=(0.13027869782)*x[0]**o+(-0.424094152568)*x[0]+(-0.197473699021)*x[1]**o+(0.0795845331856)*x[1]+(-0.511301398362)*x[2]**o+(-0.331330168767)*x[2]
        arg[3,4,1,1]=(0.476554240918)*x[0]**o+(0.18800431479)*x[0]+(0.259031626216)*x[1]**o+(-0.0532533771273)*x[1]+(0.914847196458)*x[2]**o+(0.333851370508)*x[2]
        arg[3,4,2,0]=(0.800058524963)*x[0]**o+(-0.402028928896)*x[0]+(0.913881556051)*x[1]**o+(0.895516442873)*x[1]+(-0.0227700921994)*x[2]**o+(-0.0502319834692)*x[2]
        arg[3,4,2,1]=(-0.533834741079)*x[0]**o+(0.557954226232)*x[0]+(-0.121676666641)*x[1]**o+(0.781782385789)*x[1]+(0.0682870158781)*x[2]**o+(-0.236432552686)*x[2]
        ref[0,0,0,0]=(1.46703352572)/(o+1.)+(-1.33520045704)+(0.0737627367843)*0.5**o
        ref[0,0,0,1]=(0.213535197837)/(o+1.)+(-0.325019128805)+(0.353968533247)*0.5**o
        ref[0,0,1,0]=(-0.233213543591)/(o+1.)+(0.450075189439)+(0.315704665816)*0.5**o
        ref[0,0,1,1]=(0.809431837761)/(o+1.)+(0.642450689879)+(-0.453912049362)*0.5**o
        ref[0,0,2,0]=(1.8822564746)/(o+1.)+(-0.160827113662)+(0.705225432895)*0.5**o
        ref[0,0,2,1]=(-0.851077895823)/(o+1.)+(-0.032586006392)+(-0.397162258761)*0.5**o
        ref[0,1,0,0]=(-1.75021683351)/(o+1.)+(0.282579703111)+(0.00444300957662)*0.5**o
        ref[0,1,0,1]=(-1.61043513823)/(o+1.)+(0.209579442759)+(0.726535641072)*0.5**o
        ref[0,1,1,0]=(0.550951959927)/(o+1.)+(-0.453625179799)+(-0.554579794515)*0.5**o
        ref[0,1,1,1]=(-0.809122195368)/(o+1.)+(0.609569171964)+(0.294744137189)*0.5**o
        ref[0,1,2,0]=(1.3771144222)/(o+1.)+(-0.703069822248)+(-0.30261020867)*0.5**o
        ref[0,1,2,1]=(1.08629781546)/(o+1.)+(0.0271166180983)+(0.494855964972)*0.5**o
        ref[0,2,0,0]=(-1.26922850072)/(o+1.)+(0.72637380105)+(-0.914585215232)*0.5**o
        ref[0,2,0,1]=(-0.556637167762)/(o+1.)+(-0.036494329576)+(-0.260590397572)*0.5**o
        ref[0,2,1,0]=(0.707349561895)/(o+1.)+(0.208292082952)+(0.0455797513404)*0.5**o
        ref[0,2,1,1]=(-1.47922090453)/(o+1.)+(-0.621962103251)+(-0.300900136798)*0.5**o
        ref[0,2,2,0]=(0.01827281852)/(o+1.)+(-0.159177319679)+(0.745711773093)*0.5**o
        ref[0,2,2,1]=(1.09869785336)/(o+1.)+(0.352958766464)+(0.291340077334)*0.5**o
        ref[0,3,0,0]=(-0.802486910351)/(o+1.)+(0.741067349535)+(0.89242099254)*0.5**o
        ref[0,3,0,1]=(0.187447770797)/(o+1.)+(0.0211789816838)+(0.095390666498)*0.5**o
        ref[0,3,1,0]=(1.0518488657)/(o+1.)+(-0.0260861096315)+(-0.317750910722)*0.5**o
        ref[0,3,1,1]=(1.16078906983)/(o+1.)+(0.0263064747213)+(-0.204455448392)*0.5**o
        ref[0,3,2,0]=(-0.0986989221708)/(o+1.)+(0.182363177629)+(0.38581656579)*0.5**o
        ref[0,3,2,1]=(-0.388150366913)/(o+1.)+(0.557428570571)+(-0.225571343976)*0.5**o
        ref[0,4,0,0]=(-0.623448360222)/(o+1.)+(0.64524557416)+(0.834083057621)*0.5**o
        ref[0,4,0,1]=(-0.639580493935)/(o+1.)+(-0.380689110373)+(-0.613152578007)*0.5**o
        ref[0,4,1,0]=(-0.575892151257)/(o+1.)+(-0.142294010946)+(-0.373668614153)*0.5**o
        ref[0,4,1,1]=(-0.0316225271882)/(o+1.)+(-0.115844822465)+(0.00943802199128)*0.5**o
        ref[0,4,2,0]=(-0.902549384411)/(o+1.)+(0.210416030442)+(0.157099024354)*0.5**o
        ref[0,4,2,1]=(1.2254583864)/(o+1.)+(-0.371769146738)+(0.933762543076)*0.5**o
        ref[1,0,0,0]=(0.0440432191149)/(o+1.)+(-0.303248360069)+(0.0417476479889)*0.5**o
        ref[1,0,0,1]=(-0.196875527758)/(o+1.)+(-0.39869023332)+(0.450986977446)*0.5**o
        ref[1,0,1,0]=(-0.742874955412)/(o+1.)+(-0.677197937069)+(-0.23712795738)*0.5**o
        ref[1,0,1,1]=(0.385960155907)/(o+1.)+(0.0647632575546)+(-0.308721170682)*0.5**o
        ref[1,0,2,0]=(1.11246238908)/(o+1.)+(0.0111904605859)+(-0.267232319593)*0.5**o
        ref[1,0,2,1]=(0.0798238708014)/(o+1.)+(-0.355425426066)+(0.405254374432)*0.5**o
        ref[1,1,0,0]=(0.133162474633)/(o+1.)+(-0.863343187001)+(-0.877224972777)*0.5**o
        ref[1,1,0,1]=(1.23844759518)/(o+1.)+(0.871424091405)+(-0.873928172379)*0.5**o
        ref[1,1,1,0]=(-0.0275104699246)/(o+1.)+(-0.739054056468)+(0.552826305913)*0.5**o
        ref[1,1,1,1]=(0.948010887116)/(o+1.)+(-0.0503040657995)+(-0.71233607947)*0.5**o
        ref[1,1,2,0]=(1.01678266341)/(o+1.)+(-0.425240837237)+(-0.411667274338)*0.5**o
        ref[1,1,2,1]=(1.62045693811)/(o+1.)+(0.23057679995)+(0.533005373575)*0.5**o
        ref[1,2,0,0]=(0.44625545933)/(o+1.)+(0.0233472080645)+(-0.637366803484)*0.5**o
        ref[1,2,0,1]=(0.277223414052)/(o+1.)+(0.189254359042)+(-0.404870452835)*0.5**o
        ref[1,2,1,0]=(0.880141841398)/(o+1.)+(-0.750127627837)+(0.85574176569)*0.5**o
        ref[1,2,1,1]=(0.0718593507895)/(o+1.)+(0.306599537674)+(0.576825533647)*0.5**o
        ref[1,2,2,0]=(0.35649607288)/(o+1.)+(0.242884247745)+(0.741411326194)*0.5**o
        ref[1,2,2,1]=(1.14623459997)/(o+1.)+(-0.897591123484)+(0.386515069261)*0.5**o
        ref[1,3,0,0]=(-0.796578230199)/(o+1.)+(0.200648623687)+(0.848952965744)*0.5**o
        ref[1,3,0,1]=(0.679746997945)/(o+1.)+(-0.0297741837061)+(0.357870838328)*0.5**o
        ref[1,3,1,0]=(0.0713725201916)/(o+1.)+(0.223521474626)+(-0.509333127149)*0.5**o
        ref[1,3,1,1]=(-1.07672806067)/(o+1.)+(0.402304847309)+(-0.837885550386)*0.5**o
        ref[1,3,2,0]=(0.808945160636)/(o+1.)+(0.525847284807)+(0.205889276236)*0.5**o
        ref[1,3,2,1]=(0.440475889238)/(o+1.)+(0.353047235363)+(-0.331660060588)*0.5**o
        ref[1,4,0,0]=(0.733653130588)/(o+1.)+(-0.113784463751)+(0.394201404247)*0.5**o
        ref[1,4,0,1]=(0.892965597874)/(o+1.)+(0.359236861297)+(-0.346144268453)*0.5**o
        ref[1,4,1,0]=(0.512178276429)/(o+1.)+(0.0722586978563)+(-0.907566140749)*0.5**o
        ref[1,4,1,1]=(-1.21754003377)/(o+1.)+(0.369986403191)+(0.726194147182)*0.5**o
        ref[1,4,2,0]=(0.353974376932)/(o+1.)+(0.291102967642)+(-0.940824546553)*0.5**o
        ref[1,4,2,1]=(0.166197160477)/(o+1.)+(-0.358856317187)+(0.728499572603)*0.5**o
        ref[2,0,0,0]=(1.55742864501)/(o+1.)+(0.38703652972)+(0.467025642879)*0.5**o
        ref[2,0,0,1]=(-0.867288120579)/(o+1.)+(-0.995570381672)+(-0.211129222294)*0.5**o
        ref[2,0,1,0]=(0.0896432549354)/(o+1.)+(-0.894461032367)+(-0.668065295797)*0.5**o
        ref[2,0,1,1]=(-0.240985203598)/(o+1.)+(-0.407009494534)+(0.928751918298)*0.5**o
        ref[2,0,2,0]=(0.815866062908)/(o+1.)+(-0.281559110938)+(-0.531818105049)*0.5**o
        ref[2,0,2,1]=(0.0143868099026)/(o+1.)+(0.253341199154)+(0.0150988200159)*0.5**o
        ref[2,1,0,0]=(0.745834270286)/(o+1.)+(-0.630128050767)+(-0.0829154151121)*0.5**o
        ref[2,1,0,1]=(-0.513757920038)/(o+1.)+(-0.505622196077)+(-0.834502772339)*0.5**o
        ref[2,1,1,0]=(1.38226406947)/(o+1.)+(-0.763221422703)+(0.247439924718)*0.5**o
        ref[2,1,1,1]=(-0.408281199784)/(o+1.)+(-0.811328833758)+(-0.120169534728)*0.5**o
        ref[2,1,2,0]=(-0.586214858819)/(o+1.)+(0.11721515075)+(0.703841647274)*0.5**o
        ref[2,1,2,1]=(0.0812921429368)/(o+1.)+(-0.357151740094)+(-0.19760048131)*0.5**o
        ref[2,2,0,0]=(1.65849097715)/(o+1.)+(-0.777022017999)+(-0.507872937196)*0.5**o
        ref[2,2,0,1]=(0.242698257917)/(o+1.)+(1.08992054229)+(-0.481403986338)*0.5**o
        ref[2,2,1,0]=(-0.754639453272)/(o+1.)+(0.81286785715)+(0.241675645258)*0.5**o
        ref[2,2,1,1]=(-0.438727562295)/(o+1.)+(0.217076971211)+(0.114772627763)*0.5**o
        ref[2,2,2,0]=(0.683917726753)/(o+1.)+(-0.781863136791)+(0.887023209688)*0.5**o
        ref[2,2,2,1]=(1.19838469898)/(o+1.)+(0.729462151403)+(0.836731324893)*0.5**o
        ref[2,3,0,0]=(1.28206834695)/(o+1.)+(0.583576731084)+(0.196161020545)*0.5**o
        ref[2,3,0,1]=(0.395309853285)/(o+1.)+(0.573307319258)+(-0.862326385376)*0.5**o
        ref[2,3,1,0]=(1.34760990633)/(o+1.)+(0.348652101929)+(0.571263839862)*0.5**o
        ref[2,3,1,1]=(0.293682752388)/(o+1.)+(0.080258501461)+(0.272559456412)*0.5**o
        ref[2,3,2,0]=(-1.47056068107)/(o+1.)+(-0.237909711446)+(0.3409363126)*0.5**o
        ref[2,3,2,1]=(-0.356052894952)/(o+1.)+(-0.166196236302)+(0.949170673623)*0.5**o
        ref[2,4,0,0]=(-0.21887693849)/(o+1.)+(0.365445238731)+(0.876774885889)*0.5**o
        ref[2,4,0,1]=(0.824663735425)/(o+1.)+(-0.568245545547)+(0.347441329383)*0.5**o
        ref[2,4,1,0]=(-1.22807417581)/(o+1.)+(-0.0380844812022)+(0.732123350436)*0.5**o
        ref[2,4,1,1]=(0.45997712447)/(o+1.)+(0.181215330541)+(-0.128872747083)*0.5**o
        ref[2,4,2,0]=(0.282057449847)/(o+1.)+(0.49247247384)+(0.619034818105)*0.5**o
        ref[2,4,2,1]=(0.0852215499679)/(o+1.)+(0.0807247245641)+(0.645979236759)*0.5**o
        ref[3,0,0,0]=(-0.173183992256)/(o+1.)+(0.215110333109)+(-0.265836015694)*0.5**o
        ref[3,0,0,1]=(0.88174458564)/(o+1.)+(0.282289419133)+(0.306246512933)*0.5**o
        ref[3,0,1,0]=(-0.542918101609)/(o+1.)+(0.791816447658)+(0.509420593773)*0.5**o
        ref[3,0,1,1]=(0.585362427423)/(o+1.)+(-0.229639545331)+(-0.609626065482)*0.5**o
        ref[3,0,2,0]=(-0.625654730672)/(o+1.)+(0.690821759706)+(0.269212290711)*0.5**o
        ref[3,0,2,1]=(0.971993211744)/(o+1.)+(-1.18192725526)+(-0.445116541697)*0.5**o
        ref[3,1,0,0]=(-1.04249026287)/(o+1.)+(0.0447480055334)+(0.25585673506)*0.5**o
        ref[3,1,0,1]=(-1.30475672494)/(o+1.)+(0.700582633156)+(-0.159301665223)*0.5**o
        ref[3,1,1,0]=(0.25661740363)/(o+1.)+(-0.036124371669)+(0.673375237931)*0.5**o
        ref[3,1,1,1]=(1.24932735533)/(o+1.)+(-0.597633535843)+(0.273992606955)*0.5**o
        ref[3,1,2,0]=(0.116260708817)/(o+1.)+(0.813874160367)+(-0.533129977534)*0.5**o
        ref[3,1,2,1]=(1.30199813466)/(o+1.)+(0.377333042401)+(0.977824314677)*0.5**o
        ref[3,2,0,0]=(1.03411019508)/(o+1.)+(0.386114811649)+(0.969349250234)*0.5**o
        ref[3,2,0,1]=(0.114450341594)/(o+1.)+(0.288108255139)+(-0.673721580725)*0.5**o
        ref[3,2,1,0]=(0.799847650292)/(o+1.)+(-0.0983585840958)+(-0.432873699676)*0.5**o
        ref[3,2,1,1]=(-1.18814224634)/(o+1.)+(-0.696858644421)+(0.633726614581)*0.5**o
        ref[3,2,2,0]=(0.626385522485)/(o+1.)+(0.0881472172038)+(-0.182700878265)*0.5**o
        ref[3,2,2,1]=(0.27320795045)/(o+1.)+(0.0047710521094)+(0.999790796581)*0.5**o
        ref[3,3,0,0]=(0.0826963327674)/(o+1.)+(-0.666702246557)+(-0.959143897162)*0.5**o
        ref[3,3,0,1]=(0.320563187816)/(o+1.)+(0.774809717833)+(0.385513188376)*0.5**o
        ref[3,3,1,0]=(-1.80695590229)/(o+1.)+(0.214248941855)+(0.809269330969)*0.5**o
        ref[3,3,1,1]=(-0.146050130242)/(o+1.)+(-0.625770524147)+(-0.983474390608)*0.5**o
        ref[3,3,2,0]=(0.956779785308)/(o+1.)+(0.138176772609)+(0.710820041642)*0.5**o
        ref[3,3,2,1]=(-1.39320588501)/(o+1.)+(0.571745087644)+(0.633364771689)*0.5**o
        ref[3,4,0,0]=(0.828689828045)/(o+1.)+(-0.821398677259)+(0.44039719913)*0.5**o
        ref[3,4,0,1]=(1.50194037834)/(o+1.)+(-0.535574620947)+(0.253149910502)*0.5**o
        ref[3,4,1,0]=(-0.708775097383)/(o+1.)+(-0.337919894074)+(0.13027869782)*0.5**o
        ref[3,4,1,1]=(1.17387882267)/(o+1.)+(0.234301154085)+(0.476554240918)*0.5**o
        ref[3,4,2,0]=(0.891111463851)/(o+1.)+(0.221627765254)+(0.800058524963)*0.5**o
        ref[3,4,2,1]=(-0.0533896507631)/(o+1.)+(0.551652029667)+(-0.533834741079)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.570707549273)*x[0]+(-0.726264032186)*x[1]
        ref=(-0.0423765208873)/(o+1.)+(-0.791542656673)+(0.328490252775)*0.5**o
      else:
        arg=(1.28222892931)*x[0]+(-0.798170675576)*x[1]+(1.5701208405)*x[2]
        ref=(0.721975372868)/(o+1.)+(0.344720905832)+(0.642761909705)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.575166906798)*x[0]+(-1.63698773418)*x[1]
        arg[1]=(-0.0535946584636)*x[0]+(0.105969101775)*x[1]
        ref[0]=(-0.751926426344)/(o+1.)+(-0.395800138888)+(0.481705876734)*0.5**o
        ref[1]=(0.698496252242)/(o+1.)+(-0.265927161866)+(-0.114267485199)*0.5**o
      else:
        arg[0]=(-0.17879254138)*x[0]+(-1.12752123792)*x[1]+(0.921297311605)*x[2]
        arg[1]=(1.33131824618)*x[0]+(-0.392175059176)*x[1]+(0.545754281465)*x[2]
        ref[0]=(-0.494870303061)/(o+1.)+(0.0697636237654)+(-0.0296734121631)*0.5**o
        ref[1]=(-0.394300983704)/(o+1.)+(0.722261312754)+(0.434675826665)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.338658191734)*x[0]+(1.46127609683)*x[1]
        arg[0,1]=(-0.629116867491)*x[0]+(1.40141569066)*x[1]
        arg[0,2]=(0.105569779347)*x[0]+(0.239936324967)*x[1]
        arg[0,3]=(-0.281718331718)*x[0]+(0.280323404866)*x[1]
        arg[0,4]=(0.467341113661)*x[0]+(-0.50452062075)*x[1]
        arg[1,0]=(-0.841295219186)*x[0]+(0.524038163723)*x[1]
        arg[1,1]=(1.40143704236)*x[0]+(-1.50467961637)*x[1]
        arg[1,2]=(1.21078790505)*x[0]+(-0.731453470292)*x[1]
        arg[1,3]=(0.469556922076)*x[0]+(0.0686293204391)*x[1]
        arg[1,4]=(-1.26394637585)*x[0]+(-0.0361395276585)*x[1]
        arg[2,0]=(-1.12850883148)*x[0]+(0.284296673326)*x[1]
        arg[2,1]=(-0.626362828813)*x[0]+(1.19459827573)*x[1]
        arg[2,2]=(-0.545457293083)*x[0]+(1.547337019)*x[1]
        arg[2,3]=(0.664461986629)*x[0]+(-1.69082198266)*x[1]
        arg[2,4]=(-0.100004763305)*x[0]+(0.152159044423)*x[1]
        arg[3,0]=(0.307574843061)*x[0]+(1.20209181301)*x[1]
        arg[3,1]=(-0.0663203356627)*x[0]+(-1.26669863527)*x[1]
        arg[3,2]=(-1.20436798894)*x[0]+(-0.0375693379661)*x[1]
        arg[3,3]=(0.753649087747)*x[0]+(0.69376223486)*x[1]
        arg[3,4]=(-0.0512732264834)*x[0]+(-0.658097606436)*x[1]
        ref[0,0]=(0.996821439952)/(o+1.)+(0.015088160779)+(0.0956201435879)*0.5**o
        ref[0,1]=(0.643092963726)/(o+1.)+(0.239390724844)+(-0.349575590244)*0.5**o
        ref[0,2]=(-0.198468008977)/(o+1.)+(0.00411201743342)+(0.535750078425)*0.5**o
        ref[0,3]=(0.642491201137)/(o+1.)+(-0.623413199652)+(0.602940271315)*0.5**o
        ref[0,4]=(-0.105706120549)/(o+1.)+(-0.353560054939)+(0.775646723338)*0.5**o
        ref[1,0]=(0.144540669221)/(o+1.)+(0.0228268215803)+(-0.507451367845)*0.5**o
        ref[1,1]=(-0.820379339897)/(o+1.)+(0.0951825678509)+(0.526771630185)*0.5**o
        ref[1,2]=(-0.21214585824)/(o+1.)+(-0.0889183062985)+(0.869316905591)*0.5**o
        ref[1,3]=(0.291228940674)/(o+1.)+(0.192623928703)+(-0.138290555565)*0.5**o
        ref[1,4]=(-0.0663718956434)/(o+1.)+(-0.309484335614)+(-0.614745336642)*0.5**o
        ref[2,0]=(0.484892962201)/(o+1.)+(-0.367129859793)+(-0.594845400763)*0.5**o
        ref[2,1]=(0.768457634672)/(o+1.)+(0.154447657562)+(-0.509117502876)*0.5**o
        ref[2,2]=(0.707320338363)/(o+1.)+(-0.0592149232183)+(0.412989233992)*0.5**o
        ref[2,3]=(-0.983684339171)/(o+1.)+(-0.495377645598)+(0.948079634337)*0.5**o
        ref[2,4]=(-0.0315947220465)/(o+1.)+(0.506833986759)+(-0.929918970353)*0.5**o
        ref[3,0]=(0.54141067098)/(o+1.)+(0.489049131493)+(-0.00984227789859)*0.5**o
        ref[3,1]=(-0.552335053051)/(o+1.)+(-0.383325638803)+(-0.0140326402734)*0.5**o
        ref[3,2]=(0.0821472412575)/(o+1.)+(-0.335827719718)+(-0.652429128733)*0.5**o
        ref[3,3]=(0.112917208924)/(o+1.)+(0.667772807338)+(-0.00105150099312)*0.5**o
        ref[3,4]=(-0.19331508157)/(o+1.)+(-0.102296385708)+(-0.311462979932)*0.5**o
      else:
        arg[0,0]=(0.069245963142)*x[0]+(-0.185727379494)*x[1]+(-0.887410889699)*x[2]
        arg[0,1]=(-0.816745668949)*x[0]+(-0.250435680662)*x[1]+(0.701607830582)*x[2]
        arg[0,2]=(-0.862433954308)*x[0]+(0.731874372527)*x[1]+(-0.115384982779)*x[2]
        arg[0,3]=(-0.265247631604)*x[0]+(-0.157059741958)*x[1]+(-0.368419335078)*x[2]
        arg[0,4]=(1.66720203567)*x[0]+(-0.759393708719)*x[1]+(1.36360850704)*x[2]
        arg[1,0]=(-0.192822813795)*x[0]+(0.0358776152527)*x[1]+(-0.722864520788)*x[2]
        arg[1,1]=(-0.480606215471)*x[0]+(1.13682774573)*x[1]+(0.17900467623)*x[2]
        arg[1,2]=(0.346760852584)*x[0]+(0.994833030281)*x[1]+(0.507055933913)*x[2]
        arg[1,3]=(-0.509961853478)*x[0]+(-0.852997575505)*x[1]+(0.586527030253)*x[2]
        arg[1,4]=(0.93957975846)*x[0]+(0.28878335152)*x[1]+(0.128070931738)*x[2]
        arg[2,0]=(0.338283306313)*x[0]+(0.0244500761608)*x[1]+(-0.606875554888)*x[2]
        arg[2,1]=(0.850782624552)*x[0]+(-0.836382038189)*x[1]+(-1.41808459947)*x[2]
        arg[2,2]=(-1.06176229613)*x[0]+(-0.259247046061)*x[1]+(-0.583455484089)*x[2]
        arg[2,3]=(1.17437923282)*x[0]+(1.31485601495)*x[1]+(1.3139743231)*x[2]
        arg[2,4]=(0.362851636275)*x[0]+(0.804049783156)*x[1]+(1.09205000214)*x[2]
        arg[3,0]=(0.343603837195)*x[0]+(-1.49082257981)*x[1]+(-0.628668674545)*x[2]
        arg[3,1]=(-0.676847711765)*x[0]+(1.32350560453)*x[1]+(1.07275185686)*x[2]
        arg[3,2]=(0.570784985469)*x[0]+(-1.01212360602)*x[1]+(-0.0808776667606)*x[2]
        arg[3,3]=(0.506392444236)*x[0]+(-1.00848318041)*x[1]+(0.558777901541)*x[2]
        arg[3,4]=(-0.60925738291)*x[0]+(-0.386755957809)*x[1]+(0.60930115434)*x[2]
        ref[0,0]=(-0.695230136419)/(o+1.)+(-0.388400729083)+(0.468139288533)*0.5**o
        ref[0,1]=(0.462098132187)/(o+1.)+(-0.215041893287)+(-0.397587864642)*0.5**o
        ref[0,2]=(-0.262411385363)/(o+1.)+(0.275729226874)+(-0.534991632946)*0.5**o
        ref[0,3]=(0.117822633378)/(o+1.)+(0.0286827897592)+(-0.965914921536)*0.5**o
        ref[0,4]=(0.293575681039)/(o+1.)+(0.584039692827)+(0.809761767299)*0.5**o
        ref[1,0]=(0.627951922463)/(o+1.)+(-0.994889162557)+(0.482016683321)*0.5**o
        ref[1,1]=(1.83482238662)/(o+1.)+(-0.205719558739)+(-0.588157062647)*0.5**o
        ref[1,2]=(1.13612020398)/(o+1.)+(0.0262824938134)+(0.659964625174)*0.5**o
        ref[1,3]=(0.325264487741)/(o+1.)+(-0.0752154813413)+(-0.951265923787)*0.5**o
        ref[1,4]=(-0.149919283302)/(o+1.)+(0.77701502585)+(-0.0476767266789)*0.5**o
        ref[2,0]=(0.491504686171)/(o+1.)+(-0.562534706926)+(0.389422555265)*0.5**o
        ref[2,1]=(-1.52704217417)/(o+1.)+(0.120525042098)+(-0.117691923134)*0.5**o
        ref[2,2]=(-0.779796772543)/(o+1.)+(-0.318864997267)+(-0.486938059201)*0.5**o
        ref[2,3]=(1.53343318682)/(o+1.)+(0.731149467308)+(0.807477449443)*0.5**o
        ref[2,4]=(1.15919207694)/(o+1.)+(0.113921448815)+(0.871916446996)*0.5**o
        ref[3,0]=(-1.75312920944)/(o+1.)+(-0.153798982308)+(0.284839756895)*0.5**o
        ref[3,1]=(1.30985814569)/(o+1.)+(0.438430874237)+(-0.467310144534)*0.5**o
        ref[3,2]=(-0.737480320972)/(o+1.)+(0.0604403094397)+(0.094383414777)*0.5**o
        ref[3,3]=(-0.688723266524)/(o+1.)+(0.43486027018)+(-0.124310108473)*0.5**o
        ref[3,4]=(0.128399262674)/(o+1.)+(-0.372445844122)+(0.229780239192)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.993339371954)*x[0]+(-1.35308128985)*x[1]
        arg[0,0,1]=(-1.07886149134)*x[0]+(-1.38999283206)*x[1]
        arg[0,1,0]=(0.246468026972)*x[0]+(-1.37217464882)*x[1]
        arg[0,1,1]=(0.645672648599)*x[0]+(1.8070624103)*x[1]
        arg[1,0,0]=(-1.02278989388)*x[0]+(0.151634023103)*x[1]
        arg[1,0,1]=(-0.667332860541)*x[0]+(0.312742149193)*x[1]
        arg[1,1,0]=(-0.848647385832)*x[0]+(0.466438038516)*x[1]
        arg[1,1,1]=(-0.185495911372)*x[0]+(-0.137118786891)*x[1]
        arg[2,0,0]=(-0.0250555870731)*x[0]+(0.905553573283)*x[1]
        arg[2,0,1]=(-1.03940852939)*x[0]+(-1.57326016394)*x[1]
        arg[2,1,0]=(1.48834718811)*x[0]+(-0.340889253246)*x[1]
        arg[2,1,1]=(0.875721514739)*x[0]+(-0.310857412212)*x[1]
        arg[3,0,0]=(1.25129508831)*x[0]+(-1.69312043525)*x[1]
        arg[3,0,1]=(1.50124675312)*x[0]+(0.244195748654)*x[1]
        arg[3,1,0]=(0.292060478909)*x[0]+(-0.660005422696)*x[1]
        arg[3,1,1]=(0.791335231308)*x[0]+(0.466481573591)*x[1]
        arg[4,0,0]=(-0.361241947236)*x[0]+(-1.43591648587)*x[1]
        arg[4,0,1]=(-1.10385456695)*x[0]+(-0.325314195825)*x[1]
        arg[4,1,0]=(1.01004173033)*x[0]+(0.426824552857)*x[1]
        arg[4,1,1]=(-0.631382479612)*x[0]+(0.0603792190966)*x[1]
        arg[5,0,0]=(-0.30404368786)*x[0]+(-0.0102498522932)*x[1]
        arg[5,0,1]=(0.109188848746)*x[0]+(-1.16846160105)*x[1]
        arg[5,1,0]=(-0.351916879547)*x[0]+(-0.320714795448)*x[1]
        arg[5,1,1]=(-0.534138562652)*x[0]+(-0.256013589844)*x[1]
        ref[0,0,0]=(-0.506159375456)/(o+1.)+(-0.522895228783)+(-0.794470828783)*0.5**o
        ref[0,0,1]=(-0.930984367068)/(o+1.)+(-0.636851258685)+(-0.264167438968)*0.5**o
        ref[0,1,0]=(-0.674222792172)/(o+1.)+(-0.52131379048)+(0.591143751282)*0.5**o
        ref[0,1,1]=(0.953370751283)/(o+1.)+(0.377773224645)+(0.743817858326)*0.5**o
        ref[1,0,0]=(0.700508215428)/(o+1.)+(-0.303343530867)+(-0.964977024471)*0.5**o
        ref[1,0,1]=(-0.408185116452)/(o+1.)+(0.0668983569743)+(-0.0802023088445)*0.5**o
        ref[1,1,0]=(0.80742345527)/(o+1.)+(-0.405780705683)+(-0.37807139122)*0.5**o
        ref[1,1,1]=(0.0184615568095)/(o+1.)+(0.221921562266)+(-0.784919379605)*0.5**o
        ref[2,0,0]=(0.571906256883)/(o+1.)+(-0.0415588382133)+(0.391709405753)*0.5**o
        ref[2,0,1]=(-0.778176009451)/(o+1.)+(-0.761680990566)+(-0.311130702752)*0.5**o
        ref[2,1,0]=(0.498214686321)/(o+1.)+(-0.146173705483)+(0.941590659509)*0.5**o
        ref[2,1,1]=(-0.273747071818)/(o+1.)+(-0.0319184747153)+(0.902448123776)*0.5**o
        ref[3,0,0]=(-0.730029471178)/(o+1.)+(-0.288672747933)+(0.865549620109)*0.5**o
        ref[3,0,1]=(-0.371409186475)/(o+1.)+(0.623169375401)+(0.870512937445)*0.5**o
        ref[3,1,0]=(0.0611454391867)/(o+1.)+(-0.627723874214)+(0.826357365454)*0.5**o
        ref[3,1,1]=(0.785553447935)/(o+1.)+(0.246940213405)+(-0.021617069845)*0.5**o
        ref[4,0,0]=(-0.936201955907)/(o+1.)+(-0.484823720021)+(0.108690962842)*0.5**o
        ref[4,0,1]=(-0.277830390815)/(o+1.)+(-0.151786547071)+(-0.847765277819)*0.5**o
        ref[4,1,0]=(0.12293099533)/(o+1.)+(0.295620382054)+(0.722694523749)*0.5**o
        ref[4,1,1]=(-0.5517019185)/(o+1.)+(0.0850748937088)+(-0.189451129433)*0.5**o
        ref[5,0,0]=(-0.942533148805)/(o+1.)+(0.425680504872)+(-0.223121401092)*0.5**o
        ref[5,0,1]=(-0.377134228332)/(o+1.)+(-0.721019409569)+(0.759900295162)*0.5**o
        ref[5,1,0]=(-0.643828391942)/(o+1.)+(0.399974456311)+(-0.828752195674)*0.5**o
        ref[5,1,1]=(-0.118293731439)/(o+1.)+(-0.355568544431)+(0.0392786678054)*0.5**o
      else:
        arg[0,0,0]=(0.205446621454)*x[0]+(0.10254878667)*x[1]+(1.27673728979)*x[2]
        arg[0,0,1]=(0.758846313193)*x[0]+(0.63320336713)*x[1]+(-0.42638505807)*x[2]
        arg[0,1,0]=(-0.0127116561603)*x[0]+(-0.360359420027)*x[1]+(0.540508591493)*x[2]
        arg[0,1,1]=(-0.278918629538)*x[0]+(-0.0291534835772)*x[1]+(0.708655731565)*x[2]
        arg[1,0,0]=(-0.612965210483)*x[0]+(0.787787556912)*x[1]+(0.222384448462)*x[2]
        arg[1,0,1]=(1.07409083742)*x[0]+(1.19429173706)*x[1]+(-0.484979976779)*x[2]
        arg[1,1,0]=(-1.25714999157)*x[0]+(-0.321346019622)*x[1]+(0.278329320837)*x[2]
        arg[1,1,1]=(-0.512262689707)*x[0]+(1.6891992053)*x[1]+(-1.02351830068)*x[2]
        arg[2,0,0]=(-1.57911907773)*x[0]+(0.096124258122)*x[1]+(0.336619253521)*x[2]
        arg[2,0,1]=(0.0901144254199)*x[0]+(-1.36818186255)*x[1]+(1.2969609239)*x[2]
        arg[2,1,0]=(0.158329511499)*x[0]+(0.696997185217)*x[1]+(-0.400535235927)*x[2]
        arg[2,1,1]=(-1.44570853201)*x[0]+(0.237359587195)*x[1]+(-1.17323024801)*x[2]
        arg[3,0,0]=(-1.4325812824)*x[0]+(-0.733085265902)*x[1]+(0.125883186484)*x[2]
        arg[3,0,1]=(-0.0529889021435)*x[0]+(0.335154600942)*x[1]+(0.839337375209)*x[2]
        arg[3,1,0]=(-0.423552182963)*x[0]+(1.31813289315)*x[1]+(0.13235854915)*x[2]
        arg[3,1,1]=(0.53278572217)*x[0]+(0.159434073071)*x[1]+(-0.313621199136)*x[2]
        arg[4,0,0]=(0.747779887673)*x[0]+(0.30977483911)*x[1]+(-0.0760295896739)*x[2]
        arg[4,0,1]=(0.00295118909194)*x[0]+(-0.687914867842)*x[1]+(0.0386268422915)*x[2]
        arg[4,1,0]=(-0.276363532754)*x[0]+(0.52102466785)*x[1]+(-0.684840191917)*x[2]
        arg[4,1,1]=(0.0283138898538)*x[0]+(0.116042458294)*x[1]+(0.485578401528)*x[2]
        arg[5,0,0]=(-0.118402060192)*x[0]+(-1.18409809944)*x[1]+(-0.304048852342)*x[2]
        arg[5,0,1]=(0.174393790186)*x[0]+(-0.524912689263)*x[1]+(-0.648708392355)*x[2]
        arg[5,1,0]=(-0.254958756344)*x[0]+(-0.772071348728)*x[1]+(-0.0572494064027)*x[2]
        arg[5,1,1]=(-0.74777072018)*x[0]+(0.19109939364)*x[1]+(0.498973987814)*x[2]
        ref[0,0,0]=(0.986457444152)/(o+1.)+(0.258994989412)+(0.0802852749346)*0.5**o
        ref[0,0,1]=(-0.0286717143415)/(o+1.)+(0.554867138611)+(-0.115397940627)*0.5**o
        ref[0,1,0]=(-0.886482131138)/(o+1.)+(0.464293129749)+(0.125333386946)*0.5**o
        ref[0,1,1]=(0.199214164457)/(o+1.)+(0.336882900218)+(-0.472396346443)*0.5**o
        ref[1,0,0]=(0.36021680954)/(o+1.)+(0.408903924703)+(-0.780817864055)*0.5**o
        ref[1,0,1]=(0.44795581074)/(o+1.)+(0.472159267742)+(0.391128251477)*0.5**o
        ref[1,1,0]=(1.02665734893)/(o+1.)+(-0.71418625728)+(-0.898451524719)*0.5**o
        ref[1,1,1]=(0.0322676607777)/(o+1.)+(0.105378958385)+(-0.0896073626306)*0.5**o
        ref[2,0,0]=(1.15723247637)/(o+1.)+(-0.719604538702)+(-0.86439896505)*0.5**o
        ref[2,0,1]=(-0.254100181576)/(o+1.)+(-0.0387997401585)+(0.35059314866)*0.5**o
        ref[2,1,0]=(-0.636439167982)/(o+1.)+(0.563390681981)+(-0.0355507351918)*0.5**o
        ref[2,1,1]=(-0.265265225142)/(o+1.)+(-0.635620798949)+(-0.845072369786)*0.5**o
        ref[3,0,0]=(-0.062162800164)/(o+1.)+(-0.554499928282)+(-0.868620705087)*0.5**o
        ref[3,0,1]=(0.784681961813)/(o+1.)+(0.355202655131)+(-0.373584198067)*0.5**o
        ref[3,1,0]=(1.20272043429)/(o+1.)+(0.352677694103)+(-0.881136563162)*0.5**o
        ref[3,1,1]=(-0.311152797156)/(o+1.)+(0.313736563953)+(0.0622782653549)*0.5**o
        ref[4,0,0]=(-0.365517620238)/(o+1.)+(0.733502174253)+(-0.119961591158)*0.5**o
        ref[4,0,1]=(0.0155845896619)/(o+1.)+(-0.654963543839)+(0.648005661558)*0.5**o
        ref[4,1,0]=(0.405697975915)/(o+1.)+(-0.637805133954)+(0.429733235172)*0.5**o
        ref[4,1,1]=(-0.0469738326054)/(o+1.)+(0.0590721074184)+(0.558764367445)*0.5**o
        ref[5,0,0]=(-0.333825007954)/(o+1.)+(-0.46448520882)+(-0.343753586377)*0.5**o
        ref[5,0,1]=(-0.469125954795)/(o+1.)+(-0.147192309255)+(-0.235716718127)*0.5**o
        ref[5,1,0]=(-0.573275891777)/(o+1.)+(-0.478328163121)+(0.445652706545)*0.5**o
        ref[5,1,1]=(0.0545515234726)/(o+1.)+(-0.153919681346)+(0.195590500493)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.226142137728)*x[0]+(0.85681895183)*x[1]
        arg[0,0,0,1]=(-0.824407181762)*x[0]+(-0.290023144892)*x[1]
        arg[0,0,1,0]=(-1.25987816017)*x[0]+(0.824340990595)*x[1]
        arg[0,0,1,1]=(-0.0917811606765)*x[0]+(0.660211932964)*x[1]
        arg[0,0,2,0]=(-1.31073149178)*x[0]+(0.482324984918)*x[1]
        arg[0,0,2,1]=(0.256995458227)*x[0]+(0.79244254852)*x[1]
        arg[0,1,0,0]=(-0.0682147475407)*x[0]+(0.243620052061)*x[1]
        arg[0,1,0,1]=(0.071099982687)*x[0]+(-0.62986754897)*x[1]
        arg[0,1,1,0]=(-0.0716793733131)*x[0]+(-1.15235525814)*x[1]
        arg[0,1,1,1]=(0.193107575216)*x[0]+(-0.596149957543)*x[1]
        arg[0,1,2,0]=(-0.160142629886)*x[0]+(1.36581707067)*x[1]
        arg[0,1,2,1]=(-1.21477648376)*x[0]+(-0.732000298369)*x[1]
        arg[0,2,0,0]=(-1.02468971059)*x[0]+(0.330721595217)*x[1]
        arg[0,2,0,1]=(1.11414907391)*x[0]+(1.29386127812)*x[1]
        arg[0,2,1,0]=(1.1485850154)*x[0]+(0.387585977495)*x[1]
        arg[0,2,1,1]=(0.237382583072)*x[0]+(1.56248221279)*x[1]
        arg[0,2,2,0]=(-0.949129049229)*x[0]+(0.578823379063)*x[1]
        arg[0,2,2,1]=(-1.0667182537)*x[0]+(0.462227485959)*x[1]
        arg[0,3,0,0]=(1.37954377554)*x[0]+(0.541856880139)*x[1]
        arg[0,3,0,1]=(0.401584700319)*x[0]+(0.577102327147)*x[1]
        arg[0,3,1,0]=(-0.289796408108)*x[0]+(0.807853785325)*x[1]
        arg[0,3,1,1]=(0.0038467861822)*x[0]+(-0.931311510771)*x[1]
        arg[0,3,2,0]=(0.859036254821)*x[0]+(0.198884218516)*x[1]
        arg[0,3,2,1]=(-0.00310256070968)*x[0]+(0.511838475968)*x[1]
        arg[0,4,0,0]=(0.539686886802)*x[0]+(-0.324612924267)*x[1]
        arg[0,4,0,1]=(-0.0458655898565)*x[0]+(0.211810380966)*x[1]
        arg[0,4,1,0]=(0.302333533003)*x[0]+(0.0663033316831)*x[1]
        arg[0,4,1,1]=(-0.981519161639)*x[0]+(-0.00207756909914)*x[1]
        arg[0,4,2,0]=(0.283659447004)*x[0]+(-0.231526592979)*x[1]
        arg[0,4,2,1]=(-1.17679167875)*x[0]+(0.285698985756)*x[1]
        arg[1,0,0,0]=(-0.162900059582)*x[0]+(-0.877164340783)*x[1]
        arg[1,0,0,1]=(-0.50468605065)*x[0]+(-0.160440073709)*x[1]
        arg[1,0,1,0]=(-0.100901676644)*x[0]+(-0.296111526029)*x[1]
        arg[1,0,1,1]=(0.176107421869)*x[0]+(1.74393708745)*x[1]
        arg[1,0,2,0]=(-0.979807721703)*x[0]+(-0.229400841067)*x[1]
        arg[1,0,2,1]=(0.217472795645)*x[0]+(0.0881503243323)*x[1]
        arg[1,1,0,0]=(-0.127980123625)*x[0]+(0.0932095456541)*x[1]
        arg[1,1,0,1]=(0.829674410931)*x[0]+(-0.796196041601)*x[1]
        arg[1,1,1,0]=(1.28656110233)*x[0]+(-1.27850471392)*x[1]
        arg[1,1,1,1]=(1.159269884)*x[0]+(-0.973135433631)*x[1]
        arg[1,1,2,0]=(1.46274837856)*x[0]+(-1.81554270863)*x[1]
        arg[1,1,2,1]=(0.0973472511688)*x[0]+(0.551451261495)*x[1]
        arg[1,2,0,0]=(0.510667671293)*x[0]+(-0.51075982099)*x[1]
        arg[1,2,0,1]=(-1.41612093451)*x[0]+(0.614417669629)*x[1]
        arg[1,2,1,0]=(1.46595125796)*x[0]+(0.60374279122)*x[1]
        arg[1,2,1,1]=(-0.317185218187)*x[0]+(-0.0785532882696)*x[1]
        arg[1,2,2,0]=(-0.388504041993)*x[0]+(-0.838003398775)*x[1]
        arg[1,2,2,1]=(0.0497207129012)*x[0]+(-0.622155692247)*x[1]
        arg[1,3,0,0]=(-0.0471400179563)*x[0]+(1.3378948207)*x[1]
        arg[1,3,0,1]=(-1.22250822195)*x[0]+(0.629714783138)*x[1]
        arg[1,3,1,0]=(-1.28801080408)*x[0]+(-1.54079710038)*x[1]
        arg[1,3,1,1]=(-0.426412448469)*x[0]+(-1.09410474021)*x[1]
        arg[1,3,2,0]=(0.022834167351)*x[0]+(-0.497423389788)*x[1]
        arg[1,3,2,1]=(-0.787492124301)*x[0]+(-0.245621168419)*x[1]
        arg[1,4,0,0]=(-1.37814880919)*x[0]+(-0.704805848451)*x[1]
        arg[1,4,0,1]=(-1.12494770055)*x[0]+(1.44920024305)*x[1]
        arg[1,4,1,0]=(-0.809532572761)*x[0]+(1.16638587657)*x[1]
        arg[1,4,1,1]=(0.931784486816)*x[0]+(1.34629889453)*x[1]
        arg[1,4,2,0]=(-0.172430942793)*x[0]+(-0.584736998321)*x[1]
        arg[1,4,2,1]=(0.00542948736389)*x[0]+(0.43917597284)*x[1]
        arg[2,0,0,0]=(-0.946122337406)*x[0]+(-0.161940168137)*x[1]
        arg[2,0,0,1]=(-0.7856653779)*x[0]+(-0.75287701788)*x[1]
        arg[2,0,1,0]=(-1.08858367592)*x[0]+(-0.530380818694)*x[1]
        arg[2,0,1,1]=(-1.0573555095)*x[0]+(-0.854148765785)*x[1]
        arg[2,0,2,0]=(0.723412825328)*x[0]+(0.326761950437)*x[1]
        arg[2,0,2,1]=(-0.0143142161225)*x[0]+(-0.752724311876)*x[1]
        arg[2,1,0,0]=(-0.942548092355)*x[0]+(0.700861094897)*x[1]
        arg[2,1,0,1]=(-0.712286567957)*x[0]+(0.230406545738)*x[1]
        arg[2,1,1,0]=(1.27414839774)*x[0]+(0.724784649037)*x[1]
        arg[2,1,1,1]=(0.637494163002)*x[0]+(0.442489599192)*x[1]
        arg[2,1,2,0]=(1.81617631726)*x[0]+(1.56870218809)*x[1]
        arg[2,1,2,1]=(1.16419900666)*x[0]+(-0.044314729141)*x[1]
        arg[2,2,0,0]=(0.880267125196)*x[0]+(0.77124406259)*x[1]
        arg[2,2,0,1]=(0.482402867542)*x[0]+(0.0746924305017)*x[1]
        arg[2,2,1,0]=(0.105861259696)*x[0]+(0.484292207267)*x[1]
        arg[2,2,1,1]=(-0.107785838196)*x[0]+(0.701132243209)*x[1]
        arg[2,2,2,0]=(0.559717405844)*x[0]+(-1.5585942717)*x[1]
        arg[2,2,2,1]=(-0.989130712605)*x[0]+(0.221361137556)*x[1]
        arg[2,3,0,0]=(-0.045471818095)*x[0]+(-0.437941959792)*x[1]
        arg[2,3,0,1]=(1.25840468379)*x[0]+(-1.18872011057)*x[1]
        arg[2,3,1,0]=(1.70665829408)*x[0]+(0.705273921003)*x[1]
        arg[2,3,1,1]=(-0.0928709253845)*x[0]+(0.785115449756)*x[1]
        arg[2,3,2,0]=(-0.887243418838)*x[0]+(1.06790660642)*x[1]
        arg[2,3,2,1]=(-0.572644686872)*x[0]+(-0.672831948394)*x[1]
        arg[2,4,0,0]=(1.28282566801)*x[0]+(0.651985199447)*x[1]
        arg[2,4,0,1]=(0.0965663229335)*x[0]+(1.33819998052)*x[1]
        arg[2,4,1,0]=(0.205569041098)*x[0]+(-0.873095358733)*x[1]
        arg[2,4,1,1]=(-1.04452025745)*x[0]+(-0.204917033907)*x[1]
        arg[2,4,2,0]=(-0.181163938972)*x[0]+(0.0341993409272)*x[1]
        arg[2,4,2,1]=(1.68327027059)*x[0]+(-0.289436355642)*x[1]
        arg[3,0,0,0]=(1.05712810731)*x[0]+(0.915711819735)*x[1]
        arg[3,0,0,1]=(-0.171072039482)*x[0]+(-0.160656943623)*x[1]
        arg[3,0,1,0]=(-1.27063799869)*x[0]+(-0.605131642813)*x[1]
        arg[3,0,1,1]=(0.940512990835)*x[0]+(-0.116675577531)*x[1]
        arg[3,0,2,0]=(-0.194213869952)*x[0]+(0.364066752025)*x[1]
        arg[3,0,2,1]=(-0.328780465397)*x[0]+(-0.114689413673)*x[1]
        arg[3,1,0,0]=(-0.825914305895)*x[0]+(-0.167010056731)*x[1]
        arg[3,1,0,1]=(0.657779056207)*x[0]+(0.0666252451638)*x[1]
        arg[3,1,1,0]=(-0.464004426036)*x[0]+(-0.340932918409)*x[1]
        arg[3,1,1,1]=(-1.6698652667)*x[0]+(0.731870373141)*x[1]
        arg[3,1,2,0]=(1.39626250779)*x[0]+(-0.879152855557)*x[1]
        arg[3,1,2,1]=(1.34973919508)*x[0]+(-0.129753332761)*x[1]
        arg[3,2,0,0]=(1.03443450706)*x[0]+(0.807333477811)*x[1]
        arg[3,2,0,1]=(0.366011302596)*x[0]+(0.543804800926)*x[1]
        arg[3,2,1,0]=(0.262446846036)*x[0]+(-0.247786850991)*x[1]
        arg[3,2,1,1]=(-1.04657091196)*x[0]+(-0.485443426155)*x[1]
        arg[3,2,2,0]=(0.342434327538)*x[0]+(-0.148503585588)*x[1]
        arg[3,2,2,1]=(-0.189833429716)*x[0]+(0.424391874819)*x[1]
        arg[3,3,0,0]=(-0.118348838118)*x[0]+(0.397628264341)*x[1]
        arg[3,3,0,1]=(0.481440510084)*x[0]+(-0.393576221424)*x[1]
        arg[3,3,1,0]=(0.907533552208)*x[0]+(-0.668646549376)*x[1]
        arg[3,3,1,1]=(-1.49731714213)*x[0]+(-1.18710678264)*x[1]
        arg[3,3,2,0]=(-0.460660079207)*x[0]+(1.48078266444)*x[1]
        arg[3,3,2,1]=(-0.161073556988)*x[0]+(-0.486827042754)*x[1]
        arg[3,4,0,0]=(-0.668755455623)*x[0]+(0.345804059434)*x[1]
        arg[3,4,0,1]=(1.00528889193)*x[0]+(0.978507928191)*x[1]
        arg[3,4,1,0]=(0.98955856331)*x[0]+(1.0421852144)*x[1]
        arg[3,4,1,1]=(0.711688260873)*x[0]+(1.39962222322)*x[1]
        arg[3,4,2,0]=(1.08512429209)*x[0]+(-0.555249097743)*x[1]
        arg[3,4,2,1]=(-1.10443784007)*x[0]+(-0.782274429169)*x[1]
        ref[0,0,0,0]=(0.151153165564)/(o+1.)+(0.112609656739)+(0.254304335059)*0.5**o
        ref[0,0,0,1]=(-0.530602583511)/(o+1.)+(0.1817855378)+(-0.947398818743)*0.5**o
        ref[0,0,1,0]=(0.308747437738)/(o+1.)+(-0.0338129255643)+(-0.676658756184)*0.5**o
        ref[0,0,1,1]=(0.321026553427)/(o+1.)+(0.0692705367618)+(0.108863145337)*0.5**o
        ref[0,0,2,0]=(-0.0908034666655)/(o+1.)+(-0.1100205155)+(-0.517562009199)*0.5**o
        ref[0,0,2,1]=(0.303456040903)/(o+1.)+(0.500090875336)+(-0.254199784829)*0.5**o
        ref[0,1,0,0]=(-0.440641221831)/(o+1.)+(0.456218946873)+(-0.296391367396)*0.5**o
        ref[0,1,0,1]=(0.222628561762)/(o+1.)+(-0.162064412854)+(-0.457267302337)*0.5**o
        ref[0,1,1,0]=(-0.913361887408)/(o+1.)+(-0.356168674239)+(0.401664604431)*0.5**o
        ref[0,1,1,1]=(-0.543116639248)/(o+1.)+(-0.120902439422)+(0.381879135767)*0.5**o
        ref[0,1,2,0]=(0.993649262831)/(o+1.)+(0.384695858206)+(-0.55736653846)*0.5**o
        ref[0,1,2,1]=(-0.968077797451)/(o+1.)+(-0.187765942637)+(-0.603167099406)*0.5**o
        ref[0,2,0,0]=(0.873662019918)/(o+1.)+(-0.710635360033)+(-0.146359415223)*0.5**o
        ref[0,2,0,1]=(0.466408230693)/(o+1.)+(0.77626565285)+(0.389070815631)*0.5**o
        ref[0,2,1,0]=(-0.475010813579)/(o+1.)+(0.525435880006)+(0.960310046461)*0.5**o
        ref[0,2,1,1]=(0.697127520484)/(o+1.)+(0.357204276552)+(0.388328722275)*0.5**o
        ref[0,2,2,0]=(-0.108434738487)/(o+1.)+(0.356443300466)+(-0.97475753261)*0.5**o
        ref[0,2,2,1]=(0.952274964304)/(o+1.)+(-0.497379508976)+(-0.562006714089)*0.5**o
        ref[0,3,0,0]=(-0.0368012761671)/(o+1.)+(0.594000625518)+(0.770200680808)*0.5**o
        ref[0,3,0,1]=(0.450361861836)/(o+1.)+(0.182031052848)+(0.164263059934)*0.5**o
        ref[0,3,1,0]=(0.500714741956)/(o+1.)+(0.0479711343168)+(-0.0785996333735)*0.5**o
        ref[0,3,1,1]=(-0.949280938645)/(o+1.)+(0.452441874504)+(-0.88306753495)*0.5**o
        ref[0,3,2,0]=(0.186931712837)/(o+1.)+(0.222335175903)+(0.426318408694)*0.5**o
        ref[0,3,2,1]=(-0.288015428918)/(o+1.)+(0.80948657963)+(-0.822221815085)*0.5**o
        ref[0,4,0,0]=(0.670151184695)/(o+1.)+(-0.415883847013)+(0.376690471866)*0.5**o
        ref[0,4,0,1]=(-0.107106843992)/(o+1.)+(0.555375298435)+(-0.837698961768)*0.5**o
        ref[0,4,1,0]=(-0.299999096802)/(o+1.)+(0.668996782774)+(-0.669357604061)*0.5**o
        ref[0,4,1,1]=(0.778006183401)/(o+1.)+(-0.614473529032)+(-0.532655856076)*0.5**o
        ref[0,4,2,0]=(-0.250068389818)/(o+1.)+(0.156274314197)+(-0.0103473845504)*0.5**o
        ref[0,4,2,1]=(0.902125322216)/(o+1.)+(-0.616424567683)+(-0.560368879846)*0.5**o
        ref[1,0,0,0]=(-0.133690258295)/(o+1.)+(-0.715949360249)+(0.525524578427)*0.5**o
        ref[1,0,0,1]=(-0.788097689216)/(o+1.)+(0.0998405659256)+(-0.076709566994)*0.5**o
        ref[1,0,1,0]=(0.144280886187)/(o+1.)+(0.0775377290314)+(-0.696369546924)*0.5**o
        ref[1,0,1,1]=(0.824172196605)/(o+1.)+(0.561804029444)+(-0.0277357461773)*0.5**o
        ref[1,0,2,0]=(0.422564421802)/(o+1.)+(-0.60719307099)+(-0.417386842591)*0.5**o
        ref[1,0,2,1]=(0.778908672762)/(o+1.)+(0.0163446340765)+(-0.505974820938)*0.5**o
        ref[1,1,0,0]=(0.0574071047387)/(o+1.)+(-0.350514214742)+(0.608850746775)*0.5**o
        ref[1,1,0,1]=(-0.991660961792)/(o+1.)+(0.322084279455)+(0.380970772212)*0.5**o
        ref[1,1,1,0]=(-0.823761120268)/(o+1.)+(-0.0495419092898)+(0.930901327253)*0.5**o
        ref[1,1,1,1]=(-0.384729980569)/(o+1.)+(-0.0392583789163)+(0.649381188769)*0.5**o
        ref[1,1,2,0]=(-0.840365304069)/(o+1.)+(-0.0150512662777)+(0.517673506558)*0.5**o
        ref[1,1,2,1]=(0.555976660846)/(o+1.)+(-0.307265886898)+(0.707353625614)*0.5**o
        ref[1,2,0,0]=(-0.423056705491)/(o+1.)+(0.0559404393793)+(0.311083677036)*0.5**o
        ref[1,2,0,1]=(0.15409797664)/(o+1.)+(-0.0906637067519)+(-0.774473828018)*0.5**o
        ref[1,2,1,0]=(0.796718644088)/(o+1.)+(0.248012077098)+(0.776951250898)*0.5**o
        ref[1,2,1,1]=(-0.342830889263)/(o+1.)+(-0.282162737841)+(0.511417858487)*0.5**o
        ref[1,2,2,0]=(-0.722403016671)/(o+1.)+(-0.367059457954)+(0.230014491812)*0.5**o
        ref[1,2,2,1]=(-0.773685058881)/(o+1.)+(0.412807900451)+(-0.624365721366)*0.5**o
        ref[1,3,0,0]=(0.527481434758)/(o+1.)+(0.36602961117)+(0.031214145649)*0.5**o
        ref[1,3,0,1]=(0.0630927531585)/(o+1.)+(-0.164989701995)+(-0.325906787975)*0.5**o
        ref[1,3,1,0]=(-0.612390590593)/(o+1.)+(-0.78855996274)+(-0.63929738839)*0.5**o
        ref[1,3,1,1]=(-0.962656044265)/(o+1.)+(-0.00576535037253)+(-0.546330443673)*0.5**o
        ref[1,3,2,0]=(-0.947280922008)/(o+1.)+(-0.137449579996)+(0.747590859563)*0.5**o
        ref[1,3,2,1]=(-0.946623026202)/(o+1.)+(0.0787536550213)+(-0.24399757656)*0.5**o
        ref[1,4,0,0]=(-0.174471016943)/(o+1.)+(-0.636500556039)+(-0.635482528623)*0.5**o
        ref[1,4,0,1]=(0.806379391595)/(o+1.)+(-0.00304106577093)+(-0.476044717545)*0.5**o
        ref[1,4,1,0]=(0.729649125134)/(o+1.)+(0.0151477051616)+(-0.403091231646)*0.5**o
        ref[1,4,1,1]=(0.559976454745)/(o+1.)+(0.479295896421)+(0.75951513376)*0.5**o
        ref[1,4,2,0]=(0.272199270818)/(o+1.)+(-0.0929072274332)+(-0.843552757065)*0.5**o
        ref[1,4,2,1]=(-0.552663678377)/(o+1.)+(0.613784745778)+(-0.230300352975)*0.5**o
        ref[2,0,0,0]=(-0.174247908098)/(o+1.)+(-0.108873783282)+(-0.716067030882)*0.5**o
        ref[2,0,0,1]=(-0.300584841532)/(o+1.)+(-0.485363665493)+(-0.267230223262)*0.5**o
        ref[2,0,1,0]=(0.316261805154)/(o+1.)+(-0.523474106451)+(-0.888278086869)*0.5**o
        ref[2,0,1,1]=(-0.699622562861)/(o+1.)+(-0.151089538248)+(-0.909702635929)*0.5**o
        ref[2,0,2,0]=(0.728545639829)/(o+1.)+(0.192083148877)+(-0.0625371618174)*0.5**o
        ref[2,0,2,1]=(-0.22042267239)/(o+1.)+(-0.611974599324)+(0.677333343041)*0.5**o
        ref[2,1,0,0]=(-0.0734883811113)/(o+1.)+(0.36100831022)+(-0.890215236787)*0.5**o
        ref[2,1,0,1]=(0.121656829057)/(o+1.)+(-0.130801635799)+(-0.341933579678)*0.5**o
        ref[2,1,1,0]=(0.221353168629)/(o+1.)+(0.618635673483)+(0.540308531179)*0.5**o
        ref[2,1,1,1]=(0.0152580531273)/(o+1.)+(0.252173415542)+(0.560378877984)*0.5**o
        ref[2,1,2,0]=(0.580319777731)/(o+1.)+(0.90566612137)+(0.993226484879)*0.5**o
        ref[2,1,2,1]=(-0.710250320949)/(o+1.)+(0.742177021871)+(0.345780554724)*0.5**o
        ref[2,2,0,0]=(0.770701563507)/(o+1.)+(0.240823471881)+(0.399162680517)*0.5**o
        ref[2,2,0,1]=(0.76121761761)/(o+1.)+(0.128382656738)+(-0.460887633042)*0.5**o
        ref[2,2,1,0]=(-0.269854962057)/(o+1.)+(0.690854898294)+(-0.521701367567)*0.5**o
        ref[2,2,1,1]=(0.719518595712)/(o+1.)+(0.029596430244)+(-0.185365051187)*0.5**o
        ref[2,2,2,0]=(-0.671157761165)/(o+1.)+(-0.142688285012)+(-0.0423425346644)*0.5**o
        ref[2,2,2,1]=(-0.541156097112)/(o+1.)+(0.0765180918929)+(-0.379649661722)*0.5**o
        ref[2,3,0,0]=(-0.0604347109212)/(o+1.)+(-0.19403702073)+(-0.034905025507)*0.5**o
        ref[2,3,0,1]=(-0.793398507071)/(o+1.)+(0.255681280626)+(0.351720519037)*0.5**o
        ref[2,3,1,0]=(-0.267058413731)/(o+1.)+(0.852211426049)+(0.974567776717)*0.5**o
        ref[2,3,1,1]=(0.228396641207)/(o+1.)+(0.322922896735)+(-0.181997910305)*0.5**o
        ref[2,3,2,0]=(0.480987773937)/(o+1.)+(-0.197709028569)+(0.0950934707835)*0.5**o
        ref[2,3,2,1]=(-0.648934475027)/(o+1.)+(-0.418383297875)+(0.240224435511)*0.5**o
        ref[2,4,0,0]=(0.430020261528)/(o+1.)+(0.53421018733)+(0.436370231269)*0.5**o
        ref[2,4,0,1]=(0.906390660223)/(o+1.)+(-0.056057545856)+(0.640490734948)*0.5**o
        ref[2,4,1,0]=(-0.732099770943)/(o+1.)+(-0.0485027617732)+(0.161578976854)*0.5**o
        ref[2,4,1,1]=(-0.827495680848)/(o+1.)+(0.285044728936)+(-0.992031068384)*0.5**o
        ref[2,4,2,0]=(-0.919905471787)/(o+1.)+(0.656971386264)+(-0.541001898785)*0.5**o
        ref[2,4,2,1]=(0.161975875302)/(o+1.)+(0.272665206557)+(0.686527626533)*0.5**o
        ref[3,0,0,0]=(0.584120963259)/(o+1.)+(0.207930729804)+(0.972857504177)*0.5**o
        ref[3,0,0,1]=(0.315823141491)/(o+1.)+(-0.217561208386)+(-0.212429707823)*0.5**o
        ref[3,0,1,0]=(-0.732294996139)/(o+1.)+(-0.278653143596)+(-0.586168358175)*0.5**o
        ref[3,0,1,1]=(0.481321847225)/(o+1.)+(-0.070080648792)+(0.482676863662)*0.5**o
        ref[3,0,2,0]=(0.616679859076)/(o+1.)+(-0.418784546677)+(0.390742116352)*0.5**o
        ref[3,0,2,1]=(0.866778417448)/(o+1.)+(-0.686585656643)+(0.0629230167692)*0.5**o
        ref[3,1,0,0]=(0.531890054167)/(o+1.)+(-0.516187607633)+(-0.492439201527)*0.5**o
        ref[3,1,0,1]=(-0.512361639413)/(o+1.)+(0.499557182545)+(0.237651575694)*0.5**o
        ref[3,1,1,0]=(-0.67542965405)/(o+1.)+(0.274414249559)+(-0.678336189514)*0.5**o
        ref[3,1,1,1]=(0.371936242507)/(o+1.)+(-0.1636482466)+(-0.982634642862)*0.5**o
        ref[3,1,2,0]=(-0.799408831501)/(o+1.)+(0.458633399549)+(0.399251684639)*0.5**o
        ref[3,1,2,1]=(-0.0145253903686)/(o+1.)+(0.238472345905)+(0.757566560882)*0.5**o
        ref[3,2,0,0]=(-0.130519473112)/(o+1.)+(0.695277604837)+(0.581732248306)*0.5**o
        ref[3,2,0,1]=(0.632010536438)/(o+1.)+(0.123066205529)+(0.0316731560259)*0.5**o
        ref[3,2,1,0]=(-0.709926210356)/(o+1.)+(0.0926699972302)+(0.53924621094)*0.5**o
        ref[3,2,1,1]=(0.428094002247)/(o+1.)+(-0.841997926131)+(-0.276112488096)*0.5**o
        ref[3,2,2,0]=(0.708481148086)/(o+1.)+(-0.50498161534)+(0.495412824545)*0.5**o
        ref[3,2,2,1]=(0.782824622703)/(o+1.)+(-0.151509698687)+(-0.245246780227)*0.5**o
        ref[3,3,0,0]=(0.848460868418)/(o+1.)+(-0.104639248461)+(-0.359902945273)*0.5**o
        ref[3,3,0,1]=(0.0569151307298)/(o+1.)+(-0.0674224535469)+(0.165794065024)*0.5**o
        ref[3,3,1,0]=(-0.165683739908)/(o+1.)+(-0.183388699236)+(0.771348141212)*0.5**o
        ref[3,3,1,1]=(-0.932550058069)/(o+1.)+(-0.47828109001)+(-0.795311686688)*0.5**o
        ref[3,3,2,0]=(0.821542940923)/(o+1.)+(0.166853194487)+(-0.13512674466)*0.5**o
        ref[3,3,2,1]=(-0.46375871628)/(o+1.)+(0.132701267732)+(-0.449544418926)*0.5**o
        ref[3,4,0,0]=(-0.074530542168)/(o+1.)+(-0.0567877963797)+(-0.134845261261)*0.5**o
        ref[3,4,0,1]=(0.238671520479)/(o+1.)+(0.50689253572)+(0.731340228204)*0.5**o
        ref[3,4,1,0]=(0.981601513335)/(o+1.)+(0.400717209822)+(0.248707844726)*0.5**o
        ref[3,4,1,1]=(0.592304390107)/(o+1.)+(0.342408637328)+(0.834188819333)*0.5**o
        ref[3,4,2,0]=(-0.339991330909)/(o+1.)+(-0.0145922376156)+(0.899051000485)*0.5**o
        ref[3,4,2,1]=(-0.626610688164)/(o+1.)+(-0.417008165115)+(-0.426085250846)*0.5**o
      else:
        arg[0,0,0,0]=(0.697666818785)*x[0]+(1.00202295791)*x[1]+(0.617700157184)*x[2]
        arg[0,0,0,1]=(-0.465577819457)*x[0]+(-1.40054693355)*x[1]+(1.16912653295)*x[2]
        arg[0,0,1,0]=(-1.26696500634)*x[0]+(-0.0469756169975)*x[1]+(-0.445591738991)*x[2]
        arg[0,0,1,1]=(-1.61654321709)*x[0]+(-1.69522962467)*x[1]+(-1.06089094851)*x[2]
        arg[0,0,2,0]=(0.229468378506)*x[0]+(0.716813596991)*x[1]+(-0.237703937029)*x[2]
        arg[0,0,2,1]=(-1.12052051905)*x[0]+(0.835000004618)*x[1]+(-0.853004812404)*x[2]
        arg[0,1,0,0]=(-1.75500745705)*x[0]+(0.655055836882)*x[1]+(1.05207750735)*x[2]
        arg[0,1,0,1]=(-0.0746024824253)*x[0]+(-0.421596816958)*x[1]+(-0.375715855989)*x[2]
        arg[0,1,1,0]=(-0.963269684396)*x[0]+(0.431993248836)*x[1]+(-1.63061928649)*x[2]
        arg[0,1,1,1]=(-0.418757265018)*x[0]+(-0.665626168583)*x[1]+(-0.335734432507)*x[2]
        arg[0,1,2,0]=(0.980346404307)*x[0]+(1.45442049802)*x[1]+(-0.833420308612)*x[2]
        arg[0,1,2,1]=(-1.52536017847)*x[0]+(0.99484028681)*x[1]+(0.109754321663)*x[2]
        arg[0,2,0,0]=(1.697179262)*x[0]+(-0.928732802147)*x[1]+(-0.653379513626)*x[2]
        arg[0,2,0,1]=(0.682698902948)*x[0]+(0.829307959583)*x[1]+(-1.0300112276)*x[2]
        arg[0,2,1,0]=(0.026084323266)*x[0]+(1.28865784608)*x[1]+(0.830446580645)*x[2]
        arg[0,2,1,1]=(0.486291506032)*x[0]+(-0.519637742913)*x[1]+(0.81671330629)*x[2]
        arg[0,2,2,0]=(-0.761591624634)*x[0]+(0.370951131011)*x[1]+(-0.256748479312)*x[2]
        arg[0,2,2,1]=(-0.249974326513)*x[0]+(0.267323484635)*x[1]+(-1.09642130905)*x[2]
        arg[0,3,0,0]=(0.752057563255)*x[0]+(-1.2772262175)*x[1]+(-0.264569668983)*x[2]
        arg[0,3,0,1]=(-0.173167796447)*x[0]+(0.31873208356)*x[1]+(-0.599780701112)*x[2]
        arg[0,3,1,0]=(-0.912464014941)*x[0]+(-1.56452517166)*x[1]+(1.37651660499)*x[2]
        arg[0,3,1,1]=(0.302009667854)*x[0]+(0.524213064769)*x[1]+(0.316301522683)*x[2]
        arg[0,3,2,0]=(-1.2940154673)*x[0]+(-0.900061396992)*x[1]+(0.3860029847)*x[2]
        arg[0,3,2,1]=(-0.97354131341)*x[0]+(1.08221883964)*x[1]+(-0.739323076416)*x[2]
        arg[0,4,0,0]=(1.86535317192)*x[0]+(-0.478359589546)*x[1]+(-0.798014592302)*x[2]
        arg[0,4,0,1]=(-1.08518693851)*x[0]+(-0.186048654536)*x[1]+(1.04915573267)*x[2]
        arg[0,4,1,0]=(-1.09692811055)*x[0]+(1.01042920935)*x[1]+(-0.73978089555)*x[2]
        arg[0,4,1,1]=(0.654656964663)*x[0]+(0.279204596126)*x[1]+(0.886217781361)*x[2]
        arg[0,4,2,0]=(-0.54009886291)*x[0]+(0.378037248739)*x[1]+(1.09785568886)*x[2]
        arg[0,4,2,1]=(-1.43995958913)*x[0]+(-0.720157090994)*x[1]+(0.0676362028136)*x[2]
        arg[1,0,0,0]=(1.47735412608)*x[0]+(0.342355546228)*x[1]+(-0.0134502415508)*x[2]
        arg[1,0,0,1]=(-0.419536590369)*x[0]+(-1.61108470511)*x[1]+(1.51281979355)*x[2]
        arg[1,0,1,0]=(1.31765698245)*x[0]+(-0.7499566294)*x[1]+(0.934537503478)*x[2]
        arg[1,0,1,1]=(0.688467914883)*x[0]+(-1.32681280673)*x[1]+(-0.387405705058)*x[2]
        arg[1,0,2,0]=(1.37069587556)*x[0]+(0.540830111773)*x[1]+(0.477311068263)*x[2]
        arg[1,0,2,1]=(-0.572066213761)*x[0]+(-1.11950925494)*x[1]+(1.07939217872)*x[2]
        arg[1,1,0,0]=(0.564997105326)*x[0]+(-0.73130546389)*x[1]+(0.435910937304)*x[2]
        arg[1,1,0,1]=(0.513931739674)*x[0]+(-0.944043404826)*x[1]+(0.014114227468)*x[2]
        arg[1,1,1,0]=(-0.77850597256)*x[0]+(-0.445835136484)*x[1]+(0.157478357479)*x[2]
        arg[1,1,1,1]=(0.752431176866)*x[0]+(0.659759664199)*x[1]+(-0.859515433521)*x[2]
        arg[1,1,2,0]=(0.0693158730911)*x[0]+(-0.820997827204)*x[1]+(-0.714694436829)*x[2]
        arg[1,1,2,1]=(0.604674152847)*x[0]+(0.647261183132)*x[1]+(-1.21981352028)*x[2]
        arg[1,2,0,0]=(0.0997504695984)*x[0]+(-0.271549839083)*x[1]+(-0.573467835942)*x[2]
        arg[1,2,0,1]=(0.01861845581)*x[0]+(0.157410502406)*x[1]+(0.939257868776)*x[2]
        arg[1,2,1,0]=(1.44909894631)*x[0]+(0.116038517547)*x[1]+(-0.851940489233)*x[2]
        arg[1,2,1,1]=(1.34598841007)*x[0]+(1.79163125841)*x[1]+(0.719171044245)*x[2]
        arg[1,2,2,0]=(-0.214737791448)*x[0]+(-1.85810839095)*x[1]+(-0.240776599927)*x[2]
        arg[1,2,2,1]=(-0.574707645928)*x[0]+(-0.664173831956)*x[1]+(0.310262106696)*x[2]
        arg[1,3,0,0]=(0.296213453592)*x[0]+(-0.831112482555)*x[1]+(0.130075739128)*x[2]
        arg[1,3,0,1]=(-0.369510604457)*x[0]+(0.0373909697998)*x[1]+(1.6689129783)*x[2]
        arg[1,3,1,0]=(0.0922435881419)*x[0]+(-0.280519827004)*x[1]+(0.285933839232)*x[2]
        arg[1,3,1,1]=(-0.649253282272)*x[0]+(-0.460065044123)*x[1]+(-0.0711164948063)*x[2]
        arg[1,3,2,0]=(0.793313307581)*x[0]+(-0.0587956351149)*x[1]+(0.810487259739)*x[2]
        arg[1,3,2,1]=(-1.38668597055)*x[0]+(1.51396934198)*x[1]+(-1.13918550255)*x[2]
        arg[1,4,0,0]=(1.40078167737)*x[0]+(-1.61915436345)*x[1]+(0.421710091214)*x[2]
        arg[1,4,0,1]=(-1.37982015877)*x[0]+(1.30607371738)*x[1]+(-0.608888848419)*x[2]
        arg[1,4,1,0]=(-0.558087802805)*x[0]+(-0.214169499512)*x[1]+(-0.766242546503)*x[2]
        arg[1,4,1,1]=(-1.75503436584)*x[0]+(0.183604705451)*x[1]+(-0.148071484576)*x[2]
        arg[1,4,2,0]=(0.537942069142)*x[0]+(-0.00640749928252)*x[1]+(-0.277544629657)*x[2]
        arg[1,4,2,1]=(-0.652645833371)*x[0]+(-1.33439077026)*x[1]+(-0.226308197201)*x[2]
        arg[2,0,0,0]=(1.45646289279)*x[0]+(1.40842127675)*x[1]+(1.75078281495)*x[2]
        arg[2,0,0,1]=(0.867988651995)*x[0]+(1.35061812241)*x[1]+(-1.56520776323)*x[2]
        arg[2,0,1,0]=(0.549686523522)*x[0]+(-0.528085465813)*x[1]+(-0.994978820267)*x[2]
        arg[2,0,1,1]=(0.383595996552)*x[0]+(0.612989859087)*x[1]+(-0.560401247231)*x[2]
        arg[2,0,2,0]=(-0.341259919499)*x[0]+(-0.125491675042)*x[1]+(0.545572340757)*x[2]
        arg[2,0,2,1]=(-0.414269395434)*x[0]+(-0.433969881898)*x[1]+(-1.01820199328)*x[2]
        arg[2,1,0,0]=(-0.113575280389)*x[0]+(0.326916039942)*x[1]+(-0.576072890249)*x[2]
        arg[2,1,0,1]=(-0.296473969661)*x[0]+(0.195675099431)*x[1]+(-1.72378150831)*x[2]
        arg[2,1,1,0]=(-1.31575932758)*x[0]+(-0.122557423522)*x[1]+(0.53435027695)*x[2]
        arg[2,1,1,1]=(0.299435754608)*x[0]+(0.727140413706)*x[1]+(0.848931358475)*x[2]
        arg[2,1,2,0]=(0.2618965198)*x[0]+(-0.461672573339)*x[1]+(0.923592803172)*x[2]
        arg[2,1,2,1]=(-0.326817258523)*x[0]+(-0.0363300805442)*x[1]+(1.82049875918)*x[2]
        arg[2,2,0,0]=(-0.435113320888)*x[0]+(0.0257984353786)*x[1]+(0.287846577657)*x[2]
        arg[2,2,0,1]=(1.39178068279)*x[0]+(0.136781967539)*x[1]+(-1.38161590351)*x[2]
        arg[2,2,1,0]=(-0.887874199356)*x[0]+(-0.476769153654)*x[1]+(-0.618760485829)*x[2]
        arg[2,2,1,1]=(1.58883355098)*x[0]+(-0.889469047555)*x[1]+(-0.678486755843)*x[2]
        arg[2,2,2,0]=(0.237147029365)*x[0]+(-0.509883462401)*x[1]+(0.533163151551)*x[2]
        arg[2,2,2,1]=(0.66513913749)*x[0]+(1.30317030657)*x[1]+(0.804083365617)*x[2]
        arg[2,3,0,0]=(1.01531518403)*x[0]+(1.3753053143)*x[1]+(0.000877284235964)*x[2]
        arg[2,3,0,1]=(-0.40853642543)*x[0]+(0.154794268783)*x[1]+(-1.34306671117)*x[2]
        arg[2,3,1,0]=(0.43065284359)*x[0]+(-0.28965146009)*x[1]+(0.660248828789)*x[2]
        arg[2,3,1,1]=(-1.4334922681)*x[0]+(-0.0348802858609)*x[1]+(0.502051304974)*x[2]
        arg[2,3,2,0]=(-1.17399537487)*x[0]+(0.865787043946)*x[1]+(-0.0105949677478)*x[2]
        arg[2,3,2,1]=(0.63104158794)*x[0]+(0.436196079951)*x[1]+(0.498606623061)*x[2]
        arg[2,4,0,0]=(-1.44410177354)*x[0]+(0.513406326011)*x[1]+(0.699127199447)*x[2]
        arg[2,4,0,1]=(-0.981727023077)*x[0]+(0.484182120695)*x[1]+(0.667175832045)*x[2]
        arg[2,4,1,0]=(-1.3352602588)*x[0]+(0.626269023364)*x[1]+(0.813537646695)*x[2]
        arg[2,4,1,1]=(-0.932486930114)*x[0]+(0.299448807849)*x[1]+(0.195862014797)*x[2]
        arg[2,4,2,0]=(-0.0621489644806)*x[0]+(0.131979664158)*x[1]+(0.442029269778)*x[2]
        arg[2,4,2,1]=(-0.507508608633)*x[0]+(-1.28613226533)*x[1]+(-0.697920444023)*x[2]
        arg[3,0,0,0]=(-0.984942677585)*x[0]+(-1.58913299616)*x[1]+(-0.373026784822)*x[2]
        arg[3,0,0,1]=(0.616666427113)*x[0]+(-0.625162800471)*x[1]+(-0.796307290839)*x[2]
        arg[3,0,1,0]=(-0.48469203683)*x[0]+(-0.0572105166553)*x[1]+(0.809572383581)*x[2]
        arg[3,0,1,1]=(-0.487160309892)*x[0]+(0.77646454341)*x[1]+(-0.41947544573)*x[2]
        arg[3,0,2,0]=(-1.17902121554)*x[0]+(-0.244939532033)*x[1]+(-0.282994359476)*x[2]
        arg[3,0,2,1]=(0.33515909881)*x[0]+(0.146866561705)*x[1]+(1.81486098364)*x[2]
        arg[3,1,0,0]=(0.379977901341)*x[0]+(0.834531701517)*x[1]+(-0.0242027062166)*x[2]
        arg[3,1,0,1]=(0.184318292766)*x[0]+(1.08367911805)*x[1]+(-0.164531741594)*x[2]
        arg[3,1,1,0]=(-1.00967115996)*x[0]+(0.542699640149)*x[1]+(0.707616701259)*x[2]
        arg[3,1,1,1]=(0.856948233811)*x[0]+(-1.2186550297)*x[1]+(-0.034268798159)*x[2]
        arg[3,1,2,0]=(0.259470684477)*x[0]+(-1.03586983732)*x[1]+(-0.157275979341)*x[2]
        arg[3,1,2,1]=(-0.161954890658)*x[0]+(1.65853209885)*x[1]+(0.381113456719)*x[2]
        arg[3,2,0,0]=(-0.437413004681)*x[0]+(-0.730340273929)*x[1]+(-0.687689272494)*x[2]
        arg[3,2,0,1]=(-0.336618164118)*x[0]+(-1.24388666895)*x[1]+(-0.2198689254)*x[2]
        arg[3,2,1,0]=(0.0872793087887)*x[0]+(-0.336554573215)*x[1]+(-0.0597803523262)*x[2]
        arg[3,2,1,1]=(0.192805432018)*x[0]+(0.0540495582064)*x[1]+(0.846658042071)*x[2]
        arg[3,2,2,0]=(0.00794192604414)*x[0]+(-0.33504559053)*x[1]+(0.615701149787)*x[2]
        arg[3,2,2,1]=(0.210870961503)*x[0]+(1.14164095846)*x[1]+(-0.192785742791)*x[2]
        arg[3,3,0,0]=(-0.25895162041)*x[0]+(-0.225800436488)*x[1]+(-0.198007079838)*x[2]
        arg[3,3,0,1]=(-0.150363703231)*x[0]+(1.23870429535)*x[1]+(0.28635881247)*x[2]
        arg[3,3,1,0]=(-0.865980718216)*x[0]+(-0.0849853244335)*x[1]+(-0.292164217189)*x[2]
        arg[3,3,1,1]=(-0.350465968722)*x[0]+(0.09853736936)*x[1]+(-0.476711524665)*x[2]
        arg[3,3,2,0]=(0.311396300263)*x[0]+(-0.92437612034)*x[1]+(-0.757139962928)*x[2]
        arg[3,3,2,1]=(-0.912311919837)*x[0]+(0.267793631099)*x[1]+(0.4917345401)*x[2]
        arg[3,4,0,0]=(0.613662541753)*x[0]+(0.782757341361)*x[1]+(1.04209808158)*x[2]
        arg[3,4,0,1]=(0.492828613997)*x[0]+(1.11979095664)*x[1]+(-0.843545072354)*x[2]
        arg[3,4,1,0]=(-0.51939120556)*x[0]+(0.476004484286)*x[1]+(-0.516311785432)*x[2]
        arg[3,4,1,1]=(1.00423313567)*x[0]+(0.90776890388)*x[1]+(1.34423309497)*x[2]
        arg[3,4,2,0]=(-1.44069600826)*x[0]+(-1.39515290909)*x[1]+(1.19851590247)*x[2]
        arg[3,4,2,1]=(0.391851619877)*x[0]+(0.265529120553)*x[1]+(0.996610410913)*x[2]
        ref[0,0,0,0]=(1.19330263935)/(o+1.)+(0.55394275251)+(0.0162017895041)*0.5**o
        ref[0,0,0,1]=(-0.672351783684)/(o+1.)+(-0.187889622039)+(0.351132807711)*0.5**o
        ref[0,0,1,0]=(-1.27021996264)/(o+1.)+(0.236286530101)+(-0.961885459894)*0.5**o
        ref[0,0,1,1]=(-1.43442517107)/(o+1.)+(-1.00113374859)+(-0.935971122022)*0.5**o
        ref[0,0,2,0]=(-0.434813695127)/(o+1.)+(0.606673209583)+(-0.0699546855712)*0.5**o
        ref[0,0,2,1]=(-0.0434901247948)/(o+1.)+(-0.180911877983)+(-0.733211446073)*0.5**o
        ref[0,1,0,0]=(1.41514302182)/(o+1.)+(-0.247977413325)+(-0.967062307989)*0.5**o
        ref[0,1,0,1]=(0.69128100142)/(o+1.)+(-0.621609882923)+(-0.319976390946)*0.5**o
        ref[0,1,1,0]=(-0.881870429191)/(o+1.)+(-0.654075774082)+(0.0281262553057)*0.5**o
        ref[0,1,1,1]=(-1.27239116086)/(o+1.)+(-0.0854066429492)+(0.0230865806541)*0.5**o
        ref[0,1,2,0]=(-0.222663945351)/(o+1.)+(0.685312598747)+(0.453385341573)*0.5**o
        ref[0,1,2,1]=(0.101409038706)/(o+1.)+(0.151367946744)+(-0.824910502191)*0.5**o
        ref[0,2,0,0]=(-1.19509443892)/(o+1.)+(0.225437856866)+(0.85928567141)*0.5**o
        ref[0,2,0,1]=(0.285764347754)/(o+1.)+(-0.383432127133)+(0.963095541446)*0.5**o
        ref[0,2,1,0]=(0.728010227911)/(o+1.)+(0.983790919896)+(-0.550403317717)*0.5**o
        ref[0,2,1,1]=(0.563748259036)/(o+1.)+(0.165164845927)+(-0.110710881481)*0.5**o
        ref[0,2,2,0]=(-0.361502606429)/(o+1.)+(0.0276145482611)+(-0.341115463028)*0.5**o
        ref[0,2,2,1]=(-0.293320207933)/(o+1.)+(0.0336757304385)+(-0.853103403872)*0.5**o
        ref[0,3,0,0]=(-0.410233956527)/(o+1.)+(-0.474281126245)+(0.569057885785)*0.5**o
        ref[0,3,0,1]=(-0.835861970115)/(o+1.)+(0.0173715483175)+(0.346902459481)*0.5**o
        ref[0,3,1,0]=(0.0830466456187)/(o+1.)+(-0.288204681495)+(-0.607109864241)*0.5**o
        ref[0,3,1,1]=(1.07074130823)/(o+1.)+(-0.0437983012867)+(0.159379549652)*0.5**o
        ref[0,3,2,0]=(0.172301049609)/(o+1.)+(-0.491724874152)+(-0.996925180895)*0.5**o
        ref[0,3,2,1]=(-0.24891941221)/(o+1.)+(-0.03935856269)+(-0.3030090126)*0.5**o
        ref[0,4,0,0]=(-0.778575122429)/(o+1.)+(0.213085222078)+(0.941383668349)*0.5**o
        ref[0,4,0,1]=(0.880801627174)/(o+1.)+(-0.454451723401)+(-0.193978040747)*0.5**o
        ref[0,4,1,0]=(-0.835604724882)/(o+1.)+(0.127240834083)+(-0.245156740036)*0.5**o
        ref[0,4,1,1]=(1.62336716324)/(o+1.)+(-0.101085918464)+(0.398884015839)*0.5**o
        ref[0,4,2,0]=(1.16805827702)/(o+1.)+(0.211232591688)+(-0.654729385702)*0.5**o
        ref[0,4,2,1]=(-0.305424138957)/(o+1.)+(-0.431076572693)+(-0.924903192969)*0.5**o
        ref[1,0,0,0]=(-0.481596247529)/(o+1.)+(0.769902632559)+(0.748050413171)*0.5**o
        ref[1,0,0,1]=(0.177388661252)/(o+1.)+(-0.118766442613)+(-0.457657277952)*0.5**o
        ref[1,0,1,0]=(0.709890860372)/(o+1.)+(0.0668010230727)+(0.658744950012)*0.5**o
        ref[1,0,1,1]=(-1.45685781381)/(o+1.)+(-0.178674426089)+(0.788456069079)*0.5**o
        ref[1,0,2,0]=(0.194388185612)/(o+1.)+(0.649177543063)+(0.896093783855)*0.5**o
        ref[1,0,2,1]=(0.181561279507)/(o+1.)+(-0.200618291902)+(-0.392507985678)*0.5**o
        ref[1,1,0,0]=(-0.884376814027)/(o+1.)+(0.280382736159)+(0.593213920451)*0.5**o
        ref[1,1,0,1]=(-0.250709602218)/(o+1.)+(-0.492104211211)+(0.818920586956)*0.5**o
        ref[1,1,1,0]=(0.0529723694625)/(o+1.)+(-0.502047650638)+(-0.115739819752)*0.5**o
        ref[1,1,1,1]=(0.215812202957)/(o+1.)+(-0.308796450292)+(0.954456105171)*0.5**o
        ref[1,1,2,0]=(-0.991481050902)/(o+1.)+(0.162626252838)+(-0.800147845717)*0.5**o
        ref[1,1,2,1]=(-0.433989756951)/(o+1.)+(0.362776336712)+(-0.259441100777)*0.5**o
        ref[1,2,0,0]=(-0.930287052624)/(o+1.)+(-0.347752763108)+(0.880525373413)*0.5**o
        ref[1,2,0,1]=(0.698674852861)/(o+1.)+(-0.15119613196)+(0.719004238051)*0.5**o
        ref[1,2,1,0]=(-0.777216821365)/(o+1.)+(0.436336076297)+(0.617741643399)*0.5**o
        ref[1,2,1,1]=(1.09595223135)/(o+1.)+(0.978251130983)+(0.804336219412)*0.5**o
        ref[1,2,2,0]=(-0.651012630849)/(o+1.)+(-1.04894606664)+(0.43528198181)*0.5**o
        ref[1,2,2,1]=(-1.5227353883)/(o+1.)+(0.670019307557)+(-0.745922598001)*0.5**o
        ref[1,3,0,0]=(-0.750195637997)/(o+1.)+(0.231864245348)+(-0.118356142533)*0.5**o
        ref[1,3,0,1]=(1.17136226596)/(o+1.)+(-0.114976421389)+(0.395383920468)*0.5**o
        ref[1,3,1,0]=(-0.108070636769)/(o+1.)+(0.279145429234)+(-0.352562621327)*0.5**o
        ref[1,3,1,1]=(-0.144223481714)/(o+1.)+(-0.472954352493)+(-0.0903026345013)*0.5**o
        ref[1,3,2,0]=(-0.487564307786)/(o+1.)+(0.636973351449)+(0.758622537094)*0.5**o
        ref[1,3,2,1]=(0.390860965954)/(o+1.)+(-0.36647678165)+(-0.66980953377)*0.5**o
        ref[1,4,0,0]=(-1.04824984837)/(o+1.)+(0.213932231728)+(0.823722790054)*0.5**o
        ref[1,4,0,1]=(-0.0984653601336)/(o+1.)+(0.07867877705)+(-0.741527483773)*0.5**o
        ref[1,4,1,0]=(-0.146287867383)/(o+1.)+(-0.908951868158)+(0.425691754878)*0.5**o
        ref[1,4,1,1]=(-0.94160416554)/(o+1.)+(0.0459877010016)+(-0.869872381431)*0.5**o
        ref[1,4,2,0]=(0.0115742858952)/(o+1.)+(-0.222488399041)+(0.687392452389)*0.5**o
        ref[1,4,2,1]=(-0.956561936575)/(o+1.)+(-0.59770879042)+(-0.0613652834159)*0.5**o
        ref[2,0,0,0]=(1.83987381729)/(o+1.)+(0.956579317963)+(0.862634531274)*0.5**o
        ref[2,0,0,1]=(0.105446127421)/(o+1.)+(0.180661881191)+(0.186629121375)*0.5**o
        ref[2,0,1,0]=(0.0212167440766)/(o+1.)+(-0.337825588888)+(-0.318943328859)*0.5**o
        ref[2,0,1,1]=(0.744834715631)/(o+1.)+(-0.350058416663)+(0.391466726104)*0.5**o
        ref[2,0,2,0]=(-0.788776613856)/(o+1.)+(0.233776806014)+(0.400043748043)*0.5**o
        ref[2,0,2,1]=(-0.95346976307)/(o+1.)+(0.0274998652324)+(-0.967971238003)*0.5**o
        ref[2,1,0,0]=(-0.536051408942)/(o+1.)+(-0.208498016994)+(0.590315312233)*0.5**o
        ref[2,1,0,1]=(-1.04158508788)/(o+1.)+(-0.397885793101)+(0.0127762955396)*0.5**o
        ref[2,1,1,0]=(0.529914207253)/(o+1.)+(-0.547503443987)+(-0.338873793431)*0.5**o
        ref[2,1,1,1]=(0.62298670274)/(o+1.)+(0.449860048786)+(0.352800726478)*0.5**o
        ref[2,1,2,0]=(0.965292650595)/(o+1.)+(0.00980498639)+(-0.261085873742)*0.5**o
        ref[2,1,2,1]=(0.403285360803)/(o+1.)+(0.369160592269)+(0.315744874774)*0.5**o
        ref[2,2,0,0]=(-0.457764884724)/(o+1.)+(0.563657776706)+(-0.791018976541)*0.5**o
        ref[2,2,0,1]=(0.188151274044)/(o+1.)+(-0.476470545911)+(0.911736564588)*0.5**o
        ref[2,2,1,0]=(-0.558515085739)/(o+1.)+(-0.445583796553)+(-0.533721159994)*0.5**o
        ref[2,2,1,1]=(-0.431992337265)/(o+1.)+(-0.0957726221961)+(0.64441532924)*0.5**o
        ref[2,2,2,0]=(-0.159056461947)/(o+1.)+(-0.0326864536156)+(0.484856087693)*0.5**o
        ref[2,2,2,1]=(0.533116456037)/(o+1.)+(1.10197257725)+(0.0353311991352)*0.5**o
        ref[2,3,0,0]=(0.0570625147023)/(o+1.)+(0.74170779854)+(0.851019670777)*0.5**o
        ref[2,3,0,1]=(-0.495059904832)/(o+1.)+(-0.169954912181)+(-0.761839138623)*0.5**o
        ref[2,3,1,0]=(0.482375173314)/(o+1.)+(-0.19374838651)+(0.706371811996)*0.5**o
        ref[2,3,1,1]=(0.12928100721)/(o+1.)+(-0.317327862062)+(-0.460946532077)*0.5**o
        ref[2,3,2,0]=(-0.234362084763)/(o+1.)+(0.423486279181)+(-0.93141377227)*0.5**o
        ref[2,3,2,1]=(1.25586693002)/(o+1.)+(-0.127551329993)+(0.56508002092)*0.5**o
        ref[2,4,0,0]=(0.45235297749)/(o+1.)+(-0.0217537364788)+(-0.640413752615)*0.5**o
        ref[2,4,0,1]=(0.936809207384)/(o+1.)+(-0.151519610695)+(-0.464139056331)*0.5**o
        ref[2,4,1,0]=(0.867693208863)/(o+1.)+(0.105683574314)+(-0.974513946233)*0.5**o
        ref[2,4,1,1]=(-1.27821575101)/(o+1.)+(0.463163847311)+(-0.0852880510814)*0.5**o
        ref[2,4,2,0]=(-0.846224501366)/(o+1.)+(0.432619461409)+(0.492845548002)*0.5**o
        ref[2,4,2,1]=(-1.48923229995)/(o+1.)+(-0.505747396375)+(0.00916577472118)*0.5**o
        ref[3,0,0,0]=(-0.959215731121)/(o+1.)+(-0.904609108587)+(-0.178668510268)*0.5**o
        ref[3,0,0,1]=(-1.53181231865)/(o+1.)+(-0.0758543610408)+(0.878717376536)*0.5**o
        ref[3,0,1,0]=(1.18643238831)/(o+1.)+(-0.288825159097)+(-0.341112240015)*0.5**o
        ref[3,0,1,1]=(-0.0675694917031)/(o+1.)+(-0.168955703616)+(0.275309686724)*0.5**o
        ref[3,0,2,0]=(-0.663269860133)/(o+1.)+(-0.235102421885)+(-0.573480403148)*0.5**o
        ref[3,0,2,1]=(1.0969289778)/(o+1.)+(0.46603172821)+(0.267894209932)*0.5**o
        ref[3,1,0,0]=(-0.716870513559)/(o+1.)+(0.712039409283)+(0.483098591635)*0.5**o
        ref[3,1,0,1]=(0.0328531999411)/(o+1.)+(0.379319450988)+(0.311973567308)*0.5**o
        ref[3,1,1,0]=(1.08648208064)/(o+1.)+(-0.196831959338)+(-0.452172980519)*0.5**o
        ref[3,1,1,1]=(0.0813181502612)/(o+1.)+(-0.227364064527)+(-0.0225656152548)*0.5**o
        ref[3,1,2,0]=(-0.271000454408)/(o+1.)+(-0.765711687155)+(0.868748696531)*0.5**o
        ref[3,1,2,1]=(1.60569646253)/(o+1.)+(0.137437739736)+(-0.00288127709197)*0.5**o
        ref[3,2,0,0]=(-1.25887089059)/(o+1.)+(-0.525842290945)+(0.455112921376)*0.5**o
        ref[3,2,0,1]=(-1.63787880725)/(o+1.)+(0.00214706725405)+(-0.166789085722)*0.5**o
        ref[3,2,1,0]=(-1.82111307894)/(o+1.)+(0.561806270366)+(0.388444921454)*0.5**o
        ref[3,2,1,1]=(-0.489259906414)/(o+1.)+(0.582548398381)+(0.417676141947)*0.5**o
        ref[3,2,2,0]=(-0.0674375340391)/(o+1.)+(-0.00345549227349)+(0.362946003887)*0.5**o
        ref[3,2,2,1]=(0.662780682001)/(o+1.)+(-0.15453848218)+(0.806022459534)*0.5**o
        ref[3,3,0,0]=(-1.15795406515)/(o+1.)+(0.0926725746467)+(0.289849779118)*0.5**o
        ref[3,3,0,1]=(1.27520269547)/(o+1.)+(-0.286529818361)+(0.672556345847)*0.5**o
        ref[3,3,1,0]=(-0.343290651469)/(o+1.)+(-0.0269962100904)+(-0.845847188189)*0.5**o
        ref[3,3,1,1]=(0.497715053371)/(o+1.)+(-0.158577177766)+(-0.909200821866)*0.5**o
        ref[3,3,2,0]=(-0.952431538662)/(o+1.)+(0.0816482063045)+(-0.580984656952)*0.5**o
        ref[3,3,2,1]=(-0.176227469799)/(o+1.)+(0.0456157853688)+(-0.0677878495767)*0.5**o
        ref[3,4,0,0]=(0.976460483802)/(o+1.)+(0.347595506011)+(0.766866468871)*0.5**o
        ref[3,4,0,1]=(0.0358170327662)/(o+1.)+(-0.0991456259778)+(0.931548717473)*0.5**o
        ref[3,4,1,0]=(0.114544318514)/(o+1.)+(-0.201014865792)+(-0.272213093635)*0.5**o
        ref[3,4,1,1]=(0.554509925385)/(o+1.)+(1.23219332557)+(0.237338557994)*0.5**o
        ref[3,4,2,0]=(0.0803465444931)/(o+1.)+(-0.518257569997)+(-0.681164419371)*0.5**o
        ref[3,4,2,1]=(-0.0443609570522)/(o+1.)+(0.572813368096)+(0.552725372202)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.570707549273)*x[0]+(-0.726264032186)*x[1]
        ref=(-0.0423765208873)/(o+1.)+(-0.791542656673)+(0.328490252775)*0.5**o
      else:
        arg=(1.28222892931)*x[0]+(-0.798170675576)*x[1]+(1.5701208405)*x[2]
        ref=(0.721975372868)/(o+1.)+(0.344720905832)+(0.642761909705)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.575166906798)*x[0]+(-1.63698773418)*x[1]
        arg[1]=(-0.0535946584636)*x[0]+(0.105969101775)*x[1]
        ref[0]=(-0.751926426344)/(o+1.)+(-0.395800138888)+(0.481705876734)*0.5**o
        ref[1]=(0.698496252242)/(o+1.)+(-0.265927161866)+(-0.114267485199)*0.5**o
      else:
        arg[0]=(-0.17879254138)*x[0]+(-1.12752123792)*x[1]+(0.921297311605)*x[2]
        arg[1]=(1.33131824618)*x[0]+(-0.392175059176)*x[1]+(0.545754281465)*x[2]
        ref[0]=(-0.494870303061)/(o+1.)+(0.0697636237654)+(-0.0296734121631)*0.5**o
        ref[1]=(-0.394300983704)/(o+1.)+(0.722261312754)+(0.434675826665)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.338658191734)*x[0]+(1.46127609683)*x[1]
        arg[0,1]=(-0.629116867491)*x[0]+(1.40141569066)*x[1]
        arg[0,2]=(0.105569779347)*x[0]+(0.239936324967)*x[1]
        arg[0,3]=(-0.281718331718)*x[0]+(0.280323404866)*x[1]
        arg[0,4]=(0.467341113661)*x[0]+(-0.50452062075)*x[1]
        arg[1,0]=(-0.841295219186)*x[0]+(0.524038163723)*x[1]
        arg[1,1]=(1.40143704236)*x[0]+(-1.50467961637)*x[1]
        arg[1,2]=(1.21078790505)*x[0]+(-0.731453470292)*x[1]
        arg[1,3]=(0.469556922076)*x[0]+(0.0686293204391)*x[1]
        arg[1,4]=(-1.26394637585)*x[0]+(-0.0361395276585)*x[1]
        arg[2,0]=(-1.12850883148)*x[0]+(0.284296673326)*x[1]
        arg[2,1]=(-0.626362828813)*x[0]+(1.19459827573)*x[1]
        arg[2,2]=(-0.545457293083)*x[0]+(1.547337019)*x[1]
        arg[2,3]=(0.664461986629)*x[0]+(-1.69082198266)*x[1]
        arg[2,4]=(-0.100004763305)*x[0]+(0.152159044423)*x[1]
        arg[3,0]=(0.307574843061)*x[0]+(1.20209181301)*x[1]
        arg[3,1]=(-0.0663203356627)*x[0]+(-1.26669863527)*x[1]
        arg[3,2]=(-1.20436798894)*x[0]+(-0.0375693379661)*x[1]
        arg[3,3]=(0.753649087747)*x[0]+(0.69376223486)*x[1]
        arg[3,4]=(-0.0512732264834)*x[0]+(-0.658097606436)*x[1]
        ref[0,0]=(0.996821439952)/(o+1.)+(0.015088160779)+(0.0956201435879)*0.5**o
        ref[0,1]=(0.643092963726)/(o+1.)+(0.239390724844)+(-0.349575590244)*0.5**o
        ref[0,2]=(-0.198468008977)/(o+1.)+(0.00411201743342)+(0.535750078425)*0.5**o
        ref[0,3]=(0.642491201137)/(o+1.)+(-0.623413199652)+(0.602940271315)*0.5**o
        ref[0,4]=(-0.105706120549)/(o+1.)+(-0.353560054939)+(0.775646723338)*0.5**o
        ref[1,0]=(0.144540669221)/(o+1.)+(0.0228268215803)+(-0.507451367845)*0.5**o
        ref[1,1]=(-0.820379339897)/(o+1.)+(0.0951825678509)+(0.526771630185)*0.5**o
        ref[1,2]=(-0.21214585824)/(o+1.)+(-0.0889183062985)+(0.869316905591)*0.5**o
        ref[1,3]=(0.291228940674)/(o+1.)+(0.192623928703)+(-0.138290555565)*0.5**o
        ref[1,4]=(-0.0663718956434)/(o+1.)+(-0.309484335614)+(-0.614745336642)*0.5**o
        ref[2,0]=(0.484892962201)/(o+1.)+(-0.367129859793)+(-0.594845400763)*0.5**o
        ref[2,1]=(0.768457634672)/(o+1.)+(0.154447657562)+(-0.509117502876)*0.5**o
        ref[2,2]=(0.707320338363)/(o+1.)+(-0.0592149232183)+(0.412989233992)*0.5**o
        ref[2,3]=(-0.983684339171)/(o+1.)+(-0.495377645598)+(0.948079634337)*0.5**o
        ref[2,4]=(-0.0315947220465)/(o+1.)+(0.506833986759)+(-0.929918970353)*0.5**o
        ref[3,0]=(0.54141067098)/(o+1.)+(0.489049131493)+(-0.00984227789859)*0.5**o
        ref[3,1]=(-0.552335053051)/(o+1.)+(-0.383325638803)+(-0.0140326402734)*0.5**o
        ref[3,2]=(0.0821472412575)/(o+1.)+(-0.335827719718)+(-0.652429128733)*0.5**o
        ref[3,3]=(0.112917208924)/(o+1.)+(0.667772807338)+(-0.00105150099312)*0.5**o
        ref[3,4]=(-0.19331508157)/(o+1.)+(-0.102296385708)+(-0.311462979932)*0.5**o
      else:
        arg[0,0]=(0.069245963142)*x[0]+(-0.185727379494)*x[1]+(-0.887410889699)*x[2]
        arg[0,1]=(-0.816745668949)*x[0]+(-0.250435680662)*x[1]+(0.701607830582)*x[2]
        arg[0,2]=(-0.862433954308)*x[0]+(0.731874372527)*x[1]+(-0.115384982779)*x[2]
        arg[0,3]=(-0.265247631604)*x[0]+(-0.157059741958)*x[1]+(-0.368419335078)*x[2]
        arg[0,4]=(1.66720203567)*x[0]+(-0.759393708719)*x[1]+(1.36360850704)*x[2]
        arg[1,0]=(-0.192822813795)*x[0]+(0.0358776152527)*x[1]+(-0.722864520788)*x[2]
        arg[1,1]=(-0.480606215471)*x[0]+(1.13682774573)*x[1]+(0.17900467623)*x[2]
        arg[1,2]=(0.346760852584)*x[0]+(0.994833030281)*x[1]+(0.507055933913)*x[2]
        arg[1,3]=(-0.509961853478)*x[0]+(-0.852997575505)*x[1]+(0.586527030253)*x[2]
        arg[1,4]=(0.93957975846)*x[0]+(0.28878335152)*x[1]+(0.128070931738)*x[2]
        arg[2,0]=(0.338283306313)*x[0]+(0.0244500761608)*x[1]+(-0.606875554888)*x[2]
        arg[2,1]=(0.850782624552)*x[0]+(-0.836382038189)*x[1]+(-1.41808459947)*x[2]
        arg[2,2]=(-1.06176229613)*x[0]+(-0.259247046061)*x[1]+(-0.583455484089)*x[2]
        arg[2,3]=(1.17437923282)*x[0]+(1.31485601495)*x[1]+(1.3139743231)*x[2]
        arg[2,4]=(0.362851636275)*x[0]+(0.804049783156)*x[1]+(1.09205000214)*x[2]
        arg[3,0]=(0.343603837195)*x[0]+(-1.49082257981)*x[1]+(-0.628668674545)*x[2]
        arg[3,1]=(-0.676847711765)*x[0]+(1.32350560453)*x[1]+(1.07275185686)*x[2]
        arg[3,2]=(0.570784985469)*x[0]+(-1.01212360602)*x[1]+(-0.0808776667606)*x[2]
        arg[3,3]=(0.506392444236)*x[0]+(-1.00848318041)*x[1]+(0.558777901541)*x[2]
        arg[3,4]=(-0.60925738291)*x[0]+(-0.386755957809)*x[1]+(0.60930115434)*x[2]
        ref[0,0]=(-0.695230136419)/(o+1.)+(-0.388400729083)+(0.468139288533)*0.5**o
        ref[0,1]=(0.462098132187)/(o+1.)+(-0.215041893287)+(-0.397587864642)*0.5**o
        ref[0,2]=(-0.262411385363)/(o+1.)+(0.275729226874)+(-0.534991632946)*0.5**o
        ref[0,3]=(0.117822633378)/(o+1.)+(0.0286827897592)+(-0.965914921536)*0.5**o
        ref[0,4]=(0.293575681039)/(o+1.)+(0.584039692827)+(0.809761767299)*0.5**o
        ref[1,0]=(0.627951922463)/(o+1.)+(-0.994889162557)+(0.482016683321)*0.5**o
        ref[1,1]=(1.83482238662)/(o+1.)+(-0.205719558739)+(-0.588157062647)*0.5**o
        ref[1,2]=(1.13612020398)/(o+1.)+(0.0262824938134)+(0.659964625174)*0.5**o
        ref[1,3]=(0.325264487741)/(o+1.)+(-0.0752154813413)+(-0.951265923787)*0.5**o
        ref[1,4]=(-0.149919283302)/(o+1.)+(0.77701502585)+(-0.0476767266789)*0.5**o
        ref[2,0]=(0.491504686171)/(o+1.)+(-0.562534706926)+(0.389422555265)*0.5**o
        ref[2,1]=(-1.52704217417)/(o+1.)+(0.120525042098)+(-0.117691923134)*0.5**o
        ref[2,2]=(-0.779796772543)/(o+1.)+(-0.318864997267)+(-0.486938059201)*0.5**o
        ref[2,3]=(1.53343318682)/(o+1.)+(0.731149467308)+(0.807477449443)*0.5**o
        ref[2,4]=(1.15919207694)/(o+1.)+(0.113921448815)+(0.871916446996)*0.5**o
        ref[3,0]=(-1.75312920944)/(o+1.)+(-0.153798982308)+(0.284839756895)*0.5**o
        ref[3,1]=(1.30985814569)/(o+1.)+(0.438430874237)+(-0.467310144534)*0.5**o
        ref[3,2]=(-0.737480320972)/(o+1.)+(0.0604403094397)+(0.094383414777)*0.5**o
        ref[3,3]=(-0.688723266524)/(o+1.)+(0.43486027018)+(-0.124310108473)*0.5**o
        ref[3,4]=(0.128399262674)/(o+1.)+(-0.372445844122)+(0.229780239192)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.993339371954)*x[0]+(-1.35308128985)*x[1]
        arg[0,0,1]=(-1.07886149134)*x[0]+(-1.38999283206)*x[1]
        arg[0,1,0]=(0.246468026972)*x[0]+(-1.37217464882)*x[1]
        arg[0,1,1]=(0.645672648599)*x[0]+(1.8070624103)*x[1]
        arg[1,0,0]=(-1.02278989388)*x[0]+(0.151634023103)*x[1]
        arg[1,0,1]=(-0.667332860541)*x[0]+(0.312742149193)*x[1]
        arg[1,1,0]=(-0.848647385832)*x[0]+(0.466438038516)*x[1]
        arg[1,1,1]=(-0.185495911372)*x[0]+(-0.137118786891)*x[1]
        arg[2,0,0]=(-0.0250555870731)*x[0]+(0.905553573283)*x[1]
        arg[2,0,1]=(-1.03940852939)*x[0]+(-1.57326016394)*x[1]
        arg[2,1,0]=(1.48834718811)*x[0]+(-0.340889253246)*x[1]
        arg[2,1,1]=(0.875721514739)*x[0]+(-0.310857412212)*x[1]
        arg[3,0,0]=(1.25129508831)*x[0]+(-1.69312043525)*x[1]
        arg[3,0,1]=(1.50124675312)*x[0]+(0.244195748654)*x[1]
        arg[3,1,0]=(0.292060478909)*x[0]+(-0.660005422696)*x[1]
        arg[3,1,1]=(0.791335231308)*x[0]+(0.466481573591)*x[1]
        arg[4,0,0]=(-0.361241947236)*x[0]+(-1.43591648587)*x[1]
        arg[4,0,1]=(-1.10385456695)*x[0]+(-0.325314195825)*x[1]
        arg[4,1,0]=(1.01004173033)*x[0]+(0.426824552857)*x[1]
        arg[4,1,1]=(-0.631382479612)*x[0]+(0.0603792190966)*x[1]
        arg[5,0,0]=(-0.30404368786)*x[0]+(-0.0102498522932)*x[1]
        arg[5,0,1]=(0.109188848746)*x[0]+(-1.16846160105)*x[1]
        arg[5,1,0]=(-0.351916879547)*x[0]+(-0.320714795448)*x[1]
        arg[5,1,1]=(-0.534138562652)*x[0]+(-0.256013589844)*x[1]
        ref[0,0,0]=(-0.506159375456)/(o+1.)+(-0.522895228783)+(-0.794470828783)*0.5**o
        ref[0,0,1]=(-0.930984367068)/(o+1.)+(-0.636851258685)+(-0.264167438968)*0.5**o
        ref[0,1,0]=(-0.674222792172)/(o+1.)+(-0.52131379048)+(0.591143751282)*0.5**o
        ref[0,1,1]=(0.953370751283)/(o+1.)+(0.377773224645)+(0.743817858326)*0.5**o
        ref[1,0,0]=(0.700508215428)/(o+1.)+(-0.303343530867)+(-0.964977024471)*0.5**o
        ref[1,0,1]=(-0.408185116452)/(o+1.)+(0.0668983569743)+(-0.0802023088445)*0.5**o
        ref[1,1,0]=(0.80742345527)/(o+1.)+(-0.405780705683)+(-0.37807139122)*0.5**o
        ref[1,1,1]=(0.0184615568095)/(o+1.)+(0.221921562266)+(-0.784919379605)*0.5**o
        ref[2,0,0]=(0.571906256883)/(o+1.)+(-0.0415588382133)+(0.391709405753)*0.5**o
        ref[2,0,1]=(-0.778176009451)/(o+1.)+(-0.761680990566)+(-0.311130702752)*0.5**o
        ref[2,1,0]=(0.498214686321)/(o+1.)+(-0.146173705483)+(0.941590659509)*0.5**o
        ref[2,1,1]=(-0.273747071818)/(o+1.)+(-0.0319184747153)+(0.902448123776)*0.5**o
        ref[3,0,0]=(-0.730029471178)/(o+1.)+(-0.288672747933)+(0.865549620109)*0.5**o
        ref[3,0,1]=(-0.371409186475)/(o+1.)+(0.623169375401)+(0.870512937445)*0.5**o
        ref[3,1,0]=(0.0611454391867)/(o+1.)+(-0.627723874214)+(0.826357365454)*0.5**o
        ref[3,1,1]=(0.785553447935)/(o+1.)+(0.246940213405)+(-0.021617069845)*0.5**o
        ref[4,0,0]=(-0.936201955907)/(o+1.)+(-0.484823720021)+(0.108690962842)*0.5**o
        ref[4,0,1]=(-0.277830390815)/(o+1.)+(-0.151786547071)+(-0.847765277819)*0.5**o
        ref[4,1,0]=(0.12293099533)/(o+1.)+(0.295620382054)+(0.722694523749)*0.5**o
        ref[4,1,1]=(-0.5517019185)/(o+1.)+(0.0850748937088)+(-0.189451129433)*0.5**o
        ref[5,0,0]=(-0.942533148805)/(o+1.)+(0.425680504872)+(-0.223121401092)*0.5**o
        ref[5,0,1]=(-0.377134228332)/(o+1.)+(-0.721019409569)+(0.759900295162)*0.5**o
        ref[5,1,0]=(-0.643828391942)/(o+1.)+(0.399974456311)+(-0.828752195674)*0.5**o
        ref[5,1,1]=(-0.118293731439)/(o+1.)+(-0.355568544431)+(0.0392786678054)*0.5**o
      else:
        arg[0,0,0]=(0.205446621454)*x[0]+(0.10254878667)*x[1]+(1.27673728979)*x[2]
        arg[0,0,1]=(0.758846313193)*x[0]+(0.63320336713)*x[1]+(-0.42638505807)*x[2]
        arg[0,1,0]=(-0.0127116561603)*x[0]+(-0.360359420027)*x[1]+(0.540508591493)*x[2]
        arg[0,1,1]=(-0.278918629538)*x[0]+(-0.0291534835772)*x[1]+(0.708655731565)*x[2]
        arg[1,0,0]=(-0.612965210483)*x[0]+(0.787787556912)*x[1]+(0.222384448462)*x[2]
        arg[1,0,1]=(1.07409083742)*x[0]+(1.19429173706)*x[1]+(-0.484979976779)*x[2]
        arg[1,1,0]=(-1.25714999157)*x[0]+(-0.321346019622)*x[1]+(0.278329320837)*x[2]
        arg[1,1,1]=(-0.512262689707)*x[0]+(1.6891992053)*x[1]+(-1.02351830068)*x[2]
        arg[2,0,0]=(-1.57911907773)*x[0]+(0.096124258122)*x[1]+(0.336619253521)*x[2]
        arg[2,0,1]=(0.0901144254199)*x[0]+(-1.36818186255)*x[1]+(1.2969609239)*x[2]
        arg[2,1,0]=(0.158329511499)*x[0]+(0.696997185217)*x[1]+(-0.400535235927)*x[2]
        arg[2,1,1]=(-1.44570853201)*x[0]+(0.237359587195)*x[1]+(-1.17323024801)*x[2]
        arg[3,0,0]=(-1.4325812824)*x[0]+(-0.733085265902)*x[1]+(0.125883186484)*x[2]
        arg[3,0,1]=(-0.0529889021435)*x[0]+(0.335154600942)*x[1]+(0.839337375209)*x[2]
        arg[3,1,0]=(-0.423552182963)*x[0]+(1.31813289315)*x[1]+(0.13235854915)*x[2]
        arg[3,1,1]=(0.53278572217)*x[0]+(0.159434073071)*x[1]+(-0.313621199136)*x[2]
        arg[4,0,0]=(0.747779887673)*x[0]+(0.30977483911)*x[1]+(-0.0760295896739)*x[2]
        arg[4,0,1]=(0.00295118909194)*x[0]+(-0.687914867842)*x[1]+(0.0386268422915)*x[2]
        arg[4,1,0]=(-0.276363532754)*x[0]+(0.52102466785)*x[1]+(-0.684840191917)*x[2]
        arg[4,1,1]=(0.0283138898538)*x[0]+(0.116042458294)*x[1]+(0.485578401528)*x[2]
        arg[5,0,0]=(-0.118402060192)*x[0]+(-1.18409809944)*x[1]+(-0.304048852342)*x[2]
        arg[5,0,1]=(0.174393790186)*x[0]+(-0.524912689263)*x[1]+(-0.648708392355)*x[2]
        arg[5,1,0]=(-0.254958756344)*x[0]+(-0.772071348728)*x[1]+(-0.0572494064027)*x[2]
        arg[5,1,1]=(-0.74777072018)*x[0]+(0.19109939364)*x[1]+(0.498973987814)*x[2]
        ref[0,0,0]=(0.986457444152)/(o+1.)+(0.258994989412)+(0.0802852749346)*0.5**o
        ref[0,0,1]=(-0.0286717143415)/(o+1.)+(0.554867138611)+(-0.115397940627)*0.5**o
        ref[0,1,0]=(-0.886482131138)/(o+1.)+(0.464293129749)+(0.125333386946)*0.5**o
        ref[0,1,1]=(0.199214164457)/(o+1.)+(0.336882900218)+(-0.472396346443)*0.5**o
        ref[1,0,0]=(0.36021680954)/(o+1.)+(0.408903924703)+(-0.780817864055)*0.5**o
        ref[1,0,1]=(0.44795581074)/(o+1.)+(0.472159267742)+(0.391128251477)*0.5**o
        ref[1,1,0]=(1.02665734893)/(o+1.)+(-0.71418625728)+(-0.898451524719)*0.5**o
        ref[1,1,1]=(0.0322676607777)/(o+1.)+(0.105378958385)+(-0.0896073626306)*0.5**o
        ref[2,0,0]=(1.15723247637)/(o+1.)+(-0.719604538702)+(-0.86439896505)*0.5**o
        ref[2,0,1]=(-0.254100181576)/(o+1.)+(-0.0387997401585)+(0.35059314866)*0.5**o
        ref[2,1,0]=(-0.636439167982)/(o+1.)+(0.563390681981)+(-0.0355507351918)*0.5**o
        ref[2,1,1]=(-0.265265225142)/(o+1.)+(-0.635620798949)+(-0.845072369786)*0.5**o
        ref[3,0,0]=(-0.062162800164)/(o+1.)+(-0.554499928282)+(-0.868620705087)*0.5**o
        ref[3,0,1]=(0.784681961813)/(o+1.)+(0.355202655131)+(-0.373584198067)*0.5**o
        ref[3,1,0]=(1.20272043429)/(o+1.)+(0.352677694103)+(-0.881136563162)*0.5**o
        ref[3,1,1]=(-0.311152797156)/(o+1.)+(0.313736563953)+(0.0622782653549)*0.5**o
        ref[4,0,0]=(-0.365517620238)/(o+1.)+(0.733502174253)+(-0.119961591158)*0.5**o
        ref[4,0,1]=(0.0155845896619)/(o+1.)+(-0.654963543839)+(0.648005661558)*0.5**o
        ref[4,1,0]=(0.405697975915)/(o+1.)+(-0.637805133954)+(0.429733235172)*0.5**o
        ref[4,1,1]=(-0.0469738326054)/(o+1.)+(0.0590721074184)+(0.558764367445)*0.5**o
        ref[5,0,0]=(-0.333825007954)/(o+1.)+(-0.46448520882)+(-0.343753586377)*0.5**o
        ref[5,0,1]=(-0.469125954795)/(o+1.)+(-0.147192309255)+(-0.235716718127)*0.5**o
        ref[5,1,0]=(-0.573275891777)/(o+1.)+(-0.478328163121)+(0.445652706545)*0.5**o
        ref[5,1,1]=(0.0545515234726)/(o+1.)+(-0.153919681346)+(0.195590500493)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.226142137728)*x[0]+(0.85681895183)*x[1]
        arg[0,0,0,1]=(-0.824407181762)*x[0]+(-0.290023144892)*x[1]
        arg[0,0,1,0]=(-1.25987816017)*x[0]+(0.824340990595)*x[1]
        arg[0,0,1,1]=(-0.0917811606765)*x[0]+(0.660211932964)*x[1]
        arg[0,0,2,0]=(-1.31073149178)*x[0]+(0.482324984918)*x[1]
        arg[0,0,2,1]=(0.256995458227)*x[0]+(0.79244254852)*x[1]
        arg[0,1,0,0]=(-0.0682147475407)*x[0]+(0.243620052061)*x[1]
        arg[0,1,0,1]=(0.071099982687)*x[0]+(-0.62986754897)*x[1]
        arg[0,1,1,0]=(-0.0716793733131)*x[0]+(-1.15235525814)*x[1]
        arg[0,1,1,1]=(0.193107575216)*x[0]+(-0.596149957543)*x[1]
        arg[0,1,2,0]=(-0.160142629886)*x[0]+(1.36581707067)*x[1]
        arg[0,1,2,1]=(-1.21477648376)*x[0]+(-0.732000298369)*x[1]
        arg[0,2,0,0]=(-1.02468971059)*x[0]+(0.330721595217)*x[1]
        arg[0,2,0,1]=(1.11414907391)*x[0]+(1.29386127812)*x[1]
        arg[0,2,1,0]=(1.1485850154)*x[0]+(0.387585977495)*x[1]
        arg[0,2,1,1]=(0.237382583072)*x[0]+(1.56248221279)*x[1]
        arg[0,2,2,0]=(-0.949129049229)*x[0]+(0.578823379063)*x[1]
        arg[0,2,2,1]=(-1.0667182537)*x[0]+(0.462227485959)*x[1]
        arg[0,3,0,0]=(1.37954377554)*x[0]+(0.541856880139)*x[1]
        arg[0,3,0,1]=(0.401584700319)*x[0]+(0.577102327147)*x[1]
        arg[0,3,1,0]=(-0.289796408108)*x[0]+(0.807853785325)*x[1]
        arg[0,3,1,1]=(0.0038467861822)*x[0]+(-0.931311510771)*x[1]
        arg[0,3,2,0]=(0.859036254821)*x[0]+(0.198884218516)*x[1]
        arg[0,3,2,1]=(-0.00310256070968)*x[0]+(0.511838475968)*x[1]
        arg[0,4,0,0]=(0.539686886802)*x[0]+(-0.324612924267)*x[1]
        arg[0,4,0,1]=(-0.0458655898565)*x[0]+(0.211810380966)*x[1]
        arg[0,4,1,0]=(0.302333533003)*x[0]+(0.0663033316831)*x[1]
        arg[0,4,1,1]=(-0.981519161639)*x[0]+(-0.00207756909914)*x[1]
        arg[0,4,2,0]=(0.283659447004)*x[0]+(-0.231526592979)*x[1]
        arg[0,4,2,1]=(-1.17679167875)*x[0]+(0.285698985756)*x[1]
        arg[1,0,0,0]=(-0.162900059582)*x[0]+(-0.877164340783)*x[1]
        arg[1,0,0,1]=(-0.50468605065)*x[0]+(-0.160440073709)*x[1]
        arg[1,0,1,0]=(-0.100901676644)*x[0]+(-0.296111526029)*x[1]
        arg[1,0,1,1]=(0.176107421869)*x[0]+(1.74393708745)*x[1]
        arg[1,0,2,0]=(-0.979807721703)*x[0]+(-0.229400841067)*x[1]
        arg[1,0,2,1]=(0.217472795645)*x[0]+(0.0881503243323)*x[1]
        arg[1,1,0,0]=(-0.127980123625)*x[0]+(0.0932095456541)*x[1]
        arg[1,1,0,1]=(0.829674410931)*x[0]+(-0.796196041601)*x[1]
        arg[1,1,1,0]=(1.28656110233)*x[0]+(-1.27850471392)*x[1]
        arg[1,1,1,1]=(1.159269884)*x[0]+(-0.973135433631)*x[1]
        arg[1,1,2,0]=(1.46274837856)*x[0]+(-1.81554270863)*x[1]
        arg[1,1,2,1]=(0.0973472511688)*x[0]+(0.551451261495)*x[1]
        arg[1,2,0,0]=(0.510667671293)*x[0]+(-0.51075982099)*x[1]
        arg[1,2,0,1]=(-1.41612093451)*x[0]+(0.614417669629)*x[1]
        arg[1,2,1,0]=(1.46595125796)*x[0]+(0.60374279122)*x[1]
        arg[1,2,1,1]=(-0.317185218187)*x[0]+(-0.0785532882696)*x[1]
        arg[1,2,2,0]=(-0.388504041993)*x[0]+(-0.838003398775)*x[1]
        arg[1,2,2,1]=(0.0497207129012)*x[0]+(-0.622155692247)*x[1]
        arg[1,3,0,0]=(-0.0471400179563)*x[0]+(1.3378948207)*x[1]
        arg[1,3,0,1]=(-1.22250822195)*x[0]+(0.629714783138)*x[1]
        arg[1,3,1,0]=(-1.28801080408)*x[0]+(-1.54079710038)*x[1]
        arg[1,3,1,1]=(-0.426412448469)*x[0]+(-1.09410474021)*x[1]
        arg[1,3,2,0]=(0.022834167351)*x[0]+(-0.497423389788)*x[1]
        arg[1,3,2,1]=(-0.787492124301)*x[0]+(-0.245621168419)*x[1]
        arg[1,4,0,0]=(-1.37814880919)*x[0]+(-0.704805848451)*x[1]
        arg[1,4,0,1]=(-1.12494770055)*x[0]+(1.44920024305)*x[1]
        arg[1,4,1,0]=(-0.809532572761)*x[0]+(1.16638587657)*x[1]
        arg[1,4,1,1]=(0.931784486816)*x[0]+(1.34629889453)*x[1]
        arg[1,4,2,0]=(-0.172430942793)*x[0]+(-0.584736998321)*x[1]
        arg[1,4,2,1]=(0.00542948736389)*x[0]+(0.43917597284)*x[1]
        arg[2,0,0,0]=(-0.946122337406)*x[0]+(-0.161940168137)*x[1]
        arg[2,0,0,1]=(-0.7856653779)*x[0]+(-0.75287701788)*x[1]
        arg[2,0,1,0]=(-1.08858367592)*x[0]+(-0.530380818694)*x[1]
        arg[2,0,1,1]=(-1.0573555095)*x[0]+(-0.854148765785)*x[1]
        arg[2,0,2,0]=(0.723412825328)*x[0]+(0.326761950437)*x[1]
        arg[2,0,2,1]=(-0.0143142161225)*x[0]+(-0.752724311876)*x[1]
        arg[2,1,0,0]=(-0.942548092355)*x[0]+(0.700861094897)*x[1]
        arg[2,1,0,1]=(-0.712286567957)*x[0]+(0.230406545738)*x[1]
        arg[2,1,1,0]=(1.27414839774)*x[0]+(0.724784649037)*x[1]
        arg[2,1,1,1]=(0.637494163002)*x[0]+(0.442489599192)*x[1]
        arg[2,1,2,0]=(1.81617631726)*x[0]+(1.56870218809)*x[1]
        arg[2,1,2,1]=(1.16419900666)*x[0]+(-0.044314729141)*x[1]
        arg[2,2,0,0]=(0.880267125196)*x[0]+(0.77124406259)*x[1]
        arg[2,2,0,1]=(0.482402867542)*x[0]+(0.0746924305017)*x[1]
        arg[2,2,1,0]=(0.105861259696)*x[0]+(0.484292207267)*x[1]
        arg[2,2,1,1]=(-0.107785838196)*x[0]+(0.701132243209)*x[1]
        arg[2,2,2,0]=(0.559717405844)*x[0]+(-1.5585942717)*x[1]
        arg[2,2,2,1]=(-0.989130712605)*x[0]+(0.221361137556)*x[1]
        arg[2,3,0,0]=(-0.045471818095)*x[0]+(-0.437941959792)*x[1]
        arg[2,3,0,1]=(1.25840468379)*x[0]+(-1.18872011057)*x[1]
        arg[2,3,1,0]=(1.70665829408)*x[0]+(0.705273921003)*x[1]
        arg[2,3,1,1]=(-0.0928709253845)*x[0]+(0.785115449756)*x[1]
        arg[2,3,2,0]=(-0.887243418838)*x[0]+(1.06790660642)*x[1]
        arg[2,3,2,1]=(-0.572644686872)*x[0]+(-0.672831948394)*x[1]
        arg[2,4,0,0]=(1.28282566801)*x[0]+(0.651985199447)*x[1]
        arg[2,4,0,1]=(0.0965663229335)*x[0]+(1.33819998052)*x[1]
        arg[2,4,1,0]=(0.205569041098)*x[0]+(-0.873095358733)*x[1]
        arg[2,4,1,1]=(-1.04452025745)*x[0]+(-0.204917033907)*x[1]
        arg[2,4,2,0]=(-0.181163938972)*x[0]+(0.0341993409272)*x[1]
        arg[2,4,2,1]=(1.68327027059)*x[0]+(-0.289436355642)*x[1]
        arg[3,0,0,0]=(1.05712810731)*x[0]+(0.915711819735)*x[1]
        arg[3,0,0,1]=(-0.171072039482)*x[0]+(-0.160656943623)*x[1]
        arg[3,0,1,0]=(-1.27063799869)*x[0]+(-0.605131642813)*x[1]
        arg[3,0,1,1]=(0.940512990835)*x[0]+(-0.116675577531)*x[1]
        arg[3,0,2,0]=(-0.194213869952)*x[0]+(0.364066752025)*x[1]
        arg[3,0,2,1]=(-0.328780465397)*x[0]+(-0.114689413673)*x[1]
        arg[3,1,0,0]=(-0.825914305895)*x[0]+(-0.167010056731)*x[1]
        arg[3,1,0,1]=(0.657779056207)*x[0]+(0.0666252451638)*x[1]
        arg[3,1,1,0]=(-0.464004426036)*x[0]+(-0.340932918409)*x[1]
        arg[3,1,1,1]=(-1.6698652667)*x[0]+(0.731870373141)*x[1]
        arg[3,1,2,0]=(1.39626250779)*x[0]+(-0.879152855557)*x[1]
        arg[3,1,2,1]=(1.34973919508)*x[0]+(-0.129753332761)*x[1]
        arg[3,2,0,0]=(1.03443450706)*x[0]+(0.807333477811)*x[1]
        arg[3,2,0,1]=(0.366011302596)*x[0]+(0.543804800926)*x[1]
        arg[3,2,1,0]=(0.262446846036)*x[0]+(-0.247786850991)*x[1]
        arg[3,2,1,1]=(-1.04657091196)*x[0]+(-0.485443426155)*x[1]
        arg[3,2,2,0]=(0.342434327538)*x[0]+(-0.148503585588)*x[1]
        arg[3,2,2,1]=(-0.189833429716)*x[0]+(0.424391874819)*x[1]
        arg[3,3,0,0]=(-0.118348838118)*x[0]+(0.397628264341)*x[1]
        arg[3,3,0,1]=(0.481440510084)*x[0]+(-0.393576221424)*x[1]
        arg[3,3,1,0]=(0.907533552208)*x[0]+(-0.668646549376)*x[1]
        arg[3,3,1,1]=(-1.49731714213)*x[0]+(-1.18710678264)*x[1]
        arg[3,3,2,0]=(-0.460660079207)*x[0]+(1.48078266444)*x[1]
        arg[3,3,2,1]=(-0.161073556988)*x[0]+(-0.486827042754)*x[1]
        arg[3,4,0,0]=(-0.668755455623)*x[0]+(0.345804059434)*x[1]
        arg[3,4,0,1]=(1.00528889193)*x[0]+(0.978507928191)*x[1]
        arg[3,4,1,0]=(0.98955856331)*x[0]+(1.0421852144)*x[1]
        arg[3,4,1,1]=(0.711688260873)*x[0]+(1.39962222322)*x[1]
        arg[3,4,2,0]=(1.08512429209)*x[0]+(-0.555249097743)*x[1]
        arg[3,4,2,1]=(-1.10443784007)*x[0]+(-0.782274429169)*x[1]
        ref[0,0,0,0]=(0.151153165564)/(o+1.)+(0.112609656739)+(0.254304335059)*0.5**o
        ref[0,0,0,1]=(-0.530602583511)/(o+1.)+(0.1817855378)+(-0.947398818743)*0.5**o
        ref[0,0,1,0]=(0.308747437738)/(o+1.)+(-0.0338129255643)+(-0.676658756184)*0.5**o
        ref[0,0,1,1]=(0.321026553427)/(o+1.)+(0.0692705367618)+(0.108863145337)*0.5**o
        ref[0,0,2,0]=(-0.0908034666655)/(o+1.)+(-0.1100205155)+(-0.517562009199)*0.5**o
        ref[0,0,2,1]=(0.303456040903)/(o+1.)+(0.500090875336)+(-0.254199784829)*0.5**o
        ref[0,1,0,0]=(-0.440641221831)/(o+1.)+(0.456218946873)+(-0.296391367396)*0.5**o
        ref[0,1,0,1]=(0.222628561762)/(o+1.)+(-0.162064412854)+(-0.457267302337)*0.5**o
        ref[0,1,1,0]=(-0.913361887408)/(o+1.)+(-0.356168674239)+(0.401664604431)*0.5**o
        ref[0,1,1,1]=(-0.543116639248)/(o+1.)+(-0.120902439422)+(0.381879135767)*0.5**o
        ref[0,1,2,0]=(0.993649262831)/(o+1.)+(0.384695858206)+(-0.55736653846)*0.5**o
        ref[0,1,2,1]=(-0.968077797451)/(o+1.)+(-0.187765942637)+(-0.603167099406)*0.5**o
        ref[0,2,0,0]=(0.873662019918)/(o+1.)+(-0.710635360033)+(-0.146359415223)*0.5**o
        ref[0,2,0,1]=(0.466408230693)/(o+1.)+(0.77626565285)+(0.389070815631)*0.5**o
        ref[0,2,1,0]=(-0.475010813579)/(o+1.)+(0.525435880006)+(0.960310046461)*0.5**o
        ref[0,2,1,1]=(0.697127520484)/(o+1.)+(0.357204276552)+(0.388328722275)*0.5**o
        ref[0,2,2,0]=(-0.108434738487)/(o+1.)+(0.356443300466)+(-0.97475753261)*0.5**o
        ref[0,2,2,1]=(0.952274964304)/(o+1.)+(-0.497379508976)+(-0.562006714089)*0.5**o
        ref[0,3,0,0]=(-0.0368012761671)/(o+1.)+(0.594000625518)+(0.770200680808)*0.5**o
        ref[0,3,0,1]=(0.450361861836)/(o+1.)+(0.182031052848)+(0.164263059934)*0.5**o
        ref[0,3,1,0]=(0.500714741956)/(o+1.)+(0.0479711343168)+(-0.0785996333735)*0.5**o
        ref[0,3,1,1]=(-0.949280938645)/(o+1.)+(0.452441874504)+(-0.88306753495)*0.5**o
        ref[0,3,2,0]=(0.186931712837)/(o+1.)+(0.222335175903)+(0.426318408694)*0.5**o
        ref[0,3,2,1]=(-0.288015428918)/(o+1.)+(0.80948657963)+(-0.822221815085)*0.5**o
        ref[0,4,0,0]=(0.670151184695)/(o+1.)+(-0.415883847013)+(0.376690471866)*0.5**o
        ref[0,4,0,1]=(-0.107106843992)/(o+1.)+(0.555375298435)+(-0.837698961768)*0.5**o
        ref[0,4,1,0]=(-0.299999096802)/(o+1.)+(0.668996782774)+(-0.669357604061)*0.5**o
        ref[0,4,1,1]=(0.778006183401)/(o+1.)+(-0.614473529032)+(-0.532655856076)*0.5**o
        ref[0,4,2,0]=(-0.250068389818)/(o+1.)+(0.156274314197)+(-0.0103473845504)*0.5**o
        ref[0,4,2,1]=(0.902125322216)/(o+1.)+(-0.616424567683)+(-0.560368879846)*0.5**o
        ref[1,0,0,0]=(-0.133690258295)/(o+1.)+(-0.715949360249)+(0.525524578427)*0.5**o
        ref[1,0,0,1]=(-0.788097689216)/(o+1.)+(0.0998405659256)+(-0.076709566994)*0.5**o
        ref[1,0,1,0]=(0.144280886187)/(o+1.)+(0.0775377290314)+(-0.696369546924)*0.5**o
        ref[1,0,1,1]=(0.824172196605)/(o+1.)+(0.561804029444)+(-0.0277357461773)*0.5**o
        ref[1,0,2,0]=(0.422564421802)/(o+1.)+(-0.60719307099)+(-0.417386842591)*0.5**o
        ref[1,0,2,1]=(0.778908672762)/(o+1.)+(0.0163446340765)+(-0.505974820938)*0.5**o
        ref[1,1,0,0]=(0.0574071047387)/(o+1.)+(-0.350514214742)+(0.608850746775)*0.5**o
        ref[1,1,0,1]=(-0.991660961792)/(o+1.)+(0.322084279455)+(0.380970772212)*0.5**o
        ref[1,1,1,0]=(-0.823761120268)/(o+1.)+(-0.0495419092898)+(0.930901327253)*0.5**o
        ref[1,1,1,1]=(-0.384729980569)/(o+1.)+(-0.0392583789163)+(0.649381188769)*0.5**o
        ref[1,1,2,0]=(-0.840365304069)/(o+1.)+(-0.0150512662777)+(0.517673506558)*0.5**o
        ref[1,1,2,1]=(0.555976660846)/(o+1.)+(-0.307265886898)+(0.707353625614)*0.5**o
        ref[1,2,0,0]=(-0.423056705491)/(o+1.)+(0.0559404393793)+(0.311083677036)*0.5**o
        ref[1,2,0,1]=(0.15409797664)/(o+1.)+(-0.0906637067519)+(-0.774473828018)*0.5**o
        ref[1,2,1,0]=(0.796718644088)/(o+1.)+(0.248012077098)+(0.776951250898)*0.5**o
        ref[1,2,1,1]=(-0.342830889263)/(o+1.)+(-0.282162737841)+(0.511417858487)*0.5**o
        ref[1,2,2,0]=(-0.722403016671)/(o+1.)+(-0.367059457954)+(0.230014491812)*0.5**o
        ref[1,2,2,1]=(-0.773685058881)/(o+1.)+(0.412807900451)+(-0.624365721366)*0.5**o
        ref[1,3,0,0]=(0.527481434758)/(o+1.)+(0.36602961117)+(0.031214145649)*0.5**o
        ref[1,3,0,1]=(0.0630927531585)/(o+1.)+(-0.164989701995)+(-0.325906787975)*0.5**o
        ref[1,3,1,0]=(-0.612390590593)/(o+1.)+(-0.78855996274)+(-0.63929738839)*0.5**o
        ref[1,3,1,1]=(-0.962656044265)/(o+1.)+(-0.00576535037253)+(-0.546330443673)*0.5**o
        ref[1,3,2,0]=(-0.947280922008)/(o+1.)+(-0.137449579996)+(0.747590859563)*0.5**o
        ref[1,3,2,1]=(-0.946623026202)/(o+1.)+(0.0787536550213)+(-0.24399757656)*0.5**o
        ref[1,4,0,0]=(-0.174471016943)/(o+1.)+(-0.636500556039)+(-0.635482528623)*0.5**o
        ref[1,4,0,1]=(0.806379391595)/(o+1.)+(-0.00304106577093)+(-0.476044717545)*0.5**o
        ref[1,4,1,0]=(0.729649125134)/(o+1.)+(0.0151477051616)+(-0.403091231646)*0.5**o
        ref[1,4,1,1]=(0.559976454745)/(o+1.)+(0.479295896421)+(0.75951513376)*0.5**o
        ref[1,4,2,0]=(0.272199270818)/(o+1.)+(-0.0929072274332)+(-0.843552757065)*0.5**o
        ref[1,4,2,1]=(-0.552663678377)/(o+1.)+(0.613784745778)+(-0.230300352975)*0.5**o
        ref[2,0,0,0]=(-0.174247908098)/(o+1.)+(-0.108873783282)+(-0.716067030882)*0.5**o
        ref[2,0,0,1]=(-0.300584841532)/(o+1.)+(-0.485363665493)+(-0.267230223262)*0.5**o
        ref[2,0,1,0]=(0.316261805154)/(o+1.)+(-0.523474106451)+(-0.888278086869)*0.5**o
        ref[2,0,1,1]=(-0.699622562861)/(o+1.)+(-0.151089538248)+(-0.909702635929)*0.5**o
        ref[2,0,2,0]=(0.728545639829)/(o+1.)+(0.192083148877)+(-0.0625371618174)*0.5**o
        ref[2,0,2,1]=(-0.22042267239)/(o+1.)+(-0.611974599324)+(0.677333343041)*0.5**o
        ref[2,1,0,0]=(-0.0734883811113)/(o+1.)+(0.36100831022)+(-0.890215236787)*0.5**o
        ref[2,1,0,1]=(0.121656829057)/(o+1.)+(-0.130801635799)+(-0.341933579678)*0.5**o
        ref[2,1,1,0]=(0.221353168629)/(o+1.)+(0.618635673483)+(0.540308531179)*0.5**o
        ref[2,1,1,1]=(0.0152580531273)/(o+1.)+(0.252173415542)+(0.560378877984)*0.5**o
        ref[2,1,2,0]=(0.580319777731)/(o+1.)+(0.90566612137)+(0.993226484879)*0.5**o
        ref[2,1,2,1]=(-0.710250320949)/(o+1.)+(0.742177021871)+(0.345780554724)*0.5**o
        ref[2,2,0,0]=(0.770701563507)/(o+1.)+(0.240823471881)+(0.399162680517)*0.5**o
        ref[2,2,0,1]=(0.76121761761)/(o+1.)+(0.128382656738)+(-0.460887633042)*0.5**o
        ref[2,2,1,0]=(-0.269854962057)/(o+1.)+(0.690854898294)+(-0.521701367567)*0.5**o
        ref[2,2,1,1]=(0.719518595712)/(o+1.)+(0.029596430244)+(-0.185365051187)*0.5**o
        ref[2,2,2,0]=(-0.671157761165)/(o+1.)+(-0.142688285012)+(-0.0423425346644)*0.5**o
        ref[2,2,2,1]=(-0.541156097112)/(o+1.)+(0.0765180918929)+(-0.379649661722)*0.5**o
        ref[2,3,0,0]=(-0.0604347109212)/(o+1.)+(-0.19403702073)+(-0.034905025507)*0.5**o
        ref[2,3,0,1]=(-0.793398507071)/(o+1.)+(0.255681280626)+(0.351720519037)*0.5**o
        ref[2,3,1,0]=(-0.267058413731)/(o+1.)+(0.852211426049)+(0.974567776717)*0.5**o
        ref[2,3,1,1]=(0.228396641207)/(o+1.)+(0.322922896735)+(-0.181997910305)*0.5**o
        ref[2,3,2,0]=(0.480987773937)/(o+1.)+(-0.197709028569)+(0.0950934707835)*0.5**o
        ref[2,3,2,1]=(-0.648934475027)/(o+1.)+(-0.418383297875)+(0.240224435511)*0.5**o
        ref[2,4,0,0]=(0.430020261528)/(o+1.)+(0.53421018733)+(0.436370231269)*0.5**o
        ref[2,4,0,1]=(0.906390660223)/(o+1.)+(-0.056057545856)+(0.640490734948)*0.5**o
        ref[2,4,1,0]=(-0.732099770943)/(o+1.)+(-0.0485027617732)+(0.161578976854)*0.5**o
        ref[2,4,1,1]=(-0.827495680848)/(o+1.)+(0.285044728936)+(-0.992031068384)*0.5**o
        ref[2,4,2,0]=(-0.919905471787)/(o+1.)+(0.656971386264)+(-0.541001898785)*0.5**o
        ref[2,4,2,1]=(0.161975875302)/(o+1.)+(0.272665206557)+(0.686527626533)*0.5**o
        ref[3,0,0,0]=(0.584120963259)/(o+1.)+(0.207930729804)+(0.972857504177)*0.5**o
        ref[3,0,0,1]=(0.315823141491)/(o+1.)+(-0.217561208386)+(-0.212429707823)*0.5**o
        ref[3,0,1,0]=(-0.732294996139)/(o+1.)+(-0.278653143596)+(-0.586168358175)*0.5**o
        ref[3,0,1,1]=(0.481321847225)/(o+1.)+(-0.070080648792)+(0.482676863662)*0.5**o
        ref[3,0,2,0]=(0.616679859076)/(o+1.)+(-0.418784546677)+(0.390742116352)*0.5**o
        ref[3,0,2,1]=(0.866778417448)/(o+1.)+(-0.686585656643)+(0.0629230167692)*0.5**o
        ref[3,1,0,0]=(0.531890054167)/(o+1.)+(-0.516187607633)+(-0.492439201527)*0.5**o
        ref[3,1,0,1]=(-0.512361639413)/(o+1.)+(0.499557182545)+(0.237651575694)*0.5**o
        ref[3,1,1,0]=(-0.67542965405)/(o+1.)+(0.274414249559)+(-0.678336189514)*0.5**o
        ref[3,1,1,1]=(0.371936242507)/(o+1.)+(-0.1636482466)+(-0.982634642862)*0.5**o
        ref[3,1,2,0]=(-0.799408831501)/(o+1.)+(0.458633399549)+(0.399251684639)*0.5**o
        ref[3,1,2,1]=(-0.0145253903686)/(o+1.)+(0.238472345905)+(0.757566560882)*0.5**o
        ref[3,2,0,0]=(-0.130519473112)/(o+1.)+(0.695277604837)+(0.581732248306)*0.5**o
        ref[3,2,0,1]=(0.632010536438)/(o+1.)+(0.123066205529)+(0.0316731560259)*0.5**o
        ref[3,2,1,0]=(-0.709926210356)/(o+1.)+(0.0926699972302)+(0.53924621094)*0.5**o
        ref[3,2,1,1]=(0.428094002247)/(o+1.)+(-0.841997926131)+(-0.276112488096)*0.5**o
        ref[3,2,2,0]=(0.708481148086)/(o+1.)+(-0.50498161534)+(0.495412824545)*0.5**o
        ref[3,2,2,1]=(0.782824622703)/(o+1.)+(-0.151509698687)+(-0.245246780227)*0.5**o
        ref[3,3,0,0]=(0.848460868418)/(o+1.)+(-0.104639248461)+(-0.359902945273)*0.5**o
        ref[3,3,0,1]=(0.0569151307298)/(o+1.)+(-0.0674224535469)+(0.165794065024)*0.5**o
        ref[3,3,1,0]=(-0.165683739908)/(o+1.)+(-0.183388699236)+(0.771348141212)*0.5**o
        ref[3,3,1,1]=(-0.932550058069)/(o+1.)+(-0.47828109001)+(-0.795311686688)*0.5**o
        ref[3,3,2,0]=(0.821542940923)/(o+1.)+(0.166853194487)+(-0.13512674466)*0.5**o
        ref[3,3,2,1]=(-0.46375871628)/(o+1.)+(0.132701267732)+(-0.449544418926)*0.5**o
        ref[3,4,0,0]=(-0.074530542168)/(o+1.)+(-0.0567877963797)+(-0.134845261261)*0.5**o
        ref[3,4,0,1]=(0.238671520479)/(o+1.)+(0.50689253572)+(0.731340228204)*0.5**o
        ref[3,4,1,0]=(0.981601513335)/(o+1.)+(0.400717209822)+(0.248707844726)*0.5**o
        ref[3,4,1,1]=(0.592304390107)/(o+1.)+(0.342408637328)+(0.834188819333)*0.5**o
        ref[3,4,2,0]=(-0.339991330909)/(o+1.)+(-0.0145922376156)+(0.899051000485)*0.5**o
        ref[3,4,2,1]=(-0.626610688164)/(o+1.)+(-0.417008165115)+(-0.426085250846)*0.5**o
      else:
        arg[0,0,0,0]=(0.697666818785)*x[0]+(1.00202295791)*x[1]+(0.617700157184)*x[2]
        arg[0,0,0,1]=(-0.465577819457)*x[0]+(-1.40054693355)*x[1]+(1.16912653295)*x[2]
        arg[0,0,1,0]=(-1.26696500634)*x[0]+(-0.0469756169975)*x[1]+(-0.445591738991)*x[2]
        arg[0,0,1,1]=(-1.61654321709)*x[0]+(-1.69522962467)*x[1]+(-1.06089094851)*x[2]
        arg[0,0,2,0]=(0.229468378506)*x[0]+(0.716813596991)*x[1]+(-0.237703937029)*x[2]
        arg[0,0,2,1]=(-1.12052051905)*x[0]+(0.835000004618)*x[1]+(-0.853004812404)*x[2]
        arg[0,1,0,0]=(-1.75500745705)*x[0]+(0.655055836882)*x[1]+(1.05207750735)*x[2]
        arg[0,1,0,1]=(-0.0746024824253)*x[0]+(-0.421596816958)*x[1]+(-0.375715855989)*x[2]
        arg[0,1,1,0]=(-0.963269684396)*x[0]+(0.431993248836)*x[1]+(-1.63061928649)*x[2]
        arg[0,1,1,1]=(-0.418757265018)*x[0]+(-0.665626168583)*x[1]+(-0.335734432507)*x[2]
        arg[0,1,2,0]=(0.980346404307)*x[0]+(1.45442049802)*x[1]+(-0.833420308612)*x[2]
        arg[0,1,2,1]=(-1.52536017847)*x[0]+(0.99484028681)*x[1]+(0.109754321663)*x[2]
        arg[0,2,0,0]=(1.697179262)*x[0]+(-0.928732802147)*x[1]+(-0.653379513626)*x[2]
        arg[0,2,0,1]=(0.682698902948)*x[0]+(0.829307959583)*x[1]+(-1.0300112276)*x[2]
        arg[0,2,1,0]=(0.026084323266)*x[0]+(1.28865784608)*x[1]+(0.830446580645)*x[2]
        arg[0,2,1,1]=(0.486291506032)*x[0]+(-0.519637742913)*x[1]+(0.81671330629)*x[2]
        arg[0,2,2,0]=(-0.761591624634)*x[0]+(0.370951131011)*x[1]+(-0.256748479312)*x[2]
        arg[0,2,2,1]=(-0.249974326513)*x[0]+(0.267323484635)*x[1]+(-1.09642130905)*x[2]
        arg[0,3,0,0]=(0.752057563255)*x[0]+(-1.2772262175)*x[1]+(-0.264569668983)*x[2]
        arg[0,3,0,1]=(-0.173167796447)*x[0]+(0.31873208356)*x[1]+(-0.599780701112)*x[2]
        arg[0,3,1,0]=(-0.912464014941)*x[0]+(-1.56452517166)*x[1]+(1.37651660499)*x[2]
        arg[0,3,1,1]=(0.302009667854)*x[0]+(0.524213064769)*x[1]+(0.316301522683)*x[2]
        arg[0,3,2,0]=(-1.2940154673)*x[0]+(-0.900061396992)*x[1]+(0.3860029847)*x[2]
        arg[0,3,2,1]=(-0.97354131341)*x[0]+(1.08221883964)*x[1]+(-0.739323076416)*x[2]
        arg[0,4,0,0]=(1.86535317192)*x[0]+(-0.478359589546)*x[1]+(-0.798014592302)*x[2]
        arg[0,4,0,1]=(-1.08518693851)*x[0]+(-0.186048654536)*x[1]+(1.04915573267)*x[2]
        arg[0,4,1,0]=(-1.09692811055)*x[0]+(1.01042920935)*x[1]+(-0.73978089555)*x[2]
        arg[0,4,1,1]=(0.654656964663)*x[0]+(0.279204596126)*x[1]+(0.886217781361)*x[2]
        arg[0,4,2,0]=(-0.54009886291)*x[0]+(0.378037248739)*x[1]+(1.09785568886)*x[2]
        arg[0,4,2,1]=(-1.43995958913)*x[0]+(-0.720157090994)*x[1]+(0.0676362028136)*x[2]
        arg[1,0,0,0]=(1.47735412608)*x[0]+(0.342355546228)*x[1]+(-0.0134502415508)*x[2]
        arg[1,0,0,1]=(-0.419536590369)*x[0]+(-1.61108470511)*x[1]+(1.51281979355)*x[2]
        arg[1,0,1,0]=(1.31765698245)*x[0]+(-0.7499566294)*x[1]+(0.934537503478)*x[2]
        arg[1,0,1,1]=(0.688467914883)*x[0]+(-1.32681280673)*x[1]+(-0.387405705058)*x[2]
        arg[1,0,2,0]=(1.37069587556)*x[0]+(0.540830111773)*x[1]+(0.477311068263)*x[2]
        arg[1,0,2,1]=(-0.572066213761)*x[0]+(-1.11950925494)*x[1]+(1.07939217872)*x[2]
        arg[1,1,0,0]=(0.564997105326)*x[0]+(-0.73130546389)*x[1]+(0.435910937304)*x[2]
        arg[1,1,0,1]=(0.513931739674)*x[0]+(-0.944043404826)*x[1]+(0.014114227468)*x[2]
        arg[1,1,1,0]=(-0.77850597256)*x[0]+(-0.445835136484)*x[1]+(0.157478357479)*x[2]
        arg[1,1,1,1]=(0.752431176866)*x[0]+(0.659759664199)*x[1]+(-0.859515433521)*x[2]
        arg[1,1,2,0]=(0.0693158730911)*x[0]+(-0.820997827204)*x[1]+(-0.714694436829)*x[2]
        arg[1,1,2,1]=(0.604674152847)*x[0]+(0.647261183132)*x[1]+(-1.21981352028)*x[2]
        arg[1,2,0,0]=(0.0997504695984)*x[0]+(-0.271549839083)*x[1]+(-0.573467835942)*x[2]
        arg[1,2,0,1]=(0.01861845581)*x[0]+(0.157410502406)*x[1]+(0.939257868776)*x[2]
        arg[1,2,1,0]=(1.44909894631)*x[0]+(0.116038517547)*x[1]+(-0.851940489233)*x[2]
        arg[1,2,1,1]=(1.34598841007)*x[0]+(1.79163125841)*x[1]+(0.719171044245)*x[2]
        arg[1,2,2,0]=(-0.214737791448)*x[0]+(-1.85810839095)*x[1]+(-0.240776599927)*x[2]
        arg[1,2,2,1]=(-0.574707645928)*x[0]+(-0.664173831956)*x[1]+(0.310262106696)*x[2]
        arg[1,3,0,0]=(0.296213453592)*x[0]+(-0.831112482555)*x[1]+(0.130075739128)*x[2]
        arg[1,3,0,1]=(-0.369510604457)*x[0]+(0.0373909697998)*x[1]+(1.6689129783)*x[2]
        arg[1,3,1,0]=(0.0922435881419)*x[0]+(-0.280519827004)*x[1]+(0.285933839232)*x[2]
        arg[1,3,1,1]=(-0.649253282272)*x[0]+(-0.460065044123)*x[1]+(-0.0711164948063)*x[2]
        arg[1,3,2,0]=(0.793313307581)*x[0]+(-0.0587956351149)*x[1]+(0.810487259739)*x[2]
        arg[1,3,2,1]=(-1.38668597055)*x[0]+(1.51396934198)*x[1]+(-1.13918550255)*x[2]
        arg[1,4,0,0]=(1.40078167737)*x[0]+(-1.61915436345)*x[1]+(0.421710091214)*x[2]
        arg[1,4,0,1]=(-1.37982015877)*x[0]+(1.30607371738)*x[1]+(-0.608888848419)*x[2]
        arg[1,4,1,0]=(-0.558087802805)*x[0]+(-0.214169499512)*x[1]+(-0.766242546503)*x[2]
        arg[1,4,1,1]=(-1.75503436584)*x[0]+(0.183604705451)*x[1]+(-0.148071484576)*x[2]
        arg[1,4,2,0]=(0.537942069142)*x[0]+(-0.00640749928252)*x[1]+(-0.277544629657)*x[2]
        arg[1,4,2,1]=(-0.652645833371)*x[0]+(-1.33439077026)*x[1]+(-0.226308197201)*x[2]
        arg[2,0,0,0]=(1.45646289279)*x[0]+(1.40842127675)*x[1]+(1.75078281495)*x[2]
        arg[2,0,0,1]=(0.867988651995)*x[0]+(1.35061812241)*x[1]+(-1.56520776323)*x[2]
        arg[2,0,1,0]=(0.549686523522)*x[0]+(-0.528085465813)*x[1]+(-0.994978820267)*x[2]
        arg[2,0,1,1]=(0.383595996552)*x[0]+(0.612989859087)*x[1]+(-0.560401247231)*x[2]
        arg[2,0,2,0]=(-0.341259919499)*x[0]+(-0.125491675042)*x[1]+(0.545572340757)*x[2]
        arg[2,0,2,1]=(-0.414269395434)*x[0]+(-0.433969881898)*x[1]+(-1.01820199328)*x[2]
        arg[2,1,0,0]=(-0.113575280389)*x[0]+(0.326916039942)*x[1]+(-0.576072890249)*x[2]
        arg[2,1,0,1]=(-0.296473969661)*x[0]+(0.195675099431)*x[1]+(-1.72378150831)*x[2]
        arg[2,1,1,0]=(-1.31575932758)*x[0]+(-0.122557423522)*x[1]+(0.53435027695)*x[2]
        arg[2,1,1,1]=(0.299435754608)*x[0]+(0.727140413706)*x[1]+(0.848931358475)*x[2]
        arg[2,1,2,0]=(0.2618965198)*x[0]+(-0.461672573339)*x[1]+(0.923592803172)*x[2]
        arg[2,1,2,1]=(-0.326817258523)*x[0]+(-0.0363300805442)*x[1]+(1.82049875918)*x[2]
        arg[2,2,0,0]=(-0.435113320888)*x[0]+(0.0257984353786)*x[1]+(0.287846577657)*x[2]
        arg[2,2,0,1]=(1.39178068279)*x[0]+(0.136781967539)*x[1]+(-1.38161590351)*x[2]
        arg[2,2,1,0]=(-0.887874199356)*x[0]+(-0.476769153654)*x[1]+(-0.618760485829)*x[2]
        arg[2,2,1,1]=(1.58883355098)*x[0]+(-0.889469047555)*x[1]+(-0.678486755843)*x[2]
        arg[2,2,2,0]=(0.237147029365)*x[0]+(-0.509883462401)*x[1]+(0.533163151551)*x[2]
        arg[2,2,2,1]=(0.66513913749)*x[0]+(1.30317030657)*x[1]+(0.804083365617)*x[2]
        arg[2,3,0,0]=(1.01531518403)*x[0]+(1.3753053143)*x[1]+(0.000877284235964)*x[2]
        arg[2,3,0,1]=(-0.40853642543)*x[0]+(0.154794268783)*x[1]+(-1.34306671117)*x[2]
        arg[2,3,1,0]=(0.43065284359)*x[0]+(-0.28965146009)*x[1]+(0.660248828789)*x[2]
        arg[2,3,1,1]=(-1.4334922681)*x[0]+(-0.0348802858609)*x[1]+(0.502051304974)*x[2]
        arg[2,3,2,0]=(-1.17399537487)*x[0]+(0.865787043946)*x[1]+(-0.0105949677478)*x[2]
        arg[2,3,2,1]=(0.63104158794)*x[0]+(0.436196079951)*x[1]+(0.498606623061)*x[2]
        arg[2,4,0,0]=(-1.44410177354)*x[0]+(0.513406326011)*x[1]+(0.699127199447)*x[2]
        arg[2,4,0,1]=(-0.981727023077)*x[0]+(0.484182120695)*x[1]+(0.667175832045)*x[2]
        arg[2,4,1,0]=(-1.3352602588)*x[0]+(0.626269023364)*x[1]+(0.813537646695)*x[2]
        arg[2,4,1,1]=(-0.932486930114)*x[0]+(0.299448807849)*x[1]+(0.195862014797)*x[2]
        arg[2,4,2,0]=(-0.0621489644806)*x[0]+(0.131979664158)*x[1]+(0.442029269778)*x[2]
        arg[2,4,2,1]=(-0.507508608633)*x[0]+(-1.28613226533)*x[1]+(-0.697920444023)*x[2]
        arg[3,0,0,0]=(-0.984942677585)*x[0]+(-1.58913299616)*x[1]+(-0.373026784822)*x[2]
        arg[3,0,0,1]=(0.616666427113)*x[0]+(-0.625162800471)*x[1]+(-0.796307290839)*x[2]
        arg[3,0,1,0]=(-0.48469203683)*x[0]+(-0.0572105166553)*x[1]+(0.809572383581)*x[2]
        arg[3,0,1,1]=(-0.487160309892)*x[0]+(0.77646454341)*x[1]+(-0.41947544573)*x[2]
        arg[3,0,2,0]=(-1.17902121554)*x[0]+(-0.244939532033)*x[1]+(-0.282994359476)*x[2]
        arg[3,0,2,1]=(0.33515909881)*x[0]+(0.146866561705)*x[1]+(1.81486098364)*x[2]
        arg[3,1,0,0]=(0.379977901341)*x[0]+(0.834531701517)*x[1]+(-0.0242027062166)*x[2]
        arg[3,1,0,1]=(0.184318292766)*x[0]+(1.08367911805)*x[1]+(-0.164531741594)*x[2]
        arg[3,1,1,0]=(-1.00967115996)*x[0]+(0.542699640149)*x[1]+(0.707616701259)*x[2]
        arg[3,1,1,1]=(0.856948233811)*x[0]+(-1.2186550297)*x[1]+(-0.034268798159)*x[2]
        arg[3,1,2,0]=(0.259470684477)*x[0]+(-1.03586983732)*x[1]+(-0.157275979341)*x[2]
        arg[3,1,2,1]=(-0.161954890658)*x[0]+(1.65853209885)*x[1]+(0.381113456719)*x[2]
        arg[3,2,0,0]=(-0.437413004681)*x[0]+(-0.730340273929)*x[1]+(-0.687689272494)*x[2]
        arg[3,2,0,1]=(-0.336618164118)*x[0]+(-1.24388666895)*x[1]+(-0.2198689254)*x[2]
        arg[3,2,1,0]=(0.0872793087887)*x[0]+(-0.336554573215)*x[1]+(-0.0597803523262)*x[2]
        arg[3,2,1,1]=(0.192805432018)*x[0]+(0.0540495582064)*x[1]+(0.846658042071)*x[2]
        arg[3,2,2,0]=(0.00794192604414)*x[0]+(-0.33504559053)*x[1]+(0.615701149787)*x[2]
        arg[3,2,2,1]=(0.210870961503)*x[0]+(1.14164095846)*x[1]+(-0.192785742791)*x[2]
        arg[3,3,0,0]=(-0.25895162041)*x[0]+(-0.225800436488)*x[1]+(-0.198007079838)*x[2]
        arg[3,3,0,1]=(-0.150363703231)*x[0]+(1.23870429535)*x[1]+(0.28635881247)*x[2]
        arg[3,3,1,0]=(-0.865980718216)*x[0]+(-0.0849853244335)*x[1]+(-0.292164217189)*x[2]
        arg[3,3,1,1]=(-0.350465968722)*x[0]+(0.09853736936)*x[1]+(-0.476711524665)*x[2]
        arg[3,3,2,0]=(0.311396300263)*x[0]+(-0.92437612034)*x[1]+(-0.757139962928)*x[2]
        arg[3,3,2,1]=(-0.912311919837)*x[0]+(0.267793631099)*x[1]+(0.4917345401)*x[2]
        arg[3,4,0,0]=(0.613662541753)*x[0]+(0.782757341361)*x[1]+(1.04209808158)*x[2]
        arg[3,4,0,1]=(0.492828613997)*x[0]+(1.11979095664)*x[1]+(-0.843545072354)*x[2]
        arg[3,4,1,0]=(-0.51939120556)*x[0]+(0.476004484286)*x[1]+(-0.516311785432)*x[2]
        arg[3,4,1,1]=(1.00423313567)*x[0]+(0.90776890388)*x[1]+(1.34423309497)*x[2]
        arg[3,4,2,0]=(-1.44069600826)*x[0]+(-1.39515290909)*x[1]+(1.19851590247)*x[2]
        arg[3,4,2,1]=(0.391851619877)*x[0]+(0.265529120553)*x[1]+(0.996610410913)*x[2]
        ref[0,0,0,0]=(1.19330263935)/(o+1.)+(0.55394275251)+(0.0162017895041)*0.5**o
        ref[0,0,0,1]=(-0.672351783684)/(o+1.)+(-0.187889622039)+(0.351132807711)*0.5**o
        ref[0,0,1,0]=(-1.27021996264)/(o+1.)+(0.236286530101)+(-0.961885459894)*0.5**o
        ref[0,0,1,1]=(-1.43442517107)/(o+1.)+(-1.00113374859)+(-0.935971122022)*0.5**o
        ref[0,0,2,0]=(-0.434813695127)/(o+1.)+(0.606673209583)+(-0.0699546855712)*0.5**o
        ref[0,0,2,1]=(-0.0434901247948)/(o+1.)+(-0.180911877983)+(-0.733211446073)*0.5**o
        ref[0,1,0,0]=(1.41514302182)/(o+1.)+(-0.247977413325)+(-0.967062307989)*0.5**o
        ref[0,1,0,1]=(0.69128100142)/(o+1.)+(-0.621609882923)+(-0.319976390946)*0.5**o
        ref[0,1,1,0]=(-0.881870429191)/(o+1.)+(-0.654075774082)+(0.0281262553057)*0.5**o
        ref[0,1,1,1]=(-1.27239116086)/(o+1.)+(-0.0854066429492)+(0.0230865806541)*0.5**o
        ref[0,1,2,0]=(-0.222663945351)/(o+1.)+(0.685312598747)+(0.453385341573)*0.5**o
        ref[0,1,2,1]=(0.101409038706)/(o+1.)+(0.151367946744)+(-0.824910502191)*0.5**o
        ref[0,2,0,0]=(-1.19509443892)/(o+1.)+(0.225437856866)+(0.85928567141)*0.5**o
        ref[0,2,0,1]=(0.285764347754)/(o+1.)+(-0.383432127133)+(0.963095541446)*0.5**o
        ref[0,2,1,0]=(0.728010227911)/(o+1.)+(0.983790919896)+(-0.550403317717)*0.5**o
        ref[0,2,1,1]=(0.563748259036)/(o+1.)+(0.165164845927)+(-0.110710881481)*0.5**o
        ref[0,2,2,0]=(-0.361502606429)/(o+1.)+(0.0276145482611)+(-0.341115463028)*0.5**o
        ref[0,2,2,1]=(-0.293320207933)/(o+1.)+(0.0336757304385)+(-0.853103403872)*0.5**o
        ref[0,3,0,0]=(-0.410233956527)/(o+1.)+(-0.474281126245)+(0.569057885785)*0.5**o
        ref[0,3,0,1]=(-0.835861970115)/(o+1.)+(0.0173715483175)+(0.346902459481)*0.5**o
        ref[0,3,1,0]=(0.0830466456187)/(o+1.)+(-0.288204681495)+(-0.607109864241)*0.5**o
        ref[0,3,1,1]=(1.07074130823)/(o+1.)+(-0.0437983012867)+(0.159379549652)*0.5**o
        ref[0,3,2,0]=(0.172301049609)/(o+1.)+(-0.491724874152)+(-0.996925180895)*0.5**o
        ref[0,3,2,1]=(-0.24891941221)/(o+1.)+(-0.03935856269)+(-0.3030090126)*0.5**o
        ref[0,4,0,0]=(-0.778575122429)/(o+1.)+(0.213085222078)+(0.941383668349)*0.5**o
        ref[0,4,0,1]=(0.880801627174)/(o+1.)+(-0.454451723401)+(-0.193978040747)*0.5**o
        ref[0,4,1,0]=(-0.835604724882)/(o+1.)+(0.127240834083)+(-0.245156740036)*0.5**o
        ref[0,4,1,1]=(1.62336716324)/(o+1.)+(-0.101085918464)+(0.398884015839)*0.5**o
        ref[0,4,2,0]=(1.16805827702)/(o+1.)+(0.211232591688)+(-0.654729385702)*0.5**o
        ref[0,4,2,1]=(-0.305424138957)/(o+1.)+(-0.431076572693)+(-0.924903192969)*0.5**o
        ref[1,0,0,0]=(-0.481596247529)/(o+1.)+(0.769902632559)+(0.748050413171)*0.5**o
        ref[1,0,0,1]=(0.177388661252)/(o+1.)+(-0.118766442613)+(-0.457657277952)*0.5**o
        ref[1,0,1,0]=(0.709890860372)/(o+1.)+(0.0668010230727)+(0.658744950012)*0.5**o
        ref[1,0,1,1]=(-1.45685781381)/(o+1.)+(-0.178674426089)+(0.788456069079)*0.5**o
        ref[1,0,2,0]=(0.194388185612)/(o+1.)+(0.649177543063)+(0.896093783855)*0.5**o
        ref[1,0,2,1]=(0.181561279507)/(o+1.)+(-0.200618291902)+(-0.392507985678)*0.5**o
        ref[1,1,0,0]=(-0.884376814027)/(o+1.)+(0.280382736159)+(0.593213920451)*0.5**o
        ref[1,1,0,1]=(-0.250709602218)/(o+1.)+(-0.492104211211)+(0.818920586956)*0.5**o
        ref[1,1,1,0]=(0.0529723694625)/(o+1.)+(-0.502047650638)+(-0.115739819752)*0.5**o
        ref[1,1,1,1]=(0.215812202957)/(o+1.)+(-0.308796450292)+(0.954456105171)*0.5**o
        ref[1,1,2,0]=(-0.991481050902)/(o+1.)+(0.162626252838)+(-0.800147845717)*0.5**o
        ref[1,1,2,1]=(-0.433989756951)/(o+1.)+(0.362776336712)+(-0.259441100777)*0.5**o
        ref[1,2,0,0]=(-0.930287052624)/(o+1.)+(-0.347752763108)+(0.880525373413)*0.5**o
        ref[1,2,0,1]=(0.698674852861)/(o+1.)+(-0.15119613196)+(0.719004238051)*0.5**o
        ref[1,2,1,0]=(-0.777216821365)/(o+1.)+(0.436336076297)+(0.617741643399)*0.5**o
        ref[1,2,1,1]=(1.09595223135)/(o+1.)+(0.978251130983)+(0.804336219412)*0.5**o
        ref[1,2,2,0]=(-0.651012630849)/(o+1.)+(-1.04894606664)+(0.43528198181)*0.5**o
        ref[1,2,2,1]=(-1.5227353883)/(o+1.)+(0.670019307557)+(-0.745922598001)*0.5**o
        ref[1,3,0,0]=(-0.750195637997)/(o+1.)+(0.231864245348)+(-0.118356142533)*0.5**o
        ref[1,3,0,1]=(1.17136226596)/(o+1.)+(-0.114976421389)+(0.395383920468)*0.5**o
        ref[1,3,1,0]=(-0.108070636769)/(o+1.)+(0.279145429234)+(-0.352562621327)*0.5**o
        ref[1,3,1,1]=(-0.144223481714)/(o+1.)+(-0.472954352493)+(-0.0903026345013)*0.5**o
        ref[1,3,2,0]=(-0.487564307786)/(o+1.)+(0.636973351449)+(0.758622537094)*0.5**o
        ref[1,3,2,1]=(0.390860965954)/(o+1.)+(-0.36647678165)+(-0.66980953377)*0.5**o
        ref[1,4,0,0]=(-1.04824984837)/(o+1.)+(0.213932231728)+(0.823722790054)*0.5**o
        ref[1,4,0,1]=(-0.0984653601336)/(o+1.)+(0.07867877705)+(-0.741527483773)*0.5**o
        ref[1,4,1,0]=(-0.146287867383)/(o+1.)+(-0.908951868158)+(0.425691754878)*0.5**o
        ref[1,4,1,1]=(-0.94160416554)/(o+1.)+(0.0459877010016)+(-0.869872381431)*0.5**o
        ref[1,4,2,0]=(0.0115742858952)/(o+1.)+(-0.222488399041)+(0.687392452389)*0.5**o
        ref[1,4,2,1]=(-0.956561936575)/(o+1.)+(-0.59770879042)+(-0.0613652834159)*0.5**o
        ref[2,0,0,0]=(1.83987381729)/(o+1.)+(0.956579317963)+(0.862634531274)*0.5**o
        ref[2,0,0,1]=(0.105446127421)/(o+1.)+(0.180661881191)+(0.186629121375)*0.5**o
        ref[2,0,1,0]=(0.0212167440766)/(o+1.)+(-0.337825588888)+(-0.318943328859)*0.5**o
        ref[2,0,1,1]=(0.744834715631)/(o+1.)+(-0.350058416663)+(0.391466726104)*0.5**o
        ref[2,0,2,0]=(-0.788776613856)/(o+1.)+(0.233776806014)+(0.400043748043)*0.5**o
        ref[2,0,2,1]=(-0.95346976307)/(o+1.)+(0.0274998652324)+(-0.967971238003)*0.5**o
        ref[2,1,0,0]=(-0.536051408942)/(o+1.)+(-0.208498016994)+(0.590315312233)*0.5**o
        ref[2,1,0,1]=(-1.04158508788)/(o+1.)+(-0.397885793101)+(0.0127762955396)*0.5**o
        ref[2,1,1,0]=(0.529914207253)/(o+1.)+(-0.547503443987)+(-0.338873793431)*0.5**o
        ref[2,1,1,1]=(0.62298670274)/(o+1.)+(0.449860048786)+(0.352800726478)*0.5**o
        ref[2,1,2,0]=(0.965292650595)/(o+1.)+(0.00980498639)+(-0.261085873742)*0.5**o
        ref[2,1,2,1]=(0.403285360803)/(o+1.)+(0.369160592269)+(0.315744874774)*0.5**o
        ref[2,2,0,0]=(-0.457764884724)/(o+1.)+(0.563657776706)+(-0.791018976541)*0.5**o
        ref[2,2,0,1]=(0.188151274044)/(o+1.)+(-0.476470545911)+(0.911736564588)*0.5**o
        ref[2,2,1,0]=(-0.558515085739)/(o+1.)+(-0.445583796553)+(-0.533721159994)*0.5**o
        ref[2,2,1,1]=(-0.431992337265)/(o+1.)+(-0.0957726221961)+(0.64441532924)*0.5**o
        ref[2,2,2,0]=(-0.159056461947)/(o+1.)+(-0.0326864536156)+(0.484856087693)*0.5**o
        ref[2,2,2,1]=(0.533116456037)/(o+1.)+(1.10197257725)+(0.0353311991352)*0.5**o
        ref[2,3,0,0]=(0.0570625147023)/(o+1.)+(0.74170779854)+(0.851019670777)*0.5**o
        ref[2,3,0,1]=(-0.495059904832)/(o+1.)+(-0.169954912181)+(-0.761839138623)*0.5**o
        ref[2,3,1,0]=(0.482375173314)/(o+1.)+(-0.19374838651)+(0.706371811996)*0.5**o
        ref[2,3,1,1]=(0.12928100721)/(o+1.)+(-0.317327862062)+(-0.460946532077)*0.5**o
        ref[2,3,2,0]=(-0.234362084763)/(o+1.)+(0.423486279181)+(-0.93141377227)*0.5**o
        ref[2,3,2,1]=(1.25586693002)/(o+1.)+(-0.127551329993)+(0.56508002092)*0.5**o
        ref[2,4,0,0]=(0.45235297749)/(o+1.)+(-0.0217537364788)+(-0.640413752615)*0.5**o
        ref[2,4,0,1]=(0.936809207384)/(o+1.)+(-0.151519610695)+(-0.464139056331)*0.5**o
        ref[2,4,1,0]=(0.867693208863)/(o+1.)+(0.105683574314)+(-0.974513946233)*0.5**o
        ref[2,4,1,1]=(-1.27821575101)/(o+1.)+(0.463163847311)+(-0.0852880510814)*0.5**o
        ref[2,4,2,0]=(-0.846224501366)/(o+1.)+(0.432619461409)+(0.492845548002)*0.5**o
        ref[2,4,2,1]=(-1.48923229995)/(o+1.)+(-0.505747396375)+(0.00916577472118)*0.5**o
        ref[3,0,0,0]=(-0.959215731121)/(o+1.)+(-0.904609108587)+(-0.178668510268)*0.5**o
        ref[3,0,0,1]=(-1.53181231865)/(o+1.)+(-0.0758543610408)+(0.878717376536)*0.5**o
        ref[3,0,1,0]=(1.18643238831)/(o+1.)+(-0.288825159097)+(-0.341112240015)*0.5**o
        ref[3,0,1,1]=(-0.0675694917031)/(o+1.)+(-0.168955703616)+(0.275309686724)*0.5**o
        ref[3,0,2,0]=(-0.663269860133)/(o+1.)+(-0.235102421885)+(-0.573480403148)*0.5**o
        ref[3,0,2,1]=(1.0969289778)/(o+1.)+(0.46603172821)+(0.267894209932)*0.5**o
        ref[3,1,0,0]=(-0.716870513559)/(o+1.)+(0.712039409283)+(0.483098591635)*0.5**o
        ref[3,1,0,1]=(0.0328531999411)/(o+1.)+(0.379319450988)+(0.311973567308)*0.5**o
        ref[3,1,1,0]=(1.08648208064)/(o+1.)+(-0.196831959338)+(-0.452172980519)*0.5**o
        ref[3,1,1,1]=(0.0813181502612)/(o+1.)+(-0.227364064527)+(-0.0225656152548)*0.5**o
        ref[3,1,2,0]=(-0.271000454408)/(o+1.)+(-0.765711687155)+(0.868748696531)*0.5**o
        ref[3,1,2,1]=(1.60569646253)/(o+1.)+(0.137437739736)+(-0.00288127709197)*0.5**o
        ref[3,2,0,0]=(-1.25887089059)/(o+1.)+(-0.525842290945)+(0.455112921376)*0.5**o
        ref[3,2,0,1]=(-1.63787880725)/(o+1.)+(0.00214706725405)+(-0.166789085722)*0.5**o
        ref[3,2,1,0]=(-1.82111307894)/(o+1.)+(0.561806270366)+(0.388444921454)*0.5**o
        ref[3,2,1,1]=(-0.489259906414)/(o+1.)+(0.582548398381)+(0.417676141947)*0.5**o
        ref[3,2,2,0]=(-0.0674375340391)/(o+1.)+(-0.00345549227349)+(0.362946003887)*0.5**o
        ref[3,2,2,1]=(0.662780682001)/(o+1.)+(-0.15453848218)+(0.806022459534)*0.5**o
        ref[3,3,0,0]=(-1.15795406515)/(o+1.)+(0.0926725746467)+(0.289849779118)*0.5**o
        ref[3,3,0,1]=(1.27520269547)/(o+1.)+(-0.286529818361)+(0.672556345847)*0.5**o
        ref[3,3,1,0]=(-0.343290651469)/(o+1.)+(-0.0269962100904)+(-0.845847188189)*0.5**o
        ref[3,3,1,1]=(0.497715053371)/(o+1.)+(-0.158577177766)+(-0.909200821866)*0.5**o
        ref[3,3,2,0]=(-0.952431538662)/(o+1.)+(0.0816482063045)+(-0.580984656952)*0.5**o
        ref[3,3,2,1]=(-0.176227469799)/(o+1.)+(0.0456157853688)+(-0.0677878495767)*0.5**o
        ref[3,4,0,0]=(0.976460483802)/(o+1.)+(0.347595506011)+(0.766866468871)*0.5**o
        ref[3,4,0,1]=(0.0358170327662)/(o+1.)+(-0.0991456259778)+(0.931548717473)*0.5**o
        ref[3,4,1,0]=(0.114544318514)/(o+1.)+(-0.201014865792)+(-0.272213093635)*0.5**o
        ref[3,4,1,1]=(0.554509925385)/(o+1.)+(1.23219332557)+(0.237338557994)*0.5**o
        ref[3,4,2,0]=(0.0803465444931)/(o+1.)+(-0.518257569997)+(-0.681164419371)*0.5**o
        ref[3,4,2,1]=(-0.0443609570522)/(o+1.)+(0.572813368096)+(0.552725372202)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedFunctionOnContactZero(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.114918917602)*x[0]**o+(0.304834778607)*x[0]+(0.704057176139)*x[1]**o+(-0.0664624375434)*x[1]
        ref=(0.704057176139)/(o+1.)+(0.119186170532)+(-0.114918917602)*0.5**o
      else:
        arg=(-0.134961573162)*x[0]**o+(0.585433861814)*x[0]+(0.209596676649)*x[1]**o+(0.870867416311)*x[1]+(-0.902854033973)*x[2]**o+(-0.812135783993)*x[2]
        ref=(-0.693257357324)/(o+1.)+(0.322082747066)+(-0.134961573162)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedFunctionOnContactZero(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.441677901994)*x[0]**o+(0.0234076994082)*x[0]+(-0.520465368981)*x[1]**o+(-0.668717137452)*x[1]
        arg[1]=(0.0637451453664)*x[0]**o+(0.610543665309)*x[0]+(0.639008465664)*x[1]**o+(-0.460500756968)*x[1]
        ref[0]=(-0.520465368981)/(o+1.)+(-0.322654719022)+(0.441677901994)*0.5**o
        ref[1]=(0.639008465664)/(o+1.)+(0.0750214541708)+(0.0637451453664)*0.5**o
      else:
        arg[0]=(0.207042421638)*x[0]**o+(0.586067682859)*x[0]+(-0.731189752251)*x[1]**o+(0.439755444977)*x[1]+(-0.215027146515)*x[2]**o+(-0.22109699198)*x[2]
        arg[1]=(-0.482897746923)*x[0]**o+(-0.262848758384)*x[0]+(-0.130231246709)*x[1]**o+(0.494143899685)*x[1]+(0.871226664446)*x[2]**o+(-0.45136577521)*x[2]
        ref[0]=(-0.946216898767)/(o+1.)+(0.402363067928)+(0.207042421638)*0.5**o
        ref[1]=(0.740995417737)/(o+1.)+(-0.110035316955)+(-0.482897746923)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedFunctionOnContactZero(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.418217706031)*x[0]**o+(0.817101274679)*x[0]+(0.359227724634)*x[1]**o+(0.106952446872)*x[1]
        arg[0,1]=(0.735477258623)*x[0]**o+(-0.181692404492)*x[0]+(0.118248218763)*x[1]**o+(0.910033225874)*x[1]
        arg[0,2]=(-0.370301174151)*x[0]**o+(0.421589410528)*x[0]+(-0.91974297037)*x[1]**o+(0.676441845707)*x[1]
        arg[0,3]=(0.170535646217)*x[0]**o+(0.650458506701)*x[0]+(-0.431898202846)*x[1]**o+(-0.89133664224)*x[1]
        arg[0,4]=(-0.288793192482)*x[0]**o+(-0.752066662422)*x[0]+(0.279565184651)*x[1]**o+(-0.20569858505)*x[1]
        arg[1,0]=(-0.831478902729)*x[0]**o+(-0.545347589281)*x[0]+(-0.315738709444)*x[1]**o+(-0.407156023524)*x[1]
        arg[1,1]=(0.119396512736)*x[0]**o+(0.56655186098)*x[0]+(0.798206239231)*x[1]**o+(0.0342467197178)*x[1]
        arg[1,2]=(-0.264721761214)*x[0]**o+(-0.7583989175)*x[0]+(-0.350411882147)*x[1]**o+(-0.556879396863)*x[1]
        arg[1,3]=(0.976029513102)*x[0]**o+(-0.117830794452)*x[0]+(0.318067419468)*x[1]**o+(-0.496129688948)*x[1]
        arg[1,4]=(0.462797018845)*x[0]**o+(-0.214502973992)*x[0]+(0.155052684044)*x[1]**o+(0.264794267973)*x[1]
        arg[2,0]=(0.193892132778)*x[0]**o+(-0.440406597181)*x[0]+(0.527330868842)*x[1]**o+(0.856602716155)*x[1]
        arg[2,1]=(0.0418968976051)*x[0]**o+(0.944283628708)*x[0]+(0.923548919008)*x[1]**o+(0.958011594347)*x[1]
        arg[2,2]=(0.744026460715)*x[0]**o+(0.238896690247)*x[0]+(0.968891312128)*x[1]**o+(0.0636831438113)*x[1]
        arg[2,3]=(-0.289048983574)*x[0]**o+(0.897469203226)*x[0]+(0.948131139828)*x[1]**o+(0.821749060686)*x[1]
        arg[2,4]=(-0.903766144138)*x[0]**o+(-0.453139974489)*x[0]+(-0.715818671181)*x[1]**o+(-0.712367492971)*x[1]
        arg[3,0]=(0.315277263598)*x[0]**o+(-0.676815102073)*x[0]+(0.428704415911)*x[1]**o+(0.54703485039)*x[1]
        arg[3,1]=(-0.301882901745)*x[0]**o+(0.888822293672)*x[0]+(-0.496204235393)*x[1]**o+(0.210598259748)*x[1]
        arg[3,2]=(0.568071178492)*x[0]**o+(-0.552863936415)*x[0]+(0.758619958752)*x[1]**o+(0.520428540448)*x[1]
        arg[3,3]=(0.477425351692)*x[0]**o+(-0.427405588029)*x[0]+(0.197959279245)*x[1]**o+(0.328047998649)*x[1]
        arg[3,4]=(-0.869704203328)*x[0]**o+(-0.0825164966999)*x[0]+(0.779325592334)*x[1]**o+(0.419916607608)*x[1]
        ref[0,0]=(0.359227724634)/(o+1.)+(0.462026860776)+(0.418217706031)*0.5**o
        ref[0,1]=(0.118248218763)/(o+1.)+(0.364170410691)+(0.735477258623)*0.5**o
        ref[0,2]=(-0.91974297037)/(o+1.)+(0.549015628118)+(-0.370301174151)*0.5**o
        ref[0,3]=(-0.431898202846)/(o+1.)+(-0.120439067769)+(0.170535646217)*0.5**o
        ref[0,4]=(0.279565184651)/(o+1.)+(-0.478882623736)+(-0.288793192482)*0.5**o
        ref[1,0]=(-0.315738709444)/(o+1.)+(-0.476251806402)+(-0.831478902729)*0.5**o
        ref[1,1]=(0.798206239231)/(o+1.)+(0.300399290349)+(0.119396512736)*0.5**o
        ref[1,2]=(-0.350411882147)/(o+1.)+(-0.657639157182)+(-0.264721761214)*0.5**o
        ref[1,3]=(0.318067419468)/(o+1.)+(-0.3069802417)+(0.976029513102)*0.5**o
        ref[1,4]=(0.155052684044)/(o+1.)+(0.02514564699)+(0.462797018845)*0.5**o
        ref[2,0]=(0.527330868842)/(o+1.)+(0.208098059487)+(0.193892132778)*0.5**o
        ref[2,1]=(0.923548919008)/(o+1.)+(0.951147611528)+(0.0418968976051)*0.5**o
        ref[2,2]=(0.968891312128)/(o+1.)+(0.151289917029)+(0.744026460715)*0.5**o
        ref[2,3]=(0.948131139828)/(o+1.)+(0.859609131956)+(-0.289048983574)*0.5**o
        ref[2,4]=(-0.715818671181)/(o+1.)+(-0.58275373373)+(-0.903766144138)*0.5**o
        ref[3,0]=(0.428704415911)/(o+1.)+(-0.0648901258416)+(0.315277263598)*0.5**o
        ref[3,1]=(-0.496204235393)/(o+1.)+(0.54971027671)+(-0.301882901745)*0.5**o
        ref[3,2]=(0.758619958752)/(o+1.)+(-0.0162176979831)+(0.568071178492)*0.5**o
        ref[3,3]=(0.197959279245)/(o+1.)+(-0.0496787946897)+(0.477425351692)*0.5**o
        ref[3,4]=(0.779325592334)/(o+1.)+(0.168700055454)+(-0.869704203328)*0.5**o
      else:
        arg[0,0]=(0.204452789147)*x[0]**o+(0.411843111144)*x[0]+(-0.0502866367412)*x[1]**o+(0.957356261522)*x[1]+(0.173522462943)*x[2]**o+(-0.0194908382525)*x[2]
        arg[0,1]=(-0.49641528149)*x[0]**o+(0.158625922243)*x[0]+(0.670932733725)*x[1]**o+(-0.611829064862)*x[1]+(0.758373881782)*x[2]**o+(0.151236368375)*x[2]
        arg[0,2]=(0.577504765611)*x[0]**o+(0.872996499874)*x[0]+(-0.734000014722)*x[1]**o+(-0.890824538872)*x[1]+(-0.796786080271)*x[2]**o+(-0.389005562901)*x[2]
        arg[0,3]=(-0.933166619447)*x[0]**o+(-0.528627639194)*x[0]+(-0.828111326322)*x[1]**o+(0.528345547301)*x[1]+(0.99894455165)*x[2]**o+(0.472982583662)*x[2]
        arg[0,4]=(-0.495503045054)*x[0]**o+(0.629909217292)*x[0]+(0.901515784751)*x[1]**o+(-0.863668288904)*x[1]+(0.76892679355)*x[2]**o+(0.34730747389)*x[2]
        arg[1,0]=(0.303189803506)*x[0]**o+(-0.17543401741)*x[0]+(-0.283157585975)*x[1]**o+(0.950171529024)*x[1]+(0.0483247572497)*x[2]**o+(-0.604332298954)*x[2]
        arg[1,1]=(-0.225480407643)*x[0]**o+(-0.0521067696425)*x[0]+(-0.10259234894)*x[1]**o+(0.373551917005)*x[1]+(0.89047195372)*x[2]**o+(-0.292740890417)*x[2]
        arg[1,2]=(-0.476041156832)*x[0]**o+(0.916078681628)*x[0]+(-0.393887610676)*x[1]**o+(-0.651852504525)*x[1]+(0.611404175606)*x[2]**o+(0.336594647903)*x[2]
        arg[1,3]=(-0.415178048882)*x[0]**o+(0.145394549559)*x[0]+(0.680115054186)*x[1]**o+(0.261517386893)*x[1]+(0.781566117396)*x[2]**o+(-0.659634539391)*x[2]
        arg[1,4]=(0.874821341857)*x[0]**o+(0.317418381126)*x[0]+(0.051455923305)*x[1]**o+(0.670170143071)*x[1]+(-0.211902435336)*x[2]**o+(-0.175299130233)*x[2]
        arg[2,0]=(-0.372961546517)*x[0]**o+(0.199493150648)*x[0]+(-0.630239499285)*x[1]**o+(0.657412283442)*x[1]+(-0.82211915324)*x[2]**o+(-0.468612977158)*x[2]
        arg[2,1]=(-0.503168527887)*x[0]**o+(0.904406483201)*x[0]+(-0.206179628633)*x[1]**o+(0.0787026584997)*x[1]+(0.899590466467)*x[2]**o+(-0.893401768413)*x[2]
        arg[2,2]=(-0.860556408954)*x[0]**o+(0.39021140432)*x[0]+(-0.66425383352)*x[1]**o+(0.287711780417)*x[1]+(0.776884760126)*x[2]**o+(-0.115905192334)*x[2]
        arg[2,3]=(0.461359695993)*x[0]**o+(0.997877835019)*x[0]+(0.0513514112685)*x[1]**o+(-0.904812988908)*x[1]+(-0.903222331838)*x[2]**o+(-0.470185679041)*x[2]
        arg[2,4]=(-0.646560730528)*x[0]**o+(0.10038324067)*x[0]+(0.426753245517)*x[1]**o+(-0.109889624344)*x[1]+(-0.20158193753)*x[2]**o+(0.014258531224)*x[2]
        arg[3,0]=(0.364407704271)*x[0]**o+(0.840519040103)*x[0]+(0.528721642544)*x[1]**o+(-0.853679853614)*x[1]+(0.079836122575)*x[2]**o+(-0.439039178308)*x[2]
        arg[3,1]=(-0.0726491927955)*x[0]**o+(0.217804431494)*x[0]+(-0.454354285545)*x[1]**o+(0.390539936003)*x[1]+(-0.989667912688)*x[2]**o+(-0.0674079606774)*x[2]
        arg[3,2]=(0.62239422825)*x[0]**o+(0.998733716844)*x[0]+(0.796258161536)*x[1]**o+(0.618545466026)*x[1]+(-0.243188896911)*x[2]**o+(-0.677687584224)*x[2]
        arg[3,3]=(0.444073180258)*x[0]**o+(-0.764729326792)*x[0]+(-0.823491453567)*x[1]**o+(0.985393963131)*x[1]+(0.623800470977)*x[2]**o+(0.947328422513)*x[2]
        arg[3,4]=(0.686438920325)*x[0]**o+(0.789760134176)*x[0]+(-0.26011082782)*x[1]**o+(0.0692377043339)*x[1]+(-0.3712894236)*x[2]**o+(-0.542974191103)*x[2]
        ref[0,0]=(0.123235826202)/(o+1.)+(0.674854267207)+(0.204452789147)*0.5**o
        ref[0,1]=(1.42930661551)/(o+1.)+(-0.150983387122)+(-0.49641528149)*0.5**o
        ref[0,2]=(-1.53078609499)/(o+1.)+(-0.203416800949)+(0.577504765611)*0.5**o
        ref[0,3]=(0.170833225328)/(o+1.)+(0.236350245885)+(-0.933166619447)*0.5**o
        ref[0,4]=(1.6704425783)/(o+1.)+(0.0567742011389)+(-0.495503045054)*0.5**o
        ref[1,0]=(-0.234832828725)/(o+1.)+(0.0852026063299)+(0.303189803506)*0.5**o
        ref[1,1]=(0.787879604779)/(o+1.)+(0.0143521284725)+(-0.225480407643)*0.5**o
        ref[1,2]=(0.21751656493)/(o+1.)+(0.300410412503)+(-0.476041156832)*0.5**o
        ref[1,3]=(1.46168117158)/(o+1.)+(-0.126361301469)+(-0.415178048882)*0.5**o
        ref[1,4]=(-0.160446512031)/(o+1.)+(0.406144696982)+(0.874821341857)*0.5**o
        ref[2,0]=(-1.45235865253)/(o+1.)+(0.194146228466)+(-0.372961546517)*0.5**o
        ref[2,1]=(0.693410837835)/(o+1.)+(0.0448536866438)+(-0.503168527887)*0.5**o
        ref[2,2]=(0.112630926606)/(o+1.)+(0.281008996202)+(-0.860556408954)*0.5**o
        ref[2,3]=(-0.85187092057)/(o+1.)+(-0.188560416464)+(0.461359695993)*0.5**o
        ref[2,4]=(0.225171307987)/(o+1.)+(0.00237607377473)+(-0.646560730528)*0.5**o
        ref[3,0]=(0.608557765119)/(o+1.)+(-0.22609999591)+(0.364407704271)*0.5**o
        ref[3,1]=(-1.44402219823)/(o+1.)+(0.27046820341)+(-0.0726491927955)*0.5**o
        ref[3,2]=(0.553069264625)/(o+1.)+(0.469795799323)+(0.62239422825)*0.5**o
        ref[3,3]=(-0.19969098259)/(o+1.)+(0.583996529426)+(0.444073180258)*0.5**o
        ref[3,4]=(-0.631400251421)/(o+1.)+(0.158011823704)+(0.686438920325)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedFunctionOnContactZero(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.963014866001)*x[0]**o+(-0.592091539519)*x[0]+(0.397897510985)*x[1]**o+(0.485612885034)*x[1]
        arg[0,0,1]=(-0.764849027936)*x[0]**o+(0.936217923087)*x[0]+(-0.6256860991)*x[1]**o+(0.162359755936)*x[1]
        arg[0,1,0]=(-0.269607090458)*x[0]**o+(0.197100067821)*x[0]+(-0.216513043667)*x[1]**o+(0.143038940436)*x[1]
        arg[0,1,1]=(-0.432206914977)*x[0]**o+(0.177851208963)*x[0]+(0.0477652910843)*x[1]**o+(0.278333389583)*x[1]
        arg[1,0,0]=(-0.123940629628)*x[0]**o+(0.818553665062)*x[0]+(-0.168322144522)*x[1]**o+(0.485324994762)*x[1]
        arg[1,0,1]=(0.593609516922)*x[0]**o+(0.530508390905)*x[0]+(-0.938421636915)*x[1]**o+(-0.686459217917)*x[1]
        arg[1,1,0]=(-0.874022523162)*x[0]**o+(0.694500749056)*x[0]+(-0.650951516798)*x[1]**o+(0.45826621855)*x[1]
        arg[1,1,1]=(-0.259058106634)*x[0]**o+(0.0837282545694)*x[0]+(0.465734884341)*x[1]**o+(0.185558060279)*x[1]
        arg[2,0,0]=(0.346292866419)*x[0]**o+(0.982902010078)*x[0]+(0.0297606445334)*x[1]**o+(0.339043518987)*x[1]
        arg[2,0,1]=(0.749684718966)*x[0]**o+(-0.855799436867)*x[0]+(-0.186333913876)*x[1]**o+(-0.548968316245)*x[1]
        arg[2,1,0]=(-0.255759466052)*x[0]**o+(0.441882703647)*x[0]+(-0.0239807362322)*x[1]**o+(-0.236417762222)*x[1]
        arg[2,1,1]=(-0.0586694299212)*x[0]**o+(-0.912208422332)*x[0]+(-0.11610871626)*x[1]**o+(-0.512805786403)*x[1]
        arg[3,0,0]=(0.826027179564)*x[0]**o+(-0.589876767386)*x[0]+(-0.654858828463)*x[1]**o+(0.601656609023)*x[1]
        arg[3,0,1]=(-0.148714999203)*x[0]**o+(-0.409138795316)*x[0]+(0.290726299307)*x[1]**o+(-0.94009130345)*x[1]
        arg[3,1,0]=(0.63431415979)*x[0]**o+(-0.184553980231)*x[0]+(0.4087395755)*x[1]**o+(-0.694376406449)*x[1]
        arg[3,1,1]=(-0.000928514830108)*x[0]**o+(-0.782814501236)*x[0]+(0.177390244103)*x[1]**o+(0.698562703841)*x[1]
        arg[4,0,0]=(-0.939264907756)*x[0]**o+(0.191360117367)*x[0]+(-0.318750748728)*x[1]**o+(0.910713477162)*x[1]
        arg[4,0,1]=(-0.105186863713)*x[0]**o+(0.0319652211881)*x[0]+(0.118896230389)*x[1]**o+(-0.212677289177)*x[1]
        arg[4,1,0]=(-0.702838100428)*x[0]**o+(0.881695294345)*x[0]+(0.728231539571)*x[1]**o+(-0.192311604244)*x[1]
        arg[4,1,1]=(0.25197450886)*x[0]**o+(0.464857988125)*x[0]+(0.851996888259)*x[1]**o+(-0.408525083481)*x[1]
        arg[5,0,0]=(0.665627182446)*x[0]**o+(0.5367740452)*x[0]+(0.295800616406)*x[1]**o+(0.0620589956995)*x[1]
        arg[5,0,1]=(-0.37320168697)*x[0]**o+(0.072855481498)*x[0]+(0.539806161326)*x[1]**o+(-0.999091813027)*x[1]
        arg[5,1,0]=(-0.060845772709)*x[0]**o+(-0.931869215952)*x[0]+(-0.393477853475)*x[1]**o+(0.968288081996)*x[1]
        arg[5,1,1]=(-0.0768045781739)*x[0]**o+(-0.257318730887)*x[0]+(0.781473167319)*x[1]**o+(-0.458870794544)*x[1]
        ref[0,0,0]=(0.397897510985)/(o+1.)+(-0.0532393272425)+(-0.963014866001)*0.5**o
        ref[0,0,1]=(-0.6256860991)/(o+1.)+(0.549288839511)+(-0.764849027936)*0.5**o
        ref[0,1,0]=(-0.216513043667)/(o+1.)+(0.170069504128)+(-0.269607090458)*0.5**o
        ref[0,1,1]=(0.0477652910843)/(o+1.)+(0.228092299273)+(-0.432206914977)*0.5**o
        ref[1,0,0]=(-0.168322144522)/(o+1.)+(0.651939329912)+(-0.123940629628)*0.5**o
        ref[1,0,1]=(-0.938421636915)/(o+1.)+(-0.0779754135063)+(0.593609516922)*0.5**o
        ref[1,1,0]=(-0.650951516798)/(o+1.)+(0.576383483803)+(-0.874022523162)*0.5**o
        ref[1,1,1]=(0.465734884341)/(o+1.)+(0.134643157424)+(-0.259058106634)*0.5**o
        ref[2,0,0]=(0.0297606445334)/(o+1.)+(0.660972764533)+(0.346292866419)*0.5**o
        ref[2,0,1]=(-0.186333913876)/(o+1.)+(-0.702383876556)+(0.749684718966)*0.5**o
        ref[2,1,0]=(-0.0239807362322)/(o+1.)+(0.102732470713)+(-0.255759466052)*0.5**o
        ref[2,1,1]=(-0.11610871626)/(o+1.)+(-0.712507104367)+(-0.0586694299212)*0.5**o
        ref[3,0,0]=(-0.654858828463)/(o+1.)+(0.0058899208184)+(0.826027179564)*0.5**o
        ref[3,0,1]=(0.290726299307)/(o+1.)+(-0.674615049383)+(-0.148714999203)*0.5**o
        ref[3,1,0]=(0.4087395755)/(o+1.)+(-0.43946519334)+(0.63431415979)*0.5**o
        ref[3,1,1]=(0.177390244103)/(o+1.)+(-0.0421258986973)+(-0.000928514830108)*0.5**o
        ref[4,0,0]=(-0.318750748728)/(o+1.)+(0.551036797264)+(-0.939264907756)*0.5**o
        ref[4,0,1]=(0.118896230389)/(o+1.)+(-0.0903560339946)+(-0.105186863713)*0.5**o
        ref[4,1,0]=(0.728231539571)/(o+1.)+(0.344691845051)+(-0.702838100428)*0.5**o
        ref[4,1,1]=(0.851996888259)/(o+1.)+(0.0281664523221)+(0.25197450886)*0.5**o
        ref[5,0,0]=(0.295800616406)/(o+1.)+(0.29941652045)+(0.665627182446)*0.5**o
        ref[5,0,1]=(0.539806161326)/(o+1.)+(-0.463118165764)+(-0.37320168697)*0.5**o
        ref[5,1,0]=(-0.393477853475)/(o+1.)+(0.0182094330217)+(-0.060845772709)*0.5**o
        ref[5,1,1]=(0.781473167319)/(o+1.)+(-0.358094762716)+(-0.0768045781739)*0.5**o
      else:
        arg[0,0,0]=(0.638389367885)*x[0]**o+(-0.77703387527)*x[0]+(0.951496212227)*x[1]**o+(-0.484305156823)*x[1]+(-0.582437996275)*x[2]**o+(-0.460130065965)*x[2]
        arg[0,0,1]=(0.859011160063)*x[0]**o+(0.999372374168)*x[0]+(-0.946049724069)*x[1]**o+(0.522184535496)*x[1]+(0.00820405492191)*x[2]**o+(0.769305042211)*x[2]
        arg[0,1,0]=(-0.998681640062)*x[0]**o+(0.916063974594)*x[0]+(0.0801858148899)*x[1]**o+(0.475476401258)*x[1]+(-0.805247247843)*x[2]**o+(-0.809331278141)*x[2]
        arg[0,1,1]=(0.1340261052)*x[0]**o+(0.843204972218)*x[0]+(-0.809433597166)*x[1]**o+(-0.947798215446)*x[1]+(0.268796347717)*x[2]**o+(-0.896620567629)*x[2]
        arg[1,0,0]=(0.251201824008)*x[0]**o+(0.762438451805)*x[0]+(0.892722429031)*x[1]**o+(0.146944927124)*x[1]+(0.00953973207152)*x[2]**o+(0.423262196414)*x[2]
        arg[1,0,1]=(-0.258843593921)*x[0]**o+(-0.950574260099)*x[0]+(-0.49096256134)*x[1]**o+(0.446238902975)*x[1]+(-0.880706204083)*x[2]**o+(-0.400075002454)*x[2]
        arg[1,1,0]=(-0.854529145157)*x[0]**o+(0.118165533114)*x[0]+(-0.913218271084)*x[1]**o+(0.991977570414)*x[1]+(0.626241160161)*x[2]**o+(0.667843914102)*x[2]
        arg[1,1,1]=(0.636818650335)*x[0]**o+(-0.174996100829)*x[0]+(0.77317621688)*x[1]**o+(0.369280208456)*x[1]+(-0.0130124322187)*x[2]**o+(0.857098751007)*x[2]
        arg[2,0,0]=(-0.415847479141)*x[0]**o+(0.675986357495)*x[0]+(0.704926614111)*x[1]**o+(-0.166544295134)*x[1]+(-0.00917226654826)*x[2]**o+(-0.896432190739)*x[2]
        arg[2,0,1]=(-0.882383192316)*x[0]**o+(0.11122730635)*x[0]+(-0.693587089789)*x[1]**o+(-0.0905457091132)*x[1]+(-0.610526000157)*x[2]**o+(-0.992542379681)*x[2]
        arg[2,1,0]=(-0.476987072283)*x[0]**o+(-0.450296923812)*x[0]+(0.905270434608)*x[1]**o+(0.976270820516)*x[1]+(-0.233422717373)*x[2]**o+(-0.641253735386)*x[2]
        arg[2,1,1]=(0.873832514958)*x[0]**o+(-0.478059419217)*x[0]+(-0.941650455606)*x[1]**o+(0.307250152442)*x[1]+(0.865160620819)*x[2]**o+(0.961340749451)*x[2]
        arg[3,0,0]=(-0.443837945036)*x[0]**o+(-0.688773058425)*x[0]+(-0.654988618952)*x[1]**o+(-0.259511611372)*x[1]+(-0.931332326494)*x[2]**o+(0.939379631125)*x[2]
        arg[3,0,1]=(-0.181860312567)*x[0]**o+(-0.229066064127)*x[0]+(-0.359408642857)*x[1]**o+(-0.47945907585)*x[1]+(0.312345402782)*x[2]**o+(0.559298232285)*x[2]
        arg[3,1,0]=(-0.372379819015)*x[0]**o+(0.308179383085)*x[0]+(0.123613160006)*x[1]**o+(0.191944822984)*x[1]+(0.950869988808)*x[2]**o+(0.229360419984)*x[2]
        arg[3,1,1]=(-0.803126271269)*x[0]**o+(0.833324260042)*x[0]+(-0.917001746509)*x[1]**o+(-0.649776906898)*x[1]+(0.49624769362)*x[2]**o+(0.414080802594)*x[2]
        arg[4,0,0]=(-0.951747333891)*x[0]**o+(0.5919442178)*x[0]+(0.121310623341)*x[1]**o+(-0.321939845724)*x[1]+(0.358051737083)*x[2]**o+(-0.222801347833)*x[2]
        arg[4,0,1]=(-0.59462534105)*x[0]**o+(0.943860617698)*x[0]+(0.787822231831)*x[1]**o+(-0.0913435772425)*x[1]+(0.831651946729)*x[2]**o+(-0.903604465879)*x[2]
        arg[4,1,0]=(0.68655463207)*x[0]**o+(-0.98918689213)*x[0]+(0.251300089098)*x[1]**o+(-0.317828131349)*x[1]+(-0.105934216764)*x[2]**o+(-0.874506361945)*x[2]
        arg[4,1,1]=(-0.80544616293)*x[0]**o+(-0.321523684723)*x[0]+(0.874455533992)*x[1]**o+(0.200912811442)*x[1]+(-0.646776981862)*x[2]**o+(-0.988728118812)*x[2]
        arg[5,0,0]=(0.413387141328)*x[0]**o+(0.349210032211)*x[0]+(0.80870689028)*x[1]**o+(-0.633296311752)*x[1]+(0.24621029112)*x[2]**o+(0.853398479355)*x[2]
        arg[5,0,1]=(0.398787817372)*x[0]**o+(0.662875239272)*x[0]+(-0.719962467558)*x[1]**o+(0.49562860294)*x[1]+(-0.694859272211)*x[2]**o+(-0.792616661463)*x[2]
        arg[5,1,0]=(-0.298838035128)*x[0]**o+(0.121578680468)*x[0]+(-0.931062588633)*x[1]**o+(-0.0575832984853)*x[1]+(-0.696140085453)*x[2]**o+(0.836738246058)*x[2]
        arg[5,1,1]=(-0.949626960081)*x[0]**o+(-0.449843145228)*x[0]+(-0.304260073311)*x[1]**o+(-0.258880519466)*x[1]+(0.880466679751)*x[2]**o+(0.97505579364)*x[2]
        ref[0,0,0]=(0.369058215952)/(o+1.)+(-0.860734549029)+(0.638389367885)*0.5**o
        ref[0,0,1]=(-0.937845669147)/(o+1.)+(1.14543097594)+(0.859011160063)*0.5**o
        ref[0,1,0]=(-0.725061432954)/(o+1.)+(0.291104548855)+(-0.998681640062)*0.5**o
        ref[0,1,1]=(-0.540637249448)/(o+1.)+(-0.500606905428)+(0.1340261052)*0.5**o
        ref[1,0,0]=(0.902262161102)/(o+1.)+(0.666322787672)+(0.251201824008)*0.5**o
        ref[1,0,1]=(-1.37166876542)/(o+1.)+(-0.452205179789)+(-0.258843593921)*0.5**o
        ref[1,1,0]=(-0.286977110923)/(o+1.)+(0.888993508815)+(-0.854529145157)*0.5**o
        ref[1,1,1]=(0.760163784662)/(o+1.)+(0.525691429317)+(0.636818650335)*0.5**o
        ref[2,0,0]=(0.695754347563)/(o+1.)+(-0.193495064189)+(-0.415847479141)*0.5**o
        ref[2,0,1]=(-1.30411308995)/(o+1.)+(-0.485930391222)+(-0.882383192316)*0.5**o
        ref[2,1,0]=(0.671847717235)/(o+1.)+(-0.0576399193408)+(-0.476987072283)*0.5**o
        ref[2,1,1]=(-0.0764898347868)/(o+1.)+(0.395265741338)+(0.873832514958)*0.5**o
        ref[3,0,0]=(-1.58632094545)/(o+1.)+(-0.00445251933618)+(-0.443837945036)*0.5**o
        ref[3,0,1]=(-0.0470632400747)/(o+1.)+(-0.0746134538458)+(-0.181860312567)*0.5**o
        ref[3,1,0]=(1.07448314881)/(o+1.)+(0.364742313027)+(-0.372379819015)*0.5**o
        ref[3,1,1]=(-0.420754052889)/(o+1.)+(0.298814077869)+(-0.803126271269)*0.5**o
        ref[4,0,0]=(0.479362360424)/(o+1.)+(0.0236015121217)+(-0.951747333891)*0.5**o
        ref[4,0,1]=(1.61947417856)/(o+1.)+(-0.0255437127116)+(-0.59462534105)*0.5**o
        ref[4,1,0]=(0.145365872334)/(o+1.)+(-1.09076069271)+(0.68655463207)*0.5**o
        ref[4,1,1]=(0.22767855213)/(o+1.)+(-0.554669496046)+(-0.80544616293)*0.5**o
        ref[5,0,0]=(1.0549171814)/(o+1.)+(0.284656099907)+(0.413387141328)*0.5**o
        ref[5,0,1]=(-1.41482173977)/(o+1.)+(0.182943590374)+(0.398787817372)*0.5**o
        ref[5,1,0]=(-1.62720267409)/(o+1.)+(0.450366814021)+(-0.298838035128)*0.5**o
        ref[5,1,1]=(0.576206606439)/(o+1.)+(0.133166064473)+(-0.949626960081)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactZero

      assumptions: ReducedFunctionOnContactZero(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.392628958003)*x[0]**o+(0.186084293395)*x[0]+(0.965300679335)*x[1]**o+(0.0986282244747)*x[1]
        arg[0,0,0,1]=(-0.503312425492)*x[0]**o+(-0.205791466802)*x[0]+(0.384782163488)*x[1]**o+(-0.608371097114)*x[1]
        arg[0,0,1,0]=(0.660162511699)*x[0]**o+(-0.263141686561)*x[0]+(-0.765235012795)*x[1]**o+(-0.371233144611)*x[1]
        arg[0,0,1,1]=(-0.986935412732)*x[0]**o+(0.645433351743)*x[0]+(-0.451665346257)*x[1]**o+(-0.713815759872)*x[1]
        arg[0,0,2,0]=(0.632875178502)*x[0]**o+(-0.257542208821)*x[0]+(-0.695553902133)*x[1]**o+(-0.677920876645)*x[1]
        arg[0,0,2,1]=(0.102625648731)*x[0]**o+(0.511896902175)*x[0]+(0.345909088853)*x[1]**o+(-0.0609413722414)*x[1]
        arg[0,1,0,0]=(-0.0108106931425)*x[0]**o+(-0.595497517075)*x[0]+(0.332480141972)*x[1]**o+(-0.18131109924)*x[1]
        arg[0,1,0,1]=(0.220826593491)*x[0]**o+(-0.202422451222)*x[0]+(0.328480565578)*x[1]**o+(0.851432750557)*x[1]
        arg[0,1,1,0]=(0.22735402466)*x[0]**o+(0.51437836535)*x[0]+(0.292927451477)*x[1]**o+(-0.148121856023)*x[1]
        arg[0,1,1,1]=(0.495079875966)*x[0]**o+(0.582755117761)*x[0]+(-0.322061131008)*x[1]**o+(0.953437449986)*x[1]
        arg[0,1,2,0]=(0.968583889554)*x[0]**o+(-0.593253483731)*x[0]+(0.286037351205)*x[1]**o+(0.997314411263)*x[1]
        arg[0,1,2,1]=(0.361498191707)*x[0]**o+(0.633275777697)*x[0]+(0.891305263219)*x[1]**o+(-0.194468605638)*x[1]
        arg[0,2,0,0]=(0.185647566516)*x[0]**o+(0.101448861247)*x[0]+(0.0870608077911)*x[1]**o+(-0.86477908948)*x[1]
        arg[0,2,0,1]=(0.227262327826)*x[0]**o+(0.812586702875)*x[0]+(-0.0229924234496)*x[1]**o+(-0.323581477982)*x[1]
        arg[0,2,1,0]=(0.0405753322301)*x[0]**o+(0.891289727227)*x[0]+(0.957377329048)*x[1]**o+(-0.982119724379)*x[1]
        arg[0,2,1,1]=(-0.546046301701)*x[0]**o+(0.628511534929)*x[0]+(-0.628262263897)*x[1]**o+(-0.454315708999)*x[1]
        arg[0,2,2,0]=(-0.828455793175)*x[0]**o+(0.190147126828)*x[0]+(0.499595322279)*x[1]**o+(-0.0882297087937)*x[1]
        arg[0,2,2,1]=(0.926138085222)*x[0]**o+(0.402659001583)*x[0]+(-0.678432418122)*x[1]**o+(0.236054704896)*x[1]
        arg[0,3,0,0]=(-0.473066058658)*x[0]**o+(0.57622727181)*x[0]+(0.62665109546)*x[1]**o+(0.948409659533)*x[1]
        arg[0,3,0,1]=(0.776790978673)*x[0]**o+(-0.0244752180934)*x[0]+(-0.459252843816)*x[1]**o+(0.616203849528)*x[1]
        arg[0,3,1,0]=(0.34783214571)*x[0]**o+(0.806528120142)*x[0]+(0.445864469556)*x[1]**o+(-0.221961922099)*x[1]
        arg[0,3,1,1]=(0.0717966037463)*x[0]**o+(-0.434117388366)*x[0]+(0.577708844669)*x[1]**o+(-0.548519157145)*x[1]
        arg[0,3,2,0]=(-0.548573971799)*x[0]**o+(-0.926704595492)*x[0]+(-0.33936340707)*x[1]**o+(0.0687290934591)*x[1]
        arg[0,3,2,1]=(-0.28309240461)*x[0]**o+(0.12914675768)*x[0]+(-0.833323497891)*x[1]**o+(-0.577727985559)*x[1]
        arg[0,4,0,0]=(-0.105416428922)*x[0]**o+(0.41564239907)*x[0]+(0.146657173319)*x[1]**o+(-0.0141417271845)*x[1]
        arg[0,4,0,1]=(-0.876111094069)*x[0]**o+(0.491957669628)*x[0]+(0.534941970665)*x[1]**o+(-0.995507289274)*x[1]
        arg[0,4,1,0]=(-0.678996394303)*x[0]**o+(0.183159947613)*x[0]+(-0.739368558349)*x[1]**o+(0.274319401135)*x[1]
        arg[0,4,1,1]=(0.0472145581152)*x[0]**o+(-0.669737608207)*x[0]+(-0.651647023142)*x[1]**o+(0.351208329111)*x[1]
        arg[0,4,2,0]=(0.114936857967)*x[0]**o+(0.564688790898)*x[0]+(-0.524234284672)*x[1]**o+(-0.127615425724)*x[1]
        arg[0,4,2,1]=(0.568834117773)*x[0]**o+(-0.936513147599)*x[0]+(0.133122220811)*x[1]**o+(0.31510090512)*x[1]
        arg[1,0,0,0]=(-0.821241886235)*x[0]**o+(0.779734235239)*x[0]+(-0.132449009841)*x[1]**o+(-0.0179700618848)*x[1]
        arg[1,0,0,1]=(0.624372422235)*x[0]**o+(-0.752831611955)*x[0]+(-0.573597121344)*x[1]**o+(0.376463665377)*x[1]
        arg[1,0,1,0]=(-0.568749808355)*x[0]**o+(-0.697521100346)*x[0]+(-0.249786553024)*x[1]**o+(-0.624593568423)*x[1]
        arg[1,0,1,1]=(0.0765557360835)*x[0]**o+(-0.744005823966)*x[0]+(0.59246455477)*x[1]**o+(0.231310687774)*x[1]
        arg[1,0,2,0]=(0.909614864187)*x[0]**o+(0.350543076144)*x[0]+(0.0689432766231)*x[1]**o+(-0.88465698758)*x[1]
        arg[1,0,2,1]=(0.0610206203468)*x[0]**o+(0.620401328094)*x[0]+(0.378273922837)*x[1]**o+(0.91988744934)*x[1]
        arg[1,1,0,0]=(0.917249169051)*x[0]**o+(0.756656473206)*x[0]+(0.728786835113)*x[1]**o+(0.116994159781)*x[1]
        arg[1,1,0,1]=(0.331129708654)*x[0]**o+(-0.14342302905)*x[0]+(-0.418402258918)*x[1]**o+(-0.808676850609)*x[1]
        arg[1,1,1,0]=(-0.527408405279)*x[0]**o+(-0.671440239997)*x[0]+(0.989603613532)*x[1]**o+(0.364280392731)*x[1]
        arg[1,1,1,1]=(-0.189543145217)*x[0]**o+(0.14762392216)*x[0]+(0.11459047938)*x[1]**o+(-0.760477267267)*x[1]
        arg[1,1,2,0]=(0.385274935809)*x[0]**o+(0.527744200995)*x[0]+(-0.862654310338)*x[1]**o+(0.904005467598)*x[1]
        arg[1,1,2,1]=(-0.380421555378)*x[0]**o+(0.598300323394)*x[0]+(-0.541819816686)*x[1]**o+(0.446934572528)*x[1]
        arg[1,2,0,0]=(0.446781075767)*x[0]**o+(0.906830344178)*x[0]+(0.158734254777)*x[1]**o+(-0.722982586936)*x[1]
        arg[1,2,0,1]=(0.927811104869)*x[0]**o+(-0.446795135388)*x[0]+(-0.90301726295)*x[1]**o+(0.289530807979)*x[1]
        arg[1,2,1,0]=(0.741527109588)*x[0]**o+(-0.370029728344)*x[0]+(0.157695458386)*x[1]**o+(-0.620804465293)*x[1]
        arg[1,2,1,1]=(0.193242816889)*x[0]**o+(0.455789913444)*x[0]+(0.0491001091271)*x[1]**o+(-0.942127264151)*x[1]
        arg[1,2,2,0]=(0.220709813709)*x[0]**o+(-0.713472965425)*x[0]+(-0.736262418061)*x[1]**o+(0.341347605023)*x[1]
        arg[1,2,2,1]=(0.0034506429721)*x[0]**o+(0.397289753453)*x[0]+(-0.167013881353)*x[1]**o+(0.390364445974)*x[1]
        arg[1,3,0,0]=(0.321471168172)*x[0]**o+(-0.237896433618)*x[0]+(-0.199843421009)*x[1]**o+(0.811319091471)*x[1]
        arg[1,3,0,1]=(-0.0799622908037)*x[0]**o+(-0.66155883346)*x[0]+(0.0303112997369)*x[1]**o+(-0.199256497975)*x[1]
        arg[1,3,1,0]=(-0.12051943489)*x[0]**o+(0.674374603779)*x[0]+(0.335933608833)*x[1]**o+(0.366671959553)*x[1]
        arg[1,3,1,1]=(-0.587242845986)*x[0]**o+(-0.495652656487)*x[0]+(0.10609614346)*x[1]**o+(0.177210588935)*x[1]
        arg[1,3,2,0]=(0.947782336638)*x[0]**o+(-0.327283099142)*x[0]+(-0.0280648836948)*x[1]**o+(0.133334543825)*x[1]
        arg[1,3,2,1]=(-0.0887414134508)*x[0]**o+(0.640694455701)*x[0]+(0.266916366237)*x[1]**o+(0.345886606493)*x[1]
        arg[1,4,0,0]=(-0.138184872938)*x[0]**o+(0.658635897645)*x[0]+(-0.00526240442626)*x[1]**o+(0.800443113371)*x[1]
        arg[1,4,0,1]=(-0.549831917928)*x[0]**o+(-0.666733703837)*x[0]+(0.400982699808)*x[1]**o+(0.739099662817)*x[1]
        arg[1,4,1,0]=(-0.137046722563)*x[0]**o+(-0.818450480933)*x[0]+(-0.827665726293)*x[1]**o+(-0.286585607624)*x[1]
        arg[1,4,1,1]=(-0.474156202572)*x[0]**o+(-0.855381707793)*x[0]+(0.113318564035)*x[1]**o+(0.218410511108)*x[1]
        arg[1,4,2,0]=(0.18476112439)*x[0]**o+(-0.964533442641)*x[0]+(-0.968611931598)*x[1]**o+(0.0973026144203)*x[1]
        arg[1,4,2,1]=(-0.372995418552)*x[0]**o+(0.998322146553)*x[0]+(-0.384672070744)*x[1]**o+(0.47536977268)*x[1]
        arg[2,0,0,0]=(0.882400857222)*x[0]**o+(0.455324640128)*x[0]+(0.764297272234)*x[1]**o+(0.498669000791)*x[1]
        arg[2,0,0,1]=(0.894017590794)*x[0]**o+(0.217984119634)*x[0]+(-0.0934000967188)*x[1]**o+(-0.886960376026)*x[1]
        arg[2,0,1,0]=(-0.382158057826)*x[0]**o+(-0.897639065996)*x[0]+(0.819002513335)*x[1]**o+(-0.0739374877518)*x[1]
        arg[2,0,1,1]=(-0.259771272131)*x[0]**o+(-0.533540086977)*x[0]+(0.366917477923)*x[1]**o+(-0.751904525722)*x[1]
        arg[2,0,2,0]=(-0.33946164111)*x[0]**o+(-0.317441178182)*x[0]+(-0.857625945147)*x[1]**o+(-0.84647892719)*x[1]
        arg[2,0,2,1]=(0.153353093148)*x[0]**o+(0.832883645984)*x[0]+(-0.359476461654)*x[1]**o+(0.332390899654)*x[1]
        arg[2,1,0,0]=(0.780980225775)*x[0]**o+(0.619330950566)*x[0]+(0.194312746335)*x[1]**o+(0.595311976655)*x[1]
        arg[2,1,0,1]=(-0.334149973283)*x[0]**o+(-0.670178805858)*x[0]+(-0.313038325156)*x[1]**o+(0.793149618246)*x[1]
        arg[2,1,1,0]=(-0.832852528577)*x[0]**o+(-0.606634692926)*x[0]+(-0.0643816133251)*x[1]**o+(-0.663765987285)*x[1]
        arg[2,1,1,1]=(0.300975732963)*x[0]**o+(0.965431406497)*x[0]+(-0.618456119184)*x[1]**o+(0.459530421885)*x[1]
        arg[2,1,2,0]=(0.122326105776)*x[0]**o+(0.0940802696378)*x[0]+(-0.986852501828)*x[1]**o+(0.370428043236)*x[1]
        arg[2,1,2,1]=(0.447916985436)*x[0]**o+(-0.161926311978)*x[0]+(-0.0933563424166)*x[1]**o+(0.836290263421)*x[1]
        arg[2,2,0,0]=(0.032076347277)*x[0]**o+(-0.447517143643)*x[0]+(-0.053769157756)*x[1]**o+(-0.0242981207801)*x[1]
        arg[2,2,0,1]=(-0.403935703163)*x[0]**o+(-0.960750093184)*x[0]+(-0.287956277998)*x[1]**o+(-0.168614701082)*x[1]
        arg[2,2,1,0]=(-0.205077009492)*x[0]**o+(-0.409536289697)*x[0]+(0.359071620541)*x[1]**o+(-0.466455759025)*x[1]
        arg[2,2,1,1]=(0.443658057141)*x[0]**o+(-0.748681939527)*x[0]+(0.2153167916)*x[1]**o+(-0.475899799738)*x[1]
        arg[2,2,2,0]=(-0.463879681251)*x[0]**o+(0.758317222041)*x[0]+(0.608221916958)*x[1]**o+(-0.237798399085)*x[1]
        arg[2,2,2,1]=(0.533981754168)*x[0]**o+(0.590125912931)*x[0]+(0.8957267808)*x[1]**o+(0.294612478767)*x[1]
        arg[2,3,0,0]=(0.499387701403)*x[0]**o+(0.0379132293822)*x[0]+(0.717384215206)*x[1]**o+(0.754497698503)*x[1]
        arg[2,3,0,1]=(0.866500398882)*x[0]**o+(0.23678788012)*x[0]+(0.395132533716)*x[1]**o+(-0.451619930593)*x[1]
        arg[2,3,1,0]=(-0.879663354176)*x[0]**o+(-0.328937922978)*x[0]+(0.688771917581)*x[1]**o+(-0.87392051489)*x[1]
        arg[2,3,1,1]=(0.291214284356)*x[0]**o+(-0.783159133488)*x[0]+(0.240960084233)*x[1]**o+(0.339043900275)*x[1]
        arg[2,3,2,0]=(0.47064755192)*x[0]**o+(-0.42618815064)*x[0]+(-0.505342655776)*x[1]**o+(0.408218280069)*x[1]
        arg[2,3,2,1]=(0.775731405735)*x[0]**o+(-0.768750178591)*x[0]+(-0.896333275063)*x[1]**o+(0.286558848957)*x[1]
        arg[2,4,0,0]=(-0.7221248688)*x[0]**o+(0.612509823967)*x[0]+(0.436744528143)*x[1]**o+(-0.761502758764)*x[1]
        arg[2,4,0,1]=(-0.742652222385)*x[0]**o+(0.214811997988)*x[0]+(-0.031320962337)*x[1]**o+(-0.320134664789)*x[1]
        arg[2,4,1,0]=(-0.529575527815)*x[0]**o+(0.623770482091)*x[0]+(-0.50790148262)*x[1]**o+(0.629874274271)*x[1]
        arg[2,4,1,1]=(0.535631957883)*x[0]**o+(0.439578618993)*x[0]+(-0.770747718564)*x[1]**o+(0.0798696116058)*x[1]
        arg[2,4,2,0]=(0.81688202721)*x[0]**o+(-0.0301054197201)*x[0]+(0.613495031968)*x[1]**o+(-0.153619455295)*x[1]
        arg[2,4,2,1]=(0.493471708612)*x[0]**o+(0.423626757568)*x[0]+(0.093174141001)*x[1]**o+(0.307452108722)*x[1]
        arg[3,0,0,0]=(-0.0668630243284)*x[0]**o+(-0.577709692979)*x[0]+(0.984307685082)*x[1]**o+(-0.171487030152)*x[1]
        arg[3,0,0,1]=(-0.210452484293)*x[0]**o+(-0.130476372055)*x[0]+(-0.693596086089)*x[1]**o+(-0.692143724329)*x[1]
        arg[3,0,1,0]=(-0.34853285523)*x[0]**o+(-0.369649709966)*x[0]+(0.824648362105)*x[1]**o+(0.733046240399)*x[1]
        arg[3,0,1,1]=(0.861903045965)*x[0]**o+(0.609535692452)*x[0]+(0.681514895836)*x[1]**o+(0.66773275418)*x[1]
        arg[3,0,2,0]=(0.0197670342844)*x[0]**o+(-0.251645539228)*x[0]+(-0.0390293468405)*x[1]**o+(-0.00955605638373)*x[1]
        arg[3,0,2,1]=(-0.499093940395)*x[0]**o+(0.660356623112)*x[0]+(0.969057635866)*x[1]**o+(-0.821377317197)*x[1]
        arg[3,1,0,0]=(-0.955568800095)*x[0]**o+(0.557877698917)*x[0]+(0.829599262959)*x[1]**o+(0.933736802562)*x[1]
        arg[3,1,0,1]=(-0.910307638058)*x[0]**o+(-0.3287646946)*x[0]+(-0.777548579252)*x[1]**o+(-0.971744353969)*x[1]
        arg[3,1,1,0]=(0.22353820834)*x[0]**o+(0.904479243377)*x[0]+(-0.108768166204)*x[1]**o+(0.823767232517)*x[1]
        arg[3,1,1,1]=(0.371995768061)*x[0]**o+(0.747630377273)*x[0]+(0.931152588498)*x[1]**o+(-0.0154330268188)*x[1]
        arg[3,1,2,0]=(0.4567417369)*x[0]**o+(0.424951293892)*x[0]+(-0.14032259215)*x[1]**o+(0.603329614924)*x[1]
        arg[3,1,2,1]=(-0.855100213731)*x[0]**o+(-0.667832882495)*x[0]+(0.942530922378)*x[1]**o+(0.55861156259)*x[1]
        arg[3,2,0,0]=(0.17190936572)*x[0]**o+(-0.904098615762)*x[0]+(0.991497679624)*x[1]**o+(0.373020358381)*x[1]
        arg[3,2,0,1]=(-0.683587206983)*x[0]**o+(0.718257814718)*x[0]+(-0.730243933765)*x[1]**o+(-0.583892189979)*x[1]
        arg[3,2,1,0]=(0.051672736629)*x[0]**o+(-0.991172563807)*x[0]+(0.36605487466)*x[1]**o+(0.680719877794)*x[1]
        arg[3,2,1,1]=(0.438014662973)*x[0]**o+(0.172969450125)*x[0]+(-0.424515284177)*x[1]**o+(0.615315132547)*x[1]
        arg[3,2,2,0]=(-0.709499532674)*x[0]**o+(0.985097212621)*x[0]+(0.326610298255)*x[1]**o+(0.994701669258)*x[1]
        arg[3,2,2,1]=(-0.657647707844)*x[0]**o+(-0.683954483768)*x[0]+(-0.846657945161)*x[1]**o+(0.852308904063)*x[1]
        arg[3,3,0,0]=(0.21346320453)*x[0]**o+(-0.181665119768)*x[0]+(0.549866100727)*x[1]**o+(-0.0858045316871)*x[1]
        arg[3,3,0,1]=(0.249957470685)*x[0]**o+(0.537464773638)*x[0]+(0.0224330511636)*x[1]**o+(0.114623595638)*x[1]
        arg[3,3,1,0]=(0.166507029533)*x[0]**o+(0.232150195873)*x[0]+(0.119453667263)*x[1]**o+(-0.261110960642)*x[1]
        arg[3,3,1,1]=(0.358248720077)*x[0]**o+(-0.871911261059)*x[0]+(0.57743453064)*x[1]**o+(0.950322066335)*x[1]
        arg[3,3,2,0]=(-0.896899724447)*x[0]**o+(-0.898884767716)*x[0]+(0.63832012742)*x[1]**o+(-0.833800375634)*x[1]
        arg[3,3,2,1]=(0.425194254583)*x[0]**o+(0.161328467295)*x[0]+(0.496219405916)*x[1]**o+(0.380935896093)*x[1]
        arg[3,4,0,0]=(0.793902581592)*x[0]**o+(-0.33428511164)*x[0]+(-0.283510438468)*x[1]**o+(0.228913046074)*x[1]
        arg[3,4,0,1]=(0.372028980539)*x[0]**o+(-0.388909011671)*x[0]+(-0.259885631206)*x[1]**o+(-0.950910172897)*x[1]
        arg[3,4,1,0]=(-0.80489673786)*x[0]**o+(-0.828175420038)*x[0]+(0.895070489116)*x[1]**o+(-0.742272132103)*x[1]
        arg[3,4,1,1]=(0.272535126101)*x[0]**o+(0.902668974891)*x[0]+(0.399816667015)*x[1]**o+(-0.953186908789)*x[1]
        arg[3,4,2,0]=(0.0341775288582)*x[0]**o+(0.962241492639)*x[0]+(0.906020184876)*x[1]**o+(-0.400536255726)*x[1]
        arg[3,4,2,1]=(-0.745016464485)*x[0]**o+(-0.85241829747)*x[0]+(-0.648936544722)*x[1]**o+(-0.997030074218)*x[1]
        ref[0,0,0,0]=(0.965300679335)/(o+1.)+(0.142356258935)+(0.392628958003)*0.5**o
        ref[0,0,0,1]=(0.384782163488)/(o+1.)+(-0.407081281958)+(-0.503312425492)*0.5**o
        ref[0,0,1,0]=(-0.765235012795)/(o+1.)+(-0.317187415586)+(0.660162511699)*0.5**o
        ref[0,0,1,1]=(-0.451665346257)/(o+1.)+(-0.0341912040643)+(-0.986935412732)*0.5**o
        ref[0,0,2,0]=(-0.695553902133)/(o+1.)+(-0.467731542733)+(0.632875178502)*0.5**o
        ref[0,0,2,1]=(0.345909088853)/(o+1.)+(0.225477764967)+(0.102625648731)*0.5**o
        ref[0,1,0,0]=(0.332480141972)/(o+1.)+(-0.388404308158)+(-0.0108106931425)*0.5**o
        ref[0,1,0,1]=(0.328480565578)/(o+1.)+(0.324505149668)+(0.220826593491)*0.5**o
        ref[0,1,1,0]=(0.292927451477)/(o+1.)+(0.183128254664)+(0.22735402466)*0.5**o
        ref[0,1,1,1]=(-0.322061131008)/(o+1.)+(0.768096283874)+(0.495079875966)*0.5**o
        ref[0,1,2,0]=(0.286037351205)/(o+1.)+(0.202030463766)+(0.968583889554)*0.5**o
        ref[0,1,2,1]=(0.891305263219)/(o+1.)+(0.219403586029)+(0.361498191707)*0.5**o
        ref[0,2,0,0]=(0.0870608077911)/(o+1.)+(-0.381665114117)+(0.185647566516)*0.5**o
        ref[0,2,0,1]=(-0.0229924234496)/(o+1.)+(0.244502612446)+(0.227262327826)*0.5**o
        ref[0,2,1,0]=(0.957377329048)/(o+1.)+(-0.0454149985758)+(0.0405753322301)*0.5**o
        ref[0,2,1,1]=(-0.628262263897)/(o+1.)+(0.0870979129647)+(-0.546046301701)*0.5**o
        ref[0,2,2,0]=(0.499595322279)/(o+1.)+(0.0509587090171)+(-0.828455793175)*0.5**o
        ref[0,2,2,1]=(-0.678432418122)/(o+1.)+(0.319356853239)+(0.926138085222)*0.5**o
        ref[0,3,0,0]=(0.62665109546)/(o+1.)+(0.762318465671)+(-0.473066058658)*0.5**o
        ref[0,3,0,1]=(-0.459252843816)/(o+1.)+(0.295864315717)+(0.776790978673)*0.5**o
        ref[0,3,1,0]=(0.445864469556)/(o+1.)+(0.292283099022)+(0.34783214571)*0.5**o
        ref[0,3,1,1]=(0.577708844669)/(o+1.)+(-0.491318272755)+(0.0717966037463)*0.5**o
        ref[0,3,2,0]=(-0.33936340707)/(o+1.)+(-0.428987751016)+(-0.548573971799)*0.5**o
        ref[0,3,2,1]=(-0.833323497891)/(o+1.)+(-0.224290613939)+(-0.28309240461)*0.5**o
        ref[0,4,0,0]=(0.146657173319)/(o+1.)+(0.200750335943)+(-0.105416428922)*0.5**o
        ref[0,4,0,1]=(0.534941970665)/(o+1.)+(-0.251774809823)+(-0.876111094069)*0.5**o
        ref[0,4,1,0]=(-0.739368558349)/(o+1.)+(0.228739674374)+(-0.678996394303)*0.5**o
        ref[0,4,1,1]=(-0.651647023142)/(o+1.)+(-0.159264639548)+(0.0472145581152)*0.5**o
        ref[0,4,2,0]=(-0.524234284672)/(o+1.)+(0.218536682587)+(0.114936857967)*0.5**o
        ref[0,4,2,1]=(0.133122220811)/(o+1.)+(-0.31070612124)+(0.568834117773)*0.5**o
        ref[1,0,0,0]=(-0.132449009841)/(o+1.)+(0.380882086677)+(-0.821241886235)*0.5**o
        ref[1,0,0,1]=(-0.573597121344)/(o+1.)+(-0.188183973289)+(0.624372422235)*0.5**o
        ref[1,0,1,0]=(-0.249786553024)/(o+1.)+(-0.661057334385)+(-0.568749808355)*0.5**o
        ref[1,0,1,1]=(0.59246455477)/(o+1.)+(-0.256347568096)+(0.0765557360835)*0.5**o
        ref[1,0,2,0]=(0.0689432766231)/(o+1.)+(-0.267056955718)+(0.909614864187)*0.5**o
        ref[1,0,2,1]=(0.378273922837)/(o+1.)+(0.770144388717)+(0.0610206203468)*0.5**o
        ref[1,1,0,0]=(0.728786835113)/(o+1.)+(0.436825316493)+(0.917249169051)*0.5**o
        ref[1,1,0,1]=(-0.418402258918)/(o+1.)+(-0.47604993983)+(0.331129708654)*0.5**o
        ref[1,1,1,0]=(0.989603613532)/(o+1.)+(-0.153579923633)+(-0.527408405279)*0.5**o
        ref[1,1,1,1]=(0.11459047938)/(o+1.)+(-0.306426672553)+(-0.189543145217)*0.5**o
        ref[1,1,2,0]=(-0.862654310338)/(o+1.)+(0.715874834296)+(0.385274935809)*0.5**o
        ref[1,1,2,1]=(-0.541819816686)/(o+1.)+(0.522617447961)+(-0.380421555378)*0.5**o
        ref[1,2,0,0]=(0.158734254777)/(o+1.)+(0.0919238786215)+(0.446781075767)*0.5**o
        ref[1,2,0,1]=(-0.90301726295)/(o+1.)+(-0.0786321637044)+(0.927811104869)*0.5**o
        ref[1,2,1,0]=(0.157695458386)/(o+1.)+(-0.495417096819)+(0.741527109588)*0.5**o
        ref[1,2,1,1]=(0.0491001091271)/(o+1.)+(-0.243168675353)+(0.193242816889)*0.5**o
        ref[1,2,2,0]=(-0.736262418061)/(o+1.)+(-0.186062680201)+(0.220709813709)*0.5**o
        ref[1,2,2,1]=(-0.167013881353)/(o+1.)+(0.393827099713)+(0.0034506429721)*0.5**o
        ref[1,3,0,0]=(-0.199843421009)/(o+1.)+(0.286711328927)+(0.321471168172)*0.5**o
        ref[1,3,0,1]=(0.0303112997369)/(o+1.)+(-0.430407665717)+(-0.0799622908037)*0.5**o
        ref[1,3,1,0]=(0.335933608833)/(o+1.)+(0.520523281666)+(-0.12051943489)*0.5**o
        ref[1,3,1,1]=(0.10609614346)/(o+1.)+(-0.159221033776)+(-0.587242845986)*0.5**o
        ref[1,3,2,0]=(-0.0280648836948)/(o+1.)+(-0.0969742776586)+(0.947782336638)*0.5**o
        ref[1,3,2,1]=(0.266916366237)/(o+1.)+(0.493290531097)+(-0.0887414134508)*0.5**o
        ref[1,4,0,0]=(-0.00526240442626)/(o+1.)+(0.729539505508)+(-0.138184872938)*0.5**o
        ref[1,4,0,1]=(0.400982699808)/(o+1.)+(0.0361829794896)+(-0.549831917928)*0.5**o
        ref[1,4,1,0]=(-0.827665726293)/(o+1.)+(-0.552518044279)+(-0.137046722563)*0.5**o
        ref[1,4,1,1]=(0.113318564035)/(o+1.)+(-0.318485598343)+(-0.474156202572)*0.5**o
        ref[1,4,2,0]=(-0.968611931598)/(o+1.)+(-0.433615414111)+(0.18476112439)*0.5**o
        ref[1,4,2,1]=(-0.384672070744)/(o+1.)+(0.736845959617)+(-0.372995418552)*0.5**o
        ref[2,0,0,0]=(0.764297272234)/(o+1.)+(0.476996820459)+(0.882400857222)*0.5**o
        ref[2,0,0,1]=(-0.0934000967188)/(o+1.)+(-0.334488128196)+(0.894017590794)*0.5**o
        ref[2,0,1,0]=(0.819002513335)/(o+1.)+(-0.485788276874)+(-0.382158057826)*0.5**o
        ref[2,0,1,1]=(0.366917477923)/(o+1.)+(-0.64272230635)+(-0.259771272131)*0.5**o
        ref[2,0,2,0]=(-0.857625945147)/(o+1.)+(-0.581960052686)+(-0.33946164111)*0.5**o
        ref[2,0,2,1]=(-0.359476461654)/(o+1.)+(0.582637272819)+(0.153353093148)*0.5**o
        ref[2,1,0,0]=(0.194312746335)/(o+1.)+(0.607321463611)+(0.780980225775)*0.5**o
        ref[2,1,0,1]=(-0.313038325156)/(o+1.)+(0.0614854061938)+(-0.334149973283)*0.5**o
        ref[2,1,1,0]=(-0.0643816133251)/(o+1.)+(-0.635200340106)+(-0.832852528577)*0.5**o
        ref[2,1,1,1]=(-0.618456119184)/(o+1.)+(0.712480914191)+(0.300975732963)*0.5**o
        ref[2,1,2,0]=(-0.986852501828)/(o+1.)+(0.232254156437)+(0.122326105776)*0.5**o
        ref[2,1,2,1]=(-0.0933563424166)/(o+1.)+(0.337181975722)+(0.447916985436)*0.5**o
        ref[2,2,0,0]=(-0.053769157756)/(o+1.)+(-0.235907632211)+(0.032076347277)*0.5**o
        ref[2,2,0,1]=(-0.287956277998)/(o+1.)+(-0.564682397133)+(-0.403935703163)*0.5**o
        ref[2,2,1,0]=(0.359071620541)/(o+1.)+(-0.437996024361)+(-0.205077009492)*0.5**o
        ref[2,2,1,1]=(0.2153167916)/(o+1.)+(-0.612290869632)+(0.443658057141)*0.5**o
        ref[2,2,2,0]=(0.608221916958)/(o+1.)+(0.260259411478)+(-0.463879681251)*0.5**o
        ref[2,2,2,1]=(0.8957267808)/(o+1.)+(0.442369195849)+(0.533981754168)*0.5**o
        ref[2,3,0,0]=(0.717384215206)/(o+1.)+(0.396205463942)+(0.499387701403)*0.5**o
        ref[2,3,0,1]=(0.395132533716)/(o+1.)+(-0.107416025237)+(0.866500398882)*0.5**o
        ref[2,3,1,0]=(0.688771917581)/(o+1.)+(-0.601429218934)+(-0.879663354176)*0.5**o
        ref[2,3,1,1]=(0.240960084233)/(o+1.)+(-0.222057616607)+(0.291214284356)*0.5**o
        ref[2,3,2,0]=(-0.505342655776)/(o+1.)+(-0.00898493528542)+(0.47064755192)*0.5**o
        ref[2,3,2,1]=(-0.896333275063)/(o+1.)+(-0.241095664817)+(0.775731405735)*0.5**o
        ref[2,4,0,0]=(0.436744528143)/(o+1.)+(-0.0744964673986)+(-0.7221248688)*0.5**o
        ref[2,4,0,1]=(-0.031320962337)/(o+1.)+(-0.0526613334003)+(-0.742652222385)*0.5**o
        ref[2,4,1,0]=(-0.50790148262)/(o+1.)+(0.626822378181)+(-0.529575527815)*0.5**o
        ref[2,4,1,1]=(-0.770747718564)/(o+1.)+(0.2597241153)+(0.535631957883)*0.5**o
        ref[2,4,2,0]=(0.613495031968)/(o+1.)+(-0.0918624375077)+(0.81688202721)*0.5**o
        ref[2,4,2,1]=(0.093174141001)/(o+1.)+(0.365539433145)+(0.493471708612)*0.5**o
        ref[3,0,0,0]=(0.984307685082)/(o+1.)+(-0.374598361566)+(-0.0668630243284)*0.5**o
        ref[3,0,0,1]=(-0.693596086089)/(o+1.)+(-0.411310048192)+(-0.210452484293)*0.5**o
        ref[3,0,1,0]=(0.824648362105)/(o+1.)+(0.181698265217)+(-0.34853285523)*0.5**o
        ref[3,0,1,1]=(0.681514895836)/(o+1.)+(0.638634223316)+(0.861903045965)*0.5**o
        ref[3,0,2,0]=(-0.0390293468405)/(o+1.)+(-0.130600797806)+(0.0197670342844)*0.5**o
        ref[3,0,2,1]=(0.969057635866)/(o+1.)+(-0.0805103470423)+(-0.499093940395)*0.5**o
        ref[3,1,0,0]=(0.829599262959)/(o+1.)+(0.745807250739)+(-0.955568800095)*0.5**o
        ref[3,1,0,1]=(-0.777548579252)/(o+1.)+(-0.650254524284)+(-0.910307638058)*0.5**o
        ref[3,1,1,0]=(-0.108768166204)/(o+1.)+(0.864123237947)+(0.22353820834)*0.5**o
        ref[3,1,1,1]=(0.931152588498)/(o+1.)+(0.366098675227)+(0.371995768061)*0.5**o
        ref[3,1,2,0]=(-0.14032259215)/(o+1.)+(0.514140454408)+(0.4567417369)*0.5**o
        ref[3,1,2,1]=(0.942530922378)/(o+1.)+(-0.0546106599524)+(-0.855100213731)*0.5**o
        ref[3,2,0,0]=(0.991497679624)/(o+1.)+(-0.26553912869)+(0.17190936572)*0.5**o
        ref[3,2,0,1]=(-0.730243933765)/(o+1.)+(0.0671828123695)+(-0.683587206983)*0.5**o
        ref[3,2,1,0]=(0.36605487466)/(o+1.)+(-0.155226343006)+(0.051672736629)*0.5**o
        ref[3,2,1,1]=(-0.424515284177)/(o+1.)+(0.394142291336)+(0.438014662973)*0.5**o
        ref[3,2,2,0]=(0.326610298255)/(o+1.)+(0.98989944094)+(-0.709499532674)*0.5**o
        ref[3,2,2,1]=(-0.846657945161)/(o+1.)+(0.0841772101477)+(-0.657647707844)*0.5**o
        ref[3,3,0,0]=(0.549866100727)/(o+1.)+(-0.133734825728)+(0.21346320453)*0.5**o
        ref[3,3,0,1]=(0.0224330511636)/(o+1.)+(0.326044184638)+(0.249957470685)*0.5**o
        ref[3,3,1,0]=(0.119453667263)/(o+1.)+(-0.0144803823844)+(0.166507029533)*0.5**o
        ref[3,3,1,1]=(0.57743453064)/(o+1.)+(0.0392054026384)+(0.358248720077)*0.5**o
        ref[3,3,2,0]=(0.63832012742)/(o+1.)+(-0.866342571675)+(-0.896899724447)*0.5**o
        ref[3,3,2,1]=(0.496219405916)/(o+1.)+(0.271132181694)+(0.425194254583)*0.5**o
        ref[3,4,0,0]=(-0.283510438468)/(o+1.)+(-0.0526860327828)+(0.793902581592)*0.5**o
        ref[3,4,0,1]=(-0.259885631206)/(o+1.)+(-0.669909592284)+(0.372028980539)*0.5**o
        ref[3,4,1,0]=(0.895070489116)/(o+1.)+(-0.78522377607)+(-0.80489673786)*0.5**o
        ref[3,4,1,1]=(0.399816667015)/(o+1.)+(-0.0252589669489)+(0.272535126101)*0.5**o
        ref[3,4,2,0]=(0.906020184876)/(o+1.)+(0.280852618457)+(0.0341775288582)*0.5**o
        ref[3,4,2,1]=(-0.648936544722)/(o+1.)+(-0.924724185844)+(-0.745016464485)*0.5**o
      else:
        arg[0,0,0,0]=(-0.891975745754)*x[0]**o+(0.678331209591)*x[0]+(-0.450022305538)*x[1]**o+(0.521272586024)*x[1]+(-0.28868708769)*x[2]**o+(-0.631617544339)*x[2]
        arg[0,0,0,1]=(0.428698930349)*x[0]**o+(-0.856251438077)*x[0]+(0.915766367729)*x[1]**o+(-0.35019463199)*x[1]+(-0.0029340113102)*x[2]**o+(-0.394458505835)*x[2]
        arg[0,0,1,0]=(0.974069118653)*x[0]**o+(-0.674028781962)*x[0]+(0.489718068973)*x[1]**o+(-0.471323892144)*x[1]+(0.11469271678)*x[2]**o+(-0.264584383354)*x[2]
        arg[0,0,1,1]=(0.813412823502)*x[0]**o+(0.739276532652)*x[0]+(-0.863216628203)*x[1]**o+(0.486539683681)*x[1]+(-0.771448803712)*x[2]**o+(-0.621526889056)*x[2]
        arg[0,0,2,0]=(-0.874502916672)*x[0]**o+(0.0708047915752)*x[0]+(0.917576817333)*x[1]**o+(-0.257626031306)*x[1]+(-0.161822453369)*x[2]**o+(0.488866912786)*x[2]
        arg[0,0,2,1]=(-0.258524075807)*x[0]**o+(0.206589504672)*x[0]+(0.232971263871)*x[1]**o+(-0.917991566808)*x[1]+(0.725943193794)*x[2]**o+(-0.170341108992)*x[2]
        arg[0,1,0,0]=(0.489505903018)*x[0]**o+(0.386376068728)*x[0]+(0.510797276159)*x[1]**o+(0.45356046467)*x[1]+(-0.664963144251)*x[2]**o+(0.128579947316)*x[2]
        arg[0,1,0,1]=(0.441567332096)*x[0]**o+(0.658686083904)*x[0]+(0.46601056348)*x[1]**o+(-0.130759928108)*x[1]+(0.324576589)*x[2]**o+(0.844700852997)*x[2]
        arg[0,1,1,0]=(0.388576628492)*x[0]**o+(0.41731399682)*x[0]+(-0.37037879729)*x[1]**o+(-0.00306376860454)*x[1]+(-0.464420131682)*x[2]**o+(0.570470055678)*x[2]
        arg[0,1,1,1]=(-0.0537780037583)*x[0]**o+(0.641854502644)*x[0]+(-0.475355608842)*x[1]**o+(-0.155501886306)*x[1]+(0.159833591997)*x[2]**o+(0.889951997355)*x[2]
        arg[0,1,2,0]=(0.58262140787)*x[0]**o+(0.487380368384)*x[0]+(-0.783228845976)*x[1]**o+(-0.683221249176)*x[1]+(0.98628901689)*x[2]**o+(0.591884616904)*x[2]
        arg[0,1,2,1]=(0.438472233911)*x[0]**o+(0.989088416794)*x[0]+(0.221688966923)*x[1]**o+(-0.119266918098)*x[1]+(-0.887421255415)*x[2]**o+(0.728082459156)*x[2]
        arg[0,2,0,0]=(-0.334110476968)*x[0]**o+(-0.877094545216)*x[0]+(-0.293135209996)*x[1]**o+(0.578110480224)*x[1]+(-0.898686340104)*x[2]**o+(0.528072783107)*x[2]
        arg[0,2,0,1]=(0.33041258354)*x[0]**o+(-0.70349630629)*x[0]+(0.349551161792)*x[1]**o+(-0.296687014146)*x[1]+(-0.272377440466)*x[2]**o+(-0.42221458274)*x[2]
        arg[0,2,1,0]=(-0.80021736912)*x[0]**o+(0.271727608951)*x[0]+(0.689351606243)*x[1]**o+(-0.95063104158)*x[1]+(0.447293221661)*x[2]**o+(0.962435109303)*x[2]
        arg[0,2,1,1]=(-0.773331228218)*x[0]**o+(0.569225340051)*x[0]+(-0.0330042729183)*x[1]**o+(-0.732867386826)*x[1]+(-0.399631128084)*x[2]**o+(0.235907094874)*x[2]
        arg[0,2,2,0]=(-0.943072868965)*x[0]**o+(-0.959364749946)*x[0]+(0.922934300445)*x[1]**o+(0.27257203243)*x[1]+(0.450453864748)*x[2]**o+(0.521284310775)*x[2]
        arg[0,2,2,1]=(-0.821201079286)*x[0]**o+(-0.801011107627)*x[0]+(0.554575277011)*x[1]**o+(-0.411361733513)*x[1]+(-0.858195570474)*x[2]**o+(0.420660103079)*x[2]
        arg[0,3,0,0]=(-0.0564164524098)*x[0]**o+(-0.865516143298)*x[0]+(-0.85754022189)*x[1]**o+(-0.0159800361584)*x[1]+(-0.891158469429)*x[2]**o+(-0.809909518115)*x[2]
        arg[0,3,0,1]=(-0.553858251068)*x[0]**o+(0.827558838148)*x[0]+(0.690081127613)*x[1]**o+(0.3475090336)*x[1]+(-0.190426828891)*x[2]**o+(-0.99589316535)*x[2]
        arg[0,3,1,0]=(-0.992317453163)*x[0]**o+(-0.992812531491)*x[0]+(-0.859683310869)*x[1]**o+(0.441779587739)*x[1]+(0.582457020078)*x[2]**o+(-0.959111461424)*x[2]
        arg[0,3,1,1]=(-0.713753922213)*x[0]**o+(0.504885148776)*x[0]+(-0.076988545634)*x[1]**o+(0.530828889101)*x[1]+(-0.335616242708)*x[2]**o+(-0.0421293420747)*x[2]
        arg[0,3,2,0]=(0.536024118289)*x[0]**o+(-0.527038052847)*x[0]+(-0.558392588656)*x[1]**o+(0.928618057856)*x[1]+(0.427507510033)*x[2]**o+(0.223570914141)*x[2]
        arg[0,3,2,1]=(-0.197968756117)*x[0]**o+(-0.58157606173)*x[0]+(-0.878630012825)*x[1]**o+(-0.380635552726)*x[1]+(-0.477570149125)*x[2]**o+(-0.642904015434)*x[2]
        arg[0,4,0,0]=(0.688763655127)*x[0]**o+(0.286011523933)*x[0]+(0.0933014250929)*x[1]**o+(0.355550563233)*x[1]+(0.121445291305)*x[2]**o+(-0.814538082974)*x[2]
        arg[0,4,0,1]=(-0.402717742839)*x[0]**o+(0.403946649835)*x[0]+(0.286978800579)*x[1]**o+(-0.363358056216)*x[1]+(-0.185233594102)*x[2]**o+(-0.902660250807)*x[2]
        arg[0,4,1,0]=(0.39352838659)*x[0]**o+(-0.430755011337)*x[0]+(-0.912518631162)*x[1]**o+(-0.634118795867)*x[1]+(-0.802445843416)*x[2]**o+(0.643395158048)*x[2]
        arg[0,4,1,1]=(-0.375507250175)*x[0]**o+(-0.609576942939)*x[0]+(0.35060243693)*x[1]**o+(0.384630130603)*x[1]+(0.165576820334)*x[2]**o+(-0.319012357341)*x[2]
        arg[0,4,2,0]=(0.0603590263818)*x[0]**o+(-0.435954775883)*x[0]+(-0.895329092913)*x[1]**o+(0.364809539103)*x[1]+(0.912864788565)*x[2]**o+(0.582044419963)*x[2]
        arg[0,4,2,1]=(-0.111089454685)*x[0]**o+(-0.247527778755)*x[0]+(0.897263901657)*x[1]**o+(0.168097655548)*x[1]+(-0.532273075067)*x[2]**o+(-0.158529268933)*x[2]
        arg[1,0,0,0]=(-0.940865155622)*x[0]**o+(0.639652431549)*x[0]+(0.204433172026)*x[1]**o+(-0.333435100662)*x[1]+(0.922286591304)*x[2]**o+(0.245091036189)*x[2]
        arg[1,0,0,1]=(-0.843207880309)*x[0]**o+(0.496567621809)*x[0]+(-0.747925907677)*x[1]**o+(-0.666195656311)*x[1]+(-0.836848401917)*x[2]**o+(-0.77301298659)*x[2]
        arg[1,0,1,0]=(-0.484987528212)*x[0]**o+(-0.208798022576)*x[0]+(-0.516295195217)*x[1]**o+(0.273733676249)*x[1]+(0.642865886081)*x[2]**o+(0.94556566928)*x[2]
        arg[1,0,1,1]=(0.285548625366)*x[0]**o+(-0.316870037241)*x[0]+(0.166055429845)*x[1]**o+(-0.907240066063)*x[1]+(-0.973602611383)*x[2]**o+(-0.232655535454)*x[2]
        arg[1,0,2,0]=(0.96797757731)*x[0]**o+(-0.0920831342923)*x[0]+(0.0367163548581)*x[1]**o+(0.372503137376)*x[1]+(0.114902128151)*x[2]**o+(0.147390259838)*x[2]
        arg[1,0,2,1]=(-0.486190283922)*x[0]**o+(-0.713417720041)*x[0]+(0.0212905465522)*x[1]**o+(-0.358636168795)*x[1]+(0.719370965272)*x[2]**o+(-0.506502687359)*x[2]
        arg[1,1,0,0]=(-0.216544472879)*x[0]**o+(-0.690540499698)*x[0]+(0.190155427044)*x[1]**o+(0.832206912182)*x[1]+(-0.108713936252)*x[2]**o+(-0.221373087337)*x[2]
        arg[1,1,0,1]=(0.982934604971)*x[0]**o+(0.604369826835)*x[0]+(-0.34365978148)*x[1]**o+(-0.167486779981)*x[1]+(-0.753609370582)*x[2]**o+(-0.220160888977)*x[2]
        arg[1,1,1,0]=(0.516177509446)*x[0]**o+(-0.474445723273)*x[0]+(0.813647087566)*x[1]**o+(-0.281201093802)*x[1]+(0.941465944108)*x[2]**o+(0.352728360059)*x[2]
        arg[1,1,1,1]=(-0.914732890633)*x[0]**o+(0.214962090901)*x[0]+(-0.710485058274)*x[1]**o+(-0.070756754056)*x[1]+(-0.449571590211)*x[2]**o+(0.353655711451)*x[2]
        arg[1,1,2,0]=(-0.96655555619)*x[0]**o+(-0.47460534732)*x[0]+(0.0610695053327)*x[1]**o+(0.815390116709)*x[1]+(-0.844916163552)*x[2]**o+(-0.433015304307)*x[2]
        arg[1,1,2,1]=(-0.611879453709)*x[0]**o+(0.849092848509)*x[0]+(0.0181322361566)*x[1]**o+(0.264173017827)*x[1]+(0.0452154806937)*x[2]**o+(0.0395300648465)*x[2]
        arg[1,2,0,0]=(0.558820946357)*x[0]**o+(-0.471519705598)*x[0]+(-0.200404864145)*x[1]**o+(0.452044160691)*x[1]+(-0.507559485963)*x[2]**o+(-0.993437841726)*x[2]
        arg[1,2,0,1]=(-0.944738664192)*x[0]**o+(-0.497536772117)*x[0]+(0.111399525675)*x[1]**o+(-0.93450294271)*x[1]+(-0.463702615737)*x[2]**o+(-0.330112672884)*x[2]
        arg[1,2,1,0]=(0.238225961058)*x[0]**o+(-0.471520173419)*x[0]+(0.483767084892)*x[1]**o+(-0.0359407284537)*x[1]+(-0.767379409753)*x[2]**o+(-0.283596371915)*x[2]
        arg[1,2,1,1]=(0.188699181238)*x[0]**o+(0.829593121989)*x[0]+(0.603377767593)*x[1]**o+(0.163334775468)*x[1]+(-0.695141594131)*x[2]**o+(-0.335208640181)*x[2]
        arg[1,2,2,0]=(-0.0953160757881)*x[0]**o+(0.179594592176)*x[0]+(0.95983902191)*x[1]**o+(-0.689622683527)*x[1]+(0.418442085936)*x[2]**o+(-0.271399861129)*x[2]
        arg[1,2,2,1]=(0.598691886555)*x[0]**o+(0.334120340646)*x[0]+(0.580763872443)*x[1]**o+(0.769016736836)*x[1]+(-0.442121483951)*x[2]**o+(0.0685264980499)*x[2]
        arg[1,3,0,0]=(-0.482944276372)*x[0]**o+(-0.0518057205519)*x[0]+(-0.878100108766)*x[1]**o+(-0.130957239599)*x[1]+(0.71270341547)*x[2]**o+(-0.313822116756)*x[2]
        arg[1,3,0,1]=(0.565128725562)*x[0]**o+(0.706149004661)*x[0]+(0.837747652328)*x[1]**o+(-0.429417376176)*x[1]+(0.179437627997)*x[2]**o+(-0.841831183522)*x[2]
        arg[1,3,1,0]=(0.608648887224)*x[0]**o+(0.118220851451)*x[0]+(0.492890967377)*x[1]**o+(-0.162344084552)*x[1]+(0.141203210671)*x[2]**o+(0.883457779687)*x[2]
        arg[1,3,1,1]=(0.0322521291544)*x[0]**o+(0.839748977403)*x[0]+(0.75094786544)*x[1]**o+(0.954833145672)*x[1]+(0.461457599013)*x[2]**o+(-0.761973385747)*x[2]
        arg[1,3,2,0]=(-0.757022609619)*x[0]**o+(-0.250893790099)*x[0]+(-0.973631035961)*x[1]**o+(0.506574565217)*x[1]+(0.103557430896)*x[2]**o+(0.881026337134)*x[2]
        arg[1,3,2,1]=(-0.395945302352)*x[0]**o+(-0.125428842065)*x[0]+(-0.339663004568)*x[1]**o+(-0.0272044592179)*x[1]+(0.75200506467)*x[2]**o+(-0.632922582454)*x[2]
        arg[1,4,0,0]=(0.0462127928202)*x[0]**o+(-0.745613488751)*x[0]+(0.357338297422)*x[1]**o+(0.739034094804)*x[1]+(-0.451247272253)*x[2]**o+(0.418030110847)*x[2]
        arg[1,4,0,1]=(0.914131238526)*x[0]**o+(-0.0280436066624)*x[0]+(0.861048515118)*x[1]**o+(0.148206807819)*x[1]+(-0.990856990354)*x[2]**o+(0.903615279172)*x[2]
        arg[1,4,1,0]=(0.189809824038)*x[0]**o+(-0.749477832258)*x[0]+(-0.802713794942)*x[1]**o+(0.0610727255572)*x[1]+(0.292940815327)*x[2]**o+(0.213443422205)*x[2]
        arg[1,4,1,1]=(0.792313292544)*x[0]**o+(0.321344675677)*x[0]+(-0.417516596097)*x[1]**o+(-0.63797359819)*x[1]+(-0.426112319872)*x[2]**o+(0.69281310288)*x[2]
        arg[1,4,2,0]=(-0.11017764073)*x[0]**o+(0.703037562344)*x[0]+(0.739591950873)*x[1]**o+(0.445612268331)*x[1]+(-0.071251825024)*x[2]**o+(0.00577350654185)*x[2]
        arg[1,4,2,1]=(-0.793584849832)*x[0]**o+(-0.546439256231)*x[0]+(-0.947194030803)*x[1]**o+(-0.0455700937677)*x[1]+(0.535368301918)*x[2]**o+(0.814435535521)*x[2]
        arg[2,0,0,0]=(0.249661949438)*x[0]**o+(-0.581533619092)*x[0]+(-0.674373505747)*x[1]**o+(0.464175602196)*x[1]+(-0.528086662834)*x[2]**o+(-0.212416176351)*x[2]
        arg[2,0,0,1]=(-0.588282108798)*x[0]**o+(0.761293399494)*x[0]+(0.99918874903)*x[1]**o+(0.542499992928)*x[1]+(0.433925746483)*x[2]**o+(0.972864491889)*x[2]
        arg[2,0,1,0]=(0.0572335623442)*x[0]**o+(0.444906443533)*x[0]+(-0.305538151235)*x[1]**o+(0.244530068532)*x[1]+(-0.20198975858)*x[2]**o+(-0.00490928477351)*x[2]
        arg[2,0,1,1]=(0.832496085695)*x[0]**o+(-0.797246004595)*x[0]+(0.873816520001)*x[1]**o+(0.628293733929)*x[1]+(-0.110922710624)*x[2]**o+(-0.0718221914062)*x[2]
        arg[2,0,2,0]=(0.140970192196)*x[0]**o+(0.582230323368)*x[0]+(-0.757764751267)*x[1]**o+(0.206061262231)*x[1]+(0.657237474788)*x[2]**o+(0.598632779121)*x[2]
        arg[2,0,2,1]=(-0.686218790734)*x[0]**o+(-0.157012470162)*x[0]+(0.448565721848)*x[1]**o+(0.0539087954076)*x[1]+(0.619821805007)*x[2]**o+(0.289144379283)*x[2]
        arg[2,1,0,0]=(0.787423884494)*x[0]**o+(-0.985959420171)*x[0]+(0.155584360983)*x[1]**o+(0.223292293778)*x[1]+(-0.437349719833)*x[2]**o+(0.989077189275)*x[2]
        arg[2,1,0,1]=(-0.597301631508)*x[0]**o+(-0.0414813766671)*x[0]+(-0.64883048443)*x[1]**o+(-0.18731615522)*x[1]+(0.735189823623)*x[2]**o+(-0.746633725819)*x[2]
        arg[2,1,1,0]=(-0.393486075452)*x[0]**o+(-0.480110060647)*x[0]+(0.559811870081)*x[1]**o+(0.0814282056993)*x[1]+(0.244899906416)*x[2]**o+(-0.884478480652)*x[2]
        arg[2,1,1,1]=(0.547747813175)*x[0]**o+(0.317890585903)*x[0]+(-0.70925099068)*x[1]**o+(-0.543227445702)*x[1]+(0.749693240223)*x[2]**o+(0.79984501038)*x[2]
        arg[2,1,2,0]=(-0.289583636666)*x[0]**o+(0.513926622265)*x[0]+(-0.00283544918865)*x[1]**o+(0.530724065811)*x[1]+(0.900479623243)*x[2]**o+(-0.146332301083)*x[2]
        arg[2,1,2,1]=(-0.487223003118)*x[0]**o+(0.161546483594)*x[0]+(-0.685211654772)*x[1]**o+(0.4803179816)*x[1]+(0.676120396917)*x[2]**o+(-0.174953085068)*x[2]
        arg[2,2,0,0]=(0.31746495834)*x[0]**o+(-0.513225182059)*x[0]+(0.33220677322)*x[1]**o+(-0.302371334854)*x[1]+(0.0964988251907)*x[2]**o+(-0.259532745762)*x[2]
        arg[2,2,0,1]=(0.313462846078)*x[0]**o+(0.079720902627)*x[0]+(-0.487932457563)*x[1]**o+(-0.786522786773)*x[1]+(0.357868214554)*x[2]**o+(0.251856050362)*x[2]
        arg[2,2,1,0]=(-0.180472378556)*x[0]**o+(0.345084802362)*x[0]+(0.641108557953)*x[1]**o+(0.198717742483)*x[1]+(0.0374114682326)*x[2]**o+(-0.915249907773)*x[2]
        arg[2,2,1,1]=(0.303787742688)*x[0]**o+(-0.941822487745)*x[0]+(-0.262567025842)*x[1]**o+(-0.197051445307)*x[1]+(0.381561965565)*x[2]**o+(-0.198549777232)*x[2]
        arg[2,2,2,0]=(-0.850284190461)*x[0]**o+(0.985224868544)*x[0]+(0.633868153422)*x[1]**o+(-0.57559927346)*x[1]+(0.68753726951)*x[2]**o+(-0.923696997656)*x[2]
        arg[2,2,2,1]=(0.559527478503)*x[0]**o+(-0.619764636475)*x[0]+(-0.0352163693945)*x[1]**o+(-0.37351931995)*x[1]+(0.0772333960938)*x[2]**o+(-0.353057755509)*x[2]
        arg[2,3,0,0]=(-0.274260374984)*x[0]**o+(0.365180349506)*x[0]+(-0.879564869545)*x[1]**o+(-0.0908162765131)*x[1]+(0.813491263836)*x[2]**o+(-0.396739585204)*x[2]
        arg[2,3,0,1]=(-0.793218219979)*x[0]**o+(0.227474218201)*x[0]+(0.97716610661)*x[1]**o+(0.317074032722)*x[1]+(0.076776446926)*x[2]**o+(0.069448617786)*x[2]
        arg[2,3,1,0]=(0.411400569812)*x[0]**o+(0.517743454926)*x[0]+(0.646697523033)*x[1]**o+(-0.973369313024)*x[1]+(-0.599741793926)*x[2]**o+(-0.743687020283)*x[2]
        arg[2,3,1,1]=(-0.491445354697)*x[0]**o+(-0.549840110396)*x[0]+(-0.260012915614)*x[1]**o+(0.671054626537)*x[1]+(0.0394966796969)*x[2]**o+(-0.701243931411)*x[2]
        arg[2,3,2,0]=(-0.806859723517)*x[0]**o+(-0.753940066715)*x[0]+(-0.664229274865)*x[1]**o+(0.0595484609931)*x[1]+(-0.921892262667)*x[2]**o+(-0.701770672395)*x[2]
        arg[2,3,2,1]=(-0.313979805844)*x[0]**o+(-0.542583864222)*x[0]+(-0.985990886553)*x[1]**o+(0.664926443979)*x[1]+(-0.99778334615)*x[2]**o+(0.268046810979)*x[2]
        arg[2,4,0,0]=(-0.41348240167)*x[0]**o+(0.473517720534)*x[0]+(-0.290880345559)*x[1]**o+(-0.607611624107)*x[1]+(-0.25537697471)*x[2]**o+(-0.472402477362)*x[2]
        arg[2,4,0,1]=(0.0222534371058)*x[0]**o+(-0.0739354104963)*x[0]+(-0.960057765718)*x[1]**o+(-0.772161403037)*x[1]+(-0.0356127352146)*x[2]**o+(-0.623605628308)*x[2]
        arg[2,4,1,0]=(-0.750995516053)*x[0]**o+(-0.289601799215)*x[0]+(0.294370180152)*x[1]**o+(0.956012147239)*x[1]+(-0.27958662106)*x[2]**o+(-0.370609593537)*x[2]
        arg[2,4,1,1]=(0.321767371741)*x[0]**o+(0.999223434136)*x[0]+(0.665865428241)*x[1]**o+(0.427942082269)*x[1]+(-0.135020552387)*x[2]**o+(0.692025479514)*x[2]
        arg[2,4,2,0]=(-0.519508249649)*x[0]**o+(0.0779688946471)*x[0]+(0.493489242872)*x[1]**o+(-0.0961136530093)*x[1]+(-0.0750243559261)*x[2]**o+(0.860970740104)*x[2]
        arg[2,4,2,1]=(0.807909049346)*x[0]**o+(0.388180877227)*x[0]+(-0.9044578433)*x[1]**o+(0.86765389355)*x[1]+(-0.196105779343)*x[2]**o+(0.448661151455)*x[2]
        arg[3,0,0,0]=(-0.780895567041)*x[0]**o+(-0.571253266808)*x[0]+(-0.904183399932)*x[1]**o+(-0.883296586722)*x[1]+(0.855945559487)*x[2]**o+(0.618533553098)*x[2]
        arg[3,0,0,1]=(-0.831076704037)*x[0]**o+(0.898347446798)*x[0]+(-0.716442620396)*x[1]**o+(0.761028369421)*x[1]+(-0.486505604444)*x[2]**o+(0.25474220145)*x[2]
        arg[3,0,1,0]=(-0.0127541044953)*x[0]**o+(0.759482315308)*x[0]+(-0.725655006866)*x[1]**o+(0.667310781285)*x[1]+(0.266075927439)*x[2]**o+(0.423300182252)*x[2]
        arg[3,0,1,1]=(-0.594972266882)*x[0]**o+(-0.842501149639)*x[0]+(-0.347686790781)*x[1]**o+(0.686403640474)*x[1]+(0.3452971966)*x[2]**o+(0.722258585706)*x[2]
        arg[3,0,2,0]=(-0.715711146058)*x[0]**o+(-0.835739963371)*x[0]+(0.966177361301)*x[1]**o+(-0.909111754995)*x[1]+(-0.175184576557)*x[2]**o+(0.695058239921)*x[2]
        arg[3,0,2,1]=(0.306251544915)*x[0]**o+(0.944775715803)*x[0]+(-0.740601754617)*x[1]**o+(0.434394890125)*x[1]+(-0.876527474971)*x[2]**o+(0.209321129689)*x[2]
        arg[3,1,0,0]=(-0.86788221871)*x[0]**o+(-0.833030657309)*x[0]+(0.84576621947)*x[1]**o+(-0.469186887277)*x[1]+(-0.527611740694)*x[2]**o+(-0.312054431001)*x[2]
        arg[3,1,0,1]=(-0.353678394652)*x[0]**o+(0.591287444611)*x[0]+(-0.213012375931)*x[1]**o+(-0.917259099251)*x[1]+(0.739016360622)*x[2]**o+(0.718754459378)*x[2]
        arg[3,1,1,0]=(-0.92646825221)*x[0]**o+(-0.870564973175)*x[0]+(-0.781805219404)*x[1]**o+(0.078530434642)*x[1]+(0.386018501087)*x[2]**o+(-0.120490492288)*x[2]
        arg[3,1,1,1]=(-0.824273737615)*x[0]**o+(0.401211543421)*x[0]+(0.360077381624)*x[1]**o+(-0.0599234883152)*x[1]+(0.246820480884)*x[2]**o+(-0.175313542134)*x[2]
        arg[3,1,2,0]=(0.807756928769)*x[0]**o+(0.889534807508)*x[0]+(0.633296189954)*x[1]**o+(-0.471821951252)*x[1]+(0.396011249874)*x[2]**o+(0.386800809575)*x[2]
        arg[3,1,2,1]=(-0.494145142543)*x[0]**o+(-0.514793985721)*x[0]+(0.825379848403)*x[1]**o+(-0.457860259786)*x[1]+(-0.599557868275)*x[2]**o+(-0.952704784672)*x[2]
        arg[3,2,0,0]=(0.970345295013)*x[0]**o+(-0.838468036663)*x[0]+(0.265634348738)*x[1]**o+(-0.873847124748)*x[1]+(0.416242648913)*x[2]**o+(0.59908438222)*x[2]
        arg[3,2,0,1]=(0.692377432429)*x[0]**o+(-0.338749041227)*x[0]+(-0.449383750398)*x[1]**o+(0.841669527559)*x[1]+(0.902482059741)*x[2]**o+(-0.655719647704)*x[2]
        arg[3,2,1,0]=(0.152417917522)*x[0]**o+(-0.78860144)*x[0]+(-0.881276170743)*x[1]**o+(0.692901550061)*x[1]+(-0.985377819322)*x[2]**o+(0.362747973601)*x[2]
        arg[3,2,1,1]=(0.238845792119)*x[0]**o+(-0.65838373309)*x[0]+(-0.659676514296)*x[1]**o+(0.870742876778)*x[1]+(-0.367345415947)*x[2]**o+(0.322353666541)*x[2]
        arg[3,2,2,0]=(0.456386394533)*x[0]**o+(-0.135575574979)*x[0]+(-0.731714323922)*x[1]**o+(0.936807859479)*x[1]+(0.511595594967)*x[2]**o+(0.570282464669)*x[2]
        arg[3,2,2,1]=(0.53299212624)*x[0]**o+(0.800998166434)*x[0]+(-0.431740087552)*x[1]**o+(-0.880092532661)*x[1]+(-0.418657988674)*x[2]**o+(-0.181008149338)*x[2]
        arg[3,3,0,0]=(0.60015312936)*x[0]**o+(-0.163125583081)*x[0]+(0.299668981953)*x[1]**o+(-0.761494422747)*x[1]+(0.404839953231)*x[2]**o+(-0.114039836538)*x[2]
        arg[3,3,0,1]=(-0.597481072133)*x[0]**o+(-0.506468582896)*x[0]+(0.406822348574)*x[1]**o+(0.247431348859)*x[1]+(-0.137941563908)*x[2]**o+(0.0503979259383)*x[2]
        arg[3,3,1,0]=(0.0921069092744)*x[0]**o+(-0.113960490794)*x[0]+(0.750059773909)*x[1]**o+(0.894945048248)*x[1]+(-0.154873449786)*x[2]**o+(-0.520229467775)*x[2]
        arg[3,3,1,1]=(0.349078696154)*x[0]**o+(0.141298430827)*x[0]+(0.249161007832)*x[1]**o+(0.0240310278591)*x[1]+(-0.18294527838)*x[2]**o+(-0.666698832632)*x[2]
        arg[3,3,2,0]=(-0.476684936147)*x[0]**o+(0.334093402615)*x[0]+(-0.326434684009)*x[1]**o+(0.91336431975)*x[1]+(0.483080537399)*x[2]**o+(0.439096732335)*x[2]
        arg[3,3,2,1]=(0.214713426286)*x[0]**o+(-0.973239849984)*x[0]+(0.984290637567)*x[1]**o+(-0.644668525299)*x[1]+(0.00985524319943)*x[2]**o+(0.68777403533)*x[2]
        arg[3,4,0,0]=(-0.88715423933)*x[0]**o+(0.642412222188)*x[0]+(-0.119667464132)*x[1]**o+(-0.201583116481)*x[1]+(-0.870921042381)*x[2]**o+(-0.00379039429884)*x[2]
        arg[3,4,0,1]=(0.511114986801)*x[0]**o+(-0.497617928997)*x[0]+(-0.189435233543)*x[1]**o+(-0.29288141421)*x[1]+(-0.476297664539)*x[2]**o+(-0.202915859137)*x[2]
        arg[3,4,1,0]=(0.0739878219028)*x[0]**o+(0.675050675951)*x[0]+(-0.45950101521)*x[1]**o+(0.306538598103)*x[1]+(-0.726069135031)*x[2]**o+(0.193290422951)*x[2]
        arg[3,4,1,1]=(-0.141324267565)*x[0]**o+(0.377343882109)*x[0]+(-0.16075604872)*x[1]**o+(0.263332274591)*x[1]+(0.886416182992)*x[2]**o+(0.931905996631)*x[2]
        arg[3,4,2,0]=(-0.76912732001)*x[0]**o+(-0.255011103121)*x[0]+(-0.372810348364)*x[1]**o+(0.108137766687)*x[1]+(-0.0761382180171)*x[2]**o+(0.46825331282)*x[2]
        arg[3,4,2,1]=(-0.515945940217)*x[0]**o+(-0.630017609554)*x[0]+(-0.815236764508)*x[1]**o+(0.636572471005)*x[1]+(0.244112465789)*x[2]**o+(-0.471661017589)*x[2]
        ref[0,0,0,0]=(-0.738709393227)/(o+1.)+(0.283993125638)+(-0.891975745754)*0.5**o
        ref[0,0,0,1]=(0.912832356419)/(o+1.)+(-0.800452287951)+(0.428698930349)*0.5**o
        ref[0,0,1,0]=(0.604410785754)/(o+1.)+(-0.70496852873)+(0.974069118653)*0.5**o
        ref[0,0,1,1]=(-1.63466543191)/(o+1.)+(0.302144663639)+(0.813412823502)*0.5**o
        ref[0,0,2,0]=(0.755754363963)/(o+1.)+(0.151022836528)+(-0.874502916672)*0.5**o
        ref[0,0,2,1]=(0.958914457666)/(o+1.)+(-0.440871585564)+(-0.258524075807)*0.5**o
        ref[0,1,0,0]=(-0.154165868092)/(o+1.)+(0.484258240357)+(0.489505903018)*0.5**o
        ref[0,1,0,1]=(0.79058715248)/(o+1.)+(0.686313504396)+(0.441567332096)*0.5**o
        ref[0,1,1,0]=(-0.834798928973)/(o+1.)+(0.492360141946)+(0.388576628492)*0.5**o
        ref[0,1,1,1]=(-0.315522016845)/(o+1.)+(0.688152306846)+(-0.0537780037583)*0.5**o
        ref[0,1,2,0]=(0.203060170914)/(o+1.)+(0.198021868056)+(0.58262140787)*0.5**o
        ref[0,1,2,1]=(-0.665732288492)/(o+1.)+(0.798951978927)+(0.438472233911)*0.5**o
        ref[0,2,0,0]=(-1.1918215501)/(o+1.)+(0.114544359058)+(-0.334110476968)*0.5**o
        ref[0,2,0,1]=(0.0771737213265)/(o+1.)+(-0.711198951588)+(0.33041258354)*0.5**o
        ref[0,2,1,0]=(1.1366448279)/(o+1.)+(0.141765838337)+(-0.80021736912)*0.5**o
        ref[0,2,1,1]=(-0.432635401002)/(o+1.)+(0.0361325240497)+(-0.773331228218)*0.5**o
        ref[0,2,2,0]=(1.37338816519)/(o+1.)+(-0.0827542033711)+(-0.943072868965)*0.5**o
        ref[0,2,2,1]=(-0.303620293463)/(o+1.)+(-0.39585636903)+(-0.821201079286)*0.5**o
        ref[0,3,0,0]=(-1.74869869132)/(o+1.)+(-0.845702848785)+(-0.0564164524098)*0.5**o
        ref[0,3,0,1]=(0.499654298722)/(o+1.)+(0.0895873531986)+(-0.553858251068)*0.5**o
        ref[0,3,1,0]=(-0.277226290791)/(o+1.)+(-0.755072202588)+(-0.992317453163)*0.5**o
        ref[0,3,1,1]=(-0.412604788342)/(o+1.)+(0.496792347901)+(-0.713753922213)*0.5**o
        ref[0,3,2,0]=(-0.130885078624)/(o+1.)+(0.312575459575)+(0.536024118289)*0.5**o
        ref[0,3,2,1]=(-1.35620016195)/(o+1.)+(-0.802557814945)+(-0.197968756117)*0.5**o
        ref[0,4,0,0]=(0.214746716397)/(o+1.)+(-0.0864879979042)+(0.688763655127)*0.5**o
        ref[0,4,0,1]=(0.101745206477)/(o+1.)+(-0.431035828594)+(-0.402717742839)*0.5**o
        ref[0,4,1,0]=(-1.71496447458)/(o+1.)+(-0.210739324578)+(0.39352838659)*0.5**o
        ref[0,4,1,1]=(0.516179257264)/(o+1.)+(-0.271979584839)+(-0.375507250175)*0.5**o
        ref[0,4,2,0]=(0.0175356956523)/(o+1.)+(0.255449591592)+(0.0603590263818)*0.5**o
        ref[0,4,2,1]=(0.36499082659)/(o+1.)+(-0.11897969607)+(-0.111089454685)*0.5**o
        ref[1,0,0,0]=(1.12671976333)/(o+1.)+(0.275654183538)+(-0.940865155622)*0.5**o
        ref[1,0,0,1]=(-1.58477430959)/(o+1.)+(-0.471320510546)+(-0.843207880309)*0.5**o
        ref[1,0,1,0]=(0.126570690864)/(o+1.)+(0.505250661477)+(-0.484987528212)*0.5**o
        ref[1,0,1,1]=(-0.807547181539)/(o+1.)+(-0.728382819379)+(0.285548625366)*0.5**o
        ref[1,0,2,0]=(0.151618483009)/(o+1.)+(0.213905131461)+(0.96797757731)*0.5**o
        ref[1,0,2,1]=(0.740661511824)/(o+1.)+(-0.789278288097)+(-0.486190283922)*0.5**o
        ref[1,1,0,0]=(0.0814414907926)/(o+1.)+(-0.0398533374263)+(-0.216544472879)*0.5**o
        ref[1,1,0,1]=(-1.09726915206)/(o+1.)+(0.108361078938)+(0.982934604971)*0.5**o
        ref[1,1,1,0]=(1.75511303167)/(o+1.)+(-0.201459228508)+(0.516177509446)*0.5**o
        ref[1,1,1,1]=(-1.16005664849)/(o+1.)+(0.248930524148)+(-0.914732890633)*0.5**o
        ref[1,1,2,0]=(-0.78384665822)/(o+1.)+(-0.0461152674588)+(-0.96655555619)*0.5**o
        ref[1,1,2,1]=(0.0633477168503)/(o+1.)+(0.576397965591)+(-0.611879453709)*0.5**o
        ref[1,2,0,0]=(-0.707964350108)/(o+1.)+(-0.506456693317)+(0.558820946357)*0.5**o
        ref[1,2,0,1]=(-0.352303090063)/(o+1.)+(-0.881076193856)+(-0.944738664192)*0.5**o
        ref[1,2,1,0]=(-0.283612324861)/(o+1.)+(-0.395528636894)+(0.238225961058)*0.5**o
        ref[1,2,1,1]=(-0.0917638265383)/(o+1.)+(0.328859628638)+(0.188699181238)*0.5**o
        ref[1,2,2,0]=(1.37828110785)/(o+1.)+(-0.39071397624)+(-0.0953160757881)*0.5**o
        ref[1,2,2,1]=(0.138642388492)/(o+1.)+(0.585831787766)+(0.598691886555)*0.5**o
        ref[1,3,0,0]=(-0.165396693296)/(o+1.)+(-0.248292538454)+(-0.482944276372)*0.5**o
        ref[1,3,0,1]=(1.01718528032)/(o+1.)+(-0.282549777519)+(0.565128725562)*0.5**o
        ref[1,3,1,0]=(0.634094178049)/(o+1.)+(0.419667273293)+(0.608648887224)*0.5**o
        ref[1,3,1,1]=(1.21240546445)/(o+1.)+(0.516304368664)+(0.0322521291544)*0.5**o
        ref[1,3,2,0]=(-0.870073605065)/(o+1.)+(0.568353556126)+(-0.757022609619)*0.5**o
        ref[1,3,2,1]=(0.412342060102)/(o+1.)+(-0.392777941868)+(-0.395945302352)*0.5**o
        ref[1,4,0,0]=(-0.0939089748304)/(o+1.)+(0.20572535845)+(0.0462127928202)*0.5**o
        ref[1,4,0,1]=(-0.129808475236)/(o+1.)+(0.511889240164)+(0.914131238526)*0.5**o
        ref[1,4,1,0]=(-0.509772979615)/(o+1.)+(-0.237480842248)+(0.189809824038)*0.5**o
        ref[1,4,1,1]=(-0.843628915969)/(o+1.)+(0.188092090183)+(0.792313292544)*0.5**o
        ref[1,4,2,0]=(0.668340125849)/(o+1.)+(0.577211668608)+(-0.11017764073)*0.5**o
        ref[1,4,2,1]=(-0.411825728886)/(o+1.)+(0.111213092761)+(-0.793584849832)*0.5**o
        ref[2,0,0,0]=(-1.20246016858)/(o+1.)+(-0.164887096623)+(0.249661949438)*0.5**o
        ref[2,0,0,1]=(1.43311449551)/(o+1.)+(1.13832894216)+(-0.588282108798)*0.5**o
        ref[2,0,1,0]=(-0.507527909815)/(o+1.)+(0.342263613646)+(0.0572335623442)*0.5**o
        ref[2,0,1,1]=(0.762893809377)/(o+1.)+(-0.120387231036)+(0.832496085695)*0.5**o
        ref[2,0,2,0]=(-0.100527276479)/(o+1.)+(0.69346218236)+(0.140970192196)*0.5**o
        ref[2,0,2,1]=(1.06838752686)/(o+1.)+(0.0930203522642)+(-0.686218790734)*0.5**o
        ref[2,1,0,0]=(-0.28176535885)/(o+1.)+(0.113205031441)+(0.787423884494)*0.5**o
        ref[2,1,0,1]=(0.0863593391922)/(o+1.)+(-0.487715628853)+(-0.597301631508)*0.5**o
        ref[2,1,1,0]=(0.804711776497)/(o+1.)+(-0.6415801678)+(-0.393486075452)*0.5**o
        ref[2,1,1,1]=(0.0404422495423)/(o+1.)+(0.28725407529)+(0.547747813175)*0.5**o
        ref[2,1,2,0]=(0.897644174054)/(o+1.)+(0.449159193497)+(-0.289583636666)*0.5**o
        ref[2,1,2,1]=(-0.00909125785473)/(o+1.)+(0.233455690063)+(-0.487223003118)*0.5**o
        ref[2,2,0,0]=(0.428705598411)/(o+1.)+(-0.537564631337)+(0.31746495834)*0.5**o
        ref[2,2,0,1]=(-0.13006424301)/(o+1.)+(-0.227472916892)+(0.313462846078)*0.5**o
        ref[2,2,1,0]=(0.678520026185)/(o+1.)+(-0.185723681464)+(-0.180472378556)*0.5**o
        ref[2,2,1,1]=(0.118994939723)/(o+1.)+(-0.668711855142)+(0.303787742688)*0.5**o
        ref[2,2,2,0]=(1.32140542293)/(o+1.)+(-0.257035701286)+(-0.850284190461)*0.5**o
        ref[2,2,2,1]=(0.0420170266993)/(o+1.)+(-0.673170855967)+(0.559527478503)*0.5**o
        ref[2,3,0,0]=(-0.0660736057092)/(o+1.)+(-0.0611877561056)+(-0.274260374984)*0.5**o
        ref[2,3,0,1]=(1.05394255354)/(o+1.)+(0.306998434354)+(-0.793218219979)*0.5**o
        ref[2,3,1,0]=(0.046955729107)/(o+1.)+(-0.599656439191)+(0.411400569812)*0.5**o
        ref[2,3,1,1]=(-0.220516235917)/(o+1.)+(-0.290014707635)+(-0.491445354697)*0.5**o
        ref[2,3,2,0]=(-1.58612153753)/(o+1.)+(-0.698081139059)+(-0.806859723517)*0.5**o
        ref[2,3,2,1]=(-1.9837742327)/(o+1.)+(0.195194695368)+(-0.313979805844)*0.5**o
        ref[2,4,0,0]=(-0.546257320269)/(o+1.)+(-0.303248190467)+(-0.41348240167)*0.5**o
        ref[2,4,0,1]=(-0.995670500932)/(o+1.)+(-0.734851220921)+(0.0222534371058)*0.5**o
        ref[2,4,1,0]=(0.0147835590923)/(o+1.)+(0.147900377243)+(-0.750995516053)*0.5**o
        ref[2,4,1,1]=(0.530844875854)/(o+1.)+(1.05959549796)+(0.321767371741)*0.5**o
        ref[2,4,2,0]=(0.418464886946)/(o+1.)+(0.421412990871)+(-0.519508249649)*0.5**o
        ref[2,4,2,1]=(-1.10056362264)/(o+1.)+(0.852247961116)+(0.807909049346)*0.5**o
        ref[3,0,0,0]=(-0.0482378404454)/(o+1.)+(-0.418008150216)+(-0.780895567041)*0.5**o
        ref[3,0,0,1]=(-1.20294822484)/(o+1.)+(0.957059008835)+(-0.831076704037)*0.5**o
        ref[3,0,1,0]=(-0.459579079427)/(o+1.)+(0.925046639422)+(-0.0127541044953)*0.5**o
        ref[3,0,1,1]=(-0.00238959418047)/(o+1.)+(0.28308053827)+(-0.594972266882)*0.5**o
        ref[3,0,2,0]=(0.790992784744)/(o+1.)+(-0.524896739222)+(-0.715711146058)*0.5**o
        ref[3,0,2,1]=(-1.61712922959)/(o+1.)+(0.794245867809)+(0.306251544915)*0.5**o
        ref[3,1,0,0]=(0.318154478777)/(o+1.)+(-0.807135987793)+(-0.86788221871)*0.5**o
        ref[3,1,0,1]=(0.526003984691)/(o+1.)+(0.196391402369)+(-0.353678394652)*0.5**o
        ref[3,1,1,0]=(-0.395786718317)/(o+1.)+(-0.456262515411)+(-0.92646825221)*0.5**o
        ref[3,1,1,1]=(0.606897862508)/(o+1.)+(0.082987256486)+(-0.824273737615)*0.5**o
        ref[3,1,2,0]=(1.02930743983)/(o+1.)+(0.402256832916)+(0.807756928769)*0.5**o
        ref[3,1,2,1]=(0.225821980128)/(o+1.)+(-0.96267951509)+(-0.494145142543)*0.5**o
        ref[3,2,0,0]=(0.681876997651)/(o+1.)+(-0.556615389596)+(0.970345295013)*0.5**o
        ref[3,2,0,1]=(0.453098309344)/(o+1.)+(-0.0763995806862)+(0.692377432429)*0.5**o
        ref[3,2,1,0]=(-1.86665399006)/(o+1.)+(0.133524041831)+(0.152417917522)*0.5**o
        ref[3,2,1,1]=(-1.02702193024)/(o+1.)+(0.267356405114)+(0.238845792119)*0.5**o
        ref[3,2,2,0]=(-0.220118728954)/(o+1.)+(0.685757374584)+(0.456386394533)*0.5**o
        ref[3,2,2,1]=(-0.850398076225)/(o+1.)+(-0.130051257782)+(0.53299212624)*0.5**o
        ref[3,3,0,0]=(0.704508935183)/(o+1.)+(-0.519329921183)+(0.60015312936)*0.5**o
        ref[3,3,0,1]=(0.268880784666)/(o+1.)+(-0.10431965405)+(-0.597481072133)*0.5**o
        ref[3,3,1,0]=(0.595186324123)/(o+1.)+(0.13037754484)+(0.0921069092744)*0.5**o
        ref[3,3,1,1]=(0.0662157294518)/(o+1.)+(-0.250684686973)+(0.349078696154)*0.5**o
        ref[3,3,2,0]=(0.156645853389)/(o+1.)+(0.84327722735)+(-0.476684936147)*0.5**o
        ref[3,3,2,1]=(0.994145880767)/(o+1.)+(-0.465067169976)+(0.214713426286)*0.5**o
        ref[3,4,0,0]=(-0.990588506513)/(o+1.)+(0.218519355704)+(-0.88715423933)*0.5**o
        ref[3,4,0,1]=(-0.665732898082)/(o+1.)+(-0.496707601172)+(0.511114986801)*0.5**o
        ref[3,4,1,0]=(-1.18557015024)/(o+1.)+(0.587439848502)+(0.0739878219028)*0.5**o
        ref[3,4,1,1]=(0.725660134271)/(o+1.)+(0.786291076666)+(-0.141324267565)*0.5**o
        ref[3,4,2,0]=(-0.448948566381)/(o+1.)+(0.160689988193)+(-0.76912732001)*0.5**o
        ref[3,4,2,1]=(-0.571124298719)/(o+1.)+(-0.232553078069)+(-0.515945940217)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0333010052777)*x[0]**o+(-0.242011298555)*x[0]+(0.686457578633)*x[1]**o+(0.438506741453)*x[1]
        ref=(0.686457578633)/(o+1.)+(0.0982477214491)+(-0.0333010052777)*0.5**o
      else:
        arg=(-0.365813216932)*x[0]**o+(0.969816151161)*x[0]+(0.349995764077)*x[1]**o+(0.867492181692)*x[1]+(-0.866177355796)*x[2]**o+(-0.203219031102)*x[2]
        ref=(-0.516181591719)/(o+1.)+(0.817044650875)+(-0.365813216932)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.038446371199)*x[0]**o+(0.323075821067)*x[0]+(0.770528210463)*x[1]**o+(0.0434528043023)*x[1]
        arg[1]=(-0.461876402566)*x[0]**o+(0.746965376214)*x[0]+(0.861234840159)*x[1]**o+(-0.568165009958)*x[1]
        ref[0]=(0.770528210463)/(o+1.)+(0.183264312685)+(0.038446371199)*0.5**o
        ref[1]=(0.861234840159)/(o+1.)+(0.0894001831281)+(-0.461876402566)*0.5**o
      else:
        arg[0]=(-0.472858663642)*x[0]**o+(-0.680210745369)*x[0]+(0.175425838082)*x[1]**o+(0.0189427039667)*x[1]+(0.0365198456146)*x[2]**o+(0.7840527457)*x[2]
        arg[1]=(-0.0191775972693)*x[0]**o+(0.19303923194)*x[0]+(0.775528295333)*x[1]**o+(-0.765500708496)*x[1]+(0.351369471606)*x[2]**o+(-0.351587727906)*x[2]
        ref[0]=(0.211945683696)/(o+1.)+(0.0613923521485)+(-0.472858663642)*0.5**o
        ref[1]=(1.12689776694)/(o+1.)+(-0.462024602231)+(-0.0191775972693)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.994443255559)*x[0]**o+(-0.403427184912)*x[0]+(-0.889656446862)*x[1]**o+(-0.411934106034)*x[1]
        arg[0,1]=(-0.864086447061)*x[0]**o+(-0.277489357308)*x[0]+(0.977910864864)*x[1]**o+(-0.0505775488074)*x[1]
        arg[0,2]=(-0.397179114534)*x[0]**o+(-0.448046821439)*x[0]+(-0.340253609736)*x[1]**o+(0.624988837716)*x[1]
        arg[0,3]=(-0.9155191116)*x[0]**o+(0.133777450278)*x[0]+(-0.809567026065)*x[1]**o+(-0.815374059364)*x[1]
        arg[0,4]=(-0.416764684181)*x[0]**o+(0.0965164570638)*x[0]+(0.190191799833)*x[1]**o+(0.0970709424519)*x[1]
        arg[1,0]=(0.646527294535)*x[0]**o+(-0.337119506636)*x[0]+(0.42669563245)*x[1]**o+(-0.185228568706)*x[1]
        arg[1,1]=(-0.411532647934)*x[0]**o+(0.178830973832)*x[0]+(-0.904886020892)*x[1]**o+(-0.740623071275)*x[1]
        arg[1,2]=(0.523024622808)*x[0]**o+(-0.226298848077)*x[0]+(-0.707740568617)*x[1]**o+(0.998182796571)*x[1]
        arg[1,3]=(0.840649917357)*x[0]**o+(-0.736279151571)*x[0]+(0.996017554392)*x[1]**o+(-0.397833319931)*x[1]
        arg[1,4]=(0.76711764419)*x[0]**o+(-0.905452991298)*x[0]+(-0.925675237951)*x[1]**o+(-0.982712645237)*x[1]
        arg[2,0]=(0.198547103682)*x[0]**o+(0.674537682234)*x[0]+(-0.891195618193)*x[1]**o+(0.979109691481)*x[1]
        arg[2,1]=(0.171629921737)*x[0]**o+(0.0102727001751)*x[0]+(0.467578393374)*x[1]**o+(0.583960984645)*x[1]
        arg[2,2]=(-0.637526271497)*x[0]**o+(-0.900821073531)*x[0]+(0.61637733965)*x[1]**o+(0.43717505341)*x[1]
        arg[2,3]=(-0.205861963967)*x[0]**o+(-0.104621062003)*x[0]+(0.582584591946)*x[1]**o+(0.351739662419)*x[1]
        arg[2,4]=(-0.179355911316)*x[0]**o+(-0.188415282678)*x[0]+(0.598342895007)*x[1]**o+(0.364638908577)*x[1]
        arg[3,0]=(0.150471526119)*x[0]**o+(-0.940789873202)*x[0]+(-0.524275825864)*x[1]**o+(0.729004683023)*x[1]
        arg[3,1]=(0.123933417087)*x[0]**o+(0.192413160498)*x[0]+(0.203730760507)*x[1]**o+(0.852010172966)*x[1]
        arg[3,2]=(-0.297766950847)*x[0]**o+(0.269905131596)*x[0]+(-0.0830931256847)*x[1]**o+(0.26964712569)*x[1]
        arg[3,3]=(0.96961517265)*x[0]**o+(0.363312967302)*x[0]+(-0.811142762589)*x[1]**o+(-0.420442983537)*x[1]
        arg[3,4]=(-0.314185602206)*x[0]**o+(-0.0446515646101)*x[0]+(-0.884651038732)*x[1]**o+(-0.166386477611)*x[1]
        ref[0,0]=(-0.889656446862)/(o+1.)+(-0.407680645473)+(-0.994443255559)*0.5**o
        ref[0,1]=(0.977910864864)/(o+1.)+(-0.164033453058)+(-0.864086447061)*0.5**o
        ref[0,2]=(-0.340253609736)/(o+1.)+(0.0884710081389)+(-0.397179114534)*0.5**o
        ref[0,3]=(-0.809567026065)/(o+1.)+(-0.340798304543)+(-0.9155191116)*0.5**o
        ref[0,4]=(0.190191799833)/(o+1.)+(0.0967936997578)+(-0.416764684181)*0.5**o
        ref[1,0]=(0.42669563245)/(o+1.)+(-0.261174037671)+(0.646527294535)*0.5**o
        ref[1,1]=(-0.904886020892)/(o+1.)+(-0.280896048722)+(-0.411532647934)*0.5**o
        ref[1,2]=(-0.707740568617)/(o+1.)+(0.385941974247)+(0.523024622808)*0.5**o
        ref[1,3]=(0.996017554392)/(o+1.)+(-0.567056235751)+(0.840649917357)*0.5**o
        ref[1,4]=(-0.925675237951)/(o+1.)+(-0.944082818268)+(0.76711764419)*0.5**o
        ref[2,0]=(-0.891195618193)/(o+1.)+(0.826823686857)+(0.198547103682)*0.5**o
        ref[2,1]=(0.467578393374)/(o+1.)+(0.29711684241)+(0.171629921737)*0.5**o
        ref[2,2]=(0.61637733965)/(o+1.)+(-0.231823010061)+(-0.637526271497)*0.5**o
        ref[2,3]=(0.582584591946)/(o+1.)+(0.123559300208)+(-0.205861963967)*0.5**o
        ref[2,4]=(0.598342895007)/(o+1.)+(0.0881118129495)+(-0.179355911316)*0.5**o
        ref[3,0]=(-0.524275825864)/(o+1.)+(-0.105892595089)+(0.150471526119)*0.5**o
        ref[3,1]=(0.203730760507)/(o+1.)+(0.522211666732)+(0.123933417087)*0.5**o
        ref[3,2]=(-0.0830931256847)/(o+1.)+(0.269776128643)+(-0.297766950847)*0.5**o
        ref[3,3]=(-0.811142762589)/(o+1.)+(-0.0285650081177)+(0.96961517265)*0.5**o
        ref[3,4]=(-0.884651038732)/(o+1.)+(-0.105519021111)+(-0.314185602206)*0.5**o
      else:
        arg[0,0]=(0.264691789613)*x[0]**o+(0.816541340997)*x[0]+(-0.402990464667)*x[1]**o+(-0.57479619687)*x[1]+(0.336313666995)*x[2]**o+(-0.353125285536)*x[2]
        arg[0,1]=(-0.0526559162494)*x[0]**o+(-0.901336706583)*x[0]+(0.237482536727)*x[1]**o+(-0.458564275773)*x[1]+(-0.90715992057)*x[2]**o+(-0.352062856077)*x[2]
        arg[0,2]=(0.786349112412)*x[0]**o+(-0.395569856472)*x[0]+(-0.996593023157)*x[1]**o+(-0.277001350627)*x[1]+(0.435375913289)*x[2]**o+(0.00380742297736)*x[2]
        arg[0,3]=(0.3336751197)*x[0]**o+(0.592185657129)*x[0]+(0.148956266814)*x[1]**o+(-0.852264651722)*x[1]+(-0.680249245511)*x[2]**o+(0.238843544413)*x[2]
        arg[0,4]=(0.875206676786)*x[0]**o+(0.0420423017124)*x[0]+(-0.690488811412)*x[1]**o+(-0.191328696331)*x[1]+(-0.861423624177)*x[2]**o+(0.483925405411)*x[2]
        arg[1,0]=(-0.573216715158)*x[0]**o+(-0.0668114726198)*x[0]+(0.382097918101)*x[1]**o+(-0.126485626438)*x[1]+(0.37521336616)*x[2]**o+(0.42402066058)*x[2]
        arg[1,1]=(-0.0919326514467)*x[0]**o+(-0.791351880042)*x[0]+(0.772694542694)*x[1]**o+(-0.537902003787)*x[1]+(0.424413342503)*x[2]**o+(-0.285553230731)*x[2]
        arg[1,2]=(0.855490362389)*x[0]**o+(0.60428104628)*x[0]+(-0.846565232506)*x[1]**o+(-0.243727717758)*x[1]+(0.900077399935)*x[2]**o+(-0.974305489047)*x[2]
        arg[1,3]=(-0.90802343526)*x[0]**o+(0.00950587842259)*x[0]+(0.821234711321)*x[1]**o+(-0.0628912085092)*x[1]+(-0.773563396868)*x[2]**o+(-0.403955952419)*x[2]
        arg[1,4]=(-0.33139358787)*x[0]**o+(-0.901382943687)*x[0]+(-0.981921375707)*x[1]**o+(-0.0654519807133)*x[1]+(0.880069861158)*x[2]**o+(-0.0796172270076)*x[2]
        arg[2,0]=(-0.202818692518)*x[0]**o+(-0.155436611971)*x[0]+(0.739467305515)*x[1]**o+(-0.12096877475)*x[1]+(0.479644796804)*x[2]**o+(0.0580715322389)*x[2]
        arg[2,1]=(0.0461326006936)*x[0]**o+(0.122168224674)*x[0]+(0.986788073931)*x[1]**o+(-0.435933938764)*x[1]+(-0.312527130215)*x[2]**o+(0.726805187165)*x[2]
        arg[2,2]=(-0.539659153805)*x[0]**o+(-0.26082353681)*x[0]+(0.367739719819)*x[1]**o+(0.943584828354)*x[1]+(0.688225221556)*x[2]**o+(-0.121563525696)*x[2]
        arg[2,3]=(0.861717128203)*x[0]**o+(0.0666983176426)*x[0]+(0.984510848927)*x[1]**o+(-0.321865378029)*x[1]+(0.522584734)*x[2]**o+(0.589527116625)*x[2]
        arg[2,4]=(0.778575699703)*x[0]**o+(0.515802361664)*x[0]+(0.817655905624)*x[1]**o+(0.883824717911)*x[1]+(0.656833761714)*x[2]**o+(-0.991585145611)*x[2]
        arg[3,0]=(0.752875513075)*x[0]**o+(0.615136567082)*x[0]+(-0.205265870704)*x[1]**o+(0.483918142728)*x[1]+(-0.0495322147557)*x[2]**o+(-0.132474571954)*x[2]
        arg[3,1]=(0.0563875061408)*x[0]**o+(-0.603033440119)*x[0]+(-0.0944809376206)*x[1]**o+(0.725330686731)*x[1]+(0.00527671533343)*x[2]**o+(-0.434519675586)*x[2]
        arg[3,2]=(0.100078550382)*x[0]**o+(-0.340988639989)*x[0]+(-0.392591649097)*x[1]**o+(-0.924705734588)*x[1]+(0.420410192075)*x[2]**o+(-0.94741221593)*x[2]
        arg[3,3]=(0.536414145801)*x[0]**o+(0.142421640429)*x[0]+(0.213162255423)*x[1]**o+(-0.695193008147)*x[1]+(0.9625091599)*x[2]**o+(0.534409542153)*x[2]
        arg[3,4]=(0.305683606235)*x[0]**o+(0.340964210305)*x[0]+(0.757807973246)*x[1]**o+(0.11459551493)*x[1]+(-0.81395047506)*x[2]**o+(0.636392281491)*x[2]
        ref[0,0]=(-0.0666767976714)/(o+1.)+(-0.0556900707042)+(0.264691789613)*0.5**o
        ref[0,1]=(-0.669677383843)/(o+1.)+(-0.855981919217)+(-0.0526559162494)*0.5**o
        ref[0,2]=(-0.561217109869)/(o+1.)+(-0.334381892061)+(0.786349112412)*0.5**o
        ref[0,3]=(-0.531292978697)/(o+1.)+(-0.0106177250896)+(0.3336751197)*0.5**o
        ref[0,4]=(-1.55191243559)/(o+1.)+(0.167319505396)+(0.875206676786)*0.5**o
        ref[1,0]=(0.757311284261)/(o+1.)+(0.115361780761)+(-0.573216715158)*0.5**o
        ref[1,1]=(1.1971078852)/(o+1.)+(-0.80740355728)+(-0.0919326514467)*0.5**o
        ref[1,2]=(0.0535121674292)/(o+1.)+(-0.306876080263)+(0.855490362389)*0.5**o
        ref[1,3]=(0.0476713144526)/(o+1.)+(-0.228670641253)+(-0.90802343526)*0.5**o
        ref[1,4]=(-0.101851514548)/(o+1.)+(-0.523226075704)+(-0.33139358787)*0.5**o
        ref[2,0]=(1.21911210232)/(o+1.)+(-0.109166927241)+(-0.202818692518)*0.5**o
        ref[2,1]=(0.674260943716)/(o+1.)+(0.206519736538)+(0.0461326006936)*0.5**o
        ref[2,2]=(1.05596494137)/(o+1.)+(0.280598882924)+(-0.539659153805)*0.5**o
        ref[2,3]=(1.50709558293)/(o+1.)+(0.167180028119)+(0.861717128203)*0.5**o
        ref[2,4]=(1.47448966734)/(o+1.)+(0.204020966982)+(0.778575699703)*0.5**o
        ref[3,0]=(-0.25479808546)/(o+1.)+(0.483290068928)+(0.752875513075)*0.5**o
        ref[3,1]=(-0.0892042222872)/(o+1.)+(-0.156111214487)+(0.0563875061408)*0.5**o
        ref[3,2]=(0.0278185429781)/(o+1.)+(-1.10655329525)+(0.100078550382)*0.5**o
        ref[3,3]=(1.17567141532)/(o+1.)+(-0.00918091278228)+(0.536414145801)*0.5**o
        ref[3,4]=(-0.0561425018138)/(o+1.)+(0.545976003363)+(0.305683606235)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.448877659671)*x[0]**o+(0.130599173692)*x[0]+(-0.264185555629)*x[1]**o+(-0.722726040492)*x[1]
        arg[0,0,1]=(-0.543806308372)*x[0]**o+(-0.685242399166)*x[0]+(0.325336801501)*x[1]**o+(0.572960380228)*x[1]
        arg[0,1,0]=(-0.893080563297)*x[0]**o+(-0.251553981066)*x[0]+(0.256078388483)*x[1]**o+(-0.926701685259)*x[1]
        arg[0,1,1]=(0.346088914447)*x[0]**o+(0.166180418696)*x[0]+(0.353024178635)*x[1]**o+(0.731811935721)*x[1]
        arg[1,0,0]=(0.553083243483)*x[0]**o+(0.976153803522)*x[0]+(-0.260777317634)*x[1]**o+(-0.641936915628)*x[1]
        arg[1,0,1]=(-0.561076901068)*x[0]**o+(0.837832618302)*x[0]+(-0.638970151828)*x[1]**o+(-0.495169195543)*x[1]
        arg[1,1,0]=(-0.292805925132)*x[0]**o+(0.670472035586)*x[0]+(-0.752026521307)*x[1]**o+(-0.473087472861)*x[1]
        arg[1,1,1]=(-0.432857805275)*x[0]**o+(0.495387797249)*x[0]+(0.0579881469257)*x[1]**o+(0.831644440274)*x[1]
        arg[2,0,0]=(0.413433231384)*x[0]**o+(0.919831571878)*x[0]+(-0.373581626416)*x[1]**o+(0.16423318333)*x[1]
        arg[2,0,1]=(0.857566619293)*x[0]**o+(0.340454525569)*x[0]+(0.139298007219)*x[1]**o+(0.220542978265)*x[1]
        arg[2,1,0]=(-0.914184896115)*x[0]**o+(-0.0279186159835)*x[0]+(0.893417478351)*x[1]**o+(-0.108178714604)*x[1]
        arg[2,1,1]=(-0.749629612712)*x[0]**o+(-0.649957092299)*x[0]+(0.181944194719)*x[1]**o+(-0.988782927329)*x[1]
        arg[3,0,0]=(0.292013067678)*x[0]**o+(0.491680364317)*x[0]+(0.951144487798)*x[1]**o+(0.980949630108)*x[1]
        arg[3,0,1]=(0.89779245885)*x[0]**o+(-0.147546711846)*x[0]+(0.0902450198362)*x[1]**o+(-0.669984062969)*x[1]
        arg[3,1,0]=(0.62590631314)*x[0]**o+(0.329384454978)*x[0]+(-0.676917185409)*x[1]**o+(-0.0903173226074)*x[1]
        arg[3,1,1]=(-0.668268661271)*x[0]**o+(-0.847167879584)*x[0]+(0.329431843319)*x[1]**o+(-0.989253343115)*x[1]
        arg[4,0,0]=(-0.675111374291)*x[0]**o+(0.738030423939)*x[0]+(-0.413299452027)*x[1]**o+(0.36859580301)*x[1]
        arg[4,0,1]=(-0.207625666422)*x[0]**o+(-0.455639256266)*x[0]+(0.554118743878)*x[1]**o+(0.14044849621)*x[1]
        arg[4,1,0]=(-0.313337039275)*x[0]**o+(-0.451797308521)*x[0]+(0.201296275104)*x[1]**o+(-0.779869953962)*x[1]
        arg[4,1,1]=(-0.0308697706768)*x[0]**o+(0.214664850788)*x[0]+(-0.824157101269)*x[1]**o+(0.156012836486)*x[1]
        arg[5,0,0]=(0.277885566309)*x[0]**o+(-0.387820938776)*x[0]+(0.820602896312)*x[1]**o+(0.162715513355)*x[1]
        arg[5,0,1]=(-0.216951421349)*x[0]**o+(0.906019329549)*x[0]+(-0.330132336808)*x[1]**o+(0.416109194357)*x[1]
        arg[5,1,0]=(0.923496627826)*x[0]**o+(-0.177644464371)*x[0]+(0.482278717995)*x[1]**o+(-0.362003106154)*x[1]
        arg[5,1,1]=(0.9109111591)*x[0]**o+(-0.0919782899769)*x[0]+(0.103950387446)*x[1]**o+(0.447452648868)*x[1]
        ref[0,0,0]=(-0.264185555629)/(o+1.)+(-0.2960634334)+(0.448877659671)*0.5**o
        ref[0,0,1]=(0.325336801501)/(o+1.)+(-0.0561410094689)+(-0.543806308372)*0.5**o
        ref[0,1,0]=(0.256078388483)/(o+1.)+(-0.589127833162)+(-0.893080563297)*0.5**o
        ref[0,1,1]=(0.353024178635)/(o+1.)+(0.448996177208)+(0.346088914447)*0.5**o
        ref[1,0,0]=(-0.260777317634)/(o+1.)+(0.167108443947)+(0.553083243483)*0.5**o
        ref[1,0,1]=(-0.638970151828)/(o+1.)+(0.17133171138)+(-0.561076901068)*0.5**o
        ref[1,1,0]=(-0.752026521307)/(o+1.)+(0.0986922813626)+(-0.292805925132)*0.5**o
        ref[1,1,1]=(0.0579881469257)/(o+1.)+(0.663516118762)+(-0.432857805275)*0.5**o
        ref[2,0,0]=(-0.373581626416)/(o+1.)+(0.542032377604)+(0.413433231384)*0.5**o
        ref[2,0,1]=(0.139298007219)/(o+1.)+(0.280498751917)+(0.857566619293)*0.5**o
        ref[2,1,0]=(0.893417478351)/(o+1.)+(-0.0680486652939)+(-0.914184896115)*0.5**o
        ref[2,1,1]=(0.181944194719)/(o+1.)+(-0.819370009814)+(-0.749629612712)*0.5**o
        ref[3,0,0]=(0.951144487798)/(o+1.)+(0.736314997212)+(0.292013067678)*0.5**o
        ref[3,0,1]=(0.0902450198362)/(o+1.)+(-0.408765387408)+(0.89779245885)*0.5**o
        ref[3,1,0]=(-0.676917185409)/(o+1.)+(0.119533566185)+(0.62590631314)*0.5**o
        ref[3,1,1]=(0.329431843319)/(o+1.)+(-0.91821061135)+(-0.668268661271)*0.5**o
        ref[4,0,0]=(-0.413299452027)/(o+1.)+(0.553313113475)+(-0.675111374291)*0.5**o
        ref[4,0,1]=(0.554118743878)/(o+1.)+(-0.157595380028)+(-0.207625666422)*0.5**o
        ref[4,1,0]=(0.201296275104)/(o+1.)+(-0.615833631241)+(-0.313337039275)*0.5**o
        ref[4,1,1]=(-0.824157101269)/(o+1.)+(0.185338843637)+(-0.0308697706768)*0.5**o
        ref[5,0,0]=(0.820602896312)/(o+1.)+(-0.11255271271)+(0.277885566309)*0.5**o
        ref[5,0,1]=(-0.330132336808)/(o+1.)+(0.661064261953)+(-0.216951421349)*0.5**o
        ref[5,1,0]=(0.482278717995)/(o+1.)+(-0.269823785262)+(0.923496627826)*0.5**o
        ref[5,1,1]=(0.103950387446)/(o+1.)+(0.177737179446)+(0.9109111591)*0.5**o
      else:
        arg[0,0,0]=(0.843169691966)*x[0]**o+(0.252560741792)*x[0]+(0.0428740803857)*x[1]**o+(-0.702885246223)*x[1]+(0.175406156382)*x[2]**o+(-0.819206174419)*x[2]
        arg[0,0,1]=(-0.780678250187)*x[0]**o+(-0.628781977796)*x[0]+(0.345241549754)*x[1]**o+(-0.686783366275)*x[1]+(0.298501249797)*x[2]**o+(-0.592222734778)*x[2]
        arg[0,1,0]=(0.680331984783)*x[0]**o+(0.20945731105)*x[0]+(0.582220641498)*x[1]**o+(0.159144772508)*x[1]+(0.197544685245)*x[2]**o+(-0.542847499692)*x[2]
        arg[0,1,1]=(0.561649107089)*x[0]**o+(0.409538404439)*x[0]+(0.687461001448)*x[1]**o+(0.327986995905)*x[1]+(0.701007758697)*x[2]**o+(0.95919995454)*x[2]
        arg[1,0,0]=(0.0620287124894)*x[0]**o+(-0.946536581396)*x[0]+(-0.813637608115)*x[1]**o+(0.0808967031579)*x[1]+(0.545408819059)*x[2]**o+(0.250558134093)*x[2]
        arg[1,0,1]=(-0.0778300410758)*x[0]**o+(-0.852201314682)*x[0]+(0.712528858815)*x[1]**o+(-0.69009644973)*x[1]+(-0.0836140529567)*x[2]**o+(0.687579338803)*x[2]
        arg[1,1,0]=(-0.37648103863)*x[0]**o+(-0.264461846666)*x[0]+(-0.45921598716)*x[1]**o+(-0.516551658367)*x[1]+(0.512446750786)*x[2]**o+(0.850266522322)*x[2]
        arg[1,1,1]=(-0.951741356481)*x[0]**o+(-0.989023540683)*x[0]+(0.338787640816)*x[1]**o+(0.764043925298)*x[1]+(0.524009019713)*x[2]**o+(-0.912507794534)*x[2]
        arg[2,0,0]=(0.613468181683)*x[0]**o+(-0.357325179573)*x[0]+(0.153221256731)*x[1]**o+(-0.748686771579)*x[1]+(0.364085012368)*x[2]**o+(-0.134535056195)*x[2]
        arg[2,0,1]=(-0.728948376511)*x[0]**o+(-0.0657162843876)*x[0]+(-0.336013775703)*x[1]**o+(0.299044747397)*x[1]+(-0.712877779871)*x[2]**o+(0.229974926734)*x[2]
        arg[2,1,0]=(-0.213428159952)*x[0]**o+(0.680250111549)*x[0]+(-0.986942572737)*x[1]**o+(0.549006049647)*x[1]+(0.572921637402)*x[2]**o+(-0.241851127933)*x[2]
        arg[2,1,1]=(0.0367155018841)*x[0]**o+(-0.768111161527)*x[0]+(-0.748085216726)*x[1]**o+(-0.752536272228)*x[1]+(0.555945063877)*x[2]**o+(0.012687074005)*x[2]
        arg[3,0,0]=(-0.945391902075)*x[0]**o+(-0.581330133808)*x[0]+(-0.253390460088)*x[1]**o+(-0.885589795648)*x[1]+(-0.555900062171)*x[2]**o+(-0.241006607054)*x[2]
        arg[3,0,1]=(0.924110231188)*x[0]**o+(0.428520753545)*x[0]+(0.187773835144)*x[1]**o+(-0.53787305121)*x[1]+(0.0970284978901)*x[2]**o+(0.0580212350951)*x[2]
        arg[3,1,0]=(0.426389283701)*x[0]**o+(-0.0976148877161)*x[0]+(-0.364828466078)*x[1]**o+(-0.536398529197)*x[1]+(0.648605043498)*x[2]**o+(-0.234629706559)*x[2]
        arg[3,1,1]=(-0.736098543054)*x[0]**o+(0.985014182905)*x[0]+(-0.563513361433)*x[1]**o+(0.589434310221)*x[1]+(-0.360296462891)*x[2]**o+(-0.586591636182)*x[2]
        arg[4,0,0]=(0.953916732391)*x[0]**o+(-0.835620914249)*x[0]+(0.562803727423)*x[1]**o+(-0.904263689551)*x[1]+(-0.349815820142)*x[2]**o+(-0.156410988081)*x[2]
        arg[4,0,1]=(-0.236710267891)*x[0]**o+(0.958992060128)*x[0]+(0.711565908363)*x[1]**o+(-0.639771085045)*x[1]+(-0.919042661652)*x[2]**o+(-0.198907436333)*x[2]
        arg[4,1,0]=(0.275058935551)*x[0]**o+(0.563461500878)*x[0]+(0.438833998471)*x[1]**o+(0.686375712335)*x[1]+(-0.930322605963)*x[2]**o+(0.710777122765)*x[2]
        arg[4,1,1]=(-0.507845102241)*x[0]**o+(-0.390574086909)*x[0]+(-0.170020669232)*x[1]**o+(-0.107455242464)*x[1]+(-0.97056652533)*x[2]**o+(-0.720180192255)*x[2]
        arg[5,0,0]=(-0.922597172214)*x[0]**o+(0.229197305733)*x[0]+(-0.943214499543)*x[1]**o+(0.269256599046)*x[1]+(-0.11540537404)*x[2]**o+(0.956940487934)*x[2]
        arg[5,0,1]=(-0.0350868975578)*x[0]**o+(-0.547620412319)*x[0]+(-0.818344866429)*x[1]**o+(-0.0632082353078)*x[1]+(-0.490536679501)*x[2]**o+(-0.79641747407)*x[2]
        arg[5,1,0]=(0.877279477171)*x[0]**o+(-0.408752853807)*x[0]+(0.143432191997)*x[1]**o+(0.230679627667)*x[1]+(0.0346230515425)*x[2]**o+(0.357377660033)*x[2]
        arg[5,1,1]=(0.294375780312)*x[0]**o+(-0.688447065878)*x[0]+(0.213334642918)*x[1]**o+(0.664021137697)*x[1]+(-0.532996065961)*x[2]**o+(0.794315136407)*x[2]
        ref[0,0,0]=(0.218280236768)/(o+1.)+(-0.634765339425)+(0.843169691966)*0.5**o
        ref[0,0,1]=(0.64374279955)/(o+1.)+(-0.953894039424)+(-0.780678250187)*0.5**o
        ref[0,1,0]=(0.779765326742)/(o+1.)+(-0.0871227080671)+(0.680331984783)*0.5**o
        ref[0,1,1]=(1.38846876015)/(o+1.)+(0.848362677442)+(0.561649107089)*0.5**o
        ref[1,0,0]=(-0.268228789055)/(o+1.)+(-0.307540872072)+(0.0620287124894)*0.5**o
        ref[1,0,1]=(0.628914805859)/(o+1.)+(-0.427359212804)+(-0.0778300410758)*0.5**o
        ref[1,1,0]=(0.0532307636262)/(o+1.)+(0.0346265086445)+(-0.37648103863)*0.5**o
        ref[1,1,1]=(0.862796660529)/(o+1.)+(-0.56874370496)+(-0.951741356481)*0.5**o
        ref[2,0,0]=(0.5173062691)/(o+1.)+(-0.620273503673)+(0.613468181683)*0.5**o
        ref[2,0,1]=(-1.04889155557)/(o+1.)+(0.231651694871)+(-0.728948376511)*0.5**o
        ref[2,1,0]=(-0.414020935335)/(o+1.)+(0.493702516631)+(-0.213428159952)*0.5**o
        ref[2,1,1]=(-0.192140152849)/(o+1.)+(-0.753980179875)+(0.0367155018841)*0.5**o
        ref[3,0,0]=(-0.809290522259)/(o+1.)+(-0.853963268255)+(-0.945391902075)*0.5**o
        ref[3,0,1]=(0.284802333034)/(o+1.)+(-0.0256655312851)+(0.924110231188)*0.5**o
        ref[3,1,0]=(0.28377657742)/(o+1.)+(-0.434321561736)+(0.426389283701)*0.5**o
        ref[3,1,1]=(-0.923809824324)/(o+1.)+(0.493928428472)+(-0.736098543054)*0.5**o
        ref[4,0,0]=(0.212987907281)/(o+1.)+(-0.948147795941)+(0.953916732391)*0.5**o
        ref[4,0,1]=(-0.207476753289)/(o+1.)+(0.0601567693744)+(-0.236710267891)*0.5**o
        ref[4,1,0]=(-0.491488607492)/(o+1.)+(0.980307167989)+(0.275058935551)*0.5**o
        ref[4,1,1]=(-1.14058719456)/(o+1.)+(-0.609104760814)+(-0.507845102241)*0.5**o
        ref[5,0,0]=(-1.05861987358)/(o+1.)+(0.727697196357)+(-0.922597172214)*0.5**o
        ref[5,0,1]=(-1.30888154593)/(o+1.)+(-0.703623060848)+(-0.0350868975578)*0.5**o
        ref[5,1,0]=(0.17805524354)/(o+1.)+(0.0896522169463)+(0.877279477171)*0.5**o
        ref[5,1,1]=(-0.319661423044)/(o+1.)+(0.384944604113)+(0.294375780312)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.182956096072)*x[0]**o+(0.857065806542)*x[0]+(-0.0352052809673)*x[1]**o+(0.950661773214)*x[1]
        arg[0,0,0,1]=(-0.81342189211)*x[0]**o+(0.36137453013)*x[0]+(0.0610373691872)*x[1]**o+(-0.974859653337)*x[1]
        arg[0,0,1,0]=(-0.149103531408)*x[0]**o+(0.853517194849)*x[0]+(0.623943596759)*x[1]**o+(0.441477983782)*x[1]
        arg[0,0,1,1]=(0.74912087771)*x[0]**o+(0.935896347068)*x[0]+(0.136600247183)*x[1]**o+(0.915043476739)*x[1]
        arg[0,0,2,0]=(-0.289042694587)*x[0]**o+(-0.0324580134489)*x[0]+(0.966454801911)*x[1]**o+(0.158747002668)*x[1]
        arg[0,0,2,1]=(0.0809737838902)*x[0]**o+(0.483125829124)*x[0]+(-0.156074498516)*x[1]**o+(-0.387215102608)*x[1]
        arg[0,1,0,0]=(0.638822193145)*x[0]**o+(-0.211942445721)*x[0]+(-0.016865109839)*x[1]**o+(-0.596486509057)*x[1]
        arg[0,1,0,1]=(0.611604473356)*x[0]**o+(-0.330484367791)*x[0]+(-0.897448372147)*x[1]**o+(-0.0188556077191)*x[1]
        arg[0,1,1,0]=(-0.745407006471)*x[0]**o+(0.547985829362)*x[0]+(-0.229098133178)*x[1]**o+(0.339357994536)*x[1]
        arg[0,1,1,1]=(0.756528549281)*x[0]**o+(0.0968167016962)*x[0]+(0.481206893907)*x[1]**o+(-0.320299781291)*x[1]
        arg[0,1,2,0]=(0.335544368565)*x[0]**o+(0.143490409353)*x[0]+(0.367303506039)*x[1]**o+(-0.459001464112)*x[1]
        arg[0,1,2,1]=(0.669933676654)*x[0]**o+(0.0679909847194)*x[0]+(0.182483888026)*x[1]**o+(-0.382138842505)*x[1]
        arg[0,2,0,0]=(-0.929163949925)*x[0]**o+(0.670533855076)*x[0]+(0.587675819052)*x[1]**o+(0.221939557678)*x[1]
        arg[0,2,0,1]=(0.639896773528)*x[0]**o+(-0.166901547286)*x[0]+(-0.83934259298)*x[1]**o+(-0.522536841927)*x[1]
        arg[0,2,1,0]=(-0.399556357179)*x[0]**o+(0.860582177909)*x[0]+(0.675592435624)*x[1]**o+(0.77480796903)*x[1]
        arg[0,2,1,1]=(-0.281969675471)*x[0]**o+(-0.250831251067)*x[0]+(0.14396936899)*x[1]**o+(-0.65775670156)*x[1]
        arg[0,2,2,0]=(0.339204154215)*x[0]**o+(0.125319656279)*x[0]+(0.417557775604)*x[1]**o+(-0.985431777777)*x[1]
        arg[0,2,2,1]=(-0.655500794874)*x[0]**o+(-0.419889131296)*x[0]+(-0.0763188160723)*x[1]**o+(0.995600932757)*x[1]
        arg[0,3,0,0]=(-0.483746115201)*x[0]**o+(-0.256632816498)*x[0]+(-0.0533973517656)*x[1]**o+(-0.982541041474)*x[1]
        arg[0,3,0,1]=(0.446810992304)*x[0]**o+(-0.823523374302)*x[0]+(-0.607238533539)*x[1]**o+(-0.97562384609)*x[1]
        arg[0,3,1,0]=(-0.0235907627093)*x[0]**o+(-0.0720587626585)*x[0]+(0.930589323947)*x[1]**o+(0.327916365569)*x[1]
        arg[0,3,1,1]=(0.486342062258)*x[0]**o+(0.430279082913)*x[0]+(0.237445053125)*x[1]**o+(0.577650421809)*x[1]
        arg[0,3,2,0]=(-0.581025309937)*x[0]**o+(0.97026755344)*x[0]+(0.180505264557)*x[1]**o+(0.289585805796)*x[1]
        arg[0,3,2,1]=(-0.523356041281)*x[0]**o+(-0.502230606793)*x[0]+(0.915270822508)*x[1]**o+(0.622076448072)*x[1]
        arg[0,4,0,0]=(-0.584782530121)*x[0]**o+(-0.433761163729)*x[0]+(-0.449731522195)*x[1]**o+(0.392234342377)*x[1]
        arg[0,4,0,1]=(-0.0724106581167)*x[0]**o+(0.76896002907)*x[0]+(0.0374933582543)*x[1]**o+(-0.417790382045)*x[1]
        arg[0,4,1,0]=(0.0460360663465)*x[0]**o+(-0.0343116732267)*x[0]+(0.815864788848)*x[1]**o+(-0.922236546574)*x[1]
        arg[0,4,1,1]=(-0.978584425712)*x[0]**o+(0.249337867126)*x[0]+(-0.822961904166)*x[1]**o+(0.627473480274)*x[1]
        arg[0,4,2,0]=(-0.267227111945)*x[0]**o+(0.791320996051)*x[0]+(0.721009228711)*x[1]**o+(0.00199546728243)*x[1]
        arg[0,4,2,1]=(-0.141372777283)*x[0]**o+(-0.996912487736)*x[0]+(0.94281427253)*x[1]**o+(-0.780685413978)*x[1]
        arg[1,0,0,0]=(-0.721911758033)*x[0]**o+(0.0119756829508)*x[0]+(0.583083015744)*x[1]**o+(0.80826073507)*x[1]
        arg[1,0,0,1]=(0.929636966638)*x[0]**o+(-0.35107696413)*x[0]+(-0.656523378462)*x[1]**o+(0.610681274197)*x[1]
        arg[1,0,1,0]=(0.594847872723)*x[0]**o+(-0.15013091781)*x[0]+(-0.562311637909)*x[1]**o+(0.871283992407)*x[1]
        arg[1,0,1,1]=(0.529274819592)*x[0]**o+(0.998035152453)*x[0]+(-0.265309908189)*x[1]**o+(-0.703391843389)*x[1]
        arg[1,0,2,0]=(0.2321959905)*x[0]**o+(-0.781602927232)*x[0]+(-0.0509405210126)*x[1]**o+(0.839298348419)*x[1]
        arg[1,0,2,1]=(0.0811824999287)*x[0]**o+(0.916373714851)*x[0]+(0.817375877796)*x[1]**o+(-0.926274023089)*x[1]
        arg[1,1,0,0]=(0.418525096774)*x[0]**o+(0.064653235493)*x[0]+(0.15912570069)*x[1]**o+(0.47570489259)*x[1]
        arg[1,1,0,1]=(0.784138248952)*x[0]**o+(-0.715130508682)*x[0]+(-0.299421222048)*x[1]**o+(-0.816706522916)*x[1]
        arg[1,1,1,0]=(-0.911449050332)*x[0]**o+(-0.472182975049)*x[0]+(0.935180253062)*x[1]**o+(0.527208406385)*x[1]
        arg[1,1,1,1]=(-0.378301315644)*x[0]**o+(0.430065353544)*x[0]+(0.827581865806)*x[1]**o+(-0.355263817375)*x[1]
        arg[1,1,2,0]=(-0.882543004053)*x[0]**o+(0.503794963637)*x[0]+(-0.580086073626)*x[1]**o+(0.545676378283)*x[1]
        arg[1,1,2,1]=(-0.116197074531)*x[0]**o+(-0.0629871602139)*x[0]+(0.320343551712)*x[1]**o+(0.0552550885696)*x[1]
        arg[1,2,0,0]=(-0.870347384636)*x[0]**o+(-0.611439594673)*x[0]+(0.707766015783)*x[1]**o+(0.0653036268139)*x[1]
        arg[1,2,0,1]=(-0.219872666014)*x[0]**o+(-0.435867669656)*x[0]+(0.825489444837)*x[1]**o+(-0.737253859382)*x[1]
        arg[1,2,1,0]=(-0.963149287055)*x[0]**o+(0.964306345088)*x[0]+(0.502736178959)*x[1]**o+(0.998017404878)*x[1]
        arg[1,2,1,1]=(-0.586350249194)*x[0]**o+(-0.437455605038)*x[0]+(-0.182719660538)*x[1]**o+(-0.50797820305)*x[1]
        arg[1,2,2,0]=(-0.423818475166)*x[0]**o+(0.290902418459)*x[0]+(-0.304129806418)*x[1]**o+(-0.67571878331)*x[1]
        arg[1,2,2,1]=(0.901384945886)*x[0]**o+(0.0526906910209)*x[0]+(0.438755721196)*x[1]**o+(-0.774352941544)*x[1]
        arg[1,3,0,0]=(-0.797968602257)*x[0]**o+(0.856637297517)*x[0]+(0.427243495756)*x[1]**o+(0.212942988457)*x[1]
        arg[1,3,0,1]=(-0.696126104381)*x[0]**o+(0.616445534254)*x[0]+(-0.943037028688)*x[1]**o+(0.807265721375)*x[1]
        arg[1,3,1,0]=(-0.69028368426)*x[0]**o+(-0.652017674488)*x[0]+(-0.772965740874)*x[1]**o+(-0.726830598291)*x[1]
        arg[1,3,1,1]=(-0.534347181294)*x[0]**o+(-0.861020241321)*x[0]+(-0.250549859157)*x[1]**o+(-0.638289396201)*x[1]
        arg[1,3,2,0]=(-0.827981962115)*x[0]**o+(-0.427544131008)*x[0]+(-0.516514308221)*x[1]**o+(-0.101906062547)*x[1]
        arg[1,3,2,1]=(0.864701869194)*x[0]**o+(0.876327129524)*x[0]+(0.806201346091)*x[1]**o+(0.845259832272)*x[1]
        arg[1,4,0,0]=(-0.589044559231)*x[0]**o+(0.697104039773)*x[0]+(-0.288389617381)*x[1]**o+(0.422618969092)*x[1]
        arg[1,4,0,1]=(0.814971550967)*x[0]**o+(-0.665151459593)*x[0]+(-0.696145763418)*x[1]**o+(0.906949165457)*x[1]
        arg[1,4,1,0]=(0.760190450826)*x[0]**o+(-0.883964913626)*x[0]+(-0.507159320061)*x[1]**o+(-0.15707026975)*x[1]
        arg[1,4,1,1]=(-0.0416603697257)*x[0]**o+(0.560561779164)*x[0]+(0.0153574982122)*x[1]**o+(0.75228872946)*x[1]
        arg[1,4,2,0]=(0.105929120872)*x[0]**o+(0.130738968719)*x[0]+(-0.0991159921294)*x[1]**o+(0.972627650896)*x[1]
        arg[1,4,2,1]=(-0.503911541225)*x[0]**o+(-0.00357063650385)*x[0]+(0.377083955928)*x[1]**o+(0.343010153434)*x[1]
        arg[2,0,0,0]=(0.953733725947)*x[0]**o+(-0.988629842686)*x[0]+(0.196766730386)*x[1]**o+(0.0427993038279)*x[1]
        arg[2,0,0,1]=(0.491469820597)*x[0]**o+(0.708248604732)*x[0]+(-0.600347310418)*x[1]**o+(-0.0380188651191)*x[1]
        arg[2,0,1,0]=(0.615118477478)*x[0]**o+(0.832030863133)*x[0]+(0.783037278839)*x[1]**o+(-0.331824170309)*x[1]
        arg[2,0,1,1]=(-0.806125471265)*x[0]**o+(0.763233458333)*x[0]+(-0.381389710969)*x[1]**o+(0.652974431609)*x[1]
        arg[2,0,2,0]=(-0.831727679795)*x[0]**o+(0.239296785135)*x[0]+(-0.13737609858)*x[1]**o+(0.583210422723)*x[1]
        arg[2,0,2,1]=(-0.805843206194)*x[0]**o+(-0.523576669019)*x[0]+(0.159758297984)*x[1]**o+(-0.298308874886)*x[1]
        arg[2,1,0,0]=(0.589135530105)*x[0]**o+(0.272145879514)*x[0]+(0.325741074892)*x[1]**o+(-0.579782654647)*x[1]
        arg[2,1,0,1]=(0.106049062774)*x[0]**o+(0.790280056075)*x[0]+(0.0171600098883)*x[1]**o+(-0.419540445094)*x[1]
        arg[2,1,1,0]=(-0.181511670026)*x[0]**o+(-0.814531462508)*x[0]+(-0.576165158737)*x[1]**o+(0.679678310717)*x[1]
        arg[2,1,1,1]=(-0.709695453728)*x[0]**o+(-0.489487606257)*x[0]+(-0.424093531654)*x[1]**o+(0.535427169862)*x[1]
        arg[2,1,2,0]=(0.60069280023)*x[0]**o+(-0.909882731214)*x[0]+(0.640141428017)*x[1]**o+(-0.697396064954)*x[1]
        arg[2,1,2,1]=(0.896964594562)*x[0]**o+(0.976387328697)*x[0]+(0.144061490066)*x[1]**o+(0.378042869796)*x[1]
        arg[2,2,0,0]=(0.447633246258)*x[0]**o+(-0.531416367062)*x[0]+(0.199304002742)*x[1]**o+(0.85195706213)*x[1]
        arg[2,2,0,1]=(-0.683614706554)*x[0]**o+(0.933255565209)*x[0]+(-0.328047392082)*x[1]**o+(-0.0865769137182)*x[1]
        arg[2,2,1,0]=(0.22474720984)*x[0]**o+(0.0910500178874)*x[0]+(0.449220001922)*x[1]**o+(-0.0406214557829)*x[1]
        arg[2,2,1,1]=(-0.266419887395)*x[0]**o+(-0.48883285131)*x[0]+(-0.794933934805)*x[1]**o+(0.182439649969)*x[1]
        arg[2,2,2,0]=(0.949852186215)*x[0]**o+(-0.116121704259)*x[0]+(-0.977131144423)*x[1]**o+(0.698402625747)*x[1]
        arg[2,2,2,1]=(-0.0446180768654)*x[0]**o+(-0.742263614438)*x[0]+(-0.898675367333)*x[1]**o+(-0.260891146746)*x[1]
        arg[2,3,0,0]=(0.819774310005)*x[0]**o+(-0.0677371406664)*x[0]+(0.276747165681)*x[1]**o+(-0.42828783)*x[1]
        arg[2,3,0,1]=(-0.391001671385)*x[0]**o+(0.0558814994728)*x[0]+(-0.624786978129)*x[1]**o+(0.178742280566)*x[1]
        arg[2,3,1,0]=(0.904597103646)*x[0]**o+(-0.825472271707)*x[0]+(0.112221476422)*x[1]**o+(0.269349199527)*x[1]
        arg[2,3,1,1]=(-0.190955697638)*x[0]**o+(0.441698994606)*x[0]+(-0.26229840557)*x[1]**o+(0.148974540187)*x[1]
        arg[2,3,2,0]=(-0.873098622934)*x[0]**o+(-0.624915584376)*x[0]+(-0.151646578512)*x[1]**o+(-0.914878228283)*x[1]
        arg[2,3,2,1]=(-0.712569947477)*x[0]**o+(-0.72679361801)*x[0]+(-0.0946869166491)*x[1]**o+(-0.580512407933)*x[1]
        arg[2,4,0,0]=(-0.655752680768)*x[0]**o+(-0.409544975855)*x[0]+(0.865325438086)*x[1]**o+(-0.227397026937)*x[1]
        arg[2,4,0,1]=(0.670566711209)*x[0]**o+(-0.794343533883)*x[0]+(-0.127116396219)*x[1]**o+(-0.422419866647)*x[1]
        arg[2,4,1,0]=(0.46903441444)*x[0]**o+(0.0689129759715)*x[0]+(-0.549325647826)*x[1]**o+(0.893700020399)*x[1]
        arg[2,4,1,1]=(-0.0099246750652)*x[0]**o+(0.446742615509)*x[0]+(0.900251202513)*x[1]**o+(0.221599931055)*x[1]
        arg[2,4,2,0]=(0.218854380066)*x[0]**o+(0.877849336909)*x[0]+(-0.338211083089)*x[1]**o+(0.746950973063)*x[1]
        arg[2,4,2,1]=(-0.890762146523)*x[0]**o+(0.439653904008)*x[0]+(0.0802890292215)*x[1]**o+(-0.305233739845)*x[1]
        arg[3,0,0,0]=(0.284622209957)*x[0]**o+(0.983284338924)*x[0]+(-0.732588048234)*x[1]**o+(0.654066388109)*x[1]
        arg[3,0,0,1]=(0.379628076244)*x[0]**o+(-0.75362875047)*x[0]+(0.288422119968)*x[1]**o+(0.139197401407)*x[1]
        arg[3,0,1,0]=(0.0594472092467)*x[0]**o+(-0.391581973307)*x[0]+(-0.898996153568)*x[1]**o+(-2.75165483707e-05)*x[1]
        arg[3,0,1,1]=(-0.241252797172)*x[0]**o+(-0.87384982932)*x[0]+(0.811628957112)*x[1]**o+(0.0170539509618)*x[1]
        arg[3,0,2,0]=(0.255617459837)*x[0]**o+(-0.744071491889)*x[0]+(0.745552769828)*x[1]**o+(0.128612266243)*x[1]
        arg[3,0,2,1]=(0.0725082197254)*x[0]**o+(0.779602587374)*x[0]+(-0.751879287668)*x[1]**o+(0.838360641095)*x[1]
        arg[3,1,0,0]=(-0.444520674459)*x[0]**o+(-0.811442143542)*x[0]+(-0.485010478999)*x[1]**o+(0.727600464745)*x[1]
        arg[3,1,0,1]=(0.260061065042)*x[0]**o+(0.932320346965)*x[0]+(0.556377882844)*x[1]**o+(0.284348284894)*x[1]
        arg[3,1,1,0]=(0.561755002522)*x[0]**o+(-0.842983435746)*x[0]+(0.729209244377)*x[1]**o+(-0.254683846974)*x[1]
        arg[3,1,1,1]=(-0.949556399673)*x[0]**o+(-0.0804522142379)*x[0]+(0.066141483426)*x[1]**o+(0.78626145966)*x[1]
        arg[3,1,2,0]=(0.833198651881)*x[0]**o+(-0.297955946153)*x[0]+(-0.561247934789)*x[1]**o+(-0.645124432611)*x[1]
        arg[3,1,2,1]=(-0.641727656043)*x[0]**o+(0.32981201062)*x[0]+(-0.152588766563)*x[1]**o+(-0.289611460761)*x[1]
        arg[3,2,0,0]=(0.825643539598)*x[0]**o+(0.924286336436)*x[0]+(0.413089850944)*x[1]**o+(-0.640198886209)*x[1]
        arg[3,2,0,1]=(0.0855693847128)*x[0]**o+(-0.363881253576)*x[0]+(0.0397854550832)*x[1]**o+(-0.844391996875)*x[1]
        arg[3,2,1,0]=(-0.8248537795)*x[0]**o+(0.153070857979)*x[0]+(0.970009460355)*x[1]**o+(0.431000028243)*x[1]
        arg[3,2,1,1]=(-0.978024209918)*x[0]**o+(-0.527494257814)*x[0]+(-0.814691105217)*x[1]**o+(0.458334868681)*x[1]
        arg[3,2,2,0]=(0.944359157869)*x[0]**o+(-0.868016282191)*x[0]+(0.410849160429)*x[1]**o+(0.480177162607)*x[1]
        arg[3,2,2,1]=(0.109166287134)*x[0]**o+(0.940100154901)*x[0]+(0.630946061805)*x[1]**o+(0.347160020469)*x[1]
        arg[3,3,0,0]=(0.364585502975)*x[0]**o+(0.717282598047)*x[0]+(0.517805310585)*x[1]**o+(0.963452728824)*x[1]
        arg[3,3,0,1]=(0.0373453558524)*x[0]**o+(0.943887402198)*x[0]+(0.983624425071)*x[1]**o+(-0.396242794059)*x[1]
        arg[3,3,1,0]=(-0.563724031351)*x[0]**o+(0.959753988518)*x[0]+(0.118506885299)*x[1]**o+(-0.587748202342)*x[1]
        arg[3,3,1,1]=(0.460419807013)*x[0]**o+(0.237753399499)*x[0]+(0.585589833251)*x[1]**o+(0.935177937667)*x[1]
        arg[3,3,2,0]=(-0.750248086829)*x[0]**o+(-0.561629848994)*x[0]+(-0.919919077536)*x[1]**o+(0.239984179086)*x[1]
        arg[3,3,2,1]=(-0.0969897163659)*x[0]**o+(0.185433022889)*x[0]+(0.733111631837)*x[1]**o+(-0.508147081928)*x[1]
        arg[3,4,0,0]=(0.247054568876)*x[0]**o+(0.721751593201)*x[0]+(-0.684740924498)*x[1]**o+(0.28395202476)*x[1]
        arg[3,4,0,1]=(-0.0949248638736)*x[0]**o+(0.270285273316)*x[0]+(0.850836559464)*x[1]**o+(-0.771085947382)*x[1]
        arg[3,4,1,0]=(-0.965410502362)*x[0]**o+(-0.582448742001)*x[0]+(-0.90307507827)*x[1]**o+(0.914051445745)*x[1]
        arg[3,4,1,1]=(0.443340058617)*x[0]**o+(0.434462120769)*x[0]+(0.522015964301)*x[1]**o+(-0.520184389865)*x[1]
        arg[3,4,2,0]=(-0.961375562593)*x[0]**o+(-0.474630747481)*x[0]+(0.59574636341)*x[1]**o+(0.434190359663)*x[1]
        arg[3,4,2,1]=(0.174154210501)*x[0]**o+(-0.510858596799)*x[0]+(-0.0279700806065)*x[1]**o+(-0.0401612371912)*x[1]
        ref[0,0,0,0]=(-0.0352052809673)/(o+1.)+(0.903863789878)+(0.182956096072)*0.5**o
        ref[0,0,0,1]=(0.0610373691872)/(o+1.)+(-0.306742561603)+(-0.81342189211)*0.5**o
        ref[0,0,1,0]=(0.623943596759)/(o+1.)+(0.647497589315)+(-0.149103531408)*0.5**o
        ref[0,0,1,1]=(0.136600247183)/(o+1.)+(0.925469911903)+(0.74912087771)*0.5**o
        ref[0,0,2,0]=(0.966454801911)/(o+1.)+(0.0631444946098)+(-0.289042694587)*0.5**o
        ref[0,0,2,1]=(-0.156074498516)/(o+1.)+(0.047955363258)+(0.0809737838902)*0.5**o
        ref[0,1,0,0]=(-0.016865109839)/(o+1.)+(-0.404214477389)+(0.638822193145)*0.5**o
        ref[0,1,0,1]=(-0.897448372147)/(o+1.)+(-0.174669987755)+(0.611604473356)*0.5**o
        ref[0,1,1,0]=(-0.229098133178)/(o+1.)+(0.443671911949)+(-0.745407006471)*0.5**o
        ref[0,1,1,1]=(0.481206893907)/(o+1.)+(-0.111741539797)+(0.756528549281)*0.5**o
        ref[0,1,2,0]=(0.367303506039)/(o+1.)+(-0.157755527379)+(0.335544368565)*0.5**o
        ref[0,1,2,1]=(0.182483888026)/(o+1.)+(-0.157073928893)+(0.669933676654)*0.5**o
        ref[0,2,0,0]=(0.587675819052)/(o+1.)+(0.446236706377)+(-0.929163949925)*0.5**o
        ref[0,2,0,1]=(-0.83934259298)/(o+1.)+(-0.344719194607)+(0.639896773528)*0.5**o
        ref[0,2,1,0]=(0.675592435624)/(o+1.)+(0.81769507347)+(-0.399556357179)*0.5**o
        ref[0,2,1,1]=(0.14396936899)/(o+1.)+(-0.454293976313)+(-0.281969675471)*0.5**o
        ref[0,2,2,0]=(0.417557775604)/(o+1.)+(-0.430056060749)+(0.339204154215)*0.5**o
        ref[0,2,2,1]=(-0.0763188160723)/(o+1.)+(0.28785590073)+(-0.655500794874)*0.5**o
        ref[0,3,0,0]=(-0.0533973517656)/(o+1.)+(-0.619586928986)+(-0.483746115201)*0.5**o
        ref[0,3,0,1]=(-0.607238533539)/(o+1.)+(-0.899573610196)+(0.446810992304)*0.5**o
        ref[0,3,1,0]=(0.930589323947)/(o+1.)+(0.127928801455)+(-0.0235907627093)*0.5**o
        ref[0,3,1,1]=(0.237445053125)/(o+1.)+(0.503964752361)+(0.486342062258)*0.5**o
        ref[0,3,2,0]=(0.180505264557)/(o+1.)+(0.629926679618)+(-0.581025309937)*0.5**o
        ref[0,3,2,1]=(0.915270822508)/(o+1.)+(0.0599229206396)+(-0.523356041281)*0.5**o
        ref[0,4,0,0]=(-0.449731522195)/(o+1.)+(-0.0207634106763)+(-0.584782530121)*0.5**o
        ref[0,4,0,1]=(0.0374933582543)/(o+1.)+(0.175584823513)+(-0.0724106581167)*0.5**o
        ref[0,4,1,0]=(0.815864788848)/(o+1.)+(-0.4782741099)+(0.0460360663465)*0.5**o
        ref[0,4,1,1]=(-0.822961904166)/(o+1.)+(0.4384056737)+(-0.978584425712)*0.5**o
        ref[0,4,2,0]=(0.721009228711)/(o+1.)+(0.396658231667)+(-0.267227111945)*0.5**o
        ref[0,4,2,1]=(0.94281427253)/(o+1.)+(-0.888798950857)+(-0.141372777283)*0.5**o
        ref[1,0,0,0]=(0.583083015744)/(o+1.)+(0.41011820901)+(-0.721911758033)*0.5**o
        ref[1,0,0,1]=(-0.656523378462)/(o+1.)+(0.129802155034)+(0.929636966638)*0.5**o
        ref[1,0,1,0]=(-0.562311637909)/(o+1.)+(0.360576537299)+(0.594847872723)*0.5**o
        ref[1,0,1,1]=(-0.265309908189)/(o+1.)+(0.147321654532)+(0.529274819592)*0.5**o
        ref[1,0,2,0]=(-0.0509405210126)/(o+1.)+(0.0288477105933)+(0.2321959905)*0.5**o
        ref[1,0,2,1]=(0.817375877796)/(o+1.)+(-0.00495015411903)+(0.0811824999287)*0.5**o
        ref[1,1,0,0]=(0.15912570069)/(o+1.)+(0.270179064041)+(0.418525096774)*0.5**o
        ref[1,1,0,1]=(-0.299421222048)/(o+1.)+(-0.765918515799)+(0.784138248952)*0.5**o
        ref[1,1,1,0]=(0.935180253062)/(o+1.)+(0.0275127156677)+(-0.911449050332)*0.5**o
        ref[1,1,1,1]=(0.827581865806)/(o+1.)+(0.0374007680845)+(-0.378301315644)*0.5**o
        ref[1,1,2,0]=(-0.580086073626)/(o+1.)+(0.52473567096)+(-0.882543004053)*0.5**o
        ref[1,1,2,1]=(0.320343551712)/(o+1.)+(-0.00386603582212)+(-0.116197074531)*0.5**o
        ref[1,2,0,0]=(0.707766015783)/(o+1.)+(-0.273067983929)+(-0.870347384636)*0.5**o
        ref[1,2,0,1]=(0.825489444837)/(o+1.)+(-0.586560764519)+(-0.219872666014)*0.5**o
        ref[1,2,1,0]=(0.502736178959)/(o+1.)+(0.981161874983)+(-0.963149287055)*0.5**o
        ref[1,2,1,1]=(-0.182719660538)/(o+1.)+(-0.472716904044)+(-0.586350249194)*0.5**o
        ref[1,2,2,0]=(-0.304129806418)/(o+1.)+(-0.192408182425)+(-0.423818475166)*0.5**o
        ref[1,2,2,1]=(0.438755721196)/(o+1.)+(-0.360831125262)+(0.901384945886)*0.5**o
        ref[1,3,0,0]=(0.427243495756)/(o+1.)+(0.534790142987)+(-0.797968602257)*0.5**o
        ref[1,3,0,1]=(-0.943037028688)/(o+1.)+(0.711855627815)+(-0.696126104381)*0.5**o
        ref[1,3,1,0]=(-0.772965740874)/(o+1.)+(-0.689424136389)+(-0.69028368426)*0.5**o
        ref[1,3,1,1]=(-0.250549859157)/(o+1.)+(-0.749654818761)+(-0.534347181294)*0.5**o
        ref[1,3,2,0]=(-0.516514308221)/(o+1.)+(-0.264725096778)+(-0.827981962115)*0.5**o
        ref[1,3,2,1]=(0.806201346091)/(o+1.)+(0.860793480898)+(0.864701869194)*0.5**o
        ref[1,4,0,0]=(-0.288389617381)/(o+1.)+(0.559861504432)+(-0.589044559231)*0.5**o
        ref[1,4,0,1]=(-0.696145763418)/(o+1.)+(0.120898852932)+(0.814971550967)*0.5**o
        ref[1,4,1,0]=(-0.507159320061)/(o+1.)+(-0.520517591688)+(0.760190450826)*0.5**o
        ref[1,4,1,1]=(0.0153574982122)/(o+1.)+(0.656425254312)+(-0.0416603697257)*0.5**o
        ref[1,4,2,0]=(-0.0991159921294)/(o+1.)+(0.551683309807)+(0.105929120872)*0.5**o
        ref[1,4,2,1]=(0.377083955928)/(o+1.)+(0.169719758465)+(-0.503911541225)*0.5**o
        ref[2,0,0,0]=(0.196766730386)/(o+1.)+(-0.472915269429)+(0.953733725947)*0.5**o
        ref[2,0,0,1]=(-0.600347310418)/(o+1.)+(0.335114869806)+(0.491469820597)*0.5**o
        ref[2,0,1,0]=(0.783037278839)/(o+1.)+(0.250103346412)+(0.615118477478)*0.5**o
        ref[2,0,1,1]=(-0.381389710969)/(o+1.)+(0.708103944971)+(-0.806125471265)*0.5**o
        ref[2,0,2,0]=(-0.13737609858)/(o+1.)+(0.411253603929)+(-0.831727679795)*0.5**o
        ref[2,0,2,1]=(0.159758297984)/(o+1.)+(-0.410942771953)+(-0.805843206194)*0.5**o
        ref[2,1,0,0]=(0.325741074892)/(o+1.)+(-0.153818387566)+(0.589135530105)*0.5**o
        ref[2,1,0,1]=(0.0171600098883)/(o+1.)+(0.185369805491)+(0.106049062774)*0.5**o
        ref[2,1,1,0]=(-0.576165158737)/(o+1.)+(-0.0674265758955)+(-0.181511670026)*0.5**o
        ref[2,1,1,1]=(-0.424093531654)/(o+1.)+(0.0229697818029)+(-0.709695453728)*0.5**o
        ref[2,1,2,0]=(0.640141428017)/(o+1.)+(-0.803639398084)+(0.60069280023)*0.5**o
        ref[2,1,2,1]=(0.144061490066)/(o+1.)+(0.677215099246)+(0.896964594562)*0.5**o
        ref[2,2,0,0]=(0.199304002742)/(o+1.)+(0.160270347534)+(0.447633246258)*0.5**o
        ref[2,2,0,1]=(-0.328047392082)/(o+1.)+(0.423339325745)+(-0.683614706554)*0.5**o
        ref[2,2,1,0]=(0.449220001922)/(o+1.)+(0.0252142810523)+(0.22474720984)*0.5**o
        ref[2,2,1,1]=(-0.794933934805)/(o+1.)+(-0.153196600671)+(-0.266419887395)*0.5**o
        ref[2,2,2,0]=(-0.977131144423)/(o+1.)+(0.291140460744)+(0.949852186215)*0.5**o
        ref[2,2,2,1]=(-0.898675367333)/(o+1.)+(-0.501577380592)+(-0.0446180768654)*0.5**o
        ref[2,3,0,0]=(0.276747165681)/(o+1.)+(-0.248012485333)+(0.819774310005)*0.5**o
        ref[2,3,0,1]=(-0.624786978129)/(o+1.)+(0.11731189002)+(-0.391001671385)*0.5**o
        ref[2,3,1,0]=(0.112221476422)/(o+1.)+(-0.27806153609)+(0.904597103646)*0.5**o
        ref[2,3,1,1]=(-0.26229840557)/(o+1.)+(0.295336767397)+(-0.190955697638)*0.5**o
        ref[2,3,2,0]=(-0.151646578512)/(o+1.)+(-0.76989690633)+(-0.873098622934)*0.5**o
        ref[2,3,2,1]=(-0.0946869166491)/(o+1.)+(-0.653653012971)+(-0.712569947477)*0.5**o
        ref[2,4,0,0]=(0.865325438086)/(o+1.)+(-0.318471001396)+(-0.655752680768)*0.5**o
        ref[2,4,0,1]=(-0.127116396219)/(o+1.)+(-0.608381700265)+(0.670566711209)*0.5**o
        ref[2,4,1,0]=(-0.549325647826)/(o+1.)+(0.481306498185)+(0.46903441444)*0.5**o
        ref[2,4,1,1]=(0.900251202513)/(o+1.)+(0.334171273282)+(-0.0099246750652)*0.5**o
        ref[2,4,2,0]=(-0.338211083089)/(o+1.)+(0.812400154986)+(0.218854380066)*0.5**o
        ref[2,4,2,1]=(0.0802890292215)/(o+1.)+(0.0672100820814)+(-0.890762146523)*0.5**o
        ref[3,0,0,0]=(-0.732588048234)/(o+1.)+(0.818675363516)+(0.284622209957)*0.5**o
        ref[3,0,0,1]=(0.288422119968)/(o+1.)+(-0.307215674532)+(0.379628076244)*0.5**o
        ref[3,0,1,0]=(-0.898996153568)/(o+1.)+(-0.195804744928)+(0.0594472092467)*0.5**o
        ref[3,0,1,1]=(0.811628957112)/(o+1.)+(-0.428397939179)+(-0.241252797172)*0.5**o
        ref[3,0,2,0]=(0.745552769828)/(o+1.)+(-0.307729612823)+(0.255617459837)*0.5**o
        ref[3,0,2,1]=(-0.751879287668)/(o+1.)+(0.808981614234)+(0.0725082197254)*0.5**o
        ref[3,1,0,0]=(-0.485010478999)/(o+1.)+(-0.0419208393982)+(-0.444520674459)*0.5**o
        ref[3,1,0,1]=(0.556377882844)/(o+1.)+(0.60833431593)+(0.260061065042)*0.5**o
        ref[3,1,1,0]=(0.729209244377)/(o+1.)+(-0.54883364136)+(0.561755002522)*0.5**o
        ref[3,1,1,1]=(0.066141483426)/(o+1.)+(0.352904622711)+(-0.949556399673)*0.5**o
        ref[3,1,2,0]=(-0.561247934789)/(o+1.)+(-0.471540189382)+(0.833198651881)*0.5**o
        ref[3,1,2,1]=(-0.152588766563)/(o+1.)+(0.0201002749295)+(-0.641727656043)*0.5**o
        ref[3,2,0,0]=(0.413089850944)/(o+1.)+(0.142043725114)+(0.825643539598)*0.5**o
        ref[3,2,0,1]=(0.0397854550832)/(o+1.)+(-0.604136625225)+(0.0855693847128)*0.5**o
        ref[3,2,1,0]=(0.970009460355)/(o+1.)+(0.292035443111)+(-0.8248537795)*0.5**o
        ref[3,2,1,1]=(-0.814691105217)/(o+1.)+(-0.0345796945668)+(-0.978024209918)*0.5**o
        ref[3,2,2,0]=(0.410849160429)/(o+1.)+(-0.193919559792)+(0.944359157869)*0.5**o
        ref[3,2,2,1]=(0.630946061805)/(o+1.)+(0.643630087685)+(0.109166287134)*0.5**o
        ref[3,3,0,0]=(0.517805310585)/(o+1.)+(0.840367663435)+(0.364585502975)*0.5**o
        ref[3,3,0,1]=(0.983624425071)/(o+1.)+(0.273822304069)+(0.0373453558524)*0.5**o
        ref[3,3,1,0]=(0.118506885299)/(o+1.)+(0.186002893088)+(-0.563724031351)*0.5**o
        ref[3,3,1,1]=(0.585589833251)/(o+1.)+(0.586465668583)+(0.460419807013)*0.5**o
        ref[3,3,2,0]=(-0.919919077536)/(o+1.)+(-0.160822834954)+(-0.750248086829)*0.5**o
        ref[3,3,2,1]=(0.733111631837)/(o+1.)+(-0.161357029519)+(-0.0969897163659)*0.5**o
        ref[3,4,0,0]=(-0.684740924498)/(o+1.)+(0.50285180898)+(0.247054568876)*0.5**o
        ref[3,4,0,1]=(0.850836559464)/(o+1.)+(-0.250400337033)+(-0.0949248638736)*0.5**o
        ref[3,4,1,0]=(-0.90307507827)/(o+1.)+(0.165801351872)+(-0.965410502362)*0.5**o
        ref[3,4,1,1]=(0.522015964301)/(o+1.)+(-0.0428611345479)+(0.443340058617)*0.5**o
        ref[3,4,2,0]=(0.59574636341)/(o+1.)+(-0.0202201939092)+(-0.961375562593)*0.5**o
        ref[3,4,2,1]=(-0.0279700806065)/(o+1.)+(-0.275509916995)+(0.174154210501)*0.5**o
      else:
        arg[0,0,0,0]=(-0.793251694335)*x[0]**o+(0.616970271582)*x[0]+(-0.805609243189)*x[1]**o+(0.949515692474)*x[1]+(0.484069738431)*x[2]**o+(-0.0407533911672)*x[2]
        arg[0,0,0,1]=(-0.624206604764)*x[0]**o+(-0.992961148702)*x[0]+(-0.737741763879)*x[1]**o+(0.161544434398)*x[1]+(0.47292947772)*x[2]**o+(0.525538846945)*x[2]
        arg[0,0,1,0]=(0.115999561481)*x[0]**o+(-0.612694096044)*x[0]+(0.625539443386)*x[1]**o+(0.857353715882)*x[1]+(-0.0811256709155)*x[2]**o+(0.183474706394)*x[2]
        arg[0,0,1,1]=(-0.972466831671)*x[0]**o+(0.262980887655)*x[0]+(0.920924576059)*x[1]**o+(0.983517658667)*x[1]+(0.579443973505)*x[2]**o+(0.101139612823)*x[2]
        arg[0,0,2,0]=(0.708180260094)*x[0]**o+(-0.0950913766285)*x[0]+(0.361602191798)*x[1]**o+(-0.880757848201)*x[1]+(0.326272205058)*x[2]**o+(-0.997155694972)*x[2]
        arg[0,0,2,1]=(0.0282615573045)*x[0]**o+(-0.317851994728)*x[0]+(0.449330147388)*x[1]**o+(-0.156946318079)*x[1]+(0.344196894245)*x[2]**o+(-0.045082290904)*x[2]
        arg[0,1,0,0]=(0.231621820235)*x[0]**o+(0.0985804817846)*x[0]+(-0.343623333022)*x[1]**o+(0.115043475441)*x[1]+(-0.910742772023)*x[2]**o+(-0.663592068553)*x[2]
        arg[0,1,0,1]=(0.114056801155)*x[0]**o+(-0.463711315494)*x[0]+(0.380521147429)*x[1]**o+(0.294823684238)*x[1]+(0.510501672453)*x[2]**o+(-0.161234251875)*x[2]
        arg[0,1,1,0]=(-0.973115195017)*x[0]**o+(-0.0133290715609)*x[0]+(-0.642072931588)*x[1]**o+(-0.580655605735)*x[1]+(-0.191371539061)*x[2]**o+(-0.0513051439064)*x[2]
        arg[0,1,1,1]=(-0.304758579369)*x[0]**o+(-0.321904612871)*x[0]+(-0.389525877547)*x[1]**o+(-0.0602638359252)*x[1]+(0.199726168523)*x[2]**o+(-0.10313327614)*x[2]
        arg[0,1,2,0]=(0.0923844967157)*x[0]**o+(-0.369319307184)*x[0]+(-0.614761123101)*x[1]**o+(0.78817430449)*x[1]+(-0.783809995839)*x[2]**o+(-0.701522582246)*x[2]
        arg[0,1,2,1]=(0.793568921835)*x[0]**o+(0.352244892726)*x[0]+(0.185546719529)*x[1]**o+(0.145969792058)*x[1]+(0.85147546253)*x[2]**o+(-0.862069843156)*x[2]
        arg[0,2,0,0]=(0.570673248048)*x[0]**o+(0.607924942767)*x[0]+(-0.353952625913)*x[1]**o+(0.111796842318)*x[1]+(-0.765670542263)*x[2]**o+(0.589532724922)*x[2]
        arg[0,2,0,1]=(0.299061155215)*x[0]**o+(-0.542978954882)*x[0]+(0.0946151623892)*x[1]**o+(0.310266302534)*x[1]+(-0.00880229885823)*x[2]**o+(0.585508874849)*x[2]
        arg[0,2,1,0]=(0.694725244159)*x[0]**o+(0.530004132186)*x[0]+(-0.32588478629)*x[1]**o+(0.357278329148)*x[1]+(-0.93877612203)*x[2]**o+(0.34973013935)*x[2]
        arg[0,2,1,1]=(-0.83896527178)*x[0]**o+(0.487231460025)*x[0]+(0.493952010418)*x[1]**o+(0.247442509692)*x[1]+(-0.312456515075)*x[2]**o+(0.90248200924)*x[2]
        arg[0,2,2,0]=(0.2228052704)*x[0]**o+(0.901373696962)*x[0]+(0.149826948427)*x[1]**o+(0.234013159005)*x[1]+(0.717685535862)*x[2]**o+(-0.42865545543)*x[2]
        arg[0,2,2,1]=(0.63141216243)*x[0]**o+(0.448446314659)*x[0]+(0.379625538779)*x[1]**o+(-0.807448154272)*x[1]+(0.238724835223)*x[2]**o+(-0.478216911196)*x[2]
        arg[0,3,0,0]=(0.00698107166168)*x[0]**o+(-0.727207190747)*x[0]+(-0.78731647832)*x[1]**o+(0.885371753974)*x[1]+(-0.0146821479126)*x[2]**o+(-0.763310617205)*x[2]
        arg[0,3,0,1]=(0.243423742325)*x[0]**o+(0.0587322175837)*x[0]+(-0.453213790216)*x[1]**o+(0.861270481703)*x[1]+(0.867332526083)*x[2]**o+(-0.713187214416)*x[2]
        arg[0,3,1,0]=(-0.106629554393)*x[0]**o+(0.723693136654)*x[0]+(-0.0276604293602)*x[1]**o+(0.339456904686)*x[1]+(-0.102075459215)*x[2]**o+(-0.0348969356596)*x[2]
        arg[0,3,1,1]=(0.375143733111)*x[0]**o+(-0.490781234413)*x[0]+(0.168903339622)*x[1]**o+(0.594080859808)*x[1]+(-0.0802336915361)*x[2]**o+(0.472348979787)*x[2]
        arg[0,3,2,0]=(0.955685455438)*x[0]**o+(0.105142628533)*x[0]+(-0.446384800971)*x[1]**o+(-0.905998305229)*x[1]+(-0.30011637186)*x[2]**o+(0.0218195326037)*x[2]
        arg[0,3,2,1]=(0.353654445481)*x[0]**o+(-0.0477024594709)*x[0]+(-0.495075817497)*x[1]**o+(-0.425091398447)*x[1]+(-0.571207600446)*x[2]**o+(0.177935012163)*x[2]
        arg[0,4,0,0]=(0.974759684151)*x[0]**o+(-0.876975976179)*x[0]+(-0.472336406836)*x[1]**o+(0.991410874408)*x[1]+(0.297823444758)*x[2]**o+(-0.161436760659)*x[2]
        arg[0,4,0,1]=(-0.0799922241539)*x[0]**o+(0.975369691404)*x[0]+(0.302361880712)*x[1]**o+(0.171779641966)*x[1]+(0.878428630413)*x[2]**o+(0.814497247319)*x[2]
        arg[0,4,1,0]=(0.532633280629)*x[0]**o+(0.902436316468)*x[0]+(-0.676910177358)*x[1]**o+(-0.986029752926)*x[1]+(-0.519990783496)*x[2]**o+(0.430468215617)*x[2]
        arg[0,4,1,1]=(-0.353243658293)*x[0]**o+(-0.908938062877)*x[0]+(0.922933758125)*x[1]**o+(0.00500161721147)*x[1]+(0.189861558609)*x[2]**o+(-0.623307494462)*x[2]
        arg[0,4,2,0]=(-0.720840869801)*x[0]**o+(0.0349736577597)*x[0]+(-0.516608060915)*x[1]**o+(0.688756843577)*x[1]+(-0.302494303699)*x[2]**o+(-0.0703783111937)*x[2]
        arg[0,4,2,1]=(-0.179073400214)*x[0]**o+(0.128485402427)*x[0]+(-0.726789308786)*x[1]**o+(-0.691163482057)*x[1]+(0.329009412577)*x[2]**o+(-0.79293394396)*x[2]
        arg[1,0,0,0]=(-0.384011553333)*x[0]**o+(0.285271264922)*x[0]+(0.512105930181)*x[1]**o+(0.190321101872)*x[1]+(-0.509334982029)*x[2]**o+(0.128097742502)*x[2]
        arg[1,0,0,1]=(-0.808211570649)*x[0]**o+(-0.730450956462)*x[0]+(0.92309757237)*x[1]**o+(-0.703749565272)*x[1]+(0.301129716544)*x[2]**o+(0.244379489105)*x[2]
        arg[1,0,1,0]=(0.0404274840294)*x[0]**o+(-0.483903392253)*x[0]+(0.73680959705)*x[1]**o+(0.190850758522)*x[1]+(0.133136432605)*x[2]**o+(0.910055186498)*x[2]
        arg[1,0,1,1]=(0.302394139901)*x[0]**o+(0.802524687002)*x[0]+(0.970871728408)*x[1]**o+(0.207348092291)*x[1]+(0.666310912781)*x[2]**o+(0.510392805767)*x[2]
        arg[1,0,2,0]=(-0.981981997649)*x[0]**o+(-0.465269233707)*x[0]+(0.916146418639)*x[1]**o+(0.143602438774)*x[1]+(-0.947874263512)*x[2]**o+(0.0840394012229)*x[2]
        arg[1,0,2,1]=(-0.237479805943)*x[0]**o+(-0.89993757244)*x[0]+(-0.759650487895)*x[1]**o+(-0.807139164373)*x[1]+(-0.377055204014)*x[2]**o+(-0.57656201026)*x[2]
        arg[1,1,0,0]=(-0.543719871098)*x[0]**o+(-0.726308160246)*x[0]+(0.821244616742)*x[1]**o+(0.718112961487)*x[1]+(-0.655445553192)*x[2]**o+(-0.599104608292)*x[2]
        arg[1,1,0,1]=(-0.2624625471)*x[0]**o+(0.85250071151)*x[0]+(-0.665066081561)*x[1]**o+(-0.566502138196)*x[1]+(-0.572858653628)*x[2]**o+(-0.862890633323)*x[2]
        arg[1,1,1,0]=(0.88003513115)*x[0]**o+(-0.0150959972442)*x[0]+(-0.698880737368)*x[1]**o+(-0.236909810427)*x[1]+(-0.721315930706)*x[2]**o+(-0.219894530829)*x[2]
        arg[1,1,1,1]=(0.177746615088)*x[0]**o+(0.205790814183)*x[0]+(-0.216021190604)*x[1]**o+(0.384079010991)*x[1]+(0.61634008905)*x[2]**o+(0.940900776108)*x[2]
        arg[1,1,2,0]=(0.889070121473)*x[0]**o+(0.68131533527)*x[0]+(0.709207182081)*x[1]**o+(-0.765184991252)*x[1]+(0.506328018952)*x[2]**o+(-0.46041174473)*x[2]
        arg[1,1,2,1]=(0.951736215954)*x[0]**o+(-0.147499702642)*x[0]+(0.60348358384)*x[1]**o+(0.777404433567)*x[1]+(-0.482441452679)*x[2]**o+(0.714835256434)*x[2]
        arg[1,2,0,0]=(-0.954701036263)*x[0]**o+(-0.360901887475)*x[0]+(-0.644285028623)*x[1]**o+(0.169271524632)*x[1]+(-0.687047203668)*x[2]**o+(0.205115312108)*x[2]
        arg[1,2,0,1]=(0.256489282842)*x[0]**o+(0.17406448362)*x[0]+(0.97500243816)*x[1]**o+(0.614577997602)*x[1]+(0.0537536804583)*x[2]**o+(-0.211963274032)*x[2]
        arg[1,2,1,0]=(-0.94015914517)*x[0]**o+(0.169982490736)*x[0]+(0.482085495724)*x[1]**o+(-0.556994404153)*x[1]+(-0.809311296883)*x[2]**o+(0.719907503024)*x[2]
        arg[1,2,1,1]=(0.367156421931)*x[0]**o+(0.489648825153)*x[0]+(-0.0398225517816)*x[1]**o+(-0.495726296032)*x[1]+(0.359374061053)*x[2]**o+(0.231991952482)*x[2]
        arg[1,2,2,0]=(-0.812516694781)*x[0]**o+(-0.631181257391)*x[0]+(-0.723784471687)*x[1]**o+(0.178976877342)*x[1]+(-0.629293758013)*x[2]**o+(-0.313897724608)*x[2]
        arg[1,2,2,1]=(-0.497461944186)*x[0]**o+(0.268998525103)*x[0]+(-0.478017492026)*x[1]**o+(0.026395792841)*x[1]+(0.581256386137)*x[2]**o+(-0.476961644324)*x[2]
        arg[1,3,0,0]=(-0.227077550639)*x[0]**o+(0.400407424161)*x[0]+(-0.53529812303)*x[1]**o+(-0.418594299853)*x[1]+(-0.195190916346)*x[2]**o+(-0.624284483774)*x[2]
        arg[1,3,0,1]=(0.822091021645)*x[0]**o+(-0.31133140026)*x[0]+(0.0665130965049)*x[1]**o+(0.596697216047)*x[1]+(0.466195828646)*x[2]**o+(-0.477988391299)*x[2]
        arg[1,3,1,0]=(-0.0341577310045)*x[0]**o+(-0.77936644392)*x[0]+(0.954811382784)*x[1]**o+(0.837374830631)*x[1]+(0.738992234768)*x[2]**o+(-0.501187788985)*x[2]
        arg[1,3,1,1]=(-0.102867224088)*x[0]**o+(0.00829793194765)*x[0]+(0.081790556957)*x[1]**o+(-0.779351210472)*x[1]+(-0.735321115716)*x[2]**o+(0.126919542418)*x[2]
        arg[1,3,2,0]=(0.556301076488)*x[0]**o+(0.0923434629749)*x[0]+(-0.191048735713)*x[1]**o+(-0.218833028687)*x[1]+(-0.088867516146)*x[2]**o+(-0.247214200389)*x[2]
        arg[1,3,2,1]=(0.888436531887)*x[0]**o+(-0.79225495914)*x[0]+(0.0154788262219)*x[1]**o+(-0.210622890864)*x[1]+(0.782629134632)*x[2]**o+(0.350772185791)*x[2]
        arg[1,4,0,0]=(0.651898073685)*x[0]**o+(-0.84098616689)*x[0]+(0.674088959443)*x[1]**o+(-0.658824952898)*x[1]+(-0.423551541558)*x[2]**o+(0.985098530356)*x[2]
        arg[1,4,0,1]=(-0.149187282052)*x[0]**o+(0.0880370482988)*x[0]+(0.780809276595)*x[1]**o+(-0.610833366472)*x[1]+(0.794946384304)*x[2]**o+(-0.817031796429)*x[2]
        arg[1,4,1,0]=(-0.285938055108)*x[0]**o+(-0.106165154734)*x[0]+(-0.948376845189)*x[1]**o+(0.874607682056)*x[1]+(-0.100051825909)*x[2]**o+(-0.483640835026)*x[2]
        arg[1,4,1,1]=(-0.34015136472)*x[0]**o+(0.991331180352)*x[0]+(-0.925796518928)*x[1]**o+(-0.872775691287)*x[1]+(0.953898056819)*x[2]**o+(0.985025365955)*x[2]
        arg[1,4,2,0]=(-0.464687865356)*x[0]**o+(-0.439271850879)*x[0]+(0.302716649876)*x[1]**o+(0.235186289965)*x[1]+(0.672062761925)*x[2]**o+(-0.793689262392)*x[2]
        arg[1,4,2,1]=(-0.777967458564)*x[0]**o+(0.0565891925223)*x[0]+(-0.27539733623)*x[1]**o+(0.252329868046)*x[1]+(-0.269529792944)*x[2]**o+(-0.953051723483)*x[2]
        arg[2,0,0,0]=(0.665321754689)*x[0]**o+(0.00639681939987)*x[0]+(0.262193891905)*x[1]**o+(-0.312800735151)*x[1]+(0.8403244871)*x[2]**o+(-0.607084328204)*x[2]
        arg[2,0,0,1]=(0.876435018569)*x[0]**o+(-0.388250784543)*x[0]+(-0.799866349878)*x[1]**o+(0.114896594902)*x[1]+(-0.438064877186)*x[2]**o+(-0.458871344685)*x[2]
        arg[2,0,1,0]=(-0.344552466761)*x[0]**o+(0.967217640605)*x[0]+(-0.00183230845245)*x[1]**o+(-0.861206591247)*x[1]+(0.795588257851)*x[2]**o+(-0.333607767932)*x[2]
        arg[2,0,1,1]=(0.841069648591)*x[0]**o+(0.580162460873)*x[0]+(-0.0277035983747)*x[1]**o+(-0.404616536792)*x[1]+(-0.403195438454)*x[2]**o+(-0.202877947947)*x[2]
        arg[2,0,2,0]=(0.192691357169)*x[0]**o+(-0.519968251714)*x[0]+(-0.575397578091)*x[1]**o+(-0.029378989829)*x[1]+(-0.958861195684)*x[2]**o+(0.848705509543)*x[2]
        arg[2,0,2,1]=(0.556616365284)*x[0]**o+(0.679993071506)*x[0]+(-0.800652717274)*x[1]**o+(0.979182847258)*x[1]+(-0.501953582042)*x[2]**o+(0.127003985335)*x[2]
        arg[2,1,0,0]=(0.947750123445)*x[0]**o+(-0.0350691553359)*x[0]+(0.0650485999092)*x[1]**o+(-0.131216537555)*x[1]+(0.888782559831)*x[2]**o+(0.0338703818654)*x[2]
        arg[2,1,0,1]=(0.962918564559)*x[0]**o+(-0.119367421389)*x[0]+(0.43003549616)*x[1]**o+(-0.097883763913)*x[1]+(0.230422206305)*x[2]**o+(0.187273479711)*x[2]
        arg[2,1,1,0]=(-0.549909211973)*x[0]**o+(0.263443098451)*x[0]+(-0.476703606205)*x[1]**o+(0.216159301182)*x[1]+(-0.561829938323)*x[2]**o+(0.943016557203)*x[2]
        arg[2,1,1,1]=(0.382867749635)*x[0]**o+(0.404317006935)*x[0]+(0.786117847428)*x[1]**o+(-0.80719258612)*x[1]+(0.471144889777)*x[2]**o+(0.6478813349)*x[2]
        arg[2,1,2,0]=(0.355742024596)*x[0]**o+(-0.728812391059)*x[0]+(0.766039604734)*x[1]**o+(-0.532629598363)*x[1]+(-0.518284303077)*x[2]**o+(-0.972348067947)*x[2]
        arg[2,1,2,1]=(-0.882334243081)*x[0]**o+(-0.0940633993231)*x[0]+(-0.00166643832426)*x[1]**o+(-0.971243725495)*x[1]+(-0.0422614083707)*x[2]**o+(-0.547881520447)*x[2]
        arg[2,2,0,0]=(0.372164678051)*x[0]**o+(-0.0644361811234)*x[0]+(0.886078780084)*x[1]**o+(-0.459054628604)*x[1]+(-0.983880547097)*x[2]**o+(-0.669749873181)*x[2]
        arg[2,2,0,1]=(0.126650330827)*x[0]**o+(0.375111974063)*x[0]+(-0.759797654917)*x[1]**o+(-0.663216091322)*x[1]+(-0.683191373782)*x[2]**o+(-0.607048449469)*x[2]
        arg[2,2,1,0]=(-0.805920040628)*x[0]**o+(-0.0144685877057)*x[0]+(0.00875222056291)*x[1]**o+(-0.110000267929)*x[1]+(-0.478201664476)*x[2]**o+(-0.0537439708493)*x[2]
        arg[2,2,1,1]=(-0.550624517803)*x[0]**o+(0.570407090431)*x[0]+(0.661026826148)*x[1]**o+(0.905189813242)*x[1]+(0.00707531661352)*x[2]**o+(0.711072343064)*x[2]
        arg[2,2,2,0]=(-0.644276976064)*x[0]**o+(-0.151665972531)*x[0]+(-0.645369927259)*x[1]**o+(0.915967227715)*x[1]+(-0.211750817665)*x[2]**o+(-0.596646212118)*x[2]
        arg[2,2,2,1]=(0.564688999107)*x[0]**o+(0.635094410913)*x[0]+(0.281863287848)*x[1]**o+(-0.927529659408)*x[1]+(0.581071087484)*x[2]**o+(0.164399277305)*x[2]
        arg[2,3,0,0]=(0.64754765253)*x[0]**o+(0.176693707964)*x[0]+(0.172801737347)*x[1]**o+(-0.769408746261)*x[1]+(-0.176026486601)*x[2]**o+(0.187304474822)*x[2]
        arg[2,3,0,1]=(0.365315558093)*x[0]**o+(0.888853264239)*x[0]+(0.0541759434362)*x[1]**o+(0.241489226711)*x[1]+(0.441822201832)*x[2]**o+(0.075541840076)*x[2]
        arg[2,3,1,0]=(-0.984940830356)*x[0]**o+(0.521596714942)*x[0]+(-0.933671223761)*x[1]**o+(-0.132907654205)*x[1]+(-0.225804727932)*x[2]**o+(0.987334445626)*x[2]
        arg[2,3,1,1]=(-0.785516613303)*x[0]**o+(0.0324126413814)*x[0]+(-0.084233004699)*x[1]**o+(0.45131442059)*x[1]+(0.36498759129)*x[2]**o+(0.417036982561)*x[2]
        arg[2,3,2,0]=(-0.310575473358)*x[0]**o+(-0.769219454218)*x[0]+(-0.263431454577)*x[1]**o+(0.474143448264)*x[1]+(-0.490633240671)*x[2]**o+(0.435609988849)*x[2]
        arg[2,3,2,1]=(0.230960030802)*x[0]**o+(-0.480593443732)*x[0]+(0.547888534735)*x[1]**o+(-0.818842709599)*x[1]+(-0.857141410253)*x[2]**o+(-0.287898487847)*x[2]
        arg[2,4,0,0]=(-0.811008715357)*x[0]**o+(-0.608505768579)*x[0]+(-0.117642307757)*x[1]**o+(-0.658229045784)*x[1]+(0.246542958193)*x[2]**o+(-0.42783248265)*x[2]
        arg[2,4,0,1]=(0.817994599948)*x[0]**o+(0.394931907893)*x[0]+(-0.273993145403)*x[1]**o+(0.9572397013)*x[1]+(0.101025833331)*x[2]**o+(-0.925874548613)*x[2]
        arg[2,4,1,0]=(-0.626152000269)*x[0]**o+(0.821322294557)*x[0]+(0.543761841392)*x[1]**o+(0.112341361443)*x[1]+(-0.395657376418)*x[2]**o+(0.815481187683)*x[2]
        arg[2,4,1,1]=(0.97134483493)*x[0]**o+(0.303341541402)*x[0]+(-0.270102237344)*x[1]**o+(-0.982452271352)*x[1]+(0.435568243011)*x[2]**o+(-0.200051496445)*x[2]
        arg[2,4,2,0]=(-0.827528679064)*x[0]**o+(-0.559447627475)*x[0]+(-0.867397899368)*x[1]**o+(0.915539157819)*x[1]+(0.370118757042)*x[2]**o+(0.578366179408)*x[2]
        arg[2,4,2,1]=(0.685882920875)*x[0]**o+(0.830016468585)*x[0]+(-0.603607199824)*x[1]**o+(0.888619939961)*x[1]+(-0.264173896988)*x[2]**o+(-0.588065135532)*x[2]
        arg[3,0,0,0]=(0.480000662344)*x[0]**o+(-0.240072007989)*x[0]+(0.252669294205)*x[1]**o+(-0.657466732916)*x[1]+(0.945149322572)*x[2]**o+(0.340592627149)*x[2]
        arg[3,0,0,1]=(-0.819956304014)*x[0]**o+(0.910408628621)*x[0]+(-0.875714560314)*x[1]**o+(0.653323198674)*x[1]+(0.910127896386)*x[2]**o+(0.21382952806)*x[2]
        arg[3,0,1,0]=(0.736655857249)*x[0]**o+(-0.982636511328)*x[0]+(0.747965385081)*x[1]**o+(-0.952056699676)*x[1]+(0.143168283422)*x[2]**o+(0.891386576579)*x[2]
        arg[3,0,1,1]=(0.615723453617)*x[0]**o+(0.57376968582)*x[0]+(0.766802940979)*x[1]**o+(-0.294849095727)*x[1]+(-0.26718123792)*x[2]**o+(-0.979175148343)*x[2]
        arg[3,0,2,0]=(-0.299975789693)*x[0]**o+(-0.941879795797)*x[0]+(-0.332633604206)*x[1]**o+(0.887589360228)*x[1]+(0.570321438438)*x[2]**o+(0.107180499249)*x[2]
        arg[3,0,2,1]=(-0.77479605873)*x[0]**o+(0.928359807986)*x[0]+(-0.144170026638)*x[1]**o+(0.533835316696)*x[1]+(0.183938136052)*x[2]**o+(-0.87336340483)*x[2]
        arg[3,1,0,0]=(-0.121858163593)*x[0]**o+(-0.971122306623)*x[0]+(-0.351891554168)*x[1]**o+(-0.989754547278)*x[1]+(0.48106592456)*x[2]**o+(0.750009716044)*x[2]
        arg[3,1,0,1]=(-0.332320017719)*x[0]**o+(0.229814969323)*x[0]+(0.950038106614)*x[1]**o+(-0.364221636516)*x[1]+(-0.842684764717)*x[2]**o+(-0.430003876831)*x[2]
        arg[3,1,1,0]=(0.1036461168)*x[0]**o+(0.243047542195)*x[0]+(-0.189881740385)*x[1]**o+(-0.646948186795)*x[1]+(-0.246523431838)*x[2]**o+(0.0363222140239)*x[2]
        arg[3,1,1,1]=(0.629747493968)*x[0]**o+(0.257110968013)*x[0]+(-0.594332699815)*x[1]**o+(-0.74041867705)*x[1]+(0.656583120031)*x[2]**o+(0.392776534802)*x[2]
        arg[3,1,2,0]=(0.997368307137)*x[0]**o+(-0.709004475665)*x[0]+(-0.379447862403)*x[1]**o+(-0.368700793823)*x[1]+(0.293673648471)*x[2]**o+(-0.555545691082)*x[2]
        arg[3,1,2,1]=(-0.323027033928)*x[0]**o+(0.804640557762)*x[0]+(-0.199434780533)*x[1]**o+(0.604194682736)*x[1]+(0.945516011113)*x[2]**o+(-0.831838977373)*x[2]
        arg[3,2,0,0]=(0.255071922458)*x[0]**o+(0.261075290259)*x[0]+(0.198792431961)*x[1]**o+(0.987681568369)*x[1]+(-0.257626079249)*x[2]**o+(0.817801578613)*x[2]
        arg[3,2,0,1]=(-0.732615916833)*x[0]**o+(0.8114347023)*x[0]+(0.308896981216)*x[1]**o+(-0.164981244979)*x[1]+(0.724411853266)*x[2]**o+(-0.647969863039)*x[2]
        arg[3,2,1,0]=(0.126755528121)*x[0]**o+(0.0203968097627)*x[0]+(0.350888385037)*x[1]**o+(0.765584878761)*x[1]+(0.43521215637)*x[2]**o+(0.5821821143)*x[2]
        arg[3,2,1,1]=(0.795956472753)*x[0]**o+(0.363631111688)*x[0]+(-0.296417527594)*x[1]**o+(-0.0927953540926)*x[1]+(0.396041397741)*x[2]**o+(0.43932000066)*x[2]
        arg[3,2,2,0]=(-0.926754341344)*x[0]**o+(-0.688414966522)*x[0]+(-0.660984101932)*x[1]**o+(-0.627883344982)*x[1]+(0.403036784646)*x[2]**o+(0.818277068314)*x[2]
        arg[3,2,2,1]=(-0.358921307079)*x[0]**o+(0.636847962491)*x[0]+(-0.510050460541)*x[1]**o+(-0.758190944228)*x[1]+(-0.0319326931965)*x[2]**o+(-0.643422203077)*x[2]
        arg[3,3,0,0]=(-0.930730482483)*x[0]**o+(-0.883991363505)*x[0]+(0.156751809769)*x[1]**o+(0.669949596594)*x[1]+(0.697840698407)*x[2]**o+(0.721622061073)*x[2]
        arg[3,3,0,1]=(0.108814498453)*x[0]**o+(-0.164024398834)*x[0]+(-0.172084798233)*x[1]**o+(0.031530303081)*x[1]+(-0.985663868487)*x[2]**o+(-0.290224642955)*x[2]
        arg[3,3,1,0]=(0.690575802875)*x[0]**o+(-0.720428203968)*x[0]+(-0.837111899789)*x[1]**o+(0.140256279088)*x[1]+(0.532813462778)*x[2]**o+(-0.557885762558)*x[2]
        arg[3,3,1,1]=(-0.0785388159589)*x[0]**o+(-0.40595445804)*x[0]+(-0.691255332383)*x[1]**o+(0.54108326284)*x[1]+(0.737684932212)*x[2]**o+(0.423508847197)*x[2]
        arg[3,3,2,0]=(-0.589838770269)*x[0]**o+(-0.557133995504)*x[0]+(-0.297626023793)*x[1]**o+(-0.511809000987)*x[1]+(-0.933566600332)*x[2]**o+(-0.21434993446)*x[2]
        arg[3,3,2,1]=(0.19892263441)*x[0]**o+(0.633952342142)*x[0]+(0.0179400729177)*x[1]**o+(0.154020947355)*x[1]+(-0.620690264897)*x[2]**o+(0.524853027077)*x[2]
        arg[3,4,0,0]=(0.475231682303)*x[0]**o+(-0.48174419167)*x[0]+(-0.100854189221)*x[1]**o+(0.818166518054)*x[1]+(-0.922566309657)*x[2]**o+(-0.498745068755)*x[2]
        arg[3,4,0,1]=(0.613982633884)*x[0]**o+(-0.947323459945)*x[0]+(0.17395204013)*x[1]**o+(0.929669087853)*x[1]+(-0.617278089466)*x[2]**o+(0.900537470392)*x[2]
        arg[3,4,1,0]=(-0.403633510324)*x[0]**o+(0.954097943555)*x[0]+(-0.463422043607)*x[1]**o+(0.843916487223)*x[1]+(-0.5955912138)*x[2]**o+(-0.765444657727)*x[2]
        arg[3,4,1,1]=(-0.663412886802)*x[0]**o+(-0.547440885077)*x[0]+(-0.479993570508)*x[1]**o+(0.0620303034445)*x[1]+(-0.653759917895)*x[2]**o+(-0.681004441743)*x[2]
        arg[3,4,2,0]=(-0.994529375243)*x[0]**o+(0.546983479745)*x[0]+(0.17913912435)*x[1]**o+(0.660366383627)*x[1]+(0.0747943517387)*x[2]**o+(-0.321063660452)*x[2]
        arg[3,4,2,1]=(-0.538074654323)*x[0]**o+(-0.81330034965)*x[0]+(0.660614896398)*x[1]**o+(-0.918192909648)*x[1]+(0.553635166581)*x[2]**o+(-0.918131802189)*x[2]
        ref[0,0,0,0]=(-0.321539504757)/(o+1.)+(0.762866286444)+(-0.793251694335)*0.5**o
        ref[0,0,0,1]=(-0.264812286159)/(o+1.)+(-0.15293893368)+(-0.624206604764)*0.5**o
        ref[0,0,1,0]=(0.54441377247)/(o+1.)+(0.214067163116)+(0.115999561481)*0.5**o
        ref[0,0,1,1]=(1.50036854956)/(o+1.)+(0.673819079573)+(-0.972466831671)*0.5**o
        ref[0,0,2,0]=(0.687874396856)/(o+1.)+(-0.986502459901)+(0.708180260094)*0.5**o
        ref[0,0,2,1]=(0.793527041633)/(o+1.)+(-0.259940301856)+(0.0282615573045)*0.5**o
        ref[0,1,0,0]=(-1.25436610505)/(o+1.)+(-0.224984055664)+(0.231621820235)*0.5**o
        ref[0,1,0,1]=(0.891022819882)/(o+1.)+(-0.165060941565)+(0.114056801155)*0.5**o
        ref[0,1,1,0]=(-0.833444470649)/(o+1.)+(-0.322644910601)+(-0.973115195017)*0.5**o
        ref[0,1,1,1]=(-0.189799709023)/(o+1.)+(-0.242650862468)+(-0.304758579369)*0.5**o
        ref[0,1,2,0]=(-1.39857111894)/(o+1.)+(-0.14133379247)+(0.0923844967157)*0.5**o
        ref[0,1,2,1]=(1.03702218206)/(o+1.)+(-0.181927579186)+(0.793568921835)*0.5**o
        ref[0,2,0,0]=(-1.11962316818)/(o+1.)+(0.654627255004)+(0.570673248048)*0.5**o
        ref[0,2,0,1]=(0.085812863531)/(o+1.)+(0.17639811125)+(0.299061155215)*0.5**o
        ref[0,2,1,0]=(-1.26466090832)/(o+1.)+(0.618506300342)+(0.694725244159)*0.5**o
        ref[0,2,1,1]=(0.181495495342)/(o+1.)+(0.818577989479)+(-0.83896527178)*0.5**o
        ref[0,2,2,0]=(0.867512484289)/(o+1.)+(0.353365700269)+(0.2228052704)*0.5**o
        ref[0,2,2,1]=(0.618350374002)/(o+1.)+(-0.418609375405)+(0.63141216243)*0.5**o
        ref[0,3,0,0]=(-0.801998626232)/(o+1.)+(-0.302573026989)+(0.00698107166168)*0.5**o
        ref[0,3,0,1]=(0.414118735868)/(o+1.)+(0.103407742435)+(0.243423742325)*0.5**o
        ref[0,3,1,0]=(-0.129735888575)/(o+1.)+(0.51412655284)+(-0.106629554393)*0.5**o
        ref[0,3,1,1]=(0.0886696480864)/(o+1.)+(0.287824302591)+(0.375143733111)*0.5**o
        ref[0,3,2,0]=(-0.746501172831)/(o+1.)+(-0.389518072046)+(0.955685455438)*0.5**o
        ref[0,3,2,1]=(-1.06628341794)/(o+1.)+(-0.147429422878)+(0.353654445481)*0.5**o
        ref[0,4,0,0]=(-0.174512962078)/(o+1.)+(-0.023500931215)+(0.974759684151)*0.5**o
        ref[0,4,0,1]=(1.18079051113)/(o+1.)+(0.980823290345)+(-0.0799922241539)*0.5**o
        ref[0,4,1,0]=(-1.19690096085)/(o+1.)+(0.173437389579)+(0.532633280629)*0.5**o
        ref[0,4,1,1]=(1.11279531673)/(o+1.)+(-0.763621970064)+(-0.353243658293)*0.5**o
        ref[0,4,2,0]=(-0.819102364613)/(o+1.)+(0.326676095071)+(-0.720840869801)*0.5**o
        ref[0,4,2,1]=(-0.397779896209)/(o+1.)+(-0.677806011795)+(-0.179073400214)*0.5**o
        ref[1,0,0,0]=(0.00277094815158)/(o+1.)+(0.301845054648)+(-0.384011553333)*0.5**o
        ref[1,0,0,1]=(1.22422728891)/(o+1.)+(-0.594910516315)+(-0.808211570649)*0.5**o
        ref[1,0,1,0]=(0.869946029655)/(o+1.)+(0.308501276384)+(0.0404274840294)*0.5**o
        ref[1,0,1,1]=(1.63718264119)/(o+1.)+(0.76013279253)+(0.302394139901)*0.5**o
        ref[1,0,2,0]=(-0.0317278448724)/(o+1.)+(-0.118813696855)+(-0.981981997649)*0.5**o
        ref[1,0,2,1]=(-1.13670569191)/(o+1.)+(-1.14181937354)+(-0.237479805943)*0.5**o
        ref[1,1,0,0]=(0.16579906355)/(o+1.)+(-0.303649903525)+(-0.543719871098)*0.5**o
        ref[1,1,0,1]=(-1.23792473519)/(o+1.)+(-0.288446030005)+(-0.2624625471)*0.5**o
        ref[1,1,1,0]=(-1.42019666807)/(o+1.)+(-0.23595016925)+(0.88003513115)*0.5**o
        ref[1,1,1,1]=(0.400318898446)/(o+1.)+(0.765385300641)+(0.177746615088)*0.5**o
        ref[1,1,2,0]=(1.21553520103)/(o+1.)+(-0.272140700356)+(0.889070121473)*0.5**o
        ref[1,1,2,1]=(0.121042131161)/(o+1.)+(0.672369993679)+(0.951736215954)*0.5**o
        ref[1,2,0,0]=(-1.33133223229)/(o+1.)+(0.0067424746325)+(-0.954701036263)*0.5**o
        ref[1,2,0,1]=(1.02875611862)/(o+1.)+(0.288339603595)+(0.256489282842)*0.5**o
        ref[1,2,1,0]=(-0.32722580116)/(o+1.)+(0.166447794803)+(-0.94015914517)*0.5**o
        ref[1,2,1,1]=(0.319551509271)/(o+1.)+(0.112957240802)+(0.367156421931)*0.5**o
        ref[1,2,2,0]=(-1.3530782297)/(o+1.)+(-0.383051052329)+(-0.812516694781)*0.5**o
        ref[1,2,2,1]=(0.103238894111)/(o+1.)+(-0.09078366319)+(-0.497461944186)*0.5**o
        ref[1,3,0,0]=(-0.730489039376)/(o+1.)+(-0.321235679733)+(-0.227077550639)*0.5**o
        ref[1,3,0,1]=(0.532708925151)/(o+1.)+(-0.0963112877562)+(0.822091021645)*0.5**o
        ref[1,3,1,0]=(1.69380361755)/(o+1.)+(-0.221589701137)+(-0.0341577310045)*0.5**o
        ref[1,3,1,1]=(-0.653530558759)/(o+1.)+(-0.322066868053)+(-0.102867224088)*0.5**o
        ref[1,3,2,0]=(-0.279916251859)/(o+1.)+(-0.18685188305)+(0.556301076488)*0.5**o
        ref[1,3,2,1]=(0.798107960854)/(o+1.)+(-0.326052832107)+(0.888436531887)*0.5**o
        ref[1,4,0,0]=(0.250537417885)/(o+1.)+(-0.257356294716)+(0.651898073685)*0.5**o
        ref[1,4,0,1]=(1.5757556609)/(o+1.)+(-0.669914057301)+(-0.149187282052)*0.5**o
        ref[1,4,1,0]=(-1.0484286711)/(o+1.)+(0.142400846148)+(-0.285938055108)*0.5**o
        ref[1,4,1,1]=(0.028101537891)/(o+1.)+(0.55179042751)+(-0.34015136472)*0.5**o
        ref[1,4,2,0]=(0.974779411801)/(o+1.)+(-0.498887411653)+(-0.464687865356)*0.5**o
        ref[1,4,2,1]=(-0.544927129174)/(o+1.)+(-0.322066331457)+(-0.777967458564)*0.5**o
        ref[2,0,0,0]=(1.102518379)/(o+1.)+(-0.456744121978)+(0.665321754689)*0.5**o
        ref[2,0,0,1]=(-1.23793122706)/(o+1.)+(-0.366112767163)+(0.876435018569)*0.5**o
        ref[2,0,1,0]=(0.793755949398)/(o+1.)+(-0.113798359287)+(-0.344552466761)*0.5**o
        ref[2,0,1,1]=(-0.430899036829)/(o+1.)+(-0.0136660119329)+(0.841069648591)*0.5**o
        ref[2,0,2,0]=(-1.53425877377)/(o+1.)+(0.149679134)+(0.192691357169)*0.5**o
        ref[2,0,2,1]=(-1.30260629932)/(o+1.)+(0.89308995205)+(0.556616365284)*0.5**o
        ref[2,1,0,0]=(0.95383115974)/(o+1.)+(-0.0662076555126)+(0.947750123445)*0.5**o
        ref[2,1,0,1]=(0.660457702465)/(o+1.)+(-0.0149888527952)+(0.962918564559)*0.5**o
        ref[2,1,1,0]=(-1.03853354453)/(o+1.)+(0.711309478418)+(-0.549909211973)*0.5**o
        ref[2,1,1,1]=(1.25726273721)/(o+1.)+(0.122502877857)+(0.382867749635)*0.5**o
        ref[2,1,2,0]=(0.247755301657)/(o+1.)+(-1.11689502868)+(0.355742024596)*0.5**o
        ref[2,1,2,1]=(-0.043927846695)/(o+1.)+(-0.806594322632)+(-0.882334243081)*0.5**o
        ref[2,2,0,0]=(-0.0978017670124)/(o+1.)+(-0.596620341454)+(0.372164678051)*0.5**o
        ref[2,2,0,1]=(-1.4429890287)/(o+1.)+(-0.447576283364)+(0.126650330827)*0.5**o
        ref[2,2,1,0]=(-0.469449443913)/(o+1.)+(-0.0891064132419)+(-0.805920040628)*0.5**o
        ref[2,2,1,1]=(0.668102142762)/(o+1.)+(1.09333462337)+(-0.550624517803)*0.5**o
        ref[2,2,2,0]=(-0.857120744924)/(o+1.)+(0.0838275215328)+(-0.644276976064)*0.5**o
        ref[2,2,2,1]=(0.862934375332)/(o+1.)+(-0.0640179855954)+(0.564688999107)*0.5**o
        ref[2,3,0,0]=(-0.00322474925369)/(o+1.)+(-0.202705281737)+(0.64754765253)*0.5**o
        ref[2,3,0,1]=(0.495998145268)/(o+1.)+(0.602942165513)+(0.365315558093)*0.5**o
        ref[2,3,1,0]=(-1.15947595169)/(o+1.)+(0.688011753181)+(-0.984940830356)*0.5**o
        ref[2,3,1,1]=(0.280754586591)/(o+1.)+(0.450382022266)+(-0.785516613303)*0.5**o
        ref[2,3,2,0]=(-0.754064695248)/(o+1.)+(0.0702669914474)+(-0.310575473358)*0.5**o
        ref[2,3,2,1]=(-0.309252875518)/(o+1.)+(-0.793667320589)+(0.230960030802)*0.5**o
        ref[2,4,0,0]=(0.128900650436)/(o+1.)+(-0.847283648506)+(-0.811008715357)*0.5**o
        ref[2,4,0,1]=(-0.172967312072)/(o+1.)+(0.21314853029)+(0.817994599948)*0.5**o
        ref[2,4,1,0]=(0.148104464974)/(o+1.)+(0.874572421841)+(-0.626152000269)*0.5**o
        ref[2,4,1,1]=(0.165466005667)/(o+1.)+(-0.439581113198)+(0.97134483493)*0.5**o
        ref[2,4,2,0]=(-0.497279142326)/(o+1.)+(0.467228854876)+(-0.827528679064)*0.5**o
        ref[2,4,2,1]=(-0.867781096812)/(o+1.)+(0.565285636507)+(0.685882920875)*0.5**o
        ref[3,0,0,0]=(1.19781861678)/(o+1.)+(-0.278473056878)+(0.480000662344)*0.5**o
        ref[3,0,0,1]=(0.0344133360711)/(o+1.)+(0.888780677678)+(-0.819956304014)*0.5**o
        ref[3,0,1,0]=(0.891133668503)/(o+1.)+(-0.521653317213)+(0.736655857249)*0.5**o
        ref[3,0,1,1]=(0.499621703059)/(o+1.)+(-0.350127279125)+(0.615723453617)*0.5**o
        ref[3,0,2,0]=(0.237687834232)/(o+1.)+(0.0264450318399)+(-0.299975789693)*0.5**o
        ref[3,0,2,1]=(0.0397681094138)/(o+1.)+(0.294415859926)+(-0.77479605873)*0.5**o
        ref[3,1,0,0]=(0.129174370391)/(o+1.)+(-0.605433568929)+(-0.121858163593)*0.5**o
        ref[3,1,0,1]=(0.107353341897)/(o+1.)+(-0.282205272012)+(-0.332320017719)*0.5**o
        ref[3,1,1,0]=(-0.436405172223)/(o+1.)+(-0.183789215288)+(0.1036461168)*0.5**o
        ref[3,1,1,1]=(0.0622504202157)/(o+1.)+(-0.0452655871179)+(0.629747493968)*0.5**o
        ref[3,1,2,0]=(-0.0857742139322)/(o+1.)+(-0.816625480285)+(0.997368307137)*0.5**o
        ref[3,1,2,1]=(0.746081230581)/(o+1.)+(0.288498131562)+(-0.323027033928)*0.5**o
        ref[3,2,0,0]=(-0.0588336472879)/(o+1.)+(1.03327921862)+(0.255071922458)*0.5**o
        ref[3,2,0,1]=(1.03330883448)/(o+1.)+(-0.000758202858916)+(-0.732615916833)*0.5**o
        ref[3,2,1,0]=(0.786100541406)/(o+1.)+(0.684081901412)+(0.126755528121)*0.5**o
        ref[3,2,1,1]=(0.099623870147)/(o+1.)+(0.355077879128)+(0.795956472753)*0.5**o
        ref[3,2,2,0]=(-0.257947317287)/(o+1.)+(-0.249010621595)+(-0.926754341344)*0.5**o
        ref[3,2,2,1]=(-0.541983153738)/(o+1.)+(-0.382382592407)+(-0.358921307079)*0.5**o
        ref[3,3,0,0]=(0.854592508176)/(o+1.)+(0.253790147081)+(-0.930730482483)*0.5**o
        ref[3,3,0,1]=(-1.15774866672)/(o+1.)+(-0.211359369354)+(0.108814498453)*0.5**o
        ref[3,3,1,0]=(-0.304298437011)/(o+1.)+(-0.569028843719)+(0.690575802875)*0.5**o
        ref[3,3,1,1]=(0.0464295998283)/(o+1.)+(0.279318825998)+(-0.0785388159589)*0.5**o
        ref[3,3,2,0]=(-1.23119262412)/(o+1.)+(-0.641646465476)+(-0.589838770269)*0.5**o
        ref[3,3,2,1]=(-0.60275019198)/(o+1.)+(0.656413158287)+(0.19892263441)*0.5**o
        ref[3,4,0,0]=(-1.02342049888)/(o+1.)+(-0.0811613711857)+(0.475231682303)*0.5**o
        ref[3,4,0,1]=(-0.443326049337)/(o+1.)+(0.44144154915)+(0.613982633884)*0.5**o
        ref[3,4,1,0]=(-1.05901325741)/(o+1.)+(0.516284886526)+(-0.403633510324)*0.5**o
        ref[3,4,1,1]=(-1.1337534884)/(o+1.)+(-0.583207511688)+(-0.663412886802)*0.5**o
        ref[3,4,2,0]=(0.253933476088)/(o+1.)+(0.44314310146)+(-0.994529375243)*0.5**o
        ref[3,4,2,1]=(1.21425006298)/(o+1.)+(-1.32481253074)+(-0.538074654323)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.441098788712)*x[0]**o+(0.082953252724)*x[0]+(-0.97005044508)*x[1]**o+(0.928239302707)*x[1]
        ref=(-0.97005044508)/(o+1.)+(0.505596277715)+(-0.441098788712)*0.5**o
      else:
        arg=(0.384995424392)*x[0]**o+(0.999298118235)*x[0]+(0.144745460948)*x[1]**o+(0.0545198589076)*x[1]+(-0.773375797864)*x[2]**o+(0.145524911495)*x[2]
        ref=(-0.628630336917)/(o+1.)+(0.599671444318)+(0.384995424392)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.0745889108767)*x[0]**o+(-0.694812761318)*x[0]+(-0.357791726239)*x[1]**o+(-0.785448350489)*x[1]
        arg[1]=(0.436006633056)*x[0]**o+(-0.797943632034)*x[0]+(-0.903321835716)*x[1]**o+(0.299527545727)*x[1]
        ref[0]=(-0.357791726239)/(o+1.)+(-0.740130555904)+(0.0745889108767)*0.5**o
        ref[1]=(-0.903321835716)/(o+1.)+(-0.249208043154)+(0.436006633056)*0.5**o
      else:
        arg[0]=(-0.292048194246)*x[0]**o+(-0.612402340241)*x[0]+(0.830253896674)*x[1]**o+(-0.0173245051245)*x[1]+(-0.105062444571)*x[2]**o+(0.968232377167)*x[2]
        arg[1]=(0.153118126553)*x[0]**o+(0.958628221601)*x[0]+(-0.326368437133)*x[1]**o+(0.878861879269)*x[1]+(-0.297398071379)*x[2]**o+(0.155202299471)*x[2]
        ref[0]=(0.725191452103)/(o+1.)+(0.169252765901)+(-0.292048194246)*0.5**o
        ref[1]=(-0.623766508512)/(o+1.)+(0.99634620017)+(0.153118126553)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.410346849722)*x[0]**o+(0.86102012198)*x[0]+(-0.0143604505833)*x[1]**o+(-0.333268586255)*x[1]
        arg[0,1]=(-0.104991947387)*x[0]**o+(0.51540973662)*x[0]+(0.449773354197)*x[1]**o+(0.0397236625379)*x[1]
        arg[0,2]=(0.0010671797269)*x[0]**o+(0.165202831822)*x[0]+(-0.536444395178)*x[1]**o+(0.142112903729)*x[1]
        arg[0,3]=(0.0655247081801)*x[0]**o+(-0.59093644572)*x[0]+(-0.0135713515051)*x[1]**o+(0.665578431202)*x[1]
        arg[0,4]=(0.902382714765)*x[0]**o+(-0.358716290301)*x[0]+(-0.504792876433)*x[1]**o+(-0.760982991404)*x[1]
        arg[1,0]=(0.663815334483)*x[0]**o+(-0.967978720711)*x[0]+(-0.308953852146)*x[1]**o+(-0.250315651366)*x[1]
        arg[1,1]=(0.602976841368)*x[0]**o+(0.407346455361)*x[0]+(0.820076160142)*x[1]**o+(0.720380314572)*x[1]
        arg[1,2]=(0.0899286747557)*x[0]**o+(-0.636748133402)*x[0]+(0.615853620374)*x[1]**o+(0.389136375831)*x[1]
        arg[1,3]=(-0.695859848989)*x[0]**o+(-0.160892907203)*x[0]+(0.972252211687)*x[1]**o+(0.971653054313)*x[1]
        arg[1,4]=(-0.229141368314)*x[0]**o+(0.434757135281)*x[0]+(0.508751791725)*x[1]**o+(-0.593022987916)*x[1]
        arg[2,0]=(-0.732297666821)*x[0]**o+(-0.885623150041)*x[0]+(0.355264312146)*x[1]**o+(-0.203582600498)*x[1]
        arg[2,1]=(-0.0683706126862)*x[0]**o+(0.390953181924)*x[0]+(-0.104115827873)*x[1]**o+(-0.0586349400277)*x[1]
        arg[2,2]=(0.213052201915)*x[0]**o+(0.813758731408)*x[0]+(-0.927085393446)*x[1]**o+(-0.70884993627)*x[1]
        arg[2,3]=(-0.11792311156)*x[0]**o+(0.705053056049)*x[0]+(-0.988858028527)*x[1]**o+(-0.532493233644)*x[1]
        arg[2,4]=(0.643258167584)*x[0]**o+(0.991059864292)*x[0]+(0.466734669253)*x[1]**o+(-0.763209421816)*x[1]
        arg[3,0]=(-0.686036752646)*x[0]**o+(-0.0152278415398)*x[0]+(0.572818298467)*x[1]**o+(0.0883817883272)*x[1]
        arg[3,1]=(-0.498088836117)*x[0]**o+(0.413958831603)*x[0]+(-0.788338642889)*x[1]**o+(-0.219412111863)*x[1]
        arg[3,2]=(0.895103303091)*x[0]**o+(0.967224972282)*x[0]+(0.778129303031)*x[1]**o+(-0.239067231978)*x[1]
        arg[3,3]=(-0.715385920659)*x[0]**o+(-0.928446013585)*x[0]+(-0.919578735328)*x[1]**o+(0.735438220506)*x[1]
        arg[3,4]=(0.273825880641)*x[0]**o+(0.602304312788)*x[0]+(-0.788415866411)*x[1]**o+(0.841776933557)*x[1]
        ref[0,0]=(-0.0143604505833)/(o+1.)+(0.263875767862)+(0.410346849722)*0.5**o
        ref[0,1]=(0.449773354197)/(o+1.)+(0.277566699579)+(-0.104991947387)*0.5**o
        ref[0,2]=(-0.536444395178)/(o+1.)+(0.153657867776)+(0.0010671797269)*0.5**o
        ref[0,3]=(-0.0135713515051)/(o+1.)+(0.0373209927411)+(0.0655247081801)*0.5**o
        ref[0,4]=(-0.504792876433)/(o+1.)+(-0.559849640852)+(0.902382714765)*0.5**o
        ref[1,0]=(-0.308953852146)/(o+1.)+(-0.609147186038)+(0.663815334483)*0.5**o
        ref[1,1]=(0.820076160142)/(o+1.)+(0.563863384966)+(0.602976841368)*0.5**o
        ref[1,2]=(0.615853620374)/(o+1.)+(-0.123805878786)+(0.0899286747557)*0.5**o
        ref[1,3]=(0.972252211687)/(o+1.)+(0.405380073555)+(-0.695859848989)*0.5**o
        ref[1,4]=(0.508751791725)/(o+1.)+(-0.0791329263176)+(-0.229141368314)*0.5**o
        ref[2,0]=(0.355264312146)/(o+1.)+(-0.544602875269)+(-0.732297666821)*0.5**o
        ref[2,1]=(-0.104115827873)/(o+1.)+(0.166159120948)+(-0.0683706126862)*0.5**o
        ref[2,2]=(-0.927085393446)/(o+1.)+(0.0524543975689)+(0.213052201915)*0.5**o
        ref[2,3]=(-0.988858028527)/(o+1.)+(0.0862799112023)+(-0.11792311156)*0.5**o
        ref[2,4]=(0.466734669253)/(o+1.)+(0.113925221238)+(0.643258167584)*0.5**o
        ref[3,0]=(0.572818298467)/(o+1.)+(0.0365769733937)+(-0.686036752646)*0.5**o
        ref[3,1]=(-0.788338642889)/(o+1.)+(0.0972733598697)+(-0.498088836117)*0.5**o
        ref[3,2]=(0.778129303031)/(o+1.)+(0.364078870152)+(0.895103303091)*0.5**o
        ref[3,3]=(-0.919578735328)/(o+1.)+(-0.0965038965398)+(-0.715385920659)*0.5**o
        ref[3,4]=(-0.788415866411)/(o+1.)+(0.722040623173)+(0.273825880641)*0.5**o
      else:
        arg[0,0]=(-0.903197979191)*x[0]**o+(-0.497252911882)*x[0]+(0.5689188153)*x[1]**o+(-0.569434759407)*x[1]+(-0.204533481399)*x[2]**o+(-0.0255207088467)*x[2]
        arg[0,1]=(-0.0296606374337)*x[0]**o+(0.917552577919)*x[0]+(0.746132694062)*x[1]**o+(0.285568252958)*x[1]+(0.175571603104)*x[2]**o+(0.5701931124)*x[2]
        arg[0,2]=(0.657981464039)*x[0]**o+(-0.105816675859)*x[0]+(-0.210726712386)*x[1]**o+(-0.457303699577)*x[1]+(-0.225517453595)*x[2]**o+(0.61593257413)*x[2]
        arg[0,3]=(-0.652766009035)*x[0]**o+(0.538908603782)*x[0]+(0.226594150058)*x[1]**o+(0.691428293374)*x[1]+(-0.210896072508)*x[2]**o+(0.538061905948)*x[2]
        arg[0,4]=(0.175024193446)*x[0]**o+(0.433285241007)*x[0]+(0.928411521986)*x[1]**o+(-0.241175862661)*x[1]+(-0.779192552542)*x[2]**o+(0.428100531631)*x[2]
        arg[1,0]=(-0.094230207815)*x[0]**o+(-0.113255272088)*x[0]+(-0.880816596679)*x[1]**o+(0.56261116141)*x[1]+(-0.418606336511)*x[2]**o+(0.858058623759)*x[2]
        arg[1,1]=(0.08556853498)*x[0]**o+(0.95491440456)*x[0]+(-0.435275492231)*x[1]**o+(0.339105115812)*x[1]+(-0.634101556975)*x[2]**o+(0.357557194572)*x[2]
        arg[1,2]=(-0.692883812524)*x[0]**o+(0.673540357384)*x[0]+(0.240219465837)*x[1]**o+(0.139957925862)*x[1]+(0.135892430596)*x[2]**o+(0.383465818668)*x[2]
        arg[1,3]=(-0.892921935206)*x[0]**o+(0.0694647682287)*x[0]+(0.0530888770821)*x[1]**o+(-0.0735398003251)*x[1]+(-0.40702392212)*x[2]**o+(-0.223494881098)*x[2]
        arg[1,4]=(0.743112585337)*x[0]**o+(-0.12393012942)*x[0]+(-0.660616664197)*x[1]**o+(-0.0334489536139)*x[1]+(-0.0566394013558)*x[2]**o+(-0.0281273872158)*x[2]
        arg[2,0]=(-0.500500871517)*x[0]**o+(-0.326237506809)*x[0]+(-0.960480146386)*x[1]**o+(0.0901654083154)*x[1]+(0.466124845857)*x[2]**o+(0.6908848619)*x[2]
        arg[2,1]=(0.120233846989)*x[0]**o+(0.520609850629)*x[0]+(-0.407423400749)*x[1]**o+(0.59145151128)*x[1]+(0.981171407459)*x[2]**o+(-0.703987165108)*x[2]
        arg[2,2]=(0.811775583161)*x[0]**o+(-0.209650983138)*x[0]+(-0.286223486261)*x[1]**o+(-0.864122954179)*x[1]+(0.609725467363)*x[2]**o+(0.0311464898903)*x[2]
        arg[2,3]=(-0.325647787579)*x[0]**o+(0.956001955119)*x[0]+(0.637272546129)*x[1]**o+(-0.951109798103)*x[1]+(0.406445558466)*x[2]**o+(-0.930854597943)*x[2]
        arg[2,4]=(0.0621315789778)*x[0]**o+(-0.917033227802)*x[0]+(-0.179584904974)*x[1]**o+(0.283357518809)*x[1]+(-0.778863909832)*x[2]**o+(-0.524497331838)*x[2]
        arg[3,0]=(0.140605486894)*x[0]**o+(0.172882849759)*x[0]+(-0.883586756712)*x[1]**o+(-0.160461884399)*x[1]+(-0.718343713337)*x[2]**o+(0.196937721256)*x[2]
        arg[3,1]=(-0.660232764173)*x[0]**o+(-0.92091042631)*x[0]+(-0.568852134661)*x[1]**o+(-0.331441330509)*x[1]+(-0.220286922375)*x[2]**o+(-0.256407439105)*x[2]
        arg[3,2]=(0.593940468197)*x[0]**o+(0.527756437117)*x[0]+(0.943980933807)*x[1]**o+(0.749415167795)*x[1]+(0.484043332105)*x[2]**o+(0.40135030381)*x[2]
        arg[3,3]=(0.870508483705)*x[0]**o+(-0.234209640456)*x[0]+(-0.62692310882)*x[1]**o+(0.352628981854)*x[1]+(-0.24937844453)*x[2]**o+(0.73438735864)*x[2]
        arg[3,4]=(-0.638437449353)*x[0]**o+(0.308007229542)*x[0]+(0.537590844528)*x[1]**o+(-0.755595388361)*x[1]+(-0.152781228745)*x[2]**o+(-0.450122095362)*x[2]
        ref[0,0]=(0.364385333901)/(o+1.)+(-0.546104190068)+(-0.903197979191)*0.5**o
        ref[0,1]=(0.921704297167)/(o+1.)+(0.886656971638)+(-0.0296606374337)*0.5**o
        ref[0,2]=(-0.436244165981)/(o+1.)+(0.0264060993469)+(0.657981464039)*0.5**o
        ref[0,3]=(0.0156980775503)/(o+1.)+(0.884199401552)+(-0.652766009035)*0.5**o
        ref[0,4]=(0.149218969444)/(o+1.)+(0.310104954989)+(0.175024193446)*0.5**o
        ref[1,0]=(-1.29942293319)/(o+1.)+(0.653707256541)+(-0.094230207815)*0.5**o
        ref[1,1]=(-1.06937704921)/(o+1.)+(0.825788357473)+(0.08556853498)*0.5**o
        ref[1,2]=(0.376111896433)/(o+1.)+(0.598482050957)+(-0.692883812524)*0.5**o
        ref[1,3]=(-0.353935045038)/(o+1.)+(-0.113784956597)+(-0.892921935206)*0.5**o
        ref[1,4]=(-0.717256065552)/(o+1.)+(-0.0927532351247)+(0.743112585337)*0.5**o
        ref[2,0]=(-0.494355300528)/(o+1.)+(0.227406381703)+(-0.500500871517)*0.5**o
        ref[2,1]=(0.57374800671)/(o+1.)+(0.2040370984)+(0.120233846989)*0.5**o
        ref[2,2]=(0.323501981101)/(o+1.)+(-0.521313723714)+(0.811775583161)*0.5**o
        ref[2,3]=(1.04371810459)/(o+1.)+(-0.462981220464)+(-0.325647787579)*0.5**o
        ref[2,4]=(-0.958448814807)/(o+1.)+(-0.579086520416)+(0.0621315789778)*0.5**o
        ref[3,0]=(-1.60193047005)/(o+1.)+(0.104679343308)+(0.140605486894)*0.5**o
        ref[3,1]=(-0.789139057035)/(o+1.)+(-0.754379597962)+(-0.660232764173)*0.5**o
        ref[3,2]=(1.42802426591)/(o+1.)+(0.839260954361)+(0.593940468197)*0.5**o
        ref[3,3]=(-0.87630155335)/(o+1.)+(0.426403350019)+(0.870508483705)*0.5**o
        ref[3,4]=(0.384809615783)/(o+1.)+(-0.448855127091)+(-0.638437449353)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.315756069525)*x[0]**o+(0.253240552004)*x[0]+(0.86540335823)*x[1]**o+(-0.616203780065)*x[1]
        arg[0,0,1]=(-0.948116617977)*x[0]**o+(0.892154336439)*x[0]+(0.136605864823)*x[1]**o+(-0.135599413854)*x[1]
        arg[0,1,0]=(0.249388669107)*x[0]**o+(-0.299043026617)*x[0]+(0.335963868826)*x[1]**o+(-0.147202749474)*x[1]
        arg[0,1,1]=(0.649003124445)*x[0]**o+(0.102344972673)*x[0]+(-0.795038258171)*x[1]**o+(0.149716983842)*x[1]
        arg[1,0,0]=(0.814606374941)*x[0]**o+(-0.949251236145)*x[0]+(0.723727126173)*x[1]**o+(-0.75419822708)*x[1]
        arg[1,0,1]=(0.043251358201)*x[0]**o+(-0.227748374318)*x[0]+(-0.231556019421)*x[1]**o+(-0.152524439909)*x[1]
        arg[1,1,0]=(0.565989211326)*x[0]**o+(-0.488248650798)*x[0]+(-0.353077780419)*x[1]**o+(-0.0139123365766)*x[1]
        arg[1,1,1]=(-0.106085432662)*x[0]**o+(0.627311916563)*x[0]+(-0.496301561324)*x[1]**o+(-0.648448849796)*x[1]
        arg[2,0,0]=(0.741868098354)*x[0]**o+(-0.486898566862)*x[0]+(0.0461464851414)*x[1]**o+(0.0133916949979)*x[1]
        arg[2,0,1]=(0.969108066213)*x[0]**o+(-0.688797899381)*x[0]+(0.897017408844)*x[1]**o+(0.807714585761)*x[1]
        arg[2,1,0]=(-0.138159486947)*x[0]**o+(-0.377759480876)*x[0]+(-0.764484355987)*x[1]**o+(-0.723912265361)*x[1]
        arg[2,1,1]=(-0.781167132076)*x[0]**o+(0.119512137054)*x[0]+(-0.535230596139)*x[1]**o+(0.441358968417)*x[1]
        arg[3,0,0]=(-0.663820836909)*x[0]**o+(0.907668749761)*x[0]+(-0.345114213135)*x[1]**o+(0.0580058218243)*x[1]
        arg[3,0,1]=(-0.359054701663)*x[0]**o+(0.0389188839076)*x[0]+(0.168050801441)*x[1]**o+(-0.695770024565)*x[1]
        arg[3,1,0]=(0.70118269628)*x[0]**o+(0.87976625188)*x[0]+(0.154503286758)*x[1]**o+(0.807878896951)*x[1]
        arg[3,1,1]=(-0.143516320927)*x[0]**o+(-0.369452129865)*x[0]+(-0.0392268449697)*x[1]**o+(-0.046104227955)*x[1]
        arg[4,0,0]=(0.197244254754)*x[0]**o+(-0.580444991466)*x[0]+(0.334036744845)*x[1]**o+(0.000791083620504)*x[1]
        arg[4,0,1]=(-0.892626226865)*x[0]**o+(0.140042741589)*x[0]+(0.158775259845)*x[1]**o+(-0.308266331028)*x[1]
        arg[4,1,0]=(0.650374542633)*x[0]**o+(0.586152174865)*x[0]+(0.0201324360301)*x[1]**o+(-0.813221777119)*x[1]
        arg[4,1,1]=(0.149656968314)*x[0]**o+(0.700641349434)*x[0]+(-0.529780404433)*x[1]**o+(0.512159137234)*x[1]
        arg[5,0,0]=(0.288664621864)*x[0]**o+(0.119652133593)*x[0]+(-0.599113573241)*x[1]**o+(0.879966724309)*x[1]
        arg[5,0,1]=(0.861443356689)*x[0]**o+(0.530607121787)*x[0]+(0.240306513701)*x[1]**o+(-0.575772515225)*x[1]
        arg[5,1,0]=(0.721635890008)*x[0]**o+(-0.109537583449)*x[0]+(0.162041613112)*x[1]**o+(0.229143033467)*x[1]
        arg[5,1,1]=(-0.876930820494)*x[0]**o+(0.749378811252)*x[0]+(0.886213554632)*x[1]**o+(0.433327797349)*x[1]
        ref[0,0,0]=(0.86540335823)/(o+1.)+(-0.18148161403)+(-0.315756069525)*0.5**o
        ref[0,0,1]=(0.136605864823)/(o+1.)+(0.378277461292)+(-0.948116617977)*0.5**o
        ref[0,1,0]=(0.335963868826)/(o+1.)+(-0.223122888045)+(0.249388669107)*0.5**o
        ref[0,1,1]=(-0.795038258171)/(o+1.)+(0.126030978258)+(0.649003124445)*0.5**o
        ref[1,0,0]=(0.723727126173)/(o+1.)+(-0.851724731613)+(0.814606374941)*0.5**o
        ref[1,0,1]=(-0.231556019421)/(o+1.)+(-0.190136407114)+(0.043251358201)*0.5**o
        ref[1,1,0]=(-0.353077780419)/(o+1.)+(-0.251080493687)+(0.565989211326)*0.5**o
        ref[1,1,1]=(-0.496301561324)/(o+1.)+(-0.0105684666165)+(-0.106085432662)*0.5**o
        ref[2,0,0]=(0.0461464851414)/(o+1.)+(-0.236753435932)+(0.741868098354)*0.5**o
        ref[2,0,1]=(0.897017408844)/(o+1.)+(0.05945834319)+(0.969108066213)*0.5**o
        ref[2,1,0]=(-0.764484355987)/(o+1.)+(-0.550835873118)+(-0.138159486947)*0.5**o
        ref[2,1,1]=(-0.535230596139)/(o+1.)+(0.280435552736)+(-0.781167132076)*0.5**o
        ref[3,0,0]=(-0.345114213135)/(o+1.)+(0.482837285793)+(-0.663820836909)*0.5**o
        ref[3,0,1]=(0.168050801441)/(o+1.)+(-0.328425570329)+(-0.359054701663)*0.5**o
        ref[3,1,0]=(0.154503286758)/(o+1.)+(0.843822574415)+(0.70118269628)*0.5**o
        ref[3,1,1]=(-0.0392268449697)/(o+1.)+(-0.20777817891)+(-0.143516320927)*0.5**o
        ref[4,0,0]=(0.334036744845)/(o+1.)+(-0.289826953923)+(0.197244254754)*0.5**o
        ref[4,0,1]=(0.158775259845)/(o+1.)+(-0.0841117947196)+(-0.892626226865)*0.5**o
        ref[4,1,0]=(0.0201324360301)/(o+1.)+(-0.113534801127)+(0.650374542633)*0.5**o
        ref[4,1,1]=(-0.529780404433)/(o+1.)+(0.606400243334)+(0.149656968314)*0.5**o
        ref[5,0,0]=(-0.599113573241)/(o+1.)+(0.499809428951)+(0.288664621864)*0.5**o
        ref[5,0,1]=(0.240306513701)/(o+1.)+(-0.022582696719)+(0.861443356689)*0.5**o
        ref[5,1,0]=(0.162041613112)/(o+1.)+(0.0598027250089)+(0.721635890008)*0.5**o
        ref[5,1,1]=(0.886213554632)/(o+1.)+(0.5913533043)+(-0.876930820494)*0.5**o
      else:
        arg[0,0,0]=(0.804252705823)*x[0]**o+(-0.215722921442)*x[0]+(-0.806316349529)*x[1]**o+(-0.984002993222)*x[1]+(-0.411946014731)*x[2]**o+(-0.526356191855)*x[2]
        arg[0,0,1]=(-0.845902068304)*x[0]**o+(0.0773568471048)*x[0]+(-0.746877301439)*x[1]**o+(-0.581914187059)*x[1]+(-0.0299873869697)*x[2]**o+(-0.133702230427)*x[2]
        arg[0,1,0]=(0.0113663840954)*x[0]**o+(-0.155702215771)*x[0]+(-0.134490626112)*x[1]**o+(0.973329917011)*x[1]+(-0.71585624135)*x[2]**o+(-0.840113198085)*x[2]
        arg[0,1,1]=(0.708589516593)*x[0]**o+(0.40282025998)*x[0]+(0.114498353117)*x[1]**o+(0.117244230433)*x[1]+(0.110135318853)*x[2]**o+(-0.0227879044898)*x[2]
        arg[1,0,0]=(0.836102144172)*x[0]**o+(0.212564417583)*x[0]+(0.744354739679)*x[1]**o+(-0.265633978841)*x[1]+(0.843108381264)*x[2]**o+(-0.288428427501)*x[2]
        arg[1,0,1]=(-0.0712209756059)*x[0]**o+(-0.186158140814)*x[0]+(0.184457099888)*x[1]**o+(0.381365494077)*x[1]+(-0.547874738232)*x[2]**o+(0.645124013254)*x[2]
        arg[1,1,0]=(-0.982992052158)*x[0]**o+(0.259627201342)*x[0]+(-0.885159598031)*x[1]**o+(-0.720754022982)*x[1]+(-0.762444117798)*x[2]**o+(0.737661938668)*x[2]
        arg[1,1,1]=(0.518426662725)*x[0]**o+(0.718331452733)*x[0]+(0.523870992899)*x[1]**o+(0.527831420124)*x[1]+(0.791634534316)*x[2]**o+(0.17320870773)*x[2]
        arg[2,0,0]=(-0.0368697940364)*x[0]**o+(-0.101718169515)*x[0]+(0.933712377315)*x[1]**o+(-0.465193151914)*x[1]+(-0.113637506449)*x[2]**o+(0.606619949802)*x[2]
        arg[2,0,1]=(-0.785633726304)*x[0]**o+(0.526518355588)*x[0]+(0.151859749675)*x[1]**o+(0.805094094642)*x[1]+(-0.881122119524)*x[2]**o+(0.85143600659)*x[2]
        arg[2,1,0]=(0.969244001518)*x[0]**o+(0.22600349917)*x[0]+(0.404826692149)*x[1]**o+(0.633039564539)*x[1]+(0.470567669552)*x[2]**o+(0.378348877016)*x[2]
        arg[2,1,1]=(-0.310035929524)*x[0]**o+(0.42632271287)*x[0]+(0.866930472029)*x[1]**o+(-0.124483457934)*x[1]+(0.291920406068)*x[2]**o+(0.573984377154)*x[2]
        arg[3,0,0]=(0.245553971829)*x[0]**o+(-0.567563766526)*x[0]+(0.326654705159)*x[1]**o+(0.642370867777)*x[1]+(-0.538450732704)*x[2]**o+(0.773349470647)*x[2]
        arg[3,0,1]=(0.427017913395)*x[0]**o+(0.885132021286)*x[0]+(0.0338747045378)*x[1]**o+(-0.978508839789)*x[1]+(-0.734559666249)*x[2]**o+(0.379991690191)*x[2]
        arg[3,1,0]=(-0.29533334088)*x[0]**o+(-0.894548834889)*x[0]+(0.537043269081)*x[1]**o+(-0.559173506199)*x[1]+(0.359216600837)*x[2]**o+(0.269563745611)*x[2]
        arg[3,1,1]=(-0.146152544907)*x[0]**o+(-0.95681567795)*x[0]+(-0.0323978120064)*x[1]**o+(-0.5351716012)*x[1]+(-0.548536561192)*x[2]**o+(0.489675460511)*x[2]
        arg[4,0,0]=(0.0552991386013)*x[0]**o+(0.772417012152)*x[0]+(-0.298212559744)*x[1]**o+(-0.189934984078)*x[1]+(0.74843296731)*x[2]**o+(-0.00412448772016)*x[2]
        arg[4,0,1]=(0.577046858044)*x[0]**o+(-0.804142676861)*x[0]+(-0.0703211154568)*x[1]**o+(-0.84092953544)*x[1]+(0.254119640557)*x[2]**o+(-0.765048199333)*x[2]
        arg[4,1,0]=(-0.358558179011)*x[0]**o+(-0.998072158283)*x[0]+(0.508152998513)*x[1]**o+(-0.0813296609683)*x[1]+(-0.534332816004)*x[2]**o+(-0.189623317277)*x[2]
        arg[4,1,1]=(0.840098089823)*x[0]**o+(0.966726062679)*x[0]+(0.920079609258)*x[1]**o+(-0.242532131681)*x[1]+(0.429713768025)*x[2]**o+(0.171223902751)*x[2]
        arg[5,0,0]=(0.695331559401)*x[0]**o+(0.819482211381)*x[0]+(0.550452400412)*x[1]**o+(0.518496496336)*x[1]+(-0.24727850431)*x[2]**o+(-0.796406493488)*x[2]
        arg[5,0,1]=(0.469488028534)*x[0]**o+(0.457530124732)*x[0]+(0.922206279456)*x[1]**o+(0.678758573006)*x[1]+(0.34102098746)*x[2]**o+(0.108616278675)*x[2]
        arg[5,1,0]=(0.444799933589)*x[0]**o+(-0.361568786934)*x[0]+(0.470174557338)*x[1]**o+(-0.344557214342)*x[1]+(0.894860467988)*x[2]**o+(-0.347027743579)*x[2]
        arg[5,1,1]=(0.0304781332872)*x[0]**o+(-0.69046475182)*x[0]+(0.873260975232)*x[1]**o+(-0.705100555993)*x[1]+(0.660057279013)*x[2]**o+(-0.910002146105)*x[2]
        ref[0,0,0]=(-1.21826236426)/(o+1.)+(-0.863041053259)+(0.804252705823)*0.5**o
        ref[0,0,1]=(-0.776864688409)/(o+1.)+(-0.31912978519)+(-0.845902068304)*0.5**o
        ref[0,1,0]=(-0.850346867462)/(o+1.)+(-0.0112427484226)+(0.0113663840954)*0.5**o
        ref[0,1,1]=(0.22463367197)/(o+1.)+(0.248638292962)+(0.708589516593)*0.5**o
        ref[1,0,0]=(1.58746312094)/(o+1.)+(-0.17074899438)+(0.836102144172)*0.5**o
        ref[1,0,1]=(-0.363417638344)/(o+1.)+(0.420165683258)+(-0.0712209756059)*0.5**o
        ref[1,1,0]=(-1.64760371583)/(o+1.)+(0.138267558514)+(-0.982992052158)*0.5**o
        ref[1,1,1]=(1.31550552722)/(o+1.)+(0.709685790293)+(0.518426662725)*0.5**o
        ref[2,0,0]=(0.820074870865)/(o+1.)+(0.0198543141864)+(-0.0368697940364)*0.5**o
        ref[2,0,1]=(-0.729262369848)/(o+1.)+(1.09152422841)+(-0.785633726304)*0.5**o
        ref[2,1,0]=(0.875394361701)/(o+1.)+(0.618695970363)+(0.969244001518)*0.5**o
        ref[2,1,1]=(1.1588508781)/(o+1.)+(0.437911816045)+(-0.310035929524)*0.5**o
        ref[3,0,0]=(-0.211796027546)/(o+1.)+(0.424078285949)+(0.245553971829)*0.5**o
        ref[3,0,1]=(-0.700684961711)/(o+1.)+(0.143307435844)+(0.427017913395)*0.5**o
        ref[3,1,0]=(0.896259869918)/(o+1.)+(-0.592079297738)+(-0.29533334088)*0.5**o
        ref[3,1,1]=(-0.580934373198)/(o+1.)+(-0.50115590932)+(-0.146152544907)*0.5**o
        ref[4,0,0]=(0.450220407566)/(o+1.)+(0.289178770177)+(0.0552991386013)*0.5**o
        ref[4,0,1]=(0.1837985251)/(o+1.)+(-1.20506020582)+(0.577046858044)*0.5**o
        ref[4,1,0]=(-0.0261798174917)/(o+1.)+(-0.634512568264)+(-0.358558179011)*0.5**o
        ref[4,1,1]=(1.34979337728)/(o+1.)+(0.447708916874)+(0.840098089823)*0.5**o
        ref[5,0,0]=(0.303173896103)/(o+1.)+(0.270786107114)+(0.695331559401)*0.5**o
        ref[5,0,1]=(1.26322726692)/(o+1.)+(0.622452488207)+(0.469488028534)*0.5**o
        ref[5,1,0]=(1.36503502533)/(o+1.)+(-0.526576872427)+(0.444799933589)*0.5**o
        ref[5,1,1]=(1.53331825424)/(o+1.)+(-1.15278372696)+(0.0304781332872)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.348972036919)*x[0]**o+(0.123542094911)*x[0]+(-0.596353540657)*x[1]**o+(0.717656729405)*x[1]
        arg[0,0,0,1]=(0.187733014896)*x[0]**o+(0.76244697815)*x[0]+(0.261409664458)*x[1]**o+(0.0213714031642)*x[1]
        arg[0,0,1,0]=(0.951655197513)*x[0]**o+(-0.865274104487)*x[0]+(0.59127040848)*x[1]**o+(-0.261622300136)*x[1]
        arg[0,0,1,1]=(-0.188127934304)*x[0]**o+(-0.513768204254)*x[0]+(0.77162116582)*x[1]**o+(-0.12621176914)*x[1]
        arg[0,0,2,0]=(0.770717242337)*x[0]**o+(-0.991038362841)*x[0]+(0.908566805373)*x[1]**o+(-0.24890064672)*x[1]
        arg[0,0,2,1]=(-0.543253359807)*x[0]**o+(-0.0504511921259)*x[0]+(-0.724584621389)*x[1]**o+(0.999965312141)*x[1]
        arg[0,1,0,0]=(0.986939140609)*x[0]**o+(0.236894127675)*x[0]+(0.520162431169)*x[1]**o+(0.306413415759)*x[1]
        arg[0,1,0,1]=(-0.171773212929)*x[0]**o+(-0.248436968201)*x[0]+(0.43818378458)*x[1]**o+(-0.495734982435)*x[1]
        arg[0,1,1,0]=(-0.369350989684)*x[0]**o+(0.59047357584)*x[0]+(0.422340945702)*x[1]**o+(0.326270009775)*x[1]
        arg[0,1,1,1]=(0.559606389467)*x[0]**o+(-0.0507064379501)*x[0]+(-0.0464190048363)*x[1]**o+(0.18964879235)*x[1]
        arg[0,1,2,0]=(-0.989408371198)*x[0]**o+(0.108750148268)*x[0]+(-0.347984444559)*x[1]**o+(-0.819590699683)*x[1]
        arg[0,1,2,1]=(-0.289295900777)*x[0]**o+(-0.579698633174)*x[0]+(-0.124928198101)*x[1]**o+(0.640901698875)*x[1]
        arg[0,2,0,0]=(0.132599667548)*x[0]**o+(-0.102095433147)*x[0]+(0.446169236302)*x[1]**o+(-0.844062525718)*x[1]
        arg[0,2,0,1]=(-0.452662917924)*x[0]**o+(-0.0497770418153)*x[0]+(-0.258525519083)*x[1]**o+(-0.0365197858914)*x[1]
        arg[0,2,1,0]=(0.318740871592)*x[0]**o+(0.562764515639)*x[0]+(-0.209089266173)*x[1]**o+(0.922024900771)*x[1]
        arg[0,2,1,1]=(-0.825757440563)*x[0]**o+(0.634009249217)*x[0]+(0.954836398565)*x[1]**o+(-0.910652766097)*x[1]
        arg[0,2,2,0]=(-0.640337007803)*x[0]**o+(0.862625623816)*x[0]+(-0.0327529254483)*x[1]**o+(-0.327305638603)*x[1]
        arg[0,2,2,1]=(-0.956300293885)*x[0]**o+(0.456116376134)*x[0]+(0.312023159058)*x[1]**o+(0.768546792276)*x[1]
        arg[0,3,0,0]=(-0.927040285049)*x[0]**o+(-0.614994799603)*x[0]+(-0.431859510779)*x[1]**o+(-0.088774748513)*x[1]
        arg[0,3,0,1]=(-0.807009561768)*x[0]**o+(0.00323409463398)*x[0]+(0.274814268019)*x[1]**o+(-0.132755038246)*x[1]
        arg[0,3,1,0]=(0.931774139846)*x[0]**o+(0.665327108449)*x[0]+(0.130437344004)*x[1]**o+(-0.00859025546518)*x[1]
        arg[0,3,1,1]=(-0.314188008746)*x[0]**o+(0.405251729437)*x[0]+(-0.313484462262)*x[1]**o+(-0.670765905491)*x[1]
        arg[0,3,2,0]=(0.748691505108)*x[0]**o+(0.717110963353)*x[0]+(-0.451892315269)*x[1]**o+(-0.103930621637)*x[1]
        arg[0,3,2,1]=(-0.124290384657)*x[0]**o+(0.0761550925153)*x[0]+(0.864166710803)*x[1]**o+(0.050297196468)*x[1]
        arg[0,4,0,0]=(-0.726606985714)*x[0]**o+(-0.304862755503)*x[0]+(-0.364735884875)*x[1]**o+(0.769999145518)*x[1]
        arg[0,4,0,1]=(-0.716457484159)*x[0]**o+(0.850185561874)*x[0]+(0.924535166786)*x[1]**o+(0.414539526778)*x[1]
        arg[0,4,1,0]=(-0.67154700238)*x[0]**o+(-0.0990079744678)*x[0]+(-0.738258048173)*x[1]**o+(-0.105333364414)*x[1]
        arg[0,4,1,1]=(-0.568569022142)*x[0]**o+(-0.865340476096)*x[0]+(0.592223896873)*x[1]**o+(0.545216501911)*x[1]
        arg[0,4,2,0]=(0.959619955514)*x[0]**o+(-0.625097978453)*x[0]+(0.447944032533)*x[1]**o+(-0.792670959618)*x[1]
        arg[0,4,2,1]=(-0.446578795272)*x[0]**o+(0.516341492325)*x[0]+(0.383143400712)*x[1]**o+(-0.339255681885)*x[1]
        arg[1,0,0,0]=(0.130600745253)*x[0]**o+(-0.225084866063)*x[0]+(-0.871963625443)*x[1]**o+(-0.0741251944735)*x[1]
        arg[1,0,0,1]=(-0.684630134211)*x[0]**o+(-0.298097256903)*x[0]+(0.511206122216)*x[1]**o+(-0.784104136535)*x[1]
        arg[1,0,1,0]=(0.624178362847)*x[0]**o+(0.161787148072)*x[0]+(-0.0906465041298)*x[1]**o+(-0.0302149132362)*x[1]
        arg[1,0,1,1]=(-0.403988173215)*x[0]**o+(-0.423895624882)*x[0]+(-0.0594261878161)*x[1]**o+(0.71753539369)*x[1]
        arg[1,0,2,0]=(0.411689797339)*x[0]**o+(0.0555768292842)*x[0]+(0.786614145981)*x[1]**o+(-0.701418967065)*x[1]
        arg[1,0,2,1]=(0.971933572797)*x[0]**o+(0.203165120814)*x[0]+(0.50340688009)*x[1]**o+(-0.443869607992)*x[1]
        arg[1,1,0,0]=(0.517782412737)*x[0]**o+(-0.233071827503)*x[0]+(0.310115588175)*x[1]**o+(-0.970604318783)*x[1]
        arg[1,1,0,1]=(0.250379154253)*x[0]**o+(-0.0914513321258)*x[0]+(0.466403827523)*x[1]**o+(-0.454369386353)*x[1]
        arg[1,1,1,0]=(-0.235520035004)*x[0]**o+(0.0932661387959)*x[0]+(0.0964933173918)*x[1]**o+(0.585245824454)*x[1]
        arg[1,1,1,1]=(-0.466145341257)*x[0]**o+(0.532910678544)*x[0]+(0.757545184326)*x[1]**o+(-0.488241176428)*x[1]
        arg[1,1,2,0]=(0.802070127079)*x[0]**o+(0.30227826497)*x[0]+(0.76818718588)*x[1]**o+(-0.966380392422)*x[1]
        arg[1,1,2,1]=(0.764411747219)*x[0]**o+(0.62151105703)*x[0]+(0.310875960956)*x[1]**o+(0.218417169467)*x[1]
        arg[1,2,0,0]=(0.627865841486)*x[0]**o+(0.328844171432)*x[0]+(0.47405718918)*x[1]**o+(0.238471462812)*x[1]
        arg[1,2,0,1]=(-0.890402471541)*x[0]**o+(0.0654874441479)*x[0]+(0.411199516291)*x[1]**o+(0.0669990208336)*x[1]
        arg[1,2,1,0]=(-0.789401901979)*x[0]**o+(-0.379725885915)*x[0]+(-0.103817842125)*x[1]**o+(0.358055283478)*x[1]
        arg[1,2,1,1]=(-0.619688531158)*x[0]**o+(0.314022911152)*x[0]+(-0.515963904961)*x[1]**o+(0.807912272743)*x[1]
        arg[1,2,2,0]=(0.696584946679)*x[0]**o+(-0.156405028983)*x[0]+(0.99374085906)*x[1]**o+(0.692362331558)*x[1]
        arg[1,2,2,1]=(0.463393460998)*x[0]**o+(-0.0302433372264)*x[0]+(0.243006695756)*x[1]**o+(0.934578685307)*x[1]
        arg[1,3,0,0]=(-0.760149691585)*x[0]**o+(0.305312987392)*x[0]+(0.478665319824)*x[1]**o+(-0.395734016727)*x[1]
        arg[1,3,0,1]=(-0.495638457166)*x[0]**o+(-0.352868042523)*x[0]+(0.0577132713041)*x[1]**o+(0.856507718332)*x[1]
        arg[1,3,1,0]=(-0.588160041725)*x[0]**o+(-0.941242199161)*x[0]+(-0.883563055172)*x[1]**o+(-0.474349426459)*x[1]
        arg[1,3,1,1]=(-0.822752260982)*x[0]**o+(0.284881595545)*x[0]+(-0.914198363943)*x[1]**o+(0.421107500666)*x[1]
        arg[1,3,2,0]=(-0.125099230328)*x[0]**o+(-0.0167977306674)*x[0]+(-0.880147366581)*x[1]**o+(0.947650728684)*x[1]
        arg[1,3,2,1]=(-0.770895899992)*x[0]**o+(-0.121705997003)*x[0]+(-0.0486858273347)*x[1]**o+(0.666179109672)*x[1]
        arg[1,4,0,0]=(-0.767476534476)*x[0]**o+(0.720013119195)*x[0]+(-0.540220631624)*x[1]**o+(-0.363702385125)*x[1]
        arg[1,4,0,1]=(0.90195509915)*x[0]**o+(0.104569298687)*x[0]+(0.901816813435)*x[1]**o+(0.421060929159)*x[1]
        arg[1,4,1,0]=(0.0901631681262)*x[0]**o+(-0.330570875574)*x[0]+(0.892147778885)*x[1]**o+(-0.102976747188)*x[1]
        arg[1,4,1,1]=(-0.336016437275)*x[0]**o+(-0.00913798206221)*x[0]+(-0.20211674106)*x[1]**o+(-0.0470039081275)*x[1]
        arg[1,4,2,0]=(0.194383418491)*x[0]**o+(-0.913993330791)*x[0]+(-0.0207457431142)*x[1]**o+(-0.728130242724)*x[1]
        arg[1,4,2,1]=(-0.285983741049)*x[0]**o+(-0.359824200302)*x[0]+(-0.286913656927)*x[1]**o+(-0.801477464894)*x[1]
        arg[2,0,0,0]=(-0.266085322677)*x[0]**o+(-0.477036323188)*x[0]+(0.967304544659)*x[1]**o+(-0.843721203165)*x[1]
        arg[2,0,0,1]=(-0.431692878995)*x[0]**o+(-0.326793076488)*x[0]+(0.478220893643)*x[1]**o+(-0.953607426852)*x[1]
        arg[2,0,1,0]=(0.800398295672)*x[0]**o+(-0.726923243756)*x[0]+(0.641194491232)*x[1]**o+(0.933755071699)*x[1]
        arg[2,0,1,1]=(0.778158607269)*x[0]**o+(0.192895772287)*x[0]+(0.160540751687)*x[1]**o+(0.132780133987)*x[1]
        arg[2,0,2,0]=(0.0700487728725)*x[0]**o+(0.366571342769)*x[0]+(0.213123464402)*x[1]**o+(0.0224970803568)*x[1]
        arg[2,0,2,1]=(-0.608321696505)*x[0]**o+(0.416761589908)*x[0]+(0.0499280394381)*x[1]**o+(-0.678150075911)*x[1]
        arg[2,1,0,0]=(-0.580987690899)*x[0]**o+(-0.625951863984)*x[0]+(0.0064094419355)*x[1]**o+(0.646045286912)*x[1]
        arg[2,1,0,1]=(-0.881255703221)*x[0]**o+(-0.153796152409)*x[0]+(0.320566497467)*x[1]**o+(-0.217827916224)*x[1]
        arg[2,1,1,0]=(0.981593086009)*x[0]**o+(0.659643450249)*x[0]+(0.824530802391)*x[1]**o+(0.109125926037)*x[1]
        arg[2,1,1,1]=(-0.79695617371)*x[0]**o+(-0.340030091407)*x[0]+(0.476648191338)*x[1]**o+(-0.751961375342)*x[1]
        arg[2,1,2,0]=(0.00785807621726)*x[0]**o+(-0.177037931946)*x[0]+(-0.322685883101)*x[1]**o+(0.639485466195)*x[1]
        arg[2,1,2,1]=(0.386667953213)*x[0]**o+(0.805115373435)*x[0]+(0.489529148574)*x[1]**o+(0.380650996708)*x[1]
        arg[2,2,0,0]=(-0.205678281473)*x[0]**o+(0.725798252716)*x[0]+(-0.884620261631)*x[1]**o+(0.737678769923)*x[1]
        arg[2,2,0,1]=(-0.859841848792)*x[0]**o+(0.215497396508)*x[0]+(-0.9231710768)*x[1]**o+(-0.504914651252)*x[1]
        arg[2,2,1,0]=(-0.0161889255216)*x[0]**o+(0.218150095097)*x[0]+(0.508505799895)*x[1]**o+(0.441270271663)*x[1]
        arg[2,2,1,1]=(0.645494826531)*x[0]**o+(0.812462584734)*x[0]+(0.498639261012)*x[1]**o+(0.695415642432)*x[1]
        arg[2,2,2,0]=(-0.855342734108)*x[0]**o+(0.174928383231)*x[0]+(-0.550636347213)*x[1]**o+(0.814295216919)*x[1]
        arg[2,2,2,1]=(-0.146152468476)*x[0]**o+(-0.810876776267)*x[0]+(0.77444775812)*x[1]**o+(0.758171250375)*x[1]
        arg[2,3,0,0]=(0.914771099535)*x[0]**o+(-0.801103658802)*x[0]+(-0.339761119169)*x[1]**o+(-0.0697409372498)*x[1]
        arg[2,3,0,1]=(-0.278017621273)*x[0]**o+(-0.916896227964)*x[0]+(0.0151205004002)*x[1]**o+(-0.36424189479)*x[1]
        arg[2,3,1,0]=(0.635845368235)*x[0]**o+(-0.286420165946)*x[0]+(-0.547314386502)*x[1]**o+(-0.297882680065)*x[1]
        arg[2,3,1,1]=(0.236860892332)*x[0]**o+(0.670289851551)*x[0]+(-0.437288046493)*x[1]**o+(0.0610684872034)*x[1]
        arg[2,3,2,0]=(0.096479176127)*x[0]**o+(-0.992608241155)*x[0]+(0.622650933171)*x[1]**o+(-0.171594052732)*x[1]
        arg[2,3,2,1]=(0.259913661654)*x[0]**o+(0.521353257934)*x[0]+(0.141212812345)*x[1]**o+(0.0311971708742)*x[1]
        arg[2,4,0,0]=(0.208498140469)*x[0]**o+(0.914285328223)*x[0]+(-0.368860713612)*x[1]**o+(0.744057640483)*x[1]
        arg[2,4,0,1]=(-0.0760701135919)*x[0]**o+(-0.0666410245297)*x[0]+(-0.442858585142)*x[1]**o+(-0.869273907698)*x[1]
        arg[2,4,1,0]=(0.694291134331)*x[0]**o+(0.153487646161)*x[0]+(-0.454726570824)*x[1]**o+(0.12533511793)*x[1]
        arg[2,4,1,1]=(0.505643124198)*x[0]**o+(0.935208927517)*x[0]+(0.120271053976)*x[1]**o+(0.472499209985)*x[1]
        arg[2,4,2,0]=(0.458890639035)*x[0]**o+(-0.38125264637)*x[0]+(-0.742525095759)*x[1]**o+(0.0727341692272)*x[1]
        arg[2,4,2,1]=(-0.228856512525)*x[0]**o+(0.215540806099)*x[0]+(-0.492076254398)*x[1]**o+(0.325121281149)*x[1]
        arg[3,0,0,0]=(0.791320435019)*x[0]**o+(0.264208069532)*x[0]+(0.105218141263)*x[1]**o+(-0.947073121631)*x[1]
        arg[3,0,0,1]=(0.384081306848)*x[0]**o+(-0.735420806741)*x[0]+(0.695451719819)*x[1]**o+(-0.104071285853)*x[1]
        arg[3,0,1,0]=(0.176280431402)*x[0]**o+(0.766248516714)*x[0]+(0.620920827635)*x[1]**o+(0.571778731591)*x[1]
        arg[3,0,1,1]=(0.895148060637)*x[0]**o+(0.120401904357)*x[0]+(0.500131556768)*x[1]**o+(-0.432523916076)*x[1]
        arg[3,0,2,0]=(-0.171375065793)*x[0]**o+(0.731777853738)*x[0]+(0.914774635339)*x[1]**o+(-0.329504614878)*x[1]
        arg[3,0,2,1]=(-0.65523706909)*x[0]**o+(-0.309487164889)*x[0]+(-0.244763521442)*x[1]**o+(0.790744116827)*x[1]
        arg[3,1,0,0]=(-0.515742288)*x[0]**o+(0.688021032329)*x[0]+(0.744558730681)*x[1]**o+(0.663410706435)*x[1]
        arg[3,1,0,1]=(0.299920493853)*x[0]**o+(0.221554378219)*x[0]+(-0.262111371211)*x[1]**o+(0.597219370531)*x[1]
        arg[3,1,1,0]=(-0.491648532317)*x[0]**o+(-0.0763921400417)*x[0]+(0.517994656474)*x[1]**o+(0.690649210498)*x[1]
        arg[3,1,1,1]=(0.291881729319)*x[0]**o+(0.388296683726)*x[0]+(0.304223657331)*x[1]**o+(-0.17249216188)*x[1]
        arg[3,1,2,0]=(0.309288981861)*x[0]**o+(-0.638172982762)*x[0]+(-0.566768084047)*x[1]**o+(-0.694439437371)*x[1]
        arg[3,1,2,1]=(-0.680391436031)*x[0]**o+(0.876096306554)*x[0]+(0.676979754874)*x[1]**o+(0.376507786465)*x[1]
        arg[3,2,0,0]=(-0.688464200232)*x[0]**o+(0.404140014906)*x[0]+(-0.389359624725)*x[1]**o+(0.218257423404)*x[1]
        arg[3,2,0,1]=(-0.321661287857)*x[0]**o+(0.305707842777)*x[0]+(0.0488067596906)*x[1]**o+(0.475071042874)*x[1]
        arg[3,2,1,0]=(-0.172172832131)*x[0]**o+(-0.0589598637728)*x[0]+(0.997578239775)*x[1]**o+(0.0919614884141)*x[1]
        arg[3,2,1,1]=(0.500590025777)*x[0]**o+(0.258724052781)*x[0]+(0.233819575609)*x[1]**o+(0.787949345995)*x[1]
        arg[3,2,2,0]=(-0.464931198743)*x[0]**o+(0.689246395117)*x[0]+(0.0701366334357)*x[1]**o+(-0.206391031385)*x[1]
        arg[3,2,2,1]=(-0.352492545609)*x[0]**o+(-0.160385273073)*x[0]+(0.876791526094)*x[1]**o+(0.748257988817)*x[1]
        arg[3,3,0,0]=(0.70346530112)*x[0]**o+(-0.351192123239)*x[0]+(-0.985298342079)*x[1]**o+(-0.286817451723)*x[1]
        arg[3,3,0,1]=(0.841285056577)*x[0]**o+(0.741986440026)*x[0]+(-0.758645627646)*x[1]**o+(-0.914494381783)*x[1]
        arg[3,3,1,0]=(-0.614293213484)*x[0]**o+(-0.746555667296)*x[0]+(-0.301781252786)*x[1]**o+(-0.721633195357)*x[1]
        arg[3,3,1,1]=(-0.287510519978)*x[0]**o+(-0.764960517935)*x[0]+(-0.948032248565)*x[1]**o+(-0.548676002552)*x[1]
        arg[3,3,2,0]=(0.145506943729)*x[0]**o+(-0.353925293653)*x[0]+(0.786904887352)*x[1]**o+(-0.36548133551)*x[1]
        arg[3,3,2,1]=(0.60782215735)*x[0]**o+(0.872444227513)*x[0]+(-0.237571902274)*x[1]**o+(-0.0559745367747)*x[1]
        arg[3,4,0,0]=(0.691107698337)*x[0]**o+(0.0800532750053)*x[0]+(-0.847456183316)*x[1]**o+(0.911068576288)*x[1]
        arg[3,4,0,1]=(-0.00167606578449)*x[0]**o+(-0.697603805697)*x[0]+(-0.362440886816)*x[1]**o+(-0.0604379772766)*x[1]
        arg[3,4,1,0]=(-0.74026325403)*x[0]**o+(-0.539076677004)*x[0]+(-0.317176913637)*x[1]**o+(0.759776288772)*x[1]
        arg[3,4,1,1]=(-0.230005454291)*x[0]**o+(-0.929557352119)*x[0]+(-0.435125210116)*x[1]**o+(0.106892135893)*x[1]
        arg[3,4,2,0]=(-0.638149620463)*x[0]**o+(-0.0511424128986)*x[0]+(0.460436795546)*x[1]**o+(-0.919535518181)*x[1]
        arg[3,4,2,1]=(-0.793537245531)*x[0]**o+(0.409768940347)*x[0]+(0.632560678557)*x[1]**o+(0.700596370851)*x[1]
        ref[0,0,0,0]=(-0.596353540657)/(o+1.)+(0.420599412158)+(-0.348972036919)*0.5**o
        ref[0,0,0,1]=(0.261409664458)/(o+1.)+(0.391909190657)+(0.187733014896)*0.5**o
        ref[0,0,1,0]=(0.59127040848)/(o+1.)+(-0.563448202311)+(0.951655197513)*0.5**o
        ref[0,0,1,1]=(0.77162116582)/(o+1.)+(-0.319989986697)+(-0.188127934304)*0.5**o
        ref[0,0,2,0]=(0.908566805373)/(o+1.)+(-0.619969504781)+(0.770717242337)*0.5**o
        ref[0,0,2,1]=(-0.724584621389)/(o+1.)+(0.474757060008)+(-0.543253359807)*0.5**o
        ref[0,1,0,0]=(0.520162431169)/(o+1.)+(0.271653771717)+(0.986939140609)*0.5**o
        ref[0,1,0,1]=(0.43818378458)/(o+1.)+(-0.372085975318)+(-0.171773212929)*0.5**o
        ref[0,1,1,0]=(0.422340945702)/(o+1.)+(0.458371792807)+(-0.369350989684)*0.5**o
        ref[0,1,1,1]=(-0.0464190048363)/(o+1.)+(0.0694711771999)+(0.559606389467)*0.5**o
        ref[0,1,2,0]=(-0.347984444559)/(o+1.)+(-0.355420275708)+(-0.989408371198)*0.5**o
        ref[0,1,2,1]=(-0.124928198101)/(o+1.)+(0.0306015328506)+(-0.289295900777)*0.5**o
        ref[0,2,0,0]=(0.446169236302)/(o+1.)+(-0.473078979432)+(0.132599667548)*0.5**o
        ref[0,2,0,1]=(-0.258525519083)/(o+1.)+(-0.0431484138534)+(-0.452662917924)*0.5**o
        ref[0,2,1,0]=(-0.209089266173)/(o+1.)+(0.742394708205)+(0.318740871592)*0.5**o
        ref[0,2,1,1]=(0.954836398565)/(o+1.)+(-0.13832175844)+(-0.825757440563)*0.5**o
        ref[0,2,2,0]=(-0.0327529254483)/(o+1.)+(0.267659992607)+(-0.640337007803)*0.5**o
        ref[0,2,2,1]=(0.312023159058)/(o+1.)+(0.612331584205)+(-0.956300293885)*0.5**o
        ref[0,3,0,0]=(-0.431859510779)/(o+1.)+(-0.351884774058)+(-0.927040285049)*0.5**o
        ref[0,3,0,1]=(0.274814268019)/(o+1.)+(-0.064760471806)+(-0.807009561768)*0.5**o
        ref[0,3,1,0]=(0.130437344004)/(o+1.)+(0.328368426492)+(0.931774139846)*0.5**o
        ref[0,3,1,1]=(-0.313484462262)/(o+1.)+(-0.132757088027)+(-0.314188008746)*0.5**o
        ref[0,3,2,0]=(-0.451892315269)/(o+1.)+(0.306590170858)+(0.748691505108)*0.5**o
        ref[0,3,2,1]=(0.864166710803)/(o+1.)+(0.0632261444916)+(-0.124290384657)*0.5**o
        ref[0,4,0,0]=(-0.364735884875)/(o+1.)+(0.232568195007)+(-0.726606985714)*0.5**o
        ref[0,4,0,1]=(0.924535166786)/(o+1.)+(0.632362544326)+(-0.716457484159)*0.5**o
        ref[0,4,1,0]=(-0.738258048173)/(o+1.)+(-0.102170669441)+(-0.67154700238)*0.5**o
        ref[0,4,1,1]=(0.592223896873)/(o+1.)+(-0.160061987092)+(-0.568569022142)*0.5**o
        ref[0,4,2,0]=(0.447944032533)/(o+1.)+(-0.708884469036)+(0.959619955514)*0.5**o
        ref[0,4,2,1]=(0.383143400712)/(o+1.)+(0.0885429052198)+(-0.446578795272)*0.5**o
        ref[1,0,0,0]=(-0.871963625443)/(o+1.)+(-0.149605030268)+(0.130600745253)*0.5**o
        ref[1,0,0,1]=(0.511206122216)/(o+1.)+(-0.541100696719)+(-0.684630134211)*0.5**o
        ref[1,0,1,0]=(-0.0906465041298)/(o+1.)+(0.065786117418)+(0.624178362847)*0.5**o
        ref[1,0,1,1]=(-0.0594261878161)/(o+1.)+(0.146819884404)+(-0.403988173215)*0.5**o
        ref[1,0,2,0]=(0.786614145981)/(o+1.)+(-0.32292106889)+(0.411689797339)*0.5**o
        ref[1,0,2,1]=(0.50340688009)/(o+1.)+(-0.120352243589)+(0.971933572797)*0.5**o
        ref[1,1,0,0]=(0.310115588175)/(o+1.)+(-0.601838073143)+(0.517782412737)*0.5**o
        ref[1,1,0,1]=(0.466403827523)/(o+1.)+(-0.27291035924)+(0.250379154253)*0.5**o
        ref[1,1,1,0]=(0.0964933173918)/(o+1.)+(0.339255981625)+(-0.235520035004)*0.5**o
        ref[1,1,1,1]=(0.757545184326)/(o+1.)+(0.0223347510579)+(-0.466145341257)*0.5**o
        ref[1,1,2,0]=(0.76818718588)/(o+1.)+(-0.332051063726)+(0.802070127079)*0.5**o
        ref[1,1,2,1]=(0.310875960956)/(o+1.)+(0.419964113249)+(0.764411747219)*0.5**o
        ref[1,2,0,0]=(0.47405718918)/(o+1.)+(0.283657817122)+(0.627865841486)*0.5**o
        ref[1,2,0,1]=(0.411199516291)/(o+1.)+(0.0662432324908)+(-0.890402471541)*0.5**o
        ref[1,2,1,0]=(-0.103817842125)/(o+1.)+(-0.0108353012186)+(-0.789401901979)*0.5**o
        ref[1,2,1,1]=(-0.515963904961)/(o+1.)+(0.560967591948)+(-0.619688531158)*0.5**o
        ref[1,2,2,0]=(0.99374085906)/(o+1.)+(0.267978651288)+(0.696584946679)*0.5**o
        ref[1,2,2,1]=(0.243006695756)/(o+1.)+(0.45216767404)+(0.463393460998)*0.5**o
        ref[1,3,0,0]=(0.478665319824)/(o+1.)+(-0.0452105146678)+(-0.760149691585)*0.5**o
        ref[1,3,0,1]=(0.0577132713041)/(o+1.)+(0.251819837904)+(-0.495638457166)*0.5**o
        ref[1,3,1,0]=(-0.883563055172)/(o+1.)+(-0.70779581281)+(-0.588160041725)*0.5**o
        ref[1,3,1,1]=(-0.914198363943)/(o+1.)+(0.352994548105)+(-0.822752260982)*0.5**o
        ref[1,3,2,0]=(-0.880147366581)/(o+1.)+(0.465426499008)+(-0.125099230328)*0.5**o
        ref[1,3,2,1]=(-0.0486858273347)/(o+1.)+(0.272236556335)+(-0.770895899992)*0.5**o
        ref[1,4,0,0]=(-0.540220631624)/(o+1.)+(0.178155367035)+(-0.767476534476)*0.5**o
        ref[1,4,0,1]=(0.901816813435)/(o+1.)+(0.262815113923)+(0.90195509915)*0.5**o
        ref[1,4,1,0]=(0.892147778885)/(o+1.)+(-0.216773811381)+(0.0901631681262)*0.5**o
        ref[1,4,1,1]=(-0.20211674106)/(o+1.)+(-0.0280709450949)+(-0.336016437275)*0.5**o
        ref[1,4,2,0]=(-0.0207457431142)/(o+1.)+(-0.821061786757)+(0.194383418491)*0.5**o
        ref[1,4,2,1]=(-0.286913656927)/(o+1.)+(-0.580650832598)+(-0.285983741049)*0.5**o
        ref[2,0,0,0]=(0.967304544659)/(o+1.)+(-0.660378763176)+(-0.266085322677)*0.5**o
        ref[2,0,0,1]=(0.478220893643)/(o+1.)+(-0.64020025167)+(-0.431692878995)*0.5**o
        ref[2,0,1,0]=(0.641194491232)/(o+1.)+(0.103415913972)+(0.800398295672)*0.5**o
        ref[2,0,1,1]=(0.160540751687)/(o+1.)+(0.162837953137)+(0.778158607269)*0.5**o
        ref[2,0,2,0]=(0.213123464402)/(o+1.)+(0.194534211563)+(0.0700487728725)*0.5**o
        ref[2,0,2,1]=(0.0499280394381)/(o+1.)+(-0.130694243001)+(-0.608321696505)*0.5**o
        ref[2,1,0,0]=(0.0064094419355)/(o+1.)+(0.0100467114639)+(-0.580987690899)*0.5**o
        ref[2,1,0,1]=(0.320566497467)/(o+1.)+(-0.185812034316)+(-0.881255703221)*0.5**o
        ref[2,1,1,0]=(0.824530802391)/(o+1.)+(0.384384688143)+(0.981593086009)*0.5**o
        ref[2,1,1,1]=(0.476648191338)/(o+1.)+(-0.545995733375)+(-0.79695617371)*0.5**o
        ref[2,1,2,0]=(-0.322685883101)/(o+1.)+(0.231223767125)+(0.00785807621726)*0.5**o
        ref[2,1,2,1]=(0.489529148574)/(o+1.)+(0.592883185071)+(0.386667953213)*0.5**o
        ref[2,2,0,0]=(-0.884620261631)/(o+1.)+(0.73173851132)+(-0.205678281473)*0.5**o
        ref[2,2,0,1]=(-0.9231710768)/(o+1.)+(-0.144708627372)+(-0.859841848792)*0.5**o
        ref[2,2,1,0]=(0.508505799895)/(o+1.)+(0.32971018338)+(-0.0161889255216)*0.5**o
        ref[2,2,1,1]=(0.498639261012)/(o+1.)+(0.753939113583)+(0.645494826531)*0.5**o
        ref[2,2,2,0]=(-0.550636347213)/(o+1.)+(0.494611800075)+(-0.855342734108)*0.5**o
        ref[2,2,2,1]=(0.77444775812)/(o+1.)+(-0.0263527629459)+(-0.146152468476)*0.5**o
        ref[2,3,0,0]=(-0.339761119169)/(o+1.)+(-0.435422298026)+(0.914771099535)*0.5**o
        ref[2,3,0,1]=(0.0151205004002)/(o+1.)+(-0.640569061377)+(-0.278017621273)*0.5**o
        ref[2,3,1,0]=(-0.547314386502)/(o+1.)+(-0.292151423006)+(0.635845368235)*0.5**o
        ref[2,3,1,1]=(-0.437288046493)/(o+1.)+(0.365679169377)+(0.236860892332)*0.5**o
        ref[2,3,2,0]=(0.622650933171)/(o+1.)+(-0.582101146943)+(0.096479176127)*0.5**o
        ref[2,3,2,1]=(0.141212812345)/(o+1.)+(0.276275214404)+(0.259913661654)*0.5**o
        ref[2,4,0,0]=(-0.368860713612)/(o+1.)+(0.829171484353)+(0.208498140469)*0.5**o
        ref[2,4,0,1]=(-0.442858585142)/(o+1.)+(-0.467957466114)+(-0.0760701135919)*0.5**o
        ref[2,4,1,0]=(-0.454726570824)/(o+1.)+(0.139411382046)+(0.694291134331)*0.5**o
        ref[2,4,1,1]=(0.120271053976)/(o+1.)+(0.703854068751)+(0.505643124198)*0.5**o
        ref[2,4,2,0]=(-0.742525095759)/(o+1.)+(-0.154259238572)+(0.458890639035)*0.5**o
        ref[2,4,2,1]=(-0.492076254398)/(o+1.)+(0.270331043624)+(-0.228856512525)*0.5**o
        ref[3,0,0,0]=(0.105218141263)/(o+1.)+(-0.34143252605)+(0.791320435019)*0.5**o
        ref[3,0,0,1]=(0.695451719819)/(o+1.)+(-0.419746046297)+(0.384081306848)*0.5**o
        ref[3,0,1,0]=(0.620920827635)/(o+1.)+(0.669013624153)+(0.176280431402)*0.5**o
        ref[3,0,1,1]=(0.500131556768)/(o+1.)+(-0.15606100586)+(0.895148060637)*0.5**o
        ref[3,0,2,0]=(0.914774635339)/(o+1.)+(0.20113661943)+(-0.171375065793)*0.5**o
        ref[3,0,2,1]=(-0.244763521442)/(o+1.)+(0.240628475969)+(-0.65523706909)*0.5**o
        ref[3,1,0,0]=(0.744558730681)/(o+1.)+(0.675715869382)+(-0.515742288)*0.5**o
        ref[3,1,0,1]=(-0.262111371211)/(o+1.)+(0.409386874375)+(0.299920493853)*0.5**o
        ref[3,1,1,0]=(0.517994656474)/(o+1.)+(0.307128535228)+(-0.491648532317)*0.5**o
        ref[3,1,1,1]=(0.304223657331)/(o+1.)+(0.107902260923)+(0.291881729319)*0.5**o
        ref[3,1,2,0]=(-0.566768084047)/(o+1.)+(-0.666306210066)+(0.309288981861)*0.5**o
        ref[3,1,2,1]=(0.676979754874)/(o+1.)+(0.62630204651)+(-0.680391436031)*0.5**o
        ref[3,2,0,0]=(-0.389359624725)/(o+1.)+(0.311198719155)+(-0.688464200232)*0.5**o
        ref[3,2,0,1]=(0.0488067596906)/(o+1.)+(0.390389442826)+(-0.321661287857)*0.5**o
        ref[3,2,1,0]=(0.997578239775)/(o+1.)+(0.0165008123207)+(-0.172172832131)*0.5**o
        ref[3,2,1,1]=(0.233819575609)/(o+1.)+(0.523336699388)+(0.500590025777)*0.5**o
        ref[3,2,2,0]=(0.0701366334357)/(o+1.)+(0.241427681866)+(-0.464931198743)*0.5**o
        ref[3,2,2,1]=(0.876791526094)/(o+1.)+(0.293936357872)+(-0.352492545609)*0.5**o
        ref[3,3,0,0]=(-0.985298342079)/(o+1.)+(-0.319004787481)+(0.70346530112)*0.5**o
        ref[3,3,0,1]=(-0.758645627646)/(o+1.)+(-0.0862539708788)+(0.841285056577)*0.5**o
        ref[3,3,1,0]=(-0.301781252786)/(o+1.)+(-0.734094431327)+(-0.614293213484)*0.5**o
        ref[3,3,1,1]=(-0.948032248565)/(o+1.)+(-0.656818260243)+(-0.287510519978)*0.5**o
        ref[3,3,2,0]=(0.786904887352)/(o+1.)+(-0.359703314582)+(0.145506943729)*0.5**o
        ref[3,3,2,1]=(-0.237571902274)/(o+1.)+(0.408234845369)+(0.60782215735)*0.5**o
        ref[3,4,0,0]=(-0.847456183316)/(o+1.)+(0.495560925646)+(0.691107698337)*0.5**o
        ref[3,4,0,1]=(-0.362440886816)/(o+1.)+(-0.379020891487)+(-0.00167606578449)*0.5**o
        ref[3,4,1,0]=(-0.317176913637)/(o+1.)+(0.110349805884)+(-0.74026325403)*0.5**o
        ref[3,4,1,1]=(-0.435125210116)/(o+1.)+(-0.411332608113)+(-0.230005454291)*0.5**o
        ref[3,4,2,0]=(0.460436795546)/(o+1.)+(-0.48533896554)+(-0.638149620463)*0.5**o
        ref[3,4,2,1]=(0.632560678557)/(o+1.)+(0.555182655599)+(-0.793537245531)*0.5**o
      else:
        arg[0,0,0,0]=(0.816458625566)*x[0]**o+(-0.438180707327)*x[0]+(-0.39836196112)*x[1]**o+(0.641701267855)*x[1]+(0.919064096208)*x[2]**o+(0.0155653082454)*x[2]
        arg[0,0,0,1]=(-0.709065020122)*x[0]**o+(-0.787938005558)*x[0]+(-0.459829890247)*x[1]**o+(-0.306587811189)*x[1]+(0.648348776519)*x[2]**o+(-0.782857657922)*x[2]
        arg[0,0,1,0]=(-0.09165636265)*x[0]**o+(-0.215464577259)*x[0]+(-0.445060581145)*x[1]**o+(0.541797069021)*x[1]+(-0.312096860473)*x[2]**o+(0.307061070632)*x[2]
        arg[0,0,1,1]=(0.900670566988)*x[0]**o+(-0.157226841228)*x[0]+(-0.00514019601556)*x[1]**o+(0.167616993369)*x[1]+(0.331077209212)*x[2]**o+(0.0797784869412)*x[2]
        arg[0,0,2,0]=(-0.883617789072)*x[0]**o+(-0.233507045302)*x[0]+(0.540311388635)*x[1]**o+(-0.0698351272775)*x[1]+(0.598059668701)*x[2]**o+(-0.842961707358)*x[2]
        arg[0,0,2,1]=(-0.770041366552)*x[0]**o+(-0.683157969301)*x[0]+(0.235580399076)*x[1]**o+(-0.994177047985)*x[1]+(-0.90074656484)*x[2]**o+(-0.0133600820046)*x[2]
        arg[0,1,0,0]=(0.90396926296)*x[0]**o+(0.0492056402145)*x[0]+(0.32744155595)*x[1]**o+(0.0162505824727)*x[1]+(0.386607661085)*x[2]**o+(0.759639829913)*x[2]
        arg[0,1,0,1]=(0.47956310963)*x[0]**o+(0.638067028609)*x[0]+(-0.520596788355)*x[1]**o+(0.476897948019)*x[1]+(0.950320232053)*x[2]**o+(-0.603034001297)*x[2]
        arg[0,1,1,0]=(-0.75177304495)*x[0]**o+(0.369014786745)*x[0]+(-0.332220011735)*x[1]**o+(-0.835059332186)*x[1]+(0.0494323092557)*x[2]**o+(-0.796200422375)*x[2]
        arg[0,1,1,1]=(-0.109937058453)*x[0]**o+(0.607637390544)*x[0]+(0.464147925564)*x[1]**o+(0.650325883743)*x[1]+(0.698196425121)*x[2]**o+(0.933002044297)*x[2]
        arg[0,1,2,0]=(-0.13801307581)*x[0]**o+(0.75768744715)*x[0]+(-0.35990643116)*x[1]**o+(-0.787975707318)*x[1]+(-0.259281382264)*x[2]**o+(0.971003238549)*x[2]
        arg[0,1,2,1]=(-0.518540484981)*x[0]**o+(0.197912267942)*x[0]+(0.481058074022)*x[1]**o+(0.918189335061)*x[1]+(0.160239271801)*x[2]**o+(0.75950986792)*x[2]
        arg[0,2,0,0]=(0.286922773164)*x[0]**o+(-0.874592150595)*x[0]+(0.416014232771)*x[1]**o+(-0.492379239759)*x[1]+(0.950640982197)*x[2]**o+(-0.10625158241)*x[2]
        arg[0,2,0,1]=(-0.116447648923)*x[0]**o+(-0.499481854319)*x[0]+(-0.556602044084)*x[1]**o+(0.753256358343)*x[1]+(0.280780022694)*x[2]**o+(-0.879728755507)*x[2]
        arg[0,2,1,0]=(-0.322803067618)*x[0]**o+(0.0974878233964)*x[0]+(0.420547949274)*x[1]**o+(0.304839957099)*x[1]+(-0.12605355567)*x[2]**o+(-0.166970509739)*x[2]
        arg[0,2,1,1]=(0.997823909059)*x[0]**o+(0.76749010767)*x[0]+(0.376351544673)*x[1]**o+(-0.316937252621)*x[1]+(-0.627670639733)*x[2]**o+(0.584121037987)*x[2]
        arg[0,2,2,0]=(0.516866019209)*x[0]**o+(0.413564799736)*x[0]+(-0.0730736747833)*x[1]**o+(-0.500984764568)*x[1]+(-0.989323894645)*x[2]**o+(-0.261458643035)*x[2]
        arg[0,2,2,1]=(0.777538237337)*x[0]**o+(-0.0623276952959)*x[0]+(0.167068351133)*x[1]**o+(-0.345550070545)*x[1]+(-0.28166577069)*x[2]**o+(-0.568650037569)*x[2]
        arg[0,3,0,0]=(-0.622067362782)*x[0]**o+(0.971384323624)*x[0]+(-0.099597286042)*x[1]**o+(-0.265680407306)*x[1]+(-0.446275631545)*x[2]**o+(-0.0637591060995)*x[2]
        arg[0,3,0,1]=(-0.49698112241)*x[0]**o+(0.766181263321)*x[0]+(-0.977014162391)*x[1]**o+(0.806444170647)*x[1]+(-0.441023882645)*x[2]**o+(-0.919279365683)*x[2]
        arg[0,3,1,0]=(0.917091542097)*x[0]**o+(-0.67016353228)*x[0]+(0.0468346020734)*x[1]**o+(0.81092419675)*x[1]+(0.174037096652)*x[2]**o+(0.564190407348)*x[2]
        arg[0,3,1,1]=(-0.437188100982)*x[0]**o+(-0.0857174204982)*x[0]+(-0.971592105442)*x[1]**o+(-0.122352019844)*x[1]+(-0.853225566811)*x[2]**o+(-0.861128476903)*x[2]
        arg[0,3,2,0]=(-0.841265485658)*x[0]**o+(0.553186811235)*x[0]+(0.414489259007)*x[1]**o+(-0.800997281522)*x[1]+(0.630269543743)*x[2]**o+(0.263584335528)*x[2]
        arg[0,3,2,1]=(0.383282548921)*x[0]**o+(0.260041562865)*x[0]+(-0.95180999378)*x[1]**o+(0.566954668715)*x[1]+(0.688843907846)*x[2]**o+(-0.166837115553)*x[2]
        arg[0,4,0,0]=(0.0385891845805)*x[0]**o+(0.324932244772)*x[0]+(-0.946997425504)*x[1]**o+(-0.745439090316)*x[1]+(-0.721657519976)*x[2]**o+(-0.714366519961)*x[2]
        arg[0,4,0,1]=(-0.387739686761)*x[0]**o+(-0.860915144676)*x[0]+(0.270056169204)*x[1]**o+(-0.914861208322)*x[1]+(-0.453514265988)*x[2]**o+(-0.495768314604)*x[2]
        arg[0,4,1,0]=(-0.329464483421)*x[0]**o+(0.313424566493)*x[0]+(0.185677682455)*x[1]**o+(-0.720200750426)*x[1]+(0.425742092834)*x[2]**o+(0.505209526425)*x[2]
        arg[0,4,1,1]=(0.63272559561)*x[0]**o+(-0.026429919604)*x[0]+(-0.14023723263)*x[1]**o+(-0.580805084228)*x[1]+(0.141232832837)*x[2]**o+(0.241327430534)*x[2]
        arg[0,4,2,0]=(-0.847582850604)*x[0]**o+(0.207270297356)*x[0]+(0.851958035148)*x[1]**o+(0.336594484304)*x[1]+(-0.221434386211)*x[2]**o+(-0.607602400875)*x[2]
        arg[0,4,2,1]=(-0.722990535113)*x[0]**o+(-0.379964881881)*x[0]+(0.825779173038)*x[1]**o+(-0.285509543358)*x[1]+(-0.518809494335)*x[2]**o+(-0.788507606636)*x[2]
        arg[1,0,0,0]=(0.0725951752952)*x[0]**o+(-0.282800153416)*x[0]+(-0.0228303688498)*x[1]**o+(-0.863127549259)*x[1]+(-0.383553228278)*x[2]**o+(-0.0151737809185)*x[2]
        arg[1,0,0,1]=(-0.254920010173)*x[0]**o+(0.160195189848)*x[0]+(0.216407190109)*x[1]**o+(-0.814599655672)*x[1]+(0.771746025302)*x[2]**o+(0.0636221760766)*x[2]
        arg[1,0,1,0]=(0.486342223622)*x[0]**o+(0.736242651359)*x[0]+(0.720722697611)*x[1]**o+(0.974853194395)*x[1]+(-0.834707965096)*x[2]**o+(-0.947169299038)*x[2]
        arg[1,0,1,1]=(0.777015150638)*x[0]**o+(-0.228082772046)*x[0]+(-0.51095285588)*x[1]**o+(-0.835057692508)*x[1]+(-0.564012944223)*x[2]**o+(-0.898040787549)*x[2]
        arg[1,0,2,0]=(0.240813275278)*x[0]**o+(-0.509566588932)*x[0]+(0.247803237451)*x[1]**o+(0.156442781059)*x[1]+(0.618367580757)*x[2]**o+(0.943435734207)*x[2]
        arg[1,0,2,1]=(-0.670257601629)*x[0]**o+(-0.470716006203)*x[0]+(-0.559207336128)*x[1]**o+(-0.81577896974)*x[1]+(0.0571093198914)*x[2]**o+(0.477103420898)*x[2]
        arg[1,1,0,0]=(0.985895581872)*x[0]**o+(-0.00359465828149)*x[0]+(0.00739312124399)*x[1]**o+(-0.931299968425)*x[1]+(-0.10047347666)*x[2]**o+(0.410589913002)*x[2]
        arg[1,1,0,1]=(-0.954132399275)*x[0]**o+(0.159222338169)*x[0]+(-0.865370419503)*x[1]**o+(-0.744696059283)*x[1]+(-0.794993976252)*x[2]**o+(0.961664562839)*x[2]
        arg[1,1,1,0]=(0.462810965625)*x[0]**o+(0.927378551017)*x[0]+(-0.105005418179)*x[1]**o+(0.282340667975)*x[1]+(-0.443931241323)*x[2]**o+(0.740071617715)*x[2]
        arg[1,1,1,1]=(0.645629353541)*x[0]**o+(0.848804271896)*x[0]+(-0.882497356826)*x[1]**o+(-0.74661619875)*x[1]+(-0.927133238344)*x[2]**o+(-0.502758459229)*x[2]
        arg[1,1,2,0]=(-0.554993653726)*x[0]**o+(-0.473286752336)*x[0]+(0.759624773306)*x[1]**o+(0.809169103932)*x[1]+(0.657766085675)*x[2]**o+(0.763913478843)*x[2]
        arg[1,1,2,1]=(0.1476587556)*x[0]**o+(0.385711162597)*x[0]+(-0.0767252754622)*x[1]**o+(-0.544968629318)*x[1]+(0.16751959089)*x[2]**o+(-0.742697412247)*x[2]
        arg[1,2,0,0]=(-0.715887407694)*x[0]**o+(0.273818352078)*x[0]+(0.0426629688432)*x[1]**o+(-0.808531265186)*x[1]+(-0.699839650334)*x[2]**o+(-0.332093073009)*x[2]
        arg[1,2,0,1]=(0.48440337855)*x[0]**o+(0.765443731546)*x[0]+(-0.0761968055499)*x[1]**o+(0.0516389293584)*x[1]+(-0.486224925751)*x[2]**o+(-0.796105709176)*x[2]
        arg[1,2,1,0]=(0.621562227153)*x[0]**o+(0.907213416636)*x[0]+(0.591392192921)*x[1]**o+(-0.825911160831)*x[1]+(0.690773959314)*x[2]**o+(-0.744144678185)*x[2]
        arg[1,2,1,1]=(0.472846565983)*x[0]**o+(-0.837395663533)*x[0]+(0.526438435448)*x[1]**o+(0.310053110281)*x[1]+(0.634301291566)*x[2]**o+(0.608270679412)*x[2]
        arg[1,2,2,0]=(-0.212000572333)*x[0]**o+(0.644218219058)*x[0]+(0.827779140409)*x[1]**o+(0.563968121281)*x[1]+(-0.0556319977635)*x[2]**o+(0.98126338415)*x[2]
        arg[1,2,2,1]=(-0.632752637148)*x[0]**o+(-0.999490611532)*x[0]+(0.0717950178432)*x[1]**o+(-0.0611949850049)*x[1]+(0.740177288895)*x[2]**o+(0.241010363236)*x[2]
        arg[1,3,0,0]=(0.390081216336)*x[0]**o+(-0.389460065845)*x[0]+(0.0812808695521)*x[1]**o+(0.0843698990866)*x[1]+(0.672844227186)*x[2]**o+(0.306345015188)*x[2]
        arg[1,3,0,1]=(-0.402887151891)*x[0]**o+(-0.908615473265)*x[0]+(-0.296307155209)*x[1]**o+(0.228858413043)*x[1]+(0.509401868055)*x[2]**o+(-0.793787805038)*x[2]
        arg[1,3,1,0]=(0.253947492129)*x[0]**o+(0.401465417831)*x[0]+(-0.942329176703)*x[1]**o+(-0.518371272083)*x[1]+(0.90142806242)*x[2]**o+(0.694084281869)*x[2]
        arg[1,3,1,1]=(-0.435656267291)*x[0]**o+(-0.743058541536)*x[0]+(-0.22434079396)*x[1]**o+(-0.414125965702)*x[1]+(0.743546448275)*x[2]**o+(0.323188772386)*x[2]
        arg[1,3,2,0]=(-0.069319863146)*x[0]**o+(-0.0749969565014)*x[0]+(0.29700943253)*x[1]**o+(0.972639470201)*x[1]+(-0.270592469279)*x[2]**o+(-0.355098058901)*x[2]
        arg[1,3,2,1]=(0.741825647399)*x[0]**o+(0.542055833087)*x[0]+(-0.0684409109262)*x[1]**o+(0.842094761143)*x[1]+(-0.777320219447)*x[2]**o+(0.758570478904)*x[2]
        arg[1,4,0,0]=(-0.66003338591)*x[0]**o+(-0.393755284573)*x[0]+(0.617154869065)*x[1]**o+(-0.480999198967)*x[1]+(0.350625972066)*x[2]**o+(0.733472524967)*x[2]
        arg[1,4,0,1]=(-0.0001759210396)*x[0]**o+(-0.0900574950432)*x[0]+(-0.592936627579)*x[1]**o+(-0.647214933825)*x[1]+(-0.43249743908)*x[2]**o+(-0.901387692551)*x[2]
        arg[1,4,1,0]=(0.052435899043)*x[0]**o+(0.163326392489)*x[0]+(0.433169876656)*x[1]**o+(0.977133611649)*x[1]+(-0.311959404527)*x[2]**o+(0.297218570651)*x[2]
        arg[1,4,1,1]=(0.00806522196579)*x[0]**o+(-0.121807466189)*x[0]+(-0.176557734776)*x[1]**o+(0.865146379005)*x[1]+(-0.304648040177)*x[2]**o+(0.269150163824)*x[2]
        arg[1,4,2,0]=(0.440212388104)*x[0]**o+(-0.590292628428)*x[0]+(-0.595359323633)*x[1]**o+(0.606734888384)*x[1]+(-0.951693103773)*x[2]**o+(-0.539741612207)*x[2]
        arg[1,4,2,1]=(0.346121727918)*x[0]**o+(0.188961570963)*x[0]+(0.402818736255)*x[1]**o+(-0.474029523974)*x[1]+(-0.95972237672)*x[2]**o+(0.983543070357)*x[2]
        arg[2,0,0,0]=(-0.801941188348)*x[0]**o+(-0.318215868754)*x[0]+(0.322031073259)*x[1]**o+(-0.00368402006147)*x[1]+(0.473846239338)*x[2]**o+(0.512752561753)*x[2]
        arg[2,0,0,1]=(0.786459471841)*x[0]**o+(0.941049841845)*x[0]+(0.064717660139)*x[1]**o+(-0.891518891268)*x[1]+(-0.317681565962)*x[2]**o+(-0.672310725168)*x[2]
        arg[2,0,1,0]=(0.813571093948)*x[0]**o+(-0.146551267883)*x[0]+(0.942106359835)*x[1]**o+(0.165543949281)*x[1]+(-0.0301736488709)*x[2]**o+(0.936915469225)*x[2]
        arg[2,0,1,1]=(-0.0298664722355)*x[0]**o+(-0.596644819792)*x[0]+(-0.306728572634)*x[1]**o+(0.770671782936)*x[1]+(-0.334606536003)*x[2]**o+(0.888940190159)*x[2]
        arg[2,0,2,0]=(0.688553055025)*x[0]**o+(-0.241541868482)*x[0]+(-0.694609526571)*x[1]**o+(-0.509896457784)*x[1]+(-0.0610272843414)*x[2]**o+(-0.251290389152)*x[2]
        arg[2,0,2,1]=(-0.959007707329)*x[0]**o+(-0.193856482901)*x[0]+(-0.138697126551)*x[1]**o+(0.768166305153)*x[1]+(0.951006645306)*x[2]**o+(-0.533986863953)*x[2]
        arg[2,1,0,0]=(0.648021743651)*x[0]**o+(-0.839074631201)*x[0]+(-0.445337118707)*x[1]**o+(0.997475278653)*x[1]+(-0.154475323496)*x[2]**o+(-0.45292992537)*x[2]
        arg[2,1,0,1]=(-0.134785884326)*x[0]**o+(-0.906738442209)*x[0]+(0.00509900741927)*x[1]**o+(0.489865232888)*x[1]+(0.620823018353)*x[2]**o+(-0.480127263607)*x[2]
        arg[2,1,1,0]=(-0.210371458935)*x[0]**o+(0.308453655681)*x[0]+(-0.510473100104)*x[1]**o+(-0.0776488834445)*x[1]+(0.58103002434)*x[2]**o+(0.039267622777)*x[2]
        arg[2,1,1,1]=(-0.124204683462)*x[0]**o+(0.429525935007)*x[0]+(0.183174600008)*x[1]**o+(-0.665914708196)*x[1]+(-0.576713549033)*x[2]**o+(-0.3788567811)*x[2]
        arg[2,1,2,0]=(-0.864065825608)*x[0]**o+(0.644046059464)*x[0]+(-0.265959280151)*x[1]**o+(0.989465634246)*x[1]+(0.536466956842)*x[2]**o+(0.689999246325)*x[2]
        arg[2,1,2,1]=(0.37622234855)*x[0]**o+(-0.905554688955)*x[0]+(-0.0258037833143)*x[1]**o+(-0.23415538846)*x[1]+(-0.64461795706)*x[2]**o+(0.262259040646)*x[2]
        arg[2,2,0,0]=(0.0541619627989)*x[0]**o+(-0.978918782799)*x[0]+(0.478333900479)*x[1]**o+(0.958652455667)*x[1]+(-0.746701804693)*x[2]**o+(-0.557405653804)*x[2]
        arg[2,2,0,1]=(0.611472347101)*x[0]**o+(-0.409754132769)*x[0]+(0.291274496706)*x[1]**o+(0.901983238042)*x[1]+(-0.102917678824)*x[2]**o+(-0.668205224505)*x[2]
        arg[2,2,1,0]=(-0.926062469523)*x[0]**o+(0.101032437783)*x[0]+(-0.385485330853)*x[1]**o+(0.0521756546126)*x[1]+(0.132320977758)*x[2]**o+(0.471619572767)*x[2]
        arg[2,2,1,1]=(0.0689111652346)*x[0]**o+(-0.517070173331)*x[0]+(-0.599184868344)*x[1]**o+(-0.814239996638)*x[1]+(0.97403158037)*x[2]**o+(-0.546843890524)*x[2]
        arg[2,2,2,0]=(-0.133251706696)*x[0]**o+(-0.0954736405212)*x[0]+(0.780584398494)*x[1]**o+(-0.505305860913)*x[1]+(-0.11518705364)*x[2]**o+(0.0855698333433)*x[2]
        arg[2,2,2,1]=(-0.567309977436)*x[0]**o+(-0.284182307402)*x[0]+(0.220302202269)*x[1]**o+(-0.252628588826)*x[1]+(-0.368390646272)*x[2]**o+(-0.577499260947)*x[2]
        arg[2,3,0,0]=(0.477968626266)*x[0]**o+(0.522764690241)*x[0]+(-0.0339165683557)*x[1]**o+(0.623897338329)*x[1]+(-0.169558925831)*x[2]**o+(-0.426663511213)*x[2]
        arg[2,3,0,1]=(0.185482762361)*x[0]**o+(0.496017413966)*x[0]+(-0.217189284009)*x[1]**o+(0.488205372581)*x[1]+(0.445762454794)*x[2]**o+(-0.423587587222)*x[2]
        arg[2,3,1,0]=(-0.489549524298)*x[0]**o+(-0.482299084307)*x[0]+(0.567729727254)*x[1]**o+(-0.341933025591)*x[1]+(-0.325429025641)*x[2]**o+(0.476030984261)*x[2]
        arg[2,3,1,1]=(-0.390754226646)*x[0]**o+(-0.980255652462)*x[0]+(0.109328516217)*x[1]**o+(-0.0141509875662)*x[1]+(0.486458644215)*x[2]**o+(-0.668018419371)*x[2]
        arg[2,3,2,0]=(0.352912006363)*x[0]**o+(-0.542969561227)*x[0]+(0.512067200296)*x[1]**o+(0.661110389562)*x[1]+(0.289744832231)*x[2]**o+(-0.176943484459)*x[2]
        arg[2,3,2,1]=(0.535262244127)*x[0]**o+(0.528325019892)*x[0]+(-0.106312276613)*x[1]**o+(-0.484575045858)*x[1]+(-0.407717587021)*x[2]**o+(-0.786366604025)*x[2]
        arg[2,4,0,0]=(-0.511771630262)*x[0]**o+(-0.590485876141)*x[0]+(0.493685168985)*x[1]**o+(-0.726700517739)*x[1]+(-0.213137223607)*x[2]**o+(-0.566923814015)*x[2]
        arg[2,4,0,1]=(0.421559088561)*x[0]**o+(0.838784944841)*x[0]+(-0.693654869402)*x[1]**o+(-0.0755493272975)*x[1]+(-0.174245799979)*x[2]**o+(0.242245946157)*x[2]
        arg[2,4,1,0]=(0.681235817735)*x[0]**o+(-0.313564199551)*x[0]+(0.663598935418)*x[1]**o+(-0.604327253895)*x[1]+(0.698985821444)*x[2]**o+(0.298899431146)*x[2]
        arg[2,4,1,1]=(0.432866305507)*x[0]**o+(-0.891501121013)*x[0]+(-0.290106178906)*x[1]**o+(0.268152357012)*x[1]+(0.75279498948)*x[2]**o+(0.211312203371)*x[2]
        arg[2,4,2,0]=(-0.956901622446)*x[0]**o+(0.742548241435)*x[0]+(-0.956209816095)*x[1]**o+(-0.539017080821)*x[1]+(-0.08062797007)*x[2]**o+(0.383843198102)*x[2]
        arg[2,4,2,1]=(-0.16700370288)*x[0]**o+(0.388837619298)*x[0]+(-0.190586726252)*x[1]**o+(0.916004535916)*x[1]+(0.310254387059)*x[2]**o+(-0.380371301717)*x[2]
        arg[3,0,0,0]=(0.64046380712)*x[0]**o+(0.254058760972)*x[0]+(-0.829050222873)*x[1]**o+(0.289942284927)*x[1]+(0.196268101069)*x[2]**o+(0.748880970446)*x[2]
        arg[3,0,0,1]=(-0.226057776579)*x[0]**o+(0.982685008224)*x[0]+(0.819680514589)*x[1]**o+(0.48005222662)*x[1]+(-0.615248069776)*x[2]**o+(-0.0923489923012)*x[2]
        arg[3,0,1,0]=(-0.358031084019)*x[0]**o+(-0.83941518843)*x[0]+(-0.0687218422525)*x[1]**o+(-0.648976147029)*x[1]+(0.711009285701)*x[2]**o+(0.360758898243)*x[2]
        arg[3,0,1,1]=(-0.312175299678)*x[0]**o+(0.360877864123)*x[0]+(0.653283960509)*x[1]**o+(0.323527948404)*x[1]+(-0.278454805659)*x[2]**o+(0.0585194127844)*x[2]
        arg[3,0,2,0]=(0.367207247645)*x[0]**o+(-0.0743547693027)*x[0]+(-0.473358544608)*x[1]**o+(0.332985041128)*x[1]+(-0.899612971909)*x[2]**o+(0.69409768473)*x[2]
        arg[3,0,2,1]=(-0.806787228991)*x[0]**o+(-0.110344204889)*x[0]+(-0.631610779885)*x[1]**o+(0.0239853435419)*x[1]+(0.83036063388)*x[2]**o+(0.623825256849)*x[2]
        arg[3,1,0,0]=(-0.268344201624)*x[0]**o+(-0.59230799177)*x[0]+(0.493719194706)*x[1]**o+(-0.128351291806)*x[1]+(0.0584636422077)*x[2]**o+(0.0859725996454)*x[2]
        arg[3,1,0,1]=(-0.870410307785)*x[0]**o+(-0.917370176071)*x[0]+(0.395701966434)*x[1]**o+(0.694856510628)*x[1]+(-0.748440730513)*x[2]**o+(0.748465450614)*x[2]
        arg[3,1,1,0]=(-0.178465888488)*x[0]**o+(-0.956884701156)*x[0]+(0.0472790389286)*x[1]**o+(0.946702274722)*x[1]+(-0.200668750878)*x[2]**o+(-0.0104023894447)*x[2]
        arg[3,1,1,1]=(0.113559334778)*x[0]**o+(0.270505514561)*x[0]+(-0.238008799138)*x[1]**o+(0.533067625871)*x[1]+(-0.586814102301)*x[2]**o+(0.045657074859)*x[2]
        arg[3,1,2,0]=(-0.574716699633)*x[0]**o+(-0.659499274639)*x[0]+(0.892605757655)*x[1]**o+(-0.39697940457)*x[1]+(0.363011145793)*x[2]**o+(-0.364722526091)*x[2]
        arg[3,1,2,1]=(-0.0193167048813)*x[0]**o+(0.761104649925)*x[0]+(0.490164183635)*x[1]**o+(0.10693267057)*x[1]+(0.635328989408)*x[2]**o+(-0.353114187925)*x[2]
        arg[3,2,0,0]=(0.0777423524941)*x[0]**o+(-0.431011355257)*x[0]+(-0.356762042134)*x[1]**o+(-0.94361679144)*x[1]+(-0.716191700456)*x[2]**o+(0.49906825221)*x[2]
        arg[3,2,0,1]=(0.80859948321)*x[0]**o+(0.542680788872)*x[0]+(-0.626897369633)*x[1]**o+(-0.821877255135)*x[1]+(-0.455575534023)*x[2]**o+(0.317480554025)*x[2]
        arg[3,2,1,0]=(-0.0144328272818)*x[0]**o+(-0.395367848628)*x[0]+(-0.601042089925)*x[1]**o+(-0.0822921879892)*x[1]+(0.896199761801)*x[2]**o+(-0.215055215166)*x[2]
        arg[3,2,1,1]=(0.377945289524)*x[0]**o+(-0.580064012583)*x[0]+(0.993926713349)*x[1]**o+(-0.736307101943)*x[1]+(-0.939628422411)*x[2]**o+(-0.923219473393)*x[2]
        arg[3,2,2,0]=(-0.733150287394)*x[0]**o+(0.692501827859)*x[0]+(0.904655347343)*x[1]**o+(0.310386512564)*x[1]+(-0.93795078156)*x[2]**o+(0.772534565721)*x[2]
        arg[3,2,2,1]=(-0.155229801255)*x[0]**o+(0.580356480314)*x[0]+(0.0673352927499)*x[1]**o+(-0.274975805243)*x[1]+(0.553818209037)*x[2]**o+(-0.835142674988)*x[2]
        arg[3,3,0,0]=(0.60752130978)*x[0]**o+(0.847750395856)*x[0]+(0.8635279539)*x[1]**o+(0.16341072973)*x[1]+(0.470011948821)*x[2]**o+(0.41529971393)*x[2]
        arg[3,3,0,1]=(0.755049219363)*x[0]**o+(0.66080573214)*x[0]+(-0.0256982576552)*x[1]**o+(-0.0311937109251)*x[1]+(-0.028843413794)*x[2]**o+(-0.19345503082)*x[2]
        arg[3,3,1,0]=(-0.39765237082)*x[0]**o+(-0.746359385218)*x[0]+(0.275396324626)*x[1]**o+(0.064571659107)*x[1]+(0.771289010752)*x[2]**o+(-6.69891594269e-05)*x[2]
        arg[3,3,1,1]=(0.350497253938)*x[0]**o+(-0.236419516561)*x[0]+(0.673483791847)*x[1]**o+(-0.934644138115)*x[1]+(0.994408125332)*x[2]**o+(-0.398423596854)*x[2]
        arg[3,3,2,0]=(-0.952544750261)*x[0]**o+(0.24256752245)*x[0]+(0.376782843601)*x[1]**o+(-0.349840168156)*x[1]+(-0.260325141903)*x[2]**o+(-0.452866344447)*x[2]
        arg[3,3,2,1]=(-0.0232941915849)*x[0]**o+(0.191625715695)*x[0]+(-0.490187753329)*x[1]**o+(-0.823976612466)*x[1]+(0.476804203752)*x[2]**o+(0.61057521184)*x[2]
        arg[3,4,0,0]=(0.0279130167137)*x[0]**o+(-0.134953133648)*x[0]+(0.0821614836141)*x[1]**o+(-0.894046763697)*x[1]+(-0.497197221983)*x[2]**o+(0.586297244251)*x[2]
        arg[3,4,0,1]=(0.0423746259802)*x[0]**o+(0.458075512165)*x[0]+(0.855157549678)*x[1]**o+(-0.607654212682)*x[1]+(-0.672693159955)*x[2]**o+(-0.300116892313)*x[2]
        arg[3,4,1,0]=(-0.695685046398)*x[0]**o+(-0.709807910289)*x[0]+(-0.167879145281)*x[1]**o+(0.524084329764)*x[1]+(0.763706690019)*x[2]**o+(-0.219502671318)*x[2]
        arg[3,4,1,1]=(0.864599640696)*x[0]**o+(-0.341125612414)*x[0]+(-0.548888211648)*x[1]**o+(0.918497762429)*x[1]+(-0.445600024463)*x[2]**o+(0.221658300325)*x[2]
        arg[3,4,2,0]=(-0.893419673448)*x[0]**o+(0.653718653341)*x[0]+(0.323868832441)*x[1]**o+(0.905933530129)*x[1]+(0.62045953406)*x[2]**o+(0.145474797235)*x[2]
        arg[3,4,2,1]=(0.480643160777)*x[0]**o+(-0.15896921521)*x[0]+(-0.329564714008)*x[1]**o+(-0.613303857496)*x[1]+(-0.483827450625)*x[2]**o+(-0.91023336048)*x[2]
        ref[0,0,0,0]=(0.520702135089)/(o+1.)+(0.109542934387)+(0.816458625566)*0.5**o
        ref[0,0,0,1]=(0.188518886272)/(o+1.)+(-0.938691737334)+(-0.709065020122)*0.5**o
        ref[0,0,1,0]=(-0.757157441618)/(o+1.)+(0.316696781197)+(-0.09165636265)*0.5**o
        ref[0,0,1,1]=(0.325937013197)/(o+1.)+(0.0450843195409)+(0.900670566988)*0.5**o
        ref[0,0,2,0]=(1.13837105734)/(o+1.)+(-0.573151939969)+(-0.883617789072)*0.5**o
        ref[0,0,2,1]=(-0.665166165764)/(o+1.)+(-0.845347549645)+(-0.770041366552)*0.5**o
        ref[0,1,0,0]=(0.714049217035)/(o+1.)+(0.4125480263)+(0.90396926296)*0.5**o
        ref[0,1,0,1]=(0.429723443697)/(o+1.)+(0.255965487666)+(0.47956310963)*0.5**o
        ref[0,1,1,0]=(-0.282787702479)/(o+1.)+(-0.631122483908)+(-0.75177304495)*0.5**o
        ref[0,1,1,1]=(1.16234435068)/(o+1.)+(1.09548265929)+(-0.109937058453)*0.5**o
        ref[0,1,2,0]=(-0.619187813423)/(o+1.)+(0.47035748919)+(-0.13801307581)*0.5**o
        ref[0,1,2,1]=(0.641297345823)/(o+1.)+(0.937805735461)+(-0.518540484981)*0.5**o
        ref[0,2,0,0]=(1.36665521497)/(o+1.)+(-0.736611486382)+(0.286922773164)*0.5**o
        ref[0,2,0,1]=(-0.27582202139)/(o+1.)+(-0.312977125741)+(-0.116447648923)*0.5**o
        ref[0,2,1,0]=(0.294494393603)/(o+1.)+(0.117678635379)+(-0.322803067618)*0.5**o
        ref[0,2,1,1]=(-0.251319095061)/(o+1.)+(0.517336946518)+(0.997823909059)*0.5**o
        ref[0,2,2,0]=(-1.06239756943)/(o+1.)+(-0.174439303934)+(0.516866019209)*0.5**o
        ref[0,2,2,1]=(-0.114597419557)/(o+1.)+(-0.488263901705)+(0.777538237337)*0.5**o
        ref[0,3,0,0]=(-0.545872917587)/(o+1.)+(0.320972405109)+(-0.622067362782)*0.5**o
        ref[0,3,0,1]=(-1.41803804504)/(o+1.)+(0.326673034142)+(-0.49698112241)*0.5**o
        ref[0,3,1,0]=(0.220871698725)/(o+1.)+(0.352475535909)+(0.917091542097)*0.5**o
        ref[0,3,1,1]=(-1.82481767225)/(o+1.)+(-0.534598958622)+(-0.437188100982)*0.5**o
        ref[0,3,2,0]=(1.04475880275)/(o+1.)+(0.00788693262039)+(-0.841265485658)*0.5**o
        ref[0,3,2,1]=(-0.262966085934)/(o+1.)+(0.330079558014)+(0.383282548921)*0.5**o
        ref[0,4,0,0]=(-1.66865494548)/(o+1.)+(-0.567436682752)+(0.0385891845805)*0.5**o
        ref[0,4,0,1]=(-0.183458096784)/(o+1.)+(-1.1357723338)+(-0.387739686761)*0.5**o
        ref[0,4,1,0]=(0.611419775289)/(o+1.)+(0.0492166712458)+(-0.329464483421)*0.5**o
        ref[0,4,1,1]=(0.000995600207189)/(o+1.)+(-0.182953786649)+(0.63272559561)*0.5**o
        ref[0,4,2,0]=(0.630523648936)/(o+1.)+(-0.0318688096074)+(-0.847582850604)*0.5**o
        ref[0,4,2,1]=(0.306969678703)/(o+1.)+(-0.726991015938)+(-0.722990535113)*0.5**o
        ref[1,0,0,0]=(-0.406383597128)/(o+1.)+(-0.580550741796)+(0.0725951752952)*0.5**o
        ref[1,0,0,1]=(0.988153215411)/(o+1.)+(-0.295391144874)+(-0.254920010173)*0.5**o
        ref[1,0,1,0]=(-0.113985267485)/(o+1.)+(0.381963273358)+(0.486342223622)*0.5**o
        ref[1,0,1,1]=(-1.0749658001)/(o+1.)+(-0.980590626052)+(0.777015150638)*0.5**o
        ref[1,0,2,0]=(0.866170818208)/(o+1.)+(0.295155963167)+(0.240813275278)*0.5**o
        ref[1,0,2,1]=(-0.502098016236)/(o+1.)+(-0.404695777523)+(-0.670257601629)*0.5**o
        ref[1,1,0,0]=(-0.0930803554155)/(o+1.)+(-0.262152356852)+(0.985895581872)*0.5**o
        ref[1,1,0,1]=(-1.66036439576)/(o+1.)+(0.188095420862)+(-0.954132399275)*0.5**o
        ref[1,1,1,0]=(-0.548936659502)/(o+1.)+(0.974895418354)+(0.462810965625)*0.5**o
        ref[1,1,1,1]=(-1.80963059517)/(o+1.)+(-0.200285193041)+(0.645629353541)*0.5**o
        ref[1,1,2,0]=(1.41739085898)/(o+1.)+(0.549897915219)+(-0.554993653726)*0.5**o
        ref[1,1,2,1]=(0.0907943154278)/(o+1.)+(-0.450977439484)+(0.1476587556)*0.5**o
        ref[1,2,0,0]=(-0.657176681491)/(o+1.)+(-0.433402993059)+(-0.715887407694)*0.5**o
        ref[1,2,0,1]=(-0.562421731301)/(o+1.)+(0.0104884758643)+(0.48440337855)*0.5**o
        ref[1,2,1,0]=(1.28216615223)/(o+1.)+(-0.33142121119)+(0.621562227153)*0.5**o
        ref[1,2,1,1]=(1.16073972701)/(o+1.)+(0.0404640630801)+(0.472846565983)*0.5**o
        ref[1,2,2,0]=(0.772147142645)/(o+1.)+(1.09472486224)+(-0.212000572333)*0.5**o
        ref[1,2,2,1]=(0.811972306738)/(o+1.)+(-0.409837616651)+(-0.632752637148)*0.5**o
        ref[1,3,0,0]=(0.754125096738)/(o+1.)+(0.000627424214716)+(0.390081216336)*0.5**o
        ref[1,3,0,1]=(0.213094712846)/(o+1.)+(-0.73677243263)+(-0.402887151891)*0.5**o
        ref[1,3,1,0]=(-0.0409011142831)/(o+1.)+(0.288589213809)+(0.253947492129)*0.5**o
        ref[1,3,1,1]=(0.519205654314)/(o+1.)+(-0.416997867426)+(-0.435656267291)*0.5**o
        ref[1,3,2,0]=(0.0264169632513)/(o+1.)+(0.271272227399)+(-0.069319863146)*0.5**o
        ref[1,3,2,1]=(-0.845761130374)/(o+1.)+(1.07136053657)+(0.741825647399)*0.5**o
        ref[1,4,0,0]=(0.967780841131)/(o+1.)+(-0.0706409792865)+(-0.66003338591)*0.5**o
        ref[1,4,0,1]=(-1.02543406666)/(o+1.)+(-0.81933006071)+(-0.0001759210396)*0.5**o
        ref[1,4,1,0]=(0.121210472128)/(o+1.)+(0.718839287394)+(0.052435899043)*0.5**o
        ref[1,4,1,1]=(-0.481205774953)/(o+1.)+(0.50624453832)+(0.00806522196579)*0.5**o
        ref[1,4,2,0]=(-1.54705242741)/(o+1.)+(-0.261649676126)+(0.440212388104)*0.5**o
        ref[1,4,2,1]=(-0.556903640465)/(o+1.)+(0.349237558673)+(0.346121727918)*0.5**o
        ref[2,0,0,0]=(0.795877312597)/(o+1.)+(0.0954263364687)+(-0.801941188348)*0.5**o
        ref[2,0,0,1]=(-0.252963905823)/(o+1.)+(-0.311389887295)+(0.786459471841)*0.5**o
        ref[2,0,1,0]=(0.911932710964)/(o+1.)+(0.477954075311)+(0.813571093948)*0.5**o
        ref[2,0,1,1]=(-0.641335108636)/(o+1.)+(0.531483576652)+(-0.0298664722355)*0.5**o
        ref[2,0,2,0]=(-0.755636810913)/(o+1.)+(-0.501364357709)+(0.688553055025)*0.5**o
        ref[2,0,2,1]=(0.812309518755)/(o+1.)+(0.0201614791497)+(-0.959007707329)*0.5**o
        ref[2,1,0,0]=(-0.599812442203)/(o+1.)+(-0.147264638959)+(0.648021743651)*0.5**o
        ref[2,1,0,1]=(0.625922025772)/(o+1.)+(-0.448500236464)+(-0.134785884326)*0.5**o
        ref[2,1,1,0]=(0.0705569242367)/(o+1.)+(0.135036197507)+(-0.210371458935)*0.5**o
        ref[2,1,1,1]=(-0.393538949024)/(o+1.)+(-0.307622777145)+(-0.124204683462)*0.5**o
        ref[2,1,2,0]=(0.270507676691)/(o+1.)+(1.16175547002)+(-0.864065825608)*0.5**o
        ref[2,1,2,1]=(-0.670421740374)/(o+1.)+(-0.438725518385)+(0.37622234855)*0.5**o
        ref[2,2,0,0]=(-0.268367904213)/(o+1.)+(-0.288835990468)+(0.0541619627989)*0.5**o
        ref[2,2,0,1]=(0.188356817882)/(o+1.)+(-0.0879880596161)+(0.611472347101)*0.5**o
        ref[2,2,1,0]=(-0.253164353095)/(o+1.)+(0.312413832582)+(-0.926062469523)*0.5**o
        ref[2,2,1,1]=(0.374846712027)/(o+1.)+(-0.939077030246)+(0.0689111652346)*0.5**o
        ref[2,2,2,0]=(0.665397344854)/(o+1.)+(-0.257604834045)+(-0.133251706696)*0.5**o
        ref[2,2,2,1]=(-0.148088444003)/(o+1.)+(-0.557155078588)+(-0.567309977436)*0.5**o
        ref[2,3,0,0]=(-0.203475494187)/(o+1.)+(0.359999258678)+(0.477968626266)*0.5**o
        ref[2,3,0,1]=(0.228573170785)/(o+1.)+(0.280317599662)+(0.185482762361)*0.5**o
        ref[2,3,1,0]=(0.242300701614)/(o+1.)+(-0.174100562819)+(-0.489549524298)*0.5**o
        ref[2,3,1,1]=(0.595787160431)/(o+1.)+(-0.8312125297)+(-0.390754226646)*0.5**o
        ref[2,3,2,0]=(0.801812032527)/(o+1.)+(-0.0294013280621)+(0.352912006363)*0.5**o
        ref[2,3,2,1]=(-0.514029863634)/(o+1.)+(-0.371308314996)+(0.535262244127)*0.5**o
        ref[2,4,0,0]=(0.280547945378)/(o+1.)+(-0.942055103948)+(-0.511771630262)*0.5**o
        ref[2,4,0,1]=(-0.867900669381)/(o+1.)+(0.50274078185)+(0.421559088561)*0.5**o
        ref[2,4,1,0]=(1.36258475686)/(o+1.)+(-0.30949601115)+(0.681235817735)*0.5**o
        ref[2,4,1,1]=(0.462688810574)/(o+1.)+(-0.206018280315)+(0.432866305507)*0.5**o
        ref[2,4,2,0]=(-1.03683778617)/(o+1.)+(0.293687179358)+(-0.956901622446)*0.5**o
        ref[2,4,2,1]=(0.119667660807)/(o+1.)+(0.462235426749)+(-0.16700370288)*0.5**o
        ref[3,0,0,0]=(-0.632782121804)/(o+1.)+(0.646441008172)+(0.64046380712)*0.5**o
        ref[3,0,0,1]=(0.204432444813)/(o+1.)+(0.685194121272)+(-0.226057776579)*0.5**o
        ref[3,0,1,0]=(0.642287443449)/(o+1.)+(-0.563816218608)+(-0.358031084019)*0.5**o
        ref[3,0,1,1]=(0.37482915485)/(o+1.)+(0.371462612656)+(-0.312175299678)*0.5**o
        ref[3,0,2,0]=(-1.37297151652)/(o+1.)+(0.476363978278)+(0.367207247645)*0.5**o
        ref[3,0,2,1]=(0.198749853995)/(o+1.)+(0.268733197751)+(-0.806787228991)*0.5**o
        ref[3,1,0,0]=(0.552182836914)/(o+1.)+(-0.317343341965)+(-0.268344201624)*0.5**o
        ref[3,1,0,1]=(-0.352738764079)/(o+1.)+(0.262975892585)+(-0.870410307785)*0.5**o
        ref[3,1,1,0]=(-0.153389711949)/(o+1.)+(-0.0102924079393)+(-0.178465888488)*0.5**o
        ref[3,1,1,1]=(-0.824822901439)/(o+1.)+(0.424615107645)+(0.113559334778)*0.5**o
        ref[3,1,2,0]=(1.25561690345)/(o+1.)+(-0.71060060265)+(-0.574716699633)*0.5**o
        ref[3,1,2,1]=(1.12549317304)/(o+1.)+(0.257461566285)+(-0.0193167048813)*0.5**o
        ref[3,2,0,0]=(-1.07295374259)/(o+1.)+(-0.437779947243)+(0.0777423524941)*0.5**o
        ref[3,2,0,1]=(-1.08247290366)/(o+1.)+(0.019142043881)+(0.80859948321)*0.5**o
        ref[3,2,1,0]=(0.295157671875)/(o+1.)+(-0.346357625892)+(-0.0144328272818)*0.5**o
        ref[3,2,1,1]=(0.054298290938)/(o+1.)+(-1.11979529396)+(0.377945289524)*0.5**o
        ref[3,2,2,0]=(-0.0332954342171)/(o+1.)+(0.887711453071)+(-0.733150287394)*0.5**o
        ref[3,2,2,1]=(0.621153501787)/(o+1.)+(-0.264880999958)+(-0.155229801255)*0.5**o
        ref[3,3,0,0]=(1.33353990272)/(o+1.)+(0.713230419758)+(0.60752130978)*0.5**o
        ref[3,3,0,1]=(-0.0545416714491)/(o+1.)+(0.218078495198)+(0.755049219363)*0.5**o
        ref[3,3,1,0]=(1.04668533538)/(o+1.)+(-0.340927357635)+(-0.39765237082)*0.5**o
        ref[3,3,1,1]=(1.66789191718)/(o+1.)+(-0.784743625765)+(0.350497253938)*0.5**o
        ref[3,3,2,0]=(0.116457701698)/(o+1.)+(-0.280069495076)+(-0.952544750261)*0.5**o
        ref[3,3,2,1]=(-0.0133835495768)/(o+1.)+(-0.0108878424651)+(-0.0232941915849)*0.5**o
        ref[3,4,0,0]=(-0.415035738369)/(o+1.)+(-0.221351326547)+(0.0279130167137)*0.5**o
        ref[3,4,0,1]=(0.182464389723)/(o+1.)+(-0.224847796415)+(0.0423746259802)*0.5**o
        ref[3,4,1,0]=(0.595827544738)/(o+1.)+(-0.202613125922)+(-0.695685046398)*0.5**o
        ref[3,4,1,1]=(-0.99448823611)/(o+1.)+(0.39951522517)+(0.864599640696)*0.5**o
        ref[3,4,2,0]=(0.944328366501)/(o+1.)+(0.852563490353)+(-0.893419673448)*0.5**o
        ref[3,4,2,1]=(-0.813392164634)/(o+1.)+(-0.841253216593)+(0.480643160777)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.241972460279)*x[0]+(0.539729925673)*x[1]
        ref=(-0.11832904353)/(o+1.)+(0.649911364492)+(-0.399791299503)*0.5**o
      else:
        arg=(0.53919797656)*x[0]+(0.379010664884)*x[1]+(0.274416064529)*x[2]
        ref=(0.307947301472)/(o+1.)+(0.120674157748)+(0.643329089005)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.84849017056)*x[0]+(-0.161962845202)*x[1]
        arg[1]=(-0.794717919135)*x[0]+(1.16003773358)*x[1]
        ref[0]=(0.780514448202)/(o+1.)+(-0.34628958021)+(0.598592037576)*0.5**o
        ref[1]=(0.570158156278)/(o+1.)+(0.241192316704)+(-0.687222975237)*0.5**o
      else:
        arg[0]=(-0.637352163634)*x[0]+(1.49477462643)*x[1]+(-0.284018740238)*x[2]
        arg[1]=(0.125937459904)*x[0]+(-0.206499869891)*x[1]+(-0.342856240345)*x[2]
        ref[0]=(-0.274486314451)/(o+1.)+(0.379737553853)+(0.0884149293071)*0.5**o
        ref[1]=(-0.78452814021)/(o+1.)+(-0.122017410137)+(0.605144310151)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.440868762581)*x[0]+(-0.216223666616)*x[1]
        arg[0,1]=(-1.78807355591)*x[0]+(-0.208278123838)*x[1]
        arg[0,2]=(-0.448066998103)*x[0]+(-0.819815507826)*x[1]
        arg[0,3]=(1.27247881689)*x[0]+(0.0653439567221)*x[1]
        arg[0,4]=(-1.00344990223)*x[0]+(-0.291263790583)*x[1]
        arg[1,0]=(0.440080423838)*x[0]+(0.0356836558113)*x[1]
        arg[1,1]=(-1.49545340091)*x[0]+(-0.133615590211)*x[1]
        arg[1,2]=(1.05947638385)*x[0]+(0.408685904238)*x[1]
        arg[1,3]=(-0.170309086605)*x[0]+(0.660523970285)*x[1]
        arg[1,4]=(0.764323678475)*x[0]+(-0.529415080303)*x[1]
        arg[2,0]=(0.55270081325)*x[0]+(-1.60630146231)*x[1]
        arg[2,1]=(1.57419330248)*x[0]+(-1.24644288706)*x[1]
        arg[2,2]=(0.0416757238363)*x[0]+(0.0247379738465)*x[1]
        arg[2,3]=(1.0200370393)*x[0]+(0.364209137693)*x[1]
        arg[2,4]=(-1.21858490389)*x[0]+(0.523056254546)*x[1]
        arg[3,0]=(0.0524205820842)*x[0]+(1.65953217854)*x[1]
        arg[3,1]=(1.21824883395)*x[0]+(0.905190834998)*x[1]
        arg[3,2]=(0.226344434525)*x[0]+(-0.735349531462)*x[1]
        arg[3,3]=(-1.00295245717)*x[0]+(-0.719605376499)*x[1]
        arg[3,4]=(-0.446228759693)*x[0]+(1.23441974277)*x[1]
        ref[0,0]=(0.347328582602)/(o+1.)+(-0.520484463519)+(0.0365479152384)*0.5**o
        ref[0,1]=(-0.274564442352)/(o+1.)+(-0.413557210678)+(-0.894672816039)*0.5**o
        ref[0,2]=(0.0720644688954)/(o+1.)+(-0.906636966188)+(0.473326957552)*0.5**o
        ref[0,3]=(-0.747749341777)/(o+1.)+(0.730249972378)+(0.625072170633)*0.5**o
        ref[0,4]=(-0.194031116169)/(o+1.)+(-0.432063325829)+(-0.236555924983)*0.5**o
        ref[1,0]=(0.807081538033)/(o+1.)+(-0.612266650007)+(0.893215841629)*0.5**o
        ref[1,1]=(-0.662516518957)/(o+1.)+(-0.176318925886)+(-0.61391462039)*0.5**o
        ref[1,2]=(-0.570345636156)/(o+1.)+(0.791877954963)+(0.454752014319)*0.5**o
        ref[1,3]=(-0.312400910795)/(o+1.)+(0.212935137466)+(0.376745519544)*0.5**o
        ref[1,4]=(-0.764680414264)/(o+1.)+(0.174215486579)+(0.651158039278)*0.5**o
        ref[2,0]=(-0.837942771642)/(o+1.)+(-0.187938486702)+(0.160219095985)*0.5**o
        ref[2,1]=(-0.672421039761)/(o+1.)+(0.17384002882)+(0.652491397549)*0.5**o
        ref[2,2]=(-0.656187949922)/(o+1.)+(0.559888736121)+(-0.397175824637)*0.5**o
        ref[2,3]=(0.259558368195)/(o+1.)+(0.55148644978)+(0.0217149092406)*0.5**o
        ref[2,4]=(-0.188397537901)/(o+1.)+(-0.0561706170512)+(-0.394789877345)*0.5**o
        ref[3,0]=(0.91055628617)/(o+1.)+(0.574211274098)+(-0.347026073739)*0.5**o
        ref[3,1]=(0.29225509237)/(o+1.)+(0.64447705482)+(0.542230466941)*0.5**o
        ref[3,2]=(0.121065624384)/(o+1.)+(0.0316495277836)+(-0.693369776888)*0.5**o
        ref[3,3]=(-0.943523334976)/(o+1.)+(-0.266302644031)+(-0.246429210629)*0.5**o
        ref[3,4]=(0.714915399455)/(o+1.)+(0.459963131994)+(-0.846650680369)*0.5**o
      else:
        arg[0,0]=(-0.493869713454)*x[0]+(-0.381969413515)*x[1]+(1.59862368848)*x[2]
        arg[0,1]=(-0.581245543547)*x[0]+(0.961117271889)*x[1]+(-0.158073515343)*x[2]
        arg[0,2]=(-0.635283143641)*x[0]+(0.695573236987)*x[1]+(-0.719180832278)*x[2]
        arg[0,3]=(-0.737780596733)*x[0]+(-1.11171283352)*x[1]+(0.459954346315)*x[2]
        arg[0,4]=(1.12522065074)*x[0]+(-1.08427904854)*x[1]+(-0.715679673556)*x[2]
        arg[1,0]=(0.972468186181)*x[0]+(-0.211014530435)*x[1]+(-0.734414099002)*x[2]
        arg[1,1]=(-1.67337704748)*x[0]+(0.871096213089)*x[1]+(1.51063313223)*x[2]
        arg[1,2]=(-0.222013007834)*x[0]+(1.46474508326)*x[1]+(-0.979243908862)*x[2]
        arg[1,3]=(1.21195281087)*x[0]+(0.274082389329)*x[1]+(1.56144436465)*x[2]
        arg[1,4]=(1.35812865884)*x[0]+(1.25819798853)*x[1]+(0.895334048395)*x[2]
        arg[2,0]=(-0.167404344437)*x[0]+(0.0710591934315)*x[1]+(-0.124469683665)*x[2]
        arg[2,1]=(-0.492240632996)*x[0]+(-0.317251117382)*x[1]+(0.0195461703529)*x[2]
        arg[2,2]=(-0.277590266225)*x[0]+(-1.65859476899)*x[1]+(-0.0100216405668)*x[2]
        arg[2,3]=(-0.0701281298455)*x[0]+(0.502137254703)*x[1]+(0.701220923443)*x[2]
        arg[2,4]=(-0.72186096407)*x[0]+(-0.900487568179)*x[1]+(-0.735573922826)*x[2]
        arg[3,0]=(-0.0677324474129)*x[0]+(-0.649963209718)*x[1]+(-0.770861099726)*x[2]
        arg[3,1]=(0.563588189481)*x[0]+(0.714614932251)*x[1]+(-0.0957764632886)*x[2]
        arg[3,2]=(1.13337395062)*x[0]+(0.507556316073)*x[1]+(-0.196759085949)*x[2]
        arg[3,3]=(-1.2636883168)*x[0]+(-0.811660344076)*x[1]+(0.645462052598)*x[2]
        arg[3,4]=(-0.104175186987)*x[0]+(-0.929176038155)*x[1]+(1.07530379856)*x[2]
        ref[0,0]=(0.84538674973)/(o+1.)+(0.0569713556632)+(-0.236544899547)*0.5**o
        ref[0,1]=(-0.307689589699)/(o+1.)+(0.245040337642)+(0.0394071274131)*0.5**o
        ref[0,2]=(0.955244736357)/(o+1.)+(-0.341771136504)+(-0.930593202281)*0.5**o
        ref[0,3]=(-0.353490461722)/(o+1.)+(-0.476228433901)+(-0.083591754415)*0.5**o
        ref[0,4]=(-1.31220485482)/(o+1.)+(0.0466426656676)+(0.544181452128)*0.5**o
        ref[1,0]=(0.22773475414)/(o+1.)+(-0.488454801654)+(0.776214405914)*0.5**o
        ref[1,1]=(1.50688678707)/(o+1.)+(0.0851415404695)+(-0.96881757017)*0.5**o
        ref[1,2]=(0.313471097102)/(o+1.)+(-0.0875594518156)+(0.125135973096)*0.5**o
        ref[1,3]=(0.269521220179)/(o+1.)+(1.1192007195)+(0.539556905661)*0.5**o
        ref[1,4]=(1.91499876064)/(o+1.)+(0.509063797893)+(0.578534339343)*0.5**o
        ref[2,0]=(0.941352651562)/(o+1.)+(-0.966199402186)+(0.770231318139)*0.5**o
        ref[2,1]=(-0.141490823562)/(o+1.)+(-0.20919987022)+(-0.230055016023)*0.5**o
        ref[2,2]=(-0.916148047971)/(o+1.)+(-0.371740600192)+(-0.286577427426)*0.5**o
        ref[2,3]=(0.229516219331)/(o+1.)+(0.274790532736)+(0.354132763498)*0.5**o
        ref[2,4]=(-1.19120612619)/(o+1.)+(-0.139859859385)+(-0.88699661011)*0.5**o
        ref[3,0]=(-1.48630121465)/(o+1.)+(0.0744166251063)+(-0.151088792417)*0.5**o
        ref[3,1]=(-0.335583560513)/(o+1.)+(0.277436410841)+(0.963137397276)*0.5**o
        ref[3,2]=(-0.673942688644)/(o+1.)+(0.70077509905)+(0.716563671283)*0.5**o
        ref[3,3]=(-0.148394092795)/(o+1.)+(-0.237190239072)+(-0.807112037336)*0.5**o
        ref[3,4]=(0.529983886211)/(o+1.)+(-0.371055083588)+(0.254078854382)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.661725715165)*x[0]+(0.219720717581)*x[1]
        arg[0,0,1]=(-0.642240082304)*x[0]+(-0.279652756776)*x[1]
        arg[0,1,0]=(0.0204235813151)*x[0]+(-0.374873249933)*x[1]
        arg[0,1,1]=(1.4418481252)*x[0]+(0.205656834772)*x[1]
        arg[1,0,0]=(0.65992179951)*x[0]+(-1.04887875458)*x[1]
        arg[1,0,1]=(-0.195480216199)*x[0]+(-0.373905275039)*x[1]
        arg[1,1,0]=(0.781251543352)*x[0]+(0.0416248109536)*x[1]
        arg[1,1,1]=(-0.925751319189)*x[0]+(-0.113992910046)*x[1]
        arg[2,0,0]=(-0.0767414065227)*x[0]+(-0.156682840973)*x[1]
        arg[2,0,1]=(0.688457272529)*x[0]+(0.524916336046)*x[1]
        arg[2,1,0]=(0.753722830084)*x[0]+(-0.890325839466)*x[1]
        arg[2,1,1]=(0.730789290145)*x[0]+(-0.278986703716)*x[1]
        arg[3,0,0]=(-0.287376796439)*x[0]+(-0.153987233224)*x[1]
        arg[3,0,1]=(-1.15687435255)*x[0]+(0.250005890238)*x[1]
        arg[3,1,0]=(-0.845312830629)*x[0]+(0.256211962352)*x[1]
        arg[3,1,1]=(-0.574311820519)*x[0]+(-0.440297035998)*x[1]
        arg[4,0,0]=(-0.360607647308)*x[0]+(-1.29380437557)*x[1]
        arg[4,0,1]=(1.32019433485)*x[0]+(0.295255956618)*x[1]
        arg[4,1,0]=(-1.64162941972)*x[0]+(0.117482372162)*x[1]
        arg[4,1,1]=(-0.303103646619)*x[0]+(-0.88458982681)*x[1]
        arg[5,0,0]=(-0.666761879038)*x[0]+(-0.735786278877)*x[1]
        arg[5,0,1]=(0.0892896382617)*x[0]+(-0.45478872773)*x[1]
        arg[5,1,0]=(-0.50985624415)*x[0]+(-0.400022939567)*x[1]
        arg[5,1,1]=(1.39240706368)*x[0]+(0.0149740822664)*x[1]
        ref[0,0,0]=(-0.114834538488)/(o+1.)+(0.240621006484)+(0.515038958264)*0.5**o
        ref[0,0,1]=(0.588698337077)/(o+1.)+(-0.702256377233)+(-0.10607842169)*0.5**o
        ref[0,1,0]=(0.13849818224)/(o+1.)+(-0.277555277991)+(0.0621627051247)*0.5**o
        ref[0,1,1]=(-0.0878009521163)/(o+1.)+(0.630168348947)+(0.474969214192)*0.5**o
        ref[1,0,0]=(-0.696600538111)/(o+1.)+(-0.343318260912)+(0.994280104861)*0.5**o
        ref[1,0,1]=(-0.0887026688527)/(o+1.)+(-0.385753227979)+(0.290823633572)*0.5**o
        ref[1,1,0]=(0.831408608956)/(o+1.)+(-0.242985671853)+(0.477439089057)*0.5**o
        ref[1,1,1]=(-0.621059053191)/(o+1.)+(-0.0700203481241)+(-0.278644479797)*0.5**o
        ref[2,0,0]=(0.7890720399)/(o+1.)+(-0.539721196393)+(0.0569461053905)*0.5**o
        ref[2,0,1]=(-0.228628151401)/(o+1.)+(0.826946074171)+(-0.211890388366)*0.5**o
        ref[2,1,0]=(-0.837617325493)/(o+1.)+(0.289656653751)+(0.121701008608)*0.5**o
        ref[2,1,1]=(0.246577988625)/(o+1.)+(0.0512227634376)+(0.102779070928)*0.5**o
        ref[3,0,0]=(-0.0109584684212)/(o+1.)+(-0.334888666171)+(0.239371771102)*0.5**o
        ref[3,0,1]=(-0.125987676419)/(o+1.)+(-0.170179004032)+(-0.440522777827)*0.5**o
        ref[3,1,0]=(-0.354093141636)/(o+1.)+(-0.177294275919)+(0.119580825198)*0.5**o
        ref[3,1,1]=(-0.746761075412)/(o+1.)+(0.192068760731)+(-0.651985302566)*0.5**o
        ref[4,0,0]=(-0.42411673219)/(o+1.)+(-0.579162944937)+(-0.0719694008158)*0.5**o
        ref[4,0,1]=(0.169881169584)/(o+1.)+(0.483614029132)+(0.478341063616)*0.5**o
        ref[4,1,0]=(0.943567836264)/(o+1.)+(-0.898782941264)+(-0.670149001297)*0.5**o
        ref[4,1,1]=(-0.703610807936)/(o+1.)+(0.225342628657)+(-0.934767922806)*0.5**o
        ref[5,0,0]=(-0.788792225213)/(o+1.)+(-0.1768857093)+(-0.259984514102)*0.5**o
        ref[5,0,1]=(-0.788708879899)/(o+1.)+(-0.202288491141)+(0.827786772711)*0.5**o
        ref[5,1,0]=(-0.823081809707)/(o+1.)+(0.274720963093)+(-0.636239300197)*0.5**o
        ref[5,1,1]=(0.498922549807)/(o+1.)+(0.0129388257746)+(0.882580944588)*0.5**o
      else:
        arg[0,0,0]=(-0.426613654844)*x[0]+(-1.01433980801)*x[1]+(0.564621874916)*x[2]
        arg[0,0,1]=(-0.438521968882)*x[0]+(1.6033775267)*x[1]+(0.691349995832)*x[2]
        arg[0,1,0]=(-0.04337513731)*x[0]+(0.142927668234)*x[1]+(0.506581723199)*x[2]
        arg[0,1,1]=(-0.120598868162)*x[0]+(-0.213666787813)*x[1]+(-1.04221679332)*x[2]
        arg[1,0,0]=(0.153696490712)*x[0]+(-0.122516529815)*x[1]+(-1.08837272712)*x[2]
        arg[1,0,1]=(0.410706464237)*x[0]+(0.720079363315)*x[1]+(-0.767790357539)*x[2]
        arg[1,1,0]=(-1.03726428173)*x[0]+(-0.316809853003)*x[1]+(-0.780405868082)*x[2]
        arg[1,1,1]=(-0.427547967216)*x[0]+(-0.897766643609)*x[1]+(-1.56913552404)*x[2]
        arg[2,0,0]=(-0.282789248898)*x[0]+(-0.110846936031)*x[1]+(-1.4329391906)*x[2]
        arg[2,0,1]=(1.46866588617)*x[0]+(-1.23022501959)*x[1]+(-0.3123390018)*x[2]
        arg[2,1,0]=(-0.314115718621)*x[0]+(1.7333934438)*x[1]+(0.833572700726)*x[2]
        arg[2,1,1]=(0.464246462638)*x[0]+(0.267484068646)*x[1]+(0.342009204787)*x[2]
        arg[3,0,0]=(0.578853099163)*x[0]+(-0.0453855508465)*x[1]+(-0.0751578475516)*x[2]
        arg[3,0,1]=(-0.526734485752)*x[0]+(-1.03286310016)*x[1]+(-0.398986583917)*x[2]
        arg[3,1,0]=(0.0239088277798)*x[0]+(0.205105470192)*x[1]+(-1.04986349166)*x[2]
        arg[3,1,1]=(0.130893744001)*x[0]+(-0.273702140669)*x[1]+(-0.0804222415128)*x[2]
        arg[4,0,0]=(-0.822661618584)*x[0]+(-0.267189204519)*x[1]+(-0.290013766971)*x[2]
        arg[4,0,1]=(-1.46759823734)*x[0]+(0.317465945316)*x[1]+(0.108284317752)*x[2]
        arg[4,1,0]=(-1.3920990386)*x[0]+(-1.17823778694)*x[1]+(-0.18366434307)*x[2]
        arg[4,1,1]=(-0.0915496209516)*x[0]+(-0.367624609832)*x[1]+(1.30630257323)*x[2]
        arg[5,0,0]=(0.117840830455)*x[0]+(-0.667875433194)*x[1]+(-1.59716688455)*x[2]
        arg[5,0,1]=(-0.818693227605)*x[0]+(0.483842075248)*x[1]+(-0.408373205535)*x[2]
        arg[5,1,0]=(1.36906770301)*x[0]+(0.494926179999)*x[1]+(0.717041707262)*x[2]
        arg[5,1,1]=(0.819757116081)*x[0]+(0.0982458749015)*x[1]+(0.275080391599)*x[2]
        ref[0,0,0]=(0.236560699949)/(o+1.)+(-0.392784388084)+(-0.327323511722)*0.5**o
        ref[0,0,1]=(1.32491522045)/(o+1.)+(0.525146863711)+(-0.519003394222)*0.5**o
        ref[0,1,0]=(0.433749865563)/(o+1.)+(0.23007126008)+(-0.287758131599)*0.5**o
        ref[0,1,1]=(-0.0563021060686)/(o+1.)+(-0.403921160703)+(-0.512338021825)*0.5**o
        ref[1,0,0]=(-0.170496965963)/(o+1.)+(-0.914421777259)+(0.942147754258)*0.5**o
        ref[1,0,1]=(-0.372707483726)/(o+1.)+(0.636041158872)+(-0.536379364005)*0.5**o
        ref[1,1,0]=(-0.211529878364)/(o+1.)+(-0.912151985776)+(-0.0986461529033)*0.5**o
        ref[1,1,1]=(-1.32533998437)/(o+1.)+(-0.685030198386)+(-0.19904975372)*0.5**o
        ref[2,0,0]=(0.152048902469)/(o+1.)+(-0.88880590209)+(-0.201012473813)*0.5**o
        ref[2,0,1]=(-0.748900052246)/(o+1.)+(-0.0226120471998)+(0.720226011425)*0.5**o
        ref[2,1,0]=(0.77797707941)/(o+1.)+(1.01635055442)+(-0.557827762351)*0.5**o
        ref[2,1,1]=(1.08969334716)/(o+1.)+(-0.404644728844)+(0.793335846595)*0.5**o
        ref[3,0,0]=(1.56498448669)/(o+1.)+(-0.699967188685)+(0.293259591449)*0.5**o
        ref[3,0,1]=(-0.620133427128)/(o+1.)+(-0.550037413688)+(-0.238375915329)*0.5**o
        ref[3,1,0]=(-0.156380586288)/(o+1.)+(-0.0154369973565)+(-0.633594612687)*0.5**o
        ref[3,1,1]=(-0.352989559326)/(o+1.)+(-0.290547031627)+(0.710852984398)*0.5**o
        ref[4,0,0]=(-0.147579361355)/(o+1.)+(-0.615173424765)+(-0.00193837918766)*0.5**o
        ref[4,0,1]=(0.598152345277)/(o+1.)+(-0.540248606426)+(-0.559503106696)*0.5**o
        ref[4,1,0]=(-1.17182448825)/(o+1.)+(-0.590551681576)+(-0.401073317208)*0.5**o
        ref[4,1,1]=(0.837504141173)/(o+1.)+(0.0807414267186)+(-0.15185865216)*0.5**o
        ref[5,0,0]=(-0.646601115533)/(o+1.)+(-1.20895265273)+(0.9173049337)*0.5**o
        ref[5,0,1]=(1.26626110052)/(o+1.)+(-0.56471923973)+(-0.880046978955)*0.5**o
        ref[5,1,0]=(-0.069063169298)/(o+1.)+(0.883536796964)+(0.883025165644)*0.5**o
        ref[5,1,1]=(1.05999338993)/(o+1.)+(-0.392519102893)+(0.918128198439)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.257898226293)*x[0]+(-0.01230083314)*x[1]
        arg[0,0,0,1]=(0.0531183070647)*x[0]+(-0.170192287836)*x[1]
        arg[0,0,1,0]=(-0.364644940864)*x[0]+(1.04227651446)*x[1]
        arg[0,0,1,1]=(-0.894334108203)*x[0]+(0.151390331211)*x[1]
        arg[0,0,2,0]=(1.03825401708)*x[0]+(-0.882136097149)*x[1]
        arg[0,0,2,1]=(-1.05304910155)*x[0]+(1.06866311708)*x[1]
        arg[0,1,0,0]=(1.54742542159)*x[0]+(0.941778348482)*x[1]
        arg[0,1,0,1]=(1.1797380956)*x[0]+(-0.296417138234)*x[1]
        arg[0,1,1,0]=(-1.09196028027)*x[0]+(-0.783377624006)*x[1]
        arg[0,1,1,1]=(0.370952191443)*x[0]+(-0.0854141858991)*x[1]
        arg[0,1,2,0]=(-1.23552368139)*x[0]+(1.23848088087)*x[1]
        arg[0,1,2,1]=(0.216345350114)*x[0]+(-0.136524468427)*x[1]
        arg[0,2,0,0]=(-1.21206901263)*x[0]+(0.804219582144)*x[1]
        arg[0,2,0,1]=(0.192956526208)*x[0]+(0.588417678858)*x[1]
        arg[0,2,1,0]=(-0.940175603742)*x[0]+(0.329047260784)*x[1]
        arg[0,2,1,1]=(-0.138597387792)*x[0]+(0.398298326223)*x[1]
        arg[0,2,2,0]=(0.10089766824)*x[0]+(1.25459493298)*x[1]
        arg[0,2,2,1]=(-0.401163525883)*x[0]+(0.276542765282)*x[1]
        arg[0,3,0,0]=(0.252156795268)*x[0]+(-0.50223063539)*x[1]
        arg[0,3,0,1]=(0.466244610905)*x[0]+(-0.587025224301)*x[1]
        arg[0,3,1,0]=(-0.268082012713)*x[0]+(-1.28874675764)*x[1]
        arg[0,3,1,1]=(1.10867463519)*x[0]+(-0.165500912864)*x[1]
        arg[0,3,2,0]=(-1.94681908266)*x[0]+(0.166854108476)*x[1]
        arg[0,3,2,1]=(-0.537150417922)*x[0]+(0.0872851457131)*x[1]
        arg[0,4,0,0]=(-0.376548581807)*x[0]+(-0.245201133388)*x[1]
        arg[0,4,0,1]=(-0.135287690767)*x[0]+(0.750376238942)*x[1]
        arg[0,4,1,0]=(-0.0240491596907)*x[0]+(-1.21337088424)*x[1]
        arg[0,4,1,1]=(0.699215733397)*x[0]+(0.0596356032853)*x[1]
        arg[0,4,2,0]=(-0.521316010702)*x[0]+(-1.75309802615)*x[1]
        arg[0,4,2,1]=(0.403640172305)*x[0]+(1.41465827077)*x[1]
        arg[1,0,0,0]=(-1.01538089315)*x[0]+(-1.14785578909)*x[1]
        arg[1,0,0,1]=(1.10045201832)*x[0]+(0.630761728795)*x[1]
        arg[1,0,1,0]=(0.137673646147)*x[0]+(-0.387129090253)*x[1]
        arg[1,0,1,1]=(-1.04588287863)*x[0]+(-1.22433029859)*x[1]
        arg[1,0,2,0]=(-0.567830263535)*x[0]+(-0.405808616295)*x[1]
        arg[1,0,2,1]=(0.139326102495)*x[0]+(-1.93153384363)*x[1]
        arg[1,1,0,0]=(-0.638524762921)*x[0]+(0.506405751653)*x[1]
        arg[1,1,0,1]=(0.130775153144)*x[0]+(0.531317202208)*x[1]
        arg[1,1,1,0]=(0.557476206038)*x[0]+(0.447588676184)*x[1]
        arg[1,1,1,1]=(-0.538795916875)*x[0]+(-0.593869493531)*x[1]
        arg[1,1,2,0]=(-1.09827483942)*x[0]+(0.376431061876)*x[1]
        arg[1,1,2,1]=(-0.0074505620387)*x[0]+(-0.798645151486)*x[1]
        arg[1,2,0,0]=(0.251283911459)*x[0]+(0.194595071813)*x[1]
        arg[1,2,0,1]=(1.00482763799)*x[0]+(-1.00059863394)*x[1]
        arg[1,2,1,0]=(-0.453095434684)*x[0]+(-0.513922011418)*x[1]
        arg[1,2,1,1]=(0.558548837484)*x[0]+(0.0696328406825)*x[1]
        arg[1,2,2,0]=(-1.20625552545)*x[0]+(-0.151371710692)*x[1]
        arg[1,2,2,1]=(0.100262209656)*x[0]+(-0.314628804179)*x[1]
        arg[1,3,0,0]=(0.947063141882)*x[0]+(-0.297271200562)*x[1]
        arg[1,3,0,1]=(0.0760277251261)*x[0]+(-1.17982146387)*x[1]
        arg[1,3,1,0]=(-1.46946063969)*x[0]+(-0.615302663567)*x[1]
        arg[1,3,1,1]=(0.419747293729)*x[0]+(-0.629842531949)*x[1]
        arg[1,3,2,0]=(-1.44607954917)*x[0]+(0.520024093924)*x[1]
        arg[1,3,2,1]=(-1.01556392596)*x[0]+(-0.307173811354)*x[1]
        arg[1,4,0,0]=(-0.164339335244)*x[0]+(0.390736414982)*x[1]
        arg[1,4,0,1]=(-0.266932380288)*x[0]+(-0.731533848078)*x[1]
        arg[1,4,1,0]=(-0.903825489055)*x[0]+(-1.81564784421)*x[1]
        arg[1,4,1,1]=(-0.426008882665)*x[0]+(-1.09792948908)*x[1]
        arg[1,4,2,0]=(-1.32984927589)*x[0]+(1.58196302541)*x[1]
        arg[1,4,2,1]=(-0.0581032806145)*x[0]+(-0.323073355302)*x[1]
        arg[2,0,0,0]=(-0.744014251373)*x[0]+(-0.818253254872)*x[1]
        arg[2,0,0,1]=(0.368656965396)*x[0]+(-0.0864646380237)*x[1]
        arg[2,0,1,0]=(-0.112428845773)*x[0]+(0.109170104421)*x[1]
        arg[2,0,1,1]=(0.0682241548911)*x[0]+(-1.65999482283)*x[1]
        arg[2,0,2,0]=(0.212777058993)*x[0]+(0.127163480945)*x[1]
        arg[2,0,2,1]=(-0.256989443356)*x[0]+(0.183206687414)*x[1]
        arg[2,1,0,0]=(0.578997994373)*x[0]+(0.413087453689)*x[1]
        arg[2,1,0,1]=(0.886532090807)*x[0]+(0.330645525532)*x[1]
        arg[2,1,1,0]=(0.18224742814)*x[0]+(-0.713166591262)*x[1]
        arg[2,1,1,1]=(-0.25218296706)*x[0]+(0.736797897014)*x[1]
        arg[2,1,2,0]=(0.885270371408)*x[0]+(-1.35641397127)*x[1]
        arg[2,1,2,1]=(-0.963214290879)*x[0]+(0.348421598509)*x[1]
        arg[2,2,0,0]=(0.884120745351)*x[0]+(0.956857507884)*x[1]
        arg[2,2,0,1]=(-0.00795298517561)*x[0]+(0.51243740322)*x[1]
        arg[2,2,1,0]=(-0.563655252975)*x[0]+(-0.716305131721)*x[1]
        arg[2,2,1,1]=(0.648243491623)*x[0]+(-0.142239786485)*x[1]
        arg[2,2,2,0]=(1.50677848542)*x[0]+(-0.359722006818)*x[1]
        arg[2,2,2,1]=(-0.522592086302)*x[0]+(-0.18544927497)*x[1]
        arg[2,3,0,0]=(0.0822787762966)*x[0]+(0.196309662733)*x[1]
        arg[2,3,0,1]=(0.232573258627)*x[0]+(-0.0597976775432)*x[1]
        arg[2,3,1,0]=(1.50895737052)*x[0]+(0.546625934312)*x[1]
        arg[2,3,1,1]=(-1.64023373286)*x[0]+(0.220053261777)*x[1]
        arg[2,3,2,0]=(1.5897208842)*x[0]+(-0.0355885594992)*x[1]
        arg[2,3,2,1]=(-1.52822349982)*x[0]+(-0.0906934717656)*x[1]
        arg[2,4,0,0]=(0.496696582601)*x[0]+(-0.303888675059)*x[1]
        arg[2,4,0,1]=(-0.0819586112177)*x[0]+(-0.0449047294378)*x[1]
        arg[2,4,1,0]=(-0.678530758513)*x[0]+(-1.2173963447)*x[1]
        arg[2,4,1,1]=(-0.645450475268)*x[0]+(0.356771113426)*x[1]
        arg[2,4,2,0]=(-0.486643021867)*x[0]+(0.977763869689)*x[1]
        arg[2,4,2,1]=(-0.458991807397)*x[0]+(1.80038432923)*x[1]
        arg[3,0,0,0]=(0.270795948755)*x[0]+(-0.691512096338)*x[1]
        arg[3,0,0,1]=(-0.756717130099)*x[0]+(1.16189640534)*x[1]
        arg[3,0,1,0]=(0.967554346699)*x[0]+(-0.272119008963)*x[1]
        arg[3,0,1,1]=(1.14296491195)*x[0]+(-0.707175662379)*x[1]
        arg[3,0,2,0]=(1.15665030692)*x[0]+(0.414803389409)*x[1]
        arg[3,0,2,1]=(1.38997274955)*x[0]+(1.07065568192)*x[1]
        arg[3,1,0,0]=(-0.623998925511)*x[0]+(0.375545806421)*x[1]
        arg[3,1,0,1]=(-0.363431948398)*x[0]+(-0.183100417297)*x[1]
        arg[3,1,1,0]=(0.0435866301729)*x[0]+(-0.557137572068)*x[1]
        arg[3,1,1,1]=(-0.253225217468)*x[0]+(-0.0657932188329)*x[1]
        arg[3,1,2,0]=(0.147700921064)*x[0]+(0.681624563015)*x[1]
        arg[3,1,2,1]=(-0.075394297616)*x[0]+(0.271741245789)*x[1]
        arg[3,2,0,0]=(1.34803044812)*x[0]+(0.87991065609)*x[1]
        arg[3,2,0,1]=(0.0529487135353)*x[0]+(0.806480092352)*x[1]
        arg[3,2,1,0]=(-1.52038624166)*x[0]+(-0.992718138685)*x[1]
        arg[3,2,1,1]=(0.0859852188749)*x[0]+(1.63512596202)*x[1]
        arg[3,2,2,0]=(1.46925148575)*x[0]+(-1.60383564005)*x[1]
        arg[3,2,2,1]=(0.114463395473)*x[0]+(-1.0186377376)*x[1]
        arg[3,3,0,0]=(-1.02962502658)*x[0]+(-1.06788543146)*x[1]
        arg[3,3,0,1]=(-0.184391597652)*x[0]+(0.462603936483)*x[1]
        arg[3,3,1,0]=(0.611113335782)*x[0]+(0.282417397043)*x[1]
        arg[3,3,1,1]=(-0.193423873796)*x[0]+(-0.56108614964)*x[1]
        arg[3,3,2,0]=(-0.0290687598224)*x[0]+(-0.460221158417)*x[1]
        arg[3,3,2,1]=(0.784513522959)*x[0]+(-0.484813129206)*x[1]
        arg[3,4,0,0]=(0.423780748664)*x[0]+(-0.344020787894)*x[1]
        arg[3,4,0,1]=(1.10623024451)*x[0]+(-0.775008153301)*x[1]
        arg[3,4,1,0]=(0.554847534291)*x[0]+(-0.854749107999)*x[1]
        arg[3,4,1,1]=(-0.639595693597)*x[0]+(-0.966152813212)*x[1]
        arg[3,4,2,0]=(-0.0172621996023)*x[0]+(0.321815074269)*x[1]
        arg[3,4,2,1]=(0.39605156824)*x[0]+(-0.499138231749)*x[1]
        ref[0,0,0,0]=(0.169892838337)/(o+1.)+(-0.323261281472)+(0.722227117759)*0.5**o
        ref[0,0,0,1]=(-0.899797330597)/(o+1.)+(0.731242549464)+(-0.679761749104)*0.5**o
        ref[0,0,1,0]=(0.948969615551)/(o+1.)+(-0.173478478707)+(0.0756189154575)*0.5**o
        ref[0,0,1,1]=(-0.484419525942)/(o+1.)+(0.287761319651)+(-0.834046890353)*0.5**o
        ref[0,0,2,0]=(-0.275366085384)/(o+1.)+(-0.051636507926)+(0.534757021171)*0.5**o
        ref[0,0,2,1]=(0.338862631582)/(o+1.)+(0.076489529483)+(-0.47622767501)*0.5**o
        ref[0,1,0,0]=(0.960885993642)/(o+1.)+(0.36959166751)+(0.789134441414)*0.5**o
        ref[0,1,0,1]=(-0.101908900511)/(o+1.)+(0.343409131566)+(0.298411594744)*0.5**o
        ref[0,1,1,0]=(-0.50227382988)/(o+1.)+(-0.431549057432)+(-0.509965959533)*0.5**o
        ref[0,1,1,1]=(-0.294216689224)/(o+1.)+(0.293114907564)+(-0.00647512035965)*0.5**o
        ref[0,1,2,0]=(0.330313139948)/(o+1.)+(0.0601559902451)+(-0.447667920962)*0.5**o
        ref[0,1,2,1]=(-0.855325117375)/(o+1.)+(0.455706690618)+(0.0237326178252)*0.5**o
        ref[0,2,0,0]=(0.958948557218)/(o+1.)+(-0.307015973897)+(-0.752766039908)*0.5**o
        ref[0,2,0,1]=(0.622666086232)/(o+1.)+(0.417782845058)+(-0.676857571283)*0.5**o
        ref[0,2,1,0]=(-0.476269525636)/(o+1.)+(0.203518519787)+(-0.541895856898)*0.5**o
        ref[0,2,1,1]=(0.396862410812)/(o+1.)+(-0.318442063423)+(0.499722654464)*0.5**o
        ref[0,2,2,0]=(0.291546173382)/(o+1.)+(0.345688404218)+(0.372569619405)*0.5**o
        ref[0,2,2,1]=(-0.0863432016899)/(o+1.)+(0.296494638577)+(-0.631266836064)*0.5**o
        ref[0,3,0,0]=(0.437179077921)/(o+1.)+(-0.340515851088)+(-0.00622121586734)*0.5**o
        ref[0,3,0,1]=(-0.0670787919003)/(o+1.)+(-0.492735838895)+(0.931769856293)*0.5**o
        ref[0,3,1,0]=(-0.455620785864)/(o+1.)+(-0.866886545624)+(0.632565106759)*0.5**o
        ref[0,3,1,1]=(0.769395985412)/(o+1.)+(-0.303642587154)+(0.781062911223)*0.5**o
        ref[0,3,2,0]=(-0.338160377272)/(o+1.)+(-0.237545833658)+(-0.966712929597)*0.5**o
        ref[0,3,2,1]=(-0.796435534783)/(o+1.)+(0.544922318125)+(-0.743274373676)*0.5**o
        ref[0,4,0,0]=(0.452932120968)/(o+1.)+(-0.512874769603)+(-0.048932296957)*0.5**o
        ref[0,4,0,1]=(0.361579034112)/(o+1.)+(-0.0561290568189)+(0.365767627701)*0.5**o
        ref[0,4,1,0]=(-0.626053161593)/(o+1.)+(-0.761324946295)+(0.911283010254)*0.5**o
        ref[0,4,1,1]=(-0.0455977379835)/(o+1.)+(0.292122388548)+(0.22020429757)*0.5**o
        ref[0,4,2,0]=(-0.881052488675)/(o+1.)+(-0.210171400955)+(-0.973018746263)*0.5**o
        ref[0,4,2,1]=(0.955475719603)/(o+1.)+(0.367959093676)+(0.126904536122)*0.5**o
        ref[1,0,0,0]=(-0.314523133291)/(o+1.)+(-0.649561146707)+(-0.54959125554)*0.5**o
        ref[1,0,0,1]=(0.776726613012)/(o+1.)+(0.0747896120381)+(0.804907910029)*0.5**o
        ref[1,0,1,0]=(0.416309554773)/(o+1.)+(0.0650606132082)+(-0.795886225295)*0.5**o
        ref[1,0,1,1]=(-0.642083920221)/(o+1.)+(-0.518812997818)+(-0.590503261366)*0.5**o
        ref[1,0,2,0]=(0.209400082886)/(o+1.)+(-0.592256267485)+(0.00147357225359)*0.5**o
        ref[1,0,2,1]=(-0.969945221754)/(o+1.)+(-0.364559871358)+(-0.0931427766681)*0.5**o
        ref[1,1,0,0]=(0.941138659497)/(o+1.)+(-0.103485523138)+(-0.866286624489)*0.5**o
        ref[1,1,0,1]=(0.527395399401)/(o+1.)+(0.373358049173)+(-0.612019142395)*0.5**o
        ref[1,1,1,0]=(-0.276942103917)/(o+1.)+(0.624509599487)+(0.0329877871648)*0.5**o
        ref[1,1,1,1]=(-0.153477829221)/(o+1.)+(-0.0616268433725)+(-0.85593389444)*0.5**o
        ref[1,1,2,0]=(0.2685285297)/(o+1.)+(-0.301824551915)+(-0.386723203413)*0.5**o
        ref[1,1,2,1]=(-0.530960074652)/(o+1.)+(0.248621312801)+(-0.772378264474)*0.5**o
        ref[1,2,0,0]=(0.745071931955)/(o+1.)+(-0.149268301427)+(-0.000656345828701)*0.5**o
        ref[1,2,0,1]=(-0.581822428087)/(o+1.)+(0.2315691847)+(0.122913062738)*0.5**o
        ref[1,2,1,0]=(-0.325165099021)/(o+1.)+(-0.414758139139)+(0.187663931198)*0.5**o
        ref[1,2,1,1]=(-0.389683593964)/(o+1.)+(0.265985790556)+(0.485893691018)*0.5**o
        ref[1,2,2,0]=(0.661969003098)/(o+1.)+(-0.514584521701)+(-0.990427195841)*0.5**o
        ref[1,2,2,1]=(-0.0828384828188)/(o+1.)+(-0.482166570368)+(0.832805029032)*0.5**o
        ref[1,3,0,0]=(-0.756020671713)/(o+1.)+(0.627704058228)+(0.150404496577)*0.5**o
        ref[1,3,0,1]=(-0.93648010541)/(o+1.)+(0.11022703478)+(-0.387767702896)*0.5**o
        ref[1,3,1,0]=(-0.995823513997)/(o+1.)+(-0.242072805569)+(-0.604794178119)*0.5**o
        ref[1,3,1,1]=(-0.543071878541)/(o+1.)+(-0.160890620667)+(0.654757881656)*0.5**o
        ref[1,3,2,0]=(0.803184441579)/(o+1.)+(-0.612293875048)+(-0.504652146723)*0.5**o
        ref[1,3,2,1]=(-0.786714439322)/(o+1.)+(-0.175556007517)+(-0.18491128296)*0.5**o
        ref[1,4,0,0]=(-0.302928744298)/(o+1.)+(0.476360987754)+(-0.423396151472)*0.5**o
        ref[1,4,0,1]=(-0.410240602476)/(o+1.)+(-0.49551630458)+(0.40280698327)*0.5**o
        ref[1,4,1,0]=(-0.820677400346)/(o+1.)+(-0.573037933363)+(-0.752720066188)*0.5**o
        ref[1,4,1,1]=(-0.492748923658)/(o+1.)+(-0.600739806799)+(0.170290165507)*0.5**o
        ref[1,4,2,0]=(0.926654770824)/(o+1.)+(0.011282908181)+(-0.697106837662)*0.5**o
        ref[1,4,2,1]=(-0.958477382611)/(o+1.)+(0.618515858902)+(-0.65973097111)*0.5**o
        ref[2,0,0,0]=(-0.15636950624)/(o+1.)+(-0.520670547262)+(-0.36455690548)*0.5**o
        ref[2,0,0,1]=(-0.726424878286)/(o+1.)+(0.697757734621)+(-0.386898263585)*0.5**o
        ref[2,0,1,0]=(-0.055162393508)/(o+1.)+(0.247015646426)+(-0.442127640696)*0.5**o
        ref[2,0,1,1]=(-0.769332288875)/(o+1.)+(-0.0917175733456)+(-0.639003232373)*0.5**o
        ref[2,0,2,0]=(-0.196390038297)/(o+1.)+(0.0977907270548)+(0.340749124126)*0.5**o
        ref[2,0,2,1]=(0.841927481859)/(o+1.)+(-0.489185426146)+(0.0626606144906)*0.5**o
        ref[2,1,0,0]=(0.590018249148)/(o+1.)+(0.378476967182)+(-0.354886735451)*0.5**o
        ref[2,1,0,1]=(0.178495604742)/(o+1.)+(0.233073081069)+(0.572535849459)*0.5**o
        ref[2,1,1,0]=(-0.762671064919)/(o+1.)+(-0.108539686142)+(0.448831274081)*0.5**o
        ref[2,1,1,1]=(0.927591406535)/(o+1.)+(0.0441635703243)+(-0.53130361723)*0.5**o
        ref[2,1,2,0]=(-0.604360706259)/(o+1.)+(-0.134848102332)+(0.402913311063)*0.5**o
        ref[2,1,2,1]=(0.229727627441)/(o+1.)+(-0.323296604293)+(-0.197927111225)*0.5**o
        ref[2,2,0,0]=(0.287478356112)/(o+1.)+(0.762635506201)+(0.0282288847209)*0.5**o
        ref[2,2,0,1]=(0.243518182621)/(o+1.)+(-0.241502069199)+(0.743970373822)*0.5**o
        ref[2,2,1,0]=(-0.369049760378)/(o+1.)+(-0.21941778965)+(-0.472075045017)*0.5**o
        ref[2,2,1,1]=(-0.745343658666)/(o+1.)+(0.193375557775)+(0.864596248255)*0.5**o
        ref[2,2,2,0]=(-0.491769251494)/(o+1.)+(0.442742387618)+(0.753340954862)*0.5**o
        ref[2,2,2,1]=(0.551196377158)/(o+1.)+(-0.523535406613)+(-0.212166925203)*0.5**o
        ref[2,3,0,0]=(0.192271522168)/(o+1.)+(-0.0181953944687)+(0.122707705798)*0.5**o
        ref[2,3,0,1]=(-0.593768738594)/(o+1.)+(0.199088910545)+(0.368366498588)*0.5**o
        ref[2,3,1,0]=(0.0866211808983)/(o+1.)+(0.723512694127)+(0.521936735685)*0.5**o
        ref[2,3,1,1]=(-0.453288268254)/(o+1.)+(-0.15580970281)+(-0.655272797205)*0.5**o
        ref[2,3,2,0]=(0.614105802669)/(o+1.)+(0.0411427694898)+(0.85774098305)*0.5**o
        ref[2,3,2,1]=(0.263133533654)/(o+1.)+(-0.628776355867)+(-0.624497793507)*0.5**o
        ref[2,4,0,0]=(-0.809347914722)/(o+1.)+(0.367719412237)+(0.26671699779)*0.5**o
        ref[2,4,0,1]=(0.773124137566)/(o+1.)+(-0.442496395542)+(-0.0149946871367)*0.5**o
        ref[2,4,1,0]=(-0.927311525126)/(o+1.)+(-0.0125177466568)+(-0.943580084777)*0.5**o
        ref[2,4,1,1]=(0.306409665483)/(o+1.)+(0.128569553604)+(-0.852228134533)*0.5**o
        ref[2,4,2,0]=(0.926185522145)/(o+1.)+(0.0750538146738)+(-0.58517230367)*0.5**o
        ref[2,4,2,1]=(0.858245470922)/(o+1.)+(0.0125876600893)+(0.457971730734)*0.5**o
        ref[3,0,0,0]=(-0.74501947953)/(o+1.)+(0.343942243752)+(-0.363581155556)*0.5**o
        ref[3,0,0,1]=(0.850890737948)/(o+1.)+(0.157997002133)+(-0.761705466968)*0.5**o
        ref[3,0,1,0]=(-0.0468962248591)/(o+1.)+(0.0297980532763)+(0.682735456043)*0.5**o
        ref[3,0,1,1]=(0.0126326425456)/(o+1.)+(-0.076115296096)+(0.575387199213)*0.5**o
        ref[3,0,2,0]=(-0.410549752333)/(o+1.)+(0.85556811978)+(0.270867209104)*0.5**o
        ref[3,0,2,1]=(0.580487848545)/(o+1.)+(0.59069171729)+(0.698757148346)*0.5**o
        ref[3,1,0,0]=(0.139052775786)/(o+1.)+(-0.334747475889)+(0.281989056901)*0.5**o
        ref[3,1,0,1]=(0.591083610998)/(o+1.)+(-0.322680889668)+(-0.492254197357)*0.5**o
        ref[3,1,1,0]=(0.384713965427)/(o+1.)+(-0.522411606102)+(0.146558304882)*0.5**o
        ref[3,1,1,1]=(-0.827615582009)/(o+1.)+(-0.00295218575854)+(0.514501517225)*0.5**o
        ref[3,1,2,0]=(0.84762578765)/(o+1.)+(0.208361555685)+(-0.43502341494)*0.5**o
        ref[3,1,2,1]=(-0.554668615664)/(o+1.)+(0.825347835483)+(-0.899680107129)*0.5**o
        ref[3,2,0,0]=(0.652802712143)/(o+1.)+(0.397197582731)+(0.7807432266)*0.5**o
        ref[3,2,0,1]=(0.421495251937)/(o+1.)+(-0.119571710512)+(0.677076974975)*0.5**o
        ref[3,2,1,0]=(-0.230439630375)/(o+1.)+(-0.733165339017)+(-0.816334071941)*0.5**o
        ref[3,2,1,1]=(0.64153308098)/(o+1.)+(0.778374176432)+(-0.477170252948)*0.5**o
        ref[3,2,2,0]=(-0.987848717403)/(o+1.)+(-0.0606435888479)+(0.974551740798)*0.5**o
        ref[3,2,2,1]=(-0.318593711859)/(o+1.)+(-0.459904454409)+(0.334228278552)*0.5**o
        ref[3,3,0,0]=(-0.418115413351)/(o+1.)+(-0.348616063735)+(-0.982162917221)*0.5**o
        ref[3,3,0,1]=(-0.437684435015)/(o+1.)+(0.804145379606)+(-0.892393985366)*0.5**o
        ref[3,3,1,0]=(0.944100110135)/(o+1.)+(-0.059994562245)+(0.0694197471801)*0.5**o
        ref[3,3,1,1]=(-0.860656048613)/(o+1.)+(0.178136109535)+(-0.250126193893)*0.5**o
        ref[3,3,2,0]=(-0.363631991908)/(o+1.)+(-0.298384287282)+(0.471110648232)*0.5**o
        ref[3,3,2,1]=(-0.578021736479)/(o+1.)+(0.320913094365)+(0.235895941501)*0.5**o
        ref[3,4,0,0]=(-0.946995649946)/(o+1.)+(0.0451978887455)+(0.936359833225)*0.5**o
        ref[3,4,0,1]=(-0.636305818291)/(o+1.)+(0.258932088634)+(0.449663732234)*0.5**o
        ref[3,4,1,0]=(-0.823421937366)/(o+1.)+(0.112980674356)+(0.297559014947)*0.5**o
        ref[3,4,1,1]=(-0.441768022626)/(o+1.)+(-0.240840962617)+(-0.682298558949)*0.5**o
        ref[3,4,2,0]=(-0.258808936822)/(o+1.)+(0.510306577882)+(-0.457251344276)*0.5**o
        ref[3,4,2,1]=(-0.489365749016)/(o+1.)+(-0.296617385332)+(0.979513856171)*0.5**o
      else:
        arg[0,0,0,0]=(-0.474247989728)*x[0]+(-0.768689352235)*x[1]+(-1.08527900853)*x[2]
        arg[0,0,0,1]=(-0.237113797789)*x[0]+(0.0842735741018)*x[1]+(0.203858551029)*x[2]
        arg[0,0,1,0]=(0.24237127665)*x[0]+(0.421022955803)*x[1]+(-1.52818540927)*x[2]
        arg[0,0,1,1]=(1.34624865817)*x[0]+(-0.798943272818)*x[1]+(-0.823168386315)*x[2]
        arg[0,0,2,0]=(-0.906723771267)*x[0]+(0.328893123771)*x[1]+(-0.942770202463)*x[2]
        arg[0,0,2,1]=(1.54305197694)*x[0]+(0.813063941984)*x[1]+(0.205532727168)*x[2]
        arg[0,1,0,0]=(0.399159979872)*x[0]+(-1.52089837934)*x[1]+(-1.4064251406)*x[2]
        arg[0,1,0,1]=(-0.121139354731)*x[0]+(1.12768246422)*x[1]+(0.234299409411)*x[2]
        arg[0,1,1,0]=(-1.62513899388)*x[0]+(-0.173131700583)*x[1]+(1.24199996799)*x[2]
        arg[0,1,1,1]=(-0.371929434177)*x[0]+(-0.606278610409)*x[1]+(-0.00600790063066)*x[2]
        arg[0,1,2,0]=(-0.804531294423)*x[0]+(0.869392248002)*x[1]+(0.840417359943)*x[2]
        arg[0,1,2,1]=(0.329452946275)*x[0]+(1.1072495446)*x[1]+(-0.213517932323)*x[2]
        arg[0,2,0,0]=(1.02784887224)*x[0]+(0.435517042337)*x[1]+(1.58767206677)*x[2]
        arg[0,2,0,1]=(-0.0120997177)*x[0]+(0.430135803314)*x[1]+(-0.717473068956)*x[2]
        arg[0,2,1,0]=(0.955455517302)*x[0]+(0.712976918968)*x[1]+(-0.624804508641)*x[2]
        arg[0,2,1,1]=(-1.03293986594)*x[0]+(-0.539062834482)*x[1]+(-1.61172701428)*x[2]
        arg[0,2,2,0]=(-0.177662810138)*x[0]+(0.553352366127)*x[1]+(0.695417121111)*x[2]
        arg[0,2,2,1]=(0.318133474679)*x[0]+(-0.452528947926)*x[1]+(0.468602505091)*x[2]
        arg[0,3,0,0]=(-0.294467725916)*x[0]+(1.67314409894)*x[1]+(0.100736440007)*x[2]
        arg[0,3,0,1]=(-0.517162349887)*x[0]+(-1.57430191254)*x[1]+(1.24045877402)*x[2]
        arg[0,3,1,0]=(0.667393640368)*x[0]+(0.340993632212)*x[1]+(-1.09920535559)*x[2]
        arg[0,3,1,1]=(0.37451406727)*x[0]+(0.37350649326)*x[1]+(0.357440871322)*x[2]
        arg[0,3,2,0]=(-0.989366398194)*x[0]+(0.596179154759)*x[1]+(-0.573280308087)*x[2]
        arg[0,3,2,1]=(-0.25561190141)*x[0]+(1.42296905238)*x[1]+(-1.21032031399)*x[2]
        arg[0,4,0,0]=(-0.41394085285)*x[0]+(-0.751213929646)*x[1]+(-0.68470718293)*x[2]
        arg[0,4,0,1]=(0.702985204167)*x[0]+(-0.0375821582587)*x[1]+(0.290630803819)*x[2]
        arg[0,4,1,0]=(0.0558204653441)*x[0]+(1.61940057302)*x[1]+(-0.518863528354)*x[2]
        arg[0,4,1,1]=(1.25876303604)*x[0]+(-1.40923202908)*x[1]+(1.05202333762)*x[2]
        arg[0,4,2,0]=(0.0715295483294)*x[0]+(0.883869702529)*x[1]+(-0.463253474408)*x[2]
        arg[0,4,2,1]=(0.564563789615)*x[0]+(-0.135049742988)*x[1]+(-1.3118878567)*x[2]
        arg[1,0,0,0]=(0.707267335291)*x[0]+(-0.197244449872)*x[1]+(-0.114362773781)*x[2]
        arg[1,0,0,1]=(-0.144762815659)*x[0]+(0.434177706452)*x[1]+(0.0754859099691)*x[2]
        arg[1,0,1,0]=(-0.906934991359)*x[0]+(0.179011995627)*x[1]+(0.823936036469)*x[2]
        arg[1,0,1,1]=(1.11165724605)*x[0]+(0.48023223731)*x[1]+(0.981353808181)*x[2]
        arg[1,0,2,0]=(0.499460583481)*x[0]+(0.171586243561)*x[1]+(-0.252474775657)*x[2]
        arg[1,0,2,1]=(-0.604289370884)*x[0]+(1.11610914086)*x[1]+(0.747439062711)*x[2]
        arg[1,1,0,0]=(0.696191463882)*x[0]+(0.349547714404)*x[1]+(0.651361019042)*x[2]
        arg[1,1,0,1]=(0.152679923426)*x[0]+(-0.448442911425)*x[1]+(1.33964371501)*x[2]
        arg[1,1,1,0]=(-0.74612592015)*x[0]+(-1.27089239786)*x[1]+(-0.261725631736)*x[2]
        arg[1,1,1,1]=(1.50839493929)*x[0]+(1.93229724987)*x[1]+(-0.963996845831)*x[2]
        arg[1,1,2,0]=(0.867984447375)*x[0]+(-0.777493811854)*x[1]+(0.847847553133)*x[2]
        arg[1,1,2,1]=(1.34249389498)*x[0]+(0.776160804062)*x[1]+(-0.295393972199)*x[2]
        arg[1,2,0,0]=(-0.655771613243)*x[0]+(1.58220214455)*x[1]+(-0.406292477629)*x[2]
        arg[1,2,0,1]=(0.59294895986)*x[0]+(0.544961164723)*x[1]+(0.634279670948)*x[2]
        arg[1,2,1,0]=(1.15324179862)*x[0]+(0.645764835217)*x[1]+(0.670364725309)*x[2]
        arg[1,2,1,1]=(-0.819459106271)*x[0]+(0.024815625404)*x[1]+(0.769753650676)*x[2]
        arg[1,2,2,0]=(1.19042366161)*x[0]+(0.982020548972)*x[1]+(0.92730609738)*x[2]
        arg[1,2,2,1]=(-0.742441488349)*x[0]+(-1.3101429812)*x[1]+(-0.198104246324)*x[2]
        arg[1,3,0,0]=(0.349205030437)*x[0]+(-0.745240339596)*x[1]+(-0.744742092259)*x[2]
        arg[1,3,0,1]=(0.993101296711)*x[0]+(-1.48331355036)*x[1]+(-1.00616778677)*x[2]
        arg[1,3,1,0]=(-1.10586442824)*x[0]+(1.71922449779)*x[1]+(-0.943451153259)*x[2]
        arg[1,3,1,1]=(-1.72393763035)*x[0]+(-1.14113495956)*x[1]+(-0.211909615809)*x[2]
        arg[1,3,2,0]=(-0.378302963108)*x[0]+(-0.494705322002)*x[1]+(0.940164363332)*x[2]
        arg[1,3,2,1]=(0.850541148743)*x[0]+(-1.45461945468)*x[1]+(1.20025821328)*x[2]
        arg[1,4,0,0]=(1.33882463795)*x[0]+(0.475208133119)*x[1]+(0.722686059253)*x[2]
        arg[1,4,0,1]=(0.0750422155969)*x[0]+(1.29758288845)*x[1]+(-1.41431668738)*x[2]
        arg[1,4,1,0]=(1.36072344575)*x[0]+(-0.345743547563)*x[1]+(-0.63248018754)*x[2]
        arg[1,4,1,1]=(0.579049229913)*x[0]+(0.954245074078)*x[1]+(-0.170538926965)*x[2]
        arg[1,4,2,0]=(1.67615484634)*x[0]+(0.561067019023)*x[1]+(-1.26952326914)*x[2]
        arg[1,4,2,1]=(0.700910547934)*x[0]+(0.58717328764)*x[1]+(0.795138196965)*x[2]
        arg[2,0,0,0]=(0.763193592359)*x[0]+(-0.397549328381)*x[1]+(-0.230593559545)*x[2]
        arg[2,0,0,1]=(0.454026153468)*x[0]+(-1.12464443717)*x[1]+(-0.147462098246)*x[2]
        arg[2,0,1,0]=(-0.276457034935)*x[0]+(0.118173156654)*x[1]+(0.57296366275)*x[2]
        arg[2,0,1,1]=(1.31887812253)*x[0]+(-0.544644483254)*x[1]+(0.0013728350582)*x[2]
        arg[2,0,2,0]=(-0.257589384265)*x[0]+(-0.016005453203)*x[1]+(-0.508836430965)*x[2]
        arg[2,0,2,1]=(-0.259420301773)*x[0]+(0.122645293168)*x[1]+(0.381613885289)*x[2]
        arg[2,1,0,0]=(-1.64168512418)*x[0]+(0.025182006192)*x[1]+(0.0268780458975)*x[2]
        arg[2,1,0,1]=(-0.0885344200898)*x[0]+(0.408795219114)*x[1]+(0.635798801606)*x[2]
        arg[2,1,1,0]=(-0.959658503283)*x[0]+(1.034415556)*x[1]+(0.725681105748)*x[2]
        arg[2,1,1,1]=(1.59312707789)*x[0]+(-0.116514952101)*x[1]+(-0.86121633507)*x[2]
        arg[2,1,2,0]=(-1.44859572067)*x[0]+(-1.28344247275)*x[1]+(0.299296401869)*x[2]
        arg[2,1,2,1]=(-0.488424372501)*x[0]+(1.23481402288)*x[1]+(0.01702902463)*x[2]
        arg[2,2,0,0]=(-1.29986951244)*x[0]+(-1.20190912751)*x[1]+(-0.755848291051)*x[2]
        arg[2,2,0,1]=(-0.52421656312)*x[0]+(-0.335471509689)*x[1]+(-1.01264312499)*x[2]
        arg[2,2,1,0]=(-1.12066642609)*x[0]+(-1.02138558053)*x[1]+(-1.00777755552)*x[2]
        arg[2,2,1,1]=(-1.42127024744)*x[0]+(1.63480022479)*x[1]+(0.542244906643)*x[2]
        arg[2,2,2,0]=(-0.418852880297)*x[0]+(-0.983110436453)*x[1]+(-0.622603673805)*x[2]
        arg[2,2,2,1]=(0.727366446825)*x[0]+(-0.256324856361)*x[1]+(-0.20608740758)*x[2]
        arg[2,3,0,0]=(0.548268302883)*x[0]+(0.0102173127665)*x[1]+(0.869833576217)*x[2]
        arg[2,3,0,1]=(-1.70159091061)*x[0]+(1.15876875083)*x[1]+(-0.234757706773)*x[2]
        arg[2,3,1,0]=(-1.72817024817)*x[0]+(0.690055132218)*x[1]+(-0.149398998825)*x[2]
        arg[2,3,1,1]=(1.80829412387)*x[0]+(-0.217899654855)*x[1]+(-0.922672217822)*x[2]
        arg[2,3,2,0]=(1.23855200328)*x[0]+(0.412795411508)*x[1]+(0.960476191755)*x[2]
        arg[2,3,2,1]=(-1.43061904621)*x[0]+(-0.766039000745)*x[1]+(-1.25765397361)*x[2]
        arg[2,4,0,0]=(-0.660872016687)*x[0]+(-0.000405233718708)*x[1]+(-0.178507968219)*x[2]
        arg[2,4,0,1]=(-0.168526882871)*x[0]+(-1.06936802685)*x[1]+(1.54083857104)*x[2]
        arg[2,4,1,0]=(0.565807359529)*x[0]+(-1.41401297939)*x[1]+(-1.23254260537)*x[2]
        arg[2,4,1,1]=(-0.373920075802)*x[0]+(0.407330812903)*x[1]+(-0.204667001328)*x[2]
        arg[2,4,2,0]=(0.874100879885)*x[0]+(-0.191238094267)*x[1]+(1.54890969328)*x[2]
        arg[2,4,2,1]=(0.566428125313)*x[0]+(-1.60899595601)*x[1]+(-1.05112529768)*x[2]
        arg[3,0,0,0]=(-0.688037248111)*x[0]+(-1.56483484784)*x[1]+(0.26782504619)*x[2]
        arg[3,0,0,1]=(0.418788954459)*x[0]+(-0.670455226092)*x[1]+(-0.0562760591375)*x[2]
        arg[3,0,1,0]=(-0.2614362754)*x[0]+(0.271050519855)*x[1]+(0.227811982923)*x[2]
        arg[3,0,1,1]=(-0.873198928647)*x[0]+(-0.483763141395)*x[1]+(0.33532383979)*x[2]
        arg[3,0,2,0]=(-0.864161938176)*x[0]+(-1.4682754386)*x[1]+(-0.38126411162)*x[2]
        arg[3,0,2,1]=(0.142827521381)*x[0]+(0.464483940188)*x[1]+(0.214297576606)*x[2]
        arg[3,1,0,0]=(0.188747079113)*x[0]+(0.590821253713)*x[1]+(0.908850367108)*x[2]
        arg[3,1,0,1]=(0.0587541582624)*x[0]+(0.955494266151)*x[1]+(1.70822209849)*x[2]
        arg[3,1,1,0]=(-0.420356861523)*x[0]+(-0.551445918262)*x[1]+(-0.0187114754269)*x[2]
        arg[3,1,1,1]=(1.07071658401)*x[0]+(-0.589035602194)*x[1]+(0.133725074027)*x[2]
        arg[3,1,2,0]=(0.370477043767)*x[0]+(0.226262353618)*x[1]+(0.566087784489)*x[2]
        arg[3,1,2,1]=(1.77828660075)*x[0]+(0.607419742685)*x[1]+(0.0297054744582)*x[2]
        arg[3,2,0,0]=(-1.19326939474)*x[0]+(1.20875153887)*x[1]+(-0.124756222091)*x[2]
        arg[3,2,0,1]=(0.0645190914271)*x[0]+(0.357580475716)*x[1]+(-1.59461469718)*x[2]
        arg[3,2,1,0]=(1.4853619905)*x[0]+(0.751582301648)*x[1]+(0.859363177167)*x[2]
        arg[3,2,1,1]=(1.36658655075)*x[0]+(-0.776650996613)*x[1]+(0.654743994261)*x[2]
        arg[3,2,2,0]=(0.406696005677)*x[0]+(-0.528546876384)*x[1]+(-0.439775325396)*x[2]
        arg[3,2,2,1]=(-0.725198159461)*x[0]+(-0.0933136564416)*x[1]+(-0.720365141393)*x[2]
        arg[3,3,0,0]=(-0.428669112669)*x[0]+(0.626162511175)*x[1]+(-0.382776233923)*x[2]
        arg[3,3,0,1]=(0.340474939218)*x[0]+(-0.140550383925)*x[1]+(0.353836983746)*x[2]
        arg[3,3,1,0]=(0.851646816711)*x[0]+(0.0171684195949)*x[1]+(0.38173044322)*x[2]
        arg[3,3,1,1]=(0.359647722225)*x[0]+(0.425799941165)*x[1]+(-0.80174052212)*x[2]
        arg[3,3,2,0]=(0.13012568736)*x[0]+(-0.573735693651)*x[1]+(1.08843730088)*x[2]
        arg[3,3,2,1]=(-1.01372953294)*x[0]+(0.335282956856)*x[1]+(1.24570512628)*x[2]
        arg[3,4,0,0]=(-0.05674609039)*x[0]+(-0.814879575064)*x[1]+(-0.169789408264)*x[2]
        arg[3,4,0,1]=(1.5968230938)*x[0]+(0.0679819277023)*x[1]+(0.892284314512)*x[2]
        arg[3,4,1,0]=(1.22220298478)*x[0]+(-0.0258417926472)*x[1]+(0.440670138568)*x[2]
        arg[3,4,1,1]=(0.647072768716)*x[0]+(-0.177651087906)*x[1]+(-0.115553644646)*x[2]
        arg[3,4,2,0]=(1.29816363306)*x[0]+(-1.18656044167)*x[1]+(-0.465776533157)*x[2]
        arg[3,4,2,1]=(0.328907496091)*x[0]+(-0.179857937151)*x[1]+(-0.476306938306)*x[2]
        ref[0,0,0,0]=(-0.530378359482)/(o+1.)+(-0.935143327523)+(0.0724486640358)*0.5**o
        ref[0,0,0,1]=(1.22837847871)/(o+1.)+(-0.113350954728)+(-0.950658241915)*0.5**o
        ref[0,0,1,0]=(-0.000414683535449)/(o+1.)+(-0.183158731552)+(-0.49805903018)*0.5**o
        ref[0,0,1,1]=(-0.882730389132)/(o+1.)+(-0.112174958816)+(0.831217305802)*0.5**o
        ref[0,0,2,0]=(-1.1913557966)/(o+1.)+(-0.0615750891074)+(-0.206094875146)*0.5**o
        ref[0,0,2,1]=(0.634536492589)/(o+1.)+(0.649890059464)+(0.627332034578)*0.5**o
        ref[0,1,0,0]=(-1.38665625775)/(o+1.)+(-1.0210653524)+(0.900623422471)*0.5**o
        ref[0,1,0,1]=(1.20633678024)/(o+1.)+(0.0888193121856)+(-0.143132885707)*0.5**o
        ref[0,1,1,0]=(0.452225278035)/(o+1.)+(-0.148926943801)+(-0.710642116901)*0.5**o
        ref[0,1,1,1]=(-0.244223430204)/(o+1.)+(-0.619607191996)+(0.499221868978)*0.5**o
        ref[0,1,2,0]=(0.0383849036606)/(o+1.)+(0.759484971644)+(-0.652076533427)*0.5**o
        ref[0,1,2,1]=(0.207558055759)/(o+1.)+(0.502305382384)+(0.0110157380215)*0.5**o
        ref[0,2,0,0]=(0.761374280144)/(o+1.)+(0.777616611515)+(0.734430478174)*0.5**o
        ref[0,2,0,1]=(0.554502862028)/(o+1.)+(-0.409274602016)+(-0.0353906413371)*0.5**o
        ref[0,2,1,0]=(-0.252543841709)/(o+1.)+(0.386640377894)+(0.522891013551)*0.5**o
        ref[0,2,1,1]=(-1.52497235887)/(o+1.)+(-0.807603254401)+(-0.0435508470367)*0.5**o
        ref[0,2,2,0]=(0.172485606257)/(o+1.)+(0.350297526187)+(0.19802601847)*0.5**o
        ref[0,2,2,1]=(-0.141271359379)/(o+1.)+(0.0509198134622)+(0.373638764298)*0.5**o
        ref[0,3,0,0]=(1.31278836809)/(o+1.)+(-0.0918739323256)+(0.350372309595)*0.5**o
        ref[0,3,0,1]=(-0.396933928109)/(o+1.)+(-0.436137805999)+(0.418204051696)*0.5**o
        ref[0,3,1,0]=(-0.206619840099)/(o+1.)+(-0.118703289798)+(0.353208336689)*0.5**o
        ref[0,3,1,1]=(-0.595153371293)/(o+1.)+(0.615722581562)+(0.469169640021)*0.5**o
        ref[0,3,2,0]=(-0.680858997986)/(o+1.)+(-0.0393223937964)+(-0.206963765944)*0.5**o
        ref[0,3,2,1]=(0.012753264598)/(o+1.)+(0.246188741489)+(-0.548093910595)*0.5**o
        ref[0,4,0,0]=(-0.40478987122)/(o+1.)+(-0.890399681609)+(0.33572726901)*0.5**o
        ref[0,4,0,1]=(0.0280062557264)/(o+1.)+(0.0358049413228)+(0.856417711355)*0.5**o
        ref[0,4,1,0]=(0.181560004565)/(o+1.)+(0.329885369854)+(0.31502676574)*0.5**o
        ref[0,4,1,1]=(0.0552681276841)/(o+1.)+(0.00372919426319)+(0.838827828375)*0.5**o
        ref[0,4,2,0]=(0.656057162724)/(o+1.)+(0.179315527684)+(-0.522542441641)*0.5**o
        ref[0,4,2,1]=(0.126189389317)/(o+1.)+(-0.429452001643)+(-0.149659196108)*0.5**o
        ref[1,0,0,0]=(1.43443485527)/(o+1.)+(-0.55091754491)+(0.0630603461901)*0.5**o
        ref[1,0,0,1]=(-0.612191779306)/(o+1.)+(0.604025235381)+(-0.230957890695)*0.5**o
        ref[1,0,1,0]=(0.373085423681)/(o+1.)+(-0.0817391530697)+(-0.113594076805)*0.5**o
        ref[1,0,1,1]=(0.761862677657)/(o+1.)+(0.819524484428)+(0.172331645033)*0.5**o
        ref[1,0,2,0]=(0.929265450746)/(o+1.)+(-0.382585784938)+(0.254478170515)*0.5**o
        ref[1,0,2,1]=(1.58236203463)/(o+1.)+(-0.280641188811)+(0.238179175679)*0.5**o
        ref[1,1,0,0]=(-0.117055726312)/(o+1.)+(0.957914090571)+(-0.101672257501)*0.5**o
        ref[1,1,0,1]=(0.0849062553796)/(o+1.)+(0.0703564566846)+(0.818261558262)*0.5**o
        ref[1,1,1,0]=(-1.72219113622)/(o+1.)+(-0.357454724939)+(0.15835663635)*0.5**o
        ref[1,1,1,1]=(0.739078400468)/(o+1.)+(0.532292543169)+(0.673031856531)*0.5**o
        ref[1,1,2,0]=(0.255023909903)/(o+1.)+(0.229656917655)+(0.22400044344)*0.5**o
        ref[1,1,2,1]=(0.518856765199)/(o+1.)+(0.380552790125)+(0.543298381396)*0.5**o
        ref[1,2,0,0]=(0.0321563884831)/(o+1.)+(0.25981567762)+(-0.0316496900482)*0.5**o
        ref[1,2,0,1]=(1.09318176771)/(o+1.)+(0.0572442907603)+(0.564519446305)*0.5**o
        ref[1,2,1,0]=(0.756962791387)/(o+1.)+(0.69419429624)+(0.32401997528)*0.5**o
        ref[1,2,1,1]=(1.28480299241)/(o+1.)+(-0.201990593541)+(-0.905711635518)*0.5**o
        ref[1,2,2,0]=(1.04969566442)/(o+1.)+(0.72854097372)+(0.592972696101)*0.5**o
        ref[1,2,2,1]=(-1.77086264775)/(o+1.)+(-0.261914180361)+(0.0440022926066)*0.5**o
        ref[1,3,0,0]=(-0.477780531124)/(o+1.)+(-0.181354085163)+(-0.300288699967)*0.5**o
        ref[1,3,0,1]=(-0.797627013456)/(o+1.)+(-0.523163955197)+(0.347574883435)*0.5**o
        ref[1,3,1,0]=(0.428806106821)/(o+1.)+(-0.0514584302394)+(-0.655980330052)*0.5**o
        ref[1,3,1,1]=(-0.141740459457)/(o+1.)+(-1.05699960931)+(-0.821242527644)*0.5**o
        ref[1,3,2,0]=(0.0181810985873)/(o+1.)+(-0.180697139553)+(0.41036925874)*0.5**o
        ref[1,3,2,1]=(-0.061077367)/(o+1.)+(0.250077446763)+(0.157102380812)*0.5**o
        ref[1,4,0,0]=(1.14087212913)/(o+1.)+(0.368059029864)+(0.659728641463)*0.5**o
        ref[1,4,0,1]=(0.365658713316)/(o+1.)+(0.088533821044)+(-0.584417938734)*0.5**o
        ref[1,4,1,0]=(-1.11681308321)/(o+1.)+(0.438275560627)+(0.622761672604)*0.5**o
        ref[1,4,1,1]=(1.18893372018)/(o+1.)+(0.172934240923)+(-0.172046824999)*0.5**o
        ref[1,4,2,0]=(0.16240489493)/(o+1.)+(-0.0145923852256)+(0.834478471749)*0.5**o
        ref[1,4,2,1]=(0.571355829378)/(o+1.)+(0.526896417749)+(0.458073367663)*0.5**o
        ref[2,0,0,0]=(-0.489234572937)/(o+1.)+(-0.0135589093901)+(0.65140309615)*0.5**o
        ref[2,0,0,1]=(-1.55383491931)/(o+1.)+(0.0161644619972)+(0.703425613366)*0.5**o
        ref[2,0,1,0]=(-0.258604855723)/(o+1.)+(0.0827812678671)+(0.507722104457)*0.5**o
        ref[2,0,1,1]=(0.0504178529481)/(o+1.)+(0.182991850332)+(0.359204920726)*0.5**o
        ref[2,0,2,0]=(0.652043479841)/(o+1.)+(-0.443015450013)+(-0.548443848248)*0.5**o
        ref[2,0,2,1]=(0.340317839857)/(o+1.)+(-0.379565594545)+(0.663652225917)*0.5**o
        ref[2,1,0,0]=(-0.587863196667)/(o+1.)+(-0.0137797674284)+(-0.974202340562)*0.5**o
        ref[2,1,0,1]=(1.00030747039)/(o+1.)+(-0.399997754768)+(0.755747639771)*0.5**o
        ref[2,1,1,0]=(0.703413281036)/(o+1.)+(0.487016208736)+(-0.877007540039)*0.5**o
        ref[2,1,1,1]=(0.263166830572)/(o+1.)+(-0.157259452849)+(0.666747865841)*0.5**o
        ref[2,1,2,0]=(0.104532343004)/(o+1.)+(-1.02548001493)+(-0.486314104701)*0.5**o
        ref[2,1,2,1]=(0.933518655478)/(o+1.)+(0.0369928229139)+(-0.244085626298)*0.5**o
        ref[2,2,0,0]=(-1.50550384519)/(o+1.)+(-0.590996227594)+(-0.570130630622)*0.5**o
        ref[2,2,0,1]=(-0.60318683839)/(o+1.)+(-0.38308245528)+(-0.502979448853)*0.5**o
        ref[2,2,1,0]=(-1.11258910879)/(o+1.)+(-0.518777277969)+(-0.999685897413)*0.5**o
        ref[2,2,1,1]=(1.84366617107)/(o+1.)+(-0.232866738261)+(-0.622157810558)*0.5**o
        ref[2,2,2,0]=(-0.620181988585)/(o+1.)+(-0.836162444183)+(0.267939886396)*0.5**o
        ref[2,2,2,1]=(-0.826647389099)/(o+1.)+(0.0929536518109)+(0.905694268362)*0.5**o
        ref[2,3,0,0]=(0.184067674443)/(o+1.)+(0.259745858757)+(0.724759799909)*0.5**o
        ref[2,3,0,1]=(0.0175595303049)/(o+1.)+(0.0560117358912)+(-0.907162868639)*0.5**o
        ref[2,3,1,0]=(-0.609443154617)/(o+1.)+(0.186540819401)+(-0.951152598958)*0.5**o
        ref[2,3,1,1]=(-1.61858082094)/(o+1.)+(0.679439858578)+(0.927423354971)*0.5**o
        ref[2,3,2,0]=(0.44350652486)/(o+1.)+(0.81466177481)+(0.53899353207)*0.5**o
        ref[2,3,2,1]=(-0.877289862245)/(o+1.)+(-1.05742260849)+(-0.462176941342)*0.5**o
        ref[2,4,0,0]=(-0.736381416539)/(o+1.)+(-0.0139374259612)+(-0.0755289501641)*0.5**o
        ref[2,4,0,1]=(0.690429326999)/(o+1.)+(-0.0293644018251)+(-0.328756862036)*0.5**o
        ref[2,4,1,0]=(-1.57117034002)/(o+1.)+(-0.531463765387)+(0.553349645561)*0.5**o
        ref[2,4,1,1]=(1.24898175282)/(o+1.)+(-0.452273024959)+(-0.515691967125)*0.5**o
        ref[2,4,2,0]=(0.239126191643)/(o+1.)+(0.722504563932)+(0.547637159392)*0.5**o
        ref[2,4,2,1]=(-1.40763254209)/(o+1.)+(-0.809456441012)+(0.932852295739)*0.5**o
        ref[3,0,0,0]=(-0.416554297686)/(o+1.)+(-0.632193610911)+(-0.304105530258)*0.5**o
        ref[3,0,0,1]=(-1.02983448036)/(o+1.)+(0.579212231778)+(-0.436532313966)*0.5**o
        ref[3,0,1,0]=(0.42124133216)/(o+1.)+(0.368290418827)+(-0.920395942436)*0.5**o
        ref[3,0,1,1]=(0.158195474854)/(o+1.)+(-0.17459474514)+(-0.830644214826)*0.5**o
        ref[3,0,2,0]=(-1.18749647329)/(o+1.)+(-0.506173996178)+(-0.513857022748)*0.5**o
        ref[3,0,2,1]=(0.233437753018)/(o+1.)+(0.455705145371)+(-0.323239005586)*0.5**o
        ref[3,1,0,0]=(1.42991958566)/(o+1.)+(0.053769921022)+(0.150959272234)*0.5**o
        ref[3,1,0,1]=(1.48423731111)/(o+1.)+(0.964048315997)+(-0.689863420196)*0.5**o
        ref[3,1,1,0]=(0.868837676765)/(o+1.)+(-1.01261543254)+(0.165878933098)*0.5**o
        ref[3,1,1,1]=(0.820750307831)/(o+1.)+(-0.300930223821)+(0.396516195659)*0.5**o
        ref[3,1,2,0]=(0.477077782202)/(o+1.)+(-0.0416336989494)+(0.769016797571)*0.5**o
        ref[3,1,2,1]=(0.359895440587)/(o+1.)+(0.555322395811)+(0.944871585683)*0.5**o
        ref[3,2,0,0]=(0.657668315832)/(o+1.)+(-0.240321895174)+(-0.286298603451)*0.5**o
        ref[3,2,0,1]=(-0.941360568282)/(o+1.)+(-0.0246544136593)+(-0.181845734439)*0.5**o
        ref[3,2,1,0]=(1.60316834971)/(o+1.)+(0.252291602298)+(0.988555915005)*0.5**o
        ref[3,2,1,1]=(0.833811957678)/(o+1.)+(-0.111970621871)+(0.634808834458)*0.5**o
        ref[3,2,2,0]=(-1.38877073784)/(o+1.)+(0.538356636307)+(-0.249568730877)*0.5**o
        ref[3,2,2,1]=(-0.311839184097)/(o+1.)+(-0.698114103264)+(0.169190433329)*0.5**o
        ref[3,3,0,0]=(0.743170989696)/(o+1.)+(-0.499224746788)+(0.0699956684628)*0.5**o
        ref[3,3,0,1]=(-0.293538786051)/(o+1.)+(0.26667992866)+(0.313940467768)*0.5**o
        ref[3,3,1,0]=(0.951483461345)/(o+1.)+(0.148317904561)+(0.00242640905887)*0.5**o
        ref[3,3,1,1]=(-0.110753912776)/(o+1.)+(-0.44308689138)+(0.980634836805)*0.5**o
        ref[3,3,2,0]=(-0.337895223544)/(o+1.)+(0.358941646037)+(0.264839226057)*0.5**o
        ref[3,3,2,1]=(0.857823611315)/(o+1.)+(0.30404611512)+(-0.898657291359)*0.5**o
        ref[3,4,0,0]=(0.037944509376)/(o+1.)+(-0.364801700413)+(-0.349756182268)*0.5**o
        ref[3,4,0,1]=(1.01114399396)/(o+1.)+(0.457103279485)+(0.631738783076)*0.5**o
        ref[3,4,1,0]=(0.969931913888)/(o+1.)+(-0.165922085756)+(0.998943588321)*0.5**o
        ref[3,4,1,1]=(1.35441114248)/(o+1.)+(-0.502422727199)+(0.00430234808496)*0.5**o
        ref[3,4,2,0]=(-0.367201097566)/(o+1.)+(-0.337780780531)+(0.688589316866)*0.5**o
        ref[3,4,2,1]=(-0.0547905846061)/(o+1.)+(-0.293870420685)+(0.31527404661)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.241972460279)*x[0]+(0.539729925673)*x[1]
        ref=(-0.11832904353)/(o+1.)+(0.649911364492)+(-0.399791299503)*0.5**o
      else:
        arg=(0.53919797656)*x[0]+(0.379010664884)*x[1]+(0.274416064529)*x[2]
        ref=(0.307947301472)/(o+1.)+(0.120674157748)+(0.643329089005)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.84849017056)*x[0]+(-0.161962845202)*x[1]
        arg[1]=(-0.794717919135)*x[0]+(1.16003773358)*x[1]
        ref[0]=(0.780514448202)/(o+1.)+(-0.34628958021)+(0.598592037576)*0.5**o
        ref[1]=(0.570158156278)/(o+1.)+(0.241192316704)+(-0.687222975237)*0.5**o
      else:
        arg[0]=(-0.637352163634)*x[0]+(1.49477462643)*x[1]+(-0.284018740238)*x[2]
        arg[1]=(0.125937459904)*x[0]+(-0.206499869891)*x[1]+(-0.342856240345)*x[2]
        ref[0]=(-0.274486314451)/(o+1.)+(0.379737553853)+(0.0884149293071)*0.5**o
        ref[1]=(-0.78452814021)/(o+1.)+(-0.122017410137)+(0.605144310151)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.440868762581)*x[0]+(-0.216223666616)*x[1]
        arg[0,1]=(-1.78807355591)*x[0]+(-0.208278123838)*x[1]
        arg[0,2]=(-0.448066998103)*x[0]+(-0.819815507826)*x[1]
        arg[0,3]=(1.27247881689)*x[0]+(0.0653439567221)*x[1]
        arg[0,4]=(-1.00344990223)*x[0]+(-0.291263790583)*x[1]
        arg[1,0]=(0.440080423838)*x[0]+(0.0356836558113)*x[1]
        arg[1,1]=(-1.49545340091)*x[0]+(-0.133615590211)*x[1]
        arg[1,2]=(1.05947638385)*x[0]+(0.408685904238)*x[1]
        arg[1,3]=(-0.170309086605)*x[0]+(0.660523970285)*x[1]
        arg[1,4]=(0.764323678475)*x[0]+(-0.529415080303)*x[1]
        arg[2,0]=(0.55270081325)*x[0]+(-1.60630146231)*x[1]
        arg[2,1]=(1.57419330248)*x[0]+(-1.24644288706)*x[1]
        arg[2,2]=(0.0416757238363)*x[0]+(0.0247379738465)*x[1]
        arg[2,3]=(1.0200370393)*x[0]+(0.364209137693)*x[1]
        arg[2,4]=(-1.21858490389)*x[0]+(0.523056254546)*x[1]
        arg[3,0]=(0.0524205820842)*x[0]+(1.65953217854)*x[1]
        arg[3,1]=(1.21824883395)*x[0]+(0.905190834998)*x[1]
        arg[3,2]=(0.226344434525)*x[0]+(-0.735349531462)*x[1]
        arg[3,3]=(-1.00295245717)*x[0]+(-0.719605376499)*x[1]
        arg[3,4]=(-0.446228759693)*x[0]+(1.23441974277)*x[1]
        ref[0,0]=(0.347328582602)/(o+1.)+(-0.520484463519)+(0.0365479152384)*0.5**o
        ref[0,1]=(-0.274564442352)/(o+1.)+(-0.413557210678)+(-0.894672816039)*0.5**o
        ref[0,2]=(0.0720644688954)/(o+1.)+(-0.906636966188)+(0.473326957552)*0.5**o
        ref[0,3]=(-0.747749341777)/(o+1.)+(0.730249972378)+(0.625072170633)*0.5**o
        ref[0,4]=(-0.194031116169)/(o+1.)+(-0.432063325829)+(-0.236555924983)*0.5**o
        ref[1,0]=(0.807081538033)/(o+1.)+(-0.612266650007)+(0.893215841629)*0.5**o
        ref[1,1]=(-0.662516518957)/(o+1.)+(-0.176318925886)+(-0.61391462039)*0.5**o
        ref[1,2]=(-0.570345636156)/(o+1.)+(0.791877954963)+(0.454752014319)*0.5**o
        ref[1,3]=(-0.312400910795)/(o+1.)+(0.212935137466)+(0.376745519544)*0.5**o
        ref[1,4]=(-0.764680414264)/(o+1.)+(0.174215486579)+(0.651158039278)*0.5**o
        ref[2,0]=(-0.837942771642)/(o+1.)+(-0.187938486702)+(0.160219095985)*0.5**o
        ref[2,1]=(-0.672421039761)/(o+1.)+(0.17384002882)+(0.652491397549)*0.5**o
        ref[2,2]=(-0.656187949922)/(o+1.)+(0.559888736121)+(-0.397175824637)*0.5**o
        ref[2,3]=(0.259558368195)/(o+1.)+(0.55148644978)+(0.0217149092406)*0.5**o
        ref[2,4]=(-0.188397537901)/(o+1.)+(-0.0561706170512)+(-0.394789877345)*0.5**o
        ref[3,0]=(0.91055628617)/(o+1.)+(0.574211274098)+(-0.347026073739)*0.5**o
        ref[3,1]=(0.29225509237)/(o+1.)+(0.64447705482)+(0.542230466941)*0.5**o
        ref[3,2]=(0.121065624384)/(o+1.)+(0.0316495277836)+(-0.693369776888)*0.5**o
        ref[3,3]=(-0.943523334976)/(o+1.)+(-0.266302644031)+(-0.246429210629)*0.5**o
        ref[3,4]=(0.714915399455)/(o+1.)+(0.459963131994)+(-0.846650680369)*0.5**o
      else:
        arg[0,0]=(-0.493869713454)*x[0]+(-0.381969413515)*x[1]+(1.59862368848)*x[2]
        arg[0,1]=(-0.581245543547)*x[0]+(0.961117271889)*x[1]+(-0.158073515343)*x[2]
        arg[0,2]=(-0.635283143641)*x[0]+(0.695573236987)*x[1]+(-0.719180832278)*x[2]
        arg[0,3]=(-0.737780596733)*x[0]+(-1.11171283352)*x[1]+(0.459954346315)*x[2]
        arg[0,4]=(1.12522065074)*x[0]+(-1.08427904854)*x[1]+(-0.715679673556)*x[2]
        arg[1,0]=(0.972468186181)*x[0]+(-0.211014530435)*x[1]+(-0.734414099002)*x[2]
        arg[1,1]=(-1.67337704748)*x[0]+(0.871096213089)*x[1]+(1.51063313223)*x[2]
        arg[1,2]=(-0.222013007834)*x[0]+(1.46474508326)*x[1]+(-0.979243908862)*x[2]
        arg[1,3]=(1.21195281087)*x[0]+(0.274082389329)*x[1]+(1.56144436465)*x[2]
        arg[1,4]=(1.35812865884)*x[0]+(1.25819798853)*x[1]+(0.895334048395)*x[2]
        arg[2,0]=(-0.167404344437)*x[0]+(0.0710591934315)*x[1]+(-0.124469683665)*x[2]
        arg[2,1]=(-0.492240632996)*x[0]+(-0.317251117382)*x[1]+(0.0195461703529)*x[2]
        arg[2,2]=(-0.277590266225)*x[0]+(-1.65859476899)*x[1]+(-0.0100216405668)*x[2]
        arg[2,3]=(-0.0701281298455)*x[0]+(0.502137254703)*x[1]+(0.701220923443)*x[2]
        arg[2,4]=(-0.72186096407)*x[0]+(-0.900487568179)*x[1]+(-0.735573922826)*x[2]
        arg[3,0]=(-0.0677324474129)*x[0]+(-0.649963209718)*x[1]+(-0.770861099726)*x[2]
        arg[3,1]=(0.563588189481)*x[0]+(0.714614932251)*x[1]+(-0.0957764632886)*x[2]
        arg[3,2]=(1.13337395062)*x[0]+(0.507556316073)*x[1]+(-0.196759085949)*x[2]
        arg[3,3]=(-1.2636883168)*x[0]+(-0.811660344076)*x[1]+(0.645462052598)*x[2]
        arg[3,4]=(-0.104175186987)*x[0]+(-0.929176038155)*x[1]+(1.07530379856)*x[2]
        ref[0,0]=(0.84538674973)/(o+1.)+(0.0569713556632)+(-0.236544899547)*0.5**o
        ref[0,1]=(-0.307689589699)/(o+1.)+(0.245040337642)+(0.0394071274131)*0.5**o
        ref[0,2]=(0.955244736357)/(o+1.)+(-0.341771136504)+(-0.930593202281)*0.5**o
        ref[0,3]=(-0.353490461722)/(o+1.)+(-0.476228433901)+(-0.083591754415)*0.5**o
        ref[0,4]=(-1.31220485482)/(o+1.)+(0.0466426656676)+(0.544181452128)*0.5**o
        ref[1,0]=(0.22773475414)/(o+1.)+(-0.488454801654)+(0.776214405914)*0.5**o
        ref[1,1]=(1.50688678707)/(o+1.)+(0.0851415404695)+(-0.96881757017)*0.5**o
        ref[1,2]=(0.313471097102)/(o+1.)+(-0.0875594518156)+(0.125135973096)*0.5**o
        ref[1,3]=(0.269521220179)/(o+1.)+(1.1192007195)+(0.539556905661)*0.5**o
        ref[1,4]=(1.91499876064)/(o+1.)+(0.509063797893)+(0.578534339343)*0.5**o
        ref[2,0]=(0.941352651562)/(o+1.)+(-0.966199402186)+(0.770231318139)*0.5**o
        ref[2,1]=(-0.141490823562)/(o+1.)+(-0.20919987022)+(-0.230055016023)*0.5**o
        ref[2,2]=(-0.916148047971)/(o+1.)+(-0.371740600192)+(-0.286577427426)*0.5**o
        ref[2,3]=(0.229516219331)/(o+1.)+(0.274790532736)+(0.354132763498)*0.5**o
        ref[2,4]=(-1.19120612619)/(o+1.)+(-0.139859859385)+(-0.88699661011)*0.5**o
        ref[3,0]=(-1.48630121465)/(o+1.)+(0.0744166251063)+(-0.151088792417)*0.5**o
        ref[3,1]=(-0.335583560513)/(o+1.)+(0.277436410841)+(0.963137397276)*0.5**o
        ref[3,2]=(-0.673942688644)/(o+1.)+(0.70077509905)+(0.716563671283)*0.5**o
        ref[3,3]=(-0.148394092795)/(o+1.)+(-0.237190239072)+(-0.807112037336)*0.5**o
        ref[3,4]=(0.529983886211)/(o+1.)+(-0.371055083588)+(0.254078854382)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.661725715165)*x[0]+(0.219720717581)*x[1]
        arg[0,0,1]=(-0.642240082304)*x[0]+(-0.279652756776)*x[1]
        arg[0,1,0]=(0.0204235813151)*x[0]+(-0.374873249933)*x[1]
        arg[0,1,1]=(1.4418481252)*x[0]+(0.205656834772)*x[1]
        arg[1,0,0]=(0.65992179951)*x[0]+(-1.04887875458)*x[1]
        arg[1,0,1]=(-0.195480216199)*x[0]+(-0.373905275039)*x[1]
        arg[1,1,0]=(0.781251543352)*x[0]+(0.0416248109536)*x[1]
        arg[1,1,1]=(-0.925751319189)*x[0]+(-0.113992910046)*x[1]
        arg[2,0,0]=(-0.0767414065227)*x[0]+(-0.156682840973)*x[1]
        arg[2,0,1]=(0.688457272529)*x[0]+(0.524916336046)*x[1]
        arg[2,1,0]=(0.753722830084)*x[0]+(-0.890325839466)*x[1]
        arg[2,1,1]=(0.730789290145)*x[0]+(-0.278986703716)*x[1]
        arg[3,0,0]=(-0.287376796439)*x[0]+(-0.153987233224)*x[1]
        arg[3,0,1]=(-1.15687435255)*x[0]+(0.250005890238)*x[1]
        arg[3,1,0]=(-0.845312830629)*x[0]+(0.256211962352)*x[1]
        arg[3,1,1]=(-0.574311820519)*x[0]+(-0.440297035998)*x[1]
        arg[4,0,0]=(-0.360607647308)*x[0]+(-1.29380437557)*x[1]
        arg[4,0,1]=(1.32019433485)*x[0]+(0.295255956618)*x[1]
        arg[4,1,0]=(-1.64162941972)*x[0]+(0.117482372162)*x[1]
        arg[4,1,1]=(-0.303103646619)*x[0]+(-0.88458982681)*x[1]
        arg[5,0,0]=(-0.666761879038)*x[0]+(-0.735786278877)*x[1]
        arg[5,0,1]=(0.0892896382617)*x[0]+(-0.45478872773)*x[1]
        arg[5,1,0]=(-0.50985624415)*x[0]+(-0.400022939567)*x[1]
        arg[5,1,1]=(1.39240706368)*x[0]+(0.0149740822664)*x[1]
        ref[0,0,0]=(-0.114834538488)/(o+1.)+(0.240621006484)+(0.515038958264)*0.5**o
        ref[0,0,1]=(0.588698337077)/(o+1.)+(-0.702256377233)+(-0.10607842169)*0.5**o
        ref[0,1,0]=(0.13849818224)/(o+1.)+(-0.277555277991)+(0.0621627051247)*0.5**o
        ref[0,1,1]=(-0.0878009521163)/(o+1.)+(0.630168348947)+(0.474969214192)*0.5**o
        ref[1,0,0]=(-0.696600538111)/(o+1.)+(-0.343318260912)+(0.994280104861)*0.5**o
        ref[1,0,1]=(-0.0887026688527)/(o+1.)+(-0.385753227979)+(0.290823633572)*0.5**o
        ref[1,1,0]=(0.831408608956)/(o+1.)+(-0.242985671853)+(0.477439089057)*0.5**o
        ref[1,1,1]=(-0.621059053191)/(o+1.)+(-0.0700203481241)+(-0.278644479797)*0.5**o
        ref[2,0,0]=(0.7890720399)/(o+1.)+(-0.539721196393)+(0.0569461053905)*0.5**o
        ref[2,0,1]=(-0.228628151401)/(o+1.)+(0.826946074171)+(-0.211890388366)*0.5**o
        ref[2,1,0]=(-0.837617325493)/(o+1.)+(0.289656653751)+(0.121701008608)*0.5**o
        ref[2,1,1]=(0.246577988625)/(o+1.)+(0.0512227634376)+(0.102779070928)*0.5**o
        ref[3,0,0]=(-0.0109584684212)/(o+1.)+(-0.334888666171)+(0.239371771102)*0.5**o
        ref[3,0,1]=(-0.125987676419)/(o+1.)+(-0.170179004032)+(-0.440522777827)*0.5**o
        ref[3,1,0]=(-0.354093141636)/(o+1.)+(-0.177294275919)+(0.119580825198)*0.5**o
        ref[3,1,1]=(-0.746761075412)/(o+1.)+(0.192068760731)+(-0.651985302566)*0.5**o
        ref[4,0,0]=(-0.42411673219)/(o+1.)+(-0.579162944937)+(-0.0719694008158)*0.5**o
        ref[4,0,1]=(0.169881169584)/(o+1.)+(0.483614029132)+(0.478341063616)*0.5**o
        ref[4,1,0]=(0.943567836264)/(o+1.)+(-0.898782941264)+(-0.670149001297)*0.5**o
        ref[4,1,1]=(-0.703610807936)/(o+1.)+(0.225342628657)+(-0.934767922806)*0.5**o
        ref[5,0,0]=(-0.788792225213)/(o+1.)+(-0.1768857093)+(-0.259984514102)*0.5**o
        ref[5,0,1]=(-0.788708879899)/(o+1.)+(-0.202288491141)+(0.827786772711)*0.5**o
        ref[5,1,0]=(-0.823081809707)/(o+1.)+(0.274720963093)+(-0.636239300197)*0.5**o
        ref[5,1,1]=(0.498922549807)/(o+1.)+(0.0129388257746)+(0.882580944588)*0.5**o
      else:
        arg[0,0,0]=(-0.426613654844)*x[0]+(-1.01433980801)*x[1]+(0.564621874916)*x[2]
        arg[0,0,1]=(-0.438521968882)*x[0]+(1.6033775267)*x[1]+(0.691349995832)*x[2]
        arg[0,1,0]=(-0.04337513731)*x[0]+(0.142927668234)*x[1]+(0.506581723199)*x[2]
        arg[0,1,1]=(-0.120598868162)*x[0]+(-0.213666787813)*x[1]+(-1.04221679332)*x[2]
        arg[1,0,0]=(0.153696490712)*x[0]+(-0.122516529815)*x[1]+(-1.08837272712)*x[2]
        arg[1,0,1]=(0.410706464237)*x[0]+(0.720079363315)*x[1]+(-0.767790357539)*x[2]
        arg[1,1,0]=(-1.03726428173)*x[0]+(-0.316809853003)*x[1]+(-0.780405868082)*x[2]
        arg[1,1,1]=(-0.427547967216)*x[0]+(-0.897766643609)*x[1]+(-1.56913552404)*x[2]
        arg[2,0,0]=(-0.282789248898)*x[0]+(-0.110846936031)*x[1]+(-1.4329391906)*x[2]
        arg[2,0,1]=(1.46866588617)*x[0]+(-1.23022501959)*x[1]+(-0.3123390018)*x[2]
        arg[2,1,0]=(-0.314115718621)*x[0]+(1.7333934438)*x[1]+(0.833572700726)*x[2]
        arg[2,1,1]=(0.464246462638)*x[0]+(0.267484068646)*x[1]+(0.342009204787)*x[2]
        arg[3,0,0]=(0.578853099163)*x[0]+(-0.0453855508465)*x[1]+(-0.0751578475516)*x[2]
        arg[3,0,1]=(-0.526734485752)*x[0]+(-1.03286310016)*x[1]+(-0.398986583917)*x[2]
        arg[3,1,0]=(0.0239088277798)*x[0]+(0.205105470192)*x[1]+(-1.04986349166)*x[2]
        arg[3,1,1]=(0.130893744001)*x[0]+(-0.273702140669)*x[1]+(-0.0804222415128)*x[2]
        arg[4,0,0]=(-0.822661618584)*x[0]+(-0.267189204519)*x[1]+(-0.290013766971)*x[2]
        arg[4,0,1]=(-1.46759823734)*x[0]+(0.317465945316)*x[1]+(0.108284317752)*x[2]
        arg[4,1,0]=(-1.3920990386)*x[0]+(-1.17823778694)*x[1]+(-0.18366434307)*x[2]
        arg[4,1,1]=(-0.0915496209516)*x[0]+(-0.367624609832)*x[1]+(1.30630257323)*x[2]
        arg[5,0,0]=(0.117840830455)*x[0]+(-0.667875433194)*x[1]+(-1.59716688455)*x[2]
        arg[5,0,1]=(-0.818693227605)*x[0]+(0.483842075248)*x[1]+(-0.408373205535)*x[2]
        arg[5,1,0]=(1.36906770301)*x[0]+(0.494926179999)*x[1]+(0.717041707262)*x[2]
        arg[5,1,1]=(0.819757116081)*x[0]+(0.0982458749015)*x[1]+(0.275080391599)*x[2]
        ref[0,0,0]=(0.236560699949)/(o+1.)+(-0.392784388084)+(-0.327323511722)*0.5**o
        ref[0,0,1]=(1.32491522045)/(o+1.)+(0.525146863711)+(-0.519003394222)*0.5**o
        ref[0,1,0]=(0.433749865563)/(o+1.)+(0.23007126008)+(-0.287758131599)*0.5**o
        ref[0,1,1]=(-0.0563021060686)/(o+1.)+(-0.403921160703)+(-0.512338021825)*0.5**o
        ref[1,0,0]=(-0.170496965963)/(o+1.)+(-0.914421777259)+(0.942147754258)*0.5**o
        ref[1,0,1]=(-0.372707483726)/(o+1.)+(0.636041158872)+(-0.536379364005)*0.5**o
        ref[1,1,0]=(-0.211529878364)/(o+1.)+(-0.912151985776)+(-0.0986461529033)*0.5**o
        ref[1,1,1]=(-1.32533998437)/(o+1.)+(-0.685030198386)+(-0.19904975372)*0.5**o
        ref[2,0,0]=(0.152048902469)/(o+1.)+(-0.88880590209)+(-0.201012473813)*0.5**o
        ref[2,0,1]=(-0.748900052246)/(o+1.)+(-0.0226120471998)+(0.720226011425)*0.5**o
        ref[2,1,0]=(0.77797707941)/(o+1.)+(1.01635055442)+(-0.557827762351)*0.5**o
        ref[2,1,1]=(1.08969334716)/(o+1.)+(-0.404644728844)+(0.793335846595)*0.5**o
        ref[3,0,0]=(1.56498448669)/(o+1.)+(-0.699967188685)+(0.293259591449)*0.5**o
        ref[3,0,1]=(-0.620133427128)/(o+1.)+(-0.550037413688)+(-0.238375915329)*0.5**o
        ref[3,1,0]=(-0.156380586288)/(o+1.)+(-0.0154369973565)+(-0.633594612687)*0.5**o
        ref[3,1,1]=(-0.352989559326)/(o+1.)+(-0.290547031627)+(0.710852984398)*0.5**o
        ref[4,0,0]=(-0.147579361355)/(o+1.)+(-0.615173424765)+(-0.00193837918766)*0.5**o
        ref[4,0,1]=(0.598152345277)/(o+1.)+(-0.540248606426)+(-0.559503106696)*0.5**o
        ref[4,1,0]=(-1.17182448825)/(o+1.)+(-0.590551681576)+(-0.401073317208)*0.5**o
        ref[4,1,1]=(0.837504141173)/(o+1.)+(0.0807414267186)+(-0.15185865216)*0.5**o
        ref[5,0,0]=(-0.646601115533)/(o+1.)+(-1.20895265273)+(0.9173049337)*0.5**o
        ref[5,0,1]=(1.26626110052)/(o+1.)+(-0.56471923973)+(-0.880046978955)*0.5**o
        ref[5,1,0]=(-0.069063169298)/(o+1.)+(0.883536796964)+(0.883025165644)*0.5**o
        ref[5,1,1]=(1.05999338993)/(o+1.)+(-0.392519102893)+(0.918128198439)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.257898226293)*x[0]+(-0.01230083314)*x[1]
        arg[0,0,0,1]=(0.0531183070647)*x[0]+(-0.170192287836)*x[1]
        arg[0,0,1,0]=(-0.364644940864)*x[0]+(1.04227651446)*x[1]
        arg[0,0,1,1]=(-0.894334108203)*x[0]+(0.151390331211)*x[1]
        arg[0,0,2,0]=(1.03825401708)*x[0]+(-0.882136097149)*x[1]
        arg[0,0,2,1]=(-1.05304910155)*x[0]+(1.06866311708)*x[1]
        arg[0,1,0,0]=(1.54742542159)*x[0]+(0.941778348482)*x[1]
        arg[0,1,0,1]=(1.1797380956)*x[0]+(-0.296417138234)*x[1]
        arg[0,1,1,0]=(-1.09196028027)*x[0]+(-0.783377624006)*x[1]
        arg[0,1,1,1]=(0.370952191443)*x[0]+(-0.0854141858991)*x[1]
        arg[0,1,2,0]=(-1.23552368139)*x[0]+(1.23848088087)*x[1]
        arg[0,1,2,1]=(0.216345350114)*x[0]+(-0.136524468427)*x[1]
        arg[0,2,0,0]=(-1.21206901263)*x[0]+(0.804219582144)*x[1]
        arg[0,2,0,1]=(0.192956526208)*x[0]+(0.588417678858)*x[1]
        arg[0,2,1,0]=(-0.940175603742)*x[0]+(0.329047260784)*x[1]
        arg[0,2,1,1]=(-0.138597387792)*x[0]+(0.398298326223)*x[1]
        arg[0,2,2,0]=(0.10089766824)*x[0]+(1.25459493298)*x[1]
        arg[0,2,2,1]=(-0.401163525883)*x[0]+(0.276542765282)*x[1]
        arg[0,3,0,0]=(0.252156795268)*x[0]+(-0.50223063539)*x[1]
        arg[0,3,0,1]=(0.466244610905)*x[0]+(-0.587025224301)*x[1]
        arg[0,3,1,0]=(-0.268082012713)*x[0]+(-1.28874675764)*x[1]
        arg[0,3,1,1]=(1.10867463519)*x[0]+(-0.165500912864)*x[1]
        arg[0,3,2,0]=(-1.94681908266)*x[0]+(0.166854108476)*x[1]
        arg[0,3,2,1]=(-0.537150417922)*x[0]+(0.0872851457131)*x[1]
        arg[0,4,0,0]=(-0.376548581807)*x[0]+(-0.245201133388)*x[1]
        arg[0,4,0,1]=(-0.135287690767)*x[0]+(0.750376238942)*x[1]
        arg[0,4,1,0]=(-0.0240491596907)*x[0]+(-1.21337088424)*x[1]
        arg[0,4,1,1]=(0.699215733397)*x[0]+(0.0596356032853)*x[1]
        arg[0,4,2,0]=(-0.521316010702)*x[0]+(-1.75309802615)*x[1]
        arg[0,4,2,1]=(0.403640172305)*x[0]+(1.41465827077)*x[1]
        arg[1,0,0,0]=(-1.01538089315)*x[0]+(-1.14785578909)*x[1]
        arg[1,0,0,1]=(1.10045201832)*x[0]+(0.630761728795)*x[1]
        arg[1,0,1,0]=(0.137673646147)*x[0]+(-0.387129090253)*x[1]
        arg[1,0,1,1]=(-1.04588287863)*x[0]+(-1.22433029859)*x[1]
        arg[1,0,2,0]=(-0.567830263535)*x[0]+(-0.405808616295)*x[1]
        arg[1,0,2,1]=(0.139326102495)*x[0]+(-1.93153384363)*x[1]
        arg[1,1,0,0]=(-0.638524762921)*x[0]+(0.506405751653)*x[1]
        arg[1,1,0,1]=(0.130775153144)*x[0]+(0.531317202208)*x[1]
        arg[1,1,1,0]=(0.557476206038)*x[0]+(0.447588676184)*x[1]
        arg[1,1,1,1]=(-0.538795916875)*x[0]+(-0.593869493531)*x[1]
        arg[1,1,2,0]=(-1.09827483942)*x[0]+(0.376431061876)*x[1]
        arg[1,1,2,1]=(-0.0074505620387)*x[0]+(-0.798645151486)*x[1]
        arg[1,2,0,0]=(0.251283911459)*x[0]+(0.194595071813)*x[1]
        arg[1,2,0,1]=(1.00482763799)*x[0]+(-1.00059863394)*x[1]
        arg[1,2,1,0]=(-0.453095434684)*x[0]+(-0.513922011418)*x[1]
        arg[1,2,1,1]=(0.558548837484)*x[0]+(0.0696328406825)*x[1]
        arg[1,2,2,0]=(-1.20625552545)*x[0]+(-0.151371710692)*x[1]
        arg[1,2,2,1]=(0.100262209656)*x[0]+(-0.314628804179)*x[1]
        arg[1,3,0,0]=(0.947063141882)*x[0]+(-0.297271200562)*x[1]
        arg[1,3,0,1]=(0.0760277251261)*x[0]+(-1.17982146387)*x[1]
        arg[1,3,1,0]=(-1.46946063969)*x[0]+(-0.615302663567)*x[1]
        arg[1,3,1,1]=(0.419747293729)*x[0]+(-0.629842531949)*x[1]
        arg[1,3,2,0]=(-1.44607954917)*x[0]+(0.520024093924)*x[1]
        arg[1,3,2,1]=(-1.01556392596)*x[0]+(-0.307173811354)*x[1]
        arg[1,4,0,0]=(-0.164339335244)*x[0]+(0.390736414982)*x[1]
        arg[1,4,0,1]=(-0.266932380288)*x[0]+(-0.731533848078)*x[1]
        arg[1,4,1,0]=(-0.903825489055)*x[0]+(-1.81564784421)*x[1]
        arg[1,4,1,1]=(-0.426008882665)*x[0]+(-1.09792948908)*x[1]
        arg[1,4,2,0]=(-1.32984927589)*x[0]+(1.58196302541)*x[1]
        arg[1,4,2,1]=(-0.0581032806145)*x[0]+(-0.323073355302)*x[1]
        arg[2,0,0,0]=(-0.744014251373)*x[0]+(-0.818253254872)*x[1]
        arg[2,0,0,1]=(0.368656965396)*x[0]+(-0.0864646380237)*x[1]
        arg[2,0,1,0]=(-0.112428845773)*x[0]+(0.109170104421)*x[1]
        arg[2,0,1,1]=(0.0682241548911)*x[0]+(-1.65999482283)*x[1]
        arg[2,0,2,0]=(0.212777058993)*x[0]+(0.127163480945)*x[1]
        arg[2,0,2,1]=(-0.256989443356)*x[0]+(0.183206687414)*x[1]
        arg[2,1,0,0]=(0.578997994373)*x[0]+(0.413087453689)*x[1]
        arg[2,1,0,1]=(0.886532090807)*x[0]+(0.330645525532)*x[1]
        arg[2,1,1,0]=(0.18224742814)*x[0]+(-0.713166591262)*x[1]
        arg[2,1,1,1]=(-0.25218296706)*x[0]+(0.736797897014)*x[1]
        arg[2,1,2,0]=(0.885270371408)*x[0]+(-1.35641397127)*x[1]
        arg[2,1,2,1]=(-0.963214290879)*x[0]+(0.348421598509)*x[1]
        arg[2,2,0,0]=(0.884120745351)*x[0]+(0.956857507884)*x[1]
        arg[2,2,0,1]=(-0.00795298517561)*x[0]+(0.51243740322)*x[1]
        arg[2,2,1,0]=(-0.563655252975)*x[0]+(-0.716305131721)*x[1]
        arg[2,2,1,1]=(0.648243491623)*x[0]+(-0.142239786485)*x[1]
        arg[2,2,2,0]=(1.50677848542)*x[0]+(-0.359722006818)*x[1]
        arg[2,2,2,1]=(-0.522592086302)*x[0]+(-0.18544927497)*x[1]
        arg[2,3,0,0]=(0.0822787762966)*x[0]+(0.196309662733)*x[1]
        arg[2,3,0,1]=(0.232573258627)*x[0]+(-0.0597976775432)*x[1]
        arg[2,3,1,0]=(1.50895737052)*x[0]+(0.546625934312)*x[1]
        arg[2,3,1,1]=(-1.64023373286)*x[0]+(0.220053261777)*x[1]
        arg[2,3,2,0]=(1.5897208842)*x[0]+(-0.0355885594992)*x[1]
        arg[2,3,2,1]=(-1.52822349982)*x[0]+(-0.0906934717656)*x[1]
        arg[2,4,0,0]=(0.496696582601)*x[0]+(-0.303888675059)*x[1]
        arg[2,4,0,1]=(-0.0819586112177)*x[0]+(-0.0449047294378)*x[1]
        arg[2,4,1,0]=(-0.678530758513)*x[0]+(-1.2173963447)*x[1]
        arg[2,4,1,1]=(-0.645450475268)*x[0]+(0.356771113426)*x[1]
        arg[2,4,2,0]=(-0.486643021867)*x[0]+(0.977763869689)*x[1]
        arg[2,4,2,1]=(-0.458991807397)*x[0]+(1.80038432923)*x[1]
        arg[3,0,0,0]=(0.270795948755)*x[0]+(-0.691512096338)*x[1]
        arg[3,0,0,1]=(-0.756717130099)*x[0]+(1.16189640534)*x[1]
        arg[3,0,1,0]=(0.967554346699)*x[0]+(-0.272119008963)*x[1]
        arg[3,0,1,1]=(1.14296491195)*x[0]+(-0.707175662379)*x[1]
        arg[3,0,2,0]=(1.15665030692)*x[0]+(0.414803389409)*x[1]
        arg[3,0,2,1]=(1.38997274955)*x[0]+(1.07065568192)*x[1]
        arg[3,1,0,0]=(-0.623998925511)*x[0]+(0.375545806421)*x[1]
        arg[3,1,0,1]=(-0.363431948398)*x[0]+(-0.183100417297)*x[1]
        arg[3,1,1,0]=(0.0435866301729)*x[0]+(-0.557137572068)*x[1]
        arg[3,1,1,1]=(-0.253225217468)*x[0]+(-0.0657932188329)*x[1]
        arg[3,1,2,0]=(0.147700921064)*x[0]+(0.681624563015)*x[1]
        arg[3,1,2,1]=(-0.075394297616)*x[0]+(0.271741245789)*x[1]
        arg[3,2,0,0]=(1.34803044812)*x[0]+(0.87991065609)*x[1]
        arg[3,2,0,1]=(0.0529487135353)*x[0]+(0.806480092352)*x[1]
        arg[3,2,1,0]=(-1.52038624166)*x[0]+(-0.992718138685)*x[1]
        arg[3,2,1,1]=(0.0859852188749)*x[0]+(1.63512596202)*x[1]
        arg[3,2,2,0]=(1.46925148575)*x[0]+(-1.60383564005)*x[1]
        arg[3,2,2,1]=(0.114463395473)*x[0]+(-1.0186377376)*x[1]
        arg[3,3,0,0]=(-1.02962502658)*x[0]+(-1.06788543146)*x[1]
        arg[3,3,0,1]=(-0.184391597652)*x[0]+(0.462603936483)*x[1]
        arg[3,3,1,0]=(0.611113335782)*x[0]+(0.282417397043)*x[1]
        arg[3,3,1,1]=(-0.193423873796)*x[0]+(-0.56108614964)*x[1]
        arg[3,3,2,0]=(-0.0290687598224)*x[0]+(-0.460221158417)*x[1]
        arg[3,3,2,1]=(0.784513522959)*x[0]+(-0.484813129206)*x[1]
        arg[3,4,0,0]=(0.423780748664)*x[0]+(-0.344020787894)*x[1]
        arg[3,4,0,1]=(1.10623024451)*x[0]+(-0.775008153301)*x[1]
        arg[3,4,1,0]=(0.554847534291)*x[0]+(-0.854749107999)*x[1]
        arg[3,4,1,1]=(-0.639595693597)*x[0]+(-0.966152813212)*x[1]
        arg[3,4,2,0]=(-0.0172621996023)*x[0]+(0.321815074269)*x[1]
        arg[3,4,2,1]=(0.39605156824)*x[0]+(-0.499138231749)*x[1]
        ref[0,0,0,0]=(0.169892838337)/(o+1.)+(-0.323261281472)+(0.722227117759)*0.5**o
        ref[0,0,0,1]=(-0.899797330597)/(o+1.)+(0.731242549464)+(-0.679761749104)*0.5**o
        ref[0,0,1,0]=(0.948969615551)/(o+1.)+(-0.173478478707)+(0.0756189154575)*0.5**o
        ref[0,0,1,1]=(-0.484419525942)/(o+1.)+(0.287761319651)+(-0.834046890353)*0.5**o
        ref[0,0,2,0]=(-0.275366085384)/(o+1.)+(-0.051636507926)+(0.534757021171)*0.5**o
        ref[0,0,2,1]=(0.338862631582)/(o+1.)+(0.076489529483)+(-0.47622767501)*0.5**o
        ref[0,1,0,0]=(0.960885993642)/(o+1.)+(0.36959166751)+(0.789134441414)*0.5**o
        ref[0,1,0,1]=(-0.101908900511)/(o+1.)+(0.343409131566)+(0.298411594744)*0.5**o
        ref[0,1,1,0]=(-0.50227382988)/(o+1.)+(-0.431549057432)+(-0.509965959533)*0.5**o
        ref[0,1,1,1]=(-0.294216689224)/(o+1.)+(0.293114907564)+(-0.00647512035965)*0.5**o
        ref[0,1,2,0]=(0.330313139948)/(o+1.)+(0.0601559902451)+(-0.447667920962)*0.5**o
        ref[0,1,2,1]=(-0.855325117375)/(o+1.)+(0.455706690618)+(0.0237326178252)*0.5**o
        ref[0,2,0,0]=(0.958948557218)/(o+1.)+(-0.307015973897)+(-0.752766039908)*0.5**o
        ref[0,2,0,1]=(0.622666086232)/(o+1.)+(0.417782845058)+(-0.676857571283)*0.5**o
        ref[0,2,1,0]=(-0.476269525636)/(o+1.)+(0.203518519787)+(-0.541895856898)*0.5**o
        ref[0,2,1,1]=(0.396862410812)/(o+1.)+(-0.318442063423)+(0.499722654464)*0.5**o
        ref[0,2,2,0]=(0.291546173382)/(o+1.)+(0.345688404218)+(0.372569619405)*0.5**o
        ref[0,2,2,1]=(-0.0863432016899)/(o+1.)+(0.296494638577)+(-0.631266836064)*0.5**o
        ref[0,3,0,0]=(0.437179077921)/(o+1.)+(-0.340515851088)+(-0.00622121586734)*0.5**o
        ref[0,3,0,1]=(-0.0670787919003)/(o+1.)+(-0.492735838895)+(0.931769856293)*0.5**o
        ref[0,3,1,0]=(-0.455620785864)/(o+1.)+(-0.866886545624)+(0.632565106759)*0.5**o
        ref[0,3,1,1]=(0.769395985412)/(o+1.)+(-0.303642587154)+(0.781062911223)*0.5**o
        ref[0,3,2,0]=(-0.338160377272)/(o+1.)+(-0.237545833658)+(-0.966712929597)*0.5**o
        ref[0,3,2,1]=(-0.796435534783)/(o+1.)+(0.544922318125)+(-0.743274373676)*0.5**o
        ref[0,4,0,0]=(0.452932120968)/(o+1.)+(-0.512874769603)+(-0.048932296957)*0.5**o
        ref[0,4,0,1]=(0.361579034112)/(o+1.)+(-0.0561290568189)+(0.365767627701)*0.5**o
        ref[0,4,1,0]=(-0.626053161593)/(o+1.)+(-0.761324946295)+(0.911283010254)*0.5**o
        ref[0,4,1,1]=(-0.0455977379835)/(o+1.)+(0.292122388548)+(0.22020429757)*0.5**o
        ref[0,4,2,0]=(-0.881052488675)/(o+1.)+(-0.210171400955)+(-0.973018746263)*0.5**o
        ref[0,4,2,1]=(0.955475719603)/(o+1.)+(0.367959093676)+(0.126904536122)*0.5**o
        ref[1,0,0,0]=(-0.314523133291)/(o+1.)+(-0.649561146707)+(-0.54959125554)*0.5**o
        ref[1,0,0,1]=(0.776726613012)/(o+1.)+(0.0747896120381)+(0.804907910029)*0.5**o
        ref[1,0,1,0]=(0.416309554773)/(o+1.)+(0.0650606132082)+(-0.795886225295)*0.5**o
        ref[1,0,1,1]=(-0.642083920221)/(o+1.)+(-0.518812997818)+(-0.590503261366)*0.5**o
        ref[1,0,2,0]=(0.209400082886)/(o+1.)+(-0.592256267485)+(0.00147357225359)*0.5**o
        ref[1,0,2,1]=(-0.969945221754)/(o+1.)+(-0.364559871358)+(-0.0931427766681)*0.5**o
        ref[1,1,0,0]=(0.941138659497)/(o+1.)+(-0.103485523138)+(-0.866286624489)*0.5**o
        ref[1,1,0,1]=(0.527395399401)/(o+1.)+(0.373358049173)+(-0.612019142395)*0.5**o
        ref[1,1,1,0]=(-0.276942103917)/(o+1.)+(0.624509599487)+(0.0329877871648)*0.5**o
        ref[1,1,1,1]=(-0.153477829221)/(o+1.)+(-0.0616268433725)+(-0.85593389444)*0.5**o
        ref[1,1,2,0]=(0.2685285297)/(o+1.)+(-0.301824551915)+(-0.386723203413)*0.5**o
        ref[1,1,2,1]=(-0.530960074652)/(o+1.)+(0.248621312801)+(-0.772378264474)*0.5**o
        ref[1,2,0,0]=(0.745071931955)/(o+1.)+(-0.149268301427)+(-0.000656345828701)*0.5**o
        ref[1,2,0,1]=(-0.581822428087)/(o+1.)+(0.2315691847)+(0.122913062738)*0.5**o
        ref[1,2,1,0]=(-0.325165099021)/(o+1.)+(-0.414758139139)+(0.187663931198)*0.5**o
        ref[1,2,1,1]=(-0.389683593964)/(o+1.)+(0.265985790556)+(0.485893691018)*0.5**o
        ref[1,2,2,0]=(0.661969003098)/(o+1.)+(-0.514584521701)+(-0.990427195841)*0.5**o
        ref[1,2,2,1]=(-0.0828384828188)/(o+1.)+(-0.482166570368)+(0.832805029032)*0.5**o
        ref[1,3,0,0]=(-0.756020671713)/(o+1.)+(0.627704058228)+(0.150404496577)*0.5**o
        ref[1,3,0,1]=(-0.93648010541)/(o+1.)+(0.11022703478)+(-0.387767702896)*0.5**o
        ref[1,3,1,0]=(-0.995823513997)/(o+1.)+(-0.242072805569)+(-0.604794178119)*0.5**o
        ref[1,3,1,1]=(-0.543071878541)/(o+1.)+(-0.160890620667)+(0.654757881656)*0.5**o
        ref[1,3,2,0]=(0.803184441579)/(o+1.)+(-0.612293875048)+(-0.504652146723)*0.5**o
        ref[1,3,2,1]=(-0.786714439322)/(o+1.)+(-0.175556007517)+(-0.18491128296)*0.5**o
        ref[1,4,0,0]=(-0.302928744298)/(o+1.)+(0.476360987754)+(-0.423396151472)*0.5**o
        ref[1,4,0,1]=(-0.410240602476)/(o+1.)+(-0.49551630458)+(0.40280698327)*0.5**o
        ref[1,4,1,0]=(-0.820677400346)/(o+1.)+(-0.573037933363)+(-0.752720066188)*0.5**o
        ref[1,4,1,1]=(-0.492748923658)/(o+1.)+(-0.600739806799)+(0.170290165507)*0.5**o
        ref[1,4,2,0]=(0.926654770824)/(o+1.)+(0.011282908181)+(-0.697106837662)*0.5**o
        ref[1,4,2,1]=(-0.958477382611)/(o+1.)+(0.618515858902)+(-0.65973097111)*0.5**o
        ref[2,0,0,0]=(-0.15636950624)/(o+1.)+(-0.520670547262)+(-0.36455690548)*0.5**o
        ref[2,0,0,1]=(-0.726424878286)/(o+1.)+(0.697757734621)+(-0.386898263585)*0.5**o
        ref[2,0,1,0]=(-0.055162393508)/(o+1.)+(0.247015646426)+(-0.442127640696)*0.5**o
        ref[2,0,1,1]=(-0.769332288875)/(o+1.)+(-0.0917175733456)+(-0.639003232373)*0.5**o
        ref[2,0,2,0]=(-0.196390038297)/(o+1.)+(0.0977907270548)+(0.340749124126)*0.5**o
        ref[2,0,2,1]=(0.841927481859)/(o+1.)+(-0.489185426146)+(0.0626606144906)*0.5**o
        ref[2,1,0,0]=(0.590018249148)/(o+1.)+(0.378476967182)+(-0.354886735451)*0.5**o
        ref[2,1,0,1]=(0.178495604742)/(o+1.)+(0.233073081069)+(0.572535849459)*0.5**o
        ref[2,1,1,0]=(-0.762671064919)/(o+1.)+(-0.108539686142)+(0.448831274081)*0.5**o
        ref[2,1,1,1]=(0.927591406535)/(o+1.)+(0.0441635703243)+(-0.53130361723)*0.5**o
        ref[2,1,2,0]=(-0.604360706259)/(o+1.)+(-0.134848102332)+(0.402913311063)*0.5**o
        ref[2,1,2,1]=(0.229727627441)/(o+1.)+(-0.323296604293)+(-0.197927111225)*0.5**o
        ref[2,2,0,0]=(0.287478356112)/(o+1.)+(0.762635506201)+(0.0282288847209)*0.5**o
        ref[2,2,0,1]=(0.243518182621)/(o+1.)+(-0.241502069199)+(0.743970373822)*0.5**o
        ref[2,2,1,0]=(-0.369049760378)/(o+1.)+(-0.21941778965)+(-0.472075045017)*0.5**o
        ref[2,2,1,1]=(-0.745343658666)/(o+1.)+(0.193375557775)+(0.864596248255)*0.5**o
        ref[2,2,2,0]=(-0.491769251494)/(o+1.)+(0.442742387618)+(0.753340954862)*0.5**o
        ref[2,2,2,1]=(0.551196377158)/(o+1.)+(-0.523535406613)+(-0.212166925203)*0.5**o
        ref[2,3,0,0]=(0.192271522168)/(o+1.)+(-0.0181953944687)+(0.122707705798)*0.5**o
        ref[2,3,0,1]=(-0.593768738594)/(o+1.)+(0.199088910545)+(0.368366498588)*0.5**o
        ref[2,3,1,0]=(0.0866211808983)/(o+1.)+(0.723512694127)+(0.521936735685)*0.5**o
        ref[2,3,1,1]=(-0.453288268254)/(o+1.)+(-0.15580970281)+(-0.655272797205)*0.5**o
        ref[2,3,2,0]=(0.614105802669)/(o+1.)+(0.0411427694898)+(0.85774098305)*0.5**o
        ref[2,3,2,1]=(0.263133533654)/(o+1.)+(-0.628776355867)+(-0.624497793507)*0.5**o
        ref[2,4,0,0]=(-0.809347914722)/(o+1.)+(0.367719412237)+(0.26671699779)*0.5**o
        ref[2,4,0,1]=(0.773124137566)/(o+1.)+(-0.442496395542)+(-0.0149946871367)*0.5**o
        ref[2,4,1,0]=(-0.927311525126)/(o+1.)+(-0.0125177466568)+(-0.943580084777)*0.5**o
        ref[2,4,1,1]=(0.306409665483)/(o+1.)+(0.128569553604)+(-0.852228134533)*0.5**o
        ref[2,4,2,0]=(0.926185522145)/(o+1.)+(0.0750538146738)+(-0.58517230367)*0.5**o
        ref[2,4,2,1]=(0.858245470922)/(o+1.)+(0.0125876600893)+(0.457971730734)*0.5**o
        ref[3,0,0,0]=(-0.74501947953)/(o+1.)+(0.343942243752)+(-0.363581155556)*0.5**o
        ref[3,0,0,1]=(0.850890737948)/(o+1.)+(0.157997002133)+(-0.761705466968)*0.5**o
        ref[3,0,1,0]=(-0.0468962248591)/(o+1.)+(0.0297980532763)+(0.682735456043)*0.5**o
        ref[3,0,1,1]=(0.0126326425456)/(o+1.)+(-0.076115296096)+(0.575387199213)*0.5**o
        ref[3,0,2,0]=(-0.410549752333)/(o+1.)+(0.85556811978)+(0.270867209104)*0.5**o
        ref[3,0,2,1]=(0.580487848545)/(o+1.)+(0.59069171729)+(0.698757148346)*0.5**o
        ref[3,1,0,0]=(0.139052775786)/(o+1.)+(-0.334747475889)+(0.281989056901)*0.5**o
        ref[3,1,0,1]=(0.591083610998)/(o+1.)+(-0.322680889668)+(-0.492254197357)*0.5**o
        ref[3,1,1,0]=(0.384713965427)/(o+1.)+(-0.522411606102)+(0.146558304882)*0.5**o
        ref[3,1,1,1]=(-0.827615582009)/(o+1.)+(-0.00295218575854)+(0.514501517225)*0.5**o
        ref[3,1,2,0]=(0.84762578765)/(o+1.)+(0.208361555685)+(-0.43502341494)*0.5**o
        ref[3,1,2,1]=(-0.554668615664)/(o+1.)+(0.825347835483)+(-0.899680107129)*0.5**o
        ref[3,2,0,0]=(0.652802712143)/(o+1.)+(0.397197582731)+(0.7807432266)*0.5**o
        ref[3,2,0,1]=(0.421495251937)/(o+1.)+(-0.119571710512)+(0.677076974975)*0.5**o
        ref[3,2,1,0]=(-0.230439630375)/(o+1.)+(-0.733165339017)+(-0.816334071941)*0.5**o
        ref[3,2,1,1]=(0.64153308098)/(o+1.)+(0.778374176432)+(-0.477170252948)*0.5**o
        ref[3,2,2,0]=(-0.987848717403)/(o+1.)+(-0.0606435888479)+(0.974551740798)*0.5**o
        ref[3,2,2,1]=(-0.318593711859)/(o+1.)+(-0.459904454409)+(0.334228278552)*0.5**o
        ref[3,3,0,0]=(-0.418115413351)/(o+1.)+(-0.348616063735)+(-0.982162917221)*0.5**o
        ref[3,3,0,1]=(-0.437684435015)/(o+1.)+(0.804145379606)+(-0.892393985366)*0.5**o
        ref[3,3,1,0]=(0.944100110135)/(o+1.)+(-0.059994562245)+(0.0694197471801)*0.5**o
        ref[3,3,1,1]=(-0.860656048613)/(o+1.)+(0.178136109535)+(-0.250126193893)*0.5**o
        ref[3,3,2,0]=(-0.363631991908)/(o+1.)+(-0.298384287282)+(0.471110648232)*0.5**o
        ref[3,3,2,1]=(-0.578021736479)/(o+1.)+(0.320913094365)+(0.235895941501)*0.5**o
        ref[3,4,0,0]=(-0.946995649946)/(o+1.)+(0.0451978887455)+(0.936359833225)*0.5**o
        ref[3,4,0,1]=(-0.636305818291)/(o+1.)+(0.258932088634)+(0.449663732234)*0.5**o
        ref[3,4,1,0]=(-0.823421937366)/(o+1.)+(0.112980674356)+(0.297559014947)*0.5**o
        ref[3,4,1,1]=(-0.441768022626)/(o+1.)+(-0.240840962617)+(-0.682298558949)*0.5**o
        ref[3,4,2,0]=(-0.258808936822)/(o+1.)+(0.510306577882)+(-0.457251344276)*0.5**o
        ref[3,4,2,1]=(-0.489365749016)/(o+1.)+(-0.296617385332)+(0.979513856171)*0.5**o
      else:
        arg[0,0,0,0]=(-0.474247989728)*x[0]+(-0.768689352235)*x[1]+(-1.08527900853)*x[2]
        arg[0,0,0,1]=(-0.237113797789)*x[0]+(0.0842735741018)*x[1]+(0.203858551029)*x[2]
        arg[0,0,1,0]=(0.24237127665)*x[0]+(0.421022955803)*x[1]+(-1.52818540927)*x[2]
        arg[0,0,1,1]=(1.34624865817)*x[0]+(-0.798943272818)*x[1]+(-0.823168386315)*x[2]
        arg[0,0,2,0]=(-0.906723771267)*x[0]+(0.328893123771)*x[1]+(-0.942770202463)*x[2]
        arg[0,0,2,1]=(1.54305197694)*x[0]+(0.813063941984)*x[1]+(0.205532727168)*x[2]
        arg[0,1,0,0]=(0.399159979872)*x[0]+(-1.52089837934)*x[1]+(-1.4064251406)*x[2]
        arg[0,1,0,1]=(-0.121139354731)*x[0]+(1.12768246422)*x[1]+(0.234299409411)*x[2]
        arg[0,1,1,0]=(-1.62513899388)*x[0]+(-0.173131700583)*x[1]+(1.24199996799)*x[2]
        arg[0,1,1,1]=(-0.371929434177)*x[0]+(-0.606278610409)*x[1]+(-0.00600790063066)*x[2]
        arg[0,1,2,0]=(-0.804531294423)*x[0]+(0.869392248002)*x[1]+(0.840417359943)*x[2]
        arg[0,1,2,1]=(0.329452946275)*x[0]+(1.1072495446)*x[1]+(-0.213517932323)*x[2]
        arg[0,2,0,0]=(1.02784887224)*x[0]+(0.435517042337)*x[1]+(1.58767206677)*x[2]
        arg[0,2,0,1]=(-0.0120997177)*x[0]+(0.430135803314)*x[1]+(-0.717473068956)*x[2]
        arg[0,2,1,0]=(0.955455517302)*x[0]+(0.712976918968)*x[1]+(-0.624804508641)*x[2]
        arg[0,2,1,1]=(-1.03293986594)*x[0]+(-0.539062834482)*x[1]+(-1.61172701428)*x[2]
        arg[0,2,2,0]=(-0.177662810138)*x[0]+(0.553352366127)*x[1]+(0.695417121111)*x[2]
        arg[0,2,2,1]=(0.318133474679)*x[0]+(-0.452528947926)*x[1]+(0.468602505091)*x[2]
        arg[0,3,0,0]=(-0.294467725916)*x[0]+(1.67314409894)*x[1]+(0.100736440007)*x[2]
        arg[0,3,0,1]=(-0.517162349887)*x[0]+(-1.57430191254)*x[1]+(1.24045877402)*x[2]
        arg[0,3,1,0]=(0.667393640368)*x[0]+(0.340993632212)*x[1]+(-1.09920535559)*x[2]
        arg[0,3,1,1]=(0.37451406727)*x[0]+(0.37350649326)*x[1]+(0.357440871322)*x[2]
        arg[0,3,2,0]=(-0.989366398194)*x[0]+(0.596179154759)*x[1]+(-0.573280308087)*x[2]
        arg[0,3,2,1]=(-0.25561190141)*x[0]+(1.42296905238)*x[1]+(-1.21032031399)*x[2]
        arg[0,4,0,0]=(-0.41394085285)*x[0]+(-0.751213929646)*x[1]+(-0.68470718293)*x[2]
        arg[0,4,0,1]=(0.702985204167)*x[0]+(-0.0375821582587)*x[1]+(0.290630803819)*x[2]
        arg[0,4,1,0]=(0.0558204653441)*x[0]+(1.61940057302)*x[1]+(-0.518863528354)*x[2]
        arg[0,4,1,1]=(1.25876303604)*x[0]+(-1.40923202908)*x[1]+(1.05202333762)*x[2]
        arg[0,4,2,0]=(0.0715295483294)*x[0]+(0.883869702529)*x[1]+(-0.463253474408)*x[2]
        arg[0,4,2,1]=(0.564563789615)*x[0]+(-0.135049742988)*x[1]+(-1.3118878567)*x[2]
        arg[1,0,0,0]=(0.707267335291)*x[0]+(-0.197244449872)*x[1]+(-0.114362773781)*x[2]
        arg[1,0,0,1]=(-0.144762815659)*x[0]+(0.434177706452)*x[1]+(0.0754859099691)*x[2]
        arg[1,0,1,0]=(-0.906934991359)*x[0]+(0.179011995627)*x[1]+(0.823936036469)*x[2]
        arg[1,0,1,1]=(1.11165724605)*x[0]+(0.48023223731)*x[1]+(0.981353808181)*x[2]
        arg[1,0,2,0]=(0.499460583481)*x[0]+(0.171586243561)*x[1]+(-0.252474775657)*x[2]
        arg[1,0,2,1]=(-0.604289370884)*x[0]+(1.11610914086)*x[1]+(0.747439062711)*x[2]
        arg[1,1,0,0]=(0.696191463882)*x[0]+(0.349547714404)*x[1]+(0.651361019042)*x[2]
        arg[1,1,0,1]=(0.152679923426)*x[0]+(-0.448442911425)*x[1]+(1.33964371501)*x[2]
        arg[1,1,1,0]=(-0.74612592015)*x[0]+(-1.27089239786)*x[1]+(-0.261725631736)*x[2]
        arg[1,1,1,1]=(1.50839493929)*x[0]+(1.93229724987)*x[1]+(-0.963996845831)*x[2]
        arg[1,1,2,0]=(0.867984447375)*x[0]+(-0.777493811854)*x[1]+(0.847847553133)*x[2]
        arg[1,1,2,1]=(1.34249389498)*x[0]+(0.776160804062)*x[1]+(-0.295393972199)*x[2]
        arg[1,2,0,0]=(-0.655771613243)*x[0]+(1.58220214455)*x[1]+(-0.406292477629)*x[2]
        arg[1,2,0,1]=(0.59294895986)*x[0]+(0.544961164723)*x[1]+(0.634279670948)*x[2]
        arg[1,2,1,0]=(1.15324179862)*x[0]+(0.645764835217)*x[1]+(0.670364725309)*x[2]
        arg[1,2,1,1]=(-0.819459106271)*x[0]+(0.024815625404)*x[1]+(0.769753650676)*x[2]
        arg[1,2,2,0]=(1.19042366161)*x[0]+(0.982020548972)*x[1]+(0.92730609738)*x[2]
        arg[1,2,2,1]=(-0.742441488349)*x[0]+(-1.3101429812)*x[1]+(-0.198104246324)*x[2]
        arg[1,3,0,0]=(0.349205030437)*x[0]+(-0.745240339596)*x[1]+(-0.744742092259)*x[2]
        arg[1,3,0,1]=(0.993101296711)*x[0]+(-1.48331355036)*x[1]+(-1.00616778677)*x[2]
        arg[1,3,1,0]=(-1.10586442824)*x[0]+(1.71922449779)*x[1]+(-0.943451153259)*x[2]
        arg[1,3,1,1]=(-1.72393763035)*x[0]+(-1.14113495956)*x[1]+(-0.211909615809)*x[2]
        arg[1,3,2,0]=(-0.378302963108)*x[0]+(-0.494705322002)*x[1]+(0.940164363332)*x[2]
        arg[1,3,2,1]=(0.850541148743)*x[0]+(-1.45461945468)*x[1]+(1.20025821328)*x[2]
        arg[1,4,0,0]=(1.33882463795)*x[0]+(0.475208133119)*x[1]+(0.722686059253)*x[2]
        arg[1,4,0,1]=(0.0750422155969)*x[0]+(1.29758288845)*x[1]+(-1.41431668738)*x[2]
        arg[1,4,1,0]=(1.36072344575)*x[0]+(-0.345743547563)*x[1]+(-0.63248018754)*x[2]
        arg[1,4,1,1]=(0.579049229913)*x[0]+(0.954245074078)*x[1]+(-0.170538926965)*x[2]
        arg[1,4,2,0]=(1.67615484634)*x[0]+(0.561067019023)*x[1]+(-1.26952326914)*x[2]
        arg[1,4,2,1]=(0.700910547934)*x[0]+(0.58717328764)*x[1]+(0.795138196965)*x[2]
        arg[2,0,0,0]=(0.763193592359)*x[0]+(-0.397549328381)*x[1]+(-0.230593559545)*x[2]
        arg[2,0,0,1]=(0.454026153468)*x[0]+(-1.12464443717)*x[1]+(-0.147462098246)*x[2]
        arg[2,0,1,0]=(-0.276457034935)*x[0]+(0.118173156654)*x[1]+(0.57296366275)*x[2]
        arg[2,0,1,1]=(1.31887812253)*x[0]+(-0.544644483254)*x[1]+(0.0013728350582)*x[2]
        arg[2,0,2,0]=(-0.257589384265)*x[0]+(-0.016005453203)*x[1]+(-0.508836430965)*x[2]
        arg[2,0,2,1]=(-0.259420301773)*x[0]+(0.122645293168)*x[1]+(0.381613885289)*x[2]
        arg[2,1,0,0]=(-1.64168512418)*x[0]+(0.025182006192)*x[1]+(0.0268780458975)*x[2]
        arg[2,1,0,1]=(-0.0885344200898)*x[0]+(0.408795219114)*x[1]+(0.635798801606)*x[2]
        arg[2,1,1,0]=(-0.959658503283)*x[0]+(1.034415556)*x[1]+(0.725681105748)*x[2]
        arg[2,1,1,1]=(1.59312707789)*x[0]+(-0.116514952101)*x[1]+(-0.86121633507)*x[2]
        arg[2,1,2,0]=(-1.44859572067)*x[0]+(-1.28344247275)*x[1]+(0.299296401869)*x[2]
        arg[2,1,2,1]=(-0.488424372501)*x[0]+(1.23481402288)*x[1]+(0.01702902463)*x[2]
        arg[2,2,0,0]=(-1.29986951244)*x[0]+(-1.20190912751)*x[1]+(-0.755848291051)*x[2]
        arg[2,2,0,1]=(-0.52421656312)*x[0]+(-0.335471509689)*x[1]+(-1.01264312499)*x[2]
        arg[2,2,1,0]=(-1.12066642609)*x[0]+(-1.02138558053)*x[1]+(-1.00777755552)*x[2]
        arg[2,2,1,1]=(-1.42127024744)*x[0]+(1.63480022479)*x[1]+(0.542244906643)*x[2]
        arg[2,2,2,0]=(-0.418852880297)*x[0]+(-0.983110436453)*x[1]+(-0.622603673805)*x[2]
        arg[2,2,2,1]=(0.727366446825)*x[0]+(-0.256324856361)*x[1]+(-0.20608740758)*x[2]
        arg[2,3,0,0]=(0.548268302883)*x[0]+(0.0102173127665)*x[1]+(0.869833576217)*x[2]
        arg[2,3,0,1]=(-1.70159091061)*x[0]+(1.15876875083)*x[1]+(-0.234757706773)*x[2]
        arg[2,3,1,0]=(-1.72817024817)*x[0]+(0.690055132218)*x[1]+(-0.149398998825)*x[2]
        arg[2,3,1,1]=(1.80829412387)*x[0]+(-0.217899654855)*x[1]+(-0.922672217822)*x[2]
        arg[2,3,2,0]=(1.23855200328)*x[0]+(0.412795411508)*x[1]+(0.960476191755)*x[2]
        arg[2,3,2,1]=(-1.43061904621)*x[0]+(-0.766039000745)*x[1]+(-1.25765397361)*x[2]
        arg[2,4,0,0]=(-0.660872016687)*x[0]+(-0.000405233718708)*x[1]+(-0.178507968219)*x[2]
        arg[2,4,0,1]=(-0.168526882871)*x[0]+(-1.06936802685)*x[1]+(1.54083857104)*x[2]
        arg[2,4,1,0]=(0.565807359529)*x[0]+(-1.41401297939)*x[1]+(-1.23254260537)*x[2]
        arg[2,4,1,1]=(-0.373920075802)*x[0]+(0.407330812903)*x[1]+(-0.204667001328)*x[2]
        arg[2,4,2,0]=(0.874100879885)*x[0]+(-0.191238094267)*x[1]+(1.54890969328)*x[2]
        arg[2,4,2,1]=(0.566428125313)*x[0]+(-1.60899595601)*x[1]+(-1.05112529768)*x[2]
        arg[3,0,0,0]=(-0.688037248111)*x[0]+(-1.56483484784)*x[1]+(0.26782504619)*x[2]
        arg[3,0,0,1]=(0.418788954459)*x[0]+(-0.670455226092)*x[1]+(-0.0562760591375)*x[2]
        arg[3,0,1,0]=(-0.2614362754)*x[0]+(0.271050519855)*x[1]+(0.227811982923)*x[2]
        arg[3,0,1,1]=(-0.873198928647)*x[0]+(-0.483763141395)*x[1]+(0.33532383979)*x[2]
        arg[3,0,2,0]=(-0.864161938176)*x[0]+(-1.4682754386)*x[1]+(-0.38126411162)*x[2]
        arg[3,0,2,1]=(0.142827521381)*x[0]+(0.464483940188)*x[1]+(0.214297576606)*x[2]
        arg[3,1,0,0]=(0.188747079113)*x[0]+(0.590821253713)*x[1]+(0.908850367108)*x[2]
        arg[3,1,0,1]=(0.0587541582624)*x[0]+(0.955494266151)*x[1]+(1.70822209849)*x[2]
        arg[3,1,1,0]=(-0.420356861523)*x[0]+(-0.551445918262)*x[1]+(-0.0187114754269)*x[2]
        arg[3,1,1,1]=(1.07071658401)*x[0]+(-0.589035602194)*x[1]+(0.133725074027)*x[2]
        arg[3,1,2,0]=(0.370477043767)*x[0]+(0.226262353618)*x[1]+(0.566087784489)*x[2]
        arg[3,1,2,1]=(1.77828660075)*x[0]+(0.607419742685)*x[1]+(0.0297054744582)*x[2]
        arg[3,2,0,0]=(-1.19326939474)*x[0]+(1.20875153887)*x[1]+(-0.124756222091)*x[2]
        arg[3,2,0,1]=(0.0645190914271)*x[0]+(0.357580475716)*x[1]+(-1.59461469718)*x[2]
        arg[3,2,1,0]=(1.4853619905)*x[0]+(0.751582301648)*x[1]+(0.859363177167)*x[2]
        arg[3,2,1,1]=(1.36658655075)*x[0]+(-0.776650996613)*x[1]+(0.654743994261)*x[2]
        arg[3,2,2,0]=(0.406696005677)*x[0]+(-0.528546876384)*x[1]+(-0.439775325396)*x[2]
        arg[3,2,2,1]=(-0.725198159461)*x[0]+(-0.0933136564416)*x[1]+(-0.720365141393)*x[2]
        arg[3,3,0,0]=(-0.428669112669)*x[0]+(0.626162511175)*x[1]+(-0.382776233923)*x[2]
        arg[3,3,0,1]=(0.340474939218)*x[0]+(-0.140550383925)*x[1]+(0.353836983746)*x[2]
        arg[3,3,1,0]=(0.851646816711)*x[0]+(0.0171684195949)*x[1]+(0.38173044322)*x[2]
        arg[3,3,1,1]=(0.359647722225)*x[0]+(0.425799941165)*x[1]+(-0.80174052212)*x[2]
        arg[3,3,2,0]=(0.13012568736)*x[0]+(-0.573735693651)*x[1]+(1.08843730088)*x[2]
        arg[3,3,2,1]=(-1.01372953294)*x[0]+(0.335282956856)*x[1]+(1.24570512628)*x[2]
        arg[3,4,0,0]=(-0.05674609039)*x[0]+(-0.814879575064)*x[1]+(-0.169789408264)*x[2]
        arg[3,4,0,1]=(1.5968230938)*x[0]+(0.0679819277023)*x[1]+(0.892284314512)*x[2]
        arg[3,4,1,0]=(1.22220298478)*x[0]+(-0.0258417926472)*x[1]+(0.440670138568)*x[2]
        arg[3,4,1,1]=(0.647072768716)*x[0]+(-0.177651087906)*x[1]+(-0.115553644646)*x[2]
        arg[3,4,2,0]=(1.29816363306)*x[0]+(-1.18656044167)*x[1]+(-0.465776533157)*x[2]
        arg[3,4,2,1]=(0.328907496091)*x[0]+(-0.179857937151)*x[1]+(-0.476306938306)*x[2]
        ref[0,0,0,0]=(-0.530378359482)/(o+1.)+(-0.935143327523)+(0.0724486640358)*0.5**o
        ref[0,0,0,1]=(1.22837847871)/(o+1.)+(-0.113350954728)+(-0.950658241915)*0.5**o
        ref[0,0,1,0]=(-0.000414683535449)/(o+1.)+(-0.183158731552)+(-0.49805903018)*0.5**o
        ref[0,0,1,1]=(-0.882730389132)/(o+1.)+(-0.112174958816)+(0.831217305802)*0.5**o
        ref[0,0,2,0]=(-1.1913557966)/(o+1.)+(-0.0615750891074)+(-0.206094875146)*0.5**o
        ref[0,0,2,1]=(0.634536492589)/(o+1.)+(0.649890059464)+(0.627332034578)*0.5**o
        ref[0,1,0,0]=(-1.38665625775)/(o+1.)+(-1.0210653524)+(0.900623422471)*0.5**o
        ref[0,1,0,1]=(1.20633678024)/(o+1.)+(0.0888193121856)+(-0.143132885707)*0.5**o
        ref[0,1,1,0]=(0.452225278035)/(o+1.)+(-0.148926943801)+(-0.710642116901)*0.5**o
        ref[0,1,1,1]=(-0.244223430204)/(o+1.)+(-0.619607191996)+(0.499221868978)*0.5**o
        ref[0,1,2,0]=(0.0383849036606)/(o+1.)+(0.759484971644)+(-0.652076533427)*0.5**o
        ref[0,1,2,1]=(0.207558055759)/(o+1.)+(0.502305382384)+(0.0110157380215)*0.5**o
        ref[0,2,0,0]=(0.761374280144)/(o+1.)+(0.777616611515)+(0.734430478174)*0.5**o
        ref[0,2,0,1]=(0.554502862028)/(o+1.)+(-0.409274602016)+(-0.0353906413371)*0.5**o
        ref[0,2,1,0]=(-0.252543841709)/(o+1.)+(0.386640377894)+(0.522891013551)*0.5**o
        ref[0,2,1,1]=(-1.52497235887)/(o+1.)+(-0.807603254401)+(-0.0435508470367)*0.5**o
        ref[0,2,2,0]=(0.172485606257)/(o+1.)+(0.350297526187)+(0.19802601847)*0.5**o
        ref[0,2,2,1]=(-0.141271359379)/(o+1.)+(0.0509198134622)+(0.373638764298)*0.5**o
        ref[0,3,0,0]=(1.31278836809)/(o+1.)+(-0.0918739323256)+(0.350372309595)*0.5**o
        ref[0,3,0,1]=(-0.396933928109)/(o+1.)+(-0.436137805999)+(0.418204051696)*0.5**o
        ref[0,3,1,0]=(-0.206619840099)/(o+1.)+(-0.118703289798)+(0.353208336689)*0.5**o
        ref[0,3,1,1]=(-0.595153371293)/(o+1.)+(0.615722581562)+(0.469169640021)*0.5**o
        ref[0,3,2,0]=(-0.680858997986)/(o+1.)+(-0.0393223937964)+(-0.206963765944)*0.5**o
        ref[0,3,2,1]=(0.012753264598)/(o+1.)+(0.246188741489)+(-0.548093910595)*0.5**o
        ref[0,4,0,0]=(-0.40478987122)/(o+1.)+(-0.890399681609)+(0.33572726901)*0.5**o
        ref[0,4,0,1]=(0.0280062557264)/(o+1.)+(0.0358049413228)+(0.856417711355)*0.5**o
        ref[0,4,1,0]=(0.181560004565)/(o+1.)+(0.329885369854)+(0.31502676574)*0.5**o
        ref[0,4,1,1]=(0.0552681276841)/(o+1.)+(0.00372919426319)+(0.838827828375)*0.5**o
        ref[0,4,2,0]=(0.656057162724)/(o+1.)+(0.179315527684)+(-0.522542441641)*0.5**o
        ref[0,4,2,1]=(0.126189389317)/(o+1.)+(-0.429452001643)+(-0.149659196108)*0.5**o
        ref[1,0,0,0]=(1.43443485527)/(o+1.)+(-0.55091754491)+(0.0630603461901)*0.5**o
        ref[1,0,0,1]=(-0.612191779306)/(o+1.)+(0.604025235381)+(-0.230957890695)*0.5**o
        ref[1,0,1,0]=(0.373085423681)/(o+1.)+(-0.0817391530697)+(-0.113594076805)*0.5**o
        ref[1,0,1,1]=(0.761862677657)/(o+1.)+(0.819524484428)+(0.172331645033)*0.5**o
        ref[1,0,2,0]=(0.929265450746)/(o+1.)+(-0.382585784938)+(0.254478170515)*0.5**o
        ref[1,0,2,1]=(1.58236203463)/(o+1.)+(-0.280641188811)+(0.238179175679)*0.5**o
        ref[1,1,0,0]=(-0.117055726312)/(o+1.)+(0.957914090571)+(-0.101672257501)*0.5**o
        ref[1,1,0,1]=(0.0849062553796)/(o+1.)+(0.0703564566846)+(0.818261558262)*0.5**o
        ref[1,1,1,0]=(-1.72219113622)/(o+1.)+(-0.357454724939)+(0.15835663635)*0.5**o
        ref[1,1,1,1]=(0.739078400468)/(o+1.)+(0.532292543169)+(0.673031856531)*0.5**o
        ref[1,1,2,0]=(0.255023909903)/(o+1.)+(0.229656917655)+(0.22400044344)*0.5**o
        ref[1,1,2,1]=(0.518856765199)/(o+1.)+(0.380552790125)+(0.543298381396)*0.5**o
        ref[1,2,0,0]=(0.0321563884831)/(o+1.)+(0.25981567762)+(-0.0316496900482)*0.5**o
        ref[1,2,0,1]=(1.09318176771)/(o+1.)+(0.0572442907603)+(0.564519446305)*0.5**o
        ref[1,2,1,0]=(0.756962791387)/(o+1.)+(0.69419429624)+(0.32401997528)*0.5**o
        ref[1,2,1,1]=(1.28480299241)/(o+1.)+(-0.201990593541)+(-0.905711635518)*0.5**o
        ref[1,2,2,0]=(1.04969566442)/(o+1.)+(0.72854097372)+(0.592972696101)*0.5**o
        ref[1,2,2,1]=(-1.77086264775)/(o+1.)+(-0.261914180361)+(0.0440022926066)*0.5**o
        ref[1,3,0,0]=(-0.477780531124)/(o+1.)+(-0.181354085163)+(-0.300288699967)*0.5**o
        ref[1,3,0,1]=(-0.797627013456)/(o+1.)+(-0.523163955197)+(0.347574883435)*0.5**o
        ref[1,3,1,0]=(0.428806106821)/(o+1.)+(-0.0514584302394)+(-0.655980330052)*0.5**o
        ref[1,3,1,1]=(-0.141740459457)/(o+1.)+(-1.05699960931)+(-0.821242527644)*0.5**o
        ref[1,3,2,0]=(0.0181810985873)/(o+1.)+(-0.180697139553)+(0.41036925874)*0.5**o
        ref[1,3,2,1]=(-0.061077367)/(o+1.)+(0.250077446763)+(0.157102380812)*0.5**o
        ref[1,4,0,0]=(1.14087212913)/(o+1.)+(0.368059029864)+(0.659728641463)*0.5**o
        ref[1,4,0,1]=(0.365658713316)/(o+1.)+(0.088533821044)+(-0.584417938734)*0.5**o
        ref[1,4,1,0]=(-1.11681308321)/(o+1.)+(0.438275560627)+(0.622761672604)*0.5**o
        ref[1,4,1,1]=(1.18893372018)/(o+1.)+(0.172934240923)+(-0.172046824999)*0.5**o
        ref[1,4,2,0]=(0.16240489493)/(o+1.)+(-0.0145923852256)+(0.834478471749)*0.5**o
        ref[1,4,2,1]=(0.571355829378)/(o+1.)+(0.526896417749)+(0.458073367663)*0.5**o
        ref[2,0,0,0]=(-0.489234572937)/(o+1.)+(-0.0135589093901)+(0.65140309615)*0.5**o
        ref[2,0,0,1]=(-1.55383491931)/(o+1.)+(0.0161644619972)+(0.703425613366)*0.5**o
        ref[2,0,1,0]=(-0.258604855723)/(o+1.)+(0.0827812678671)+(0.507722104457)*0.5**o
        ref[2,0,1,1]=(0.0504178529481)/(o+1.)+(0.182991850332)+(0.359204920726)*0.5**o
        ref[2,0,2,0]=(0.652043479841)/(o+1.)+(-0.443015450013)+(-0.548443848248)*0.5**o
        ref[2,0,2,1]=(0.340317839857)/(o+1.)+(-0.379565594545)+(0.663652225917)*0.5**o
        ref[2,1,0,0]=(-0.587863196667)/(o+1.)+(-0.0137797674284)+(-0.974202340562)*0.5**o
        ref[2,1,0,1]=(1.00030747039)/(o+1.)+(-0.399997754768)+(0.755747639771)*0.5**o
        ref[2,1,1,0]=(0.703413281036)/(o+1.)+(0.487016208736)+(-0.877007540039)*0.5**o
        ref[2,1,1,1]=(0.263166830572)/(o+1.)+(-0.157259452849)+(0.666747865841)*0.5**o
        ref[2,1,2,0]=(0.104532343004)/(o+1.)+(-1.02548001493)+(-0.486314104701)*0.5**o
        ref[2,1,2,1]=(0.933518655478)/(o+1.)+(0.0369928229139)+(-0.244085626298)*0.5**o
        ref[2,2,0,0]=(-1.50550384519)/(o+1.)+(-0.590996227594)+(-0.570130630622)*0.5**o
        ref[2,2,0,1]=(-0.60318683839)/(o+1.)+(-0.38308245528)+(-0.502979448853)*0.5**o
        ref[2,2,1,0]=(-1.11258910879)/(o+1.)+(-0.518777277969)+(-0.999685897413)*0.5**o
        ref[2,2,1,1]=(1.84366617107)/(o+1.)+(-0.232866738261)+(-0.622157810558)*0.5**o
        ref[2,2,2,0]=(-0.620181988585)/(o+1.)+(-0.836162444183)+(0.267939886396)*0.5**o
        ref[2,2,2,1]=(-0.826647389099)/(o+1.)+(0.0929536518109)+(0.905694268362)*0.5**o
        ref[2,3,0,0]=(0.184067674443)/(o+1.)+(0.259745858757)+(0.724759799909)*0.5**o
        ref[2,3,0,1]=(0.0175595303049)/(o+1.)+(0.0560117358912)+(-0.907162868639)*0.5**o
        ref[2,3,1,0]=(-0.609443154617)/(o+1.)+(0.186540819401)+(-0.951152598958)*0.5**o
        ref[2,3,1,1]=(-1.61858082094)/(o+1.)+(0.679439858578)+(0.927423354971)*0.5**o
        ref[2,3,2,0]=(0.44350652486)/(o+1.)+(0.81466177481)+(0.53899353207)*0.5**o
        ref[2,3,2,1]=(-0.877289862245)/(o+1.)+(-1.05742260849)+(-0.462176941342)*0.5**o
        ref[2,4,0,0]=(-0.736381416539)/(o+1.)+(-0.0139374259612)+(-0.0755289501641)*0.5**o
        ref[2,4,0,1]=(0.690429326999)/(o+1.)+(-0.0293644018251)+(-0.328756862036)*0.5**o
        ref[2,4,1,0]=(-1.57117034002)/(o+1.)+(-0.531463765387)+(0.553349645561)*0.5**o
        ref[2,4,1,1]=(1.24898175282)/(o+1.)+(-0.452273024959)+(-0.515691967125)*0.5**o
        ref[2,4,2,0]=(0.239126191643)/(o+1.)+(0.722504563932)+(0.547637159392)*0.5**o
        ref[2,4,2,1]=(-1.40763254209)/(o+1.)+(-0.809456441012)+(0.932852295739)*0.5**o
        ref[3,0,0,0]=(-0.416554297686)/(o+1.)+(-0.632193610911)+(-0.304105530258)*0.5**o
        ref[3,0,0,1]=(-1.02983448036)/(o+1.)+(0.579212231778)+(-0.436532313966)*0.5**o
        ref[3,0,1,0]=(0.42124133216)/(o+1.)+(0.368290418827)+(-0.920395942436)*0.5**o
        ref[3,0,1,1]=(0.158195474854)/(o+1.)+(-0.17459474514)+(-0.830644214826)*0.5**o
        ref[3,0,2,0]=(-1.18749647329)/(o+1.)+(-0.506173996178)+(-0.513857022748)*0.5**o
        ref[3,0,2,1]=(0.233437753018)/(o+1.)+(0.455705145371)+(-0.323239005586)*0.5**o
        ref[3,1,0,0]=(1.42991958566)/(o+1.)+(0.053769921022)+(0.150959272234)*0.5**o
        ref[3,1,0,1]=(1.48423731111)/(o+1.)+(0.964048315997)+(-0.689863420196)*0.5**o
        ref[3,1,1,0]=(0.868837676765)/(o+1.)+(-1.01261543254)+(0.165878933098)*0.5**o
        ref[3,1,1,1]=(0.820750307831)/(o+1.)+(-0.300930223821)+(0.396516195659)*0.5**o
        ref[3,1,2,0]=(0.477077782202)/(o+1.)+(-0.0416336989494)+(0.769016797571)*0.5**o
        ref[3,1,2,1]=(0.359895440587)/(o+1.)+(0.555322395811)+(0.944871585683)*0.5**o
        ref[3,2,0,0]=(0.657668315832)/(o+1.)+(-0.240321895174)+(-0.286298603451)*0.5**o
        ref[3,2,0,1]=(-0.941360568282)/(o+1.)+(-0.0246544136593)+(-0.181845734439)*0.5**o
        ref[3,2,1,0]=(1.60316834971)/(o+1.)+(0.252291602298)+(0.988555915005)*0.5**o
        ref[3,2,1,1]=(0.833811957678)/(o+1.)+(-0.111970621871)+(0.634808834458)*0.5**o
        ref[3,2,2,0]=(-1.38877073784)/(o+1.)+(0.538356636307)+(-0.249568730877)*0.5**o
        ref[3,2,2,1]=(-0.311839184097)/(o+1.)+(-0.698114103264)+(0.169190433329)*0.5**o
        ref[3,3,0,0]=(0.743170989696)/(o+1.)+(-0.499224746788)+(0.0699956684628)*0.5**o
        ref[3,3,0,1]=(-0.293538786051)/(o+1.)+(0.26667992866)+(0.313940467768)*0.5**o
        ref[3,3,1,0]=(0.951483461345)/(o+1.)+(0.148317904561)+(0.00242640905887)*0.5**o
        ref[3,3,1,1]=(-0.110753912776)/(o+1.)+(-0.44308689138)+(0.980634836805)*0.5**o
        ref[3,3,2,0]=(-0.337895223544)/(o+1.)+(0.358941646037)+(0.264839226057)*0.5**o
        ref[3,3,2,1]=(0.857823611315)/(o+1.)+(0.30404611512)+(-0.898657291359)*0.5**o
        ref[3,4,0,0]=(0.037944509376)/(o+1.)+(-0.364801700413)+(-0.349756182268)*0.5**o
        ref[3,4,0,1]=(1.01114399396)/(o+1.)+(0.457103279485)+(0.631738783076)*0.5**o
        ref[3,4,1,0]=(0.969931913888)/(o+1.)+(-0.165922085756)+(0.998943588321)*0.5**o
        ref[3,4,1,1]=(1.35441114248)/(o+1.)+(-0.502422727199)+(0.00430234808496)*0.5**o
        ref[3,4,2,0]=(-0.367201097566)/(o+1.)+(-0.337780780531)+(0.688589316866)*0.5**o
        ref[3,4,2,1]=(-0.0547905846061)/(o+1.)+(-0.293870420685)+(0.31527404661)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_FunctionOnContactOne_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnContactOne

      assumptions: FunctionOnContactOne(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.311787641998)*x[0]**o+(0.161673282875)*x[0]+(0.32570909237)*x[1]**o+(-0.249404742764)*x[1]
        ref=(0.32570909237)/(o+1.)+(-0.0438657299447)+(0.311787641998)*0.5**o
      else:
        arg=(0.0675036998003)*x[0]**o+(-0.247751392136)*x[0]+(0.791676220029)*x[1]**o+(-0.269015236163)*x[1]+(0.265735156875)*x[2]**o+(0.820319192519)*x[2]
        ref=(1.0574113769)/(o+1.)+(0.15177628211)+(0.0675036998003)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_FunctionOnContactOne_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnContactOne

      assumptions: FunctionOnContactOne(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.99235950851)*x[0]**o+(0.00867392831518)*x[0]+(-0.598024390869)*x[1]**o+(0.00945232777857)*x[1]
        arg[1]=(-0.140261338996)*x[0]**o+(-0.166524016889)*x[0]+(0.59850076881)*x[1]**o+(-0.805756215354)*x[1]
        ref[0]=(-0.598024390869)/(o+1.)+(0.00906312804687)+(0.99235950851)*0.5**o
        ref[1]=(0.59850076881)/(o+1.)+(-0.486140116122)+(-0.140261338996)*0.5**o
      else:
        arg[0]=(-0.158686556523)*x[0]**o+(0.249193164275)*x[0]+(-0.801383648208)*x[1]**o+(-0.439021533504)*x[1]+(-0.112035191366)*x[2]**o+(-0.323385862625)*x[2]
        arg[1]=(0.392997873484)*x[0]**o+(-0.867468854104)*x[0]+(0.804459243476)*x[1]**o+(0.129028860314)*x[1]+(-0.252715502862)*x[2]**o+(0.457355923469)*x[2]
        ref[0]=(-0.913418839574)/(o+1.)+(-0.256607115927)+(-0.158686556523)*0.5**o
        ref[1]=(0.551743740614)/(o+1.)+(-0.140542035161)+(0.392997873484)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_FunctionOnContactOne_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnContactOne

      assumptions: FunctionOnContactOne(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.330520725992)*x[0]**o+(0.160118740246)*x[0]+(-0.178670895269)*x[1]**o+(-0.281660040648)*x[1]
        arg[0,1]=(-0.871801084852)*x[0]**o+(0.692549035429)*x[0]+(-0.265102051042)*x[1]**o+(-0.799752901882)*x[1]
        arg[0,2]=(-0.816985024574)*x[0]**o+(0.534242721856)*x[0]+(0.81492666918)*x[1]**o+(0.891189158633)*x[1]
        arg[0,3]=(-0.20728122917)*x[0]**o+(-0.656873565737)*x[0]+(-0.924660993906)*x[1]**o+(0.453699593426)*x[1]
        arg[0,4]=(-0.541377726375)*x[0]**o+(0.916287049432)*x[0]+(-0.15879052465)*x[1]**o+(-0.319875161126)*x[1]
        arg[1,0]=(0.296966289838)*x[0]**o+(0.170347587334)*x[0]+(0.755207250839)*x[1]**o+(0.677658805212)*x[1]
        arg[1,1]=(-0.507591010573)*x[0]**o+(0.299638021889)*x[0]+(-0.757702071557)*x[1]**o+(-0.194564860748)*x[1]
        arg[1,2]=(0.821972467197)*x[0]**o+(-0.75902704466)*x[0]+(0.19424332718)*x[1]**o+(-0.070182646035)*x[1]
        arg[1,3]=(-0.340805687409)*x[0]**o+(0.585961973082)*x[0]+(-0.602942732318)*x[1]**o+(-0.288941395094)*x[1]
        arg[1,4]=(0.0314374795024)*x[0]**o+(0.293710824013)*x[0]+(0.576509865762)*x[1]**o+(-0.481377812302)*x[1]
        arg[2,0]=(0.250117528833)*x[0]**o+(-0.318419344173)*x[0]+(-0.793830015706)*x[1]**o+(0.0445553063586)*x[1]
        arg[2,1]=(-0.347974394766)*x[0]**o+(-0.569347105354)*x[0]+(0.838687005759)*x[1]**o+(0.534306628908)*x[1]
        arg[2,2]=(0.53063272248)*x[0]**o+(0.426504428158)*x[0]+(0.215398162753)*x[1]**o+(0.324348187788)*x[1]
        arg[2,3]=(-0.336752121527)*x[0]**o+(0.666074768462)*x[0]+(-0.739123251932)*x[1]**o+(-0.538899653147)*x[1]
        arg[2,4]=(-0.832783810519)*x[0]**o+(0.558572116384)*x[0]+(0.119988536199)*x[1]**o+(-0.664170918761)*x[1]
        arg[3,0]=(0.4941356722)*x[0]**o+(-0.403591783006)*x[0]+(0.205050527245)*x[1]**o+(0.351642574939)*x[1]
        arg[3,1]=(0.12538190695)*x[0]**o+(-0.890722638937)*x[0]+(-0.897011486902)*x[1]**o+(-0.465522770696)*x[1]
        arg[3,2]=(-0.862829007552)*x[0]**o+(-0.950531038019)*x[0]+(-0.970016379643)*x[1]**o+(-0.500580755888)*x[1]
        arg[3,3]=(0.131451743264)*x[0]**o+(0.43845495141)*x[0]+(-0.929808474589)*x[1]**o+(-0.104714452616)*x[1]
        arg[3,4]=(-0.81935224348)*x[0]**o+(-0.53791483312)*x[0]+(-0.714512816761)*x[1]**o+(0.329859667225)*x[1]
        ref[0,0]=(-0.178670895269)/(o+1.)+(-0.0607706502011)+(0.330520725992)*0.5**o
        ref[0,1]=(-0.265102051042)/(o+1.)+(-0.0536019332266)+(-0.871801084852)*0.5**o
        ref[0,2]=(0.81492666918)/(o+1.)+(0.712715940244)+(-0.816985024574)*0.5**o
        ref[0,3]=(-0.924660993906)/(o+1.)+(-0.101586986156)+(-0.20728122917)*0.5**o
        ref[0,4]=(-0.15879052465)/(o+1.)+(0.298205944153)+(-0.541377726375)*0.5**o
        ref[1,0]=(0.755207250839)/(o+1.)+(0.424003196273)+(0.296966289838)*0.5**o
        ref[1,1]=(-0.757702071557)/(o+1.)+(0.0525365805706)+(-0.507591010573)*0.5**o
        ref[1,2]=(0.19424332718)/(o+1.)+(-0.414604845348)+(0.821972467197)*0.5**o
        ref[1,3]=(-0.602942732318)/(o+1.)+(0.148510288994)+(-0.340805687409)*0.5**o
        ref[1,4]=(0.576509865762)/(o+1.)+(-0.0938334941445)+(0.0314374795024)*0.5**o
        ref[2,0]=(-0.793830015706)/(o+1.)+(-0.136932018907)+(0.250117528833)*0.5**o
        ref[2,1]=(0.838687005759)/(o+1.)+(-0.0175202382229)+(-0.347974394766)*0.5**o
        ref[2,2]=(0.215398162753)/(o+1.)+(0.375426307973)+(0.53063272248)*0.5**o
        ref[2,3]=(-0.739123251932)/(o+1.)+(0.0635875576574)+(-0.336752121527)*0.5**o
        ref[2,4]=(0.119988536199)/(o+1.)+(-0.0527994011886)+(-0.832783810519)*0.5**o
        ref[3,0]=(0.205050527245)/(o+1.)+(-0.0259746040335)+(0.4941356722)*0.5**o
        ref[3,1]=(-0.897011486902)/(o+1.)+(-0.678122704817)+(0.12538190695)*0.5**o
        ref[3,2]=(-0.970016379643)/(o+1.)+(-0.725555896953)+(-0.862829007552)*0.5**o
        ref[3,3]=(-0.929808474589)/(o+1.)+(0.166870249397)+(0.131451743264)*0.5**o
        ref[3,4]=(-0.714512816761)/(o+1.)+(-0.104027582947)+(-0.81935224348)*0.5**o
      else:
        arg[0,0]=(-0.916251178344)*x[0]**o+(0.858306033367)*x[0]+(0.814515071272)*x[1]**o+(-0.31988121358)*x[1]+(-0.973228319626)*x[2]**o+(-0.448894628432)*x[2]
        arg[0,1]=(0.394307707909)*x[0]**o+(-0.270473398988)*x[0]+(0.904831123674)*x[1]**o+(-0.949869961069)*x[1]+(0.771889549168)*x[2]**o+(0.650652953462)*x[2]
        arg[0,2]=(-0.0387282383528)*x[0]**o+(0.0292919821249)*x[0]+(-0.165189629855)*x[1]**o+(-0.265182468614)*x[1]+(0.919042632361)*x[2]**o+(0.710368688769)*x[2]
        arg[0,3]=(0.680006301334)*x[0]**o+(0.96959940913)*x[0]+(0.773153544603)*x[1]**o+(-0.460435754764)*x[1]+(-0.202448514604)*x[2]**o+(-0.495739672831)*x[2]
        arg[0,4]=(-0.972790915366)*x[0]**o+(-0.14148426268)*x[0]+(0.511995390348)*x[1]**o+(-0.252881217719)*x[1]+(0.255711210131)*x[2]**o+(0.729991325048)*x[2]
        arg[1,0]=(-0.105824689194)*x[0]**o+(-0.201427239595)*x[0]+(-0.00550922402684)*x[1]**o+(-0.732794556814)*x[1]+(-0.0820953451459)*x[2]**o+(0.289904794134)*x[2]
        arg[1,1]=(-0.378707957559)*x[0]**o+(0.776904510463)*x[0]+(0.0406721415351)*x[1]**o+(0.159288221237)*x[1]+(0.961004037621)*x[2]**o+(-0.0866180279992)*x[2]
        arg[1,2]=(-0.0226058491089)*x[0]**o+(-0.569995151391)*x[0]+(-0.0675265986408)*x[1]**o+(-0.141280346236)*x[1]+(0.116098815262)*x[2]**o+(0.654630437206)*x[2]
        arg[1,3]=(-0.567270755897)*x[0]**o+(0.580139672878)*x[0]+(-0.411816626246)*x[1]**o+(-0.490525831459)*x[1]+(0.9322522728)*x[2]**o+(-0.948230068413)*x[2]
        arg[1,4]=(0.977251285351)*x[0]**o+(-0.457921996364)*x[0]+(-0.586207130671)*x[1]**o+(0.801370818191)*x[1]+(-0.667233775445)*x[2]**o+(0.277054304674)*x[2]
        arg[2,0]=(0.997778241906)*x[0]**o+(0.490242237175)*x[0]+(0.406384527224)*x[1]**o+(-0.00724475574186)*x[1]+(-0.59799957401)*x[2]**o+(0.520598391022)*x[2]
        arg[2,1]=(0.59766081214)*x[0]**o+(-0.144793824209)*x[0]+(-0.191121009265)*x[1]**o+(-0.178016091024)*x[1]+(0.168431462883)*x[2]**o+(-0.766200678925)*x[2]
        arg[2,2]=(0.633988266946)*x[0]**o+(-0.880495194817)*x[0]+(-0.217502420499)*x[1]**o+(-0.730159316723)*x[1]+(0.255599562732)*x[2]**o+(-0.510511087905)*x[2]
        arg[2,3]=(0.102332382261)*x[0]**o+(-0.370884332534)*x[0]+(-0.0926186085648)*x[1]**o+(0.86446607832)*x[1]+(0.0643112986026)*x[2]**o+(0.950118360785)*x[2]
        arg[2,4]=(0.928323090624)*x[0]**o+(0.942943574803)*x[0]+(0.162570981489)*x[1]**o+(-0.971092675635)*x[1]+(0.641489870001)*x[2]**o+(0.695989498774)*x[2]
        arg[3,0]=(-0.988813741403)*x[0]**o+(0.407960892494)*x[0]+(0.671572481461)*x[1]**o+(0.119756938854)*x[1]+(-0.923671788537)*x[2]**o+(-0.458931381612)*x[2]
        arg[3,1]=(-0.241202616681)*x[0]**o+(-0.369134099039)*x[0]+(-0.129164268709)*x[1]**o+(-0.469311903775)*x[1]+(0.633909205241)*x[2]**o+(0.564043934538)*x[2]
        arg[3,2]=(-0.397081811407)*x[0]**o+(0.909387986899)*x[0]+(0.260590665694)*x[1]**o+(-0.371472818152)*x[1]+(-0.0369834262513)*x[2]**o+(-0.887598789153)*x[2]
        arg[3,3]=(-0.304262529036)*x[0]**o+(0.741316478862)*x[0]+(-0.384317206475)*x[1]**o+(-0.854685760579)*x[1]+(0.502373892581)*x[2]**o+(-0.923561844368)*x[2]
        arg[3,4]=(-0.720555420423)*x[0]**o+(-0.115786593478)*x[0]+(0.832043243085)*x[1]**o+(-0.912631896773)*x[1]+(-0.583397798668)*x[2]**o+(-0.508174721594)*x[2]
        ref[0,0]=(-0.158713248354)/(o+1.)+(0.0447650956777)+(-0.916251178344)*0.5**o
        ref[0,1]=(1.67672067284)/(o+1.)+(-0.284845203298)+(0.394307707909)*0.5**o
        ref[0,2]=(0.753853002506)/(o+1.)+(0.23723910114)+(-0.0387282383528)*0.5**o
        ref[0,3]=(0.570705029999)/(o+1.)+(0.00671199076774)+(0.680006301334)*0.5**o
        ref[0,4]=(0.767706600479)/(o+1.)+(0.167812922325)+(-0.972790915366)*0.5**o
        ref[1,0]=(-0.0876045691728)/(o+1.)+(-0.322158501138)+(-0.105824689194)*0.5**o
        ref[1,1]=(1.00167617916)/(o+1.)+(0.42478735185)+(-0.378707957559)*0.5**o
        ref[1,2]=(0.0485722166212)/(o+1.)+(-0.0283225302105)+(-0.0226058491089)*0.5**o
        ref[1,3]=(0.520435646554)/(o+1.)+(-0.429308113497)+(-0.567270755897)*0.5**o
        ref[1,4]=(-1.25344090612)/(o+1.)+(0.31025156325)+(0.977251285351)*0.5**o
        ref[2,0]=(-0.191615046786)/(o+1.)+(0.501797936227)+(0.997778241906)*0.5**o
        ref[2,1]=(-0.0226895463818)/(o+1.)+(-0.544505297079)+(0.59766081214)*0.5**o
        ref[2,2]=(0.0380971422336)/(o+1.)+(-1.06058279972)+(0.633988266946)*0.5**o
        ref[2,3]=(-0.0283073099622)/(o+1.)+(0.721850053286)+(0.102332382261)*0.5**o
        ref[2,4]=(0.80406085149)/(o+1.)+(0.333920198971)+(0.928323090624)*0.5**o
        ref[3,0]=(-0.252099307076)/(o+1.)+(0.0343932248679)+(-0.988813741403)*0.5**o
        ref[3,1]=(0.504744936532)/(o+1.)+(-0.137201034138)+(-0.241202616681)*0.5**o
        ref[3,2]=(0.223607239442)/(o+1.)+(-0.174841810203)+(-0.397081811407)*0.5**o
        ref[3,3]=(0.118056686106)/(o+1.)+(-0.518465563043)+(-0.304262529036)*0.5**o
        ref[3,4]=(0.248645444416)/(o+1.)+(-0.768296605922)+(-0.720555420423)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_FunctionOnContactOne_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnContactOne

      assumptions: FunctionOnContactOne(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.347273720148)*x[0]**o+(0.325185727721)*x[0]+(-0.855515194735)*x[1]**o+(0.825807975597)*x[1]
        arg[0,0,1]=(0.842054669437)*x[0]**o+(0.353768311241)*x[0]+(0.0148900864182)*x[1]**o+(-0.402154654143)*x[1]
        arg[0,1,0]=(-0.191525791617)*x[0]**o+(-0.299181220926)*x[0]+(-0.696260869529)*x[1]**o+(0.285645399154)*x[1]
        arg[0,1,1]=(0.190151633835)*x[0]**o+(0.184074338131)*x[0]+(-0.497689504982)*x[1]**o+(-0.813410371313)*x[1]
        arg[1,0,0]=(-0.695114302024)*x[0]**o+(0.427133485674)*x[0]+(-0.711899825683)*x[1]**o+(-0.296004734911)*x[1]
        arg[1,0,1]=(0.37620789583)*x[0]**o+(-0.944850444544)*x[0]+(0.666194812839)*x[1]**o+(-0.0715430628168)*x[1]
        arg[1,1,0]=(-0.0312291891427)*x[0]**o+(0.403009732443)*x[0]+(-0.378486005359)*x[1]**o+(-0.992729541609)*x[1]
        arg[1,1,1]=(0.0675984132698)*x[0]**o+(0.241001129867)*x[0]+(0.340546237588)*x[1]**o+(-0.024830810543)*x[1]
        arg[2,0,0]=(-0.146870758676)*x[0]**o+(-0.170434897147)*x[0]+(-0.579618469996)*x[1]**o+(-0.770817574749)*x[1]
        arg[2,0,1]=(-0.3029068201)*x[0]**o+(-0.110636302601)*x[0]+(-0.905841665338)*x[1]**o+(-0.280977886629)*x[1]
        arg[2,1,0]=(0.0547810600344)*x[0]**o+(0.00763321517272)*x[0]+(0.0156822194106)*x[1]**o+(-0.528337857094)*x[1]
        arg[2,1,1]=(0.12016647407)*x[0]**o+(0.4590743733)*x[0]+(-0.574499998403)*x[1]**o+(0.698871172503)*x[1]
        arg[3,0,0]=(0.481811757276)*x[0]**o+(0.748356476205)*x[0]+(-0.300272959749)*x[1]**o+(-0.707230263391)*x[1]
        arg[3,0,1]=(-0.347858576311)*x[0]**o+(-0.177810365974)*x[0]+(0.384658344075)*x[1]**o+(-0.837385668636)*x[1]
        arg[3,1,0]=(0.943559460135)*x[0]**o+(0.178778474835)*x[0]+(-0.831384000346)*x[1]**o+(-0.474975888568)*x[1]
        arg[3,1,1]=(-0.589903969401)*x[0]**o+(0.271114795948)*x[0]+(-0.946594221253)*x[1]**o+(-0.964029170264)*x[1]
        arg[4,0,0]=(0.221077049797)*x[0]**o+(0.986052914095)*x[0]+(-0.392982812685)*x[1]**o+(0.370828365061)*x[1]
        arg[4,0,1]=(0.463121232737)*x[0]**o+(-0.91920434872)*x[0]+(0.830655891827)*x[1]**o+(0.193876462558)*x[1]
        arg[4,1,0]=(0.925740559506)*x[0]**o+(0.534342942418)*x[0]+(0.446730463706)*x[1]**o+(0.181095371535)*x[1]
        arg[4,1,1]=(-0.777093674353)*x[0]**o+(-0.435130879793)*x[0]+(0.134406853212)*x[1]**o+(0.861099956528)*x[1]
        arg[5,0,0]=(-0.443943796359)*x[0]**o+(0.359190127208)*x[0]+(0.0570788998709)*x[1]**o+(0.548070594378)*x[1]
        arg[5,0,1]=(0.762962865085)*x[0]**o+(0.969248002849)*x[0]+(0.729222816543)*x[1]**o+(-0.865271983926)*x[1]
        arg[5,1,0]=(-0.915609563739)*x[0]**o+(-0.338560068696)*x[0]+(0.972145654544)*x[1]**o+(0.0941277349395)*x[1]
        arg[5,1,1]=(-0.486023356984)*x[0]**o+(-0.426382776921)*x[0]+(-0.114606039743)*x[1]**o+(-0.992152486495)*x[1]
        ref[0,0,0]=(-0.855515194735)/(o+1.)+(0.575496851659)+(-0.347273720148)*0.5**o
        ref[0,0,1]=(0.0148900864182)/(o+1.)+(-0.0241931714509)+(0.842054669437)*0.5**o
        ref[0,1,0]=(-0.696260869529)/(o+1.)+(-0.00676791088595)+(-0.191525791617)*0.5**o
        ref[0,1,1]=(-0.497689504982)/(o+1.)+(-0.314668016591)+(0.190151633835)*0.5**o
        ref[1,0,0]=(-0.711899825683)/(o+1.)+(0.0655643753813)+(-0.695114302024)*0.5**o
        ref[1,0,1]=(0.666194812839)/(o+1.)+(-0.50819675368)+(0.37620789583)*0.5**o
        ref[1,1,0]=(-0.378486005359)/(o+1.)+(-0.294859904583)+(-0.0312291891427)*0.5**o
        ref[1,1,1]=(0.340546237588)/(o+1.)+(0.108085159662)+(0.0675984132698)*0.5**o
        ref[2,0,0]=(-0.579618469996)/(o+1.)+(-0.470626235948)+(-0.146870758676)*0.5**o
        ref[2,0,1]=(-0.905841665338)/(o+1.)+(-0.195807094615)+(-0.3029068201)*0.5**o
        ref[2,1,0]=(0.0156822194106)/(o+1.)+(-0.260352320961)+(0.0547810600344)*0.5**o
        ref[2,1,1]=(-0.574499998403)/(o+1.)+(0.578972772902)+(0.12016647407)*0.5**o
        ref[3,0,0]=(-0.300272959749)/(o+1.)+(0.0205631064069)+(0.481811757276)*0.5**o
        ref[3,0,1]=(0.384658344075)/(o+1.)+(-0.507598017305)+(-0.347858576311)*0.5**o
        ref[3,1,0]=(-0.831384000346)/(o+1.)+(-0.148098706867)+(0.943559460135)*0.5**o
        ref[3,1,1]=(-0.946594221253)/(o+1.)+(-0.346457187158)+(-0.589903969401)*0.5**o
        ref[4,0,0]=(-0.392982812685)/(o+1.)+(0.678440639578)+(0.221077049797)*0.5**o
        ref[4,0,1]=(0.830655891827)/(o+1.)+(-0.362663943081)+(0.463121232737)*0.5**o
        ref[4,1,0]=(0.446730463706)/(o+1.)+(0.357719156977)+(0.925740559506)*0.5**o
        ref[4,1,1]=(0.134406853212)/(o+1.)+(0.212984538367)+(-0.777093674353)*0.5**o
        ref[5,0,0]=(0.0570788998709)/(o+1.)+(0.453630360793)+(-0.443943796359)*0.5**o
        ref[5,0,1]=(0.729222816543)/(o+1.)+(0.0519880094615)+(0.762962865085)*0.5**o
        ref[5,1,0]=(0.972145654544)/(o+1.)+(-0.122216166878)+(-0.915609563739)*0.5**o
        ref[5,1,1]=(-0.114606039743)/(o+1.)+(-0.709267631708)+(-0.486023356984)*0.5**o
      else:
        arg[0,0,0]=(0.455440122065)*x[0]**o+(0.911340571887)*x[0]+(0.158409397783)*x[1]**o+(0.353755257124)*x[1]+(-0.392556456322)*x[2]**o+(0.707232271095)*x[2]
        arg[0,0,1]=(-0.816405087957)*x[0]**o+(0.0691299991887)*x[0]+(0.178677657085)*x[1]**o+(0.526176369634)*x[1]+(-0.159459641444)*x[2]**o+(-0.197907778155)*x[2]
        arg[0,1,0]=(-0.463418217546)*x[0]**o+(-0.485835567862)*x[0]+(0.760458548482)*x[1]**o+(0.902873590314)*x[1]+(-0.760679635511)*x[2]**o+(-0.483163544794)*x[2]
        arg[0,1,1]=(0.743029549626)*x[0]**o+(-0.39131913217)*x[0]+(0.488809682734)*x[1]**o+(-0.256115251135)*x[1]+(0.0918611991036)*x[2]**o+(-0.281291060967)*x[2]
        arg[1,0,0]=(0.794367740289)*x[0]**o+(0.727663222288)*x[0]+(-0.631739003283)*x[1]**o+(0.733862609049)*x[1]+(-0.721230552103)*x[2]**o+(0.787774790111)*x[2]
        arg[1,0,1]=(0.505578181889)*x[0]**o+(0.616497234291)*x[0]+(-0.616903198005)*x[1]**o+(-0.355969913594)*x[1]+(0.0105854597103)*x[2]**o+(0.510273945857)*x[2]
        arg[1,1,0]=(0.35370601878)*x[0]**o+(-0.0414232536827)*x[0]+(0.0546619209334)*x[1]**o+(0.0459241421788)*x[1]+(0.796266758386)*x[2]**o+(0.303998757002)*x[2]
        arg[1,1,1]=(-0.245326163282)*x[0]**o+(0.0893045737315)*x[0]+(0.019401528413)*x[1]**o+(-0.838249399348)*x[1]+(0.510320088517)*x[2]**o+(0.544558918049)*x[2]
        arg[2,0,0]=(-0.660208039977)*x[0]**o+(-0.224874460802)*x[0]+(-0.0467836584153)*x[1]**o+(-0.463886703951)*x[1]+(-0.0951679104966)*x[2]**o+(-0.735787783779)*x[2]
        arg[2,0,1]=(-0.588259372806)*x[0]**o+(-0.18950606147)*x[0]+(-0.373439432467)*x[1]**o+(-0.116285718036)*x[1]+(-0.944687791765)*x[2]**o+(-0.243001707318)*x[2]
        arg[2,1,0]=(0.798498386971)*x[0]**o+(0.500125941024)*x[0]+(0.197339376576)*x[1]**o+(-0.068777545221)*x[1]+(0.36649695624)*x[2]**o+(0.193956440608)*x[2]
        arg[2,1,1]=(0.27501104189)*x[0]**o+(0.203066488761)*x[0]+(0.379278873515)*x[1]**o+(-0.789154057614)*x[1]+(0.893889499095)*x[2]**o+(0.867881975803)*x[2]
        arg[3,0,0]=(0.0579414247006)*x[0]**o+(0.661826954985)*x[0]+(0.313969775635)*x[1]**o+(-0.125420102972)*x[1]+(-0.808513922748)*x[2]**o+(0.0301073398588)*x[2]
        arg[3,0,1]=(-0.254544981247)*x[0]**o+(0.755983567644)*x[0]+(0.129524533092)*x[1]**o+(0.562828535682)*x[1]+(0.712881323083)*x[2]**o+(0.375450604678)*x[2]
        arg[3,1,0]=(-0.100221176995)*x[0]**o+(-0.470987646782)*x[0]+(-0.444245257427)*x[1]**o+(-0.0112626639353)*x[1]+(-0.516917833845)*x[2]**o+(-0.664158938756)*x[2]
        arg[3,1,1]=(0.259306174564)*x[0]**o+(0.367178744389)*x[0]+(0.177897569676)*x[1]**o+(-0.355724319562)*x[1]+(-0.809141944951)*x[2]**o+(0.962429694607)*x[2]
        arg[4,0,0]=(0.351990799011)*x[0]**o+(-0.41657457182)*x[0]+(-0.769748341564)*x[1]**o+(-0.856355494001)*x[1]+(-0.777902936204)*x[2]**o+(-0.388980567423)*x[2]
        arg[4,0,1]=(0.875307372684)*x[0]**o+(0.061375057221)*x[0]+(-0.262062525469)*x[1]**o+(0.0950324968281)*x[1]+(-0.145145751811)*x[2]**o+(-0.246872045818)*x[2]
        arg[4,1,0]=(0.115340652181)*x[0]**o+(0.910414483949)*x[0]+(0.347463744624)*x[1]**o+(0.399961393031)*x[1]+(0.0491863598064)*x[2]**o+(-0.992304094864)*x[2]
        arg[4,1,1]=(-0.125763267852)*x[0]**o+(0.485867057874)*x[0]+(-0.414172783931)*x[1]**o+(0.164145362806)*x[1]+(-0.488738184728)*x[2]**o+(-0.232758639076)*x[2]
        arg[5,0,0]=(-0.748184063441)*x[0]**o+(-0.252715593115)*x[0]+(0.688609235584)*x[1]**o+(0.0645665122312)*x[1]+(0.365465805991)*x[2]**o+(-0.217347848328)*x[2]
        arg[5,0,1]=(-0.298018437594)*x[0]**o+(-0.39683066279)*x[0]+(-0.0609074331294)*x[1]**o+(0.146930937933)*x[1]+(-0.00737018012329)*x[2]**o+(-0.700002160047)*x[2]
        arg[5,1,0]=(0.974781669859)*x[0]**o+(0.0634072121163)*x[0]+(0.0194044697894)*x[1]**o+(-0.919375545742)*x[1]+(-0.779806244021)*x[2]**o+(0.641761171945)*x[2]
        arg[5,1,1]=(0.485215976154)*x[0]**o+(0.879804212343)*x[0]+(-0.765170516406)*x[1]**o+(-0.205158612954)*x[1]+(-0.586708070331)*x[2]**o+(0.506670980992)*x[2]
        ref[0,0,0]=(-0.23414705854)/(o+1.)+(0.986164050053)+(0.455440122065)*0.5**o
        ref[0,0,1]=(0.0192180156412)/(o+1.)+(0.198699295334)+(-0.816405087957)*0.5**o
        ref[0,1,0]=(-0.000221087029094)/(o+1.)+(-0.0330627611712)+(-0.463418217546)*0.5**o
        ref[0,1,1]=(0.580670881837)/(o+1.)+(-0.464362722135)+(0.743029549626)*0.5**o
        ref[1,0,0]=(-1.35296955539)/(o+1.)+(1.12465031072)+(0.794367740289)*0.5**o
        ref[1,0,1]=(-0.606317738295)/(o+1.)+(0.385400633277)+(0.505578181889)*0.5**o
        ref[1,1,0]=(0.85092867932)/(o+1.)+(0.154249822749)+(0.35370601878)*0.5**o
        ref[1,1,1]=(0.52972161693)/(o+1.)+(-0.102192953784)+(-0.245326163282)*0.5**o
        ref[2,0,0]=(-0.141951568912)/(o+1.)+(-0.712274474266)+(-0.660208039977)*0.5**o
        ref[2,0,1]=(-1.31812722423)/(o+1.)+(-0.274396743412)+(-0.588259372806)*0.5**o
        ref[2,1,0]=(0.563836332815)/(o+1.)+(0.312652418205)+(0.798498386971)*0.5**o
        ref[2,1,1]=(1.27316837261)/(o+1.)+(0.140897203475)+(0.27501104189)*0.5**o
        ref[3,0,0]=(-0.494544147113)/(o+1.)+(0.283257095936)+(0.0579414247006)*0.5**o
        ref[3,0,1]=(0.842405856175)/(o+1.)+(0.847131354002)+(-0.254544981247)*0.5**o
        ref[3,1,0]=(-0.961163091272)/(o+1.)+(-0.573204624737)+(-0.100221176995)*0.5**o
        ref[3,1,1]=(-0.631244375275)/(o+1.)+(0.486942059717)+(0.259306174564)*0.5**o
        ref[4,0,0]=(-1.54765127777)/(o+1.)+(-0.830955316622)+(0.351990799011)*0.5**o
        ref[4,0,1]=(-0.407208277279)/(o+1.)+(-0.0452322458845)+(0.875307372684)*0.5**o
        ref[4,1,0]=(0.39665010443)/(o+1.)+(0.159035891058)+(0.115340652181)*0.5**o
        ref[4,1,1]=(-0.902910968659)/(o+1.)+(0.208626890802)+(-0.125763267852)*0.5**o
        ref[5,0,0]=(1.05407504158)/(o+1.)+(-0.202748464606)+(-0.748184063441)*0.5**o
        ref[5,0,1]=(-0.0682776132527)/(o+1.)+(-0.474950942452)+(-0.298018437594)*0.5**o
        ref[5,1,0]=(-0.760401774232)/(o+1.)+(-0.10710358084)+(0.974781669859)*0.5**o
        ref[5,1,1]=(-1.35187858674)/(o+1.)+(0.590658290191)+(0.485215976154)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnContactOne_fromData_FunctionOnContactOne_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnContactOne

      assumptions: FunctionOnContactOne(self.domain) exists
                   self.domain supports integral on FunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnContactOne(self.domain)
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.177078428867)*x[0]**o+(0.338116995819)*x[0]+(-0.750823220445)*x[1]**o+(0.0630074420245)*x[1]
        arg[0,0,0,1]=(-0.682838671264)*x[0]**o+(-0.893797611147)*x[0]+(-0.0127659464437)*x[1]**o+(0.144120749888)*x[1]
        arg[0,0,1,0]=(0.352439319413)*x[0]**o+(0.380064866457)*x[0]+(-0.932452125435)*x[1]**o+(0.914519937443)*x[1]
        arg[0,0,1,1]=(-0.354033408023)*x[0]**o+(0.771069338542)*x[0]+(-0.702758889809)*x[1]**o+(-0.484485925087)*x[1]
        arg[0,0,2,0]=(0.706826713479)*x[0]**o+(0.0526230471775)*x[0]+(-0.535244256862)*x[1]**o+(-0.085929844746)*x[1]
        arg[0,0,2,1]=(-0.893497856528)*x[0]**o+(-0.892309840177)*x[0]+(0.742770093639)*x[1]**o+(0.987917060326)*x[1]
        arg[0,1,0,0]=(-0.536624329097)*x[0]**o+(-0.569536051343)*x[0]+(-0.7740511254)*x[1]**o+(0.873150177653)*x[1]
        arg[0,1,0,1]=(0.388396767346)*x[0]**o+(-0.902569510334)*x[0]+(0.196975646933)*x[1]**o+(0.184178333918)*x[1]
        arg[0,1,1,0]=(-0.125771297178)*x[0]**o+(-0.335093336446)*x[0]+(0.372752949038)*x[1]**o+(-0.883928748878)*x[1]
        arg[0,1,1,1]=(-0.313637256508)*x[0]**o+(-0.49607955109)*x[0]+(-0.726599229867)*x[1]**o+(0.360837115079)*x[1]
        arg[0,1,2,0]=(0.92425817047)*x[0]**o+(0.734728600171)*x[0]+(1.60583902129e-05)*x[1]**o+(-0.124600971339)*x[1]
        arg[0,1,2,1]=(-0.0981827607557)*x[0]**o+(-0.556005942436)*x[0]+(-0.888060617134)*x[1]**o+(0.857508807636)*x[1]
        arg[0,2,0,0]=(-0.871590768072)*x[0]**o+(-0.208546938841)*x[0]+(-0.228336105988)*x[1]**o+(0.444098195066)*x[1]
        arg[0,2,0,1]=(-0.628675127006)*x[0]**o+(0.834043549769)*x[0]+(-0.509662470381)*x[1]**o+(-0.801907226651)*x[1]
        arg[0,2,1,0]=(-0.767528433339)*x[0]**o+(0.305967685996)*x[0]+(-0.587902586916)*x[1]**o+(0.38652781193)*x[1]
        arg[0,2,1,1]=(0.388984380682)*x[0]**o+(0.490673256209)*x[0]+(-0.78470263647)*x[1]**o+(0.34909383158)*x[1]
        arg[0,2,2,0]=(-0.910514531543)*x[0]**o+(-0.418810366626)*x[0]+(-0.637159792471)*x[1]**o+(0.794653160897)*x[1]
        arg[0,2,2,1]=(0.570535552656)*x[0]**o+(-0.590385806831)*x[0]+(-0.298806908011)*x[1]**o+(-0.901531327347)*x[1]
        arg[0,3,0,0]=(0.916864925136)*x[0]**o+(-0.019733476459)*x[0]+(-0.91370863065)*x[1]**o+(0.291932128733)*x[1]
        arg[0,3,0,1]=(-0.851147484632)*x[0]**o+(-0.159304969602)*x[0]+(0.699832871063)*x[1]**o+(0.327932729929)*x[1]
        arg[0,3,1,0]=(0.461273862468)*x[0]**o+(0.40068516653)*x[0]+(0.0203170502559)*x[1]**o+(0.973716167207)*x[1]
        arg[0,3,1,1]=(0.63984877633)*x[0]**o+(0.580871491133)*x[0]+(-0.00587359808953)*x[1]**o+(-0.104805621161)*x[1]
        arg[0,3,2,0]=(-0.159416471436)*x[0]**o+(-0.585039922994)*x[0]+(0.689553978928)*x[1]**o+(0.569617375406)*x[1]
        arg[0,3,2,1]=(0.782397836312)*x[0]**o+(-0.581894025693)*x[0]+(0.0370277809126)*x[1]**o+(0.597309139432)*x[1]
        arg[0,4,0,0]=(-0.0063433205485)*x[0]**o+(-0.242291627085)*x[0]+(0.110218625254)*x[1]**o+(0.819656578408)*x[1]
        arg[0,4,0,1]=(-0.128215646666)*x[0]**o+(0.821312510893)*x[0]+(0.910723121611)*x[1]**o+(-0.0174623980585)*x[1]
        arg[0,4,1,0]=(-0.604528735513)*x[0]**o+(-0.358705990504)*x[0]+(-0.281954207008)*x[1]**o+(0.123469639952)*x[1]
        arg[0,4,1,1]=(0.65096807032)*x[0]**o+(0.882935341204)*x[0]+(-0.761226249914)*x[1]**o+(-0.830800218214)*x[1]
        arg[0,4,2,0]=(-0.710718948634)*x[0]**o+(0.452128874613)*x[0]+(-0.780671499903)*x[1]**o+(-0.942033435912)*x[1]
        arg[0,4,2,1]=(-0.0443187506467)*x[0]**o+(0.513943853321)*x[0]+(0.709902333411)*x[1]**o+(0.755430555318)*x[1]
        arg[1,0,0,0]=(0.731890221753)*x[0]**o+(-0.821835735249)*x[0]+(-0.713098084853)*x[1]**o+(0.875738077525)*x[1]
        arg[1,0,0,1]=(-0.130335125365)*x[0]**o+(0.580579210083)*x[0]+(-0.427105331069)*x[1]**o+(0.0435798068353)*x[1]
        arg[1,0,1,0]=(-0.203908509107)*x[0]**o+(-0.941974700448)*x[0]+(0.344459670502)*x[1]**o+(0.758336135081)*x[1]
        arg[1,0,1,1]=(0.186770929724)*x[0]**o+(0.348332058705)*x[0]+(-0.0683630778679)*x[1]**o+(0.480425212189)*x[1]
        arg[1,0,2,0]=(-0.891275312975)*x[0]**o+(0.260630881729)*x[0]+(0.127955472707)*x[1]**o+(-0.554992279729)*x[1]
        arg[1,0,2,1]=(-0.74620344519)*x[0]**o+(0.944025669845)*x[0]+(-0.052031562679)*x[1]**o+(0.966060927262)*x[1]
        arg[1,1,0,0]=(0.871423198418)*x[0]**o+(-0.0769723128617)*x[0]+(0.783714758356)*x[1]**o+(0.530209967997)*x[1]
        arg[1,1,0,1]=(0.943197824902)*x[0]**o+(-0.62624855429)*x[0]+(-0.00888785913795)*x[1]**o+(-0.0408857497082)*x[1]
        arg[1,1,1,0]=(-0.74971981056)*x[0]**o+(0.925175163834)*x[0]+(0.481063204532)*x[1]**o+(-0.476700447185)*x[1]
        arg[1,1,1,1]=(-0.431776603043)*x[0]**o+(-0.495358165507)*x[0]+(-0.0110588803484)*x[1]**o+(-0.575147325496)*x[1]
        arg[1,1,2,0]=(0.578105297721)*x[0]**o+(-0.88169100464)*x[0]+(-0.646224518073)*x[1]**o+(0.465619837803)*x[1]
        arg[1,1,2,1]=(0.442066689582)*x[0]**o+(0.0937839719519)*x[0]+(-0.101534585934)*x[1]**o+(-0.974003442605)*x[1]
        arg[1,2,0,0]=(-0.206702379228)*x[0]**o+(0.330161049866)*x[0]+(0.185493587693)*x[1]**o+(0.0391986992855)*x[1]
        arg[1,2,0,1]=(0.79092697362)*x[0]**o+(0.692054695304)*x[0]+(0.141622971369)*x[1]**o+(0.643869474153)*x[1]
        arg[1,2,1,0]=(0.914281902407)*x[0]**o+(0.936365302055)*x[0]+(0.249095964198)*x[1]**o+(-0.440892551355)*x[1]
        arg[1,2,1,1]=(-0.602926709108)*x[0]**o+(0.817198874641)*x[0]+(0.262528290838)*x[1]**o+(0.493607450371)*x[1]
        arg[1,2,2,0]=(-0.724435531493)*x[0]**o+(-0.455127093687)*x[0]+(0.144972402754)*x[1]**o+(-0.361255031527)*x[1]
        arg[1,2,2,1]=(-0.611538590545)*x[0]**o+(-0.780730652602)*x[0]+(-0.150439352532)*x[1]**o+(0.450875229163)*x[1]
        arg[1,3,0,0]=(-0.913553947133)*x[0]**o+(-0.930172924067)*x[0]+(-0.509426399026)*x[1]**o+(0.3953360754)*x[1]
        arg[1,3,0,1]=(-0.709834132519)*x[0]**o+(0.972632526254)*x[0]+(0.540358729462)*x[1]**o+(-0.0295254269136)*x[1]
        arg[1,3,1,0]=(-0.30788139278)*x[0]**o+(-0.362561853154)*x[0]+(0.0135015706177)*x[1]**o+(0.558034983013)*x[1]
        arg[1,3,1,1]=(-0.243327801257)*x[0]**o+(-0.156242539226)*x[0]+(0.076082439418)*x[1]**o+(0.818613998695)*x[1]
        arg[1,3,2,0]=(-0.0290765899224)*x[0]**o+(-0.108604693002)*x[0]+(-0.0741034599215)*x[1]**o+(0.44467098076)*x[1]
        arg[1,3,2,1]=(-0.751374584848)*x[0]**o+(0.568700665267)*x[0]+(-0.431665912498)*x[1]**o+(0.881337230634)*x[1]
        arg[1,4,0,0]=(-0.940864464113)*x[0]**o+(-0.204392585977)*x[0]+(-0.872729326815)*x[1]**o+(0.454069061239)*x[1]
        arg[1,4,0,1]=(0.620544898911)*x[0]**o+(-0.194050689442)*x[0]+(0.399921161647)*x[1]**o+(0.812456444893)*x[1]
        arg[1,4,1,0]=(-0.398096053023)*x[0]**o+(-0.521802671926)*x[0]+(-0.525522030503)*x[1]**o+(-0.795276120062)*x[1]
        arg[1,4,1,1]=(0.277239330261)*x[0]**o+(-0.376159942175)*x[0]+(0.785902674573)*x[1]**o+(-0.00686958652716)*x[1]
        arg[1,4,2,0]=(-0.469242672532)*x[0]**o+(0.289432797089)*x[0]+(0.728305062026)*x[1]**o+(0.821761742668)*x[1]
        arg[1,4,2,1]=(0.828113301834)*x[0]**o+(-0.149301261166)*x[0]+(0.279230829643)*x[1]**o+(-0.971944918618)*x[1]
        arg[2,0,0,0]=(-0.724500208669)*x[0]**o+(-0.761179166402)*x[0]+(0.302409732402)*x[1]**o+(0.509101646177)*x[1]
        arg[2,0,0,1]=(-0.133300605917)*x[0]**o+(-0.0299057843108)*x[0]+(0.190028790905)*x[1]**o+(-0.668097775221)*x[1]
        arg[2,0,1,0]=(0.828019154028)*x[0]**o+(-0.446721913762)*x[0]+(0.745204383075)*x[1]**o+(0.363968853359)*x[1]
        arg[2,0,1,1]=(-0.653962912288)*x[0]**o+(0.753529804971)*x[0]+(-0.104712073949)*x[1]**o+(-0.44152589042)*x[1]
        arg[2,0,2,0]=(0.21572653818)*x[0]**o+(-0.553685840495)*x[0]+(0.372296357092)*x[1]**o+(-0.23566755183)*x[1]
        arg[2,0,2,1]=(-0.0943992743407)*x[0]**o+(0.677836956763)*x[0]+(-0.33055202826)*x[1]**o+(0.444414877097)*x[1]
        arg[2,1,0,0]=(-0.812500024009)*x[0]**o+(0.4856670967)*x[0]+(-0.891481528328)*x[1]**o+(-0.244043957515)*x[1]
        arg[2,1,0,1]=(-0.958152880518)*x[0]**o+(0.494470945352)*x[0]+(-0.000696143847979)*x[1]**o+(-0.036680781152)*x[1]
        arg[2,1,1,0]=(-0.81109120479)*x[0]**o+(-0.929246624754)*x[0]+(0.128356074186)*x[1]**o+(0.0107006959497)*x[1]
        arg[2,1,1,1]=(-0.72470858095)*x[0]**o+(-0.202254265463)*x[0]+(-0.995176229907)*x[1]**o+(-0.752592973024)*x[1]
        arg[2,1,2,0]=(-0.191810147267)*x[0]**o+(-0.819397723879)*x[0]+(0.306860971474)*x[1]**o+(0.681274705052)*x[1]
        arg[2,1,2,1]=(-0.867447323913)*x[0]**o+(-0.279148200325)*x[0]+(-0.195907590382)*x[1]**o+(-0.305826952304)*x[1]
        arg[2,2,0,0]=(0.983782298896)*x[0]**o+(-0.284064487432)*x[0]+(-0.0928333646602)*x[1]**o+(-0.929708520581)*x[1]
        arg[2,2,0,1]=(0.121619221979)*x[0]**o+(-0.876309915691)*x[0]+(0.455030678927)*x[1]**o+(0.635763125356)*x[1]
        arg[2,2,1,0]=(0.378726440151)*x[0]**o+(-0.626329947078)*x[0]+(-0.0239503446136)*x[1]**o+(0.32696940295)*x[1]
        arg[2,2,1,1]=(0.442586981104)*x[0]**o+(-0.427714534417)*x[0]+(0.906235883406)*x[1]**o+(0.819595435756)*x[1]
        arg[2,2,2,0]=(-0.564560890159)*x[0]**o+(-0.115075498726)*x[0]+(-0.520596320159)*x[1]**o+(-0.368366112222)*x[1]
        arg[2,2,2,1]=(0.215858165597)*x[0]**o+(-0.575665537635)*x[0]+(0.792289256264)*x[1]**o+(0.215575294284)*x[1]
        arg[2,3,0,0]=(-0.596335978041)*x[0]**o+(-0.677983108509)*x[0]+(-0.551360906631)*x[1]**o+(-0.874453959198)*x[1]
        arg[2,3,0,1]=(-0.913806992124)*x[0]**o+(0.192343143898)*x[0]+(0.821070238869)*x[1]**o+(-0.188359117641)*x[1]
        arg[2,3,1,0]=(-0.0831972104302)*x[0]**o+(-0.927005811846)*x[0]+(-0.43240265235)*x[1]**o+(-0.144332400422)*x[1]
        arg[2,3,1,1]=(0.643878326697)*x[0]**o+(0.212328499471)*x[0]+(0.72072318435)*x[1]**o+(-0.146144387035)*x[1]
        arg[2,3,2,0]=(0.843652876743)*x[0]**o+(-0.0386075132797)*x[0]+(0.754511827443)*x[1]**o+(-0.26945353599)*x[1]
        arg[2,3,2,1]=(0.0413857115672)*x[0]**o+(0.682727309008)*x[0]+(0.261249117327)*x[1]**o+(-0.753970539308)*x[1]
        arg[2,4,0,0]=(-0.607477693828)*x[0]**o+(-0.5870344058)*x[0]+(-0.0678823396451)*x[1]**o+(0.201869223616)*x[1]
        arg[2,4,0,1]=(-0.618172945146)*x[0]**o+(0.685084114438)*x[0]+(0.0457149578306)*x[1]**o+(-0.877971038996)*x[1]
        arg[2,4,1,0]=(0.722668769621)*x[0]**o+(0.302142466253)*x[0]+(0.630735604709)*x[1]**o+(-0.427754766567)*x[1]
        arg[2,4,1,1]=(-0.847113860795)*x[0]**o+(-0.779747246887)*x[0]+(-0.693659242703)*x[1]**o+(0.897451884568)*x[1]
        arg[2,4,2,0]=(-0.437578500878)*x[0]**o+(0.399407036331)*x[0]+(0.705207362097)*x[1]**o+(-0.271060427211)*x[1]
        arg[2,4,2,1]=(0.329570005179)*x[0]**o+(-0.712089906899)*x[0]+(0.860707329294)*x[1]**o+(0.553218273352)*x[1]
        arg[3,0,0,0]=(0.932602317046)*x[0]**o+(-0.948445141026)*x[0]+(0.549409370786)*x[1]**o+(0.782964260926)*x[1]
        arg[3,0,0,1]=(0.975116319103)*x[0]**o+(-0.0131313829738)*x[0]+(0.222376331882)*x[1]**o+(0.545375534163)*x[1]
        arg[3,0,1,0]=(0.019317277942)*x[0]**o+(0.918003018679)*x[0]+(-0.481194397466)*x[1]**o+(0.762780571026)*x[1]
        arg[3,0,1,1]=(-0.393162543578)*x[0]**o+(0.984304464298)*x[0]+(0.737901448601)*x[1]**o+(-0.890589811376)*x[1]
        arg[3,0,2,0]=(0.458096711279)*x[0]**o+(0.0343270226524)*x[0]+(0.278242150139)*x[1]**o+(-0.158378755134)*x[1]
        arg[3,0,2,1]=(0.307111052767)*x[0]**o+(-0.294724312785)*x[0]+(-0.489134379327)*x[1]**o+(-0.211139185303)*x[1]
        arg[3,1,0,0]=(0.420707084879)*x[0]**o+(-0.998173294986)*x[0]+(0.508808683)*x[1]**o+(-0.189902487212)*x[1]
        arg[3,1,0,1]=(-0.745253594401)*x[0]**o+(-0.344809947215)*x[0]+(0.198212694401)*x[1]**o+(-0.125428245606)*x[1]
        arg[3,1,1,0]=(0.535215252931)*x[0]**o+(0.565625295858)*x[0]+(-0.0741524230667)*x[1]**o+(-0.898600294889)*x[1]
        arg[3,1,1,1]=(0.193417593556)*x[0]**o+(-0.80053921694)*x[0]+(-0.664372557476)*x[1]**o+(0.47786060257)*x[1]
        arg[3,1,2,0]=(0.553690237434)*x[0]**o+(0.970763166888)*x[0]+(0.408815221365)*x[1]**o+(0.143175119269)*x[1]
        arg[3,1,2,1]=(-0.211683249218)*x[0]**o+(0.0643055656631)*x[0]+(-0.111028404397)*x[1]**o+(0.506093872018)*x[1]
        arg[3,2,0,0]=(-0.866119051438)*x[0]**o+(-0.417310807983)*x[0]+(-0.228385234131)*x[1]**o+(0.554977378653)*x[1]
        arg[3,2,0,1]=(-0.146618266885)*x[0]**o+(0.817733306502)*x[0]+(-0.709059272627)*x[1]**o+(-0.271140815458)*x[1]
        arg[3,2,1,0]=(0.622973500199)*x[0]**o+(-0.568858605993)*x[0]+(-0.651609755171)*x[1]**o+(-0.866184190672)*x[1]
        arg[3,2,1,1]=(-0.538565077894)*x[0]**o+(-0.676237465393)*x[0]+(0.106659392436)*x[1]**o+(0.488735158765)*x[1]
        arg[3,2,2,0]=(-0.325897413686)*x[0]**o+(0.99502293321)*x[0]+(-0.339096900793)*x[1]**o+(0.238083096263)*x[1]
        arg[3,2,2,1]=(0.861430484293)*x[0]**o+(-0.761739126386)*x[0]+(-0.10651751912)*x[1]**o+(0.688993400068)*x[1]
        arg[3,3,0,0]=(0.0890838652267)*x[0]**o+(0.151733228604)*x[0]+(0.230063809846)*x[1]**o+(-0.135041740614)*x[1]
        arg[3,3,0,1]=(-0.969251937403)*x[0]**o+(0.345709865473)*x[0]+(-0.309971427898)*x[1]**o+(-0.262217856737)*x[1]
        arg[3,3,1,0]=(0.640075232231)*x[0]**o+(-0.00817560718155)*x[0]+(0.55773788655)*x[1]**o+(0.642205213407)*x[1]
        arg[3,3,1,1]=(-0.350002630494)*x[0]**o+(0.373270925939)*x[0]+(-0.655757122598)*x[1]**o+(-0.74598967873)*x[1]
        arg[3,3,2,0]=(-0.262399316954)*x[0]**o+(-0.37991094527)*x[0]+(0.310310045807)*x[1]**o+(0.838942372822)*x[1]
        arg[3,3,2,1]=(-0.977039760257)*x[0]**o+(-0.0804751651971)*x[0]+(-0.844329846245)*x[1]**o+(0.905827257908)*x[1]
        arg[3,4,0,0]=(0.749594099085)*x[0]**o+(0.382880202285)*x[0]+(0.39699250956)*x[1]**o+(-0.708672318437)*x[1]
        arg[3,4,0,1]=(0.913703789336)*x[0]**o+(-0.497792024208)*x[0]+(-0.34125011876)*x[1]**o+(-0.657159430634)*x[1]
        arg[3,4,1,0]=(-0.397439923687)*x[0]**o+(0.595274583577)*x[0]+(-0.292338728991)*x[1]**o+(-0.314101057271)*x[1]
        arg[3,4,1,1]=(-0.403584993513)*x[0]**o+(-0.758728905384)*x[0]+(-0.182501473325)*x[1]**o+(-0.0942108622967)*x[1]
        arg[3,4,2,0]=(-0.263608698751)*x[0]**o+(-0.591451983536)*x[0]+(0.620412677389)*x[1]**o+(0.600005456466)*x[1]
        arg[3,4,2,1]=(-0.633502278259)*x[0]**o+(0.733127573934)*x[0]+(-0.765468966987)*x[1]**o+(-0.24211874772)*x[1]
        ref[0,0,0,0]=(-0.750823220445)/(o+1.)+(0.200562218922)+(-0.177078428867)*0.5**o
        ref[0,0,0,1]=(-0.0127659464437)/(o+1.)+(-0.374838430629)+(-0.682838671264)*0.5**o
        ref[0,0,1,0]=(-0.932452125435)/(o+1.)+(0.64729240195)+(0.352439319413)*0.5**o
        ref[0,0,1,1]=(-0.702758889809)/(o+1.)+(0.143291706728)+(-0.354033408023)*0.5**o
        ref[0,0,2,0]=(-0.535244256862)/(o+1.)+(-0.0166533987843)+(0.706826713479)*0.5**o
        ref[0,0,2,1]=(0.742770093639)/(o+1.)+(0.0478036100749)+(-0.893497856528)*0.5**o
        ref[0,1,0,0]=(-0.7740511254)/(o+1.)+(0.151807063155)+(-0.536624329097)*0.5**o
        ref[0,1,0,1]=(0.196975646933)/(o+1.)+(-0.359195588208)+(0.388396767346)*0.5**o
        ref[0,1,1,0]=(0.372752949038)/(o+1.)+(-0.609511042662)+(-0.125771297178)*0.5**o
        ref[0,1,1,1]=(-0.726599229867)/(o+1.)+(-0.0676212180057)+(-0.313637256508)*0.5**o
        ref[0,1,2,0]=(1.60583902129e-05)/(o+1.)+(0.305063814416)+(0.92425817047)*0.5**o
        ref[0,1,2,1]=(-0.888060617134)/(o+1.)+(0.1507514326)+(-0.0981827607557)*0.5**o
        ref[0,2,0,0]=(-0.228336105988)/(o+1.)+(0.117775628113)+(-0.871590768072)*0.5**o
        ref[0,2,0,1]=(-0.509662470381)/(o+1.)+(0.0160681615592)+(-0.628675127006)*0.5**o
        ref[0,2,1,0]=(-0.587902586916)/(o+1.)+(0.346247748963)+(-0.767528433339)*0.5**o
        ref[0,2,1,1]=(-0.78470263647)/(o+1.)+(0.419883543895)+(0.388984380682)*0.5**o
        ref[0,2,2,0]=(-0.637159792471)/(o+1.)+(0.187921397135)+(-0.910514531543)*0.5**o
        ref[0,2,2,1]=(-0.298806908011)/(o+1.)+(-0.745958567089)+(0.570535552656)*0.5**o
        ref[0,3,0,0]=(-0.91370863065)/(o+1.)+(0.136099326137)+(0.916864925136)*0.5**o
        ref[0,3,0,1]=(0.699832871063)/(o+1.)+(0.0843138801635)+(-0.851147484632)*0.5**o
        ref[0,3,1,0]=(0.0203170502559)/(o+1.)+(0.687200666868)+(0.461273862468)*0.5**o
        ref[0,3,1,1]=(-0.00587359808953)/(o+1.)+(0.238032934986)+(0.63984877633)*0.5**o
        ref[0,3,2,0]=(0.689553978928)/(o+1.)+(-0.00771127379434)+(-0.159416471436)*0.5**o
        ref[0,3,2,1]=(0.0370277809126)/(o+1.)+(0.00770755686953)+(0.782397836312)*0.5**o
        ref[0,4,0,0]=(0.110218625254)/(o+1.)+(0.288682475662)+(-0.0063433205485)*0.5**o
        ref[0,4,0,1]=(0.910723121611)/(o+1.)+(0.401925056417)+(-0.128215646666)*0.5**o
        ref[0,4,1,0]=(-0.281954207008)/(o+1.)+(-0.117618175276)+(-0.604528735513)*0.5**o
        ref[0,4,1,1]=(-0.761226249914)/(o+1.)+(0.0260675614951)+(0.65096807032)*0.5**o
        ref[0,4,2,0]=(-0.780671499903)/(o+1.)+(-0.244952280649)+(-0.710718948634)*0.5**o
        ref[0,4,2,1]=(0.709902333411)/(o+1.)+(0.63468720432)+(-0.0443187506467)*0.5**o
        ref[1,0,0,0]=(-0.713098084853)/(o+1.)+(0.026951171138)+(0.731890221753)*0.5**o
        ref[1,0,0,1]=(-0.427105331069)/(o+1.)+(0.312079508459)+(-0.130335125365)*0.5**o
        ref[1,0,1,0]=(0.344459670502)/(o+1.)+(-0.0918192826838)+(-0.203908509107)*0.5**o
        ref[1,0,1,1]=(-0.0683630778679)/(o+1.)+(0.414378635447)+(0.186770929724)*0.5**o
        ref[1,0,2,0]=(0.127955472707)/(o+1.)+(-0.147180699)+(-0.891275312975)*0.5**o
        ref[1,0,2,1]=(-0.052031562679)/(o+1.)+(0.955043298554)+(-0.74620344519)*0.5**o
        ref[1,1,0,0]=(0.783714758356)/(o+1.)+(0.226618827568)+(0.871423198418)*0.5**o
        ref[1,1,0,1]=(-0.00888785913795)/(o+1.)+(-0.333567151999)+(0.943197824902)*0.5**o
        ref[1,1,1,0]=(0.481063204532)/(o+1.)+(0.224237358325)+(-0.74971981056)*0.5**o
        ref[1,1,1,1]=(-0.0110588803484)/(o+1.)+(-0.535252745501)+(-0.431776603043)*0.5**o
        ref[1,1,2,0]=(-0.646224518073)/(o+1.)+(-0.208035583419)+(0.578105297721)*0.5**o
        ref[1,1,2,1]=(-0.101534585934)/(o+1.)+(-0.440109735326)+(0.442066689582)*0.5**o
        ref[1,2,0,0]=(0.185493587693)/(o+1.)+(0.184679874576)+(-0.206702379228)*0.5**o
        ref[1,2,0,1]=(0.141622971369)/(o+1.)+(0.667962084728)+(0.79092697362)*0.5**o
        ref[1,2,1,0]=(0.249095964198)/(o+1.)+(0.24773637535)+(0.914281902407)*0.5**o
        ref[1,2,1,1]=(0.262528290838)/(o+1.)+(0.655403162506)+(-0.602926709108)*0.5**o
        ref[1,2,2,0]=(0.144972402754)/(o+1.)+(-0.408191062607)+(-0.724435531493)*0.5**o
        ref[1,2,2,1]=(-0.150439352532)/(o+1.)+(-0.16492771172)+(-0.611538590545)*0.5**o
        ref[1,3,0,0]=(-0.509426399026)/(o+1.)+(-0.267418424334)+(-0.913553947133)*0.5**o
        ref[1,3,0,1]=(0.540358729462)/(o+1.)+(0.47155354967)+(-0.709834132519)*0.5**o
        ref[1,3,1,0]=(0.0135015706177)/(o+1.)+(0.0977365649293)+(-0.30788139278)*0.5**o
        ref[1,3,1,1]=(0.076082439418)/(o+1.)+(0.331185729735)+(-0.243327801257)*0.5**o
        ref[1,3,2,0]=(-0.0741034599215)/(o+1.)+(0.168033143879)+(-0.0290765899224)*0.5**o
        ref[1,3,2,1]=(-0.431665912498)/(o+1.)+(0.72501894795)+(-0.751374584848)*0.5**o
        ref[1,4,0,0]=(-0.872729326815)/(o+1.)+(0.124838237631)+(-0.940864464113)*0.5**o
        ref[1,4,0,1]=(0.399921161647)/(o+1.)+(0.309202877726)+(0.620544898911)*0.5**o
        ref[1,4,1,0]=(-0.525522030503)/(o+1.)+(-0.658539395994)+(-0.398096053023)*0.5**o
        ref[1,4,1,1]=(0.785902674573)/(o+1.)+(-0.191514764351)+(0.277239330261)*0.5**o
        ref[1,4,2,0]=(0.728305062026)/(o+1.)+(0.555597269878)+(-0.469242672532)*0.5**o
        ref[1,4,2,1]=(0.279230829643)/(o+1.)+(-0.560623089892)+(0.828113301834)*0.5**o
        ref[2,0,0,0]=(0.302409732402)/(o+1.)+(-0.126038760113)+(-0.724500208669)*0.5**o
        ref[2,0,0,1]=(0.190028790905)/(o+1.)+(-0.349001779766)+(-0.133300605917)*0.5**o
        ref[2,0,1,0]=(0.745204383075)/(o+1.)+(-0.0413765302013)+(0.828019154028)*0.5**o
        ref[2,0,1,1]=(-0.104712073949)/(o+1.)+(0.156001957276)+(-0.653962912288)*0.5**o
        ref[2,0,2,0]=(0.372296357092)/(o+1.)+(-0.394676696163)+(0.21572653818)*0.5**o
        ref[2,0,2,1]=(-0.33055202826)/(o+1.)+(0.56112591693)+(-0.0943992743407)*0.5**o
        ref[2,1,0,0]=(-0.891481528328)/(o+1.)+(0.120811569592)+(-0.812500024009)*0.5**o
        ref[2,1,0,1]=(-0.000696143847979)/(o+1.)+(0.2288950821)+(-0.958152880518)*0.5**o
        ref[2,1,1,0]=(0.128356074186)/(o+1.)+(-0.459272964402)+(-0.81109120479)*0.5**o
        ref[2,1,1,1]=(-0.995176229907)/(o+1.)+(-0.477423619244)+(-0.72470858095)*0.5**o
        ref[2,1,2,0]=(0.306860971474)/(o+1.)+(-0.0690615094134)+(-0.191810147267)*0.5**o
        ref[2,1,2,1]=(-0.195907590382)/(o+1.)+(-0.292487576314)+(-0.867447323913)*0.5**o
        ref[2,2,0,0]=(-0.0928333646602)/(o+1.)+(-0.606886504006)+(0.983782298896)*0.5**o
        ref[2,2,0,1]=(0.455030678927)/(o+1.)+(-0.120273395167)+(0.121619221979)*0.5**o
        ref[2,2,1,0]=(-0.0239503446136)/(o+1.)+(-0.149680272064)+(0.378726440151)*0.5**o
        ref[2,2,1,1]=(0.906235883406)/(o+1.)+(0.19594045067)+(0.442586981104)*0.5**o
        ref[2,2,2,0]=(-0.520596320159)/(o+1.)+(-0.241720805474)+(-0.564560890159)*0.5**o
        ref[2,2,2,1]=(0.792289256264)/(o+1.)+(-0.180045121676)+(0.215858165597)*0.5**o
        ref[2,3,0,0]=(-0.551360906631)/(o+1.)+(-0.776218533854)+(-0.596335978041)*0.5**o
        ref[2,3,0,1]=(0.821070238869)/(o+1.)+(0.00199201312847)+(-0.913806992124)*0.5**o
        ref[2,3,1,0]=(-0.43240265235)/(o+1.)+(-0.535669106134)+(-0.0831972104302)*0.5**o
        ref[2,3,1,1]=(0.72072318435)/(o+1.)+(0.0330920562179)+(0.643878326697)*0.5**o
        ref[2,3,2,0]=(0.754511827443)/(o+1.)+(-0.154030524635)+(0.843652876743)*0.5**o
        ref[2,3,2,1]=(0.261249117327)/(o+1.)+(-0.03562161515)+(0.0413857115672)*0.5**o
        ref[2,4,0,0]=(-0.0678823396451)/(o+1.)+(-0.192582591092)+(-0.607477693828)*0.5**o
        ref[2,4,0,1]=(0.0457149578306)/(o+1.)+(-0.0964434622791)+(-0.618172945146)*0.5**o
        ref[2,4,1,0]=(0.630735604709)/(o+1.)+(-0.0628061501569)+(0.722668769621)*0.5**o
        ref[2,4,1,1]=(-0.693659242703)/(o+1.)+(0.0588523188405)+(-0.847113860795)*0.5**o
        ref[2,4,2,0]=(0.705207362097)/(o+1.)+(0.0641733045603)+(-0.437578500878)*0.5**o
        ref[2,4,2,1]=(0.860707329294)/(o+1.)+(-0.0794358167735)+(0.329570005179)*0.5**o
        ref[3,0,0,0]=(0.549409370786)/(o+1.)+(-0.0827404400499)+(0.932602317046)*0.5**o
        ref[3,0,0,1]=(0.222376331882)/(o+1.)+(0.266122075595)+(0.975116319103)*0.5**o
        ref[3,0,1,0]=(-0.481194397466)/(o+1.)+(0.840391794852)+(0.019317277942)*0.5**o
        ref[3,0,1,1]=(0.737901448601)/(o+1.)+(0.0468573264614)+(-0.393162543578)*0.5**o
        ref[3,0,2,0]=(0.278242150139)/(o+1.)+(-0.0620258662406)+(0.458096711279)*0.5**o
        ref[3,0,2,1]=(-0.489134379327)/(o+1.)+(-0.252931749044)+(0.307111052767)*0.5**o
        ref[3,1,0,0]=(0.508808683)/(o+1.)+(-0.594037891099)+(0.420707084879)*0.5**o
        ref[3,1,0,1]=(0.198212694401)/(o+1.)+(-0.23511909641)+(-0.745253594401)*0.5**o
        ref[3,1,1,0]=(-0.0741524230667)/(o+1.)+(-0.166487499516)+(0.535215252931)*0.5**o
        ref[3,1,1,1]=(-0.664372557476)/(o+1.)+(-0.161339307185)+(0.193417593556)*0.5**o
        ref[3,1,2,0]=(0.408815221365)/(o+1.)+(0.556969143078)+(0.553690237434)*0.5**o
        ref[3,1,2,1]=(-0.111028404397)/(o+1.)+(0.28519971884)+(-0.211683249218)*0.5**o
        ref[3,2,0,0]=(-0.228385234131)/(o+1.)+(0.068833285335)+(-0.866119051438)*0.5**o
        ref[3,2,0,1]=(-0.709059272627)/(o+1.)+(0.273296245522)+(-0.146618266885)*0.5**o
        ref[3,2,1,0]=(-0.651609755171)/(o+1.)+(-0.717521398332)+(0.622973500199)*0.5**o
        ref[3,2,1,1]=(0.106659392436)/(o+1.)+(-0.0937511533137)+(-0.538565077894)*0.5**o
        ref[3,2,2,0]=(-0.339096900793)/(o+1.)+(0.616553014737)+(-0.325897413686)*0.5**o
        ref[3,2,2,1]=(-0.10651751912)/(o+1.)+(-0.0363728631586)+(0.861430484293)*0.5**o
        ref[3,3,0,0]=(0.230063809846)/(o+1.)+(0.0083457439947)+(0.0890838652267)*0.5**o
        ref[3,3,0,1]=(-0.309971427898)/(o+1.)+(0.0417460043682)+(-0.969251937403)*0.5**o
        ref[3,3,1,0]=(0.55773788655)/(o+1.)+(0.317014803113)+(0.640075232231)*0.5**o
        ref[3,3,1,1]=(-0.655757122598)/(o+1.)+(-0.186359376395)+(-0.350002630494)*0.5**o
        ref[3,3,2,0]=(0.310310045807)/(o+1.)+(0.229515713776)+(-0.262399316954)*0.5**o
        ref[3,3,2,1]=(-0.844329846245)/(o+1.)+(0.412676046356)+(-0.977039760257)*0.5**o
        ref[3,4,0,0]=(0.39699250956)/(o+1.)+(-0.162896058076)+(0.749594099085)*0.5**o
        ref[3,4,0,1]=(-0.34125011876)/(o+1.)+(-0.577475727421)+(0.913703789336)*0.5**o
        ref[3,4,1,0]=(-0.292338728991)/(o+1.)+(0.140586763153)+(-0.397439923687)*0.5**o
        ref[3,4,1,1]=(-0.182501473325)/(o+1.)+(-0.42646988384)+(-0.403584993513)*0.5**o
        ref[3,4,2,0]=(0.620412677389)/(o+1.)+(0.00427673646471)+(-0.263608698751)*0.5**o
        ref[3,4,2,1]=(-0.765468966987)/(o+1.)+(0.245504413107)+(-0.633502278259)*0.5**o
      else:
        arg[0,0,0,0]=(0.480896043196)*x[0]**o+(0.113355160927)*x[0]+(0.498630339771)*x[1]**o+(0.0149514785179)*x[1]+(-0.944439525708)*x[2]**o+(-0.349299625744)*x[2]
        arg[0,0,0,1]=(-0.609632888023)*x[0]**o+(-0.576508904483)*x[0]+(-0.0610487008887)*x[1]**o+(-0.844087384492)*x[1]+(-0.715427733517)*x[2]**o+(-0.263424969452)*x[2]
        arg[0,0,1,0]=(0.205065167181)*x[0]**o+(-0.840225514826)*x[0]+(0.808161497555)*x[1]**o+(0.685902712814)*x[1]+(-0.625183952417)*x[2]**o+(-0.0947932775112)*x[2]
        arg[0,0,1,1]=(0.875716461075)*x[0]**o+(0.814425331995)*x[0]+(-0.805331622235)*x[1]**o+(-0.287737068993)*x[1]+(-0.365715978159)*x[2]**o+(0.689907502778)*x[2]
        arg[0,0,2,0]=(0.711372030289)*x[0]**o+(-0.393612062405)*x[0]+(0.642793863702)*x[1]**o+(-0.66636178423)*x[1]+(0.00488425355017)*x[2]**o+(-0.444523817626)*x[2]
        arg[0,0,2,1]=(-0.0536842545617)*x[0]**o+(-0.714279418499)*x[0]+(-0.487452457649)*x[1]**o+(0.917805820162)*x[1]+(0.445007135027)*x[2]**o+(0.101201310289)*x[2]
        arg[0,1,0,0]=(-0.710091748251)*x[0]**o+(0.470243562863)*x[0]+(-0.0178862343634)*x[1]**o+(0.393644042332)*x[1]+(-0.952839337577)*x[2]**o+(-0.152392013939)*x[2]
        arg[0,1,0,1]=(0.250989366649)*x[0]**o+(0.592158236555)*x[0]+(0.698185234877)*x[1]**o+(0.112154263635)*x[1]+(0.843915127813)*x[2]**o+(-0.825331842242)*x[2]
        arg[0,1,1,0]=(0.690360096543)*x[0]**o+(0.78906430602)*x[0]+(0.596189773531)*x[1]**o+(-0.293452572358)*x[1]+(0.737565980909)*x[2]**o+(-0.96988090057)*x[2]
        arg[0,1,1,1]=(-0.815220933301)*x[0]**o+(-0.875224889387)*x[0]+(-0.0562950344908)*x[1]**o+(0.56567431969)*x[1]+(0.133489969036)*x[2]**o+(0.299403989148)*x[2]
        arg[0,1,2,0]=(0.169442617085)*x[0]**o+(-0.413300800662)*x[0]+(0.580773652363)*x[1]**o+(-0.627706012397)*x[1]+(-0.730218498677)*x[2]**o+(0.587233195345)*x[2]
        arg[0,1,2,1]=(-0.110636656112)*x[0]**o+(-0.302213944624)*x[0]+(0.643229732019)*x[1]**o+(-0.651805494828)*x[1]+(-0.528276901455)*x[2]**o+(-0.272057635286)*x[2]
        arg[0,2,0,0]=(0.340661037085)*x[0]**o+(-0.672207039496)*x[0]+(0.676291394824)*x[1]**o+(0.596998737167)*x[1]+(-0.689800391841)*x[2]**o+(-0.134273731064)*x[2]
        arg[0,2,0,1]=(0.185282186176)*x[0]**o+(0.252489447393)*x[0]+(-0.0447305595755)*x[1]**o+(-0.558361084175)*x[1]+(0.922681220429)*x[2]**o+(-0.484588607971)*x[2]
        arg[0,2,1,0]=(0.238772787004)*x[0]**o+(0.987311136495)*x[0]+(0.215079265839)*x[1]**o+(-0.797762453244)*x[1]+(0.0329055379323)*x[2]**o+(-0.88999503049)*x[2]
        arg[0,2,1,1]=(-0.175884373587)*x[0]**o+(0.806690051494)*x[0]+(0.0086744009376)*x[1]**o+(0.286131063746)*x[1]+(0.536839534689)*x[2]**o+(-0.0946282917135)*x[2]
        arg[0,2,2,0]=(-0.615695203188)*x[0]**o+(0.600565627235)*x[0]+(-0.869759996675)*x[1]**o+(0.648192341374)*x[1]+(-0.952728999251)*x[2]**o+(0.564068973312)*x[2]
        arg[0,2,2,1]=(-0.0992379560876)*x[0]**o+(-0.897179656441)*x[0]+(-0.519600820875)*x[1]**o+(0.70507167679)*x[1]+(0.903640275297)*x[2]**o+(-0.645039734963)*x[2]
        arg[0,3,0,0]=(0.432415072474)*x[0]**o+(-0.748137008742)*x[0]+(0.875883324091)*x[1]**o+(0.706172632952)*x[1]+(-0.47486904255)*x[2]**o+(0.238106343647)*x[2]
        arg[0,3,0,1]=(-0.656397468061)*x[0]**o+(-0.530867058489)*x[0]+(-0.904148914114)*x[1]**o+(0.87213389632)*x[1]+(0.0612056123548)*x[2]**o+(0.288620109264)*x[2]
        arg[0,3,1,0]=(-0.472896985897)*x[0]**o+(0.83328574859)*x[0]+(0.744310819801)*x[1]**o+(0.0601035189502)*x[1]+(-0.198693260491)*x[2]**o+(-0.882304016543)*x[2]
        arg[0,3,1,1]=(-0.515568063685)*x[0]**o+(0.484765564236)*x[0]+(0.754304059424)*x[1]**o+(0.654790289063)*x[1]+(-0.102335538236)*x[2]**o+(-0.0705158755887)*x[2]
        arg[0,3,2,0]=(-0.709802317012)*x[0]**o+(-0.346551215683)*x[0]+(0.312943005623)*x[1]**o+(0.476828461567)*x[1]+(0.901171463566)*x[2]**o+(0.538981245022)*x[2]
        arg[0,3,2,1]=(-0.20057625974)*x[0]**o+(0.0458666109661)*x[0]+(0.909610730794)*x[1]**o+(0.145637065192)*x[1]+(0.666420073244)*x[2]**o+(0.0550124082656)*x[2]
        arg[0,4,0,0]=(-0.726921583373)*x[0]**o+(0.891306726707)*x[0]+(-0.189151461127)*x[1]**o+(0.479274047923)*x[1]+(-0.129077261233)*x[2]**o+(-0.873599130413)*x[2]
        arg[0,4,0,1]=(0.765272127757)*x[0]**o+(0.159307695759)*x[0]+(-0.569162341331)*x[1]**o+(-0.530158933095)*x[1]+(-0.216727788754)*x[2]**o+(-0.519551969741)*x[2]
        arg[0,4,1,0]=(-0.613269202518)*x[0]**o+(0.612846846793)*x[0]+(-0.187285796021)*x[1]**o+(-0.462603518143)*x[1]+(-0.739162067312)*x[2]**o+(0.0737393951189)*x[2]
        arg[0,4,1,1]=(-0.727485674691)*x[0]**o+(-0.461427895477)*x[0]+(0.917029475887)*x[1]**o+(0.265874989879)*x[1]+(0.602505162351)*x[2]**o+(-0.372206475027)*x[2]
        arg[0,4,2,0]=(-0.663071984573)*x[0]**o+(0.511217165998)*x[0]+(-0.472765672078)*x[1]**o+(0.814962161218)*x[1]+(-0.608890631297)*x[2]**o+(-0.284872941025)*x[2]
        arg[0,4,2,1]=(0.893396132015)*x[0]**o+(0.183468149957)*x[0]+(0.647812435076)*x[1]**o+(-0.398672312216)*x[1]+(0.6359553902)*x[2]**o+(0.485509522238)*x[2]
        arg[1,0,0,0]=(-0.0771632829864)*x[0]**o+(-0.723025089864)*x[0]+(0.486209418481)*x[1]**o+(0.0869704919133)*x[1]+(-0.365499197901)*x[2]**o+(-0.0796713538898)*x[2]
        arg[1,0,0,1]=(0.133617787097)*x[0]**o+(0.993872304226)*x[0]+(-0.898504795319)*x[1]**o+(0.656793416434)*x[1]+(-0.821841885906)*x[2]**o+(0.726972748423)*x[2]
        arg[1,0,1,0]=(0.174277385391)*x[0]**o+(-0.301857401564)*x[0]+(-0.846153302312)*x[1]**o+(-0.559076632516)*x[1]+(-0.89870762709)*x[2]**o+(0.640171641512)*x[2]
        arg[1,0,1,1]=(0.795443369588)*x[0]**o+(-0.346082013333)*x[0]+(-0.439713049567)*x[1]**o+(0.860680213906)*x[1]+(-0.466761099917)*x[2]**o+(-0.843656974958)*x[2]
        arg[1,0,2,0]=(0.482814141263)*x[0]**o+(0.142978643528)*x[0]+(-0.735291704767)*x[1]**o+(-0.925675892787)*x[1]+(0.747235439093)*x[2]**o+(-0.569454067374)*x[2]
        arg[1,0,2,1]=(0.619359374773)*x[0]**o+(-0.547172728307)*x[0]+(0.0829642963797)*x[1]**o+(-0.0180882163346)*x[1]+(0.328627445196)*x[2]**o+(-0.274608686384)*x[2]
        arg[1,1,0,0]=(-0.510375454058)*x[0]**o+(-0.779894165181)*x[0]+(0.656168894408)*x[1]**o+(-0.449484133421)*x[1]+(-0.147284453744)*x[2]**o+(-0.488384108271)*x[2]
        arg[1,1,0,1]=(-0.302705575836)*x[0]**o+(0.256670307547)*x[0]+(0.110465379068)*x[1]**o+(0.319291387545)*x[1]+(0.408832172454)*x[2]**o+(0.718793126931)*x[2]
        arg[1,1,1,0]=(0.520882033451)*x[0]**o+(0.412524861348)*x[0]+(0.761080123209)*x[1]**o+(0.401223078157)*x[1]+(-0.65945839721)*x[2]**o+(0.71349489395)*x[2]
        arg[1,1,1,1]=(0.624622427594)*x[0]**o+(0.847907716976)*x[0]+(0.855268296474)*x[1]**o+(-0.671685662164)*x[1]+(-0.407895715308)*x[2]**o+(-0.719054767733)*x[2]
        arg[1,1,2,0]=(-0.044254701409)*x[0]**o+(0.72911290175)*x[0]+(0.869467934211)*x[1]**o+(0.613006764744)*x[1]+(0.678602796496)*x[2]**o+(0.388923815748)*x[2]
        arg[1,1,2,1]=(-0.240571107795)*x[0]**o+(0.838382917252)*x[0]+(0.0342933298383)*x[1]**o+(0.595421789197)*x[1]+(-0.989088178853)*x[2]**o+(-0.20002339454)*x[2]
        arg[1,2,0,0]=(0.0596937021722)*x[0]**o+(0.196768537308)*x[0]+(0.0834459161806)*x[1]**o+(0.194049508373)*x[1]+(-0.097309631361)*x[2]**o+(-0.867555781902)*x[2]
        arg[1,2,0,1]=(-0.935541542647)*x[0]**o+(-0.00702430636955)*x[0]+(-0.384657002538)*x[1]**o+(0.186884376746)*x[1]+(0.97207387841)*x[2]**o+(0.452460932869)*x[2]
        arg[1,2,1,0]=(0.276741406646)*x[0]**o+(0.838348610585)*x[0]+(-0.403054578066)*x[1]**o+(0.655318135938)*x[1]+(0.37614761635)*x[2]**o+(0.82736662494)*x[2]
        arg[1,2,1,1]=(0.828731604537)*x[0]**o+(0.762067030749)*x[0]+(-0.881827056079)*x[1]**o+(0.250928989647)*x[1]+(-0.771040367979)*x[2]**o+(0.239950799138)*x[2]
        arg[1,2,2,0]=(-0.380797583569)*x[0]**o+(0.112805267222)*x[0]+(-0.454730562411)*x[1]**o+(0.926061735219)*x[1]+(0.348331256905)*x[2]**o+(-0.725641369775)*x[2]
        arg[1,2,2,1]=(0.257022926473)*x[0]**o+(0.0160557858816)*x[0]+(-0.291367922966)*x[1]**o+(-0.611341260765)*x[1]+(0.232472950268)*x[2]**o+(-0.738139185521)*x[2]
        arg[1,3,0,0]=(-0.645340088797)*x[0]**o+(0.480235460274)*x[0]+(0.135346246274)*x[1]**o+(-0.560730820351)*x[1]+(-0.635582025321)*x[2]**o+(-0.476912924392)*x[2]
        arg[1,3,0,1]=(0.124132693667)*x[0]**o+(-0.813993821983)*x[0]+(-0.108593659227)*x[1]**o+(-0.432186668568)*x[1]+(0.209382070426)*x[2]**o+(0.928912669667)*x[2]
        arg[1,3,1,0]=(-0.7660131868)*x[0]**o+(-0.30274186797)*x[0]+(0.140498262981)*x[1]**o+(0.993084314471)*x[1]+(0.684043026981)*x[2]**o+(-0.339724714081)*x[2]
        arg[1,3,1,1]=(-0.150578257351)*x[0]**o+(-0.674840941109)*x[0]+(0.469240334383)*x[1]**o+(0.729963338278)*x[1]+(-0.800829664334)*x[2]**o+(0.0467922871395)*x[2]
        arg[1,3,2,0]=(-0.336661555676)*x[0]**o+(0.705661383379)*x[0]+(-0.76993460355)*x[1]**o+(0.459144989218)*x[1]+(0.545901025663)*x[2]**o+(0.385988013614)*x[2]
        arg[1,3,2,1]=(0.165928745462)*x[0]**o+(0.540718600752)*x[0]+(0.172773708803)*x[1]**o+(0.253664803784)*x[1]+(-0.264896626947)*x[2]**o+(-0.478577475495)*x[2]
        arg[1,4,0,0]=(0.283586653542)*x[0]**o+(-0.104364884571)*x[0]+(-0.840555806526)*x[1]**o+(0.701868415915)*x[1]+(-0.290943689679)*x[2]**o+(-0.73024942052)*x[2]
        arg[1,4,0,1]=(-0.772849835334)*x[0]**o+(-0.432170589951)*x[0]+(-0.376099423267)*x[1]**o+(0.0215357491826)*x[1]+(-0.346787011464)*x[2]**o+(0.620411709813)*x[2]
        arg[1,4,1,0]=(-0.718524564964)*x[0]**o+(0.119936682579)*x[0]+(-0.994276256462)*x[1]**o+(-0.839783920473)*x[1]+(-0.454576336782)*x[2]**o+(-0.515489988915)*x[2]
        arg[1,4,1,1]=(-0.771664154234)*x[0]**o+(-0.584052299067)*x[0]+(-0.0233756210307)*x[1]**o+(-0.174442816636)*x[1]+(-0.702383477714)*x[2]**o+(-0.279952858144)*x[2]
        arg[1,4,2,0]=(-0.591902790978)*x[0]**o+(-0.151826578123)*x[0]+(0.923906788219)*x[1]**o+(0.301090558476)*x[1]+(-0.518314647811)*x[2]**o+(-0.784331717681)*x[2]
        arg[1,4,2,1]=(0.303567440092)*x[0]**o+(0.744948852961)*x[0]+(0.719939071212)*x[1]**o+(-0.161138638319)*x[1]+(0.924953261766)*x[2]**o+(-0.875565390131)*x[2]
        arg[2,0,0,0]=(0.197574550087)*x[0]**o+(-0.962018785397)*x[0]+(0.498378037582)*x[1]**o+(-0.334199169606)*x[1]+(0.852426750588)*x[2]**o+(-0.650173474618)*x[2]
        arg[2,0,0,1]=(0.0663499012532)*x[0]**o+(0.00812461763474)*x[0]+(0.237233309304)*x[1]**o+(-0.839377881658)*x[1]+(-0.67246872567)*x[2]**o+(0.677471261363)*x[2]
        arg[2,0,1,0]=(-0.789090971877)*x[0]**o+(-0.387724517273)*x[0]+(0.455954877641)*x[1]**o+(-0.356995254842)*x[1]+(-0.897019846935)*x[2]**o+(-0.0703543939474)*x[2]
        arg[2,0,1,1]=(0.222873274437)*x[0]**o+(0.405228982254)*x[0]+(-0.532697993554)*x[1]**o+(-0.836340143789)*x[1]+(-0.139525062661)*x[2]**o+(-0.916307770461)*x[2]
        arg[2,0,2,0]=(-0.300828039715)*x[0]**o+(-0.584457605313)*x[0]+(-0.443125109903)*x[1]**o+(0.292734004842)*x[1]+(-0.55875691535)*x[2]**o+(0.391237822545)*x[2]
        arg[2,0,2,1]=(-0.501507304569)*x[0]**o+(-0.203549089015)*x[0]+(-0.113050138121)*x[1]**o+(0.712978728546)*x[1]+(-0.642617274635)*x[2]**o+(-0.307119076865)*x[2]
        arg[2,1,0,0]=(-0.809237327867)*x[0]**o+(0.984684980548)*x[0]+(-0.643310231704)*x[1]**o+(0.759430082065)*x[1]+(-0.393902220205)*x[2]**o+(-0.900173371318)*x[2]
        arg[2,1,0,1]=(-0.722795722506)*x[0]**o+(0.433309366611)*x[0]+(-0.559871890844)*x[1]**o+(0.417301432603)*x[1]+(-0.364192390887)*x[2]**o+(0.64549736058)*x[2]
        arg[2,1,1,0]=(-0.551325611251)*x[0]**o+(-0.638520772682)*x[0]+(0.222655558982)*x[1]**o+(-0.230002147207)*x[1]+(-0.290821163957)*x[2]**o+(0.0647728628594)*x[2]
        arg[2,1,1,1]=(0.743453045588)*x[0]**o+(-0.901163108996)*x[0]+(-0.546031302921)*x[1]**o+(-0.512614476111)*x[1]+(-0.886430682715)*x[2]**o+(0.720650678026)*x[2]
        arg[2,1,2,0]=(-0.033787671958)*x[0]**o+(0.930927279443)*x[0]+(-0.530482747719)*x[1]**o+(0.714053784656)*x[1]+(-0.0978565618487)*x[2]**o+(-0.0423072209447)*x[2]
        arg[2,1,2,1]=(0.127121989623)*x[0]**o+(0.809654796091)*x[0]+(0.718533701795)*x[1]**o+(0.542738494557)*x[1]+(-0.342918607246)*x[2]**o+(-0.627991145352)*x[2]
        arg[2,2,0,0]=(-0.867530011926)*x[0]**o+(0.870295856967)*x[0]+(-0.13043222912)*x[1]**o+(0.2670058777)*x[1]+(0.395016123247)*x[2]**o+(-0.811207959629)*x[2]
        arg[2,2,0,1]=(-0.00858729113056)*x[0]**o+(-0.885551110851)*x[0]+(0.820441870015)*x[1]**o+(-0.348896120549)*x[1]+(0.295475751596)*x[2]**o+(-0.672688762902)*x[2]
        arg[2,2,1,0]=(-0.982123891355)*x[0]**o+(0.232770408062)*x[0]+(0.374641341895)*x[1]**o+(-0.21375197203)*x[1]+(0.436992590849)*x[2]**o+(-0.473399371298)*x[2]
        arg[2,2,1,1]=(-0.415533190881)*x[0]**o+(0.0945615271507)*x[0]+(-0.281959759072)*x[1]**o+(0.0807942449559)*x[1]+(0.225501997354)*x[2]**o+(0.305247648387)*x[2]
        arg[2,2,2,0]=(-0.559916632426)*x[0]**o+(-0.372679495574)*x[0]+(0.392857820158)*x[1]**o+(0.726824319268)*x[1]+(0.36923535014)*x[2]**o+(-0.208169256022)*x[2]
        arg[2,2,2,1]=(-0.869055045156)*x[0]**o+(0.359188490398)*x[0]+(0.702110073265)*x[1]**o+(0.668757858248)*x[1]+(-0.499300337779)*x[2]**o+(0.0471054271112)*x[2]
        arg[2,3,0,0]=(0.754089271004)*x[0]**o+(-0.0240479506036)*x[0]+(-0.0775830745475)*x[1]**o+(0.215564363064)*x[1]+(-0.791969192801)*x[2]**o+(-0.130308630692)*x[2]
        arg[2,3,0,1]=(0.448970570476)*x[0]**o+(0.510346880872)*x[0]+(-0.561996764085)*x[1]**o+(0.217897696493)*x[1]+(-0.0610826665756)*x[2]**o+(0.387896322636)*x[2]
        arg[2,3,1,0]=(0.737635683404)*x[0]**o+(-0.35395540781)*x[0]+(0.0419652573208)*x[1]**o+(-0.519735381528)*x[1]+(0.739523085214)*x[2]**o+(0.808529094094)*x[2]
        arg[2,3,1,1]=(0.551348499194)*x[0]**o+(-0.40145493368)*x[0]+(0.211723169473)*x[1]**o+(-0.762751144824)*x[1]+(-0.17710033067)*x[2]**o+(0.688961224051)*x[2]
        arg[2,3,2,0]=(-0.870721462645)*x[0]**o+(-0.0356869329313)*x[0]+(0.277142920424)*x[1]**o+(-0.154474999666)*x[1]+(0.837001124911)*x[2]**o+(-0.233028214567)*x[2]
        arg[2,3,2,1]=(-0.234260935528)*x[0]**o+(0.913718165996)*x[0]+(0.385678747942)*x[1]**o+(0.237966356277)*x[1]+(0.382072927695)*x[2]**o+(0.627666395757)*x[2]
        arg[2,4,0,0]=(-0.474674316363)*x[0]**o+(-0.575533683959)*x[0]+(0.751489557666)*x[1]**o+(-0.0623565445313)*x[1]+(-0.517274743843)*x[2]**o+(-0.712143034168)*x[2]
        arg[2,4,0,1]=(0.259723799797)*x[0]**o+(0.345952822815)*x[0]+(0.169940774832)*x[1]**o+(-0.185154259609)*x[1]+(0.418430344681)*x[2]**o+(0.861837327531)*x[2]
        arg[2,4,1,0]=(-0.960970748682)*x[0]**o+(0.842706511282)*x[0]+(-0.771779045108)*x[1]**o+(-0.63675621795)*x[1]+(-0.317416556239)*x[2]**o+(0.287317968341)*x[2]
        arg[2,4,1,1]=(-0.753944077832)*x[0]**o+(0.570245260805)*x[0]+(-0.0226797611003)*x[1]**o+(0.0776773843077)*x[1]+(0.629536717673)*x[2]**o+(0.271216337699)*x[2]
        arg[2,4,2,0]=(-0.233304828171)*x[0]**o+(-0.217436697538)*x[0]+(-0.373618121496)*x[1]**o+(-0.342584906659)*x[1]+(0.0114484228954)*x[2]**o+(-0.639753178531)*x[2]
        arg[2,4,2,1]=(0.341373328517)*x[0]**o+(-0.432795268788)*x[0]+(-0.871263113926)*x[1]**o+(0.843590946949)*x[1]+(-0.653058566267)*x[2]**o+(-0.894366240333)*x[2]
        arg[3,0,0,0]=(-0.472652289489)*x[0]**o+(0.697119499826)*x[0]+(-0.195781503945)*x[1]**o+(-0.716872496009)*x[1]+(-0.159070341485)*x[2]**o+(-0.700548071467)*x[2]
        arg[3,0,0,1]=(-0.398674023611)*x[0]**o+(-0.690683017472)*x[0]+(0.957112237189)*x[1]**o+(-0.782702815224)*x[1]+(-0.249336230999)*x[2]**o+(0.22734983387)*x[2]
        arg[3,0,1,0]=(-0.0216721345971)*x[0]**o+(0.283765576696)*x[0]+(-0.35531833417)*x[1]**o+(0.974402352836)*x[1]+(-0.568820213603)*x[2]**o+(0.957610279127)*x[2]
        arg[3,0,1,1]=(0.0606265274442)*x[0]**o+(-0.941169961629)*x[0]+(0.804902933622)*x[1]**o+(-0.857158317629)*x[1]+(-0.227167312547)*x[2]**o+(0.945175823328)*x[2]
        arg[3,0,2,0]=(0.630444608821)*x[0]**o+(-0.192343222551)*x[0]+(0.0342249337232)*x[1]**o+(0.970401334961)*x[1]+(-0.608568046213)*x[2]**o+(-0.539374299133)*x[2]
        arg[3,0,2,1]=(-0.348127940577)*x[0]**o+(-0.895424161569)*x[0]+(0.918173840926)*x[1]**o+(0.99658998821)*x[1]+(-0.28733372377)*x[2]**o+(-0.516441029544)*x[2]
        arg[3,1,0,0]=(0.735069593965)*x[0]**o+(-0.727899981109)*x[0]+(-0.325639355247)*x[1]**o+(0.751064830331)*x[1]+(-0.297126073245)*x[2]**o+(-0.673421779764)*x[2]
        arg[3,1,0,1]=(0.00281719106153)*x[0]**o+(0.246929980679)*x[0]+(-0.48091940116)*x[1]**o+(0.501927059357)*x[1]+(-0.520099802736)*x[2]**o+(-0.902706012118)*x[2]
        arg[3,1,1,0]=(-0.341732637339)*x[0]**o+(-0.169802234575)*x[0]+(0.156725375114)*x[1]**o+(0.739176104936)*x[1]+(-0.374572295196)*x[2]**o+(-0.436641001832)*x[2]
        arg[3,1,1,1]=(0.644746847185)*x[0]**o+(-0.992126325722)*x[0]+(0.551602368714)*x[1]**o+(-0.839306768893)*x[1]+(-0.298981617653)*x[2]**o+(0.089795271605)*x[2]
        arg[3,1,2,0]=(0.382425223801)*x[0]**o+(-0.949095041278)*x[0]+(-0.135997266858)*x[1]**o+(-0.871114033611)*x[1]+(-0.422984170709)*x[2]**o+(-0.698818784989)*x[2]
        arg[3,1,2,1]=(0.630558573276)*x[0]**o+(0.90644556976)*x[0]+(0.512953870249)*x[1]**o+(-0.58514271049)*x[1]+(-0.0587225891088)*x[2]**o+(-0.390796465925)*x[2]
        arg[3,2,0,0]=(-0.404884771697)*x[0]**o+(-0.178144326223)*x[0]+(-0.606503995405)*x[1]**o+(0.190025878248)*x[1]+(0.735831203504)*x[2]**o+(-0.567394607398)*x[2]
        arg[3,2,0,1]=(0.28819895431)*x[0]**o+(0.456573144203)*x[0]+(0.00856899200079)*x[1]**o+(0.547827005333)*x[1]+(0.087428132552)*x[2]**o+(0.759650060552)*x[2]
        arg[3,2,1,0]=(0.949790710961)*x[0]**o+(-0.685602682574)*x[0]+(0.75278809536)*x[1]**o+(-0.876689834021)*x[1]+(0.0116394079709)*x[2]**o+(0.319113513167)*x[2]
        arg[3,2,1,1]=(0.619659549755)*x[0]**o+(-0.397388124185)*x[0]+(0.460679661901)*x[1]**o+(-0.507122241204)*x[1]+(0.985741194076)*x[2]**o+(0.916231863626)*x[2]
        arg[3,2,2,0]=(-0.457269257294)*x[0]**o+(-0.47265466694)*x[0]+(0.900051837969)*x[1]**o+(-0.148802803439)*x[1]+(-0.195540487054)*x[2]**o+(0.0784524130754)*x[2]
        arg[3,2,2,1]=(0.933553130797)*x[0]**o+(0.617362831046)*x[0]+(0.263616954879)*x[1]**o+(-0.0658887625105)*x[1]+(0.48091174024)*x[2]**o+(0.452935484504)*x[2]
        arg[3,3,0,0]=(0.825594594128)*x[0]**o+(0.350203407373)*x[0]+(0.211125321741)*x[1]**o+(0.194914147746)*x[1]+(-0.639432220749)*x[2]**o+(-0.678718852935)*x[2]
        arg[3,3,0,1]=(0.514394313629)*x[0]**o+(-0.595261005308)*x[0]+(-0.0219208527235)*x[1]**o+(-0.77204975709)*x[1]+(-0.483587370025)*x[2]**o+(-0.559464911044)*x[2]
        arg[3,3,1,0]=(0.821240511475)*x[0]**o+(-0.403492177329)*x[0]+(0.486132478046)*x[1]**o+(-0.0187529457547)*x[1]+(0.73209003456)*x[2]**o+(0.226791335797)*x[2]
        arg[3,3,1,1]=(-0.644256768839)*x[0]**o+(0.0372588784686)*x[0]+(0.726561650796)*x[1]**o+(0.518003809017)*x[1]+(0.738146681601)*x[2]**o+(-0.256036433171)*x[2]
        arg[3,3,2,0]=(0.534862520097)*x[0]**o+(0.275559826342)*x[0]+(0.292383677513)*x[1]**o+(-0.448938027415)*x[1]+(0.281053776971)*x[2]**o+(0.0119964402657)*x[2]
        arg[3,3,2,1]=(0.722071427786)*x[0]**o+(0.150280744346)*x[0]+(-0.0354786178164)*x[1]**o+(-0.703984035201)*x[1]+(0.105365795033)*x[2]**o+(-0.556567650379)*x[2]
        arg[3,4,0,0]=(0.188625467698)*x[0]**o+(0.737108753179)*x[0]+(0.481952725951)*x[1]**o+(0.841958412185)*x[1]+(-0.329504392569)*x[2]**o+(0.228456367852)*x[2]
        arg[3,4,0,1]=(0.181430484096)*x[0]**o+(0.307078178179)*x[0]+(0.262079226853)*x[1]**o+(0.111656765153)*x[1]+(0.15340094176)*x[2]**o+(-0.859774252333)*x[2]
        arg[3,4,1,0]=(0.213491785855)*x[0]**o+(-0.479441582203)*x[0]+(0.0527576231613)*x[1]**o+(0.328465118263)*x[1]+(-0.645719217198)*x[2]**o+(-0.550569446193)*x[2]
        arg[3,4,1,1]=(-0.401744411188)*x[0]**o+(-0.626446165761)*x[0]+(-0.191670239416)*x[1]**o+(0.866888407009)*x[1]+(-0.341220740645)*x[2]**o+(-0.742685031762)*x[2]
        arg[3,4,2,0]=(-0.0813611687514)*x[0]**o+(0.318680684889)*x[0]+(-0.780434724843)*x[1]**o+(-0.0902735483921)*x[1]+(0.586108990767)*x[2]**o+(0.934041577436)*x[2]
        arg[3,4,2,1]=(-0.0991425040958)*x[0]**o+(0.99274076773)*x[0]+(0.924248544389)*x[1]**o+(0.290576034321)*x[1]+(0.452104871421)*x[2]**o+(0.774017129246)*x[2]
        ref[0,0,0,0]=(-0.445809185937)/(o+1.)+(-0.11049649315)+(0.480896043196)*0.5**o
        ref[0,0,0,1]=(-0.776476434405)/(o+1.)+(-0.842010629213)+(-0.609632888023)*0.5**o
        ref[0,0,1,0]=(0.182977545138)/(o+1.)+(-0.124558039762)+(0.205065167181)*0.5**o
        ref[0,0,1,1]=(-1.17104760039)/(o+1.)+(0.608297882891)+(0.875716461075)*0.5**o
        ref[0,0,2,0]=(0.647678117252)/(o+1.)+(-0.752248832131)+(0.711372030289)*0.5**o
        ref[0,0,2,1]=(-0.0424453226222)/(o+1.)+(0.152363855976)+(-0.0536842545617)*0.5**o
        ref[0,1,0,0]=(-0.97072557194)/(o+1.)+(0.355747795628)+(-0.710091748251)*0.5**o
        ref[0,1,0,1]=(1.54210036269)/(o+1.)+(-0.0605096710259)+(0.250989366649)*0.5**o
        ref[0,1,1,0]=(1.33375575444)/(o+1.)+(-0.237134583454)+(0.690360096543)*0.5**o
        ref[0,1,1,1]=(0.0771949345448)/(o+1.)+(-0.00507329027461)+(-0.815220933301)*0.5**o
        ref[0,1,2,0]=(-0.149444846314)/(o+1.)+(-0.226886808857)+(0.169442617085)*0.5**o
        ref[0,1,2,1]=(0.114952830564)/(o+1.)+(-0.613038537369)+(-0.110636656112)*0.5**o
        ref[0,2,0,0]=(-0.0135089970172)/(o+1.)+(-0.104741016697)+(0.340661037085)*0.5**o
        ref[0,2,0,1]=(0.877950660854)/(o+1.)+(-0.395230122376)+(0.185282186176)*0.5**o
        ref[0,2,1,0]=(0.247984803771)/(o+1.)+(-0.35022317362)+(0.238772787004)*0.5**o
        ref[0,2,1,1]=(0.545513935627)/(o+1.)+(0.499096411763)+(-0.175884373587)*0.5**o
        ref[0,2,2,0]=(-1.82248899593)/(o+1.)+(0.90641347096)+(-0.615695203188)*0.5**o
        ref[0,2,2,1]=(0.384039454422)/(o+1.)+(-0.418573857307)+(-0.0992379560876)*0.5**o
        ref[0,3,0,0]=(0.401014281541)/(o+1.)+(0.0980709839284)+(0.432415072474)*0.5**o
        ref[0,3,0,1]=(-0.842943301759)/(o+1.)+(0.314943473548)+(-0.656397468061)*0.5**o
        ref[0,3,1,0]=(0.54561755931)/(o+1.)+(0.00554262549843)+(-0.472896985897)*0.5**o
        ref[0,3,1,1]=(0.651968521189)/(o+1.)+(0.534519988855)+(-0.515568063685)*0.5**o
        ref[0,3,2,0]=(1.21411446919)/(o+1.)+(0.334629245453)+(-0.709802317012)*0.5**o
        ref[0,3,2,1]=(1.57603080404)/(o+1.)+(0.123258042212)+(-0.20057625974)*0.5**o
        ref[0,4,0,0]=(-0.318228722361)/(o+1.)+(0.248490822109)+(-0.726921583373)*0.5**o
        ref[0,4,0,1]=(-0.785890130085)/(o+1.)+(-0.445201603538)+(0.765272127757)*0.5**o
        ref[0,4,1,0]=(-0.926447863334)/(o+1.)+(0.111991361885)+(-0.613269202518)*0.5**o
        ref[0,4,1,1]=(1.51953463824)/(o+1.)+(-0.283879690312)+(-0.727485674691)*0.5**o
        ref[0,4,2,0]=(-1.08165630338)/(o+1.)+(0.520653193096)+(-0.663071984573)*0.5**o
        ref[0,4,2,1]=(1.28376782528)/(o+1.)+(0.135152679989)+(0.893396132015)*0.5**o
        ref[1,0,0,0]=(0.12071022058)/(o+1.)+(-0.35786297592)+(-0.0771632829864)*0.5**o
        ref[1,0,0,1]=(-1.72034668123)/(o+1.)+(1.18881923454)+(0.133617787097)*0.5**o
        ref[1,0,1,0]=(-1.7448609294)/(o+1.)+(-0.110381196284)+(0.174277385391)*0.5**o
        ref[1,0,1,1]=(-0.906474149485)/(o+1.)+(-0.164529387192)+(0.795443369588)*0.5**o
        ref[1,0,2,0]=(0.0119437343259)/(o+1.)+(-0.676075658316)+(0.482814141263)*0.5**o
        ref[1,0,2,1]=(0.411591741576)/(o+1.)+(-0.419934815512)+(0.619359374773)*0.5**o
        ref[1,1,0,0]=(0.508884440665)/(o+1.)+(-0.858881203437)+(-0.510375454058)*0.5**o
        ref[1,1,0,1]=(0.519297551522)/(o+1.)+(0.647377411012)+(-0.302705575836)*0.5**o
        ref[1,1,1,0]=(0.101621725999)/(o+1.)+(0.763621416728)+(0.520882033451)*0.5**o
        ref[1,1,1,1]=(0.447372581166)/(o+1.)+(-0.27141635646)+(0.624622427594)*0.5**o
        ref[1,1,2,0]=(1.54807073071)/(o+1.)+(0.865521741121)+(-0.044254701409)*0.5**o
        ref[1,1,2,1]=(-0.954794849015)/(o+1.)+(0.616890655954)+(-0.240571107795)*0.5**o
        ref[1,2,0,0]=(-0.0138637151804)/(o+1.)+(-0.23836886811)+(0.0596937021722)*0.5**o
        ref[1,2,0,1]=(0.587416875871)/(o+1.)+(0.316160501623)+(-0.935541542647)*0.5**o
        ref[1,2,1,0]=(-0.0269069617163)/(o+1.)+(1.16051668573)+(0.276741406646)*0.5**o
        ref[1,2,1,1]=(-1.65286742406)/(o+1.)+(0.626473409767)+(0.828731604537)*0.5**o
        ref[1,2,2,0]=(-0.106399305507)/(o+1.)+(0.156612816333)+(-0.380797583569)*0.5**o
        ref[1,2,2,1]=(-0.0588949726984)/(o+1.)+(-0.666712330202)+(0.257022926473)*0.5**o
        ref[1,3,0,0]=(-0.500235779047)/(o+1.)+(-0.278704142234)+(-0.645340088797)*0.5**o
        ref[1,3,0,1]=(0.100788411199)/(o+1.)+(-0.158633910442)+(0.124132693667)*0.5**o
        ref[1,3,1,0]=(0.824541289962)/(o+1.)+(0.17530886621)+(-0.7660131868)*0.5**o
        ref[1,3,1,1]=(-0.331589329951)/(o+1.)+(0.0509573421542)+(-0.150578257351)*0.5**o
        ref[1,3,2,0]=(-0.224033577887)/(o+1.)+(0.775397193106)+(-0.336661555676)*0.5**o
        ref[1,3,2,1]=(-0.0921229181433)/(o+1.)+(0.15790296452)+(0.165928745462)*0.5**o
        ref[1,4,0,0]=(-1.13149949621)/(o+1.)+(-0.0663729445879)+(0.283586653542)*0.5**o
        ref[1,4,0,1]=(-0.722886434731)/(o+1.)+(0.104888434522)+(-0.772849835334)*0.5**o
        ref[1,4,1,0]=(-1.44885259324)/(o+1.)+(-0.617668613405)+(-0.718524564964)*0.5**o
        ref[1,4,1,1]=(-0.725759098745)/(o+1.)+(-0.519223986923)+(-0.771664154234)*0.5**o
        ref[1,4,2,0]=(0.405592140408)/(o+1.)+(-0.317533868664)+(-0.591902790978)*0.5**o
        ref[1,4,2,1]=(1.64489233298)/(o+1.)+(-0.145877587744)+(0.303567440092)*0.5**o
        ref[2,0,0,0]=(1.35080478817)/(o+1.)+(-0.97319571481)+(0.197574550087)*0.5**o
        ref[2,0,0,1]=(-0.435235416366)/(o+1.)+(-0.0768910013301)+(0.0663499012532)*0.5**o
        ref[2,0,1,0]=(-0.441064969294)/(o+1.)+(-0.407537083031)+(-0.789090971877)*0.5**o
        ref[2,0,1,1]=(-0.672223056216)/(o+1.)+(-0.673709465998)+(0.222873274437)*0.5**o
        ref[2,0,2,0]=(-1.00188202525)/(o+1.)+(0.0497571110368)+(-0.300828039715)*0.5**o
        ref[2,0,2,1]=(-0.755667412756)/(o+1.)+(0.101155281333)+(-0.501507304569)*0.5**o
        ref[2,1,0,0]=(-1.03721245191)/(o+1.)+(0.421970845648)+(-0.809237327867)*0.5**o
        ref[2,1,0,1]=(-0.924064281731)/(o+1.)+(0.748054079897)+(-0.722795722506)*0.5**o
        ref[2,1,1,0]=(-0.0681656049743)/(o+1.)+(-0.401875028515)+(-0.551325611251)*0.5**o
        ref[2,1,1,1]=(-1.43246198564)/(o+1.)+(-0.346563453541)+(0.743453045588)*0.5**o
        ref[2,1,2,0]=(-0.628339309568)/(o+1.)+(0.801336921577)+(-0.033787671958)*0.5**o
        ref[2,1,2,1]=(0.375615094549)/(o+1.)+(0.362201072648)+(0.127121989623)*0.5**o
        ref[2,2,0,0]=(0.264583894127)/(o+1.)+(0.163046887519)+(-0.867530011926)*0.5**o
        ref[2,2,0,1]=(1.11591762161)/(o+1.)+(-0.953567997151)+(-0.00858729113056)*0.5**o
        ref[2,2,1,0]=(0.811633932744)/(o+1.)+(-0.227190467633)+(-0.982123891355)*0.5**o
        ref[2,2,1,1]=(-0.0564577617182)/(o+1.)+(0.240301710247)+(-0.415533190881)*0.5**o
        ref[2,2,2,0]=(0.762093170298)/(o+1.)+(0.072987783836)+(-0.559916632426)*0.5**o
        ref[2,2,2,1]=(0.202809735486)/(o+1.)+(0.537525887878)+(-0.869055045156)*0.5**o
        ref[2,3,0,0]=(-0.869552267348)/(o+1.)+(0.0306038908843)+(0.754089271004)*0.5**o
        ref[2,3,0,1]=(-0.62307943066)/(o+1.)+(0.55807045)+(0.448970570476)*0.5**o
        ref[2,3,1,0]=(0.781488342534)/(o+1.)+(-0.0325808476221)+(0.737635683404)*0.5**o
        ref[2,3,1,1]=(0.0346228388027)/(o+1.)+(-0.237622427226)+(0.551348499194)*0.5**o
        ref[2,3,2,0]=(1.11414404534)/(o+1.)+(-0.211595073582)+(-0.870721462645)*0.5**o
        ref[2,3,2,1]=(0.767751675637)/(o+1.)+(0.889675459015)+(-0.234260935528)*0.5**o
        ref[2,4,0,0]=(0.234214813823)/(o+1.)+(-0.675016631329)+(-0.474674316363)*0.5**o
        ref[2,4,0,1]=(0.588371119514)/(o+1.)+(0.511317945368)+(0.259723799797)*0.5**o
        ref[2,4,1,0]=(-1.08919560135)/(o+1.)+(0.246634130836)+(-0.960970748682)*0.5**o
        ref[2,4,1,1]=(0.606856956573)/(o+1.)+(0.459569491406)+(-0.753944077832)*0.5**o
        ref[2,4,2,0]=(-0.3621696986)/(o+1.)+(-0.599887391364)+(-0.233304828171)*0.5**o
        ref[2,4,2,1]=(-1.52432168019)/(o+1.)+(-0.241785281086)+(0.341373328517)*0.5**o
        ref[3,0,0,0]=(-0.35485184543)/(o+1.)+(-0.360150533825)+(-0.472652289489)*0.5**o
        ref[3,0,0,1]=(0.70777600619)/(o+1.)+(-0.623017999413)+(-0.398674023611)*0.5**o
        ref[3,0,1,0]=(-0.924138547773)/(o+1.)+(1.10788910433)+(-0.0216721345971)*0.5**o
        ref[3,0,1,1]=(0.577735621075)/(o+1.)+(-0.426576227965)+(0.0606265274442)*0.5**o
        ref[3,0,2,0]=(-0.57434311249)/(o+1.)+(0.119341906638)+(0.630444608821)*0.5**o
        ref[3,0,2,1]=(0.630840117156)/(o+1.)+(-0.207637601452)+(-0.348127940577)*0.5**o
        ref[3,1,0,0]=(-0.622765428492)/(o+1.)+(-0.32512846527)+(0.735069593965)*0.5**o
        ref[3,1,0,1]=(-1.0010192039)/(o+1.)+(-0.076924486041)+(0.00281719106153)*0.5**o
        ref[3,1,1,0]=(-0.217846920082)/(o+1.)+(0.0663664342643)+(-0.341732637339)*0.5**o
        ref[3,1,1,1]=(0.25262075106)/(o+1.)+(-0.870818911505)+(0.644746847185)*0.5**o
        ref[3,1,2,0]=(-0.558981437567)/(o+1.)+(-1.25951392994)+(0.382425223801)*0.5**o
        ref[3,1,2,1]=(0.454231281141)/(o+1.)+(-0.034746803327)+(0.630558573276)*0.5**o
        ref[3,2,0,0]=(0.129327208099)/(o+1.)+(-0.277756527686)+(-0.404884771697)*0.5**o
        ref[3,2,0,1]=(0.0959971245528)/(o+1.)+(0.882025105044)+(0.28819895431)*0.5**o
        ref[3,2,1,0]=(0.764427503331)/(o+1.)+(-0.621589501714)+(0.949790710961)*0.5**o
        ref[3,2,1,1]=(1.44642085598)/(o+1.)+(0.00586074911794)+(0.619659549755)*0.5**o
        ref[3,2,2,0]=(0.704511350915)/(o+1.)+(-0.271502528652)+(-0.457269257294)*0.5**o
        ref[3,2,2,1]=(0.744528695119)/(o+1.)+(0.50220477652)+(0.933553130797)*0.5**o
        ref[3,3,0,0]=(-0.428306899008)/(o+1.)+(-0.0668006489082)+(0.825594594128)*0.5**o
        ref[3,3,0,1]=(-0.505508222749)/(o+1.)+(-0.963387836721)+(0.514394313629)*0.5**o
        ref[3,3,1,0]=(1.21822251261)/(o+1.)+(-0.0977268936434)+(0.821240511475)*0.5**o
        ref[3,3,1,1]=(1.4647083324)/(o+1.)+(0.149613127157)+(-0.644256768839)*0.5**o
        ref[3,3,2,0]=(0.573437454483)/(o+1.)+(-0.0806908804035)+(0.534862520097)*0.5**o
        ref[3,3,2,1]=(0.069887177217)/(o+1.)+(-0.555135470617)+(0.722071427786)*0.5**o
        ref[3,4,0,0]=(0.152448333383)/(o+1.)+(0.903761766608)+(0.188625467698)*0.5**o
        ref[3,4,0,1]=(0.415480168613)/(o+1.)+(-0.220519654501)+(0.181430484096)*0.5**o
        ref[3,4,1,0]=(-0.592961594036)/(o+1.)+(-0.350772955066)+(0.213491785855)*0.5**o
        ref[3,4,1,1]=(-0.532890980061)/(o+1.)+(-0.251121395257)+(-0.401744411188)*0.5**o
        ref[3,4,2,0]=(-0.194325734076)/(o+1.)+(0.581224356966)+(-0.0813611687514)*0.5**o
        ref[3,4,2,1]=(1.37635341581)/(o+1.)+(1.02866696565)+(-0.0991425040958)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0333010052777)*x[0]**o+(-0.242011298555)*x[0]+(0.686457578633)*x[1]**o+(0.438506741453)*x[1]
        ref=(0.686457578633)/(o+1.)+(0.0982477214491)+(-0.0333010052777)*0.5**o
      else:
        arg=(-0.365813216932)*x[0]**o+(0.969816151161)*x[0]+(0.349995764077)*x[1]**o+(0.867492181692)*x[1]+(-0.866177355796)*x[2]**o+(-0.203219031102)*x[2]
        ref=(-0.516181591719)/(o+1.)+(0.817044650875)+(-0.365813216932)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.038446371199)*x[0]**o+(0.323075821067)*x[0]+(0.770528210463)*x[1]**o+(0.0434528043023)*x[1]
        arg[1]=(-0.461876402566)*x[0]**o+(0.746965376214)*x[0]+(0.861234840159)*x[1]**o+(-0.568165009958)*x[1]
        ref[0]=(0.770528210463)/(o+1.)+(0.183264312685)+(0.038446371199)*0.5**o
        ref[1]=(0.861234840159)/(o+1.)+(0.0894001831281)+(-0.461876402566)*0.5**o
      else:
        arg[0]=(-0.472858663642)*x[0]**o+(-0.680210745369)*x[0]+(0.175425838082)*x[1]**o+(0.0189427039667)*x[1]+(0.0365198456146)*x[2]**o+(0.7840527457)*x[2]
        arg[1]=(-0.0191775972693)*x[0]**o+(0.19303923194)*x[0]+(0.775528295333)*x[1]**o+(-0.765500708496)*x[1]+(0.351369471606)*x[2]**o+(-0.351587727906)*x[2]
        ref[0]=(0.211945683696)/(o+1.)+(0.0613923521485)+(-0.472858663642)*0.5**o
        ref[1]=(1.12689776694)/(o+1.)+(-0.462024602231)+(-0.0191775972693)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.994443255559)*x[0]**o+(-0.403427184912)*x[0]+(-0.889656446862)*x[1]**o+(-0.411934106034)*x[1]
        arg[0,1]=(-0.864086447061)*x[0]**o+(-0.277489357308)*x[0]+(0.977910864864)*x[1]**o+(-0.0505775488074)*x[1]
        arg[0,2]=(-0.397179114534)*x[0]**o+(-0.448046821439)*x[0]+(-0.340253609736)*x[1]**o+(0.624988837716)*x[1]
        arg[0,3]=(-0.9155191116)*x[0]**o+(0.133777450278)*x[0]+(-0.809567026065)*x[1]**o+(-0.815374059364)*x[1]
        arg[0,4]=(-0.416764684181)*x[0]**o+(0.0965164570638)*x[0]+(0.190191799833)*x[1]**o+(0.0970709424519)*x[1]
        arg[1,0]=(0.646527294535)*x[0]**o+(-0.337119506636)*x[0]+(0.42669563245)*x[1]**o+(-0.185228568706)*x[1]
        arg[1,1]=(-0.411532647934)*x[0]**o+(0.178830973832)*x[0]+(-0.904886020892)*x[1]**o+(-0.740623071275)*x[1]
        arg[1,2]=(0.523024622808)*x[0]**o+(-0.226298848077)*x[0]+(-0.707740568617)*x[1]**o+(0.998182796571)*x[1]
        arg[1,3]=(0.840649917357)*x[0]**o+(-0.736279151571)*x[0]+(0.996017554392)*x[1]**o+(-0.397833319931)*x[1]
        arg[1,4]=(0.76711764419)*x[0]**o+(-0.905452991298)*x[0]+(-0.925675237951)*x[1]**o+(-0.982712645237)*x[1]
        arg[2,0]=(0.198547103682)*x[0]**o+(0.674537682234)*x[0]+(-0.891195618193)*x[1]**o+(0.979109691481)*x[1]
        arg[2,1]=(0.171629921737)*x[0]**o+(0.0102727001751)*x[0]+(0.467578393374)*x[1]**o+(0.583960984645)*x[1]
        arg[2,2]=(-0.637526271497)*x[0]**o+(-0.900821073531)*x[0]+(0.61637733965)*x[1]**o+(0.43717505341)*x[1]
        arg[2,3]=(-0.205861963967)*x[0]**o+(-0.104621062003)*x[0]+(0.582584591946)*x[1]**o+(0.351739662419)*x[1]
        arg[2,4]=(-0.179355911316)*x[0]**o+(-0.188415282678)*x[0]+(0.598342895007)*x[1]**o+(0.364638908577)*x[1]
        arg[3,0]=(0.150471526119)*x[0]**o+(-0.940789873202)*x[0]+(-0.524275825864)*x[1]**o+(0.729004683023)*x[1]
        arg[3,1]=(0.123933417087)*x[0]**o+(0.192413160498)*x[0]+(0.203730760507)*x[1]**o+(0.852010172966)*x[1]
        arg[3,2]=(-0.297766950847)*x[0]**o+(0.269905131596)*x[0]+(-0.0830931256847)*x[1]**o+(0.26964712569)*x[1]
        arg[3,3]=(0.96961517265)*x[0]**o+(0.363312967302)*x[0]+(-0.811142762589)*x[1]**o+(-0.420442983537)*x[1]
        arg[3,4]=(-0.314185602206)*x[0]**o+(-0.0446515646101)*x[0]+(-0.884651038732)*x[1]**o+(-0.166386477611)*x[1]
        ref[0,0]=(-0.889656446862)/(o+1.)+(-0.407680645473)+(-0.994443255559)*0.5**o
        ref[0,1]=(0.977910864864)/(o+1.)+(-0.164033453058)+(-0.864086447061)*0.5**o
        ref[0,2]=(-0.340253609736)/(o+1.)+(0.0884710081389)+(-0.397179114534)*0.5**o
        ref[0,3]=(-0.809567026065)/(o+1.)+(-0.340798304543)+(-0.9155191116)*0.5**o
        ref[0,4]=(0.190191799833)/(o+1.)+(0.0967936997578)+(-0.416764684181)*0.5**o
        ref[1,0]=(0.42669563245)/(o+1.)+(-0.261174037671)+(0.646527294535)*0.5**o
        ref[1,1]=(-0.904886020892)/(o+1.)+(-0.280896048722)+(-0.411532647934)*0.5**o
        ref[1,2]=(-0.707740568617)/(o+1.)+(0.385941974247)+(0.523024622808)*0.5**o
        ref[1,3]=(0.996017554392)/(o+1.)+(-0.567056235751)+(0.840649917357)*0.5**o
        ref[1,4]=(-0.925675237951)/(o+1.)+(-0.944082818268)+(0.76711764419)*0.5**o
        ref[2,0]=(-0.891195618193)/(o+1.)+(0.826823686857)+(0.198547103682)*0.5**o
        ref[2,1]=(0.467578393374)/(o+1.)+(0.29711684241)+(0.171629921737)*0.5**o
        ref[2,2]=(0.61637733965)/(o+1.)+(-0.231823010061)+(-0.637526271497)*0.5**o
        ref[2,3]=(0.582584591946)/(o+1.)+(0.123559300208)+(-0.205861963967)*0.5**o
        ref[2,4]=(0.598342895007)/(o+1.)+(0.0881118129495)+(-0.179355911316)*0.5**o
        ref[3,0]=(-0.524275825864)/(o+1.)+(-0.105892595089)+(0.150471526119)*0.5**o
        ref[3,1]=(0.203730760507)/(o+1.)+(0.522211666732)+(0.123933417087)*0.5**o
        ref[3,2]=(-0.0830931256847)/(o+1.)+(0.269776128643)+(-0.297766950847)*0.5**o
        ref[3,3]=(-0.811142762589)/(o+1.)+(-0.0285650081177)+(0.96961517265)*0.5**o
        ref[3,4]=(-0.884651038732)/(o+1.)+(-0.105519021111)+(-0.314185602206)*0.5**o
      else:
        arg[0,0]=(0.264691789613)*x[0]**o+(0.816541340997)*x[0]+(-0.402990464667)*x[1]**o+(-0.57479619687)*x[1]+(0.336313666995)*x[2]**o+(-0.353125285536)*x[2]
        arg[0,1]=(-0.0526559162494)*x[0]**o+(-0.901336706583)*x[0]+(0.237482536727)*x[1]**o+(-0.458564275773)*x[1]+(-0.90715992057)*x[2]**o+(-0.352062856077)*x[2]
        arg[0,2]=(0.786349112412)*x[0]**o+(-0.395569856472)*x[0]+(-0.996593023157)*x[1]**o+(-0.277001350627)*x[1]+(0.435375913289)*x[2]**o+(0.00380742297736)*x[2]
        arg[0,3]=(0.3336751197)*x[0]**o+(0.592185657129)*x[0]+(0.148956266814)*x[1]**o+(-0.852264651722)*x[1]+(-0.680249245511)*x[2]**o+(0.238843544413)*x[2]
        arg[0,4]=(0.875206676786)*x[0]**o+(0.0420423017124)*x[0]+(-0.690488811412)*x[1]**o+(-0.191328696331)*x[1]+(-0.861423624177)*x[2]**o+(0.483925405411)*x[2]
        arg[1,0]=(-0.573216715158)*x[0]**o+(-0.0668114726198)*x[0]+(0.382097918101)*x[1]**o+(-0.126485626438)*x[1]+(0.37521336616)*x[2]**o+(0.42402066058)*x[2]
        arg[1,1]=(-0.0919326514467)*x[0]**o+(-0.791351880042)*x[0]+(0.772694542694)*x[1]**o+(-0.537902003787)*x[1]+(0.424413342503)*x[2]**o+(-0.285553230731)*x[2]
        arg[1,2]=(0.855490362389)*x[0]**o+(0.60428104628)*x[0]+(-0.846565232506)*x[1]**o+(-0.243727717758)*x[1]+(0.900077399935)*x[2]**o+(-0.974305489047)*x[2]
        arg[1,3]=(-0.90802343526)*x[0]**o+(0.00950587842259)*x[0]+(0.821234711321)*x[1]**o+(-0.0628912085092)*x[1]+(-0.773563396868)*x[2]**o+(-0.403955952419)*x[2]
        arg[1,4]=(-0.33139358787)*x[0]**o+(-0.901382943687)*x[0]+(-0.981921375707)*x[1]**o+(-0.0654519807133)*x[1]+(0.880069861158)*x[2]**o+(-0.0796172270076)*x[2]
        arg[2,0]=(-0.202818692518)*x[0]**o+(-0.155436611971)*x[0]+(0.739467305515)*x[1]**o+(-0.12096877475)*x[1]+(0.479644796804)*x[2]**o+(0.0580715322389)*x[2]
        arg[2,1]=(0.0461326006936)*x[0]**o+(0.122168224674)*x[0]+(0.986788073931)*x[1]**o+(-0.435933938764)*x[1]+(-0.312527130215)*x[2]**o+(0.726805187165)*x[2]
        arg[2,2]=(-0.539659153805)*x[0]**o+(-0.26082353681)*x[0]+(0.367739719819)*x[1]**o+(0.943584828354)*x[1]+(0.688225221556)*x[2]**o+(-0.121563525696)*x[2]
        arg[2,3]=(0.861717128203)*x[0]**o+(0.0666983176426)*x[0]+(0.984510848927)*x[1]**o+(-0.321865378029)*x[1]+(0.522584734)*x[2]**o+(0.589527116625)*x[2]
        arg[2,4]=(0.778575699703)*x[0]**o+(0.515802361664)*x[0]+(0.817655905624)*x[1]**o+(0.883824717911)*x[1]+(0.656833761714)*x[2]**o+(-0.991585145611)*x[2]
        arg[3,0]=(0.752875513075)*x[0]**o+(0.615136567082)*x[0]+(-0.205265870704)*x[1]**o+(0.483918142728)*x[1]+(-0.0495322147557)*x[2]**o+(-0.132474571954)*x[2]
        arg[3,1]=(0.0563875061408)*x[0]**o+(-0.603033440119)*x[0]+(-0.0944809376206)*x[1]**o+(0.725330686731)*x[1]+(0.00527671533343)*x[2]**o+(-0.434519675586)*x[2]
        arg[3,2]=(0.100078550382)*x[0]**o+(-0.340988639989)*x[0]+(-0.392591649097)*x[1]**o+(-0.924705734588)*x[1]+(0.420410192075)*x[2]**o+(-0.94741221593)*x[2]
        arg[3,3]=(0.536414145801)*x[0]**o+(0.142421640429)*x[0]+(0.213162255423)*x[1]**o+(-0.695193008147)*x[1]+(0.9625091599)*x[2]**o+(0.534409542153)*x[2]
        arg[3,4]=(0.305683606235)*x[0]**o+(0.340964210305)*x[0]+(0.757807973246)*x[1]**o+(0.11459551493)*x[1]+(-0.81395047506)*x[2]**o+(0.636392281491)*x[2]
        ref[0,0]=(-0.0666767976714)/(o+1.)+(-0.0556900707042)+(0.264691789613)*0.5**o
        ref[0,1]=(-0.669677383843)/(o+1.)+(-0.855981919217)+(-0.0526559162494)*0.5**o
        ref[0,2]=(-0.561217109869)/(o+1.)+(-0.334381892061)+(0.786349112412)*0.5**o
        ref[0,3]=(-0.531292978697)/(o+1.)+(-0.0106177250896)+(0.3336751197)*0.5**o
        ref[0,4]=(-1.55191243559)/(o+1.)+(0.167319505396)+(0.875206676786)*0.5**o
        ref[1,0]=(0.757311284261)/(o+1.)+(0.115361780761)+(-0.573216715158)*0.5**o
        ref[1,1]=(1.1971078852)/(o+1.)+(-0.80740355728)+(-0.0919326514467)*0.5**o
        ref[1,2]=(0.0535121674292)/(o+1.)+(-0.306876080263)+(0.855490362389)*0.5**o
        ref[1,3]=(0.0476713144526)/(o+1.)+(-0.228670641253)+(-0.90802343526)*0.5**o
        ref[1,4]=(-0.101851514548)/(o+1.)+(-0.523226075704)+(-0.33139358787)*0.5**o
        ref[2,0]=(1.21911210232)/(o+1.)+(-0.109166927241)+(-0.202818692518)*0.5**o
        ref[2,1]=(0.674260943716)/(o+1.)+(0.206519736538)+(0.0461326006936)*0.5**o
        ref[2,2]=(1.05596494137)/(o+1.)+(0.280598882924)+(-0.539659153805)*0.5**o
        ref[2,3]=(1.50709558293)/(o+1.)+(0.167180028119)+(0.861717128203)*0.5**o
        ref[2,4]=(1.47448966734)/(o+1.)+(0.204020966982)+(0.778575699703)*0.5**o
        ref[3,0]=(-0.25479808546)/(o+1.)+(0.483290068928)+(0.752875513075)*0.5**o
        ref[3,1]=(-0.0892042222872)/(o+1.)+(-0.156111214487)+(0.0563875061408)*0.5**o
        ref[3,2]=(0.0278185429781)/(o+1.)+(-1.10655329525)+(0.100078550382)*0.5**o
        ref[3,3]=(1.17567141532)/(o+1.)+(-0.00918091278228)+(0.536414145801)*0.5**o
        ref[3,4]=(-0.0561425018138)/(o+1.)+(0.545976003363)+(0.305683606235)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.448877659671)*x[0]**o+(0.130599173692)*x[0]+(-0.264185555629)*x[1]**o+(-0.722726040492)*x[1]
        arg[0,0,1]=(-0.543806308372)*x[0]**o+(-0.685242399166)*x[0]+(0.325336801501)*x[1]**o+(0.572960380228)*x[1]
        arg[0,1,0]=(-0.893080563297)*x[0]**o+(-0.251553981066)*x[0]+(0.256078388483)*x[1]**o+(-0.926701685259)*x[1]
        arg[0,1,1]=(0.346088914447)*x[0]**o+(0.166180418696)*x[0]+(0.353024178635)*x[1]**o+(0.731811935721)*x[1]
        arg[1,0,0]=(0.553083243483)*x[0]**o+(0.976153803522)*x[0]+(-0.260777317634)*x[1]**o+(-0.641936915628)*x[1]
        arg[1,0,1]=(-0.561076901068)*x[0]**o+(0.837832618302)*x[0]+(-0.638970151828)*x[1]**o+(-0.495169195543)*x[1]
        arg[1,1,0]=(-0.292805925132)*x[0]**o+(0.670472035586)*x[0]+(-0.752026521307)*x[1]**o+(-0.473087472861)*x[1]
        arg[1,1,1]=(-0.432857805275)*x[0]**o+(0.495387797249)*x[0]+(0.0579881469257)*x[1]**o+(0.831644440274)*x[1]
        arg[2,0,0]=(0.413433231384)*x[0]**o+(0.919831571878)*x[0]+(-0.373581626416)*x[1]**o+(0.16423318333)*x[1]
        arg[2,0,1]=(0.857566619293)*x[0]**o+(0.340454525569)*x[0]+(0.139298007219)*x[1]**o+(0.220542978265)*x[1]
        arg[2,1,0]=(-0.914184896115)*x[0]**o+(-0.0279186159835)*x[0]+(0.893417478351)*x[1]**o+(-0.108178714604)*x[1]
        arg[2,1,1]=(-0.749629612712)*x[0]**o+(-0.649957092299)*x[0]+(0.181944194719)*x[1]**o+(-0.988782927329)*x[1]
        arg[3,0,0]=(0.292013067678)*x[0]**o+(0.491680364317)*x[0]+(0.951144487798)*x[1]**o+(0.980949630108)*x[1]
        arg[3,0,1]=(0.89779245885)*x[0]**o+(-0.147546711846)*x[0]+(0.0902450198362)*x[1]**o+(-0.669984062969)*x[1]
        arg[3,1,0]=(0.62590631314)*x[0]**o+(0.329384454978)*x[0]+(-0.676917185409)*x[1]**o+(-0.0903173226074)*x[1]
        arg[3,1,1]=(-0.668268661271)*x[0]**o+(-0.847167879584)*x[0]+(0.329431843319)*x[1]**o+(-0.989253343115)*x[1]
        arg[4,0,0]=(-0.675111374291)*x[0]**o+(0.738030423939)*x[0]+(-0.413299452027)*x[1]**o+(0.36859580301)*x[1]
        arg[4,0,1]=(-0.207625666422)*x[0]**o+(-0.455639256266)*x[0]+(0.554118743878)*x[1]**o+(0.14044849621)*x[1]
        arg[4,1,0]=(-0.313337039275)*x[0]**o+(-0.451797308521)*x[0]+(0.201296275104)*x[1]**o+(-0.779869953962)*x[1]
        arg[4,1,1]=(-0.0308697706768)*x[0]**o+(0.214664850788)*x[0]+(-0.824157101269)*x[1]**o+(0.156012836486)*x[1]
        arg[5,0,0]=(0.277885566309)*x[0]**o+(-0.387820938776)*x[0]+(0.820602896312)*x[1]**o+(0.162715513355)*x[1]
        arg[5,0,1]=(-0.216951421349)*x[0]**o+(0.906019329549)*x[0]+(-0.330132336808)*x[1]**o+(0.416109194357)*x[1]
        arg[5,1,0]=(0.923496627826)*x[0]**o+(-0.177644464371)*x[0]+(0.482278717995)*x[1]**o+(-0.362003106154)*x[1]
        arg[5,1,1]=(0.9109111591)*x[0]**o+(-0.0919782899769)*x[0]+(0.103950387446)*x[1]**o+(0.447452648868)*x[1]
        ref[0,0,0]=(-0.264185555629)/(o+1.)+(-0.2960634334)+(0.448877659671)*0.5**o
        ref[0,0,1]=(0.325336801501)/(o+1.)+(-0.0561410094689)+(-0.543806308372)*0.5**o
        ref[0,1,0]=(0.256078388483)/(o+1.)+(-0.589127833162)+(-0.893080563297)*0.5**o
        ref[0,1,1]=(0.353024178635)/(o+1.)+(0.448996177208)+(0.346088914447)*0.5**o
        ref[1,0,0]=(-0.260777317634)/(o+1.)+(0.167108443947)+(0.553083243483)*0.5**o
        ref[1,0,1]=(-0.638970151828)/(o+1.)+(0.17133171138)+(-0.561076901068)*0.5**o
        ref[1,1,0]=(-0.752026521307)/(o+1.)+(0.0986922813626)+(-0.292805925132)*0.5**o
        ref[1,1,1]=(0.0579881469257)/(o+1.)+(0.663516118762)+(-0.432857805275)*0.5**o
        ref[2,0,0]=(-0.373581626416)/(o+1.)+(0.542032377604)+(0.413433231384)*0.5**o
        ref[2,0,1]=(0.139298007219)/(o+1.)+(0.280498751917)+(0.857566619293)*0.5**o
        ref[2,1,0]=(0.893417478351)/(o+1.)+(-0.0680486652939)+(-0.914184896115)*0.5**o
        ref[2,1,1]=(0.181944194719)/(o+1.)+(-0.819370009814)+(-0.749629612712)*0.5**o
        ref[3,0,0]=(0.951144487798)/(o+1.)+(0.736314997212)+(0.292013067678)*0.5**o
        ref[3,0,1]=(0.0902450198362)/(o+1.)+(-0.408765387408)+(0.89779245885)*0.5**o
        ref[3,1,0]=(-0.676917185409)/(o+1.)+(0.119533566185)+(0.62590631314)*0.5**o
        ref[3,1,1]=(0.329431843319)/(o+1.)+(-0.91821061135)+(-0.668268661271)*0.5**o
        ref[4,0,0]=(-0.413299452027)/(o+1.)+(0.553313113475)+(-0.675111374291)*0.5**o
        ref[4,0,1]=(0.554118743878)/(o+1.)+(-0.157595380028)+(-0.207625666422)*0.5**o
        ref[4,1,0]=(0.201296275104)/(o+1.)+(-0.615833631241)+(-0.313337039275)*0.5**o
        ref[4,1,1]=(-0.824157101269)/(o+1.)+(0.185338843637)+(-0.0308697706768)*0.5**o
        ref[5,0,0]=(0.820602896312)/(o+1.)+(-0.11255271271)+(0.277885566309)*0.5**o
        ref[5,0,1]=(-0.330132336808)/(o+1.)+(0.661064261953)+(-0.216951421349)*0.5**o
        ref[5,1,0]=(0.482278717995)/(o+1.)+(-0.269823785262)+(0.923496627826)*0.5**o
        ref[5,1,1]=(0.103950387446)/(o+1.)+(0.177737179446)+(0.9109111591)*0.5**o
      else:
        arg[0,0,0]=(0.843169691966)*x[0]**o+(0.252560741792)*x[0]+(0.0428740803857)*x[1]**o+(-0.702885246223)*x[1]+(0.175406156382)*x[2]**o+(-0.819206174419)*x[2]
        arg[0,0,1]=(-0.780678250187)*x[0]**o+(-0.628781977796)*x[0]+(0.345241549754)*x[1]**o+(-0.686783366275)*x[1]+(0.298501249797)*x[2]**o+(-0.592222734778)*x[2]
        arg[0,1,0]=(0.680331984783)*x[0]**o+(0.20945731105)*x[0]+(0.582220641498)*x[1]**o+(0.159144772508)*x[1]+(0.197544685245)*x[2]**o+(-0.542847499692)*x[2]
        arg[0,1,1]=(0.561649107089)*x[0]**o+(0.409538404439)*x[0]+(0.687461001448)*x[1]**o+(0.327986995905)*x[1]+(0.701007758697)*x[2]**o+(0.95919995454)*x[2]
        arg[1,0,0]=(0.0620287124894)*x[0]**o+(-0.946536581396)*x[0]+(-0.813637608115)*x[1]**o+(0.0808967031579)*x[1]+(0.545408819059)*x[2]**o+(0.250558134093)*x[2]
        arg[1,0,1]=(-0.0778300410758)*x[0]**o+(-0.852201314682)*x[0]+(0.712528858815)*x[1]**o+(-0.69009644973)*x[1]+(-0.0836140529567)*x[2]**o+(0.687579338803)*x[2]
        arg[1,1,0]=(-0.37648103863)*x[0]**o+(-0.264461846666)*x[0]+(-0.45921598716)*x[1]**o+(-0.516551658367)*x[1]+(0.512446750786)*x[2]**o+(0.850266522322)*x[2]
        arg[1,1,1]=(-0.951741356481)*x[0]**o+(-0.989023540683)*x[0]+(0.338787640816)*x[1]**o+(0.764043925298)*x[1]+(0.524009019713)*x[2]**o+(-0.912507794534)*x[2]
        arg[2,0,0]=(0.613468181683)*x[0]**o+(-0.357325179573)*x[0]+(0.153221256731)*x[1]**o+(-0.748686771579)*x[1]+(0.364085012368)*x[2]**o+(-0.134535056195)*x[2]
        arg[2,0,1]=(-0.728948376511)*x[0]**o+(-0.0657162843876)*x[0]+(-0.336013775703)*x[1]**o+(0.299044747397)*x[1]+(-0.712877779871)*x[2]**o+(0.229974926734)*x[2]
        arg[2,1,0]=(-0.213428159952)*x[0]**o+(0.680250111549)*x[0]+(-0.986942572737)*x[1]**o+(0.549006049647)*x[1]+(0.572921637402)*x[2]**o+(-0.241851127933)*x[2]
        arg[2,1,1]=(0.0367155018841)*x[0]**o+(-0.768111161527)*x[0]+(-0.748085216726)*x[1]**o+(-0.752536272228)*x[1]+(0.555945063877)*x[2]**o+(0.012687074005)*x[2]
        arg[3,0,0]=(-0.945391902075)*x[0]**o+(-0.581330133808)*x[0]+(-0.253390460088)*x[1]**o+(-0.885589795648)*x[1]+(-0.555900062171)*x[2]**o+(-0.241006607054)*x[2]
        arg[3,0,1]=(0.924110231188)*x[0]**o+(0.428520753545)*x[0]+(0.187773835144)*x[1]**o+(-0.53787305121)*x[1]+(0.0970284978901)*x[2]**o+(0.0580212350951)*x[2]
        arg[3,1,0]=(0.426389283701)*x[0]**o+(-0.0976148877161)*x[0]+(-0.364828466078)*x[1]**o+(-0.536398529197)*x[1]+(0.648605043498)*x[2]**o+(-0.234629706559)*x[2]
        arg[3,1,1]=(-0.736098543054)*x[0]**o+(0.985014182905)*x[0]+(-0.563513361433)*x[1]**o+(0.589434310221)*x[1]+(-0.360296462891)*x[2]**o+(-0.586591636182)*x[2]
        arg[4,0,0]=(0.953916732391)*x[0]**o+(-0.835620914249)*x[0]+(0.562803727423)*x[1]**o+(-0.904263689551)*x[1]+(-0.349815820142)*x[2]**o+(-0.156410988081)*x[2]
        arg[4,0,1]=(-0.236710267891)*x[0]**o+(0.958992060128)*x[0]+(0.711565908363)*x[1]**o+(-0.639771085045)*x[1]+(-0.919042661652)*x[2]**o+(-0.198907436333)*x[2]
        arg[4,1,0]=(0.275058935551)*x[0]**o+(0.563461500878)*x[0]+(0.438833998471)*x[1]**o+(0.686375712335)*x[1]+(-0.930322605963)*x[2]**o+(0.710777122765)*x[2]
        arg[4,1,1]=(-0.507845102241)*x[0]**o+(-0.390574086909)*x[0]+(-0.170020669232)*x[1]**o+(-0.107455242464)*x[1]+(-0.97056652533)*x[2]**o+(-0.720180192255)*x[2]
        arg[5,0,0]=(-0.922597172214)*x[0]**o+(0.229197305733)*x[0]+(-0.943214499543)*x[1]**o+(0.269256599046)*x[1]+(-0.11540537404)*x[2]**o+(0.956940487934)*x[2]
        arg[5,0,1]=(-0.0350868975578)*x[0]**o+(-0.547620412319)*x[0]+(-0.818344866429)*x[1]**o+(-0.0632082353078)*x[1]+(-0.490536679501)*x[2]**o+(-0.79641747407)*x[2]
        arg[5,1,0]=(0.877279477171)*x[0]**o+(-0.408752853807)*x[0]+(0.143432191997)*x[1]**o+(0.230679627667)*x[1]+(0.0346230515425)*x[2]**o+(0.357377660033)*x[2]
        arg[5,1,1]=(0.294375780312)*x[0]**o+(-0.688447065878)*x[0]+(0.213334642918)*x[1]**o+(0.664021137697)*x[1]+(-0.532996065961)*x[2]**o+(0.794315136407)*x[2]
        ref[0,0,0]=(0.218280236768)/(o+1.)+(-0.634765339425)+(0.843169691966)*0.5**o
        ref[0,0,1]=(0.64374279955)/(o+1.)+(-0.953894039424)+(-0.780678250187)*0.5**o
        ref[0,1,0]=(0.779765326742)/(o+1.)+(-0.0871227080671)+(0.680331984783)*0.5**o
        ref[0,1,1]=(1.38846876015)/(o+1.)+(0.848362677442)+(0.561649107089)*0.5**o
        ref[1,0,0]=(-0.268228789055)/(o+1.)+(-0.307540872072)+(0.0620287124894)*0.5**o
        ref[1,0,1]=(0.628914805859)/(o+1.)+(-0.427359212804)+(-0.0778300410758)*0.5**o
        ref[1,1,0]=(0.0532307636262)/(o+1.)+(0.0346265086445)+(-0.37648103863)*0.5**o
        ref[1,1,1]=(0.862796660529)/(o+1.)+(-0.56874370496)+(-0.951741356481)*0.5**o
        ref[2,0,0]=(0.5173062691)/(o+1.)+(-0.620273503673)+(0.613468181683)*0.5**o
        ref[2,0,1]=(-1.04889155557)/(o+1.)+(0.231651694871)+(-0.728948376511)*0.5**o
        ref[2,1,0]=(-0.414020935335)/(o+1.)+(0.493702516631)+(-0.213428159952)*0.5**o
        ref[2,1,1]=(-0.192140152849)/(o+1.)+(-0.753980179875)+(0.0367155018841)*0.5**o
        ref[3,0,0]=(-0.809290522259)/(o+1.)+(-0.853963268255)+(-0.945391902075)*0.5**o
        ref[3,0,1]=(0.284802333034)/(o+1.)+(-0.0256655312851)+(0.924110231188)*0.5**o
        ref[3,1,0]=(0.28377657742)/(o+1.)+(-0.434321561736)+(0.426389283701)*0.5**o
        ref[3,1,1]=(-0.923809824324)/(o+1.)+(0.493928428472)+(-0.736098543054)*0.5**o
        ref[4,0,0]=(0.212987907281)/(o+1.)+(-0.948147795941)+(0.953916732391)*0.5**o
        ref[4,0,1]=(-0.207476753289)/(o+1.)+(0.0601567693744)+(-0.236710267891)*0.5**o
        ref[4,1,0]=(-0.491488607492)/(o+1.)+(0.980307167989)+(0.275058935551)*0.5**o
        ref[4,1,1]=(-1.14058719456)/(o+1.)+(-0.609104760814)+(-0.507845102241)*0.5**o
        ref[5,0,0]=(-1.05861987358)/(o+1.)+(0.727697196357)+(-0.922597172214)*0.5**o
        ref[5,0,1]=(-1.30888154593)/(o+1.)+(-0.703623060848)+(-0.0350868975578)*0.5**o
        ref[5,1,0]=(0.17805524354)/(o+1.)+(0.0896522169463)+(0.877279477171)*0.5**o
        ref[5,1,1]=(-0.319661423044)/(o+1.)+(0.384944604113)+(0.294375780312)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactOne

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.182956096072)*x[0]**o+(0.857065806542)*x[0]+(-0.0352052809673)*x[1]**o+(0.950661773214)*x[1]
        arg[0,0,0,1]=(-0.81342189211)*x[0]**o+(0.36137453013)*x[0]+(0.0610373691872)*x[1]**o+(-0.974859653337)*x[1]
        arg[0,0,1,0]=(-0.149103531408)*x[0]**o+(0.853517194849)*x[0]+(0.623943596759)*x[1]**o+(0.441477983782)*x[1]
        arg[0,0,1,1]=(0.74912087771)*x[0]**o+(0.935896347068)*x[0]+(0.136600247183)*x[1]**o+(0.915043476739)*x[1]
        arg[0,0,2,0]=(-0.289042694587)*x[0]**o+(-0.0324580134489)*x[0]+(0.966454801911)*x[1]**o+(0.158747002668)*x[1]
        arg[0,0,2,1]=(0.0809737838902)*x[0]**o+(0.483125829124)*x[0]+(-0.156074498516)*x[1]**o+(-0.387215102608)*x[1]
        arg[0,1,0,0]=(0.638822193145)*x[0]**o+(-0.211942445721)*x[0]+(-0.016865109839)*x[1]**o+(-0.596486509057)*x[1]
        arg[0,1,0,1]=(0.611604473356)*x[0]**o+(-0.330484367791)*x[0]+(-0.897448372147)*x[1]**o+(-0.0188556077191)*x[1]
        arg[0,1,1,0]=(-0.745407006471)*x[0]**o+(0.547985829362)*x[0]+(-0.229098133178)*x[1]**o+(0.339357994536)*x[1]
        arg[0,1,1,1]=(0.756528549281)*x[0]**o+(0.0968167016962)*x[0]+(0.481206893907)*x[1]**o+(-0.320299781291)*x[1]
        arg[0,1,2,0]=(0.335544368565)*x[0]**o+(0.143490409353)*x[0]+(0.367303506039)*x[1]**o+(-0.459001464112)*x[1]
        arg[0,1,2,1]=(0.669933676654)*x[0]**o+(0.0679909847194)*x[0]+(0.182483888026)*x[1]**o+(-0.382138842505)*x[1]
        arg[0,2,0,0]=(-0.929163949925)*x[0]**o+(0.670533855076)*x[0]+(0.587675819052)*x[1]**o+(0.221939557678)*x[1]
        arg[0,2,0,1]=(0.639896773528)*x[0]**o+(-0.166901547286)*x[0]+(-0.83934259298)*x[1]**o+(-0.522536841927)*x[1]
        arg[0,2,1,0]=(-0.399556357179)*x[0]**o+(0.860582177909)*x[0]+(0.675592435624)*x[1]**o+(0.77480796903)*x[1]
        arg[0,2,1,1]=(-0.281969675471)*x[0]**o+(-0.250831251067)*x[0]+(0.14396936899)*x[1]**o+(-0.65775670156)*x[1]
        arg[0,2,2,0]=(0.339204154215)*x[0]**o+(0.125319656279)*x[0]+(0.417557775604)*x[1]**o+(-0.985431777777)*x[1]
        arg[0,2,2,1]=(-0.655500794874)*x[0]**o+(-0.419889131296)*x[0]+(-0.0763188160723)*x[1]**o+(0.995600932757)*x[1]
        arg[0,3,0,0]=(-0.483746115201)*x[0]**o+(-0.256632816498)*x[0]+(-0.0533973517656)*x[1]**o+(-0.982541041474)*x[1]
        arg[0,3,0,1]=(0.446810992304)*x[0]**o+(-0.823523374302)*x[0]+(-0.607238533539)*x[1]**o+(-0.97562384609)*x[1]
        arg[0,3,1,0]=(-0.0235907627093)*x[0]**o+(-0.0720587626585)*x[0]+(0.930589323947)*x[1]**o+(0.327916365569)*x[1]
        arg[0,3,1,1]=(0.486342062258)*x[0]**o+(0.430279082913)*x[0]+(0.237445053125)*x[1]**o+(0.577650421809)*x[1]
        arg[0,3,2,0]=(-0.581025309937)*x[0]**o+(0.97026755344)*x[0]+(0.180505264557)*x[1]**o+(0.289585805796)*x[1]
        arg[0,3,2,1]=(-0.523356041281)*x[0]**o+(-0.502230606793)*x[0]+(0.915270822508)*x[1]**o+(0.622076448072)*x[1]
        arg[0,4,0,0]=(-0.584782530121)*x[0]**o+(-0.433761163729)*x[0]+(-0.449731522195)*x[1]**o+(0.392234342377)*x[1]
        arg[0,4,0,1]=(-0.0724106581167)*x[0]**o+(0.76896002907)*x[0]+(0.0374933582543)*x[1]**o+(-0.417790382045)*x[1]
        arg[0,4,1,0]=(0.0460360663465)*x[0]**o+(-0.0343116732267)*x[0]+(0.815864788848)*x[1]**o+(-0.922236546574)*x[1]
        arg[0,4,1,1]=(-0.978584425712)*x[0]**o+(0.249337867126)*x[0]+(-0.822961904166)*x[1]**o+(0.627473480274)*x[1]
        arg[0,4,2,0]=(-0.267227111945)*x[0]**o+(0.791320996051)*x[0]+(0.721009228711)*x[1]**o+(0.00199546728243)*x[1]
        arg[0,4,2,1]=(-0.141372777283)*x[0]**o+(-0.996912487736)*x[0]+(0.94281427253)*x[1]**o+(-0.780685413978)*x[1]
        arg[1,0,0,0]=(-0.721911758033)*x[0]**o+(0.0119756829508)*x[0]+(0.583083015744)*x[1]**o+(0.80826073507)*x[1]
        arg[1,0,0,1]=(0.929636966638)*x[0]**o+(-0.35107696413)*x[0]+(-0.656523378462)*x[1]**o+(0.610681274197)*x[1]
        arg[1,0,1,0]=(0.594847872723)*x[0]**o+(-0.15013091781)*x[0]+(-0.562311637909)*x[1]**o+(0.871283992407)*x[1]
        arg[1,0,1,1]=(0.529274819592)*x[0]**o+(0.998035152453)*x[0]+(-0.265309908189)*x[1]**o+(-0.703391843389)*x[1]
        arg[1,0,2,0]=(0.2321959905)*x[0]**o+(-0.781602927232)*x[0]+(-0.0509405210126)*x[1]**o+(0.839298348419)*x[1]
        arg[1,0,2,1]=(0.0811824999287)*x[0]**o+(0.916373714851)*x[0]+(0.817375877796)*x[1]**o+(-0.926274023089)*x[1]
        arg[1,1,0,0]=(0.418525096774)*x[0]**o+(0.064653235493)*x[0]+(0.15912570069)*x[1]**o+(0.47570489259)*x[1]
        arg[1,1,0,1]=(0.784138248952)*x[0]**o+(-0.715130508682)*x[0]+(-0.299421222048)*x[1]**o+(-0.816706522916)*x[1]
        arg[1,1,1,0]=(-0.911449050332)*x[0]**o+(-0.472182975049)*x[0]+(0.935180253062)*x[1]**o+(0.527208406385)*x[1]
        arg[1,1,1,1]=(-0.378301315644)*x[0]**o+(0.430065353544)*x[0]+(0.827581865806)*x[1]**o+(-0.355263817375)*x[1]
        arg[1,1,2,0]=(-0.882543004053)*x[0]**o+(0.503794963637)*x[0]+(-0.580086073626)*x[1]**o+(0.545676378283)*x[1]
        arg[1,1,2,1]=(-0.116197074531)*x[0]**o+(-0.0629871602139)*x[0]+(0.320343551712)*x[1]**o+(0.0552550885696)*x[1]
        arg[1,2,0,0]=(-0.870347384636)*x[0]**o+(-0.611439594673)*x[0]+(0.707766015783)*x[1]**o+(0.0653036268139)*x[1]
        arg[1,2,0,1]=(-0.219872666014)*x[0]**o+(-0.435867669656)*x[0]+(0.825489444837)*x[1]**o+(-0.737253859382)*x[1]
        arg[1,2,1,0]=(-0.963149287055)*x[0]**o+(0.964306345088)*x[0]+(0.502736178959)*x[1]**o+(0.998017404878)*x[1]
        arg[1,2,1,1]=(-0.586350249194)*x[0]**o+(-0.437455605038)*x[0]+(-0.182719660538)*x[1]**o+(-0.50797820305)*x[1]
        arg[1,2,2,0]=(-0.423818475166)*x[0]**o+(0.290902418459)*x[0]+(-0.304129806418)*x[1]**o+(-0.67571878331)*x[1]
        arg[1,2,2,1]=(0.901384945886)*x[0]**o+(0.0526906910209)*x[0]+(0.438755721196)*x[1]**o+(-0.774352941544)*x[1]
        arg[1,3,0,0]=(-0.797968602257)*x[0]**o+(0.856637297517)*x[0]+(0.427243495756)*x[1]**o+(0.212942988457)*x[1]
        arg[1,3,0,1]=(-0.696126104381)*x[0]**o+(0.616445534254)*x[0]+(-0.943037028688)*x[1]**o+(0.807265721375)*x[1]
        arg[1,3,1,0]=(-0.69028368426)*x[0]**o+(-0.652017674488)*x[0]+(-0.772965740874)*x[1]**o+(-0.726830598291)*x[1]
        arg[1,3,1,1]=(-0.534347181294)*x[0]**o+(-0.861020241321)*x[0]+(-0.250549859157)*x[1]**o+(-0.638289396201)*x[1]
        arg[1,3,2,0]=(-0.827981962115)*x[0]**o+(-0.427544131008)*x[0]+(-0.516514308221)*x[1]**o+(-0.101906062547)*x[1]
        arg[1,3,2,1]=(0.864701869194)*x[0]**o+(0.876327129524)*x[0]+(0.806201346091)*x[1]**o+(0.845259832272)*x[1]
        arg[1,4,0,0]=(-0.589044559231)*x[0]**o+(0.697104039773)*x[0]+(-0.288389617381)*x[1]**o+(0.422618969092)*x[1]
        arg[1,4,0,1]=(0.814971550967)*x[0]**o+(-0.665151459593)*x[0]+(-0.696145763418)*x[1]**o+(0.906949165457)*x[1]
        arg[1,4,1,0]=(0.760190450826)*x[0]**o+(-0.883964913626)*x[0]+(-0.507159320061)*x[1]**o+(-0.15707026975)*x[1]
        arg[1,4,1,1]=(-0.0416603697257)*x[0]**o+(0.560561779164)*x[0]+(0.0153574982122)*x[1]**o+(0.75228872946)*x[1]
        arg[1,4,2,0]=(0.105929120872)*x[0]**o+(0.130738968719)*x[0]+(-0.0991159921294)*x[1]**o+(0.972627650896)*x[1]
        arg[1,4,2,1]=(-0.503911541225)*x[0]**o+(-0.00357063650385)*x[0]+(0.377083955928)*x[1]**o+(0.343010153434)*x[1]
        arg[2,0,0,0]=(0.953733725947)*x[0]**o+(-0.988629842686)*x[0]+(0.196766730386)*x[1]**o+(0.0427993038279)*x[1]
        arg[2,0,0,1]=(0.491469820597)*x[0]**o+(0.708248604732)*x[0]+(-0.600347310418)*x[1]**o+(-0.0380188651191)*x[1]
        arg[2,0,1,0]=(0.615118477478)*x[0]**o+(0.832030863133)*x[0]+(0.783037278839)*x[1]**o+(-0.331824170309)*x[1]
        arg[2,0,1,1]=(-0.806125471265)*x[0]**o+(0.763233458333)*x[0]+(-0.381389710969)*x[1]**o+(0.652974431609)*x[1]
        arg[2,0,2,0]=(-0.831727679795)*x[0]**o+(0.239296785135)*x[0]+(-0.13737609858)*x[1]**o+(0.583210422723)*x[1]
        arg[2,0,2,1]=(-0.805843206194)*x[0]**o+(-0.523576669019)*x[0]+(0.159758297984)*x[1]**o+(-0.298308874886)*x[1]
        arg[2,1,0,0]=(0.589135530105)*x[0]**o+(0.272145879514)*x[0]+(0.325741074892)*x[1]**o+(-0.579782654647)*x[1]
        arg[2,1,0,1]=(0.106049062774)*x[0]**o+(0.790280056075)*x[0]+(0.0171600098883)*x[1]**o+(-0.419540445094)*x[1]
        arg[2,1,1,0]=(-0.181511670026)*x[0]**o+(-0.814531462508)*x[0]+(-0.576165158737)*x[1]**o+(0.679678310717)*x[1]
        arg[2,1,1,1]=(-0.709695453728)*x[0]**o+(-0.489487606257)*x[0]+(-0.424093531654)*x[1]**o+(0.535427169862)*x[1]
        arg[2,1,2,0]=(0.60069280023)*x[0]**o+(-0.909882731214)*x[0]+(0.640141428017)*x[1]**o+(-0.697396064954)*x[1]
        arg[2,1,2,1]=(0.896964594562)*x[0]**o+(0.976387328697)*x[0]+(0.144061490066)*x[1]**o+(0.378042869796)*x[1]
        arg[2,2,0,0]=(0.447633246258)*x[0]**o+(-0.531416367062)*x[0]+(0.199304002742)*x[1]**o+(0.85195706213)*x[1]
        arg[2,2,0,1]=(-0.683614706554)*x[0]**o+(0.933255565209)*x[0]+(-0.328047392082)*x[1]**o+(-0.0865769137182)*x[1]
        arg[2,2,1,0]=(0.22474720984)*x[0]**o+(0.0910500178874)*x[0]+(0.449220001922)*x[1]**o+(-0.0406214557829)*x[1]
        arg[2,2,1,1]=(-0.266419887395)*x[0]**o+(-0.48883285131)*x[0]+(-0.794933934805)*x[1]**o+(0.182439649969)*x[1]
        arg[2,2,2,0]=(0.949852186215)*x[0]**o+(-0.116121704259)*x[0]+(-0.977131144423)*x[1]**o+(0.698402625747)*x[1]
        arg[2,2,2,1]=(-0.0446180768654)*x[0]**o+(-0.742263614438)*x[0]+(-0.898675367333)*x[1]**o+(-0.260891146746)*x[1]
        arg[2,3,0,0]=(0.819774310005)*x[0]**o+(-0.0677371406664)*x[0]+(0.276747165681)*x[1]**o+(-0.42828783)*x[1]
        arg[2,3,0,1]=(-0.391001671385)*x[0]**o+(0.0558814994728)*x[0]+(-0.624786978129)*x[1]**o+(0.178742280566)*x[1]
        arg[2,3,1,0]=(0.904597103646)*x[0]**o+(-0.825472271707)*x[0]+(0.112221476422)*x[1]**o+(0.269349199527)*x[1]
        arg[2,3,1,1]=(-0.190955697638)*x[0]**o+(0.441698994606)*x[0]+(-0.26229840557)*x[1]**o+(0.148974540187)*x[1]
        arg[2,3,2,0]=(-0.873098622934)*x[0]**o+(-0.624915584376)*x[0]+(-0.151646578512)*x[1]**o+(-0.914878228283)*x[1]
        arg[2,3,2,1]=(-0.712569947477)*x[0]**o+(-0.72679361801)*x[0]+(-0.0946869166491)*x[1]**o+(-0.580512407933)*x[1]
        arg[2,4,0,0]=(-0.655752680768)*x[0]**o+(-0.409544975855)*x[0]+(0.865325438086)*x[1]**o+(-0.227397026937)*x[1]
        arg[2,4,0,1]=(0.670566711209)*x[0]**o+(-0.794343533883)*x[0]+(-0.127116396219)*x[1]**o+(-0.422419866647)*x[1]
        arg[2,4,1,0]=(0.46903441444)*x[0]**o+(0.0689129759715)*x[0]+(-0.549325647826)*x[1]**o+(0.893700020399)*x[1]
        arg[2,4,1,1]=(-0.0099246750652)*x[0]**o+(0.446742615509)*x[0]+(0.900251202513)*x[1]**o+(0.221599931055)*x[1]
        arg[2,4,2,0]=(0.218854380066)*x[0]**o+(0.877849336909)*x[0]+(-0.338211083089)*x[1]**o+(0.746950973063)*x[1]
        arg[2,4,2,1]=(-0.890762146523)*x[0]**o+(0.439653904008)*x[0]+(0.0802890292215)*x[1]**o+(-0.305233739845)*x[1]
        arg[3,0,0,0]=(0.284622209957)*x[0]**o+(0.983284338924)*x[0]+(-0.732588048234)*x[1]**o+(0.654066388109)*x[1]
        arg[3,0,0,1]=(0.379628076244)*x[0]**o+(-0.75362875047)*x[0]+(0.288422119968)*x[1]**o+(0.139197401407)*x[1]
        arg[3,0,1,0]=(0.0594472092467)*x[0]**o+(-0.391581973307)*x[0]+(-0.898996153568)*x[1]**o+(-2.75165483707e-05)*x[1]
        arg[3,0,1,1]=(-0.241252797172)*x[0]**o+(-0.87384982932)*x[0]+(0.811628957112)*x[1]**o+(0.0170539509618)*x[1]
        arg[3,0,2,0]=(0.255617459837)*x[0]**o+(-0.744071491889)*x[0]+(0.745552769828)*x[1]**o+(0.128612266243)*x[1]
        arg[3,0,2,1]=(0.0725082197254)*x[0]**o+(0.779602587374)*x[0]+(-0.751879287668)*x[1]**o+(0.838360641095)*x[1]
        arg[3,1,0,0]=(-0.444520674459)*x[0]**o+(-0.811442143542)*x[0]+(-0.485010478999)*x[1]**o+(0.727600464745)*x[1]
        arg[3,1,0,1]=(0.260061065042)*x[0]**o+(0.932320346965)*x[0]+(0.556377882844)*x[1]**o+(0.284348284894)*x[1]
        arg[3,1,1,0]=(0.561755002522)*x[0]**o+(-0.842983435746)*x[0]+(0.729209244377)*x[1]**o+(-0.254683846974)*x[1]
        arg[3,1,1,1]=(-0.949556399673)*x[0]**o+(-0.0804522142379)*x[0]+(0.066141483426)*x[1]**o+(0.78626145966)*x[1]
        arg[3,1,2,0]=(0.833198651881)*x[0]**o+(-0.297955946153)*x[0]+(-0.561247934789)*x[1]**o+(-0.645124432611)*x[1]
        arg[3,1,2,1]=(-0.641727656043)*x[0]**o+(0.32981201062)*x[0]+(-0.152588766563)*x[1]**o+(-0.289611460761)*x[1]
        arg[3,2,0,0]=(0.825643539598)*x[0]**o+(0.924286336436)*x[0]+(0.413089850944)*x[1]**o+(-0.640198886209)*x[1]
        arg[3,2,0,1]=(0.0855693847128)*x[0]**o+(-0.363881253576)*x[0]+(0.0397854550832)*x[1]**o+(-0.844391996875)*x[1]
        arg[3,2,1,0]=(-0.8248537795)*x[0]**o+(0.153070857979)*x[0]+(0.970009460355)*x[1]**o+(0.431000028243)*x[1]
        arg[3,2,1,1]=(-0.978024209918)*x[0]**o+(-0.527494257814)*x[0]+(-0.814691105217)*x[1]**o+(0.458334868681)*x[1]
        arg[3,2,2,0]=(0.944359157869)*x[0]**o+(-0.868016282191)*x[0]+(0.410849160429)*x[1]**o+(0.480177162607)*x[1]
        arg[3,2,2,1]=(0.109166287134)*x[0]**o+(0.940100154901)*x[0]+(0.630946061805)*x[1]**o+(0.347160020469)*x[1]
        arg[3,3,0,0]=(0.364585502975)*x[0]**o+(0.717282598047)*x[0]+(0.517805310585)*x[1]**o+(0.963452728824)*x[1]
        arg[3,3,0,1]=(0.0373453558524)*x[0]**o+(0.943887402198)*x[0]+(0.983624425071)*x[1]**o+(-0.396242794059)*x[1]
        arg[3,3,1,0]=(-0.563724031351)*x[0]**o+(0.959753988518)*x[0]+(0.118506885299)*x[1]**o+(-0.587748202342)*x[1]
        arg[3,3,1,1]=(0.460419807013)*x[0]**o+(0.237753399499)*x[0]+(0.585589833251)*x[1]**o+(0.935177937667)*x[1]
        arg[3,3,2,0]=(-0.750248086829)*x[0]**o+(-0.561629848994)*x[0]+(-0.919919077536)*x[1]**o+(0.239984179086)*x[1]
        arg[3,3,2,1]=(-0.0969897163659)*x[0]**o+(0.185433022889)*x[0]+(0.733111631837)*x[1]**o+(-0.508147081928)*x[1]
        arg[3,4,0,0]=(0.247054568876)*x[0]**o+(0.721751593201)*x[0]+(-0.684740924498)*x[1]**o+(0.28395202476)*x[1]
        arg[3,4,0,1]=(-0.0949248638736)*x[0]**o+(0.270285273316)*x[0]+(0.850836559464)*x[1]**o+(-0.771085947382)*x[1]
        arg[3,4,1,0]=(-0.965410502362)*x[0]**o+(-0.582448742001)*x[0]+(-0.90307507827)*x[1]**o+(0.914051445745)*x[1]
        arg[3,4,1,1]=(0.443340058617)*x[0]**o+(0.434462120769)*x[0]+(0.522015964301)*x[1]**o+(-0.520184389865)*x[1]
        arg[3,4,2,0]=(-0.961375562593)*x[0]**o+(-0.474630747481)*x[0]+(0.59574636341)*x[1]**o+(0.434190359663)*x[1]
        arg[3,4,2,1]=(0.174154210501)*x[0]**o+(-0.510858596799)*x[0]+(-0.0279700806065)*x[1]**o+(-0.0401612371912)*x[1]
        ref[0,0,0,0]=(-0.0352052809673)/(o+1.)+(0.903863789878)+(0.182956096072)*0.5**o
        ref[0,0,0,1]=(0.0610373691872)/(o+1.)+(-0.306742561603)+(-0.81342189211)*0.5**o
        ref[0,0,1,0]=(0.623943596759)/(o+1.)+(0.647497589315)+(-0.149103531408)*0.5**o
        ref[0,0,1,1]=(0.136600247183)/(o+1.)+(0.925469911903)+(0.74912087771)*0.5**o
        ref[0,0,2,0]=(0.966454801911)/(o+1.)+(0.0631444946098)+(-0.289042694587)*0.5**o
        ref[0,0,2,1]=(-0.156074498516)/(o+1.)+(0.047955363258)+(0.0809737838902)*0.5**o
        ref[0,1,0,0]=(-0.016865109839)/(o+1.)+(-0.404214477389)+(0.638822193145)*0.5**o
        ref[0,1,0,1]=(-0.897448372147)/(o+1.)+(-0.174669987755)+(0.611604473356)*0.5**o
        ref[0,1,1,0]=(-0.229098133178)/(o+1.)+(0.443671911949)+(-0.745407006471)*0.5**o
        ref[0,1,1,1]=(0.481206893907)/(o+1.)+(-0.111741539797)+(0.756528549281)*0.5**o
        ref[0,1,2,0]=(0.367303506039)/(o+1.)+(-0.157755527379)+(0.335544368565)*0.5**o
        ref[0,1,2,1]=(0.182483888026)/(o+1.)+(-0.157073928893)+(0.669933676654)*0.5**o
        ref[0,2,0,0]=(0.587675819052)/(o+1.)+(0.446236706377)+(-0.929163949925)*0.5**o
        ref[0,2,0,1]=(-0.83934259298)/(o+1.)+(-0.344719194607)+(0.639896773528)*0.5**o
        ref[0,2,1,0]=(0.675592435624)/(o+1.)+(0.81769507347)+(-0.399556357179)*0.5**o
        ref[0,2,1,1]=(0.14396936899)/(o+1.)+(-0.454293976313)+(-0.281969675471)*0.5**o
        ref[0,2,2,0]=(0.417557775604)/(o+1.)+(-0.430056060749)+(0.339204154215)*0.5**o
        ref[0,2,2,1]=(-0.0763188160723)/(o+1.)+(0.28785590073)+(-0.655500794874)*0.5**o
        ref[0,3,0,0]=(-0.0533973517656)/(o+1.)+(-0.619586928986)+(-0.483746115201)*0.5**o
        ref[0,3,0,1]=(-0.607238533539)/(o+1.)+(-0.899573610196)+(0.446810992304)*0.5**o
        ref[0,3,1,0]=(0.930589323947)/(o+1.)+(0.127928801455)+(-0.0235907627093)*0.5**o
        ref[0,3,1,1]=(0.237445053125)/(o+1.)+(0.503964752361)+(0.486342062258)*0.5**o
        ref[0,3,2,0]=(0.180505264557)/(o+1.)+(0.629926679618)+(-0.581025309937)*0.5**o
        ref[0,3,2,1]=(0.915270822508)/(o+1.)+(0.0599229206396)+(-0.523356041281)*0.5**o
        ref[0,4,0,0]=(-0.449731522195)/(o+1.)+(-0.0207634106763)+(-0.584782530121)*0.5**o
        ref[0,4,0,1]=(0.0374933582543)/(o+1.)+(0.175584823513)+(-0.0724106581167)*0.5**o
        ref[0,4,1,0]=(0.815864788848)/(o+1.)+(-0.4782741099)+(0.0460360663465)*0.5**o
        ref[0,4,1,1]=(-0.822961904166)/(o+1.)+(0.4384056737)+(-0.978584425712)*0.5**o
        ref[0,4,2,0]=(0.721009228711)/(o+1.)+(0.396658231667)+(-0.267227111945)*0.5**o
        ref[0,4,2,1]=(0.94281427253)/(o+1.)+(-0.888798950857)+(-0.141372777283)*0.5**o
        ref[1,0,0,0]=(0.583083015744)/(o+1.)+(0.41011820901)+(-0.721911758033)*0.5**o
        ref[1,0,0,1]=(-0.656523378462)/(o+1.)+(0.129802155034)+(0.929636966638)*0.5**o
        ref[1,0,1,0]=(-0.562311637909)/(o+1.)+(0.360576537299)+(0.594847872723)*0.5**o
        ref[1,0,1,1]=(-0.265309908189)/(o+1.)+(0.147321654532)+(0.529274819592)*0.5**o
        ref[1,0,2,0]=(-0.0509405210126)/(o+1.)+(0.0288477105933)+(0.2321959905)*0.5**o
        ref[1,0,2,1]=(0.817375877796)/(o+1.)+(-0.00495015411903)+(0.0811824999287)*0.5**o
        ref[1,1,0,0]=(0.15912570069)/(o+1.)+(0.270179064041)+(0.418525096774)*0.5**o
        ref[1,1,0,1]=(-0.299421222048)/(o+1.)+(-0.765918515799)+(0.784138248952)*0.5**o
        ref[1,1,1,0]=(0.935180253062)/(o+1.)+(0.0275127156677)+(-0.911449050332)*0.5**o
        ref[1,1,1,1]=(0.827581865806)/(o+1.)+(0.0374007680845)+(-0.378301315644)*0.5**o
        ref[1,1,2,0]=(-0.580086073626)/(o+1.)+(0.52473567096)+(-0.882543004053)*0.5**o
        ref[1,1,2,1]=(0.320343551712)/(o+1.)+(-0.00386603582212)+(-0.116197074531)*0.5**o
        ref[1,2,0,0]=(0.707766015783)/(o+1.)+(-0.273067983929)+(-0.870347384636)*0.5**o
        ref[1,2,0,1]=(0.825489444837)/(o+1.)+(-0.586560764519)+(-0.219872666014)*0.5**o
        ref[1,2,1,0]=(0.502736178959)/(o+1.)+(0.981161874983)+(-0.963149287055)*0.5**o
        ref[1,2,1,1]=(-0.182719660538)/(o+1.)+(-0.472716904044)+(-0.586350249194)*0.5**o
        ref[1,2,2,0]=(-0.304129806418)/(o+1.)+(-0.192408182425)+(-0.423818475166)*0.5**o
        ref[1,2,2,1]=(0.438755721196)/(o+1.)+(-0.360831125262)+(0.901384945886)*0.5**o
        ref[1,3,0,0]=(0.427243495756)/(o+1.)+(0.534790142987)+(-0.797968602257)*0.5**o
        ref[1,3,0,1]=(-0.943037028688)/(o+1.)+(0.711855627815)+(-0.696126104381)*0.5**o
        ref[1,3,1,0]=(-0.772965740874)/(o+1.)+(-0.689424136389)+(-0.69028368426)*0.5**o
        ref[1,3,1,1]=(-0.250549859157)/(o+1.)+(-0.749654818761)+(-0.534347181294)*0.5**o
        ref[1,3,2,0]=(-0.516514308221)/(o+1.)+(-0.264725096778)+(-0.827981962115)*0.5**o
        ref[1,3,2,1]=(0.806201346091)/(o+1.)+(0.860793480898)+(0.864701869194)*0.5**o
        ref[1,4,0,0]=(-0.288389617381)/(o+1.)+(0.559861504432)+(-0.589044559231)*0.5**o
        ref[1,4,0,1]=(-0.696145763418)/(o+1.)+(0.120898852932)+(0.814971550967)*0.5**o
        ref[1,4,1,0]=(-0.507159320061)/(o+1.)+(-0.520517591688)+(0.760190450826)*0.5**o
        ref[1,4,1,1]=(0.0153574982122)/(o+1.)+(0.656425254312)+(-0.0416603697257)*0.5**o
        ref[1,4,2,0]=(-0.0991159921294)/(o+1.)+(0.551683309807)+(0.105929120872)*0.5**o
        ref[1,4,2,1]=(0.377083955928)/(o+1.)+(0.169719758465)+(-0.503911541225)*0.5**o
        ref[2,0,0,0]=(0.196766730386)/(o+1.)+(-0.472915269429)+(0.953733725947)*0.5**o
        ref[2,0,0,1]=(-0.600347310418)/(o+1.)+(0.335114869806)+(0.491469820597)*0.5**o
        ref[2,0,1,0]=(0.783037278839)/(o+1.)+(0.250103346412)+(0.615118477478)*0.5**o
        ref[2,0,1,1]=(-0.381389710969)/(o+1.)+(0.708103944971)+(-0.806125471265)*0.5**o
        ref[2,0,2,0]=(-0.13737609858)/(o+1.)+(0.411253603929)+(-0.831727679795)*0.5**o
        ref[2,0,2,1]=(0.159758297984)/(o+1.)+(-0.410942771953)+(-0.805843206194)*0.5**o
        ref[2,1,0,0]=(0.325741074892)/(o+1.)+(-0.153818387566)+(0.589135530105)*0.5**o
        ref[2,1,0,1]=(0.0171600098883)/(o+1.)+(0.185369805491)+(0.106049062774)*0.5**o
        ref[2,1,1,0]=(-0.576165158737)/(o+1.)+(-0.0674265758955)+(-0.181511670026)*0.5**o
        ref[2,1,1,1]=(-0.424093531654)/(o+1.)+(0.0229697818029)+(-0.709695453728)*0.5**o
        ref[2,1,2,0]=(0.640141428017)/(o+1.)+(-0.803639398084)+(0.60069280023)*0.5**o
        ref[2,1,2,1]=(0.144061490066)/(o+1.)+(0.677215099246)+(0.896964594562)*0.5**o
        ref[2,2,0,0]=(0.199304002742)/(o+1.)+(0.160270347534)+(0.447633246258)*0.5**o
        ref[2,2,0,1]=(-0.328047392082)/(o+1.)+(0.423339325745)+(-0.683614706554)*0.5**o
        ref[2,2,1,0]=(0.449220001922)/(o+1.)+(0.0252142810523)+(0.22474720984)*0.5**o
        ref[2,2,1,1]=(-0.794933934805)/(o+1.)+(-0.153196600671)+(-0.266419887395)*0.5**o
        ref[2,2,2,0]=(-0.977131144423)/(o+1.)+(0.291140460744)+(0.949852186215)*0.5**o
        ref[2,2,2,1]=(-0.898675367333)/(o+1.)+(-0.501577380592)+(-0.0446180768654)*0.5**o
        ref[2,3,0,0]=(0.276747165681)/(o+1.)+(-0.248012485333)+(0.819774310005)*0.5**o
        ref[2,3,0,1]=(-0.624786978129)/(o+1.)+(0.11731189002)+(-0.391001671385)*0.5**o
        ref[2,3,1,0]=(0.112221476422)/(o+1.)+(-0.27806153609)+(0.904597103646)*0.5**o
        ref[2,3,1,1]=(-0.26229840557)/(o+1.)+(0.295336767397)+(-0.190955697638)*0.5**o
        ref[2,3,2,0]=(-0.151646578512)/(o+1.)+(-0.76989690633)+(-0.873098622934)*0.5**o
        ref[2,3,2,1]=(-0.0946869166491)/(o+1.)+(-0.653653012971)+(-0.712569947477)*0.5**o
        ref[2,4,0,0]=(0.865325438086)/(o+1.)+(-0.318471001396)+(-0.655752680768)*0.5**o
        ref[2,4,0,1]=(-0.127116396219)/(o+1.)+(-0.608381700265)+(0.670566711209)*0.5**o
        ref[2,4,1,0]=(-0.549325647826)/(o+1.)+(0.481306498185)+(0.46903441444)*0.5**o
        ref[2,4,1,1]=(0.900251202513)/(o+1.)+(0.334171273282)+(-0.0099246750652)*0.5**o
        ref[2,4,2,0]=(-0.338211083089)/(o+1.)+(0.812400154986)+(0.218854380066)*0.5**o
        ref[2,4,2,1]=(0.0802890292215)/(o+1.)+(0.0672100820814)+(-0.890762146523)*0.5**o
        ref[3,0,0,0]=(-0.732588048234)/(o+1.)+(0.818675363516)+(0.284622209957)*0.5**o
        ref[3,0,0,1]=(0.288422119968)/(o+1.)+(-0.307215674532)+(0.379628076244)*0.5**o
        ref[3,0,1,0]=(-0.898996153568)/(o+1.)+(-0.195804744928)+(0.0594472092467)*0.5**o
        ref[3,0,1,1]=(0.811628957112)/(o+1.)+(-0.428397939179)+(-0.241252797172)*0.5**o
        ref[3,0,2,0]=(0.745552769828)/(o+1.)+(-0.307729612823)+(0.255617459837)*0.5**o
        ref[3,0,2,1]=(-0.751879287668)/(o+1.)+(0.808981614234)+(0.0725082197254)*0.5**o
        ref[3,1,0,0]=(-0.485010478999)/(o+1.)+(-0.0419208393982)+(-0.444520674459)*0.5**o
        ref[3,1,0,1]=(0.556377882844)/(o+1.)+(0.60833431593)+(0.260061065042)*0.5**o
        ref[3,1,1,0]=(0.729209244377)/(o+1.)+(-0.54883364136)+(0.561755002522)*0.5**o
        ref[3,1,1,1]=(0.066141483426)/(o+1.)+(0.352904622711)+(-0.949556399673)*0.5**o
        ref[3,1,2,0]=(-0.561247934789)/(o+1.)+(-0.471540189382)+(0.833198651881)*0.5**o
        ref[3,1,2,1]=(-0.152588766563)/(o+1.)+(0.0201002749295)+(-0.641727656043)*0.5**o
        ref[3,2,0,0]=(0.413089850944)/(o+1.)+(0.142043725114)+(0.825643539598)*0.5**o
        ref[3,2,0,1]=(0.0397854550832)/(o+1.)+(-0.604136625225)+(0.0855693847128)*0.5**o
        ref[3,2,1,0]=(0.970009460355)/(o+1.)+(0.292035443111)+(-0.8248537795)*0.5**o
        ref[3,2,1,1]=(-0.814691105217)/(o+1.)+(-0.0345796945668)+(-0.978024209918)*0.5**o
        ref[3,2,2,0]=(0.410849160429)/(o+1.)+(-0.193919559792)+(0.944359157869)*0.5**o
        ref[3,2,2,1]=(0.630946061805)/(o+1.)+(0.643630087685)+(0.109166287134)*0.5**o
        ref[3,3,0,0]=(0.517805310585)/(o+1.)+(0.840367663435)+(0.364585502975)*0.5**o
        ref[3,3,0,1]=(0.983624425071)/(o+1.)+(0.273822304069)+(0.0373453558524)*0.5**o
        ref[3,3,1,0]=(0.118506885299)/(o+1.)+(0.186002893088)+(-0.563724031351)*0.5**o
        ref[3,3,1,1]=(0.585589833251)/(o+1.)+(0.586465668583)+(0.460419807013)*0.5**o
        ref[3,3,2,0]=(-0.919919077536)/(o+1.)+(-0.160822834954)+(-0.750248086829)*0.5**o
        ref[3,3,2,1]=(0.733111631837)/(o+1.)+(-0.161357029519)+(-0.0969897163659)*0.5**o
        ref[3,4,0,0]=(-0.684740924498)/(o+1.)+(0.50285180898)+(0.247054568876)*0.5**o
        ref[3,4,0,1]=(0.850836559464)/(o+1.)+(-0.250400337033)+(-0.0949248638736)*0.5**o
        ref[3,4,1,0]=(-0.90307507827)/(o+1.)+(0.165801351872)+(-0.965410502362)*0.5**o
        ref[3,4,1,1]=(0.522015964301)/(o+1.)+(-0.0428611345479)+(0.443340058617)*0.5**o
        ref[3,4,2,0]=(0.59574636341)/(o+1.)+(-0.0202201939092)+(-0.961375562593)*0.5**o
        ref[3,4,2,1]=(-0.0279700806065)/(o+1.)+(-0.275509916995)+(0.174154210501)*0.5**o
      else:
        arg[0,0,0,0]=(-0.793251694335)*x[0]**o+(0.616970271582)*x[0]+(-0.805609243189)*x[1]**o+(0.949515692474)*x[1]+(0.484069738431)*x[2]**o+(-0.0407533911672)*x[2]
        arg[0,0,0,1]=(-0.624206604764)*x[0]**o+(-0.992961148702)*x[0]+(-0.737741763879)*x[1]**o+(0.161544434398)*x[1]+(0.47292947772)*x[2]**o+(0.525538846945)*x[2]
        arg[0,0,1,0]=(0.115999561481)*x[0]**o+(-0.612694096044)*x[0]+(0.625539443386)*x[1]**o+(0.857353715882)*x[1]+(-0.0811256709155)*x[2]**o+(0.183474706394)*x[2]
        arg[0,0,1,1]=(-0.972466831671)*x[0]**o+(0.262980887655)*x[0]+(0.920924576059)*x[1]**o+(0.983517658667)*x[1]+(0.579443973505)*x[2]**o+(0.101139612823)*x[2]
        arg[0,0,2,0]=(0.708180260094)*x[0]**o+(-0.0950913766285)*x[0]+(0.361602191798)*x[1]**o+(-0.880757848201)*x[1]+(0.326272205058)*x[2]**o+(-0.997155694972)*x[2]
        arg[0,0,2,1]=(0.0282615573045)*x[0]**o+(-0.317851994728)*x[0]+(0.449330147388)*x[1]**o+(-0.156946318079)*x[1]+(0.344196894245)*x[2]**o+(-0.045082290904)*x[2]
        arg[0,1,0,0]=(0.231621820235)*x[0]**o+(0.0985804817846)*x[0]+(-0.343623333022)*x[1]**o+(0.115043475441)*x[1]+(-0.910742772023)*x[2]**o+(-0.663592068553)*x[2]
        arg[0,1,0,1]=(0.114056801155)*x[0]**o+(-0.463711315494)*x[0]+(0.380521147429)*x[1]**o+(0.294823684238)*x[1]+(0.510501672453)*x[2]**o+(-0.161234251875)*x[2]
        arg[0,1,1,0]=(-0.973115195017)*x[0]**o+(-0.0133290715609)*x[0]+(-0.642072931588)*x[1]**o+(-0.580655605735)*x[1]+(-0.191371539061)*x[2]**o+(-0.0513051439064)*x[2]
        arg[0,1,1,1]=(-0.304758579369)*x[0]**o+(-0.321904612871)*x[0]+(-0.389525877547)*x[1]**o+(-0.0602638359252)*x[1]+(0.199726168523)*x[2]**o+(-0.10313327614)*x[2]
        arg[0,1,2,0]=(0.0923844967157)*x[0]**o+(-0.369319307184)*x[0]+(-0.614761123101)*x[1]**o+(0.78817430449)*x[1]+(-0.783809995839)*x[2]**o+(-0.701522582246)*x[2]
        arg[0,1,2,1]=(0.793568921835)*x[0]**o+(0.352244892726)*x[0]+(0.185546719529)*x[1]**o+(0.145969792058)*x[1]+(0.85147546253)*x[2]**o+(-0.862069843156)*x[2]
        arg[0,2,0,0]=(0.570673248048)*x[0]**o+(0.607924942767)*x[0]+(-0.353952625913)*x[1]**o+(0.111796842318)*x[1]+(-0.765670542263)*x[2]**o+(0.589532724922)*x[2]
        arg[0,2,0,1]=(0.299061155215)*x[0]**o+(-0.542978954882)*x[0]+(0.0946151623892)*x[1]**o+(0.310266302534)*x[1]+(-0.00880229885823)*x[2]**o+(0.585508874849)*x[2]
        arg[0,2,1,0]=(0.694725244159)*x[0]**o+(0.530004132186)*x[0]+(-0.32588478629)*x[1]**o+(0.357278329148)*x[1]+(-0.93877612203)*x[2]**o+(0.34973013935)*x[2]
        arg[0,2,1,1]=(-0.83896527178)*x[0]**o+(0.487231460025)*x[0]+(0.493952010418)*x[1]**o+(0.247442509692)*x[1]+(-0.312456515075)*x[2]**o+(0.90248200924)*x[2]
        arg[0,2,2,0]=(0.2228052704)*x[0]**o+(0.901373696962)*x[0]+(0.149826948427)*x[1]**o+(0.234013159005)*x[1]+(0.717685535862)*x[2]**o+(-0.42865545543)*x[2]
        arg[0,2,2,1]=(0.63141216243)*x[0]**o+(0.448446314659)*x[0]+(0.379625538779)*x[1]**o+(-0.807448154272)*x[1]+(0.238724835223)*x[2]**o+(-0.478216911196)*x[2]
        arg[0,3,0,0]=(0.00698107166168)*x[0]**o+(-0.727207190747)*x[0]+(-0.78731647832)*x[1]**o+(0.885371753974)*x[1]+(-0.0146821479126)*x[2]**o+(-0.763310617205)*x[2]
        arg[0,3,0,1]=(0.243423742325)*x[0]**o+(0.0587322175837)*x[0]+(-0.453213790216)*x[1]**o+(0.861270481703)*x[1]+(0.867332526083)*x[2]**o+(-0.713187214416)*x[2]
        arg[0,3,1,0]=(-0.106629554393)*x[0]**o+(0.723693136654)*x[0]+(-0.0276604293602)*x[1]**o+(0.339456904686)*x[1]+(-0.102075459215)*x[2]**o+(-0.0348969356596)*x[2]
        arg[0,3,1,1]=(0.375143733111)*x[0]**o+(-0.490781234413)*x[0]+(0.168903339622)*x[1]**o+(0.594080859808)*x[1]+(-0.0802336915361)*x[2]**o+(0.472348979787)*x[2]
        arg[0,3,2,0]=(0.955685455438)*x[0]**o+(0.105142628533)*x[0]+(-0.446384800971)*x[1]**o+(-0.905998305229)*x[1]+(-0.30011637186)*x[2]**o+(0.0218195326037)*x[2]
        arg[0,3,2,1]=(0.353654445481)*x[0]**o+(-0.0477024594709)*x[0]+(-0.495075817497)*x[1]**o+(-0.425091398447)*x[1]+(-0.571207600446)*x[2]**o+(0.177935012163)*x[2]
        arg[0,4,0,0]=(0.974759684151)*x[0]**o+(-0.876975976179)*x[0]+(-0.472336406836)*x[1]**o+(0.991410874408)*x[1]+(0.297823444758)*x[2]**o+(-0.161436760659)*x[2]
        arg[0,4,0,1]=(-0.0799922241539)*x[0]**o+(0.975369691404)*x[0]+(0.302361880712)*x[1]**o+(0.171779641966)*x[1]+(0.878428630413)*x[2]**o+(0.814497247319)*x[2]
        arg[0,4,1,0]=(0.532633280629)*x[0]**o+(0.902436316468)*x[0]+(-0.676910177358)*x[1]**o+(-0.986029752926)*x[1]+(-0.519990783496)*x[2]**o+(0.430468215617)*x[2]
        arg[0,4,1,1]=(-0.353243658293)*x[0]**o+(-0.908938062877)*x[0]+(0.922933758125)*x[1]**o+(0.00500161721147)*x[1]+(0.189861558609)*x[2]**o+(-0.623307494462)*x[2]
        arg[0,4,2,0]=(-0.720840869801)*x[0]**o+(0.0349736577597)*x[0]+(-0.516608060915)*x[1]**o+(0.688756843577)*x[1]+(-0.302494303699)*x[2]**o+(-0.0703783111937)*x[2]
        arg[0,4,2,1]=(-0.179073400214)*x[0]**o+(0.128485402427)*x[0]+(-0.726789308786)*x[1]**o+(-0.691163482057)*x[1]+(0.329009412577)*x[2]**o+(-0.79293394396)*x[2]
        arg[1,0,0,0]=(-0.384011553333)*x[0]**o+(0.285271264922)*x[0]+(0.512105930181)*x[1]**o+(0.190321101872)*x[1]+(-0.509334982029)*x[2]**o+(0.128097742502)*x[2]
        arg[1,0,0,1]=(-0.808211570649)*x[0]**o+(-0.730450956462)*x[0]+(0.92309757237)*x[1]**o+(-0.703749565272)*x[1]+(0.301129716544)*x[2]**o+(0.244379489105)*x[2]
        arg[1,0,1,0]=(0.0404274840294)*x[0]**o+(-0.483903392253)*x[0]+(0.73680959705)*x[1]**o+(0.190850758522)*x[1]+(0.133136432605)*x[2]**o+(0.910055186498)*x[2]
        arg[1,0,1,1]=(0.302394139901)*x[0]**o+(0.802524687002)*x[0]+(0.970871728408)*x[1]**o+(0.207348092291)*x[1]+(0.666310912781)*x[2]**o+(0.510392805767)*x[2]
        arg[1,0,2,0]=(-0.981981997649)*x[0]**o+(-0.465269233707)*x[0]+(0.916146418639)*x[1]**o+(0.143602438774)*x[1]+(-0.947874263512)*x[2]**o+(0.0840394012229)*x[2]
        arg[1,0,2,1]=(-0.237479805943)*x[0]**o+(-0.89993757244)*x[0]+(-0.759650487895)*x[1]**o+(-0.807139164373)*x[1]+(-0.377055204014)*x[2]**o+(-0.57656201026)*x[2]
        arg[1,1,0,0]=(-0.543719871098)*x[0]**o+(-0.726308160246)*x[0]+(0.821244616742)*x[1]**o+(0.718112961487)*x[1]+(-0.655445553192)*x[2]**o+(-0.599104608292)*x[2]
        arg[1,1,0,1]=(-0.2624625471)*x[0]**o+(0.85250071151)*x[0]+(-0.665066081561)*x[1]**o+(-0.566502138196)*x[1]+(-0.572858653628)*x[2]**o+(-0.862890633323)*x[2]
        arg[1,1,1,0]=(0.88003513115)*x[0]**o+(-0.0150959972442)*x[0]+(-0.698880737368)*x[1]**o+(-0.236909810427)*x[1]+(-0.721315930706)*x[2]**o+(-0.219894530829)*x[2]
        arg[1,1,1,1]=(0.177746615088)*x[0]**o+(0.205790814183)*x[0]+(-0.216021190604)*x[1]**o+(0.384079010991)*x[1]+(0.61634008905)*x[2]**o+(0.940900776108)*x[2]
        arg[1,1,2,0]=(0.889070121473)*x[0]**o+(0.68131533527)*x[0]+(0.709207182081)*x[1]**o+(-0.765184991252)*x[1]+(0.506328018952)*x[2]**o+(-0.46041174473)*x[2]
        arg[1,1,2,1]=(0.951736215954)*x[0]**o+(-0.147499702642)*x[0]+(0.60348358384)*x[1]**o+(0.777404433567)*x[1]+(-0.482441452679)*x[2]**o+(0.714835256434)*x[2]
        arg[1,2,0,0]=(-0.954701036263)*x[0]**o+(-0.360901887475)*x[0]+(-0.644285028623)*x[1]**o+(0.169271524632)*x[1]+(-0.687047203668)*x[2]**o+(0.205115312108)*x[2]
        arg[1,2,0,1]=(0.256489282842)*x[0]**o+(0.17406448362)*x[0]+(0.97500243816)*x[1]**o+(0.614577997602)*x[1]+(0.0537536804583)*x[2]**o+(-0.211963274032)*x[2]
        arg[1,2,1,0]=(-0.94015914517)*x[0]**o+(0.169982490736)*x[0]+(0.482085495724)*x[1]**o+(-0.556994404153)*x[1]+(-0.809311296883)*x[2]**o+(0.719907503024)*x[2]
        arg[1,2,1,1]=(0.367156421931)*x[0]**o+(0.489648825153)*x[0]+(-0.0398225517816)*x[1]**o+(-0.495726296032)*x[1]+(0.359374061053)*x[2]**o+(0.231991952482)*x[2]
        arg[1,2,2,0]=(-0.812516694781)*x[0]**o+(-0.631181257391)*x[0]+(-0.723784471687)*x[1]**o+(0.178976877342)*x[1]+(-0.629293758013)*x[2]**o+(-0.313897724608)*x[2]
        arg[1,2,2,1]=(-0.497461944186)*x[0]**o+(0.268998525103)*x[0]+(-0.478017492026)*x[1]**o+(0.026395792841)*x[1]+(0.581256386137)*x[2]**o+(-0.476961644324)*x[2]
        arg[1,3,0,0]=(-0.227077550639)*x[0]**o+(0.400407424161)*x[0]+(-0.53529812303)*x[1]**o+(-0.418594299853)*x[1]+(-0.195190916346)*x[2]**o+(-0.624284483774)*x[2]
        arg[1,3,0,1]=(0.822091021645)*x[0]**o+(-0.31133140026)*x[0]+(0.0665130965049)*x[1]**o+(0.596697216047)*x[1]+(0.466195828646)*x[2]**o+(-0.477988391299)*x[2]
        arg[1,3,1,0]=(-0.0341577310045)*x[0]**o+(-0.77936644392)*x[0]+(0.954811382784)*x[1]**o+(0.837374830631)*x[1]+(0.738992234768)*x[2]**o+(-0.501187788985)*x[2]
        arg[1,3,1,1]=(-0.102867224088)*x[0]**o+(0.00829793194765)*x[0]+(0.081790556957)*x[1]**o+(-0.779351210472)*x[1]+(-0.735321115716)*x[2]**o+(0.126919542418)*x[2]
        arg[1,3,2,0]=(0.556301076488)*x[0]**o+(0.0923434629749)*x[0]+(-0.191048735713)*x[1]**o+(-0.218833028687)*x[1]+(-0.088867516146)*x[2]**o+(-0.247214200389)*x[2]
        arg[1,3,2,1]=(0.888436531887)*x[0]**o+(-0.79225495914)*x[0]+(0.0154788262219)*x[1]**o+(-0.210622890864)*x[1]+(0.782629134632)*x[2]**o+(0.350772185791)*x[2]
        arg[1,4,0,0]=(0.651898073685)*x[0]**o+(-0.84098616689)*x[0]+(0.674088959443)*x[1]**o+(-0.658824952898)*x[1]+(-0.423551541558)*x[2]**o+(0.985098530356)*x[2]
        arg[1,4,0,1]=(-0.149187282052)*x[0]**o+(0.0880370482988)*x[0]+(0.780809276595)*x[1]**o+(-0.610833366472)*x[1]+(0.794946384304)*x[2]**o+(-0.817031796429)*x[2]
        arg[1,4,1,0]=(-0.285938055108)*x[0]**o+(-0.106165154734)*x[0]+(-0.948376845189)*x[1]**o+(0.874607682056)*x[1]+(-0.100051825909)*x[2]**o+(-0.483640835026)*x[2]
        arg[1,4,1,1]=(-0.34015136472)*x[0]**o+(0.991331180352)*x[0]+(-0.925796518928)*x[1]**o+(-0.872775691287)*x[1]+(0.953898056819)*x[2]**o+(0.985025365955)*x[2]
        arg[1,4,2,0]=(-0.464687865356)*x[0]**o+(-0.439271850879)*x[0]+(0.302716649876)*x[1]**o+(0.235186289965)*x[1]+(0.672062761925)*x[2]**o+(-0.793689262392)*x[2]
        arg[1,4,2,1]=(-0.777967458564)*x[0]**o+(0.0565891925223)*x[0]+(-0.27539733623)*x[1]**o+(0.252329868046)*x[1]+(-0.269529792944)*x[2]**o+(-0.953051723483)*x[2]
        arg[2,0,0,0]=(0.665321754689)*x[0]**o+(0.00639681939987)*x[0]+(0.262193891905)*x[1]**o+(-0.312800735151)*x[1]+(0.8403244871)*x[2]**o+(-0.607084328204)*x[2]
        arg[2,0,0,1]=(0.876435018569)*x[0]**o+(-0.388250784543)*x[0]+(-0.799866349878)*x[1]**o+(0.114896594902)*x[1]+(-0.438064877186)*x[2]**o+(-0.458871344685)*x[2]
        arg[2,0,1,0]=(-0.344552466761)*x[0]**o+(0.967217640605)*x[0]+(-0.00183230845245)*x[1]**o+(-0.861206591247)*x[1]+(0.795588257851)*x[2]**o+(-0.333607767932)*x[2]
        arg[2,0,1,1]=(0.841069648591)*x[0]**o+(0.580162460873)*x[0]+(-0.0277035983747)*x[1]**o+(-0.404616536792)*x[1]+(-0.403195438454)*x[2]**o+(-0.202877947947)*x[2]
        arg[2,0,2,0]=(0.192691357169)*x[0]**o+(-0.519968251714)*x[0]+(-0.575397578091)*x[1]**o+(-0.029378989829)*x[1]+(-0.958861195684)*x[2]**o+(0.848705509543)*x[2]
        arg[2,0,2,1]=(0.556616365284)*x[0]**o+(0.679993071506)*x[0]+(-0.800652717274)*x[1]**o+(0.979182847258)*x[1]+(-0.501953582042)*x[2]**o+(0.127003985335)*x[2]
        arg[2,1,0,0]=(0.947750123445)*x[0]**o+(-0.0350691553359)*x[0]+(0.0650485999092)*x[1]**o+(-0.131216537555)*x[1]+(0.888782559831)*x[2]**o+(0.0338703818654)*x[2]
        arg[2,1,0,1]=(0.962918564559)*x[0]**o+(-0.119367421389)*x[0]+(0.43003549616)*x[1]**o+(-0.097883763913)*x[1]+(0.230422206305)*x[2]**o+(0.187273479711)*x[2]
        arg[2,1,1,0]=(-0.549909211973)*x[0]**o+(0.263443098451)*x[0]+(-0.476703606205)*x[1]**o+(0.216159301182)*x[1]+(-0.561829938323)*x[2]**o+(0.943016557203)*x[2]
        arg[2,1,1,1]=(0.382867749635)*x[0]**o+(0.404317006935)*x[0]+(0.786117847428)*x[1]**o+(-0.80719258612)*x[1]+(0.471144889777)*x[2]**o+(0.6478813349)*x[2]
        arg[2,1,2,0]=(0.355742024596)*x[0]**o+(-0.728812391059)*x[0]+(0.766039604734)*x[1]**o+(-0.532629598363)*x[1]+(-0.518284303077)*x[2]**o+(-0.972348067947)*x[2]
        arg[2,1,2,1]=(-0.882334243081)*x[0]**o+(-0.0940633993231)*x[0]+(-0.00166643832426)*x[1]**o+(-0.971243725495)*x[1]+(-0.0422614083707)*x[2]**o+(-0.547881520447)*x[2]
        arg[2,2,0,0]=(0.372164678051)*x[0]**o+(-0.0644361811234)*x[0]+(0.886078780084)*x[1]**o+(-0.459054628604)*x[1]+(-0.983880547097)*x[2]**o+(-0.669749873181)*x[2]
        arg[2,2,0,1]=(0.126650330827)*x[0]**o+(0.375111974063)*x[0]+(-0.759797654917)*x[1]**o+(-0.663216091322)*x[1]+(-0.683191373782)*x[2]**o+(-0.607048449469)*x[2]
        arg[2,2,1,0]=(-0.805920040628)*x[0]**o+(-0.0144685877057)*x[0]+(0.00875222056291)*x[1]**o+(-0.110000267929)*x[1]+(-0.478201664476)*x[2]**o+(-0.0537439708493)*x[2]
        arg[2,2,1,1]=(-0.550624517803)*x[0]**o+(0.570407090431)*x[0]+(0.661026826148)*x[1]**o+(0.905189813242)*x[1]+(0.00707531661352)*x[2]**o+(0.711072343064)*x[2]
        arg[2,2,2,0]=(-0.644276976064)*x[0]**o+(-0.151665972531)*x[0]+(-0.645369927259)*x[1]**o+(0.915967227715)*x[1]+(-0.211750817665)*x[2]**o+(-0.596646212118)*x[2]
        arg[2,2,2,1]=(0.564688999107)*x[0]**o+(0.635094410913)*x[0]+(0.281863287848)*x[1]**o+(-0.927529659408)*x[1]+(0.581071087484)*x[2]**o+(0.164399277305)*x[2]
        arg[2,3,0,0]=(0.64754765253)*x[0]**o+(0.176693707964)*x[0]+(0.172801737347)*x[1]**o+(-0.769408746261)*x[1]+(-0.176026486601)*x[2]**o+(0.187304474822)*x[2]
        arg[2,3,0,1]=(0.365315558093)*x[0]**o+(0.888853264239)*x[0]+(0.0541759434362)*x[1]**o+(0.241489226711)*x[1]+(0.441822201832)*x[2]**o+(0.075541840076)*x[2]
        arg[2,3,1,0]=(-0.984940830356)*x[0]**o+(0.521596714942)*x[0]+(-0.933671223761)*x[1]**o+(-0.132907654205)*x[1]+(-0.225804727932)*x[2]**o+(0.987334445626)*x[2]
        arg[2,3,1,1]=(-0.785516613303)*x[0]**o+(0.0324126413814)*x[0]+(-0.084233004699)*x[1]**o+(0.45131442059)*x[1]+(0.36498759129)*x[2]**o+(0.417036982561)*x[2]
        arg[2,3,2,0]=(-0.310575473358)*x[0]**o+(-0.769219454218)*x[0]+(-0.263431454577)*x[1]**o+(0.474143448264)*x[1]+(-0.490633240671)*x[2]**o+(0.435609988849)*x[2]
        arg[2,3,2,1]=(0.230960030802)*x[0]**o+(-0.480593443732)*x[0]+(0.547888534735)*x[1]**o+(-0.818842709599)*x[1]+(-0.857141410253)*x[2]**o+(-0.287898487847)*x[2]
        arg[2,4,0,0]=(-0.811008715357)*x[0]**o+(-0.608505768579)*x[0]+(-0.117642307757)*x[1]**o+(-0.658229045784)*x[1]+(0.246542958193)*x[2]**o+(-0.42783248265)*x[2]
        arg[2,4,0,1]=(0.817994599948)*x[0]**o+(0.394931907893)*x[0]+(-0.273993145403)*x[1]**o+(0.9572397013)*x[1]+(0.101025833331)*x[2]**o+(-0.925874548613)*x[2]
        arg[2,4,1,0]=(-0.626152000269)*x[0]**o+(0.821322294557)*x[0]+(0.543761841392)*x[1]**o+(0.112341361443)*x[1]+(-0.395657376418)*x[2]**o+(0.815481187683)*x[2]
        arg[2,4,1,1]=(0.97134483493)*x[0]**o+(0.303341541402)*x[0]+(-0.270102237344)*x[1]**o+(-0.982452271352)*x[1]+(0.435568243011)*x[2]**o+(-0.200051496445)*x[2]
        arg[2,4,2,0]=(-0.827528679064)*x[0]**o+(-0.559447627475)*x[0]+(-0.867397899368)*x[1]**o+(0.915539157819)*x[1]+(0.370118757042)*x[2]**o+(0.578366179408)*x[2]
        arg[2,4,2,1]=(0.685882920875)*x[0]**o+(0.830016468585)*x[0]+(-0.603607199824)*x[1]**o+(0.888619939961)*x[1]+(-0.264173896988)*x[2]**o+(-0.588065135532)*x[2]
        arg[3,0,0,0]=(0.480000662344)*x[0]**o+(-0.240072007989)*x[0]+(0.252669294205)*x[1]**o+(-0.657466732916)*x[1]+(0.945149322572)*x[2]**o+(0.340592627149)*x[2]
        arg[3,0,0,1]=(-0.819956304014)*x[0]**o+(0.910408628621)*x[0]+(-0.875714560314)*x[1]**o+(0.653323198674)*x[1]+(0.910127896386)*x[2]**o+(0.21382952806)*x[2]
        arg[3,0,1,0]=(0.736655857249)*x[0]**o+(-0.982636511328)*x[0]+(0.747965385081)*x[1]**o+(-0.952056699676)*x[1]+(0.143168283422)*x[2]**o+(0.891386576579)*x[2]
        arg[3,0,1,1]=(0.615723453617)*x[0]**o+(0.57376968582)*x[0]+(0.766802940979)*x[1]**o+(-0.294849095727)*x[1]+(-0.26718123792)*x[2]**o+(-0.979175148343)*x[2]
        arg[3,0,2,0]=(-0.299975789693)*x[0]**o+(-0.941879795797)*x[0]+(-0.332633604206)*x[1]**o+(0.887589360228)*x[1]+(0.570321438438)*x[2]**o+(0.107180499249)*x[2]
        arg[3,0,2,1]=(-0.77479605873)*x[0]**o+(0.928359807986)*x[0]+(-0.144170026638)*x[1]**o+(0.533835316696)*x[1]+(0.183938136052)*x[2]**o+(-0.87336340483)*x[2]
        arg[3,1,0,0]=(-0.121858163593)*x[0]**o+(-0.971122306623)*x[0]+(-0.351891554168)*x[1]**o+(-0.989754547278)*x[1]+(0.48106592456)*x[2]**o+(0.750009716044)*x[2]
        arg[3,1,0,1]=(-0.332320017719)*x[0]**o+(0.229814969323)*x[0]+(0.950038106614)*x[1]**o+(-0.364221636516)*x[1]+(-0.842684764717)*x[2]**o+(-0.430003876831)*x[2]
        arg[3,1,1,0]=(0.1036461168)*x[0]**o+(0.243047542195)*x[0]+(-0.189881740385)*x[1]**o+(-0.646948186795)*x[1]+(-0.246523431838)*x[2]**o+(0.0363222140239)*x[2]
        arg[3,1,1,1]=(0.629747493968)*x[0]**o+(0.257110968013)*x[0]+(-0.594332699815)*x[1]**o+(-0.74041867705)*x[1]+(0.656583120031)*x[2]**o+(0.392776534802)*x[2]
        arg[3,1,2,0]=(0.997368307137)*x[0]**o+(-0.709004475665)*x[0]+(-0.379447862403)*x[1]**o+(-0.368700793823)*x[1]+(0.293673648471)*x[2]**o+(-0.555545691082)*x[2]
        arg[3,1,2,1]=(-0.323027033928)*x[0]**o+(0.804640557762)*x[0]+(-0.199434780533)*x[1]**o+(0.604194682736)*x[1]+(0.945516011113)*x[2]**o+(-0.831838977373)*x[2]
        arg[3,2,0,0]=(0.255071922458)*x[0]**o+(0.261075290259)*x[0]+(0.198792431961)*x[1]**o+(0.987681568369)*x[1]+(-0.257626079249)*x[2]**o+(0.817801578613)*x[2]
        arg[3,2,0,1]=(-0.732615916833)*x[0]**o+(0.8114347023)*x[0]+(0.308896981216)*x[1]**o+(-0.164981244979)*x[1]+(0.724411853266)*x[2]**o+(-0.647969863039)*x[2]
        arg[3,2,1,0]=(0.126755528121)*x[0]**o+(0.0203968097627)*x[0]+(0.350888385037)*x[1]**o+(0.765584878761)*x[1]+(0.43521215637)*x[2]**o+(0.5821821143)*x[2]
        arg[3,2,1,1]=(0.795956472753)*x[0]**o+(0.363631111688)*x[0]+(-0.296417527594)*x[1]**o+(-0.0927953540926)*x[1]+(0.396041397741)*x[2]**o+(0.43932000066)*x[2]
        arg[3,2,2,0]=(-0.926754341344)*x[0]**o+(-0.688414966522)*x[0]+(-0.660984101932)*x[1]**o+(-0.627883344982)*x[1]+(0.403036784646)*x[2]**o+(0.818277068314)*x[2]
        arg[3,2,2,1]=(-0.358921307079)*x[0]**o+(0.636847962491)*x[0]+(-0.510050460541)*x[1]**o+(-0.758190944228)*x[1]+(-0.0319326931965)*x[2]**o+(-0.643422203077)*x[2]
        arg[3,3,0,0]=(-0.930730482483)*x[0]**o+(-0.883991363505)*x[0]+(0.156751809769)*x[1]**o+(0.669949596594)*x[1]+(0.697840698407)*x[2]**o+(0.721622061073)*x[2]
        arg[3,3,0,1]=(0.108814498453)*x[0]**o+(-0.164024398834)*x[0]+(-0.172084798233)*x[1]**o+(0.031530303081)*x[1]+(-0.985663868487)*x[2]**o+(-0.290224642955)*x[2]
        arg[3,3,1,0]=(0.690575802875)*x[0]**o+(-0.720428203968)*x[0]+(-0.837111899789)*x[1]**o+(0.140256279088)*x[1]+(0.532813462778)*x[2]**o+(-0.557885762558)*x[2]
        arg[3,3,1,1]=(-0.0785388159589)*x[0]**o+(-0.40595445804)*x[0]+(-0.691255332383)*x[1]**o+(0.54108326284)*x[1]+(0.737684932212)*x[2]**o+(0.423508847197)*x[2]
        arg[3,3,2,0]=(-0.589838770269)*x[0]**o+(-0.557133995504)*x[0]+(-0.297626023793)*x[1]**o+(-0.511809000987)*x[1]+(-0.933566600332)*x[2]**o+(-0.21434993446)*x[2]
        arg[3,3,2,1]=(0.19892263441)*x[0]**o+(0.633952342142)*x[0]+(0.0179400729177)*x[1]**o+(0.154020947355)*x[1]+(-0.620690264897)*x[2]**o+(0.524853027077)*x[2]
        arg[3,4,0,0]=(0.475231682303)*x[0]**o+(-0.48174419167)*x[0]+(-0.100854189221)*x[1]**o+(0.818166518054)*x[1]+(-0.922566309657)*x[2]**o+(-0.498745068755)*x[2]
        arg[3,4,0,1]=(0.613982633884)*x[0]**o+(-0.947323459945)*x[0]+(0.17395204013)*x[1]**o+(0.929669087853)*x[1]+(-0.617278089466)*x[2]**o+(0.900537470392)*x[2]
        arg[3,4,1,0]=(-0.403633510324)*x[0]**o+(0.954097943555)*x[0]+(-0.463422043607)*x[1]**o+(0.843916487223)*x[1]+(-0.5955912138)*x[2]**o+(-0.765444657727)*x[2]
        arg[3,4,1,1]=(-0.663412886802)*x[0]**o+(-0.547440885077)*x[0]+(-0.479993570508)*x[1]**o+(0.0620303034445)*x[1]+(-0.653759917895)*x[2]**o+(-0.681004441743)*x[2]
        arg[3,4,2,0]=(-0.994529375243)*x[0]**o+(0.546983479745)*x[0]+(0.17913912435)*x[1]**o+(0.660366383627)*x[1]+(0.0747943517387)*x[2]**o+(-0.321063660452)*x[2]
        arg[3,4,2,1]=(-0.538074654323)*x[0]**o+(-0.81330034965)*x[0]+(0.660614896398)*x[1]**o+(-0.918192909648)*x[1]+(0.553635166581)*x[2]**o+(-0.918131802189)*x[2]
        ref[0,0,0,0]=(-0.321539504757)/(o+1.)+(0.762866286444)+(-0.793251694335)*0.5**o
        ref[0,0,0,1]=(-0.264812286159)/(o+1.)+(-0.15293893368)+(-0.624206604764)*0.5**o
        ref[0,0,1,0]=(0.54441377247)/(o+1.)+(0.214067163116)+(0.115999561481)*0.5**o
        ref[0,0,1,1]=(1.50036854956)/(o+1.)+(0.673819079573)+(-0.972466831671)*0.5**o
        ref[0,0,2,0]=(0.687874396856)/(o+1.)+(-0.986502459901)+(0.708180260094)*0.5**o
        ref[0,0,2,1]=(0.793527041633)/(o+1.)+(-0.259940301856)+(0.0282615573045)*0.5**o
        ref[0,1,0,0]=(-1.25436610505)/(o+1.)+(-0.224984055664)+(0.231621820235)*0.5**o
        ref[0,1,0,1]=(0.891022819882)/(o+1.)+(-0.165060941565)+(0.114056801155)*0.5**o
        ref[0,1,1,0]=(-0.833444470649)/(o+1.)+(-0.322644910601)+(-0.973115195017)*0.5**o
        ref[0,1,1,1]=(-0.189799709023)/(o+1.)+(-0.242650862468)+(-0.304758579369)*0.5**o
        ref[0,1,2,0]=(-1.39857111894)/(o+1.)+(-0.14133379247)+(0.0923844967157)*0.5**o
        ref[0,1,2,1]=(1.03702218206)/(o+1.)+(-0.181927579186)+(0.793568921835)*0.5**o
        ref[0,2,0,0]=(-1.11962316818)/(o+1.)+(0.654627255004)+(0.570673248048)*0.5**o
        ref[0,2,0,1]=(0.085812863531)/(o+1.)+(0.17639811125)+(0.299061155215)*0.5**o
        ref[0,2,1,0]=(-1.26466090832)/(o+1.)+(0.618506300342)+(0.694725244159)*0.5**o
        ref[0,2,1,1]=(0.181495495342)/(o+1.)+(0.818577989479)+(-0.83896527178)*0.5**o
        ref[0,2,2,0]=(0.867512484289)/(o+1.)+(0.353365700269)+(0.2228052704)*0.5**o
        ref[0,2,2,1]=(0.618350374002)/(o+1.)+(-0.418609375405)+(0.63141216243)*0.5**o
        ref[0,3,0,0]=(-0.801998626232)/(o+1.)+(-0.302573026989)+(0.00698107166168)*0.5**o
        ref[0,3,0,1]=(0.414118735868)/(o+1.)+(0.103407742435)+(0.243423742325)*0.5**o
        ref[0,3,1,0]=(-0.129735888575)/(o+1.)+(0.51412655284)+(-0.106629554393)*0.5**o
        ref[0,3,1,1]=(0.0886696480864)/(o+1.)+(0.287824302591)+(0.375143733111)*0.5**o
        ref[0,3,2,0]=(-0.746501172831)/(o+1.)+(-0.389518072046)+(0.955685455438)*0.5**o
        ref[0,3,2,1]=(-1.06628341794)/(o+1.)+(-0.147429422878)+(0.353654445481)*0.5**o
        ref[0,4,0,0]=(-0.174512962078)/(o+1.)+(-0.023500931215)+(0.974759684151)*0.5**o
        ref[0,4,0,1]=(1.18079051113)/(o+1.)+(0.980823290345)+(-0.0799922241539)*0.5**o
        ref[0,4,1,0]=(-1.19690096085)/(o+1.)+(0.173437389579)+(0.532633280629)*0.5**o
        ref[0,4,1,1]=(1.11279531673)/(o+1.)+(-0.763621970064)+(-0.353243658293)*0.5**o
        ref[0,4,2,0]=(-0.819102364613)/(o+1.)+(0.326676095071)+(-0.720840869801)*0.5**o
        ref[0,4,2,1]=(-0.397779896209)/(o+1.)+(-0.677806011795)+(-0.179073400214)*0.5**o
        ref[1,0,0,0]=(0.00277094815158)/(o+1.)+(0.301845054648)+(-0.384011553333)*0.5**o
        ref[1,0,0,1]=(1.22422728891)/(o+1.)+(-0.594910516315)+(-0.808211570649)*0.5**o
        ref[1,0,1,0]=(0.869946029655)/(o+1.)+(0.308501276384)+(0.0404274840294)*0.5**o
        ref[1,0,1,1]=(1.63718264119)/(o+1.)+(0.76013279253)+(0.302394139901)*0.5**o
        ref[1,0,2,0]=(-0.0317278448724)/(o+1.)+(-0.118813696855)+(-0.981981997649)*0.5**o
        ref[1,0,2,1]=(-1.13670569191)/(o+1.)+(-1.14181937354)+(-0.237479805943)*0.5**o
        ref[1,1,0,0]=(0.16579906355)/(o+1.)+(-0.303649903525)+(-0.543719871098)*0.5**o
        ref[1,1,0,1]=(-1.23792473519)/(o+1.)+(-0.288446030005)+(-0.2624625471)*0.5**o
        ref[1,1,1,0]=(-1.42019666807)/(o+1.)+(-0.23595016925)+(0.88003513115)*0.5**o
        ref[1,1,1,1]=(0.400318898446)/(o+1.)+(0.765385300641)+(0.177746615088)*0.5**o
        ref[1,1,2,0]=(1.21553520103)/(o+1.)+(-0.272140700356)+(0.889070121473)*0.5**o
        ref[1,1,2,1]=(0.121042131161)/(o+1.)+(0.672369993679)+(0.951736215954)*0.5**o
        ref[1,2,0,0]=(-1.33133223229)/(o+1.)+(0.0067424746325)+(-0.954701036263)*0.5**o
        ref[1,2,0,1]=(1.02875611862)/(o+1.)+(0.288339603595)+(0.256489282842)*0.5**o
        ref[1,2,1,0]=(-0.32722580116)/(o+1.)+(0.166447794803)+(-0.94015914517)*0.5**o
        ref[1,2,1,1]=(0.319551509271)/(o+1.)+(0.112957240802)+(0.367156421931)*0.5**o
        ref[1,2,2,0]=(-1.3530782297)/(o+1.)+(-0.383051052329)+(-0.812516694781)*0.5**o
        ref[1,2,2,1]=(0.103238894111)/(o+1.)+(-0.09078366319)+(-0.497461944186)*0.5**o
        ref[1,3,0,0]=(-0.730489039376)/(o+1.)+(-0.321235679733)+(-0.227077550639)*0.5**o
        ref[1,3,0,1]=(0.532708925151)/(o+1.)+(-0.0963112877562)+(0.822091021645)*0.5**o
        ref[1,3,1,0]=(1.69380361755)/(o+1.)+(-0.221589701137)+(-0.0341577310045)*0.5**o
        ref[1,3,1,1]=(-0.653530558759)/(o+1.)+(-0.322066868053)+(-0.102867224088)*0.5**o
        ref[1,3,2,0]=(-0.279916251859)/(o+1.)+(-0.18685188305)+(0.556301076488)*0.5**o
        ref[1,3,2,1]=(0.798107960854)/(o+1.)+(-0.326052832107)+(0.888436531887)*0.5**o
        ref[1,4,0,0]=(0.250537417885)/(o+1.)+(-0.257356294716)+(0.651898073685)*0.5**o
        ref[1,4,0,1]=(1.5757556609)/(o+1.)+(-0.669914057301)+(-0.149187282052)*0.5**o
        ref[1,4,1,0]=(-1.0484286711)/(o+1.)+(0.142400846148)+(-0.285938055108)*0.5**o
        ref[1,4,1,1]=(0.028101537891)/(o+1.)+(0.55179042751)+(-0.34015136472)*0.5**o
        ref[1,4,2,0]=(0.974779411801)/(o+1.)+(-0.498887411653)+(-0.464687865356)*0.5**o
        ref[1,4,2,1]=(-0.544927129174)/(o+1.)+(-0.322066331457)+(-0.777967458564)*0.5**o
        ref[2,0,0,0]=(1.102518379)/(o+1.)+(-0.456744121978)+(0.665321754689)*0.5**o
        ref[2,0,0,1]=(-1.23793122706)/(o+1.)+(-0.366112767163)+(0.876435018569)*0.5**o
        ref[2,0,1,0]=(0.793755949398)/(o+1.)+(-0.113798359287)+(-0.344552466761)*0.5**o
        ref[2,0,1,1]=(-0.430899036829)/(o+1.)+(-0.0136660119329)+(0.841069648591)*0.5**o
        ref[2,0,2,0]=(-1.53425877377)/(o+1.)+(0.149679134)+(0.192691357169)*0.5**o
        ref[2,0,2,1]=(-1.30260629932)/(o+1.)+(0.89308995205)+(0.556616365284)*0.5**o
        ref[2,1,0,0]=(0.95383115974)/(o+1.)+(-0.0662076555126)+(0.947750123445)*0.5**o
        ref[2,1,0,1]=(0.660457702465)/(o+1.)+(-0.0149888527952)+(0.962918564559)*0.5**o
        ref[2,1,1,0]=(-1.03853354453)/(o+1.)+(0.711309478418)+(-0.549909211973)*0.5**o
        ref[2,1,1,1]=(1.25726273721)/(o+1.)+(0.122502877857)+(0.382867749635)*0.5**o
        ref[2,1,2,0]=(0.247755301657)/(o+1.)+(-1.11689502868)+(0.355742024596)*0.5**o
        ref[2,1,2,1]=(-0.043927846695)/(o+1.)+(-0.806594322632)+(-0.882334243081)*0.5**o
        ref[2,2,0,0]=(-0.0978017670124)/(o+1.)+(-0.596620341454)+(0.372164678051)*0.5**o
        ref[2,2,0,1]=(-1.4429890287)/(o+1.)+(-0.447576283364)+(0.126650330827)*0.5**o
        ref[2,2,1,0]=(-0.469449443913)/(o+1.)+(-0.0891064132419)+(-0.805920040628)*0.5**o
        ref[2,2,1,1]=(0.668102142762)/(o+1.)+(1.09333462337)+(-0.550624517803)*0.5**o
        ref[2,2,2,0]=(-0.857120744924)/(o+1.)+(0.0838275215328)+(-0.644276976064)*0.5**o
        ref[2,2,2,1]=(0.862934375332)/(o+1.)+(-0.0640179855954)+(0.564688999107)*0.5**o
        ref[2,3,0,0]=(-0.00322474925369)/(o+1.)+(-0.202705281737)+(0.64754765253)*0.5**o
        ref[2,3,0,1]=(0.495998145268)/(o+1.)+(0.602942165513)+(0.365315558093)*0.5**o
        ref[2,3,1,0]=(-1.15947595169)/(o+1.)+(0.688011753181)+(-0.984940830356)*0.5**o
        ref[2,3,1,1]=(0.280754586591)/(o+1.)+(0.450382022266)+(-0.785516613303)*0.5**o
        ref[2,3,2,0]=(-0.754064695248)/(o+1.)+(0.0702669914474)+(-0.310575473358)*0.5**o
        ref[2,3,2,1]=(-0.309252875518)/(o+1.)+(-0.793667320589)+(0.230960030802)*0.5**o
        ref[2,4,0,0]=(0.128900650436)/(o+1.)+(-0.847283648506)+(-0.811008715357)*0.5**o
        ref[2,4,0,1]=(-0.172967312072)/(o+1.)+(0.21314853029)+(0.817994599948)*0.5**o
        ref[2,4,1,0]=(0.148104464974)/(o+1.)+(0.874572421841)+(-0.626152000269)*0.5**o
        ref[2,4,1,1]=(0.165466005667)/(o+1.)+(-0.439581113198)+(0.97134483493)*0.5**o
        ref[2,4,2,0]=(-0.497279142326)/(o+1.)+(0.467228854876)+(-0.827528679064)*0.5**o
        ref[2,4,2,1]=(-0.867781096812)/(o+1.)+(0.565285636507)+(0.685882920875)*0.5**o
        ref[3,0,0,0]=(1.19781861678)/(o+1.)+(-0.278473056878)+(0.480000662344)*0.5**o
        ref[3,0,0,1]=(0.0344133360711)/(o+1.)+(0.888780677678)+(-0.819956304014)*0.5**o
        ref[3,0,1,0]=(0.891133668503)/(o+1.)+(-0.521653317213)+(0.736655857249)*0.5**o
        ref[3,0,1,1]=(0.499621703059)/(o+1.)+(-0.350127279125)+(0.615723453617)*0.5**o
        ref[3,0,2,0]=(0.237687834232)/(o+1.)+(0.0264450318399)+(-0.299975789693)*0.5**o
        ref[3,0,2,1]=(0.0397681094138)/(o+1.)+(0.294415859926)+(-0.77479605873)*0.5**o
        ref[3,1,0,0]=(0.129174370391)/(o+1.)+(-0.605433568929)+(-0.121858163593)*0.5**o
        ref[3,1,0,1]=(0.107353341897)/(o+1.)+(-0.282205272012)+(-0.332320017719)*0.5**o
        ref[3,1,1,0]=(-0.436405172223)/(o+1.)+(-0.183789215288)+(0.1036461168)*0.5**o
        ref[3,1,1,1]=(0.0622504202157)/(o+1.)+(-0.0452655871179)+(0.629747493968)*0.5**o
        ref[3,1,2,0]=(-0.0857742139322)/(o+1.)+(-0.816625480285)+(0.997368307137)*0.5**o
        ref[3,1,2,1]=(0.746081230581)/(o+1.)+(0.288498131562)+(-0.323027033928)*0.5**o
        ref[3,2,0,0]=(-0.0588336472879)/(o+1.)+(1.03327921862)+(0.255071922458)*0.5**o
        ref[3,2,0,1]=(1.03330883448)/(o+1.)+(-0.000758202858916)+(-0.732615916833)*0.5**o
        ref[3,2,1,0]=(0.786100541406)/(o+1.)+(0.684081901412)+(0.126755528121)*0.5**o
        ref[3,2,1,1]=(0.099623870147)/(o+1.)+(0.355077879128)+(0.795956472753)*0.5**o
        ref[3,2,2,0]=(-0.257947317287)/(o+1.)+(-0.249010621595)+(-0.926754341344)*0.5**o
        ref[3,2,2,1]=(-0.541983153738)/(o+1.)+(-0.382382592407)+(-0.358921307079)*0.5**o
        ref[3,3,0,0]=(0.854592508176)/(o+1.)+(0.253790147081)+(-0.930730482483)*0.5**o
        ref[3,3,0,1]=(-1.15774866672)/(o+1.)+(-0.211359369354)+(0.108814498453)*0.5**o
        ref[3,3,1,0]=(-0.304298437011)/(o+1.)+(-0.569028843719)+(0.690575802875)*0.5**o
        ref[3,3,1,1]=(0.0464295998283)/(o+1.)+(0.279318825998)+(-0.0785388159589)*0.5**o
        ref[3,3,2,0]=(-1.23119262412)/(o+1.)+(-0.641646465476)+(-0.589838770269)*0.5**o
        ref[3,3,2,1]=(-0.60275019198)/(o+1.)+(0.656413158287)+(0.19892263441)*0.5**o
        ref[3,4,0,0]=(-1.02342049888)/(o+1.)+(-0.0811613711857)+(0.475231682303)*0.5**o
        ref[3,4,0,1]=(-0.443326049337)/(o+1.)+(0.44144154915)+(0.613982633884)*0.5**o
        ref[3,4,1,0]=(-1.05901325741)/(o+1.)+(0.516284886526)+(-0.403633510324)*0.5**o
        ref[3,4,1,1]=(-1.1337534884)/(o+1.)+(-0.583207511688)+(-0.663412886802)*0.5**o
        ref[3,4,2,0]=(0.253933476088)/(o+1.)+(0.44314310146)+(-0.994529375243)*0.5**o
        ref[3,4,2,1]=(1.21425006298)/(o+1.)+(-1.32481253074)+(-0.538074654323)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.441098788712)*x[0]**o+(0.082953252724)*x[0]+(-0.97005044508)*x[1]**o+(0.928239302707)*x[1]
        ref=(-0.97005044508)/(o+1.)+(0.505596277715)+(-0.441098788712)*0.5**o
      else:
        arg=(0.384995424392)*x[0]**o+(0.999298118235)*x[0]+(0.144745460948)*x[1]**o+(0.0545198589076)*x[1]+(-0.773375797864)*x[2]**o+(0.145524911495)*x[2]
        ref=(-0.628630336917)/(o+1.)+(0.599671444318)+(0.384995424392)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.0745889108767)*x[0]**o+(-0.694812761318)*x[0]+(-0.357791726239)*x[1]**o+(-0.785448350489)*x[1]
        arg[1]=(0.436006633056)*x[0]**o+(-0.797943632034)*x[0]+(-0.903321835716)*x[1]**o+(0.299527545727)*x[1]
        ref[0]=(-0.357791726239)/(o+1.)+(-0.740130555904)+(0.0745889108767)*0.5**o
        ref[1]=(-0.903321835716)/(o+1.)+(-0.249208043154)+(0.436006633056)*0.5**o
      else:
        arg[0]=(-0.292048194246)*x[0]**o+(-0.612402340241)*x[0]+(0.830253896674)*x[1]**o+(-0.0173245051245)*x[1]+(-0.105062444571)*x[2]**o+(0.968232377167)*x[2]
        arg[1]=(0.153118126553)*x[0]**o+(0.958628221601)*x[0]+(-0.326368437133)*x[1]**o+(0.878861879269)*x[1]+(-0.297398071379)*x[2]**o+(0.155202299471)*x[2]
        ref[0]=(0.725191452103)/(o+1.)+(0.169252765901)+(-0.292048194246)*0.5**o
        ref[1]=(-0.623766508512)/(o+1.)+(0.99634620017)+(0.153118126553)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.410346849722)*x[0]**o+(0.86102012198)*x[0]+(-0.0143604505833)*x[1]**o+(-0.333268586255)*x[1]
        arg[0,1]=(-0.104991947387)*x[0]**o+(0.51540973662)*x[0]+(0.449773354197)*x[1]**o+(0.0397236625379)*x[1]
        arg[0,2]=(0.0010671797269)*x[0]**o+(0.165202831822)*x[0]+(-0.536444395178)*x[1]**o+(0.142112903729)*x[1]
        arg[0,3]=(0.0655247081801)*x[0]**o+(-0.59093644572)*x[0]+(-0.0135713515051)*x[1]**o+(0.665578431202)*x[1]
        arg[0,4]=(0.902382714765)*x[0]**o+(-0.358716290301)*x[0]+(-0.504792876433)*x[1]**o+(-0.760982991404)*x[1]
        arg[1,0]=(0.663815334483)*x[0]**o+(-0.967978720711)*x[0]+(-0.308953852146)*x[1]**o+(-0.250315651366)*x[1]
        arg[1,1]=(0.602976841368)*x[0]**o+(0.407346455361)*x[0]+(0.820076160142)*x[1]**o+(0.720380314572)*x[1]
        arg[1,2]=(0.0899286747557)*x[0]**o+(-0.636748133402)*x[0]+(0.615853620374)*x[1]**o+(0.389136375831)*x[1]
        arg[1,3]=(-0.695859848989)*x[0]**o+(-0.160892907203)*x[0]+(0.972252211687)*x[1]**o+(0.971653054313)*x[1]
        arg[1,4]=(-0.229141368314)*x[0]**o+(0.434757135281)*x[0]+(0.508751791725)*x[1]**o+(-0.593022987916)*x[1]
        arg[2,0]=(-0.732297666821)*x[0]**o+(-0.885623150041)*x[0]+(0.355264312146)*x[1]**o+(-0.203582600498)*x[1]
        arg[2,1]=(-0.0683706126862)*x[0]**o+(0.390953181924)*x[0]+(-0.104115827873)*x[1]**o+(-0.0586349400277)*x[1]
        arg[2,2]=(0.213052201915)*x[0]**o+(0.813758731408)*x[0]+(-0.927085393446)*x[1]**o+(-0.70884993627)*x[1]
        arg[2,3]=(-0.11792311156)*x[0]**o+(0.705053056049)*x[0]+(-0.988858028527)*x[1]**o+(-0.532493233644)*x[1]
        arg[2,4]=(0.643258167584)*x[0]**o+(0.991059864292)*x[0]+(0.466734669253)*x[1]**o+(-0.763209421816)*x[1]
        arg[3,0]=(-0.686036752646)*x[0]**o+(-0.0152278415398)*x[0]+(0.572818298467)*x[1]**o+(0.0883817883272)*x[1]
        arg[3,1]=(-0.498088836117)*x[0]**o+(0.413958831603)*x[0]+(-0.788338642889)*x[1]**o+(-0.219412111863)*x[1]
        arg[3,2]=(0.895103303091)*x[0]**o+(0.967224972282)*x[0]+(0.778129303031)*x[1]**o+(-0.239067231978)*x[1]
        arg[3,3]=(-0.715385920659)*x[0]**o+(-0.928446013585)*x[0]+(-0.919578735328)*x[1]**o+(0.735438220506)*x[1]
        arg[3,4]=(0.273825880641)*x[0]**o+(0.602304312788)*x[0]+(-0.788415866411)*x[1]**o+(0.841776933557)*x[1]
        ref[0,0]=(-0.0143604505833)/(o+1.)+(0.263875767862)+(0.410346849722)*0.5**o
        ref[0,1]=(0.449773354197)/(o+1.)+(0.277566699579)+(-0.104991947387)*0.5**o
        ref[0,2]=(-0.536444395178)/(o+1.)+(0.153657867776)+(0.0010671797269)*0.5**o
        ref[0,3]=(-0.0135713515051)/(o+1.)+(0.0373209927411)+(0.0655247081801)*0.5**o
        ref[0,4]=(-0.504792876433)/(o+1.)+(-0.559849640852)+(0.902382714765)*0.5**o
        ref[1,0]=(-0.308953852146)/(o+1.)+(-0.609147186038)+(0.663815334483)*0.5**o
        ref[1,1]=(0.820076160142)/(o+1.)+(0.563863384966)+(0.602976841368)*0.5**o
        ref[1,2]=(0.615853620374)/(o+1.)+(-0.123805878786)+(0.0899286747557)*0.5**o
        ref[1,3]=(0.972252211687)/(o+1.)+(0.405380073555)+(-0.695859848989)*0.5**o
        ref[1,4]=(0.508751791725)/(o+1.)+(-0.0791329263176)+(-0.229141368314)*0.5**o
        ref[2,0]=(0.355264312146)/(o+1.)+(-0.544602875269)+(-0.732297666821)*0.5**o
        ref[2,1]=(-0.104115827873)/(o+1.)+(0.166159120948)+(-0.0683706126862)*0.5**o
        ref[2,2]=(-0.927085393446)/(o+1.)+(0.0524543975689)+(0.213052201915)*0.5**o
        ref[2,3]=(-0.988858028527)/(o+1.)+(0.0862799112023)+(-0.11792311156)*0.5**o
        ref[2,4]=(0.466734669253)/(o+1.)+(0.113925221238)+(0.643258167584)*0.5**o
        ref[3,0]=(0.572818298467)/(o+1.)+(0.0365769733937)+(-0.686036752646)*0.5**o
        ref[3,1]=(-0.788338642889)/(o+1.)+(0.0972733598697)+(-0.498088836117)*0.5**o
        ref[3,2]=(0.778129303031)/(o+1.)+(0.364078870152)+(0.895103303091)*0.5**o
        ref[3,3]=(-0.919578735328)/(o+1.)+(-0.0965038965398)+(-0.715385920659)*0.5**o
        ref[3,4]=(-0.788415866411)/(o+1.)+(0.722040623173)+(0.273825880641)*0.5**o
      else:
        arg[0,0]=(-0.903197979191)*x[0]**o+(-0.497252911882)*x[0]+(0.5689188153)*x[1]**o+(-0.569434759407)*x[1]+(-0.204533481399)*x[2]**o+(-0.0255207088467)*x[2]
        arg[0,1]=(-0.0296606374337)*x[0]**o+(0.917552577919)*x[0]+(0.746132694062)*x[1]**o+(0.285568252958)*x[1]+(0.175571603104)*x[2]**o+(0.5701931124)*x[2]
        arg[0,2]=(0.657981464039)*x[0]**o+(-0.105816675859)*x[0]+(-0.210726712386)*x[1]**o+(-0.457303699577)*x[1]+(-0.225517453595)*x[2]**o+(0.61593257413)*x[2]
        arg[0,3]=(-0.652766009035)*x[0]**o+(0.538908603782)*x[0]+(0.226594150058)*x[1]**o+(0.691428293374)*x[1]+(-0.210896072508)*x[2]**o+(0.538061905948)*x[2]
        arg[0,4]=(0.175024193446)*x[0]**o+(0.433285241007)*x[0]+(0.928411521986)*x[1]**o+(-0.241175862661)*x[1]+(-0.779192552542)*x[2]**o+(0.428100531631)*x[2]
        arg[1,0]=(-0.094230207815)*x[0]**o+(-0.113255272088)*x[0]+(-0.880816596679)*x[1]**o+(0.56261116141)*x[1]+(-0.418606336511)*x[2]**o+(0.858058623759)*x[2]
        arg[1,1]=(0.08556853498)*x[0]**o+(0.95491440456)*x[0]+(-0.435275492231)*x[1]**o+(0.339105115812)*x[1]+(-0.634101556975)*x[2]**o+(0.357557194572)*x[2]
        arg[1,2]=(-0.692883812524)*x[0]**o+(0.673540357384)*x[0]+(0.240219465837)*x[1]**o+(0.139957925862)*x[1]+(0.135892430596)*x[2]**o+(0.383465818668)*x[2]
        arg[1,3]=(-0.892921935206)*x[0]**o+(0.0694647682287)*x[0]+(0.0530888770821)*x[1]**o+(-0.0735398003251)*x[1]+(-0.40702392212)*x[2]**o+(-0.223494881098)*x[2]
        arg[1,4]=(0.743112585337)*x[0]**o+(-0.12393012942)*x[0]+(-0.660616664197)*x[1]**o+(-0.0334489536139)*x[1]+(-0.0566394013558)*x[2]**o+(-0.0281273872158)*x[2]
        arg[2,0]=(-0.500500871517)*x[0]**o+(-0.326237506809)*x[0]+(-0.960480146386)*x[1]**o+(0.0901654083154)*x[1]+(0.466124845857)*x[2]**o+(0.6908848619)*x[2]
        arg[2,1]=(0.120233846989)*x[0]**o+(0.520609850629)*x[0]+(-0.407423400749)*x[1]**o+(0.59145151128)*x[1]+(0.981171407459)*x[2]**o+(-0.703987165108)*x[2]
        arg[2,2]=(0.811775583161)*x[0]**o+(-0.209650983138)*x[0]+(-0.286223486261)*x[1]**o+(-0.864122954179)*x[1]+(0.609725467363)*x[2]**o+(0.0311464898903)*x[2]
        arg[2,3]=(-0.325647787579)*x[0]**o+(0.956001955119)*x[0]+(0.637272546129)*x[1]**o+(-0.951109798103)*x[1]+(0.406445558466)*x[2]**o+(-0.930854597943)*x[2]
        arg[2,4]=(0.0621315789778)*x[0]**o+(-0.917033227802)*x[0]+(-0.179584904974)*x[1]**o+(0.283357518809)*x[1]+(-0.778863909832)*x[2]**o+(-0.524497331838)*x[2]
        arg[3,0]=(0.140605486894)*x[0]**o+(0.172882849759)*x[0]+(-0.883586756712)*x[1]**o+(-0.160461884399)*x[1]+(-0.718343713337)*x[2]**o+(0.196937721256)*x[2]
        arg[3,1]=(-0.660232764173)*x[0]**o+(-0.92091042631)*x[0]+(-0.568852134661)*x[1]**o+(-0.331441330509)*x[1]+(-0.220286922375)*x[2]**o+(-0.256407439105)*x[2]
        arg[3,2]=(0.593940468197)*x[0]**o+(0.527756437117)*x[0]+(0.943980933807)*x[1]**o+(0.749415167795)*x[1]+(0.484043332105)*x[2]**o+(0.40135030381)*x[2]
        arg[3,3]=(0.870508483705)*x[0]**o+(-0.234209640456)*x[0]+(-0.62692310882)*x[1]**o+(0.352628981854)*x[1]+(-0.24937844453)*x[2]**o+(0.73438735864)*x[2]
        arg[3,4]=(-0.638437449353)*x[0]**o+(0.308007229542)*x[0]+(0.537590844528)*x[1]**o+(-0.755595388361)*x[1]+(-0.152781228745)*x[2]**o+(-0.450122095362)*x[2]
        ref[0,0]=(0.364385333901)/(o+1.)+(-0.546104190068)+(-0.903197979191)*0.5**o
        ref[0,1]=(0.921704297167)/(o+1.)+(0.886656971638)+(-0.0296606374337)*0.5**o
        ref[0,2]=(-0.436244165981)/(o+1.)+(0.0264060993469)+(0.657981464039)*0.5**o
        ref[0,3]=(0.0156980775503)/(o+1.)+(0.884199401552)+(-0.652766009035)*0.5**o
        ref[0,4]=(0.149218969444)/(o+1.)+(0.310104954989)+(0.175024193446)*0.5**o
        ref[1,0]=(-1.29942293319)/(o+1.)+(0.653707256541)+(-0.094230207815)*0.5**o
        ref[1,1]=(-1.06937704921)/(o+1.)+(0.825788357473)+(0.08556853498)*0.5**o
        ref[1,2]=(0.376111896433)/(o+1.)+(0.598482050957)+(-0.692883812524)*0.5**o
        ref[1,3]=(-0.353935045038)/(o+1.)+(-0.113784956597)+(-0.892921935206)*0.5**o
        ref[1,4]=(-0.717256065552)/(o+1.)+(-0.0927532351247)+(0.743112585337)*0.5**o
        ref[2,0]=(-0.494355300528)/(o+1.)+(0.227406381703)+(-0.500500871517)*0.5**o
        ref[2,1]=(0.57374800671)/(o+1.)+(0.2040370984)+(0.120233846989)*0.5**o
        ref[2,2]=(0.323501981101)/(o+1.)+(-0.521313723714)+(0.811775583161)*0.5**o
        ref[2,3]=(1.04371810459)/(o+1.)+(-0.462981220464)+(-0.325647787579)*0.5**o
        ref[2,4]=(-0.958448814807)/(o+1.)+(-0.579086520416)+(0.0621315789778)*0.5**o
        ref[3,0]=(-1.60193047005)/(o+1.)+(0.104679343308)+(0.140605486894)*0.5**o
        ref[3,1]=(-0.789139057035)/(o+1.)+(-0.754379597962)+(-0.660232764173)*0.5**o
        ref[3,2]=(1.42802426591)/(o+1.)+(0.839260954361)+(0.593940468197)*0.5**o
        ref[3,3]=(-0.87630155335)/(o+1.)+(0.426403350019)+(0.870508483705)*0.5**o
        ref[3,4]=(0.384809615783)/(o+1.)+(-0.448855127091)+(-0.638437449353)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.315756069525)*x[0]**o+(0.253240552004)*x[0]+(0.86540335823)*x[1]**o+(-0.616203780065)*x[1]
        arg[0,0,1]=(-0.948116617977)*x[0]**o+(0.892154336439)*x[0]+(0.136605864823)*x[1]**o+(-0.135599413854)*x[1]
        arg[0,1,0]=(0.249388669107)*x[0]**o+(-0.299043026617)*x[0]+(0.335963868826)*x[1]**o+(-0.147202749474)*x[1]
        arg[0,1,1]=(0.649003124445)*x[0]**o+(0.102344972673)*x[0]+(-0.795038258171)*x[1]**o+(0.149716983842)*x[1]
        arg[1,0,0]=(0.814606374941)*x[0]**o+(-0.949251236145)*x[0]+(0.723727126173)*x[1]**o+(-0.75419822708)*x[1]
        arg[1,0,1]=(0.043251358201)*x[0]**o+(-0.227748374318)*x[0]+(-0.231556019421)*x[1]**o+(-0.152524439909)*x[1]
        arg[1,1,0]=(0.565989211326)*x[0]**o+(-0.488248650798)*x[0]+(-0.353077780419)*x[1]**o+(-0.0139123365766)*x[1]
        arg[1,1,1]=(-0.106085432662)*x[0]**o+(0.627311916563)*x[0]+(-0.496301561324)*x[1]**o+(-0.648448849796)*x[1]
        arg[2,0,0]=(0.741868098354)*x[0]**o+(-0.486898566862)*x[0]+(0.0461464851414)*x[1]**o+(0.0133916949979)*x[1]
        arg[2,0,1]=(0.969108066213)*x[0]**o+(-0.688797899381)*x[0]+(0.897017408844)*x[1]**o+(0.807714585761)*x[1]
        arg[2,1,0]=(-0.138159486947)*x[0]**o+(-0.377759480876)*x[0]+(-0.764484355987)*x[1]**o+(-0.723912265361)*x[1]
        arg[2,1,1]=(-0.781167132076)*x[0]**o+(0.119512137054)*x[0]+(-0.535230596139)*x[1]**o+(0.441358968417)*x[1]
        arg[3,0,0]=(-0.663820836909)*x[0]**o+(0.907668749761)*x[0]+(-0.345114213135)*x[1]**o+(0.0580058218243)*x[1]
        arg[3,0,1]=(-0.359054701663)*x[0]**o+(0.0389188839076)*x[0]+(0.168050801441)*x[1]**o+(-0.695770024565)*x[1]
        arg[3,1,0]=(0.70118269628)*x[0]**o+(0.87976625188)*x[0]+(0.154503286758)*x[1]**o+(0.807878896951)*x[1]
        arg[3,1,1]=(-0.143516320927)*x[0]**o+(-0.369452129865)*x[0]+(-0.0392268449697)*x[1]**o+(-0.046104227955)*x[1]
        arg[4,0,0]=(0.197244254754)*x[0]**o+(-0.580444991466)*x[0]+(0.334036744845)*x[1]**o+(0.000791083620504)*x[1]
        arg[4,0,1]=(-0.892626226865)*x[0]**o+(0.140042741589)*x[0]+(0.158775259845)*x[1]**o+(-0.308266331028)*x[1]
        arg[4,1,0]=(0.650374542633)*x[0]**o+(0.586152174865)*x[0]+(0.0201324360301)*x[1]**o+(-0.813221777119)*x[1]
        arg[4,1,1]=(0.149656968314)*x[0]**o+(0.700641349434)*x[0]+(-0.529780404433)*x[1]**o+(0.512159137234)*x[1]
        arg[5,0,0]=(0.288664621864)*x[0]**o+(0.119652133593)*x[0]+(-0.599113573241)*x[1]**o+(0.879966724309)*x[1]
        arg[5,0,1]=(0.861443356689)*x[0]**o+(0.530607121787)*x[0]+(0.240306513701)*x[1]**o+(-0.575772515225)*x[1]
        arg[5,1,0]=(0.721635890008)*x[0]**o+(-0.109537583449)*x[0]+(0.162041613112)*x[1]**o+(0.229143033467)*x[1]
        arg[5,1,1]=(-0.876930820494)*x[0]**o+(0.749378811252)*x[0]+(0.886213554632)*x[1]**o+(0.433327797349)*x[1]
        ref[0,0,0]=(0.86540335823)/(o+1.)+(-0.18148161403)+(-0.315756069525)*0.5**o
        ref[0,0,1]=(0.136605864823)/(o+1.)+(0.378277461292)+(-0.948116617977)*0.5**o
        ref[0,1,0]=(0.335963868826)/(o+1.)+(-0.223122888045)+(0.249388669107)*0.5**o
        ref[0,1,1]=(-0.795038258171)/(o+1.)+(0.126030978258)+(0.649003124445)*0.5**o
        ref[1,0,0]=(0.723727126173)/(o+1.)+(-0.851724731613)+(0.814606374941)*0.5**o
        ref[1,0,1]=(-0.231556019421)/(o+1.)+(-0.190136407114)+(0.043251358201)*0.5**o
        ref[1,1,0]=(-0.353077780419)/(o+1.)+(-0.251080493687)+(0.565989211326)*0.5**o
        ref[1,1,1]=(-0.496301561324)/(o+1.)+(-0.0105684666165)+(-0.106085432662)*0.5**o
        ref[2,0,0]=(0.0461464851414)/(o+1.)+(-0.236753435932)+(0.741868098354)*0.5**o
        ref[2,0,1]=(0.897017408844)/(o+1.)+(0.05945834319)+(0.969108066213)*0.5**o
        ref[2,1,0]=(-0.764484355987)/(o+1.)+(-0.550835873118)+(-0.138159486947)*0.5**o
        ref[2,1,1]=(-0.535230596139)/(o+1.)+(0.280435552736)+(-0.781167132076)*0.5**o
        ref[3,0,0]=(-0.345114213135)/(o+1.)+(0.482837285793)+(-0.663820836909)*0.5**o
        ref[3,0,1]=(0.168050801441)/(o+1.)+(-0.328425570329)+(-0.359054701663)*0.5**o
        ref[3,1,0]=(0.154503286758)/(o+1.)+(0.843822574415)+(0.70118269628)*0.5**o
        ref[3,1,1]=(-0.0392268449697)/(o+1.)+(-0.20777817891)+(-0.143516320927)*0.5**o
        ref[4,0,0]=(0.334036744845)/(o+1.)+(-0.289826953923)+(0.197244254754)*0.5**o
        ref[4,0,1]=(0.158775259845)/(o+1.)+(-0.0841117947196)+(-0.892626226865)*0.5**o
        ref[4,1,0]=(0.0201324360301)/(o+1.)+(-0.113534801127)+(0.650374542633)*0.5**o
        ref[4,1,1]=(-0.529780404433)/(o+1.)+(0.606400243334)+(0.149656968314)*0.5**o
        ref[5,0,0]=(-0.599113573241)/(o+1.)+(0.499809428951)+(0.288664621864)*0.5**o
        ref[5,0,1]=(0.240306513701)/(o+1.)+(-0.022582696719)+(0.861443356689)*0.5**o
        ref[5,1,0]=(0.162041613112)/(o+1.)+(0.0598027250089)+(0.721635890008)*0.5**o
        ref[5,1,1]=(0.886213554632)/(o+1.)+(0.5913533043)+(-0.876930820494)*0.5**o
      else:
        arg[0,0,0]=(0.804252705823)*x[0]**o+(-0.215722921442)*x[0]+(-0.806316349529)*x[1]**o+(-0.984002993222)*x[1]+(-0.411946014731)*x[2]**o+(-0.526356191855)*x[2]
        arg[0,0,1]=(-0.845902068304)*x[0]**o+(0.0773568471048)*x[0]+(-0.746877301439)*x[1]**o+(-0.581914187059)*x[1]+(-0.0299873869697)*x[2]**o+(-0.133702230427)*x[2]
        arg[0,1,0]=(0.0113663840954)*x[0]**o+(-0.155702215771)*x[0]+(-0.134490626112)*x[1]**o+(0.973329917011)*x[1]+(-0.71585624135)*x[2]**o+(-0.840113198085)*x[2]
        arg[0,1,1]=(0.708589516593)*x[0]**o+(0.40282025998)*x[0]+(0.114498353117)*x[1]**o+(0.117244230433)*x[1]+(0.110135318853)*x[2]**o+(-0.0227879044898)*x[2]
        arg[1,0,0]=(0.836102144172)*x[0]**o+(0.212564417583)*x[0]+(0.744354739679)*x[1]**o+(-0.265633978841)*x[1]+(0.843108381264)*x[2]**o+(-0.288428427501)*x[2]
        arg[1,0,1]=(-0.0712209756059)*x[0]**o+(-0.186158140814)*x[0]+(0.184457099888)*x[1]**o+(0.381365494077)*x[1]+(-0.547874738232)*x[2]**o+(0.645124013254)*x[2]
        arg[1,1,0]=(-0.982992052158)*x[0]**o+(0.259627201342)*x[0]+(-0.885159598031)*x[1]**o+(-0.720754022982)*x[1]+(-0.762444117798)*x[2]**o+(0.737661938668)*x[2]
        arg[1,1,1]=(0.518426662725)*x[0]**o+(0.718331452733)*x[0]+(0.523870992899)*x[1]**o+(0.527831420124)*x[1]+(0.791634534316)*x[2]**o+(0.17320870773)*x[2]
        arg[2,0,0]=(-0.0368697940364)*x[0]**o+(-0.101718169515)*x[0]+(0.933712377315)*x[1]**o+(-0.465193151914)*x[1]+(-0.113637506449)*x[2]**o+(0.606619949802)*x[2]
        arg[2,0,1]=(-0.785633726304)*x[0]**o+(0.526518355588)*x[0]+(0.151859749675)*x[1]**o+(0.805094094642)*x[1]+(-0.881122119524)*x[2]**o+(0.85143600659)*x[2]
        arg[2,1,0]=(0.969244001518)*x[0]**o+(0.22600349917)*x[0]+(0.404826692149)*x[1]**o+(0.633039564539)*x[1]+(0.470567669552)*x[2]**o+(0.378348877016)*x[2]
        arg[2,1,1]=(-0.310035929524)*x[0]**o+(0.42632271287)*x[0]+(0.866930472029)*x[1]**o+(-0.124483457934)*x[1]+(0.291920406068)*x[2]**o+(0.573984377154)*x[2]
        arg[3,0,0]=(0.245553971829)*x[0]**o+(-0.567563766526)*x[0]+(0.326654705159)*x[1]**o+(0.642370867777)*x[1]+(-0.538450732704)*x[2]**o+(0.773349470647)*x[2]
        arg[3,0,1]=(0.427017913395)*x[0]**o+(0.885132021286)*x[0]+(0.0338747045378)*x[1]**o+(-0.978508839789)*x[1]+(-0.734559666249)*x[2]**o+(0.379991690191)*x[2]
        arg[3,1,0]=(-0.29533334088)*x[0]**o+(-0.894548834889)*x[0]+(0.537043269081)*x[1]**o+(-0.559173506199)*x[1]+(0.359216600837)*x[2]**o+(0.269563745611)*x[2]
        arg[3,1,1]=(-0.146152544907)*x[0]**o+(-0.95681567795)*x[0]+(-0.0323978120064)*x[1]**o+(-0.5351716012)*x[1]+(-0.548536561192)*x[2]**o+(0.489675460511)*x[2]
        arg[4,0,0]=(0.0552991386013)*x[0]**o+(0.772417012152)*x[0]+(-0.298212559744)*x[1]**o+(-0.189934984078)*x[1]+(0.74843296731)*x[2]**o+(-0.00412448772016)*x[2]
        arg[4,0,1]=(0.577046858044)*x[0]**o+(-0.804142676861)*x[0]+(-0.0703211154568)*x[1]**o+(-0.84092953544)*x[1]+(0.254119640557)*x[2]**o+(-0.765048199333)*x[2]
        arg[4,1,0]=(-0.358558179011)*x[0]**o+(-0.998072158283)*x[0]+(0.508152998513)*x[1]**o+(-0.0813296609683)*x[1]+(-0.534332816004)*x[2]**o+(-0.189623317277)*x[2]
        arg[4,1,1]=(0.840098089823)*x[0]**o+(0.966726062679)*x[0]+(0.920079609258)*x[1]**o+(-0.242532131681)*x[1]+(0.429713768025)*x[2]**o+(0.171223902751)*x[2]
        arg[5,0,0]=(0.695331559401)*x[0]**o+(0.819482211381)*x[0]+(0.550452400412)*x[1]**o+(0.518496496336)*x[1]+(-0.24727850431)*x[2]**o+(-0.796406493488)*x[2]
        arg[5,0,1]=(0.469488028534)*x[0]**o+(0.457530124732)*x[0]+(0.922206279456)*x[1]**o+(0.678758573006)*x[1]+(0.34102098746)*x[2]**o+(0.108616278675)*x[2]
        arg[5,1,0]=(0.444799933589)*x[0]**o+(-0.361568786934)*x[0]+(0.470174557338)*x[1]**o+(-0.344557214342)*x[1]+(0.894860467988)*x[2]**o+(-0.347027743579)*x[2]
        arg[5,1,1]=(0.0304781332872)*x[0]**o+(-0.69046475182)*x[0]+(0.873260975232)*x[1]**o+(-0.705100555993)*x[1]+(0.660057279013)*x[2]**o+(-0.910002146105)*x[2]
        ref[0,0,0]=(-1.21826236426)/(o+1.)+(-0.863041053259)+(0.804252705823)*0.5**o
        ref[0,0,1]=(-0.776864688409)/(o+1.)+(-0.31912978519)+(-0.845902068304)*0.5**o
        ref[0,1,0]=(-0.850346867462)/(o+1.)+(-0.0112427484226)+(0.0113663840954)*0.5**o
        ref[0,1,1]=(0.22463367197)/(o+1.)+(0.248638292962)+(0.708589516593)*0.5**o
        ref[1,0,0]=(1.58746312094)/(o+1.)+(-0.17074899438)+(0.836102144172)*0.5**o
        ref[1,0,1]=(-0.363417638344)/(o+1.)+(0.420165683258)+(-0.0712209756059)*0.5**o
        ref[1,1,0]=(-1.64760371583)/(o+1.)+(0.138267558514)+(-0.982992052158)*0.5**o
        ref[1,1,1]=(1.31550552722)/(o+1.)+(0.709685790293)+(0.518426662725)*0.5**o
        ref[2,0,0]=(0.820074870865)/(o+1.)+(0.0198543141864)+(-0.0368697940364)*0.5**o
        ref[2,0,1]=(-0.729262369848)/(o+1.)+(1.09152422841)+(-0.785633726304)*0.5**o
        ref[2,1,0]=(0.875394361701)/(o+1.)+(0.618695970363)+(0.969244001518)*0.5**o
        ref[2,1,1]=(1.1588508781)/(o+1.)+(0.437911816045)+(-0.310035929524)*0.5**o
        ref[3,0,0]=(-0.211796027546)/(o+1.)+(0.424078285949)+(0.245553971829)*0.5**o
        ref[3,0,1]=(-0.700684961711)/(o+1.)+(0.143307435844)+(0.427017913395)*0.5**o
        ref[3,1,0]=(0.896259869918)/(o+1.)+(-0.592079297738)+(-0.29533334088)*0.5**o
        ref[3,1,1]=(-0.580934373198)/(o+1.)+(-0.50115590932)+(-0.146152544907)*0.5**o
        ref[4,0,0]=(0.450220407566)/(o+1.)+(0.289178770177)+(0.0552991386013)*0.5**o
        ref[4,0,1]=(0.1837985251)/(o+1.)+(-1.20506020582)+(0.577046858044)*0.5**o
        ref[4,1,0]=(-0.0261798174917)/(o+1.)+(-0.634512568264)+(-0.358558179011)*0.5**o
        ref[4,1,1]=(1.34979337728)/(o+1.)+(0.447708916874)+(0.840098089823)*0.5**o
        ref[5,0,0]=(0.303173896103)/(o+1.)+(0.270786107114)+(0.695331559401)*0.5**o
        ref[5,0,1]=(1.26322726692)/(o+1.)+(0.622452488207)+(0.469488028534)*0.5**o
        ref[5,1,0]=(1.36503502533)/(o+1.)+(-0.526576872427)+(0.444799933589)*0.5**o
        ref[5,1,1]=(1.53331825424)/(o+1.)+(-1.15278372696)+(0.0304781332872)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactOne

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.348972036919)*x[0]**o+(0.123542094911)*x[0]+(-0.596353540657)*x[1]**o+(0.717656729405)*x[1]
        arg[0,0,0,1]=(0.187733014896)*x[0]**o+(0.76244697815)*x[0]+(0.261409664458)*x[1]**o+(0.0213714031642)*x[1]
        arg[0,0,1,0]=(0.951655197513)*x[0]**o+(-0.865274104487)*x[0]+(0.59127040848)*x[1]**o+(-0.261622300136)*x[1]
        arg[0,0,1,1]=(-0.188127934304)*x[0]**o+(-0.513768204254)*x[0]+(0.77162116582)*x[1]**o+(-0.12621176914)*x[1]
        arg[0,0,2,0]=(0.770717242337)*x[0]**o+(-0.991038362841)*x[0]+(0.908566805373)*x[1]**o+(-0.24890064672)*x[1]
        arg[0,0,2,1]=(-0.543253359807)*x[0]**o+(-0.0504511921259)*x[0]+(-0.724584621389)*x[1]**o+(0.999965312141)*x[1]
        arg[0,1,0,0]=(0.986939140609)*x[0]**o+(0.236894127675)*x[0]+(0.520162431169)*x[1]**o+(0.306413415759)*x[1]
        arg[0,1,0,1]=(-0.171773212929)*x[0]**o+(-0.248436968201)*x[0]+(0.43818378458)*x[1]**o+(-0.495734982435)*x[1]
        arg[0,1,1,0]=(-0.369350989684)*x[0]**o+(0.59047357584)*x[0]+(0.422340945702)*x[1]**o+(0.326270009775)*x[1]
        arg[0,1,1,1]=(0.559606389467)*x[0]**o+(-0.0507064379501)*x[0]+(-0.0464190048363)*x[1]**o+(0.18964879235)*x[1]
        arg[0,1,2,0]=(-0.989408371198)*x[0]**o+(0.108750148268)*x[0]+(-0.347984444559)*x[1]**o+(-0.819590699683)*x[1]
        arg[0,1,2,1]=(-0.289295900777)*x[0]**o+(-0.579698633174)*x[0]+(-0.124928198101)*x[1]**o+(0.640901698875)*x[1]
        arg[0,2,0,0]=(0.132599667548)*x[0]**o+(-0.102095433147)*x[0]+(0.446169236302)*x[1]**o+(-0.844062525718)*x[1]
        arg[0,2,0,1]=(-0.452662917924)*x[0]**o+(-0.0497770418153)*x[0]+(-0.258525519083)*x[1]**o+(-0.0365197858914)*x[1]
        arg[0,2,1,0]=(0.318740871592)*x[0]**o+(0.562764515639)*x[0]+(-0.209089266173)*x[1]**o+(0.922024900771)*x[1]
        arg[0,2,1,1]=(-0.825757440563)*x[0]**o+(0.634009249217)*x[0]+(0.954836398565)*x[1]**o+(-0.910652766097)*x[1]
        arg[0,2,2,0]=(-0.640337007803)*x[0]**o+(0.862625623816)*x[0]+(-0.0327529254483)*x[1]**o+(-0.327305638603)*x[1]
        arg[0,2,2,1]=(-0.956300293885)*x[0]**o+(0.456116376134)*x[0]+(0.312023159058)*x[1]**o+(0.768546792276)*x[1]
        arg[0,3,0,0]=(-0.927040285049)*x[0]**o+(-0.614994799603)*x[0]+(-0.431859510779)*x[1]**o+(-0.088774748513)*x[1]
        arg[0,3,0,1]=(-0.807009561768)*x[0]**o+(0.00323409463398)*x[0]+(0.274814268019)*x[1]**o+(-0.132755038246)*x[1]
        arg[0,3,1,0]=(0.931774139846)*x[0]**o+(0.665327108449)*x[0]+(0.130437344004)*x[1]**o+(-0.00859025546518)*x[1]
        arg[0,3,1,1]=(-0.314188008746)*x[0]**o+(0.405251729437)*x[0]+(-0.313484462262)*x[1]**o+(-0.670765905491)*x[1]
        arg[0,3,2,0]=(0.748691505108)*x[0]**o+(0.717110963353)*x[0]+(-0.451892315269)*x[1]**o+(-0.103930621637)*x[1]
        arg[0,3,2,1]=(-0.124290384657)*x[0]**o+(0.0761550925153)*x[0]+(0.864166710803)*x[1]**o+(0.050297196468)*x[1]
        arg[0,4,0,0]=(-0.726606985714)*x[0]**o+(-0.304862755503)*x[0]+(-0.364735884875)*x[1]**o+(0.769999145518)*x[1]
        arg[0,4,0,1]=(-0.716457484159)*x[0]**o+(0.850185561874)*x[0]+(0.924535166786)*x[1]**o+(0.414539526778)*x[1]
        arg[0,4,1,0]=(-0.67154700238)*x[0]**o+(-0.0990079744678)*x[0]+(-0.738258048173)*x[1]**o+(-0.105333364414)*x[1]
        arg[0,4,1,1]=(-0.568569022142)*x[0]**o+(-0.865340476096)*x[0]+(0.592223896873)*x[1]**o+(0.545216501911)*x[1]
        arg[0,4,2,0]=(0.959619955514)*x[0]**o+(-0.625097978453)*x[0]+(0.447944032533)*x[1]**o+(-0.792670959618)*x[1]
        arg[0,4,2,1]=(-0.446578795272)*x[0]**o+(0.516341492325)*x[0]+(0.383143400712)*x[1]**o+(-0.339255681885)*x[1]
        arg[1,0,0,0]=(0.130600745253)*x[0]**o+(-0.225084866063)*x[0]+(-0.871963625443)*x[1]**o+(-0.0741251944735)*x[1]
        arg[1,0,0,1]=(-0.684630134211)*x[0]**o+(-0.298097256903)*x[0]+(0.511206122216)*x[1]**o+(-0.784104136535)*x[1]
        arg[1,0,1,0]=(0.624178362847)*x[0]**o+(0.161787148072)*x[0]+(-0.0906465041298)*x[1]**o+(-0.0302149132362)*x[1]
        arg[1,0,1,1]=(-0.403988173215)*x[0]**o+(-0.423895624882)*x[0]+(-0.0594261878161)*x[1]**o+(0.71753539369)*x[1]
        arg[1,0,2,0]=(0.411689797339)*x[0]**o+(0.0555768292842)*x[0]+(0.786614145981)*x[1]**o+(-0.701418967065)*x[1]
        arg[1,0,2,1]=(0.971933572797)*x[0]**o+(0.203165120814)*x[0]+(0.50340688009)*x[1]**o+(-0.443869607992)*x[1]
        arg[1,1,0,0]=(0.517782412737)*x[0]**o+(-0.233071827503)*x[0]+(0.310115588175)*x[1]**o+(-0.970604318783)*x[1]
        arg[1,1,0,1]=(0.250379154253)*x[0]**o+(-0.0914513321258)*x[0]+(0.466403827523)*x[1]**o+(-0.454369386353)*x[1]
        arg[1,1,1,0]=(-0.235520035004)*x[0]**o+(0.0932661387959)*x[0]+(0.0964933173918)*x[1]**o+(0.585245824454)*x[1]
        arg[1,1,1,1]=(-0.466145341257)*x[0]**o+(0.532910678544)*x[0]+(0.757545184326)*x[1]**o+(-0.488241176428)*x[1]
        arg[1,1,2,0]=(0.802070127079)*x[0]**o+(0.30227826497)*x[0]+(0.76818718588)*x[1]**o+(-0.966380392422)*x[1]
        arg[1,1,2,1]=(0.764411747219)*x[0]**o+(0.62151105703)*x[0]+(0.310875960956)*x[1]**o+(0.218417169467)*x[1]
        arg[1,2,0,0]=(0.627865841486)*x[0]**o+(0.328844171432)*x[0]+(0.47405718918)*x[1]**o+(0.238471462812)*x[1]
        arg[1,2,0,1]=(-0.890402471541)*x[0]**o+(0.0654874441479)*x[0]+(0.411199516291)*x[1]**o+(0.0669990208336)*x[1]
        arg[1,2,1,0]=(-0.789401901979)*x[0]**o+(-0.379725885915)*x[0]+(-0.103817842125)*x[1]**o+(0.358055283478)*x[1]
        arg[1,2,1,1]=(-0.619688531158)*x[0]**o+(0.314022911152)*x[0]+(-0.515963904961)*x[1]**o+(0.807912272743)*x[1]
        arg[1,2,2,0]=(0.696584946679)*x[0]**o+(-0.156405028983)*x[0]+(0.99374085906)*x[1]**o+(0.692362331558)*x[1]
        arg[1,2,2,1]=(0.463393460998)*x[0]**o+(-0.0302433372264)*x[0]+(0.243006695756)*x[1]**o+(0.934578685307)*x[1]
        arg[1,3,0,0]=(-0.760149691585)*x[0]**o+(0.305312987392)*x[0]+(0.478665319824)*x[1]**o+(-0.395734016727)*x[1]
        arg[1,3,0,1]=(-0.495638457166)*x[0]**o+(-0.352868042523)*x[0]+(0.0577132713041)*x[1]**o+(0.856507718332)*x[1]
        arg[1,3,1,0]=(-0.588160041725)*x[0]**o+(-0.941242199161)*x[0]+(-0.883563055172)*x[1]**o+(-0.474349426459)*x[1]
        arg[1,3,1,1]=(-0.822752260982)*x[0]**o+(0.284881595545)*x[0]+(-0.914198363943)*x[1]**o+(0.421107500666)*x[1]
        arg[1,3,2,0]=(-0.125099230328)*x[0]**o+(-0.0167977306674)*x[0]+(-0.880147366581)*x[1]**o+(0.947650728684)*x[1]
        arg[1,3,2,1]=(-0.770895899992)*x[0]**o+(-0.121705997003)*x[0]+(-0.0486858273347)*x[1]**o+(0.666179109672)*x[1]
        arg[1,4,0,0]=(-0.767476534476)*x[0]**o+(0.720013119195)*x[0]+(-0.540220631624)*x[1]**o+(-0.363702385125)*x[1]
        arg[1,4,0,1]=(0.90195509915)*x[0]**o+(0.104569298687)*x[0]+(0.901816813435)*x[1]**o+(0.421060929159)*x[1]
        arg[1,4,1,0]=(0.0901631681262)*x[0]**o+(-0.330570875574)*x[0]+(0.892147778885)*x[1]**o+(-0.102976747188)*x[1]
        arg[1,4,1,1]=(-0.336016437275)*x[0]**o+(-0.00913798206221)*x[0]+(-0.20211674106)*x[1]**o+(-0.0470039081275)*x[1]
        arg[1,4,2,0]=(0.194383418491)*x[0]**o+(-0.913993330791)*x[0]+(-0.0207457431142)*x[1]**o+(-0.728130242724)*x[1]
        arg[1,4,2,1]=(-0.285983741049)*x[0]**o+(-0.359824200302)*x[0]+(-0.286913656927)*x[1]**o+(-0.801477464894)*x[1]
        arg[2,0,0,0]=(-0.266085322677)*x[0]**o+(-0.477036323188)*x[0]+(0.967304544659)*x[1]**o+(-0.843721203165)*x[1]
        arg[2,0,0,1]=(-0.431692878995)*x[0]**o+(-0.326793076488)*x[0]+(0.478220893643)*x[1]**o+(-0.953607426852)*x[1]
        arg[2,0,1,0]=(0.800398295672)*x[0]**o+(-0.726923243756)*x[0]+(0.641194491232)*x[1]**o+(0.933755071699)*x[1]
        arg[2,0,1,1]=(0.778158607269)*x[0]**o+(0.192895772287)*x[0]+(0.160540751687)*x[1]**o+(0.132780133987)*x[1]
        arg[2,0,2,0]=(0.0700487728725)*x[0]**o+(0.366571342769)*x[0]+(0.213123464402)*x[1]**o+(0.0224970803568)*x[1]
        arg[2,0,2,1]=(-0.608321696505)*x[0]**o+(0.416761589908)*x[0]+(0.0499280394381)*x[1]**o+(-0.678150075911)*x[1]
        arg[2,1,0,0]=(-0.580987690899)*x[0]**o+(-0.625951863984)*x[0]+(0.0064094419355)*x[1]**o+(0.646045286912)*x[1]
        arg[2,1,0,1]=(-0.881255703221)*x[0]**o+(-0.153796152409)*x[0]+(0.320566497467)*x[1]**o+(-0.217827916224)*x[1]
        arg[2,1,1,0]=(0.981593086009)*x[0]**o+(0.659643450249)*x[0]+(0.824530802391)*x[1]**o+(0.109125926037)*x[1]
        arg[2,1,1,1]=(-0.79695617371)*x[0]**o+(-0.340030091407)*x[0]+(0.476648191338)*x[1]**o+(-0.751961375342)*x[1]
        arg[2,1,2,0]=(0.00785807621726)*x[0]**o+(-0.177037931946)*x[0]+(-0.322685883101)*x[1]**o+(0.639485466195)*x[1]
        arg[2,1,2,1]=(0.386667953213)*x[0]**o+(0.805115373435)*x[0]+(0.489529148574)*x[1]**o+(0.380650996708)*x[1]
        arg[2,2,0,0]=(-0.205678281473)*x[0]**o+(0.725798252716)*x[0]+(-0.884620261631)*x[1]**o+(0.737678769923)*x[1]
        arg[2,2,0,1]=(-0.859841848792)*x[0]**o+(0.215497396508)*x[0]+(-0.9231710768)*x[1]**o+(-0.504914651252)*x[1]
        arg[2,2,1,0]=(-0.0161889255216)*x[0]**o+(0.218150095097)*x[0]+(0.508505799895)*x[1]**o+(0.441270271663)*x[1]
        arg[2,2,1,1]=(0.645494826531)*x[0]**o+(0.812462584734)*x[0]+(0.498639261012)*x[1]**o+(0.695415642432)*x[1]
        arg[2,2,2,0]=(-0.855342734108)*x[0]**o+(0.174928383231)*x[0]+(-0.550636347213)*x[1]**o+(0.814295216919)*x[1]
        arg[2,2,2,1]=(-0.146152468476)*x[0]**o+(-0.810876776267)*x[0]+(0.77444775812)*x[1]**o+(0.758171250375)*x[1]
        arg[2,3,0,0]=(0.914771099535)*x[0]**o+(-0.801103658802)*x[0]+(-0.339761119169)*x[1]**o+(-0.0697409372498)*x[1]
        arg[2,3,0,1]=(-0.278017621273)*x[0]**o+(-0.916896227964)*x[0]+(0.0151205004002)*x[1]**o+(-0.36424189479)*x[1]
        arg[2,3,1,0]=(0.635845368235)*x[0]**o+(-0.286420165946)*x[0]+(-0.547314386502)*x[1]**o+(-0.297882680065)*x[1]
        arg[2,3,1,1]=(0.236860892332)*x[0]**o+(0.670289851551)*x[0]+(-0.437288046493)*x[1]**o+(0.0610684872034)*x[1]
        arg[2,3,2,0]=(0.096479176127)*x[0]**o+(-0.992608241155)*x[0]+(0.622650933171)*x[1]**o+(-0.171594052732)*x[1]
        arg[2,3,2,1]=(0.259913661654)*x[0]**o+(0.521353257934)*x[0]+(0.141212812345)*x[1]**o+(0.0311971708742)*x[1]
        arg[2,4,0,0]=(0.208498140469)*x[0]**o+(0.914285328223)*x[0]+(-0.368860713612)*x[1]**o+(0.744057640483)*x[1]
        arg[2,4,0,1]=(-0.0760701135919)*x[0]**o+(-0.0666410245297)*x[0]+(-0.442858585142)*x[1]**o+(-0.869273907698)*x[1]
        arg[2,4,1,0]=(0.694291134331)*x[0]**o+(0.153487646161)*x[0]+(-0.454726570824)*x[1]**o+(0.12533511793)*x[1]
        arg[2,4,1,1]=(0.505643124198)*x[0]**o+(0.935208927517)*x[0]+(0.120271053976)*x[1]**o+(0.472499209985)*x[1]
        arg[2,4,2,0]=(0.458890639035)*x[0]**o+(-0.38125264637)*x[0]+(-0.742525095759)*x[1]**o+(0.0727341692272)*x[1]
        arg[2,4,2,1]=(-0.228856512525)*x[0]**o+(0.215540806099)*x[0]+(-0.492076254398)*x[1]**o+(0.325121281149)*x[1]
        arg[3,0,0,0]=(0.791320435019)*x[0]**o+(0.264208069532)*x[0]+(0.105218141263)*x[1]**o+(-0.947073121631)*x[1]
        arg[3,0,0,1]=(0.384081306848)*x[0]**o+(-0.735420806741)*x[0]+(0.695451719819)*x[1]**o+(-0.104071285853)*x[1]
        arg[3,0,1,0]=(0.176280431402)*x[0]**o+(0.766248516714)*x[0]+(0.620920827635)*x[1]**o+(0.571778731591)*x[1]
        arg[3,0,1,1]=(0.895148060637)*x[0]**o+(0.120401904357)*x[0]+(0.500131556768)*x[1]**o+(-0.432523916076)*x[1]
        arg[3,0,2,0]=(-0.171375065793)*x[0]**o+(0.731777853738)*x[0]+(0.914774635339)*x[1]**o+(-0.329504614878)*x[1]
        arg[3,0,2,1]=(-0.65523706909)*x[0]**o+(-0.309487164889)*x[0]+(-0.244763521442)*x[1]**o+(0.790744116827)*x[1]
        arg[3,1,0,0]=(-0.515742288)*x[0]**o+(0.688021032329)*x[0]+(0.744558730681)*x[1]**o+(0.663410706435)*x[1]
        arg[3,1,0,1]=(0.299920493853)*x[0]**o+(0.221554378219)*x[0]+(-0.262111371211)*x[1]**o+(0.597219370531)*x[1]
        arg[3,1,1,0]=(-0.491648532317)*x[0]**o+(-0.0763921400417)*x[0]+(0.517994656474)*x[1]**o+(0.690649210498)*x[1]
        arg[3,1,1,1]=(0.291881729319)*x[0]**o+(0.388296683726)*x[0]+(0.304223657331)*x[1]**o+(-0.17249216188)*x[1]
        arg[3,1,2,0]=(0.309288981861)*x[0]**o+(-0.638172982762)*x[0]+(-0.566768084047)*x[1]**o+(-0.694439437371)*x[1]
        arg[3,1,2,1]=(-0.680391436031)*x[0]**o+(0.876096306554)*x[0]+(0.676979754874)*x[1]**o+(0.376507786465)*x[1]
        arg[3,2,0,0]=(-0.688464200232)*x[0]**o+(0.404140014906)*x[0]+(-0.389359624725)*x[1]**o+(0.218257423404)*x[1]
        arg[3,2,0,1]=(-0.321661287857)*x[0]**o+(0.305707842777)*x[0]+(0.0488067596906)*x[1]**o+(0.475071042874)*x[1]
        arg[3,2,1,0]=(-0.172172832131)*x[0]**o+(-0.0589598637728)*x[0]+(0.997578239775)*x[1]**o+(0.0919614884141)*x[1]
        arg[3,2,1,1]=(0.500590025777)*x[0]**o+(0.258724052781)*x[0]+(0.233819575609)*x[1]**o+(0.787949345995)*x[1]
        arg[3,2,2,0]=(-0.464931198743)*x[0]**o+(0.689246395117)*x[0]+(0.0701366334357)*x[1]**o+(-0.206391031385)*x[1]
        arg[3,2,2,1]=(-0.352492545609)*x[0]**o+(-0.160385273073)*x[0]+(0.876791526094)*x[1]**o+(0.748257988817)*x[1]
        arg[3,3,0,0]=(0.70346530112)*x[0]**o+(-0.351192123239)*x[0]+(-0.985298342079)*x[1]**o+(-0.286817451723)*x[1]
        arg[3,3,0,1]=(0.841285056577)*x[0]**o+(0.741986440026)*x[0]+(-0.758645627646)*x[1]**o+(-0.914494381783)*x[1]
        arg[3,3,1,0]=(-0.614293213484)*x[0]**o+(-0.746555667296)*x[0]+(-0.301781252786)*x[1]**o+(-0.721633195357)*x[1]
        arg[3,3,1,1]=(-0.287510519978)*x[0]**o+(-0.764960517935)*x[0]+(-0.948032248565)*x[1]**o+(-0.548676002552)*x[1]
        arg[3,3,2,0]=(0.145506943729)*x[0]**o+(-0.353925293653)*x[0]+(0.786904887352)*x[1]**o+(-0.36548133551)*x[1]
        arg[3,3,2,1]=(0.60782215735)*x[0]**o+(0.872444227513)*x[0]+(-0.237571902274)*x[1]**o+(-0.0559745367747)*x[1]
        arg[3,4,0,0]=(0.691107698337)*x[0]**o+(0.0800532750053)*x[0]+(-0.847456183316)*x[1]**o+(0.911068576288)*x[1]
        arg[3,4,0,1]=(-0.00167606578449)*x[0]**o+(-0.697603805697)*x[0]+(-0.362440886816)*x[1]**o+(-0.0604379772766)*x[1]
        arg[3,4,1,0]=(-0.74026325403)*x[0]**o+(-0.539076677004)*x[0]+(-0.317176913637)*x[1]**o+(0.759776288772)*x[1]
        arg[3,4,1,1]=(-0.230005454291)*x[0]**o+(-0.929557352119)*x[0]+(-0.435125210116)*x[1]**o+(0.106892135893)*x[1]
        arg[3,4,2,0]=(-0.638149620463)*x[0]**o+(-0.0511424128986)*x[0]+(0.460436795546)*x[1]**o+(-0.919535518181)*x[1]
        arg[3,4,2,1]=(-0.793537245531)*x[0]**o+(0.409768940347)*x[0]+(0.632560678557)*x[1]**o+(0.700596370851)*x[1]
        ref[0,0,0,0]=(-0.596353540657)/(o+1.)+(0.420599412158)+(-0.348972036919)*0.5**o
        ref[0,0,0,1]=(0.261409664458)/(o+1.)+(0.391909190657)+(0.187733014896)*0.5**o
        ref[0,0,1,0]=(0.59127040848)/(o+1.)+(-0.563448202311)+(0.951655197513)*0.5**o
        ref[0,0,1,1]=(0.77162116582)/(o+1.)+(-0.319989986697)+(-0.188127934304)*0.5**o
        ref[0,0,2,0]=(0.908566805373)/(o+1.)+(-0.619969504781)+(0.770717242337)*0.5**o
        ref[0,0,2,1]=(-0.724584621389)/(o+1.)+(0.474757060008)+(-0.543253359807)*0.5**o
        ref[0,1,0,0]=(0.520162431169)/(o+1.)+(0.271653771717)+(0.986939140609)*0.5**o
        ref[0,1,0,1]=(0.43818378458)/(o+1.)+(-0.372085975318)+(-0.171773212929)*0.5**o
        ref[0,1,1,0]=(0.422340945702)/(o+1.)+(0.458371792807)+(-0.369350989684)*0.5**o
        ref[0,1,1,1]=(-0.0464190048363)/(o+1.)+(0.0694711771999)+(0.559606389467)*0.5**o
        ref[0,1,2,0]=(-0.347984444559)/(o+1.)+(-0.355420275708)+(-0.989408371198)*0.5**o
        ref[0,1,2,1]=(-0.124928198101)/(o+1.)+(0.0306015328506)+(-0.289295900777)*0.5**o
        ref[0,2,0,0]=(0.446169236302)/(o+1.)+(-0.473078979432)+(0.132599667548)*0.5**o
        ref[0,2,0,1]=(-0.258525519083)/(o+1.)+(-0.0431484138534)+(-0.452662917924)*0.5**o
        ref[0,2,1,0]=(-0.209089266173)/(o+1.)+(0.742394708205)+(0.318740871592)*0.5**o
        ref[0,2,1,1]=(0.954836398565)/(o+1.)+(-0.13832175844)+(-0.825757440563)*0.5**o
        ref[0,2,2,0]=(-0.0327529254483)/(o+1.)+(0.267659992607)+(-0.640337007803)*0.5**o
        ref[0,2,2,1]=(0.312023159058)/(o+1.)+(0.612331584205)+(-0.956300293885)*0.5**o
        ref[0,3,0,0]=(-0.431859510779)/(o+1.)+(-0.351884774058)+(-0.927040285049)*0.5**o
        ref[0,3,0,1]=(0.274814268019)/(o+1.)+(-0.064760471806)+(-0.807009561768)*0.5**o
        ref[0,3,1,0]=(0.130437344004)/(o+1.)+(0.328368426492)+(0.931774139846)*0.5**o
        ref[0,3,1,1]=(-0.313484462262)/(o+1.)+(-0.132757088027)+(-0.314188008746)*0.5**o
        ref[0,3,2,0]=(-0.451892315269)/(o+1.)+(0.306590170858)+(0.748691505108)*0.5**o
        ref[0,3,2,1]=(0.864166710803)/(o+1.)+(0.0632261444916)+(-0.124290384657)*0.5**o
        ref[0,4,0,0]=(-0.364735884875)/(o+1.)+(0.232568195007)+(-0.726606985714)*0.5**o
        ref[0,4,0,1]=(0.924535166786)/(o+1.)+(0.632362544326)+(-0.716457484159)*0.5**o
        ref[0,4,1,0]=(-0.738258048173)/(o+1.)+(-0.102170669441)+(-0.67154700238)*0.5**o
        ref[0,4,1,1]=(0.592223896873)/(o+1.)+(-0.160061987092)+(-0.568569022142)*0.5**o
        ref[0,4,2,0]=(0.447944032533)/(o+1.)+(-0.708884469036)+(0.959619955514)*0.5**o
        ref[0,4,2,1]=(0.383143400712)/(o+1.)+(0.0885429052198)+(-0.446578795272)*0.5**o
        ref[1,0,0,0]=(-0.871963625443)/(o+1.)+(-0.149605030268)+(0.130600745253)*0.5**o
        ref[1,0,0,1]=(0.511206122216)/(o+1.)+(-0.541100696719)+(-0.684630134211)*0.5**o
        ref[1,0,1,0]=(-0.0906465041298)/(o+1.)+(0.065786117418)+(0.624178362847)*0.5**o
        ref[1,0,1,1]=(-0.0594261878161)/(o+1.)+(0.146819884404)+(-0.403988173215)*0.5**o
        ref[1,0,2,0]=(0.786614145981)/(o+1.)+(-0.32292106889)+(0.411689797339)*0.5**o
        ref[1,0,2,1]=(0.50340688009)/(o+1.)+(-0.120352243589)+(0.971933572797)*0.5**o
        ref[1,1,0,0]=(0.310115588175)/(o+1.)+(-0.601838073143)+(0.517782412737)*0.5**o
        ref[1,1,0,1]=(0.466403827523)/(o+1.)+(-0.27291035924)+(0.250379154253)*0.5**o
        ref[1,1,1,0]=(0.0964933173918)/(o+1.)+(0.339255981625)+(-0.235520035004)*0.5**o
        ref[1,1,1,1]=(0.757545184326)/(o+1.)+(0.0223347510579)+(-0.466145341257)*0.5**o
        ref[1,1,2,0]=(0.76818718588)/(o+1.)+(-0.332051063726)+(0.802070127079)*0.5**o
        ref[1,1,2,1]=(0.310875960956)/(o+1.)+(0.419964113249)+(0.764411747219)*0.5**o
        ref[1,2,0,0]=(0.47405718918)/(o+1.)+(0.283657817122)+(0.627865841486)*0.5**o
        ref[1,2,0,1]=(0.411199516291)/(o+1.)+(0.0662432324908)+(-0.890402471541)*0.5**o
        ref[1,2,1,0]=(-0.103817842125)/(o+1.)+(-0.0108353012186)+(-0.789401901979)*0.5**o
        ref[1,2,1,1]=(-0.515963904961)/(o+1.)+(0.560967591948)+(-0.619688531158)*0.5**o
        ref[1,2,2,0]=(0.99374085906)/(o+1.)+(0.267978651288)+(0.696584946679)*0.5**o
        ref[1,2,2,1]=(0.243006695756)/(o+1.)+(0.45216767404)+(0.463393460998)*0.5**o
        ref[1,3,0,0]=(0.478665319824)/(o+1.)+(-0.0452105146678)+(-0.760149691585)*0.5**o
        ref[1,3,0,1]=(0.0577132713041)/(o+1.)+(0.251819837904)+(-0.495638457166)*0.5**o
        ref[1,3,1,0]=(-0.883563055172)/(o+1.)+(-0.70779581281)+(-0.588160041725)*0.5**o
        ref[1,3,1,1]=(-0.914198363943)/(o+1.)+(0.352994548105)+(-0.822752260982)*0.5**o
        ref[1,3,2,0]=(-0.880147366581)/(o+1.)+(0.465426499008)+(-0.125099230328)*0.5**o
        ref[1,3,2,1]=(-0.0486858273347)/(o+1.)+(0.272236556335)+(-0.770895899992)*0.5**o
        ref[1,4,0,0]=(-0.540220631624)/(o+1.)+(0.178155367035)+(-0.767476534476)*0.5**o
        ref[1,4,0,1]=(0.901816813435)/(o+1.)+(0.262815113923)+(0.90195509915)*0.5**o
        ref[1,4,1,0]=(0.892147778885)/(o+1.)+(-0.216773811381)+(0.0901631681262)*0.5**o
        ref[1,4,1,1]=(-0.20211674106)/(o+1.)+(-0.0280709450949)+(-0.336016437275)*0.5**o
        ref[1,4,2,0]=(-0.0207457431142)/(o+1.)+(-0.821061786757)+(0.194383418491)*0.5**o
        ref[1,4,2,1]=(-0.286913656927)/(o+1.)+(-0.580650832598)+(-0.285983741049)*0.5**o
        ref[2,0,0,0]=(0.967304544659)/(o+1.)+(-0.660378763176)+(-0.266085322677)*0.5**o
        ref[2,0,0,1]=(0.478220893643)/(o+1.)+(-0.64020025167)+(-0.431692878995)*0.5**o
        ref[2,0,1,0]=(0.641194491232)/(o+1.)+(0.103415913972)+(0.800398295672)*0.5**o
        ref[2,0,1,1]=(0.160540751687)/(o+1.)+(0.162837953137)+(0.778158607269)*0.5**o
        ref[2,0,2,0]=(0.213123464402)/(o+1.)+(0.194534211563)+(0.0700487728725)*0.5**o
        ref[2,0,2,1]=(0.0499280394381)/(o+1.)+(-0.130694243001)+(-0.608321696505)*0.5**o
        ref[2,1,0,0]=(0.0064094419355)/(o+1.)+(0.0100467114639)+(-0.580987690899)*0.5**o
        ref[2,1,0,1]=(0.320566497467)/(o+1.)+(-0.185812034316)+(-0.881255703221)*0.5**o
        ref[2,1,1,0]=(0.824530802391)/(o+1.)+(0.384384688143)+(0.981593086009)*0.5**o
        ref[2,1,1,1]=(0.476648191338)/(o+1.)+(-0.545995733375)+(-0.79695617371)*0.5**o
        ref[2,1,2,0]=(-0.322685883101)/(o+1.)+(0.231223767125)+(0.00785807621726)*0.5**o
        ref[2,1,2,1]=(0.489529148574)/(o+1.)+(0.592883185071)+(0.386667953213)*0.5**o
        ref[2,2,0,0]=(-0.884620261631)/(o+1.)+(0.73173851132)+(-0.205678281473)*0.5**o
        ref[2,2,0,1]=(-0.9231710768)/(o+1.)+(-0.144708627372)+(-0.859841848792)*0.5**o
        ref[2,2,1,0]=(0.508505799895)/(o+1.)+(0.32971018338)+(-0.0161889255216)*0.5**o
        ref[2,2,1,1]=(0.498639261012)/(o+1.)+(0.753939113583)+(0.645494826531)*0.5**o
        ref[2,2,2,0]=(-0.550636347213)/(o+1.)+(0.494611800075)+(-0.855342734108)*0.5**o
        ref[2,2,2,1]=(0.77444775812)/(o+1.)+(-0.0263527629459)+(-0.146152468476)*0.5**o
        ref[2,3,0,0]=(-0.339761119169)/(o+1.)+(-0.435422298026)+(0.914771099535)*0.5**o
        ref[2,3,0,1]=(0.0151205004002)/(o+1.)+(-0.640569061377)+(-0.278017621273)*0.5**o
        ref[2,3,1,0]=(-0.547314386502)/(o+1.)+(-0.292151423006)+(0.635845368235)*0.5**o
        ref[2,3,1,1]=(-0.437288046493)/(o+1.)+(0.365679169377)+(0.236860892332)*0.5**o
        ref[2,3,2,0]=(0.622650933171)/(o+1.)+(-0.582101146943)+(0.096479176127)*0.5**o
        ref[2,3,2,1]=(0.141212812345)/(o+1.)+(0.276275214404)+(0.259913661654)*0.5**o
        ref[2,4,0,0]=(-0.368860713612)/(o+1.)+(0.829171484353)+(0.208498140469)*0.5**o
        ref[2,4,0,1]=(-0.442858585142)/(o+1.)+(-0.467957466114)+(-0.0760701135919)*0.5**o
        ref[2,4,1,0]=(-0.454726570824)/(o+1.)+(0.139411382046)+(0.694291134331)*0.5**o
        ref[2,4,1,1]=(0.120271053976)/(o+1.)+(0.703854068751)+(0.505643124198)*0.5**o
        ref[2,4,2,0]=(-0.742525095759)/(o+1.)+(-0.154259238572)+(0.458890639035)*0.5**o
        ref[2,4,2,1]=(-0.492076254398)/(o+1.)+(0.270331043624)+(-0.228856512525)*0.5**o
        ref[3,0,0,0]=(0.105218141263)/(o+1.)+(-0.34143252605)+(0.791320435019)*0.5**o
        ref[3,0,0,1]=(0.695451719819)/(o+1.)+(-0.419746046297)+(0.384081306848)*0.5**o
        ref[3,0,1,0]=(0.620920827635)/(o+1.)+(0.669013624153)+(0.176280431402)*0.5**o
        ref[3,0,1,1]=(0.500131556768)/(o+1.)+(-0.15606100586)+(0.895148060637)*0.5**o
        ref[3,0,2,0]=(0.914774635339)/(o+1.)+(0.20113661943)+(-0.171375065793)*0.5**o
        ref[3,0,2,1]=(-0.244763521442)/(o+1.)+(0.240628475969)+(-0.65523706909)*0.5**o
        ref[3,1,0,0]=(0.744558730681)/(o+1.)+(0.675715869382)+(-0.515742288)*0.5**o
        ref[3,1,0,1]=(-0.262111371211)/(o+1.)+(0.409386874375)+(0.299920493853)*0.5**o
        ref[3,1,1,0]=(0.517994656474)/(o+1.)+(0.307128535228)+(-0.491648532317)*0.5**o
        ref[3,1,1,1]=(0.304223657331)/(o+1.)+(0.107902260923)+(0.291881729319)*0.5**o
        ref[3,1,2,0]=(-0.566768084047)/(o+1.)+(-0.666306210066)+(0.309288981861)*0.5**o
        ref[3,1,2,1]=(0.676979754874)/(o+1.)+(0.62630204651)+(-0.680391436031)*0.5**o
        ref[3,2,0,0]=(-0.389359624725)/(o+1.)+(0.311198719155)+(-0.688464200232)*0.5**o
        ref[3,2,0,1]=(0.0488067596906)/(o+1.)+(0.390389442826)+(-0.321661287857)*0.5**o
        ref[3,2,1,0]=(0.997578239775)/(o+1.)+(0.0165008123207)+(-0.172172832131)*0.5**o
        ref[3,2,1,1]=(0.233819575609)/(o+1.)+(0.523336699388)+(0.500590025777)*0.5**o
        ref[3,2,2,0]=(0.0701366334357)/(o+1.)+(0.241427681866)+(-0.464931198743)*0.5**o
        ref[3,2,2,1]=(0.876791526094)/(o+1.)+(0.293936357872)+(-0.352492545609)*0.5**o
        ref[3,3,0,0]=(-0.985298342079)/(o+1.)+(-0.319004787481)+(0.70346530112)*0.5**o
        ref[3,3,0,1]=(-0.758645627646)/(o+1.)+(-0.0862539708788)+(0.841285056577)*0.5**o
        ref[3,3,1,0]=(-0.301781252786)/(o+1.)+(-0.734094431327)+(-0.614293213484)*0.5**o
        ref[3,3,1,1]=(-0.948032248565)/(o+1.)+(-0.656818260243)+(-0.287510519978)*0.5**o
        ref[3,3,2,0]=(0.786904887352)/(o+1.)+(-0.359703314582)+(0.145506943729)*0.5**o
        ref[3,3,2,1]=(-0.237571902274)/(o+1.)+(0.408234845369)+(0.60782215735)*0.5**o
        ref[3,4,0,0]=(-0.847456183316)/(o+1.)+(0.495560925646)+(0.691107698337)*0.5**o
        ref[3,4,0,1]=(-0.362440886816)/(o+1.)+(-0.379020891487)+(-0.00167606578449)*0.5**o
        ref[3,4,1,0]=(-0.317176913637)/(o+1.)+(0.110349805884)+(-0.74026325403)*0.5**o
        ref[3,4,1,1]=(-0.435125210116)/(o+1.)+(-0.411332608113)+(-0.230005454291)*0.5**o
        ref[3,4,2,0]=(0.460436795546)/(o+1.)+(-0.48533896554)+(-0.638149620463)*0.5**o
        ref[3,4,2,1]=(0.632560678557)/(o+1.)+(0.555182655599)+(-0.793537245531)*0.5**o
      else:
        arg[0,0,0,0]=(0.816458625566)*x[0]**o+(-0.438180707327)*x[0]+(-0.39836196112)*x[1]**o+(0.641701267855)*x[1]+(0.919064096208)*x[2]**o+(0.0155653082454)*x[2]
        arg[0,0,0,1]=(-0.709065020122)*x[0]**o+(-0.787938005558)*x[0]+(-0.459829890247)*x[1]**o+(-0.306587811189)*x[1]+(0.648348776519)*x[2]**o+(-0.782857657922)*x[2]
        arg[0,0,1,0]=(-0.09165636265)*x[0]**o+(-0.215464577259)*x[0]+(-0.445060581145)*x[1]**o+(0.541797069021)*x[1]+(-0.312096860473)*x[2]**o+(0.307061070632)*x[2]
        arg[0,0,1,1]=(0.900670566988)*x[0]**o+(-0.157226841228)*x[0]+(-0.00514019601556)*x[1]**o+(0.167616993369)*x[1]+(0.331077209212)*x[2]**o+(0.0797784869412)*x[2]
        arg[0,0,2,0]=(-0.883617789072)*x[0]**o+(-0.233507045302)*x[0]+(0.540311388635)*x[1]**o+(-0.0698351272775)*x[1]+(0.598059668701)*x[2]**o+(-0.842961707358)*x[2]
        arg[0,0,2,1]=(-0.770041366552)*x[0]**o+(-0.683157969301)*x[0]+(0.235580399076)*x[1]**o+(-0.994177047985)*x[1]+(-0.90074656484)*x[2]**o+(-0.0133600820046)*x[2]
        arg[0,1,0,0]=(0.90396926296)*x[0]**o+(0.0492056402145)*x[0]+(0.32744155595)*x[1]**o+(0.0162505824727)*x[1]+(0.386607661085)*x[2]**o+(0.759639829913)*x[2]
        arg[0,1,0,1]=(0.47956310963)*x[0]**o+(0.638067028609)*x[0]+(-0.520596788355)*x[1]**o+(0.476897948019)*x[1]+(0.950320232053)*x[2]**o+(-0.603034001297)*x[2]
        arg[0,1,1,0]=(-0.75177304495)*x[0]**o+(0.369014786745)*x[0]+(-0.332220011735)*x[1]**o+(-0.835059332186)*x[1]+(0.0494323092557)*x[2]**o+(-0.796200422375)*x[2]
        arg[0,1,1,1]=(-0.109937058453)*x[0]**o+(0.607637390544)*x[0]+(0.464147925564)*x[1]**o+(0.650325883743)*x[1]+(0.698196425121)*x[2]**o+(0.933002044297)*x[2]
        arg[0,1,2,0]=(-0.13801307581)*x[0]**o+(0.75768744715)*x[0]+(-0.35990643116)*x[1]**o+(-0.787975707318)*x[1]+(-0.259281382264)*x[2]**o+(0.971003238549)*x[2]
        arg[0,1,2,1]=(-0.518540484981)*x[0]**o+(0.197912267942)*x[0]+(0.481058074022)*x[1]**o+(0.918189335061)*x[1]+(0.160239271801)*x[2]**o+(0.75950986792)*x[2]
        arg[0,2,0,0]=(0.286922773164)*x[0]**o+(-0.874592150595)*x[0]+(0.416014232771)*x[1]**o+(-0.492379239759)*x[1]+(0.950640982197)*x[2]**o+(-0.10625158241)*x[2]
        arg[0,2,0,1]=(-0.116447648923)*x[0]**o+(-0.499481854319)*x[0]+(-0.556602044084)*x[1]**o+(0.753256358343)*x[1]+(0.280780022694)*x[2]**o+(-0.879728755507)*x[2]
        arg[0,2,1,0]=(-0.322803067618)*x[0]**o+(0.0974878233964)*x[0]+(0.420547949274)*x[1]**o+(0.304839957099)*x[1]+(-0.12605355567)*x[2]**o+(-0.166970509739)*x[2]
        arg[0,2,1,1]=(0.997823909059)*x[0]**o+(0.76749010767)*x[0]+(0.376351544673)*x[1]**o+(-0.316937252621)*x[1]+(-0.627670639733)*x[2]**o+(0.584121037987)*x[2]
        arg[0,2,2,0]=(0.516866019209)*x[0]**o+(0.413564799736)*x[0]+(-0.0730736747833)*x[1]**o+(-0.500984764568)*x[1]+(-0.989323894645)*x[2]**o+(-0.261458643035)*x[2]
        arg[0,2,2,1]=(0.777538237337)*x[0]**o+(-0.0623276952959)*x[0]+(0.167068351133)*x[1]**o+(-0.345550070545)*x[1]+(-0.28166577069)*x[2]**o+(-0.568650037569)*x[2]
        arg[0,3,0,0]=(-0.622067362782)*x[0]**o+(0.971384323624)*x[0]+(-0.099597286042)*x[1]**o+(-0.265680407306)*x[1]+(-0.446275631545)*x[2]**o+(-0.0637591060995)*x[2]
        arg[0,3,0,1]=(-0.49698112241)*x[0]**o+(0.766181263321)*x[0]+(-0.977014162391)*x[1]**o+(0.806444170647)*x[1]+(-0.441023882645)*x[2]**o+(-0.919279365683)*x[2]
        arg[0,3,1,0]=(0.917091542097)*x[0]**o+(-0.67016353228)*x[0]+(0.0468346020734)*x[1]**o+(0.81092419675)*x[1]+(0.174037096652)*x[2]**o+(0.564190407348)*x[2]
        arg[0,3,1,1]=(-0.437188100982)*x[0]**o+(-0.0857174204982)*x[0]+(-0.971592105442)*x[1]**o+(-0.122352019844)*x[1]+(-0.853225566811)*x[2]**o+(-0.861128476903)*x[2]
        arg[0,3,2,0]=(-0.841265485658)*x[0]**o+(0.553186811235)*x[0]+(0.414489259007)*x[1]**o+(-0.800997281522)*x[1]+(0.630269543743)*x[2]**o+(0.263584335528)*x[2]
        arg[0,3,2,1]=(0.383282548921)*x[0]**o+(0.260041562865)*x[0]+(-0.95180999378)*x[1]**o+(0.566954668715)*x[1]+(0.688843907846)*x[2]**o+(-0.166837115553)*x[2]
        arg[0,4,0,0]=(0.0385891845805)*x[0]**o+(0.324932244772)*x[0]+(-0.946997425504)*x[1]**o+(-0.745439090316)*x[1]+(-0.721657519976)*x[2]**o+(-0.714366519961)*x[2]
        arg[0,4,0,1]=(-0.387739686761)*x[0]**o+(-0.860915144676)*x[0]+(0.270056169204)*x[1]**o+(-0.914861208322)*x[1]+(-0.453514265988)*x[2]**o+(-0.495768314604)*x[2]
        arg[0,4,1,0]=(-0.329464483421)*x[0]**o+(0.313424566493)*x[0]+(0.185677682455)*x[1]**o+(-0.720200750426)*x[1]+(0.425742092834)*x[2]**o+(0.505209526425)*x[2]
        arg[0,4,1,1]=(0.63272559561)*x[0]**o+(-0.026429919604)*x[0]+(-0.14023723263)*x[1]**o+(-0.580805084228)*x[1]+(0.141232832837)*x[2]**o+(0.241327430534)*x[2]
        arg[0,4,2,0]=(-0.847582850604)*x[0]**o+(0.207270297356)*x[0]+(0.851958035148)*x[1]**o+(0.336594484304)*x[1]+(-0.221434386211)*x[2]**o+(-0.607602400875)*x[2]
        arg[0,4,2,1]=(-0.722990535113)*x[0]**o+(-0.379964881881)*x[0]+(0.825779173038)*x[1]**o+(-0.285509543358)*x[1]+(-0.518809494335)*x[2]**o+(-0.788507606636)*x[2]
        arg[1,0,0,0]=(0.0725951752952)*x[0]**o+(-0.282800153416)*x[0]+(-0.0228303688498)*x[1]**o+(-0.863127549259)*x[1]+(-0.383553228278)*x[2]**o+(-0.0151737809185)*x[2]
        arg[1,0,0,1]=(-0.254920010173)*x[0]**o+(0.160195189848)*x[0]+(0.216407190109)*x[1]**o+(-0.814599655672)*x[1]+(0.771746025302)*x[2]**o+(0.0636221760766)*x[2]
        arg[1,0,1,0]=(0.486342223622)*x[0]**o+(0.736242651359)*x[0]+(0.720722697611)*x[1]**o+(0.974853194395)*x[1]+(-0.834707965096)*x[2]**o+(-0.947169299038)*x[2]
        arg[1,0,1,1]=(0.777015150638)*x[0]**o+(-0.228082772046)*x[0]+(-0.51095285588)*x[1]**o+(-0.835057692508)*x[1]+(-0.564012944223)*x[2]**o+(-0.898040787549)*x[2]
        arg[1,0,2,0]=(0.240813275278)*x[0]**o+(-0.509566588932)*x[0]+(0.247803237451)*x[1]**o+(0.156442781059)*x[1]+(0.618367580757)*x[2]**o+(0.943435734207)*x[2]
        arg[1,0,2,1]=(-0.670257601629)*x[0]**o+(-0.470716006203)*x[0]+(-0.559207336128)*x[1]**o+(-0.81577896974)*x[1]+(0.0571093198914)*x[2]**o+(0.477103420898)*x[2]
        arg[1,1,0,0]=(0.985895581872)*x[0]**o+(-0.00359465828149)*x[0]+(0.00739312124399)*x[1]**o+(-0.931299968425)*x[1]+(-0.10047347666)*x[2]**o+(0.410589913002)*x[2]
        arg[1,1,0,1]=(-0.954132399275)*x[0]**o+(0.159222338169)*x[0]+(-0.865370419503)*x[1]**o+(-0.744696059283)*x[1]+(-0.794993976252)*x[2]**o+(0.961664562839)*x[2]
        arg[1,1,1,0]=(0.462810965625)*x[0]**o+(0.927378551017)*x[0]+(-0.105005418179)*x[1]**o+(0.282340667975)*x[1]+(-0.443931241323)*x[2]**o+(0.740071617715)*x[2]
        arg[1,1,1,1]=(0.645629353541)*x[0]**o+(0.848804271896)*x[0]+(-0.882497356826)*x[1]**o+(-0.74661619875)*x[1]+(-0.927133238344)*x[2]**o+(-0.502758459229)*x[2]
        arg[1,1,2,0]=(-0.554993653726)*x[0]**o+(-0.473286752336)*x[0]+(0.759624773306)*x[1]**o+(0.809169103932)*x[1]+(0.657766085675)*x[2]**o+(0.763913478843)*x[2]
        arg[1,1,2,1]=(0.1476587556)*x[0]**o+(0.385711162597)*x[0]+(-0.0767252754622)*x[1]**o+(-0.544968629318)*x[1]+(0.16751959089)*x[2]**o+(-0.742697412247)*x[2]
        arg[1,2,0,0]=(-0.715887407694)*x[0]**o+(0.273818352078)*x[0]+(0.0426629688432)*x[1]**o+(-0.808531265186)*x[1]+(-0.699839650334)*x[2]**o+(-0.332093073009)*x[2]
        arg[1,2,0,1]=(0.48440337855)*x[0]**o+(0.765443731546)*x[0]+(-0.0761968055499)*x[1]**o+(0.0516389293584)*x[1]+(-0.486224925751)*x[2]**o+(-0.796105709176)*x[2]
        arg[1,2,1,0]=(0.621562227153)*x[0]**o+(0.907213416636)*x[0]+(0.591392192921)*x[1]**o+(-0.825911160831)*x[1]+(0.690773959314)*x[2]**o+(-0.744144678185)*x[2]
        arg[1,2,1,1]=(0.472846565983)*x[0]**o+(-0.837395663533)*x[0]+(0.526438435448)*x[1]**o+(0.310053110281)*x[1]+(0.634301291566)*x[2]**o+(0.608270679412)*x[2]
        arg[1,2,2,0]=(-0.212000572333)*x[0]**o+(0.644218219058)*x[0]+(0.827779140409)*x[1]**o+(0.563968121281)*x[1]+(-0.0556319977635)*x[2]**o+(0.98126338415)*x[2]
        arg[1,2,2,1]=(-0.632752637148)*x[0]**o+(-0.999490611532)*x[0]+(0.0717950178432)*x[1]**o+(-0.0611949850049)*x[1]+(0.740177288895)*x[2]**o+(0.241010363236)*x[2]
        arg[1,3,0,0]=(0.390081216336)*x[0]**o+(-0.389460065845)*x[0]+(0.0812808695521)*x[1]**o+(0.0843698990866)*x[1]+(0.672844227186)*x[2]**o+(0.306345015188)*x[2]
        arg[1,3,0,1]=(-0.402887151891)*x[0]**o+(-0.908615473265)*x[0]+(-0.296307155209)*x[1]**o+(0.228858413043)*x[1]+(0.509401868055)*x[2]**o+(-0.793787805038)*x[2]
        arg[1,3,1,0]=(0.253947492129)*x[0]**o+(0.401465417831)*x[0]+(-0.942329176703)*x[1]**o+(-0.518371272083)*x[1]+(0.90142806242)*x[2]**o+(0.694084281869)*x[2]
        arg[1,3,1,1]=(-0.435656267291)*x[0]**o+(-0.743058541536)*x[0]+(-0.22434079396)*x[1]**o+(-0.414125965702)*x[1]+(0.743546448275)*x[2]**o+(0.323188772386)*x[2]
        arg[1,3,2,0]=(-0.069319863146)*x[0]**o+(-0.0749969565014)*x[0]+(0.29700943253)*x[1]**o+(0.972639470201)*x[1]+(-0.270592469279)*x[2]**o+(-0.355098058901)*x[2]
        arg[1,3,2,1]=(0.741825647399)*x[0]**o+(0.542055833087)*x[0]+(-0.0684409109262)*x[1]**o+(0.842094761143)*x[1]+(-0.777320219447)*x[2]**o+(0.758570478904)*x[2]
        arg[1,4,0,0]=(-0.66003338591)*x[0]**o+(-0.393755284573)*x[0]+(0.617154869065)*x[1]**o+(-0.480999198967)*x[1]+(0.350625972066)*x[2]**o+(0.733472524967)*x[2]
        arg[1,4,0,1]=(-0.0001759210396)*x[0]**o+(-0.0900574950432)*x[0]+(-0.592936627579)*x[1]**o+(-0.647214933825)*x[1]+(-0.43249743908)*x[2]**o+(-0.901387692551)*x[2]
        arg[1,4,1,0]=(0.052435899043)*x[0]**o+(0.163326392489)*x[0]+(0.433169876656)*x[1]**o+(0.977133611649)*x[1]+(-0.311959404527)*x[2]**o+(0.297218570651)*x[2]
        arg[1,4,1,1]=(0.00806522196579)*x[0]**o+(-0.121807466189)*x[0]+(-0.176557734776)*x[1]**o+(0.865146379005)*x[1]+(-0.304648040177)*x[2]**o+(0.269150163824)*x[2]
        arg[1,4,2,0]=(0.440212388104)*x[0]**o+(-0.590292628428)*x[0]+(-0.595359323633)*x[1]**o+(0.606734888384)*x[1]+(-0.951693103773)*x[2]**o+(-0.539741612207)*x[2]
        arg[1,4,2,1]=(0.346121727918)*x[0]**o+(0.188961570963)*x[0]+(0.402818736255)*x[1]**o+(-0.474029523974)*x[1]+(-0.95972237672)*x[2]**o+(0.983543070357)*x[2]
        arg[2,0,0,0]=(-0.801941188348)*x[0]**o+(-0.318215868754)*x[0]+(0.322031073259)*x[1]**o+(-0.00368402006147)*x[1]+(0.473846239338)*x[2]**o+(0.512752561753)*x[2]
        arg[2,0,0,1]=(0.786459471841)*x[0]**o+(0.941049841845)*x[0]+(0.064717660139)*x[1]**o+(-0.891518891268)*x[1]+(-0.317681565962)*x[2]**o+(-0.672310725168)*x[2]
        arg[2,0,1,0]=(0.813571093948)*x[0]**o+(-0.146551267883)*x[0]+(0.942106359835)*x[1]**o+(0.165543949281)*x[1]+(-0.0301736488709)*x[2]**o+(0.936915469225)*x[2]
        arg[2,0,1,1]=(-0.0298664722355)*x[0]**o+(-0.596644819792)*x[0]+(-0.306728572634)*x[1]**o+(0.770671782936)*x[1]+(-0.334606536003)*x[2]**o+(0.888940190159)*x[2]
        arg[2,0,2,0]=(0.688553055025)*x[0]**o+(-0.241541868482)*x[0]+(-0.694609526571)*x[1]**o+(-0.509896457784)*x[1]+(-0.0610272843414)*x[2]**o+(-0.251290389152)*x[2]
        arg[2,0,2,1]=(-0.959007707329)*x[0]**o+(-0.193856482901)*x[0]+(-0.138697126551)*x[1]**o+(0.768166305153)*x[1]+(0.951006645306)*x[2]**o+(-0.533986863953)*x[2]
        arg[2,1,0,0]=(0.648021743651)*x[0]**o+(-0.839074631201)*x[0]+(-0.445337118707)*x[1]**o+(0.997475278653)*x[1]+(-0.154475323496)*x[2]**o+(-0.45292992537)*x[2]
        arg[2,1,0,1]=(-0.134785884326)*x[0]**o+(-0.906738442209)*x[0]+(0.00509900741927)*x[1]**o+(0.489865232888)*x[1]+(0.620823018353)*x[2]**o+(-0.480127263607)*x[2]
        arg[2,1,1,0]=(-0.210371458935)*x[0]**o+(0.308453655681)*x[0]+(-0.510473100104)*x[1]**o+(-0.0776488834445)*x[1]+(0.58103002434)*x[2]**o+(0.039267622777)*x[2]
        arg[2,1,1,1]=(-0.124204683462)*x[0]**o+(0.429525935007)*x[0]+(0.183174600008)*x[1]**o+(-0.665914708196)*x[1]+(-0.576713549033)*x[2]**o+(-0.3788567811)*x[2]
        arg[2,1,2,0]=(-0.864065825608)*x[0]**o+(0.644046059464)*x[0]+(-0.265959280151)*x[1]**o+(0.989465634246)*x[1]+(0.536466956842)*x[2]**o+(0.689999246325)*x[2]
        arg[2,1,2,1]=(0.37622234855)*x[0]**o+(-0.905554688955)*x[0]+(-0.0258037833143)*x[1]**o+(-0.23415538846)*x[1]+(-0.64461795706)*x[2]**o+(0.262259040646)*x[2]
        arg[2,2,0,0]=(0.0541619627989)*x[0]**o+(-0.978918782799)*x[0]+(0.478333900479)*x[1]**o+(0.958652455667)*x[1]+(-0.746701804693)*x[2]**o+(-0.557405653804)*x[2]
        arg[2,2,0,1]=(0.611472347101)*x[0]**o+(-0.409754132769)*x[0]+(0.291274496706)*x[1]**o+(0.901983238042)*x[1]+(-0.102917678824)*x[2]**o+(-0.668205224505)*x[2]
        arg[2,2,1,0]=(-0.926062469523)*x[0]**o+(0.101032437783)*x[0]+(-0.385485330853)*x[1]**o+(0.0521756546126)*x[1]+(0.132320977758)*x[2]**o+(0.471619572767)*x[2]
        arg[2,2,1,1]=(0.0689111652346)*x[0]**o+(-0.517070173331)*x[0]+(-0.599184868344)*x[1]**o+(-0.814239996638)*x[1]+(0.97403158037)*x[2]**o+(-0.546843890524)*x[2]
        arg[2,2,2,0]=(-0.133251706696)*x[0]**o+(-0.0954736405212)*x[0]+(0.780584398494)*x[1]**o+(-0.505305860913)*x[1]+(-0.11518705364)*x[2]**o+(0.0855698333433)*x[2]
        arg[2,2,2,1]=(-0.567309977436)*x[0]**o+(-0.284182307402)*x[0]+(0.220302202269)*x[1]**o+(-0.252628588826)*x[1]+(-0.368390646272)*x[2]**o+(-0.577499260947)*x[2]
        arg[2,3,0,0]=(0.477968626266)*x[0]**o+(0.522764690241)*x[0]+(-0.0339165683557)*x[1]**o+(0.623897338329)*x[1]+(-0.169558925831)*x[2]**o+(-0.426663511213)*x[2]
        arg[2,3,0,1]=(0.185482762361)*x[0]**o+(0.496017413966)*x[0]+(-0.217189284009)*x[1]**o+(0.488205372581)*x[1]+(0.445762454794)*x[2]**o+(-0.423587587222)*x[2]
        arg[2,3,1,0]=(-0.489549524298)*x[0]**o+(-0.482299084307)*x[0]+(0.567729727254)*x[1]**o+(-0.341933025591)*x[1]+(-0.325429025641)*x[2]**o+(0.476030984261)*x[2]
        arg[2,3,1,1]=(-0.390754226646)*x[0]**o+(-0.980255652462)*x[0]+(0.109328516217)*x[1]**o+(-0.0141509875662)*x[1]+(0.486458644215)*x[2]**o+(-0.668018419371)*x[2]
        arg[2,3,2,0]=(0.352912006363)*x[0]**o+(-0.542969561227)*x[0]+(0.512067200296)*x[1]**o+(0.661110389562)*x[1]+(0.289744832231)*x[2]**o+(-0.176943484459)*x[2]
        arg[2,3,2,1]=(0.535262244127)*x[0]**o+(0.528325019892)*x[0]+(-0.106312276613)*x[1]**o+(-0.484575045858)*x[1]+(-0.407717587021)*x[2]**o+(-0.786366604025)*x[2]
        arg[2,4,0,0]=(-0.511771630262)*x[0]**o+(-0.590485876141)*x[0]+(0.493685168985)*x[1]**o+(-0.726700517739)*x[1]+(-0.213137223607)*x[2]**o+(-0.566923814015)*x[2]
        arg[2,4,0,1]=(0.421559088561)*x[0]**o+(0.838784944841)*x[0]+(-0.693654869402)*x[1]**o+(-0.0755493272975)*x[1]+(-0.174245799979)*x[2]**o+(0.242245946157)*x[2]
        arg[2,4,1,0]=(0.681235817735)*x[0]**o+(-0.313564199551)*x[0]+(0.663598935418)*x[1]**o+(-0.604327253895)*x[1]+(0.698985821444)*x[2]**o+(0.298899431146)*x[2]
        arg[2,4,1,1]=(0.432866305507)*x[0]**o+(-0.891501121013)*x[0]+(-0.290106178906)*x[1]**o+(0.268152357012)*x[1]+(0.75279498948)*x[2]**o+(0.211312203371)*x[2]
        arg[2,4,2,0]=(-0.956901622446)*x[0]**o+(0.742548241435)*x[0]+(-0.956209816095)*x[1]**o+(-0.539017080821)*x[1]+(-0.08062797007)*x[2]**o+(0.383843198102)*x[2]
        arg[2,4,2,1]=(-0.16700370288)*x[0]**o+(0.388837619298)*x[0]+(-0.190586726252)*x[1]**o+(0.916004535916)*x[1]+(0.310254387059)*x[2]**o+(-0.380371301717)*x[2]
        arg[3,0,0,0]=(0.64046380712)*x[0]**o+(0.254058760972)*x[0]+(-0.829050222873)*x[1]**o+(0.289942284927)*x[1]+(0.196268101069)*x[2]**o+(0.748880970446)*x[2]
        arg[3,0,0,1]=(-0.226057776579)*x[0]**o+(0.982685008224)*x[0]+(0.819680514589)*x[1]**o+(0.48005222662)*x[1]+(-0.615248069776)*x[2]**o+(-0.0923489923012)*x[2]
        arg[3,0,1,0]=(-0.358031084019)*x[0]**o+(-0.83941518843)*x[0]+(-0.0687218422525)*x[1]**o+(-0.648976147029)*x[1]+(0.711009285701)*x[2]**o+(0.360758898243)*x[2]
        arg[3,0,1,1]=(-0.312175299678)*x[0]**o+(0.360877864123)*x[0]+(0.653283960509)*x[1]**o+(0.323527948404)*x[1]+(-0.278454805659)*x[2]**o+(0.0585194127844)*x[2]
        arg[3,0,2,0]=(0.367207247645)*x[0]**o+(-0.0743547693027)*x[0]+(-0.473358544608)*x[1]**o+(0.332985041128)*x[1]+(-0.899612971909)*x[2]**o+(0.69409768473)*x[2]
        arg[3,0,2,1]=(-0.806787228991)*x[0]**o+(-0.110344204889)*x[0]+(-0.631610779885)*x[1]**o+(0.0239853435419)*x[1]+(0.83036063388)*x[2]**o+(0.623825256849)*x[2]
        arg[3,1,0,0]=(-0.268344201624)*x[0]**o+(-0.59230799177)*x[0]+(0.493719194706)*x[1]**o+(-0.128351291806)*x[1]+(0.0584636422077)*x[2]**o+(0.0859725996454)*x[2]
        arg[3,1,0,1]=(-0.870410307785)*x[0]**o+(-0.917370176071)*x[0]+(0.395701966434)*x[1]**o+(0.694856510628)*x[1]+(-0.748440730513)*x[2]**o+(0.748465450614)*x[2]
        arg[3,1,1,0]=(-0.178465888488)*x[0]**o+(-0.956884701156)*x[0]+(0.0472790389286)*x[1]**o+(0.946702274722)*x[1]+(-0.200668750878)*x[2]**o+(-0.0104023894447)*x[2]
        arg[3,1,1,1]=(0.113559334778)*x[0]**o+(0.270505514561)*x[0]+(-0.238008799138)*x[1]**o+(0.533067625871)*x[1]+(-0.586814102301)*x[2]**o+(0.045657074859)*x[2]
        arg[3,1,2,0]=(-0.574716699633)*x[0]**o+(-0.659499274639)*x[0]+(0.892605757655)*x[1]**o+(-0.39697940457)*x[1]+(0.363011145793)*x[2]**o+(-0.364722526091)*x[2]
        arg[3,1,2,1]=(-0.0193167048813)*x[0]**o+(0.761104649925)*x[0]+(0.490164183635)*x[1]**o+(0.10693267057)*x[1]+(0.635328989408)*x[2]**o+(-0.353114187925)*x[2]
        arg[3,2,0,0]=(0.0777423524941)*x[0]**o+(-0.431011355257)*x[0]+(-0.356762042134)*x[1]**o+(-0.94361679144)*x[1]+(-0.716191700456)*x[2]**o+(0.49906825221)*x[2]
        arg[3,2,0,1]=(0.80859948321)*x[0]**o+(0.542680788872)*x[0]+(-0.626897369633)*x[1]**o+(-0.821877255135)*x[1]+(-0.455575534023)*x[2]**o+(0.317480554025)*x[2]
        arg[3,2,1,0]=(-0.0144328272818)*x[0]**o+(-0.395367848628)*x[0]+(-0.601042089925)*x[1]**o+(-0.0822921879892)*x[1]+(0.896199761801)*x[2]**o+(-0.215055215166)*x[2]
        arg[3,2,1,1]=(0.377945289524)*x[0]**o+(-0.580064012583)*x[0]+(0.993926713349)*x[1]**o+(-0.736307101943)*x[1]+(-0.939628422411)*x[2]**o+(-0.923219473393)*x[2]
        arg[3,2,2,0]=(-0.733150287394)*x[0]**o+(0.692501827859)*x[0]+(0.904655347343)*x[1]**o+(0.310386512564)*x[1]+(-0.93795078156)*x[2]**o+(0.772534565721)*x[2]
        arg[3,2,2,1]=(-0.155229801255)*x[0]**o+(0.580356480314)*x[0]+(0.0673352927499)*x[1]**o+(-0.274975805243)*x[1]+(0.553818209037)*x[2]**o+(-0.835142674988)*x[2]
        arg[3,3,0,0]=(0.60752130978)*x[0]**o+(0.847750395856)*x[0]+(0.8635279539)*x[1]**o+(0.16341072973)*x[1]+(0.470011948821)*x[2]**o+(0.41529971393)*x[2]
        arg[3,3,0,1]=(0.755049219363)*x[0]**o+(0.66080573214)*x[0]+(-0.0256982576552)*x[1]**o+(-0.0311937109251)*x[1]+(-0.028843413794)*x[2]**o+(-0.19345503082)*x[2]
        arg[3,3,1,0]=(-0.39765237082)*x[0]**o+(-0.746359385218)*x[0]+(0.275396324626)*x[1]**o+(0.064571659107)*x[1]+(0.771289010752)*x[2]**o+(-6.69891594269e-05)*x[2]
        arg[3,3,1,1]=(0.350497253938)*x[0]**o+(-0.236419516561)*x[0]+(0.673483791847)*x[1]**o+(-0.934644138115)*x[1]+(0.994408125332)*x[2]**o+(-0.398423596854)*x[2]
        arg[3,3,2,0]=(-0.952544750261)*x[0]**o+(0.24256752245)*x[0]+(0.376782843601)*x[1]**o+(-0.349840168156)*x[1]+(-0.260325141903)*x[2]**o+(-0.452866344447)*x[2]
        arg[3,3,2,1]=(-0.0232941915849)*x[0]**o+(0.191625715695)*x[0]+(-0.490187753329)*x[1]**o+(-0.823976612466)*x[1]+(0.476804203752)*x[2]**o+(0.61057521184)*x[2]
        arg[3,4,0,0]=(0.0279130167137)*x[0]**o+(-0.134953133648)*x[0]+(0.0821614836141)*x[1]**o+(-0.894046763697)*x[1]+(-0.497197221983)*x[2]**o+(0.586297244251)*x[2]
        arg[3,4,0,1]=(0.0423746259802)*x[0]**o+(0.458075512165)*x[0]+(0.855157549678)*x[1]**o+(-0.607654212682)*x[1]+(-0.672693159955)*x[2]**o+(-0.300116892313)*x[2]
        arg[3,4,1,0]=(-0.695685046398)*x[0]**o+(-0.709807910289)*x[0]+(-0.167879145281)*x[1]**o+(0.524084329764)*x[1]+(0.763706690019)*x[2]**o+(-0.219502671318)*x[2]
        arg[3,4,1,1]=(0.864599640696)*x[0]**o+(-0.341125612414)*x[0]+(-0.548888211648)*x[1]**o+(0.918497762429)*x[1]+(-0.445600024463)*x[2]**o+(0.221658300325)*x[2]
        arg[3,4,2,0]=(-0.893419673448)*x[0]**o+(0.653718653341)*x[0]+(0.323868832441)*x[1]**o+(0.905933530129)*x[1]+(0.62045953406)*x[2]**o+(0.145474797235)*x[2]
        arg[3,4,2,1]=(0.480643160777)*x[0]**o+(-0.15896921521)*x[0]+(-0.329564714008)*x[1]**o+(-0.613303857496)*x[1]+(-0.483827450625)*x[2]**o+(-0.91023336048)*x[2]
        ref[0,0,0,0]=(0.520702135089)/(o+1.)+(0.109542934387)+(0.816458625566)*0.5**o
        ref[0,0,0,1]=(0.188518886272)/(o+1.)+(-0.938691737334)+(-0.709065020122)*0.5**o
        ref[0,0,1,0]=(-0.757157441618)/(o+1.)+(0.316696781197)+(-0.09165636265)*0.5**o
        ref[0,0,1,1]=(0.325937013197)/(o+1.)+(0.0450843195409)+(0.900670566988)*0.5**o
        ref[0,0,2,0]=(1.13837105734)/(o+1.)+(-0.573151939969)+(-0.883617789072)*0.5**o
        ref[0,0,2,1]=(-0.665166165764)/(o+1.)+(-0.845347549645)+(-0.770041366552)*0.5**o
        ref[0,1,0,0]=(0.714049217035)/(o+1.)+(0.4125480263)+(0.90396926296)*0.5**o
        ref[0,1,0,1]=(0.429723443697)/(o+1.)+(0.255965487666)+(0.47956310963)*0.5**o
        ref[0,1,1,0]=(-0.282787702479)/(o+1.)+(-0.631122483908)+(-0.75177304495)*0.5**o
        ref[0,1,1,1]=(1.16234435068)/(o+1.)+(1.09548265929)+(-0.109937058453)*0.5**o
        ref[0,1,2,0]=(-0.619187813423)/(o+1.)+(0.47035748919)+(-0.13801307581)*0.5**o
        ref[0,1,2,1]=(0.641297345823)/(o+1.)+(0.937805735461)+(-0.518540484981)*0.5**o
        ref[0,2,0,0]=(1.36665521497)/(o+1.)+(-0.736611486382)+(0.286922773164)*0.5**o
        ref[0,2,0,1]=(-0.27582202139)/(o+1.)+(-0.312977125741)+(-0.116447648923)*0.5**o
        ref[0,2,1,0]=(0.294494393603)/(o+1.)+(0.117678635379)+(-0.322803067618)*0.5**o
        ref[0,2,1,1]=(-0.251319095061)/(o+1.)+(0.517336946518)+(0.997823909059)*0.5**o
        ref[0,2,2,0]=(-1.06239756943)/(o+1.)+(-0.174439303934)+(0.516866019209)*0.5**o
        ref[0,2,2,1]=(-0.114597419557)/(o+1.)+(-0.488263901705)+(0.777538237337)*0.5**o
        ref[0,3,0,0]=(-0.545872917587)/(o+1.)+(0.320972405109)+(-0.622067362782)*0.5**o
        ref[0,3,0,1]=(-1.41803804504)/(o+1.)+(0.326673034142)+(-0.49698112241)*0.5**o
        ref[0,3,1,0]=(0.220871698725)/(o+1.)+(0.352475535909)+(0.917091542097)*0.5**o
        ref[0,3,1,1]=(-1.82481767225)/(o+1.)+(-0.534598958622)+(-0.437188100982)*0.5**o
        ref[0,3,2,0]=(1.04475880275)/(o+1.)+(0.00788693262039)+(-0.841265485658)*0.5**o
        ref[0,3,2,1]=(-0.262966085934)/(o+1.)+(0.330079558014)+(0.383282548921)*0.5**o
        ref[0,4,0,0]=(-1.66865494548)/(o+1.)+(-0.567436682752)+(0.0385891845805)*0.5**o
        ref[0,4,0,1]=(-0.183458096784)/(o+1.)+(-1.1357723338)+(-0.387739686761)*0.5**o
        ref[0,4,1,0]=(0.611419775289)/(o+1.)+(0.0492166712458)+(-0.329464483421)*0.5**o
        ref[0,4,1,1]=(0.000995600207189)/(o+1.)+(-0.182953786649)+(0.63272559561)*0.5**o
        ref[0,4,2,0]=(0.630523648936)/(o+1.)+(-0.0318688096074)+(-0.847582850604)*0.5**o
        ref[0,4,2,1]=(0.306969678703)/(o+1.)+(-0.726991015938)+(-0.722990535113)*0.5**o
        ref[1,0,0,0]=(-0.406383597128)/(o+1.)+(-0.580550741796)+(0.0725951752952)*0.5**o
        ref[1,0,0,1]=(0.988153215411)/(o+1.)+(-0.295391144874)+(-0.254920010173)*0.5**o
        ref[1,0,1,0]=(-0.113985267485)/(o+1.)+(0.381963273358)+(0.486342223622)*0.5**o
        ref[1,0,1,1]=(-1.0749658001)/(o+1.)+(-0.980590626052)+(0.777015150638)*0.5**o
        ref[1,0,2,0]=(0.866170818208)/(o+1.)+(0.295155963167)+(0.240813275278)*0.5**o
        ref[1,0,2,1]=(-0.502098016236)/(o+1.)+(-0.404695777523)+(-0.670257601629)*0.5**o
        ref[1,1,0,0]=(-0.0930803554155)/(o+1.)+(-0.262152356852)+(0.985895581872)*0.5**o
        ref[1,1,0,1]=(-1.66036439576)/(o+1.)+(0.188095420862)+(-0.954132399275)*0.5**o
        ref[1,1,1,0]=(-0.548936659502)/(o+1.)+(0.974895418354)+(0.462810965625)*0.5**o
        ref[1,1,1,1]=(-1.80963059517)/(o+1.)+(-0.200285193041)+(0.645629353541)*0.5**o
        ref[1,1,2,0]=(1.41739085898)/(o+1.)+(0.549897915219)+(-0.554993653726)*0.5**o
        ref[1,1,2,1]=(0.0907943154278)/(o+1.)+(-0.450977439484)+(0.1476587556)*0.5**o
        ref[1,2,0,0]=(-0.657176681491)/(o+1.)+(-0.433402993059)+(-0.715887407694)*0.5**o
        ref[1,2,0,1]=(-0.562421731301)/(o+1.)+(0.0104884758643)+(0.48440337855)*0.5**o
        ref[1,2,1,0]=(1.28216615223)/(o+1.)+(-0.33142121119)+(0.621562227153)*0.5**o
        ref[1,2,1,1]=(1.16073972701)/(o+1.)+(0.0404640630801)+(0.472846565983)*0.5**o
        ref[1,2,2,0]=(0.772147142645)/(o+1.)+(1.09472486224)+(-0.212000572333)*0.5**o
        ref[1,2,2,1]=(0.811972306738)/(o+1.)+(-0.409837616651)+(-0.632752637148)*0.5**o
        ref[1,3,0,0]=(0.754125096738)/(o+1.)+(0.000627424214716)+(0.390081216336)*0.5**o
        ref[1,3,0,1]=(0.213094712846)/(o+1.)+(-0.73677243263)+(-0.402887151891)*0.5**o
        ref[1,3,1,0]=(-0.0409011142831)/(o+1.)+(0.288589213809)+(0.253947492129)*0.5**o
        ref[1,3,1,1]=(0.519205654314)/(o+1.)+(-0.416997867426)+(-0.435656267291)*0.5**o
        ref[1,3,2,0]=(0.0264169632513)/(o+1.)+(0.271272227399)+(-0.069319863146)*0.5**o
        ref[1,3,2,1]=(-0.845761130374)/(o+1.)+(1.07136053657)+(0.741825647399)*0.5**o
        ref[1,4,0,0]=(0.967780841131)/(o+1.)+(-0.0706409792865)+(-0.66003338591)*0.5**o
        ref[1,4,0,1]=(-1.02543406666)/(o+1.)+(-0.81933006071)+(-0.0001759210396)*0.5**o
        ref[1,4,1,0]=(0.121210472128)/(o+1.)+(0.718839287394)+(0.052435899043)*0.5**o
        ref[1,4,1,1]=(-0.481205774953)/(o+1.)+(0.50624453832)+(0.00806522196579)*0.5**o
        ref[1,4,2,0]=(-1.54705242741)/(o+1.)+(-0.261649676126)+(0.440212388104)*0.5**o
        ref[1,4,2,1]=(-0.556903640465)/(o+1.)+(0.349237558673)+(0.346121727918)*0.5**o
        ref[2,0,0,0]=(0.795877312597)/(o+1.)+(0.0954263364687)+(-0.801941188348)*0.5**o
        ref[2,0,0,1]=(-0.252963905823)/(o+1.)+(-0.311389887295)+(0.786459471841)*0.5**o
        ref[2,0,1,0]=(0.911932710964)/(o+1.)+(0.477954075311)+(0.813571093948)*0.5**o
        ref[2,0,1,1]=(-0.641335108636)/(o+1.)+(0.531483576652)+(-0.0298664722355)*0.5**o
        ref[2,0,2,0]=(-0.755636810913)/(o+1.)+(-0.501364357709)+(0.688553055025)*0.5**o
        ref[2,0,2,1]=(0.812309518755)/(o+1.)+(0.0201614791497)+(-0.959007707329)*0.5**o
        ref[2,1,0,0]=(-0.599812442203)/(o+1.)+(-0.147264638959)+(0.648021743651)*0.5**o
        ref[2,1,0,1]=(0.625922025772)/(o+1.)+(-0.448500236464)+(-0.134785884326)*0.5**o
        ref[2,1,1,0]=(0.0705569242367)/(o+1.)+(0.135036197507)+(-0.210371458935)*0.5**o
        ref[2,1,1,1]=(-0.393538949024)/(o+1.)+(-0.307622777145)+(-0.124204683462)*0.5**o
        ref[2,1,2,0]=(0.270507676691)/(o+1.)+(1.16175547002)+(-0.864065825608)*0.5**o
        ref[2,1,2,1]=(-0.670421740374)/(o+1.)+(-0.438725518385)+(0.37622234855)*0.5**o
        ref[2,2,0,0]=(-0.268367904213)/(o+1.)+(-0.288835990468)+(0.0541619627989)*0.5**o
        ref[2,2,0,1]=(0.188356817882)/(o+1.)+(-0.0879880596161)+(0.611472347101)*0.5**o
        ref[2,2,1,0]=(-0.253164353095)/(o+1.)+(0.312413832582)+(-0.926062469523)*0.5**o
        ref[2,2,1,1]=(0.374846712027)/(o+1.)+(-0.939077030246)+(0.0689111652346)*0.5**o
        ref[2,2,2,0]=(0.665397344854)/(o+1.)+(-0.257604834045)+(-0.133251706696)*0.5**o
        ref[2,2,2,1]=(-0.148088444003)/(o+1.)+(-0.557155078588)+(-0.567309977436)*0.5**o
        ref[2,3,0,0]=(-0.203475494187)/(o+1.)+(0.359999258678)+(0.477968626266)*0.5**o
        ref[2,3,0,1]=(0.228573170785)/(o+1.)+(0.280317599662)+(0.185482762361)*0.5**o
        ref[2,3,1,0]=(0.242300701614)/(o+1.)+(-0.174100562819)+(-0.489549524298)*0.5**o
        ref[2,3,1,1]=(0.595787160431)/(o+1.)+(-0.8312125297)+(-0.390754226646)*0.5**o
        ref[2,3,2,0]=(0.801812032527)/(o+1.)+(-0.0294013280621)+(0.352912006363)*0.5**o
        ref[2,3,2,1]=(-0.514029863634)/(o+1.)+(-0.371308314996)+(0.535262244127)*0.5**o
        ref[2,4,0,0]=(0.280547945378)/(o+1.)+(-0.942055103948)+(-0.511771630262)*0.5**o
        ref[2,4,0,1]=(-0.867900669381)/(o+1.)+(0.50274078185)+(0.421559088561)*0.5**o
        ref[2,4,1,0]=(1.36258475686)/(o+1.)+(-0.30949601115)+(0.681235817735)*0.5**o
        ref[2,4,1,1]=(0.462688810574)/(o+1.)+(-0.206018280315)+(0.432866305507)*0.5**o
        ref[2,4,2,0]=(-1.03683778617)/(o+1.)+(0.293687179358)+(-0.956901622446)*0.5**o
        ref[2,4,2,1]=(0.119667660807)/(o+1.)+(0.462235426749)+(-0.16700370288)*0.5**o
        ref[3,0,0,0]=(-0.632782121804)/(o+1.)+(0.646441008172)+(0.64046380712)*0.5**o
        ref[3,0,0,1]=(0.204432444813)/(o+1.)+(0.685194121272)+(-0.226057776579)*0.5**o
        ref[3,0,1,0]=(0.642287443449)/(o+1.)+(-0.563816218608)+(-0.358031084019)*0.5**o
        ref[3,0,1,1]=(0.37482915485)/(o+1.)+(0.371462612656)+(-0.312175299678)*0.5**o
        ref[3,0,2,0]=(-1.37297151652)/(o+1.)+(0.476363978278)+(0.367207247645)*0.5**o
        ref[3,0,2,1]=(0.198749853995)/(o+1.)+(0.268733197751)+(-0.806787228991)*0.5**o
        ref[3,1,0,0]=(0.552182836914)/(o+1.)+(-0.317343341965)+(-0.268344201624)*0.5**o
        ref[3,1,0,1]=(-0.352738764079)/(o+1.)+(0.262975892585)+(-0.870410307785)*0.5**o
        ref[3,1,1,0]=(-0.153389711949)/(o+1.)+(-0.0102924079393)+(-0.178465888488)*0.5**o
        ref[3,1,1,1]=(-0.824822901439)/(o+1.)+(0.424615107645)+(0.113559334778)*0.5**o
        ref[3,1,2,0]=(1.25561690345)/(o+1.)+(-0.71060060265)+(-0.574716699633)*0.5**o
        ref[3,1,2,1]=(1.12549317304)/(o+1.)+(0.257461566285)+(-0.0193167048813)*0.5**o
        ref[3,2,0,0]=(-1.07295374259)/(o+1.)+(-0.437779947243)+(0.0777423524941)*0.5**o
        ref[3,2,0,1]=(-1.08247290366)/(o+1.)+(0.019142043881)+(0.80859948321)*0.5**o
        ref[3,2,1,0]=(0.295157671875)/(o+1.)+(-0.346357625892)+(-0.0144328272818)*0.5**o
        ref[3,2,1,1]=(0.054298290938)/(o+1.)+(-1.11979529396)+(0.377945289524)*0.5**o
        ref[3,2,2,0]=(-0.0332954342171)/(o+1.)+(0.887711453071)+(-0.733150287394)*0.5**o
        ref[3,2,2,1]=(0.621153501787)/(o+1.)+(-0.264880999958)+(-0.155229801255)*0.5**o
        ref[3,3,0,0]=(1.33353990272)/(o+1.)+(0.713230419758)+(0.60752130978)*0.5**o
        ref[3,3,0,1]=(-0.0545416714491)/(o+1.)+(0.218078495198)+(0.755049219363)*0.5**o
        ref[3,3,1,0]=(1.04668533538)/(o+1.)+(-0.340927357635)+(-0.39765237082)*0.5**o
        ref[3,3,1,1]=(1.66789191718)/(o+1.)+(-0.784743625765)+(0.350497253938)*0.5**o
        ref[3,3,2,0]=(0.116457701698)/(o+1.)+(-0.280069495076)+(-0.952544750261)*0.5**o
        ref[3,3,2,1]=(-0.0133835495768)/(o+1.)+(-0.0108878424651)+(-0.0232941915849)*0.5**o
        ref[3,4,0,0]=(-0.415035738369)/(o+1.)+(-0.221351326547)+(0.0279130167137)*0.5**o
        ref[3,4,0,1]=(0.182464389723)/(o+1.)+(-0.224847796415)+(0.0423746259802)*0.5**o
        ref[3,4,1,0]=(0.595827544738)/(o+1.)+(-0.202613125922)+(-0.695685046398)*0.5**o
        ref[3,4,1,1]=(-0.99448823611)/(o+1.)+(0.39951522517)+(0.864599640696)*0.5**o
        ref[3,4,2,0]=(0.944328366501)/(o+1.)+(0.852563490353)+(-0.893419673448)*0.5**o
        ref[3,4,2,1]=(-0.813392164634)/(o+1.)+(-0.841253216593)+(0.480643160777)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.241972460279)*x[0]+(0.539729925673)*x[1]
        ref=(-0.11832904353)/(o+1.)+(0.649911364492)+(-0.399791299503)*0.5**o
      else:
        arg=(0.53919797656)*x[0]+(0.379010664884)*x[1]+(0.274416064529)*x[2]
        ref=(0.307947301472)/(o+1.)+(0.120674157748)+(0.643329089005)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.84849017056)*x[0]+(-0.161962845202)*x[1]
        arg[1]=(-0.794717919135)*x[0]+(1.16003773358)*x[1]
        ref[0]=(0.780514448202)/(o+1.)+(-0.34628958021)+(0.598592037576)*0.5**o
        ref[1]=(0.570158156278)/(o+1.)+(0.241192316704)+(-0.687222975237)*0.5**o
      else:
        arg[0]=(-0.637352163634)*x[0]+(1.49477462643)*x[1]+(-0.284018740238)*x[2]
        arg[1]=(0.125937459904)*x[0]+(-0.206499869891)*x[1]+(-0.342856240345)*x[2]
        ref[0]=(-0.274486314451)/(o+1.)+(0.379737553853)+(0.0884149293071)*0.5**o
        ref[1]=(-0.78452814021)/(o+1.)+(-0.122017410137)+(0.605144310151)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.440868762581)*x[0]+(-0.216223666616)*x[1]
        arg[0,1]=(-1.78807355591)*x[0]+(-0.208278123838)*x[1]
        arg[0,2]=(-0.448066998103)*x[0]+(-0.819815507826)*x[1]
        arg[0,3]=(1.27247881689)*x[0]+(0.0653439567221)*x[1]
        arg[0,4]=(-1.00344990223)*x[0]+(-0.291263790583)*x[1]
        arg[1,0]=(0.440080423838)*x[0]+(0.0356836558113)*x[1]
        arg[1,1]=(-1.49545340091)*x[0]+(-0.133615590211)*x[1]
        arg[1,2]=(1.05947638385)*x[0]+(0.408685904238)*x[1]
        arg[1,3]=(-0.170309086605)*x[0]+(0.660523970285)*x[1]
        arg[1,4]=(0.764323678475)*x[0]+(-0.529415080303)*x[1]
        arg[2,0]=(0.55270081325)*x[0]+(-1.60630146231)*x[1]
        arg[2,1]=(1.57419330248)*x[0]+(-1.24644288706)*x[1]
        arg[2,2]=(0.0416757238363)*x[0]+(0.0247379738465)*x[1]
        arg[2,3]=(1.0200370393)*x[0]+(0.364209137693)*x[1]
        arg[2,4]=(-1.21858490389)*x[0]+(0.523056254546)*x[1]
        arg[3,0]=(0.0524205820842)*x[0]+(1.65953217854)*x[1]
        arg[3,1]=(1.21824883395)*x[0]+(0.905190834998)*x[1]
        arg[3,2]=(0.226344434525)*x[0]+(-0.735349531462)*x[1]
        arg[3,3]=(-1.00295245717)*x[0]+(-0.719605376499)*x[1]
        arg[3,4]=(-0.446228759693)*x[0]+(1.23441974277)*x[1]
        ref[0,0]=(0.347328582602)/(o+1.)+(-0.520484463519)+(0.0365479152384)*0.5**o
        ref[0,1]=(-0.274564442352)/(o+1.)+(-0.413557210678)+(-0.894672816039)*0.5**o
        ref[0,2]=(0.0720644688954)/(o+1.)+(-0.906636966188)+(0.473326957552)*0.5**o
        ref[0,3]=(-0.747749341777)/(o+1.)+(0.730249972378)+(0.625072170633)*0.5**o
        ref[0,4]=(-0.194031116169)/(o+1.)+(-0.432063325829)+(-0.236555924983)*0.5**o
        ref[1,0]=(0.807081538033)/(o+1.)+(-0.612266650007)+(0.893215841629)*0.5**o
        ref[1,1]=(-0.662516518957)/(o+1.)+(-0.176318925886)+(-0.61391462039)*0.5**o
        ref[1,2]=(-0.570345636156)/(o+1.)+(0.791877954963)+(0.454752014319)*0.5**o
        ref[1,3]=(-0.312400910795)/(o+1.)+(0.212935137466)+(0.376745519544)*0.5**o
        ref[1,4]=(-0.764680414264)/(o+1.)+(0.174215486579)+(0.651158039278)*0.5**o
        ref[2,0]=(-0.837942771642)/(o+1.)+(-0.187938486702)+(0.160219095985)*0.5**o
        ref[2,1]=(-0.672421039761)/(o+1.)+(0.17384002882)+(0.652491397549)*0.5**o
        ref[2,2]=(-0.656187949922)/(o+1.)+(0.559888736121)+(-0.397175824637)*0.5**o
        ref[2,3]=(0.259558368195)/(o+1.)+(0.55148644978)+(0.0217149092406)*0.5**o
        ref[2,4]=(-0.188397537901)/(o+1.)+(-0.0561706170512)+(-0.394789877345)*0.5**o
        ref[3,0]=(0.91055628617)/(o+1.)+(0.574211274098)+(-0.347026073739)*0.5**o
        ref[3,1]=(0.29225509237)/(o+1.)+(0.64447705482)+(0.542230466941)*0.5**o
        ref[3,2]=(0.121065624384)/(o+1.)+(0.0316495277836)+(-0.693369776888)*0.5**o
        ref[3,3]=(-0.943523334976)/(o+1.)+(-0.266302644031)+(-0.246429210629)*0.5**o
        ref[3,4]=(0.714915399455)/(o+1.)+(0.459963131994)+(-0.846650680369)*0.5**o
      else:
        arg[0,0]=(-0.493869713454)*x[0]+(-0.381969413515)*x[1]+(1.59862368848)*x[2]
        arg[0,1]=(-0.581245543547)*x[0]+(0.961117271889)*x[1]+(-0.158073515343)*x[2]
        arg[0,2]=(-0.635283143641)*x[0]+(0.695573236987)*x[1]+(-0.719180832278)*x[2]
        arg[0,3]=(-0.737780596733)*x[0]+(-1.11171283352)*x[1]+(0.459954346315)*x[2]
        arg[0,4]=(1.12522065074)*x[0]+(-1.08427904854)*x[1]+(-0.715679673556)*x[2]
        arg[1,0]=(0.972468186181)*x[0]+(-0.211014530435)*x[1]+(-0.734414099002)*x[2]
        arg[1,1]=(-1.67337704748)*x[0]+(0.871096213089)*x[1]+(1.51063313223)*x[2]
        arg[1,2]=(-0.222013007834)*x[0]+(1.46474508326)*x[1]+(-0.979243908862)*x[2]
        arg[1,3]=(1.21195281087)*x[0]+(0.274082389329)*x[1]+(1.56144436465)*x[2]
        arg[1,4]=(1.35812865884)*x[0]+(1.25819798853)*x[1]+(0.895334048395)*x[2]
        arg[2,0]=(-0.167404344437)*x[0]+(0.0710591934315)*x[1]+(-0.124469683665)*x[2]
        arg[2,1]=(-0.492240632996)*x[0]+(-0.317251117382)*x[1]+(0.0195461703529)*x[2]
        arg[2,2]=(-0.277590266225)*x[0]+(-1.65859476899)*x[1]+(-0.0100216405668)*x[2]
        arg[2,3]=(-0.0701281298455)*x[0]+(0.502137254703)*x[1]+(0.701220923443)*x[2]
        arg[2,4]=(-0.72186096407)*x[0]+(-0.900487568179)*x[1]+(-0.735573922826)*x[2]
        arg[3,0]=(-0.0677324474129)*x[0]+(-0.649963209718)*x[1]+(-0.770861099726)*x[2]
        arg[3,1]=(0.563588189481)*x[0]+(0.714614932251)*x[1]+(-0.0957764632886)*x[2]
        arg[3,2]=(1.13337395062)*x[0]+(0.507556316073)*x[1]+(-0.196759085949)*x[2]
        arg[3,3]=(-1.2636883168)*x[0]+(-0.811660344076)*x[1]+(0.645462052598)*x[2]
        arg[3,4]=(-0.104175186987)*x[0]+(-0.929176038155)*x[1]+(1.07530379856)*x[2]
        ref[0,0]=(0.84538674973)/(o+1.)+(0.0569713556632)+(-0.236544899547)*0.5**o
        ref[0,1]=(-0.307689589699)/(o+1.)+(0.245040337642)+(0.0394071274131)*0.5**o
        ref[0,2]=(0.955244736357)/(o+1.)+(-0.341771136504)+(-0.930593202281)*0.5**o
        ref[0,3]=(-0.353490461722)/(o+1.)+(-0.476228433901)+(-0.083591754415)*0.5**o
        ref[0,4]=(-1.31220485482)/(o+1.)+(0.0466426656676)+(0.544181452128)*0.5**o
        ref[1,0]=(0.22773475414)/(o+1.)+(-0.488454801654)+(0.776214405914)*0.5**o
        ref[1,1]=(1.50688678707)/(o+1.)+(0.0851415404695)+(-0.96881757017)*0.5**o
        ref[1,2]=(0.313471097102)/(o+1.)+(-0.0875594518156)+(0.125135973096)*0.5**o
        ref[1,3]=(0.269521220179)/(o+1.)+(1.1192007195)+(0.539556905661)*0.5**o
        ref[1,4]=(1.91499876064)/(o+1.)+(0.509063797893)+(0.578534339343)*0.5**o
        ref[2,0]=(0.941352651562)/(o+1.)+(-0.966199402186)+(0.770231318139)*0.5**o
        ref[2,1]=(-0.141490823562)/(o+1.)+(-0.20919987022)+(-0.230055016023)*0.5**o
        ref[2,2]=(-0.916148047971)/(o+1.)+(-0.371740600192)+(-0.286577427426)*0.5**o
        ref[2,3]=(0.229516219331)/(o+1.)+(0.274790532736)+(0.354132763498)*0.5**o
        ref[2,4]=(-1.19120612619)/(o+1.)+(-0.139859859385)+(-0.88699661011)*0.5**o
        ref[3,0]=(-1.48630121465)/(o+1.)+(0.0744166251063)+(-0.151088792417)*0.5**o
        ref[3,1]=(-0.335583560513)/(o+1.)+(0.277436410841)+(0.963137397276)*0.5**o
        ref[3,2]=(-0.673942688644)/(o+1.)+(0.70077509905)+(0.716563671283)*0.5**o
        ref[3,3]=(-0.148394092795)/(o+1.)+(-0.237190239072)+(-0.807112037336)*0.5**o
        ref[3,4]=(0.529983886211)/(o+1.)+(-0.371055083588)+(0.254078854382)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.661725715165)*x[0]+(0.219720717581)*x[1]
        arg[0,0,1]=(-0.642240082304)*x[0]+(-0.279652756776)*x[1]
        arg[0,1,0]=(0.0204235813151)*x[0]+(-0.374873249933)*x[1]
        arg[0,1,1]=(1.4418481252)*x[0]+(0.205656834772)*x[1]
        arg[1,0,0]=(0.65992179951)*x[0]+(-1.04887875458)*x[1]
        arg[1,0,1]=(-0.195480216199)*x[0]+(-0.373905275039)*x[1]
        arg[1,1,0]=(0.781251543352)*x[0]+(0.0416248109536)*x[1]
        arg[1,1,1]=(-0.925751319189)*x[0]+(-0.113992910046)*x[1]
        arg[2,0,0]=(-0.0767414065227)*x[0]+(-0.156682840973)*x[1]
        arg[2,0,1]=(0.688457272529)*x[0]+(0.524916336046)*x[1]
        arg[2,1,0]=(0.753722830084)*x[0]+(-0.890325839466)*x[1]
        arg[2,1,1]=(0.730789290145)*x[0]+(-0.278986703716)*x[1]
        arg[3,0,0]=(-0.287376796439)*x[0]+(-0.153987233224)*x[1]
        arg[3,0,1]=(-1.15687435255)*x[0]+(0.250005890238)*x[1]
        arg[3,1,0]=(-0.845312830629)*x[0]+(0.256211962352)*x[1]
        arg[3,1,1]=(-0.574311820519)*x[0]+(-0.440297035998)*x[1]
        arg[4,0,0]=(-0.360607647308)*x[0]+(-1.29380437557)*x[1]
        arg[4,0,1]=(1.32019433485)*x[0]+(0.295255956618)*x[1]
        arg[4,1,0]=(-1.64162941972)*x[0]+(0.117482372162)*x[1]
        arg[4,1,1]=(-0.303103646619)*x[0]+(-0.88458982681)*x[1]
        arg[5,0,0]=(-0.666761879038)*x[0]+(-0.735786278877)*x[1]
        arg[5,0,1]=(0.0892896382617)*x[0]+(-0.45478872773)*x[1]
        arg[5,1,0]=(-0.50985624415)*x[0]+(-0.400022939567)*x[1]
        arg[5,1,1]=(1.39240706368)*x[0]+(0.0149740822664)*x[1]
        ref[0,0,0]=(-0.114834538488)/(o+1.)+(0.240621006484)+(0.515038958264)*0.5**o
        ref[0,0,1]=(0.588698337077)/(o+1.)+(-0.702256377233)+(-0.10607842169)*0.5**o
        ref[0,1,0]=(0.13849818224)/(o+1.)+(-0.277555277991)+(0.0621627051247)*0.5**o
        ref[0,1,1]=(-0.0878009521163)/(o+1.)+(0.630168348947)+(0.474969214192)*0.5**o
        ref[1,0,0]=(-0.696600538111)/(o+1.)+(-0.343318260912)+(0.994280104861)*0.5**o
        ref[1,0,1]=(-0.0887026688527)/(o+1.)+(-0.385753227979)+(0.290823633572)*0.5**o
        ref[1,1,0]=(0.831408608956)/(o+1.)+(-0.242985671853)+(0.477439089057)*0.5**o
        ref[1,1,1]=(-0.621059053191)/(o+1.)+(-0.0700203481241)+(-0.278644479797)*0.5**o
        ref[2,0,0]=(0.7890720399)/(o+1.)+(-0.539721196393)+(0.0569461053905)*0.5**o
        ref[2,0,1]=(-0.228628151401)/(o+1.)+(0.826946074171)+(-0.211890388366)*0.5**o
        ref[2,1,0]=(-0.837617325493)/(o+1.)+(0.289656653751)+(0.121701008608)*0.5**o
        ref[2,1,1]=(0.246577988625)/(o+1.)+(0.0512227634376)+(0.102779070928)*0.5**o
        ref[3,0,0]=(-0.0109584684212)/(o+1.)+(-0.334888666171)+(0.239371771102)*0.5**o
        ref[3,0,1]=(-0.125987676419)/(o+1.)+(-0.170179004032)+(-0.440522777827)*0.5**o
        ref[3,1,0]=(-0.354093141636)/(o+1.)+(-0.177294275919)+(0.119580825198)*0.5**o
        ref[3,1,1]=(-0.746761075412)/(o+1.)+(0.192068760731)+(-0.651985302566)*0.5**o
        ref[4,0,0]=(-0.42411673219)/(o+1.)+(-0.579162944937)+(-0.0719694008158)*0.5**o
        ref[4,0,1]=(0.169881169584)/(o+1.)+(0.483614029132)+(0.478341063616)*0.5**o
        ref[4,1,0]=(0.943567836264)/(o+1.)+(-0.898782941264)+(-0.670149001297)*0.5**o
        ref[4,1,1]=(-0.703610807936)/(o+1.)+(0.225342628657)+(-0.934767922806)*0.5**o
        ref[5,0,0]=(-0.788792225213)/(o+1.)+(-0.1768857093)+(-0.259984514102)*0.5**o
        ref[5,0,1]=(-0.788708879899)/(o+1.)+(-0.202288491141)+(0.827786772711)*0.5**o
        ref[5,1,0]=(-0.823081809707)/(o+1.)+(0.274720963093)+(-0.636239300197)*0.5**o
        ref[5,1,1]=(0.498922549807)/(o+1.)+(0.0129388257746)+(0.882580944588)*0.5**o
      else:
        arg[0,0,0]=(-0.426613654844)*x[0]+(-1.01433980801)*x[1]+(0.564621874916)*x[2]
        arg[0,0,1]=(-0.438521968882)*x[0]+(1.6033775267)*x[1]+(0.691349995832)*x[2]
        arg[0,1,0]=(-0.04337513731)*x[0]+(0.142927668234)*x[1]+(0.506581723199)*x[2]
        arg[0,1,1]=(-0.120598868162)*x[0]+(-0.213666787813)*x[1]+(-1.04221679332)*x[2]
        arg[1,0,0]=(0.153696490712)*x[0]+(-0.122516529815)*x[1]+(-1.08837272712)*x[2]
        arg[1,0,1]=(0.410706464237)*x[0]+(0.720079363315)*x[1]+(-0.767790357539)*x[2]
        arg[1,1,0]=(-1.03726428173)*x[0]+(-0.316809853003)*x[1]+(-0.780405868082)*x[2]
        arg[1,1,1]=(-0.427547967216)*x[0]+(-0.897766643609)*x[1]+(-1.56913552404)*x[2]
        arg[2,0,0]=(-0.282789248898)*x[0]+(-0.110846936031)*x[1]+(-1.4329391906)*x[2]
        arg[2,0,1]=(1.46866588617)*x[0]+(-1.23022501959)*x[1]+(-0.3123390018)*x[2]
        arg[2,1,0]=(-0.314115718621)*x[0]+(1.7333934438)*x[1]+(0.833572700726)*x[2]
        arg[2,1,1]=(0.464246462638)*x[0]+(0.267484068646)*x[1]+(0.342009204787)*x[2]
        arg[3,0,0]=(0.578853099163)*x[0]+(-0.0453855508465)*x[1]+(-0.0751578475516)*x[2]
        arg[3,0,1]=(-0.526734485752)*x[0]+(-1.03286310016)*x[1]+(-0.398986583917)*x[2]
        arg[3,1,0]=(0.0239088277798)*x[0]+(0.205105470192)*x[1]+(-1.04986349166)*x[2]
        arg[3,1,1]=(0.130893744001)*x[0]+(-0.273702140669)*x[1]+(-0.0804222415128)*x[2]
        arg[4,0,0]=(-0.822661618584)*x[0]+(-0.267189204519)*x[1]+(-0.290013766971)*x[2]
        arg[4,0,1]=(-1.46759823734)*x[0]+(0.317465945316)*x[1]+(0.108284317752)*x[2]
        arg[4,1,0]=(-1.3920990386)*x[0]+(-1.17823778694)*x[1]+(-0.18366434307)*x[2]
        arg[4,1,1]=(-0.0915496209516)*x[0]+(-0.367624609832)*x[1]+(1.30630257323)*x[2]
        arg[5,0,0]=(0.117840830455)*x[0]+(-0.667875433194)*x[1]+(-1.59716688455)*x[2]
        arg[5,0,1]=(-0.818693227605)*x[0]+(0.483842075248)*x[1]+(-0.408373205535)*x[2]
        arg[5,1,0]=(1.36906770301)*x[0]+(0.494926179999)*x[1]+(0.717041707262)*x[2]
        arg[5,1,1]=(0.819757116081)*x[0]+(0.0982458749015)*x[1]+(0.275080391599)*x[2]
        ref[0,0,0]=(0.236560699949)/(o+1.)+(-0.392784388084)+(-0.327323511722)*0.5**o
        ref[0,0,1]=(1.32491522045)/(o+1.)+(0.525146863711)+(-0.519003394222)*0.5**o
        ref[0,1,0]=(0.433749865563)/(o+1.)+(0.23007126008)+(-0.287758131599)*0.5**o
        ref[0,1,1]=(-0.0563021060686)/(o+1.)+(-0.403921160703)+(-0.512338021825)*0.5**o
        ref[1,0,0]=(-0.170496965963)/(o+1.)+(-0.914421777259)+(0.942147754258)*0.5**o
        ref[1,0,1]=(-0.372707483726)/(o+1.)+(0.636041158872)+(-0.536379364005)*0.5**o
        ref[1,1,0]=(-0.211529878364)/(o+1.)+(-0.912151985776)+(-0.0986461529033)*0.5**o
        ref[1,1,1]=(-1.32533998437)/(o+1.)+(-0.685030198386)+(-0.19904975372)*0.5**o
        ref[2,0,0]=(0.152048902469)/(o+1.)+(-0.88880590209)+(-0.201012473813)*0.5**o
        ref[2,0,1]=(-0.748900052246)/(o+1.)+(-0.0226120471998)+(0.720226011425)*0.5**o
        ref[2,1,0]=(0.77797707941)/(o+1.)+(1.01635055442)+(-0.557827762351)*0.5**o
        ref[2,1,1]=(1.08969334716)/(o+1.)+(-0.404644728844)+(0.793335846595)*0.5**o
        ref[3,0,0]=(1.56498448669)/(o+1.)+(-0.699967188685)+(0.293259591449)*0.5**o
        ref[3,0,1]=(-0.620133427128)/(o+1.)+(-0.550037413688)+(-0.238375915329)*0.5**o
        ref[3,1,0]=(-0.156380586288)/(o+1.)+(-0.0154369973565)+(-0.633594612687)*0.5**o
        ref[3,1,1]=(-0.352989559326)/(o+1.)+(-0.290547031627)+(0.710852984398)*0.5**o
        ref[4,0,0]=(-0.147579361355)/(o+1.)+(-0.615173424765)+(-0.00193837918766)*0.5**o
        ref[4,0,1]=(0.598152345277)/(o+1.)+(-0.540248606426)+(-0.559503106696)*0.5**o
        ref[4,1,0]=(-1.17182448825)/(o+1.)+(-0.590551681576)+(-0.401073317208)*0.5**o
        ref[4,1,1]=(0.837504141173)/(o+1.)+(0.0807414267186)+(-0.15185865216)*0.5**o
        ref[5,0,0]=(-0.646601115533)/(o+1.)+(-1.20895265273)+(0.9173049337)*0.5**o
        ref[5,0,1]=(1.26626110052)/(o+1.)+(-0.56471923973)+(-0.880046978955)*0.5**o
        ref[5,1,0]=(-0.069063169298)/(o+1.)+(0.883536796964)+(0.883025165644)*0.5**o
        ref[5,1,1]=(1.05999338993)/(o+1.)+(-0.392519102893)+(0.918128198439)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.257898226293)*x[0]+(-0.01230083314)*x[1]
        arg[0,0,0,1]=(0.0531183070647)*x[0]+(-0.170192287836)*x[1]
        arg[0,0,1,0]=(-0.364644940864)*x[0]+(1.04227651446)*x[1]
        arg[0,0,1,1]=(-0.894334108203)*x[0]+(0.151390331211)*x[1]
        arg[0,0,2,0]=(1.03825401708)*x[0]+(-0.882136097149)*x[1]
        arg[0,0,2,1]=(-1.05304910155)*x[0]+(1.06866311708)*x[1]
        arg[0,1,0,0]=(1.54742542159)*x[0]+(0.941778348482)*x[1]
        arg[0,1,0,1]=(1.1797380956)*x[0]+(-0.296417138234)*x[1]
        arg[0,1,1,0]=(-1.09196028027)*x[0]+(-0.783377624006)*x[1]
        arg[0,1,1,1]=(0.370952191443)*x[0]+(-0.0854141858991)*x[1]
        arg[0,1,2,0]=(-1.23552368139)*x[0]+(1.23848088087)*x[1]
        arg[0,1,2,1]=(0.216345350114)*x[0]+(-0.136524468427)*x[1]
        arg[0,2,0,0]=(-1.21206901263)*x[0]+(0.804219582144)*x[1]
        arg[0,2,0,1]=(0.192956526208)*x[0]+(0.588417678858)*x[1]
        arg[0,2,1,0]=(-0.940175603742)*x[0]+(0.329047260784)*x[1]
        arg[0,2,1,1]=(-0.138597387792)*x[0]+(0.398298326223)*x[1]
        arg[0,2,2,0]=(0.10089766824)*x[0]+(1.25459493298)*x[1]
        arg[0,2,2,1]=(-0.401163525883)*x[0]+(0.276542765282)*x[1]
        arg[0,3,0,0]=(0.252156795268)*x[0]+(-0.50223063539)*x[1]
        arg[0,3,0,1]=(0.466244610905)*x[0]+(-0.587025224301)*x[1]
        arg[0,3,1,0]=(-0.268082012713)*x[0]+(-1.28874675764)*x[1]
        arg[0,3,1,1]=(1.10867463519)*x[0]+(-0.165500912864)*x[1]
        arg[0,3,2,0]=(-1.94681908266)*x[0]+(0.166854108476)*x[1]
        arg[0,3,2,1]=(-0.537150417922)*x[0]+(0.0872851457131)*x[1]
        arg[0,4,0,0]=(-0.376548581807)*x[0]+(-0.245201133388)*x[1]
        arg[0,4,0,1]=(-0.135287690767)*x[0]+(0.750376238942)*x[1]
        arg[0,4,1,0]=(-0.0240491596907)*x[0]+(-1.21337088424)*x[1]
        arg[0,4,1,1]=(0.699215733397)*x[0]+(0.0596356032853)*x[1]
        arg[0,4,2,0]=(-0.521316010702)*x[0]+(-1.75309802615)*x[1]
        arg[0,4,2,1]=(0.403640172305)*x[0]+(1.41465827077)*x[1]
        arg[1,0,0,0]=(-1.01538089315)*x[0]+(-1.14785578909)*x[1]
        arg[1,0,0,1]=(1.10045201832)*x[0]+(0.630761728795)*x[1]
        arg[1,0,1,0]=(0.137673646147)*x[0]+(-0.387129090253)*x[1]
        arg[1,0,1,1]=(-1.04588287863)*x[0]+(-1.22433029859)*x[1]
        arg[1,0,2,0]=(-0.567830263535)*x[0]+(-0.405808616295)*x[1]
        arg[1,0,2,1]=(0.139326102495)*x[0]+(-1.93153384363)*x[1]
        arg[1,1,0,0]=(-0.638524762921)*x[0]+(0.506405751653)*x[1]
        arg[1,1,0,1]=(0.130775153144)*x[0]+(0.531317202208)*x[1]
        arg[1,1,1,0]=(0.557476206038)*x[0]+(0.447588676184)*x[1]
        arg[1,1,1,1]=(-0.538795916875)*x[0]+(-0.593869493531)*x[1]
        arg[1,1,2,0]=(-1.09827483942)*x[0]+(0.376431061876)*x[1]
        arg[1,1,2,1]=(-0.0074505620387)*x[0]+(-0.798645151486)*x[1]
        arg[1,2,0,0]=(0.251283911459)*x[0]+(0.194595071813)*x[1]
        arg[1,2,0,1]=(1.00482763799)*x[0]+(-1.00059863394)*x[1]
        arg[1,2,1,0]=(-0.453095434684)*x[0]+(-0.513922011418)*x[1]
        arg[1,2,1,1]=(0.558548837484)*x[0]+(0.0696328406825)*x[1]
        arg[1,2,2,0]=(-1.20625552545)*x[0]+(-0.151371710692)*x[1]
        arg[1,2,2,1]=(0.100262209656)*x[0]+(-0.314628804179)*x[1]
        arg[1,3,0,0]=(0.947063141882)*x[0]+(-0.297271200562)*x[1]
        arg[1,3,0,1]=(0.0760277251261)*x[0]+(-1.17982146387)*x[1]
        arg[1,3,1,0]=(-1.46946063969)*x[0]+(-0.615302663567)*x[1]
        arg[1,3,1,1]=(0.419747293729)*x[0]+(-0.629842531949)*x[1]
        arg[1,3,2,0]=(-1.44607954917)*x[0]+(0.520024093924)*x[1]
        arg[1,3,2,1]=(-1.01556392596)*x[0]+(-0.307173811354)*x[1]
        arg[1,4,0,0]=(-0.164339335244)*x[0]+(0.390736414982)*x[1]
        arg[1,4,0,1]=(-0.266932380288)*x[0]+(-0.731533848078)*x[1]
        arg[1,4,1,0]=(-0.903825489055)*x[0]+(-1.81564784421)*x[1]
        arg[1,4,1,1]=(-0.426008882665)*x[0]+(-1.09792948908)*x[1]
        arg[1,4,2,0]=(-1.32984927589)*x[0]+(1.58196302541)*x[1]
        arg[1,4,2,1]=(-0.0581032806145)*x[0]+(-0.323073355302)*x[1]
        arg[2,0,0,0]=(-0.744014251373)*x[0]+(-0.818253254872)*x[1]
        arg[2,0,0,1]=(0.368656965396)*x[0]+(-0.0864646380237)*x[1]
        arg[2,0,1,0]=(-0.112428845773)*x[0]+(0.109170104421)*x[1]
        arg[2,0,1,1]=(0.0682241548911)*x[0]+(-1.65999482283)*x[1]
        arg[2,0,2,0]=(0.212777058993)*x[0]+(0.127163480945)*x[1]
        arg[2,0,2,1]=(-0.256989443356)*x[0]+(0.183206687414)*x[1]
        arg[2,1,0,0]=(0.578997994373)*x[0]+(0.413087453689)*x[1]
        arg[2,1,0,1]=(0.886532090807)*x[0]+(0.330645525532)*x[1]
        arg[2,1,1,0]=(0.18224742814)*x[0]+(-0.713166591262)*x[1]
        arg[2,1,1,1]=(-0.25218296706)*x[0]+(0.736797897014)*x[1]
        arg[2,1,2,0]=(0.885270371408)*x[0]+(-1.35641397127)*x[1]
        arg[2,1,2,1]=(-0.963214290879)*x[0]+(0.348421598509)*x[1]
        arg[2,2,0,0]=(0.884120745351)*x[0]+(0.956857507884)*x[1]
        arg[2,2,0,1]=(-0.00795298517561)*x[0]+(0.51243740322)*x[1]
        arg[2,2,1,0]=(-0.563655252975)*x[0]+(-0.716305131721)*x[1]
        arg[2,2,1,1]=(0.648243491623)*x[0]+(-0.142239786485)*x[1]
        arg[2,2,2,0]=(1.50677848542)*x[0]+(-0.359722006818)*x[1]
        arg[2,2,2,1]=(-0.522592086302)*x[0]+(-0.18544927497)*x[1]
        arg[2,3,0,0]=(0.0822787762966)*x[0]+(0.196309662733)*x[1]
        arg[2,3,0,1]=(0.232573258627)*x[0]+(-0.0597976775432)*x[1]
        arg[2,3,1,0]=(1.50895737052)*x[0]+(0.546625934312)*x[1]
        arg[2,3,1,1]=(-1.64023373286)*x[0]+(0.220053261777)*x[1]
        arg[2,3,2,0]=(1.5897208842)*x[0]+(-0.0355885594992)*x[1]
        arg[2,3,2,1]=(-1.52822349982)*x[0]+(-0.0906934717656)*x[1]
        arg[2,4,0,0]=(0.496696582601)*x[0]+(-0.303888675059)*x[1]
        arg[2,4,0,1]=(-0.0819586112177)*x[0]+(-0.0449047294378)*x[1]
        arg[2,4,1,0]=(-0.678530758513)*x[0]+(-1.2173963447)*x[1]
        arg[2,4,1,1]=(-0.645450475268)*x[0]+(0.356771113426)*x[1]
        arg[2,4,2,0]=(-0.486643021867)*x[0]+(0.977763869689)*x[1]
        arg[2,4,2,1]=(-0.458991807397)*x[0]+(1.80038432923)*x[1]
        arg[3,0,0,0]=(0.270795948755)*x[0]+(-0.691512096338)*x[1]
        arg[3,0,0,1]=(-0.756717130099)*x[0]+(1.16189640534)*x[1]
        arg[3,0,1,0]=(0.967554346699)*x[0]+(-0.272119008963)*x[1]
        arg[3,0,1,1]=(1.14296491195)*x[0]+(-0.707175662379)*x[1]
        arg[3,0,2,0]=(1.15665030692)*x[0]+(0.414803389409)*x[1]
        arg[3,0,2,1]=(1.38997274955)*x[0]+(1.07065568192)*x[1]
        arg[3,1,0,0]=(-0.623998925511)*x[0]+(0.375545806421)*x[1]
        arg[3,1,0,1]=(-0.363431948398)*x[0]+(-0.183100417297)*x[1]
        arg[3,1,1,0]=(0.0435866301729)*x[0]+(-0.557137572068)*x[1]
        arg[3,1,1,1]=(-0.253225217468)*x[0]+(-0.0657932188329)*x[1]
        arg[3,1,2,0]=(0.147700921064)*x[0]+(0.681624563015)*x[1]
        arg[3,1,2,1]=(-0.075394297616)*x[0]+(0.271741245789)*x[1]
        arg[3,2,0,0]=(1.34803044812)*x[0]+(0.87991065609)*x[1]
        arg[3,2,0,1]=(0.0529487135353)*x[0]+(0.806480092352)*x[1]
        arg[3,2,1,0]=(-1.52038624166)*x[0]+(-0.992718138685)*x[1]
        arg[3,2,1,1]=(0.0859852188749)*x[0]+(1.63512596202)*x[1]
        arg[3,2,2,0]=(1.46925148575)*x[0]+(-1.60383564005)*x[1]
        arg[3,2,2,1]=(0.114463395473)*x[0]+(-1.0186377376)*x[1]
        arg[3,3,0,0]=(-1.02962502658)*x[0]+(-1.06788543146)*x[1]
        arg[3,3,0,1]=(-0.184391597652)*x[0]+(0.462603936483)*x[1]
        arg[3,3,1,0]=(0.611113335782)*x[0]+(0.282417397043)*x[1]
        arg[3,3,1,1]=(-0.193423873796)*x[0]+(-0.56108614964)*x[1]
        arg[3,3,2,0]=(-0.0290687598224)*x[0]+(-0.460221158417)*x[1]
        arg[3,3,2,1]=(0.784513522959)*x[0]+(-0.484813129206)*x[1]
        arg[3,4,0,0]=(0.423780748664)*x[0]+(-0.344020787894)*x[1]
        arg[3,4,0,1]=(1.10623024451)*x[0]+(-0.775008153301)*x[1]
        arg[3,4,1,0]=(0.554847534291)*x[0]+(-0.854749107999)*x[1]
        arg[3,4,1,1]=(-0.639595693597)*x[0]+(-0.966152813212)*x[1]
        arg[3,4,2,0]=(-0.0172621996023)*x[0]+(0.321815074269)*x[1]
        arg[3,4,2,1]=(0.39605156824)*x[0]+(-0.499138231749)*x[1]
        ref[0,0,0,0]=(0.169892838337)/(o+1.)+(-0.323261281472)+(0.722227117759)*0.5**o
        ref[0,0,0,1]=(-0.899797330597)/(o+1.)+(0.731242549464)+(-0.679761749104)*0.5**o
        ref[0,0,1,0]=(0.948969615551)/(o+1.)+(-0.173478478707)+(0.0756189154575)*0.5**o
        ref[0,0,1,1]=(-0.484419525942)/(o+1.)+(0.287761319651)+(-0.834046890353)*0.5**o
        ref[0,0,2,0]=(-0.275366085384)/(o+1.)+(-0.051636507926)+(0.534757021171)*0.5**o
        ref[0,0,2,1]=(0.338862631582)/(o+1.)+(0.076489529483)+(-0.47622767501)*0.5**o
        ref[0,1,0,0]=(0.960885993642)/(o+1.)+(0.36959166751)+(0.789134441414)*0.5**o
        ref[0,1,0,1]=(-0.101908900511)/(o+1.)+(0.343409131566)+(0.298411594744)*0.5**o
        ref[0,1,1,0]=(-0.50227382988)/(o+1.)+(-0.431549057432)+(-0.509965959533)*0.5**o
        ref[0,1,1,1]=(-0.294216689224)/(o+1.)+(0.293114907564)+(-0.00647512035965)*0.5**o
        ref[0,1,2,0]=(0.330313139948)/(o+1.)+(0.0601559902451)+(-0.447667920962)*0.5**o
        ref[0,1,2,1]=(-0.855325117375)/(o+1.)+(0.455706690618)+(0.0237326178252)*0.5**o
        ref[0,2,0,0]=(0.958948557218)/(o+1.)+(-0.307015973897)+(-0.752766039908)*0.5**o
        ref[0,2,0,1]=(0.622666086232)/(o+1.)+(0.417782845058)+(-0.676857571283)*0.5**o
        ref[0,2,1,0]=(-0.476269525636)/(o+1.)+(0.203518519787)+(-0.541895856898)*0.5**o
        ref[0,2,1,1]=(0.396862410812)/(o+1.)+(-0.318442063423)+(0.499722654464)*0.5**o
        ref[0,2,2,0]=(0.291546173382)/(o+1.)+(0.345688404218)+(0.372569619405)*0.5**o
        ref[0,2,2,1]=(-0.0863432016899)/(o+1.)+(0.296494638577)+(-0.631266836064)*0.5**o
        ref[0,3,0,0]=(0.437179077921)/(o+1.)+(-0.340515851088)+(-0.00622121586734)*0.5**o
        ref[0,3,0,1]=(-0.0670787919003)/(o+1.)+(-0.492735838895)+(0.931769856293)*0.5**o
        ref[0,3,1,0]=(-0.455620785864)/(o+1.)+(-0.866886545624)+(0.632565106759)*0.5**o
        ref[0,3,1,1]=(0.769395985412)/(o+1.)+(-0.303642587154)+(0.781062911223)*0.5**o
        ref[0,3,2,0]=(-0.338160377272)/(o+1.)+(-0.237545833658)+(-0.966712929597)*0.5**o
        ref[0,3,2,1]=(-0.796435534783)/(o+1.)+(0.544922318125)+(-0.743274373676)*0.5**o
        ref[0,4,0,0]=(0.452932120968)/(o+1.)+(-0.512874769603)+(-0.048932296957)*0.5**o
        ref[0,4,0,1]=(0.361579034112)/(o+1.)+(-0.0561290568189)+(0.365767627701)*0.5**o
        ref[0,4,1,0]=(-0.626053161593)/(o+1.)+(-0.761324946295)+(0.911283010254)*0.5**o
        ref[0,4,1,1]=(-0.0455977379835)/(o+1.)+(0.292122388548)+(0.22020429757)*0.5**o
        ref[0,4,2,0]=(-0.881052488675)/(o+1.)+(-0.210171400955)+(-0.973018746263)*0.5**o
        ref[0,4,2,1]=(0.955475719603)/(o+1.)+(0.367959093676)+(0.126904536122)*0.5**o
        ref[1,0,0,0]=(-0.314523133291)/(o+1.)+(-0.649561146707)+(-0.54959125554)*0.5**o
        ref[1,0,0,1]=(0.776726613012)/(o+1.)+(0.0747896120381)+(0.804907910029)*0.5**o
        ref[1,0,1,0]=(0.416309554773)/(o+1.)+(0.0650606132082)+(-0.795886225295)*0.5**o
        ref[1,0,1,1]=(-0.642083920221)/(o+1.)+(-0.518812997818)+(-0.590503261366)*0.5**o
        ref[1,0,2,0]=(0.209400082886)/(o+1.)+(-0.592256267485)+(0.00147357225359)*0.5**o
        ref[1,0,2,1]=(-0.969945221754)/(o+1.)+(-0.364559871358)+(-0.0931427766681)*0.5**o
        ref[1,1,0,0]=(0.941138659497)/(o+1.)+(-0.103485523138)+(-0.866286624489)*0.5**o
        ref[1,1,0,1]=(0.527395399401)/(o+1.)+(0.373358049173)+(-0.612019142395)*0.5**o
        ref[1,1,1,0]=(-0.276942103917)/(o+1.)+(0.624509599487)+(0.0329877871648)*0.5**o
        ref[1,1,1,1]=(-0.153477829221)/(o+1.)+(-0.0616268433725)+(-0.85593389444)*0.5**o
        ref[1,1,2,0]=(0.2685285297)/(o+1.)+(-0.301824551915)+(-0.386723203413)*0.5**o
        ref[1,1,2,1]=(-0.530960074652)/(o+1.)+(0.248621312801)+(-0.772378264474)*0.5**o
        ref[1,2,0,0]=(0.745071931955)/(o+1.)+(-0.149268301427)+(-0.000656345828701)*0.5**o
        ref[1,2,0,1]=(-0.581822428087)/(o+1.)+(0.2315691847)+(0.122913062738)*0.5**o
        ref[1,2,1,0]=(-0.325165099021)/(o+1.)+(-0.414758139139)+(0.187663931198)*0.5**o
        ref[1,2,1,1]=(-0.389683593964)/(o+1.)+(0.265985790556)+(0.485893691018)*0.5**o
        ref[1,2,2,0]=(0.661969003098)/(o+1.)+(-0.514584521701)+(-0.990427195841)*0.5**o
        ref[1,2,2,1]=(-0.0828384828188)/(o+1.)+(-0.482166570368)+(0.832805029032)*0.5**o
        ref[1,3,0,0]=(-0.756020671713)/(o+1.)+(0.627704058228)+(0.150404496577)*0.5**o
        ref[1,3,0,1]=(-0.93648010541)/(o+1.)+(0.11022703478)+(-0.387767702896)*0.5**o
        ref[1,3,1,0]=(-0.995823513997)/(o+1.)+(-0.242072805569)+(-0.604794178119)*0.5**o
        ref[1,3,1,1]=(-0.543071878541)/(o+1.)+(-0.160890620667)+(0.654757881656)*0.5**o
        ref[1,3,2,0]=(0.803184441579)/(o+1.)+(-0.612293875048)+(-0.504652146723)*0.5**o
        ref[1,3,2,1]=(-0.786714439322)/(o+1.)+(-0.175556007517)+(-0.18491128296)*0.5**o
        ref[1,4,0,0]=(-0.302928744298)/(o+1.)+(0.476360987754)+(-0.423396151472)*0.5**o
        ref[1,4,0,1]=(-0.410240602476)/(o+1.)+(-0.49551630458)+(0.40280698327)*0.5**o
        ref[1,4,1,0]=(-0.820677400346)/(o+1.)+(-0.573037933363)+(-0.752720066188)*0.5**o
        ref[1,4,1,1]=(-0.492748923658)/(o+1.)+(-0.600739806799)+(0.170290165507)*0.5**o
        ref[1,4,2,0]=(0.926654770824)/(o+1.)+(0.011282908181)+(-0.697106837662)*0.5**o
        ref[1,4,2,1]=(-0.958477382611)/(o+1.)+(0.618515858902)+(-0.65973097111)*0.5**o
        ref[2,0,0,0]=(-0.15636950624)/(o+1.)+(-0.520670547262)+(-0.36455690548)*0.5**o
        ref[2,0,0,1]=(-0.726424878286)/(o+1.)+(0.697757734621)+(-0.386898263585)*0.5**o
        ref[2,0,1,0]=(-0.055162393508)/(o+1.)+(0.247015646426)+(-0.442127640696)*0.5**o
        ref[2,0,1,1]=(-0.769332288875)/(o+1.)+(-0.0917175733456)+(-0.639003232373)*0.5**o
        ref[2,0,2,0]=(-0.196390038297)/(o+1.)+(0.0977907270548)+(0.340749124126)*0.5**o
        ref[2,0,2,1]=(0.841927481859)/(o+1.)+(-0.489185426146)+(0.0626606144906)*0.5**o
        ref[2,1,0,0]=(0.590018249148)/(o+1.)+(0.378476967182)+(-0.354886735451)*0.5**o
        ref[2,1,0,1]=(0.178495604742)/(o+1.)+(0.233073081069)+(0.572535849459)*0.5**o
        ref[2,1,1,0]=(-0.762671064919)/(o+1.)+(-0.108539686142)+(0.448831274081)*0.5**o
        ref[2,1,1,1]=(0.927591406535)/(o+1.)+(0.0441635703243)+(-0.53130361723)*0.5**o
        ref[2,1,2,0]=(-0.604360706259)/(o+1.)+(-0.134848102332)+(0.402913311063)*0.5**o
        ref[2,1,2,1]=(0.229727627441)/(o+1.)+(-0.323296604293)+(-0.197927111225)*0.5**o
        ref[2,2,0,0]=(0.287478356112)/(o+1.)+(0.762635506201)+(0.0282288847209)*0.5**o
        ref[2,2,0,1]=(0.243518182621)/(o+1.)+(-0.241502069199)+(0.743970373822)*0.5**o
        ref[2,2,1,0]=(-0.369049760378)/(o+1.)+(-0.21941778965)+(-0.472075045017)*0.5**o
        ref[2,2,1,1]=(-0.745343658666)/(o+1.)+(0.193375557775)+(0.864596248255)*0.5**o
        ref[2,2,2,0]=(-0.491769251494)/(o+1.)+(0.442742387618)+(0.753340954862)*0.5**o
        ref[2,2,2,1]=(0.551196377158)/(o+1.)+(-0.523535406613)+(-0.212166925203)*0.5**o
        ref[2,3,0,0]=(0.192271522168)/(o+1.)+(-0.0181953944687)+(0.122707705798)*0.5**o
        ref[2,3,0,1]=(-0.593768738594)/(o+1.)+(0.199088910545)+(0.368366498588)*0.5**o
        ref[2,3,1,0]=(0.0866211808983)/(o+1.)+(0.723512694127)+(0.521936735685)*0.5**o
        ref[2,3,1,1]=(-0.453288268254)/(o+1.)+(-0.15580970281)+(-0.655272797205)*0.5**o
        ref[2,3,2,0]=(0.614105802669)/(o+1.)+(0.0411427694898)+(0.85774098305)*0.5**o
        ref[2,3,2,1]=(0.263133533654)/(o+1.)+(-0.628776355867)+(-0.624497793507)*0.5**o
        ref[2,4,0,0]=(-0.809347914722)/(o+1.)+(0.367719412237)+(0.26671699779)*0.5**o
        ref[2,4,0,1]=(0.773124137566)/(o+1.)+(-0.442496395542)+(-0.0149946871367)*0.5**o
        ref[2,4,1,0]=(-0.927311525126)/(o+1.)+(-0.0125177466568)+(-0.943580084777)*0.5**o
        ref[2,4,1,1]=(0.306409665483)/(o+1.)+(0.128569553604)+(-0.852228134533)*0.5**o
        ref[2,4,2,0]=(0.926185522145)/(o+1.)+(0.0750538146738)+(-0.58517230367)*0.5**o
        ref[2,4,2,1]=(0.858245470922)/(o+1.)+(0.0125876600893)+(0.457971730734)*0.5**o
        ref[3,0,0,0]=(-0.74501947953)/(o+1.)+(0.343942243752)+(-0.363581155556)*0.5**o
        ref[3,0,0,1]=(0.850890737948)/(o+1.)+(0.157997002133)+(-0.761705466968)*0.5**o
        ref[3,0,1,0]=(-0.0468962248591)/(o+1.)+(0.0297980532763)+(0.682735456043)*0.5**o
        ref[3,0,1,1]=(0.0126326425456)/(o+1.)+(-0.076115296096)+(0.575387199213)*0.5**o
        ref[3,0,2,0]=(-0.410549752333)/(o+1.)+(0.85556811978)+(0.270867209104)*0.5**o
        ref[3,0,2,1]=(0.580487848545)/(o+1.)+(0.59069171729)+(0.698757148346)*0.5**o
        ref[3,1,0,0]=(0.139052775786)/(o+1.)+(-0.334747475889)+(0.281989056901)*0.5**o
        ref[3,1,0,1]=(0.591083610998)/(o+1.)+(-0.322680889668)+(-0.492254197357)*0.5**o
        ref[3,1,1,0]=(0.384713965427)/(o+1.)+(-0.522411606102)+(0.146558304882)*0.5**o
        ref[3,1,1,1]=(-0.827615582009)/(o+1.)+(-0.00295218575854)+(0.514501517225)*0.5**o
        ref[3,1,2,0]=(0.84762578765)/(o+1.)+(0.208361555685)+(-0.43502341494)*0.5**o
        ref[3,1,2,1]=(-0.554668615664)/(o+1.)+(0.825347835483)+(-0.899680107129)*0.5**o
        ref[3,2,0,0]=(0.652802712143)/(o+1.)+(0.397197582731)+(0.7807432266)*0.5**o
        ref[3,2,0,1]=(0.421495251937)/(o+1.)+(-0.119571710512)+(0.677076974975)*0.5**o
        ref[3,2,1,0]=(-0.230439630375)/(o+1.)+(-0.733165339017)+(-0.816334071941)*0.5**o
        ref[3,2,1,1]=(0.64153308098)/(o+1.)+(0.778374176432)+(-0.477170252948)*0.5**o
        ref[3,2,2,0]=(-0.987848717403)/(o+1.)+(-0.0606435888479)+(0.974551740798)*0.5**o
        ref[3,2,2,1]=(-0.318593711859)/(o+1.)+(-0.459904454409)+(0.334228278552)*0.5**o
        ref[3,3,0,0]=(-0.418115413351)/(o+1.)+(-0.348616063735)+(-0.982162917221)*0.5**o
        ref[3,3,0,1]=(-0.437684435015)/(o+1.)+(0.804145379606)+(-0.892393985366)*0.5**o
        ref[3,3,1,0]=(0.944100110135)/(o+1.)+(-0.059994562245)+(0.0694197471801)*0.5**o
        ref[3,3,1,1]=(-0.860656048613)/(o+1.)+(0.178136109535)+(-0.250126193893)*0.5**o
        ref[3,3,2,0]=(-0.363631991908)/(o+1.)+(-0.298384287282)+(0.471110648232)*0.5**o
        ref[3,3,2,1]=(-0.578021736479)/(o+1.)+(0.320913094365)+(0.235895941501)*0.5**o
        ref[3,4,0,0]=(-0.946995649946)/(o+1.)+(0.0451978887455)+(0.936359833225)*0.5**o
        ref[3,4,0,1]=(-0.636305818291)/(o+1.)+(0.258932088634)+(0.449663732234)*0.5**o
        ref[3,4,1,0]=(-0.823421937366)/(o+1.)+(0.112980674356)+(0.297559014947)*0.5**o
        ref[3,4,1,1]=(-0.441768022626)/(o+1.)+(-0.240840962617)+(-0.682298558949)*0.5**o
        ref[3,4,2,0]=(-0.258808936822)/(o+1.)+(0.510306577882)+(-0.457251344276)*0.5**o
        ref[3,4,2,1]=(-0.489365749016)/(o+1.)+(-0.296617385332)+(0.979513856171)*0.5**o
      else:
        arg[0,0,0,0]=(-0.474247989728)*x[0]+(-0.768689352235)*x[1]+(-1.08527900853)*x[2]
        arg[0,0,0,1]=(-0.237113797789)*x[0]+(0.0842735741018)*x[1]+(0.203858551029)*x[2]
        arg[0,0,1,0]=(0.24237127665)*x[0]+(0.421022955803)*x[1]+(-1.52818540927)*x[2]
        arg[0,0,1,1]=(1.34624865817)*x[0]+(-0.798943272818)*x[1]+(-0.823168386315)*x[2]
        arg[0,0,2,0]=(-0.906723771267)*x[0]+(0.328893123771)*x[1]+(-0.942770202463)*x[2]
        arg[0,0,2,1]=(1.54305197694)*x[0]+(0.813063941984)*x[1]+(0.205532727168)*x[2]
        arg[0,1,0,0]=(0.399159979872)*x[0]+(-1.52089837934)*x[1]+(-1.4064251406)*x[2]
        arg[0,1,0,1]=(-0.121139354731)*x[0]+(1.12768246422)*x[1]+(0.234299409411)*x[2]
        arg[0,1,1,0]=(-1.62513899388)*x[0]+(-0.173131700583)*x[1]+(1.24199996799)*x[2]
        arg[0,1,1,1]=(-0.371929434177)*x[0]+(-0.606278610409)*x[1]+(-0.00600790063066)*x[2]
        arg[0,1,2,0]=(-0.804531294423)*x[0]+(0.869392248002)*x[1]+(0.840417359943)*x[2]
        arg[0,1,2,1]=(0.329452946275)*x[0]+(1.1072495446)*x[1]+(-0.213517932323)*x[2]
        arg[0,2,0,0]=(1.02784887224)*x[0]+(0.435517042337)*x[1]+(1.58767206677)*x[2]
        arg[0,2,0,1]=(-0.0120997177)*x[0]+(0.430135803314)*x[1]+(-0.717473068956)*x[2]
        arg[0,2,1,0]=(0.955455517302)*x[0]+(0.712976918968)*x[1]+(-0.624804508641)*x[2]
        arg[0,2,1,1]=(-1.03293986594)*x[0]+(-0.539062834482)*x[1]+(-1.61172701428)*x[2]
        arg[0,2,2,0]=(-0.177662810138)*x[0]+(0.553352366127)*x[1]+(0.695417121111)*x[2]
        arg[0,2,2,1]=(0.318133474679)*x[0]+(-0.452528947926)*x[1]+(0.468602505091)*x[2]
        arg[0,3,0,0]=(-0.294467725916)*x[0]+(1.67314409894)*x[1]+(0.100736440007)*x[2]
        arg[0,3,0,1]=(-0.517162349887)*x[0]+(-1.57430191254)*x[1]+(1.24045877402)*x[2]
        arg[0,3,1,0]=(0.667393640368)*x[0]+(0.340993632212)*x[1]+(-1.09920535559)*x[2]
        arg[0,3,1,1]=(0.37451406727)*x[0]+(0.37350649326)*x[1]+(0.357440871322)*x[2]
        arg[0,3,2,0]=(-0.989366398194)*x[0]+(0.596179154759)*x[1]+(-0.573280308087)*x[2]
        arg[0,3,2,1]=(-0.25561190141)*x[0]+(1.42296905238)*x[1]+(-1.21032031399)*x[2]
        arg[0,4,0,0]=(-0.41394085285)*x[0]+(-0.751213929646)*x[1]+(-0.68470718293)*x[2]
        arg[0,4,0,1]=(0.702985204167)*x[0]+(-0.0375821582587)*x[1]+(0.290630803819)*x[2]
        arg[0,4,1,0]=(0.0558204653441)*x[0]+(1.61940057302)*x[1]+(-0.518863528354)*x[2]
        arg[0,4,1,1]=(1.25876303604)*x[0]+(-1.40923202908)*x[1]+(1.05202333762)*x[2]
        arg[0,4,2,0]=(0.0715295483294)*x[0]+(0.883869702529)*x[1]+(-0.463253474408)*x[2]
        arg[0,4,2,1]=(0.564563789615)*x[0]+(-0.135049742988)*x[1]+(-1.3118878567)*x[2]
        arg[1,0,0,0]=(0.707267335291)*x[0]+(-0.197244449872)*x[1]+(-0.114362773781)*x[2]
        arg[1,0,0,1]=(-0.144762815659)*x[0]+(0.434177706452)*x[1]+(0.0754859099691)*x[2]
        arg[1,0,1,0]=(-0.906934991359)*x[0]+(0.179011995627)*x[1]+(0.823936036469)*x[2]
        arg[1,0,1,1]=(1.11165724605)*x[0]+(0.48023223731)*x[1]+(0.981353808181)*x[2]
        arg[1,0,2,0]=(0.499460583481)*x[0]+(0.171586243561)*x[1]+(-0.252474775657)*x[2]
        arg[1,0,2,1]=(-0.604289370884)*x[0]+(1.11610914086)*x[1]+(0.747439062711)*x[2]
        arg[1,1,0,0]=(0.696191463882)*x[0]+(0.349547714404)*x[1]+(0.651361019042)*x[2]
        arg[1,1,0,1]=(0.152679923426)*x[0]+(-0.448442911425)*x[1]+(1.33964371501)*x[2]
        arg[1,1,1,0]=(-0.74612592015)*x[0]+(-1.27089239786)*x[1]+(-0.261725631736)*x[2]
        arg[1,1,1,1]=(1.50839493929)*x[0]+(1.93229724987)*x[1]+(-0.963996845831)*x[2]
        arg[1,1,2,0]=(0.867984447375)*x[0]+(-0.777493811854)*x[1]+(0.847847553133)*x[2]
        arg[1,1,2,1]=(1.34249389498)*x[0]+(0.776160804062)*x[1]+(-0.295393972199)*x[2]
        arg[1,2,0,0]=(-0.655771613243)*x[0]+(1.58220214455)*x[1]+(-0.406292477629)*x[2]
        arg[1,2,0,1]=(0.59294895986)*x[0]+(0.544961164723)*x[1]+(0.634279670948)*x[2]
        arg[1,2,1,0]=(1.15324179862)*x[0]+(0.645764835217)*x[1]+(0.670364725309)*x[2]
        arg[1,2,1,1]=(-0.819459106271)*x[0]+(0.024815625404)*x[1]+(0.769753650676)*x[2]
        arg[1,2,2,0]=(1.19042366161)*x[0]+(0.982020548972)*x[1]+(0.92730609738)*x[2]
        arg[1,2,2,1]=(-0.742441488349)*x[0]+(-1.3101429812)*x[1]+(-0.198104246324)*x[2]
        arg[1,3,0,0]=(0.349205030437)*x[0]+(-0.745240339596)*x[1]+(-0.744742092259)*x[2]
        arg[1,3,0,1]=(0.993101296711)*x[0]+(-1.48331355036)*x[1]+(-1.00616778677)*x[2]
        arg[1,3,1,0]=(-1.10586442824)*x[0]+(1.71922449779)*x[1]+(-0.943451153259)*x[2]
        arg[1,3,1,1]=(-1.72393763035)*x[0]+(-1.14113495956)*x[1]+(-0.211909615809)*x[2]
        arg[1,3,2,0]=(-0.378302963108)*x[0]+(-0.494705322002)*x[1]+(0.940164363332)*x[2]
        arg[1,3,2,1]=(0.850541148743)*x[0]+(-1.45461945468)*x[1]+(1.20025821328)*x[2]
        arg[1,4,0,0]=(1.33882463795)*x[0]+(0.475208133119)*x[1]+(0.722686059253)*x[2]
        arg[1,4,0,1]=(0.0750422155969)*x[0]+(1.29758288845)*x[1]+(-1.41431668738)*x[2]
        arg[1,4,1,0]=(1.36072344575)*x[0]+(-0.345743547563)*x[1]+(-0.63248018754)*x[2]
        arg[1,4,1,1]=(0.579049229913)*x[0]+(0.954245074078)*x[1]+(-0.170538926965)*x[2]
        arg[1,4,2,0]=(1.67615484634)*x[0]+(0.561067019023)*x[1]+(-1.26952326914)*x[2]
        arg[1,4,2,1]=(0.700910547934)*x[0]+(0.58717328764)*x[1]+(0.795138196965)*x[2]
        arg[2,0,0,0]=(0.763193592359)*x[0]+(-0.397549328381)*x[1]+(-0.230593559545)*x[2]
        arg[2,0,0,1]=(0.454026153468)*x[0]+(-1.12464443717)*x[1]+(-0.147462098246)*x[2]
        arg[2,0,1,0]=(-0.276457034935)*x[0]+(0.118173156654)*x[1]+(0.57296366275)*x[2]
        arg[2,0,1,1]=(1.31887812253)*x[0]+(-0.544644483254)*x[1]+(0.0013728350582)*x[2]
        arg[2,0,2,0]=(-0.257589384265)*x[0]+(-0.016005453203)*x[1]+(-0.508836430965)*x[2]
        arg[2,0,2,1]=(-0.259420301773)*x[0]+(0.122645293168)*x[1]+(0.381613885289)*x[2]
        arg[2,1,0,0]=(-1.64168512418)*x[0]+(0.025182006192)*x[1]+(0.0268780458975)*x[2]
        arg[2,1,0,1]=(-0.0885344200898)*x[0]+(0.408795219114)*x[1]+(0.635798801606)*x[2]
        arg[2,1,1,0]=(-0.959658503283)*x[0]+(1.034415556)*x[1]+(0.725681105748)*x[2]
        arg[2,1,1,1]=(1.59312707789)*x[0]+(-0.116514952101)*x[1]+(-0.86121633507)*x[2]
        arg[2,1,2,0]=(-1.44859572067)*x[0]+(-1.28344247275)*x[1]+(0.299296401869)*x[2]
        arg[2,1,2,1]=(-0.488424372501)*x[0]+(1.23481402288)*x[1]+(0.01702902463)*x[2]
        arg[2,2,0,0]=(-1.29986951244)*x[0]+(-1.20190912751)*x[1]+(-0.755848291051)*x[2]
        arg[2,2,0,1]=(-0.52421656312)*x[0]+(-0.335471509689)*x[1]+(-1.01264312499)*x[2]
        arg[2,2,1,0]=(-1.12066642609)*x[0]+(-1.02138558053)*x[1]+(-1.00777755552)*x[2]
        arg[2,2,1,1]=(-1.42127024744)*x[0]+(1.63480022479)*x[1]+(0.542244906643)*x[2]
        arg[2,2,2,0]=(-0.418852880297)*x[0]+(-0.983110436453)*x[1]+(-0.622603673805)*x[2]
        arg[2,2,2,1]=(0.727366446825)*x[0]+(-0.256324856361)*x[1]+(-0.20608740758)*x[2]
        arg[2,3,0,0]=(0.548268302883)*x[0]+(0.0102173127665)*x[1]+(0.869833576217)*x[2]
        arg[2,3,0,1]=(-1.70159091061)*x[0]+(1.15876875083)*x[1]+(-0.234757706773)*x[2]
        arg[2,3,1,0]=(-1.72817024817)*x[0]+(0.690055132218)*x[1]+(-0.149398998825)*x[2]
        arg[2,3,1,1]=(1.80829412387)*x[0]+(-0.217899654855)*x[1]+(-0.922672217822)*x[2]
        arg[2,3,2,0]=(1.23855200328)*x[0]+(0.412795411508)*x[1]+(0.960476191755)*x[2]
        arg[2,3,2,1]=(-1.43061904621)*x[0]+(-0.766039000745)*x[1]+(-1.25765397361)*x[2]
        arg[2,4,0,0]=(-0.660872016687)*x[0]+(-0.000405233718708)*x[1]+(-0.178507968219)*x[2]
        arg[2,4,0,1]=(-0.168526882871)*x[0]+(-1.06936802685)*x[1]+(1.54083857104)*x[2]
        arg[2,4,1,0]=(0.565807359529)*x[0]+(-1.41401297939)*x[1]+(-1.23254260537)*x[2]
        arg[2,4,1,1]=(-0.373920075802)*x[0]+(0.407330812903)*x[1]+(-0.204667001328)*x[2]
        arg[2,4,2,0]=(0.874100879885)*x[0]+(-0.191238094267)*x[1]+(1.54890969328)*x[2]
        arg[2,4,2,1]=(0.566428125313)*x[0]+(-1.60899595601)*x[1]+(-1.05112529768)*x[2]
        arg[3,0,0,0]=(-0.688037248111)*x[0]+(-1.56483484784)*x[1]+(0.26782504619)*x[2]
        arg[3,0,0,1]=(0.418788954459)*x[0]+(-0.670455226092)*x[1]+(-0.0562760591375)*x[2]
        arg[3,0,1,0]=(-0.2614362754)*x[0]+(0.271050519855)*x[1]+(0.227811982923)*x[2]
        arg[3,0,1,1]=(-0.873198928647)*x[0]+(-0.483763141395)*x[1]+(0.33532383979)*x[2]
        arg[3,0,2,0]=(-0.864161938176)*x[0]+(-1.4682754386)*x[1]+(-0.38126411162)*x[2]
        arg[3,0,2,1]=(0.142827521381)*x[0]+(0.464483940188)*x[1]+(0.214297576606)*x[2]
        arg[3,1,0,0]=(0.188747079113)*x[0]+(0.590821253713)*x[1]+(0.908850367108)*x[2]
        arg[3,1,0,1]=(0.0587541582624)*x[0]+(0.955494266151)*x[1]+(1.70822209849)*x[2]
        arg[3,1,1,0]=(-0.420356861523)*x[0]+(-0.551445918262)*x[1]+(-0.0187114754269)*x[2]
        arg[3,1,1,1]=(1.07071658401)*x[0]+(-0.589035602194)*x[1]+(0.133725074027)*x[2]
        arg[3,1,2,0]=(0.370477043767)*x[0]+(0.226262353618)*x[1]+(0.566087784489)*x[2]
        arg[3,1,2,1]=(1.77828660075)*x[0]+(0.607419742685)*x[1]+(0.0297054744582)*x[2]
        arg[3,2,0,0]=(-1.19326939474)*x[0]+(1.20875153887)*x[1]+(-0.124756222091)*x[2]
        arg[3,2,0,1]=(0.0645190914271)*x[0]+(0.357580475716)*x[1]+(-1.59461469718)*x[2]
        arg[3,2,1,0]=(1.4853619905)*x[0]+(0.751582301648)*x[1]+(0.859363177167)*x[2]
        arg[3,2,1,1]=(1.36658655075)*x[0]+(-0.776650996613)*x[1]+(0.654743994261)*x[2]
        arg[3,2,2,0]=(0.406696005677)*x[0]+(-0.528546876384)*x[1]+(-0.439775325396)*x[2]
        arg[3,2,2,1]=(-0.725198159461)*x[0]+(-0.0933136564416)*x[1]+(-0.720365141393)*x[2]
        arg[3,3,0,0]=(-0.428669112669)*x[0]+(0.626162511175)*x[1]+(-0.382776233923)*x[2]
        arg[3,3,0,1]=(0.340474939218)*x[0]+(-0.140550383925)*x[1]+(0.353836983746)*x[2]
        arg[3,3,1,0]=(0.851646816711)*x[0]+(0.0171684195949)*x[1]+(0.38173044322)*x[2]
        arg[3,3,1,1]=(0.359647722225)*x[0]+(0.425799941165)*x[1]+(-0.80174052212)*x[2]
        arg[3,3,2,0]=(0.13012568736)*x[0]+(-0.573735693651)*x[1]+(1.08843730088)*x[2]
        arg[3,3,2,1]=(-1.01372953294)*x[0]+(0.335282956856)*x[1]+(1.24570512628)*x[2]
        arg[3,4,0,0]=(-0.05674609039)*x[0]+(-0.814879575064)*x[1]+(-0.169789408264)*x[2]
        arg[3,4,0,1]=(1.5968230938)*x[0]+(0.0679819277023)*x[1]+(0.892284314512)*x[2]
        arg[3,4,1,0]=(1.22220298478)*x[0]+(-0.0258417926472)*x[1]+(0.440670138568)*x[2]
        arg[3,4,1,1]=(0.647072768716)*x[0]+(-0.177651087906)*x[1]+(-0.115553644646)*x[2]
        arg[3,4,2,0]=(1.29816363306)*x[0]+(-1.18656044167)*x[1]+(-0.465776533157)*x[2]
        arg[3,4,2,1]=(0.328907496091)*x[0]+(-0.179857937151)*x[1]+(-0.476306938306)*x[2]
        ref[0,0,0,0]=(-0.530378359482)/(o+1.)+(-0.935143327523)+(0.0724486640358)*0.5**o
        ref[0,0,0,1]=(1.22837847871)/(o+1.)+(-0.113350954728)+(-0.950658241915)*0.5**o
        ref[0,0,1,0]=(-0.000414683535449)/(o+1.)+(-0.183158731552)+(-0.49805903018)*0.5**o
        ref[0,0,1,1]=(-0.882730389132)/(o+1.)+(-0.112174958816)+(0.831217305802)*0.5**o
        ref[0,0,2,0]=(-1.1913557966)/(o+1.)+(-0.0615750891074)+(-0.206094875146)*0.5**o
        ref[0,0,2,1]=(0.634536492589)/(o+1.)+(0.649890059464)+(0.627332034578)*0.5**o
        ref[0,1,0,0]=(-1.38665625775)/(o+1.)+(-1.0210653524)+(0.900623422471)*0.5**o
        ref[0,1,0,1]=(1.20633678024)/(o+1.)+(0.0888193121856)+(-0.143132885707)*0.5**o
        ref[0,1,1,0]=(0.452225278035)/(o+1.)+(-0.148926943801)+(-0.710642116901)*0.5**o
        ref[0,1,1,1]=(-0.244223430204)/(o+1.)+(-0.619607191996)+(0.499221868978)*0.5**o
        ref[0,1,2,0]=(0.0383849036606)/(o+1.)+(0.759484971644)+(-0.652076533427)*0.5**o
        ref[0,1,2,1]=(0.207558055759)/(o+1.)+(0.502305382384)+(0.0110157380215)*0.5**o
        ref[0,2,0,0]=(0.761374280144)/(o+1.)+(0.777616611515)+(0.734430478174)*0.5**o
        ref[0,2,0,1]=(0.554502862028)/(o+1.)+(-0.409274602016)+(-0.0353906413371)*0.5**o
        ref[0,2,1,0]=(-0.252543841709)/(o+1.)+(0.386640377894)+(0.522891013551)*0.5**o
        ref[0,2,1,1]=(-1.52497235887)/(o+1.)+(-0.807603254401)+(-0.0435508470367)*0.5**o
        ref[0,2,2,0]=(0.172485606257)/(o+1.)+(0.350297526187)+(0.19802601847)*0.5**o
        ref[0,2,2,1]=(-0.141271359379)/(o+1.)+(0.0509198134622)+(0.373638764298)*0.5**o
        ref[0,3,0,0]=(1.31278836809)/(o+1.)+(-0.0918739323256)+(0.350372309595)*0.5**o
        ref[0,3,0,1]=(-0.396933928109)/(o+1.)+(-0.436137805999)+(0.418204051696)*0.5**o
        ref[0,3,1,0]=(-0.206619840099)/(o+1.)+(-0.118703289798)+(0.353208336689)*0.5**o
        ref[0,3,1,1]=(-0.595153371293)/(o+1.)+(0.615722581562)+(0.469169640021)*0.5**o
        ref[0,3,2,0]=(-0.680858997986)/(o+1.)+(-0.0393223937964)+(-0.206963765944)*0.5**o
        ref[0,3,2,1]=(0.012753264598)/(o+1.)+(0.246188741489)+(-0.548093910595)*0.5**o
        ref[0,4,0,0]=(-0.40478987122)/(o+1.)+(-0.890399681609)+(0.33572726901)*0.5**o
        ref[0,4,0,1]=(0.0280062557264)/(o+1.)+(0.0358049413228)+(0.856417711355)*0.5**o
        ref[0,4,1,0]=(0.181560004565)/(o+1.)+(0.329885369854)+(0.31502676574)*0.5**o
        ref[0,4,1,1]=(0.0552681276841)/(o+1.)+(0.00372919426319)+(0.838827828375)*0.5**o
        ref[0,4,2,0]=(0.656057162724)/(o+1.)+(0.179315527684)+(-0.522542441641)*0.5**o
        ref[0,4,2,1]=(0.126189389317)/(o+1.)+(-0.429452001643)+(-0.149659196108)*0.5**o
        ref[1,0,0,0]=(1.43443485527)/(o+1.)+(-0.55091754491)+(0.0630603461901)*0.5**o
        ref[1,0,0,1]=(-0.612191779306)/(o+1.)+(0.604025235381)+(-0.230957890695)*0.5**o
        ref[1,0,1,0]=(0.373085423681)/(o+1.)+(-0.0817391530697)+(-0.113594076805)*0.5**o
        ref[1,0,1,1]=(0.761862677657)/(o+1.)+(0.819524484428)+(0.172331645033)*0.5**o
        ref[1,0,2,0]=(0.929265450746)/(o+1.)+(-0.382585784938)+(0.254478170515)*0.5**o
        ref[1,0,2,1]=(1.58236203463)/(o+1.)+(-0.280641188811)+(0.238179175679)*0.5**o
        ref[1,1,0,0]=(-0.117055726312)/(o+1.)+(0.957914090571)+(-0.101672257501)*0.5**o
        ref[1,1,0,1]=(0.0849062553796)/(o+1.)+(0.0703564566846)+(0.818261558262)*0.5**o
        ref[1,1,1,0]=(-1.72219113622)/(o+1.)+(-0.357454724939)+(0.15835663635)*0.5**o
        ref[1,1,1,1]=(0.739078400468)/(o+1.)+(0.532292543169)+(0.673031856531)*0.5**o
        ref[1,1,2,0]=(0.255023909903)/(o+1.)+(0.229656917655)+(0.22400044344)*0.5**o
        ref[1,1,2,1]=(0.518856765199)/(o+1.)+(0.380552790125)+(0.543298381396)*0.5**o
        ref[1,2,0,0]=(0.0321563884831)/(o+1.)+(0.25981567762)+(-0.0316496900482)*0.5**o
        ref[1,2,0,1]=(1.09318176771)/(o+1.)+(0.0572442907603)+(0.564519446305)*0.5**o
        ref[1,2,1,0]=(0.756962791387)/(o+1.)+(0.69419429624)+(0.32401997528)*0.5**o
        ref[1,2,1,1]=(1.28480299241)/(o+1.)+(-0.201990593541)+(-0.905711635518)*0.5**o
        ref[1,2,2,0]=(1.04969566442)/(o+1.)+(0.72854097372)+(0.592972696101)*0.5**o
        ref[1,2,2,1]=(-1.77086264775)/(o+1.)+(-0.261914180361)+(0.0440022926066)*0.5**o
        ref[1,3,0,0]=(-0.477780531124)/(o+1.)+(-0.181354085163)+(-0.300288699967)*0.5**o
        ref[1,3,0,1]=(-0.797627013456)/(o+1.)+(-0.523163955197)+(0.347574883435)*0.5**o
        ref[1,3,1,0]=(0.428806106821)/(o+1.)+(-0.0514584302394)+(-0.655980330052)*0.5**o
        ref[1,3,1,1]=(-0.141740459457)/(o+1.)+(-1.05699960931)+(-0.821242527644)*0.5**o
        ref[1,3,2,0]=(0.0181810985873)/(o+1.)+(-0.180697139553)+(0.41036925874)*0.5**o
        ref[1,3,2,1]=(-0.061077367)/(o+1.)+(0.250077446763)+(0.157102380812)*0.5**o
        ref[1,4,0,0]=(1.14087212913)/(o+1.)+(0.368059029864)+(0.659728641463)*0.5**o
        ref[1,4,0,1]=(0.365658713316)/(o+1.)+(0.088533821044)+(-0.584417938734)*0.5**o
        ref[1,4,1,0]=(-1.11681308321)/(o+1.)+(0.438275560627)+(0.622761672604)*0.5**o
        ref[1,4,1,1]=(1.18893372018)/(o+1.)+(0.172934240923)+(-0.172046824999)*0.5**o
        ref[1,4,2,0]=(0.16240489493)/(o+1.)+(-0.0145923852256)+(0.834478471749)*0.5**o
        ref[1,4,2,1]=(0.571355829378)/(o+1.)+(0.526896417749)+(0.458073367663)*0.5**o
        ref[2,0,0,0]=(-0.489234572937)/(o+1.)+(-0.0135589093901)+(0.65140309615)*0.5**o
        ref[2,0,0,1]=(-1.55383491931)/(o+1.)+(0.0161644619972)+(0.703425613366)*0.5**o
        ref[2,0,1,0]=(-0.258604855723)/(o+1.)+(0.0827812678671)+(0.507722104457)*0.5**o
        ref[2,0,1,1]=(0.0504178529481)/(o+1.)+(0.182991850332)+(0.359204920726)*0.5**o
        ref[2,0,2,0]=(0.652043479841)/(o+1.)+(-0.443015450013)+(-0.548443848248)*0.5**o
        ref[2,0,2,1]=(0.340317839857)/(o+1.)+(-0.379565594545)+(0.663652225917)*0.5**o
        ref[2,1,0,0]=(-0.587863196667)/(o+1.)+(-0.0137797674284)+(-0.974202340562)*0.5**o
        ref[2,1,0,1]=(1.00030747039)/(o+1.)+(-0.399997754768)+(0.755747639771)*0.5**o
        ref[2,1,1,0]=(0.703413281036)/(o+1.)+(0.487016208736)+(-0.877007540039)*0.5**o
        ref[2,1,1,1]=(0.263166830572)/(o+1.)+(-0.157259452849)+(0.666747865841)*0.5**o
        ref[2,1,2,0]=(0.104532343004)/(o+1.)+(-1.02548001493)+(-0.486314104701)*0.5**o
        ref[2,1,2,1]=(0.933518655478)/(o+1.)+(0.0369928229139)+(-0.244085626298)*0.5**o
        ref[2,2,0,0]=(-1.50550384519)/(o+1.)+(-0.590996227594)+(-0.570130630622)*0.5**o
        ref[2,2,0,1]=(-0.60318683839)/(o+1.)+(-0.38308245528)+(-0.502979448853)*0.5**o
        ref[2,2,1,0]=(-1.11258910879)/(o+1.)+(-0.518777277969)+(-0.999685897413)*0.5**o
        ref[2,2,1,1]=(1.84366617107)/(o+1.)+(-0.232866738261)+(-0.622157810558)*0.5**o
        ref[2,2,2,0]=(-0.620181988585)/(o+1.)+(-0.836162444183)+(0.267939886396)*0.5**o
        ref[2,2,2,1]=(-0.826647389099)/(o+1.)+(0.0929536518109)+(0.905694268362)*0.5**o
        ref[2,3,0,0]=(0.184067674443)/(o+1.)+(0.259745858757)+(0.724759799909)*0.5**o
        ref[2,3,0,1]=(0.0175595303049)/(o+1.)+(0.0560117358912)+(-0.907162868639)*0.5**o
        ref[2,3,1,0]=(-0.609443154617)/(o+1.)+(0.186540819401)+(-0.951152598958)*0.5**o
        ref[2,3,1,1]=(-1.61858082094)/(o+1.)+(0.679439858578)+(0.927423354971)*0.5**o
        ref[2,3,2,0]=(0.44350652486)/(o+1.)+(0.81466177481)+(0.53899353207)*0.5**o
        ref[2,3,2,1]=(-0.877289862245)/(o+1.)+(-1.05742260849)+(-0.462176941342)*0.5**o
        ref[2,4,0,0]=(-0.736381416539)/(o+1.)+(-0.0139374259612)+(-0.0755289501641)*0.5**o
        ref[2,4,0,1]=(0.690429326999)/(o+1.)+(-0.0293644018251)+(-0.328756862036)*0.5**o
        ref[2,4,1,0]=(-1.57117034002)/(o+1.)+(-0.531463765387)+(0.553349645561)*0.5**o
        ref[2,4,1,1]=(1.24898175282)/(o+1.)+(-0.452273024959)+(-0.515691967125)*0.5**o
        ref[2,4,2,0]=(0.239126191643)/(o+1.)+(0.722504563932)+(0.547637159392)*0.5**o
        ref[2,4,2,1]=(-1.40763254209)/(o+1.)+(-0.809456441012)+(0.932852295739)*0.5**o
        ref[3,0,0,0]=(-0.416554297686)/(o+1.)+(-0.632193610911)+(-0.304105530258)*0.5**o
        ref[3,0,0,1]=(-1.02983448036)/(o+1.)+(0.579212231778)+(-0.436532313966)*0.5**o
        ref[3,0,1,0]=(0.42124133216)/(o+1.)+(0.368290418827)+(-0.920395942436)*0.5**o
        ref[3,0,1,1]=(0.158195474854)/(o+1.)+(-0.17459474514)+(-0.830644214826)*0.5**o
        ref[3,0,2,0]=(-1.18749647329)/(o+1.)+(-0.506173996178)+(-0.513857022748)*0.5**o
        ref[3,0,2,1]=(0.233437753018)/(o+1.)+(0.455705145371)+(-0.323239005586)*0.5**o
        ref[3,1,0,0]=(1.42991958566)/(o+1.)+(0.053769921022)+(0.150959272234)*0.5**o
        ref[3,1,0,1]=(1.48423731111)/(o+1.)+(0.964048315997)+(-0.689863420196)*0.5**o
        ref[3,1,1,0]=(0.868837676765)/(o+1.)+(-1.01261543254)+(0.165878933098)*0.5**o
        ref[3,1,1,1]=(0.820750307831)/(o+1.)+(-0.300930223821)+(0.396516195659)*0.5**o
        ref[3,1,2,0]=(0.477077782202)/(o+1.)+(-0.0416336989494)+(0.769016797571)*0.5**o
        ref[3,1,2,1]=(0.359895440587)/(o+1.)+(0.555322395811)+(0.944871585683)*0.5**o
        ref[3,2,0,0]=(0.657668315832)/(o+1.)+(-0.240321895174)+(-0.286298603451)*0.5**o
        ref[3,2,0,1]=(-0.941360568282)/(o+1.)+(-0.0246544136593)+(-0.181845734439)*0.5**o
        ref[3,2,1,0]=(1.60316834971)/(o+1.)+(0.252291602298)+(0.988555915005)*0.5**o
        ref[3,2,1,1]=(0.833811957678)/(o+1.)+(-0.111970621871)+(0.634808834458)*0.5**o
        ref[3,2,2,0]=(-1.38877073784)/(o+1.)+(0.538356636307)+(-0.249568730877)*0.5**o
        ref[3,2,2,1]=(-0.311839184097)/(o+1.)+(-0.698114103264)+(0.169190433329)*0.5**o
        ref[3,3,0,0]=(0.743170989696)/(o+1.)+(-0.499224746788)+(0.0699956684628)*0.5**o
        ref[3,3,0,1]=(-0.293538786051)/(o+1.)+(0.26667992866)+(0.313940467768)*0.5**o
        ref[3,3,1,0]=(0.951483461345)/(o+1.)+(0.148317904561)+(0.00242640905887)*0.5**o
        ref[3,3,1,1]=(-0.110753912776)/(o+1.)+(-0.44308689138)+(0.980634836805)*0.5**o
        ref[3,3,2,0]=(-0.337895223544)/(o+1.)+(0.358941646037)+(0.264839226057)*0.5**o
        ref[3,3,2,1]=(0.857823611315)/(o+1.)+(0.30404611512)+(-0.898657291359)*0.5**o
        ref[3,4,0,0]=(0.037944509376)/(o+1.)+(-0.364801700413)+(-0.349756182268)*0.5**o
        ref[3,4,0,1]=(1.01114399396)/(o+1.)+(0.457103279485)+(0.631738783076)*0.5**o
        ref[3,4,1,0]=(0.969931913888)/(o+1.)+(-0.165922085756)+(0.998943588321)*0.5**o
        ref[3,4,1,1]=(1.35441114248)/(o+1.)+(-0.502422727199)+(0.00430234808496)*0.5**o
        ref[3,4,2,0]=(-0.367201097566)/(o+1.)+(-0.337780780531)+(0.688589316866)*0.5**o
        ref[3,4,2,1]=(-0.0547905846061)/(o+1.)+(-0.293870420685)+(0.31527404661)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.241972460279)*x[0]+(0.539729925673)*x[1]
        ref=(-0.11832904353)/(o+1.)+(0.649911364492)+(-0.399791299503)*0.5**o
      else:
        arg=(0.53919797656)*x[0]+(0.379010664884)*x[1]+(0.274416064529)*x[2]
        ref=(0.307947301472)/(o+1.)+(0.120674157748)+(0.643329089005)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.84849017056)*x[0]+(-0.161962845202)*x[1]
        arg[1]=(-0.794717919135)*x[0]+(1.16003773358)*x[1]
        ref[0]=(0.780514448202)/(o+1.)+(-0.34628958021)+(0.598592037576)*0.5**o
        ref[1]=(0.570158156278)/(o+1.)+(0.241192316704)+(-0.687222975237)*0.5**o
      else:
        arg[0]=(-0.637352163634)*x[0]+(1.49477462643)*x[1]+(-0.284018740238)*x[2]
        arg[1]=(0.125937459904)*x[0]+(-0.206499869891)*x[1]+(-0.342856240345)*x[2]
        ref[0]=(-0.274486314451)/(o+1.)+(0.379737553853)+(0.0884149293071)*0.5**o
        ref[1]=(-0.78452814021)/(o+1.)+(-0.122017410137)+(0.605144310151)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.440868762581)*x[0]+(-0.216223666616)*x[1]
        arg[0,1]=(-1.78807355591)*x[0]+(-0.208278123838)*x[1]
        arg[0,2]=(-0.448066998103)*x[0]+(-0.819815507826)*x[1]
        arg[0,3]=(1.27247881689)*x[0]+(0.0653439567221)*x[1]
        arg[0,4]=(-1.00344990223)*x[0]+(-0.291263790583)*x[1]
        arg[1,0]=(0.440080423838)*x[0]+(0.0356836558113)*x[1]
        arg[1,1]=(-1.49545340091)*x[0]+(-0.133615590211)*x[1]
        arg[1,2]=(1.05947638385)*x[0]+(0.408685904238)*x[1]
        arg[1,3]=(-0.170309086605)*x[0]+(0.660523970285)*x[1]
        arg[1,4]=(0.764323678475)*x[0]+(-0.529415080303)*x[1]
        arg[2,0]=(0.55270081325)*x[0]+(-1.60630146231)*x[1]
        arg[2,1]=(1.57419330248)*x[0]+(-1.24644288706)*x[1]
        arg[2,2]=(0.0416757238363)*x[0]+(0.0247379738465)*x[1]
        arg[2,3]=(1.0200370393)*x[0]+(0.364209137693)*x[1]
        arg[2,4]=(-1.21858490389)*x[0]+(0.523056254546)*x[1]
        arg[3,0]=(0.0524205820842)*x[0]+(1.65953217854)*x[1]
        arg[3,1]=(1.21824883395)*x[0]+(0.905190834998)*x[1]
        arg[3,2]=(0.226344434525)*x[0]+(-0.735349531462)*x[1]
        arg[3,3]=(-1.00295245717)*x[0]+(-0.719605376499)*x[1]
        arg[3,4]=(-0.446228759693)*x[0]+(1.23441974277)*x[1]
        ref[0,0]=(0.347328582602)/(o+1.)+(-0.520484463519)+(0.0365479152384)*0.5**o
        ref[0,1]=(-0.274564442352)/(o+1.)+(-0.413557210678)+(-0.894672816039)*0.5**o
        ref[0,2]=(0.0720644688954)/(o+1.)+(-0.906636966188)+(0.473326957552)*0.5**o
        ref[0,3]=(-0.747749341777)/(o+1.)+(0.730249972378)+(0.625072170633)*0.5**o
        ref[0,4]=(-0.194031116169)/(o+1.)+(-0.432063325829)+(-0.236555924983)*0.5**o
        ref[1,0]=(0.807081538033)/(o+1.)+(-0.612266650007)+(0.893215841629)*0.5**o
        ref[1,1]=(-0.662516518957)/(o+1.)+(-0.176318925886)+(-0.61391462039)*0.5**o
        ref[1,2]=(-0.570345636156)/(o+1.)+(0.791877954963)+(0.454752014319)*0.5**o
        ref[1,3]=(-0.312400910795)/(o+1.)+(0.212935137466)+(0.376745519544)*0.5**o
        ref[1,4]=(-0.764680414264)/(o+1.)+(0.174215486579)+(0.651158039278)*0.5**o
        ref[2,0]=(-0.837942771642)/(o+1.)+(-0.187938486702)+(0.160219095985)*0.5**o
        ref[2,1]=(-0.672421039761)/(o+1.)+(0.17384002882)+(0.652491397549)*0.5**o
        ref[2,2]=(-0.656187949922)/(o+1.)+(0.559888736121)+(-0.397175824637)*0.5**o
        ref[2,3]=(0.259558368195)/(o+1.)+(0.55148644978)+(0.0217149092406)*0.5**o
        ref[2,4]=(-0.188397537901)/(o+1.)+(-0.0561706170512)+(-0.394789877345)*0.5**o
        ref[3,0]=(0.91055628617)/(o+1.)+(0.574211274098)+(-0.347026073739)*0.5**o
        ref[3,1]=(0.29225509237)/(o+1.)+(0.64447705482)+(0.542230466941)*0.5**o
        ref[3,2]=(0.121065624384)/(o+1.)+(0.0316495277836)+(-0.693369776888)*0.5**o
        ref[3,3]=(-0.943523334976)/(o+1.)+(-0.266302644031)+(-0.246429210629)*0.5**o
        ref[3,4]=(0.714915399455)/(o+1.)+(0.459963131994)+(-0.846650680369)*0.5**o
      else:
        arg[0,0]=(-0.493869713454)*x[0]+(-0.381969413515)*x[1]+(1.59862368848)*x[2]
        arg[0,1]=(-0.581245543547)*x[0]+(0.961117271889)*x[1]+(-0.158073515343)*x[2]
        arg[0,2]=(-0.635283143641)*x[0]+(0.695573236987)*x[1]+(-0.719180832278)*x[2]
        arg[0,3]=(-0.737780596733)*x[0]+(-1.11171283352)*x[1]+(0.459954346315)*x[2]
        arg[0,4]=(1.12522065074)*x[0]+(-1.08427904854)*x[1]+(-0.715679673556)*x[2]
        arg[1,0]=(0.972468186181)*x[0]+(-0.211014530435)*x[1]+(-0.734414099002)*x[2]
        arg[1,1]=(-1.67337704748)*x[0]+(0.871096213089)*x[1]+(1.51063313223)*x[2]
        arg[1,2]=(-0.222013007834)*x[0]+(1.46474508326)*x[1]+(-0.979243908862)*x[2]
        arg[1,3]=(1.21195281087)*x[0]+(0.274082389329)*x[1]+(1.56144436465)*x[2]
        arg[1,4]=(1.35812865884)*x[0]+(1.25819798853)*x[1]+(0.895334048395)*x[2]
        arg[2,0]=(-0.167404344437)*x[0]+(0.0710591934315)*x[1]+(-0.124469683665)*x[2]
        arg[2,1]=(-0.492240632996)*x[0]+(-0.317251117382)*x[1]+(0.0195461703529)*x[2]
        arg[2,2]=(-0.277590266225)*x[0]+(-1.65859476899)*x[1]+(-0.0100216405668)*x[2]
        arg[2,3]=(-0.0701281298455)*x[0]+(0.502137254703)*x[1]+(0.701220923443)*x[2]
        arg[2,4]=(-0.72186096407)*x[0]+(-0.900487568179)*x[1]+(-0.735573922826)*x[2]
        arg[3,0]=(-0.0677324474129)*x[0]+(-0.649963209718)*x[1]+(-0.770861099726)*x[2]
        arg[3,1]=(0.563588189481)*x[0]+(0.714614932251)*x[1]+(-0.0957764632886)*x[2]
        arg[3,2]=(1.13337395062)*x[0]+(0.507556316073)*x[1]+(-0.196759085949)*x[2]
        arg[3,3]=(-1.2636883168)*x[0]+(-0.811660344076)*x[1]+(0.645462052598)*x[2]
        arg[3,4]=(-0.104175186987)*x[0]+(-0.929176038155)*x[1]+(1.07530379856)*x[2]
        ref[0,0]=(0.84538674973)/(o+1.)+(0.0569713556632)+(-0.236544899547)*0.5**o
        ref[0,1]=(-0.307689589699)/(o+1.)+(0.245040337642)+(0.0394071274131)*0.5**o
        ref[0,2]=(0.955244736357)/(o+1.)+(-0.341771136504)+(-0.930593202281)*0.5**o
        ref[0,3]=(-0.353490461722)/(o+1.)+(-0.476228433901)+(-0.083591754415)*0.5**o
        ref[0,4]=(-1.31220485482)/(o+1.)+(0.0466426656676)+(0.544181452128)*0.5**o
        ref[1,0]=(0.22773475414)/(o+1.)+(-0.488454801654)+(0.776214405914)*0.5**o
        ref[1,1]=(1.50688678707)/(o+1.)+(0.0851415404695)+(-0.96881757017)*0.5**o
        ref[1,2]=(0.313471097102)/(o+1.)+(-0.0875594518156)+(0.125135973096)*0.5**o
        ref[1,3]=(0.269521220179)/(o+1.)+(1.1192007195)+(0.539556905661)*0.5**o
        ref[1,4]=(1.91499876064)/(o+1.)+(0.509063797893)+(0.578534339343)*0.5**o
        ref[2,0]=(0.941352651562)/(o+1.)+(-0.966199402186)+(0.770231318139)*0.5**o
        ref[2,1]=(-0.141490823562)/(o+1.)+(-0.20919987022)+(-0.230055016023)*0.5**o
        ref[2,2]=(-0.916148047971)/(o+1.)+(-0.371740600192)+(-0.286577427426)*0.5**o
        ref[2,3]=(0.229516219331)/(o+1.)+(0.274790532736)+(0.354132763498)*0.5**o
        ref[2,4]=(-1.19120612619)/(o+1.)+(-0.139859859385)+(-0.88699661011)*0.5**o
        ref[3,0]=(-1.48630121465)/(o+1.)+(0.0744166251063)+(-0.151088792417)*0.5**o
        ref[3,1]=(-0.335583560513)/(o+1.)+(0.277436410841)+(0.963137397276)*0.5**o
        ref[3,2]=(-0.673942688644)/(o+1.)+(0.70077509905)+(0.716563671283)*0.5**o
        ref[3,3]=(-0.148394092795)/(o+1.)+(-0.237190239072)+(-0.807112037336)*0.5**o
        ref[3,4]=(0.529983886211)/(o+1.)+(-0.371055083588)+(0.254078854382)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.661725715165)*x[0]+(0.219720717581)*x[1]
        arg[0,0,1]=(-0.642240082304)*x[0]+(-0.279652756776)*x[1]
        arg[0,1,0]=(0.0204235813151)*x[0]+(-0.374873249933)*x[1]
        arg[0,1,1]=(1.4418481252)*x[0]+(0.205656834772)*x[1]
        arg[1,0,0]=(0.65992179951)*x[0]+(-1.04887875458)*x[1]
        arg[1,0,1]=(-0.195480216199)*x[0]+(-0.373905275039)*x[1]
        arg[1,1,0]=(0.781251543352)*x[0]+(0.0416248109536)*x[1]
        arg[1,1,1]=(-0.925751319189)*x[0]+(-0.113992910046)*x[1]
        arg[2,0,0]=(-0.0767414065227)*x[0]+(-0.156682840973)*x[1]
        arg[2,0,1]=(0.688457272529)*x[0]+(0.524916336046)*x[1]
        arg[2,1,0]=(0.753722830084)*x[0]+(-0.890325839466)*x[1]
        arg[2,1,1]=(0.730789290145)*x[0]+(-0.278986703716)*x[1]
        arg[3,0,0]=(-0.287376796439)*x[0]+(-0.153987233224)*x[1]
        arg[3,0,1]=(-1.15687435255)*x[0]+(0.250005890238)*x[1]
        arg[3,1,0]=(-0.845312830629)*x[0]+(0.256211962352)*x[1]
        arg[3,1,1]=(-0.574311820519)*x[0]+(-0.440297035998)*x[1]
        arg[4,0,0]=(-0.360607647308)*x[0]+(-1.29380437557)*x[1]
        arg[4,0,1]=(1.32019433485)*x[0]+(0.295255956618)*x[1]
        arg[4,1,0]=(-1.64162941972)*x[0]+(0.117482372162)*x[1]
        arg[4,1,1]=(-0.303103646619)*x[0]+(-0.88458982681)*x[1]
        arg[5,0,0]=(-0.666761879038)*x[0]+(-0.735786278877)*x[1]
        arg[5,0,1]=(0.0892896382617)*x[0]+(-0.45478872773)*x[1]
        arg[5,1,0]=(-0.50985624415)*x[0]+(-0.400022939567)*x[1]
        arg[5,1,1]=(1.39240706368)*x[0]+(0.0149740822664)*x[1]
        ref[0,0,0]=(-0.114834538488)/(o+1.)+(0.240621006484)+(0.515038958264)*0.5**o
        ref[0,0,1]=(0.588698337077)/(o+1.)+(-0.702256377233)+(-0.10607842169)*0.5**o
        ref[0,1,0]=(0.13849818224)/(o+1.)+(-0.277555277991)+(0.0621627051247)*0.5**o
        ref[0,1,1]=(-0.0878009521163)/(o+1.)+(0.630168348947)+(0.474969214192)*0.5**o
        ref[1,0,0]=(-0.696600538111)/(o+1.)+(-0.343318260912)+(0.994280104861)*0.5**o
        ref[1,0,1]=(-0.0887026688527)/(o+1.)+(-0.385753227979)+(0.290823633572)*0.5**o
        ref[1,1,0]=(0.831408608956)/(o+1.)+(-0.242985671853)+(0.477439089057)*0.5**o
        ref[1,1,1]=(-0.621059053191)/(o+1.)+(-0.0700203481241)+(-0.278644479797)*0.5**o
        ref[2,0,0]=(0.7890720399)/(o+1.)+(-0.539721196393)+(0.0569461053905)*0.5**o
        ref[2,0,1]=(-0.228628151401)/(o+1.)+(0.826946074171)+(-0.211890388366)*0.5**o
        ref[2,1,0]=(-0.837617325493)/(o+1.)+(0.289656653751)+(0.121701008608)*0.5**o
        ref[2,1,1]=(0.246577988625)/(o+1.)+(0.0512227634376)+(0.102779070928)*0.5**o
        ref[3,0,0]=(-0.0109584684212)/(o+1.)+(-0.334888666171)+(0.239371771102)*0.5**o
        ref[3,0,1]=(-0.125987676419)/(o+1.)+(-0.170179004032)+(-0.440522777827)*0.5**o
        ref[3,1,0]=(-0.354093141636)/(o+1.)+(-0.177294275919)+(0.119580825198)*0.5**o
        ref[3,1,1]=(-0.746761075412)/(o+1.)+(0.192068760731)+(-0.651985302566)*0.5**o
        ref[4,0,0]=(-0.42411673219)/(o+1.)+(-0.579162944937)+(-0.0719694008158)*0.5**o
        ref[4,0,1]=(0.169881169584)/(o+1.)+(0.483614029132)+(0.478341063616)*0.5**o
        ref[4,1,0]=(0.943567836264)/(o+1.)+(-0.898782941264)+(-0.670149001297)*0.5**o
        ref[4,1,1]=(-0.703610807936)/(o+1.)+(0.225342628657)+(-0.934767922806)*0.5**o
        ref[5,0,0]=(-0.788792225213)/(o+1.)+(-0.1768857093)+(-0.259984514102)*0.5**o
        ref[5,0,1]=(-0.788708879899)/(o+1.)+(-0.202288491141)+(0.827786772711)*0.5**o
        ref[5,1,0]=(-0.823081809707)/(o+1.)+(0.274720963093)+(-0.636239300197)*0.5**o
        ref[5,1,1]=(0.498922549807)/(o+1.)+(0.0129388257746)+(0.882580944588)*0.5**o
      else:
        arg[0,0,0]=(-0.426613654844)*x[0]+(-1.01433980801)*x[1]+(0.564621874916)*x[2]
        arg[0,0,1]=(-0.438521968882)*x[0]+(1.6033775267)*x[1]+(0.691349995832)*x[2]
        arg[0,1,0]=(-0.04337513731)*x[0]+(0.142927668234)*x[1]+(0.506581723199)*x[2]
        arg[0,1,1]=(-0.120598868162)*x[0]+(-0.213666787813)*x[1]+(-1.04221679332)*x[2]
        arg[1,0,0]=(0.153696490712)*x[0]+(-0.122516529815)*x[1]+(-1.08837272712)*x[2]
        arg[1,0,1]=(0.410706464237)*x[0]+(0.720079363315)*x[1]+(-0.767790357539)*x[2]
        arg[1,1,0]=(-1.03726428173)*x[0]+(-0.316809853003)*x[1]+(-0.780405868082)*x[2]
        arg[1,1,1]=(-0.427547967216)*x[0]+(-0.897766643609)*x[1]+(-1.56913552404)*x[2]
        arg[2,0,0]=(-0.282789248898)*x[0]+(-0.110846936031)*x[1]+(-1.4329391906)*x[2]
        arg[2,0,1]=(1.46866588617)*x[0]+(-1.23022501959)*x[1]+(-0.3123390018)*x[2]
        arg[2,1,0]=(-0.314115718621)*x[0]+(1.7333934438)*x[1]+(0.833572700726)*x[2]
        arg[2,1,1]=(0.464246462638)*x[0]+(0.267484068646)*x[1]+(0.342009204787)*x[2]
        arg[3,0,0]=(0.578853099163)*x[0]+(-0.0453855508465)*x[1]+(-0.0751578475516)*x[2]
        arg[3,0,1]=(-0.526734485752)*x[0]+(-1.03286310016)*x[1]+(-0.398986583917)*x[2]
        arg[3,1,0]=(0.0239088277798)*x[0]+(0.205105470192)*x[1]+(-1.04986349166)*x[2]
        arg[3,1,1]=(0.130893744001)*x[0]+(-0.273702140669)*x[1]+(-0.0804222415128)*x[2]
        arg[4,0,0]=(-0.822661618584)*x[0]+(-0.267189204519)*x[1]+(-0.290013766971)*x[2]
        arg[4,0,1]=(-1.46759823734)*x[0]+(0.317465945316)*x[1]+(0.108284317752)*x[2]
        arg[4,1,0]=(-1.3920990386)*x[0]+(-1.17823778694)*x[1]+(-0.18366434307)*x[2]
        arg[4,1,1]=(-0.0915496209516)*x[0]+(-0.367624609832)*x[1]+(1.30630257323)*x[2]
        arg[5,0,0]=(0.117840830455)*x[0]+(-0.667875433194)*x[1]+(-1.59716688455)*x[2]
        arg[5,0,1]=(-0.818693227605)*x[0]+(0.483842075248)*x[1]+(-0.408373205535)*x[2]
        arg[5,1,0]=(1.36906770301)*x[0]+(0.494926179999)*x[1]+(0.717041707262)*x[2]
        arg[5,1,1]=(0.819757116081)*x[0]+(0.0982458749015)*x[1]+(0.275080391599)*x[2]
        ref[0,0,0]=(0.236560699949)/(o+1.)+(-0.392784388084)+(-0.327323511722)*0.5**o
        ref[0,0,1]=(1.32491522045)/(o+1.)+(0.525146863711)+(-0.519003394222)*0.5**o
        ref[0,1,0]=(0.433749865563)/(o+1.)+(0.23007126008)+(-0.287758131599)*0.5**o
        ref[0,1,1]=(-0.0563021060686)/(o+1.)+(-0.403921160703)+(-0.512338021825)*0.5**o
        ref[1,0,0]=(-0.170496965963)/(o+1.)+(-0.914421777259)+(0.942147754258)*0.5**o
        ref[1,0,1]=(-0.372707483726)/(o+1.)+(0.636041158872)+(-0.536379364005)*0.5**o
        ref[1,1,0]=(-0.211529878364)/(o+1.)+(-0.912151985776)+(-0.0986461529033)*0.5**o
        ref[1,1,1]=(-1.32533998437)/(o+1.)+(-0.685030198386)+(-0.19904975372)*0.5**o
        ref[2,0,0]=(0.152048902469)/(o+1.)+(-0.88880590209)+(-0.201012473813)*0.5**o
        ref[2,0,1]=(-0.748900052246)/(o+1.)+(-0.0226120471998)+(0.720226011425)*0.5**o
        ref[2,1,0]=(0.77797707941)/(o+1.)+(1.01635055442)+(-0.557827762351)*0.5**o
        ref[2,1,1]=(1.08969334716)/(o+1.)+(-0.404644728844)+(0.793335846595)*0.5**o
        ref[3,0,0]=(1.56498448669)/(o+1.)+(-0.699967188685)+(0.293259591449)*0.5**o
        ref[3,0,1]=(-0.620133427128)/(o+1.)+(-0.550037413688)+(-0.238375915329)*0.5**o
        ref[3,1,0]=(-0.156380586288)/(o+1.)+(-0.0154369973565)+(-0.633594612687)*0.5**o
        ref[3,1,1]=(-0.352989559326)/(o+1.)+(-0.290547031627)+(0.710852984398)*0.5**o
        ref[4,0,0]=(-0.147579361355)/(o+1.)+(-0.615173424765)+(-0.00193837918766)*0.5**o
        ref[4,0,1]=(0.598152345277)/(o+1.)+(-0.540248606426)+(-0.559503106696)*0.5**o
        ref[4,1,0]=(-1.17182448825)/(o+1.)+(-0.590551681576)+(-0.401073317208)*0.5**o
        ref[4,1,1]=(0.837504141173)/(o+1.)+(0.0807414267186)+(-0.15185865216)*0.5**o
        ref[5,0,0]=(-0.646601115533)/(o+1.)+(-1.20895265273)+(0.9173049337)*0.5**o
        ref[5,0,1]=(1.26626110052)/(o+1.)+(-0.56471923973)+(-0.880046978955)*0.5**o
        ref[5,1,0]=(-0.069063169298)/(o+1.)+(0.883536796964)+(0.883025165644)*0.5**o
        ref[5,1,1]=(1.05999338993)/(o+1.)+(-0.392519102893)+(0.918128198439)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.257898226293)*x[0]+(-0.01230083314)*x[1]
        arg[0,0,0,1]=(0.0531183070647)*x[0]+(-0.170192287836)*x[1]
        arg[0,0,1,0]=(-0.364644940864)*x[0]+(1.04227651446)*x[1]
        arg[0,0,1,1]=(-0.894334108203)*x[0]+(0.151390331211)*x[1]
        arg[0,0,2,0]=(1.03825401708)*x[0]+(-0.882136097149)*x[1]
        arg[0,0,2,1]=(-1.05304910155)*x[0]+(1.06866311708)*x[1]
        arg[0,1,0,0]=(1.54742542159)*x[0]+(0.941778348482)*x[1]
        arg[0,1,0,1]=(1.1797380956)*x[0]+(-0.296417138234)*x[1]
        arg[0,1,1,0]=(-1.09196028027)*x[0]+(-0.783377624006)*x[1]
        arg[0,1,1,1]=(0.370952191443)*x[0]+(-0.0854141858991)*x[1]
        arg[0,1,2,0]=(-1.23552368139)*x[0]+(1.23848088087)*x[1]
        arg[0,1,2,1]=(0.216345350114)*x[0]+(-0.136524468427)*x[1]
        arg[0,2,0,0]=(-1.21206901263)*x[0]+(0.804219582144)*x[1]
        arg[0,2,0,1]=(0.192956526208)*x[0]+(0.588417678858)*x[1]
        arg[0,2,1,0]=(-0.940175603742)*x[0]+(0.329047260784)*x[1]
        arg[0,2,1,1]=(-0.138597387792)*x[0]+(0.398298326223)*x[1]
        arg[0,2,2,0]=(0.10089766824)*x[0]+(1.25459493298)*x[1]
        arg[0,2,2,1]=(-0.401163525883)*x[0]+(0.276542765282)*x[1]
        arg[0,3,0,0]=(0.252156795268)*x[0]+(-0.50223063539)*x[1]
        arg[0,3,0,1]=(0.466244610905)*x[0]+(-0.587025224301)*x[1]
        arg[0,3,1,0]=(-0.268082012713)*x[0]+(-1.28874675764)*x[1]
        arg[0,3,1,1]=(1.10867463519)*x[0]+(-0.165500912864)*x[1]
        arg[0,3,2,0]=(-1.94681908266)*x[0]+(0.166854108476)*x[1]
        arg[0,3,2,1]=(-0.537150417922)*x[0]+(0.0872851457131)*x[1]
        arg[0,4,0,0]=(-0.376548581807)*x[0]+(-0.245201133388)*x[1]
        arg[0,4,0,1]=(-0.135287690767)*x[0]+(0.750376238942)*x[1]
        arg[0,4,1,0]=(-0.0240491596907)*x[0]+(-1.21337088424)*x[1]
        arg[0,4,1,1]=(0.699215733397)*x[0]+(0.0596356032853)*x[1]
        arg[0,4,2,0]=(-0.521316010702)*x[0]+(-1.75309802615)*x[1]
        arg[0,4,2,1]=(0.403640172305)*x[0]+(1.41465827077)*x[1]
        arg[1,0,0,0]=(-1.01538089315)*x[0]+(-1.14785578909)*x[1]
        arg[1,0,0,1]=(1.10045201832)*x[0]+(0.630761728795)*x[1]
        arg[1,0,1,0]=(0.137673646147)*x[0]+(-0.387129090253)*x[1]
        arg[1,0,1,1]=(-1.04588287863)*x[0]+(-1.22433029859)*x[1]
        arg[1,0,2,0]=(-0.567830263535)*x[0]+(-0.405808616295)*x[1]
        arg[1,0,2,1]=(0.139326102495)*x[0]+(-1.93153384363)*x[1]
        arg[1,1,0,0]=(-0.638524762921)*x[0]+(0.506405751653)*x[1]
        arg[1,1,0,1]=(0.130775153144)*x[0]+(0.531317202208)*x[1]
        arg[1,1,1,0]=(0.557476206038)*x[0]+(0.447588676184)*x[1]
        arg[1,1,1,1]=(-0.538795916875)*x[0]+(-0.593869493531)*x[1]
        arg[1,1,2,0]=(-1.09827483942)*x[0]+(0.376431061876)*x[1]
        arg[1,1,2,1]=(-0.0074505620387)*x[0]+(-0.798645151486)*x[1]
        arg[1,2,0,0]=(0.251283911459)*x[0]+(0.194595071813)*x[1]
        arg[1,2,0,1]=(1.00482763799)*x[0]+(-1.00059863394)*x[1]
        arg[1,2,1,0]=(-0.453095434684)*x[0]+(-0.513922011418)*x[1]
        arg[1,2,1,1]=(0.558548837484)*x[0]+(0.0696328406825)*x[1]
        arg[1,2,2,0]=(-1.20625552545)*x[0]+(-0.151371710692)*x[1]
        arg[1,2,2,1]=(0.100262209656)*x[0]+(-0.314628804179)*x[1]
        arg[1,3,0,0]=(0.947063141882)*x[0]+(-0.297271200562)*x[1]
        arg[1,3,0,1]=(0.0760277251261)*x[0]+(-1.17982146387)*x[1]
        arg[1,3,1,0]=(-1.46946063969)*x[0]+(-0.615302663567)*x[1]
        arg[1,3,1,1]=(0.419747293729)*x[0]+(-0.629842531949)*x[1]
        arg[1,3,2,0]=(-1.44607954917)*x[0]+(0.520024093924)*x[1]
        arg[1,3,2,1]=(-1.01556392596)*x[0]+(-0.307173811354)*x[1]
        arg[1,4,0,0]=(-0.164339335244)*x[0]+(0.390736414982)*x[1]
        arg[1,4,0,1]=(-0.266932380288)*x[0]+(-0.731533848078)*x[1]
        arg[1,4,1,0]=(-0.903825489055)*x[0]+(-1.81564784421)*x[1]
        arg[1,4,1,1]=(-0.426008882665)*x[0]+(-1.09792948908)*x[1]
        arg[1,4,2,0]=(-1.32984927589)*x[0]+(1.58196302541)*x[1]
        arg[1,4,2,1]=(-0.0581032806145)*x[0]+(-0.323073355302)*x[1]
        arg[2,0,0,0]=(-0.744014251373)*x[0]+(-0.818253254872)*x[1]
        arg[2,0,0,1]=(0.368656965396)*x[0]+(-0.0864646380237)*x[1]
        arg[2,0,1,0]=(-0.112428845773)*x[0]+(0.109170104421)*x[1]
        arg[2,0,1,1]=(0.0682241548911)*x[0]+(-1.65999482283)*x[1]
        arg[2,0,2,0]=(0.212777058993)*x[0]+(0.127163480945)*x[1]
        arg[2,0,2,1]=(-0.256989443356)*x[0]+(0.183206687414)*x[1]
        arg[2,1,0,0]=(0.578997994373)*x[0]+(0.413087453689)*x[1]
        arg[2,1,0,1]=(0.886532090807)*x[0]+(0.330645525532)*x[1]
        arg[2,1,1,0]=(0.18224742814)*x[0]+(-0.713166591262)*x[1]
        arg[2,1,1,1]=(-0.25218296706)*x[0]+(0.736797897014)*x[1]
        arg[2,1,2,0]=(0.885270371408)*x[0]+(-1.35641397127)*x[1]
        arg[2,1,2,1]=(-0.963214290879)*x[0]+(0.348421598509)*x[1]
        arg[2,2,0,0]=(0.884120745351)*x[0]+(0.956857507884)*x[1]
        arg[2,2,0,1]=(-0.00795298517561)*x[0]+(0.51243740322)*x[1]
        arg[2,2,1,0]=(-0.563655252975)*x[0]+(-0.716305131721)*x[1]
        arg[2,2,1,1]=(0.648243491623)*x[0]+(-0.142239786485)*x[1]
        arg[2,2,2,0]=(1.50677848542)*x[0]+(-0.359722006818)*x[1]
        arg[2,2,2,1]=(-0.522592086302)*x[0]+(-0.18544927497)*x[1]
        arg[2,3,0,0]=(0.0822787762966)*x[0]+(0.196309662733)*x[1]
        arg[2,3,0,1]=(0.232573258627)*x[0]+(-0.0597976775432)*x[1]
        arg[2,3,1,0]=(1.50895737052)*x[0]+(0.546625934312)*x[1]
        arg[2,3,1,1]=(-1.64023373286)*x[0]+(0.220053261777)*x[1]
        arg[2,3,2,0]=(1.5897208842)*x[0]+(-0.0355885594992)*x[1]
        arg[2,3,2,1]=(-1.52822349982)*x[0]+(-0.0906934717656)*x[1]
        arg[2,4,0,0]=(0.496696582601)*x[0]+(-0.303888675059)*x[1]
        arg[2,4,0,1]=(-0.0819586112177)*x[0]+(-0.0449047294378)*x[1]
        arg[2,4,1,0]=(-0.678530758513)*x[0]+(-1.2173963447)*x[1]
        arg[2,4,1,1]=(-0.645450475268)*x[0]+(0.356771113426)*x[1]
        arg[2,4,2,0]=(-0.486643021867)*x[0]+(0.977763869689)*x[1]
        arg[2,4,2,1]=(-0.458991807397)*x[0]+(1.80038432923)*x[1]
        arg[3,0,0,0]=(0.270795948755)*x[0]+(-0.691512096338)*x[1]
        arg[3,0,0,1]=(-0.756717130099)*x[0]+(1.16189640534)*x[1]
        arg[3,0,1,0]=(0.967554346699)*x[0]+(-0.272119008963)*x[1]
        arg[3,0,1,1]=(1.14296491195)*x[0]+(-0.707175662379)*x[1]
        arg[3,0,2,0]=(1.15665030692)*x[0]+(0.414803389409)*x[1]
        arg[3,0,2,1]=(1.38997274955)*x[0]+(1.07065568192)*x[1]
        arg[3,1,0,0]=(-0.623998925511)*x[0]+(0.375545806421)*x[1]
        arg[3,1,0,1]=(-0.363431948398)*x[0]+(-0.183100417297)*x[1]
        arg[3,1,1,0]=(0.0435866301729)*x[0]+(-0.557137572068)*x[1]
        arg[3,1,1,1]=(-0.253225217468)*x[0]+(-0.0657932188329)*x[1]
        arg[3,1,2,0]=(0.147700921064)*x[0]+(0.681624563015)*x[1]
        arg[3,1,2,1]=(-0.075394297616)*x[0]+(0.271741245789)*x[1]
        arg[3,2,0,0]=(1.34803044812)*x[0]+(0.87991065609)*x[1]
        arg[3,2,0,1]=(0.0529487135353)*x[0]+(0.806480092352)*x[1]
        arg[3,2,1,0]=(-1.52038624166)*x[0]+(-0.992718138685)*x[1]
        arg[3,2,1,1]=(0.0859852188749)*x[0]+(1.63512596202)*x[1]
        arg[3,2,2,0]=(1.46925148575)*x[0]+(-1.60383564005)*x[1]
        arg[3,2,2,1]=(0.114463395473)*x[0]+(-1.0186377376)*x[1]
        arg[3,3,0,0]=(-1.02962502658)*x[0]+(-1.06788543146)*x[1]
        arg[3,3,0,1]=(-0.184391597652)*x[0]+(0.462603936483)*x[1]
        arg[3,3,1,0]=(0.611113335782)*x[0]+(0.282417397043)*x[1]
        arg[3,3,1,1]=(-0.193423873796)*x[0]+(-0.56108614964)*x[1]
        arg[3,3,2,0]=(-0.0290687598224)*x[0]+(-0.460221158417)*x[1]
        arg[3,3,2,1]=(0.784513522959)*x[0]+(-0.484813129206)*x[1]
        arg[3,4,0,0]=(0.423780748664)*x[0]+(-0.344020787894)*x[1]
        arg[3,4,0,1]=(1.10623024451)*x[0]+(-0.775008153301)*x[1]
        arg[3,4,1,0]=(0.554847534291)*x[0]+(-0.854749107999)*x[1]
        arg[3,4,1,1]=(-0.639595693597)*x[0]+(-0.966152813212)*x[1]
        arg[3,4,2,0]=(-0.0172621996023)*x[0]+(0.321815074269)*x[1]
        arg[3,4,2,1]=(0.39605156824)*x[0]+(-0.499138231749)*x[1]
        ref[0,0,0,0]=(0.169892838337)/(o+1.)+(-0.323261281472)+(0.722227117759)*0.5**o
        ref[0,0,0,1]=(-0.899797330597)/(o+1.)+(0.731242549464)+(-0.679761749104)*0.5**o
        ref[0,0,1,0]=(0.948969615551)/(o+1.)+(-0.173478478707)+(0.0756189154575)*0.5**o
        ref[0,0,1,1]=(-0.484419525942)/(o+1.)+(0.287761319651)+(-0.834046890353)*0.5**o
        ref[0,0,2,0]=(-0.275366085384)/(o+1.)+(-0.051636507926)+(0.534757021171)*0.5**o
        ref[0,0,2,1]=(0.338862631582)/(o+1.)+(0.076489529483)+(-0.47622767501)*0.5**o
        ref[0,1,0,0]=(0.960885993642)/(o+1.)+(0.36959166751)+(0.789134441414)*0.5**o
        ref[0,1,0,1]=(-0.101908900511)/(o+1.)+(0.343409131566)+(0.298411594744)*0.5**o
        ref[0,1,1,0]=(-0.50227382988)/(o+1.)+(-0.431549057432)+(-0.509965959533)*0.5**o
        ref[0,1,1,1]=(-0.294216689224)/(o+1.)+(0.293114907564)+(-0.00647512035965)*0.5**o
        ref[0,1,2,0]=(0.330313139948)/(o+1.)+(0.0601559902451)+(-0.447667920962)*0.5**o
        ref[0,1,2,1]=(-0.855325117375)/(o+1.)+(0.455706690618)+(0.0237326178252)*0.5**o
        ref[0,2,0,0]=(0.958948557218)/(o+1.)+(-0.307015973897)+(-0.752766039908)*0.5**o
        ref[0,2,0,1]=(0.622666086232)/(o+1.)+(0.417782845058)+(-0.676857571283)*0.5**o
        ref[0,2,1,0]=(-0.476269525636)/(o+1.)+(0.203518519787)+(-0.541895856898)*0.5**o
        ref[0,2,1,1]=(0.396862410812)/(o+1.)+(-0.318442063423)+(0.499722654464)*0.5**o
        ref[0,2,2,0]=(0.291546173382)/(o+1.)+(0.345688404218)+(0.372569619405)*0.5**o
        ref[0,2,2,1]=(-0.0863432016899)/(o+1.)+(0.296494638577)+(-0.631266836064)*0.5**o
        ref[0,3,0,0]=(0.437179077921)/(o+1.)+(-0.340515851088)+(-0.00622121586734)*0.5**o
        ref[0,3,0,1]=(-0.0670787919003)/(o+1.)+(-0.492735838895)+(0.931769856293)*0.5**o
        ref[0,3,1,0]=(-0.455620785864)/(o+1.)+(-0.866886545624)+(0.632565106759)*0.5**o
        ref[0,3,1,1]=(0.769395985412)/(o+1.)+(-0.303642587154)+(0.781062911223)*0.5**o
        ref[0,3,2,0]=(-0.338160377272)/(o+1.)+(-0.237545833658)+(-0.966712929597)*0.5**o
        ref[0,3,2,1]=(-0.796435534783)/(o+1.)+(0.544922318125)+(-0.743274373676)*0.5**o
        ref[0,4,0,0]=(0.452932120968)/(o+1.)+(-0.512874769603)+(-0.048932296957)*0.5**o
        ref[0,4,0,1]=(0.361579034112)/(o+1.)+(-0.0561290568189)+(0.365767627701)*0.5**o
        ref[0,4,1,0]=(-0.626053161593)/(o+1.)+(-0.761324946295)+(0.911283010254)*0.5**o
        ref[0,4,1,1]=(-0.0455977379835)/(o+1.)+(0.292122388548)+(0.22020429757)*0.5**o
        ref[0,4,2,0]=(-0.881052488675)/(o+1.)+(-0.210171400955)+(-0.973018746263)*0.5**o
        ref[0,4,2,1]=(0.955475719603)/(o+1.)+(0.367959093676)+(0.126904536122)*0.5**o
        ref[1,0,0,0]=(-0.314523133291)/(o+1.)+(-0.649561146707)+(-0.54959125554)*0.5**o
        ref[1,0,0,1]=(0.776726613012)/(o+1.)+(0.0747896120381)+(0.804907910029)*0.5**o
        ref[1,0,1,0]=(0.416309554773)/(o+1.)+(0.0650606132082)+(-0.795886225295)*0.5**o
        ref[1,0,1,1]=(-0.642083920221)/(o+1.)+(-0.518812997818)+(-0.590503261366)*0.5**o
        ref[1,0,2,0]=(0.209400082886)/(o+1.)+(-0.592256267485)+(0.00147357225359)*0.5**o
        ref[1,0,2,1]=(-0.969945221754)/(o+1.)+(-0.364559871358)+(-0.0931427766681)*0.5**o
        ref[1,1,0,0]=(0.941138659497)/(o+1.)+(-0.103485523138)+(-0.866286624489)*0.5**o
        ref[1,1,0,1]=(0.527395399401)/(o+1.)+(0.373358049173)+(-0.612019142395)*0.5**o
        ref[1,1,1,0]=(-0.276942103917)/(o+1.)+(0.624509599487)+(0.0329877871648)*0.5**o
        ref[1,1,1,1]=(-0.153477829221)/(o+1.)+(-0.0616268433725)+(-0.85593389444)*0.5**o
        ref[1,1,2,0]=(0.2685285297)/(o+1.)+(-0.301824551915)+(-0.386723203413)*0.5**o
        ref[1,1,2,1]=(-0.530960074652)/(o+1.)+(0.248621312801)+(-0.772378264474)*0.5**o
        ref[1,2,0,0]=(0.745071931955)/(o+1.)+(-0.149268301427)+(-0.000656345828701)*0.5**o
        ref[1,2,0,1]=(-0.581822428087)/(o+1.)+(0.2315691847)+(0.122913062738)*0.5**o
        ref[1,2,1,0]=(-0.325165099021)/(o+1.)+(-0.414758139139)+(0.187663931198)*0.5**o
        ref[1,2,1,1]=(-0.389683593964)/(o+1.)+(0.265985790556)+(0.485893691018)*0.5**o
        ref[1,2,2,0]=(0.661969003098)/(o+1.)+(-0.514584521701)+(-0.990427195841)*0.5**o
        ref[1,2,2,1]=(-0.0828384828188)/(o+1.)+(-0.482166570368)+(0.832805029032)*0.5**o
        ref[1,3,0,0]=(-0.756020671713)/(o+1.)+(0.627704058228)+(0.150404496577)*0.5**o
        ref[1,3,0,1]=(-0.93648010541)/(o+1.)+(0.11022703478)+(-0.387767702896)*0.5**o
        ref[1,3,1,0]=(-0.995823513997)/(o+1.)+(-0.242072805569)+(-0.604794178119)*0.5**o
        ref[1,3,1,1]=(-0.543071878541)/(o+1.)+(-0.160890620667)+(0.654757881656)*0.5**o
        ref[1,3,2,0]=(0.803184441579)/(o+1.)+(-0.612293875048)+(-0.504652146723)*0.5**o
        ref[1,3,2,1]=(-0.786714439322)/(o+1.)+(-0.175556007517)+(-0.18491128296)*0.5**o
        ref[1,4,0,0]=(-0.302928744298)/(o+1.)+(0.476360987754)+(-0.423396151472)*0.5**o
        ref[1,4,0,1]=(-0.410240602476)/(o+1.)+(-0.49551630458)+(0.40280698327)*0.5**o
        ref[1,4,1,0]=(-0.820677400346)/(o+1.)+(-0.573037933363)+(-0.752720066188)*0.5**o
        ref[1,4,1,1]=(-0.492748923658)/(o+1.)+(-0.600739806799)+(0.170290165507)*0.5**o
        ref[1,4,2,0]=(0.926654770824)/(o+1.)+(0.011282908181)+(-0.697106837662)*0.5**o
        ref[1,4,2,1]=(-0.958477382611)/(o+1.)+(0.618515858902)+(-0.65973097111)*0.5**o
        ref[2,0,0,0]=(-0.15636950624)/(o+1.)+(-0.520670547262)+(-0.36455690548)*0.5**o
        ref[2,0,0,1]=(-0.726424878286)/(o+1.)+(0.697757734621)+(-0.386898263585)*0.5**o
        ref[2,0,1,0]=(-0.055162393508)/(o+1.)+(0.247015646426)+(-0.442127640696)*0.5**o
        ref[2,0,1,1]=(-0.769332288875)/(o+1.)+(-0.0917175733456)+(-0.639003232373)*0.5**o
        ref[2,0,2,0]=(-0.196390038297)/(o+1.)+(0.0977907270548)+(0.340749124126)*0.5**o
        ref[2,0,2,1]=(0.841927481859)/(o+1.)+(-0.489185426146)+(0.0626606144906)*0.5**o
        ref[2,1,0,0]=(0.590018249148)/(o+1.)+(0.378476967182)+(-0.354886735451)*0.5**o
        ref[2,1,0,1]=(0.178495604742)/(o+1.)+(0.233073081069)+(0.572535849459)*0.5**o
        ref[2,1,1,0]=(-0.762671064919)/(o+1.)+(-0.108539686142)+(0.448831274081)*0.5**o
        ref[2,1,1,1]=(0.927591406535)/(o+1.)+(0.0441635703243)+(-0.53130361723)*0.5**o
        ref[2,1,2,0]=(-0.604360706259)/(o+1.)+(-0.134848102332)+(0.402913311063)*0.5**o
        ref[2,1,2,1]=(0.229727627441)/(o+1.)+(-0.323296604293)+(-0.197927111225)*0.5**o
        ref[2,2,0,0]=(0.287478356112)/(o+1.)+(0.762635506201)+(0.0282288847209)*0.5**o
        ref[2,2,0,1]=(0.243518182621)/(o+1.)+(-0.241502069199)+(0.743970373822)*0.5**o
        ref[2,2,1,0]=(-0.369049760378)/(o+1.)+(-0.21941778965)+(-0.472075045017)*0.5**o
        ref[2,2,1,1]=(-0.745343658666)/(o+1.)+(0.193375557775)+(0.864596248255)*0.5**o
        ref[2,2,2,0]=(-0.491769251494)/(o+1.)+(0.442742387618)+(0.753340954862)*0.5**o
        ref[2,2,2,1]=(0.551196377158)/(o+1.)+(-0.523535406613)+(-0.212166925203)*0.5**o
        ref[2,3,0,0]=(0.192271522168)/(o+1.)+(-0.0181953944687)+(0.122707705798)*0.5**o
        ref[2,3,0,1]=(-0.593768738594)/(o+1.)+(0.199088910545)+(0.368366498588)*0.5**o
        ref[2,3,1,0]=(0.0866211808983)/(o+1.)+(0.723512694127)+(0.521936735685)*0.5**o
        ref[2,3,1,1]=(-0.453288268254)/(o+1.)+(-0.15580970281)+(-0.655272797205)*0.5**o
        ref[2,3,2,0]=(0.614105802669)/(o+1.)+(0.0411427694898)+(0.85774098305)*0.5**o
        ref[2,3,2,1]=(0.263133533654)/(o+1.)+(-0.628776355867)+(-0.624497793507)*0.5**o
        ref[2,4,0,0]=(-0.809347914722)/(o+1.)+(0.367719412237)+(0.26671699779)*0.5**o
        ref[2,4,0,1]=(0.773124137566)/(o+1.)+(-0.442496395542)+(-0.0149946871367)*0.5**o
        ref[2,4,1,0]=(-0.927311525126)/(o+1.)+(-0.0125177466568)+(-0.943580084777)*0.5**o
        ref[2,4,1,1]=(0.306409665483)/(o+1.)+(0.128569553604)+(-0.852228134533)*0.5**o
        ref[2,4,2,0]=(0.926185522145)/(o+1.)+(0.0750538146738)+(-0.58517230367)*0.5**o
        ref[2,4,2,1]=(0.858245470922)/(o+1.)+(0.0125876600893)+(0.457971730734)*0.5**o
        ref[3,0,0,0]=(-0.74501947953)/(o+1.)+(0.343942243752)+(-0.363581155556)*0.5**o
        ref[3,0,0,1]=(0.850890737948)/(o+1.)+(0.157997002133)+(-0.761705466968)*0.5**o
        ref[3,0,1,0]=(-0.0468962248591)/(o+1.)+(0.0297980532763)+(0.682735456043)*0.5**o
        ref[3,0,1,1]=(0.0126326425456)/(o+1.)+(-0.076115296096)+(0.575387199213)*0.5**o
        ref[3,0,2,0]=(-0.410549752333)/(o+1.)+(0.85556811978)+(0.270867209104)*0.5**o
        ref[3,0,2,1]=(0.580487848545)/(o+1.)+(0.59069171729)+(0.698757148346)*0.5**o
        ref[3,1,0,0]=(0.139052775786)/(o+1.)+(-0.334747475889)+(0.281989056901)*0.5**o
        ref[3,1,0,1]=(0.591083610998)/(o+1.)+(-0.322680889668)+(-0.492254197357)*0.5**o
        ref[3,1,1,0]=(0.384713965427)/(o+1.)+(-0.522411606102)+(0.146558304882)*0.5**o
        ref[3,1,1,1]=(-0.827615582009)/(o+1.)+(-0.00295218575854)+(0.514501517225)*0.5**o
        ref[3,1,2,0]=(0.84762578765)/(o+1.)+(0.208361555685)+(-0.43502341494)*0.5**o
        ref[3,1,2,1]=(-0.554668615664)/(o+1.)+(0.825347835483)+(-0.899680107129)*0.5**o
        ref[3,2,0,0]=(0.652802712143)/(o+1.)+(0.397197582731)+(0.7807432266)*0.5**o
        ref[3,2,0,1]=(0.421495251937)/(o+1.)+(-0.119571710512)+(0.677076974975)*0.5**o
        ref[3,2,1,0]=(-0.230439630375)/(o+1.)+(-0.733165339017)+(-0.816334071941)*0.5**o
        ref[3,2,1,1]=(0.64153308098)/(o+1.)+(0.778374176432)+(-0.477170252948)*0.5**o
        ref[3,2,2,0]=(-0.987848717403)/(o+1.)+(-0.0606435888479)+(0.974551740798)*0.5**o
        ref[3,2,2,1]=(-0.318593711859)/(o+1.)+(-0.459904454409)+(0.334228278552)*0.5**o
        ref[3,3,0,0]=(-0.418115413351)/(o+1.)+(-0.348616063735)+(-0.982162917221)*0.5**o
        ref[3,3,0,1]=(-0.437684435015)/(o+1.)+(0.804145379606)+(-0.892393985366)*0.5**o
        ref[3,3,1,0]=(0.944100110135)/(o+1.)+(-0.059994562245)+(0.0694197471801)*0.5**o
        ref[3,3,1,1]=(-0.860656048613)/(o+1.)+(0.178136109535)+(-0.250126193893)*0.5**o
        ref[3,3,2,0]=(-0.363631991908)/(o+1.)+(-0.298384287282)+(0.471110648232)*0.5**o
        ref[3,3,2,1]=(-0.578021736479)/(o+1.)+(0.320913094365)+(0.235895941501)*0.5**o
        ref[3,4,0,0]=(-0.946995649946)/(o+1.)+(0.0451978887455)+(0.936359833225)*0.5**o
        ref[3,4,0,1]=(-0.636305818291)/(o+1.)+(0.258932088634)+(0.449663732234)*0.5**o
        ref[3,4,1,0]=(-0.823421937366)/(o+1.)+(0.112980674356)+(0.297559014947)*0.5**o
        ref[3,4,1,1]=(-0.441768022626)/(o+1.)+(-0.240840962617)+(-0.682298558949)*0.5**o
        ref[3,4,2,0]=(-0.258808936822)/(o+1.)+(0.510306577882)+(-0.457251344276)*0.5**o
        ref[3,4,2,1]=(-0.489365749016)/(o+1.)+(-0.296617385332)+(0.979513856171)*0.5**o
      else:
        arg[0,0,0,0]=(-0.474247989728)*x[0]+(-0.768689352235)*x[1]+(-1.08527900853)*x[2]
        arg[0,0,0,1]=(-0.237113797789)*x[0]+(0.0842735741018)*x[1]+(0.203858551029)*x[2]
        arg[0,0,1,0]=(0.24237127665)*x[0]+(0.421022955803)*x[1]+(-1.52818540927)*x[2]
        arg[0,0,1,1]=(1.34624865817)*x[0]+(-0.798943272818)*x[1]+(-0.823168386315)*x[2]
        arg[0,0,2,0]=(-0.906723771267)*x[0]+(0.328893123771)*x[1]+(-0.942770202463)*x[2]
        arg[0,0,2,1]=(1.54305197694)*x[0]+(0.813063941984)*x[1]+(0.205532727168)*x[2]
        arg[0,1,0,0]=(0.399159979872)*x[0]+(-1.52089837934)*x[1]+(-1.4064251406)*x[2]
        arg[0,1,0,1]=(-0.121139354731)*x[0]+(1.12768246422)*x[1]+(0.234299409411)*x[2]
        arg[0,1,1,0]=(-1.62513899388)*x[0]+(-0.173131700583)*x[1]+(1.24199996799)*x[2]
        arg[0,1,1,1]=(-0.371929434177)*x[0]+(-0.606278610409)*x[1]+(-0.00600790063066)*x[2]
        arg[0,1,2,0]=(-0.804531294423)*x[0]+(0.869392248002)*x[1]+(0.840417359943)*x[2]
        arg[0,1,2,1]=(0.329452946275)*x[0]+(1.1072495446)*x[1]+(-0.213517932323)*x[2]
        arg[0,2,0,0]=(1.02784887224)*x[0]+(0.435517042337)*x[1]+(1.58767206677)*x[2]
        arg[0,2,0,1]=(-0.0120997177)*x[0]+(0.430135803314)*x[1]+(-0.717473068956)*x[2]
        arg[0,2,1,0]=(0.955455517302)*x[0]+(0.712976918968)*x[1]+(-0.624804508641)*x[2]
        arg[0,2,1,1]=(-1.03293986594)*x[0]+(-0.539062834482)*x[1]+(-1.61172701428)*x[2]
        arg[0,2,2,0]=(-0.177662810138)*x[0]+(0.553352366127)*x[1]+(0.695417121111)*x[2]
        arg[0,2,2,1]=(0.318133474679)*x[0]+(-0.452528947926)*x[1]+(0.468602505091)*x[2]
        arg[0,3,0,0]=(-0.294467725916)*x[0]+(1.67314409894)*x[1]+(0.100736440007)*x[2]
        arg[0,3,0,1]=(-0.517162349887)*x[0]+(-1.57430191254)*x[1]+(1.24045877402)*x[2]
        arg[0,3,1,0]=(0.667393640368)*x[0]+(0.340993632212)*x[1]+(-1.09920535559)*x[2]
        arg[0,3,1,1]=(0.37451406727)*x[0]+(0.37350649326)*x[1]+(0.357440871322)*x[2]
        arg[0,3,2,0]=(-0.989366398194)*x[0]+(0.596179154759)*x[1]+(-0.573280308087)*x[2]
        arg[0,3,2,1]=(-0.25561190141)*x[0]+(1.42296905238)*x[1]+(-1.21032031399)*x[2]
        arg[0,4,0,0]=(-0.41394085285)*x[0]+(-0.751213929646)*x[1]+(-0.68470718293)*x[2]
        arg[0,4,0,1]=(0.702985204167)*x[0]+(-0.0375821582587)*x[1]+(0.290630803819)*x[2]
        arg[0,4,1,0]=(0.0558204653441)*x[0]+(1.61940057302)*x[1]+(-0.518863528354)*x[2]
        arg[0,4,1,1]=(1.25876303604)*x[0]+(-1.40923202908)*x[1]+(1.05202333762)*x[2]
        arg[0,4,2,0]=(0.0715295483294)*x[0]+(0.883869702529)*x[1]+(-0.463253474408)*x[2]
        arg[0,4,2,1]=(0.564563789615)*x[0]+(-0.135049742988)*x[1]+(-1.3118878567)*x[2]
        arg[1,0,0,0]=(0.707267335291)*x[0]+(-0.197244449872)*x[1]+(-0.114362773781)*x[2]
        arg[1,0,0,1]=(-0.144762815659)*x[0]+(0.434177706452)*x[1]+(0.0754859099691)*x[2]
        arg[1,0,1,0]=(-0.906934991359)*x[0]+(0.179011995627)*x[1]+(0.823936036469)*x[2]
        arg[1,0,1,1]=(1.11165724605)*x[0]+(0.48023223731)*x[1]+(0.981353808181)*x[2]
        arg[1,0,2,0]=(0.499460583481)*x[0]+(0.171586243561)*x[1]+(-0.252474775657)*x[2]
        arg[1,0,2,1]=(-0.604289370884)*x[0]+(1.11610914086)*x[1]+(0.747439062711)*x[2]
        arg[1,1,0,0]=(0.696191463882)*x[0]+(0.349547714404)*x[1]+(0.651361019042)*x[2]
        arg[1,1,0,1]=(0.152679923426)*x[0]+(-0.448442911425)*x[1]+(1.33964371501)*x[2]
        arg[1,1,1,0]=(-0.74612592015)*x[0]+(-1.27089239786)*x[1]+(-0.261725631736)*x[2]
        arg[1,1,1,1]=(1.50839493929)*x[0]+(1.93229724987)*x[1]+(-0.963996845831)*x[2]
        arg[1,1,2,0]=(0.867984447375)*x[0]+(-0.777493811854)*x[1]+(0.847847553133)*x[2]
        arg[1,1,2,1]=(1.34249389498)*x[0]+(0.776160804062)*x[1]+(-0.295393972199)*x[2]
        arg[1,2,0,0]=(-0.655771613243)*x[0]+(1.58220214455)*x[1]+(-0.406292477629)*x[2]
        arg[1,2,0,1]=(0.59294895986)*x[0]+(0.544961164723)*x[1]+(0.634279670948)*x[2]
        arg[1,2,1,0]=(1.15324179862)*x[0]+(0.645764835217)*x[1]+(0.670364725309)*x[2]
        arg[1,2,1,1]=(-0.819459106271)*x[0]+(0.024815625404)*x[1]+(0.769753650676)*x[2]
        arg[1,2,2,0]=(1.19042366161)*x[0]+(0.982020548972)*x[1]+(0.92730609738)*x[2]
        arg[1,2,2,1]=(-0.742441488349)*x[0]+(-1.3101429812)*x[1]+(-0.198104246324)*x[2]
        arg[1,3,0,0]=(0.349205030437)*x[0]+(-0.745240339596)*x[1]+(-0.744742092259)*x[2]
        arg[1,3,0,1]=(0.993101296711)*x[0]+(-1.48331355036)*x[1]+(-1.00616778677)*x[2]
        arg[1,3,1,0]=(-1.10586442824)*x[0]+(1.71922449779)*x[1]+(-0.943451153259)*x[2]
        arg[1,3,1,1]=(-1.72393763035)*x[0]+(-1.14113495956)*x[1]+(-0.211909615809)*x[2]
        arg[1,3,2,0]=(-0.378302963108)*x[0]+(-0.494705322002)*x[1]+(0.940164363332)*x[2]
        arg[1,3,2,1]=(0.850541148743)*x[0]+(-1.45461945468)*x[1]+(1.20025821328)*x[2]
        arg[1,4,0,0]=(1.33882463795)*x[0]+(0.475208133119)*x[1]+(0.722686059253)*x[2]
        arg[1,4,0,1]=(0.0750422155969)*x[0]+(1.29758288845)*x[1]+(-1.41431668738)*x[2]
        arg[1,4,1,0]=(1.36072344575)*x[0]+(-0.345743547563)*x[1]+(-0.63248018754)*x[2]
        arg[1,4,1,1]=(0.579049229913)*x[0]+(0.954245074078)*x[1]+(-0.170538926965)*x[2]
        arg[1,4,2,0]=(1.67615484634)*x[0]+(0.561067019023)*x[1]+(-1.26952326914)*x[2]
        arg[1,4,2,1]=(0.700910547934)*x[0]+(0.58717328764)*x[1]+(0.795138196965)*x[2]
        arg[2,0,0,0]=(0.763193592359)*x[0]+(-0.397549328381)*x[1]+(-0.230593559545)*x[2]
        arg[2,0,0,1]=(0.454026153468)*x[0]+(-1.12464443717)*x[1]+(-0.147462098246)*x[2]
        arg[2,0,1,0]=(-0.276457034935)*x[0]+(0.118173156654)*x[1]+(0.57296366275)*x[2]
        arg[2,0,1,1]=(1.31887812253)*x[0]+(-0.544644483254)*x[1]+(0.0013728350582)*x[2]
        arg[2,0,2,0]=(-0.257589384265)*x[0]+(-0.016005453203)*x[1]+(-0.508836430965)*x[2]
        arg[2,0,2,1]=(-0.259420301773)*x[0]+(0.122645293168)*x[1]+(0.381613885289)*x[2]
        arg[2,1,0,0]=(-1.64168512418)*x[0]+(0.025182006192)*x[1]+(0.0268780458975)*x[2]
        arg[2,1,0,1]=(-0.0885344200898)*x[0]+(0.408795219114)*x[1]+(0.635798801606)*x[2]
        arg[2,1,1,0]=(-0.959658503283)*x[0]+(1.034415556)*x[1]+(0.725681105748)*x[2]
        arg[2,1,1,1]=(1.59312707789)*x[0]+(-0.116514952101)*x[1]+(-0.86121633507)*x[2]
        arg[2,1,2,0]=(-1.44859572067)*x[0]+(-1.28344247275)*x[1]+(0.299296401869)*x[2]
        arg[2,1,2,1]=(-0.488424372501)*x[0]+(1.23481402288)*x[1]+(0.01702902463)*x[2]
        arg[2,2,0,0]=(-1.29986951244)*x[0]+(-1.20190912751)*x[1]+(-0.755848291051)*x[2]
        arg[2,2,0,1]=(-0.52421656312)*x[0]+(-0.335471509689)*x[1]+(-1.01264312499)*x[2]
        arg[2,2,1,0]=(-1.12066642609)*x[0]+(-1.02138558053)*x[1]+(-1.00777755552)*x[2]
        arg[2,2,1,1]=(-1.42127024744)*x[0]+(1.63480022479)*x[1]+(0.542244906643)*x[2]
        arg[2,2,2,0]=(-0.418852880297)*x[0]+(-0.983110436453)*x[1]+(-0.622603673805)*x[2]
        arg[2,2,2,1]=(0.727366446825)*x[0]+(-0.256324856361)*x[1]+(-0.20608740758)*x[2]
        arg[2,3,0,0]=(0.548268302883)*x[0]+(0.0102173127665)*x[1]+(0.869833576217)*x[2]
        arg[2,3,0,1]=(-1.70159091061)*x[0]+(1.15876875083)*x[1]+(-0.234757706773)*x[2]
        arg[2,3,1,0]=(-1.72817024817)*x[0]+(0.690055132218)*x[1]+(-0.149398998825)*x[2]
        arg[2,3,1,1]=(1.80829412387)*x[0]+(-0.217899654855)*x[1]+(-0.922672217822)*x[2]
        arg[2,3,2,0]=(1.23855200328)*x[0]+(0.412795411508)*x[1]+(0.960476191755)*x[2]
        arg[2,3,2,1]=(-1.43061904621)*x[0]+(-0.766039000745)*x[1]+(-1.25765397361)*x[2]
        arg[2,4,0,0]=(-0.660872016687)*x[0]+(-0.000405233718708)*x[1]+(-0.178507968219)*x[2]
        arg[2,4,0,1]=(-0.168526882871)*x[0]+(-1.06936802685)*x[1]+(1.54083857104)*x[2]
        arg[2,4,1,0]=(0.565807359529)*x[0]+(-1.41401297939)*x[1]+(-1.23254260537)*x[2]
        arg[2,4,1,1]=(-0.373920075802)*x[0]+(0.407330812903)*x[1]+(-0.204667001328)*x[2]
        arg[2,4,2,0]=(0.874100879885)*x[0]+(-0.191238094267)*x[1]+(1.54890969328)*x[2]
        arg[2,4,2,1]=(0.566428125313)*x[0]+(-1.60899595601)*x[1]+(-1.05112529768)*x[2]
        arg[3,0,0,0]=(-0.688037248111)*x[0]+(-1.56483484784)*x[1]+(0.26782504619)*x[2]
        arg[3,0,0,1]=(0.418788954459)*x[0]+(-0.670455226092)*x[1]+(-0.0562760591375)*x[2]
        arg[3,0,1,0]=(-0.2614362754)*x[0]+(0.271050519855)*x[1]+(0.227811982923)*x[2]
        arg[3,0,1,1]=(-0.873198928647)*x[0]+(-0.483763141395)*x[1]+(0.33532383979)*x[2]
        arg[3,0,2,0]=(-0.864161938176)*x[0]+(-1.4682754386)*x[1]+(-0.38126411162)*x[2]
        arg[3,0,2,1]=(0.142827521381)*x[0]+(0.464483940188)*x[1]+(0.214297576606)*x[2]
        arg[3,1,0,0]=(0.188747079113)*x[0]+(0.590821253713)*x[1]+(0.908850367108)*x[2]
        arg[3,1,0,1]=(0.0587541582624)*x[0]+(0.955494266151)*x[1]+(1.70822209849)*x[2]
        arg[3,1,1,0]=(-0.420356861523)*x[0]+(-0.551445918262)*x[1]+(-0.0187114754269)*x[2]
        arg[3,1,1,1]=(1.07071658401)*x[0]+(-0.589035602194)*x[1]+(0.133725074027)*x[2]
        arg[3,1,2,0]=(0.370477043767)*x[0]+(0.226262353618)*x[1]+(0.566087784489)*x[2]
        arg[3,1,2,1]=(1.77828660075)*x[0]+(0.607419742685)*x[1]+(0.0297054744582)*x[2]
        arg[3,2,0,0]=(-1.19326939474)*x[0]+(1.20875153887)*x[1]+(-0.124756222091)*x[2]
        arg[3,2,0,1]=(0.0645190914271)*x[0]+(0.357580475716)*x[1]+(-1.59461469718)*x[2]
        arg[3,2,1,0]=(1.4853619905)*x[0]+(0.751582301648)*x[1]+(0.859363177167)*x[2]
        arg[3,2,1,1]=(1.36658655075)*x[0]+(-0.776650996613)*x[1]+(0.654743994261)*x[2]
        arg[3,2,2,0]=(0.406696005677)*x[0]+(-0.528546876384)*x[1]+(-0.439775325396)*x[2]
        arg[3,2,2,1]=(-0.725198159461)*x[0]+(-0.0933136564416)*x[1]+(-0.720365141393)*x[2]
        arg[3,3,0,0]=(-0.428669112669)*x[0]+(0.626162511175)*x[1]+(-0.382776233923)*x[2]
        arg[3,3,0,1]=(0.340474939218)*x[0]+(-0.140550383925)*x[1]+(0.353836983746)*x[2]
        arg[3,3,1,0]=(0.851646816711)*x[0]+(0.0171684195949)*x[1]+(0.38173044322)*x[2]
        arg[3,3,1,1]=(0.359647722225)*x[0]+(0.425799941165)*x[1]+(-0.80174052212)*x[2]
        arg[3,3,2,0]=(0.13012568736)*x[0]+(-0.573735693651)*x[1]+(1.08843730088)*x[2]
        arg[3,3,2,1]=(-1.01372953294)*x[0]+(0.335282956856)*x[1]+(1.24570512628)*x[2]
        arg[3,4,0,0]=(-0.05674609039)*x[0]+(-0.814879575064)*x[1]+(-0.169789408264)*x[2]
        arg[3,4,0,1]=(1.5968230938)*x[0]+(0.0679819277023)*x[1]+(0.892284314512)*x[2]
        arg[3,4,1,0]=(1.22220298478)*x[0]+(-0.0258417926472)*x[1]+(0.440670138568)*x[2]
        arg[3,4,1,1]=(0.647072768716)*x[0]+(-0.177651087906)*x[1]+(-0.115553644646)*x[2]
        arg[3,4,2,0]=(1.29816363306)*x[0]+(-1.18656044167)*x[1]+(-0.465776533157)*x[2]
        arg[3,4,2,1]=(0.328907496091)*x[0]+(-0.179857937151)*x[1]+(-0.476306938306)*x[2]
        ref[0,0,0,0]=(-0.530378359482)/(o+1.)+(-0.935143327523)+(0.0724486640358)*0.5**o
        ref[0,0,0,1]=(1.22837847871)/(o+1.)+(-0.113350954728)+(-0.950658241915)*0.5**o
        ref[0,0,1,0]=(-0.000414683535449)/(o+1.)+(-0.183158731552)+(-0.49805903018)*0.5**o
        ref[0,0,1,1]=(-0.882730389132)/(o+1.)+(-0.112174958816)+(0.831217305802)*0.5**o
        ref[0,0,2,0]=(-1.1913557966)/(o+1.)+(-0.0615750891074)+(-0.206094875146)*0.5**o
        ref[0,0,2,1]=(0.634536492589)/(o+1.)+(0.649890059464)+(0.627332034578)*0.5**o
        ref[0,1,0,0]=(-1.38665625775)/(o+1.)+(-1.0210653524)+(0.900623422471)*0.5**o
        ref[0,1,0,1]=(1.20633678024)/(o+1.)+(0.0888193121856)+(-0.143132885707)*0.5**o
        ref[0,1,1,0]=(0.452225278035)/(o+1.)+(-0.148926943801)+(-0.710642116901)*0.5**o
        ref[0,1,1,1]=(-0.244223430204)/(o+1.)+(-0.619607191996)+(0.499221868978)*0.5**o
        ref[0,1,2,0]=(0.0383849036606)/(o+1.)+(0.759484971644)+(-0.652076533427)*0.5**o
        ref[0,1,2,1]=(0.207558055759)/(o+1.)+(0.502305382384)+(0.0110157380215)*0.5**o
        ref[0,2,0,0]=(0.761374280144)/(o+1.)+(0.777616611515)+(0.734430478174)*0.5**o
        ref[0,2,0,1]=(0.554502862028)/(o+1.)+(-0.409274602016)+(-0.0353906413371)*0.5**o
        ref[0,2,1,0]=(-0.252543841709)/(o+1.)+(0.386640377894)+(0.522891013551)*0.5**o
        ref[0,2,1,1]=(-1.52497235887)/(o+1.)+(-0.807603254401)+(-0.0435508470367)*0.5**o
        ref[0,2,2,0]=(0.172485606257)/(o+1.)+(0.350297526187)+(0.19802601847)*0.5**o
        ref[0,2,2,1]=(-0.141271359379)/(o+1.)+(0.0509198134622)+(0.373638764298)*0.5**o
        ref[0,3,0,0]=(1.31278836809)/(o+1.)+(-0.0918739323256)+(0.350372309595)*0.5**o
        ref[0,3,0,1]=(-0.396933928109)/(o+1.)+(-0.436137805999)+(0.418204051696)*0.5**o
        ref[0,3,1,0]=(-0.206619840099)/(o+1.)+(-0.118703289798)+(0.353208336689)*0.5**o
        ref[0,3,1,1]=(-0.595153371293)/(o+1.)+(0.615722581562)+(0.469169640021)*0.5**o
        ref[0,3,2,0]=(-0.680858997986)/(o+1.)+(-0.0393223937964)+(-0.206963765944)*0.5**o
        ref[0,3,2,1]=(0.012753264598)/(o+1.)+(0.246188741489)+(-0.548093910595)*0.5**o
        ref[0,4,0,0]=(-0.40478987122)/(o+1.)+(-0.890399681609)+(0.33572726901)*0.5**o
        ref[0,4,0,1]=(0.0280062557264)/(o+1.)+(0.0358049413228)+(0.856417711355)*0.5**o
        ref[0,4,1,0]=(0.181560004565)/(o+1.)+(0.329885369854)+(0.31502676574)*0.5**o
        ref[0,4,1,1]=(0.0552681276841)/(o+1.)+(0.00372919426319)+(0.838827828375)*0.5**o
        ref[0,4,2,0]=(0.656057162724)/(o+1.)+(0.179315527684)+(-0.522542441641)*0.5**o
        ref[0,4,2,1]=(0.126189389317)/(o+1.)+(-0.429452001643)+(-0.149659196108)*0.5**o
        ref[1,0,0,0]=(1.43443485527)/(o+1.)+(-0.55091754491)+(0.0630603461901)*0.5**o
        ref[1,0,0,1]=(-0.612191779306)/(o+1.)+(0.604025235381)+(-0.230957890695)*0.5**o
        ref[1,0,1,0]=(0.373085423681)/(o+1.)+(-0.0817391530697)+(-0.113594076805)*0.5**o
        ref[1,0,1,1]=(0.761862677657)/(o+1.)+(0.819524484428)+(0.172331645033)*0.5**o
        ref[1,0,2,0]=(0.929265450746)/(o+1.)+(-0.382585784938)+(0.254478170515)*0.5**o
        ref[1,0,2,1]=(1.58236203463)/(o+1.)+(-0.280641188811)+(0.238179175679)*0.5**o
        ref[1,1,0,0]=(-0.117055726312)/(o+1.)+(0.957914090571)+(-0.101672257501)*0.5**o
        ref[1,1,0,1]=(0.0849062553796)/(o+1.)+(0.0703564566846)+(0.818261558262)*0.5**o
        ref[1,1,1,0]=(-1.72219113622)/(o+1.)+(-0.357454724939)+(0.15835663635)*0.5**o
        ref[1,1,1,1]=(0.739078400468)/(o+1.)+(0.532292543169)+(0.673031856531)*0.5**o
        ref[1,1,2,0]=(0.255023909903)/(o+1.)+(0.229656917655)+(0.22400044344)*0.5**o
        ref[1,1,2,1]=(0.518856765199)/(o+1.)+(0.380552790125)+(0.543298381396)*0.5**o
        ref[1,2,0,0]=(0.0321563884831)/(o+1.)+(0.25981567762)+(-0.0316496900482)*0.5**o
        ref[1,2,0,1]=(1.09318176771)/(o+1.)+(0.0572442907603)+(0.564519446305)*0.5**o
        ref[1,2,1,0]=(0.756962791387)/(o+1.)+(0.69419429624)+(0.32401997528)*0.5**o
        ref[1,2,1,1]=(1.28480299241)/(o+1.)+(-0.201990593541)+(-0.905711635518)*0.5**o
        ref[1,2,2,0]=(1.04969566442)/(o+1.)+(0.72854097372)+(0.592972696101)*0.5**o
        ref[1,2,2,1]=(-1.77086264775)/(o+1.)+(-0.261914180361)+(0.0440022926066)*0.5**o
        ref[1,3,0,0]=(-0.477780531124)/(o+1.)+(-0.181354085163)+(-0.300288699967)*0.5**o
        ref[1,3,0,1]=(-0.797627013456)/(o+1.)+(-0.523163955197)+(0.347574883435)*0.5**o
        ref[1,3,1,0]=(0.428806106821)/(o+1.)+(-0.0514584302394)+(-0.655980330052)*0.5**o
        ref[1,3,1,1]=(-0.141740459457)/(o+1.)+(-1.05699960931)+(-0.821242527644)*0.5**o
        ref[1,3,2,0]=(0.0181810985873)/(o+1.)+(-0.180697139553)+(0.41036925874)*0.5**o
        ref[1,3,2,1]=(-0.061077367)/(o+1.)+(0.250077446763)+(0.157102380812)*0.5**o
        ref[1,4,0,0]=(1.14087212913)/(o+1.)+(0.368059029864)+(0.659728641463)*0.5**o
        ref[1,4,0,1]=(0.365658713316)/(o+1.)+(0.088533821044)+(-0.584417938734)*0.5**o
        ref[1,4,1,0]=(-1.11681308321)/(o+1.)+(0.438275560627)+(0.622761672604)*0.5**o
        ref[1,4,1,1]=(1.18893372018)/(o+1.)+(0.172934240923)+(-0.172046824999)*0.5**o
        ref[1,4,2,0]=(0.16240489493)/(o+1.)+(-0.0145923852256)+(0.834478471749)*0.5**o
        ref[1,4,2,1]=(0.571355829378)/(o+1.)+(0.526896417749)+(0.458073367663)*0.5**o
        ref[2,0,0,0]=(-0.489234572937)/(o+1.)+(-0.0135589093901)+(0.65140309615)*0.5**o
        ref[2,0,0,1]=(-1.55383491931)/(o+1.)+(0.0161644619972)+(0.703425613366)*0.5**o
        ref[2,0,1,0]=(-0.258604855723)/(o+1.)+(0.0827812678671)+(0.507722104457)*0.5**o
        ref[2,0,1,1]=(0.0504178529481)/(o+1.)+(0.182991850332)+(0.359204920726)*0.5**o
        ref[2,0,2,0]=(0.652043479841)/(o+1.)+(-0.443015450013)+(-0.548443848248)*0.5**o
        ref[2,0,2,1]=(0.340317839857)/(o+1.)+(-0.379565594545)+(0.663652225917)*0.5**o
        ref[2,1,0,0]=(-0.587863196667)/(o+1.)+(-0.0137797674284)+(-0.974202340562)*0.5**o
        ref[2,1,0,1]=(1.00030747039)/(o+1.)+(-0.399997754768)+(0.755747639771)*0.5**o
        ref[2,1,1,0]=(0.703413281036)/(o+1.)+(0.487016208736)+(-0.877007540039)*0.5**o
        ref[2,1,1,1]=(0.263166830572)/(o+1.)+(-0.157259452849)+(0.666747865841)*0.5**o
        ref[2,1,2,0]=(0.104532343004)/(o+1.)+(-1.02548001493)+(-0.486314104701)*0.5**o
        ref[2,1,2,1]=(0.933518655478)/(o+1.)+(0.0369928229139)+(-0.244085626298)*0.5**o
        ref[2,2,0,0]=(-1.50550384519)/(o+1.)+(-0.590996227594)+(-0.570130630622)*0.5**o
        ref[2,2,0,1]=(-0.60318683839)/(o+1.)+(-0.38308245528)+(-0.502979448853)*0.5**o
        ref[2,2,1,0]=(-1.11258910879)/(o+1.)+(-0.518777277969)+(-0.999685897413)*0.5**o
        ref[2,2,1,1]=(1.84366617107)/(o+1.)+(-0.232866738261)+(-0.622157810558)*0.5**o
        ref[2,2,2,0]=(-0.620181988585)/(o+1.)+(-0.836162444183)+(0.267939886396)*0.5**o
        ref[2,2,2,1]=(-0.826647389099)/(o+1.)+(0.0929536518109)+(0.905694268362)*0.5**o
        ref[2,3,0,0]=(0.184067674443)/(o+1.)+(0.259745858757)+(0.724759799909)*0.5**o
        ref[2,3,0,1]=(0.0175595303049)/(o+1.)+(0.0560117358912)+(-0.907162868639)*0.5**o
        ref[2,3,1,0]=(-0.609443154617)/(o+1.)+(0.186540819401)+(-0.951152598958)*0.5**o
        ref[2,3,1,1]=(-1.61858082094)/(o+1.)+(0.679439858578)+(0.927423354971)*0.5**o
        ref[2,3,2,0]=(0.44350652486)/(o+1.)+(0.81466177481)+(0.53899353207)*0.5**o
        ref[2,3,2,1]=(-0.877289862245)/(o+1.)+(-1.05742260849)+(-0.462176941342)*0.5**o
        ref[2,4,0,0]=(-0.736381416539)/(o+1.)+(-0.0139374259612)+(-0.0755289501641)*0.5**o
        ref[2,4,0,1]=(0.690429326999)/(o+1.)+(-0.0293644018251)+(-0.328756862036)*0.5**o
        ref[2,4,1,0]=(-1.57117034002)/(o+1.)+(-0.531463765387)+(0.553349645561)*0.5**o
        ref[2,4,1,1]=(1.24898175282)/(o+1.)+(-0.452273024959)+(-0.515691967125)*0.5**o
        ref[2,4,2,0]=(0.239126191643)/(o+1.)+(0.722504563932)+(0.547637159392)*0.5**o
        ref[2,4,2,1]=(-1.40763254209)/(o+1.)+(-0.809456441012)+(0.932852295739)*0.5**o
        ref[3,0,0,0]=(-0.416554297686)/(o+1.)+(-0.632193610911)+(-0.304105530258)*0.5**o
        ref[3,0,0,1]=(-1.02983448036)/(o+1.)+(0.579212231778)+(-0.436532313966)*0.5**o
        ref[3,0,1,0]=(0.42124133216)/(o+1.)+(0.368290418827)+(-0.920395942436)*0.5**o
        ref[3,0,1,1]=(0.158195474854)/(o+1.)+(-0.17459474514)+(-0.830644214826)*0.5**o
        ref[3,0,2,0]=(-1.18749647329)/(o+1.)+(-0.506173996178)+(-0.513857022748)*0.5**o
        ref[3,0,2,1]=(0.233437753018)/(o+1.)+(0.455705145371)+(-0.323239005586)*0.5**o
        ref[3,1,0,0]=(1.42991958566)/(o+1.)+(0.053769921022)+(0.150959272234)*0.5**o
        ref[3,1,0,1]=(1.48423731111)/(o+1.)+(0.964048315997)+(-0.689863420196)*0.5**o
        ref[3,1,1,0]=(0.868837676765)/(o+1.)+(-1.01261543254)+(0.165878933098)*0.5**o
        ref[3,1,1,1]=(0.820750307831)/(o+1.)+(-0.300930223821)+(0.396516195659)*0.5**o
        ref[3,1,2,0]=(0.477077782202)/(o+1.)+(-0.0416336989494)+(0.769016797571)*0.5**o
        ref[3,1,2,1]=(0.359895440587)/(o+1.)+(0.555322395811)+(0.944871585683)*0.5**o
        ref[3,2,0,0]=(0.657668315832)/(o+1.)+(-0.240321895174)+(-0.286298603451)*0.5**o
        ref[3,2,0,1]=(-0.941360568282)/(o+1.)+(-0.0246544136593)+(-0.181845734439)*0.5**o
        ref[3,2,1,0]=(1.60316834971)/(o+1.)+(0.252291602298)+(0.988555915005)*0.5**o
        ref[3,2,1,1]=(0.833811957678)/(o+1.)+(-0.111970621871)+(0.634808834458)*0.5**o
        ref[3,2,2,0]=(-1.38877073784)/(o+1.)+(0.538356636307)+(-0.249568730877)*0.5**o
        ref[3,2,2,1]=(-0.311839184097)/(o+1.)+(-0.698114103264)+(0.169190433329)*0.5**o
        ref[3,3,0,0]=(0.743170989696)/(o+1.)+(-0.499224746788)+(0.0699956684628)*0.5**o
        ref[3,3,0,1]=(-0.293538786051)/(o+1.)+(0.26667992866)+(0.313940467768)*0.5**o
        ref[3,3,1,0]=(0.951483461345)/(o+1.)+(0.148317904561)+(0.00242640905887)*0.5**o
        ref[3,3,1,1]=(-0.110753912776)/(o+1.)+(-0.44308689138)+(0.980634836805)*0.5**o
        ref[3,3,2,0]=(-0.337895223544)/(o+1.)+(0.358941646037)+(0.264839226057)*0.5**o
        ref[3,3,2,1]=(0.857823611315)/(o+1.)+(0.30404611512)+(-0.898657291359)*0.5**o
        ref[3,4,0,0]=(0.037944509376)/(o+1.)+(-0.364801700413)+(-0.349756182268)*0.5**o
        ref[3,4,0,1]=(1.01114399396)/(o+1.)+(0.457103279485)+(0.631738783076)*0.5**o
        ref[3,4,1,0]=(0.969931913888)/(o+1.)+(-0.165922085756)+(0.998943588321)*0.5**o
        ref[3,4,1,1]=(1.35441114248)/(o+1.)+(-0.502422727199)+(0.00430234808496)*0.5**o
        ref[3,4,2,0]=(-0.367201097566)/(o+1.)+(-0.337780780531)+(0.688589316866)*0.5**o
        ref[3,4,2,1]=(-0.0547905846061)/(o+1.)+(-0.293870420685)+(0.31527404661)*0.5**o
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedFunctionOnContactOne(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.311787641998)*x[0]**o+(0.161673282875)*x[0]+(0.32570909237)*x[1]**o+(-0.249404742764)*x[1]
        ref=(0.32570909237)/(o+1.)+(-0.0438657299447)+(0.311787641998)*0.5**o
      else:
        arg=(0.0675036998003)*x[0]**o+(-0.247751392136)*x[0]+(0.791676220029)*x[1]**o+(-0.269015236163)*x[1]+(0.265735156875)*x[2]**o+(0.820319192519)*x[2]
        ref=(1.0574113769)/(o+1.)+(0.15177628211)+(0.0675036998003)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedFunctionOnContactOne(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.99235950851)*x[0]**o+(0.00867392831518)*x[0]+(-0.598024390869)*x[1]**o+(0.00945232777857)*x[1]
        arg[1]=(-0.140261338996)*x[0]**o+(-0.166524016889)*x[0]+(0.59850076881)*x[1]**o+(-0.805756215354)*x[1]
        ref[0]=(-0.598024390869)/(o+1.)+(0.00906312804687)+(0.99235950851)*0.5**o
        ref[1]=(0.59850076881)/(o+1.)+(-0.486140116122)+(-0.140261338996)*0.5**o
      else:
        arg[0]=(-0.158686556523)*x[0]**o+(0.249193164275)*x[0]+(-0.801383648208)*x[1]**o+(-0.439021533504)*x[1]+(-0.112035191366)*x[2]**o+(-0.323385862625)*x[2]
        arg[1]=(0.392997873484)*x[0]**o+(-0.867468854104)*x[0]+(0.804459243476)*x[1]**o+(0.129028860314)*x[1]+(-0.252715502862)*x[2]**o+(0.457355923469)*x[2]
        ref[0]=(-0.913418839574)/(o+1.)+(-0.256607115927)+(-0.158686556523)*0.5**o
        ref[1]=(0.551743740614)/(o+1.)+(-0.140542035161)+(0.392997873484)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedFunctionOnContactOne(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.330520725992)*x[0]**o+(0.160118740246)*x[0]+(-0.178670895269)*x[1]**o+(-0.281660040648)*x[1]
        arg[0,1]=(-0.871801084852)*x[0]**o+(0.692549035429)*x[0]+(-0.265102051042)*x[1]**o+(-0.799752901882)*x[1]
        arg[0,2]=(-0.816985024574)*x[0]**o+(0.534242721856)*x[0]+(0.81492666918)*x[1]**o+(0.891189158633)*x[1]
        arg[0,3]=(-0.20728122917)*x[0]**o+(-0.656873565737)*x[0]+(-0.924660993906)*x[1]**o+(0.453699593426)*x[1]
        arg[0,4]=(-0.541377726375)*x[0]**o+(0.916287049432)*x[0]+(-0.15879052465)*x[1]**o+(-0.319875161126)*x[1]
        arg[1,0]=(0.296966289838)*x[0]**o+(0.170347587334)*x[0]+(0.755207250839)*x[1]**o+(0.677658805212)*x[1]
        arg[1,1]=(-0.507591010573)*x[0]**o+(0.299638021889)*x[0]+(-0.757702071557)*x[1]**o+(-0.194564860748)*x[1]
        arg[1,2]=(0.821972467197)*x[0]**o+(-0.75902704466)*x[0]+(0.19424332718)*x[1]**o+(-0.070182646035)*x[1]
        arg[1,3]=(-0.340805687409)*x[0]**o+(0.585961973082)*x[0]+(-0.602942732318)*x[1]**o+(-0.288941395094)*x[1]
        arg[1,4]=(0.0314374795024)*x[0]**o+(0.293710824013)*x[0]+(0.576509865762)*x[1]**o+(-0.481377812302)*x[1]
        arg[2,0]=(0.250117528833)*x[0]**o+(-0.318419344173)*x[0]+(-0.793830015706)*x[1]**o+(0.0445553063586)*x[1]
        arg[2,1]=(-0.347974394766)*x[0]**o+(-0.569347105354)*x[0]+(0.838687005759)*x[1]**o+(0.534306628908)*x[1]
        arg[2,2]=(0.53063272248)*x[0]**o+(0.426504428158)*x[0]+(0.215398162753)*x[1]**o+(0.324348187788)*x[1]
        arg[2,3]=(-0.336752121527)*x[0]**o+(0.666074768462)*x[0]+(-0.739123251932)*x[1]**o+(-0.538899653147)*x[1]
        arg[2,4]=(-0.832783810519)*x[0]**o+(0.558572116384)*x[0]+(0.119988536199)*x[1]**o+(-0.664170918761)*x[1]
        arg[3,0]=(0.4941356722)*x[0]**o+(-0.403591783006)*x[0]+(0.205050527245)*x[1]**o+(0.351642574939)*x[1]
        arg[3,1]=(0.12538190695)*x[0]**o+(-0.890722638937)*x[0]+(-0.897011486902)*x[1]**o+(-0.465522770696)*x[1]
        arg[3,2]=(-0.862829007552)*x[0]**o+(-0.950531038019)*x[0]+(-0.970016379643)*x[1]**o+(-0.500580755888)*x[1]
        arg[3,3]=(0.131451743264)*x[0]**o+(0.43845495141)*x[0]+(-0.929808474589)*x[1]**o+(-0.104714452616)*x[1]
        arg[3,4]=(-0.81935224348)*x[0]**o+(-0.53791483312)*x[0]+(-0.714512816761)*x[1]**o+(0.329859667225)*x[1]
        ref[0,0]=(-0.178670895269)/(o+1.)+(-0.0607706502011)+(0.330520725992)*0.5**o
        ref[0,1]=(-0.265102051042)/(o+1.)+(-0.0536019332266)+(-0.871801084852)*0.5**o
        ref[0,2]=(0.81492666918)/(o+1.)+(0.712715940244)+(-0.816985024574)*0.5**o
        ref[0,3]=(-0.924660993906)/(o+1.)+(-0.101586986156)+(-0.20728122917)*0.5**o
        ref[0,4]=(-0.15879052465)/(o+1.)+(0.298205944153)+(-0.541377726375)*0.5**o
        ref[1,0]=(0.755207250839)/(o+1.)+(0.424003196273)+(0.296966289838)*0.5**o
        ref[1,1]=(-0.757702071557)/(o+1.)+(0.0525365805706)+(-0.507591010573)*0.5**o
        ref[1,2]=(0.19424332718)/(o+1.)+(-0.414604845348)+(0.821972467197)*0.5**o
        ref[1,3]=(-0.602942732318)/(o+1.)+(0.148510288994)+(-0.340805687409)*0.5**o
        ref[1,4]=(0.576509865762)/(o+1.)+(-0.0938334941445)+(0.0314374795024)*0.5**o
        ref[2,0]=(-0.793830015706)/(o+1.)+(-0.136932018907)+(0.250117528833)*0.5**o
        ref[2,1]=(0.838687005759)/(o+1.)+(-0.0175202382229)+(-0.347974394766)*0.5**o
        ref[2,2]=(0.215398162753)/(o+1.)+(0.375426307973)+(0.53063272248)*0.5**o
        ref[2,3]=(-0.739123251932)/(o+1.)+(0.0635875576574)+(-0.336752121527)*0.5**o
        ref[2,4]=(0.119988536199)/(o+1.)+(-0.0527994011886)+(-0.832783810519)*0.5**o
        ref[3,0]=(0.205050527245)/(o+1.)+(-0.0259746040335)+(0.4941356722)*0.5**o
        ref[3,1]=(-0.897011486902)/(o+1.)+(-0.678122704817)+(0.12538190695)*0.5**o
        ref[3,2]=(-0.970016379643)/(o+1.)+(-0.725555896953)+(-0.862829007552)*0.5**o
        ref[3,3]=(-0.929808474589)/(o+1.)+(0.166870249397)+(0.131451743264)*0.5**o
        ref[3,4]=(-0.714512816761)/(o+1.)+(-0.104027582947)+(-0.81935224348)*0.5**o
      else:
        arg[0,0]=(-0.916251178344)*x[0]**o+(0.858306033367)*x[0]+(0.814515071272)*x[1]**o+(-0.31988121358)*x[1]+(-0.973228319626)*x[2]**o+(-0.448894628432)*x[2]
        arg[0,1]=(0.394307707909)*x[0]**o+(-0.270473398988)*x[0]+(0.904831123674)*x[1]**o+(-0.949869961069)*x[1]+(0.771889549168)*x[2]**o+(0.650652953462)*x[2]
        arg[0,2]=(-0.0387282383528)*x[0]**o+(0.0292919821249)*x[0]+(-0.165189629855)*x[1]**o+(-0.265182468614)*x[1]+(0.919042632361)*x[2]**o+(0.710368688769)*x[2]
        arg[0,3]=(0.680006301334)*x[0]**o+(0.96959940913)*x[0]+(0.773153544603)*x[1]**o+(-0.460435754764)*x[1]+(-0.202448514604)*x[2]**o+(-0.495739672831)*x[2]
        arg[0,4]=(-0.972790915366)*x[0]**o+(-0.14148426268)*x[0]+(0.511995390348)*x[1]**o+(-0.252881217719)*x[1]+(0.255711210131)*x[2]**o+(0.729991325048)*x[2]
        arg[1,0]=(-0.105824689194)*x[0]**o+(-0.201427239595)*x[0]+(-0.00550922402684)*x[1]**o+(-0.732794556814)*x[1]+(-0.0820953451459)*x[2]**o+(0.289904794134)*x[2]
        arg[1,1]=(-0.378707957559)*x[0]**o+(0.776904510463)*x[0]+(0.0406721415351)*x[1]**o+(0.159288221237)*x[1]+(0.961004037621)*x[2]**o+(-0.0866180279992)*x[2]
        arg[1,2]=(-0.0226058491089)*x[0]**o+(-0.569995151391)*x[0]+(-0.0675265986408)*x[1]**o+(-0.141280346236)*x[1]+(0.116098815262)*x[2]**o+(0.654630437206)*x[2]
        arg[1,3]=(-0.567270755897)*x[0]**o+(0.580139672878)*x[0]+(-0.411816626246)*x[1]**o+(-0.490525831459)*x[1]+(0.9322522728)*x[2]**o+(-0.948230068413)*x[2]
        arg[1,4]=(0.977251285351)*x[0]**o+(-0.457921996364)*x[0]+(-0.586207130671)*x[1]**o+(0.801370818191)*x[1]+(-0.667233775445)*x[2]**o+(0.277054304674)*x[2]
        arg[2,0]=(0.997778241906)*x[0]**o+(0.490242237175)*x[0]+(0.406384527224)*x[1]**o+(-0.00724475574186)*x[1]+(-0.59799957401)*x[2]**o+(0.520598391022)*x[2]
        arg[2,1]=(0.59766081214)*x[0]**o+(-0.144793824209)*x[0]+(-0.191121009265)*x[1]**o+(-0.178016091024)*x[1]+(0.168431462883)*x[2]**o+(-0.766200678925)*x[2]
        arg[2,2]=(0.633988266946)*x[0]**o+(-0.880495194817)*x[0]+(-0.217502420499)*x[1]**o+(-0.730159316723)*x[1]+(0.255599562732)*x[2]**o+(-0.510511087905)*x[2]
        arg[2,3]=(0.102332382261)*x[0]**o+(-0.370884332534)*x[0]+(-0.0926186085648)*x[1]**o+(0.86446607832)*x[1]+(0.0643112986026)*x[2]**o+(0.950118360785)*x[2]
        arg[2,4]=(0.928323090624)*x[0]**o+(0.942943574803)*x[0]+(0.162570981489)*x[1]**o+(-0.971092675635)*x[1]+(0.641489870001)*x[2]**o+(0.695989498774)*x[2]
        arg[3,0]=(-0.988813741403)*x[0]**o+(0.407960892494)*x[0]+(0.671572481461)*x[1]**o+(0.119756938854)*x[1]+(-0.923671788537)*x[2]**o+(-0.458931381612)*x[2]
        arg[3,1]=(-0.241202616681)*x[0]**o+(-0.369134099039)*x[0]+(-0.129164268709)*x[1]**o+(-0.469311903775)*x[1]+(0.633909205241)*x[2]**o+(0.564043934538)*x[2]
        arg[3,2]=(-0.397081811407)*x[0]**o+(0.909387986899)*x[0]+(0.260590665694)*x[1]**o+(-0.371472818152)*x[1]+(-0.0369834262513)*x[2]**o+(-0.887598789153)*x[2]
        arg[3,3]=(-0.304262529036)*x[0]**o+(0.741316478862)*x[0]+(-0.384317206475)*x[1]**o+(-0.854685760579)*x[1]+(0.502373892581)*x[2]**o+(-0.923561844368)*x[2]
        arg[3,4]=(-0.720555420423)*x[0]**o+(-0.115786593478)*x[0]+(0.832043243085)*x[1]**o+(-0.912631896773)*x[1]+(-0.583397798668)*x[2]**o+(-0.508174721594)*x[2]
        ref[0,0]=(-0.158713248354)/(o+1.)+(0.0447650956777)+(-0.916251178344)*0.5**o
        ref[0,1]=(1.67672067284)/(o+1.)+(-0.284845203298)+(0.394307707909)*0.5**o
        ref[0,2]=(0.753853002506)/(o+1.)+(0.23723910114)+(-0.0387282383528)*0.5**o
        ref[0,3]=(0.570705029999)/(o+1.)+(0.00671199076774)+(0.680006301334)*0.5**o
        ref[0,4]=(0.767706600479)/(o+1.)+(0.167812922325)+(-0.972790915366)*0.5**o
        ref[1,0]=(-0.0876045691728)/(o+1.)+(-0.322158501138)+(-0.105824689194)*0.5**o
        ref[1,1]=(1.00167617916)/(o+1.)+(0.42478735185)+(-0.378707957559)*0.5**o
        ref[1,2]=(0.0485722166212)/(o+1.)+(-0.0283225302105)+(-0.0226058491089)*0.5**o
        ref[1,3]=(0.520435646554)/(o+1.)+(-0.429308113497)+(-0.567270755897)*0.5**o
        ref[1,4]=(-1.25344090612)/(o+1.)+(0.31025156325)+(0.977251285351)*0.5**o
        ref[2,0]=(-0.191615046786)/(o+1.)+(0.501797936227)+(0.997778241906)*0.5**o
        ref[2,1]=(-0.0226895463818)/(o+1.)+(-0.544505297079)+(0.59766081214)*0.5**o
        ref[2,2]=(0.0380971422336)/(o+1.)+(-1.06058279972)+(0.633988266946)*0.5**o
        ref[2,3]=(-0.0283073099622)/(o+1.)+(0.721850053286)+(0.102332382261)*0.5**o
        ref[2,4]=(0.80406085149)/(o+1.)+(0.333920198971)+(0.928323090624)*0.5**o
        ref[3,0]=(-0.252099307076)/(o+1.)+(0.0343932248679)+(-0.988813741403)*0.5**o
        ref[3,1]=(0.504744936532)/(o+1.)+(-0.137201034138)+(-0.241202616681)*0.5**o
        ref[3,2]=(0.223607239442)/(o+1.)+(-0.174841810203)+(-0.397081811407)*0.5**o
        ref[3,3]=(0.118056686106)/(o+1.)+(-0.518465563043)+(-0.304262529036)*0.5**o
        ref[3,4]=(0.248645444416)/(o+1.)+(-0.768296605922)+(-0.720555420423)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedFunctionOnContactOne(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.347273720148)*x[0]**o+(0.325185727721)*x[0]+(-0.855515194735)*x[1]**o+(0.825807975597)*x[1]
        arg[0,0,1]=(0.842054669437)*x[0]**o+(0.353768311241)*x[0]+(0.0148900864182)*x[1]**o+(-0.402154654143)*x[1]
        arg[0,1,0]=(-0.191525791617)*x[0]**o+(-0.299181220926)*x[0]+(-0.696260869529)*x[1]**o+(0.285645399154)*x[1]
        arg[0,1,1]=(0.190151633835)*x[0]**o+(0.184074338131)*x[0]+(-0.497689504982)*x[1]**o+(-0.813410371313)*x[1]
        arg[1,0,0]=(-0.695114302024)*x[0]**o+(0.427133485674)*x[0]+(-0.711899825683)*x[1]**o+(-0.296004734911)*x[1]
        arg[1,0,1]=(0.37620789583)*x[0]**o+(-0.944850444544)*x[0]+(0.666194812839)*x[1]**o+(-0.0715430628168)*x[1]
        arg[1,1,0]=(-0.0312291891427)*x[0]**o+(0.403009732443)*x[0]+(-0.378486005359)*x[1]**o+(-0.992729541609)*x[1]
        arg[1,1,1]=(0.0675984132698)*x[0]**o+(0.241001129867)*x[0]+(0.340546237588)*x[1]**o+(-0.024830810543)*x[1]
        arg[2,0,0]=(-0.146870758676)*x[0]**o+(-0.170434897147)*x[0]+(-0.579618469996)*x[1]**o+(-0.770817574749)*x[1]
        arg[2,0,1]=(-0.3029068201)*x[0]**o+(-0.110636302601)*x[0]+(-0.905841665338)*x[1]**o+(-0.280977886629)*x[1]
        arg[2,1,0]=(0.0547810600344)*x[0]**o+(0.00763321517272)*x[0]+(0.0156822194106)*x[1]**o+(-0.528337857094)*x[1]
        arg[2,1,1]=(0.12016647407)*x[0]**o+(0.4590743733)*x[0]+(-0.574499998403)*x[1]**o+(0.698871172503)*x[1]
        arg[3,0,0]=(0.481811757276)*x[0]**o+(0.748356476205)*x[0]+(-0.300272959749)*x[1]**o+(-0.707230263391)*x[1]
        arg[3,0,1]=(-0.347858576311)*x[0]**o+(-0.177810365974)*x[0]+(0.384658344075)*x[1]**o+(-0.837385668636)*x[1]
        arg[3,1,0]=(0.943559460135)*x[0]**o+(0.178778474835)*x[0]+(-0.831384000346)*x[1]**o+(-0.474975888568)*x[1]
        arg[3,1,1]=(-0.589903969401)*x[0]**o+(0.271114795948)*x[0]+(-0.946594221253)*x[1]**o+(-0.964029170264)*x[1]
        arg[4,0,0]=(0.221077049797)*x[0]**o+(0.986052914095)*x[0]+(-0.392982812685)*x[1]**o+(0.370828365061)*x[1]
        arg[4,0,1]=(0.463121232737)*x[0]**o+(-0.91920434872)*x[0]+(0.830655891827)*x[1]**o+(0.193876462558)*x[1]
        arg[4,1,0]=(0.925740559506)*x[0]**o+(0.534342942418)*x[0]+(0.446730463706)*x[1]**o+(0.181095371535)*x[1]
        arg[4,1,1]=(-0.777093674353)*x[0]**o+(-0.435130879793)*x[0]+(0.134406853212)*x[1]**o+(0.861099956528)*x[1]
        arg[5,0,0]=(-0.443943796359)*x[0]**o+(0.359190127208)*x[0]+(0.0570788998709)*x[1]**o+(0.548070594378)*x[1]
        arg[5,0,1]=(0.762962865085)*x[0]**o+(0.969248002849)*x[0]+(0.729222816543)*x[1]**o+(-0.865271983926)*x[1]
        arg[5,1,0]=(-0.915609563739)*x[0]**o+(-0.338560068696)*x[0]+(0.972145654544)*x[1]**o+(0.0941277349395)*x[1]
        arg[5,1,1]=(-0.486023356984)*x[0]**o+(-0.426382776921)*x[0]+(-0.114606039743)*x[1]**o+(-0.992152486495)*x[1]
        ref[0,0,0]=(-0.855515194735)/(o+1.)+(0.575496851659)+(-0.347273720148)*0.5**o
        ref[0,0,1]=(0.0148900864182)/(o+1.)+(-0.0241931714509)+(0.842054669437)*0.5**o
        ref[0,1,0]=(-0.696260869529)/(o+1.)+(-0.00676791088595)+(-0.191525791617)*0.5**o
        ref[0,1,1]=(-0.497689504982)/(o+1.)+(-0.314668016591)+(0.190151633835)*0.5**o
        ref[1,0,0]=(-0.711899825683)/(o+1.)+(0.0655643753813)+(-0.695114302024)*0.5**o
        ref[1,0,1]=(0.666194812839)/(o+1.)+(-0.50819675368)+(0.37620789583)*0.5**o
        ref[1,1,0]=(-0.378486005359)/(o+1.)+(-0.294859904583)+(-0.0312291891427)*0.5**o
        ref[1,1,1]=(0.340546237588)/(o+1.)+(0.108085159662)+(0.0675984132698)*0.5**o
        ref[2,0,0]=(-0.579618469996)/(o+1.)+(-0.470626235948)+(-0.146870758676)*0.5**o
        ref[2,0,1]=(-0.905841665338)/(o+1.)+(-0.195807094615)+(-0.3029068201)*0.5**o
        ref[2,1,0]=(0.0156822194106)/(o+1.)+(-0.260352320961)+(0.0547810600344)*0.5**o
        ref[2,1,1]=(-0.574499998403)/(o+1.)+(0.578972772902)+(0.12016647407)*0.5**o
        ref[3,0,0]=(-0.300272959749)/(o+1.)+(0.0205631064069)+(0.481811757276)*0.5**o
        ref[3,0,1]=(0.384658344075)/(o+1.)+(-0.507598017305)+(-0.347858576311)*0.5**o
        ref[3,1,0]=(-0.831384000346)/(o+1.)+(-0.148098706867)+(0.943559460135)*0.5**o
        ref[3,1,1]=(-0.946594221253)/(o+1.)+(-0.346457187158)+(-0.589903969401)*0.5**o
        ref[4,0,0]=(-0.392982812685)/(o+1.)+(0.678440639578)+(0.221077049797)*0.5**o
        ref[4,0,1]=(0.830655891827)/(o+1.)+(-0.362663943081)+(0.463121232737)*0.5**o
        ref[4,1,0]=(0.446730463706)/(o+1.)+(0.357719156977)+(0.925740559506)*0.5**o
        ref[4,1,1]=(0.134406853212)/(o+1.)+(0.212984538367)+(-0.777093674353)*0.5**o
        ref[5,0,0]=(0.0570788998709)/(o+1.)+(0.453630360793)+(-0.443943796359)*0.5**o
        ref[5,0,1]=(0.729222816543)/(o+1.)+(0.0519880094615)+(0.762962865085)*0.5**o
        ref[5,1,0]=(0.972145654544)/(o+1.)+(-0.122216166878)+(-0.915609563739)*0.5**o
        ref[5,1,1]=(-0.114606039743)/(o+1.)+(-0.709267631708)+(-0.486023356984)*0.5**o
      else:
        arg[0,0,0]=(0.455440122065)*x[0]**o+(0.911340571887)*x[0]+(0.158409397783)*x[1]**o+(0.353755257124)*x[1]+(-0.392556456322)*x[2]**o+(0.707232271095)*x[2]
        arg[0,0,1]=(-0.816405087957)*x[0]**o+(0.0691299991887)*x[0]+(0.178677657085)*x[1]**o+(0.526176369634)*x[1]+(-0.159459641444)*x[2]**o+(-0.197907778155)*x[2]
        arg[0,1,0]=(-0.463418217546)*x[0]**o+(-0.485835567862)*x[0]+(0.760458548482)*x[1]**o+(0.902873590314)*x[1]+(-0.760679635511)*x[2]**o+(-0.483163544794)*x[2]
        arg[0,1,1]=(0.743029549626)*x[0]**o+(-0.39131913217)*x[0]+(0.488809682734)*x[1]**o+(-0.256115251135)*x[1]+(0.0918611991036)*x[2]**o+(-0.281291060967)*x[2]
        arg[1,0,0]=(0.794367740289)*x[0]**o+(0.727663222288)*x[0]+(-0.631739003283)*x[1]**o+(0.733862609049)*x[1]+(-0.721230552103)*x[2]**o+(0.787774790111)*x[2]
        arg[1,0,1]=(0.505578181889)*x[0]**o+(0.616497234291)*x[0]+(-0.616903198005)*x[1]**o+(-0.355969913594)*x[1]+(0.0105854597103)*x[2]**o+(0.510273945857)*x[2]
        arg[1,1,0]=(0.35370601878)*x[0]**o+(-0.0414232536827)*x[0]+(0.0546619209334)*x[1]**o+(0.0459241421788)*x[1]+(0.796266758386)*x[2]**o+(0.303998757002)*x[2]
        arg[1,1,1]=(-0.245326163282)*x[0]**o+(0.0893045737315)*x[0]+(0.019401528413)*x[1]**o+(-0.838249399348)*x[1]+(0.510320088517)*x[2]**o+(0.544558918049)*x[2]
        arg[2,0,0]=(-0.660208039977)*x[0]**o+(-0.224874460802)*x[0]+(-0.0467836584153)*x[1]**o+(-0.463886703951)*x[1]+(-0.0951679104966)*x[2]**o+(-0.735787783779)*x[2]
        arg[2,0,1]=(-0.588259372806)*x[0]**o+(-0.18950606147)*x[0]+(-0.373439432467)*x[1]**o+(-0.116285718036)*x[1]+(-0.944687791765)*x[2]**o+(-0.243001707318)*x[2]
        arg[2,1,0]=(0.798498386971)*x[0]**o+(0.500125941024)*x[0]+(0.197339376576)*x[1]**o+(-0.068777545221)*x[1]+(0.36649695624)*x[2]**o+(0.193956440608)*x[2]
        arg[2,1,1]=(0.27501104189)*x[0]**o+(0.203066488761)*x[0]+(0.379278873515)*x[1]**o+(-0.789154057614)*x[1]+(0.893889499095)*x[2]**o+(0.867881975803)*x[2]
        arg[3,0,0]=(0.0579414247006)*x[0]**o+(0.661826954985)*x[0]+(0.313969775635)*x[1]**o+(-0.125420102972)*x[1]+(-0.808513922748)*x[2]**o+(0.0301073398588)*x[2]
        arg[3,0,1]=(-0.254544981247)*x[0]**o+(0.755983567644)*x[0]+(0.129524533092)*x[1]**o+(0.562828535682)*x[1]+(0.712881323083)*x[2]**o+(0.375450604678)*x[2]
        arg[3,1,0]=(-0.100221176995)*x[0]**o+(-0.470987646782)*x[0]+(-0.444245257427)*x[1]**o+(-0.0112626639353)*x[1]+(-0.516917833845)*x[2]**o+(-0.664158938756)*x[2]
        arg[3,1,1]=(0.259306174564)*x[0]**o+(0.367178744389)*x[0]+(0.177897569676)*x[1]**o+(-0.355724319562)*x[1]+(-0.809141944951)*x[2]**o+(0.962429694607)*x[2]
        arg[4,0,0]=(0.351990799011)*x[0]**o+(-0.41657457182)*x[0]+(-0.769748341564)*x[1]**o+(-0.856355494001)*x[1]+(-0.777902936204)*x[2]**o+(-0.388980567423)*x[2]
        arg[4,0,1]=(0.875307372684)*x[0]**o+(0.061375057221)*x[0]+(-0.262062525469)*x[1]**o+(0.0950324968281)*x[1]+(-0.145145751811)*x[2]**o+(-0.246872045818)*x[2]
        arg[4,1,0]=(0.115340652181)*x[0]**o+(0.910414483949)*x[0]+(0.347463744624)*x[1]**o+(0.399961393031)*x[1]+(0.0491863598064)*x[2]**o+(-0.992304094864)*x[2]
        arg[4,1,1]=(-0.125763267852)*x[0]**o+(0.485867057874)*x[0]+(-0.414172783931)*x[1]**o+(0.164145362806)*x[1]+(-0.488738184728)*x[2]**o+(-0.232758639076)*x[2]
        arg[5,0,0]=(-0.748184063441)*x[0]**o+(-0.252715593115)*x[0]+(0.688609235584)*x[1]**o+(0.0645665122312)*x[1]+(0.365465805991)*x[2]**o+(-0.217347848328)*x[2]
        arg[5,0,1]=(-0.298018437594)*x[0]**o+(-0.39683066279)*x[0]+(-0.0609074331294)*x[1]**o+(0.146930937933)*x[1]+(-0.00737018012329)*x[2]**o+(-0.700002160047)*x[2]
        arg[5,1,0]=(0.974781669859)*x[0]**o+(0.0634072121163)*x[0]+(0.0194044697894)*x[1]**o+(-0.919375545742)*x[1]+(-0.779806244021)*x[2]**o+(0.641761171945)*x[2]
        arg[5,1,1]=(0.485215976154)*x[0]**o+(0.879804212343)*x[0]+(-0.765170516406)*x[1]**o+(-0.205158612954)*x[1]+(-0.586708070331)*x[2]**o+(0.506670980992)*x[2]
        ref[0,0,0]=(-0.23414705854)/(o+1.)+(0.986164050053)+(0.455440122065)*0.5**o
        ref[0,0,1]=(0.0192180156412)/(o+1.)+(0.198699295334)+(-0.816405087957)*0.5**o
        ref[0,1,0]=(-0.000221087029094)/(o+1.)+(-0.0330627611712)+(-0.463418217546)*0.5**o
        ref[0,1,1]=(0.580670881837)/(o+1.)+(-0.464362722135)+(0.743029549626)*0.5**o
        ref[1,0,0]=(-1.35296955539)/(o+1.)+(1.12465031072)+(0.794367740289)*0.5**o
        ref[1,0,1]=(-0.606317738295)/(o+1.)+(0.385400633277)+(0.505578181889)*0.5**o
        ref[1,1,0]=(0.85092867932)/(o+1.)+(0.154249822749)+(0.35370601878)*0.5**o
        ref[1,1,1]=(0.52972161693)/(o+1.)+(-0.102192953784)+(-0.245326163282)*0.5**o
        ref[2,0,0]=(-0.141951568912)/(o+1.)+(-0.712274474266)+(-0.660208039977)*0.5**o
        ref[2,0,1]=(-1.31812722423)/(o+1.)+(-0.274396743412)+(-0.588259372806)*0.5**o
        ref[2,1,0]=(0.563836332815)/(o+1.)+(0.312652418205)+(0.798498386971)*0.5**o
        ref[2,1,1]=(1.27316837261)/(o+1.)+(0.140897203475)+(0.27501104189)*0.5**o
        ref[3,0,0]=(-0.494544147113)/(o+1.)+(0.283257095936)+(0.0579414247006)*0.5**o
        ref[3,0,1]=(0.842405856175)/(o+1.)+(0.847131354002)+(-0.254544981247)*0.5**o
        ref[3,1,0]=(-0.961163091272)/(o+1.)+(-0.573204624737)+(-0.100221176995)*0.5**o
        ref[3,1,1]=(-0.631244375275)/(o+1.)+(0.486942059717)+(0.259306174564)*0.5**o
        ref[4,0,0]=(-1.54765127777)/(o+1.)+(-0.830955316622)+(0.351990799011)*0.5**o
        ref[4,0,1]=(-0.407208277279)/(o+1.)+(-0.0452322458845)+(0.875307372684)*0.5**o
        ref[4,1,0]=(0.39665010443)/(o+1.)+(0.159035891058)+(0.115340652181)*0.5**o
        ref[4,1,1]=(-0.902910968659)/(o+1.)+(0.208626890802)+(-0.125763267852)*0.5**o
        ref[5,0,0]=(1.05407504158)/(o+1.)+(-0.202748464606)+(-0.748184063441)*0.5**o
        ref[5,0,1]=(-0.0682776132527)/(o+1.)+(-0.474950942452)+(-0.298018437594)*0.5**o
        ref[5,1,0]=(-0.760401774232)/(o+1.)+(-0.10710358084)+(0.974781669859)*0.5**o
        ref[5,1,1]=(-1.35187858674)/(o+1.)+(0.590658290191)+(0.485215976154)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnContactOne

      assumptions: ReducedFunctionOnContactOne(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.177078428867)*x[0]**o+(0.338116995819)*x[0]+(-0.750823220445)*x[1]**o+(0.0630074420245)*x[1]
        arg[0,0,0,1]=(-0.682838671264)*x[0]**o+(-0.893797611147)*x[0]+(-0.0127659464437)*x[1]**o+(0.144120749888)*x[1]
        arg[0,0,1,0]=(0.352439319413)*x[0]**o+(0.380064866457)*x[0]+(-0.932452125435)*x[1]**o+(0.914519937443)*x[1]
        arg[0,0,1,1]=(-0.354033408023)*x[0]**o+(0.771069338542)*x[0]+(-0.702758889809)*x[1]**o+(-0.484485925087)*x[1]
        arg[0,0,2,0]=(0.706826713479)*x[0]**o+(0.0526230471775)*x[0]+(-0.535244256862)*x[1]**o+(-0.085929844746)*x[1]
        arg[0,0,2,1]=(-0.893497856528)*x[0]**o+(-0.892309840177)*x[0]+(0.742770093639)*x[1]**o+(0.987917060326)*x[1]
        arg[0,1,0,0]=(-0.536624329097)*x[0]**o+(-0.569536051343)*x[0]+(-0.7740511254)*x[1]**o+(0.873150177653)*x[1]
        arg[0,1,0,1]=(0.388396767346)*x[0]**o+(-0.902569510334)*x[0]+(0.196975646933)*x[1]**o+(0.184178333918)*x[1]
        arg[0,1,1,0]=(-0.125771297178)*x[0]**o+(-0.335093336446)*x[0]+(0.372752949038)*x[1]**o+(-0.883928748878)*x[1]
        arg[0,1,1,1]=(-0.313637256508)*x[0]**o+(-0.49607955109)*x[0]+(-0.726599229867)*x[1]**o+(0.360837115079)*x[1]
        arg[0,1,2,0]=(0.92425817047)*x[0]**o+(0.734728600171)*x[0]+(1.60583902129e-05)*x[1]**o+(-0.124600971339)*x[1]
        arg[0,1,2,1]=(-0.0981827607557)*x[0]**o+(-0.556005942436)*x[0]+(-0.888060617134)*x[1]**o+(0.857508807636)*x[1]
        arg[0,2,0,0]=(-0.871590768072)*x[0]**o+(-0.208546938841)*x[0]+(-0.228336105988)*x[1]**o+(0.444098195066)*x[1]
        arg[0,2,0,1]=(-0.628675127006)*x[0]**o+(0.834043549769)*x[0]+(-0.509662470381)*x[1]**o+(-0.801907226651)*x[1]
        arg[0,2,1,0]=(-0.767528433339)*x[0]**o+(0.305967685996)*x[0]+(-0.587902586916)*x[1]**o+(0.38652781193)*x[1]
        arg[0,2,1,1]=(0.388984380682)*x[0]**o+(0.490673256209)*x[0]+(-0.78470263647)*x[1]**o+(0.34909383158)*x[1]
        arg[0,2,2,0]=(-0.910514531543)*x[0]**o+(-0.418810366626)*x[0]+(-0.637159792471)*x[1]**o+(0.794653160897)*x[1]
        arg[0,2,2,1]=(0.570535552656)*x[0]**o+(-0.590385806831)*x[0]+(-0.298806908011)*x[1]**o+(-0.901531327347)*x[1]
        arg[0,3,0,0]=(0.916864925136)*x[0]**o+(-0.019733476459)*x[0]+(-0.91370863065)*x[1]**o+(0.291932128733)*x[1]
        arg[0,3,0,1]=(-0.851147484632)*x[0]**o+(-0.159304969602)*x[0]+(0.699832871063)*x[1]**o+(0.327932729929)*x[1]
        arg[0,3,1,0]=(0.461273862468)*x[0]**o+(0.40068516653)*x[0]+(0.0203170502559)*x[1]**o+(0.973716167207)*x[1]
        arg[0,3,1,1]=(0.63984877633)*x[0]**o+(0.580871491133)*x[0]+(-0.00587359808953)*x[1]**o+(-0.104805621161)*x[1]
        arg[0,3,2,0]=(-0.159416471436)*x[0]**o+(-0.585039922994)*x[0]+(0.689553978928)*x[1]**o+(0.569617375406)*x[1]
        arg[0,3,2,1]=(0.782397836312)*x[0]**o+(-0.581894025693)*x[0]+(0.0370277809126)*x[1]**o+(0.597309139432)*x[1]
        arg[0,4,0,0]=(-0.0063433205485)*x[0]**o+(-0.242291627085)*x[0]+(0.110218625254)*x[1]**o+(0.819656578408)*x[1]
        arg[0,4,0,1]=(-0.128215646666)*x[0]**o+(0.821312510893)*x[0]+(0.910723121611)*x[1]**o+(-0.0174623980585)*x[1]
        arg[0,4,1,0]=(-0.604528735513)*x[0]**o+(-0.358705990504)*x[0]+(-0.281954207008)*x[1]**o+(0.123469639952)*x[1]
        arg[0,4,1,1]=(0.65096807032)*x[0]**o+(0.882935341204)*x[0]+(-0.761226249914)*x[1]**o+(-0.830800218214)*x[1]
        arg[0,4,2,0]=(-0.710718948634)*x[0]**o+(0.452128874613)*x[0]+(-0.780671499903)*x[1]**o+(-0.942033435912)*x[1]
        arg[0,4,2,1]=(-0.0443187506467)*x[0]**o+(0.513943853321)*x[0]+(0.709902333411)*x[1]**o+(0.755430555318)*x[1]
        arg[1,0,0,0]=(0.731890221753)*x[0]**o+(-0.821835735249)*x[0]+(-0.713098084853)*x[1]**o+(0.875738077525)*x[1]
        arg[1,0,0,1]=(-0.130335125365)*x[0]**o+(0.580579210083)*x[0]+(-0.427105331069)*x[1]**o+(0.0435798068353)*x[1]
        arg[1,0,1,0]=(-0.203908509107)*x[0]**o+(-0.941974700448)*x[0]+(0.344459670502)*x[1]**o+(0.758336135081)*x[1]
        arg[1,0,1,1]=(0.186770929724)*x[0]**o+(0.348332058705)*x[0]+(-0.0683630778679)*x[1]**o+(0.480425212189)*x[1]
        arg[1,0,2,0]=(-0.891275312975)*x[0]**o+(0.260630881729)*x[0]+(0.127955472707)*x[1]**o+(-0.554992279729)*x[1]
        arg[1,0,2,1]=(-0.74620344519)*x[0]**o+(0.944025669845)*x[0]+(-0.052031562679)*x[1]**o+(0.966060927262)*x[1]
        arg[1,1,0,0]=(0.871423198418)*x[0]**o+(-0.0769723128617)*x[0]+(0.783714758356)*x[1]**o+(0.530209967997)*x[1]
        arg[1,1,0,1]=(0.943197824902)*x[0]**o+(-0.62624855429)*x[0]+(-0.00888785913795)*x[1]**o+(-0.0408857497082)*x[1]
        arg[1,1,1,0]=(-0.74971981056)*x[0]**o+(0.925175163834)*x[0]+(0.481063204532)*x[1]**o+(-0.476700447185)*x[1]
        arg[1,1,1,1]=(-0.431776603043)*x[0]**o+(-0.495358165507)*x[0]+(-0.0110588803484)*x[1]**o+(-0.575147325496)*x[1]
        arg[1,1,2,0]=(0.578105297721)*x[0]**o+(-0.88169100464)*x[0]+(-0.646224518073)*x[1]**o+(0.465619837803)*x[1]
        arg[1,1,2,1]=(0.442066689582)*x[0]**o+(0.0937839719519)*x[0]+(-0.101534585934)*x[1]**o+(-0.974003442605)*x[1]
        arg[1,2,0,0]=(-0.206702379228)*x[0]**o+(0.330161049866)*x[0]+(0.185493587693)*x[1]**o+(0.0391986992855)*x[1]
        arg[1,2,0,1]=(0.79092697362)*x[0]**o+(0.692054695304)*x[0]+(0.141622971369)*x[1]**o+(0.643869474153)*x[1]
        arg[1,2,1,0]=(0.914281902407)*x[0]**o+(0.936365302055)*x[0]+(0.249095964198)*x[1]**o+(-0.440892551355)*x[1]
        arg[1,2,1,1]=(-0.602926709108)*x[0]**o+(0.817198874641)*x[0]+(0.262528290838)*x[1]**o+(0.493607450371)*x[1]
        arg[1,2,2,0]=(-0.724435531493)*x[0]**o+(-0.455127093687)*x[0]+(0.144972402754)*x[1]**o+(-0.361255031527)*x[1]
        arg[1,2,2,1]=(-0.611538590545)*x[0]**o+(-0.780730652602)*x[0]+(-0.150439352532)*x[1]**o+(0.450875229163)*x[1]
        arg[1,3,0,0]=(-0.913553947133)*x[0]**o+(-0.930172924067)*x[0]+(-0.509426399026)*x[1]**o+(0.3953360754)*x[1]
        arg[1,3,0,1]=(-0.709834132519)*x[0]**o+(0.972632526254)*x[0]+(0.540358729462)*x[1]**o+(-0.0295254269136)*x[1]
        arg[1,3,1,0]=(-0.30788139278)*x[0]**o+(-0.362561853154)*x[0]+(0.0135015706177)*x[1]**o+(0.558034983013)*x[1]
        arg[1,3,1,1]=(-0.243327801257)*x[0]**o+(-0.156242539226)*x[0]+(0.076082439418)*x[1]**o+(0.818613998695)*x[1]
        arg[1,3,2,0]=(-0.0290765899224)*x[0]**o+(-0.108604693002)*x[0]+(-0.0741034599215)*x[1]**o+(0.44467098076)*x[1]
        arg[1,3,2,1]=(-0.751374584848)*x[0]**o+(0.568700665267)*x[0]+(-0.431665912498)*x[1]**o+(0.881337230634)*x[1]
        arg[1,4,0,0]=(-0.940864464113)*x[0]**o+(-0.204392585977)*x[0]+(-0.872729326815)*x[1]**o+(0.454069061239)*x[1]
        arg[1,4,0,1]=(0.620544898911)*x[0]**o+(-0.194050689442)*x[0]+(0.399921161647)*x[1]**o+(0.812456444893)*x[1]
        arg[1,4,1,0]=(-0.398096053023)*x[0]**o+(-0.521802671926)*x[0]+(-0.525522030503)*x[1]**o+(-0.795276120062)*x[1]
        arg[1,4,1,1]=(0.277239330261)*x[0]**o+(-0.376159942175)*x[0]+(0.785902674573)*x[1]**o+(-0.00686958652716)*x[1]
        arg[1,4,2,0]=(-0.469242672532)*x[0]**o+(0.289432797089)*x[0]+(0.728305062026)*x[1]**o+(0.821761742668)*x[1]
        arg[1,4,2,1]=(0.828113301834)*x[0]**o+(-0.149301261166)*x[0]+(0.279230829643)*x[1]**o+(-0.971944918618)*x[1]
        arg[2,0,0,0]=(-0.724500208669)*x[0]**o+(-0.761179166402)*x[0]+(0.302409732402)*x[1]**o+(0.509101646177)*x[1]
        arg[2,0,0,1]=(-0.133300605917)*x[0]**o+(-0.0299057843108)*x[0]+(0.190028790905)*x[1]**o+(-0.668097775221)*x[1]
        arg[2,0,1,0]=(0.828019154028)*x[0]**o+(-0.446721913762)*x[0]+(0.745204383075)*x[1]**o+(0.363968853359)*x[1]
        arg[2,0,1,1]=(-0.653962912288)*x[0]**o+(0.753529804971)*x[0]+(-0.104712073949)*x[1]**o+(-0.44152589042)*x[1]
        arg[2,0,2,0]=(0.21572653818)*x[0]**o+(-0.553685840495)*x[0]+(0.372296357092)*x[1]**o+(-0.23566755183)*x[1]
        arg[2,0,2,1]=(-0.0943992743407)*x[0]**o+(0.677836956763)*x[0]+(-0.33055202826)*x[1]**o+(0.444414877097)*x[1]
        arg[2,1,0,0]=(-0.812500024009)*x[0]**o+(0.4856670967)*x[0]+(-0.891481528328)*x[1]**o+(-0.244043957515)*x[1]
        arg[2,1,0,1]=(-0.958152880518)*x[0]**o+(0.494470945352)*x[0]+(-0.000696143847979)*x[1]**o+(-0.036680781152)*x[1]
        arg[2,1,1,0]=(-0.81109120479)*x[0]**o+(-0.929246624754)*x[0]+(0.128356074186)*x[1]**o+(0.0107006959497)*x[1]
        arg[2,1,1,1]=(-0.72470858095)*x[0]**o+(-0.202254265463)*x[0]+(-0.995176229907)*x[1]**o+(-0.752592973024)*x[1]
        arg[2,1,2,0]=(-0.191810147267)*x[0]**o+(-0.819397723879)*x[0]+(0.306860971474)*x[1]**o+(0.681274705052)*x[1]
        arg[2,1,2,1]=(-0.867447323913)*x[0]**o+(-0.279148200325)*x[0]+(-0.195907590382)*x[1]**o+(-0.305826952304)*x[1]
        arg[2,2,0,0]=(0.983782298896)*x[0]**o+(-0.284064487432)*x[0]+(-0.0928333646602)*x[1]**o+(-0.929708520581)*x[1]
        arg[2,2,0,1]=(0.121619221979)*x[0]**o+(-0.876309915691)*x[0]+(0.455030678927)*x[1]**o+(0.635763125356)*x[1]
        arg[2,2,1,0]=(0.378726440151)*x[0]**o+(-0.626329947078)*x[0]+(-0.0239503446136)*x[1]**o+(0.32696940295)*x[1]
        arg[2,2,1,1]=(0.442586981104)*x[0]**o+(-0.427714534417)*x[0]+(0.906235883406)*x[1]**o+(0.819595435756)*x[1]
        arg[2,2,2,0]=(-0.564560890159)*x[0]**o+(-0.115075498726)*x[0]+(-0.520596320159)*x[1]**o+(-0.368366112222)*x[1]
        arg[2,2,2,1]=(0.215858165597)*x[0]**o+(-0.575665537635)*x[0]+(0.792289256264)*x[1]**o+(0.215575294284)*x[1]
        arg[2,3,0,0]=(-0.596335978041)*x[0]**o+(-0.677983108509)*x[0]+(-0.551360906631)*x[1]**o+(-0.874453959198)*x[1]
        arg[2,3,0,1]=(-0.913806992124)*x[0]**o+(0.192343143898)*x[0]+(0.821070238869)*x[1]**o+(-0.188359117641)*x[1]
        arg[2,3,1,0]=(-0.0831972104302)*x[0]**o+(-0.927005811846)*x[0]+(-0.43240265235)*x[1]**o+(-0.144332400422)*x[1]
        arg[2,3,1,1]=(0.643878326697)*x[0]**o+(0.212328499471)*x[0]+(0.72072318435)*x[1]**o+(-0.146144387035)*x[1]
        arg[2,3,2,0]=(0.843652876743)*x[0]**o+(-0.0386075132797)*x[0]+(0.754511827443)*x[1]**o+(-0.26945353599)*x[1]
        arg[2,3,2,1]=(0.0413857115672)*x[0]**o+(0.682727309008)*x[0]+(0.261249117327)*x[1]**o+(-0.753970539308)*x[1]
        arg[2,4,0,0]=(-0.607477693828)*x[0]**o+(-0.5870344058)*x[0]+(-0.0678823396451)*x[1]**o+(0.201869223616)*x[1]
        arg[2,4,0,1]=(-0.618172945146)*x[0]**o+(0.685084114438)*x[0]+(0.0457149578306)*x[1]**o+(-0.877971038996)*x[1]
        arg[2,4,1,0]=(0.722668769621)*x[0]**o+(0.302142466253)*x[0]+(0.630735604709)*x[1]**o+(-0.427754766567)*x[1]
        arg[2,4,1,1]=(-0.847113860795)*x[0]**o+(-0.779747246887)*x[0]+(-0.693659242703)*x[1]**o+(0.897451884568)*x[1]
        arg[2,4,2,0]=(-0.437578500878)*x[0]**o+(0.399407036331)*x[0]+(0.705207362097)*x[1]**o+(-0.271060427211)*x[1]
        arg[2,4,2,1]=(0.329570005179)*x[0]**o+(-0.712089906899)*x[0]+(0.860707329294)*x[1]**o+(0.553218273352)*x[1]
        arg[3,0,0,0]=(0.932602317046)*x[0]**o+(-0.948445141026)*x[0]+(0.549409370786)*x[1]**o+(0.782964260926)*x[1]
        arg[3,0,0,1]=(0.975116319103)*x[0]**o+(-0.0131313829738)*x[0]+(0.222376331882)*x[1]**o+(0.545375534163)*x[1]
        arg[3,0,1,0]=(0.019317277942)*x[0]**o+(0.918003018679)*x[0]+(-0.481194397466)*x[1]**o+(0.762780571026)*x[1]
        arg[3,0,1,1]=(-0.393162543578)*x[0]**o+(0.984304464298)*x[0]+(0.737901448601)*x[1]**o+(-0.890589811376)*x[1]
        arg[3,0,2,0]=(0.458096711279)*x[0]**o+(0.0343270226524)*x[0]+(0.278242150139)*x[1]**o+(-0.158378755134)*x[1]
        arg[3,0,2,1]=(0.307111052767)*x[0]**o+(-0.294724312785)*x[0]+(-0.489134379327)*x[1]**o+(-0.211139185303)*x[1]
        arg[3,1,0,0]=(0.420707084879)*x[0]**o+(-0.998173294986)*x[0]+(0.508808683)*x[1]**o+(-0.189902487212)*x[1]
        arg[3,1,0,1]=(-0.745253594401)*x[0]**o+(-0.344809947215)*x[0]+(0.198212694401)*x[1]**o+(-0.125428245606)*x[1]
        arg[3,1,1,0]=(0.535215252931)*x[0]**o+(0.565625295858)*x[0]+(-0.0741524230667)*x[1]**o+(-0.898600294889)*x[1]
        arg[3,1,1,1]=(0.193417593556)*x[0]**o+(-0.80053921694)*x[0]+(-0.664372557476)*x[1]**o+(0.47786060257)*x[1]
        arg[3,1,2,0]=(0.553690237434)*x[0]**o+(0.970763166888)*x[0]+(0.408815221365)*x[1]**o+(0.143175119269)*x[1]
        arg[3,1,2,1]=(-0.211683249218)*x[0]**o+(0.0643055656631)*x[0]+(-0.111028404397)*x[1]**o+(0.506093872018)*x[1]
        arg[3,2,0,0]=(-0.866119051438)*x[0]**o+(-0.417310807983)*x[0]+(-0.228385234131)*x[1]**o+(0.554977378653)*x[1]
        arg[3,2,0,1]=(-0.146618266885)*x[0]**o+(0.817733306502)*x[0]+(-0.709059272627)*x[1]**o+(-0.271140815458)*x[1]
        arg[3,2,1,0]=(0.622973500199)*x[0]**o+(-0.568858605993)*x[0]+(-0.651609755171)*x[1]**o+(-0.866184190672)*x[1]
        arg[3,2,1,1]=(-0.538565077894)*x[0]**o+(-0.676237465393)*x[0]+(0.106659392436)*x[1]**o+(0.488735158765)*x[1]
        arg[3,2,2,0]=(-0.325897413686)*x[0]**o+(0.99502293321)*x[0]+(-0.339096900793)*x[1]**o+(0.238083096263)*x[1]
        arg[3,2,2,1]=(0.861430484293)*x[0]**o+(-0.761739126386)*x[0]+(-0.10651751912)*x[1]**o+(0.688993400068)*x[1]
        arg[3,3,0,0]=(0.0890838652267)*x[0]**o+(0.151733228604)*x[0]+(0.230063809846)*x[1]**o+(-0.135041740614)*x[1]
        arg[3,3,0,1]=(-0.969251937403)*x[0]**o+(0.345709865473)*x[0]+(-0.309971427898)*x[1]**o+(-0.262217856737)*x[1]
        arg[3,3,1,0]=(0.640075232231)*x[0]**o+(-0.00817560718155)*x[0]+(0.55773788655)*x[1]**o+(0.642205213407)*x[1]
        arg[3,3,1,1]=(-0.350002630494)*x[0]**o+(0.373270925939)*x[0]+(-0.655757122598)*x[1]**o+(-0.74598967873)*x[1]
        arg[3,3,2,0]=(-0.262399316954)*x[0]**o+(-0.37991094527)*x[0]+(0.310310045807)*x[1]**o+(0.838942372822)*x[1]
        arg[3,3,2,1]=(-0.977039760257)*x[0]**o+(-0.0804751651971)*x[0]+(-0.844329846245)*x[1]**o+(0.905827257908)*x[1]
        arg[3,4,0,0]=(0.749594099085)*x[0]**o+(0.382880202285)*x[0]+(0.39699250956)*x[1]**o+(-0.708672318437)*x[1]
        arg[3,4,0,1]=(0.913703789336)*x[0]**o+(-0.497792024208)*x[0]+(-0.34125011876)*x[1]**o+(-0.657159430634)*x[1]
        arg[3,4,1,0]=(-0.397439923687)*x[0]**o+(0.595274583577)*x[0]+(-0.292338728991)*x[1]**o+(-0.314101057271)*x[1]
        arg[3,4,1,1]=(-0.403584993513)*x[0]**o+(-0.758728905384)*x[0]+(-0.182501473325)*x[1]**o+(-0.0942108622967)*x[1]
        arg[3,4,2,0]=(-0.263608698751)*x[0]**o+(-0.591451983536)*x[0]+(0.620412677389)*x[1]**o+(0.600005456466)*x[1]
        arg[3,4,2,1]=(-0.633502278259)*x[0]**o+(0.733127573934)*x[0]+(-0.765468966987)*x[1]**o+(-0.24211874772)*x[1]
        ref[0,0,0,0]=(-0.750823220445)/(o+1.)+(0.200562218922)+(-0.177078428867)*0.5**o
        ref[0,0,0,1]=(-0.0127659464437)/(o+1.)+(-0.374838430629)+(-0.682838671264)*0.5**o
        ref[0,0,1,0]=(-0.932452125435)/(o+1.)+(0.64729240195)+(0.352439319413)*0.5**o
        ref[0,0,1,1]=(-0.702758889809)/(o+1.)+(0.143291706728)+(-0.354033408023)*0.5**o
        ref[0,0,2,0]=(-0.535244256862)/(o+1.)+(-0.0166533987843)+(0.706826713479)*0.5**o
        ref[0,0,2,1]=(0.742770093639)/(o+1.)+(0.0478036100749)+(-0.893497856528)*0.5**o
        ref[0,1,0,0]=(-0.7740511254)/(o+1.)+(0.151807063155)+(-0.536624329097)*0.5**o
        ref[0,1,0,1]=(0.196975646933)/(o+1.)+(-0.359195588208)+(0.388396767346)*0.5**o
        ref[0,1,1,0]=(0.372752949038)/(o+1.)+(-0.609511042662)+(-0.125771297178)*0.5**o
        ref[0,1,1,1]=(-0.726599229867)/(o+1.)+(-0.0676212180057)+(-0.313637256508)*0.5**o
        ref[0,1,2,0]=(1.60583902129e-05)/(o+1.)+(0.305063814416)+(0.92425817047)*0.5**o
        ref[0,1,2,1]=(-0.888060617134)/(o+1.)+(0.1507514326)+(-0.0981827607557)*0.5**o
        ref[0,2,0,0]=(-0.228336105988)/(o+1.)+(0.117775628113)+(-0.871590768072)*0.5**o
        ref[0,2,0,1]=(-0.509662470381)/(o+1.)+(0.0160681615592)+(-0.628675127006)*0.5**o
        ref[0,2,1,0]=(-0.587902586916)/(o+1.)+(0.346247748963)+(-0.767528433339)*0.5**o
        ref[0,2,1,1]=(-0.78470263647)/(o+1.)+(0.419883543895)+(0.388984380682)*0.5**o
        ref[0,2,2,0]=(-0.637159792471)/(o+1.)+(0.187921397135)+(-0.910514531543)*0.5**o
        ref[0,2,2,1]=(-0.298806908011)/(o+1.)+(-0.745958567089)+(0.570535552656)*0.5**o
        ref[0,3,0,0]=(-0.91370863065)/(o+1.)+(0.136099326137)+(0.916864925136)*0.5**o
        ref[0,3,0,1]=(0.699832871063)/(o+1.)+(0.0843138801635)+(-0.851147484632)*0.5**o
        ref[0,3,1,0]=(0.0203170502559)/(o+1.)+(0.687200666868)+(0.461273862468)*0.5**o
        ref[0,3,1,1]=(-0.00587359808953)/(o+1.)+(0.238032934986)+(0.63984877633)*0.5**o
        ref[0,3,2,0]=(0.689553978928)/(o+1.)+(-0.00771127379434)+(-0.159416471436)*0.5**o
        ref[0,3,2,1]=(0.0370277809126)/(o+1.)+(0.00770755686953)+(0.782397836312)*0.5**o
        ref[0,4,0,0]=(0.110218625254)/(o+1.)+(0.288682475662)+(-0.0063433205485)*0.5**o
        ref[0,4,0,1]=(0.910723121611)/(o+1.)+(0.401925056417)+(-0.128215646666)*0.5**o
        ref[0,4,1,0]=(-0.281954207008)/(o+1.)+(-0.117618175276)+(-0.604528735513)*0.5**o
        ref[0,4,1,1]=(-0.761226249914)/(o+1.)+(0.0260675614951)+(0.65096807032)*0.5**o
        ref[0,4,2,0]=(-0.780671499903)/(o+1.)+(-0.244952280649)+(-0.710718948634)*0.5**o
        ref[0,4,2,1]=(0.709902333411)/(o+1.)+(0.63468720432)+(-0.0443187506467)*0.5**o
        ref[1,0,0,0]=(-0.713098084853)/(o+1.)+(0.026951171138)+(0.731890221753)*0.5**o
        ref[1,0,0,1]=(-0.427105331069)/(o+1.)+(0.312079508459)+(-0.130335125365)*0.5**o
        ref[1,0,1,0]=(0.344459670502)/(o+1.)+(-0.0918192826838)+(-0.203908509107)*0.5**o
        ref[1,0,1,1]=(-0.0683630778679)/(o+1.)+(0.414378635447)+(0.186770929724)*0.5**o
        ref[1,0,2,0]=(0.127955472707)/(o+1.)+(-0.147180699)+(-0.891275312975)*0.5**o
        ref[1,0,2,1]=(-0.052031562679)/(o+1.)+(0.955043298554)+(-0.74620344519)*0.5**o
        ref[1,1,0,0]=(0.783714758356)/(o+1.)+(0.226618827568)+(0.871423198418)*0.5**o
        ref[1,1,0,1]=(-0.00888785913795)/(o+1.)+(-0.333567151999)+(0.943197824902)*0.5**o
        ref[1,1,1,0]=(0.481063204532)/(o+1.)+(0.224237358325)+(-0.74971981056)*0.5**o
        ref[1,1,1,1]=(-0.0110588803484)/(o+1.)+(-0.535252745501)+(-0.431776603043)*0.5**o
        ref[1,1,2,0]=(-0.646224518073)/(o+1.)+(-0.208035583419)+(0.578105297721)*0.5**o
        ref[1,1,2,1]=(-0.101534585934)/(o+1.)+(-0.440109735326)+(0.442066689582)*0.5**o
        ref[1,2,0,0]=(0.185493587693)/(o+1.)+(0.184679874576)+(-0.206702379228)*0.5**o
        ref[1,2,0,1]=(0.141622971369)/(o+1.)+(0.667962084728)+(0.79092697362)*0.5**o
        ref[1,2,1,0]=(0.249095964198)/(o+1.)+(0.24773637535)+(0.914281902407)*0.5**o
        ref[1,2,1,1]=(0.262528290838)/(o+1.)+(0.655403162506)+(-0.602926709108)*0.5**o
        ref[1,2,2,0]=(0.144972402754)/(o+1.)+(-0.408191062607)+(-0.724435531493)*0.5**o
        ref[1,2,2,1]=(-0.150439352532)/(o+1.)+(-0.16492771172)+(-0.611538590545)*0.5**o
        ref[1,3,0,0]=(-0.509426399026)/(o+1.)+(-0.267418424334)+(-0.913553947133)*0.5**o
        ref[1,3,0,1]=(0.540358729462)/(o+1.)+(0.47155354967)+(-0.709834132519)*0.5**o
        ref[1,3,1,0]=(0.0135015706177)/(o+1.)+(0.0977365649293)+(-0.30788139278)*0.5**o
        ref[1,3,1,1]=(0.076082439418)/(o+1.)+(0.331185729735)+(-0.243327801257)*0.5**o
        ref[1,3,2,0]=(-0.0741034599215)/(o+1.)+(0.168033143879)+(-0.0290765899224)*0.5**o
        ref[1,3,2,1]=(-0.431665912498)/(o+1.)+(0.72501894795)+(-0.751374584848)*0.5**o
        ref[1,4,0,0]=(-0.872729326815)/(o+1.)+(0.124838237631)+(-0.940864464113)*0.5**o
        ref[1,4,0,1]=(0.399921161647)/(o+1.)+(0.309202877726)+(0.620544898911)*0.5**o
        ref[1,4,1,0]=(-0.525522030503)/(o+1.)+(-0.658539395994)+(-0.398096053023)*0.5**o
        ref[1,4,1,1]=(0.785902674573)/(o+1.)+(-0.191514764351)+(0.277239330261)*0.5**o
        ref[1,4,2,0]=(0.728305062026)/(o+1.)+(0.555597269878)+(-0.469242672532)*0.5**o
        ref[1,4,2,1]=(0.279230829643)/(o+1.)+(-0.560623089892)+(0.828113301834)*0.5**o
        ref[2,0,0,0]=(0.302409732402)/(o+1.)+(-0.126038760113)+(-0.724500208669)*0.5**o
        ref[2,0,0,1]=(0.190028790905)/(o+1.)+(-0.349001779766)+(-0.133300605917)*0.5**o
        ref[2,0,1,0]=(0.745204383075)/(o+1.)+(-0.0413765302013)+(0.828019154028)*0.5**o
        ref[2,0,1,1]=(-0.104712073949)/(o+1.)+(0.156001957276)+(-0.653962912288)*0.5**o
        ref[2,0,2,0]=(0.372296357092)/(o+1.)+(-0.394676696163)+(0.21572653818)*0.5**o
        ref[2,0,2,1]=(-0.33055202826)/(o+1.)+(0.56112591693)+(-0.0943992743407)*0.5**o
        ref[2,1,0,0]=(-0.891481528328)/(o+1.)+(0.120811569592)+(-0.812500024009)*0.5**o
        ref[2,1,0,1]=(-0.000696143847979)/(o+1.)+(0.2288950821)+(-0.958152880518)*0.5**o
        ref[2,1,1,0]=(0.128356074186)/(o+1.)+(-0.459272964402)+(-0.81109120479)*0.5**o
        ref[2,1,1,1]=(-0.995176229907)/(o+1.)+(-0.477423619244)+(-0.72470858095)*0.5**o
        ref[2,1,2,0]=(0.306860971474)/(o+1.)+(-0.0690615094134)+(-0.191810147267)*0.5**o
        ref[2,1,2,1]=(-0.195907590382)/(o+1.)+(-0.292487576314)+(-0.867447323913)*0.5**o
        ref[2,2,0,0]=(-0.0928333646602)/(o+1.)+(-0.606886504006)+(0.983782298896)*0.5**o
        ref[2,2,0,1]=(0.455030678927)/(o+1.)+(-0.120273395167)+(0.121619221979)*0.5**o
        ref[2,2,1,0]=(-0.0239503446136)/(o+1.)+(-0.149680272064)+(0.378726440151)*0.5**o
        ref[2,2,1,1]=(0.906235883406)/(o+1.)+(0.19594045067)+(0.442586981104)*0.5**o
        ref[2,2,2,0]=(-0.520596320159)/(o+1.)+(-0.241720805474)+(-0.564560890159)*0.5**o
        ref[2,2,2,1]=(0.792289256264)/(o+1.)+(-0.180045121676)+(0.215858165597)*0.5**o
        ref[2,3,0,0]=(-0.551360906631)/(o+1.)+(-0.776218533854)+(-0.596335978041)*0.5**o
        ref[2,3,0,1]=(0.821070238869)/(o+1.)+(0.00199201312847)+(-0.913806992124)*0.5**o
        ref[2,3,1,0]=(-0.43240265235)/(o+1.)+(-0.535669106134)+(-0.0831972104302)*0.5**o
        ref[2,3,1,1]=(0.72072318435)/(o+1.)+(0.0330920562179)+(0.643878326697)*0.5**o
        ref[2,3,2,0]=(0.754511827443)/(o+1.)+(-0.154030524635)+(0.843652876743)*0.5**o
        ref[2,3,2,1]=(0.261249117327)/(o+1.)+(-0.03562161515)+(0.0413857115672)*0.5**o
        ref[2,4,0,0]=(-0.0678823396451)/(o+1.)+(-0.192582591092)+(-0.607477693828)*0.5**o
        ref[2,4,0,1]=(0.0457149578306)/(o+1.)+(-0.0964434622791)+(-0.618172945146)*0.5**o
        ref[2,4,1,0]=(0.630735604709)/(o+1.)+(-0.0628061501569)+(0.722668769621)*0.5**o
        ref[2,4,1,1]=(-0.693659242703)/(o+1.)+(0.0588523188405)+(-0.847113860795)*0.5**o
        ref[2,4,2,0]=(0.705207362097)/(o+1.)+(0.0641733045603)+(-0.437578500878)*0.5**o
        ref[2,4,2,1]=(0.860707329294)/(o+1.)+(-0.0794358167735)+(0.329570005179)*0.5**o
        ref[3,0,0,0]=(0.549409370786)/(o+1.)+(-0.0827404400499)+(0.932602317046)*0.5**o
        ref[3,0,0,1]=(0.222376331882)/(o+1.)+(0.266122075595)+(0.975116319103)*0.5**o
        ref[3,0,1,0]=(-0.481194397466)/(o+1.)+(0.840391794852)+(0.019317277942)*0.5**o
        ref[3,0,1,1]=(0.737901448601)/(o+1.)+(0.0468573264614)+(-0.393162543578)*0.5**o
        ref[3,0,2,0]=(0.278242150139)/(o+1.)+(-0.0620258662406)+(0.458096711279)*0.5**o
        ref[3,0,2,1]=(-0.489134379327)/(o+1.)+(-0.252931749044)+(0.307111052767)*0.5**o
        ref[3,1,0,0]=(0.508808683)/(o+1.)+(-0.594037891099)+(0.420707084879)*0.5**o
        ref[3,1,0,1]=(0.198212694401)/(o+1.)+(-0.23511909641)+(-0.745253594401)*0.5**o
        ref[3,1,1,0]=(-0.0741524230667)/(o+1.)+(-0.166487499516)+(0.535215252931)*0.5**o
        ref[3,1,1,1]=(-0.664372557476)/(o+1.)+(-0.161339307185)+(0.193417593556)*0.5**o
        ref[3,1,2,0]=(0.408815221365)/(o+1.)+(0.556969143078)+(0.553690237434)*0.5**o
        ref[3,1,2,1]=(-0.111028404397)/(o+1.)+(0.28519971884)+(-0.211683249218)*0.5**o
        ref[3,2,0,0]=(-0.228385234131)/(o+1.)+(0.068833285335)+(-0.866119051438)*0.5**o
        ref[3,2,0,1]=(-0.709059272627)/(o+1.)+(0.273296245522)+(-0.146618266885)*0.5**o
        ref[3,2,1,0]=(-0.651609755171)/(o+1.)+(-0.717521398332)+(0.622973500199)*0.5**o
        ref[3,2,1,1]=(0.106659392436)/(o+1.)+(-0.0937511533137)+(-0.538565077894)*0.5**o
        ref[3,2,2,0]=(-0.339096900793)/(o+1.)+(0.616553014737)+(-0.325897413686)*0.5**o
        ref[3,2,2,1]=(-0.10651751912)/(o+1.)+(-0.0363728631586)+(0.861430484293)*0.5**o
        ref[3,3,0,0]=(0.230063809846)/(o+1.)+(0.0083457439947)+(0.0890838652267)*0.5**o
        ref[3,3,0,1]=(-0.309971427898)/(o+1.)+(0.0417460043682)+(-0.969251937403)*0.5**o
        ref[3,3,1,0]=(0.55773788655)/(o+1.)+(0.317014803113)+(0.640075232231)*0.5**o
        ref[3,3,1,1]=(-0.655757122598)/(o+1.)+(-0.186359376395)+(-0.350002630494)*0.5**o
        ref[3,3,2,0]=(0.310310045807)/(o+1.)+(0.229515713776)+(-0.262399316954)*0.5**o
        ref[3,3,2,1]=(-0.844329846245)/(o+1.)+(0.412676046356)+(-0.977039760257)*0.5**o
        ref[3,4,0,0]=(0.39699250956)/(o+1.)+(-0.162896058076)+(0.749594099085)*0.5**o
        ref[3,4,0,1]=(-0.34125011876)/(o+1.)+(-0.577475727421)+(0.913703789336)*0.5**o
        ref[3,4,1,0]=(-0.292338728991)/(o+1.)+(0.140586763153)+(-0.397439923687)*0.5**o
        ref[3,4,1,1]=(-0.182501473325)/(o+1.)+(-0.42646988384)+(-0.403584993513)*0.5**o
        ref[3,4,2,0]=(0.620412677389)/(o+1.)+(0.00427673646471)+(-0.263608698751)*0.5**o
        ref[3,4,2,1]=(-0.765468966987)/(o+1.)+(0.245504413107)+(-0.633502278259)*0.5**o
      else:
        arg[0,0,0,0]=(0.480896043196)*x[0]**o+(0.113355160927)*x[0]+(0.498630339771)*x[1]**o+(0.0149514785179)*x[1]+(-0.944439525708)*x[2]**o+(-0.349299625744)*x[2]
        arg[0,0,0,1]=(-0.609632888023)*x[0]**o+(-0.576508904483)*x[0]+(-0.0610487008887)*x[1]**o+(-0.844087384492)*x[1]+(-0.715427733517)*x[2]**o+(-0.263424969452)*x[2]
        arg[0,0,1,0]=(0.205065167181)*x[0]**o+(-0.840225514826)*x[0]+(0.808161497555)*x[1]**o+(0.685902712814)*x[1]+(-0.625183952417)*x[2]**o+(-0.0947932775112)*x[2]
        arg[0,0,1,1]=(0.875716461075)*x[0]**o+(0.814425331995)*x[0]+(-0.805331622235)*x[1]**o+(-0.287737068993)*x[1]+(-0.365715978159)*x[2]**o+(0.689907502778)*x[2]
        arg[0,0,2,0]=(0.711372030289)*x[0]**o+(-0.393612062405)*x[0]+(0.642793863702)*x[1]**o+(-0.66636178423)*x[1]+(0.00488425355017)*x[2]**o+(-0.444523817626)*x[2]
        arg[0,0,2,1]=(-0.0536842545617)*x[0]**o+(-0.714279418499)*x[0]+(-0.487452457649)*x[1]**o+(0.917805820162)*x[1]+(0.445007135027)*x[2]**o+(0.101201310289)*x[2]
        arg[0,1,0,0]=(-0.710091748251)*x[0]**o+(0.470243562863)*x[0]+(-0.0178862343634)*x[1]**o+(0.393644042332)*x[1]+(-0.952839337577)*x[2]**o+(-0.152392013939)*x[2]
        arg[0,1,0,1]=(0.250989366649)*x[0]**o+(0.592158236555)*x[0]+(0.698185234877)*x[1]**o+(0.112154263635)*x[1]+(0.843915127813)*x[2]**o+(-0.825331842242)*x[2]
        arg[0,1,1,0]=(0.690360096543)*x[0]**o+(0.78906430602)*x[0]+(0.596189773531)*x[1]**o+(-0.293452572358)*x[1]+(0.737565980909)*x[2]**o+(-0.96988090057)*x[2]
        arg[0,1,1,1]=(-0.815220933301)*x[0]**o+(-0.875224889387)*x[0]+(-0.0562950344908)*x[1]**o+(0.56567431969)*x[1]+(0.133489969036)*x[2]**o+(0.299403989148)*x[2]
        arg[0,1,2,0]=(0.169442617085)*x[0]**o+(-0.413300800662)*x[0]+(0.580773652363)*x[1]**o+(-0.627706012397)*x[1]+(-0.730218498677)*x[2]**o+(0.587233195345)*x[2]
        arg[0,1,2,1]=(-0.110636656112)*x[0]**o+(-0.302213944624)*x[0]+(0.643229732019)*x[1]**o+(-0.651805494828)*x[1]+(-0.528276901455)*x[2]**o+(-0.272057635286)*x[2]
        arg[0,2,0,0]=(0.340661037085)*x[0]**o+(-0.672207039496)*x[0]+(0.676291394824)*x[1]**o+(0.596998737167)*x[1]+(-0.689800391841)*x[2]**o+(-0.134273731064)*x[2]
        arg[0,2,0,1]=(0.185282186176)*x[0]**o+(0.252489447393)*x[0]+(-0.0447305595755)*x[1]**o+(-0.558361084175)*x[1]+(0.922681220429)*x[2]**o+(-0.484588607971)*x[2]
        arg[0,2,1,0]=(0.238772787004)*x[0]**o+(0.987311136495)*x[0]+(0.215079265839)*x[1]**o+(-0.797762453244)*x[1]+(0.0329055379323)*x[2]**o+(-0.88999503049)*x[2]
        arg[0,2,1,1]=(-0.175884373587)*x[0]**o+(0.806690051494)*x[0]+(0.0086744009376)*x[1]**o+(0.286131063746)*x[1]+(0.536839534689)*x[2]**o+(-0.0946282917135)*x[2]
        arg[0,2,2,0]=(-0.615695203188)*x[0]**o+(0.600565627235)*x[0]+(-0.869759996675)*x[1]**o+(0.648192341374)*x[1]+(-0.952728999251)*x[2]**o+(0.564068973312)*x[2]
        arg[0,2,2,1]=(-0.0992379560876)*x[0]**o+(-0.897179656441)*x[0]+(-0.519600820875)*x[1]**o+(0.70507167679)*x[1]+(0.903640275297)*x[2]**o+(-0.645039734963)*x[2]
        arg[0,3,0,0]=(0.432415072474)*x[0]**o+(-0.748137008742)*x[0]+(0.875883324091)*x[1]**o+(0.706172632952)*x[1]+(-0.47486904255)*x[2]**o+(0.238106343647)*x[2]
        arg[0,3,0,1]=(-0.656397468061)*x[0]**o+(-0.530867058489)*x[0]+(-0.904148914114)*x[1]**o+(0.87213389632)*x[1]+(0.0612056123548)*x[2]**o+(0.288620109264)*x[2]
        arg[0,3,1,0]=(-0.472896985897)*x[0]**o+(0.83328574859)*x[0]+(0.744310819801)*x[1]**o+(0.0601035189502)*x[1]+(-0.198693260491)*x[2]**o+(-0.882304016543)*x[2]
        arg[0,3,1,1]=(-0.515568063685)*x[0]**o+(0.484765564236)*x[0]+(0.754304059424)*x[1]**o+(0.654790289063)*x[1]+(-0.102335538236)*x[2]**o+(-0.0705158755887)*x[2]
        arg[0,3,2,0]=(-0.709802317012)*x[0]**o+(-0.346551215683)*x[0]+(0.312943005623)*x[1]**o+(0.476828461567)*x[1]+(0.901171463566)*x[2]**o+(0.538981245022)*x[2]
        arg[0,3,2,1]=(-0.20057625974)*x[0]**o+(0.0458666109661)*x[0]+(0.909610730794)*x[1]**o+(0.145637065192)*x[1]+(0.666420073244)*x[2]**o+(0.0550124082656)*x[2]
        arg[0,4,0,0]=(-0.726921583373)*x[0]**o+(0.891306726707)*x[0]+(-0.189151461127)*x[1]**o+(0.479274047923)*x[1]+(-0.129077261233)*x[2]**o+(-0.873599130413)*x[2]
        arg[0,4,0,1]=(0.765272127757)*x[0]**o+(0.159307695759)*x[0]+(-0.569162341331)*x[1]**o+(-0.530158933095)*x[1]+(-0.216727788754)*x[2]**o+(-0.519551969741)*x[2]
        arg[0,4,1,0]=(-0.613269202518)*x[0]**o+(0.612846846793)*x[0]+(-0.187285796021)*x[1]**o+(-0.462603518143)*x[1]+(-0.739162067312)*x[2]**o+(0.0737393951189)*x[2]
        arg[0,4,1,1]=(-0.727485674691)*x[0]**o+(-0.461427895477)*x[0]+(0.917029475887)*x[1]**o+(0.265874989879)*x[1]+(0.602505162351)*x[2]**o+(-0.372206475027)*x[2]
        arg[0,4,2,0]=(-0.663071984573)*x[0]**o+(0.511217165998)*x[0]+(-0.472765672078)*x[1]**o+(0.814962161218)*x[1]+(-0.608890631297)*x[2]**o+(-0.284872941025)*x[2]
        arg[0,4,2,1]=(0.893396132015)*x[0]**o+(0.183468149957)*x[0]+(0.647812435076)*x[1]**o+(-0.398672312216)*x[1]+(0.6359553902)*x[2]**o+(0.485509522238)*x[2]
        arg[1,0,0,0]=(-0.0771632829864)*x[0]**o+(-0.723025089864)*x[0]+(0.486209418481)*x[1]**o+(0.0869704919133)*x[1]+(-0.365499197901)*x[2]**o+(-0.0796713538898)*x[2]
        arg[1,0,0,1]=(0.133617787097)*x[0]**o+(0.993872304226)*x[0]+(-0.898504795319)*x[1]**o+(0.656793416434)*x[1]+(-0.821841885906)*x[2]**o+(0.726972748423)*x[2]
        arg[1,0,1,0]=(0.174277385391)*x[0]**o+(-0.301857401564)*x[0]+(-0.846153302312)*x[1]**o+(-0.559076632516)*x[1]+(-0.89870762709)*x[2]**o+(0.640171641512)*x[2]
        arg[1,0,1,1]=(0.795443369588)*x[0]**o+(-0.346082013333)*x[0]+(-0.439713049567)*x[1]**o+(0.860680213906)*x[1]+(-0.466761099917)*x[2]**o+(-0.843656974958)*x[2]
        arg[1,0,2,0]=(0.482814141263)*x[0]**o+(0.142978643528)*x[0]+(-0.735291704767)*x[1]**o+(-0.925675892787)*x[1]+(0.747235439093)*x[2]**o+(-0.569454067374)*x[2]
        arg[1,0,2,1]=(0.619359374773)*x[0]**o+(-0.547172728307)*x[0]+(0.0829642963797)*x[1]**o+(-0.0180882163346)*x[1]+(0.328627445196)*x[2]**o+(-0.274608686384)*x[2]
        arg[1,1,0,0]=(-0.510375454058)*x[0]**o+(-0.779894165181)*x[0]+(0.656168894408)*x[1]**o+(-0.449484133421)*x[1]+(-0.147284453744)*x[2]**o+(-0.488384108271)*x[2]
        arg[1,1,0,1]=(-0.302705575836)*x[0]**o+(0.256670307547)*x[0]+(0.110465379068)*x[1]**o+(0.319291387545)*x[1]+(0.408832172454)*x[2]**o+(0.718793126931)*x[2]
        arg[1,1,1,0]=(0.520882033451)*x[0]**o+(0.412524861348)*x[0]+(0.761080123209)*x[1]**o+(0.401223078157)*x[1]+(-0.65945839721)*x[2]**o+(0.71349489395)*x[2]
        arg[1,1,1,1]=(0.624622427594)*x[0]**o+(0.847907716976)*x[0]+(0.855268296474)*x[1]**o+(-0.671685662164)*x[1]+(-0.407895715308)*x[2]**o+(-0.719054767733)*x[2]
        arg[1,1,2,0]=(-0.044254701409)*x[0]**o+(0.72911290175)*x[0]+(0.869467934211)*x[1]**o+(0.613006764744)*x[1]+(0.678602796496)*x[2]**o+(0.388923815748)*x[2]
        arg[1,1,2,1]=(-0.240571107795)*x[0]**o+(0.838382917252)*x[0]+(0.0342933298383)*x[1]**o+(0.595421789197)*x[1]+(-0.989088178853)*x[2]**o+(-0.20002339454)*x[2]
        arg[1,2,0,0]=(0.0596937021722)*x[0]**o+(0.196768537308)*x[0]+(0.0834459161806)*x[1]**o+(0.194049508373)*x[1]+(-0.097309631361)*x[2]**o+(-0.867555781902)*x[2]
        arg[1,2,0,1]=(-0.935541542647)*x[0]**o+(-0.00702430636955)*x[0]+(-0.384657002538)*x[1]**o+(0.186884376746)*x[1]+(0.97207387841)*x[2]**o+(0.452460932869)*x[2]
        arg[1,2,1,0]=(0.276741406646)*x[0]**o+(0.838348610585)*x[0]+(-0.403054578066)*x[1]**o+(0.655318135938)*x[1]+(0.37614761635)*x[2]**o+(0.82736662494)*x[2]
        arg[1,2,1,1]=(0.828731604537)*x[0]**o+(0.762067030749)*x[0]+(-0.881827056079)*x[1]**o+(0.250928989647)*x[1]+(-0.771040367979)*x[2]**o+(0.239950799138)*x[2]
        arg[1,2,2,0]=(-0.380797583569)*x[0]**o+(0.112805267222)*x[0]+(-0.454730562411)*x[1]**o+(0.926061735219)*x[1]+(0.348331256905)*x[2]**o+(-0.725641369775)*x[2]
        arg[1,2,2,1]=(0.257022926473)*x[0]**o+(0.0160557858816)*x[0]+(-0.291367922966)*x[1]**o+(-0.611341260765)*x[1]+(0.232472950268)*x[2]**o+(-0.738139185521)*x[2]
        arg[1,3,0,0]=(-0.645340088797)*x[0]**o+(0.480235460274)*x[0]+(0.135346246274)*x[1]**o+(-0.560730820351)*x[1]+(-0.635582025321)*x[2]**o+(-0.476912924392)*x[2]
        arg[1,3,0,1]=(0.124132693667)*x[0]**o+(-0.813993821983)*x[0]+(-0.108593659227)*x[1]**o+(-0.432186668568)*x[1]+(0.209382070426)*x[2]**o+(0.928912669667)*x[2]
        arg[1,3,1,0]=(-0.7660131868)*x[0]**o+(-0.30274186797)*x[0]+(0.140498262981)*x[1]**o+(0.993084314471)*x[1]+(0.684043026981)*x[2]**o+(-0.339724714081)*x[2]
        arg[1,3,1,1]=(-0.150578257351)*x[0]**o+(-0.674840941109)*x[0]+(0.469240334383)*x[1]**o+(0.729963338278)*x[1]+(-0.800829664334)*x[2]**o+(0.0467922871395)*x[2]
        arg[1,3,2,0]=(-0.336661555676)*x[0]**o+(0.705661383379)*x[0]+(-0.76993460355)*x[1]**o+(0.459144989218)*x[1]+(0.545901025663)*x[2]**o+(0.385988013614)*x[2]
        arg[1,3,2,1]=(0.165928745462)*x[0]**o+(0.540718600752)*x[0]+(0.172773708803)*x[1]**o+(0.253664803784)*x[1]+(-0.264896626947)*x[2]**o+(-0.478577475495)*x[2]
        arg[1,4,0,0]=(0.283586653542)*x[0]**o+(-0.104364884571)*x[0]+(-0.840555806526)*x[1]**o+(0.701868415915)*x[1]+(-0.290943689679)*x[2]**o+(-0.73024942052)*x[2]
        arg[1,4,0,1]=(-0.772849835334)*x[0]**o+(-0.432170589951)*x[0]+(-0.376099423267)*x[1]**o+(0.0215357491826)*x[1]+(-0.346787011464)*x[2]**o+(0.620411709813)*x[2]
        arg[1,4,1,0]=(-0.718524564964)*x[0]**o+(0.119936682579)*x[0]+(-0.994276256462)*x[1]**o+(-0.839783920473)*x[1]+(-0.454576336782)*x[2]**o+(-0.515489988915)*x[2]
        arg[1,4,1,1]=(-0.771664154234)*x[0]**o+(-0.584052299067)*x[0]+(-0.0233756210307)*x[1]**o+(-0.174442816636)*x[1]+(-0.702383477714)*x[2]**o+(-0.279952858144)*x[2]
        arg[1,4,2,0]=(-0.591902790978)*x[0]**o+(-0.151826578123)*x[0]+(0.923906788219)*x[1]**o+(0.301090558476)*x[1]+(-0.518314647811)*x[2]**o+(-0.784331717681)*x[2]
        arg[1,4,2,1]=(0.303567440092)*x[0]**o+(0.744948852961)*x[0]+(0.719939071212)*x[1]**o+(-0.161138638319)*x[1]+(0.924953261766)*x[2]**o+(-0.875565390131)*x[2]
        arg[2,0,0,0]=(0.197574550087)*x[0]**o+(-0.962018785397)*x[0]+(0.498378037582)*x[1]**o+(-0.334199169606)*x[1]+(0.852426750588)*x[2]**o+(-0.650173474618)*x[2]
        arg[2,0,0,1]=(0.0663499012532)*x[0]**o+(0.00812461763474)*x[0]+(0.237233309304)*x[1]**o+(-0.839377881658)*x[1]+(-0.67246872567)*x[2]**o+(0.677471261363)*x[2]
        arg[2,0,1,0]=(-0.789090971877)*x[0]**o+(-0.387724517273)*x[0]+(0.455954877641)*x[1]**o+(-0.356995254842)*x[1]+(-0.897019846935)*x[2]**o+(-0.0703543939474)*x[2]
        arg[2,0,1,1]=(0.222873274437)*x[0]**o+(0.405228982254)*x[0]+(-0.532697993554)*x[1]**o+(-0.836340143789)*x[1]+(-0.139525062661)*x[2]**o+(-0.916307770461)*x[2]
        arg[2,0,2,0]=(-0.300828039715)*x[0]**o+(-0.584457605313)*x[0]+(-0.443125109903)*x[1]**o+(0.292734004842)*x[1]+(-0.55875691535)*x[2]**o+(0.391237822545)*x[2]
        arg[2,0,2,1]=(-0.501507304569)*x[0]**o+(-0.203549089015)*x[0]+(-0.113050138121)*x[1]**o+(0.712978728546)*x[1]+(-0.642617274635)*x[2]**o+(-0.307119076865)*x[2]
        arg[2,1,0,0]=(-0.809237327867)*x[0]**o+(0.984684980548)*x[0]+(-0.643310231704)*x[1]**o+(0.759430082065)*x[1]+(-0.393902220205)*x[2]**o+(-0.900173371318)*x[2]
        arg[2,1,0,1]=(-0.722795722506)*x[0]**o+(0.433309366611)*x[0]+(-0.559871890844)*x[1]**o+(0.417301432603)*x[1]+(-0.364192390887)*x[2]**o+(0.64549736058)*x[2]
        arg[2,1,1,0]=(-0.551325611251)*x[0]**o+(-0.638520772682)*x[0]+(0.222655558982)*x[1]**o+(-0.230002147207)*x[1]+(-0.290821163957)*x[2]**o+(0.0647728628594)*x[2]
        arg[2,1,1,1]=(0.743453045588)*x[0]**o+(-0.901163108996)*x[0]+(-0.546031302921)*x[1]**o+(-0.512614476111)*x[1]+(-0.886430682715)*x[2]**o+(0.720650678026)*x[2]
        arg[2,1,2,0]=(-0.033787671958)*x[0]**o+(0.930927279443)*x[0]+(-0.530482747719)*x[1]**o+(0.714053784656)*x[1]+(-0.0978565618487)*x[2]**o+(-0.0423072209447)*x[2]
        arg[2,1,2,1]=(0.127121989623)*x[0]**o+(0.809654796091)*x[0]+(0.718533701795)*x[1]**o+(0.542738494557)*x[1]+(-0.342918607246)*x[2]**o+(-0.627991145352)*x[2]
        arg[2,2,0,0]=(-0.867530011926)*x[0]**o+(0.870295856967)*x[0]+(-0.13043222912)*x[1]**o+(0.2670058777)*x[1]+(0.395016123247)*x[2]**o+(-0.811207959629)*x[2]
        arg[2,2,0,1]=(-0.00858729113056)*x[0]**o+(-0.885551110851)*x[0]+(0.820441870015)*x[1]**o+(-0.348896120549)*x[1]+(0.295475751596)*x[2]**o+(-0.672688762902)*x[2]
        arg[2,2,1,0]=(-0.982123891355)*x[0]**o+(0.232770408062)*x[0]+(0.374641341895)*x[1]**o+(-0.21375197203)*x[1]+(0.436992590849)*x[2]**o+(-0.473399371298)*x[2]
        arg[2,2,1,1]=(-0.415533190881)*x[0]**o+(0.0945615271507)*x[0]+(-0.281959759072)*x[1]**o+(0.0807942449559)*x[1]+(0.225501997354)*x[2]**o+(0.305247648387)*x[2]
        arg[2,2,2,0]=(-0.559916632426)*x[0]**o+(-0.372679495574)*x[0]+(0.392857820158)*x[1]**o+(0.726824319268)*x[1]+(0.36923535014)*x[2]**o+(-0.208169256022)*x[2]
        arg[2,2,2,1]=(-0.869055045156)*x[0]**o+(0.359188490398)*x[0]+(0.702110073265)*x[1]**o+(0.668757858248)*x[1]+(-0.499300337779)*x[2]**o+(0.0471054271112)*x[2]
        arg[2,3,0,0]=(0.754089271004)*x[0]**o+(-0.0240479506036)*x[0]+(-0.0775830745475)*x[1]**o+(0.215564363064)*x[1]+(-0.791969192801)*x[2]**o+(-0.130308630692)*x[2]
        arg[2,3,0,1]=(0.448970570476)*x[0]**o+(0.510346880872)*x[0]+(-0.561996764085)*x[1]**o+(0.217897696493)*x[1]+(-0.0610826665756)*x[2]**o+(0.387896322636)*x[2]
        arg[2,3,1,0]=(0.737635683404)*x[0]**o+(-0.35395540781)*x[0]+(0.0419652573208)*x[1]**o+(-0.519735381528)*x[1]+(0.739523085214)*x[2]**o+(0.808529094094)*x[2]
        arg[2,3,1,1]=(0.551348499194)*x[0]**o+(-0.40145493368)*x[0]+(0.211723169473)*x[1]**o+(-0.762751144824)*x[1]+(-0.17710033067)*x[2]**o+(0.688961224051)*x[2]
        arg[2,3,2,0]=(-0.870721462645)*x[0]**o+(-0.0356869329313)*x[0]+(0.277142920424)*x[1]**o+(-0.154474999666)*x[1]+(0.837001124911)*x[2]**o+(-0.233028214567)*x[2]
        arg[2,3,2,1]=(-0.234260935528)*x[0]**o+(0.913718165996)*x[0]+(0.385678747942)*x[1]**o+(0.237966356277)*x[1]+(0.382072927695)*x[2]**o+(0.627666395757)*x[2]
        arg[2,4,0,0]=(-0.474674316363)*x[0]**o+(-0.575533683959)*x[0]+(0.751489557666)*x[1]**o+(-0.0623565445313)*x[1]+(-0.517274743843)*x[2]**o+(-0.712143034168)*x[2]
        arg[2,4,0,1]=(0.259723799797)*x[0]**o+(0.345952822815)*x[0]+(0.169940774832)*x[1]**o+(-0.185154259609)*x[1]+(0.418430344681)*x[2]**o+(0.861837327531)*x[2]
        arg[2,4,1,0]=(-0.960970748682)*x[0]**o+(0.842706511282)*x[0]+(-0.771779045108)*x[1]**o+(-0.63675621795)*x[1]+(-0.317416556239)*x[2]**o+(0.287317968341)*x[2]
        arg[2,4,1,1]=(-0.753944077832)*x[0]**o+(0.570245260805)*x[0]+(-0.0226797611003)*x[1]**o+(0.0776773843077)*x[1]+(0.629536717673)*x[2]**o+(0.271216337699)*x[2]
        arg[2,4,2,0]=(-0.233304828171)*x[0]**o+(-0.217436697538)*x[0]+(-0.373618121496)*x[1]**o+(-0.342584906659)*x[1]+(0.0114484228954)*x[2]**o+(-0.639753178531)*x[2]
        arg[2,4,2,1]=(0.341373328517)*x[0]**o+(-0.432795268788)*x[0]+(-0.871263113926)*x[1]**o+(0.843590946949)*x[1]+(-0.653058566267)*x[2]**o+(-0.894366240333)*x[2]
        arg[3,0,0,0]=(-0.472652289489)*x[0]**o+(0.697119499826)*x[0]+(-0.195781503945)*x[1]**o+(-0.716872496009)*x[1]+(-0.159070341485)*x[2]**o+(-0.700548071467)*x[2]
        arg[3,0,0,1]=(-0.398674023611)*x[0]**o+(-0.690683017472)*x[0]+(0.957112237189)*x[1]**o+(-0.782702815224)*x[1]+(-0.249336230999)*x[2]**o+(0.22734983387)*x[2]
        arg[3,0,1,0]=(-0.0216721345971)*x[0]**o+(0.283765576696)*x[0]+(-0.35531833417)*x[1]**o+(0.974402352836)*x[1]+(-0.568820213603)*x[2]**o+(0.957610279127)*x[2]
        arg[3,0,1,1]=(0.0606265274442)*x[0]**o+(-0.941169961629)*x[0]+(0.804902933622)*x[1]**o+(-0.857158317629)*x[1]+(-0.227167312547)*x[2]**o+(0.945175823328)*x[2]
        arg[3,0,2,0]=(0.630444608821)*x[0]**o+(-0.192343222551)*x[0]+(0.0342249337232)*x[1]**o+(0.970401334961)*x[1]+(-0.608568046213)*x[2]**o+(-0.539374299133)*x[2]
        arg[3,0,2,1]=(-0.348127940577)*x[0]**o+(-0.895424161569)*x[0]+(0.918173840926)*x[1]**o+(0.99658998821)*x[1]+(-0.28733372377)*x[2]**o+(-0.516441029544)*x[2]
        arg[3,1,0,0]=(0.735069593965)*x[0]**o+(-0.727899981109)*x[0]+(-0.325639355247)*x[1]**o+(0.751064830331)*x[1]+(-0.297126073245)*x[2]**o+(-0.673421779764)*x[2]
        arg[3,1,0,1]=(0.00281719106153)*x[0]**o+(0.246929980679)*x[0]+(-0.48091940116)*x[1]**o+(0.501927059357)*x[1]+(-0.520099802736)*x[2]**o+(-0.902706012118)*x[2]
        arg[3,1,1,0]=(-0.341732637339)*x[0]**o+(-0.169802234575)*x[0]+(0.156725375114)*x[1]**o+(0.739176104936)*x[1]+(-0.374572295196)*x[2]**o+(-0.436641001832)*x[2]
        arg[3,1,1,1]=(0.644746847185)*x[0]**o+(-0.992126325722)*x[0]+(0.551602368714)*x[1]**o+(-0.839306768893)*x[1]+(-0.298981617653)*x[2]**o+(0.089795271605)*x[2]
        arg[3,1,2,0]=(0.382425223801)*x[0]**o+(-0.949095041278)*x[0]+(-0.135997266858)*x[1]**o+(-0.871114033611)*x[1]+(-0.422984170709)*x[2]**o+(-0.698818784989)*x[2]
        arg[3,1,2,1]=(0.630558573276)*x[0]**o+(0.90644556976)*x[0]+(0.512953870249)*x[1]**o+(-0.58514271049)*x[1]+(-0.0587225891088)*x[2]**o+(-0.390796465925)*x[2]
        arg[3,2,0,0]=(-0.404884771697)*x[0]**o+(-0.178144326223)*x[0]+(-0.606503995405)*x[1]**o+(0.190025878248)*x[1]+(0.735831203504)*x[2]**o+(-0.567394607398)*x[2]
        arg[3,2,0,1]=(0.28819895431)*x[0]**o+(0.456573144203)*x[0]+(0.00856899200079)*x[1]**o+(0.547827005333)*x[1]+(0.087428132552)*x[2]**o+(0.759650060552)*x[2]
        arg[3,2,1,0]=(0.949790710961)*x[0]**o+(-0.685602682574)*x[0]+(0.75278809536)*x[1]**o+(-0.876689834021)*x[1]+(0.0116394079709)*x[2]**o+(0.319113513167)*x[2]
        arg[3,2,1,1]=(0.619659549755)*x[0]**o+(-0.397388124185)*x[0]+(0.460679661901)*x[1]**o+(-0.507122241204)*x[1]+(0.985741194076)*x[2]**o+(0.916231863626)*x[2]
        arg[3,2,2,0]=(-0.457269257294)*x[0]**o+(-0.47265466694)*x[0]+(0.900051837969)*x[1]**o+(-0.148802803439)*x[1]+(-0.195540487054)*x[2]**o+(0.0784524130754)*x[2]
        arg[3,2,2,1]=(0.933553130797)*x[0]**o+(0.617362831046)*x[0]+(0.263616954879)*x[1]**o+(-0.0658887625105)*x[1]+(0.48091174024)*x[2]**o+(0.452935484504)*x[2]
        arg[3,3,0,0]=(0.825594594128)*x[0]**o+(0.350203407373)*x[0]+(0.211125321741)*x[1]**o+(0.194914147746)*x[1]+(-0.639432220749)*x[2]**o+(-0.678718852935)*x[2]
        arg[3,3,0,1]=(0.514394313629)*x[0]**o+(-0.595261005308)*x[0]+(-0.0219208527235)*x[1]**o+(-0.77204975709)*x[1]+(-0.483587370025)*x[2]**o+(-0.559464911044)*x[2]
        arg[3,3,1,0]=(0.821240511475)*x[0]**o+(-0.403492177329)*x[0]+(0.486132478046)*x[1]**o+(-0.0187529457547)*x[1]+(0.73209003456)*x[2]**o+(0.226791335797)*x[2]
        arg[3,3,1,1]=(-0.644256768839)*x[0]**o+(0.0372588784686)*x[0]+(0.726561650796)*x[1]**o+(0.518003809017)*x[1]+(0.738146681601)*x[2]**o+(-0.256036433171)*x[2]
        arg[3,3,2,0]=(0.534862520097)*x[0]**o+(0.275559826342)*x[0]+(0.292383677513)*x[1]**o+(-0.448938027415)*x[1]+(0.281053776971)*x[2]**o+(0.0119964402657)*x[2]
        arg[3,3,2,1]=(0.722071427786)*x[0]**o+(0.150280744346)*x[0]+(-0.0354786178164)*x[1]**o+(-0.703984035201)*x[1]+(0.105365795033)*x[2]**o+(-0.556567650379)*x[2]
        arg[3,4,0,0]=(0.188625467698)*x[0]**o+(0.737108753179)*x[0]+(0.481952725951)*x[1]**o+(0.841958412185)*x[1]+(-0.329504392569)*x[2]**o+(0.228456367852)*x[2]
        arg[3,4,0,1]=(0.181430484096)*x[0]**o+(0.307078178179)*x[0]+(0.262079226853)*x[1]**o+(0.111656765153)*x[1]+(0.15340094176)*x[2]**o+(-0.859774252333)*x[2]
        arg[3,4,1,0]=(0.213491785855)*x[0]**o+(-0.479441582203)*x[0]+(0.0527576231613)*x[1]**o+(0.328465118263)*x[1]+(-0.645719217198)*x[2]**o+(-0.550569446193)*x[2]
        arg[3,4,1,1]=(-0.401744411188)*x[0]**o+(-0.626446165761)*x[0]+(-0.191670239416)*x[1]**o+(0.866888407009)*x[1]+(-0.341220740645)*x[2]**o+(-0.742685031762)*x[2]
        arg[3,4,2,0]=(-0.0813611687514)*x[0]**o+(0.318680684889)*x[0]+(-0.780434724843)*x[1]**o+(-0.0902735483921)*x[1]+(0.586108990767)*x[2]**o+(0.934041577436)*x[2]
        arg[3,4,2,1]=(-0.0991425040958)*x[0]**o+(0.99274076773)*x[0]+(0.924248544389)*x[1]**o+(0.290576034321)*x[1]+(0.452104871421)*x[2]**o+(0.774017129246)*x[2]
        ref[0,0,0,0]=(-0.445809185937)/(o+1.)+(-0.11049649315)+(0.480896043196)*0.5**o
        ref[0,0,0,1]=(-0.776476434405)/(o+1.)+(-0.842010629213)+(-0.609632888023)*0.5**o
        ref[0,0,1,0]=(0.182977545138)/(o+1.)+(-0.124558039762)+(0.205065167181)*0.5**o
        ref[0,0,1,1]=(-1.17104760039)/(o+1.)+(0.608297882891)+(0.875716461075)*0.5**o
        ref[0,0,2,0]=(0.647678117252)/(o+1.)+(-0.752248832131)+(0.711372030289)*0.5**o
        ref[0,0,2,1]=(-0.0424453226222)/(o+1.)+(0.152363855976)+(-0.0536842545617)*0.5**o
        ref[0,1,0,0]=(-0.97072557194)/(o+1.)+(0.355747795628)+(-0.710091748251)*0.5**o
        ref[0,1,0,1]=(1.54210036269)/(o+1.)+(-0.0605096710259)+(0.250989366649)*0.5**o
        ref[0,1,1,0]=(1.33375575444)/(o+1.)+(-0.237134583454)+(0.690360096543)*0.5**o
        ref[0,1,1,1]=(0.0771949345448)/(o+1.)+(-0.00507329027461)+(-0.815220933301)*0.5**o
        ref[0,1,2,0]=(-0.149444846314)/(o+1.)+(-0.226886808857)+(0.169442617085)*0.5**o
        ref[0,1,2,1]=(0.114952830564)/(o+1.)+(-0.613038537369)+(-0.110636656112)*0.5**o
        ref[0,2,0,0]=(-0.0135089970172)/(o+1.)+(-0.104741016697)+(0.340661037085)*0.5**o
        ref[0,2,0,1]=(0.877950660854)/(o+1.)+(-0.395230122376)+(0.185282186176)*0.5**o
        ref[0,2,1,0]=(0.247984803771)/(o+1.)+(-0.35022317362)+(0.238772787004)*0.5**o
        ref[0,2,1,1]=(0.545513935627)/(o+1.)+(0.499096411763)+(-0.175884373587)*0.5**o
        ref[0,2,2,0]=(-1.82248899593)/(o+1.)+(0.90641347096)+(-0.615695203188)*0.5**o
        ref[0,2,2,1]=(0.384039454422)/(o+1.)+(-0.418573857307)+(-0.0992379560876)*0.5**o
        ref[0,3,0,0]=(0.401014281541)/(o+1.)+(0.0980709839284)+(0.432415072474)*0.5**o
        ref[0,3,0,1]=(-0.842943301759)/(o+1.)+(0.314943473548)+(-0.656397468061)*0.5**o
        ref[0,3,1,0]=(0.54561755931)/(o+1.)+(0.00554262549843)+(-0.472896985897)*0.5**o
        ref[0,3,1,1]=(0.651968521189)/(o+1.)+(0.534519988855)+(-0.515568063685)*0.5**o
        ref[0,3,2,0]=(1.21411446919)/(o+1.)+(0.334629245453)+(-0.709802317012)*0.5**o
        ref[0,3,2,1]=(1.57603080404)/(o+1.)+(0.123258042212)+(-0.20057625974)*0.5**o
        ref[0,4,0,0]=(-0.318228722361)/(o+1.)+(0.248490822109)+(-0.726921583373)*0.5**o
        ref[0,4,0,1]=(-0.785890130085)/(o+1.)+(-0.445201603538)+(0.765272127757)*0.5**o
        ref[0,4,1,0]=(-0.926447863334)/(o+1.)+(0.111991361885)+(-0.613269202518)*0.5**o
        ref[0,4,1,1]=(1.51953463824)/(o+1.)+(-0.283879690312)+(-0.727485674691)*0.5**o
        ref[0,4,2,0]=(-1.08165630338)/(o+1.)+(0.520653193096)+(-0.663071984573)*0.5**o
        ref[0,4,2,1]=(1.28376782528)/(o+1.)+(0.135152679989)+(0.893396132015)*0.5**o
        ref[1,0,0,0]=(0.12071022058)/(o+1.)+(-0.35786297592)+(-0.0771632829864)*0.5**o
        ref[1,0,0,1]=(-1.72034668123)/(o+1.)+(1.18881923454)+(0.133617787097)*0.5**o
        ref[1,0,1,0]=(-1.7448609294)/(o+1.)+(-0.110381196284)+(0.174277385391)*0.5**o
        ref[1,0,1,1]=(-0.906474149485)/(o+1.)+(-0.164529387192)+(0.795443369588)*0.5**o
        ref[1,0,2,0]=(0.0119437343259)/(o+1.)+(-0.676075658316)+(0.482814141263)*0.5**o
        ref[1,0,2,1]=(0.411591741576)/(o+1.)+(-0.419934815512)+(0.619359374773)*0.5**o
        ref[1,1,0,0]=(0.508884440665)/(o+1.)+(-0.858881203437)+(-0.510375454058)*0.5**o
        ref[1,1,0,1]=(0.519297551522)/(o+1.)+(0.647377411012)+(-0.302705575836)*0.5**o
        ref[1,1,1,0]=(0.101621725999)/(o+1.)+(0.763621416728)+(0.520882033451)*0.5**o
        ref[1,1,1,1]=(0.447372581166)/(o+1.)+(-0.27141635646)+(0.624622427594)*0.5**o
        ref[1,1,2,0]=(1.54807073071)/(o+1.)+(0.865521741121)+(-0.044254701409)*0.5**o
        ref[1,1,2,1]=(-0.954794849015)/(o+1.)+(0.616890655954)+(-0.240571107795)*0.5**o
        ref[1,2,0,0]=(-0.0138637151804)/(o+1.)+(-0.23836886811)+(0.0596937021722)*0.5**o
        ref[1,2,0,1]=(0.587416875871)/(o+1.)+(0.316160501623)+(-0.935541542647)*0.5**o
        ref[1,2,1,0]=(-0.0269069617163)/(o+1.)+(1.16051668573)+(0.276741406646)*0.5**o
        ref[1,2,1,1]=(-1.65286742406)/(o+1.)+(0.626473409767)+(0.828731604537)*0.5**o
        ref[1,2,2,0]=(-0.106399305507)/(o+1.)+(0.156612816333)+(-0.380797583569)*0.5**o
        ref[1,2,2,1]=(-0.0588949726984)/(o+1.)+(-0.666712330202)+(0.257022926473)*0.5**o
        ref[1,3,0,0]=(-0.500235779047)/(o+1.)+(-0.278704142234)+(-0.645340088797)*0.5**o
        ref[1,3,0,1]=(0.100788411199)/(o+1.)+(-0.158633910442)+(0.124132693667)*0.5**o
        ref[1,3,1,0]=(0.824541289962)/(o+1.)+(0.17530886621)+(-0.7660131868)*0.5**o
        ref[1,3,1,1]=(-0.331589329951)/(o+1.)+(0.0509573421542)+(-0.150578257351)*0.5**o
        ref[1,3,2,0]=(-0.224033577887)/(o+1.)+(0.775397193106)+(-0.336661555676)*0.5**o
        ref[1,3,2,1]=(-0.0921229181433)/(o+1.)+(0.15790296452)+(0.165928745462)*0.5**o
        ref[1,4,0,0]=(-1.13149949621)/(o+1.)+(-0.0663729445879)+(0.283586653542)*0.5**o
        ref[1,4,0,1]=(-0.722886434731)/(o+1.)+(0.104888434522)+(-0.772849835334)*0.5**o
        ref[1,4,1,0]=(-1.44885259324)/(o+1.)+(-0.617668613405)+(-0.718524564964)*0.5**o
        ref[1,4,1,1]=(-0.725759098745)/(o+1.)+(-0.519223986923)+(-0.771664154234)*0.5**o
        ref[1,4,2,0]=(0.405592140408)/(o+1.)+(-0.317533868664)+(-0.591902790978)*0.5**o
        ref[1,4,2,1]=(1.64489233298)/(o+1.)+(-0.145877587744)+(0.303567440092)*0.5**o
        ref[2,0,0,0]=(1.35080478817)/(o+1.)+(-0.97319571481)+(0.197574550087)*0.5**o
        ref[2,0,0,1]=(-0.435235416366)/(o+1.)+(-0.0768910013301)+(0.0663499012532)*0.5**o
        ref[2,0,1,0]=(-0.441064969294)/(o+1.)+(-0.407537083031)+(-0.789090971877)*0.5**o
        ref[2,0,1,1]=(-0.672223056216)/(o+1.)+(-0.673709465998)+(0.222873274437)*0.5**o
        ref[2,0,2,0]=(-1.00188202525)/(o+1.)+(0.0497571110368)+(-0.300828039715)*0.5**o
        ref[2,0,2,1]=(-0.755667412756)/(o+1.)+(0.101155281333)+(-0.501507304569)*0.5**o
        ref[2,1,0,0]=(-1.03721245191)/(o+1.)+(0.421970845648)+(-0.809237327867)*0.5**o
        ref[2,1,0,1]=(-0.924064281731)/(o+1.)+(0.748054079897)+(-0.722795722506)*0.5**o
        ref[2,1,1,0]=(-0.0681656049743)/(o+1.)+(-0.401875028515)+(-0.551325611251)*0.5**o
        ref[2,1,1,1]=(-1.43246198564)/(o+1.)+(-0.346563453541)+(0.743453045588)*0.5**o
        ref[2,1,2,0]=(-0.628339309568)/(o+1.)+(0.801336921577)+(-0.033787671958)*0.5**o
        ref[2,1,2,1]=(0.375615094549)/(o+1.)+(0.362201072648)+(0.127121989623)*0.5**o
        ref[2,2,0,0]=(0.264583894127)/(o+1.)+(0.163046887519)+(-0.867530011926)*0.5**o
        ref[2,2,0,1]=(1.11591762161)/(o+1.)+(-0.953567997151)+(-0.00858729113056)*0.5**o
        ref[2,2,1,0]=(0.811633932744)/(o+1.)+(-0.227190467633)+(-0.982123891355)*0.5**o
        ref[2,2,1,1]=(-0.0564577617182)/(o+1.)+(0.240301710247)+(-0.415533190881)*0.5**o
        ref[2,2,2,0]=(0.762093170298)/(o+1.)+(0.072987783836)+(-0.559916632426)*0.5**o
        ref[2,2,2,1]=(0.202809735486)/(o+1.)+(0.537525887878)+(-0.869055045156)*0.5**o
        ref[2,3,0,0]=(-0.869552267348)/(o+1.)+(0.0306038908843)+(0.754089271004)*0.5**o
        ref[2,3,0,1]=(-0.62307943066)/(o+1.)+(0.55807045)+(0.448970570476)*0.5**o
        ref[2,3,1,0]=(0.781488342534)/(o+1.)+(-0.0325808476221)+(0.737635683404)*0.5**o
        ref[2,3,1,1]=(0.0346228388027)/(o+1.)+(-0.237622427226)+(0.551348499194)*0.5**o
        ref[2,3,2,0]=(1.11414404534)/(o+1.)+(-0.211595073582)+(-0.870721462645)*0.5**o
        ref[2,3,2,1]=(0.767751675637)/(o+1.)+(0.889675459015)+(-0.234260935528)*0.5**o
        ref[2,4,0,0]=(0.234214813823)/(o+1.)+(-0.675016631329)+(-0.474674316363)*0.5**o
        ref[2,4,0,1]=(0.588371119514)/(o+1.)+(0.511317945368)+(0.259723799797)*0.5**o
        ref[2,4,1,0]=(-1.08919560135)/(o+1.)+(0.246634130836)+(-0.960970748682)*0.5**o
        ref[2,4,1,1]=(0.606856956573)/(o+1.)+(0.459569491406)+(-0.753944077832)*0.5**o
        ref[2,4,2,0]=(-0.3621696986)/(o+1.)+(-0.599887391364)+(-0.233304828171)*0.5**o
        ref[2,4,2,1]=(-1.52432168019)/(o+1.)+(-0.241785281086)+(0.341373328517)*0.5**o
        ref[3,0,0,0]=(-0.35485184543)/(o+1.)+(-0.360150533825)+(-0.472652289489)*0.5**o
        ref[3,0,0,1]=(0.70777600619)/(o+1.)+(-0.623017999413)+(-0.398674023611)*0.5**o
        ref[3,0,1,0]=(-0.924138547773)/(o+1.)+(1.10788910433)+(-0.0216721345971)*0.5**o
        ref[3,0,1,1]=(0.577735621075)/(o+1.)+(-0.426576227965)+(0.0606265274442)*0.5**o
        ref[3,0,2,0]=(-0.57434311249)/(o+1.)+(0.119341906638)+(0.630444608821)*0.5**o
        ref[3,0,2,1]=(0.630840117156)/(o+1.)+(-0.207637601452)+(-0.348127940577)*0.5**o
        ref[3,1,0,0]=(-0.622765428492)/(o+1.)+(-0.32512846527)+(0.735069593965)*0.5**o
        ref[3,1,0,1]=(-1.0010192039)/(o+1.)+(-0.076924486041)+(0.00281719106153)*0.5**o
        ref[3,1,1,0]=(-0.217846920082)/(o+1.)+(0.0663664342643)+(-0.341732637339)*0.5**o
        ref[3,1,1,1]=(0.25262075106)/(o+1.)+(-0.870818911505)+(0.644746847185)*0.5**o
        ref[3,1,2,0]=(-0.558981437567)/(o+1.)+(-1.25951392994)+(0.382425223801)*0.5**o
        ref[3,1,2,1]=(0.454231281141)/(o+1.)+(-0.034746803327)+(0.630558573276)*0.5**o
        ref[3,2,0,0]=(0.129327208099)/(o+1.)+(-0.277756527686)+(-0.404884771697)*0.5**o
        ref[3,2,0,1]=(0.0959971245528)/(o+1.)+(0.882025105044)+(0.28819895431)*0.5**o
        ref[3,2,1,0]=(0.764427503331)/(o+1.)+(-0.621589501714)+(0.949790710961)*0.5**o
        ref[3,2,1,1]=(1.44642085598)/(o+1.)+(0.00586074911794)+(0.619659549755)*0.5**o
        ref[3,2,2,0]=(0.704511350915)/(o+1.)+(-0.271502528652)+(-0.457269257294)*0.5**o
        ref[3,2,2,1]=(0.744528695119)/(o+1.)+(0.50220477652)+(0.933553130797)*0.5**o
        ref[3,3,0,0]=(-0.428306899008)/(o+1.)+(-0.0668006489082)+(0.825594594128)*0.5**o
        ref[3,3,0,1]=(-0.505508222749)/(o+1.)+(-0.963387836721)+(0.514394313629)*0.5**o
        ref[3,3,1,0]=(1.21822251261)/(o+1.)+(-0.0977268936434)+(0.821240511475)*0.5**o
        ref[3,3,1,1]=(1.4647083324)/(o+1.)+(0.149613127157)+(-0.644256768839)*0.5**o
        ref[3,3,2,0]=(0.573437454483)/(o+1.)+(-0.0806908804035)+(0.534862520097)*0.5**o
        ref[3,3,2,1]=(0.069887177217)/(o+1.)+(-0.555135470617)+(0.722071427786)*0.5**o
        ref[3,4,0,0]=(0.152448333383)/(o+1.)+(0.903761766608)+(0.188625467698)*0.5**o
        ref[3,4,0,1]=(0.415480168613)/(o+1.)+(-0.220519654501)+(0.181430484096)*0.5**o
        ref[3,4,1,0]=(-0.592961594036)/(o+1.)+(-0.350772955066)+(0.213491785855)*0.5**o
        ref[3,4,1,1]=(-0.532890980061)/(o+1.)+(-0.251121395257)+(-0.401744411188)*0.5**o
        ref[3,4,2,0]=(-0.194325734076)/(o+1.)+(0.581224356966)+(-0.0813611687514)*0.5**o
        ref[3,4,2,1]=(1.37635341581)/(o+1.)+(1.02866696565)+(-0.0991425040958)*0.5**o
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.492060381705)*x[0]**o+(0.578791868713)*x[0]+(-0.439849757114)*x[1]**o+(-0.52629328967)*x[1]
        ref=(0.492060381705)*x_ref[0]**o+(0.578791868713)*x_ref[0]+(-0.439849757114)*x_ref[1]**o+(-0.52629328967)*x_ref[1]
      else:
        arg=(0.954737963273)*x[0]**o+(0.452578946878)*x[0]+(0.0471085060698)*x[1]**o+(-0.640169119832)*x[1]+(0.710401490627)*x[2]**o+(-0.382589884739)*x[2]
        ref=(0.954737963273)*x_ref[0]**o+(0.452578946878)*x_ref[0]+(0.0471085060698)*x_ref[1]**o+(-0.640169119832)*x_ref[1]+(0.710401490627)*x_ref[2]**o+(-0.382589884739)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.59604585527)*x[0]**o+(0.826981169401)*x[0]+(0.227617349484)*x[1]**o+(0.184907819029)*x[1]
        arg[1]=(-0.424684036731)*x[0]**o+(0.346537997905)*x[0]+(0.891921272597)*x[1]**o+(-0.145924182872)*x[1]
        ref[0]=(-0.59604585527)*x_ref[0]**o+(0.826981169401)*x_ref[0]+(0.227617349484)*x_ref[1]**o+(0.184907819029)*x_ref[1]
        ref[1]=(-0.424684036731)*x_ref[0]**o+(0.346537997905)*x_ref[0]+(0.891921272597)*x_ref[1]**o+(-0.145924182872)*x_ref[1]
      else:
        arg[0]=(0.777228173978)*x[0]**o+(-0.812215827155)*x[0]+(-0.618922680968)*x[1]**o+(0.0660312103206)*x[1]+(-0.917545504466)*x[2]**o+(0.988524619414)*x[2]
        arg[1]=(0.110219469565)*x[0]**o+(-0.441866866062)*x[0]+(-0.654447764335)*x[1]**o+(0.230716133289)*x[1]+(0.347487317165)*x[2]**o+(-0.948877903043)*x[2]
        ref[0]=(0.777228173978)*x_ref[0]**o+(-0.812215827155)*x_ref[0]+(-0.618922680968)*x_ref[1]**o+(0.0660312103206)*x_ref[1]+(-0.917545504466)*x_ref[2]**o+(0.988524619414)*x_ref[2]
        ref[1]=(0.110219469565)*x_ref[0]**o+(-0.441866866062)*x_ref[0]+(-0.654447764335)*x_ref[1]**o+(0.230716133289)*x_ref[1]+(0.347487317165)*x_ref[2]**o+(-0.948877903043)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.958809754913)*x[0]**o+(-0.844717049041)*x[0]+(0.462564514427)*x[1]**o+(-0.247678335594)*x[1]
        arg[0,1]=(0.700700054189)*x[0]**o+(-0.653724611457)*x[0]+(0.52659177803)*x[1]**o+(-0.387559446418)*x[1]
        arg[0,2]=(-0.326124300134)*x[0]**o+(0.115197897152)*x[0]+(-0.876340641846)*x[1]**o+(0.433060472581)*x[1]
        arg[0,3]=(-0.101215736407)*x[0]**o+(-0.483684510661)*x[0]+(0.0418770907036)*x[1]**o+(-0.845564414421)*x[1]
        arg[0,4]=(-0.293094855718)*x[0]**o+(0.710954930773)*x[0]+(0.238671713869)*x[1]**o+(0.576638343412)*x[1]
        arg[1,0]=(0.529861280122)*x[0]**o+(-0.259437904048)*x[0]+(-0.620725288158)*x[1]**o+(0.326787999382)*x[1]
        arg[1,1]=(0.421245543012)*x[0]**o+(-0.954139636698)*x[0]+(0.0498199734926)*x[1]**o+(0.24010238499)*x[1]
        arg[1,2]=(0.333362362373)*x[0]**o+(-0.999953175952)*x[0]+(0.866193849246)*x[1]**o+(0.764287410177)*x[1]
        arg[1,3]=(0.470415957539)*x[0]**o+(-0.133477966871)*x[0]+(-0.0416970809832)*x[1]**o+(-0.416352740444)*x[1]
        arg[1,4]=(-0.178613018492)*x[0]**o+(-0.967264985978)*x[0]+(-0.297102423687)*x[1]**o+(0.884184973413)*x[1]
        arg[2,0]=(-0.29363453294)*x[0]**o+(-0.0237360594862)*x[0]+(-0.300776450747)*x[1]**o+(-0.493932721521)*x[1]
        arg[2,1]=(0.795045745699)*x[0]**o+(0.984956851154)*x[0]+(-0.724208405235)*x[1]**o+(0.59029067239)*x[1]
        arg[2,2]=(-0.803012112531)*x[0]**o+(-0.831727785428)*x[0]+(0.330295112332)*x[1]**o+(0.0966317908147)*x[1]
        arg[2,3]=(-0.652332429144)*x[0]**o+(-0.38689457016)*x[0]+(-0.391780825295)*x[1]**o+(-0.444284790848)*x[1]
        arg[2,4]=(0.416775161398)*x[0]**o+(-0.16281006536)*x[0]+(-0.742938932867)*x[1]**o+(0.456486183714)*x[1]
        arg[3,0]=(0.506731632086)*x[0]**o+(-0.222468320035)*x[0]+(-0.906398223556)*x[1]**o+(-0.0250630448611)*x[1]
        arg[3,1]=(-0.39207398783)*x[0]**o+(-0.574739374252)*x[0]+(0.262101584569)*x[1]**o+(0.249587535087)*x[1]
        arg[3,2]=(-0.252593913503)*x[0]**o+(-0.864294052989)*x[0]+(0.0110434296336)*x[1]**o+(-0.224585040999)*x[1]
        arg[3,3]=(-0.631704469212)*x[0]**o+(-0.495153785048)*x[0]+(-0.304420746672)*x[1]**o+(-0.254005222021)*x[1]
        arg[3,4]=(0.775631055421)*x[0]**o+(0.370095409961)*x[0]+(0.894627612994)*x[1]**o+(-0.329716708629)*x[1]
        ref[0,0]=(-0.958809754913)*x_ref[0]**o+(-0.844717049041)*x_ref[0]+(0.462564514427)*x_ref[1]**o+(-0.247678335594)*x_ref[1]
        ref[0,1]=(0.700700054189)*x_ref[0]**o+(-0.653724611457)*x_ref[0]+(0.52659177803)*x_ref[1]**o+(-0.387559446418)*x_ref[1]
        ref[0,2]=(-0.326124300134)*x_ref[0]**o+(0.115197897152)*x_ref[0]+(-0.876340641846)*x_ref[1]**o+(0.433060472581)*x_ref[1]
        ref[0,3]=(-0.101215736407)*x_ref[0]**o+(-0.483684510661)*x_ref[0]+(0.0418770907036)*x_ref[1]**o+(-0.845564414421)*x_ref[1]
        ref[0,4]=(-0.293094855718)*x_ref[0]**o+(0.710954930773)*x_ref[0]+(0.238671713869)*x_ref[1]**o+(0.576638343412)*x_ref[1]
        ref[1,0]=(0.529861280122)*x_ref[0]**o+(-0.259437904048)*x_ref[0]+(-0.620725288158)*x_ref[1]**o+(0.326787999382)*x_ref[1]
        ref[1,1]=(0.421245543012)*x_ref[0]**o+(-0.954139636698)*x_ref[0]+(0.0498199734926)*x_ref[1]**o+(0.24010238499)*x_ref[1]
        ref[1,2]=(0.333362362373)*x_ref[0]**o+(-0.999953175952)*x_ref[0]+(0.866193849246)*x_ref[1]**o+(0.764287410177)*x_ref[1]
        ref[1,3]=(0.470415957539)*x_ref[0]**o+(-0.133477966871)*x_ref[0]+(-0.0416970809832)*x_ref[1]**o+(-0.416352740444)*x_ref[1]
        ref[1,4]=(-0.178613018492)*x_ref[0]**o+(-0.967264985978)*x_ref[0]+(-0.297102423687)*x_ref[1]**o+(0.884184973413)*x_ref[1]
        ref[2,0]=(-0.29363453294)*x_ref[0]**o+(-0.0237360594862)*x_ref[0]+(-0.300776450747)*x_ref[1]**o+(-0.493932721521)*x_ref[1]
        ref[2,1]=(0.795045745699)*x_ref[0]**o+(0.984956851154)*x_ref[0]+(-0.724208405235)*x_ref[1]**o+(0.59029067239)*x_ref[1]
        ref[2,2]=(-0.803012112531)*x_ref[0]**o+(-0.831727785428)*x_ref[0]+(0.330295112332)*x_ref[1]**o+(0.0966317908147)*x_ref[1]
        ref[2,3]=(-0.652332429144)*x_ref[0]**o+(-0.38689457016)*x_ref[0]+(-0.391780825295)*x_ref[1]**o+(-0.444284790848)*x_ref[1]
        ref[2,4]=(0.416775161398)*x_ref[0]**o+(-0.16281006536)*x_ref[0]+(-0.742938932867)*x_ref[1]**o+(0.456486183714)*x_ref[1]
        ref[3,0]=(0.506731632086)*x_ref[0]**o+(-0.222468320035)*x_ref[0]+(-0.906398223556)*x_ref[1]**o+(-0.0250630448611)*x_ref[1]
        ref[3,1]=(-0.39207398783)*x_ref[0]**o+(-0.574739374252)*x_ref[0]+(0.262101584569)*x_ref[1]**o+(0.249587535087)*x_ref[1]
        ref[3,2]=(-0.252593913503)*x_ref[0]**o+(-0.864294052989)*x_ref[0]+(0.0110434296336)*x_ref[1]**o+(-0.224585040999)*x_ref[1]
        ref[3,3]=(-0.631704469212)*x_ref[0]**o+(-0.495153785048)*x_ref[0]+(-0.304420746672)*x_ref[1]**o+(-0.254005222021)*x_ref[1]
        ref[3,4]=(0.775631055421)*x_ref[0]**o+(0.370095409961)*x_ref[0]+(0.894627612994)*x_ref[1]**o+(-0.329716708629)*x_ref[1]
      else:
        arg[0,0]=(0.392450639899)*x[0]**o+(-0.685488230002)*x[0]+(-0.41545928428)*x[1]**o+(0.244144392379)*x[1]+(0.602642156528)*x[2]**o+(-0.348717378461)*x[2]
        arg[0,1]=(-0.341243165541)*x[0]**o+(-0.616870164204)*x[0]+(-0.510613066362)*x[1]**o+(-0.81387010212)*x[1]+(0.856239691119)*x[2]**o+(0.0744651854239)*x[2]
        arg[0,2]=(-0.765983413604)*x[0]**o+(0.474551394174)*x[0]+(-0.939524929765)*x[1]**o+(0.418297898065)*x[1]+(0.00589909539531)*x[2]**o+(0.0889012994897)*x[2]
        arg[0,3]=(-0.521217563352)*x[0]**o+(-0.256316459647)*x[0]+(-0.0265965098379)*x[1]**o+(-0.597562278966)*x[1]+(0.0389151661339)*x[2]**o+(-0.572109451352)*x[2]
        arg[0,4]=(-0.853801268491)*x[0]**o+(-0.691274812057)*x[0]+(-0.682435618581)*x[1]**o+(-0.974334375359)*x[1]+(-0.642333316794)*x[2]**o+(0.353816751252)*x[2]
        arg[1,0]=(-0.657545394322)*x[0]**o+(-0.334172338096)*x[0]+(0.016748279459)*x[1]**o+(-0.294858621125)*x[1]+(-0.49403880046)*x[2]**o+(-0.152416774822)*x[2]
        arg[1,1]=(-0.602731976042)*x[0]**o+(0.685046816545)*x[0]+(0.750001644716)*x[1]**o+(-0.202076972289)*x[1]+(-0.0232888078434)*x[2]**o+(-0.0774477649878)*x[2]
        arg[1,2]=(-0.707321785527)*x[0]**o+(0.618526464205)*x[0]+(-0.578548442468)*x[1]**o+(-0.411374744442)*x[1]+(-0.944277708787)*x[2]**o+(0.647856975707)*x[2]
        arg[1,3]=(-0.222503271022)*x[0]**o+(-0.929428092728)*x[0]+(-0.976553212424)*x[1]**o+(0.0376439617084)*x[1]+(0.0428078568512)*x[2]**o+(-0.441850944272)*x[2]
        arg[1,4]=(-0.862164898455)*x[0]**o+(-0.989900221991)*x[0]+(0.110610844567)*x[1]**o+(0.419384612236)*x[1]+(-0.00271948841262)*x[2]**o+(0.301635101629)*x[2]
        arg[2,0]=(-0.0950421881942)*x[0]**o+(-0.741293053909)*x[0]+(-0.461538879874)*x[1]**o+(-0.8199842553)*x[1]+(0.0590344916829)*x[2]**o+(0.352541518434)*x[2]
        arg[2,1]=(-0.734407995647)*x[0]**o+(-0.726167721992)*x[0]+(0.700521229916)*x[1]**o+(0.753555863428)*x[1]+(-0.891638139696)*x[2]**o+(-0.996550711213)*x[2]
        arg[2,2]=(0.777110352629)*x[0]**o+(0.309719201157)*x[0]+(0.026816547522)*x[1]**o+(0.33161010287)*x[1]+(0.257929504352)*x[2]**o+(0.71871314654)*x[2]
        arg[2,3]=(-0.0982522143825)*x[0]**o+(0.0532144836415)*x[0]+(0.155895442381)*x[1]**o+(-0.872222127033)*x[1]+(0.810538165596)*x[2]**o+(-0.917071363009)*x[2]
        arg[2,4]=(0.882674783877)*x[0]**o+(-0.246439463648)*x[0]+(-0.478821457401)*x[1]**o+(-0.629183697296)*x[1]+(0.833283185393)*x[2]**o+(-0.895796651184)*x[2]
        arg[3,0]=(0.949019461395)*x[0]**o+(0.437511668107)*x[0]+(0.46679355983)*x[1]**o+(0.375391913775)*x[1]+(0.992336073925)*x[2]**o+(0.784847856608)*x[2]
        arg[3,1]=(0.10730061928)*x[0]**o+(0.624678931808)*x[0]+(-0.105084917663)*x[1]**o+(-0.0650034223387)*x[1]+(-0.271922353868)*x[2]**o+(-0.182967296038)*x[2]
        arg[3,2]=(0.115856234055)*x[0]**o+(0.711912354922)*x[0]+(-0.317962972122)*x[1]**o+(0.577856092951)*x[1]+(-0.246833922867)*x[2]**o+(0.354116462218)*x[2]
        arg[3,3]=(0.40937236699)*x[0]**o+(0.639252679403)*x[0]+(0.540206943219)*x[1]**o+(0.287006139758)*x[1]+(0.123311433598)*x[2]**o+(-0.0300672780213)*x[2]
        arg[3,4]=(-0.190299257713)*x[0]**o+(-0.022055675261)*x[0]+(-0.640751994447)*x[1]**o+(-0.411801947317)*x[1]+(0.0675008125057)*x[2]**o+(-0.659424255203)*x[2]
        ref[0,0]=(0.392450639899)*x_ref[0]**o+(-0.685488230002)*x_ref[0]+(-0.41545928428)*x_ref[1]**o+(0.244144392379)*x_ref[1]+(0.602642156528)*x_ref[2]**o+(-0.348717378461)*x_ref[2]
        ref[0,1]=(-0.341243165541)*x_ref[0]**o+(-0.616870164204)*x_ref[0]+(-0.510613066362)*x_ref[1]**o+(-0.81387010212)*x_ref[1]+(0.856239691119)*x_ref[2]**o+(0.0744651854239)*x_ref[2]
        ref[0,2]=(-0.765983413604)*x_ref[0]**o+(0.474551394174)*x_ref[0]+(-0.939524929765)*x_ref[1]**o+(0.418297898065)*x_ref[1]+(0.00589909539531)*x_ref[2]**o+(0.0889012994897)*x_ref[2]
        ref[0,3]=(-0.521217563352)*x_ref[0]**o+(-0.256316459647)*x_ref[0]+(-0.0265965098379)*x_ref[1]**o+(-0.597562278966)*x_ref[1]+(0.0389151661339)*x_ref[2]**o+(-0.572109451352)*x_ref[2]
        ref[0,4]=(-0.853801268491)*x_ref[0]**o+(-0.691274812057)*x_ref[0]+(-0.682435618581)*x_ref[1]**o+(-0.974334375359)*x_ref[1]+(-0.642333316794)*x_ref[2]**o+(0.353816751252)*x_ref[2]
        ref[1,0]=(-0.657545394322)*x_ref[0]**o+(-0.334172338096)*x_ref[0]+(0.016748279459)*x_ref[1]**o+(-0.294858621125)*x_ref[1]+(-0.49403880046)*x_ref[2]**o+(-0.152416774822)*x_ref[2]
        ref[1,1]=(-0.602731976042)*x_ref[0]**o+(0.685046816545)*x_ref[0]+(0.750001644716)*x_ref[1]**o+(-0.202076972289)*x_ref[1]+(-0.0232888078434)*x_ref[2]**o+(-0.0774477649878)*x_ref[2]
        ref[1,2]=(-0.707321785527)*x_ref[0]**o+(0.618526464205)*x_ref[0]+(-0.578548442468)*x_ref[1]**o+(-0.411374744442)*x_ref[1]+(-0.944277708787)*x_ref[2]**o+(0.647856975707)*x_ref[2]
        ref[1,3]=(-0.222503271022)*x_ref[0]**o+(-0.929428092728)*x_ref[0]+(-0.976553212424)*x_ref[1]**o+(0.0376439617084)*x_ref[1]+(0.0428078568512)*x_ref[2]**o+(-0.441850944272)*x_ref[2]
        ref[1,4]=(-0.862164898455)*x_ref[0]**o+(-0.989900221991)*x_ref[0]+(0.110610844567)*x_ref[1]**o+(0.419384612236)*x_ref[1]+(-0.00271948841262)*x_ref[2]**o+(0.301635101629)*x_ref[2]
        ref[2,0]=(-0.0950421881942)*x_ref[0]**o+(-0.741293053909)*x_ref[0]+(-0.461538879874)*x_ref[1]**o+(-0.8199842553)*x_ref[1]+(0.0590344916829)*x_ref[2]**o+(0.352541518434)*x_ref[2]
        ref[2,1]=(-0.734407995647)*x_ref[0]**o+(-0.726167721992)*x_ref[0]+(0.700521229916)*x_ref[1]**o+(0.753555863428)*x_ref[1]+(-0.891638139696)*x_ref[2]**o+(-0.996550711213)*x_ref[2]
        ref[2,2]=(0.777110352629)*x_ref[0]**o+(0.309719201157)*x_ref[0]+(0.026816547522)*x_ref[1]**o+(0.33161010287)*x_ref[1]+(0.257929504352)*x_ref[2]**o+(0.71871314654)*x_ref[2]
        ref[2,3]=(-0.0982522143825)*x_ref[0]**o+(0.0532144836415)*x_ref[0]+(0.155895442381)*x_ref[1]**o+(-0.872222127033)*x_ref[1]+(0.810538165596)*x_ref[2]**o+(-0.917071363009)*x_ref[2]
        ref[2,4]=(0.882674783877)*x_ref[0]**o+(-0.246439463648)*x_ref[0]+(-0.478821457401)*x_ref[1]**o+(-0.629183697296)*x_ref[1]+(0.833283185393)*x_ref[2]**o+(-0.895796651184)*x_ref[2]
        ref[3,0]=(0.949019461395)*x_ref[0]**o+(0.437511668107)*x_ref[0]+(0.46679355983)*x_ref[1]**o+(0.375391913775)*x_ref[1]+(0.992336073925)*x_ref[2]**o+(0.784847856608)*x_ref[2]
        ref[3,1]=(0.10730061928)*x_ref[0]**o+(0.624678931808)*x_ref[0]+(-0.105084917663)*x_ref[1]**o+(-0.0650034223387)*x_ref[1]+(-0.271922353868)*x_ref[2]**o+(-0.182967296038)*x_ref[2]
        ref[3,2]=(0.115856234055)*x_ref[0]**o+(0.711912354922)*x_ref[0]+(-0.317962972122)*x_ref[1]**o+(0.577856092951)*x_ref[1]+(-0.246833922867)*x_ref[2]**o+(0.354116462218)*x_ref[2]
        ref[3,3]=(0.40937236699)*x_ref[0]**o+(0.639252679403)*x_ref[0]+(0.540206943219)*x_ref[1]**o+(0.287006139758)*x_ref[1]+(0.123311433598)*x_ref[2]**o+(-0.0300672780213)*x_ref[2]
        ref[3,4]=(-0.190299257713)*x_ref[0]**o+(-0.022055675261)*x_ref[0]+(-0.640751994447)*x_ref[1]**o+(-0.411801947317)*x_ref[1]+(0.0675008125057)*x_ref[2]**o+(-0.659424255203)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.697323110242)*x[0]**o+(-0.806494185218)*x[0]+(0.387914602484)*x[1]**o+(0.841835487965)*x[1]
        arg[0,0,1]=(-0.77670653908)*x[0]**o+(0.311710212706)*x[0]+(-0.0951864020181)*x[1]**o+(-0.14909139946)*x[1]
        arg[0,1,0]=(0.80933791543)*x[0]**o+(0.72441968276)*x[0]+(-0.446054850085)*x[1]**o+(0.419303534988)*x[1]
        arg[0,1,1]=(0.174464190017)*x[0]**o+(0.330609661878)*x[0]+(0.798365390058)*x[1]**o+(0.354899223817)*x[1]
        arg[1,0,0]=(0.36073672839)*x[0]**o+(0.831900846214)*x[0]+(-0.656354645044)*x[1]**o+(0.200645044646)*x[1]
        arg[1,0,1]=(0.817889801065)*x[0]**o+(0.34052889294)*x[0]+(-0.2416944924)*x[1]**o+(0.0922342148934)*x[1]
        arg[1,1,0]=(-0.962116509558)*x[0]**o+(-0.0785232077542)*x[0]+(-0.981399563356)*x[1]**o+(0.785509095614)*x[1]
        arg[1,1,1]=(-0.688444030508)*x[0]**o+(0.651617638972)*x[0]+(0.160584046401)*x[1]**o+(0.253258991923)*x[1]
        arg[2,0,0]=(0.0728970412574)*x[0]**o+(-0.647328240217)*x[0]+(0.517665494535)*x[1]**o+(-0.416383753573)*x[1]
        arg[2,0,1]=(0.148104739733)*x[0]**o+(-0.378371639837)*x[0]+(0.848608631336)*x[1]**o+(-0.881516167852)*x[1]
        arg[2,1,0]=(-0.506376676256)*x[0]**o+(0.234090340446)*x[0]+(-0.737828959566)*x[1]**o+(-0.74093763135)*x[1]
        arg[2,1,1]=(0.259605354135)*x[0]**o+(0.537424356513)*x[0]+(-0.405718119693)*x[1]**o+(0.120663712044)*x[1]
        arg[3,0,0]=(-0.609659975269)*x[0]**o+(-0.455923293874)*x[0]+(-0.00765676965971)*x[1]**o+(-0.710899349259)*x[1]
        arg[3,0,1]=(-0.731011612904)*x[0]**o+(-0.355060765358)*x[0]+(-0.127369372538)*x[1]**o+(0.694265371908)*x[1]
        arg[3,1,0]=(0.188766337132)*x[0]**o+(-0.194878304215)*x[0]+(-0.963356218109)*x[1]**o+(0.327053372293)*x[1]
        arg[3,1,1]=(-0.566712779324)*x[0]**o+(-0.251473305071)*x[0]+(0.0181297146138)*x[1]**o+(0.257664697754)*x[1]
        arg[4,0,0]=(0.574100764569)*x[0]**o+(-0.717776999574)*x[0]+(0.880998124155)*x[1]**o+(-0.810615928848)*x[1]
        arg[4,0,1]=(0.42552502518)*x[0]**o+(0.456751341147)*x[0]+(-0.484998619005)*x[1]**o+(0.932151187696)*x[1]
        arg[4,1,0]=(0.56865980485)*x[0]**o+(0.593928872734)*x[0]+(-0.475639894836)*x[1]**o+(-0.990718402693)*x[1]
        arg[4,1,1]=(0.129586342261)*x[0]**o+(-0.408951830619)*x[0]+(0.132518475289)*x[1]**o+(0.708612820692)*x[1]
        arg[5,0,0]=(0.0708038739082)*x[0]**o+(0.524270558649)*x[0]+(0.368504533482)*x[1]**o+(0.935828810393)*x[1]
        arg[5,0,1]=(-0.383650306829)*x[0]**o+(-0.0439362943961)*x[0]+(0.80939275198)*x[1]**o+(-0.61335508023)*x[1]
        arg[5,1,0]=(-0.621571339247)*x[0]**o+(-0.185069293199)*x[0]+(0.368472011431)*x[1]**o+(-0.388752495932)*x[1]
        arg[5,1,1]=(-0.263310763174)*x[0]**o+(-0.694616630015)*x[0]+(0.271238274203)*x[1]**o+(-0.477861402347)*x[1]
        ref[0,0,0]=(-0.697323110242)*x_ref[0]**o+(-0.806494185218)*x_ref[0]+(0.387914602484)*x_ref[1]**o+(0.841835487965)*x_ref[1]
        ref[0,0,1]=(-0.77670653908)*x_ref[0]**o+(0.311710212706)*x_ref[0]+(-0.0951864020181)*x_ref[1]**o+(-0.14909139946)*x_ref[1]
        ref[0,1,0]=(0.80933791543)*x_ref[0]**o+(0.72441968276)*x_ref[0]+(-0.446054850085)*x_ref[1]**o+(0.419303534988)*x_ref[1]
        ref[0,1,1]=(0.174464190017)*x_ref[0]**o+(0.330609661878)*x_ref[0]+(0.798365390058)*x_ref[1]**o+(0.354899223817)*x_ref[1]
        ref[1,0,0]=(0.36073672839)*x_ref[0]**o+(0.831900846214)*x_ref[0]+(-0.656354645044)*x_ref[1]**o+(0.200645044646)*x_ref[1]
        ref[1,0,1]=(0.817889801065)*x_ref[0]**o+(0.34052889294)*x_ref[0]+(-0.2416944924)*x_ref[1]**o+(0.0922342148934)*x_ref[1]
        ref[1,1,0]=(-0.962116509558)*x_ref[0]**o+(-0.0785232077542)*x_ref[0]+(-0.981399563356)*x_ref[1]**o+(0.785509095614)*x_ref[1]
        ref[1,1,1]=(-0.688444030508)*x_ref[0]**o+(0.651617638972)*x_ref[0]+(0.160584046401)*x_ref[1]**o+(0.253258991923)*x_ref[1]
        ref[2,0,0]=(0.0728970412574)*x_ref[0]**o+(-0.647328240217)*x_ref[0]+(0.517665494535)*x_ref[1]**o+(-0.416383753573)*x_ref[1]
        ref[2,0,1]=(0.148104739733)*x_ref[0]**o+(-0.378371639837)*x_ref[0]+(0.848608631336)*x_ref[1]**o+(-0.881516167852)*x_ref[1]
        ref[2,1,0]=(-0.506376676256)*x_ref[0]**o+(0.234090340446)*x_ref[0]+(-0.737828959566)*x_ref[1]**o+(-0.74093763135)*x_ref[1]
        ref[2,1,1]=(0.259605354135)*x_ref[0]**o+(0.537424356513)*x_ref[0]+(-0.405718119693)*x_ref[1]**o+(0.120663712044)*x_ref[1]
        ref[3,0,0]=(-0.609659975269)*x_ref[0]**o+(-0.455923293874)*x_ref[0]+(-0.00765676965971)*x_ref[1]**o+(-0.710899349259)*x_ref[1]
        ref[3,0,1]=(-0.731011612904)*x_ref[0]**o+(-0.355060765358)*x_ref[0]+(-0.127369372538)*x_ref[1]**o+(0.694265371908)*x_ref[1]
        ref[3,1,0]=(0.188766337132)*x_ref[0]**o+(-0.194878304215)*x_ref[0]+(-0.963356218109)*x_ref[1]**o+(0.327053372293)*x_ref[1]
        ref[3,1,1]=(-0.566712779324)*x_ref[0]**o+(-0.251473305071)*x_ref[0]+(0.0181297146138)*x_ref[1]**o+(0.257664697754)*x_ref[1]
        ref[4,0,0]=(0.574100764569)*x_ref[0]**o+(-0.717776999574)*x_ref[0]+(0.880998124155)*x_ref[1]**o+(-0.810615928848)*x_ref[1]
        ref[4,0,1]=(0.42552502518)*x_ref[0]**o+(0.456751341147)*x_ref[0]+(-0.484998619005)*x_ref[1]**o+(0.932151187696)*x_ref[1]
        ref[4,1,0]=(0.56865980485)*x_ref[0]**o+(0.593928872734)*x_ref[0]+(-0.475639894836)*x_ref[1]**o+(-0.990718402693)*x_ref[1]
        ref[4,1,1]=(0.129586342261)*x_ref[0]**o+(-0.408951830619)*x_ref[0]+(0.132518475289)*x_ref[1]**o+(0.708612820692)*x_ref[1]
        ref[5,0,0]=(0.0708038739082)*x_ref[0]**o+(0.524270558649)*x_ref[0]+(0.368504533482)*x_ref[1]**o+(0.935828810393)*x_ref[1]
        ref[5,0,1]=(-0.383650306829)*x_ref[0]**o+(-0.0439362943961)*x_ref[0]+(0.80939275198)*x_ref[1]**o+(-0.61335508023)*x_ref[1]
        ref[5,1,0]=(-0.621571339247)*x_ref[0]**o+(-0.185069293199)*x_ref[0]+(0.368472011431)*x_ref[1]**o+(-0.388752495932)*x_ref[1]
        ref[5,1,1]=(-0.263310763174)*x_ref[0]**o+(-0.694616630015)*x_ref[0]+(0.271238274203)*x_ref[1]**o+(-0.477861402347)*x_ref[1]
      else:
        arg[0,0,0]=(0.287534645778)*x[0]**o+(0.94865229379)*x[0]+(0.655299232864)*x[1]**o+(0.350034924637)*x[1]+(0.556341391183)*x[2]**o+(0.840643265765)*x[2]
        arg[0,0,1]=(-0.207213781549)*x[0]**o+(-0.0394497047797)*x[0]+(-0.0236016308575)*x[1]**o+(0.344122299199)*x[1]+(0.389588325996)*x[2]**o+(0.941980468657)*x[2]
        arg[0,1,0]=(0.00980412235862)*x[0]**o+(0.74862549447)*x[0]+(0.630165284383)*x[1]**o+(0.234711268989)*x[1]+(-0.22509981272)*x[2]**o+(-0.0153888168649)*x[2]
        arg[0,1,1]=(-0.87309186262)*x[0]**o+(0.833526923917)*x[0]+(0.78785479415)*x[1]**o+(0.823967259171)*x[1]+(0.320244316109)*x[2]**o+(-0.735225854417)*x[2]
        arg[1,0,0]=(-0.0892232551051)*x[0]**o+(-0.346728833563)*x[0]+(0.422774634671)*x[1]**o+(0.793222381957)*x[1]+(-0.670476677006)*x[2]**o+(0.896884540116)*x[2]
        arg[1,0,1]=(-0.0023382136345)*x[0]**o+(0.73058281545)*x[0]+(0.11828880374)*x[1]**o+(-0.691839091619)*x[1]+(-0.39086463392)*x[2]**o+(0.66200209782)*x[2]
        arg[1,1,0]=(-0.924238055558)*x[0]**o+(-0.334208271705)*x[0]+(-0.57106045555)*x[1]**o+(-0.765650999193)*x[1]+(0.0404533468054)*x[2]**o+(-0.225975260777)*x[2]
        arg[1,1,1]=(-0.351110999108)*x[0]**o+(0.626166346251)*x[0]+(0.385737198467)*x[1]**o+(0.577561466908)*x[1]+(0.708325217242)*x[2]**o+(0.193727923904)*x[2]
        arg[2,0,0]=(0.278942080997)*x[0]**o+(-0.569761653572)*x[0]+(-0.624879270918)*x[1]**o+(-0.0151754112896)*x[1]+(0.411399083825)*x[2]**o+(-0.846669568641)*x[2]
        arg[2,0,1]=(-0.543586963879)*x[0]**o+(-0.383959956625)*x[0]+(0.184375944938)*x[1]**o+(-0.104178233562)*x[1]+(-0.223101894662)*x[2]**o+(-0.17221835595)*x[2]
        arg[2,1,0]=(0.317103921207)*x[0]**o+(-0.919542461354)*x[0]+(0.56666554454)*x[1]**o+(0.895772298847)*x[1]+(0.714550707822)*x[2]**o+(0.634702089556)*x[2]
        arg[2,1,1]=(0.986576254164)*x[0]**o+(0.544493787633)*x[0]+(0.964620027203)*x[1]**o+(0.259991384678)*x[1]+(-0.577361835739)*x[2]**o+(0.651593456822)*x[2]
        arg[3,0,0]=(0.362841472614)*x[0]**o+(-0.521771228041)*x[0]+(-0.594530755517)*x[1]**o+(0.700647681241)*x[1]+(-0.131683781311)*x[2]**o+(0.636310189834)*x[2]
        arg[3,0,1]=(0.885891862321)*x[0]**o+(0.397438320764)*x[0]+(0.711380452937)*x[1]**o+(0.962571325144)*x[1]+(-0.300114113339)*x[2]**o+(0.34700388868)*x[2]
        arg[3,1,0]=(0.52140726408)*x[0]**o+(-0.0921295295522)*x[0]+(-0.823076091446)*x[1]**o+(0.477401876125)*x[1]+(-0.973057905067)*x[2]**o+(-0.975526794128)*x[2]
        arg[3,1,1]=(-0.241503887113)*x[0]**o+(0.133988088285)*x[0]+(0.243026393121)*x[1]**o+(0.0848716779473)*x[1]+(0.140349614791)*x[2]**o+(-0.217182025695)*x[2]
        arg[4,0,0]=(0.680080194053)*x[0]**o+(0.512009633844)*x[0]+(-0.640214066363)*x[1]**o+(0.571800633214)*x[1]+(-0.749395617545)*x[2]**o+(0.448487668786)*x[2]
        arg[4,0,1]=(-0.792959916349)*x[0]**o+(-0.433770516776)*x[0]+(-0.380980758215)*x[1]**o+(-0.831154769101)*x[1]+(0.277551460651)*x[2]**o+(0.626476984233)*x[2]
        arg[4,1,0]=(-0.285102566546)*x[0]**o+(0.745636275253)*x[0]+(0.4999427836)*x[1]**o+(0.561343302884)*x[1]+(0.796220302361)*x[2]**o+(-0.830020520143)*x[2]
        arg[4,1,1]=(0.237420507314)*x[0]**o+(0.327181858092)*x[0]+(-0.895743235593)*x[1]**o+(-0.496964477443)*x[1]+(0.57336428204)*x[2]**o+(-0.154267071287)*x[2]
        arg[5,0,0]=(0.0401925997607)*x[0]**o+(-0.0155957424748)*x[0]+(-0.658363451239)*x[1]**o+(-0.175223211879)*x[1]+(-0.743729768327)*x[2]**o+(0.391531327128)*x[2]
        arg[5,0,1]=(0.836239801824)*x[0]**o+(-0.72288085675)*x[0]+(0.177981042887)*x[1]**o+(0.354505260098)*x[1]+(-0.180520504267)*x[2]**o+(-0.380306239974)*x[2]
        arg[5,1,0]=(-0.65536853024)*x[0]**o+(-0.00573854242476)*x[0]+(-0.261460514314)*x[1]**o+(-0.289771405693)*x[1]+(0.750775397523)*x[2]**o+(-0.333215608562)*x[2]
        arg[5,1,1]=(-0.579532884024)*x[0]**o+(-0.208677373724)*x[0]+(-0.379612024024)*x[1]**o+(-0.0565813995114)*x[1]+(-0.867913562592)*x[2]**o+(-0.254654298293)*x[2]
        ref[0,0,0]=(0.287534645778)*x_ref[0]**o+(0.94865229379)*x_ref[0]+(0.655299232864)*x_ref[1]**o+(0.350034924637)*x_ref[1]+(0.556341391183)*x_ref[2]**o+(0.840643265765)*x_ref[2]
        ref[0,0,1]=(-0.207213781549)*x_ref[0]**o+(-0.0394497047797)*x_ref[0]+(-0.0236016308575)*x_ref[1]**o+(0.344122299199)*x_ref[1]+(0.389588325996)*x_ref[2]**o+(0.941980468657)*x_ref[2]
        ref[0,1,0]=(0.00980412235862)*x_ref[0]**o+(0.74862549447)*x_ref[0]+(0.630165284383)*x_ref[1]**o+(0.234711268989)*x_ref[1]+(-0.22509981272)*x_ref[2]**o+(-0.0153888168649)*x_ref[2]
        ref[0,1,1]=(-0.87309186262)*x_ref[0]**o+(0.833526923917)*x_ref[0]+(0.78785479415)*x_ref[1]**o+(0.823967259171)*x_ref[1]+(0.320244316109)*x_ref[2]**o+(-0.735225854417)*x_ref[2]
        ref[1,0,0]=(-0.0892232551051)*x_ref[0]**o+(-0.346728833563)*x_ref[0]+(0.422774634671)*x_ref[1]**o+(0.793222381957)*x_ref[1]+(-0.670476677006)*x_ref[2]**o+(0.896884540116)*x_ref[2]
        ref[1,0,1]=(-0.0023382136345)*x_ref[0]**o+(0.73058281545)*x_ref[0]+(0.11828880374)*x_ref[1]**o+(-0.691839091619)*x_ref[1]+(-0.39086463392)*x_ref[2]**o+(0.66200209782)*x_ref[2]
        ref[1,1,0]=(-0.924238055558)*x_ref[0]**o+(-0.334208271705)*x_ref[0]+(-0.57106045555)*x_ref[1]**o+(-0.765650999193)*x_ref[1]+(0.0404533468054)*x_ref[2]**o+(-0.225975260777)*x_ref[2]
        ref[1,1,1]=(-0.351110999108)*x_ref[0]**o+(0.626166346251)*x_ref[0]+(0.385737198467)*x_ref[1]**o+(0.577561466908)*x_ref[1]+(0.708325217242)*x_ref[2]**o+(0.193727923904)*x_ref[2]
        ref[2,0,0]=(0.278942080997)*x_ref[0]**o+(-0.569761653572)*x_ref[0]+(-0.624879270918)*x_ref[1]**o+(-0.0151754112896)*x_ref[1]+(0.411399083825)*x_ref[2]**o+(-0.846669568641)*x_ref[2]
        ref[2,0,1]=(-0.543586963879)*x_ref[0]**o+(-0.383959956625)*x_ref[0]+(0.184375944938)*x_ref[1]**o+(-0.104178233562)*x_ref[1]+(-0.223101894662)*x_ref[2]**o+(-0.17221835595)*x_ref[2]
        ref[2,1,0]=(0.317103921207)*x_ref[0]**o+(-0.919542461354)*x_ref[0]+(0.56666554454)*x_ref[1]**o+(0.895772298847)*x_ref[1]+(0.714550707822)*x_ref[2]**o+(0.634702089556)*x_ref[2]
        ref[2,1,1]=(0.986576254164)*x_ref[0]**o+(0.544493787633)*x_ref[0]+(0.964620027203)*x_ref[1]**o+(0.259991384678)*x_ref[1]+(-0.577361835739)*x_ref[2]**o+(0.651593456822)*x_ref[2]
        ref[3,0,0]=(0.362841472614)*x_ref[0]**o+(-0.521771228041)*x_ref[0]+(-0.594530755517)*x_ref[1]**o+(0.700647681241)*x_ref[1]+(-0.131683781311)*x_ref[2]**o+(0.636310189834)*x_ref[2]
        ref[3,0,1]=(0.885891862321)*x_ref[0]**o+(0.397438320764)*x_ref[0]+(0.711380452937)*x_ref[1]**o+(0.962571325144)*x_ref[1]+(-0.300114113339)*x_ref[2]**o+(0.34700388868)*x_ref[2]
        ref[3,1,0]=(0.52140726408)*x_ref[0]**o+(-0.0921295295522)*x_ref[0]+(-0.823076091446)*x_ref[1]**o+(0.477401876125)*x_ref[1]+(-0.973057905067)*x_ref[2]**o+(-0.975526794128)*x_ref[2]
        ref[3,1,1]=(-0.241503887113)*x_ref[0]**o+(0.133988088285)*x_ref[0]+(0.243026393121)*x_ref[1]**o+(0.0848716779473)*x_ref[1]+(0.140349614791)*x_ref[2]**o+(-0.217182025695)*x_ref[2]
        ref[4,0,0]=(0.680080194053)*x_ref[0]**o+(0.512009633844)*x_ref[0]+(-0.640214066363)*x_ref[1]**o+(0.571800633214)*x_ref[1]+(-0.749395617545)*x_ref[2]**o+(0.448487668786)*x_ref[2]
        ref[4,0,1]=(-0.792959916349)*x_ref[0]**o+(-0.433770516776)*x_ref[0]+(-0.380980758215)*x_ref[1]**o+(-0.831154769101)*x_ref[1]+(0.277551460651)*x_ref[2]**o+(0.626476984233)*x_ref[2]
        ref[4,1,0]=(-0.285102566546)*x_ref[0]**o+(0.745636275253)*x_ref[0]+(0.4999427836)*x_ref[1]**o+(0.561343302884)*x_ref[1]+(0.796220302361)*x_ref[2]**o+(-0.830020520143)*x_ref[2]
        ref[4,1,1]=(0.237420507314)*x_ref[0]**o+(0.327181858092)*x_ref[0]+(-0.895743235593)*x_ref[1]**o+(-0.496964477443)*x_ref[1]+(0.57336428204)*x_ref[2]**o+(-0.154267071287)*x_ref[2]
        ref[5,0,0]=(0.0401925997607)*x_ref[0]**o+(-0.0155957424748)*x_ref[0]+(-0.658363451239)*x_ref[1]**o+(-0.175223211879)*x_ref[1]+(-0.743729768327)*x_ref[2]**o+(0.391531327128)*x_ref[2]
        ref[5,0,1]=(0.836239801824)*x_ref[0]**o+(-0.72288085675)*x_ref[0]+(0.177981042887)*x_ref[1]**o+(0.354505260098)*x_ref[1]+(-0.180520504267)*x_ref[2]**o+(-0.380306239974)*x_ref[2]
        ref[5,1,0]=(-0.65536853024)*x_ref[0]**o+(-0.00573854242476)*x_ref[0]+(-0.261460514314)*x_ref[1]**o+(-0.289771405693)*x_ref[1]+(0.750775397523)*x_ref[2]**o+(-0.333215608562)*x_ref[2]
        ref[5,1,1]=(-0.579532884024)*x_ref[0]**o+(-0.208677373724)*x_ref[0]+(-0.379612024024)*x_ref[1]**o+(-0.0565813995114)*x_ref[1]+(-0.867913562592)*x_ref[2]**o+(-0.254654298293)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.274472293956)*x[0]**o+(-0.293347130929)*x[0]+(0.392586448254)*x[1]**o+(-0.403066854054)*x[1]
        arg[0,0,0,1]=(0.385933192907)*x[0]**o+(0.484240065867)*x[0]+(0.144565631551)*x[1]**o+(0.882508492739)*x[1]
        arg[0,0,1,0]=(0.993368905137)*x[0]**o+(0.340349919343)*x[0]+(-0.956806746524)*x[1]**o+(0.990607958001)*x[1]
        arg[0,0,1,1]=(-0.88610250102)*x[0]**o+(-0.543705587259)*x[0]+(0.170798164924)*x[1]**o+(0.685331195754)*x[1]
        arg[0,0,2,0]=(-0.922211611143)*x[0]**o+(-0.103640037893)*x[0]+(0.401354703434)*x[1]**o+(-0.117826261635)*x[1]
        arg[0,0,2,1]=(-0.391601387132)*x[0]**o+(-0.472753344804)*x[0]+(-0.473433838993)*x[1]**o+(-0.173501635989)*x[1]
        arg[0,1,0,0]=(0.747353274916)*x[0]**o+(-0.960856612086)*x[0]+(-0.54809012978)*x[1]**o+(0.862023893332)*x[1]
        arg[0,1,0,1]=(-0.256359001149)*x[0]**o+(-0.868975044191)*x[0]+(0.494314269744)*x[1]**o+(0.949957481172)*x[1]
        arg[0,1,1,0]=(-0.0256658750601)*x[0]**o+(0.145499104203)*x[0]+(-0.68669791629)*x[1]**o+(0.447326207295)*x[1]
        arg[0,1,1,1]=(0.150617284467)*x[0]**o+(-0.350046088196)*x[0]+(-0.367671302053)*x[1]**o+(0.633220122068)*x[1]
        arg[0,1,2,0]=(-0.714145362271)*x[0]**o+(-0.957661490758)*x[0]+(0.386253770747)*x[1]**o+(0.982634262634)*x[1]
        arg[0,1,2,1]=(0.707565096824)*x[0]**o+(0.281758058573)*x[0]+(-0.407791315587)*x[1]**o+(0.86069036983)*x[1]
        arg[0,2,0,0]=(-0.453860220598)*x[0]**o+(0.944457388061)*x[0]+(-0.903818960162)*x[1]**o+(-0.799423205297)*x[1]
        arg[0,2,0,1]=(-0.621257500941)*x[0]**o+(0.748963737937)*x[0]+(-0.12660934458)*x[1]**o+(-0.362830639592)*x[1]
        arg[0,2,1,0]=(-0.702485820037)*x[0]**o+(0.355584898593)*x[0]+(0.657127357751)*x[1]**o+(0.535533390762)*x[1]
        arg[0,2,1,1]=(0.795733125008)*x[0]**o+(0.983607327249)*x[0]+(-0.463818780644)*x[1]**o+(-0.681831588244)*x[1]
        arg[0,2,2,0]=(0.335783204138)*x[0]**o+(-0.712225782679)*x[0]+(0.853426644096)*x[1]**o+(-0.143845263686)*x[1]
        arg[0,2,2,1]=(0.389724375144)*x[0]**o+(0.715176227705)*x[0]+(-0.660690893888)*x[1]**o+(0.428289724709)*x[1]
        arg[0,3,0,0]=(-0.363836203724)*x[0]**o+(-0.257775427457)*x[0]+(0.407904889098)*x[1]**o+(-0.731273814284)*x[1]
        arg[0,3,0,1]=(0.402038262494)*x[0]**o+(0.33289103638)*x[0]+(0.444957562329)*x[1]**o+(0.0657971935664)*x[1]
        arg[0,3,1,0]=(0.979673814602)*x[0]**o+(-0.133405332848)*x[0]+(0.23884689304)*x[1]**o+(0.728464997681)*x[1]
        arg[0,3,1,1]=(0.724603429246)*x[0]**o+(-0.391040898144)*x[0]+(-0.966816499003)*x[1]**o+(0.0984302578905)*x[1]
        arg[0,3,2,0]=(-0.986139645754)*x[0]**o+(0.60807351165)*x[0]+(0.799807413231)*x[1]**o+(0.12091479711)*x[1]
        arg[0,3,2,1]=(0.387453999377)*x[0]**o+(0.212725006628)*x[0]+(-0.120374039139)*x[1]**o+(-0.262304757951)*x[1]
        arg[0,4,0,0]=(-0.260545986276)*x[0]**o+(0.251085312932)*x[0]+(0.738082708518)*x[1]**o+(-0.0804883797938)*x[1]
        arg[0,4,0,1]=(-0.190701356138)*x[0]**o+(-0.0205295935522)*x[0]+(0.904261681819)*x[1]**o+(-0.859647709957)*x[1]
        arg[0,4,1,0]=(-0.945357492849)*x[0]**o+(-0.85298463389)*x[0]+(0.688964336673)*x[1]**o+(-0.672881696932)*x[1]
        arg[0,4,1,1]=(0.877720258045)*x[0]**o+(-0.361228026805)*x[0]+(-0.326626527006)*x[1]**o+(-0.274100421232)*x[1]
        arg[0,4,2,0]=(0.202358729084)*x[0]**o+(-0.997571481837)*x[0]+(0.431193711862)*x[1]**o+(0.513561123915)*x[1]
        arg[0,4,2,1]=(-0.136602954592)*x[0]**o+(-0.40864934437)*x[0]+(0.675945574351)*x[1]**o+(0.728757793536)*x[1]
        arg[1,0,0,0]=(0.835464261851)*x[0]**o+(-0.131447020601)*x[0]+(0.0810352521446)*x[1]**o+(0.664744750899)*x[1]
        arg[1,0,0,1]=(0.239716928689)*x[0]**o+(-0.192258999531)*x[0]+(0.615877054036)*x[1]**o+(0.835274774437)*x[1]
        arg[1,0,1,0]=(0.509342803767)*x[0]**o+(-0.633944481882)*x[0]+(0.262098006086)*x[1]**o+(0.159902070799)*x[1]
        arg[1,0,1,1]=(0.894432440588)*x[0]**o+(-0.167778431319)*x[0]+(-0.984183148477)*x[1]**o+(-0.554981658632)*x[1]
        arg[1,0,2,0]=(-0.762472607231)*x[0]**o+(0.405550139256)*x[0]+(0.4051290641)*x[1]**o+(-0.27180186672)*x[1]
        arg[1,0,2,1]=(0.663869273816)*x[0]**o+(0.867453974628)*x[0]+(-0.279659969132)*x[1]**o+(-0.370539101204)*x[1]
        arg[1,1,0,0]=(-0.969326778164)*x[0]**o+(0.965471349497)*x[0]+(-0.529983232946)*x[1]**o+(0.714091610977)*x[1]
        arg[1,1,0,1]=(-0.840360667093)*x[0]**o+(-0.399319808256)*x[0]+(0.568769915582)*x[1]**o+(0.872427494442)*x[1]
        arg[1,1,1,0]=(0.46099116941)*x[0]**o+(-0.986008874142)*x[0]+(-0.53999023556)*x[1]**o+(-0.549695567855)*x[1]
        arg[1,1,1,1]=(0.485944918559)*x[0]**o+(0.189773828651)*x[0]+(-0.611654980493)*x[1]**o+(-0.337071911657)*x[1]
        arg[1,1,2,0]=(0.191663603248)*x[0]**o+(0.0906374965004)*x[0]+(-0.271051598251)*x[1]**o+(0.880507788825)*x[1]
        arg[1,1,2,1]=(-0.439514997001)*x[0]**o+(0.701230426224)*x[0]+(-0.193231641029)*x[1]**o+(0.950584054332)*x[1]
        arg[1,2,0,0]=(0.439456565259)*x[0]**o+(-0.25664171484)*x[0]+(-0.653059819329)*x[1]**o+(-0.680544747777)*x[1]
        arg[1,2,0,1]=(0.34890454443)*x[0]**o+(-0.69851200963)*x[0]+(-0.00159381688549)*x[1]**o+(-0.612359675499)*x[1]
        arg[1,2,1,0]=(0.132759671375)*x[0]**o+(-0.235126716401)*x[0]+(0.652068885826)*x[1]**o+(0.749698645371)*x[1]
        arg[1,2,1,1]=(0.934980403627)*x[0]**o+(0.321409273437)*x[0]+(-0.605909929553)*x[1]**o+(0.135554927418)*x[1]
        arg[1,2,2,0]=(-0.68327709329)*x[0]**o+(-0.0381292756005)*x[0]+(0.710953191114)*x[1]**o+(-0.344522256689)*x[1]
        arg[1,2,2,1]=(-0.813976235652)*x[0]**o+(0.613879708123)*x[0]+(0.451599277293)*x[1]**o+(-0.6134026375)*x[1]
        arg[1,3,0,0]=(0.595310128627)*x[0]**o+(-0.247621573131)*x[0]+(0.165607783104)*x[1]**o+(0.578132180805)*x[1]
        arg[1,3,0,1]=(0.848827238882)*x[0]**o+(0.173154067999)*x[0]+(0.340662419437)*x[1]**o+(-0.389656447251)*x[1]
        arg[1,3,1,0]=(0.250888415228)*x[0]**o+(-0.278203136198)*x[0]+(-0.693211186734)*x[1]**o+(0.613812194735)*x[1]
        arg[1,3,1,1]=(0.634533003162)*x[0]**o+(-0.0924665762471)*x[0]+(0.749575894384)*x[1]**o+(-0.737205017415)*x[1]
        arg[1,3,2,0]=(0.607750966973)*x[0]**o+(0.905016928404)*x[0]+(-0.171049542163)*x[1]**o+(-0.20526244601)*x[1]
        arg[1,3,2,1]=(0.723946025405)*x[0]**o+(0.0157576743987)*x[0]+(0.190875428312)*x[1]**o+(0.407925899509)*x[1]
        arg[1,4,0,0]=(-0.561582695803)*x[0]**o+(-0.350262461371)*x[0]+(0.57610409872)*x[1]**o+(0.881780963877)*x[1]
        arg[1,4,0,1]=(0.077066498695)*x[0]**o+(0.515597802516)*x[0]+(0.332011413016)*x[1]**o+(0.215491334787)*x[1]
        arg[1,4,1,0]=(0.33858727737)*x[0]**o+(-0.573231800768)*x[0]+(-0.559915809132)*x[1]**o+(0.0240869142176)*x[1]
        arg[1,4,1,1]=(-0.684581727952)*x[0]**o+(0.365541904201)*x[0]+(0.204252397376)*x[1]**o+(0.149300766054)*x[1]
        arg[1,4,2,0]=(-0.574225088383)*x[0]**o+(0.0286339776067)*x[0]+(-0.358732122607)*x[1]**o+(-0.416932738957)*x[1]
        arg[1,4,2,1]=(-0.308460184932)*x[0]**o+(0.389214987836)*x[0]+(-0.358303257031)*x[1]**o+(0.348109483688)*x[1]
        arg[2,0,0,0]=(-0.811318668804)*x[0]**o+(-0.298028116982)*x[0]+(0.679871523483)*x[1]**o+(0.958042925287)*x[1]
        arg[2,0,0,1]=(-0.616521818338)*x[0]**o+(-0.0817124854496)*x[0]+(0.403257366893)*x[1]**o+(0.740689710435)*x[1]
        arg[2,0,1,0]=(0.271046179127)*x[0]**o+(0.814928093947)*x[0]+(0.423722781579)*x[1]**o+(0.925333209811)*x[1]
        arg[2,0,1,1]=(0.0596651806816)*x[0]**o+(0.493854605359)*x[0]+(0.539998537439)*x[1]**o+(-0.910865639375)*x[1]
        arg[2,0,2,0]=(0.503267602327)*x[0]**o+(-0.270966471743)*x[0]+(-0.342063366676)*x[1]**o+(-0.975700224705)*x[1]
        arg[2,0,2,1]=(-0.0362358939195)*x[0]**o+(-0.766033869339)*x[0]+(0.891365100088)*x[1]**o+(0.815025898929)*x[1]
        arg[2,1,0,0]=(-0.667240459984)*x[0]**o+(0.8659076701)*x[0]+(0.215089397562)*x[1]**o+(0.464769182307)*x[1]
        arg[2,1,0,1]=(-0.298625221931)*x[0]**o+(-0.504329743472)*x[0]+(0.412472510208)*x[1]**o+(0.241693152889)*x[1]
        arg[2,1,1,0]=(-0.670730179133)*x[0]**o+(0.0201070113832)*x[0]+(-0.572725898919)*x[1]**o+(0.665609259014)*x[1]
        arg[2,1,1,1]=(-0.319080984089)*x[0]**o+(0.0299423045224)*x[0]+(0.0152311810613)*x[1]**o+(-0.784646189026)*x[1]
        arg[2,1,2,0]=(0.153308789716)*x[0]**o+(0.725808150613)*x[0]+(0.62876493338)*x[1]**o+(0.80275284252)*x[1]
        arg[2,1,2,1]=(0.804698870013)*x[0]**o+(-0.631078773061)*x[0]+(-0.015221975698)*x[1]**o+(-0.954169686746)*x[1]
        arg[2,2,0,0]=(0.169022797498)*x[0]**o+(0.733025791099)*x[0]+(0.321554996633)*x[1]**o+(0.292173365364)*x[1]
        arg[2,2,0,1]=(0.327345955109)*x[0]**o+(0.21757823896)*x[0]+(0.873027160446)*x[1]**o+(-0.250192861579)*x[1]
        arg[2,2,1,0]=(0.394387190832)*x[0]**o+(-0.613661883482)*x[0]+(0.806965291386)*x[1]**o+(0.244513244016)*x[1]
        arg[2,2,1,1]=(0.043913907536)*x[0]**o+(-0.303567806424)*x[0]+(0.547037093323)*x[1]**o+(-0.598687385328)*x[1]
        arg[2,2,2,0]=(0.875663154787)*x[0]**o+(-0.34984832772)*x[0]+(0.879143174396)*x[1]**o+(-0.193154051109)*x[1]
        arg[2,2,2,1]=(0.940311132918)*x[0]**o+(-0.358708743382)*x[0]+(0.117370945146)*x[1]**o+(-0.664208034757)*x[1]
        arg[2,3,0,0]=(0.0334525660588)*x[0]**o+(0.592805915255)*x[0]+(-0.278216157971)*x[1]**o+(0.894369967671)*x[1]
        arg[2,3,0,1]=(-0.966591106326)*x[0]**o+(-0.882160390768)*x[0]+(0.843982702732)*x[1]**o+(0.363868921656)*x[1]
        arg[2,3,1,0]=(-0.246392342771)*x[0]**o+(-0.934459096242)*x[0]+(-0.472771368537)*x[1]**o+(0.148854304706)*x[1]
        arg[2,3,1,1]=(0.248218424051)*x[0]**o+(0.0980396576111)*x[0]+(0.413322166489)*x[1]**o+(0.429747261346)*x[1]
        arg[2,3,2,0]=(0.742987565733)*x[0]**o+(-0.213736450136)*x[0]+(0.996975741314)*x[1]**o+(0.960709958665)*x[1]
        arg[2,3,2,1]=(0.436149490977)*x[0]**o+(0.87632181197)*x[0]+(-0.881934669912)*x[1]**o+(-0.457520504175)*x[1]
        arg[2,4,0,0]=(-0.519014330693)*x[0]**o+(0.938619608407)*x[0]+(-0.847954563609)*x[1]**o+(0.791774379621)*x[1]
        arg[2,4,0,1]=(-0.599487299453)*x[0]**o+(0.977867555132)*x[0]+(0.693105630579)*x[1]**o+(-0.688827812942)*x[1]
        arg[2,4,1,0]=(0.955907964163)*x[0]**o+(0.0724056986172)*x[0]+(0.254773619772)*x[1]**o+(0.0428861146964)*x[1]
        arg[2,4,1,1]=(-0.422202101998)*x[0]**o+(0.803263223544)*x[0]+(0.456048140785)*x[1]**o+(0.0419133126183)*x[1]
        arg[2,4,2,0]=(-0.182308704781)*x[0]**o+(0.862282443488)*x[0]+(-0.844027007955)*x[1]**o+(0.586220526376)*x[1]
        arg[2,4,2,1]=(-0.801542666229)*x[0]**o+(0.903685299455)*x[0]+(0.428926544377)*x[1]**o+(-0.891220471549)*x[1]
        arg[3,0,0,0]=(0.602917787713)*x[0]**o+(-0.592239549743)*x[0]+(0.761830006718)*x[1]**o+(-0.784264167777)*x[1]
        arg[3,0,0,1]=(0.819962957732)*x[0]**o+(-0.455126790372)*x[0]+(-0.144898728525)*x[1]**o+(0.121566345339)*x[1]
        arg[3,0,1,0]=(0.570694941559)*x[0]**o+(-0.505678656386)*x[0]+(0.295536587355)*x[1]**o+(-0.852779737814)*x[1]
        arg[3,0,1,1]=(-0.226246858753)*x[0]**o+(-0.513900306228)*x[0]+(-0.389483591407)*x[1]**o+(-0.735105335539)*x[1]
        arg[3,0,2,0]=(0.941877610938)*x[0]**o+(0.93808330298)*x[0]+(0.189785162182)*x[1]**o+(0.546525366304)*x[1]
        arg[3,0,2,1]=(0.10639100485)*x[0]**o+(-0.773361007972)*x[0]+(-0.470531427579)*x[1]**o+(0.467359342808)*x[1]
        arg[3,1,0,0]=(0.857725599589)*x[0]**o+(0.0881746413132)*x[0]+(0.040713151152)*x[1]**o+(-0.52070729883)*x[1]
        arg[3,1,0,1]=(0.266807652146)*x[0]**o+(0.337496673892)*x[0]+(-0.562266297445)*x[1]**o+(0.326015195113)*x[1]
        arg[3,1,1,0]=(0.177275509465)*x[0]**o+(-0.647855761301)*x[0]+(-0.0963373631995)*x[1]**o+(-0.535256252994)*x[1]
        arg[3,1,1,1]=(0.430309295606)*x[0]**o+(-0.0894749557428)*x[0]+(0.503470724483)*x[1]**o+(-0.202361122248)*x[1]
        arg[3,1,2,0]=(0.283995014194)*x[0]**o+(-0.544419907194)*x[0]+(0.78478571712)*x[1]**o+(0.341540664147)*x[1]
        arg[3,1,2,1]=(-0.281042869936)*x[0]**o+(0.728771766492)*x[0]+(-0.298974388557)*x[1]**o+(-0.836789906985)*x[1]
        arg[3,2,0,0]=(-0.415323371561)*x[0]**o+(0.583971137371)*x[0]+(-0.265658679489)*x[1]**o+(-0.432447373809)*x[1]
        arg[3,2,0,1]=(0.690486298855)*x[0]**o+(0.552521632226)*x[0]+(0.237622739042)*x[1]**o+(-0.254513693924)*x[1]
        arg[3,2,1,0]=(0.453985687207)*x[0]**o+(-0.785385580962)*x[0]+(0.0190987841634)*x[1]**o+(-0.7923288861)*x[1]
        arg[3,2,1,1]=(-0.964082656379)*x[0]**o+(-0.806315937371)*x[0]+(0.271145420366)*x[1]**o+(-0.839014535945)*x[1]
        arg[3,2,2,0]=(-0.372062773997)*x[0]**o+(-0.069749560567)*x[0]+(-0.126310395383)*x[1]**o+(0.972025865285)*x[1]
        arg[3,2,2,1]=(-0.580353020086)*x[0]**o+(-0.152746821477)*x[0]+(0.297553660535)*x[1]**o+(0.444791568145)*x[1]
        arg[3,3,0,0]=(-0.361520029439)*x[0]**o+(-0.810391070499)*x[0]+(0.462123882642)*x[1]**o+(-0.974534227742)*x[1]
        arg[3,3,0,1]=(0.113370617837)*x[0]**o+(0.428257355435)*x[0]+(-0.18656623758)*x[1]**o+(0.206503327394)*x[1]
        arg[3,3,1,0]=(0.341328445689)*x[0]**o+(0.34002947054)*x[0]+(0.997677286362)*x[1]**o+(0.43300073225)*x[1]
        arg[3,3,1,1]=(-0.513912982823)*x[0]**o+(0.826573827064)*x[0]+(-0.701417864085)*x[1]**o+(0.567078883219)*x[1]
        arg[3,3,2,0]=(-0.325536866965)*x[0]**o+(0.924405767451)*x[0]+(-0.0481381954346)*x[1]**o+(0.912210617073)*x[1]
        arg[3,3,2,1]=(0.491532434302)*x[0]**o+(-0.584980848036)*x[0]+(-0.788317530858)*x[1]**o+(-0.900681893172)*x[1]
        arg[3,4,0,0]=(0.432249855148)*x[0]**o+(-0.673369312552)*x[0]+(0.244704845849)*x[1]**o+(-0.300995595367)*x[1]
        arg[3,4,0,1]=(0.383037211984)*x[0]**o+(0.299711400488)*x[0]+(0.870584255454)*x[1]**o+(-0.33571823329)*x[1]
        arg[3,4,1,0]=(0.737134180177)*x[0]**o+(-0.101307090628)*x[0]+(-0.970632061143)*x[1]**o+(-0.449042684137)*x[1]
        arg[3,4,1,1]=(0.00684250643632)*x[0]**o+(0.676964654022)*x[0]+(0.789069834858)*x[1]**o+(-0.829739508197)*x[1]
        arg[3,4,2,0]=(0.7279044944)*x[0]**o+(0.215711122038)*x[0]+(0.385631252632)*x[1]**o+(0.629390433529)*x[1]
        arg[3,4,2,1]=(-0.380750716629)*x[0]**o+(-0.413040692437)*x[0]+(0.26795244639)*x[1]**o+(0.492475504659)*x[1]
        ref[0,0,0,0]=(-0.274472293956)*x_ref[0]**o+(-0.293347130929)*x_ref[0]+(0.392586448254)*x_ref[1]**o+(-0.403066854054)*x_ref[1]
        ref[0,0,0,1]=(0.385933192907)*x_ref[0]**o+(0.484240065867)*x_ref[0]+(0.144565631551)*x_ref[1]**o+(0.882508492739)*x_ref[1]
        ref[0,0,1,0]=(0.993368905137)*x_ref[0]**o+(0.340349919343)*x_ref[0]+(-0.956806746524)*x_ref[1]**o+(0.990607958001)*x_ref[1]
        ref[0,0,1,1]=(-0.88610250102)*x_ref[0]**o+(-0.543705587259)*x_ref[0]+(0.170798164924)*x_ref[1]**o+(0.685331195754)*x_ref[1]
        ref[0,0,2,0]=(-0.922211611143)*x_ref[0]**o+(-0.103640037893)*x_ref[0]+(0.401354703434)*x_ref[1]**o+(-0.117826261635)*x_ref[1]
        ref[0,0,2,1]=(-0.391601387132)*x_ref[0]**o+(-0.472753344804)*x_ref[0]+(-0.473433838993)*x_ref[1]**o+(-0.173501635989)*x_ref[1]
        ref[0,1,0,0]=(0.747353274916)*x_ref[0]**o+(-0.960856612086)*x_ref[0]+(-0.54809012978)*x_ref[1]**o+(0.862023893332)*x_ref[1]
        ref[0,1,0,1]=(-0.256359001149)*x_ref[0]**o+(-0.868975044191)*x_ref[0]+(0.494314269744)*x_ref[1]**o+(0.949957481172)*x_ref[1]
        ref[0,1,1,0]=(-0.0256658750601)*x_ref[0]**o+(0.145499104203)*x_ref[0]+(-0.68669791629)*x_ref[1]**o+(0.447326207295)*x_ref[1]
        ref[0,1,1,1]=(0.150617284467)*x_ref[0]**o+(-0.350046088196)*x_ref[0]+(-0.367671302053)*x_ref[1]**o+(0.633220122068)*x_ref[1]
        ref[0,1,2,0]=(-0.714145362271)*x_ref[0]**o+(-0.957661490758)*x_ref[0]+(0.386253770747)*x_ref[1]**o+(0.982634262634)*x_ref[1]
        ref[0,1,2,1]=(0.707565096824)*x_ref[0]**o+(0.281758058573)*x_ref[0]+(-0.407791315587)*x_ref[1]**o+(0.86069036983)*x_ref[1]
        ref[0,2,0,0]=(-0.453860220598)*x_ref[0]**o+(0.944457388061)*x_ref[0]+(-0.903818960162)*x_ref[1]**o+(-0.799423205297)*x_ref[1]
        ref[0,2,0,1]=(-0.621257500941)*x_ref[0]**o+(0.748963737937)*x_ref[0]+(-0.12660934458)*x_ref[1]**o+(-0.362830639592)*x_ref[1]
        ref[0,2,1,0]=(-0.702485820037)*x_ref[0]**o+(0.355584898593)*x_ref[0]+(0.657127357751)*x_ref[1]**o+(0.535533390762)*x_ref[1]
        ref[0,2,1,1]=(0.795733125008)*x_ref[0]**o+(0.983607327249)*x_ref[0]+(-0.463818780644)*x_ref[1]**o+(-0.681831588244)*x_ref[1]
        ref[0,2,2,0]=(0.335783204138)*x_ref[0]**o+(-0.712225782679)*x_ref[0]+(0.853426644096)*x_ref[1]**o+(-0.143845263686)*x_ref[1]
        ref[0,2,2,1]=(0.389724375144)*x_ref[0]**o+(0.715176227705)*x_ref[0]+(-0.660690893888)*x_ref[1]**o+(0.428289724709)*x_ref[1]
        ref[0,3,0,0]=(-0.363836203724)*x_ref[0]**o+(-0.257775427457)*x_ref[0]+(0.407904889098)*x_ref[1]**o+(-0.731273814284)*x_ref[1]
        ref[0,3,0,1]=(0.402038262494)*x_ref[0]**o+(0.33289103638)*x_ref[0]+(0.444957562329)*x_ref[1]**o+(0.0657971935664)*x_ref[1]
        ref[0,3,1,0]=(0.979673814602)*x_ref[0]**o+(-0.133405332848)*x_ref[0]+(0.23884689304)*x_ref[1]**o+(0.728464997681)*x_ref[1]
        ref[0,3,1,1]=(0.724603429246)*x_ref[0]**o+(-0.391040898144)*x_ref[0]+(-0.966816499003)*x_ref[1]**o+(0.0984302578905)*x_ref[1]
        ref[0,3,2,0]=(-0.986139645754)*x_ref[0]**o+(0.60807351165)*x_ref[0]+(0.799807413231)*x_ref[1]**o+(0.12091479711)*x_ref[1]
        ref[0,3,2,1]=(0.387453999377)*x_ref[0]**o+(0.212725006628)*x_ref[0]+(-0.120374039139)*x_ref[1]**o+(-0.262304757951)*x_ref[1]
        ref[0,4,0,0]=(-0.260545986276)*x_ref[0]**o+(0.251085312932)*x_ref[0]+(0.738082708518)*x_ref[1]**o+(-0.0804883797938)*x_ref[1]
        ref[0,4,0,1]=(-0.190701356138)*x_ref[0]**o+(-0.0205295935522)*x_ref[0]+(0.904261681819)*x_ref[1]**o+(-0.859647709957)*x_ref[1]
        ref[0,4,1,0]=(-0.945357492849)*x_ref[0]**o+(-0.85298463389)*x_ref[0]+(0.688964336673)*x_ref[1]**o+(-0.672881696932)*x_ref[1]
        ref[0,4,1,1]=(0.877720258045)*x_ref[0]**o+(-0.361228026805)*x_ref[0]+(-0.326626527006)*x_ref[1]**o+(-0.274100421232)*x_ref[1]
        ref[0,4,2,0]=(0.202358729084)*x_ref[0]**o+(-0.997571481837)*x_ref[0]+(0.431193711862)*x_ref[1]**o+(0.513561123915)*x_ref[1]
        ref[0,4,2,1]=(-0.136602954592)*x_ref[0]**o+(-0.40864934437)*x_ref[0]+(0.675945574351)*x_ref[1]**o+(0.728757793536)*x_ref[1]
        ref[1,0,0,0]=(0.835464261851)*x_ref[0]**o+(-0.131447020601)*x_ref[0]+(0.0810352521446)*x_ref[1]**o+(0.664744750899)*x_ref[1]
        ref[1,0,0,1]=(0.239716928689)*x_ref[0]**o+(-0.192258999531)*x_ref[0]+(0.615877054036)*x_ref[1]**o+(0.835274774437)*x_ref[1]
        ref[1,0,1,0]=(0.509342803767)*x_ref[0]**o+(-0.633944481882)*x_ref[0]+(0.262098006086)*x_ref[1]**o+(0.159902070799)*x_ref[1]
        ref[1,0,1,1]=(0.894432440588)*x_ref[0]**o+(-0.167778431319)*x_ref[0]+(-0.984183148477)*x_ref[1]**o+(-0.554981658632)*x_ref[1]
        ref[1,0,2,0]=(-0.762472607231)*x_ref[0]**o+(0.405550139256)*x_ref[0]+(0.4051290641)*x_ref[1]**o+(-0.27180186672)*x_ref[1]
        ref[1,0,2,1]=(0.663869273816)*x_ref[0]**o+(0.867453974628)*x_ref[0]+(-0.279659969132)*x_ref[1]**o+(-0.370539101204)*x_ref[1]
        ref[1,1,0,0]=(-0.969326778164)*x_ref[0]**o+(0.965471349497)*x_ref[0]+(-0.529983232946)*x_ref[1]**o+(0.714091610977)*x_ref[1]
        ref[1,1,0,1]=(-0.840360667093)*x_ref[0]**o+(-0.399319808256)*x_ref[0]+(0.568769915582)*x_ref[1]**o+(0.872427494442)*x_ref[1]
        ref[1,1,1,0]=(0.46099116941)*x_ref[0]**o+(-0.986008874142)*x_ref[0]+(-0.53999023556)*x_ref[1]**o+(-0.549695567855)*x_ref[1]
        ref[1,1,1,1]=(0.485944918559)*x_ref[0]**o+(0.189773828651)*x_ref[0]+(-0.611654980493)*x_ref[1]**o+(-0.337071911657)*x_ref[1]
        ref[1,1,2,0]=(0.191663603248)*x_ref[0]**o+(0.0906374965004)*x_ref[0]+(-0.271051598251)*x_ref[1]**o+(0.880507788825)*x_ref[1]
        ref[1,1,2,1]=(-0.439514997001)*x_ref[0]**o+(0.701230426224)*x_ref[0]+(-0.193231641029)*x_ref[1]**o+(0.950584054332)*x_ref[1]
        ref[1,2,0,0]=(0.439456565259)*x_ref[0]**o+(-0.25664171484)*x_ref[0]+(-0.653059819329)*x_ref[1]**o+(-0.680544747777)*x_ref[1]
        ref[1,2,0,1]=(0.34890454443)*x_ref[0]**o+(-0.69851200963)*x_ref[0]+(-0.00159381688549)*x_ref[1]**o+(-0.612359675499)*x_ref[1]
        ref[1,2,1,0]=(0.132759671375)*x_ref[0]**o+(-0.235126716401)*x_ref[0]+(0.652068885826)*x_ref[1]**o+(0.749698645371)*x_ref[1]
        ref[1,2,1,1]=(0.934980403627)*x_ref[0]**o+(0.321409273437)*x_ref[0]+(-0.605909929553)*x_ref[1]**o+(0.135554927418)*x_ref[1]
        ref[1,2,2,0]=(-0.68327709329)*x_ref[0]**o+(-0.0381292756005)*x_ref[0]+(0.710953191114)*x_ref[1]**o+(-0.344522256689)*x_ref[1]
        ref[1,2,2,1]=(-0.813976235652)*x_ref[0]**o+(0.613879708123)*x_ref[0]+(0.451599277293)*x_ref[1]**o+(-0.6134026375)*x_ref[1]
        ref[1,3,0,0]=(0.595310128627)*x_ref[0]**o+(-0.247621573131)*x_ref[0]+(0.165607783104)*x_ref[1]**o+(0.578132180805)*x_ref[1]
        ref[1,3,0,1]=(0.848827238882)*x_ref[0]**o+(0.173154067999)*x_ref[0]+(0.340662419437)*x_ref[1]**o+(-0.389656447251)*x_ref[1]
        ref[1,3,1,0]=(0.250888415228)*x_ref[0]**o+(-0.278203136198)*x_ref[0]+(-0.693211186734)*x_ref[1]**o+(0.613812194735)*x_ref[1]
        ref[1,3,1,1]=(0.634533003162)*x_ref[0]**o+(-0.0924665762471)*x_ref[0]+(0.749575894384)*x_ref[1]**o+(-0.737205017415)*x_ref[1]
        ref[1,3,2,0]=(0.607750966973)*x_ref[0]**o+(0.905016928404)*x_ref[0]+(-0.171049542163)*x_ref[1]**o+(-0.20526244601)*x_ref[1]
        ref[1,3,2,1]=(0.723946025405)*x_ref[0]**o+(0.0157576743987)*x_ref[0]+(0.190875428312)*x_ref[1]**o+(0.407925899509)*x_ref[1]
        ref[1,4,0,0]=(-0.561582695803)*x_ref[0]**o+(-0.350262461371)*x_ref[0]+(0.57610409872)*x_ref[1]**o+(0.881780963877)*x_ref[1]
        ref[1,4,0,1]=(0.077066498695)*x_ref[0]**o+(0.515597802516)*x_ref[0]+(0.332011413016)*x_ref[1]**o+(0.215491334787)*x_ref[1]
        ref[1,4,1,0]=(0.33858727737)*x_ref[0]**o+(-0.573231800768)*x_ref[0]+(-0.559915809132)*x_ref[1]**o+(0.0240869142176)*x_ref[1]
        ref[1,4,1,1]=(-0.684581727952)*x_ref[0]**o+(0.365541904201)*x_ref[0]+(0.204252397376)*x_ref[1]**o+(0.149300766054)*x_ref[1]
        ref[1,4,2,0]=(-0.574225088383)*x_ref[0]**o+(0.0286339776067)*x_ref[0]+(-0.358732122607)*x_ref[1]**o+(-0.416932738957)*x_ref[1]
        ref[1,4,2,1]=(-0.308460184932)*x_ref[0]**o+(0.389214987836)*x_ref[0]+(-0.358303257031)*x_ref[1]**o+(0.348109483688)*x_ref[1]
        ref[2,0,0,0]=(-0.811318668804)*x_ref[0]**o+(-0.298028116982)*x_ref[0]+(0.679871523483)*x_ref[1]**o+(0.958042925287)*x_ref[1]
        ref[2,0,0,1]=(-0.616521818338)*x_ref[0]**o+(-0.0817124854496)*x_ref[0]+(0.403257366893)*x_ref[1]**o+(0.740689710435)*x_ref[1]
        ref[2,0,1,0]=(0.271046179127)*x_ref[0]**o+(0.814928093947)*x_ref[0]+(0.423722781579)*x_ref[1]**o+(0.925333209811)*x_ref[1]
        ref[2,0,1,1]=(0.0596651806816)*x_ref[0]**o+(0.493854605359)*x_ref[0]+(0.539998537439)*x_ref[1]**o+(-0.910865639375)*x_ref[1]
        ref[2,0,2,0]=(0.503267602327)*x_ref[0]**o+(-0.270966471743)*x_ref[0]+(-0.342063366676)*x_ref[1]**o+(-0.975700224705)*x_ref[1]
        ref[2,0,2,1]=(-0.0362358939195)*x_ref[0]**o+(-0.766033869339)*x_ref[0]+(0.891365100088)*x_ref[1]**o+(0.815025898929)*x_ref[1]
        ref[2,1,0,0]=(-0.667240459984)*x_ref[0]**o+(0.8659076701)*x_ref[0]+(0.215089397562)*x_ref[1]**o+(0.464769182307)*x_ref[1]
        ref[2,1,0,1]=(-0.298625221931)*x_ref[0]**o+(-0.504329743472)*x_ref[0]+(0.412472510208)*x_ref[1]**o+(0.241693152889)*x_ref[1]
        ref[2,1,1,0]=(-0.670730179133)*x_ref[0]**o+(0.0201070113832)*x_ref[0]+(-0.572725898919)*x_ref[1]**o+(0.665609259014)*x_ref[1]
        ref[2,1,1,1]=(-0.319080984089)*x_ref[0]**o+(0.0299423045224)*x_ref[0]+(0.0152311810613)*x_ref[1]**o+(-0.784646189026)*x_ref[1]
        ref[2,1,2,0]=(0.153308789716)*x_ref[0]**o+(0.725808150613)*x_ref[0]+(0.62876493338)*x_ref[1]**o+(0.80275284252)*x_ref[1]
        ref[2,1,2,1]=(0.804698870013)*x_ref[0]**o+(-0.631078773061)*x_ref[0]+(-0.015221975698)*x_ref[1]**o+(-0.954169686746)*x_ref[1]
        ref[2,2,0,0]=(0.169022797498)*x_ref[0]**o+(0.733025791099)*x_ref[0]+(0.321554996633)*x_ref[1]**o+(0.292173365364)*x_ref[1]
        ref[2,2,0,1]=(0.327345955109)*x_ref[0]**o+(0.21757823896)*x_ref[0]+(0.873027160446)*x_ref[1]**o+(-0.250192861579)*x_ref[1]
        ref[2,2,1,0]=(0.394387190832)*x_ref[0]**o+(-0.613661883482)*x_ref[0]+(0.806965291386)*x_ref[1]**o+(0.244513244016)*x_ref[1]
        ref[2,2,1,1]=(0.043913907536)*x_ref[0]**o+(-0.303567806424)*x_ref[0]+(0.547037093323)*x_ref[1]**o+(-0.598687385328)*x_ref[1]
        ref[2,2,2,0]=(0.875663154787)*x_ref[0]**o+(-0.34984832772)*x_ref[0]+(0.879143174396)*x_ref[1]**o+(-0.193154051109)*x_ref[1]
        ref[2,2,2,1]=(0.940311132918)*x_ref[0]**o+(-0.358708743382)*x_ref[0]+(0.117370945146)*x_ref[1]**o+(-0.664208034757)*x_ref[1]
        ref[2,3,0,0]=(0.0334525660588)*x_ref[0]**o+(0.592805915255)*x_ref[0]+(-0.278216157971)*x_ref[1]**o+(0.894369967671)*x_ref[1]
        ref[2,3,0,1]=(-0.966591106326)*x_ref[0]**o+(-0.882160390768)*x_ref[0]+(0.843982702732)*x_ref[1]**o+(0.363868921656)*x_ref[1]
        ref[2,3,1,0]=(-0.246392342771)*x_ref[0]**o+(-0.934459096242)*x_ref[0]+(-0.472771368537)*x_ref[1]**o+(0.148854304706)*x_ref[1]
        ref[2,3,1,1]=(0.248218424051)*x_ref[0]**o+(0.0980396576111)*x_ref[0]+(0.413322166489)*x_ref[1]**o+(0.429747261346)*x_ref[1]
        ref[2,3,2,0]=(0.742987565733)*x_ref[0]**o+(-0.213736450136)*x_ref[0]+(0.996975741314)*x_ref[1]**o+(0.960709958665)*x_ref[1]
        ref[2,3,2,1]=(0.436149490977)*x_ref[0]**o+(0.87632181197)*x_ref[0]+(-0.881934669912)*x_ref[1]**o+(-0.457520504175)*x_ref[1]
        ref[2,4,0,0]=(-0.519014330693)*x_ref[0]**o+(0.938619608407)*x_ref[0]+(-0.847954563609)*x_ref[1]**o+(0.791774379621)*x_ref[1]
        ref[2,4,0,1]=(-0.599487299453)*x_ref[0]**o+(0.977867555132)*x_ref[0]+(0.693105630579)*x_ref[1]**o+(-0.688827812942)*x_ref[1]
        ref[2,4,1,0]=(0.955907964163)*x_ref[0]**o+(0.0724056986172)*x_ref[0]+(0.254773619772)*x_ref[1]**o+(0.0428861146964)*x_ref[1]
        ref[2,4,1,1]=(-0.422202101998)*x_ref[0]**o+(0.803263223544)*x_ref[0]+(0.456048140785)*x_ref[1]**o+(0.0419133126183)*x_ref[1]
        ref[2,4,2,0]=(-0.182308704781)*x_ref[0]**o+(0.862282443488)*x_ref[0]+(-0.844027007955)*x_ref[1]**o+(0.586220526376)*x_ref[1]
        ref[2,4,2,1]=(-0.801542666229)*x_ref[0]**o+(0.903685299455)*x_ref[0]+(0.428926544377)*x_ref[1]**o+(-0.891220471549)*x_ref[1]
        ref[3,0,0,0]=(0.602917787713)*x_ref[0]**o+(-0.592239549743)*x_ref[0]+(0.761830006718)*x_ref[1]**o+(-0.784264167777)*x_ref[1]
        ref[3,0,0,1]=(0.819962957732)*x_ref[0]**o+(-0.455126790372)*x_ref[0]+(-0.144898728525)*x_ref[1]**o+(0.121566345339)*x_ref[1]
        ref[3,0,1,0]=(0.570694941559)*x_ref[0]**o+(-0.505678656386)*x_ref[0]+(0.295536587355)*x_ref[1]**o+(-0.852779737814)*x_ref[1]
        ref[3,0,1,1]=(-0.226246858753)*x_ref[0]**o+(-0.513900306228)*x_ref[0]+(-0.389483591407)*x_ref[1]**o+(-0.735105335539)*x_ref[1]
        ref[3,0,2,0]=(0.941877610938)*x_ref[0]**o+(0.93808330298)*x_ref[0]+(0.189785162182)*x_ref[1]**o+(0.546525366304)*x_ref[1]
        ref[3,0,2,1]=(0.10639100485)*x_ref[0]**o+(-0.773361007972)*x_ref[0]+(-0.470531427579)*x_ref[1]**o+(0.467359342808)*x_ref[1]
        ref[3,1,0,0]=(0.857725599589)*x_ref[0]**o+(0.0881746413132)*x_ref[0]+(0.040713151152)*x_ref[1]**o+(-0.52070729883)*x_ref[1]
        ref[3,1,0,1]=(0.266807652146)*x_ref[0]**o+(0.337496673892)*x_ref[0]+(-0.562266297445)*x_ref[1]**o+(0.326015195113)*x_ref[1]
        ref[3,1,1,0]=(0.177275509465)*x_ref[0]**o+(-0.647855761301)*x_ref[0]+(-0.0963373631995)*x_ref[1]**o+(-0.535256252994)*x_ref[1]
        ref[3,1,1,1]=(0.430309295606)*x_ref[0]**o+(-0.0894749557428)*x_ref[0]+(0.503470724483)*x_ref[1]**o+(-0.202361122248)*x_ref[1]
        ref[3,1,2,0]=(0.283995014194)*x_ref[0]**o+(-0.544419907194)*x_ref[0]+(0.78478571712)*x_ref[1]**o+(0.341540664147)*x_ref[1]
        ref[3,1,2,1]=(-0.281042869936)*x_ref[0]**o+(0.728771766492)*x_ref[0]+(-0.298974388557)*x_ref[1]**o+(-0.836789906985)*x_ref[1]
        ref[3,2,0,0]=(-0.415323371561)*x_ref[0]**o+(0.583971137371)*x_ref[0]+(-0.265658679489)*x_ref[1]**o+(-0.432447373809)*x_ref[1]
        ref[3,2,0,1]=(0.690486298855)*x_ref[0]**o+(0.552521632226)*x_ref[0]+(0.237622739042)*x_ref[1]**o+(-0.254513693924)*x_ref[1]
        ref[3,2,1,0]=(0.453985687207)*x_ref[0]**o+(-0.785385580962)*x_ref[0]+(0.0190987841634)*x_ref[1]**o+(-0.7923288861)*x_ref[1]
        ref[3,2,1,1]=(-0.964082656379)*x_ref[0]**o+(-0.806315937371)*x_ref[0]+(0.271145420366)*x_ref[1]**o+(-0.839014535945)*x_ref[1]
        ref[3,2,2,0]=(-0.372062773997)*x_ref[0]**o+(-0.069749560567)*x_ref[0]+(-0.126310395383)*x_ref[1]**o+(0.972025865285)*x_ref[1]
        ref[3,2,2,1]=(-0.580353020086)*x_ref[0]**o+(-0.152746821477)*x_ref[0]+(0.297553660535)*x_ref[1]**o+(0.444791568145)*x_ref[1]
        ref[3,3,0,0]=(-0.361520029439)*x_ref[0]**o+(-0.810391070499)*x_ref[0]+(0.462123882642)*x_ref[1]**o+(-0.974534227742)*x_ref[1]
        ref[3,3,0,1]=(0.113370617837)*x_ref[0]**o+(0.428257355435)*x_ref[0]+(-0.18656623758)*x_ref[1]**o+(0.206503327394)*x_ref[1]
        ref[3,3,1,0]=(0.341328445689)*x_ref[0]**o+(0.34002947054)*x_ref[0]+(0.997677286362)*x_ref[1]**o+(0.43300073225)*x_ref[1]
        ref[3,3,1,1]=(-0.513912982823)*x_ref[0]**o+(0.826573827064)*x_ref[0]+(-0.701417864085)*x_ref[1]**o+(0.567078883219)*x_ref[1]
        ref[3,3,2,0]=(-0.325536866965)*x_ref[0]**o+(0.924405767451)*x_ref[0]+(-0.0481381954346)*x_ref[1]**o+(0.912210617073)*x_ref[1]
        ref[3,3,2,1]=(0.491532434302)*x_ref[0]**o+(-0.584980848036)*x_ref[0]+(-0.788317530858)*x_ref[1]**o+(-0.900681893172)*x_ref[1]
        ref[3,4,0,0]=(0.432249855148)*x_ref[0]**o+(-0.673369312552)*x_ref[0]+(0.244704845849)*x_ref[1]**o+(-0.300995595367)*x_ref[1]
        ref[3,4,0,1]=(0.383037211984)*x_ref[0]**o+(0.299711400488)*x_ref[0]+(0.870584255454)*x_ref[1]**o+(-0.33571823329)*x_ref[1]
        ref[3,4,1,0]=(0.737134180177)*x_ref[0]**o+(-0.101307090628)*x_ref[0]+(-0.970632061143)*x_ref[1]**o+(-0.449042684137)*x_ref[1]
        ref[3,4,1,1]=(0.00684250643632)*x_ref[0]**o+(0.676964654022)*x_ref[0]+(0.789069834858)*x_ref[1]**o+(-0.829739508197)*x_ref[1]
        ref[3,4,2,0]=(0.7279044944)*x_ref[0]**o+(0.215711122038)*x_ref[0]+(0.385631252632)*x_ref[1]**o+(0.629390433529)*x_ref[1]
        ref[3,4,2,1]=(-0.380750716629)*x_ref[0]**o+(-0.413040692437)*x_ref[0]+(0.26795244639)*x_ref[1]**o+(0.492475504659)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.287160004495)*x[0]**o+(0.35112250158)*x[0]+(0.061693427026)*x[1]**o+(-0.752505628029)*x[1]+(-0.57137756028)*x[2]**o+(-0.0397598892504)*x[2]
        arg[0,0,0,1]=(0.272799584607)*x[0]**o+(-0.569947956933)*x[0]+(0.322189261657)*x[1]**o+(-0.440552272224)*x[1]+(-0.10460241481)*x[2]**o+(-0.884848984752)*x[2]
        arg[0,0,1,0]=(0.147766301604)*x[0]**o+(0.405651654428)*x[0]+(-0.440827209529)*x[1]**o+(-0.399952864271)*x[1]+(-0.770191009459)*x[2]**o+(-0.789853968049)*x[2]
        arg[0,0,1,1]=(0.688053165375)*x[0]**o+(-0.154890421924)*x[0]+(0.583684164319)*x[1]**o+(-0.398082068217)*x[1]+(-0.573142923099)*x[2]**o+(-0.205271877337)*x[2]
        arg[0,0,2,0]=(0.807043275877)*x[0]**o+(0.843046383231)*x[0]+(-0.588146471402)*x[1]**o+(0.157787304557)*x[1]+(-0.771009748568)*x[2]**o+(0.162090791609)*x[2]
        arg[0,0,2,1]=(-0.0706800489392)*x[0]**o+(0.499961344665)*x[0]+(0.208926178797)*x[1]**o+(-0.7462472769)*x[1]+(-0.105475405648)*x[2]**o+(-0.18670229948)*x[2]
        arg[0,1,0,0]=(-0.696657281943)*x[0]**o+(-0.0251785345759)*x[0]+(-0.579575530451)*x[1]**o+(-0.0628378665725)*x[1]+(0.668395441302)*x[2]**o+(0.813478105092)*x[2]
        arg[0,1,0,1]=(-0.574784314876)*x[0]**o+(-0.321268045204)*x[0]+(0.160704464213)*x[1]**o+(-0.676016748103)*x[1]+(0.918971102865)*x[2]**o+(0.252082410081)*x[2]
        arg[0,1,1,0]=(-0.905062846817)*x[0]**o+(0.788906658477)*x[0]+(-0.408807611936)*x[1]**o+(0.114686186437)*x[1]+(-0.654685568493)*x[2]**o+(0.0879202139254)*x[2]
        arg[0,1,1,1]=(-0.68020231443)*x[0]**o+(-0.82049202902)*x[0]+(-0.958352740187)*x[1]**o+(-0.33201745802)*x[1]+(0.539810363682)*x[2]**o+(0.744988573087)*x[2]
        arg[0,1,2,0]=(-0.960884744997)*x[0]**o+(-0.704157132503)*x[0]+(0.282236905414)*x[1]**o+(0.756363456433)*x[1]+(-0.365400101792)*x[2]**o+(0.975735846554)*x[2]
        arg[0,1,2,1]=(-0.691780076951)*x[0]**o+(-0.952575440059)*x[0]+(-0.921247808098)*x[1]**o+(-0.88872834095)*x[1]+(0.534103013833)*x[2]**o+(-0.602295186104)*x[2]
        arg[0,2,0,0]=(0.460698415704)*x[0]**o+(0.770243814897)*x[0]+(-0.380381727574)*x[1]**o+(0.476360845668)*x[1]+(-0.681040425102)*x[2]**o+(-0.244369435116)*x[2]
        arg[0,2,0,1]=(0.251591652281)*x[0]**o+(0.748234919686)*x[0]+(-0.698189037307)*x[1]**o+(-0.292400412508)*x[1]+(0.613850696577)*x[2]**o+(-0.942243226834)*x[2]
        arg[0,2,1,0]=(-0.718090066328)*x[0]**o+(-0.322499511794)*x[0]+(-0.75105623397)*x[1]**o+(0.526680271832)*x[1]+(-0.186671531793)*x[2]**o+(-0.449273875991)*x[2]
        arg[0,2,1,1]=(-0.52165552452)*x[0]**o+(0.902446006085)*x[0]+(-0.915575172702)*x[1]**o+(-0.150816194662)*x[1]+(0.596782785689)*x[2]**o+(0.546880865642)*x[2]
        arg[0,2,2,0]=(-0.104486704816)*x[0]**o+(0.723863699457)*x[0]+(0.758622117279)*x[1]**o+(0.372271580145)*x[1]+(-0.0820570383792)*x[2]**o+(-0.63539781104)*x[2]
        arg[0,2,2,1]=(0.0355930823603)*x[0]**o+(0.0931942600432)*x[0]+(-0.118096985978)*x[1]**o+(0.296983782212)*x[1]+(-0.864491763712)*x[2]**o+(0.942230663978)*x[2]
        arg[0,3,0,0]=(0.371153709815)*x[0]**o+(0.0427309083893)*x[0]+(-0.886563919084)*x[1]**o+(-0.14946513404)*x[1]+(0.80903804519)*x[2]**o+(0.69728533073)*x[2]
        arg[0,3,0,1]=(0.0593930396142)*x[0]**o+(-0.370327974089)*x[0]+(-0.388354464683)*x[1]**o+(-0.367648771812)*x[1]+(-0.573754821458)*x[2]**o+(-0.947499155479)*x[2]
        arg[0,3,1,0]=(-0.869817317)*x[0]**o+(0.874317772678)*x[0]+(-0.870060178375)*x[1]**o+(0.66571074524)*x[1]+(-0.669052377378)*x[2]**o+(0.940212680575)*x[2]
        arg[0,3,1,1]=(-0.417445606772)*x[0]**o+(0.942606057619)*x[0]+(-0.133936034409)*x[1]**o+(-0.363274280866)*x[1]+(0.83487439935)*x[2]**o+(-0.108398981461)*x[2]
        arg[0,3,2,0]=(0.363623383293)*x[0]**o+(0.624267309217)*x[0]+(0.801719984899)*x[1]**o+(0.189212370122)*x[1]+(0.464633630928)*x[2]**o+(-0.979808318438)*x[2]
        arg[0,3,2,1]=(0.85384408805)*x[0]**o+(-0.889214754116)*x[0]+(0.164915391224)*x[1]**o+(0.0168991405201)*x[1]+(0.097171994236)*x[2]**o+(-0.146094197339)*x[2]
        arg[0,4,0,0]=(0.778356158972)*x[0]**o+(0.910241265149)*x[0]+(0.123690863272)*x[1]**o+(0.87802121779)*x[1]+(-0.714615668595)*x[2]**o+(-0.437472831583)*x[2]
        arg[0,4,0,1]=(-0.157185481783)*x[0]**o+(0.352802163968)*x[0]+(0.390184123612)*x[1]**o+(-0.680415093098)*x[1]+(-0.121195042454)*x[2]**o+(-0.713693712117)*x[2]
        arg[0,4,1,0]=(0.778115668646)*x[0]**o+(-0.111729042892)*x[0]+(0.451518002735)*x[1]**o+(0.232731183374)*x[1]+(-0.764246158086)*x[2]**o+(-0.795525155328)*x[2]
        arg[0,4,1,1]=(-0.880652367027)*x[0]**o+(-0.314966079126)*x[0]+(-0.640720373065)*x[1]**o+(0.336230541901)*x[1]+(0.566956999526)*x[2]**o+(0.74098258811)*x[2]
        arg[0,4,2,0]=(0.908657274301)*x[0]**o+(0.636427823614)*x[0]+(0.650137414386)*x[1]**o+(0.695558379005)*x[1]+(-0.489852401087)*x[2]**o+(-0.479227108044)*x[2]
        arg[0,4,2,1]=(0.015734241525)*x[0]**o+(0.0719884727088)*x[0]+(0.811864895261)*x[1]**o+(0.959127986784)*x[1]+(-0.238146663368)*x[2]**o+(-0.454939322571)*x[2]
        arg[1,0,0,0]=(-0.918218246559)*x[0]**o+(-0.662787169269)*x[0]+(-0.954035144111)*x[1]**o+(-0.647264685228)*x[1]+(-0.834085960213)*x[2]**o+(-0.167541951024)*x[2]
        arg[1,0,0,1]=(-0.670950869382)*x[0]**o+(0.826472159989)*x[0]+(0.180410872786)*x[1]**o+(-0.715162871594)*x[1]+(0.653287853169)*x[2]**o+(0.892060529625)*x[2]
        arg[1,0,1,0]=(0.398789786407)*x[0]**o+(0.689408033554)*x[0]+(0.124338245128)*x[1]**o+(0.748280808587)*x[1]+(0.418357122274)*x[2]**o+(0.230707440075)*x[2]
        arg[1,0,1,1]=(-0.912778964055)*x[0]**o+(0.904288708694)*x[0]+(-0.874744856901)*x[1]**o+(0.550162195719)*x[1]+(-0.7526423373)*x[2]**o+(0.820993231654)*x[2]
        arg[1,0,2,0]=(-0.843459684648)*x[0]**o+(-0.579708764091)*x[0]+(-0.876072131849)*x[1]**o+(0.933694599696)*x[1]+(0.507781209021)*x[2]**o+(-0.719729482133)*x[2]
        arg[1,0,2,1]=(0.594566925796)*x[0]**o+(-0.994301292358)*x[0]+(0.808114389564)*x[1]**o+(-0.970803712864)*x[1]+(0.433544910389)*x[2]**o+(0.260659440299)*x[2]
        arg[1,1,0,0]=(0.997377553172)*x[0]**o+(0.458428616383)*x[0]+(0.357280778493)*x[1]**o+(0.349837110741)*x[1]+(-0.709592214445)*x[2]**o+(-0.630000739033)*x[2]
        arg[1,1,0,1]=(0.121595650501)*x[0]**o+(0.40870120924)*x[0]+(0.576917123807)*x[1]**o+(-0.932268308325)*x[1]+(0.710043472116)*x[2]**o+(-0.716528412077)*x[2]
        arg[1,1,1,0]=(-0.151416035033)*x[0]**o+(0.0255347649787)*x[0]+(-0.752987701889)*x[1]**o+(-0.480969875813)*x[1]+(-0.610514504179)*x[2]**o+(0.959274757402)*x[2]
        arg[1,1,1,1]=(0.76695706106)*x[0]**o+(-0.0750965975512)*x[0]+(0.846833900735)*x[1]**o+(-0.105838465439)*x[1]+(0.541027486457)*x[2]**o+(0.152543017589)*x[2]
        arg[1,1,2,0]=(0.0856977181865)*x[0]**o+(-0.188139369706)*x[0]+(0.936071565289)*x[1]**o+(-0.3958713036)*x[1]+(-0.350945878893)*x[2]**o+(0.382302674883)*x[2]
        arg[1,1,2,1]=(0.873054234789)*x[0]**o+(-0.490869716146)*x[0]+(0.46720526212)*x[1]**o+(-0.334203412334)*x[1]+(-0.81140895636)*x[2]**o+(-0.973858971749)*x[2]
        arg[1,2,0,0]=(-0.716676352443)*x[0]**o+(0.659821396277)*x[0]+(0.335964292954)*x[1]**o+(0.11609664773)*x[1]+(-0.889186749466)*x[2]**o+(0.454231973946)*x[2]
        arg[1,2,0,1]=(0.367048872648)*x[0]**o+(0.276083713126)*x[0]+(0.686639128043)*x[1]**o+(-0.991414415293)*x[1]+(0.25360515964)*x[2]**o+(-0.377623315093)*x[2]
        arg[1,2,1,0]=(-0.723089358777)*x[0]**o+(-0.516690112406)*x[0]+(0.262334273979)*x[1]**o+(0.140773502537)*x[1]+(0.447333972396)*x[2]**o+(-0.551395692966)*x[2]
        arg[1,2,1,1]=(0.738791182491)*x[0]**o+(-0.978120626668)*x[0]+(-0.217970433713)*x[1]**o+(0.0181011067492)*x[1]+(0.756322867523)*x[2]**o+(0.495269195978)*x[2]
        arg[1,2,2,0]=(-0.630506449854)*x[0]**o+(-0.0865482364231)*x[0]+(-0.415896403157)*x[1]**o+(-0.712833149895)*x[1]+(-0.180482789777)*x[2]**o+(0.00788290661235)*x[2]
        arg[1,2,2,1]=(0.0946913865377)*x[0]**o+(0.752854827716)*x[0]+(0.257846220843)*x[1]**o+(-0.953926983578)*x[1]+(-0.388505920682)*x[2]**o+(-0.10751401251)*x[2]
        arg[1,3,0,0]=(0.694014030611)*x[0]**o+(0.0433103547797)*x[0]+(-0.0805025225569)*x[1]**o+(0.389052254754)*x[1]+(-0.813210939614)*x[2]**o+(-0.593788787896)*x[2]
        arg[1,3,0,1]=(0.0101606353406)*x[0]**o+(-0.849874258855)*x[0]+(-0.0578973959381)*x[1]**o+(0.942113140424)*x[1]+(0.613716459216)*x[2]**o+(0.72631176833)*x[2]
        arg[1,3,1,0]=(0.264703608811)*x[0]**o+(0.457962548326)*x[0]+(-0.433795242673)*x[1]**o+(0.650995712848)*x[1]+(0.372471807967)*x[2]**o+(0.00416636841047)*x[2]
        arg[1,3,1,1]=(-0.73686451669)*x[0]**o+(-0.0425614910776)*x[0]+(-0.587820431311)*x[1]**o+(-0.377628617833)*x[1]+(-0.463964014406)*x[2]**o+(0.283273874989)*x[2]
        arg[1,3,2,0]=(-0.689341439143)*x[0]**o+(-0.675861918889)*x[0]+(-0.918622662352)*x[1]**o+(0.740990261888)*x[1]+(0.0454007677041)*x[2]**o+(0.513623361074)*x[2]
        arg[1,3,2,1]=(0.13229101848)*x[0]**o+(0.171873799817)*x[0]+(-0.154491174434)*x[1]**o+(0.897976884089)*x[1]+(-0.598886311821)*x[2]**o+(0.738696710918)*x[2]
        arg[1,4,0,0]=(0.59403636966)*x[0]**o+(0.93856848969)*x[0]+(0.80164849167)*x[1]**o+(-0.753322979568)*x[1]+(-0.107098073916)*x[2]**o+(-0.623047487973)*x[2]
        arg[1,4,0,1]=(0.291046099082)*x[0]**o+(-0.437982471049)*x[0]+(0.017395762947)*x[1]**o+(0.852440593923)*x[1]+(0.630825424359)*x[2]**o+(0.419347491727)*x[2]
        arg[1,4,1,0]=(-0.747962012701)*x[0]**o+(-0.11622032094)*x[0]+(0.324950549781)*x[1]**o+(0.950911398327)*x[1]+(-0.18098990205)*x[2]**o+(-0.676060633522)*x[2]
        arg[1,4,1,1]=(0.545224363128)*x[0]**o+(0.753376367121)*x[0]+(0.271426695753)*x[1]**o+(-0.139309844675)*x[1]+(-0.923434716375)*x[2]**o+(-0.985988324389)*x[2]
        arg[1,4,2,0]=(-0.532260822047)*x[0]**o+(0.901407135724)*x[0]+(0.937451330141)*x[1]**o+(0.285217666559)*x[1]+(-0.133028984824)*x[2]**o+(0.993965133696)*x[2]
        arg[1,4,2,1]=(0.2231815607)*x[0]**o+(0.994845537582)*x[0]+(0.0368727378195)*x[1]**o+(0.511772933777)*x[1]+(-0.871580215976)*x[2]**o+(-0.587010057329)*x[2]
        arg[2,0,0,0]=(-0.55149411913)*x[0]**o+(-0.905541278559)*x[0]+(0.461217373096)*x[1]**o+(-0.53359853241)*x[1]+(-0.666643290883)*x[2]**o+(0.151495826926)*x[2]
        arg[2,0,0,1]=(-0.980485062148)*x[0]**o+(0.150165616162)*x[0]+(0.59165725487)*x[1]**o+(-0.243853536017)*x[1]+(0.66816261686)*x[2]**o+(0.2425696131)*x[2]
        arg[2,0,1,0]=(-0.509337213758)*x[0]**o+(0.652936706687)*x[0]+(-0.849214338091)*x[1]**o+(-0.523354123549)*x[1]+(0.371515689887)*x[2]**o+(0.799989815414)*x[2]
        arg[2,0,1,1]=(0.461584544541)*x[0]**o+(0.662931355954)*x[0]+(0.744415634786)*x[1]**o+(-0.016743801256)*x[1]+(0.172945846325)*x[2]**o+(-0.0889827089341)*x[2]
        arg[2,0,2,0]=(0.506514791562)*x[0]**o+(0.974477160694)*x[0]+(-0.968485343167)*x[1]**o+(0.427480013599)*x[1]+(0.0460755879358)*x[2]**o+(0.186500994948)*x[2]
        arg[2,0,2,1]=(0.259831607993)*x[0]**o+(-0.0141290444125)*x[0]+(-0.968452177682)*x[1]**o+(0.635619125009)*x[1]+(-0.958848431827)*x[2]**o+(-0.0977241523308)*x[2]
        arg[2,1,0,0]=(0.204233222933)*x[0]**o+(0.834739381877)*x[0]+(0.19961185048)*x[1]**o+(-0.238470380578)*x[1]+(-0.749979713773)*x[2]**o+(-0.268168886251)*x[2]
        arg[2,1,0,1]=(0.998123643088)*x[0]**o+(0.635823733041)*x[0]+(-0.265447124419)*x[1]**o+(-0.225582751946)*x[1]+(0.91705937439)*x[2]**o+(-0.739461510792)*x[2]
        arg[2,1,1,0]=(0.0800715288646)*x[0]**o+(-0.37677196659)*x[0]+(-0.0737286350219)*x[1]**o+(0.032076062933)*x[1]+(0.245421912709)*x[2]**o+(-0.8378023935)*x[2]
        arg[2,1,1,1]=(-0.572376055603)*x[0]**o+(0.8675468469)*x[0]+(-0.533012889867)*x[1]**o+(-0.247215658297)*x[1]+(0.673144564693)*x[2]**o+(-0.749136108985)*x[2]
        arg[2,1,2,0]=(-0.661188192488)*x[0]**o+(-0.924435593008)*x[0]+(-0.791845808916)*x[1]**o+(0.00133765351092)*x[1]+(0.233676862873)*x[2]**o+(0.981812737962)*x[2]
        arg[2,1,2,1]=(0.332072445986)*x[0]**o+(-0.909373798588)*x[0]+(0.927025647819)*x[1]**o+(-0.936348942617)*x[1]+(0.251508592152)*x[2]**o+(-0.468147973987)*x[2]
        arg[2,2,0,0]=(0.892377893841)*x[0]**o+(-0.147933293091)*x[0]+(0.950491102076)*x[1]**o+(0.595250355926)*x[1]+(0.0169764649992)*x[2]**o+(-0.683137752586)*x[2]
        arg[2,2,0,1]=(0.00579397005645)*x[0]**o+(0.627197373917)*x[0]+(0.663574756196)*x[1]**o+(0.934106767513)*x[1]+(0.105844878022)*x[2]**o+(0.513322917688)*x[2]
        arg[2,2,1,0]=(-0.168776016133)*x[0]**o+(-0.940071612944)*x[0]+(0.0366053554566)*x[1]**o+(-0.628010909771)*x[1]+(0.250023045906)*x[2]**o+(-0.00327322332166)*x[2]
        arg[2,2,1,1]=(0.904155764376)*x[0]**o+(-0.258018307306)*x[0]+(0.65862820521)*x[1]**o+(0.434480798515)*x[1]+(0.4089419173)*x[2]**o+(-0.518032695598)*x[2]
        arg[2,2,2,0]=(-0.41470070958)*x[0]**o+(-0.783186751952)*x[0]+(0.548309561942)*x[1]**o+(0.837130592803)*x[1]+(-0.346808491848)*x[2]**o+(0.165416040985)*x[2]
        arg[2,2,2,1]=(0.997845909195)*x[0]**o+(0.448428042158)*x[0]+(-0.657050004311)*x[1]**o+(0.493104851747)*x[1]+(0.558315071197)*x[2]**o+(0.206659432923)*x[2]
        arg[2,3,0,0]=(0.433874360232)*x[0]**o+(-0.0635646197131)*x[0]+(-0.521690521523)*x[1]**o+(-0.298130684888)*x[1]+(-0.131905313644)*x[2]**o+(0.64389451112)*x[2]
        arg[2,3,0,1]=(0.457729315216)*x[0]**o+(-0.419323242135)*x[0]+(0.614544051135)*x[1]**o+(-0.701671999275)*x[1]+(0.905494300727)*x[2]**o+(-0.565716985337)*x[2]
        arg[2,3,1,0]=(0.256570728097)*x[0]**o+(0.866349285103)*x[0]+(-0.115768354896)*x[1]**o+(-0.613044456549)*x[1]+(-0.20074185156)*x[2]**o+(-0.188403571776)*x[2]
        arg[2,3,1,1]=(0.767479943414)*x[0]**o+(0.884066215007)*x[0]+(0.363855420056)*x[1]**o+(0.919698981104)*x[1]+(0.225747762317)*x[2]**o+(-0.220362271104)*x[2]
        arg[2,3,2,0]=(-0.746469501382)*x[0]**o+(0.680520398001)*x[0]+(0.926950996325)*x[1]**o+(-0.122831303044)*x[1]+(-0.27226561675)*x[2]**o+(0.933304087951)*x[2]
        arg[2,3,2,1]=(-0.0961434747216)*x[0]**o+(0.987660516767)*x[0]+(0.288175205045)*x[1]**o+(0.683720820403)*x[1]+(0.354403143642)*x[2]**o+(-0.745694089132)*x[2]
        arg[2,4,0,0]=(0.921750835986)*x[0]**o+(-0.0358861598394)*x[0]+(0.190129645259)*x[1]**o+(0.377067803331)*x[1]+(-0.0188615236582)*x[2]**o+(0.804533888824)*x[2]
        arg[2,4,0,1]=(-0.256132234484)*x[0]**o+(-0.102661411017)*x[0]+(-0.903465157009)*x[1]**o+(0.50133310168)*x[1]+(-0.0889102598619)*x[2]**o+(-0.0310174238365)*x[2]
        arg[2,4,1,0]=(0.58483798855)*x[0]**o+(0.967943344305)*x[0]+(-0.0701080005438)*x[1]**o+(0.489476104406)*x[1]+(0.615308427038)*x[2]**o+(-0.38065505709)*x[2]
        arg[2,4,1,1]=(0.647734130724)*x[0]**o+(-0.930934392719)*x[0]+(0.267370540491)*x[1]**o+(-0.0175019424556)*x[1]+(0.38593657129)*x[2]**o+(-0.985922292359)*x[2]
        arg[2,4,2,0]=(-0.289971479933)*x[0]**o+(0.882014313157)*x[0]+(0.147321067579)*x[1]**o+(-0.176771838489)*x[1]+(0.710916531544)*x[2]**o+(-0.0856290170401)*x[2]
        arg[2,4,2,1]=(-0.465930997733)*x[0]**o+(-0.305563942094)*x[0]+(-0.824535262061)*x[1]**o+(-0.917892226972)*x[1]+(0.624462871955)*x[2]**o+(0.544736500065)*x[2]
        arg[3,0,0,0]=(-0.673230815097)*x[0]**o+(-0.726292712917)*x[0]+(0.0531037021688)*x[1]**o+(-0.467219030159)*x[1]+(-0.0729861369718)*x[2]**o+(-0.440394335015)*x[2]
        arg[3,0,0,1]=(-0.358235295558)*x[0]**o+(-0.301184713671)*x[0]+(-0.496030649086)*x[1]**o+(-0.125922184777)*x[1]+(0.946715036973)*x[2]**o+(0.846370337893)*x[2]
        arg[3,0,1,0]=(-0.394175109612)*x[0]**o+(-0.85137582307)*x[0]+(-0.667804785044)*x[1]**o+(0.156762109891)*x[1]+(-0.616444700525)*x[2]**o+(0.0738729399356)*x[2]
        arg[3,0,1,1]=(-0.652875462534)*x[0]**o+(-0.0605650192541)*x[0]+(0.840539933823)*x[1]**o+(-0.422819423301)*x[1]+(0.113440226819)*x[2]**o+(-0.915154532306)*x[2]
        arg[3,0,2,0]=(-0.162863352328)*x[0]**o+(-0.364461332292)*x[0]+(-0.222181741398)*x[1]**o+(-0.563668008365)*x[1]+(0.118491770105)*x[2]**o+(0.0809155053022)*x[2]
        arg[3,0,2,1]=(-0.985278929437)*x[0]**o+(-0.919440760157)*x[0]+(-0.776910686)*x[1]**o+(-0.884270488943)*x[1]+(0.984279010363)*x[2]**o+(-0.489503278582)*x[2]
        arg[3,1,0,0]=(-0.806565449823)*x[0]**o+(-0.0851385375281)*x[0]+(-0.88851381773)*x[1]**o+(-0.666539867528)*x[1]+(0.154825531411)*x[2]**o+(0.764069849144)*x[2]
        arg[3,1,0,1]=(-0.222125783918)*x[0]**o+(0.568035675467)*x[0]+(0.470171778932)*x[1]**o+(-0.307608710881)*x[1]+(-0.836083742471)*x[2]**o+(0.553597553381)*x[2]
        arg[3,1,1,0]=(0.431237389847)*x[0]**o+(-0.141391012471)*x[0]+(0.726895025519)*x[1]**o+(-0.527868751679)*x[1]+(0.711761083137)*x[2]**o+(0.996992864103)*x[2]
        arg[3,1,1,1]=(-0.123885959924)*x[0]**o+(0.877034475529)*x[0]+(-0.0432492682928)*x[1]**o+(-0.370315494618)*x[1]+(0.243139719071)*x[2]**o+(-0.684318634902)*x[2]
        arg[3,1,2,0]=(-0.477214489355)*x[0]**o+(0.340808576619)*x[0]+(-0.0991802306558)*x[1]**o+(0.462599821908)*x[1]+(0.091768281564)*x[2]**o+(0.0315552838839)*x[2]
        arg[3,1,2,1]=(0.676189142894)*x[0]**o+(0.510136386913)*x[0]+(-0.558579041952)*x[1]**o+(0.961615047877)*x[1]+(0.510717386124)*x[2]**o+(0.774430183432)*x[2]
        arg[3,2,0,0]=(0.158370232444)*x[0]**o+(0.209424855992)*x[0]+(0.691438753013)*x[1]**o+(-0.781594298738)*x[1]+(-0.0188536627517)*x[2]**o+(0.136400798995)*x[2]
        arg[3,2,0,1]=(-0.440674414328)*x[0]**o+(0.118228612687)*x[0]+(-0.700026492613)*x[1]**o+(-0.981341215714)*x[1]+(-0.172847519578)*x[2]**o+(-0.423095202172)*x[2]
        arg[3,2,1,0]=(-0.0694960391695)*x[0]**o+(-0.66448610135)*x[0]+(-0.177572703748)*x[1]**o+(-0.060488161727)*x[1]+(0.306021853885)*x[2]**o+(0.0730338716187)*x[2]
        arg[3,2,1,1]=(-0.360662154547)*x[0]**o+(0.577679359648)*x[0]+(-0.515313534966)*x[1]**o+(0.217352687358)*x[1]+(0.728367577882)*x[2]**o+(0.38505849834)*x[2]
        arg[3,2,2,0]=(0.568671861232)*x[0]**o+(-0.0792490917882)*x[0]+(0.0188002885229)*x[1]**o+(-0.956862575759)*x[1]+(-0.844332442983)*x[2]**o+(0.670792928042)*x[2]
        arg[3,2,2,1]=(0.833099204713)*x[0]**o+(0.368482169933)*x[0]+(-0.926002498892)*x[1]**o+(-0.130646446568)*x[1]+(0.81164928914)*x[2]**o+(-0.641079919927)*x[2]
        arg[3,3,0,0]=(-0.277769350989)*x[0]**o+(0.753452018649)*x[0]+(-0.766322344505)*x[1]**o+(-0.679108224612)*x[1]+(0.0799499795424)*x[2]**o+(0.255751308975)*x[2]
        arg[3,3,0,1]=(0.429176753733)*x[0]**o+(0.123106050259)*x[0]+(-0.169037728992)*x[1]**o+(0.749245933687)*x[1]+(0.560140859355)*x[2]**o+(-0.937091840021)*x[2]
        arg[3,3,1,0]=(0.648792557593)*x[0]**o+(-0.057555187513)*x[0]+(0.883290331726)*x[1]**o+(0.539276905995)*x[1]+(0.473614408583)*x[2]**o+(0.723258340333)*x[2]
        arg[3,3,1,1]=(0.527298833572)*x[0]**o+(0.0753824058279)*x[0]+(0.719196099257)*x[1]**o+(-0.0124603350035)*x[1]+(-0.964484650253)*x[2]**o+(0.404444269694)*x[2]
        arg[3,3,2,0]=(-0.403640677462)*x[0]**o+(0.906195964766)*x[0]+(-0.561419078404)*x[1]**o+(0.645408717838)*x[1]+(0.671624799609)*x[2]**o+(-0.573901868621)*x[2]
        arg[3,3,2,1]=(-0.288117182165)*x[0]**o+(-0.391003584945)*x[0]+(-0.555037182038)*x[1]**o+(0.299362209017)*x[1]+(-0.552429724326)*x[2]**o+(-0.416923306774)*x[2]
        arg[3,4,0,0]=(-0.19594719192)*x[0]**o+(0.150898934692)*x[0]+(0.210935232044)*x[1]**o+(-0.319764293347)*x[1]+(-0.642450751195)*x[2]**o+(0.22231085168)*x[2]
        arg[3,4,0,1]=(-0.0514589460225)*x[0]**o+(0.0366025646754)*x[0]+(0.416280576691)*x[1]**o+(0.609178063951)*x[1]+(0.835611677062)*x[2]**o+(-0.131630571478)*x[2]
        arg[3,4,1,0]=(0.171509152214)*x[0]**o+(-0.761188271093)*x[0]+(-0.915951802896)*x[1]**o+(-0.451796709915)*x[1]+(-0.979847828162)*x[2]**o+(0.0566453393315)*x[2]
        arg[3,4,1,1]=(0.690427998102)*x[0]**o+(-0.358318625695)*x[0]+(0.330270871613)*x[1]**o+(-0.724392522137)*x[1]+(-0.535474623489)*x[2]**o+(-0.12341743885)*x[2]
        arg[3,4,2,0]=(0.488090753426)*x[0]**o+(-0.352417078372)*x[0]+(0.971988099319)*x[1]**o+(-0.288929516009)*x[1]+(0.612576734094)*x[2]**o+(-0.43799514603)*x[2]
        arg[3,4,2,1]=(-0.098768101356)*x[0]**o+(0.123300661895)*x[0]+(-0.0489724280789)*x[1]**o+(-0.422508669137)*x[1]+(-0.861483104139)*x[2]**o+(0.950198316001)*x[2]
        ref[0,0,0,0]=(0.287160004495)*x_ref[0]**o+(0.35112250158)*x_ref[0]+(0.061693427026)*x_ref[1]**o+(-0.752505628029)*x_ref[1]+(-0.57137756028)*x_ref[2]**o+(-0.0397598892504)*x_ref[2]
        ref[0,0,0,1]=(0.272799584607)*x_ref[0]**o+(-0.569947956933)*x_ref[0]+(0.322189261657)*x_ref[1]**o+(-0.440552272224)*x_ref[1]+(-0.10460241481)*x_ref[2]**o+(-0.884848984752)*x_ref[2]
        ref[0,0,1,0]=(0.147766301604)*x_ref[0]**o+(0.405651654428)*x_ref[0]+(-0.440827209529)*x_ref[1]**o+(-0.399952864271)*x_ref[1]+(-0.770191009459)*x_ref[2]**o+(-0.789853968049)*x_ref[2]
        ref[0,0,1,1]=(0.688053165375)*x_ref[0]**o+(-0.154890421924)*x_ref[0]+(0.583684164319)*x_ref[1]**o+(-0.398082068217)*x_ref[1]+(-0.573142923099)*x_ref[2]**o+(-0.205271877337)*x_ref[2]
        ref[0,0,2,0]=(0.807043275877)*x_ref[0]**o+(0.843046383231)*x_ref[0]+(-0.588146471402)*x_ref[1]**o+(0.157787304557)*x_ref[1]+(-0.771009748568)*x_ref[2]**o+(0.162090791609)*x_ref[2]
        ref[0,0,2,1]=(-0.0706800489392)*x_ref[0]**o+(0.499961344665)*x_ref[0]+(0.208926178797)*x_ref[1]**o+(-0.7462472769)*x_ref[1]+(-0.105475405648)*x_ref[2]**o+(-0.18670229948)*x_ref[2]
        ref[0,1,0,0]=(-0.696657281943)*x_ref[0]**o+(-0.0251785345759)*x_ref[0]+(-0.579575530451)*x_ref[1]**o+(-0.0628378665725)*x_ref[1]+(0.668395441302)*x_ref[2]**o+(0.813478105092)*x_ref[2]
        ref[0,1,0,1]=(-0.574784314876)*x_ref[0]**o+(-0.321268045204)*x_ref[0]+(0.160704464213)*x_ref[1]**o+(-0.676016748103)*x_ref[1]+(0.918971102865)*x_ref[2]**o+(0.252082410081)*x_ref[2]
        ref[0,1,1,0]=(-0.905062846817)*x_ref[0]**o+(0.788906658477)*x_ref[0]+(-0.408807611936)*x_ref[1]**o+(0.114686186437)*x_ref[1]+(-0.654685568493)*x_ref[2]**o+(0.0879202139254)*x_ref[2]
        ref[0,1,1,1]=(-0.68020231443)*x_ref[0]**o+(-0.82049202902)*x_ref[0]+(-0.958352740187)*x_ref[1]**o+(-0.33201745802)*x_ref[1]+(0.539810363682)*x_ref[2]**o+(0.744988573087)*x_ref[2]
        ref[0,1,2,0]=(-0.960884744997)*x_ref[0]**o+(-0.704157132503)*x_ref[0]+(0.282236905414)*x_ref[1]**o+(0.756363456433)*x_ref[1]+(-0.365400101792)*x_ref[2]**o+(0.975735846554)*x_ref[2]
        ref[0,1,2,1]=(-0.691780076951)*x_ref[0]**o+(-0.952575440059)*x_ref[0]+(-0.921247808098)*x_ref[1]**o+(-0.88872834095)*x_ref[1]+(0.534103013833)*x_ref[2]**o+(-0.602295186104)*x_ref[2]
        ref[0,2,0,0]=(0.460698415704)*x_ref[0]**o+(0.770243814897)*x_ref[0]+(-0.380381727574)*x_ref[1]**o+(0.476360845668)*x_ref[1]+(-0.681040425102)*x_ref[2]**o+(-0.244369435116)*x_ref[2]
        ref[0,2,0,1]=(0.251591652281)*x_ref[0]**o+(0.748234919686)*x_ref[0]+(-0.698189037307)*x_ref[1]**o+(-0.292400412508)*x_ref[1]+(0.613850696577)*x_ref[2]**o+(-0.942243226834)*x_ref[2]
        ref[0,2,1,0]=(-0.718090066328)*x_ref[0]**o+(-0.322499511794)*x_ref[0]+(-0.75105623397)*x_ref[1]**o+(0.526680271832)*x_ref[1]+(-0.186671531793)*x_ref[2]**o+(-0.449273875991)*x_ref[2]
        ref[0,2,1,1]=(-0.52165552452)*x_ref[0]**o+(0.902446006085)*x_ref[0]+(-0.915575172702)*x_ref[1]**o+(-0.150816194662)*x_ref[1]+(0.596782785689)*x_ref[2]**o+(0.546880865642)*x_ref[2]
        ref[0,2,2,0]=(-0.104486704816)*x_ref[0]**o+(0.723863699457)*x_ref[0]+(0.758622117279)*x_ref[1]**o+(0.372271580145)*x_ref[1]+(-0.0820570383792)*x_ref[2]**o+(-0.63539781104)*x_ref[2]
        ref[0,2,2,1]=(0.0355930823603)*x_ref[0]**o+(0.0931942600432)*x_ref[0]+(-0.118096985978)*x_ref[1]**o+(0.296983782212)*x_ref[1]+(-0.864491763712)*x_ref[2]**o+(0.942230663978)*x_ref[2]
        ref[0,3,0,0]=(0.371153709815)*x_ref[0]**o+(0.0427309083893)*x_ref[0]+(-0.886563919084)*x_ref[1]**o+(-0.14946513404)*x_ref[1]+(0.80903804519)*x_ref[2]**o+(0.69728533073)*x_ref[2]
        ref[0,3,0,1]=(0.0593930396142)*x_ref[0]**o+(-0.370327974089)*x_ref[0]+(-0.388354464683)*x_ref[1]**o+(-0.367648771812)*x_ref[1]+(-0.573754821458)*x_ref[2]**o+(-0.947499155479)*x_ref[2]
        ref[0,3,1,0]=(-0.869817317)*x_ref[0]**o+(0.874317772678)*x_ref[0]+(-0.870060178375)*x_ref[1]**o+(0.66571074524)*x_ref[1]+(-0.669052377378)*x_ref[2]**o+(0.940212680575)*x_ref[2]
        ref[0,3,1,1]=(-0.417445606772)*x_ref[0]**o+(0.942606057619)*x_ref[0]+(-0.133936034409)*x_ref[1]**o+(-0.363274280866)*x_ref[1]+(0.83487439935)*x_ref[2]**o+(-0.108398981461)*x_ref[2]
        ref[0,3,2,0]=(0.363623383293)*x_ref[0]**o+(0.624267309217)*x_ref[0]+(0.801719984899)*x_ref[1]**o+(0.189212370122)*x_ref[1]+(0.464633630928)*x_ref[2]**o+(-0.979808318438)*x_ref[2]
        ref[0,3,2,1]=(0.85384408805)*x_ref[0]**o+(-0.889214754116)*x_ref[0]+(0.164915391224)*x_ref[1]**o+(0.0168991405201)*x_ref[1]+(0.097171994236)*x_ref[2]**o+(-0.146094197339)*x_ref[2]
        ref[0,4,0,0]=(0.778356158972)*x_ref[0]**o+(0.910241265149)*x_ref[0]+(0.123690863272)*x_ref[1]**o+(0.87802121779)*x_ref[1]+(-0.714615668595)*x_ref[2]**o+(-0.437472831583)*x_ref[2]
        ref[0,4,0,1]=(-0.157185481783)*x_ref[0]**o+(0.352802163968)*x_ref[0]+(0.390184123612)*x_ref[1]**o+(-0.680415093098)*x_ref[1]+(-0.121195042454)*x_ref[2]**o+(-0.713693712117)*x_ref[2]
        ref[0,4,1,0]=(0.778115668646)*x_ref[0]**o+(-0.111729042892)*x_ref[0]+(0.451518002735)*x_ref[1]**o+(0.232731183374)*x_ref[1]+(-0.764246158086)*x_ref[2]**o+(-0.795525155328)*x_ref[2]
        ref[0,4,1,1]=(-0.880652367027)*x_ref[0]**o+(-0.314966079126)*x_ref[0]+(-0.640720373065)*x_ref[1]**o+(0.336230541901)*x_ref[1]+(0.566956999526)*x_ref[2]**o+(0.74098258811)*x_ref[2]
        ref[0,4,2,0]=(0.908657274301)*x_ref[0]**o+(0.636427823614)*x_ref[0]+(0.650137414386)*x_ref[1]**o+(0.695558379005)*x_ref[1]+(-0.489852401087)*x_ref[2]**o+(-0.479227108044)*x_ref[2]
        ref[0,4,2,1]=(0.015734241525)*x_ref[0]**o+(0.0719884727088)*x_ref[0]+(0.811864895261)*x_ref[1]**o+(0.959127986784)*x_ref[1]+(-0.238146663368)*x_ref[2]**o+(-0.454939322571)*x_ref[2]
        ref[1,0,0,0]=(-0.918218246559)*x_ref[0]**o+(-0.662787169269)*x_ref[0]+(-0.954035144111)*x_ref[1]**o+(-0.647264685228)*x_ref[1]+(-0.834085960213)*x_ref[2]**o+(-0.167541951024)*x_ref[2]
        ref[1,0,0,1]=(-0.670950869382)*x_ref[0]**o+(0.826472159989)*x_ref[0]+(0.180410872786)*x_ref[1]**o+(-0.715162871594)*x_ref[1]+(0.653287853169)*x_ref[2]**o+(0.892060529625)*x_ref[2]
        ref[1,0,1,0]=(0.398789786407)*x_ref[0]**o+(0.689408033554)*x_ref[0]+(0.124338245128)*x_ref[1]**o+(0.748280808587)*x_ref[1]+(0.418357122274)*x_ref[2]**o+(0.230707440075)*x_ref[2]
        ref[1,0,1,1]=(-0.912778964055)*x_ref[0]**o+(0.904288708694)*x_ref[0]+(-0.874744856901)*x_ref[1]**o+(0.550162195719)*x_ref[1]+(-0.7526423373)*x_ref[2]**o+(0.820993231654)*x_ref[2]
        ref[1,0,2,0]=(-0.843459684648)*x_ref[0]**o+(-0.579708764091)*x_ref[0]+(-0.876072131849)*x_ref[1]**o+(0.933694599696)*x_ref[1]+(0.507781209021)*x_ref[2]**o+(-0.719729482133)*x_ref[2]
        ref[1,0,2,1]=(0.594566925796)*x_ref[0]**o+(-0.994301292358)*x_ref[0]+(0.808114389564)*x_ref[1]**o+(-0.970803712864)*x_ref[1]+(0.433544910389)*x_ref[2]**o+(0.260659440299)*x_ref[2]
        ref[1,1,0,0]=(0.997377553172)*x_ref[0]**o+(0.458428616383)*x_ref[0]+(0.357280778493)*x_ref[1]**o+(0.349837110741)*x_ref[1]+(-0.709592214445)*x_ref[2]**o+(-0.630000739033)*x_ref[2]
        ref[1,1,0,1]=(0.121595650501)*x_ref[0]**o+(0.40870120924)*x_ref[0]+(0.576917123807)*x_ref[1]**o+(-0.932268308325)*x_ref[1]+(0.710043472116)*x_ref[2]**o+(-0.716528412077)*x_ref[2]
        ref[1,1,1,0]=(-0.151416035033)*x_ref[0]**o+(0.0255347649787)*x_ref[0]+(-0.752987701889)*x_ref[1]**o+(-0.480969875813)*x_ref[1]+(-0.610514504179)*x_ref[2]**o+(0.959274757402)*x_ref[2]
        ref[1,1,1,1]=(0.76695706106)*x_ref[0]**o+(-0.0750965975512)*x_ref[0]+(0.846833900735)*x_ref[1]**o+(-0.105838465439)*x_ref[1]+(0.541027486457)*x_ref[2]**o+(0.152543017589)*x_ref[2]
        ref[1,1,2,0]=(0.0856977181865)*x_ref[0]**o+(-0.188139369706)*x_ref[0]+(0.936071565289)*x_ref[1]**o+(-0.3958713036)*x_ref[1]+(-0.350945878893)*x_ref[2]**o+(0.382302674883)*x_ref[2]
        ref[1,1,2,1]=(0.873054234789)*x_ref[0]**o+(-0.490869716146)*x_ref[0]+(0.46720526212)*x_ref[1]**o+(-0.334203412334)*x_ref[1]+(-0.81140895636)*x_ref[2]**o+(-0.973858971749)*x_ref[2]
        ref[1,2,0,0]=(-0.716676352443)*x_ref[0]**o+(0.659821396277)*x_ref[0]+(0.335964292954)*x_ref[1]**o+(0.11609664773)*x_ref[1]+(-0.889186749466)*x_ref[2]**o+(0.454231973946)*x_ref[2]
        ref[1,2,0,1]=(0.367048872648)*x_ref[0]**o+(0.276083713126)*x_ref[0]+(0.686639128043)*x_ref[1]**o+(-0.991414415293)*x_ref[1]+(0.25360515964)*x_ref[2]**o+(-0.377623315093)*x_ref[2]
        ref[1,2,1,0]=(-0.723089358777)*x_ref[0]**o+(-0.516690112406)*x_ref[0]+(0.262334273979)*x_ref[1]**o+(0.140773502537)*x_ref[1]+(0.447333972396)*x_ref[2]**o+(-0.551395692966)*x_ref[2]
        ref[1,2,1,1]=(0.738791182491)*x_ref[0]**o+(-0.978120626668)*x_ref[0]+(-0.217970433713)*x_ref[1]**o+(0.0181011067492)*x_ref[1]+(0.756322867523)*x_ref[2]**o+(0.495269195978)*x_ref[2]
        ref[1,2,2,0]=(-0.630506449854)*x_ref[0]**o+(-0.0865482364231)*x_ref[0]+(-0.415896403157)*x_ref[1]**o+(-0.712833149895)*x_ref[1]+(-0.180482789777)*x_ref[2]**o+(0.00788290661235)*x_ref[2]
        ref[1,2,2,1]=(0.0946913865377)*x_ref[0]**o+(0.752854827716)*x_ref[0]+(0.257846220843)*x_ref[1]**o+(-0.953926983578)*x_ref[1]+(-0.388505920682)*x_ref[2]**o+(-0.10751401251)*x_ref[2]
        ref[1,3,0,0]=(0.694014030611)*x_ref[0]**o+(0.0433103547797)*x_ref[0]+(-0.0805025225569)*x_ref[1]**o+(0.389052254754)*x_ref[1]+(-0.813210939614)*x_ref[2]**o+(-0.593788787896)*x_ref[2]
        ref[1,3,0,1]=(0.0101606353406)*x_ref[0]**o+(-0.849874258855)*x_ref[0]+(-0.0578973959381)*x_ref[1]**o+(0.942113140424)*x_ref[1]+(0.613716459216)*x_ref[2]**o+(0.72631176833)*x_ref[2]
        ref[1,3,1,0]=(0.264703608811)*x_ref[0]**o+(0.457962548326)*x_ref[0]+(-0.433795242673)*x_ref[1]**o+(0.650995712848)*x_ref[1]+(0.372471807967)*x_ref[2]**o+(0.00416636841047)*x_ref[2]
        ref[1,3,1,1]=(-0.73686451669)*x_ref[0]**o+(-0.0425614910776)*x_ref[0]+(-0.587820431311)*x_ref[1]**o+(-0.377628617833)*x_ref[1]+(-0.463964014406)*x_ref[2]**o+(0.283273874989)*x_ref[2]
        ref[1,3,2,0]=(-0.689341439143)*x_ref[0]**o+(-0.675861918889)*x_ref[0]+(-0.918622662352)*x_ref[1]**o+(0.740990261888)*x_ref[1]+(0.0454007677041)*x_ref[2]**o+(0.513623361074)*x_ref[2]
        ref[1,3,2,1]=(0.13229101848)*x_ref[0]**o+(0.171873799817)*x_ref[0]+(-0.154491174434)*x_ref[1]**o+(0.897976884089)*x_ref[1]+(-0.598886311821)*x_ref[2]**o+(0.738696710918)*x_ref[2]
        ref[1,4,0,0]=(0.59403636966)*x_ref[0]**o+(0.93856848969)*x_ref[0]+(0.80164849167)*x_ref[1]**o+(-0.753322979568)*x_ref[1]+(-0.107098073916)*x_ref[2]**o+(-0.623047487973)*x_ref[2]
        ref[1,4,0,1]=(0.291046099082)*x_ref[0]**o+(-0.437982471049)*x_ref[0]+(0.017395762947)*x_ref[1]**o+(0.852440593923)*x_ref[1]+(0.630825424359)*x_ref[2]**o+(0.419347491727)*x_ref[2]
        ref[1,4,1,0]=(-0.747962012701)*x_ref[0]**o+(-0.11622032094)*x_ref[0]+(0.324950549781)*x_ref[1]**o+(0.950911398327)*x_ref[1]+(-0.18098990205)*x_ref[2]**o+(-0.676060633522)*x_ref[2]
        ref[1,4,1,1]=(0.545224363128)*x_ref[0]**o+(0.753376367121)*x_ref[0]+(0.271426695753)*x_ref[1]**o+(-0.139309844675)*x_ref[1]+(-0.923434716375)*x_ref[2]**o+(-0.985988324389)*x_ref[2]
        ref[1,4,2,0]=(-0.532260822047)*x_ref[0]**o+(0.901407135724)*x_ref[0]+(0.937451330141)*x_ref[1]**o+(0.285217666559)*x_ref[1]+(-0.133028984824)*x_ref[2]**o+(0.993965133696)*x_ref[2]
        ref[1,4,2,1]=(0.2231815607)*x_ref[0]**o+(0.994845537582)*x_ref[0]+(0.0368727378195)*x_ref[1]**o+(0.511772933777)*x_ref[1]+(-0.871580215976)*x_ref[2]**o+(-0.587010057329)*x_ref[2]
        ref[2,0,0,0]=(-0.55149411913)*x_ref[0]**o+(-0.905541278559)*x_ref[0]+(0.461217373096)*x_ref[1]**o+(-0.53359853241)*x_ref[1]+(-0.666643290883)*x_ref[2]**o+(0.151495826926)*x_ref[2]
        ref[2,0,0,1]=(-0.980485062148)*x_ref[0]**o+(0.150165616162)*x_ref[0]+(0.59165725487)*x_ref[1]**o+(-0.243853536017)*x_ref[1]+(0.66816261686)*x_ref[2]**o+(0.2425696131)*x_ref[2]
        ref[2,0,1,0]=(-0.509337213758)*x_ref[0]**o+(0.652936706687)*x_ref[0]+(-0.849214338091)*x_ref[1]**o+(-0.523354123549)*x_ref[1]+(0.371515689887)*x_ref[2]**o+(0.799989815414)*x_ref[2]
        ref[2,0,1,1]=(0.461584544541)*x_ref[0]**o+(0.662931355954)*x_ref[0]+(0.744415634786)*x_ref[1]**o+(-0.016743801256)*x_ref[1]+(0.172945846325)*x_ref[2]**o+(-0.0889827089341)*x_ref[2]
        ref[2,0,2,0]=(0.506514791562)*x_ref[0]**o+(0.974477160694)*x_ref[0]+(-0.968485343167)*x_ref[1]**o+(0.427480013599)*x_ref[1]+(0.0460755879358)*x_ref[2]**o+(0.186500994948)*x_ref[2]
        ref[2,0,2,1]=(0.259831607993)*x_ref[0]**o+(-0.0141290444125)*x_ref[0]+(-0.968452177682)*x_ref[1]**o+(0.635619125009)*x_ref[1]+(-0.958848431827)*x_ref[2]**o+(-0.0977241523308)*x_ref[2]
        ref[2,1,0,0]=(0.204233222933)*x_ref[0]**o+(0.834739381877)*x_ref[0]+(0.19961185048)*x_ref[1]**o+(-0.238470380578)*x_ref[1]+(-0.749979713773)*x_ref[2]**o+(-0.268168886251)*x_ref[2]
        ref[2,1,0,1]=(0.998123643088)*x_ref[0]**o+(0.635823733041)*x_ref[0]+(-0.265447124419)*x_ref[1]**o+(-0.225582751946)*x_ref[1]+(0.91705937439)*x_ref[2]**o+(-0.739461510792)*x_ref[2]
        ref[2,1,1,0]=(0.0800715288646)*x_ref[0]**o+(-0.37677196659)*x_ref[0]+(-0.0737286350219)*x_ref[1]**o+(0.032076062933)*x_ref[1]+(0.245421912709)*x_ref[2]**o+(-0.8378023935)*x_ref[2]
        ref[2,1,1,1]=(-0.572376055603)*x_ref[0]**o+(0.8675468469)*x_ref[0]+(-0.533012889867)*x_ref[1]**o+(-0.247215658297)*x_ref[1]+(0.673144564693)*x_ref[2]**o+(-0.749136108985)*x_ref[2]
        ref[2,1,2,0]=(-0.661188192488)*x_ref[0]**o+(-0.924435593008)*x_ref[0]+(-0.791845808916)*x_ref[1]**o+(0.00133765351092)*x_ref[1]+(0.233676862873)*x_ref[2]**o+(0.981812737962)*x_ref[2]
        ref[2,1,2,1]=(0.332072445986)*x_ref[0]**o+(-0.909373798588)*x_ref[0]+(0.927025647819)*x_ref[1]**o+(-0.936348942617)*x_ref[1]+(0.251508592152)*x_ref[2]**o+(-0.468147973987)*x_ref[2]
        ref[2,2,0,0]=(0.892377893841)*x_ref[0]**o+(-0.147933293091)*x_ref[0]+(0.950491102076)*x_ref[1]**o+(0.595250355926)*x_ref[1]+(0.0169764649992)*x_ref[2]**o+(-0.683137752586)*x_ref[2]
        ref[2,2,0,1]=(0.00579397005645)*x_ref[0]**o+(0.627197373917)*x_ref[0]+(0.663574756196)*x_ref[1]**o+(0.934106767513)*x_ref[1]+(0.105844878022)*x_ref[2]**o+(0.513322917688)*x_ref[2]
        ref[2,2,1,0]=(-0.168776016133)*x_ref[0]**o+(-0.940071612944)*x_ref[0]+(0.0366053554566)*x_ref[1]**o+(-0.628010909771)*x_ref[1]+(0.250023045906)*x_ref[2]**o+(-0.00327322332166)*x_ref[2]
        ref[2,2,1,1]=(0.904155764376)*x_ref[0]**o+(-0.258018307306)*x_ref[0]+(0.65862820521)*x_ref[1]**o+(0.434480798515)*x_ref[1]+(0.4089419173)*x_ref[2]**o+(-0.518032695598)*x_ref[2]
        ref[2,2,2,0]=(-0.41470070958)*x_ref[0]**o+(-0.783186751952)*x_ref[0]+(0.548309561942)*x_ref[1]**o+(0.837130592803)*x_ref[1]+(-0.346808491848)*x_ref[2]**o+(0.165416040985)*x_ref[2]
        ref[2,2,2,1]=(0.997845909195)*x_ref[0]**o+(0.448428042158)*x_ref[0]+(-0.657050004311)*x_ref[1]**o+(0.493104851747)*x_ref[1]+(0.558315071197)*x_ref[2]**o+(0.206659432923)*x_ref[2]
        ref[2,3,0,0]=(0.433874360232)*x_ref[0]**o+(-0.0635646197131)*x_ref[0]+(-0.521690521523)*x_ref[1]**o+(-0.298130684888)*x_ref[1]+(-0.131905313644)*x_ref[2]**o+(0.64389451112)*x_ref[2]
        ref[2,3,0,1]=(0.457729315216)*x_ref[0]**o+(-0.419323242135)*x_ref[0]+(0.614544051135)*x_ref[1]**o+(-0.701671999275)*x_ref[1]+(0.905494300727)*x_ref[2]**o+(-0.565716985337)*x_ref[2]
        ref[2,3,1,0]=(0.256570728097)*x_ref[0]**o+(0.866349285103)*x_ref[0]+(-0.115768354896)*x_ref[1]**o+(-0.613044456549)*x_ref[1]+(-0.20074185156)*x_ref[2]**o+(-0.188403571776)*x_ref[2]
        ref[2,3,1,1]=(0.767479943414)*x_ref[0]**o+(0.884066215007)*x_ref[0]+(0.363855420056)*x_ref[1]**o+(0.919698981104)*x_ref[1]+(0.225747762317)*x_ref[2]**o+(-0.220362271104)*x_ref[2]
        ref[2,3,2,0]=(-0.746469501382)*x_ref[0]**o+(0.680520398001)*x_ref[0]+(0.926950996325)*x_ref[1]**o+(-0.122831303044)*x_ref[1]+(-0.27226561675)*x_ref[2]**o+(0.933304087951)*x_ref[2]
        ref[2,3,2,1]=(-0.0961434747216)*x_ref[0]**o+(0.987660516767)*x_ref[0]+(0.288175205045)*x_ref[1]**o+(0.683720820403)*x_ref[1]+(0.354403143642)*x_ref[2]**o+(-0.745694089132)*x_ref[2]
        ref[2,4,0,0]=(0.921750835986)*x_ref[0]**o+(-0.0358861598394)*x_ref[0]+(0.190129645259)*x_ref[1]**o+(0.377067803331)*x_ref[1]+(-0.0188615236582)*x_ref[2]**o+(0.804533888824)*x_ref[2]
        ref[2,4,0,1]=(-0.256132234484)*x_ref[0]**o+(-0.102661411017)*x_ref[0]+(-0.903465157009)*x_ref[1]**o+(0.50133310168)*x_ref[1]+(-0.0889102598619)*x_ref[2]**o+(-0.0310174238365)*x_ref[2]
        ref[2,4,1,0]=(0.58483798855)*x_ref[0]**o+(0.967943344305)*x_ref[0]+(-0.0701080005438)*x_ref[1]**o+(0.489476104406)*x_ref[1]+(0.615308427038)*x_ref[2]**o+(-0.38065505709)*x_ref[2]
        ref[2,4,1,1]=(0.647734130724)*x_ref[0]**o+(-0.930934392719)*x_ref[0]+(0.267370540491)*x_ref[1]**o+(-0.0175019424556)*x_ref[1]+(0.38593657129)*x_ref[2]**o+(-0.985922292359)*x_ref[2]
        ref[2,4,2,0]=(-0.289971479933)*x_ref[0]**o+(0.882014313157)*x_ref[0]+(0.147321067579)*x_ref[1]**o+(-0.176771838489)*x_ref[1]+(0.710916531544)*x_ref[2]**o+(-0.0856290170401)*x_ref[2]
        ref[2,4,2,1]=(-0.465930997733)*x_ref[0]**o+(-0.305563942094)*x_ref[0]+(-0.824535262061)*x_ref[1]**o+(-0.917892226972)*x_ref[1]+(0.624462871955)*x_ref[2]**o+(0.544736500065)*x_ref[2]
        ref[3,0,0,0]=(-0.673230815097)*x_ref[0]**o+(-0.726292712917)*x_ref[0]+(0.0531037021688)*x_ref[1]**o+(-0.467219030159)*x_ref[1]+(-0.0729861369718)*x_ref[2]**o+(-0.440394335015)*x_ref[2]
        ref[3,0,0,1]=(-0.358235295558)*x_ref[0]**o+(-0.301184713671)*x_ref[0]+(-0.496030649086)*x_ref[1]**o+(-0.125922184777)*x_ref[1]+(0.946715036973)*x_ref[2]**o+(0.846370337893)*x_ref[2]
        ref[3,0,1,0]=(-0.394175109612)*x_ref[0]**o+(-0.85137582307)*x_ref[0]+(-0.667804785044)*x_ref[1]**o+(0.156762109891)*x_ref[1]+(-0.616444700525)*x_ref[2]**o+(0.0738729399356)*x_ref[2]
        ref[3,0,1,1]=(-0.652875462534)*x_ref[0]**o+(-0.0605650192541)*x_ref[0]+(0.840539933823)*x_ref[1]**o+(-0.422819423301)*x_ref[1]+(0.113440226819)*x_ref[2]**o+(-0.915154532306)*x_ref[2]
        ref[3,0,2,0]=(-0.162863352328)*x_ref[0]**o+(-0.364461332292)*x_ref[0]+(-0.222181741398)*x_ref[1]**o+(-0.563668008365)*x_ref[1]+(0.118491770105)*x_ref[2]**o+(0.0809155053022)*x_ref[2]
        ref[3,0,2,1]=(-0.985278929437)*x_ref[0]**o+(-0.919440760157)*x_ref[0]+(-0.776910686)*x_ref[1]**o+(-0.884270488943)*x_ref[1]+(0.984279010363)*x_ref[2]**o+(-0.489503278582)*x_ref[2]
        ref[3,1,0,0]=(-0.806565449823)*x_ref[0]**o+(-0.0851385375281)*x_ref[0]+(-0.88851381773)*x_ref[1]**o+(-0.666539867528)*x_ref[1]+(0.154825531411)*x_ref[2]**o+(0.764069849144)*x_ref[2]
        ref[3,1,0,1]=(-0.222125783918)*x_ref[0]**o+(0.568035675467)*x_ref[0]+(0.470171778932)*x_ref[1]**o+(-0.307608710881)*x_ref[1]+(-0.836083742471)*x_ref[2]**o+(0.553597553381)*x_ref[2]
        ref[3,1,1,0]=(0.431237389847)*x_ref[0]**o+(-0.141391012471)*x_ref[0]+(0.726895025519)*x_ref[1]**o+(-0.527868751679)*x_ref[1]+(0.711761083137)*x_ref[2]**o+(0.996992864103)*x_ref[2]
        ref[3,1,1,1]=(-0.123885959924)*x_ref[0]**o+(0.877034475529)*x_ref[0]+(-0.0432492682928)*x_ref[1]**o+(-0.370315494618)*x_ref[1]+(0.243139719071)*x_ref[2]**o+(-0.684318634902)*x_ref[2]
        ref[3,1,2,0]=(-0.477214489355)*x_ref[0]**o+(0.340808576619)*x_ref[0]+(-0.0991802306558)*x_ref[1]**o+(0.462599821908)*x_ref[1]+(0.091768281564)*x_ref[2]**o+(0.0315552838839)*x_ref[2]
        ref[3,1,2,1]=(0.676189142894)*x_ref[0]**o+(0.510136386913)*x_ref[0]+(-0.558579041952)*x_ref[1]**o+(0.961615047877)*x_ref[1]+(0.510717386124)*x_ref[2]**o+(0.774430183432)*x_ref[2]
        ref[3,2,0,0]=(0.158370232444)*x_ref[0]**o+(0.209424855992)*x_ref[0]+(0.691438753013)*x_ref[1]**o+(-0.781594298738)*x_ref[1]+(-0.0188536627517)*x_ref[2]**o+(0.136400798995)*x_ref[2]
        ref[3,2,0,1]=(-0.440674414328)*x_ref[0]**o+(0.118228612687)*x_ref[0]+(-0.700026492613)*x_ref[1]**o+(-0.981341215714)*x_ref[1]+(-0.172847519578)*x_ref[2]**o+(-0.423095202172)*x_ref[2]
        ref[3,2,1,0]=(-0.0694960391695)*x_ref[0]**o+(-0.66448610135)*x_ref[0]+(-0.177572703748)*x_ref[1]**o+(-0.060488161727)*x_ref[1]+(0.306021853885)*x_ref[2]**o+(0.0730338716187)*x_ref[2]
        ref[3,2,1,1]=(-0.360662154547)*x_ref[0]**o+(0.577679359648)*x_ref[0]+(-0.515313534966)*x_ref[1]**o+(0.217352687358)*x_ref[1]+(0.728367577882)*x_ref[2]**o+(0.38505849834)*x_ref[2]
        ref[3,2,2,0]=(0.568671861232)*x_ref[0]**o+(-0.0792490917882)*x_ref[0]+(0.0188002885229)*x_ref[1]**o+(-0.956862575759)*x_ref[1]+(-0.844332442983)*x_ref[2]**o+(0.670792928042)*x_ref[2]
        ref[3,2,2,1]=(0.833099204713)*x_ref[0]**o+(0.368482169933)*x_ref[0]+(-0.926002498892)*x_ref[1]**o+(-0.130646446568)*x_ref[1]+(0.81164928914)*x_ref[2]**o+(-0.641079919927)*x_ref[2]
        ref[3,3,0,0]=(-0.277769350989)*x_ref[0]**o+(0.753452018649)*x_ref[0]+(-0.766322344505)*x_ref[1]**o+(-0.679108224612)*x_ref[1]+(0.0799499795424)*x_ref[2]**o+(0.255751308975)*x_ref[2]
        ref[3,3,0,1]=(0.429176753733)*x_ref[0]**o+(0.123106050259)*x_ref[0]+(-0.169037728992)*x_ref[1]**o+(0.749245933687)*x_ref[1]+(0.560140859355)*x_ref[2]**o+(-0.937091840021)*x_ref[2]
        ref[3,3,1,0]=(0.648792557593)*x_ref[0]**o+(-0.057555187513)*x_ref[0]+(0.883290331726)*x_ref[1]**o+(0.539276905995)*x_ref[1]+(0.473614408583)*x_ref[2]**o+(0.723258340333)*x_ref[2]
        ref[3,3,1,1]=(0.527298833572)*x_ref[0]**o+(0.0753824058279)*x_ref[0]+(0.719196099257)*x_ref[1]**o+(-0.0124603350035)*x_ref[1]+(-0.964484650253)*x_ref[2]**o+(0.404444269694)*x_ref[2]
        ref[3,3,2,0]=(-0.403640677462)*x_ref[0]**o+(0.906195964766)*x_ref[0]+(-0.561419078404)*x_ref[1]**o+(0.645408717838)*x_ref[1]+(0.671624799609)*x_ref[2]**o+(-0.573901868621)*x_ref[2]
        ref[3,3,2,1]=(-0.288117182165)*x_ref[0]**o+(-0.391003584945)*x_ref[0]+(-0.555037182038)*x_ref[1]**o+(0.299362209017)*x_ref[1]+(-0.552429724326)*x_ref[2]**o+(-0.416923306774)*x_ref[2]
        ref[3,4,0,0]=(-0.19594719192)*x_ref[0]**o+(0.150898934692)*x_ref[0]+(0.210935232044)*x_ref[1]**o+(-0.319764293347)*x_ref[1]+(-0.642450751195)*x_ref[2]**o+(0.22231085168)*x_ref[2]
        ref[3,4,0,1]=(-0.0514589460225)*x_ref[0]**o+(0.0366025646754)*x_ref[0]+(0.416280576691)*x_ref[1]**o+(0.609178063951)*x_ref[1]+(0.835611677062)*x_ref[2]**o+(-0.131630571478)*x_ref[2]
        ref[3,4,1,0]=(0.171509152214)*x_ref[0]**o+(-0.761188271093)*x_ref[0]+(-0.915951802896)*x_ref[1]**o+(-0.451796709915)*x_ref[1]+(-0.979847828162)*x_ref[2]**o+(0.0566453393315)*x_ref[2]
        ref[3,4,1,1]=(0.690427998102)*x_ref[0]**o+(-0.358318625695)*x_ref[0]+(0.330270871613)*x_ref[1]**o+(-0.724392522137)*x_ref[1]+(-0.535474623489)*x_ref[2]**o+(-0.12341743885)*x_ref[2]
        ref[3,4,2,0]=(0.488090753426)*x_ref[0]**o+(-0.352417078372)*x_ref[0]+(0.971988099319)*x_ref[1]**o+(-0.288929516009)*x_ref[1]+(0.612576734094)*x_ref[2]**o+(-0.43799514603)*x_ref[2]
        ref[3,4,2,1]=(-0.098768101356)*x_ref[0]**o+(0.123300661895)*x_ref[0]+(-0.0489724280789)*x_ref[1]**o+(-0.422508669137)*x_ref[1]+(-0.861483104139)*x_ref[2]**o+(0.950198316001)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.724373964817)*x[0]**o+(-0.0560850774901)*x[0]+(0.201815078281)*x[1]**o+(-0.721683697048)*x[1]
        ref=(0.724373964817)*x_ref[0]**o+(-0.0560850774901)*x_ref[0]+(0.201815078281)*x_ref[1]**o+(-0.721683697048)*x_ref[1]
      else:
        arg=(0.390296010052)*x[0]**o+(-0.279464234123)*x[0]+(-0.791334947488)*x[1]**o+(-0.0210629346704)*x[1]+(0.0816556526012)*x[2]**o+(0.77326269137)*x[2]
        ref=(0.390296010052)*x_ref[0]**o+(-0.279464234123)*x_ref[0]+(-0.791334947488)*x_ref[1]**o+(-0.0210629346704)*x_ref[1]+(0.0816556526012)*x_ref[2]**o+(0.77326269137)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.645585823917)*x[0]**o+(-0.203443728436)*x[0]+(0.278126891692)*x[1]**o+(0.922093564911)*x[1]
        arg[1]=(0.23370137029)*x[0]**o+(0.84371173541)*x[0]+(0.99297574548)*x[1]**o+(0.72481785931)*x[1]
        ref[0]=(0.645585823917)*x_ref[0]**o+(-0.203443728436)*x_ref[0]+(0.278126891692)*x_ref[1]**o+(0.922093564911)*x_ref[1]
        ref[1]=(0.23370137029)*x_ref[0]**o+(0.84371173541)*x_ref[0]+(0.99297574548)*x_ref[1]**o+(0.72481785931)*x_ref[1]
      else:
        arg[0]=(0.159998049951)*x[0]**o+(-0.863367327023)*x[0]+(-0.0628498246638)*x[1]**o+(0.685139358273)*x[1]+(0.783213360641)*x[2]**o+(0.397526421338)*x[2]
        arg[1]=(0.165519279053)*x[0]**o+(-0.0936018967661)*x[0]+(0.863047696619)*x[1]**o+(0.653118186434)*x[1]+(-0.408224131944)*x[2]**o+(0.82400132854)*x[2]
        ref[0]=(0.159998049951)*x_ref[0]**o+(-0.863367327023)*x_ref[0]+(-0.0628498246638)*x_ref[1]**o+(0.685139358273)*x_ref[1]+(0.783213360641)*x_ref[2]**o+(0.397526421338)*x_ref[2]
        ref[1]=(0.165519279053)*x_ref[0]**o+(-0.0936018967661)*x_ref[0]+(0.863047696619)*x_ref[1]**o+(0.653118186434)*x_ref[1]+(-0.408224131944)*x_ref[2]**o+(0.82400132854)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.490709482882)*x[0]**o+(0.518382584887)*x[0]+(-0.940707425885)*x[1]**o+(0.224578726673)*x[1]
        arg[0,1]=(-0.747621141516)*x[0]**o+(-0.217452489097)*x[0]+(0.929120128008)*x[1]**o+(0.55263700209)*x[1]
        arg[0,2]=(-0.183545119672)*x[0]**o+(0.148192386895)*x[0]+(-0.22046309708)*x[1]**o+(0.433620446668)*x[1]
        arg[0,3]=(-0.0724840266959)*x[0]**o+(0.835144620602)*x[0]+(-0.962971534144)*x[1]**o+(-0.149681811534)*x[1]
        arg[0,4]=(-0.380150192228)*x[0]**o+(-0.309494603814)*x[0]+(-0.217425635182)*x[1]**o+(0.93567286482)*x[1]
        arg[1,0]=(0.0367921563947)*x[0]**o+(0.220217625576)*x[0]+(-0.120595819238)*x[1]**o+(0.308108793279)*x[1]
        arg[1,1]=(-0.142860237109)*x[0]**o+(-0.789758888253)*x[0]+(0.748085258229)*x[1]**o+(0.784767485883)*x[1]
        arg[1,2]=(-0.154690818845)*x[0]**o+(-0.678458370784)*x[0]+(0.750742996193)*x[1]**o+(-0.350199095531)*x[1]
        arg[1,3]=(-0.893295215013)*x[0]**o+(0.0295805814985)*x[0]+(0.265893730887)*x[1]**o+(0.435487860418)*x[1]
        arg[1,4]=(-0.302701704727)*x[0]**o+(0.618274652266)*x[0]+(-0.688386407308)*x[1]**o+(-0.832515064466)*x[1]
        arg[2,0]=(-0.926658745469)*x[0]**o+(0.548174350496)*x[0]+(-0.350700366641)*x[1]**o+(0.530188602011)*x[1]
        arg[2,1]=(0.900965385167)*x[0]**o+(0.0234051845124)*x[0]+(-0.00783834941699)*x[1]**o+(-0.766352156598)*x[1]
        arg[2,2]=(0.694733282444)*x[0]**o+(-0.0644519234093)*x[0]+(0.0493475528391)*x[1]**o+(-0.589101546403)*x[1]
        arg[2,3]=(-0.100333102737)*x[0]**o+(-0.633740106784)*x[0]+(0.408697826034)*x[1]**o+(-0.429059439472)*x[1]
        arg[2,4]=(0.635563324404)*x[0]**o+(0.955881346028)*x[0]+(0.0755013269452)*x[1]**o+(-0.743681282337)*x[1]
        arg[3,0]=(0.604891134834)*x[0]**o+(0.942047409702)*x[0]+(-0.00544373299112)*x[1]**o+(0.447523533166)*x[1]
        arg[3,1]=(-0.980967446837)*x[0]**o+(0.019824933239)*x[0]+(-0.826876930392)*x[1]**o+(0.113548968693)*x[1]
        arg[3,2]=(-0.731417395373)*x[0]**o+(-0.275072472108)*x[0]+(0.933867711492)*x[1]**o+(-0.444847972819)*x[1]
        arg[3,3]=(0.00121172413626)*x[0]**o+(-0.171073908218)*x[0]+(0.913182873344)*x[1]**o+(0.960256570926)*x[1]
        arg[3,4]=(-0.395134372727)*x[0]**o+(0.646385082409)*x[0]+(0.17387019219)*x[1]**o+(-0.287550462397)*x[1]
        ref[0,0]=(0.490709482882)*x_ref[0]**o+(0.518382584887)*x_ref[0]+(-0.940707425885)*x_ref[1]**o+(0.224578726673)*x_ref[1]
        ref[0,1]=(-0.747621141516)*x_ref[0]**o+(-0.217452489097)*x_ref[0]+(0.929120128008)*x_ref[1]**o+(0.55263700209)*x_ref[1]
        ref[0,2]=(-0.183545119672)*x_ref[0]**o+(0.148192386895)*x_ref[0]+(-0.22046309708)*x_ref[1]**o+(0.433620446668)*x_ref[1]
        ref[0,3]=(-0.0724840266959)*x_ref[0]**o+(0.835144620602)*x_ref[0]+(-0.962971534144)*x_ref[1]**o+(-0.149681811534)*x_ref[1]
        ref[0,4]=(-0.380150192228)*x_ref[0]**o+(-0.309494603814)*x_ref[0]+(-0.217425635182)*x_ref[1]**o+(0.93567286482)*x_ref[1]
        ref[1,0]=(0.0367921563947)*x_ref[0]**o+(0.220217625576)*x_ref[0]+(-0.120595819238)*x_ref[1]**o+(0.308108793279)*x_ref[1]
        ref[1,1]=(-0.142860237109)*x_ref[0]**o+(-0.789758888253)*x_ref[0]+(0.748085258229)*x_ref[1]**o+(0.784767485883)*x_ref[1]
        ref[1,2]=(-0.154690818845)*x_ref[0]**o+(-0.678458370784)*x_ref[0]+(0.750742996193)*x_ref[1]**o+(-0.350199095531)*x_ref[1]
        ref[1,3]=(-0.893295215013)*x_ref[0]**o+(0.0295805814985)*x_ref[0]+(0.265893730887)*x_ref[1]**o+(0.435487860418)*x_ref[1]
        ref[1,4]=(-0.302701704727)*x_ref[0]**o+(0.618274652266)*x_ref[0]+(-0.688386407308)*x_ref[1]**o+(-0.832515064466)*x_ref[1]
        ref[2,0]=(-0.926658745469)*x_ref[0]**o+(0.548174350496)*x_ref[0]+(-0.350700366641)*x_ref[1]**o+(0.530188602011)*x_ref[1]
        ref[2,1]=(0.900965385167)*x_ref[0]**o+(0.0234051845124)*x_ref[0]+(-0.00783834941699)*x_ref[1]**o+(-0.766352156598)*x_ref[1]
        ref[2,2]=(0.694733282444)*x_ref[0]**o+(-0.0644519234093)*x_ref[0]+(0.0493475528391)*x_ref[1]**o+(-0.589101546403)*x_ref[1]
        ref[2,3]=(-0.100333102737)*x_ref[0]**o+(-0.633740106784)*x_ref[0]+(0.408697826034)*x_ref[1]**o+(-0.429059439472)*x_ref[1]
        ref[2,4]=(0.635563324404)*x_ref[0]**o+(0.955881346028)*x_ref[0]+(0.0755013269452)*x_ref[1]**o+(-0.743681282337)*x_ref[1]
        ref[3,0]=(0.604891134834)*x_ref[0]**o+(0.942047409702)*x_ref[0]+(-0.00544373299112)*x_ref[1]**o+(0.447523533166)*x_ref[1]
        ref[3,1]=(-0.980967446837)*x_ref[0]**o+(0.019824933239)*x_ref[0]+(-0.826876930392)*x_ref[1]**o+(0.113548968693)*x_ref[1]
        ref[3,2]=(-0.731417395373)*x_ref[0]**o+(-0.275072472108)*x_ref[0]+(0.933867711492)*x_ref[1]**o+(-0.444847972819)*x_ref[1]
        ref[3,3]=(0.00121172413626)*x_ref[0]**o+(-0.171073908218)*x_ref[0]+(0.913182873344)*x_ref[1]**o+(0.960256570926)*x_ref[1]
        ref[3,4]=(-0.395134372727)*x_ref[0]**o+(0.646385082409)*x_ref[0]+(0.17387019219)*x_ref[1]**o+(-0.287550462397)*x_ref[1]
      else:
        arg[0,0]=(0.386545340192)*x[0]**o+(-0.397636296285)*x[0]+(-0.849437816516)*x[1]**o+(-0.83083406493)*x[1]+(0.479216612031)*x[2]**o+(0.0903787541308)*x[2]
        arg[0,1]=(-0.331651855726)*x[0]**o+(0.849421626918)*x[0]+(0.528368477362)*x[1]**o+(-0.015122214824)*x[1]+(0.494310578452)*x[2]**o+(0.190239928161)*x[2]
        arg[0,2]=(-0.662389759646)*x[0]**o+(0.204378343962)*x[0]+(0.486802744897)*x[1]**o+(0.822174784922)*x[1]+(-0.309203812647)*x[2]**o+(-0.101665040799)*x[2]
        arg[0,3]=(0.609106341828)*x[0]**o+(0.658404088333)*x[0]+(-0.0092991120732)*x[1]**o+(-0.73794934348)*x[1]+(0.270245135914)*x[2]**o+(-0.594067838407)*x[2]
        arg[0,4]=(-0.291752213357)*x[0]**o+(-0.596751343066)*x[0]+(-0.109196903078)*x[1]**o+(-0.868610307708)*x[1]+(0.622541308544)*x[2]**o+(0.482202762224)*x[2]
        arg[1,0]=(-0.113988933227)*x[0]**o+(-0.846328804695)*x[0]+(0.691769553626)*x[1]**o+(-0.377441645148)*x[1]+(-0.769857970682)*x[2]**o+(-0.583083729037)*x[2]
        arg[1,1]=(-0.635333757145)*x[0]**o+(0.194761117849)*x[0]+(-0.268287296731)*x[1]**o+(0.463906930647)*x[1]+(0.768293821029)*x[2]**o+(-0.532824906687)*x[2]
        arg[1,2]=(0.355380090138)*x[0]**o+(-0.181191575848)*x[0]+(-0.558667766467)*x[1]**o+(0.561707671872)*x[1]+(-0.427004232725)*x[2]**o+(-0.996564299989)*x[2]
        arg[1,3]=(-0.891638892893)*x[0]**o+(0.0867350850512)*x[0]+(0.880258390584)*x[1]**o+(-0.798123407525)*x[1]+(-0.571505202201)*x[2]**o+(0.35374245063)*x[2]
        arg[1,4]=(0.543632496941)*x[0]**o+(-0.630363803582)*x[0]+(0.38107312058)*x[1]**o+(0.222401831883)*x[1]+(-0.932212080175)*x[2]**o+(-0.0482899826174)*x[2]
        arg[2,0]=(0.46514308325)*x[0]**o+(-0.907464652384)*x[0]+(-0.706105595148)*x[1]**o+(0.486989958675)*x[1]+(0.422702683485)*x[2]**o+(0.640916925932)*x[2]
        arg[2,1]=(0.912808317479)*x[0]**o+(-0.860779312763)*x[0]+(-0.915364956133)*x[1]**o+(-0.314186117224)*x[1]+(-0.860814410035)*x[2]**o+(0.837570442722)*x[2]
        arg[2,2]=(0.00932917919464)*x[0]**o+(0.393773283049)*x[0]+(0.699998572668)*x[1]**o+(0.153407885795)*x[1]+(-0.166245189281)*x[2]**o+(-0.430541214236)*x[2]
        arg[2,3]=(-0.00127511229816)*x[0]**o+(-0.327443174924)*x[0]+(0.61881322537)*x[1]**o+(-0.725854102962)*x[1]+(0.650649062427)*x[2]**o+(0.246373184041)*x[2]
        arg[2,4]=(-0.530580898049)*x[0]**o+(-0.208506132901)*x[0]+(-0.102049332015)*x[1]**o+(-0.537037644189)*x[1]+(0.740113470729)*x[2]**o+(0.12094405264)*x[2]
        arg[3,0]=(-0.103423602042)*x[0]**o+(-0.535742323666)*x[0]+(0.678663019405)*x[1]**o+(0.579375950057)*x[1]+(0.323627490052)*x[2]**o+(-0.681323301356)*x[2]
        arg[3,1]=(0.670239087519)*x[0]**o+(0.209657968128)*x[0]+(-0.357490300727)*x[1]**o+(-0.441776782321)*x[1]+(0.00927389037562)*x[2]**o+(0.633527599723)*x[2]
        arg[3,2]=(0.415569606138)*x[0]**o+(-0.422211616068)*x[0]+(-0.695086235645)*x[1]**o+(0.709158289848)*x[1]+(-0.523450845183)*x[2]**o+(0.555778994398)*x[2]
        arg[3,3]=(0.0188993892075)*x[0]**o+(0.611088737828)*x[0]+(0.0313795167921)*x[1]**o+(0.0842651755016)*x[1]+(0.288414199415)*x[2]**o+(0.695910780235)*x[2]
        arg[3,4]=(-0.500338800879)*x[0]**o+(0.0984192369328)*x[0]+(0.71163172992)*x[1]**o+(-0.102123116139)*x[1]+(0.954613955678)*x[2]**o+(0.384872409422)*x[2]
        ref[0,0]=(0.386545340192)*x_ref[0]**o+(-0.397636296285)*x_ref[0]+(-0.849437816516)*x_ref[1]**o+(-0.83083406493)*x_ref[1]+(0.479216612031)*x_ref[2]**o+(0.0903787541308)*x_ref[2]
        ref[0,1]=(-0.331651855726)*x_ref[0]**o+(0.849421626918)*x_ref[0]+(0.528368477362)*x_ref[1]**o+(-0.015122214824)*x_ref[1]+(0.494310578452)*x_ref[2]**o+(0.190239928161)*x_ref[2]
        ref[0,2]=(-0.662389759646)*x_ref[0]**o+(0.204378343962)*x_ref[0]+(0.486802744897)*x_ref[1]**o+(0.822174784922)*x_ref[1]+(-0.309203812647)*x_ref[2]**o+(-0.101665040799)*x_ref[2]
        ref[0,3]=(0.609106341828)*x_ref[0]**o+(0.658404088333)*x_ref[0]+(-0.0092991120732)*x_ref[1]**o+(-0.73794934348)*x_ref[1]+(0.270245135914)*x_ref[2]**o+(-0.594067838407)*x_ref[2]
        ref[0,4]=(-0.291752213357)*x_ref[0]**o+(-0.596751343066)*x_ref[0]+(-0.109196903078)*x_ref[1]**o+(-0.868610307708)*x_ref[1]+(0.622541308544)*x_ref[2]**o+(0.482202762224)*x_ref[2]
        ref[1,0]=(-0.113988933227)*x_ref[0]**o+(-0.846328804695)*x_ref[0]+(0.691769553626)*x_ref[1]**o+(-0.377441645148)*x_ref[1]+(-0.769857970682)*x_ref[2]**o+(-0.583083729037)*x_ref[2]
        ref[1,1]=(-0.635333757145)*x_ref[0]**o+(0.194761117849)*x_ref[0]+(-0.268287296731)*x_ref[1]**o+(0.463906930647)*x_ref[1]+(0.768293821029)*x_ref[2]**o+(-0.532824906687)*x_ref[2]
        ref[1,2]=(0.355380090138)*x_ref[0]**o+(-0.181191575848)*x_ref[0]+(-0.558667766467)*x_ref[1]**o+(0.561707671872)*x_ref[1]+(-0.427004232725)*x_ref[2]**o+(-0.996564299989)*x_ref[2]
        ref[1,3]=(-0.891638892893)*x_ref[0]**o+(0.0867350850512)*x_ref[0]+(0.880258390584)*x_ref[1]**o+(-0.798123407525)*x_ref[1]+(-0.571505202201)*x_ref[2]**o+(0.35374245063)*x_ref[2]
        ref[1,4]=(0.543632496941)*x_ref[0]**o+(-0.630363803582)*x_ref[0]+(0.38107312058)*x_ref[1]**o+(0.222401831883)*x_ref[1]+(-0.932212080175)*x_ref[2]**o+(-0.0482899826174)*x_ref[2]
        ref[2,0]=(0.46514308325)*x_ref[0]**o+(-0.907464652384)*x_ref[0]+(-0.706105595148)*x_ref[1]**o+(0.486989958675)*x_ref[1]+(0.422702683485)*x_ref[2]**o+(0.640916925932)*x_ref[2]
        ref[2,1]=(0.912808317479)*x_ref[0]**o+(-0.860779312763)*x_ref[0]+(-0.915364956133)*x_ref[1]**o+(-0.314186117224)*x_ref[1]+(-0.860814410035)*x_ref[2]**o+(0.837570442722)*x_ref[2]
        ref[2,2]=(0.00932917919464)*x_ref[0]**o+(0.393773283049)*x_ref[0]+(0.699998572668)*x_ref[1]**o+(0.153407885795)*x_ref[1]+(-0.166245189281)*x_ref[2]**o+(-0.430541214236)*x_ref[2]
        ref[2,3]=(-0.00127511229816)*x_ref[0]**o+(-0.327443174924)*x_ref[0]+(0.61881322537)*x_ref[1]**o+(-0.725854102962)*x_ref[1]+(0.650649062427)*x_ref[2]**o+(0.246373184041)*x_ref[2]
        ref[2,4]=(-0.530580898049)*x_ref[0]**o+(-0.208506132901)*x_ref[0]+(-0.102049332015)*x_ref[1]**o+(-0.537037644189)*x_ref[1]+(0.740113470729)*x_ref[2]**o+(0.12094405264)*x_ref[2]
        ref[3,0]=(-0.103423602042)*x_ref[0]**o+(-0.535742323666)*x_ref[0]+(0.678663019405)*x_ref[1]**o+(0.579375950057)*x_ref[1]+(0.323627490052)*x_ref[2]**o+(-0.681323301356)*x_ref[2]
        ref[3,1]=(0.670239087519)*x_ref[0]**o+(0.209657968128)*x_ref[0]+(-0.357490300727)*x_ref[1]**o+(-0.441776782321)*x_ref[1]+(0.00927389037562)*x_ref[2]**o+(0.633527599723)*x_ref[2]
        ref[3,2]=(0.415569606138)*x_ref[0]**o+(-0.422211616068)*x_ref[0]+(-0.695086235645)*x_ref[1]**o+(0.709158289848)*x_ref[1]+(-0.523450845183)*x_ref[2]**o+(0.555778994398)*x_ref[2]
        ref[3,3]=(0.0188993892075)*x_ref[0]**o+(0.611088737828)*x_ref[0]+(0.0313795167921)*x_ref[1]**o+(0.0842651755016)*x_ref[1]+(0.288414199415)*x_ref[2]**o+(0.695910780235)*x_ref[2]
        ref[3,4]=(-0.500338800879)*x_ref[0]**o+(0.0984192369328)*x_ref[0]+(0.71163172992)*x_ref[1]**o+(-0.102123116139)*x_ref[1]+(0.954613955678)*x_ref[2]**o+(0.384872409422)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.317274364174)*x[0]**o+(-0.565307843766)*x[0]+(-0.367874638208)*x[1]**o+(0.0029581033546)*x[1]
        arg[0,0,1]=(0.832728755858)*x[0]**o+(0.245271193641)*x[0]+(-0.727130029377)*x[1]**o+(0.137912975829)*x[1]
        arg[0,1,0]=(-0.634508968932)*x[0]**o+(-0.864200861637)*x[0]+(-0.566985971867)*x[1]**o+(0.458187145519)*x[1]
        arg[0,1,1]=(0.133772683885)*x[0]**o+(-0.630800341215)*x[0]+(0.958934971335)*x[1]**o+(-0.15839182257)*x[1]
        arg[1,0,0]=(-0.0948724752087)*x[0]**o+(-0.25589337469)*x[0]+(-0.315713312867)*x[1]**o+(0.223487524432)*x[1]
        arg[1,0,1]=(0.794634271698)*x[0]**o+(-0.505797079712)*x[0]+(0.54367349338)*x[1]**o+(0.610149581933)*x[1]
        arg[1,1,0]=(-0.447039667341)*x[0]**o+(-0.917009838715)*x[0]+(0.525331113522)*x[1]**o+(-0.626176620973)*x[1]
        arg[1,1,1]=(0.0749028334839)*x[0]**o+(0.106950341725)*x[0]+(-0.375321288987)*x[1]**o+(0.510260762626)*x[1]
        arg[2,0,0]=(-0.286661970669)*x[0]**o+(0.584169107201)*x[0]+(0.189852415568)*x[1]**o+(-0.685695910848)*x[1]
        arg[2,0,1]=(-0.689509796724)*x[0]**o+(-0.316067007053)*x[0]+(0.93758916753)*x[1]**o+(0.987368705364)*x[1]
        arg[2,1,0]=(-0.411695937554)*x[0]**o+(0.858528628791)*x[0]+(-0.579361651418)*x[1]**o+(-0.179617869494)*x[1]
        arg[2,1,1]=(0.871441175049)*x[0]**o+(0.445176827889)*x[0]+(0.282613647019)*x[1]**o+(-0.0414230375843)*x[1]
        arg[3,0,0]=(-0.242136375691)*x[0]**o+(-0.91917484201)*x[0]+(-0.88026838607)*x[1]**o+(0.363309786006)*x[1]
        arg[3,0,1]=(-0.991221762386)*x[0]**o+(-0.0149280738883)*x[0]+(0.420322977563)*x[1]**o+(-0.0533076687981)*x[1]
        arg[3,1,0]=(-0.889003524303)*x[0]**o+(-0.023519502168)*x[0]+(0.826015855889)*x[1]**o+(0.0568467842617)*x[1]
        arg[3,1,1]=(0.608567002824)*x[0]**o+(-0.531288593441)*x[0]+(0.0886678034719)*x[1]**o+(-0.732353220319)*x[1]
        arg[4,0,0]=(-0.0537755786538)*x[0]**o+(-0.475886396481)*x[0]+(-0.153787571145)*x[1]**o+(0.689850737569)*x[1]
        arg[4,0,1]=(0.413017674549)*x[0]**o+(-0.993353261494)*x[0]+(-0.944597571268)*x[1]**o+(0.566423551758)*x[1]
        arg[4,1,0]=(-0.716864417268)*x[0]**o+(-0.871574272865)*x[0]+(0.394071572628)*x[1]**o+(0.452903048597)*x[1]
        arg[4,1,1]=(0.861790785883)*x[0]**o+(-0.511789401943)*x[0]+(-0.272452527007)*x[1]**o+(-0.379021521619)*x[1]
        arg[5,0,0]=(0.509668147054)*x[0]**o+(0.919355584411)*x[0]+(-0.673215166917)*x[1]**o+(0.766459631931)*x[1]
        arg[5,0,1]=(0.177006227916)*x[0]**o+(-0.275118477)*x[0]+(0.0302367843701)*x[1]**o+(-0.761226360328)*x[1]
        arg[5,1,0]=(0.013432534638)*x[0]**o+(-0.911626567027)*x[0]+(-0.710520442529)*x[1]**o+(-0.882317448423)*x[1]
        arg[5,1,1]=(0.595583093615)*x[0]**o+(0.163527905156)*x[0]+(-0.122165051333)*x[1]**o+(-0.526359288181)*x[1]
        ref[0,0,0]=(0.317274364174)*x_ref[0]**o+(-0.565307843766)*x_ref[0]+(-0.367874638208)*x_ref[1]**o+(0.0029581033546)*x_ref[1]
        ref[0,0,1]=(0.832728755858)*x_ref[0]**o+(0.245271193641)*x_ref[0]+(-0.727130029377)*x_ref[1]**o+(0.137912975829)*x_ref[1]
        ref[0,1,0]=(-0.634508968932)*x_ref[0]**o+(-0.864200861637)*x_ref[0]+(-0.566985971867)*x_ref[1]**o+(0.458187145519)*x_ref[1]
        ref[0,1,1]=(0.133772683885)*x_ref[0]**o+(-0.630800341215)*x_ref[0]+(0.958934971335)*x_ref[1]**o+(-0.15839182257)*x_ref[1]
        ref[1,0,0]=(-0.0948724752087)*x_ref[0]**o+(-0.25589337469)*x_ref[0]+(-0.315713312867)*x_ref[1]**o+(0.223487524432)*x_ref[1]
        ref[1,0,1]=(0.794634271698)*x_ref[0]**o+(-0.505797079712)*x_ref[0]+(0.54367349338)*x_ref[1]**o+(0.610149581933)*x_ref[1]
        ref[1,1,0]=(-0.447039667341)*x_ref[0]**o+(-0.917009838715)*x_ref[0]+(0.525331113522)*x_ref[1]**o+(-0.626176620973)*x_ref[1]
        ref[1,1,1]=(0.0749028334839)*x_ref[0]**o+(0.106950341725)*x_ref[0]+(-0.375321288987)*x_ref[1]**o+(0.510260762626)*x_ref[1]
        ref[2,0,0]=(-0.286661970669)*x_ref[0]**o+(0.584169107201)*x_ref[0]+(0.189852415568)*x_ref[1]**o+(-0.685695910848)*x_ref[1]
        ref[2,0,1]=(-0.689509796724)*x_ref[0]**o+(-0.316067007053)*x_ref[0]+(0.93758916753)*x_ref[1]**o+(0.987368705364)*x_ref[1]
        ref[2,1,0]=(-0.411695937554)*x_ref[0]**o+(0.858528628791)*x_ref[0]+(-0.579361651418)*x_ref[1]**o+(-0.179617869494)*x_ref[1]
        ref[2,1,1]=(0.871441175049)*x_ref[0]**o+(0.445176827889)*x_ref[0]+(0.282613647019)*x_ref[1]**o+(-0.0414230375843)*x_ref[1]
        ref[3,0,0]=(-0.242136375691)*x_ref[0]**o+(-0.91917484201)*x_ref[0]+(-0.88026838607)*x_ref[1]**o+(0.363309786006)*x_ref[1]
        ref[3,0,1]=(-0.991221762386)*x_ref[0]**o+(-0.0149280738883)*x_ref[0]+(0.420322977563)*x_ref[1]**o+(-0.0533076687981)*x_ref[1]
        ref[3,1,0]=(-0.889003524303)*x_ref[0]**o+(-0.023519502168)*x_ref[0]+(0.826015855889)*x_ref[1]**o+(0.0568467842617)*x_ref[1]
        ref[3,1,1]=(0.608567002824)*x_ref[0]**o+(-0.531288593441)*x_ref[0]+(0.0886678034719)*x_ref[1]**o+(-0.732353220319)*x_ref[1]
        ref[4,0,0]=(-0.0537755786538)*x_ref[0]**o+(-0.475886396481)*x_ref[0]+(-0.153787571145)*x_ref[1]**o+(0.689850737569)*x_ref[1]
        ref[4,0,1]=(0.413017674549)*x_ref[0]**o+(-0.993353261494)*x_ref[0]+(-0.944597571268)*x_ref[1]**o+(0.566423551758)*x_ref[1]
        ref[4,1,0]=(-0.716864417268)*x_ref[0]**o+(-0.871574272865)*x_ref[0]+(0.394071572628)*x_ref[1]**o+(0.452903048597)*x_ref[1]
        ref[4,1,1]=(0.861790785883)*x_ref[0]**o+(-0.511789401943)*x_ref[0]+(-0.272452527007)*x_ref[1]**o+(-0.379021521619)*x_ref[1]
        ref[5,0,0]=(0.509668147054)*x_ref[0]**o+(0.919355584411)*x_ref[0]+(-0.673215166917)*x_ref[1]**o+(0.766459631931)*x_ref[1]
        ref[5,0,1]=(0.177006227916)*x_ref[0]**o+(-0.275118477)*x_ref[0]+(0.0302367843701)*x_ref[1]**o+(-0.761226360328)*x_ref[1]
        ref[5,1,0]=(0.013432534638)*x_ref[0]**o+(-0.911626567027)*x_ref[0]+(-0.710520442529)*x_ref[1]**o+(-0.882317448423)*x_ref[1]
        ref[5,1,1]=(0.595583093615)*x_ref[0]**o+(0.163527905156)*x_ref[0]+(-0.122165051333)*x_ref[1]**o+(-0.526359288181)*x_ref[1]
      else:
        arg[0,0,0]=(0.230695674262)*x[0]**o+(-0.987212801716)*x[0]+(-0.0599001434987)*x[1]**o+(0.719591259555)*x[1]+(-0.665186369759)*x[2]**o+(-0.563552602515)*x[2]
        arg[0,0,1]=(-0.775441966131)*x[0]**o+(-0.152166849702)*x[0]+(0.626466789363)*x[1]**o+(0.179356346581)*x[1]+(-0.687406927434)*x[2]**o+(-0.563492313114)*x[2]
        arg[0,1,0]=(0.749522748133)*x[0]**o+(-0.582456308337)*x[0]+(0.676882643058)*x[1]**o+(0.365363182598)*x[1]+(-0.550867708812)*x[2]**o+(0.777938151183)*x[2]
        arg[0,1,1]=(0.389788552775)*x[0]**o+(-0.298877489997)*x[0]+(0.560326209269)*x[1]**o+(-0.113249832155)*x[1]+(0.588087510254)*x[2]**o+(-0.593503776461)*x[2]
        arg[1,0,0]=(-0.354661103909)*x[0]**o+(-0.195375369549)*x[0]+(-0.798795069895)*x[1]**o+(0.0968631048579)*x[1]+(-0.016118190391)*x[2]**o+(-0.903745407705)*x[2]
        arg[1,0,1]=(-0.383688834784)*x[0]**o+(0.861256231933)*x[0]+(-0.767204834667)*x[1]**o+(0.338875482372)*x[1]+(0.117146310891)*x[2]**o+(-0.302651846083)*x[2]
        arg[1,1,0]=(0.646263670969)*x[0]**o+(-0.447984008871)*x[0]+(-0.169322492785)*x[1]**o+(-0.570614724481)*x[1]+(-0.848518457618)*x[2]**o+(-0.0750005539216)*x[2]
        arg[1,1,1]=(-0.232496745304)*x[0]**o+(0.182581173413)*x[0]+(0.8032436357)*x[1]**o+(-0.391266449476)*x[1]+(0.491677629766)*x[2]**o+(-0.188361708708)*x[2]
        arg[2,0,0]=(-0.571145099091)*x[0]**o+(-0.486855242188)*x[0]+(0.52833582073)*x[1]**o+(0.326716348242)*x[1]+(0.528475192869)*x[2]**o+(0.163280575137)*x[2]
        arg[2,0,1]=(-0.368889674601)*x[0]**o+(0.601775816775)*x[0]+(0.0281742455403)*x[1]**o+(-0.198869716243)*x[1]+(0.193343065506)*x[2]**o+(-0.0614451571108)*x[2]
        arg[2,1,0]=(0.724595792492)*x[0]**o+(-0.206983407651)*x[0]+(-0.33966393719)*x[1]**o+(0.929465830796)*x[1]+(-0.938517195863)*x[2]**o+(0.70874077222)*x[2]
        arg[2,1,1]=(-0.536303362606)*x[0]**o+(-0.430663025812)*x[0]+(-0.0845923469062)*x[1]**o+(0.866655631144)*x[1]+(-0.2378102403)*x[2]**o+(0.969070797268)*x[2]
        arg[3,0,0]=(0.345137352559)*x[0]**o+(-0.696652518382)*x[0]+(0.914998974008)*x[1]**o+(-0.874935163156)*x[1]+(0.947812661283)*x[2]**o+(0.868611662789)*x[2]
        arg[3,0,1]=(-0.421640746358)*x[0]**o+(-0.37925991208)*x[0]+(0.468281241452)*x[1]**o+(0.439086689805)*x[1]+(-0.607986317001)*x[2]**o+(0.269488168937)*x[2]
        arg[3,1,0]=(-0.550485456473)*x[0]**o+(-0.520018714495)*x[0]+(0.439679309654)*x[1]**o+(-0.0652434084668)*x[1]+(-0.584838668708)*x[2]**o+(-0.46047281034)*x[2]
        arg[3,1,1]=(-0.421746456038)*x[0]**o+(0.370489549278)*x[0]+(-0.683682295105)*x[1]**o+(0.324051718035)*x[1]+(-0.579046430636)*x[2]**o+(0.587623604941)*x[2]
        arg[4,0,0]=(-0.167582584882)*x[0]**o+(0.524904122243)*x[0]+(0.665109348362)*x[1]**o+(-0.093655366509)*x[1]+(0.704404656254)*x[2]**o+(-0.402602441741)*x[2]
        arg[4,0,1]=(0.633854964824)*x[0]**o+(-0.836385840194)*x[0]+(-0.0995437996628)*x[1]**o+(-0.200285338869)*x[1]+(0.535023783881)*x[2]**o+(0.884702543593)*x[2]
        arg[4,1,0]=(-0.953619378208)*x[0]**o+(-0.390518815669)*x[0]+(-0.0590889178236)*x[1]**o+(-0.997225825019)*x[1]+(0.0520418736402)*x[2]**o+(0.484749465262)*x[2]
        arg[4,1,1]=(-0.440389854756)*x[0]**o+(0.15520431778)*x[0]+(0.346839174513)*x[1]**o+(0.0172676873159)*x[1]+(0.86275919224)*x[2]**o+(0.568069148992)*x[2]
        arg[5,0,0]=(-0.303687941217)*x[0]**o+(0.516115915687)*x[0]+(-0.750276597519)*x[1]**o+(0.734039058778)*x[1]+(-0.624499210252)*x[2]**o+(-0.926338135421)*x[2]
        arg[5,0,1]=(-0.472230777895)*x[0]**o+(-0.557827728373)*x[0]+(-0.403783023157)*x[1]**o+(0.70433336955)*x[1]+(-0.0211737454531)*x[2]**o+(0.0774406125836)*x[2]
        arg[5,1,0]=(0.718149875204)*x[0]**o+(-0.183000739787)*x[0]+(-0.655724252373)*x[1]**o+(-0.634941710776)*x[1]+(-0.26300163323)*x[2]**o+(-0.25105424715)*x[2]
        arg[5,1,1]=(0.109601904441)*x[0]**o+(-0.00640764631554)*x[0]+(-0.488035610009)*x[1]**o+(0.909668323769)*x[1]+(0.229692168501)*x[2]**o+(-0.739241309497)*x[2]
        ref[0,0,0]=(0.230695674262)*x_ref[0]**o+(-0.987212801716)*x_ref[0]+(-0.0599001434987)*x_ref[1]**o+(0.719591259555)*x_ref[1]+(-0.665186369759)*x_ref[2]**o+(-0.563552602515)*x_ref[2]
        ref[0,0,1]=(-0.775441966131)*x_ref[0]**o+(-0.152166849702)*x_ref[0]+(0.626466789363)*x_ref[1]**o+(0.179356346581)*x_ref[1]+(-0.687406927434)*x_ref[2]**o+(-0.563492313114)*x_ref[2]
        ref[0,1,0]=(0.749522748133)*x_ref[0]**o+(-0.582456308337)*x_ref[0]+(0.676882643058)*x_ref[1]**o+(0.365363182598)*x_ref[1]+(-0.550867708812)*x_ref[2]**o+(0.777938151183)*x_ref[2]
        ref[0,1,1]=(0.389788552775)*x_ref[0]**o+(-0.298877489997)*x_ref[0]+(0.560326209269)*x_ref[1]**o+(-0.113249832155)*x_ref[1]+(0.588087510254)*x_ref[2]**o+(-0.593503776461)*x_ref[2]
        ref[1,0,0]=(-0.354661103909)*x_ref[0]**o+(-0.195375369549)*x_ref[0]+(-0.798795069895)*x_ref[1]**o+(0.0968631048579)*x_ref[1]+(-0.016118190391)*x_ref[2]**o+(-0.903745407705)*x_ref[2]
        ref[1,0,1]=(-0.383688834784)*x_ref[0]**o+(0.861256231933)*x_ref[0]+(-0.767204834667)*x_ref[1]**o+(0.338875482372)*x_ref[1]+(0.117146310891)*x_ref[2]**o+(-0.302651846083)*x_ref[2]
        ref[1,1,0]=(0.646263670969)*x_ref[0]**o+(-0.447984008871)*x_ref[0]+(-0.169322492785)*x_ref[1]**o+(-0.570614724481)*x_ref[1]+(-0.848518457618)*x_ref[2]**o+(-0.0750005539216)*x_ref[2]
        ref[1,1,1]=(-0.232496745304)*x_ref[0]**o+(0.182581173413)*x_ref[0]+(0.8032436357)*x_ref[1]**o+(-0.391266449476)*x_ref[1]+(0.491677629766)*x_ref[2]**o+(-0.188361708708)*x_ref[2]
        ref[2,0,0]=(-0.571145099091)*x_ref[0]**o+(-0.486855242188)*x_ref[0]+(0.52833582073)*x_ref[1]**o+(0.326716348242)*x_ref[1]+(0.528475192869)*x_ref[2]**o+(0.163280575137)*x_ref[2]
        ref[2,0,1]=(-0.368889674601)*x_ref[0]**o+(0.601775816775)*x_ref[0]+(0.0281742455403)*x_ref[1]**o+(-0.198869716243)*x_ref[1]+(0.193343065506)*x_ref[2]**o+(-0.0614451571108)*x_ref[2]
        ref[2,1,0]=(0.724595792492)*x_ref[0]**o+(-0.206983407651)*x_ref[0]+(-0.33966393719)*x_ref[1]**o+(0.929465830796)*x_ref[1]+(-0.938517195863)*x_ref[2]**o+(0.70874077222)*x_ref[2]
        ref[2,1,1]=(-0.536303362606)*x_ref[0]**o+(-0.430663025812)*x_ref[0]+(-0.0845923469062)*x_ref[1]**o+(0.866655631144)*x_ref[1]+(-0.2378102403)*x_ref[2]**o+(0.969070797268)*x_ref[2]
        ref[3,0,0]=(0.345137352559)*x_ref[0]**o+(-0.696652518382)*x_ref[0]+(0.914998974008)*x_ref[1]**o+(-0.874935163156)*x_ref[1]+(0.947812661283)*x_ref[2]**o+(0.868611662789)*x_ref[2]
        ref[3,0,1]=(-0.421640746358)*x_ref[0]**o+(-0.37925991208)*x_ref[0]+(0.468281241452)*x_ref[1]**o+(0.439086689805)*x_ref[1]+(-0.607986317001)*x_ref[2]**o+(0.269488168937)*x_ref[2]
        ref[3,1,0]=(-0.550485456473)*x_ref[0]**o+(-0.520018714495)*x_ref[0]+(0.439679309654)*x_ref[1]**o+(-0.0652434084668)*x_ref[1]+(-0.584838668708)*x_ref[2]**o+(-0.46047281034)*x_ref[2]
        ref[3,1,1]=(-0.421746456038)*x_ref[0]**o+(0.370489549278)*x_ref[0]+(-0.683682295105)*x_ref[1]**o+(0.324051718035)*x_ref[1]+(-0.579046430636)*x_ref[2]**o+(0.587623604941)*x_ref[2]
        ref[4,0,0]=(-0.167582584882)*x_ref[0]**o+(0.524904122243)*x_ref[0]+(0.665109348362)*x_ref[1]**o+(-0.093655366509)*x_ref[1]+(0.704404656254)*x_ref[2]**o+(-0.402602441741)*x_ref[2]
        ref[4,0,1]=(0.633854964824)*x_ref[0]**o+(-0.836385840194)*x_ref[0]+(-0.0995437996628)*x_ref[1]**o+(-0.200285338869)*x_ref[1]+(0.535023783881)*x_ref[2]**o+(0.884702543593)*x_ref[2]
        ref[4,1,0]=(-0.953619378208)*x_ref[0]**o+(-0.390518815669)*x_ref[0]+(-0.0590889178236)*x_ref[1]**o+(-0.997225825019)*x_ref[1]+(0.0520418736402)*x_ref[2]**o+(0.484749465262)*x_ref[2]
        ref[4,1,1]=(-0.440389854756)*x_ref[0]**o+(0.15520431778)*x_ref[0]+(0.346839174513)*x_ref[1]**o+(0.0172676873159)*x_ref[1]+(0.86275919224)*x_ref[2]**o+(0.568069148992)*x_ref[2]
        ref[5,0,0]=(-0.303687941217)*x_ref[0]**o+(0.516115915687)*x_ref[0]+(-0.750276597519)*x_ref[1]**o+(0.734039058778)*x_ref[1]+(-0.624499210252)*x_ref[2]**o+(-0.926338135421)*x_ref[2]
        ref[5,0,1]=(-0.472230777895)*x_ref[0]**o+(-0.557827728373)*x_ref[0]+(-0.403783023157)*x_ref[1]**o+(0.70433336955)*x_ref[1]+(-0.0211737454531)*x_ref[2]**o+(0.0774406125836)*x_ref[2]
        ref[5,1,0]=(0.718149875204)*x_ref[0]**o+(-0.183000739787)*x_ref[0]+(-0.655724252373)*x_ref[1]**o+(-0.634941710776)*x_ref[1]+(-0.26300163323)*x_ref[2]**o+(-0.25105424715)*x_ref[2]
        ref[5,1,1]=(0.109601904441)*x_ref[0]**o+(-0.00640764631554)*x_ref[0]+(-0.488035610009)*x_ref[1]**o+(0.909668323769)*x_ref[1]+(0.229692168501)*x_ref[2]**o+(-0.739241309497)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.912518456494)*x[0]**o+(-0.828176151903)*x[0]+(-0.313940298121)*x[1]**o+(-0.850120336531)*x[1]
        arg[0,0,0,1]=(0.14415852783)*x[0]**o+(-0.294155246668)*x[0]+(0.562517248232)*x[1]**o+(-0.204505938286)*x[1]
        arg[0,0,1,0]=(0.555455404368)*x[0]**o+(0.916425343576)*x[0]+(0.547252315693)*x[1]**o+(-0.0992842367479)*x[1]
        arg[0,0,1,1]=(0.603632473928)*x[0]**o+(0.505571549062)*x[0]+(-0.490402921542)*x[1]**o+(0.908624427316)*x[1]
        arg[0,0,2,0]=(0.729542308985)*x[0]**o+(-0.654310868638)*x[0]+(0.0845981276245)*x[1]**o+(-0.95231604686)*x[1]
        arg[0,0,2,1]=(0.0511615606745)*x[0]**o+(-0.939469887975)*x[0]+(-0.206229418155)*x[1]**o+(0.254879844701)*x[1]
        arg[0,1,0,0]=(0.257917344507)*x[0]**o+(0.860729621288)*x[0]+(0.489966193348)*x[1]**o+(0.174897233858)*x[1]
        arg[0,1,0,1]=(0.336588362051)*x[0]**o+(-0.690598635114)*x[0]+(-0.0913735300483)*x[1]**o+(-0.932914439004)*x[1]
        arg[0,1,1,0]=(0.244583093108)*x[0]**o+(-0.827768167758)*x[0]+(0.573065359412)*x[1]**o+(0.176685515596)*x[1]
        arg[0,1,1,1]=(-0.520550726127)*x[0]**o+(-0.605243177286)*x[0]+(0.130549556226)*x[1]**o+(0.750019884067)*x[1]
        arg[0,1,2,0]=(-0.680201627614)*x[0]**o+(0.679925324882)*x[0]+(-0.060408604263)*x[1]**o+(0.568125689359)*x[1]
        arg[0,1,2,1]=(-0.111908859432)*x[0]**o+(0.256521037018)*x[0]+(0.332050616611)*x[1]**o+(-0.0500016397683)*x[1]
        arg[0,2,0,0]=(-0.484953933663)*x[0]**o+(0.628032211831)*x[0]+(-0.63297141454)*x[1]**o+(0.141016662755)*x[1]
        arg[0,2,0,1]=(-0.463920680465)*x[0]**o+(-0.868825520511)*x[0]+(-0.36594425006)*x[1]**o+(-0.823052216578)*x[1]
        arg[0,2,1,0]=(0.972337802357)*x[0]**o+(0.943605157343)*x[0]+(-0.113950717844)*x[1]**o+(-0.839596452897)*x[1]
        arg[0,2,1,1]=(-0.295562401302)*x[0]**o+(0.735770621017)*x[0]+(-0.183838912348)*x[1]**o+(-0.929365380252)*x[1]
        arg[0,2,2,0]=(0.912886298388)*x[0]**o+(0.847507982065)*x[0]+(-0.341306036323)*x[1]**o+(0.393627025978)*x[1]
        arg[0,2,2,1]=(-0.769650638386)*x[0]**o+(-0.619672356488)*x[0]+(-0.126894288906)*x[1]**o+(0.231547632139)*x[1]
        arg[0,3,0,0]=(-0.884442976834)*x[0]**o+(-0.0194505760676)*x[0]+(-0.338764675072)*x[1]**o+(0.723481600171)*x[1]
        arg[0,3,0,1]=(0.655973555091)*x[0]**o+(-0.556230649412)*x[0]+(-0.850356832441)*x[1]**o+(0.115267314853)*x[1]
        arg[0,3,1,0]=(0.659628443559)*x[0]**o+(-0.586677822491)*x[0]+(0.135930588518)*x[1]**o+(0.206866895948)*x[1]
        arg[0,3,1,1]=(-0.700258957557)*x[0]**o+(-0.699774198764)*x[0]+(-0.294009479186)*x[1]**o+(0.848159666503)*x[1]
        arg[0,3,2,0]=(-0.927501754488)*x[0]**o+(-0.981178071331)*x[0]+(0.614074194451)*x[1]**o+(0.711549573086)*x[1]
        arg[0,3,2,1]=(0.312434473198)*x[0]**o+(-0.241506125991)*x[0]+(-0.0840979088912)*x[1]**o+(0.902948463117)*x[1]
        arg[0,4,0,0]=(-0.892269860242)*x[0]**o+(0.823434435855)*x[0]+(0.50499383365)*x[1]**o+(0.763736541034)*x[1]
        arg[0,4,0,1]=(-0.606959691191)*x[0]**o+(0.945562536791)*x[0]+(-0.46271929481)*x[1]**o+(0.929514432096)*x[1]
        arg[0,4,1,0]=(0.953382685939)*x[0]**o+(0.553766359805)*x[0]+(-0.191004010293)*x[1]**o+(0.556369491698)*x[1]
        arg[0,4,1,1]=(-0.269402649165)*x[0]**o+(-0.158926232076)*x[0]+(-0.390038749681)*x[1]**o+(0.553655192666)*x[1]
        arg[0,4,2,0]=(0.0785586372006)*x[0]**o+(-0.778452221862)*x[0]+(-0.927627273292)*x[1]**o+(-0.187742405246)*x[1]
        arg[0,4,2,1]=(-0.965329945447)*x[0]**o+(0.304539006089)*x[0]+(0.213031689568)*x[1]**o+(-0.2807790149)*x[1]
        arg[1,0,0,0]=(0.981534698569)*x[0]**o+(0.466451685806)*x[0]+(0.845322931856)*x[1]**o+(0.582422032334)*x[1]
        arg[1,0,0,1]=(-0.0645798740394)*x[0]**o+(0.0475214206619)*x[0]+(-0.687759433076)*x[1]**o+(0.403440938285)*x[1]
        arg[1,0,1,0]=(-0.317887773392)*x[0]**o+(0.59101163416)*x[0]+(0.658626263291)*x[1]**o+(-0.618167594137)*x[1]
        arg[1,0,1,1]=(-0.314093092813)*x[0]**o+(0.699933899942)*x[0]+(0.824195260822)*x[1]**o+(-0.42875327345)*x[1]
        arg[1,0,2,0]=(-0.254254956887)*x[0]**o+(-0.912137080163)*x[0]+(0.360457121452)*x[1]**o+(-0.902596912588)*x[1]
        arg[1,0,2,1]=(-0.577214593683)*x[0]**o+(-0.419495663008)*x[0]+(-0.803459783331)*x[1]**o+(0.490885513568)*x[1]
        arg[1,1,0,0]=(0.477947766892)*x[0]**o+(-0.853046971581)*x[0]+(0.576044041601)*x[1]**o+(-0.447007701252)*x[1]
        arg[1,1,0,1]=(0.5886933409)*x[0]**o+(0.253650778351)*x[0]+(-0.198070684431)*x[1]**o+(0.189479364528)*x[1]
        arg[1,1,1,0]=(-0.528744478872)*x[0]**o+(-0.504072059557)*x[0]+(-0.0161832474015)*x[1]**o+(0.639592055966)*x[1]
        arg[1,1,1,1]=(-0.581844827169)*x[0]**o+(0.997883224538)*x[0]+(0.9354364489)*x[1]**o+(0.884060785136)*x[1]
        arg[1,1,2,0]=(0.858226955817)*x[0]**o+(0.707261698312)*x[0]+(-0.738309111076)*x[1]**o+(-0.644655976448)*x[1]
        arg[1,1,2,1]=(-0.254976445964)*x[0]**o+(-0.624982185158)*x[0]+(0.304854486996)*x[1]**o+(0.597089811682)*x[1]
        arg[1,2,0,0]=(-0.625190759215)*x[0]**o+(-0.267820197905)*x[0]+(-0.721452024377)*x[1]**o+(-0.830984245819)*x[1]
        arg[1,2,0,1]=(-0.972257608766)*x[0]**o+(0.981368890544)*x[0]+(-0.925886510064)*x[1]**o+(-0.337107356862)*x[1]
        arg[1,2,1,0]=(-0.178932954207)*x[0]**o+(-0.941789507374)*x[0]+(0.360726744794)*x[1]**o+(0.347621863123)*x[1]
        arg[1,2,1,1]=(-0.59790695611)*x[0]**o+(0.552509013166)*x[0]+(0.592412443998)*x[1]**o+(0.105369637971)*x[1]
        arg[1,2,2,0]=(0.300713195225)*x[0]**o+(0.814257503574)*x[0]+(-0.281407483492)*x[1]**o+(-0.803431681095)*x[1]
        arg[1,2,2,1]=(0.80113022942)*x[0]**o+(0.749336171665)*x[0]+(-0.0842928982283)*x[1]**o+(-0.974041792058)*x[1]
        arg[1,3,0,0]=(0.533118876666)*x[0]**o+(-0.00888178987567)*x[0]+(0.372869213762)*x[1]**o+(-0.652513683144)*x[1]
        arg[1,3,0,1]=(-0.295090222537)*x[0]**o+(0.772749124303)*x[0]+(-0.00857029468729)*x[1]**o+(0.927794431125)*x[1]
        arg[1,3,1,0]=(-0.152537465441)*x[0]**o+(0.154900534979)*x[0]+(-0.488743959543)*x[1]**o+(-0.0318497199232)*x[1]
        arg[1,3,1,1]=(-0.754333787921)*x[0]**o+(-0.0583007358701)*x[0]+(-0.050921106891)*x[1]**o+(0.161426452156)*x[1]
        arg[1,3,2,0]=(0.387357469451)*x[0]**o+(-0.213640305885)*x[0]+(-0.844609663023)*x[1]**o+(0.0761914594192)*x[1]
        arg[1,3,2,1]=(0.36964862459)*x[0]**o+(0.887888401952)*x[0]+(-0.933930511093)*x[1]**o+(0.769046614932)*x[1]
        arg[1,4,0,0]=(0.680878887266)*x[0]**o+(0.62930033679)*x[0]+(0.894793542585)*x[1]**o+(-0.636881609067)*x[1]
        arg[1,4,0,1]=(-0.346053756598)*x[0]**o+(0.225254433923)*x[0]+(-0.861052985621)*x[1]**o+(0.415123108136)*x[1]
        arg[1,4,1,0]=(-0.997758408183)*x[0]**o+(-0.265055614341)*x[0]+(-0.7568881617)*x[1]**o+(0.95144584777)*x[1]
        arg[1,4,1,1]=(0.227777912008)*x[0]**o+(-0.83647660817)*x[0]+(-0.275733410165)*x[1]**o+(-0.971560248343)*x[1]
        arg[1,4,2,0]=(0.718803859114)*x[0]**o+(-0.978966510682)*x[0]+(0.463102438649)*x[1]**o+(0.987805610534)*x[1]
        arg[1,4,2,1]=(-0.107543838917)*x[0]**o+(-0.362684018694)*x[0]+(-0.310926939373)*x[1]**o+(-0.571571044712)*x[1]
        arg[2,0,0,0]=(-0.475977416182)*x[0]**o+(-0.627173002609)*x[0]+(-0.369962970984)*x[1]**o+(-0.245077302537)*x[1]
        arg[2,0,0,1]=(0.913949611751)*x[0]**o+(0.645611424617)*x[0]+(0.639304734076)*x[1]**o+(-0.12171441998)*x[1]
        arg[2,0,1,0]=(0.251434098263)*x[0]**o+(-0.119898861407)*x[0]+(-0.365614523709)*x[1]**o+(-0.802928738296)*x[1]
        arg[2,0,1,1]=(0.756390927006)*x[0]**o+(0.246578936199)*x[0]+(0.714968085461)*x[1]**o+(-0.929428808096)*x[1]
        arg[2,0,2,0]=(-0.516888361447)*x[0]**o+(0.0344339250886)*x[0]+(-0.383709788774)*x[1]**o+(-0.437344417623)*x[1]
        arg[2,0,2,1]=(-0.219307162317)*x[0]**o+(0.704909718842)*x[0]+(-0.340454491855)*x[1]**o+(-0.933866231773)*x[1]
        arg[2,1,0,0]=(-0.0618967891627)*x[0]**o+(-0.0590712030789)*x[0]+(0.116776433077)*x[1]**o+(0.0241239764104)*x[1]
        arg[2,1,0,1]=(-0.802207645638)*x[0]**o+(0.590732740531)*x[0]+(-0.25990272933)*x[1]**o+(-0.257633805416)*x[1]
        arg[2,1,1,0]=(0.827623274379)*x[0]**o+(-0.450310573599)*x[0]+(0.213809775826)*x[1]**o+(-0.905410777787)*x[1]
        arg[2,1,1,1]=(-0.624510830356)*x[0]**o+(-0.705682607787)*x[0]+(0.430105716993)*x[1]**o+(0.975510832521)*x[1]
        arg[2,1,2,0]=(-0.297562223309)*x[0]**o+(-0.99090348994)*x[0]+(0.548589416291)*x[1]**o+(0.591718478583)*x[1]
        arg[2,1,2,1]=(-0.531899919376)*x[0]**o+(-0.227453719581)*x[0]+(-0.131664806128)*x[1]**o+(0.0521895817509)*x[1]
        arg[2,2,0,0]=(-0.592291745786)*x[0]**o+(-0.586879961972)*x[0]+(0.563196552327)*x[1]**o+(0.876989321187)*x[1]
        arg[2,2,0,1]=(-0.273360445369)*x[0]**o+(-0.714740527775)*x[0]+(-0.647699497837)*x[1]**o+(-0.0405089612762)*x[1]
        arg[2,2,1,0]=(-0.536629732224)*x[0]**o+(0.72656373966)*x[0]+(0.986630594165)*x[1]**o+(0.0741630711667)*x[1]
        arg[2,2,1,1]=(0.770746088838)*x[0]**o+(-0.219279833426)*x[0]+(0.875445174779)*x[1]**o+(-0.947889926444)*x[1]
        arg[2,2,2,0]=(-0.41428815119)*x[0]**o+(-0.696166268903)*x[0]+(-0.96838062553)*x[1]**o+(0.781394205899)*x[1]
        arg[2,2,2,1]=(-0.810267960752)*x[0]**o+(-0.873632687702)*x[0]+(0.0131971321611)*x[1]**o+(-0.552767192193)*x[1]
        arg[2,3,0,0]=(-0.216091366291)*x[0]**o+(-0.17824047472)*x[0]+(-0.4644515725)*x[1]**o+(0.135819458957)*x[1]
        arg[2,3,0,1]=(-0.0823616523105)*x[0]**o+(0.33121362151)*x[0]+(0.926719919839)*x[1]**o+(0.663475641636)*x[1]
        arg[2,3,1,0]=(-0.278303066047)*x[0]**o+(-0.457286378314)*x[0]+(0.186792859092)*x[1]**o+(-0.734245612464)*x[1]
        arg[2,3,1,1]=(0.810871459779)*x[0]**o+(-0.706641995324)*x[0]+(-0.287827176091)*x[1]**o+(0.453216090282)*x[1]
        arg[2,3,2,0]=(0.83826221049)*x[0]**o+(0.556887792111)*x[0]+(-0.000231102245991)*x[1]**o+(-0.440730958954)*x[1]
        arg[2,3,2,1]=(-0.455087877907)*x[0]**o+(-0.132259634036)*x[0]+(-0.492728335525)*x[1]**o+(-0.880311317777)*x[1]
        arg[2,4,0,0]=(-0.303141352388)*x[0]**o+(0.426137046804)*x[0]+(0.788934776738)*x[1]**o+(0.157477785179)*x[1]
        arg[2,4,0,1]=(0.603411288614)*x[0]**o+(-0.533561451987)*x[0]+(-0.901259210942)*x[1]**o+(0.547372624746)*x[1]
        arg[2,4,1,0]=(0.69517961739)*x[0]**o+(-0.106282888246)*x[0]+(0.257484565681)*x[1]**o+(0.655749633294)*x[1]
        arg[2,4,1,1]=(-0.072748573118)*x[0]**o+(0.905966560113)*x[0]+(0.433145718123)*x[1]**o+(0.187620401615)*x[1]
        arg[2,4,2,0]=(-0.0280176325954)*x[0]**o+(0.489610131347)*x[0]+(-0.740554922134)*x[1]**o+(-0.85045302096)*x[1]
        arg[2,4,2,1]=(-0.158503180709)*x[0]**o+(0.525929643075)*x[0]+(-0.352022652515)*x[1]**o+(-0.706553104932)*x[1]
        arg[3,0,0,0]=(-0.837316508127)*x[0]**o+(-0.89490181779)*x[0]+(0.510084075291)*x[1]**o+(0.321206310024)*x[1]
        arg[3,0,0,1]=(-0.325038591079)*x[0]**o+(0.140342922753)*x[0]+(-0.937466403004)*x[1]**o+(0.580557383497)*x[1]
        arg[3,0,1,0]=(-0.939713654062)*x[0]**o+(-0.394660244166)*x[0]+(0.762376614557)*x[1]**o+(-0.135565547532)*x[1]
        arg[3,0,1,1]=(0.0275332524464)*x[0]**o+(0.368737701882)*x[0]+(0.983641372218)*x[1]**o+(0.573815516366)*x[1]
        arg[3,0,2,0]=(0.816957570314)*x[0]**o+(-0.357626892562)*x[0]+(-0.368933149642)*x[1]**o+(-0.692502486914)*x[1]
        arg[3,0,2,1]=(-0.372408810926)*x[0]**o+(-0.113036390416)*x[0]+(-0.355880297003)*x[1]**o+(0.755478191481)*x[1]
        arg[3,1,0,0]=(0.0399012900832)*x[0]**o+(-0.879576900086)*x[0]+(-0.660858776134)*x[1]**o+(-0.59837040429)*x[1]
        arg[3,1,0,1]=(0.569397128496)*x[0]**o+(-0.138333417746)*x[0]+(0.257769147727)*x[1]**o+(0.174737141027)*x[1]
        arg[3,1,1,0]=(-0.67938408823)*x[0]**o+(0.862731161981)*x[0]+(-0.792538485264)*x[1]**o+(-0.0805197796157)*x[1]
        arg[3,1,1,1]=(0.0113523229279)*x[0]**o+(0.475393597561)*x[0]+(-0.705748021886)*x[1]**o+(0.00836460135285)*x[1]
        arg[3,1,2,0]=(0.463839509799)*x[0]**o+(-0.12663544643)*x[0]+(0.353158399738)*x[1]**o+(0.674929537597)*x[1]
        arg[3,1,2,1]=(-0.6019896019)*x[0]**o+(-0.638280685565)*x[0]+(0.280047072988)*x[1]**o+(0.569239958138)*x[1]
        arg[3,2,0,0]=(-0.113906500346)*x[0]**o+(-0.493494498501)*x[0]+(-0.686419229162)*x[1]**o+(-0.0727046763516)*x[1]
        arg[3,2,0,1]=(-0.302092112795)*x[0]**o+(-0.74819167)*x[0]+(-0.478908975282)*x[1]**o+(0.430024789627)*x[1]
        arg[3,2,1,0]=(0.353535608984)*x[0]**o+(-0.743408549264)*x[0]+(-0.069122881538)*x[1]**o+(-0.0767595089529)*x[1]
        arg[3,2,1,1]=(-0.629946386234)*x[0]**o+(0.851779590894)*x[0]+(0.728577108695)*x[1]**o+(-0.366983040147)*x[1]
        arg[3,2,2,0]=(0.1685167508)*x[0]**o+(0.149999747637)*x[0]+(0.709320430374)*x[1]**o+(-0.280950677222)*x[1]
        arg[3,2,2,1]=(-0.956164974578)*x[0]**o+(0.240460380901)*x[0]+(-0.2820176577)*x[1]**o+(-0.595667156955)*x[1]
        arg[3,3,0,0]=(-0.0711295055938)*x[0]**o+(-0.314513251791)*x[0]+(-0.41924591293)*x[1]**o+(0.0813197780959)*x[1]
        arg[3,3,0,1]=(-0.230592584877)*x[0]**o+(-0.488005144715)*x[0]+(-0.450611759976)*x[1]**o+(-0.0571152769768)*x[1]
        arg[3,3,1,0]=(-0.608255699014)*x[0]**o+(-0.697324108575)*x[0]+(-0.885935900746)*x[1]**o+(-0.884497550016)*x[1]
        arg[3,3,1,1]=(0.857286225292)*x[0]**o+(0.509954550626)*x[0]+(0.182030821433)*x[1]**o+(-0.268076060316)*x[1]
        arg[3,3,2,0]=(-0.77322653479)*x[0]**o+(-0.583013210018)*x[0]+(-0.683867300277)*x[1]**o+(0.754530541251)*x[1]
        arg[3,3,2,1]=(-0.744667985181)*x[0]**o+(0.164165128978)*x[0]+(0.412340155958)*x[1]**o+(-0.38451107118)*x[1]
        arg[3,4,0,0]=(0.216720755049)*x[0]**o+(-0.347803433722)*x[0]+(0.417376268797)*x[1]**o+(0.475787304652)*x[1]
        arg[3,4,0,1]=(-0.737038295974)*x[0]**o+(0.579172519178)*x[0]+(-0.900272993672)*x[1]**o+(0.407018603625)*x[1]
        arg[3,4,1,0]=(-0.898959236474)*x[0]**o+(-0.886636008121)*x[0]+(-0.901972610044)*x[1]**o+(-0.518393782318)*x[1]
        arg[3,4,1,1]=(0.976342884921)*x[0]**o+(-0.273568319493)*x[0]+(-0.332424980846)*x[1]**o+(-0.0146661731139)*x[1]
        arg[3,4,2,0]=(0.794792214027)*x[0]**o+(0.846927984413)*x[0]+(0.535818809709)*x[1]**o+(-0.49324400033)*x[1]
        arg[3,4,2,1]=(0.695766308812)*x[0]**o+(-0.37923153689)*x[0]+(0.0633513026782)*x[1]**o+(0.12334892105)*x[1]
        ref[0,0,0,0]=(0.912518456494)*x_ref[0]**o+(-0.828176151903)*x_ref[0]+(-0.313940298121)*x_ref[1]**o+(-0.850120336531)*x_ref[1]
        ref[0,0,0,1]=(0.14415852783)*x_ref[0]**o+(-0.294155246668)*x_ref[0]+(0.562517248232)*x_ref[1]**o+(-0.204505938286)*x_ref[1]
        ref[0,0,1,0]=(0.555455404368)*x_ref[0]**o+(0.916425343576)*x_ref[0]+(0.547252315693)*x_ref[1]**o+(-0.0992842367479)*x_ref[1]
        ref[0,0,1,1]=(0.603632473928)*x_ref[0]**o+(0.505571549062)*x_ref[0]+(-0.490402921542)*x_ref[1]**o+(0.908624427316)*x_ref[1]
        ref[0,0,2,0]=(0.729542308985)*x_ref[0]**o+(-0.654310868638)*x_ref[0]+(0.0845981276245)*x_ref[1]**o+(-0.95231604686)*x_ref[1]
        ref[0,0,2,1]=(0.0511615606745)*x_ref[0]**o+(-0.939469887975)*x_ref[0]+(-0.206229418155)*x_ref[1]**o+(0.254879844701)*x_ref[1]
        ref[0,1,0,0]=(0.257917344507)*x_ref[0]**o+(0.860729621288)*x_ref[0]+(0.489966193348)*x_ref[1]**o+(0.174897233858)*x_ref[1]
        ref[0,1,0,1]=(0.336588362051)*x_ref[0]**o+(-0.690598635114)*x_ref[0]+(-0.0913735300483)*x_ref[1]**o+(-0.932914439004)*x_ref[1]
        ref[0,1,1,0]=(0.244583093108)*x_ref[0]**o+(-0.827768167758)*x_ref[0]+(0.573065359412)*x_ref[1]**o+(0.176685515596)*x_ref[1]
        ref[0,1,1,1]=(-0.520550726127)*x_ref[0]**o+(-0.605243177286)*x_ref[0]+(0.130549556226)*x_ref[1]**o+(0.750019884067)*x_ref[1]
        ref[0,1,2,0]=(-0.680201627614)*x_ref[0]**o+(0.679925324882)*x_ref[0]+(-0.060408604263)*x_ref[1]**o+(0.568125689359)*x_ref[1]
        ref[0,1,2,1]=(-0.111908859432)*x_ref[0]**o+(0.256521037018)*x_ref[0]+(0.332050616611)*x_ref[1]**o+(-0.0500016397683)*x_ref[1]
        ref[0,2,0,0]=(-0.484953933663)*x_ref[0]**o+(0.628032211831)*x_ref[0]+(-0.63297141454)*x_ref[1]**o+(0.141016662755)*x_ref[1]
        ref[0,2,0,1]=(-0.463920680465)*x_ref[0]**o+(-0.868825520511)*x_ref[0]+(-0.36594425006)*x_ref[1]**o+(-0.823052216578)*x_ref[1]
        ref[0,2,1,0]=(0.972337802357)*x_ref[0]**o+(0.943605157343)*x_ref[0]+(-0.113950717844)*x_ref[1]**o+(-0.839596452897)*x_ref[1]
        ref[0,2,1,1]=(-0.295562401302)*x_ref[0]**o+(0.735770621017)*x_ref[0]+(-0.183838912348)*x_ref[1]**o+(-0.929365380252)*x_ref[1]
        ref[0,2,2,0]=(0.912886298388)*x_ref[0]**o+(0.847507982065)*x_ref[0]+(-0.341306036323)*x_ref[1]**o+(0.393627025978)*x_ref[1]
        ref[0,2,2,1]=(-0.769650638386)*x_ref[0]**o+(-0.619672356488)*x_ref[0]+(-0.126894288906)*x_ref[1]**o+(0.231547632139)*x_ref[1]
        ref[0,3,0,0]=(-0.884442976834)*x_ref[0]**o+(-0.0194505760676)*x_ref[0]+(-0.338764675072)*x_ref[1]**o+(0.723481600171)*x_ref[1]
        ref[0,3,0,1]=(0.655973555091)*x_ref[0]**o+(-0.556230649412)*x_ref[0]+(-0.850356832441)*x_ref[1]**o+(0.115267314853)*x_ref[1]
        ref[0,3,1,0]=(0.659628443559)*x_ref[0]**o+(-0.586677822491)*x_ref[0]+(0.135930588518)*x_ref[1]**o+(0.206866895948)*x_ref[1]
        ref[0,3,1,1]=(-0.700258957557)*x_ref[0]**o+(-0.699774198764)*x_ref[0]+(-0.294009479186)*x_ref[1]**o+(0.848159666503)*x_ref[1]
        ref[0,3,2,0]=(-0.927501754488)*x_ref[0]**o+(-0.981178071331)*x_ref[0]+(0.614074194451)*x_ref[1]**o+(0.711549573086)*x_ref[1]
        ref[0,3,2,1]=(0.312434473198)*x_ref[0]**o+(-0.241506125991)*x_ref[0]+(-0.0840979088912)*x_ref[1]**o+(0.902948463117)*x_ref[1]
        ref[0,4,0,0]=(-0.892269860242)*x_ref[0]**o+(0.823434435855)*x_ref[0]+(0.50499383365)*x_ref[1]**o+(0.763736541034)*x_ref[1]
        ref[0,4,0,1]=(-0.606959691191)*x_ref[0]**o+(0.945562536791)*x_ref[0]+(-0.46271929481)*x_ref[1]**o+(0.929514432096)*x_ref[1]
        ref[0,4,1,0]=(0.953382685939)*x_ref[0]**o+(0.553766359805)*x_ref[0]+(-0.191004010293)*x_ref[1]**o+(0.556369491698)*x_ref[1]
        ref[0,4,1,1]=(-0.269402649165)*x_ref[0]**o+(-0.158926232076)*x_ref[0]+(-0.390038749681)*x_ref[1]**o+(0.553655192666)*x_ref[1]
        ref[0,4,2,0]=(0.0785586372006)*x_ref[0]**o+(-0.778452221862)*x_ref[0]+(-0.927627273292)*x_ref[1]**o+(-0.187742405246)*x_ref[1]
        ref[0,4,2,1]=(-0.965329945447)*x_ref[0]**o+(0.304539006089)*x_ref[0]+(0.213031689568)*x_ref[1]**o+(-0.2807790149)*x_ref[1]
        ref[1,0,0,0]=(0.981534698569)*x_ref[0]**o+(0.466451685806)*x_ref[0]+(0.845322931856)*x_ref[1]**o+(0.582422032334)*x_ref[1]
        ref[1,0,0,1]=(-0.0645798740394)*x_ref[0]**o+(0.0475214206619)*x_ref[0]+(-0.687759433076)*x_ref[1]**o+(0.403440938285)*x_ref[1]
        ref[1,0,1,0]=(-0.317887773392)*x_ref[0]**o+(0.59101163416)*x_ref[0]+(0.658626263291)*x_ref[1]**o+(-0.618167594137)*x_ref[1]
        ref[1,0,1,1]=(-0.314093092813)*x_ref[0]**o+(0.699933899942)*x_ref[0]+(0.824195260822)*x_ref[1]**o+(-0.42875327345)*x_ref[1]
        ref[1,0,2,0]=(-0.254254956887)*x_ref[0]**o+(-0.912137080163)*x_ref[0]+(0.360457121452)*x_ref[1]**o+(-0.902596912588)*x_ref[1]
        ref[1,0,2,1]=(-0.577214593683)*x_ref[0]**o+(-0.419495663008)*x_ref[0]+(-0.803459783331)*x_ref[1]**o+(0.490885513568)*x_ref[1]
        ref[1,1,0,0]=(0.477947766892)*x_ref[0]**o+(-0.853046971581)*x_ref[0]+(0.576044041601)*x_ref[1]**o+(-0.447007701252)*x_ref[1]
        ref[1,1,0,1]=(0.5886933409)*x_ref[0]**o+(0.253650778351)*x_ref[0]+(-0.198070684431)*x_ref[1]**o+(0.189479364528)*x_ref[1]
        ref[1,1,1,0]=(-0.528744478872)*x_ref[0]**o+(-0.504072059557)*x_ref[0]+(-0.0161832474015)*x_ref[1]**o+(0.639592055966)*x_ref[1]
        ref[1,1,1,1]=(-0.581844827169)*x_ref[0]**o+(0.997883224538)*x_ref[0]+(0.9354364489)*x_ref[1]**o+(0.884060785136)*x_ref[1]
        ref[1,1,2,0]=(0.858226955817)*x_ref[0]**o+(0.707261698312)*x_ref[0]+(-0.738309111076)*x_ref[1]**o+(-0.644655976448)*x_ref[1]
        ref[1,1,2,1]=(-0.254976445964)*x_ref[0]**o+(-0.624982185158)*x_ref[0]+(0.304854486996)*x_ref[1]**o+(0.597089811682)*x_ref[1]
        ref[1,2,0,0]=(-0.625190759215)*x_ref[0]**o+(-0.267820197905)*x_ref[0]+(-0.721452024377)*x_ref[1]**o+(-0.830984245819)*x_ref[1]
        ref[1,2,0,1]=(-0.972257608766)*x_ref[0]**o+(0.981368890544)*x_ref[0]+(-0.925886510064)*x_ref[1]**o+(-0.337107356862)*x_ref[1]
        ref[1,2,1,0]=(-0.178932954207)*x_ref[0]**o+(-0.941789507374)*x_ref[0]+(0.360726744794)*x_ref[1]**o+(0.347621863123)*x_ref[1]
        ref[1,2,1,1]=(-0.59790695611)*x_ref[0]**o+(0.552509013166)*x_ref[0]+(0.592412443998)*x_ref[1]**o+(0.105369637971)*x_ref[1]
        ref[1,2,2,0]=(0.300713195225)*x_ref[0]**o+(0.814257503574)*x_ref[0]+(-0.281407483492)*x_ref[1]**o+(-0.803431681095)*x_ref[1]
        ref[1,2,2,1]=(0.80113022942)*x_ref[0]**o+(0.749336171665)*x_ref[0]+(-0.0842928982283)*x_ref[1]**o+(-0.974041792058)*x_ref[1]
        ref[1,3,0,0]=(0.533118876666)*x_ref[0]**o+(-0.00888178987567)*x_ref[0]+(0.372869213762)*x_ref[1]**o+(-0.652513683144)*x_ref[1]
        ref[1,3,0,1]=(-0.295090222537)*x_ref[0]**o+(0.772749124303)*x_ref[0]+(-0.00857029468729)*x_ref[1]**o+(0.927794431125)*x_ref[1]
        ref[1,3,1,0]=(-0.152537465441)*x_ref[0]**o+(0.154900534979)*x_ref[0]+(-0.488743959543)*x_ref[1]**o+(-0.0318497199232)*x_ref[1]
        ref[1,3,1,1]=(-0.754333787921)*x_ref[0]**o+(-0.0583007358701)*x_ref[0]+(-0.050921106891)*x_ref[1]**o+(0.161426452156)*x_ref[1]
        ref[1,3,2,0]=(0.387357469451)*x_ref[0]**o+(-0.213640305885)*x_ref[0]+(-0.844609663023)*x_ref[1]**o+(0.0761914594192)*x_ref[1]
        ref[1,3,2,1]=(0.36964862459)*x_ref[0]**o+(0.887888401952)*x_ref[0]+(-0.933930511093)*x_ref[1]**o+(0.769046614932)*x_ref[1]
        ref[1,4,0,0]=(0.680878887266)*x_ref[0]**o+(0.62930033679)*x_ref[0]+(0.894793542585)*x_ref[1]**o+(-0.636881609067)*x_ref[1]
        ref[1,4,0,1]=(-0.346053756598)*x_ref[0]**o+(0.225254433923)*x_ref[0]+(-0.861052985621)*x_ref[1]**o+(0.415123108136)*x_ref[1]
        ref[1,4,1,0]=(-0.997758408183)*x_ref[0]**o+(-0.265055614341)*x_ref[0]+(-0.7568881617)*x_ref[1]**o+(0.95144584777)*x_ref[1]
        ref[1,4,1,1]=(0.227777912008)*x_ref[0]**o+(-0.83647660817)*x_ref[0]+(-0.275733410165)*x_ref[1]**o+(-0.971560248343)*x_ref[1]
        ref[1,4,2,0]=(0.718803859114)*x_ref[0]**o+(-0.978966510682)*x_ref[0]+(0.463102438649)*x_ref[1]**o+(0.987805610534)*x_ref[1]
        ref[1,4,2,1]=(-0.107543838917)*x_ref[0]**o+(-0.362684018694)*x_ref[0]+(-0.310926939373)*x_ref[1]**o+(-0.571571044712)*x_ref[1]
        ref[2,0,0,0]=(-0.475977416182)*x_ref[0]**o+(-0.627173002609)*x_ref[0]+(-0.369962970984)*x_ref[1]**o+(-0.245077302537)*x_ref[1]
        ref[2,0,0,1]=(0.913949611751)*x_ref[0]**o+(0.645611424617)*x_ref[0]+(0.639304734076)*x_ref[1]**o+(-0.12171441998)*x_ref[1]
        ref[2,0,1,0]=(0.251434098263)*x_ref[0]**o+(-0.119898861407)*x_ref[0]+(-0.365614523709)*x_ref[1]**o+(-0.802928738296)*x_ref[1]
        ref[2,0,1,1]=(0.756390927006)*x_ref[0]**o+(0.246578936199)*x_ref[0]+(0.714968085461)*x_ref[1]**o+(-0.929428808096)*x_ref[1]
        ref[2,0,2,0]=(-0.516888361447)*x_ref[0]**o+(0.0344339250886)*x_ref[0]+(-0.383709788774)*x_ref[1]**o+(-0.437344417623)*x_ref[1]
        ref[2,0,2,1]=(-0.219307162317)*x_ref[0]**o+(0.704909718842)*x_ref[0]+(-0.340454491855)*x_ref[1]**o+(-0.933866231773)*x_ref[1]
        ref[2,1,0,0]=(-0.0618967891627)*x_ref[0]**o+(-0.0590712030789)*x_ref[0]+(0.116776433077)*x_ref[1]**o+(0.0241239764104)*x_ref[1]
        ref[2,1,0,1]=(-0.802207645638)*x_ref[0]**o+(0.590732740531)*x_ref[0]+(-0.25990272933)*x_ref[1]**o+(-0.257633805416)*x_ref[1]
        ref[2,1,1,0]=(0.827623274379)*x_ref[0]**o+(-0.450310573599)*x_ref[0]+(0.213809775826)*x_ref[1]**o+(-0.905410777787)*x_ref[1]
        ref[2,1,1,1]=(-0.624510830356)*x_ref[0]**o+(-0.705682607787)*x_ref[0]+(0.430105716993)*x_ref[1]**o+(0.975510832521)*x_ref[1]
        ref[2,1,2,0]=(-0.297562223309)*x_ref[0]**o+(-0.99090348994)*x_ref[0]+(0.548589416291)*x_ref[1]**o+(0.591718478583)*x_ref[1]
        ref[2,1,2,1]=(-0.531899919376)*x_ref[0]**o+(-0.227453719581)*x_ref[0]+(-0.131664806128)*x_ref[1]**o+(0.0521895817509)*x_ref[1]
        ref[2,2,0,0]=(-0.592291745786)*x_ref[0]**o+(-0.586879961972)*x_ref[0]+(0.563196552327)*x_ref[1]**o+(0.876989321187)*x_ref[1]
        ref[2,2,0,1]=(-0.273360445369)*x_ref[0]**o+(-0.714740527775)*x_ref[0]+(-0.647699497837)*x_ref[1]**o+(-0.0405089612762)*x_ref[1]
        ref[2,2,1,0]=(-0.536629732224)*x_ref[0]**o+(0.72656373966)*x_ref[0]+(0.986630594165)*x_ref[1]**o+(0.0741630711667)*x_ref[1]
        ref[2,2,1,1]=(0.770746088838)*x_ref[0]**o+(-0.219279833426)*x_ref[0]+(0.875445174779)*x_ref[1]**o+(-0.947889926444)*x_ref[1]
        ref[2,2,2,0]=(-0.41428815119)*x_ref[0]**o+(-0.696166268903)*x_ref[0]+(-0.96838062553)*x_ref[1]**o+(0.781394205899)*x_ref[1]
        ref[2,2,2,1]=(-0.810267960752)*x_ref[0]**o+(-0.873632687702)*x_ref[0]+(0.0131971321611)*x_ref[1]**o+(-0.552767192193)*x_ref[1]
        ref[2,3,0,0]=(-0.216091366291)*x_ref[0]**o+(-0.17824047472)*x_ref[0]+(-0.4644515725)*x_ref[1]**o+(0.135819458957)*x_ref[1]
        ref[2,3,0,1]=(-0.0823616523105)*x_ref[0]**o+(0.33121362151)*x_ref[0]+(0.926719919839)*x_ref[1]**o+(0.663475641636)*x_ref[1]
        ref[2,3,1,0]=(-0.278303066047)*x_ref[0]**o+(-0.457286378314)*x_ref[0]+(0.186792859092)*x_ref[1]**o+(-0.734245612464)*x_ref[1]
        ref[2,3,1,1]=(0.810871459779)*x_ref[0]**o+(-0.706641995324)*x_ref[0]+(-0.287827176091)*x_ref[1]**o+(0.453216090282)*x_ref[1]
        ref[2,3,2,0]=(0.83826221049)*x_ref[0]**o+(0.556887792111)*x_ref[0]+(-0.000231102245991)*x_ref[1]**o+(-0.440730958954)*x_ref[1]
        ref[2,3,2,1]=(-0.455087877907)*x_ref[0]**o+(-0.132259634036)*x_ref[0]+(-0.492728335525)*x_ref[1]**o+(-0.880311317777)*x_ref[1]
        ref[2,4,0,0]=(-0.303141352388)*x_ref[0]**o+(0.426137046804)*x_ref[0]+(0.788934776738)*x_ref[1]**o+(0.157477785179)*x_ref[1]
        ref[2,4,0,1]=(0.603411288614)*x_ref[0]**o+(-0.533561451987)*x_ref[0]+(-0.901259210942)*x_ref[1]**o+(0.547372624746)*x_ref[1]
        ref[2,4,1,0]=(0.69517961739)*x_ref[0]**o+(-0.106282888246)*x_ref[0]+(0.257484565681)*x_ref[1]**o+(0.655749633294)*x_ref[1]
        ref[2,4,1,1]=(-0.072748573118)*x_ref[0]**o+(0.905966560113)*x_ref[0]+(0.433145718123)*x_ref[1]**o+(0.187620401615)*x_ref[1]
        ref[2,4,2,0]=(-0.0280176325954)*x_ref[0]**o+(0.489610131347)*x_ref[0]+(-0.740554922134)*x_ref[1]**o+(-0.85045302096)*x_ref[1]
        ref[2,4,2,1]=(-0.158503180709)*x_ref[0]**o+(0.525929643075)*x_ref[0]+(-0.352022652515)*x_ref[1]**o+(-0.706553104932)*x_ref[1]
        ref[3,0,0,0]=(-0.837316508127)*x_ref[0]**o+(-0.89490181779)*x_ref[0]+(0.510084075291)*x_ref[1]**o+(0.321206310024)*x_ref[1]
        ref[3,0,0,1]=(-0.325038591079)*x_ref[0]**o+(0.140342922753)*x_ref[0]+(-0.937466403004)*x_ref[1]**o+(0.580557383497)*x_ref[1]
        ref[3,0,1,0]=(-0.939713654062)*x_ref[0]**o+(-0.394660244166)*x_ref[0]+(0.762376614557)*x_ref[1]**o+(-0.135565547532)*x_ref[1]
        ref[3,0,1,1]=(0.0275332524464)*x_ref[0]**o+(0.368737701882)*x_ref[0]+(0.983641372218)*x_ref[1]**o+(0.573815516366)*x_ref[1]
        ref[3,0,2,0]=(0.816957570314)*x_ref[0]**o+(-0.357626892562)*x_ref[0]+(-0.368933149642)*x_ref[1]**o+(-0.692502486914)*x_ref[1]
        ref[3,0,2,1]=(-0.372408810926)*x_ref[0]**o+(-0.113036390416)*x_ref[0]+(-0.355880297003)*x_ref[1]**o+(0.755478191481)*x_ref[1]
        ref[3,1,0,0]=(0.0399012900832)*x_ref[0]**o+(-0.879576900086)*x_ref[0]+(-0.660858776134)*x_ref[1]**o+(-0.59837040429)*x_ref[1]
        ref[3,1,0,1]=(0.569397128496)*x_ref[0]**o+(-0.138333417746)*x_ref[0]+(0.257769147727)*x_ref[1]**o+(0.174737141027)*x_ref[1]
        ref[3,1,1,0]=(-0.67938408823)*x_ref[0]**o+(0.862731161981)*x_ref[0]+(-0.792538485264)*x_ref[1]**o+(-0.0805197796157)*x_ref[1]
        ref[3,1,1,1]=(0.0113523229279)*x_ref[0]**o+(0.475393597561)*x_ref[0]+(-0.705748021886)*x_ref[1]**o+(0.00836460135285)*x_ref[1]
        ref[3,1,2,0]=(0.463839509799)*x_ref[0]**o+(-0.12663544643)*x_ref[0]+(0.353158399738)*x_ref[1]**o+(0.674929537597)*x_ref[1]
        ref[3,1,2,1]=(-0.6019896019)*x_ref[0]**o+(-0.638280685565)*x_ref[0]+(0.280047072988)*x_ref[1]**o+(0.569239958138)*x_ref[1]
        ref[3,2,0,0]=(-0.113906500346)*x_ref[0]**o+(-0.493494498501)*x_ref[0]+(-0.686419229162)*x_ref[1]**o+(-0.0727046763516)*x_ref[1]
        ref[3,2,0,1]=(-0.302092112795)*x_ref[0]**o+(-0.74819167)*x_ref[0]+(-0.478908975282)*x_ref[1]**o+(0.430024789627)*x_ref[1]
        ref[3,2,1,0]=(0.353535608984)*x_ref[0]**o+(-0.743408549264)*x_ref[0]+(-0.069122881538)*x_ref[1]**o+(-0.0767595089529)*x_ref[1]
        ref[3,2,1,1]=(-0.629946386234)*x_ref[0]**o+(0.851779590894)*x_ref[0]+(0.728577108695)*x_ref[1]**o+(-0.366983040147)*x_ref[1]
        ref[3,2,2,0]=(0.1685167508)*x_ref[0]**o+(0.149999747637)*x_ref[0]+(0.709320430374)*x_ref[1]**o+(-0.280950677222)*x_ref[1]
        ref[3,2,2,1]=(-0.956164974578)*x_ref[0]**o+(0.240460380901)*x_ref[0]+(-0.2820176577)*x_ref[1]**o+(-0.595667156955)*x_ref[1]
        ref[3,3,0,0]=(-0.0711295055938)*x_ref[0]**o+(-0.314513251791)*x_ref[0]+(-0.41924591293)*x_ref[1]**o+(0.0813197780959)*x_ref[1]
        ref[3,3,0,1]=(-0.230592584877)*x_ref[0]**o+(-0.488005144715)*x_ref[0]+(-0.450611759976)*x_ref[1]**o+(-0.0571152769768)*x_ref[1]
        ref[3,3,1,0]=(-0.608255699014)*x_ref[0]**o+(-0.697324108575)*x_ref[0]+(-0.885935900746)*x_ref[1]**o+(-0.884497550016)*x_ref[1]
        ref[3,3,1,1]=(0.857286225292)*x_ref[0]**o+(0.509954550626)*x_ref[0]+(0.182030821433)*x_ref[1]**o+(-0.268076060316)*x_ref[1]
        ref[3,3,2,0]=(-0.77322653479)*x_ref[0]**o+(-0.583013210018)*x_ref[0]+(-0.683867300277)*x_ref[1]**o+(0.754530541251)*x_ref[1]
        ref[3,3,2,1]=(-0.744667985181)*x_ref[0]**o+(0.164165128978)*x_ref[0]+(0.412340155958)*x_ref[1]**o+(-0.38451107118)*x_ref[1]
        ref[3,4,0,0]=(0.216720755049)*x_ref[0]**o+(-0.347803433722)*x_ref[0]+(0.417376268797)*x_ref[1]**o+(0.475787304652)*x_ref[1]
        ref[3,4,0,1]=(-0.737038295974)*x_ref[0]**o+(0.579172519178)*x_ref[0]+(-0.900272993672)*x_ref[1]**o+(0.407018603625)*x_ref[1]
        ref[3,4,1,0]=(-0.898959236474)*x_ref[0]**o+(-0.886636008121)*x_ref[0]+(-0.901972610044)*x_ref[1]**o+(-0.518393782318)*x_ref[1]
        ref[3,4,1,1]=(0.976342884921)*x_ref[0]**o+(-0.273568319493)*x_ref[0]+(-0.332424980846)*x_ref[1]**o+(-0.0146661731139)*x_ref[1]
        ref[3,4,2,0]=(0.794792214027)*x_ref[0]**o+(0.846927984413)*x_ref[0]+(0.535818809709)*x_ref[1]**o+(-0.49324400033)*x_ref[1]
        ref[3,4,2,1]=(0.695766308812)*x_ref[0]**o+(-0.37923153689)*x_ref[0]+(0.0633513026782)*x_ref[1]**o+(0.12334892105)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.401901998082)*x[0]**o+(-0.805091116808)*x[0]+(0.997659882792)*x[1]**o+(-0.953957389492)*x[1]+(-0.450185408289)*x[2]**o+(0.636472207773)*x[2]
        arg[0,0,0,1]=(-0.786386817128)*x[0]**o+(-0.882559775393)*x[0]+(0.117871400819)*x[1]**o+(-0.625922274774)*x[1]+(0.410431886931)*x[2]**o+(-0.979356501067)*x[2]
        arg[0,0,1,0]=(0.686025409275)*x[0]**o+(-0.19806907659)*x[0]+(-0.710420224449)*x[1]**o+(-0.839236953983)*x[1]+(-0.79971896093)*x[2]**o+(0.378944111319)*x[2]
        arg[0,0,1,1]=(0.771205112484)*x[0]**o+(-0.007083192859)*x[0]+(0.672918588001)*x[1]**o+(-0.351652495793)*x[1]+(-0.841767161187)*x[2]**o+(-0.345642373014)*x[2]
        arg[0,0,2,0]=(0.266930729015)*x[0]**o+(-0.873808152307)*x[0]+(0.206466759135)*x[1]**o+(0.539888649948)*x[1]+(0.17985862191)*x[2]**o+(0.240651460153)*x[2]
        arg[0,0,2,1]=(-0.232569768041)*x[0]**o+(-0.579841454072)*x[0]+(-0.706956104891)*x[1]**o+(0.822863815698)*x[1]+(0.345774991358)*x[2]**o+(-0.0848567668311)*x[2]
        arg[0,1,0,0]=(0.51593851507)*x[0]**o+(-0.233342434349)*x[0]+(-0.92498298026)*x[1]**o+(0.164540234829)*x[1]+(0.356559744499)*x[2]**o+(0.483902511717)*x[2]
        arg[0,1,0,1]=(-0.811509654443)*x[0]**o+(0.73929453189)*x[0]+(0.256205706418)*x[1]**o+(-0.667924913793)*x[1]+(0.266065641257)*x[2]**o+(0.856490258029)*x[2]
        arg[0,1,1,0]=(0.502051590284)*x[0]**o+(0.927283797128)*x[0]+(0.500854521411)*x[1]**o+(-0.486326996369)*x[1]+(-0.327578523685)*x[2]**o+(-0.28106070397)*x[2]
        arg[0,1,1,1]=(0.201740014513)*x[0]**o+(-0.166988241759)*x[0]+(-0.513072955566)*x[1]**o+(0.67179532417)*x[1]+(-0.72135980627)*x[2]**o+(0.763548886467)*x[2]
        arg[0,1,2,0]=(0.3995590338)*x[0]**o+(0.433286929621)*x[0]+(-0.287309453792)*x[1]**o+(-0.0671679443074)*x[1]+(-0.80815399993)*x[2]**o+(-0.398740362114)*x[2]
        arg[0,1,2,1]=(-0.16314640684)*x[0]**o+(-0.698681697367)*x[0]+(-0.11609738418)*x[1]**o+(-0.798465154739)*x[1]+(-0.169052628971)*x[2]**o+(-0.880316962539)*x[2]
        arg[0,2,0,0]=(-0.219731279442)*x[0]**o+(0.731464233112)*x[0]+(0.637811533993)*x[1]**o+(0.381218512692)*x[1]+(-0.181357790048)*x[2]**o+(-0.296793024723)*x[2]
        arg[0,2,0,1]=(0.566451624572)*x[0]**o+(0.302904935559)*x[0]+(-0.771526665232)*x[1]**o+(0.460577683302)*x[1]+(-0.921312455507)*x[2]**o+(0.604299765394)*x[2]
        arg[0,2,1,0]=(-0.0887516895197)*x[0]**o+(0.872208558285)*x[0]+(-0.306564750999)*x[1]**o+(0.841982500263)*x[1]+(-0.550398737046)*x[2]**o+(-0.869697162427)*x[2]
        arg[0,2,1,1]=(-0.652471383912)*x[0]**o+(0.634592304393)*x[0]+(0.0709212473167)*x[1]**o+(-0.671330082579)*x[1]+(-0.232895004478)*x[2]**o+(0.479008351165)*x[2]
        arg[0,2,2,0]=(-0.0764115046136)*x[0]**o+(0.649230346025)*x[0]+(0.544503480999)*x[1]**o+(0.788838981797)*x[1]+(0.416236148267)*x[2]**o+(0.722979541918)*x[2]
        arg[0,2,2,1]=(-0.0776368628837)*x[0]**o+(0.801984674386)*x[0]+(0.443670313486)*x[1]**o+(0.424676792255)*x[1]+(-0.895989937982)*x[2]**o+(0.0774648143589)*x[2]
        arg[0,3,0,0]=(-0.39041318438)*x[0]**o+(-0.174335409262)*x[0]+(0.658666482337)*x[1]**o+(0.23060605921)*x[1]+(-0.336603364419)*x[2]**o+(0.871530192198)*x[2]
        arg[0,3,0,1]=(-0.493104661363)*x[0]**o+(0.590779697499)*x[0]+(-0.903897457443)*x[1]**o+(-0.377640520704)*x[1]+(0.227174544833)*x[2]**o+(-0.860150483539)*x[2]
        arg[0,3,1,0]=(-0.093367253954)*x[0]**o+(0.159230499811)*x[0]+(-0.731396509936)*x[1]**o+(0.600088148838)*x[1]+(0.829175660147)*x[2]**o+(-0.435878280919)*x[2]
        arg[0,3,1,1]=(0.665125540711)*x[0]**o+(-0.92128737118)*x[0]+(-0.33260297543)*x[1]**o+(0.682569476816)*x[1]+(0.999761405058)*x[2]**o+(-0.748303702618)*x[2]
        arg[0,3,2,0]=(-0.430442277525)*x[0]**o+(-0.841437664378)*x[0]+(-0.374267371055)*x[1]**o+(0.667860327373)*x[1]+(0.694185699238)*x[2]**o+(-0.673629629622)*x[2]
        arg[0,3,2,1]=(0.304808617243)*x[0]**o+(-0.894407776037)*x[0]+(-0.758585281699)*x[1]**o+(0.20519963241)*x[1]+(0.538222015181)*x[2]**o+(-0.792962435573)*x[2]
        arg[0,4,0,0]=(0.568300300269)*x[0]**o+(0.407455936625)*x[0]+(0.980898344535)*x[1]**o+(0.988220474461)*x[1]+(0.0598892620089)*x[2]**o+(-0.279555207723)*x[2]
        arg[0,4,0,1]=(-0.208391854989)*x[0]**o+(0.9428019078)*x[0]+(-0.540916877275)*x[1]**o+(-0.427185538555)*x[1]+(-0.879268085857)*x[2]**o+(0.025102321099)*x[2]
        arg[0,4,1,0]=(0.412690112527)*x[0]**o+(-0.957376863734)*x[0]+(-0.913623139719)*x[1]**o+(-0.462418849979)*x[1]+(0.463336747019)*x[2]**o+(0.330747531293)*x[2]
        arg[0,4,1,1]=(-0.632140552118)*x[0]**o+(0.830769577671)*x[0]+(-0.888012251479)*x[1]**o+(-0.0560181569699)*x[1]+(-0.70068099248)*x[2]**o+(0.834483253039)*x[2]
        arg[0,4,2,0]=(0.0429717567385)*x[0]**o+(-0.691784042595)*x[0]+(-0.264988764713)*x[1]**o+(0.359873170931)*x[1]+(-0.116275128437)*x[2]**o+(0.0348873917345)*x[2]
        arg[0,4,2,1]=(-0.203754358311)*x[0]**o+(0.177896174646)*x[0]+(-0.0785920580727)*x[1]**o+(-0.124905949184)*x[1]+(0.968871355367)*x[2]**o+(-0.0369722247589)*x[2]
        arg[1,0,0,0]=(0.594851756786)*x[0]**o+(-0.594400822936)*x[0]+(0.868401350949)*x[1]**o+(0.186739616297)*x[1]+(-0.937219946754)*x[2]**o+(-0.248916422131)*x[2]
        arg[1,0,0,1]=(-0.639865978203)*x[0]**o+(-0.00990375501714)*x[0]+(-0.038266357411)*x[1]**o+(-0.741522380136)*x[1]+(0.460572922772)*x[2]**o+(0.4314205896)*x[2]
        arg[1,0,1,0]=(-0.460110413486)*x[0]**o+(-0.0964934735182)*x[0]+(0.474209024352)*x[1]**o+(-0.928476110805)*x[1]+(0.598621411487)*x[2]**o+(-0.00858812754345)*x[2]
        arg[1,0,1,1]=(0.936466172807)*x[0]**o+(-0.363937334197)*x[0]+(-0.540224651799)*x[1]**o+(0.484530722263)*x[1]+(0.675416993296)*x[2]**o+(0.89657344343)*x[2]
        arg[1,0,2,0]=(-0.713050692555)*x[0]**o+(-0.00851892179229)*x[0]+(0.689494113809)*x[1]**o+(0.633341028044)*x[1]+(0.272726911113)*x[2]**o+(-0.122017566819)*x[2]
        arg[1,0,2,1]=(0.793152080981)*x[0]**o+(0.287379739639)*x[0]+(0.304047372771)*x[1]**o+(-0.509079361053)*x[1]+(0.888302006985)*x[2]**o+(-0.297023045775)*x[2]
        arg[1,1,0,0]=(-0.69695118901)*x[0]**o+(0.505182668512)*x[0]+(0.827555590952)*x[1]**o+(-0.777232447056)*x[1]+(-0.460918138036)*x[2]**o+(0.782090206922)*x[2]
        arg[1,1,0,1]=(-0.435398535566)*x[0]**o+(-0.0997593681984)*x[0]+(0.140250252691)*x[1]**o+(-0.0204469256488)*x[1]+(-0.445836771028)*x[2]**o+(-0.775366037861)*x[2]
        arg[1,1,1,0]=(0.918015347957)*x[0]**o+(-0.491940111489)*x[0]+(0.793071617663)*x[1]**o+(-0.607035294859)*x[1]+(0.610076618566)*x[2]**o+(0.148116733513)*x[2]
        arg[1,1,1,1]=(0.224062518811)*x[0]**o+(-0.145897156735)*x[0]+(0.165380866901)*x[1]**o+(0.910562981814)*x[1]+(-0.406620561548)*x[2]**o+(0.528862812468)*x[2]
        arg[1,1,2,0]=(0.85148444093)*x[0]**o+(0.13210280653)*x[0]+(0.364136256712)*x[1]**o+(0.520388575386)*x[1]+(0.0145641287346)*x[2]**o+(-0.7849036351)*x[2]
        arg[1,1,2,1]=(0.493837193832)*x[0]**o+(-0.193326176569)*x[0]+(-0.665920972154)*x[1]**o+(-0.321645070095)*x[1]+(0.0209354853869)*x[2]**o+(-0.328315981752)*x[2]
        arg[1,2,0,0]=(-0.349281164766)*x[0]**o+(-0.318415198113)*x[0]+(-0.687972781234)*x[1]**o+(0.91825998605)*x[1]+(0.396868773798)*x[2]**o+(0.998204458027)*x[2]
        arg[1,2,0,1]=(-0.967493222935)*x[0]**o+(-0.721128465653)*x[0]+(-0.672178802974)*x[1]**o+(0.0924145224431)*x[1]+(0.114040587594)*x[2]**o+(0.707623895537)*x[2]
        arg[1,2,1,0]=(-0.253327088773)*x[0]**o+(0.206063827625)*x[0]+(0.355152654895)*x[1]**o+(-0.27339285621)*x[1]+(-0.0135567122291)*x[2]**o+(-0.983990446362)*x[2]
        arg[1,2,1,1]=(0.98471687132)*x[0]**o+(0.262902745655)*x[0]+(-0.589359138227)*x[1]**o+(-0.0982548958556)*x[1]+(-0.526161200133)*x[2]**o+(-0.850939484752)*x[2]
        arg[1,2,2,0]=(0.37583596705)*x[0]**o+(0.274845981222)*x[0]+(0.601331110943)*x[1]**o+(0.788294448218)*x[1]+(0.366445166459)*x[2]**o+(-0.554935072394)*x[2]
        arg[1,2,2,1]=(-0.986387850237)*x[0]**o+(0.68772989781)*x[0]+(0.345166604933)*x[1]**o+(0.930376726384)*x[1]+(0.760974421701)*x[2]**o+(0.261797396808)*x[2]
        arg[1,3,0,0]=(-0.171005013523)*x[0]**o+(-0.339229596812)*x[0]+(-0.150317419556)*x[1]**o+(-0.0146533427463)*x[1]+(0.957020663875)*x[2]**o+(0.616513940347)*x[2]
        arg[1,3,0,1]=(-0.0727328269499)*x[0]**o+(-0.635591748073)*x[0]+(-0.502426989774)*x[1]**o+(0.772595119619)*x[1]+(-0.76006511271)*x[2]**o+(-0.665920288084)*x[2]
        arg[1,3,1,0]=(-0.394284205622)*x[0]**o+(-0.105959165811)*x[0]+(0.72649045701)*x[1]**o+(-0.970276379473)*x[1]+(-0.236520349415)*x[2]**o+(0.994973267783)*x[2]
        arg[1,3,1,1]=(-0.444149867906)*x[0]**o+(0.748449201475)*x[0]+(0.419604170756)*x[1]**o+(-0.0291702705178)*x[1]+(-0.109580142628)*x[2]**o+(-0.0791870699838)*x[2]
        arg[1,3,2,0]=(-0.629657820523)*x[0]**o+(-0.995576552624)*x[0]+(0.530135462206)*x[1]**o+(0.197336335845)*x[1]+(0.0224815558807)*x[2]**o+(0.240498463912)*x[2]
        arg[1,3,2,1]=(0.52933948346)*x[0]**o+(0.98353913011)*x[0]+(0.0217614943138)*x[1]**o+(0.0157778453682)*x[1]+(0.177956932986)*x[2]**o+(0.234629778706)*x[2]
        arg[1,4,0,0]=(0.593028367137)*x[0]**o+(0.0340892538994)*x[0]+(-0.0295133170051)*x[1]**o+(0.0865921041963)*x[1]+(0.641131718656)*x[2]**o+(0.788858018609)*x[2]
        arg[1,4,0,1]=(-0.283333820634)*x[0]**o+(0.864482725099)*x[0]+(-0.560218794286)*x[1]**o+(-0.0083115362546)*x[1]+(0.726423285129)*x[2]**o+(0.031531339166)*x[2]
        arg[1,4,1,0]=(0.801452411554)*x[0]**o+(-0.59916393038)*x[0]+(-0.106568863591)*x[1]**o+(0.357442125133)*x[1]+(-0.203380280132)*x[2]**o+(-0.129031606441)*x[2]
        arg[1,4,1,1]=(-0.589220928637)*x[0]**o+(-0.308949389428)*x[0]+(0.785549858338)*x[1]**o+(-0.131153853015)*x[1]+(-0.30571855555)*x[2]**o+(-0.329634802588)*x[2]
        arg[1,4,2,0]=(-0.479686944)*x[0]**o+(-0.707702026873)*x[0]+(0.635894172352)*x[1]**o+(0.370673635069)*x[1]+(0.942862875991)*x[2]**o+(-0.846943456738)*x[2]
        arg[1,4,2,1]=(-0.403737324968)*x[0]**o+(0.101664426734)*x[0]+(-0.677093482354)*x[1]**o+(-0.36526066745)*x[1]+(-0.0889619136182)*x[2]**o+(-0.772491940844)*x[2]
        arg[2,0,0,0]=(0.753148325987)*x[0]**o+(-0.342843083128)*x[0]+(0.812661029553)*x[1]**o+(-0.112683035459)*x[1]+(0.900976470275)*x[2]**o+(0.00934086955102)*x[2]
        arg[2,0,0,1]=(-0.682315007143)*x[0]**o+(-0.694069900861)*x[0]+(-0.469082247747)*x[1]**o+(-0.522165903301)*x[1]+(-0.0997886221924)*x[2]**o+(-0.954161552419)*x[2]
        arg[2,0,1,0]=(0.780125539705)*x[0]**o+(-0.445331450225)*x[0]+(-0.282765510424)*x[1]**o+(-0.0131898361592)*x[1]+(-0.200186433923)*x[2]**o+(0.715139435696)*x[2]
        arg[2,0,1,1]=(-0.132310660786)*x[0]**o+(-0.0981344234369)*x[0]+(0.959297529475)*x[1]**o+(-0.85944894515)*x[1]+(-0.396258659619)*x[2]**o+(-0.09672066185)*x[2]
        arg[2,0,2,0]=(-0.344134014191)*x[0]**o+(0.534345864287)*x[0]+(0.797209509484)*x[1]**o+(0.0479050101499)*x[1]+(0.13620476987)*x[2]**o+(0.82708115383)*x[2]
        arg[2,0,2,1]=(-0.342509328238)*x[0]**o+(-0.833164493006)*x[0]+(0.472047721532)*x[1]**o+(0.301613158637)*x[1]+(0.858691679638)*x[2]**o+(-0.691744164006)*x[2]
        arg[2,1,0,0]=(0.498979390299)*x[0]**o+(0.371513881824)*x[0]+(-0.243506167971)*x[1]**o+(-0.245423890439)*x[1]+(-0.0494707840274)*x[2]**o+(0.0965573709928)*x[2]
        arg[2,1,0,1]=(0.665982455862)*x[0]**o+(-0.409710866336)*x[0]+(0.669257568046)*x[1]**o+(0.711947704788)*x[1]+(-0.0362568635785)*x[2]**o+(-0.30394278188)*x[2]
        arg[2,1,1,0]=(0.296245249928)*x[0]**o+(0.132631019281)*x[0]+(-0.376880152158)*x[1]**o+(0.303184156241)*x[1]+(0.836046560882)*x[2]**o+(0.195205645106)*x[2]
        arg[2,1,1,1]=(-0.971045820953)*x[0]**o+(0.154708896467)*x[0]+(-0.663844656602)*x[1]**o+(-0.0873797588245)*x[1]+(0.985123372542)*x[2]**o+(0.349554922131)*x[2]
        arg[2,1,2,0]=(-0.0117443431348)*x[0]**o+(-0.71980503985)*x[0]+(0.175348078994)*x[1]**o+(-0.883987892161)*x[1]+(0.0587275450458)*x[2]**o+(0.464811715548)*x[2]
        arg[2,1,2,1]=(-0.309202575765)*x[0]**o+(-0.528513248453)*x[0]+(-0.509298303972)*x[1]**o+(-0.228278863496)*x[1]+(0.933917053676)*x[2]**o+(-0.823678942264)*x[2]
        arg[2,2,0,0]=(0.374387386929)*x[0]**o+(0.64969993847)*x[0]+(-0.939822635641)*x[1]**o+(0.578249430422)*x[1]+(-0.973412660635)*x[2]**o+(0.484052718386)*x[2]
        arg[2,2,0,1]=(-0.972635162618)*x[0]**o+(-0.255007433802)*x[0]+(0.134165187097)*x[1]**o+(0.313213615976)*x[1]+(0.143298148686)*x[2]**o+(0.901819440229)*x[2]
        arg[2,2,1,0]=(0.868173256315)*x[0]**o+(-0.532325343453)*x[0]+(-0.712500880845)*x[1]**o+(0.216943462308)*x[1]+(0.733920091212)*x[2]**o+(0.0992400603383)*x[2]
        arg[2,2,1,1]=(0.642839350527)*x[0]**o+(0.127721543559)*x[0]+(-0.152686523032)*x[1]**o+(-0.399032361234)*x[1]+(0.317014518842)*x[2]**o+(0.905538790115)*x[2]
        arg[2,2,2,0]=(0.685267932413)*x[0]**o+(-0.167571547782)*x[0]+(-0.213446144405)*x[1]**o+(0.420270456377)*x[1]+(0.827725350621)*x[2]**o+(-0.920242274399)*x[2]
        arg[2,2,2,1]=(-0.507424593437)*x[0]**o+(-0.694222840019)*x[0]+(-0.904108970188)*x[1]**o+(-0.750899614966)*x[1]+(0.783997553302)*x[2]**o+(0.196327513447)*x[2]
        arg[2,3,0,0]=(-0.473900600479)*x[0]**o+(-0.464347097789)*x[0]+(-0.472423358671)*x[1]**o+(0.373786456364)*x[1]+(-0.989371300414)*x[2]**o+(-0.959508065239)*x[2]
        arg[2,3,0,1]=(0.174375189399)*x[0]**o+(0.925503059886)*x[0]+(-0.647274384723)*x[1]**o+(-0.577149479679)*x[1]+(0.171298877768)*x[2]**o+(-0.936612100882)*x[2]
        arg[2,3,1,0]=(0.482959343396)*x[0]**o+(0.842243554011)*x[0]+(0.814423633314)*x[1]**o+(-0.989445050315)*x[1]+(0.817933433065)*x[2]**o+(0.552396881404)*x[2]
        arg[2,3,1,1]=(0.55641736229)*x[0]**o+(0.498094640058)*x[0]+(0.0890113606037)*x[1]**o+(0.36592723553)*x[1]+(0.458607593386)*x[2]**o+(0.603047420063)*x[2]
        arg[2,3,2,0]=(0.884515080124)*x[0]**o+(-0.561898381022)*x[0]+(-0.292475503834)*x[1]**o+(0.425634423478)*x[1]+(0.65813323252)*x[2]**o+(-0.548850399842)*x[2]
        arg[2,3,2,1]=(-0.72895566982)*x[0]**o+(-0.569405588005)*x[0]+(0.117770875594)*x[1]**o+(-0.33515735969)*x[1]+(0.102062736382)*x[2]**o+(0.611532417683)*x[2]
        arg[2,4,0,0]=(-0.669208630424)*x[0]**o+(0.944393014177)*x[0]+(-0.00345511889683)*x[1]**o+(0.0695279595783)*x[1]+(0.496948250833)*x[2]**o+(-0.899750463971)*x[2]
        arg[2,4,0,1]=(-0.387828717292)*x[0]**o+(0.621172267004)*x[0]+(-0.280632674111)*x[1]**o+(0.401702567575)*x[1]+(-0.65554489378)*x[2]**o+(0.403018047007)*x[2]
        arg[2,4,1,0]=(-0.0770246925803)*x[0]**o+(0.52913302745)*x[0]+(0.922468724357)*x[1]**o+(-0.304442750439)*x[1]+(-0.556575079059)*x[2]**o+(-0.725790672251)*x[2]
        arg[2,4,1,1]=(-0.828378342911)*x[0]**o+(-0.0696370821776)*x[0]+(0.394018581667)*x[1]**o+(-0.251099848269)*x[1]+(-0.29878985656)*x[2]**o+(0.925753371056)*x[2]
        arg[2,4,2,0]=(-0.818277786029)*x[0]**o+(0.225862354706)*x[0]+(-0.629050212595)*x[1]**o+(0.139186107062)*x[1]+(-0.0134835519238)*x[2]**o+(0.301899528641)*x[2]
        arg[2,4,2,1]=(-0.783053684767)*x[0]**o+(-0.0926938783599)*x[0]+(0.519034210352)*x[1]**o+(-0.494564657126)*x[1]+(0.536117310622)*x[2]**o+(0.6278756634)*x[2]
        arg[3,0,0,0]=(-0.694353957685)*x[0]**o+(-0.361037363965)*x[0]+(-0.327004091885)*x[1]**o+(-0.765735142884)*x[1]+(-0.810915729713)*x[2]**o+(0.0640539311609)*x[2]
        arg[3,0,0,1]=(-0.131601548312)*x[0]**o+(0.271587756708)*x[0]+(0.322486168005)*x[1]**o+(0.238086581801)*x[1]+(-0.488081720555)*x[2]**o+(0.607278995635)*x[2]
        arg[3,0,1,0]=(0.903075418102)*x[0]**o+(-0.738609834848)*x[0]+(-0.165097989126)*x[1]**o+(-0.25967140914)*x[1]+(0.271649887428)*x[2]**o+(-0.786131648727)*x[2]
        arg[3,0,1,1]=(-0.963442989157)*x[0]**o+(-0.240205602334)*x[0]+(0.102153389731)*x[1]**o+(-0.106402434995)*x[1]+(-0.383097724211)*x[2]**o+(0.680290963195)*x[2]
        arg[3,0,2,0]=(-0.373162835796)*x[0]**o+(-0.186323761814)*x[0]+(0.661494315705)*x[1]**o+(0.320072189494)*x[1]+(0.0964443431067)*x[2]**o+(0.829015838241)*x[2]
        arg[3,0,2,1]=(0.527719514862)*x[0]**o+(0.0819948261665)*x[0]+(0.826598712178)*x[1]**o+(-0.588102975125)*x[1]+(-0.632584402292)*x[2]**o+(0.0826992233702)*x[2]
        arg[3,1,0,0]=(-0.102243686287)*x[0]**o+(-0.497336112929)*x[0]+(0.941827592628)*x[1]**o+(0.68678991944)*x[1]+(-0.508684841667)*x[2]**o+(-0.752939965161)*x[2]
        arg[3,1,0,1]=(-0.896417523183)*x[0]**o+(0.759224781685)*x[0]+(0.138537695823)*x[1]**o+(0.466727480471)*x[1]+(-0.838380824412)*x[2]**o+(0.285761157398)*x[2]
        arg[3,1,1,0]=(-0.547853182919)*x[0]**o+(-0.41595413414)*x[0]+(0.76894221584)*x[1]**o+(-0.142394275192)*x[1]+(-0.96349294581)*x[2]**o+(-0.873639964693)*x[2]
        arg[3,1,1,1]=(-0.151299403888)*x[0]**o+(0.789406063878)*x[0]+(0.163280670253)*x[1]**o+(-0.287315494745)*x[1]+(-0.627764670218)*x[2]**o+(0.842131632598)*x[2]
        arg[3,1,2,0]=(-0.957096489189)*x[0]**o+(0.0616037805677)*x[0]+(0.255858200166)*x[1]**o+(-0.430555450798)*x[1]+(0.730436930737)*x[2]**o+(0.405094271653)*x[2]
        arg[3,1,2,1]=(-0.924756832665)*x[0]**o+(0.525754162391)*x[0]+(0.281303228038)*x[1]**o+(-0.479500161878)*x[1]+(0.658416209181)*x[2]**o+(-0.0413881595557)*x[2]
        arg[3,2,0,0]=(0.516429381997)*x[0]**o+(-0.8387318467)*x[0]+(0.329773978227)*x[1]**o+(0.599641753085)*x[1]+(0.939937743271)*x[2]**o+(-0.247331169626)*x[2]
        arg[3,2,0,1]=(-0.644982044558)*x[0]**o+(-0.242033352774)*x[0]+(0.531835185779)*x[1]**o+(-0.851565064183)*x[1]+(-0.79213624915)*x[2]**o+(0.0607365690732)*x[2]
        arg[3,2,1,0]=(0.359781383537)*x[0]**o+(-0.0529212842074)*x[0]+(0.845915169452)*x[1]**o+(-0.323979570458)*x[1]+(0.828909997631)*x[2]**o+(-0.839886339138)*x[2]
        arg[3,2,1,1]=(0.559476630421)*x[0]**o+(-0.189725214615)*x[0]+(-0.00927446700413)*x[1]**o+(-0.234703720782)*x[1]+(-0.608362806755)*x[2]**o+(0.968571765534)*x[2]
        arg[3,2,2,0]=(0.924837507913)*x[0]**o+(0.846397611121)*x[0]+(-0.837765472403)*x[1]**o+(0.0225105985057)*x[1]+(-0.834187513844)*x[2]**o+(0.194311845578)*x[2]
        arg[3,2,2,1]=(-0.183561089961)*x[0]**o+(0.0745071729471)*x[0]+(-0.660415615202)*x[1]**o+(-0.528184997345)*x[1]+(0.205554094557)*x[2]**o+(-0.0309246318944)*x[2]
        arg[3,3,0,0]=(-0.67124095974)*x[0]**o+(-0.48501497083)*x[0]+(-0.0394345869094)*x[1]**o+(0.586692270429)*x[1]+(-0.593760749687)*x[2]**o+(-0.124979751327)*x[2]
        arg[3,3,0,1]=(0.778544173855)*x[0]**o+(-0.931466478494)*x[0]+(-0.746894777998)*x[1]**o+(0.806417936297)*x[1]+(0.989460929848)*x[2]**o+(0.666534366369)*x[2]
        arg[3,3,1,0]=(-0.798870838602)*x[0]**o+(-0.294618868937)*x[0]+(0.275029092968)*x[1]**o+(0.45390580364)*x[1]+(-0.311797818508)*x[2]**o+(0.751606624393)*x[2]
        arg[3,3,1,1]=(-0.924260987477)*x[0]**o+(0.771970726902)*x[0]+(-0.909099224213)*x[1]**o+(-0.45204206766)*x[1]+(0.630042614017)*x[2]**o+(-0.990915130464)*x[2]
        arg[3,3,2,0]=(0.734962154852)*x[0]**o+(-0.00621898011475)*x[0]+(-0.786362527189)*x[1]**o+(0.623507947406)*x[1]+(0.867175567104)*x[2]**o+(-0.383283255448)*x[2]
        arg[3,3,2,1]=(-0.955613156717)*x[0]**o+(0.688013754841)*x[0]+(-0.600559375208)*x[1]**o+(0.453342626645)*x[1]+(-0.864890477961)*x[2]**o+(-0.458029873063)*x[2]
        arg[3,4,0,0]=(-0.897010884834)*x[0]**o+(-0.251223576458)*x[0]+(-0.0338052047964)*x[1]**o+(-0.903421466132)*x[1]+(0.836864217874)*x[2]**o+(0.660805905175)*x[2]
        arg[3,4,0,1]=(-0.286025690215)*x[0]**o+(0.746732975604)*x[0]+(0.375810853218)*x[1]**o+(-0.985192974869)*x[1]+(-0.731463214455)*x[2]**o+(-0.465450352444)*x[2]
        arg[3,4,1,0]=(-0.500910000652)*x[0]**o+(0.387502419727)*x[0]+(0.962369164802)*x[1]**o+(-0.10933824251)*x[1]+(-0.33642703989)*x[2]**o+(0.546851133256)*x[2]
        arg[3,4,1,1]=(-0.117959459645)*x[0]**o+(0.867699323392)*x[0]+(0.950506403861)*x[1]**o+(0.613806207745)*x[1]+(0.425723173807)*x[2]**o+(-0.906937548619)*x[2]
        arg[3,4,2,0]=(0.331114892194)*x[0]**o+(-0.168829643424)*x[0]+(-0.855962311982)*x[1]**o+(-0.144202406673)*x[1]+(0.0769985713088)*x[2]**o+(0.856003746423)*x[2]
        arg[3,4,2,1]=(-0.439397439364)*x[0]**o+(0.686116343481)*x[0]+(0.519883854523)*x[1]**o+(0.716231469289)*x[1]+(0.270098231149)*x[2]**o+(-0.448525436427)*x[2]
        ref[0,0,0,0]=(-0.401901998082)*x_ref[0]**o+(-0.805091116808)*x_ref[0]+(0.997659882792)*x_ref[1]**o+(-0.953957389492)*x_ref[1]+(-0.450185408289)*x_ref[2]**o+(0.636472207773)*x_ref[2]
        ref[0,0,0,1]=(-0.786386817128)*x_ref[0]**o+(-0.882559775393)*x_ref[0]+(0.117871400819)*x_ref[1]**o+(-0.625922274774)*x_ref[1]+(0.410431886931)*x_ref[2]**o+(-0.979356501067)*x_ref[2]
        ref[0,0,1,0]=(0.686025409275)*x_ref[0]**o+(-0.19806907659)*x_ref[0]+(-0.710420224449)*x_ref[1]**o+(-0.839236953983)*x_ref[1]+(-0.79971896093)*x_ref[2]**o+(0.378944111319)*x_ref[2]
        ref[0,0,1,1]=(0.771205112484)*x_ref[0]**o+(-0.007083192859)*x_ref[0]+(0.672918588001)*x_ref[1]**o+(-0.351652495793)*x_ref[1]+(-0.841767161187)*x_ref[2]**o+(-0.345642373014)*x_ref[2]
        ref[0,0,2,0]=(0.266930729015)*x_ref[0]**o+(-0.873808152307)*x_ref[0]+(0.206466759135)*x_ref[1]**o+(0.539888649948)*x_ref[1]+(0.17985862191)*x_ref[2]**o+(0.240651460153)*x_ref[2]
        ref[0,0,2,1]=(-0.232569768041)*x_ref[0]**o+(-0.579841454072)*x_ref[0]+(-0.706956104891)*x_ref[1]**o+(0.822863815698)*x_ref[1]+(0.345774991358)*x_ref[2]**o+(-0.0848567668311)*x_ref[2]
        ref[0,1,0,0]=(0.51593851507)*x_ref[0]**o+(-0.233342434349)*x_ref[0]+(-0.92498298026)*x_ref[1]**o+(0.164540234829)*x_ref[1]+(0.356559744499)*x_ref[2]**o+(0.483902511717)*x_ref[2]
        ref[0,1,0,1]=(-0.811509654443)*x_ref[0]**o+(0.73929453189)*x_ref[0]+(0.256205706418)*x_ref[1]**o+(-0.667924913793)*x_ref[1]+(0.266065641257)*x_ref[2]**o+(0.856490258029)*x_ref[2]
        ref[0,1,1,0]=(0.502051590284)*x_ref[0]**o+(0.927283797128)*x_ref[0]+(0.500854521411)*x_ref[1]**o+(-0.486326996369)*x_ref[1]+(-0.327578523685)*x_ref[2]**o+(-0.28106070397)*x_ref[2]
        ref[0,1,1,1]=(0.201740014513)*x_ref[0]**o+(-0.166988241759)*x_ref[0]+(-0.513072955566)*x_ref[1]**o+(0.67179532417)*x_ref[1]+(-0.72135980627)*x_ref[2]**o+(0.763548886467)*x_ref[2]
        ref[0,1,2,0]=(0.3995590338)*x_ref[0]**o+(0.433286929621)*x_ref[0]+(-0.287309453792)*x_ref[1]**o+(-0.0671679443074)*x_ref[1]+(-0.80815399993)*x_ref[2]**o+(-0.398740362114)*x_ref[2]
        ref[0,1,2,1]=(-0.16314640684)*x_ref[0]**o+(-0.698681697367)*x_ref[0]+(-0.11609738418)*x_ref[1]**o+(-0.798465154739)*x_ref[1]+(-0.169052628971)*x_ref[2]**o+(-0.880316962539)*x_ref[2]
        ref[0,2,0,0]=(-0.219731279442)*x_ref[0]**o+(0.731464233112)*x_ref[0]+(0.637811533993)*x_ref[1]**o+(0.381218512692)*x_ref[1]+(-0.181357790048)*x_ref[2]**o+(-0.296793024723)*x_ref[2]
        ref[0,2,0,1]=(0.566451624572)*x_ref[0]**o+(0.302904935559)*x_ref[0]+(-0.771526665232)*x_ref[1]**o+(0.460577683302)*x_ref[1]+(-0.921312455507)*x_ref[2]**o+(0.604299765394)*x_ref[2]
        ref[0,2,1,0]=(-0.0887516895197)*x_ref[0]**o+(0.872208558285)*x_ref[0]+(-0.306564750999)*x_ref[1]**o+(0.841982500263)*x_ref[1]+(-0.550398737046)*x_ref[2]**o+(-0.869697162427)*x_ref[2]
        ref[0,2,1,1]=(-0.652471383912)*x_ref[0]**o+(0.634592304393)*x_ref[0]+(0.0709212473167)*x_ref[1]**o+(-0.671330082579)*x_ref[1]+(-0.232895004478)*x_ref[2]**o+(0.479008351165)*x_ref[2]
        ref[0,2,2,0]=(-0.0764115046136)*x_ref[0]**o+(0.649230346025)*x_ref[0]+(0.544503480999)*x_ref[1]**o+(0.788838981797)*x_ref[1]+(0.416236148267)*x_ref[2]**o+(0.722979541918)*x_ref[2]
        ref[0,2,2,1]=(-0.0776368628837)*x_ref[0]**o+(0.801984674386)*x_ref[0]+(0.443670313486)*x_ref[1]**o+(0.424676792255)*x_ref[1]+(-0.895989937982)*x_ref[2]**o+(0.0774648143589)*x_ref[2]
        ref[0,3,0,0]=(-0.39041318438)*x_ref[0]**o+(-0.174335409262)*x_ref[0]+(0.658666482337)*x_ref[1]**o+(0.23060605921)*x_ref[1]+(-0.336603364419)*x_ref[2]**o+(0.871530192198)*x_ref[2]
        ref[0,3,0,1]=(-0.493104661363)*x_ref[0]**o+(0.590779697499)*x_ref[0]+(-0.903897457443)*x_ref[1]**o+(-0.377640520704)*x_ref[1]+(0.227174544833)*x_ref[2]**o+(-0.860150483539)*x_ref[2]
        ref[0,3,1,0]=(-0.093367253954)*x_ref[0]**o+(0.159230499811)*x_ref[0]+(-0.731396509936)*x_ref[1]**o+(0.600088148838)*x_ref[1]+(0.829175660147)*x_ref[2]**o+(-0.435878280919)*x_ref[2]
        ref[0,3,1,1]=(0.665125540711)*x_ref[0]**o+(-0.92128737118)*x_ref[0]+(-0.33260297543)*x_ref[1]**o+(0.682569476816)*x_ref[1]+(0.999761405058)*x_ref[2]**o+(-0.748303702618)*x_ref[2]
        ref[0,3,2,0]=(-0.430442277525)*x_ref[0]**o+(-0.841437664378)*x_ref[0]+(-0.374267371055)*x_ref[1]**o+(0.667860327373)*x_ref[1]+(0.694185699238)*x_ref[2]**o+(-0.673629629622)*x_ref[2]
        ref[0,3,2,1]=(0.304808617243)*x_ref[0]**o+(-0.894407776037)*x_ref[0]+(-0.758585281699)*x_ref[1]**o+(0.20519963241)*x_ref[1]+(0.538222015181)*x_ref[2]**o+(-0.792962435573)*x_ref[2]
        ref[0,4,0,0]=(0.568300300269)*x_ref[0]**o+(0.407455936625)*x_ref[0]+(0.980898344535)*x_ref[1]**o+(0.988220474461)*x_ref[1]+(0.0598892620089)*x_ref[2]**o+(-0.279555207723)*x_ref[2]
        ref[0,4,0,1]=(-0.208391854989)*x_ref[0]**o+(0.9428019078)*x_ref[0]+(-0.540916877275)*x_ref[1]**o+(-0.427185538555)*x_ref[1]+(-0.879268085857)*x_ref[2]**o+(0.025102321099)*x_ref[2]
        ref[0,4,1,0]=(0.412690112527)*x_ref[0]**o+(-0.957376863734)*x_ref[0]+(-0.913623139719)*x_ref[1]**o+(-0.462418849979)*x_ref[1]+(0.463336747019)*x_ref[2]**o+(0.330747531293)*x_ref[2]
        ref[0,4,1,1]=(-0.632140552118)*x_ref[0]**o+(0.830769577671)*x_ref[0]+(-0.888012251479)*x_ref[1]**o+(-0.0560181569699)*x_ref[1]+(-0.70068099248)*x_ref[2]**o+(0.834483253039)*x_ref[2]
        ref[0,4,2,0]=(0.0429717567385)*x_ref[0]**o+(-0.691784042595)*x_ref[0]+(-0.264988764713)*x_ref[1]**o+(0.359873170931)*x_ref[1]+(-0.116275128437)*x_ref[2]**o+(0.0348873917345)*x_ref[2]
        ref[0,4,2,1]=(-0.203754358311)*x_ref[0]**o+(0.177896174646)*x_ref[0]+(-0.0785920580727)*x_ref[1]**o+(-0.124905949184)*x_ref[1]+(0.968871355367)*x_ref[2]**o+(-0.0369722247589)*x_ref[2]
        ref[1,0,0,0]=(0.594851756786)*x_ref[0]**o+(-0.594400822936)*x_ref[0]+(0.868401350949)*x_ref[1]**o+(0.186739616297)*x_ref[1]+(-0.937219946754)*x_ref[2]**o+(-0.248916422131)*x_ref[2]
        ref[1,0,0,1]=(-0.639865978203)*x_ref[0]**o+(-0.00990375501714)*x_ref[0]+(-0.038266357411)*x_ref[1]**o+(-0.741522380136)*x_ref[1]+(0.460572922772)*x_ref[2]**o+(0.4314205896)*x_ref[2]
        ref[1,0,1,0]=(-0.460110413486)*x_ref[0]**o+(-0.0964934735182)*x_ref[0]+(0.474209024352)*x_ref[1]**o+(-0.928476110805)*x_ref[1]+(0.598621411487)*x_ref[2]**o+(-0.00858812754345)*x_ref[2]
        ref[1,0,1,1]=(0.936466172807)*x_ref[0]**o+(-0.363937334197)*x_ref[0]+(-0.540224651799)*x_ref[1]**o+(0.484530722263)*x_ref[1]+(0.675416993296)*x_ref[2]**o+(0.89657344343)*x_ref[2]
        ref[1,0,2,0]=(-0.713050692555)*x_ref[0]**o+(-0.00851892179229)*x_ref[0]+(0.689494113809)*x_ref[1]**o+(0.633341028044)*x_ref[1]+(0.272726911113)*x_ref[2]**o+(-0.122017566819)*x_ref[2]
        ref[1,0,2,1]=(0.793152080981)*x_ref[0]**o+(0.287379739639)*x_ref[0]+(0.304047372771)*x_ref[1]**o+(-0.509079361053)*x_ref[1]+(0.888302006985)*x_ref[2]**o+(-0.297023045775)*x_ref[2]
        ref[1,1,0,0]=(-0.69695118901)*x_ref[0]**o+(0.505182668512)*x_ref[0]+(0.827555590952)*x_ref[1]**o+(-0.777232447056)*x_ref[1]+(-0.460918138036)*x_ref[2]**o+(0.782090206922)*x_ref[2]
        ref[1,1,0,1]=(-0.435398535566)*x_ref[0]**o+(-0.0997593681984)*x_ref[0]+(0.140250252691)*x_ref[1]**o+(-0.0204469256488)*x_ref[1]+(-0.445836771028)*x_ref[2]**o+(-0.775366037861)*x_ref[2]
        ref[1,1,1,0]=(0.918015347957)*x_ref[0]**o+(-0.491940111489)*x_ref[0]+(0.793071617663)*x_ref[1]**o+(-0.607035294859)*x_ref[1]+(0.610076618566)*x_ref[2]**o+(0.148116733513)*x_ref[2]
        ref[1,1,1,1]=(0.224062518811)*x_ref[0]**o+(-0.145897156735)*x_ref[0]+(0.165380866901)*x_ref[1]**o+(0.910562981814)*x_ref[1]+(-0.406620561548)*x_ref[2]**o+(0.528862812468)*x_ref[2]
        ref[1,1,2,0]=(0.85148444093)*x_ref[0]**o+(0.13210280653)*x_ref[0]+(0.364136256712)*x_ref[1]**o+(0.520388575386)*x_ref[1]+(0.0145641287346)*x_ref[2]**o+(-0.7849036351)*x_ref[2]
        ref[1,1,2,1]=(0.493837193832)*x_ref[0]**o+(-0.193326176569)*x_ref[0]+(-0.665920972154)*x_ref[1]**o+(-0.321645070095)*x_ref[1]+(0.0209354853869)*x_ref[2]**o+(-0.328315981752)*x_ref[2]
        ref[1,2,0,0]=(-0.349281164766)*x_ref[0]**o+(-0.318415198113)*x_ref[0]+(-0.687972781234)*x_ref[1]**o+(0.91825998605)*x_ref[1]+(0.396868773798)*x_ref[2]**o+(0.998204458027)*x_ref[2]
        ref[1,2,0,1]=(-0.967493222935)*x_ref[0]**o+(-0.721128465653)*x_ref[0]+(-0.672178802974)*x_ref[1]**o+(0.0924145224431)*x_ref[1]+(0.114040587594)*x_ref[2]**o+(0.707623895537)*x_ref[2]
        ref[1,2,1,0]=(-0.253327088773)*x_ref[0]**o+(0.206063827625)*x_ref[0]+(0.355152654895)*x_ref[1]**o+(-0.27339285621)*x_ref[1]+(-0.0135567122291)*x_ref[2]**o+(-0.983990446362)*x_ref[2]
        ref[1,2,1,1]=(0.98471687132)*x_ref[0]**o+(0.262902745655)*x_ref[0]+(-0.589359138227)*x_ref[1]**o+(-0.0982548958556)*x_ref[1]+(-0.526161200133)*x_ref[2]**o+(-0.850939484752)*x_ref[2]
        ref[1,2,2,0]=(0.37583596705)*x_ref[0]**o+(0.274845981222)*x_ref[0]+(0.601331110943)*x_ref[1]**o+(0.788294448218)*x_ref[1]+(0.366445166459)*x_ref[2]**o+(-0.554935072394)*x_ref[2]
        ref[1,2,2,1]=(-0.986387850237)*x_ref[0]**o+(0.68772989781)*x_ref[0]+(0.345166604933)*x_ref[1]**o+(0.930376726384)*x_ref[1]+(0.760974421701)*x_ref[2]**o+(0.261797396808)*x_ref[2]
        ref[1,3,0,0]=(-0.171005013523)*x_ref[0]**o+(-0.339229596812)*x_ref[0]+(-0.150317419556)*x_ref[1]**o+(-0.0146533427463)*x_ref[1]+(0.957020663875)*x_ref[2]**o+(0.616513940347)*x_ref[2]
        ref[1,3,0,1]=(-0.0727328269499)*x_ref[0]**o+(-0.635591748073)*x_ref[0]+(-0.502426989774)*x_ref[1]**o+(0.772595119619)*x_ref[1]+(-0.76006511271)*x_ref[2]**o+(-0.665920288084)*x_ref[2]
        ref[1,3,1,0]=(-0.394284205622)*x_ref[0]**o+(-0.105959165811)*x_ref[0]+(0.72649045701)*x_ref[1]**o+(-0.970276379473)*x_ref[1]+(-0.236520349415)*x_ref[2]**o+(0.994973267783)*x_ref[2]
        ref[1,3,1,1]=(-0.444149867906)*x_ref[0]**o+(0.748449201475)*x_ref[0]+(0.419604170756)*x_ref[1]**o+(-0.0291702705178)*x_ref[1]+(-0.109580142628)*x_ref[2]**o+(-0.0791870699838)*x_ref[2]
        ref[1,3,2,0]=(-0.629657820523)*x_ref[0]**o+(-0.995576552624)*x_ref[0]+(0.530135462206)*x_ref[1]**o+(0.197336335845)*x_ref[1]+(0.0224815558807)*x_ref[2]**o+(0.240498463912)*x_ref[2]
        ref[1,3,2,1]=(0.52933948346)*x_ref[0]**o+(0.98353913011)*x_ref[0]+(0.0217614943138)*x_ref[1]**o+(0.0157778453682)*x_ref[1]+(0.177956932986)*x_ref[2]**o+(0.234629778706)*x_ref[2]
        ref[1,4,0,0]=(0.593028367137)*x_ref[0]**o+(0.0340892538994)*x_ref[0]+(-0.0295133170051)*x_ref[1]**o+(0.0865921041963)*x_ref[1]+(0.641131718656)*x_ref[2]**o+(0.788858018609)*x_ref[2]
        ref[1,4,0,1]=(-0.283333820634)*x_ref[0]**o+(0.864482725099)*x_ref[0]+(-0.560218794286)*x_ref[1]**o+(-0.0083115362546)*x_ref[1]+(0.726423285129)*x_ref[2]**o+(0.031531339166)*x_ref[2]
        ref[1,4,1,0]=(0.801452411554)*x_ref[0]**o+(-0.59916393038)*x_ref[0]+(-0.106568863591)*x_ref[1]**o+(0.357442125133)*x_ref[1]+(-0.203380280132)*x_ref[2]**o+(-0.129031606441)*x_ref[2]
        ref[1,4,1,1]=(-0.589220928637)*x_ref[0]**o+(-0.308949389428)*x_ref[0]+(0.785549858338)*x_ref[1]**o+(-0.131153853015)*x_ref[1]+(-0.30571855555)*x_ref[2]**o+(-0.329634802588)*x_ref[2]
        ref[1,4,2,0]=(-0.479686944)*x_ref[0]**o+(-0.707702026873)*x_ref[0]+(0.635894172352)*x_ref[1]**o+(0.370673635069)*x_ref[1]+(0.942862875991)*x_ref[2]**o+(-0.846943456738)*x_ref[2]
        ref[1,4,2,1]=(-0.403737324968)*x_ref[0]**o+(0.101664426734)*x_ref[0]+(-0.677093482354)*x_ref[1]**o+(-0.36526066745)*x_ref[1]+(-0.0889619136182)*x_ref[2]**o+(-0.772491940844)*x_ref[2]
        ref[2,0,0,0]=(0.753148325987)*x_ref[0]**o+(-0.342843083128)*x_ref[0]+(0.812661029553)*x_ref[1]**o+(-0.112683035459)*x_ref[1]+(0.900976470275)*x_ref[2]**o+(0.00934086955102)*x_ref[2]
        ref[2,0,0,1]=(-0.682315007143)*x_ref[0]**o+(-0.694069900861)*x_ref[0]+(-0.469082247747)*x_ref[1]**o+(-0.522165903301)*x_ref[1]+(-0.0997886221924)*x_ref[2]**o+(-0.954161552419)*x_ref[2]
        ref[2,0,1,0]=(0.780125539705)*x_ref[0]**o+(-0.445331450225)*x_ref[0]+(-0.282765510424)*x_ref[1]**o+(-0.0131898361592)*x_ref[1]+(-0.200186433923)*x_ref[2]**o+(0.715139435696)*x_ref[2]
        ref[2,0,1,1]=(-0.132310660786)*x_ref[0]**o+(-0.0981344234369)*x_ref[0]+(0.959297529475)*x_ref[1]**o+(-0.85944894515)*x_ref[1]+(-0.396258659619)*x_ref[2]**o+(-0.09672066185)*x_ref[2]
        ref[2,0,2,0]=(-0.344134014191)*x_ref[0]**o+(0.534345864287)*x_ref[0]+(0.797209509484)*x_ref[1]**o+(0.0479050101499)*x_ref[1]+(0.13620476987)*x_ref[2]**o+(0.82708115383)*x_ref[2]
        ref[2,0,2,1]=(-0.342509328238)*x_ref[0]**o+(-0.833164493006)*x_ref[0]+(0.472047721532)*x_ref[1]**o+(0.301613158637)*x_ref[1]+(0.858691679638)*x_ref[2]**o+(-0.691744164006)*x_ref[2]
        ref[2,1,0,0]=(0.498979390299)*x_ref[0]**o+(0.371513881824)*x_ref[0]+(-0.243506167971)*x_ref[1]**o+(-0.245423890439)*x_ref[1]+(-0.0494707840274)*x_ref[2]**o+(0.0965573709928)*x_ref[2]
        ref[2,1,0,1]=(0.665982455862)*x_ref[0]**o+(-0.409710866336)*x_ref[0]+(0.669257568046)*x_ref[1]**o+(0.711947704788)*x_ref[1]+(-0.0362568635785)*x_ref[2]**o+(-0.30394278188)*x_ref[2]
        ref[2,1,1,0]=(0.296245249928)*x_ref[0]**o+(0.132631019281)*x_ref[0]+(-0.376880152158)*x_ref[1]**o+(0.303184156241)*x_ref[1]+(0.836046560882)*x_ref[2]**o+(0.195205645106)*x_ref[2]
        ref[2,1,1,1]=(-0.971045820953)*x_ref[0]**o+(0.154708896467)*x_ref[0]+(-0.663844656602)*x_ref[1]**o+(-0.0873797588245)*x_ref[1]+(0.985123372542)*x_ref[2]**o+(0.349554922131)*x_ref[2]
        ref[2,1,2,0]=(-0.0117443431348)*x_ref[0]**o+(-0.71980503985)*x_ref[0]+(0.175348078994)*x_ref[1]**o+(-0.883987892161)*x_ref[1]+(0.0587275450458)*x_ref[2]**o+(0.464811715548)*x_ref[2]
        ref[2,1,2,1]=(-0.309202575765)*x_ref[0]**o+(-0.528513248453)*x_ref[0]+(-0.509298303972)*x_ref[1]**o+(-0.228278863496)*x_ref[1]+(0.933917053676)*x_ref[2]**o+(-0.823678942264)*x_ref[2]
        ref[2,2,0,0]=(0.374387386929)*x_ref[0]**o+(0.64969993847)*x_ref[0]+(-0.939822635641)*x_ref[1]**o+(0.578249430422)*x_ref[1]+(-0.973412660635)*x_ref[2]**o+(0.484052718386)*x_ref[2]
        ref[2,2,0,1]=(-0.972635162618)*x_ref[0]**o+(-0.255007433802)*x_ref[0]+(0.134165187097)*x_ref[1]**o+(0.313213615976)*x_ref[1]+(0.143298148686)*x_ref[2]**o+(0.901819440229)*x_ref[2]
        ref[2,2,1,0]=(0.868173256315)*x_ref[0]**o+(-0.532325343453)*x_ref[0]+(-0.712500880845)*x_ref[1]**o+(0.216943462308)*x_ref[1]+(0.733920091212)*x_ref[2]**o+(0.0992400603383)*x_ref[2]
        ref[2,2,1,1]=(0.642839350527)*x_ref[0]**o+(0.127721543559)*x_ref[0]+(-0.152686523032)*x_ref[1]**o+(-0.399032361234)*x_ref[1]+(0.317014518842)*x_ref[2]**o+(0.905538790115)*x_ref[2]
        ref[2,2,2,0]=(0.685267932413)*x_ref[0]**o+(-0.167571547782)*x_ref[0]+(-0.213446144405)*x_ref[1]**o+(0.420270456377)*x_ref[1]+(0.827725350621)*x_ref[2]**o+(-0.920242274399)*x_ref[2]
        ref[2,2,2,1]=(-0.507424593437)*x_ref[0]**o+(-0.694222840019)*x_ref[0]+(-0.904108970188)*x_ref[1]**o+(-0.750899614966)*x_ref[1]+(0.783997553302)*x_ref[2]**o+(0.196327513447)*x_ref[2]
        ref[2,3,0,0]=(-0.473900600479)*x_ref[0]**o+(-0.464347097789)*x_ref[0]+(-0.472423358671)*x_ref[1]**o+(0.373786456364)*x_ref[1]+(-0.989371300414)*x_ref[2]**o+(-0.959508065239)*x_ref[2]
        ref[2,3,0,1]=(0.174375189399)*x_ref[0]**o+(0.925503059886)*x_ref[0]+(-0.647274384723)*x_ref[1]**o+(-0.577149479679)*x_ref[1]+(0.171298877768)*x_ref[2]**o+(-0.936612100882)*x_ref[2]
        ref[2,3,1,0]=(0.482959343396)*x_ref[0]**o+(0.842243554011)*x_ref[0]+(0.814423633314)*x_ref[1]**o+(-0.989445050315)*x_ref[1]+(0.817933433065)*x_ref[2]**o+(0.552396881404)*x_ref[2]
        ref[2,3,1,1]=(0.55641736229)*x_ref[0]**o+(0.498094640058)*x_ref[0]+(0.0890113606037)*x_ref[1]**o+(0.36592723553)*x_ref[1]+(0.458607593386)*x_ref[2]**o+(0.603047420063)*x_ref[2]
        ref[2,3,2,0]=(0.884515080124)*x_ref[0]**o+(-0.561898381022)*x_ref[0]+(-0.292475503834)*x_ref[1]**o+(0.425634423478)*x_ref[1]+(0.65813323252)*x_ref[2]**o+(-0.548850399842)*x_ref[2]
        ref[2,3,2,1]=(-0.72895566982)*x_ref[0]**o+(-0.569405588005)*x_ref[0]+(0.117770875594)*x_ref[1]**o+(-0.33515735969)*x_ref[1]+(0.102062736382)*x_ref[2]**o+(0.611532417683)*x_ref[2]
        ref[2,4,0,0]=(-0.669208630424)*x_ref[0]**o+(0.944393014177)*x_ref[0]+(-0.00345511889683)*x_ref[1]**o+(0.0695279595783)*x_ref[1]+(0.496948250833)*x_ref[2]**o+(-0.899750463971)*x_ref[2]
        ref[2,4,0,1]=(-0.387828717292)*x_ref[0]**o+(0.621172267004)*x_ref[0]+(-0.280632674111)*x_ref[1]**o+(0.401702567575)*x_ref[1]+(-0.65554489378)*x_ref[2]**o+(0.403018047007)*x_ref[2]
        ref[2,4,1,0]=(-0.0770246925803)*x_ref[0]**o+(0.52913302745)*x_ref[0]+(0.922468724357)*x_ref[1]**o+(-0.304442750439)*x_ref[1]+(-0.556575079059)*x_ref[2]**o+(-0.725790672251)*x_ref[2]
        ref[2,4,1,1]=(-0.828378342911)*x_ref[0]**o+(-0.0696370821776)*x_ref[0]+(0.394018581667)*x_ref[1]**o+(-0.251099848269)*x_ref[1]+(-0.29878985656)*x_ref[2]**o+(0.925753371056)*x_ref[2]
        ref[2,4,2,0]=(-0.818277786029)*x_ref[0]**o+(0.225862354706)*x_ref[0]+(-0.629050212595)*x_ref[1]**o+(0.139186107062)*x_ref[1]+(-0.0134835519238)*x_ref[2]**o+(0.301899528641)*x_ref[2]
        ref[2,4,2,1]=(-0.783053684767)*x_ref[0]**o+(-0.0926938783599)*x_ref[0]+(0.519034210352)*x_ref[1]**o+(-0.494564657126)*x_ref[1]+(0.536117310622)*x_ref[2]**o+(0.6278756634)*x_ref[2]
        ref[3,0,0,0]=(-0.694353957685)*x_ref[0]**o+(-0.361037363965)*x_ref[0]+(-0.327004091885)*x_ref[1]**o+(-0.765735142884)*x_ref[1]+(-0.810915729713)*x_ref[2]**o+(0.0640539311609)*x_ref[2]
        ref[3,0,0,1]=(-0.131601548312)*x_ref[0]**o+(0.271587756708)*x_ref[0]+(0.322486168005)*x_ref[1]**o+(0.238086581801)*x_ref[1]+(-0.488081720555)*x_ref[2]**o+(0.607278995635)*x_ref[2]
        ref[3,0,1,0]=(0.903075418102)*x_ref[0]**o+(-0.738609834848)*x_ref[0]+(-0.165097989126)*x_ref[1]**o+(-0.25967140914)*x_ref[1]+(0.271649887428)*x_ref[2]**o+(-0.786131648727)*x_ref[2]
        ref[3,0,1,1]=(-0.963442989157)*x_ref[0]**o+(-0.240205602334)*x_ref[0]+(0.102153389731)*x_ref[1]**o+(-0.106402434995)*x_ref[1]+(-0.383097724211)*x_ref[2]**o+(0.680290963195)*x_ref[2]
        ref[3,0,2,0]=(-0.373162835796)*x_ref[0]**o+(-0.186323761814)*x_ref[0]+(0.661494315705)*x_ref[1]**o+(0.320072189494)*x_ref[1]+(0.0964443431067)*x_ref[2]**o+(0.829015838241)*x_ref[2]
        ref[3,0,2,1]=(0.527719514862)*x_ref[0]**o+(0.0819948261665)*x_ref[0]+(0.826598712178)*x_ref[1]**o+(-0.588102975125)*x_ref[1]+(-0.632584402292)*x_ref[2]**o+(0.0826992233702)*x_ref[2]
        ref[3,1,0,0]=(-0.102243686287)*x_ref[0]**o+(-0.497336112929)*x_ref[0]+(0.941827592628)*x_ref[1]**o+(0.68678991944)*x_ref[1]+(-0.508684841667)*x_ref[2]**o+(-0.752939965161)*x_ref[2]
        ref[3,1,0,1]=(-0.896417523183)*x_ref[0]**o+(0.759224781685)*x_ref[0]+(0.138537695823)*x_ref[1]**o+(0.466727480471)*x_ref[1]+(-0.838380824412)*x_ref[2]**o+(0.285761157398)*x_ref[2]
        ref[3,1,1,0]=(-0.547853182919)*x_ref[0]**o+(-0.41595413414)*x_ref[0]+(0.76894221584)*x_ref[1]**o+(-0.142394275192)*x_ref[1]+(-0.96349294581)*x_ref[2]**o+(-0.873639964693)*x_ref[2]
        ref[3,1,1,1]=(-0.151299403888)*x_ref[0]**o+(0.789406063878)*x_ref[0]+(0.163280670253)*x_ref[1]**o+(-0.287315494745)*x_ref[1]+(-0.627764670218)*x_ref[2]**o+(0.842131632598)*x_ref[2]
        ref[3,1,2,0]=(-0.957096489189)*x_ref[0]**o+(0.0616037805677)*x_ref[0]+(0.255858200166)*x_ref[1]**o+(-0.430555450798)*x_ref[1]+(0.730436930737)*x_ref[2]**o+(0.405094271653)*x_ref[2]
        ref[3,1,2,1]=(-0.924756832665)*x_ref[0]**o+(0.525754162391)*x_ref[0]+(0.281303228038)*x_ref[1]**o+(-0.479500161878)*x_ref[1]+(0.658416209181)*x_ref[2]**o+(-0.0413881595557)*x_ref[2]
        ref[3,2,0,0]=(0.516429381997)*x_ref[0]**o+(-0.8387318467)*x_ref[0]+(0.329773978227)*x_ref[1]**o+(0.599641753085)*x_ref[1]+(0.939937743271)*x_ref[2]**o+(-0.247331169626)*x_ref[2]
        ref[3,2,0,1]=(-0.644982044558)*x_ref[0]**o+(-0.242033352774)*x_ref[0]+(0.531835185779)*x_ref[1]**o+(-0.851565064183)*x_ref[1]+(-0.79213624915)*x_ref[2]**o+(0.0607365690732)*x_ref[2]
        ref[3,2,1,0]=(0.359781383537)*x_ref[0]**o+(-0.0529212842074)*x_ref[0]+(0.845915169452)*x_ref[1]**o+(-0.323979570458)*x_ref[1]+(0.828909997631)*x_ref[2]**o+(-0.839886339138)*x_ref[2]
        ref[3,2,1,1]=(0.559476630421)*x_ref[0]**o+(-0.189725214615)*x_ref[0]+(-0.00927446700413)*x_ref[1]**o+(-0.234703720782)*x_ref[1]+(-0.608362806755)*x_ref[2]**o+(0.968571765534)*x_ref[2]
        ref[3,2,2,0]=(0.924837507913)*x_ref[0]**o+(0.846397611121)*x_ref[0]+(-0.837765472403)*x_ref[1]**o+(0.0225105985057)*x_ref[1]+(-0.834187513844)*x_ref[2]**o+(0.194311845578)*x_ref[2]
        ref[3,2,2,1]=(-0.183561089961)*x_ref[0]**o+(0.0745071729471)*x_ref[0]+(-0.660415615202)*x_ref[1]**o+(-0.528184997345)*x_ref[1]+(0.205554094557)*x_ref[2]**o+(-0.0309246318944)*x_ref[2]
        ref[3,3,0,0]=(-0.67124095974)*x_ref[0]**o+(-0.48501497083)*x_ref[0]+(-0.0394345869094)*x_ref[1]**o+(0.586692270429)*x_ref[1]+(-0.593760749687)*x_ref[2]**o+(-0.124979751327)*x_ref[2]
        ref[3,3,0,1]=(0.778544173855)*x_ref[0]**o+(-0.931466478494)*x_ref[0]+(-0.746894777998)*x_ref[1]**o+(0.806417936297)*x_ref[1]+(0.989460929848)*x_ref[2]**o+(0.666534366369)*x_ref[2]
        ref[3,3,1,0]=(-0.798870838602)*x_ref[0]**o+(-0.294618868937)*x_ref[0]+(0.275029092968)*x_ref[1]**o+(0.45390580364)*x_ref[1]+(-0.311797818508)*x_ref[2]**o+(0.751606624393)*x_ref[2]
        ref[3,3,1,1]=(-0.924260987477)*x_ref[0]**o+(0.771970726902)*x_ref[0]+(-0.909099224213)*x_ref[1]**o+(-0.45204206766)*x_ref[1]+(0.630042614017)*x_ref[2]**o+(-0.990915130464)*x_ref[2]
        ref[3,3,2,0]=(0.734962154852)*x_ref[0]**o+(-0.00621898011475)*x_ref[0]+(-0.786362527189)*x_ref[1]**o+(0.623507947406)*x_ref[1]+(0.867175567104)*x_ref[2]**o+(-0.383283255448)*x_ref[2]
        ref[3,3,2,1]=(-0.955613156717)*x_ref[0]**o+(0.688013754841)*x_ref[0]+(-0.600559375208)*x_ref[1]**o+(0.453342626645)*x_ref[1]+(-0.864890477961)*x_ref[2]**o+(-0.458029873063)*x_ref[2]
        ref[3,4,0,0]=(-0.897010884834)*x_ref[0]**o+(-0.251223576458)*x_ref[0]+(-0.0338052047964)*x_ref[1]**o+(-0.903421466132)*x_ref[1]+(0.836864217874)*x_ref[2]**o+(0.660805905175)*x_ref[2]
        ref[3,4,0,1]=(-0.286025690215)*x_ref[0]**o+(0.746732975604)*x_ref[0]+(0.375810853218)*x_ref[1]**o+(-0.985192974869)*x_ref[1]+(-0.731463214455)*x_ref[2]**o+(-0.465450352444)*x_ref[2]
        ref[3,4,1,0]=(-0.500910000652)*x_ref[0]**o+(0.387502419727)*x_ref[0]+(0.962369164802)*x_ref[1]**o+(-0.10933824251)*x_ref[1]+(-0.33642703989)*x_ref[2]**o+(0.546851133256)*x_ref[2]
        ref[3,4,1,1]=(-0.117959459645)*x_ref[0]**o+(0.867699323392)*x_ref[0]+(0.950506403861)*x_ref[1]**o+(0.613806207745)*x_ref[1]+(0.425723173807)*x_ref[2]**o+(-0.906937548619)*x_ref[2]
        ref[3,4,2,0]=(0.331114892194)*x_ref[0]**o+(-0.168829643424)*x_ref[0]+(-0.855962311982)*x_ref[1]**o+(-0.144202406673)*x_ref[1]+(0.0769985713088)*x_ref[2]**o+(0.856003746423)*x_ref[2]
        ref[3,4,2,1]=(-0.439397439364)*x_ref[0]**o+(0.686116343481)*x_ref[0]+(0.519883854523)*x_ref[1]**o+(0.716231469289)*x_ref[1]+(0.270098231149)*x_ref[2]**o+(-0.448525436427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.203238386406)*x[0]**o+(0.734542238717)*x[0]+(-0.627248060056)*x[1]**o+(0.347224669854)*x[1]
        ref=(0.203238386406)*x_ref[0]**o+(0.734542238717)*x_ref[0]+(-0.627248060056)*x_ref[1]**o+(0.347224669854)*x_ref[1]
      else:
        arg=(-0.387501489724)*x[0]**o+(0.752554332333)*x[0]+(-0.0831588271355)*x[1]**o+(0.615191870812)*x[1]+(-0.264702923622)*x[2]**o+(0.404594629993)*x[2]
        ref=(-0.387501489724)*x_ref[0]**o+(0.752554332333)*x_ref[0]+(-0.0831588271355)*x_ref[1]**o+(0.615191870812)*x_ref[1]+(-0.264702923622)*x_ref[2]**o+(0.404594629993)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.938444792136)*x[0]**o+(-0.809407040724)*x[0]+(-0.963148615762)*x[1]**o+(-0.257000244129)*x[1]
        arg[1]=(-0.791280640039)*x[0]**o+(0.293858623095)*x[0]+(0.964942589784)*x[1]**o+(0.522071353081)*x[1]
        ref[0]=(0.938444792136)*x_ref[0]**o+(-0.809407040724)*x_ref[0]+(-0.963148615762)*x_ref[1]**o+(-0.257000244129)*x_ref[1]
        ref[1]=(-0.791280640039)*x_ref[0]**o+(0.293858623095)*x_ref[0]+(0.964942589784)*x_ref[1]**o+(0.522071353081)*x_ref[1]
      else:
        arg[0]=(0.566110580666)*x[0]**o+(-0.89723913213)*x[0]+(-0.54781629706)*x[1]**o+(-0.0405507445628)*x[1]+(0.435477036501)*x[2]**o+(-0.296404435183)*x[2]
        arg[1]=(-0.436126359475)*x[0]**o+(0.863706244241)*x[0]+(-0.149818409243)*x[1]**o+(-0.654801920098)*x[1]+(-0.281005092809)*x[2]**o+(-0.0257179420057)*x[2]
        ref[0]=(0.566110580666)*x_ref[0]**o+(-0.89723913213)*x_ref[0]+(-0.54781629706)*x_ref[1]**o+(-0.0405507445628)*x_ref[1]+(0.435477036501)*x_ref[2]**o+(-0.296404435183)*x_ref[2]
        ref[1]=(-0.436126359475)*x_ref[0]**o+(0.863706244241)*x_ref[0]+(-0.149818409243)*x_ref[1]**o+(-0.654801920098)*x_ref[1]+(-0.281005092809)*x_ref[2]**o+(-0.0257179420057)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0615713819144)*x[0]**o+(0.325133297443)*x[0]+(-0.529769594257)*x[1]**o+(-0.206126939263)*x[1]
        arg[0,1]=(0.090956391947)*x[0]**o+(0.105841035331)*x[0]+(-0.537033610998)*x[1]**o+(0.429033781495)*x[1]
        arg[0,2]=(-0.568216263139)*x[0]**o+(0.698057796806)*x[0]+(-0.0981614063428)*x[1]**o+(-0.952212332523)*x[1]
        arg[0,3]=(0.128505813477)*x[0]**o+(0.206828305852)*x[0]+(0.938851873532)*x[1]**o+(-0.339234589281)*x[1]
        arg[0,4]=(0.581805477855)*x[0]**o+(-0.621122347153)*x[0]+(-0.601029625565)*x[1]**o+(0.243084366725)*x[1]
        arg[1,0]=(-0.342411991491)*x[0]**o+(0.642434320731)*x[0]+(-0.452657094515)*x[1]**o+(-0.899779109444)*x[1]
        arg[1,1]=(-0.423230603077)*x[0]**o+(0.739600521332)*x[0]+(0.916226480419)*x[1]**o+(0.755264173162)*x[1]
        arg[1,2]=(0.381864221314)*x[0]**o+(0.806402187902)*x[0]+(0.410055116458)*x[1]**o+(-0.264405213472)*x[1]
        arg[1,3]=(0.293401427429)*x[0]**o+(-0.439209675154)*x[0]+(0.299588226145)*x[1]**o+(-0.289652201976)*x[1]
        arg[1,4]=(0.157626502865)*x[0]**o+(0.32264788209)*x[0]+(0.284522039757)*x[1]**o+(-0.460413671191)*x[1]
        arg[2,0]=(0.486148207)*x[0]**o+(0.0366950072658)*x[0]+(0.315299830901)*x[1]**o+(-0.320405042262)*x[1]
        arg[2,1]=(0.207440897542)*x[0]**o+(-0.604659890692)*x[0]+(-0.64841431916)*x[1]**o+(0.0881921209641)*x[1]
        arg[2,2]=(-0.61376840347)*x[0]**o+(0.421837679185)*x[0]+(-0.0445850531916)*x[1]**o+(0.771892018903)*x[1]
        arg[2,3]=(0.632443182263)*x[0]**o+(-0.965297013012)*x[0]+(-0.029677959681)*x[1]**o+(0.598431910551)*x[1]
        arg[2,4]=(-0.457656411094)*x[0]**o+(0.122435925199)*x[0]+(0.772690251937)*x[1]**o+(0.954829922926)*x[1]
        arg[3,0]=(0.929112925461)*x[0]**o+(-0.35689901655)*x[0]+(-0.139192866782)*x[1]**o+(-0.677883436952)*x[1]
        arg[3,1]=(-0.410318721103)*x[0]**o+(0.370830312699)*x[0]+(0.429011619434)*x[1]**o+(0.81780420854)*x[1]
        arg[3,2]=(0.312456682985)*x[0]**o+(0.526447664999)*x[0]+(0.763610570932)*x[1]**o+(-0.631887901017)*x[1]
        arg[3,3]=(0.749121567955)*x[0]**o+(0.168525340464)*x[0]+(-0.838409619895)*x[1]**o+(0.912745137098)*x[1]
        arg[3,4]=(-0.665336604643)*x[0]**o+(0.638293226479)*x[0]+(0.515957084712)*x[1]**o+(-0.261426720977)*x[1]
        ref[0,0]=(0.0615713819144)*x_ref[0]**o+(0.325133297443)*x_ref[0]+(-0.529769594257)*x_ref[1]**o+(-0.206126939263)*x_ref[1]
        ref[0,1]=(0.090956391947)*x_ref[0]**o+(0.105841035331)*x_ref[0]+(-0.537033610998)*x_ref[1]**o+(0.429033781495)*x_ref[1]
        ref[0,2]=(-0.568216263139)*x_ref[0]**o+(0.698057796806)*x_ref[0]+(-0.0981614063428)*x_ref[1]**o+(-0.952212332523)*x_ref[1]
        ref[0,3]=(0.128505813477)*x_ref[0]**o+(0.206828305852)*x_ref[0]+(0.938851873532)*x_ref[1]**o+(-0.339234589281)*x_ref[1]
        ref[0,4]=(0.581805477855)*x_ref[0]**o+(-0.621122347153)*x_ref[0]+(-0.601029625565)*x_ref[1]**o+(0.243084366725)*x_ref[1]
        ref[1,0]=(-0.342411991491)*x_ref[0]**o+(0.642434320731)*x_ref[0]+(-0.452657094515)*x_ref[1]**o+(-0.899779109444)*x_ref[1]
        ref[1,1]=(-0.423230603077)*x_ref[0]**o+(0.739600521332)*x_ref[0]+(0.916226480419)*x_ref[1]**o+(0.755264173162)*x_ref[1]
        ref[1,2]=(0.381864221314)*x_ref[0]**o+(0.806402187902)*x_ref[0]+(0.410055116458)*x_ref[1]**o+(-0.264405213472)*x_ref[1]
        ref[1,3]=(0.293401427429)*x_ref[0]**o+(-0.439209675154)*x_ref[0]+(0.299588226145)*x_ref[1]**o+(-0.289652201976)*x_ref[1]
        ref[1,4]=(0.157626502865)*x_ref[0]**o+(0.32264788209)*x_ref[0]+(0.284522039757)*x_ref[1]**o+(-0.460413671191)*x_ref[1]
        ref[2,0]=(0.486148207)*x_ref[0]**o+(0.0366950072658)*x_ref[0]+(0.315299830901)*x_ref[1]**o+(-0.320405042262)*x_ref[1]
        ref[2,1]=(0.207440897542)*x_ref[0]**o+(-0.604659890692)*x_ref[0]+(-0.64841431916)*x_ref[1]**o+(0.0881921209641)*x_ref[1]
        ref[2,2]=(-0.61376840347)*x_ref[0]**o+(0.421837679185)*x_ref[0]+(-0.0445850531916)*x_ref[1]**o+(0.771892018903)*x_ref[1]
        ref[2,3]=(0.632443182263)*x_ref[0]**o+(-0.965297013012)*x_ref[0]+(-0.029677959681)*x_ref[1]**o+(0.598431910551)*x_ref[1]
        ref[2,4]=(-0.457656411094)*x_ref[0]**o+(0.122435925199)*x_ref[0]+(0.772690251937)*x_ref[1]**o+(0.954829922926)*x_ref[1]
        ref[3,0]=(0.929112925461)*x_ref[0]**o+(-0.35689901655)*x_ref[0]+(-0.139192866782)*x_ref[1]**o+(-0.677883436952)*x_ref[1]
        ref[3,1]=(-0.410318721103)*x_ref[0]**o+(0.370830312699)*x_ref[0]+(0.429011619434)*x_ref[1]**o+(0.81780420854)*x_ref[1]
        ref[3,2]=(0.312456682985)*x_ref[0]**o+(0.526447664999)*x_ref[0]+(0.763610570932)*x_ref[1]**o+(-0.631887901017)*x_ref[1]
        ref[3,3]=(0.749121567955)*x_ref[0]**o+(0.168525340464)*x_ref[0]+(-0.838409619895)*x_ref[1]**o+(0.912745137098)*x_ref[1]
        ref[3,4]=(-0.665336604643)*x_ref[0]**o+(0.638293226479)*x_ref[0]+(0.515957084712)*x_ref[1]**o+(-0.261426720977)*x_ref[1]
      else:
        arg[0,0]=(-0.330165086386)*x[0]**o+(0.60381119065)*x[0]+(0.981018763011)*x[1]**o+(0.0487965999608)*x[1]+(-0.389835388349)*x[2]**o+(0.111199857376)*x[2]
        arg[0,1]=(0.518239033178)*x[0]**o+(0.60358784684)*x[0]+(-0.476990159655)*x[1]**o+(-0.93871565843)*x[1]+(-0.869845295741)*x[2]**o+(0.391409745802)*x[2]
        arg[0,2]=(0.0927396115951)*x[0]**o+(0.546122822928)*x[0]+(-0.557424056733)*x[1]**o+(0.0623374966436)*x[1]+(-0.414595153536)*x[2]**o+(-0.196069073366)*x[2]
        arg[0,3]=(-0.678432580882)*x[0]**o+(0.281850265141)*x[0]+(-0.323978960602)*x[1]**o+(0.655558769921)*x[1]+(-0.295747032089)*x[2]**o+(-0.152719838982)*x[2]
        arg[0,4]=(-0.904177669203)*x[0]**o+(0.750395234205)*x[0]+(0.267431783461)*x[1]**o+(-0.781406235306)*x[1]+(0.856906655757)*x[2]**o+(-0.875554940224)*x[2]
        arg[1,0]=(-0.243144846569)*x[0]**o+(0.970584922467)*x[0]+(-0.977250652473)*x[1]**o+(0.0239563255869)*x[1]+(-0.769423602851)*x[2]**o+(-0.433896361294)*x[2]
        arg[1,1]=(0.31104031046)*x[0]**o+(0.899859709831)*x[0]+(0.692351561919)*x[1]**o+(-0.0838870550002)*x[1]+(-0.276105616798)*x[2]**o+(0.115501348248)*x[2]
        arg[1,2]=(0.56702114595)*x[0]**o+(0.00240284646131)*x[0]+(-0.676314888884)*x[1]**o+(-0.678204346328)*x[1]+(0.472697088984)*x[2]**o+(-0.493949191712)*x[2]
        arg[1,3]=(-0.242548554067)*x[0]**o+(0.980927507053)*x[0]+(-0.955706390436)*x[1]**o+(0.347031550126)*x[1]+(0.234352268238)*x[2]**o+(0.321126345333)*x[2]
        arg[1,4]=(-0.390890566228)*x[0]**o+(-0.958497753144)*x[0]+(0.316397716043)*x[1]**o+(0.952289821769)*x[1]+(-0.027182029514)*x[2]**o+(-0.454422377872)*x[2]
        arg[2,0]=(0.295865166318)*x[0]**o+(-0.301129718468)*x[0]+(-0.748135755269)*x[1]**o+(-0.762413814085)*x[1]+(0.572270067591)*x[2]**o+(-0.393847032684)*x[2]
        arg[2,1]=(0.20229063622)*x[0]**o+(-0.768444028163)*x[0]+(0.378669775381)*x[1]**o+(-0.703384410944)*x[1]+(0.355653859396)*x[2]**o+(-0.228168949639)*x[2]
        arg[2,2]=(-0.31725134742)*x[0]**o+(0.525802867508)*x[0]+(0.13375126758)*x[1]**o+(-0.673551491186)*x[1]+(0.498905883728)*x[2]**o+(-0.102794461221)*x[2]
        arg[2,3]=(-0.396813838168)*x[0]**o+(0.0892605807843)*x[0]+(0.244133298111)*x[1]**o+(0.310900636168)*x[1]+(-0.256807063422)*x[2]**o+(0.565370697688)*x[2]
        arg[2,4]=(0.42011835881)*x[0]**o+(0.439847594375)*x[0]+(0.0731897272006)*x[1]**o+(-0.373911054753)*x[1]+(-0.50313315963)*x[2]**o+(0.356362677601)*x[2]
        arg[3,0]=(-0.444672845607)*x[0]**o+(-0.978424347535)*x[0]+(0.421699748442)*x[1]**o+(-0.0111798084161)*x[1]+(0.206042685881)*x[2]**o+(0.990523831215)*x[2]
        arg[3,1]=(0.942429627068)*x[0]**o+(-0.769989484282)*x[0]+(-0.879916738075)*x[1]**o+(-0.746448112871)*x[1]+(0.280824650861)*x[2]**o+(0.0632084823022)*x[2]
        arg[3,2]=(0.281070481177)*x[0]**o+(0.628020001876)*x[0]+(0.347767303962)*x[1]**o+(-0.276761252142)*x[1]+(0.950345260734)*x[2]**o+(-0.00148186335195)*x[2]
        arg[3,3]=(-0.37590210465)*x[0]**o+(0.034147350068)*x[0]+(0.245173418336)*x[1]**o+(0.0564051241759)*x[1]+(-0.157928106127)*x[2]**o+(-0.611677441847)*x[2]
        arg[3,4]=(-0.279066532276)*x[0]**o+(0.316033119063)*x[0]+(0.655948999472)*x[1]**o+(-0.520504180943)*x[1]+(0.531738158819)*x[2]**o+(0.899490827443)*x[2]
        ref[0,0]=(-0.330165086386)*x_ref[0]**o+(0.60381119065)*x_ref[0]+(0.981018763011)*x_ref[1]**o+(0.0487965999608)*x_ref[1]+(-0.389835388349)*x_ref[2]**o+(0.111199857376)*x_ref[2]
        ref[0,1]=(0.518239033178)*x_ref[0]**o+(0.60358784684)*x_ref[0]+(-0.476990159655)*x_ref[1]**o+(-0.93871565843)*x_ref[1]+(-0.869845295741)*x_ref[2]**o+(0.391409745802)*x_ref[2]
        ref[0,2]=(0.0927396115951)*x_ref[0]**o+(0.546122822928)*x_ref[0]+(-0.557424056733)*x_ref[1]**o+(0.0623374966436)*x_ref[1]+(-0.414595153536)*x_ref[2]**o+(-0.196069073366)*x_ref[2]
        ref[0,3]=(-0.678432580882)*x_ref[0]**o+(0.281850265141)*x_ref[0]+(-0.323978960602)*x_ref[1]**o+(0.655558769921)*x_ref[1]+(-0.295747032089)*x_ref[2]**o+(-0.152719838982)*x_ref[2]
        ref[0,4]=(-0.904177669203)*x_ref[0]**o+(0.750395234205)*x_ref[0]+(0.267431783461)*x_ref[1]**o+(-0.781406235306)*x_ref[1]+(0.856906655757)*x_ref[2]**o+(-0.875554940224)*x_ref[2]
        ref[1,0]=(-0.243144846569)*x_ref[0]**o+(0.970584922467)*x_ref[0]+(-0.977250652473)*x_ref[1]**o+(0.0239563255869)*x_ref[1]+(-0.769423602851)*x_ref[2]**o+(-0.433896361294)*x_ref[2]
        ref[1,1]=(0.31104031046)*x_ref[0]**o+(0.899859709831)*x_ref[0]+(0.692351561919)*x_ref[1]**o+(-0.0838870550002)*x_ref[1]+(-0.276105616798)*x_ref[2]**o+(0.115501348248)*x_ref[2]
        ref[1,2]=(0.56702114595)*x_ref[0]**o+(0.00240284646131)*x_ref[0]+(-0.676314888884)*x_ref[1]**o+(-0.678204346328)*x_ref[1]+(0.472697088984)*x_ref[2]**o+(-0.493949191712)*x_ref[2]
        ref[1,3]=(-0.242548554067)*x_ref[0]**o+(0.980927507053)*x_ref[0]+(-0.955706390436)*x_ref[1]**o+(0.347031550126)*x_ref[1]+(0.234352268238)*x_ref[2]**o+(0.321126345333)*x_ref[2]
        ref[1,4]=(-0.390890566228)*x_ref[0]**o+(-0.958497753144)*x_ref[0]+(0.316397716043)*x_ref[1]**o+(0.952289821769)*x_ref[1]+(-0.027182029514)*x_ref[2]**o+(-0.454422377872)*x_ref[2]
        ref[2,0]=(0.295865166318)*x_ref[0]**o+(-0.301129718468)*x_ref[0]+(-0.748135755269)*x_ref[1]**o+(-0.762413814085)*x_ref[1]+(0.572270067591)*x_ref[2]**o+(-0.393847032684)*x_ref[2]
        ref[2,1]=(0.20229063622)*x_ref[0]**o+(-0.768444028163)*x_ref[0]+(0.378669775381)*x_ref[1]**o+(-0.703384410944)*x_ref[1]+(0.355653859396)*x_ref[2]**o+(-0.228168949639)*x_ref[2]
        ref[2,2]=(-0.31725134742)*x_ref[0]**o+(0.525802867508)*x_ref[0]+(0.13375126758)*x_ref[1]**o+(-0.673551491186)*x_ref[1]+(0.498905883728)*x_ref[2]**o+(-0.102794461221)*x_ref[2]
        ref[2,3]=(-0.396813838168)*x_ref[0]**o+(0.0892605807843)*x_ref[0]+(0.244133298111)*x_ref[1]**o+(0.310900636168)*x_ref[1]+(-0.256807063422)*x_ref[2]**o+(0.565370697688)*x_ref[2]
        ref[2,4]=(0.42011835881)*x_ref[0]**o+(0.439847594375)*x_ref[0]+(0.0731897272006)*x_ref[1]**o+(-0.373911054753)*x_ref[1]+(-0.50313315963)*x_ref[2]**o+(0.356362677601)*x_ref[2]
        ref[3,0]=(-0.444672845607)*x_ref[0]**o+(-0.978424347535)*x_ref[0]+(0.421699748442)*x_ref[1]**o+(-0.0111798084161)*x_ref[1]+(0.206042685881)*x_ref[2]**o+(0.990523831215)*x_ref[2]
        ref[3,1]=(0.942429627068)*x_ref[0]**o+(-0.769989484282)*x_ref[0]+(-0.879916738075)*x_ref[1]**o+(-0.746448112871)*x_ref[1]+(0.280824650861)*x_ref[2]**o+(0.0632084823022)*x_ref[2]
        ref[3,2]=(0.281070481177)*x_ref[0]**o+(0.628020001876)*x_ref[0]+(0.347767303962)*x_ref[1]**o+(-0.276761252142)*x_ref[1]+(0.950345260734)*x_ref[2]**o+(-0.00148186335195)*x_ref[2]
        ref[3,3]=(-0.37590210465)*x_ref[0]**o+(0.034147350068)*x_ref[0]+(0.245173418336)*x_ref[1]**o+(0.0564051241759)*x_ref[1]+(-0.157928106127)*x_ref[2]**o+(-0.611677441847)*x_ref[2]
        ref[3,4]=(-0.279066532276)*x_ref[0]**o+(0.316033119063)*x_ref[0]+(0.655948999472)*x_ref[1]**o+(-0.520504180943)*x_ref[1]+(0.531738158819)*x_ref[2]**o+(0.899490827443)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.415438516191)*x[0]**o+(0.374659323173)*x[0]+(0.688470746611)*x[1]**o+(0.498183692176)*x[1]
        arg[0,0,1]=(-0.846725329522)*x[0]**o+(0.414376816326)*x[0]+(0.651609661615)*x[1]**o+(0.357606396667)*x[1]
        arg[0,1,0]=(0.728029172023)*x[0]**o+(0.975728459112)*x[0]+(-0.658211371639)*x[1]**o+(0.411894140351)*x[1]
        arg[0,1,1]=(0.807350281817)*x[0]**o+(-0.369755733522)*x[0]+(-0.519483139314)*x[1]**o+(-0.709353537648)*x[1]
        arg[1,0,0]=(0.0588259329479)*x[0]**o+(0.218362870441)*x[0]+(-0.0279239318934)*x[1]**o+(-0.75783635329)*x[1]
        arg[1,0,1]=(-0.429573558505)*x[0]**o+(-0.339046134618)*x[0]+(0.412777825743)*x[1]**o+(-0.306607333437)*x[1]
        arg[1,1,0]=(0.0222661407543)*x[0]**o+(0.927923674781)*x[0]+(-0.767468570033)*x[1]**o+(-0.0659150590367)*x[1]
        arg[1,1,1]=(0.372811733787)*x[0]**o+(-0.286500260061)*x[0]+(-0.807504375569)*x[1]**o+(-0.992850148234)*x[1]
        arg[2,0,0]=(0.368229828596)*x[0]**o+(0.672540290379)*x[0]+(-0.760462595095)*x[1]**o+(-0.761242492425)*x[1]
        arg[2,0,1]=(0.306249523133)*x[0]**o+(0.769928924717)*x[0]+(0.610293800742)*x[1]**o+(-0.221996271153)*x[1]
        arg[2,1,0]=(-0.909148832991)*x[0]**o+(-0.0293038712703)*x[0]+(-0.161707473269)*x[1]**o+(-0.534345319122)*x[1]
        arg[2,1,1]=(-0.146432487737)*x[0]**o+(0.93729266958)*x[0]+(-0.561919238789)*x[1]**o+(-0.449440501171)*x[1]
        arg[3,0,0]=(0.411711661035)*x[0]**o+(-0.0699878047274)*x[0]+(-0.0227688826645)*x[1]**o+(-0.939787633547)*x[1]
        arg[3,0,1]=(-0.82108820245)*x[0]**o+(0.225501168847)*x[0]+(0.622921674682)*x[1]**o+(0.61092413094)*x[1]
        arg[3,1,0]=(0.758116155912)*x[0]**o+(-0.117316615447)*x[0]+(0.0702245119124)*x[1]**o+(-0.275114508561)*x[1]
        arg[3,1,1]=(-0.32568550821)*x[0]**o+(-0.280144846326)*x[0]+(0.194971187876)*x[1]**o+(-0.728447244746)*x[1]
        arg[4,0,0]=(0.880042622981)*x[0]**o+(-0.395684607563)*x[0]+(-0.563304061265)*x[1]**o+(0.968663606772)*x[1]
        arg[4,0,1]=(-0.515392712458)*x[0]**o+(0.820857647342)*x[0]+(-0.476289639495)*x[1]**o+(-0.478062589609)*x[1]
        arg[4,1,0]=(0.534322425269)*x[0]**o+(0.975618695265)*x[0]+(-0.195764005045)*x[1]**o+(0.561996200643)*x[1]
        arg[4,1,1]=(-0.906243138672)*x[0]**o+(-0.580985047294)*x[0]+(-0.094609405051)*x[1]**o+(0.734326711649)*x[1]
        arg[5,0,0]=(-0.0655516121632)*x[0]**o+(0.785493123174)*x[0]+(-0.760068333675)*x[1]**o+(0.693869247554)*x[1]
        arg[5,0,1]=(-0.931019838892)*x[0]**o+(0.80958291621)*x[0]+(-0.590591206111)*x[1]**o+(-0.800900346601)*x[1]
        arg[5,1,0]=(0.448779048952)*x[0]**o+(-0.793450503268)*x[0]+(0.161067403301)*x[1]**o+(0.432054043544)*x[1]
        arg[5,1,1]=(0.0423607361972)*x[0]**o+(-0.664958045702)*x[0]+(-0.174728816918)*x[1]**o+(-0.394390068699)*x[1]
        ref[0,0,0]=(0.415438516191)*x_ref[0]**o+(0.374659323173)*x_ref[0]+(0.688470746611)*x_ref[1]**o+(0.498183692176)*x_ref[1]
        ref[0,0,1]=(-0.846725329522)*x_ref[0]**o+(0.414376816326)*x_ref[0]+(0.651609661615)*x_ref[1]**o+(0.357606396667)*x_ref[1]
        ref[0,1,0]=(0.728029172023)*x_ref[0]**o+(0.975728459112)*x_ref[0]+(-0.658211371639)*x_ref[1]**o+(0.411894140351)*x_ref[1]
        ref[0,1,1]=(0.807350281817)*x_ref[0]**o+(-0.369755733522)*x_ref[0]+(-0.519483139314)*x_ref[1]**o+(-0.709353537648)*x_ref[1]
        ref[1,0,0]=(0.0588259329479)*x_ref[0]**o+(0.218362870441)*x_ref[0]+(-0.0279239318934)*x_ref[1]**o+(-0.75783635329)*x_ref[1]
        ref[1,0,1]=(-0.429573558505)*x_ref[0]**o+(-0.339046134618)*x_ref[0]+(0.412777825743)*x_ref[1]**o+(-0.306607333437)*x_ref[1]
        ref[1,1,0]=(0.0222661407543)*x_ref[0]**o+(0.927923674781)*x_ref[0]+(-0.767468570033)*x_ref[1]**o+(-0.0659150590367)*x_ref[1]
        ref[1,1,1]=(0.372811733787)*x_ref[0]**o+(-0.286500260061)*x_ref[0]+(-0.807504375569)*x_ref[1]**o+(-0.992850148234)*x_ref[1]
        ref[2,0,0]=(0.368229828596)*x_ref[0]**o+(0.672540290379)*x_ref[0]+(-0.760462595095)*x_ref[1]**o+(-0.761242492425)*x_ref[1]
        ref[2,0,1]=(0.306249523133)*x_ref[0]**o+(0.769928924717)*x_ref[0]+(0.610293800742)*x_ref[1]**o+(-0.221996271153)*x_ref[1]
        ref[2,1,0]=(-0.909148832991)*x_ref[0]**o+(-0.0293038712703)*x_ref[0]+(-0.161707473269)*x_ref[1]**o+(-0.534345319122)*x_ref[1]
        ref[2,1,1]=(-0.146432487737)*x_ref[0]**o+(0.93729266958)*x_ref[0]+(-0.561919238789)*x_ref[1]**o+(-0.449440501171)*x_ref[1]
        ref[3,0,0]=(0.411711661035)*x_ref[0]**o+(-0.0699878047274)*x_ref[0]+(-0.0227688826645)*x_ref[1]**o+(-0.939787633547)*x_ref[1]
        ref[3,0,1]=(-0.82108820245)*x_ref[0]**o+(0.225501168847)*x_ref[0]+(0.622921674682)*x_ref[1]**o+(0.61092413094)*x_ref[1]
        ref[3,1,0]=(0.758116155912)*x_ref[0]**o+(-0.117316615447)*x_ref[0]+(0.0702245119124)*x_ref[1]**o+(-0.275114508561)*x_ref[1]
        ref[3,1,1]=(-0.32568550821)*x_ref[0]**o+(-0.280144846326)*x_ref[0]+(0.194971187876)*x_ref[1]**o+(-0.728447244746)*x_ref[1]
        ref[4,0,0]=(0.880042622981)*x_ref[0]**o+(-0.395684607563)*x_ref[0]+(-0.563304061265)*x_ref[1]**o+(0.968663606772)*x_ref[1]
        ref[4,0,1]=(-0.515392712458)*x_ref[0]**o+(0.820857647342)*x_ref[0]+(-0.476289639495)*x_ref[1]**o+(-0.478062589609)*x_ref[1]
        ref[4,1,0]=(0.534322425269)*x_ref[0]**o+(0.975618695265)*x_ref[0]+(-0.195764005045)*x_ref[1]**o+(0.561996200643)*x_ref[1]
        ref[4,1,1]=(-0.906243138672)*x_ref[0]**o+(-0.580985047294)*x_ref[0]+(-0.094609405051)*x_ref[1]**o+(0.734326711649)*x_ref[1]
        ref[5,0,0]=(-0.0655516121632)*x_ref[0]**o+(0.785493123174)*x_ref[0]+(-0.760068333675)*x_ref[1]**o+(0.693869247554)*x_ref[1]
        ref[5,0,1]=(-0.931019838892)*x_ref[0]**o+(0.80958291621)*x_ref[0]+(-0.590591206111)*x_ref[1]**o+(-0.800900346601)*x_ref[1]
        ref[5,1,0]=(0.448779048952)*x_ref[0]**o+(-0.793450503268)*x_ref[0]+(0.161067403301)*x_ref[1]**o+(0.432054043544)*x_ref[1]
        ref[5,1,1]=(0.0423607361972)*x_ref[0]**o+(-0.664958045702)*x_ref[0]+(-0.174728816918)*x_ref[1]**o+(-0.394390068699)*x_ref[1]
      else:
        arg[0,0,0]=(0.34355250678)*x[0]**o+(0.554590991175)*x[0]+(0.00673862595967)*x[1]**o+(-0.405764224977)*x[1]+(0.0865564883661)*x[2]**o+(0.102294137607)*x[2]
        arg[0,0,1]=(0.855234460708)*x[0]**o+(-0.710901486015)*x[0]+(0.546393138087)*x[1]**o+(0.847113479239)*x[1]+(-0.650701604407)*x[2]**o+(-0.914755510262)*x[2]
        arg[0,1,0]=(0.252544568503)*x[0]**o+(-0.473471976093)*x[0]+(-0.587315274764)*x[1]**o+(-0.269940894375)*x[1]+(0.360340735507)*x[2]**o+(-0.745440004827)*x[2]
        arg[0,1,1]=(0.11967506555)*x[0]**o+(0.696836453942)*x[0]+(-0.871162187396)*x[1]**o+(-0.315170111067)*x[1]+(0.232066109266)*x[2]**o+(-0.464730524838)*x[2]
        arg[1,0,0]=(-0.790535662997)*x[0]**o+(-0.964779159884)*x[0]+(-0.689937881374)*x[1]**o+(0.422799231262)*x[1]+(-0.97315874804)*x[2]**o+(-0.532103115205)*x[2]
        arg[1,0,1]=(-0.0487589554945)*x[0]**o+(-0.622735713658)*x[0]+(-0.520842613265)*x[1]**o+(0.803711422599)*x[1]+(0.440648875556)*x[2]**o+(0.98786672574)*x[2]
        arg[1,1,0]=(0.811189977538)*x[0]**o+(0.0175150486788)*x[0]+(-0.124328899976)*x[1]**o+(-0.942398680503)*x[1]+(-0.149270209895)*x[2]**o+(-0.853084407696)*x[2]
        arg[1,1,1]=(0.688378159744)*x[0]**o+(0.408766927878)*x[0]+(0.834128919609)*x[1]**o+(0.253605923886)*x[1]+(0.424992121054)*x[2]**o+(0.21016117963)*x[2]
        arg[2,0,0]=(0.277469067087)*x[0]**o+(0.539821557611)*x[0]+(0.348679402521)*x[1]**o+(-0.469772506537)*x[1]+(0.454583229137)*x[2]**o+(0.653681481133)*x[2]
        arg[2,0,1]=(-0.934067178483)*x[0]**o+(-0.172735098497)*x[0]+(-0.0347256850948)*x[1]**o+(-0.459971424944)*x[1]+(0.932661853019)*x[2]**o+(-0.186822590955)*x[2]
        arg[2,1,0]=(0.71698574321)*x[0]**o+(-0.197244831261)*x[0]+(0.183419576182)*x[1]**o+(-0.871451480853)*x[1]+(-0.349396269955)*x[2]**o+(0.095909246936)*x[2]
        arg[2,1,1]=(-0.00972985962106)*x[0]**o+(-0.919204152811)*x[0]+(-0.751744404365)*x[1]**o+(0.377096436852)*x[1]+(-0.924228885362)*x[2]**o+(0.451053813959)*x[2]
        arg[3,0,0]=(0.951119968373)*x[0]**o+(-0.0680027504528)*x[0]+(-0.113660990287)*x[1]**o+(-0.428999453003)*x[1]+(0.171804544103)*x[2]**o+(0.677635036297)*x[2]
        arg[3,0,1]=(0.62534864581)*x[0]**o+(0.389496246771)*x[0]+(0.621295327948)*x[1]**o+(-0.684165739023)*x[1]+(0.258637161492)*x[2]**o+(0.0441460764076)*x[2]
        arg[3,1,0]=(0.0348922721145)*x[0]**o+(0.091201286767)*x[0]+(0.421079169132)*x[1]**o+(-0.616504862591)*x[1]+(0.563927673176)*x[2]**o+(-0.183993394822)*x[2]
        arg[3,1,1]=(0.657959811783)*x[0]**o+(0.229729671382)*x[0]+(0.549853155207)*x[1]**o+(-0.847412930565)*x[1]+(-0.0780682500337)*x[2]**o+(0.574435563406)*x[2]
        arg[4,0,0]=(-0.679988515344)*x[0]**o+(0.0517331630343)*x[0]+(-0.160281064879)*x[1]**o+(-0.0672405944261)*x[1]+(-0.118968444607)*x[2]**o+(0.00980065834148)*x[2]
        arg[4,0,1]=(-0.150457972951)*x[0]**o+(0.671586511406)*x[0]+(-0.844112675259)*x[1]**o+(0.42766255101)*x[1]+(-0.22449721018)*x[2]**o+(-0.150416151059)*x[2]
        arg[4,1,0]=(0.0768992597753)*x[0]**o+(0.0139605476692)*x[0]+(-0.831269554863)*x[1]**o+(-0.279339344511)*x[1]+(0.387318821185)*x[2]**o+(0.967687121678)*x[2]
        arg[4,1,1]=(-0.670585948598)*x[0]**o+(0.422715818654)*x[0]+(0.623702390878)*x[1]**o+(0.922646081646)*x[1]+(-0.714191287039)*x[2]**o+(-0.99018084869)*x[2]
        arg[5,0,0]=(0.100273819257)*x[0]**o+(-0.769422291863)*x[0]+(-0.152611272068)*x[1]**o+(-0.396798631305)*x[1]+(0.502251047462)*x[2]**o+(0.993791216449)*x[2]
        arg[5,0,1]=(-0.0222540900657)*x[0]**o+(-0.339810353987)*x[0]+(0.241665078696)*x[1]**o+(0.111020568564)*x[1]+(0.298147316931)*x[2]**o+(0.391278660594)*x[2]
        arg[5,1,0]=(0.828297599555)*x[0]**o+(0.237678875371)*x[0]+(0.924551126274)*x[1]**o+(-0.663695049485)*x[1]+(-0.547123493132)*x[2]**o+(-0.788635745118)*x[2]
        arg[5,1,1]=(0.0726334370982)*x[0]**o+(-0.875688492957)*x[0]+(0.473726235108)*x[1]**o+(0.023025596151)*x[1]+(0.827449396402)*x[2]**o+(-0.154698941604)*x[2]
        ref[0,0,0]=(0.34355250678)*x_ref[0]**o+(0.554590991175)*x_ref[0]+(0.00673862595967)*x_ref[1]**o+(-0.405764224977)*x_ref[1]+(0.0865564883661)*x_ref[2]**o+(0.102294137607)*x_ref[2]
        ref[0,0,1]=(0.855234460708)*x_ref[0]**o+(-0.710901486015)*x_ref[0]+(0.546393138087)*x_ref[1]**o+(0.847113479239)*x_ref[1]+(-0.650701604407)*x_ref[2]**o+(-0.914755510262)*x_ref[2]
        ref[0,1,0]=(0.252544568503)*x_ref[0]**o+(-0.473471976093)*x_ref[0]+(-0.587315274764)*x_ref[1]**o+(-0.269940894375)*x_ref[1]+(0.360340735507)*x_ref[2]**o+(-0.745440004827)*x_ref[2]
        ref[0,1,1]=(0.11967506555)*x_ref[0]**o+(0.696836453942)*x_ref[0]+(-0.871162187396)*x_ref[1]**o+(-0.315170111067)*x_ref[1]+(0.232066109266)*x_ref[2]**o+(-0.464730524838)*x_ref[2]
        ref[1,0,0]=(-0.790535662997)*x_ref[0]**o+(-0.964779159884)*x_ref[0]+(-0.689937881374)*x_ref[1]**o+(0.422799231262)*x_ref[1]+(-0.97315874804)*x_ref[2]**o+(-0.532103115205)*x_ref[2]
        ref[1,0,1]=(-0.0487589554945)*x_ref[0]**o+(-0.622735713658)*x_ref[0]+(-0.520842613265)*x_ref[1]**o+(0.803711422599)*x_ref[1]+(0.440648875556)*x_ref[2]**o+(0.98786672574)*x_ref[2]
        ref[1,1,0]=(0.811189977538)*x_ref[0]**o+(0.0175150486788)*x_ref[0]+(-0.124328899976)*x_ref[1]**o+(-0.942398680503)*x_ref[1]+(-0.149270209895)*x_ref[2]**o+(-0.853084407696)*x_ref[2]
        ref[1,1,1]=(0.688378159744)*x_ref[0]**o+(0.408766927878)*x_ref[0]+(0.834128919609)*x_ref[1]**o+(0.253605923886)*x_ref[1]+(0.424992121054)*x_ref[2]**o+(0.21016117963)*x_ref[2]
        ref[2,0,0]=(0.277469067087)*x_ref[0]**o+(0.539821557611)*x_ref[0]+(0.348679402521)*x_ref[1]**o+(-0.469772506537)*x_ref[1]+(0.454583229137)*x_ref[2]**o+(0.653681481133)*x_ref[2]
        ref[2,0,1]=(-0.934067178483)*x_ref[0]**o+(-0.172735098497)*x_ref[0]+(-0.0347256850948)*x_ref[1]**o+(-0.459971424944)*x_ref[1]+(0.932661853019)*x_ref[2]**o+(-0.186822590955)*x_ref[2]
        ref[2,1,0]=(0.71698574321)*x_ref[0]**o+(-0.197244831261)*x_ref[0]+(0.183419576182)*x_ref[1]**o+(-0.871451480853)*x_ref[1]+(-0.349396269955)*x_ref[2]**o+(0.095909246936)*x_ref[2]
        ref[2,1,1]=(-0.00972985962106)*x_ref[0]**o+(-0.919204152811)*x_ref[0]+(-0.751744404365)*x_ref[1]**o+(0.377096436852)*x_ref[1]+(-0.924228885362)*x_ref[2]**o+(0.451053813959)*x_ref[2]
        ref[3,0,0]=(0.951119968373)*x_ref[0]**o+(-0.0680027504528)*x_ref[0]+(-0.113660990287)*x_ref[1]**o+(-0.428999453003)*x_ref[1]+(0.171804544103)*x_ref[2]**o+(0.677635036297)*x_ref[2]
        ref[3,0,1]=(0.62534864581)*x_ref[0]**o+(0.389496246771)*x_ref[0]+(0.621295327948)*x_ref[1]**o+(-0.684165739023)*x_ref[1]+(0.258637161492)*x_ref[2]**o+(0.0441460764076)*x_ref[2]
        ref[3,1,0]=(0.0348922721145)*x_ref[0]**o+(0.091201286767)*x_ref[0]+(0.421079169132)*x_ref[1]**o+(-0.616504862591)*x_ref[1]+(0.563927673176)*x_ref[2]**o+(-0.183993394822)*x_ref[2]
        ref[3,1,1]=(0.657959811783)*x_ref[0]**o+(0.229729671382)*x_ref[0]+(0.549853155207)*x_ref[1]**o+(-0.847412930565)*x_ref[1]+(-0.0780682500337)*x_ref[2]**o+(0.574435563406)*x_ref[2]
        ref[4,0,0]=(-0.679988515344)*x_ref[0]**o+(0.0517331630343)*x_ref[0]+(-0.160281064879)*x_ref[1]**o+(-0.0672405944261)*x_ref[1]+(-0.118968444607)*x_ref[2]**o+(0.00980065834148)*x_ref[2]
        ref[4,0,1]=(-0.150457972951)*x_ref[0]**o+(0.671586511406)*x_ref[0]+(-0.844112675259)*x_ref[1]**o+(0.42766255101)*x_ref[1]+(-0.22449721018)*x_ref[2]**o+(-0.150416151059)*x_ref[2]
        ref[4,1,0]=(0.0768992597753)*x_ref[0]**o+(0.0139605476692)*x_ref[0]+(-0.831269554863)*x_ref[1]**o+(-0.279339344511)*x_ref[1]+(0.387318821185)*x_ref[2]**o+(0.967687121678)*x_ref[2]
        ref[4,1,1]=(-0.670585948598)*x_ref[0]**o+(0.422715818654)*x_ref[0]+(0.623702390878)*x_ref[1]**o+(0.922646081646)*x_ref[1]+(-0.714191287039)*x_ref[2]**o+(-0.99018084869)*x_ref[2]
        ref[5,0,0]=(0.100273819257)*x_ref[0]**o+(-0.769422291863)*x_ref[0]+(-0.152611272068)*x_ref[1]**o+(-0.396798631305)*x_ref[1]+(0.502251047462)*x_ref[2]**o+(0.993791216449)*x_ref[2]
        ref[5,0,1]=(-0.0222540900657)*x_ref[0]**o+(-0.339810353987)*x_ref[0]+(0.241665078696)*x_ref[1]**o+(0.111020568564)*x_ref[1]+(0.298147316931)*x_ref[2]**o+(0.391278660594)*x_ref[2]
        ref[5,1,0]=(0.828297599555)*x_ref[0]**o+(0.237678875371)*x_ref[0]+(0.924551126274)*x_ref[1]**o+(-0.663695049485)*x_ref[1]+(-0.547123493132)*x_ref[2]**o+(-0.788635745118)*x_ref[2]
        ref[5,1,1]=(0.0726334370982)*x_ref[0]**o+(-0.875688492957)*x_ref[0]+(0.473726235108)*x_ref[1]**o+(0.023025596151)*x_ref[1]+(0.827449396402)*x_ref[2]**o+(-0.154698941604)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0338299192888)*x[0]**o+(-0.114906274359)*x[0]+(0.15010339398)*x[1]**o+(0.625295991328)*x[1]
        arg[0,0,0,1]=(-0.1311901039)*x[0]**o+(-0.551712861217)*x[0]+(-0.287411404506)*x[1]**o+(0.0242820629572)*x[1]
        arg[0,0,1,0]=(0.121962094111)*x[0]**o+(0.203169622345)*x[0]+(0.241504218053)*x[1]**o+(-0.0696356239498)*x[1]
        arg[0,0,1,1]=(0.0461537352329)*x[0]**o+(-0.275908938567)*x[0]+(-0.271184121027)*x[1]**o+(-0.364024729567)*x[1]
        arg[0,0,2,0]=(0.199272234315)*x[0]**o+(-0.906364258855)*x[0]+(-0.92102930037)*x[1]**o+(0.0718981903454)*x[1]
        arg[0,0,2,1]=(-0.742950958383)*x[0]**o+(0.0301464847722)*x[0]+(0.466993477687)*x[1]**o+(-0.402131750698)*x[1]
        arg[0,1,0,0]=(-0.0653776546375)*x[0]**o+(-0.679360612372)*x[0]+(-0.32783837768)*x[1]**o+(-0.113753016428)*x[1]
        arg[0,1,0,1]=(0.58610119414)*x[0]**o+(0.142613166231)*x[0]+(0.0409105619986)*x[1]**o+(-0.660641663558)*x[1]
        arg[0,1,1,0]=(0.212172308585)*x[0]**o+(-0.452767768344)*x[0]+(0.963053969889)*x[1]**o+(0.752691773953)*x[1]
        arg[0,1,1,1]=(-0.514998377887)*x[0]**o+(0.0872464329801)*x[0]+(0.265590295505)*x[1]**o+(0.557238826549)*x[1]
        arg[0,1,2,0]=(0.360523954724)*x[0]**o+(0.550238402994)*x[0]+(-0.206349415539)*x[1]**o+(0.357304730511)*x[1]
        arg[0,1,2,1]=(-0.872755217435)*x[0]**o+(0.190943683664)*x[0]+(-0.245454292354)*x[1]**o+(0.507229578506)*x[1]
        arg[0,2,0,0]=(0.994619505257)*x[0]**o+(0.33482820568)*x[0]+(-0.12721791236)*x[1]**o+(-0.505873171195)*x[1]
        arg[0,2,0,1]=(0.224805131522)*x[0]**o+(0.332022434603)*x[0]+(0.0583494507268)*x[1]**o+(0.857185410994)*x[1]
        arg[0,2,1,0]=(-0.425367077516)*x[0]**o+(-0.25207101061)*x[0]+(-0.11164491753)*x[1]**o+(0.44752084285)*x[1]
        arg[0,2,1,1]=(-0.158725210751)*x[0]**o+(-0.0692363138932)*x[0]+(0.214803128256)*x[1]**o+(0.886699106748)*x[1]
        arg[0,2,2,0]=(-0.332146331257)*x[0]**o+(0.114845050128)*x[0]+(-0.0781411611124)*x[1]**o+(0.474972057387)*x[1]
        arg[0,2,2,1]=(-0.481949630694)*x[0]**o+(0.644984972151)*x[0]+(0.171146519527)*x[1]**o+(0.57534803966)*x[1]
        arg[0,3,0,0]=(0.7580633024)*x[0]**o+(0.0280511776577)*x[0]+(0.375583652078)*x[1]**o+(-0.809112787284)*x[1]
        arg[0,3,0,1]=(-0.466737159934)*x[0]**o+(0.410032775686)*x[0]+(0.176141189584)*x[1]**o+(-0.228288309456)*x[1]
        arg[0,3,1,0]=(0.427129582196)*x[0]**o+(-0.718385859483)*x[0]+(-0.172261212989)*x[1]**o+(0.913157476188)*x[1]
        arg[0,3,1,1]=(0.952506706487)*x[0]**o+(-0.497478163899)*x[0]+(-0.930273038716)*x[1]**o+(0.673573149025)*x[1]
        arg[0,3,2,0]=(-0.0074302930747)*x[0]**o+(0.226424842035)*x[0]+(0.871190620671)*x[1]**o+(0.277695333679)*x[1]
        arg[0,3,2,1]=(-0.540820239946)*x[0]**o+(0.33044763211)*x[0]+(-0.772648820924)*x[1]**o+(-0.362410088021)*x[1]
        arg[0,4,0,0]=(-0.748396039111)*x[0]**o+(0.215131518212)*x[0]+(0.514087824403)*x[1]**o+(0.335563854205)*x[1]
        arg[0,4,0,1]=(-0.443263590391)*x[0]**o+(-0.523167282542)*x[0]+(0.122612452799)*x[1]**o+(0.216527857615)*x[1]
        arg[0,4,1,0]=(0.448176296777)*x[0]**o+(0.478292146131)*x[0]+(-0.558921083502)*x[1]**o+(-0.17018258654)*x[1]
        arg[0,4,1,1]=(-0.912765673953)*x[0]**o+(0.751746623223)*x[0]+(-0.104811579057)*x[1]**o+(0.456436870323)*x[1]
        arg[0,4,2,0]=(0.157946450932)*x[0]**o+(0.147858587857)*x[0]+(0.775747362874)*x[1]**o+(0.478478862463)*x[1]
        arg[0,4,2,1]=(-0.819625058185)*x[0]**o+(-0.0231980393919)*x[0]+(-0.186647585522)*x[1]**o+(0.513350371371)*x[1]
        arg[1,0,0,0]=(-0.942373946995)*x[0]**o+(-0.761326160931)*x[0]+(0.192511867167)*x[1]**o+(-0.50895969579)*x[1]
        arg[1,0,0,1]=(0.0670813349951)*x[0]**o+(0.55602953512)*x[0]+(0.757621178635)*x[1]**o+(0.741550201759)*x[1]
        arg[1,0,1,0]=(-0.114642609493)*x[0]**o+(-0.639500048447)*x[0]+(-0.861813039243)*x[1]**o+(0.36413406744)*x[1]
        arg[1,0,1,1]=(0.126555451594)*x[0]**o+(-0.885172557082)*x[0]+(-0.376073252921)*x[1]**o+(0.166102850653)*x[1]
        arg[1,0,2,0]=(-0.381517068219)*x[0]**o+(0.94870328267)*x[0]+(-0.129934777755)*x[1]**o+(0.78090715758)*x[1]
        arg[1,0,2,1]=(-0.535852302247)*x[0]**o+(0.63751039318)*x[0]+(-0.537556159925)*x[1]**o+(0.147766714221)*x[1]
        arg[1,1,0,0]=(-0.605543047533)*x[0]**o+(-0.366612785372)*x[0]+(0.544900149212)*x[1]**o+(0.722108900447)*x[1]
        arg[1,1,0,1]=(0.133378561142)*x[0]**o+(-0.227808772494)*x[0]+(-0.729399624175)*x[1]**o+(0.307571743455)*x[1]
        arg[1,1,1,0]=(0.0458830538139)*x[0]**o+(0.0585044493026)*x[0]+(-0.136630547605)*x[1]**o+(-0.0823288389309)*x[1]
        arg[1,1,1,1]=(0.339263808784)*x[0]**o+(0.00988917109076)*x[0]+(-0.786591838083)*x[1]**o+(-0.361686799179)*x[1]
        arg[1,1,2,0]=(0.789942037592)*x[0]**o+(-0.266689758676)*x[0]+(0.464856327495)*x[1]**o+(0.138022774293)*x[1]
        arg[1,1,2,1]=(-0.723544867217)*x[0]**o+(-0.523023674205)*x[0]+(0.629726026331)*x[1]**o+(-0.867808120997)*x[1]
        arg[1,2,0,0]=(-0.495589518787)*x[0]**o+(-0.561664880262)*x[0]+(0.0319320417078)*x[1]**o+(-0.153720661498)*x[1]
        arg[1,2,0,1]=(0.293168114024)*x[0]**o+(-0.344107163981)*x[0]+(-0.275629670068)*x[1]**o+(0.06103653487)*x[1]
        arg[1,2,1,0]=(0.495640927339)*x[0]**o+(0.596540095043)*x[0]+(0.208792567756)*x[1]**o+(0.886040724145)*x[1]
        arg[1,2,1,1]=(0.331476398093)*x[0]**o+(-0.472588572455)*x[0]+(-0.455670457285)*x[1]**o+(-0.731223615642)*x[1]
        arg[1,2,2,0]=(-0.127107829674)*x[0]**o+(-0.21113128456)*x[0]+(-0.733928194168)*x[1]**o+(-0.174127696853)*x[1]
        arg[1,2,2,1]=(0.521879316151)*x[0]**o+(0.983424557865)*x[0]+(-0.288659660182)*x[1]**o+(-0.431045203113)*x[1]
        arg[1,3,0,0]=(0.292348765746)*x[0]**o+(0.0730741240915)*x[0]+(0.079624022683)*x[1]**o+(0.237483071633)*x[1]
        arg[1,3,0,1]=(0.291035771848)*x[0]**o+(-0.482909477855)*x[0]+(-0.999142129615)*x[1]**o+(-0.559723367197)*x[1]
        arg[1,3,1,0]=(0.747096275653)*x[0]**o+(0.762757649624)*x[0]+(-0.821650889507)*x[1]**o+(0.301128670189)*x[1]
        arg[1,3,1,1]=(0.82703681414)*x[0]**o+(-0.336570070436)*x[0]+(0.798131952979)*x[1]**o+(0.603562737956)*x[1]
        arg[1,3,2,0]=(-0.164875914126)*x[0]**o+(0.219409873902)*x[0]+(0.784253160678)*x[1]**o+(0.0219597460707)*x[1]
        arg[1,3,2,1]=(0.124778899794)*x[0]**o+(0.785367842145)*x[0]+(0.948111125908)*x[1]**o+(-0.887427847089)*x[1]
        arg[1,4,0,0]=(0.0911816002883)*x[0]**o+(0.645410022441)*x[0]+(0.720551531965)*x[1]**o+(-0.388407683413)*x[1]
        arg[1,4,0,1]=(-0.0416762613062)*x[0]**o+(-0.548874829834)*x[0]+(-0.377952555557)*x[1]**o+(0.742795234361)*x[1]
        arg[1,4,1,0]=(-0.896929017971)*x[0]**o+(0.678343686516)*x[0]+(0.711572763082)*x[1]**o+(-0.43397331651)*x[1]
        arg[1,4,1,1]=(-0.190804421566)*x[0]**o+(-0.0848405124789)*x[0]+(0.36246957905)*x[1]**o+(0.963599086227)*x[1]
        arg[1,4,2,0]=(-0.0210248567982)*x[0]**o+(0.545977814581)*x[0]+(0.858962864401)*x[1]**o+(0.113792851997)*x[1]
        arg[1,4,2,1]=(0.229641892622)*x[0]**o+(-0.739561617198)*x[0]+(-0.164372740617)*x[1]**o+(0.537097868843)*x[1]
        arg[2,0,0,0]=(-0.13321389343)*x[0]**o+(-0.668403350503)*x[0]+(-0.924879105983)*x[1]**o+(-0.5161790717)*x[1]
        arg[2,0,0,1]=(0.26582844685)*x[0]**o+(0.89048660684)*x[0]+(-0.377225981719)*x[1]**o+(-0.303790710194)*x[1]
        arg[2,0,1,0]=(0.0997470860283)*x[0]**o+(0.0385427264217)*x[0]+(-0.71387347273)*x[1]**o+(-0.621376499192)*x[1]
        arg[2,0,1,1]=(0.595544616326)*x[0]**o+(-0.850529503975)*x[0]+(0.602203957938)*x[1]**o+(-0.329484843907)*x[1]
        arg[2,0,2,0]=(-0.106031967029)*x[0]**o+(-0.165122821622)*x[0]+(0.647473815919)*x[1]**o+(0.28942226062)*x[1]
        arg[2,0,2,1]=(0.585955875472)*x[0]**o+(0.922667710403)*x[0]+(-0.0710905957841)*x[1]**o+(-0.669138163697)*x[1]
        arg[2,1,0,0]=(0.461941211081)*x[0]**o+(0.723022466263)*x[0]+(-0.118551893472)*x[1]**o+(0.00445187353637)*x[1]
        arg[2,1,0,1]=(-0.253168248108)*x[0]**o+(0.476938842605)*x[0]+(0.476641658034)*x[1]**o+(-0.916886387595)*x[1]
        arg[2,1,1,0]=(0.722029309503)*x[0]**o+(0.0588848593425)*x[0]+(0.84737370043)*x[1]**o+(-0.674604415213)*x[1]
        arg[2,1,1,1]=(-0.644736759298)*x[0]**o+(-0.212847303893)*x[0]+(0.25394587552)*x[1]**o+(-0.366955175081)*x[1]
        arg[2,1,2,0]=(-0.657395296801)*x[0]**o+(-0.967570027599)*x[0]+(0.180926628541)*x[1]**o+(-0.940599897942)*x[1]
        arg[2,1,2,1]=(0.849897328801)*x[0]**o+(-0.299199474051)*x[0]+(0.427221207827)*x[1]**o+(0.212027985441)*x[1]
        arg[2,2,0,0]=(-0.561573507381)*x[0]**o+(-0.446925224908)*x[0]+(0.132091224948)*x[1]**o+(-0.673375661849)*x[1]
        arg[2,2,0,1]=(-0.386657291574)*x[0]**o+(0.356784637112)*x[0]+(0.644292006233)*x[1]**o+(0.922937167842)*x[1]
        arg[2,2,1,0]=(0.484446757285)*x[0]**o+(-0.822469870455)*x[0]+(-0.89605890123)*x[1]**o+(-0.895891500031)*x[1]
        arg[2,2,1,1]=(-0.0930740520192)*x[0]**o+(0.953356473678)*x[0]+(-0.714418458806)*x[1]**o+(-0.100584950946)*x[1]
        arg[2,2,2,0]=(0.945572242667)*x[0]**o+(-0.259280849605)*x[0]+(-0.925459846341)*x[1]**o+(0.906002758873)*x[1]
        arg[2,2,2,1]=(-0.638993332975)*x[0]**o+(-0.668594641703)*x[0]+(0.012666231369)*x[1]**o+(0.89732685856)*x[1]
        arg[2,3,0,0]=(-0.885061116144)*x[0]**o+(-0.576097536483)*x[0]+(-0.43166075788)*x[1]**o+(0.538821620917)*x[1]
        arg[2,3,0,1]=(-0.0743638381797)*x[0]**o+(0.251037500035)*x[0]+(-0.905989928076)*x[1]**o+(0.243771000867)*x[1]
        arg[2,3,1,0]=(-0.016045622351)*x[0]**o+(-0.632944072174)*x[0]+(-0.296175918793)*x[1]**o+(-0.00120123641365)*x[1]
        arg[2,3,1,1]=(-0.896354921812)*x[0]**o+(-0.331348989027)*x[0]+(-0.450865285201)*x[1]**o+(-0.243268582009)*x[1]
        arg[2,3,2,0]=(0.609332927431)*x[0]**o+(-0.0815856802115)*x[0]+(0.201924792273)*x[1]**o+(-0.890424451348)*x[1]
        arg[2,3,2,1]=(0.395090137264)*x[0]**o+(0.294230442573)*x[0]+(-0.345371510364)*x[1]**o+(0.181229663241)*x[1]
        arg[2,4,0,0]=(-0.823684509034)*x[0]**o+(0.431019301942)*x[0]+(0.554707015054)*x[1]**o+(-0.846322903003)*x[1]
        arg[2,4,0,1]=(0.862984476658)*x[0]**o+(0.20512500102)*x[0]+(0.0247756159287)*x[1]**o+(-0.374795100946)*x[1]
        arg[2,4,1,0]=(-0.0849914269517)*x[0]**o+(0.754908200339)*x[0]+(0.593363754807)*x[1]**o+(0.124923961126)*x[1]
        arg[2,4,1,1]=(-0.691440264234)*x[0]**o+(0.825739298015)*x[0]+(0.426452231478)*x[1]**o+(0.0995610721341)*x[1]
        arg[2,4,2,0]=(-0.175322750567)*x[0]**o+(0.419750865569)*x[0]+(0.208448987712)*x[1]**o+(0.150584798952)*x[1]
        arg[2,4,2,1]=(0.931680152344)*x[0]**o+(0.390947206307)*x[0]+(0.388437348378)*x[1]**o+(-0.808413079032)*x[1]
        arg[3,0,0,0]=(0.58855431402)*x[0]**o+(0.383865913364)*x[0]+(-0.691445323219)*x[1]**o+(-0.443370658523)*x[1]
        arg[3,0,0,1]=(-0.657371481478)*x[0]**o+(0.746891082673)*x[0]+(-0.829229947773)*x[1]**o+(0.761915670404)*x[1]
        arg[3,0,1,0]=(-0.431502932714)*x[0]**o+(-0.961472285716)*x[0]+(0.0423384388031)*x[1]**o+(0.339373458118)*x[1]
        arg[3,0,1,1]=(-0.272831163259)*x[0]**o+(-0.533438904494)*x[0]+(-0.595625384875)*x[1]**o+(0.0366857022323)*x[1]
        arg[3,0,2,0]=(-0.319684662306)*x[0]**o+(0.672185828912)*x[0]+(0.483126548994)*x[1]**o+(0.745872629604)*x[1]
        arg[3,0,2,1]=(0.437895277404)*x[0]**o+(-0.337209416963)*x[0]+(0.101080522788)*x[1]**o+(0.892022488644)*x[1]
        arg[3,1,0,0]=(-0.582325078455)*x[0]**o+(-0.854689347157)*x[0]+(0.0308638312259)*x[1]**o+(0.633382365704)*x[1]
        arg[3,1,0,1]=(0.319735178376)*x[0]**o+(-0.782764593952)*x[0]+(0.082230132244)*x[1]**o+(-0.946010937852)*x[1]
        arg[3,1,1,0]=(-0.698490410666)*x[0]**o+(-0.801358802689)*x[0]+(0.20695648235)*x[1]**o+(0.75257523471)*x[1]
        arg[3,1,1,1]=(0.66321696527)*x[0]**o+(0.704572918868)*x[0]+(0.528208879002)*x[1]**o+(-0.408378325513)*x[1]
        arg[3,1,2,0]=(-0.925177198417)*x[0]**o+(-0.152427811549)*x[0]+(0.341781993583)*x[1]**o+(-0.589976456377)*x[1]
        arg[3,1,2,1]=(0.206706628353)*x[0]**o+(0.673777173329)*x[0]+(-0.994138314254)*x[1]**o+(-0.843696565863)*x[1]
        arg[3,2,0,0]=(-0.588881067333)*x[0]**o+(-0.537572388341)*x[0]+(-0.206570371675)*x[1]**o+(-0.127468554294)*x[1]
        arg[3,2,0,1]=(0.498638284096)*x[0]**o+(-0.488278831086)*x[0]+(-0.28320874594)*x[1]**o+(-0.638241576615)*x[1]
        arg[3,2,1,0]=(-0.445780790822)*x[0]**o+(-0.347493953089)*x[0]+(-0.558010086232)*x[1]**o+(-0.0287339307689)*x[1]
        arg[3,2,1,1]=(-0.361727811272)*x[0]**o+(-0.559565378007)*x[0]+(-0.211685558972)*x[1]**o+(0.330088396811)*x[1]
        arg[3,2,2,0]=(-0.42600117831)*x[0]**o+(-0.488997111988)*x[0]+(0.127421278051)*x[1]**o+(-0.48233167806)*x[1]
        arg[3,2,2,1]=(0.120468829686)*x[0]**o+(0.0675603906118)*x[0]+(-0.60674361864)*x[1]**o+(0.639115708258)*x[1]
        arg[3,3,0,0]=(0.369343033015)*x[0]**o+(0.707735743593)*x[0]+(-0.426275591836)*x[1]**o+(-0.733377765625)*x[1]
        arg[3,3,0,1]=(-0.931245338156)*x[0]**o+(-0.465861794185)*x[0]+(-0.650664500687)*x[1]**o+(0.103077662402)*x[1]
        arg[3,3,1,0]=(0.731834744512)*x[0]**o+(0.267840571071)*x[0]+(0.952106457495)*x[1]**o+(-0.943845364366)*x[1]
        arg[3,3,1,1]=(-0.21692720916)*x[0]**o+(0.580841374997)*x[0]+(-0.412129341223)*x[1]**o+(0.40414445304)*x[1]
        arg[3,3,2,0]=(-0.447758598726)*x[0]**o+(0.0616926352912)*x[0]+(-0.824624226122)*x[1]**o+(0.19099833104)*x[1]
        arg[3,3,2,1]=(-0.615519138165)*x[0]**o+(0.233926822164)*x[0]+(-0.969014060236)*x[1]**o+(0.835103627893)*x[1]
        arg[3,4,0,0]=(0.550778983301)*x[0]**o+(0.0198253530882)*x[0]+(-0.39383849187)*x[1]**o+(0.367721826304)*x[1]
        arg[3,4,0,1]=(0.444877923149)*x[0]**o+(-0.899159835223)*x[0]+(0.923392540452)*x[1]**o+(0.512905141718)*x[1]
        arg[3,4,1,0]=(-0.563807675145)*x[0]**o+(0.994405383899)*x[0]+(-0.777352948218)*x[1]**o+(0.865994906227)*x[1]
        arg[3,4,1,1]=(-0.593800452139)*x[0]**o+(0.613778803237)*x[0]+(0.582200979718)*x[1]**o+(0.448026243271)*x[1]
        arg[3,4,2,0]=(-0.900331125416)*x[0]**o+(0.234613965424)*x[0]+(0.510400273805)*x[1]**o+(0.0836773641661)*x[1]
        arg[3,4,2,1]=(0.710999097307)*x[0]**o+(0.886603988265)*x[0]+(-0.405644887679)*x[1]**o+(0.362257258809)*x[1]
        ref[0,0,0,0]=(-0.0338299192888)*x_ref[0]**o+(-0.114906274359)*x_ref[0]+(0.15010339398)*x_ref[1]**o+(0.625295991328)*x_ref[1]
        ref[0,0,0,1]=(-0.1311901039)*x_ref[0]**o+(-0.551712861217)*x_ref[0]+(-0.287411404506)*x_ref[1]**o+(0.0242820629572)*x_ref[1]
        ref[0,0,1,0]=(0.121962094111)*x_ref[0]**o+(0.203169622345)*x_ref[0]+(0.241504218053)*x_ref[1]**o+(-0.0696356239498)*x_ref[1]
        ref[0,0,1,1]=(0.0461537352329)*x_ref[0]**o+(-0.275908938567)*x_ref[0]+(-0.271184121027)*x_ref[1]**o+(-0.364024729567)*x_ref[1]
        ref[0,0,2,0]=(0.199272234315)*x_ref[0]**o+(-0.906364258855)*x_ref[0]+(-0.92102930037)*x_ref[1]**o+(0.0718981903454)*x_ref[1]
        ref[0,0,2,1]=(-0.742950958383)*x_ref[0]**o+(0.0301464847722)*x_ref[0]+(0.466993477687)*x_ref[1]**o+(-0.402131750698)*x_ref[1]
        ref[0,1,0,0]=(-0.0653776546375)*x_ref[0]**o+(-0.679360612372)*x_ref[0]+(-0.32783837768)*x_ref[1]**o+(-0.113753016428)*x_ref[1]
        ref[0,1,0,1]=(0.58610119414)*x_ref[0]**o+(0.142613166231)*x_ref[0]+(0.0409105619986)*x_ref[1]**o+(-0.660641663558)*x_ref[1]
        ref[0,1,1,0]=(0.212172308585)*x_ref[0]**o+(-0.452767768344)*x_ref[0]+(0.963053969889)*x_ref[1]**o+(0.752691773953)*x_ref[1]
        ref[0,1,1,1]=(-0.514998377887)*x_ref[0]**o+(0.0872464329801)*x_ref[0]+(0.265590295505)*x_ref[1]**o+(0.557238826549)*x_ref[1]
        ref[0,1,2,0]=(0.360523954724)*x_ref[0]**o+(0.550238402994)*x_ref[0]+(-0.206349415539)*x_ref[1]**o+(0.357304730511)*x_ref[1]
        ref[0,1,2,1]=(-0.872755217435)*x_ref[0]**o+(0.190943683664)*x_ref[0]+(-0.245454292354)*x_ref[1]**o+(0.507229578506)*x_ref[1]
        ref[0,2,0,0]=(0.994619505257)*x_ref[0]**o+(0.33482820568)*x_ref[0]+(-0.12721791236)*x_ref[1]**o+(-0.505873171195)*x_ref[1]
        ref[0,2,0,1]=(0.224805131522)*x_ref[0]**o+(0.332022434603)*x_ref[0]+(0.0583494507268)*x_ref[1]**o+(0.857185410994)*x_ref[1]
        ref[0,2,1,0]=(-0.425367077516)*x_ref[0]**o+(-0.25207101061)*x_ref[0]+(-0.11164491753)*x_ref[1]**o+(0.44752084285)*x_ref[1]
        ref[0,2,1,1]=(-0.158725210751)*x_ref[0]**o+(-0.0692363138932)*x_ref[0]+(0.214803128256)*x_ref[1]**o+(0.886699106748)*x_ref[1]
        ref[0,2,2,0]=(-0.332146331257)*x_ref[0]**o+(0.114845050128)*x_ref[0]+(-0.0781411611124)*x_ref[1]**o+(0.474972057387)*x_ref[1]
        ref[0,2,2,1]=(-0.481949630694)*x_ref[0]**o+(0.644984972151)*x_ref[0]+(0.171146519527)*x_ref[1]**o+(0.57534803966)*x_ref[1]
        ref[0,3,0,0]=(0.7580633024)*x_ref[0]**o+(0.0280511776577)*x_ref[0]+(0.375583652078)*x_ref[1]**o+(-0.809112787284)*x_ref[1]
        ref[0,3,0,1]=(-0.466737159934)*x_ref[0]**o+(0.410032775686)*x_ref[0]+(0.176141189584)*x_ref[1]**o+(-0.228288309456)*x_ref[1]
        ref[0,3,1,0]=(0.427129582196)*x_ref[0]**o+(-0.718385859483)*x_ref[0]+(-0.172261212989)*x_ref[1]**o+(0.913157476188)*x_ref[1]
        ref[0,3,1,1]=(0.952506706487)*x_ref[0]**o+(-0.497478163899)*x_ref[0]+(-0.930273038716)*x_ref[1]**o+(0.673573149025)*x_ref[1]
        ref[0,3,2,0]=(-0.0074302930747)*x_ref[0]**o+(0.226424842035)*x_ref[0]+(0.871190620671)*x_ref[1]**o+(0.277695333679)*x_ref[1]
        ref[0,3,2,1]=(-0.540820239946)*x_ref[0]**o+(0.33044763211)*x_ref[0]+(-0.772648820924)*x_ref[1]**o+(-0.362410088021)*x_ref[1]
        ref[0,4,0,0]=(-0.748396039111)*x_ref[0]**o+(0.215131518212)*x_ref[0]+(0.514087824403)*x_ref[1]**o+(0.335563854205)*x_ref[1]
        ref[0,4,0,1]=(-0.443263590391)*x_ref[0]**o+(-0.523167282542)*x_ref[0]+(0.122612452799)*x_ref[1]**o+(0.216527857615)*x_ref[1]
        ref[0,4,1,0]=(0.448176296777)*x_ref[0]**o+(0.478292146131)*x_ref[0]+(-0.558921083502)*x_ref[1]**o+(-0.17018258654)*x_ref[1]
        ref[0,4,1,1]=(-0.912765673953)*x_ref[0]**o+(0.751746623223)*x_ref[0]+(-0.104811579057)*x_ref[1]**o+(0.456436870323)*x_ref[1]
        ref[0,4,2,0]=(0.157946450932)*x_ref[0]**o+(0.147858587857)*x_ref[0]+(0.775747362874)*x_ref[1]**o+(0.478478862463)*x_ref[1]
        ref[0,4,2,1]=(-0.819625058185)*x_ref[0]**o+(-0.0231980393919)*x_ref[0]+(-0.186647585522)*x_ref[1]**o+(0.513350371371)*x_ref[1]
        ref[1,0,0,0]=(-0.942373946995)*x_ref[0]**o+(-0.761326160931)*x_ref[0]+(0.192511867167)*x_ref[1]**o+(-0.50895969579)*x_ref[1]
        ref[1,0,0,1]=(0.0670813349951)*x_ref[0]**o+(0.55602953512)*x_ref[0]+(0.757621178635)*x_ref[1]**o+(0.741550201759)*x_ref[1]
        ref[1,0,1,0]=(-0.114642609493)*x_ref[0]**o+(-0.639500048447)*x_ref[0]+(-0.861813039243)*x_ref[1]**o+(0.36413406744)*x_ref[1]
        ref[1,0,1,1]=(0.126555451594)*x_ref[0]**o+(-0.885172557082)*x_ref[0]+(-0.376073252921)*x_ref[1]**o+(0.166102850653)*x_ref[1]
        ref[1,0,2,0]=(-0.381517068219)*x_ref[0]**o+(0.94870328267)*x_ref[0]+(-0.129934777755)*x_ref[1]**o+(0.78090715758)*x_ref[1]
        ref[1,0,2,1]=(-0.535852302247)*x_ref[0]**o+(0.63751039318)*x_ref[0]+(-0.537556159925)*x_ref[1]**o+(0.147766714221)*x_ref[1]
        ref[1,1,0,0]=(-0.605543047533)*x_ref[0]**o+(-0.366612785372)*x_ref[0]+(0.544900149212)*x_ref[1]**o+(0.722108900447)*x_ref[1]
        ref[1,1,0,1]=(0.133378561142)*x_ref[0]**o+(-0.227808772494)*x_ref[0]+(-0.729399624175)*x_ref[1]**o+(0.307571743455)*x_ref[1]
        ref[1,1,1,0]=(0.0458830538139)*x_ref[0]**o+(0.0585044493026)*x_ref[0]+(-0.136630547605)*x_ref[1]**o+(-0.0823288389309)*x_ref[1]
        ref[1,1,1,1]=(0.339263808784)*x_ref[0]**o+(0.00988917109076)*x_ref[0]+(-0.786591838083)*x_ref[1]**o+(-0.361686799179)*x_ref[1]
        ref[1,1,2,0]=(0.789942037592)*x_ref[0]**o+(-0.266689758676)*x_ref[0]+(0.464856327495)*x_ref[1]**o+(0.138022774293)*x_ref[1]
        ref[1,1,2,1]=(-0.723544867217)*x_ref[0]**o+(-0.523023674205)*x_ref[0]+(0.629726026331)*x_ref[1]**o+(-0.867808120997)*x_ref[1]
        ref[1,2,0,0]=(-0.495589518787)*x_ref[0]**o+(-0.561664880262)*x_ref[0]+(0.0319320417078)*x_ref[1]**o+(-0.153720661498)*x_ref[1]
        ref[1,2,0,1]=(0.293168114024)*x_ref[0]**o+(-0.344107163981)*x_ref[0]+(-0.275629670068)*x_ref[1]**o+(0.06103653487)*x_ref[1]
        ref[1,2,1,0]=(0.495640927339)*x_ref[0]**o+(0.596540095043)*x_ref[0]+(0.208792567756)*x_ref[1]**o+(0.886040724145)*x_ref[1]
        ref[1,2,1,1]=(0.331476398093)*x_ref[0]**o+(-0.472588572455)*x_ref[0]+(-0.455670457285)*x_ref[1]**o+(-0.731223615642)*x_ref[1]
        ref[1,2,2,0]=(-0.127107829674)*x_ref[0]**o+(-0.21113128456)*x_ref[0]+(-0.733928194168)*x_ref[1]**o+(-0.174127696853)*x_ref[1]
        ref[1,2,2,1]=(0.521879316151)*x_ref[0]**o+(0.983424557865)*x_ref[0]+(-0.288659660182)*x_ref[1]**o+(-0.431045203113)*x_ref[1]
        ref[1,3,0,0]=(0.292348765746)*x_ref[0]**o+(0.0730741240915)*x_ref[0]+(0.079624022683)*x_ref[1]**o+(0.237483071633)*x_ref[1]
        ref[1,3,0,1]=(0.291035771848)*x_ref[0]**o+(-0.482909477855)*x_ref[0]+(-0.999142129615)*x_ref[1]**o+(-0.559723367197)*x_ref[1]
        ref[1,3,1,0]=(0.747096275653)*x_ref[0]**o+(0.762757649624)*x_ref[0]+(-0.821650889507)*x_ref[1]**o+(0.301128670189)*x_ref[1]
        ref[1,3,1,1]=(0.82703681414)*x_ref[0]**o+(-0.336570070436)*x_ref[0]+(0.798131952979)*x_ref[1]**o+(0.603562737956)*x_ref[1]
        ref[1,3,2,0]=(-0.164875914126)*x_ref[0]**o+(0.219409873902)*x_ref[0]+(0.784253160678)*x_ref[1]**o+(0.0219597460707)*x_ref[1]
        ref[1,3,2,1]=(0.124778899794)*x_ref[0]**o+(0.785367842145)*x_ref[0]+(0.948111125908)*x_ref[1]**o+(-0.887427847089)*x_ref[1]
        ref[1,4,0,0]=(0.0911816002883)*x_ref[0]**o+(0.645410022441)*x_ref[0]+(0.720551531965)*x_ref[1]**o+(-0.388407683413)*x_ref[1]
        ref[1,4,0,1]=(-0.0416762613062)*x_ref[0]**o+(-0.548874829834)*x_ref[0]+(-0.377952555557)*x_ref[1]**o+(0.742795234361)*x_ref[1]
        ref[1,4,1,0]=(-0.896929017971)*x_ref[0]**o+(0.678343686516)*x_ref[0]+(0.711572763082)*x_ref[1]**o+(-0.43397331651)*x_ref[1]
        ref[1,4,1,1]=(-0.190804421566)*x_ref[0]**o+(-0.0848405124789)*x_ref[0]+(0.36246957905)*x_ref[1]**o+(0.963599086227)*x_ref[1]
        ref[1,4,2,0]=(-0.0210248567982)*x_ref[0]**o+(0.545977814581)*x_ref[0]+(0.858962864401)*x_ref[1]**o+(0.113792851997)*x_ref[1]
        ref[1,4,2,1]=(0.229641892622)*x_ref[0]**o+(-0.739561617198)*x_ref[0]+(-0.164372740617)*x_ref[1]**o+(0.537097868843)*x_ref[1]
        ref[2,0,0,0]=(-0.13321389343)*x_ref[0]**o+(-0.668403350503)*x_ref[0]+(-0.924879105983)*x_ref[1]**o+(-0.5161790717)*x_ref[1]
        ref[2,0,0,1]=(0.26582844685)*x_ref[0]**o+(0.89048660684)*x_ref[0]+(-0.377225981719)*x_ref[1]**o+(-0.303790710194)*x_ref[1]
        ref[2,0,1,0]=(0.0997470860283)*x_ref[0]**o+(0.0385427264217)*x_ref[0]+(-0.71387347273)*x_ref[1]**o+(-0.621376499192)*x_ref[1]
        ref[2,0,1,1]=(0.595544616326)*x_ref[0]**o+(-0.850529503975)*x_ref[0]+(0.602203957938)*x_ref[1]**o+(-0.329484843907)*x_ref[1]
        ref[2,0,2,0]=(-0.106031967029)*x_ref[0]**o+(-0.165122821622)*x_ref[0]+(0.647473815919)*x_ref[1]**o+(0.28942226062)*x_ref[1]
        ref[2,0,2,1]=(0.585955875472)*x_ref[0]**o+(0.922667710403)*x_ref[0]+(-0.0710905957841)*x_ref[1]**o+(-0.669138163697)*x_ref[1]
        ref[2,1,0,0]=(0.461941211081)*x_ref[0]**o+(0.723022466263)*x_ref[0]+(-0.118551893472)*x_ref[1]**o+(0.00445187353637)*x_ref[1]
        ref[2,1,0,1]=(-0.253168248108)*x_ref[0]**o+(0.476938842605)*x_ref[0]+(0.476641658034)*x_ref[1]**o+(-0.916886387595)*x_ref[1]
        ref[2,1,1,0]=(0.722029309503)*x_ref[0]**o+(0.0588848593425)*x_ref[0]+(0.84737370043)*x_ref[1]**o+(-0.674604415213)*x_ref[1]
        ref[2,1,1,1]=(-0.644736759298)*x_ref[0]**o+(-0.212847303893)*x_ref[0]+(0.25394587552)*x_ref[1]**o+(-0.366955175081)*x_ref[1]
        ref[2,1,2,0]=(-0.657395296801)*x_ref[0]**o+(-0.967570027599)*x_ref[0]+(0.180926628541)*x_ref[1]**o+(-0.940599897942)*x_ref[1]
        ref[2,1,2,1]=(0.849897328801)*x_ref[0]**o+(-0.299199474051)*x_ref[0]+(0.427221207827)*x_ref[1]**o+(0.212027985441)*x_ref[1]
        ref[2,2,0,0]=(-0.561573507381)*x_ref[0]**o+(-0.446925224908)*x_ref[0]+(0.132091224948)*x_ref[1]**o+(-0.673375661849)*x_ref[1]
        ref[2,2,0,1]=(-0.386657291574)*x_ref[0]**o+(0.356784637112)*x_ref[0]+(0.644292006233)*x_ref[1]**o+(0.922937167842)*x_ref[1]
        ref[2,2,1,0]=(0.484446757285)*x_ref[0]**o+(-0.822469870455)*x_ref[0]+(-0.89605890123)*x_ref[1]**o+(-0.895891500031)*x_ref[1]
        ref[2,2,1,1]=(-0.0930740520192)*x_ref[0]**o+(0.953356473678)*x_ref[0]+(-0.714418458806)*x_ref[1]**o+(-0.100584950946)*x_ref[1]
        ref[2,2,2,0]=(0.945572242667)*x_ref[0]**o+(-0.259280849605)*x_ref[0]+(-0.925459846341)*x_ref[1]**o+(0.906002758873)*x_ref[1]
        ref[2,2,2,1]=(-0.638993332975)*x_ref[0]**o+(-0.668594641703)*x_ref[0]+(0.012666231369)*x_ref[1]**o+(0.89732685856)*x_ref[1]
        ref[2,3,0,0]=(-0.885061116144)*x_ref[0]**o+(-0.576097536483)*x_ref[0]+(-0.43166075788)*x_ref[1]**o+(0.538821620917)*x_ref[1]
        ref[2,3,0,1]=(-0.0743638381797)*x_ref[0]**o+(0.251037500035)*x_ref[0]+(-0.905989928076)*x_ref[1]**o+(0.243771000867)*x_ref[1]
        ref[2,3,1,0]=(-0.016045622351)*x_ref[0]**o+(-0.632944072174)*x_ref[0]+(-0.296175918793)*x_ref[1]**o+(-0.00120123641365)*x_ref[1]
        ref[2,3,1,1]=(-0.896354921812)*x_ref[0]**o+(-0.331348989027)*x_ref[0]+(-0.450865285201)*x_ref[1]**o+(-0.243268582009)*x_ref[1]
        ref[2,3,2,0]=(0.609332927431)*x_ref[0]**o+(-0.0815856802115)*x_ref[0]+(0.201924792273)*x_ref[1]**o+(-0.890424451348)*x_ref[1]
        ref[2,3,2,1]=(0.395090137264)*x_ref[0]**o+(0.294230442573)*x_ref[0]+(-0.345371510364)*x_ref[1]**o+(0.181229663241)*x_ref[1]
        ref[2,4,0,0]=(-0.823684509034)*x_ref[0]**o+(0.431019301942)*x_ref[0]+(0.554707015054)*x_ref[1]**o+(-0.846322903003)*x_ref[1]
        ref[2,4,0,1]=(0.862984476658)*x_ref[0]**o+(0.20512500102)*x_ref[0]+(0.0247756159287)*x_ref[1]**o+(-0.374795100946)*x_ref[1]
        ref[2,4,1,0]=(-0.0849914269517)*x_ref[0]**o+(0.754908200339)*x_ref[0]+(0.593363754807)*x_ref[1]**o+(0.124923961126)*x_ref[1]
        ref[2,4,1,1]=(-0.691440264234)*x_ref[0]**o+(0.825739298015)*x_ref[0]+(0.426452231478)*x_ref[1]**o+(0.0995610721341)*x_ref[1]
        ref[2,4,2,0]=(-0.175322750567)*x_ref[0]**o+(0.419750865569)*x_ref[0]+(0.208448987712)*x_ref[1]**o+(0.150584798952)*x_ref[1]
        ref[2,4,2,1]=(0.931680152344)*x_ref[0]**o+(0.390947206307)*x_ref[0]+(0.388437348378)*x_ref[1]**o+(-0.808413079032)*x_ref[1]
        ref[3,0,0,0]=(0.58855431402)*x_ref[0]**o+(0.383865913364)*x_ref[0]+(-0.691445323219)*x_ref[1]**o+(-0.443370658523)*x_ref[1]
        ref[3,0,0,1]=(-0.657371481478)*x_ref[0]**o+(0.746891082673)*x_ref[0]+(-0.829229947773)*x_ref[1]**o+(0.761915670404)*x_ref[1]
        ref[3,0,1,0]=(-0.431502932714)*x_ref[0]**o+(-0.961472285716)*x_ref[0]+(0.0423384388031)*x_ref[1]**o+(0.339373458118)*x_ref[1]
        ref[3,0,1,1]=(-0.272831163259)*x_ref[0]**o+(-0.533438904494)*x_ref[0]+(-0.595625384875)*x_ref[1]**o+(0.0366857022323)*x_ref[1]
        ref[3,0,2,0]=(-0.319684662306)*x_ref[0]**o+(0.672185828912)*x_ref[0]+(0.483126548994)*x_ref[1]**o+(0.745872629604)*x_ref[1]
        ref[3,0,2,1]=(0.437895277404)*x_ref[0]**o+(-0.337209416963)*x_ref[0]+(0.101080522788)*x_ref[1]**o+(0.892022488644)*x_ref[1]
        ref[3,1,0,0]=(-0.582325078455)*x_ref[0]**o+(-0.854689347157)*x_ref[0]+(0.0308638312259)*x_ref[1]**o+(0.633382365704)*x_ref[1]
        ref[3,1,0,1]=(0.319735178376)*x_ref[0]**o+(-0.782764593952)*x_ref[0]+(0.082230132244)*x_ref[1]**o+(-0.946010937852)*x_ref[1]
        ref[3,1,1,0]=(-0.698490410666)*x_ref[0]**o+(-0.801358802689)*x_ref[0]+(0.20695648235)*x_ref[1]**o+(0.75257523471)*x_ref[1]
        ref[3,1,1,1]=(0.66321696527)*x_ref[0]**o+(0.704572918868)*x_ref[0]+(0.528208879002)*x_ref[1]**o+(-0.408378325513)*x_ref[1]
        ref[3,1,2,0]=(-0.925177198417)*x_ref[0]**o+(-0.152427811549)*x_ref[0]+(0.341781993583)*x_ref[1]**o+(-0.589976456377)*x_ref[1]
        ref[3,1,2,1]=(0.206706628353)*x_ref[0]**o+(0.673777173329)*x_ref[0]+(-0.994138314254)*x_ref[1]**o+(-0.843696565863)*x_ref[1]
        ref[3,2,0,0]=(-0.588881067333)*x_ref[0]**o+(-0.537572388341)*x_ref[0]+(-0.206570371675)*x_ref[1]**o+(-0.127468554294)*x_ref[1]
        ref[3,2,0,1]=(0.498638284096)*x_ref[0]**o+(-0.488278831086)*x_ref[0]+(-0.28320874594)*x_ref[1]**o+(-0.638241576615)*x_ref[1]
        ref[3,2,1,0]=(-0.445780790822)*x_ref[0]**o+(-0.347493953089)*x_ref[0]+(-0.558010086232)*x_ref[1]**o+(-0.0287339307689)*x_ref[1]
        ref[3,2,1,1]=(-0.361727811272)*x_ref[0]**o+(-0.559565378007)*x_ref[0]+(-0.211685558972)*x_ref[1]**o+(0.330088396811)*x_ref[1]
        ref[3,2,2,0]=(-0.42600117831)*x_ref[0]**o+(-0.488997111988)*x_ref[0]+(0.127421278051)*x_ref[1]**o+(-0.48233167806)*x_ref[1]
        ref[3,2,2,1]=(0.120468829686)*x_ref[0]**o+(0.0675603906118)*x_ref[0]+(-0.60674361864)*x_ref[1]**o+(0.639115708258)*x_ref[1]
        ref[3,3,0,0]=(0.369343033015)*x_ref[0]**o+(0.707735743593)*x_ref[0]+(-0.426275591836)*x_ref[1]**o+(-0.733377765625)*x_ref[1]
        ref[3,3,0,1]=(-0.931245338156)*x_ref[0]**o+(-0.465861794185)*x_ref[0]+(-0.650664500687)*x_ref[1]**o+(0.103077662402)*x_ref[1]
        ref[3,3,1,0]=(0.731834744512)*x_ref[0]**o+(0.267840571071)*x_ref[0]+(0.952106457495)*x_ref[1]**o+(-0.943845364366)*x_ref[1]
        ref[3,3,1,1]=(-0.21692720916)*x_ref[0]**o+(0.580841374997)*x_ref[0]+(-0.412129341223)*x_ref[1]**o+(0.40414445304)*x_ref[1]
        ref[3,3,2,0]=(-0.447758598726)*x_ref[0]**o+(0.0616926352912)*x_ref[0]+(-0.824624226122)*x_ref[1]**o+(0.19099833104)*x_ref[1]
        ref[3,3,2,1]=(-0.615519138165)*x_ref[0]**o+(0.233926822164)*x_ref[0]+(-0.969014060236)*x_ref[1]**o+(0.835103627893)*x_ref[1]
        ref[3,4,0,0]=(0.550778983301)*x_ref[0]**o+(0.0198253530882)*x_ref[0]+(-0.39383849187)*x_ref[1]**o+(0.367721826304)*x_ref[1]
        ref[3,4,0,1]=(0.444877923149)*x_ref[0]**o+(-0.899159835223)*x_ref[0]+(0.923392540452)*x_ref[1]**o+(0.512905141718)*x_ref[1]
        ref[3,4,1,0]=(-0.563807675145)*x_ref[0]**o+(0.994405383899)*x_ref[0]+(-0.777352948218)*x_ref[1]**o+(0.865994906227)*x_ref[1]
        ref[3,4,1,1]=(-0.593800452139)*x_ref[0]**o+(0.613778803237)*x_ref[0]+(0.582200979718)*x_ref[1]**o+(0.448026243271)*x_ref[1]
        ref[3,4,2,0]=(-0.900331125416)*x_ref[0]**o+(0.234613965424)*x_ref[0]+(0.510400273805)*x_ref[1]**o+(0.0836773641661)*x_ref[1]
        ref[3,4,2,1]=(0.710999097307)*x_ref[0]**o+(0.886603988265)*x_ref[0]+(-0.405644887679)*x_ref[1]**o+(0.362257258809)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.310272924207)*x[0]**o+(-0.664497806351)*x[0]+(0.741942132609)*x[1]**o+(0.636051298779)*x[1]+(0.76627159146)*x[2]**o+(0.492960732588)*x[2]
        arg[0,0,0,1]=(-0.513740801773)*x[0]**o+(-0.633330750909)*x[0]+(0.177934618983)*x[1]**o+(-0.323745723166)*x[1]+(0.694167180904)*x[2]**o+(0.441239251435)*x[2]
        arg[0,0,1,0]=(0.417477695848)*x[0]**o+(-0.923415976002)*x[0]+(0.0765555276662)*x[1]**o+(-0.277290890834)*x[1]+(0.0177191326614)*x[2]**o+(-0.0842520100387)*x[2]
        arg[0,0,1,1]=(-0.761916249894)*x[0]**o+(0.169673139552)*x[0]+(0.0306290828795)*x[1]**o+(0.842234592613)*x[1]+(-0.350342692669)*x[2]**o+(0.321292270658)*x[2]
        arg[0,0,2,0]=(-0.707494137624)*x[0]**o+(0.346873490395)*x[0]+(-0.68281925138)*x[1]**o+(-0.830538083274)*x[1]+(0.329372141613)*x[2]**o+(-0.121366157163)*x[2]
        arg[0,0,2,1]=(-0.708018186507)*x[0]**o+(-0.778051239314)*x[0]+(-0.43106291563)*x[1]**o+(-0.0115914251143)*x[1]+(0.487334215892)*x[2]**o+(-0.289232284658)*x[2]
        arg[0,1,0,0]=(0.530614704199)*x[0]**o+(0.518622028317)*x[0]+(-0.0143129589551)*x[1]**o+(0.808469352257)*x[1]+(-0.156787373225)*x[2]**o+(0.668887813143)*x[2]
        arg[0,1,0,1]=(0.953223233706)*x[0]**o+(-0.164323838353)*x[0]+(0.336717320858)*x[1]**o+(-0.520230488172)*x[1]+(0.689960738002)*x[2]**o+(0.487630947578)*x[2]
        arg[0,1,1,0]=(-0.418095373429)*x[0]**o+(-0.941817844939)*x[0]+(0.88105115607)*x[1]**o+(-0.820502501439)*x[1]+(-0.693858267667)*x[2]**o+(0.530567713127)*x[2]
        arg[0,1,1,1]=(-0.716861320218)*x[0]**o+(-0.736619859467)*x[0]+(-0.885478605601)*x[1]**o+(-0.129375578967)*x[1]+(-0.0181433907318)*x[2]**o+(0.373255537959)*x[2]
        arg[0,1,2,0]=(-0.174742583167)*x[0]**o+(-0.0887967480112)*x[0]+(-0.291168698662)*x[1]**o+(-0.188608506769)*x[1]+(0.22479613454)*x[2]**o+(0.34173928582)*x[2]
        arg[0,1,2,1]=(0.117208929504)*x[0]**o+(-0.420969282079)*x[0]+(0.585298109836)*x[1]**o+(0.771355162745)*x[1]+(-0.51374493747)*x[2]**o+(0.785016548855)*x[2]
        arg[0,2,0,0]=(0.737394887886)*x[0]**o+(0.966080271342)*x[0]+(-0.373178384492)*x[1]**o+(-0.845505386142)*x[1]+(-0.274654519371)*x[2]**o+(0.709356557887)*x[2]
        arg[0,2,0,1]=(0.0365864090636)*x[0]**o+(-0.393382479787)*x[0]+(0.018273360977)*x[1]**o+(-0.509201909328)*x[1]+(-0.777852392534)*x[2]**o+(0.0417280247507)*x[2]
        arg[0,2,1,0]=(0.637829302751)*x[0]**o+(-0.767824790186)*x[0]+(0.843382009511)*x[1]**o+(-0.599633234403)*x[1]+(0.14283138941)*x[2]**o+(-0.354404633991)*x[2]
        arg[0,2,1,1]=(-0.465258613599)*x[0]**o+(0.09341154383)*x[0]+(-0.629114782988)*x[1]**o+(0.91608655383)*x[1]+(-0.692901135848)*x[2]**o+(-0.707225310667)*x[2]
        arg[0,2,2,0]=(0.9935958759)*x[0]**o+(-0.385008742466)*x[0]+(0.153885356854)*x[1]**o+(0.146270680656)*x[1]+(0.305838410343)*x[2]**o+(0.141193552717)*x[2]
        arg[0,2,2,1]=(-0.487538023894)*x[0]**o+(-0.814283442647)*x[0]+(0.166008520199)*x[1]**o+(0.313655774416)*x[1]+(0.282308782973)*x[2]**o+(0.309176062484)*x[2]
        arg[0,3,0,0]=(0.00203507798913)*x[0]**o+(-0.19902542604)*x[0]+(0.88110918624)*x[1]**o+(0.223549219288)*x[1]+(0.135758952369)*x[2]**o+(0.59116783639)*x[2]
        arg[0,3,0,1]=(-0.185063700767)*x[0]**o+(0.833862155173)*x[0]+(-0.0939117509645)*x[1]**o+(0.288003692231)*x[1]+(0.580196157839)*x[2]**o+(0.299320177754)*x[2]
        arg[0,3,1,0]=(-0.039100718826)*x[0]**o+(0.423179869395)*x[0]+(-0.013669578975)*x[1]**o+(-0.355053343085)*x[1]+(0.89285158975)*x[2]**o+(0.392984367877)*x[2]
        arg[0,3,1,1]=(0.0988989416429)*x[0]**o+(-0.446243235174)*x[0]+(-0.691253313617)*x[1]**o+(0.338823693086)*x[1]+(0.707054126334)*x[2]**o+(0.601121388278)*x[2]
        arg[0,3,2,0]=(-0.191366770484)*x[0]**o+(0.961791275903)*x[0]+(-0.103547475035)*x[1]**o+(-0.745236510079)*x[1]+(-0.157506981712)*x[2]**o+(0.986229506787)*x[2]
        arg[0,3,2,1]=(-0.886091883647)*x[0]**o+(0.6979804248)*x[0]+(-0.366575419428)*x[1]**o+(-0.958601153259)*x[1]+(-0.0119724511494)*x[2]**o+(-0.76101132076)*x[2]
        arg[0,4,0,0]=(0.923290554434)*x[0]**o+(0.411359492198)*x[0]+(-0.977999562875)*x[1]**o+(0.352085161527)*x[1]+(0.955745235349)*x[2]**o+(-0.978885609257)*x[2]
        arg[0,4,0,1]=(0.226973479105)*x[0]**o+(-0.423881897251)*x[0]+(0.889101969481)*x[1]**o+(-0.994236389028)*x[1]+(0.0159169033361)*x[2]**o+(-0.559935154203)*x[2]
        arg[0,4,1,0]=(-0.225427291797)*x[0]**o+(-0.557334621848)*x[0]+(-0.486252586142)*x[1]**o+(0.0497120905216)*x[1]+(0.84165736691)*x[2]**o+(-0.185010156797)*x[2]
        arg[0,4,1,1]=(0.762694602609)*x[0]**o+(-0.447367170727)*x[0]+(-0.61103181818)*x[1]**o+(0.689761057656)*x[1]+(-0.631855120122)*x[2]**o+(0.33274655602)*x[2]
        arg[0,4,2,0]=(-0.198884868914)*x[0]**o+(0.716371959278)*x[0]+(0.260871768919)*x[1]**o+(0.622241740082)*x[1]+(0.774103524035)*x[2]**o+(0.186132703676)*x[2]
        arg[0,4,2,1]=(-0.309450645317)*x[0]**o+(0.204754571336)*x[0]+(0.483402810171)*x[1]**o+(0.404620744188)*x[1]+(-0.699137246959)*x[2]**o+(0.467428706196)*x[2]
        arg[1,0,0,0]=(0.642289506448)*x[0]**o+(0.383990304394)*x[0]+(0.556940973318)*x[1]**o+(0.135219729041)*x[1]+(0.577195836358)*x[2]**o+(-0.338924419674)*x[2]
        arg[1,0,0,1]=(-0.920873911295)*x[0]**o+(-0.330610192304)*x[0]+(-0.117226803856)*x[1]**o+(0.172482867684)*x[1]+(0.909140902709)*x[2]**o+(-0.590494167913)*x[2]
        arg[1,0,1,0]=(0.210161538861)*x[0]**o+(-0.83860241761)*x[0]+(0.0257180382661)*x[1]**o+(0.877633854348)*x[1]+(0.26045311754)*x[2]**o+(-0.971349902714)*x[2]
        arg[1,0,1,1]=(0.70736843125)*x[0]**o+(-0.765121985431)*x[0]+(-0.196525225943)*x[1]**o+(-0.175928729346)*x[1]+(0.945523028748)*x[2]**o+(0.877830322823)*x[2]
        arg[1,0,2,0]=(-0.963513016065)*x[0]**o+(0.0820537856351)*x[0]+(0.0122070196118)*x[1]**o+(-0.154147449592)*x[1]+(0.323844633268)*x[2]**o+(-0.124897698214)*x[2]
        arg[1,0,2,1]=(0.656267566405)*x[0]**o+(0.0174786620768)*x[0]+(0.763902435618)*x[1]**o+(-0.986021506043)*x[1]+(0.632625555994)*x[2]**o+(-0.110657743523)*x[2]
        arg[1,1,0,0]=(0.00354144688732)*x[0]**o+(-0.813515808171)*x[0]+(-0.100717717446)*x[1]**o+(0.0543708163629)*x[1]+(0.785927880762)*x[2]**o+(0.0109166652227)*x[2]
        arg[1,1,0,1]=(-0.239705688631)*x[0]**o+(0.402915642509)*x[0]+(0.463371068163)*x[1]**o+(-0.507302462928)*x[1]+(-0.0789844128295)*x[2]**o+(0.669144891499)*x[2]
        arg[1,1,1,0]=(-0.491891367001)*x[0]**o+(-0.524377636928)*x[0]+(0.903091863917)*x[1]**o+(-0.554368807183)*x[1]+(0.332899195339)*x[2]**o+(-0.645216147049)*x[2]
        arg[1,1,1,1]=(-0.915843941243)*x[0]**o+(-0.233175654748)*x[0]+(0.467730425037)*x[1]**o+(-0.765957936683)*x[1]+(0.726678877276)*x[2]**o+(0.770533947942)*x[2]
        arg[1,1,2,0]=(0.678071852032)*x[0]**o+(0.266537844237)*x[0]+(0.324335300151)*x[1]**o+(-0.286713971373)*x[1]+(-0.702384414014)*x[2]**o+(0.887198276519)*x[2]
        arg[1,1,2,1]=(0.0466444959648)*x[0]**o+(0.689162912677)*x[0]+(-0.35033923676)*x[1]**o+(-0.0536380766876)*x[1]+(-0.511833442517)*x[2]**o+(0.125566299829)*x[2]
        arg[1,2,0,0]=(-0.803170431622)*x[0]**o+(0.476211169752)*x[0]+(-0.170510267953)*x[1]**o+(-0.765945735317)*x[1]+(-0.169966329291)*x[2]**o+(-0.988525918724)*x[2]
        arg[1,2,0,1]=(0.289618751212)*x[0]**o+(0.294379040976)*x[0]+(-0.922478416838)*x[1]**o+(0.210546495537)*x[1]+(-0.87266381496)*x[2]**o+(-0.865571669577)*x[2]
        arg[1,2,1,0]=(-0.292811054005)*x[0]**o+(0.651280833746)*x[0]+(-0.625004203736)*x[1]**o+(0.372078169172)*x[1]+(-0.0539099055169)*x[2]**o+(0.0911869270008)*x[2]
        arg[1,2,1,1]=(0.496217698294)*x[0]**o+(-0.986021787133)*x[0]+(0.246417059532)*x[1]**o+(0.610226802401)*x[1]+(0.991534665298)*x[2]**o+(-0.507441185982)*x[2]
        arg[1,2,2,0]=(-0.839902565947)*x[0]**o+(-0.279607147278)*x[0]+(0.278212307963)*x[1]**o+(0.609793640036)*x[1]+(0.940471913774)*x[2]**o+(0.21860828705)*x[2]
        arg[1,2,2,1]=(0.789578112278)*x[0]**o+(-0.0226697818608)*x[0]+(0.216321340291)*x[1]**o+(-0.701818463454)*x[1]+(-0.166689030646)*x[2]**o+(0.966131542292)*x[2]
        arg[1,3,0,0]=(0.149337042551)*x[0]**o+(-0.785050579674)*x[0]+(0.530219924721)*x[1]**o+(-0.971927239398)*x[1]+(0.950917908603)*x[2]**o+(0.132668077447)*x[2]
        arg[1,3,0,1]=(0.983081559724)*x[0]**o+(-0.358694498455)*x[0]+(0.181080099281)*x[1]**o+(-0.699756291643)*x[1]+(-0.744335953304)*x[2]**o+(0.00262043221382)*x[2]
        arg[1,3,1,0]=(0.31776032045)*x[0]**o+(-0.0726184220029)*x[0]+(-0.759883735971)*x[1]**o+(-0.6389710873)*x[1]+(-0.607858499502)*x[2]**o+(-0.397169895784)*x[2]
        arg[1,3,1,1]=(-0.950468389728)*x[0]**o+(0.831121915521)*x[0]+(0.552857011083)*x[1]**o+(0.848599672594)*x[1]+(0.432183564231)*x[2]**o+(0.57537721362)*x[2]
        arg[1,3,2,0]=(0.736080143601)*x[0]**o+(-0.804140307747)*x[0]+(-0.163864867884)*x[1]**o+(0.950874042315)*x[1]+(-0.99676139412)*x[2]**o+(0.332320294246)*x[2]
        arg[1,3,2,1]=(0.119082582041)*x[0]**o+(-0.969740778852)*x[0]+(0.828432583532)*x[1]**o+(0.190324050395)*x[1]+(-0.42639142674)*x[2]**o+(0.315519578192)*x[2]
        arg[1,4,0,0]=(-0.798465141775)*x[0]**o+(-0.717282977908)*x[0]+(0.935655148713)*x[1]**o+(-0.655752964484)*x[1]+(0.159990149099)*x[2]**o+(-0.537115062687)*x[2]
        arg[1,4,0,1]=(0.445151775007)*x[0]**o+(-0.867956302051)*x[0]+(-0.809856879825)*x[1]**o+(-0.326897750981)*x[1]+(0.756352311987)*x[2]**o+(0.577585513459)*x[2]
        arg[1,4,1,0]=(0.294243347591)*x[0]**o+(0.774741656931)*x[0]+(-0.0799985737984)*x[1]**o+(0.179789107412)*x[1]+(-0.612198025189)*x[2]**o+(-0.07086009611)*x[2]
        arg[1,4,1,1]=(0.0692586287045)*x[0]**o+(-0.671956300122)*x[0]+(-0.948667104683)*x[1]**o+(0.68656246801)*x[1]+(-0.352975297696)*x[2]**o+(0.277952113716)*x[2]
        arg[1,4,2,0]=(-0.491316095517)*x[0]**o+(0.725899087239)*x[0]+(0.634113963431)*x[1]**o+(0.235694393453)*x[1]+(-0.15600720713)*x[2]**o+(0.954814976539)*x[2]
        arg[1,4,2,1]=(0.120206996339)*x[0]**o+(-0.86151411823)*x[0]+(-0.137093650097)*x[1]**o+(0.633402978246)*x[1]+(-0.42515364728)*x[2]**o+(-0.987634210617)*x[2]
        arg[2,0,0,0]=(0.539463971222)*x[0]**o+(-0.03435650579)*x[0]+(-0.884395295544)*x[1]**o+(0.215454879022)*x[1]+(0.842377328246)*x[2]**o+(0.370920136871)*x[2]
        arg[2,0,0,1]=(-0.411505463039)*x[0]**o+(0.446494039439)*x[0]+(0.0313771969556)*x[1]**o+(0.602893223061)*x[1]+(-0.30383490021)*x[2]**o+(-0.788028756758)*x[2]
        arg[2,0,1,0]=(0.489009615999)*x[0]**o+(-0.95508248016)*x[0]+(-0.21489475022)*x[1]**o+(-0.857263478431)*x[1]+(-0.123703409045)*x[2]**o+(0.663389562431)*x[2]
        arg[2,0,1,1]=(-0.0224763310929)*x[0]**o+(-0.315886249237)*x[0]+(-0.318201849695)*x[1]**o+(0.423732682462)*x[1]+(0.396065636381)*x[2]**o+(0.0816717878865)*x[2]
        arg[2,0,2,0]=(-0.191432618224)*x[0]**o+(0.688389390795)*x[0]+(-0.382948112159)*x[1]**o+(0.562905173818)*x[1]+(-0.118396714029)*x[2]**o+(0.234478642577)*x[2]
        arg[2,0,2,1]=(-0.348008173171)*x[0]**o+(0.191994166636)*x[0]+(0.450556057147)*x[1]**o+(-0.710821693343)*x[1]+(-0.245119333857)*x[2]**o+(0.262869525239)*x[2]
        arg[2,1,0,0]=(-0.160444506203)*x[0]**o+(-0.816142837432)*x[0]+(-0.619722480492)*x[1]**o+(-0.989077039675)*x[1]+(-0.0282349103977)*x[2]**o+(-0.629849819691)*x[2]
        arg[2,1,0,1]=(0.379513765297)*x[0]**o+(0.934652105586)*x[0]+(-0.632636427868)*x[1]**o+(-0.0492550496553)*x[1]+(0.137733572307)*x[2]**o+(0.779782637942)*x[2]
        arg[2,1,1,0]=(0.469499320417)*x[0]**o+(0.570880924125)*x[0]+(-0.0329833251419)*x[1]**o+(-0.0388751034604)*x[1]+(-0.878520094389)*x[2]**o+(0.992656649048)*x[2]
        arg[2,1,1,1]=(0.592005181995)*x[0]**o+(-0.553980493526)*x[0]+(0.359837665019)*x[1]**o+(0.811107941179)*x[1]+(-0.441929650912)*x[2]**o+(-0.768618378636)*x[2]
        arg[2,1,2,0]=(0.741376258816)*x[0]**o+(-0.238909756717)*x[0]+(0.204631082449)*x[1]**o+(0.970501662949)*x[1]+(-0.615638092239)*x[2]**o+(-0.575948750844)*x[2]
        arg[2,1,2,1]=(0.774050933112)*x[0]**o+(-0.872234076721)*x[0]+(-0.0936189240061)*x[1]**o+(-0.773608756961)*x[1]+(-0.119171907808)*x[2]**o+(-0.256124796845)*x[2]
        arg[2,2,0,0]=(0.459481754425)*x[0]**o+(0.478364929493)*x[0]+(0.460199733287)*x[1]**o+(0.66880294659)*x[1]+(-0.859525058389)*x[2]**o+(0.77960284414)*x[2]
        arg[2,2,0,1]=(0.690557835207)*x[0]**o+(-0.812305038592)*x[0]+(0.837924124791)*x[1]**o+(-0.785979206364)*x[1]+(-0.295553949819)*x[2]**o+(0.575096295897)*x[2]
        arg[2,2,1,0]=(0.443193683205)*x[0]**o+(0.0355033222253)*x[0]+(0.497641002766)*x[1]**o+(-0.636955142451)*x[1]+(-0.265324357454)*x[2]**o+(0.734148569078)*x[2]
        arg[2,2,1,1]=(-0.088707242599)*x[0]**o+(0.513403518669)*x[0]+(0.337741737651)*x[1]**o+(-0.174843924478)*x[1]+(0.687291255986)*x[2]**o+(0.473850502258)*x[2]
        arg[2,2,2,0]=(0.5305624642)*x[0]**o+(-0.972997835953)*x[0]+(0.00197107287888)*x[1]**o+(-0.75490872634)*x[1]+(-0.0569624569435)*x[2]**o+(-0.0780532868631)*x[2]
        arg[2,2,2,1]=(-0.744584471866)*x[0]**o+(0.915494909272)*x[0]+(0.239235050038)*x[1]**o+(-0.192466753187)*x[1]+(0.933556550824)*x[2]**o+(-0.786420251487)*x[2]
        arg[2,3,0,0]=(0.472114530627)*x[0]**o+(0.0693211807557)*x[0]+(-0.274273425713)*x[1]**o+(0.136242554894)*x[1]+(-0.329240675785)*x[2]**o+(0.0212539247343)*x[2]
        arg[2,3,0,1]=(-0.780426108099)*x[0]**o+(-0.632382551025)*x[0]+(0.554926949239)*x[1]**o+(0.264204009059)*x[1]+(-0.868007525824)*x[2]**o+(-0.937631120201)*x[2]
        arg[2,3,1,0]=(-0.826692080947)*x[0]**o+(0.915241858538)*x[0]+(-0.49986160859)*x[1]**o+(0.921686464658)*x[1]+(0.71996666304)*x[2]**o+(0.297609501096)*x[2]
        arg[2,3,1,1]=(0.991896994614)*x[0]**o+(-0.543310008103)*x[0]+(0.0791899631111)*x[1]**o+(-0.914088200145)*x[1]+(-0.496337624951)*x[2]**o+(0.708950898614)*x[2]
        arg[2,3,2,0]=(-0.235116966432)*x[0]**o+(-0.492848730207)*x[0]+(0.428728646789)*x[1]**o+(-0.410747634691)*x[1]+(0.705059815458)*x[2]**o+(-0.185604633831)*x[2]
        arg[2,3,2,1]=(0.838829037117)*x[0]**o+(0.670561147563)*x[0]+(0.621407965075)*x[1]**o+(-0.10624946305)*x[1]+(0.893703402687)*x[2]**o+(0.985504995821)*x[2]
        arg[2,4,0,0]=(0.635072504682)*x[0]**o+(-0.681962427555)*x[0]+(-0.00774518011081)*x[1]**o+(-0.264996228668)*x[1]+(0.690668403198)*x[2]**o+(0.883573761062)*x[2]
        arg[2,4,0,1]=(0.545632268651)*x[0]**o+(0.439668519394)*x[0]+(-0.482479808491)*x[1]**o+(0.752028962391)*x[1]+(0.0596773623939)*x[2]**o+(-0.369987181127)*x[2]
        arg[2,4,1,0]=(0.659005322796)*x[0]**o+(0.503940833428)*x[0]+(0.643633321586)*x[1]**o+(0.927082323252)*x[1]+(0.354072146184)*x[2]**o+(-0.638737560608)*x[2]
        arg[2,4,1,1]=(-0.443548860913)*x[0]**o+(0.707607048458)*x[0]+(-0.955225746402)*x[1]**o+(0.515995563585)*x[1]+(-0.288406637165)*x[2]**o+(-0.760864937318)*x[2]
        arg[2,4,2,0]=(0.898146580355)*x[0]**o+(-0.590650787106)*x[0]+(0.036095789753)*x[1]**o+(0.326040260581)*x[1]+(0.680069361306)*x[2]**o+(0.333839099341)*x[2]
        arg[2,4,2,1]=(-0.151364260271)*x[0]**o+(-0.675198183832)*x[0]+(-0.183626298978)*x[1]**o+(-0.182565217146)*x[1]+(0.0345632129814)*x[2]**o+(-0.581534100763)*x[2]
        arg[3,0,0,0]=(-0.942351703246)*x[0]**o+(-0.699567649453)*x[0]+(-0.832068665702)*x[1]**o+(0.163439416504)*x[1]+(0.742500339519)*x[2]**o+(-0.492882887489)*x[2]
        arg[3,0,0,1]=(0.990078296484)*x[0]**o+(-0.999837336587)*x[0]+(0.197589456459)*x[1]**o+(0.431403059019)*x[1]+(0.859864015871)*x[2]**o+(-0.267474829282)*x[2]
        arg[3,0,1,0]=(-0.776773660723)*x[0]**o+(-0.189048370486)*x[0]+(0.926953856802)*x[1]**o+(-0.72314744652)*x[1]+(0.330064363809)*x[2]**o+(0.188533334359)*x[2]
        arg[3,0,1,1]=(-0.854830058762)*x[0]**o+(0.832073883276)*x[0]+(-0.425691165124)*x[1]**o+(0.162351500316)*x[1]+(0.370604864512)*x[2]**o+(-0.581907071484)*x[2]
        arg[3,0,2,0]=(-0.316716909548)*x[0]**o+(0.335417284383)*x[0]+(-0.783061160352)*x[1]**o+(-0.253972519239)*x[1]+(-0.826432933721)*x[2]**o+(0.267368378363)*x[2]
        arg[3,0,2,1]=(-0.765899705462)*x[0]**o+(-0.0993875861476)*x[0]+(-0.211785004352)*x[1]**o+(-0.635330151497)*x[1]+(0.414575412599)*x[2]**o+(0.586882274623)*x[2]
        arg[3,1,0,0]=(0.987776377237)*x[0]**o+(0.796660669191)*x[0]+(0.580724891862)*x[1]**o+(-0.24149253995)*x[1]+(-0.633421016909)*x[2]**o+(-0.220110913711)*x[2]
        arg[3,1,0,1]=(-0.182240918081)*x[0]**o+(-0.290526050572)*x[0]+(0.00283230959976)*x[1]**o+(-0.933282166805)*x[1]+(-0.857893235883)*x[2]**o+(0.551414074127)*x[2]
        arg[3,1,1,0]=(-0.854435997816)*x[0]**o+(0.706455144263)*x[0]+(0.406237965254)*x[1]**o+(0.192071142415)*x[1]+(0.357966796834)*x[2]**o+(0.887011087987)*x[2]
        arg[3,1,1,1]=(0.832013396996)*x[0]**o+(0.696133072321)*x[0]+(0.682726968712)*x[1]**o+(0.613316240514)*x[1]+(0.12600339887)*x[2]**o+(-0.620088231382)*x[2]
        arg[3,1,2,0]=(-0.178681615221)*x[0]**o+(-0.952835986168)*x[0]+(0.299692693746)*x[1]**o+(0.892451908416)*x[1]+(0.202403273634)*x[2]**o+(0.677212897943)*x[2]
        arg[3,1,2,1]=(0.0592594053342)*x[0]**o+(0.56391153959)*x[0]+(-0.874866627667)*x[1]**o+(0.432102900769)*x[1]+(-0.49073126285)*x[2]**o+(-0.010495860028)*x[2]
        arg[3,2,0,0]=(0.548006445814)*x[0]**o+(0.667313392202)*x[0]+(0.397805355499)*x[1]**o+(-0.788100267808)*x[1]+(0.869234055258)*x[2]**o+(-0.830241520401)*x[2]
        arg[3,2,0,1]=(0.72586139216)*x[0]**o+(-0.931878957878)*x[0]+(-0.94119555858)*x[1]**o+(-0.133544806169)*x[1]+(0.754263323382)*x[2]**o+(-0.446604740157)*x[2]
        arg[3,2,1,0]=(-0.28135830707)*x[0]**o+(-0.478744058518)*x[0]+(0.88219033731)*x[1]**o+(0.460587286945)*x[1]+(-0.849692146096)*x[2]**o+(0.0927701345107)*x[2]
        arg[3,2,1,1]=(-0.681859641611)*x[0]**o+(-0.670164293088)*x[0]+(0.796366374435)*x[1]**o+(0.837230317054)*x[1]+(-0.413283915904)*x[2]**o+(0.779624227187)*x[2]
        arg[3,2,2,0]=(-0.803242771255)*x[0]**o+(-0.0687616271776)*x[0]+(-0.948856027999)*x[1]**o+(0.762568790931)*x[1]+(0.937588292604)*x[2]**o+(-0.0224232611283)*x[2]
        arg[3,2,2,1]=(-0.215752944937)*x[0]**o+(-0.807114296075)*x[0]+(0.970625375964)*x[1]**o+(0.68578124692)*x[1]+(-0.165007047419)*x[2]**o+(0.942718439981)*x[2]
        arg[3,3,0,0]=(0.152001104716)*x[0]**o+(-0.657242342357)*x[0]+(0.559608661705)*x[1]**o+(0.0764403273251)*x[1]+(0.633414754695)*x[2]**o+(0.430777868344)*x[2]
        arg[3,3,0,1]=(-0.898109497759)*x[0]**o+(0.736574249406)*x[0]+(-0.569584184574)*x[1]**o+(-0.145501395506)*x[1]+(-0.412424116475)*x[2]**o+(-0.416350923096)*x[2]
        arg[3,3,1,0]=(0.602022795614)*x[0]**o+(-0.848525569825)*x[0]+(-0.713426287505)*x[1]**o+(-0.829278452687)*x[1]+(-0.491257659316)*x[2]**o+(0.90219908999)*x[2]
        arg[3,3,1,1]=(-0.769110216066)*x[0]**o+(-0.789464037338)*x[0]+(-0.363686172584)*x[1]**o+(-0.86190787068)*x[1]+(0.275438119232)*x[2]**o+(0.285209834248)*x[2]
        arg[3,3,2,0]=(0.351662405853)*x[0]**o+(0.8403302307)*x[0]+(-0.449869332519)*x[1]**o+(0.936043614838)*x[1]+(-0.656908777335)*x[2]**o+(-0.231923872531)*x[2]
        arg[3,3,2,1]=(0.873848945653)*x[0]**o+(0.551429883959)*x[0]+(-0.577942548701)*x[1]**o+(-0.254317163049)*x[1]+(0.64878089878)*x[2]**o+(0.7135339515)*x[2]
        arg[3,4,0,0]=(0.891435610494)*x[0]**o+(0.598893704498)*x[0]+(0.24682094756)*x[1]**o+(0.0895612710338)*x[1]+(-0.513686244953)*x[2]**o+(0.307205505502)*x[2]
        arg[3,4,0,1]=(-0.577752944196)*x[0]**o+(-0.458383751623)*x[0]+(-0.913956988143)*x[1]**o+(0.470857602235)*x[1]+(0.631726850665)*x[2]**o+(-0.112962289904)*x[2]
        arg[3,4,1,0]=(-0.65648177489)*x[0]**o+(-0.843530972747)*x[0]+(-0.0332153723809)*x[1]**o+(0.230484263355)*x[1]+(-0.865687508592)*x[2]**o+(0.122751735082)*x[2]
        arg[3,4,1,1]=(-0.353502640734)*x[0]**o+(-0.816020900282)*x[0]+(0.943688796868)*x[1]**o+(-0.526707881958)*x[1]+(-0.790258298117)*x[2]**o+(-0.748706809042)*x[2]
        arg[3,4,2,0]=(-0.596345912175)*x[0]**o+(0.872771121837)*x[0]+(0.100107395575)*x[1]**o+(0.950482249217)*x[1]+(0.0487634582783)*x[2]**o+(-0.0285218097887)*x[2]
        arg[3,4,2,1]=(-0.827443553933)*x[0]**o+(-0.922564800149)*x[0]+(-0.810354385287)*x[1]**o+(0.630079217293)*x[1]+(-0.316280559723)*x[2]**o+(0.977830435871)*x[2]
        ref[0,0,0,0]=(0.310272924207)*x_ref[0]**o+(-0.664497806351)*x_ref[0]+(0.741942132609)*x_ref[1]**o+(0.636051298779)*x_ref[1]+(0.76627159146)*x_ref[2]**o+(0.492960732588)*x_ref[2]
        ref[0,0,0,1]=(-0.513740801773)*x_ref[0]**o+(-0.633330750909)*x_ref[0]+(0.177934618983)*x_ref[1]**o+(-0.323745723166)*x_ref[1]+(0.694167180904)*x_ref[2]**o+(0.441239251435)*x_ref[2]
        ref[0,0,1,0]=(0.417477695848)*x_ref[0]**o+(-0.923415976002)*x_ref[0]+(0.0765555276662)*x_ref[1]**o+(-0.277290890834)*x_ref[1]+(0.0177191326614)*x_ref[2]**o+(-0.0842520100387)*x_ref[2]
        ref[0,0,1,1]=(-0.761916249894)*x_ref[0]**o+(0.169673139552)*x_ref[0]+(0.0306290828795)*x_ref[1]**o+(0.842234592613)*x_ref[1]+(-0.350342692669)*x_ref[2]**o+(0.321292270658)*x_ref[2]
        ref[0,0,2,0]=(-0.707494137624)*x_ref[0]**o+(0.346873490395)*x_ref[0]+(-0.68281925138)*x_ref[1]**o+(-0.830538083274)*x_ref[1]+(0.329372141613)*x_ref[2]**o+(-0.121366157163)*x_ref[2]
        ref[0,0,2,1]=(-0.708018186507)*x_ref[0]**o+(-0.778051239314)*x_ref[0]+(-0.43106291563)*x_ref[1]**o+(-0.0115914251143)*x_ref[1]+(0.487334215892)*x_ref[2]**o+(-0.289232284658)*x_ref[2]
        ref[0,1,0,0]=(0.530614704199)*x_ref[0]**o+(0.518622028317)*x_ref[0]+(-0.0143129589551)*x_ref[1]**o+(0.808469352257)*x_ref[1]+(-0.156787373225)*x_ref[2]**o+(0.668887813143)*x_ref[2]
        ref[0,1,0,1]=(0.953223233706)*x_ref[0]**o+(-0.164323838353)*x_ref[0]+(0.336717320858)*x_ref[1]**o+(-0.520230488172)*x_ref[1]+(0.689960738002)*x_ref[2]**o+(0.487630947578)*x_ref[2]
        ref[0,1,1,0]=(-0.418095373429)*x_ref[0]**o+(-0.941817844939)*x_ref[0]+(0.88105115607)*x_ref[1]**o+(-0.820502501439)*x_ref[1]+(-0.693858267667)*x_ref[2]**o+(0.530567713127)*x_ref[2]
        ref[0,1,1,1]=(-0.716861320218)*x_ref[0]**o+(-0.736619859467)*x_ref[0]+(-0.885478605601)*x_ref[1]**o+(-0.129375578967)*x_ref[1]+(-0.0181433907318)*x_ref[2]**o+(0.373255537959)*x_ref[2]
        ref[0,1,2,0]=(-0.174742583167)*x_ref[0]**o+(-0.0887967480112)*x_ref[0]+(-0.291168698662)*x_ref[1]**o+(-0.188608506769)*x_ref[1]+(0.22479613454)*x_ref[2]**o+(0.34173928582)*x_ref[2]
        ref[0,1,2,1]=(0.117208929504)*x_ref[0]**o+(-0.420969282079)*x_ref[0]+(0.585298109836)*x_ref[1]**o+(0.771355162745)*x_ref[1]+(-0.51374493747)*x_ref[2]**o+(0.785016548855)*x_ref[2]
        ref[0,2,0,0]=(0.737394887886)*x_ref[0]**o+(0.966080271342)*x_ref[0]+(-0.373178384492)*x_ref[1]**o+(-0.845505386142)*x_ref[1]+(-0.274654519371)*x_ref[2]**o+(0.709356557887)*x_ref[2]
        ref[0,2,0,1]=(0.0365864090636)*x_ref[0]**o+(-0.393382479787)*x_ref[0]+(0.018273360977)*x_ref[1]**o+(-0.509201909328)*x_ref[1]+(-0.777852392534)*x_ref[2]**o+(0.0417280247507)*x_ref[2]
        ref[0,2,1,0]=(0.637829302751)*x_ref[0]**o+(-0.767824790186)*x_ref[0]+(0.843382009511)*x_ref[1]**o+(-0.599633234403)*x_ref[1]+(0.14283138941)*x_ref[2]**o+(-0.354404633991)*x_ref[2]
        ref[0,2,1,1]=(-0.465258613599)*x_ref[0]**o+(0.09341154383)*x_ref[0]+(-0.629114782988)*x_ref[1]**o+(0.91608655383)*x_ref[1]+(-0.692901135848)*x_ref[2]**o+(-0.707225310667)*x_ref[2]
        ref[0,2,2,0]=(0.9935958759)*x_ref[0]**o+(-0.385008742466)*x_ref[0]+(0.153885356854)*x_ref[1]**o+(0.146270680656)*x_ref[1]+(0.305838410343)*x_ref[2]**o+(0.141193552717)*x_ref[2]
        ref[0,2,2,1]=(-0.487538023894)*x_ref[0]**o+(-0.814283442647)*x_ref[0]+(0.166008520199)*x_ref[1]**o+(0.313655774416)*x_ref[1]+(0.282308782973)*x_ref[2]**o+(0.309176062484)*x_ref[2]
        ref[0,3,0,0]=(0.00203507798913)*x_ref[0]**o+(-0.19902542604)*x_ref[0]+(0.88110918624)*x_ref[1]**o+(0.223549219288)*x_ref[1]+(0.135758952369)*x_ref[2]**o+(0.59116783639)*x_ref[2]
        ref[0,3,0,1]=(-0.185063700767)*x_ref[0]**o+(0.833862155173)*x_ref[0]+(-0.0939117509645)*x_ref[1]**o+(0.288003692231)*x_ref[1]+(0.580196157839)*x_ref[2]**o+(0.299320177754)*x_ref[2]
        ref[0,3,1,0]=(-0.039100718826)*x_ref[0]**o+(0.423179869395)*x_ref[0]+(-0.013669578975)*x_ref[1]**o+(-0.355053343085)*x_ref[1]+(0.89285158975)*x_ref[2]**o+(0.392984367877)*x_ref[2]
        ref[0,3,1,1]=(0.0988989416429)*x_ref[0]**o+(-0.446243235174)*x_ref[0]+(-0.691253313617)*x_ref[1]**o+(0.338823693086)*x_ref[1]+(0.707054126334)*x_ref[2]**o+(0.601121388278)*x_ref[2]
        ref[0,3,2,0]=(-0.191366770484)*x_ref[0]**o+(0.961791275903)*x_ref[0]+(-0.103547475035)*x_ref[1]**o+(-0.745236510079)*x_ref[1]+(-0.157506981712)*x_ref[2]**o+(0.986229506787)*x_ref[2]
        ref[0,3,2,1]=(-0.886091883647)*x_ref[0]**o+(0.6979804248)*x_ref[0]+(-0.366575419428)*x_ref[1]**o+(-0.958601153259)*x_ref[1]+(-0.0119724511494)*x_ref[2]**o+(-0.76101132076)*x_ref[2]
        ref[0,4,0,0]=(0.923290554434)*x_ref[0]**o+(0.411359492198)*x_ref[0]+(-0.977999562875)*x_ref[1]**o+(0.352085161527)*x_ref[1]+(0.955745235349)*x_ref[2]**o+(-0.978885609257)*x_ref[2]
        ref[0,4,0,1]=(0.226973479105)*x_ref[0]**o+(-0.423881897251)*x_ref[0]+(0.889101969481)*x_ref[1]**o+(-0.994236389028)*x_ref[1]+(0.0159169033361)*x_ref[2]**o+(-0.559935154203)*x_ref[2]
        ref[0,4,1,0]=(-0.225427291797)*x_ref[0]**o+(-0.557334621848)*x_ref[0]+(-0.486252586142)*x_ref[1]**o+(0.0497120905216)*x_ref[1]+(0.84165736691)*x_ref[2]**o+(-0.185010156797)*x_ref[2]
        ref[0,4,1,1]=(0.762694602609)*x_ref[0]**o+(-0.447367170727)*x_ref[0]+(-0.61103181818)*x_ref[1]**o+(0.689761057656)*x_ref[1]+(-0.631855120122)*x_ref[2]**o+(0.33274655602)*x_ref[2]
        ref[0,4,2,0]=(-0.198884868914)*x_ref[0]**o+(0.716371959278)*x_ref[0]+(0.260871768919)*x_ref[1]**o+(0.622241740082)*x_ref[1]+(0.774103524035)*x_ref[2]**o+(0.186132703676)*x_ref[2]
        ref[0,4,2,1]=(-0.309450645317)*x_ref[0]**o+(0.204754571336)*x_ref[0]+(0.483402810171)*x_ref[1]**o+(0.404620744188)*x_ref[1]+(-0.699137246959)*x_ref[2]**o+(0.467428706196)*x_ref[2]
        ref[1,0,0,0]=(0.642289506448)*x_ref[0]**o+(0.383990304394)*x_ref[0]+(0.556940973318)*x_ref[1]**o+(0.135219729041)*x_ref[1]+(0.577195836358)*x_ref[2]**o+(-0.338924419674)*x_ref[2]
        ref[1,0,0,1]=(-0.920873911295)*x_ref[0]**o+(-0.330610192304)*x_ref[0]+(-0.117226803856)*x_ref[1]**o+(0.172482867684)*x_ref[1]+(0.909140902709)*x_ref[2]**o+(-0.590494167913)*x_ref[2]
        ref[1,0,1,0]=(0.210161538861)*x_ref[0]**o+(-0.83860241761)*x_ref[0]+(0.0257180382661)*x_ref[1]**o+(0.877633854348)*x_ref[1]+(0.26045311754)*x_ref[2]**o+(-0.971349902714)*x_ref[2]
        ref[1,0,1,1]=(0.70736843125)*x_ref[0]**o+(-0.765121985431)*x_ref[0]+(-0.196525225943)*x_ref[1]**o+(-0.175928729346)*x_ref[1]+(0.945523028748)*x_ref[2]**o+(0.877830322823)*x_ref[2]
        ref[1,0,2,0]=(-0.963513016065)*x_ref[0]**o+(0.0820537856351)*x_ref[0]+(0.0122070196118)*x_ref[1]**o+(-0.154147449592)*x_ref[1]+(0.323844633268)*x_ref[2]**o+(-0.124897698214)*x_ref[2]
        ref[1,0,2,1]=(0.656267566405)*x_ref[0]**o+(0.0174786620768)*x_ref[0]+(0.763902435618)*x_ref[1]**o+(-0.986021506043)*x_ref[1]+(0.632625555994)*x_ref[2]**o+(-0.110657743523)*x_ref[2]
        ref[1,1,0,0]=(0.00354144688732)*x_ref[0]**o+(-0.813515808171)*x_ref[0]+(-0.100717717446)*x_ref[1]**o+(0.0543708163629)*x_ref[1]+(0.785927880762)*x_ref[2]**o+(0.0109166652227)*x_ref[2]
        ref[1,1,0,1]=(-0.239705688631)*x_ref[0]**o+(0.402915642509)*x_ref[0]+(0.463371068163)*x_ref[1]**o+(-0.507302462928)*x_ref[1]+(-0.0789844128295)*x_ref[2]**o+(0.669144891499)*x_ref[2]
        ref[1,1,1,0]=(-0.491891367001)*x_ref[0]**o+(-0.524377636928)*x_ref[0]+(0.903091863917)*x_ref[1]**o+(-0.554368807183)*x_ref[1]+(0.332899195339)*x_ref[2]**o+(-0.645216147049)*x_ref[2]
        ref[1,1,1,1]=(-0.915843941243)*x_ref[0]**o+(-0.233175654748)*x_ref[0]+(0.467730425037)*x_ref[1]**o+(-0.765957936683)*x_ref[1]+(0.726678877276)*x_ref[2]**o+(0.770533947942)*x_ref[2]
        ref[1,1,2,0]=(0.678071852032)*x_ref[0]**o+(0.266537844237)*x_ref[0]+(0.324335300151)*x_ref[1]**o+(-0.286713971373)*x_ref[1]+(-0.702384414014)*x_ref[2]**o+(0.887198276519)*x_ref[2]
        ref[1,1,2,1]=(0.0466444959648)*x_ref[0]**o+(0.689162912677)*x_ref[0]+(-0.35033923676)*x_ref[1]**o+(-0.0536380766876)*x_ref[1]+(-0.511833442517)*x_ref[2]**o+(0.125566299829)*x_ref[2]
        ref[1,2,0,0]=(-0.803170431622)*x_ref[0]**o+(0.476211169752)*x_ref[0]+(-0.170510267953)*x_ref[1]**o+(-0.765945735317)*x_ref[1]+(-0.169966329291)*x_ref[2]**o+(-0.988525918724)*x_ref[2]
        ref[1,2,0,1]=(0.289618751212)*x_ref[0]**o+(0.294379040976)*x_ref[0]+(-0.922478416838)*x_ref[1]**o+(0.210546495537)*x_ref[1]+(-0.87266381496)*x_ref[2]**o+(-0.865571669577)*x_ref[2]
        ref[1,2,1,0]=(-0.292811054005)*x_ref[0]**o+(0.651280833746)*x_ref[0]+(-0.625004203736)*x_ref[1]**o+(0.372078169172)*x_ref[1]+(-0.0539099055169)*x_ref[2]**o+(0.0911869270008)*x_ref[2]
        ref[1,2,1,1]=(0.496217698294)*x_ref[0]**o+(-0.986021787133)*x_ref[0]+(0.246417059532)*x_ref[1]**o+(0.610226802401)*x_ref[1]+(0.991534665298)*x_ref[2]**o+(-0.507441185982)*x_ref[2]
        ref[1,2,2,0]=(-0.839902565947)*x_ref[0]**o+(-0.279607147278)*x_ref[0]+(0.278212307963)*x_ref[1]**o+(0.609793640036)*x_ref[1]+(0.940471913774)*x_ref[2]**o+(0.21860828705)*x_ref[2]
        ref[1,2,2,1]=(0.789578112278)*x_ref[0]**o+(-0.0226697818608)*x_ref[0]+(0.216321340291)*x_ref[1]**o+(-0.701818463454)*x_ref[1]+(-0.166689030646)*x_ref[2]**o+(0.966131542292)*x_ref[2]
        ref[1,3,0,0]=(0.149337042551)*x_ref[0]**o+(-0.785050579674)*x_ref[0]+(0.530219924721)*x_ref[1]**o+(-0.971927239398)*x_ref[1]+(0.950917908603)*x_ref[2]**o+(0.132668077447)*x_ref[2]
        ref[1,3,0,1]=(0.983081559724)*x_ref[0]**o+(-0.358694498455)*x_ref[0]+(0.181080099281)*x_ref[1]**o+(-0.699756291643)*x_ref[1]+(-0.744335953304)*x_ref[2]**o+(0.00262043221382)*x_ref[2]
        ref[1,3,1,0]=(0.31776032045)*x_ref[0]**o+(-0.0726184220029)*x_ref[0]+(-0.759883735971)*x_ref[1]**o+(-0.6389710873)*x_ref[1]+(-0.607858499502)*x_ref[2]**o+(-0.397169895784)*x_ref[2]
        ref[1,3,1,1]=(-0.950468389728)*x_ref[0]**o+(0.831121915521)*x_ref[0]+(0.552857011083)*x_ref[1]**o+(0.848599672594)*x_ref[1]+(0.432183564231)*x_ref[2]**o+(0.57537721362)*x_ref[2]
        ref[1,3,2,0]=(0.736080143601)*x_ref[0]**o+(-0.804140307747)*x_ref[0]+(-0.163864867884)*x_ref[1]**o+(0.950874042315)*x_ref[1]+(-0.99676139412)*x_ref[2]**o+(0.332320294246)*x_ref[2]
        ref[1,3,2,1]=(0.119082582041)*x_ref[0]**o+(-0.969740778852)*x_ref[0]+(0.828432583532)*x_ref[1]**o+(0.190324050395)*x_ref[1]+(-0.42639142674)*x_ref[2]**o+(0.315519578192)*x_ref[2]
        ref[1,4,0,0]=(-0.798465141775)*x_ref[0]**o+(-0.717282977908)*x_ref[0]+(0.935655148713)*x_ref[1]**o+(-0.655752964484)*x_ref[1]+(0.159990149099)*x_ref[2]**o+(-0.537115062687)*x_ref[2]
        ref[1,4,0,1]=(0.445151775007)*x_ref[0]**o+(-0.867956302051)*x_ref[0]+(-0.809856879825)*x_ref[1]**o+(-0.326897750981)*x_ref[1]+(0.756352311987)*x_ref[2]**o+(0.577585513459)*x_ref[2]
        ref[1,4,1,0]=(0.294243347591)*x_ref[0]**o+(0.774741656931)*x_ref[0]+(-0.0799985737984)*x_ref[1]**o+(0.179789107412)*x_ref[1]+(-0.612198025189)*x_ref[2]**o+(-0.07086009611)*x_ref[2]
        ref[1,4,1,1]=(0.0692586287045)*x_ref[0]**o+(-0.671956300122)*x_ref[0]+(-0.948667104683)*x_ref[1]**o+(0.68656246801)*x_ref[1]+(-0.352975297696)*x_ref[2]**o+(0.277952113716)*x_ref[2]
        ref[1,4,2,0]=(-0.491316095517)*x_ref[0]**o+(0.725899087239)*x_ref[0]+(0.634113963431)*x_ref[1]**o+(0.235694393453)*x_ref[1]+(-0.15600720713)*x_ref[2]**o+(0.954814976539)*x_ref[2]
        ref[1,4,2,1]=(0.120206996339)*x_ref[0]**o+(-0.86151411823)*x_ref[0]+(-0.137093650097)*x_ref[1]**o+(0.633402978246)*x_ref[1]+(-0.42515364728)*x_ref[2]**o+(-0.987634210617)*x_ref[2]
        ref[2,0,0,0]=(0.539463971222)*x_ref[0]**o+(-0.03435650579)*x_ref[0]+(-0.884395295544)*x_ref[1]**o+(0.215454879022)*x_ref[1]+(0.842377328246)*x_ref[2]**o+(0.370920136871)*x_ref[2]
        ref[2,0,0,1]=(-0.411505463039)*x_ref[0]**o+(0.446494039439)*x_ref[0]+(0.0313771969556)*x_ref[1]**o+(0.602893223061)*x_ref[1]+(-0.30383490021)*x_ref[2]**o+(-0.788028756758)*x_ref[2]
        ref[2,0,1,0]=(0.489009615999)*x_ref[0]**o+(-0.95508248016)*x_ref[0]+(-0.21489475022)*x_ref[1]**o+(-0.857263478431)*x_ref[1]+(-0.123703409045)*x_ref[2]**o+(0.663389562431)*x_ref[2]
        ref[2,0,1,1]=(-0.0224763310929)*x_ref[0]**o+(-0.315886249237)*x_ref[0]+(-0.318201849695)*x_ref[1]**o+(0.423732682462)*x_ref[1]+(0.396065636381)*x_ref[2]**o+(0.0816717878865)*x_ref[2]
        ref[2,0,2,0]=(-0.191432618224)*x_ref[0]**o+(0.688389390795)*x_ref[0]+(-0.382948112159)*x_ref[1]**o+(0.562905173818)*x_ref[1]+(-0.118396714029)*x_ref[2]**o+(0.234478642577)*x_ref[2]
        ref[2,0,2,1]=(-0.348008173171)*x_ref[0]**o+(0.191994166636)*x_ref[0]+(0.450556057147)*x_ref[1]**o+(-0.710821693343)*x_ref[1]+(-0.245119333857)*x_ref[2]**o+(0.262869525239)*x_ref[2]
        ref[2,1,0,0]=(-0.160444506203)*x_ref[0]**o+(-0.816142837432)*x_ref[0]+(-0.619722480492)*x_ref[1]**o+(-0.989077039675)*x_ref[1]+(-0.0282349103977)*x_ref[2]**o+(-0.629849819691)*x_ref[2]
        ref[2,1,0,1]=(0.379513765297)*x_ref[0]**o+(0.934652105586)*x_ref[0]+(-0.632636427868)*x_ref[1]**o+(-0.0492550496553)*x_ref[1]+(0.137733572307)*x_ref[2]**o+(0.779782637942)*x_ref[2]
        ref[2,1,1,0]=(0.469499320417)*x_ref[0]**o+(0.570880924125)*x_ref[0]+(-0.0329833251419)*x_ref[1]**o+(-0.0388751034604)*x_ref[1]+(-0.878520094389)*x_ref[2]**o+(0.992656649048)*x_ref[2]
        ref[2,1,1,1]=(0.592005181995)*x_ref[0]**o+(-0.553980493526)*x_ref[0]+(0.359837665019)*x_ref[1]**o+(0.811107941179)*x_ref[1]+(-0.441929650912)*x_ref[2]**o+(-0.768618378636)*x_ref[2]
        ref[2,1,2,0]=(0.741376258816)*x_ref[0]**o+(-0.238909756717)*x_ref[0]+(0.204631082449)*x_ref[1]**o+(0.970501662949)*x_ref[1]+(-0.615638092239)*x_ref[2]**o+(-0.575948750844)*x_ref[2]
        ref[2,1,2,1]=(0.774050933112)*x_ref[0]**o+(-0.872234076721)*x_ref[0]+(-0.0936189240061)*x_ref[1]**o+(-0.773608756961)*x_ref[1]+(-0.119171907808)*x_ref[2]**o+(-0.256124796845)*x_ref[2]
        ref[2,2,0,0]=(0.459481754425)*x_ref[0]**o+(0.478364929493)*x_ref[0]+(0.460199733287)*x_ref[1]**o+(0.66880294659)*x_ref[1]+(-0.859525058389)*x_ref[2]**o+(0.77960284414)*x_ref[2]
        ref[2,2,0,1]=(0.690557835207)*x_ref[0]**o+(-0.812305038592)*x_ref[0]+(0.837924124791)*x_ref[1]**o+(-0.785979206364)*x_ref[1]+(-0.295553949819)*x_ref[2]**o+(0.575096295897)*x_ref[2]
        ref[2,2,1,0]=(0.443193683205)*x_ref[0]**o+(0.0355033222253)*x_ref[0]+(0.497641002766)*x_ref[1]**o+(-0.636955142451)*x_ref[1]+(-0.265324357454)*x_ref[2]**o+(0.734148569078)*x_ref[2]
        ref[2,2,1,1]=(-0.088707242599)*x_ref[0]**o+(0.513403518669)*x_ref[0]+(0.337741737651)*x_ref[1]**o+(-0.174843924478)*x_ref[1]+(0.687291255986)*x_ref[2]**o+(0.473850502258)*x_ref[2]
        ref[2,2,2,0]=(0.5305624642)*x_ref[0]**o+(-0.972997835953)*x_ref[0]+(0.00197107287888)*x_ref[1]**o+(-0.75490872634)*x_ref[1]+(-0.0569624569435)*x_ref[2]**o+(-0.0780532868631)*x_ref[2]
        ref[2,2,2,1]=(-0.744584471866)*x_ref[0]**o+(0.915494909272)*x_ref[0]+(0.239235050038)*x_ref[1]**o+(-0.192466753187)*x_ref[1]+(0.933556550824)*x_ref[2]**o+(-0.786420251487)*x_ref[2]
        ref[2,3,0,0]=(0.472114530627)*x_ref[0]**o+(0.0693211807557)*x_ref[0]+(-0.274273425713)*x_ref[1]**o+(0.136242554894)*x_ref[1]+(-0.329240675785)*x_ref[2]**o+(0.0212539247343)*x_ref[2]
        ref[2,3,0,1]=(-0.780426108099)*x_ref[0]**o+(-0.632382551025)*x_ref[0]+(0.554926949239)*x_ref[1]**o+(0.264204009059)*x_ref[1]+(-0.868007525824)*x_ref[2]**o+(-0.937631120201)*x_ref[2]
        ref[2,3,1,0]=(-0.826692080947)*x_ref[0]**o+(0.915241858538)*x_ref[0]+(-0.49986160859)*x_ref[1]**o+(0.921686464658)*x_ref[1]+(0.71996666304)*x_ref[2]**o+(0.297609501096)*x_ref[2]
        ref[2,3,1,1]=(0.991896994614)*x_ref[0]**o+(-0.543310008103)*x_ref[0]+(0.0791899631111)*x_ref[1]**o+(-0.914088200145)*x_ref[1]+(-0.496337624951)*x_ref[2]**o+(0.708950898614)*x_ref[2]
        ref[2,3,2,0]=(-0.235116966432)*x_ref[0]**o+(-0.492848730207)*x_ref[0]+(0.428728646789)*x_ref[1]**o+(-0.410747634691)*x_ref[1]+(0.705059815458)*x_ref[2]**o+(-0.185604633831)*x_ref[2]
        ref[2,3,2,1]=(0.838829037117)*x_ref[0]**o+(0.670561147563)*x_ref[0]+(0.621407965075)*x_ref[1]**o+(-0.10624946305)*x_ref[1]+(0.893703402687)*x_ref[2]**o+(0.985504995821)*x_ref[2]
        ref[2,4,0,0]=(0.635072504682)*x_ref[0]**o+(-0.681962427555)*x_ref[0]+(-0.00774518011081)*x_ref[1]**o+(-0.264996228668)*x_ref[1]+(0.690668403198)*x_ref[2]**o+(0.883573761062)*x_ref[2]
        ref[2,4,0,1]=(0.545632268651)*x_ref[0]**o+(0.439668519394)*x_ref[0]+(-0.482479808491)*x_ref[1]**o+(0.752028962391)*x_ref[1]+(0.0596773623939)*x_ref[2]**o+(-0.369987181127)*x_ref[2]
        ref[2,4,1,0]=(0.659005322796)*x_ref[0]**o+(0.503940833428)*x_ref[0]+(0.643633321586)*x_ref[1]**o+(0.927082323252)*x_ref[1]+(0.354072146184)*x_ref[2]**o+(-0.638737560608)*x_ref[2]
        ref[2,4,1,1]=(-0.443548860913)*x_ref[0]**o+(0.707607048458)*x_ref[0]+(-0.955225746402)*x_ref[1]**o+(0.515995563585)*x_ref[1]+(-0.288406637165)*x_ref[2]**o+(-0.760864937318)*x_ref[2]
        ref[2,4,2,0]=(0.898146580355)*x_ref[0]**o+(-0.590650787106)*x_ref[0]+(0.036095789753)*x_ref[1]**o+(0.326040260581)*x_ref[1]+(0.680069361306)*x_ref[2]**o+(0.333839099341)*x_ref[2]
        ref[2,4,2,1]=(-0.151364260271)*x_ref[0]**o+(-0.675198183832)*x_ref[0]+(-0.183626298978)*x_ref[1]**o+(-0.182565217146)*x_ref[1]+(0.0345632129814)*x_ref[2]**o+(-0.581534100763)*x_ref[2]
        ref[3,0,0,0]=(-0.942351703246)*x_ref[0]**o+(-0.699567649453)*x_ref[0]+(-0.832068665702)*x_ref[1]**o+(0.163439416504)*x_ref[1]+(0.742500339519)*x_ref[2]**o+(-0.492882887489)*x_ref[2]
        ref[3,0,0,1]=(0.990078296484)*x_ref[0]**o+(-0.999837336587)*x_ref[0]+(0.197589456459)*x_ref[1]**o+(0.431403059019)*x_ref[1]+(0.859864015871)*x_ref[2]**o+(-0.267474829282)*x_ref[2]
        ref[3,0,1,0]=(-0.776773660723)*x_ref[0]**o+(-0.189048370486)*x_ref[0]+(0.926953856802)*x_ref[1]**o+(-0.72314744652)*x_ref[1]+(0.330064363809)*x_ref[2]**o+(0.188533334359)*x_ref[2]
        ref[3,0,1,1]=(-0.854830058762)*x_ref[0]**o+(0.832073883276)*x_ref[0]+(-0.425691165124)*x_ref[1]**o+(0.162351500316)*x_ref[1]+(0.370604864512)*x_ref[2]**o+(-0.581907071484)*x_ref[2]
        ref[3,0,2,0]=(-0.316716909548)*x_ref[0]**o+(0.335417284383)*x_ref[0]+(-0.783061160352)*x_ref[1]**o+(-0.253972519239)*x_ref[1]+(-0.826432933721)*x_ref[2]**o+(0.267368378363)*x_ref[2]
        ref[3,0,2,1]=(-0.765899705462)*x_ref[0]**o+(-0.0993875861476)*x_ref[0]+(-0.211785004352)*x_ref[1]**o+(-0.635330151497)*x_ref[1]+(0.414575412599)*x_ref[2]**o+(0.586882274623)*x_ref[2]
        ref[3,1,0,0]=(0.987776377237)*x_ref[0]**o+(0.796660669191)*x_ref[0]+(0.580724891862)*x_ref[1]**o+(-0.24149253995)*x_ref[1]+(-0.633421016909)*x_ref[2]**o+(-0.220110913711)*x_ref[2]
        ref[3,1,0,1]=(-0.182240918081)*x_ref[0]**o+(-0.290526050572)*x_ref[0]+(0.00283230959976)*x_ref[1]**o+(-0.933282166805)*x_ref[1]+(-0.857893235883)*x_ref[2]**o+(0.551414074127)*x_ref[2]
        ref[3,1,1,0]=(-0.854435997816)*x_ref[0]**o+(0.706455144263)*x_ref[0]+(0.406237965254)*x_ref[1]**o+(0.192071142415)*x_ref[1]+(0.357966796834)*x_ref[2]**o+(0.887011087987)*x_ref[2]
        ref[3,1,1,1]=(0.832013396996)*x_ref[0]**o+(0.696133072321)*x_ref[0]+(0.682726968712)*x_ref[1]**o+(0.613316240514)*x_ref[1]+(0.12600339887)*x_ref[2]**o+(-0.620088231382)*x_ref[2]
        ref[3,1,2,0]=(-0.178681615221)*x_ref[0]**o+(-0.952835986168)*x_ref[0]+(0.299692693746)*x_ref[1]**o+(0.892451908416)*x_ref[1]+(0.202403273634)*x_ref[2]**o+(0.677212897943)*x_ref[2]
        ref[3,1,2,1]=(0.0592594053342)*x_ref[0]**o+(0.56391153959)*x_ref[0]+(-0.874866627667)*x_ref[1]**o+(0.432102900769)*x_ref[1]+(-0.49073126285)*x_ref[2]**o+(-0.010495860028)*x_ref[2]
        ref[3,2,0,0]=(0.548006445814)*x_ref[0]**o+(0.667313392202)*x_ref[0]+(0.397805355499)*x_ref[1]**o+(-0.788100267808)*x_ref[1]+(0.869234055258)*x_ref[2]**o+(-0.830241520401)*x_ref[2]
        ref[3,2,0,1]=(0.72586139216)*x_ref[0]**o+(-0.931878957878)*x_ref[0]+(-0.94119555858)*x_ref[1]**o+(-0.133544806169)*x_ref[1]+(0.754263323382)*x_ref[2]**o+(-0.446604740157)*x_ref[2]
        ref[3,2,1,0]=(-0.28135830707)*x_ref[0]**o+(-0.478744058518)*x_ref[0]+(0.88219033731)*x_ref[1]**o+(0.460587286945)*x_ref[1]+(-0.849692146096)*x_ref[2]**o+(0.0927701345107)*x_ref[2]
        ref[3,2,1,1]=(-0.681859641611)*x_ref[0]**o+(-0.670164293088)*x_ref[0]+(0.796366374435)*x_ref[1]**o+(0.837230317054)*x_ref[1]+(-0.413283915904)*x_ref[2]**o+(0.779624227187)*x_ref[2]
        ref[3,2,2,0]=(-0.803242771255)*x_ref[0]**o+(-0.0687616271776)*x_ref[0]+(-0.948856027999)*x_ref[1]**o+(0.762568790931)*x_ref[1]+(0.937588292604)*x_ref[2]**o+(-0.0224232611283)*x_ref[2]
        ref[3,2,2,1]=(-0.215752944937)*x_ref[0]**o+(-0.807114296075)*x_ref[0]+(0.970625375964)*x_ref[1]**o+(0.68578124692)*x_ref[1]+(-0.165007047419)*x_ref[2]**o+(0.942718439981)*x_ref[2]
        ref[3,3,0,0]=(0.152001104716)*x_ref[0]**o+(-0.657242342357)*x_ref[0]+(0.559608661705)*x_ref[1]**o+(0.0764403273251)*x_ref[1]+(0.633414754695)*x_ref[2]**o+(0.430777868344)*x_ref[2]
        ref[3,3,0,1]=(-0.898109497759)*x_ref[0]**o+(0.736574249406)*x_ref[0]+(-0.569584184574)*x_ref[1]**o+(-0.145501395506)*x_ref[1]+(-0.412424116475)*x_ref[2]**o+(-0.416350923096)*x_ref[2]
        ref[3,3,1,0]=(0.602022795614)*x_ref[0]**o+(-0.848525569825)*x_ref[0]+(-0.713426287505)*x_ref[1]**o+(-0.829278452687)*x_ref[1]+(-0.491257659316)*x_ref[2]**o+(0.90219908999)*x_ref[2]
        ref[3,3,1,1]=(-0.769110216066)*x_ref[0]**o+(-0.789464037338)*x_ref[0]+(-0.363686172584)*x_ref[1]**o+(-0.86190787068)*x_ref[1]+(0.275438119232)*x_ref[2]**o+(0.285209834248)*x_ref[2]
        ref[3,3,2,0]=(0.351662405853)*x_ref[0]**o+(0.8403302307)*x_ref[0]+(-0.449869332519)*x_ref[1]**o+(0.936043614838)*x_ref[1]+(-0.656908777335)*x_ref[2]**o+(-0.231923872531)*x_ref[2]
        ref[3,3,2,1]=(0.873848945653)*x_ref[0]**o+(0.551429883959)*x_ref[0]+(-0.577942548701)*x_ref[1]**o+(-0.254317163049)*x_ref[1]+(0.64878089878)*x_ref[2]**o+(0.7135339515)*x_ref[2]
        ref[3,4,0,0]=(0.891435610494)*x_ref[0]**o+(0.598893704498)*x_ref[0]+(0.24682094756)*x_ref[1]**o+(0.0895612710338)*x_ref[1]+(-0.513686244953)*x_ref[2]**o+(0.307205505502)*x_ref[2]
        ref[3,4,0,1]=(-0.577752944196)*x_ref[0]**o+(-0.458383751623)*x_ref[0]+(-0.913956988143)*x_ref[1]**o+(0.470857602235)*x_ref[1]+(0.631726850665)*x_ref[2]**o+(-0.112962289904)*x_ref[2]
        ref[3,4,1,0]=(-0.65648177489)*x_ref[0]**o+(-0.843530972747)*x_ref[0]+(-0.0332153723809)*x_ref[1]**o+(0.230484263355)*x_ref[1]+(-0.865687508592)*x_ref[2]**o+(0.122751735082)*x_ref[2]
        ref[3,4,1,1]=(-0.353502640734)*x_ref[0]**o+(-0.816020900282)*x_ref[0]+(0.943688796868)*x_ref[1]**o+(-0.526707881958)*x_ref[1]+(-0.790258298117)*x_ref[2]**o+(-0.748706809042)*x_ref[2]
        ref[3,4,2,0]=(-0.596345912175)*x_ref[0]**o+(0.872771121837)*x_ref[0]+(0.100107395575)*x_ref[1]**o+(0.950482249217)*x_ref[1]+(0.0487634582783)*x_ref[2]**o+(-0.0285218097887)*x_ref[2]
        ref[3,4,2,1]=(-0.827443553933)*x_ref[0]**o+(-0.922564800149)*x_ref[0]+(-0.810354385287)*x_ref[1]**o+(0.630079217293)*x_ref[1]+(-0.316280559723)*x_ref[2]**o+(0.977830435871)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.142445811229)*x[0]**o+(-0.170111870439)*x[0]+(-0.571150729608)*x[1]**o+(-0.799596368319)*x[1]
        ref=(0.142445811229)*x_ref[0]**o+(-0.170111870439)*x_ref[0]+(-0.571150729608)*x_ref[1]**o+(-0.799596368319)*x_ref[1]
      else:
        arg=(0.131269935442)*x[0]**o+(0.67764849927)*x[0]+(0.199946665)*x[1]**o+(0.775678673404)*x[1]+(-0.818505707274)*x[2]**o+(-0.141160005914)*x[2]
        ref=(0.131269935442)*x_ref[0]**o+(0.67764849927)*x_ref[0]+(0.199946665)*x_ref[1]**o+(0.775678673404)*x_ref[1]+(-0.818505707274)*x_ref[2]**o+(-0.141160005914)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.271030346059)*x[0]**o+(-0.604054986035)*x[0]+(-0.856492962015)*x[1]**o+(0.229582187423)*x[1]
        arg[1]=(-0.152539485929)*x[0]**o+(0.936634973224)*x[0]+(0.805016132937)*x[1]**o+(-0.6832854412)*x[1]
        ref[0]=(-0.271030346059)*x_ref[0]**o+(-0.604054986035)*x_ref[0]+(-0.856492962015)*x_ref[1]**o+(0.229582187423)*x_ref[1]
        ref[1]=(-0.152539485929)*x_ref[0]**o+(0.936634973224)*x_ref[0]+(0.805016132937)*x_ref[1]**o+(-0.6832854412)*x_ref[1]
      else:
        arg[0]=(-0.484580515247)*x[0]**o+(-0.134539997002)*x[0]+(0.589742989514)*x[1]**o+(-0.950803827747)*x[1]+(-0.852976687869)*x[2]**o+(0.0999526415345)*x[2]
        arg[1]=(-0.863620642149)*x[0]**o+(-0.16255544441)*x[0]+(-0.4963232161)*x[1]**o+(0.180988811217)*x[1]+(-0.787186716126)*x[2]**o+(-0.359825851725)*x[2]
        ref[0]=(-0.484580515247)*x_ref[0]**o+(-0.134539997002)*x_ref[0]+(0.589742989514)*x_ref[1]**o+(-0.950803827747)*x_ref[1]+(-0.852976687869)*x_ref[2]**o+(0.0999526415345)*x_ref[2]
        ref[1]=(-0.863620642149)*x_ref[0]**o+(-0.16255544441)*x_ref[0]+(-0.4963232161)*x_ref[1]**o+(0.180988811217)*x_ref[1]+(-0.787186716126)*x_ref[2]**o+(-0.359825851725)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.88064835903)*x[0]**o+(-0.140271048585)*x[0]+(-0.363467917345)*x[1]**o+(0.382530733181)*x[1]
        arg[0,1]=(0.835096280471)*x[0]**o+(0.132077921375)*x[0]+(-0.814392115734)*x[1]**o+(-0.379227781873)*x[1]
        arg[0,2]=(0.257391174596)*x[0]**o+(-0.135017751364)*x[0]+(-0.462760793484)*x[1]**o+(-0.332510860616)*x[1]
        arg[0,3]=(0.303541666957)*x[0]**o+(0.729054207976)*x[0]+(-0.0474210800089)*x[1]**o+(0.213859363809)*x[1]
        arg[0,4]=(0.564466740909)*x[0]**o+(0.684761789227)*x[0]+(0.432934636771)*x[1]**o+(0.261269614075)*x[1]
        arg[1,0]=(0.83080596334)*x[0]**o+(0.968795153198)*x[0]+(0.835115581928)*x[1]**o+(-0.134688885728)*x[1]
        arg[1,1]=(-0.876235093042)*x[0]**o+(-0.43923440557)*x[0]+(-0.641186633503)*x[1]**o+(-0.987129244994)*x[1]
        arg[1,2]=(-0.994007922096)*x[0]**o+(0.534914510762)*x[0]+(0.754420969331)*x[1]**o+(0.241533891038)*x[1]
        arg[1,3]=(-0.196027247194)*x[0]**o+(0.962231147494)*x[0]+(-0.819146467532)*x[1]**o+(0.966024528731)*x[1]
        arg[1,4]=(-0.577014723723)*x[0]**o+(0.350310620855)*x[0]+(0.16266475329)*x[1]**o+(-0.926739533525)*x[1]
        arg[2,0]=(-0.262964780787)*x[0]**o+(-0.824445469277)*x[0]+(-0.982497605841)*x[1]**o+(0.900080022932)*x[1]
        arg[2,1]=(0.54409272334)*x[0]**o+(0.789421020125)*x[0]+(0.672239490644)*x[1]**o+(-0.0306885937363)*x[1]
        arg[2,2]=(-0.130272075526)*x[0]**o+(0.51213564573)*x[0]+(-0.819817699321)*x[1]**o+(-0.259068167835)*x[1]
        arg[2,3]=(-0.61526267092)*x[0]**o+(-0.504362726269)*x[0]+(0.362429515671)*x[1]**o+(-0.0921852465615)*x[1]
        arg[2,4]=(-0.344772989987)*x[0]**o+(0.260747743828)*x[0]+(0.474457474131)*x[1]**o+(0.436008642452)*x[1]
        arg[3,0]=(-0.72740499107)*x[0]**o+(-0.297524430077)*x[0]+(-0.809254720524)*x[1]**o+(-0.230695822532)*x[1]
        arg[3,1]=(-0.0817042383885)*x[0]**o+(-0.324890823023)*x[0]+(0.923635342281)*x[1]**o+(-0.336825317481)*x[1]
        arg[3,2]=(-0.872954045409)*x[0]**o+(0.242165239629)*x[0]+(0.919264824591)*x[1]**o+(0.424247774945)*x[1]
        arg[3,3]=(0.146344350418)*x[0]**o+(-0.705285807962)*x[0]+(-0.0735095471635)*x[1]**o+(-0.464180824267)*x[1]
        arg[3,4]=(0.624085113245)*x[0]**o+(0.719452515121)*x[0]+(0.28331108038)*x[1]**o+(-0.897435702442)*x[1]
        ref[0,0]=(-0.88064835903)*x_ref[0]**o+(-0.140271048585)*x_ref[0]+(-0.363467917345)*x_ref[1]**o+(0.382530733181)*x_ref[1]
        ref[0,1]=(0.835096280471)*x_ref[0]**o+(0.132077921375)*x_ref[0]+(-0.814392115734)*x_ref[1]**o+(-0.379227781873)*x_ref[1]
        ref[0,2]=(0.257391174596)*x_ref[0]**o+(-0.135017751364)*x_ref[0]+(-0.462760793484)*x_ref[1]**o+(-0.332510860616)*x_ref[1]
        ref[0,3]=(0.303541666957)*x_ref[0]**o+(0.729054207976)*x_ref[0]+(-0.0474210800089)*x_ref[1]**o+(0.213859363809)*x_ref[1]
        ref[0,4]=(0.564466740909)*x_ref[0]**o+(0.684761789227)*x_ref[0]+(0.432934636771)*x_ref[1]**o+(0.261269614075)*x_ref[1]
        ref[1,0]=(0.83080596334)*x_ref[0]**o+(0.968795153198)*x_ref[0]+(0.835115581928)*x_ref[1]**o+(-0.134688885728)*x_ref[1]
        ref[1,1]=(-0.876235093042)*x_ref[0]**o+(-0.43923440557)*x_ref[0]+(-0.641186633503)*x_ref[1]**o+(-0.987129244994)*x_ref[1]
        ref[1,2]=(-0.994007922096)*x_ref[0]**o+(0.534914510762)*x_ref[0]+(0.754420969331)*x_ref[1]**o+(0.241533891038)*x_ref[1]
        ref[1,3]=(-0.196027247194)*x_ref[0]**o+(0.962231147494)*x_ref[0]+(-0.819146467532)*x_ref[1]**o+(0.966024528731)*x_ref[1]
        ref[1,4]=(-0.577014723723)*x_ref[0]**o+(0.350310620855)*x_ref[0]+(0.16266475329)*x_ref[1]**o+(-0.926739533525)*x_ref[1]
        ref[2,0]=(-0.262964780787)*x_ref[0]**o+(-0.824445469277)*x_ref[0]+(-0.982497605841)*x_ref[1]**o+(0.900080022932)*x_ref[1]
        ref[2,1]=(0.54409272334)*x_ref[0]**o+(0.789421020125)*x_ref[0]+(0.672239490644)*x_ref[1]**o+(-0.0306885937363)*x_ref[1]
        ref[2,2]=(-0.130272075526)*x_ref[0]**o+(0.51213564573)*x_ref[0]+(-0.819817699321)*x_ref[1]**o+(-0.259068167835)*x_ref[1]
        ref[2,3]=(-0.61526267092)*x_ref[0]**o+(-0.504362726269)*x_ref[0]+(0.362429515671)*x_ref[1]**o+(-0.0921852465615)*x_ref[1]
        ref[2,4]=(-0.344772989987)*x_ref[0]**o+(0.260747743828)*x_ref[0]+(0.474457474131)*x_ref[1]**o+(0.436008642452)*x_ref[1]
        ref[3,0]=(-0.72740499107)*x_ref[0]**o+(-0.297524430077)*x_ref[0]+(-0.809254720524)*x_ref[1]**o+(-0.230695822532)*x_ref[1]
        ref[3,1]=(-0.0817042383885)*x_ref[0]**o+(-0.324890823023)*x_ref[0]+(0.923635342281)*x_ref[1]**o+(-0.336825317481)*x_ref[1]
        ref[3,2]=(-0.872954045409)*x_ref[0]**o+(0.242165239629)*x_ref[0]+(0.919264824591)*x_ref[1]**o+(0.424247774945)*x_ref[1]
        ref[3,3]=(0.146344350418)*x_ref[0]**o+(-0.705285807962)*x_ref[0]+(-0.0735095471635)*x_ref[1]**o+(-0.464180824267)*x_ref[1]
        ref[3,4]=(0.624085113245)*x_ref[0]**o+(0.719452515121)*x_ref[0]+(0.28331108038)*x_ref[1]**o+(-0.897435702442)*x_ref[1]
      else:
        arg[0,0]=(0.47332206873)*x[0]**o+(0.20564453167)*x[0]+(-0.175485469473)*x[1]**o+(-0.25166656853)*x[1]+(-0.242693318919)*x[2]**o+(-0.773417228526)*x[2]
        arg[0,1]=(0.788563488587)*x[0]**o+(0.598162739778)*x[0]+(-0.736463777722)*x[1]**o+(-0.161555683484)*x[1]+(0.480782882175)*x[2]**o+(-0.500552084345)*x[2]
        arg[0,2]=(-0.207893050318)*x[0]**o+(0.206221241577)*x[0]+(0.306329036151)*x[1]**o+(0.585237890094)*x[1]+(0.0276469766188)*x[2]**o+(0.0502503670529)*x[2]
        arg[0,3]=(0.914950534537)*x[0]**o+(0.401315333176)*x[0]+(0.245224334898)*x[1]**o+(0.526049350134)*x[1]+(-0.735617023846)*x[2]**o+(-0.674692140531)*x[2]
        arg[0,4]=(-0.359566330106)*x[0]**o+(-0.279627730979)*x[0]+(-0.08398829656)*x[1]**o+(-0.789678621983)*x[1]+(-0.357374658443)*x[2]**o+(0.795724504883)*x[2]
        arg[1,0]=(-0.0715278193161)*x[0]**o+(-0.562120168502)*x[0]+(0.428614380101)*x[1]**o+(0.115408373591)*x[1]+(-0.154674397872)*x[2]**o+(0.233468972682)*x[2]
        arg[1,1]=(0.707445640039)*x[0]**o+(-0.719649334705)*x[0]+(-0.349743923514)*x[1]**o+(0.369262203159)*x[1]+(0.712786695739)*x[2]**o+(-0.962450154391)*x[2]
        arg[1,2]=(-0.306983094408)*x[0]**o+(-0.933322822964)*x[0]+(-0.525515899443)*x[1]**o+(0.642320930433)*x[1]+(-0.950421802297)*x[2]**o+(-0.392142385683)*x[2]
        arg[1,3]=(-0.765158324235)*x[0]**o+(-0.362010493356)*x[0]+(-0.918673930909)*x[1]**o+(0.72772784184)*x[1]+(-0.392848438408)*x[2]**o+(0.566119231816)*x[2]
        arg[1,4]=(0.429984662315)*x[0]**o+(-0.165476415697)*x[0]+(0.778539911928)*x[1]**o+(0.411518911472)*x[1]+(0.418323432603)*x[2]**o+(-0.846495913126)*x[2]
        arg[2,0]=(-0.836556750858)*x[0]**o+(0.976027316021)*x[0]+(-0.556499096672)*x[1]**o+(0.320038395915)*x[1]+(0.377663884516)*x[2]**o+(0.491202727923)*x[2]
        arg[2,1]=(-0.329597826951)*x[0]**o+(-0.555819984044)*x[0]+(0.110029358927)*x[1]**o+(-0.737070187836)*x[1]+(-0.587142473737)*x[2]**o+(-0.0388834177112)*x[2]
        arg[2,2]=(0.442251030646)*x[0]**o+(0.285010574287)*x[0]+(0.603267537331)*x[1]**o+(-0.0776884302384)*x[1]+(-0.123541484708)*x[2]**o+(0.44717125529)*x[2]
        arg[2,3]=(0.474458930313)*x[0]**o+(0.415947401939)*x[0]+(0.381539850696)*x[1]**o+(-0.307590909422)*x[1]+(0.488348451586)*x[2]**o+(0.336883810306)*x[2]
        arg[2,4]=(0.804466542611)*x[0]**o+(0.891337158669)*x[0]+(-0.599119487307)*x[1]**o+(-0.92047732727)*x[1]+(-0.251489480657)*x[2]**o+(-0.266483991526)*x[2]
        arg[3,0]=(-0.733633520827)*x[0]**o+(0.494805537624)*x[0]+(-0.117702558274)*x[1]**o+(-0.000461693303442)*x[1]+(-0.102033981202)*x[2]**o+(-0.772865985272)*x[2]
        arg[3,1]=(-0.636901276465)*x[0]**o+(-0.436330072041)*x[0]+(0.403048850222)*x[1]**o+(-0.94737100009)*x[1]+(0.781059874101)*x[2]**o+(-0.952781516037)*x[2]
        arg[3,2]=(-0.489372336164)*x[0]**o+(0.221211043152)*x[0]+(0.793975373906)*x[1]**o+(0.424349965288)*x[1]+(0.453332885662)*x[2]**o+(-0.513541753005)*x[2]
        arg[3,3]=(-0.700899677836)*x[0]**o+(-0.984271896306)*x[0]+(-0.663909486898)*x[1]**o+(0.461302174316)*x[1]+(-0.906457707724)*x[2]**o+(-0.0440003164038)*x[2]
        arg[3,4]=(0.838326295366)*x[0]**o+(-0.483407593962)*x[0]+(-0.122458752397)*x[1]**o+(0.607337951095)*x[1]+(0.750481384434)*x[2]**o+(-0.677155633204)*x[2]
        ref[0,0]=(0.47332206873)*x_ref[0]**o+(0.20564453167)*x_ref[0]+(-0.175485469473)*x_ref[1]**o+(-0.25166656853)*x_ref[1]+(-0.242693318919)*x_ref[2]**o+(-0.773417228526)*x_ref[2]
        ref[0,1]=(0.788563488587)*x_ref[0]**o+(0.598162739778)*x_ref[0]+(-0.736463777722)*x_ref[1]**o+(-0.161555683484)*x_ref[1]+(0.480782882175)*x_ref[2]**o+(-0.500552084345)*x_ref[2]
        ref[0,2]=(-0.207893050318)*x_ref[0]**o+(0.206221241577)*x_ref[0]+(0.306329036151)*x_ref[1]**o+(0.585237890094)*x_ref[1]+(0.0276469766188)*x_ref[2]**o+(0.0502503670529)*x_ref[2]
        ref[0,3]=(0.914950534537)*x_ref[0]**o+(0.401315333176)*x_ref[0]+(0.245224334898)*x_ref[1]**o+(0.526049350134)*x_ref[1]+(-0.735617023846)*x_ref[2]**o+(-0.674692140531)*x_ref[2]
        ref[0,4]=(-0.359566330106)*x_ref[0]**o+(-0.279627730979)*x_ref[0]+(-0.08398829656)*x_ref[1]**o+(-0.789678621983)*x_ref[1]+(-0.357374658443)*x_ref[2]**o+(0.795724504883)*x_ref[2]
        ref[1,0]=(-0.0715278193161)*x_ref[0]**o+(-0.562120168502)*x_ref[0]+(0.428614380101)*x_ref[1]**o+(0.115408373591)*x_ref[1]+(-0.154674397872)*x_ref[2]**o+(0.233468972682)*x_ref[2]
        ref[1,1]=(0.707445640039)*x_ref[0]**o+(-0.719649334705)*x_ref[0]+(-0.349743923514)*x_ref[1]**o+(0.369262203159)*x_ref[1]+(0.712786695739)*x_ref[2]**o+(-0.962450154391)*x_ref[2]
        ref[1,2]=(-0.306983094408)*x_ref[0]**o+(-0.933322822964)*x_ref[0]+(-0.525515899443)*x_ref[1]**o+(0.642320930433)*x_ref[1]+(-0.950421802297)*x_ref[2]**o+(-0.392142385683)*x_ref[2]
        ref[1,3]=(-0.765158324235)*x_ref[0]**o+(-0.362010493356)*x_ref[0]+(-0.918673930909)*x_ref[1]**o+(0.72772784184)*x_ref[1]+(-0.392848438408)*x_ref[2]**o+(0.566119231816)*x_ref[2]
        ref[1,4]=(0.429984662315)*x_ref[0]**o+(-0.165476415697)*x_ref[0]+(0.778539911928)*x_ref[1]**o+(0.411518911472)*x_ref[1]+(0.418323432603)*x_ref[2]**o+(-0.846495913126)*x_ref[2]
        ref[2,0]=(-0.836556750858)*x_ref[0]**o+(0.976027316021)*x_ref[0]+(-0.556499096672)*x_ref[1]**o+(0.320038395915)*x_ref[1]+(0.377663884516)*x_ref[2]**o+(0.491202727923)*x_ref[2]
        ref[2,1]=(-0.329597826951)*x_ref[0]**o+(-0.555819984044)*x_ref[0]+(0.110029358927)*x_ref[1]**o+(-0.737070187836)*x_ref[1]+(-0.587142473737)*x_ref[2]**o+(-0.0388834177112)*x_ref[2]
        ref[2,2]=(0.442251030646)*x_ref[0]**o+(0.285010574287)*x_ref[0]+(0.603267537331)*x_ref[1]**o+(-0.0776884302384)*x_ref[1]+(-0.123541484708)*x_ref[2]**o+(0.44717125529)*x_ref[2]
        ref[2,3]=(0.474458930313)*x_ref[0]**o+(0.415947401939)*x_ref[0]+(0.381539850696)*x_ref[1]**o+(-0.307590909422)*x_ref[1]+(0.488348451586)*x_ref[2]**o+(0.336883810306)*x_ref[2]
        ref[2,4]=(0.804466542611)*x_ref[0]**o+(0.891337158669)*x_ref[0]+(-0.599119487307)*x_ref[1]**o+(-0.92047732727)*x_ref[1]+(-0.251489480657)*x_ref[2]**o+(-0.266483991526)*x_ref[2]
        ref[3,0]=(-0.733633520827)*x_ref[0]**o+(0.494805537624)*x_ref[0]+(-0.117702558274)*x_ref[1]**o+(-0.000461693303442)*x_ref[1]+(-0.102033981202)*x_ref[2]**o+(-0.772865985272)*x_ref[2]
        ref[3,1]=(-0.636901276465)*x_ref[0]**o+(-0.436330072041)*x_ref[0]+(0.403048850222)*x_ref[1]**o+(-0.94737100009)*x_ref[1]+(0.781059874101)*x_ref[2]**o+(-0.952781516037)*x_ref[2]
        ref[3,2]=(-0.489372336164)*x_ref[0]**o+(0.221211043152)*x_ref[0]+(0.793975373906)*x_ref[1]**o+(0.424349965288)*x_ref[1]+(0.453332885662)*x_ref[2]**o+(-0.513541753005)*x_ref[2]
        ref[3,3]=(-0.700899677836)*x_ref[0]**o+(-0.984271896306)*x_ref[0]+(-0.663909486898)*x_ref[1]**o+(0.461302174316)*x_ref[1]+(-0.906457707724)*x_ref[2]**o+(-0.0440003164038)*x_ref[2]
        ref[3,4]=(0.838326295366)*x_ref[0]**o+(-0.483407593962)*x_ref[0]+(-0.122458752397)*x_ref[1]**o+(0.607337951095)*x_ref[1]+(0.750481384434)*x_ref[2]**o+(-0.677155633204)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.111473275114)*x[0]**o+(-0.552264448532)*x[0]+(0.67142534913)*x[1]**o+(-0.642305751076)*x[1]
        arg[0,0,1]=(-0.431075800974)*x[0]**o+(0.911382078454)*x[0]+(0.546541644612)*x[1]**o+(-0.67679652536)*x[1]
        arg[0,1,0]=(0.103056292292)*x[0]**o+(-0.640858804416)*x[0]+(0.159596964992)*x[1]**o+(-0.237782007316)*x[1]
        arg[0,1,1]=(0.975898170389)*x[0]**o+(0.255916326383)*x[0]+(-0.692770462322)*x[1]**o+(0.786355114846)*x[1]
        arg[1,0,0]=(-0.669378370759)*x[0]**o+(-0.0153788709146)*x[0]+(0.924383188557)*x[1]**o+(0.885037556829)*x[1]
        arg[1,0,1]=(-0.509891819881)*x[0]**o+(-0.377779514697)*x[0]+(-0.343176473036)*x[1]**o+(-0.39978416725)*x[1]
        arg[1,1,0]=(0.457959467665)*x[0]**o+(0.786511967341)*x[0]+(0.256561665146)*x[1]**o+(-0.261300099643)*x[1]
        arg[1,1,1]=(0.219754515189)*x[0]**o+(-0.131747762867)*x[0]+(0.736017818893)*x[1]**o+(-0.450963700533)*x[1]
        arg[2,0,0]=(-0.182946832789)*x[0]**o+(-0.103064731935)*x[0]+(-0.223584334222)*x[1]**o+(-0.578088201628)*x[1]
        arg[2,0,1]=(-0.830744157425)*x[0]**o+(0.590627167207)*x[0]+(-0.538048378049)*x[1]**o+(-0.0529793352741)*x[1]
        arg[2,1,0]=(-0.307908670388)*x[0]**o+(-0.929987734045)*x[0]+(-0.167190563092)*x[1]**o+(0.126466628227)*x[1]
        arg[2,1,1]=(-0.812045378916)*x[0]**o+(0.676973938072)*x[0]+(0.528243909204)*x[1]**o+(-0.594241387691)*x[1]
        arg[3,0,0]=(-0.0198606164114)*x[0]**o+(-0.00574708685554)*x[0]+(0.908946312745)*x[1]**o+(0.775054099788)*x[1]
        arg[3,0,1]=(-0.810078120518)*x[0]**o+(-0.960585888485)*x[0]+(-0.0747272427198)*x[1]**o+(0.164107641817)*x[1]
        arg[3,1,0]=(-0.535791917139)*x[0]**o+(0.0424226440074)*x[0]+(0.701479263902)*x[1]**o+(0.787745423088)*x[1]
        arg[3,1,1]=(-0.896532152793)*x[0]**o+(-0.0224566920251)*x[0]+(-0.554655626689)*x[1]**o+(0.41012867952)*x[1]
        arg[4,0,0]=(0.972433817496)*x[0]**o+(-0.440662536033)*x[0]+(-0.910456530943)*x[1]**o+(-0.333429535028)*x[1]
        arg[4,0,1]=(-0.415824313942)*x[0]**o+(0.00255557712922)*x[0]+(0.304897639195)*x[1]**o+(-0.866600848052)*x[1]
        arg[4,1,0]=(-0.866277739076)*x[0]**o+(0.282976169776)*x[0]+(0.477528278501)*x[1]**o+(-0.233506497695)*x[1]
        arg[4,1,1]=(-0.12729626381)*x[0]**o+(-0.723204796785)*x[0]+(0.0856495713442)*x[1]**o+(-0.214819437731)*x[1]
        arg[5,0,0]=(0.208133503644)*x[0]**o+(-0.955529202895)*x[0]+(0.0291569127831)*x[1]**o+(0.489913918305)*x[1]
        arg[5,0,1]=(0.96923667477)*x[0]**o+(0.543776525315)*x[0]+(0.38666450029)*x[1]**o+(-0.193493550272)*x[1]
        arg[5,1,0]=(-0.0912940016888)*x[0]**o+(-0.477561072681)*x[0]+(0.474816021613)*x[1]**o+(-0.252664229253)*x[1]
        arg[5,1,1]=(0.748337265159)*x[0]**o+(-0.706194454618)*x[0]+(-0.504466281507)*x[1]**o+(0.582500817679)*x[1]
        ref[0,0,0]=(0.111473275114)*x_ref[0]**o+(-0.552264448532)*x_ref[0]+(0.67142534913)*x_ref[1]**o+(-0.642305751076)*x_ref[1]
        ref[0,0,1]=(-0.431075800974)*x_ref[0]**o+(0.911382078454)*x_ref[0]+(0.546541644612)*x_ref[1]**o+(-0.67679652536)*x_ref[1]
        ref[0,1,0]=(0.103056292292)*x_ref[0]**o+(-0.640858804416)*x_ref[0]+(0.159596964992)*x_ref[1]**o+(-0.237782007316)*x_ref[1]
        ref[0,1,1]=(0.975898170389)*x_ref[0]**o+(0.255916326383)*x_ref[0]+(-0.692770462322)*x_ref[1]**o+(0.786355114846)*x_ref[1]
        ref[1,0,0]=(-0.669378370759)*x_ref[0]**o+(-0.0153788709146)*x_ref[0]+(0.924383188557)*x_ref[1]**o+(0.885037556829)*x_ref[1]
        ref[1,0,1]=(-0.509891819881)*x_ref[0]**o+(-0.377779514697)*x_ref[0]+(-0.343176473036)*x_ref[1]**o+(-0.39978416725)*x_ref[1]
        ref[1,1,0]=(0.457959467665)*x_ref[0]**o+(0.786511967341)*x_ref[0]+(0.256561665146)*x_ref[1]**o+(-0.261300099643)*x_ref[1]
        ref[1,1,1]=(0.219754515189)*x_ref[0]**o+(-0.131747762867)*x_ref[0]+(0.736017818893)*x_ref[1]**o+(-0.450963700533)*x_ref[1]
        ref[2,0,0]=(-0.182946832789)*x_ref[0]**o+(-0.103064731935)*x_ref[0]+(-0.223584334222)*x_ref[1]**o+(-0.578088201628)*x_ref[1]
        ref[2,0,1]=(-0.830744157425)*x_ref[0]**o+(0.590627167207)*x_ref[0]+(-0.538048378049)*x_ref[1]**o+(-0.0529793352741)*x_ref[1]
        ref[2,1,0]=(-0.307908670388)*x_ref[0]**o+(-0.929987734045)*x_ref[0]+(-0.167190563092)*x_ref[1]**o+(0.126466628227)*x_ref[1]
        ref[2,1,1]=(-0.812045378916)*x_ref[0]**o+(0.676973938072)*x_ref[0]+(0.528243909204)*x_ref[1]**o+(-0.594241387691)*x_ref[1]
        ref[3,0,0]=(-0.0198606164114)*x_ref[0]**o+(-0.00574708685554)*x_ref[0]+(0.908946312745)*x_ref[1]**o+(0.775054099788)*x_ref[1]
        ref[3,0,1]=(-0.810078120518)*x_ref[0]**o+(-0.960585888485)*x_ref[0]+(-0.0747272427198)*x_ref[1]**o+(0.164107641817)*x_ref[1]
        ref[3,1,0]=(-0.535791917139)*x_ref[0]**o+(0.0424226440074)*x_ref[0]+(0.701479263902)*x_ref[1]**o+(0.787745423088)*x_ref[1]
        ref[3,1,1]=(-0.896532152793)*x_ref[0]**o+(-0.0224566920251)*x_ref[0]+(-0.554655626689)*x_ref[1]**o+(0.41012867952)*x_ref[1]
        ref[4,0,0]=(0.972433817496)*x_ref[0]**o+(-0.440662536033)*x_ref[0]+(-0.910456530943)*x_ref[1]**o+(-0.333429535028)*x_ref[1]
        ref[4,0,1]=(-0.415824313942)*x_ref[0]**o+(0.00255557712922)*x_ref[0]+(0.304897639195)*x_ref[1]**o+(-0.866600848052)*x_ref[1]
        ref[4,1,0]=(-0.866277739076)*x_ref[0]**o+(0.282976169776)*x_ref[0]+(0.477528278501)*x_ref[1]**o+(-0.233506497695)*x_ref[1]
        ref[4,1,1]=(-0.12729626381)*x_ref[0]**o+(-0.723204796785)*x_ref[0]+(0.0856495713442)*x_ref[1]**o+(-0.214819437731)*x_ref[1]
        ref[5,0,0]=(0.208133503644)*x_ref[0]**o+(-0.955529202895)*x_ref[0]+(0.0291569127831)*x_ref[1]**o+(0.489913918305)*x_ref[1]
        ref[5,0,1]=(0.96923667477)*x_ref[0]**o+(0.543776525315)*x_ref[0]+(0.38666450029)*x_ref[1]**o+(-0.193493550272)*x_ref[1]
        ref[5,1,0]=(-0.0912940016888)*x_ref[0]**o+(-0.477561072681)*x_ref[0]+(0.474816021613)*x_ref[1]**o+(-0.252664229253)*x_ref[1]
        ref[5,1,1]=(0.748337265159)*x_ref[0]**o+(-0.706194454618)*x_ref[0]+(-0.504466281507)*x_ref[1]**o+(0.582500817679)*x_ref[1]
      else:
        arg[0,0,0]=(-0.179972720095)*x[0]**o+(-0.546715268523)*x[0]+(0.835867431319)*x[1]**o+(0.626798499259)*x[1]+(0.397356013278)*x[2]**o+(0.0771586784158)*x[2]
        arg[0,0,1]=(0.625676031039)*x[0]**o+(0.65843849475)*x[0]+(0.456469199307)*x[1]**o+(-0.970709759479)*x[1]+(-0.699606143992)*x[2]**o+(0.76429400724)*x[2]
        arg[0,1,0]=(0.613916424165)*x[0]**o+(0.556778588303)*x[0]+(-0.509658743529)*x[1]**o+(-0.978718968218)*x[1]+(-0.603928369428)*x[2]**o+(-0.550813953692)*x[2]
        arg[0,1,1]=(-0.935214755436)*x[0]**o+(0.44745901813)*x[0]+(-0.582828846453)*x[1]**o+(-0.181547183233)*x[1]+(-0.183912053485)*x[2]**o+(0.446152494296)*x[2]
        arg[1,0,0]=(-0.486345219779)*x[0]**o+(0.753529436833)*x[0]+(0.138157856552)*x[1]**o+(-0.874849020539)*x[1]+(0.374396543792)*x[2]**o+(0.26548456046)*x[2]
        arg[1,0,1]=(0.492908753036)*x[0]**o+(-0.418118685105)*x[0]+(0.880817569823)*x[1]**o+(-0.827691145524)*x[1]+(0.238633156133)*x[2]**o+(-0.463687731341)*x[2]
        arg[1,1,0]=(-0.397373578442)*x[0]**o+(-0.316321105607)*x[0]+(0.647511225459)*x[1]**o+(-0.82553752195)*x[1]+(-0.715549303171)*x[2]**o+(0.804623957459)*x[2]
        arg[1,1,1]=(0.414743130234)*x[0]**o+(0.0201794413683)*x[0]+(-0.636337599584)*x[1]**o+(-0.230053616026)*x[1]+(-0.4302153311)*x[2]**o+(-0.387468029626)*x[2]
        arg[2,0,0]=(0.715100517336)*x[0]**o+(-0.4286120178)*x[0]+(0.0987858542117)*x[1]**o+(0.815093470301)*x[1]+(0.293455353257)*x[2]**o+(0.200576090379)*x[2]
        arg[2,0,1]=(-0.158330015822)*x[0]**o+(0.363339023457)*x[0]+(-0.796055594971)*x[1]**o+(-0.213299205306)*x[1]+(-0.136620988812)*x[2]**o+(0.10733886947)*x[2]
        arg[2,1,0]=(-0.258729405118)*x[0]**o+(0.0669553480038)*x[0]+(0.997663166613)*x[1]**o+(0.00398861063931)*x[1]+(-0.989459856803)*x[2]**o+(0.932687384325)*x[2]
        arg[2,1,1]=(-0.974431322319)*x[0]**o+(-0.486447431804)*x[0]+(0.871784873347)*x[1]**o+(-0.978904038765)*x[1]+(0.647341201506)*x[2]**o+(-0.269683556528)*x[2]
        arg[3,0,0]=(-0.369507398073)*x[0]**o+(-0.772085607074)*x[0]+(0.417550881937)*x[1]**o+(0.271743719652)*x[1]+(0.881807154737)*x[2]**o+(0.54725428039)*x[2]
        arg[3,0,1]=(0.440089335941)*x[0]**o+(0.0844171421294)*x[0]+(0.926389704362)*x[1]**o+(-0.0709182998297)*x[1]+(0.903978660317)*x[2]**o+(-0.160556206716)*x[2]
        arg[3,1,0]=(-0.49936391282)*x[0]**o+(-0.632753931881)*x[0]+(-0.604686084554)*x[1]**o+(-0.954205199531)*x[1]+(-0.535102184845)*x[2]**o+(0.621683814123)*x[2]
        arg[3,1,1]=(-0.117790973132)*x[0]**o+(-0.649091727896)*x[0]+(0.925682916785)*x[1]**o+(0.852053515469)*x[1]+(-0.479218429897)*x[2]**o+(0.426823478575)*x[2]
        arg[4,0,0]=(0.454174564319)*x[0]**o+(-0.855975647406)*x[0]+(0.775972969288)*x[1]**o+(0.267900999893)*x[1]+(0.221387501995)*x[2]**o+(-0.26413791201)*x[2]
        arg[4,0,1]=(0.875191302431)*x[0]**o+(-0.362782502152)*x[0]+(0.24600070339)*x[1]**o+(-0.566709533298)*x[1]+(-0.82187455531)*x[2]**o+(-0.943929640301)*x[2]
        arg[4,1,0]=(-0.409313259973)*x[0]**o+(-0.784833509001)*x[0]+(-0.939690505056)*x[1]**o+(-0.0736280226974)*x[1]+(0.939072745486)*x[2]**o+(0.611697159959)*x[2]
        arg[4,1,1]=(0.893374246808)*x[0]**o+(0.36113513561)*x[0]+(0.227899271271)*x[1]**o+(-0.847221657012)*x[1]+(-0.0106368947417)*x[2]**o+(0.0358896817595)*x[2]
        arg[5,0,0]=(0.0329230524924)*x[0]**o+(0.0565192084014)*x[0]+(-0.68398448658)*x[1]**o+(0.909035618076)*x[1]+(-0.900060440872)*x[2]**o+(-0.481014787833)*x[2]
        arg[5,0,1]=(0.404966692391)*x[0]**o+(0.647468764373)*x[0]+(-0.723847499346)*x[1]**o+(-0.0740896957606)*x[1]+(-0.00628515079923)*x[2]**o+(0.906667863539)*x[2]
        arg[5,1,0]=(0.983181781531)*x[0]**o+(-0.297325990236)*x[0]+(0.00489064543352)*x[1]**o+(0.667560767038)*x[1]+(-0.665768141301)*x[2]**o+(0.772761393608)*x[2]
        arg[5,1,1]=(-0.496581691354)*x[0]**o+(0.0121337979505)*x[0]+(0.492281382726)*x[1]**o+(0.545232714025)*x[1]+(0.418321481532)*x[2]**o+(0.489635969764)*x[2]
        ref[0,0,0]=(-0.179972720095)*x_ref[0]**o+(-0.546715268523)*x_ref[0]+(0.835867431319)*x_ref[1]**o+(0.626798499259)*x_ref[1]+(0.397356013278)*x_ref[2]**o+(0.0771586784158)*x_ref[2]
        ref[0,0,1]=(0.625676031039)*x_ref[0]**o+(0.65843849475)*x_ref[0]+(0.456469199307)*x_ref[1]**o+(-0.970709759479)*x_ref[1]+(-0.699606143992)*x_ref[2]**o+(0.76429400724)*x_ref[2]
        ref[0,1,0]=(0.613916424165)*x_ref[0]**o+(0.556778588303)*x_ref[0]+(-0.509658743529)*x_ref[1]**o+(-0.978718968218)*x_ref[1]+(-0.603928369428)*x_ref[2]**o+(-0.550813953692)*x_ref[2]
        ref[0,1,1]=(-0.935214755436)*x_ref[0]**o+(0.44745901813)*x_ref[0]+(-0.582828846453)*x_ref[1]**o+(-0.181547183233)*x_ref[1]+(-0.183912053485)*x_ref[2]**o+(0.446152494296)*x_ref[2]
        ref[1,0,0]=(-0.486345219779)*x_ref[0]**o+(0.753529436833)*x_ref[0]+(0.138157856552)*x_ref[1]**o+(-0.874849020539)*x_ref[1]+(0.374396543792)*x_ref[2]**o+(0.26548456046)*x_ref[2]
        ref[1,0,1]=(0.492908753036)*x_ref[0]**o+(-0.418118685105)*x_ref[0]+(0.880817569823)*x_ref[1]**o+(-0.827691145524)*x_ref[1]+(0.238633156133)*x_ref[2]**o+(-0.463687731341)*x_ref[2]
        ref[1,1,0]=(-0.397373578442)*x_ref[0]**o+(-0.316321105607)*x_ref[0]+(0.647511225459)*x_ref[1]**o+(-0.82553752195)*x_ref[1]+(-0.715549303171)*x_ref[2]**o+(0.804623957459)*x_ref[2]
        ref[1,1,1]=(0.414743130234)*x_ref[0]**o+(0.0201794413683)*x_ref[0]+(-0.636337599584)*x_ref[1]**o+(-0.230053616026)*x_ref[1]+(-0.4302153311)*x_ref[2]**o+(-0.387468029626)*x_ref[2]
        ref[2,0,0]=(0.715100517336)*x_ref[0]**o+(-0.4286120178)*x_ref[0]+(0.0987858542117)*x_ref[1]**o+(0.815093470301)*x_ref[1]+(0.293455353257)*x_ref[2]**o+(0.200576090379)*x_ref[2]
        ref[2,0,1]=(-0.158330015822)*x_ref[0]**o+(0.363339023457)*x_ref[0]+(-0.796055594971)*x_ref[1]**o+(-0.213299205306)*x_ref[1]+(-0.136620988812)*x_ref[2]**o+(0.10733886947)*x_ref[2]
        ref[2,1,0]=(-0.258729405118)*x_ref[0]**o+(0.0669553480038)*x_ref[0]+(0.997663166613)*x_ref[1]**o+(0.00398861063931)*x_ref[1]+(-0.989459856803)*x_ref[2]**o+(0.932687384325)*x_ref[2]
        ref[2,1,1]=(-0.974431322319)*x_ref[0]**o+(-0.486447431804)*x_ref[0]+(0.871784873347)*x_ref[1]**o+(-0.978904038765)*x_ref[1]+(0.647341201506)*x_ref[2]**o+(-0.269683556528)*x_ref[2]
        ref[3,0,0]=(-0.369507398073)*x_ref[0]**o+(-0.772085607074)*x_ref[0]+(0.417550881937)*x_ref[1]**o+(0.271743719652)*x_ref[1]+(0.881807154737)*x_ref[2]**o+(0.54725428039)*x_ref[2]
        ref[3,0,1]=(0.440089335941)*x_ref[0]**o+(0.0844171421294)*x_ref[0]+(0.926389704362)*x_ref[1]**o+(-0.0709182998297)*x_ref[1]+(0.903978660317)*x_ref[2]**o+(-0.160556206716)*x_ref[2]
        ref[3,1,0]=(-0.49936391282)*x_ref[0]**o+(-0.632753931881)*x_ref[0]+(-0.604686084554)*x_ref[1]**o+(-0.954205199531)*x_ref[1]+(-0.535102184845)*x_ref[2]**o+(0.621683814123)*x_ref[2]
        ref[3,1,1]=(-0.117790973132)*x_ref[0]**o+(-0.649091727896)*x_ref[0]+(0.925682916785)*x_ref[1]**o+(0.852053515469)*x_ref[1]+(-0.479218429897)*x_ref[2]**o+(0.426823478575)*x_ref[2]
        ref[4,0,0]=(0.454174564319)*x_ref[0]**o+(-0.855975647406)*x_ref[0]+(0.775972969288)*x_ref[1]**o+(0.267900999893)*x_ref[1]+(0.221387501995)*x_ref[2]**o+(-0.26413791201)*x_ref[2]
        ref[4,0,1]=(0.875191302431)*x_ref[0]**o+(-0.362782502152)*x_ref[0]+(0.24600070339)*x_ref[1]**o+(-0.566709533298)*x_ref[1]+(-0.82187455531)*x_ref[2]**o+(-0.943929640301)*x_ref[2]
        ref[4,1,0]=(-0.409313259973)*x_ref[0]**o+(-0.784833509001)*x_ref[0]+(-0.939690505056)*x_ref[1]**o+(-0.0736280226974)*x_ref[1]+(0.939072745486)*x_ref[2]**o+(0.611697159959)*x_ref[2]
        ref[4,1,1]=(0.893374246808)*x_ref[0]**o+(0.36113513561)*x_ref[0]+(0.227899271271)*x_ref[1]**o+(-0.847221657012)*x_ref[1]+(-0.0106368947417)*x_ref[2]**o+(0.0358896817595)*x_ref[2]
        ref[5,0,0]=(0.0329230524924)*x_ref[0]**o+(0.0565192084014)*x_ref[0]+(-0.68398448658)*x_ref[1]**o+(0.909035618076)*x_ref[1]+(-0.900060440872)*x_ref[2]**o+(-0.481014787833)*x_ref[2]
        ref[5,0,1]=(0.404966692391)*x_ref[0]**o+(0.647468764373)*x_ref[0]+(-0.723847499346)*x_ref[1]**o+(-0.0740896957606)*x_ref[1]+(-0.00628515079923)*x_ref[2]**o+(0.906667863539)*x_ref[2]
        ref[5,1,0]=(0.983181781531)*x_ref[0]**o+(-0.297325990236)*x_ref[0]+(0.00489064543352)*x_ref[1]**o+(0.667560767038)*x_ref[1]+(-0.665768141301)*x_ref[2]**o+(0.772761393608)*x_ref[2]
        ref[5,1,1]=(-0.496581691354)*x_ref[0]**o+(0.0121337979505)*x_ref[0]+(0.492281382726)*x_ref[1]**o+(0.545232714025)*x_ref[1]+(0.418321481532)*x_ref[2]**o+(0.489635969764)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.282394950702)*x[0]**o+(0.836580157584)*x[0]+(0.337652262648)*x[1]**o+(-0.442694263325)*x[1]
        arg[0,0,0,1]=(0.85150774028)*x[0]**o+(0.998297296593)*x[0]+(-0.710029954342)*x[1]**o+(0.0816168287837)*x[1]
        arg[0,0,1,0]=(-0.636435102838)*x[0]**o+(-0.613479863553)*x[0]+(-0.861506513168)*x[1]**o+(-0.242998961478)*x[1]
        arg[0,0,1,1]=(-0.00671863978521)*x[0]**o+(0.909404084068)*x[0]+(-0.734956067026)*x[1]**o+(0.256530538264)*x[1]
        arg[0,0,2,0]=(-0.646631608999)*x[0]**o+(-0.521130455843)*x[0]+(-0.44328404557)*x[1]**o+(-0.156860325817)*x[1]
        arg[0,0,2,1]=(-0.817550008949)*x[0]**o+(0.702414497795)*x[0]+(0.946449680763)*x[1]**o+(-0.61576361925)*x[1]
        arg[0,1,0,0]=(-0.527206691595)*x[0]**o+(-0.43469216984)*x[0]+(-0.108455901154)*x[1]**o+(0.435646962156)*x[1]
        arg[0,1,0,1]=(-0.148522219535)*x[0]**o+(0.146912249755)*x[0]+(-0.414934945665)*x[1]**o+(-0.422754988714)*x[1]
        arg[0,1,1,0]=(0.177760972947)*x[0]**o+(-0.938522530862)*x[0]+(0.566229652264)*x[1]**o+(0.255269743628)*x[1]
        arg[0,1,1,1]=(-0.317807933792)*x[0]**o+(0.432181451246)*x[0]+(-0.358821513362)*x[1]**o+(-0.470517232585)*x[1]
        arg[0,1,2,0]=(0.486536430227)*x[0]**o+(-0.533653809349)*x[0]+(-0.126612514045)*x[1]**o+(0.103088115487)*x[1]
        arg[0,1,2,1]=(-0.995976559589)*x[0]**o+(0.965537167421)*x[0]+(-0.351307699531)*x[1]**o+(-0.684367434666)*x[1]
        arg[0,2,0,0]=(0.976008613768)*x[0]**o+(-0.890573007002)*x[0]+(-0.440709423944)*x[1]**o+(0.315724459367)*x[1]
        arg[0,2,0,1]=(-0.595275811962)*x[0]**o+(-0.599634144509)*x[0]+(0.646466002346)*x[1]**o+(-0.94529568651)*x[1]
        arg[0,2,1,0]=(0.977691521821)*x[0]**o+(-0.0730997643682)*x[0]+(-0.338897498737)*x[1]**o+(0.246308678824)*x[1]
        arg[0,2,1,1]=(-0.111317410732)*x[0]**o+(-0.574381793398)*x[0]+(-0.0314431899843)*x[1]**o+(0.393105047096)*x[1]
        arg[0,2,2,0]=(0.886320020757)*x[0]**o+(-0.785042003544)*x[0]+(0.82609269979)*x[1]**o+(-0.677169484337)*x[1]
        arg[0,2,2,1]=(0.811027138079)*x[0]**o+(-0.375108470948)*x[0]+(0.781472464253)*x[1]**o+(0.217640794143)*x[1]
        arg[0,3,0,0]=(0.264884117091)*x[0]**o+(0.116591319557)*x[0]+(0.282667679763)*x[1]**o+(0.111123454337)*x[1]
        arg[0,3,0,1]=(0.0451626985922)*x[0]**o+(-0.93771279728)*x[0]+(0.129171764072)*x[1]**o+(0.598367954162)*x[1]
        arg[0,3,1,0]=(0.905243511086)*x[0]**o+(0.0837062573586)*x[0]+(0.0657223248433)*x[1]**o+(-0.746644690072)*x[1]
        arg[0,3,1,1]=(-0.0547856755485)*x[0]**o+(-0.22157806755)*x[0]+(0.124623625108)*x[1]**o+(0.0436590157081)*x[1]
        arg[0,3,2,0]=(-0.558323744577)*x[0]**o+(-0.646043519007)*x[0]+(0.991885458004)*x[1]**o+(-0.634140808163)*x[1]
        arg[0,3,2,1]=(-0.27247983028)*x[0]**o+(0.803906404485)*x[0]+(-0.816826609924)*x[1]**o+(-0.969532783336)*x[1]
        arg[0,4,0,0]=(0.638733842891)*x[0]**o+(-0.972793270208)*x[0]+(0.714611008099)*x[1]**o+(0.586988595847)*x[1]
        arg[0,4,0,1]=(0.63584760734)*x[0]**o+(-0.653740236747)*x[0]+(-0.596304731446)*x[1]**o+(0.42519786429)*x[1]
        arg[0,4,1,0]=(0.0665562973529)*x[0]**o+(0.362152092093)*x[0]+(0.732530335739)*x[1]**o+(0.618536326941)*x[1]
        arg[0,4,1,1]=(0.486060667608)*x[0]**o+(0.59489013279)*x[0]+(-0.547237017498)*x[1]**o+(0.0371894705857)*x[1]
        arg[0,4,2,0]=(0.395505844397)*x[0]**o+(-0.869090434535)*x[0]+(-0.606640827437)*x[1]**o+(0.269995459083)*x[1]
        arg[0,4,2,1]=(0.304671547176)*x[0]**o+(-0.515772523849)*x[0]+(0.546114300876)*x[1]**o+(0.530247404341)*x[1]
        arg[1,0,0,0]=(0.408332677433)*x[0]**o+(-0.0778807359918)*x[0]+(-0.803337489297)*x[1]**o+(-0.123835718422)*x[1]
        arg[1,0,0,1]=(-0.00441620254944)*x[0]**o+(0.0321259598831)*x[0]+(-0.998381508472)*x[1]**o+(0.800767747511)*x[1]
        arg[1,0,1,0]=(-0.923683057418)*x[0]**o+(0.579405114902)*x[0]+(-0.939639823585)*x[1]**o+(-0.138423358134)*x[1]
        arg[1,0,1,1]=(-0.558420234848)*x[0]**o+(0.133316569813)*x[0]+(-0.294134717329)*x[1]**o+(-0.750897749044)*x[1]
        arg[1,0,2,0]=(0.478951298545)*x[0]**o+(-0.807173978799)*x[0]+(0.852438146763)*x[1]**o+(-0.103814229563)*x[1]
        arg[1,0,2,1]=(-0.875250726603)*x[0]**o+(-0.934803428491)*x[0]+(0.143352873672)*x[1]**o+(0.600934881206)*x[1]
        arg[1,1,0,0]=(-0.0539476321018)*x[0]**o+(0.194177663417)*x[0]+(-0.862877991009)*x[1]**o+(-0.438337993664)*x[1]
        arg[1,1,0,1]=(-0.623379355474)*x[0]**o+(-0.0746994739748)*x[0]+(0.365005438956)*x[1]**o+(-0.431737688205)*x[1]
        arg[1,1,1,0]=(0.0936054461735)*x[0]**o+(-0.974594269446)*x[0]+(-0.121917598428)*x[1]**o+(-0.84429083727)*x[1]
        arg[1,1,1,1]=(0.967763348672)*x[0]**o+(0.331698826475)*x[0]+(0.66448515877)*x[1]**o+(0.00214985930661)*x[1]
        arg[1,1,2,0]=(-0.312146580762)*x[0]**o+(0.998152287706)*x[0]+(-0.474242942777)*x[1]**o+(0.637827347846)*x[1]
        arg[1,1,2,1]=(-0.556765443436)*x[0]**o+(0.227327929774)*x[0]+(0.739457845569)*x[1]**o+(-0.615070017663)*x[1]
        arg[1,2,0,0]=(-0.23030660552)*x[0]**o+(0.750479465953)*x[0]+(0.278215408392)*x[1]**o+(-0.640618698224)*x[1]
        arg[1,2,0,1]=(-0.0713217101129)*x[0]**o+(0.500737477887)*x[0]+(-0.972834770005)*x[1]**o+(-0.420386056033)*x[1]
        arg[1,2,1,0]=(-0.497322989847)*x[0]**o+(-0.528685124841)*x[0]+(-0.934386348496)*x[1]**o+(0.191930553839)*x[1]
        arg[1,2,1,1]=(0.124362419781)*x[0]**o+(-0.0457849787456)*x[0]+(-0.360596154712)*x[1]**o+(0.799736253715)*x[1]
        arg[1,2,2,0]=(-0.346294878893)*x[0]**o+(0.427772309361)*x[0]+(-0.0642666791012)*x[1]**o+(-0.214442238311)*x[1]
        arg[1,2,2,1]=(0.24434837411)*x[0]**o+(-0.913373159786)*x[0]+(0.158750862391)*x[1]**o+(-0.544328807576)*x[1]
        arg[1,3,0,0]=(0.870847002687)*x[0]**o+(-0.466884462301)*x[0]+(-0.343358829445)*x[1]**o+(-0.180956419766)*x[1]
        arg[1,3,0,1]=(-0.813518850863)*x[0]**o+(0.960810467501)*x[0]+(0.35091416904)*x[1]**o+(0.84534544948)*x[1]
        arg[1,3,1,0]=(-0.284687751218)*x[0]**o+(-0.274583519355)*x[0]+(0.338892764102)*x[1]**o+(-0.145142443753)*x[1]
        arg[1,3,1,1]=(0.921064262137)*x[0]**o+(0.152552421678)*x[0]+(0.827247693375)*x[1]**o+(-0.569529848619)*x[1]
        arg[1,3,2,0]=(0.133567630356)*x[0]**o+(-0.0429397358364)*x[0]+(-0.175383170663)*x[1]**o+(0.584407631334)*x[1]
        arg[1,3,2,1]=(0.203880167371)*x[0]**o+(-0.422978073825)*x[0]+(-0.210109267588)*x[1]**o+(-0.902327373219)*x[1]
        arg[1,4,0,0]=(-0.566386864876)*x[0]**o+(-0.139795085483)*x[0]+(-0.00852913464238)*x[1]**o+(0.569151381884)*x[1]
        arg[1,4,0,1]=(-0.937256625212)*x[0]**o+(0.563438253128)*x[0]+(-0.586182238715)*x[1]**o+(-0.566169678723)*x[1]
        arg[1,4,1,0]=(-0.73296890712)*x[0]**o+(0.359249594406)*x[0]+(0.66050790616)*x[1]**o+(-0.367919797426)*x[1]
        arg[1,4,1,1]=(-0.81165734248)*x[0]**o+(0.460592817838)*x[0]+(0.953573909492)*x[1]**o+(0.329659021274)*x[1]
        arg[1,4,2,0]=(0.345994483595)*x[0]**o+(-0.18547324343)*x[0]+(0.669452926238)*x[1]**o+(0.42513249272)*x[1]
        arg[1,4,2,1]=(0.250677642616)*x[0]**o+(0.466118496803)*x[0]+(-0.776608543595)*x[1]**o+(0.449144243682)*x[1]
        arg[2,0,0,0]=(0.244172909021)*x[0]**o+(0.0311068322898)*x[0]+(-0.909348039508)*x[1]**o+(0.376500719414)*x[1]
        arg[2,0,0,1]=(-0.033516640667)*x[0]**o+(0.698424718328)*x[0]+(-0.282814785542)*x[1]**o+(0.61118022947)*x[1]
        arg[2,0,1,0]=(-0.946672649538)*x[0]**o+(0.814325244591)*x[0]+(-0.353215562825)*x[1]**o+(-0.138150257429)*x[1]
        arg[2,0,1,1]=(-0.886509051692)*x[0]**o+(-0.0468334385819)*x[0]+(0.71984719189)*x[1]**o+(-0.324744971493)*x[1]
        arg[2,0,2,0]=(-0.780921647053)*x[0]**o+(-0.401418496932)*x[0]+(0.647847274332)*x[1]**o+(-0.831834322278)*x[1]
        arg[2,0,2,1]=(0.835155004862)*x[0]**o+(0.684459986902)*x[0]+(-0.301244240677)*x[1]**o+(-0.161292780187)*x[1]
        arg[2,1,0,0]=(-0.652199398394)*x[0]**o+(0.62480722178)*x[0]+(-0.474535712914)*x[1]**o+(0.926061797766)*x[1]
        arg[2,1,0,1]=(-0.813365478568)*x[0]**o+(-0.38795792417)*x[0]+(0.378161353697)*x[1]**o+(0.70581806211)*x[1]
        arg[2,1,1,0]=(0.729265594934)*x[0]**o+(-0.469397842359)*x[0]+(0.642472199926)*x[1]**o+(-0.779527283011)*x[1]
        arg[2,1,1,1]=(0.0517342917574)*x[0]**o+(-0.414111007917)*x[0]+(-0.89437375241)*x[1]**o+(-0.830119585842)*x[1]
        arg[2,1,2,0]=(0.769898486257)*x[0]**o+(0.791725624832)*x[0]+(0.297519329396)*x[1]**o+(-0.959462006611)*x[1]
        arg[2,1,2,1]=(0.817802144251)*x[0]**o+(0.69851258911)*x[0]+(-0.174401735228)*x[1]**o+(0.13449696498)*x[1]
        arg[2,2,0,0]=(0.547972662256)*x[0]**o+(-0.66491929979)*x[0]+(0.330783041672)*x[1]**o+(0.816196730606)*x[1]
        arg[2,2,0,1]=(0.245971097763)*x[0]**o+(0.213087481481)*x[0]+(-0.840766291392)*x[1]**o+(0.388414557376)*x[1]
        arg[2,2,1,0]=(0.523151676444)*x[0]**o+(0.854817486027)*x[0]+(-0.53570188528)*x[1]**o+(-0.68149362242)*x[1]
        arg[2,2,1,1]=(-0.984200247787)*x[0]**o+(0.83324365092)*x[0]+(-0.896184408816)*x[1]**o+(0.102276243487)*x[1]
        arg[2,2,2,0]=(-0.216357164308)*x[0]**o+(-0.597984882107)*x[0]+(-0.205435921511)*x[1]**o+(0.484947652759)*x[1]
        arg[2,2,2,1]=(-0.650046253868)*x[0]**o+(-0.624878186105)*x[0]+(0.786519914406)*x[1]**o+(0.28027273095)*x[1]
        arg[2,3,0,0]=(-0.0821718346579)*x[0]**o+(-0.687049917467)*x[0]+(0.380792333736)*x[1]**o+(0.151185837715)*x[1]
        arg[2,3,0,1]=(0.61796113854)*x[0]**o+(0.844624413065)*x[0]+(-0.980619401098)*x[1]**o+(-0.854924862641)*x[1]
        arg[2,3,1,0]=(0.812293906883)*x[0]**o+(-0.225593208768)*x[0]+(0.964175374163)*x[1]**o+(0.848850064926)*x[1]
        arg[2,3,1,1]=(0.376453572937)*x[0]**o+(0.994818812059)*x[0]+(0.58872018303)*x[1]**o+(0.68332473361)*x[1]
        arg[2,3,2,0]=(-0.00948421972234)*x[0]**o+(-0.680685462738)*x[0]+(-0.767076278167)*x[1]**o+(0.210165437284)*x[1]
        arg[2,3,2,1]=(-0.455199698263)*x[0]**o+(-0.763430401419)*x[0]+(0.183773350212)*x[1]**o+(0.43105940923)*x[1]
        arg[2,4,0,0]=(-0.673962025988)*x[0]**o+(-0.679073069725)*x[0]+(0.11018431715)*x[1]**o+(0.349020527482)*x[1]
        arg[2,4,0,1]=(-0.876691282076)*x[0]**o+(0.920902226631)*x[0]+(-0.477417470246)*x[1]**o+(0.836417982216)*x[1]
        arg[2,4,1,0]=(0.567966851024)*x[0]**o+(0.583241431716)*x[0]+(-0.271851271008)*x[1]**o+(0.678727902863)*x[1]
        arg[2,4,1,1]=(0.707400094)*x[0]**o+(0.802820856444)*x[0]+(0.84518937616)*x[1]**o+(-0.0267193533403)*x[1]
        arg[2,4,2,0]=(0.213268920036)*x[0]**o+(0.780721415328)*x[0]+(0.839292948154)*x[1]**o+(-0.0376334744443)*x[1]
        arg[2,4,2,1]=(0.136255687031)*x[0]**o+(-0.582766278693)*x[0]+(0.257662521321)*x[1]**o+(0.808337521979)*x[1]
        arg[3,0,0,0]=(-0.524068770822)*x[0]**o+(0.942236878842)*x[0]+(0.889496255793)*x[1]**o+(-0.904810742393)*x[1]
        arg[3,0,0,1]=(0.780157465475)*x[0]**o+(-0.763811668832)*x[0]+(-0.517482070639)*x[1]**o+(0.662770780676)*x[1]
        arg[3,0,1,0]=(-0.693227564812)*x[0]**o+(-0.763506211841)*x[0]+(-0.427709881126)*x[1]**o+(0.159109827204)*x[1]
        arg[3,0,1,1]=(0.644975629084)*x[0]**o+(-0.928112761899)*x[0]+(-0.530772740083)*x[1]**o+(-0.723174790087)*x[1]
        arg[3,0,2,0]=(-0.923467274097)*x[0]**o+(0.806128117288)*x[0]+(0.177779363207)*x[1]**o+(0.862243885542)*x[1]
        arg[3,0,2,1]=(-0.0823252901358)*x[0]**o+(0.289022916284)*x[0]+(-0.518508771021)*x[1]**o+(0.174564769845)*x[1]
        arg[3,1,0,0]=(0.608837578444)*x[0]**o+(0.898216758361)*x[0]+(-0.844730109761)*x[1]**o+(0.745974173207)*x[1]
        arg[3,1,0,1]=(-0.11776411078)*x[0]**o+(0.186256502356)*x[0]+(0.565289400333)*x[1]**o+(-0.776173716206)*x[1]
        arg[3,1,1,0]=(-0.170736580152)*x[0]**o+(0.829581661683)*x[0]+(0.310135833859)*x[1]**o+(-0.509629545687)*x[1]
        arg[3,1,1,1]=(-0.225246122751)*x[0]**o+(0.586754377294)*x[0]+(-0.726064725635)*x[1]**o+(0.136952605704)*x[1]
        arg[3,1,2,0]=(-0.287932115478)*x[0]**o+(0.858031396153)*x[0]+(-0.493441745353)*x[1]**o+(0.601827664994)*x[1]
        arg[3,1,2,1]=(0.665503842862)*x[0]**o+(-0.515998970452)*x[0]+(-0.890055558624)*x[1]**o+(-0.0446371451457)*x[1]
        arg[3,2,0,0]=(-0.33741557585)*x[0]**o+(-0.416183752333)*x[0]+(-0.92430073538)*x[1]**o+(-0.486265090904)*x[1]
        arg[3,2,0,1]=(0.910860925994)*x[0]**o+(-0.896869755389)*x[0]+(0.0634658185324)*x[1]**o+(-0.383136455739)*x[1]
        arg[3,2,1,0]=(0.386839297976)*x[0]**o+(-0.190659927071)*x[0]+(-0.304846978185)*x[1]**o+(-0.92705511899)*x[1]
        arg[3,2,1,1]=(-0.0197812018683)*x[0]**o+(-0.479083095775)*x[0]+(0.502489976089)*x[1]**o+(-0.859440665475)*x[1]
        arg[3,2,2,0]=(-0.507672045299)*x[0]**o+(0.498689676072)*x[0]+(-0.719735804645)*x[1]**o+(-0.522053306674)*x[1]
        arg[3,2,2,1]=(0.46207629211)*x[0]**o+(-0.496697066039)*x[0]+(0.994069339904)*x[1]**o+(0.873222754449)*x[1]
        arg[3,3,0,0]=(0.126410029692)*x[0]**o+(0.51067050196)*x[0]+(-0.978384824088)*x[1]**o+(-0.68606930503)*x[1]
        arg[3,3,0,1]=(0.376888157008)*x[0]**o+(-0.14015764025)*x[0]+(0.523667700758)*x[1]**o+(0.0277726855258)*x[1]
        arg[3,3,1,0]=(-0.105865233294)*x[0]**o+(-0.0515230741203)*x[0]+(0.203766525036)*x[1]**o+(-0.0431893014387)*x[1]
        arg[3,3,1,1]=(-0.404351448486)*x[0]**o+(-0.639161429157)*x[0]+(-0.957341091708)*x[1]**o+(-0.103346466984)*x[1]
        arg[3,3,2,0]=(-0.0523508780623)*x[0]**o+(-0.563589575544)*x[0]+(0.158487727325)*x[1]**o+(-0.519455492178)*x[1]
        arg[3,3,2,1]=(-0.823286342218)*x[0]**o+(-0.493663499535)*x[0]+(-0.409825574064)*x[1]**o+(-0.0406101279806)*x[1]
        arg[3,4,0,0]=(0.926212910387)*x[0]**o+(-0.0944507963)*x[0]+(0.640740499391)*x[1]**o+(-0.106969076047)*x[1]
        arg[3,4,0,1]=(-0.912044993941)*x[0]**o+(-0.495323777918)*x[0]+(-0.595831845108)*x[1]**o+(-0.313723426318)*x[1]
        arg[3,4,1,0]=(-0.0730898742548)*x[0]**o+(0.0524008484492)*x[0]+(-0.948232935497)*x[1]**o+(0.952670824572)*x[1]
        arg[3,4,1,1]=(-0.81869438138)*x[0]**o+(-0.894654266952)*x[0]+(0.414078621872)*x[1]**o+(-0.876690817847)*x[1]
        arg[3,4,2,0]=(0.00688155078441)*x[0]**o+(0.845964586551)*x[0]+(0.0183149371892)*x[1]**o+(-0.882931394537)*x[1]
        arg[3,4,2,1]=(-0.734830097149)*x[0]**o+(-0.37824588859)*x[0]+(-0.815402620311)*x[1]**o+(-0.357905685517)*x[1]
        ref[0,0,0,0]=(-0.282394950702)*x_ref[0]**o+(0.836580157584)*x_ref[0]+(0.337652262648)*x_ref[1]**o+(-0.442694263325)*x_ref[1]
        ref[0,0,0,1]=(0.85150774028)*x_ref[0]**o+(0.998297296593)*x_ref[0]+(-0.710029954342)*x_ref[1]**o+(0.0816168287837)*x_ref[1]
        ref[0,0,1,0]=(-0.636435102838)*x_ref[0]**o+(-0.613479863553)*x_ref[0]+(-0.861506513168)*x_ref[1]**o+(-0.242998961478)*x_ref[1]
        ref[0,0,1,1]=(-0.00671863978521)*x_ref[0]**o+(0.909404084068)*x_ref[0]+(-0.734956067026)*x_ref[1]**o+(0.256530538264)*x_ref[1]
        ref[0,0,2,0]=(-0.646631608999)*x_ref[0]**o+(-0.521130455843)*x_ref[0]+(-0.44328404557)*x_ref[1]**o+(-0.156860325817)*x_ref[1]
        ref[0,0,2,1]=(-0.817550008949)*x_ref[0]**o+(0.702414497795)*x_ref[0]+(0.946449680763)*x_ref[1]**o+(-0.61576361925)*x_ref[1]
        ref[0,1,0,0]=(-0.527206691595)*x_ref[0]**o+(-0.43469216984)*x_ref[0]+(-0.108455901154)*x_ref[1]**o+(0.435646962156)*x_ref[1]
        ref[0,1,0,1]=(-0.148522219535)*x_ref[0]**o+(0.146912249755)*x_ref[0]+(-0.414934945665)*x_ref[1]**o+(-0.422754988714)*x_ref[1]
        ref[0,1,1,0]=(0.177760972947)*x_ref[0]**o+(-0.938522530862)*x_ref[0]+(0.566229652264)*x_ref[1]**o+(0.255269743628)*x_ref[1]
        ref[0,1,1,1]=(-0.317807933792)*x_ref[0]**o+(0.432181451246)*x_ref[0]+(-0.358821513362)*x_ref[1]**o+(-0.470517232585)*x_ref[1]
        ref[0,1,2,0]=(0.486536430227)*x_ref[0]**o+(-0.533653809349)*x_ref[0]+(-0.126612514045)*x_ref[1]**o+(0.103088115487)*x_ref[1]
        ref[0,1,2,1]=(-0.995976559589)*x_ref[0]**o+(0.965537167421)*x_ref[0]+(-0.351307699531)*x_ref[1]**o+(-0.684367434666)*x_ref[1]
        ref[0,2,0,0]=(0.976008613768)*x_ref[0]**o+(-0.890573007002)*x_ref[0]+(-0.440709423944)*x_ref[1]**o+(0.315724459367)*x_ref[1]
        ref[0,2,0,1]=(-0.595275811962)*x_ref[0]**o+(-0.599634144509)*x_ref[0]+(0.646466002346)*x_ref[1]**o+(-0.94529568651)*x_ref[1]
        ref[0,2,1,0]=(0.977691521821)*x_ref[0]**o+(-0.0730997643682)*x_ref[0]+(-0.338897498737)*x_ref[1]**o+(0.246308678824)*x_ref[1]
        ref[0,2,1,1]=(-0.111317410732)*x_ref[0]**o+(-0.574381793398)*x_ref[0]+(-0.0314431899843)*x_ref[1]**o+(0.393105047096)*x_ref[1]
        ref[0,2,2,0]=(0.886320020757)*x_ref[0]**o+(-0.785042003544)*x_ref[0]+(0.82609269979)*x_ref[1]**o+(-0.677169484337)*x_ref[1]
        ref[0,2,2,1]=(0.811027138079)*x_ref[0]**o+(-0.375108470948)*x_ref[0]+(0.781472464253)*x_ref[1]**o+(0.217640794143)*x_ref[1]
        ref[0,3,0,0]=(0.264884117091)*x_ref[0]**o+(0.116591319557)*x_ref[0]+(0.282667679763)*x_ref[1]**o+(0.111123454337)*x_ref[1]
        ref[0,3,0,1]=(0.0451626985922)*x_ref[0]**o+(-0.93771279728)*x_ref[0]+(0.129171764072)*x_ref[1]**o+(0.598367954162)*x_ref[1]
        ref[0,3,1,0]=(0.905243511086)*x_ref[0]**o+(0.0837062573586)*x_ref[0]+(0.0657223248433)*x_ref[1]**o+(-0.746644690072)*x_ref[1]
        ref[0,3,1,1]=(-0.0547856755485)*x_ref[0]**o+(-0.22157806755)*x_ref[0]+(0.124623625108)*x_ref[1]**o+(0.0436590157081)*x_ref[1]
        ref[0,3,2,0]=(-0.558323744577)*x_ref[0]**o+(-0.646043519007)*x_ref[0]+(0.991885458004)*x_ref[1]**o+(-0.634140808163)*x_ref[1]
        ref[0,3,2,1]=(-0.27247983028)*x_ref[0]**o+(0.803906404485)*x_ref[0]+(-0.816826609924)*x_ref[1]**o+(-0.969532783336)*x_ref[1]
        ref[0,4,0,0]=(0.638733842891)*x_ref[0]**o+(-0.972793270208)*x_ref[0]+(0.714611008099)*x_ref[1]**o+(0.586988595847)*x_ref[1]
        ref[0,4,0,1]=(0.63584760734)*x_ref[0]**o+(-0.653740236747)*x_ref[0]+(-0.596304731446)*x_ref[1]**o+(0.42519786429)*x_ref[1]
        ref[0,4,1,0]=(0.0665562973529)*x_ref[0]**o+(0.362152092093)*x_ref[0]+(0.732530335739)*x_ref[1]**o+(0.618536326941)*x_ref[1]
        ref[0,4,1,1]=(0.486060667608)*x_ref[0]**o+(0.59489013279)*x_ref[0]+(-0.547237017498)*x_ref[1]**o+(0.0371894705857)*x_ref[1]
        ref[0,4,2,0]=(0.395505844397)*x_ref[0]**o+(-0.869090434535)*x_ref[0]+(-0.606640827437)*x_ref[1]**o+(0.269995459083)*x_ref[1]
        ref[0,4,2,1]=(0.304671547176)*x_ref[0]**o+(-0.515772523849)*x_ref[0]+(0.546114300876)*x_ref[1]**o+(0.530247404341)*x_ref[1]
        ref[1,0,0,0]=(0.408332677433)*x_ref[0]**o+(-0.0778807359918)*x_ref[0]+(-0.803337489297)*x_ref[1]**o+(-0.123835718422)*x_ref[1]
        ref[1,0,0,1]=(-0.00441620254944)*x_ref[0]**o+(0.0321259598831)*x_ref[0]+(-0.998381508472)*x_ref[1]**o+(0.800767747511)*x_ref[1]
        ref[1,0,1,0]=(-0.923683057418)*x_ref[0]**o+(0.579405114902)*x_ref[0]+(-0.939639823585)*x_ref[1]**o+(-0.138423358134)*x_ref[1]
        ref[1,0,1,1]=(-0.558420234848)*x_ref[0]**o+(0.133316569813)*x_ref[0]+(-0.294134717329)*x_ref[1]**o+(-0.750897749044)*x_ref[1]
        ref[1,0,2,0]=(0.478951298545)*x_ref[0]**o+(-0.807173978799)*x_ref[0]+(0.852438146763)*x_ref[1]**o+(-0.103814229563)*x_ref[1]
        ref[1,0,2,1]=(-0.875250726603)*x_ref[0]**o+(-0.934803428491)*x_ref[0]+(0.143352873672)*x_ref[1]**o+(0.600934881206)*x_ref[1]
        ref[1,1,0,0]=(-0.0539476321018)*x_ref[0]**o+(0.194177663417)*x_ref[0]+(-0.862877991009)*x_ref[1]**o+(-0.438337993664)*x_ref[1]
        ref[1,1,0,1]=(-0.623379355474)*x_ref[0]**o+(-0.0746994739748)*x_ref[0]+(0.365005438956)*x_ref[1]**o+(-0.431737688205)*x_ref[1]
        ref[1,1,1,0]=(0.0936054461735)*x_ref[0]**o+(-0.974594269446)*x_ref[0]+(-0.121917598428)*x_ref[1]**o+(-0.84429083727)*x_ref[1]
        ref[1,1,1,1]=(0.967763348672)*x_ref[0]**o+(0.331698826475)*x_ref[0]+(0.66448515877)*x_ref[1]**o+(0.00214985930661)*x_ref[1]
        ref[1,1,2,0]=(-0.312146580762)*x_ref[0]**o+(0.998152287706)*x_ref[0]+(-0.474242942777)*x_ref[1]**o+(0.637827347846)*x_ref[1]
        ref[1,1,2,1]=(-0.556765443436)*x_ref[0]**o+(0.227327929774)*x_ref[0]+(0.739457845569)*x_ref[1]**o+(-0.615070017663)*x_ref[1]
        ref[1,2,0,0]=(-0.23030660552)*x_ref[0]**o+(0.750479465953)*x_ref[0]+(0.278215408392)*x_ref[1]**o+(-0.640618698224)*x_ref[1]
        ref[1,2,0,1]=(-0.0713217101129)*x_ref[0]**o+(0.500737477887)*x_ref[0]+(-0.972834770005)*x_ref[1]**o+(-0.420386056033)*x_ref[1]
        ref[1,2,1,0]=(-0.497322989847)*x_ref[0]**o+(-0.528685124841)*x_ref[0]+(-0.934386348496)*x_ref[1]**o+(0.191930553839)*x_ref[1]
        ref[1,2,1,1]=(0.124362419781)*x_ref[0]**o+(-0.0457849787456)*x_ref[0]+(-0.360596154712)*x_ref[1]**o+(0.799736253715)*x_ref[1]
        ref[1,2,2,0]=(-0.346294878893)*x_ref[0]**o+(0.427772309361)*x_ref[0]+(-0.0642666791012)*x_ref[1]**o+(-0.214442238311)*x_ref[1]
        ref[1,2,2,1]=(0.24434837411)*x_ref[0]**o+(-0.913373159786)*x_ref[0]+(0.158750862391)*x_ref[1]**o+(-0.544328807576)*x_ref[1]
        ref[1,3,0,0]=(0.870847002687)*x_ref[0]**o+(-0.466884462301)*x_ref[0]+(-0.343358829445)*x_ref[1]**o+(-0.180956419766)*x_ref[1]
        ref[1,3,0,1]=(-0.813518850863)*x_ref[0]**o+(0.960810467501)*x_ref[0]+(0.35091416904)*x_ref[1]**o+(0.84534544948)*x_ref[1]
        ref[1,3,1,0]=(-0.284687751218)*x_ref[0]**o+(-0.274583519355)*x_ref[0]+(0.338892764102)*x_ref[1]**o+(-0.145142443753)*x_ref[1]
        ref[1,3,1,1]=(0.921064262137)*x_ref[0]**o+(0.152552421678)*x_ref[0]+(0.827247693375)*x_ref[1]**o+(-0.569529848619)*x_ref[1]
        ref[1,3,2,0]=(0.133567630356)*x_ref[0]**o+(-0.0429397358364)*x_ref[0]+(-0.175383170663)*x_ref[1]**o+(0.584407631334)*x_ref[1]
        ref[1,3,2,1]=(0.203880167371)*x_ref[0]**o+(-0.422978073825)*x_ref[0]+(-0.210109267588)*x_ref[1]**o+(-0.902327373219)*x_ref[1]
        ref[1,4,0,0]=(-0.566386864876)*x_ref[0]**o+(-0.139795085483)*x_ref[0]+(-0.00852913464238)*x_ref[1]**o+(0.569151381884)*x_ref[1]
        ref[1,4,0,1]=(-0.937256625212)*x_ref[0]**o+(0.563438253128)*x_ref[0]+(-0.586182238715)*x_ref[1]**o+(-0.566169678723)*x_ref[1]
        ref[1,4,1,0]=(-0.73296890712)*x_ref[0]**o+(0.359249594406)*x_ref[0]+(0.66050790616)*x_ref[1]**o+(-0.367919797426)*x_ref[1]
        ref[1,4,1,1]=(-0.81165734248)*x_ref[0]**o+(0.460592817838)*x_ref[0]+(0.953573909492)*x_ref[1]**o+(0.329659021274)*x_ref[1]
        ref[1,4,2,0]=(0.345994483595)*x_ref[0]**o+(-0.18547324343)*x_ref[0]+(0.669452926238)*x_ref[1]**o+(0.42513249272)*x_ref[1]
        ref[1,4,2,1]=(0.250677642616)*x_ref[0]**o+(0.466118496803)*x_ref[0]+(-0.776608543595)*x_ref[1]**o+(0.449144243682)*x_ref[1]
        ref[2,0,0,0]=(0.244172909021)*x_ref[0]**o+(0.0311068322898)*x_ref[0]+(-0.909348039508)*x_ref[1]**o+(0.376500719414)*x_ref[1]
        ref[2,0,0,1]=(-0.033516640667)*x_ref[0]**o+(0.698424718328)*x_ref[0]+(-0.282814785542)*x_ref[1]**o+(0.61118022947)*x_ref[1]
        ref[2,0,1,0]=(-0.946672649538)*x_ref[0]**o+(0.814325244591)*x_ref[0]+(-0.353215562825)*x_ref[1]**o+(-0.138150257429)*x_ref[1]
        ref[2,0,1,1]=(-0.886509051692)*x_ref[0]**o+(-0.0468334385819)*x_ref[0]+(0.71984719189)*x_ref[1]**o+(-0.324744971493)*x_ref[1]
        ref[2,0,2,0]=(-0.780921647053)*x_ref[0]**o+(-0.401418496932)*x_ref[0]+(0.647847274332)*x_ref[1]**o+(-0.831834322278)*x_ref[1]
        ref[2,0,2,1]=(0.835155004862)*x_ref[0]**o+(0.684459986902)*x_ref[0]+(-0.301244240677)*x_ref[1]**o+(-0.161292780187)*x_ref[1]
        ref[2,1,0,0]=(-0.652199398394)*x_ref[0]**o+(0.62480722178)*x_ref[0]+(-0.474535712914)*x_ref[1]**o+(0.926061797766)*x_ref[1]
        ref[2,1,0,1]=(-0.813365478568)*x_ref[0]**o+(-0.38795792417)*x_ref[0]+(0.378161353697)*x_ref[1]**o+(0.70581806211)*x_ref[1]
        ref[2,1,1,0]=(0.729265594934)*x_ref[0]**o+(-0.469397842359)*x_ref[0]+(0.642472199926)*x_ref[1]**o+(-0.779527283011)*x_ref[1]
        ref[2,1,1,1]=(0.0517342917574)*x_ref[0]**o+(-0.414111007917)*x_ref[0]+(-0.89437375241)*x_ref[1]**o+(-0.830119585842)*x_ref[1]
        ref[2,1,2,0]=(0.769898486257)*x_ref[0]**o+(0.791725624832)*x_ref[0]+(0.297519329396)*x_ref[1]**o+(-0.959462006611)*x_ref[1]
        ref[2,1,2,1]=(0.817802144251)*x_ref[0]**o+(0.69851258911)*x_ref[0]+(-0.174401735228)*x_ref[1]**o+(0.13449696498)*x_ref[1]
        ref[2,2,0,0]=(0.547972662256)*x_ref[0]**o+(-0.66491929979)*x_ref[0]+(0.330783041672)*x_ref[1]**o+(0.816196730606)*x_ref[1]
        ref[2,2,0,1]=(0.245971097763)*x_ref[0]**o+(0.213087481481)*x_ref[0]+(-0.840766291392)*x_ref[1]**o+(0.388414557376)*x_ref[1]
        ref[2,2,1,0]=(0.523151676444)*x_ref[0]**o+(0.854817486027)*x_ref[0]+(-0.53570188528)*x_ref[1]**o+(-0.68149362242)*x_ref[1]
        ref[2,2,1,1]=(-0.984200247787)*x_ref[0]**o+(0.83324365092)*x_ref[0]+(-0.896184408816)*x_ref[1]**o+(0.102276243487)*x_ref[1]
        ref[2,2,2,0]=(-0.216357164308)*x_ref[0]**o+(-0.597984882107)*x_ref[0]+(-0.205435921511)*x_ref[1]**o+(0.484947652759)*x_ref[1]
        ref[2,2,2,1]=(-0.650046253868)*x_ref[0]**o+(-0.624878186105)*x_ref[0]+(0.786519914406)*x_ref[1]**o+(0.28027273095)*x_ref[1]
        ref[2,3,0,0]=(-0.0821718346579)*x_ref[0]**o+(-0.687049917467)*x_ref[0]+(0.380792333736)*x_ref[1]**o+(0.151185837715)*x_ref[1]
        ref[2,3,0,1]=(0.61796113854)*x_ref[0]**o+(0.844624413065)*x_ref[0]+(-0.980619401098)*x_ref[1]**o+(-0.854924862641)*x_ref[1]
        ref[2,3,1,0]=(0.812293906883)*x_ref[0]**o+(-0.225593208768)*x_ref[0]+(0.964175374163)*x_ref[1]**o+(0.848850064926)*x_ref[1]
        ref[2,3,1,1]=(0.376453572937)*x_ref[0]**o+(0.994818812059)*x_ref[0]+(0.58872018303)*x_ref[1]**o+(0.68332473361)*x_ref[1]
        ref[2,3,2,0]=(-0.00948421972234)*x_ref[0]**o+(-0.680685462738)*x_ref[0]+(-0.767076278167)*x_ref[1]**o+(0.210165437284)*x_ref[1]
        ref[2,3,2,1]=(-0.455199698263)*x_ref[0]**o+(-0.763430401419)*x_ref[0]+(0.183773350212)*x_ref[1]**o+(0.43105940923)*x_ref[1]
        ref[2,4,0,0]=(-0.673962025988)*x_ref[0]**o+(-0.679073069725)*x_ref[0]+(0.11018431715)*x_ref[1]**o+(0.349020527482)*x_ref[1]
        ref[2,4,0,1]=(-0.876691282076)*x_ref[0]**o+(0.920902226631)*x_ref[0]+(-0.477417470246)*x_ref[1]**o+(0.836417982216)*x_ref[1]
        ref[2,4,1,0]=(0.567966851024)*x_ref[0]**o+(0.583241431716)*x_ref[0]+(-0.271851271008)*x_ref[1]**o+(0.678727902863)*x_ref[1]
        ref[2,4,1,1]=(0.707400094)*x_ref[0]**o+(0.802820856444)*x_ref[0]+(0.84518937616)*x_ref[1]**o+(-0.0267193533403)*x_ref[1]
        ref[2,4,2,0]=(0.213268920036)*x_ref[0]**o+(0.780721415328)*x_ref[0]+(0.839292948154)*x_ref[1]**o+(-0.0376334744443)*x_ref[1]
        ref[2,4,2,1]=(0.136255687031)*x_ref[0]**o+(-0.582766278693)*x_ref[0]+(0.257662521321)*x_ref[1]**o+(0.808337521979)*x_ref[1]
        ref[3,0,0,0]=(-0.524068770822)*x_ref[0]**o+(0.942236878842)*x_ref[0]+(0.889496255793)*x_ref[1]**o+(-0.904810742393)*x_ref[1]
        ref[3,0,0,1]=(0.780157465475)*x_ref[0]**o+(-0.763811668832)*x_ref[0]+(-0.517482070639)*x_ref[1]**o+(0.662770780676)*x_ref[1]
        ref[3,0,1,0]=(-0.693227564812)*x_ref[0]**o+(-0.763506211841)*x_ref[0]+(-0.427709881126)*x_ref[1]**o+(0.159109827204)*x_ref[1]
        ref[3,0,1,1]=(0.644975629084)*x_ref[0]**o+(-0.928112761899)*x_ref[0]+(-0.530772740083)*x_ref[1]**o+(-0.723174790087)*x_ref[1]
        ref[3,0,2,0]=(-0.923467274097)*x_ref[0]**o+(0.806128117288)*x_ref[0]+(0.177779363207)*x_ref[1]**o+(0.862243885542)*x_ref[1]
        ref[3,0,2,1]=(-0.0823252901358)*x_ref[0]**o+(0.289022916284)*x_ref[0]+(-0.518508771021)*x_ref[1]**o+(0.174564769845)*x_ref[1]
        ref[3,1,0,0]=(0.608837578444)*x_ref[0]**o+(0.898216758361)*x_ref[0]+(-0.844730109761)*x_ref[1]**o+(0.745974173207)*x_ref[1]
        ref[3,1,0,1]=(-0.11776411078)*x_ref[0]**o+(0.186256502356)*x_ref[0]+(0.565289400333)*x_ref[1]**o+(-0.776173716206)*x_ref[1]
        ref[3,1,1,0]=(-0.170736580152)*x_ref[0]**o+(0.829581661683)*x_ref[0]+(0.310135833859)*x_ref[1]**o+(-0.509629545687)*x_ref[1]
        ref[3,1,1,1]=(-0.225246122751)*x_ref[0]**o+(0.586754377294)*x_ref[0]+(-0.726064725635)*x_ref[1]**o+(0.136952605704)*x_ref[1]
        ref[3,1,2,0]=(-0.287932115478)*x_ref[0]**o+(0.858031396153)*x_ref[0]+(-0.493441745353)*x_ref[1]**o+(0.601827664994)*x_ref[1]
        ref[3,1,2,1]=(0.665503842862)*x_ref[0]**o+(-0.515998970452)*x_ref[0]+(-0.890055558624)*x_ref[1]**o+(-0.0446371451457)*x_ref[1]
        ref[3,2,0,0]=(-0.33741557585)*x_ref[0]**o+(-0.416183752333)*x_ref[0]+(-0.92430073538)*x_ref[1]**o+(-0.486265090904)*x_ref[1]
        ref[3,2,0,1]=(0.910860925994)*x_ref[0]**o+(-0.896869755389)*x_ref[0]+(0.0634658185324)*x_ref[1]**o+(-0.383136455739)*x_ref[1]
        ref[3,2,1,0]=(0.386839297976)*x_ref[0]**o+(-0.190659927071)*x_ref[0]+(-0.304846978185)*x_ref[1]**o+(-0.92705511899)*x_ref[1]
        ref[3,2,1,1]=(-0.0197812018683)*x_ref[0]**o+(-0.479083095775)*x_ref[0]+(0.502489976089)*x_ref[1]**o+(-0.859440665475)*x_ref[1]
        ref[3,2,2,0]=(-0.507672045299)*x_ref[0]**o+(0.498689676072)*x_ref[0]+(-0.719735804645)*x_ref[1]**o+(-0.522053306674)*x_ref[1]
        ref[3,2,2,1]=(0.46207629211)*x_ref[0]**o+(-0.496697066039)*x_ref[0]+(0.994069339904)*x_ref[1]**o+(0.873222754449)*x_ref[1]
        ref[3,3,0,0]=(0.126410029692)*x_ref[0]**o+(0.51067050196)*x_ref[0]+(-0.978384824088)*x_ref[1]**o+(-0.68606930503)*x_ref[1]
        ref[3,3,0,1]=(0.376888157008)*x_ref[0]**o+(-0.14015764025)*x_ref[0]+(0.523667700758)*x_ref[1]**o+(0.0277726855258)*x_ref[1]
        ref[3,3,1,0]=(-0.105865233294)*x_ref[0]**o+(-0.0515230741203)*x_ref[0]+(0.203766525036)*x_ref[1]**o+(-0.0431893014387)*x_ref[1]
        ref[3,3,1,1]=(-0.404351448486)*x_ref[0]**o+(-0.639161429157)*x_ref[0]+(-0.957341091708)*x_ref[1]**o+(-0.103346466984)*x_ref[1]
        ref[3,3,2,0]=(-0.0523508780623)*x_ref[0]**o+(-0.563589575544)*x_ref[0]+(0.158487727325)*x_ref[1]**o+(-0.519455492178)*x_ref[1]
        ref[3,3,2,1]=(-0.823286342218)*x_ref[0]**o+(-0.493663499535)*x_ref[0]+(-0.409825574064)*x_ref[1]**o+(-0.0406101279806)*x_ref[1]
        ref[3,4,0,0]=(0.926212910387)*x_ref[0]**o+(-0.0944507963)*x_ref[0]+(0.640740499391)*x_ref[1]**o+(-0.106969076047)*x_ref[1]
        ref[3,4,0,1]=(-0.912044993941)*x_ref[0]**o+(-0.495323777918)*x_ref[0]+(-0.595831845108)*x_ref[1]**o+(-0.313723426318)*x_ref[1]
        ref[3,4,1,0]=(-0.0730898742548)*x_ref[0]**o+(0.0524008484492)*x_ref[0]+(-0.948232935497)*x_ref[1]**o+(0.952670824572)*x_ref[1]
        ref[3,4,1,1]=(-0.81869438138)*x_ref[0]**o+(-0.894654266952)*x_ref[0]+(0.414078621872)*x_ref[1]**o+(-0.876690817847)*x_ref[1]
        ref[3,4,2,0]=(0.00688155078441)*x_ref[0]**o+(0.845964586551)*x_ref[0]+(0.0183149371892)*x_ref[1]**o+(-0.882931394537)*x_ref[1]
        ref[3,4,2,1]=(-0.734830097149)*x_ref[0]**o+(-0.37824588859)*x_ref[0]+(-0.815402620311)*x_ref[1]**o+(-0.357905685517)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.110519315354)*x[0]**o+(0.904227374292)*x[0]+(0.459005710707)*x[1]**o+(-0.980231422256)*x[1]+(-0.819713355475)*x[2]**o+(-0.93459517989)*x[2]
        arg[0,0,0,1]=(-0.326902946413)*x[0]**o+(0.489327766219)*x[0]+(-0.965308741061)*x[1]**o+(0.0351399111471)*x[1]+(0.257760096595)*x[2]**o+(0.125123023244)*x[2]
        arg[0,0,1,0]=(-0.584360710004)*x[0]**o+(0.681892732544)*x[0]+(-0.118257741614)*x[1]**o+(0.839875738892)*x[1]+(0.107934154241)*x[2]**o+(0.279757054665)*x[2]
        arg[0,0,1,1]=(0.90973214794)*x[0]**o+(-0.776804618826)*x[0]+(-0.910633853381)*x[1]**o+(-0.25678788529)*x[1]+(0.783709363646)*x[2]**o+(-0.922660855915)*x[2]
        arg[0,0,2,0]=(-0.802273263394)*x[0]**o+(-0.917749665952)*x[0]+(0.796206817226)*x[1]**o+(0.917284637885)*x[1]+(0.219841508893)*x[2]**o+(0.0655848200083)*x[2]
        arg[0,0,2,1]=(0.550596792343)*x[0]**o+(-0.760207057617)*x[0]+(0.0368469683098)*x[1]**o+(0.509347717331)*x[1]+(0.813780778487)*x[2]**o+(-0.432656298451)*x[2]
        arg[0,1,0,0]=(-0.238502042941)*x[0]**o+(-0.966632221032)*x[0]+(-0.373745652933)*x[1]**o+(-0.436176336389)*x[1]+(0.124280197699)*x[2]**o+(0.552700562803)*x[2]
        arg[0,1,0,1]=(-0.19848053276)*x[0]**o+(0.110167634957)*x[0]+(-0.748553857473)*x[1]**o+(0.355932508959)*x[1]+(-0.226987914753)*x[2]**o+(-0.68870238197)*x[2]
        arg[0,1,1,0]=(0.365822518882)*x[0]**o+(0.327285102897)*x[0]+(0.549212504635)*x[1]**o+(0.918179946504)*x[1]+(0.20644004679)*x[2]**o+(0.821299409756)*x[2]
        arg[0,1,1,1]=(-0.366167489101)*x[0]**o+(0.529933186572)*x[0]+(-0.604965310967)*x[1]**o+(-0.833093092352)*x[1]+(-0.73593913912)*x[2]**o+(0.872325614453)*x[2]
        arg[0,1,2,0]=(0.289123646902)*x[0]**o+(0.705862153789)*x[0]+(-0.319756886822)*x[1]**o+(-0.453220232524)*x[1]+(-0.192336374167)*x[2]**o+(0.363575495006)*x[2]
        arg[0,1,2,1]=(-0.213014833353)*x[0]**o+(-0.0706784230627)*x[0]+(-0.941272982229)*x[1]**o+(-0.96641361036)*x[1]+(0.289919014964)*x[2]**o+(0.57658772066)*x[2]
        arg[0,2,0,0]=(-0.406683991704)*x[0]**o+(-0.0110042213592)*x[0]+(-0.273148083258)*x[1]**o+(0.054257030447)*x[1]+(0.387598147681)*x[2]**o+(0.480398109069)*x[2]
        arg[0,2,0,1]=(0.318713057562)*x[0]**o+(-0.449467749585)*x[0]+(0.669991267829)*x[1]**o+(0.237986860249)*x[1]+(0.894370523471)*x[2]**o+(-0.397832436233)*x[2]
        arg[0,2,1,0]=(-0.858603253642)*x[0]**o+(-0.713122015919)*x[0]+(0.870251906188)*x[1]**o+(-0.781601254578)*x[1]+(-0.504531525822)*x[2]**o+(-0.699819349556)*x[2]
        arg[0,2,1,1]=(-0.825591116361)*x[0]**o+(-0.552452876394)*x[0]+(-0.262627261618)*x[1]**o+(-0.147127256184)*x[1]+(-0.171485665103)*x[2]**o+(0.823279741501)*x[2]
        arg[0,2,2,0]=(-0.0416279689068)*x[0]**o+(-0.0903338000001)*x[0]+(0.911673780635)*x[1]**o+(0.86298673049)*x[1]+(-0.60251179847)*x[2]**o+(0.378303877746)*x[2]
        arg[0,2,2,1]=(0.261384106387)*x[0]**o+(0.0326416199811)*x[0]+(-0.426740102269)*x[1]**o+(-0.445401626921)*x[1]+(-0.388668254526)*x[2]**o+(-0.977942913284)*x[2]
        arg[0,3,0,0]=(-0.0195053813483)*x[0]**o+(-0.294061998664)*x[0]+(0.790461098426)*x[1]**o+(-0.506334268838)*x[1]+(0.394015541605)*x[2]**o+(-0.562248105432)*x[2]
        arg[0,3,0,1]=(0.390726516735)*x[0]**o+(-0.757597437859)*x[0]+(0.861882149861)*x[1]**o+(0.621870451314)*x[1]+(0.900777666162)*x[2]**o+(0.271507131938)*x[2]
        arg[0,3,1,0]=(-0.725635966475)*x[0]**o+(-0.695837548737)*x[0]+(-0.163078176328)*x[1]**o+(-0.0435808418729)*x[1]+(0.0890756544611)*x[2]**o+(-0.270422627949)*x[2]
        arg[0,3,1,1]=(-0.175498720647)*x[0]**o+(-0.213605441632)*x[0]+(-0.358632620362)*x[1]**o+(-0.193273004152)*x[1]+(0.937704823805)*x[2]**o+(0.861707832291)*x[2]
        arg[0,3,2,0]=(0.0976010265611)*x[0]**o+(0.995135925341)*x[0]+(0.755339287387)*x[1]**o+(-0.924722068193)*x[1]+(-0.480212181904)*x[2]**o+(-0.40237760439)*x[2]
        arg[0,3,2,1]=(-0.679215392952)*x[0]**o+(-0.00796096253297)*x[0]+(-0.964972599421)*x[1]**o+(-0.410415435656)*x[1]+(-0.292638067785)*x[2]**o+(-0.950383976561)*x[2]
        arg[0,4,0,0]=(0.245586594928)*x[0]**o+(-0.593420401896)*x[0]+(0.553450486534)*x[1]**o+(0.173814965689)*x[1]+(0.0719392610563)*x[2]**o+(0.676231437486)*x[2]
        arg[0,4,0,1]=(0.154382756425)*x[0]**o+(-0.0655740293432)*x[0]+(0.924689745142)*x[1]**o+(-0.246876531175)*x[1]+(-0.241460713543)*x[2]**o+(-0.489840108568)*x[2]
        arg[0,4,1,0]=(-0.482705840359)*x[0]**o+(-0.392960380114)*x[0]+(-0.503463823972)*x[1]**o+(0.317163204738)*x[1]+(0.775048877356)*x[2]**o+(0.599889732326)*x[2]
        arg[0,4,1,1]=(-0.122555861716)*x[0]**o+(0.930175770449)*x[0]+(0.483503764941)*x[1]**o+(0.340001033217)*x[1]+(0.672924280979)*x[2]**o+(0.746489359683)*x[2]
        arg[0,4,2,0]=(-0.339421141033)*x[0]**o+(-0.604966575169)*x[0]+(0.862413017648)*x[1]**o+(0.936752078454)*x[1]+(-0.0549628353285)*x[2]**o+(-0.0858885835314)*x[2]
        arg[0,4,2,1]=(-0.581767357421)*x[0]**o+(-0.644286027548)*x[0]+(0.76948396865)*x[1]**o+(0.692265904744)*x[1]+(0.955871091531)*x[2]**o+(-0.378888503457)*x[2]
        arg[1,0,0,0]=(-0.0407846595046)*x[0]**o+(-0.293709943251)*x[0]+(-0.310552931582)*x[1]**o+(-0.696244178466)*x[1]+(0.962985284252)*x[2]**o+(-0.0266457621404)*x[2]
        arg[1,0,0,1]=(-0.480942811584)*x[0]**o+(-0.400158759679)*x[0]+(0.311171869807)*x[1]**o+(-0.79518809427)*x[1]+(-0.497004450478)*x[2]**o+(-0.639822005752)*x[2]
        arg[1,0,1,0]=(0.158738648483)*x[0]**o+(0.274859816468)*x[0]+(-0.208677612654)*x[1]**o+(0.911107598366)*x[1]+(-0.400163007778)*x[2]**o+(0.276296890267)*x[2]
        arg[1,0,1,1]=(0.613312276864)*x[0]**o+(-0.644237218776)*x[0]+(0.81927272895)*x[1]**o+(-0.12971860556)*x[1]+(0.214379406799)*x[2]**o+(0.856335675043)*x[2]
        arg[1,0,2,0]=(-0.699925173766)*x[0]**o+(-0.560613547952)*x[0]+(-0.355325107637)*x[1]**o+(-0.598747822041)*x[1]+(0.243087037553)*x[2]**o+(-0.484176993015)*x[2]
        arg[1,0,2,1]=(-0.218912300848)*x[0]**o+(-0.259483312969)*x[0]+(-0.863788186295)*x[1]**o+(0.924423774757)*x[1]+(-0.0823752192204)*x[2]**o+(-0.928377900602)*x[2]
        arg[1,1,0,0]=(-0.368901144425)*x[0]**o+(0.544104655592)*x[0]+(0.321295151176)*x[1]**o+(0.395692754134)*x[1]+(-0.721319111754)*x[2]**o+(-0.698518442137)*x[2]
        arg[1,1,0,1]=(-0.592867336047)*x[0]**o+(0.145120231038)*x[0]+(0.110390771366)*x[1]**o+(0.168333368086)*x[1]+(-0.577866419921)*x[2]**o+(0.705027708469)*x[2]
        arg[1,1,1,0]=(-0.210603428235)*x[0]**o+(-0.250156447116)*x[0]+(-0.397803248976)*x[1]**o+(0.332163380867)*x[1]+(-0.250648216541)*x[2]**o+(-0.0693729055697)*x[2]
        arg[1,1,1,1]=(0.324927558096)*x[0]**o+(-0.262726116223)*x[0]+(-0.555439758488)*x[1]**o+(-0.865019248609)*x[1]+(-0.0971907425857)*x[2]**o+(-0.742359236711)*x[2]
        arg[1,1,2,0]=(0.161058869414)*x[0]**o+(0.349191500791)*x[0]+(0.468336473315)*x[1]**o+(-0.856364775995)*x[1]+(0.491736696689)*x[2]**o+(0.352984104173)*x[2]
        arg[1,1,2,1]=(-0.243275976508)*x[0]**o+(-0.785605388878)*x[0]+(-0.282478919296)*x[1]**o+(-0.728851522341)*x[1]+(0.375134848237)*x[2]**o+(-0.486200561256)*x[2]
        arg[1,2,0,0]=(-0.172241451405)*x[0]**o+(0.422361498196)*x[0]+(0.464862755164)*x[1]**o+(0.466661717105)*x[1]+(-0.886209677512)*x[2]**o+(-0.43529076789)*x[2]
        arg[1,2,0,1]=(0.946582222405)*x[0]**o+(0.513640973894)*x[0]+(-0.923788000633)*x[1]**o+(-0.358163567195)*x[1]+(-0.709999132129)*x[2]**o+(-0.761163299694)*x[2]
        arg[1,2,1,0]=(0.236339981889)*x[0]**o+(0.259874866134)*x[0]+(-0.32460762691)*x[1]**o+(-0.629786950313)*x[1]+(-0.679261685402)*x[2]**o+(0.593069192918)*x[2]
        arg[1,2,1,1]=(0.11006800009)*x[0]**o+(0.970532991787)*x[0]+(-0.467608188361)*x[1]**o+(-0.959281984552)*x[1]+(0.610396622519)*x[2]**o+(0.282468600058)*x[2]
        arg[1,2,2,0]=(-0.656419157742)*x[0]**o+(0.234027308171)*x[0]+(-0.67027206978)*x[1]**o+(-0.746034436817)*x[1]+(-0.771894073652)*x[2]**o+(0.949545830287)*x[2]
        arg[1,2,2,1]=(0.615884236811)*x[0]**o+(-0.670156612555)*x[0]+(0.543165493345)*x[1]**o+(-0.618294277594)*x[1]+(-0.974114255061)*x[2]**o+(-0.762320542298)*x[2]
        arg[1,3,0,0]=(0.096201957648)*x[0]**o+(0.440870816208)*x[0]+(0.124448859607)*x[1]**o+(0.507919407561)*x[1]+(-0.401666263687)*x[2]**o+(-0.437653713723)*x[2]
        arg[1,3,0,1]=(0.165129229294)*x[0]**o+(-0.353217428955)*x[0]+(0.473666065504)*x[1]**o+(0.907667166988)*x[1]+(-0.864298533288)*x[2]**o+(0.641505117281)*x[2]
        arg[1,3,1,0]=(0.381900624425)*x[0]**o+(0.427520178597)*x[0]+(-0.560279540293)*x[1]**o+(0.190814532271)*x[1]+(-0.187947688629)*x[2]**o+(-0.127694148593)*x[2]
        arg[1,3,1,1]=(-0.903921363895)*x[0]**o+(-0.659942040828)*x[0]+(-0.549460213492)*x[1]**o+(0.869909651895)*x[1]+(-0.962273439484)*x[2]**o+(-0.046365857751)*x[2]
        arg[1,3,2,0]=(-0.021688482187)*x[0]**o+(-0.209531000717)*x[0]+(0.668564395133)*x[1]**o+(-0.0347978700975)*x[1]+(0.947677147488)*x[2]**o+(-0.0152385052475)*x[2]
        arg[1,3,2,1]=(0.880839485817)*x[0]**o+(-0.319275929312)*x[0]+(-0.940056657485)*x[1]**o+(0.570855397726)*x[1]+(0.279710812074)*x[2]**o+(0.496622215051)*x[2]
        arg[1,4,0,0]=(-0.139033645615)*x[0]**o+(-0.271453722366)*x[0]+(-0.268516720244)*x[1]**o+(-0.0467555017438)*x[1]+(-0.41618262265)*x[2]**o+(0.315320363813)*x[2]
        arg[1,4,0,1]=(0.296108607204)*x[0]**o+(-0.184582165661)*x[0]+(0.121416543744)*x[1]**o+(-0.922284926751)*x[1]+(-0.437600089668)*x[2]**o+(-0.608800626513)*x[2]
        arg[1,4,1,0]=(0.8942028423)*x[0]**o+(0.673402831188)*x[0]+(-0.622529195562)*x[1]**o+(-0.266052338798)*x[1]+(-0.273167851702)*x[2]**o+(-0.312211982747)*x[2]
        arg[1,4,1,1]=(-0.102315711178)*x[0]**o+(-0.885515719371)*x[0]+(-0.532910101204)*x[1]**o+(0.798525944348)*x[1]+(0.640586802303)*x[2]**o+(0.36308806536)*x[2]
        arg[1,4,2,0]=(0.376793506541)*x[0]**o+(-0.667063564875)*x[0]+(0.609591985816)*x[1]**o+(0.28347745142)*x[1]+(-0.751832936302)*x[2]**o+(-0.683738856965)*x[2]
        arg[1,4,2,1]=(-0.51412252819)*x[0]**o+(0.559560297206)*x[0]+(0.751134603429)*x[1]**o+(-0.0031812310797)*x[1]+(-0.525776122178)*x[2]**o+(0.338471942848)*x[2]
        arg[2,0,0,0]=(0.937572244905)*x[0]**o+(-0.293020870379)*x[0]+(-0.918591520817)*x[1]**o+(0.622194477261)*x[1]+(0.31641482851)*x[2]**o+(0.276912266864)*x[2]
        arg[2,0,0,1]=(0.853041599998)*x[0]**o+(-0.0220623724171)*x[0]+(0.89909044296)*x[1]**o+(-0.838782642732)*x[1]+(0.0133197592042)*x[2]**o+(-0.868977618993)*x[2]
        arg[2,0,1,0]=(-0.325799042976)*x[0]**o+(0.746269971994)*x[0]+(0.882671061757)*x[1]**o+(-0.284133373252)*x[1]+(0.120086549521)*x[2]**o+(-0.397352610969)*x[2]
        arg[2,0,1,1]=(0.282336934951)*x[0]**o+(-0.339893978547)*x[0]+(-0.0861660125386)*x[1]**o+(0.283063579798)*x[1]+(0.714762478569)*x[2]**o+(-0.497021718549)*x[2]
        arg[2,0,2,0]=(0.891782779491)*x[0]**o+(-0.817647577644)*x[0]+(0.364088142085)*x[1]**o+(-0.950577637048)*x[1]+(0.4070663557)*x[2]**o+(0.270753179246)*x[2]
        arg[2,0,2,1]=(0.316594913808)*x[0]**o+(0.310731486851)*x[0]+(0.250421828765)*x[1]**o+(0.772786141458)*x[1]+(0.64843040679)*x[2]**o+(-0.691903332511)*x[2]
        arg[2,1,0,0]=(0.87180715028)*x[0]**o+(-0.953173230375)*x[0]+(-0.823516344548)*x[1]**o+(0.934653756548)*x[1]+(-0.537124962489)*x[2]**o+(-0.233365939659)*x[2]
        arg[2,1,0,1]=(-0.407189744631)*x[0]**o+(-0.754425720288)*x[0]+(0.737526767897)*x[1]**o+(0.745393943502)*x[1]+(0.811744293685)*x[2]**o+(-0.459556659177)*x[2]
        arg[2,1,1,0]=(-0.693529259877)*x[0]**o+(0.580197938048)*x[0]+(0.506055440405)*x[1]**o+(0.122531014759)*x[1]+(0.427736137775)*x[2]**o+(-0.83260140401)*x[2]
        arg[2,1,1,1]=(0.542621975026)*x[0]**o+(-0.568425713106)*x[0]+(0.303002043446)*x[1]**o+(0.363286261651)*x[1]+(0.2729476289)*x[2]**o+(-0.0113629744659)*x[2]
        arg[2,1,2,0]=(-0.266750235694)*x[0]**o+(0.10116400306)*x[0]+(0.315977962897)*x[1]**o+(-0.406236201454)*x[1]+(-0.586845072374)*x[2]**o+(0.513937047282)*x[2]
        arg[2,1,2,1]=(-0.518780811975)*x[0]**o+(0.275053640829)*x[0]+(0.4234970646)*x[1]**o+(0.389585163376)*x[1]+(-0.609771068778)*x[2]**o+(-0.121467061755)*x[2]
        arg[2,2,0,0]=(0.138823207466)*x[0]**o+(-0.779837455014)*x[0]+(0.70167049555)*x[1]**o+(0.766428813369)*x[1]+(0.114023530366)*x[2]**o+(0.650098536187)*x[2]
        arg[2,2,0,1]=(0.974998667633)*x[0]**o+(0.958471050654)*x[0]+(0.193515162503)*x[1]**o+(-0.252148167253)*x[1]+(0.495475460125)*x[2]**o+(0.562693347983)*x[2]
        arg[2,2,1,0]=(-0.552379548988)*x[0]**o+(-0.831399971618)*x[0]+(0.437581216647)*x[1]**o+(-0.775331267696)*x[1]+(0.729041628844)*x[2]**o+(0.633113359893)*x[2]
        arg[2,2,1,1]=(0.379858437049)*x[0]**o+(0.923586504203)*x[0]+(0.707244783969)*x[1]**o+(-0.293157375059)*x[1]+(-0.368717941876)*x[2]**o+(-0.202610288616)*x[2]
        arg[2,2,2,0]=(-0.804826108231)*x[0]**o+(-0.0208712096312)*x[0]+(0.0482547072498)*x[1]**o+(0.852366543609)*x[1]+(0.328962500136)*x[2]**o+(-0.132777835512)*x[2]
        arg[2,2,2,1]=(0.91373752217)*x[0]**o+(-0.792112120729)*x[0]+(-0.0457564087072)*x[1]**o+(0.309694852949)*x[1]+(0.089219285337)*x[2]**o+(-0.516000514614)*x[2]
        arg[2,3,0,0]=(0.343215129115)*x[0]**o+(0.40804099837)*x[0]+(-0.865942442914)*x[1]**o+(0.842241683043)*x[1]+(0.0727803297601)*x[2]**o+(-0.395225682949)*x[2]
        arg[2,3,0,1]=(-0.965340522771)*x[0]**o+(-0.897056478677)*x[0]+(-0.305205944062)*x[1]**o+(0.393941789685)*x[1]+(-0.253328379131)*x[2]**o+(0.0951510469271)*x[2]
        arg[2,3,1,0]=(0.536251514752)*x[0]**o+(0.0629776386779)*x[0]+(-0.706311222379)*x[1]**o+(-0.84362242029)*x[1]+(0.00110678760539)*x[2]**o+(-0.30286059941)*x[2]
        arg[2,3,1,1]=(0.114584737295)*x[0]**o+(0.0708763738067)*x[0]+(-0.454760062397)*x[1]**o+(0.228742539922)*x[1]+(0.514960810352)*x[2]**o+(-0.815749211085)*x[2]
        arg[2,3,2,0]=(0.246335688008)*x[0]**o+(-0.239199299885)*x[0]+(-0.447390151659)*x[1]**o+(0.918949356391)*x[1]+(-0.307864840175)*x[2]**o+(0.560139080481)*x[2]
        arg[2,3,2,1]=(0.118766035723)*x[0]**o+(0.610711164899)*x[0]+(0.336628771776)*x[1]**o+(0.833699268751)*x[1]+(-0.424694647429)*x[2]**o+(-0.639929688104)*x[2]
        arg[2,4,0,0]=(0.814543506987)*x[0]**o+(-0.797684321368)*x[0]+(0.337242635783)*x[1]**o+(0.278139837072)*x[1]+(0.633511526441)*x[2]**o+(0.372175615028)*x[2]
        arg[2,4,0,1]=(-0.0590668694094)*x[0]**o+(0.256732067267)*x[0]+(0.519737187577)*x[1]**o+(-0.376391855948)*x[1]+(-0.810934658651)*x[2]**o+(0.826899476928)*x[2]
        arg[2,4,1,0]=(-0.428976843661)*x[0]**o+(0.672667833313)*x[0]+(0.263194890946)*x[1]**o+(-0.894790773661)*x[1]+(-0.550298557295)*x[2]**o+(-0.00560384035502)*x[2]
        arg[2,4,1,1]=(-0.818372430144)*x[0]**o+(-0.829920833073)*x[0]+(-0.359560636119)*x[1]**o+(-0.113484563205)*x[1]+(0.586796110365)*x[2]**o+(0.296820898426)*x[2]
        arg[2,4,2,0]=(-0.691597969566)*x[0]**o+(0.715813153538)*x[0]+(0.954093612503)*x[1]**o+(0.88803420754)*x[1]+(0.697447667765)*x[2]**o+(0.793730496625)*x[2]
        arg[2,4,2,1]=(0.351099833336)*x[0]**o+(-0.716419272807)*x[0]+(0.318726674998)*x[1]**o+(0.246151914414)*x[1]+(-0.945520673884)*x[2]**o+(-0.489026233954)*x[2]
        arg[3,0,0,0]=(0.511598022804)*x[0]**o+(0.567617269582)*x[0]+(-0.526293319137)*x[1]**o+(0.333750098146)*x[1]+(-0.0869228849144)*x[2]**o+(-0.204072834867)*x[2]
        arg[3,0,0,1]=(0.905384007172)*x[0]**o+(-0.661067390314)*x[0]+(-0.470790632295)*x[1]**o+(-0.149608017616)*x[1]+(0.427879278267)*x[2]**o+(0.987040179184)*x[2]
        arg[3,0,1,0]=(-0.946562006809)*x[0]**o+(0.64141909633)*x[0]+(-0.762076826276)*x[1]**o+(-0.642912850794)*x[1]+(0.712567898559)*x[2]**o+(-0.292128944205)*x[2]
        arg[3,0,1,1]=(-0.955087949361)*x[0]**o+(0.426588041133)*x[0]+(0.339313104127)*x[1]**o+(-0.375147344045)*x[1]+(0.967259014679)*x[2]**o+(0.504521081858)*x[2]
        arg[3,0,2,0]=(0.141876598344)*x[0]**o+(-0.335579252825)*x[0]+(0.514552993602)*x[1]**o+(-0.531116948499)*x[1]+(-0.369319413141)*x[2]**o+(0.0374413262362)*x[2]
        arg[3,0,2,1]=(0.678360463658)*x[0]**o+(-0.357900119646)*x[0]+(0.300750717059)*x[1]**o+(0.516196189448)*x[1]+(0.0535466018203)*x[2]**o+(-0.149961940786)*x[2]
        arg[3,1,0,0]=(0.0975994351079)*x[0]**o+(0.634544068499)*x[0]+(-0.963217046352)*x[1]**o+(-0.804658091456)*x[1]+(-0.620132720823)*x[2]**o+(0.862250619369)*x[2]
        arg[3,1,0,1]=(0.210306328879)*x[0]**o+(-0.457341187966)*x[0]+(0.957556084801)*x[1]**o+(0.704073170928)*x[1]+(0.423134492133)*x[2]**o+(-0.167268375412)*x[2]
        arg[3,1,1,0]=(-0.909228817744)*x[0]**o+(0.110665018921)*x[0]+(0.36275630147)*x[1]**o+(0.346866824763)*x[1]+(-0.535600547489)*x[2]**o+(0.627638814375)*x[2]
        arg[3,1,1,1]=(0.328780933823)*x[0]**o+(-0.249563994161)*x[0]+(-0.555789062892)*x[1]**o+(-0.917287525036)*x[1]+(0.386191175017)*x[2]**o+(0.123904319061)*x[2]
        arg[3,1,2,0]=(-0.838409901623)*x[0]**o+(0.673479705393)*x[0]+(0.482921104917)*x[1]**o+(0.136411388588)*x[1]+(0.583674116225)*x[2]**o+(0.305100836877)*x[2]
        arg[3,1,2,1]=(0.952961008312)*x[0]**o+(-0.0501981079532)*x[0]+(0.661434464642)*x[1]**o+(0.65030469399)*x[1]+(-0.157562001295)*x[2]**o+(-0.249551892917)*x[2]
        arg[3,2,0,0]=(-0.0232804500286)*x[0]**o+(-0.559229392203)*x[0]+(0.127967930273)*x[1]**o+(-0.635827181645)*x[1]+(-0.0426703224383)*x[2]**o+(0.47720230258)*x[2]
        arg[3,2,0,1]=(-0.421658996106)*x[0]**o+(-0.468260038077)*x[0]+(0.479083572554)*x[1]**o+(-0.214417881841)*x[1]+(0.372430534261)*x[2]**o+(0.133063789546)*x[2]
        arg[3,2,1,0]=(0.905811586567)*x[0]**o+(0.245447340384)*x[0]+(-0.337650999146)*x[1]**o+(0.908188235976)*x[1]+(-0.804029751277)*x[2]**o+(0.0512676456643)*x[2]
        arg[3,2,1,1]=(-0.891486839409)*x[0]**o+(0.20753794329)*x[0]+(0.879183464252)*x[1]**o+(0.902028906992)*x[1]+(0.416776813598)*x[2]**o+(0.491964374373)*x[2]
        arg[3,2,2,0]=(0.653067027119)*x[0]**o+(0.765293890745)*x[0]+(-0.348902169508)*x[1]**o+(0.324429035849)*x[1]+(-0.363026365065)*x[2]**o+(0.0816186405511)*x[2]
        arg[3,2,2,1]=(-0.0415375052325)*x[0]**o+(-0.480779075784)*x[0]+(-0.137223552938)*x[1]**o+(-0.915733625512)*x[1]+(0.750199445159)*x[2]**o+(-0.789818945171)*x[2]
        arg[3,3,0,0]=(0.0744921239083)*x[0]**o+(0.0700589144159)*x[0]+(-0.327105425297)*x[1]**o+(0.469843257765)*x[1]+(-0.170191600844)*x[2]**o+(0.732071534155)*x[2]
        arg[3,3,0,1]=(0.172643927441)*x[0]**o+(0.848924114383)*x[0]+(-0.79125762228)*x[1]**o+(-0.694725775976)*x[1]+(-0.309710503985)*x[2]**o+(-0.405310234927)*x[2]
        arg[3,3,1,0]=(-0.465911787595)*x[0]**o+(-0.71080474709)*x[0]+(0.959925934203)*x[1]**o+(-0.111110644825)*x[1]+(-0.145349925218)*x[2]**o+(0.122955165811)*x[2]
        arg[3,3,1,1]=(-0.769111863754)*x[0]**o+(0.348036938532)*x[0]+(0.485890856474)*x[1]**o+(0.382595218985)*x[1]+(-0.897488692981)*x[2]**o+(0.830822333819)*x[2]
        arg[3,3,2,0]=(0.341890793681)*x[0]**o+(-0.667145936571)*x[0]+(-0.145094108268)*x[1]**o+(0.771433911212)*x[1]+(-0.299752164325)*x[2]**o+(-0.357856549677)*x[2]
        arg[3,3,2,1]=(0.26381838796)*x[0]**o+(-0.44474724998)*x[0]+(0.310283988195)*x[1]**o+(0.698675181314)*x[1]+(0.437624133238)*x[2]**o+(0.394916054445)*x[2]
        arg[3,4,0,0]=(-0.576499406674)*x[0]**o+(0.770936634959)*x[0]+(0.707441735783)*x[1]**o+(-0.794953257043)*x[1]+(-0.535342943209)*x[2]**o+(-0.509446811787)*x[2]
        arg[3,4,0,1]=(0.161115012113)*x[0]**o+(0.11079092252)*x[0]+(0.0633804181164)*x[1]**o+(-0.406049293019)*x[1]+(-0.824283994336)*x[2]**o+(-0.481107560499)*x[2]
        arg[3,4,1,0]=(-0.411133810295)*x[0]**o+(0.988486625099)*x[0]+(-0.36050191367)*x[1]**o+(0.469479821965)*x[1]+(-0.431175357325)*x[2]**o+(-0.966295015263)*x[2]
        arg[3,4,1,1]=(0.233731410367)*x[0]**o+(-0.821900740766)*x[0]+(-0.823123461385)*x[1]**o+(-0.380537914421)*x[1]+(-0.896146300355)*x[2]**o+(-0.775172990257)*x[2]
        arg[3,4,2,0]=(-0.96323363586)*x[0]**o+(0.0935412405245)*x[0]+(0.929333026259)*x[1]**o+(-0.322236627348)*x[1]+(-0.390898135964)*x[2]**o+(-0.766178290162)*x[2]
        arg[3,4,2,1]=(-0.773670533268)*x[0]**o+(-0.156975157098)*x[0]+(-0.91512996795)*x[1]**o+(0.0247803682126)*x[1]+(0.995159733782)*x[2]**o+(-0.669224144225)*x[2]
        ref[0,0,0,0]=(-0.110519315354)*x_ref[0]**o+(0.904227374292)*x_ref[0]+(0.459005710707)*x_ref[1]**o+(-0.980231422256)*x_ref[1]+(-0.819713355475)*x_ref[2]**o+(-0.93459517989)*x_ref[2]
        ref[0,0,0,1]=(-0.326902946413)*x_ref[0]**o+(0.489327766219)*x_ref[0]+(-0.965308741061)*x_ref[1]**o+(0.0351399111471)*x_ref[1]+(0.257760096595)*x_ref[2]**o+(0.125123023244)*x_ref[2]
        ref[0,0,1,0]=(-0.584360710004)*x_ref[0]**o+(0.681892732544)*x_ref[0]+(-0.118257741614)*x_ref[1]**o+(0.839875738892)*x_ref[1]+(0.107934154241)*x_ref[2]**o+(0.279757054665)*x_ref[2]
        ref[0,0,1,1]=(0.90973214794)*x_ref[0]**o+(-0.776804618826)*x_ref[0]+(-0.910633853381)*x_ref[1]**o+(-0.25678788529)*x_ref[1]+(0.783709363646)*x_ref[2]**o+(-0.922660855915)*x_ref[2]
        ref[0,0,2,0]=(-0.802273263394)*x_ref[0]**o+(-0.917749665952)*x_ref[0]+(0.796206817226)*x_ref[1]**o+(0.917284637885)*x_ref[1]+(0.219841508893)*x_ref[2]**o+(0.0655848200083)*x_ref[2]
        ref[0,0,2,1]=(0.550596792343)*x_ref[0]**o+(-0.760207057617)*x_ref[0]+(0.0368469683098)*x_ref[1]**o+(0.509347717331)*x_ref[1]+(0.813780778487)*x_ref[2]**o+(-0.432656298451)*x_ref[2]
        ref[0,1,0,0]=(-0.238502042941)*x_ref[0]**o+(-0.966632221032)*x_ref[0]+(-0.373745652933)*x_ref[1]**o+(-0.436176336389)*x_ref[1]+(0.124280197699)*x_ref[2]**o+(0.552700562803)*x_ref[2]
        ref[0,1,0,1]=(-0.19848053276)*x_ref[0]**o+(0.110167634957)*x_ref[0]+(-0.748553857473)*x_ref[1]**o+(0.355932508959)*x_ref[1]+(-0.226987914753)*x_ref[2]**o+(-0.68870238197)*x_ref[2]
        ref[0,1,1,0]=(0.365822518882)*x_ref[0]**o+(0.327285102897)*x_ref[0]+(0.549212504635)*x_ref[1]**o+(0.918179946504)*x_ref[1]+(0.20644004679)*x_ref[2]**o+(0.821299409756)*x_ref[2]
        ref[0,1,1,1]=(-0.366167489101)*x_ref[0]**o+(0.529933186572)*x_ref[0]+(-0.604965310967)*x_ref[1]**o+(-0.833093092352)*x_ref[1]+(-0.73593913912)*x_ref[2]**o+(0.872325614453)*x_ref[2]
        ref[0,1,2,0]=(0.289123646902)*x_ref[0]**o+(0.705862153789)*x_ref[0]+(-0.319756886822)*x_ref[1]**o+(-0.453220232524)*x_ref[1]+(-0.192336374167)*x_ref[2]**o+(0.363575495006)*x_ref[2]
        ref[0,1,2,1]=(-0.213014833353)*x_ref[0]**o+(-0.0706784230627)*x_ref[0]+(-0.941272982229)*x_ref[1]**o+(-0.96641361036)*x_ref[1]+(0.289919014964)*x_ref[2]**o+(0.57658772066)*x_ref[2]
        ref[0,2,0,0]=(-0.406683991704)*x_ref[0]**o+(-0.0110042213592)*x_ref[0]+(-0.273148083258)*x_ref[1]**o+(0.054257030447)*x_ref[1]+(0.387598147681)*x_ref[2]**o+(0.480398109069)*x_ref[2]
        ref[0,2,0,1]=(0.318713057562)*x_ref[0]**o+(-0.449467749585)*x_ref[0]+(0.669991267829)*x_ref[1]**o+(0.237986860249)*x_ref[1]+(0.894370523471)*x_ref[2]**o+(-0.397832436233)*x_ref[2]
        ref[0,2,1,0]=(-0.858603253642)*x_ref[0]**o+(-0.713122015919)*x_ref[0]+(0.870251906188)*x_ref[1]**o+(-0.781601254578)*x_ref[1]+(-0.504531525822)*x_ref[2]**o+(-0.699819349556)*x_ref[2]
        ref[0,2,1,1]=(-0.825591116361)*x_ref[0]**o+(-0.552452876394)*x_ref[0]+(-0.262627261618)*x_ref[1]**o+(-0.147127256184)*x_ref[1]+(-0.171485665103)*x_ref[2]**o+(0.823279741501)*x_ref[2]
        ref[0,2,2,0]=(-0.0416279689068)*x_ref[0]**o+(-0.0903338000001)*x_ref[0]+(0.911673780635)*x_ref[1]**o+(0.86298673049)*x_ref[1]+(-0.60251179847)*x_ref[2]**o+(0.378303877746)*x_ref[2]
        ref[0,2,2,1]=(0.261384106387)*x_ref[0]**o+(0.0326416199811)*x_ref[0]+(-0.426740102269)*x_ref[1]**o+(-0.445401626921)*x_ref[1]+(-0.388668254526)*x_ref[2]**o+(-0.977942913284)*x_ref[2]
        ref[0,3,0,0]=(-0.0195053813483)*x_ref[0]**o+(-0.294061998664)*x_ref[0]+(0.790461098426)*x_ref[1]**o+(-0.506334268838)*x_ref[1]+(0.394015541605)*x_ref[2]**o+(-0.562248105432)*x_ref[2]
        ref[0,3,0,1]=(0.390726516735)*x_ref[0]**o+(-0.757597437859)*x_ref[0]+(0.861882149861)*x_ref[1]**o+(0.621870451314)*x_ref[1]+(0.900777666162)*x_ref[2]**o+(0.271507131938)*x_ref[2]
        ref[0,3,1,0]=(-0.725635966475)*x_ref[0]**o+(-0.695837548737)*x_ref[0]+(-0.163078176328)*x_ref[1]**o+(-0.0435808418729)*x_ref[1]+(0.0890756544611)*x_ref[2]**o+(-0.270422627949)*x_ref[2]
        ref[0,3,1,1]=(-0.175498720647)*x_ref[0]**o+(-0.213605441632)*x_ref[0]+(-0.358632620362)*x_ref[1]**o+(-0.193273004152)*x_ref[1]+(0.937704823805)*x_ref[2]**o+(0.861707832291)*x_ref[2]
        ref[0,3,2,0]=(0.0976010265611)*x_ref[0]**o+(0.995135925341)*x_ref[0]+(0.755339287387)*x_ref[1]**o+(-0.924722068193)*x_ref[1]+(-0.480212181904)*x_ref[2]**o+(-0.40237760439)*x_ref[2]
        ref[0,3,2,1]=(-0.679215392952)*x_ref[0]**o+(-0.00796096253297)*x_ref[0]+(-0.964972599421)*x_ref[1]**o+(-0.410415435656)*x_ref[1]+(-0.292638067785)*x_ref[2]**o+(-0.950383976561)*x_ref[2]
        ref[0,4,0,0]=(0.245586594928)*x_ref[0]**o+(-0.593420401896)*x_ref[0]+(0.553450486534)*x_ref[1]**o+(0.173814965689)*x_ref[1]+(0.0719392610563)*x_ref[2]**o+(0.676231437486)*x_ref[2]
        ref[0,4,0,1]=(0.154382756425)*x_ref[0]**o+(-0.0655740293432)*x_ref[0]+(0.924689745142)*x_ref[1]**o+(-0.246876531175)*x_ref[1]+(-0.241460713543)*x_ref[2]**o+(-0.489840108568)*x_ref[2]
        ref[0,4,1,0]=(-0.482705840359)*x_ref[0]**o+(-0.392960380114)*x_ref[0]+(-0.503463823972)*x_ref[1]**o+(0.317163204738)*x_ref[1]+(0.775048877356)*x_ref[2]**o+(0.599889732326)*x_ref[2]
        ref[0,4,1,1]=(-0.122555861716)*x_ref[0]**o+(0.930175770449)*x_ref[0]+(0.483503764941)*x_ref[1]**o+(0.340001033217)*x_ref[1]+(0.672924280979)*x_ref[2]**o+(0.746489359683)*x_ref[2]
        ref[0,4,2,0]=(-0.339421141033)*x_ref[0]**o+(-0.604966575169)*x_ref[0]+(0.862413017648)*x_ref[1]**o+(0.936752078454)*x_ref[1]+(-0.0549628353285)*x_ref[2]**o+(-0.0858885835314)*x_ref[2]
        ref[0,4,2,1]=(-0.581767357421)*x_ref[0]**o+(-0.644286027548)*x_ref[0]+(0.76948396865)*x_ref[1]**o+(0.692265904744)*x_ref[1]+(0.955871091531)*x_ref[2]**o+(-0.378888503457)*x_ref[2]
        ref[1,0,0,0]=(-0.0407846595046)*x_ref[0]**o+(-0.293709943251)*x_ref[0]+(-0.310552931582)*x_ref[1]**o+(-0.696244178466)*x_ref[1]+(0.962985284252)*x_ref[2]**o+(-0.0266457621404)*x_ref[2]
        ref[1,0,0,1]=(-0.480942811584)*x_ref[0]**o+(-0.400158759679)*x_ref[0]+(0.311171869807)*x_ref[1]**o+(-0.79518809427)*x_ref[1]+(-0.497004450478)*x_ref[2]**o+(-0.639822005752)*x_ref[2]
        ref[1,0,1,0]=(0.158738648483)*x_ref[0]**o+(0.274859816468)*x_ref[0]+(-0.208677612654)*x_ref[1]**o+(0.911107598366)*x_ref[1]+(-0.400163007778)*x_ref[2]**o+(0.276296890267)*x_ref[2]
        ref[1,0,1,1]=(0.613312276864)*x_ref[0]**o+(-0.644237218776)*x_ref[0]+(0.81927272895)*x_ref[1]**o+(-0.12971860556)*x_ref[1]+(0.214379406799)*x_ref[2]**o+(0.856335675043)*x_ref[2]
        ref[1,0,2,0]=(-0.699925173766)*x_ref[0]**o+(-0.560613547952)*x_ref[0]+(-0.355325107637)*x_ref[1]**o+(-0.598747822041)*x_ref[1]+(0.243087037553)*x_ref[2]**o+(-0.484176993015)*x_ref[2]
        ref[1,0,2,1]=(-0.218912300848)*x_ref[0]**o+(-0.259483312969)*x_ref[0]+(-0.863788186295)*x_ref[1]**o+(0.924423774757)*x_ref[1]+(-0.0823752192204)*x_ref[2]**o+(-0.928377900602)*x_ref[2]
        ref[1,1,0,0]=(-0.368901144425)*x_ref[0]**o+(0.544104655592)*x_ref[0]+(0.321295151176)*x_ref[1]**o+(0.395692754134)*x_ref[1]+(-0.721319111754)*x_ref[2]**o+(-0.698518442137)*x_ref[2]
        ref[1,1,0,1]=(-0.592867336047)*x_ref[0]**o+(0.145120231038)*x_ref[0]+(0.110390771366)*x_ref[1]**o+(0.168333368086)*x_ref[1]+(-0.577866419921)*x_ref[2]**o+(0.705027708469)*x_ref[2]
        ref[1,1,1,0]=(-0.210603428235)*x_ref[0]**o+(-0.250156447116)*x_ref[0]+(-0.397803248976)*x_ref[1]**o+(0.332163380867)*x_ref[1]+(-0.250648216541)*x_ref[2]**o+(-0.0693729055697)*x_ref[2]
        ref[1,1,1,1]=(0.324927558096)*x_ref[0]**o+(-0.262726116223)*x_ref[0]+(-0.555439758488)*x_ref[1]**o+(-0.865019248609)*x_ref[1]+(-0.0971907425857)*x_ref[2]**o+(-0.742359236711)*x_ref[2]
        ref[1,1,2,0]=(0.161058869414)*x_ref[0]**o+(0.349191500791)*x_ref[0]+(0.468336473315)*x_ref[1]**o+(-0.856364775995)*x_ref[1]+(0.491736696689)*x_ref[2]**o+(0.352984104173)*x_ref[2]
        ref[1,1,2,1]=(-0.243275976508)*x_ref[0]**o+(-0.785605388878)*x_ref[0]+(-0.282478919296)*x_ref[1]**o+(-0.728851522341)*x_ref[1]+(0.375134848237)*x_ref[2]**o+(-0.486200561256)*x_ref[2]
        ref[1,2,0,0]=(-0.172241451405)*x_ref[0]**o+(0.422361498196)*x_ref[0]+(0.464862755164)*x_ref[1]**o+(0.466661717105)*x_ref[1]+(-0.886209677512)*x_ref[2]**o+(-0.43529076789)*x_ref[2]
        ref[1,2,0,1]=(0.946582222405)*x_ref[0]**o+(0.513640973894)*x_ref[0]+(-0.923788000633)*x_ref[1]**o+(-0.358163567195)*x_ref[1]+(-0.709999132129)*x_ref[2]**o+(-0.761163299694)*x_ref[2]
        ref[1,2,1,0]=(0.236339981889)*x_ref[0]**o+(0.259874866134)*x_ref[0]+(-0.32460762691)*x_ref[1]**o+(-0.629786950313)*x_ref[1]+(-0.679261685402)*x_ref[2]**o+(0.593069192918)*x_ref[2]
        ref[1,2,1,1]=(0.11006800009)*x_ref[0]**o+(0.970532991787)*x_ref[0]+(-0.467608188361)*x_ref[1]**o+(-0.959281984552)*x_ref[1]+(0.610396622519)*x_ref[2]**o+(0.282468600058)*x_ref[2]
        ref[1,2,2,0]=(-0.656419157742)*x_ref[0]**o+(0.234027308171)*x_ref[0]+(-0.67027206978)*x_ref[1]**o+(-0.746034436817)*x_ref[1]+(-0.771894073652)*x_ref[2]**o+(0.949545830287)*x_ref[2]
        ref[1,2,2,1]=(0.615884236811)*x_ref[0]**o+(-0.670156612555)*x_ref[0]+(0.543165493345)*x_ref[1]**o+(-0.618294277594)*x_ref[1]+(-0.974114255061)*x_ref[2]**o+(-0.762320542298)*x_ref[2]
        ref[1,3,0,0]=(0.096201957648)*x_ref[0]**o+(0.440870816208)*x_ref[0]+(0.124448859607)*x_ref[1]**o+(0.507919407561)*x_ref[1]+(-0.401666263687)*x_ref[2]**o+(-0.437653713723)*x_ref[2]
        ref[1,3,0,1]=(0.165129229294)*x_ref[0]**o+(-0.353217428955)*x_ref[0]+(0.473666065504)*x_ref[1]**o+(0.907667166988)*x_ref[1]+(-0.864298533288)*x_ref[2]**o+(0.641505117281)*x_ref[2]
        ref[1,3,1,0]=(0.381900624425)*x_ref[0]**o+(0.427520178597)*x_ref[0]+(-0.560279540293)*x_ref[1]**o+(0.190814532271)*x_ref[1]+(-0.187947688629)*x_ref[2]**o+(-0.127694148593)*x_ref[2]
        ref[1,3,1,1]=(-0.903921363895)*x_ref[0]**o+(-0.659942040828)*x_ref[0]+(-0.549460213492)*x_ref[1]**o+(0.869909651895)*x_ref[1]+(-0.962273439484)*x_ref[2]**o+(-0.046365857751)*x_ref[2]
        ref[1,3,2,0]=(-0.021688482187)*x_ref[0]**o+(-0.209531000717)*x_ref[0]+(0.668564395133)*x_ref[1]**o+(-0.0347978700975)*x_ref[1]+(0.947677147488)*x_ref[2]**o+(-0.0152385052475)*x_ref[2]
        ref[1,3,2,1]=(0.880839485817)*x_ref[0]**o+(-0.319275929312)*x_ref[0]+(-0.940056657485)*x_ref[1]**o+(0.570855397726)*x_ref[1]+(0.279710812074)*x_ref[2]**o+(0.496622215051)*x_ref[2]
        ref[1,4,0,0]=(-0.139033645615)*x_ref[0]**o+(-0.271453722366)*x_ref[0]+(-0.268516720244)*x_ref[1]**o+(-0.0467555017438)*x_ref[1]+(-0.41618262265)*x_ref[2]**o+(0.315320363813)*x_ref[2]
        ref[1,4,0,1]=(0.296108607204)*x_ref[0]**o+(-0.184582165661)*x_ref[0]+(0.121416543744)*x_ref[1]**o+(-0.922284926751)*x_ref[1]+(-0.437600089668)*x_ref[2]**o+(-0.608800626513)*x_ref[2]
        ref[1,4,1,0]=(0.8942028423)*x_ref[0]**o+(0.673402831188)*x_ref[0]+(-0.622529195562)*x_ref[1]**o+(-0.266052338798)*x_ref[1]+(-0.273167851702)*x_ref[2]**o+(-0.312211982747)*x_ref[2]
        ref[1,4,1,1]=(-0.102315711178)*x_ref[0]**o+(-0.885515719371)*x_ref[0]+(-0.532910101204)*x_ref[1]**o+(0.798525944348)*x_ref[1]+(0.640586802303)*x_ref[2]**o+(0.36308806536)*x_ref[2]
        ref[1,4,2,0]=(0.376793506541)*x_ref[0]**o+(-0.667063564875)*x_ref[0]+(0.609591985816)*x_ref[1]**o+(0.28347745142)*x_ref[1]+(-0.751832936302)*x_ref[2]**o+(-0.683738856965)*x_ref[2]
        ref[1,4,2,1]=(-0.51412252819)*x_ref[0]**o+(0.559560297206)*x_ref[0]+(0.751134603429)*x_ref[1]**o+(-0.0031812310797)*x_ref[1]+(-0.525776122178)*x_ref[2]**o+(0.338471942848)*x_ref[2]
        ref[2,0,0,0]=(0.937572244905)*x_ref[0]**o+(-0.293020870379)*x_ref[0]+(-0.918591520817)*x_ref[1]**o+(0.622194477261)*x_ref[1]+(0.31641482851)*x_ref[2]**o+(0.276912266864)*x_ref[2]
        ref[2,0,0,1]=(0.853041599998)*x_ref[0]**o+(-0.0220623724171)*x_ref[0]+(0.89909044296)*x_ref[1]**o+(-0.838782642732)*x_ref[1]+(0.0133197592042)*x_ref[2]**o+(-0.868977618993)*x_ref[2]
        ref[2,0,1,0]=(-0.325799042976)*x_ref[0]**o+(0.746269971994)*x_ref[0]+(0.882671061757)*x_ref[1]**o+(-0.284133373252)*x_ref[1]+(0.120086549521)*x_ref[2]**o+(-0.397352610969)*x_ref[2]
        ref[2,0,1,1]=(0.282336934951)*x_ref[0]**o+(-0.339893978547)*x_ref[0]+(-0.0861660125386)*x_ref[1]**o+(0.283063579798)*x_ref[1]+(0.714762478569)*x_ref[2]**o+(-0.497021718549)*x_ref[2]
        ref[2,0,2,0]=(0.891782779491)*x_ref[0]**o+(-0.817647577644)*x_ref[0]+(0.364088142085)*x_ref[1]**o+(-0.950577637048)*x_ref[1]+(0.4070663557)*x_ref[2]**o+(0.270753179246)*x_ref[2]
        ref[2,0,2,1]=(0.316594913808)*x_ref[0]**o+(0.310731486851)*x_ref[0]+(0.250421828765)*x_ref[1]**o+(0.772786141458)*x_ref[1]+(0.64843040679)*x_ref[2]**o+(-0.691903332511)*x_ref[2]
        ref[2,1,0,0]=(0.87180715028)*x_ref[0]**o+(-0.953173230375)*x_ref[0]+(-0.823516344548)*x_ref[1]**o+(0.934653756548)*x_ref[1]+(-0.537124962489)*x_ref[2]**o+(-0.233365939659)*x_ref[2]
        ref[2,1,0,1]=(-0.407189744631)*x_ref[0]**o+(-0.754425720288)*x_ref[0]+(0.737526767897)*x_ref[1]**o+(0.745393943502)*x_ref[1]+(0.811744293685)*x_ref[2]**o+(-0.459556659177)*x_ref[2]
        ref[2,1,1,0]=(-0.693529259877)*x_ref[0]**o+(0.580197938048)*x_ref[0]+(0.506055440405)*x_ref[1]**o+(0.122531014759)*x_ref[1]+(0.427736137775)*x_ref[2]**o+(-0.83260140401)*x_ref[2]
        ref[2,1,1,1]=(0.542621975026)*x_ref[0]**o+(-0.568425713106)*x_ref[0]+(0.303002043446)*x_ref[1]**o+(0.363286261651)*x_ref[1]+(0.2729476289)*x_ref[2]**o+(-0.0113629744659)*x_ref[2]
        ref[2,1,2,0]=(-0.266750235694)*x_ref[0]**o+(0.10116400306)*x_ref[0]+(0.315977962897)*x_ref[1]**o+(-0.406236201454)*x_ref[1]+(-0.586845072374)*x_ref[2]**o+(0.513937047282)*x_ref[2]
        ref[2,1,2,1]=(-0.518780811975)*x_ref[0]**o+(0.275053640829)*x_ref[0]+(0.4234970646)*x_ref[1]**o+(0.389585163376)*x_ref[1]+(-0.609771068778)*x_ref[2]**o+(-0.121467061755)*x_ref[2]
        ref[2,2,0,0]=(0.138823207466)*x_ref[0]**o+(-0.779837455014)*x_ref[0]+(0.70167049555)*x_ref[1]**o+(0.766428813369)*x_ref[1]+(0.114023530366)*x_ref[2]**o+(0.650098536187)*x_ref[2]
        ref[2,2,0,1]=(0.974998667633)*x_ref[0]**o+(0.958471050654)*x_ref[0]+(0.193515162503)*x_ref[1]**o+(-0.252148167253)*x_ref[1]+(0.495475460125)*x_ref[2]**o+(0.562693347983)*x_ref[2]
        ref[2,2,1,0]=(-0.552379548988)*x_ref[0]**o+(-0.831399971618)*x_ref[0]+(0.437581216647)*x_ref[1]**o+(-0.775331267696)*x_ref[1]+(0.729041628844)*x_ref[2]**o+(0.633113359893)*x_ref[2]
        ref[2,2,1,1]=(0.379858437049)*x_ref[0]**o+(0.923586504203)*x_ref[0]+(0.707244783969)*x_ref[1]**o+(-0.293157375059)*x_ref[1]+(-0.368717941876)*x_ref[2]**o+(-0.202610288616)*x_ref[2]
        ref[2,2,2,0]=(-0.804826108231)*x_ref[0]**o+(-0.0208712096312)*x_ref[0]+(0.0482547072498)*x_ref[1]**o+(0.852366543609)*x_ref[1]+(0.328962500136)*x_ref[2]**o+(-0.132777835512)*x_ref[2]
        ref[2,2,2,1]=(0.91373752217)*x_ref[0]**o+(-0.792112120729)*x_ref[0]+(-0.0457564087072)*x_ref[1]**o+(0.309694852949)*x_ref[1]+(0.089219285337)*x_ref[2]**o+(-0.516000514614)*x_ref[2]
        ref[2,3,0,0]=(0.343215129115)*x_ref[0]**o+(0.40804099837)*x_ref[0]+(-0.865942442914)*x_ref[1]**o+(0.842241683043)*x_ref[1]+(0.0727803297601)*x_ref[2]**o+(-0.395225682949)*x_ref[2]
        ref[2,3,0,1]=(-0.965340522771)*x_ref[0]**o+(-0.897056478677)*x_ref[0]+(-0.305205944062)*x_ref[1]**o+(0.393941789685)*x_ref[1]+(-0.253328379131)*x_ref[2]**o+(0.0951510469271)*x_ref[2]
        ref[2,3,1,0]=(0.536251514752)*x_ref[0]**o+(0.0629776386779)*x_ref[0]+(-0.706311222379)*x_ref[1]**o+(-0.84362242029)*x_ref[1]+(0.00110678760539)*x_ref[2]**o+(-0.30286059941)*x_ref[2]
        ref[2,3,1,1]=(0.114584737295)*x_ref[0]**o+(0.0708763738067)*x_ref[0]+(-0.454760062397)*x_ref[1]**o+(0.228742539922)*x_ref[1]+(0.514960810352)*x_ref[2]**o+(-0.815749211085)*x_ref[2]
        ref[2,3,2,0]=(0.246335688008)*x_ref[0]**o+(-0.239199299885)*x_ref[0]+(-0.447390151659)*x_ref[1]**o+(0.918949356391)*x_ref[1]+(-0.307864840175)*x_ref[2]**o+(0.560139080481)*x_ref[2]
        ref[2,3,2,1]=(0.118766035723)*x_ref[0]**o+(0.610711164899)*x_ref[0]+(0.336628771776)*x_ref[1]**o+(0.833699268751)*x_ref[1]+(-0.424694647429)*x_ref[2]**o+(-0.639929688104)*x_ref[2]
        ref[2,4,0,0]=(0.814543506987)*x_ref[0]**o+(-0.797684321368)*x_ref[0]+(0.337242635783)*x_ref[1]**o+(0.278139837072)*x_ref[1]+(0.633511526441)*x_ref[2]**o+(0.372175615028)*x_ref[2]
        ref[2,4,0,1]=(-0.0590668694094)*x_ref[0]**o+(0.256732067267)*x_ref[0]+(0.519737187577)*x_ref[1]**o+(-0.376391855948)*x_ref[1]+(-0.810934658651)*x_ref[2]**o+(0.826899476928)*x_ref[2]
        ref[2,4,1,0]=(-0.428976843661)*x_ref[0]**o+(0.672667833313)*x_ref[0]+(0.263194890946)*x_ref[1]**o+(-0.894790773661)*x_ref[1]+(-0.550298557295)*x_ref[2]**o+(-0.00560384035502)*x_ref[2]
        ref[2,4,1,1]=(-0.818372430144)*x_ref[0]**o+(-0.829920833073)*x_ref[0]+(-0.359560636119)*x_ref[1]**o+(-0.113484563205)*x_ref[1]+(0.586796110365)*x_ref[2]**o+(0.296820898426)*x_ref[2]
        ref[2,4,2,0]=(-0.691597969566)*x_ref[0]**o+(0.715813153538)*x_ref[0]+(0.954093612503)*x_ref[1]**o+(0.88803420754)*x_ref[1]+(0.697447667765)*x_ref[2]**o+(0.793730496625)*x_ref[2]
        ref[2,4,2,1]=(0.351099833336)*x_ref[0]**o+(-0.716419272807)*x_ref[0]+(0.318726674998)*x_ref[1]**o+(0.246151914414)*x_ref[1]+(-0.945520673884)*x_ref[2]**o+(-0.489026233954)*x_ref[2]
        ref[3,0,0,0]=(0.511598022804)*x_ref[0]**o+(0.567617269582)*x_ref[0]+(-0.526293319137)*x_ref[1]**o+(0.333750098146)*x_ref[1]+(-0.0869228849144)*x_ref[2]**o+(-0.204072834867)*x_ref[2]
        ref[3,0,0,1]=(0.905384007172)*x_ref[0]**o+(-0.661067390314)*x_ref[0]+(-0.470790632295)*x_ref[1]**o+(-0.149608017616)*x_ref[1]+(0.427879278267)*x_ref[2]**o+(0.987040179184)*x_ref[2]
        ref[3,0,1,0]=(-0.946562006809)*x_ref[0]**o+(0.64141909633)*x_ref[0]+(-0.762076826276)*x_ref[1]**o+(-0.642912850794)*x_ref[1]+(0.712567898559)*x_ref[2]**o+(-0.292128944205)*x_ref[2]
        ref[3,0,1,1]=(-0.955087949361)*x_ref[0]**o+(0.426588041133)*x_ref[0]+(0.339313104127)*x_ref[1]**o+(-0.375147344045)*x_ref[1]+(0.967259014679)*x_ref[2]**o+(0.504521081858)*x_ref[2]
        ref[3,0,2,0]=(0.141876598344)*x_ref[0]**o+(-0.335579252825)*x_ref[0]+(0.514552993602)*x_ref[1]**o+(-0.531116948499)*x_ref[1]+(-0.369319413141)*x_ref[2]**o+(0.0374413262362)*x_ref[2]
        ref[3,0,2,1]=(0.678360463658)*x_ref[0]**o+(-0.357900119646)*x_ref[0]+(0.300750717059)*x_ref[1]**o+(0.516196189448)*x_ref[1]+(0.0535466018203)*x_ref[2]**o+(-0.149961940786)*x_ref[2]
        ref[3,1,0,0]=(0.0975994351079)*x_ref[0]**o+(0.634544068499)*x_ref[0]+(-0.963217046352)*x_ref[1]**o+(-0.804658091456)*x_ref[1]+(-0.620132720823)*x_ref[2]**o+(0.862250619369)*x_ref[2]
        ref[3,1,0,1]=(0.210306328879)*x_ref[0]**o+(-0.457341187966)*x_ref[0]+(0.957556084801)*x_ref[1]**o+(0.704073170928)*x_ref[1]+(0.423134492133)*x_ref[2]**o+(-0.167268375412)*x_ref[2]
        ref[3,1,1,0]=(-0.909228817744)*x_ref[0]**o+(0.110665018921)*x_ref[0]+(0.36275630147)*x_ref[1]**o+(0.346866824763)*x_ref[1]+(-0.535600547489)*x_ref[2]**o+(0.627638814375)*x_ref[2]
        ref[3,1,1,1]=(0.328780933823)*x_ref[0]**o+(-0.249563994161)*x_ref[0]+(-0.555789062892)*x_ref[1]**o+(-0.917287525036)*x_ref[1]+(0.386191175017)*x_ref[2]**o+(0.123904319061)*x_ref[2]
        ref[3,1,2,0]=(-0.838409901623)*x_ref[0]**o+(0.673479705393)*x_ref[0]+(0.482921104917)*x_ref[1]**o+(0.136411388588)*x_ref[1]+(0.583674116225)*x_ref[2]**o+(0.305100836877)*x_ref[2]
        ref[3,1,2,1]=(0.952961008312)*x_ref[0]**o+(-0.0501981079532)*x_ref[0]+(0.661434464642)*x_ref[1]**o+(0.65030469399)*x_ref[1]+(-0.157562001295)*x_ref[2]**o+(-0.249551892917)*x_ref[2]
        ref[3,2,0,0]=(-0.0232804500286)*x_ref[0]**o+(-0.559229392203)*x_ref[0]+(0.127967930273)*x_ref[1]**o+(-0.635827181645)*x_ref[1]+(-0.0426703224383)*x_ref[2]**o+(0.47720230258)*x_ref[2]
        ref[3,2,0,1]=(-0.421658996106)*x_ref[0]**o+(-0.468260038077)*x_ref[0]+(0.479083572554)*x_ref[1]**o+(-0.214417881841)*x_ref[1]+(0.372430534261)*x_ref[2]**o+(0.133063789546)*x_ref[2]
        ref[3,2,1,0]=(0.905811586567)*x_ref[0]**o+(0.245447340384)*x_ref[0]+(-0.337650999146)*x_ref[1]**o+(0.908188235976)*x_ref[1]+(-0.804029751277)*x_ref[2]**o+(0.0512676456643)*x_ref[2]
        ref[3,2,1,1]=(-0.891486839409)*x_ref[0]**o+(0.20753794329)*x_ref[0]+(0.879183464252)*x_ref[1]**o+(0.902028906992)*x_ref[1]+(0.416776813598)*x_ref[2]**o+(0.491964374373)*x_ref[2]
        ref[3,2,2,0]=(0.653067027119)*x_ref[0]**o+(0.765293890745)*x_ref[0]+(-0.348902169508)*x_ref[1]**o+(0.324429035849)*x_ref[1]+(-0.363026365065)*x_ref[2]**o+(0.0816186405511)*x_ref[2]
        ref[3,2,2,1]=(-0.0415375052325)*x_ref[0]**o+(-0.480779075784)*x_ref[0]+(-0.137223552938)*x_ref[1]**o+(-0.915733625512)*x_ref[1]+(0.750199445159)*x_ref[2]**o+(-0.789818945171)*x_ref[2]
        ref[3,3,0,0]=(0.0744921239083)*x_ref[0]**o+(0.0700589144159)*x_ref[0]+(-0.327105425297)*x_ref[1]**o+(0.469843257765)*x_ref[1]+(-0.170191600844)*x_ref[2]**o+(0.732071534155)*x_ref[2]
        ref[3,3,0,1]=(0.172643927441)*x_ref[0]**o+(0.848924114383)*x_ref[0]+(-0.79125762228)*x_ref[1]**o+(-0.694725775976)*x_ref[1]+(-0.309710503985)*x_ref[2]**o+(-0.405310234927)*x_ref[2]
        ref[3,3,1,0]=(-0.465911787595)*x_ref[0]**o+(-0.71080474709)*x_ref[0]+(0.959925934203)*x_ref[1]**o+(-0.111110644825)*x_ref[1]+(-0.145349925218)*x_ref[2]**o+(0.122955165811)*x_ref[2]
        ref[3,3,1,1]=(-0.769111863754)*x_ref[0]**o+(0.348036938532)*x_ref[0]+(0.485890856474)*x_ref[1]**o+(0.382595218985)*x_ref[1]+(-0.897488692981)*x_ref[2]**o+(0.830822333819)*x_ref[2]
        ref[3,3,2,0]=(0.341890793681)*x_ref[0]**o+(-0.667145936571)*x_ref[0]+(-0.145094108268)*x_ref[1]**o+(0.771433911212)*x_ref[1]+(-0.299752164325)*x_ref[2]**o+(-0.357856549677)*x_ref[2]
        ref[3,3,2,1]=(0.26381838796)*x_ref[0]**o+(-0.44474724998)*x_ref[0]+(0.310283988195)*x_ref[1]**o+(0.698675181314)*x_ref[1]+(0.437624133238)*x_ref[2]**o+(0.394916054445)*x_ref[2]
        ref[3,4,0,0]=(-0.576499406674)*x_ref[0]**o+(0.770936634959)*x_ref[0]+(0.707441735783)*x_ref[1]**o+(-0.794953257043)*x_ref[1]+(-0.535342943209)*x_ref[2]**o+(-0.509446811787)*x_ref[2]
        ref[3,4,0,1]=(0.161115012113)*x_ref[0]**o+(0.11079092252)*x_ref[0]+(0.0633804181164)*x_ref[1]**o+(-0.406049293019)*x_ref[1]+(-0.824283994336)*x_ref[2]**o+(-0.481107560499)*x_ref[2]
        ref[3,4,1,0]=(-0.411133810295)*x_ref[0]**o+(0.988486625099)*x_ref[0]+(-0.36050191367)*x_ref[1]**o+(0.469479821965)*x_ref[1]+(-0.431175357325)*x_ref[2]**o+(-0.966295015263)*x_ref[2]
        ref[3,4,1,1]=(0.233731410367)*x_ref[0]**o+(-0.821900740766)*x_ref[0]+(-0.823123461385)*x_ref[1]**o+(-0.380537914421)*x_ref[1]+(-0.896146300355)*x_ref[2]**o+(-0.775172990257)*x_ref[2]
        ref[3,4,2,0]=(-0.96323363586)*x_ref[0]**o+(0.0935412405245)*x_ref[0]+(0.929333026259)*x_ref[1]**o+(-0.322236627348)*x_ref[1]+(-0.390898135964)*x_ref[2]**o+(-0.766178290162)*x_ref[2]
        ref[3,4,2,1]=(-0.773670533268)*x_ref[0]**o+(-0.156975157098)*x_ref[0]+(-0.91512996795)*x_ref[1]**o+(0.0247803682126)*x_ref[1]+(0.995159733782)*x_ref[2]**o+(-0.669224144225)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.492060381705)*x[0]**o+(0.578791868713)*x[0]+(-0.439849757114)*x[1]**o+(-0.52629328967)*x[1]
        ref=(0.492060381705)*x_ref[0]**o+(0.578791868713)*x_ref[0]+(-0.439849757114)*x_ref[1]**o+(-0.52629328967)*x_ref[1]
      else:
        arg=(0.954737963273)*x[0]**o+(0.452578946878)*x[0]+(0.0471085060698)*x[1]**o+(-0.640169119832)*x[1]+(0.710401490627)*x[2]**o+(-0.382589884739)*x[2]
        ref=(0.954737963273)*x_ref[0]**o+(0.452578946878)*x_ref[0]+(0.0471085060698)*x_ref[1]**o+(-0.640169119832)*x_ref[1]+(0.710401490627)*x_ref[2]**o+(-0.382589884739)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.59604585527)*x[0]**o+(0.826981169401)*x[0]+(0.227617349484)*x[1]**o+(0.184907819029)*x[1]
        arg[1]=(-0.424684036731)*x[0]**o+(0.346537997905)*x[0]+(0.891921272597)*x[1]**o+(-0.145924182872)*x[1]
        ref[0]=(-0.59604585527)*x_ref[0]**o+(0.826981169401)*x_ref[0]+(0.227617349484)*x_ref[1]**o+(0.184907819029)*x_ref[1]
        ref[1]=(-0.424684036731)*x_ref[0]**o+(0.346537997905)*x_ref[0]+(0.891921272597)*x_ref[1]**o+(-0.145924182872)*x_ref[1]
      else:
        arg[0]=(0.777228173978)*x[0]**o+(-0.812215827155)*x[0]+(-0.618922680968)*x[1]**o+(0.0660312103206)*x[1]+(-0.917545504466)*x[2]**o+(0.988524619414)*x[2]
        arg[1]=(0.110219469565)*x[0]**o+(-0.441866866062)*x[0]+(-0.654447764335)*x[1]**o+(0.230716133289)*x[1]+(0.347487317165)*x[2]**o+(-0.948877903043)*x[2]
        ref[0]=(0.777228173978)*x_ref[0]**o+(-0.812215827155)*x_ref[0]+(-0.618922680968)*x_ref[1]**o+(0.0660312103206)*x_ref[1]+(-0.917545504466)*x_ref[2]**o+(0.988524619414)*x_ref[2]
        ref[1]=(0.110219469565)*x_ref[0]**o+(-0.441866866062)*x_ref[0]+(-0.654447764335)*x_ref[1]**o+(0.230716133289)*x_ref[1]+(0.347487317165)*x_ref[2]**o+(-0.948877903043)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.958809754913)*x[0]**o+(-0.844717049041)*x[0]+(0.462564514427)*x[1]**o+(-0.247678335594)*x[1]
        arg[0,1]=(0.700700054189)*x[0]**o+(-0.653724611457)*x[0]+(0.52659177803)*x[1]**o+(-0.387559446418)*x[1]
        arg[0,2]=(-0.326124300134)*x[0]**o+(0.115197897152)*x[0]+(-0.876340641846)*x[1]**o+(0.433060472581)*x[1]
        arg[0,3]=(-0.101215736407)*x[0]**o+(-0.483684510661)*x[0]+(0.0418770907036)*x[1]**o+(-0.845564414421)*x[1]
        arg[0,4]=(-0.293094855718)*x[0]**o+(0.710954930773)*x[0]+(0.238671713869)*x[1]**o+(0.576638343412)*x[1]
        arg[1,0]=(0.529861280122)*x[0]**o+(-0.259437904048)*x[0]+(-0.620725288158)*x[1]**o+(0.326787999382)*x[1]
        arg[1,1]=(0.421245543012)*x[0]**o+(-0.954139636698)*x[0]+(0.0498199734926)*x[1]**o+(0.24010238499)*x[1]
        arg[1,2]=(0.333362362373)*x[0]**o+(-0.999953175952)*x[0]+(0.866193849246)*x[1]**o+(0.764287410177)*x[1]
        arg[1,3]=(0.470415957539)*x[0]**o+(-0.133477966871)*x[0]+(-0.0416970809832)*x[1]**o+(-0.416352740444)*x[1]
        arg[1,4]=(-0.178613018492)*x[0]**o+(-0.967264985978)*x[0]+(-0.297102423687)*x[1]**o+(0.884184973413)*x[1]
        arg[2,0]=(-0.29363453294)*x[0]**o+(-0.0237360594862)*x[0]+(-0.300776450747)*x[1]**o+(-0.493932721521)*x[1]
        arg[2,1]=(0.795045745699)*x[0]**o+(0.984956851154)*x[0]+(-0.724208405235)*x[1]**o+(0.59029067239)*x[1]
        arg[2,2]=(-0.803012112531)*x[0]**o+(-0.831727785428)*x[0]+(0.330295112332)*x[1]**o+(0.0966317908147)*x[1]
        arg[2,3]=(-0.652332429144)*x[0]**o+(-0.38689457016)*x[0]+(-0.391780825295)*x[1]**o+(-0.444284790848)*x[1]
        arg[2,4]=(0.416775161398)*x[0]**o+(-0.16281006536)*x[0]+(-0.742938932867)*x[1]**o+(0.456486183714)*x[1]
        arg[3,0]=(0.506731632086)*x[0]**o+(-0.222468320035)*x[0]+(-0.906398223556)*x[1]**o+(-0.0250630448611)*x[1]
        arg[3,1]=(-0.39207398783)*x[0]**o+(-0.574739374252)*x[0]+(0.262101584569)*x[1]**o+(0.249587535087)*x[1]
        arg[3,2]=(-0.252593913503)*x[0]**o+(-0.864294052989)*x[0]+(0.0110434296336)*x[1]**o+(-0.224585040999)*x[1]
        arg[3,3]=(-0.631704469212)*x[0]**o+(-0.495153785048)*x[0]+(-0.304420746672)*x[1]**o+(-0.254005222021)*x[1]
        arg[3,4]=(0.775631055421)*x[0]**o+(0.370095409961)*x[0]+(0.894627612994)*x[1]**o+(-0.329716708629)*x[1]
        ref[0,0]=(-0.958809754913)*x_ref[0]**o+(-0.844717049041)*x_ref[0]+(0.462564514427)*x_ref[1]**o+(-0.247678335594)*x_ref[1]
        ref[0,1]=(0.700700054189)*x_ref[0]**o+(-0.653724611457)*x_ref[0]+(0.52659177803)*x_ref[1]**o+(-0.387559446418)*x_ref[1]
        ref[0,2]=(-0.326124300134)*x_ref[0]**o+(0.115197897152)*x_ref[0]+(-0.876340641846)*x_ref[1]**o+(0.433060472581)*x_ref[1]
        ref[0,3]=(-0.101215736407)*x_ref[0]**o+(-0.483684510661)*x_ref[0]+(0.0418770907036)*x_ref[1]**o+(-0.845564414421)*x_ref[1]
        ref[0,4]=(-0.293094855718)*x_ref[0]**o+(0.710954930773)*x_ref[0]+(0.238671713869)*x_ref[1]**o+(0.576638343412)*x_ref[1]
        ref[1,0]=(0.529861280122)*x_ref[0]**o+(-0.259437904048)*x_ref[0]+(-0.620725288158)*x_ref[1]**o+(0.326787999382)*x_ref[1]
        ref[1,1]=(0.421245543012)*x_ref[0]**o+(-0.954139636698)*x_ref[0]+(0.0498199734926)*x_ref[1]**o+(0.24010238499)*x_ref[1]
        ref[1,2]=(0.333362362373)*x_ref[0]**o+(-0.999953175952)*x_ref[0]+(0.866193849246)*x_ref[1]**o+(0.764287410177)*x_ref[1]
        ref[1,3]=(0.470415957539)*x_ref[0]**o+(-0.133477966871)*x_ref[0]+(-0.0416970809832)*x_ref[1]**o+(-0.416352740444)*x_ref[1]
        ref[1,4]=(-0.178613018492)*x_ref[0]**o+(-0.967264985978)*x_ref[0]+(-0.297102423687)*x_ref[1]**o+(0.884184973413)*x_ref[1]
        ref[2,0]=(-0.29363453294)*x_ref[0]**o+(-0.0237360594862)*x_ref[0]+(-0.300776450747)*x_ref[1]**o+(-0.493932721521)*x_ref[1]
        ref[2,1]=(0.795045745699)*x_ref[0]**o+(0.984956851154)*x_ref[0]+(-0.724208405235)*x_ref[1]**o+(0.59029067239)*x_ref[1]
        ref[2,2]=(-0.803012112531)*x_ref[0]**o+(-0.831727785428)*x_ref[0]+(0.330295112332)*x_ref[1]**o+(0.0966317908147)*x_ref[1]
        ref[2,3]=(-0.652332429144)*x_ref[0]**o+(-0.38689457016)*x_ref[0]+(-0.391780825295)*x_ref[1]**o+(-0.444284790848)*x_ref[1]
        ref[2,4]=(0.416775161398)*x_ref[0]**o+(-0.16281006536)*x_ref[0]+(-0.742938932867)*x_ref[1]**o+(0.456486183714)*x_ref[1]
        ref[3,0]=(0.506731632086)*x_ref[0]**o+(-0.222468320035)*x_ref[0]+(-0.906398223556)*x_ref[1]**o+(-0.0250630448611)*x_ref[1]
        ref[3,1]=(-0.39207398783)*x_ref[0]**o+(-0.574739374252)*x_ref[0]+(0.262101584569)*x_ref[1]**o+(0.249587535087)*x_ref[1]
        ref[3,2]=(-0.252593913503)*x_ref[0]**o+(-0.864294052989)*x_ref[0]+(0.0110434296336)*x_ref[1]**o+(-0.224585040999)*x_ref[1]
        ref[3,3]=(-0.631704469212)*x_ref[0]**o+(-0.495153785048)*x_ref[0]+(-0.304420746672)*x_ref[1]**o+(-0.254005222021)*x_ref[1]
        ref[3,4]=(0.775631055421)*x_ref[0]**o+(0.370095409961)*x_ref[0]+(0.894627612994)*x_ref[1]**o+(-0.329716708629)*x_ref[1]
      else:
        arg[0,0]=(0.392450639899)*x[0]**o+(-0.685488230002)*x[0]+(-0.41545928428)*x[1]**o+(0.244144392379)*x[1]+(0.602642156528)*x[2]**o+(-0.348717378461)*x[2]
        arg[0,1]=(-0.341243165541)*x[0]**o+(-0.616870164204)*x[0]+(-0.510613066362)*x[1]**o+(-0.81387010212)*x[1]+(0.856239691119)*x[2]**o+(0.0744651854239)*x[2]
        arg[0,2]=(-0.765983413604)*x[0]**o+(0.474551394174)*x[0]+(-0.939524929765)*x[1]**o+(0.418297898065)*x[1]+(0.00589909539531)*x[2]**o+(0.0889012994897)*x[2]
        arg[0,3]=(-0.521217563352)*x[0]**o+(-0.256316459647)*x[0]+(-0.0265965098379)*x[1]**o+(-0.597562278966)*x[1]+(0.0389151661339)*x[2]**o+(-0.572109451352)*x[2]
        arg[0,4]=(-0.853801268491)*x[0]**o+(-0.691274812057)*x[0]+(-0.682435618581)*x[1]**o+(-0.974334375359)*x[1]+(-0.642333316794)*x[2]**o+(0.353816751252)*x[2]
        arg[1,0]=(-0.657545394322)*x[0]**o+(-0.334172338096)*x[0]+(0.016748279459)*x[1]**o+(-0.294858621125)*x[1]+(-0.49403880046)*x[2]**o+(-0.152416774822)*x[2]
        arg[1,1]=(-0.602731976042)*x[0]**o+(0.685046816545)*x[0]+(0.750001644716)*x[1]**o+(-0.202076972289)*x[1]+(-0.0232888078434)*x[2]**o+(-0.0774477649878)*x[2]
        arg[1,2]=(-0.707321785527)*x[0]**o+(0.618526464205)*x[0]+(-0.578548442468)*x[1]**o+(-0.411374744442)*x[1]+(-0.944277708787)*x[2]**o+(0.647856975707)*x[2]
        arg[1,3]=(-0.222503271022)*x[0]**o+(-0.929428092728)*x[0]+(-0.976553212424)*x[1]**o+(0.0376439617084)*x[1]+(0.0428078568512)*x[2]**o+(-0.441850944272)*x[2]
        arg[1,4]=(-0.862164898455)*x[0]**o+(-0.989900221991)*x[0]+(0.110610844567)*x[1]**o+(0.419384612236)*x[1]+(-0.00271948841262)*x[2]**o+(0.301635101629)*x[2]
        arg[2,0]=(-0.0950421881942)*x[0]**o+(-0.741293053909)*x[0]+(-0.461538879874)*x[1]**o+(-0.8199842553)*x[1]+(0.0590344916829)*x[2]**o+(0.352541518434)*x[2]
        arg[2,1]=(-0.734407995647)*x[0]**o+(-0.726167721992)*x[0]+(0.700521229916)*x[1]**o+(0.753555863428)*x[1]+(-0.891638139696)*x[2]**o+(-0.996550711213)*x[2]
        arg[2,2]=(0.777110352629)*x[0]**o+(0.309719201157)*x[0]+(0.026816547522)*x[1]**o+(0.33161010287)*x[1]+(0.257929504352)*x[2]**o+(0.71871314654)*x[2]
        arg[2,3]=(-0.0982522143825)*x[0]**o+(0.0532144836415)*x[0]+(0.155895442381)*x[1]**o+(-0.872222127033)*x[1]+(0.810538165596)*x[2]**o+(-0.917071363009)*x[2]
        arg[2,4]=(0.882674783877)*x[0]**o+(-0.246439463648)*x[0]+(-0.478821457401)*x[1]**o+(-0.629183697296)*x[1]+(0.833283185393)*x[2]**o+(-0.895796651184)*x[2]
        arg[3,0]=(0.949019461395)*x[0]**o+(0.437511668107)*x[0]+(0.46679355983)*x[1]**o+(0.375391913775)*x[1]+(0.992336073925)*x[2]**o+(0.784847856608)*x[2]
        arg[3,1]=(0.10730061928)*x[0]**o+(0.624678931808)*x[0]+(-0.105084917663)*x[1]**o+(-0.0650034223387)*x[1]+(-0.271922353868)*x[2]**o+(-0.182967296038)*x[2]
        arg[3,2]=(0.115856234055)*x[0]**o+(0.711912354922)*x[0]+(-0.317962972122)*x[1]**o+(0.577856092951)*x[1]+(-0.246833922867)*x[2]**o+(0.354116462218)*x[2]
        arg[3,3]=(0.40937236699)*x[0]**o+(0.639252679403)*x[0]+(0.540206943219)*x[1]**o+(0.287006139758)*x[1]+(0.123311433598)*x[2]**o+(-0.0300672780213)*x[2]
        arg[3,4]=(-0.190299257713)*x[0]**o+(-0.022055675261)*x[0]+(-0.640751994447)*x[1]**o+(-0.411801947317)*x[1]+(0.0675008125057)*x[2]**o+(-0.659424255203)*x[2]
        ref[0,0]=(0.392450639899)*x_ref[0]**o+(-0.685488230002)*x_ref[0]+(-0.41545928428)*x_ref[1]**o+(0.244144392379)*x_ref[1]+(0.602642156528)*x_ref[2]**o+(-0.348717378461)*x_ref[2]
        ref[0,1]=(-0.341243165541)*x_ref[0]**o+(-0.616870164204)*x_ref[0]+(-0.510613066362)*x_ref[1]**o+(-0.81387010212)*x_ref[1]+(0.856239691119)*x_ref[2]**o+(0.0744651854239)*x_ref[2]
        ref[0,2]=(-0.765983413604)*x_ref[0]**o+(0.474551394174)*x_ref[0]+(-0.939524929765)*x_ref[1]**o+(0.418297898065)*x_ref[1]+(0.00589909539531)*x_ref[2]**o+(0.0889012994897)*x_ref[2]
        ref[0,3]=(-0.521217563352)*x_ref[0]**o+(-0.256316459647)*x_ref[0]+(-0.0265965098379)*x_ref[1]**o+(-0.597562278966)*x_ref[1]+(0.0389151661339)*x_ref[2]**o+(-0.572109451352)*x_ref[2]
        ref[0,4]=(-0.853801268491)*x_ref[0]**o+(-0.691274812057)*x_ref[0]+(-0.682435618581)*x_ref[1]**o+(-0.974334375359)*x_ref[1]+(-0.642333316794)*x_ref[2]**o+(0.353816751252)*x_ref[2]
        ref[1,0]=(-0.657545394322)*x_ref[0]**o+(-0.334172338096)*x_ref[0]+(0.016748279459)*x_ref[1]**o+(-0.294858621125)*x_ref[1]+(-0.49403880046)*x_ref[2]**o+(-0.152416774822)*x_ref[2]
        ref[1,1]=(-0.602731976042)*x_ref[0]**o+(0.685046816545)*x_ref[0]+(0.750001644716)*x_ref[1]**o+(-0.202076972289)*x_ref[1]+(-0.0232888078434)*x_ref[2]**o+(-0.0774477649878)*x_ref[2]
        ref[1,2]=(-0.707321785527)*x_ref[0]**o+(0.618526464205)*x_ref[0]+(-0.578548442468)*x_ref[1]**o+(-0.411374744442)*x_ref[1]+(-0.944277708787)*x_ref[2]**o+(0.647856975707)*x_ref[2]
        ref[1,3]=(-0.222503271022)*x_ref[0]**o+(-0.929428092728)*x_ref[0]+(-0.976553212424)*x_ref[1]**o+(0.0376439617084)*x_ref[1]+(0.0428078568512)*x_ref[2]**o+(-0.441850944272)*x_ref[2]
        ref[1,4]=(-0.862164898455)*x_ref[0]**o+(-0.989900221991)*x_ref[0]+(0.110610844567)*x_ref[1]**o+(0.419384612236)*x_ref[1]+(-0.00271948841262)*x_ref[2]**o+(0.301635101629)*x_ref[2]
        ref[2,0]=(-0.0950421881942)*x_ref[0]**o+(-0.741293053909)*x_ref[0]+(-0.461538879874)*x_ref[1]**o+(-0.8199842553)*x_ref[1]+(0.0590344916829)*x_ref[2]**o+(0.352541518434)*x_ref[2]
        ref[2,1]=(-0.734407995647)*x_ref[0]**o+(-0.726167721992)*x_ref[0]+(0.700521229916)*x_ref[1]**o+(0.753555863428)*x_ref[1]+(-0.891638139696)*x_ref[2]**o+(-0.996550711213)*x_ref[2]
        ref[2,2]=(0.777110352629)*x_ref[0]**o+(0.309719201157)*x_ref[0]+(0.026816547522)*x_ref[1]**o+(0.33161010287)*x_ref[1]+(0.257929504352)*x_ref[2]**o+(0.71871314654)*x_ref[2]
        ref[2,3]=(-0.0982522143825)*x_ref[0]**o+(0.0532144836415)*x_ref[0]+(0.155895442381)*x_ref[1]**o+(-0.872222127033)*x_ref[1]+(0.810538165596)*x_ref[2]**o+(-0.917071363009)*x_ref[2]
        ref[2,4]=(0.882674783877)*x_ref[0]**o+(-0.246439463648)*x_ref[0]+(-0.478821457401)*x_ref[1]**o+(-0.629183697296)*x_ref[1]+(0.833283185393)*x_ref[2]**o+(-0.895796651184)*x_ref[2]
        ref[3,0]=(0.949019461395)*x_ref[0]**o+(0.437511668107)*x_ref[0]+(0.46679355983)*x_ref[1]**o+(0.375391913775)*x_ref[1]+(0.992336073925)*x_ref[2]**o+(0.784847856608)*x_ref[2]
        ref[3,1]=(0.10730061928)*x_ref[0]**o+(0.624678931808)*x_ref[0]+(-0.105084917663)*x_ref[1]**o+(-0.0650034223387)*x_ref[1]+(-0.271922353868)*x_ref[2]**o+(-0.182967296038)*x_ref[2]
        ref[3,2]=(0.115856234055)*x_ref[0]**o+(0.711912354922)*x_ref[0]+(-0.317962972122)*x_ref[1]**o+(0.577856092951)*x_ref[1]+(-0.246833922867)*x_ref[2]**o+(0.354116462218)*x_ref[2]
        ref[3,3]=(0.40937236699)*x_ref[0]**o+(0.639252679403)*x_ref[0]+(0.540206943219)*x_ref[1]**o+(0.287006139758)*x_ref[1]+(0.123311433598)*x_ref[2]**o+(-0.0300672780213)*x_ref[2]
        ref[3,4]=(-0.190299257713)*x_ref[0]**o+(-0.022055675261)*x_ref[0]+(-0.640751994447)*x_ref[1]**o+(-0.411801947317)*x_ref[1]+(0.0675008125057)*x_ref[2]**o+(-0.659424255203)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.697323110242)*x[0]**o+(-0.806494185218)*x[0]+(0.387914602484)*x[1]**o+(0.841835487965)*x[1]
        arg[0,0,1]=(-0.77670653908)*x[0]**o+(0.311710212706)*x[0]+(-0.0951864020181)*x[1]**o+(-0.14909139946)*x[1]
        arg[0,1,0]=(0.80933791543)*x[0]**o+(0.72441968276)*x[0]+(-0.446054850085)*x[1]**o+(0.419303534988)*x[1]
        arg[0,1,1]=(0.174464190017)*x[0]**o+(0.330609661878)*x[0]+(0.798365390058)*x[1]**o+(0.354899223817)*x[1]
        arg[1,0,0]=(0.36073672839)*x[0]**o+(0.831900846214)*x[0]+(-0.656354645044)*x[1]**o+(0.200645044646)*x[1]
        arg[1,0,1]=(0.817889801065)*x[0]**o+(0.34052889294)*x[0]+(-0.2416944924)*x[1]**o+(0.0922342148934)*x[1]
        arg[1,1,0]=(-0.962116509558)*x[0]**o+(-0.0785232077542)*x[0]+(-0.981399563356)*x[1]**o+(0.785509095614)*x[1]
        arg[1,1,1]=(-0.688444030508)*x[0]**o+(0.651617638972)*x[0]+(0.160584046401)*x[1]**o+(0.253258991923)*x[1]
        arg[2,0,0]=(0.0728970412574)*x[0]**o+(-0.647328240217)*x[0]+(0.517665494535)*x[1]**o+(-0.416383753573)*x[1]
        arg[2,0,1]=(0.148104739733)*x[0]**o+(-0.378371639837)*x[0]+(0.848608631336)*x[1]**o+(-0.881516167852)*x[1]
        arg[2,1,0]=(-0.506376676256)*x[0]**o+(0.234090340446)*x[0]+(-0.737828959566)*x[1]**o+(-0.74093763135)*x[1]
        arg[2,1,1]=(0.259605354135)*x[0]**o+(0.537424356513)*x[0]+(-0.405718119693)*x[1]**o+(0.120663712044)*x[1]
        arg[3,0,0]=(-0.609659975269)*x[0]**o+(-0.455923293874)*x[0]+(-0.00765676965971)*x[1]**o+(-0.710899349259)*x[1]
        arg[3,0,1]=(-0.731011612904)*x[0]**o+(-0.355060765358)*x[0]+(-0.127369372538)*x[1]**o+(0.694265371908)*x[1]
        arg[3,1,0]=(0.188766337132)*x[0]**o+(-0.194878304215)*x[0]+(-0.963356218109)*x[1]**o+(0.327053372293)*x[1]
        arg[3,1,1]=(-0.566712779324)*x[0]**o+(-0.251473305071)*x[0]+(0.0181297146138)*x[1]**o+(0.257664697754)*x[1]
        arg[4,0,0]=(0.574100764569)*x[0]**o+(-0.717776999574)*x[0]+(0.880998124155)*x[1]**o+(-0.810615928848)*x[1]
        arg[4,0,1]=(0.42552502518)*x[0]**o+(0.456751341147)*x[0]+(-0.484998619005)*x[1]**o+(0.932151187696)*x[1]
        arg[4,1,0]=(0.56865980485)*x[0]**o+(0.593928872734)*x[0]+(-0.475639894836)*x[1]**o+(-0.990718402693)*x[1]
        arg[4,1,1]=(0.129586342261)*x[0]**o+(-0.408951830619)*x[0]+(0.132518475289)*x[1]**o+(0.708612820692)*x[1]
        arg[5,0,0]=(0.0708038739082)*x[0]**o+(0.524270558649)*x[0]+(0.368504533482)*x[1]**o+(0.935828810393)*x[1]
        arg[5,0,1]=(-0.383650306829)*x[0]**o+(-0.0439362943961)*x[0]+(0.80939275198)*x[1]**o+(-0.61335508023)*x[1]
        arg[5,1,0]=(-0.621571339247)*x[0]**o+(-0.185069293199)*x[0]+(0.368472011431)*x[1]**o+(-0.388752495932)*x[1]
        arg[5,1,1]=(-0.263310763174)*x[0]**o+(-0.694616630015)*x[0]+(0.271238274203)*x[1]**o+(-0.477861402347)*x[1]
        ref[0,0,0]=(-0.697323110242)*x_ref[0]**o+(-0.806494185218)*x_ref[0]+(0.387914602484)*x_ref[1]**o+(0.841835487965)*x_ref[1]
        ref[0,0,1]=(-0.77670653908)*x_ref[0]**o+(0.311710212706)*x_ref[0]+(-0.0951864020181)*x_ref[1]**o+(-0.14909139946)*x_ref[1]
        ref[0,1,0]=(0.80933791543)*x_ref[0]**o+(0.72441968276)*x_ref[0]+(-0.446054850085)*x_ref[1]**o+(0.419303534988)*x_ref[1]
        ref[0,1,1]=(0.174464190017)*x_ref[0]**o+(0.330609661878)*x_ref[0]+(0.798365390058)*x_ref[1]**o+(0.354899223817)*x_ref[1]
        ref[1,0,0]=(0.36073672839)*x_ref[0]**o+(0.831900846214)*x_ref[0]+(-0.656354645044)*x_ref[1]**o+(0.200645044646)*x_ref[1]
        ref[1,0,1]=(0.817889801065)*x_ref[0]**o+(0.34052889294)*x_ref[0]+(-0.2416944924)*x_ref[1]**o+(0.0922342148934)*x_ref[1]
        ref[1,1,0]=(-0.962116509558)*x_ref[0]**o+(-0.0785232077542)*x_ref[0]+(-0.981399563356)*x_ref[1]**o+(0.785509095614)*x_ref[1]
        ref[1,1,1]=(-0.688444030508)*x_ref[0]**o+(0.651617638972)*x_ref[0]+(0.160584046401)*x_ref[1]**o+(0.253258991923)*x_ref[1]
        ref[2,0,0]=(0.0728970412574)*x_ref[0]**o+(-0.647328240217)*x_ref[0]+(0.517665494535)*x_ref[1]**o+(-0.416383753573)*x_ref[1]
        ref[2,0,1]=(0.148104739733)*x_ref[0]**o+(-0.378371639837)*x_ref[0]+(0.848608631336)*x_ref[1]**o+(-0.881516167852)*x_ref[1]
        ref[2,1,0]=(-0.506376676256)*x_ref[0]**o+(0.234090340446)*x_ref[0]+(-0.737828959566)*x_ref[1]**o+(-0.74093763135)*x_ref[1]
        ref[2,1,1]=(0.259605354135)*x_ref[0]**o+(0.537424356513)*x_ref[0]+(-0.405718119693)*x_ref[1]**o+(0.120663712044)*x_ref[1]
        ref[3,0,0]=(-0.609659975269)*x_ref[0]**o+(-0.455923293874)*x_ref[0]+(-0.00765676965971)*x_ref[1]**o+(-0.710899349259)*x_ref[1]
        ref[3,0,1]=(-0.731011612904)*x_ref[0]**o+(-0.355060765358)*x_ref[0]+(-0.127369372538)*x_ref[1]**o+(0.694265371908)*x_ref[1]
        ref[3,1,0]=(0.188766337132)*x_ref[0]**o+(-0.194878304215)*x_ref[0]+(-0.963356218109)*x_ref[1]**o+(0.327053372293)*x_ref[1]
        ref[3,1,1]=(-0.566712779324)*x_ref[0]**o+(-0.251473305071)*x_ref[0]+(0.0181297146138)*x_ref[1]**o+(0.257664697754)*x_ref[1]
        ref[4,0,0]=(0.574100764569)*x_ref[0]**o+(-0.717776999574)*x_ref[0]+(0.880998124155)*x_ref[1]**o+(-0.810615928848)*x_ref[1]
        ref[4,0,1]=(0.42552502518)*x_ref[0]**o+(0.456751341147)*x_ref[0]+(-0.484998619005)*x_ref[1]**o+(0.932151187696)*x_ref[1]
        ref[4,1,0]=(0.56865980485)*x_ref[0]**o+(0.593928872734)*x_ref[0]+(-0.475639894836)*x_ref[1]**o+(-0.990718402693)*x_ref[1]
        ref[4,1,1]=(0.129586342261)*x_ref[0]**o+(-0.408951830619)*x_ref[0]+(0.132518475289)*x_ref[1]**o+(0.708612820692)*x_ref[1]
        ref[5,0,0]=(0.0708038739082)*x_ref[0]**o+(0.524270558649)*x_ref[0]+(0.368504533482)*x_ref[1]**o+(0.935828810393)*x_ref[1]
        ref[5,0,1]=(-0.383650306829)*x_ref[0]**o+(-0.0439362943961)*x_ref[0]+(0.80939275198)*x_ref[1]**o+(-0.61335508023)*x_ref[1]
        ref[5,1,0]=(-0.621571339247)*x_ref[0]**o+(-0.185069293199)*x_ref[0]+(0.368472011431)*x_ref[1]**o+(-0.388752495932)*x_ref[1]
        ref[5,1,1]=(-0.263310763174)*x_ref[0]**o+(-0.694616630015)*x_ref[0]+(0.271238274203)*x_ref[1]**o+(-0.477861402347)*x_ref[1]
      else:
        arg[0,0,0]=(0.287534645778)*x[0]**o+(0.94865229379)*x[0]+(0.655299232864)*x[1]**o+(0.350034924637)*x[1]+(0.556341391183)*x[2]**o+(0.840643265765)*x[2]
        arg[0,0,1]=(-0.207213781549)*x[0]**o+(-0.0394497047797)*x[0]+(-0.0236016308575)*x[1]**o+(0.344122299199)*x[1]+(0.389588325996)*x[2]**o+(0.941980468657)*x[2]
        arg[0,1,0]=(0.00980412235862)*x[0]**o+(0.74862549447)*x[0]+(0.630165284383)*x[1]**o+(0.234711268989)*x[1]+(-0.22509981272)*x[2]**o+(-0.0153888168649)*x[2]
        arg[0,1,1]=(-0.87309186262)*x[0]**o+(0.833526923917)*x[0]+(0.78785479415)*x[1]**o+(0.823967259171)*x[1]+(0.320244316109)*x[2]**o+(-0.735225854417)*x[2]
        arg[1,0,0]=(-0.0892232551051)*x[0]**o+(-0.346728833563)*x[0]+(0.422774634671)*x[1]**o+(0.793222381957)*x[1]+(-0.670476677006)*x[2]**o+(0.896884540116)*x[2]
        arg[1,0,1]=(-0.0023382136345)*x[0]**o+(0.73058281545)*x[0]+(0.11828880374)*x[1]**o+(-0.691839091619)*x[1]+(-0.39086463392)*x[2]**o+(0.66200209782)*x[2]
        arg[1,1,0]=(-0.924238055558)*x[0]**o+(-0.334208271705)*x[0]+(-0.57106045555)*x[1]**o+(-0.765650999193)*x[1]+(0.0404533468054)*x[2]**o+(-0.225975260777)*x[2]
        arg[1,1,1]=(-0.351110999108)*x[0]**o+(0.626166346251)*x[0]+(0.385737198467)*x[1]**o+(0.577561466908)*x[1]+(0.708325217242)*x[2]**o+(0.193727923904)*x[2]
        arg[2,0,0]=(0.278942080997)*x[0]**o+(-0.569761653572)*x[0]+(-0.624879270918)*x[1]**o+(-0.0151754112896)*x[1]+(0.411399083825)*x[2]**o+(-0.846669568641)*x[2]
        arg[2,0,1]=(-0.543586963879)*x[0]**o+(-0.383959956625)*x[0]+(0.184375944938)*x[1]**o+(-0.104178233562)*x[1]+(-0.223101894662)*x[2]**o+(-0.17221835595)*x[2]
        arg[2,1,0]=(0.317103921207)*x[0]**o+(-0.919542461354)*x[0]+(0.56666554454)*x[1]**o+(0.895772298847)*x[1]+(0.714550707822)*x[2]**o+(0.634702089556)*x[2]
        arg[2,1,1]=(0.986576254164)*x[0]**o+(0.544493787633)*x[0]+(0.964620027203)*x[1]**o+(0.259991384678)*x[1]+(-0.577361835739)*x[2]**o+(0.651593456822)*x[2]
        arg[3,0,0]=(0.362841472614)*x[0]**o+(-0.521771228041)*x[0]+(-0.594530755517)*x[1]**o+(0.700647681241)*x[1]+(-0.131683781311)*x[2]**o+(0.636310189834)*x[2]
        arg[3,0,1]=(0.885891862321)*x[0]**o+(0.397438320764)*x[0]+(0.711380452937)*x[1]**o+(0.962571325144)*x[1]+(-0.300114113339)*x[2]**o+(0.34700388868)*x[2]
        arg[3,1,0]=(0.52140726408)*x[0]**o+(-0.0921295295522)*x[0]+(-0.823076091446)*x[1]**o+(0.477401876125)*x[1]+(-0.973057905067)*x[2]**o+(-0.975526794128)*x[2]
        arg[3,1,1]=(-0.241503887113)*x[0]**o+(0.133988088285)*x[0]+(0.243026393121)*x[1]**o+(0.0848716779473)*x[1]+(0.140349614791)*x[2]**o+(-0.217182025695)*x[2]
        arg[4,0,0]=(0.680080194053)*x[0]**o+(0.512009633844)*x[0]+(-0.640214066363)*x[1]**o+(0.571800633214)*x[1]+(-0.749395617545)*x[2]**o+(0.448487668786)*x[2]
        arg[4,0,1]=(-0.792959916349)*x[0]**o+(-0.433770516776)*x[0]+(-0.380980758215)*x[1]**o+(-0.831154769101)*x[1]+(0.277551460651)*x[2]**o+(0.626476984233)*x[2]
        arg[4,1,0]=(-0.285102566546)*x[0]**o+(0.745636275253)*x[0]+(0.4999427836)*x[1]**o+(0.561343302884)*x[1]+(0.796220302361)*x[2]**o+(-0.830020520143)*x[2]
        arg[4,1,1]=(0.237420507314)*x[0]**o+(0.327181858092)*x[0]+(-0.895743235593)*x[1]**o+(-0.496964477443)*x[1]+(0.57336428204)*x[2]**o+(-0.154267071287)*x[2]
        arg[5,0,0]=(0.0401925997607)*x[0]**o+(-0.0155957424748)*x[0]+(-0.658363451239)*x[1]**o+(-0.175223211879)*x[1]+(-0.743729768327)*x[2]**o+(0.391531327128)*x[2]
        arg[5,0,1]=(0.836239801824)*x[0]**o+(-0.72288085675)*x[0]+(0.177981042887)*x[1]**o+(0.354505260098)*x[1]+(-0.180520504267)*x[2]**o+(-0.380306239974)*x[2]
        arg[5,1,0]=(-0.65536853024)*x[0]**o+(-0.00573854242476)*x[0]+(-0.261460514314)*x[1]**o+(-0.289771405693)*x[1]+(0.750775397523)*x[2]**o+(-0.333215608562)*x[2]
        arg[5,1,1]=(-0.579532884024)*x[0]**o+(-0.208677373724)*x[0]+(-0.379612024024)*x[1]**o+(-0.0565813995114)*x[1]+(-0.867913562592)*x[2]**o+(-0.254654298293)*x[2]
        ref[0,0,0]=(0.287534645778)*x_ref[0]**o+(0.94865229379)*x_ref[0]+(0.655299232864)*x_ref[1]**o+(0.350034924637)*x_ref[1]+(0.556341391183)*x_ref[2]**o+(0.840643265765)*x_ref[2]
        ref[0,0,1]=(-0.207213781549)*x_ref[0]**o+(-0.0394497047797)*x_ref[0]+(-0.0236016308575)*x_ref[1]**o+(0.344122299199)*x_ref[1]+(0.389588325996)*x_ref[2]**o+(0.941980468657)*x_ref[2]
        ref[0,1,0]=(0.00980412235862)*x_ref[0]**o+(0.74862549447)*x_ref[0]+(0.630165284383)*x_ref[1]**o+(0.234711268989)*x_ref[1]+(-0.22509981272)*x_ref[2]**o+(-0.0153888168649)*x_ref[2]
        ref[0,1,1]=(-0.87309186262)*x_ref[0]**o+(0.833526923917)*x_ref[0]+(0.78785479415)*x_ref[1]**o+(0.823967259171)*x_ref[1]+(0.320244316109)*x_ref[2]**o+(-0.735225854417)*x_ref[2]
        ref[1,0,0]=(-0.0892232551051)*x_ref[0]**o+(-0.346728833563)*x_ref[0]+(0.422774634671)*x_ref[1]**o+(0.793222381957)*x_ref[1]+(-0.670476677006)*x_ref[2]**o+(0.896884540116)*x_ref[2]
        ref[1,0,1]=(-0.0023382136345)*x_ref[0]**o+(0.73058281545)*x_ref[0]+(0.11828880374)*x_ref[1]**o+(-0.691839091619)*x_ref[1]+(-0.39086463392)*x_ref[2]**o+(0.66200209782)*x_ref[2]
        ref[1,1,0]=(-0.924238055558)*x_ref[0]**o+(-0.334208271705)*x_ref[0]+(-0.57106045555)*x_ref[1]**o+(-0.765650999193)*x_ref[1]+(0.0404533468054)*x_ref[2]**o+(-0.225975260777)*x_ref[2]
        ref[1,1,1]=(-0.351110999108)*x_ref[0]**o+(0.626166346251)*x_ref[0]+(0.385737198467)*x_ref[1]**o+(0.577561466908)*x_ref[1]+(0.708325217242)*x_ref[2]**o+(0.193727923904)*x_ref[2]
        ref[2,0,0]=(0.278942080997)*x_ref[0]**o+(-0.569761653572)*x_ref[0]+(-0.624879270918)*x_ref[1]**o+(-0.0151754112896)*x_ref[1]+(0.411399083825)*x_ref[2]**o+(-0.846669568641)*x_ref[2]
        ref[2,0,1]=(-0.543586963879)*x_ref[0]**o+(-0.383959956625)*x_ref[0]+(0.184375944938)*x_ref[1]**o+(-0.104178233562)*x_ref[1]+(-0.223101894662)*x_ref[2]**o+(-0.17221835595)*x_ref[2]
        ref[2,1,0]=(0.317103921207)*x_ref[0]**o+(-0.919542461354)*x_ref[0]+(0.56666554454)*x_ref[1]**o+(0.895772298847)*x_ref[1]+(0.714550707822)*x_ref[2]**o+(0.634702089556)*x_ref[2]
        ref[2,1,1]=(0.986576254164)*x_ref[0]**o+(0.544493787633)*x_ref[0]+(0.964620027203)*x_ref[1]**o+(0.259991384678)*x_ref[1]+(-0.577361835739)*x_ref[2]**o+(0.651593456822)*x_ref[2]
        ref[3,0,0]=(0.362841472614)*x_ref[0]**o+(-0.521771228041)*x_ref[0]+(-0.594530755517)*x_ref[1]**o+(0.700647681241)*x_ref[1]+(-0.131683781311)*x_ref[2]**o+(0.636310189834)*x_ref[2]
        ref[3,0,1]=(0.885891862321)*x_ref[0]**o+(0.397438320764)*x_ref[0]+(0.711380452937)*x_ref[1]**o+(0.962571325144)*x_ref[1]+(-0.300114113339)*x_ref[2]**o+(0.34700388868)*x_ref[2]
        ref[3,1,0]=(0.52140726408)*x_ref[0]**o+(-0.0921295295522)*x_ref[0]+(-0.823076091446)*x_ref[1]**o+(0.477401876125)*x_ref[1]+(-0.973057905067)*x_ref[2]**o+(-0.975526794128)*x_ref[2]
        ref[3,1,1]=(-0.241503887113)*x_ref[0]**o+(0.133988088285)*x_ref[0]+(0.243026393121)*x_ref[1]**o+(0.0848716779473)*x_ref[1]+(0.140349614791)*x_ref[2]**o+(-0.217182025695)*x_ref[2]
        ref[4,0,0]=(0.680080194053)*x_ref[0]**o+(0.512009633844)*x_ref[0]+(-0.640214066363)*x_ref[1]**o+(0.571800633214)*x_ref[1]+(-0.749395617545)*x_ref[2]**o+(0.448487668786)*x_ref[2]
        ref[4,0,1]=(-0.792959916349)*x_ref[0]**o+(-0.433770516776)*x_ref[0]+(-0.380980758215)*x_ref[1]**o+(-0.831154769101)*x_ref[1]+(0.277551460651)*x_ref[2]**o+(0.626476984233)*x_ref[2]
        ref[4,1,0]=(-0.285102566546)*x_ref[0]**o+(0.745636275253)*x_ref[0]+(0.4999427836)*x_ref[1]**o+(0.561343302884)*x_ref[1]+(0.796220302361)*x_ref[2]**o+(-0.830020520143)*x_ref[2]
        ref[4,1,1]=(0.237420507314)*x_ref[0]**o+(0.327181858092)*x_ref[0]+(-0.895743235593)*x_ref[1]**o+(-0.496964477443)*x_ref[1]+(0.57336428204)*x_ref[2]**o+(-0.154267071287)*x_ref[2]
        ref[5,0,0]=(0.0401925997607)*x_ref[0]**o+(-0.0155957424748)*x_ref[0]+(-0.658363451239)*x_ref[1]**o+(-0.175223211879)*x_ref[1]+(-0.743729768327)*x_ref[2]**o+(0.391531327128)*x_ref[2]
        ref[5,0,1]=(0.836239801824)*x_ref[0]**o+(-0.72288085675)*x_ref[0]+(0.177981042887)*x_ref[1]**o+(0.354505260098)*x_ref[1]+(-0.180520504267)*x_ref[2]**o+(-0.380306239974)*x_ref[2]
        ref[5,1,0]=(-0.65536853024)*x_ref[0]**o+(-0.00573854242476)*x_ref[0]+(-0.261460514314)*x_ref[1]**o+(-0.289771405693)*x_ref[1]+(0.750775397523)*x_ref[2]**o+(-0.333215608562)*x_ref[2]
        ref[5,1,1]=(-0.579532884024)*x_ref[0]**o+(-0.208677373724)*x_ref[0]+(-0.379612024024)*x_ref[1]**o+(-0.0565813995114)*x_ref[1]+(-0.867913562592)*x_ref[2]**o+(-0.254654298293)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.274472293956)*x[0]**o+(-0.293347130929)*x[0]+(0.392586448254)*x[1]**o+(-0.403066854054)*x[1]
        arg[0,0,0,1]=(0.385933192907)*x[0]**o+(0.484240065867)*x[0]+(0.144565631551)*x[1]**o+(0.882508492739)*x[1]
        arg[0,0,1,0]=(0.993368905137)*x[0]**o+(0.340349919343)*x[0]+(-0.956806746524)*x[1]**o+(0.990607958001)*x[1]
        arg[0,0,1,1]=(-0.88610250102)*x[0]**o+(-0.543705587259)*x[0]+(0.170798164924)*x[1]**o+(0.685331195754)*x[1]
        arg[0,0,2,0]=(-0.922211611143)*x[0]**o+(-0.103640037893)*x[0]+(0.401354703434)*x[1]**o+(-0.117826261635)*x[1]
        arg[0,0,2,1]=(-0.391601387132)*x[0]**o+(-0.472753344804)*x[0]+(-0.473433838993)*x[1]**o+(-0.173501635989)*x[1]
        arg[0,1,0,0]=(0.747353274916)*x[0]**o+(-0.960856612086)*x[0]+(-0.54809012978)*x[1]**o+(0.862023893332)*x[1]
        arg[0,1,0,1]=(-0.256359001149)*x[0]**o+(-0.868975044191)*x[0]+(0.494314269744)*x[1]**o+(0.949957481172)*x[1]
        arg[0,1,1,0]=(-0.0256658750601)*x[0]**o+(0.145499104203)*x[0]+(-0.68669791629)*x[1]**o+(0.447326207295)*x[1]
        arg[0,1,1,1]=(0.150617284467)*x[0]**o+(-0.350046088196)*x[0]+(-0.367671302053)*x[1]**o+(0.633220122068)*x[1]
        arg[0,1,2,0]=(-0.714145362271)*x[0]**o+(-0.957661490758)*x[0]+(0.386253770747)*x[1]**o+(0.982634262634)*x[1]
        arg[0,1,2,1]=(0.707565096824)*x[0]**o+(0.281758058573)*x[0]+(-0.407791315587)*x[1]**o+(0.86069036983)*x[1]
        arg[0,2,0,0]=(-0.453860220598)*x[0]**o+(0.944457388061)*x[0]+(-0.903818960162)*x[1]**o+(-0.799423205297)*x[1]
        arg[0,2,0,1]=(-0.621257500941)*x[0]**o+(0.748963737937)*x[0]+(-0.12660934458)*x[1]**o+(-0.362830639592)*x[1]
        arg[0,2,1,0]=(-0.702485820037)*x[0]**o+(0.355584898593)*x[0]+(0.657127357751)*x[1]**o+(0.535533390762)*x[1]
        arg[0,2,1,1]=(0.795733125008)*x[0]**o+(0.983607327249)*x[0]+(-0.463818780644)*x[1]**o+(-0.681831588244)*x[1]
        arg[0,2,2,0]=(0.335783204138)*x[0]**o+(-0.712225782679)*x[0]+(0.853426644096)*x[1]**o+(-0.143845263686)*x[1]
        arg[0,2,2,1]=(0.389724375144)*x[0]**o+(0.715176227705)*x[0]+(-0.660690893888)*x[1]**o+(0.428289724709)*x[1]
        arg[0,3,0,0]=(-0.363836203724)*x[0]**o+(-0.257775427457)*x[0]+(0.407904889098)*x[1]**o+(-0.731273814284)*x[1]
        arg[0,3,0,1]=(0.402038262494)*x[0]**o+(0.33289103638)*x[0]+(0.444957562329)*x[1]**o+(0.0657971935664)*x[1]
        arg[0,3,1,0]=(0.979673814602)*x[0]**o+(-0.133405332848)*x[0]+(0.23884689304)*x[1]**o+(0.728464997681)*x[1]
        arg[0,3,1,1]=(0.724603429246)*x[0]**o+(-0.391040898144)*x[0]+(-0.966816499003)*x[1]**o+(0.0984302578905)*x[1]
        arg[0,3,2,0]=(-0.986139645754)*x[0]**o+(0.60807351165)*x[0]+(0.799807413231)*x[1]**o+(0.12091479711)*x[1]
        arg[0,3,2,1]=(0.387453999377)*x[0]**o+(0.212725006628)*x[0]+(-0.120374039139)*x[1]**o+(-0.262304757951)*x[1]
        arg[0,4,0,0]=(-0.260545986276)*x[0]**o+(0.251085312932)*x[0]+(0.738082708518)*x[1]**o+(-0.0804883797938)*x[1]
        arg[0,4,0,1]=(-0.190701356138)*x[0]**o+(-0.0205295935522)*x[0]+(0.904261681819)*x[1]**o+(-0.859647709957)*x[1]
        arg[0,4,1,0]=(-0.945357492849)*x[0]**o+(-0.85298463389)*x[0]+(0.688964336673)*x[1]**o+(-0.672881696932)*x[1]
        arg[0,4,1,1]=(0.877720258045)*x[0]**o+(-0.361228026805)*x[0]+(-0.326626527006)*x[1]**o+(-0.274100421232)*x[1]
        arg[0,4,2,0]=(0.202358729084)*x[0]**o+(-0.997571481837)*x[0]+(0.431193711862)*x[1]**o+(0.513561123915)*x[1]
        arg[0,4,2,1]=(-0.136602954592)*x[0]**o+(-0.40864934437)*x[0]+(0.675945574351)*x[1]**o+(0.728757793536)*x[1]
        arg[1,0,0,0]=(0.835464261851)*x[0]**o+(-0.131447020601)*x[0]+(0.0810352521446)*x[1]**o+(0.664744750899)*x[1]
        arg[1,0,0,1]=(0.239716928689)*x[0]**o+(-0.192258999531)*x[0]+(0.615877054036)*x[1]**o+(0.835274774437)*x[1]
        arg[1,0,1,0]=(0.509342803767)*x[0]**o+(-0.633944481882)*x[0]+(0.262098006086)*x[1]**o+(0.159902070799)*x[1]
        arg[1,0,1,1]=(0.894432440588)*x[0]**o+(-0.167778431319)*x[0]+(-0.984183148477)*x[1]**o+(-0.554981658632)*x[1]
        arg[1,0,2,0]=(-0.762472607231)*x[0]**o+(0.405550139256)*x[0]+(0.4051290641)*x[1]**o+(-0.27180186672)*x[1]
        arg[1,0,2,1]=(0.663869273816)*x[0]**o+(0.867453974628)*x[0]+(-0.279659969132)*x[1]**o+(-0.370539101204)*x[1]
        arg[1,1,0,0]=(-0.969326778164)*x[0]**o+(0.965471349497)*x[0]+(-0.529983232946)*x[1]**o+(0.714091610977)*x[1]
        arg[1,1,0,1]=(-0.840360667093)*x[0]**o+(-0.399319808256)*x[0]+(0.568769915582)*x[1]**o+(0.872427494442)*x[1]
        arg[1,1,1,0]=(0.46099116941)*x[0]**o+(-0.986008874142)*x[0]+(-0.53999023556)*x[1]**o+(-0.549695567855)*x[1]
        arg[1,1,1,1]=(0.485944918559)*x[0]**o+(0.189773828651)*x[0]+(-0.611654980493)*x[1]**o+(-0.337071911657)*x[1]
        arg[1,1,2,0]=(0.191663603248)*x[0]**o+(0.0906374965004)*x[0]+(-0.271051598251)*x[1]**o+(0.880507788825)*x[1]
        arg[1,1,2,1]=(-0.439514997001)*x[0]**o+(0.701230426224)*x[0]+(-0.193231641029)*x[1]**o+(0.950584054332)*x[1]
        arg[1,2,0,0]=(0.439456565259)*x[0]**o+(-0.25664171484)*x[0]+(-0.653059819329)*x[1]**o+(-0.680544747777)*x[1]
        arg[1,2,0,1]=(0.34890454443)*x[0]**o+(-0.69851200963)*x[0]+(-0.00159381688549)*x[1]**o+(-0.612359675499)*x[1]
        arg[1,2,1,0]=(0.132759671375)*x[0]**o+(-0.235126716401)*x[0]+(0.652068885826)*x[1]**o+(0.749698645371)*x[1]
        arg[1,2,1,1]=(0.934980403627)*x[0]**o+(0.321409273437)*x[0]+(-0.605909929553)*x[1]**o+(0.135554927418)*x[1]
        arg[1,2,2,0]=(-0.68327709329)*x[0]**o+(-0.0381292756005)*x[0]+(0.710953191114)*x[1]**o+(-0.344522256689)*x[1]
        arg[1,2,2,1]=(-0.813976235652)*x[0]**o+(0.613879708123)*x[0]+(0.451599277293)*x[1]**o+(-0.6134026375)*x[1]
        arg[1,3,0,0]=(0.595310128627)*x[0]**o+(-0.247621573131)*x[0]+(0.165607783104)*x[1]**o+(0.578132180805)*x[1]
        arg[1,3,0,1]=(0.848827238882)*x[0]**o+(0.173154067999)*x[0]+(0.340662419437)*x[1]**o+(-0.389656447251)*x[1]
        arg[1,3,1,0]=(0.250888415228)*x[0]**o+(-0.278203136198)*x[0]+(-0.693211186734)*x[1]**o+(0.613812194735)*x[1]
        arg[1,3,1,1]=(0.634533003162)*x[0]**o+(-0.0924665762471)*x[0]+(0.749575894384)*x[1]**o+(-0.737205017415)*x[1]
        arg[1,3,2,0]=(0.607750966973)*x[0]**o+(0.905016928404)*x[0]+(-0.171049542163)*x[1]**o+(-0.20526244601)*x[1]
        arg[1,3,2,1]=(0.723946025405)*x[0]**o+(0.0157576743987)*x[0]+(0.190875428312)*x[1]**o+(0.407925899509)*x[1]
        arg[1,4,0,0]=(-0.561582695803)*x[0]**o+(-0.350262461371)*x[0]+(0.57610409872)*x[1]**o+(0.881780963877)*x[1]
        arg[1,4,0,1]=(0.077066498695)*x[0]**o+(0.515597802516)*x[0]+(0.332011413016)*x[1]**o+(0.215491334787)*x[1]
        arg[1,4,1,0]=(0.33858727737)*x[0]**o+(-0.573231800768)*x[0]+(-0.559915809132)*x[1]**o+(0.0240869142176)*x[1]
        arg[1,4,1,1]=(-0.684581727952)*x[0]**o+(0.365541904201)*x[0]+(0.204252397376)*x[1]**o+(0.149300766054)*x[1]
        arg[1,4,2,0]=(-0.574225088383)*x[0]**o+(0.0286339776067)*x[0]+(-0.358732122607)*x[1]**o+(-0.416932738957)*x[1]
        arg[1,4,2,1]=(-0.308460184932)*x[0]**o+(0.389214987836)*x[0]+(-0.358303257031)*x[1]**o+(0.348109483688)*x[1]
        arg[2,0,0,0]=(-0.811318668804)*x[0]**o+(-0.298028116982)*x[0]+(0.679871523483)*x[1]**o+(0.958042925287)*x[1]
        arg[2,0,0,1]=(-0.616521818338)*x[0]**o+(-0.0817124854496)*x[0]+(0.403257366893)*x[1]**o+(0.740689710435)*x[1]
        arg[2,0,1,0]=(0.271046179127)*x[0]**o+(0.814928093947)*x[0]+(0.423722781579)*x[1]**o+(0.925333209811)*x[1]
        arg[2,0,1,1]=(0.0596651806816)*x[0]**o+(0.493854605359)*x[0]+(0.539998537439)*x[1]**o+(-0.910865639375)*x[1]
        arg[2,0,2,0]=(0.503267602327)*x[0]**o+(-0.270966471743)*x[0]+(-0.342063366676)*x[1]**o+(-0.975700224705)*x[1]
        arg[2,0,2,1]=(-0.0362358939195)*x[0]**o+(-0.766033869339)*x[0]+(0.891365100088)*x[1]**o+(0.815025898929)*x[1]
        arg[2,1,0,0]=(-0.667240459984)*x[0]**o+(0.8659076701)*x[0]+(0.215089397562)*x[1]**o+(0.464769182307)*x[1]
        arg[2,1,0,1]=(-0.298625221931)*x[0]**o+(-0.504329743472)*x[0]+(0.412472510208)*x[1]**o+(0.241693152889)*x[1]
        arg[2,1,1,0]=(-0.670730179133)*x[0]**o+(0.0201070113832)*x[0]+(-0.572725898919)*x[1]**o+(0.665609259014)*x[1]
        arg[2,1,1,1]=(-0.319080984089)*x[0]**o+(0.0299423045224)*x[0]+(0.0152311810613)*x[1]**o+(-0.784646189026)*x[1]
        arg[2,1,2,0]=(0.153308789716)*x[0]**o+(0.725808150613)*x[0]+(0.62876493338)*x[1]**o+(0.80275284252)*x[1]
        arg[2,1,2,1]=(0.804698870013)*x[0]**o+(-0.631078773061)*x[0]+(-0.015221975698)*x[1]**o+(-0.954169686746)*x[1]
        arg[2,2,0,0]=(0.169022797498)*x[0]**o+(0.733025791099)*x[0]+(0.321554996633)*x[1]**o+(0.292173365364)*x[1]
        arg[2,2,0,1]=(0.327345955109)*x[0]**o+(0.21757823896)*x[0]+(0.873027160446)*x[1]**o+(-0.250192861579)*x[1]
        arg[2,2,1,0]=(0.394387190832)*x[0]**o+(-0.613661883482)*x[0]+(0.806965291386)*x[1]**o+(0.244513244016)*x[1]
        arg[2,2,1,1]=(0.043913907536)*x[0]**o+(-0.303567806424)*x[0]+(0.547037093323)*x[1]**o+(-0.598687385328)*x[1]
        arg[2,2,2,0]=(0.875663154787)*x[0]**o+(-0.34984832772)*x[0]+(0.879143174396)*x[1]**o+(-0.193154051109)*x[1]
        arg[2,2,2,1]=(0.940311132918)*x[0]**o+(-0.358708743382)*x[0]+(0.117370945146)*x[1]**o+(-0.664208034757)*x[1]
        arg[2,3,0,0]=(0.0334525660588)*x[0]**o+(0.592805915255)*x[0]+(-0.278216157971)*x[1]**o+(0.894369967671)*x[1]
        arg[2,3,0,1]=(-0.966591106326)*x[0]**o+(-0.882160390768)*x[0]+(0.843982702732)*x[1]**o+(0.363868921656)*x[1]
        arg[2,3,1,0]=(-0.246392342771)*x[0]**o+(-0.934459096242)*x[0]+(-0.472771368537)*x[1]**o+(0.148854304706)*x[1]
        arg[2,3,1,1]=(0.248218424051)*x[0]**o+(0.0980396576111)*x[0]+(0.413322166489)*x[1]**o+(0.429747261346)*x[1]
        arg[2,3,2,0]=(0.742987565733)*x[0]**o+(-0.213736450136)*x[0]+(0.996975741314)*x[1]**o+(0.960709958665)*x[1]
        arg[2,3,2,1]=(0.436149490977)*x[0]**o+(0.87632181197)*x[0]+(-0.881934669912)*x[1]**o+(-0.457520504175)*x[1]
        arg[2,4,0,0]=(-0.519014330693)*x[0]**o+(0.938619608407)*x[0]+(-0.847954563609)*x[1]**o+(0.791774379621)*x[1]
        arg[2,4,0,1]=(-0.599487299453)*x[0]**o+(0.977867555132)*x[0]+(0.693105630579)*x[1]**o+(-0.688827812942)*x[1]
        arg[2,4,1,0]=(0.955907964163)*x[0]**o+(0.0724056986172)*x[0]+(0.254773619772)*x[1]**o+(0.0428861146964)*x[1]
        arg[2,4,1,1]=(-0.422202101998)*x[0]**o+(0.803263223544)*x[0]+(0.456048140785)*x[1]**o+(0.0419133126183)*x[1]
        arg[2,4,2,0]=(-0.182308704781)*x[0]**o+(0.862282443488)*x[0]+(-0.844027007955)*x[1]**o+(0.586220526376)*x[1]
        arg[2,4,2,1]=(-0.801542666229)*x[0]**o+(0.903685299455)*x[0]+(0.428926544377)*x[1]**o+(-0.891220471549)*x[1]
        arg[3,0,0,0]=(0.602917787713)*x[0]**o+(-0.592239549743)*x[0]+(0.761830006718)*x[1]**o+(-0.784264167777)*x[1]
        arg[3,0,0,1]=(0.819962957732)*x[0]**o+(-0.455126790372)*x[0]+(-0.144898728525)*x[1]**o+(0.121566345339)*x[1]
        arg[3,0,1,0]=(0.570694941559)*x[0]**o+(-0.505678656386)*x[0]+(0.295536587355)*x[1]**o+(-0.852779737814)*x[1]
        arg[3,0,1,1]=(-0.226246858753)*x[0]**o+(-0.513900306228)*x[0]+(-0.389483591407)*x[1]**o+(-0.735105335539)*x[1]
        arg[3,0,2,0]=(0.941877610938)*x[0]**o+(0.93808330298)*x[0]+(0.189785162182)*x[1]**o+(0.546525366304)*x[1]
        arg[3,0,2,1]=(0.10639100485)*x[0]**o+(-0.773361007972)*x[0]+(-0.470531427579)*x[1]**o+(0.467359342808)*x[1]
        arg[3,1,0,0]=(0.857725599589)*x[0]**o+(0.0881746413132)*x[0]+(0.040713151152)*x[1]**o+(-0.52070729883)*x[1]
        arg[3,1,0,1]=(0.266807652146)*x[0]**o+(0.337496673892)*x[0]+(-0.562266297445)*x[1]**o+(0.326015195113)*x[1]
        arg[3,1,1,0]=(0.177275509465)*x[0]**o+(-0.647855761301)*x[0]+(-0.0963373631995)*x[1]**o+(-0.535256252994)*x[1]
        arg[3,1,1,1]=(0.430309295606)*x[0]**o+(-0.0894749557428)*x[0]+(0.503470724483)*x[1]**o+(-0.202361122248)*x[1]
        arg[3,1,2,0]=(0.283995014194)*x[0]**o+(-0.544419907194)*x[0]+(0.78478571712)*x[1]**o+(0.341540664147)*x[1]
        arg[3,1,2,1]=(-0.281042869936)*x[0]**o+(0.728771766492)*x[0]+(-0.298974388557)*x[1]**o+(-0.836789906985)*x[1]
        arg[3,2,0,0]=(-0.415323371561)*x[0]**o+(0.583971137371)*x[0]+(-0.265658679489)*x[1]**o+(-0.432447373809)*x[1]
        arg[3,2,0,1]=(0.690486298855)*x[0]**o+(0.552521632226)*x[0]+(0.237622739042)*x[1]**o+(-0.254513693924)*x[1]
        arg[3,2,1,0]=(0.453985687207)*x[0]**o+(-0.785385580962)*x[0]+(0.0190987841634)*x[1]**o+(-0.7923288861)*x[1]
        arg[3,2,1,1]=(-0.964082656379)*x[0]**o+(-0.806315937371)*x[0]+(0.271145420366)*x[1]**o+(-0.839014535945)*x[1]
        arg[3,2,2,0]=(-0.372062773997)*x[0]**o+(-0.069749560567)*x[0]+(-0.126310395383)*x[1]**o+(0.972025865285)*x[1]
        arg[3,2,2,1]=(-0.580353020086)*x[0]**o+(-0.152746821477)*x[0]+(0.297553660535)*x[1]**o+(0.444791568145)*x[1]
        arg[3,3,0,0]=(-0.361520029439)*x[0]**o+(-0.810391070499)*x[0]+(0.462123882642)*x[1]**o+(-0.974534227742)*x[1]
        arg[3,3,0,1]=(0.113370617837)*x[0]**o+(0.428257355435)*x[0]+(-0.18656623758)*x[1]**o+(0.206503327394)*x[1]
        arg[3,3,1,0]=(0.341328445689)*x[0]**o+(0.34002947054)*x[0]+(0.997677286362)*x[1]**o+(0.43300073225)*x[1]
        arg[3,3,1,1]=(-0.513912982823)*x[0]**o+(0.826573827064)*x[0]+(-0.701417864085)*x[1]**o+(0.567078883219)*x[1]
        arg[3,3,2,0]=(-0.325536866965)*x[0]**o+(0.924405767451)*x[0]+(-0.0481381954346)*x[1]**o+(0.912210617073)*x[1]
        arg[3,3,2,1]=(0.491532434302)*x[0]**o+(-0.584980848036)*x[0]+(-0.788317530858)*x[1]**o+(-0.900681893172)*x[1]
        arg[3,4,0,0]=(0.432249855148)*x[0]**o+(-0.673369312552)*x[0]+(0.244704845849)*x[1]**o+(-0.300995595367)*x[1]
        arg[3,4,0,1]=(0.383037211984)*x[0]**o+(0.299711400488)*x[0]+(0.870584255454)*x[1]**o+(-0.33571823329)*x[1]
        arg[3,4,1,0]=(0.737134180177)*x[0]**o+(-0.101307090628)*x[0]+(-0.970632061143)*x[1]**o+(-0.449042684137)*x[1]
        arg[3,4,1,1]=(0.00684250643632)*x[0]**o+(0.676964654022)*x[0]+(0.789069834858)*x[1]**o+(-0.829739508197)*x[1]
        arg[3,4,2,0]=(0.7279044944)*x[0]**o+(0.215711122038)*x[0]+(0.385631252632)*x[1]**o+(0.629390433529)*x[1]
        arg[3,4,2,1]=(-0.380750716629)*x[0]**o+(-0.413040692437)*x[0]+(0.26795244639)*x[1]**o+(0.492475504659)*x[1]
        ref[0,0,0,0]=(-0.274472293956)*x_ref[0]**o+(-0.293347130929)*x_ref[0]+(0.392586448254)*x_ref[1]**o+(-0.403066854054)*x_ref[1]
        ref[0,0,0,1]=(0.385933192907)*x_ref[0]**o+(0.484240065867)*x_ref[0]+(0.144565631551)*x_ref[1]**o+(0.882508492739)*x_ref[1]
        ref[0,0,1,0]=(0.993368905137)*x_ref[0]**o+(0.340349919343)*x_ref[0]+(-0.956806746524)*x_ref[1]**o+(0.990607958001)*x_ref[1]
        ref[0,0,1,1]=(-0.88610250102)*x_ref[0]**o+(-0.543705587259)*x_ref[0]+(0.170798164924)*x_ref[1]**o+(0.685331195754)*x_ref[1]
        ref[0,0,2,0]=(-0.922211611143)*x_ref[0]**o+(-0.103640037893)*x_ref[0]+(0.401354703434)*x_ref[1]**o+(-0.117826261635)*x_ref[1]
        ref[0,0,2,1]=(-0.391601387132)*x_ref[0]**o+(-0.472753344804)*x_ref[0]+(-0.473433838993)*x_ref[1]**o+(-0.173501635989)*x_ref[1]
        ref[0,1,0,0]=(0.747353274916)*x_ref[0]**o+(-0.960856612086)*x_ref[0]+(-0.54809012978)*x_ref[1]**o+(0.862023893332)*x_ref[1]
        ref[0,1,0,1]=(-0.256359001149)*x_ref[0]**o+(-0.868975044191)*x_ref[0]+(0.494314269744)*x_ref[1]**o+(0.949957481172)*x_ref[1]
        ref[0,1,1,0]=(-0.0256658750601)*x_ref[0]**o+(0.145499104203)*x_ref[0]+(-0.68669791629)*x_ref[1]**o+(0.447326207295)*x_ref[1]
        ref[0,1,1,1]=(0.150617284467)*x_ref[0]**o+(-0.350046088196)*x_ref[0]+(-0.367671302053)*x_ref[1]**o+(0.633220122068)*x_ref[1]
        ref[0,1,2,0]=(-0.714145362271)*x_ref[0]**o+(-0.957661490758)*x_ref[0]+(0.386253770747)*x_ref[1]**o+(0.982634262634)*x_ref[1]
        ref[0,1,2,1]=(0.707565096824)*x_ref[0]**o+(0.281758058573)*x_ref[0]+(-0.407791315587)*x_ref[1]**o+(0.86069036983)*x_ref[1]
        ref[0,2,0,0]=(-0.453860220598)*x_ref[0]**o+(0.944457388061)*x_ref[0]+(-0.903818960162)*x_ref[1]**o+(-0.799423205297)*x_ref[1]
        ref[0,2,0,1]=(-0.621257500941)*x_ref[0]**o+(0.748963737937)*x_ref[0]+(-0.12660934458)*x_ref[1]**o+(-0.362830639592)*x_ref[1]
        ref[0,2,1,0]=(-0.702485820037)*x_ref[0]**o+(0.355584898593)*x_ref[0]+(0.657127357751)*x_ref[1]**o+(0.535533390762)*x_ref[1]
        ref[0,2,1,1]=(0.795733125008)*x_ref[0]**o+(0.983607327249)*x_ref[0]+(-0.463818780644)*x_ref[1]**o+(-0.681831588244)*x_ref[1]
        ref[0,2,2,0]=(0.335783204138)*x_ref[0]**o+(-0.712225782679)*x_ref[0]+(0.853426644096)*x_ref[1]**o+(-0.143845263686)*x_ref[1]
        ref[0,2,2,1]=(0.389724375144)*x_ref[0]**o+(0.715176227705)*x_ref[0]+(-0.660690893888)*x_ref[1]**o+(0.428289724709)*x_ref[1]
        ref[0,3,0,0]=(-0.363836203724)*x_ref[0]**o+(-0.257775427457)*x_ref[0]+(0.407904889098)*x_ref[1]**o+(-0.731273814284)*x_ref[1]
        ref[0,3,0,1]=(0.402038262494)*x_ref[0]**o+(0.33289103638)*x_ref[0]+(0.444957562329)*x_ref[1]**o+(0.0657971935664)*x_ref[1]
        ref[0,3,1,0]=(0.979673814602)*x_ref[0]**o+(-0.133405332848)*x_ref[0]+(0.23884689304)*x_ref[1]**o+(0.728464997681)*x_ref[1]
        ref[0,3,1,1]=(0.724603429246)*x_ref[0]**o+(-0.391040898144)*x_ref[0]+(-0.966816499003)*x_ref[1]**o+(0.0984302578905)*x_ref[1]
        ref[0,3,2,0]=(-0.986139645754)*x_ref[0]**o+(0.60807351165)*x_ref[0]+(0.799807413231)*x_ref[1]**o+(0.12091479711)*x_ref[1]
        ref[0,3,2,1]=(0.387453999377)*x_ref[0]**o+(0.212725006628)*x_ref[0]+(-0.120374039139)*x_ref[1]**o+(-0.262304757951)*x_ref[1]
        ref[0,4,0,0]=(-0.260545986276)*x_ref[0]**o+(0.251085312932)*x_ref[0]+(0.738082708518)*x_ref[1]**o+(-0.0804883797938)*x_ref[1]
        ref[0,4,0,1]=(-0.190701356138)*x_ref[0]**o+(-0.0205295935522)*x_ref[0]+(0.904261681819)*x_ref[1]**o+(-0.859647709957)*x_ref[1]
        ref[0,4,1,0]=(-0.945357492849)*x_ref[0]**o+(-0.85298463389)*x_ref[0]+(0.688964336673)*x_ref[1]**o+(-0.672881696932)*x_ref[1]
        ref[0,4,1,1]=(0.877720258045)*x_ref[0]**o+(-0.361228026805)*x_ref[0]+(-0.326626527006)*x_ref[1]**o+(-0.274100421232)*x_ref[1]
        ref[0,4,2,0]=(0.202358729084)*x_ref[0]**o+(-0.997571481837)*x_ref[0]+(0.431193711862)*x_ref[1]**o+(0.513561123915)*x_ref[1]
        ref[0,4,2,1]=(-0.136602954592)*x_ref[0]**o+(-0.40864934437)*x_ref[0]+(0.675945574351)*x_ref[1]**o+(0.728757793536)*x_ref[1]
        ref[1,0,0,0]=(0.835464261851)*x_ref[0]**o+(-0.131447020601)*x_ref[0]+(0.0810352521446)*x_ref[1]**o+(0.664744750899)*x_ref[1]
        ref[1,0,0,1]=(0.239716928689)*x_ref[0]**o+(-0.192258999531)*x_ref[0]+(0.615877054036)*x_ref[1]**o+(0.835274774437)*x_ref[1]
        ref[1,0,1,0]=(0.509342803767)*x_ref[0]**o+(-0.633944481882)*x_ref[0]+(0.262098006086)*x_ref[1]**o+(0.159902070799)*x_ref[1]
        ref[1,0,1,1]=(0.894432440588)*x_ref[0]**o+(-0.167778431319)*x_ref[0]+(-0.984183148477)*x_ref[1]**o+(-0.554981658632)*x_ref[1]
        ref[1,0,2,0]=(-0.762472607231)*x_ref[0]**o+(0.405550139256)*x_ref[0]+(0.4051290641)*x_ref[1]**o+(-0.27180186672)*x_ref[1]
        ref[1,0,2,1]=(0.663869273816)*x_ref[0]**o+(0.867453974628)*x_ref[0]+(-0.279659969132)*x_ref[1]**o+(-0.370539101204)*x_ref[1]
        ref[1,1,0,0]=(-0.969326778164)*x_ref[0]**o+(0.965471349497)*x_ref[0]+(-0.529983232946)*x_ref[1]**o+(0.714091610977)*x_ref[1]
        ref[1,1,0,1]=(-0.840360667093)*x_ref[0]**o+(-0.399319808256)*x_ref[0]+(0.568769915582)*x_ref[1]**o+(0.872427494442)*x_ref[1]
        ref[1,1,1,0]=(0.46099116941)*x_ref[0]**o+(-0.986008874142)*x_ref[0]+(-0.53999023556)*x_ref[1]**o+(-0.549695567855)*x_ref[1]
        ref[1,1,1,1]=(0.485944918559)*x_ref[0]**o+(0.189773828651)*x_ref[0]+(-0.611654980493)*x_ref[1]**o+(-0.337071911657)*x_ref[1]
        ref[1,1,2,0]=(0.191663603248)*x_ref[0]**o+(0.0906374965004)*x_ref[0]+(-0.271051598251)*x_ref[1]**o+(0.880507788825)*x_ref[1]
        ref[1,1,2,1]=(-0.439514997001)*x_ref[0]**o+(0.701230426224)*x_ref[0]+(-0.193231641029)*x_ref[1]**o+(0.950584054332)*x_ref[1]
        ref[1,2,0,0]=(0.439456565259)*x_ref[0]**o+(-0.25664171484)*x_ref[0]+(-0.653059819329)*x_ref[1]**o+(-0.680544747777)*x_ref[1]
        ref[1,2,0,1]=(0.34890454443)*x_ref[0]**o+(-0.69851200963)*x_ref[0]+(-0.00159381688549)*x_ref[1]**o+(-0.612359675499)*x_ref[1]
        ref[1,2,1,0]=(0.132759671375)*x_ref[0]**o+(-0.235126716401)*x_ref[0]+(0.652068885826)*x_ref[1]**o+(0.749698645371)*x_ref[1]
        ref[1,2,1,1]=(0.934980403627)*x_ref[0]**o+(0.321409273437)*x_ref[0]+(-0.605909929553)*x_ref[1]**o+(0.135554927418)*x_ref[1]
        ref[1,2,2,0]=(-0.68327709329)*x_ref[0]**o+(-0.0381292756005)*x_ref[0]+(0.710953191114)*x_ref[1]**o+(-0.344522256689)*x_ref[1]
        ref[1,2,2,1]=(-0.813976235652)*x_ref[0]**o+(0.613879708123)*x_ref[0]+(0.451599277293)*x_ref[1]**o+(-0.6134026375)*x_ref[1]
        ref[1,3,0,0]=(0.595310128627)*x_ref[0]**o+(-0.247621573131)*x_ref[0]+(0.165607783104)*x_ref[1]**o+(0.578132180805)*x_ref[1]
        ref[1,3,0,1]=(0.848827238882)*x_ref[0]**o+(0.173154067999)*x_ref[0]+(0.340662419437)*x_ref[1]**o+(-0.389656447251)*x_ref[1]
        ref[1,3,1,0]=(0.250888415228)*x_ref[0]**o+(-0.278203136198)*x_ref[0]+(-0.693211186734)*x_ref[1]**o+(0.613812194735)*x_ref[1]
        ref[1,3,1,1]=(0.634533003162)*x_ref[0]**o+(-0.0924665762471)*x_ref[0]+(0.749575894384)*x_ref[1]**o+(-0.737205017415)*x_ref[1]
        ref[1,3,2,0]=(0.607750966973)*x_ref[0]**o+(0.905016928404)*x_ref[0]+(-0.171049542163)*x_ref[1]**o+(-0.20526244601)*x_ref[1]
        ref[1,3,2,1]=(0.723946025405)*x_ref[0]**o+(0.0157576743987)*x_ref[0]+(0.190875428312)*x_ref[1]**o+(0.407925899509)*x_ref[1]
        ref[1,4,0,0]=(-0.561582695803)*x_ref[0]**o+(-0.350262461371)*x_ref[0]+(0.57610409872)*x_ref[1]**o+(0.881780963877)*x_ref[1]
        ref[1,4,0,1]=(0.077066498695)*x_ref[0]**o+(0.515597802516)*x_ref[0]+(0.332011413016)*x_ref[1]**o+(0.215491334787)*x_ref[1]
        ref[1,4,1,0]=(0.33858727737)*x_ref[0]**o+(-0.573231800768)*x_ref[0]+(-0.559915809132)*x_ref[1]**o+(0.0240869142176)*x_ref[1]
        ref[1,4,1,1]=(-0.684581727952)*x_ref[0]**o+(0.365541904201)*x_ref[0]+(0.204252397376)*x_ref[1]**o+(0.149300766054)*x_ref[1]
        ref[1,4,2,0]=(-0.574225088383)*x_ref[0]**o+(0.0286339776067)*x_ref[0]+(-0.358732122607)*x_ref[1]**o+(-0.416932738957)*x_ref[1]
        ref[1,4,2,1]=(-0.308460184932)*x_ref[0]**o+(0.389214987836)*x_ref[0]+(-0.358303257031)*x_ref[1]**o+(0.348109483688)*x_ref[1]
        ref[2,0,0,0]=(-0.811318668804)*x_ref[0]**o+(-0.298028116982)*x_ref[0]+(0.679871523483)*x_ref[1]**o+(0.958042925287)*x_ref[1]
        ref[2,0,0,1]=(-0.616521818338)*x_ref[0]**o+(-0.0817124854496)*x_ref[0]+(0.403257366893)*x_ref[1]**o+(0.740689710435)*x_ref[1]
        ref[2,0,1,0]=(0.271046179127)*x_ref[0]**o+(0.814928093947)*x_ref[0]+(0.423722781579)*x_ref[1]**o+(0.925333209811)*x_ref[1]
        ref[2,0,1,1]=(0.0596651806816)*x_ref[0]**o+(0.493854605359)*x_ref[0]+(0.539998537439)*x_ref[1]**o+(-0.910865639375)*x_ref[1]
        ref[2,0,2,0]=(0.503267602327)*x_ref[0]**o+(-0.270966471743)*x_ref[0]+(-0.342063366676)*x_ref[1]**o+(-0.975700224705)*x_ref[1]
        ref[2,0,2,1]=(-0.0362358939195)*x_ref[0]**o+(-0.766033869339)*x_ref[0]+(0.891365100088)*x_ref[1]**o+(0.815025898929)*x_ref[1]
        ref[2,1,0,0]=(-0.667240459984)*x_ref[0]**o+(0.8659076701)*x_ref[0]+(0.215089397562)*x_ref[1]**o+(0.464769182307)*x_ref[1]
        ref[2,1,0,1]=(-0.298625221931)*x_ref[0]**o+(-0.504329743472)*x_ref[0]+(0.412472510208)*x_ref[1]**o+(0.241693152889)*x_ref[1]
        ref[2,1,1,0]=(-0.670730179133)*x_ref[0]**o+(0.0201070113832)*x_ref[0]+(-0.572725898919)*x_ref[1]**o+(0.665609259014)*x_ref[1]
        ref[2,1,1,1]=(-0.319080984089)*x_ref[0]**o+(0.0299423045224)*x_ref[0]+(0.0152311810613)*x_ref[1]**o+(-0.784646189026)*x_ref[1]
        ref[2,1,2,0]=(0.153308789716)*x_ref[0]**o+(0.725808150613)*x_ref[0]+(0.62876493338)*x_ref[1]**o+(0.80275284252)*x_ref[1]
        ref[2,1,2,1]=(0.804698870013)*x_ref[0]**o+(-0.631078773061)*x_ref[0]+(-0.015221975698)*x_ref[1]**o+(-0.954169686746)*x_ref[1]
        ref[2,2,0,0]=(0.169022797498)*x_ref[0]**o+(0.733025791099)*x_ref[0]+(0.321554996633)*x_ref[1]**o+(0.292173365364)*x_ref[1]
        ref[2,2,0,1]=(0.327345955109)*x_ref[0]**o+(0.21757823896)*x_ref[0]+(0.873027160446)*x_ref[1]**o+(-0.250192861579)*x_ref[1]
        ref[2,2,1,0]=(0.394387190832)*x_ref[0]**o+(-0.613661883482)*x_ref[0]+(0.806965291386)*x_ref[1]**o+(0.244513244016)*x_ref[1]
        ref[2,2,1,1]=(0.043913907536)*x_ref[0]**o+(-0.303567806424)*x_ref[0]+(0.547037093323)*x_ref[1]**o+(-0.598687385328)*x_ref[1]
        ref[2,2,2,0]=(0.875663154787)*x_ref[0]**o+(-0.34984832772)*x_ref[0]+(0.879143174396)*x_ref[1]**o+(-0.193154051109)*x_ref[1]
        ref[2,2,2,1]=(0.940311132918)*x_ref[0]**o+(-0.358708743382)*x_ref[0]+(0.117370945146)*x_ref[1]**o+(-0.664208034757)*x_ref[1]
        ref[2,3,0,0]=(0.0334525660588)*x_ref[0]**o+(0.592805915255)*x_ref[0]+(-0.278216157971)*x_ref[1]**o+(0.894369967671)*x_ref[1]
        ref[2,3,0,1]=(-0.966591106326)*x_ref[0]**o+(-0.882160390768)*x_ref[0]+(0.843982702732)*x_ref[1]**o+(0.363868921656)*x_ref[1]
        ref[2,3,1,0]=(-0.246392342771)*x_ref[0]**o+(-0.934459096242)*x_ref[0]+(-0.472771368537)*x_ref[1]**o+(0.148854304706)*x_ref[1]
        ref[2,3,1,1]=(0.248218424051)*x_ref[0]**o+(0.0980396576111)*x_ref[0]+(0.413322166489)*x_ref[1]**o+(0.429747261346)*x_ref[1]
        ref[2,3,2,0]=(0.742987565733)*x_ref[0]**o+(-0.213736450136)*x_ref[0]+(0.996975741314)*x_ref[1]**o+(0.960709958665)*x_ref[1]
        ref[2,3,2,1]=(0.436149490977)*x_ref[0]**o+(0.87632181197)*x_ref[0]+(-0.881934669912)*x_ref[1]**o+(-0.457520504175)*x_ref[1]
        ref[2,4,0,0]=(-0.519014330693)*x_ref[0]**o+(0.938619608407)*x_ref[0]+(-0.847954563609)*x_ref[1]**o+(0.791774379621)*x_ref[1]
        ref[2,4,0,1]=(-0.599487299453)*x_ref[0]**o+(0.977867555132)*x_ref[0]+(0.693105630579)*x_ref[1]**o+(-0.688827812942)*x_ref[1]
        ref[2,4,1,0]=(0.955907964163)*x_ref[0]**o+(0.0724056986172)*x_ref[0]+(0.254773619772)*x_ref[1]**o+(0.0428861146964)*x_ref[1]
        ref[2,4,1,1]=(-0.422202101998)*x_ref[0]**o+(0.803263223544)*x_ref[0]+(0.456048140785)*x_ref[1]**o+(0.0419133126183)*x_ref[1]
        ref[2,4,2,0]=(-0.182308704781)*x_ref[0]**o+(0.862282443488)*x_ref[0]+(-0.844027007955)*x_ref[1]**o+(0.586220526376)*x_ref[1]
        ref[2,4,2,1]=(-0.801542666229)*x_ref[0]**o+(0.903685299455)*x_ref[0]+(0.428926544377)*x_ref[1]**o+(-0.891220471549)*x_ref[1]
        ref[3,0,0,0]=(0.602917787713)*x_ref[0]**o+(-0.592239549743)*x_ref[0]+(0.761830006718)*x_ref[1]**o+(-0.784264167777)*x_ref[1]
        ref[3,0,0,1]=(0.819962957732)*x_ref[0]**o+(-0.455126790372)*x_ref[0]+(-0.144898728525)*x_ref[1]**o+(0.121566345339)*x_ref[1]
        ref[3,0,1,0]=(0.570694941559)*x_ref[0]**o+(-0.505678656386)*x_ref[0]+(0.295536587355)*x_ref[1]**o+(-0.852779737814)*x_ref[1]
        ref[3,0,1,1]=(-0.226246858753)*x_ref[0]**o+(-0.513900306228)*x_ref[0]+(-0.389483591407)*x_ref[1]**o+(-0.735105335539)*x_ref[1]
        ref[3,0,2,0]=(0.941877610938)*x_ref[0]**o+(0.93808330298)*x_ref[0]+(0.189785162182)*x_ref[1]**o+(0.546525366304)*x_ref[1]
        ref[3,0,2,1]=(0.10639100485)*x_ref[0]**o+(-0.773361007972)*x_ref[0]+(-0.470531427579)*x_ref[1]**o+(0.467359342808)*x_ref[1]
        ref[3,1,0,0]=(0.857725599589)*x_ref[0]**o+(0.0881746413132)*x_ref[0]+(0.040713151152)*x_ref[1]**o+(-0.52070729883)*x_ref[1]
        ref[3,1,0,1]=(0.266807652146)*x_ref[0]**o+(0.337496673892)*x_ref[0]+(-0.562266297445)*x_ref[1]**o+(0.326015195113)*x_ref[1]
        ref[3,1,1,0]=(0.177275509465)*x_ref[0]**o+(-0.647855761301)*x_ref[0]+(-0.0963373631995)*x_ref[1]**o+(-0.535256252994)*x_ref[1]
        ref[3,1,1,1]=(0.430309295606)*x_ref[0]**o+(-0.0894749557428)*x_ref[0]+(0.503470724483)*x_ref[1]**o+(-0.202361122248)*x_ref[1]
        ref[3,1,2,0]=(0.283995014194)*x_ref[0]**o+(-0.544419907194)*x_ref[0]+(0.78478571712)*x_ref[1]**o+(0.341540664147)*x_ref[1]
        ref[3,1,2,1]=(-0.281042869936)*x_ref[0]**o+(0.728771766492)*x_ref[0]+(-0.298974388557)*x_ref[1]**o+(-0.836789906985)*x_ref[1]
        ref[3,2,0,0]=(-0.415323371561)*x_ref[0]**o+(0.583971137371)*x_ref[0]+(-0.265658679489)*x_ref[1]**o+(-0.432447373809)*x_ref[1]
        ref[3,2,0,1]=(0.690486298855)*x_ref[0]**o+(0.552521632226)*x_ref[0]+(0.237622739042)*x_ref[1]**o+(-0.254513693924)*x_ref[1]
        ref[3,2,1,0]=(0.453985687207)*x_ref[0]**o+(-0.785385580962)*x_ref[0]+(0.0190987841634)*x_ref[1]**o+(-0.7923288861)*x_ref[1]
        ref[3,2,1,1]=(-0.964082656379)*x_ref[0]**o+(-0.806315937371)*x_ref[0]+(0.271145420366)*x_ref[1]**o+(-0.839014535945)*x_ref[1]
        ref[3,2,2,0]=(-0.372062773997)*x_ref[0]**o+(-0.069749560567)*x_ref[0]+(-0.126310395383)*x_ref[1]**o+(0.972025865285)*x_ref[1]
        ref[3,2,2,1]=(-0.580353020086)*x_ref[0]**o+(-0.152746821477)*x_ref[0]+(0.297553660535)*x_ref[1]**o+(0.444791568145)*x_ref[1]
        ref[3,3,0,0]=(-0.361520029439)*x_ref[0]**o+(-0.810391070499)*x_ref[0]+(0.462123882642)*x_ref[1]**o+(-0.974534227742)*x_ref[1]
        ref[3,3,0,1]=(0.113370617837)*x_ref[0]**o+(0.428257355435)*x_ref[0]+(-0.18656623758)*x_ref[1]**o+(0.206503327394)*x_ref[1]
        ref[3,3,1,0]=(0.341328445689)*x_ref[0]**o+(0.34002947054)*x_ref[0]+(0.997677286362)*x_ref[1]**o+(0.43300073225)*x_ref[1]
        ref[3,3,1,1]=(-0.513912982823)*x_ref[0]**o+(0.826573827064)*x_ref[0]+(-0.701417864085)*x_ref[1]**o+(0.567078883219)*x_ref[1]
        ref[3,3,2,0]=(-0.325536866965)*x_ref[0]**o+(0.924405767451)*x_ref[0]+(-0.0481381954346)*x_ref[1]**o+(0.912210617073)*x_ref[1]
        ref[3,3,2,1]=(0.491532434302)*x_ref[0]**o+(-0.584980848036)*x_ref[0]+(-0.788317530858)*x_ref[1]**o+(-0.900681893172)*x_ref[1]
        ref[3,4,0,0]=(0.432249855148)*x_ref[0]**o+(-0.673369312552)*x_ref[0]+(0.244704845849)*x_ref[1]**o+(-0.300995595367)*x_ref[1]
        ref[3,4,0,1]=(0.383037211984)*x_ref[0]**o+(0.299711400488)*x_ref[0]+(0.870584255454)*x_ref[1]**o+(-0.33571823329)*x_ref[1]
        ref[3,4,1,0]=(0.737134180177)*x_ref[0]**o+(-0.101307090628)*x_ref[0]+(-0.970632061143)*x_ref[1]**o+(-0.449042684137)*x_ref[1]
        ref[3,4,1,1]=(0.00684250643632)*x_ref[0]**o+(0.676964654022)*x_ref[0]+(0.789069834858)*x_ref[1]**o+(-0.829739508197)*x_ref[1]
        ref[3,4,2,0]=(0.7279044944)*x_ref[0]**o+(0.215711122038)*x_ref[0]+(0.385631252632)*x_ref[1]**o+(0.629390433529)*x_ref[1]
        ref[3,4,2,1]=(-0.380750716629)*x_ref[0]**o+(-0.413040692437)*x_ref[0]+(0.26795244639)*x_ref[1]**o+(0.492475504659)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.287160004495)*x[0]**o+(0.35112250158)*x[0]+(0.061693427026)*x[1]**o+(-0.752505628029)*x[1]+(-0.57137756028)*x[2]**o+(-0.0397598892504)*x[2]
        arg[0,0,0,1]=(0.272799584607)*x[0]**o+(-0.569947956933)*x[0]+(0.322189261657)*x[1]**o+(-0.440552272224)*x[1]+(-0.10460241481)*x[2]**o+(-0.884848984752)*x[2]
        arg[0,0,1,0]=(0.147766301604)*x[0]**o+(0.405651654428)*x[0]+(-0.440827209529)*x[1]**o+(-0.399952864271)*x[1]+(-0.770191009459)*x[2]**o+(-0.789853968049)*x[2]
        arg[0,0,1,1]=(0.688053165375)*x[0]**o+(-0.154890421924)*x[0]+(0.583684164319)*x[1]**o+(-0.398082068217)*x[1]+(-0.573142923099)*x[2]**o+(-0.205271877337)*x[2]
        arg[0,0,2,0]=(0.807043275877)*x[0]**o+(0.843046383231)*x[0]+(-0.588146471402)*x[1]**o+(0.157787304557)*x[1]+(-0.771009748568)*x[2]**o+(0.162090791609)*x[2]
        arg[0,0,2,1]=(-0.0706800489392)*x[0]**o+(0.499961344665)*x[0]+(0.208926178797)*x[1]**o+(-0.7462472769)*x[1]+(-0.105475405648)*x[2]**o+(-0.18670229948)*x[2]
        arg[0,1,0,0]=(-0.696657281943)*x[0]**o+(-0.0251785345759)*x[0]+(-0.579575530451)*x[1]**o+(-0.0628378665725)*x[1]+(0.668395441302)*x[2]**o+(0.813478105092)*x[2]
        arg[0,1,0,1]=(-0.574784314876)*x[0]**o+(-0.321268045204)*x[0]+(0.160704464213)*x[1]**o+(-0.676016748103)*x[1]+(0.918971102865)*x[2]**o+(0.252082410081)*x[2]
        arg[0,1,1,0]=(-0.905062846817)*x[0]**o+(0.788906658477)*x[0]+(-0.408807611936)*x[1]**o+(0.114686186437)*x[1]+(-0.654685568493)*x[2]**o+(0.0879202139254)*x[2]
        arg[0,1,1,1]=(-0.68020231443)*x[0]**o+(-0.82049202902)*x[0]+(-0.958352740187)*x[1]**o+(-0.33201745802)*x[1]+(0.539810363682)*x[2]**o+(0.744988573087)*x[2]
        arg[0,1,2,0]=(-0.960884744997)*x[0]**o+(-0.704157132503)*x[0]+(0.282236905414)*x[1]**o+(0.756363456433)*x[1]+(-0.365400101792)*x[2]**o+(0.975735846554)*x[2]
        arg[0,1,2,1]=(-0.691780076951)*x[0]**o+(-0.952575440059)*x[0]+(-0.921247808098)*x[1]**o+(-0.88872834095)*x[1]+(0.534103013833)*x[2]**o+(-0.602295186104)*x[2]
        arg[0,2,0,0]=(0.460698415704)*x[0]**o+(0.770243814897)*x[0]+(-0.380381727574)*x[1]**o+(0.476360845668)*x[1]+(-0.681040425102)*x[2]**o+(-0.244369435116)*x[2]
        arg[0,2,0,1]=(0.251591652281)*x[0]**o+(0.748234919686)*x[0]+(-0.698189037307)*x[1]**o+(-0.292400412508)*x[1]+(0.613850696577)*x[2]**o+(-0.942243226834)*x[2]
        arg[0,2,1,0]=(-0.718090066328)*x[0]**o+(-0.322499511794)*x[0]+(-0.75105623397)*x[1]**o+(0.526680271832)*x[1]+(-0.186671531793)*x[2]**o+(-0.449273875991)*x[2]
        arg[0,2,1,1]=(-0.52165552452)*x[0]**o+(0.902446006085)*x[0]+(-0.915575172702)*x[1]**o+(-0.150816194662)*x[1]+(0.596782785689)*x[2]**o+(0.546880865642)*x[2]
        arg[0,2,2,0]=(-0.104486704816)*x[0]**o+(0.723863699457)*x[0]+(0.758622117279)*x[1]**o+(0.372271580145)*x[1]+(-0.0820570383792)*x[2]**o+(-0.63539781104)*x[2]
        arg[0,2,2,1]=(0.0355930823603)*x[0]**o+(0.0931942600432)*x[0]+(-0.118096985978)*x[1]**o+(0.296983782212)*x[1]+(-0.864491763712)*x[2]**o+(0.942230663978)*x[2]
        arg[0,3,0,0]=(0.371153709815)*x[0]**o+(0.0427309083893)*x[0]+(-0.886563919084)*x[1]**o+(-0.14946513404)*x[1]+(0.80903804519)*x[2]**o+(0.69728533073)*x[2]
        arg[0,3,0,1]=(0.0593930396142)*x[0]**o+(-0.370327974089)*x[0]+(-0.388354464683)*x[1]**o+(-0.367648771812)*x[1]+(-0.573754821458)*x[2]**o+(-0.947499155479)*x[2]
        arg[0,3,1,0]=(-0.869817317)*x[0]**o+(0.874317772678)*x[0]+(-0.870060178375)*x[1]**o+(0.66571074524)*x[1]+(-0.669052377378)*x[2]**o+(0.940212680575)*x[2]
        arg[0,3,1,1]=(-0.417445606772)*x[0]**o+(0.942606057619)*x[0]+(-0.133936034409)*x[1]**o+(-0.363274280866)*x[1]+(0.83487439935)*x[2]**o+(-0.108398981461)*x[2]
        arg[0,3,2,0]=(0.363623383293)*x[0]**o+(0.624267309217)*x[0]+(0.801719984899)*x[1]**o+(0.189212370122)*x[1]+(0.464633630928)*x[2]**o+(-0.979808318438)*x[2]
        arg[0,3,2,1]=(0.85384408805)*x[0]**o+(-0.889214754116)*x[0]+(0.164915391224)*x[1]**o+(0.0168991405201)*x[1]+(0.097171994236)*x[2]**o+(-0.146094197339)*x[2]
        arg[0,4,0,0]=(0.778356158972)*x[0]**o+(0.910241265149)*x[0]+(0.123690863272)*x[1]**o+(0.87802121779)*x[1]+(-0.714615668595)*x[2]**o+(-0.437472831583)*x[2]
        arg[0,4,0,1]=(-0.157185481783)*x[0]**o+(0.352802163968)*x[0]+(0.390184123612)*x[1]**o+(-0.680415093098)*x[1]+(-0.121195042454)*x[2]**o+(-0.713693712117)*x[2]
        arg[0,4,1,0]=(0.778115668646)*x[0]**o+(-0.111729042892)*x[0]+(0.451518002735)*x[1]**o+(0.232731183374)*x[1]+(-0.764246158086)*x[2]**o+(-0.795525155328)*x[2]
        arg[0,4,1,1]=(-0.880652367027)*x[0]**o+(-0.314966079126)*x[0]+(-0.640720373065)*x[1]**o+(0.336230541901)*x[1]+(0.566956999526)*x[2]**o+(0.74098258811)*x[2]
        arg[0,4,2,0]=(0.908657274301)*x[0]**o+(0.636427823614)*x[0]+(0.650137414386)*x[1]**o+(0.695558379005)*x[1]+(-0.489852401087)*x[2]**o+(-0.479227108044)*x[2]
        arg[0,4,2,1]=(0.015734241525)*x[0]**o+(0.0719884727088)*x[0]+(0.811864895261)*x[1]**o+(0.959127986784)*x[1]+(-0.238146663368)*x[2]**o+(-0.454939322571)*x[2]
        arg[1,0,0,0]=(-0.918218246559)*x[0]**o+(-0.662787169269)*x[0]+(-0.954035144111)*x[1]**o+(-0.647264685228)*x[1]+(-0.834085960213)*x[2]**o+(-0.167541951024)*x[2]
        arg[1,0,0,1]=(-0.670950869382)*x[0]**o+(0.826472159989)*x[0]+(0.180410872786)*x[1]**o+(-0.715162871594)*x[1]+(0.653287853169)*x[2]**o+(0.892060529625)*x[2]
        arg[1,0,1,0]=(0.398789786407)*x[0]**o+(0.689408033554)*x[0]+(0.124338245128)*x[1]**o+(0.748280808587)*x[1]+(0.418357122274)*x[2]**o+(0.230707440075)*x[2]
        arg[1,0,1,1]=(-0.912778964055)*x[0]**o+(0.904288708694)*x[0]+(-0.874744856901)*x[1]**o+(0.550162195719)*x[1]+(-0.7526423373)*x[2]**o+(0.820993231654)*x[2]
        arg[1,0,2,0]=(-0.843459684648)*x[0]**o+(-0.579708764091)*x[0]+(-0.876072131849)*x[1]**o+(0.933694599696)*x[1]+(0.507781209021)*x[2]**o+(-0.719729482133)*x[2]
        arg[1,0,2,1]=(0.594566925796)*x[0]**o+(-0.994301292358)*x[0]+(0.808114389564)*x[1]**o+(-0.970803712864)*x[1]+(0.433544910389)*x[2]**o+(0.260659440299)*x[2]
        arg[1,1,0,0]=(0.997377553172)*x[0]**o+(0.458428616383)*x[0]+(0.357280778493)*x[1]**o+(0.349837110741)*x[1]+(-0.709592214445)*x[2]**o+(-0.630000739033)*x[2]
        arg[1,1,0,1]=(0.121595650501)*x[0]**o+(0.40870120924)*x[0]+(0.576917123807)*x[1]**o+(-0.932268308325)*x[1]+(0.710043472116)*x[2]**o+(-0.716528412077)*x[2]
        arg[1,1,1,0]=(-0.151416035033)*x[0]**o+(0.0255347649787)*x[0]+(-0.752987701889)*x[1]**o+(-0.480969875813)*x[1]+(-0.610514504179)*x[2]**o+(0.959274757402)*x[2]
        arg[1,1,1,1]=(0.76695706106)*x[0]**o+(-0.0750965975512)*x[0]+(0.846833900735)*x[1]**o+(-0.105838465439)*x[1]+(0.541027486457)*x[2]**o+(0.152543017589)*x[2]
        arg[1,1,2,0]=(0.0856977181865)*x[0]**o+(-0.188139369706)*x[0]+(0.936071565289)*x[1]**o+(-0.3958713036)*x[1]+(-0.350945878893)*x[2]**o+(0.382302674883)*x[2]
        arg[1,1,2,1]=(0.873054234789)*x[0]**o+(-0.490869716146)*x[0]+(0.46720526212)*x[1]**o+(-0.334203412334)*x[1]+(-0.81140895636)*x[2]**o+(-0.973858971749)*x[2]
        arg[1,2,0,0]=(-0.716676352443)*x[0]**o+(0.659821396277)*x[0]+(0.335964292954)*x[1]**o+(0.11609664773)*x[1]+(-0.889186749466)*x[2]**o+(0.454231973946)*x[2]
        arg[1,2,0,1]=(0.367048872648)*x[0]**o+(0.276083713126)*x[0]+(0.686639128043)*x[1]**o+(-0.991414415293)*x[1]+(0.25360515964)*x[2]**o+(-0.377623315093)*x[2]
        arg[1,2,1,0]=(-0.723089358777)*x[0]**o+(-0.516690112406)*x[0]+(0.262334273979)*x[1]**o+(0.140773502537)*x[1]+(0.447333972396)*x[2]**o+(-0.551395692966)*x[2]
        arg[1,2,1,1]=(0.738791182491)*x[0]**o+(-0.978120626668)*x[0]+(-0.217970433713)*x[1]**o+(0.0181011067492)*x[1]+(0.756322867523)*x[2]**o+(0.495269195978)*x[2]
        arg[1,2,2,0]=(-0.630506449854)*x[0]**o+(-0.0865482364231)*x[0]+(-0.415896403157)*x[1]**o+(-0.712833149895)*x[1]+(-0.180482789777)*x[2]**o+(0.00788290661235)*x[2]
        arg[1,2,2,1]=(0.0946913865377)*x[0]**o+(0.752854827716)*x[0]+(0.257846220843)*x[1]**o+(-0.953926983578)*x[1]+(-0.388505920682)*x[2]**o+(-0.10751401251)*x[2]
        arg[1,3,0,0]=(0.694014030611)*x[0]**o+(0.0433103547797)*x[0]+(-0.0805025225569)*x[1]**o+(0.389052254754)*x[1]+(-0.813210939614)*x[2]**o+(-0.593788787896)*x[2]
        arg[1,3,0,1]=(0.0101606353406)*x[0]**o+(-0.849874258855)*x[0]+(-0.0578973959381)*x[1]**o+(0.942113140424)*x[1]+(0.613716459216)*x[2]**o+(0.72631176833)*x[2]
        arg[1,3,1,0]=(0.264703608811)*x[0]**o+(0.457962548326)*x[0]+(-0.433795242673)*x[1]**o+(0.650995712848)*x[1]+(0.372471807967)*x[2]**o+(0.00416636841047)*x[2]
        arg[1,3,1,1]=(-0.73686451669)*x[0]**o+(-0.0425614910776)*x[0]+(-0.587820431311)*x[1]**o+(-0.377628617833)*x[1]+(-0.463964014406)*x[2]**o+(0.283273874989)*x[2]
        arg[1,3,2,0]=(-0.689341439143)*x[0]**o+(-0.675861918889)*x[0]+(-0.918622662352)*x[1]**o+(0.740990261888)*x[1]+(0.0454007677041)*x[2]**o+(0.513623361074)*x[2]
        arg[1,3,2,1]=(0.13229101848)*x[0]**o+(0.171873799817)*x[0]+(-0.154491174434)*x[1]**o+(0.897976884089)*x[1]+(-0.598886311821)*x[2]**o+(0.738696710918)*x[2]
        arg[1,4,0,0]=(0.59403636966)*x[0]**o+(0.93856848969)*x[0]+(0.80164849167)*x[1]**o+(-0.753322979568)*x[1]+(-0.107098073916)*x[2]**o+(-0.623047487973)*x[2]
        arg[1,4,0,1]=(0.291046099082)*x[0]**o+(-0.437982471049)*x[0]+(0.017395762947)*x[1]**o+(0.852440593923)*x[1]+(0.630825424359)*x[2]**o+(0.419347491727)*x[2]
        arg[1,4,1,0]=(-0.747962012701)*x[0]**o+(-0.11622032094)*x[0]+(0.324950549781)*x[1]**o+(0.950911398327)*x[1]+(-0.18098990205)*x[2]**o+(-0.676060633522)*x[2]
        arg[1,4,1,1]=(0.545224363128)*x[0]**o+(0.753376367121)*x[0]+(0.271426695753)*x[1]**o+(-0.139309844675)*x[1]+(-0.923434716375)*x[2]**o+(-0.985988324389)*x[2]
        arg[1,4,2,0]=(-0.532260822047)*x[0]**o+(0.901407135724)*x[0]+(0.937451330141)*x[1]**o+(0.285217666559)*x[1]+(-0.133028984824)*x[2]**o+(0.993965133696)*x[2]
        arg[1,4,2,1]=(0.2231815607)*x[0]**o+(0.994845537582)*x[0]+(0.0368727378195)*x[1]**o+(0.511772933777)*x[1]+(-0.871580215976)*x[2]**o+(-0.587010057329)*x[2]
        arg[2,0,0,0]=(-0.55149411913)*x[0]**o+(-0.905541278559)*x[0]+(0.461217373096)*x[1]**o+(-0.53359853241)*x[1]+(-0.666643290883)*x[2]**o+(0.151495826926)*x[2]
        arg[2,0,0,1]=(-0.980485062148)*x[0]**o+(0.150165616162)*x[0]+(0.59165725487)*x[1]**o+(-0.243853536017)*x[1]+(0.66816261686)*x[2]**o+(0.2425696131)*x[2]
        arg[2,0,1,0]=(-0.509337213758)*x[0]**o+(0.652936706687)*x[0]+(-0.849214338091)*x[1]**o+(-0.523354123549)*x[1]+(0.371515689887)*x[2]**o+(0.799989815414)*x[2]
        arg[2,0,1,1]=(0.461584544541)*x[0]**o+(0.662931355954)*x[0]+(0.744415634786)*x[1]**o+(-0.016743801256)*x[1]+(0.172945846325)*x[2]**o+(-0.0889827089341)*x[2]
        arg[2,0,2,0]=(0.506514791562)*x[0]**o+(0.974477160694)*x[0]+(-0.968485343167)*x[1]**o+(0.427480013599)*x[1]+(0.0460755879358)*x[2]**o+(0.186500994948)*x[2]
        arg[2,0,2,1]=(0.259831607993)*x[0]**o+(-0.0141290444125)*x[0]+(-0.968452177682)*x[1]**o+(0.635619125009)*x[1]+(-0.958848431827)*x[2]**o+(-0.0977241523308)*x[2]
        arg[2,1,0,0]=(0.204233222933)*x[0]**o+(0.834739381877)*x[0]+(0.19961185048)*x[1]**o+(-0.238470380578)*x[1]+(-0.749979713773)*x[2]**o+(-0.268168886251)*x[2]
        arg[2,1,0,1]=(0.998123643088)*x[0]**o+(0.635823733041)*x[0]+(-0.265447124419)*x[1]**o+(-0.225582751946)*x[1]+(0.91705937439)*x[2]**o+(-0.739461510792)*x[2]
        arg[2,1,1,0]=(0.0800715288646)*x[0]**o+(-0.37677196659)*x[0]+(-0.0737286350219)*x[1]**o+(0.032076062933)*x[1]+(0.245421912709)*x[2]**o+(-0.8378023935)*x[2]
        arg[2,1,1,1]=(-0.572376055603)*x[0]**o+(0.8675468469)*x[0]+(-0.533012889867)*x[1]**o+(-0.247215658297)*x[1]+(0.673144564693)*x[2]**o+(-0.749136108985)*x[2]
        arg[2,1,2,0]=(-0.661188192488)*x[0]**o+(-0.924435593008)*x[0]+(-0.791845808916)*x[1]**o+(0.00133765351092)*x[1]+(0.233676862873)*x[2]**o+(0.981812737962)*x[2]
        arg[2,1,2,1]=(0.332072445986)*x[0]**o+(-0.909373798588)*x[0]+(0.927025647819)*x[1]**o+(-0.936348942617)*x[1]+(0.251508592152)*x[2]**o+(-0.468147973987)*x[2]
        arg[2,2,0,0]=(0.892377893841)*x[0]**o+(-0.147933293091)*x[0]+(0.950491102076)*x[1]**o+(0.595250355926)*x[1]+(0.0169764649992)*x[2]**o+(-0.683137752586)*x[2]
        arg[2,2,0,1]=(0.00579397005645)*x[0]**o+(0.627197373917)*x[0]+(0.663574756196)*x[1]**o+(0.934106767513)*x[1]+(0.105844878022)*x[2]**o+(0.513322917688)*x[2]
        arg[2,2,1,0]=(-0.168776016133)*x[0]**o+(-0.940071612944)*x[0]+(0.0366053554566)*x[1]**o+(-0.628010909771)*x[1]+(0.250023045906)*x[2]**o+(-0.00327322332166)*x[2]
        arg[2,2,1,1]=(0.904155764376)*x[0]**o+(-0.258018307306)*x[0]+(0.65862820521)*x[1]**o+(0.434480798515)*x[1]+(0.4089419173)*x[2]**o+(-0.518032695598)*x[2]
        arg[2,2,2,0]=(-0.41470070958)*x[0]**o+(-0.783186751952)*x[0]+(0.548309561942)*x[1]**o+(0.837130592803)*x[1]+(-0.346808491848)*x[2]**o+(0.165416040985)*x[2]
        arg[2,2,2,1]=(0.997845909195)*x[0]**o+(0.448428042158)*x[0]+(-0.657050004311)*x[1]**o+(0.493104851747)*x[1]+(0.558315071197)*x[2]**o+(0.206659432923)*x[2]
        arg[2,3,0,0]=(0.433874360232)*x[0]**o+(-0.0635646197131)*x[0]+(-0.521690521523)*x[1]**o+(-0.298130684888)*x[1]+(-0.131905313644)*x[2]**o+(0.64389451112)*x[2]
        arg[2,3,0,1]=(0.457729315216)*x[0]**o+(-0.419323242135)*x[0]+(0.614544051135)*x[1]**o+(-0.701671999275)*x[1]+(0.905494300727)*x[2]**o+(-0.565716985337)*x[2]
        arg[2,3,1,0]=(0.256570728097)*x[0]**o+(0.866349285103)*x[0]+(-0.115768354896)*x[1]**o+(-0.613044456549)*x[1]+(-0.20074185156)*x[2]**o+(-0.188403571776)*x[2]
        arg[2,3,1,1]=(0.767479943414)*x[0]**o+(0.884066215007)*x[0]+(0.363855420056)*x[1]**o+(0.919698981104)*x[1]+(0.225747762317)*x[2]**o+(-0.220362271104)*x[2]
        arg[2,3,2,0]=(-0.746469501382)*x[0]**o+(0.680520398001)*x[0]+(0.926950996325)*x[1]**o+(-0.122831303044)*x[1]+(-0.27226561675)*x[2]**o+(0.933304087951)*x[2]
        arg[2,3,2,1]=(-0.0961434747216)*x[0]**o+(0.987660516767)*x[0]+(0.288175205045)*x[1]**o+(0.683720820403)*x[1]+(0.354403143642)*x[2]**o+(-0.745694089132)*x[2]
        arg[2,4,0,0]=(0.921750835986)*x[0]**o+(-0.0358861598394)*x[0]+(0.190129645259)*x[1]**o+(0.377067803331)*x[1]+(-0.0188615236582)*x[2]**o+(0.804533888824)*x[2]
        arg[2,4,0,1]=(-0.256132234484)*x[0]**o+(-0.102661411017)*x[0]+(-0.903465157009)*x[1]**o+(0.50133310168)*x[1]+(-0.0889102598619)*x[2]**o+(-0.0310174238365)*x[2]
        arg[2,4,1,0]=(0.58483798855)*x[0]**o+(0.967943344305)*x[0]+(-0.0701080005438)*x[1]**o+(0.489476104406)*x[1]+(0.615308427038)*x[2]**o+(-0.38065505709)*x[2]
        arg[2,4,1,1]=(0.647734130724)*x[0]**o+(-0.930934392719)*x[0]+(0.267370540491)*x[1]**o+(-0.0175019424556)*x[1]+(0.38593657129)*x[2]**o+(-0.985922292359)*x[2]
        arg[2,4,2,0]=(-0.289971479933)*x[0]**o+(0.882014313157)*x[0]+(0.147321067579)*x[1]**o+(-0.176771838489)*x[1]+(0.710916531544)*x[2]**o+(-0.0856290170401)*x[2]
        arg[2,4,2,1]=(-0.465930997733)*x[0]**o+(-0.305563942094)*x[0]+(-0.824535262061)*x[1]**o+(-0.917892226972)*x[1]+(0.624462871955)*x[2]**o+(0.544736500065)*x[2]
        arg[3,0,0,0]=(-0.673230815097)*x[0]**o+(-0.726292712917)*x[0]+(0.0531037021688)*x[1]**o+(-0.467219030159)*x[1]+(-0.0729861369718)*x[2]**o+(-0.440394335015)*x[2]
        arg[3,0,0,1]=(-0.358235295558)*x[0]**o+(-0.301184713671)*x[0]+(-0.496030649086)*x[1]**o+(-0.125922184777)*x[1]+(0.946715036973)*x[2]**o+(0.846370337893)*x[2]
        arg[3,0,1,0]=(-0.394175109612)*x[0]**o+(-0.85137582307)*x[0]+(-0.667804785044)*x[1]**o+(0.156762109891)*x[1]+(-0.616444700525)*x[2]**o+(0.0738729399356)*x[2]
        arg[3,0,1,1]=(-0.652875462534)*x[0]**o+(-0.0605650192541)*x[0]+(0.840539933823)*x[1]**o+(-0.422819423301)*x[1]+(0.113440226819)*x[2]**o+(-0.915154532306)*x[2]
        arg[3,0,2,0]=(-0.162863352328)*x[0]**o+(-0.364461332292)*x[0]+(-0.222181741398)*x[1]**o+(-0.563668008365)*x[1]+(0.118491770105)*x[2]**o+(0.0809155053022)*x[2]
        arg[3,0,2,1]=(-0.985278929437)*x[0]**o+(-0.919440760157)*x[0]+(-0.776910686)*x[1]**o+(-0.884270488943)*x[1]+(0.984279010363)*x[2]**o+(-0.489503278582)*x[2]
        arg[3,1,0,0]=(-0.806565449823)*x[0]**o+(-0.0851385375281)*x[0]+(-0.88851381773)*x[1]**o+(-0.666539867528)*x[1]+(0.154825531411)*x[2]**o+(0.764069849144)*x[2]
        arg[3,1,0,1]=(-0.222125783918)*x[0]**o+(0.568035675467)*x[0]+(0.470171778932)*x[1]**o+(-0.307608710881)*x[1]+(-0.836083742471)*x[2]**o+(0.553597553381)*x[2]
        arg[3,1,1,0]=(0.431237389847)*x[0]**o+(-0.141391012471)*x[0]+(0.726895025519)*x[1]**o+(-0.527868751679)*x[1]+(0.711761083137)*x[2]**o+(0.996992864103)*x[2]
        arg[3,1,1,1]=(-0.123885959924)*x[0]**o+(0.877034475529)*x[0]+(-0.0432492682928)*x[1]**o+(-0.370315494618)*x[1]+(0.243139719071)*x[2]**o+(-0.684318634902)*x[2]
        arg[3,1,2,0]=(-0.477214489355)*x[0]**o+(0.340808576619)*x[0]+(-0.0991802306558)*x[1]**o+(0.462599821908)*x[1]+(0.091768281564)*x[2]**o+(0.0315552838839)*x[2]
        arg[3,1,2,1]=(0.676189142894)*x[0]**o+(0.510136386913)*x[0]+(-0.558579041952)*x[1]**o+(0.961615047877)*x[1]+(0.510717386124)*x[2]**o+(0.774430183432)*x[2]
        arg[3,2,0,0]=(0.158370232444)*x[0]**o+(0.209424855992)*x[0]+(0.691438753013)*x[1]**o+(-0.781594298738)*x[1]+(-0.0188536627517)*x[2]**o+(0.136400798995)*x[2]
        arg[3,2,0,1]=(-0.440674414328)*x[0]**o+(0.118228612687)*x[0]+(-0.700026492613)*x[1]**o+(-0.981341215714)*x[1]+(-0.172847519578)*x[2]**o+(-0.423095202172)*x[2]
        arg[3,2,1,0]=(-0.0694960391695)*x[0]**o+(-0.66448610135)*x[0]+(-0.177572703748)*x[1]**o+(-0.060488161727)*x[1]+(0.306021853885)*x[2]**o+(0.0730338716187)*x[2]
        arg[3,2,1,1]=(-0.360662154547)*x[0]**o+(0.577679359648)*x[0]+(-0.515313534966)*x[1]**o+(0.217352687358)*x[1]+(0.728367577882)*x[2]**o+(0.38505849834)*x[2]
        arg[3,2,2,0]=(0.568671861232)*x[0]**o+(-0.0792490917882)*x[0]+(0.0188002885229)*x[1]**o+(-0.956862575759)*x[1]+(-0.844332442983)*x[2]**o+(0.670792928042)*x[2]
        arg[3,2,2,1]=(0.833099204713)*x[0]**o+(0.368482169933)*x[0]+(-0.926002498892)*x[1]**o+(-0.130646446568)*x[1]+(0.81164928914)*x[2]**o+(-0.641079919927)*x[2]
        arg[3,3,0,0]=(-0.277769350989)*x[0]**o+(0.753452018649)*x[0]+(-0.766322344505)*x[1]**o+(-0.679108224612)*x[1]+(0.0799499795424)*x[2]**o+(0.255751308975)*x[2]
        arg[3,3,0,1]=(0.429176753733)*x[0]**o+(0.123106050259)*x[0]+(-0.169037728992)*x[1]**o+(0.749245933687)*x[1]+(0.560140859355)*x[2]**o+(-0.937091840021)*x[2]
        arg[3,3,1,0]=(0.648792557593)*x[0]**o+(-0.057555187513)*x[0]+(0.883290331726)*x[1]**o+(0.539276905995)*x[1]+(0.473614408583)*x[2]**o+(0.723258340333)*x[2]
        arg[3,3,1,1]=(0.527298833572)*x[0]**o+(0.0753824058279)*x[0]+(0.719196099257)*x[1]**o+(-0.0124603350035)*x[1]+(-0.964484650253)*x[2]**o+(0.404444269694)*x[2]
        arg[3,3,2,0]=(-0.403640677462)*x[0]**o+(0.906195964766)*x[0]+(-0.561419078404)*x[1]**o+(0.645408717838)*x[1]+(0.671624799609)*x[2]**o+(-0.573901868621)*x[2]
        arg[3,3,2,1]=(-0.288117182165)*x[0]**o+(-0.391003584945)*x[0]+(-0.555037182038)*x[1]**o+(0.299362209017)*x[1]+(-0.552429724326)*x[2]**o+(-0.416923306774)*x[2]
        arg[3,4,0,0]=(-0.19594719192)*x[0]**o+(0.150898934692)*x[0]+(0.210935232044)*x[1]**o+(-0.319764293347)*x[1]+(-0.642450751195)*x[2]**o+(0.22231085168)*x[2]
        arg[3,4,0,1]=(-0.0514589460225)*x[0]**o+(0.0366025646754)*x[0]+(0.416280576691)*x[1]**o+(0.609178063951)*x[1]+(0.835611677062)*x[2]**o+(-0.131630571478)*x[2]
        arg[3,4,1,0]=(0.171509152214)*x[0]**o+(-0.761188271093)*x[0]+(-0.915951802896)*x[1]**o+(-0.451796709915)*x[1]+(-0.979847828162)*x[2]**o+(0.0566453393315)*x[2]
        arg[3,4,1,1]=(0.690427998102)*x[0]**o+(-0.358318625695)*x[0]+(0.330270871613)*x[1]**o+(-0.724392522137)*x[1]+(-0.535474623489)*x[2]**o+(-0.12341743885)*x[2]
        arg[3,4,2,0]=(0.488090753426)*x[0]**o+(-0.352417078372)*x[0]+(0.971988099319)*x[1]**o+(-0.288929516009)*x[1]+(0.612576734094)*x[2]**o+(-0.43799514603)*x[2]
        arg[3,4,2,1]=(-0.098768101356)*x[0]**o+(0.123300661895)*x[0]+(-0.0489724280789)*x[1]**o+(-0.422508669137)*x[1]+(-0.861483104139)*x[2]**o+(0.950198316001)*x[2]
        ref[0,0,0,0]=(0.287160004495)*x_ref[0]**o+(0.35112250158)*x_ref[0]+(0.061693427026)*x_ref[1]**o+(-0.752505628029)*x_ref[1]+(-0.57137756028)*x_ref[2]**o+(-0.0397598892504)*x_ref[2]
        ref[0,0,0,1]=(0.272799584607)*x_ref[0]**o+(-0.569947956933)*x_ref[0]+(0.322189261657)*x_ref[1]**o+(-0.440552272224)*x_ref[1]+(-0.10460241481)*x_ref[2]**o+(-0.884848984752)*x_ref[2]
        ref[0,0,1,0]=(0.147766301604)*x_ref[0]**o+(0.405651654428)*x_ref[0]+(-0.440827209529)*x_ref[1]**o+(-0.399952864271)*x_ref[1]+(-0.770191009459)*x_ref[2]**o+(-0.789853968049)*x_ref[2]
        ref[0,0,1,1]=(0.688053165375)*x_ref[0]**o+(-0.154890421924)*x_ref[0]+(0.583684164319)*x_ref[1]**o+(-0.398082068217)*x_ref[1]+(-0.573142923099)*x_ref[2]**o+(-0.205271877337)*x_ref[2]
        ref[0,0,2,0]=(0.807043275877)*x_ref[0]**o+(0.843046383231)*x_ref[0]+(-0.588146471402)*x_ref[1]**o+(0.157787304557)*x_ref[1]+(-0.771009748568)*x_ref[2]**o+(0.162090791609)*x_ref[2]
        ref[0,0,2,1]=(-0.0706800489392)*x_ref[0]**o+(0.499961344665)*x_ref[0]+(0.208926178797)*x_ref[1]**o+(-0.7462472769)*x_ref[1]+(-0.105475405648)*x_ref[2]**o+(-0.18670229948)*x_ref[2]
        ref[0,1,0,0]=(-0.696657281943)*x_ref[0]**o+(-0.0251785345759)*x_ref[0]+(-0.579575530451)*x_ref[1]**o+(-0.0628378665725)*x_ref[1]+(0.668395441302)*x_ref[2]**o+(0.813478105092)*x_ref[2]
        ref[0,1,0,1]=(-0.574784314876)*x_ref[0]**o+(-0.321268045204)*x_ref[0]+(0.160704464213)*x_ref[1]**o+(-0.676016748103)*x_ref[1]+(0.918971102865)*x_ref[2]**o+(0.252082410081)*x_ref[2]
        ref[0,1,1,0]=(-0.905062846817)*x_ref[0]**o+(0.788906658477)*x_ref[0]+(-0.408807611936)*x_ref[1]**o+(0.114686186437)*x_ref[1]+(-0.654685568493)*x_ref[2]**o+(0.0879202139254)*x_ref[2]
        ref[0,1,1,1]=(-0.68020231443)*x_ref[0]**o+(-0.82049202902)*x_ref[0]+(-0.958352740187)*x_ref[1]**o+(-0.33201745802)*x_ref[1]+(0.539810363682)*x_ref[2]**o+(0.744988573087)*x_ref[2]
        ref[0,1,2,0]=(-0.960884744997)*x_ref[0]**o+(-0.704157132503)*x_ref[0]+(0.282236905414)*x_ref[1]**o+(0.756363456433)*x_ref[1]+(-0.365400101792)*x_ref[2]**o+(0.975735846554)*x_ref[2]
        ref[0,1,2,1]=(-0.691780076951)*x_ref[0]**o+(-0.952575440059)*x_ref[0]+(-0.921247808098)*x_ref[1]**o+(-0.88872834095)*x_ref[1]+(0.534103013833)*x_ref[2]**o+(-0.602295186104)*x_ref[2]
        ref[0,2,0,0]=(0.460698415704)*x_ref[0]**o+(0.770243814897)*x_ref[0]+(-0.380381727574)*x_ref[1]**o+(0.476360845668)*x_ref[1]+(-0.681040425102)*x_ref[2]**o+(-0.244369435116)*x_ref[2]
        ref[0,2,0,1]=(0.251591652281)*x_ref[0]**o+(0.748234919686)*x_ref[0]+(-0.698189037307)*x_ref[1]**o+(-0.292400412508)*x_ref[1]+(0.613850696577)*x_ref[2]**o+(-0.942243226834)*x_ref[2]
        ref[0,2,1,0]=(-0.718090066328)*x_ref[0]**o+(-0.322499511794)*x_ref[0]+(-0.75105623397)*x_ref[1]**o+(0.526680271832)*x_ref[1]+(-0.186671531793)*x_ref[2]**o+(-0.449273875991)*x_ref[2]
        ref[0,2,1,1]=(-0.52165552452)*x_ref[0]**o+(0.902446006085)*x_ref[0]+(-0.915575172702)*x_ref[1]**o+(-0.150816194662)*x_ref[1]+(0.596782785689)*x_ref[2]**o+(0.546880865642)*x_ref[2]
        ref[0,2,2,0]=(-0.104486704816)*x_ref[0]**o+(0.723863699457)*x_ref[0]+(0.758622117279)*x_ref[1]**o+(0.372271580145)*x_ref[1]+(-0.0820570383792)*x_ref[2]**o+(-0.63539781104)*x_ref[2]
        ref[0,2,2,1]=(0.0355930823603)*x_ref[0]**o+(0.0931942600432)*x_ref[0]+(-0.118096985978)*x_ref[1]**o+(0.296983782212)*x_ref[1]+(-0.864491763712)*x_ref[2]**o+(0.942230663978)*x_ref[2]
        ref[0,3,0,0]=(0.371153709815)*x_ref[0]**o+(0.0427309083893)*x_ref[0]+(-0.886563919084)*x_ref[1]**o+(-0.14946513404)*x_ref[1]+(0.80903804519)*x_ref[2]**o+(0.69728533073)*x_ref[2]
        ref[0,3,0,1]=(0.0593930396142)*x_ref[0]**o+(-0.370327974089)*x_ref[0]+(-0.388354464683)*x_ref[1]**o+(-0.367648771812)*x_ref[1]+(-0.573754821458)*x_ref[2]**o+(-0.947499155479)*x_ref[2]
        ref[0,3,1,0]=(-0.869817317)*x_ref[0]**o+(0.874317772678)*x_ref[0]+(-0.870060178375)*x_ref[1]**o+(0.66571074524)*x_ref[1]+(-0.669052377378)*x_ref[2]**o+(0.940212680575)*x_ref[2]
        ref[0,3,1,1]=(-0.417445606772)*x_ref[0]**o+(0.942606057619)*x_ref[0]+(-0.133936034409)*x_ref[1]**o+(-0.363274280866)*x_ref[1]+(0.83487439935)*x_ref[2]**o+(-0.108398981461)*x_ref[2]
        ref[0,3,2,0]=(0.363623383293)*x_ref[0]**o+(0.624267309217)*x_ref[0]+(0.801719984899)*x_ref[1]**o+(0.189212370122)*x_ref[1]+(0.464633630928)*x_ref[2]**o+(-0.979808318438)*x_ref[2]
        ref[0,3,2,1]=(0.85384408805)*x_ref[0]**o+(-0.889214754116)*x_ref[0]+(0.164915391224)*x_ref[1]**o+(0.0168991405201)*x_ref[1]+(0.097171994236)*x_ref[2]**o+(-0.146094197339)*x_ref[2]
        ref[0,4,0,0]=(0.778356158972)*x_ref[0]**o+(0.910241265149)*x_ref[0]+(0.123690863272)*x_ref[1]**o+(0.87802121779)*x_ref[1]+(-0.714615668595)*x_ref[2]**o+(-0.437472831583)*x_ref[2]
        ref[0,4,0,1]=(-0.157185481783)*x_ref[0]**o+(0.352802163968)*x_ref[0]+(0.390184123612)*x_ref[1]**o+(-0.680415093098)*x_ref[1]+(-0.121195042454)*x_ref[2]**o+(-0.713693712117)*x_ref[2]
        ref[0,4,1,0]=(0.778115668646)*x_ref[0]**o+(-0.111729042892)*x_ref[0]+(0.451518002735)*x_ref[1]**o+(0.232731183374)*x_ref[1]+(-0.764246158086)*x_ref[2]**o+(-0.795525155328)*x_ref[2]
        ref[0,4,1,1]=(-0.880652367027)*x_ref[0]**o+(-0.314966079126)*x_ref[0]+(-0.640720373065)*x_ref[1]**o+(0.336230541901)*x_ref[1]+(0.566956999526)*x_ref[2]**o+(0.74098258811)*x_ref[2]
        ref[0,4,2,0]=(0.908657274301)*x_ref[0]**o+(0.636427823614)*x_ref[0]+(0.650137414386)*x_ref[1]**o+(0.695558379005)*x_ref[1]+(-0.489852401087)*x_ref[2]**o+(-0.479227108044)*x_ref[2]
        ref[0,4,2,1]=(0.015734241525)*x_ref[0]**o+(0.0719884727088)*x_ref[0]+(0.811864895261)*x_ref[1]**o+(0.959127986784)*x_ref[1]+(-0.238146663368)*x_ref[2]**o+(-0.454939322571)*x_ref[2]
        ref[1,0,0,0]=(-0.918218246559)*x_ref[0]**o+(-0.662787169269)*x_ref[0]+(-0.954035144111)*x_ref[1]**o+(-0.647264685228)*x_ref[1]+(-0.834085960213)*x_ref[2]**o+(-0.167541951024)*x_ref[2]
        ref[1,0,0,1]=(-0.670950869382)*x_ref[0]**o+(0.826472159989)*x_ref[0]+(0.180410872786)*x_ref[1]**o+(-0.715162871594)*x_ref[1]+(0.653287853169)*x_ref[2]**o+(0.892060529625)*x_ref[2]
        ref[1,0,1,0]=(0.398789786407)*x_ref[0]**o+(0.689408033554)*x_ref[0]+(0.124338245128)*x_ref[1]**o+(0.748280808587)*x_ref[1]+(0.418357122274)*x_ref[2]**o+(0.230707440075)*x_ref[2]
        ref[1,0,1,1]=(-0.912778964055)*x_ref[0]**o+(0.904288708694)*x_ref[0]+(-0.874744856901)*x_ref[1]**o+(0.550162195719)*x_ref[1]+(-0.7526423373)*x_ref[2]**o+(0.820993231654)*x_ref[2]
        ref[1,0,2,0]=(-0.843459684648)*x_ref[0]**o+(-0.579708764091)*x_ref[0]+(-0.876072131849)*x_ref[1]**o+(0.933694599696)*x_ref[1]+(0.507781209021)*x_ref[2]**o+(-0.719729482133)*x_ref[2]
        ref[1,0,2,1]=(0.594566925796)*x_ref[0]**o+(-0.994301292358)*x_ref[0]+(0.808114389564)*x_ref[1]**o+(-0.970803712864)*x_ref[1]+(0.433544910389)*x_ref[2]**o+(0.260659440299)*x_ref[2]
        ref[1,1,0,0]=(0.997377553172)*x_ref[0]**o+(0.458428616383)*x_ref[0]+(0.357280778493)*x_ref[1]**o+(0.349837110741)*x_ref[1]+(-0.709592214445)*x_ref[2]**o+(-0.630000739033)*x_ref[2]
        ref[1,1,0,1]=(0.121595650501)*x_ref[0]**o+(0.40870120924)*x_ref[0]+(0.576917123807)*x_ref[1]**o+(-0.932268308325)*x_ref[1]+(0.710043472116)*x_ref[2]**o+(-0.716528412077)*x_ref[2]
        ref[1,1,1,0]=(-0.151416035033)*x_ref[0]**o+(0.0255347649787)*x_ref[0]+(-0.752987701889)*x_ref[1]**o+(-0.480969875813)*x_ref[1]+(-0.610514504179)*x_ref[2]**o+(0.959274757402)*x_ref[2]
        ref[1,1,1,1]=(0.76695706106)*x_ref[0]**o+(-0.0750965975512)*x_ref[0]+(0.846833900735)*x_ref[1]**o+(-0.105838465439)*x_ref[1]+(0.541027486457)*x_ref[2]**o+(0.152543017589)*x_ref[2]
        ref[1,1,2,0]=(0.0856977181865)*x_ref[0]**o+(-0.188139369706)*x_ref[0]+(0.936071565289)*x_ref[1]**o+(-0.3958713036)*x_ref[1]+(-0.350945878893)*x_ref[2]**o+(0.382302674883)*x_ref[2]
        ref[1,1,2,1]=(0.873054234789)*x_ref[0]**o+(-0.490869716146)*x_ref[0]+(0.46720526212)*x_ref[1]**o+(-0.334203412334)*x_ref[1]+(-0.81140895636)*x_ref[2]**o+(-0.973858971749)*x_ref[2]
        ref[1,2,0,0]=(-0.716676352443)*x_ref[0]**o+(0.659821396277)*x_ref[0]+(0.335964292954)*x_ref[1]**o+(0.11609664773)*x_ref[1]+(-0.889186749466)*x_ref[2]**o+(0.454231973946)*x_ref[2]
        ref[1,2,0,1]=(0.367048872648)*x_ref[0]**o+(0.276083713126)*x_ref[0]+(0.686639128043)*x_ref[1]**o+(-0.991414415293)*x_ref[1]+(0.25360515964)*x_ref[2]**o+(-0.377623315093)*x_ref[2]
        ref[1,2,1,0]=(-0.723089358777)*x_ref[0]**o+(-0.516690112406)*x_ref[0]+(0.262334273979)*x_ref[1]**o+(0.140773502537)*x_ref[1]+(0.447333972396)*x_ref[2]**o+(-0.551395692966)*x_ref[2]
        ref[1,2,1,1]=(0.738791182491)*x_ref[0]**o+(-0.978120626668)*x_ref[0]+(-0.217970433713)*x_ref[1]**o+(0.0181011067492)*x_ref[1]+(0.756322867523)*x_ref[2]**o+(0.495269195978)*x_ref[2]
        ref[1,2,2,0]=(-0.630506449854)*x_ref[0]**o+(-0.0865482364231)*x_ref[0]+(-0.415896403157)*x_ref[1]**o+(-0.712833149895)*x_ref[1]+(-0.180482789777)*x_ref[2]**o+(0.00788290661235)*x_ref[2]
        ref[1,2,2,1]=(0.0946913865377)*x_ref[0]**o+(0.752854827716)*x_ref[0]+(0.257846220843)*x_ref[1]**o+(-0.953926983578)*x_ref[1]+(-0.388505920682)*x_ref[2]**o+(-0.10751401251)*x_ref[2]
        ref[1,3,0,0]=(0.694014030611)*x_ref[0]**o+(0.0433103547797)*x_ref[0]+(-0.0805025225569)*x_ref[1]**o+(0.389052254754)*x_ref[1]+(-0.813210939614)*x_ref[2]**o+(-0.593788787896)*x_ref[2]
        ref[1,3,0,1]=(0.0101606353406)*x_ref[0]**o+(-0.849874258855)*x_ref[0]+(-0.0578973959381)*x_ref[1]**o+(0.942113140424)*x_ref[1]+(0.613716459216)*x_ref[2]**o+(0.72631176833)*x_ref[2]
        ref[1,3,1,0]=(0.264703608811)*x_ref[0]**o+(0.457962548326)*x_ref[0]+(-0.433795242673)*x_ref[1]**o+(0.650995712848)*x_ref[1]+(0.372471807967)*x_ref[2]**o+(0.00416636841047)*x_ref[2]
        ref[1,3,1,1]=(-0.73686451669)*x_ref[0]**o+(-0.0425614910776)*x_ref[0]+(-0.587820431311)*x_ref[1]**o+(-0.377628617833)*x_ref[1]+(-0.463964014406)*x_ref[2]**o+(0.283273874989)*x_ref[2]
        ref[1,3,2,0]=(-0.689341439143)*x_ref[0]**o+(-0.675861918889)*x_ref[0]+(-0.918622662352)*x_ref[1]**o+(0.740990261888)*x_ref[1]+(0.0454007677041)*x_ref[2]**o+(0.513623361074)*x_ref[2]
        ref[1,3,2,1]=(0.13229101848)*x_ref[0]**o+(0.171873799817)*x_ref[0]+(-0.154491174434)*x_ref[1]**o+(0.897976884089)*x_ref[1]+(-0.598886311821)*x_ref[2]**o+(0.738696710918)*x_ref[2]
        ref[1,4,0,0]=(0.59403636966)*x_ref[0]**o+(0.93856848969)*x_ref[0]+(0.80164849167)*x_ref[1]**o+(-0.753322979568)*x_ref[1]+(-0.107098073916)*x_ref[2]**o+(-0.623047487973)*x_ref[2]
        ref[1,4,0,1]=(0.291046099082)*x_ref[0]**o+(-0.437982471049)*x_ref[0]+(0.017395762947)*x_ref[1]**o+(0.852440593923)*x_ref[1]+(0.630825424359)*x_ref[2]**o+(0.419347491727)*x_ref[2]
        ref[1,4,1,0]=(-0.747962012701)*x_ref[0]**o+(-0.11622032094)*x_ref[0]+(0.324950549781)*x_ref[1]**o+(0.950911398327)*x_ref[1]+(-0.18098990205)*x_ref[2]**o+(-0.676060633522)*x_ref[2]
        ref[1,4,1,1]=(0.545224363128)*x_ref[0]**o+(0.753376367121)*x_ref[0]+(0.271426695753)*x_ref[1]**o+(-0.139309844675)*x_ref[1]+(-0.923434716375)*x_ref[2]**o+(-0.985988324389)*x_ref[2]
        ref[1,4,2,0]=(-0.532260822047)*x_ref[0]**o+(0.901407135724)*x_ref[0]+(0.937451330141)*x_ref[1]**o+(0.285217666559)*x_ref[1]+(-0.133028984824)*x_ref[2]**o+(0.993965133696)*x_ref[2]
        ref[1,4,2,1]=(0.2231815607)*x_ref[0]**o+(0.994845537582)*x_ref[0]+(0.0368727378195)*x_ref[1]**o+(0.511772933777)*x_ref[1]+(-0.871580215976)*x_ref[2]**o+(-0.587010057329)*x_ref[2]
        ref[2,0,0,0]=(-0.55149411913)*x_ref[0]**o+(-0.905541278559)*x_ref[0]+(0.461217373096)*x_ref[1]**o+(-0.53359853241)*x_ref[1]+(-0.666643290883)*x_ref[2]**o+(0.151495826926)*x_ref[2]
        ref[2,0,0,1]=(-0.980485062148)*x_ref[0]**o+(0.150165616162)*x_ref[0]+(0.59165725487)*x_ref[1]**o+(-0.243853536017)*x_ref[1]+(0.66816261686)*x_ref[2]**o+(0.2425696131)*x_ref[2]
        ref[2,0,1,0]=(-0.509337213758)*x_ref[0]**o+(0.652936706687)*x_ref[0]+(-0.849214338091)*x_ref[1]**o+(-0.523354123549)*x_ref[1]+(0.371515689887)*x_ref[2]**o+(0.799989815414)*x_ref[2]
        ref[2,0,1,1]=(0.461584544541)*x_ref[0]**o+(0.662931355954)*x_ref[0]+(0.744415634786)*x_ref[1]**o+(-0.016743801256)*x_ref[1]+(0.172945846325)*x_ref[2]**o+(-0.0889827089341)*x_ref[2]
        ref[2,0,2,0]=(0.506514791562)*x_ref[0]**o+(0.974477160694)*x_ref[0]+(-0.968485343167)*x_ref[1]**o+(0.427480013599)*x_ref[1]+(0.0460755879358)*x_ref[2]**o+(0.186500994948)*x_ref[2]
        ref[2,0,2,1]=(0.259831607993)*x_ref[0]**o+(-0.0141290444125)*x_ref[0]+(-0.968452177682)*x_ref[1]**o+(0.635619125009)*x_ref[1]+(-0.958848431827)*x_ref[2]**o+(-0.0977241523308)*x_ref[2]
        ref[2,1,0,0]=(0.204233222933)*x_ref[0]**o+(0.834739381877)*x_ref[0]+(0.19961185048)*x_ref[1]**o+(-0.238470380578)*x_ref[1]+(-0.749979713773)*x_ref[2]**o+(-0.268168886251)*x_ref[2]
        ref[2,1,0,1]=(0.998123643088)*x_ref[0]**o+(0.635823733041)*x_ref[0]+(-0.265447124419)*x_ref[1]**o+(-0.225582751946)*x_ref[1]+(0.91705937439)*x_ref[2]**o+(-0.739461510792)*x_ref[2]
        ref[2,1,1,0]=(0.0800715288646)*x_ref[0]**o+(-0.37677196659)*x_ref[0]+(-0.0737286350219)*x_ref[1]**o+(0.032076062933)*x_ref[1]+(0.245421912709)*x_ref[2]**o+(-0.8378023935)*x_ref[2]
        ref[2,1,1,1]=(-0.572376055603)*x_ref[0]**o+(0.8675468469)*x_ref[0]+(-0.533012889867)*x_ref[1]**o+(-0.247215658297)*x_ref[1]+(0.673144564693)*x_ref[2]**o+(-0.749136108985)*x_ref[2]
        ref[2,1,2,0]=(-0.661188192488)*x_ref[0]**o+(-0.924435593008)*x_ref[0]+(-0.791845808916)*x_ref[1]**o+(0.00133765351092)*x_ref[1]+(0.233676862873)*x_ref[2]**o+(0.981812737962)*x_ref[2]
        ref[2,1,2,1]=(0.332072445986)*x_ref[0]**o+(-0.909373798588)*x_ref[0]+(0.927025647819)*x_ref[1]**o+(-0.936348942617)*x_ref[1]+(0.251508592152)*x_ref[2]**o+(-0.468147973987)*x_ref[2]
        ref[2,2,0,0]=(0.892377893841)*x_ref[0]**o+(-0.147933293091)*x_ref[0]+(0.950491102076)*x_ref[1]**o+(0.595250355926)*x_ref[1]+(0.0169764649992)*x_ref[2]**o+(-0.683137752586)*x_ref[2]
        ref[2,2,0,1]=(0.00579397005645)*x_ref[0]**o+(0.627197373917)*x_ref[0]+(0.663574756196)*x_ref[1]**o+(0.934106767513)*x_ref[1]+(0.105844878022)*x_ref[2]**o+(0.513322917688)*x_ref[2]
        ref[2,2,1,0]=(-0.168776016133)*x_ref[0]**o+(-0.940071612944)*x_ref[0]+(0.0366053554566)*x_ref[1]**o+(-0.628010909771)*x_ref[1]+(0.250023045906)*x_ref[2]**o+(-0.00327322332166)*x_ref[2]
        ref[2,2,1,1]=(0.904155764376)*x_ref[0]**o+(-0.258018307306)*x_ref[0]+(0.65862820521)*x_ref[1]**o+(0.434480798515)*x_ref[1]+(0.4089419173)*x_ref[2]**o+(-0.518032695598)*x_ref[2]
        ref[2,2,2,0]=(-0.41470070958)*x_ref[0]**o+(-0.783186751952)*x_ref[0]+(0.548309561942)*x_ref[1]**o+(0.837130592803)*x_ref[1]+(-0.346808491848)*x_ref[2]**o+(0.165416040985)*x_ref[2]
        ref[2,2,2,1]=(0.997845909195)*x_ref[0]**o+(0.448428042158)*x_ref[0]+(-0.657050004311)*x_ref[1]**o+(0.493104851747)*x_ref[1]+(0.558315071197)*x_ref[2]**o+(0.206659432923)*x_ref[2]
        ref[2,3,0,0]=(0.433874360232)*x_ref[0]**o+(-0.0635646197131)*x_ref[0]+(-0.521690521523)*x_ref[1]**o+(-0.298130684888)*x_ref[1]+(-0.131905313644)*x_ref[2]**o+(0.64389451112)*x_ref[2]
        ref[2,3,0,1]=(0.457729315216)*x_ref[0]**o+(-0.419323242135)*x_ref[0]+(0.614544051135)*x_ref[1]**o+(-0.701671999275)*x_ref[1]+(0.905494300727)*x_ref[2]**o+(-0.565716985337)*x_ref[2]
        ref[2,3,1,0]=(0.256570728097)*x_ref[0]**o+(0.866349285103)*x_ref[0]+(-0.115768354896)*x_ref[1]**o+(-0.613044456549)*x_ref[1]+(-0.20074185156)*x_ref[2]**o+(-0.188403571776)*x_ref[2]
        ref[2,3,1,1]=(0.767479943414)*x_ref[0]**o+(0.884066215007)*x_ref[0]+(0.363855420056)*x_ref[1]**o+(0.919698981104)*x_ref[1]+(0.225747762317)*x_ref[2]**o+(-0.220362271104)*x_ref[2]
        ref[2,3,2,0]=(-0.746469501382)*x_ref[0]**o+(0.680520398001)*x_ref[0]+(0.926950996325)*x_ref[1]**o+(-0.122831303044)*x_ref[1]+(-0.27226561675)*x_ref[2]**o+(0.933304087951)*x_ref[2]
        ref[2,3,2,1]=(-0.0961434747216)*x_ref[0]**o+(0.987660516767)*x_ref[0]+(0.288175205045)*x_ref[1]**o+(0.683720820403)*x_ref[1]+(0.354403143642)*x_ref[2]**o+(-0.745694089132)*x_ref[2]
        ref[2,4,0,0]=(0.921750835986)*x_ref[0]**o+(-0.0358861598394)*x_ref[0]+(0.190129645259)*x_ref[1]**o+(0.377067803331)*x_ref[1]+(-0.0188615236582)*x_ref[2]**o+(0.804533888824)*x_ref[2]
        ref[2,4,0,1]=(-0.256132234484)*x_ref[0]**o+(-0.102661411017)*x_ref[0]+(-0.903465157009)*x_ref[1]**o+(0.50133310168)*x_ref[1]+(-0.0889102598619)*x_ref[2]**o+(-0.0310174238365)*x_ref[2]
        ref[2,4,1,0]=(0.58483798855)*x_ref[0]**o+(0.967943344305)*x_ref[0]+(-0.0701080005438)*x_ref[1]**o+(0.489476104406)*x_ref[1]+(0.615308427038)*x_ref[2]**o+(-0.38065505709)*x_ref[2]
        ref[2,4,1,1]=(0.647734130724)*x_ref[0]**o+(-0.930934392719)*x_ref[0]+(0.267370540491)*x_ref[1]**o+(-0.0175019424556)*x_ref[1]+(0.38593657129)*x_ref[2]**o+(-0.985922292359)*x_ref[2]
        ref[2,4,2,0]=(-0.289971479933)*x_ref[0]**o+(0.882014313157)*x_ref[0]+(0.147321067579)*x_ref[1]**o+(-0.176771838489)*x_ref[1]+(0.710916531544)*x_ref[2]**o+(-0.0856290170401)*x_ref[2]
        ref[2,4,2,1]=(-0.465930997733)*x_ref[0]**o+(-0.305563942094)*x_ref[0]+(-0.824535262061)*x_ref[1]**o+(-0.917892226972)*x_ref[1]+(0.624462871955)*x_ref[2]**o+(0.544736500065)*x_ref[2]
        ref[3,0,0,0]=(-0.673230815097)*x_ref[0]**o+(-0.726292712917)*x_ref[0]+(0.0531037021688)*x_ref[1]**o+(-0.467219030159)*x_ref[1]+(-0.0729861369718)*x_ref[2]**o+(-0.440394335015)*x_ref[2]
        ref[3,0,0,1]=(-0.358235295558)*x_ref[0]**o+(-0.301184713671)*x_ref[0]+(-0.496030649086)*x_ref[1]**o+(-0.125922184777)*x_ref[1]+(0.946715036973)*x_ref[2]**o+(0.846370337893)*x_ref[2]
        ref[3,0,1,0]=(-0.394175109612)*x_ref[0]**o+(-0.85137582307)*x_ref[0]+(-0.667804785044)*x_ref[1]**o+(0.156762109891)*x_ref[1]+(-0.616444700525)*x_ref[2]**o+(0.0738729399356)*x_ref[2]
        ref[3,0,1,1]=(-0.652875462534)*x_ref[0]**o+(-0.0605650192541)*x_ref[0]+(0.840539933823)*x_ref[1]**o+(-0.422819423301)*x_ref[1]+(0.113440226819)*x_ref[2]**o+(-0.915154532306)*x_ref[2]
        ref[3,0,2,0]=(-0.162863352328)*x_ref[0]**o+(-0.364461332292)*x_ref[0]+(-0.222181741398)*x_ref[1]**o+(-0.563668008365)*x_ref[1]+(0.118491770105)*x_ref[2]**o+(0.0809155053022)*x_ref[2]
        ref[3,0,2,1]=(-0.985278929437)*x_ref[0]**o+(-0.919440760157)*x_ref[0]+(-0.776910686)*x_ref[1]**o+(-0.884270488943)*x_ref[1]+(0.984279010363)*x_ref[2]**o+(-0.489503278582)*x_ref[2]
        ref[3,1,0,0]=(-0.806565449823)*x_ref[0]**o+(-0.0851385375281)*x_ref[0]+(-0.88851381773)*x_ref[1]**o+(-0.666539867528)*x_ref[1]+(0.154825531411)*x_ref[2]**o+(0.764069849144)*x_ref[2]
        ref[3,1,0,1]=(-0.222125783918)*x_ref[0]**o+(0.568035675467)*x_ref[0]+(0.470171778932)*x_ref[1]**o+(-0.307608710881)*x_ref[1]+(-0.836083742471)*x_ref[2]**o+(0.553597553381)*x_ref[2]
        ref[3,1,1,0]=(0.431237389847)*x_ref[0]**o+(-0.141391012471)*x_ref[0]+(0.726895025519)*x_ref[1]**o+(-0.527868751679)*x_ref[1]+(0.711761083137)*x_ref[2]**o+(0.996992864103)*x_ref[2]
        ref[3,1,1,1]=(-0.123885959924)*x_ref[0]**o+(0.877034475529)*x_ref[0]+(-0.0432492682928)*x_ref[1]**o+(-0.370315494618)*x_ref[1]+(0.243139719071)*x_ref[2]**o+(-0.684318634902)*x_ref[2]
        ref[3,1,2,0]=(-0.477214489355)*x_ref[0]**o+(0.340808576619)*x_ref[0]+(-0.0991802306558)*x_ref[1]**o+(0.462599821908)*x_ref[1]+(0.091768281564)*x_ref[2]**o+(0.0315552838839)*x_ref[2]
        ref[3,1,2,1]=(0.676189142894)*x_ref[0]**o+(0.510136386913)*x_ref[0]+(-0.558579041952)*x_ref[1]**o+(0.961615047877)*x_ref[1]+(0.510717386124)*x_ref[2]**o+(0.774430183432)*x_ref[2]
        ref[3,2,0,0]=(0.158370232444)*x_ref[0]**o+(0.209424855992)*x_ref[0]+(0.691438753013)*x_ref[1]**o+(-0.781594298738)*x_ref[1]+(-0.0188536627517)*x_ref[2]**o+(0.136400798995)*x_ref[2]
        ref[3,2,0,1]=(-0.440674414328)*x_ref[0]**o+(0.118228612687)*x_ref[0]+(-0.700026492613)*x_ref[1]**o+(-0.981341215714)*x_ref[1]+(-0.172847519578)*x_ref[2]**o+(-0.423095202172)*x_ref[2]
        ref[3,2,1,0]=(-0.0694960391695)*x_ref[0]**o+(-0.66448610135)*x_ref[0]+(-0.177572703748)*x_ref[1]**o+(-0.060488161727)*x_ref[1]+(0.306021853885)*x_ref[2]**o+(0.0730338716187)*x_ref[2]
        ref[3,2,1,1]=(-0.360662154547)*x_ref[0]**o+(0.577679359648)*x_ref[0]+(-0.515313534966)*x_ref[1]**o+(0.217352687358)*x_ref[1]+(0.728367577882)*x_ref[2]**o+(0.38505849834)*x_ref[2]
        ref[3,2,2,0]=(0.568671861232)*x_ref[0]**o+(-0.0792490917882)*x_ref[0]+(0.0188002885229)*x_ref[1]**o+(-0.956862575759)*x_ref[1]+(-0.844332442983)*x_ref[2]**o+(0.670792928042)*x_ref[2]
        ref[3,2,2,1]=(0.833099204713)*x_ref[0]**o+(0.368482169933)*x_ref[0]+(-0.926002498892)*x_ref[1]**o+(-0.130646446568)*x_ref[1]+(0.81164928914)*x_ref[2]**o+(-0.641079919927)*x_ref[2]
        ref[3,3,0,0]=(-0.277769350989)*x_ref[0]**o+(0.753452018649)*x_ref[0]+(-0.766322344505)*x_ref[1]**o+(-0.679108224612)*x_ref[1]+(0.0799499795424)*x_ref[2]**o+(0.255751308975)*x_ref[2]
        ref[3,3,0,1]=(0.429176753733)*x_ref[0]**o+(0.123106050259)*x_ref[0]+(-0.169037728992)*x_ref[1]**o+(0.749245933687)*x_ref[1]+(0.560140859355)*x_ref[2]**o+(-0.937091840021)*x_ref[2]
        ref[3,3,1,0]=(0.648792557593)*x_ref[0]**o+(-0.057555187513)*x_ref[0]+(0.883290331726)*x_ref[1]**o+(0.539276905995)*x_ref[1]+(0.473614408583)*x_ref[2]**o+(0.723258340333)*x_ref[2]
        ref[3,3,1,1]=(0.527298833572)*x_ref[0]**o+(0.0753824058279)*x_ref[0]+(0.719196099257)*x_ref[1]**o+(-0.0124603350035)*x_ref[1]+(-0.964484650253)*x_ref[2]**o+(0.404444269694)*x_ref[2]
        ref[3,3,2,0]=(-0.403640677462)*x_ref[0]**o+(0.906195964766)*x_ref[0]+(-0.561419078404)*x_ref[1]**o+(0.645408717838)*x_ref[1]+(0.671624799609)*x_ref[2]**o+(-0.573901868621)*x_ref[2]
        ref[3,3,2,1]=(-0.288117182165)*x_ref[0]**o+(-0.391003584945)*x_ref[0]+(-0.555037182038)*x_ref[1]**o+(0.299362209017)*x_ref[1]+(-0.552429724326)*x_ref[2]**o+(-0.416923306774)*x_ref[2]
        ref[3,4,0,0]=(-0.19594719192)*x_ref[0]**o+(0.150898934692)*x_ref[0]+(0.210935232044)*x_ref[1]**o+(-0.319764293347)*x_ref[1]+(-0.642450751195)*x_ref[2]**o+(0.22231085168)*x_ref[2]
        ref[3,4,0,1]=(-0.0514589460225)*x_ref[0]**o+(0.0366025646754)*x_ref[0]+(0.416280576691)*x_ref[1]**o+(0.609178063951)*x_ref[1]+(0.835611677062)*x_ref[2]**o+(-0.131630571478)*x_ref[2]
        ref[3,4,1,0]=(0.171509152214)*x_ref[0]**o+(-0.761188271093)*x_ref[0]+(-0.915951802896)*x_ref[1]**o+(-0.451796709915)*x_ref[1]+(-0.979847828162)*x_ref[2]**o+(0.0566453393315)*x_ref[2]
        ref[3,4,1,1]=(0.690427998102)*x_ref[0]**o+(-0.358318625695)*x_ref[0]+(0.330270871613)*x_ref[1]**o+(-0.724392522137)*x_ref[1]+(-0.535474623489)*x_ref[2]**o+(-0.12341743885)*x_ref[2]
        ref[3,4,2,0]=(0.488090753426)*x_ref[0]**o+(-0.352417078372)*x_ref[0]+(0.971988099319)*x_ref[1]**o+(-0.288929516009)*x_ref[1]+(0.612576734094)*x_ref[2]**o+(-0.43799514603)*x_ref[2]
        ref[3,4,2,1]=(-0.098768101356)*x_ref[0]**o+(0.123300661895)*x_ref[0]+(-0.0489724280789)*x_ref[1]**o+(-0.422508669137)*x_ref[1]+(-0.861483104139)*x_ref[2]**o+(0.950198316001)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.724373964817)*x[0]**o+(-0.0560850774901)*x[0]+(0.201815078281)*x[1]**o+(-0.721683697048)*x[1]
        ref=(0.724373964817)*x_ref[0]**o+(-0.0560850774901)*x_ref[0]+(0.201815078281)*x_ref[1]**o+(-0.721683697048)*x_ref[1]
      else:
        arg=(0.390296010052)*x[0]**o+(-0.279464234123)*x[0]+(-0.791334947488)*x[1]**o+(-0.0210629346704)*x[1]+(0.0816556526012)*x[2]**o+(0.77326269137)*x[2]
        ref=(0.390296010052)*x_ref[0]**o+(-0.279464234123)*x_ref[0]+(-0.791334947488)*x_ref[1]**o+(-0.0210629346704)*x_ref[1]+(0.0816556526012)*x_ref[2]**o+(0.77326269137)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.645585823917)*x[0]**o+(-0.203443728436)*x[0]+(0.278126891692)*x[1]**o+(0.922093564911)*x[1]
        arg[1]=(0.23370137029)*x[0]**o+(0.84371173541)*x[0]+(0.99297574548)*x[1]**o+(0.72481785931)*x[1]
        ref[0]=(0.645585823917)*x_ref[0]**o+(-0.203443728436)*x_ref[0]+(0.278126891692)*x_ref[1]**o+(0.922093564911)*x_ref[1]
        ref[1]=(0.23370137029)*x_ref[0]**o+(0.84371173541)*x_ref[0]+(0.99297574548)*x_ref[1]**o+(0.72481785931)*x_ref[1]
      else:
        arg[0]=(0.159998049951)*x[0]**o+(-0.863367327023)*x[0]+(-0.0628498246638)*x[1]**o+(0.685139358273)*x[1]+(0.783213360641)*x[2]**o+(0.397526421338)*x[2]
        arg[1]=(0.165519279053)*x[0]**o+(-0.0936018967661)*x[0]+(0.863047696619)*x[1]**o+(0.653118186434)*x[1]+(-0.408224131944)*x[2]**o+(0.82400132854)*x[2]
        ref[0]=(0.159998049951)*x_ref[0]**o+(-0.863367327023)*x_ref[0]+(-0.0628498246638)*x_ref[1]**o+(0.685139358273)*x_ref[1]+(0.783213360641)*x_ref[2]**o+(0.397526421338)*x_ref[2]
        ref[1]=(0.165519279053)*x_ref[0]**o+(-0.0936018967661)*x_ref[0]+(0.863047696619)*x_ref[1]**o+(0.653118186434)*x_ref[1]+(-0.408224131944)*x_ref[2]**o+(0.82400132854)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.490709482882)*x[0]**o+(0.518382584887)*x[0]+(-0.940707425885)*x[1]**o+(0.224578726673)*x[1]
        arg[0,1]=(-0.747621141516)*x[0]**o+(-0.217452489097)*x[0]+(0.929120128008)*x[1]**o+(0.55263700209)*x[1]
        arg[0,2]=(-0.183545119672)*x[0]**o+(0.148192386895)*x[0]+(-0.22046309708)*x[1]**o+(0.433620446668)*x[1]
        arg[0,3]=(-0.0724840266959)*x[0]**o+(0.835144620602)*x[0]+(-0.962971534144)*x[1]**o+(-0.149681811534)*x[1]
        arg[0,4]=(-0.380150192228)*x[0]**o+(-0.309494603814)*x[0]+(-0.217425635182)*x[1]**o+(0.93567286482)*x[1]
        arg[1,0]=(0.0367921563947)*x[0]**o+(0.220217625576)*x[0]+(-0.120595819238)*x[1]**o+(0.308108793279)*x[1]
        arg[1,1]=(-0.142860237109)*x[0]**o+(-0.789758888253)*x[0]+(0.748085258229)*x[1]**o+(0.784767485883)*x[1]
        arg[1,2]=(-0.154690818845)*x[0]**o+(-0.678458370784)*x[0]+(0.750742996193)*x[1]**o+(-0.350199095531)*x[1]
        arg[1,3]=(-0.893295215013)*x[0]**o+(0.0295805814985)*x[0]+(0.265893730887)*x[1]**o+(0.435487860418)*x[1]
        arg[1,4]=(-0.302701704727)*x[0]**o+(0.618274652266)*x[0]+(-0.688386407308)*x[1]**o+(-0.832515064466)*x[1]
        arg[2,0]=(-0.926658745469)*x[0]**o+(0.548174350496)*x[0]+(-0.350700366641)*x[1]**o+(0.530188602011)*x[1]
        arg[2,1]=(0.900965385167)*x[0]**o+(0.0234051845124)*x[0]+(-0.00783834941699)*x[1]**o+(-0.766352156598)*x[1]
        arg[2,2]=(0.694733282444)*x[0]**o+(-0.0644519234093)*x[0]+(0.0493475528391)*x[1]**o+(-0.589101546403)*x[1]
        arg[2,3]=(-0.100333102737)*x[0]**o+(-0.633740106784)*x[0]+(0.408697826034)*x[1]**o+(-0.429059439472)*x[1]
        arg[2,4]=(0.635563324404)*x[0]**o+(0.955881346028)*x[0]+(0.0755013269452)*x[1]**o+(-0.743681282337)*x[1]
        arg[3,0]=(0.604891134834)*x[0]**o+(0.942047409702)*x[0]+(-0.00544373299112)*x[1]**o+(0.447523533166)*x[1]
        arg[3,1]=(-0.980967446837)*x[0]**o+(0.019824933239)*x[0]+(-0.826876930392)*x[1]**o+(0.113548968693)*x[1]
        arg[3,2]=(-0.731417395373)*x[0]**o+(-0.275072472108)*x[0]+(0.933867711492)*x[1]**o+(-0.444847972819)*x[1]
        arg[3,3]=(0.00121172413626)*x[0]**o+(-0.171073908218)*x[0]+(0.913182873344)*x[1]**o+(0.960256570926)*x[1]
        arg[3,4]=(-0.395134372727)*x[0]**o+(0.646385082409)*x[0]+(0.17387019219)*x[1]**o+(-0.287550462397)*x[1]
        ref[0,0]=(0.490709482882)*x_ref[0]**o+(0.518382584887)*x_ref[0]+(-0.940707425885)*x_ref[1]**o+(0.224578726673)*x_ref[1]
        ref[0,1]=(-0.747621141516)*x_ref[0]**o+(-0.217452489097)*x_ref[0]+(0.929120128008)*x_ref[1]**o+(0.55263700209)*x_ref[1]
        ref[0,2]=(-0.183545119672)*x_ref[0]**o+(0.148192386895)*x_ref[0]+(-0.22046309708)*x_ref[1]**o+(0.433620446668)*x_ref[1]
        ref[0,3]=(-0.0724840266959)*x_ref[0]**o+(0.835144620602)*x_ref[0]+(-0.962971534144)*x_ref[1]**o+(-0.149681811534)*x_ref[1]
        ref[0,4]=(-0.380150192228)*x_ref[0]**o+(-0.309494603814)*x_ref[0]+(-0.217425635182)*x_ref[1]**o+(0.93567286482)*x_ref[1]
        ref[1,0]=(0.0367921563947)*x_ref[0]**o+(0.220217625576)*x_ref[0]+(-0.120595819238)*x_ref[1]**o+(0.308108793279)*x_ref[1]
        ref[1,1]=(-0.142860237109)*x_ref[0]**o+(-0.789758888253)*x_ref[0]+(0.748085258229)*x_ref[1]**o+(0.784767485883)*x_ref[1]
        ref[1,2]=(-0.154690818845)*x_ref[0]**o+(-0.678458370784)*x_ref[0]+(0.750742996193)*x_ref[1]**o+(-0.350199095531)*x_ref[1]
        ref[1,3]=(-0.893295215013)*x_ref[0]**o+(0.0295805814985)*x_ref[0]+(0.265893730887)*x_ref[1]**o+(0.435487860418)*x_ref[1]
        ref[1,4]=(-0.302701704727)*x_ref[0]**o+(0.618274652266)*x_ref[0]+(-0.688386407308)*x_ref[1]**o+(-0.832515064466)*x_ref[1]
        ref[2,0]=(-0.926658745469)*x_ref[0]**o+(0.548174350496)*x_ref[0]+(-0.350700366641)*x_ref[1]**o+(0.530188602011)*x_ref[1]
        ref[2,1]=(0.900965385167)*x_ref[0]**o+(0.0234051845124)*x_ref[0]+(-0.00783834941699)*x_ref[1]**o+(-0.766352156598)*x_ref[1]
        ref[2,2]=(0.694733282444)*x_ref[0]**o+(-0.0644519234093)*x_ref[0]+(0.0493475528391)*x_ref[1]**o+(-0.589101546403)*x_ref[1]
        ref[2,3]=(-0.100333102737)*x_ref[0]**o+(-0.633740106784)*x_ref[0]+(0.408697826034)*x_ref[1]**o+(-0.429059439472)*x_ref[1]
        ref[2,4]=(0.635563324404)*x_ref[0]**o+(0.955881346028)*x_ref[0]+(0.0755013269452)*x_ref[1]**o+(-0.743681282337)*x_ref[1]
        ref[3,0]=(0.604891134834)*x_ref[0]**o+(0.942047409702)*x_ref[0]+(-0.00544373299112)*x_ref[1]**o+(0.447523533166)*x_ref[1]
        ref[3,1]=(-0.980967446837)*x_ref[0]**o+(0.019824933239)*x_ref[0]+(-0.826876930392)*x_ref[1]**o+(0.113548968693)*x_ref[1]
        ref[3,2]=(-0.731417395373)*x_ref[0]**o+(-0.275072472108)*x_ref[0]+(0.933867711492)*x_ref[1]**o+(-0.444847972819)*x_ref[1]
        ref[3,3]=(0.00121172413626)*x_ref[0]**o+(-0.171073908218)*x_ref[0]+(0.913182873344)*x_ref[1]**o+(0.960256570926)*x_ref[1]
        ref[3,4]=(-0.395134372727)*x_ref[0]**o+(0.646385082409)*x_ref[0]+(0.17387019219)*x_ref[1]**o+(-0.287550462397)*x_ref[1]
      else:
        arg[0,0]=(0.386545340192)*x[0]**o+(-0.397636296285)*x[0]+(-0.849437816516)*x[1]**o+(-0.83083406493)*x[1]+(0.479216612031)*x[2]**o+(0.0903787541308)*x[2]
        arg[0,1]=(-0.331651855726)*x[0]**o+(0.849421626918)*x[0]+(0.528368477362)*x[1]**o+(-0.015122214824)*x[1]+(0.494310578452)*x[2]**o+(0.190239928161)*x[2]
        arg[0,2]=(-0.662389759646)*x[0]**o+(0.204378343962)*x[0]+(0.486802744897)*x[1]**o+(0.822174784922)*x[1]+(-0.309203812647)*x[2]**o+(-0.101665040799)*x[2]
        arg[0,3]=(0.609106341828)*x[0]**o+(0.658404088333)*x[0]+(-0.0092991120732)*x[1]**o+(-0.73794934348)*x[1]+(0.270245135914)*x[2]**o+(-0.594067838407)*x[2]
        arg[0,4]=(-0.291752213357)*x[0]**o+(-0.596751343066)*x[0]+(-0.109196903078)*x[1]**o+(-0.868610307708)*x[1]+(0.622541308544)*x[2]**o+(0.482202762224)*x[2]
        arg[1,0]=(-0.113988933227)*x[0]**o+(-0.846328804695)*x[0]+(0.691769553626)*x[1]**o+(-0.377441645148)*x[1]+(-0.769857970682)*x[2]**o+(-0.583083729037)*x[2]
        arg[1,1]=(-0.635333757145)*x[0]**o+(0.194761117849)*x[0]+(-0.268287296731)*x[1]**o+(0.463906930647)*x[1]+(0.768293821029)*x[2]**o+(-0.532824906687)*x[2]
        arg[1,2]=(0.355380090138)*x[0]**o+(-0.181191575848)*x[0]+(-0.558667766467)*x[1]**o+(0.561707671872)*x[1]+(-0.427004232725)*x[2]**o+(-0.996564299989)*x[2]
        arg[1,3]=(-0.891638892893)*x[0]**o+(0.0867350850512)*x[0]+(0.880258390584)*x[1]**o+(-0.798123407525)*x[1]+(-0.571505202201)*x[2]**o+(0.35374245063)*x[2]
        arg[1,4]=(0.543632496941)*x[0]**o+(-0.630363803582)*x[0]+(0.38107312058)*x[1]**o+(0.222401831883)*x[1]+(-0.932212080175)*x[2]**o+(-0.0482899826174)*x[2]
        arg[2,0]=(0.46514308325)*x[0]**o+(-0.907464652384)*x[0]+(-0.706105595148)*x[1]**o+(0.486989958675)*x[1]+(0.422702683485)*x[2]**o+(0.640916925932)*x[2]
        arg[2,1]=(0.912808317479)*x[0]**o+(-0.860779312763)*x[0]+(-0.915364956133)*x[1]**o+(-0.314186117224)*x[1]+(-0.860814410035)*x[2]**o+(0.837570442722)*x[2]
        arg[2,2]=(0.00932917919464)*x[0]**o+(0.393773283049)*x[0]+(0.699998572668)*x[1]**o+(0.153407885795)*x[1]+(-0.166245189281)*x[2]**o+(-0.430541214236)*x[2]
        arg[2,3]=(-0.00127511229816)*x[0]**o+(-0.327443174924)*x[0]+(0.61881322537)*x[1]**o+(-0.725854102962)*x[1]+(0.650649062427)*x[2]**o+(0.246373184041)*x[2]
        arg[2,4]=(-0.530580898049)*x[0]**o+(-0.208506132901)*x[0]+(-0.102049332015)*x[1]**o+(-0.537037644189)*x[1]+(0.740113470729)*x[2]**o+(0.12094405264)*x[2]
        arg[3,0]=(-0.103423602042)*x[0]**o+(-0.535742323666)*x[0]+(0.678663019405)*x[1]**o+(0.579375950057)*x[1]+(0.323627490052)*x[2]**o+(-0.681323301356)*x[2]
        arg[3,1]=(0.670239087519)*x[0]**o+(0.209657968128)*x[0]+(-0.357490300727)*x[1]**o+(-0.441776782321)*x[1]+(0.00927389037562)*x[2]**o+(0.633527599723)*x[2]
        arg[3,2]=(0.415569606138)*x[0]**o+(-0.422211616068)*x[0]+(-0.695086235645)*x[1]**o+(0.709158289848)*x[1]+(-0.523450845183)*x[2]**o+(0.555778994398)*x[2]
        arg[3,3]=(0.0188993892075)*x[0]**o+(0.611088737828)*x[0]+(0.0313795167921)*x[1]**o+(0.0842651755016)*x[1]+(0.288414199415)*x[2]**o+(0.695910780235)*x[2]
        arg[3,4]=(-0.500338800879)*x[0]**o+(0.0984192369328)*x[0]+(0.71163172992)*x[1]**o+(-0.102123116139)*x[1]+(0.954613955678)*x[2]**o+(0.384872409422)*x[2]
        ref[0,0]=(0.386545340192)*x_ref[0]**o+(-0.397636296285)*x_ref[0]+(-0.849437816516)*x_ref[1]**o+(-0.83083406493)*x_ref[1]+(0.479216612031)*x_ref[2]**o+(0.0903787541308)*x_ref[2]
        ref[0,1]=(-0.331651855726)*x_ref[0]**o+(0.849421626918)*x_ref[0]+(0.528368477362)*x_ref[1]**o+(-0.015122214824)*x_ref[1]+(0.494310578452)*x_ref[2]**o+(0.190239928161)*x_ref[2]
        ref[0,2]=(-0.662389759646)*x_ref[0]**o+(0.204378343962)*x_ref[0]+(0.486802744897)*x_ref[1]**o+(0.822174784922)*x_ref[1]+(-0.309203812647)*x_ref[2]**o+(-0.101665040799)*x_ref[2]
        ref[0,3]=(0.609106341828)*x_ref[0]**o+(0.658404088333)*x_ref[0]+(-0.0092991120732)*x_ref[1]**o+(-0.73794934348)*x_ref[1]+(0.270245135914)*x_ref[2]**o+(-0.594067838407)*x_ref[2]
        ref[0,4]=(-0.291752213357)*x_ref[0]**o+(-0.596751343066)*x_ref[0]+(-0.109196903078)*x_ref[1]**o+(-0.868610307708)*x_ref[1]+(0.622541308544)*x_ref[2]**o+(0.482202762224)*x_ref[2]
        ref[1,0]=(-0.113988933227)*x_ref[0]**o+(-0.846328804695)*x_ref[0]+(0.691769553626)*x_ref[1]**o+(-0.377441645148)*x_ref[1]+(-0.769857970682)*x_ref[2]**o+(-0.583083729037)*x_ref[2]
        ref[1,1]=(-0.635333757145)*x_ref[0]**o+(0.194761117849)*x_ref[0]+(-0.268287296731)*x_ref[1]**o+(0.463906930647)*x_ref[1]+(0.768293821029)*x_ref[2]**o+(-0.532824906687)*x_ref[2]
        ref[1,2]=(0.355380090138)*x_ref[0]**o+(-0.181191575848)*x_ref[0]+(-0.558667766467)*x_ref[1]**o+(0.561707671872)*x_ref[1]+(-0.427004232725)*x_ref[2]**o+(-0.996564299989)*x_ref[2]
        ref[1,3]=(-0.891638892893)*x_ref[0]**o+(0.0867350850512)*x_ref[0]+(0.880258390584)*x_ref[1]**o+(-0.798123407525)*x_ref[1]+(-0.571505202201)*x_ref[2]**o+(0.35374245063)*x_ref[2]
        ref[1,4]=(0.543632496941)*x_ref[0]**o+(-0.630363803582)*x_ref[0]+(0.38107312058)*x_ref[1]**o+(0.222401831883)*x_ref[1]+(-0.932212080175)*x_ref[2]**o+(-0.0482899826174)*x_ref[2]
        ref[2,0]=(0.46514308325)*x_ref[0]**o+(-0.907464652384)*x_ref[0]+(-0.706105595148)*x_ref[1]**o+(0.486989958675)*x_ref[1]+(0.422702683485)*x_ref[2]**o+(0.640916925932)*x_ref[2]
        ref[2,1]=(0.912808317479)*x_ref[0]**o+(-0.860779312763)*x_ref[0]+(-0.915364956133)*x_ref[1]**o+(-0.314186117224)*x_ref[1]+(-0.860814410035)*x_ref[2]**o+(0.837570442722)*x_ref[2]
        ref[2,2]=(0.00932917919464)*x_ref[0]**o+(0.393773283049)*x_ref[0]+(0.699998572668)*x_ref[1]**o+(0.153407885795)*x_ref[1]+(-0.166245189281)*x_ref[2]**o+(-0.430541214236)*x_ref[2]
        ref[2,3]=(-0.00127511229816)*x_ref[0]**o+(-0.327443174924)*x_ref[0]+(0.61881322537)*x_ref[1]**o+(-0.725854102962)*x_ref[1]+(0.650649062427)*x_ref[2]**o+(0.246373184041)*x_ref[2]
        ref[2,4]=(-0.530580898049)*x_ref[0]**o+(-0.208506132901)*x_ref[0]+(-0.102049332015)*x_ref[1]**o+(-0.537037644189)*x_ref[1]+(0.740113470729)*x_ref[2]**o+(0.12094405264)*x_ref[2]
        ref[3,0]=(-0.103423602042)*x_ref[0]**o+(-0.535742323666)*x_ref[0]+(0.678663019405)*x_ref[1]**o+(0.579375950057)*x_ref[1]+(0.323627490052)*x_ref[2]**o+(-0.681323301356)*x_ref[2]
        ref[3,1]=(0.670239087519)*x_ref[0]**o+(0.209657968128)*x_ref[0]+(-0.357490300727)*x_ref[1]**o+(-0.441776782321)*x_ref[1]+(0.00927389037562)*x_ref[2]**o+(0.633527599723)*x_ref[2]
        ref[3,2]=(0.415569606138)*x_ref[0]**o+(-0.422211616068)*x_ref[0]+(-0.695086235645)*x_ref[1]**o+(0.709158289848)*x_ref[1]+(-0.523450845183)*x_ref[2]**o+(0.555778994398)*x_ref[2]
        ref[3,3]=(0.0188993892075)*x_ref[0]**o+(0.611088737828)*x_ref[0]+(0.0313795167921)*x_ref[1]**o+(0.0842651755016)*x_ref[1]+(0.288414199415)*x_ref[2]**o+(0.695910780235)*x_ref[2]
        ref[3,4]=(-0.500338800879)*x_ref[0]**o+(0.0984192369328)*x_ref[0]+(0.71163172992)*x_ref[1]**o+(-0.102123116139)*x_ref[1]+(0.954613955678)*x_ref[2]**o+(0.384872409422)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.317274364174)*x[0]**o+(-0.565307843766)*x[0]+(-0.367874638208)*x[1]**o+(0.0029581033546)*x[1]
        arg[0,0,1]=(0.832728755858)*x[0]**o+(0.245271193641)*x[0]+(-0.727130029377)*x[1]**o+(0.137912975829)*x[1]
        arg[0,1,0]=(-0.634508968932)*x[0]**o+(-0.864200861637)*x[0]+(-0.566985971867)*x[1]**o+(0.458187145519)*x[1]
        arg[0,1,1]=(0.133772683885)*x[0]**o+(-0.630800341215)*x[0]+(0.958934971335)*x[1]**o+(-0.15839182257)*x[1]
        arg[1,0,0]=(-0.0948724752087)*x[0]**o+(-0.25589337469)*x[0]+(-0.315713312867)*x[1]**o+(0.223487524432)*x[1]
        arg[1,0,1]=(0.794634271698)*x[0]**o+(-0.505797079712)*x[0]+(0.54367349338)*x[1]**o+(0.610149581933)*x[1]
        arg[1,1,0]=(-0.447039667341)*x[0]**o+(-0.917009838715)*x[0]+(0.525331113522)*x[1]**o+(-0.626176620973)*x[1]
        arg[1,1,1]=(0.0749028334839)*x[0]**o+(0.106950341725)*x[0]+(-0.375321288987)*x[1]**o+(0.510260762626)*x[1]
        arg[2,0,0]=(-0.286661970669)*x[0]**o+(0.584169107201)*x[0]+(0.189852415568)*x[1]**o+(-0.685695910848)*x[1]
        arg[2,0,1]=(-0.689509796724)*x[0]**o+(-0.316067007053)*x[0]+(0.93758916753)*x[1]**o+(0.987368705364)*x[1]
        arg[2,1,0]=(-0.411695937554)*x[0]**o+(0.858528628791)*x[0]+(-0.579361651418)*x[1]**o+(-0.179617869494)*x[1]
        arg[2,1,1]=(0.871441175049)*x[0]**o+(0.445176827889)*x[0]+(0.282613647019)*x[1]**o+(-0.0414230375843)*x[1]
        arg[3,0,0]=(-0.242136375691)*x[0]**o+(-0.91917484201)*x[0]+(-0.88026838607)*x[1]**o+(0.363309786006)*x[1]
        arg[3,0,1]=(-0.991221762386)*x[0]**o+(-0.0149280738883)*x[0]+(0.420322977563)*x[1]**o+(-0.0533076687981)*x[1]
        arg[3,1,0]=(-0.889003524303)*x[0]**o+(-0.023519502168)*x[0]+(0.826015855889)*x[1]**o+(0.0568467842617)*x[1]
        arg[3,1,1]=(0.608567002824)*x[0]**o+(-0.531288593441)*x[0]+(0.0886678034719)*x[1]**o+(-0.732353220319)*x[1]
        arg[4,0,0]=(-0.0537755786538)*x[0]**o+(-0.475886396481)*x[0]+(-0.153787571145)*x[1]**o+(0.689850737569)*x[1]
        arg[4,0,1]=(0.413017674549)*x[0]**o+(-0.993353261494)*x[0]+(-0.944597571268)*x[1]**o+(0.566423551758)*x[1]
        arg[4,1,0]=(-0.716864417268)*x[0]**o+(-0.871574272865)*x[0]+(0.394071572628)*x[1]**o+(0.452903048597)*x[1]
        arg[4,1,1]=(0.861790785883)*x[0]**o+(-0.511789401943)*x[0]+(-0.272452527007)*x[1]**o+(-0.379021521619)*x[1]
        arg[5,0,0]=(0.509668147054)*x[0]**o+(0.919355584411)*x[0]+(-0.673215166917)*x[1]**o+(0.766459631931)*x[1]
        arg[5,0,1]=(0.177006227916)*x[0]**o+(-0.275118477)*x[0]+(0.0302367843701)*x[1]**o+(-0.761226360328)*x[1]
        arg[5,1,0]=(0.013432534638)*x[0]**o+(-0.911626567027)*x[0]+(-0.710520442529)*x[1]**o+(-0.882317448423)*x[1]
        arg[5,1,1]=(0.595583093615)*x[0]**o+(0.163527905156)*x[0]+(-0.122165051333)*x[1]**o+(-0.526359288181)*x[1]
        ref[0,0,0]=(0.317274364174)*x_ref[0]**o+(-0.565307843766)*x_ref[0]+(-0.367874638208)*x_ref[1]**o+(0.0029581033546)*x_ref[1]
        ref[0,0,1]=(0.832728755858)*x_ref[0]**o+(0.245271193641)*x_ref[0]+(-0.727130029377)*x_ref[1]**o+(0.137912975829)*x_ref[1]
        ref[0,1,0]=(-0.634508968932)*x_ref[0]**o+(-0.864200861637)*x_ref[0]+(-0.566985971867)*x_ref[1]**o+(0.458187145519)*x_ref[1]
        ref[0,1,1]=(0.133772683885)*x_ref[0]**o+(-0.630800341215)*x_ref[0]+(0.958934971335)*x_ref[1]**o+(-0.15839182257)*x_ref[1]
        ref[1,0,0]=(-0.0948724752087)*x_ref[0]**o+(-0.25589337469)*x_ref[0]+(-0.315713312867)*x_ref[1]**o+(0.223487524432)*x_ref[1]
        ref[1,0,1]=(0.794634271698)*x_ref[0]**o+(-0.505797079712)*x_ref[0]+(0.54367349338)*x_ref[1]**o+(0.610149581933)*x_ref[1]
        ref[1,1,0]=(-0.447039667341)*x_ref[0]**o+(-0.917009838715)*x_ref[0]+(0.525331113522)*x_ref[1]**o+(-0.626176620973)*x_ref[1]
        ref[1,1,1]=(0.0749028334839)*x_ref[0]**o+(0.106950341725)*x_ref[0]+(-0.375321288987)*x_ref[1]**o+(0.510260762626)*x_ref[1]
        ref[2,0,0]=(-0.286661970669)*x_ref[0]**o+(0.584169107201)*x_ref[0]+(0.189852415568)*x_ref[1]**o+(-0.685695910848)*x_ref[1]
        ref[2,0,1]=(-0.689509796724)*x_ref[0]**o+(-0.316067007053)*x_ref[0]+(0.93758916753)*x_ref[1]**o+(0.987368705364)*x_ref[1]
        ref[2,1,0]=(-0.411695937554)*x_ref[0]**o+(0.858528628791)*x_ref[0]+(-0.579361651418)*x_ref[1]**o+(-0.179617869494)*x_ref[1]
        ref[2,1,1]=(0.871441175049)*x_ref[0]**o+(0.445176827889)*x_ref[0]+(0.282613647019)*x_ref[1]**o+(-0.0414230375843)*x_ref[1]
        ref[3,0,0]=(-0.242136375691)*x_ref[0]**o+(-0.91917484201)*x_ref[0]+(-0.88026838607)*x_ref[1]**o+(0.363309786006)*x_ref[1]
        ref[3,0,1]=(-0.991221762386)*x_ref[0]**o+(-0.0149280738883)*x_ref[0]+(0.420322977563)*x_ref[1]**o+(-0.0533076687981)*x_ref[1]
        ref[3,1,0]=(-0.889003524303)*x_ref[0]**o+(-0.023519502168)*x_ref[0]+(0.826015855889)*x_ref[1]**o+(0.0568467842617)*x_ref[1]
        ref[3,1,1]=(0.608567002824)*x_ref[0]**o+(-0.531288593441)*x_ref[0]+(0.0886678034719)*x_ref[1]**o+(-0.732353220319)*x_ref[1]
        ref[4,0,0]=(-0.0537755786538)*x_ref[0]**o+(-0.475886396481)*x_ref[0]+(-0.153787571145)*x_ref[1]**o+(0.689850737569)*x_ref[1]
        ref[4,0,1]=(0.413017674549)*x_ref[0]**o+(-0.993353261494)*x_ref[0]+(-0.944597571268)*x_ref[1]**o+(0.566423551758)*x_ref[1]
        ref[4,1,0]=(-0.716864417268)*x_ref[0]**o+(-0.871574272865)*x_ref[0]+(0.394071572628)*x_ref[1]**o+(0.452903048597)*x_ref[1]
        ref[4,1,1]=(0.861790785883)*x_ref[0]**o+(-0.511789401943)*x_ref[0]+(-0.272452527007)*x_ref[1]**o+(-0.379021521619)*x_ref[1]
        ref[5,0,0]=(0.509668147054)*x_ref[0]**o+(0.919355584411)*x_ref[0]+(-0.673215166917)*x_ref[1]**o+(0.766459631931)*x_ref[1]
        ref[5,0,1]=(0.177006227916)*x_ref[0]**o+(-0.275118477)*x_ref[0]+(0.0302367843701)*x_ref[1]**o+(-0.761226360328)*x_ref[1]
        ref[5,1,0]=(0.013432534638)*x_ref[0]**o+(-0.911626567027)*x_ref[0]+(-0.710520442529)*x_ref[1]**o+(-0.882317448423)*x_ref[1]
        ref[5,1,1]=(0.595583093615)*x_ref[0]**o+(0.163527905156)*x_ref[0]+(-0.122165051333)*x_ref[1]**o+(-0.526359288181)*x_ref[1]
      else:
        arg[0,0,0]=(0.230695674262)*x[0]**o+(-0.987212801716)*x[0]+(-0.0599001434987)*x[1]**o+(0.719591259555)*x[1]+(-0.665186369759)*x[2]**o+(-0.563552602515)*x[2]
        arg[0,0,1]=(-0.775441966131)*x[0]**o+(-0.152166849702)*x[0]+(0.626466789363)*x[1]**o+(0.179356346581)*x[1]+(-0.687406927434)*x[2]**o+(-0.563492313114)*x[2]
        arg[0,1,0]=(0.749522748133)*x[0]**o+(-0.582456308337)*x[0]+(0.676882643058)*x[1]**o+(0.365363182598)*x[1]+(-0.550867708812)*x[2]**o+(0.777938151183)*x[2]
        arg[0,1,1]=(0.389788552775)*x[0]**o+(-0.298877489997)*x[0]+(0.560326209269)*x[1]**o+(-0.113249832155)*x[1]+(0.588087510254)*x[2]**o+(-0.593503776461)*x[2]
        arg[1,0,0]=(-0.354661103909)*x[0]**o+(-0.195375369549)*x[0]+(-0.798795069895)*x[1]**o+(0.0968631048579)*x[1]+(-0.016118190391)*x[2]**o+(-0.903745407705)*x[2]
        arg[1,0,1]=(-0.383688834784)*x[0]**o+(0.861256231933)*x[0]+(-0.767204834667)*x[1]**o+(0.338875482372)*x[1]+(0.117146310891)*x[2]**o+(-0.302651846083)*x[2]
        arg[1,1,0]=(0.646263670969)*x[0]**o+(-0.447984008871)*x[0]+(-0.169322492785)*x[1]**o+(-0.570614724481)*x[1]+(-0.848518457618)*x[2]**o+(-0.0750005539216)*x[2]
        arg[1,1,1]=(-0.232496745304)*x[0]**o+(0.182581173413)*x[0]+(0.8032436357)*x[1]**o+(-0.391266449476)*x[1]+(0.491677629766)*x[2]**o+(-0.188361708708)*x[2]
        arg[2,0,0]=(-0.571145099091)*x[0]**o+(-0.486855242188)*x[0]+(0.52833582073)*x[1]**o+(0.326716348242)*x[1]+(0.528475192869)*x[2]**o+(0.163280575137)*x[2]
        arg[2,0,1]=(-0.368889674601)*x[0]**o+(0.601775816775)*x[0]+(0.0281742455403)*x[1]**o+(-0.198869716243)*x[1]+(0.193343065506)*x[2]**o+(-0.0614451571108)*x[2]
        arg[2,1,0]=(0.724595792492)*x[0]**o+(-0.206983407651)*x[0]+(-0.33966393719)*x[1]**o+(0.929465830796)*x[1]+(-0.938517195863)*x[2]**o+(0.70874077222)*x[2]
        arg[2,1,1]=(-0.536303362606)*x[0]**o+(-0.430663025812)*x[0]+(-0.0845923469062)*x[1]**o+(0.866655631144)*x[1]+(-0.2378102403)*x[2]**o+(0.969070797268)*x[2]
        arg[3,0,0]=(0.345137352559)*x[0]**o+(-0.696652518382)*x[0]+(0.914998974008)*x[1]**o+(-0.874935163156)*x[1]+(0.947812661283)*x[2]**o+(0.868611662789)*x[2]
        arg[3,0,1]=(-0.421640746358)*x[0]**o+(-0.37925991208)*x[0]+(0.468281241452)*x[1]**o+(0.439086689805)*x[1]+(-0.607986317001)*x[2]**o+(0.269488168937)*x[2]
        arg[3,1,0]=(-0.550485456473)*x[0]**o+(-0.520018714495)*x[0]+(0.439679309654)*x[1]**o+(-0.0652434084668)*x[1]+(-0.584838668708)*x[2]**o+(-0.46047281034)*x[2]
        arg[3,1,1]=(-0.421746456038)*x[0]**o+(0.370489549278)*x[0]+(-0.683682295105)*x[1]**o+(0.324051718035)*x[1]+(-0.579046430636)*x[2]**o+(0.587623604941)*x[2]
        arg[4,0,0]=(-0.167582584882)*x[0]**o+(0.524904122243)*x[0]+(0.665109348362)*x[1]**o+(-0.093655366509)*x[1]+(0.704404656254)*x[2]**o+(-0.402602441741)*x[2]
        arg[4,0,1]=(0.633854964824)*x[0]**o+(-0.836385840194)*x[0]+(-0.0995437996628)*x[1]**o+(-0.200285338869)*x[1]+(0.535023783881)*x[2]**o+(0.884702543593)*x[2]
        arg[4,1,0]=(-0.953619378208)*x[0]**o+(-0.390518815669)*x[0]+(-0.0590889178236)*x[1]**o+(-0.997225825019)*x[1]+(0.0520418736402)*x[2]**o+(0.484749465262)*x[2]
        arg[4,1,1]=(-0.440389854756)*x[0]**o+(0.15520431778)*x[0]+(0.346839174513)*x[1]**o+(0.0172676873159)*x[1]+(0.86275919224)*x[2]**o+(0.568069148992)*x[2]
        arg[5,0,0]=(-0.303687941217)*x[0]**o+(0.516115915687)*x[0]+(-0.750276597519)*x[1]**o+(0.734039058778)*x[1]+(-0.624499210252)*x[2]**o+(-0.926338135421)*x[2]
        arg[5,0,1]=(-0.472230777895)*x[0]**o+(-0.557827728373)*x[0]+(-0.403783023157)*x[1]**o+(0.70433336955)*x[1]+(-0.0211737454531)*x[2]**o+(0.0774406125836)*x[2]
        arg[5,1,0]=(0.718149875204)*x[0]**o+(-0.183000739787)*x[0]+(-0.655724252373)*x[1]**o+(-0.634941710776)*x[1]+(-0.26300163323)*x[2]**o+(-0.25105424715)*x[2]
        arg[5,1,1]=(0.109601904441)*x[0]**o+(-0.00640764631554)*x[0]+(-0.488035610009)*x[1]**o+(0.909668323769)*x[1]+(0.229692168501)*x[2]**o+(-0.739241309497)*x[2]
        ref[0,0,0]=(0.230695674262)*x_ref[0]**o+(-0.987212801716)*x_ref[0]+(-0.0599001434987)*x_ref[1]**o+(0.719591259555)*x_ref[1]+(-0.665186369759)*x_ref[2]**o+(-0.563552602515)*x_ref[2]
        ref[0,0,1]=(-0.775441966131)*x_ref[0]**o+(-0.152166849702)*x_ref[0]+(0.626466789363)*x_ref[1]**o+(0.179356346581)*x_ref[1]+(-0.687406927434)*x_ref[2]**o+(-0.563492313114)*x_ref[2]
        ref[0,1,0]=(0.749522748133)*x_ref[0]**o+(-0.582456308337)*x_ref[0]+(0.676882643058)*x_ref[1]**o+(0.365363182598)*x_ref[1]+(-0.550867708812)*x_ref[2]**o+(0.777938151183)*x_ref[2]
        ref[0,1,1]=(0.389788552775)*x_ref[0]**o+(-0.298877489997)*x_ref[0]+(0.560326209269)*x_ref[1]**o+(-0.113249832155)*x_ref[1]+(0.588087510254)*x_ref[2]**o+(-0.593503776461)*x_ref[2]
        ref[1,0,0]=(-0.354661103909)*x_ref[0]**o+(-0.195375369549)*x_ref[0]+(-0.798795069895)*x_ref[1]**o+(0.0968631048579)*x_ref[1]+(-0.016118190391)*x_ref[2]**o+(-0.903745407705)*x_ref[2]
        ref[1,0,1]=(-0.383688834784)*x_ref[0]**o+(0.861256231933)*x_ref[0]+(-0.767204834667)*x_ref[1]**o+(0.338875482372)*x_ref[1]+(0.117146310891)*x_ref[2]**o+(-0.302651846083)*x_ref[2]
        ref[1,1,0]=(0.646263670969)*x_ref[0]**o+(-0.447984008871)*x_ref[0]+(-0.169322492785)*x_ref[1]**o+(-0.570614724481)*x_ref[1]+(-0.848518457618)*x_ref[2]**o+(-0.0750005539216)*x_ref[2]
        ref[1,1,1]=(-0.232496745304)*x_ref[0]**o+(0.182581173413)*x_ref[0]+(0.8032436357)*x_ref[1]**o+(-0.391266449476)*x_ref[1]+(0.491677629766)*x_ref[2]**o+(-0.188361708708)*x_ref[2]
        ref[2,0,0]=(-0.571145099091)*x_ref[0]**o+(-0.486855242188)*x_ref[0]+(0.52833582073)*x_ref[1]**o+(0.326716348242)*x_ref[1]+(0.528475192869)*x_ref[2]**o+(0.163280575137)*x_ref[2]
        ref[2,0,1]=(-0.368889674601)*x_ref[0]**o+(0.601775816775)*x_ref[0]+(0.0281742455403)*x_ref[1]**o+(-0.198869716243)*x_ref[1]+(0.193343065506)*x_ref[2]**o+(-0.0614451571108)*x_ref[2]
        ref[2,1,0]=(0.724595792492)*x_ref[0]**o+(-0.206983407651)*x_ref[0]+(-0.33966393719)*x_ref[1]**o+(0.929465830796)*x_ref[1]+(-0.938517195863)*x_ref[2]**o+(0.70874077222)*x_ref[2]
        ref[2,1,1]=(-0.536303362606)*x_ref[0]**o+(-0.430663025812)*x_ref[0]+(-0.0845923469062)*x_ref[1]**o+(0.866655631144)*x_ref[1]+(-0.2378102403)*x_ref[2]**o+(0.969070797268)*x_ref[2]
        ref[3,0,0]=(0.345137352559)*x_ref[0]**o+(-0.696652518382)*x_ref[0]+(0.914998974008)*x_ref[1]**o+(-0.874935163156)*x_ref[1]+(0.947812661283)*x_ref[2]**o+(0.868611662789)*x_ref[2]
        ref[3,0,1]=(-0.421640746358)*x_ref[0]**o+(-0.37925991208)*x_ref[0]+(0.468281241452)*x_ref[1]**o+(0.439086689805)*x_ref[1]+(-0.607986317001)*x_ref[2]**o+(0.269488168937)*x_ref[2]
        ref[3,1,0]=(-0.550485456473)*x_ref[0]**o+(-0.520018714495)*x_ref[0]+(0.439679309654)*x_ref[1]**o+(-0.0652434084668)*x_ref[1]+(-0.584838668708)*x_ref[2]**o+(-0.46047281034)*x_ref[2]
        ref[3,1,1]=(-0.421746456038)*x_ref[0]**o+(0.370489549278)*x_ref[0]+(-0.683682295105)*x_ref[1]**o+(0.324051718035)*x_ref[1]+(-0.579046430636)*x_ref[2]**o+(0.587623604941)*x_ref[2]
        ref[4,0,0]=(-0.167582584882)*x_ref[0]**o+(0.524904122243)*x_ref[0]+(0.665109348362)*x_ref[1]**o+(-0.093655366509)*x_ref[1]+(0.704404656254)*x_ref[2]**o+(-0.402602441741)*x_ref[2]
        ref[4,0,1]=(0.633854964824)*x_ref[0]**o+(-0.836385840194)*x_ref[0]+(-0.0995437996628)*x_ref[1]**o+(-0.200285338869)*x_ref[1]+(0.535023783881)*x_ref[2]**o+(0.884702543593)*x_ref[2]
        ref[4,1,0]=(-0.953619378208)*x_ref[0]**o+(-0.390518815669)*x_ref[0]+(-0.0590889178236)*x_ref[1]**o+(-0.997225825019)*x_ref[1]+(0.0520418736402)*x_ref[2]**o+(0.484749465262)*x_ref[2]
        ref[4,1,1]=(-0.440389854756)*x_ref[0]**o+(0.15520431778)*x_ref[0]+(0.346839174513)*x_ref[1]**o+(0.0172676873159)*x_ref[1]+(0.86275919224)*x_ref[2]**o+(0.568069148992)*x_ref[2]
        ref[5,0,0]=(-0.303687941217)*x_ref[0]**o+(0.516115915687)*x_ref[0]+(-0.750276597519)*x_ref[1]**o+(0.734039058778)*x_ref[1]+(-0.624499210252)*x_ref[2]**o+(-0.926338135421)*x_ref[2]
        ref[5,0,1]=(-0.472230777895)*x_ref[0]**o+(-0.557827728373)*x_ref[0]+(-0.403783023157)*x_ref[1]**o+(0.70433336955)*x_ref[1]+(-0.0211737454531)*x_ref[2]**o+(0.0774406125836)*x_ref[2]
        ref[5,1,0]=(0.718149875204)*x_ref[0]**o+(-0.183000739787)*x_ref[0]+(-0.655724252373)*x_ref[1]**o+(-0.634941710776)*x_ref[1]+(-0.26300163323)*x_ref[2]**o+(-0.25105424715)*x_ref[2]
        ref[5,1,1]=(0.109601904441)*x_ref[0]**o+(-0.00640764631554)*x_ref[0]+(-0.488035610009)*x_ref[1]**o+(0.909668323769)*x_ref[1]+(0.229692168501)*x_ref[2]**o+(-0.739241309497)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.912518456494)*x[0]**o+(-0.828176151903)*x[0]+(-0.313940298121)*x[1]**o+(-0.850120336531)*x[1]
        arg[0,0,0,1]=(0.14415852783)*x[0]**o+(-0.294155246668)*x[0]+(0.562517248232)*x[1]**o+(-0.204505938286)*x[1]
        arg[0,0,1,0]=(0.555455404368)*x[0]**o+(0.916425343576)*x[0]+(0.547252315693)*x[1]**o+(-0.0992842367479)*x[1]
        arg[0,0,1,1]=(0.603632473928)*x[0]**o+(0.505571549062)*x[0]+(-0.490402921542)*x[1]**o+(0.908624427316)*x[1]
        arg[0,0,2,0]=(0.729542308985)*x[0]**o+(-0.654310868638)*x[0]+(0.0845981276245)*x[1]**o+(-0.95231604686)*x[1]
        arg[0,0,2,1]=(0.0511615606745)*x[0]**o+(-0.939469887975)*x[0]+(-0.206229418155)*x[1]**o+(0.254879844701)*x[1]
        arg[0,1,0,0]=(0.257917344507)*x[0]**o+(0.860729621288)*x[0]+(0.489966193348)*x[1]**o+(0.174897233858)*x[1]
        arg[0,1,0,1]=(0.336588362051)*x[0]**o+(-0.690598635114)*x[0]+(-0.0913735300483)*x[1]**o+(-0.932914439004)*x[1]
        arg[0,1,1,0]=(0.244583093108)*x[0]**o+(-0.827768167758)*x[0]+(0.573065359412)*x[1]**o+(0.176685515596)*x[1]
        arg[0,1,1,1]=(-0.520550726127)*x[0]**o+(-0.605243177286)*x[0]+(0.130549556226)*x[1]**o+(0.750019884067)*x[1]
        arg[0,1,2,0]=(-0.680201627614)*x[0]**o+(0.679925324882)*x[0]+(-0.060408604263)*x[1]**o+(0.568125689359)*x[1]
        arg[0,1,2,1]=(-0.111908859432)*x[0]**o+(0.256521037018)*x[0]+(0.332050616611)*x[1]**o+(-0.0500016397683)*x[1]
        arg[0,2,0,0]=(-0.484953933663)*x[0]**o+(0.628032211831)*x[0]+(-0.63297141454)*x[1]**o+(0.141016662755)*x[1]
        arg[0,2,0,1]=(-0.463920680465)*x[0]**o+(-0.868825520511)*x[0]+(-0.36594425006)*x[1]**o+(-0.823052216578)*x[1]
        arg[0,2,1,0]=(0.972337802357)*x[0]**o+(0.943605157343)*x[0]+(-0.113950717844)*x[1]**o+(-0.839596452897)*x[1]
        arg[0,2,1,1]=(-0.295562401302)*x[0]**o+(0.735770621017)*x[0]+(-0.183838912348)*x[1]**o+(-0.929365380252)*x[1]
        arg[0,2,2,0]=(0.912886298388)*x[0]**o+(0.847507982065)*x[0]+(-0.341306036323)*x[1]**o+(0.393627025978)*x[1]
        arg[0,2,2,1]=(-0.769650638386)*x[0]**o+(-0.619672356488)*x[0]+(-0.126894288906)*x[1]**o+(0.231547632139)*x[1]
        arg[0,3,0,0]=(-0.884442976834)*x[0]**o+(-0.0194505760676)*x[0]+(-0.338764675072)*x[1]**o+(0.723481600171)*x[1]
        arg[0,3,0,1]=(0.655973555091)*x[0]**o+(-0.556230649412)*x[0]+(-0.850356832441)*x[1]**o+(0.115267314853)*x[1]
        arg[0,3,1,0]=(0.659628443559)*x[0]**o+(-0.586677822491)*x[0]+(0.135930588518)*x[1]**o+(0.206866895948)*x[1]
        arg[0,3,1,1]=(-0.700258957557)*x[0]**o+(-0.699774198764)*x[0]+(-0.294009479186)*x[1]**o+(0.848159666503)*x[1]
        arg[0,3,2,0]=(-0.927501754488)*x[0]**o+(-0.981178071331)*x[0]+(0.614074194451)*x[1]**o+(0.711549573086)*x[1]
        arg[0,3,2,1]=(0.312434473198)*x[0]**o+(-0.241506125991)*x[0]+(-0.0840979088912)*x[1]**o+(0.902948463117)*x[1]
        arg[0,4,0,0]=(-0.892269860242)*x[0]**o+(0.823434435855)*x[0]+(0.50499383365)*x[1]**o+(0.763736541034)*x[1]
        arg[0,4,0,1]=(-0.606959691191)*x[0]**o+(0.945562536791)*x[0]+(-0.46271929481)*x[1]**o+(0.929514432096)*x[1]
        arg[0,4,1,0]=(0.953382685939)*x[0]**o+(0.553766359805)*x[0]+(-0.191004010293)*x[1]**o+(0.556369491698)*x[1]
        arg[0,4,1,1]=(-0.269402649165)*x[0]**o+(-0.158926232076)*x[0]+(-0.390038749681)*x[1]**o+(0.553655192666)*x[1]
        arg[0,4,2,0]=(0.0785586372006)*x[0]**o+(-0.778452221862)*x[0]+(-0.927627273292)*x[1]**o+(-0.187742405246)*x[1]
        arg[0,4,2,1]=(-0.965329945447)*x[0]**o+(0.304539006089)*x[0]+(0.213031689568)*x[1]**o+(-0.2807790149)*x[1]
        arg[1,0,0,0]=(0.981534698569)*x[0]**o+(0.466451685806)*x[0]+(0.845322931856)*x[1]**o+(0.582422032334)*x[1]
        arg[1,0,0,1]=(-0.0645798740394)*x[0]**o+(0.0475214206619)*x[0]+(-0.687759433076)*x[1]**o+(0.403440938285)*x[1]
        arg[1,0,1,0]=(-0.317887773392)*x[0]**o+(0.59101163416)*x[0]+(0.658626263291)*x[1]**o+(-0.618167594137)*x[1]
        arg[1,0,1,1]=(-0.314093092813)*x[0]**o+(0.699933899942)*x[0]+(0.824195260822)*x[1]**o+(-0.42875327345)*x[1]
        arg[1,0,2,0]=(-0.254254956887)*x[0]**o+(-0.912137080163)*x[0]+(0.360457121452)*x[1]**o+(-0.902596912588)*x[1]
        arg[1,0,2,1]=(-0.577214593683)*x[0]**o+(-0.419495663008)*x[0]+(-0.803459783331)*x[1]**o+(0.490885513568)*x[1]
        arg[1,1,0,0]=(0.477947766892)*x[0]**o+(-0.853046971581)*x[0]+(0.576044041601)*x[1]**o+(-0.447007701252)*x[1]
        arg[1,1,0,1]=(0.5886933409)*x[0]**o+(0.253650778351)*x[0]+(-0.198070684431)*x[1]**o+(0.189479364528)*x[1]
        arg[1,1,1,0]=(-0.528744478872)*x[0]**o+(-0.504072059557)*x[0]+(-0.0161832474015)*x[1]**o+(0.639592055966)*x[1]
        arg[1,1,1,1]=(-0.581844827169)*x[0]**o+(0.997883224538)*x[0]+(0.9354364489)*x[1]**o+(0.884060785136)*x[1]
        arg[1,1,2,0]=(0.858226955817)*x[0]**o+(0.707261698312)*x[0]+(-0.738309111076)*x[1]**o+(-0.644655976448)*x[1]
        arg[1,1,2,1]=(-0.254976445964)*x[0]**o+(-0.624982185158)*x[0]+(0.304854486996)*x[1]**o+(0.597089811682)*x[1]
        arg[1,2,0,0]=(-0.625190759215)*x[0]**o+(-0.267820197905)*x[0]+(-0.721452024377)*x[1]**o+(-0.830984245819)*x[1]
        arg[1,2,0,1]=(-0.972257608766)*x[0]**o+(0.981368890544)*x[0]+(-0.925886510064)*x[1]**o+(-0.337107356862)*x[1]
        arg[1,2,1,0]=(-0.178932954207)*x[0]**o+(-0.941789507374)*x[0]+(0.360726744794)*x[1]**o+(0.347621863123)*x[1]
        arg[1,2,1,1]=(-0.59790695611)*x[0]**o+(0.552509013166)*x[0]+(0.592412443998)*x[1]**o+(0.105369637971)*x[1]
        arg[1,2,2,0]=(0.300713195225)*x[0]**o+(0.814257503574)*x[0]+(-0.281407483492)*x[1]**o+(-0.803431681095)*x[1]
        arg[1,2,2,1]=(0.80113022942)*x[0]**o+(0.749336171665)*x[0]+(-0.0842928982283)*x[1]**o+(-0.974041792058)*x[1]
        arg[1,3,0,0]=(0.533118876666)*x[0]**o+(-0.00888178987567)*x[0]+(0.372869213762)*x[1]**o+(-0.652513683144)*x[1]
        arg[1,3,0,1]=(-0.295090222537)*x[0]**o+(0.772749124303)*x[0]+(-0.00857029468729)*x[1]**o+(0.927794431125)*x[1]
        arg[1,3,1,0]=(-0.152537465441)*x[0]**o+(0.154900534979)*x[0]+(-0.488743959543)*x[1]**o+(-0.0318497199232)*x[1]
        arg[1,3,1,1]=(-0.754333787921)*x[0]**o+(-0.0583007358701)*x[0]+(-0.050921106891)*x[1]**o+(0.161426452156)*x[1]
        arg[1,3,2,0]=(0.387357469451)*x[0]**o+(-0.213640305885)*x[0]+(-0.844609663023)*x[1]**o+(0.0761914594192)*x[1]
        arg[1,3,2,1]=(0.36964862459)*x[0]**o+(0.887888401952)*x[0]+(-0.933930511093)*x[1]**o+(0.769046614932)*x[1]
        arg[1,4,0,0]=(0.680878887266)*x[0]**o+(0.62930033679)*x[0]+(0.894793542585)*x[1]**o+(-0.636881609067)*x[1]
        arg[1,4,0,1]=(-0.346053756598)*x[0]**o+(0.225254433923)*x[0]+(-0.861052985621)*x[1]**o+(0.415123108136)*x[1]
        arg[1,4,1,0]=(-0.997758408183)*x[0]**o+(-0.265055614341)*x[0]+(-0.7568881617)*x[1]**o+(0.95144584777)*x[1]
        arg[1,4,1,1]=(0.227777912008)*x[0]**o+(-0.83647660817)*x[0]+(-0.275733410165)*x[1]**o+(-0.971560248343)*x[1]
        arg[1,4,2,0]=(0.718803859114)*x[0]**o+(-0.978966510682)*x[0]+(0.463102438649)*x[1]**o+(0.987805610534)*x[1]
        arg[1,4,2,1]=(-0.107543838917)*x[0]**o+(-0.362684018694)*x[0]+(-0.310926939373)*x[1]**o+(-0.571571044712)*x[1]
        arg[2,0,0,0]=(-0.475977416182)*x[0]**o+(-0.627173002609)*x[0]+(-0.369962970984)*x[1]**o+(-0.245077302537)*x[1]
        arg[2,0,0,1]=(0.913949611751)*x[0]**o+(0.645611424617)*x[0]+(0.639304734076)*x[1]**o+(-0.12171441998)*x[1]
        arg[2,0,1,0]=(0.251434098263)*x[0]**o+(-0.119898861407)*x[0]+(-0.365614523709)*x[1]**o+(-0.802928738296)*x[1]
        arg[2,0,1,1]=(0.756390927006)*x[0]**o+(0.246578936199)*x[0]+(0.714968085461)*x[1]**o+(-0.929428808096)*x[1]
        arg[2,0,2,0]=(-0.516888361447)*x[0]**o+(0.0344339250886)*x[0]+(-0.383709788774)*x[1]**o+(-0.437344417623)*x[1]
        arg[2,0,2,1]=(-0.219307162317)*x[0]**o+(0.704909718842)*x[0]+(-0.340454491855)*x[1]**o+(-0.933866231773)*x[1]
        arg[2,1,0,0]=(-0.0618967891627)*x[0]**o+(-0.0590712030789)*x[0]+(0.116776433077)*x[1]**o+(0.0241239764104)*x[1]
        arg[2,1,0,1]=(-0.802207645638)*x[0]**o+(0.590732740531)*x[0]+(-0.25990272933)*x[1]**o+(-0.257633805416)*x[1]
        arg[2,1,1,0]=(0.827623274379)*x[0]**o+(-0.450310573599)*x[0]+(0.213809775826)*x[1]**o+(-0.905410777787)*x[1]
        arg[2,1,1,1]=(-0.624510830356)*x[0]**o+(-0.705682607787)*x[0]+(0.430105716993)*x[1]**o+(0.975510832521)*x[1]
        arg[2,1,2,0]=(-0.297562223309)*x[0]**o+(-0.99090348994)*x[0]+(0.548589416291)*x[1]**o+(0.591718478583)*x[1]
        arg[2,1,2,1]=(-0.531899919376)*x[0]**o+(-0.227453719581)*x[0]+(-0.131664806128)*x[1]**o+(0.0521895817509)*x[1]
        arg[2,2,0,0]=(-0.592291745786)*x[0]**o+(-0.586879961972)*x[0]+(0.563196552327)*x[1]**o+(0.876989321187)*x[1]
        arg[2,2,0,1]=(-0.273360445369)*x[0]**o+(-0.714740527775)*x[0]+(-0.647699497837)*x[1]**o+(-0.0405089612762)*x[1]
        arg[2,2,1,0]=(-0.536629732224)*x[0]**o+(0.72656373966)*x[0]+(0.986630594165)*x[1]**o+(0.0741630711667)*x[1]
        arg[2,2,1,1]=(0.770746088838)*x[0]**o+(-0.219279833426)*x[0]+(0.875445174779)*x[1]**o+(-0.947889926444)*x[1]
        arg[2,2,2,0]=(-0.41428815119)*x[0]**o+(-0.696166268903)*x[0]+(-0.96838062553)*x[1]**o+(0.781394205899)*x[1]
        arg[2,2,2,1]=(-0.810267960752)*x[0]**o+(-0.873632687702)*x[0]+(0.0131971321611)*x[1]**o+(-0.552767192193)*x[1]
        arg[2,3,0,0]=(-0.216091366291)*x[0]**o+(-0.17824047472)*x[0]+(-0.4644515725)*x[1]**o+(0.135819458957)*x[1]
        arg[2,3,0,1]=(-0.0823616523105)*x[0]**o+(0.33121362151)*x[0]+(0.926719919839)*x[1]**o+(0.663475641636)*x[1]
        arg[2,3,1,0]=(-0.278303066047)*x[0]**o+(-0.457286378314)*x[0]+(0.186792859092)*x[1]**o+(-0.734245612464)*x[1]
        arg[2,3,1,1]=(0.810871459779)*x[0]**o+(-0.706641995324)*x[0]+(-0.287827176091)*x[1]**o+(0.453216090282)*x[1]
        arg[2,3,2,0]=(0.83826221049)*x[0]**o+(0.556887792111)*x[0]+(-0.000231102245991)*x[1]**o+(-0.440730958954)*x[1]
        arg[2,3,2,1]=(-0.455087877907)*x[0]**o+(-0.132259634036)*x[0]+(-0.492728335525)*x[1]**o+(-0.880311317777)*x[1]
        arg[2,4,0,0]=(-0.303141352388)*x[0]**o+(0.426137046804)*x[0]+(0.788934776738)*x[1]**o+(0.157477785179)*x[1]
        arg[2,4,0,1]=(0.603411288614)*x[0]**o+(-0.533561451987)*x[0]+(-0.901259210942)*x[1]**o+(0.547372624746)*x[1]
        arg[2,4,1,0]=(0.69517961739)*x[0]**o+(-0.106282888246)*x[0]+(0.257484565681)*x[1]**o+(0.655749633294)*x[1]
        arg[2,4,1,1]=(-0.072748573118)*x[0]**o+(0.905966560113)*x[0]+(0.433145718123)*x[1]**o+(0.187620401615)*x[1]
        arg[2,4,2,0]=(-0.0280176325954)*x[0]**o+(0.489610131347)*x[0]+(-0.740554922134)*x[1]**o+(-0.85045302096)*x[1]
        arg[2,4,2,1]=(-0.158503180709)*x[0]**o+(0.525929643075)*x[0]+(-0.352022652515)*x[1]**o+(-0.706553104932)*x[1]
        arg[3,0,0,0]=(-0.837316508127)*x[0]**o+(-0.89490181779)*x[0]+(0.510084075291)*x[1]**o+(0.321206310024)*x[1]
        arg[3,0,0,1]=(-0.325038591079)*x[0]**o+(0.140342922753)*x[0]+(-0.937466403004)*x[1]**o+(0.580557383497)*x[1]
        arg[3,0,1,0]=(-0.939713654062)*x[0]**o+(-0.394660244166)*x[0]+(0.762376614557)*x[1]**o+(-0.135565547532)*x[1]
        arg[3,0,1,1]=(0.0275332524464)*x[0]**o+(0.368737701882)*x[0]+(0.983641372218)*x[1]**o+(0.573815516366)*x[1]
        arg[3,0,2,0]=(0.816957570314)*x[0]**o+(-0.357626892562)*x[0]+(-0.368933149642)*x[1]**o+(-0.692502486914)*x[1]
        arg[3,0,2,1]=(-0.372408810926)*x[0]**o+(-0.113036390416)*x[0]+(-0.355880297003)*x[1]**o+(0.755478191481)*x[1]
        arg[3,1,0,0]=(0.0399012900832)*x[0]**o+(-0.879576900086)*x[0]+(-0.660858776134)*x[1]**o+(-0.59837040429)*x[1]
        arg[3,1,0,1]=(0.569397128496)*x[0]**o+(-0.138333417746)*x[0]+(0.257769147727)*x[1]**o+(0.174737141027)*x[1]
        arg[3,1,1,0]=(-0.67938408823)*x[0]**o+(0.862731161981)*x[0]+(-0.792538485264)*x[1]**o+(-0.0805197796157)*x[1]
        arg[3,1,1,1]=(0.0113523229279)*x[0]**o+(0.475393597561)*x[0]+(-0.705748021886)*x[1]**o+(0.00836460135285)*x[1]
        arg[3,1,2,0]=(0.463839509799)*x[0]**o+(-0.12663544643)*x[0]+(0.353158399738)*x[1]**o+(0.674929537597)*x[1]
        arg[3,1,2,1]=(-0.6019896019)*x[0]**o+(-0.638280685565)*x[0]+(0.280047072988)*x[1]**o+(0.569239958138)*x[1]
        arg[3,2,0,0]=(-0.113906500346)*x[0]**o+(-0.493494498501)*x[0]+(-0.686419229162)*x[1]**o+(-0.0727046763516)*x[1]
        arg[3,2,0,1]=(-0.302092112795)*x[0]**o+(-0.74819167)*x[0]+(-0.478908975282)*x[1]**o+(0.430024789627)*x[1]
        arg[3,2,1,0]=(0.353535608984)*x[0]**o+(-0.743408549264)*x[0]+(-0.069122881538)*x[1]**o+(-0.0767595089529)*x[1]
        arg[3,2,1,1]=(-0.629946386234)*x[0]**o+(0.851779590894)*x[0]+(0.728577108695)*x[1]**o+(-0.366983040147)*x[1]
        arg[3,2,2,0]=(0.1685167508)*x[0]**o+(0.149999747637)*x[0]+(0.709320430374)*x[1]**o+(-0.280950677222)*x[1]
        arg[3,2,2,1]=(-0.956164974578)*x[0]**o+(0.240460380901)*x[0]+(-0.2820176577)*x[1]**o+(-0.595667156955)*x[1]
        arg[3,3,0,0]=(-0.0711295055938)*x[0]**o+(-0.314513251791)*x[0]+(-0.41924591293)*x[1]**o+(0.0813197780959)*x[1]
        arg[3,3,0,1]=(-0.230592584877)*x[0]**o+(-0.488005144715)*x[0]+(-0.450611759976)*x[1]**o+(-0.0571152769768)*x[1]
        arg[3,3,1,0]=(-0.608255699014)*x[0]**o+(-0.697324108575)*x[0]+(-0.885935900746)*x[1]**o+(-0.884497550016)*x[1]
        arg[3,3,1,1]=(0.857286225292)*x[0]**o+(0.509954550626)*x[0]+(0.182030821433)*x[1]**o+(-0.268076060316)*x[1]
        arg[3,3,2,0]=(-0.77322653479)*x[0]**o+(-0.583013210018)*x[0]+(-0.683867300277)*x[1]**o+(0.754530541251)*x[1]
        arg[3,3,2,1]=(-0.744667985181)*x[0]**o+(0.164165128978)*x[0]+(0.412340155958)*x[1]**o+(-0.38451107118)*x[1]
        arg[3,4,0,0]=(0.216720755049)*x[0]**o+(-0.347803433722)*x[0]+(0.417376268797)*x[1]**o+(0.475787304652)*x[1]
        arg[3,4,0,1]=(-0.737038295974)*x[0]**o+(0.579172519178)*x[0]+(-0.900272993672)*x[1]**o+(0.407018603625)*x[1]
        arg[3,4,1,0]=(-0.898959236474)*x[0]**o+(-0.886636008121)*x[0]+(-0.901972610044)*x[1]**o+(-0.518393782318)*x[1]
        arg[3,4,1,1]=(0.976342884921)*x[0]**o+(-0.273568319493)*x[0]+(-0.332424980846)*x[1]**o+(-0.0146661731139)*x[1]
        arg[3,4,2,0]=(0.794792214027)*x[0]**o+(0.846927984413)*x[0]+(0.535818809709)*x[1]**o+(-0.49324400033)*x[1]
        arg[3,4,2,1]=(0.695766308812)*x[0]**o+(-0.37923153689)*x[0]+(0.0633513026782)*x[1]**o+(0.12334892105)*x[1]
        ref[0,0,0,0]=(0.912518456494)*x_ref[0]**o+(-0.828176151903)*x_ref[0]+(-0.313940298121)*x_ref[1]**o+(-0.850120336531)*x_ref[1]
        ref[0,0,0,1]=(0.14415852783)*x_ref[0]**o+(-0.294155246668)*x_ref[0]+(0.562517248232)*x_ref[1]**o+(-0.204505938286)*x_ref[1]
        ref[0,0,1,0]=(0.555455404368)*x_ref[0]**o+(0.916425343576)*x_ref[0]+(0.547252315693)*x_ref[1]**o+(-0.0992842367479)*x_ref[1]
        ref[0,0,1,1]=(0.603632473928)*x_ref[0]**o+(0.505571549062)*x_ref[0]+(-0.490402921542)*x_ref[1]**o+(0.908624427316)*x_ref[1]
        ref[0,0,2,0]=(0.729542308985)*x_ref[0]**o+(-0.654310868638)*x_ref[0]+(0.0845981276245)*x_ref[1]**o+(-0.95231604686)*x_ref[1]
        ref[0,0,2,1]=(0.0511615606745)*x_ref[0]**o+(-0.939469887975)*x_ref[0]+(-0.206229418155)*x_ref[1]**o+(0.254879844701)*x_ref[1]
        ref[0,1,0,0]=(0.257917344507)*x_ref[0]**o+(0.860729621288)*x_ref[0]+(0.489966193348)*x_ref[1]**o+(0.174897233858)*x_ref[1]
        ref[0,1,0,1]=(0.336588362051)*x_ref[0]**o+(-0.690598635114)*x_ref[0]+(-0.0913735300483)*x_ref[1]**o+(-0.932914439004)*x_ref[1]
        ref[0,1,1,0]=(0.244583093108)*x_ref[0]**o+(-0.827768167758)*x_ref[0]+(0.573065359412)*x_ref[1]**o+(0.176685515596)*x_ref[1]
        ref[0,1,1,1]=(-0.520550726127)*x_ref[0]**o+(-0.605243177286)*x_ref[0]+(0.130549556226)*x_ref[1]**o+(0.750019884067)*x_ref[1]
        ref[0,1,2,0]=(-0.680201627614)*x_ref[0]**o+(0.679925324882)*x_ref[0]+(-0.060408604263)*x_ref[1]**o+(0.568125689359)*x_ref[1]
        ref[0,1,2,1]=(-0.111908859432)*x_ref[0]**o+(0.256521037018)*x_ref[0]+(0.332050616611)*x_ref[1]**o+(-0.0500016397683)*x_ref[1]
        ref[0,2,0,0]=(-0.484953933663)*x_ref[0]**o+(0.628032211831)*x_ref[0]+(-0.63297141454)*x_ref[1]**o+(0.141016662755)*x_ref[1]
        ref[0,2,0,1]=(-0.463920680465)*x_ref[0]**o+(-0.868825520511)*x_ref[0]+(-0.36594425006)*x_ref[1]**o+(-0.823052216578)*x_ref[1]
        ref[0,2,1,0]=(0.972337802357)*x_ref[0]**o+(0.943605157343)*x_ref[0]+(-0.113950717844)*x_ref[1]**o+(-0.839596452897)*x_ref[1]
        ref[0,2,1,1]=(-0.295562401302)*x_ref[0]**o+(0.735770621017)*x_ref[0]+(-0.183838912348)*x_ref[1]**o+(-0.929365380252)*x_ref[1]
        ref[0,2,2,0]=(0.912886298388)*x_ref[0]**o+(0.847507982065)*x_ref[0]+(-0.341306036323)*x_ref[1]**o+(0.393627025978)*x_ref[1]
        ref[0,2,2,1]=(-0.769650638386)*x_ref[0]**o+(-0.619672356488)*x_ref[0]+(-0.126894288906)*x_ref[1]**o+(0.231547632139)*x_ref[1]
        ref[0,3,0,0]=(-0.884442976834)*x_ref[0]**o+(-0.0194505760676)*x_ref[0]+(-0.338764675072)*x_ref[1]**o+(0.723481600171)*x_ref[1]
        ref[0,3,0,1]=(0.655973555091)*x_ref[0]**o+(-0.556230649412)*x_ref[0]+(-0.850356832441)*x_ref[1]**o+(0.115267314853)*x_ref[1]
        ref[0,3,1,0]=(0.659628443559)*x_ref[0]**o+(-0.586677822491)*x_ref[0]+(0.135930588518)*x_ref[1]**o+(0.206866895948)*x_ref[1]
        ref[0,3,1,1]=(-0.700258957557)*x_ref[0]**o+(-0.699774198764)*x_ref[0]+(-0.294009479186)*x_ref[1]**o+(0.848159666503)*x_ref[1]
        ref[0,3,2,0]=(-0.927501754488)*x_ref[0]**o+(-0.981178071331)*x_ref[0]+(0.614074194451)*x_ref[1]**o+(0.711549573086)*x_ref[1]
        ref[0,3,2,1]=(0.312434473198)*x_ref[0]**o+(-0.241506125991)*x_ref[0]+(-0.0840979088912)*x_ref[1]**o+(0.902948463117)*x_ref[1]
        ref[0,4,0,0]=(-0.892269860242)*x_ref[0]**o+(0.823434435855)*x_ref[0]+(0.50499383365)*x_ref[1]**o+(0.763736541034)*x_ref[1]
        ref[0,4,0,1]=(-0.606959691191)*x_ref[0]**o+(0.945562536791)*x_ref[0]+(-0.46271929481)*x_ref[1]**o+(0.929514432096)*x_ref[1]
        ref[0,4,1,0]=(0.953382685939)*x_ref[0]**o+(0.553766359805)*x_ref[0]+(-0.191004010293)*x_ref[1]**o+(0.556369491698)*x_ref[1]
        ref[0,4,1,1]=(-0.269402649165)*x_ref[0]**o+(-0.158926232076)*x_ref[0]+(-0.390038749681)*x_ref[1]**o+(0.553655192666)*x_ref[1]
        ref[0,4,2,0]=(0.0785586372006)*x_ref[0]**o+(-0.778452221862)*x_ref[0]+(-0.927627273292)*x_ref[1]**o+(-0.187742405246)*x_ref[1]
        ref[0,4,2,1]=(-0.965329945447)*x_ref[0]**o+(0.304539006089)*x_ref[0]+(0.213031689568)*x_ref[1]**o+(-0.2807790149)*x_ref[1]
        ref[1,0,0,0]=(0.981534698569)*x_ref[0]**o+(0.466451685806)*x_ref[0]+(0.845322931856)*x_ref[1]**o+(0.582422032334)*x_ref[1]
        ref[1,0,0,1]=(-0.0645798740394)*x_ref[0]**o+(0.0475214206619)*x_ref[0]+(-0.687759433076)*x_ref[1]**o+(0.403440938285)*x_ref[1]
        ref[1,0,1,0]=(-0.317887773392)*x_ref[0]**o+(0.59101163416)*x_ref[0]+(0.658626263291)*x_ref[1]**o+(-0.618167594137)*x_ref[1]
        ref[1,0,1,1]=(-0.314093092813)*x_ref[0]**o+(0.699933899942)*x_ref[0]+(0.824195260822)*x_ref[1]**o+(-0.42875327345)*x_ref[1]
        ref[1,0,2,0]=(-0.254254956887)*x_ref[0]**o+(-0.912137080163)*x_ref[0]+(0.360457121452)*x_ref[1]**o+(-0.902596912588)*x_ref[1]
        ref[1,0,2,1]=(-0.577214593683)*x_ref[0]**o+(-0.419495663008)*x_ref[0]+(-0.803459783331)*x_ref[1]**o+(0.490885513568)*x_ref[1]
        ref[1,1,0,0]=(0.477947766892)*x_ref[0]**o+(-0.853046971581)*x_ref[0]+(0.576044041601)*x_ref[1]**o+(-0.447007701252)*x_ref[1]
        ref[1,1,0,1]=(0.5886933409)*x_ref[0]**o+(0.253650778351)*x_ref[0]+(-0.198070684431)*x_ref[1]**o+(0.189479364528)*x_ref[1]
        ref[1,1,1,0]=(-0.528744478872)*x_ref[0]**o+(-0.504072059557)*x_ref[0]+(-0.0161832474015)*x_ref[1]**o+(0.639592055966)*x_ref[1]
        ref[1,1,1,1]=(-0.581844827169)*x_ref[0]**o+(0.997883224538)*x_ref[0]+(0.9354364489)*x_ref[1]**o+(0.884060785136)*x_ref[1]
        ref[1,1,2,0]=(0.858226955817)*x_ref[0]**o+(0.707261698312)*x_ref[0]+(-0.738309111076)*x_ref[1]**o+(-0.644655976448)*x_ref[1]
        ref[1,1,2,1]=(-0.254976445964)*x_ref[0]**o+(-0.624982185158)*x_ref[0]+(0.304854486996)*x_ref[1]**o+(0.597089811682)*x_ref[1]
        ref[1,2,0,0]=(-0.625190759215)*x_ref[0]**o+(-0.267820197905)*x_ref[0]+(-0.721452024377)*x_ref[1]**o+(-0.830984245819)*x_ref[1]
        ref[1,2,0,1]=(-0.972257608766)*x_ref[0]**o+(0.981368890544)*x_ref[0]+(-0.925886510064)*x_ref[1]**o+(-0.337107356862)*x_ref[1]
        ref[1,2,1,0]=(-0.178932954207)*x_ref[0]**o+(-0.941789507374)*x_ref[0]+(0.360726744794)*x_ref[1]**o+(0.347621863123)*x_ref[1]
        ref[1,2,1,1]=(-0.59790695611)*x_ref[0]**o+(0.552509013166)*x_ref[0]+(0.592412443998)*x_ref[1]**o+(0.105369637971)*x_ref[1]
        ref[1,2,2,0]=(0.300713195225)*x_ref[0]**o+(0.814257503574)*x_ref[0]+(-0.281407483492)*x_ref[1]**o+(-0.803431681095)*x_ref[1]
        ref[1,2,2,1]=(0.80113022942)*x_ref[0]**o+(0.749336171665)*x_ref[0]+(-0.0842928982283)*x_ref[1]**o+(-0.974041792058)*x_ref[1]
        ref[1,3,0,0]=(0.533118876666)*x_ref[0]**o+(-0.00888178987567)*x_ref[0]+(0.372869213762)*x_ref[1]**o+(-0.652513683144)*x_ref[1]
        ref[1,3,0,1]=(-0.295090222537)*x_ref[0]**o+(0.772749124303)*x_ref[0]+(-0.00857029468729)*x_ref[1]**o+(0.927794431125)*x_ref[1]
        ref[1,3,1,0]=(-0.152537465441)*x_ref[0]**o+(0.154900534979)*x_ref[0]+(-0.488743959543)*x_ref[1]**o+(-0.0318497199232)*x_ref[1]
        ref[1,3,1,1]=(-0.754333787921)*x_ref[0]**o+(-0.0583007358701)*x_ref[0]+(-0.050921106891)*x_ref[1]**o+(0.161426452156)*x_ref[1]
        ref[1,3,2,0]=(0.387357469451)*x_ref[0]**o+(-0.213640305885)*x_ref[0]+(-0.844609663023)*x_ref[1]**o+(0.0761914594192)*x_ref[1]
        ref[1,3,2,1]=(0.36964862459)*x_ref[0]**o+(0.887888401952)*x_ref[0]+(-0.933930511093)*x_ref[1]**o+(0.769046614932)*x_ref[1]
        ref[1,4,0,0]=(0.680878887266)*x_ref[0]**o+(0.62930033679)*x_ref[0]+(0.894793542585)*x_ref[1]**o+(-0.636881609067)*x_ref[1]
        ref[1,4,0,1]=(-0.346053756598)*x_ref[0]**o+(0.225254433923)*x_ref[0]+(-0.861052985621)*x_ref[1]**o+(0.415123108136)*x_ref[1]
        ref[1,4,1,0]=(-0.997758408183)*x_ref[0]**o+(-0.265055614341)*x_ref[0]+(-0.7568881617)*x_ref[1]**o+(0.95144584777)*x_ref[1]
        ref[1,4,1,1]=(0.227777912008)*x_ref[0]**o+(-0.83647660817)*x_ref[0]+(-0.275733410165)*x_ref[1]**o+(-0.971560248343)*x_ref[1]
        ref[1,4,2,0]=(0.718803859114)*x_ref[0]**o+(-0.978966510682)*x_ref[0]+(0.463102438649)*x_ref[1]**o+(0.987805610534)*x_ref[1]
        ref[1,4,2,1]=(-0.107543838917)*x_ref[0]**o+(-0.362684018694)*x_ref[0]+(-0.310926939373)*x_ref[1]**o+(-0.571571044712)*x_ref[1]
        ref[2,0,0,0]=(-0.475977416182)*x_ref[0]**o+(-0.627173002609)*x_ref[0]+(-0.369962970984)*x_ref[1]**o+(-0.245077302537)*x_ref[1]
        ref[2,0,0,1]=(0.913949611751)*x_ref[0]**o+(0.645611424617)*x_ref[0]+(0.639304734076)*x_ref[1]**o+(-0.12171441998)*x_ref[1]
        ref[2,0,1,0]=(0.251434098263)*x_ref[0]**o+(-0.119898861407)*x_ref[0]+(-0.365614523709)*x_ref[1]**o+(-0.802928738296)*x_ref[1]
        ref[2,0,1,1]=(0.756390927006)*x_ref[0]**o+(0.246578936199)*x_ref[0]+(0.714968085461)*x_ref[1]**o+(-0.929428808096)*x_ref[1]
        ref[2,0,2,0]=(-0.516888361447)*x_ref[0]**o+(0.0344339250886)*x_ref[0]+(-0.383709788774)*x_ref[1]**o+(-0.437344417623)*x_ref[1]
        ref[2,0,2,1]=(-0.219307162317)*x_ref[0]**o+(0.704909718842)*x_ref[0]+(-0.340454491855)*x_ref[1]**o+(-0.933866231773)*x_ref[1]
        ref[2,1,0,0]=(-0.0618967891627)*x_ref[0]**o+(-0.0590712030789)*x_ref[0]+(0.116776433077)*x_ref[1]**o+(0.0241239764104)*x_ref[1]
        ref[2,1,0,1]=(-0.802207645638)*x_ref[0]**o+(0.590732740531)*x_ref[0]+(-0.25990272933)*x_ref[1]**o+(-0.257633805416)*x_ref[1]
        ref[2,1,1,0]=(0.827623274379)*x_ref[0]**o+(-0.450310573599)*x_ref[0]+(0.213809775826)*x_ref[1]**o+(-0.905410777787)*x_ref[1]
        ref[2,1,1,1]=(-0.624510830356)*x_ref[0]**o+(-0.705682607787)*x_ref[0]+(0.430105716993)*x_ref[1]**o+(0.975510832521)*x_ref[1]
        ref[2,1,2,0]=(-0.297562223309)*x_ref[0]**o+(-0.99090348994)*x_ref[0]+(0.548589416291)*x_ref[1]**o+(0.591718478583)*x_ref[1]
        ref[2,1,2,1]=(-0.531899919376)*x_ref[0]**o+(-0.227453719581)*x_ref[0]+(-0.131664806128)*x_ref[1]**o+(0.0521895817509)*x_ref[1]
        ref[2,2,0,0]=(-0.592291745786)*x_ref[0]**o+(-0.586879961972)*x_ref[0]+(0.563196552327)*x_ref[1]**o+(0.876989321187)*x_ref[1]
        ref[2,2,0,1]=(-0.273360445369)*x_ref[0]**o+(-0.714740527775)*x_ref[0]+(-0.647699497837)*x_ref[1]**o+(-0.0405089612762)*x_ref[1]
        ref[2,2,1,0]=(-0.536629732224)*x_ref[0]**o+(0.72656373966)*x_ref[0]+(0.986630594165)*x_ref[1]**o+(0.0741630711667)*x_ref[1]
        ref[2,2,1,1]=(0.770746088838)*x_ref[0]**o+(-0.219279833426)*x_ref[0]+(0.875445174779)*x_ref[1]**o+(-0.947889926444)*x_ref[1]
        ref[2,2,2,0]=(-0.41428815119)*x_ref[0]**o+(-0.696166268903)*x_ref[0]+(-0.96838062553)*x_ref[1]**o+(0.781394205899)*x_ref[1]
        ref[2,2,2,1]=(-0.810267960752)*x_ref[0]**o+(-0.873632687702)*x_ref[0]+(0.0131971321611)*x_ref[1]**o+(-0.552767192193)*x_ref[1]
        ref[2,3,0,0]=(-0.216091366291)*x_ref[0]**o+(-0.17824047472)*x_ref[0]+(-0.4644515725)*x_ref[1]**o+(0.135819458957)*x_ref[1]
        ref[2,3,0,1]=(-0.0823616523105)*x_ref[0]**o+(0.33121362151)*x_ref[0]+(0.926719919839)*x_ref[1]**o+(0.663475641636)*x_ref[1]
        ref[2,3,1,0]=(-0.278303066047)*x_ref[0]**o+(-0.457286378314)*x_ref[0]+(0.186792859092)*x_ref[1]**o+(-0.734245612464)*x_ref[1]
        ref[2,3,1,1]=(0.810871459779)*x_ref[0]**o+(-0.706641995324)*x_ref[0]+(-0.287827176091)*x_ref[1]**o+(0.453216090282)*x_ref[1]
        ref[2,3,2,0]=(0.83826221049)*x_ref[0]**o+(0.556887792111)*x_ref[0]+(-0.000231102245991)*x_ref[1]**o+(-0.440730958954)*x_ref[1]
        ref[2,3,2,1]=(-0.455087877907)*x_ref[0]**o+(-0.132259634036)*x_ref[0]+(-0.492728335525)*x_ref[1]**o+(-0.880311317777)*x_ref[1]
        ref[2,4,0,0]=(-0.303141352388)*x_ref[0]**o+(0.426137046804)*x_ref[0]+(0.788934776738)*x_ref[1]**o+(0.157477785179)*x_ref[1]
        ref[2,4,0,1]=(0.603411288614)*x_ref[0]**o+(-0.533561451987)*x_ref[0]+(-0.901259210942)*x_ref[1]**o+(0.547372624746)*x_ref[1]
        ref[2,4,1,0]=(0.69517961739)*x_ref[0]**o+(-0.106282888246)*x_ref[0]+(0.257484565681)*x_ref[1]**o+(0.655749633294)*x_ref[1]
        ref[2,4,1,1]=(-0.072748573118)*x_ref[0]**o+(0.905966560113)*x_ref[0]+(0.433145718123)*x_ref[1]**o+(0.187620401615)*x_ref[1]
        ref[2,4,2,0]=(-0.0280176325954)*x_ref[0]**o+(0.489610131347)*x_ref[0]+(-0.740554922134)*x_ref[1]**o+(-0.85045302096)*x_ref[1]
        ref[2,4,2,1]=(-0.158503180709)*x_ref[0]**o+(0.525929643075)*x_ref[0]+(-0.352022652515)*x_ref[1]**o+(-0.706553104932)*x_ref[1]
        ref[3,0,0,0]=(-0.837316508127)*x_ref[0]**o+(-0.89490181779)*x_ref[0]+(0.510084075291)*x_ref[1]**o+(0.321206310024)*x_ref[1]
        ref[3,0,0,1]=(-0.325038591079)*x_ref[0]**o+(0.140342922753)*x_ref[0]+(-0.937466403004)*x_ref[1]**o+(0.580557383497)*x_ref[1]
        ref[3,0,1,0]=(-0.939713654062)*x_ref[0]**o+(-0.394660244166)*x_ref[0]+(0.762376614557)*x_ref[1]**o+(-0.135565547532)*x_ref[1]
        ref[3,0,1,1]=(0.0275332524464)*x_ref[0]**o+(0.368737701882)*x_ref[0]+(0.983641372218)*x_ref[1]**o+(0.573815516366)*x_ref[1]
        ref[3,0,2,0]=(0.816957570314)*x_ref[0]**o+(-0.357626892562)*x_ref[0]+(-0.368933149642)*x_ref[1]**o+(-0.692502486914)*x_ref[1]
        ref[3,0,2,1]=(-0.372408810926)*x_ref[0]**o+(-0.113036390416)*x_ref[0]+(-0.355880297003)*x_ref[1]**o+(0.755478191481)*x_ref[1]
        ref[3,1,0,0]=(0.0399012900832)*x_ref[0]**o+(-0.879576900086)*x_ref[0]+(-0.660858776134)*x_ref[1]**o+(-0.59837040429)*x_ref[1]
        ref[3,1,0,1]=(0.569397128496)*x_ref[0]**o+(-0.138333417746)*x_ref[0]+(0.257769147727)*x_ref[1]**o+(0.174737141027)*x_ref[1]
        ref[3,1,1,0]=(-0.67938408823)*x_ref[0]**o+(0.862731161981)*x_ref[0]+(-0.792538485264)*x_ref[1]**o+(-0.0805197796157)*x_ref[1]
        ref[3,1,1,1]=(0.0113523229279)*x_ref[0]**o+(0.475393597561)*x_ref[0]+(-0.705748021886)*x_ref[1]**o+(0.00836460135285)*x_ref[1]
        ref[3,1,2,0]=(0.463839509799)*x_ref[0]**o+(-0.12663544643)*x_ref[0]+(0.353158399738)*x_ref[1]**o+(0.674929537597)*x_ref[1]
        ref[3,1,2,1]=(-0.6019896019)*x_ref[0]**o+(-0.638280685565)*x_ref[0]+(0.280047072988)*x_ref[1]**o+(0.569239958138)*x_ref[1]
        ref[3,2,0,0]=(-0.113906500346)*x_ref[0]**o+(-0.493494498501)*x_ref[0]+(-0.686419229162)*x_ref[1]**o+(-0.0727046763516)*x_ref[1]
        ref[3,2,0,1]=(-0.302092112795)*x_ref[0]**o+(-0.74819167)*x_ref[0]+(-0.478908975282)*x_ref[1]**o+(0.430024789627)*x_ref[1]
        ref[3,2,1,0]=(0.353535608984)*x_ref[0]**o+(-0.743408549264)*x_ref[0]+(-0.069122881538)*x_ref[1]**o+(-0.0767595089529)*x_ref[1]
        ref[3,2,1,1]=(-0.629946386234)*x_ref[0]**o+(0.851779590894)*x_ref[0]+(0.728577108695)*x_ref[1]**o+(-0.366983040147)*x_ref[1]
        ref[3,2,2,0]=(0.1685167508)*x_ref[0]**o+(0.149999747637)*x_ref[0]+(0.709320430374)*x_ref[1]**o+(-0.280950677222)*x_ref[1]
        ref[3,2,2,1]=(-0.956164974578)*x_ref[0]**o+(0.240460380901)*x_ref[0]+(-0.2820176577)*x_ref[1]**o+(-0.595667156955)*x_ref[1]
        ref[3,3,0,0]=(-0.0711295055938)*x_ref[0]**o+(-0.314513251791)*x_ref[0]+(-0.41924591293)*x_ref[1]**o+(0.0813197780959)*x_ref[1]
        ref[3,3,0,1]=(-0.230592584877)*x_ref[0]**o+(-0.488005144715)*x_ref[0]+(-0.450611759976)*x_ref[1]**o+(-0.0571152769768)*x_ref[1]
        ref[3,3,1,0]=(-0.608255699014)*x_ref[0]**o+(-0.697324108575)*x_ref[0]+(-0.885935900746)*x_ref[1]**o+(-0.884497550016)*x_ref[1]
        ref[3,3,1,1]=(0.857286225292)*x_ref[0]**o+(0.509954550626)*x_ref[0]+(0.182030821433)*x_ref[1]**o+(-0.268076060316)*x_ref[1]
        ref[3,3,2,0]=(-0.77322653479)*x_ref[0]**o+(-0.583013210018)*x_ref[0]+(-0.683867300277)*x_ref[1]**o+(0.754530541251)*x_ref[1]
        ref[3,3,2,1]=(-0.744667985181)*x_ref[0]**o+(0.164165128978)*x_ref[0]+(0.412340155958)*x_ref[1]**o+(-0.38451107118)*x_ref[1]
        ref[3,4,0,0]=(0.216720755049)*x_ref[0]**o+(-0.347803433722)*x_ref[0]+(0.417376268797)*x_ref[1]**o+(0.475787304652)*x_ref[1]
        ref[3,4,0,1]=(-0.737038295974)*x_ref[0]**o+(0.579172519178)*x_ref[0]+(-0.900272993672)*x_ref[1]**o+(0.407018603625)*x_ref[1]
        ref[3,4,1,0]=(-0.898959236474)*x_ref[0]**o+(-0.886636008121)*x_ref[0]+(-0.901972610044)*x_ref[1]**o+(-0.518393782318)*x_ref[1]
        ref[3,4,1,1]=(0.976342884921)*x_ref[0]**o+(-0.273568319493)*x_ref[0]+(-0.332424980846)*x_ref[1]**o+(-0.0146661731139)*x_ref[1]
        ref[3,4,2,0]=(0.794792214027)*x_ref[0]**o+(0.846927984413)*x_ref[0]+(0.535818809709)*x_ref[1]**o+(-0.49324400033)*x_ref[1]
        ref[3,4,2,1]=(0.695766308812)*x_ref[0]**o+(-0.37923153689)*x_ref[0]+(0.0633513026782)*x_ref[1]**o+(0.12334892105)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.401901998082)*x[0]**o+(-0.805091116808)*x[0]+(0.997659882792)*x[1]**o+(-0.953957389492)*x[1]+(-0.450185408289)*x[2]**o+(0.636472207773)*x[2]
        arg[0,0,0,1]=(-0.786386817128)*x[0]**o+(-0.882559775393)*x[0]+(0.117871400819)*x[1]**o+(-0.625922274774)*x[1]+(0.410431886931)*x[2]**o+(-0.979356501067)*x[2]
        arg[0,0,1,0]=(0.686025409275)*x[0]**o+(-0.19806907659)*x[0]+(-0.710420224449)*x[1]**o+(-0.839236953983)*x[1]+(-0.79971896093)*x[2]**o+(0.378944111319)*x[2]
        arg[0,0,1,1]=(0.771205112484)*x[0]**o+(-0.007083192859)*x[0]+(0.672918588001)*x[1]**o+(-0.351652495793)*x[1]+(-0.841767161187)*x[2]**o+(-0.345642373014)*x[2]
        arg[0,0,2,0]=(0.266930729015)*x[0]**o+(-0.873808152307)*x[0]+(0.206466759135)*x[1]**o+(0.539888649948)*x[1]+(0.17985862191)*x[2]**o+(0.240651460153)*x[2]
        arg[0,0,2,1]=(-0.232569768041)*x[0]**o+(-0.579841454072)*x[0]+(-0.706956104891)*x[1]**o+(0.822863815698)*x[1]+(0.345774991358)*x[2]**o+(-0.0848567668311)*x[2]
        arg[0,1,0,0]=(0.51593851507)*x[0]**o+(-0.233342434349)*x[0]+(-0.92498298026)*x[1]**o+(0.164540234829)*x[1]+(0.356559744499)*x[2]**o+(0.483902511717)*x[2]
        arg[0,1,0,1]=(-0.811509654443)*x[0]**o+(0.73929453189)*x[0]+(0.256205706418)*x[1]**o+(-0.667924913793)*x[1]+(0.266065641257)*x[2]**o+(0.856490258029)*x[2]
        arg[0,1,1,0]=(0.502051590284)*x[0]**o+(0.927283797128)*x[0]+(0.500854521411)*x[1]**o+(-0.486326996369)*x[1]+(-0.327578523685)*x[2]**o+(-0.28106070397)*x[2]
        arg[0,1,1,1]=(0.201740014513)*x[0]**o+(-0.166988241759)*x[0]+(-0.513072955566)*x[1]**o+(0.67179532417)*x[1]+(-0.72135980627)*x[2]**o+(0.763548886467)*x[2]
        arg[0,1,2,0]=(0.3995590338)*x[0]**o+(0.433286929621)*x[0]+(-0.287309453792)*x[1]**o+(-0.0671679443074)*x[1]+(-0.80815399993)*x[2]**o+(-0.398740362114)*x[2]
        arg[0,1,2,1]=(-0.16314640684)*x[0]**o+(-0.698681697367)*x[0]+(-0.11609738418)*x[1]**o+(-0.798465154739)*x[1]+(-0.169052628971)*x[2]**o+(-0.880316962539)*x[2]
        arg[0,2,0,0]=(-0.219731279442)*x[0]**o+(0.731464233112)*x[0]+(0.637811533993)*x[1]**o+(0.381218512692)*x[1]+(-0.181357790048)*x[2]**o+(-0.296793024723)*x[2]
        arg[0,2,0,1]=(0.566451624572)*x[0]**o+(0.302904935559)*x[0]+(-0.771526665232)*x[1]**o+(0.460577683302)*x[1]+(-0.921312455507)*x[2]**o+(0.604299765394)*x[2]
        arg[0,2,1,0]=(-0.0887516895197)*x[0]**o+(0.872208558285)*x[0]+(-0.306564750999)*x[1]**o+(0.841982500263)*x[1]+(-0.550398737046)*x[2]**o+(-0.869697162427)*x[2]
        arg[0,2,1,1]=(-0.652471383912)*x[0]**o+(0.634592304393)*x[0]+(0.0709212473167)*x[1]**o+(-0.671330082579)*x[1]+(-0.232895004478)*x[2]**o+(0.479008351165)*x[2]
        arg[0,2,2,0]=(-0.0764115046136)*x[0]**o+(0.649230346025)*x[0]+(0.544503480999)*x[1]**o+(0.788838981797)*x[1]+(0.416236148267)*x[2]**o+(0.722979541918)*x[2]
        arg[0,2,2,1]=(-0.0776368628837)*x[0]**o+(0.801984674386)*x[0]+(0.443670313486)*x[1]**o+(0.424676792255)*x[1]+(-0.895989937982)*x[2]**o+(0.0774648143589)*x[2]
        arg[0,3,0,0]=(-0.39041318438)*x[0]**o+(-0.174335409262)*x[0]+(0.658666482337)*x[1]**o+(0.23060605921)*x[1]+(-0.336603364419)*x[2]**o+(0.871530192198)*x[2]
        arg[0,3,0,1]=(-0.493104661363)*x[0]**o+(0.590779697499)*x[0]+(-0.903897457443)*x[1]**o+(-0.377640520704)*x[1]+(0.227174544833)*x[2]**o+(-0.860150483539)*x[2]
        arg[0,3,1,0]=(-0.093367253954)*x[0]**o+(0.159230499811)*x[0]+(-0.731396509936)*x[1]**o+(0.600088148838)*x[1]+(0.829175660147)*x[2]**o+(-0.435878280919)*x[2]
        arg[0,3,1,1]=(0.665125540711)*x[0]**o+(-0.92128737118)*x[0]+(-0.33260297543)*x[1]**o+(0.682569476816)*x[1]+(0.999761405058)*x[2]**o+(-0.748303702618)*x[2]
        arg[0,3,2,0]=(-0.430442277525)*x[0]**o+(-0.841437664378)*x[0]+(-0.374267371055)*x[1]**o+(0.667860327373)*x[1]+(0.694185699238)*x[2]**o+(-0.673629629622)*x[2]
        arg[0,3,2,1]=(0.304808617243)*x[0]**o+(-0.894407776037)*x[0]+(-0.758585281699)*x[1]**o+(0.20519963241)*x[1]+(0.538222015181)*x[2]**o+(-0.792962435573)*x[2]
        arg[0,4,0,0]=(0.568300300269)*x[0]**o+(0.407455936625)*x[0]+(0.980898344535)*x[1]**o+(0.988220474461)*x[1]+(0.0598892620089)*x[2]**o+(-0.279555207723)*x[2]
        arg[0,4,0,1]=(-0.208391854989)*x[0]**o+(0.9428019078)*x[0]+(-0.540916877275)*x[1]**o+(-0.427185538555)*x[1]+(-0.879268085857)*x[2]**o+(0.025102321099)*x[2]
        arg[0,4,1,0]=(0.412690112527)*x[0]**o+(-0.957376863734)*x[0]+(-0.913623139719)*x[1]**o+(-0.462418849979)*x[1]+(0.463336747019)*x[2]**o+(0.330747531293)*x[2]
        arg[0,4,1,1]=(-0.632140552118)*x[0]**o+(0.830769577671)*x[0]+(-0.888012251479)*x[1]**o+(-0.0560181569699)*x[1]+(-0.70068099248)*x[2]**o+(0.834483253039)*x[2]
        arg[0,4,2,0]=(0.0429717567385)*x[0]**o+(-0.691784042595)*x[0]+(-0.264988764713)*x[1]**o+(0.359873170931)*x[1]+(-0.116275128437)*x[2]**o+(0.0348873917345)*x[2]
        arg[0,4,2,1]=(-0.203754358311)*x[0]**o+(0.177896174646)*x[0]+(-0.0785920580727)*x[1]**o+(-0.124905949184)*x[1]+(0.968871355367)*x[2]**o+(-0.0369722247589)*x[2]
        arg[1,0,0,0]=(0.594851756786)*x[0]**o+(-0.594400822936)*x[0]+(0.868401350949)*x[1]**o+(0.186739616297)*x[1]+(-0.937219946754)*x[2]**o+(-0.248916422131)*x[2]
        arg[1,0,0,1]=(-0.639865978203)*x[0]**o+(-0.00990375501714)*x[0]+(-0.038266357411)*x[1]**o+(-0.741522380136)*x[1]+(0.460572922772)*x[2]**o+(0.4314205896)*x[2]
        arg[1,0,1,0]=(-0.460110413486)*x[0]**o+(-0.0964934735182)*x[0]+(0.474209024352)*x[1]**o+(-0.928476110805)*x[1]+(0.598621411487)*x[2]**o+(-0.00858812754345)*x[2]
        arg[1,0,1,1]=(0.936466172807)*x[0]**o+(-0.363937334197)*x[0]+(-0.540224651799)*x[1]**o+(0.484530722263)*x[1]+(0.675416993296)*x[2]**o+(0.89657344343)*x[2]
        arg[1,0,2,0]=(-0.713050692555)*x[0]**o+(-0.00851892179229)*x[0]+(0.689494113809)*x[1]**o+(0.633341028044)*x[1]+(0.272726911113)*x[2]**o+(-0.122017566819)*x[2]
        arg[1,0,2,1]=(0.793152080981)*x[0]**o+(0.287379739639)*x[0]+(0.304047372771)*x[1]**o+(-0.509079361053)*x[1]+(0.888302006985)*x[2]**o+(-0.297023045775)*x[2]
        arg[1,1,0,0]=(-0.69695118901)*x[0]**o+(0.505182668512)*x[0]+(0.827555590952)*x[1]**o+(-0.777232447056)*x[1]+(-0.460918138036)*x[2]**o+(0.782090206922)*x[2]
        arg[1,1,0,1]=(-0.435398535566)*x[0]**o+(-0.0997593681984)*x[0]+(0.140250252691)*x[1]**o+(-0.0204469256488)*x[1]+(-0.445836771028)*x[2]**o+(-0.775366037861)*x[2]
        arg[1,1,1,0]=(0.918015347957)*x[0]**o+(-0.491940111489)*x[0]+(0.793071617663)*x[1]**o+(-0.607035294859)*x[1]+(0.610076618566)*x[2]**o+(0.148116733513)*x[2]
        arg[1,1,1,1]=(0.224062518811)*x[0]**o+(-0.145897156735)*x[0]+(0.165380866901)*x[1]**o+(0.910562981814)*x[1]+(-0.406620561548)*x[2]**o+(0.528862812468)*x[2]
        arg[1,1,2,0]=(0.85148444093)*x[0]**o+(0.13210280653)*x[0]+(0.364136256712)*x[1]**o+(0.520388575386)*x[1]+(0.0145641287346)*x[2]**o+(-0.7849036351)*x[2]
        arg[1,1,2,1]=(0.493837193832)*x[0]**o+(-0.193326176569)*x[0]+(-0.665920972154)*x[1]**o+(-0.321645070095)*x[1]+(0.0209354853869)*x[2]**o+(-0.328315981752)*x[2]
        arg[1,2,0,0]=(-0.349281164766)*x[0]**o+(-0.318415198113)*x[0]+(-0.687972781234)*x[1]**o+(0.91825998605)*x[1]+(0.396868773798)*x[2]**o+(0.998204458027)*x[2]
        arg[1,2,0,1]=(-0.967493222935)*x[0]**o+(-0.721128465653)*x[0]+(-0.672178802974)*x[1]**o+(0.0924145224431)*x[1]+(0.114040587594)*x[2]**o+(0.707623895537)*x[2]
        arg[1,2,1,0]=(-0.253327088773)*x[0]**o+(0.206063827625)*x[0]+(0.355152654895)*x[1]**o+(-0.27339285621)*x[1]+(-0.0135567122291)*x[2]**o+(-0.983990446362)*x[2]
        arg[1,2,1,1]=(0.98471687132)*x[0]**o+(0.262902745655)*x[0]+(-0.589359138227)*x[1]**o+(-0.0982548958556)*x[1]+(-0.526161200133)*x[2]**o+(-0.850939484752)*x[2]
        arg[1,2,2,0]=(0.37583596705)*x[0]**o+(0.274845981222)*x[0]+(0.601331110943)*x[1]**o+(0.788294448218)*x[1]+(0.366445166459)*x[2]**o+(-0.554935072394)*x[2]
        arg[1,2,2,1]=(-0.986387850237)*x[0]**o+(0.68772989781)*x[0]+(0.345166604933)*x[1]**o+(0.930376726384)*x[1]+(0.760974421701)*x[2]**o+(0.261797396808)*x[2]
        arg[1,3,0,0]=(-0.171005013523)*x[0]**o+(-0.339229596812)*x[0]+(-0.150317419556)*x[1]**o+(-0.0146533427463)*x[1]+(0.957020663875)*x[2]**o+(0.616513940347)*x[2]
        arg[1,3,0,1]=(-0.0727328269499)*x[0]**o+(-0.635591748073)*x[0]+(-0.502426989774)*x[1]**o+(0.772595119619)*x[1]+(-0.76006511271)*x[2]**o+(-0.665920288084)*x[2]
        arg[1,3,1,0]=(-0.394284205622)*x[0]**o+(-0.105959165811)*x[0]+(0.72649045701)*x[1]**o+(-0.970276379473)*x[1]+(-0.236520349415)*x[2]**o+(0.994973267783)*x[2]
        arg[1,3,1,1]=(-0.444149867906)*x[0]**o+(0.748449201475)*x[0]+(0.419604170756)*x[1]**o+(-0.0291702705178)*x[1]+(-0.109580142628)*x[2]**o+(-0.0791870699838)*x[2]
        arg[1,3,2,0]=(-0.629657820523)*x[0]**o+(-0.995576552624)*x[0]+(0.530135462206)*x[1]**o+(0.197336335845)*x[1]+(0.0224815558807)*x[2]**o+(0.240498463912)*x[2]
        arg[1,3,2,1]=(0.52933948346)*x[0]**o+(0.98353913011)*x[0]+(0.0217614943138)*x[1]**o+(0.0157778453682)*x[1]+(0.177956932986)*x[2]**o+(0.234629778706)*x[2]
        arg[1,4,0,0]=(0.593028367137)*x[0]**o+(0.0340892538994)*x[0]+(-0.0295133170051)*x[1]**o+(0.0865921041963)*x[1]+(0.641131718656)*x[2]**o+(0.788858018609)*x[2]
        arg[1,4,0,1]=(-0.283333820634)*x[0]**o+(0.864482725099)*x[0]+(-0.560218794286)*x[1]**o+(-0.0083115362546)*x[1]+(0.726423285129)*x[2]**o+(0.031531339166)*x[2]
        arg[1,4,1,0]=(0.801452411554)*x[0]**o+(-0.59916393038)*x[0]+(-0.106568863591)*x[1]**o+(0.357442125133)*x[1]+(-0.203380280132)*x[2]**o+(-0.129031606441)*x[2]
        arg[1,4,1,1]=(-0.589220928637)*x[0]**o+(-0.308949389428)*x[0]+(0.785549858338)*x[1]**o+(-0.131153853015)*x[1]+(-0.30571855555)*x[2]**o+(-0.329634802588)*x[2]
        arg[1,4,2,0]=(-0.479686944)*x[0]**o+(-0.707702026873)*x[0]+(0.635894172352)*x[1]**o+(0.370673635069)*x[1]+(0.942862875991)*x[2]**o+(-0.846943456738)*x[2]
        arg[1,4,2,1]=(-0.403737324968)*x[0]**o+(0.101664426734)*x[0]+(-0.677093482354)*x[1]**o+(-0.36526066745)*x[1]+(-0.0889619136182)*x[2]**o+(-0.772491940844)*x[2]
        arg[2,0,0,0]=(0.753148325987)*x[0]**o+(-0.342843083128)*x[0]+(0.812661029553)*x[1]**o+(-0.112683035459)*x[1]+(0.900976470275)*x[2]**o+(0.00934086955102)*x[2]
        arg[2,0,0,1]=(-0.682315007143)*x[0]**o+(-0.694069900861)*x[0]+(-0.469082247747)*x[1]**o+(-0.522165903301)*x[1]+(-0.0997886221924)*x[2]**o+(-0.954161552419)*x[2]
        arg[2,0,1,0]=(0.780125539705)*x[0]**o+(-0.445331450225)*x[0]+(-0.282765510424)*x[1]**o+(-0.0131898361592)*x[1]+(-0.200186433923)*x[2]**o+(0.715139435696)*x[2]
        arg[2,0,1,1]=(-0.132310660786)*x[0]**o+(-0.0981344234369)*x[0]+(0.959297529475)*x[1]**o+(-0.85944894515)*x[1]+(-0.396258659619)*x[2]**o+(-0.09672066185)*x[2]
        arg[2,0,2,0]=(-0.344134014191)*x[0]**o+(0.534345864287)*x[0]+(0.797209509484)*x[1]**o+(0.0479050101499)*x[1]+(0.13620476987)*x[2]**o+(0.82708115383)*x[2]
        arg[2,0,2,1]=(-0.342509328238)*x[0]**o+(-0.833164493006)*x[0]+(0.472047721532)*x[1]**o+(0.301613158637)*x[1]+(0.858691679638)*x[2]**o+(-0.691744164006)*x[2]
        arg[2,1,0,0]=(0.498979390299)*x[0]**o+(0.371513881824)*x[0]+(-0.243506167971)*x[1]**o+(-0.245423890439)*x[1]+(-0.0494707840274)*x[2]**o+(0.0965573709928)*x[2]
        arg[2,1,0,1]=(0.665982455862)*x[0]**o+(-0.409710866336)*x[0]+(0.669257568046)*x[1]**o+(0.711947704788)*x[1]+(-0.0362568635785)*x[2]**o+(-0.30394278188)*x[2]
        arg[2,1,1,0]=(0.296245249928)*x[0]**o+(0.132631019281)*x[0]+(-0.376880152158)*x[1]**o+(0.303184156241)*x[1]+(0.836046560882)*x[2]**o+(0.195205645106)*x[2]
        arg[2,1,1,1]=(-0.971045820953)*x[0]**o+(0.154708896467)*x[0]+(-0.663844656602)*x[1]**o+(-0.0873797588245)*x[1]+(0.985123372542)*x[2]**o+(0.349554922131)*x[2]
        arg[2,1,2,0]=(-0.0117443431348)*x[0]**o+(-0.71980503985)*x[0]+(0.175348078994)*x[1]**o+(-0.883987892161)*x[1]+(0.0587275450458)*x[2]**o+(0.464811715548)*x[2]
        arg[2,1,2,1]=(-0.309202575765)*x[0]**o+(-0.528513248453)*x[0]+(-0.509298303972)*x[1]**o+(-0.228278863496)*x[1]+(0.933917053676)*x[2]**o+(-0.823678942264)*x[2]
        arg[2,2,0,0]=(0.374387386929)*x[0]**o+(0.64969993847)*x[0]+(-0.939822635641)*x[1]**o+(0.578249430422)*x[1]+(-0.973412660635)*x[2]**o+(0.484052718386)*x[2]
        arg[2,2,0,1]=(-0.972635162618)*x[0]**o+(-0.255007433802)*x[0]+(0.134165187097)*x[1]**o+(0.313213615976)*x[1]+(0.143298148686)*x[2]**o+(0.901819440229)*x[2]
        arg[2,2,1,0]=(0.868173256315)*x[0]**o+(-0.532325343453)*x[0]+(-0.712500880845)*x[1]**o+(0.216943462308)*x[1]+(0.733920091212)*x[2]**o+(0.0992400603383)*x[2]
        arg[2,2,1,1]=(0.642839350527)*x[0]**o+(0.127721543559)*x[0]+(-0.152686523032)*x[1]**o+(-0.399032361234)*x[1]+(0.317014518842)*x[2]**o+(0.905538790115)*x[2]
        arg[2,2,2,0]=(0.685267932413)*x[0]**o+(-0.167571547782)*x[0]+(-0.213446144405)*x[1]**o+(0.420270456377)*x[1]+(0.827725350621)*x[2]**o+(-0.920242274399)*x[2]
        arg[2,2,2,1]=(-0.507424593437)*x[0]**o+(-0.694222840019)*x[0]+(-0.904108970188)*x[1]**o+(-0.750899614966)*x[1]+(0.783997553302)*x[2]**o+(0.196327513447)*x[2]
        arg[2,3,0,0]=(-0.473900600479)*x[0]**o+(-0.464347097789)*x[0]+(-0.472423358671)*x[1]**o+(0.373786456364)*x[1]+(-0.989371300414)*x[2]**o+(-0.959508065239)*x[2]
        arg[2,3,0,1]=(0.174375189399)*x[0]**o+(0.925503059886)*x[0]+(-0.647274384723)*x[1]**o+(-0.577149479679)*x[1]+(0.171298877768)*x[2]**o+(-0.936612100882)*x[2]
        arg[2,3,1,0]=(0.482959343396)*x[0]**o+(0.842243554011)*x[0]+(0.814423633314)*x[1]**o+(-0.989445050315)*x[1]+(0.817933433065)*x[2]**o+(0.552396881404)*x[2]
        arg[2,3,1,1]=(0.55641736229)*x[0]**o+(0.498094640058)*x[0]+(0.0890113606037)*x[1]**o+(0.36592723553)*x[1]+(0.458607593386)*x[2]**o+(0.603047420063)*x[2]
        arg[2,3,2,0]=(0.884515080124)*x[0]**o+(-0.561898381022)*x[0]+(-0.292475503834)*x[1]**o+(0.425634423478)*x[1]+(0.65813323252)*x[2]**o+(-0.548850399842)*x[2]
        arg[2,3,2,1]=(-0.72895566982)*x[0]**o+(-0.569405588005)*x[0]+(0.117770875594)*x[1]**o+(-0.33515735969)*x[1]+(0.102062736382)*x[2]**o+(0.611532417683)*x[2]
        arg[2,4,0,0]=(-0.669208630424)*x[0]**o+(0.944393014177)*x[0]+(-0.00345511889683)*x[1]**o+(0.0695279595783)*x[1]+(0.496948250833)*x[2]**o+(-0.899750463971)*x[2]
        arg[2,4,0,1]=(-0.387828717292)*x[0]**o+(0.621172267004)*x[0]+(-0.280632674111)*x[1]**o+(0.401702567575)*x[1]+(-0.65554489378)*x[2]**o+(0.403018047007)*x[2]
        arg[2,4,1,0]=(-0.0770246925803)*x[0]**o+(0.52913302745)*x[0]+(0.922468724357)*x[1]**o+(-0.304442750439)*x[1]+(-0.556575079059)*x[2]**o+(-0.725790672251)*x[2]
        arg[2,4,1,1]=(-0.828378342911)*x[0]**o+(-0.0696370821776)*x[0]+(0.394018581667)*x[1]**o+(-0.251099848269)*x[1]+(-0.29878985656)*x[2]**o+(0.925753371056)*x[2]
        arg[2,4,2,0]=(-0.818277786029)*x[0]**o+(0.225862354706)*x[0]+(-0.629050212595)*x[1]**o+(0.139186107062)*x[1]+(-0.0134835519238)*x[2]**o+(0.301899528641)*x[2]
        arg[2,4,2,1]=(-0.783053684767)*x[0]**o+(-0.0926938783599)*x[0]+(0.519034210352)*x[1]**o+(-0.494564657126)*x[1]+(0.536117310622)*x[2]**o+(0.6278756634)*x[2]
        arg[3,0,0,0]=(-0.694353957685)*x[0]**o+(-0.361037363965)*x[0]+(-0.327004091885)*x[1]**o+(-0.765735142884)*x[1]+(-0.810915729713)*x[2]**o+(0.0640539311609)*x[2]
        arg[3,0,0,1]=(-0.131601548312)*x[0]**o+(0.271587756708)*x[0]+(0.322486168005)*x[1]**o+(0.238086581801)*x[1]+(-0.488081720555)*x[2]**o+(0.607278995635)*x[2]
        arg[3,0,1,0]=(0.903075418102)*x[0]**o+(-0.738609834848)*x[0]+(-0.165097989126)*x[1]**o+(-0.25967140914)*x[1]+(0.271649887428)*x[2]**o+(-0.786131648727)*x[2]
        arg[3,0,1,1]=(-0.963442989157)*x[0]**o+(-0.240205602334)*x[0]+(0.102153389731)*x[1]**o+(-0.106402434995)*x[1]+(-0.383097724211)*x[2]**o+(0.680290963195)*x[2]
        arg[3,0,2,0]=(-0.373162835796)*x[0]**o+(-0.186323761814)*x[0]+(0.661494315705)*x[1]**o+(0.320072189494)*x[1]+(0.0964443431067)*x[2]**o+(0.829015838241)*x[2]
        arg[3,0,2,1]=(0.527719514862)*x[0]**o+(0.0819948261665)*x[0]+(0.826598712178)*x[1]**o+(-0.588102975125)*x[1]+(-0.632584402292)*x[2]**o+(0.0826992233702)*x[2]
        arg[3,1,0,0]=(-0.102243686287)*x[0]**o+(-0.497336112929)*x[0]+(0.941827592628)*x[1]**o+(0.68678991944)*x[1]+(-0.508684841667)*x[2]**o+(-0.752939965161)*x[2]
        arg[3,1,0,1]=(-0.896417523183)*x[0]**o+(0.759224781685)*x[0]+(0.138537695823)*x[1]**o+(0.466727480471)*x[1]+(-0.838380824412)*x[2]**o+(0.285761157398)*x[2]
        arg[3,1,1,0]=(-0.547853182919)*x[0]**o+(-0.41595413414)*x[0]+(0.76894221584)*x[1]**o+(-0.142394275192)*x[1]+(-0.96349294581)*x[2]**o+(-0.873639964693)*x[2]
        arg[3,1,1,1]=(-0.151299403888)*x[0]**o+(0.789406063878)*x[0]+(0.163280670253)*x[1]**o+(-0.287315494745)*x[1]+(-0.627764670218)*x[2]**o+(0.842131632598)*x[2]
        arg[3,1,2,0]=(-0.957096489189)*x[0]**o+(0.0616037805677)*x[0]+(0.255858200166)*x[1]**o+(-0.430555450798)*x[1]+(0.730436930737)*x[2]**o+(0.405094271653)*x[2]
        arg[3,1,2,1]=(-0.924756832665)*x[0]**o+(0.525754162391)*x[0]+(0.281303228038)*x[1]**o+(-0.479500161878)*x[1]+(0.658416209181)*x[2]**o+(-0.0413881595557)*x[2]
        arg[3,2,0,0]=(0.516429381997)*x[0]**o+(-0.8387318467)*x[0]+(0.329773978227)*x[1]**o+(0.599641753085)*x[1]+(0.939937743271)*x[2]**o+(-0.247331169626)*x[2]
        arg[3,2,0,1]=(-0.644982044558)*x[0]**o+(-0.242033352774)*x[0]+(0.531835185779)*x[1]**o+(-0.851565064183)*x[1]+(-0.79213624915)*x[2]**o+(0.0607365690732)*x[2]
        arg[3,2,1,0]=(0.359781383537)*x[0]**o+(-0.0529212842074)*x[0]+(0.845915169452)*x[1]**o+(-0.323979570458)*x[1]+(0.828909997631)*x[2]**o+(-0.839886339138)*x[2]
        arg[3,2,1,1]=(0.559476630421)*x[0]**o+(-0.189725214615)*x[0]+(-0.00927446700413)*x[1]**o+(-0.234703720782)*x[1]+(-0.608362806755)*x[2]**o+(0.968571765534)*x[2]
        arg[3,2,2,0]=(0.924837507913)*x[0]**o+(0.846397611121)*x[0]+(-0.837765472403)*x[1]**o+(0.0225105985057)*x[1]+(-0.834187513844)*x[2]**o+(0.194311845578)*x[2]
        arg[3,2,2,1]=(-0.183561089961)*x[0]**o+(0.0745071729471)*x[0]+(-0.660415615202)*x[1]**o+(-0.528184997345)*x[1]+(0.205554094557)*x[2]**o+(-0.0309246318944)*x[2]
        arg[3,3,0,0]=(-0.67124095974)*x[0]**o+(-0.48501497083)*x[0]+(-0.0394345869094)*x[1]**o+(0.586692270429)*x[1]+(-0.593760749687)*x[2]**o+(-0.124979751327)*x[2]
        arg[3,3,0,1]=(0.778544173855)*x[0]**o+(-0.931466478494)*x[0]+(-0.746894777998)*x[1]**o+(0.806417936297)*x[1]+(0.989460929848)*x[2]**o+(0.666534366369)*x[2]
        arg[3,3,1,0]=(-0.798870838602)*x[0]**o+(-0.294618868937)*x[0]+(0.275029092968)*x[1]**o+(0.45390580364)*x[1]+(-0.311797818508)*x[2]**o+(0.751606624393)*x[2]
        arg[3,3,1,1]=(-0.924260987477)*x[0]**o+(0.771970726902)*x[0]+(-0.909099224213)*x[1]**o+(-0.45204206766)*x[1]+(0.630042614017)*x[2]**o+(-0.990915130464)*x[2]
        arg[3,3,2,0]=(0.734962154852)*x[0]**o+(-0.00621898011475)*x[0]+(-0.786362527189)*x[1]**o+(0.623507947406)*x[1]+(0.867175567104)*x[2]**o+(-0.383283255448)*x[2]
        arg[3,3,2,1]=(-0.955613156717)*x[0]**o+(0.688013754841)*x[0]+(-0.600559375208)*x[1]**o+(0.453342626645)*x[1]+(-0.864890477961)*x[2]**o+(-0.458029873063)*x[2]
        arg[3,4,0,0]=(-0.897010884834)*x[0]**o+(-0.251223576458)*x[0]+(-0.0338052047964)*x[1]**o+(-0.903421466132)*x[1]+(0.836864217874)*x[2]**o+(0.660805905175)*x[2]
        arg[3,4,0,1]=(-0.286025690215)*x[0]**o+(0.746732975604)*x[0]+(0.375810853218)*x[1]**o+(-0.985192974869)*x[1]+(-0.731463214455)*x[2]**o+(-0.465450352444)*x[2]
        arg[3,4,1,0]=(-0.500910000652)*x[0]**o+(0.387502419727)*x[0]+(0.962369164802)*x[1]**o+(-0.10933824251)*x[1]+(-0.33642703989)*x[2]**o+(0.546851133256)*x[2]
        arg[3,4,1,1]=(-0.117959459645)*x[0]**o+(0.867699323392)*x[0]+(0.950506403861)*x[1]**o+(0.613806207745)*x[1]+(0.425723173807)*x[2]**o+(-0.906937548619)*x[2]
        arg[3,4,2,0]=(0.331114892194)*x[0]**o+(-0.168829643424)*x[0]+(-0.855962311982)*x[1]**o+(-0.144202406673)*x[1]+(0.0769985713088)*x[2]**o+(0.856003746423)*x[2]
        arg[3,4,2,1]=(-0.439397439364)*x[0]**o+(0.686116343481)*x[0]+(0.519883854523)*x[1]**o+(0.716231469289)*x[1]+(0.270098231149)*x[2]**o+(-0.448525436427)*x[2]
        ref[0,0,0,0]=(-0.401901998082)*x_ref[0]**o+(-0.805091116808)*x_ref[0]+(0.997659882792)*x_ref[1]**o+(-0.953957389492)*x_ref[1]+(-0.450185408289)*x_ref[2]**o+(0.636472207773)*x_ref[2]
        ref[0,0,0,1]=(-0.786386817128)*x_ref[0]**o+(-0.882559775393)*x_ref[0]+(0.117871400819)*x_ref[1]**o+(-0.625922274774)*x_ref[1]+(0.410431886931)*x_ref[2]**o+(-0.979356501067)*x_ref[2]
        ref[0,0,1,0]=(0.686025409275)*x_ref[0]**o+(-0.19806907659)*x_ref[0]+(-0.710420224449)*x_ref[1]**o+(-0.839236953983)*x_ref[1]+(-0.79971896093)*x_ref[2]**o+(0.378944111319)*x_ref[2]
        ref[0,0,1,1]=(0.771205112484)*x_ref[0]**o+(-0.007083192859)*x_ref[0]+(0.672918588001)*x_ref[1]**o+(-0.351652495793)*x_ref[1]+(-0.841767161187)*x_ref[2]**o+(-0.345642373014)*x_ref[2]
        ref[0,0,2,0]=(0.266930729015)*x_ref[0]**o+(-0.873808152307)*x_ref[0]+(0.206466759135)*x_ref[1]**o+(0.539888649948)*x_ref[1]+(0.17985862191)*x_ref[2]**o+(0.240651460153)*x_ref[2]
        ref[0,0,2,1]=(-0.232569768041)*x_ref[0]**o+(-0.579841454072)*x_ref[0]+(-0.706956104891)*x_ref[1]**o+(0.822863815698)*x_ref[1]+(0.345774991358)*x_ref[2]**o+(-0.0848567668311)*x_ref[2]
        ref[0,1,0,0]=(0.51593851507)*x_ref[0]**o+(-0.233342434349)*x_ref[0]+(-0.92498298026)*x_ref[1]**o+(0.164540234829)*x_ref[1]+(0.356559744499)*x_ref[2]**o+(0.483902511717)*x_ref[2]
        ref[0,1,0,1]=(-0.811509654443)*x_ref[0]**o+(0.73929453189)*x_ref[0]+(0.256205706418)*x_ref[1]**o+(-0.667924913793)*x_ref[1]+(0.266065641257)*x_ref[2]**o+(0.856490258029)*x_ref[2]
        ref[0,1,1,0]=(0.502051590284)*x_ref[0]**o+(0.927283797128)*x_ref[0]+(0.500854521411)*x_ref[1]**o+(-0.486326996369)*x_ref[1]+(-0.327578523685)*x_ref[2]**o+(-0.28106070397)*x_ref[2]
        ref[0,1,1,1]=(0.201740014513)*x_ref[0]**o+(-0.166988241759)*x_ref[0]+(-0.513072955566)*x_ref[1]**o+(0.67179532417)*x_ref[1]+(-0.72135980627)*x_ref[2]**o+(0.763548886467)*x_ref[2]
        ref[0,1,2,0]=(0.3995590338)*x_ref[0]**o+(0.433286929621)*x_ref[0]+(-0.287309453792)*x_ref[1]**o+(-0.0671679443074)*x_ref[1]+(-0.80815399993)*x_ref[2]**o+(-0.398740362114)*x_ref[2]
        ref[0,1,2,1]=(-0.16314640684)*x_ref[0]**o+(-0.698681697367)*x_ref[0]+(-0.11609738418)*x_ref[1]**o+(-0.798465154739)*x_ref[1]+(-0.169052628971)*x_ref[2]**o+(-0.880316962539)*x_ref[2]
        ref[0,2,0,0]=(-0.219731279442)*x_ref[0]**o+(0.731464233112)*x_ref[0]+(0.637811533993)*x_ref[1]**o+(0.381218512692)*x_ref[1]+(-0.181357790048)*x_ref[2]**o+(-0.296793024723)*x_ref[2]
        ref[0,2,0,1]=(0.566451624572)*x_ref[0]**o+(0.302904935559)*x_ref[0]+(-0.771526665232)*x_ref[1]**o+(0.460577683302)*x_ref[1]+(-0.921312455507)*x_ref[2]**o+(0.604299765394)*x_ref[2]
        ref[0,2,1,0]=(-0.0887516895197)*x_ref[0]**o+(0.872208558285)*x_ref[0]+(-0.306564750999)*x_ref[1]**o+(0.841982500263)*x_ref[1]+(-0.550398737046)*x_ref[2]**o+(-0.869697162427)*x_ref[2]
        ref[0,2,1,1]=(-0.652471383912)*x_ref[0]**o+(0.634592304393)*x_ref[0]+(0.0709212473167)*x_ref[1]**o+(-0.671330082579)*x_ref[1]+(-0.232895004478)*x_ref[2]**o+(0.479008351165)*x_ref[2]
        ref[0,2,2,0]=(-0.0764115046136)*x_ref[0]**o+(0.649230346025)*x_ref[0]+(0.544503480999)*x_ref[1]**o+(0.788838981797)*x_ref[1]+(0.416236148267)*x_ref[2]**o+(0.722979541918)*x_ref[2]
        ref[0,2,2,1]=(-0.0776368628837)*x_ref[0]**o+(0.801984674386)*x_ref[0]+(0.443670313486)*x_ref[1]**o+(0.424676792255)*x_ref[1]+(-0.895989937982)*x_ref[2]**o+(0.0774648143589)*x_ref[2]
        ref[0,3,0,0]=(-0.39041318438)*x_ref[0]**o+(-0.174335409262)*x_ref[0]+(0.658666482337)*x_ref[1]**o+(0.23060605921)*x_ref[1]+(-0.336603364419)*x_ref[2]**o+(0.871530192198)*x_ref[2]
        ref[0,3,0,1]=(-0.493104661363)*x_ref[0]**o+(0.590779697499)*x_ref[0]+(-0.903897457443)*x_ref[1]**o+(-0.377640520704)*x_ref[1]+(0.227174544833)*x_ref[2]**o+(-0.860150483539)*x_ref[2]
        ref[0,3,1,0]=(-0.093367253954)*x_ref[0]**o+(0.159230499811)*x_ref[0]+(-0.731396509936)*x_ref[1]**o+(0.600088148838)*x_ref[1]+(0.829175660147)*x_ref[2]**o+(-0.435878280919)*x_ref[2]
        ref[0,3,1,1]=(0.665125540711)*x_ref[0]**o+(-0.92128737118)*x_ref[0]+(-0.33260297543)*x_ref[1]**o+(0.682569476816)*x_ref[1]+(0.999761405058)*x_ref[2]**o+(-0.748303702618)*x_ref[2]
        ref[0,3,2,0]=(-0.430442277525)*x_ref[0]**o+(-0.841437664378)*x_ref[0]+(-0.374267371055)*x_ref[1]**o+(0.667860327373)*x_ref[1]+(0.694185699238)*x_ref[2]**o+(-0.673629629622)*x_ref[2]
        ref[0,3,2,1]=(0.304808617243)*x_ref[0]**o+(-0.894407776037)*x_ref[0]+(-0.758585281699)*x_ref[1]**o+(0.20519963241)*x_ref[1]+(0.538222015181)*x_ref[2]**o+(-0.792962435573)*x_ref[2]
        ref[0,4,0,0]=(0.568300300269)*x_ref[0]**o+(0.407455936625)*x_ref[0]+(0.980898344535)*x_ref[1]**o+(0.988220474461)*x_ref[1]+(0.0598892620089)*x_ref[2]**o+(-0.279555207723)*x_ref[2]
        ref[0,4,0,1]=(-0.208391854989)*x_ref[0]**o+(0.9428019078)*x_ref[0]+(-0.540916877275)*x_ref[1]**o+(-0.427185538555)*x_ref[1]+(-0.879268085857)*x_ref[2]**o+(0.025102321099)*x_ref[2]
        ref[0,4,1,0]=(0.412690112527)*x_ref[0]**o+(-0.957376863734)*x_ref[0]+(-0.913623139719)*x_ref[1]**o+(-0.462418849979)*x_ref[1]+(0.463336747019)*x_ref[2]**o+(0.330747531293)*x_ref[2]
        ref[0,4,1,1]=(-0.632140552118)*x_ref[0]**o+(0.830769577671)*x_ref[0]+(-0.888012251479)*x_ref[1]**o+(-0.0560181569699)*x_ref[1]+(-0.70068099248)*x_ref[2]**o+(0.834483253039)*x_ref[2]
        ref[0,4,2,0]=(0.0429717567385)*x_ref[0]**o+(-0.691784042595)*x_ref[0]+(-0.264988764713)*x_ref[1]**o+(0.359873170931)*x_ref[1]+(-0.116275128437)*x_ref[2]**o+(0.0348873917345)*x_ref[2]
        ref[0,4,2,1]=(-0.203754358311)*x_ref[0]**o+(0.177896174646)*x_ref[0]+(-0.0785920580727)*x_ref[1]**o+(-0.124905949184)*x_ref[1]+(0.968871355367)*x_ref[2]**o+(-0.0369722247589)*x_ref[2]
        ref[1,0,0,0]=(0.594851756786)*x_ref[0]**o+(-0.594400822936)*x_ref[0]+(0.868401350949)*x_ref[1]**o+(0.186739616297)*x_ref[1]+(-0.937219946754)*x_ref[2]**o+(-0.248916422131)*x_ref[2]
        ref[1,0,0,1]=(-0.639865978203)*x_ref[0]**o+(-0.00990375501714)*x_ref[0]+(-0.038266357411)*x_ref[1]**o+(-0.741522380136)*x_ref[1]+(0.460572922772)*x_ref[2]**o+(0.4314205896)*x_ref[2]
        ref[1,0,1,0]=(-0.460110413486)*x_ref[0]**o+(-0.0964934735182)*x_ref[0]+(0.474209024352)*x_ref[1]**o+(-0.928476110805)*x_ref[1]+(0.598621411487)*x_ref[2]**o+(-0.00858812754345)*x_ref[2]
        ref[1,0,1,1]=(0.936466172807)*x_ref[0]**o+(-0.363937334197)*x_ref[0]+(-0.540224651799)*x_ref[1]**o+(0.484530722263)*x_ref[1]+(0.675416993296)*x_ref[2]**o+(0.89657344343)*x_ref[2]
        ref[1,0,2,0]=(-0.713050692555)*x_ref[0]**o+(-0.00851892179229)*x_ref[0]+(0.689494113809)*x_ref[1]**o+(0.633341028044)*x_ref[1]+(0.272726911113)*x_ref[2]**o+(-0.122017566819)*x_ref[2]
        ref[1,0,2,1]=(0.793152080981)*x_ref[0]**o+(0.287379739639)*x_ref[0]+(0.304047372771)*x_ref[1]**o+(-0.509079361053)*x_ref[1]+(0.888302006985)*x_ref[2]**o+(-0.297023045775)*x_ref[2]
        ref[1,1,0,0]=(-0.69695118901)*x_ref[0]**o+(0.505182668512)*x_ref[0]+(0.827555590952)*x_ref[1]**o+(-0.777232447056)*x_ref[1]+(-0.460918138036)*x_ref[2]**o+(0.782090206922)*x_ref[2]
        ref[1,1,0,1]=(-0.435398535566)*x_ref[0]**o+(-0.0997593681984)*x_ref[0]+(0.140250252691)*x_ref[1]**o+(-0.0204469256488)*x_ref[1]+(-0.445836771028)*x_ref[2]**o+(-0.775366037861)*x_ref[2]
        ref[1,1,1,0]=(0.918015347957)*x_ref[0]**o+(-0.491940111489)*x_ref[0]+(0.793071617663)*x_ref[1]**o+(-0.607035294859)*x_ref[1]+(0.610076618566)*x_ref[2]**o+(0.148116733513)*x_ref[2]
        ref[1,1,1,1]=(0.224062518811)*x_ref[0]**o+(-0.145897156735)*x_ref[0]+(0.165380866901)*x_ref[1]**o+(0.910562981814)*x_ref[1]+(-0.406620561548)*x_ref[2]**o+(0.528862812468)*x_ref[2]
        ref[1,1,2,0]=(0.85148444093)*x_ref[0]**o+(0.13210280653)*x_ref[0]+(0.364136256712)*x_ref[1]**o+(0.520388575386)*x_ref[1]+(0.0145641287346)*x_ref[2]**o+(-0.7849036351)*x_ref[2]
        ref[1,1,2,1]=(0.493837193832)*x_ref[0]**o+(-0.193326176569)*x_ref[0]+(-0.665920972154)*x_ref[1]**o+(-0.321645070095)*x_ref[1]+(0.0209354853869)*x_ref[2]**o+(-0.328315981752)*x_ref[2]
        ref[1,2,0,0]=(-0.349281164766)*x_ref[0]**o+(-0.318415198113)*x_ref[0]+(-0.687972781234)*x_ref[1]**o+(0.91825998605)*x_ref[1]+(0.396868773798)*x_ref[2]**o+(0.998204458027)*x_ref[2]
        ref[1,2,0,1]=(-0.967493222935)*x_ref[0]**o+(-0.721128465653)*x_ref[0]+(-0.672178802974)*x_ref[1]**o+(0.0924145224431)*x_ref[1]+(0.114040587594)*x_ref[2]**o+(0.707623895537)*x_ref[2]
        ref[1,2,1,0]=(-0.253327088773)*x_ref[0]**o+(0.206063827625)*x_ref[0]+(0.355152654895)*x_ref[1]**o+(-0.27339285621)*x_ref[1]+(-0.0135567122291)*x_ref[2]**o+(-0.983990446362)*x_ref[2]
        ref[1,2,1,1]=(0.98471687132)*x_ref[0]**o+(0.262902745655)*x_ref[0]+(-0.589359138227)*x_ref[1]**o+(-0.0982548958556)*x_ref[1]+(-0.526161200133)*x_ref[2]**o+(-0.850939484752)*x_ref[2]
        ref[1,2,2,0]=(0.37583596705)*x_ref[0]**o+(0.274845981222)*x_ref[0]+(0.601331110943)*x_ref[1]**o+(0.788294448218)*x_ref[1]+(0.366445166459)*x_ref[2]**o+(-0.554935072394)*x_ref[2]
        ref[1,2,2,1]=(-0.986387850237)*x_ref[0]**o+(0.68772989781)*x_ref[0]+(0.345166604933)*x_ref[1]**o+(0.930376726384)*x_ref[1]+(0.760974421701)*x_ref[2]**o+(0.261797396808)*x_ref[2]
        ref[1,3,0,0]=(-0.171005013523)*x_ref[0]**o+(-0.339229596812)*x_ref[0]+(-0.150317419556)*x_ref[1]**o+(-0.0146533427463)*x_ref[1]+(0.957020663875)*x_ref[2]**o+(0.616513940347)*x_ref[2]
        ref[1,3,0,1]=(-0.0727328269499)*x_ref[0]**o+(-0.635591748073)*x_ref[0]+(-0.502426989774)*x_ref[1]**o+(0.772595119619)*x_ref[1]+(-0.76006511271)*x_ref[2]**o+(-0.665920288084)*x_ref[2]
        ref[1,3,1,0]=(-0.394284205622)*x_ref[0]**o+(-0.105959165811)*x_ref[0]+(0.72649045701)*x_ref[1]**o+(-0.970276379473)*x_ref[1]+(-0.236520349415)*x_ref[2]**o+(0.994973267783)*x_ref[2]
        ref[1,3,1,1]=(-0.444149867906)*x_ref[0]**o+(0.748449201475)*x_ref[0]+(0.419604170756)*x_ref[1]**o+(-0.0291702705178)*x_ref[1]+(-0.109580142628)*x_ref[2]**o+(-0.0791870699838)*x_ref[2]
        ref[1,3,2,0]=(-0.629657820523)*x_ref[0]**o+(-0.995576552624)*x_ref[0]+(0.530135462206)*x_ref[1]**o+(0.197336335845)*x_ref[1]+(0.0224815558807)*x_ref[2]**o+(0.240498463912)*x_ref[2]
        ref[1,3,2,1]=(0.52933948346)*x_ref[0]**o+(0.98353913011)*x_ref[0]+(0.0217614943138)*x_ref[1]**o+(0.0157778453682)*x_ref[1]+(0.177956932986)*x_ref[2]**o+(0.234629778706)*x_ref[2]
        ref[1,4,0,0]=(0.593028367137)*x_ref[0]**o+(0.0340892538994)*x_ref[0]+(-0.0295133170051)*x_ref[1]**o+(0.0865921041963)*x_ref[1]+(0.641131718656)*x_ref[2]**o+(0.788858018609)*x_ref[2]
        ref[1,4,0,1]=(-0.283333820634)*x_ref[0]**o+(0.864482725099)*x_ref[0]+(-0.560218794286)*x_ref[1]**o+(-0.0083115362546)*x_ref[1]+(0.726423285129)*x_ref[2]**o+(0.031531339166)*x_ref[2]
        ref[1,4,1,0]=(0.801452411554)*x_ref[0]**o+(-0.59916393038)*x_ref[0]+(-0.106568863591)*x_ref[1]**o+(0.357442125133)*x_ref[1]+(-0.203380280132)*x_ref[2]**o+(-0.129031606441)*x_ref[2]
        ref[1,4,1,1]=(-0.589220928637)*x_ref[0]**o+(-0.308949389428)*x_ref[0]+(0.785549858338)*x_ref[1]**o+(-0.131153853015)*x_ref[1]+(-0.30571855555)*x_ref[2]**o+(-0.329634802588)*x_ref[2]
        ref[1,4,2,0]=(-0.479686944)*x_ref[0]**o+(-0.707702026873)*x_ref[0]+(0.635894172352)*x_ref[1]**o+(0.370673635069)*x_ref[1]+(0.942862875991)*x_ref[2]**o+(-0.846943456738)*x_ref[2]
        ref[1,4,2,1]=(-0.403737324968)*x_ref[0]**o+(0.101664426734)*x_ref[0]+(-0.677093482354)*x_ref[1]**o+(-0.36526066745)*x_ref[1]+(-0.0889619136182)*x_ref[2]**o+(-0.772491940844)*x_ref[2]
        ref[2,0,0,0]=(0.753148325987)*x_ref[0]**o+(-0.342843083128)*x_ref[0]+(0.812661029553)*x_ref[1]**o+(-0.112683035459)*x_ref[1]+(0.900976470275)*x_ref[2]**o+(0.00934086955102)*x_ref[2]
        ref[2,0,0,1]=(-0.682315007143)*x_ref[0]**o+(-0.694069900861)*x_ref[0]+(-0.469082247747)*x_ref[1]**o+(-0.522165903301)*x_ref[1]+(-0.0997886221924)*x_ref[2]**o+(-0.954161552419)*x_ref[2]
        ref[2,0,1,0]=(0.780125539705)*x_ref[0]**o+(-0.445331450225)*x_ref[0]+(-0.282765510424)*x_ref[1]**o+(-0.0131898361592)*x_ref[1]+(-0.200186433923)*x_ref[2]**o+(0.715139435696)*x_ref[2]
        ref[2,0,1,1]=(-0.132310660786)*x_ref[0]**o+(-0.0981344234369)*x_ref[0]+(0.959297529475)*x_ref[1]**o+(-0.85944894515)*x_ref[1]+(-0.396258659619)*x_ref[2]**o+(-0.09672066185)*x_ref[2]
        ref[2,0,2,0]=(-0.344134014191)*x_ref[0]**o+(0.534345864287)*x_ref[0]+(0.797209509484)*x_ref[1]**o+(0.0479050101499)*x_ref[1]+(0.13620476987)*x_ref[2]**o+(0.82708115383)*x_ref[2]
        ref[2,0,2,1]=(-0.342509328238)*x_ref[0]**o+(-0.833164493006)*x_ref[0]+(0.472047721532)*x_ref[1]**o+(0.301613158637)*x_ref[1]+(0.858691679638)*x_ref[2]**o+(-0.691744164006)*x_ref[2]
        ref[2,1,0,0]=(0.498979390299)*x_ref[0]**o+(0.371513881824)*x_ref[0]+(-0.243506167971)*x_ref[1]**o+(-0.245423890439)*x_ref[1]+(-0.0494707840274)*x_ref[2]**o+(0.0965573709928)*x_ref[2]
        ref[2,1,0,1]=(0.665982455862)*x_ref[0]**o+(-0.409710866336)*x_ref[0]+(0.669257568046)*x_ref[1]**o+(0.711947704788)*x_ref[1]+(-0.0362568635785)*x_ref[2]**o+(-0.30394278188)*x_ref[2]
        ref[2,1,1,0]=(0.296245249928)*x_ref[0]**o+(0.132631019281)*x_ref[0]+(-0.376880152158)*x_ref[1]**o+(0.303184156241)*x_ref[1]+(0.836046560882)*x_ref[2]**o+(0.195205645106)*x_ref[2]
        ref[2,1,1,1]=(-0.971045820953)*x_ref[0]**o+(0.154708896467)*x_ref[0]+(-0.663844656602)*x_ref[1]**o+(-0.0873797588245)*x_ref[1]+(0.985123372542)*x_ref[2]**o+(0.349554922131)*x_ref[2]
        ref[2,1,2,0]=(-0.0117443431348)*x_ref[0]**o+(-0.71980503985)*x_ref[0]+(0.175348078994)*x_ref[1]**o+(-0.883987892161)*x_ref[1]+(0.0587275450458)*x_ref[2]**o+(0.464811715548)*x_ref[2]
        ref[2,1,2,1]=(-0.309202575765)*x_ref[0]**o+(-0.528513248453)*x_ref[0]+(-0.509298303972)*x_ref[1]**o+(-0.228278863496)*x_ref[1]+(0.933917053676)*x_ref[2]**o+(-0.823678942264)*x_ref[2]
        ref[2,2,0,0]=(0.374387386929)*x_ref[0]**o+(0.64969993847)*x_ref[0]+(-0.939822635641)*x_ref[1]**o+(0.578249430422)*x_ref[1]+(-0.973412660635)*x_ref[2]**o+(0.484052718386)*x_ref[2]
        ref[2,2,0,1]=(-0.972635162618)*x_ref[0]**o+(-0.255007433802)*x_ref[0]+(0.134165187097)*x_ref[1]**o+(0.313213615976)*x_ref[1]+(0.143298148686)*x_ref[2]**o+(0.901819440229)*x_ref[2]
        ref[2,2,1,0]=(0.868173256315)*x_ref[0]**o+(-0.532325343453)*x_ref[0]+(-0.712500880845)*x_ref[1]**o+(0.216943462308)*x_ref[1]+(0.733920091212)*x_ref[2]**o+(0.0992400603383)*x_ref[2]
        ref[2,2,1,1]=(0.642839350527)*x_ref[0]**o+(0.127721543559)*x_ref[0]+(-0.152686523032)*x_ref[1]**o+(-0.399032361234)*x_ref[1]+(0.317014518842)*x_ref[2]**o+(0.905538790115)*x_ref[2]
        ref[2,2,2,0]=(0.685267932413)*x_ref[0]**o+(-0.167571547782)*x_ref[0]+(-0.213446144405)*x_ref[1]**o+(0.420270456377)*x_ref[1]+(0.827725350621)*x_ref[2]**o+(-0.920242274399)*x_ref[2]
        ref[2,2,2,1]=(-0.507424593437)*x_ref[0]**o+(-0.694222840019)*x_ref[0]+(-0.904108970188)*x_ref[1]**o+(-0.750899614966)*x_ref[1]+(0.783997553302)*x_ref[2]**o+(0.196327513447)*x_ref[2]
        ref[2,3,0,0]=(-0.473900600479)*x_ref[0]**o+(-0.464347097789)*x_ref[0]+(-0.472423358671)*x_ref[1]**o+(0.373786456364)*x_ref[1]+(-0.989371300414)*x_ref[2]**o+(-0.959508065239)*x_ref[2]
        ref[2,3,0,1]=(0.174375189399)*x_ref[0]**o+(0.925503059886)*x_ref[0]+(-0.647274384723)*x_ref[1]**o+(-0.577149479679)*x_ref[1]+(0.171298877768)*x_ref[2]**o+(-0.936612100882)*x_ref[2]
        ref[2,3,1,0]=(0.482959343396)*x_ref[0]**o+(0.842243554011)*x_ref[0]+(0.814423633314)*x_ref[1]**o+(-0.989445050315)*x_ref[1]+(0.817933433065)*x_ref[2]**o+(0.552396881404)*x_ref[2]
        ref[2,3,1,1]=(0.55641736229)*x_ref[0]**o+(0.498094640058)*x_ref[0]+(0.0890113606037)*x_ref[1]**o+(0.36592723553)*x_ref[1]+(0.458607593386)*x_ref[2]**o+(0.603047420063)*x_ref[2]
        ref[2,3,2,0]=(0.884515080124)*x_ref[0]**o+(-0.561898381022)*x_ref[0]+(-0.292475503834)*x_ref[1]**o+(0.425634423478)*x_ref[1]+(0.65813323252)*x_ref[2]**o+(-0.548850399842)*x_ref[2]
        ref[2,3,2,1]=(-0.72895566982)*x_ref[0]**o+(-0.569405588005)*x_ref[0]+(0.117770875594)*x_ref[1]**o+(-0.33515735969)*x_ref[1]+(0.102062736382)*x_ref[2]**o+(0.611532417683)*x_ref[2]
        ref[2,4,0,0]=(-0.669208630424)*x_ref[0]**o+(0.944393014177)*x_ref[0]+(-0.00345511889683)*x_ref[1]**o+(0.0695279595783)*x_ref[1]+(0.496948250833)*x_ref[2]**o+(-0.899750463971)*x_ref[2]
        ref[2,4,0,1]=(-0.387828717292)*x_ref[0]**o+(0.621172267004)*x_ref[0]+(-0.280632674111)*x_ref[1]**o+(0.401702567575)*x_ref[1]+(-0.65554489378)*x_ref[2]**o+(0.403018047007)*x_ref[2]
        ref[2,4,1,0]=(-0.0770246925803)*x_ref[0]**o+(0.52913302745)*x_ref[0]+(0.922468724357)*x_ref[1]**o+(-0.304442750439)*x_ref[1]+(-0.556575079059)*x_ref[2]**o+(-0.725790672251)*x_ref[2]
        ref[2,4,1,1]=(-0.828378342911)*x_ref[0]**o+(-0.0696370821776)*x_ref[0]+(0.394018581667)*x_ref[1]**o+(-0.251099848269)*x_ref[1]+(-0.29878985656)*x_ref[2]**o+(0.925753371056)*x_ref[2]
        ref[2,4,2,0]=(-0.818277786029)*x_ref[0]**o+(0.225862354706)*x_ref[0]+(-0.629050212595)*x_ref[1]**o+(0.139186107062)*x_ref[1]+(-0.0134835519238)*x_ref[2]**o+(0.301899528641)*x_ref[2]
        ref[2,4,2,1]=(-0.783053684767)*x_ref[0]**o+(-0.0926938783599)*x_ref[0]+(0.519034210352)*x_ref[1]**o+(-0.494564657126)*x_ref[1]+(0.536117310622)*x_ref[2]**o+(0.6278756634)*x_ref[2]
        ref[3,0,0,0]=(-0.694353957685)*x_ref[0]**o+(-0.361037363965)*x_ref[0]+(-0.327004091885)*x_ref[1]**o+(-0.765735142884)*x_ref[1]+(-0.810915729713)*x_ref[2]**o+(0.0640539311609)*x_ref[2]
        ref[3,0,0,1]=(-0.131601548312)*x_ref[0]**o+(0.271587756708)*x_ref[0]+(0.322486168005)*x_ref[1]**o+(0.238086581801)*x_ref[1]+(-0.488081720555)*x_ref[2]**o+(0.607278995635)*x_ref[2]
        ref[3,0,1,0]=(0.903075418102)*x_ref[0]**o+(-0.738609834848)*x_ref[0]+(-0.165097989126)*x_ref[1]**o+(-0.25967140914)*x_ref[1]+(0.271649887428)*x_ref[2]**o+(-0.786131648727)*x_ref[2]
        ref[3,0,1,1]=(-0.963442989157)*x_ref[0]**o+(-0.240205602334)*x_ref[0]+(0.102153389731)*x_ref[1]**o+(-0.106402434995)*x_ref[1]+(-0.383097724211)*x_ref[2]**o+(0.680290963195)*x_ref[2]
        ref[3,0,2,0]=(-0.373162835796)*x_ref[0]**o+(-0.186323761814)*x_ref[0]+(0.661494315705)*x_ref[1]**o+(0.320072189494)*x_ref[1]+(0.0964443431067)*x_ref[2]**o+(0.829015838241)*x_ref[2]
        ref[3,0,2,1]=(0.527719514862)*x_ref[0]**o+(0.0819948261665)*x_ref[0]+(0.826598712178)*x_ref[1]**o+(-0.588102975125)*x_ref[1]+(-0.632584402292)*x_ref[2]**o+(0.0826992233702)*x_ref[2]
        ref[3,1,0,0]=(-0.102243686287)*x_ref[0]**o+(-0.497336112929)*x_ref[0]+(0.941827592628)*x_ref[1]**o+(0.68678991944)*x_ref[1]+(-0.508684841667)*x_ref[2]**o+(-0.752939965161)*x_ref[2]
        ref[3,1,0,1]=(-0.896417523183)*x_ref[0]**o+(0.759224781685)*x_ref[0]+(0.138537695823)*x_ref[1]**o+(0.466727480471)*x_ref[1]+(-0.838380824412)*x_ref[2]**o+(0.285761157398)*x_ref[2]
        ref[3,1,1,0]=(-0.547853182919)*x_ref[0]**o+(-0.41595413414)*x_ref[0]+(0.76894221584)*x_ref[1]**o+(-0.142394275192)*x_ref[1]+(-0.96349294581)*x_ref[2]**o+(-0.873639964693)*x_ref[2]
        ref[3,1,1,1]=(-0.151299403888)*x_ref[0]**o+(0.789406063878)*x_ref[0]+(0.163280670253)*x_ref[1]**o+(-0.287315494745)*x_ref[1]+(-0.627764670218)*x_ref[2]**o+(0.842131632598)*x_ref[2]
        ref[3,1,2,0]=(-0.957096489189)*x_ref[0]**o+(0.0616037805677)*x_ref[0]+(0.255858200166)*x_ref[1]**o+(-0.430555450798)*x_ref[1]+(0.730436930737)*x_ref[2]**o+(0.405094271653)*x_ref[2]
        ref[3,1,2,1]=(-0.924756832665)*x_ref[0]**o+(0.525754162391)*x_ref[0]+(0.281303228038)*x_ref[1]**o+(-0.479500161878)*x_ref[1]+(0.658416209181)*x_ref[2]**o+(-0.0413881595557)*x_ref[2]
        ref[3,2,0,0]=(0.516429381997)*x_ref[0]**o+(-0.8387318467)*x_ref[0]+(0.329773978227)*x_ref[1]**o+(0.599641753085)*x_ref[1]+(0.939937743271)*x_ref[2]**o+(-0.247331169626)*x_ref[2]
        ref[3,2,0,1]=(-0.644982044558)*x_ref[0]**o+(-0.242033352774)*x_ref[0]+(0.531835185779)*x_ref[1]**o+(-0.851565064183)*x_ref[1]+(-0.79213624915)*x_ref[2]**o+(0.0607365690732)*x_ref[2]
        ref[3,2,1,0]=(0.359781383537)*x_ref[0]**o+(-0.0529212842074)*x_ref[0]+(0.845915169452)*x_ref[1]**o+(-0.323979570458)*x_ref[1]+(0.828909997631)*x_ref[2]**o+(-0.839886339138)*x_ref[2]
        ref[3,2,1,1]=(0.559476630421)*x_ref[0]**o+(-0.189725214615)*x_ref[0]+(-0.00927446700413)*x_ref[1]**o+(-0.234703720782)*x_ref[1]+(-0.608362806755)*x_ref[2]**o+(0.968571765534)*x_ref[2]
        ref[3,2,2,0]=(0.924837507913)*x_ref[0]**o+(0.846397611121)*x_ref[0]+(-0.837765472403)*x_ref[1]**o+(0.0225105985057)*x_ref[1]+(-0.834187513844)*x_ref[2]**o+(0.194311845578)*x_ref[2]
        ref[3,2,2,1]=(-0.183561089961)*x_ref[0]**o+(0.0745071729471)*x_ref[0]+(-0.660415615202)*x_ref[1]**o+(-0.528184997345)*x_ref[1]+(0.205554094557)*x_ref[2]**o+(-0.0309246318944)*x_ref[2]
        ref[3,3,0,0]=(-0.67124095974)*x_ref[0]**o+(-0.48501497083)*x_ref[0]+(-0.0394345869094)*x_ref[1]**o+(0.586692270429)*x_ref[1]+(-0.593760749687)*x_ref[2]**o+(-0.124979751327)*x_ref[2]
        ref[3,3,0,1]=(0.778544173855)*x_ref[0]**o+(-0.931466478494)*x_ref[0]+(-0.746894777998)*x_ref[1]**o+(0.806417936297)*x_ref[1]+(0.989460929848)*x_ref[2]**o+(0.666534366369)*x_ref[2]
        ref[3,3,1,0]=(-0.798870838602)*x_ref[0]**o+(-0.294618868937)*x_ref[0]+(0.275029092968)*x_ref[1]**o+(0.45390580364)*x_ref[1]+(-0.311797818508)*x_ref[2]**o+(0.751606624393)*x_ref[2]
        ref[3,3,1,1]=(-0.924260987477)*x_ref[0]**o+(0.771970726902)*x_ref[0]+(-0.909099224213)*x_ref[1]**o+(-0.45204206766)*x_ref[1]+(0.630042614017)*x_ref[2]**o+(-0.990915130464)*x_ref[2]
        ref[3,3,2,0]=(0.734962154852)*x_ref[0]**o+(-0.00621898011475)*x_ref[0]+(-0.786362527189)*x_ref[1]**o+(0.623507947406)*x_ref[1]+(0.867175567104)*x_ref[2]**o+(-0.383283255448)*x_ref[2]
        ref[3,3,2,1]=(-0.955613156717)*x_ref[0]**o+(0.688013754841)*x_ref[0]+(-0.600559375208)*x_ref[1]**o+(0.453342626645)*x_ref[1]+(-0.864890477961)*x_ref[2]**o+(-0.458029873063)*x_ref[2]
        ref[3,4,0,0]=(-0.897010884834)*x_ref[0]**o+(-0.251223576458)*x_ref[0]+(-0.0338052047964)*x_ref[1]**o+(-0.903421466132)*x_ref[1]+(0.836864217874)*x_ref[2]**o+(0.660805905175)*x_ref[2]
        ref[3,4,0,1]=(-0.286025690215)*x_ref[0]**o+(0.746732975604)*x_ref[0]+(0.375810853218)*x_ref[1]**o+(-0.985192974869)*x_ref[1]+(-0.731463214455)*x_ref[2]**o+(-0.465450352444)*x_ref[2]
        ref[3,4,1,0]=(-0.500910000652)*x_ref[0]**o+(0.387502419727)*x_ref[0]+(0.962369164802)*x_ref[1]**o+(-0.10933824251)*x_ref[1]+(-0.33642703989)*x_ref[2]**o+(0.546851133256)*x_ref[2]
        ref[3,4,1,1]=(-0.117959459645)*x_ref[0]**o+(0.867699323392)*x_ref[0]+(0.950506403861)*x_ref[1]**o+(0.613806207745)*x_ref[1]+(0.425723173807)*x_ref[2]**o+(-0.906937548619)*x_ref[2]
        ref[3,4,2,0]=(0.331114892194)*x_ref[0]**o+(-0.168829643424)*x_ref[0]+(-0.855962311982)*x_ref[1]**o+(-0.144202406673)*x_ref[1]+(0.0769985713088)*x_ref[2]**o+(0.856003746423)*x_ref[2]
        ref[3,4,2,1]=(-0.439397439364)*x_ref[0]**o+(0.686116343481)*x_ref[0]+(0.519883854523)*x_ref[1]**o+(0.716231469289)*x_ref[1]+(0.270098231149)*x_ref[2]**o+(-0.448525436427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.203238386406)*x[0]**o+(0.734542238717)*x[0]+(-0.627248060056)*x[1]**o+(0.347224669854)*x[1]
        ref=(0.203238386406)*x_ref[0]**o+(0.734542238717)*x_ref[0]+(-0.627248060056)*x_ref[1]**o+(0.347224669854)*x_ref[1]
      else:
        arg=(-0.387501489724)*x[0]**o+(0.752554332333)*x[0]+(-0.0831588271355)*x[1]**o+(0.615191870812)*x[1]+(-0.264702923622)*x[2]**o+(0.404594629993)*x[2]
        ref=(-0.387501489724)*x_ref[0]**o+(0.752554332333)*x_ref[0]+(-0.0831588271355)*x_ref[1]**o+(0.615191870812)*x_ref[1]+(-0.264702923622)*x_ref[2]**o+(0.404594629993)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.938444792136)*x[0]**o+(-0.809407040724)*x[0]+(-0.963148615762)*x[1]**o+(-0.257000244129)*x[1]
        arg[1]=(-0.791280640039)*x[0]**o+(0.293858623095)*x[0]+(0.964942589784)*x[1]**o+(0.522071353081)*x[1]
        ref[0]=(0.938444792136)*x_ref[0]**o+(-0.809407040724)*x_ref[0]+(-0.963148615762)*x_ref[1]**o+(-0.257000244129)*x_ref[1]
        ref[1]=(-0.791280640039)*x_ref[0]**o+(0.293858623095)*x_ref[0]+(0.964942589784)*x_ref[1]**o+(0.522071353081)*x_ref[1]
      else:
        arg[0]=(0.566110580666)*x[0]**o+(-0.89723913213)*x[0]+(-0.54781629706)*x[1]**o+(-0.0405507445628)*x[1]+(0.435477036501)*x[2]**o+(-0.296404435183)*x[2]
        arg[1]=(-0.436126359475)*x[0]**o+(0.863706244241)*x[0]+(-0.149818409243)*x[1]**o+(-0.654801920098)*x[1]+(-0.281005092809)*x[2]**o+(-0.0257179420057)*x[2]
        ref[0]=(0.566110580666)*x_ref[0]**o+(-0.89723913213)*x_ref[0]+(-0.54781629706)*x_ref[1]**o+(-0.0405507445628)*x_ref[1]+(0.435477036501)*x_ref[2]**o+(-0.296404435183)*x_ref[2]
        ref[1]=(-0.436126359475)*x_ref[0]**o+(0.863706244241)*x_ref[0]+(-0.149818409243)*x_ref[1]**o+(-0.654801920098)*x_ref[1]+(-0.281005092809)*x_ref[2]**o+(-0.0257179420057)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0615713819144)*x[0]**o+(0.325133297443)*x[0]+(-0.529769594257)*x[1]**o+(-0.206126939263)*x[1]
        arg[0,1]=(0.090956391947)*x[0]**o+(0.105841035331)*x[0]+(-0.537033610998)*x[1]**o+(0.429033781495)*x[1]
        arg[0,2]=(-0.568216263139)*x[0]**o+(0.698057796806)*x[0]+(-0.0981614063428)*x[1]**o+(-0.952212332523)*x[1]
        arg[0,3]=(0.128505813477)*x[0]**o+(0.206828305852)*x[0]+(0.938851873532)*x[1]**o+(-0.339234589281)*x[1]
        arg[0,4]=(0.581805477855)*x[0]**o+(-0.621122347153)*x[0]+(-0.601029625565)*x[1]**o+(0.243084366725)*x[1]
        arg[1,0]=(-0.342411991491)*x[0]**o+(0.642434320731)*x[0]+(-0.452657094515)*x[1]**o+(-0.899779109444)*x[1]
        arg[1,1]=(-0.423230603077)*x[0]**o+(0.739600521332)*x[0]+(0.916226480419)*x[1]**o+(0.755264173162)*x[1]
        arg[1,2]=(0.381864221314)*x[0]**o+(0.806402187902)*x[0]+(0.410055116458)*x[1]**o+(-0.264405213472)*x[1]
        arg[1,3]=(0.293401427429)*x[0]**o+(-0.439209675154)*x[0]+(0.299588226145)*x[1]**o+(-0.289652201976)*x[1]
        arg[1,4]=(0.157626502865)*x[0]**o+(0.32264788209)*x[0]+(0.284522039757)*x[1]**o+(-0.460413671191)*x[1]
        arg[2,0]=(0.486148207)*x[0]**o+(0.0366950072658)*x[0]+(0.315299830901)*x[1]**o+(-0.320405042262)*x[1]
        arg[2,1]=(0.207440897542)*x[0]**o+(-0.604659890692)*x[0]+(-0.64841431916)*x[1]**o+(0.0881921209641)*x[1]
        arg[2,2]=(-0.61376840347)*x[0]**o+(0.421837679185)*x[0]+(-0.0445850531916)*x[1]**o+(0.771892018903)*x[1]
        arg[2,3]=(0.632443182263)*x[0]**o+(-0.965297013012)*x[0]+(-0.029677959681)*x[1]**o+(0.598431910551)*x[1]
        arg[2,4]=(-0.457656411094)*x[0]**o+(0.122435925199)*x[0]+(0.772690251937)*x[1]**o+(0.954829922926)*x[1]
        arg[3,0]=(0.929112925461)*x[0]**o+(-0.35689901655)*x[0]+(-0.139192866782)*x[1]**o+(-0.677883436952)*x[1]
        arg[3,1]=(-0.410318721103)*x[0]**o+(0.370830312699)*x[0]+(0.429011619434)*x[1]**o+(0.81780420854)*x[1]
        arg[3,2]=(0.312456682985)*x[0]**o+(0.526447664999)*x[0]+(0.763610570932)*x[1]**o+(-0.631887901017)*x[1]
        arg[3,3]=(0.749121567955)*x[0]**o+(0.168525340464)*x[0]+(-0.838409619895)*x[1]**o+(0.912745137098)*x[1]
        arg[3,4]=(-0.665336604643)*x[0]**o+(0.638293226479)*x[0]+(0.515957084712)*x[1]**o+(-0.261426720977)*x[1]
        ref[0,0]=(0.0615713819144)*x_ref[0]**o+(0.325133297443)*x_ref[0]+(-0.529769594257)*x_ref[1]**o+(-0.206126939263)*x_ref[1]
        ref[0,1]=(0.090956391947)*x_ref[0]**o+(0.105841035331)*x_ref[0]+(-0.537033610998)*x_ref[1]**o+(0.429033781495)*x_ref[1]
        ref[0,2]=(-0.568216263139)*x_ref[0]**o+(0.698057796806)*x_ref[0]+(-0.0981614063428)*x_ref[1]**o+(-0.952212332523)*x_ref[1]
        ref[0,3]=(0.128505813477)*x_ref[0]**o+(0.206828305852)*x_ref[0]+(0.938851873532)*x_ref[1]**o+(-0.339234589281)*x_ref[1]
        ref[0,4]=(0.581805477855)*x_ref[0]**o+(-0.621122347153)*x_ref[0]+(-0.601029625565)*x_ref[1]**o+(0.243084366725)*x_ref[1]
        ref[1,0]=(-0.342411991491)*x_ref[0]**o+(0.642434320731)*x_ref[0]+(-0.452657094515)*x_ref[1]**o+(-0.899779109444)*x_ref[1]
        ref[1,1]=(-0.423230603077)*x_ref[0]**o+(0.739600521332)*x_ref[0]+(0.916226480419)*x_ref[1]**o+(0.755264173162)*x_ref[1]
        ref[1,2]=(0.381864221314)*x_ref[0]**o+(0.806402187902)*x_ref[0]+(0.410055116458)*x_ref[1]**o+(-0.264405213472)*x_ref[1]
        ref[1,3]=(0.293401427429)*x_ref[0]**o+(-0.439209675154)*x_ref[0]+(0.299588226145)*x_ref[1]**o+(-0.289652201976)*x_ref[1]
        ref[1,4]=(0.157626502865)*x_ref[0]**o+(0.32264788209)*x_ref[0]+(0.284522039757)*x_ref[1]**o+(-0.460413671191)*x_ref[1]
        ref[2,0]=(0.486148207)*x_ref[0]**o+(0.0366950072658)*x_ref[0]+(0.315299830901)*x_ref[1]**o+(-0.320405042262)*x_ref[1]
        ref[2,1]=(0.207440897542)*x_ref[0]**o+(-0.604659890692)*x_ref[0]+(-0.64841431916)*x_ref[1]**o+(0.0881921209641)*x_ref[1]
        ref[2,2]=(-0.61376840347)*x_ref[0]**o+(0.421837679185)*x_ref[0]+(-0.0445850531916)*x_ref[1]**o+(0.771892018903)*x_ref[1]
        ref[2,3]=(0.632443182263)*x_ref[0]**o+(-0.965297013012)*x_ref[0]+(-0.029677959681)*x_ref[1]**o+(0.598431910551)*x_ref[1]
        ref[2,4]=(-0.457656411094)*x_ref[0]**o+(0.122435925199)*x_ref[0]+(0.772690251937)*x_ref[1]**o+(0.954829922926)*x_ref[1]
        ref[3,0]=(0.929112925461)*x_ref[0]**o+(-0.35689901655)*x_ref[0]+(-0.139192866782)*x_ref[1]**o+(-0.677883436952)*x_ref[1]
        ref[3,1]=(-0.410318721103)*x_ref[0]**o+(0.370830312699)*x_ref[0]+(0.429011619434)*x_ref[1]**o+(0.81780420854)*x_ref[1]
        ref[3,2]=(0.312456682985)*x_ref[0]**o+(0.526447664999)*x_ref[0]+(0.763610570932)*x_ref[1]**o+(-0.631887901017)*x_ref[1]
        ref[3,3]=(0.749121567955)*x_ref[0]**o+(0.168525340464)*x_ref[0]+(-0.838409619895)*x_ref[1]**o+(0.912745137098)*x_ref[1]
        ref[3,4]=(-0.665336604643)*x_ref[0]**o+(0.638293226479)*x_ref[0]+(0.515957084712)*x_ref[1]**o+(-0.261426720977)*x_ref[1]
      else:
        arg[0,0]=(-0.330165086386)*x[0]**o+(0.60381119065)*x[0]+(0.981018763011)*x[1]**o+(0.0487965999608)*x[1]+(-0.389835388349)*x[2]**o+(0.111199857376)*x[2]
        arg[0,1]=(0.518239033178)*x[0]**o+(0.60358784684)*x[0]+(-0.476990159655)*x[1]**o+(-0.93871565843)*x[1]+(-0.869845295741)*x[2]**o+(0.391409745802)*x[2]
        arg[0,2]=(0.0927396115951)*x[0]**o+(0.546122822928)*x[0]+(-0.557424056733)*x[1]**o+(0.0623374966436)*x[1]+(-0.414595153536)*x[2]**o+(-0.196069073366)*x[2]
        arg[0,3]=(-0.678432580882)*x[0]**o+(0.281850265141)*x[0]+(-0.323978960602)*x[1]**o+(0.655558769921)*x[1]+(-0.295747032089)*x[2]**o+(-0.152719838982)*x[2]
        arg[0,4]=(-0.904177669203)*x[0]**o+(0.750395234205)*x[0]+(0.267431783461)*x[1]**o+(-0.781406235306)*x[1]+(0.856906655757)*x[2]**o+(-0.875554940224)*x[2]
        arg[1,0]=(-0.243144846569)*x[0]**o+(0.970584922467)*x[0]+(-0.977250652473)*x[1]**o+(0.0239563255869)*x[1]+(-0.769423602851)*x[2]**o+(-0.433896361294)*x[2]
        arg[1,1]=(0.31104031046)*x[0]**o+(0.899859709831)*x[0]+(0.692351561919)*x[1]**o+(-0.0838870550002)*x[1]+(-0.276105616798)*x[2]**o+(0.115501348248)*x[2]
        arg[1,2]=(0.56702114595)*x[0]**o+(0.00240284646131)*x[0]+(-0.676314888884)*x[1]**o+(-0.678204346328)*x[1]+(0.472697088984)*x[2]**o+(-0.493949191712)*x[2]
        arg[1,3]=(-0.242548554067)*x[0]**o+(0.980927507053)*x[0]+(-0.955706390436)*x[1]**o+(0.347031550126)*x[1]+(0.234352268238)*x[2]**o+(0.321126345333)*x[2]
        arg[1,4]=(-0.390890566228)*x[0]**o+(-0.958497753144)*x[0]+(0.316397716043)*x[1]**o+(0.952289821769)*x[1]+(-0.027182029514)*x[2]**o+(-0.454422377872)*x[2]
        arg[2,0]=(0.295865166318)*x[0]**o+(-0.301129718468)*x[0]+(-0.748135755269)*x[1]**o+(-0.762413814085)*x[1]+(0.572270067591)*x[2]**o+(-0.393847032684)*x[2]
        arg[2,1]=(0.20229063622)*x[0]**o+(-0.768444028163)*x[0]+(0.378669775381)*x[1]**o+(-0.703384410944)*x[1]+(0.355653859396)*x[2]**o+(-0.228168949639)*x[2]
        arg[2,2]=(-0.31725134742)*x[0]**o+(0.525802867508)*x[0]+(0.13375126758)*x[1]**o+(-0.673551491186)*x[1]+(0.498905883728)*x[2]**o+(-0.102794461221)*x[2]
        arg[2,3]=(-0.396813838168)*x[0]**o+(0.0892605807843)*x[0]+(0.244133298111)*x[1]**o+(0.310900636168)*x[1]+(-0.256807063422)*x[2]**o+(0.565370697688)*x[2]
        arg[2,4]=(0.42011835881)*x[0]**o+(0.439847594375)*x[0]+(0.0731897272006)*x[1]**o+(-0.373911054753)*x[1]+(-0.50313315963)*x[2]**o+(0.356362677601)*x[2]
        arg[3,0]=(-0.444672845607)*x[0]**o+(-0.978424347535)*x[0]+(0.421699748442)*x[1]**o+(-0.0111798084161)*x[1]+(0.206042685881)*x[2]**o+(0.990523831215)*x[2]
        arg[3,1]=(0.942429627068)*x[0]**o+(-0.769989484282)*x[0]+(-0.879916738075)*x[1]**o+(-0.746448112871)*x[1]+(0.280824650861)*x[2]**o+(0.0632084823022)*x[2]
        arg[3,2]=(0.281070481177)*x[0]**o+(0.628020001876)*x[0]+(0.347767303962)*x[1]**o+(-0.276761252142)*x[1]+(0.950345260734)*x[2]**o+(-0.00148186335195)*x[2]
        arg[3,3]=(-0.37590210465)*x[0]**o+(0.034147350068)*x[0]+(0.245173418336)*x[1]**o+(0.0564051241759)*x[1]+(-0.157928106127)*x[2]**o+(-0.611677441847)*x[2]
        arg[3,4]=(-0.279066532276)*x[0]**o+(0.316033119063)*x[0]+(0.655948999472)*x[1]**o+(-0.520504180943)*x[1]+(0.531738158819)*x[2]**o+(0.899490827443)*x[2]
        ref[0,0]=(-0.330165086386)*x_ref[0]**o+(0.60381119065)*x_ref[0]+(0.981018763011)*x_ref[1]**o+(0.0487965999608)*x_ref[1]+(-0.389835388349)*x_ref[2]**o+(0.111199857376)*x_ref[2]
        ref[0,1]=(0.518239033178)*x_ref[0]**o+(0.60358784684)*x_ref[0]+(-0.476990159655)*x_ref[1]**o+(-0.93871565843)*x_ref[1]+(-0.869845295741)*x_ref[2]**o+(0.391409745802)*x_ref[2]
        ref[0,2]=(0.0927396115951)*x_ref[0]**o+(0.546122822928)*x_ref[0]+(-0.557424056733)*x_ref[1]**o+(0.0623374966436)*x_ref[1]+(-0.414595153536)*x_ref[2]**o+(-0.196069073366)*x_ref[2]
        ref[0,3]=(-0.678432580882)*x_ref[0]**o+(0.281850265141)*x_ref[0]+(-0.323978960602)*x_ref[1]**o+(0.655558769921)*x_ref[1]+(-0.295747032089)*x_ref[2]**o+(-0.152719838982)*x_ref[2]
        ref[0,4]=(-0.904177669203)*x_ref[0]**o+(0.750395234205)*x_ref[0]+(0.267431783461)*x_ref[1]**o+(-0.781406235306)*x_ref[1]+(0.856906655757)*x_ref[2]**o+(-0.875554940224)*x_ref[2]
        ref[1,0]=(-0.243144846569)*x_ref[0]**o+(0.970584922467)*x_ref[0]+(-0.977250652473)*x_ref[1]**o+(0.0239563255869)*x_ref[1]+(-0.769423602851)*x_ref[2]**o+(-0.433896361294)*x_ref[2]
        ref[1,1]=(0.31104031046)*x_ref[0]**o+(0.899859709831)*x_ref[0]+(0.692351561919)*x_ref[1]**o+(-0.0838870550002)*x_ref[1]+(-0.276105616798)*x_ref[2]**o+(0.115501348248)*x_ref[2]
        ref[1,2]=(0.56702114595)*x_ref[0]**o+(0.00240284646131)*x_ref[0]+(-0.676314888884)*x_ref[1]**o+(-0.678204346328)*x_ref[1]+(0.472697088984)*x_ref[2]**o+(-0.493949191712)*x_ref[2]
        ref[1,3]=(-0.242548554067)*x_ref[0]**o+(0.980927507053)*x_ref[0]+(-0.955706390436)*x_ref[1]**o+(0.347031550126)*x_ref[1]+(0.234352268238)*x_ref[2]**o+(0.321126345333)*x_ref[2]
        ref[1,4]=(-0.390890566228)*x_ref[0]**o+(-0.958497753144)*x_ref[0]+(0.316397716043)*x_ref[1]**o+(0.952289821769)*x_ref[1]+(-0.027182029514)*x_ref[2]**o+(-0.454422377872)*x_ref[2]
        ref[2,0]=(0.295865166318)*x_ref[0]**o+(-0.301129718468)*x_ref[0]+(-0.748135755269)*x_ref[1]**o+(-0.762413814085)*x_ref[1]+(0.572270067591)*x_ref[2]**o+(-0.393847032684)*x_ref[2]
        ref[2,1]=(0.20229063622)*x_ref[0]**o+(-0.768444028163)*x_ref[0]+(0.378669775381)*x_ref[1]**o+(-0.703384410944)*x_ref[1]+(0.355653859396)*x_ref[2]**o+(-0.228168949639)*x_ref[2]
        ref[2,2]=(-0.31725134742)*x_ref[0]**o+(0.525802867508)*x_ref[0]+(0.13375126758)*x_ref[1]**o+(-0.673551491186)*x_ref[1]+(0.498905883728)*x_ref[2]**o+(-0.102794461221)*x_ref[2]
        ref[2,3]=(-0.396813838168)*x_ref[0]**o+(0.0892605807843)*x_ref[0]+(0.244133298111)*x_ref[1]**o+(0.310900636168)*x_ref[1]+(-0.256807063422)*x_ref[2]**o+(0.565370697688)*x_ref[2]
        ref[2,4]=(0.42011835881)*x_ref[0]**o+(0.439847594375)*x_ref[0]+(0.0731897272006)*x_ref[1]**o+(-0.373911054753)*x_ref[1]+(-0.50313315963)*x_ref[2]**o+(0.356362677601)*x_ref[2]
        ref[3,0]=(-0.444672845607)*x_ref[0]**o+(-0.978424347535)*x_ref[0]+(0.421699748442)*x_ref[1]**o+(-0.0111798084161)*x_ref[1]+(0.206042685881)*x_ref[2]**o+(0.990523831215)*x_ref[2]
        ref[3,1]=(0.942429627068)*x_ref[0]**o+(-0.769989484282)*x_ref[0]+(-0.879916738075)*x_ref[1]**o+(-0.746448112871)*x_ref[1]+(0.280824650861)*x_ref[2]**o+(0.0632084823022)*x_ref[2]
        ref[3,2]=(0.281070481177)*x_ref[0]**o+(0.628020001876)*x_ref[0]+(0.347767303962)*x_ref[1]**o+(-0.276761252142)*x_ref[1]+(0.950345260734)*x_ref[2]**o+(-0.00148186335195)*x_ref[2]
        ref[3,3]=(-0.37590210465)*x_ref[0]**o+(0.034147350068)*x_ref[0]+(0.245173418336)*x_ref[1]**o+(0.0564051241759)*x_ref[1]+(-0.157928106127)*x_ref[2]**o+(-0.611677441847)*x_ref[2]
        ref[3,4]=(-0.279066532276)*x_ref[0]**o+(0.316033119063)*x_ref[0]+(0.655948999472)*x_ref[1]**o+(-0.520504180943)*x_ref[1]+(0.531738158819)*x_ref[2]**o+(0.899490827443)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.415438516191)*x[0]**o+(0.374659323173)*x[0]+(0.688470746611)*x[1]**o+(0.498183692176)*x[1]
        arg[0,0,1]=(-0.846725329522)*x[0]**o+(0.414376816326)*x[0]+(0.651609661615)*x[1]**o+(0.357606396667)*x[1]
        arg[0,1,0]=(0.728029172023)*x[0]**o+(0.975728459112)*x[0]+(-0.658211371639)*x[1]**o+(0.411894140351)*x[1]
        arg[0,1,1]=(0.807350281817)*x[0]**o+(-0.369755733522)*x[0]+(-0.519483139314)*x[1]**o+(-0.709353537648)*x[1]
        arg[1,0,0]=(0.0588259329479)*x[0]**o+(0.218362870441)*x[0]+(-0.0279239318934)*x[1]**o+(-0.75783635329)*x[1]
        arg[1,0,1]=(-0.429573558505)*x[0]**o+(-0.339046134618)*x[0]+(0.412777825743)*x[1]**o+(-0.306607333437)*x[1]
        arg[1,1,0]=(0.0222661407543)*x[0]**o+(0.927923674781)*x[0]+(-0.767468570033)*x[1]**o+(-0.0659150590367)*x[1]
        arg[1,1,1]=(0.372811733787)*x[0]**o+(-0.286500260061)*x[0]+(-0.807504375569)*x[1]**o+(-0.992850148234)*x[1]
        arg[2,0,0]=(0.368229828596)*x[0]**o+(0.672540290379)*x[0]+(-0.760462595095)*x[1]**o+(-0.761242492425)*x[1]
        arg[2,0,1]=(0.306249523133)*x[0]**o+(0.769928924717)*x[0]+(0.610293800742)*x[1]**o+(-0.221996271153)*x[1]
        arg[2,1,0]=(-0.909148832991)*x[0]**o+(-0.0293038712703)*x[0]+(-0.161707473269)*x[1]**o+(-0.534345319122)*x[1]
        arg[2,1,1]=(-0.146432487737)*x[0]**o+(0.93729266958)*x[0]+(-0.561919238789)*x[1]**o+(-0.449440501171)*x[1]
        arg[3,0,0]=(0.411711661035)*x[0]**o+(-0.0699878047274)*x[0]+(-0.0227688826645)*x[1]**o+(-0.939787633547)*x[1]
        arg[3,0,1]=(-0.82108820245)*x[0]**o+(0.225501168847)*x[0]+(0.622921674682)*x[1]**o+(0.61092413094)*x[1]
        arg[3,1,0]=(0.758116155912)*x[0]**o+(-0.117316615447)*x[0]+(0.0702245119124)*x[1]**o+(-0.275114508561)*x[1]
        arg[3,1,1]=(-0.32568550821)*x[0]**o+(-0.280144846326)*x[0]+(0.194971187876)*x[1]**o+(-0.728447244746)*x[1]
        arg[4,0,0]=(0.880042622981)*x[0]**o+(-0.395684607563)*x[0]+(-0.563304061265)*x[1]**o+(0.968663606772)*x[1]
        arg[4,0,1]=(-0.515392712458)*x[0]**o+(0.820857647342)*x[0]+(-0.476289639495)*x[1]**o+(-0.478062589609)*x[1]
        arg[4,1,0]=(0.534322425269)*x[0]**o+(0.975618695265)*x[0]+(-0.195764005045)*x[1]**o+(0.561996200643)*x[1]
        arg[4,1,1]=(-0.906243138672)*x[0]**o+(-0.580985047294)*x[0]+(-0.094609405051)*x[1]**o+(0.734326711649)*x[1]
        arg[5,0,0]=(-0.0655516121632)*x[0]**o+(0.785493123174)*x[0]+(-0.760068333675)*x[1]**o+(0.693869247554)*x[1]
        arg[5,0,1]=(-0.931019838892)*x[0]**o+(0.80958291621)*x[0]+(-0.590591206111)*x[1]**o+(-0.800900346601)*x[1]
        arg[5,1,0]=(0.448779048952)*x[0]**o+(-0.793450503268)*x[0]+(0.161067403301)*x[1]**o+(0.432054043544)*x[1]
        arg[5,1,1]=(0.0423607361972)*x[0]**o+(-0.664958045702)*x[0]+(-0.174728816918)*x[1]**o+(-0.394390068699)*x[1]
        ref[0,0,0]=(0.415438516191)*x_ref[0]**o+(0.374659323173)*x_ref[0]+(0.688470746611)*x_ref[1]**o+(0.498183692176)*x_ref[1]
        ref[0,0,1]=(-0.846725329522)*x_ref[0]**o+(0.414376816326)*x_ref[0]+(0.651609661615)*x_ref[1]**o+(0.357606396667)*x_ref[1]
        ref[0,1,0]=(0.728029172023)*x_ref[0]**o+(0.975728459112)*x_ref[0]+(-0.658211371639)*x_ref[1]**o+(0.411894140351)*x_ref[1]
        ref[0,1,1]=(0.807350281817)*x_ref[0]**o+(-0.369755733522)*x_ref[0]+(-0.519483139314)*x_ref[1]**o+(-0.709353537648)*x_ref[1]
        ref[1,0,0]=(0.0588259329479)*x_ref[0]**o+(0.218362870441)*x_ref[0]+(-0.0279239318934)*x_ref[1]**o+(-0.75783635329)*x_ref[1]
        ref[1,0,1]=(-0.429573558505)*x_ref[0]**o+(-0.339046134618)*x_ref[0]+(0.412777825743)*x_ref[1]**o+(-0.306607333437)*x_ref[1]
        ref[1,1,0]=(0.0222661407543)*x_ref[0]**o+(0.927923674781)*x_ref[0]+(-0.767468570033)*x_ref[1]**o+(-0.0659150590367)*x_ref[1]
        ref[1,1,1]=(0.372811733787)*x_ref[0]**o+(-0.286500260061)*x_ref[0]+(-0.807504375569)*x_ref[1]**o+(-0.992850148234)*x_ref[1]
        ref[2,0,0]=(0.368229828596)*x_ref[0]**o+(0.672540290379)*x_ref[0]+(-0.760462595095)*x_ref[1]**o+(-0.761242492425)*x_ref[1]
        ref[2,0,1]=(0.306249523133)*x_ref[0]**o+(0.769928924717)*x_ref[0]+(0.610293800742)*x_ref[1]**o+(-0.221996271153)*x_ref[1]
        ref[2,1,0]=(-0.909148832991)*x_ref[0]**o+(-0.0293038712703)*x_ref[0]+(-0.161707473269)*x_ref[1]**o+(-0.534345319122)*x_ref[1]
        ref[2,1,1]=(-0.146432487737)*x_ref[0]**o+(0.93729266958)*x_ref[0]+(-0.561919238789)*x_ref[1]**o+(-0.449440501171)*x_ref[1]
        ref[3,0,0]=(0.411711661035)*x_ref[0]**o+(-0.0699878047274)*x_ref[0]+(-0.0227688826645)*x_ref[1]**o+(-0.939787633547)*x_ref[1]
        ref[3,0,1]=(-0.82108820245)*x_ref[0]**o+(0.225501168847)*x_ref[0]+(0.622921674682)*x_ref[1]**o+(0.61092413094)*x_ref[1]
        ref[3,1,0]=(0.758116155912)*x_ref[0]**o+(-0.117316615447)*x_ref[0]+(0.0702245119124)*x_ref[1]**o+(-0.275114508561)*x_ref[1]
        ref[3,1,1]=(-0.32568550821)*x_ref[0]**o+(-0.280144846326)*x_ref[0]+(0.194971187876)*x_ref[1]**o+(-0.728447244746)*x_ref[1]
        ref[4,0,0]=(0.880042622981)*x_ref[0]**o+(-0.395684607563)*x_ref[0]+(-0.563304061265)*x_ref[1]**o+(0.968663606772)*x_ref[1]
        ref[4,0,1]=(-0.515392712458)*x_ref[0]**o+(0.820857647342)*x_ref[0]+(-0.476289639495)*x_ref[1]**o+(-0.478062589609)*x_ref[1]
        ref[4,1,0]=(0.534322425269)*x_ref[0]**o+(0.975618695265)*x_ref[0]+(-0.195764005045)*x_ref[1]**o+(0.561996200643)*x_ref[1]
        ref[4,1,1]=(-0.906243138672)*x_ref[0]**o+(-0.580985047294)*x_ref[0]+(-0.094609405051)*x_ref[1]**o+(0.734326711649)*x_ref[1]
        ref[5,0,0]=(-0.0655516121632)*x_ref[0]**o+(0.785493123174)*x_ref[0]+(-0.760068333675)*x_ref[1]**o+(0.693869247554)*x_ref[1]
        ref[5,0,1]=(-0.931019838892)*x_ref[0]**o+(0.80958291621)*x_ref[0]+(-0.590591206111)*x_ref[1]**o+(-0.800900346601)*x_ref[1]
        ref[5,1,0]=(0.448779048952)*x_ref[0]**o+(-0.793450503268)*x_ref[0]+(0.161067403301)*x_ref[1]**o+(0.432054043544)*x_ref[1]
        ref[5,1,1]=(0.0423607361972)*x_ref[0]**o+(-0.664958045702)*x_ref[0]+(-0.174728816918)*x_ref[1]**o+(-0.394390068699)*x_ref[1]
      else:
        arg[0,0,0]=(0.34355250678)*x[0]**o+(0.554590991175)*x[0]+(0.00673862595967)*x[1]**o+(-0.405764224977)*x[1]+(0.0865564883661)*x[2]**o+(0.102294137607)*x[2]
        arg[0,0,1]=(0.855234460708)*x[0]**o+(-0.710901486015)*x[0]+(0.546393138087)*x[1]**o+(0.847113479239)*x[1]+(-0.650701604407)*x[2]**o+(-0.914755510262)*x[2]
        arg[0,1,0]=(0.252544568503)*x[0]**o+(-0.473471976093)*x[0]+(-0.587315274764)*x[1]**o+(-0.269940894375)*x[1]+(0.360340735507)*x[2]**o+(-0.745440004827)*x[2]
        arg[0,1,1]=(0.11967506555)*x[0]**o+(0.696836453942)*x[0]+(-0.871162187396)*x[1]**o+(-0.315170111067)*x[1]+(0.232066109266)*x[2]**o+(-0.464730524838)*x[2]
        arg[1,0,0]=(-0.790535662997)*x[0]**o+(-0.964779159884)*x[0]+(-0.689937881374)*x[1]**o+(0.422799231262)*x[1]+(-0.97315874804)*x[2]**o+(-0.532103115205)*x[2]
        arg[1,0,1]=(-0.0487589554945)*x[0]**o+(-0.622735713658)*x[0]+(-0.520842613265)*x[1]**o+(0.803711422599)*x[1]+(0.440648875556)*x[2]**o+(0.98786672574)*x[2]
        arg[1,1,0]=(0.811189977538)*x[0]**o+(0.0175150486788)*x[0]+(-0.124328899976)*x[1]**o+(-0.942398680503)*x[1]+(-0.149270209895)*x[2]**o+(-0.853084407696)*x[2]
        arg[1,1,1]=(0.688378159744)*x[0]**o+(0.408766927878)*x[0]+(0.834128919609)*x[1]**o+(0.253605923886)*x[1]+(0.424992121054)*x[2]**o+(0.21016117963)*x[2]
        arg[2,0,0]=(0.277469067087)*x[0]**o+(0.539821557611)*x[0]+(0.348679402521)*x[1]**o+(-0.469772506537)*x[1]+(0.454583229137)*x[2]**o+(0.653681481133)*x[2]
        arg[2,0,1]=(-0.934067178483)*x[0]**o+(-0.172735098497)*x[0]+(-0.0347256850948)*x[1]**o+(-0.459971424944)*x[1]+(0.932661853019)*x[2]**o+(-0.186822590955)*x[2]
        arg[2,1,0]=(0.71698574321)*x[0]**o+(-0.197244831261)*x[0]+(0.183419576182)*x[1]**o+(-0.871451480853)*x[1]+(-0.349396269955)*x[2]**o+(0.095909246936)*x[2]
        arg[2,1,1]=(-0.00972985962106)*x[0]**o+(-0.919204152811)*x[0]+(-0.751744404365)*x[1]**o+(0.377096436852)*x[1]+(-0.924228885362)*x[2]**o+(0.451053813959)*x[2]
        arg[3,0,0]=(0.951119968373)*x[0]**o+(-0.0680027504528)*x[0]+(-0.113660990287)*x[1]**o+(-0.428999453003)*x[1]+(0.171804544103)*x[2]**o+(0.677635036297)*x[2]
        arg[3,0,1]=(0.62534864581)*x[0]**o+(0.389496246771)*x[0]+(0.621295327948)*x[1]**o+(-0.684165739023)*x[1]+(0.258637161492)*x[2]**o+(0.0441460764076)*x[2]
        arg[3,1,0]=(0.0348922721145)*x[0]**o+(0.091201286767)*x[0]+(0.421079169132)*x[1]**o+(-0.616504862591)*x[1]+(0.563927673176)*x[2]**o+(-0.183993394822)*x[2]
        arg[3,1,1]=(0.657959811783)*x[0]**o+(0.229729671382)*x[0]+(0.549853155207)*x[1]**o+(-0.847412930565)*x[1]+(-0.0780682500337)*x[2]**o+(0.574435563406)*x[2]
        arg[4,0,0]=(-0.679988515344)*x[0]**o+(0.0517331630343)*x[0]+(-0.160281064879)*x[1]**o+(-0.0672405944261)*x[1]+(-0.118968444607)*x[2]**o+(0.00980065834148)*x[2]
        arg[4,0,1]=(-0.150457972951)*x[0]**o+(0.671586511406)*x[0]+(-0.844112675259)*x[1]**o+(0.42766255101)*x[1]+(-0.22449721018)*x[2]**o+(-0.150416151059)*x[2]
        arg[4,1,0]=(0.0768992597753)*x[0]**o+(0.0139605476692)*x[0]+(-0.831269554863)*x[1]**o+(-0.279339344511)*x[1]+(0.387318821185)*x[2]**o+(0.967687121678)*x[2]
        arg[4,1,1]=(-0.670585948598)*x[0]**o+(0.422715818654)*x[0]+(0.623702390878)*x[1]**o+(0.922646081646)*x[1]+(-0.714191287039)*x[2]**o+(-0.99018084869)*x[2]
        arg[5,0,0]=(0.100273819257)*x[0]**o+(-0.769422291863)*x[0]+(-0.152611272068)*x[1]**o+(-0.396798631305)*x[1]+(0.502251047462)*x[2]**o+(0.993791216449)*x[2]
        arg[5,0,1]=(-0.0222540900657)*x[0]**o+(-0.339810353987)*x[0]+(0.241665078696)*x[1]**o+(0.111020568564)*x[1]+(0.298147316931)*x[2]**o+(0.391278660594)*x[2]
        arg[5,1,0]=(0.828297599555)*x[0]**o+(0.237678875371)*x[0]+(0.924551126274)*x[1]**o+(-0.663695049485)*x[1]+(-0.547123493132)*x[2]**o+(-0.788635745118)*x[2]
        arg[5,1,1]=(0.0726334370982)*x[0]**o+(-0.875688492957)*x[0]+(0.473726235108)*x[1]**o+(0.023025596151)*x[1]+(0.827449396402)*x[2]**o+(-0.154698941604)*x[2]
        ref[0,0,0]=(0.34355250678)*x_ref[0]**o+(0.554590991175)*x_ref[0]+(0.00673862595967)*x_ref[1]**o+(-0.405764224977)*x_ref[1]+(0.0865564883661)*x_ref[2]**o+(0.102294137607)*x_ref[2]
        ref[0,0,1]=(0.855234460708)*x_ref[0]**o+(-0.710901486015)*x_ref[0]+(0.546393138087)*x_ref[1]**o+(0.847113479239)*x_ref[1]+(-0.650701604407)*x_ref[2]**o+(-0.914755510262)*x_ref[2]
        ref[0,1,0]=(0.252544568503)*x_ref[0]**o+(-0.473471976093)*x_ref[0]+(-0.587315274764)*x_ref[1]**o+(-0.269940894375)*x_ref[1]+(0.360340735507)*x_ref[2]**o+(-0.745440004827)*x_ref[2]
        ref[0,1,1]=(0.11967506555)*x_ref[0]**o+(0.696836453942)*x_ref[0]+(-0.871162187396)*x_ref[1]**o+(-0.315170111067)*x_ref[1]+(0.232066109266)*x_ref[2]**o+(-0.464730524838)*x_ref[2]
        ref[1,0,0]=(-0.790535662997)*x_ref[0]**o+(-0.964779159884)*x_ref[0]+(-0.689937881374)*x_ref[1]**o+(0.422799231262)*x_ref[1]+(-0.97315874804)*x_ref[2]**o+(-0.532103115205)*x_ref[2]
        ref[1,0,1]=(-0.0487589554945)*x_ref[0]**o+(-0.622735713658)*x_ref[0]+(-0.520842613265)*x_ref[1]**o+(0.803711422599)*x_ref[1]+(0.440648875556)*x_ref[2]**o+(0.98786672574)*x_ref[2]
        ref[1,1,0]=(0.811189977538)*x_ref[0]**o+(0.0175150486788)*x_ref[0]+(-0.124328899976)*x_ref[1]**o+(-0.942398680503)*x_ref[1]+(-0.149270209895)*x_ref[2]**o+(-0.853084407696)*x_ref[2]
        ref[1,1,1]=(0.688378159744)*x_ref[0]**o+(0.408766927878)*x_ref[0]+(0.834128919609)*x_ref[1]**o+(0.253605923886)*x_ref[1]+(0.424992121054)*x_ref[2]**o+(0.21016117963)*x_ref[2]
        ref[2,0,0]=(0.277469067087)*x_ref[0]**o+(0.539821557611)*x_ref[0]+(0.348679402521)*x_ref[1]**o+(-0.469772506537)*x_ref[1]+(0.454583229137)*x_ref[2]**o+(0.653681481133)*x_ref[2]
        ref[2,0,1]=(-0.934067178483)*x_ref[0]**o+(-0.172735098497)*x_ref[0]+(-0.0347256850948)*x_ref[1]**o+(-0.459971424944)*x_ref[1]+(0.932661853019)*x_ref[2]**o+(-0.186822590955)*x_ref[2]
        ref[2,1,0]=(0.71698574321)*x_ref[0]**o+(-0.197244831261)*x_ref[0]+(0.183419576182)*x_ref[1]**o+(-0.871451480853)*x_ref[1]+(-0.349396269955)*x_ref[2]**o+(0.095909246936)*x_ref[2]
        ref[2,1,1]=(-0.00972985962106)*x_ref[0]**o+(-0.919204152811)*x_ref[0]+(-0.751744404365)*x_ref[1]**o+(0.377096436852)*x_ref[1]+(-0.924228885362)*x_ref[2]**o+(0.451053813959)*x_ref[2]
        ref[3,0,0]=(0.951119968373)*x_ref[0]**o+(-0.0680027504528)*x_ref[0]+(-0.113660990287)*x_ref[1]**o+(-0.428999453003)*x_ref[1]+(0.171804544103)*x_ref[2]**o+(0.677635036297)*x_ref[2]
        ref[3,0,1]=(0.62534864581)*x_ref[0]**o+(0.389496246771)*x_ref[0]+(0.621295327948)*x_ref[1]**o+(-0.684165739023)*x_ref[1]+(0.258637161492)*x_ref[2]**o+(0.0441460764076)*x_ref[2]
        ref[3,1,0]=(0.0348922721145)*x_ref[0]**o+(0.091201286767)*x_ref[0]+(0.421079169132)*x_ref[1]**o+(-0.616504862591)*x_ref[1]+(0.563927673176)*x_ref[2]**o+(-0.183993394822)*x_ref[2]
        ref[3,1,1]=(0.657959811783)*x_ref[0]**o+(0.229729671382)*x_ref[0]+(0.549853155207)*x_ref[1]**o+(-0.847412930565)*x_ref[1]+(-0.0780682500337)*x_ref[2]**o+(0.574435563406)*x_ref[2]
        ref[4,0,0]=(-0.679988515344)*x_ref[0]**o+(0.0517331630343)*x_ref[0]+(-0.160281064879)*x_ref[1]**o+(-0.0672405944261)*x_ref[1]+(-0.118968444607)*x_ref[2]**o+(0.00980065834148)*x_ref[2]
        ref[4,0,1]=(-0.150457972951)*x_ref[0]**o+(0.671586511406)*x_ref[0]+(-0.844112675259)*x_ref[1]**o+(0.42766255101)*x_ref[1]+(-0.22449721018)*x_ref[2]**o+(-0.150416151059)*x_ref[2]
        ref[4,1,0]=(0.0768992597753)*x_ref[0]**o+(0.0139605476692)*x_ref[0]+(-0.831269554863)*x_ref[1]**o+(-0.279339344511)*x_ref[1]+(0.387318821185)*x_ref[2]**o+(0.967687121678)*x_ref[2]
        ref[4,1,1]=(-0.670585948598)*x_ref[0]**o+(0.422715818654)*x_ref[0]+(0.623702390878)*x_ref[1]**o+(0.922646081646)*x_ref[1]+(-0.714191287039)*x_ref[2]**o+(-0.99018084869)*x_ref[2]
        ref[5,0,0]=(0.100273819257)*x_ref[0]**o+(-0.769422291863)*x_ref[0]+(-0.152611272068)*x_ref[1]**o+(-0.396798631305)*x_ref[1]+(0.502251047462)*x_ref[2]**o+(0.993791216449)*x_ref[2]
        ref[5,0,1]=(-0.0222540900657)*x_ref[0]**o+(-0.339810353987)*x_ref[0]+(0.241665078696)*x_ref[1]**o+(0.111020568564)*x_ref[1]+(0.298147316931)*x_ref[2]**o+(0.391278660594)*x_ref[2]
        ref[5,1,0]=(0.828297599555)*x_ref[0]**o+(0.237678875371)*x_ref[0]+(0.924551126274)*x_ref[1]**o+(-0.663695049485)*x_ref[1]+(-0.547123493132)*x_ref[2]**o+(-0.788635745118)*x_ref[2]
        ref[5,1,1]=(0.0726334370982)*x_ref[0]**o+(-0.875688492957)*x_ref[0]+(0.473726235108)*x_ref[1]**o+(0.023025596151)*x_ref[1]+(0.827449396402)*x_ref[2]**o+(-0.154698941604)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0338299192888)*x[0]**o+(-0.114906274359)*x[0]+(0.15010339398)*x[1]**o+(0.625295991328)*x[1]
        arg[0,0,0,1]=(-0.1311901039)*x[0]**o+(-0.551712861217)*x[0]+(-0.287411404506)*x[1]**o+(0.0242820629572)*x[1]
        arg[0,0,1,0]=(0.121962094111)*x[0]**o+(0.203169622345)*x[0]+(0.241504218053)*x[1]**o+(-0.0696356239498)*x[1]
        arg[0,0,1,1]=(0.0461537352329)*x[0]**o+(-0.275908938567)*x[0]+(-0.271184121027)*x[1]**o+(-0.364024729567)*x[1]
        arg[0,0,2,0]=(0.199272234315)*x[0]**o+(-0.906364258855)*x[0]+(-0.92102930037)*x[1]**o+(0.0718981903454)*x[1]
        arg[0,0,2,1]=(-0.742950958383)*x[0]**o+(0.0301464847722)*x[0]+(0.466993477687)*x[1]**o+(-0.402131750698)*x[1]
        arg[0,1,0,0]=(-0.0653776546375)*x[0]**o+(-0.679360612372)*x[0]+(-0.32783837768)*x[1]**o+(-0.113753016428)*x[1]
        arg[0,1,0,1]=(0.58610119414)*x[0]**o+(0.142613166231)*x[0]+(0.0409105619986)*x[1]**o+(-0.660641663558)*x[1]
        arg[0,1,1,0]=(0.212172308585)*x[0]**o+(-0.452767768344)*x[0]+(0.963053969889)*x[1]**o+(0.752691773953)*x[1]
        arg[0,1,1,1]=(-0.514998377887)*x[0]**o+(0.0872464329801)*x[0]+(0.265590295505)*x[1]**o+(0.557238826549)*x[1]
        arg[0,1,2,0]=(0.360523954724)*x[0]**o+(0.550238402994)*x[0]+(-0.206349415539)*x[1]**o+(0.357304730511)*x[1]
        arg[0,1,2,1]=(-0.872755217435)*x[0]**o+(0.190943683664)*x[0]+(-0.245454292354)*x[1]**o+(0.507229578506)*x[1]
        arg[0,2,0,0]=(0.994619505257)*x[0]**o+(0.33482820568)*x[0]+(-0.12721791236)*x[1]**o+(-0.505873171195)*x[1]
        arg[0,2,0,1]=(0.224805131522)*x[0]**o+(0.332022434603)*x[0]+(0.0583494507268)*x[1]**o+(0.857185410994)*x[1]
        arg[0,2,1,0]=(-0.425367077516)*x[0]**o+(-0.25207101061)*x[0]+(-0.11164491753)*x[1]**o+(0.44752084285)*x[1]
        arg[0,2,1,1]=(-0.158725210751)*x[0]**o+(-0.0692363138932)*x[0]+(0.214803128256)*x[1]**o+(0.886699106748)*x[1]
        arg[0,2,2,0]=(-0.332146331257)*x[0]**o+(0.114845050128)*x[0]+(-0.0781411611124)*x[1]**o+(0.474972057387)*x[1]
        arg[0,2,2,1]=(-0.481949630694)*x[0]**o+(0.644984972151)*x[0]+(0.171146519527)*x[1]**o+(0.57534803966)*x[1]
        arg[0,3,0,0]=(0.7580633024)*x[0]**o+(0.0280511776577)*x[0]+(0.375583652078)*x[1]**o+(-0.809112787284)*x[1]
        arg[0,3,0,1]=(-0.466737159934)*x[0]**o+(0.410032775686)*x[0]+(0.176141189584)*x[1]**o+(-0.228288309456)*x[1]
        arg[0,3,1,0]=(0.427129582196)*x[0]**o+(-0.718385859483)*x[0]+(-0.172261212989)*x[1]**o+(0.913157476188)*x[1]
        arg[0,3,1,1]=(0.952506706487)*x[0]**o+(-0.497478163899)*x[0]+(-0.930273038716)*x[1]**o+(0.673573149025)*x[1]
        arg[0,3,2,0]=(-0.0074302930747)*x[0]**o+(0.226424842035)*x[0]+(0.871190620671)*x[1]**o+(0.277695333679)*x[1]
        arg[0,3,2,1]=(-0.540820239946)*x[0]**o+(0.33044763211)*x[0]+(-0.772648820924)*x[1]**o+(-0.362410088021)*x[1]
        arg[0,4,0,0]=(-0.748396039111)*x[0]**o+(0.215131518212)*x[0]+(0.514087824403)*x[1]**o+(0.335563854205)*x[1]
        arg[0,4,0,1]=(-0.443263590391)*x[0]**o+(-0.523167282542)*x[0]+(0.122612452799)*x[1]**o+(0.216527857615)*x[1]
        arg[0,4,1,0]=(0.448176296777)*x[0]**o+(0.478292146131)*x[0]+(-0.558921083502)*x[1]**o+(-0.17018258654)*x[1]
        arg[0,4,1,1]=(-0.912765673953)*x[0]**o+(0.751746623223)*x[0]+(-0.104811579057)*x[1]**o+(0.456436870323)*x[1]
        arg[0,4,2,0]=(0.157946450932)*x[0]**o+(0.147858587857)*x[0]+(0.775747362874)*x[1]**o+(0.478478862463)*x[1]
        arg[0,4,2,1]=(-0.819625058185)*x[0]**o+(-0.0231980393919)*x[0]+(-0.186647585522)*x[1]**o+(0.513350371371)*x[1]
        arg[1,0,0,0]=(-0.942373946995)*x[0]**o+(-0.761326160931)*x[0]+(0.192511867167)*x[1]**o+(-0.50895969579)*x[1]
        arg[1,0,0,1]=(0.0670813349951)*x[0]**o+(0.55602953512)*x[0]+(0.757621178635)*x[1]**o+(0.741550201759)*x[1]
        arg[1,0,1,0]=(-0.114642609493)*x[0]**o+(-0.639500048447)*x[0]+(-0.861813039243)*x[1]**o+(0.36413406744)*x[1]
        arg[1,0,1,1]=(0.126555451594)*x[0]**o+(-0.885172557082)*x[0]+(-0.376073252921)*x[1]**o+(0.166102850653)*x[1]
        arg[1,0,2,0]=(-0.381517068219)*x[0]**o+(0.94870328267)*x[0]+(-0.129934777755)*x[1]**o+(0.78090715758)*x[1]
        arg[1,0,2,1]=(-0.535852302247)*x[0]**o+(0.63751039318)*x[0]+(-0.537556159925)*x[1]**o+(0.147766714221)*x[1]
        arg[1,1,0,0]=(-0.605543047533)*x[0]**o+(-0.366612785372)*x[0]+(0.544900149212)*x[1]**o+(0.722108900447)*x[1]
        arg[1,1,0,1]=(0.133378561142)*x[0]**o+(-0.227808772494)*x[0]+(-0.729399624175)*x[1]**o+(0.307571743455)*x[1]
        arg[1,1,1,0]=(0.0458830538139)*x[0]**o+(0.0585044493026)*x[0]+(-0.136630547605)*x[1]**o+(-0.0823288389309)*x[1]
        arg[1,1,1,1]=(0.339263808784)*x[0]**o+(0.00988917109076)*x[0]+(-0.786591838083)*x[1]**o+(-0.361686799179)*x[1]
        arg[1,1,2,0]=(0.789942037592)*x[0]**o+(-0.266689758676)*x[0]+(0.464856327495)*x[1]**o+(0.138022774293)*x[1]
        arg[1,1,2,1]=(-0.723544867217)*x[0]**o+(-0.523023674205)*x[0]+(0.629726026331)*x[1]**o+(-0.867808120997)*x[1]
        arg[1,2,0,0]=(-0.495589518787)*x[0]**o+(-0.561664880262)*x[0]+(0.0319320417078)*x[1]**o+(-0.153720661498)*x[1]
        arg[1,2,0,1]=(0.293168114024)*x[0]**o+(-0.344107163981)*x[0]+(-0.275629670068)*x[1]**o+(0.06103653487)*x[1]
        arg[1,2,1,0]=(0.495640927339)*x[0]**o+(0.596540095043)*x[0]+(0.208792567756)*x[1]**o+(0.886040724145)*x[1]
        arg[1,2,1,1]=(0.331476398093)*x[0]**o+(-0.472588572455)*x[0]+(-0.455670457285)*x[1]**o+(-0.731223615642)*x[1]
        arg[1,2,2,0]=(-0.127107829674)*x[0]**o+(-0.21113128456)*x[0]+(-0.733928194168)*x[1]**o+(-0.174127696853)*x[1]
        arg[1,2,2,1]=(0.521879316151)*x[0]**o+(0.983424557865)*x[0]+(-0.288659660182)*x[1]**o+(-0.431045203113)*x[1]
        arg[1,3,0,0]=(0.292348765746)*x[0]**o+(0.0730741240915)*x[0]+(0.079624022683)*x[1]**o+(0.237483071633)*x[1]
        arg[1,3,0,1]=(0.291035771848)*x[0]**o+(-0.482909477855)*x[0]+(-0.999142129615)*x[1]**o+(-0.559723367197)*x[1]
        arg[1,3,1,0]=(0.747096275653)*x[0]**o+(0.762757649624)*x[0]+(-0.821650889507)*x[1]**o+(0.301128670189)*x[1]
        arg[1,3,1,1]=(0.82703681414)*x[0]**o+(-0.336570070436)*x[0]+(0.798131952979)*x[1]**o+(0.603562737956)*x[1]
        arg[1,3,2,0]=(-0.164875914126)*x[0]**o+(0.219409873902)*x[0]+(0.784253160678)*x[1]**o+(0.0219597460707)*x[1]
        arg[1,3,2,1]=(0.124778899794)*x[0]**o+(0.785367842145)*x[0]+(0.948111125908)*x[1]**o+(-0.887427847089)*x[1]
        arg[1,4,0,0]=(0.0911816002883)*x[0]**o+(0.645410022441)*x[0]+(0.720551531965)*x[1]**o+(-0.388407683413)*x[1]
        arg[1,4,0,1]=(-0.0416762613062)*x[0]**o+(-0.548874829834)*x[0]+(-0.377952555557)*x[1]**o+(0.742795234361)*x[1]
        arg[1,4,1,0]=(-0.896929017971)*x[0]**o+(0.678343686516)*x[0]+(0.711572763082)*x[1]**o+(-0.43397331651)*x[1]
        arg[1,4,1,1]=(-0.190804421566)*x[0]**o+(-0.0848405124789)*x[0]+(0.36246957905)*x[1]**o+(0.963599086227)*x[1]
        arg[1,4,2,0]=(-0.0210248567982)*x[0]**o+(0.545977814581)*x[0]+(0.858962864401)*x[1]**o+(0.113792851997)*x[1]
        arg[1,4,2,1]=(0.229641892622)*x[0]**o+(-0.739561617198)*x[0]+(-0.164372740617)*x[1]**o+(0.537097868843)*x[1]
        arg[2,0,0,0]=(-0.13321389343)*x[0]**o+(-0.668403350503)*x[0]+(-0.924879105983)*x[1]**o+(-0.5161790717)*x[1]
        arg[2,0,0,1]=(0.26582844685)*x[0]**o+(0.89048660684)*x[0]+(-0.377225981719)*x[1]**o+(-0.303790710194)*x[1]
        arg[2,0,1,0]=(0.0997470860283)*x[0]**o+(0.0385427264217)*x[0]+(-0.71387347273)*x[1]**o+(-0.621376499192)*x[1]
        arg[2,0,1,1]=(0.595544616326)*x[0]**o+(-0.850529503975)*x[0]+(0.602203957938)*x[1]**o+(-0.329484843907)*x[1]
        arg[2,0,2,0]=(-0.106031967029)*x[0]**o+(-0.165122821622)*x[0]+(0.647473815919)*x[1]**o+(0.28942226062)*x[1]
        arg[2,0,2,1]=(0.585955875472)*x[0]**o+(0.922667710403)*x[0]+(-0.0710905957841)*x[1]**o+(-0.669138163697)*x[1]
        arg[2,1,0,0]=(0.461941211081)*x[0]**o+(0.723022466263)*x[0]+(-0.118551893472)*x[1]**o+(0.00445187353637)*x[1]
        arg[2,1,0,1]=(-0.253168248108)*x[0]**o+(0.476938842605)*x[0]+(0.476641658034)*x[1]**o+(-0.916886387595)*x[1]
        arg[2,1,1,0]=(0.722029309503)*x[0]**o+(0.0588848593425)*x[0]+(0.84737370043)*x[1]**o+(-0.674604415213)*x[1]
        arg[2,1,1,1]=(-0.644736759298)*x[0]**o+(-0.212847303893)*x[0]+(0.25394587552)*x[1]**o+(-0.366955175081)*x[1]
        arg[2,1,2,0]=(-0.657395296801)*x[0]**o+(-0.967570027599)*x[0]+(0.180926628541)*x[1]**o+(-0.940599897942)*x[1]
        arg[2,1,2,1]=(0.849897328801)*x[0]**o+(-0.299199474051)*x[0]+(0.427221207827)*x[1]**o+(0.212027985441)*x[1]
        arg[2,2,0,0]=(-0.561573507381)*x[0]**o+(-0.446925224908)*x[0]+(0.132091224948)*x[1]**o+(-0.673375661849)*x[1]
        arg[2,2,0,1]=(-0.386657291574)*x[0]**o+(0.356784637112)*x[0]+(0.644292006233)*x[1]**o+(0.922937167842)*x[1]
        arg[2,2,1,0]=(0.484446757285)*x[0]**o+(-0.822469870455)*x[0]+(-0.89605890123)*x[1]**o+(-0.895891500031)*x[1]
        arg[2,2,1,1]=(-0.0930740520192)*x[0]**o+(0.953356473678)*x[0]+(-0.714418458806)*x[1]**o+(-0.100584950946)*x[1]
        arg[2,2,2,0]=(0.945572242667)*x[0]**o+(-0.259280849605)*x[0]+(-0.925459846341)*x[1]**o+(0.906002758873)*x[1]
        arg[2,2,2,1]=(-0.638993332975)*x[0]**o+(-0.668594641703)*x[0]+(0.012666231369)*x[1]**o+(0.89732685856)*x[1]
        arg[2,3,0,0]=(-0.885061116144)*x[0]**o+(-0.576097536483)*x[0]+(-0.43166075788)*x[1]**o+(0.538821620917)*x[1]
        arg[2,3,0,1]=(-0.0743638381797)*x[0]**o+(0.251037500035)*x[0]+(-0.905989928076)*x[1]**o+(0.243771000867)*x[1]
        arg[2,3,1,0]=(-0.016045622351)*x[0]**o+(-0.632944072174)*x[0]+(-0.296175918793)*x[1]**o+(-0.00120123641365)*x[1]
        arg[2,3,1,1]=(-0.896354921812)*x[0]**o+(-0.331348989027)*x[0]+(-0.450865285201)*x[1]**o+(-0.243268582009)*x[1]
        arg[2,3,2,0]=(0.609332927431)*x[0]**o+(-0.0815856802115)*x[0]+(0.201924792273)*x[1]**o+(-0.890424451348)*x[1]
        arg[2,3,2,1]=(0.395090137264)*x[0]**o+(0.294230442573)*x[0]+(-0.345371510364)*x[1]**o+(0.181229663241)*x[1]
        arg[2,4,0,0]=(-0.823684509034)*x[0]**o+(0.431019301942)*x[0]+(0.554707015054)*x[1]**o+(-0.846322903003)*x[1]
        arg[2,4,0,1]=(0.862984476658)*x[0]**o+(0.20512500102)*x[0]+(0.0247756159287)*x[1]**o+(-0.374795100946)*x[1]
        arg[2,4,1,0]=(-0.0849914269517)*x[0]**o+(0.754908200339)*x[0]+(0.593363754807)*x[1]**o+(0.124923961126)*x[1]
        arg[2,4,1,1]=(-0.691440264234)*x[0]**o+(0.825739298015)*x[0]+(0.426452231478)*x[1]**o+(0.0995610721341)*x[1]
        arg[2,4,2,0]=(-0.175322750567)*x[0]**o+(0.419750865569)*x[0]+(0.208448987712)*x[1]**o+(0.150584798952)*x[1]
        arg[2,4,2,1]=(0.931680152344)*x[0]**o+(0.390947206307)*x[0]+(0.388437348378)*x[1]**o+(-0.808413079032)*x[1]
        arg[3,0,0,0]=(0.58855431402)*x[0]**o+(0.383865913364)*x[0]+(-0.691445323219)*x[1]**o+(-0.443370658523)*x[1]
        arg[3,0,0,1]=(-0.657371481478)*x[0]**o+(0.746891082673)*x[0]+(-0.829229947773)*x[1]**o+(0.761915670404)*x[1]
        arg[3,0,1,0]=(-0.431502932714)*x[0]**o+(-0.961472285716)*x[0]+(0.0423384388031)*x[1]**o+(0.339373458118)*x[1]
        arg[3,0,1,1]=(-0.272831163259)*x[0]**o+(-0.533438904494)*x[0]+(-0.595625384875)*x[1]**o+(0.0366857022323)*x[1]
        arg[3,0,2,0]=(-0.319684662306)*x[0]**o+(0.672185828912)*x[0]+(0.483126548994)*x[1]**o+(0.745872629604)*x[1]
        arg[3,0,2,1]=(0.437895277404)*x[0]**o+(-0.337209416963)*x[0]+(0.101080522788)*x[1]**o+(0.892022488644)*x[1]
        arg[3,1,0,0]=(-0.582325078455)*x[0]**o+(-0.854689347157)*x[0]+(0.0308638312259)*x[1]**o+(0.633382365704)*x[1]
        arg[3,1,0,1]=(0.319735178376)*x[0]**o+(-0.782764593952)*x[0]+(0.082230132244)*x[1]**o+(-0.946010937852)*x[1]
        arg[3,1,1,0]=(-0.698490410666)*x[0]**o+(-0.801358802689)*x[0]+(0.20695648235)*x[1]**o+(0.75257523471)*x[1]
        arg[3,1,1,1]=(0.66321696527)*x[0]**o+(0.704572918868)*x[0]+(0.528208879002)*x[1]**o+(-0.408378325513)*x[1]
        arg[3,1,2,0]=(-0.925177198417)*x[0]**o+(-0.152427811549)*x[0]+(0.341781993583)*x[1]**o+(-0.589976456377)*x[1]
        arg[3,1,2,1]=(0.206706628353)*x[0]**o+(0.673777173329)*x[0]+(-0.994138314254)*x[1]**o+(-0.843696565863)*x[1]
        arg[3,2,0,0]=(-0.588881067333)*x[0]**o+(-0.537572388341)*x[0]+(-0.206570371675)*x[1]**o+(-0.127468554294)*x[1]
        arg[3,2,0,1]=(0.498638284096)*x[0]**o+(-0.488278831086)*x[0]+(-0.28320874594)*x[1]**o+(-0.638241576615)*x[1]
        arg[3,2,1,0]=(-0.445780790822)*x[0]**o+(-0.347493953089)*x[0]+(-0.558010086232)*x[1]**o+(-0.0287339307689)*x[1]
        arg[3,2,1,1]=(-0.361727811272)*x[0]**o+(-0.559565378007)*x[0]+(-0.211685558972)*x[1]**o+(0.330088396811)*x[1]
        arg[3,2,2,0]=(-0.42600117831)*x[0]**o+(-0.488997111988)*x[0]+(0.127421278051)*x[1]**o+(-0.48233167806)*x[1]
        arg[3,2,2,1]=(0.120468829686)*x[0]**o+(0.0675603906118)*x[0]+(-0.60674361864)*x[1]**o+(0.639115708258)*x[1]
        arg[3,3,0,0]=(0.369343033015)*x[0]**o+(0.707735743593)*x[0]+(-0.426275591836)*x[1]**o+(-0.733377765625)*x[1]
        arg[3,3,0,1]=(-0.931245338156)*x[0]**o+(-0.465861794185)*x[0]+(-0.650664500687)*x[1]**o+(0.103077662402)*x[1]
        arg[3,3,1,0]=(0.731834744512)*x[0]**o+(0.267840571071)*x[0]+(0.952106457495)*x[1]**o+(-0.943845364366)*x[1]
        arg[3,3,1,1]=(-0.21692720916)*x[0]**o+(0.580841374997)*x[0]+(-0.412129341223)*x[1]**o+(0.40414445304)*x[1]
        arg[3,3,2,0]=(-0.447758598726)*x[0]**o+(0.0616926352912)*x[0]+(-0.824624226122)*x[1]**o+(0.19099833104)*x[1]
        arg[3,3,2,1]=(-0.615519138165)*x[0]**o+(0.233926822164)*x[0]+(-0.969014060236)*x[1]**o+(0.835103627893)*x[1]
        arg[3,4,0,0]=(0.550778983301)*x[0]**o+(0.0198253530882)*x[0]+(-0.39383849187)*x[1]**o+(0.367721826304)*x[1]
        arg[3,4,0,1]=(0.444877923149)*x[0]**o+(-0.899159835223)*x[0]+(0.923392540452)*x[1]**o+(0.512905141718)*x[1]
        arg[3,4,1,0]=(-0.563807675145)*x[0]**o+(0.994405383899)*x[0]+(-0.777352948218)*x[1]**o+(0.865994906227)*x[1]
        arg[3,4,1,1]=(-0.593800452139)*x[0]**o+(0.613778803237)*x[0]+(0.582200979718)*x[1]**o+(0.448026243271)*x[1]
        arg[3,4,2,0]=(-0.900331125416)*x[0]**o+(0.234613965424)*x[0]+(0.510400273805)*x[1]**o+(0.0836773641661)*x[1]
        arg[3,4,2,1]=(0.710999097307)*x[0]**o+(0.886603988265)*x[0]+(-0.405644887679)*x[1]**o+(0.362257258809)*x[1]
        ref[0,0,0,0]=(-0.0338299192888)*x_ref[0]**o+(-0.114906274359)*x_ref[0]+(0.15010339398)*x_ref[1]**o+(0.625295991328)*x_ref[1]
        ref[0,0,0,1]=(-0.1311901039)*x_ref[0]**o+(-0.551712861217)*x_ref[0]+(-0.287411404506)*x_ref[1]**o+(0.0242820629572)*x_ref[1]
        ref[0,0,1,0]=(0.121962094111)*x_ref[0]**o+(0.203169622345)*x_ref[0]+(0.241504218053)*x_ref[1]**o+(-0.0696356239498)*x_ref[1]
        ref[0,0,1,1]=(0.0461537352329)*x_ref[0]**o+(-0.275908938567)*x_ref[0]+(-0.271184121027)*x_ref[1]**o+(-0.364024729567)*x_ref[1]
        ref[0,0,2,0]=(0.199272234315)*x_ref[0]**o+(-0.906364258855)*x_ref[0]+(-0.92102930037)*x_ref[1]**o+(0.0718981903454)*x_ref[1]
        ref[0,0,2,1]=(-0.742950958383)*x_ref[0]**o+(0.0301464847722)*x_ref[0]+(0.466993477687)*x_ref[1]**o+(-0.402131750698)*x_ref[1]
        ref[0,1,0,0]=(-0.0653776546375)*x_ref[0]**o+(-0.679360612372)*x_ref[0]+(-0.32783837768)*x_ref[1]**o+(-0.113753016428)*x_ref[1]
        ref[0,1,0,1]=(0.58610119414)*x_ref[0]**o+(0.142613166231)*x_ref[0]+(0.0409105619986)*x_ref[1]**o+(-0.660641663558)*x_ref[1]
        ref[0,1,1,0]=(0.212172308585)*x_ref[0]**o+(-0.452767768344)*x_ref[0]+(0.963053969889)*x_ref[1]**o+(0.752691773953)*x_ref[1]
        ref[0,1,1,1]=(-0.514998377887)*x_ref[0]**o+(0.0872464329801)*x_ref[0]+(0.265590295505)*x_ref[1]**o+(0.557238826549)*x_ref[1]
        ref[0,1,2,0]=(0.360523954724)*x_ref[0]**o+(0.550238402994)*x_ref[0]+(-0.206349415539)*x_ref[1]**o+(0.357304730511)*x_ref[1]
        ref[0,1,2,1]=(-0.872755217435)*x_ref[0]**o+(0.190943683664)*x_ref[0]+(-0.245454292354)*x_ref[1]**o+(0.507229578506)*x_ref[1]
        ref[0,2,0,0]=(0.994619505257)*x_ref[0]**o+(0.33482820568)*x_ref[0]+(-0.12721791236)*x_ref[1]**o+(-0.505873171195)*x_ref[1]
        ref[0,2,0,1]=(0.224805131522)*x_ref[0]**o+(0.332022434603)*x_ref[0]+(0.0583494507268)*x_ref[1]**o+(0.857185410994)*x_ref[1]
        ref[0,2,1,0]=(-0.425367077516)*x_ref[0]**o+(-0.25207101061)*x_ref[0]+(-0.11164491753)*x_ref[1]**o+(0.44752084285)*x_ref[1]
        ref[0,2,1,1]=(-0.158725210751)*x_ref[0]**o+(-0.0692363138932)*x_ref[0]+(0.214803128256)*x_ref[1]**o+(0.886699106748)*x_ref[1]
        ref[0,2,2,0]=(-0.332146331257)*x_ref[0]**o+(0.114845050128)*x_ref[0]+(-0.0781411611124)*x_ref[1]**o+(0.474972057387)*x_ref[1]
        ref[0,2,2,1]=(-0.481949630694)*x_ref[0]**o+(0.644984972151)*x_ref[0]+(0.171146519527)*x_ref[1]**o+(0.57534803966)*x_ref[1]
        ref[0,3,0,0]=(0.7580633024)*x_ref[0]**o+(0.0280511776577)*x_ref[0]+(0.375583652078)*x_ref[1]**o+(-0.809112787284)*x_ref[1]
        ref[0,3,0,1]=(-0.466737159934)*x_ref[0]**o+(0.410032775686)*x_ref[0]+(0.176141189584)*x_ref[1]**o+(-0.228288309456)*x_ref[1]
        ref[0,3,1,0]=(0.427129582196)*x_ref[0]**o+(-0.718385859483)*x_ref[0]+(-0.172261212989)*x_ref[1]**o+(0.913157476188)*x_ref[1]
        ref[0,3,1,1]=(0.952506706487)*x_ref[0]**o+(-0.497478163899)*x_ref[0]+(-0.930273038716)*x_ref[1]**o+(0.673573149025)*x_ref[1]
        ref[0,3,2,0]=(-0.0074302930747)*x_ref[0]**o+(0.226424842035)*x_ref[0]+(0.871190620671)*x_ref[1]**o+(0.277695333679)*x_ref[1]
        ref[0,3,2,1]=(-0.540820239946)*x_ref[0]**o+(0.33044763211)*x_ref[0]+(-0.772648820924)*x_ref[1]**o+(-0.362410088021)*x_ref[1]
        ref[0,4,0,0]=(-0.748396039111)*x_ref[0]**o+(0.215131518212)*x_ref[0]+(0.514087824403)*x_ref[1]**o+(0.335563854205)*x_ref[1]
        ref[0,4,0,1]=(-0.443263590391)*x_ref[0]**o+(-0.523167282542)*x_ref[0]+(0.122612452799)*x_ref[1]**o+(0.216527857615)*x_ref[1]
        ref[0,4,1,0]=(0.448176296777)*x_ref[0]**o+(0.478292146131)*x_ref[0]+(-0.558921083502)*x_ref[1]**o+(-0.17018258654)*x_ref[1]
        ref[0,4,1,1]=(-0.912765673953)*x_ref[0]**o+(0.751746623223)*x_ref[0]+(-0.104811579057)*x_ref[1]**o+(0.456436870323)*x_ref[1]
        ref[0,4,2,0]=(0.157946450932)*x_ref[0]**o+(0.147858587857)*x_ref[0]+(0.775747362874)*x_ref[1]**o+(0.478478862463)*x_ref[1]
        ref[0,4,2,1]=(-0.819625058185)*x_ref[0]**o+(-0.0231980393919)*x_ref[0]+(-0.186647585522)*x_ref[1]**o+(0.513350371371)*x_ref[1]
        ref[1,0,0,0]=(-0.942373946995)*x_ref[0]**o+(-0.761326160931)*x_ref[0]+(0.192511867167)*x_ref[1]**o+(-0.50895969579)*x_ref[1]
        ref[1,0,0,1]=(0.0670813349951)*x_ref[0]**o+(0.55602953512)*x_ref[0]+(0.757621178635)*x_ref[1]**o+(0.741550201759)*x_ref[1]
        ref[1,0,1,0]=(-0.114642609493)*x_ref[0]**o+(-0.639500048447)*x_ref[0]+(-0.861813039243)*x_ref[1]**o+(0.36413406744)*x_ref[1]
        ref[1,0,1,1]=(0.126555451594)*x_ref[0]**o+(-0.885172557082)*x_ref[0]+(-0.376073252921)*x_ref[1]**o+(0.166102850653)*x_ref[1]
        ref[1,0,2,0]=(-0.381517068219)*x_ref[0]**o+(0.94870328267)*x_ref[0]+(-0.129934777755)*x_ref[1]**o+(0.78090715758)*x_ref[1]
        ref[1,0,2,1]=(-0.535852302247)*x_ref[0]**o+(0.63751039318)*x_ref[0]+(-0.537556159925)*x_ref[1]**o+(0.147766714221)*x_ref[1]
        ref[1,1,0,0]=(-0.605543047533)*x_ref[0]**o+(-0.366612785372)*x_ref[0]+(0.544900149212)*x_ref[1]**o+(0.722108900447)*x_ref[1]
        ref[1,1,0,1]=(0.133378561142)*x_ref[0]**o+(-0.227808772494)*x_ref[0]+(-0.729399624175)*x_ref[1]**o+(0.307571743455)*x_ref[1]
        ref[1,1,1,0]=(0.0458830538139)*x_ref[0]**o+(0.0585044493026)*x_ref[0]+(-0.136630547605)*x_ref[1]**o+(-0.0823288389309)*x_ref[1]
        ref[1,1,1,1]=(0.339263808784)*x_ref[0]**o+(0.00988917109076)*x_ref[0]+(-0.786591838083)*x_ref[1]**o+(-0.361686799179)*x_ref[1]
        ref[1,1,2,0]=(0.789942037592)*x_ref[0]**o+(-0.266689758676)*x_ref[0]+(0.464856327495)*x_ref[1]**o+(0.138022774293)*x_ref[1]
        ref[1,1,2,1]=(-0.723544867217)*x_ref[0]**o+(-0.523023674205)*x_ref[0]+(0.629726026331)*x_ref[1]**o+(-0.867808120997)*x_ref[1]
        ref[1,2,0,0]=(-0.495589518787)*x_ref[0]**o+(-0.561664880262)*x_ref[0]+(0.0319320417078)*x_ref[1]**o+(-0.153720661498)*x_ref[1]
        ref[1,2,0,1]=(0.293168114024)*x_ref[0]**o+(-0.344107163981)*x_ref[0]+(-0.275629670068)*x_ref[1]**o+(0.06103653487)*x_ref[1]
        ref[1,2,1,0]=(0.495640927339)*x_ref[0]**o+(0.596540095043)*x_ref[0]+(0.208792567756)*x_ref[1]**o+(0.886040724145)*x_ref[1]
        ref[1,2,1,1]=(0.331476398093)*x_ref[0]**o+(-0.472588572455)*x_ref[0]+(-0.455670457285)*x_ref[1]**o+(-0.731223615642)*x_ref[1]
        ref[1,2,2,0]=(-0.127107829674)*x_ref[0]**o+(-0.21113128456)*x_ref[0]+(-0.733928194168)*x_ref[1]**o+(-0.174127696853)*x_ref[1]
        ref[1,2,2,1]=(0.521879316151)*x_ref[0]**o+(0.983424557865)*x_ref[0]+(-0.288659660182)*x_ref[1]**o+(-0.431045203113)*x_ref[1]
        ref[1,3,0,0]=(0.292348765746)*x_ref[0]**o+(0.0730741240915)*x_ref[0]+(0.079624022683)*x_ref[1]**o+(0.237483071633)*x_ref[1]
        ref[1,3,0,1]=(0.291035771848)*x_ref[0]**o+(-0.482909477855)*x_ref[0]+(-0.999142129615)*x_ref[1]**o+(-0.559723367197)*x_ref[1]
        ref[1,3,1,0]=(0.747096275653)*x_ref[0]**o+(0.762757649624)*x_ref[0]+(-0.821650889507)*x_ref[1]**o+(0.301128670189)*x_ref[1]
        ref[1,3,1,1]=(0.82703681414)*x_ref[0]**o+(-0.336570070436)*x_ref[0]+(0.798131952979)*x_ref[1]**o+(0.603562737956)*x_ref[1]
        ref[1,3,2,0]=(-0.164875914126)*x_ref[0]**o+(0.219409873902)*x_ref[0]+(0.784253160678)*x_ref[1]**o+(0.0219597460707)*x_ref[1]
        ref[1,3,2,1]=(0.124778899794)*x_ref[0]**o+(0.785367842145)*x_ref[0]+(0.948111125908)*x_ref[1]**o+(-0.887427847089)*x_ref[1]
        ref[1,4,0,0]=(0.0911816002883)*x_ref[0]**o+(0.645410022441)*x_ref[0]+(0.720551531965)*x_ref[1]**o+(-0.388407683413)*x_ref[1]
        ref[1,4,0,1]=(-0.0416762613062)*x_ref[0]**o+(-0.548874829834)*x_ref[0]+(-0.377952555557)*x_ref[1]**o+(0.742795234361)*x_ref[1]
        ref[1,4,1,0]=(-0.896929017971)*x_ref[0]**o+(0.678343686516)*x_ref[0]+(0.711572763082)*x_ref[1]**o+(-0.43397331651)*x_ref[1]
        ref[1,4,1,1]=(-0.190804421566)*x_ref[0]**o+(-0.0848405124789)*x_ref[0]+(0.36246957905)*x_ref[1]**o+(0.963599086227)*x_ref[1]
        ref[1,4,2,0]=(-0.0210248567982)*x_ref[0]**o+(0.545977814581)*x_ref[0]+(0.858962864401)*x_ref[1]**o+(0.113792851997)*x_ref[1]
        ref[1,4,2,1]=(0.229641892622)*x_ref[0]**o+(-0.739561617198)*x_ref[0]+(-0.164372740617)*x_ref[1]**o+(0.537097868843)*x_ref[1]
        ref[2,0,0,0]=(-0.13321389343)*x_ref[0]**o+(-0.668403350503)*x_ref[0]+(-0.924879105983)*x_ref[1]**o+(-0.5161790717)*x_ref[1]
        ref[2,0,0,1]=(0.26582844685)*x_ref[0]**o+(0.89048660684)*x_ref[0]+(-0.377225981719)*x_ref[1]**o+(-0.303790710194)*x_ref[1]
        ref[2,0,1,0]=(0.0997470860283)*x_ref[0]**o+(0.0385427264217)*x_ref[0]+(-0.71387347273)*x_ref[1]**o+(-0.621376499192)*x_ref[1]
        ref[2,0,1,1]=(0.595544616326)*x_ref[0]**o+(-0.850529503975)*x_ref[0]+(0.602203957938)*x_ref[1]**o+(-0.329484843907)*x_ref[1]
        ref[2,0,2,0]=(-0.106031967029)*x_ref[0]**o+(-0.165122821622)*x_ref[0]+(0.647473815919)*x_ref[1]**o+(0.28942226062)*x_ref[1]
        ref[2,0,2,1]=(0.585955875472)*x_ref[0]**o+(0.922667710403)*x_ref[0]+(-0.0710905957841)*x_ref[1]**o+(-0.669138163697)*x_ref[1]
        ref[2,1,0,0]=(0.461941211081)*x_ref[0]**o+(0.723022466263)*x_ref[0]+(-0.118551893472)*x_ref[1]**o+(0.00445187353637)*x_ref[1]
        ref[2,1,0,1]=(-0.253168248108)*x_ref[0]**o+(0.476938842605)*x_ref[0]+(0.476641658034)*x_ref[1]**o+(-0.916886387595)*x_ref[1]
        ref[2,1,1,0]=(0.722029309503)*x_ref[0]**o+(0.0588848593425)*x_ref[0]+(0.84737370043)*x_ref[1]**o+(-0.674604415213)*x_ref[1]
        ref[2,1,1,1]=(-0.644736759298)*x_ref[0]**o+(-0.212847303893)*x_ref[0]+(0.25394587552)*x_ref[1]**o+(-0.366955175081)*x_ref[1]
        ref[2,1,2,0]=(-0.657395296801)*x_ref[0]**o+(-0.967570027599)*x_ref[0]+(0.180926628541)*x_ref[1]**o+(-0.940599897942)*x_ref[1]
        ref[2,1,2,1]=(0.849897328801)*x_ref[0]**o+(-0.299199474051)*x_ref[0]+(0.427221207827)*x_ref[1]**o+(0.212027985441)*x_ref[1]
        ref[2,2,0,0]=(-0.561573507381)*x_ref[0]**o+(-0.446925224908)*x_ref[0]+(0.132091224948)*x_ref[1]**o+(-0.673375661849)*x_ref[1]
        ref[2,2,0,1]=(-0.386657291574)*x_ref[0]**o+(0.356784637112)*x_ref[0]+(0.644292006233)*x_ref[1]**o+(0.922937167842)*x_ref[1]
        ref[2,2,1,0]=(0.484446757285)*x_ref[0]**o+(-0.822469870455)*x_ref[0]+(-0.89605890123)*x_ref[1]**o+(-0.895891500031)*x_ref[1]
        ref[2,2,1,1]=(-0.0930740520192)*x_ref[0]**o+(0.953356473678)*x_ref[0]+(-0.714418458806)*x_ref[1]**o+(-0.100584950946)*x_ref[1]
        ref[2,2,2,0]=(0.945572242667)*x_ref[0]**o+(-0.259280849605)*x_ref[0]+(-0.925459846341)*x_ref[1]**o+(0.906002758873)*x_ref[1]
        ref[2,2,2,1]=(-0.638993332975)*x_ref[0]**o+(-0.668594641703)*x_ref[0]+(0.012666231369)*x_ref[1]**o+(0.89732685856)*x_ref[1]
        ref[2,3,0,0]=(-0.885061116144)*x_ref[0]**o+(-0.576097536483)*x_ref[0]+(-0.43166075788)*x_ref[1]**o+(0.538821620917)*x_ref[1]
        ref[2,3,0,1]=(-0.0743638381797)*x_ref[0]**o+(0.251037500035)*x_ref[0]+(-0.905989928076)*x_ref[1]**o+(0.243771000867)*x_ref[1]
        ref[2,3,1,0]=(-0.016045622351)*x_ref[0]**o+(-0.632944072174)*x_ref[0]+(-0.296175918793)*x_ref[1]**o+(-0.00120123641365)*x_ref[1]
        ref[2,3,1,1]=(-0.896354921812)*x_ref[0]**o+(-0.331348989027)*x_ref[0]+(-0.450865285201)*x_ref[1]**o+(-0.243268582009)*x_ref[1]
        ref[2,3,2,0]=(0.609332927431)*x_ref[0]**o+(-0.0815856802115)*x_ref[0]+(0.201924792273)*x_ref[1]**o+(-0.890424451348)*x_ref[1]
        ref[2,3,2,1]=(0.395090137264)*x_ref[0]**o+(0.294230442573)*x_ref[0]+(-0.345371510364)*x_ref[1]**o+(0.181229663241)*x_ref[1]
        ref[2,4,0,0]=(-0.823684509034)*x_ref[0]**o+(0.431019301942)*x_ref[0]+(0.554707015054)*x_ref[1]**o+(-0.846322903003)*x_ref[1]
        ref[2,4,0,1]=(0.862984476658)*x_ref[0]**o+(0.20512500102)*x_ref[0]+(0.0247756159287)*x_ref[1]**o+(-0.374795100946)*x_ref[1]
        ref[2,4,1,0]=(-0.0849914269517)*x_ref[0]**o+(0.754908200339)*x_ref[0]+(0.593363754807)*x_ref[1]**o+(0.124923961126)*x_ref[1]
        ref[2,4,1,1]=(-0.691440264234)*x_ref[0]**o+(0.825739298015)*x_ref[0]+(0.426452231478)*x_ref[1]**o+(0.0995610721341)*x_ref[1]
        ref[2,4,2,0]=(-0.175322750567)*x_ref[0]**o+(0.419750865569)*x_ref[0]+(0.208448987712)*x_ref[1]**o+(0.150584798952)*x_ref[1]
        ref[2,4,2,1]=(0.931680152344)*x_ref[0]**o+(0.390947206307)*x_ref[0]+(0.388437348378)*x_ref[1]**o+(-0.808413079032)*x_ref[1]
        ref[3,0,0,0]=(0.58855431402)*x_ref[0]**o+(0.383865913364)*x_ref[0]+(-0.691445323219)*x_ref[1]**o+(-0.443370658523)*x_ref[1]
        ref[3,0,0,1]=(-0.657371481478)*x_ref[0]**o+(0.746891082673)*x_ref[0]+(-0.829229947773)*x_ref[1]**o+(0.761915670404)*x_ref[1]
        ref[3,0,1,0]=(-0.431502932714)*x_ref[0]**o+(-0.961472285716)*x_ref[0]+(0.0423384388031)*x_ref[1]**o+(0.339373458118)*x_ref[1]
        ref[3,0,1,1]=(-0.272831163259)*x_ref[0]**o+(-0.533438904494)*x_ref[0]+(-0.595625384875)*x_ref[1]**o+(0.0366857022323)*x_ref[1]
        ref[3,0,2,0]=(-0.319684662306)*x_ref[0]**o+(0.672185828912)*x_ref[0]+(0.483126548994)*x_ref[1]**o+(0.745872629604)*x_ref[1]
        ref[3,0,2,1]=(0.437895277404)*x_ref[0]**o+(-0.337209416963)*x_ref[0]+(0.101080522788)*x_ref[1]**o+(0.892022488644)*x_ref[1]
        ref[3,1,0,0]=(-0.582325078455)*x_ref[0]**o+(-0.854689347157)*x_ref[0]+(0.0308638312259)*x_ref[1]**o+(0.633382365704)*x_ref[1]
        ref[3,1,0,1]=(0.319735178376)*x_ref[0]**o+(-0.782764593952)*x_ref[0]+(0.082230132244)*x_ref[1]**o+(-0.946010937852)*x_ref[1]
        ref[3,1,1,0]=(-0.698490410666)*x_ref[0]**o+(-0.801358802689)*x_ref[0]+(0.20695648235)*x_ref[1]**o+(0.75257523471)*x_ref[1]
        ref[3,1,1,1]=(0.66321696527)*x_ref[0]**o+(0.704572918868)*x_ref[0]+(0.528208879002)*x_ref[1]**o+(-0.408378325513)*x_ref[1]
        ref[3,1,2,0]=(-0.925177198417)*x_ref[0]**o+(-0.152427811549)*x_ref[0]+(0.341781993583)*x_ref[1]**o+(-0.589976456377)*x_ref[1]
        ref[3,1,2,1]=(0.206706628353)*x_ref[0]**o+(0.673777173329)*x_ref[0]+(-0.994138314254)*x_ref[1]**o+(-0.843696565863)*x_ref[1]
        ref[3,2,0,0]=(-0.588881067333)*x_ref[0]**o+(-0.537572388341)*x_ref[0]+(-0.206570371675)*x_ref[1]**o+(-0.127468554294)*x_ref[1]
        ref[3,2,0,1]=(0.498638284096)*x_ref[0]**o+(-0.488278831086)*x_ref[0]+(-0.28320874594)*x_ref[1]**o+(-0.638241576615)*x_ref[1]
        ref[3,2,1,0]=(-0.445780790822)*x_ref[0]**o+(-0.347493953089)*x_ref[0]+(-0.558010086232)*x_ref[1]**o+(-0.0287339307689)*x_ref[1]
        ref[3,2,1,1]=(-0.361727811272)*x_ref[0]**o+(-0.559565378007)*x_ref[0]+(-0.211685558972)*x_ref[1]**o+(0.330088396811)*x_ref[1]
        ref[3,2,2,0]=(-0.42600117831)*x_ref[0]**o+(-0.488997111988)*x_ref[0]+(0.127421278051)*x_ref[1]**o+(-0.48233167806)*x_ref[1]
        ref[3,2,2,1]=(0.120468829686)*x_ref[0]**o+(0.0675603906118)*x_ref[0]+(-0.60674361864)*x_ref[1]**o+(0.639115708258)*x_ref[1]
        ref[3,3,0,0]=(0.369343033015)*x_ref[0]**o+(0.707735743593)*x_ref[0]+(-0.426275591836)*x_ref[1]**o+(-0.733377765625)*x_ref[1]
        ref[3,3,0,1]=(-0.931245338156)*x_ref[0]**o+(-0.465861794185)*x_ref[0]+(-0.650664500687)*x_ref[1]**o+(0.103077662402)*x_ref[1]
        ref[3,3,1,0]=(0.731834744512)*x_ref[0]**o+(0.267840571071)*x_ref[0]+(0.952106457495)*x_ref[1]**o+(-0.943845364366)*x_ref[1]
        ref[3,3,1,1]=(-0.21692720916)*x_ref[0]**o+(0.580841374997)*x_ref[0]+(-0.412129341223)*x_ref[1]**o+(0.40414445304)*x_ref[1]
        ref[3,3,2,0]=(-0.447758598726)*x_ref[0]**o+(0.0616926352912)*x_ref[0]+(-0.824624226122)*x_ref[1]**o+(0.19099833104)*x_ref[1]
        ref[3,3,2,1]=(-0.615519138165)*x_ref[0]**o+(0.233926822164)*x_ref[0]+(-0.969014060236)*x_ref[1]**o+(0.835103627893)*x_ref[1]
        ref[3,4,0,0]=(0.550778983301)*x_ref[0]**o+(0.0198253530882)*x_ref[0]+(-0.39383849187)*x_ref[1]**o+(0.367721826304)*x_ref[1]
        ref[3,4,0,1]=(0.444877923149)*x_ref[0]**o+(-0.899159835223)*x_ref[0]+(0.923392540452)*x_ref[1]**o+(0.512905141718)*x_ref[1]
        ref[3,4,1,0]=(-0.563807675145)*x_ref[0]**o+(0.994405383899)*x_ref[0]+(-0.777352948218)*x_ref[1]**o+(0.865994906227)*x_ref[1]
        ref[3,4,1,1]=(-0.593800452139)*x_ref[0]**o+(0.613778803237)*x_ref[0]+(0.582200979718)*x_ref[1]**o+(0.448026243271)*x_ref[1]
        ref[3,4,2,0]=(-0.900331125416)*x_ref[0]**o+(0.234613965424)*x_ref[0]+(0.510400273805)*x_ref[1]**o+(0.0836773641661)*x_ref[1]
        ref[3,4,2,1]=(0.710999097307)*x_ref[0]**o+(0.886603988265)*x_ref[0]+(-0.405644887679)*x_ref[1]**o+(0.362257258809)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.310272924207)*x[0]**o+(-0.664497806351)*x[0]+(0.741942132609)*x[1]**o+(0.636051298779)*x[1]+(0.76627159146)*x[2]**o+(0.492960732588)*x[2]
        arg[0,0,0,1]=(-0.513740801773)*x[0]**o+(-0.633330750909)*x[0]+(0.177934618983)*x[1]**o+(-0.323745723166)*x[1]+(0.694167180904)*x[2]**o+(0.441239251435)*x[2]
        arg[0,0,1,0]=(0.417477695848)*x[0]**o+(-0.923415976002)*x[0]+(0.0765555276662)*x[1]**o+(-0.277290890834)*x[1]+(0.0177191326614)*x[2]**o+(-0.0842520100387)*x[2]
        arg[0,0,1,1]=(-0.761916249894)*x[0]**o+(0.169673139552)*x[0]+(0.0306290828795)*x[1]**o+(0.842234592613)*x[1]+(-0.350342692669)*x[2]**o+(0.321292270658)*x[2]
        arg[0,0,2,0]=(-0.707494137624)*x[0]**o+(0.346873490395)*x[0]+(-0.68281925138)*x[1]**o+(-0.830538083274)*x[1]+(0.329372141613)*x[2]**o+(-0.121366157163)*x[2]
        arg[0,0,2,1]=(-0.708018186507)*x[0]**o+(-0.778051239314)*x[0]+(-0.43106291563)*x[1]**o+(-0.0115914251143)*x[1]+(0.487334215892)*x[2]**o+(-0.289232284658)*x[2]
        arg[0,1,0,0]=(0.530614704199)*x[0]**o+(0.518622028317)*x[0]+(-0.0143129589551)*x[1]**o+(0.808469352257)*x[1]+(-0.156787373225)*x[2]**o+(0.668887813143)*x[2]
        arg[0,1,0,1]=(0.953223233706)*x[0]**o+(-0.164323838353)*x[0]+(0.336717320858)*x[1]**o+(-0.520230488172)*x[1]+(0.689960738002)*x[2]**o+(0.487630947578)*x[2]
        arg[0,1,1,0]=(-0.418095373429)*x[0]**o+(-0.941817844939)*x[0]+(0.88105115607)*x[1]**o+(-0.820502501439)*x[1]+(-0.693858267667)*x[2]**o+(0.530567713127)*x[2]
        arg[0,1,1,1]=(-0.716861320218)*x[0]**o+(-0.736619859467)*x[0]+(-0.885478605601)*x[1]**o+(-0.129375578967)*x[1]+(-0.0181433907318)*x[2]**o+(0.373255537959)*x[2]
        arg[0,1,2,0]=(-0.174742583167)*x[0]**o+(-0.0887967480112)*x[0]+(-0.291168698662)*x[1]**o+(-0.188608506769)*x[1]+(0.22479613454)*x[2]**o+(0.34173928582)*x[2]
        arg[0,1,2,1]=(0.117208929504)*x[0]**o+(-0.420969282079)*x[0]+(0.585298109836)*x[1]**o+(0.771355162745)*x[1]+(-0.51374493747)*x[2]**o+(0.785016548855)*x[2]
        arg[0,2,0,0]=(0.737394887886)*x[0]**o+(0.966080271342)*x[0]+(-0.373178384492)*x[1]**o+(-0.845505386142)*x[1]+(-0.274654519371)*x[2]**o+(0.709356557887)*x[2]
        arg[0,2,0,1]=(0.0365864090636)*x[0]**o+(-0.393382479787)*x[0]+(0.018273360977)*x[1]**o+(-0.509201909328)*x[1]+(-0.777852392534)*x[2]**o+(0.0417280247507)*x[2]
        arg[0,2,1,0]=(0.637829302751)*x[0]**o+(-0.767824790186)*x[0]+(0.843382009511)*x[1]**o+(-0.599633234403)*x[1]+(0.14283138941)*x[2]**o+(-0.354404633991)*x[2]
        arg[0,2,1,1]=(-0.465258613599)*x[0]**o+(0.09341154383)*x[0]+(-0.629114782988)*x[1]**o+(0.91608655383)*x[1]+(-0.692901135848)*x[2]**o+(-0.707225310667)*x[2]
        arg[0,2,2,0]=(0.9935958759)*x[0]**o+(-0.385008742466)*x[0]+(0.153885356854)*x[1]**o+(0.146270680656)*x[1]+(0.305838410343)*x[2]**o+(0.141193552717)*x[2]
        arg[0,2,2,1]=(-0.487538023894)*x[0]**o+(-0.814283442647)*x[0]+(0.166008520199)*x[1]**o+(0.313655774416)*x[1]+(0.282308782973)*x[2]**o+(0.309176062484)*x[2]
        arg[0,3,0,0]=(0.00203507798913)*x[0]**o+(-0.19902542604)*x[0]+(0.88110918624)*x[1]**o+(0.223549219288)*x[1]+(0.135758952369)*x[2]**o+(0.59116783639)*x[2]
        arg[0,3,0,1]=(-0.185063700767)*x[0]**o+(0.833862155173)*x[0]+(-0.0939117509645)*x[1]**o+(0.288003692231)*x[1]+(0.580196157839)*x[2]**o+(0.299320177754)*x[2]
        arg[0,3,1,0]=(-0.039100718826)*x[0]**o+(0.423179869395)*x[0]+(-0.013669578975)*x[1]**o+(-0.355053343085)*x[1]+(0.89285158975)*x[2]**o+(0.392984367877)*x[2]
        arg[0,3,1,1]=(0.0988989416429)*x[0]**o+(-0.446243235174)*x[0]+(-0.691253313617)*x[1]**o+(0.338823693086)*x[1]+(0.707054126334)*x[2]**o+(0.601121388278)*x[2]
        arg[0,3,2,0]=(-0.191366770484)*x[0]**o+(0.961791275903)*x[0]+(-0.103547475035)*x[1]**o+(-0.745236510079)*x[1]+(-0.157506981712)*x[2]**o+(0.986229506787)*x[2]
        arg[0,3,2,1]=(-0.886091883647)*x[0]**o+(0.6979804248)*x[0]+(-0.366575419428)*x[1]**o+(-0.958601153259)*x[1]+(-0.0119724511494)*x[2]**o+(-0.76101132076)*x[2]
        arg[0,4,0,0]=(0.923290554434)*x[0]**o+(0.411359492198)*x[0]+(-0.977999562875)*x[1]**o+(0.352085161527)*x[1]+(0.955745235349)*x[2]**o+(-0.978885609257)*x[2]
        arg[0,4,0,1]=(0.226973479105)*x[0]**o+(-0.423881897251)*x[0]+(0.889101969481)*x[1]**o+(-0.994236389028)*x[1]+(0.0159169033361)*x[2]**o+(-0.559935154203)*x[2]
        arg[0,4,1,0]=(-0.225427291797)*x[0]**o+(-0.557334621848)*x[0]+(-0.486252586142)*x[1]**o+(0.0497120905216)*x[1]+(0.84165736691)*x[2]**o+(-0.185010156797)*x[2]
        arg[0,4,1,1]=(0.762694602609)*x[0]**o+(-0.447367170727)*x[0]+(-0.61103181818)*x[1]**o+(0.689761057656)*x[1]+(-0.631855120122)*x[2]**o+(0.33274655602)*x[2]
        arg[0,4,2,0]=(-0.198884868914)*x[0]**o+(0.716371959278)*x[0]+(0.260871768919)*x[1]**o+(0.622241740082)*x[1]+(0.774103524035)*x[2]**o+(0.186132703676)*x[2]
        arg[0,4,2,1]=(-0.309450645317)*x[0]**o+(0.204754571336)*x[0]+(0.483402810171)*x[1]**o+(0.404620744188)*x[1]+(-0.699137246959)*x[2]**o+(0.467428706196)*x[2]
        arg[1,0,0,0]=(0.642289506448)*x[0]**o+(0.383990304394)*x[0]+(0.556940973318)*x[1]**o+(0.135219729041)*x[1]+(0.577195836358)*x[2]**o+(-0.338924419674)*x[2]
        arg[1,0,0,1]=(-0.920873911295)*x[0]**o+(-0.330610192304)*x[0]+(-0.117226803856)*x[1]**o+(0.172482867684)*x[1]+(0.909140902709)*x[2]**o+(-0.590494167913)*x[2]
        arg[1,0,1,0]=(0.210161538861)*x[0]**o+(-0.83860241761)*x[0]+(0.0257180382661)*x[1]**o+(0.877633854348)*x[1]+(0.26045311754)*x[2]**o+(-0.971349902714)*x[2]
        arg[1,0,1,1]=(0.70736843125)*x[0]**o+(-0.765121985431)*x[0]+(-0.196525225943)*x[1]**o+(-0.175928729346)*x[1]+(0.945523028748)*x[2]**o+(0.877830322823)*x[2]
        arg[1,0,2,0]=(-0.963513016065)*x[0]**o+(0.0820537856351)*x[0]+(0.0122070196118)*x[1]**o+(-0.154147449592)*x[1]+(0.323844633268)*x[2]**o+(-0.124897698214)*x[2]
        arg[1,0,2,1]=(0.656267566405)*x[0]**o+(0.0174786620768)*x[0]+(0.763902435618)*x[1]**o+(-0.986021506043)*x[1]+(0.632625555994)*x[2]**o+(-0.110657743523)*x[2]
        arg[1,1,0,0]=(0.00354144688732)*x[0]**o+(-0.813515808171)*x[0]+(-0.100717717446)*x[1]**o+(0.0543708163629)*x[1]+(0.785927880762)*x[2]**o+(0.0109166652227)*x[2]
        arg[1,1,0,1]=(-0.239705688631)*x[0]**o+(0.402915642509)*x[0]+(0.463371068163)*x[1]**o+(-0.507302462928)*x[1]+(-0.0789844128295)*x[2]**o+(0.669144891499)*x[2]
        arg[1,1,1,0]=(-0.491891367001)*x[0]**o+(-0.524377636928)*x[0]+(0.903091863917)*x[1]**o+(-0.554368807183)*x[1]+(0.332899195339)*x[2]**o+(-0.645216147049)*x[2]
        arg[1,1,1,1]=(-0.915843941243)*x[0]**o+(-0.233175654748)*x[0]+(0.467730425037)*x[1]**o+(-0.765957936683)*x[1]+(0.726678877276)*x[2]**o+(0.770533947942)*x[2]
        arg[1,1,2,0]=(0.678071852032)*x[0]**o+(0.266537844237)*x[0]+(0.324335300151)*x[1]**o+(-0.286713971373)*x[1]+(-0.702384414014)*x[2]**o+(0.887198276519)*x[2]
        arg[1,1,2,1]=(0.0466444959648)*x[0]**o+(0.689162912677)*x[0]+(-0.35033923676)*x[1]**o+(-0.0536380766876)*x[1]+(-0.511833442517)*x[2]**o+(0.125566299829)*x[2]
        arg[1,2,0,0]=(-0.803170431622)*x[0]**o+(0.476211169752)*x[0]+(-0.170510267953)*x[1]**o+(-0.765945735317)*x[1]+(-0.169966329291)*x[2]**o+(-0.988525918724)*x[2]
        arg[1,2,0,1]=(0.289618751212)*x[0]**o+(0.294379040976)*x[0]+(-0.922478416838)*x[1]**o+(0.210546495537)*x[1]+(-0.87266381496)*x[2]**o+(-0.865571669577)*x[2]
        arg[1,2,1,0]=(-0.292811054005)*x[0]**o+(0.651280833746)*x[0]+(-0.625004203736)*x[1]**o+(0.372078169172)*x[1]+(-0.0539099055169)*x[2]**o+(0.0911869270008)*x[2]
        arg[1,2,1,1]=(0.496217698294)*x[0]**o+(-0.986021787133)*x[0]+(0.246417059532)*x[1]**o+(0.610226802401)*x[1]+(0.991534665298)*x[2]**o+(-0.507441185982)*x[2]
        arg[1,2,2,0]=(-0.839902565947)*x[0]**o+(-0.279607147278)*x[0]+(0.278212307963)*x[1]**o+(0.609793640036)*x[1]+(0.940471913774)*x[2]**o+(0.21860828705)*x[2]
        arg[1,2,2,1]=(0.789578112278)*x[0]**o+(-0.0226697818608)*x[0]+(0.216321340291)*x[1]**o+(-0.701818463454)*x[1]+(-0.166689030646)*x[2]**o+(0.966131542292)*x[2]
        arg[1,3,0,0]=(0.149337042551)*x[0]**o+(-0.785050579674)*x[0]+(0.530219924721)*x[1]**o+(-0.971927239398)*x[1]+(0.950917908603)*x[2]**o+(0.132668077447)*x[2]
        arg[1,3,0,1]=(0.983081559724)*x[0]**o+(-0.358694498455)*x[0]+(0.181080099281)*x[1]**o+(-0.699756291643)*x[1]+(-0.744335953304)*x[2]**o+(0.00262043221382)*x[2]
        arg[1,3,1,0]=(0.31776032045)*x[0]**o+(-0.0726184220029)*x[0]+(-0.759883735971)*x[1]**o+(-0.6389710873)*x[1]+(-0.607858499502)*x[2]**o+(-0.397169895784)*x[2]
        arg[1,3,1,1]=(-0.950468389728)*x[0]**o+(0.831121915521)*x[0]+(0.552857011083)*x[1]**o+(0.848599672594)*x[1]+(0.432183564231)*x[2]**o+(0.57537721362)*x[2]
        arg[1,3,2,0]=(0.736080143601)*x[0]**o+(-0.804140307747)*x[0]+(-0.163864867884)*x[1]**o+(0.950874042315)*x[1]+(-0.99676139412)*x[2]**o+(0.332320294246)*x[2]
        arg[1,3,2,1]=(0.119082582041)*x[0]**o+(-0.969740778852)*x[0]+(0.828432583532)*x[1]**o+(0.190324050395)*x[1]+(-0.42639142674)*x[2]**o+(0.315519578192)*x[2]
        arg[1,4,0,0]=(-0.798465141775)*x[0]**o+(-0.717282977908)*x[0]+(0.935655148713)*x[1]**o+(-0.655752964484)*x[1]+(0.159990149099)*x[2]**o+(-0.537115062687)*x[2]
        arg[1,4,0,1]=(0.445151775007)*x[0]**o+(-0.867956302051)*x[0]+(-0.809856879825)*x[1]**o+(-0.326897750981)*x[1]+(0.756352311987)*x[2]**o+(0.577585513459)*x[2]
        arg[1,4,1,0]=(0.294243347591)*x[0]**o+(0.774741656931)*x[0]+(-0.0799985737984)*x[1]**o+(0.179789107412)*x[1]+(-0.612198025189)*x[2]**o+(-0.07086009611)*x[2]
        arg[1,4,1,1]=(0.0692586287045)*x[0]**o+(-0.671956300122)*x[0]+(-0.948667104683)*x[1]**o+(0.68656246801)*x[1]+(-0.352975297696)*x[2]**o+(0.277952113716)*x[2]
        arg[1,4,2,0]=(-0.491316095517)*x[0]**o+(0.725899087239)*x[0]+(0.634113963431)*x[1]**o+(0.235694393453)*x[1]+(-0.15600720713)*x[2]**o+(0.954814976539)*x[2]
        arg[1,4,2,1]=(0.120206996339)*x[0]**o+(-0.86151411823)*x[0]+(-0.137093650097)*x[1]**o+(0.633402978246)*x[1]+(-0.42515364728)*x[2]**o+(-0.987634210617)*x[2]
        arg[2,0,0,0]=(0.539463971222)*x[0]**o+(-0.03435650579)*x[0]+(-0.884395295544)*x[1]**o+(0.215454879022)*x[1]+(0.842377328246)*x[2]**o+(0.370920136871)*x[2]
        arg[2,0,0,1]=(-0.411505463039)*x[0]**o+(0.446494039439)*x[0]+(0.0313771969556)*x[1]**o+(0.602893223061)*x[1]+(-0.30383490021)*x[2]**o+(-0.788028756758)*x[2]
        arg[2,0,1,0]=(0.489009615999)*x[0]**o+(-0.95508248016)*x[0]+(-0.21489475022)*x[1]**o+(-0.857263478431)*x[1]+(-0.123703409045)*x[2]**o+(0.663389562431)*x[2]
        arg[2,0,1,1]=(-0.0224763310929)*x[0]**o+(-0.315886249237)*x[0]+(-0.318201849695)*x[1]**o+(0.423732682462)*x[1]+(0.396065636381)*x[2]**o+(0.0816717878865)*x[2]
        arg[2,0,2,0]=(-0.191432618224)*x[0]**o+(0.688389390795)*x[0]+(-0.382948112159)*x[1]**o+(0.562905173818)*x[1]+(-0.118396714029)*x[2]**o+(0.234478642577)*x[2]
        arg[2,0,2,1]=(-0.348008173171)*x[0]**o+(0.191994166636)*x[0]+(0.450556057147)*x[1]**o+(-0.710821693343)*x[1]+(-0.245119333857)*x[2]**o+(0.262869525239)*x[2]
        arg[2,1,0,0]=(-0.160444506203)*x[0]**o+(-0.816142837432)*x[0]+(-0.619722480492)*x[1]**o+(-0.989077039675)*x[1]+(-0.0282349103977)*x[2]**o+(-0.629849819691)*x[2]
        arg[2,1,0,1]=(0.379513765297)*x[0]**o+(0.934652105586)*x[0]+(-0.632636427868)*x[1]**o+(-0.0492550496553)*x[1]+(0.137733572307)*x[2]**o+(0.779782637942)*x[2]
        arg[2,1,1,0]=(0.469499320417)*x[0]**o+(0.570880924125)*x[0]+(-0.0329833251419)*x[1]**o+(-0.0388751034604)*x[1]+(-0.878520094389)*x[2]**o+(0.992656649048)*x[2]
        arg[2,1,1,1]=(0.592005181995)*x[0]**o+(-0.553980493526)*x[0]+(0.359837665019)*x[1]**o+(0.811107941179)*x[1]+(-0.441929650912)*x[2]**o+(-0.768618378636)*x[2]
        arg[2,1,2,0]=(0.741376258816)*x[0]**o+(-0.238909756717)*x[0]+(0.204631082449)*x[1]**o+(0.970501662949)*x[1]+(-0.615638092239)*x[2]**o+(-0.575948750844)*x[2]
        arg[2,1,2,1]=(0.774050933112)*x[0]**o+(-0.872234076721)*x[0]+(-0.0936189240061)*x[1]**o+(-0.773608756961)*x[1]+(-0.119171907808)*x[2]**o+(-0.256124796845)*x[2]
        arg[2,2,0,0]=(0.459481754425)*x[0]**o+(0.478364929493)*x[0]+(0.460199733287)*x[1]**o+(0.66880294659)*x[1]+(-0.859525058389)*x[2]**o+(0.77960284414)*x[2]
        arg[2,2,0,1]=(0.690557835207)*x[0]**o+(-0.812305038592)*x[0]+(0.837924124791)*x[1]**o+(-0.785979206364)*x[1]+(-0.295553949819)*x[2]**o+(0.575096295897)*x[2]
        arg[2,2,1,0]=(0.443193683205)*x[0]**o+(0.0355033222253)*x[0]+(0.497641002766)*x[1]**o+(-0.636955142451)*x[1]+(-0.265324357454)*x[2]**o+(0.734148569078)*x[2]
        arg[2,2,1,1]=(-0.088707242599)*x[0]**o+(0.513403518669)*x[0]+(0.337741737651)*x[1]**o+(-0.174843924478)*x[1]+(0.687291255986)*x[2]**o+(0.473850502258)*x[2]
        arg[2,2,2,0]=(0.5305624642)*x[0]**o+(-0.972997835953)*x[0]+(0.00197107287888)*x[1]**o+(-0.75490872634)*x[1]+(-0.0569624569435)*x[2]**o+(-0.0780532868631)*x[2]
        arg[2,2,2,1]=(-0.744584471866)*x[0]**o+(0.915494909272)*x[0]+(0.239235050038)*x[1]**o+(-0.192466753187)*x[1]+(0.933556550824)*x[2]**o+(-0.786420251487)*x[2]
        arg[2,3,0,0]=(0.472114530627)*x[0]**o+(0.0693211807557)*x[0]+(-0.274273425713)*x[1]**o+(0.136242554894)*x[1]+(-0.329240675785)*x[2]**o+(0.0212539247343)*x[2]
        arg[2,3,0,1]=(-0.780426108099)*x[0]**o+(-0.632382551025)*x[0]+(0.554926949239)*x[1]**o+(0.264204009059)*x[1]+(-0.868007525824)*x[2]**o+(-0.937631120201)*x[2]
        arg[2,3,1,0]=(-0.826692080947)*x[0]**o+(0.915241858538)*x[0]+(-0.49986160859)*x[1]**o+(0.921686464658)*x[1]+(0.71996666304)*x[2]**o+(0.297609501096)*x[2]
        arg[2,3,1,1]=(0.991896994614)*x[0]**o+(-0.543310008103)*x[0]+(0.0791899631111)*x[1]**o+(-0.914088200145)*x[1]+(-0.496337624951)*x[2]**o+(0.708950898614)*x[2]
        arg[2,3,2,0]=(-0.235116966432)*x[0]**o+(-0.492848730207)*x[0]+(0.428728646789)*x[1]**o+(-0.410747634691)*x[1]+(0.705059815458)*x[2]**o+(-0.185604633831)*x[2]
        arg[2,3,2,1]=(0.838829037117)*x[0]**o+(0.670561147563)*x[0]+(0.621407965075)*x[1]**o+(-0.10624946305)*x[1]+(0.893703402687)*x[2]**o+(0.985504995821)*x[2]
        arg[2,4,0,0]=(0.635072504682)*x[0]**o+(-0.681962427555)*x[0]+(-0.00774518011081)*x[1]**o+(-0.264996228668)*x[1]+(0.690668403198)*x[2]**o+(0.883573761062)*x[2]
        arg[2,4,0,1]=(0.545632268651)*x[0]**o+(0.439668519394)*x[0]+(-0.482479808491)*x[1]**o+(0.752028962391)*x[1]+(0.0596773623939)*x[2]**o+(-0.369987181127)*x[2]
        arg[2,4,1,0]=(0.659005322796)*x[0]**o+(0.503940833428)*x[0]+(0.643633321586)*x[1]**o+(0.927082323252)*x[1]+(0.354072146184)*x[2]**o+(-0.638737560608)*x[2]
        arg[2,4,1,1]=(-0.443548860913)*x[0]**o+(0.707607048458)*x[0]+(-0.955225746402)*x[1]**o+(0.515995563585)*x[1]+(-0.288406637165)*x[2]**o+(-0.760864937318)*x[2]
        arg[2,4,2,0]=(0.898146580355)*x[0]**o+(-0.590650787106)*x[0]+(0.036095789753)*x[1]**o+(0.326040260581)*x[1]+(0.680069361306)*x[2]**o+(0.333839099341)*x[2]
        arg[2,4,2,1]=(-0.151364260271)*x[0]**o+(-0.675198183832)*x[0]+(-0.183626298978)*x[1]**o+(-0.182565217146)*x[1]+(0.0345632129814)*x[2]**o+(-0.581534100763)*x[2]
        arg[3,0,0,0]=(-0.942351703246)*x[0]**o+(-0.699567649453)*x[0]+(-0.832068665702)*x[1]**o+(0.163439416504)*x[1]+(0.742500339519)*x[2]**o+(-0.492882887489)*x[2]
        arg[3,0,0,1]=(0.990078296484)*x[0]**o+(-0.999837336587)*x[0]+(0.197589456459)*x[1]**o+(0.431403059019)*x[1]+(0.859864015871)*x[2]**o+(-0.267474829282)*x[2]
        arg[3,0,1,0]=(-0.776773660723)*x[0]**o+(-0.189048370486)*x[0]+(0.926953856802)*x[1]**o+(-0.72314744652)*x[1]+(0.330064363809)*x[2]**o+(0.188533334359)*x[2]
        arg[3,0,1,1]=(-0.854830058762)*x[0]**o+(0.832073883276)*x[0]+(-0.425691165124)*x[1]**o+(0.162351500316)*x[1]+(0.370604864512)*x[2]**o+(-0.581907071484)*x[2]
        arg[3,0,2,0]=(-0.316716909548)*x[0]**o+(0.335417284383)*x[0]+(-0.783061160352)*x[1]**o+(-0.253972519239)*x[1]+(-0.826432933721)*x[2]**o+(0.267368378363)*x[2]
        arg[3,0,2,1]=(-0.765899705462)*x[0]**o+(-0.0993875861476)*x[0]+(-0.211785004352)*x[1]**o+(-0.635330151497)*x[1]+(0.414575412599)*x[2]**o+(0.586882274623)*x[2]
        arg[3,1,0,0]=(0.987776377237)*x[0]**o+(0.796660669191)*x[0]+(0.580724891862)*x[1]**o+(-0.24149253995)*x[1]+(-0.633421016909)*x[2]**o+(-0.220110913711)*x[2]
        arg[3,1,0,1]=(-0.182240918081)*x[0]**o+(-0.290526050572)*x[0]+(0.00283230959976)*x[1]**o+(-0.933282166805)*x[1]+(-0.857893235883)*x[2]**o+(0.551414074127)*x[2]
        arg[3,1,1,0]=(-0.854435997816)*x[0]**o+(0.706455144263)*x[0]+(0.406237965254)*x[1]**o+(0.192071142415)*x[1]+(0.357966796834)*x[2]**o+(0.887011087987)*x[2]
        arg[3,1,1,1]=(0.832013396996)*x[0]**o+(0.696133072321)*x[0]+(0.682726968712)*x[1]**o+(0.613316240514)*x[1]+(0.12600339887)*x[2]**o+(-0.620088231382)*x[2]
        arg[3,1,2,0]=(-0.178681615221)*x[0]**o+(-0.952835986168)*x[0]+(0.299692693746)*x[1]**o+(0.892451908416)*x[1]+(0.202403273634)*x[2]**o+(0.677212897943)*x[2]
        arg[3,1,2,1]=(0.0592594053342)*x[0]**o+(0.56391153959)*x[0]+(-0.874866627667)*x[1]**o+(0.432102900769)*x[1]+(-0.49073126285)*x[2]**o+(-0.010495860028)*x[2]
        arg[3,2,0,0]=(0.548006445814)*x[0]**o+(0.667313392202)*x[0]+(0.397805355499)*x[1]**o+(-0.788100267808)*x[1]+(0.869234055258)*x[2]**o+(-0.830241520401)*x[2]
        arg[3,2,0,1]=(0.72586139216)*x[0]**o+(-0.931878957878)*x[0]+(-0.94119555858)*x[1]**o+(-0.133544806169)*x[1]+(0.754263323382)*x[2]**o+(-0.446604740157)*x[2]
        arg[3,2,1,0]=(-0.28135830707)*x[0]**o+(-0.478744058518)*x[0]+(0.88219033731)*x[1]**o+(0.460587286945)*x[1]+(-0.849692146096)*x[2]**o+(0.0927701345107)*x[2]
        arg[3,2,1,1]=(-0.681859641611)*x[0]**o+(-0.670164293088)*x[0]+(0.796366374435)*x[1]**o+(0.837230317054)*x[1]+(-0.413283915904)*x[2]**o+(0.779624227187)*x[2]
        arg[3,2,2,0]=(-0.803242771255)*x[0]**o+(-0.0687616271776)*x[0]+(-0.948856027999)*x[1]**o+(0.762568790931)*x[1]+(0.937588292604)*x[2]**o+(-0.0224232611283)*x[2]
        arg[3,2,2,1]=(-0.215752944937)*x[0]**o+(-0.807114296075)*x[0]+(0.970625375964)*x[1]**o+(0.68578124692)*x[1]+(-0.165007047419)*x[2]**o+(0.942718439981)*x[2]
        arg[3,3,0,0]=(0.152001104716)*x[0]**o+(-0.657242342357)*x[0]+(0.559608661705)*x[1]**o+(0.0764403273251)*x[1]+(0.633414754695)*x[2]**o+(0.430777868344)*x[2]
        arg[3,3,0,1]=(-0.898109497759)*x[0]**o+(0.736574249406)*x[0]+(-0.569584184574)*x[1]**o+(-0.145501395506)*x[1]+(-0.412424116475)*x[2]**o+(-0.416350923096)*x[2]
        arg[3,3,1,0]=(0.602022795614)*x[0]**o+(-0.848525569825)*x[0]+(-0.713426287505)*x[1]**o+(-0.829278452687)*x[1]+(-0.491257659316)*x[2]**o+(0.90219908999)*x[2]
        arg[3,3,1,1]=(-0.769110216066)*x[0]**o+(-0.789464037338)*x[0]+(-0.363686172584)*x[1]**o+(-0.86190787068)*x[1]+(0.275438119232)*x[2]**o+(0.285209834248)*x[2]
        arg[3,3,2,0]=(0.351662405853)*x[0]**o+(0.8403302307)*x[0]+(-0.449869332519)*x[1]**o+(0.936043614838)*x[1]+(-0.656908777335)*x[2]**o+(-0.231923872531)*x[2]
        arg[3,3,2,1]=(0.873848945653)*x[0]**o+(0.551429883959)*x[0]+(-0.577942548701)*x[1]**o+(-0.254317163049)*x[1]+(0.64878089878)*x[2]**o+(0.7135339515)*x[2]
        arg[3,4,0,0]=(0.891435610494)*x[0]**o+(0.598893704498)*x[0]+(0.24682094756)*x[1]**o+(0.0895612710338)*x[1]+(-0.513686244953)*x[2]**o+(0.307205505502)*x[2]
        arg[3,4,0,1]=(-0.577752944196)*x[0]**o+(-0.458383751623)*x[0]+(-0.913956988143)*x[1]**o+(0.470857602235)*x[1]+(0.631726850665)*x[2]**o+(-0.112962289904)*x[2]
        arg[3,4,1,0]=(-0.65648177489)*x[0]**o+(-0.843530972747)*x[0]+(-0.0332153723809)*x[1]**o+(0.230484263355)*x[1]+(-0.865687508592)*x[2]**o+(0.122751735082)*x[2]
        arg[3,4,1,1]=(-0.353502640734)*x[0]**o+(-0.816020900282)*x[0]+(0.943688796868)*x[1]**o+(-0.526707881958)*x[1]+(-0.790258298117)*x[2]**o+(-0.748706809042)*x[2]
        arg[3,4,2,0]=(-0.596345912175)*x[0]**o+(0.872771121837)*x[0]+(0.100107395575)*x[1]**o+(0.950482249217)*x[1]+(0.0487634582783)*x[2]**o+(-0.0285218097887)*x[2]
        arg[3,4,2,1]=(-0.827443553933)*x[0]**o+(-0.922564800149)*x[0]+(-0.810354385287)*x[1]**o+(0.630079217293)*x[1]+(-0.316280559723)*x[2]**o+(0.977830435871)*x[2]
        ref[0,0,0,0]=(0.310272924207)*x_ref[0]**o+(-0.664497806351)*x_ref[0]+(0.741942132609)*x_ref[1]**o+(0.636051298779)*x_ref[1]+(0.76627159146)*x_ref[2]**o+(0.492960732588)*x_ref[2]
        ref[0,0,0,1]=(-0.513740801773)*x_ref[0]**o+(-0.633330750909)*x_ref[0]+(0.177934618983)*x_ref[1]**o+(-0.323745723166)*x_ref[1]+(0.694167180904)*x_ref[2]**o+(0.441239251435)*x_ref[2]
        ref[0,0,1,0]=(0.417477695848)*x_ref[0]**o+(-0.923415976002)*x_ref[0]+(0.0765555276662)*x_ref[1]**o+(-0.277290890834)*x_ref[1]+(0.0177191326614)*x_ref[2]**o+(-0.0842520100387)*x_ref[2]
        ref[0,0,1,1]=(-0.761916249894)*x_ref[0]**o+(0.169673139552)*x_ref[0]+(0.0306290828795)*x_ref[1]**o+(0.842234592613)*x_ref[1]+(-0.350342692669)*x_ref[2]**o+(0.321292270658)*x_ref[2]
        ref[0,0,2,0]=(-0.707494137624)*x_ref[0]**o+(0.346873490395)*x_ref[0]+(-0.68281925138)*x_ref[1]**o+(-0.830538083274)*x_ref[1]+(0.329372141613)*x_ref[2]**o+(-0.121366157163)*x_ref[2]
        ref[0,0,2,1]=(-0.708018186507)*x_ref[0]**o+(-0.778051239314)*x_ref[0]+(-0.43106291563)*x_ref[1]**o+(-0.0115914251143)*x_ref[1]+(0.487334215892)*x_ref[2]**o+(-0.289232284658)*x_ref[2]
        ref[0,1,0,0]=(0.530614704199)*x_ref[0]**o+(0.518622028317)*x_ref[0]+(-0.0143129589551)*x_ref[1]**o+(0.808469352257)*x_ref[1]+(-0.156787373225)*x_ref[2]**o+(0.668887813143)*x_ref[2]
        ref[0,1,0,1]=(0.953223233706)*x_ref[0]**o+(-0.164323838353)*x_ref[0]+(0.336717320858)*x_ref[1]**o+(-0.520230488172)*x_ref[1]+(0.689960738002)*x_ref[2]**o+(0.487630947578)*x_ref[2]
        ref[0,1,1,0]=(-0.418095373429)*x_ref[0]**o+(-0.941817844939)*x_ref[0]+(0.88105115607)*x_ref[1]**o+(-0.820502501439)*x_ref[1]+(-0.693858267667)*x_ref[2]**o+(0.530567713127)*x_ref[2]
        ref[0,1,1,1]=(-0.716861320218)*x_ref[0]**o+(-0.736619859467)*x_ref[0]+(-0.885478605601)*x_ref[1]**o+(-0.129375578967)*x_ref[1]+(-0.0181433907318)*x_ref[2]**o+(0.373255537959)*x_ref[2]
        ref[0,1,2,0]=(-0.174742583167)*x_ref[0]**o+(-0.0887967480112)*x_ref[0]+(-0.291168698662)*x_ref[1]**o+(-0.188608506769)*x_ref[1]+(0.22479613454)*x_ref[2]**o+(0.34173928582)*x_ref[2]
        ref[0,1,2,1]=(0.117208929504)*x_ref[0]**o+(-0.420969282079)*x_ref[0]+(0.585298109836)*x_ref[1]**o+(0.771355162745)*x_ref[1]+(-0.51374493747)*x_ref[2]**o+(0.785016548855)*x_ref[2]
        ref[0,2,0,0]=(0.737394887886)*x_ref[0]**o+(0.966080271342)*x_ref[0]+(-0.373178384492)*x_ref[1]**o+(-0.845505386142)*x_ref[1]+(-0.274654519371)*x_ref[2]**o+(0.709356557887)*x_ref[2]
        ref[0,2,0,1]=(0.0365864090636)*x_ref[0]**o+(-0.393382479787)*x_ref[0]+(0.018273360977)*x_ref[1]**o+(-0.509201909328)*x_ref[1]+(-0.777852392534)*x_ref[2]**o+(0.0417280247507)*x_ref[2]
        ref[0,2,1,0]=(0.637829302751)*x_ref[0]**o+(-0.767824790186)*x_ref[0]+(0.843382009511)*x_ref[1]**o+(-0.599633234403)*x_ref[1]+(0.14283138941)*x_ref[2]**o+(-0.354404633991)*x_ref[2]
        ref[0,2,1,1]=(-0.465258613599)*x_ref[0]**o+(0.09341154383)*x_ref[0]+(-0.629114782988)*x_ref[1]**o+(0.91608655383)*x_ref[1]+(-0.692901135848)*x_ref[2]**o+(-0.707225310667)*x_ref[2]
        ref[0,2,2,0]=(0.9935958759)*x_ref[0]**o+(-0.385008742466)*x_ref[0]+(0.153885356854)*x_ref[1]**o+(0.146270680656)*x_ref[1]+(0.305838410343)*x_ref[2]**o+(0.141193552717)*x_ref[2]
        ref[0,2,2,1]=(-0.487538023894)*x_ref[0]**o+(-0.814283442647)*x_ref[0]+(0.166008520199)*x_ref[1]**o+(0.313655774416)*x_ref[1]+(0.282308782973)*x_ref[2]**o+(0.309176062484)*x_ref[2]
        ref[0,3,0,0]=(0.00203507798913)*x_ref[0]**o+(-0.19902542604)*x_ref[0]+(0.88110918624)*x_ref[1]**o+(0.223549219288)*x_ref[1]+(0.135758952369)*x_ref[2]**o+(0.59116783639)*x_ref[2]
        ref[0,3,0,1]=(-0.185063700767)*x_ref[0]**o+(0.833862155173)*x_ref[0]+(-0.0939117509645)*x_ref[1]**o+(0.288003692231)*x_ref[1]+(0.580196157839)*x_ref[2]**o+(0.299320177754)*x_ref[2]
        ref[0,3,1,0]=(-0.039100718826)*x_ref[0]**o+(0.423179869395)*x_ref[0]+(-0.013669578975)*x_ref[1]**o+(-0.355053343085)*x_ref[1]+(0.89285158975)*x_ref[2]**o+(0.392984367877)*x_ref[2]
        ref[0,3,1,1]=(0.0988989416429)*x_ref[0]**o+(-0.446243235174)*x_ref[0]+(-0.691253313617)*x_ref[1]**o+(0.338823693086)*x_ref[1]+(0.707054126334)*x_ref[2]**o+(0.601121388278)*x_ref[2]
        ref[0,3,2,0]=(-0.191366770484)*x_ref[0]**o+(0.961791275903)*x_ref[0]+(-0.103547475035)*x_ref[1]**o+(-0.745236510079)*x_ref[1]+(-0.157506981712)*x_ref[2]**o+(0.986229506787)*x_ref[2]
        ref[0,3,2,1]=(-0.886091883647)*x_ref[0]**o+(0.6979804248)*x_ref[0]+(-0.366575419428)*x_ref[1]**o+(-0.958601153259)*x_ref[1]+(-0.0119724511494)*x_ref[2]**o+(-0.76101132076)*x_ref[2]
        ref[0,4,0,0]=(0.923290554434)*x_ref[0]**o+(0.411359492198)*x_ref[0]+(-0.977999562875)*x_ref[1]**o+(0.352085161527)*x_ref[1]+(0.955745235349)*x_ref[2]**o+(-0.978885609257)*x_ref[2]
        ref[0,4,0,1]=(0.226973479105)*x_ref[0]**o+(-0.423881897251)*x_ref[0]+(0.889101969481)*x_ref[1]**o+(-0.994236389028)*x_ref[1]+(0.0159169033361)*x_ref[2]**o+(-0.559935154203)*x_ref[2]
        ref[0,4,1,0]=(-0.225427291797)*x_ref[0]**o+(-0.557334621848)*x_ref[0]+(-0.486252586142)*x_ref[1]**o+(0.0497120905216)*x_ref[1]+(0.84165736691)*x_ref[2]**o+(-0.185010156797)*x_ref[2]
        ref[0,4,1,1]=(0.762694602609)*x_ref[0]**o+(-0.447367170727)*x_ref[0]+(-0.61103181818)*x_ref[1]**o+(0.689761057656)*x_ref[1]+(-0.631855120122)*x_ref[2]**o+(0.33274655602)*x_ref[2]
        ref[0,4,2,0]=(-0.198884868914)*x_ref[0]**o+(0.716371959278)*x_ref[0]+(0.260871768919)*x_ref[1]**o+(0.622241740082)*x_ref[1]+(0.774103524035)*x_ref[2]**o+(0.186132703676)*x_ref[2]
        ref[0,4,2,1]=(-0.309450645317)*x_ref[0]**o+(0.204754571336)*x_ref[0]+(0.483402810171)*x_ref[1]**o+(0.404620744188)*x_ref[1]+(-0.699137246959)*x_ref[2]**o+(0.467428706196)*x_ref[2]
        ref[1,0,0,0]=(0.642289506448)*x_ref[0]**o+(0.383990304394)*x_ref[0]+(0.556940973318)*x_ref[1]**o+(0.135219729041)*x_ref[1]+(0.577195836358)*x_ref[2]**o+(-0.338924419674)*x_ref[2]
        ref[1,0,0,1]=(-0.920873911295)*x_ref[0]**o+(-0.330610192304)*x_ref[0]+(-0.117226803856)*x_ref[1]**o+(0.172482867684)*x_ref[1]+(0.909140902709)*x_ref[2]**o+(-0.590494167913)*x_ref[2]
        ref[1,0,1,0]=(0.210161538861)*x_ref[0]**o+(-0.83860241761)*x_ref[0]+(0.0257180382661)*x_ref[1]**o+(0.877633854348)*x_ref[1]+(0.26045311754)*x_ref[2]**o+(-0.971349902714)*x_ref[2]
        ref[1,0,1,1]=(0.70736843125)*x_ref[0]**o+(-0.765121985431)*x_ref[0]+(-0.196525225943)*x_ref[1]**o+(-0.175928729346)*x_ref[1]+(0.945523028748)*x_ref[2]**o+(0.877830322823)*x_ref[2]
        ref[1,0,2,0]=(-0.963513016065)*x_ref[0]**o+(0.0820537856351)*x_ref[0]+(0.0122070196118)*x_ref[1]**o+(-0.154147449592)*x_ref[1]+(0.323844633268)*x_ref[2]**o+(-0.124897698214)*x_ref[2]
        ref[1,0,2,1]=(0.656267566405)*x_ref[0]**o+(0.0174786620768)*x_ref[0]+(0.763902435618)*x_ref[1]**o+(-0.986021506043)*x_ref[1]+(0.632625555994)*x_ref[2]**o+(-0.110657743523)*x_ref[2]
        ref[1,1,0,0]=(0.00354144688732)*x_ref[0]**o+(-0.813515808171)*x_ref[0]+(-0.100717717446)*x_ref[1]**o+(0.0543708163629)*x_ref[1]+(0.785927880762)*x_ref[2]**o+(0.0109166652227)*x_ref[2]
        ref[1,1,0,1]=(-0.239705688631)*x_ref[0]**o+(0.402915642509)*x_ref[0]+(0.463371068163)*x_ref[1]**o+(-0.507302462928)*x_ref[1]+(-0.0789844128295)*x_ref[2]**o+(0.669144891499)*x_ref[2]
        ref[1,1,1,0]=(-0.491891367001)*x_ref[0]**o+(-0.524377636928)*x_ref[0]+(0.903091863917)*x_ref[1]**o+(-0.554368807183)*x_ref[1]+(0.332899195339)*x_ref[2]**o+(-0.645216147049)*x_ref[2]
        ref[1,1,1,1]=(-0.915843941243)*x_ref[0]**o+(-0.233175654748)*x_ref[0]+(0.467730425037)*x_ref[1]**o+(-0.765957936683)*x_ref[1]+(0.726678877276)*x_ref[2]**o+(0.770533947942)*x_ref[2]
        ref[1,1,2,0]=(0.678071852032)*x_ref[0]**o+(0.266537844237)*x_ref[0]+(0.324335300151)*x_ref[1]**o+(-0.286713971373)*x_ref[1]+(-0.702384414014)*x_ref[2]**o+(0.887198276519)*x_ref[2]
        ref[1,1,2,1]=(0.0466444959648)*x_ref[0]**o+(0.689162912677)*x_ref[0]+(-0.35033923676)*x_ref[1]**o+(-0.0536380766876)*x_ref[1]+(-0.511833442517)*x_ref[2]**o+(0.125566299829)*x_ref[2]
        ref[1,2,0,0]=(-0.803170431622)*x_ref[0]**o+(0.476211169752)*x_ref[0]+(-0.170510267953)*x_ref[1]**o+(-0.765945735317)*x_ref[1]+(-0.169966329291)*x_ref[2]**o+(-0.988525918724)*x_ref[2]
        ref[1,2,0,1]=(0.289618751212)*x_ref[0]**o+(0.294379040976)*x_ref[0]+(-0.922478416838)*x_ref[1]**o+(0.210546495537)*x_ref[1]+(-0.87266381496)*x_ref[2]**o+(-0.865571669577)*x_ref[2]
        ref[1,2,1,0]=(-0.292811054005)*x_ref[0]**o+(0.651280833746)*x_ref[0]+(-0.625004203736)*x_ref[1]**o+(0.372078169172)*x_ref[1]+(-0.0539099055169)*x_ref[2]**o+(0.0911869270008)*x_ref[2]
        ref[1,2,1,1]=(0.496217698294)*x_ref[0]**o+(-0.986021787133)*x_ref[0]+(0.246417059532)*x_ref[1]**o+(0.610226802401)*x_ref[1]+(0.991534665298)*x_ref[2]**o+(-0.507441185982)*x_ref[2]
        ref[1,2,2,0]=(-0.839902565947)*x_ref[0]**o+(-0.279607147278)*x_ref[0]+(0.278212307963)*x_ref[1]**o+(0.609793640036)*x_ref[1]+(0.940471913774)*x_ref[2]**o+(0.21860828705)*x_ref[2]
        ref[1,2,2,1]=(0.789578112278)*x_ref[0]**o+(-0.0226697818608)*x_ref[0]+(0.216321340291)*x_ref[1]**o+(-0.701818463454)*x_ref[1]+(-0.166689030646)*x_ref[2]**o+(0.966131542292)*x_ref[2]
        ref[1,3,0,0]=(0.149337042551)*x_ref[0]**o+(-0.785050579674)*x_ref[0]+(0.530219924721)*x_ref[1]**o+(-0.971927239398)*x_ref[1]+(0.950917908603)*x_ref[2]**o+(0.132668077447)*x_ref[2]
        ref[1,3,0,1]=(0.983081559724)*x_ref[0]**o+(-0.358694498455)*x_ref[0]+(0.181080099281)*x_ref[1]**o+(-0.699756291643)*x_ref[1]+(-0.744335953304)*x_ref[2]**o+(0.00262043221382)*x_ref[2]
        ref[1,3,1,0]=(0.31776032045)*x_ref[0]**o+(-0.0726184220029)*x_ref[0]+(-0.759883735971)*x_ref[1]**o+(-0.6389710873)*x_ref[1]+(-0.607858499502)*x_ref[2]**o+(-0.397169895784)*x_ref[2]
        ref[1,3,1,1]=(-0.950468389728)*x_ref[0]**o+(0.831121915521)*x_ref[0]+(0.552857011083)*x_ref[1]**o+(0.848599672594)*x_ref[1]+(0.432183564231)*x_ref[2]**o+(0.57537721362)*x_ref[2]
        ref[1,3,2,0]=(0.736080143601)*x_ref[0]**o+(-0.804140307747)*x_ref[0]+(-0.163864867884)*x_ref[1]**o+(0.950874042315)*x_ref[1]+(-0.99676139412)*x_ref[2]**o+(0.332320294246)*x_ref[2]
        ref[1,3,2,1]=(0.119082582041)*x_ref[0]**o+(-0.969740778852)*x_ref[0]+(0.828432583532)*x_ref[1]**o+(0.190324050395)*x_ref[1]+(-0.42639142674)*x_ref[2]**o+(0.315519578192)*x_ref[2]
        ref[1,4,0,0]=(-0.798465141775)*x_ref[0]**o+(-0.717282977908)*x_ref[0]+(0.935655148713)*x_ref[1]**o+(-0.655752964484)*x_ref[1]+(0.159990149099)*x_ref[2]**o+(-0.537115062687)*x_ref[2]
        ref[1,4,0,1]=(0.445151775007)*x_ref[0]**o+(-0.867956302051)*x_ref[0]+(-0.809856879825)*x_ref[1]**o+(-0.326897750981)*x_ref[1]+(0.756352311987)*x_ref[2]**o+(0.577585513459)*x_ref[2]
        ref[1,4,1,0]=(0.294243347591)*x_ref[0]**o+(0.774741656931)*x_ref[0]+(-0.0799985737984)*x_ref[1]**o+(0.179789107412)*x_ref[1]+(-0.612198025189)*x_ref[2]**o+(-0.07086009611)*x_ref[2]
        ref[1,4,1,1]=(0.0692586287045)*x_ref[0]**o+(-0.671956300122)*x_ref[0]+(-0.948667104683)*x_ref[1]**o+(0.68656246801)*x_ref[1]+(-0.352975297696)*x_ref[2]**o+(0.277952113716)*x_ref[2]
        ref[1,4,2,0]=(-0.491316095517)*x_ref[0]**o+(0.725899087239)*x_ref[0]+(0.634113963431)*x_ref[1]**o+(0.235694393453)*x_ref[1]+(-0.15600720713)*x_ref[2]**o+(0.954814976539)*x_ref[2]
        ref[1,4,2,1]=(0.120206996339)*x_ref[0]**o+(-0.86151411823)*x_ref[0]+(-0.137093650097)*x_ref[1]**o+(0.633402978246)*x_ref[1]+(-0.42515364728)*x_ref[2]**o+(-0.987634210617)*x_ref[2]
        ref[2,0,0,0]=(0.539463971222)*x_ref[0]**o+(-0.03435650579)*x_ref[0]+(-0.884395295544)*x_ref[1]**o+(0.215454879022)*x_ref[1]+(0.842377328246)*x_ref[2]**o+(0.370920136871)*x_ref[2]
        ref[2,0,0,1]=(-0.411505463039)*x_ref[0]**o+(0.446494039439)*x_ref[0]+(0.0313771969556)*x_ref[1]**o+(0.602893223061)*x_ref[1]+(-0.30383490021)*x_ref[2]**o+(-0.788028756758)*x_ref[2]
        ref[2,0,1,0]=(0.489009615999)*x_ref[0]**o+(-0.95508248016)*x_ref[0]+(-0.21489475022)*x_ref[1]**o+(-0.857263478431)*x_ref[1]+(-0.123703409045)*x_ref[2]**o+(0.663389562431)*x_ref[2]
        ref[2,0,1,1]=(-0.0224763310929)*x_ref[0]**o+(-0.315886249237)*x_ref[0]+(-0.318201849695)*x_ref[1]**o+(0.423732682462)*x_ref[1]+(0.396065636381)*x_ref[2]**o+(0.0816717878865)*x_ref[2]
        ref[2,0,2,0]=(-0.191432618224)*x_ref[0]**o+(0.688389390795)*x_ref[0]+(-0.382948112159)*x_ref[1]**o+(0.562905173818)*x_ref[1]+(-0.118396714029)*x_ref[2]**o+(0.234478642577)*x_ref[2]
        ref[2,0,2,1]=(-0.348008173171)*x_ref[0]**o+(0.191994166636)*x_ref[0]+(0.450556057147)*x_ref[1]**o+(-0.710821693343)*x_ref[1]+(-0.245119333857)*x_ref[2]**o+(0.262869525239)*x_ref[2]
        ref[2,1,0,0]=(-0.160444506203)*x_ref[0]**o+(-0.816142837432)*x_ref[0]+(-0.619722480492)*x_ref[1]**o+(-0.989077039675)*x_ref[1]+(-0.0282349103977)*x_ref[2]**o+(-0.629849819691)*x_ref[2]
        ref[2,1,0,1]=(0.379513765297)*x_ref[0]**o+(0.934652105586)*x_ref[0]+(-0.632636427868)*x_ref[1]**o+(-0.0492550496553)*x_ref[1]+(0.137733572307)*x_ref[2]**o+(0.779782637942)*x_ref[2]
        ref[2,1,1,0]=(0.469499320417)*x_ref[0]**o+(0.570880924125)*x_ref[0]+(-0.0329833251419)*x_ref[1]**o+(-0.0388751034604)*x_ref[1]+(-0.878520094389)*x_ref[2]**o+(0.992656649048)*x_ref[2]
        ref[2,1,1,1]=(0.592005181995)*x_ref[0]**o+(-0.553980493526)*x_ref[0]+(0.359837665019)*x_ref[1]**o+(0.811107941179)*x_ref[1]+(-0.441929650912)*x_ref[2]**o+(-0.768618378636)*x_ref[2]
        ref[2,1,2,0]=(0.741376258816)*x_ref[0]**o+(-0.238909756717)*x_ref[0]+(0.204631082449)*x_ref[1]**o+(0.970501662949)*x_ref[1]+(-0.615638092239)*x_ref[2]**o+(-0.575948750844)*x_ref[2]
        ref[2,1,2,1]=(0.774050933112)*x_ref[0]**o+(-0.872234076721)*x_ref[0]+(-0.0936189240061)*x_ref[1]**o+(-0.773608756961)*x_ref[1]+(-0.119171907808)*x_ref[2]**o+(-0.256124796845)*x_ref[2]
        ref[2,2,0,0]=(0.459481754425)*x_ref[0]**o+(0.478364929493)*x_ref[0]+(0.460199733287)*x_ref[1]**o+(0.66880294659)*x_ref[1]+(-0.859525058389)*x_ref[2]**o+(0.77960284414)*x_ref[2]
        ref[2,2,0,1]=(0.690557835207)*x_ref[0]**o+(-0.812305038592)*x_ref[0]+(0.837924124791)*x_ref[1]**o+(-0.785979206364)*x_ref[1]+(-0.295553949819)*x_ref[2]**o+(0.575096295897)*x_ref[2]
        ref[2,2,1,0]=(0.443193683205)*x_ref[0]**o+(0.0355033222253)*x_ref[0]+(0.497641002766)*x_ref[1]**o+(-0.636955142451)*x_ref[1]+(-0.265324357454)*x_ref[2]**o+(0.734148569078)*x_ref[2]
        ref[2,2,1,1]=(-0.088707242599)*x_ref[0]**o+(0.513403518669)*x_ref[0]+(0.337741737651)*x_ref[1]**o+(-0.174843924478)*x_ref[1]+(0.687291255986)*x_ref[2]**o+(0.473850502258)*x_ref[2]
        ref[2,2,2,0]=(0.5305624642)*x_ref[0]**o+(-0.972997835953)*x_ref[0]+(0.00197107287888)*x_ref[1]**o+(-0.75490872634)*x_ref[1]+(-0.0569624569435)*x_ref[2]**o+(-0.0780532868631)*x_ref[2]
        ref[2,2,2,1]=(-0.744584471866)*x_ref[0]**o+(0.915494909272)*x_ref[0]+(0.239235050038)*x_ref[1]**o+(-0.192466753187)*x_ref[1]+(0.933556550824)*x_ref[2]**o+(-0.786420251487)*x_ref[2]
        ref[2,3,0,0]=(0.472114530627)*x_ref[0]**o+(0.0693211807557)*x_ref[0]+(-0.274273425713)*x_ref[1]**o+(0.136242554894)*x_ref[1]+(-0.329240675785)*x_ref[2]**o+(0.0212539247343)*x_ref[2]
        ref[2,3,0,1]=(-0.780426108099)*x_ref[0]**o+(-0.632382551025)*x_ref[0]+(0.554926949239)*x_ref[1]**o+(0.264204009059)*x_ref[1]+(-0.868007525824)*x_ref[2]**o+(-0.937631120201)*x_ref[2]
        ref[2,3,1,0]=(-0.826692080947)*x_ref[0]**o+(0.915241858538)*x_ref[0]+(-0.49986160859)*x_ref[1]**o+(0.921686464658)*x_ref[1]+(0.71996666304)*x_ref[2]**o+(0.297609501096)*x_ref[2]
        ref[2,3,1,1]=(0.991896994614)*x_ref[0]**o+(-0.543310008103)*x_ref[0]+(0.0791899631111)*x_ref[1]**o+(-0.914088200145)*x_ref[1]+(-0.496337624951)*x_ref[2]**o+(0.708950898614)*x_ref[2]
        ref[2,3,2,0]=(-0.235116966432)*x_ref[0]**o+(-0.492848730207)*x_ref[0]+(0.428728646789)*x_ref[1]**o+(-0.410747634691)*x_ref[1]+(0.705059815458)*x_ref[2]**o+(-0.185604633831)*x_ref[2]
        ref[2,3,2,1]=(0.838829037117)*x_ref[0]**o+(0.670561147563)*x_ref[0]+(0.621407965075)*x_ref[1]**o+(-0.10624946305)*x_ref[1]+(0.893703402687)*x_ref[2]**o+(0.985504995821)*x_ref[2]
        ref[2,4,0,0]=(0.635072504682)*x_ref[0]**o+(-0.681962427555)*x_ref[0]+(-0.00774518011081)*x_ref[1]**o+(-0.264996228668)*x_ref[1]+(0.690668403198)*x_ref[2]**o+(0.883573761062)*x_ref[2]
        ref[2,4,0,1]=(0.545632268651)*x_ref[0]**o+(0.439668519394)*x_ref[0]+(-0.482479808491)*x_ref[1]**o+(0.752028962391)*x_ref[1]+(0.0596773623939)*x_ref[2]**o+(-0.369987181127)*x_ref[2]
        ref[2,4,1,0]=(0.659005322796)*x_ref[0]**o+(0.503940833428)*x_ref[0]+(0.643633321586)*x_ref[1]**o+(0.927082323252)*x_ref[1]+(0.354072146184)*x_ref[2]**o+(-0.638737560608)*x_ref[2]
        ref[2,4,1,1]=(-0.443548860913)*x_ref[0]**o+(0.707607048458)*x_ref[0]+(-0.955225746402)*x_ref[1]**o+(0.515995563585)*x_ref[1]+(-0.288406637165)*x_ref[2]**o+(-0.760864937318)*x_ref[2]
        ref[2,4,2,0]=(0.898146580355)*x_ref[0]**o+(-0.590650787106)*x_ref[0]+(0.036095789753)*x_ref[1]**o+(0.326040260581)*x_ref[1]+(0.680069361306)*x_ref[2]**o+(0.333839099341)*x_ref[2]
        ref[2,4,2,1]=(-0.151364260271)*x_ref[0]**o+(-0.675198183832)*x_ref[0]+(-0.183626298978)*x_ref[1]**o+(-0.182565217146)*x_ref[1]+(0.0345632129814)*x_ref[2]**o+(-0.581534100763)*x_ref[2]
        ref[3,0,0,0]=(-0.942351703246)*x_ref[0]**o+(-0.699567649453)*x_ref[0]+(-0.832068665702)*x_ref[1]**o+(0.163439416504)*x_ref[1]+(0.742500339519)*x_ref[2]**o+(-0.492882887489)*x_ref[2]
        ref[3,0,0,1]=(0.990078296484)*x_ref[0]**o+(-0.999837336587)*x_ref[0]+(0.197589456459)*x_ref[1]**o+(0.431403059019)*x_ref[1]+(0.859864015871)*x_ref[2]**o+(-0.267474829282)*x_ref[2]
        ref[3,0,1,0]=(-0.776773660723)*x_ref[0]**o+(-0.189048370486)*x_ref[0]+(0.926953856802)*x_ref[1]**o+(-0.72314744652)*x_ref[1]+(0.330064363809)*x_ref[2]**o+(0.188533334359)*x_ref[2]
        ref[3,0,1,1]=(-0.854830058762)*x_ref[0]**o+(0.832073883276)*x_ref[0]+(-0.425691165124)*x_ref[1]**o+(0.162351500316)*x_ref[1]+(0.370604864512)*x_ref[2]**o+(-0.581907071484)*x_ref[2]
        ref[3,0,2,0]=(-0.316716909548)*x_ref[0]**o+(0.335417284383)*x_ref[0]+(-0.783061160352)*x_ref[1]**o+(-0.253972519239)*x_ref[1]+(-0.826432933721)*x_ref[2]**o+(0.267368378363)*x_ref[2]
        ref[3,0,2,1]=(-0.765899705462)*x_ref[0]**o+(-0.0993875861476)*x_ref[0]+(-0.211785004352)*x_ref[1]**o+(-0.635330151497)*x_ref[1]+(0.414575412599)*x_ref[2]**o+(0.586882274623)*x_ref[2]
        ref[3,1,0,0]=(0.987776377237)*x_ref[0]**o+(0.796660669191)*x_ref[0]+(0.580724891862)*x_ref[1]**o+(-0.24149253995)*x_ref[1]+(-0.633421016909)*x_ref[2]**o+(-0.220110913711)*x_ref[2]
        ref[3,1,0,1]=(-0.182240918081)*x_ref[0]**o+(-0.290526050572)*x_ref[0]+(0.00283230959976)*x_ref[1]**o+(-0.933282166805)*x_ref[1]+(-0.857893235883)*x_ref[2]**o+(0.551414074127)*x_ref[2]
        ref[3,1,1,0]=(-0.854435997816)*x_ref[0]**o+(0.706455144263)*x_ref[0]+(0.406237965254)*x_ref[1]**o+(0.192071142415)*x_ref[1]+(0.357966796834)*x_ref[2]**o+(0.887011087987)*x_ref[2]
        ref[3,1,1,1]=(0.832013396996)*x_ref[0]**o+(0.696133072321)*x_ref[0]+(0.682726968712)*x_ref[1]**o+(0.613316240514)*x_ref[1]+(0.12600339887)*x_ref[2]**o+(-0.620088231382)*x_ref[2]
        ref[3,1,2,0]=(-0.178681615221)*x_ref[0]**o+(-0.952835986168)*x_ref[0]+(0.299692693746)*x_ref[1]**o+(0.892451908416)*x_ref[1]+(0.202403273634)*x_ref[2]**o+(0.677212897943)*x_ref[2]
        ref[3,1,2,1]=(0.0592594053342)*x_ref[0]**o+(0.56391153959)*x_ref[0]+(-0.874866627667)*x_ref[1]**o+(0.432102900769)*x_ref[1]+(-0.49073126285)*x_ref[2]**o+(-0.010495860028)*x_ref[2]
        ref[3,2,0,0]=(0.548006445814)*x_ref[0]**o+(0.667313392202)*x_ref[0]+(0.397805355499)*x_ref[1]**o+(-0.788100267808)*x_ref[1]+(0.869234055258)*x_ref[2]**o+(-0.830241520401)*x_ref[2]
        ref[3,2,0,1]=(0.72586139216)*x_ref[0]**o+(-0.931878957878)*x_ref[0]+(-0.94119555858)*x_ref[1]**o+(-0.133544806169)*x_ref[1]+(0.754263323382)*x_ref[2]**o+(-0.446604740157)*x_ref[2]
        ref[3,2,1,0]=(-0.28135830707)*x_ref[0]**o+(-0.478744058518)*x_ref[0]+(0.88219033731)*x_ref[1]**o+(0.460587286945)*x_ref[1]+(-0.849692146096)*x_ref[2]**o+(0.0927701345107)*x_ref[2]
        ref[3,2,1,1]=(-0.681859641611)*x_ref[0]**o+(-0.670164293088)*x_ref[0]+(0.796366374435)*x_ref[1]**o+(0.837230317054)*x_ref[1]+(-0.413283915904)*x_ref[2]**o+(0.779624227187)*x_ref[2]
        ref[3,2,2,0]=(-0.803242771255)*x_ref[0]**o+(-0.0687616271776)*x_ref[0]+(-0.948856027999)*x_ref[1]**o+(0.762568790931)*x_ref[1]+(0.937588292604)*x_ref[2]**o+(-0.0224232611283)*x_ref[2]
        ref[3,2,2,1]=(-0.215752944937)*x_ref[0]**o+(-0.807114296075)*x_ref[0]+(0.970625375964)*x_ref[1]**o+(0.68578124692)*x_ref[1]+(-0.165007047419)*x_ref[2]**o+(0.942718439981)*x_ref[2]
        ref[3,3,0,0]=(0.152001104716)*x_ref[0]**o+(-0.657242342357)*x_ref[0]+(0.559608661705)*x_ref[1]**o+(0.0764403273251)*x_ref[1]+(0.633414754695)*x_ref[2]**o+(0.430777868344)*x_ref[2]
        ref[3,3,0,1]=(-0.898109497759)*x_ref[0]**o+(0.736574249406)*x_ref[0]+(-0.569584184574)*x_ref[1]**o+(-0.145501395506)*x_ref[1]+(-0.412424116475)*x_ref[2]**o+(-0.416350923096)*x_ref[2]
        ref[3,3,1,0]=(0.602022795614)*x_ref[0]**o+(-0.848525569825)*x_ref[0]+(-0.713426287505)*x_ref[1]**o+(-0.829278452687)*x_ref[1]+(-0.491257659316)*x_ref[2]**o+(0.90219908999)*x_ref[2]
        ref[3,3,1,1]=(-0.769110216066)*x_ref[0]**o+(-0.789464037338)*x_ref[0]+(-0.363686172584)*x_ref[1]**o+(-0.86190787068)*x_ref[1]+(0.275438119232)*x_ref[2]**o+(0.285209834248)*x_ref[2]
        ref[3,3,2,0]=(0.351662405853)*x_ref[0]**o+(0.8403302307)*x_ref[0]+(-0.449869332519)*x_ref[1]**o+(0.936043614838)*x_ref[1]+(-0.656908777335)*x_ref[2]**o+(-0.231923872531)*x_ref[2]
        ref[3,3,2,1]=(0.873848945653)*x_ref[0]**o+(0.551429883959)*x_ref[0]+(-0.577942548701)*x_ref[1]**o+(-0.254317163049)*x_ref[1]+(0.64878089878)*x_ref[2]**o+(0.7135339515)*x_ref[2]
        ref[3,4,0,0]=(0.891435610494)*x_ref[0]**o+(0.598893704498)*x_ref[0]+(0.24682094756)*x_ref[1]**o+(0.0895612710338)*x_ref[1]+(-0.513686244953)*x_ref[2]**o+(0.307205505502)*x_ref[2]
        ref[3,4,0,1]=(-0.577752944196)*x_ref[0]**o+(-0.458383751623)*x_ref[0]+(-0.913956988143)*x_ref[1]**o+(0.470857602235)*x_ref[1]+(0.631726850665)*x_ref[2]**o+(-0.112962289904)*x_ref[2]
        ref[3,4,1,0]=(-0.65648177489)*x_ref[0]**o+(-0.843530972747)*x_ref[0]+(-0.0332153723809)*x_ref[1]**o+(0.230484263355)*x_ref[1]+(-0.865687508592)*x_ref[2]**o+(0.122751735082)*x_ref[2]
        ref[3,4,1,1]=(-0.353502640734)*x_ref[0]**o+(-0.816020900282)*x_ref[0]+(0.943688796868)*x_ref[1]**o+(-0.526707881958)*x_ref[1]+(-0.790258298117)*x_ref[2]**o+(-0.748706809042)*x_ref[2]
        ref[3,4,2,0]=(-0.596345912175)*x_ref[0]**o+(0.872771121837)*x_ref[0]+(0.100107395575)*x_ref[1]**o+(0.950482249217)*x_ref[1]+(0.0487634582783)*x_ref[2]**o+(-0.0285218097887)*x_ref[2]
        ref[3,4,2,1]=(-0.827443553933)*x_ref[0]**o+(-0.922564800149)*x_ref[0]+(-0.810354385287)*x_ref[1]**o+(0.630079217293)*x_ref[1]+(-0.316280559723)*x_ref[2]**o+(0.977830435871)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.142445811229)*x[0]**o+(-0.170111870439)*x[0]+(-0.571150729608)*x[1]**o+(-0.799596368319)*x[1]
        ref=(0.142445811229)*x_ref[0]**o+(-0.170111870439)*x_ref[0]+(-0.571150729608)*x_ref[1]**o+(-0.799596368319)*x_ref[1]
      else:
        arg=(0.131269935442)*x[0]**o+(0.67764849927)*x[0]+(0.199946665)*x[1]**o+(0.775678673404)*x[1]+(-0.818505707274)*x[2]**o+(-0.141160005914)*x[2]
        ref=(0.131269935442)*x_ref[0]**o+(0.67764849927)*x_ref[0]+(0.199946665)*x_ref[1]**o+(0.775678673404)*x_ref[1]+(-0.818505707274)*x_ref[2]**o+(-0.141160005914)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.271030346059)*x[0]**o+(-0.604054986035)*x[0]+(-0.856492962015)*x[1]**o+(0.229582187423)*x[1]
        arg[1]=(-0.152539485929)*x[0]**o+(0.936634973224)*x[0]+(0.805016132937)*x[1]**o+(-0.6832854412)*x[1]
        ref[0]=(-0.271030346059)*x_ref[0]**o+(-0.604054986035)*x_ref[0]+(-0.856492962015)*x_ref[1]**o+(0.229582187423)*x_ref[1]
        ref[1]=(-0.152539485929)*x_ref[0]**o+(0.936634973224)*x_ref[0]+(0.805016132937)*x_ref[1]**o+(-0.6832854412)*x_ref[1]
      else:
        arg[0]=(-0.484580515247)*x[0]**o+(-0.134539997002)*x[0]+(0.589742989514)*x[1]**o+(-0.950803827747)*x[1]+(-0.852976687869)*x[2]**o+(0.0999526415345)*x[2]
        arg[1]=(-0.863620642149)*x[0]**o+(-0.16255544441)*x[0]+(-0.4963232161)*x[1]**o+(0.180988811217)*x[1]+(-0.787186716126)*x[2]**o+(-0.359825851725)*x[2]
        ref[0]=(-0.484580515247)*x_ref[0]**o+(-0.134539997002)*x_ref[0]+(0.589742989514)*x_ref[1]**o+(-0.950803827747)*x_ref[1]+(-0.852976687869)*x_ref[2]**o+(0.0999526415345)*x_ref[2]
        ref[1]=(-0.863620642149)*x_ref[0]**o+(-0.16255544441)*x_ref[0]+(-0.4963232161)*x_ref[1]**o+(0.180988811217)*x_ref[1]+(-0.787186716126)*x_ref[2]**o+(-0.359825851725)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.88064835903)*x[0]**o+(-0.140271048585)*x[0]+(-0.363467917345)*x[1]**o+(0.382530733181)*x[1]
        arg[0,1]=(0.835096280471)*x[0]**o+(0.132077921375)*x[0]+(-0.814392115734)*x[1]**o+(-0.379227781873)*x[1]
        arg[0,2]=(0.257391174596)*x[0]**o+(-0.135017751364)*x[0]+(-0.462760793484)*x[1]**o+(-0.332510860616)*x[1]
        arg[0,3]=(0.303541666957)*x[0]**o+(0.729054207976)*x[0]+(-0.0474210800089)*x[1]**o+(0.213859363809)*x[1]
        arg[0,4]=(0.564466740909)*x[0]**o+(0.684761789227)*x[0]+(0.432934636771)*x[1]**o+(0.261269614075)*x[1]
        arg[1,0]=(0.83080596334)*x[0]**o+(0.968795153198)*x[0]+(0.835115581928)*x[1]**o+(-0.134688885728)*x[1]
        arg[1,1]=(-0.876235093042)*x[0]**o+(-0.43923440557)*x[0]+(-0.641186633503)*x[1]**o+(-0.987129244994)*x[1]
        arg[1,2]=(-0.994007922096)*x[0]**o+(0.534914510762)*x[0]+(0.754420969331)*x[1]**o+(0.241533891038)*x[1]
        arg[1,3]=(-0.196027247194)*x[0]**o+(0.962231147494)*x[0]+(-0.819146467532)*x[1]**o+(0.966024528731)*x[1]
        arg[1,4]=(-0.577014723723)*x[0]**o+(0.350310620855)*x[0]+(0.16266475329)*x[1]**o+(-0.926739533525)*x[1]
        arg[2,0]=(-0.262964780787)*x[0]**o+(-0.824445469277)*x[0]+(-0.982497605841)*x[1]**o+(0.900080022932)*x[1]
        arg[2,1]=(0.54409272334)*x[0]**o+(0.789421020125)*x[0]+(0.672239490644)*x[1]**o+(-0.0306885937363)*x[1]
        arg[2,2]=(-0.130272075526)*x[0]**o+(0.51213564573)*x[0]+(-0.819817699321)*x[1]**o+(-0.259068167835)*x[1]
        arg[2,3]=(-0.61526267092)*x[0]**o+(-0.504362726269)*x[0]+(0.362429515671)*x[1]**o+(-0.0921852465615)*x[1]
        arg[2,4]=(-0.344772989987)*x[0]**o+(0.260747743828)*x[0]+(0.474457474131)*x[1]**o+(0.436008642452)*x[1]
        arg[3,0]=(-0.72740499107)*x[0]**o+(-0.297524430077)*x[0]+(-0.809254720524)*x[1]**o+(-0.230695822532)*x[1]
        arg[3,1]=(-0.0817042383885)*x[0]**o+(-0.324890823023)*x[0]+(0.923635342281)*x[1]**o+(-0.336825317481)*x[1]
        arg[3,2]=(-0.872954045409)*x[0]**o+(0.242165239629)*x[0]+(0.919264824591)*x[1]**o+(0.424247774945)*x[1]
        arg[3,3]=(0.146344350418)*x[0]**o+(-0.705285807962)*x[0]+(-0.0735095471635)*x[1]**o+(-0.464180824267)*x[1]
        arg[3,4]=(0.624085113245)*x[0]**o+(0.719452515121)*x[0]+(0.28331108038)*x[1]**o+(-0.897435702442)*x[1]
        ref[0,0]=(-0.88064835903)*x_ref[0]**o+(-0.140271048585)*x_ref[0]+(-0.363467917345)*x_ref[1]**o+(0.382530733181)*x_ref[1]
        ref[0,1]=(0.835096280471)*x_ref[0]**o+(0.132077921375)*x_ref[0]+(-0.814392115734)*x_ref[1]**o+(-0.379227781873)*x_ref[1]
        ref[0,2]=(0.257391174596)*x_ref[0]**o+(-0.135017751364)*x_ref[0]+(-0.462760793484)*x_ref[1]**o+(-0.332510860616)*x_ref[1]
        ref[0,3]=(0.303541666957)*x_ref[0]**o+(0.729054207976)*x_ref[0]+(-0.0474210800089)*x_ref[1]**o+(0.213859363809)*x_ref[1]
        ref[0,4]=(0.564466740909)*x_ref[0]**o+(0.684761789227)*x_ref[0]+(0.432934636771)*x_ref[1]**o+(0.261269614075)*x_ref[1]
        ref[1,0]=(0.83080596334)*x_ref[0]**o+(0.968795153198)*x_ref[0]+(0.835115581928)*x_ref[1]**o+(-0.134688885728)*x_ref[1]
        ref[1,1]=(-0.876235093042)*x_ref[0]**o+(-0.43923440557)*x_ref[0]+(-0.641186633503)*x_ref[1]**o+(-0.987129244994)*x_ref[1]
        ref[1,2]=(-0.994007922096)*x_ref[0]**o+(0.534914510762)*x_ref[0]+(0.754420969331)*x_ref[1]**o+(0.241533891038)*x_ref[1]
        ref[1,3]=(-0.196027247194)*x_ref[0]**o+(0.962231147494)*x_ref[0]+(-0.819146467532)*x_ref[1]**o+(0.966024528731)*x_ref[1]
        ref[1,4]=(-0.577014723723)*x_ref[0]**o+(0.350310620855)*x_ref[0]+(0.16266475329)*x_ref[1]**o+(-0.926739533525)*x_ref[1]
        ref[2,0]=(-0.262964780787)*x_ref[0]**o+(-0.824445469277)*x_ref[0]+(-0.982497605841)*x_ref[1]**o+(0.900080022932)*x_ref[1]
        ref[2,1]=(0.54409272334)*x_ref[0]**o+(0.789421020125)*x_ref[0]+(0.672239490644)*x_ref[1]**o+(-0.0306885937363)*x_ref[1]
        ref[2,2]=(-0.130272075526)*x_ref[0]**o+(0.51213564573)*x_ref[0]+(-0.819817699321)*x_ref[1]**o+(-0.259068167835)*x_ref[1]
        ref[2,3]=(-0.61526267092)*x_ref[0]**o+(-0.504362726269)*x_ref[0]+(0.362429515671)*x_ref[1]**o+(-0.0921852465615)*x_ref[1]
        ref[2,4]=(-0.344772989987)*x_ref[0]**o+(0.260747743828)*x_ref[0]+(0.474457474131)*x_ref[1]**o+(0.436008642452)*x_ref[1]
        ref[3,0]=(-0.72740499107)*x_ref[0]**o+(-0.297524430077)*x_ref[0]+(-0.809254720524)*x_ref[1]**o+(-0.230695822532)*x_ref[1]
        ref[3,1]=(-0.0817042383885)*x_ref[0]**o+(-0.324890823023)*x_ref[0]+(0.923635342281)*x_ref[1]**o+(-0.336825317481)*x_ref[1]
        ref[3,2]=(-0.872954045409)*x_ref[0]**o+(0.242165239629)*x_ref[0]+(0.919264824591)*x_ref[1]**o+(0.424247774945)*x_ref[1]
        ref[3,3]=(0.146344350418)*x_ref[0]**o+(-0.705285807962)*x_ref[0]+(-0.0735095471635)*x_ref[1]**o+(-0.464180824267)*x_ref[1]
        ref[3,4]=(0.624085113245)*x_ref[0]**o+(0.719452515121)*x_ref[0]+(0.28331108038)*x_ref[1]**o+(-0.897435702442)*x_ref[1]
      else:
        arg[0,0]=(0.47332206873)*x[0]**o+(0.20564453167)*x[0]+(-0.175485469473)*x[1]**o+(-0.25166656853)*x[1]+(-0.242693318919)*x[2]**o+(-0.773417228526)*x[2]
        arg[0,1]=(0.788563488587)*x[0]**o+(0.598162739778)*x[0]+(-0.736463777722)*x[1]**o+(-0.161555683484)*x[1]+(0.480782882175)*x[2]**o+(-0.500552084345)*x[2]
        arg[0,2]=(-0.207893050318)*x[0]**o+(0.206221241577)*x[0]+(0.306329036151)*x[1]**o+(0.585237890094)*x[1]+(0.0276469766188)*x[2]**o+(0.0502503670529)*x[2]
        arg[0,3]=(0.914950534537)*x[0]**o+(0.401315333176)*x[0]+(0.245224334898)*x[1]**o+(0.526049350134)*x[1]+(-0.735617023846)*x[2]**o+(-0.674692140531)*x[2]
        arg[0,4]=(-0.359566330106)*x[0]**o+(-0.279627730979)*x[0]+(-0.08398829656)*x[1]**o+(-0.789678621983)*x[1]+(-0.357374658443)*x[2]**o+(0.795724504883)*x[2]
        arg[1,0]=(-0.0715278193161)*x[0]**o+(-0.562120168502)*x[0]+(0.428614380101)*x[1]**o+(0.115408373591)*x[1]+(-0.154674397872)*x[2]**o+(0.233468972682)*x[2]
        arg[1,1]=(0.707445640039)*x[0]**o+(-0.719649334705)*x[0]+(-0.349743923514)*x[1]**o+(0.369262203159)*x[1]+(0.712786695739)*x[2]**o+(-0.962450154391)*x[2]
        arg[1,2]=(-0.306983094408)*x[0]**o+(-0.933322822964)*x[0]+(-0.525515899443)*x[1]**o+(0.642320930433)*x[1]+(-0.950421802297)*x[2]**o+(-0.392142385683)*x[2]
        arg[1,3]=(-0.765158324235)*x[0]**o+(-0.362010493356)*x[0]+(-0.918673930909)*x[1]**o+(0.72772784184)*x[1]+(-0.392848438408)*x[2]**o+(0.566119231816)*x[2]
        arg[1,4]=(0.429984662315)*x[0]**o+(-0.165476415697)*x[0]+(0.778539911928)*x[1]**o+(0.411518911472)*x[1]+(0.418323432603)*x[2]**o+(-0.846495913126)*x[2]
        arg[2,0]=(-0.836556750858)*x[0]**o+(0.976027316021)*x[0]+(-0.556499096672)*x[1]**o+(0.320038395915)*x[1]+(0.377663884516)*x[2]**o+(0.491202727923)*x[2]
        arg[2,1]=(-0.329597826951)*x[0]**o+(-0.555819984044)*x[0]+(0.110029358927)*x[1]**o+(-0.737070187836)*x[1]+(-0.587142473737)*x[2]**o+(-0.0388834177112)*x[2]
        arg[2,2]=(0.442251030646)*x[0]**o+(0.285010574287)*x[0]+(0.603267537331)*x[1]**o+(-0.0776884302384)*x[1]+(-0.123541484708)*x[2]**o+(0.44717125529)*x[2]
        arg[2,3]=(0.474458930313)*x[0]**o+(0.415947401939)*x[0]+(0.381539850696)*x[1]**o+(-0.307590909422)*x[1]+(0.488348451586)*x[2]**o+(0.336883810306)*x[2]
        arg[2,4]=(0.804466542611)*x[0]**o+(0.891337158669)*x[0]+(-0.599119487307)*x[1]**o+(-0.92047732727)*x[1]+(-0.251489480657)*x[2]**o+(-0.266483991526)*x[2]
        arg[3,0]=(-0.733633520827)*x[0]**o+(0.494805537624)*x[0]+(-0.117702558274)*x[1]**o+(-0.000461693303442)*x[1]+(-0.102033981202)*x[2]**o+(-0.772865985272)*x[2]
        arg[3,1]=(-0.636901276465)*x[0]**o+(-0.436330072041)*x[0]+(0.403048850222)*x[1]**o+(-0.94737100009)*x[1]+(0.781059874101)*x[2]**o+(-0.952781516037)*x[2]
        arg[3,2]=(-0.489372336164)*x[0]**o+(0.221211043152)*x[0]+(0.793975373906)*x[1]**o+(0.424349965288)*x[1]+(0.453332885662)*x[2]**o+(-0.513541753005)*x[2]
        arg[3,3]=(-0.700899677836)*x[0]**o+(-0.984271896306)*x[0]+(-0.663909486898)*x[1]**o+(0.461302174316)*x[1]+(-0.906457707724)*x[2]**o+(-0.0440003164038)*x[2]
        arg[3,4]=(0.838326295366)*x[0]**o+(-0.483407593962)*x[0]+(-0.122458752397)*x[1]**o+(0.607337951095)*x[1]+(0.750481384434)*x[2]**o+(-0.677155633204)*x[2]
        ref[0,0]=(0.47332206873)*x_ref[0]**o+(0.20564453167)*x_ref[0]+(-0.175485469473)*x_ref[1]**o+(-0.25166656853)*x_ref[1]+(-0.242693318919)*x_ref[2]**o+(-0.773417228526)*x_ref[2]
        ref[0,1]=(0.788563488587)*x_ref[0]**o+(0.598162739778)*x_ref[0]+(-0.736463777722)*x_ref[1]**o+(-0.161555683484)*x_ref[1]+(0.480782882175)*x_ref[2]**o+(-0.500552084345)*x_ref[2]
        ref[0,2]=(-0.207893050318)*x_ref[0]**o+(0.206221241577)*x_ref[0]+(0.306329036151)*x_ref[1]**o+(0.585237890094)*x_ref[1]+(0.0276469766188)*x_ref[2]**o+(0.0502503670529)*x_ref[2]
        ref[0,3]=(0.914950534537)*x_ref[0]**o+(0.401315333176)*x_ref[0]+(0.245224334898)*x_ref[1]**o+(0.526049350134)*x_ref[1]+(-0.735617023846)*x_ref[2]**o+(-0.674692140531)*x_ref[2]
        ref[0,4]=(-0.359566330106)*x_ref[0]**o+(-0.279627730979)*x_ref[0]+(-0.08398829656)*x_ref[1]**o+(-0.789678621983)*x_ref[1]+(-0.357374658443)*x_ref[2]**o+(0.795724504883)*x_ref[2]
        ref[1,0]=(-0.0715278193161)*x_ref[0]**o+(-0.562120168502)*x_ref[0]+(0.428614380101)*x_ref[1]**o+(0.115408373591)*x_ref[1]+(-0.154674397872)*x_ref[2]**o+(0.233468972682)*x_ref[2]
        ref[1,1]=(0.707445640039)*x_ref[0]**o+(-0.719649334705)*x_ref[0]+(-0.349743923514)*x_ref[1]**o+(0.369262203159)*x_ref[1]+(0.712786695739)*x_ref[2]**o+(-0.962450154391)*x_ref[2]
        ref[1,2]=(-0.306983094408)*x_ref[0]**o+(-0.933322822964)*x_ref[0]+(-0.525515899443)*x_ref[1]**o+(0.642320930433)*x_ref[1]+(-0.950421802297)*x_ref[2]**o+(-0.392142385683)*x_ref[2]
        ref[1,3]=(-0.765158324235)*x_ref[0]**o+(-0.362010493356)*x_ref[0]+(-0.918673930909)*x_ref[1]**o+(0.72772784184)*x_ref[1]+(-0.392848438408)*x_ref[2]**o+(0.566119231816)*x_ref[2]
        ref[1,4]=(0.429984662315)*x_ref[0]**o+(-0.165476415697)*x_ref[0]+(0.778539911928)*x_ref[1]**o+(0.411518911472)*x_ref[1]+(0.418323432603)*x_ref[2]**o+(-0.846495913126)*x_ref[2]
        ref[2,0]=(-0.836556750858)*x_ref[0]**o+(0.976027316021)*x_ref[0]+(-0.556499096672)*x_ref[1]**o+(0.320038395915)*x_ref[1]+(0.377663884516)*x_ref[2]**o+(0.491202727923)*x_ref[2]
        ref[2,1]=(-0.329597826951)*x_ref[0]**o+(-0.555819984044)*x_ref[0]+(0.110029358927)*x_ref[1]**o+(-0.737070187836)*x_ref[1]+(-0.587142473737)*x_ref[2]**o+(-0.0388834177112)*x_ref[2]
        ref[2,2]=(0.442251030646)*x_ref[0]**o+(0.285010574287)*x_ref[0]+(0.603267537331)*x_ref[1]**o+(-0.0776884302384)*x_ref[1]+(-0.123541484708)*x_ref[2]**o+(0.44717125529)*x_ref[2]
        ref[2,3]=(0.474458930313)*x_ref[0]**o+(0.415947401939)*x_ref[0]+(0.381539850696)*x_ref[1]**o+(-0.307590909422)*x_ref[1]+(0.488348451586)*x_ref[2]**o+(0.336883810306)*x_ref[2]
        ref[2,4]=(0.804466542611)*x_ref[0]**o+(0.891337158669)*x_ref[0]+(-0.599119487307)*x_ref[1]**o+(-0.92047732727)*x_ref[1]+(-0.251489480657)*x_ref[2]**o+(-0.266483991526)*x_ref[2]
        ref[3,0]=(-0.733633520827)*x_ref[0]**o+(0.494805537624)*x_ref[0]+(-0.117702558274)*x_ref[1]**o+(-0.000461693303442)*x_ref[1]+(-0.102033981202)*x_ref[2]**o+(-0.772865985272)*x_ref[2]
        ref[3,1]=(-0.636901276465)*x_ref[0]**o+(-0.436330072041)*x_ref[0]+(0.403048850222)*x_ref[1]**o+(-0.94737100009)*x_ref[1]+(0.781059874101)*x_ref[2]**o+(-0.952781516037)*x_ref[2]
        ref[3,2]=(-0.489372336164)*x_ref[0]**o+(0.221211043152)*x_ref[0]+(0.793975373906)*x_ref[1]**o+(0.424349965288)*x_ref[1]+(0.453332885662)*x_ref[2]**o+(-0.513541753005)*x_ref[2]
        ref[3,3]=(-0.700899677836)*x_ref[0]**o+(-0.984271896306)*x_ref[0]+(-0.663909486898)*x_ref[1]**o+(0.461302174316)*x_ref[1]+(-0.906457707724)*x_ref[2]**o+(-0.0440003164038)*x_ref[2]
        ref[3,4]=(0.838326295366)*x_ref[0]**o+(-0.483407593962)*x_ref[0]+(-0.122458752397)*x_ref[1]**o+(0.607337951095)*x_ref[1]+(0.750481384434)*x_ref[2]**o+(-0.677155633204)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.111473275114)*x[0]**o+(-0.552264448532)*x[0]+(0.67142534913)*x[1]**o+(-0.642305751076)*x[1]
        arg[0,0,1]=(-0.431075800974)*x[0]**o+(0.911382078454)*x[0]+(0.546541644612)*x[1]**o+(-0.67679652536)*x[1]
        arg[0,1,0]=(0.103056292292)*x[0]**o+(-0.640858804416)*x[0]+(0.159596964992)*x[1]**o+(-0.237782007316)*x[1]
        arg[0,1,1]=(0.975898170389)*x[0]**o+(0.255916326383)*x[0]+(-0.692770462322)*x[1]**o+(0.786355114846)*x[1]
        arg[1,0,0]=(-0.669378370759)*x[0]**o+(-0.0153788709146)*x[0]+(0.924383188557)*x[1]**o+(0.885037556829)*x[1]
        arg[1,0,1]=(-0.509891819881)*x[0]**o+(-0.377779514697)*x[0]+(-0.343176473036)*x[1]**o+(-0.39978416725)*x[1]
        arg[1,1,0]=(0.457959467665)*x[0]**o+(0.786511967341)*x[0]+(0.256561665146)*x[1]**o+(-0.261300099643)*x[1]
        arg[1,1,1]=(0.219754515189)*x[0]**o+(-0.131747762867)*x[0]+(0.736017818893)*x[1]**o+(-0.450963700533)*x[1]
        arg[2,0,0]=(-0.182946832789)*x[0]**o+(-0.103064731935)*x[0]+(-0.223584334222)*x[1]**o+(-0.578088201628)*x[1]
        arg[2,0,1]=(-0.830744157425)*x[0]**o+(0.590627167207)*x[0]+(-0.538048378049)*x[1]**o+(-0.0529793352741)*x[1]
        arg[2,1,0]=(-0.307908670388)*x[0]**o+(-0.929987734045)*x[0]+(-0.167190563092)*x[1]**o+(0.126466628227)*x[1]
        arg[2,1,1]=(-0.812045378916)*x[0]**o+(0.676973938072)*x[0]+(0.528243909204)*x[1]**o+(-0.594241387691)*x[1]
        arg[3,0,0]=(-0.0198606164114)*x[0]**o+(-0.00574708685554)*x[0]+(0.908946312745)*x[1]**o+(0.775054099788)*x[1]
        arg[3,0,1]=(-0.810078120518)*x[0]**o+(-0.960585888485)*x[0]+(-0.0747272427198)*x[1]**o+(0.164107641817)*x[1]
        arg[3,1,0]=(-0.535791917139)*x[0]**o+(0.0424226440074)*x[0]+(0.701479263902)*x[1]**o+(0.787745423088)*x[1]
        arg[3,1,1]=(-0.896532152793)*x[0]**o+(-0.0224566920251)*x[0]+(-0.554655626689)*x[1]**o+(0.41012867952)*x[1]
        arg[4,0,0]=(0.972433817496)*x[0]**o+(-0.440662536033)*x[0]+(-0.910456530943)*x[1]**o+(-0.333429535028)*x[1]
        arg[4,0,1]=(-0.415824313942)*x[0]**o+(0.00255557712922)*x[0]+(0.304897639195)*x[1]**o+(-0.866600848052)*x[1]
        arg[4,1,0]=(-0.866277739076)*x[0]**o+(0.282976169776)*x[0]+(0.477528278501)*x[1]**o+(-0.233506497695)*x[1]
        arg[4,1,1]=(-0.12729626381)*x[0]**o+(-0.723204796785)*x[0]+(0.0856495713442)*x[1]**o+(-0.214819437731)*x[1]
        arg[5,0,0]=(0.208133503644)*x[0]**o+(-0.955529202895)*x[0]+(0.0291569127831)*x[1]**o+(0.489913918305)*x[1]
        arg[5,0,1]=(0.96923667477)*x[0]**o+(0.543776525315)*x[0]+(0.38666450029)*x[1]**o+(-0.193493550272)*x[1]
        arg[5,1,0]=(-0.0912940016888)*x[0]**o+(-0.477561072681)*x[0]+(0.474816021613)*x[1]**o+(-0.252664229253)*x[1]
        arg[5,1,1]=(0.748337265159)*x[0]**o+(-0.706194454618)*x[0]+(-0.504466281507)*x[1]**o+(0.582500817679)*x[1]
        ref[0,0,0]=(0.111473275114)*x_ref[0]**o+(-0.552264448532)*x_ref[0]+(0.67142534913)*x_ref[1]**o+(-0.642305751076)*x_ref[1]
        ref[0,0,1]=(-0.431075800974)*x_ref[0]**o+(0.911382078454)*x_ref[0]+(0.546541644612)*x_ref[1]**o+(-0.67679652536)*x_ref[1]
        ref[0,1,0]=(0.103056292292)*x_ref[0]**o+(-0.640858804416)*x_ref[0]+(0.159596964992)*x_ref[1]**o+(-0.237782007316)*x_ref[1]
        ref[0,1,1]=(0.975898170389)*x_ref[0]**o+(0.255916326383)*x_ref[0]+(-0.692770462322)*x_ref[1]**o+(0.786355114846)*x_ref[1]
        ref[1,0,0]=(-0.669378370759)*x_ref[0]**o+(-0.0153788709146)*x_ref[0]+(0.924383188557)*x_ref[1]**o+(0.885037556829)*x_ref[1]
        ref[1,0,1]=(-0.509891819881)*x_ref[0]**o+(-0.377779514697)*x_ref[0]+(-0.343176473036)*x_ref[1]**o+(-0.39978416725)*x_ref[1]
        ref[1,1,0]=(0.457959467665)*x_ref[0]**o+(0.786511967341)*x_ref[0]+(0.256561665146)*x_ref[1]**o+(-0.261300099643)*x_ref[1]
        ref[1,1,1]=(0.219754515189)*x_ref[0]**o+(-0.131747762867)*x_ref[0]+(0.736017818893)*x_ref[1]**o+(-0.450963700533)*x_ref[1]
        ref[2,0,0]=(-0.182946832789)*x_ref[0]**o+(-0.103064731935)*x_ref[0]+(-0.223584334222)*x_ref[1]**o+(-0.578088201628)*x_ref[1]
        ref[2,0,1]=(-0.830744157425)*x_ref[0]**o+(0.590627167207)*x_ref[0]+(-0.538048378049)*x_ref[1]**o+(-0.0529793352741)*x_ref[1]
        ref[2,1,0]=(-0.307908670388)*x_ref[0]**o+(-0.929987734045)*x_ref[0]+(-0.167190563092)*x_ref[1]**o+(0.126466628227)*x_ref[1]
        ref[2,1,1]=(-0.812045378916)*x_ref[0]**o+(0.676973938072)*x_ref[0]+(0.528243909204)*x_ref[1]**o+(-0.594241387691)*x_ref[1]
        ref[3,0,0]=(-0.0198606164114)*x_ref[0]**o+(-0.00574708685554)*x_ref[0]+(0.908946312745)*x_ref[1]**o+(0.775054099788)*x_ref[1]
        ref[3,0,1]=(-0.810078120518)*x_ref[0]**o+(-0.960585888485)*x_ref[0]+(-0.0747272427198)*x_ref[1]**o+(0.164107641817)*x_ref[1]
        ref[3,1,0]=(-0.535791917139)*x_ref[0]**o+(0.0424226440074)*x_ref[0]+(0.701479263902)*x_ref[1]**o+(0.787745423088)*x_ref[1]
        ref[3,1,1]=(-0.896532152793)*x_ref[0]**o+(-0.0224566920251)*x_ref[0]+(-0.554655626689)*x_ref[1]**o+(0.41012867952)*x_ref[1]
        ref[4,0,0]=(0.972433817496)*x_ref[0]**o+(-0.440662536033)*x_ref[0]+(-0.910456530943)*x_ref[1]**o+(-0.333429535028)*x_ref[1]
        ref[4,0,1]=(-0.415824313942)*x_ref[0]**o+(0.00255557712922)*x_ref[0]+(0.304897639195)*x_ref[1]**o+(-0.866600848052)*x_ref[1]
        ref[4,1,0]=(-0.866277739076)*x_ref[0]**o+(0.282976169776)*x_ref[0]+(0.477528278501)*x_ref[1]**o+(-0.233506497695)*x_ref[1]
        ref[4,1,1]=(-0.12729626381)*x_ref[0]**o+(-0.723204796785)*x_ref[0]+(0.0856495713442)*x_ref[1]**o+(-0.214819437731)*x_ref[1]
        ref[5,0,0]=(0.208133503644)*x_ref[0]**o+(-0.955529202895)*x_ref[0]+(0.0291569127831)*x_ref[1]**o+(0.489913918305)*x_ref[1]
        ref[5,0,1]=(0.96923667477)*x_ref[0]**o+(0.543776525315)*x_ref[0]+(0.38666450029)*x_ref[1]**o+(-0.193493550272)*x_ref[1]
        ref[5,1,0]=(-0.0912940016888)*x_ref[0]**o+(-0.477561072681)*x_ref[0]+(0.474816021613)*x_ref[1]**o+(-0.252664229253)*x_ref[1]
        ref[5,1,1]=(0.748337265159)*x_ref[0]**o+(-0.706194454618)*x_ref[0]+(-0.504466281507)*x_ref[1]**o+(0.582500817679)*x_ref[1]
      else:
        arg[0,0,0]=(-0.179972720095)*x[0]**o+(-0.546715268523)*x[0]+(0.835867431319)*x[1]**o+(0.626798499259)*x[1]+(0.397356013278)*x[2]**o+(0.0771586784158)*x[2]
        arg[0,0,1]=(0.625676031039)*x[0]**o+(0.65843849475)*x[0]+(0.456469199307)*x[1]**o+(-0.970709759479)*x[1]+(-0.699606143992)*x[2]**o+(0.76429400724)*x[2]
        arg[0,1,0]=(0.613916424165)*x[0]**o+(0.556778588303)*x[0]+(-0.509658743529)*x[1]**o+(-0.978718968218)*x[1]+(-0.603928369428)*x[2]**o+(-0.550813953692)*x[2]
        arg[0,1,1]=(-0.935214755436)*x[0]**o+(0.44745901813)*x[0]+(-0.582828846453)*x[1]**o+(-0.181547183233)*x[1]+(-0.183912053485)*x[2]**o+(0.446152494296)*x[2]
        arg[1,0,0]=(-0.486345219779)*x[0]**o+(0.753529436833)*x[0]+(0.138157856552)*x[1]**o+(-0.874849020539)*x[1]+(0.374396543792)*x[2]**o+(0.26548456046)*x[2]
        arg[1,0,1]=(0.492908753036)*x[0]**o+(-0.418118685105)*x[0]+(0.880817569823)*x[1]**o+(-0.827691145524)*x[1]+(0.238633156133)*x[2]**o+(-0.463687731341)*x[2]
        arg[1,1,0]=(-0.397373578442)*x[0]**o+(-0.316321105607)*x[0]+(0.647511225459)*x[1]**o+(-0.82553752195)*x[1]+(-0.715549303171)*x[2]**o+(0.804623957459)*x[2]
        arg[1,1,1]=(0.414743130234)*x[0]**o+(0.0201794413683)*x[0]+(-0.636337599584)*x[1]**o+(-0.230053616026)*x[1]+(-0.4302153311)*x[2]**o+(-0.387468029626)*x[2]
        arg[2,0,0]=(0.715100517336)*x[0]**o+(-0.4286120178)*x[0]+(0.0987858542117)*x[1]**o+(0.815093470301)*x[1]+(0.293455353257)*x[2]**o+(0.200576090379)*x[2]
        arg[2,0,1]=(-0.158330015822)*x[0]**o+(0.363339023457)*x[0]+(-0.796055594971)*x[1]**o+(-0.213299205306)*x[1]+(-0.136620988812)*x[2]**o+(0.10733886947)*x[2]
        arg[2,1,0]=(-0.258729405118)*x[0]**o+(0.0669553480038)*x[0]+(0.997663166613)*x[1]**o+(0.00398861063931)*x[1]+(-0.989459856803)*x[2]**o+(0.932687384325)*x[2]
        arg[2,1,1]=(-0.974431322319)*x[0]**o+(-0.486447431804)*x[0]+(0.871784873347)*x[1]**o+(-0.978904038765)*x[1]+(0.647341201506)*x[2]**o+(-0.269683556528)*x[2]
        arg[3,0,0]=(-0.369507398073)*x[0]**o+(-0.772085607074)*x[0]+(0.417550881937)*x[1]**o+(0.271743719652)*x[1]+(0.881807154737)*x[2]**o+(0.54725428039)*x[2]
        arg[3,0,1]=(0.440089335941)*x[0]**o+(0.0844171421294)*x[0]+(0.926389704362)*x[1]**o+(-0.0709182998297)*x[1]+(0.903978660317)*x[2]**o+(-0.160556206716)*x[2]
        arg[3,1,0]=(-0.49936391282)*x[0]**o+(-0.632753931881)*x[0]+(-0.604686084554)*x[1]**o+(-0.954205199531)*x[1]+(-0.535102184845)*x[2]**o+(0.621683814123)*x[2]
        arg[3,1,1]=(-0.117790973132)*x[0]**o+(-0.649091727896)*x[0]+(0.925682916785)*x[1]**o+(0.852053515469)*x[1]+(-0.479218429897)*x[2]**o+(0.426823478575)*x[2]
        arg[4,0,0]=(0.454174564319)*x[0]**o+(-0.855975647406)*x[0]+(0.775972969288)*x[1]**o+(0.267900999893)*x[1]+(0.221387501995)*x[2]**o+(-0.26413791201)*x[2]
        arg[4,0,1]=(0.875191302431)*x[0]**o+(-0.362782502152)*x[0]+(0.24600070339)*x[1]**o+(-0.566709533298)*x[1]+(-0.82187455531)*x[2]**o+(-0.943929640301)*x[2]
        arg[4,1,0]=(-0.409313259973)*x[0]**o+(-0.784833509001)*x[0]+(-0.939690505056)*x[1]**o+(-0.0736280226974)*x[1]+(0.939072745486)*x[2]**o+(0.611697159959)*x[2]
        arg[4,1,1]=(0.893374246808)*x[0]**o+(0.36113513561)*x[0]+(0.227899271271)*x[1]**o+(-0.847221657012)*x[1]+(-0.0106368947417)*x[2]**o+(0.0358896817595)*x[2]
        arg[5,0,0]=(0.0329230524924)*x[0]**o+(0.0565192084014)*x[0]+(-0.68398448658)*x[1]**o+(0.909035618076)*x[1]+(-0.900060440872)*x[2]**o+(-0.481014787833)*x[2]
        arg[5,0,1]=(0.404966692391)*x[0]**o+(0.647468764373)*x[0]+(-0.723847499346)*x[1]**o+(-0.0740896957606)*x[1]+(-0.00628515079923)*x[2]**o+(0.906667863539)*x[2]
        arg[5,1,0]=(0.983181781531)*x[0]**o+(-0.297325990236)*x[0]+(0.00489064543352)*x[1]**o+(0.667560767038)*x[1]+(-0.665768141301)*x[2]**o+(0.772761393608)*x[2]
        arg[5,1,1]=(-0.496581691354)*x[0]**o+(0.0121337979505)*x[0]+(0.492281382726)*x[1]**o+(0.545232714025)*x[1]+(0.418321481532)*x[2]**o+(0.489635969764)*x[2]
        ref[0,0,0]=(-0.179972720095)*x_ref[0]**o+(-0.546715268523)*x_ref[0]+(0.835867431319)*x_ref[1]**o+(0.626798499259)*x_ref[1]+(0.397356013278)*x_ref[2]**o+(0.0771586784158)*x_ref[2]
        ref[0,0,1]=(0.625676031039)*x_ref[0]**o+(0.65843849475)*x_ref[0]+(0.456469199307)*x_ref[1]**o+(-0.970709759479)*x_ref[1]+(-0.699606143992)*x_ref[2]**o+(0.76429400724)*x_ref[2]
        ref[0,1,0]=(0.613916424165)*x_ref[0]**o+(0.556778588303)*x_ref[0]+(-0.509658743529)*x_ref[1]**o+(-0.978718968218)*x_ref[1]+(-0.603928369428)*x_ref[2]**o+(-0.550813953692)*x_ref[2]
        ref[0,1,1]=(-0.935214755436)*x_ref[0]**o+(0.44745901813)*x_ref[0]+(-0.582828846453)*x_ref[1]**o+(-0.181547183233)*x_ref[1]+(-0.183912053485)*x_ref[2]**o+(0.446152494296)*x_ref[2]
        ref[1,0,0]=(-0.486345219779)*x_ref[0]**o+(0.753529436833)*x_ref[0]+(0.138157856552)*x_ref[1]**o+(-0.874849020539)*x_ref[1]+(0.374396543792)*x_ref[2]**o+(0.26548456046)*x_ref[2]
        ref[1,0,1]=(0.492908753036)*x_ref[0]**o+(-0.418118685105)*x_ref[0]+(0.880817569823)*x_ref[1]**o+(-0.827691145524)*x_ref[1]+(0.238633156133)*x_ref[2]**o+(-0.463687731341)*x_ref[2]
        ref[1,1,0]=(-0.397373578442)*x_ref[0]**o+(-0.316321105607)*x_ref[0]+(0.647511225459)*x_ref[1]**o+(-0.82553752195)*x_ref[1]+(-0.715549303171)*x_ref[2]**o+(0.804623957459)*x_ref[2]
        ref[1,1,1]=(0.414743130234)*x_ref[0]**o+(0.0201794413683)*x_ref[0]+(-0.636337599584)*x_ref[1]**o+(-0.230053616026)*x_ref[1]+(-0.4302153311)*x_ref[2]**o+(-0.387468029626)*x_ref[2]
        ref[2,0,0]=(0.715100517336)*x_ref[0]**o+(-0.4286120178)*x_ref[0]+(0.0987858542117)*x_ref[1]**o+(0.815093470301)*x_ref[1]+(0.293455353257)*x_ref[2]**o+(0.200576090379)*x_ref[2]
        ref[2,0,1]=(-0.158330015822)*x_ref[0]**o+(0.363339023457)*x_ref[0]+(-0.796055594971)*x_ref[1]**o+(-0.213299205306)*x_ref[1]+(-0.136620988812)*x_ref[2]**o+(0.10733886947)*x_ref[2]
        ref[2,1,0]=(-0.258729405118)*x_ref[0]**o+(0.0669553480038)*x_ref[0]+(0.997663166613)*x_ref[1]**o+(0.00398861063931)*x_ref[1]+(-0.989459856803)*x_ref[2]**o+(0.932687384325)*x_ref[2]
        ref[2,1,1]=(-0.974431322319)*x_ref[0]**o+(-0.486447431804)*x_ref[0]+(0.871784873347)*x_ref[1]**o+(-0.978904038765)*x_ref[1]+(0.647341201506)*x_ref[2]**o+(-0.269683556528)*x_ref[2]
        ref[3,0,0]=(-0.369507398073)*x_ref[0]**o+(-0.772085607074)*x_ref[0]+(0.417550881937)*x_ref[1]**o+(0.271743719652)*x_ref[1]+(0.881807154737)*x_ref[2]**o+(0.54725428039)*x_ref[2]
        ref[3,0,1]=(0.440089335941)*x_ref[0]**o+(0.0844171421294)*x_ref[0]+(0.926389704362)*x_ref[1]**o+(-0.0709182998297)*x_ref[1]+(0.903978660317)*x_ref[2]**o+(-0.160556206716)*x_ref[2]
        ref[3,1,0]=(-0.49936391282)*x_ref[0]**o+(-0.632753931881)*x_ref[0]+(-0.604686084554)*x_ref[1]**o+(-0.954205199531)*x_ref[1]+(-0.535102184845)*x_ref[2]**o+(0.621683814123)*x_ref[2]
        ref[3,1,1]=(-0.117790973132)*x_ref[0]**o+(-0.649091727896)*x_ref[0]+(0.925682916785)*x_ref[1]**o+(0.852053515469)*x_ref[1]+(-0.479218429897)*x_ref[2]**o+(0.426823478575)*x_ref[2]
        ref[4,0,0]=(0.454174564319)*x_ref[0]**o+(-0.855975647406)*x_ref[0]+(0.775972969288)*x_ref[1]**o+(0.267900999893)*x_ref[1]+(0.221387501995)*x_ref[2]**o+(-0.26413791201)*x_ref[2]
        ref[4,0,1]=(0.875191302431)*x_ref[0]**o+(-0.362782502152)*x_ref[0]+(0.24600070339)*x_ref[1]**o+(-0.566709533298)*x_ref[1]+(-0.82187455531)*x_ref[2]**o+(-0.943929640301)*x_ref[2]
        ref[4,1,0]=(-0.409313259973)*x_ref[0]**o+(-0.784833509001)*x_ref[0]+(-0.939690505056)*x_ref[1]**o+(-0.0736280226974)*x_ref[1]+(0.939072745486)*x_ref[2]**o+(0.611697159959)*x_ref[2]
        ref[4,1,1]=(0.893374246808)*x_ref[0]**o+(0.36113513561)*x_ref[0]+(0.227899271271)*x_ref[1]**o+(-0.847221657012)*x_ref[1]+(-0.0106368947417)*x_ref[2]**o+(0.0358896817595)*x_ref[2]
        ref[5,0,0]=(0.0329230524924)*x_ref[0]**o+(0.0565192084014)*x_ref[0]+(-0.68398448658)*x_ref[1]**o+(0.909035618076)*x_ref[1]+(-0.900060440872)*x_ref[2]**o+(-0.481014787833)*x_ref[2]
        ref[5,0,1]=(0.404966692391)*x_ref[0]**o+(0.647468764373)*x_ref[0]+(-0.723847499346)*x_ref[1]**o+(-0.0740896957606)*x_ref[1]+(-0.00628515079923)*x_ref[2]**o+(0.906667863539)*x_ref[2]
        ref[5,1,0]=(0.983181781531)*x_ref[0]**o+(-0.297325990236)*x_ref[0]+(0.00489064543352)*x_ref[1]**o+(0.667560767038)*x_ref[1]+(-0.665768141301)*x_ref[2]**o+(0.772761393608)*x_ref[2]
        ref[5,1,1]=(-0.496581691354)*x_ref[0]**o+(0.0121337979505)*x_ref[0]+(0.492281382726)*x_ref[1]**o+(0.545232714025)*x_ref[1]+(0.418321481532)*x_ref[2]**o+(0.489635969764)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.282394950702)*x[0]**o+(0.836580157584)*x[0]+(0.337652262648)*x[1]**o+(-0.442694263325)*x[1]
        arg[0,0,0,1]=(0.85150774028)*x[0]**o+(0.998297296593)*x[0]+(-0.710029954342)*x[1]**o+(0.0816168287837)*x[1]
        arg[0,0,1,0]=(-0.636435102838)*x[0]**o+(-0.613479863553)*x[0]+(-0.861506513168)*x[1]**o+(-0.242998961478)*x[1]
        arg[0,0,1,1]=(-0.00671863978521)*x[0]**o+(0.909404084068)*x[0]+(-0.734956067026)*x[1]**o+(0.256530538264)*x[1]
        arg[0,0,2,0]=(-0.646631608999)*x[0]**o+(-0.521130455843)*x[0]+(-0.44328404557)*x[1]**o+(-0.156860325817)*x[1]
        arg[0,0,2,1]=(-0.817550008949)*x[0]**o+(0.702414497795)*x[0]+(0.946449680763)*x[1]**o+(-0.61576361925)*x[1]
        arg[0,1,0,0]=(-0.527206691595)*x[0]**o+(-0.43469216984)*x[0]+(-0.108455901154)*x[1]**o+(0.435646962156)*x[1]
        arg[0,1,0,1]=(-0.148522219535)*x[0]**o+(0.146912249755)*x[0]+(-0.414934945665)*x[1]**o+(-0.422754988714)*x[1]
        arg[0,1,1,0]=(0.177760972947)*x[0]**o+(-0.938522530862)*x[0]+(0.566229652264)*x[1]**o+(0.255269743628)*x[1]
        arg[0,1,1,1]=(-0.317807933792)*x[0]**o+(0.432181451246)*x[0]+(-0.358821513362)*x[1]**o+(-0.470517232585)*x[1]
        arg[0,1,2,0]=(0.486536430227)*x[0]**o+(-0.533653809349)*x[0]+(-0.126612514045)*x[1]**o+(0.103088115487)*x[1]
        arg[0,1,2,1]=(-0.995976559589)*x[0]**o+(0.965537167421)*x[0]+(-0.351307699531)*x[1]**o+(-0.684367434666)*x[1]
        arg[0,2,0,0]=(0.976008613768)*x[0]**o+(-0.890573007002)*x[0]+(-0.440709423944)*x[1]**o+(0.315724459367)*x[1]
        arg[0,2,0,1]=(-0.595275811962)*x[0]**o+(-0.599634144509)*x[0]+(0.646466002346)*x[1]**o+(-0.94529568651)*x[1]
        arg[0,2,1,0]=(0.977691521821)*x[0]**o+(-0.0730997643682)*x[0]+(-0.338897498737)*x[1]**o+(0.246308678824)*x[1]
        arg[0,2,1,1]=(-0.111317410732)*x[0]**o+(-0.574381793398)*x[0]+(-0.0314431899843)*x[1]**o+(0.393105047096)*x[1]
        arg[0,2,2,0]=(0.886320020757)*x[0]**o+(-0.785042003544)*x[0]+(0.82609269979)*x[1]**o+(-0.677169484337)*x[1]
        arg[0,2,2,1]=(0.811027138079)*x[0]**o+(-0.375108470948)*x[0]+(0.781472464253)*x[1]**o+(0.217640794143)*x[1]
        arg[0,3,0,0]=(0.264884117091)*x[0]**o+(0.116591319557)*x[0]+(0.282667679763)*x[1]**o+(0.111123454337)*x[1]
        arg[0,3,0,1]=(0.0451626985922)*x[0]**o+(-0.93771279728)*x[0]+(0.129171764072)*x[1]**o+(0.598367954162)*x[1]
        arg[0,3,1,0]=(0.905243511086)*x[0]**o+(0.0837062573586)*x[0]+(0.0657223248433)*x[1]**o+(-0.746644690072)*x[1]
        arg[0,3,1,1]=(-0.0547856755485)*x[0]**o+(-0.22157806755)*x[0]+(0.124623625108)*x[1]**o+(0.0436590157081)*x[1]
        arg[0,3,2,0]=(-0.558323744577)*x[0]**o+(-0.646043519007)*x[0]+(0.991885458004)*x[1]**o+(-0.634140808163)*x[1]
        arg[0,3,2,1]=(-0.27247983028)*x[0]**o+(0.803906404485)*x[0]+(-0.816826609924)*x[1]**o+(-0.969532783336)*x[1]
        arg[0,4,0,0]=(0.638733842891)*x[0]**o+(-0.972793270208)*x[0]+(0.714611008099)*x[1]**o+(0.586988595847)*x[1]
        arg[0,4,0,1]=(0.63584760734)*x[0]**o+(-0.653740236747)*x[0]+(-0.596304731446)*x[1]**o+(0.42519786429)*x[1]
        arg[0,4,1,0]=(0.0665562973529)*x[0]**o+(0.362152092093)*x[0]+(0.732530335739)*x[1]**o+(0.618536326941)*x[1]
        arg[0,4,1,1]=(0.486060667608)*x[0]**o+(0.59489013279)*x[0]+(-0.547237017498)*x[1]**o+(0.0371894705857)*x[1]
        arg[0,4,2,0]=(0.395505844397)*x[0]**o+(-0.869090434535)*x[0]+(-0.606640827437)*x[1]**o+(0.269995459083)*x[1]
        arg[0,4,2,1]=(0.304671547176)*x[0]**o+(-0.515772523849)*x[0]+(0.546114300876)*x[1]**o+(0.530247404341)*x[1]
        arg[1,0,0,0]=(0.408332677433)*x[0]**o+(-0.0778807359918)*x[0]+(-0.803337489297)*x[1]**o+(-0.123835718422)*x[1]
        arg[1,0,0,1]=(-0.00441620254944)*x[0]**o+(0.0321259598831)*x[0]+(-0.998381508472)*x[1]**o+(0.800767747511)*x[1]
        arg[1,0,1,0]=(-0.923683057418)*x[0]**o+(0.579405114902)*x[0]+(-0.939639823585)*x[1]**o+(-0.138423358134)*x[1]
        arg[1,0,1,1]=(-0.558420234848)*x[0]**o+(0.133316569813)*x[0]+(-0.294134717329)*x[1]**o+(-0.750897749044)*x[1]
        arg[1,0,2,0]=(0.478951298545)*x[0]**o+(-0.807173978799)*x[0]+(0.852438146763)*x[1]**o+(-0.103814229563)*x[1]
        arg[1,0,2,1]=(-0.875250726603)*x[0]**o+(-0.934803428491)*x[0]+(0.143352873672)*x[1]**o+(0.600934881206)*x[1]
        arg[1,1,0,0]=(-0.0539476321018)*x[0]**o+(0.194177663417)*x[0]+(-0.862877991009)*x[1]**o+(-0.438337993664)*x[1]
        arg[1,1,0,1]=(-0.623379355474)*x[0]**o+(-0.0746994739748)*x[0]+(0.365005438956)*x[1]**o+(-0.431737688205)*x[1]
        arg[1,1,1,0]=(0.0936054461735)*x[0]**o+(-0.974594269446)*x[0]+(-0.121917598428)*x[1]**o+(-0.84429083727)*x[1]
        arg[1,1,1,1]=(0.967763348672)*x[0]**o+(0.331698826475)*x[0]+(0.66448515877)*x[1]**o+(0.00214985930661)*x[1]
        arg[1,1,2,0]=(-0.312146580762)*x[0]**o+(0.998152287706)*x[0]+(-0.474242942777)*x[1]**o+(0.637827347846)*x[1]
        arg[1,1,2,1]=(-0.556765443436)*x[0]**o+(0.227327929774)*x[0]+(0.739457845569)*x[1]**o+(-0.615070017663)*x[1]
        arg[1,2,0,0]=(-0.23030660552)*x[0]**o+(0.750479465953)*x[0]+(0.278215408392)*x[1]**o+(-0.640618698224)*x[1]
        arg[1,2,0,1]=(-0.0713217101129)*x[0]**o+(0.500737477887)*x[0]+(-0.972834770005)*x[1]**o+(-0.420386056033)*x[1]
        arg[1,2,1,0]=(-0.497322989847)*x[0]**o+(-0.528685124841)*x[0]+(-0.934386348496)*x[1]**o+(0.191930553839)*x[1]
        arg[1,2,1,1]=(0.124362419781)*x[0]**o+(-0.0457849787456)*x[0]+(-0.360596154712)*x[1]**o+(0.799736253715)*x[1]
        arg[1,2,2,0]=(-0.346294878893)*x[0]**o+(0.427772309361)*x[0]+(-0.0642666791012)*x[1]**o+(-0.214442238311)*x[1]
        arg[1,2,2,1]=(0.24434837411)*x[0]**o+(-0.913373159786)*x[0]+(0.158750862391)*x[1]**o+(-0.544328807576)*x[1]
        arg[1,3,0,0]=(0.870847002687)*x[0]**o+(-0.466884462301)*x[0]+(-0.343358829445)*x[1]**o+(-0.180956419766)*x[1]
        arg[1,3,0,1]=(-0.813518850863)*x[0]**o+(0.960810467501)*x[0]+(0.35091416904)*x[1]**o+(0.84534544948)*x[1]
        arg[1,3,1,0]=(-0.284687751218)*x[0]**o+(-0.274583519355)*x[0]+(0.338892764102)*x[1]**o+(-0.145142443753)*x[1]
        arg[1,3,1,1]=(0.921064262137)*x[0]**o+(0.152552421678)*x[0]+(0.827247693375)*x[1]**o+(-0.569529848619)*x[1]
        arg[1,3,2,0]=(0.133567630356)*x[0]**o+(-0.0429397358364)*x[0]+(-0.175383170663)*x[1]**o+(0.584407631334)*x[1]
        arg[1,3,2,1]=(0.203880167371)*x[0]**o+(-0.422978073825)*x[0]+(-0.210109267588)*x[1]**o+(-0.902327373219)*x[1]
        arg[1,4,0,0]=(-0.566386864876)*x[0]**o+(-0.139795085483)*x[0]+(-0.00852913464238)*x[1]**o+(0.569151381884)*x[1]
        arg[1,4,0,1]=(-0.937256625212)*x[0]**o+(0.563438253128)*x[0]+(-0.586182238715)*x[1]**o+(-0.566169678723)*x[1]
        arg[1,4,1,0]=(-0.73296890712)*x[0]**o+(0.359249594406)*x[0]+(0.66050790616)*x[1]**o+(-0.367919797426)*x[1]
        arg[1,4,1,1]=(-0.81165734248)*x[0]**o+(0.460592817838)*x[0]+(0.953573909492)*x[1]**o+(0.329659021274)*x[1]
        arg[1,4,2,0]=(0.345994483595)*x[0]**o+(-0.18547324343)*x[0]+(0.669452926238)*x[1]**o+(0.42513249272)*x[1]
        arg[1,4,2,1]=(0.250677642616)*x[0]**o+(0.466118496803)*x[0]+(-0.776608543595)*x[1]**o+(0.449144243682)*x[1]
        arg[2,0,0,0]=(0.244172909021)*x[0]**o+(0.0311068322898)*x[0]+(-0.909348039508)*x[1]**o+(0.376500719414)*x[1]
        arg[2,0,0,1]=(-0.033516640667)*x[0]**o+(0.698424718328)*x[0]+(-0.282814785542)*x[1]**o+(0.61118022947)*x[1]
        arg[2,0,1,0]=(-0.946672649538)*x[0]**o+(0.814325244591)*x[0]+(-0.353215562825)*x[1]**o+(-0.138150257429)*x[1]
        arg[2,0,1,1]=(-0.886509051692)*x[0]**o+(-0.0468334385819)*x[0]+(0.71984719189)*x[1]**o+(-0.324744971493)*x[1]
        arg[2,0,2,0]=(-0.780921647053)*x[0]**o+(-0.401418496932)*x[0]+(0.647847274332)*x[1]**o+(-0.831834322278)*x[1]
        arg[2,0,2,1]=(0.835155004862)*x[0]**o+(0.684459986902)*x[0]+(-0.301244240677)*x[1]**o+(-0.161292780187)*x[1]
        arg[2,1,0,0]=(-0.652199398394)*x[0]**o+(0.62480722178)*x[0]+(-0.474535712914)*x[1]**o+(0.926061797766)*x[1]
        arg[2,1,0,1]=(-0.813365478568)*x[0]**o+(-0.38795792417)*x[0]+(0.378161353697)*x[1]**o+(0.70581806211)*x[1]
        arg[2,1,1,0]=(0.729265594934)*x[0]**o+(-0.469397842359)*x[0]+(0.642472199926)*x[1]**o+(-0.779527283011)*x[1]
        arg[2,1,1,1]=(0.0517342917574)*x[0]**o+(-0.414111007917)*x[0]+(-0.89437375241)*x[1]**o+(-0.830119585842)*x[1]
        arg[2,1,2,0]=(0.769898486257)*x[0]**o+(0.791725624832)*x[0]+(0.297519329396)*x[1]**o+(-0.959462006611)*x[1]
        arg[2,1,2,1]=(0.817802144251)*x[0]**o+(0.69851258911)*x[0]+(-0.174401735228)*x[1]**o+(0.13449696498)*x[1]
        arg[2,2,0,0]=(0.547972662256)*x[0]**o+(-0.66491929979)*x[0]+(0.330783041672)*x[1]**o+(0.816196730606)*x[1]
        arg[2,2,0,1]=(0.245971097763)*x[0]**o+(0.213087481481)*x[0]+(-0.840766291392)*x[1]**o+(0.388414557376)*x[1]
        arg[2,2,1,0]=(0.523151676444)*x[0]**o+(0.854817486027)*x[0]+(-0.53570188528)*x[1]**o+(-0.68149362242)*x[1]
        arg[2,2,1,1]=(-0.984200247787)*x[0]**o+(0.83324365092)*x[0]+(-0.896184408816)*x[1]**o+(0.102276243487)*x[1]
        arg[2,2,2,0]=(-0.216357164308)*x[0]**o+(-0.597984882107)*x[0]+(-0.205435921511)*x[1]**o+(0.484947652759)*x[1]
        arg[2,2,2,1]=(-0.650046253868)*x[0]**o+(-0.624878186105)*x[0]+(0.786519914406)*x[1]**o+(0.28027273095)*x[1]
        arg[2,3,0,0]=(-0.0821718346579)*x[0]**o+(-0.687049917467)*x[0]+(0.380792333736)*x[1]**o+(0.151185837715)*x[1]
        arg[2,3,0,1]=(0.61796113854)*x[0]**o+(0.844624413065)*x[0]+(-0.980619401098)*x[1]**o+(-0.854924862641)*x[1]
        arg[2,3,1,0]=(0.812293906883)*x[0]**o+(-0.225593208768)*x[0]+(0.964175374163)*x[1]**o+(0.848850064926)*x[1]
        arg[2,3,1,1]=(0.376453572937)*x[0]**o+(0.994818812059)*x[0]+(0.58872018303)*x[1]**o+(0.68332473361)*x[1]
        arg[2,3,2,0]=(-0.00948421972234)*x[0]**o+(-0.680685462738)*x[0]+(-0.767076278167)*x[1]**o+(0.210165437284)*x[1]
        arg[2,3,2,1]=(-0.455199698263)*x[0]**o+(-0.763430401419)*x[0]+(0.183773350212)*x[1]**o+(0.43105940923)*x[1]
        arg[2,4,0,0]=(-0.673962025988)*x[0]**o+(-0.679073069725)*x[0]+(0.11018431715)*x[1]**o+(0.349020527482)*x[1]
        arg[2,4,0,1]=(-0.876691282076)*x[0]**o+(0.920902226631)*x[0]+(-0.477417470246)*x[1]**o+(0.836417982216)*x[1]
        arg[2,4,1,0]=(0.567966851024)*x[0]**o+(0.583241431716)*x[0]+(-0.271851271008)*x[1]**o+(0.678727902863)*x[1]
        arg[2,4,1,1]=(0.707400094)*x[0]**o+(0.802820856444)*x[0]+(0.84518937616)*x[1]**o+(-0.0267193533403)*x[1]
        arg[2,4,2,0]=(0.213268920036)*x[0]**o+(0.780721415328)*x[0]+(0.839292948154)*x[1]**o+(-0.0376334744443)*x[1]
        arg[2,4,2,1]=(0.136255687031)*x[0]**o+(-0.582766278693)*x[0]+(0.257662521321)*x[1]**o+(0.808337521979)*x[1]
        arg[3,0,0,0]=(-0.524068770822)*x[0]**o+(0.942236878842)*x[0]+(0.889496255793)*x[1]**o+(-0.904810742393)*x[1]
        arg[3,0,0,1]=(0.780157465475)*x[0]**o+(-0.763811668832)*x[0]+(-0.517482070639)*x[1]**o+(0.662770780676)*x[1]
        arg[3,0,1,0]=(-0.693227564812)*x[0]**o+(-0.763506211841)*x[0]+(-0.427709881126)*x[1]**o+(0.159109827204)*x[1]
        arg[3,0,1,1]=(0.644975629084)*x[0]**o+(-0.928112761899)*x[0]+(-0.530772740083)*x[1]**o+(-0.723174790087)*x[1]
        arg[3,0,2,0]=(-0.923467274097)*x[0]**o+(0.806128117288)*x[0]+(0.177779363207)*x[1]**o+(0.862243885542)*x[1]
        arg[3,0,2,1]=(-0.0823252901358)*x[0]**o+(0.289022916284)*x[0]+(-0.518508771021)*x[1]**o+(0.174564769845)*x[1]
        arg[3,1,0,0]=(0.608837578444)*x[0]**o+(0.898216758361)*x[0]+(-0.844730109761)*x[1]**o+(0.745974173207)*x[1]
        arg[3,1,0,1]=(-0.11776411078)*x[0]**o+(0.186256502356)*x[0]+(0.565289400333)*x[1]**o+(-0.776173716206)*x[1]
        arg[3,1,1,0]=(-0.170736580152)*x[0]**o+(0.829581661683)*x[0]+(0.310135833859)*x[1]**o+(-0.509629545687)*x[1]
        arg[3,1,1,1]=(-0.225246122751)*x[0]**o+(0.586754377294)*x[0]+(-0.726064725635)*x[1]**o+(0.136952605704)*x[1]
        arg[3,1,2,0]=(-0.287932115478)*x[0]**o+(0.858031396153)*x[0]+(-0.493441745353)*x[1]**o+(0.601827664994)*x[1]
        arg[3,1,2,1]=(0.665503842862)*x[0]**o+(-0.515998970452)*x[0]+(-0.890055558624)*x[1]**o+(-0.0446371451457)*x[1]
        arg[3,2,0,0]=(-0.33741557585)*x[0]**o+(-0.416183752333)*x[0]+(-0.92430073538)*x[1]**o+(-0.486265090904)*x[1]
        arg[3,2,0,1]=(0.910860925994)*x[0]**o+(-0.896869755389)*x[0]+(0.0634658185324)*x[1]**o+(-0.383136455739)*x[1]
        arg[3,2,1,0]=(0.386839297976)*x[0]**o+(-0.190659927071)*x[0]+(-0.304846978185)*x[1]**o+(-0.92705511899)*x[1]
        arg[3,2,1,1]=(-0.0197812018683)*x[0]**o+(-0.479083095775)*x[0]+(0.502489976089)*x[1]**o+(-0.859440665475)*x[1]
        arg[3,2,2,0]=(-0.507672045299)*x[0]**o+(0.498689676072)*x[0]+(-0.719735804645)*x[1]**o+(-0.522053306674)*x[1]
        arg[3,2,2,1]=(0.46207629211)*x[0]**o+(-0.496697066039)*x[0]+(0.994069339904)*x[1]**o+(0.873222754449)*x[1]
        arg[3,3,0,0]=(0.126410029692)*x[0]**o+(0.51067050196)*x[0]+(-0.978384824088)*x[1]**o+(-0.68606930503)*x[1]
        arg[3,3,0,1]=(0.376888157008)*x[0]**o+(-0.14015764025)*x[0]+(0.523667700758)*x[1]**o+(0.0277726855258)*x[1]
        arg[3,3,1,0]=(-0.105865233294)*x[0]**o+(-0.0515230741203)*x[0]+(0.203766525036)*x[1]**o+(-0.0431893014387)*x[1]
        arg[3,3,1,1]=(-0.404351448486)*x[0]**o+(-0.639161429157)*x[0]+(-0.957341091708)*x[1]**o+(-0.103346466984)*x[1]
        arg[3,3,2,0]=(-0.0523508780623)*x[0]**o+(-0.563589575544)*x[0]+(0.158487727325)*x[1]**o+(-0.519455492178)*x[1]
        arg[3,3,2,1]=(-0.823286342218)*x[0]**o+(-0.493663499535)*x[0]+(-0.409825574064)*x[1]**o+(-0.0406101279806)*x[1]
        arg[3,4,0,0]=(0.926212910387)*x[0]**o+(-0.0944507963)*x[0]+(0.640740499391)*x[1]**o+(-0.106969076047)*x[1]
        arg[3,4,0,1]=(-0.912044993941)*x[0]**o+(-0.495323777918)*x[0]+(-0.595831845108)*x[1]**o+(-0.313723426318)*x[1]
        arg[3,4,1,0]=(-0.0730898742548)*x[0]**o+(0.0524008484492)*x[0]+(-0.948232935497)*x[1]**o+(0.952670824572)*x[1]
        arg[3,4,1,1]=(-0.81869438138)*x[0]**o+(-0.894654266952)*x[0]+(0.414078621872)*x[1]**o+(-0.876690817847)*x[1]
        arg[3,4,2,0]=(0.00688155078441)*x[0]**o+(0.845964586551)*x[0]+(0.0183149371892)*x[1]**o+(-0.882931394537)*x[1]
        arg[3,4,2,1]=(-0.734830097149)*x[0]**o+(-0.37824588859)*x[0]+(-0.815402620311)*x[1]**o+(-0.357905685517)*x[1]
        ref[0,0,0,0]=(-0.282394950702)*x_ref[0]**o+(0.836580157584)*x_ref[0]+(0.337652262648)*x_ref[1]**o+(-0.442694263325)*x_ref[1]
        ref[0,0,0,1]=(0.85150774028)*x_ref[0]**o+(0.998297296593)*x_ref[0]+(-0.710029954342)*x_ref[1]**o+(0.0816168287837)*x_ref[1]
        ref[0,0,1,0]=(-0.636435102838)*x_ref[0]**o+(-0.613479863553)*x_ref[0]+(-0.861506513168)*x_ref[1]**o+(-0.242998961478)*x_ref[1]
        ref[0,0,1,1]=(-0.00671863978521)*x_ref[0]**o+(0.909404084068)*x_ref[0]+(-0.734956067026)*x_ref[1]**o+(0.256530538264)*x_ref[1]
        ref[0,0,2,0]=(-0.646631608999)*x_ref[0]**o+(-0.521130455843)*x_ref[0]+(-0.44328404557)*x_ref[1]**o+(-0.156860325817)*x_ref[1]
        ref[0,0,2,1]=(-0.817550008949)*x_ref[0]**o+(0.702414497795)*x_ref[0]+(0.946449680763)*x_ref[1]**o+(-0.61576361925)*x_ref[1]
        ref[0,1,0,0]=(-0.527206691595)*x_ref[0]**o+(-0.43469216984)*x_ref[0]+(-0.108455901154)*x_ref[1]**o+(0.435646962156)*x_ref[1]
        ref[0,1,0,1]=(-0.148522219535)*x_ref[0]**o+(0.146912249755)*x_ref[0]+(-0.414934945665)*x_ref[1]**o+(-0.422754988714)*x_ref[1]
        ref[0,1,1,0]=(0.177760972947)*x_ref[0]**o+(-0.938522530862)*x_ref[0]+(0.566229652264)*x_ref[1]**o+(0.255269743628)*x_ref[1]
        ref[0,1,1,1]=(-0.317807933792)*x_ref[0]**o+(0.432181451246)*x_ref[0]+(-0.358821513362)*x_ref[1]**o+(-0.470517232585)*x_ref[1]
        ref[0,1,2,0]=(0.486536430227)*x_ref[0]**o+(-0.533653809349)*x_ref[0]+(-0.126612514045)*x_ref[1]**o+(0.103088115487)*x_ref[1]
        ref[0,1,2,1]=(-0.995976559589)*x_ref[0]**o+(0.965537167421)*x_ref[0]+(-0.351307699531)*x_ref[1]**o+(-0.684367434666)*x_ref[1]
        ref[0,2,0,0]=(0.976008613768)*x_ref[0]**o+(-0.890573007002)*x_ref[0]+(-0.440709423944)*x_ref[1]**o+(0.315724459367)*x_ref[1]
        ref[0,2,0,1]=(-0.595275811962)*x_ref[0]**o+(-0.599634144509)*x_ref[0]+(0.646466002346)*x_ref[1]**o+(-0.94529568651)*x_ref[1]
        ref[0,2,1,0]=(0.977691521821)*x_ref[0]**o+(-0.0730997643682)*x_ref[0]+(-0.338897498737)*x_ref[1]**o+(0.246308678824)*x_ref[1]
        ref[0,2,1,1]=(-0.111317410732)*x_ref[0]**o+(-0.574381793398)*x_ref[0]+(-0.0314431899843)*x_ref[1]**o+(0.393105047096)*x_ref[1]
        ref[0,2,2,0]=(0.886320020757)*x_ref[0]**o+(-0.785042003544)*x_ref[0]+(0.82609269979)*x_ref[1]**o+(-0.677169484337)*x_ref[1]
        ref[0,2,2,1]=(0.811027138079)*x_ref[0]**o+(-0.375108470948)*x_ref[0]+(0.781472464253)*x_ref[1]**o+(0.217640794143)*x_ref[1]
        ref[0,3,0,0]=(0.264884117091)*x_ref[0]**o+(0.116591319557)*x_ref[0]+(0.282667679763)*x_ref[1]**o+(0.111123454337)*x_ref[1]
        ref[0,3,0,1]=(0.0451626985922)*x_ref[0]**o+(-0.93771279728)*x_ref[0]+(0.129171764072)*x_ref[1]**o+(0.598367954162)*x_ref[1]
        ref[0,3,1,0]=(0.905243511086)*x_ref[0]**o+(0.0837062573586)*x_ref[0]+(0.0657223248433)*x_ref[1]**o+(-0.746644690072)*x_ref[1]
        ref[0,3,1,1]=(-0.0547856755485)*x_ref[0]**o+(-0.22157806755)*x_ref[0]+(0.124623625108)*x_ref[1]**o+(0.0436590157081)*x_ref[1]
        ref[0,3,2,0]=(-0.558323744577)*x_ref[0]**o+(-0.646043519007)*x_ref[0]+(0.991885458004)*x_ref[1]**o+(-0.634140808163)*x_ref[1]
        ref[0,3,2,1]=(-0.27247983028)*x_ref[0]**o+(0.803906404485)*x_ref[0]+(-0.816826609924)*x_ref[1]**o+(-0.969532783336)*x_ref[1]
        ref[0,4,0,0]=(0.638733842891)*x_ref[0]**o+(-0.972793270208)*x_ref[0]+(0.714611008099)*x_ref[1]**o+(0.586988595847)*x_ref[1]
        ref[0,4,0,1]=(0.63584760734)*x_ref[0]**o+(-0.653740236747)*x_ref[0]+(-0.596304731446)*x_ref[1]**o+(0.42519786429)*x_ref[1]
        ref[0,4,1,0]=(0.0665562973529)*x_ref[0]**o+(0.362152092093)*x_ref[0]+(0.732530335739)*x_ref[1]**o+(0.618536326941)*x_ref[1]
        ref[0,4,1,1]=(0.486060667608)*x_ref[0]**o+(0.59489013279)*x_ref[0]+(-0.547237017498)*x_ref[1]**o+(0.0371894705857)*x_ref[1]
        ref[0,4,2,0]=(0.395505844397)*x_ref[0]**o+(-0.869090434535)*x_ref[0]+(-0.606640827437)*x_ref[1]**o+(0.269995459083)*x_ref[1]
        ref[0,4,2,1]=(0.304671547176)*x_ref[0]**o+(-0.515772523849)*x_ref[0]+(0.546114300876)*x_ref[1]**o+(0.530247404341)*x_ref[1]
        ref[1,0,0,0]=(0.408332677433)*x_ref[0]**o+(-0.0778807359918)*x_ref[0]+(-0.803337489297)*x_ref[1]**o+(-0.123835718422)*x_ref[1]
        ref[1,0,0,1]=(-0.00441620254944)*x_ref[0]**o+(0.0321259598831)*x_ref[0]+(-0.998381508472)*x_ref[1]**o+(0.800767747511)*x_ref[1]
        ref[1,0,1,0]=(-0.923683057418)*x_ref[0]**o+(0.579405114902)*x_ref[0]+(-0.939639823585)*x_ref[1]**o+(-0.138423358134)*x_ref[1]
        ref[1,0,1,1]=(-0.558420234848)*x_ref[0]**o+(0.133316569813)*x_ref[0]+(-0.294134717329)*x_ref[1]**o+(-0.750897749044)*x_ref[1]
        ref[1,0,2,0]=(0.478951298545)*x_ref[0]**o+(-0.807173978799)*x_ref[0]+(0.852438146763)*x_ref[1]**o+(-0.103814229563)*x_ref[1]
        ref[1,0,2,1]=(-0.875250726603)*x_ref[0]**o+(-0.934803428491)*x_ref[0]+(0.143352873672)*x_ref[1]**o+(0.600934881206)*x_ref[1]
        ref[1,1,0,0]=(-0.0539476321018)*x_ref[0]**o+(0.194177663417)*x_ref[0]+(-0.862877991009)*x_ref[1]**o+(-0.438337993664)*x_ref[1]
        ref[1,1,0,1]=(-0.623379355474)*x_ref[0]**o+(-0.0746994739748)*x_ref[0]+(0.365005438956)*x_ref[1]**o+(-0.431737688205)*x_ref[1]
        ref[1,1,1,0]=(0.0936054461735)*x_ref[0]**o+(-0.974594269446)*x_ref[0]+(-0.121917598428)*x_ref[1]**o+(-0.84429083727)*x_ref[1]
        ref[1,1,1,1]=(0.967763348672)*x_ref[0]**o+(0.331698826475)*x_ref[0]+(0.66448515877)*x_ref[1]**o+(0.00214985930661)*x_ref[1]
        ref[1,1,2,0]=(-0.312146580762)*x_ref[0]**o+(0.998152287706)*x_ref[0]+(-0.474242942777)*x_ref[1]**o+(0.637827347846)*x_ref[1]
        ref[1,1,2,1]=(-0.556765443436)*x_ref[0]**o+(0.227327929774)*x_ref[0]+(0.739457845569)*x_ref[1]**o+(-0.615070017663)*x_ref[1]
        ref[1,2,0,0]=(-0.23030660552)*x_ref[0]**o+(0.750479465953)*x_ref[0]+(0.278215408392)*x_ref[1]**o+(-0.640618698224)*x_ref[1]
        ref[1,2,0,1]=(-0.0713217101129)*x_ref[0]**o+(0.500737477887)*x_ref[0]+(-0.972834770005)*x_ref[1]**o+(-0.420386056033)*x_ref[1]
        ref[1,2,1,0]=(-0.497322989847)*x_ref[0]**o+(-0.528685124841)*x_ref[0]+(-0.934386348496)*x_ref[1]**o+(0.191930553839)*x_ref[1]
        ref[1,2,1,1]=(0.124362419781)*x_ref[0]**o+(-0.0457849787456)*x_ref[0]+(-0.360596154712)*x_ref[1]**o+(0.799736253715)*x_ref[1]
        ref[1,2,2,0]=(-0.346294878893)*x_ref[0]**o+(0.427772309361)*x_ref[0]+(-0.0642666791012)*x_ref[1]**o+(-0.214442238311)*x_ref[1]
        ref[1,2,2,1]=(0.24434837411)*x_ref[0]**o+(-0.913373159786)*x_ref[0]+(0.158750862391)*x_ref[1]**o+(-0.544328807576)*x_ref[1]
        ref[1,3,0,0]=(0.870847002687)*x_ref[0]**o+(-0.466884462301)*x_ref[0]+(-0.343358829445)*x_ref[1]**o+(-0.180956419766)*x_ref[1]
        ref[1,3,0,1]=(-0.813518850863)*x_ref[0]**o+(0.960810467501)*x_ref[0]+(0.35091416904)*x_ref[1]**o+(0.84534544948)*x_ref[1]
        ref[1,3,1,0]=(-0.284687751218)*x_ref[0]**o+(-0.274583519355)*x_ref[0]+(0.338892764102)*x_ref[1]**o+(-0.145142443753)*x_ref[1]
        ref[1,3,1,1]=(0.921064262137)*x_ref[0]**o+(0.152552421678)*x_ref[0]+(0.827247693375)*x_ref[1]**o+(-0.569529848619)*x_ref[1]
        ref[1,3,2,0]=(0.133567630356)*x_ref[0]**o+(-0.0429397358364)*x_ref[0]+(-0.175383170663)*x_ref[1]**o+(0.584407631334)*x_ref[1]
        ref[1,3,2,1]=(0.203880167371)*x_ref[0]**o+(-0.422978073825)*x_ref[0]+(-0.210109267588)*x_ref[1]**o+(-0.902327373219)*x_ref[1]
        ref[1,4,0,0]=(-0.566386864876)*x_ref[0]**o+(-0.139795085483)*x_ref[0]+(-0.00852913464238)*x_ref[1]**o+(0.569151381884)*x_ref[1]
        ref[1,4,0,1]=(-0.937256625212)*x_ref[0]**o+(0.563438253128)*x_ref[0]+(-0.586182238715)*x_ref[1]**o+(-0.566169678723)*x_ref[1]
        ref[1,4,1,0]=(-0.73296890712)*x_ref[0]**o+(0.359249594406)*x_ref[0]+(0.66050790616)*x_ref[1]**o+(-0.367919797426)*x_ref[1]
        ref[1,4,1,1]=(-0.81165734248)*x_ref[0]**o+(0.460592817838)*x_ref[0]+(0.953573909492)*x_ref[1]**o+(0.329659021274)*x_ref[1]
        ref[1,4,2,0]=(0.345994483595)*x_ref[0]**o+(-0.18547324343)*x_ref[0]+(0.669452926238)*x_ref[1]**o+(0.42513249272)*x_ref[1]
        ref[1,4,2,1]=(0.250677642616)*x_ref[0]**o+(0.466118496803)*x_ref[0]+(-0.776608543595)*x_ref[1]**o+(0.449144243682)*x_ref[1]
        ref[2,0,0,0]=(0.244172909021)*x_ref[0]**o+(0.0311068322898)*x_ref[0]+(-0.909348039508)*x_ref[1]**o+(0.376500719414)*x_ref[1]
        ref[2,0,0,1]=(-0.033516640667)*x_ref[0]**o+(0.698424718328)*x_ref[0]+(-0.282814785542)*x_ref[1]**o+(0.61118022947)*x_ref[1]
        ref[2,0,1,0]=(-0.946672649538)*x_ref[0]**o+(0.814325244591)*x_ref[0]+(-0.353215562825)*x_ref[1]**o+(-0.138150257429)*x_ref[1]
        ref[2,0,1,1]=(-0.886509051692)*x_ref[0]**o+(-0.0468334385819)*x_ref[0]+(0.71984719189)*x_ref[1]**o+(-0.324744971493)*x_ref[1]
        ref[2,0,2,0]=(-0.780921647053)*x_ref[0]**o+(-0.401418496932)*x_ref[0]+(0.647847274332)*x_ref[1]**o+(-0.831834322278)*x_ref[1]
        ref[2,0,2,1]=(0.835155004862)*x_ref[0]**o+(0.684459986902)*x_ref[0]+(-0.301244240677)*x_ref[1]**o+(-0.161292780187)*x_ref[1]
        ref[2,1,0,0]=(-0.652199398394)*x_ref[0]**o+(0.62480722178)*x_ref[0]+(-0.474535712914)*x_ref[1]**o+(0.926061797766)*x_ref[1]
        ref[2,1,0,1]=(-0.813365478568)*x_ref[0]**o+(-0.38795792417)*x_ref[0]+(0.378161353697)*x_ref[1]**o+(0.70581806211)*x_ref[1]
        ref[2,1,1,0]=(0.729265594934)*x_ref[0]**o+(-0.469397842359)*x_ref[0]+(0.642472199926)*x_ref[1]**o+(-0.779527283011)*x_ref[1]
        ref[2,1,1,1]=(0.0517342917574)*x_ref[0]**o+(-0.414111007917)*x_ref[0]+(-0.89437375241)*x_ref[1]**o+(-0.830119585842)*x_ref[1]
        ref[2,1,2,0]=(0.769898486257)*x_ref[0]**o+(0.791725624832)*x_ref[0]+(0.297519329396)*x_ref[1]**o+(-0.959462006611)*x_ref[1]
        ref[2,1,2,1]=(0.817802144251)*x_ref[0]**o+(0.69851258911)*x_ref[0]+(-0.174401735228)*x_ref[1]**o+(0.13449696498)*x_ref[1]
        ref[2,2,0,0]=(0.547972662256)*x_ref[0]**o+(-0.66491929979)*x_ref[0]+(0.330783041672)*x_ref[1]**o+(0.816196730606)*x_ref[1]
        ref[2,2,0,1]=(0.245971097763)*x_ref[0]**o+(0.213087481481)*x_ref[0]+(-0.840766291392)*x_ref[1]**o+(0.388414557376)*x_ref[1]
        ref[2,2,1,0]=(0.523151676444)*x_ref[0]**o+(0.854817486027)*x_ref[0]+(-0.53570188528)*x_ref[1]**o+(-0.68149362242)*x_ref[1]
        ref[2,2,1,1]=(-0.984200247787)*x_ref[0]**o+(0.83324365092)*x_ref[0]+(-0.896184408816)*x_ref[1]**o+(0.102276243487)*x_ref[1]
        ref[2,2,2,0]=(-0.216357164308)*x_ref[0]**o+(-0.597984882107)*x_ref[0]+(-0.205435921511)*x_ref[1]**o+(0.484947652759)*x_ref[1]
        ref[2,2,2,1]=(-0.650046253868)*x_ref[0]**o+(-0.624878186105)*x_ref[0]+(0.786519914406)*x_ref[1]**o+(0.28027273095)*x_ref[1]
        ref[2,3,0,0]=(-0.0821718346579)*x_ref[0]**o+(-0.687049917467)*x_ref[0]+(0.380792333736)*x_ref[1]**o+(0.151185837715)*x_ref[1]
        ref[2,3,0,1]=(0.61796113854)*x_ref[0]**o+(0.844624413065)*x_ref[0]+(-0.980619401098)*x_ref[1]**o+(-0.854924862641)*x_ref[1]
        ref[2,3,1,0]=(0.812293906883)*x_ref[0]**o+(-0.225593208768)*x_ref[0]+(0.964175374163)*x_ref[1]**o+(0.848850064926)*x_ref[1]
        ref[2,3,1,1]=(0.376453572937)*x_ref[0]**o+(0.994818812059)*x_ref[0]+(0.58872018303)*x_ref[1]**o+(0.68332473361)*x_ref[1]
        ref[2,3,2,0]=(-0.00948421972234)*x_ref[0]**o+(-0.680685462738)*x_ref[0]+(-0.767076278167)*x_ref[1]**o+(0.210165437284)*x_ref[1]
        ref[2,3,2,1]=(-0.455199698263)*x_ref[0]**o+(-0.763430401419)*x_ref[0]+(0.183773350212)*x_ref[1]**o+(0.43105940923)*x_ref[1]
        ref[2,4,0,0]=(-0.673962025988)*x_ref[0]**o+(-0.679073069725)*x_ref[0]+(0.11018431715)*x_ref[1]**o+(0.349020527482)*x_ref[1]
        ref[2,4,0,1]=(-0.876691282076)*x_ref[0]**o+(0.920902226631)*x_ref[0]+(-0.477417470246)*x_ref[1]**o+(0.836417982216)*x_ref[1]
        ref[2,4,1,0]=(0.567966851024)*x_ref[0]**o+(0.583241431716)*x_ref[0]+(-0.271851271008)*x_ref[1]**o+(0.678727902863)*x_ref[1]
        ref[2,4,1,1]=(0.707400094)*x_ref[0]**o+(0.802820856444)*x_ref[0]+(0.84518937616)*x_ref[1]**o+(-0.0267193533403)*x_ref[1]
        ref[2,4,2,0]=(0.213268920036)*x_ref[0]**o+(0.780721415328)*x_ref[0]+(0.839292948154)*x_ref[1]**o+(-0.0376334744443)*x_ref[1]
        ref[2,4,2,1]=(0.136255687031)*x_ref[0]**o+(-0.582766278693)*x_ref[0]+(0.257662521321)*x_ref[1]**o+(0.808337521979)*x_ref[1]
        ref[3,0,0,0]=(-0.524068770822)*x_ref[0]**o+(0.942236878842)*x_ref[0]+(0.889496255793)*x_ref[1]**o+(-0.904810742393)*x_ref[1]
        ref[3,0,0,1]=(0.780157465475)*x_ref[0]**o+(-0.763811668832)*x_ref[0]+(-0.517482070639)*x_ref[1]**o+(0.662770780676)*x_ref[1]
        ref[3,0,1,0]=(-0.693227564812)*x_ref[0]**o+(-0.763506211841)*x_ref[0]+(-0.427709881126)*x_ref[1]**o+(0.159109827204)*x_ref[1]
        ref[3,0,1,1]=(0.644975629084)*x_ref[0]**o+(-0.928112761899)*x_ref[0]+(-0.530772740083)*x_ref[1]**o+(-0.723174790087)*x_ref[1]
        ref[3,0,2,0]=(-0.923467274097)*x_ref[0]**o+(0.806128117288)*x_ref[0]+(0.177779363207)*x_ref[1]**o+(0.862243885542)*x_ref[1]
        ref[3,0,2,1]=(-0.0823252901358)*x_ref[0]**o+(0.289022916284)*x_ref[0]+(-0.518508771021)*x_ref[1]**o+(0.174564769845)*x_ref[1]
        ref[3,1,0,0]=(0.608837578444)*x_ref[0]**o+(0.898216758361)*x_ref[0]+(-0.844730109761)*x_ref[1]**o+(0.745974173207)*x_ref[1]
        ref[3,1,0,1]=(-0.11776411078)*x_ref[0]**o+(0.186256502356)*x_ref[0]+(0.565289400333)*x_ref[1]**o+(-0.776173716206)*x_ref[1]
        ref[3,1,1,0]=(-0.170736580152)*x_ref[0]**o+(0.829581661683)*x_ref[0]+(0.310135833859)*x_ref[1]**o+(-0.509629545687)*x_ref[1]
        ref[3,1,1,1]=(-0.225246122751)*x_ref[0]**o+(0.586754377294)*x_ref[0]+(-0.726064725635)*x_ref[1]**o+(0.136952605704)*x_ref[1]
        ref[3,1,2,0]=(-0.287932115478)*x_ref[0]**o+(0.858031396153)*x_ref[0]+(-0.493441745353)*x_ref[1]**o+(0.601827664994)*x_ref[1]
        ref[3,1,2,1]=(0.665503842862)*x_ref[0]**o+(-0.515998970452)*x_ref[0]+(-0.890055558624)*x_ref[1]**o+(-0.0446371451457)*x_ref[1]
        ref[3,2,0,0]=(-0.33741557585)*x_ref[0]**o+(-0.416183752333)*x_ref[0]+(-0.92430073538)*x_ref[1]**o+(-0.486265090904)*x_ref[1]
        ref[3,2,0,1]=(0.910860925994)*x_ref[0]**o+(-0.896869755389)*x_ref[0]+(0.0634658185324)*x_ref[1]**o+(-0.383136455739)*x_ref[1]
        ref[3,2,1,0]=(0.386839297976)*x_ref[0]**o+(-0.190659927071)*x_ref[0]+(-0.304846978185)*x_ref[1]**o+(-0.92705511899)*x_ref[1]
        ref[3,2,1,1]=(-0.0197812018683)*x_ref[0]**o+(-0.479083095775)*x_ref[0]+(0.502489976089)*x_ref[1]**o+(-0.859440665475)*x_ref[1]
        ref[3,2,2,0]=(-0.507672045299)*x_ref[0]**o+(0.498689676072)*x_ref[0]+(-0.719735804645)*x_ref[1]**o+(-0.522053306674)*x_ref[1]
        ref[3,2,2,1]=(0.46207629211)*x_ref[0]**o+(-0.496697066039)*x_ref[0]+(0.994069339904)*x_ref[1]**o+(0.873222754449)*x_ref[1]
        ref[3,3,0,0]=(0.126410029692)*x_ref[0]**o+(0.51067050196)*x_ref[0]+(-0.978384824088)*x_ref[1]**o+(-0.68606930503)*x_ref[1]
        ref[3,3,0,1]=(0.376888157008)*x_ref[0]**o+(-0.14015764025)*x_ref[0]+(0.523667700758)*x_ref[1]**o+(0.0277726855258)*x_ref[1]
        ref[3,3,1,0]=(-0.105865233294)*x_ref[0]**o+(-0.0515230741203)*x_ref[0]+(0.203766525036)*x_ref[1]**o+(-0.0431893014387)*x_ref[1]
        ref[3,3,1,1]=(-0.404351448486)*x_ref[0]**o+(-0.639161429157)*x_ref[0]+(-0.957341091708)*x_ref[1]**o+(-0.103346466984)*x_ref[1]
        ref[3,3,2,0]=(-0.0523508780623)*x_ref[0]**o+(-0.563589575544)*x_ref[0]+(0.158487727325)*x_ref[1]**o+(-0.519455492178)*x_ref[1]
        ref[3,3,2,1]=(-0.823286342218)*x_ref[0]**o+(-0.493663499535)*x_ref[0]+(-0.409825574064)*x_ref[1]**o+(-0.0406101279806)*x_ref[1]
        ref[3,4,0,0]=(0.926212910387)*x_ref[0]**o+(-0.0944507963)*x_ref[0]+(0.640740499391)*x_ref[1]**o+(-0.106969076047)*x_ref[1]
        ref[3,4,0,1]=(-0.912044993941)*x_ref[0]**o+(-0.495323777918)*x_ref[0]+(-0.595831845108)*x_ref[1]**o+(-0.313723426318)*x_ref[1]
        ref[3,4,1,0]=(-0.0730898742548)*x_ref[0]**o+(0.0524008484492)*x_ref[0]+(-0.948232935497)*x_ref[1]**o+(0.952670824572)*x_ref[1]
        ref[3,4,1,1]=(-0.81869438138)*x_ref[0]**o+(-0.894654266952)*x_ref[0]+(0.414078621872)*x_ref[1]**o+(-0.876690817847)*x_ref[1]
        ref[3,4,2,0]=(0.00688155078441)*x_ref[0]**o+(0.845964586551)*x_ref[0]+(0.0183149371892)*x_ref[1]**o+(-0.882931394537)*x_ref[1]
        ref[3,4,2,1]=(-0.734830097149)*x_ref[0]**o+(-0.37824588859)*x_ref[0]+(-0.815402620311)*x_ref[1]**o+(-0.357905685517)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.110519315354)*x[0]**o+(0.904227374292)*x[0]+(0.459005710707)*x[1]**o+(-0.980231422256)*x[1]+(-0.819713355475)*x[2]**o+(-0.93459517989)*x[2]
        arg[0,0,0,1]=(-0.326902946413)*x[0]**o+(0.489327766219)*x[0]+(-0.965308741061)*x[1]**o+(0.0351399111471)*x[1]+(0.257760096595)*x[2]**o+(0.125123023244)*x[2]
        arg[0,0,1,0]=(-0.584360710004)*x[0]**o+(0.681892732544)*x[0]+(-0.118257741614)*x[1]**o+(0.839875738892)*x[1]+(0.107934154241)*x[2]**o+(0.279757054665)*x[2]
        arg[0,0,1,1]=(0.90973214794)*x[0]**o+(-0.776804618826)*x[0]+(-0.910633853381)*x[1]**o+(-0.25678788529)*x[1]+(0.783709363646)*x[2]**o+(-0.922660855915)*x[2]
        arg[0,0,2,0]=(-0.802273263394)*x[0]**o+(-0.917749665952)*x[0]+(0.796206817226)*x[1]**o+(0.917284637885)*x[1]+(0.219841508893)*x[2]**o+(0.0655848200083)*x[2]
        arg[0,0,2,1]=(0.550596792343)*x[0]**o+(-0.760207057617)*x[0]+(0.0368469683098)*x[1]**o+(0.509347717331)*x[1]+(0.813780778487)*x[2]**o+(-0.432656298451)*x[2]
        arg[0,1,0,0]=(-0.238502042941)*x[0]**o+(-0.966632221032)*x[0]+(-0.373745652933)*x[1]**o+(-0.436176336389)*x[1]+(0.124280197699)*x[2]**o+(0.552700562803)*x[2]
        arg[0,1,0,1]=(-0.19848053276)*x[0]**o+(0.110167634957)*x[0]+(-0.748553857473)*x[1]**o+(0.355932508959)*x[1]+(-0.226987914753)*x[2]**o+(-0.68870238197)*x[2]
        arg[0,1,1,0]=(0.365822518882)*x[0]**o+(0.327285102897)*x[0]+(0.549212504635)*x[1]**o+(0.918179946504)*x[1]+(0.20644004679)*x[2]**o+(0.821299409756)*x[2]
        arg[0,1,1,1]=(-0.366167489101)*x[0]**o+(0.529933186572)*x[0]+(-0.604965310967)*x[1]**o+(-0.833093092352)*x[1]+(-0.73593913912)*x[2]**o+(0.872325614453)*x[2]
        arg[0,1,2,0]=(0.289123646902)*x[0]**o+(0.705862153789)*x[0]+(-0.319756886822)*x[1]**o+(-0.453220232524)*x[1]+(-0.192336374167)*x[2]**o+(0.363575495006)*x[2]
        arg[0,1,2,1]=(-0.213014833353)*x[0]**o+(-0.0706784230627)*x[0]+(-0.941272982229)*x[1]**o+(-0.96641361036)*x[1]+(0.289919014964)*x[2]**o+(0.57658772066)*x[2]
        arg[0,2,0,0]=(-0.406683991704)*x[0]**o+(-0.0110042213592)*x[0]+(-0.273148083258)*x[1]**o+(0.054257030447)*x[1]+(0.387598147681)*x[2]**o+(0.480398109069)*x[2]
        arg[0,2,0,1]=(0.318713057562)*x[0]**o+(-0.449467749585)*x[0]+(0.669991267829)*x[1]**o+(0.237986860249)*x[1]+(0.894370523471)*x[2]**o+(-0.397832436233)*x[2]
        arg[0,2,1,0]=(-0.858603253642)*x[0]**o+(-0.713122015919)*x[0]+(0.870251906188)*x[1]**o+(-0.781601254578)*x[1]+(-0.504531525822)*x[2]**o+(-0.699819349556)*x[2]
        arg[0,2,1,1]=(-0.825591116361)*x[0]**o+(-0.552452876394)*x[0]+(-0.262627261618)*x[1]**o+(-0.147127256184)*x[1]+(-0.171485665103)*x[2]**o+(0.823279741501)*x[2]
        arg[0,2,2,0]=(-0.0416279689068)*x[0]**o+(-0.0903338000001)*x[0]+(0.911673780635)*x[1]**o+(0.86298673049)*x[1]+(-0.60251179847)*x[2]**o+(0.378303877746)*x[2]
        arg[0,2,2,1]=(0.261384106387)*x[0]**o+(0.0326416199811)*x[0]+(-0.426740102269)*x[1]**o+(-0.445401626921)*x[1]+(-0.388668254526)*x[2]**o+(-0.977942913284)*x[2]
        arg[0,3,0,0]=(-0.0195053813483)*x[0]**o+(-0.294061998664)*x[0]+(0.790461098426)*x[1]**o+(-0.506334268838)*x[1]+(0.394015541605)*x[2]**o+(-0.562248105432)*x[2]
        arg[0,3,0,1]=(0.390726516735)*x[0]**o+(-0.757597437859)*x[0]+(0.861882149861)*x[1]**o+(0.621870451314)*x[1]+(0.900777666162)*x[2]**o+(0.271507131938)*x[2]
        arg[0,3,1,0]=(-0.725635966475)*x[0]**o+(-0.695837548737)*x[0]+(-0.163078176328)*x[1]**o+(-0.0435808418729)*x[1]+(0.0890756544611)*x[2]**o+(-0.270422627949)*x[2]
        arg[0,3,1,1]=(-0.175498720647)*x[0]**o+(-0.213605441632)*x[0]+(-0.358632620362)*x[1]**o+(-0.193273004152)*x[1]+(0.937704823805)*x[2]**o+(0.861707832291)*x[2]
        arg[0,3,2,0]=(0.0976010265611)*x[0]**o+(0.995135925341)*x[0]+(0.755339287387)*x[1]**o+(-0.924722068193)*x[1]+(-0.480212181904)*x[2]**o+(-0.40237760439)*x[2]
        arg[0,3,2,1]=(-0.679215392952)*x[0]**o+(-0.00796096253297)*x[0]+(-0.964972599421)*x[1]**o+(-0.410415435656)*x[1]+(-0.292638067785)*x[2]**o+(-0.950383976561)*x[2]
        arg[0,4,0,0]=(0.245586594928)*x[0]**o+(-0.593420401896)*x[0]+(0.553450486534)*x[1]**o+(0.173814965689)*x[1]+(0.0719392610563)*x[2]**o+(0.676231437486)*x[2]
        arg[0,4,0,1]=(0.154382756425)*x[0]**o+(-0.0655740293432)*x[0]+(0.924689745142)*x[1]**o+(-0.246876531175)*x[1]+(-0.241460713543)*x[2]**o+(-0.489840108568)*x[2]
        arg[0,4,1,0]=(-0.482705840359)*x[0]**o+(-0.392960380114)*x[0]+(-0.503463823972)*x[1]**o+(0.317163204738)*x[1]+(0.775048877356)*x[2]**o+(0.599889732326)*x[2]
        arg[0,4,1,1]=(-0.122555861716)*x[0]**o+(0.930175770449)*x[0]+(0.483503764941)*x[1]**o+(0.340001033217)*x[1]+(0.672924280979)*x[2]**o+(0.746489359683)*x[2]
        arg[0,4,2,0]=(-0.339421141033)*x[0]**o+(-0.604966575169)*x[0]+(0.862413017648)*x[1]**o+(0.936752078454)*x[1]+(-0.0549628353285)*x[2]**o+(-0.0858885835314)*x[2]
        arg[0,4,2,1]=(-0.581767357421)*x[0]**o+(-0.644286027548)*x[0]+(0.76948396865)*x[1]**o+(0.692265904744)*x[1]+(0.955871091531)*x[2]**o+(-0.378888503457)*x[2]
        arg[1,0,0,0]=(-0.0407846595046)*x[0]**o+(-0.293709943251)*x[0]+(-0.310552931582)*x[1]**o+(-0.696244178466)*x[1]+(0.962985284252)*x[2]**o+(-0.0266457621404)*x[2]
        arg[1,0,0,1]=(-0.480942811584)*x[0]**o+(-0.400158759679)*x[0]+(0.311171869807)*x[1]**o+(-0.79518809427)*x[1]+(-0.497004450478)*x[2]**o+(-0.639822005752)*x[2]
        arg[1,0,1,0]=(0.158738648483)*x[0]**o+(0.274859816468)*x[0]+(-0.208677612654)*x[1]**o+(0.911107598366)*x[1]+(-0.400163007778)*x[2]**o+(0.276296890267)*x[2]
        arg[1,0,1,1]=(0.613312276864)*x[0]**o+(-0.644237218776)*x[0]+(0.81927272895)*x[1]**o+(-0.12971860556)*x[1]+(0.214379406799)*x[2]**o+(0.856335675043)*x[2]
        arg[1,0,2,0]=(-0.699925173766)*x[0]**o+(-0.560613547952)*x[0]+(-0.355325107637)*x[1]**o+(-0.598747822041)*x[1]+(0.243087037553)*x[2]**o+(-0.484176993015)*x[2]
        arg[1,0,2,1]=(-0.218912300848)*x[0]**o+(-0.259483312969)*x[0]+(-0.863788186295)*x[1]**o+(0.924423774757)*x[1]+(-0.0823752192204)*x[2]**o+(-0.928377900602)*x[2]
        arg[1,1,0,0]=(-0.368901144425)*x[0]**o+(0.544104655592)*x[0]+(0.321295151176)*x[1]**o+(0.395692754134)*x[1]+(-0.721319111754)*x[2]**o+(-0.698518442137)*x[2]
        arg[1,1,0,1]=(-0.592867336047)*x[0]**o+(0.145120231038)*x[0]+(0.110390771366)*x[1]**o+(0.168333368086)*x[1]+(-0.577866419921)*x[2]**o+(0.705027708469)*x[2]
        arg[1,1,1,0]=(-0.210603428235)*x[0]**o+(-0.250156447116)*x[0]+(-0.397803248976)*x[1]**o+(0.332163380867)*x[1]+(-0.250648216541)*x[2]**o+(-0.0693729055697)*x[2]
        arg[1,1,1,1]=(0.324927558096)*x[0]**o+(-0.262726116223)*x[0]+(-0.555439758488)*x[1]**o+(-0.865019248609)*x[1]+(-0.0971907425857)*x[2]**o+(-0.742359236711)*x[2]
        arg[1,1,2,0]=(0.161058869414)*x[0]**o+(0.349191500791)*x[0]+(0.468336473315)*x[1]**o+(-0.856364775995)*x[1]+(0.491736696689)*x[2]**o+(0.352984104173)*x[2]
        arg[1,1,2,1]=(-0.243275976508)*x[0]**o+(-0.785605388878)*x[0]+(-0.282478919296)*x[1]**o+(-0.728851522341)*x[1]+(0.375134848237)*x[2]**o+(-0.486200561256)*x[2]
        arg[1,2,0,0]=(-0.172241451405)*x[0]**o+(0.422361498196)*x[0]+(0.464862755164)*x[1]**o+(0.466661717105)*x[1]+(-0.886209677512)*x[2]**o+(-0.43529076789)*x[2]
        arg[1,2,0,1]=(0.946582222405)*x[0]**o+(0.513640973894)*x[0]+(-0.923788000633)*x[1]**o+(-0.358163567195)*x[1]+(-0.709999132129)*x[2]**o+(-0.761163299694)*x[2]
        arg[1,2,1,0]=(0.236339981889)*x[0]**o+(0.259874866134)*x[0]+(-0.32460762691)*x[1]**o+(-0.629786950313)*x[1]+(-0.679261685402)*x[2]**o+(0.593069192918)*x[2]
        arg[1,2,1,1]=(0.11006800009)*x[0]**o+(0.970532991787)*x[0]+(-0.467608188361)*x[1]**o+(-0.959281984552)*x[1]+(0.610396622519)*x[2]**o+(0.282468600058)*x[2]
        arg[1,2,2,0]=(-0.656419157742)*x[0]**o+(0.234027308171)*x[0]+(-0.67027206978)*x[1]**o+(-0.746034436817)*x[1]+(-0.771894073652)*x[2]**o+(0.949545830287)*x[2]
        arg[1,2,2,1]=(0.615884236811)*x[0]**o+(-0.670156612555)*x[0]+(0.543165493345)*x[1]**o+(-0.618294277594)*x[1]+(-0.974114255061)*x[2]**o+(-0.762320542298)*x[2]
        arg[1,3,0,0]=(0.096201957648)*x[0]**o+(0.440870816208)*x[0]+(0.124448859607)*x[1]**o+(0.507919407561)*x[1]+(-0.401666263687)*x[2]**o+(-0.437653713723)*x[2]
        arg[1,3,0,1]=(0.165129229294)*x[0]**o+(-0.353217428955)*x[0]+(0.473666065504)*x[1]**o+(0.907667166988)*x[1]+(-0.864298533288)*x[2]**o+(0.641505117281)*x[2]
        arg[1,3,1,0]=(0.381900624425)*x[0]**o+(0.427520178597)*x[0]+(-0.560279540293)*x[1]**o+(0.190814532271)*x[1]+(-0.187947688629)*x[2]**o+(-0.127694148593)*x[2]
        arg[1,3,1,1]=(-0.903921363895)*x[0]**o+(-0.659942040828)*x[0]+(-0.549460213492)*x[1]**o+(0.869909651895)*x[1]+(-0.962273439484)*x[2]**o+(-0.046365857751)*x[2]
        arg[1,3,2,0]=(-0.021688482187)*x[0]**o+(-0.209531000717)*x[0]+(0.668564395133)*x[1]**o+(-0.0347978700975)*x[1]+(0.947677147488)*x[2]**o+(-0.0152385052475)*x[2]
        arg[1,3,2,1]=(0.880839485817)*x[0]**o+(-0.319275929312)*x[0]+(-0.940056657485)*x[1]**o+(0.570855397726)*x[1]+(0.279710812074)*x[2]**o+(0.496622215051)*x[2]
        arg[1,4,0,0]=(-0.139033645615)*x[0]**o+(-0.271453722366)*x[0]+(-0.268516720244)*x[1]**o+(-0.0467555017438)*x[1]+(-0.41618262265)*x[2]**o+(0.315320363813)*x[2]
        arg[1,4,0,1]=(0.296108607204)*x[0]**o+(-0.184582165661)*x[0]+(0.121416543744)*x[1]**o+(-0.922284926751)*x[1]+(-0.437600089668)*x[2]**o+(-0.608800626513)*x[2]
        arg[1,4,1,0]=(0.8942028423)*x[0]**o+(0.673402831188)*x[0]+(-0.622529195562)*x[1]**o+(-0.266052338798)*x[1]+(-0.273167851702)*x[2]**o+(-0.312211982747)*x[2]
        arg[1,4,1,1]=(-0.102315711178)*x[0]**o+(-0.885515719371)*x[0]+(-0.532910101204)*x[1]**o+(0.798525944348)*x[1]+(0.640586802303)*x[2]**o+(0.36308806536)*x[2]
        arg[1,4,2,0]=(0.376793506541)*x[0]**o+(-0.667063564875)*x[0]+(0.609591985816)*x[1]**o+(0.28347745142)*x[1]+(-0.751832936302)*x[2]**o+(-0.683738856965)*x[2]
        arg[1,4,2,1]=(-0.51412252819)*x[0]**o+(0.559560297206)*x[0]+(0.751134603429)*x[1]**o+(-0.0031812310797)*x[1]+(-0.525776122178)*x[2]**o+(0.338471942848)*x[2]
        arg[2,0,0,0]=(0.937572244905)*x[0]**o+(-0.293020870379)*x[0]+(-0.918591520817)*x[1]**o+(0.622194477261)*x[1]+(0.31641482851)*x[2]**o+(0.276912266864)*x[2]
        arg[2,0,0,1]=(0.853041599998)*x[0]**o+(-0.0220623724171)*x[0]+(0.89909044296)*x[1]**o+(-0.838782642732)*x[1]+(0.0133197592042)*x[2]**o+(-0.868977618993)*x[2]
        arg[2,0,1,0]=(-0.325799042976)*x[0]**o+(0.746269971994)*x[0]+(0.882671061757)*x[1]**o+(-0.284133373252)*x[1]+(0.120086549521)*x[2]**o+(-0.397352610969)*x[2]
        arg[2,0,1,1]=(0.282336934951)*x[0]**o+(-0.339893978547)*x[0]+(-0.0861660125386)*x[1]**o+(0.283063579798)*x[1]+(0.714762478569)*x[2]**o+(-0.497021718549)*x[2]
        arg[2,0,2,0]=(0.891782779491)*x[0]**o+(-0.817647577644)*x[0]+(0.364088142085)*x[1]**o+(-0.950577637048)*x[1]+(0.4070663557)*x[2]**o+(0.270753179246)*x[2]
        arg[2,0,2,1]=(0.316594913808)*x[0]**o+(0.310731486851)*x[0]+(0.250421828765)*x[1]**o+(0.772786141458)*x[1]+(0.64843040679)*x[2]**o+(-0.691903332511)*x[2]
        arg[2,1,0,0]=(0.87180715028)*x[0]**o+(-0.953173230375)*x[0]+(-0.823516344548)*x[1]**o+(0.934653756548)*x[1]+(-0.537124962489)*x[2]**o+(-0.233365939659)*x[2]
        arg[2,1,0,1]=(-0.407189744631)*x[0]**o+(-0.754425720288)*x[0]+(0.737526767897)*x[1]**o+(0.745393943502)*x[1]+(0.811744293685)*x[2]**o+(-0.459556659177)*x[2]
        arg[2,1,1,0]=(-0.693529259877)*x[0]**o+(0.580197938048)*x[0]+(0.506055440405)*x[1]**o+(0.122531014759)*x[1]+(0.427736137775)*x[2]**o+(-0.83260140401)*x[2]
        arg[2,1,1,1]=(0.542621975026)*x[0]**o+(-0.568425713106)*x[0]+(0.303002043446)*x[1]**o+(0.363286261651)*x[1]+(0.2729476289)*x[2]**o+(-0.0113629744659)*x[2]
        arg[2,1,2,0]=(-0.266750235694)*x[0]**o+(0.10116400306)*x[0]+(0.315977962897)*x[1]**o+(-0.406236201454)*x[1]+(-0.586845072374)*x[2]**o+(0.513937047282)*x[2]
        arg[2,1,2,1]=(-0.518780811975)*x[0]**o+(0.275053640829)*x[0]+(0.4234970646)*x[1]**o+(0.389585163376)*x[1]+(-0.609771068778)*x[2]**o+(-0.121467061755)*x[2]
        arg[2,2,0,0]=(0.138823207466)*x[0]**o+(-0.779837455014)*x[0]+(0.70167049555)*x[1]**o+(0.766428813369)*x[1]+(0.114023530366)*x[2]**o+(0.650098536187)*x[2]
        arg[2,2,0,1]=(0.974998667633)*x[0]**o+(0.958471050654)*x[0]+(0.193515162503)*x[1]**o+(-0.252148167253)*x[1]+(0.495475460125)*x[2]**o+(0.562693347983)*x[2]
        arg[2,2,1,0]=(-0.552379548988)*x[0]**o+(-0.831399971618)*x[0]+(0.437581216647)*x[1]**o+(-0.775331267696)*x[1]+(0.729041628844)*x[2]**o+(0.633113359893)*x[2]
        arg[2,2,1,1]=(0.379858437049)*x[0]**o+(0.923586504203)*x[0]+(0.707244783969)*x[1]**o+(-0.293157375059)*x[1]+(-0.368717941876)*x[2]**o+(-0.202610288616)*x[2]
        arg[2,2,2,0]=(-0.804826108231)*x[0]**o+(-0.0208712096312)*x[0]+(0.0482547072498)*x[1]**o+(0.852366543609)*x[1]+(0.328962500136)*x[2]**o+(-0.132777835512)*x[2]
        arg[2,2,2,1]=(0.91373752217)*x[0]**o+(-0.792112120729)*x[0]+(-0.0457564087072)*x[1]**o+(0.309694852949)*x[1]+(0.089219285337)*x[2]**o+(-0.516000514614)*x[2]
        arg[2,3,0,0]=(0.343215129115)*x[0]**o+(0.40804099837)*x[0]+(-0.865942442914)*x[1]**o+(0.842241683043)*x[1]+(0.0727803297601)*x[2]**o+(-0.395225682949)*x[2]
        arg[2,3,0,1]=(-0.965340522771)*x[0]**o+(-0.897056478677)*x[0]+(-0.305205944062)*x[1]**o+(0.393941789685)*x[1]+(-0.253328379131)*x[2]**o+(0.0951510469271)*x[2]
        arg[2,3,1,0]=(0.536251514752)*x[0]**o+(0.0629776386779)*x[0]+(-0.706311222379)*x[1]**o+(-0.84362242029)*x[1]+(0.00110678760539)*x[2]**o+(-0.30286059941)*x[2]
        arg[2,3,1,1]=(0.114584737295)*x[0]**o+(0.0708763738067)*x[0]+(-0.454760062397)*x[1]**o+(0.228742539922)*x[1]+(0.514960810352)*x[2]**o+(-0.815749211085)*x[2]
        arg[2,3,2,0]=(0.246335688008)*x[0]**o+(-0.239199299885)*x[0]+(-0.447390151659)*x[1]**o+(0.918949356391)*x[1]+(-0.307864840175)*x[2]**o+(0.560139080481)*x[2]
        arg[2,3,2,1]=(0.118766035723)*x[0]**o+(0.610711164899)*x[0]+(0.336628771776)*x[1]**o+(0.833699268751)*x[1]+(-0.424694647429)*x[2]**o+(-0.639929688104)*x[2]
        arg[2,4,0,0]=(0.814543506987)*x[0]**o+(-0.797684321368)*x[0]+(0.337242635783)*x[1]**o+(0.278139837072)*x[1]+(0.633511526441)*x[2]**o+(0.372175615028)*x[2]
        arg[2,4,0,1]=(-0.0590668694094)*x[0]**o+(0.256732067267)*x[0]+(0.519737187577)*x[1]**o+(-0.376391855948)*x[1]+(-0.810934658651)*x[2]**o+(0.826899476928)*x[2]
        arg[2,4,1,0]=(-0.428976843661)*x[0]**o+(0.672667833313)*x[0]+(0.263194890946)*x[1]**o+(-0.894790773661)*x[1]+(-0.550298557295)*x[2]**o+(-0.00560384035502)*x[2]
        arg[2,4,1,1]=(-0.818372430144)*x[0]**o+(-0.829920833073)*x[0]+(-0.359560636119)*x[1]**o+(-0.113484563205)*x[1]+(0.586796110365)*x[2]**o+(0.296820898426)*x[2]
        arg[2,4,2,0]=(-0.691597969566)*x[0]**o+(0.715813153538)*x[0]+(0.954093612503)*x[1]**o+(0.88803420754)*x[1]+(0.697447667765)*x[2]**o+(0.793730496625)*x[2]
        arg[2,4,2,1]=(0.351099833336)*x[0]**o+(-0.716419272807)*x[0]+(0.318726674998)*x[1]**o+(0.246151914414)*x[1]+(-0.945520673884)*x[2]**o+(-0.489026233954)*x[2]
        arg[3,0,0,0]=(0.511598022804)*x[0]**o+(0.567617269582)*x[0]+(-0.526293319137)*x[1]**o+(0.333750098146)*x[1]+(-0.0869228849144)*x[2]**o+(-0.204072834867)*x[2]
        arg[3,0,0,1]=(0.905384007172)*x[0]**o+(-0.661067390314)*x[0]+(-0.470790632295)*x[1]**o+(-0.149608017616)*x[1]+(0.427879278267)*x[2]**o+(0.987040179184)*x[2]
        arg[3,0,1,0]=(-0.946562006809)*x[0]**o+(0.64141909633)*x[0]+(-0.762076826276)*x[1]**o+(-0.642912850794)*x[1]+(0.712567898559)*x[2]**o+(-0.292128944205)*x[2]
        arg[3,0,1,1]=(-0.955087949361)*x[0]**o+(0.426588041133)*x[0]+(0.339313104127)*x[1]**o+(-0.375147344045)*x[1]+(0.967259014679)*x[2]**o+(0.504521081858)*x[2]
        arg[3,0,2,0]=(0.141876598344)*x[0]**o+(-0.335579252825)*x[0]+(0.514552993602)*x[1]**o+(-0.531116948499)*x[1]+(-0.369319413141)*x[2]**o+(0.0374413262362)*x[2]
        arg[3,0,2,1]=(0.678360463658)*x[0]**o+(-0.357900119646)*x[0]+(0.300750717059)*x[1]**o+(0.516196189448)*x[1]+(0.0535466018203)*x[2]**o+(-0.149961940786)*x[2]
        arg[3,1,0,0]=(0.0975994351079)*x[0]**o+(0.634544068499)*x[0]+(-0.963217046352)*x[1]**o+(-0.804658091456)*x[1]+(-0.620132720823)*x[2]**o+(0.862250619369)*x[2]
        arg[3,1,0,1]=(0.210306328879)*x[0]**o+(-0.457341187966)*x[0]+(0.957556084801)*x[1]**o+(0.704073170928)*x[1]+(0.423134492133)*x[2]**o+(-0.167268375412)*x[2]
        arg[3,1,1,0]=(-0.909228817744)*x[0]**o+(0.110665018921)*x[0]+(0.36275630147)*x[1]**o+(0.346866824763)*x[1]+(-0.535600547489)*x[2]**o+(0.627638814375)*x[2]
        arg[3,1,1,1]=(0.328780933823)*x[0]**o+(-0.249563994161)*x[0]+(-0.555789062892)*x[1]**o+(-0.917287525036)*x[1]+(0.386191175017)*x[2]**o+(0.123904319061)*x[2]
        arg[3,1,2,0]=(-0.838409901623)*x[0]**o+(0.673479705393)*x[0]+(0.482921104917)*x[1]**o+(0.136411388588)*x[1]+(0.583674116225)*x[2]**o+(0.305100836877)*x[2]
        arg[3,1,2,1]=(0.952961008312)*x[0]**o+(-0.0501981079532)*x[0]+(0.661434464642)*x[1]**o+(0.65030469399)*x[1]+(-0.157562001295)*x[2]**o+(-0.249551892917)*x[2]
        arg[3,2,0,0]=(-0.0232804500286)*x[0]**o+(-0.559229392203)*x[0]+(0.127967930273)*x[1]**o+(-0.635827181645)*x[1]+(-0.0426703224383)*x[2]**o+(0.47720230258)*x[2]
        arg[3,2,0,1]=(-0.421658996106)*x[0]**o+(-0.468260038077)*x[0]+(0.479083572554)*x[1]**o+(-0.214417881841)*x[1]+(0.372430534261)*x[2]**o+(0.133063789546)*x[2]
        arg[3,2,1,0]=(0.905811586567)*x[0]**o+(0.245447340384)*x[0]+(-0.337650999146)*x[1]**o+(0.908188235976)*x[1]+(-0.804029751277)*x[2]**o+(0.0512676456643)*x[2]
        arg[3,2,1,1]=(-0.891486839409)*x[0]**o+(0.20753794329)*x[0]+(0.879183464252)*x[1]**o+(0.902028906992)*x[1]+(0.416776813598)*x[2]**o+(0.491964374373)*x[2]
        arg[3,2,2,0]=(0.653067027119)*x[0]**o+(0.765293890745)*x[0]+(-0.348902169508)*x[1]**o+(0.324429035849)*x[1]+(-0.363026365065)*x[2]**o+(0.0816186405511)*x[2]
        arg[3,2,2,1]=(-0.0415375052325)*x[0]**o+(-0.480779075784)*x[0]+(-0.137223552938)*x[1]**o+(-0.915733625512)*x[1]+(0.750199445159)*x[2]**o+(-0.789818945171)*x[2]
        arg[3,3,0,0]=(0.0744921239083)*x[0]**o+(0.0700589144159)*x[0]+(-0.327105425297)*x[1]**o+(0.469843257765)*x[1]+(-0.170191600844)*x[2]**o+(0.732071534155)*x[2]
        arg[3,3,0,1]=(0.172643927441)*x[0]**o+(0.848924114383)*x[0]+(-0.79125762228)*x[1]**o+(-0.694725775976)*x[1]+(-0.309710503985)*x[2]**o+(-0.405310234927)*x[2]
        arg[3,3,1,0]=(-0.465911787595)*x[0]**o+(-0.71080474709)*x[0]+(0.959925934203)*x[1]**o+(-0.111110644825)*x[1]+(-0.145349925218)*x[2]**o+(0.122955165811)*x[2]
        arg[3,3,1,1]=(-0.769111863754)*x[0]**o+(0.348036938532)*x[0]+(0.485890856474)*x[1]**o+(0.382595218985)*x[1]+(-0.897488692981)*x[2]**o+(0.830822333819)*x[2]
        arg[3,3,2,0]=(0.341890793681)*x[0]**o+(-0.667145936571)*x[0]+(-0.145094108268)*x[1]**o+(0.771433911212)*x[1]+(-0.299752164325)*x[2]**o+(-0.357856549677)*x[2]
        arg[3,3,2,1]=(0.26381838796)*x[0]**o+(-0.44474724998)*x[0]+(0.310283988195)*x[1]**o+(0.698675181314)*x[1]+(0.437624133238)*x[2]**o+(0.394916054445)*x[2]
        arg[3,4,0,0]=(-0.576499406674)*x[0]**o+(0.770936634959)*x[0]+(0.707441735783)*x[1]**o+(-0.794953257043)*x[1]+(-0.535342943209)*x[2]**o+(-0.509446811787)*x[2]
        arg[3,4,0,1]=(0.161115012113)*x[0]**o+(0.11079092252)*x[0]+(0.0633804181164)*x[1]**o+(-0.406049293019)*x[1]+(-0.824283994336)*x[2]**o+(-0.481107560499)*x[2]
        arg[3,4,1,0]=(-0.411133810295)*x[0]**o+(0.988486625099)*x[0]+(-0.36050191367)*x[1]**o+(0.469479821965)*x[1]+(-0.431175357325)*x[2]**o+(-0.966295015263)*x[2]
        arg[3,4,1,1]=(0.233731410367)*x[0]**o+(-0.821900740766)*x[0]+(-0.823123461385)*x[1]**o+(-0.380537914421)*x[1]+(-0.896146300355)*x[2]**o+(-0.775172990257)*x[2]
        arg[3,4,2,0]=(-0.96323363586)*x[0]**o+(0.0935412405245)*x[0]+(0.929333026259)*x[1]**o+(-0.322236627348)*x[1]+(-0.390898135964)*x[2]**o+(-0.766178290162)*x[2]
        arg[3,4,2,1]=(-0.773670533268)*x[0]**o+(-0.156975157098)*x[0]+(-0.91512996795)*x[1]**o+(0.0247803682126)*x[1]+(0.995159733782)*x[2]**o+(-0.669224144225)*x[2]
        ref[0,0,0,0]=(-0.110519315354)*x_ref[0]**o+(0.904227374292)*x_ref[0]+(0.459005710707)*x_ref[1]**o+(-0.980231422256)*x_ref[1]+(-0.819713355475)*x_ref[2]**o+(-0.93459517989)*x_ref[2]
        ref[0,0,0,1]=(-0.326902946413)*x_ref[0]**o+(0.489327766219)*x_ref[0]+(-0.965308741061)*x_ref[1]**o+(0.0351399111471)*x_ref[1]+(0.257760096595)*x_ref[2]**o+(0.125123023244)*x_ref[2]
        ref[0,0,1,0]=(-0.584360710004)*x_ref[0]**o+(0.681892732544)*x_ref[0]+(-0.118257741614)*x_ref[1]**o+(0.839875738892)*x_ref[1]+(0.107934154241)*x_ref[2]**o+(0.279757054665)*x_ref[2]
        ref[0,0,1,1]=(0.90973214794)*x_ref[0]**o+(-0.776804618826)*x_ref[0]+(-0.910633853381)*x_ref[1]**o+(-0.25678788529)*x_ref[1]+(0.783709363646)*x_ref[2]**o+(-0.922660855915)*x_ref[2]
        ref[0,0,2,0]=(-0.802273263394)*x_ref[0]**o+(-0.917749665952)*x_ref[0]+(0.796206817226)*x_ref[1]**o+(0.917284637885)*x_ref[1]+(0.219841508893)*x_ref[2]**o+(0.0655848200083)*x_ref[2]
        ref[0,0,2,1]=(0.550596792343)*x_ref[0]**o+(-0.760207057617)*x_ref[0]+(0.0368469683098)*x_ref[1]**o+(0.509347717331)*x_ref[1]+(0.813780778487)*x_ref[2]**o+(-0.432656298451)*x_ref[2]
        ref[0,1,0,0]=(-0.238502042941)*x_ref[0]**o+(-0.966632221032)*x_ref[0]+(-0.373745652933)*x_ref[1]**o+(-0.436176336389)*x_ref[1]+(0.124280197699)*x_ref[2]**o+(0.552700562803)*x_ref[2]
        ref[0,1,0,1]=(-0.19848053276)*x_ref[0]**o+(0.110167634957)*x_ref[0]+(-0.748553857473)*x_ref[1]**o+(0.355932508959)*x_ref[1]+(-0.226987914753)*x_ref[2]**o+(-0.68870238197)*x_ref[2]
        ref[0,1,1,0]=(0.365822518882)*x_ref[0]**o+(0.327285102897)*x_ref[0]+(0.549212504635)*x_ref[1]**o+(0.918179946504)*x_ref[1]+(0.20644004679)*x_ref[2]**o+(0.821299409756)*x_ref[2]
        ref[0,1,1,1]=(-0.366167489101)*x_ref[0]**o+(0.529933186572)*x_ref[0]+(-0.604965310967)*x_ref[1]**o+(-0.833093092352)*x_ref[1]+(-0.73593913912)*x_ref[2]**o+(0.872325614453)*x_ref[2]
        ref[0,1,2,0]=(0.289123646902)*x_ref[0]**o+(0.705862153789)*x_ref[0]+(-0.319756886822)*x_ref[1]**o+(-0.453220232524)*x_ref[1]+(-0.192336374167)*x_ref[2]**o+(0.363575495006)*x_ref[2]
        ref[0,1,2,1]=(-0.213014833353)*x_ref[0]**o+(-0.0706784230627)*x_ref[0]+(-0.941272982229)*x_ref[1]**o+(-0.96641361036)*x_ref[1]+(0.289919014964)*x_ref[2]**o+(0.57658772066)*x_ref[2]
        ref[0,2,0,0]=(-0.406683991704)*x_ref[0]**o+(-0.0110042213592)*x_ref[0]+(-0.273148083258)*x_ref[1]**o+(0.054257030447)*x_ref[1]+(0.387598147681)*x_ref[2]**o+(0.480398109069)*x_ref[2]
        ref[0,2,0,1]=(0.318713057562)*x_ref[0]**o+(-0.449467749585)*x_ref[0]+(0.669991267829)*x_ref[1]**o+(0.237986860249)*x_ref[1]+(0.894370523471)*x_ref[2]**o+(-0.397832436233)*x_ref[2]
        ref[0,2,1,0]=(-0.858603253642)*x_ref[0]**o+(-0.713122015919)*x_ref[0]+(0.870251906188)*x_ref[1]**o+(-0.781601254578)*x_ref[1]+(-0.504531525822)*x_ref[2]**o+(-0.699819349556)*x_ref[2]
        ref[0,2,1,1]=(-0.825591116361)*x_ref[0]**o+(-0.552452876394)*x_ref[0]+(-0.262627261618)*x_ref[1]**o+(-0.147127256184)*x_ref[1]+(-0.171485665103)*x_ref[2]**o+(0.823279741501)*x_ref[2]
        ref[0,2,2,0]=(-0.0416279689068)*x_ref[0]**o+(-0.0903338000001)*x_ref[0]+(0.911673780635)*x_ref[1]**o+(0.86298673049)*x_ref[1]+(-0.60251179847)*x_ref[2]**o+(0.378303877746)*x_ref[2]
        ref[0,2,2,1]=(0.261384106387)*x_ref[0]**o+(0.0326416199811)*x_ref[0]+(-0.426740102269)*x_ref[1]**o+(-0.445401626921)*x_ref[1]+(-0.388668254526)*x_ref[2]**o+(-0.977942913284)*x_ref[2]
        ref[0,3,0,0]=(-0.0195053813483)*x_ref[0]**o+(-0.294061998664)*x_ref[0]+(0.790461098426)*x_ref[1]**o+(-0.506334268838)*x_ref[1]+(0.394015541605)*x_ref[2]**o+(-0.562248105432)*x_ref[2]
        ref[0,3,0,1]=(0.390726516735)*x_ref[0]**o+(-0.757597437859)*x_ref[0]+(0.861882149861)*x_ref[1]**o+(0.621870451314)*x_ref[1]+(0.900777666162)*x_ref[2]**o+(0.271507131938)*x_ref[2]
        ref[0,3,1,0]=(-0.725635966475)*x_ref[0]**o+(-0.695837548737)*x_ref[0]+(-0.163078176328)*x_ref[1]**o+(-0.0435808418729)*x_ref[1]+(0.0890756544611)*x_ref[2]**o+(-0.270422627949)*x_ref[2]
        ref[0,3,1,1]=(-0.175498720647)*x_ref[0]**o+(-0.213605441632)*x_ref[0]+(-0.358632620362)*x_ref[1]**o+(-0.193273004152)*x_ref[1]+(0.937704823805)*x_ref[2]**o+(0.861707832291)*x_ref[2]
        ref[0,3,2,0]=(0.0976010265611)*x_ref[0]**o+(0.995135925341)*x_ref[0]+(0.755339287387)*x_ref[1]**o+(-0.924722068193)*x_ref[1]+(-0.480212181904)*x_ref[2]**o+(-0.40237760439)*x_ref[2]
        ref[0,3,2,1]=(-0.679215392952)*x_ref[0]**o+(-0.00796096253297)*x_ref[0]+(-0.964972599421)*x_ref[1]**o+(-0.410415435656)*x_ref[1]+(-0.292638067785)*x_ref[2]**o+(-0.950383976561)*x_ref[2]
        ref[0,4,0,0]=(0.245586594928)*x_ref[0]**o+(-0.593420401896)*x_ref[0]+(0.553450486534)*x_ref[1]**o+(0.173814965689)*x_ref[1]+(0.0719392610563)*x_ref[2]**o+(0.676231437486)*x_ref[2]
        ref[0,4,0,1]=(0.154382756425)*x_ref[0]**o+(-0.0655740293432)*x_ref[0]+(0.924689745142)*x_ref[1]**o+(-0.246876531175)*x_ref[1]+(-0.241460713543)*x_ref[2]**o+(-0.489840108568)*x_ref[2]
        ref[0,4,1,0]=(-0.482705840359)*x_ref[0]**o+(-0.392960380114)*x_ref[0]+(-0.503463823972)*x_ref[1]**o+(0.317163204738)*x_ref[1]+(0.775048877356)*x_ref[2]**o+(0.599889732326)*x_ref[2]
        ref[0,4,1,1]=(-0.122555861716)*x_ref[0]**o+(0.930175770449)*x_ref[0]+(0.483503764941)*x_ref[1]**o+(0.340001033217)*x_ref[1]+(0.672924280979)*x_ref[2]**o+(0.746489359683)*x_ref[2]
        ref[0,4,2,0]=(-0.339421141033)*x_ref[0]**o+(-0.604966575169)*x_ref[0]+(0.862413017648)*x_ref[1]**o+(0.936752078454)*x_ref[1]+(-0.0549628353285)*x_ref[2]**o+(-0.0858885835314)*x_ref[2]
        ref[0,4,2,1]=(-0.581767357421)*x_ref[0]**o+(-0.644286027548)*x_ref[0]+(0.76948396865)*x_ref[1]**o+(0.692265904744)*x_ref[1]+(0.955871091531)*x_ref[2]**o+(-0.378888503457)*x_ref[2]
        ref[1,0,0,0]=(-0.0407846595046)*x_ref[0]**o+(-0.293709943251)*x_ref[0]+(-0.310552931582)*x_ref[1]**o+(-0.696244178466)*x_ref[1]+(0.962985284252)*x_ref[2]**o+(-0.0266457621404)*x_ref[2]
        ref[1,0,0,1]=(-0.480942811584)*x_ref[0]**o+(-0.400158759679)*x_ref[0]+(0.311171869807)*x_ref[1]**o+(-0.79518809427)*x_ref[1]+(-0.497004450478)*x_ref[2]**o+(-0.639822005752)*x_ref[2]
        ref[1,0,1,0]=(0.158738648483)*x_ref[0]**o+(0.274859816468)*x_ref[0]+(-0.208677612654)*x_ref[1]**o+(0.911107598366)*x_ref[1]+(-0.400163007778)*x_ref[2]**o+(0.276296890267)*x_ref[2]
        ref[1,0,1,1]=(0.613312276864)*x_ref[0]**o+(-0.644237218776)*x_ref[0]+(0.81927272895)*x_ref[1]**o+(-0.12971860556)*x_ref[1]+(0.214379406799)*x_ref[2]**o+(0.856335675043)*x_ref[2]
        ref[1,0,2,0]=(-0.699925173766)*x_ref[0]**o+(-0.560613547952)*x_ref[0]+(-0.355325107637)*x_ref[1]**o+(-0.598747822041)*x_ref[1]+(0.243087037553)*x_ref[2]**o+(-0.484176993015)*x_ref[2]
        ref[1,0,2,1]=(-0.218912300848)*x_ref[0]**o+(-0.259483312969)*x_ref[0]+(-0.863788186295)*x_ref[1]**o+(0.924423774757)*x_ref[1]+(-0.0823752192204)*x_ref[2]**o+(-0.928377900602)*x_ref[2]
        ref[1,1,0,0]=(-0.368901144425)*x_ref[0]**o+(0.544104655592)*x_ref[0]+(0.321295151176)*x_ref[1]**o+(0.395692754134)*x_ref[1]+(-0.721319111754)*x_ref[2]**o+(-0.698518442137)*x_ref[2]
        ref[1,1,0,1]=(-0.592867336047)*x_ref[0]**o+(0.145120231038)*x_ref[0]+(0.110390771366)*x_ref[1]**o+(0.168333368086)*x_ref[1]+(-0.577866419921)*x_ref[2]**o+(0.705027708469)*x_ref[2]
        ref[1,1,1,0]=(-0.210603428235)*x_ref[0]**o+(-0.250156447116)*x_ref[0]+(-0.397803248976)*x_ref[1]**o+(0.332163380867)*x_ref[1]+(-0.250648216541)*x_ref[2]**o+(-0.0693729055697)*x_ref[2]
        ref[1,1,1,1]=(0.324927558096)*x_ref[0]**o+(-0.262726116223)*x_ref[0]+(-0.555439758488)*x_ref[1]**o+(-0.865019248609)*x_ref[1]+(-0.0971907425857)*x_ref[2]**o+(-0.742359236711)*x_ref[2]
        ref[1,1,2,0]=(0.161058869414)*x_ref[0]**o+(0.349191500791)*x_ref[0]+(0.468336473315)*x_ref[1]**o+(-0.856364775995)*x_ref[1]+(0.491736696689)*x_ref[2]**o+(0.352984104173)*x_ref[2]
        ref[1,1,2,1]=(-0.243275976508)*x_ref[0]**o+(-0.785605388878)*x_ref[0]+(-0.282478919296)*x_ref[1]**o+(-0.728851522341)*x_ref[1]+(0.375134848237)*x_ref[2]**o+(-0.486200561256)*x_ref[2]
        ref[1,2,0,0]=(-0.172241451405)*x_ref[0]**o+(0.422361498196)*x_ref[0]+(0.464862755164)*x_ref[1]**o+(0.466661717105)*x_ref[1]+(-0.886209677512)*x_ref[2]**o+(-0.43529076789)*x_ref[2]
        ref[1,2,0,1]=(0.946582222405)*x_ref[0]**o+(0.513640973894)*x_ref[0]+(-0.923788000633)*x_ref[1]**o+(-0.358163567195)*x_ref[1]+(-0.709999132129)*x_ref[2]**o+(-0.761163299694)*x_ref[2]
        ref[1,2,1,0]=(0.236339981889)*x_ref[0]**o+(0.259874866134)*x_ref[0]+(-0.32460762691)*x_ref[1]**o+(-0.629786950313)*x_ref[1]+(-0.679261685402)*x_ref[2]**o+(0.593069192918)*x_ref[2]
        ref[1,2,1,1]=(0.11006800009)*x_ref[0]**o+(0.970532991787)*x_ref[0]+(-0.467608188361)*x_ref[1]**o+(-0.959281984552)*x_ref[1]+(0.610396622519)*x_ref[2]**o+(0.282468600058)*x_ref[2]
        ref[1,2,2,0]=(-0.656419157742)*x_ref[0]**o+(0.234027308171)*x_ref[0]+(-0.67027206978)*x_ref[1]**o+(-0.746034436817)*x_ref[1]+(-0.771894073652)*x_ref[2]**o+(0.949545830287)*x_ref[2]
        ref[1,2,2,1]=(0.615884236811)*x_ref[0]**o+(-0.670156612555)*x_ref[0]+(0.543165493345)*x_ref[1]**o+(-0.618294277594)*x_ref[1]+(-0.974114255061)*x_ref[2]**o+(-0.762320542298)*x_ref[2]
        ref[1,3,0,0]=(0.096201957648)*x_ref[0]**o+(0.440870816208)*x_ref[0]+(0.124448859607)*x_ref[1]**o+(0.507919407561)*x_ref[1]+(-0.401666263687)*x_ref[2]**o+(-0.437653713723)*x_ref[2]
        ref[1,3,0,1]=(0.165129229294)*x_ref[0]**o+(-0.353217428955)*x_ref[0]+(0.473666065504)*x_ref[1]**o+(0.907667166988)*x_ref[1]+(-0.864298533288)*x_ref[2]**o+(0.641505117281)*x_ref[2]
        ref[1,3,1,0]=(0.381900624425)*x_ref[0]**o+(0.427520178597)*x_ref[0]+(-0.560279540293)*x_ref[1]**o+(0.190814532271)*x_ref[1]+(-0.187947688629)*x_ref[2]**o+(-0.127694148593)*x_ref[2]
        ref[1,3,1,1]=(-0.903921363895)*x_ref[0]**o+(-0.659942040828)*x_ref[0]+(-0.549460213492)*x_ref[1]**o+(0.869909651895)*x_ref[1]+(-0.962273439484)*x_ref[2]**o+(-0.046365857751)*x_ref[2]
        ref[1,3,2,0]=(-0.021688482187)*x_ref[0]**o+(-0.209531000717)*x_ref[0]+(0.668564395133)*x_ref[1]**o+(-0.0347978700975)*x_ref[1]+(0.947677147488)*x_ref[2]**o+(-0.0152385052475)*x_ref[2]
        ref[1,3,2,1]=(0.880839485817)*x_ref[0]**o+(-0.319275929312)*x_ref[0]+(-0.940056657485)*x_ref[1]**o+(0.570855397726)*x_ref[1]+(0.279710812074)*x_ref[2]**o+(0.496622215051)*x_ref[2]
        ref[1,4,0,0]=(-0.139033645615)*x_ref[0]**o+(-0.271453722366)*x_ref[0]+(-0.268516720244)*x_ref[1]**o+(-0.0467555017438)*x_ref[1]+(-0.41618262265)*x_ref[2]**o+(0.315320363813)*x_ref[2]
        ref[1,4,0,1]=(0.296108607204)*x_ref[0]**o+(-0.184582165661)*x_ref[0]+(0.121416543744)*x_ref[1]**o+(-0.922284926751)*x_ref[1]+(-0.437600089668)*x_ref[2]**o+(-0.608800626513)*x_ref[2]
        ref[1,4,1,0]=(0.8942028423)*x_ref[0]**o+(0.673402831188)*x_ref[0]+(-0.622529195562)*x_ref[1]**o+(-0.266052338798)*x_ref[1]+(-0.273167851702)*x_ref[2]**o+(-0.312211982747)*x_ref[2]
        ref[1,4,1,1]=(-0.102315711178)*x_ref[0]**o+(-0.885515719371)*x_ref[0]+(-0.532910101204)*x_ref[1]**o+(0.798525944348)*x_ref[1]+(0.640586802303)*x_ref[2]**o+(0.36308806536)*x_ref[2]
        ref[1,4,2,0]=(0.376793506541)*x_ref[0]**o+(-0.667063564875)*x_ref[0]+(0.609591985816)*x_ref[1]**o+(0.28347745142)*x_ref[1]+(-0.751832936302)*x_ref[2]**o+(-0.683738856965)*x_ref[2]
        ref[1,4,2,1]=(-0.51412252819)*x_ref[0]**o+(0.559560297206)*x_ref[0]+(0.751134603429)*x_ref[1]**o+(-0.0031812310797)*x_ref[1]+(-0.525776122178)*x_ref[2]**o+(0.338471942848)*x_ref[2]
        ref[2,0,0,0]=(0.937572244905)*x_ref[0]**o+(-0.293020870379)*x_ref[0]+(-0.918591520817)*x_ref[1]**o+(0.622194477261)*x_ref[1]+(0.31641482851)*x_ref[2]**o+(0.276912266864)*x_ref[2]
        ref[2,0,0,1]=(0.853041599998)*x_ref[0]**o+(-0.0220623724171)*x_ref[0]+(0.89909044296)*x_ref[1]**o+(-0.838782642732)*x_ref[1]+(0.0133197592042)*x_ref[2]**o+(-0.868977618993)*x_ref[2]
        ref[2,0,1,0]=(-0.325799042976)*x_ref[0]**o+(0.746269971994)*x_ref[0]+(0.882671061757)*x_ref[1]**o+(-0.284133373252)*x_ref[1]+(0.120086549521)*x_ref[2]**o+(-0.397352610969)*x_ref[2]
        ref[2,0,1,1]=(0.282336934951)*x_ref[0]**o+(-0.339893978547)*x_ref[0]+(-0.0861660125386)*x_ref[1]**o+(0.283063579798)*x_ref[1]+(0.714762478569)*x_ref[2]**o+(-0.497021718549)*x_ref[2]
        ref[2,0,2,0]=(0.891782779491)*x_ref[0]**o+(-0.817647577644)*x_ref[0]+(0.364088142085)*x_ref[1]**o+(-0.950577637048)*x_ref[1]+(0.4070663557)*x_ref[2]**o+(0.270753179246)*x_ref[2]
        ref[2,0,2,1]=(0.316594913808)*x_ref[0]**o+(0.310731486851)*x_ref[0]+(0.250421828765)*x_ref[1]**o+(0.772786141458)*x_ref[1]+(0.64843040679)*x_ref[2]**o+(-0.691903332511)*x_ref[2]
        ref[2,1,0,0]=(0.87180715028)*x_ref[0]**o+(-0.953173230375)*x_ref[0]+(-0.823516344548)*x_ref[1]**o+(0.934653756548)*x_ref[1]+(-0.537124962489)*x_ref[2]**o+(-0.233365939659)*x_ref[2]
        ref[2,1,0,1]=(-0.407189744631)*x_ref[0]**o+(-0.754425720288)*x_ref[0]+(0.737526767897)*x_ref[1]**o+(0.745393943502)*x_ref[1]+(0.811744293685)*x_ref[2]**o+(-0.459556659177)*x_ref[2]
        ref[2,1,1,0]=(-0.693529259877)*x_ref[0]**o+(0.580197938048)*x_ref[0]+(0.506055440405)*x_ref[1]**o+(0.122531014759)*x_ref[1]+(0.427736137775)*x_ref[2]**o+(-0.83260140401)*x_ref[2]
        ref[2,1,1,1]=(0.542621975026)*x_ref[0]**o+(-0.568425713106)*x_ref[0]+(0.303002043446)*x_ref[1]**o+(0.363286261651)*x_ref[1]+(0.2729476289)*x_ref[2]**o+(-0.0113629744659)*x_ref[2]
        ref[2,1,2,0]=(-0.266750235694)*x_ref[0]**o+(0.10116400306)*x_ref[0]+(0.315977962897)*x_ref[1]**o+(-0.406236201454)*x_ref[1]+(-0.586845072374)*x_ref[2]**o+(0.513937047282)*x_ref[2]
        ref[2,1,2,1]=(-0.518780811975)*x_ref[0]**o+(0.275053640829)*x_ref[0]+(0.4234970646)*x_ref[1]**o+(0.389585163376)*x_ref[1]+(-0.609771068778)*x_ref[2]**o+(-0.121467061755)*x_ref[2]
        ref[2,2,0,0]=(0.138823207466)*x_ref[0]**o+(-0.779837455014)*x_ref[0]+(0.70167049555)*x_ref[1]**o+(0.766428813369)*x_ref[1]+(0.114023530366)*x_ref[2]**o+(0.650098536187)*x_ref[2]
        ref[2,2,0,1]=(0.974998667633)*x_ref[0]**o+(0.958471050654)*x_ref[0]+(0.193515162503)*x_ref[1]**o+(-0.252148167253)*x_ref[1]+(0.495475460125)*x_ref[2]**o+(0.562693347983)*x_ref[2]
        ref[2,2,1,0]=(-0.552379548988)*x_ref[0]**o+(-0.831399971618)*x_ref[0]+(0.437581216647)*x_ref[1]**o+(-0.775331267696)*x_ref[1]+(0.729041628844)*x_ref[2]**o+(0.633113359893)*x_ref[2]
        ref[2,2,1,1]=(0.379858437049)*x_ref[0]**o+(0.923586504203)*x_ref[0]+(0.707244783969)*x_ref[1]**o+(-0.293157375059)*x_ref[1]+(-0.368717941876)*x_ref[2]**o+(-0.202610288616)*x_ref[2]
        ref[2,2,2,0]=(-0.804826108231)*x_ref[0]**o+(-0.0208712096312)*x_ref[0]+(0.0482547072498)*x_ref[1]**o+(0.852366543609)*x_ref[1]+(0.328962500136)*x_ref[2]**o+(-0.132777835512)*x_ref[2]
        ref[2,2,2,1]=(0.91373752217)*x_ref[0]**o+(-0.792112120729)*x_ref[0]+(-0.0457564087072)*x_ref[1]**o+(0.309694852949)*x_ref[1]+(0.089219285337)*x_ref[2]**o+(-0.516000514614)*x_ref[2]
        ref[2,3,0,0]=(0.343215129115)*x_ref[0]**o+(0.40804099837)*x_ref[0]+(-0.865942442914)*x_ref[1]**o+(0.842241683043)*x_ref[1]+(0.0727803297601)*x_ref[2]**o+(-0.395225682949)*x_ref[2]
        ref[2,3,0,1]=(-0.965340522771)*x_ref[0]**o+(-0.897056478677)*x_ref[0]+(-0.305205944062)*x_ref[1]**o+(0.393941789685)*x_ref[1]+(-0.253328379131)*x_ref[2]**o+(0.0951510469271)*x_ref[2]
        ref[2,3,1,0]=(0.536251514752)*x_ref[0]**o+(0.0629776386779)*x_ref[0]+(-0.706311222379)*x_ref[1]**o+(-0.84362242029)*x_ref[1]+(0.00110678760539)*x_ref[2]**o+(-0.30286059941)*x_ref[2]
        ref[2,3,1,1]=(0.114584737295)*x_ref[0]**o+(0.0708763738067)*x_ref[0]+(-0.454760062397)*x_ref[1]**o+(0.228742539922)*x_ref[1]+(0.514960810352)*x_ref[2]**o+(-0.815749211085)*x_ref[2]
        ref[2,3,2,0]=(0.246335688008)*x_ref[0]**o+(-0.239199299885)*x_ref[0]+(-0.447390151659)*x_ref[1]**o+(0.918949356391)*x_ref[1]+(-0.307864840175)*x_ref[2]**o+(0.560139080481)*x_ref[2]
        ref[2,3,2,1]=(0.118766035723)*x_ref[0]**o+(0.610711164899)*x_ref[0]+(0.336628771776)*x_ref[1]**o+(0.833699268751)*x_ref[1]+(-0.424694647429)*x_ref[2]**o+(-0.639929688104)*x_ref[2]
        ref[2,4,0,0]=(0.814543506987)*x_ref[0]**o+(-0.797684321368)*x_ref[0]+(0.337242635783)*x_ref[1]**o+(0.278139837072)*x_ref[1]+(0.633511526441)*x_ref[2]**o+(0.372175615028)*x_ref[2]
        ref[2,4,0,1]=(-0.0590668694094)*x_ref[0]**o+(0.256732067267)*x_ref[0]+(0.519737187577)*x_ref[1]**o+(-0.376391855948)*x_ref[1]+(-0.810934658651)*x_ref[2]**o+(0.826899476928)*x_ref[2]
        ref[2,4,1,0]=(-0.428976843661)*x_ref[0]**o+(0.672667833313)*x_ref[0]+(0.263194890946)*x_ref[1]**o+(-0.894790773661)*x_ref[1]+(-0.550298557295)*x_ref[2]**o+(-0.00560384035502)*x_ref[2]
        ref[2,4,1,1]=(-0.818372430144)*x_ref[0]**o+(-0.829920833073)*x_ref[0]+(-0.359560636119)*x_ref[1]**o+(-0.113484563205)*x_ref[1]+(0.586796110365)*x_ref[2]**o+(0.296820898426)*x_ref[2]
        ref[2,4,2,0]=(-0.691597969566)*x_ref[0]**o+(0.715813153538)*x_ref[0]+(0.954093612503)*x_ref[1]**o+(0.88803420754)*x_ref[1]+(0.697447667765)*x_ref[2]**o+(0.793730496625)*x_ref[2]
        ref[2,4,2,1]=(0.351099833336)*x_ref[0]**o+(-0.716419272807)*x_ref[0]+(0.318726674998)*x_ref[1]**o+(0.246151914414)*x_ref[1]+(-0.945520673884)*x_ref[2]**o+(-0.489026233954)*x_ref[2]
        ref[3,0,0,0]=(0.511598022804)*x_ref[0]**o+(0.567617269582)*x_ref[0]+(-0.526293319137)*x_ref[1]**o+(0.333750098146)*x_ref[1]+(-0.0869228849144)*x_ref[2]**o+(-0.204072834867)*x_ref[2]
        ref[3,0,0,1]=(0.905384007172)*x_ref[0]**o+(-0.661067390314)*x_ref[0]+(-0.470790632295)*x_ref[1]**o+(-0.149608017616)*x_ref[1]+(0.427879278267)*x_ref[2]**o+(0.987040179184)*x_ref[2]
        ref[3,0,1,0]=(-0.946562006809)*x_ref[0]**o+(0.64141909633)*x_ref[0]+(-0.762076826276)*x_ref[1]**o+(-0.642912850794)*x_ref[1]+(0.712567898559)*x_ref[2]**o+(-0.292128944205)*x_ref[2]
        ref[3,0,1,1]=(-0.955087949361)*x_ref[0]**o+(0.426588041133)*x_ref[0]+(0.339313104127)*x_ref[1]**o+(-0.375147344045)*x_ref[1]+(0.967259014679)*x_ref[2]**o+(0.504521081858)*x_ref[2]
        ref[3,0,2,0]=(0.141876598344)*x_ref[0]**o+(-0.335579252825)*x_ref[0]+(0.514552993602)*x_ref[1]**o+(-0.531116948499)*x_ref[1]+(-0.369319413141)*x_ref[2]**o+(0.0374413262362)*x_ref[2]
        ref[3,0,2,1]=(0.678360463658)*x_ref[0]**o+(-0.357900119646)*x_ref[0]+(0.300750717059)*x_ref[1]**o+(0.516196189448)*x_ref[1]+(0.0535466018203)*x_ref[2]**o+(-0.149961940786)*x_ref[2]
        ref[3,1,0,0]=(0.0975994351079)*x_ref[0]**o+(0.634544068499)*x_ref[0]+(-0.963217046352)*x_ref[1]**o+(-0.804658091456)*x_ref[1]+(-0.620132720823)*x_ref[2]**o+(0.862250619369)*x_ref[2]
        ref[3,1,0,1]=(0.210306328879)*x_ref[0]**o+(-0.457341187966)*x_ref[0]+(0.957556084801)*x_ref[1]**o+(0.704073170928)*x_ref[1]+(0.423134492133)*x_ref[2]**o+(-0.167268375412)*x_ref[2]
        ref[3,1,1,0]=(-0.909228817744)*x_ref[0]**o+(0.110665018921)*x_ref[0]+(0.36275630147)*x_ref[1]**o+(0.346866824763)*x_ref[1]+(-0.535600547489)*x_ref[2]**o+(0.627638814375)*x_ref[2]
        ref[3,1,1,1]=(0.328780933823)*x_ref[0]**o+(-0.249563994161)*x_ref[0]+(-0.555789062892)*x_ref[1]**o+(-0.917287525036)*x_ref[1]+(0.386191175017)*x_ref[2]**o+(0.123904319061)*x_ref[2]
        ref[3,1,2,0]=(-0.838409901623)*x_ref[0]**o+(0.673479705393)*x_ref[0]+(0.482921104917)*x_ref[1]**o+(0.136411388588)*x_ref[1]+(0.583674116225)*x_ref[2]**o+(0.305100836877)*x_ref[2]
        ref[3,1,2,1]=(0.952961008312)*x_ref[0]**o+(-0.0501981079532)*x_ref[0]+(0.661434464642)*x_ref[1]**o+(0.65030469399)*x_ref[1]+(-0.157562001295)*x_ref[2]**o+(-0.249551892917)*x_ref[2]
        ref[3,2,0,0]=(-0.0232804500286)*x_ref[0]**o+(-0.559229392203)*x_ref[0]+(0.127967930273)*x_ref[1]**o+(-0.635827181645)*x_ref[1]+(-0.0426703224383)*x_ref[2]**o+(0.47720230258)*x_ref[2]
        ref[3,2,0,1]=(-0.421658996106)*x_ref[0]**o+(-0.468260038077)*x_ref[0]+(0.479083572554)*x_ref[1]**o+(-0.214417881841)*x_ref[1]+(0.372430534261)*x_ref[2]**o+(0.133063789546)*x_ref[2]
        ref[3,2,1,0]=(0.905811586567)*x_ref[0]**o+(0.245447340384)*x_ref[0]+(-0.337650999146)*x_ref[1]**o+(0.908188235976)*x_ref[1]+(-0.804029751277)*x_ref[2]**o+(0.0512676456643)*x_ref[2]
        ref[3,2,1,1]=(-0.891486839409)*x_ref[0]**o+(0.20753794329)*x_ref[0]+(0.879183464252)*x_ref[1]**o+(0.902028906992)*x_ref[1]+(0.416776813598)*x_ref[2]**o+(0.491964374373)*x_ref[2]
        ref[3,2,2,0]=(0.653067027119)*x_ref[0]**o+(0.765293890745)*x_ref[0]+(-0.348902169508)*x_ref[1]**o+(0.324429035849)*x_ref[1]+(-0.363026365065)*x_ref[2]**o+(0.0816186405511)*x_ref[2]
        ref[3,2,2,1]=(-0.0415375052325)*x_ref[0]**o+(-0.480779075784)*x_ref[0]+(-0.137223552938)*x_ref[1]**o+(-0.915733625512)*x_ref[1]+(0.750199445159)*x_ref[2]**o+(-0.789818945171)*x_ref[2]
        ref[3,3,0,0]=(0.0744921239083)*x_ref[0]**o+(0.0700589144159)*x_ref[0]+(-0.327105425297)*x_ref[1]**o+(0.469843257765)*x_ref[1]+(-0.170191600844)*x_ref[2]**o+(0.732071534155)*x_ref[2]
        ref[3,3,0,1]=(0.172643927441)*x_ref[0]**o+(0.848924114383)*x_ref[0]+(-0.79125762228)*x_ref[1]**o+(-0.694725775976)*x_ref[1]+(-0.309710503985)*x_ref[2]**o+(-0.405310234927)*x_ref[2]
        ref[3,3,1,0]=(-0.465911787595)*x_ref[0]**o+(-0.71080474709)*x_ref[0]+(0.959925934203)*x_ref[1]**o+(-0.111110644825)*x_ref[1]+(-0.145349925218)*x_ref[2]**o+(0.122955165811)*x_ref[2]
        ref[3,3,1,1]=(-0.769111863754)*x_ref[0]**o+(0.348036938532)*x_ref[0]+(0.485890856474)*x_ref[1]**o+(0.382595218985)*x_ref[1]+(-0.897488692981)*x_ref[2]**o+(0.830822333819)*x_ref[2]
        ref[3,3,2,0]=(0.341890793681)*x_ref[0]**o+(-0.667145936571)*x_ref[0]+(-0.145094108268)*x_ref[1]**o+(0.771433911212)*x_ref[1]+(-0.299752164325)*x_ref[2]**o+(-0.357856549677)*x_ref[2]
        ref[3,3,2,1]=(0.26381838796)*x_ref[0]**o+(-0.44474724998)*x_ref[0]+(0.310283988195)*x_ref[1]**o+(0.698675181314)*x_ref[1]+(0.437624133238)*x_ref[2]**o+(0.394916054445)*x_ref[2]
        ref[3,4,0,0]=(-0.576499406674)*x_ref[0]**o+(0.770936634959)*x_ref[0]+(0.707441735783)*x_ref[1]**o+(-0.794953257043)*x_ref[1]+(-0.535342943209)*x_ref[2]**o+(-0.509446811787)*x_ref[2]
        ref[3,4,0,1]=(0.161115012113)*x_ref[0]**o+(0.11079092252)*x_ref[0]+(0.0633804181164)*x_ref[1]**o+(-0.406049293019)*x_ref[1]+(-0.824283994336)*x_ref[2]**o+(-0.481107560499)*x_ref[2]
        ref[3,4,1,0]=(-0.411133810295)*x_ref[0]**o+(0.988486625099)*x_ref[0]+(-0.36050191367)*x_ref[1]**o+(0.469479821965)*x_ref[1]+(-0.431175357325)*x_ref[2]**o+(-0.966295015263)*x_ref[2]
        ref[3,4,1,1]=(0.233731410367)*x_ref[0]**o+(-0.821900740766)*x_ref[0]+(-0.823123461385)*x_ref[1]**o+(-0.380537914421)*x_ref[1]+(-0.896146300355)*x_ref[2]**o+(-0.775172990257)*x_ref[2]
        ref[3,4,2,0]=(-0.96323363586)*x_ref[0]**o+(0.0935412405245)*x_ref[0]+(0.929333026259)*x_ref[1]**o+(-0.322236627348)*x_ref[1]+(-0.390898135964)*x_ref[2]**o+(-0.766178290162)*x_ref[2]
        ref[3,4,2,1]=(-0.773670533268)*x_ref[0]**o+(-0.156975157098)*x_ref[0]+(-0.91512996795)*x_ref[1]**o+(0.0247803682126)*x_ref[1]+(0.995159733782)*x_ref[2]**o+(-0.669224144225)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.0071553637759)*x[0]**o+(0.357615094637)*x[0]+(0.980385970972)*x[1]**o+(0.423505273491)*x[1]
        ref=(-0.0071553637759)*x_ref[0]**o+(0.357615094637)*x_ref[0]+(0.980385970972)*x_ref[1]**o+(0.423505273491)*x_ref[1]
      else:
        arg=(0.546251547514)*x[0]**o+(-0.530405447164)*x[0]+(-0.901465213708)*x[1]**o+(-0.986019835171)*x[1]+(-0.731836172089)*x[2]**o+(0.377317674082)*x[2]
        ref=(0.546251547514)*x_ref[0]**o+(-0.530405447164)*x_ref[0]+(-0.901465213708)*x_ref[1]**o+(-0.986019835171)*x_ref[1]+(-0.731836172089)*x_ref[2]**o+(0.377317674082)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.138216822552)*x[0]**o+(0.908302047805)*x[0]+(-0.247600939209)*x[1]**o+(0.776239000167)*x[1]
        arg[1]=(-0.577926899073)*x[0]**o+(0.714263102346)*x[0]+(-0.343808625475)*x[1]**o+(0.786295563423)*x[1]
        ref[0]=(-0.138216822552)*x_ref[0]**o+(0.908302047805)*x_ref[0]+(-0.247600939209)*x_ref[1]**o+(0.776239000167)*x_ref[1]
        ref[1]=(-0.577926899073)*x_ref[0]**o+(0.714263102346)*x_ref[0]+(-0.343808625475)*x_ref[1]**o+(0.786295563423)*x_ref[1]
      else:
        arg[0]=(0.020232467504)*x[0]**o+(-0.0449727381653)*x[0]+(0.00132143537494)*x[1]**o+(-0.870292414095)*x[1]+(-0.926731511382)*x[2]**o+(-0.398915622332)*x[2]
        arg[1]=(0.203864545604)*x[0]**o+(0.150501673382)*x[0]+(0.132132290114)*x[1]**o+(-0.720617456803)*x[1]+(0.673909788445)*x[2]**o+(0.145376368499)*x[2]
        ref[0]=(0.020232467504)*x_ref[0]**o+(-0.0449727381653)*x_ref[0]+(0.00132143537494)*x_ref[1]**o+(-0.870292414095)*x_ref[1]+(-0.926731511382)*x_ref[2]**o+(-0.398915622332)*x_ref[2]
        ref[1]=(0.203864545604)*x_ref[0]**o+(0.150501673382)*x_ref[0]+(0.132132290114)*x_ref[1]**o+(-0.720617456803)*x_ref[1]+(0.673909788445)*x_ref[2]**o+(0.145376368499)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.344685653483)*x[0]**o+(-0.618443209029)*x[0]+(-0.338246837583)*x[1]**o+(-0.709311881709)*x[1]
        arg[0,1]=(-0.176341370817)*x[0]**o+(-0.230030832268)*x[0]+(-0.335163038431)*x[1]**o+(0.278627093006)*x[1]
        arg[0,2]=(0.376270607923)*x[0]**o+(0.895822309578)*x[0]+(0.418316736671)*x[1]**o+(-0.951461687587)*x[1]
        arg[0,3]=(0.69925470532)*x[0]**o+(-0.829917904706)*x[0]+(0.60851481091)*x[1]**o+(-0.880924691692)*x[1]
        arg[0,4]=(0.537570501499)*x[0]**o+(0.584091505535)*x[0]+(-0.0045669307163)*x[1]**o+(0.162714723047)*x[1]
        arg[1,0]=(-0.244615538677)*x[0]**o+(-0.0168359545772)*x[0]+(-0.511097014251)*x[1]**o+(-0.476830089968)*x[1]
        arg[1,1]=(0.350953788855)*x[0]**o+(0.0859931352733)*x[0]+(-0.565664846934)*x[1]**o+(-0.230383210864)*x[1]
        arg[1,2]=(0.787537227138)*x[0]**o+(-0.97589028834)*x[0]+(0.978870857538)*x[1]**o+(0.436540016432)*x[1]
        arg[1,3]=(-0.806777674064)*x[0]**o+(-0.384971723467)*x[0]+(-0.423836322283)*x[1]**o+(0.470609842503)*x[1]
        arg[1,4]=(0.288422170494)*x[0]**o+(-0.467670464119)*x[0]+(-0.616952048135)*x[1]**o+(0.604831335817)*x[1]
        arg[2,0]=(0.0461969205986)*x[0]**o+(0.663083190578)*x[0]+(-0.673597938582)*x[1]**o+(0.217389040208)*x[1]
        arg[2,1]=(-0.32008443048)*x[0]**o+(-0.970854886309)*x[0]+(0.630220968872)*x[1]**o+(0.827824011117)*x[1]
        arg[2,2]=(-0.232867522682)*x[0]**o+(-0.17921445933)*x[0]+(-0.565289194319)*x[1]**o+(0.0251729876325)*x[1]
        arg[2,3]=(-0.862002880133)*x[0]**o+(0.64790108714)*x[0]+(-0.469428882373)*x[1]**o+(0.50581485158)*x[1]
        arg[2,4]=(-0.957389923449)*x[0]**o+(-0.716972488151)*x[0]+(0.848168933491)*x[1]**o+(0.78614576485)*x[1]
        arg[3,0]=(0.417890378265)*x[0]**o+(-0.128884271967)*x[0]+(0.981251819451)*x[1]**o+(0.837945499182)*x[1]
        arg[3,1]=(-0.296791907211)*x[0]**o+(0.913966659909)*x[0]+(0.0171079547828)*x[1]**o+(0.472922286638)*x[1]
        arg[3,2]=(-0.549504779354)*x[0]**o+(0.356989069435)*x[0]+(-0.0978494385269)*x[1]**o+(0.59354694599)*x[1]
        arg[3,3]=(0.0272055308458)*x[0]**o+(0.560184544705)*x[0]+(-0.825452907104)*x[1]**o+(0.701919616153)*x[1]
        arg[3,4]=(-0.33094753848)*x[0]**o+(0.877723269108)*x[0]+(-0.491915661514)*x[1]**o+(-0.98784807255)*x[1]
        ref[0,0]=(-0.344685653483)*x_ref[0]**o+(-0.618443209029)*x_ref[0]+(-0.338246837583)*x_ref[1]**o+(-0.709311881709)*x_ref[1]
        ref[0,1]=(-0.176341370817)*x_ref[0]**o+(-0.230030832268)*x_ref[0]+(-0.335163038431)*x_ref[1]**o+(0.278627093006)*x_ref[1]
        ref[0,2]=(0.376270607923)*x_ref[0]**o+(0.895822309578)*x_ref[0]+(0.418316736671)*x_ref[1]**o+(-0.951461687587)*x_ref[1]
        ref[0,3]=(0.69925470532)*x_ref[0]**o+(-0.829917904706)*x_ref[0]+(0.60851481091)*x_ref[1]**o+(-0.880924691692)*x_ref[1]
        ref[0,4]=(0.537570501499)*x_ref[0]**o+(0.584091505535)*x_ref[0]+(-0.0045669307163)*x_ref[1]**o+(0.162714723047)*x_ref[1]
        ref[1,0]=(-0.244615538677)*x_ref[0]**o+(-0.0168359545772)*x_ref[0]+(-0.511097014251)*x_ref[1]**o+(-0.476830089968)*x_ref[1]
        ref[1,1]=(0.350953788855)*x_ref[0]**o+(0.0859931352733)*x_ref[0]+(-0.565664846934)*x_ref[1]**o+(-0.230383210864)*x_ref[1]
        ref[1,2]=(0.787537227138)*x_ref[0]**o+(-0.97589028834)*x_ref[0]+(0.978870857538)*x_ref[1]**o+(0.436540016432)*x_ref[1]
        ref[1,3]=(-0.806777674064)*x_ref[0]**o+(-0.384971723467)*x_ref[0]+(-0.423836322283)*x_ref[1]**o+(0.470609842503)*x_ref[1]
        ref[1,4]=(0.288422170494)*x_ref[0]**o+(-0.467670464119)*x_ref[0]+(-0.616952048135)*x_ref[1]**o+(0.604831335817)*x_ref[1]
        ref[2,0]=(0.0461969205986)*x_ref[0]**o+(0.663083190578)*x_ref[0]+(-0.673597938582)*x_ref[1]**o+(0.217389040208)*x_ref[1]
        ref[2,1]=(-0.32008443048)*x_ref[0]**o+(-0.970854886309)*x_ref[0]+(0.630220968872)*x_ref[1]**o+(0.827824011117)*x_ref[1]
        ref[2,2]=(-0.232867522682)*x_ref[0]**o+(-0.17921445933)*x_ref[0]+(-0.565289194319)*x_ref[1]**o+(0.0251729876325)*x_ref[1]
        ref[2,3]=(-0.862002880133)*x_ref[0]**o+(0.64790108714)*x_ref[0]+(-0.469428882373)*x_ref[1]**o+(0.50581485158)*x_ref[1]
        ref[2,4]=(-0.957389923449)*x_ref[0]**o+(-0.716972488151)*x_ref[0]+(0.848168933491)*x_ref[1]**o+(0.78614576485)*x_ref[1]
        ref[3,0]=(0.417890378265)*x_ref[0]**o+(-0.128884271967)*x_ref[0]+(0.981251819451)*x_ref[1]**o+(0.837945499182)*x_ref[1]
        ref[3,1]=(-0.296791907211)*x_ref[0]**o+(0.913966659909)*x_ref[0]+(0.0171079547828)*x_ref[1]**o+(0.472922286638)*x_ref[1]
        ref[3,2]=(-0.549504779354)*x_ref[0]**o+(0.356989069435)*x_ref[0]+(-0.0978494385269)*x_ref[1]**o+(0.59354694599)*x_ref[1]
        ref[3,3]=(0.0272055308458)*x_ref[0]**o+(0.560184544705)*x_ref[0]+(-0.825452907104)*x_ref[1]**o+(0.701919616153)*x_ref[1]
        ref[3,4]=(-0.33094753848)*x_ref[0]**o+(0.877723269108)*x_ref[0]+(-0.491915661514)*x_ref[1]**o+(-0.98784807255)*x_ref[1]
      else:
        arg[0,0]=(-0.434721514594)*x[0]**o+(-0.28074158231)*x[0]+(0.102692621316)*x[1]**o+(0.399677492352)*x[1]+(0.39799987301)*x[2]**o+(-0.134547332796)*x[2]
        arg[0,1]=(-0.814630031471)*x[0]**o+(-0.0722898377573)*x[0]+(0.356831166041)*x[1]**o+(0.0556663518903)*x[1]+(-0.335764152958)*x[2]**o+(0.344564519072)*x[2]
        arg[0,2]=(-0.881698115197)*x[0]**o+(-0.030759288922)*x[0]+(-0.23126177081)*x[1]**o+(0.865011678258)*x[1]+(0.961867471932)*x[2]**o+(-0.04809978604)*x[2]
        arg[0,3]=(-0.931319970946)*x[0]**o+(0.807195614255)*x[0]+(0.116305510032)*x[1]**o+(0.337873586448)*x[1]+(0.523766276637)*x[2]**o+(-0.890502808117)*x[2]
        arg[0,4]=(0.952369561411)*x[0]**o+(0.518186569223)*x[0]+(0.184454818734)*x[1]**o+(-0.198618411156)*x[1]+(0.891317797589)*x[2]**o+(-0.312268228065)*x[2]
        arg[1,0]=(-0.770379397538)*x[0]**o+(0.11068893635)*x[0]+(0.435574518842)*x[1]**o+(-0.895094805657)*x[1]+(0.719886389608)*x[2]**o+(0.617254266829)*x[2]
        arg[1,1]=(-0.968566082536)*x[0]**o+(0.70331821608)*x[0]+(-0.256092697223)*x[1]**o+(0.214482838048)*x[1]+(-0.20817108493)*x[2]**o+(-0.741961694045)*x[2]
        arg[1,2]=(0.928194134464)*x[0]**o+(0.748278982508)*x[0]+(0.459001496702)*x[1]**o+(-0.333954496033)*x[1]+(0.0451659439418)*x[2]**o+(0.0355669781059)*x[2]
        arg[1,3]=(-0.285564661682)*x[0]**o+(-0.780972389596)*x[0]+(0.20442278849)*x[1]**o+(-0.425673038027)*x[1]+(-0.886021022849)*x[2]**o+(0.2231241391)*x[2]
        arg[1,4]=(0.48756511391)*x[0]**o+(0.904219456651)*x[0]+(-0.723245954496)*x[1]**o+(-0.876483077689)*x[1]+(0.454913273237)*x[2]**o+(0.907651543551)*x[2]
        arg[2,0]=(-0.490623674798)*x[0]**o+(0.179561344392)*x[0]+(-0.257796709265)*x[1]**o+(0.470216680739)*x[1]+(-0.334946805819)*x[2]**o+(-0.295308708488)*x[2]
        arg[2,1]=(-0.0753731003603)*x[0]**o+(0.848622766023)*x[0]+(-0.168762035291)*x[1]**o+(-0.132048839333)*x[1]+(0.229089231051)*x[2]**o+(0.280032186468)*x[2]
        arg[2,2]=(0.601432167354)*x[0]**o+(-0.0857688429764)*x[0]+(0.556763499853)*x[1]**o+(-0.0953568285359)*x[1]+(-0.396589175262)*x[2]**o+(0.437257052414)*x[2]
        arg[2,3]=(0.239710792478)*x[0]**o+(0.255863660922)*x[0]+(-0.860831814227)*x[1]**o+(0.416057968294)*x[1]+(0.284515440811)*x[2]**o+(-0.0737992548315)*x[2]
        arg[2,4]=(-0.86884373286)*x[0]**o+(0.306811019609)*x[0]+(0.955007347168)*x[1]**o+(0.731959544948)*x[1]+(0.326947842619)*x[2]**o+(0.430618227986)*x[2]
        arg[3,0]=(0.450341287007)*x[0]**o+(-0.0318463261779)*x[0]+(-0.755083793007)*x[1]**o+(0.427706480425)*x[1]+(0.224383931841)*x[2]**o+(-0.801200151542)*x[2]
        arg[3,1]=(-0.750405028813)*x[0]**o+(-0.531057162102)*x[0]+(0.120116357521)*x[1]**o+(-0.247640504917)*x[1]+(-0.360494371461)*x[2]**o+(-0.496404068337)*x[2]
        arg[3,2]=(-0.202775617449)*x[0]**o+(0.833337604045)*x[0]+(0.551285270602)*x[1]**o+(0.266630410234)*x[1]+(0.903041333927)*x[2]**o+(0.527233939028)*x[2]
        arg[3,3]=(0.12892683791)*x[0]**o+(-0.486726243104)*x[0]+(-0.340728397711)*x[1]**o+(-0.670692588507)*x[1]+(0.832250132781)*x[2]**o+(0.276897913372)*x[2]
        arg[3,4]=(0.839441639419)*x[0]**o+(0.362150671815)*x[0]+(0.728370029046)*x[1]**o+(0.149540296968)*x[1]+(-0.0418719598944)*x[2]**o+(-0.331805297414)*x[2]
        ref[0,0]=(-0.434721514594)*x_ref[0]**o+(-0.28074158231)*x_ref[0]+(0.102692621316)*x_ref[1]**o+(0.399677492352)*x_ref[1]+(0.39799987301)*x_ref[2]**o+(-0.134547332796)*x_ref[2]
        ref[0,1]=(-0.814630031471)*x_ref[0]**o+(-0.0722898377573)*x_ref[0]+(0.356831166041)*x_ref[1]**o+(0.0556663518903)*x_ref[1]+(-0.335764152958)*x_ref[2]**o+(0.344564519072)*x_ref[2]
        ref[0,2]=(-0.881698115197)*x_ref[0]**o+(-0.030759288922)*x_ref[0]+(-0.23126177081)*x_ref[1]**o+(0.865011678258)*x_ref[1]+(0.961867471932)*x_ref[2]**o+(-0.04809978604)*x_ref[2]
        ref[0,3]=(-0.931319970946)*x_ref[0]**o+(0.807195614255)*x_ref[0]+(0.116305510032)*x_ref[1]**o+(0.337873586448)*x_ref[1]+(0.523766276637)*x_ref[2]**o+(-0.890502808117)*x_ref[2]
        ref[0,4]=(0.952369561411)*x_ref[0]**o+(0.518186569223)*x_ref[0]+(0.184454818734)*x_ref[1]**o+(-0.198618411156)*x_ref[1]+(0.891317797589)*x_ref[2]**o+(-0.312268228065)*x_ref[2]
        ref[1,0]=(-0.770379397538)*x_ref[0]**o+(0.11068893635)*x_ref[0]+(0.435574518842)*x_ref[1]**o+(-0.895094805657)*x_ref[1]+(0.719886389608)*x_ref[2]**o+(0.617254266829)*x_ref[2]
        ref[1,1]=(-0.968566082536)*x_ref[0]**o+(0.70331821608)*x_ref[0]+(-0.256092697223)*x_ref[1]**o+(0.214482838048)*x_ref[1]+(-0.20817108493)*x_ref[2]**o+(-0.741961694045)*x_ref[2]
        ref[1,2]=(0.928194134464)*x_ref[0]**o+(0.748278982508)*x_ref[0]+(0.459001496702)*x_ref[1]**o+(-0.333954496033)*x_ref[1]+(0.0451659439418)*x_ref[2]**o+(0.0355669781059)*x_ref[2]
        ref[1,3]=(-0.285564661682)*x_ref[0]**o+(-0.780972389596)*x_ref[0]+(0.20442278849)*x_ref[1]**o+(-0.425673038027)*x_ref[1]+(-0.886021022849)*x_ref[2]**o+(0.2231241391)*x_ref[2]
        ref[1,4]=(0.48756511391)*x_ref[0]**o+(0.904219456651)*x_ref[0]+(-0.723245954496)*x_ref[1]**o+(-0.876483077689)*x_ref[1]+(0.454913273237)*x_ref[2]**o+(0.907651543551)*x_ref[2]
        ref[2,0]=(-0.490623674798)*x_ref[0]**o+(0.179561344392)*x_ref[0]+(-0.257796709265)*x_ref[1]**o+(0.470216680739)*x_ref[1]+(-0.334946805819)*x_ref[2]**o+(-0.295308708488)*x_ref[2]
        ref[2,1]=(-0.0753731003603)*x_ref[0]**o+(0.848622766023)*x_ref[0]+(-0.168762035291)*x_ref[1]**o+(-0.132048839333)*x_ref[1]+(0.229089231051)*x_ref[2]**o+(0.280032186468)*x_ref[2]
        ref[2,2]=(0.601432167354)*x_ref[0]**o+(-0.0857688429764)*x_ref[0]+(0.556763499853)*x_ref[1]**o+(-0.0953568285359)*x_ref[1]+(-0.396589175262)*x_ref[2]**o+(0.437257052414)*x_ref[2]
        ref[2,3]=(0.239710792478)*x_ref[0]**o+(0.255863660922)*x_ref[0]+(-0.860831814227)*x_ref[1]**o+(0.416057968294)*x_ref[1]+(0.284515440811)*x_ref[2]**o+(-0.0737992548315)*x_ref[2]
        ref[2,4]=(-0.86884373286)*x_ref[0]**o+(0.306811019609)*x_ref[0]+(0.955007347168)*x_ref[1]**o+(0.731959544948)*x_ref[1]+(0.326947842619)*x_ref[2]**o+(0.430618227986)*x_ref[2]
        ref[3,0]=(0.450341287007)*x_ref[0]**o+(-0.0318463261779)*x_ref[0]+(-0.755083793007)*x_ref[1]**o+(0.427706480425)*x_ref[1]+(0.224383931841)*x_ref[2]**o+(-0.801200151542)*x_ref[2]
        ref[3,1]=(-0.750405028813)*x_ref[0]**o+(-0.531057162102)*x_ref[0]+(0.120116357521)*x_ref[1]**o+(-0.247640504917)*x_ref[1]+(-0.360494371461)*x_ref[2]**o+(-0.496404068337)*x_ref[2]
        ref[3,2]=(-0.202775617449)*x_ref[0]**o+(0.833337604045)*x_ref[0]+(0.551285270602)*x_ref[1]**o+(0.266630410234)*x_ref[1]+(0.903041333927)*x_ref[2]**o+(0.527233939028)*x_ref[2]
        ref[3,3]=(0.12892683791)*x_ref[0]**o+(-0.486726243104)*x_ref[0]+(-0.340728397711)*x_ref[1]**o+(-0.670692588507)*x_ref[1]+(0.832250132781)*x_ref[2]**o+(0.276897913372)*x_ref[2]
        ref[3,4]=(0.839441639419)*x_ref[0]**o+(0.362150671815)*x_ref[0]+(0.728370029046)*x_ref[1]**o+(0.149540296968)*x_ref[1]+(-0.0418719598944)*x_ref[2]**o+(-0.331805297414)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.475615092269)*x[0]**o+(-0.0893751722575)*x[0]+(0.595881277166)*x[1]**o+(0.397900173915)*x[1]
        arg[0,0,1]=(-0.552614805712)*x[0]**o+(0.14072435445)*x[0]+(0.290231845094)*x[1]**o+(0.573878059041)*x[1]
        arg[0,1,0]=(-0.226148312038)*x[0]**o+(-0.737524507359)*x[0]+(-0.704362709293)*x[1]**o+(-0.197223346266)*x[1]
        arg[0,1,1]=(0.844313529642)*x[0]**o+(-0.948514286846)*x[0]+(0.383754990882)*x[1]**o+(0.629819148632)*x[1]
        arg[1,0,0]=(-0.519122333314)*x[0]**o+(-0.895333584981)*x[0]+(-0.941503934561)*x[1]**o+(0.574440495425)*x[1]
        arg[1,0,1]=(0.34967649382)*x[0]**o+(-0.332598318229)*x[0]+(0.0587818420974)*x[1]**o+(0.0848419508498)*x[1]
        arg[1,1,0]=(0.126943263196)*x[0]**o+(0.20287950383)*x[0]+(-0.305454685691)*x[1]**o+(-0.357695790052)*x[1]
        arg[1,1,1]=(0.104154708153)*x[0]**o+(0.867435765411)*x[0]+(0.656946071961)*x[1]**o+(0.96433811362)*x[1]
        arg[2,0,0]=(-0.407362949457)*x[0]**o+(-0.171021521273)*x[0]+(0.901342300398)*x[1]**o+(-0.0360476097399)*x[1]
        arg[2,0,1]=(-0.709421035652)*x[0]**o+(0.949318093375)*x[0]+(0.155061840049)*x[1]**o+(0.828999317354)*x[1]
        arg[2,1,0]=(-0.573031939083)*x[0]**o+(-0.766186602896)*x[0]+(-0.0483071013367)*x[1]**o+(0.741988754559)*x[1]
        arg[2,1,1]=(-0.826747727748)*x[0]**o+(0.481970220735)*x[0]+(0.735617076451)*x[1]**o+(-0.64482586429)*x[1]
        arg[3,0,0]=(0.0834619008603)*x[0]**o+(-0.122260004323)*x[0]+(-0.193792114997)*x[1]**o+(0.635792828013)*x[1]
        arg[3,0,1]=(0.868630978759)*x[0]**o+(0.54474614716)*x[0]+(-0.514377753448)*x[1]**o+(0.58889412844)*x[1]
        arg[3,1,0]=(-0.233744744882)*x[0]**o+(0.883601383004)*x[0]+(0.575012113003)*x[1]**o+(0.185704140149)*x[1]
        arg[3,1,1]=(-0.948893544046)*x[0]**o+(0.851870038501)*x[0]+(-0.591835259957)*x[1]**o+(-0.958375628808)*x[1]
        arg[4,0,0]=(0.981685480559)*x[0]**o+(0.718249744763)*x[0]+(0.589534430546)*x[1]**o+(0.72294148429)*x[1]
        arg[4,0,1]=(-0.603601154465)*x[0]**o+(0.293688738803)*x[0]+(0.151726724518)*x[1]**o+(0.56316377967)*x[1]
        arg[4,1,0]=(0.539033109011)*x[0]**o+(0.791686885872)*x[0]+(0.878973377358)*x[1]**o+(0.869606341625)*x[1]
        arg[4,1,1]=(-0.0920873290752)*x[0]**o+(-0.132682060309)*x[0]+(0.34023826981)*x[1]**o+(0.585476816741)*x[1]
        arg[5,0,0]=(-0.430989548175)*x[0]**o+(-0.525387682044)*x[0]+(-0.996483829778)*x[1]**o+(-0.564503119535)*x[1]
        arg[5,0,1]=(0.602911018749)*x[0]**o+(-0.348972500704)*x[0]+(0.174681048071)*x[1]**o+(-0.94551825287)*x[1]
        arg[5,1,0]=(0.780926392259)*x[0]**o+(-0.164977652547)*x[0]+(0.224726701017)*x[1]**o+(-0.779109318124)*x[1]
        arg[5,1,1]=(0.470928696402)*x[0]**o+(-0.505871268699)*x[0]+(0.407158974816)*x[1]**o+(0.392856659954)*x[1]
        ref[0,0,0]=(-0.475615092269)*x_ref[0]**o+(-0.0893751722575)*x_ref[0]+(0.595881277166)*x_ref[1]**o+(0.397900173915)*x_ref[1]
        ref[0,0,1]=(-0.552614805712)*x_ref[0]**o+(0.14072435445)*x_ref[0]+(0.290231845094)*x_ref[1]**o+(0.573878059041)*x_ref[1]
        ref[0,1,0]=(-0.226148312038)*x_ref[0]**o+(-0.737524507359)*x_ref[0]+(-0.704362709293)*x_ref[1]**o+(-0.197223346266)*x_ref[1]
        ref[0,1,1]=(0.844313529642)*x_ref[0]**o+(-0.948514286846)*x_ref[0]+(0.383754990882)*x_ref[1]**o+(0.629819148632)*x_ref[1]
        ref[1,0,0]=(-0.519122333314)*x_ref[0]**o+(-0.895333584981)*x_ref[0]+(-0.941503934561)*x_ref[1]**o+(0.574440495425)*x_ref[1]
        ref[1,0,1]=(0.34967649382)*x_ref[0]**o+(-0.332598318229)*x_ref[0]+(0.0587818420974)*x_ref[1]**o+(0.0848419508498)*x_ref[1]
        ref[1,1,0]=(0.126943263196)*x_ref[0]**o+(0.20287950383)*x_ref[0]+(-0.305454685691)*x_ref[1]**o+(-0.357695790052)*x_ref[1]
        ref[1,1,1]=(0.104154708153)*x_ref[0]**o+(0.867435765411)*x_ref[0]+(0.656946071961)*x_ref[1]**o+(0.96433811362)*x_ref[1]
        ref[2,0,0]=(-0.407362949457)*x_ref[0]**o+(-0.171021521273)*x_ref[0]+(0.901342300398)*x_ref[1]**o+(-0.0360476097399)*x_ref[1]
        ref[2,0,1]=(-0.709421035652)*x_ref[0]**o+(0.949318093375)*x_ref[0]+(0.155061840049)*x_ref[1]**o+(0.828999317354)*x_ref[1]
        ref[2,1,0]=(-0.573031939083)*x_ref[0]**o+(-0.766186602896)*x_ref[0]+(-0.0483071013367)*x_ref[1]**o+(0.741988754559)*x_ref[1]
        ref[2,1,1]=(-0.826747727748)*x_ref[0]**o+(0.481970220735)*x_ref[0]+(0.735617076451)*x_ref[1]**o+(-0.64482586429)*x_ref[1]
        ref[3,0,0]=(0.0834619008603)*x_ref[0]**o+(-0.122260004323)*x_ref[0]+(-0.193792114997)*x_ref[1]**o+(0.635792828013)*x_ref[1]
        ref[3,0,1]=(0.868630978759)*x_ref[0]**o+(0.54474614716)*x_ref[0]+(-0.514377753448)*x_ref[1]**o+(0.58889412844)*x_ref[1]
        ref[3,1,0]=(-0.233744744882)*x_ref[0]**o+(0.883601383004)*x_ref[0]+(0.575012113003)*x_ref[1]**o+(0.185704140149)*x_ref[1]
        ref[3,1,1]=(-0.948893544046)*x_ref[0]**o+(0.851870038501)*x_ref[0]+(-0.591835259957)*x_ref[1]**o+(-0.958375628808)*x_ref[1]
        ref[4,0,0]=(0.981685480559)*x_ref[0]**o+(0.718249744763)*x_ref[0]+(0.589534430546)*x_ref[1]**o+(0.72294148429)*x_ref[1]
        ref[4,0,1]=(-0.603601154465)*x_ref[0]**o+(0.293688738803)*x_ref[0]+(0.151726724518)*x_ref[1]**o+(0.56316377967)*x_ref[1]
        ref[4,1,0]=(0.539033109011)*x_ref[0]**o+(0.791686885872)*x_ref[0]+(0.878973377358)*x_ref[1]**o+(0.869606341625)*x_ref[1]
        ref[4,1,1]=(-0.0920873290752)*x_ref[0]**o+(-0.132682060309)*x_ref[0]+(0.34023826981)*x_ref[1]**o+(0.585476816741)*x_ref[1]
        ref[5,0,0]=(-0.430989548175)*x_ref[0]**o+(-0.525387682044)*x_ref[0]+(-0.996483829778)*x_ref[1]**o+(-0.564503119535)*x_ref[1]
        ref[5,0,1]=(0.602911018749)*x_ref[0]**o+(-0.348972500704)*x_ref[0]+(0.174681048071)*x_ref[1]**o+(-0.94551825287)*x_ref[1]
        ref[5,1,0]=(0.780926392259)*x_ref[0]**o+(-0.164977652547)*x_ref[0]+(0.224726701017)*x_ref[1]**o+(-0.779109318124)*x_ref[1]
        ref[5,1,1]=(0.470928696402)*x_ref[0]**o+(-0.505871268699)*x_ref[0]+(0.407158974816)*x_ref[1]**o+(0.392856659954)*x_ref[1]
      else:
        arg[0,0,0]=(0.759590374826)*x[0]**o+(-0.776182110267)*x[0]+(0.713871101133)*x[1]**o+(-0.0991655752543)*x[1]+(0.339491998788)*x[2]**o+(-0.624141085669)*x[2]
        arg[0,0,1]=(-0.948679111098)*x[0]**o+(0.0918139408797)*x[0]+(-0.519481784001)*x[1]**o+(-0.480754367303)*x[1]+(0.996421472879)*x[2]**o+(-0.933032853651)*x[2]
        arg[0,1,0]=(-0.824750788431)*x[0]**o+(0.846735099726)*x[0]+(-0.436835905895)*x[1]**o+(-0.732363323611)*x[1]+(-0.865610972073)*x[2]**o+(0.754759759319)*x[2]
        arg[0,1,1]=(-0.403783782613)*x[0]**o+(0.918875237045)*x[0]+(0.0934155750366)*x[1]**o+(0.832477256884)*x[1]+(0.615080190605)*x[2]**o+(0.000920074523901)*x[2]
        arg[1,0,0]=(0.260531941775)*x[0]**o+(-0.514647294043)*x[0]+(0.237113978946)*x[1]**o+(-0.116626766689)*x[1]+(-0.670422853592)*x[2]**o+(0.574941151693)*x[2]
        arg[1,0,1]=(-0.0133773984155)*x[0]**o+(0.286273972547)*x[0]+(0.0493985332233)*x[1]**o+(-0.567835671454)*x[1]+(0.508314781519)*x[2]**o+(-0.603506084784)*x[2]
        arg[1,1,0]=(0.0237413792908)*x[0]**o+(-0.758893281773)*x[0]+(0.150670689461)*x[1]**o+(0.136386516616)*x[1]+(-0.440203817756)*x[2]**o+(0.796610908529)*x[2]
        arg[1,1,1]=(-0.223549634917)*x[0]**o+(0.396472946824)*x[0]+(0.47830083818)*x[1]**o+(0.059933592631)*x[1]+(0.709371074379)*x[2]**o+(-0.129442424949)*x[2]
        arg[2,0,0]=(0.686170679191)*x[0]**o+(-0.118668021586)*x[0]+(0.033954449978)*x[1]**o+(-0.0508661210104)*x[1]+(0.352824724076)*x[2]**o+(-0.772874133692)*x[2]
        arg[2,0,1]=(0.434879772789)*x[0]**o+(-0.148309931356)*x[0]+(-0.013792688421)*x[1]**o+(0.597899709109)*x[1]+(-0.280336465565)*x[2]**o+(-0.177134087728)*x[2]
        arg[2,1,0]=(-0.636844238715)*x[0]**o+(-0.646269255495)*x[0]+(0.243236542558)*x[1]**o+(0.203480511801)*x[1]+(0.693290035164)*x[2]**o+(0.773354114662)*x[2]
        arg[2,1,1]=(-0.161998867128)*x[0]**o+(0.193608078775)*x[0]+(0.607954851405)*x[1]**o+(0.648078354644)*x[1]+(-0.488891750936)*x[2]**o+(0.912860597431)*x[2]
        arg[3,0,0]=(-0.696909279028)*x[0]**o+(0.039558795649)*x[0]+(0.441121646773)*x[1]**o+(-0.0681790925087)*x[1]+(-0.792437910441)*x[2]**o+(-0.512942395878)*x[2]
        arg[3,0,1]=(-0.264832490825)*x[0]**o+(-0.955976365474)*x[0]+(-0.884641387524)*x[1]**o+(-0.736970630699)*x[1]+(-0.0525378155791)*x[2]**o+(0.149164361449)*x[2]
        arg[3,1,0]=(0.434888951175)*x[0]**o+(-0.218511432369)*x[0]+(0.564377624401)*x[1]**o+(-0.0978716780106)*x[1]+(0.447857191989)*x[2]**o+(-0.466572586856)*x[2]
        arg[3,1,1]=(0.457689985738)*x[0]**o+(0.00322996620178)*x[0]+(0.307787601144)*x[1]**o+(-0.119455639704)*x[1]+(0.70367723461)*x[2]**o+(0.561913096313)*x[2]
        arg[4,0,0]=(-0.759802700221)*x[0]**o+(0.469649690329)*x[0]+(0.0913972297116)*x[1]**o+(0.310277900593)*x[1]+(-0.590266889319)*x[2]**o+(-0.89711266859)*x[2]
        arg[4,0,1]=(0.85506147544)*x[0]**o+(-0.357476737069)*x[0]+(-0.646898911915)*x[1]**o+(-0.674811608049)*x[1]+(0.865265982147)*x[2]**o+(0.720153126392)*x[2]
        arg[4,1,0]=(0.848146446583)*x[0]**o+(0.868166234685)*x[0]+(0.0477110054774)*x[1]**o+(-0.0338098833736)*x[1]+(-0.360024439914)*x[2]**o+(-0.634359083766)*x[2]
        arg[4,1,1]=(-0.672530624126)*x[0]**o+(0.354170615043)*x[0]+(0.486028226561)*x[1]**o+(0.857532578327)*x[1]+(0.15798171243)*x[2]**o+(0.0579857791862)*x[2]
        arg[5,0,0]=(0.653171799956)*x[0]**o+(0.605427034155)*x[0]+(0.7414424011)*x[1]**o+(-0.807505211605)*x[1]+(-0.0411776868808)*x[2]**o+(0.24301961063)*x[2]
        arg[5,0,1]=(-0.336860102074)*x[0]**o+(0.235697646908)*x[0]+(-0.173217953629)*x[1]**o+(0.989643649714)*x[1]+(0.977460054831)*x[2]**o+(-0.743825027125)*x[2]
        arg[5,1,0]=(0.794904875581)*x[0]**o+(0.607548574046)*x[0]+(0.603180304387)*x[1]**o+(0.276995429843)*x[1]+(0.698731315308)*x[2]**o+(-0.966744108727)*x[2]
        arg[5,1,1]=(-0.0230004766002)*x[0]**o+(0.489585418104)*x[0]+(0.382764947116)*x[1]**o+(0.143873532935)*x[1]+(0.932000957196)*x[2]**o+(0.739975795806)*x[2]
        ref[0,0,0]=(0.759590374826)*x_ref[0]**o+(-0.776182110267)*x_ref[0]+(0.713871101133)*x_ref[1]**o+(-0.0991655752543)*x_ref[1]+(0.339491998788)*x_ref[2]**o+(-0.624141085669)*x_ref[2]
        ref[0,0,1]=(-0.948679111098)*x_ref[0]**o+(0.0918139408797)*x_ref[0]+(-0.519481784001)*x_ref[1]**o+(-0.480754367303)*x_ref[1]+(0.996421472879)*x_ref[2]**o+(-0.933032853651)*x_ref[2]
        ref[0,1,0]=(-0.824750788431)*x_ref[0]**o+(0.846735099726)*x_ref[0]+(-0.436835905895)*x_ref[1]**o+(-0.732363323611)*x_ref[1]+(-0.865610972073)*x_ref[2]**o+(0.754759759319)*x_ref[2]
        ref[0,1,1]=(-0.403783782613)*x_ref[0]**o+(0.918875237045)*x_ref[0]+(0.0934155750366)*x_ref[1]**o+(0.832477256884)*x_ref[1]+(0.615080190605)*x_ref[2]**o+(0.000920074523901)*x_ref[2]
        ref[1,0,0]=(0.260531941775)*x_ref[0]**o+(-0.514647294043)*x_ref[0]+(0.237113978946)*x_ref[1]**o+(-0.116626766689)*x_ref[1]+(-0.670422853592)*x_ref[2]**o+(0.574941151693)*x_ref[2]
        ref[1,0,1]=(-0.0133773984155)*x_ref[0]**o+(0.286273972547)*x_ref[0]+(0.0493985332233)*x_ref[1]**o+(-0.567835671454)*x_ref[1]+(0.508314781519)*x_ref[2]**o+(-0.603506084784)*x_ref[2]
        ref[1,1,0]=(0.0237413792908)*x_ref[0]**o+(-0.758893281773)*x_ref[0]+(0.150670689461)*x_ref[1]**o+(0.136386516616)*x_ref[1]+(-0.440203817756)*x_ref[2]**o+(0.796610908529)*x_ref[2]
        ref[1,1,1]=(-0.223549634917)*x_ref[0]**o+(0.396472946824)*x_ref[0]+(0.47830083818)*x_ref[1]**o+(0.059933592631)*x_ref[1]+(0.709371074379)*x_ref[2]**o+(-0.129442424949)*x_ref[2]
        ref[2,0,0]=(0.686170679191)*x_ref[0]**o+(-0.118668021586)*x_ref[0]+(0.033954449978)*x_ref[1]**o+(-0.0508661210104)*x_ref[1]+(0.352824724076)*x_ref[2]**o+(-0.772874133692)*x_ref[2]
        ref[2,0,1]=(0.434879772789)*x_ref[0]**o+(-0.148309931356)*x_ref[0]+(-0.013792688421)*x_ref[1]**o+(0.597899709109)*x_ref[1]+(-0.280336465565)*x_ref[2]**o+(-0.177134087728)*x_ref[2]
        ref[2,1,0]=(-0.636844238715)*x_ref[0]**o+(-0.646269255495)*x_ref[0]+(0.243236542558)*x_ref[1]**o+(0.203480511801)*x_ref[1]+(0.693290035164)*x_ref[2]**o+(0.773354114662)*x_ref[2]
        ref[2,1,1]=(-0.161998867128)*x_ref[0]**o+(0.193608078775)*x_ref[0]+(0.607954851405)*x_ref[1]**o+(0.648078354644)*x_ref[1]+(-0.488891750936)*x_ref[2]**o+(0.912860597431)*x_ref[2]
        ref[3,0,0]=(-0.696909279028)*x_ref[0]**o+(0.039558795649)*x_ref[0]+(0.441121646773)*x_ref[1]**o+(-0.0681790925087)*x_ref[1]+(-0.792437910441)*x_ref[2]**o+(-0.512942395878)*x_ref[2]
        ref[3,0,1]=(-0.264832490825)*x_ref[0]**o+(-0.955976365474)*x_ref[0]+(-0.884641387524)*x_ref[1]**o+(-0.736970630699)*x_ref[1]+(-0.0525378155791)*x_ref[2]**o+(0.149164361449)*x_ref[2]
        ref[3,1,0]=(0.434888951175)*x_ref[0]**o+(-0.218511432369)*x_ref[0]+(0.564377624401)*x_ref[1]**o+(-0.0978716780106)*x_ref[1]+(0.447857191989)*x_ref[2]**o+(-0.466572586856)*x_ref[2]
        ref[3,1,1]=(0.457689985738)*x_ref[0]**o+(0.00322996620178)*x_ref[0]+(0.307787601144)*x_ref[1]**o+(-0.119455639704)*x_ref[1]+(0.70367723461)*x_ref[2]**o+(0.561913096313)*x_ref[2]
        ref[4,0,0]=(-0.759802700221)*x_ref[0]**o+(0.469649690329)*x_ref[0]+(0.0913972297116)*x_ref[1]**o+(0.310277900593)*x_ref[1]+(-0.590266889319)*x_ref[2]**o+(-0.89711266859)*x_ref[2]
        ref[4,0,1]=(0.85506147544)*x_ref[0]**o+(-0.357476737069)*x_ref[0]+(-0.646898911915)*x_ref[1]**o+(-0.674811608049)*x_ref[1]+(0.865265982147)*x_ref[2]**o+(0.720153126392)*x_ref[2]
        ref[4,1,0]=(0.848146446583)*x_ref[0]**o+(0.868166234685)*x_ref[0]+(0.0477110054774)*x_ref[1]**o+(-0.0338098833736)*x_ref[1]+(-0.360024439914)*x_ref[2]**o+(-0.634359083766)*x_ref[2]
        ref[4,1,1]=(-0.672530624126)*x_ref[0]**o+(0.354170615043)*x_ref[0]+(0.486028226561)*x_ref[1]**o+(0.857532578327)*x_ref[1]+(0.15798171243)*x_ref[2]**o+(0.0579857791862)*x_ref[2]
        ref[5,0,0]=(0.653171799956)*x_ref[0]**o+(0.605427034155)*x_ref[0]+(0.7414424011)*x_ref[1]**o+(-0.807505211605)*x_ref[1]+(-0.0411776868808)*x_ref[2]**o+(0.24301961063)*x_ref[2]
        ref[5,0,1]=(-0.336860102074)*x_ref[0]**o+(0.235697646908)*x_ref[0]+(-0.173217953629)*x_ref[1]**o+(0.989643649714)*x_ref[1]+(0.977460054831)*x_ref[2]**o+(-0.743825027125)*x_ref[2]
        ref[5,1,0]=(0.794904875581)*x_ref[0]**o+(0.607548574046)*x_ref[0]+(0.603180304387)*x_ref[1]**o+(0.276995429843)*x_ref[1]+(0.698731315308)*x_ref[2]**o+(-0.966744108727)*x_ref[2]
        ref[5,1,1]=(-0.0230004766002)*x_ref[0]**o+(0.489585418104)*x_ref[0]+(0.382764947116)*x_ref[1]**o+(0.143873532935)*x_ref[1]+(0.932000957196)*x_ref[2]**o+(0.739975795806)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.922177773381)*x[0]**o+(0.37324175781)*x[0]+(-0.169208776011)*x[1]**o+(0.72134249236)*x[1]
        arg[0,0,0,1]=(0.85312270129)*x[0]**o+(0.0512919651836)*x[0]+(0.968880249675)*x[1]**o+(0.740058819096)*x[1]
        arg[0,0,1,0]=(-0.018205657901)*x[0]**o+(0.601838710779)*x[0]+(0.211002282731)*x[1]**o+(0.801382123792)*x[1]
        arg[0,0,1,1]=(-0.249128714489)*x[0]**o+(-0.408809542089)*x[0]+(-0.684434920261)*x[1]**o+(0.223674849818)*x[1]
        arg[0,0,2,0]=(0.396627551183)*x[0]**o+(-0.223592778052)*x[0]+(0.227333834259)*x[1]**o+(0.66580805616)*x[1]
        arg[0,0,2,1]=(0.365216692731)*x[0]**o+(-0.670947594503)*x[0]+(0.400260365509)*x[1]**o+(0.10903299253)*x[1]
        arg[0,1,0,0]=(0.823297765985)*x[0]**o+(0.60337946486)*x[0]+(-0.314265083693)*x[1]**o+(-0.00636816227929)*x[1]
        arg[0,1,0,1]=(-0.26216625982)*x[0]**o+(0.767412606588)*x[0]+(-0.284917238441)*x[1]**o+(0.56385025143)*x[1]
        arg[0,1,1,0]=(-0.732326448281)*x[0]**o+(-0.669197323856)*x[0]+(-0.608751752289)*x[1]**o+(-0.853458985188)*x[1]
        arg[0,1,1,1]=(0.637275594334)*x[0]**o+(0.696570850488)*x[0]+(0.012026820657)*x[1]**o+(0.558317439186)*x[1]
        arg[0,1,2,0]=(0.925899092402)*x[0]**o+(0.568045453255)*x[0]+(-0.766701931715)*x[1]**o+(-0.95082586036)*x[1]
        arg[0,1,2,1]=(0.127524211574)*x[0]**o+(-0.65150542164)*x[0]+(-0.365773940068)*x[1]**o+(0.990465711754)*x[1]
        arg[0,2,0,0]=(0.456469052217)*x[0]**o+(0.403391798397)*x[0]+(-0.21921915949)*x[1]**o+(0.171493695945)*x[1]
        arg[0,2,0,1]=(-0.980242506743)*x[0]**o+(0.880225959645)*x[0]+(0.682653186307)*x[1]**o+(-0.561867916377)*x[1]
        arg[0,2,1,0]=(-0.329881998938)*x[0]**o+(0.464976480407)*x[0]+(-0.812583693347)*x[1]**o+(0.0103356333568)*x[1]
        arg[0,2,1,1]=(-0.710350374606)*x[0]**o+(-0.445981415127)*x[0]+(-0.159054554695)*x[1]**o+(-0.0306514318922)*x[1]
        arg[0,2,2,0]=(0.0903142465124)*x[0]**o+(0.872680271888)*x[0]+(0.230242338668)*x[1]**o+(0.557126778217)*x[1]
        arg[0,2,2,1]=(-0.842046475241)*x[0]**o+(0.350544067128)*x[0]+(0.486186516633)*x[1]**o+(-0.550801066658)*x[1]
        arg[0,3,0,0]=(0.740313870334)*x[0]**o+(-0.177105021967)*x[0]+(0.00114640526837)*x[1]**o+(0.349602955151)*x[1]
        arg[0,3,0,1]=(0.898906724506)*x[0]**o+(-0.882729119203)*x[0]+(0.000278645863091)*x[1]**o+(-0.450247260284)*x[1]
        arg[0,3,1,0]=(0.411917691981)*x[0]**o+(0.0461900014479)*x[0]+(0.1608400364)*x[1]**o+(0.130550626888)*x[1]
        arg[0,3,1,1]=(0.610378099059)*x[0]**o+(-0.903971943528)*x[0]+(-0.0170960427518)*x[1]**o+(0.765729630725)*x[1]
        arg[0,3,2,0]=(-0.811421756909)*x[0]**o+(-0.220065414196)*x[0]+(0.529467921065)*x[1]**o+(0.13896086667)*x[1]
        arg[0,3,2,1]=(-0.621021589728)*x[0]**o+(0.181903892108)*x[0]+(0.724296638902)*x[1]**o+(0.651705107858)*x[1]
        arg[0,4,0,0]=(0.323254788695)*x[0]**o+(0.758728125147)*x[0]+(0.0719056895797)*x[1]**o+(-0.486299017994)*x[1]
        arg[0,4,0,1]=(-0.426525215977)*x[0]**o+(-0.70344592953)*x[0]+(-0.467799544116)*x[1]**o+(0.0394017649712)*x[1]
        arg[0,4,1,0]=(-0.559984242808)*x[0]**o+(-0.963639424043)*x[0]+(0.884232194806)*x[1]**o+(0.948111505594)*x[1]
        arg[0,4,1,1]=(-0.825627656302)*x[0]**o+(0.167871540452)*x[0]+(0.0969950885663)*x[1]**o+(-0.477024385485)*x[1]
        arg[0,4,2,0]=(-0.890191205005)*x[0]**o+(-0.291891624031)*x[0]+(-0.924812934321)*x[1]**o+(-0.898567598264)*x[1]
        arg[0,4,2,1]=(0.0552241756422)*x[0]**o+(-0.47878688784)*x[0]+(-0.208893161197)*x[1]**o+(-0.671733441515)*x[1]
        arg[1,0,0,0]=(0.609961055126)*x[0]**o+(0.389225845794)*x[0]+(0.147977819654)*x[1]**o+(-0.75096024248)*x[1]
        arg[1,0,0,1]=(-0.407376527411)*x[0]**o+(-0.696660373977)*x[0]+(-0.803862178008)*x[1]**o+(-0.367087112506)*x[1]
        arg[1,0,1,0]=(0.89146134332)*x[0]**o+(0.388607608477)*x[0]+(0.763190367241)*x[1]**o+(0.169275499463)*x[1]
        arg[1,0,1,1]=(-0.870871474063)*x[0]**o+(-0.21526798243)*x[0]+(0.159168404985)*x[1]**o+(0.715049896538)*x[1]
        arg[1,0,2,0]=(-0.842978554795)*x[0]**o+(0.209649486001)*x[0]+(-0.767629750054)*x[1]**o+(-0.126634026874)*x[1]
        arg[1,0,2,1]=(-0.728571756044)*x[0]**o+(-0.247852313042)*x[0]+(0.818809803616)*x[1]**o+(0.736163163331)*x[1]
        arg[1,1,0,0]=(-0.670608139072)*x[0]**o+(0.330167259903)*x[0]+(0.275974484086)*x[1]**o+(0.492110644529)*x[1]
        arg[1,1,0,1]=(0.724664287226)*x[0]**o+(0.738855592065)*x[0]+(0.511369798951)*x[1]**o+(0.455794112842)*x[1]
        arg[1,1,1,0]=(0.783504599076)*x[0]**o+(-0.558579246854)*x[0]+(0.608044912637)*x[1]**o+(-0.401058633268)*x[1]
        arg[1,1,1,1]=(0.406697861117)*x[0]**o+(-0.31302718156)*x[0]+(-0.999553049864)*x[1]**o+(0.647698243869)*x[1]
        arg[1,1,2,0]=(-0.788346252365)*x[0]**o+(-0.809895265253)*x[0]+(0.441738551485)*x[1]**o+(-0.686785865679)*x[1]
        arg[1,1,2,1]=(0.794571346519)*x[0]**o+(-0.552071697712)*x[0]+(0.77260218173)*x[1]**o+(-0.969584549528)*x[1]
        arg[1,2,0,0]=(-0.867948229325)*x[0]**o+(-0.582387320474)*x[0]+(0.0387726756335)*x[1]**o+(0.4143770831)*x[1]
        arg[1,2,0,1]=(-0.913557875303)*x[0]**o+(-0.517970834439)*x[0]+(-0.922517499063)*x[1]**o+(0.858552054359)*x[1]
        arg[1,2,1,0]=(0.409302016181)*x[0]**o+(-0.684529177441)*x[0]+(0.818122023218)*x[1]**o+(-0.583589419623)*x[1]
        arg[1,2,1,1]=(0.750795883056)*x[0]**o+(-0.710882726624)*x[0]+(0.915210035315)*x[1]**o+(0.802588664027)*x[1]
        arg[1,2,2,0]=(0.639216566804)*x[0]**o+(-0.300954214125)*x[0]+(-0.26724393342)*x[1]**o+(0.426742851293)*x[1]
        arg[1,2,2,1]=(-0.00337538020121)*x[0]**o+(0.283012169215)*x[0]+(0.725040731001)*x[1]**o+(-0.708238265029)*x[1]
        arg[1,3,0,0]=(0.217404922432)*x[0]**o+(-0.590883003502)*x[0]+(0.451142849723)*x[1]**o+(0.624698767392)*x[1]
        arg[1,3,0,1]=(0.0351258739828)*x[0]**o+(-0.784068492686)*x[0]+(0.469956184167)*x[1]**o+(0.634124784387)*x[1]
        arg[1,3,1,0]=(0.998414775431)*x[0]**o+(0.388202946918)*x[0]+(-0.849223443255)*x[1]**o+(-0.804590028218)*x[1]
        arg[1,3,1,1]=(0.0219748851856)*x[0]**o+(0.464069403336)*x[0]+(-0.470759541095)*x[1]**o+(-0.21773573235)*x[1]
        arg[1,3,2,0]=(-0.879175905661)*x[0]**o+(0.100238280861)*x[0]+(0.331472085056)*x[1]**o+(0.0310947651268)*x[1]
        arg[1,3,2,1]=(-0.195155740038)*x[0]**o+(0.532078676231)*x[0]+(-0.863288082011)*x[1]**o+(-0.0905878788265)*x[1]
        arg[1,4,0,0]=(0.470883637487)*x[0]**o+(0.0291028245963)*x[0]+(-0.152462235474)*x[1]**o+(0.890883963301)*x[1]
        arg[1,4,0,1]=(-0.0780797534409)*x[0]**o+(-0.999649003413)*x[0]+(-0.79446765004)*x[1]**o+(-0.820520881428)*x[1]
        arg[1,4,1,0]=(-0.82479629638)*x[0]**o+(-0.28924829922)*x[0]+(-0.960430867949)*x[1]**o+(-0.514730558673)*x[1]
        arg[1,4,1,1]=(-0.0853911110391)*x[0]**o+(0.726747270178)*x[0]+(-0.70925711586)*x[1]**o+(-0.331433386035)*x[1]
        arg[1,4,2,0]=(-0.434350343402)*x[0]**o+(0.335341284415)*x[0]+(0.496818240476)*x[1]**o+(-0.925006191184)*x[1]
        arg[1,4,2,1]=(0.0753516047957)*x[0]**o+(0.171090641529)*x[0]+(-0.836002762987)*x[1]**o+(-0.753836171615)*x[1]
        arg[2,0,0,0]=(-0.331008073553)*x[0]**o+(-0.299051880715)*x[0]+(-0.84520124413)*x[1]**o+(0.780471473975)*x[1]
        arg[2,0,0,1]=(-0.947973851318)*x[0]**o+(0.579142212501)*x[0]+(0.137317760268)*x[1]**o+(-0.379509530229)*x[1]
        arg[2,0,1,0]=(-0.741320730287)*x[0]**o+(0.642110710824)*x[0]+(-0.632595944259)*x[1]**o+(0.944980046528)*x[1]
        arg[2,0,1,1]=(0.616943682269)*x[0]**o+(-0.348189507611)*x[0]+(-0.414080070092)*x[1]**o+(-0.733041081362)*x[1]
        arg[2,0,2,0]=(-0.021261659709)*x[0]**o+(0.790041129031)*x[0]+(0.788378634137)*x[1]**o+(0.391057560074)*x[1]
        arg[2,0,2,1]=(0.563737321295)*x[0]**o+(-0.110516278477)*x[0]+(0.383563652425)*x[1]**o+(-0.857100387101)*x[1]
        arg[2,1,0,0]=(-0.0839346911978)*x[0]**o+(0.699255867073)*x[0]+(-0.698488781835)*x[1]**o+(0.225940657322)*x[1]
        arg[2,1,0,1]=(0.93638601214)*x[0]**o+(-0.593516691147)*x[0]+(-0.445863239521)*x[1]**o+(0.601067357746)*x[1]
        arg[2,1,1,0]=(0.114814804735)*x[0]**o+(0.717600499991)*x[0]+(0.7371911818)*x[1]**o+(0.136012416552)*x[1]
        arg[2,1,1,1]=(0.794499830148)*x[0]**o+(0.0581947415855)*x[0]+(0.626382859976)*x[1]**o+(-0.668272098674)*x[1]
        arg[2,1,2,0]=(0.564628771073)*x[0]**o+(0.102820241267)*x[0]+(0.397960234363)*x[1]**o+(0.379632827896)*x[1]
        arg[2,1,2,1]=(-0.0880639775913)*x[0]**o+(0.325978376801)*x[0]+(-0.505498229733)*x[1]**o+(-0.595922646655)*x[1]
        arg[2,2,0,0]=(-0.132672737928)*x[0]**o+(-0.0291240195523)*x[0]+(-0.00825944211706)*x[1]**o+(0.985993963145)*x[1]
        arg[2,2,0,1]=(0.244321497515)*x[0]**o+(-0.111329005236)*x[0]+(0.769353471123)*x[1]**o+(-0.174257027136)*x[1]
        arg[2,2,1,0]=(0.252286330282)*x[0]**o+(0.450975615825)*x[0]+(-0.515016578606)*x[1]**o+(0.834736757555)*x[1]
        arg[2,2,1,1]=(0.0588273338962)*x[0]**o+(-0.865406338788)*x[0]+(0.482095645149)*x[1]**o+(-0.595126490156)*x[1]
        arg[2,2,2,0]=(0.244580063555)*x[0]**o+(0.576709884892)*x[0]+(-0.212616562122)*x[1]**o+(-0.407508050723)*x[1]
        arg[2,2,2,1]=(0.416825783755)*x[0]**o+(0.430959078988)*x[0]+(-0.398840301481)*x[1]**o+(0.889105964295)*x[1]
        arg[2,3,0,0]=(-0.731872136283)*x[0]**o+(0.237034931745)*x[0]+(0.547737117698)*x[1]**o+(0.364733861661)*x[1]
        arg[2,3,0,1]=(0.971716549689)*x[0]**o+(0.466300970081)*x[0]+(-0.51802162747)*x[1]**o+(-0.859910778018)*x[1]
        arg[2,3,1,0]=(-0.562918219893)*x[0]**o+(-0.0532549959391)*x[0]+(0.842033495359)*x[1]**o+(-0.708554283815)*x[1]
        arg[2,3,1,1]=(0.437125534104)*x[0]**o+(-0.376751615116)*x[0]+(-0.330889383623)*x[1]**o+(0.958884035289)*x[1]
        arg[2,3,2,0]=(-0.194290501566)*x[0]**o+(-0.368131904292)*x[0]+(0.802917650409)*x[1]**o+(0.628603987906)*x[1]
        arg[2,3,2,1]=(0.70219875881)*x[0]**o+(0.822181339005)*x[0]+(0.768635447234)*x[1]**o+(0.871676010439)*x[1]
        arg[2,4,0,0]=(-0.654498700662)*x[0]**o+(0.851143676147)*x[0]+(0.468468395415)*x[1]**o+(-0.447240713576)*x[1]
        arg[2,4,0,1]=(0.893177052112)*x[0]**o+(0.652743480988)*x[0]+(0.726646163682)*x[1]**o+(-0.854895731892)*x[1]
        arg[2,4,1,0]=(0.0540542301334)*x[0]**o+(-0.208654443984)*x[0]+(0.925127576853)*x[1]**o+(-0.932489814868)*x[1]
        arg[2,4,1,1]=(-0.460268214104)*x[0]**o+(-0.303569308977)*x[0]+(-0.381458880316)*x[1]**o+(0.690537661673)*x[1]
        arg[2,4,2,0]=(0.401012366366)*x[0]**o+(-0.970993441679)*x[0]+(0.0249535421254)*x[1]**o+(-0.499737254014)*x[1]
        arg[2,4,2,1]=(-0.692604790859)*x[0]**o+(-0.685292098725)*x[0]+(0.22550357354)*x[1]**o+(0.908152783462)*x[1]
        arg[3,0,0,0]=(-0.634246407254)*x[0]**o+(-0.571223033336)*x[0]+(0.0574319920181)*x[1]**o+(0.123763929908)*x[1]
        arg[3,0,0,1]=(-0.569358219213)*x[0]**o+(-0.107237509632)*x[0]+(0.97672999497)*x[1]**o+(-0.114939366813)*x[1]
        arg[3,0,1,0]=(-0.767885152479)*x[0]**o+(-0.105924102555)*x[0]+(0.539361607398)*x[1]**o+(0.546186446469)*x[1]
        arg[3,0,1,1]=(-0.58818696897)*x[0]**o+(0.629371516566)*x[0]+(-0.598173904376)*x[1]**o+(0.959043112088)*x[1]
        arg[3,0,2,0]=(0.751165393927)*x[0]**o+(0.732151619397)*x[0]+(-0.435276735735)*x[1]**o+(0.947142140722)*x[1]
        arg[3,0,2,1]=(0.463904349547)*x[0]**o+(-0.947270841014)*x[0]+(0.304239969784)*x[1]**o+(-0.233566944292)*x[1]
        arg[3,1,0,0]=(0.961122232395)*x[0]**o+(-0.989040215156)*x[0]+(0.833186746232)*x[1]**o+(0.226673105305)*x[1]
        arg[3,1,0,1]=(-0.156957616578)*x[0]**o+(-0.361665587314)*x[0]+(0.987490273947)*x[1]**o+(-0.890271058746)*x[1]
        arg[3,1,1,0]=(-0.259959222953)*x[0]**o+(-0.696135598749)*x[0]+(0.234411151165)*x[1]**o+(0.783105897256)*x[1]
        arg[3,1,1,1]=(0.020886322199)*x[0]**o+(0.887445555864)*x[0]+(0.709489180254)*x[1]**o+(0.990590112507)*x[1]
        arg[3,1,2,0]=(-0.963790247866)*x[0]**o+(-0.417180971158)*x[0]+(0.303859261646)*x[1]**o+(0.745552024399)*x[1]
        arg[3,1,2,1]=(-0.364448135641)*x[0]**o+(0.760841571414)*x[0]+(-0.0283374186763)*x[1]**o+(-0.491433855748)*x[1]
        arg[3,2,0,0]=(-0.649766142291)*x[0]**o+(-0.177410238297)*x[0]+(0.137316641289)*x[1]**o+(-0.40890299152)*x[1]
        arg[3,2,0,1]=(0.282922984657)*x[0]**o+(-0.660238152943)*x[0]+(0.481726052488)*x[1]**o+(0.976011254218)*x[1]
        arg[3,2,1,0]=(0.989761520473)*x[0]**o+(0.194082255647)*x[0]+(-0.648435682551)*x[1]**o+(-0.721569599311)*x[1]
        arg[3,2,1,1]=(0.734634540883)*x[0]**o+(0.524638529049)*x[0]+(-0.737636791717)*x[1]**o+(-0.995904525822)*x[1]
        arg[3,2,2,0]=(0.422396808663)*x[0]**o+(-0.820639154105)*x[0]+(-0.566207315373)*x[1]**o+(-0.224174711984)*x[1]
        arg[3,2,2,1]=(0.719145374997)*x[0]**o+(-0.885860499468)*x[0]+(0.184062878609)*x[1]**o+(0.47660543007)*x[1]
        arg[3,3,0,0]=(-0.0122290849915)*x[0]**o+(-0.0314298339876)*x[0]+(0.937255407134)*x[1]**o+(-0.754858547072)*x[1]
        arg[3,3,0,1]=(-0.507826956669)*x[0]**o+(0.951205020689)*x[0]+(0.837096245109)*x[1]**o+(-0.576692399518)*x[1]
        arg[3,3,1,0]=(0.602119808238)*x[0]**o+(-0.943937134298)*x[0]+(-0.244250254626)*x[1]**o+(-0.814103542727)*x[1]
        arg[3,3,1,1]=(-0.51811661822)*x[0]**o+(-0.188969078762)*x[0]+(-0.697470816837)*x[1]**o+(-0.741593014841)*x[1]
        arg[3,3,2,0]=(-0.673948933786)*x[0]**o+(0.816404336566)*x[0]+(0.339239090092)*x[1]**o+(-0.0173550947113)*x[1]
        arg[3,3,2,1]=(0.499109852716)*x[0]**o+(0.515501552893)*x[0]+(0.926442939974)*x[1]**o+(0.157227173254)*x[1]
        arg[3,4,0,0]=(-0.1837389273)*x[0]**o+(0.78490229721)*x[0]+(-0.248033788374)*x[1]**o+(0.735136386405)*x[1]
        arg[3,4,0,1]=(0.121169550194)*x[0]**o+(-0.854027732077)*x[0]+(0.937680848475)*x[1]**o+(-0.447184197264)*x[1]
        arg[3,4,1,0]=(-0.788610934694)*x[0]**o+(-0.397825388144)*x[0]+(0.604421361449)*x[1]**o+(0.937184552821)*x[1]
        arg[3,4,1,1]=(-0.493380143496)*x[0]**o+(0.958197000398)*x[0]+(0.282372561751)*x[1]**o+(-0.55322669826)*x[1]
        arg[3,4,2,0]=(0.11333865923)*x[0]**o+(-0.542805511569)*x[0]+(-0.640646701928)*x[1]**o+(-0.342030716597)*x[1]
        arg[3,4,2,1]=(0.505356266297)*x[0]**o+(-0.107422782294)*x[0]+(0.542833447968)*x[1]**o+(-0.615275716592)*x[1]
        ref[0,0,0,0]=(0.922177773381)*x_ref[0]**o+(0.37324175781)*x_ref[0]+(-0.169208776011)*x_ref[1]**o+(0.72134249236)*x_ref[1]
        ref[0,0,0,1]=(0.85312270129)*x_ref[0]**o+(0.0512919651836)*x_ref[0]+(0.968880249675)*x_ref[1]**o+(0.740058819096)*x_ref[1]
        ref[0,0,1,0]=(-0.018205657901)*x_ref[0]**o+(0.601838710779)*x_ref[0]+(0.211002282731)*x_ref[1]**o+(0.801382123792)*x_ref[1]
        ref[0,0,1,1]=(-0.249128714489)*x_ref[0]**o+(-0.408809542089)*x_ref[0]+(-0.684434920261)*x_ref[1]**o+(0.223674849818)*x_ref[1]
        ref[0,0,2,0]=(0.396627551183)*x_ref[0]**o+(-0.223592778052)*x_ref[0]+(0.227333834259)*x_ref[1]**o+(0.66580805616)*x_ref[1]
        ref[0,0,2,1]=(0.365216692731)*x_ref[0]**o+(-0.670947594503)*x_ref[0]+(0.400260365509)*x_ref[1]**o+(0.10903299253)*x_ref[1]
        ref[0,1,0,0]=(0.823297765985)*x_ref[0]**o+(0.60337946486)*x_ref[0]+(-0.314265083693)*x_ref[1]**o+(-0.00636816227929)*x_ref[1]
        ref[0,1,0,1]=(-0.26216625982)*x_ref[0]**o+(0.767412606588)*x_ref[0]+(-0.284917238441)*x_ref[1]**o+(0.56385025143)*x_ref[1]
        ref[0,1,1,0]=(-0.732326448281)*x_ref[0]**o+(-0.669197323856)*x_ref[0]+(-0.608751752289)*x_ref[1]**o+(-0.853458985188)*x_ref[1]
        ref[0,1,1,1]=(0.637275594334)*x_ref[0]**o+(0.696570850488)*x_ref[0]+(0.012026820657)*x_ref[1]**o+(0.558317439186)*x_ref[1]
        ref[0,1,2,0]=(0.925899092402)*x_ref[0]**o+(0.568045453255)*x_ref[0]+(-0.766701931715)*x_ref[1]**o+(-0.95082586036)*x_ref[1]
        ref[0,1,2,1]=(0.127524211574)*x_ref[0]**o+(-0.65150542164)*x_ref[0]+(-0.365773940068)*x_ref[1]**o+(0.990465711754)*x_ref[1]
        ref[0,2,0,0]=(0.456469052217)*x_ref[0]**o+(0.403391798397)*x_ref[0]+(-0.21921915949)*x_ref[1]**o+(0.171493695945)*x_ref[1]
        ref[0,2,0,1]=(-0.980242506743)*x_ref[0]**o+(0.880225959645)*x_ref[0]+(0.682653186307)*x_ref[1]**o+(-0.561867916377)*x_ref[1]
        ref[0,2,1,0]=(-0.329881998938)*x_ref[0]**o+(0.464976480407)*x_ref[0]+(-0.812583693347)*x_ref[1]**o+(0.0103356333568)*x_ref[1]
        ref[0,2,1,1]=(-0.710350374606)*x_ref[0]**o+(-0.445981415127)*x_ref[0]+(-0.159054554695)*x_ref[1]**o+(-0.0306514318922)*x_ref[1]
        ref[0,2,2,0]=(0.0903142465124)*x_ref[0]**o+(0.872680271888)*x_ref[0]+(0.230242338668)*x_ref[1]**o+(0.557126778217)*x_ref[1]
        ref[0,2,2,1]=(-0.842046475241)*x_ref[0]**o+(0.350544067128)*x_ref[0]+(0.486186516633)*x_ref[1]**o+(-0.550801066658)*x_ref[1]
        ref[0,3,0,0]=(0.740313870334)*x_ref[0]**o+(-0.177105021967)*x_ref[0]+(0.00114640526837)*x_ref[1]**o+(0.349602955151)*x_ref[1]
        ref[0,3,0,1]=(0.898906724506)*x_ref[0]**o+(-0.882729119203)*x_ref[0]+(0.000278645863091)*x_ref[1]**o+(-0.450247260284)*x_ref[1]
        ref[0,3,1,0]=(0.411917691981)*x_ref[0]**o+(0.0461900014479)*x_ref[0]+(0.1608400364)*x_ref[1]**o+(0.130550626888)*x_ref[1]
        ref[0,3,1,1]=(0.610378099059)*x_ref[0]**o+(-0.903971943528)*x_ref[0]+(-0.0170960427518)*x_ref[1]**o+(0.765729630725)*x_ref[1]
        ref[0,3,2,0]=(-0.811421756909)*x_ref[0]**o+(-0.220065414196)*x_ref[0]+(0.529467921065)*x_ref[1]**o+(0.13896086667)*x_ref[1]
        ref[0,3,2,1]=(-0.621021589728)*x_ref[0]**o+(0.181903892108)*x_ref[0]+(0.724296638902)*x_ref[1]**o+(0.651705107858)*x_ref[1]
        ref[0,4,0,0]=(0.323254788695)*x_ref[0]**o+(0.758728125147)*x_ref[0]+(0.0719056895797)*x_ref[1]**o+(-0.486299017994)*x_ref[1]
        ref[0,4,0,1]=(-0.426525215977)*x_ref[0]**o+(-0.70344592953)*x_ref[0]+(-0.467799544116)*x_ref[1]**o+(0.0394017649712)*x_ref[1]
        ref[0,4,1,0]=(-0.559984242808)*x_ref[0]**o+(-0.963639424043)*x_ref[0]+(0.884232194806)*x_ref[1]**o+(0.948111505594)*x_ref[1]
        ref[0,4,1,1]=(-0.825627656302)*x_ref[0]**o+(0.167871540452)*x_ref[0]+(0.0969950885663)*x_ref[1]**o+(-0.477024385485)*x_ref[1]
        ref[0,4,2,0]=(-0.890191205005)*x_ref[0]**o+(-0.291891624031)*x_ref[0]+(-0.924812934321)*x_ref[1]**o+(-0.898567598264)*x_ref[1]
        ref[0,4,2,1]=(0.0552241756422)*x_ref[0]**o+(-0.47878688784)*x_ref[0]+(-0.208893161197)*x_ref[1]**o+(-0.671733441515)*x_ref[1]
        ref[1,0,0,0]=(0.609961055126)*x_ref[0]**o+(0.389225845794)*x_ref[0]+(0.147977819654)*x_ref[1]**o+(-0.75096024248)*x_ref[1]
        ref[1,0,0,1]=(-0.407376527411)*x_ref[0]**o+(-0.696660373977)*x_ref[0]+(-0.803862178008)*x_ref[1]**o+(-0.367087112506)*x_ref[1]
        ref[1,0,1,0]=(0.89146134332)*x_ref[0]**o+(0.388607608477)*x_ref[0]+(0.763190367241)*x_ref[1]**o+(0.169275499463)*x_ref[1]
        ref[1,0,1,1]=(-0.870871474063)*x_ref[0]**o+(-0.21526798243)*x_ref[0]+(0.159168404985)*x_ref[1]**o+(0.715049896538)*x_ref[1]
        ref[1,0,2,0]=(-0.842978554795)*x_ref[0]**o+(0.209649486001)*x_ref[0]+(-0.767629750054)*x_ref[1]**o+(-0.126634026874)*x_ref[1]
        ref[1,0,2,1]=(-0.728571756044)*x_ref[0]**o+(-0.247852313042)*x_ref[0]+(0.818809803616)*x_ref[1]**o+(0.736163163331)*x_ref[1]
        ref[1,1,0,0]=(-0.670608139072)*x_ref[0]**o+(0.330167259903)*x_ref[0]+(0.275974484086)*x_ref[1]**o+(0.492110644529)*x_ref[1]
        ref[1,1,0,1]=(0.724664287226)*x_ref[0]**o+(0.738855592065)*x_ref[0]+(0.511369798951)*x_ref[1]**o+(0.455794112842)*x_ref[1]
        ref[1,1,1,0]=(0.783504599076)*x_ref[0]**o+(-0.558579246854)*x_ref[0]+(0.608044912637)*x_ref[1]**o+(-0.401058633268)*x_ref[1]
        ref[1,1,1,1]=(0.406697861117)*x_ref[0]**o+(-0.31302718156)*x_ref[0]+(-0.999553049864)*x_ref[1]**o+(0.647698243869)*x_ref[1]
        ref[1,1,2,0]=(-0.788346252365)*x_ref[0]**o+(-0.809895265253)*x_ref[0]+(0.441738551485)*x_ref[1]**o+(-0.686785865679)*x_ref[1]
        ref[1,1,2,1]=(0.794571346519)*x_ref[0]**o+(-0.552071697712)*x_ref[0]+(0.77260218173)*x_ref[1]**o+(-0.969584549528)*x_ref[1]
        ref[1,2,0,0]=(-0.867948229325)*x_ref[0]**o+(-0.582387320474)*x_ref[0]+(0.0387726756335)*x_ref[1]**o+(0.4143770831)*x_ref[1]
        ref[1,2,0,1]=(-0.913557875303)*x_ref[0]**o+(-0.517970834439)*x_ref[0]+(-0.922517499063)*x_ref[1]**o+(0.858552054359)*x_ref[1]
        ref[1,2,1,0]=(0.409302016181)*x_ref[0]**o+(-0.684529177441)*x_ref[0]+(0.818122023218)*x_ref[1]**o+(-0.583589419623)*x_ref[1]
        ref[1,2,1,1]=(0.750795883056)*x_ref[0]**o+(-0.710882726624)*x_ref[0]+(0.915210035315)*x_ref[1]**o+(0.802588664027)*x_ref[1]
        ref[1,2,2,0]=(0.639216566804)*x_ref[0]**o+(-0.300954214125)*x_ref[0]+(-0.26724393342)*x_ref[1]**o+(0.426742851293)*x_ref[1]
        ref[1,2,2,1]=(-0.00337538020121)*x_ref[0]**o+(0.283012169215)*x_ref[0]+(0.725040731001)*x_ref[1]**o+(-0.708238265029)*x_ref[1]
        ref[1,3,0,0]=(0.217404922432)*x_ref[0]**o+(-0.590883003502)*x_ref[0]+(0.451142849723)*x_ref[1]**o+(0.624698767392)*x_ref[1]
        ref[1,3,0,1]=(0.0351258739828)*x_ref[0]**o+(-0.784068492686)*x_ref[0]+(0.469956184167)*x_ref[1]**o+(0.634124784387)*x_ref[1]
        ref[1,3,1,0]=(0.998414775431)*x_ref[0]**o+(0.388202946918)*x_ref[0]+(-0.849223443255)*x_ref[1]**o+(-0.804590028218)*x_ref[1]
        ref[1,3,1,1]=(0.0219748851856)*x_ref[0]**o+(0.464069403336)*x_ref[0]+(-0.470759541095)*x_ref[1]**o+(-0.21773573235)*x_ref[1]
        ref[1,3,2,0]=(-0.879175905661)*x_ref[0]**o+(0.100238280861)*x_ref[0]+(0.331472085056)*x_ref[1]**o+(0.0310947651268)*x_ref[1]
        ref[1,3,2,1]=(-0.195155740038)*x_ref[0]**o+(0.532078676231)*x_ref[0]+(-0.863288082011)*x_ref[1]**o+(-0.0905878788265)*x_ref[1]
        ref[1,4,0,0]=(0.470883637487)*x_ref[0]**o+(0.0291028245963)*x_ref[0]+(-0.152462235474)*x_ref[1]**o+(0.890883963301)*x_ref[1]
        ref[1,4,0,1]=(-0.0780797534409)*x_ref[0]**o+(-0.999649003413)*x_ref[0]+(-0.79446765004)*x_ref[1]**o+(-0.820520881428)*x_ref[1]
        ref[1,4,1,0]=(-0.82479629638)*x_ref[0]**o+(-0.28924829922)*x_ref[0]+(-0.960430867949)*x_ref[1]**o+(-0.514730558673)*x_ref[1]
        ref[1,4,1,1]=(-0.0853911110391)*x_ref[0]**o+(0.726747270178)*x_ref[0]+(-0.70925711586)*x_ref[1]**o+(-0.331433386035)*x_ref[1]
        ref[1,4,2,0]=(-0.434350343402)*x_ref[0]**o+(0.335341284415)*x_ref[0]+(0.496818240476)*x_ref[1]**o+(-0.925006191184)*x_ref[1]
        ref[1,4,2,1]=(0.0753516047957)*x_ref[0]**o+(0.171090641529)*x_ref[0]+(-0.836002762987)*x_ref[1]**o+(-0.753836171615)*x_ref[1]
        ref[2,0,0,0]=(-0.331008073553)*x_ref[0]**o+(-0.299051880715)*x_ref[0]+(-0.84520124413)*x_ref[1]**o+(0.780471473975)*x_ref[1]
        ref[2,0,0,1]=(-0.947973851318)*x_ref[0]**o+(0.579142212501)*x_ref[0]+(0.137317760268)*x_ref[1]**o+(-0.379509530229)*x_ref[1]
        ref[2,0,1,0]=(-0.741320730287)*x_ref[0]**o+(0.642110710824)*x_ref[0]+(-0.632595944259)*x_ref[1]**o+(0.944980046528)*x_ref[1]
        ref[2,0,1,1]=(0.616943682269)*x_ref[0]**o+(-0.348189507611)*x_ref[0]+(-0.414080070092)*x_ref[1]**o+(-0.733041081362)*x_ref[1]
        ref[2,0,2,0]=(-0.021261659709)*x_ref[0]**o+(0.790041129031)*x_ref[0]+(0.788378634137)*x_ref[1]**o+(0.391057560074)*x_ref[1]
        ref[2,0,2,1]=(0.563737321295)*x_ref[0]**o+(-0.110516278477)*x_ref[0]+(0.383563652425)*x_ref[1]**o+(-0.857100387101)*x_ref[1]
        ref[2,1,0,0]=(-0.0839346911978)*x_ref[0]**o+(0.699255867073)*x_ref[0]+(-0.698488781835)*x_ref[1]**o+(0.225940657322)*x_ref[1]
        ref[2,1,0,1]=(0.93638601214)*x_ref[0]**o+(-0.593516691147)*x_ref[0]+(-0.445863239521)*x_ref[1]**o+(0.601067357746)*x_ref[1]
        ref[2,1,1,0]=(0.114814804735)*x_ref[0]**o+(0.717600499991)*x_ref[0]+(0.7371911818)*x_ref[1]**o+(0.136012416552)*x_ref[1]
        ref[2,1,1,1]=(0.794499830148)*x_ref[0]**o+(0.0581947415855)*x_ref[0]+(0.626382859976)*x_ref[1]**o+(-0.668272098674)*x_ref[1]
        ref[2,1,2,0]=(0.564628771073)*x_ref[0]**o+(0.102820241267)*x_ref[0]+(0.397960234363)*x_ref[1]**o+(0.379632827896)*x_ref[1]
        ref[2,1,2,1]=(-0.0880639775913)*x_ref[0]**o+(0.325978376801)*x_ref[0]+(-0.505498229733)*x_ref[1]**o+(-0.595922646655)*x_ref[1]
        ref[2,2,0,0]=(-0.132672737928)*x_ref[0]**o+(-0.0291240195523)*x_ref[0]+(-0.00825944211706)*x_ref[1]**o+(0.985993963145)*x_ref[1]
        ref[2,2,0,1]=(0.244321497515)*x_ref[0]**o+(-0.111329005236)*x_ref[0]+(0.769353471123)*x_ref[1]**o+(-0.174257027136)*x_ref[1]
        ref[2,2,1,0]=(0.252286330282)*x_ref[0]**o+(0.450975615825)*x_ref[0]+(-0.515016578606)*x_ref[1]**o+(0.834736757555)*x_ref[1]
        ref[2,2,1,1]=(0.0588273338962)*x_ref[0]**o+(-0.865406338788)*x_ref[0]+(0.482095645149)*x_ref[1]**o+(-0.595126490156)*x_ref[1]
        ref[2,2,2,0]=(0.244580063555)*x_ref[0]**o+(0.576709884892)*x_ref[0]+(-0.212616562122)*x_ref[1]**o+(-0.407508050723)*x_ref[1]
        ref[2,2,2,1]=(0.416825783755)*x_ref[0]**o+(0.430959078988)*x_ref[0]+(-0.398840301481)*x_ref[1]**o+(0.889105964295)*x_ref[1]
        ref[2,3,0,0]=(-0.731872136283)*x_ref[0]**o+(0.237034931745)*x_ref[0]+(0.547737117698)*x_ref[1]**o+(0.364733861661)*x_ref[1]
        ref[2,3,0,1]=(0.971716549689)*x_ref[0]**o+(0.466300970081)*x_ref[0]+(-0.51802162747)*x_ref[1]**o+(-0.859910778018)*x_ref[1]
        ref[2,3,1,0]=(-0.562918219893)*x_ref[0]**o+(-0.0532549959391)*x_ref[0]+(0.842033495359)*x_ref[1]**o+(-0.708554283815)*x_ref[1]
        ref[2,3,1,1]=(0.437125534104)*x_ref[0]**o+(-0.376751615116)*x_ref[0]+(-0.330889383623)*x_ref[1]**o+(0.958884035289)*x_ref[1]
        ref[2,3,2,0]=(-0.194290501566)*x_ref[0]**o+(-0.368131904292)*x_ref[0]+(0.802917650409)*x_ref[1]**o+(0.628603987906)*x_ref[1]
        ref[2,3,2,1]=(0.70219875881)*x_ref[0]**o+(0.822181339005)*x_ref[0]+(0.768635447234)*x_ref[1]**o+(0.871676010439)*x_ref[1]
        ref[2,4,0,0]=(-0.654498700662)*x_ref[0]**o+(0.851143676147)*x_ref[0]+(0.468468395415)*x_ref[1]**o+(-0.447240713576)*x_ref[1]
        ref[2,4,0,1]=(0.893177052112)*x_ref[0]**o+(0.652743480988)*x_ref[0]+(0.726646163682)*x_ref[1]**o+(-0.854895731892)*x_ref[1]
        ref[2,4,1,0]=(0.0540542301334)*x_ref[0]**o+(-0.208654443984)*x_ref[0]+(0.925127576853)*x_ref[1]**o+(-0.932489814868)*x_ref[1]
        ref[2,4,1,1]=(-0.460268214104)*x_ref[0]**o+(-0.303569308977)*x_ref[0]+(-0.381458880316)*x_ref[1]**o+(0.690537661673)*x_ref[1]
        ref[2,4,2,0]=(0.401012366366)*x_ref[0]**o+(-0.970993441679)*x_ref[0]+(0.0249535421254)*x_ref[1]**o+(-0.499737254014)*x_ref[1]
        ref[2,4,2,1]=(-0.692604790859)*x_ref[0]**o+(-0.685292098725)*x_ref[0]+(0.22550357354)*x_ref[1]**o+(0.908152783462)*x_ref[1]
        ref[3,0,0,0]=(-0.634246407254)*x_ref[0]**o+(-0.571223033336)*x_ref[0]+(0.0574319920181)*x_ref[1]**o+(0.123763929908)*x_ref[1]
        ref[3,0,0,1]=(-0.569358219213)*x_ref[0]**o+(-0.107237509632)*x_ref[0]+(0.97672999497)*x_ref[1]**o+(-0.114939366813)*x_ref[1]
        ref[3,0,1,0]=(-0.767885152479)*x_ref[0]**o+(-0.105924102555)*x_ref[0]+(0.539361607398)*x_ref[1]**o+(0.546186446469)*x_ref[1]
        ref[3,0,1,1]=(-0.58818696897)*x_ref[0]**o+(0.629371516566)*x_ref[0]+(-0.598173904376)*x_ref[1]**o+(0.959043112088)*x_ref[1]
        ref[3,0,2,0]=(0.751165393927)*x_ref[0]**o+(0.732151619397)*x_ref[0]+(-0.435276735735)*x_ref[1]**o+(0.947142140722)*x_ref[1]
        ref[3,0,2,1]=(0.463904349547)*x_ref[0]**o+(-0.947270841014)*x_ref[0]+(0.304239969784)*x_ref[1]**o+(-0.233566944292)*x_ref[1]
        ref[3,1,0,0]=(0.961122232395)*x_ref[0]**o+(-0.989040215156)*x_ref[0]+(0.833186746232)*x_ref[1]**o+(0.226673105305)*x_ref[1]
        ref[3,1,0,1]=(-0.156957616578)*x_ref[0]**o+(-0.361665587314)*x_ref[0]+(0.987490273947)*x_ref[1]**o+(-0.890271058746)*x_ref[1]
        ref[3,1,1,0]=(-0.259959222953)*x_ref[0]**o+(-0.696135598749)*x_ref[0]+(0.234411151165)*x_ref[1]**o+(0.783105897256)*x_ref[1]
        ref[3,1,1,1]=(0.020886322199)*x_ref[0]**o+(0.887445555864)*x_ref[0]+(0.709489180254)*x_ref[1]**o+(0.990590112507)*x_ref[1]
        ref[3,1,2,0]=(-0.963790247866)*x_ref[0]**o+(-0.417180971158)*x_ref[0]+(0.303859261646)*x_ref[1]**o+(0.745552024399)*x_ref[1]
        ref[3,1,2,1]=(-0.364448135641)*x_ref[0]**o+(0.760841571414)*x_ref[0]+(-0.0283374186763)*x_ref[1]**o+(-0.491433855748)*x_ref[1]
        ref[3,2,0,0]=(-0.649766142291)*x_ref[0]**o+(-0.177410238297)*x_ref[0]+(0.137316641289)*x_ref[1]**o+(-0.40890299152)*x_ref[1]
        ref[3,2,0,1]=(0.282922984657)*x_ref[0]**o+(-0.660238152943)*x_ref[0]+(0.481726052488)*x_ref[1]**o+(0.976011254218)*x_ref[1]
        ref[3,2,1,0]=(0.989761520473)*x_ref[0]**o+(0.194082255647)*x_ref[0]+(-0.648435682551)*x_ref[1]**o+(-0.721569599311)*x_ref[1]
        ref[3,2,1,1]=(0.734634540883)*x_ref[0]**o+(0.524638529049)*x_ref[0]+(-0.737636791717)*x_ref[1]**o+(-0.995904525822)*x_ref[1]
        ref[3,2,2,0]=(0.422396808663)*x_ref[0]**o+(-0.820639154105)*x_ref[0]+(-0.566207315373)*x_ref[1]**o+(-0.224174711984)*x_ref[1]
        ref[3,2,2,1]=(0.719145374997)*x_ref[0]**o+(-0.885860499468)*x_ref[0]+(0.184062878609)*x_ref[1]**o+(0.47660543007)*x_ref[1]
        ref[3,3,0,0]=(-0.0122290849915)*x_ref[0]**o+(-0.0314298339876)*x_ref[0]+(0.937255407134)*x_ref[1]**o+(-0.754858547072)*x_ref[1]
        ref[3,3,0,1]=(-0.507826956669)*x_ref[0]**o+(0.951205020689)*x_ref[0]+(0.837096245109)*x_ref[1]**o+(-0.576692399518)*x_ref[1]
        ref[3,3,1,0]=(0.602119808238)*x_ref[0]**o+(-0.943937134298)*x_ref[0]+(-0.244250254626)*x_ref[1]**o+(-0.814103542727)*x_ref[1]
        ref[3,3,1,1]=(-0.51811661822)*x_ref[0]**o+(-0.188969078762)*x_ref[0]+(-0.697470816837)*x_ref[1]**o+(-0.741593014841)*x_ref[1]
        ref[3,3,2,0]=(-0.673948933786)*x_ref[0]**o+(0.816404336566)*x_ref[0]+(0.339239090092)*x_ref[1]**o+(-0.0173550947113)*x_ref[1]
        ref[3,3,2,1]=(0.499109852716)*x_ref[0]**o+(0.515501552893)*x_ref[0]+(0.926442939974)*x_ref[1]**o+(0.157227173254)*x_ref[1]
        ref[3,4,0,0]=(-0.1837389273)*x_ref[0]**o+(0.78490229721)*x_ref[0]+(-0.248033788374)*x_ref[1]**o+(0.735136386405)*x_ref[1]
        ref[3,4,0,1]=(0.121169550194)*x_ref[0]**o+(-0.854027732077)*x_ref[0]+(0.937680848475)*x_ref[1]**o+(-0.447184197264)*x_ref[1]
        ref[3,4,1,0]=(-0.788610934694)*x_ref[0]**o+(-0.397825388144)*x_ref[0]+(0.604421361449)*x_ref[1]**o+(0.937184552821)*x_ref[1]
        ref[3,4,1,1]=(-0.493380143496)*x_ref[0]**o+(0.958197000398)*x_ref[0]+(0.282372561751)*x_ref[1]**o+(-0.55322669826)*x_ref[1]
        ref[3,4,2,0]=(0.11333865923)*x_ref[0]**o+(-0.542805511569)*x_ref[0]+(-0.640646701928)*x_ref[1]**o+(-0.342030716597)*x_ref[1]
        ref[3,4,2,1]=(0.505356266297)*x_ref[0]**o+(-0.107422782294)*x_ref[0]+(0.542833447968)*x_ref[1]**o+(-0.615275716592)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.507422888369)*x[0]**o+(0.808798513305)*x[0]+(0.51493073805)*x[1]**o+(-0.928527686947)*x[1]+(-0.788010447498)*x[2]**o+(-0.378389833131)*x[2]
        arg[0,0,0,1]=(-0.475851243177)*x[0]**o+(0.746578446988)*x[0]+(0.504778622494)*x[1]**o+(-0.677770398899)*x[1]+(-0.862535648403)*x[2]**o+(0.620414141691)*x[2]
        arg[0,0,1,0]=(-0.361726879582)*x[0]**o+(0.554584127587)*x[0]+(0.436610414772)*x[1]**o+(-0.686534492717)*x[1]+(0.757226061419)*x[2]**o+(0.0948537725389)*x[2]
        arg[0,0,1,1]=(-0.362160696355)*x[0]**o+(0.766524051703)*x[0]+(0.209980857795)*x[1]**o+(-0.478398061437)*x[1]+(-0.410960179122)*x[2]**o+(-0.213272138462)*x[2]
        arg[0,0,2,0]=(0.87154872554)*x[0]**o+(-0.727915301904)*x[0]+(0.683580698056)*x[1]**o+(0.215052521277)*x[1]+(0.267231507898)*x[2]**o+(-0.0124189912449)*x[2]
        arg[0,0,2,1]=(0.631400944876)*x[0]**o+(0.537293628606)*x[0]+(0.944754350994)*x[1]**o+(0.108068908012)*x[1]+(0.890117096365)*x[2]**o+(0.164706867235)*x[2]
        arg[0,1,0,0]=(-0.935871759207)*x[0]**o+(0.350133758077)*x[0]+(-0.966482181015)*x[1]**o+(0.440902701528)*x[1]+(-0.893329541322)*x[2]**o+(0.350317827938)*x[2]
        arg[0,1,0,1]=(0.465305203601)*x[0]**o+(0.711131797928)*x[0]+(0.104091536498)*x[1]**o+(0.484811628127)*x[1]+(0.676878954516)*x[2]**o+(-0.99327016995)*x[2]
        arg[0,1,1,0]=(-0.361189016151)*x[0]**o+(0.962468573492)*x[0]+(-0.158571122564)*x[1]**o+(-0.293583660912)*x[1]+(-0.713699173608)*x[2]**o+(-0.507430163361)*x[2]
        arg[0,1,1,1]=(0.172564527054)*x[0]**o+(0.273438145027)*x[0]+(0.137378266239)*x[1]**o+(-0.675788606841)*x[1]+(-0.362143416121)*x[2]**o+(0.0703897749775)*x[2]
        arg[0,1,2,0]=(-0.941374717271)*x[0]**o+(0.289433165458)*x[0]+(0.318615227028)*x[1]**o+(0.524362568204)*x[1]+(-0.612590305095)*x[2]**o+(-0.117555538957)*x[2]
        arg[0,1,2,1]=(-0.442920825551)*x[0]**o+(-0.454982146579)*x[0]+(0.355353737589)*x[1]**o+(-0.729623917004)*x[1]+(-0.567220335937)*x[2]**o+(0.977652639976)*x[2]
        arg[0,2,0,0]=(0.942485930373)*x[0]**o+(-0.605823961033)*x[0]+(-0.143465779338)*x[1]**o+(0.327354308987)*x[1]+(0.691900417542)*x[2]**o+(-0.514165213849)*x[2]
        arg[0,2,0,1]=(0.018029272977)*x[0]**o+(-0.209992705469)*x[0]+(-0.224614727554)*x[1]**o+(-0.862697462048)*x[1]+(0.274791387928)*x[2]**o+(-0.907445869506)*x[2]
        arg[0,2,1,0]=(0.862755248265)*x[0]**o+(0.020831641939)*x[0]+(0.496662599261)*x[1]**o+(0.79318975953)*x[1]+(0.0912166895541)*x[2]**o+(0.556337610316)*x[2]
        arg[0,2,1,1]=(-0.09652997071)*x[0]**o+(-0.954279421394)*x[0]+(-0.601053524018)*x[1]**o+(0.642528012068)*x[1]+(0.295748113888)*x[2]**o+(-0.318166346186)*x[2]
        arg[0,2,2,0]=(-0.914878854129)*x[0]**o+(0.159201123642)*x[0]+(0.589819056368)*x[1]**o+(-0.715763226823)*x[1]+(-0.444572583567)*x[2]**o+(-0.487866651367)*x[2]
        arg[0,2,2,1]=(-0.71271969175)*x[0]**o+(0.454616308594)*x[0]+(0.510270095021)*x[1]**o+(-0.0385822967955)*x[1]+(-0.519736540313)*x[2]**o+(0.0556733091556)*x[2]
        arg[0,3,0,0]=(0.355768034846)*x[0]**o+(0.152240594627)*x[0]+(-0.81081131843)*x[1]**o+(-0.913373483156)*x[1]+(-0.319859744799)*x[2]**o+(-0.306218435393)*x[2]
        arg[0,3,0,1]=(-0.656932063693)*x[0]**o+(0.832886861783)*x[0]+(-0.417112750683)*x[1]**o+(-0.388190265957)*x[1]+(0.246534947096)*x[2]**o+(-0.896822839189)*x[2]
        arg[0,3,1,0]=(0.204713418891)*x[0]**o+(-0.150097488817)*x[0]+(0.806578426941)*x[1]**o+(-0.590052708995)*x[1]+(-0.758958608419)*x[2]**o+(-0.221973528232)*x[2]
        arg[0,3,1,1]=(-0.854308184618)*x[0]**o+(-0.292526203646)*x[0]+(-0.662403504833)*x[1]**o+(0.0955657982227)*x[1]+(0.69471076173)*x[2]**o+(0.384077309264)*x[2]
        arg[0,3,2,0]=(0.982057200864)*x[0]**o+(-0.64758727704)*x[0]+(0.15941493222)*x[1]**o+(-0.17745437115)*x[1]+(-0.72953252955)*x[2]**o+(0.986979270802)*x[2]
        arg[0,3,2,1]=(0.764526698779)*x[0]**o+(0.22809521115)*x[0]+(-0.0762092087998)*x[1]**o+(0.415212410471)*x[1]+(-0.263638988463)*x[2]**o+(-0.934669942336)*x[2]
        arg[0,4,0,0]=(0.0812656932707)*x[0]**o+(0.0397235062518)*x[0]+(-0.0496912823983)*x[1]**o+(0.189891585189)*x[1]+(0.243537823293)*x[2]**o+(0.461256134208)*x[2]
        arg[0,4,0,1]=(0.167596358861)*x[0]**o+(0.441650690481)*x[0]+(0.638100632205)*x[1]**o+(-0.69240615765)*x[1]+(-0.501855369512)*x[2]**o+(0.769465683782)*x[2]
        arg[0,4,1,0]=(-0.137149398927)*x[0]**o+(-0.186548325285)*x[0]+(-0.504656948352)*x[1]**o+(0.228896656134)*x[1]+(-0.472425358061)*x[2]**o+(-0.0769834754996)*x[2]
        arg[0,4,1,1]=(0.311655648626)*x[0]**o+(0.856513683959)*x[0]+(0.47173626537)*x[1]**o+(0.245497333467)*x[1]+(-0.513727691305)*x[2]**o+(-0.588614785682)*x[2]
        arg[0,4,2,0]=(-0.152907854625)*x[0]**o+(-0.86229950575)*x[0]+(-0.931356414564)*x[1]**o+(-0.915667991864)*x[1]+(0.322316118527)*x[2]**o+(-0.987559462539)*x[2]
        arg[0,4,2,1]=(-0.705947769164)*x[0]**o+(0.104081366627)*x[0]+(-0.540454644392)*x[1]**o+(0.404522880836)*x[1]+(0.146464264141)*x[2]**o+(0.586755030006)*x[2]
        arg[1,0,0,0]=(-0.466614768106)*x[0]**o+(0.482681414943)*x[0]+(-0.176553738108)*x[1]**o+(-0.39162272813)*x[1]+(0.74312023324)*x[2]**o+(0.824538905894)*x[2]
        arg[1,0,0,1]=(-0.9351073631)*x[0]**o+(-0.640226640449)*x[0]+(0.966208762054)*x[1]**o+(0.101792167944)*x[1]+(-0.732983657038)*x[2]**o+(-0.112313375937)*x[2]
        arg[1,0,1,0]=(-0.305403470434)*x[0]**o+(-0.414397930347)*x[0]+(0.356093254322)*x[1]**o+(-0.52607560749)*x[1]+(-0.70070674473)*x[2]**o+(-0.0192901616015)*x[2]
        arg[1,0,1,1]=(0.450046989211)*x[0]**o+(0.602770057268)*x[0]+(0.0193812509113)*x[1]**o+(-0.275615149191)*x[1]+(0.0180468267836)*x[2]**o+(-0.81388255113)*x[2]
        arg[1,0,2,0]=(0.913651483249)*x[0]**o+(-0.92188313743)*x[0]+(0.680007150195)*x[1]**o+(-0.300145584708)*x[1]+(-0.398883520002)*x[2]**o+(-0.752781496463)*x[2]
        arg[1,0,2,1]=(-0.265749827845)*x[0]**o+(0.594379728352)*x[0]+(0.463409943587)*x[1]**o+(-0.85840420426)*x[1]+(-0.705963756674)*x[2]**o+(0.251711112887)*x[2]
        arg[1,1,0,0]=(-0.28895690781)*x[0]**o+(0.476461656166)*x[0]+(0.0692567954597)*x[1]**o+(0.88763437878)*x[1]+(0.0872256197554)*x[2]**o+(0.19467039694)*x[2]
        arg[1,1,0,1]=(0.741045418981)*x[0]**o+(-0.726935151332)*x[0]+(0.899418662131)*x[1]**o+(0.27516497286)*x[1]+(-0.715927474971)*x[2]**o+(-0.709283098732)*x[2]
        arg[1,1,1,0]=(0.991407130174)*x[0]**o+(-0.163484265852)*x[0]+(-0.983254548659)*x[1]**o+(0.0568443855477)*x[1]+(0.309664644588)*x[2]**o+(0.904334718845)*x[2]
        arg[1,1,1,1]=(0.237261672833)*x[0]**o+(0.0861047746345)*x[0]+(-0.494898018662)*x[1]**o+(-0.433075334646)*x[1]+(0.588597346505)*x[2]**o+(-0.185559567553)*x[2]
        arg[1,1,2,0]=(-0.969384091709)*x[0]**o+(-0.0229777223491)*x[0]+(0.0945623059666)*x[1]**o+(0.303543647034)*x[1]+(0.659000135381)*x[2]**o+(0.00873791711254)*x[2]
        arg[1,1,2,1]=(-0.421406731882)*x[0]**o+(-0.5017394377)*x[0]+(0.027747130694)*x[1]**o+(0.327278873996)*x[1]+(-0.754133439304)*x[2]**o+(0.673525803845)*x[2]
        arg[1,2,0,0]=(-0.104620319685)*x[0]**o+(-0.685602393278)*x[0]+(0.972173717014)*x[1]**o+(-0.440573017063)*x[1]+(-0.709529395743)*x[2]**o+(0.634170741457)*x[2]
        arg[1,2,0,1]=(-0.560386754264)*x[0]**o+(0.570894623654)*x[0]+(-0.455122833362)*x[1]**o+(-0.00322505937935)*x[1]+(0.959400336149)*x[2]**o+(0.64014859954)*x[2]
        arg[1,2,1,0]=(0.100585233221)*x[0]**o+(0.939532360957)*x[0]+(-0.900301955112)*x[1]**o+(-0.776375896786)*x[1]+(-0.779373294746)*x[2]**o+(0.546157461116)*x[2]
        arg[1,2,1,1]=(-0.680457315071)*x[0]**o+(-0.0278844539208)*x[0]+(-0.629673590044)*x[1]**o+(0.813218376348)*x[1]+(-0.449484131841)*x[2]**o+(0.641682150048)*x[2]
        arg[1,2,2,0]=(0.60056954386)*x[0]**o+(-0.763169434657)*x[0]+(-0.672890915035)*x[1]**o+(-0.368711154571)*x[1]+(0.476810195318)*x[2]**o+(-0.935071124382)*x[2]
        arg[1,2,2,1]=(0.994100148333)*x[0]**o+(0.471647033381)*x[0]+(0.341677296654)*x[1]**o+(0.466442262068)*x[1]+(-0.0634919249117)*x[2]**o+(0.0637348830912)*x[2]
        arg[1,3,0,0]=(-0.120744286726)*x[0]**o+(0.630378984537)*x[0]+(0.0837817796507)*x[1]**o+(-0.850115274881)*x[1]+(0.574342626561)*x[2]**o+(-0.705409915626)*x[2]
        arg[1,3,0,1]=(-0.51181061391)*x[0]**o+(0.711479747616)*x[0]+(-0.00800928076734)*x[1]**o+(-0.703424580277)*x[1]+(-0.925935026157)*x[2]**o+(0.488175992827)*x[2]
        arg[1,3,1,0]=(0.531879615751)*x[0]**o+(-0.214270521934)*x[0]+(-0.432344726015)*x[1]**o+(-0.107660826449)*x[1]+(0.763406427398)*x[2]**o+(-0.511000150429)*x[2]
        arg[1,3,1,1]=(-0.867467568206)*x[0]**o+(-0.738425961596)*x[0]+(-0.925337667706)*x[1]**o+(-0.586227765107)*x[1]+(0.909987211462)*x[2]**o+(0.533347022582)*x[2]
        arg[1,3,2,0]=(-0.47713113862)*x[0]**o+(-0.130876138883)*x[0]+(-0.607996835981)*x[1]**o+(-0.0306261551172)*x[1]+(0.956207500475)*x[2]**o+(-0.287731719382)*x[2]
        arg[1,3,2,1]=(0.179479486943)*x[0]**o+(0.361266562479)*x[0]+(0.374904472267)*x[1]**o+(0.00788379106301)*x[1]+(-0.34090684265)*x[2]**o+(-0.190875910791)*x[2]
        arg[1,4,0,0]=(0.685464203729)*x[0]**o+(0.0634561930987)*x[0]+(-0.662534110385)*x[1]**o+(0.383849898086)*x[1]+(0.948230524205)*x[2]**o+(-0.347042281976)*x[2]
        arg[1,4,0,1]=(-0.484499755322)*x[0]**o+(-0.765655349243)*x[0]+(0.805331514606)*x[1]**o+(-0.0635912547496)*x[1]+(-0.139457134772)*x[2]**o+(0.227446734221)*x[2]
        arg[1,4,1,0]=(-0.913971229647)*x[0]**o+(0.042531889985)*x[0]+(0.143956410289)*x[1]**o+(0.597990264168)*x[1]+(-0.85926903006)*x[2]**o+(0.205900361792)*x[2]
        arg[1,4,1,1]=(-0.955540264653)*x[0]**o+(0.983439073814)*x[0]+(0.246870351555)*x[1]**o+(-0.610917180843)*x[1]+(-0.0418134447092)*x[2]**o+(-0.58407787472)*x[2]
        arg[1,4,2,0]=(0.787274711453)*x[0]**o+(-0.899651339467)*x[0]+(-0.62195254225)*x[1]**o+(-0.516447289251)*x[1]+(0.267082643601)*x[2]**o+(0.0408186039688)*x[2]
        arg[1,4,2,1]=(-0.485773801895)*x[0]**o+(0.663478067655)*x[0]+(0.928973237869)*x[1]**o+(0.353735145156)*x[1]+(0.128409748978)*x[2]**o+(-0.476479648592)*x[2]
        arg[2,0,0,0]=(0.491574821811)*x[0]**o+(0.455303820526)*x[0]+(-0.747773112075)*x[1]**o+(-0.653255987137)*x[1]+(-0.738293040346)*x[2]**o+(0.948070477272)*x[2]
        arg[2,0,0,1]=(-0.697412782403)*x[0]**o+(-0.738389544586)*x[0]+(0.231728579935)*x[1]**o+(0.248503931032)*x[1]+(-0.561155590703)*x[2]**o+(0.324682113147)*x[2]
        arg[2,0,1,0]=(-0.00507638382064)*x[0]**o+(-0.983105866448)*x[0]+(-0.253141627484)*x[1]**o+(-0.0664178091069)*x[1]+(0.194460220144)*x[2]**o+(-0.174257469198)*x[2]
        arg[2,0,1,1]=(-0.644447560463)*x[0]**o+(0.658703761133)*x[0]+(-0.685392825104)*x[1]**o+(-0.766295260796)*x[1]+(-0.922004388493)*x[2]**o+(-0.0745707720923)*x[2]
        arg[2,0,2,0]=(-0.69924009606)*x[0]**o+(-0.398212597328)*x[0]+(-0.364991550186)*x[1]**o+(0.243474288977)*x[1]+(0.129545162767)*x[2]**o+(0.920611954845)*x[2]
        arg[2,0,2,1]=(-0.494476126875)*x[0]**o+(-0.131503103148)*x[0]+(-0.241670500874)*x[1]**o+(0.40439125905)*x[1]+(0.134166779399)*x[2]**o+(-0.507385751809)*x[2]
        arg[2,1,0,0]=(0.301148115913)*x[0]**o+(0.189635133395)*x[0]+(0.744167404774)*x[1]**o+(0.995754344203)*x[1]+(-0.334760662218)*x[2]**o+(-0.182462923741)*x[2]
        arg[2,1,0,1]=(0.523500022508)*x[0]**o+(-0.257042661558)*x[0]+(-0.940233100381)*x[1]**o+(-0.202431731891)*x[1]+(-0.653480190287)*x[2]**o+(0.993156915081)*x[2]
        arg[2,1,1,0]=(-0.502452234661)*x[0]**o+(-0.711581805303)*x[0]+(0.696450091152)*x[1]**o+(0.436746953802)*x[1]+(0.695969418438)*x[2]**o+(-0.671400022595)*x[2]
        arg[2,1,1,1]=(0.0705944125182)*x[0]**o+(-0.865755834485)*x[0]+(0.921555996641)*x[1]**o+(-0.567855814858)*x[1]+(-0.213455192388)*x[2]**o+(0.820868399259)*x[2]
        arg[2,1,2,0]=(0.744232126707)*x[0]**o+(0.336969176485)*x[0]+(0.339407729288)*x[1]**o+(-0.465154968917)*x[1]+(0.389524167202)*x[2]**o+(0.645394954477)*x[2]
        arg[2,1,2,1]=(-0.299085687516)*x[0]**o+(0.220795864743)*x[0]+(-0.19353913538)*x[1]**o+(0.445131396297)*x[1]+(-0.779605318636)*x[2]**o+(-0.76707264371)*x[2]
        arg[2,2,0,0]=(-0.357585997171)*x[0]**o+(-0.248795296406)*x[0]+(0.0876364714478)*x[1]**o+(-0.682833482179)*x[1]+(0.497052575986)*x[2]**o+(0.893378007889)*x[2]
        arg[2,2,0,1]=(0.38067999218)*x[0]**o+(-0.857246232026)*x[0]+(0.0785818284359)*x[1]**o+(-0.473287028123)*x[1]+(-0.763041318956)*x[2]**o+(-0.1119710251)*x[2]
        arg[2,2,1,0]=(-0.0278856926369)*x[0]**o+(0.435955626553)*x[0]+(-0.676274741782)*x[1]**o+(-0.719721825467)*x[1]+(0.195007734703)*x[2]**o+(0.504771278902)*x[2]
        arg[2,2,1,1]=(0.459547475772)*x[0]**o+(-0.669041681613)*x[0]+(-0.741957874364)*x[1]**o+(0.604585337407)*x[1]+(-0.886718742046)*x[2]**o+(0.338113237761)*x[2]
        arg[2,2,2,0]=(0.392926733678)*x[0]**o+(-0.923680589861)*x[0]+(-0.425503328457)*x[1]**o+(-0.0902669250117)*x[1]+(-0.346726072703)*x[2]**o+(0.541482976709)*x[2]
        arg[2,2,2,1]=(0.134867023616)*x[0]**o+(0.478349607348)*x[0]+(-0.334617836366)*x[1]**o+(-0.74587411744)*x[1]+(-0.880365937567)*x[2]**o+(0.820743064452)*x[2]
        arg[2,3,0,0]=(0.081184173895)*x[0]**o+(0.315011929905)*x[0]+(-0.137083926616)*x[1]**o+(0.103926223299)*x[1]+(0.363270989968)*x[2]**o+(-0.454075563606)*x[2]
        arg[2,3,0,1]=(0.457730701321)*x[0]**o+(0.922345625691)*x[0]+(-0.412933933474)*x[1]**o+(-0.546359655103)*x[1]+(0.719101029933)*x[2]**o+(-0.790522859322)*x[2]
        arg[2,3,1,0]=(-0.811816419841)*x[0]**o+(0.0379298408788)*x[0]+(-0.852643125131)*x[1]**o+(0.76881274254)*x[1]+(0.75214553496)*x[2]**o+(0.800382814403)*x[2]
        arg[2,3,1,1]=(0.524802143618)*x[0]**o+(-0.717457078371)*x[0]+(0.389026887965)*x[1]**o+(-0.231912558721)*x[1]+(-0.394705562333)*x[2]**o+(-0.794723802655)*x[2]
        arg[2,3,2,0]=(-0.232215520303)*x[0]**o+(0.994932250617)*x[0]+(-0.760425274222)*x[1]**o+(0.0910727760744)*x[1]+(0.621642024132)*x[2]**o+(0.263978719304)*x[2]
        arg[2,3,2,1]=(0.515269404424)*x[0]**o+(-0.998266958288)*x[0]+(-0.485078033534)*x[1]**o+(-0.960043247553)*x[1]+(-0.926562413314)*x[2]**o+(0.612710762712)*x[2]
        arg[2,4,0,0]=(0.419808446591)*x[0]**o+(0.132792190191)*x[0]+(-0.690603562736)*x[1]**o+(0.45137757885)*x[1]+(-0.262249041567)*x[2]**o+(-0.432995254783)*x[2]
        arg[2,4,0,1]=(0.62413323935)*x[0]**o+(0.963559148607)*x[0]+(-0.879044261015)*x[1]**o+(-0.504981574957)*x[1]+(-0.575204029973)*x[2]**o+(0.066855791448)*x[2]
        arg[2,4,1,0]=(0.321218662791)*x[0]**o+(0.975103526588)*x[0]+(0.783716185844)*x[1]**o+(-0.816368139389)*x[1]+(0.778169798128)*x[2]**o+(-0.527443068085)*x[2]
        arg[2,4,1,1]=(-0.8801082648)*x[0]**o+(-0.0313845929272)*x[0]+(0.485262237838)*x[1]**o+(0.0778888535573)*x[1]+(-0.900948214929)*x[2]**o+(-0.98830300089)*x[2]
        arg[2,4,2,0]=(-0.996395042226)*x[0]**o+(0.253635852642)*x[0]+(0.180965173659)*x[1]**o+(-0.488294662459)*x[1]+(-0.144583511741)*x[2]**o+(0.285854081388)*x[2]
        arg[2,4,2,1]=(0.00988506292061)*x[0]**o+(-0.00192786315452)*x[0]+(0.720775470647)*x[1]**o+(-0.587673311506)*x[1]+(0.313836238675)*x[2]**o+(0.45975125457)*x[2]
        arg[3,0,0,0]=(-0.822049489749)*x[0]**o+(-0.989589435175)*x[0]+(-0.0943830434118)*x[1]**o+(-0.789229900035)*x[1]+(0.311382619317)*x[2]**o+(-0.316646938685)*x[2]
        arg[3,0,0,1]=(0.501517916677)*x[0]**o+(-0.704502365436)*x[0]+(0.913184728006)*x[1]**o+(-0.843712848791)*x[1]+(0.901479696878)*x[2]**o+(-0.713893336542)*x[2]
        arg[3,0,1,0]=(0.0771592811086)*x[0]**o+(0.796038032206)*x[0]+(0.566256212871)*x[1]**o+(0.394136659986)*x[1]+(-0.794422065388)*x[2]**o+(0.662174992848)*x[2]
        arg[3,0,1,1]=(0.180225115167)*x[0]**o+(0.366965867971)*x[0]+(-0.244015756153)*x[1]**o+(0.433571067335)*x[1]+(0.129075150185)*x[2]**o+(-0.216304414605)*x[2]
        arg[3,0,2,0]=(0.100277967034)*x[0]**o+(-0.590583546786)*x[0]+(-0.595553697152)*x[1]**o+(0.923573252239)*x[1]+(0.578966663184)*x[2]**o+(-0.411082913097)*x[2]
        arg[3,0,2,1]=(-0.575315696811)*x[0]**o+(0.608760285918)*x[0]+(-0.3592074605)*x[1]**o+(0.515119945308)*x[1]+(-0.4000138306)*x[2]**o+(0.477685324937)*x[2]
        arg[3,1,0,0]=(0.0780014276919)*x[0]**o+(-0.0932830671471)*x[0]+(0.39477236642)*x[1]**o+(0.406673135556)*x[1]+(-0.235480707568)*x[2]**o+(-0.872302257144)*x[2]
        arg[3,1,0,1]=(0.170573072646)*x[0]**o+(0.868146561326)*x[0]+(0.262136343896)*x[1]**o+(0.192343529912)*x[1]+(0.157400965693)*x[2]**o+(-0.932981619473)*x[2]
        arg[3,1,1,0]=(0.891433179291)*x[0]**o+(-0.281396449997)*x[0]+(-0.0803173619772)*x[1]**o+(-0.378460054356)*x[1]+(-0.581578656014)*x[2]**o+(-0.65856067509)*x[2]
        arg[3,1,1,1]=(0.343599956626)*x[0]**o+(-0.92063694125)*x[0]+(0.440698102072)*x[1]**o+(0.767124166388)*x[1]+(0.050964317543)*x[2]**o+(1.35605562239e-05)*x[2]
        arg[3,1,2,0]=(-0.614268150987)*x[0]**o+(0.363806471556)*x[0]+(0.92193457905)*x[1]**o+(0.157919580642)*x[1]+(-0.955575321124)*x[2]**o+(-0.984808437666)*x[2]
        arg[3,1,2,1]=(0.030099648966)*x[0]**o+(-0.375999369491)*x[0]+(-0.0992643926008)*x[1]**o+(0.109209161093)*x[1]+(-0.974043780088)*x[2]**o+(0.908049658745)*x[2]
        arg[3,2,0,0]=(0.297771341106)*x[0]**o+(-0.637802228169)*x[0]+(0.398100441783)*x[1]**o+(-0.0643425181651)*x[1]+(0.662775339033)*x[2]**o+(0.744835182052)*x[2]
        arg[3,2,0,1]=(0.312026420583)*x[0]**o+(0.865855052574)*x[0]+(0.0195001783165)*x[1]**o+(0.537899238473)*x[1]+(0.554893357481)*x[2]**o+(0.71555610145)*x[2]
        arg[3,2,1,0]=(-0.199138146186)*x[0]**o+(0.827546489855)*x[0]+(-0.480965464736)*x[1]**o+(-0.473796093187)*x[1]+(0.880264675049)*x[2]**o+(0.544506670844)*x[2]
        arg[3,2,1,1]=(0.773798005008)*x[0]**o+(0.744945241634)*x[0]+(0.548170564608)*x[1]**o+(-0.560503692823)*x[1]+(-0.435630215427)*x[2]**o+(0.080015730707)*x[2]
        arg[3,2,2,0]=(0.36735714937)*x[0]**o+(-0.867071557873)*x[0]+(0.936135650431)*x[1]**o+(-0.432533600524)*x[1]+(-0.0224430204069)*x[2]**o+(0.218590302621)*x[2]
        arg[3,2,2,1]=(0.932808382499)*x[0]**o+(-0.0970016713319)*x[0]+(0.458943467579)*x[1]**o+(-0.134338789898)*x[1]+(-0.426892557314)*x[2]**o+(0.627071826863)*x[2]
        arg[3,3,0,0]=(-0.623804149075)*x[0]**o+(0.910423590172)*x[0]+(0.465938181097)*x[1]**o+(0.495479245391)*x[1]+(0.693745442185)*x[2]**o+(-0.437014226757)*x[2]
        arg[3,3,0,1]=(0.598807017496)*x[0]**o+(-0.112361840211)*x[0]+(-0.550899685213)*x[1]**o+(-0.0478222019564)*x[1]+(-0.0110129517693)*x[2]**o+(-0.059534786374)*x[2]
        arg[3,3,1,0]=(-0.561155633409)*x[0]**o+(-0.463563057636)*x[0]+(-0.170463996009)*x[1]**o+(-0.214509204942)*x[1]+(0.851200198666)*x[2]**o+(-0.721036329689)*x[2]
        arg[3,3,1,1]=(0.608666240405)*x[0]**o+(0.814232455115)*x[0]+(-0.604645629689)*x[1]**o+(0.960976509673)*x[1]+(0.365807996219)*x[2]**o+(-0.87491909238)*x[2]
        arg[3,3,2,0]=(0.120126988418)*x[0]**o+(-0.745687314825)*x[0]+(-0.332697826984)*x[1]**o+(-0.823683406516)*x[1]+(0.305444079613)*x[2]**o+(0.901972115441)*x[2]
        arg[3,3,2,1]=(0.0470293848904)*x[0]**o+(-0.195496754814)*x[0]+(0.671870114105)*x[1]**o+(0.170273146371)*x[1]+(-0.851927986188)*x[2]**o+(0.772192863973)*x[2]
        arg[3,4,0,0]=(-0.95392341727)*x[0]**o+(-0.665694215521)*x[0]+(-0.0247396499378)*x[1]**o+(0.441352857387)*x[1]+(0.767156547791)*x[2]**o+(-0.14696372018)*x[2]
        arg[3,4,0,1]=(0.209761166815)*x[0]**o+(0.978934473425)*x[0]+(0.943806891053)*x[1]**o+(-0.756619917601)*x[1]+(0.635699949764)*x[2]**o+(-0.345248269568)*x[2]
        arg[3,4,1,0]=(0.748020785785)*x[0]**o+(0.311104549542)*x[0]+(0.739127551248)*x[1]**o+(-0.635238575859)*x[1]+(0.623898875653)*x[2]**o+(0.343407392068)*x[2]
        arg[3,4,1,1]=(-0.422500892471)*x[0]**o+(0.0688495013346)*x[0]+(-0.919653906232)*x[1]**o+(0.531036586396)*x[1]+(-0.22487910672)*x[2]**o+(-0.848225994541)*x[2]
        arg[3,4,2,0]=(0.723965744673)*x[0]**o+(0.558602959224)*x[0]+(0.699539243308)*x[1]**o+(-0.597426541715)*x[1]+(-0.0128675070522)*x[2]**o+(-0.874796626242)*x[2]
        arg[3,4,2,1]=(-0.962970021216)*x[0]**o+(0.532886481874)*x[0]+(-0.523159705965)*x[1]**o+(-0.82887407638)*x[1]+(0.874288116843)*x[2]**o+(0.520024515804)*x[2]
        ref[0,0,0,0]=(0.507422888369)*x_ref[0]**o+(0.808798513305)*x_ref[0]+(0.51493073805)*x_ref[1]**o+(-0.928527686947)*x_ref[1]+(-0.788010447498)*x_ref[2]**o+(-0.378389833131)*x_ref[2]
        ref[0,0,0,1]=(-0.475851243177)*x_ref[0]**o+(0.746578446988)*x_ref[0]+(0.504778622494)*x_ref[1]**o+(-0.677770398899)*x_ref[1]+(-0.862535648403)*x_ref[2]**o+(0.620414141691)*x_ref[2]
        ref[0,0,1,0]=(-0.361726879582)*x_ref[0]**o+(0.554584127587)*x_ref[0]+(0.436610414772)*x_ref[1]**o+(-0.686534492717)*x_ref[1]+(0.757226061419)*x_ref[2]**o+(0.0948537725389)*x_ref[2]
        ref[0,0,1,1]=(-0.362160696355)*x_ref[0]**o+(0.766524051703)*x_ref[0]+(0.209980857795)*x_ref[1]**o+(-0.478398061437)*x_ref[1]+(-0.410960179122)*x_ref[2]**o+(-0.213272138462)*x_ref[2]
        ref[0,0,2,0]=(0.87154872554)*x_ref[0]**o+(-0.727915301904)*x_ref[0]+(0.683580698056)*x_ref[1]**o+(0.215052521277)*x_ref[1]+(0.267231507898)*x_ref[2]**o+(-0.0124189912449)*x_ref[2]
        ref[0,0,2,1]=(0.631400944876)*x_ref[0]**o+(0.537293628606)*x_ref[0]+(0.944754350994)*x_ref[1]**o+(0.108068908012)*x_ref[1]+(0.890117096365)*x_ref[2]**o+(0.164706867235)*x_ref[2]
        ref[0,1,0,0]=(-0.935871759207)*x_ref[0]**o+(0.350133758077)*x_ref[0]+(-0.966482181015)*x_ref[1]**o+(0.440902701528)*x_ref[1]+(-0.893329541322)*x_ref[2]**o+(0.350317827938)*x_ref[2]
        ref[0,1,0,1]=(0.465305203601)*x_ref[0]**o+(0.711131797928)*x_ref[0]+(0.104091536498)*x_ref[1]**o+(0.484811628127)*x_ref[1]+(0.676878954516)*x_ref[2]**o+(-0.99327016995)*x_ref[2]
        ref[0,1,1,0]=(-0.361189016151)*x_ref[0]**o+(0.962468573492)*x_ref[0]+(-0.158571122564)*x_ref[1]**o+(-0.293583660912)*x_ref[1]+(-0.713699173608)*x_ref[2]**o+(-0.507430163361)*x_ref[2]
        ref[0,1,1,1]=(0.172564527054)*x_ref[0]**o+(0.273438145027)*x_ref[0]+(0.137378266239)*x_ref[1]**o+(-0.675788606841)*x_ref[1]+(-0.362143416121)*x_ref[2]**o+(0.0703897749775)*x_ref[2]
        ref[0,1,2,0]=(-0.941374717271)*x_ref[0]**o+(0.289433165458)*x_ref[0]+(0.318615227028)*x_ref[1]**o+(0.524362568204)*x_ref[1]+(-0.612590305095)*x_ref[2]**o+(-0.117555538957)*x_ref[2]
        ref[0,1,2,1]=(-0.442920825551)*x_ref[0]**o+(-0.454982146579)*x_ref[0]+(0.355353737589)*x_ref[1]**o+(-0.729623917004)*x_ref[1]+(-0.567220335937)*x_ref[2]**o+(0.977652639976)*x_ref[2]
        ref[0,2,0,0]=(0.942485930373)*x_ref[0]**o+(-0.605823961033)*x_ref[0]+(-0.143465779338)*x_ref[1]**o+(0.327354308987)*x_ref[1]+(0.691900417542)*x_ref[2]**o+(-0.514165213849)*x_ref[2]
        ref[0,2,0,1]=(0.018029272977)*x_ref[0]**o+(-0.209992705469)*x_ref[0]+(-0.224614727554)*x_ref[1]**o+(-0.862697462048)*x_ref[1]+(0.274791387928)*x_ref[2]**o+(-0.907445869506)*x_ref[2]
        ref[0,2,1,0]=(0.862755248265)*x_ref[0]**o+(0.020831641939)*x_ref[0]+(0.496662599261)*x_ref[1]**o+(0.79318975953)*x_ref[1]+(0.0912166895541)*x_ref[2]**o+(0.556337610316)*x_ref[2]
        ref[0,2,1,1]=(-0.09652997071)*x_ref[0]**o+(-0.954279421394)*x_ref[0]+(-0.601053524018)*x_ref[1]**o+(0.642528012068)*x_ref[1]+(0.295748113888)*x_ref[2]**o+(-0.318166346186)*x_ref[2]
        ref[0,2,2,0]=(-0.914878854129)*x_ref[0]**o+(0.159201123642)*x_ref[0]+(0.589819056368)*x_ref[1]**o+(-0.715763226823)*x_ref[1]+(-0.444572583567)*x_ref[2]**o+(-0.487866651367)*x_ref[2]
        ref[0,2,2,1]=(-0.71271969175)*x_ref[0]**o+(0.454616308594)*x_ref[0]+(0.510270095021)*x_ref[1]**o+(-0.0385822967955)*x_ref[1]+(-0.519736540313)*x_ref[2]**o+(0.0556733091556)*x_ref[2]
        ref[0,3,0,0]=(0.355768034846)*x_ref[0]**o+(0.152240594627)*x_ref[0]+(-0.81081131843)*x_ref[1]**o+(-0.913373483156)*x_ref[1]+(-0.319859744799)*x_ref[2]**o+(-0.306218435393)*x_ref[2]
        ref[0,3,0,1]=(-0.656932063693)*x_ref[0]**o+(0.832886861783)*x_ref[0]+(-0.417112750683)*x_ref[1]**o+(-0.388190265957)*x_ref[1]+(0.246534947096)*x_ref[2]**o+(-0.896822839189)*x_ref[2]
        ref[0,3,1,0]=(0.204713418891)*x_ref[0]**o+(-0.150097488817)*x_ref[0]+(0.806578426941)*x_ref[1]**o+(-0.590052708995)*x_ref[1]+(-0.758958608419)*x_ref[2]**o+(-0.221973528232)*x_ref[2]
        ref[0,3,1,1]=(-0.854308184618)*x_ref[0]**o+(-0.292526203646)*x_ref[0]+(-0.662403504833)*x_ref[1]**o+(0.0955657982227)*x_ref[1]+(0.69471076173)*x_ref[2]**o+(0.384077309264)*x_ref[2]
        ref[0,3,2,0]=(0.982057200864)*x_ref[0]**o+(-0.64758727704)*x_ref[0]+(0.15941493222)*x_ref[1]**o+(-0.17745437115)*x_ref[1]+(-0.72953252955)*x_ref[2]**o+(0.986979270802)*x_ref[2]
        ref[0,3,2,1]=(0.764526698779)*x_ref[0]**o+(0.22809521115)*x_ref[0]+(-0.0762092087998)*x_ref[1]**o+(0.415212410471)*x_ref[1]+(-0.263638988463)*x_ref[2]**o+(-0.934669942336)*x_ref[2]
        ref[0,4,0,0]=(0.0812656932707)*x_ref[0]**o+(0.0397235062518)*x_ref[0]+(-0.0496912823983)*x_ref[1]**o+(0.189891585189)*x_ref[1]+(0.243537823293)*x_ref[2]**o+(0.461256134208)*x_ref[2]
        ref[0,4,0,1]=(0.167596358861)*x_ref[0]**o+(0.441650690481)*x_ref[0]+(0.638100632205)*x_ref[1]**o+(-0.69240615765)*x_ref[1]+(-0.501855369512)*x_ref[2]**o+(0.769465683782)*x_ref[2]
        ref[0,4,1,0]=(-0.137149398927)*x_ref[0]**o+(-0.186548325285)*x_ref[0]+(-0.504656948352)*x_ref[1]**o+(0.228896656134)*x_ref[1]+(-0.472425358061)*x_ref[2]**o+(-0.0769834754996)*x_ref[2]
        ref[0,4,1,1]=(0.311655648626)*x_ref[0]**o+(0.856513683959)*x_ref[0]+(0.47173626537)*x_ref[1]**o+(0.245497333467)*x_ref[1]+(-0.513727691305)*x_ref[2]**o+(-0.588614785682)*x_ref[2]
        ref[0,4,2,0]=(-0.152907854625)*x_ref[0]**o+(-0.86229950575)*x_ref[0]+(-0.931356414564)*x_ref[1]**o+(-0.915667991864)*x_ref[1]+(0.322316118527)*x_ref[2]**o+(-0.987559462539)*x_ref[2]
        ref[0,4,2,1]=(-0.705947769164)*x_ref[0]**o+(0.104081366627)*x_ref[0]+(-0.540454644392)*x_ref[1]**o+(0.404522880836)*x_ref[1]+(0.146464264141)*x_ref[2]**o+(0.586755030006)*x_ref[2]
        ref[1,0,0,0]=(-0.466614768106)*x_ref[0]**o+(0.482681414943)*x_ref[0]+(-0.176553738108)*x_ref[1]**o+(-0.39162272813)*x_ref[1]+(0.74312023324)*x_ref[2]**o+(0.824538905894)*x_ref[2]
        ref[1,0,0,1]=(-0.9351073631)*x_ref[0]**o+(-0.640226640449)*x_ref[0]+(0.966208762054)*x_ref[1]**o+(0.101792167944)*x_ref[1]+(-0.732983657038)*x_ref[2]**o+(-0.112313375937)*x_ref[2]
        ref[1,0,1,0]=(-0.305403470434)*x_ref[0]**o+(-0.414397930347)*x_ref[0]+(0.356093254322)*x_ref[1]**o+(-0.52607560749)*x_ref[1]+(-0.70070674473)*x_ref[2]**o+(-0.0192901616015)*x_ref[2]
        ref[1,0,1,1]=(0.450046989211)*x_ref[0]**o+(0.602770057268)*x_ref[0]+(0.0193812509113)*x_ref[1]**o+(-0.275615149191)*x_ref[1]+(0.0180468267836)*x_ref[2]**o+(-0.81388255113)*x_ref[2]
        ref[1,0,2,0]=(0.913651483249)*x_ref[0]**o+(-0.92188313743)*x_ref[0]+(0.680007150195)*x_ref[1]**o+(-0.300145584708)*x_ref[1]+(-0.398883520002)*x_ref[2]**o+(-0.752781496463)*x_ref[2]
        ref[1,0,2,1]=(-0.265749827845)*x_ref[0]**o+(0.594379728352)*x_ref[0]+(0.463409943587)*x_ref[1]**o+(-0.85840420426)*x_ref[1]+(-0.705963756674)*x_ref[2]**o+(0.251711112887)*x_ref[2]
        ref[1,1,0,0]=(-0.28895690781)*x_ref[0]**o+(0.476461656166)*x_ref[0]+(0.0692567954597)*x_ref[1]**o+(0.88763437878)*x_ref[1]+(0.0872256197554)*x_ref[2]**o+(0.19467039694)*x_ref[2]
        ref[1,1,0,1]=(0.741045418981)*x_ref[0]**o+(-0.726935151332)*x_ref[0]+(0.899418662131)*x_ref[1]**o+(0.27516497286)*x_ref[1]+(-0.715927474971)*x_ref[2]**o+(-0.709283098732)*x_ref[2]
        ref[1,1,1,0]=(0.991407130174)*x_ref[0]**o+(-0.163484265852)*x_ref[0]+(-0.983254548659)*x_ref[1]**o+(0.0568443855477)*x_ref[1]+(0.309664644588)*x_ref[2]**o+(0.904334718845)*x_ref[2]
        ref[1,1,1,1]=(0.237261672833)*x_ref[0]**o+(0.0861047746345)*x_ref[0]+(-0.494898018662)*x_ref[1]**o+(-0.433075334646)*x_ref[1]+(0.588597346505)*x_ref[2]**o+(-0.185559567553)*x_ref[2]
        ref[1,1,2,0]=(-0.969384091709)*x_ref[0]**o+(-0.0229777223491)*x_ref[0]+(0.0945623059666)*x_ref[1]**o+(0.303543647034)*x_ref[1]+(0.659000135381)*x_ref[2]**o+(0.00873791711254)*x_ref[2]
        ref[1,1,2,1]=(-0.421406731882)*x_ref[0]**o+(-0.5017394377)*x_ref[0]+(0.027747130694)*x_ref[1]**o+(0.327278873996)*x_ref[1]+(-0.754133439304)*x_ref[2]**o+(0.673525803845)*x_ref[2]
        ref[1,2,0,0]=(-0.104620319685)*x_ref[0]**o+(-0.685602393278)*x_ref[0]+(0.972173717014)*x_ref[1]**o+(-0.440573017063)*x_ref[1]+(-0.709529395743)*x_ref[2]**o+(0.634170741457)*x_ref[2]
        ref[1,2,0,1]=(-0.560386754264)*x_ref[0]**o+(0.570894623654)*x_ref[0]+(-0.455122833362)*x_ref[1]**o+(-0.00322505937935)*x_ref[1]+(0.959400336149)*x_ref[2]**o+(0.64014859954)*x_ref[2]
        ref[1,2,1,0]=(0.100585233221)*x_ref[0]**o+(0.939532360957)*x_ref[0]+(-0.900301955112)*x_ref[1]**o+(-0.776375896786)*x_ref[1]+(-0.779373294746)*x_ref[2]**o+(0.546157461116)*x_ref[2]
        ref[1,2,1,1]=(-0.680457315071)*x_ref[0]**o+(-0.0278844539208)*x_ref[0]+(-0.629673590044)*x_ref[1]**o+(0.813218376348)*x_ref[1]+(-0.449484131841)*x_ref[2]**o+(0.641682150048)*x_ref[2]
        ref[1,2,2,0]=(0.60056954386)*x_ref[0]**o+(-0.763169434657)*x_ref[0]+(-0.672890915035)*x_ref[1]**o+(-0.368711154571)*x_ref[1]+(0.476810195318)*x_ref[2]**o+(-0.935071124382)*x_ref[2]
        ref[1,2,2,1]=(0.994100148333)*x_ref[0]**o+(0.471647033381)*x_ref[0]+(0.341677296654)*x_ref[1]**o+(0.466442262068)*x_ref[1]+(-0.0634919249117)*x_ref[2]**o+(0.0637348830912)*x_ref[2]
        ref[1,3,0,0]=(-0.120744286726)*x_ref[0]**o+(0.630378984537)*x_ref[0]+(0.0837817796507)*x_ref[1]**o+(-0.850115274881)*x_ref[1]+(0.574342626561)*x_ref[2]**o+(-0.705409915626)*x_ref[2]
        ref[1,3,0,1]=(-0.51181061391)*x_ref[0]**o+(0.711479747616)*x_ref[0]+(-0.00800928076734)*x_ref[1]**o+(-0.703424580277)*x_ref[1]+(-0.925935026157)*x_ref[2]**o+(0.488175992827)*x_ref[2]
        ref[1,3,1,0]=(0.531879615751)*x_ref[0]**o+(-0.214270521934)*x_ref[0]+(-0.432344726015)*x_ref[1]**o+(-0.107660826449)*x_ref[1]+(0.763406427398)*x_ref[2]**o+(-0.511000150429)*x_ref[2]
        ref[1,3,1,1]=(-0.867467568206)*x_ref[0]**o+(-0.738425961596)*x_ref[0]+(-0.925337667706)*x_ref[1]**o+(-0.586227765107)*x_ref[1]+(0.909987211462)*x_ref[2]**o+(0.533347022582)*x_ref[2]
        ref[1,3,2,0]=(-0.47713113862)*x_ref[0]**o+(-0.130876138883)*x_ref[0]+(-0.607996835981)*x_ref[1]**o+(-0.0306261551172)*x_ref[1]+(0.956207500475)*x_ref[2]**o+(-0.287731719382)*x_ref[2]
        ref[1,3,2,1]=(0.179479486943)*x_ref[0]**o+(0.361266562479)*x_ref[0]+(0.374904472267)*x_ref[1]**o+(0.00788379106301)*x_ref[1]+(-0.34090684265)*x_ref[2]**o+(-0.190875910791)*x_ref[2]
        ref[1,4,0,0]=(0.685464203729)*x_ref[0]**o+(0.0634561930987)*x_ref[0]+(-0.662534110385)*x_ref[1]**o+(0.383849898086)*x_ref[1]+(0.948230524205)*x_ref[2]**o+(-0.347042281976)*x_ref[2]
        ref[1,4,0,1]=(-0.484499755322)*x_ref[0]**o+(-0.765655349243)*x_ref[0]+(0.805331514606)*x_ref[1]**o+(-0.0635912547496)*x_ref[1]+(-0.139457134772)*x_ref[2]**o+(0.227446734221)*x_ref[2]
        ref[1,4,1,0]=(-0.913971229647)*x_ref[0]**o+(0.042531889985)*x_ref[0]+(0.143956410289)*x_ref[1]**o+(0.597990264168)*x_ref[1]+(-0.85926903006)*x_ref[2]**o+(0.205900361792)*x_ref[2]
        ref[1,4,1,1]=(-0.955540264653)*x_ref[0]**o+(0.983439073814)*x_ref[0]+(0.246870351555)*x_ref[1]**o+(-0.610917180843)*x_ref[1]+(-0.0418134447092)*x_ref[2]**o+(-0.58407787472)*x_ref[2]
        ref[1,4,2,0]=(0.787274711453)*x_ref[0]**o+(-0.899651339467)*x_ref[0]+(-0.62195254225)*x_ref[1]**o+(-0.516447289251)*x_ref[1]+(0.267082643601)*x_ref[2]**o+(0.0408186039688)*x_ref[2]
        ref[1,4,2,1]=(-0.485773801895)*x_ref[0]**o+(0.663478067655)*x_ref[0]+(0.928973237869)*x_ref[1]**o+(0.353735145156)*x_ref[1]+(0.128409748978)*x_ref[2]**o+(-0.476479648592)*x_ref[2]
        ref[2,0,0,0]=(0.491574821811)*x_ref[0]**o+(0.455303820526)*x_ref[0]+(-0.747773112075)*x_ref[1]**o+(-0.653255987137)*x_ref[1]+(-0.738293040346)*x_ref[2]**o+(0.948070477272)*x_ref[2]
        ref[2,0,0,1]=(-0.697412782403)*x_ref[0]**o+(-0.738389544586)*x_ref[0]+(0.231728579935)*x_ref[1]**o+(0.248503931032)*x_ref[1]+(-0.561155590703)*x_ref[2]**o+(0.324682113147)*x_ref[2]
        ref[2,0,1,0]=(-0.00507638382064)*x_ref[0]**o+(-0.983105866448)*x_ref[0]+(-0.253141627484)*x_ref[1]**o+(-0.0664178091069)*x_ref[1]+(0.194460220144)*x_ref[2]**o+(-0.174257469198)*x_ref[2]
        ref[2,0,1,1]=(-0.644447560463)*x_ref[0]**o+(0.658703761133)*x_ref[0]+(-0.685392825104)*x_ref[1]**o+(-0.766295260796)*x_ref[1]+(-0.922004388493)*x_ref[2]**o+(-0.0745707720923)*x_ref[2]
        ref[2,0,2,0]=(-0.69924009606)*x_ref[0]**o+(-0.398212597328)*x_ref[0]+(-0.364991550186)*x_ref[1]**o+(0.243474288977)*x_ref[1]+(0.129545162767)*x_ref[2]**o+(0.920611954845)*x_ref[2]
        ref[2,0,2,1]=(-0.494476126875)*x_ref[0]**o+(-0.131503103148)*x_ref[0]+(-0.241670500874)*x_ref[1]**o+(0.40439125905)*x_ref[1]+(0.134166779399)*x_ref[2]**o+(-0.507385751809)*x_ref[2]
        ref[2,1,0,0]=(0.301148115913)*x_ref[0]**o+(0.189635133395)*x_ref[0]+(0.744167404774)*x_ref[1]**o+(0.995754344203)*x_ref[1]+(-0.334760662218)*x_ref[2]**o+(-0.182462923741)*x_ref[2]
        ref[2,1,0,1]=(0.523500022508)*x_ref[0]**o+(-0.257042661558)*x_ref[0]+(-0.940233100381)*x_ref[1]**o+(-0.202431731891)*x_ref[1]+(-0.653480190287)*x_ref[2]**o+(0.993156915081)*x_ref[2]
        ref[2,1,1,0]=(-0.502452234661)*x_ref[0]**o+(-0.711581805303)*x_ref[0]+(0.696450091152)*x_ref[1]**o+(0.436746953802)*x_ref[1]+(0.695969418438)*x_ref[2]**o+(-0.671400022595)*x_ref[2]
        ref[2,1,1,1]=(0.0705944125182)*x_ref[0]**o+(-0.865755834485)*x_ref[0]+(0.921555996641)*x_ref[1]**o+(-0.567855814858)*x_ref[1]+(-0.213455192388)*x_ref[2]**o+(0.820868399259)*x_ref[2]
        ref[2,1,2,0]=(0.744232126707)*x_ref[0]**o+(0.336969176485)*x_ref[0]+(0.339407729288)*x_ref[1]**o+(-0.465154968917)*x_ref[1]+(0.389524167202)*x_ref[2]**o+(0.645394954477)*x_ref[2]
        ref[2,1,2,1]=(-0.299085687516)*x_ref[0]**o+(0.220795864743)*x_ref[0]+(-0.19353913538)*x_ref[1]**o+(0.445131396297)*x_ref[1]+(-0.779605318636)*x_ref[2]**o+(-0.76707264371)*x_ref[2]
        ref[2,2,0,0]=(-0.357585997171)*x_ref[0]**o+(-0.248795296406)*x_ref[0]+(0.0876364714478)*x_ref[1]**o+(-0.682833482179)*x_ref[1]+(0.497052575986)*x_ref[2]**o+(0.893378007889)*x_ref[2]
        ref[2,2,0,1]=(0.38067999218)*x_ref[0]**o+(-0.857246232026)*x_ref[0]+(0.0785818284359)*x_ref[1]**o+(-0.473287028123)*x_ref[1]+(-0.763041318956)*x_ref[2]**o+(-0.1119710251)*x_ref[2]
        ref[2,2,1,0]=(-0.0278856926369)*x_ref[0]**o+(0.435955626553)*x_ref[0]+(-0.676274741782)*x_ref[1]**o+(-0.719721825467)*x_ref[1]+(0.195007734703)*x_ref[2]**o+(0.504771278902)*x_ref[2]
        ref[2,2,1,1]=(0.459547475772)*x_ref[0]**o+(-0.669041681613)*x_ref[0]+(-0.741957874364)*x_ref[1]**o+(0.604585337407)*x_ref[1]+(-0.886718742046)*x_ref[2]**o+(0.338113237761)*x_ref[2]
        ref[2,2,2,0]=(0.392926733678)*x_ref[0]**o+(-0.923680589861)*x_ref[0]+(-0.425503328457)*x_ref[1]**o+(-0.0902669250117)*x_ref[1]+(-0.346726072703)*x_ref[2]**o+(0.541482976709)*x_ref[2]
        ref[2,2,2,1]=(0.134867023616)*x_ref[0]**o+(0.478349607348)*x_ref[0]+(-0.334617836366)*x_ref[1]**o+(-0.74587411744)*x_ref[1]+(-0.880365937567)*x_ref[2]**o+(0.820743064452)*x_ref[2]
        ref[2,3,0,0]=(0.081184173895)*x_ref[0]**o+(0.315011929905)*x_ref[0]+(-0.137083926616)*x_ref[1]**o+(0.103926223299)*x_ref[1]+(0.363270989968)*x_ref[2]**o+(-0.454075563606)*x_ref[2]
        ref[2,3,0,1]=(0.457730701321)*x_ref[0]**o+(0.922345625691)*x_ref[0]+(-0.412933933474)*x_ref[1]**o+(-0.546359655103)*x_ref[1]+(0.719101029933)*x_ref[2]**o+(-0.790522859322)*x_ref[2]
        ref[2,3,1,0]=(-0.811816419841)*x_ref[0]**o+(0.0379298408788)*x_ref[0]+(-0.852643125131)*x_ref[1]**o+(0.76881274254)*x_ref[1]+(0.75214553496)*x_ref[2]**o+(0.800382814403)*x_ref[2]
        ref[2,3,1,1]=(0.524802143618)*x_ref[0]**o+(-0.717457078371)*x_ref[0]+(0.389026887965)*x_ref[1]**o+(-0.231912558721)*x_ref[1]+(-0.394705562333)*x_ref[2]**o+(-0.794723802655)*x_ref[2]
        ref[2,3,2,0]=(-0.232215520303)*x_ref[0]**o+(0.994932250617)*x_ref[0]+(-0.760425274222)*x_ref[1]**o+(0.0910727760744)*x_ref[1]+(0.621642024132)*x_ref[2]**o+(0.263978719304)*x_ref[2]
        ref[2,3,2,1]=(0.515269404424)*x_ref[0]**o+(-0.998266958288)*x_ref[0]+(-0.485078033534)*x_ref[1]**o+(-0.960043247553)*x_ref[1]+(-0.926562413314)*x_ref[2]**o+(0.612710762712)*x_ref[2]
        ref[2,4,0,0]=(0.419808446591)*x_ref[0]**o+(0.132792190191)*x_ref[0]+(-0.690603562736)*x_ref[1]**o+(0.45137757885)*x_ref[1]+(-0.262249041567)*x_ref[2]**o+(-0.432995254783)*x_ref[2]
        ref[2,4,0,1]=(0.62413323935)*x_ref[0]**o+(0.963559148607)*x_ref[0]+(-0.879044261015)*x_ref[1]**o+(-0.504981574957)*x_ref[1]+(-0.575204029973)*x_ref[2]**o+(0.066855791448)*x_ref[2]
        ref[2,4,1,0]=(0.321218662791)*x_ref[0]**o+(0.975103526588)*x_ref[0]+(0.783716185844)*x_ref[1]**o+(-0.816368139389)*x_ref[1]+(0.778169798128)*x_ref[2]**o+(-0.527443068085)*x_ref[2]
        ref[2,4,1,1]=(-0.8801082648)*x_ref[0]**o+(-0.0313845929272)*x_ref[0]+(0.485262237838)*x_ref[1]**o+(0.0778888535573)*x_ref[1]+(-0.900948214929)*x_ref[2]**o+(-0.98830300089)*x_ref[2]
        ref[2,4,2,0]=(-0.996395042226)*x_ref[0]**o+(0.253635852642)*x_ref[0]+(0.180965173659)*x_ref[1]**o+(-0.488294662459)*x_ref[1]+(-0.144583511741)*x_ref[2]**o+(0.285854081388)*x_ref[2]
        ref[2,4,2,1]=(0.00988506292061)*x_ref[0]**o+(-0.00192786315452)*x_ref[0]+(0.720775470647)*x_ref[1]**o+(-0.587673311506)*x_ref[1]+(0.313836238675)*x_ref[2]**o+(0.45975125457)*x_ref[2]
        ref[3,0,0,0]=(-0.822049489749)*x_ref[0]**o+(-0.989589435175)*x_ref[0]+(-0.0943830434118)*x_ref[1]**o+(-0.789229900035)*x_ref[1]+(0.311382619317)*x_ref[2]**o+(-0.316646938685)*x_ref[2]
        ref[3,0,0,1]=(0.501517916677)*x_ref[0]**o+(-0.704502365436)*x_ref[0]+(0.913184728006)*x_ref[1]**o+(-0.843712848791)*x_ref[1]+(0.901479696878)*x_ref[2]**o+(-0.713893336542)*x_ref[2]
        ref[3,0,1,0]=(0.0771592811086)*x_ref[0]**o+(0.796038032206)*x_ref[0]+(0.566256212871)*x_ref[1]**o+(0.394136659986)*x_ref[1]+(-0.794422065388)*x_ref[2]**o+(0.662174992848)*x_ref[2]
        ref[3,0,1,1]=(0.180225115167)*x_ref[0]**o+(0.366965867971)*x_ref[0]+(-0.244015756153)*x_ref[1]**o+(0.433571067335)*x_ref[1]+(0.129075150185)*x_ref[2]**o+(-0.216304414605)*x_ref[2]
        ref[3,0,2,0]=(0.100277967034)*x_ref[0]**o+(-0.590583546786)*x_ref[0]+(-0.595553697152)*x_ref[1]**o+(0.923573252239)*x_ref[1]+(0.578966663184)*x_ref[2]**o+(-0.411082913097)*x_ref[2]
        ref[3,0,2,1]=(-0.575315696811)*x_ref[0]**o+(0.608760285918)*x_ref[0]+(-0.3592074605)*x_ref[1]**o+(0.515119945308)*x_ref[1]+(-0.4000138306)*x_ref[2]**o+(0.477685324937)*x_ref[2]
        ref[3,1,0,0]=(0.0780014276919)*x_ref[0]**o+(-0.0932830671471)*x_ref[0]+(0.39477236642)*x_ref[1]**o+(0.406673135556)*x_ref[1]+(-0.235480707568)*x_ref[2]**o+(-0.872302257144)*x_ref[2]
        ref[3,1,0,1]=(0.170573072646)*x_ref[0]**o+(0.868146561326)*x_ref[0]+(0.262136343896)*x_ref[1]**o+(0.192343529912)*x_ref[1]+(0.157400965693)*x_ref[2]**o+(-0.932981619473)*x_ref[2]
        ref[3,1,1,0]=(0.891433179291)*x_ref[0]**o+(-0.281396449997)*x_ref[0]+(-0.0803173619772)*x_ref[1]**o+(-0.378460054356)*x_ref[1]+(-0.581578656014)*x_ref[2]**o+(-0.65856067509)*x_ref[2]
        ref[3,1,1,1]=(0.343599956626)*x_ref[0]**o+(-0.92063694125)*x_ref[0]+(0.440698102072)*x_ref[1]**o+(0.767124166388)*x_ref[1]+(0.050964317543)*x_ref[2]**o+(1.35605562239e-05)*x_ref[2]
        ref[3,1,2,0]=(-0.614268150987)*x_ref[0]**o+(0.363806471556)*x_ref[0]+(0.92193457905)*x_ref[1]**o+(0.157919580642)*x_ref[1]+(-0.955575321124)*x_ref[2]**o+(-0.984808437666)*x_ref[2]
        ref[3,1,2,1]=(0.030099648966)*x_ref[0]**o+(-0.375999369491)*x_ref[0]+(-0.0992643926008)*x_ref[1]**o+(0.109209161093)*x_ref[1]+(-0.974043780088)*x_ref[2]**o+(0.908049658745)*x_ref[2]
        ref[3,2,0,0]=(0.297771341106)*x_ref[0]**o+(-0.637802228169)*x_ref[0]+(0.398100441783)*x_ref[1]**o+(-0.0643425181651)*x_ref[1]+(0.662775339033)*x_ref[2]**o+(0.744835182052)*x_ref[2]
        ref[3,2,0,1]=(0.312026420583)*x_ref[0]**o+(0.865855052574)*x_ref[0]+(0.0195001783165)*x_ref[1]**o+(0.537899238473)*x_ref[1]+(0.554893357481)*x_ref[2]**o+(0.71555610145)*x_ref[2]
        ref[3,2,1,0]=(-0.199138146186)*x_ref[0]**o+(0.827546489855)*x_ref[0]+(-0.480965464736)*x_ref[1]**o+(-0.473796093187)*x_ref[1]+(0.880264675049)*x_ref[2]**o+(0.544506670844)*x_ref[2]
        ref[3,2,1,1]=(0.773798005008)*x_ref[0]**o+(0.744945241634)*x_ref[0]+(0.548170564608)*x_ref[1]**o+(-0.560503692823)*x_ref[1]+(-0.435630215427)*x_ref[2]**o+(0.080015730707)*x_ref[2]
        ref[3,2,2,0]=(0.36735714937)*x_ref[0]**o+(-0.867071557873)*x_ref[0]+(0.936135650431)*x_ref[1]**o+(-0.432533600524)*x_ref[1]+(-0.0224430204069)*x_ref[2]**o+(0.218590302621)*x_ref[2]
        ref[3,2,2,1]=(0.932808382499)*x_ref[0]**o+(-0.0970016713319)*x_ref[0]+(0.458943467579)*x_ref[1]**o+(-0.134338789898)*x_ref[1]+(-0.426892557314)*x_ref[2]**o+(0.627071826863)*x_ref[2]
        ref[3,3,0,0]=(-0.623804149075)*x_ref[0]**o+(0.910423590172)*x_ref[0]+(0.465938181097)*x_ref[1]**o+(0.495479245391)*x_ref[1]+(0.693745442185)*x_ref[2]**o+(-0.437014226757)*x_ref[2]
        ref[3,3,0,1]=(0.598807017496)*x_ref[0]**o+(-0.112361840211)*x_ref[0]+(-0.550899685213)*x_ref[1]**o+(-0.0478222019564)*x_ref[1]+(-0.0110129517693)*x_ref[2]**o+(-0.059534786374)*x_ref[2]
        ref[3,3,1,0]=(-0.561155633409)*x_ref[0]**o+(-0.463563057636)*x_ref[0]+(-0.170463996009)*x_ref[1]**o+(-0.214509204942)*x_ref[1]+(0.851200198666)*x_ref[2]**o+(-0.721036329689)*x_ref[2]
        ref[3,3,1,1]=(0.608666240405)*x_ref[0]**o+(0.814232455115)*x_ref[0]+(-0.604645629689)*x_ref[1]**o+(0.960976509673)*x_ref[1]+(0.365807996219)*x_ref[2]**o+(-0.87491909238)*x_ref[2]
        ref[3,3,2,0]=(0.120126988418)*x_ref[0]**o+(-0.745687314825)*x_ref[0]+(-0.332697826984)*x_ref[1]**o+(-0.823683406516)*x_ref[1]+(0.305444079613)*x_ref[2]**o+(0.901972115441)*x_ref[2]
        ref[3,3,2,1]=(0.0470293848904)*x_ref[0]**o+(-0.195496754814)*x_ref[0]+(0.671870114105)*x_ref[1]**o+(0.170273146371)*x_ref[1]+(-0.851927986188)*x_ref[2]**o+(0.772192863973)*x_ref[2]
        ref[3,4,0,0]=(-0.95392341727)*x_ref[0]**o+(-0.665694215521)*x_ref[0]+(-0.0247396499378)*x_ref[1]**o+(0.441352857387)*x_ref[1]+(0.767156547791)*x_ref[2]**o+(-0.14696372018)*x_ref[2]
        ref[3,4,0,1]=(0.209761166815)*x_ref[0]**o+(0.978934473425)*x_ref[0]+(0.943806891053)*x_ref[1]**o+(-0.756619917601)*x_ref[1]+(0.635699949764)*x_ref[2]**o+(-0.345248269568)*x_ref[2]
        ref[3,4,1,0]=(0.748020785785)*x_ref[0]**o+(0.311104549542)*x_ref[0]+(0.739127551248)*x_ref[1]**o+(-0.635238575859)*x_ref[1]+(0.623898875653)*x_ref[2]**o+(0.343407392068)*x_ref[2]
        ref[3,4,1,1]=(-0.422500892471)*x_ref[0]**o+(0.0688495013346)*x_ref[0]+(-0.919653906232)*x_ref[1]**o+(0.531036586396)*x_ref[1]+(-0.22487910672)*x_ref[2]**o+(-0.848225994541)*x_ref[2]
        ref[3,4,2,0]=(0.723965744673)*x_ref[0]**o+(0.558602959224)*x_ref[0]+(0.699539243308)*x_ref[1]**o+(-0.597426541715)*x_ref[1]+(-0.0128675070522)*x_ref[2]**o+(-0.874796626242)*x_ref[2]
        ref[3,4,2,1]=(-0.962970021216)*x_ref[0]**o+(0.532886481874)*x_ref[0]+(-0.523159705965)*x_ref[1]**o+(-0.82887407638)*x_ref[1]+(0.874288116843)*x_ref[2]**o+(0.520024515804)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.0603413935246)*x[0]**o+(-0.70184894912)*x[0]+(-0.206341392562)*x[1]**o+(0.892585857385)*x[1]
        ref=(-0.0603413935246)*x_ref[0]**o+(-0.70184894912)*x_ref[0]+(-0.206341392562)*x_ref[1]**o+(0.892585857385)*x_ref[1]
      else:
        arg=(0.120763612793)*x[0]**o+(-0.298506957699)*x[0]+(0.0620633729217)*x[1]**o+(0.807000465544)*x[1]+(-0.564443939995)*x[2]**o+(0.820783035273)*x[2]
        ref=(0.120763612793)*x_ref[0]**o+(-0.298506957699)*x_ref[0]+(0.0620633729217)*x_ref[1]**o+(0.807000465544)*x_ref[1]+(-0.564443939995)*x_ref[2]**o+(0.820783035273)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.69193390973)*x[0]**o+(-0.768888031033)*x[0]+(-0.813493496124)*x[1]**o+(-0.767059499773)*x[1]
        arg[1]=(0.111870073987)*x[0]**o+(-0.196406293238)*x[0]+(-0.10683601134)*x[1]**o+(0.69972228684)*x[1]
        ref[0]=(-0.69193390973)*x_ref[0]**o+(-0.768888031033)*x_ref[0]+(-0.813493496124)*x_ref[1]**o+(-0.767059499773)*x_ref[1]
        ref[1]=(0.111870073987)*x_ref[0]**o+(-0.196406293238)*x_ref[0]+(-0.10683601134)*x_ref[1]**o+(0.69972228684)*x_ref[1]
      else:
        arg[0]=(-0.20606732369)*x[0]**o+(0.773916734719)*x[0]+(-0.915594370888)*x[1]**o+(-0.684202475504)*x[1]+(0.161644377631)*x[2]**o+(-0.834568364942)*x[2]
        arg[1]=(-0.475848931399)*x[0]**o+(-0.967325026091)*x[0]+(0.371032251427)*x[1]**o+(0.976933923378)*x[1]+(0.326121409381)*x[2]**o+(0.880666151318)*x[2]
        ref[0]=(-0.20606732369)*x_ref[0]**o+(0.773916734719)*x_ref[0]+(-0.915594370888)*x_ref[1]**o+(-0.684202475504)*x_ref[1]+(0.161644377631)*x_ref[2]**o+(-0.834568364942)*x_ref[2]
        ref[1]=(-0.475848931399)*x_ref[0]**o+(-0.967325026091)*x_ref[0]+(0.371032251427)*x_ref[1]**o+(0.976933923378)*x_ref[1]+(0.326121409381)*x_ref[2]**o+(0.880666151318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.664616598004)*x[0]**o+(0.159791014359)*x[0]+(-0.3747223278)*x[1]**o+(0.302802720584)*x[1]
        arg[0,1]=(-0.334311194265)*x[0]**o+(0.865855840771)*x[0]+(-0.0184933229609)*x[1]**o+(0.589277746683)*x[1]
        arg[0,2]=(0.685266031246)*x[0]**o+(0.846357835994)*x[0]+(0.233422031025)*x[1]**o+(0.391225449794)*x[1]
        arg[0,3]=(-0.889412463645)*x[0]**o+(0.336985584587)*x[0]+(-0.311629085429)*x[1]**o+(-0.574756125055)*x[1]
        arg[0,4]=(-0.786434255596)*x[0]**o+(0.897529643329)*x[0]+(0.112631455142)*x[1]**o+(0.501957817255)*x[1]
        arg[1,0]=(-0.841995669352)*x[0]**o+(-0.372241861224)*x[0]+(0.669429619157)*x[1]**o+(0.913007793035)*x[1]
        arg[1,1]=(-0.965486931849)*x[0]**o+(-0.163330142621)*x[0]+(0.302996628649)*x[1]**o+(-0.214176383433)*x[1]
        arg[1,2]=(-0.717841677595)*x[0]**o+(-0.560453345617)*x[0]+(0.885221610478)*x[1]**o+(0.0925933897521)*x[1]
        arg[1,3]=(0.827380987548)*x[0]**o+(-0.185460343082)*x[0]+(-0.618848726331)*x[1]**o+(-0.25427152262)*x[1]
        arg[1,4]=(-0.232848766689)*x[0]**o+(-0.812427149527)*x[0]+(-0.6328125174)*x[1]**o+(0.377849173478)*x[1]
        arg[2,0]=(0.725568000689)*x[0]**o+(-0.0119648722611)*x[0]+(0.0551179058345)*x[1]**o+(-0.313601095535)*x[1]
        arg[2,1]=(-0.371331280463)*x[0]**o+(0.164812693414)*x[0]+(0.140689113041)*x[1]**o+(0.866863122864)*x[1]
        arg[2,2]=(0.445443472696)*x[0]**o+(0.885420986537)*x[0]+(0.434326737693)*x[1]**o+(0.013677800309)*x[1]
        arg[2,3]=(0.186221146816)*x[0]**o+(0.919506151782)*x[0]+(-0.163957717851)*x[1]**o+(-0.533309231769)*x[1]
        arg[2,4]=(0.137164729623)*x[0]**o+(0.561715777859)*x[0]+(0.33724606372)*x[1]**o+(-0.554531877504)*x[1]
        arg[3,0]=(-0.0632399630146)*x[0]**o+(0.712014652956)*x[0]+(0.107109261787)*x[1]**o+(0.228670262705)*x[1]
        arg[3,1]=(-0.973793135038)*x[0]**o+(0.923465546039)*x[0]+(0.189825544105)*x[1]**o+(-0.33784496447)*x[1]
        arg[3,2]=(-0.12686258826)*x[0]**o+(-0.582653698893)*x[0]+(0.860503162549)*x[1]**o+(-0.477305983221)*x[1]
        arg[3,3]=(-0.407988306429)*x[0]**o+(-0.174396602789)*x[0]+(0.646815890222)*x[1]**o+(-0.450974163079)*x[1]
        arg[3,4]=(0.162458379489)*x[0]**o+(-0.0569354582388)*x[0]+(0.0463216806125)*x[1]**o+(-0.381777259635)*x[1]
        ref[0,0]=(-0.664616598004)*x_ref[0]**o+(0.159791014359)*x_ref[0]+(-0.3747223278)*x_ref[1]**o+(0.302802720584)*x_ref[1]
        ref[0,1]=(-0.334311194265)*x_ref[0]**o+(0.865855840771)*x_ref[0]+(-0.0184933229609)*x_ref[1]**o+(0.589277746683)*x_ref[1]
        ref[0,2]=(0.685266031246)*x_ref[0]**o+(0.846357835994)*x_ref[0]+(0.233422031025)*x_ref[1]**o+(0.391225449794)*x_ref[1]
        ref[0,3]=(-0.889412463645)*x_ref[0]**o+(0.336985584587)*x_ref[0]+(-0.311629085429)*x_ref[1]**o+(-0.574756125055)*x_ref[1]
        ref[0,4]=(-0.786434255596)*x_ref[0]**o+(0.897529643329)*x_ref[0]+(0.112631455142)*x_ref[1]**o+(0.501957817255)*x_ref[1]
        ref[1,0]=(-0.841995669352)*x_ref[0]**o+(-0.372241861224)*x_ref[0]+(0.669429619157)*x_ref[1]**o+(0.913007793035)*x_ref[1]
        ref[1,1]=(-0.965486931849)*x_ref[0]**o+(-0.163330142621)*x_ref[0]+(0.302996628649)*x_ref[1]**o+(-0.214176383433)*x_ref[1]
        ref[1,2]=(-0.717841677595)*x_ref[0]**o+(-0.560453345617)*x_ref[0]+(0.885221610478)*x_ref[1]**o+(0.0925933897521)*x_ref[1]
        ref[1,3]=(0.827380987548)*x_ref[0]**o+(-0.185460343082)*x_ref[0]+(-0.618848726331)*x_ref[1]**o+(-0.25427152262)*x_ref[1]
        ref[1,4]=(-0.232848766689)*x_ref[0]**o+(-0.812427149527)*x_ref[0]+(-0.6328125174)*x_ref[1]**o+(0.377849173478)*x_ref[1]
        ref[2,0]=(0.725568000689)*x_ref[0]**o+(-0.0119648722611)*x_ref[0]+(0.0551179058345)*x_ref[1]**o+(-0.313601095535)*x_ref[1]
        ref[2,1]=(-0.371331280463)*x_ref[0]**o+(0.164812693414)*x_ref[0]+(0.140689113041)*x_ref[1]**o+(0.866863122864)*x_ref[1]
        ref[2,2]=(0.445443472696)*x_ref[0]**o+(0.885420986537)*x_ref[0]+(0.434326737693)*x_ref[1]**o+(0.013677800309)*x_ref[1]
        ref[2,3]=(0.186221146816)*x_ref[0]**o+(0.919506151782)*x_ref[0]+(-0.163957717851)*x_ref[1]**o+(-0.533309231769)*x_ref[1]
        ref[2,4]=(0.137164729623)*x_ref[0]**o+(0.561715777859)*x_ref[0]+(0.33724606372)*x_ref[1]**o+(-0.554531877504)*x_ref[1]
        ref[3,0]=(-0.0632399630146)*x_ref[0]**o+(0.712014652956)*x_ref[0]+(0.107109261787)*x_ref[1]**o+(0.228670262705)*x_ref[1]
        ref[3,1]=(-0.973793135038)*x_ref[0]**o+(0.923465546039)*x_ref[0]+(0.189825544105)*x_ref[1]**o+(-0.33784496447)*x_ref[1]
        ref[3,2]=(-0.12686258826)*x_ref[0]**o+(-0.582653698893)*x_ref[0]+(0.860503162549)*x_ref[1]**o+(-0.477305983221)*x_ref[1]
        ref[3,3]=(-0.407988306429)*x_ref[0]**o+(-0.174396602789)*x_ref[0]+(0.646815890222)*x_ref[1]**o+(-0.450974163079)*x_ref[1]
        ref[3,4]=(0.162458379489)*x_ref[0]**o+(-0.0569354582388)*x_ref[0]+(0.0463216806125)*x_ref[1]**o+(-0.381777259635)*x_ref[1]
      else:
        arg[0,0]=(0.865079887449)*x[0]**o+(-0.833757275547)*x[0]+(-0.967658088049)*x[1]**o+(-0.168314490396)*x[1]+(0.151618201772)*x[2]**o+(-0.251062354355)*x[2]
        arg[0,1]=(0.110207997589)*x[0]**o+(-0.510994295641)*x[0]+(0.855215902178)*x[1]**o+(0.941502315255)*x[1]+(-0.794874945971)*x[2]**o+(-0.309190709938)*x[2]
        arg[0,2]=(0.71701217173)*x[0]**o+(-0.221333357165)*x[0]+(0.689963224026)*x[1]**o+(0.221517236085)*x[1]+(-0.52608140393)*x[2]**o+(0.371488496481)*x[2]
        arg[0,3]=(0.618194041415)*x[0]**o+(0.454387091678)*x[0]+(-0.707329629946)*x[1]**o+(0.41972715681)*x[1]+(0.0180789814415)*x[2]**o+(-0.176487980225)*x[2]
        arg[0,4]=(0.329275243989)*x[0]**o+(-0.679005993002)*x[0]+(-0.261017436418)*x[1]**o+(0.151780687062)*x[1]+(-0.480936330475)*x[2]**o+(-0.697077234587)*x[2]
        arg[1,0]=(0.720913763906)*x[0]**o+(0.283486705277)*x[0]+(0.498289077088)*x[1]**o+(0.657543220863)*x[1]+(-0.666293219679)*x[2]**o+(0.901456618353)*x[2]
        arg[1,1]=(-0.984526350563)*x[0]**o+(0.677118292603)*x[0]+(0.614707818107)*x[1]**o+(0.272173514554)*x[1]+(0.41492971828)*x[2]**o+(0.274511869871)*x[2]
        arg[1,2]=(0.322224985373)*x[0]**o+(0.567272204545)*x[0]+(-0.437313912332)*x[1]**o+(0.847486025896)*x[1]+(0.769434908781)*x[2]**o+(0.325917607497)*x[2]
        arg[1,3]=(0.709357837264)*x[0]**o+(0.927425103914)*x[0]+(-0.622876388501)*x[1]**o+(-0.296141603555)*x[1]+(-0.038448878923)*x[2]**o+(-0.320692580169)*x[2]
        arg[1,4]=(-0.27890333885)*x[0]**o+(0.803346569394)*x[0]+(0.00665934218255)*x[1]**o+(-0.89057979428)*x[1]+(-0.727135553706)*x[2]**o+(0.892946022497)*x[2]
        arg[2,0]=(0.579951394388)*x[0]**o+(0.0406506913385)*x[0]+(0.363944368798)*x[1]**o+(0.582732756951)*x[1]+(0.542548838027)*x[2]**o+(0.365742797343)*x[2]
        arg[2,1]=(0.253260777874)*x[0]**o+(-0.69077350875)*x[0]+(0.458080700492)*x[1]**o+(-0.79224006672)*x[1]+(-0.504865679524)*x[2]**o+(0.681425632809)*x[2]
        arg[2,2]=(0.600427718261)*x[0]**o+(0.91761242151)*x[0]+(0.602316047463)*x[1]**o+(0.890383614185)*x[1]+(-0.386691039525)*x[2]**o+(0.688659949109)*x[2]
        arg[2,3]=(-0.181338810914)*x[0]**o+(0.933744241999)*x[0]+(0.0212774093137)*x[1]**o+(0.284972854884)*x[1]+(-0.150219922419)*x[2]**o+(0.0878011402107)*x[2]
        arg[2,4]=(-0.848985879538)*x[0]**o+(0.298962025531)*x[0]+(0.593300576179)*x[1]**o+(0.301509253541)*x[1]+(0.971171084219)*x[2]**o+(-0.973386482857)*x[2]
        arg[3,0]=(-0.735608389534)*x[0]**o+(0.44902436426)*x[0]+(0.0536456095512)*x[1]**o+(0.208969935548)*x[1]+(-0.787609300211)*x[2]**o+(0.726024812477)*x[2]
        arg[3,1]=(0.200882171943)*x[0]**o+(-0.644461780745)*x[0]+(-0.0995704510264)*x[1]**o+(-0.857730241944)*x[1]+(-0.372302969302)*x[2]**o+(-0.171367954739)*x[2]
        arg[3,2]=(-0.673606332627)*x[0]**o+(-0.828183678442)*x[0]+(-0.236856099365)*x[1]**o+(-0.266339053288)*x[1]+(-0.972034740204)*x[2]**o+(-0.33151750348)*x[2]
        arg[3,3]=(-0.312470720017)*x[0]**o+(-0.657433718629)*x[0]+(-0.984940036612)*x[1]**o+(0.197707053145)*x[1]+(0.512009064643)*x[2]**o+(0.855601033248)*x[2]
        arg[3,4]=(-0.674854508228)*x[0]**o+(-0.394035682831)*x[0]+(-0.749446232511)*x[1]**o+(0.866992763518)*x[1]+(-0.945837361158)*x[2]**o+(0.593490782481)*x[2]
        ref[0,0]=(0.865079887449)*x_ref[0]**o+(-0.833757275547)*x_ref[0]+(-0.967658088049)*x_ref[1]**o+(-0.168314490396)*x_ref[1]+(0.151618201772)*x_ref[2]**o+(-0.251062354355)*x_ref[2]
        ref[0,1]=(0.110207997589)*x_ref[0]**o+(-0.510994295641)*x_ref[0]+(0.855215902178)*x_ref[1]**o+(0.941502315255)*x_ref[1]+(-0.794874945971)*x_ref[2]**o+(-0.309190709938)*x_ref[2]
        ref[0,2]=(0.71701217173)*x_ref[0]**o+(-0.221333357165)*x_ref[0]+(0.689963224026)*x_ref[1]**o+(0.221517236085)*x_ref[1]+(-0.52608140393)*x_ref[2]**o+(0.371488496481)*x_ref[2]
        ref[0,3]=(0.618194041415)*x_ref[0]**o+(0.454387091678)*x_ref[0]+(-0.707329629946)*x_ref[1]**o+(0.41972715681)*x_ref[1]+(0.0180789814415)*x_ref[2]**o+(-0.176487980225)*x_ref[2]
        ref[0,4]=(0.329275243989)*x_ref[0]**o+(-0.679005993002)*x_ref[0]+(-0.261017436418)*x_ref[1]**o+(0.151780687062)*x_ref[1]+(-0.480936330475)*x_ref[2]**o+(-0.697077234587)*x_ref[2]
        ref[1,0]=(0.720913763906)*x_ref[0]**o+(0.283486705277)*x_ref[0]+(0.498289077088)*x_ref[1]**o+(0.657543220863)*x_ref[1]+(-0.666293219679)*x_ref[2]**o+(0.901456618353)*x_ref[2]
        ref[1,1]=(-0.984526350563)*x_ref[0]**o+(0.677118292603)*x_ref[0]+(0.614707818107)*x_ref[1]**o+(0.272173514554)*x_ref[1]+(0.41492971828)*x_ref[2]**o+(0.274511869871)*x_ref[2]
        ref[1,2]=(0.322224985373)*x_ref[0]**o+(0.567272204545)*x_ref[0]+(-0.437313912332)*x_ref[1]**o+(0.847486025896)*x_ref[1]+(0.769434908781)*x_ref[2]**o+(0.325917607497)*x_ref[2]
        ref[1,3]=(0.709357837264)*x_ref[0]**o+(0.927425103914)*x_ref[0]+(-0.622876388501)*x_ref[1]**o+(-0.296141603555)*x_ref[1]+(-0.038448878923)*x_ref[2]**o+(-0.320692580169)*x_ref[2]
        ref[1,4]=(-0.27890333885)*x_ref[0]**o+(0.803346569394)*x_ref[0]+(0.00665934218255)*x_ref[1]**o+(-0.89057979428)*x_ref[1]+(-0.727135553706)*x_ref[2]**o+(0.892946022497)*x_ref[2]
        ref[2,0]=(0.579951394388)*x_ref[0]**o+(0.0406506913385)*x_ref[0]+(0.363944368798)*x_ref[1]**o+(0.582732756951)*x_ref[1]+(0.542548838027)*x_ref[2]**o+(0.365742797343)*x_ref[2]
        ref[2,1]=(0.253260777874)*x_ref[0]**o+(-0.69077350875)*x_ref[0]+(0.458080700492)*x_ref[1]**o+(-0.79224006672)*x_ref[1]+(-0.504865679524)*x_ref[2]**o+(0.681425632809)*x_ref[2]
        ref[2,2]=(0.600427718261)*x_ref[0]**o+(0.91761242151)*x_ref[0]+(0.602316047463)*x_ref[1]**o+(0.890383614185)*x_ref[1]+(-0.386691039525)*x_ref[2]**o+(0.688659949109)*x_ref[2]
        ref[2,3]=(-0.181338810914)*x_ref[0]**o+(0.933744241999)*x_ref[0]+(0.0212774093137)*x_ref[1]**o+(0.284972854884)*x_ref[1]+(-0.150219922419)*x_ref[2]**o+(0.0878011402107)*x_ref[2]
        ref[2,4]=(-0.848985879538)*x_ref[0]**o+(0.298962025531)*x_ref[0]+(0.593300576179)*x_ref[1]**o+(0.301509253541)*x_ref[1]+(0.971171084219)*x_ref[2]**o+(-0.973386482857)*x_ref[2]
        ref[3,0]=(-0.735608389534)*x_ref[0]**o+(0.44902436426)*x_ref[0]+(0.0536456095512)*x_ref[1]**o+(0.208969935548)*x_ref[1]+(-0.787609300211)*x_ref[2]**o+(0.726024812477)*x_ref[2]
        ref[3,1]=(0.200882171943)*x_ref[0]**o+(-0.644461780745)*x_ref[0]+(-0.0995704510264)*x_ref[1]**o+(-0.857730241944)*x_ref[1]+(-0.372302969302)*x_ref[2]**o+(-0.171367954739)*x_ref[2]
        ref[3,2]=(-0.673606332627)*x_ref[0]**o+(-0.828183678442)*x_ref[0]+(-0.236856099365)*x_ref[1]**o+(-0.266339053288)*x_ref[1]+(-0.972034740204)*x_ref[2]**o+(-0.33151750348)*x_ref[2]
        ref[3,3]=(-0.312470720017)*x_ref[0]**o+(-0.657433718629)*x_ref[0]+(-0.984940036612)*x_ref[1]**o+(0.197707053145)*x_ref[1]+(0.512009064643)*x_ref[2]**o+(0.855601033248)*x_ref[2]
        ref[3,4]=(-0.674854508228)*x_ref[0]**o+(-0.394035682831)*x_ref[0]+(-0.749446232511)*x_ref[1]**o+(0.866992763518)*x_ref[1]+(-0.945837361158)*x_ref[2]**o+(0.593490782481)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.511844892619)*x[0]**o+(0.72447480361)*x[0]+(-0.415725587152)*x[1]**o+(-0.495233818217)*x[1]
        arg[0,0,1]=(0.513861713949)*x[0]**o+(-0.918192893835)*x[0]+(0.563871531176)*x[1]**o+(0.885768393284)*x[1]
        arg[0,1,0]=(-0.412320762031)*x[0]**o+(0.576776422795)*x[0]+(0.0103535112247)*x[1]**o+(-0.738144219037)*x[1]
        arg[0,1,1]=(0.69941105604)*x[0]**o+(0.923470022558)*x[0]+(0.568449180547)*x[1]**o+(0.361429609343)*x[1]
        arg[1,0,0]=(-0.574139432427)*x[0]**o+(0.968980609212)*x[0]+(-0.198809151818)*x[1]**o+(0.751623099192)*x[1]
        arg[1,0,1]=(-0.743433955873)*x[0]**o+(-0.608013190658)*x[0]+(-0.0625473402572)*x[1]**o+(0.168418241981)*x[1]
        arg[1,1,0]=(-0.632180449517)*x[0]**o+(-0.892666803222)*x[0]+(-0.558127678438)*x[1]**o+(0.212462582041)*x[1]
        arg[1,1,1]=(-0.666381769622)*x[0]**o+(-0.0299308846938)*x[0]+(-0.450221481906)*x[1]**o+(0.701099654576)*x[1]
        arg[2,0,0]=(-0.927832262778)*x[0]**o+(0.901665098705)*x[0]+(-0.0462955000541)*x[1]**o+(-0.92139600993)*x[1]
        arg[2,0,1]=(-0.813732112672)*x[0]**o+(0.569194058283)*x[0]+(-0.527298120198)*x[1]**o+(-0.531607593369)*x[1]
        arg[2,1,0]=(-0.760238058111)*x[0]**o+(0.628718839289)*x[0]+(-0.420066582474)*x[1]**o+(-0.205537665056)*x[1]
        arg[2,1,1]=(0.421784629744)*x[0]**o+(-0.781451031751)*x[0]+(-0.0401156726902)*x[1]**o+(0.437044655245)*x[1]
        arg[3,0,0]=(0.324136561188)*x[0]**o+(-0.920235828584)*x[0]+(0.677675509043)*x[1]**o+(0.322337884254)*x[1]
        arg[3,0,1]=(-0.0711813236306)*x[0]**o+(0.0770312838088)*x[0]+(0.259613502243)*x[1]**o+(0.508154611576)*x[1]
        arg[3,1,0]=(0.944048224599)*x[0]**o+(0.039954080407)*x[0]+(-0.503309434151)*x[1]**o+(0.747109874504)*x[1]
        arg[3,1,1]=(0.692616323513)*x[0]**o+(0.0114568310884)*x[0]+(0.0708318026263)*x[1]**o+(0.822167043168)*x[1]
        arg[4,0,0]=(0.392645136008)*x[0]**o+(0.160054219263)*x[0]+(-0.32320145108)*x[1]**o+(0.598963113077)*x[1]
        arg[4,0,1]=(0.124778736846)*x[0]**o+(0.949589553291)*x[0]+(0.0985585514202)*x[1]**o+(-0.536684039884)*x[1]
        arg[4,1,0]=(-0.983690591076)*x[0]**o+(-0.115321512368)*x[0]+(0.0875590878092)*x[1]**o+(0.992162964289)*x[1]
        arg[4,1,1]=(-0.89264628952)*x[0]**o+(0.414589880453)*x[0]+(-0.40959764673)*x[1]**o+(-0.261016337239)*x[1]
        arg[5,0,0]=(-0.613553175002)*x[0]**o+(0.320685999723)*x[0]+(-0.523099455238)*x[1]**o+(0.855780524769)*x[1]
        arg[5,0,1]=(-0.45911360029)*x[0]**o+(-0.285942390871)*x[0]+(0.7279702976)*x[1]**o+(0.981483922949)*x[1]
        arg[5,1,0]=(0.577254626037)*x[0]**o+(0.146911866492)*x[0]+(-0.414545328014)*x[1]**o+(-0.979179708006)*x[1]
        arg[5,1,1]=(-0.838392720256)*x[0]**o+(-0.51717223091)*x[0]+(0.635616756614)*x[1]**o+(0.112711970543)*x[1]
        ref[0,0,0]=(-0.511844892619)*x_ref[0]**o+(0.72447480361)*x_ref[0]+(-0.415725587152)*x_ref[1]**o+(-0.495233818217)*x_ref[1]
        ref[0,0,1]=(0.513861713949)*x_ref[0]**o+(-0.918192893835)*x_ref[0]+(0.563871531176)*x_ref[1]**o+(0.885768393284)*x_ref[1]
        ref[0,1,0]=(-0.412320762031)*x_ref[0]**o+(0.576776422795)*x_ref[0]+(0.0103535112247)*x_ref[1]**o+(-0.738144219037)*x_ref[1]
        ref[0,1,1]=(0.69941105604)*x_ref[0]**o+(0.923470022558)*x_ref[0]+(0.568449180547)*x_ref[1]**o+(0.361429609343)*x_ref[1]
        ref[1,0,0]=(-0.574139432427)*x_ref[0]**o+(0.968980609212)*x_ref[0]+(-0.198809151818)*x_ref[1]**o+(0.751623099192)*x_ref[1]
        ref[1,0,1]=(-0.743433955873)*x_ref[0]**o+(-0.608013190658)*x_ref[0]+(-0.0625473402572)*x_ref[1]**o+(0.168418241981)*x_ref[1]
        ref[1,1,0]=(-0.632180449517)*x_ref[0]**o+(-0.892666803222)*x_ref[0]+(-0.558127678438)*x_ref[1]**o+(0.212462582041)*x_ref[1]
        ref[1,1,1]=(-0.666381769622)*x_ref[0]**o+(-0.0299308846938)*x_ref[0]+(-0.450221481906)*x_ref[1]**o+(0.701099654576)*x_ref[1]
        ref[2,0,0]=(-0.927832262778)*x_ref[0]**o+(0.901665098705)*x_ref[0]+(-0.0462955000541)*x_ref[1]**o+(-0.92139600993)*x_ref[1]
        ref[2,0,1]=(-0.813732112672)*x_ref[0]**o+(0.569194058283)*x_ref[0]+(-0.527298120198)*x_ref[1]**o+(-0.531607593369)*x_ref[1]
        ref[2,1,0]=(-0.760238058111)*x_ref[0]**o+(0.628718839289)*x_ref[0]+(-0.420066582474)*x_ref[1]**o+(-0.205537665056)*x_ref[1]
        ref[2,1,1]=(0.421784629744)*x_ref[0]**o+(-0.781451031751)*x_ref[0]+(-0.0401156726902)*x_ref[1]**o+(0.437044655245)*x_ref[1]
        ref[3,0,0]=(0.324136561188)*x_ref[0]**o+(-0.920235828584)*x_ref[0]+(0.677675509043)*x_ref[1]**o+(0.322337884254)*x_ref[1]
        ref[3,0,1]=(-0.0711813236306)*x_ref[0]**o+(0.0770312838088)*x_ref[0]+(0.259613502243)*x_ref[1]**o+(0.508154611576)*x_ref[1]
        ref[3,1,0]=(0.944048224599)*x_ref[0]**o+(0.039954080407)*x_ref[0]+(-0.503309434151)*x_ref[1]**o+(0.747109874504)*x_ref[1]
        ref[3,1,1]=(0.692616323513)*x_ref[0]**o+(0.0114568310884)*x_ref[0]+(0.0708318026263)*x_ref[1]**o+(0.822167043168)*x_ref[1]
        ref[4,0,0]=(0.392645136008)*x_ref[0]**o+(0.160054219263)*x_ref[0]+(-0.32320145108)*x_ref[1]**o+(0.598963113077)*x_ref[1]
        ref[4,0,1]=(0.124778736846)*x_ref[0]**o+(0.949589553291)*x_ref[0]+(0.0985585514202)*x_ref[1]**o+(-0.536684039884)*x_ref[1]
        ref[4,1,0]=(-0.983690591076)*x_ref[0]**o+(-0.115321512368)*x_ref[0]+(0.0875590878092)*x_ref[1]**o+(0.992162964289)*x_ref[1]
        ref[4,1,1]=(-0.89264628952)*x_ref[0]**o+(0.414589880453)*x_ref[0]+(-0.40959764673)*x_ref[1]**o+(-0.261016337239)*x_ref[1]
        ref[5,0,0]=(-0.613553175002)*x_ref[0]**o+(0.320685999723)*x_ref[0]+(-0.523099455238)*x_ref[1]**o+(0.855780524769)*x_ref[1]
        ref[5,0,1]=(-0.45911360029)*x_ref[0]**o+(-0.285942390871)*x_ref[0]+(0.7279702976)*x_ref[1]**o+(0.981483922949)*x_ref[1]
        ref[5,1,0]=(0.577254626037)*x_ref[0]**o+(0.146911866492)*x_ref[0]+(-0.414545328014)*x_ref[1]**o+(-0.979179708006)*x_ref[1]
        ref[5,1,1]=(-0.838392720256)*x_ref[0]**o+(-0.51717223091)*x_ref[0]+(0.635616756614)*x_ref[1]**o+(0.112711970543)*x_ref[1]
      else:
        arg[0,0,0]=(0.807360140818)*x[0]**o+(-0.399141641865)*x[0]+(0.752910658124)*x[1]**o+(0.0998734354996)*x[1]+(-0.146785318133)*x[2]**o+(-0.393796605729)*x[2]
        arg[0,0,1]=(-0.502209573844)*x[0]**o+(0.501684846477)*x[0]+(0.634537290794)*x[1]**o+(0.481836589517)*x[1]+(-0.0221536339712)*x[2]**o+(-0.868693970024)*x[2]
        arg[0,1,0]=(-0.268854282014)*x[0]**o+(0.0516535086968)*x[0]+(-0.505997013698)*x[1]**o+(0.414205349613)*x[1]+(0.0119995662599)*x[2]**o+(0.584354222675)*x[2]
        arg[0,1,1]=(0.81923283939)*x[0]**o+(-0.174637610292)*x[0]+(0.283785251937)*x[1]**o+(0.716260504685)*x[1]+(-0.957279283093)*x[2]**o+(0.632750238692)*x[2]
        arg[1,0,0]=(0.601003888536)*x[0]**o+(-0.575011763358)*x[0]+(0.095159775614)*x[1]**o+(0.141200996742)*x[1]+(0.0173869440863)*x[2]**o+(0.478733688259)*x[2]
        arg[1,0,1]=(-0.783826426764)*x[0]**o+(-0.40604430175)*x[0]+(-0.4574879786)*x[1]**o+(0.020222886424)*x[1]+(0.225419750344)*x[2]**o+(-0.530026190372)*x[2]
        arg[1,1,0]=(0.542744341911)*x[0]**o+(-0.501169367089)*x[0]+(0.447661946532)*x[1]**o+(0.384528043888)*x[1]+(-0.599271261928)*x[2]**o+(-0.535447844217)*x[2]
        arg[1,1,1]=(-0.457163465824)*x[0]**o+(0.0475382663524)*x[0]+(0.369626614251)*x[1]**o+(0.1125159395)*x[1]+(-0.252442923096)*x[2]**o+(0.0717205095822)*x[2]
        arg[2,0,0]=(0.341536174626)*x[0]**o+(0.4408059057)*x[0]+(0.196877383546)*x[1]**o+(0.903072917866)*x[1]+(0.423403716275)*x[2]**o+(-0.521563534689)*x[2]
        arg[2,0,1]=(0.553187361262)*x[0]**o+(-0.539057536895)*x[0]+(-0.814990210449)*x[1]**o+(-0.418614789609)*x[1]+(0.857298489434)*x[2]**o+(0.629708672)*x[2]
        arg[2,1,0]=(0.494803348038)*x[0]**o+(0.143644562028)*x[0]+(-0.0425813919235)*x[1]**o+(-0.818562153097)*x[1]+(-0.730690839032)*x[2]**o+(-0.569426137866)*x[2]
        arg[2,1,1]=(-0.573254727715)*x[0]**o+(-0.536459946225)*x[0]+(-0.244717967781)*x[1]**o+(-0.67463285343)*x[1]+(-0.383959754049)*x[2]**o+(-0.87595126443)*x[2]
        arg[3,0,0]=(0.758065154533)*x[0]**o+(0.634049823109)*x[0]+(0.851352543082)*x[1]**o+(0.302790462394)*x[1]+(-0.920790730269)*x[2]**o+(0.460419670912)*x[2]
        arg[3,0,1]=(0.415682634939)*x[0]**o+(-0.664821834506)*x[0]+(-0.314068887264)*x[1]**o+(-0.322413102029)*x[1]+(-0.498768865759)*x[2]**o+(-0.165338012626)*x[2]
        arg[3,1,0]=(-0.398599595126)*x[0]**o+(0.744397086526)*x[0]+(-0.02321839482)*x[1]**o+(0.989743131098)*x[1]+(0.0713981107777)*x[2]**o+(-0.7862942966)*x[2]
        arg[3,1,1]=(0.90835870295)*x[0]**o+(0.285965355995)*x[0]+(-0.302780058578)*x[1]**o+(0.396923771953)*x[1]+(0.717312324488)*x[2]**o+(0.967488138744)*x[2]
        arg[4,0,0]=(-0.286827561624)*x[0]**o+(0.67973673275)*x[0]+(0.325169520671)*x[1]**o+(0.482002981846)*x[1]+(-0.826037251282)*x[2]**o+(-0.0567054906659)*x[2]
        arg[4,0,1]=(-0.667297458554)*x[0]**o+(-0.207369757609)*x[0]+(-0.982746021703)*x[1]**o+(0.963677089363)*x[1]+(0.917328237047)*x[2]**o+(-0.684386577728)*x[2]
        arg[4,1,0]=(-0.835525763762)*x[0]**o+(-0.113324184435)*x[0]+(0.465855661969)*x[1]**o+(0.656047307997)*x[1]+(0.331675778117)*x[2]**o+(-0.686067177565)*x[2]
        arg[4,1,1]=(-0.468668496499)*x[0]**o+(0.295197251458)*x[0]+(0.065102699741)*x[1]**o+(-0.378381176309)*x[1]+(0.933627973066)*x[2]**o+(-0.91462136931)*x[2]
        arg[5,0,0]=(-0.161916994488)*x[0]**o+(0.3070005414)*x[0]+(0.195946524325)*x[1]**o+(-0.453290146527)*x[1]+(0.751532065651)*x[2]**o+(0.82676125654)*x[2]
        arg[5,0,1]=(-0.925362363492)*x[0]**o+(-0.0295351854097)*x[0]+(0.533980764998)*x[1]**o+(0.128080489262)*x[1]+(0.502858052015)*x[2]**o+(0.813196540163)*x[2]
        arg[5,1,0]=(-0.740146099318)*x[0]**o+(0.27623528724)*x[0]+(-0.361620073401)*x[1]**o+(-0.310534085108)*x[1]+(-0.0363428185757)*x[2]**o+(-0.145397849357)*x[2]
        arg[5,1,1]=(0.362089087335)*x[0]**o+(0.539359659929)*x[0]+(0.00904990097799)*x[1]**o+(-0.150401175722)*x[1]+(-0.20465882105)*x[2]**o+(-0.519331232192)*x[2]
        ref[0,0,0]=(0.807360140818)*x_ref[0]**o+(-0.399141641865)*x_ref[0]+(0.752910658124)*x_ref[1]**o+(0.0998734354996)*x_ref[1]+(-0.146785318133)*x_ref[2]**o+(-0.393796605729)*x_ref[2]
        ref[0,0,1]=(-0.502209573844)*x_ref[0]**o+(0.501684846477)*x_ref[0]+(0.634537290794)*x_ref[1]**o+(0.481836589517)*x_ref[1]+(-0.0221536339712)*x_ref[2]**o+(-0.868693970024)*x_ref[2]
        ref[0,1,0]=(-0.268854282014)*x_ref[0]**o+(0.0516535086968)*x_ref[0]+(-0.505997013698)*x_ref[1]**o+(0.414205349613)*x_ref[1]+(0.0119995662599)*x_ref[2]**o+(0.584354222675)*x_ref[2]
        ref[0,1,1]=(0.81923283939)*x_ref[0]**o+(-0.174637610292)*x_ref[0]+(0.283785251937)*x_ref[1]**o+(0.716260504685)*x_ref[1]+(-0.957279283093)*x_ref[2]**o+(0.632750238692)*x_ref[2]
        ref[1,0,0]=(0.601003888536)*x_ref[0]**o+(-0.575011763358)*x_ref[0]+(0.095159775614)*x_ref[1]**o+(0.141200996742)*x_ref[1]+(0.0173869440863)*x_ref[2]**o+(0.478733688259)*x_ref[2]
        ref[1,0,1]=(-0.783826426764)*x_ref[0]**o+(-0.40604430175)*x_ref[0]+(-0.4574879786)*x_ref[1]**o+(0.020222886424)*x_ref[1]+(0.225419750344)*x_ref[2]**o+(-0.530026190372)*x_ref[2]
        ref[1,1,0]=(0.542744341911)*x_ref[0]**o+(-0.501169367089)*x_ref[0]+(0.447661946532)*x_ref[1]**o+(0.384528043888)*x_ref[1]+(-0.599271261928)*x_ref[2]**o+(-0.535447844217)*x_ref[2]
        ref[1,1,1]=(-0.457163465824)*x_ref[0]**o+(0.0475382663524)*x_ref[0]+(0.369626614251)*x_ref[1]**o+(0.1125159395)*x_ref[1]+(-0.252442923096)*x_ref[2]**o+(0.0717205095822)*x_ref[2]
        ref[2,0,0]=(0.341536174626)*x_ref[0]**o+(0.4408059057)*x_ref[0]+(0.196877383546)*x_ref[1]**o+(0.903072917866)*x_ref[1]+(0.423403716275)*x_ref[2]**o+(-0.521563534689)*x_ref[2]
        ref[2,0,1]=(0.553187361262)*x_ref[0]**o+(-0.539057536895)*x_ref[0]+(-0.814990210449)*x_ref[1]**o+(-0.418614789609)*x_ref[1]+(0.857298489434)*x_ref[2]**o+(0.629708672)*x_ref[2]
        ref[2,1,0]=(0.494803348038)*x_ref[0]**o+(0.143644562028)*x_ref[0]+(-0.0425813919235)*x_ref[1]**o+(-0.818562153097)*x_ref[1]+(-0.730690839032)*x_ref[2]**o+(-0.569426137866)*x_ref[2]
        ref[2,1,1]=(-0.573254727715)*x_ref[0]**o+(-0.536459946225)*x_ref[0]+(-0.244717967781)*x_ref[1]**o+(-0.67463285343)*x_ref[1]+(-0.383959754049)*x_ref[2]**o+(-0.87595126443)*x_ref[2]
        ref[3,0,0]=(0.758065154533)*x_ref[0]**o+(0.634049823109)*x_ref[0]+(0.851352543082)*x_ref[1]**o+(0.302790462394)*x_ref[1]+(-0.920790730269)*x_ref[2]**o+(0.460419670912)*x_ref[2]
        ref[3,0,1]=(0.415682634939)*x_ref[0]**o+(-0.664821834506)*x_ref[0]+(-0.314068887264)*x_ref[1]**o+(-0.322413102029)*x_ref[1]+(-0.498768865759)*x_ref[2]**o+(-0.165338012626)*x_ref[2]
        ref[3,1,0]=(-0.398599595126)*x_ref[0]**o+(0.744397086526)*x_ref[0]+(-0.02321839482)*x_ref[1]**o+(0.989743131098)*x_ref[1]+(0.0713981107777)*x_ref[2]**o+(-0.7862942966)*x_ref[2]
        ref[3,1,1]=(0.90835870295)*x_ref[0]**o+(0.285965355995)*x_ref[0]+(-0.302780058578)*x_ref[1]**o+(0.396923771953)*x_ref[1]+(0.717312324488)*x_ref[2]**o+(0.967488138744)*x_ref[2]
        ref[4,0,0]=(-0.286827561624)*x_ref[0]**o+(0.67973673275)*x_ref[0]+(0.325169520671)*x_ref[1]**o+(0.482002981846)*x_ref[1]+(-0.826037251282)*x_ref[2]**o+(-0.0567054906659)*x_ref[2]
        ref[4,0,1]=(-0.667297458554)*x_ref[0]**o+(-0.207369757609)*x_ref[0]+(-0.982746021703)*x_ref[1]**o+(0.963677089363)*x_ref[1]+(0.917328237047)*x_ref[2]**o+(-0.684386577728)*x_ref[2]
        ref[4,1,0]=(-0.835525763762)*x_ref[0]**o+(-0.113324184435)*x_ref[0]+(0.465855661969)*x_ref[1]**o+(0.656047307997)*x_ref[1]+(0.331675778117)*x_ref[2]**o+(-0.686067177565)*x_ref[2]
        ref[4,1,1]=(-0.468668496499)*x_ref[0]**o+(0.295197251458)*x_ref[0]+(0.065102699741)*x_ref[1]**o+(-0.378381176309)*x_ref[1]+(0.933627973066)*x_ref[2]**o+(-0.91462136931)*x_ref[2]
        ref[5,0,0]=(-0.161916994488)*x_ref[0]**o+(0.3070005414)*x_ref[0]+(0.195946524325)*x_ref[1]**o+(-0.453290146527)*x_ref[1]+(0.751532065651)*x_ref[2]**o+(0.82676125654)*x_ref[2]
        ref[5,0,1]=(-0.925362363492)*x_ref[0]**o+(-0.0295351854097)*x_ref[0]+(0.533980764998)*x_ref[1]**o+(0.128080489262)*x_ref[1]+(0.502858052015)*x_ref[2]**o+(0.813196540163)*x_ref[2]
        ref[5,1,0]=(-0.740146099318)*x_ref[0]**o+(0.27623528724)*x_ref[0]+(-0.361620073401)*x_ref[1]**o+(-0.310534085108)*x_ref[1]+(-0.0363428185757)*x_ref[2]**o+(-0.145397849357)*x_ref[2]
        ref[5,1,1]=(0.362089087335)*x_ref[0]**o+(0.539359659929)*x_ref[0]+(0.00904990097799)*x_ref[1]**o+(-0.150401175722)*x_ref[1]+(-0.20465882105)*x_ref[2]**o+(-0.519331232192)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.14016992168)*x[0]**o+(-0.396856222091)*x[0]+(0.747446299477)*x[1]**o+(-0.299063042467)*x[1]
        arg[0,0,0,1]=(0.126297805139)*x[0]**o+(0.317089000483)*x[0]+(0.0418478916266)*x[1]**o+(-0.892825689806)*x[1]
        arg[0,0,1,0]=(0.127263485532)*x[0]**o+(0.400880552184)*x[0]+(-0.858713979216)*x[1]**o+(0.238693631738)*x[1]
        arg[0,0,1,1]=(-0.985703521737)*x[0]**o+(-0.0192845851979)*x[0]+(-0.442493677109)*x[1]**o+(0.188038174612)*x[1]
        arg[0,0,2,0]=(0.814303535641)*x[0]**o+(-0.201938227136)*x[0]+(-0.134135136581)*x[1]**o+(0.939272592089)*x[1]
        arg[0,0,2,1]=(-0.609570412575)*x[0]**o+(0.181669701944)*x[0]+(0.882708213835)*x[1]**o+(0.315561722535)*x[1]
        arg[0,1,0,0]=(0.180272217085)*x[0]**o+(-0.921231698951)*x[0]+(-0.670170775864)*x[1]**o+(0.36959006719)*x[1]
        arg[0,1,0,1]=(-0.907447754449)*x[0]**o+(0.984356686322)*x[0]+(0.300621254386)*x[1]**o+(0.491995830235)*x[1]
        arg[0,1,1,0]=(-0.535808887861)*x[0]**o+(0.991647965865)*x[0]+(-0.505667913884)*x[1]**o+(-0.403245981754)*x[1]
        arg[0,1,1,1]=(-0.315687983887)*x[0]**o+(-0.570118136173)*x[0]+(0.327842038508)*x[1]**o+(-0.475833876716)*x[1]
        arg[0,1,2,0]=(0.598143043008)*x[0]**o+(0.119789453354)*x[0]+(-0.307994361027)*x[1]**o+(-0.747443620209)*x[1]
        arg[0,1,2,1]=(-0.384576138426)*x[0]**o+(-0.21431195392)*x[0]+(0.622157923623)*x[1]**o+(-0.535976986322)*x[1]
        arg[0,2,0,0]=(-0.823640169824)*x[0]**o+(-0.946436729867)*x[0]+(0.158819942161)*x[1]**o+(-0.175783146941)*x[1]
        arg[0,2,0,1]=(0.31586435609)*x[0]**o+(0.480823809413)*x[0]+(-0.155590673671)*x[1]**o+(-0.283774897002)*x[1]
        arg[0,2,1,0]=(0.973677960338)*x[0]**o+(-0.42119441385)*x[0]+(0.936434159723)*x[1]**o+(-0.271328279678)*x[1]
        arg[0,2,1,1]=(0.995977746066)*x[0]**o+(0.115566554312)*x[0]+(0.6023712777)*x[1]**o+(0.523708594385)*x[1]
        arg[0,2,2,0]=(0.690331991941)*x[0]**o+(0.708864222789)*x[0]+(-0.0209681974418)*x[1]**o+(-0.402664547241)*x[1]
        arg[0,2,2,1]=(0.793403379571)*x[0]**o+(-0.44195907862)*x[0]+(-0.750943552401)*x[1]**o+(0.87817105061)*x[1]
        arg[0,3,0,0]=(-0.960650344452)*x[0]**o+(0.918969286865)*x[0]+(-0.788611475275)*x[1]**o+(-0.917265819096)*x[1]
        arg[0,3,0,1]=(0.379234695969)*x[0]**o+(0.288784268667)*x[0]+(-0.231788371542)*x[1]**o+(-0.521488838978)*x[1]
        arg[0,3,1,0]=(-0.482353410226)*x[0]**o+(0.666269361063)*x[0]+(0.207628082266)*x[1]**o+(0.890536697637)*x[1]
        arg[0,3,1,1]=(-0.256826971067)*x[0]**o+(0.244573109457)*x[0]+(0.435182409565)*x[1]**o+(-0.98882467576)*x[1]
        arg[0,3,2,0]=(0.727681501266)*x[0]**o+(0.174117939756)*x[0]+(0.626998345198)*x[1]**o+(0.666518756047)*x[1]
        arg[0,3,2,1]=(-0.125855300157)*x[0]**o+(-0.750291117974)*x[0]+(0.840722993443)*x[1]**o+(0.0798848819291)*x[1]
        arg[0,4,0,0]=(-0.156808480304)*x[0]**o+(-0.864855000875)*x[0]+(-0.825786767174)*x[1]**o+(-0.563999207624)*x[1]
        arg[0,4,0,1]=(-0.890116284779)*x[0]**o+(0.795471417269)*x[0]+(0.326454331959)*x[1]**o+(-0.376644334859)*x[1]
        arg[0,4,1,0]=(-0.453181887597)*x[0]**o+(0.488149970589)*x[0]+(0.506352954527)*x[1]**o+(0.971016325918)*x[1]
        arg[0,4,1,1]=(0.156976484311)*x[0]**o+(0.988818166411)*x[0]+(0.948579296281)*x[1]**o+(-0.80303322078)*x[1]
        arg[0,4,2,0]=(-0.0625333956516)*x[0]**o+(0.357609677642)*x[0]+(0.00245182913993)*x[1]**o+(-0.162058851234)*x[1]
        arg[0,4,2,1]=(0.645504631508)*x[0]**o+(-0.0107543576265)*x[0]+(0.15490632437)*x[1]**o+(0.841985786269)*x[1]
        arg[1,0,0,0]=(0.703923364197)*x[0]**o+(-0.424818178123)*x[0]+(-0.744399116071)*x[1]**o+(0.303429987787)*x[1]
        arg[1,0,0,1]=(-0.92849671168)*x[0]**o+(0.050597602551)*x[0]+(0.91000243103)*x[1]**o+(-0.742775174195)*x[1]
        arg[1,0,1,0]=(0.281218131435)*x[0]**o+(0.171029842909)*x[0]+(-0.65806576709)*x[1]**o+(-0.718069456213)*x[1]
        arg[1,0,1,1]=(0.312255067232)*x[0]**o+(0.566401030665)*x[0]+(-0.941709776271)*x[1]**o+(0.0111896041808)*x[1]
        arg[1,0,2,0]=(-0.886019321094)*x[0]**o+(-0.688815756004)*x[0]+(-0.595491288141)*x[1]**o+(0.343402495621)*x[1]
        arg[1,0,2,1]=(-0.681063845863)*x[0]**o+(-0.676550403066)*x[0]+(0.186106532969)*x[1]**o+(0.181779537727)*x[1]
        arg[1,1,0,0]=(0.128588939989)*x[0]**o+(0.477711214734)*x[0]+(-0.17697460919)*x[1]**o+(0.0217208670244)*x[1]
        arg[1,1,0,1]=(-0.652028522296)*x[0]**o+(-0.503146425209)*x[0]+(-0.427973128189)*x[1]**o+(-0.182775422425)*x[1]
        arg[1,1,1,0]=(-0.0499349942904)*x[0]**o+(-0.0183804745613)*x[0]+(-0.912604094211)*x[1]**o+(-0.949181938689)*x[1]
        arg[1,1,1,1]=(-0.472737626621)*x[0]**o+(-0.142195314687)*x[0]+(0.580775749383)*x[1]**o+(-0.832549796746)*x[1]
        arg[1,1,2,0]=(-0.288756468146)*x[0]**o+(-0.50657056474)*x[0]+(-0.809713268333)*x[1]**o+(0.598450102365)*x[1]
        arg[1,1,2,1]=(-0.259779636241)*x[0]**o+(-0.227925001879)*x[0]+(-0.0752800702413)*x[1]**o+(-0.438841969161)*x[1]
        arg[1,2,0,0]=(-0.617835522076)*x[0]**o+(-0.969275786274)*x[0]+(0.899717033555)*x[1]**o+(0.77648944909)*x[1]
        arg[1,2,0,1]=(-0.0619395581419)*x[0]**o+(0.926394874574)*x[0]+(0.984605961354)*x[1]**o+(0.253347734181)*x[1]
        arg[1,2,1,0]=(0.748474342269)*x[0]**o+(0.0794003341104)*x[0]+(-0.244607682532)*x[1]**o+(-0.228787873564)*x[1]
        arg[1,2,1,1]=(-0.266853205536)*x[0]**o+(0.529714058158)*x[0]+(-0.421599080872)*x[1]**o+(-0.595557599245)*x[1]
        arg[1,2,2,0]=(-0.377962443753)*x[0]**o+(0.609550409525)*x[0]+(0.951904771097)*x[1]**o+(-0.123551801933)*x[1]
        arg[1,2,2,1]=(0.3262671136)*x[0]**o+(-0.574281078833)*x[0]+(0.308261022622)*x[1]**o+(-0.784762723843)*x[1]
        arg[1,3,0,0]=(-0.419601410829)*x[0]**o+(-0.576064062071)*x[0]+(-0.431574278091)*x[1]**o+(-0.264303418881)*x[1]
        arg[1,3,0,1]=(0.176058414693)*x[0]**o+(0.672752046411)*x[0]+(-0.8579992002)*x[1]**o+(-0.135904959185)*x[1]
        arg[1,3,1,0]=(0.84480795105)*x[0]**o+(0.846846799225)*x[0]+(0.134417432017)*x[1]**o+(-0.725316162128)*x[1]
        arg[1,3,1,1]=(-0.102162279538)*x[0]**o+(0.451180031789)*x[0]+(0.490175190736)*x[1]**o+(0.811823281177)*x[1]
        arg[1,3,2,0]=(-0.582177348293)*x[0]**o+(-0.74631626712)*x[0]+(-0.275524221012)*x[1]**o+(-0.658276720144)*x[1]
        arg[1,3,2,1]=(0.304511057297)*x[0]**o+(-0.354531411968)*x[0]+(0.0944491594546)*x[1]**o+(0.619039121597)*x[1]
        arg[1,4,0,0]=(0.754898520324)*x[0]**o+(-0.28914218119)*x[0]+(0.8660675945)*x[1]**o+(0.526777161228)*x[1]
        arg[1,4,0,1]=(-0.820150637429)*x[0]**o+(-0.399203443857)*x[0]+(0.118863382167)*x[1]**o+(0.269335037373)*x[1]
        arg[1,4,1,0]=(-0.669389232967)*x[0]**o+(-0.119080399209)*x[0]+(0.999919873616)*x[1]**o+(0.733390578378)*x[1]
        arg[1,4,1,1]=(-0.85147101348)*x[0]**o+(-0.0453996523642)*x[0]+(0.00891698330334)*x[1]**o+(-0.27128670805)*x[1]
        arg[1,4,2,0]=(-0.946558097166)*x[0]**o+(0.0141300604882)*x[0]+(0.04075090617)*x[1]**o+(-0.141095386138)*x[1]
        arg[1,4,2,1]=(0.613700159605)*x[0]**o+(0.115107829201)*x[0]+(0.907429968641)*x[1]**o+(0.901820725424)*x[1]
        arg[2,0,0,0]=(-0.629486588984)*x[0]**o+(-0.590908868493)*x[0]+(0.689333052519)*x[1]**o+(0.22096464697)*x[1]
        arg[2,0,0,1]=(-0.475364708492)*x[0]**o+(-0.3195126467)*x[0]+(-0.688082519853)*x[1]**o+(-0.889722517035)*x[1]
        arg[2,0,1,0]=(0.931261765284)*x[0]**o+(0.399101548845)*x[0]+(-0.185342345742)*x[1]**o+(0.188927724838)*x[1]
        arg[2,0,1,1]=(-0.195889209768)*x[0]**o+(-0.0878952476483)*x[0]+(0.636980893598)*x[1]**o+(0.827709469156)*x[1]
        arg[2,0,2,0]=(0.644658731705)*x[0]**o+(0.275771474382)*x[0]+(0.0802903363405)*x[1]**o+(-0.157785300253)*x[1]
        arg[2,0,2,1]=(0.324153739483)*x[0]**o+(-0.978043993375)*x[0]+(0.440060321393)*x[1]**o+(0.103952166755)*x[1]
        arg[2,1,0,0]=(-0.233567312787)*x[0]**o+(0.139529308019)*x[0]+(-0.200168759944)*x[1]**o+(-0.629057762953)*x[1]
        arg[2,1,0,1]=(-0.407808993281)*x[0]**o+(-0.208049914815)*x[0]+(0.676817990818)*x[1]**o+(-0.27174220248)*x[1]
        arg[2,1,1,0]=(-0.159113456499)*x[0]**o+(0.586749846436)*x[0]+(0.112658884741)*x[1]**o+(0.521917298039)*x[1]
        arg[2,1,1,1]=(-0.377986619436)*x[0]**o+(0.920912593255)*x[0]+(0.61285598265)*x[1]**o+(0.491316381102)*x[1]
        arg[2,1,2,0]=(-0.594573349393)*x[0]**o+(-0.784940398408)*x[0]+(0.0602427709094)*x[1]**o+(0.381334187216)*x[1]
        arg[2,1,2,1]=(-0.359402310466)*x[0]**o+(0.243014104221)*x[0]+(0.960587701667)*x[1]**o+(0.653124687952)*x[1]
        arg[2,2,0,0]=(0.667031182369)*x[0]**o+(0.099374903559)*x[0]+(0.00279841725989)*x[1]**o+(-0.413751543412)*x[1]
        arg[2,2,0,1]=(0.793557898241)*x[0]**o+(0.146572317359)*x[0]+(0.812964170008)*x[1]**o+(-0.988110793397)*x[1]
        arg[2,2,1,0]=(0.900661246137)*x[0]**o+(-0.997975978586)*x[0]+(0.63300171387)*x[1]**o+(-0.103878548778)*x[1]
        arg[2,2,1,1]=(-0.456464050469)*x[0]**o+(0.830699811714)*x[0]+(-0.186689927425)*x[1]**o+(-0.100170151164)*x[1]
        arg[2,2,2,0]=(0.732282428164)*x[0]**o+(0.928947589228)*x[0]+(-0.282773641579)*x[1]**o+(-0.927533112027)*x[1]
        arg[2,2,2,1]=(0.768111868708)*x[0]**o+(-0.308231472858)*x[0]+(0.890807565689)*x[1]**o+(-0.731196891492)*x[1]
        arg[2,3,0,0]=(0.559658559574)*x[0]**o+(0.104536615096)*x[0]+(-0.162594824125)*x[1]**o+(-0.172859200721)*x[1]
        arg[2,3,0,1]=(0.769297853152)*x[0]**o+(0.346598673904)*x[0]+(0.0313124779153)*x[1]**o+(-0.751767979251)*x[1]
        arg[2,3,1,0]=(0.870887050816)*x[0]**o+(0.593904859463)*x[0]+(-0.108083837899)*x[1]**o+(0.5467383682)*x[1]
        arg[2,3,1,1]=(0.244253227645)*x[0]**o+(0.911176738654)*x[0]+(-0.497855487714)*x[1]**o+(0.226443212585)*x[1]
        arg[2,3,2,0]=(-0.599893560596)*x[0]**o+(-0.893030581717)*x[0]+(-0.745249903413)*x[1]**o+(0.563740068619)*x[1]
        arg[2,3,2,1]=(-0.198088241103)*x[0]**o+(-0.171306730647)*x[0]+(-0.917417138299)*x[1]**o+(0.233920617125)*x[1]
        arg[2,4,0,0]=(0.804585270618)*x[0]**o+(0.3784879961)*x[0]+(0.621076476928)*x[1]**o+(-0.534671153952)*x[1]
        arg[2,4,0,1]=(0.380555159947)*x[0]**o+(-0.630487045841)*x[0]+(0.917413683072)*x[1]**o+(0.491925783834)*x[1]
        arg[2,4,1,0]=(0.946642644964)*x[0]**o+(-0.236438579482)*x[0]+(0.525585953739)*x[1]**o+(0.499940951133)*x[1]
        arg[2,4,1,1]=(0.827450889548)*x[0]**o+(-0.73264543626)*x[0]+(0.0626492901072)*x[1]**o+(-0.882759632862)*x[1]
        arg[2,4,2,0]=(-0.0671814496498)*x[0]**o+(0.108634446052)*x[0]+(-0.421081640868)*x[1]**o+(-0.546455336395)*x[1]
        arg[2,4,2,1]=(-0.601339784479)*x[0]**o+(-0.206937965916)*x[0]+(-0.849120792543)*x[1]**o+(0.615842037441)*x[1]
        arg[3,0,0,0]=(0.640466650913)*x[0]**o+(-0.936176165762)*x[0]+(0.12758686525)*x[1]**o+(-0.104771286099)*x[1]
        arg[3,0,0,1]=(-0.868109993006)*x[0]**o+(0.470815003671)*x[0]+(-0.301208743474)*x[1]**o+(-0.296271408095)*x[1]
        arg[3,0,1,0]=(0.66584252787)*x[0]**o+(-0.889549573986)*x[0]+(-0.0876076820512)*x[1]**o+(-0.249000029178)*x[1]
        arg[3,0,1,1]=(0.289388044606)*x[0]**o+(0.674519856916)*x[0]+(-0.721031236925)*x[1]**o+(0.093281544052)*x[1]
        arg[3,0,2,0]=(-0.0558863543119)*x[0]**o+(-0.367544430651)*x[0]+(-0.85580152565)*x[1]**o+(-0.140166367203)*x[1]
        arg[3,0,2,1]=(0.600709511675)*x[0]**o+(0.286322916564)*x[0]+(-0.544094007372)*x[1]**o+(0.470305833347)*x[1]
        arg[3,1,0,0]=(-0.828780197168)*x[0]**o+(-0.568096532278)*x[0]+(0.435127414299)*x[1]**o+(0.96890877255)*x[1]
        arg[3,1,0,1]=(0.191455614175)*x[0]**o+(-0.99507613827)*x[0]+(0.0735829936134)*x[1]**o+(0.635849313153)*x[1]
        arg[3,1,1,0]=(-0.839049013591)*x[0]**o+(-0.718731394184)*x[0]+(0.575853740085)*x[1]**o+(-0.57924857477)*x[1]
        arg[3,1,1,1]=(0.116876810453)*x[0]**o+(0.145727923885)*x[0]+(0.0155069936368)*x[1]**o+(0.223214885433)*x[1]
        arg[3,1,2,0]=(-0.674811720223)*x[0]**o+(0.651640807063)*x[0]+(-0.652022625703)*x[1]**o+(-0.119110788344)*x[1]
        arg[3,1,2,1]=(-0.544748794193)*x[0]**o+(-0.689440165483)*x[0]+(0.936900319026)*x[1]**o+(-0.337287311564)*x[1]
        arg[3,2,0,0]=(-0.355059804554)*x[0]**o+(0.458306329535)*x[0]+(0.172261907494)*x[1]**o+(0.498548138594)*x[1]
        arg[3,2,0,1]=(0.624420187129)*x[0]**o+(0.808931844722)*x[0]+(-0.690635894194)*x[1]**o+(-0.156122444554)*x[1]
        arg[3,2,1,0]=(-0.190029529776)*x[0]**o+(0.925600418137)*x[0]+(0.601040324953)*x[1]**o+(0.615485749173)*x[1]
        arg[3,2,1,1]=(0.679796674662)*x[0]**o+(0.879285148869)*x[0]+(0.0476373514288)*x[1]**o+(0.37655760501)*x[1]
        arg[3,2,2,0]=(0.231748550393)*x[0]**o+(0.209858138221)*x[0]+(-0.88372579866)*x[1]**o+(0.30534744967)*x[1]
        arg[3,2,2,1]=(-0.772177697671)*x[0]**o+(0.012439044435)*x[0]+(-0.293218976284)*x[1]**o+(0.306955066595)*x[1]
        arg[3,3,0,0]=(0.387256929042)*x[0]**o+(-0.434015953267)*x[0]+(-0.098736946868)*x[1]**o+(-0.0837712534186)*x[1]
        arg[3,3,0,1]=(0.396246252967)*x[0]**o+(-0.338955181605)*x[0]+(-0.353504371554)*x[1]**o+(-0.969811402293)*x[1]
        arg[3,3,1,0]=(-0.336508964723)*x[0]**o+(-0.411882233342)*x[0]+(-0.603965321132)*x[1]**o+(0.177221705416)*x[1]
        arg[3,3,1,1]=(0.786971174387)*x[0]**o+(0.759003522166)*x[0]+(-0.0961289050409)*x[1]**o+(-0.0811258682982)*x[1]
        arg[3,3,2,0]=(0.526266925685)*x[0]**o+(-0.113145219611)*x[0]+(-0.208427816884)*x[1]**o+(0.291484710079)*x[1]
        arg[3,3,2,1]=(-0.591153907468)*x[0]**o+(-0.342892047641)*x[0]+(0.865954283634)*x[1]**o+(-0.22112030903)*x[1]
        arg[3,4,0,0]=(0.755162128186)*x[0]**o+(0.494457469003)*x[0]+(0.889254517137)*x[1]**o+(0.880654056969)*x[1]
        arg[3,4,0,1]=(0.450983401629)*x[0]**o+(-0.0855489110466)*x[0]+(0.0601625202676)*x[1]**o+(-0.0730383237708)*x[1]
        arg[3,4,1,0]=(0.495514588456)*x[0]**o+(-0.0895269876334)*x[0]+(-0.89729630673)*x[1]**o+(0.759465768614)*x[1]
        arg[3,4,1,1]=(-0.666472982468)*x[0]**o+(-0.251715028171)*x[0]+(-0.0623287375223)*x[1]**o+(-0.895663326995)*x[1]
        arg[3,4,2,0]=(0.489737923112)*x[0]**o+(0.868066981889)*x[0]+(0.0371747456704)*x[1]**o+(-0.528951636028)*x[1]
        arg[3,4,2,1]=(0.556312598455)*x[0]**o+(-0.0979773350236)*x[0]+(-0.82785928691)*x[1]**o+(-0.567596765858)*x[1]
        ref[0,0,0,0]=(0.14016992168)*x_ref[0]**o+(-0.396856222091)*x_ref[0]+(0.747446299477)*x_ref[1]**o+(-0.299063042467)*x_ref[1]
        ref[0,0,0,1]=(0.126297805139)*x_ref[0]**o+(0.317089000483)*x_ref[0]+(0.0418478916266)*x_ref[1]**o+(-0.892825689806)*x_ref[1]
        ref[0,0,1,0]=(0.127263485532)*x_ref[0]**o+(0.400880552184)*x_ref[0]+(-0.858713979216)*x_ref[1]**o+(0.238693631738)*x_ref[1]
        ref[0,0,1,1]=(-0.985703521737)*x_ref[0]**o+(-0.0192845851979)*x_ref[0]+(-0.442493677109)*x_ref[1]**o+(0.188038174612)*x_ref[1]
        ref[0,0,2,0]=(0.814303535641)*x_ref[0]**o+(-0.201938227136)*x_ref[0]+(-0.134135136581)*x_ref[1]**o+(0.939272592089)*x_ref[1]
        ref[0,0,2,1]=(-0.609570412575)*x_ref[0]**o+(0.181669701944)*x_ref[0]+(0.882708213835)*x_ref[1]**o+(0.315561722535)*x_ref[1]
        ref[0,1,0,0]=(0.180272217085)*x_ref[0]**o+(-0.921231698951)*x_ref[0]+(-0.670170775864)*x_ref[1]**o+(0.36959006719)*x_ref[1]
        ref[0,1,0,1]=(-0.907447754449)*x_ref[0]**o+(0.984356686322)*x_ref[0]+(0.300621254386)*x_ref[1]**o+(0.491995830235)*x_ref[1]
        ref[0,1,1,0]=(-0.535808887861)*x_ref[0]**o+(0.991647965865)*x_ref[0]+(-0.505667913884)*x_ref[1]**o+(-0.403245981754)*x_ref[1]
        ref[0,1,1,1]=(-0.315687983887)*x_ref[0]**o+(-0.570118136173)*x_ref[0]+(0.327842038508)*x_ref[1]**o+(-0.475833876716)*x_ref[1]
        ref[0,1,2,0]=(0.598143043008)*x_ref[0]**o+(0.119789453354)*x_ref[0]+(-0.307994361027)*x_ref[1]**o+(-0.747443620209)*x_ref[1]
        ref[0,1,2,1]=(-0.384576138426)*x_ref[0]**o+(-0.21431195392)*x_ref[0]+(0.622157923623)*x_ref[1]**o+(-0.535976986322)*x_ref[1]
        ref[0,2,0,0]=(-0.823640169824)*x_ref[0]**o+(-0.946436729867)*x_ref[0]+(0.158819942161)*x_ref[1]**o+(-0.175783146941)*x_ref[1]
        ref[0,2,0,1]=(0.31586435609)*x_ref[0]**o+(0.480823809413)*x_ref[0]+(-0.155590673671)*x_ref[1]**o+(-0.283774897002)*x_ref[1]
        ref[0,2,1,0]=(0.973677960338)*x_ref[0]**o+(-0.42119441385)*x_ref[0]+(0.936434159723)*x_ref[1]**o+(-0.271328279678)*x_ref[1]
        ref[0,2,1,1]=(0.995977746066)*x_ref[0]**o+(0.115566554312)*x_ref[0]+(0.6023712777)*x_ref[1]**o+(0.523708594385)*x_ref[1]
        ref[0,2,2,0]=(0.690331991941)*x_ref[0]**o+(0.708864222789)*x_ref[0]+(-0.0209681974418)*x_ref[1]**o+(-0.402664547241)*x_ref[1]
        ref[0,2,2,1]=(0.793403379571)*x_ref[0]**o+(-0.44195907862)*x_ref[0]+(-0.750943552401)*x_ref[1]**o+(0.87817105061)*x_ref[1]
        ref[0,3,0,0]=(-0.960650344452)*x_ref[0]**o+(0.918969286865)*x_ref[0]+(-0.788611475275)*x_ref[1]**o+(-0.917265819096)*x_ref[1]
        ref[0,3,0,1]=(0.379234695969)*x_ref[0]**o+(0.288784268667)*x_ref[0]+(-0.231788371542)*x_ref[1]**o+(-0.521488838978)*x_ref[1]
        ref[0,3,1,0]=(-0.482353410226)*x_ref[0]**o+(0.666269361063)*x_ref[0]+(0.207628082266)*x_ref[1]**o+(0.890536697637)*x_ref[1]
        ref[0,3,1,1]=(-0.256826971067)*x_ref[0]**o+(0.244573109457)*x_ref[0]+(0.435182409565)*x_ref[1]**o+(-0.98882467576)*x_ref[1]
        ref[0,3,2,0]=(0.727681501266)*x_ref[0]**o+(0.174117939756)*x_ref[0]+(0.626998345198)*x_ref[1]**o+(0.666518756047)*x_ref[1]
        ref[0,3,2,1]=(-0.125855300157)*x_ref[0]**o+(-0.750291117974)*x_ref[0]+(0.840722993443)*x_ref[1]**o+(0.0798848819291)*x_ref[1]
        ref[0,4,0,0]=(-0.156808480304)*x_ref[0]**o+(-0.864855000875)*x_ref[0]+(-0.825786767174)*x_ref[1]**o+(-0.563999207624)*x_ref[1]
        ref[0,4,0,1]=(-0.890116284779)*x_ref[0]**o+(0.795471417269)*x_ref[0]+(0.326454331959)*x_ref[1]**o+(-0.376644334859)*x_ref[1]
        ref[0,4,1,0]=(-0.453181887597)*x_ref[0]**o+(0.488149970589)*x_ref[0]+(0.506352954527)*x_ref[1]**o+(0.971016325918)*x_ref[1]
        ref[0,4,1,1]=(0.156976484311)*x_ref[0]**o+(0.988818166411)*x_ref[0]+(0.948579296281)*x_ref[1]**o+(-0.80303322078)*x_ref[1]
        ref[0,4,2,0]=(-0.0625333956516)*x_ref[0]**o+(0.357609677642)*x_ref[0]+(0.00245182913993)*x_ref[1]**o+(-0.162058851234)*x_ref[1]
        ref[0,4,2,1]=(0.645504631508)*x_ref[0]**o+(-0.0107543576265)*x_ref[0]+(0.15490632437)*x_ref[1]**o+(0.841985786269)*x_ref[1]
        ref[1,0,0,0]=(0.703923364197)*x_ref[0]**o+(-0.424818178123)*x_ref[0]+(-0.744399116071)*x_ref[1]**o+(0.303429987787)*x_ref[1]
        ref[1,0,0,1]=(-0.92849671168)*x_ref[0]**o+(0.050597602551)*x_ref[0]+(0.91000243103)*x_ref[1]**o+(-0.742775174195)*x_ref[1]
        ref[1,0,1,0]=(0.281218131435)*x_ref[0]**o+(0.171029842909)*x_ref[0]+(-0.65806576709)*x_ref[1]**o+(-0.718069456213)*x_ref[1]
        ref[1,0,1,1]=(0.312255067232)*x_ref[0]**o+(0.566401030665)*x_ref[0]+(-0.941709776271)*x_ref[1]**o+(0.0111896041808)*x_ref[1]
        ref[1,0,2,0]=(-0.886019321094)*x_ref[0]**o+(-0.688815756004)*x_ref[0]+(-0.595491288141)*x_ref[1]**o+(0.343402495621)*x_ref[1]
        ref[1,0,2,1]=(-0.681063845863)*x_ref[0]**o+(-0.676550403066)*x_ref[0]+(0.186106532969)*x_ref[1]**o+(0.181779537727)*x_ref[1]
        ref[1,1,0,0]=(0.128588939989)*x_ref[0]**o+(0.477711214734)*x_ref[0]+(-0.17697460919)*x_ref[1]**o+(0.0217208670244)*x_ref[1]
        ref[1,1,0,1]=(-0.652028522296)*x_ref[0]**o+(-0.503146425209)*x_ref[0]+(-0.427973128189)*x_ref[1]**o+(-0.182775422425)*x_ref[1]
        ref[1,1,1,0]=(-0.0499349942904)*x_ref[0]**o+(-0.0183804745613)*x_ref[0]+(-0.912604094211)*x_ref[1]**o+(-0.949181938689)*x_ref[1]
        ref[1,1,1,1]=(-0.472737626621)*x_ref[0]**o+(-0.142195314687)*x_ref[0]+(0.580775749383)*x_ref[1]**o+(-0.832549796746)*x_ref[1]
        ref[1,1,2,0]=(-0.288756468146)*x_ref[0]**o+(-0.50657056474)*x_ref[0]+(-0.809713268333)*x_ref[1]**o+(0.598450102365)*x_ref[1]
        ref[1,1,2,1]=(-0.259779636241)*x_ref[0]**o+(-0.227925001879)*x_ref[0]+(-0.0752800702413)*x_ref[1]**o+(-0.438841969161)*x_ref[1]
        ref[1,2,0,0]=(-0.617835522076)*x_ref[0]**o+(-0.969275786274)*x_ref[0]+(0.899717033555)*x_ref[1]**o+(0.77648944909)*x_ref[1]
        ref[1,2,0,1]=(-0.0619395581419)*x_ref[0]**o+(0.926394874574)*x_ref[0]+(0.984605961354)*x_ref[1]**o+(0.253347734181)*x_ref[1]
        ref[1,2,1,0]=(0.748474342269)*x_ref[0]**o+(0.0794003341104)*x_ref[0]+(-0.244607682532)*x_ref[1]**o+(-0.228787873564)*x_ref[1]
        ref[1,2,1,1]=(-0.266853205536)*x_ref[0]**o+(0.529714058158)*x_ref[0]+(-0.421599080872)*x_ref[1]**o+(-0.595557599245)*x_ref[1]
        ref[1,2,2,0]=(-0.377962443753)*x_ref[0]**o+(0.609550409525)*x_ref[0]+(0.951904771097)*x_ref[1]**o+(-0.123551801933)*x_ref[1]
        ref[1,2,2,1]=(0.3262671136)*x_ref[0]**o+(-0.574281078833)*x_ref[0]+(0.308261022622)*x_ref[1]**o+(-0.784762723843)*x_ref[1]
        ref[1,3,0,0]=(-0.419601410829)*x_ref[0]**o+(-0.576064062071)*x_ref[0]+(-0.431574278091)*x_ref[1]**o+(-0.264303418881)*x_ref[1]
        ref[1,3,0,1]=(0.176058414693)*x_ref[0]**o+(0.672752046411)*x_ref[0]+(-0.8579992002)*x_ref[1]**o+(-0.135904959185)*x_ref[1]
        ref[1,3,1,0]=(0.84480795105)*x_ref[0]**o+(0.846846799225)*x_ref[0]+(0.134417432017)*x_ref[1]**o+(-0.725316162128)*x_ref[1]
        ref[1,3,1,1]=(-0.102162279538)*x_ref[0]**o+(0.451180031789)*x_ref[0]+(0.490175190736)*x_ref[1]**o+(0.811823281177)*x_ref[1]
        ref[1,3,2,0]=(-0.582177348293)*x_ref[0]**o+(-0.74631626712)*x_ref[0]+(-0.275524221012)*x_ref[1]**o+(-0.658276720144)*x_ref[1]
        ref[1,3,2,1]=(0.304511057297)*x_ref[0]**o+(-0.354531411968)*x_ref[0]+(0.0944491594546)*x_ref[1]**o+(0.619039121597)*x_ref[1]
        ref[1,4,0,0]=(0.754898520324)*x_ref[0]**o+(-0.28914218119)*x_ref[0]+(0.8660675945)*x_ref[1]**o+(0.526777161228)*x_ref[1]
        ref[1,4,0,1]=(-0.820150637429)*x_ref[0]**o+(-0.399203443857)*x_ref[0]+(0.118863382167)*x_ref[1]**o+(0.269335037373)*x_ref[1]
        ref[1,4,1,0]=(-0.669389232967)*x_ref[0]**o+(-0.119080399209)*x_ref[0]+(0.999919873616)*x_ref[1]**o+(0.733390578378)*x_ref[1]
        ref[1,4,1,1]=(-0.85147101348)*x_ref[0]**o+(-0.0453996523642)*x_ref[0]+(0.00891698330334)*x_ref[1]**o+(-0.27128670805)*x_ref[1]
        ref[1,4,2,0]=(-0.946558097166)*x_ref[0]**o+(0.0141300604882)*x_ref[0]+(0.04075090617)*x_ref[1]**o+(-0.141095386138)*x_ref[1]
        ref[1,4,2,1]=(0.613700159605)*x_ref[0]**o+(0.115107829201)*x_ref[0]+(0.907429968641)*x_ref[1]**o+(0.901820725424)*x_ref[1]
        ref[2,0,0,0]=(-0.629486588984)*x_ref[0]**o+(-0.590908868493)*x_ref[0]+(0.689333052519)*x_ref[1]**o+(0.22096464697)*x_ref[1]
        ref[2,0,0,1]=(-0.475364708492)*x_ref[0]**o+(-0.3195126467)*x_ref[0]+(-0.688082519853)*x_ref[1]**o+(-0.889722517035)*x_ref[1]
        ref[2,0,1,0]=(0.931261765284)*x_ref[0]**o+(0.399101548845)*x_ref[0]+(-0.185342345742)*x_ref[1]**o+(0.188927724838)*x_ref[1]
        ref[2,0,1,1]=(-0.195889209768)*x_ref[0]**o+(-0.0878952476483)*x_ref[0]+(0.636980893598)*x_ref[1]**o+(0.827709469156)*x_ref[1]
        ref[2,0,2,0]=(0.644658731705)*x_ref[0]**o+(0.275771474382)*x_ref[0]+(0.0802903363405)*x_ref[1]**o+(-0.157785300253)*x_ref[1]
        ref[2,0,2,1]=(0.324153739483)*x_ref[0]**o+(-0.978043993375)*x_ref[0]+(0.440060321393)*x_ref[1]**o+(0.103952166755)*x_ref[1]
        ref[2,1,0,0]=(-0.233567312787)*x_ref[0]**o+(0.139529308019)*x_ref[0]+(-0.200168759944)*x_ref[1]**o+(-0.629057762953)*x_ref[1]
        ref[2,1,0,1]=(-0.407808993281)*x_ref[0]**o+(-0.208049914815)*x_ref[0]+(0.676817990818)*x_ref[1]**o+(-0.27174220248)*x_ref[1]
        ref[2,1,1,0]=(-0.159113456499)*x_ref[0]**o+(0.586749846436)*x_ref[0]+(0.112658884741)*x_ref[1]**o+(0.521917298039)*x_ref[1]
        ref[2,1,1,1]=(-0.377986619436)*x_ref[0]**o+(0.920912593255)*x_ref[0]+(0.61285598265)*x_ref[1]**o+(0.491316381102)*x_ref[1]
        ref[2,1,2,0]=(-0.594573349393)*x_ref[0]**o+(-0.784940398408)*x_ref[0]+(0.0602427709094)*x_ref[1]**o+(0.381334187216)*x_ref[1]
        ref[2,1,2,1]=(-0.359402310466)*x_ref[0]**o+(0.243014104221)*x_ref[0]+(0.960587701667)*x_ref[1]**o+(0.653124687952)*x_ref[1]
        ref[2,2,0,0]=(0.667031182369)*x_ref[0]**o+(0.099374903559)*x_ref[0]+(0.00279841725989)*x_ref[1]**o+(-0.413751543412)*x_ref[1]
        ref[2,2,0,1]=(0.793557898241)*x_ref[0]**o+(0.146572317359)*x_ref[0]+(0.812964170008)*x_ref[1]**o+(-0.988110793397)*x_ref[1]
        ref[2,2,1,0]=(0.900661246137)*x_ref[0]**o+(-0.997975978586)*x_ref[0]+(0.63300171387)*x_ref[1]**o+(-0.103878548778)*x_ref[1]
        ref[2,2,1,1]=(-0.456464050469)*x_ref[0]**o+(0.830699811714)*x_ref[0]+(-0.186689927425)*x_ref[1]**o+(-0.100170151164)*x_ref[1]
        ref[2,2,2,0]=(0.732282428164)*x_ref[0]**o+(0.928947589228)*x_ref[0]+(-0.282773641579)*x_ref[1]**o+(-0.927533112027)*x_ref[1]
        ref[2,2,2,1]=(0.768111868708)*x_ref[0]**o+(-0.308231472858)*x_ref[0]+(0.890807565689)*x_ref[1]**o+(-0.731196891492)*x_ref[1]
        ref[2,3,0,0]=(0.559658559574)*x_ref[0]**o+(0.104536615096)*x_ref[0]+(-0.162594824125)*x_ref[1]**o+(-0.172859200721)*x_ref[1]
        ref[2,3,0,1]=(0.769297853152)*x_ref[0]**o+(0.346598673904)*x_ref[0]+(0.0313124779153)*x_ref[1]**o+(-0.751767979251)*x_ref[1]
        ref[2,3,1,0]=(0.870887050816)*x_ref[0]**o+(0.593904859463)*x_ref[0]+(-0.108083837899)*x_ref[1]**o+(0.5467383682)*x_ref[1]
        ref[2,3,1,1]=(0.244253227645)*x_ref[0]**o+(0.911176738654)*x_ref[0]+(-0.497855487714)*x_ref[1]**o+(0.226443212585)*x_ref[1]
        ref[2,3,2,0]=(-0.599893560596)*x_ref[0]**o+(-0.893030581717)*x_ref[0]+(-0.745249903413)*x_ref[1]**o+(0.563740068619)*x_ref[1]
        ref[2,3,2,1]=(-0.198088241103)*x_ref[0]**o+(-0.171306730647)*x_ref[0]+(-0.917417138299)*x_ref[1]**o+(0.233920617125)*x_ref[1]
        ref[2,4,0,0]=(0.804585270618)*x_ref[0]**o+(0.3784879961)*x_ref[0]+(0.621076476928)*x_ref[1]**o+(-0.534671153952)*x_ref[1]
        ref[2,4,0,1]=(0.380555159947)*x_ref[0]**o+(-0.630487045841)*x_ref[0]+(0.917413683072)*x_ref[1]**o+(0.491925783834)*x_ref[1]
        ref[2,4,1,0]=(0.946642644964)*x_ref[0]**o+(-0.236438579482)*x_ref[0]+(0.525585953739)*x_ref[1]**o+(0.499940951133)*x_ref[1]
        ref[2,4,1,1]=(0.827450889548)*x_ref[0]**o+(-0.73264543626)*x_ref[0]+(0.0626492901072)*x_ref[1]**o+(-0.882759632862)*x_ref[1]
        ref[2,4,2,0]=(-0.0671814496498)*x_ref[0]**o+(0.108634446052)*x_ref[0]+(-0.421081640868)*x_ref[1]**o+(-0.546455336395)*x_ref[1]
        ref[2,4,2,1]=(-0.601339784479)*x_ref[0]**o+(-0.206937965916)*x_ref[0]+(-0.849120792543)*x_ref[1]**o+(0.615842037441)*x_ref[1]
        ref[3,0,0,0]=(0.640466650913)*x_ref[0]**o+(-0.936176165762)*x_ref[0]+(0.12758686525)*x_ref[1]**o+(-0.104771286099)*x_ref[1]
        ref[3,0,0,1]=(-0.868109993006)*x_ref[0]**o+(0.470815003671)*x_ref[0]+(-0.301208743474)*x_ref[1]**o+(-0.296271408095)*x_ref[1]
        ref[3,0,1,0]=(0.66584252787)*x_ref[0]**o+(-0.889549573986)*x_ref[0]+(-0.0876076820512)*x_ref[1]**o+(-0.249000029178)*x_ref[1]
        ref[3,0,1,1]=(0.289388044606)*x_ref[0]**o+(0.674519856916)*x_ref[0]+(-0.721031236925)*x_ref[1]**o+(0.093281544052)*x_ref[1]
        ref[3,0,2,0]=(-0.0558863543119)*x_ref[0]**o+(-0.367544430651)*x_ref[0]+(-0.85580152565)*x_ref[1]**o+(-0.140166367203)*x_ref[1]
        ref[3,0,2,1]=(0.600709511675)*x_ref[0]**o+(0.286322916564)*x_ref[0]+(-0.544094007372)*x_ref[1]**o+(0.470305833347)*x_ref[1]
        ref[3,1,0,0]=(-0.828780197168)*x_ref[0]**o+(-0.568096532278)*x_ref[0]+(0.435127414299)*x_ref[1]**o+(0.96890877255)*x_ref[1]
        ref[3,1,0,1]=(0.191455614175)*x_ref[0]**o+(-0.99507613827)*x_ref[0]+(0.0735829936134)*x_ref[1]**o+(0.635849313153)*x_ref[1]
        ref[3,1,1,0]=(-0.839049013591)*x_ref[0]**o+(-0.718731394184)*x_ref[0]+(0.575853740085)*x_ref[1]**o+(-0.57924857477)*x_ref[1]
        ref[3,1,1,1]=(0.116876810453)*x_ref[0]**o+(0.145727923885)*x_ref[0]+(0.0155069936368)*x_ref[1]**o+(0.223214885433)*x_ref[1]
        ref[3,1,2,0]=(-0.674811720223)*x_ref[0]**o+(0.651640807063)*x_ref[0]+(-0.652022625703)*x_ref[1]**o+(-0.119110788344)*x_ref[1]
        ref[3,1,2,1]=(-0.544748794193)*x_ref[0]**o+(-0.689440165483)*x_ref[0]+(0.936900319026)*x_ref[1]**o+(-0.337287311564)*x_ref[1]
        ref[3,2,0,0]=(-0.355059804554)*x_ref[0]**o+(0.458306329535)*x_ref[0]+(0.172261907494)*x_ref[1]**o+(0.498548138594)*x_ref[1]
        ref[3,2,0,1]=(0.624420187129)*x_ref[0]**o+(0.808931844722)*x_ref[0]+(-0.690635894194)*x_ref[1]**o+(-0.156122444554)*x_ref[1]
        ref[3,2,1,0]=(-0.190029529776)*x_ref[0]**o+(0.925600418137)*x_ref[0]+(0.601040324953)*x_ref[1]**o+(0.615485749173)*x_ref[1]
        ref[3,2,1,1]=(0.679796674662)*x_ref[0]**o+(0.879285148869)*x_ref[0]+(0.0476373514288)*x_ref[1]**o+(0.37655760501)*x_ref[1]
        ref[3,2,2,0]=(0.231748550393)*x_ref[0]**o+(0.209858138221)*x_ref[0]+(-0.88372579866)*x_ref[1]**o+(0.30534744967)*x_ref[1]
        ref[3,2,2,1]=(-0.772177697671)*x_ref[0]**o+(0.012439044435)*x_ref[0]+(-0.293218976284)*x_ref[1]**o+(0.306955066595)*x_ref[1]
        ref[3,3,0,0]=(0.387256929042)*x_ref[0]**o+(-0.434015953267)*x_ref[0]+(-0.098736946868)*x_ref[1]**o+(-0.0837712534186)*x_ref[1]
        ref[3,3,0,1]=(0.396246252967)*x_ref[0]**o+(-0.338955181605)*x_ref[0]+(-0.353504371554)*x_ref[1]**o+(-0.969811402293)*x_ref[1]
        ref[3,3,1,0]=(-0.336508964723)*x_ref[0]**o+(-0.411882233342)*x_ref[0]+(-0.603965321132)*x_ref[1]**o+(0.177221705416)*x_ref[1]
        ref[3,3,1,1]=(0.786971174387)*x_ref[0]**o+(0.759003522166)*x_ref[0]+(-0.0961289050409)*x_ref[1]**o+(-0.0811258682982)*x_ref[1]
        ref[3,3,2,0]=(0.526266925685)*x_ref[0]**o+(-0.113145219611)*x_ref[0]+(-0.208427816884)*x_ref[1]**o+(0.291484710079)*x_ref[1]
        ref[3,3,2,1]=(-0.591153907468)*x_ref[0]**o+(-0.342892047641)*x_ref[0]+(0.865954283634)*x_ref[1]**o+(-0.22112030903)*x_ref[1]
        ref[3,4,0,0]=(0.755162128186)*x_ref[0]**o+(0.494457469003)*x_ref[0]+(0.889254517137)*x_ref[1]**o+(0.880654056969)*x_ref[1]
        ref[3,4,0,1]=(0.450983401629)*x_ref[0]**o+(-0.0855489110466)*x_ref[0]+(0.0601625202676)*x_ref[1]**o+(-0.0730383237708)*x_ref[1]
        ref[3,4,1,0]=(0.495514588456)*x_ref[0]**o+(-0.0895269876334)*x_ref[0]+(-0.89729630673)*x_ref[1]**o+(0.759465768614)*x_ref[1]
        ref[3,4,1,1]=(-0.666472982468)*x_ref[0]**o+(-0.251715028171)*x_ref[0]+(-0.0623287375223)*x_ref[1]**o+(-0.895663326995)*x_ref[1]
        ref[3,4,2,0]=(0.489737923112)*x_ref[0]**o+(0.868066981889)*x_ref[0]+(0.0371747456704)*x_ref[1]**o+(-0.528951636028)*x_ref[1]
        ref[3,4,2,1]=(0.556312598455)*x_ref[0]**o+(-0.0979773350236)*x_ref[0]+(-0.82785928691)*x_ref[1]**o+(-0.567596765858)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.952750275738)*x[0]**o+(0.971489204087)*x[0]+(-0.320706487584)*x[1]**o+(-0.853260471578)*x[1]+(-0.696434862068)*x[2]**o+(0.641824397433)*x[2]
        arg[0,0,0,1]=(0.115484192786)*x[0]**o+(0.147813687309)*x[0]+(-0.998010548186)*x[1]**o+(0.0274712088882)*x[1]+(0.136937971489)*x[2]**o+(-0.657986135444)*x[2]
        arg[0,0,1,0]=(0.805857504779)*x[0]**o+(0.196278916938)*x[0]+(-0.969340785263)*x[1]**o+(-0.543890174888)*x[1]+(0.49118423357)*x[2]**o+(-0.883493800593)*x[2]
        arg[0,0,1,1]=(-0.808802136474)*x[0]**o+(-0.866039395649)*x[0]+(0.0339207025941)*x[1]**o+(-0.441368773534)*x[1]+(0.673688211653)*x[2]**o+(0.917894389398)*x[2]
        arg[0,0,2,0]=(0.0459902210078)*x[0]**o+(0.862332666528)*x[0]+(-0.547792306389)*x[1]**o+(0.722176431732)*x[1]+(-0.511278989032)*x[2]**o+(-0.692775336577)*x[2]
        arg[0,0,2,1]=(-0.641904406513)*x[0]**o+(-0.854450656063)*x[0]+(0.81138894849)*x[1]**o+(0.328886625074)*x[1]+(0.400935831363)*x[2]**o+(-0.25184059766)*x[2]
        arg[0,1,0,0]=(0.69922840942)*x[0]**o+(-0.798085508443)*x[0]+(-0.85191383431)*x[1]**o+(-0.162940829331)*x[1]+(0.504020642931)*x[2]**o+(0.0950957436173)*x[2]
        arg[0,1,0,1]=(0.28901429037)*x[0]**o+(-0.624508745729)*x[0]+(0.921288077478)*x[1]**o+(-0.769207901119)*x[1]+(0.0852374599763)*x[2]**o+(-0.402583033986)*x[2]
        arg[0,1,1,0]=(0.715817282215)*x[0]**o+(0.68735256665)*x[0]+(0.610212058649)*x[1]**o+(0.53560824609)*x[1]+(-0.238908897264)*x[2]**o+(0.619799158474)*x[2]
        arg[0,1,1,1]=(-0.371797414191)*x[0]**o+(0.106263350514)*x[0]+(0.583768355362)*x[1]**o+(0.818797667369)*x[1]+(0.0567571152493)*x[2]**o+(-0.242244867169)*x[2]
        arg[0,1,2,0]=(0.190480659729)*x[0]**o+(-0.632415604771)*x[0]+(0.252513826138)*x[1]**o+(0.874732840469)*x[1]+(-0.367769326078)*x[2]**o+(0.364453292404)*x[2]
        arg[0,1,2,1]=(-0.704507136581)*x[0]**o+(-0.614211737547)*x[0]+(0.533383507313)*x[1]**o+(-0.940007583478)*x[1]+(0.558260002746)*x[2]**o+(0.742565692751)*x[2]
        arg[0,2,0,0]=(0.132581156766)*x[0]**o+(0.671793183183)*x[0]+(-0.308255110769)*x[1]**o+(-0.956002957061)*x[1]+(0.300091541301)*x[2]**o+(-0.714260746791)*x[2]
        arg[0,2,0,1]=(-0.836648080859)*x[0]**o+(0.123813861017)*x[0]+(0.866569934894)*x[1]**o+(-0.957824572351)*x[1]+(-0.904478410952)*x[2]**o+(0.691673758521)*x[2]
        arg[0,2,1,0]=(0.411624721878)*x[0]**o+(0.594054722352)*x[0]+(0.150177701135)*x[1]**o+(0.945855261084)*x[1]+(0.295930158817)*x[2]**o+(-0.607980954424)*x[2]
        arg[0,2,1,1]=(-0.0777741308373)*x[0]**o+(0.697501295655)*x[0]+(0.981797797329)*x[1]**o+(-0.671827077949)*x[1]+(0.292310146229)*x[2]**o+(0.484652357857)*x[2]
        arg[0,2,2,0]=(0.194910469807)*x[0]**o+(-0.70008515623)*x[0]+(0.192808336639)*x[1]**o+(-0.962764477371)*x[1]+(-0.34292240753)*x[2]**o+(0.525470829202)*x[2]
        arg[0,2,2,1]=(0.850368420228)*x[0]**o+(-0.516355386368)*x[0]+(0.395764367873)*x[1]**o+(0.320830752607)*x[1]+(-0.26288679044)*x[2]**o+(0.174457723191)*x[2]
        arg[0,3,0,0]=(-0.713987556845)*x[0]**o+(0.0391066248253)*x[0]+(-0.63911733479)*x[1]**o+(-0.119953032121)*x[1]+(0.697879836447)*x[2]**o+(-0.474957057702)*x[2]
        arg[0,3,0,1]=(-0.128614897423)*x[0]**o+(0.7166036403)*x[0]+(0.678415978447)*x[1]**o+(-0.191453169618)*x[1]+(-0.755261361956)*x[2]**o+(-0.839956004196)*x[2]
        arg[0,3,1,0]=(0.258158740934)*x[0]**o+(-0.412199346967)*x[0]+(-0.781106341272)*x[1]**o+(-0.561162823101)*x[1]+(-0.59549864377)*x[2]**o+(-0.748038415721)*x[2]
        arg[0,3,1,1]=(0.68950222305)*x[0]**o+(-0.904027774445)*x[0]+(0.769744453441)*x[1]**o+(-0.265549152752)*x[1]+(-0.375978837131)*x[2]**o+(-0.387010981761)*x[2]
        arg[0,3,2,0]=(0.46852681319)*x[0]**o+(-0.887650258615)*x[0]+(-0.780275624319)*x[1]**o+(0.527897372599)*x[1]+(0.889281778411)*x[2]**o+(0.82538563877)*x[2]
        arg[0,3,2,1]=(-0.817514448184)*x[0]**o+(-0.230190768519)*x[0]+(-0.101550849344)*x[1]**o+(0.1654885857)*x[1]+(-0.231317749498)*x[2]**o+(0.580153306395)*x[2]
        arg[0,4,0,0]=(0.394993007242)*x[0]**o+(0.305948764395)*x[0]+(-0.66853441368)*x[1]**o+(0.279027498094)*x[1]+(0.161366934709)*x[2]**o+(-0.742002471552)*x[2]
        arg[0,4,0,1]=(-0.661214709471)*x[0]**o+(0.548880317935)*x[0]+(0.472682748503)*x[1]**o+(0.266797505699)*x[1]+(0.405618413683)*x[2]**o+(-0.0539677542667)*x[2]
        arg[0,4,1,0]=(-0.713892424429)*x[0]**o+(0.568075831592)*x[0]+(-0.956180895599)*x[1]**o+(0.63618957695)*x[1]+(0.983696160047)*x[2]**o+(-0.308157929002)*x[2]
        arg[0,4,1,1]=(-0.666582919353)*x[0]**o+(0.0816884254484)*x[0]+(0.190810048896)*x[1]**o+(0.482322905132)*x[1]+(-0.576597070954)*x[2]**o+(-0.722381611553)*x[2]
        arg[0,4,2,0]=(-0.0229650731041)*x[0]**o+(-0.0867816754979)*x[0]+(0.0905042102365)*x[1]**o+(-0.994963424765)*x[1]+(0.74504757815)*x[2]**o+(0.71462549644)*x[2]
        arg[0,4,2,1]=(-0.0464278496147)*x[0]**o+(0.122501026996)*x[0]+(-0.049450141901)*x[1]**o+(-0.388739464086)*x[1]+(-0.151389885067)*x[2]**o+(-0.0316948088803)*x[2]
        arg[1,0,0,0]=(-0.303653386851)*x[0]**o+(-0.435866753659)*x[0]+(-0.419078175082)*x[1]**o+(-0.907595244917)*x[1]+(0.442156528837)*x[2]**o+(-0.76481785523)*x[2]
        arg[1,0,0,1]=(0.879454551618)*x[0]**o+(-0.530789858588)*x[0]+(0.942203776542)*x[1]**o+(0.864362935436)*x[1]+(-0.252832164588)*x[2]**o+(0.00758252581015)*x[2]
        arg[1,0,1,0]=(0.522433142898)*x[0]**o+(0.564002257569)*x[0]+(0.149811689377)*x[1]**o+(-0.757726948165)*x[1]+(-0.767052869719)*x[2]**o+(0.743541806001)*x[2]
        arg[1,0,1,1]=(0.535612198186)*x[0]**o+(0.0874480088401)*x[0]+(0.913151292296)*x[1]**o+(-0.495024655308)*x[1]+(-0.42308905182)*x[2]**o+(-0.983623551153)*x[2]
        arg[1,0,2,0]=(0.587221456352)*x[0]**o+(0.0384438535009)*x[0]+(0.301554087927)*x[1]**o+(-0.944074061806)*x[1]+(0.808195432991)*x[2]**o+(-0.46545743768)*x[2]
        arg[1,0,2,1]=(0.136251169669)*x[0]**o+(0.772190271875)*x[0]+(-0.490981224022)*x[1]**o+(-0.191915588448)*x[1]+(0.792617864405)*x[2]**o+(0.139425359773)*x[2]
        arg[1,1,0,0]=(0.0440493146329)*x[0]**o+(-0.948951787109)*x[0]+(-0.564104026105)*x[1]**o+(0.177457883377)*x[1]+(0.578169345377)*x[2]**o+(0.35574407049)*x[2]
        arg[1,1,0,1]=(-0.740305651819)*x[0]**o+(-0.340976006932)*x[0]+(-0.657257828033)*x[1]**o+(-0.456240473558)*x[1]+(0.526236681039)*x[2]**o+(0.804762801115)*x[2]
        arg[1,1,1,0]=(-0.753480536304)*x[0]**o+(0.424080516297)*x[0]+(-0.97529611006)*x[1]**o+(-0.511450162624)*x[1]+(-0.248982226385)*x[2]**o+(0.497803766242)*x[2]
        arg[1,1,1,1]=(-0.273280180629)*x[0]**o+(0.798945176097)*x[0]+(0.552152546131)*x[1]**o+(0.737594404744)*x[1]+(0.0339269287984)*x[2]**o+(-0.773791176515)*x[2]
        arg[1,1,2,0]=(0.553404075378)*x[0]**o+(-0.630969888585)*x[0]+(0.455072287913)*x[1]**o+(0.145928503781)*x[1]+(0.632480470504)*x[2]**o+(0.363815684296)*x[2]
        arg[1,1,2,1]=(-0.426756293656)*x[0]**o+(-0.449867410734)*x[0]+(-0.875891527959)*x[1]**o+(-0.125243159763)*x[1]+(0.887344501669)*x[2]**o+(-0.0569005112264)*x[2]
        arg[1,2,0,0]=(0.762313320036)*x[0]**o+(-0.514266376348)*x[0]+(-0.915248032301)*x[1]**o+(0.50418081892)*x[1]+(-0.867662577318)*x[2]**o+(0.731635970592)*x[2]
        arg[1,2,0,1]=(-0.0866106252969)*x[0]**o+(0.063250309716)*x[0]+(0.598723588976)*x[1]**o+(0.317235331686)*x[1]+(-0.412892601258)*x[2]**o+(0.771147670726)*x[2]
        arg[1,2,1,0]=(0.57639148624)*x[0]**o+(0.773916349481)*x[0]+(0.401319543026)*x[1]**o+(-0.510387398986)*x[1]+(-0.796773375436)*x[2]**o+(-0.765544756429)*x[2]
        arg[1,2,1,1]=(-0.686952109389)*x[0]**o+(0.428892331428)*x[0]+(-0.139971762047)*x[1]**o+(-0.747581454837)*x[1]+(0.152196734461)*x[2]**o+(-0.778424993847)*x[2]
        arg[1,2,2,0]=(0.729199771391)*x[0]**o+(0.185870258363)*x[0]+(0.996198489286)*x[1]**o+(0.0343703739018)*x[1]+(-0.6637833208)*x[2]**o+(-0.738510210409)*x[2]
        arg[1,2,2,1]=(0.705773326438)*x[0]**o+(-0.633406525448)*x[0]+(0.252620923822)*x[1]**o+(-0.760785834478)*x[1]+(0.995411456622)*x[2]**o+(-0.365718663741)*x[2]
        arg[1,3,0,0]=(0.826142335301)*x[0]**o+(-0.869335317945)*x[0]+(0.660360312658)*x[1]**o+(0.96157015332)*x[1]+(-0.268246324116)*x[2]**o+(-0.575288823295)*x[2]
        arg[1,3,0,1]=(0.620117943072)*x[0]**o+(0.51267609202)*x[0]+(-0.882468485772)*x[1]**o+(0.771021246475)*x[1]+(0.398548927211)*x[2]**o+(0.275983192446)*x[2]
        arg[1,3,1,0]=(0.716523839796)*x[0]**o+(-0.0868143348547)*x[0]+(-0.456384710114)*x[1]**o+(-0.646489063979)*x[1]+(0.594500214771)*x[2]**o+(-0.299956296467)*x[2]
        arg[1,3,1,1]=(-0.453257208906)*x[0]**o+(0.583260425119)*x[0]+(0.263463648634)*x[1]**o+(0.493693446665)*x[1]+(-0.529884773858)*x[2]**o+(0.220733486264)*x[2]
        arg[1,3,2,0]=(0.669052014654)*x[0]**o+(-0.247902780475)*x[0]+(0.44071786167)*x[1]**o+(-0.21199653848)*x[1]+(-0.0713603448391)*x[2]**o+(-0.66681439981)*x[2]
        arg[1,3,2,1]=(-0.341533813758)*x[0]**o+(0.770581995272)*x[0]+(0.03376913861)*x[1]**o+(0.000270719073091)*x[1]+(-0.774936686899)*x[2]**o+(-0.0196300167016)*x[2]
        arg[1,4,0,0]=(-0.869695667108)*x[0]**o+(0.515676789402)*x[0]+(0.67731822534)*x[1]**o+(0.126887846346)*x[1]+(0.843200966455)*x[2]**o+(-0.59930363301)*x[2]
        arg[1,4,0,1]=(0.909369352728)*x[0]**o+(-0.110828876636)*x[0]+(-0.65561953478)*x[1]**o+(0.598491879954)*x[1]+(-0.557567770163)*x[2]**o+(-0.391108926502)*x[2]
        arg[1,4,1,0]=(0.320646291934)*x[0]**o+(0.164062007697)*x[0]+(-0.546323998218)*x[1]**o+(0.72651711346)*x[1]+(-0.613496738898)*x[2]**o+(0.430761082137)*x[2]
        arg[1,4,1,1]=(-0.721651561555)*x[0]**o+(-0.0647310624509)*x[0]+(-0.360525345485)*x[1]**o+(0.408239277468)*x[1]+(0.600551705094)*x[2]**o+(0.506237401392)*x[2]
        arg[1,4,2,0]=(-0.391702965605)*x[0]**o+(-0.574470582913)*x[0]+(-0.729724609696)*x[1]**o+(0.91381180302)*x[1]+(-0.544328104929)*x[2]**o+(-0.516992534607)*x[2]
        arg[1,4,2,1]=(-0.0409666962711)*x[0]**o+(-0.524913112823)*x[0]+(-0.677030607856)*x[1]**o+(-0.530859496807)*x[1]+(0.137890059737)*x[2]**o+(0.0669856870322)*x[2]
        arg[2,0,0,0]=(0.10369318023)*x[0]**o+(0.345691611083)*x[0]+(-0.222795950561)*x[1]**o+(-0.797690270935)*x[1]+(-0.703457203234)*x[2]**o+(-0.751193209871)*x[2]
        arg[2,0,0,1]=(0.166457278787)*x[0]**o+(-0.0304766240065)*x[0]+(0.159987374581)*x[1]**o+(0.828012214748)*x[1]+(0.0670696566332)*x[2]**o+(-0.498252024306)*x[2]
        arg[2,0,1,0]=(0.901095992383)*x[0]**o+(0.400296735981)*x[0]+(0.394687295672)*x[1]**o+(-0.986496611224)*x[1]+(-0.768945602918)*x[2]**o+(-0.782762026735)*x[2]
        arg[2,0,1,1]=(0.47969549703)*x[0]**o+(0.981251232699)*x[0]+(-0.69807700503)*x[1]**o+(-0.813549271424)*x[1]+(-0.135970806025)*x[2]**o+(0.288738861745)*x[2]
        arg[2,0,2,0]=(0.315262113284)*x[0]**o+(0.0380265323361)*x[0]+(0.259894600802)*x[1]**o+(0.538712801332)*x[1]+(-0.385783597257)*x[2]**o+(0.666670257561)*x[2]
        arg[2,0,2,1]=(-0.756715539091)*x[0]**o+(-0.823809574081)*x[0]+(-0.283842954238)*x[1]**o+(-0.640166311159)*x[1]+(-0.39947011053)*x[2]**o+(-0.393543798757)*x[2]
        arg[2,1,0,0]=(-0.292033523499)*x[0]**o+(-0.940246412652)*x[0]+(-0.327571437608)*x[1]**o+(0.837077706655)*x[1]+(0.446471499818)*x[2]**o+(-0.747249149016)*x[2]
        arg[2,1,0,1]=(-0.867504469067)*x[0]**o+(0.635684140727)*x[0]+(0.488043128661)*x[1]**o+(-0.163389002402)*x[1]+(0.0440527250714)*x[2]**o+(0.605766722432)*x[2]
        arg[2,1,1,0]=(-0.0622304802635)*x[0]**o+(0.607299318698)*x[0]+(0.974933906045)*x[1]**o+(0.0632442428777)*x[1]+(0.0672808299851)*x[2]**o+(-0.555700803374)*x[2]
        arg[2,1,1,1]=(-0.350438453721)*x[0]**o+(-0.826289294567)*x[0]+(-0.473764081208)*x[1]**o+(-0.247804269494)*x[1]+(0.834301955859)*x[2]**o+(0.443058363661)*x[2]
        arg[2,1,2,0]=(0.5046026606)*x[0]**o+(0.305502438052)*x[0]+(-0.201816724695)*x[1]**o+(-0.101544576856)*x[1]+(0.729817329051)*x[2]**o+(-0.851479834334)*x[2]
        arg[2,1,2,1]=(0.917039424926)*x[0]**o+(0.260565966637)*x[0]+(-0.595846674057)*x[1]**o+(0.654773238108)*x[1]+(0.544664592565)*x[2]**o+(0.363343885358)*x[2]
        arg[2,2,0,0]=(0.243879200917)*x[0]**o+(0.335612613868)*x[0]+(-0.287066716076)*x[1]**o+(0.506206687739)*x[1]+(-0.902513036448)*x[2]**o+(-0.427017746025)*x[2]
        arg[2,2,0,1]=(0.965102871162)*x[0]**o+(0.161476912964)*x[0]+(0.97007160306)*x[1]**o+(0.870104939402)*x[1]+(0.552404991536)*x[2]**o+(-0.319082395552)*x[2]
        arg[2,2,1,0]=(0.17611459545)*x[0]**o+(-0.457419791602)*x[0]+(-0.00217752259743)*x[1]**o+(-0.632177103931)*x[1]+(0.521123797224)*x[2]**o+(0.427935307605)*x[2]
        arg[2,2,1,1]=(-0.993298850316)*x[0]**o+(0.277763727112)*x[0]+(0.849464768611)*x[1]**o+(0.890919639607)*x[1]+(0.01030894969)*x[2]**o+(-0.76051772768)*x[2]
        arg[2,2,2,0]=(-0.622887788117)*x[0]**o+(0.40187802561)*x[0]+(0.836685367934)*x[1]**o+(0.700379534201)*x[1]+(0.683627950354)*x[2]**o+(0.445351804296)*x[2]
        arg[2,2,2,1]=(0.124965536288)*x[0]**o+(-0.198462168996)*x[0]+(-0.657084126986)*x[1]**o+(-0.420469996385)*x[1]+(0.237164091597)*x[2]**o+(0.161926160295)*x[2]
        arg[2,3,0,0]=(-0.622682941897)*x[0]**o+(-0.6279729398)*x[0]+(0.189086870792)*x[1]**o+(-0.350822257394)*x[1]+(-0.505481371359)*x[2]**o+(0.812944216714)*x[2]
        arg[2,3,0,1]=(-0.246704719057)*x[0]**o+(0.513716384141)*x[0]+(-0.264914192193)*x[1]**o+(-0.32984984663)*x[1]+(-0.80409813849)*x[2]**o+(-0.825568736974)*x[2]
        arg[2,3,1,0]=(-0.534782640399)*x[0]**o+(-0.917823259622)*x[0]+(0.691827074405)*x[1]**o+(-0.818629872154)*x[1]+(0.870046992385)*x[2]**o+(0.00423218787803)*x[2]
        arg[2,3,1,1]=(0.633068676613)*x[0]**o+(-0.895613811402)*x[0]+(0.62242069294)*x[1]**o+(-0.272136050955)*x[1]+(-0.199008776357)*x[2]**o+(0.836119124508)*x[2]
        arg[2,3,2,0]=(0.335884106464)*x[0]**o+(-0.857917873801)*x[0]+(-0.42653278071)*x[1]**o+(0.375200781448)*x[1]+(0.378427453722)*x[2]**o+(0.398603623445)*x[2]
        arg[2,3,2,1]=(0.0602499604703)*x[0]**o+(-0.921459358089)*x[0]+(0.558180675902)*x[1]**o+(0.0863734870743)*x[1]+(0.215167095483)*x[2]**o+(-0.37908830738)*x[2]
        arg[2,4,0,0]=(0.866321987007)*x[0]**o+(0.81950553744)*x[0]+(0.0422552872889)*x[1]**o+(0.369328458453)*x[1]+(0.0547909452885)*x[2]**o+(-0.396367133883)*x[2]
        arg[2,4,0,1]=(0.438824337571)*x[0]**o+(0.211435787811)*x[0]+(-0.801273340861)*x[1]**o+(-0.181654271172)*x[1]+(0.374171148243)*x[2]**o+(0.945633108881)*x[2]
        arg[2,4,1,0]=(0.588044951473)*x[0]**o+(0.401384824924)*x[0]+(-0.191996429616)*x[1]**o+(0.0472579697166)*x[1]+(0.470962958517)*x[2]**o+(0.52718875619)*x[2]
        arg[2,4,1,1]=(0.371613549998)*x[0]**o+(-0.414998926249)*x[0]+(-0.762013182749)*x[1]**o+(0.796144230249)*x[1]+(-0.555041053148)*x[2]**o+(-0.595298275916)*x[2]
        arg[2,4,2,0]=(-0.992527136862)*x[0]**o+(-0.995041267483)*x[0]+(-0.730305990844)*x[1]**o+(-0.52470798566)*x[1]+(-0.938945944067)*x[2]**o+(0.33396135548)*x[2]
        arg[2,4,2,1]=(0.0333921786723)*x[0]**o+(0.707990374515)*x[0]+(-0.106667413755)*x[1]**o+(0.282989655082)*x[1]+(-0.152535982788)*x[2]**o+(-0.75130096727)*x[2]
        arg[3,0,0,0]=(-0.822675771493)*x[0]**o+(0.726468488157)*x[0]+(0.972567560814)*x[1]**o+(0.308266449093)*x[1]+(0.71069864654)*x[2]**o+(-0.683195491271)*x[2]
        arg[3,0,0,1]=(-0.505155072199)*x[0]**o+(0.339013729552)*x[0]+(-0.81565931012)*x[1]**o+(0.362458680494)*x[1]+(0.960221320064)*x[2]**o+(-0.17234580875)*x[2]
        arg[3,0,1,0]=(0.67972657926)*x[0]**o+(0.143085543693)*x[0]+(-0.699516563264)*x[1]**o+(-0.892226133751)*x[1]+(-0.898541541049)*x[2]**o+(-0.736249681714)*x[2]
        arg[3,0,1,1]=(-0.144423319215)*x[0]**o+(-0.509713525676)*x[0]+(0.347584762861)*x[1]**o+(0.666169972794)*x[1]+(0.691734080163)*x[2]**o+(0.9338062148)*x[2]
        arg[3,0,2,0]=(-0.47323843589)*x[0]**o+(0.0819655059363)*x[0]+(0.0863030211133)*x[1]**o+(0.127803089307)*x[1]+(-0.155852489756)*x[2]**o+(-0.00470398386987)*x[2]
        arg[3,0,2,1]=(0.372452726232)*x[0]**o+(0.133808609857)*x[0]+(0.962945642209)*x[1]**o+(0.0446609213929)*x[1]+(-0.446207951293)*x[2]**o+(0.0400733439868)*x[2]
        arg[3,1,0,0]=(-0.668575734235)*x[0]**o+(0.905913572459)*x[0]+(-0.321327880802)*x[1]**o+(0.0263209456188)*x[1]+(-0.0355325780368)*x[2]**o+(-0.624130395105)*x[2]
        arg[3,1,0,1]=(0.859810828483)*x[0]**o+(-0.235372125226)*x[0]+(-0.16715244612)*x[1]**o+(-0.793399872713)*x[1]+(0.303973795556)*x[2]**o+(0.176740655638)*x[2]
        arg[3,1,1,0]=(0.736448874258)*x[0]**o+(-0.448956585466)*x[0]+(0.782095902992)*x[1]**o+(0.796524196866)*x[1]+(-0.732812289541)*x[2]**o+(0.687089667291)*x[2]
        arg[3,1,1,1]=(-0.797708068194)*x[0]**o+(-0.691308116296)*x[0]+(-0.644065410151)*x[1]**o+(-0.929624672017)*x[1]+(0.265040940279)*x[2]**o+(0.424929324889)*x[2]
        arg[3,1,2,0]=(0.555032482777)*x[0]**o+(-0.564617314931)*x[0]+(-0.612721853869)*x[1]**o+(-0.048466699593)*x[1]+(0.753265037897)*x[2]**o+(0.913086239874)*x[2]
        arg[3,1,2,1]=(0.580881807771)*x[0]**o+(0.970973128604)*x[0]+(-0.603610563753)*x[1]**o+(0.979985814438)*x[1]+(-0.639531987694)*x[2]**o+(-0.773094138073)*x[2]
        arg[3,2,0,0]=(0.574306840544)*x[0]**o+(0.449094769988)*x[0]+(-0.406303076614)*x[1]**o+(-0.776648781906)*x[1]+(-0.59829476647)*x[2]**o+(0.257632446666)*x[2]
        arg[3,2,0,1]=(0.689106127051)*x[0]**o+(0.717184851708)*x[0]+(-0.495581943615)*x[1]**o+(-0.30329139788)*x[1]+(0.901753128535)*x[2]**o+(0.48496505727)*x[2]
        arg[3,2,1,0]=(0.252050902256)*x[0]**o+(-0.0414281706943)*x[0]+(-0.284125627654)*x[1]**o+(0.477450554269)*x[1]+(0.849202242148)*x[2]**o+(0.211826565889)*x[2]
        arg[3,2,1,1]=(-0.319900673571)*x[0]**o+(0.965993723739)*x[0]+(0.671360092314)*x[1]**o+(-0.652805407934)*x[1]+(-0.0352316646853)*x[2]**o+(-0.442493499306)*x[2]
        arg[3,2,2,0]=(0.525577660439)*x[0]**o+(0.660882378915)*x[0]+(-0.419826738105)*x[1]**o+(-0.798251889077)*x[1]+(-0.430382393461)*x[2]**o+(0.714357242496)*x[2]
        arg[3,2,2,1]=(-0.200634944034)*x[0]**o+(0.211086366034)*x[0]+(0.116289218188)*x[1]**o+(-0.785792860501)*x[1]+(0.763320327438)*x[2]**o+(-0.457355826175)*x[2]
        arg[3,3,0,0]=(-0.890873783782)*x[0]**o+(0.322610697657)*x[0]+(-0.351437390332)*x[1]**o+(-0.720698169708)*x[1]+(-0.664444440194)*x[2]**o+(0.274865394829)*x[2]
        arg[3,3,0,1]=(-0.384885652342)*x[0]**o+(-0.409799972422)*x[0]+(0.679315041587)*x[1]**o+(0.0859788488759)*x[1]+(-0.136173937432)*x[2]**o+(-0.136591429212)*x[2]
        arg[3,3,1,0]=(0.147934741321)*x[0]**o+(0.246941956666)*x[0]+(0.696831123708)*x[1]**o+(-0.905966322958)*x[1]+(0.672705695255)*x[2]**o+(-0.776017262395)*x[2]
        arg[3,3,1,1]=(0.933982654679)*x[0]**o+(-0.367246348547)*x[0]+(-0.254076834401)*x[1]**o+(0.148617981234)*x[1]+(-0.486881004358)*x[2]**o+(0.636371570082)*x[2]
        arg[3,3,2,0]=(-0.536125392071)*x[0]**o+(-0.663380779479)*x[0]+(0.0655694689011)*x[1]**o+(0.0390002326894)*x[1]+(-0.731730129742)*x[2]**o+(-0.578791485067)*x[2]
        arg[3,3,2,1]=(0.705873902409)*x[0]**o+(-0.0444303506955)*x[0]+(-0.978519191225)*x[1]**o+(-0.0280878186958)*x[1]+(0.182378726121)*x[2]**o+(-0.555345919696)*x[2]
        arg[3,4,0,0]=(0.945945862395)*x[0]**o+(-0.448487950632)*x[0]+(-0.477570573229)*x[1]**o+(0.313871838624)*x[1]+(0.237944220287)*x[2]**o+(-0.218095130482)*x[2]
        arg[3,4,0,1]=(-0.758647277269)*x[0]**o+(-0.550604667208)*x[0]+(-0.262944399053)*x[1]**o+(0.994302652859)*x[1]+(0.363440310008)*x[2]**o+(-0.127863970613)*x[2]
        arg[3,4,1,0]=(0.667202433965)*x[0]**o+(0.954901068821)*x[0]+(0.719745510318)*x[1]**o+(-0.763882742007)*x[1]+(0.880927476123)*x[2]**o+(0.102034630893)*x[2]
        arg[3,4,1,1]=(-0.684644024456)*x[0]**o+(0.88037275093)*x[0]+(-0.772213766289)*x[1]**o+(-0.702615214094)*x[1]+(-0.25661708053)*x[2]**o+(0.126779040694)*x[2]
        arg[3,4,2,0]=(-0.589611385303)*x[0]**o+(0.638712635252)*x[0]+(0.745450741856)*x[1]**o+(0.399494092815)*x[1]+(-0.717293809038)*x[2]**o+(0.305662737751)*x[2]
        arg[3,4,2,1]=(0.0572426997796)*x[0]**o+(-0.764600144578)*x[0]+(0.482545219565)*x[1]**o+(0.525883069411)*x[1]+(-0.612224327195)*x[2]**o+(0.354344820256)*x[2]
        ref[0,0,0,0]=(0.952750275738)*x_ref[0]**o+(0.971489204087)*x_ref[0]+(-0.320706487584)*x_ref[1]**o+(-0.853260471578)*x_ref[1]+(-0.696434862068)*x_ref[2]**o+(0.641824397433)*x_ref[2]
        ref[0,0,0,1]=(0.115484192786)*x_ref[0]**o+(0.147813687309)*x_ref[0]+(-0.998010548186)*x_ref[1]**o+(0.0274712088882)*x_ref[1]+(0.136937971489)*x_ref[2]**o+(-0.657986135444)*x_ref[2]
        ref[0,0,1,0]=(0.805857504779)*x_ref[0]**o+(0.196278916938)*x_ref[0]+(-0.969340785263)*x_ref[1]**o+(-0.543890174888)*x_ref[1]+(0.49118423357)*x_ref[2]**o+(-0.883493800593)*x_ref[2]
        ref[0,0,1,1]=(-0.808802136474)*x_ref[0]**o+(-0.866039395649)*x_ref[0]+(0.0339207025941)*x_ref[1]**o+(-0.441368773534)*x_ref[1]+(0.673688211653)*x_ref[2]**o+(0.917894389398)*x_ref[2]
        ref[0,0,2,0]=(0.0459902210078)*x_ref[0]**o+(0.862332666528)*x_ref[0]+(-0.547792306389)*x_ref[1]**o+(0.722176431732)*x_ref[1]+(-0.511278989032)*x_ref[2]**o+(-0.692775336577)*x_ref[2]
        ref[0,0,2,1]=(-0.641904406513)*x_ref[0]**o+(-0.854450656063)*x_ref[0]+(0.81138894849)*x_ref[1]**o+(0.328886625074)*x_ref[1]+(0.400935831363)*x_ref[2]**o+(-0.25184059766)*x_ref[2]
        ref[0,1,0,0]=(0.69922840942)*x_ref[0]**o+(-0.798085508443)*x_ref[0]+(-0.85191383431)*x_ref[1]**o+(-0.162940829331)*x_ref[1]+(0.504020642931)*x_ref[2]**o+(0.0950957436173)*x_ref[2]
        ref[0,1,0,1]=(0.28901429037)*x_ref[0]**o+(-0.624508745729)*x_ref[0]+(0.921288077478)*x_ref[1]**o+(-0.769207901119)*x_ref[1]+(0.0852374599763)*x_ref[2]**o+(-0.402583033986)*x_ref[2]
        ref[0,1,1,0]=(0.715817282215)*x_ref[0]**o+(0.68735256665)*x_ref[0]+(0.610212058649)*x_ref[1]**o+(0.53560824609)*x_ref[1]+(-0.238908897264)*x_ref[2]**o+(0.619799158474)*x_ref[2]
        ref[0,1,1,1]=(-0.371797414191)*x_ref[0]**o+(0.106263350514)*x_ref[0]+(0.583768355362)*x_ref[1]**o+(0.818797667369)*x_ref[1]+(0.0567571152493)*x_ref[2]**o+(-0.242244867169)*x_ref[2]
        ref[0,1,2,0]=(0.190480659729)*x_ref[0]**o+(-0.632415604771)*x_ref[0]+(0.252513826138)*x_ref[1]**o+(0.874732840469)*x_ref[1]+(-0.367769326078)*x_ref[2]**o+(0.364453292404)*x_ref[2]
        ref[0,1,2,1]=(-0.704507136581)*x_ref[0]**o+(-0.614211737547)*x_ref[0]+(0.533383507313)*x_ref[1]**o+(-0.940007583478)*x_ref[1]+(0.558260002746)*x_ref[2]**o+(0.742565692751)*x_ref[2]
        ref[0,2,0,0]=(0.132581156766)*x_ref[0]**o+(0.671793183183)*x_ref[0]+(-0.308255110769)*x_ref[1]**o+(-0.956002957061)*x_ref[1]+(0.300091541301)*x_ref[2]**o+(-0.714260746791)*x_ref[2]
        ref[0,2,0,1]=(-0.836648080859)*x_ref[0]**o+(0.123813861017)*x_ref[0]+(0.866569934894)*x_ref[1]**o+(-0.957824572351)*x_ref[1]+(-0.904478410952)*x_ref[2]**o+(0.691673758521)*x_ref[2]
        ref[0,2,1,0]=(0.411624721878)*x_ref[0]**o+(0.594054722352)*x_ref[0]+(0.150177701135)*x_ref[1]**o+(0.945855261084)*x_ref[1]+(0.295930158817)*x_ref[2]**o+(-0.607980954424)*x_ref[2]
        ref[0,2,1,1]=(-0.0777741308373)*x_ref[0]**o+(0.697501295655)*x_ref[0]+(0.981797797329)*x_ref[1]**o+(-0.671827077949)*x_ref[1]+(0.292310146229)*x_ref[2]**o+(0.484652357857)*x_ref[2]
        ref[0,2,2,0]=(0.194910469807)*x_ref[0]**o+(-0.70008515623)*x_ref[0]+(0.192808336639)*x_ref[1]**o+(-0.962764477371)*x_ref[1]+(-0.34292240753)*x_ref[2]**o+(0.525470829202)*x_ref[2]
        ref[0,2,2,1]=(0.850368420228)*x_ref[0]**o+(-0.516355386368)*x_ref[0]+(0.395764367873)*x_ref[1]**o+(0.320830752607)*x_ref[1]+(-0.26288679044)*x_ref[2]**o+(0.174457723191)*x_ref[2]
        ref[0,3,0,0]=(-0.713987556845)*x_ref[0]**o+(0.0391066248253)*x_ref[0]+(-0.63911733479)*x_ref[1]**o+(-0.119953032121)*x_ref[1]+(0.697879836447)*x_ref[2]**o+(-0.474957057702)*x_ref[2]
        ref[0,3,0,1]=(-0.128614897423)*x_ref[0]**o+(0.7166036403)*x_ref[0]+(0.678415978447)*x_ref[1]**o+(-0.191453169618)*x_ref[1]+(-0.755261361956)*x_ref[2]**o+(-0.839956004196)*x_ref[2]
        ref[0,3,1,0]=(0.258158740934)*x_ref[0]**o+(-0.412199346967)*x_ref[0]+(-0.781106341272)*x_ref[1]**o+(-0.561162823101)*x_ref[1]+(-0.59549864377)*x_ref[2]**o+(-0.748038415721)*x_ref[2]
        ref[0,3,1,1]=(0.68950222305)*x_ref[0]**o+(-0.904027774445)*x_ref[0]+(0.769744453441)*x_ref[1]**o+(-0.265549152752)*x_ref[1]+(-0.375978837131)*x_ref[2]**o+(-0.387010981761)*x_ref[2]
        ref[0,3,2,0]=(0.46852681319)*x_ref[0]**o+(-0.887650258615)*x_ref[0]+(-0.780275624319)*x_ref[1]**o+(0.527897372599)*x_ref[1]+(0.889281778411)*x_ref[2]**o+(0.82538563877)*x_ref[2]
        ref[0,3,2,1]=(-0.817514448184)*x_ref[0]**o+(-0.230190768519)*x_ref[0]+(-0.101550849344)*x_ref[1]**o+(0.1654885857)*x_ref[1]+(-0.231317749498)*x_ref[2]**o+(0.580153306395)*x_ref[2]
        ref[0,4,0,0]=(0.394993007242)*x_ref[0]**o+(0.305948764395)*x_ref[0]+(-0.66853441368)*x_ref[1]**o+(0.279027498094)*x_ref[1]+(0.161366934709)*x_ref[2]**o+(-0.742002471552)*x_ref[2]
        ref[0,4,0,1]=(-0.661214709471)*x_ref[0]**o+(0.548880317935)*x_ref[0]+(0.472682748503)*x_ref[1]**o+(0.266797505699)*x_ref[1]+(0.405618413683)*x_ref[2]**o+(-0.0539677542667)*x_ref[2]
        ref[0,4,1,0]=(-0.713892424429)*x_ref[0]**o+(0.568075831592)*x_ref[0]+(-0.956180895599)*x_ref[1]**o+(0.63618957695)*x_ref[1]+(0.983696160047)*x_ref[2]**o+(-0.308157929002)*x_ref[2]
        ref[0,4,1,1]=(-0.666582919353)*x_ref[0]**o+(0.0816884254484)*x_ref[0]+(0.190810048896)*x_ref[1]**o+(0.482322905132)*x_ref[1]+(-0.576597070954)*x_ref[2]**o+(-0.722381611553)*x_ref[2]
        ref[0,4,2,0]=(-0.0229650731041)*x_ref[0]**o+(-0.0867816754979)*x_ref[0]+(0.0905042102365)*x_ref[1]**o+(-0.994963424765)*x_ref[1]+(0.74504757815)*x_ref[2]**o+(0.71462549644)*x_ref[2]
        ref[0,4,2,1]=(-0.0464278496147)*x_ref[0]**o+(0.122501026996)*x_ref[0]+(-0.049450141901)*x_ref[1]**o+(-0.388739464086)*x_ref[1]+(-0.151389885067)*x_ref[2]**o+(-0.0316948088803)*x_ref[2]
        ref[1,0,0,0]=(-0.303653386851)*x_ref[0]**o+(-0.435866753659)*x_ref[0]+(-0.419078175082)*x_ref[1]**o+(-0.907595244917)*x_ref[1]+(0.442156528837)*x_ref[2]**o+(-0.76481785523)*x_ref[2]
        ref[1,0,0,1]=(0.879454551618)*x_ref[0]**o+(-0.530789858588)*x_ref[0]+(0.942203776542)*x_ref[1]**o+(0.864362935436)*x_ref[1]+(-0.252832164588)*x_ref[2]**o+(0.00758252581015)*x_ref[2]
        ref[1,0,1,0]=(0.522433142898)*x_ref[0]**o+(0.564002257569)*x_ref[0]+(0.149811689377)*x_ref[1]**o+(-0.757726948165)*x_ref[1]+(-0.767052869719)*x_ref[2]**o+(0.743541806001)*x_ref[2]
        ref[1,0,1,1]=(0.535612198186)*x_ref[0]**o+(0.0874480088401)*x_ref[0]+(0.913151292296)*x_ref[1]**o+(-0.495024655308)*x_ref[1]+(-0.42308905182)*x_ref[2]**o+(-0.983623551153)*x_ref[2]
        ref[1,0,2,0]=(0.587221456352)*x_ref[0]**o+(0.0384438535009)*x_ref[0]+(0.301554087927)*x_ref[1]**o+(-0.944074061806)*x_ref[1]+(0.808195432991)*x_ref[2]**o+(-0.46545743768)*x_ref[2]
        ref[1,0,2,1]=(0.136251169669)*x_ref[0]**o+(0.772190271875)*x_ref[0]+(-0.490981224022)*x_ref[1]**o+(-0.191915588448)*x_ref[1]+(0.792617864405)*x_ref[2]**o+(0.139425359773)*x_ref[2]
        ref[1,1,0,0]=(0.0440493146329)*x_ref[0]**o+(-0.948951787109)*x_ref[0]+(-0.564104026105)*x_ref[1]**o+(0.177457883377)*x_ref[1]+(0.578169345377)*x_ref[2]**o+(0.35574407049)*x_ref[2]
        ref[1,1,0,1]=(-0.740305651819)*x_ref[0]**o+(-0.340976006932)*x_ref[0]+(-0.657257828033)*x_ref[1]**o+(-0.456240473558)*x_ref[1]+(0.526236681039)*x_ref[2]**o+(0.804762801115)*x_ref[2]
        ref[1,1,1,0]=(-0.753480536304)*x_ref[0]**o+(0.424080516297)*x_ref[0]+(-0.97529611006)*x_ref[1]**o+(-0.511450162624)*x_ref[1]+(-0.248982226385)*x_ref[2]**o+(0.497803766242)*x_ref[2]
        ref[1,1,1,1]=(-0.273280180629)*x_ref[0]**o+(0.798945176097)*x_ref[0]+(0.552152546131)*x_ref[1]**o+(0.737594404744)*x_ref[1]+(0.0339269287984)*x_ref[2]**o+(-0.773791176515)*x_ref[2]
        ref[1,1,2,0]=(0.553404075378)*x_ref[0]**o+(-0.630969888585)*x_ref[0]+(0.455072287913)*x_ref[1]**o+(0.145928503781)*x_ref[1]+(0.632480470504)*x_ref[2]**o+(0.363815684296)*x_ref[2]
        ref[1,1,2,1]=(-0.426756293656)*x_ref[0]**o+(-0.449867410734)*x_ref[0]+(-0.875891527959)*x_ref[1]**o+(-0.125243159763)*x_ref[1]+(0.887344501669)*x_ref[2]**o+(-0.0569005112264)*x_ref[2]
        ref[1,2,0,0]=(0.762313320036)*x_ref[0]**o+(-0.514266376348)*x_ref[0]+(-0.915248032301)*x_ref[1]**o+(0.50418081892)*x_ref[1]+(-0.867662577318)*x_ref[2]**o+(0.731635970592)*x_ref[2]
        ref[1,2,0,1]=(-0.0866106252969)*x_ref[0]**o+(0.063250309716)*x_ref[0]+(0.598723588976)*x_ref[1]**o+(0.317235331686)*x_ref[1]+(-0.412892601258)*x_ref[2]**o+(0.771147670726)*x_ref[2]
        ref[1,2,1,0]=(0.57639148624)*x_ref[0]**o+(0.773916349481)*x_ref[0]+(0.401319543026)*x_ref[1]**o+(-0.510387398986)*x_ref[1]+(-0.796773375436)*x_ref[2]**o+(-0.765544756429)*x_ref[2]
        ref[1,2,1,1]=(-0.686952109389)*x_ref[0]**o+(0.428892331428)*x_ref[0]+(-0.139971762047)*x_ref[1]**o+(-0.747581454837)*x_ref[1]+(0.152196734461)*x_ref[2]**o+(-0.778424993847)*x_ref[2]
        ref[1,2,2,0]=(0.729199771391)*x_ref[0]**o+(0.185870258363)*x_ref[0]+(0.996198489286)*x_ref[1]**o+(0.0343703739018)*x_ref[1]+(-0.6637833208)*x_ref[2]**o+(-0.738510210409)*x_ref[2]
        ref[1,2,2,1]=(0.705773326438)*x_ref[0]**o+(-0.633406525448)*x_ref[0]+(0.252620923822)*x_ref[1]**o+(-0.760785834478)*x_ref[1]+(0.995411456622)*x_ref[2]**o+(-0.365718663741)*x_ref[2]
        ref[1,3,0,0]=(0.826142335301)*x_ref[0]**o+(-0.869335317945)*x_ref[0]+(0.660360312658)*x_ref[1]**o+(0.96157015332)*x_ref[1]+(-0.268246324116)*x_ref[2]**o+(-0.575288823295)*x_ref[2]
        ref[1,3,0,1]=(0.620117943072)*x_ref[0]**o+(0.51267609202)*x_ref[0]+(-0.882468485772)*x_ref[1]**o+(0.771021246475)*x_ref[1]+(0.398548927211)*x_ref[2]**o+(0.275983192446)*x_ref[2]
        ref[1,3,1,0]=(0.716523839796)*x_ref[0]**o+(-0.0868143348547)*x_ref[0]+(-0.456384710114)*x_ref[1]**o+(-0.646489063979)*x_ref[1]+(0.594500214771)*x_ref[2]**o+(-0.299956296467)*x_ref[2]
        ref[1,3,1,1]=(-0.453257208906)*x_ref[0]**o+(0.583260425119)*x_ref[0]+(0.263463648634)*x_ref[1]**o+(0.493693446665)*x_ref[1]+(-0.529884773858)*x_ref[2]**o+(0.220733486264)*x_ref[2]
        ref[1,3,2,0]=(0.669052014654)*x_ref[0]**o+(-0.247902780475)*x_ref[0]+(0.44071786167)*x_ref[1]**o+(-0.21199653848)*x_ref[1]+(-0.0713603448391)*x_ref[2]**o+(-0.66681439981)*x_ref[2]
        ref[1,3,2,1]=(-0.341533813758)*x_ref[0]**o+(0.770581995272)*x_ref[0]+(0.03376913861)*x_ref[1]**o+(0.000270719073091)*x_ref[1]+(-0.774936686899)*x_ref[2]**o+(-0.0196300167016)*x_ref[2]
        ref[1,4,0,0]=(-0.869695667108)*x_ref[0]**o+(0.515676789402)*x_ref[0]+(0.67731822534)*x_ref[1]**o+(0.126887846346)*x_ref[1]+(0.843200966455)*x_ref[2]**o+(-0.59930363301)*x_ref[2]
        ref[1,4,0,1]=(0.909369352728)*x_ref[0]**o+(-0.110828876636)*x_ref[0]+(-0.65561953478)*x_ref[1]**o+(0.598491879954)*x_ref[1]+(-0.557567770163)*x_ref[2]**o+(-0.391108926502)*x_ref[2]
        ref[1,4,1,0]=(0.320646291934)*x_ref[0]**o+(0.164062007697)*x_ref[0]+(-0.546323998218)*x_ref[1]**o+(0.72651711346)*x_ref[1]+(-0.613496738898)*x_ref[2]**o+(0.430761082137)*x_ref[2]
        ref[1,4,1,1]=(-0.721651561555)*x_ref[0]**o+(-0.0647310624509)*x_ref[0]+(-0.360525345485)*x_ref[1]**o+(0.408239277468)*x_ref[1]+(0.600551705094)*x_ref[2]**o+(0.506237401392)*x_ref[2]
        ref[1,4,2,0]=(-0.391702965605)*x_ref[0]**o+(-0.574470582913)*x_ref[0]+(-0.729724609696)*x_ref[1]**o+(0.91381180302)*x_ref[1]+(-0.544328104929)*x_ref[2]**o+(-0.516992534607)*x_ref[2]
        ref[1,4,2,1]=(-0.0409666962711)*x_ref[0]**o+(-0.524913112823)*x_ref[0]+(-0.677030607856)*x_ref[1]**o+(-0.530859496807)*x_ref[1]+(0.137890059737)*x_ref[2]**o+(0.0669856870322)*x_ref[2]
        ref[2,0,0,0]=(0.10369318023)*x_ref[0]**o+(0.345691611083)*x_ref[0]+(-0.222795950561)*x_ref[1]**o+(-0.797690270935)*x_ref[1]+(-0.703457203234)*x_ref[2]**o+(-0.751193209871)*x_ref[2]
        ref[2,0,0,1]=(0.166457278787)*x_ref[0]**o+(-0.0304766240065)*x_ref[0]+(0.159987374581)*x_ref[1]**o+(0.828012214748)*x_ref[1]+(0.0670696566332)*x_ref[2]**o+(-0.498252024306)*x_ref[2]
        ref[2,0,1,0]=(0.901095992383)*x_ref[0]**o+(0.400296735981)*x_ref[0]+(0.394687295672)*x_ref[1]**o+(-0.986496611224)*x_ref[1]+(-0.768945602918)*x_ref[2]**o+(-0.782762026735)*x_ref[2]
        ref[2,0,1,1]=(0.47969549703)*x_ref[0]**o+(0.981251232699)*x_ref[0]+(-0.69807700503)*x_ref[1]**o+(-0.813549271424)*x_ref[1]+(-0.135970806025)*x_ref[2]**o+(0.288738861745)*x_ref[2]
        ref[2,0,2,0]=(0.315262113284)*x_ref[0]**o+(0.0380265323361)*x_ref[0]+(0.259894600802)*x_ref[1]**o+(0.538712801332)*x_ref[1]+(-0.385783597257)*x_ref[2]**o+(0.666670257561)*x_ref[2]
        ref[2,0,2,1]=(-0.756715539091)*x_ref[0]**o+(-0.823809574081)*x_ref[0]+(-0.283842954238)*x_ref[1]**o+(-0.640166311159)*x_ref[1]+(-0.39947011053)*x_ref[2]**o+(-0.393543798757)*x_ref[2]
        ref[2,1,0,0]=(-0.292033523499)*x_ref[0]**o+(-0.940246412652)*x_ref[0]+(-0.327571437608)*x_ref[1]**o+(0.837077706655)*x_ref[1]+(0.446471499818)*x_ref[2]**o+(-0.747249149016)*x_ref[2]
        ref[2,1,0,1]=(-0.867504469067)*x_ref[0]**o+(0.635684140727)*x_ref[0]+(0.488043128661)*x_ref[1]**o+(-0.163389002402)*x_ref[1]+(0.0440527250714)*x_ref[2]**o+(0.605766722432)*x_ref[2]
        ref[2,1,1,0]=(-0.0622304802635)*x_ref[0]**o+(0.607299318698)*x_ref[0]+(0.974933906045)*x_ref[1]**o+(0.0632442428777)*x_ref[1]+(0.0672808299851)*x_ref[2]**o+(-0.555700803374)*x_ref[2]
        ref[2,1,1,1]=(-0.350438453721)*x_ref[0]**o+(-0.826289294567)*x_ref[0]+(-0.473764081208)*x_ref[1]**o+(-0.247804269494)*x_ref[1]+(0.834301955859)*x_ref[2]**o+(0.443058363661)*x_ref[2]
        ref[2,1,2,0]=(0.5046026606)*x_ref[0]**o+(0.305502438052)*x_ref[0]+(-0.201816724695)*x_ref[1]**o+(-0.101544576856)*x_ref[1]+(0.729817329051)*x_ref[2]**o+(-0.851479834334)*x_ref[2]
        ref[2,1,2,1]=(0.917039424926)*x_ref[0]**o+(0.260565966637)*x_ref[0]+(-0.595846674057)*x_ref[1]**o+(0.654773238108)*x_ref[1]+(0.544664592565)*x_ref[2]**o+(0.363343885358)*x_ref[2]
        ref[2,2,0,0]=(0.243879200917)*x_ref[0]**o+(0.335612613868)*x_ref[0]+(-0.287066716076)*x_ref[1]**o+(0.506206687739)*x_ref[1]+(-0.902513036448)*x_ref[2]**o+(-0.427017746025)*x_ref[2]
        ref[2,2,0,1]=(0.965102871162)*x_ref[0]**o+(0.161476912964)*x_ref[0]+(0.97007160306)*x_ref[1]**o+(0.870104939402)*x_ref[1]+(0.552404991536)*x_ref[2]**o+(-0.319082395552)*x_ref[2]
        ref[2,2,1,0]=(0.17611459545)*x_ref[0]**o+(-0.457419791602)*x_ref[0]+(-0.00217752259743)*x_ref[1]**o+(-0.632177103931)*x_ref[1]+(0.521123797224)*x_ref[2]**o+(0.427935307605)*x_ref[2]
        ref[2,2,1,1]=(-0.993298850316)*x_ref[0]**o+(0.277763727112)*x_ref[0]+(0.849464768611)*x_ref[1]**o+(0.890919639607)*x_ref[1]+(0.01030894969)*x_ref[2]**o+(-0.76051772768)*x_ref[2]
        ref[2,2,2,0]=(-0.622887788117)*x_ref[0]**o+(0.40187802561)*x_ref[0]+(0.836685367934)*x_ref[1]**o+(0.700379534201)*x_ref[1]+(0.683627950354)*x_ref[2]**o+(0.445351804296)*x_ref[2]
        ref[2,2,2,1]=(0.124965536288)*x_ref[0]**o+(-0.198462168996)*x_ref[0]+(-0.657084126986)*x_ref[1]**o+(-0.420469996385)*x_ref[1]+(0.237164091597)*x_ref[2]**o+(0.161926160295)*x_ref[2]
        ref[2,3,0,0]=(-0.622682941897)*x_ref[0]**o+(-0.6279729398)*x_ref[0]+(0.189086870792)*x_ref[1]**o+(-0.350822257394)*x_ref[1]+(-0.505481371359)*x_ref[2]**o+(0.812944216714)*x_ref[2]
        ref[2,3,0,1]=(-0.246704719057)*x_ref[0]**o+(0.513716384141)*x_ref[0]+(-0.264914192193)*x_ref[1]**o+(-0.32984984663)*x_ref[1]+(-0.80409813849)*x_ref[2]**o+(-0.825568736974)*x_ref[2]
        ref[2,3,1,0]=(-0.534782640399)*x_ref[0]**o+(-0.917823259622)*x_ref[0]+(0.691827074405)*x_ref[1]**o+(-0.818629872154)*x_ref[1]+(0.870046992385)*x_ref[2]**o+(0.00423218787803)*x_ref[2]
        ref[2,3,1,1]=(0.633068676613)*x_ref[0]**o+(-0.895613811402)*x_ref[0]+(0.62242069294)*x_ref[1]**o+(-0.272136050955)*x_ref[1]+(-0.199008776357)*x_ref[2]**o+(0.836119124508)*x_ref[2]
        ref[2,3,2,0]=(0.335884106464)*x_ref[0]**o+(-0.857917873801)*x_ref[0]+(-0.42653278071)*x_ref[1]**o+(0.375200781448)*x_ref[1]+(0.378427453722)*x_ref[2]**o+(0.398603623445)*x_ref[2]
        ref[2,3,2,1]=(0.0602499604703)*x_ref[0]**o+(-0.921459358089)*x_ref[0]+(0.558180675902)*x_ref[1]**o+(0.0863734870743)*x_ref[1]+(0.215167095483)*x_ref[2]**o+(-0.37908830738)*x_ref[2]
        ref[2,4,0,0]=(0.866321987007)*x_ref[0]**o+(0.81950553744)*x_ref[0]+(0.0422552872889)*x_ref[1]**o+(0.369328458453)*x_ref[1]+(0.0547909452885)*x_ref[2]**o+(-0.396367133883)*x_ref[2]
        ref[2,4,0,1]=(0.438824337571)*x_ref[0]**o+(0.211435787811)*x_ref[0]+(-0.801273340861)*x_ref[1]**o+(-0.181654271172)*x_ref[1]+(0.374171148243)*x_ref[2]**o+(0.945633108881)*x_ref[2]
        ref[2,4,1,0]=(0.588044951473)*x_ref[0]**o+(0.401384824924)*x_ref[0]+(-0.191996429616)*x_ref[1]**o+(0.0472579697166)*x_ref[1]+(0.470962958517)*x_ref[2]**o+(0.52718875619)*x_ref[2]
        ref[2,4,1,1]=(0.371613549998)*x_ref[0]**o+(-0.414998926249)*x_ref[0]+(-0.762013182749)*x_ref[1]**o+(0.796144230249)*x_ref[1]+(-0.555041053148)*x_ref[2]**o+(-0.595298275916)*x_ref[2]
        ref[2,4,2,0]=(-0.992527136862)*x_ref[0]**o+(-0.995041267483)*x_ref[0]+(-0.730305990844)*x_ref[1]**o+(-0.52470798566)*x_ref[1]+(-0.938945944067)*x_ref[2]**o+(0.33396135548)*x_ref[2]
        ref[2,4,2,1]=(0.0333921786723)*x_ref[0]**o+(0.707990374515)*x_ref[0]+(-0.106667413755)*x_ref[1]**o+(0.282989655082)*x_ref[1]+(-0.152535982788)*x_ref[2]**o+(-0.75130096727)*x_ref[2]
        ref[3,0,0,0]=(-0.822675771493)*x_ref[0]**o+(0.726468488157)*x_ref[0]+(0.972567560814)*x_ref[1]**o+(0.308266449093)*x_ref[1]+(0.71069864654)*x_ref[2]**o+(-0.683195491271)*x_ref[2]
        ref[3,0,0,1]=(-0.505155072199)*x_ref[0]**o+(0.339013729552)*x_ref[0]+(-0.81565931012)*x_ref[1]**o+(0.362458680494)*x_ref[1]+(0.960221320064)*x_ref[2]**o+(-0.17234580875)*x_ref[2]
        ref[3,0,1,0]=(0.67972657926)*x_ref[0]**o+(0.143085543693)*x_ref[0]+(-0.699516563264)*x_ref[1]**o+(-0.892226133751)*x_ref[1]+(-0.898541541049)*x_ref[2]**o+(-0.736249681714)*x_ref[2]
        ref[3,0,1,1]=(-0.144423319215)*x_ref[0]**o+(-0.509713525676)*x_ref[0]+(0.347584762861)*x_ref[1]**o+(0.666169972794)*x_ref[1]+(0.691734080163)*x_ref[2]**o+(0.9338062148)*x_ref[2]
        ref[3,0,2,0]=(-0.47323843589)*x_ref[0]**o+(0.0819655059363)*x_ref[0]+(0.0863030211133)*x_ref[1]**o+(0.127803089307)*x_ref[1]+(-0.155852489756)*x_ref[2]**o+(-0.00470398386987)*x_ref[2]
        ref[3,0,2,1]=(0.372452726232)*x_ref[0]**o+(0.133808609857)*x_ref[0]+(0.962945642209)*x_ref[1]**o+(0.0446609213929)*x_ref[1]+(-0.446207951293)*x_ref[2]**o+(0.0400733439868)*x_ref[2]
        ref[3,1,0,0]=(-0.668575734235)*x_ref[0]**o+(0.905913572459)*x_ref[0]+(-0.321327880802)*x_ref[1]**o+(0.0263209456188)*x_ref[1]+(-0.0355325780368)*x_ref[2]**o+(-0.624130395105)*x_ref[2]
        ref[3,1,0,1]=(0.859810828483)*x_ref[0]**o+(-0.235372125226)*x_ref[0]+(-0.16715244612)*x_ref[1]**o+(-0.793399872713)*x_ref[1]+(0.303973795556)*x_ref[2]**o+(0.176740655638)*x_ref[2]
        ref[3,1,1,0]=(0.736448874258)*x_ref[0]**o+(-0.448956585466)*x_ref[0]+(0.782095902992)*x_ref[1]**o+(0.796524196866)*x_ref[1]+(-0.732812289541)*x_ref[2]**o+(0.687089667291)*x_ref[2]
        ref[3,1,1,1]=(-0.797708068194)*x_ref[0]**o+(-0.691308116296)*x_ref[0]+(-0.644065410151)*x_ref[1]**o+(-0.929624672017)*x_ref[1]+(0.265040940279)*x_ref[2]**o+(0.424929324889)*x_ref[2]
        ref[3,1,2,0]=(0.555032482777)*x_ref[0]**o+(-0.564617314931)*x_ref[0]+(-0.612721853869)*x_ref[1]**o+(-0.048466699593)*x_ref[1]+(0.753265037897)*x_ref[2]**o+(0.913086239874)*x_ref[2]
        ref[3,1,2,1]=(0.580881807771)*x_ref[0]**o+(0.970973128604)*x_ref[0]+(-0.603610563753)*x_ref[1]**o+(0.979985814438)*x_ref[1]+(-0.639531987694)*x_ref[2]**o+(-0.773094138073)*x_ref[2]
        ref[3,2,0,0]=(0.574306840544)*x_ref[0]**o+(0.449094769988)*x_ref[0]+(-0.406303076614)*x_ref[1]**o+(-0.776648781906)*x_ref[1]+(-0.59829476647)*x_ref[2]**o+(0.257632446666)*x_ref[2]
        ref[3,2,0,1]=(0.689106127051)*x_ref[0]**o+(0.717184851708)*x_ref[0]+(-0.495581943615)*x_ref[1]**o+(-0.30329139788)*x_ref[1]+(0.901753128535)*x_ref[2]**o+(0.48496505727)*x_ref[2]
        ref[3,2,1,0]=(0.252050902256)*x_ref[0]**o+(-0.0414281706943)*x_ref[0]+(-0.284125627654)*x_ref[1]**o+(0.477450554269)*x_ref[1]+(0.849202242148)*x_ref[2]**o+(0.211826565889)*x_ref[2]
        ref[3,2,1,1]=(-0.319900673571)*x_ref[0]**o+(0.965993723739)*x_ref[0]+(0.671360092314)*x_ref[1]**o+(-0.652805407934)*x_ref[1]+(-0.0352316646853)*x_ref[2]**o+(-0.442493499306)*x_ref[2]
        ref[3,2,2,0]=(0.525577660439)*x_ref[0]**o+(0.660882378915)*x_ref[0]+(-0.419826738105)*x_ref[1]**o+(-0.798251889077)*x_ref[1]+(-0.430382393461)*x_ref[2]**o+(0.714357242496)*x_ref[2]
        ref[3,2,2,1]=(-0.200634944034)*x_ref[0]**o+(0.211086366034)*x_ref[0]+(0.116289218188)*x_ref[1]**o+(-0.785792860501)*x_ref[1]+(0.763320327438)*x_ref[2]**o+(-0.457355826175)*x_ref[2]
        ref[3,3,0,0]=(-0.890873783782)*x_ref[0]**o+(0.322610697657)*x_ref[0]+(-0.351437390332)*x_ref[1]**o+(-0.720698169708)*x_ref[1]+(-0.664444440194)*x_ref[2]**o+(0.274865394829)*x_ref[2]
        ref[3,3,0,1]=(-0.384885652342)*x_ref[0]**o+(-0.409799972422)*x_ref[0]+(0.679315041587)*x_ref[1]**o+(0.0859788488759)*x_ref[1]+(-0.136173937432)*x_ref[2]**o+(-0.136591429212)*x_ref[2]
        ref[3,3,1,0]=(0.147934741321)*x_ref[0]**o+(0.246941956666)*x_ref[0]+(0.696831123708)*x_ref[1]**o+(-0.905966322958)*x_ref[1]+(0.672705695255)*x_ref[2]**o+(-0.776017262395)*x_ref[2]
        ref[3,3,1,1]=(0.933982654679)*x_ref[0]**o+(-0.367246348547)*x_ref[0]+(-0.254076834401)*x_ref[1]**o+(0.148617981234)*x_ref[1]+(-0.486881004358)*x_ref[2]**o+(0.636371570082)*x_ref[2]
        ref[3,3,2,0]=(-0.536125392071)*x_ref[0]**o+(-0.663380779479)*x_ref[0]+(0.0655694689011)*x_ref[1]**o+(0.0390002326894)*x_ref[1]+(-0.731730129742)*x_ref[2]**o+(-0.578791485067)*x_ref[2]
        ref[3,3,2,1]=(0.705873902409)*x_ref[0]**o+(-0.0444303506955)*x_ref[0]+(-0.978519191225)*x_ref[1]**o+(-0.0280878186958)*x_ref[1]+(0.182378726121)*x_ref[2]**o+(-0.555345919696)*x_ref[2]
        ref[3,4,0,0]=(0.945945862395)*x_ref[0]**o+(-0.448487950632)*x_ref[0]+(-0.477570573229)*x_ref[1]**o+(0.313871838624)*x_ref[1]+(0.237944220287)*x_ref[2]**o+(-0.218095130482)*x_ref[2]
        ref[3,4,0,1]=(-0.758647277269)*x_ref[0]**o+(-0.550604667208)*x_ref[0]+(-0.262944399053)*x_ref[1]**o+(0.994302652859)*x_ref[1]+(0.363440310008)*x_ref[2]**o+(-0.127863970613)*x_ref[2]
        ref[3,4,1,0]=(0.667202433965)*x_ref[0]**o+(0.954901068821)*x_ref[0]+(0.719745510318)*x_ref[1]**o+(-0.763882742007)*x_ref[1]+(0.880927476123)*x_ref[2]**o+(0.102034630893)*x_ref[2]
        ref[3,4,1,1]=(-0.684644024456)*x_ref[0]**o+(0.88037275093)*x_ref[0]+(-0.772213766289)*x_ref[1]**o+(-0.702615214094)*x_ref[1]+(-0.25661708053)*x_ref[2]**o+(0.126779040694)*x_ref[2]
        ref[3,4,2,0]=(-0.589611385303)*x_ref[0]**o+(0.638712635252)*x_ref[0]+(0.745450741856)*x_ref[1]**o+(0.399494092815)*x_ref[1]+(-0.717293809038)*x_ref[2]**o+(0.305662737751)*x_ref[2]
        ref[3,4,2,1]=(0.0572426997796)*x_ref[0]**o+(-0.764600144578)*x_ref[0]+(0.482545219565)*x_ref[1]**o+(0.525883069411)*x_ref[1]+(-0.612224327195)*x_ref[2]**o+(0.354344820256)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.100067148208)*x[0]+(-0.8375743725)*x[1]
        ref=(-0.100067148208)*x_ref[0]+(-0.8375743725)*x_ref[1]
      else:
        arg=(-0.247343416726)*x[0]+(0.604193484164)*x[1]+(-0.219899438287)*x[2]
        ref=(-0.247343416726)*x_ref[0]+(0.604193484164)*x_ref[1]+(-0.219899438287)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        arg[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
        ref[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        ref[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
      else:
        arg[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        arg[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
        ref[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        ref[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.00153698933322)*x[0]+(0.392281645764)*x[1]
        arg[0,1]=(-1.2794655544)*x[0]+(-0.373519639818)*x[1]
        arg[0,2]=(0.30127033321)*x[0]+(-0.0996865853738)*x[1]
        arg[0,3]=(1.23405799349)*x[0]+(0.498558200339)*x[1]
        arg[0,4]=(0.85519350742)*x[0]+(-0.19188046135)*x[1]
        arg[1,0]=(0.417367834962)*x[0]+(0.992318201639)*x[1]
        arg[1,1]=(-0.646432128833)*x[0]+(-0.872750321582)*x[1]
        arg[1,2]=(-0.806117185061)*x[0]+(-0.89396408325)*x[1]
        arg[1,3]=(0.768017237627)*x[0]+(-1.2904111912)*x[1]
        arg[1,4]=(-0.0159685523188)*x[0]+(0.401410193169)*x[1]
        arg[2,0]=(-0.844501716181)*x[0]+(-0.123115343819)*x[1]
        arg[2,1]=(1.08441909409)*x[0]+(0.152432282539)*x[1]
        arg[2,2]=(-0.60551037413)*x[0]+(1.15692828826)*x[1]
        arg[2,3]=(-1.21509168541)*x[0]+(1.03683236377)*x[1]
        arg[2,4]=(1.16850730134)*x[0]+(0.162202615909)*x[1]
        arg[3,0]=(0.727772166897)*x[0]+(-0.296815840935)*x[1]
        arg[3,1]=(-0.0880577151489)*x[0]+(1.66427691426)*x[1]
        arg[3,2]=(0.415144534458)*x[0]+(-0.555833723295)*x[1]
        arg[3,3]=(1.09444635829)*x[0]+(-0.335209417536)*x[1]
        arg[3,4]=(-1.05047769585)*x[0]+(0.684427039986)*x[1]
        ref[0,0]=(0.00153698933322)*x_ref[0]+(0.392281645764)*x_ref[1]
        ref[0,1]=(-1.2794655544)*x_ref[0]+(-0.373519639818)*x_ref[1]
        ref[0,2]=(0.30127033321)*x_ref[0]+(-0.0996865853738)*x_ref[1]
        ref[0,3]=(1.23405799349)*x_ref[0]+(0.498558200339)*x_ref[1]
        ref[0,4]=(0.85519350742)*x_ref[0]+(-0.19188046135)*x_ref[1]
        ref[1,0]=(0.417367834962)*x_ref[0]+(0.992318201639)*x_ref[1]
        ref[1,1]=(-0.646432128833)*x_ref[0]+(-0.872750321582)*x_ref[1]
        ref[1,2]=(-0.806117185061)*x_ref[0]+(-0.89396408325)*x_ref[1]
        ref[1,3]=(0.768017237627)*x_ref[0]+(-1.2904111912)*x_ref[1]
        ref[1,4]=(-0.0159685523188)*x_ref[0]+(0.401410193169)*x_ref[1]
        ref[2,0]=(-0.844501716181)*x_ref[0]+(-0.123115343819)*x_ref[1]
        ref[2,1]=(1.08441909409)*x_ref[0]+(0.152432282539)*x_ref[1]
        ref[2,2]=(-0.60551037413)*x_ref[0]+(1.15692828826)*x_ref[1]
        ref[2,3]=(-1.21509168541)*x_ref[0]+(1.03683236377)*x_ref[1]
        ref[2,4]=(1.16850730134)*x_ref[0]+(0.162202615909)*x_ref[1]
        ref[3,0]=(0.727772166897)*x_ref[0]+(-0.296815840935)*x_ref[1]
        ref[3,1]=(-0.0880577151489)*x_ref[0]+(1.66427691426)*x_ref[1]
        ref[3,2]=(0.415144534458)*x_ref[0]+(-0.555833723295)*x_ref[1]
        ref[3,3]=(1.09444635829)*x_ref[0]+(-0.335209417536)*x_ref[1]
        ref[3,4]=(-1.05047769585)*x_ref[0]+(0.684427039986)*x_ref[1]
      else:
        arg[0,0]=(-0.611607640773)*x[0]+(0.584485971628)*x[1]+(0.0090316337643)*x[2]
        arg[0,1]=(0.905278285857)*x[0]+(0.332523151242)*x[1]+(0.9909841807)*x[2]
        arg[0,2]=(-1.06629333132)*x[0]+(-0.38478180064)*x[1]+(-0.63309211256)*x[2]
        arg[0,3]=(0.122436386344)*x[0]+(-0.218752671204)*x[1]+(-0.746695172216)*x[2]
        arg[0,4]=(1.03152124874)*x[0]+(-1.11327271563)*x[1]+(-0.257591112135)*x[2]
        arg[1,0]=(0.969019600299)*x[0]+(-0.320279309689)*x[1]+(0.327767746545)*x[2]
        arg[1,1]=(0.291023069052)*x[0]+(0.78112901064)*x[1]+(0.879257239042)*x[2]
        arg[1,2]=(-0.131262891536)*x[0]+(-0.0109690229153)*x[1]+(-0.171900332406)*x[2]
        arg[1,3]=(-1.32272684486)*x[0]+(-0.256869536085)*x[1]+(0.0308524567462)*x[2]
        arg[1,4]=(1.30707061279)*x[0]+(0.972866282634)*x[1]+(-0.474948537882)*x[2]
        arg[2,0]=(0.169034210421)*x[0]+(-1.08714896226)*x[1]+(-0.636838450708)*x[2]
        arg[2,1]=(-0.470752780773)*x[0]+(-0.695286088059)*x[1]+(-1.19197058712)*x[2]
        arg[2,2]=(0.701403721228)*x[0]+(0.551002482436)*x[1]+(-0.295853851234)*x[2]
        arg[2,3]=(-0.796431361745)*x[0]+(0.0581884766856)*x[1]+(-1.13668563089)*x[2]
        arg[2,4]=(-0.733185798017)*x[0]+(0.178900103735)*x[1]+(-0.845671526781)*x[2]
        arg[3,0]=(-1.47064927107)*x[0]+(1.33393324965)*x[1]+(-0.767393460937)*x[2]
        arg[3,1]=(1.36784513226)*x[0]+(0.100008910431)*x[1]+(-1.09210256713)*x[2]
        arg[3,2]=(-0.873076669977)*x[0]+(0.367213631778)*x[1]+(0.0698296920039)*x[2]
        arg[3,3]=(-1.76609214158)*x[0]+(-1.26186286566)*x[1]+(0.377817035113)*x[2]
        arg[3,4]=(-1.28261709396)*x[0]+(0.240480767533)*x[1]+(0.307813279357)*x[2]
        ref[0,0]=(-0.611607640773)*x_ref[0]+(0.584485971628)*x_ref[1]+(0.0090316337643)*x_ref[2]
        ref[0,1]=(0.905278285857)*x_ref[0]+(0.332523151242)*x_ref[1]+(0.9909841807)*x_ref[2]
        ref[0,2]=(-1.06629333132)*x_ref[0]+(-0.38478180064)*x_ref[1]+(-0.63309211256)*x_ref[2]
        ref[0,3]=(0.122436386344)*x_ref[0]+(-0.218752671204)*x_ref[1]+(-0.746695172216)*x_ref[2]
        ref[0,4]=(1.03152124874)*x_ref[0]+(-1.11327271563)*x_ref[1]+(-0.257591112135)*x_ref[2]
        ref[1,0]=(0.969019600299)*x_ref[0]+(-0.320279309689)*x_ref[1]+(0.327767746545)*x_ref[2]
        ref[1,1]=(0.291023069052)*x_ref[0]+(0.78112901064)*x_ref[1]+(0.879257239042)*x_ref[2]
        ref[1,2]=(-0.131262891536)*x_ref[0]+(-0.0109690229153)*x_ref[1]+(-0.171900332406)*x_ref[2]
        ref[1,3]=(-1.32272684486)*x_ref[0]+(-0.256869536085)*x_ref[1]+(0.0308524567462)*x_ref[2]
        ref[1,4]=(1.30707061279)*x_ref[0]+(0.972866282634)*x_ref[1]+(-0.474948537882)*x_ref[2]
        ref[2,0]=(0.169034210421)*x_ref[0]+(-1.08714896226)*x_ref[1]+(-0.636838450708)*x_ref[2]
        ref[2,1]=(-0.470752780773)*x_ref[0]+(-0.695286088059)*x_ref[1]+(-1.19197058712)*x_ref[2]
        ref[2,2]=(0.701403721228)*x_ref[0]+(0.551002482436)*x_ref[1]+(-0.295853851234)*x_ref[2]
        ref[2,3]=(-0.796431361745)*x_ref[0]+(0.0581884766856)*x_ref[1]+(-1.13668563089)*x_ref[2]
        ref[2,4]=(-0.733185798017)*x_ref[0]+(0.178900103735)*x_ref[1]+(-0.845671526781)*x_ref[2]
        ref[3,0]=(-1.47064927107)*x_ref[0]+(1.33393324965)*x_ref[1]+(-0.767393460937)*x_ref[2]
        ref[3,1]=(1.36784513226)*x_ref[0]+(0.100008910431)*x_ref[1]+(-1.09210256713)*x_ref[2]
        ref[3,2]=(-0.873076669977)*x_ref[0]+(0.367213631778)*x_ref[1]+(0.0698296920039)*x_ref[2]
        ref[3,3]=(-1.76609214158)*x_ref[0]+(-1.26186286566)*x_ref[1]+(0.377817035113)*x_ref[2]
        ref[3,4]=(-1.28261709396)*x_ref[0]+(0.240480767533)*x_ref[1]+(0.307813279357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.423505169404)*x[0]+(-1.10390334264)*x[1]
        arg[0,0,1]=(-0.181820089254)*x[0]+(-0.374602268791)*x[1]
        arg[0,1,0]=(-1.36906917075)*x[0]+(0.26554385908)*x[1]
        arg[0,1,1]=(-0.280661572374)*x[0]+(-0.239232524287)*x[1]
        arg[1,0,0]=(-1.25670251934)*x[0]+(-0.429353977735)*x[1]
        arg[1,0,1]=(-0.747928970735)*x[0]+(1.17874317611)*x[1]
        arg[1,1,0]=(-0.548814611802)*x[0]+(-0.0249315764285)*x[1]
        arg[1,1,1]=(-0.481123117522)*x[0]+(1.23402203879)*x[1]
        arg[2,0,0]=(-0.0639512689884)*x[0]+(-0.0278213371376)*x[1]
        arg[2,0,1]=(-1.17292797621)*x[0]+(0.42140600304)*x[1]
        arg[2,1,0]=(-0.358052601083)*x[0]+(-1.66807880791)*x[1]
        arg[2,1,1]=(-0.291433251349)*x[0]+(0.111473162108)*x[1]
        arg[3,0,0]=(0.742643625989)*x[0]+(-1.29214106228)*x[1]
        arg[3,0,1]=(0.319887321826)*x[0]+(-0.40213369238)*x[1]
        arg[3,1,0]=(-0.449412064097)*x[0]+(-1.10654262009)*x[1]
        arg[3,1,1]=(1.19722227632)*x[0]+(1.3455799026)*x[1]
        arg[4,0,0]=(-1.77453188487)*x[0]+(0.634838793872)*x[1]
        arg[4,0,1]=(-0.215461066907)*x[0]+(1.11635233844)*x[1]
        arg[4,1,0]=(0.835816652592)*x[0]+(0.668378554214)*x[1]
        arg[4,1,1]=(-1.22133225406)*x[0]+(-0.148780252398)*x[1]
        arg[5,0,0]=(0.320245409639)*x[0]+(-1.08828513776)*x[1]
        arg[5,0,1]=(-0.1865835417)*x[0]+(-0.353282061874)*x[1]
        arg[5,1,0]=(1.38858040685)*x[0]+(0.166398261098)*x[1]
        arg[5,1,1]=(0.947718875755)*x[0]+(-1.06171934302)*x[1]
        ref[0,0,0]=(-0.423505169404)*x_ref[0]+(-1.10390334264)*x_ref[1]
        ref[0,0,1]=(-0.181820089254)*x_ref[0]+(-0.374602268791)*x_ref[1]
        ref[0,1,0]=(-1.36906917075)*x_ref[0]+(0.26554385908)*x_ref[1]
        ref[0,1,1]=(-0.280661572374)*x_ref[0]+(-0.239232524287)*x_ref[1]
        ref[1,0,0]=(-1.25670251934)*x_ref[0]+(-0.429353977735)*x_ref[1]
        ref[1,0,1]=(-0.747928970735)*x_ref[0]+(1.17874317611)*x_ref[1]
        ref[1,1,0]=(-0.548814611802)*x_ref[0]+(-0.0249315764285)*x_ref[1]
        ref[1,1,1]=(-0.481123117522)*x_ref[0]+(1.23402203879)*x_ref[1]
        ref[2,0,0]=(-0.0639512689884)*x_ref[0]+(-0.0278213371376)*x_ref[1]
        ref[2,0,1]=(-1.17292797621)*x_ref[0]+(0.42140600304)*x_ref[1]
        ref[2,1,0]=(-0.358052601083)*x_ref[0]+(-1.66807880791)*x_ref[1]
        ref[2,1,1]=(-0.291433251349)*x_ref[0]+(0.111473162108)*x_ref[1]
        ref[3,0,0]=(0.742643625989)*x_ref[0]+(-1.29214106228)*x_ref[1]
        ref[3,0,1]=(0.319887321826)*x_ref[0]+(-0.40213369238)*x_ref[1]
        ref[3,1,0]=(-0.449412064097)*x_ref[0]+(-1.10654262009)*x_ref[1]
        ref[3,1,1]=(1.19722227632)*x_ref[0]+(1.3455799026)*x_ref[1]
        ref[4,0,0]=(-1.77453188487)*x_ref[0]+(0.634838793872)*x_ref[1]
        ref[4,0,1]=(-0.215461066907)*x_ref[0]+(1.11635233844)*x_ref[1]
        ref[4,1,0]=(0.835816652592)*x_ref[0]+(0.668378554214)*x_ref[1]
        ref[4,1,1]=(-1.22133225406)*x_ref[0]+(-0.148780252398)*x_ref[1]
        ref[5,0,0]=(0.320245409639)*x_ref[0]+(-1.08828513776)*x_ref[1]
        ref[5,0,1]=(-0.1865835417)*x_ref[0]+(-0.353282061874)*x_ref[1]
        ref[5,1,0]=(1.38858040685)*x_ref[0]+(0.166398261098)*x_ref[1]
        ref[5,1,1]=(0.947718875755)*x_ref[0]+(-1.06171934302)*x_ref[1]
      else:
        arg[0,0,0]=(-0.892096322478)*x[0]+(0.720363183451)*x[1]+(0.600609272191)*x[2]
        arg[0,0,1]=(-0.418024461935)*x[0]+(-0.819275292035)*x[1]+(0.134962557692)*x[2]
        arg[0,1,0]=(-0.449783838361)*x[0]+(0.943677173533)*x[1]+(0.209424190991)*x[2]
        arg[0,1,1]=(-0.424386775004)*x[0]+(-0.213024388094)*x[1]+(-1.05927943384)*x[2]
        arg[1,0,0]=(-0.699504096539)*x[0]+(-0.153570308377)*x[1]+(0.576902392658)*x[2]
        arg[1,0,1]=(-0.426914055661)*x[0]+(-0.330405267059)*x[1]+(1.07656632155)*x[2]
        arg[1,1,0]=(0.775238005352)*x[0]+(1.11012072575)*x[1]+(0.764627098603)*x[2]
        arg[1,1,1]=(0.0130424708408)*x[0]+(-0.671778102107)*x[1]+(0.780684368983)*x[2]
        arg[2,0,0]=(-0.0512224258498)*x[0]+(-1.50515803984)*x[1]+(0.342824143946)*x[2]
        arg[2,0,1]=(0.117541624176)*x[0]+(0.372415524552)*x[1]+(0.255218095016)*x[2]
        arg[2,1,0]=(-0.574508895703)*x[0]+(0.345217292359)*x[1]+(0.209211956821)*x[2]
        arg[2,1,1]=(0.93643587147)*x[0]+(-0.096209009902)*x[1]+(0.205150859527)*x[2]
        arg[3,0,0]=(1.7541929884)*x[0]+(0.169061478447)*x[1]+(0.534746695515)*x[2]
        arg[3,0,1]=(0.442329266782)*x[0]+(0.334971710414)*x[1]+(1.66892557698)*x[2]
        arg[3,1,0]=(0.316436716758)*x[0]+(-1.25442030987)*x[1]+(0.372183105374)*x[2]
        arg[3,1,1]=(-0.254157820265)*x[0]+(0.287381300023)*x[1]+(0.0574117423352)*x[2]
        arg[4,0,0]=(-1.08380257521)*x[0]+(-1.31461890032)*x[1]+(-0.503802775099)*x[2]
        arg[4,0,1]=(-0.558293486798)*x[0]+(0.702317635925)*x[1]+(0.463814829137)*x[2]
        arg[4,1,0]=(-0.677160971974)*x[0]+(-0.14921275312)*x[1]+(-1.6501459435)*x[2]
        arg[4,1,1]=(-0.176672532474)*x[0]+(0.945017605811)*x[1]+(-0.790342939788)*x[2]
        arg[5,0,0]=(-1.45668070675)*x[0]+(-0.404957129204)*x[1]+(1.37383065954)*x[2]
        arg[5,0,1]=(0.53829326756)*x[0]+(0.543333031479)*x[1]+(-0.854057296362)*x[2]
        arg[5,1,0]=(0.805864283921)*x[0]+(-0.00186634953513)*x[1]+(-1.97111362539)*x[2]
        arg[5,1,1]=(-0.387746238967)*x[0]+(-0.739158353882)*x[1]+(-0.134082808118)*x[2]
        ref[0,0,0]=(-0.892096322478)*x_ref[0]+(0.720363183451)*x_ref[1]+(0.600609272191)*x_ref[2]
        ref[0,0,1]=(-0.418024461935)*x_ref[0]+(-0.819275292035)*x_ref[1]+(0.134962557692)*x_ref[2]
        ref[0,1,0]=(-0.449783838361)*x_ref[0]+(0.943677173533)*x_ref[1]+(0.209424190991)*x_ref[2]
        ref[0,1,1]=(-0.424386775004)*x_ref[0]+(-0.213024388094)*x_ref[1]+(-1.05927943384)*x_ref[2]
        ref[1,0,0]=(-0.699504096539)*x_ref[0]+(-0.153570308377)*x_ref[1]+(0.576902392658)*x_ref[2]
        ref[1,0,1]=(-0.426914055661)*x_ref[0]+(-0.330405267059)*x_ref[1]+(1.07656632155)*x_ref[2]
        ref[1,1,0]=(0.775238005352)*x_ref[0]+(1.11012072575)*x_ref[1]+(0.764627098603)*x_ref[2]
        ref[1,1,1]=(0.0130424708408)*x_ref[0]+(-0.671778102107)*x_ref[1]+(0.780684368983)*x_ref[2]
        ref[2,0,0]=(-0.0512224258498)*x_ref[0]+(-1.50515803984)*x_ref[1]+(0.342824143946)*x_ref[2]
        ref[2,0,1]=(0.117541624176)*x_ref[0]+(0.372415524552)*x_ref[1]+(0.255218095016)*x_ref[2]
        ref[2,1,0]=(-0.574508895703)*x_ref[0]+(0.345217292359)*x_ref[1]+(0.209211956821)*x_ref[2]
        ref[2,1,1]=(0.93643587147)*x_ref[0]+(-0.096209009902)*x_ref[1]+(0.205150859527)*x_ref[2]
        ref[3,0,0]=(1.7541929884)*x_ref[0]+(0.169061478447)*x_ref[1]+(0.534746695515)*x_ref[2]
        ref[3,0,1]=(0.442329266782)*x_ref[0]+(0.334971710414)*x_ref[1]+(1.66892557698)*x_ref[2]
        ref[3,1,0]=(0.316436716758)*x_ref[0]+(-1.25442030987)*x_ref[1]+(0.372183105374)*x_ref[2]
        ref[3,1,1]=(-0.254157820265)*x_ref[0]+(0.287381300023)*x_ref[1]+(0.0574117423352)*x_ref[2]
        ref[4,0,0]=(-1.08380257521)*x_ref[0]+(-1.31461890032)*x_ref[1]+(-0.503802775099)*x_ref[2]
        ref[4,0,1]=(-0.558293486798)*x_ref[0]+(0.702317635925)*x_ref[1]+(0.463814829137)*x_ref[2]
        ref[4,1,0]=(-0.677160971974)*x_ref[0]+(-0.14921275312)*x_ref[1]+(-1.6501459435)*x_ref[2]
        ref[4,1,1]=(-0.176672532474)*x_ref[0]+(0.945017605811)*x_ref[1]+(-0.790342939788)*x_ref[2]
        ref[5,0,0]=(-1.45668070675)*x_ref[0]+(-0.404957129204)*x_ref[1]+(1.37383065954)*x_ref[2]
        ref[5,0,1]=(0.53829326756)*x_ref[0]+(0.543333031479)*x_ref[1]+(-0.854057296362)*x_ref[2]
        ref[5,1,0]=(0.805864283921)*x_ref[0]+(-0.00186634953513)*x_ref[1]+(-1.97111362539)*x_ref[2]
        ref[5,1,1]=(-0.387746238967)*x_ref[0]+(-0.739158353882)*x_ref[1]+(-0.134082808118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.754311658086)*x[0]+(0.235157034342)*x[1]
        arg[0,0,0,1]=(-1.33786315024)*x[0]+(0.243349279742)*x[1]
        arg[0,0,1,0]=(1.48657750088)*x[0]+(1.09853108305)*x[1]
        arg[0,0,1,1]=(1.4501362266)*x[0]+(0.824811567155)*x[1]
        arg[0,0,2,0]=(1.12567566082)*x[0]+(0.663525681174)*x[1]
        arg[0,0,2,1]=(1.48991565019)*x[0]+(-0.680795365011)*x[1]
        arg[0,1,0,0]=(0.268571449528)*x[0]+(0.818595834672)*x[1]
        arg[0,1,0,1]=(-1.55490945555)*x[0]+(0.783595197888)*x[1]
        arg[0,1,1,0]=(0.913201382585)*x[0]+(0.855491379376)*x[1]
        arg[0,1,1,1]=(-0.179192075116)*x[0]+(-0.690415079523)*x[1]
        arg[0,1,2,0]=(-0.0393574502912)*x[0]+(-0.690157650634)*x[1]
        arg[0,1,2,1]=(0.784216863217)*x[0]+(0.498735486159)*x[1]
        arg[0,2,0,0]=(0.137939554165)*x[0]+(-1.33961519113)*x[1]
        arg[0,2,0,1]=(-1.03897311736)*x[0]+(-1.37797405312)*x[1]
        arg[0,2,1,0]=(0.199812003756)*x[0]+(-0.0388499674037)*x[1]
        arg[0,2,1,1]=(1.15530534202)*x[0]+(-1.28280869078)*x[1]
        arg[0,2,2,0]=(0.836447791155)*x[0]+(1.2627066026)*x[1]
        arg[0,2,2,1]=(-0.584608492688)*x[0]+(-1.55305974305)*x[1]
        arg[0,3,0,0]=(-0.661952381799)*x[0]+(-1.27062403778)*x[1]
        arg[0,3,0,1]=(-1.47789532259)*x[0]+(-0.279583524936)*x[1]
        arg[0,3,1,0]=(1.40729771357)*x[0]+(0.0971284144528)*x[1]
        arg[0,3,1,1]=(-0.459167955911)*x[0]+(-1.29033281524)*x[1]
        arg[0,3,2,0]=(-1.63189436824)*x[0]+(0.75729808422)*x[1]
        arg[0,3,2,1]=(0.522479828718)*x[0]+(-0.0231299515383)*x[1]
        arg[0,4,0,0]=(0.241891446615)*x[0]+(0.803327107651)*x[1]
        arg[0,4,0,1]=(0.273613502836)*x[0]+(-0.0115180464171)*x[1]
        arg[0,4,1,0]=(1.96219818696)*x[0]+(-0.349137094743)*x[1]
        arg[0,4,1,1]=(0.365690821719)*x[0]+(-1.6943347181)*x[1]
        arg[0,4,2,0]=(-0.283970779872)*x[0]+(-0.258537179804)*x[1]
        arg[0,4,2,1]=(-0.973333430598)*x[0]+(1.48706276212)*x[1]
        arg[1,0,0,0]=(1.18757466971)*x[0]+(0.838804656357)*x[1]
        arg[1,0,0,1]=(0.840859904702)*x[0]+(1.09955195682)*x[1]
        arg[1,0,1,0]=(1.46190513227)*x[0]+(-0.596959830422)*x[1]
        arg[1,0,1,1]=(0.718689296745)*x[0]+(-0.628074620222)*x[1]
        arg[1,0,2,0]=(0.440761293041)*x[0]+(0.115800854194)*x[1]
        arg[1,0,2,1]=(1.34891146904)*x[0]+(0.813690260652)*x[1]
        arg[1,1,0,0]=(0.183071638098)*x[0]+(-1.37996210755)*x[1]
        arg[1,1,0,1]=(0.258142115988)*x[0]+(-0.757709381529)*x[1]
        arg[1,1,1,0]=(1.72392829309)*x[0]+(0.444995102781)*x[1]
        arg[1,1,1,1]=(-0.380898209988)*x[0]+(-0.0950246860347)*x[1]
        arg[1,1,2,0]=(0.203007226016)*x[0]+(1.09786001832)*x[1]
        arg[1,1,2,1]=(-0.502500865341)*x[0]+(-0.629980527995)*x[1]
        arg[1,2,0,0]=(0.482549518436)*x[0]+(0.565840824146)*x[1]
        arg[1,2,0,1]=(0.355818565068)*x[0]+(-0.78273295574)*x[1]
        arg[1,2,1,0]=(1.79204033071)*x[0]+(-0.698040851585)*x[1]
        arg[1,2,1,1]=(-0.534827741044)*x[0]+(0.00943131627498)*x[1]
        arg[1,2,2,0]=(-1.00663524444)*x[0]+(0.0207854553456)*x[1]
        arg[1,2,2,1]=(-0.338048040159)*x[0]+(-0.380607406217)*x[1]
        arg[1,3,0,0]=(0.586310549044)*x[0]+(0.276326798805)*x[1]
        arg[1,3,0,1]=(-0.260482580702)*x[0]+(-0.349279395321)*x[1]
        arg[1,3,1,0]=(-0.196533829895)*x[0]+(0.614805272738)*x[1]
        arg[1,3,1,1]=(-0.363176483895)*x[0]+(-0.925817017731)*x[1]
        arg[1,3,2,0]=(-1.24182075421)*x[0]+(0.215966446013)*x[1]
        arg[1,3,2,1]=(1.07546652642)*x[0]+(0.383803991974)*x[1]
        arg[1,4,0,0]=(0.199312303525)*x[0]+(-0.314520883859)*x[1]
        arg[1,4,0,1]=(-1.56890826152)*x[0]+(-0.133086662241)*x[1]
        arg[1,4,1,0]=(1.1307241936)*x[0]+(-1.32905744984)*x[1]
        arg[1,4,1,1]=(-0.385991455298)*x[0]+(0.292913024507)*x[1]
        arg[1,4,2,0]=(1.14561678056)*x[0]+(-0.662696162276)*x[1]
        arg[1,4,2,1]=(-1.49533948954)*x[0]+(1.43396138744)*x[1]
        arg[2,0,0,0]=(-0.63160396493)*x[0]+(0.813120998776)*x[1]
        arg[2,0,0,1]=(0.65598998434)*x[0]+(-1.02445537553)*x[1]
        arg[2,0,1,0]=(-1.13473104075)*x[0]+(0.057331183449)*x[1]
        arg[2,0,1,1]=(-0.154892482107)*x[0]+(0.481969957579)*x[1]
        arg[2,0,2,0]=(0.521916671496)*x[0]+(0.546676943059)*x[1]
        arg[2,0,2,1]=(-1.68919379028)*x[0]+(1.08292038981)*x[1]
        arg[2,1,0,0]=(-0.643846595888)*x[0]+(0.528306097755)*x[1]
        arg[2,1,0,1]=(1.3745090497)*x[0]+(-0.0801648509184)*x[1]
        arg[2,1,1,0]=(0.129096817054)*x[0]+(0.90750370855)*x[1]
        arg[2,1,1,1]=(0.539309513305)*x[0]+(1.34949470961)*x[1]
        arg[2,1,2,0]=(1.4189703896)*x[0]+(-1.75011201893)*x[1]
        arg[2,1,2,1]=(-1.09325548386)*x[0]+(0.148279301716)*x[1]
        arg[2,2,0,0]=(-1.25785133743)*x[0]+(0.275071306594)*x[1]
        arg[2,2,0,1]=(-0.550152068761)*x[0]+(-1.20422484275)*x[1]
        arg[2,2,1,0]=(-0.102062863668)*x[0]+(-1.34431887541)*x[1]
        arg[2,2,1,1]=(0.103354724169)*x[0]+(-0.941952180446)*x[1]
        arg[2,2,2,0]=(1.2090339604)*x[0]+(-0.330357976796)*x[1]
        arg[2,2,2,1]=(0.203162449188)*x[0]+(-1.03640702014)*x[1]
        arg[2,3,0,0]=(-0.324576353514)*x[0]+(0.582614539267)*x[1]
        arg[2,3,0,1]=(-0.213823136215)*x[0]+(-0.473199105214)*x[1]
        arg[2,3,1,0]=(0.395144702599)*x[0]+(0.0358194467152)*x[1]
        arg[2,3,1,1]=(1.32860563396)*x[0]+(-0.577308310732)*x[1]
        arg[2,3,2,0]=(0.994205029432)*x[0]+(1.13433517636)*x[1]
        arg[2,3,2,1]=(-0.488499286742)*x[0]+(0.224253820738)*x[1]
        arg[2,4,0,0]=(1.48035663008)*x[0]+(-1.09697986914)*x[1]
        arg[2,4,0,1]=(-1.09991915151)*x[0]+(-0.557951199985)*x[1]
        arg[2,4,1,0]=(-1.23862699804)*x[0]+(-0.230602187713)*x[1]
        arg[2,4,1,1]=(-0.922609370448)*x[0]+(1.76673953413)*x[1]
        arg[2,4,2,0]=(0.731769419709)*x[0]+(-1.1094516608)*x[1]
        arg[2,4,2,1]=(-0.145672504805)*x[0]+(0.168194421355)*x[1]
        arg[3,0,0,0]=(1.02559707602)*x[0]+(0.260431368523)*x[1]
        arg[3,0,0,1]=(0.111899568824)*x[0]+(-0.90725475423)*x[1]
        arg[3,0,1,0]=(-0.226547745683)*x[0]+(-0.232830393372)*x[1]
        arg[3,0,1,1]=(0.347004318471)*x[0]+(-0.196130970001)*x[1]
        arg[3,0,2,0]=(-1.72155871442)*x[0]+(-1.28372407452)*x[1]
        arg[3,0,2,1]=(1.43817914693)*x[0]+(-0.354093253684)*x[1]
        arg[3,1,0,0]=(-0.293939245102)*x[0]+(0.0255562139103)*x[1]
        arg[3,1,0,1]=(0.139306213018)*x[0]+(0.595062226642)*x[1]
        arg[3,1,1,0]=(-0.670339920672)*x[0]+(0.387520583192)*x[1]
        arg[3,1,1,1]=(-0.0323325470954)*x[0]+(-0.681162999279)*x[1]
        arg[3,1,2,0]=(0.235275048739)*x[0]+(0.144725473773)*x[1]
        arg[3,1,2,1]=(0.132211795214)*x[0]+(0.262780216845)*x[1]
        arg[3,2,0,0]=(0.590514326163)*x[0]+(-0.84273677918)*x[1]
        arg[3,2,0,1]=(-0.925181556185)*x[0]+(-0.300244447001)*x[1]
        arg[3,2,1,0]=(-0.0509700795112)*x[0]+(-1.31919389616)*x[1]
        arg[3,2,1,1]=(1.50577291846)*x[0]+(-0.219284607769)*x[1]
        arg[3,2,2,0]=(-0.0460616673346)*x[0]+(-0.690312685466)*x[1]
        arg[3,2,2,1]=(-0.673765477936)*x[0]+(-0.0965860475641)*x[1]
        arg[3,3,0,0]=(-0.6577594174)*x[0]+(0.171886277016)*x[1]
        arg[3,3,0,1]=(-1.08810434718)*x[0]+(0.333153250183)*x[1]
        arg[3,3,1,0]=(0.150968371983)*x[0]+(0.290859317391)*x[1]
        arg[3,3,1,1]=(1.62096265131)*x[0]+(-1.10309602741)*x[1]
        arg[3,3,2,0]=(-0.932063639972)*x[0]+(-0.897131561126)*x[1]
        arg[3,3,2,1]=(-0.819110834345)*x[0]+(1.59159109562)*x[1]
        arg[3,4,0,0]=(0.272662931155)*x[0]+(0.69026285756)*x[1]
        arg[3,4,0,1]=(-0.600556370757)*x[0]+(-0.0853533546302)*x[1]
        arg[3,4,1,0]=(1.18544735281)*x[0]+(0.217751189034)*x[1]
        arg[3,4,1,1]=(-0.594609531358)*x[0]+(0.882158349788)*x[1]
        arg[3,4,2,0]=(1.18876765552)*x[0]+(-0.411481586979)*x[1]
        arg[3,4,2,1]=(-1.093067792)*x[0]+(-0.266744260768)*x[1]
        ref[0,0,0,0]=(-0.754311658086)*x_ref[0]+(0.235157034342)*x_ref[1]
        ref[0,0,0,1]=(-1.33786315024)*x_ref[0]+(0.243349279742)*x_ref[1]
        ref[0,0,1,0]=(1.48657750088)*x_ref[0]+(1.09853108305)*x_ref[1]
        ref[0,0,1,1]=(1.4501362266)*x_ref[0]+(0.824811567155)*x_ref[1]
        ref[0,0,2,0]=(1.12567566082)*x_ref[0]+(0.663525681174)*x_ref[1]
        ref[0,0,2,1]=(1.48991565019)*x_ref[0]+(-0.680795365011)*x_ref[1]
        ref[0,1,0,0]=(0.268571449528)*x_ref[0]+(0.818595834672)*x_ref[1]
        ref[0,1,0,1]=(-1.55490945555)*x_ref[0]+(0.783595197888)*x_ref[1]
        ref[0,1,1,0]=(0.913201382585)*x_ref[0]+(0.855491379376)*x_ref[1]
        ref[0,1,1,1]=(-0.179192075116)*x_ref[0]+(-0.690415079523)*x_ref[1]
        ref[0,1,2,0]=(-0.0393574502912)*x_ref[0]+(-0.690157650634)*x_ref[1]
        ref[0,1,2,1]=(0.784216863217)*x_ref[0]+(0.498735486159)*x_ref[1]
        ref[0,2,0,0]=(0.137939554165)*x_ref[0]+(-1.33961519113)*x_ref[1]
        ref[0,2,0,1]=(-1.03897311736)*x_ref[0]+(-1.37797405312)*x_ref[1]
        ref[0,2,1,0]=(0.199812003756)*x_ref[0]+(-0.0388499674037)*x_ref[1]
        ref[0,2,1,1]=(1.15530534202)*x_ref[0]+(-1.28280869078)*x_ref[1]
        ref[0,2,2,0]=(0.836447791155)*x_ref[0]+(1.2627066026)*x_ref[1]
        ref[0,2,2,1]=(-0.584608492688)*x_ref[0]+(-1.55305974305)*x_ref[1]
        ref[0,3,0,0]=(-0.661952381799)*x_ref[0]+(-1.27062403778)*x_ref[1]
        ref[0,3,0,1]=(-1.47789532259)*x_ref[0]+(-0.279583524936)*x_ref[1]
        ref[0,3,1,0]=(1.40729771357)*x_ref[0]+(0.0971284144528)*x_ref[1]
        ref[0,3,1,1]=(-0.459167955911)*x_ref[0]+(-1.29033281524)*x_ref[1]
        ref[0,3,2,0]=(-1.63189436824)*x_ref[0]+(0.75729808422)*x_ref[1]
        ref[0,3,2,1]=(0.522479828718)*x_ref[0]+(-0.0231299515383)*x_ref[1]
        ref[0,4,0,0]=(0.241891446615)*x_ref[0]+(0.803327107651)*x_ref[1]
        ref[0,4,0,1]=(0.273613502836)*x_ref[0]+(-0.0115180464171)*x_ref[1]
        ref[0,4,1,0]=(1.96219818696)*x_ref[0]+(-0.349137094743)*x_ref[1]
        ref[0,4,1,1]=(0.365690821719)*x_ref[0]+(-1.6943347181)*x_ref[1]
        ref[0,4,2,0]=(-0.283970779872)*x_ref[0]+(-0.258537179804)*x_ref[1]
        ref[0,4,2,1]=(-0.973333430598)*x_ref[0]+(1.48706276212)*x_ref[1]
        ref[1,0,0,0]=(1.18757466971)*x_ref[0]+(0.838804656357)*x_ref[1]
        ref[1,0,0,1]=(0.840859904702)*x_ref[0]+(1.09955195682)*x_ref[1]
        ref[1,0,1,0]=(1.46190513227)*x_ref[0]+(-0.596959830422)*x_ref[1]
        ref[1,0,1,1]=(0.718689296745)*x_ref[0]+(-0.628074620222)*x_ref[1]
        ref[1,0,2,0]=(0.440761293041)*x_ref[0]+(0.115800854194)*x_ref[1]
        ref[1,0,2,1]=(1.34891146904)*x_ref[0]+(0.813690260652)*x_ref[1]
        ref[1,1,0,0]=(0.183071638098)*x_ref[0]+(-1.37996210755)*x_ref[1]
        ref[1,1,0,1]=(0.258142115988)*x_ref[0]+(-0.757709381529)*x_ref[1]
        ref[1,1,1,0]=(1.72392829309)*x_ref[0]+(0.444995102781)*x_ref[1]
        ref[1,1,1,1]=(-0.380898209988)*x_ref[0]+(-0.0950246860347)*x_ref[1]
        ref[1,1,2,0]=(0.203007226016)*x_ref[0]+(1.09786001832)*x_ref[1]
        ref[1,1,2,1]=(-0.502500865341)*x_ref[0]+(-0.629980527995)*x_ref[1]
        ref[1,2,0,0]=(0.482549518436)*x_ref[0]+(0.565840824146)*x_ref[1]
        ref[1,2,0,1]=(0.355818565068)*x_ref[0]+(-0.78273295574)*x_ref[1]
        ref[1,2,1,0]=(1.79204033071)*x_ref[0]+(-0.698040851585)*x_ref[1]
        ref[1,2,1,1]=(-0.534827741044)*x_ref[0]+(0.00943131627498)*x_ref[1]
        ref[1,2,2,0]=(-1.00663524444)*x_ref[0]+(0.0207854553456)*x_ref[1]
        ref[1,2,2,1]=(-0.338048040159)*x_ref[0]+(-0.380607406217)*x_ref[1]
        ref[1,3,0,0]=(0.586310549044)*x_ref[0]+(0.276326798805)*x_ref[1]
        ref[1,3,0,1]=(-0.260482580702)*x_ref[0]+(-0.349279395321)*x_ref[1]
        ref[1,3,1,0]=(-0.196533829895)*x_ref[0]+(0.614805272738)*x_ref[1]
        ref[1,3,1,1]=(-0.363176483895)*x_ref[0]+(-0.925817017731)*x_ref[1]
        ref[1,3,2,0]=(-1.24182075421)*x_ref[0]+(0.215966446013)*x_ref[1]
        ref[1,3,2,1]=(1.07546652642)*x_ref[0]+(0.383803991974)*x_ref[1]
        ref[1,4,0,0]=(0.199312303525)*x_ref[0]+(-0.314520883859)*x_ref[1]
        ref[1,4,0,1]=(-1.56890826152)*x_ref[0]+(-0.133086662241)*x_ref[1]
        ref[1,4,1,0]=(1.1307241936)*x_ref[0]+(-1.32905744984)*x_ref[1]
        ref[1,4,1,1]=(-0.385991455298)*x_ref[0]+(0.292913024507)*x_ref[1]
        ref[1,4,2,0]=(1.14561678056)*x_ref[0]+(-0.662696162276)*x_ref[1]
        ref[1,4,2,1]=(-1.49533948954)*x_ref[0]+(1.43396138744)*x_ref[1]
        ref[2,0,0,0]=(-0.63160396493)*x_ref[0]+(0.813120998776)*x_ref[1]
        ref[2,0,0,1]=(0.65598998434)*x_ref[0]+(-1.02445537553)*x_ref[1]
        ref[2,0,1,0]=(-1.13473104075)*x_ref[0]+(0.057331183449)*x_ref[1]
        ref[2,0,1,1]=(-0.154892482107)*x_ref[0]+(0.481969957579)*x_ref[1]
        ref[2,0,2,0]=(0.521916671496)*x_ref[0]+(0.546676943059)*x_ref[1]
        ref[2,0,2,1]=(-1.68919379028)*x_ref[0]+(1.08292038981)*x_ref[1]
        ref[2,1,0,0]=(-0.643846595888)*x_ref[0]+(0.528306097755)*x_ref[1]
        ref[2,1,0,1]=(1.3745090497)*x_ref[0]+(-0.0801648509184)*x_ref[1]
        ref[2,1,1,0]=(0.129096817054)*x_ref[0]+(0.90750370855)*x_ref[1]
        ref[2,1,1,1]=(0.539309513305)*x_ref[0]+(1.34949470961)*x_ref[1]
        ref[2,1,2,0]=(1.4189703896)*x_ref[0]+(-1.75011201893)*x_ref[1]
        ref[2,1,2,1]=(-1.09325548386)*x_ref[0]+(0.148279301716)*x_ref[1]
        ref[2,2,0,0]=(-1.25785133743)*x_ref[0]+(0.275071306594)*x_ref[1]
        ref[2,2,0,1]=(-0.550152068761)*x_ref[0]+(-1.20422484275)*x_ref[1]
        ref[2,2,1,0]=(-0.102062863668)*x_ref[0]+(-1.34431887541)*x_ref[1]
        ref[2,2,1,1]=(0.103354724169)*x_ref[0]+(-0.941952180446)*x_ref[1]
        ref[2,2,2,0]=(1.2090339604)*x_ref[0]+(-0.330357976796)*x_ref[1]
        ref[2,2,2,1]=(0.203162449188)*x_ref[0]+(-1.03640702014)*x_ref[1]
        ref[2,3,0,0]=(-0.324576353514)*x_ref[0]+(0.582614539267)*x_ref[1]
        ref[2,3,0,1]=(-0.213823136215)*x_ref[0]+(-0.473199105214)*x_ref[1]
        ref[2,3,1,0]=(0.395144702599)*x_ref[0]+(0.0358194467152)*x_ref[1]
        ref[2,3,1,1]=(1.32860563396)*x_ref[0]+(-0.577308310732)*x_ref[1]
        ref[2,3,2,0]=(0.994205029432)*x_ref[0]+(1.13433517636)*x_ref[1]
        ref[2,3,2,1]=(-0.488499286742)*x_ref[0]+(0.224253820738)*x_ref[1]
        ref[2,4,0,0]=(1.48035663008)*x_ref[0]+(-1.09697986914)*x_ref[1]
        ref[2,4,0,1]=(-1.09991915151)*x_ref[0]+(-0.557951199985)*x_ref[1]
        ref[2,4,1,0]=(-1.23862699804)*x_ref[0]+(-0.230602187713)*x_ref[1]
        ref[2,4,1,1]=(-0.922609370448)*x_ref[0]+(1.76673953413)*x_ref[1]
        ref[2,4,2,0]=(0.731769419709)*x_ref[0]+(-1.1094516608)*x_ref[1]
        ref[2,4,2,1]=(-0.145672504805)*x_ref[0]+(0.168194421355)*x_ref[1]
        ref[3,0,0,0]=(1.02559707602)*x_ref[0]+(0.260431368523)*x_ref[1]
        ref[3,0,0,1]=(0.111899568824)*x_ref[0]+(-0.90725475423)*x_ref[1]
        ref[3,0,1,0]=(-0.226547745683)*x_ref[0]+(-0.232830393372)*x_ref[1]
        ref[3,0,1,1]=(0.347004318471)*x_ref[0]+(-0.196130970001)*x_ref[1]
        ref[3,0,2,0]=(-1.72155871442)*x_ref[0]+(-1.28372407452)*x_ref[1]
        ref[3,0,2,1]=(1.43817914693)*x_ref[0]+(-0.354093253684)*x_ref[1]
        ref[3,1,0,0]=(-0.293939245102)*x_ref[0]+(0.0255562139103)*x_ref[1]
        ref[3,1,0,1]=(0.139306213018)*x_ref[0]+(0.595062226642)*x_ref[1]
        ref[3,1,1,0]=(-0.670339920672)*x_ref[0]+(0.387520583192)*x_ref[1]
        ref[3,1,1,1]=(-0.0323325470954)*x_ref[0]+(-0.681162999279)*x_ref[1]
        ref[3,1,2,0]=(0.235275048739)*x_ref[0]+(0.144725473773)*x_ref[1]
        ref[3,1,2,1]=(0.132211795214)*x_ref[0]+(0.262780216845)*x_ref[1]
        ref[3,2,0,0]=(0.590514326163)*x_ref[0]+(-0.84273677918)*x_ref[1]
        ref[3,2,0,1]=(-0.925181556185)*x_ref[0]+(-0.300244447001)*x_ref[1]
        ref[3,2,1,0]=(-0.0509700795112)*x_ref[0]+(-1.31919389616)*x_ref[1]
        ref[3,2,1,1]=(1.50577291846)*x_ref[0]+(-0.219284607769)*x_ref[1]
        ref[3,2,2,0]=(-0.0460616673346)*x_ref[0]+(-0.690312685466)*x_ref[1]
        ref[3,2,2,1]=(-0.673765477936)*x_ref[0]+(-0.0965860475641)*x_ref[1]
        ref[3,3,0,0]=(-0.6577594174)*x_ref[0]+(0.171886277016)*x_ref[1]
        ref[3,3,0,1]=(-1.08810434718)*x_ref[0]+(0.333153250183)*x_ref[1]
        ref[3,3,1,0]=(0.150968371983)*x_ref[0]+(0.290859317391)*x_ref[1]
        ref[3,3,1,1]=(1.62096265131)*x_ref[0]+(-1.10309602741)*x_ref[1]
        ref[3,3,2,0]=(-0.932063639972)*x_ref[0]+(-0.897131561126)*x_ref[1]
        ref[3,3,2,1]=(-0.819110834345)*x_ref[0]+(1.59159109562)*x_ref[1]
        ref[3,4,0,0]=(0.272662931155)*x_ref[0]+(0.69026285756)*x_ref[1]
        ref[3,4,0,1]=(-0.600556370757)*x_ref[0]+(-0.0853533546302)*x_ref[1]
        ref[3,4,1,0]=(1.18544735281)*x_ref[0]+(0.217751189034)*x_ref[1]
        ref[3,4,1,1]=(-0.594609531358)*x_ref[0]+(0.882158349788)*x_ref[1]
        ref[3,4,2,0]=(1.18876765552)*x_ref[0]+(-0.411481586979)*x_ref[1]
        ref[3,4,2,1]=(-1.093067792)*x_ref[0]+(-0.266744260768)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.29218217256)*x[0]+(-0.382937892213)*x[1]+(0.14363896447)*x[2]
        arg[0,0,0,1]=(1.39856942421)*x[0]+(0.302082267835)*x[1]+(-0.40870736584)*x[2]
        arg[0,0,1,0]=(0.523846108043)*x[0]+(1.19961531204)*x[1]+(0.838881761349)*x[2]
        arg[0,0,1,1]=(0.634542282764)*x[0]+(-0.475514686466)*x[1]+(-1.68745428486)*x[2]
        arg[0,0,2,0]=(1.16379585731)*x[0]+(-0.928037595416)*x[1]+(0.998857178353)*x[2]
        arg[0,0,2,1]=(-0.804515607022)*x[0]+(-0.915469493439)*x[1]+(-0.143888000623)*x[2]
        arg[0,1,0,0]=(0.925843155486)*x[0]+(-0.168243418953)*x[1]+(-0.887839555187)*x[2]
        arg[0,1,0,1]=(-0.913103105329)*x[0]+(1.38011681732)*x[1]+(-0.992478793291)*x[2]
        arg[0,1,1,0]=(0.759685289903)*x[0]+(-0.579481939982)*x[1]+(0.577682929138)*x[2]
        arg[0,1,1,1]=(0.813155548228)*x[0]+(-0.178317843217)*x[1]+(-0.0460113194526)*x[2]
        arg[0,1,2,0]=(-0.195764860141)*x[0]+(1.44256979628)*x[1]+(0.283992441343)*x[2]
        arg[0,1,2,1]=(-0.921229459072)*x[0]+(-1.19295761547)*x[1]+(0.215736801944)*x[2]
        arg[0,2,0,0]=(-0.741726331752)*x[0]+(-0.225585937382)*x[1]+(1.12654556466)*x[2]
        arg[0,2,0,1]=(-0.709656162421)*x[0]+(0.298862010425)*x[1]+(-0.502443284827)*x[2]
        arg[0,2,1,0]=(-1.01166188669)*x[0]+(-1.43763408953)*x[1]+(-0.835160398488)*x[2]
        arg[0,2,1,1]=(1.80968424545)*x[0]+(-0.802666103495)*x[1]+(0.290283606115)*x[2]
        arg[0,2,2,0]=(0.346481981585)*x[0]+(-0.444176962215)*x[1]+(0.689127612227)*x[2]
        arg[0,2,2,1]=(1.06925494095)*x[0]+(0.452854988449)*x[1]+(-1.19974690364)*x[2]
        arg[0,3,0,0]=(-0.534896036328)*x[0]+(-0.312000770615)*x[1]+(-0.440033746693)*x[2]
        arg[0,3,0,1]=(0.580351431838)*x[0]+(-0.36623848451)*x[1]+(-0.507837490037)*x[2]
        arg[0,3,1,0]=(0.383180910522)*x[0]+(-0.0907152443563)*x[1]+(-0.200103743943)*x[2]
        arg[0,3,1,1]=(0.0160194687513)*x[0]+(0.862509612939)*x[1]+(-1.14643594474)*x[2]
        arg[0,3,2,0]=(-0.224390850199)*x[0]+(-0.176927715447)*x[1]+(0.213480525831)*x[2]
        arg[0,3,2,1]=(-0.0508511448148)*x[0]+(-1.28943467389)*x[1]+(0.902285302487)*x[2]
        arg[0,4,0,0]=(-0.633106625232)*x[0]+(-0.430651917485)*x[1]+(0.583149841952)*x[2]
        arg[0,4,0,1]=(0.579035528956)*x[0]+(0.159439682519)*x[1]+(-0.0926277063021)*x[2]
        arg[0,4,1,0]=(0.00484756366973)*x[0]+(0.0324822270755)*x[1]+(1.20052413378)*x[2]
        arg[0,4,1,1]=(-0.337383994688)*x[0]+(0.902076367981)*x[1]+(-0.1956377395)*x[2]
        arg[0,4,2,0]=(0.928434343456)*x[0]+(0.0376569884355)*x[1]+(-0.155274118568)*x[2]
        arg[0,4,2,1]=(-0.541990213238)*x[0]+(-0.0165076697829)*x[1]+(-0.210319224859)*x[2]
        arg[1,0,0,0]=(-1.67762025181)*x[0]+(1.61434755579)*x[1]+(0.357272049999)*x[2]
        arg[1,0,0,1]=(-1.62532544549)*x[0]+(-0.271426231027)*x[1]+(0.338680991778)*x[2]
        arg[1,0,1,0]=(1.73523408033)*x[0]+(1.11987092368)*x[1]+(0.126769364273)*x[2]
        arg[1,0,1,1]=(-1.49220786558)*x[0]+(1.51915566092)*x[1]+(1.24788192157)*x[2]
        arg[1,0,2,0]=(-0.687808766765)*x[0]+(-1.02318754737)*x[1]+(0.590653972416)*x[2]
        arg[1,0,2,1]=(0.756067354802)*x[0]+(-0.66964706965)*x[1]+(-1.18986499158)*x[2]
        arg[1,1,0,0]=(0.371345732177)*x[0]+(0.212739617115)*x[1]+(-1.13439047405)*x[2]
        arg[1,1,0,1]=(0.525761804111)*x[0]+(0.144130399691)*x[1]+(0.715517587677)*x[2]
        arg[1,1,1,0]=(-0.376160293215)*x[0]+(-0.469964935263)*x[1]+(0.413518194707)*x[2]
        arg[1,1,1,1]=(-0.787748398144)*x[0]+(-0.810090402237)*x[1]+(-0.50127109256)*x[2]
        arg[1,1,2,0]=(0.654555083692)*x[0]+(-0.142134700515)*x[1]+(-1.55891930521)*x[2]
        arg[1,1,2,1]=(-0.423058885456)*x[0]+(-0.65930158641)*x[1]+(0.94440057744)*x[2]
        arg[1,2,0,0]=(-0.924832007053)*x[0]+(1.0154890382)*x[1]+(0.11938079748)*x[2]
        arg[1,2,0,1]=(0.134642686931)*x[0]+(-1.30908067998)*x[1]+(0.367206139074)*x[2]
        arg[1,2,1,0]=(1.55524055201)*x[0]+(0.727208660637)*x[1]+(0.809810917377)*x[2]
        arg[1,2,1,1]=(0.94813391462)*x[0]+(-1.49402524308)*x[1]+(0.367824542434)*x[2]
        arg[1,2,2,0]=(-0.608295798731)*x[0]+(-0.142302252814)*x[1]+(-0.986529251308)*x[2]
        arg[1,2,2,1]=(0.0259394619607)*x[0]+(-1.60728202467)*x[1]+(0.312534656619)*x[2]
        arg[1,3,0,0]=(-1.34063373729)*x[0]+(-0.268499839329)*x[1]+(-0.335125886299)*x[2]
        arg[1,3,0,1]=(1.53280698072)*x[0]+(1.24369684711)*x[1]+(0.766369156132)*x[2]
        arg[1,3,1,0]=(0.902067176876)*x[0]+(0.791102994814)*x[1]+(-0.89266430493)*x[2]
        arg[1,3,1,1]=(0.704997629232)*x[0]+(-0.551611360332)*x[1]+(1.39134537491)*x[2]
        arg[1,3,2,0]=(-0.151690675211)*x[0]+(-1.01762975893)*x[1]+(-0.627493214284)*x[2]
        arg[1,3,2,1]=(1.00936645922)*x[0]+(0.0322458997248)*x[1]+(1.45039938194)*x[2]
        arg[1,4,0,0]=(-0.0916390388904)*x[0]+(-0.783111746469)*x[1]+(0.389725342225)*x[2]
        arg[1,4,0,1]=(1.18859328243)*x[0]+(0.871689661558)*x[1]+(-0.775464713448)*x[2]
        arg[1,4,1,0]=(0.913118342488)*x[0]+(1.53285648884)*x[1]+(0.0947365101243)*x[2]
        arg[1,4,1,1]=(-1.00735841224)*x[0]+(-1.12544193607)*x[1]+(0.267347229372)*x[2]
        arg[1,4,2,0]=(0.285943438284)*x[0]+(-1.0889389244)*x[1]+(-1.03054842327)*x[2]
        arg[1,4,2,1]=(-1.5768940139)*x[0]+(0.91237754587)*x[1]+(1.30220394199)*x[2]
        arg[2,0,0,0]=(-0.0634935883277)*x[0]+(0.157304131712)*x[1]+(-1.2995042074)*x[2]
        arg[2,0,0,1]=(-0.802705266871)*x[0]+(0.238233162774)*x[1]+(1.29463414013)*x[2]
        arg[2,0,1,0]=(-0.925468462671)*x[0]+(0.870960795966)*x[1]+(0.156522765713)*x[2]
        arg[2,0,1,1]=(-0.546666833309)*x[0]+(-1.14649607995)*x[1]+(-1.53906505048)*x[2]
        arg[2,0,2,0]=(0.464952910857)*x[0]+(0.408605786027)*x[1]+(-0.886974881114)*x[2]
        arg[2,0,2,1]=(0.491723247948)*x[0]+(0.194616420077)*x[1]+(0.409049782749)*x[2]
        arg[2,1,0,0]=(-1.19836883114)*x[0]+(1.0622722337)*x[1]+(-0.109193867727)*x[2]
        arg[2,1,0,1]=(1.01973600383)*x[0]+(-0.154482220539)*x[1]+(0.378696371956)*x[2]
        arg[2,1,1,0]=(-0.0812015296246)*x[0]+(-1.10252794568)*x[1]+(-0.00585426581857)*x[2]
        arg[2,1,1,1]=(0.536042665272)*x[0]+(0.685507117255)*x[1]+(-0.179652683139)*x[2]
        arg[2,1,2,0]=(-0.36750467811)*x[0]+(-1.23674550869)*x[1]+(-0.762905702801)*x[2]
        arg[2,1,2,1]=(0.329665965651)*x[0]+(0.101195810737)*x[1]+(-0.610015052375)*x[2]
        arg[2,2,0,0]=(-0.81644635185)*x[0]+(-0.604578476306)*x[1]+(0.0440925706991)*x[2]
        arg[2,2,0,1]=(-0.958241437971)*x[0]+(0.38563498206)*x[1]+(0.913406762318)*x[2]
        arg[2,2,1,0]=(1.22027492224)*x[0]+(-1.00726209939)*x[1]+(-1.05083068303)*x[2]
        arg[2,2,1,1]=(-1.47561192645)*x[0]+(-0.903549355608)*x[1]+(0.236771324671)*x[2]
        arg[2,2,2,0]=(0.284995864449)*x[0]+(-0.175760829921)*x[1]+(-0.720607220001)*x[2]
        arg[2,2,2,1]=(-0.535894515963)*x[0]+(-0.821302957162)*x[1]+(-1.75141540241)*x[2]
        arg[2,3,0,0]=(-0.333082106674)*x[0]+(0.579471826412)*x[1]+(0.561184138108)*x[2]
        arg[2,3,0,1]=(1.34803519557)*x[0]+(-0.441299088849)*x[1]+(-1.04945951616)*x[2]
        arg[2,3,1,0]=(0.299011096602)*x[0]+(1.32726458235)*x[1]+(-0.196298687359)*x[2]
        arg[2,3,1,1]=(1.48698834204)*x[0]+(0.527221496971)*x[1]+(0.0819813519422)*x[2]
        arg[2,3,2,0]=(0.681680955153)*x[0]+(0.503653972357)*x[1]+(0.434669347891)*x[2]
        arg[2,3,2,1]=(-0.152586789311)*x[0]+(-0.765242635579)*x[1]+(0.712663994387)*x[2]
        arg[2,4,0,0]=(0.630152300463)*x[0]+(1.12366733588)*x[1]+(-0.662966700638)*x[2]
        arg[2,4,0,1]=(1.02796709755)*x[0]+(-0.24093881584)*x[1]+(1.79445417019)*x[2]
        arg[2,4,1,0]=(-0.954575760758)*x[0]+(0.16331782608)*x[1]+(1.38309921277)*x[2]
        arg[2,4,1,1]=(0.536092514281)*x[0]+(-0.276364952673)*x[1]+(0.779619701528)*x[2]
        arg[2,4,2,0]=(0.267270444141)*x[0]+(-0.810746424331)*x[1]+(-0.733877853264)*x[2]
        arg[2,4,2,1]=(-0.0641688847923)*x[0]+(-0.386279280667)*x[1]+(0.0637571595255)*x[2]
        arg[3,0,0,0]=(0.469903180687)*x[0]+(-0.793812397342)*x[1]+(1.60841921779)*x[2]
        arg[3,0,0,1]=(0.720057600102)*x[0]+(0.502582253261)*x[1]+(0.460147791003)*x[2]
        arg[3,0,1,0]=(0.0327862564137)*x[0]+(-1.142962965)*x[1]+(0.114799131894)*x[2]
        arg[3,0,1,1]=(-0.997412656978)*x[0]+(-0.678175468576)*x[1]+(-0.152663147682)*x[2]
        arg[3,0,2,0]=(0.810424599078)*x[0]+(0.75522727786)*x[1]+(-0.205880941501)*x[2]
        arg[3,0,2,1]=(0.353601949476)*x[0]+(0.0963200813379)*x[1]+(0.0578829287146)*x[2]
        arg[3,1,0,0]=(0.346590730062)*x[0]+(1.07953790092)*x[1]+(0.647586851217)*x[2]
        arg[3,1,0,1]=(0.889879195334)*x[0]+(-0.352788186249)*x[1]+(0.434900135803)*x[2]
        arg[3,1,1,0]=(-0.721742734959)*x[0]+(0.166066672134)*x[1]+(-0.489266924809)*x[2]
        arg[3,1,1,1]=(-0.810487755181)*x[0]+(0.0354916979253)*x[1]+(0.0735520654395)*x[2]
        arg[3,1,2,0]=(1.02414338939)*x[0]+(-0.851076284556)*x[1]+(-0.232174494405)*x[2]
        arg[3,1,2,1]=(-0.327246566026)*x[0]+(-0.0499396023605)*x[1]+(0.611585402379)*x[2]
        arg[3,2,0,0]=(-0.365757628557)*x[0]+(0.868697342115)*x[1]+(-1.102545972)*x[2]
        arg[3,2,0,1]=(0.628592129227)*x[0]+(1.25545515311)*x[1]+(-1.0577476264)*x[2]
        arg[3,2,1,0]=(0.160995978044)*x[0]+(-0.947389499556)*x[1]+(-0.870476160656)*x[2]
        arg[3,2,1,1]=(0.910787139938)*x[0]+(-0.0681493345139)*x[1]+(-0.561910153781)*x[2]
        arg[3,2,2,0]=(0.0761644634428)*x[0]+(-0.700940967343)*x[1]+(1.19407210904)*x[2]
        arg[3,2,2,1]=(1.75717437065)*x[0]+(-0.144457317494)*x[1]+(-0.523233453079)*x[2]
        arg[3,3,0,0]=(0.448782872606)*x[0]+(0.650464256547)*x[1]+(1.19476547456)*x[2]
        arg[3,3,0,1]=(0.118586027201)*x[0]+(0.449971816918)*x[1]+(-0.510812733389)*x[2]
        arg[3,3,1,0]=(-0.330032265081)*x[0]+(-0.13657534891)*x[1]+(0.456365289556)*x[2]
        arg[3,3,1,1]=(0.697584212635)*x[0]+(0.897604241724)*x[1]+(1.31517116663)*x[2]
        arg[3,3,2,0]=(-0.940538707365)*x[0]+(-0.334862221665)*x[1]+(0.0969742096754)*x[2]
        arg[3,3,2,1]=(-0.985788813516)*x[0]+(0.348389725548)*x[1]+(-0.260816590361)*x[2]
        arg[3,4,0,0]=(0.754692421417)*x[0]+(0.775985781769)*x[1]+(-0.0866159222457)*x[2]
        arg[3,4,0,1]=(-1.17670502429)*x[0]+(1.12268279393)*x[1]+(-1.1279822095)*x[2]
        arg[3,4,1,0]=(-1.48674370666)*x[0]+(1.26685445713)*x[1]+(-0.0522828809306)*x[2]
        arg[3,4,1,1]=(0.401760689843)*x[0]+(-1.35707797971)*x[1]+(-0.0580654273828)*x[2]
        arg[3,4,2,0]=(-1.40697830852)*x[0]+(0.933923369958)*x[1]+(0.230152169159)*x[2]
        arg[3,4,2,1]=(-1.21242130555)*x[0]+(0.858316339663)*x[1]+(-0.839747078526)*x[2]
        ref[0,0,0,0]=(-1.29218217256)*x_ref[0]+(-0.382937892213)*x_ref[1]+(0.14363896447)*x_ref[2]
        ref[0,0,0,1]=(1.39856942421)*x_ref[0]+(0.302082267835)*x_ref[1]+(-0.40870736584)*x_ref[2]
        ref[0,0,1,0]=(0.523846108043)*x_ref[0]+(1.19961531204)*x_ref[1]+(0.838881761349)*x_ref[2]
        ref[0,0,1,1]=(0.634542282764)*x_ref[0]+(-0.475514686466)*x_ref[1]+(-1.68745428486)*x_ref[2]
        ref[0,0,2,0]=(1.16379585731)*x_ref[0]+(-0.928037595416)*x_ref[1]+(0.998857178353)*x_ref[2]
        ref[0,0,2,1]=(-0.804515607022)*x_ref[0]+(-0.915469493439)*x_ref[1]+(-0.143888000623)*x_ref[2]
        ref[0,1,0,0]=(0.925843155486)*x_ref[0]+(-0.168243418953)*x_ref[1]+(-0.887839555187)*x_ref[2]
        ref[0,1,0,1]=(-0.913103105329)*x_ref[0]+(1.38011681732)*x_ref[1]+(-0.992478793291)*x_ref[2]
        ref[0,1,1,0]=(0.759685289903)*x_ref[0]+(-0.579481939982)*x_ref[1]+(0.577682929138)*x_ref[2]
        ref[0,1,1,1]=(0.813155548228)*x_ref[0]+(-0.178317843217)*x_ref[1]+(-0.0460113194526)*x_ref[2]
        ref[0,1,2,0]=(-0.195764860141)*x_ref[0]+(1.44256979628)*x_ref[1]+(0.283992441343)*x_ref[2]
        ref[0,1,2,1]=(-0.921229459072)*x_ref[0]+(-1.19295761547)*x_ref[1]+(0.215736801944)*x_ref[2]
        ref[0,2,0,0]=(-0.741726331752)*x_ref[0]+(-0.225585937382)*x_ref[1]+(1.12654556466)*x_ref[2]
        ref[0,2,0,1]=(-0.709656162421)*x_ref[0]+(0.298862010425)*x_ref[1]+(-0.502443284827)*x_ref[2]
        ref[0,2,1,0]=(-1.01166188669)*x_ref[0]+(-1.43763408953)*x_ref[1]+(-0.835160398488)*x_ref[2]
        ref[0,2,1,1]=(1.80968424545)*x_ref[0]+(-0.802666103495)*x_ref[1]+(0.290283606115)*x_ref[2]
        ref[0,2,2,0]=(0.346481981585)*x_ref[0]+(-0.444176962215)*x_ref[1]+(0.689127612227)*x_ref[2]
        ref[0,2,2,1]=(1.06925494095)*x_ref[0]+(0.452854988449)*x_ref[1]+(-1.19974690364)*x_ref[2]
        ref[0,3,0,0]=(-0.534896036328)*x_ref[0]+(-0.312000770615)*x_ref[1]+(-0.440033746693)*x_ref[2]
        ref[0,3,0,1]=(0.580351431838)*x_ref[0]+(-0.36623848451)*x_ref[1]+(-0.507837490037)*x_ref[2]
        ref[0,3,1,0]=(0.383180910522)*x_ref[0]+(-0.0907152443563)*x_ref[1]+(-0.200103743943)*x_ref[2]
        ref[0,3,1,1]=(0.0160194687513)*x_ref[0]+(0.862509612939)*x_ref[1]+(-1.14643594474)*x_ref[2]
        ref[0,3,2,0]=(-0.224390850199)*x_ref[0]+(-0.176927715447)*x_ref[1]+(0.213480525831)*x_ref[2]
        ref[0,3,2,1]=(-0.0508511448148)*x_ref[0]+(-1.28943467389)*x_ref[1]+(0.902285302487)*x_ref[2]
        ref[0,4,0,0]=(-0.633106625232)*x_ref[0]+(-0.430651917485)*x_ref[1]+(0.583149841952)*x_ref[2]
        ref[0,4,0,1]=(0.579035528956)*x_ref[0]+(0.159439682519)*x_ref[1]+(-0.0926277063021)*x_ref[2]
        ref[0,4,1,0]=(0.00484756366973)*x_ref[0]+(0.0324822270755)*x_ref[1]+(1.20052413378)*x_ref[2]
        ref[0,4,1,1]=(-0.337383994688)*x_ref[0]+(0.902076367981)*x_ref[1]+(-0.1956377395)*x_ref[2]
        ref[0,4,2,0]=(0.928434343456)*x_ref[0]+(0.0376569884355)*x_ref[1]+(-0.155274118568)*x_ref[2]
        ref[0,4,2,1]=(-0.541990213238)*x_ref[0]+(-0.0165076697829)*x_ref[1]+(-0.210319224859)*x_ref[2]
        ref[1,0,0,0]=(-1.67762025181)*x_ref[0]+(1.61434755579)*x_ref[1]+(0.357272049999)*x_ref[2]
        ref[1,0,0,1]=(-1.62532544549)*x_ref[0]+(-0.271426231027)*x_ref[1]+(0.338680991778)*x_ref[2]
        ref[1,0,1,0]=(1.73523408033)*x_ref[0]+(1.11987092368)*x_ref[1]+(0.126769364273)*x_ref[2]
        ref[1,0,1,1]=(-1.49220786558)*x_ref[0]+(1.51915566092)*x_ref[1]+(1.24788192157)*x_ref[2]
        ref[1,0,2,0]=(-0.687808766765)*x_ref[0]+(-1.02318754737)*x_ref[1]+(0.590653972416)*x_ref[2]
        ref[1,0,2,1]=(0.756067354802)*x_ref[0]+(-0.66964706965)*x_ref[1]+(-1.18986499158)*x_ref[2]
        ref[1,1,0,0]=(0.371345732177)*x_ref[0]+(0.212739617115)*x_ref[1]+(-1.13439047405)*x_ref[2]
        ref[1,1,0,1]=(0.525761804111)*x_ref[0]+(0.144130399691)*x_ref[1]+(0.715517587677)*x_ref[2]
        ref[1,1,1,0]=(-0.376160293215)*x_ref[0]+(-0.469964935263)*x_ref[1]+(0.413518194707)*x_ref[2]
        ref[1,1,1,1]=(-0.787748398144)*x_ref[0]+(-0.810090402237)*x_ref[1]+(-0.50127109256)*x_ref[2]
        ref[1,1,2,0]=(0.654555083692)*x_ref[0]+(-0.142134700515)*x_ref[1]+(-1.55891930521)*x_ref[2]
        ref[1,1,2,1]=(-0.423058885456)*x_ref[0]+(-0.65930158641)*x_ref[1]+(0.94440057744)*x_ref[2]
        ref[1,2,0,0]=(-0.924832007053)*x_ref[0]+(1.0154890382)*x_ref[1]+(0.11938079748)*x_ref[2]
        ref[1,2,0,1]=(0.134642686931)*x_ref[0]+(-1.30908067998)*x_ref[1]+(0.367206139074)*x_ref[2]
        ref[1,2,1,0]=(1.55524055201)*x_ref[0]+(0.727208660637)*x_ref[1]+(0.809810917377)*x_ref[2]
        ref[1,2,1,1]=(0.94813391462)*x_ref[0]+(-1.49402524308)*x_ref[1]+(0.367824542434)*x_ref[2]
        ref[1,2,2,0]=(-0.608295798731)*x_ref[0]+(-0.142302252814)*x_ref[1]+(-0.986529251308)*x_ref[2]
        ref[1,2,2,1]=(0.0259394619607)*x_ref[0]+(-1.60728202467)*x_ref[1]+(0.312534656619)*x_ref[2]
        ref[1,3,0,0]=(-1.34063373729)*x_ref[0]+(-0.268499839329)*x_ref[1]+(-0.335125886299)*x_ref[2]
        ref[1,3,0,1]=(1.53280698072)*x_ref[0]+(1.24369684711)*x_ref[1]+(0.766369156132)*x_ref[2]
        ref[1,3,1,0]=(0.902067176876)*x_ref[0]+(0.791102994814)*x_ref[1]+(-0.89266430493)*x_ref[2]
        ref[1,3,1,1]=(0.704997629232)*x_ref[0]+(-0.551611360332)*x_ref[1]+(1.39134537491)*x_ref[2]
        ref[1,3,2,0]=(-0.151690675211)*x_ref[0]+(-1.01762975893)*x_ref[1]+(-0.627493214284)*x_ref[2]
        ref[1,3,2,1]=(1.00936645922)*x_ref[0]+(0.0322458997248)*x_ref[1]+(1.45039938194)*x_ref[2]
        ref[1,4,0,0]=(-0.0916390388904)*x_ref[0]+(-0.783111746469)*x_ref[1]+(0.389725342225)*x_ref[2]
        ref[1,4,0,1]=(1.18859328243)*x_ref[0]+(0.871689661558)*x_ref[1]+(-0.775464713448)*x_ref[2]
        ref[1,4,1,0]=(0.913118342488)*x_ref[0]+(1.53285648884)*x_ref[1]+(0.0947365101243)*x_ref[2]
        ref[1,4,1,1]=(-1.00735841224)*x_ref[0]+(-1.12544193607)*x_ref[1]+(0.267347229372)*x_ref[2]
        ref[1,4,2,0]=(0.285943438284)*x_ref[0]+(-1.0889389244)*x_ref[1]+(-1.03054842327)*x_ref[2]
        ref[1,4,2,1]=(-1.5768940139)*x_ref[0]+(0.91237754587)*x_ref[1]+(1.30220394199)*x_ref[2]
        ref[2,0,0,0]=(-0.0634935883277)*x_ref[0]+(0.157304131712)*x_ref[1]+(-1.2995042074)*x_ref[2]
        ref[2,0,0,1]=(-0.802705266871)*x_ref[0]+(0.238233162774)*x_ref[1]+(1.29463414013)*x_ref[2]
        ref[2,0,1,0]=(-0.925468462671)*x_ref[0]+(0.870960795966)*x_ref[1]+(0.156522765713)*x_ref[2]
        ref[2,0,1,1]=(-0.546666833309)*x_ref[0]+(-1.14649607995)*x_ref[1]+(-1.53906505048)*x_ref[2]
        ref[2,0,2,0]=(0.464952910857)*x_ref[0]+(0.408605786027)*x_ref[1]+(-0.886974881114)*x_ref[2]
        ref[2,0,2,1]=(0.491723247948)*x_ref[0]+(0.194616420077)*x_ref[1]+(0.409049782749)*x_ref[2]
        ref[2,1,0,0]=(-1.19836883114)*x_ref[0]+(1.0622722337)*x_ref[1]+(-0.109193867727)*x_ref[2]
        ref[2,1,0,1]=(1.01973600383)*x_ref[0]+(-0.154482220539)*x_ref[1]+(0.378696371956)*x_ref[2]
        ref[2,1,1,0]=(-0.0812015296246)*x_ref[0]+(-1.10252794568)*x_ref[1]+(-0.00585426581857)*x_ref[2]
        ref[2,1,1,1]=(0.536042665272)*x_ref[0]+(0.685507117255)*x_ref[1]+(-0.179652683139)*x_ref[2]
        ref[2,1,2,0]=(-0.36750467811)*x_ref[0]+(-1.23674550869)*x_ref[1]+(-0.762905702801)*x_ref[2]
        ref[2,1,2,1]=(0.329665965651)*x_ref[0]+(0.101195810737)*x_ref[1]+(-0.610015052375)*x_ref[2]
        ref[2,2,0,0]=(-0.81644635185)*x_ref[0]+(-0.604578476306)*x_ref[1]+(0.0440925706991)*x_ref[2]
        ref[2,2,0,1]=(-0.958241437971)*x_ref[0]+(0.38563498206)*x_ref[1]+(0.913406762318)*x_ref[2]
        ref[2,2,1,0]=(1.22027492224)*x_ref[0]+(-1.00726209939)*x_ref[1]+(-1.05083068303)*x_ref[2]
        ref[2,2,1,1]=(-1.47561192645)*x_ref[0]+(-0.903549355608)*x_ref[1]+(0.236771324671)*x_ref[2]
        ref[2,2,2,0]=(0.284995864449)*x_ref[0]+(-0.175760829921)*x_ref[1]+(-0.720607220001)*x_ref[2]
        ref[2,2,2,1]=(-0.535894515963)*x_ref[0]+(-0.821302957162)*x_ref[1]+(-1.75141540241)*x_ref[2]
        ref[2,3,0,0]=(-0.333082106674)*x_ref[0]+(0.579471826412)*x_ref[1]+(0.561184138108)*x_ref[2]
        ref[2,3,0,1]=(1.34803519557)*x_ref[0]+(-0.441299088849)*x_ref[1]+(-1.04945951616)*x_ref[2]
        ref[2,3,1,0]=(0.299011096602)*x_ref[0]+(1.32726458235)*x_ref[1]+(-0.196298687359)*x_ref[2]
        ref[2,3,1,1]=(1.48698834204)*x_ref[0]+(0.527221496971)*x_ref[1]+(0.0819813519422)*x_ref[2]
        ref[2,3,2,0]=(0.681680955153)*x_ref[0]+(0.503653972357)*x_ref[1]+(0.434669347891)*x_ref[2]
        ref[2,3,2,1]=(-0.152586789311)*x_ref[0]+(-0.765242635579)*x_ref[1]+(0.712663994387)*x_ref[2]
        ref[2,4,0,0]=(0.630152300463)*x_ref[0]+(1.12366733588)*x_ref[1]+(-0.662966700638)*x_ref[2]
        ref[2,4,0,1]=(1.02796709755)*x_ref[0]+(-0.24093881584)*x_ref[1]+(1.79445417019)*x_ref[2]
        ref[2,4,1,0]=(-0.954575760758)*x_ref[0]+(0.16331782608)*x_ref[1]+(1.38309921277)*x_ref[2]
        ref[2,4,1,1]=(0.536092514281)*x_ref[0]+(-0.276364952673)*x_ref[1]+(0.779619701528)*x_ref[2]
        ref[2,4,2,0]=(0.267270444141)*x_ref[0]+(-0.810746424331)*x_ref[1]+(-0.733877853264)*x_ref[2]
        ref[2,4,2,1]=(-0.0641688847923)*x_ref[0]+(-0.386279280667)*x_ref[1]+(0.0637571595255)*x_ref[2]
        ref[3,0,0,0]=(0.469903180687)*x_ref[0]+(-0.793812397342)*x_ref[1]+(1.60841921779)*x_ref[2]
        ref[3,0,0,1]=(0.720057600102)*x_ref[0]+(0.502582253261)*x_ref[1]+(0.460147791003)*x_ref[2]
        ref[3,0,1,0]=(0.0327862564137)*x_ref[0]+(-1.142962965)*x_ref[1]+(0.114799131894)*x_ref[2]
        ref[3,0,1,1]=(-0.997412656978)*x_ref[0]+(-0.678175468576)*x_ref[1]+(-0.152663147682)*x_ref[2]
        ref[3,0,2,0]=(0.810424599078)*x_ref[0]+(0.75522727786)*x_ref[1]+(-0.205880941501)*x_ref[2]
        ref[3,0,2,1]=(0.353601949476)*x_ref[0]+(0.0963200813379)*x_ref[1]+(0.0578829287146)*x_ref[2]
        ref[3,1,0,0]=(0.346590730062)*x_ref[0]+(1.07953790092)*x_ref[1]+(0.647586851217)*x_ref[2]
        ref[3,1,0,1]=(0.889879195334)*x_ref[0]+(-0.352788186249)*x_ref[1]+(0.434900135803)*x_ref[2]
        ref[3,1,1,0]=(-0.721742734959)*x_ref[0]+(0.166066672134)*x_ref[1]+(-0.489266924809)*x_ref[2]
        ref[3,1,1,1]=(-0.810487755181)*x_ref[0]+(0.0354916979253)*x_ref[1]+(0.0735520654395)*x_ref[2]
        ref[3,1,2,0]=(1.02414338939)*x_ref[0]+(-0.851076284556)*x_ref[1]+(-0.232174494405)*x_ref[2]
        ref[3,1,2,1]=(-0.327246566026)*x_ref[0]+(-0.0499396023605)*x_ref[1]+(0.611585402379)*x_ref[2]
        ref[3,2,0,0]=(-0.365757628557)*x_ref[0]+(0.868697342115)*x_ref[1]+(-1.102545972)*x_ref[2]
        ref[3,2,0,1]=(0.628592129227)*x_ref[0]+(1.25545515311)*x_ref[1]+(-1.0577476264)*x_ref[2]
        ref[3,2,1,0]=(0.160995978044)*x_ref[0]+(-0.947389499556)*x_ref[1]+(-0.870476160656)*x_ref[2]
        ref[3,2,1,1]=(0.910787139938)*x_ref[0]+(-0.0681493345139)*x_ref[1]+(-0.561910153781)*x_ref[2]
        ref[3,2,2,0]=(0.0761644634428)*x_ref[0]+(-0.700940967343)*x_ref[1]+(1.19407210904)*x_ref[2]
        ref[3,2,2,1]=(1.75717437065)*x_ref[0]+(-0.144457317494)*x_ref[1]+(-0.523233453079)*x_ref[2]
        ref[3,3,0,0]=(0.448782872606)*x_ref[0]+(0.650464256547)*x_ref[1]+(1.19476547456)*x_ref[2]
        ref[3,3,0,1]=(0.118586027201)*x_ref[0]+(0.449971816918)*x_ref[1]+(-0.510812733389)*x_ref[2]
        ref[3,3,1,0]=(-0.330032265081)*x_ref[0]+(-0.13657534891)*x_ref[1]+(0.456365289556)*x_ref[2]
        ref[3,3,1,1]=(0.697584212635)*x_ref[0]+(0.897604241724)*x_ref[1]+(1.31517116663)*x_ref[2]
        ref[3,3,2,0]=(-0.940538707365)*x_ref[0]+(-0.334862221665)*x_ref[1]+(0.0969742096754)*x_ref[2]
        ref[3,3,2,1]=(-0.985788813516)*x_ref[0]+(0.348389725548)*x_ref[1]+(-0.260816590361)*x_ref[2]
        ref[3,4,0,0]=(0.754692421417)*x_ref[0]+(0.775985781769)*x_ref[1]+(-0.0866159222457)*x_ref[2]
        ref[3,4,0,1]=(-1.17670502429)*x_ref[0]+(1.12268279393)*x_ref[1]+(-1.1279822095)*x_ref[2]
        ref[3,4,1,0]=(-1.48674370666)*x_ref[0]+(1.26685445713)*x_ref[1]+(-0.0522828809306)*x_ref[2]
        ref[3,4,1,1]=(0.401760689843)*x_ref[0]+(-1.35707797971)*x_ref[1]+(-0.0580654273828)*x_ref[2]
        ref[3,4,2,0]=(-1.40697830852)*x_ref[0]+(0.933923369958)*x_ref[1]+(0.230152169159)*x_ref[2]
        ref[3,4,2,1]=(-1.21242130555)*x_ref[0]+(0.858316339663)*x_ref[1]+(-0.839747078526)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.100067148208)*x[0]+(-0.8375743725)*x[1]
        ref=(-0.100067148208)*x_ref[0]+(-0.8375743725)*x_ref[1]
      else:
        arg=(-0.247343416726)*x[0]+(0.604193484164)*x[1]+(-0.219899438287)*x[2]
        ref=(-0.247343416726)*x_ref[0]+(0.604193484164)*x_ref[1]+(-0.219899438287)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        arg[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
        ref[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        ref[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
      else:
        arg[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        arg[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
        ref[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        ref[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.00153698933322)*x[0]+(0.392281645764)*x[1]
        arg[0,1]=(-1.2794655544)*x[0]+(-0.373519639818)*x[1]
        arg[0,2]=(0.30127033321)*x[0]+(-0.0996865853738)*x[1]
        arg[0,3]=(1.23405799349)*x[0]+(0.498558200339)*x[1]
        arg[0,4]=(0.85519350742)*x[0]+(-0.19188046135)*x[1]
        arg[1,0]=(0.417367834962)*x[0]+(0.992318201639)*x[1]
        arg[1,1]=(-0.646432128833)*x[0]+(-0.872750321582)*x[1]
        arg[1,2]=(-0.806117185061)*x[0]+(-0.89396408325)*x[1]
        arg[1,3]=(0.768017237627)*x[0]+(-1.2904111912)*x[1]
        arg[1,4]=(-0.0159685523188)*x[0]+(0.401410193169)*x[1]
        arg[2,0]=(-0.844501716181)*x[0]+(-0.123115343819)*x[1]
        arg[2,1]=(1.08441909409)*x[0]+(0.152432282539)*x[1]
        arg[2,2]=(-0.60551037413)*x[0]+(1.15692828826)*x[1]
        arg[2,3]=(-1.21509168541)*x[0]+(1.03683236377)*x[1]
        arg[2,4]=(1.16850730134)*x[0]+(0.162202615909)*x[1]
        arg[3,0]=(0.727772166897)*x[0]+(-0.296815840935)*x[1]
        arg[3,1]=(-0.0880577151489)*x[0]+(1.66427691426)*x[1]
        arg[3,2]=(0.415144534458)*x[0]+(-0.555833723295)*x[1]
        arg[3,3]=(1.09444635829)*x[0]+(-0.335209417536)*x[1]
        arg[3,4]=(-1.05047769585)*x[0]+(0.684427039986)*x[1]
        ref[0,0]=(0.00153698933322)*x_ref[0]+(0.392281645764)*x_ref[1]
        ref[0,1]=(-1.2794655544)*x_ref[0]+(-0.373519639818)*x_ref[1]
        ref[0,2]=(0.30127033321)*x_ref[0]+(-0.0996865853738)*x_ref[1]
        ref[0,3]=(1.23405799349)*x_ref[0]+(0.498558200339)*x_ref[1]
        ref[0,4]=(0.85519350742)*x_ref[0]+(-0.19188046135)*x_ref[1]
        ref[1,0]=(0.417367834962)*x_ref[0]+(0.992318201639)*x_ref[1]
        ref[1,1]=(-0.646432128833)*x_ref[0]+(-0.872750321582)*x_ref[1]
        ref[1,2]=(-0.806117185061)*x_ref[0]+(-0.89396408325)*x_ref[1]
        ref[1,3]=(0.768017237627)*x_ref[0]+(-1.2904111912)*x_ref[1]
        ref[1,4]=(-0.0159685523188)*x_ref[0]+(0.401410193169)*x_ref[1]
        ref[2,0]=(-0.844501716181)*x_ref[0]+(-0.123115343819)*x_ref[1]
        ref[2,1]=(1.08441909409)*x_ref[0]+(0.152432282539)*x_ref[1]
        ref[2,2]=(-0.60551037413)*x_ref[0]+(1.15692828826)*x_ref[1]
        ref[2,3]=(-1.21509168541)*x_ref[0]+(1.03683236377)*x_ref[1]
        ref[2,4]=(1.16850730134)*x_ref[0]+(0.162202615909)*x_ref[1]
        ref[3,0]=(0.727772166897)*x_ref[0]+(-0.296815840935)*x_ref[1]
        ref[3,1]=(-0.0880577151489)*x_ref[0]+(1.66427691426)*x_ref[1]
        ref[3,2]=(0.415144534458)*x_ref[0]+(-0.555833723295)*x_ref[1]
        ref[3,3]=(1.09444635829)*x_ref[0]+(-0.335209417536)*x_ref[1]
        ref[3,4]=(-1.05047769585)*x_ref[0]+(0.684427039986)*x_ref[1]
      else:
        arg[0,0]=(-0.611607640773)*x[0]+(0.584485971628)*x[1]+(0.0090316337643)*x[2]
        arg[0,1]=(0.905278285857)*x[0]+(0.332523151242)*x[1]+(0.9909841807)*x[2]
        arg[0,2]=(-1.06629333132)*x[0]+(-0.38478180064)*x[1]+(-0.63309211256)*x[2]
        arg[0,3]=(0.122436386344)*x[0]+(-0.218752671204)*x[1]+(-0.746695172216)*x[2]
        arg[0,4]=(1.03152124874)*x[0]+(-1.11327271563)*x[1]+(-0.257591112135)*x[2]
        arg[1,0]=(0.969019600299)*x[0]+(-0.320279309689)*x[1]+(0.327767746545)*x[2]
        arg[1,1]=(0.291023069052)*x[0]+(0.78112901064)*x[1]+(0.879257239042)*x[2]
        arg[1,2]=(-0.131262891536)*x[0]+(-0.0109690229153)*x[1]+(-0.171900332406)*x[2]
        arg[1,3]=(-1.32272684486)*x[0]+(-0.256869536085)*x[1]+(0.0308524567462)*x[2]
        arg[1,4]=(1.30707061279)*x[0]+(0.972866282634)*x[1]+(-0.474948537882)*x[2]
        arg[2,0]=(0.169034210421)*x[0]+(-1.08714896226)*x[1]+(-0.636838450708)*x[2]
        arg[2,1]=(-0.470752780773)*x[0]+(-0.695286088059)*x[1]+(-1.19197058712)*x[2]
        arg[2,2]=(0.701403721228)*x[0]+(0.551002482436)*x[1]+(-0.295853851234)*x[2]
        arg[2,3]=(-0.796431361745)*x[0]+(0.0581884766856)*x[1]+(-1.13668563089)*x[2]
        arg[2,4]=(-0.733185798017)*x[0]+(0.178900103735)*x[1]+(-0.845671526781)*x[2]
        arg[3,0]=(-1.47064927107)*x[0]+(1.33393324965)*x[1]+(-0.767393460937)*x[2]
        arg[3,1]=(1.36784513226)*x[0]+(0.100008910431)*x[1]+(-1.09210256713)*x[2]
        arg[3,2]=(-0.873076669977)*x[0]+(0.367213631778)*x[1]+(0.0698296920039)*x[2]
        arg[3,3]=(-1.76609214158)*x[0]+(-1.26186286566)*x[1]+(0.377817035113)*x[2]
        arg[3,4]=(-1.28261709396)*x[0]+(0.240480767533)*x[1]+(0.307813279357)*x[2]
        ref[0,0]=(-0.611607640773)*x_ref[0]+(0.584485971628)*x_ref[1]+(0.0090316337643)*x_ref[2]
        ref[0,1]=(0.905278285857)*x_ref[0]+(0.332523151242)*x_ref[1]+(0.9909841807)*x_ref[2]
        ref[0,2]=(-1.06629333132)*x_ref[0]+(-0.38478180064)*x_ref[1]+(-0.63309211256)*x_ref[2]
        ref[0,3]=(0.122436386344)*x_ref[0]+(-0.218752671204)*x_ref[1]+(-0.746695172216)*x_ref[2]
        ref[0,4]=(1.03152124874)*x_ref[0]+(-1.11327271563)*x_ref[1]+(-0.257591112135)*x_ref[2]
        ref[1,0]=(0.969019600299)*x_ref[0]+(-0.320279309689)*x_ref[1]+(0.327767746545)*x_ref[2]
        ref[1,1]=(0.291023069052)*x_ref[0]+(0.78112901064)*x_ref[1]+(0.879257239042)*x_ref[2]
        ref[1,2]=(-0.131262891536)*x_ref[0]+(-0.0109690229153)*x_ref[1]+(-0.171900332406)*x_ref[2]
        ref[1,3]=(-1.32272684486)*x_ref[0]+(-0.256869536085)*x_ref[1]+(0.0308524567462)*x_ref[2]
        ref[1,4]=(1.30707061279)*x_ref[0]+(0.972866282634)*x_ref[1]+(-0.474948537882)*x_ref[2]
        ref[2,0]=(0.169034210421)*x_ref[0]+(-1.08714896226)*x_ref[1]+(-0.636838450708)*x_ref[2]
        ref[2,1]=(-0.470752780773)*x_ref[0]+(-0.695286088059)*x_ref[1]+(-1.19197058712)*x_ref[2]
        ref[2,2]=(0.701403721228)*x_ref[0]+(0.551002482436)*x_ref[1]+(-0.295853851234)*x_ref[2]
        ref[2,3]=(-0.796431361745)*x_ref[0]+(0.0581884766856)*x_ref[1]+(-1.13668563089)*x_ref[2]
        ref[2,4]=(-0.733185798017)*x_ref[0]+(0.178900103735)*x_ref[1]+(-0.845671526781)*x_ref[2]
        ref[3,0]=(-1.47064927107)*x_ref[0]+(1.33393324965)*x_ref[1]+(-0.767393460937)*x_ref[2]
        ref[3,1]=(1.36784513226)*x_ref[0]+(0.100008910431)*x_ref[1]+(-1.09210256713)*x_ref[2]
        ref[3,2]=(-0.873076669977)*x_ref[0]+(0.367213631778)*x_ref[1]+(0.0698296920039)*x_ref[2]
        ref[3,3]=(-1.76609214158)*x_ref[0]+(-1.26186286566)*x_ref[1]+(0.377817035113)*x_ref[2]
        ref[3,4]=(-1.28261709396)*x_ref[0]+(0.240480767533)*x_ref[1]+(0.307813279357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.423505169404)*x[0]+(-1.10390334264)*x[1]
        arg[0,0,1]=(-0.181820089254)*x[0]+(-0.374602268791)*x[1]
        arg[0,1,0]=(-1.36906917075)*x[0]+(0.26554385908)*x[1]
        arg[0,1,1]=(-0.280661572374)*x[0]+(-0.239232524287)*x[1]
        arg[1,0,0]=(-1.25670251934)*x[0]+(-0.429353977735)*x[1]
        arg[1,0,1]=(-0.747928970735)*x[0]+(1.17874317611)*x[1]
        arg[1,1,0]=(-0.548814611802)*x[0]+(-0.0249315764285)*x[1]
        arg[1,1,1]=(-0.481123117522)*x[0]+(1.23402203879)*x[1]
        arg[2,0,0]=(-0.0639512689884)*x[0]+(-0.0278213371376)*x[1]
        arg[2,0,1]=(-1.17292797621)*x[0]+(0.42140600304)*x[1]
        arg[2,1,0]=(-0.358052601083)*x[0]+(-1.66807880791)*x[1]
        arg[2,1,1]=(-0.291433251349)*x[0]+(0.111473162108)*x[1]
        arg[3,0,0]=(0.742643625989)*x[0]+(-1.29214106228)*x[1]
        arg[3,0,1]=(0.319887321826)*x[0]+(-0.40213369238)*x[1]
        arg[3,1,0]=(-0.449412064097)*x[0]+(-1.10654262009)*x[1]
        arg[3,1,1]=(1.19722227632)*x[0]+(1.3455799026)*x[1]
        arg[4,0,0]=(-1.77453188487)*x[0]+(0.634838793872)*x[1]
        arg[4,0,1]=(-0.215461066907)*x[0]+(1.11635233844)*x[1]
        arg[4,1,0]=(0.835816652592)*x[0]+(0.668378554214)*x[1]
        arg[4,1,1]=(-1.22133225406)*x[0]+(-0.148780252398)*x[1]
        arg[5,0,0]=(0.320245409639)*x[0]+(-1.08828513776)*x[1]
        arg[5,0,1]=(-0.1865835417)*x[0]+(-0.353282061874)*x[1]
        arg[5,1,0]=(1.38858040685)*x[0]+(0.166398261098)*x[1]
        arg[5,1,1]=(0.947718875755)*x[0]+(-1.06171934302)*x[1]
        ref[0,0,0]=(-0.423505169404)*x_ref[0]+(-1.10390334264)*x_ref[1]
        ref[0,0,1]=(-0.181820089254)*x_ref[0]+(-0.374602268791)*x_ref[1]
        ref[0,1,0]=(-1.36906917075)*x_ref[0]+(0.26554385908)*x_ref[1]
        ref[0,1,1]=(-0.280661572374)*x_ref[0]+(-0.239232524287)*x_ref[1]
        ref[1,0,0]=(-1.25670251934)*x_ref[0]+(-0.429353977735)*x_ref[1]
        ref[1,0,1]=(-0.747928970735)*x_ref[0]+(1.17874317611)*x_ref[1]
        ref[1,1,0]=(-0.548814611802)*x_ref[0]+(-0.0249315764285)*x_ref[1]
        ref[1,1,1]=(-0.481123117522)*x_ref[0]+(1.23402203879)*x_ref[1]
        ref[2,0,0]=(-0.0639512689884)*x_ref[0]+(-0.0278213371376)*x_ref[1]
        ref[2,0,1]=(-1.17292797621)*x_ref[0]+(0.42140600304)*x_ref[1]
        ref[2,1,0]=(-0.358052601083)*x_ref[0]+(-1.66807880791)*x_ref[1]
        ref[2,1,1]=(-0.291433251349)*x_ref[0]+(0.111473162108)*x_ref[1]
        ref[3,0,0]=(0.742643625989)*x_ref[0]+(-1.29214106228)*x_ref[1]
        ref[3,0,1]=(0.319887321826)*x_ref[0]+(-0.40213369238)*x_ref[1]
        ref[3,1,0]=(-0.449412064097)*x_ref[0]+(-1.10654262009)*x_ref[1]
        ref[3,1,1]=(1.19722227632)*x_ref[0]+(1.3455799026)*x_ref[1]
        ref[4,0,0]=(-1.77453188487)*x_ref[0]+(0.634838793872)*x_ref[1]
        ref[4,0,1]=(-0.215461066907)*x_ref[0]+(1.11635233844)*x_ref[1]
        ref[4,1,0]=(0.835816652592)*x_ref[0]+(0.668378554214)*x_ref[1]
        ref[4,1,1]=(-1.22133225406)*x_ref[0]+(-0.148780252398)*x_ref[1]
        ref[5,0,0]=(0.320245409639)*x_ref[0]+(-1.08828513776)*x_ref[1]
        ref[5,0,1]=(-0.1865835417)*x_ref[0]+(-0.353282061874)*x_ref[1]
        ref[5,1,0]=(1.38858040685)*x_ref[0]+(0.166398261098)*x_ref[1]
        ref[5,1,1]=(0.947718875755)*x_ref[0]+(-1.06171934302)*x_ref[1]
      else:
        arg[0,0,0]=(-0.892096322478)*x[0]+(0.720363183451)*x[1]+(0.600609272191)*x[2]
        arg[0,0,1]=(-0.418024461935)*x[0]+(-0.819275292035)*x[1]+(0.134962557692)*x[2]
        arg[0,1,0]=(-0.449783838361)*x[0]+(0.943677173533)*x[1]+(0.209424190991)*x[2]
        arg[0,1,1]=(-0.424386775004)*x[0]+(-0.213024388094)*x[1]+(-1.05927943384)*x[2]
        arg[1,0,0]=(-0.699504096539)*x[0]+(-0.153570308377)*x[1]+(0.576902392658)*x[2]
        arg[1,0,1]=(-0.426914055661)*x[0]+(-0.330405267059)*x[1]+(1.07656632155)*x[2]
        arg[1,1,0]=(0.775238005352)*x[0]+(1.11012072575)*x[1]+(0.764627098603)*x[2]
        arg[1,1,1]=(0.0130424708408)*x[0]+(-0.671778102107)*x[1]+(0.780684368983)*x[2]
        arg[2,0,0]=(-0.0512224258498)*x[0]+(-1.50515803984)*x[1]+(0.342824143946)*x[2]
        arg[2,0,1]=(0.117541624176)*x[0]+(0.372415524552)*x[1]+(0.255218095016)*x[2]
        arg[2,1,0]=(-0.574508895703)*x[0]+(0.345217292359)*x[1]+(0.209211956821)*x[2]
        arg[2,1,1]=(0.93643587147)*x[0]+(-0.096209009902)*x[1]+(0.205150859527)*x[2]
        arg[3,0,0]=(1.7541929884)*x[0]+(0.169061478447)*x[1]+(0.534746695515)*x[2]
        arg[3,0,1]=(0.442329266782)*x[0]+(0.334971710414)*x[1]+(1.66892557698)*x[2]
        arg[3,1,0]=(0.316436716758)*x[0]+(-1.25442030987)*x[1]+(0.372183105374)*x[2]
        arg[3,1,1]=(-0.254157820265)*x[0]+(0.287381300023)*x[1]+(0.0574117423352)*x[2]
        arg[4,0,0]=(-1.08380257521)*x[0]+(-1.31461890032)*x[1]+(-0.503802775099)*x[2]
        arg[4,0,1]=(-0.558293486798)*x[0]+(0.702317635925)*x[1]+(0.463814829137)*x[2]
        arg[4,1,0]=(-0.677160971974)*x[0]+(-0.14921275312)*x[1]+(-1.6501459435)*x[2]
        arg[4,1,1]=(-0.176672532474)*x[0]+(0.945017605811)*x[1]+(-0.790342939788)*x[2]
        arg[5,0,0]=(-1.45668070675)*x[0]+(-0.404957129204)*x[1]+(1.37383065954)*x[2]
        arg[5,0,1]=(0.53829326756)*x[0]+(0.543333031479)*x[1]+(-0.854057296362)*x[2]
        arg[5,1,0]=(0.805864283921)*x[0]+(-0.00186634953513)*x[1]+(-1.97111362539)*x[2]
        arg[5,1,1]=(-0.387746238967)*x[0]+(-0.739158353882)*x[1]+(-0.134082808118)*x[2]
        ref[0,0,0]=(-0.892096322478)*x_ref[0]+(0.720363183451)*x_ref[1]+(0.600609272191)*x_ref[2]
        ref[0,0,1]=(-0.418024461935)*x_ref[0]+(-0.819275292035)*x_ref[1]+(0.134962557692)*x_ref[2]
        ref[0,1,0]=(-0.449783838361)*x_ref[0]+(0.943677173533)*x_ref[1]+(0.209424190991)*x_ref[2]
        ref[0,1,1]=(-0.424386775004)*x_ref[0]+(-0.213024388094)*x_ref[1]+(-1.05927943384)*x_ref[2]
        ref[1,0,0]=(-0.699504096539)*x_ref[0]+(-0.153570308377)*x_ref[1]+(0.576902392658)*x_ref[2]
        ref[1,0,1]=(-0.426914055661)*x_ref[0]+(-0.330405267059)*x_ref[1]+(1.07656632155)*x_ref[2]
        ref[1,1,0]=(0.775238005352)*x_ref[0]+(1.11012072575)*x_ref[1]+(0.764627098603)*x_ref[2]
        ref[1,1,1]=(0.0130424708408)*x_ref[0]+(-0.671778102107)*x_ref[1]+(0.780684368983)*x_ref[2]
        ref[2,0,0]=(-0.0512224258498)*x_ref[0]+(-1.50515803984)*x_ref[1]+(0.342824143946)*x_ref[2]
        ref[2,0,1]=(0.117541624176)*x_ref[0]+(0.372415524552)*x_ref[1]+(0.255218095016)*x_ref[2]
        ref[2,1,0]=(-0.574508895703)*x_ref[0]+(0.345217292359)*x_ref[1]+(0.209211956821)*x_ref[2]
        ref[2,1,1]=(0.93643587147)*x_ref[0]+(-0.096209009902)*x_ref[1]+(0.205150859527)*x_ref[2]
        ref[3,0,0]=(1.7541929884)*x_ref[0]+(0.169061478447)*x_ref[1]+(0.534746695515)*x_ref[2]
        ref[3,0,1]=(0.442329266782)*x_ref[0]+(0.334971710414)*x_ref[1]+(1.66892557698)*x_ref[2]
        ref[3,1,0]=(0.316436716758)*x_ref[0]+(-1.25442030987)*x_ref[1]+(0.372183105374)*x_ref[2]
        ref[3,1,1]=(-0.254157820265)*x_ref[0]+(0.287381300023)*x_ref[1]+(0.0574117423352)*x_ref[2]
        ref[4,0,0]=(-1.08380257521)*x_ref[0]+(-1.31461890032)*x_ref[1]+(-0.503802775099)*x_ref[2]
        ref[4,0,1]=(-0.558293486798)*x_ref[0]+(0.702317635925)*x_ref[1]+(0.463814829137)*x_ref[2]
        ref[4,1,0]=(-0.677160971974)*x_ref[0]+(-0.14921275312)*x_ref[1]+(-1.6501459435)*x_ref[2]
        ref[4,1,1]=(-0.176672532474)*x_ref[0]+(0.945017605811)*x_ref[1]+(-0.790342939788)*x_ref[2]
        ref[5,0,0]=(-1.45668070675)*x_ref[0]+(-0.404957129204)*x_ref[1]+(1.37383065954)*x_ref[2]
        ref[5,0,1]=(0.53829326756)*x_ref[0]+(0.543333031479)*x_ref[1]+(-0.854057296362)*x_ref[2]
        ref[5,1,0]=(0.805864283921)*x_ref[0]+(-0.00186634953513)*x_ref[1]+(-1.97111362539)*x_ref[2]
        ref[5,1,1]=(-0.387746238967)*x_ref[0]+(-0.739158353882)*x_ref[1]+(-0.134082808118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.754311658086)*x[0]+(0.235157034342)*x[1]
        arg[0,0,0,1]=(-1.33786315024)*x[0]+(0.243349279742)*x[1]
        arg[0,0,1,0]=(1.48657750088)*x[0]+(1.09853108305)*x[1]
        arg[0,0,1,1]=(1.4501362266)*x[0]+(0.824811567155)*x[1]
        arg[0,0,2,0]=(1.12567566082)*x[0]+(0.663525681174)*x[1]
        arg[0,0,2,1]=(1.48991565019)*x[0]+(-0.680795365011)*x[1]
        arg[0,1,0,0]=(0.268571449528)*x[0]+(0.818595834672)*x[1]
        arg[0,1,0,1]=(-1.55490945555)*x[0]+(0.783595197888)*x[1]
        arg[0,1,1,0]=(0.913201382585)*x[0]+(0.855491379376)*x[1]
        arg[0,1,1,1]=(-0.179192075116)*x[0]+(-0.690415079523)*x[1]
        arg[0,1,2,0]=(-0.0393574502912)*x[0]+(-0.690157650634)*x[1]
        arg[0,1,2,1]=(0.784216863217)*x[0]+(0.498735486159)*x[1]
        arg[0,2,0,0]=(0.137939554165)*x[0]+(-1.33961519113)*x[1]
        arg[0,2,0,1]=(-1.03897311736)*x[0]+(-1.37797405312)*x[1]
        arg[0,2,1,0]=(0.199812003756)*x[0]+(-0.0388499674037)*x[1]
        arg[0,2,1,1]=(1.15530534202)*x[0]+(-1.28280869078)*x[1]
        arg[0,2,2,0]=(0.836447791155)*x[0]+(1.2627066026)*x[1]
        arg[0,2,2,1]=(-0.584608492688)*x[0]+(-1.55305974305)*x[1]
        arg[0,3,0,0]=(-0.661952381799)*x[0]+(-1.27062403778)*x[1]
        arg[0,3,0,1]=(-1.47789532259)*x[0]+(-0.279583524936)*x[1]
        arg[0,3,1,0]=(1.40729771357)*x[0]+(0.0971284144528)*x[1]
        arg[0,3,1,1]=(-0.459167955911)*x[0]+(-1.29033281524)*x[1]
        arg[0,3,2,0]=(-1.63189436824)*x[0]+(0.75729808422)*x[1]
        arg[0,3,2,1]=(0.522479828718)*x[0]+(-0.0231299515383)*x[1]
        arg[0,4,0,0]=(0.241891446615)*x[0]+(0.803327107651)*x[1]
        arg[0,4,0,1]=(0.273613502836)*x[0]+(-0.0115180464171)*x[1]
        arg[0,4,1,0]=(1.96219818696)*x[0]+(-0.349137094743)*x[1]
        arg[0,4,1,1]=(0.365690821719)*x[0]+(-1.6943347181)*x[1]
        arg[0,4,2,0]=(-0.283970779872)*x[0]+(-0.258537179804)*x[1]
        arg[0,4,2,1]=(-0.973333430598)*x[0]+(1.48706276212)*x[1]
        arg[1,0,0,0]=(1.18757466971)*x[0]+(0.838804656357)*x[1]
        arg[1,0,0,1]=(0.840859904702)*x[0]+(1.09955195682)*x[1]
        arg[1,0,1,0]=(1.46190513227)*x[0]+(-0.596959830422)*x[1]
        arg[1,0,1,1]=(0.718689296745)*x[0]+(-0.628074620222)*x[1]
        arg[1,0,2,0]=(0.440761293041)*x[0]+(0.115800854194)*x[1]
        arg[1,0,2,1]=(1.34891146904)*x[0]+(0.813690260652)*x[1]
        arg[1,1,0,0]=(0.183071638098)*x[0]+(-1.37996210755)*x[1]
        arg[1,1,0,1]=(0.258142115988)*x[0]+(-0.757709381529)*x[1]
        arg[1,1,1,0]=(1.72392829309)*x[0]+(0.444995102781)*x[1]
        arg[1,1,1,1]=(-0.380898209988)*x[0]+(-0.0950246860347)*x[1]
        arg[1,1,2,0]=(0.203007226016)*x[0]+(1.09786001832)*x[1]
        arg[1,1,2,1]=(-0.502500865341)*x[0]+(-0.629980527995)*x[1]
        arg[1,2,0,0]=(0.482549518436)*x[0]+(0.565840824146)*x[1]
        arg[1,2,0,1]=(0.355818565068)*x[0]+(-0.78273295574)*x[1]
        arg[1,2,1,0]=(1.79204033071)*x[0]+(-0.698040851585)*x[1]
        arg[1,2,1,1]=(-0.534827741044)*x[0]+(0.00943131627498)*x[1]
        arg[1,2,2,0]=(-1.00663524444)*x[0]+(0.0207854553456)*x[1]
        arg[1,2,2,1]=(-0.338048040159)*x[0]+(-0.380607406217)*x[1]
        arg[1,3,0,0]=(0.586310549044)*x[0]+(0.276326798805)*x[1]
        arg[1,3,0,1]=(-0.260482580702)*x[0]+(-0.349279395321)*x[1]
        arg[1,3,1,0]=(-0.196533829895)*x[0]+(0.614805272738)*x[1]
        arg[1,3,1,1]=(-0.363176483895)*x[0]+(-0.925817017731)*x[1]
        arg[1,3,2,0]=(-1.24182075421)*x[0]+(0.215966446013)*x[1]
        arg[1,3,2,1]=(1.07546652642)*x[0]+(0.383803991974)*x[1]
        arg[1,4,0,0]=(0.199312303525)*x[0]+(-0.314520883859)*x[1]
        arg[1,4,0,1]=(-1.56890826152)*x[0]+(-0.133086662241)*x[1]
        arg[1,4,1,0]=(1.1307241936)*x[0]+(-1.32905744984)*x[1]
        arg[1,4,1,1]=(-0.385991455298)*x[0]+(0.292913024507)*x[1]
        arg[1,4,2,0]=(1.14561678056)*x[0]+(-0.662696162276)*x[1]
        arg[1,4,2,1]=(-1.49533948954)*x[0]+(1.43396138744)*x[1]
        arg[2,0,0,0]=(-0.63160396493)*x[0]+(0.813120998776)*x[1]
        arg[2,0,0,1]=(0.65598998434)*x[0]+(-1.02445537553)*x[1]
        arg[2,0,1,0]=(-1.13473104075)*x[0]+(0.057331183449)*x[1]
        arg[2,0,1,1]=(-0.154892482107)*x[0]+(0.481969957579)*x[1]
        arg[2,0,2,0]=(0.521916671496)*x[0]+(0.546676943059)*x[1]
        arg[2,0,2,1]=(-1.68919379028)*x[0]+(1.08292038981)*x[1]
        arg[2,1,0,0]=(-0.643846595888)*x[0]+(0.528306097755)*x[1]
        arg[2,1,0,1]=(1.3745090497)*x[0]+(-0.0801648509184)*x[1]
        arg[2,1,1,0]=(0.129096817054)*x[0]+(0.90750370855)*x[1]
        arg[2,1,1,1]=(0.539309513305)*x[0]+(1.34949470961)*x[1]
        arg[2,1,2,0]=(1.4189703896)*x[0]+(-1.75011201893)*x[1]
        arg[2,1,2,1]=(-1.09325548386)*x[0]+(0.148279301716)*x[1]
        arg[2,2,0,0]=(-1.25785133743)*x[0]+(0.275071306594)*x[1]
        arg[2,2,0,1]=(-0.550152068761)*x[0]+(-1.20422484275)*x[1]
        arg[2,2,1,0]=(-0.102062863668)*x[0]+(-1.34431887541)*x[1]
        arg[2,2,1,1]=(0.103354724169)*x[0]+(-0.941952180446)*x[1]
        arg[2,2,2,0]=(1.2090339604)*x[0]+(-0.330357976796)*x[1]
        arg[2,2,2,1]=(0.203162449188)*x[0]+(-1.03640702014)*x[1]
        arg[2,3,0,0]=(-0.324576353514)*x[0]+(0.582614539267)*x[1]
        arg[2,3,0,1]=(-0.213823136215)*x[0]+(-0.473199105214)*x[1]
        arg[2,3,1,0]=(0.395144702599)*x[0]+(0.0358194467152)*x[1]
        arg[2,3,1,1]=(1.32860563396)*x[0]+(-0.577308310732)*x[1]
        arg[2,3,2,0]=(0.994205029432)*x[0]+(1.13433517636)*x[1]
        arg[2,3,2,1]=(-0.488499286742)*x[0]+(0.224253820738)*x[1]
        arg[2,4,0,0]=(1.48035663008)*x[0]+(-1.09697986914)*x[1]
        arg[2,4,0,1]=(-1.09991915151)*x[0]+(-0.557951199985)*x[1]
        arg[2,4,1,0]=(-1.23862699804)*x[0]+(-0.230602187713)*x[1]
        arg[2,4,1,1]=(-0.922609370448)*x[0]+(1.76673953413)*x[1]
        arg[2,4,2,0]=(0.731769419709)*x[0]+(-1.1094516608)*x[1]
        arg[2,4,2,1]=(-0.145672504805)*x[0]+(0.168194421355)*x[1]
        arg[3,0,0,0]=(1.02559707602)*x[0]+(0.260431368523)*x[1]
        arg[3,0,0,1]=(0.111899568824)*x[0]+(-0.90725475423)*x[1]
        arg[3,0,1,0]=(-0.226547745683)*x[0]+(-0.232830393372)*x[1]
        arg[3,0,1,1]=(0.347004318471)*x[0]+(-0.196130970001)*x[1]
        arg[3,0,2,0]=(-1.72155871442)*x[0]+(-1.28372407452)*x[1]
        arg[3,0,2,1]=(1.43817914693)*x[0]+(-0.354093253684)*x[1]
        arg[3,1,0,0]=(-0.293939245102)*x[0]+(0.0255562139103)*x[1]
        arg[3,1,0,1]=(0.139306213018)*x[0]+(0.595062226642)*x[1]
        arg[3,1,1,0]=(-0.670339920672)*x[0]+(0.387520583192)*x[1]
        arg[3,1,1,1]=(-0.0323325470954)*x[0]+(-0.681162999279)*x[1]
        arg[3,1,2,0]=(0.235275048739)*x[0]+(0.144725473773)*x[1]
        arg[3,1,2,1]=(0.132211795214)*x[0]+(0.262780216845)*x[1]
        arg[3,2,0,0]=(0.590514326163)*x[0]+(-0.84273677918)*x[1]
        arg[3,2,0,1]=(-0.925181556185)*x[0]+(-0.300244447001)*x[1]
        arg[3,2,1,0]=(-0.0509700795112)*x[0]+(-1.31919389616)*x[1]
        arg[3,2,1,1]=(1.50577291846)*x[0]+(-0.219284607769)*x[1]
        arg[3,2,2,0]=(-0.0460616673346)*x[0]+(-0.690312685466)*x[1]
        arg[3,2,2,1]=(-0.673765477936)*x[0]+(-0.0965860475641)*x[1]
        arg[3,3,0,0]=(-0.6577594174)*x[0]+(0.171886277016)*x[1]
        arg[3,3,0,1]=(-1.08810434718)*x[0]+(0.333153250183)*x[1]
        arg[3,3,1,0]=(0.150968371983)*x[0]+(0.290859317391)*x[1]
        arg[3,3,1,1]=(1.62096265131)*x[0]+(-1.10309602741)*x[1]
        arg[3,3,2,0]=(-0.932063639972)*x[0]+(-0.897131561126)*x[1]
        arg[3,3,2,1]=(-0.819110834345)*x[0]+(1.59159109562)*x[1]
        arg[3,4,0,0]=(0.272662931155)*x[0]+(0.69026285756)*x[1]
        arg[3,4,0,1]=(-0.600556370757)*x[0]+(-0.0853533546302)*x[1]
        arg[3,4,1,0]=(1.18544735281)*x[0]+(0.217751189034)*x[1]
        arg[3,4,1,1]=(-0.594609531358)*x[0]+(0.882158349788)*x[1]
        arg[3,4,2,0]=(1.18876765552)*x[0]+(-0.411481586979)*x[1]
        arg[3,4,2,1]=(-1.093067792)*x[0]+(-0.266744260768)*x[1]
        ref[0,0,0,0]=(-0.754311658086)*x_ref[0]+(0.235157034342)*x_ref[1]
        ref[0,0,0,1]=(-1.33786315024)*x_ref[0]+(0.243349279742)*x_ref[1]
        ref[0,0,1,0]=(1.48657750088)*x_ref[0]+(1.09853108305)*x_ref[1]
        ref[0,0,1,1]=(1.4501362266)*x_ref[0]+(0.824811567155)*x_ref[1]
        ref[0,0,2,0]=(1.12567566082)*x_ref[0]+(0.663525681174)*x_ref[1]
        ref[0,0,2,1]=(1.48991565019)*x_ref[0]+(-0.680795365011)*x_ref[1]
        ref[0,1,0,0]=(0.268571449528)*x_ref[0]+(0.818595834672)*x_ref[1]
        ref[0,1,0,1]=(-1.55490945555)*x_ref[0]+(0.783595197888)*x_ref[1]
        ref[0,1,1,0]=(0.913201382585)*x_ref[0]+(0.855491379376)*x_ref[1]
        ref[0,1,1,1]=(-0.179192075116)*x_ref[0]+(-0.690415079523)*x_ref[1]
        ref[0,1,2,0]=(-0.0393574502912)*x_ref[0]+(-0.690157650634)*x_ref[1]
        ref[0,1,2,1]=(0.784216863217)*x_ref[0]+(0.498735486159)*x_ref[1]
        ref[0,2,0,0]=(0.137939554165)*x_ref[0]+(-1.33961519113)*x_ref[1]
        ref[0,2,0,1]=(-1.03897311736)*x_ref[0]+(-1.37797405312)*x_ref[1]
        ref[0,2,1,0]=(0.199812003756)*x_ref[0]+(-0.0388499674037)*x_ref[1]
        ref[0,2,1,1]=(1.15530534202)*x_ref[0]+(-1.28280869078)*x_ref[1]
        ref[0,2,2,0]=(0.836447791155)*x_ref[0]+(1.2627066026)*x_ref[1]
        ref[0,2,2,1]=(-0.584608492688)*x_ref[0]+(-1.55305974305)*x_ref[1]
        ref[0,3,0,0]=(-0.661952381799)*x_ref[0]+(-1.27062403778)*x_ref[1]
        ref[0,3,0,1]=(-1.47789532259)*x_ref[0]+(-0.279583524936)*x_ref[1]
        ref[0,3,1,0]=(1.40729771357)*x_ref[0]+(0.0971284144528)*x_ref[1]
        ref[0,3,1,1]=(-0.459167955911)*x_ref[0]+(-1.29033281524)*x_ref[1]
        ref[0,3,2,0]=(-1.63189436824)*x_ref[0]+(0.75729808422)*x_ref[1]
        ref[0,3,2,1]=(0.522479828718)*x_ref[0]+(-0.0231299515383)*x_ref[1]
        ref[0,4,0,0]=(0.241891446615)*x_ref[0]+(0.803327107651)*x_ref[1]
        ref[0,4,0,1]=(0.273613502836)*x_ref[0]+(-0.0115180464171)*x_ref[1]
        ref[0,4,1,0]=(1.96219818696)*x_ref[0]+(-0.349137094743)*x_ref[1]
        ref[0,4,1,1]=(0.365690821719)*x_ref[0]+(-1.6943347181)*x_ref[1]
        ref[0,4,2,0]=(-0.283970779872)*x_ref[0]+(-0.258537179804)*x_ref[1]
        ref[0,4,2,1]=(-0.973333430598)*x_ref[0]+(1.48706276212)*x_ref[1]
        ref[1,0,0,0]=(1.18757466971)*x_ref[0]+(0.838804656357)*x_ref[1]
        ref[1,0,0,1]=(0.840859904702)*x_ref[0]+(1.09955195682)*x_ref[1]
        ref[1,0,1,0]=(1.46190513227)*x_ref[0]+(-0.596959830422)*x_ref[1]
        ref[1,0,1,1]=(0.718689296745)*x_ref[0]+(-0.628074620222)*x_ref[1]
        ref[1,0,2,0]=(0.440761293041)*x_ref[0]+(0.115800854194)*x_ref[1]
        ref[1,0,2,1]=(1.34891146904)*x_ref[0]+(0.813690260652)*x_ref[1]
        ref[1,1,0,0]=(0.183071638098)*x_ref[0]+(-1.37996210755)*x_ref[1]
        ref[1,1,0,1]=(0.258142115988)*x_ref[0]+(-0.757709381529)*x_ref[1]
        ref[1,1,1,0]=(1.72392829309)*x_ref[0]+(0.444995102781)*x_ref[1]
        ref[1,1,1,1]=(-0.380898209988)*x_ref[0]+(-0.0950246860347)*x_ref[1]
        ref[1,1,2,0]=(0.203007226016)*x_ref[0]+(1.09786001832)*x_ref[1]
        ref[1,1,2,1]=(-0.502500865341)*x_ref[0]+(-0.629980527995)*x_ref[1]
        ref[1,2,0,0]=(0.482549518436)*x_ref[0]+(0.565840824146)*x_ref[1]
        ref[1,2,0,1]=(0.355818565068)*x_ref[0]+(-0.78273295574)*x_ref[1]
        ref[1,2,1,0]=(1.79204033071)*x_ref[0]+(-0.698040851585)*x_ref[1]
        ref[1,2,1,1]=(-0.534827741044)*x_ref[0]+(0.00943131627498)*x_ref[1]
        ref[1,2,2,0]=(-1.00663524444)*x_ref[0]+(0.0207854553456)*x_ref[1]
        ref[1,2,2,1]=(-0.338048040159)*x_ref[0]+(-0.380607406217)*x_ref[1]
        ref[1,3,0,0]=(0.586310549044)*x_ref[0]+(0.276326798805)*x_ref[1]
        ref[1,3,0,1]=(-0.260482580702)*x_ref[0]+(-0.349279395321)*x_ref[1]
        ref[1,3,1,0]=(-0.196533829895)*x_ref[0]+(0.614805272738)*x_ref[1]
        ref[1,3,1,1]=(-0.363176483895)*x_ref[0]+(-0.925817017731)*x_ref[1]
        ref[1,3,2,0]=(-1.24182075421)*x_ref[0]+(0.215966446013)*x_ref[1]
        ref[1,3,2,1]=(1.07546652642)*x_ref[0]+(0.383803991974)*x_ref[1]
        ref[1,4,0,0]=(0.199312303525)*x_ref[0]+(-0.314520883859)*x_ref[1]
        ref[1,4,0,1]=(-1.56890826152)*x_ref[0]+(-0.133086662241)*x_ref[1]
        ref[1,4,1,0]=(1.1307241936)*x_ref[0]+(-1.32905744984)*x_ref[1]
        ref[1,4,1,1]=(-0.385991455298)*x_ref[0]+(0.292913024507)*x_ref[1]
        ref[1,4,2,0]=(1.14561678056)*x_ref[0]+(-0.662696162276)*x_ref[1]
        ref[1,4,2,1]=(-1.49533948954)*x_ref[0]+(1.43396138744)*x_ref[1]
        ref[2,0,0,0]=(-0.63160396493)*x_ref[0]+(0.813120998776)*x_ref[1]
        ref[2,0,0,1]=(0.65598998434)*x_ref[0]+(-1.02445537553)*x_ref[1]
        ref[2,0,1,0]=(-1.13473104075)*x_ref[0]+(0.057331183449)*x_ref[1]
        ref[2,0,1,1]=(-0.154892482107)*x_ref[0]+(0.481969957579)*x_ref[1]
        ref[2,0,2,0]=(0.521916671496)*x_ref[0]+(0.546676943059)*x_ref[1]
        ref[2,0,2,1]=(-1.68919379028)*x_ref[0]+(1.08292038981)*x_ref[1]
        ref[2,1,0,0]=(-0.643846595888)*x_ref[0]+(0.528306097755)*x_ref[1]
        ref[2,1,0,1]=(1.3745090497)*x_ref[0]+(-0.0801648509184)*x_ref[1]
        ref[2,1,1,0]=(0.129096817054)*x_ref[0]+(0.90750370855)*x_ref[1]
        ref[2,1,1,1]=(0.539309513305)*x_ref[0]+(1.34949470961)*x_ref[1]
        ref[2,1,2,0]=(1.4189703896)*x_ref[0]+(-1.75011201893)*x_ref[1]
        ref[2,1,2,1]=(-1.09325548386)*x_ref[0]+(0.148279301716)*x_ref[1]
        ref[2,2,0,0]=(-1.25785133743)*x_ref[0]+(0.275071306594)*x_ref[1]
        ref[2,2,0,1]=(-0.550152068761)*x_ref[0]+(-1.20422484275)*x_ref[1]
        ref[2,2,1,0]=(-0.102062863668)*x_ref[0]+(-1.34431887541)*x_ref[1]
        ref[2,2,1,1]=(0.103354724169)*x_ref[0]+(-0.941952180446)*x_ref[1]
        ref[2,2,2,0]=(1.2090339604)*x_ref[0]+(-0.330357976796)*x_ref[1]
        ref[2,2,2,1]=(0.203162449188)*x_ref[0]+(-1.03640702014)*x_ref[1]
        ref[2,3,0,0]=(-0.324576353514)*x_ref[0]+(0.582614539267)*x_ref[1]
        ref[2,3,0,1]=(-0.213823136215)*x_ref[0]+(-0.473199105214)*x_ref[1]
        ref[2,3,1,0]=(0.395144702599)*x_ref[0]+(0.0358194467152)*x_ref[1]
        ref[2,3,1,1]=(1.32860563396)*x_ref[0]+(-0.577308310732)*x_ref[1]
        ref[2,3,2,0]=(0.994205029432)*x_ref[0]+(1.13433517636)*x_ref[1]
        ref[2,3,2,1]=(-0.488499286742)*x_ref[0]+(0.224253820738)*x_ref[1]
        ref[2,4,0,0]=(1.48035663008)*x_ref[0]+(-1.09697986914)*x_ref[1]
        ref[2,4,0,1]=(-1.09991915151)*x_ref[0]+(-0.557951199985)*x_ref[1]
        ref[2,4,1,0]=(-1.23862699804)*x_ref[0]+(-0.230602187713)*x_ref[1]
        ref[2,4,1,1]=(-0.922609370448)*x_ref[0]+(1.76673953413)*x_ref[1]
        ref[2,4,2,0]=(0.731769419709)*x_ref[0]+(-1.1094516608)*x_ref[1]
        ref[2,4,2,1]=(-0.145672504805)*x_ref[0]+(0.168194421355)*x_ref[1]
        ref[3,0,0,0]=(1.02559707602)*x_ref[0]+(0.260431368523)*x_ref[1]
        ref[3,0,0,1]=(0.111899568824)*x_ref[0]+(-0.90725475423)*x_ref[1]
        ref[3,0,1,0]=(-0.226547745683)*x_ref[0]+(-0.232830393372)*x_ref[1]
        ref[3,0,1,1]=(0.347004318471)*x_ref[0]+(-0.196130970001)*x_ref[1]
        ref[3,0,2,0]=(-1.72155871442)*x_ref[0]+(-1.28372407452)*x_ref[1]
        ref[3,0,2,1]=(1.43817914693)*x_ref[0]+(-0.354093253684)*x_ref[1]
        ref[3,1,0,0]=(-0.293939245102)*x_ref[0]+(0.0255562139103)*x_ref[1]
        ref[3,1,0,1]=(0.139306213018)*x_ref[0]+(0.595062226642)*x_ref[1]
        ref[3,1,1,0]=(-0.670339920672)*x_ref[0]+(0.387520583192)*x_ref[1]
        ref[3,1,1,1]=(-0.0323325470954)*x_ref[0]+(-0.681162999279)*x_ref[1]
        ref[3,1,2,0]=(0.235275048739)*x_ref[0]+(0.144725473773)*x_ref[1]
        ref[3,1,2,1]=(0.132211795214)*x_ref[0]+(0.262780216845)*x_ref[1]
        ref[3,2,0,0]=(0.590514326163)*x_ref[0]+(-0.84273677918)*x_ref[1]
        ref[3,2,0,1]=(-0.925181556185)*x_ref[0]+(-0.300244447001)*x_ref[1]
        ref[3,2,1,0]=(-0.0509700795112)*x_ref[0]+(-1.31919389616)*x_ref[1]
        ref[3,2,1,1]=(1.50577291846)*x_ref[0]+(-0.219284607769)*x_ref[1]
        ref[3,2,2,0]=(-0.0460616673346)*x_ref[0]+(-0.690312685466)*x_ref[1]
        ref[3,2,2,1]=(-0.673765477936)*x_ref[0]+(-0.0965860475641)*x_ref[1]
        ref[3,3,0,0]=(-0.6577594174)*x_ref[0]+(0.171886277016)*x_ref[1]
        ref[3,3,0,1]=(-1.08810434718)*x_ref[0]+(0.333153250183)*x_ref[1]
        ref[3,3,1,0]=(0.150968371983)*x_ref[0]+(0.290859317391)*x_ref[1]
        ref[3,3,1,1]=(1.62096265131)*x_ref[0]+(-1.10309602741)*x_ref[1]
        ref[3,3,2,0]=(-0.932063639972)*x_ref[0]+(-0.897131561126)*x_ref[1]
        ref[3,3,2,1]=(-0.819110834345)*x_ref[0]+(1.59159109562)*x_ref[1]
        ref[3,4,0,0]=(0.272662931155)*x_ref[0]+(0.69026285756)*x_ref[1]
        ref[3,4,0,1]=(-0.600556370757)*x_ref[0]+(-0.0853533546302)*x_ref[1]
        ref[3,4,1,0]=(1.18544735281)*x_ref[0]+(0.217751189034)*x_ref[1]
        ref[3,4,1,1]=(-0.594609531358)*x_ref[0]+(0.882158349788)*x_ref[1]
        ref[3,4,2,0]=(1.18876765552)*x_ref[0]+(-0.411481586979)*x_ref[1]
        ref[3,4,2,1]=(-1.093067792)*x_ref[0]+(-0.266744260768)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.29218217256)*x[0]+(-0.382937892213)*x[1]+(0.14363896447)*x[2]
        arg[0,0,0,1]=(1.39856942421)*x[0]+(0.302082267835)*x[1]+(-0.40870736584)*x[2]
        arg[0,0,1,0]=(0.523846108043)*x[0]+(1.19961531204)*x[1]+(0.838881761349)*x[2]
        arg[0,0,1,1]=(0.634542282764)*x[0]+(-0.475514686466)*x[1]+(-1.68745428486)*x[2]
        arg[0,0,2,0]=(1.16379585731)*x[0]+(-0.928037595416)*x[1]+(0.998857178353)*x[2]
        arg[0,0,2,1]=(-0.804515607022)*x[0]+(-0.915469493439)*x[1]+(-0.143888000623)*x[2]
        arg[0,1,0,0]=(0.925843155486)*x[0]+(-0.168243418953)*x[1]+(-0.887839555187)*x[2]
        arg[0,1,0,1]=(-0.913103105329)*x[0]+(1.38011681732)*x[1]+(-0.992478793291)*x[2]
        arg[0,1,1,0]=(0.759685289903)*x[0]+(-0.579481939982)*x[1]+(0.577682929138)*x[2]
        arg[0,1,1,1]=(0.813155548228)*x[0]+(-0.178317843217)*x[1]+(-0.0460113194526)*x[2]
        arg[0,1,2,0]=(-0.195764860141)*x[0]+(1.44256979628)*x[1]+(0.283992441343)*x[2]
        arg[0,1,2,1]=(-0.921229459072)*x[0]+(-1.19295761547)*x[1]+(0.215736801944)*x[2]
        arg[0,2,0,0]=(-0.741726331752)*x[0]+(-0.225585937382)*x[1]+(1.12654556466)*x[2]
        arg[0,2,0,1]=(-0.709656162421)*x[0]+(0.298862010425)*x[1]+(-0.502443284827)*x[2]
        arg[0,2,1,0]=(-1.01166188669)*x[0]+(-1.43763408953)*x[1]+(-0.835160398488)*x[2]
        arg[0,2,1,1]=(1.80968424545)*x[0]+(-0.802666103495)*x[1]+(0.290283606115)*x[2]
        arg[0,2,2,0]=(0.346481981585)*x[0]+(-0.444176962215)*x[1]+(0.689127612227)*x[2]
        arg[0,2,2,1]=(1.06925494095)*x[0]+(0.452854988449)*x[1]+(-1.19974690364)*x[2]
        arg[0,3,0,0]=(-0.534896036328)*x[0]+(-0.312000770615)*x[1]+(-0.440033746693)*x[2]
        arg[0,3,0,1]=(0.580351431838)*x[0]+(-0.36623848451)*x[1]+(-0.507837490037)*x[2]
        arg[0,3,1,0]=(0.383180910522)*x[0]+(-0.0907152443563)*x[1]+(-0.200103743943)*x[2]
        arg[0,3,1,1]=(0.0160194687513)*x[0]+(0.862509612939)*x[1]+(-1.14643594474)*x[2]
        arg[0,3,2,0]=(-0.224390850199)*x[0]+(-0.176927715447)*x[1]+(0.213480525831)*x[2]
        arg[0,3,2,1]=(-0.0508511448148)*x[0]+(-1.28943467389)*x[1]+(0.902285302487)*x[2]
        arg[0,4,0,0]=(-0.633106625232)*x[0]+(-0.430651917485)*x[1]+(0.583149841952)*x[2]
        arg[0,4,0,1]=(0.579035528956)*x[0]+(0.159439682519)*x[1]+(-0.0926277063021)*x[2]
        arg[0,4,1,0]=(0.00484756366973)*x[0]+(0.0324822270755)*x[1]+(1.20052413378)*x[2]
        arg[0,4,1,1]=(-0.337383994688)*x[0]+(0.902076367981)*x[1]+(-0.1956377395)*x[2]
        arg[0,4,2,0]=(0.928434343456)*x[0]+(0.0376569884355)*x[1]+(-0.155274118568)*x[2]
        arg[0,4,2,1]=(-0.541990213238)*x[0]+(-0.0165076697829)*x[1]+(-0.210319224859)*x[2]
        arg[1,0,0,0]=(-1.67762025181)*x[0]+(1.61434755579)*x[1]+(0.357272049999)*x[2]
        arg[1,0,0,1]=(-1.62532544549)*x[0]+(-0.271426231027)*x[1]+(0.338680991778)*x[2]
        arg[1,0,1,0]=(1.73523408033)*x[0]+(1.11987092368)*x[1]+(0.126769364273)*x[2]
        arg[1,0,1,1]=(-1.49220786558)*x[0]+(1.51915566092)*x[1]+(1.24788192157)*x[2]
        arg[1,0,2,0]=(-0.687808766765)*x[0]+(-1.02318754737)*x[1]+(0.590653972416)*x[2]
        arg[1,0,2,1]=(0.756067354802)*x[0]+(-0.66964706965)*x[1]+(-1.18986499158)*x[2]
        arg[1,1,0,0]=(0.371345732177)*x[0]+(0.212739617115)*x[1]+(-1.13439047405)*x[2]
        arg[1,1,0,1]=(0.525761804111)*x[0]+(0.144130399691)*x[1]+(0.715517587677)*x[2]
        arg[1,1,1,0]=(-0.376160293215)*x[0]+(-0.469964935263)*x[1]+(0.413518194707)*x[2]
        arg[1,1,1,1]=(-0.787748398144)*x[0]+(-0.810090402237)*x[1]+(-0.50127109256)*x[2]
        arg[1,1,2,0]=(0.654555083692)*x[0]+(-0.142134700515)*x[1]+(-1.55891930521)*x[2]
        arg[1,1,2,1]=(-0.423058885456)*x[0]+(-0.65930158641)*x[1]+(0.94440057744)*x[2]
        arg[1,2,0,0]=(-0.924832007053)*x[0]+(1.0154890382)*x[1]+(0.11938079748)*x[2]
        arg[1,2,0,1]=(0.134642686931)*x[0]+(-1.30908067998)*x[1]+(0.367206139074)*x[2]
        arg[1,2,1,0]=(1.55524055201)*x[0]+(0.727208660637)*x[1]+(0.809810917377)*x[2]
        arg[1,2,1,1]=(0.94813391462)*x[0]+(-1.49402524308)*x[1]+(0.367824542434)*x[2]
        arg[1,2,2,0]=(-0.608295798731)*x[0]+(-0.142302252814)*x[1]+(-0.986529251308)*x[2]
        arg[1,2,2,1]=(0.0259394619607)*x[0]+(-1.60728202467)*x[1]+(0.312534656619)*x[2]
        arg[1,3,0,0]=(-1.34063373729)*x[0]+(-0.268499839329)*x[1]+(-0.335125886299)*x[2]
        arg[1,3,0,1]=(1.53280698072)*x[0]+(1.24369684711)*x[1]+(0.766369156132)*x[2]
        arg[1,3,1,0]=(0.902067176876)*x[0]+(0.791102994814)*x[1]+(-0.89266430493)*x[2]
        arg[1,3,1,1]=(0.704997629232)*x[0]+(-0.551611360332)*x[1]+(1.39134537491)*x[2]
        arg[1,3,2,0]=(-0.151690675211)*x[0]+(-1.01762975893)*x[1]+(-0.627493214284)*x[2]
        arg[1,3,2,1]=(1.00936645922)*x[0]+(0.0322458997248)*x[1]+(1.45039938194)*x[2]
        arg[1,4,0,0]=(-0.0916390388904)*x[0]+(-0.783111746469)*x[1]+(0.389725342225)*x[2]
        arg[1,4,0,1]=(1.18859328243)*x[0]+(0.871689661558)*x[1]+(-0.775464713448)*x[2]
        arg[1,4,1,0]=(0.913118342488)*x[0]+(1.53285648884)*x[1]+(0.0947365101243)*x[2]
        arg[1,4,1,1]=(-1.00735841224)*x[0]+(-1.12544193607)*x[1]+(0.267347229372)*x[2]
        arg[1,4,2,0]=(0.285943438284)*x[0]+(-1.0889389244)*x[1]+(-1.03054842327)*x[2]
        arg[1,4,2,1]=(-1.5768940139)*x[0]+(0.91237754587)*x[1]+(1.30220394199)*x[2]
        arg[2,0,0,0]=(-0.0634935883277)*x[0]+(0.157304131712)*x[1]+(-1.2995042074)*x[2]
        arg[2,0,0,1]=(-0.802705266871)*x[0]+(0.238233162774)*x[1]+(1.29463414013)*x[2]
        arg[2,0,1,0]=(-0.925468462671)*x[0]+(0.870960795966)*x[1]+(0.156522765713)*x[2]
        arg[2,0,1,1]=(-0.546666833309)*x[0]+(-1.14649607995)*x[1]+(-1.53906505048)*x[2]
        arg[2,0,2,0]=(0.464952910857)*x[0]+(0.408605786027)*x[1]+(-0.886974881114)*x[2]
        arg[2,0,2,1]=(0.491723247948)*x[0]+(0.194616420077)*x[1]+(0.409049782749)*x[2]
        arg[2,1,0,0]=(-1.19836883114)*x[0]+(1.0622722337)*x[1]+(-0.109193867727)*x[2]
        arg[2,1,0,1]=(1.01973600383)*x[0]+(-0.154482220539)*x[1]+(0.378696371956)*x[2]
        arg[2,1,1,0]=(-0.0812015296246)*x[0]+(-1.10252794568)*x[1]+(-0.00585426581857)*x[2]
        arg[2,1,1,1]=(0.536042665272)*x[0]+(0.685507117255)*x[1]+(-0.179652683139)*x[2]
        arg[2,1,2,0]=(-0.36750467811)*x[0]+(-1.23674550869)*x[1]+(-0.762905702801)*x[2]
        arg[2,1,2,1]=(0.329665965651)*x[0]+(0.101195810737)*x[1]+(-0.610015052375)*x[2]
        arg[2,2,0,0]=(-0.81644635185)*x[0]+(-0.604578476306)*x[1]+(0.0440925706991)*x[2]
        arg[2,2,0,1]=(-0.958241437971)*x[0]+(0.38563498206)*x[1]+(0.913406762318)*x[2]
        arg[2,2,1,0]=(1.22027492224)*x[0]+(-1.00726209939)*x[1]+(-1.05083068303)*x[2]
        arg[2,2,1,1]=(-1.47561192645)*x[0]+(-0.903549355608)*x[1]+(0.236771324671)*x[2]
        arg[2,2,2,0]=(0.284995864449)*x[0]+(-0.175760829921)*x[1]+(-0.720607220001)*x[2]
        arg[2,2,2,1]=(-0.535894515963)*x[0]+(-0.821302957162)*x[1]+(-1.75141540241)*x[2]
        arg[2,3,0,0]=(-0.333082106674)*x[0]+(0.579471826412)*x[1]+(0.561184138108)*x[2]
        arg[2,3,0,1]=(1.34803519557)*x[0]+(-0.441299088849)*x[1]+(-1.04945951616)*x[2]
        arg[2,3,1,0]=(0.299011096602)*x[0]+(1.32726458235)*x[1]+(-0.196298687359)*x[2]
        arg[2,3,1,1]=(1.48698834204)*x[0]+(0.527221496971)*x[1]+(0.0819813519422)*x[2]
        arg[2,3,2,0]=(0.681680955153)*x[0]+(0.503653972357)*x[1]+(0.434669347891)*x[2]
        arg[2,3,2,1]=(-0.152586789311)*x[0]+(-0.765242635579)*x[1]+(0.712663994387)*x[2]
        arg[2,4,0,0]=(0.630152300463)*x[0]+(1.12366733588)*x[1]+(-0.662966700638)*x[2]
        arg[2,4,0,1]=(1.02796709755)*x[0]+(-0.24093881584)*x[1]+(1.79445417019)*x[2]
        arg[2,4,1,0]=(-0.954575760758)*x[0]+(0.16331782608)*x[1]+(1.38309921277)*x[2]
        arg[2,4,1,1]=(0.536092514281)*x[0]+(-0.276364952673)*x[1]+(0.779619701528)*x[2]
        arg[2,4,2,0]=(0.267270444141)*x[0]+(-0.810746424331)*x[1]+(-0.733877853264)*x[2]
        arg[2,4,2,1]=(-0.0641688847923)*x[0]+(-0.386279280667)*x[1]+(0.0637571595255)*x[2]
        arg[3,0,0,0]=(0.469903180687)*x[0]+(-0.793812397342)*x[1]+(1.60841921779)*x[2]
        arg[3,0,0,1]=(0.720057600102)*x[0]+(0.502582253261)*x[1]+(0.460147791003)*x[2]
        arg[3,0,1,0]=(0.0327862564137)*x[0]+(-1.142962965)*x[1]+(0.114799131894)*x[2]
        arg[3,0,1,1]=(-0.997412656978)*x[0]+(-0.678175468576)*x[1]+(-0.152663147682)*x[2]
        arg[3,0,2,0]=(0.810424599078)*x[0]+(0.75522727786)*x[1]+(-0.205880941501)*x[2]
        arg[3,0,2,1]=(0.353601949476)*x[0]+(0.0963200813379)*x[1]+(0.0578829287146)*x[2]
        arg[3,1,0,0]=(0.346590730062)*x[0]+(1.07953790092)*x[1]+(0.647586851217)*x[2]
        arg[3,1,0,1]=(0.889879195334)*x[0]+(-0.352788186249)*x[1]+(0.434900135803)*x[2]
        arg[3,1,1,0]=(-0.721742734959)*x[0]+(0.166066672134)*x[1]+(-0.489266924809)*x[2]
        arg[3,1,1,1]=(-0.810487755181)*x[0]+(0.0354916979253)*x[1]+(0.0735520654395)*x[2]
        arg[3,1,2,0]=(1.02414338939)*x[0]+(-0.851076284556)*x[1]+(-0.232174494405)*x[2]
        arg[3,1,2,1]=(-0.327246566026)*x[0]+(-0.0499396023605)*x[1]+(0.611585402379)*x[2]
        arg[3,2,0,0]=(-0.365757628557)*x[0]+(0.868697342115)*x[1]+(-1.102545972)*x[2]
        arg[3,2,0,1]=(0.628592129227)*x[0]+(1.25545515311)*x[1]+(-1.0577476264)*x[2]
        arg[3,2,1,0]=(0.160995978044)*x[0]+(-0.947389499556)*x[1]+(-0.870476160656)*x[2]
        arg[3,2,1,1]=(0.910787139938)*x[0]+(-0.0681493345139)*x[1]+(-0.561910153781)*x[2]
        arg[3,2,2,0]=(0.0761644634428)*x[0]+(-0.700940967343)*x[1]+(1.19407210904)*x[2]
        arg[3,2,2,1]=(1.75717437065)*x[0]+(-0.144457317494)*x[1]+(-0.523233453079)*x[2]
        arg[3,3,0,0]=(0.448782872606)*x[0]+(0.650464256547)*x[1]+(1.19476547456)*x[2]
        arg[3,3,0,1]=(0.118586027201)*x[0]+(0.449971816918)*x[1]+(-0.510812733389)*x[2]
        arg[3,3,1,0]=(-0.330032265081)*x[0]+(-0.13657534891)*x[1]+(0.456365289556)*x[2]
        arg[3,3,1,1]=(0.697584212635)*x[0]+(0.897604241724)*x[1]+(1.31517116663)*x[2]
        arg[3,3,2,0]=(-0.940538707365)*x[0]+(-0.334862221665)*x[1]+(0.0969742096754)*x[2]
        arg[3,3,2,1]=(-0.985788813516)*x[0]+(0.348389725548)*x[1]+(-0.260816590361)*x[2]
        arg[3,4,0,0]=(0.754692421417)*x[0]+(0.775985781769)*x[1]+(-0.0866159222457)*x[2]
        arg[3,4,0,1]=(-1.17670502429)*x[0]+(1.12268279393)*x[1]+(-1.1279822095)*x[2]
        arg[3,4,1,0]=(-1.48674370666)*x[0]+(1.26685445713)*x[1]+(-0.0522828809306)*x[2]
        arg[3,4,1,1]=(0.401760689843)*x[0]+(-1.35707797971)*x[1]+(-0.0580654273828)*x[2]
        arg[3,4,2,0]=(-1.40697830852)*x[0]+(0.933923369958)*x[1]+(0.230152169159)*x[2]
        arg[3,4,2,1]=(-1.21242130555)*x[0]+(0.858316339663)*x[1]+(-0.839747078526)*x[2]
        ref[0,0,0,0]=(-1.29218217256)*x_ref[0]+(-0.382937892213)*x_ref[1]+(0.14363896447)*x_ref[2]
        ref[0,0,0,1]=(1.39856942421)*x_ref[0]+(0.302082267835)*x_ref[1]+(-0.40870736584)*x_ref[2]
        ref[0,0,1,0]=(0.523846108043)*x_ref[0]+(1.19961531204)*x_ref[1]+(0.838881761349)*x_ref[2]
        ref[0,0,1,1]=(0.634542282764)*x_ref[0]+(-0.475514686466)*x_ref[1]+(-1.68745428486)*x_ref[2]
        ref[0,0,2,0]=(1.16379585731)*x_ref[0]+(-0.928037595416)*x_ref[1]+(0.998857178353)*x_ref[2]
        ref[0,0,2,1]=(-0.804515607022)*x_ref[0]+(-0.915469493439)*x_ref[1]+(-0.143888000623)*x_ref[2]
        ref[0,1,0,0]=(0.925843155486)*x_ref[0]+(-0.168243418953)*x_ref[1]+(-0.887839555187)*x_ref[2]
        ref[0,1,0,1]=(-0.913103105329)*x_ref[0]+(1.38011681732)*x_ref[1]+(-0.992478793291)*x_ref[2]
        ref[0,1,1,0]=(0.759685289903)*x_ref[0]+(-0.579481939982)*x_ref[1]+(0.577682929138)*x_ref[2]
        ref[0,1,1,1]=(0.813155548228)*x_ref[0]+(-0.178317843217)*x_ref[1]+(-0.0460113194526)*x_ref[2]
        ref[0,1,2,0]=(-0.195764860141)*x_ref[0]+(1.44256979628)*x_ref[1]+(0.283992441343)*x_ref[2]
        ref[0,1,2,1]=(-0.921229459072)*x_ref[0]+(-1.19295761547)*x_ref[1]+(0.215736801944)*x_ref[2]
        ref[0,2,0,0]=(-0.741726331752)*x_ref[0]+(-0.225585937382)*x_ref[1]+(1.12654556466)*x_ref[2]
        ref[0,2,0,1]=(-0.709656162421)*x_ref[0]+(0.298862010425)*x_ref[1]+(-0.502443284827)*x_ref[2]
        ref[0,2,1,0]=(-1.01166188669)*x_ref[0]+(-1.43763408953)*x_ref[1]+(-0.835160398488)*x_ref[2]
        ref[0,2,1,1]=(1.80968424545)*x_ref[0]+(-0.802666103495)*x_ref[1]+(0.290283606115)*x_ref[2]
        ref[0,2,2,0]=(0.346481981585)*x_ref[0]+(-0.444176962215)*x_ref[1]+(0.689127612227)*x_ref[2]
        ref[0,2,2,1]=(1.06925494095)*x_ref[0]+(0.452854988449)*x_ref[1]+(-1.19974690364)*x_ref[2]
        ref[0,3,0,0]=(-0.534896036328)*x_ref[0]+(-0.312000770615)*x_ref[1]+(-0.440033746693)*x_ref[2]
        ref[0,3,0,1]=(0.580351431838)*x_ref[0]+(-0.36623848451)*x_ref[1]+(-0.507837490037)*x_ref[2]
        ref[0,3,1,0]=(0.383180910522)*x_ref[0]+(-0.0907152443563)*x_ref[1]+(-0.200103743943)*x_ref[2]
        ref[0,3,1,1]=(0.0160194687513)*x_ref[0]+(0.862509612939)*x_ref[1]+(-1.14643594474)*x_ref[2]
        ref[0,3,2,0]=(-0.224390850199)*x_ref[0]+(-0.176927715447)*x_ref[1]+(0.213480525831)*x_ref[2]
        ref[0,3,2,1]=(-0.0508511448148)*x_ref[0]+(-1.28943467389)*x_ref[1]+(0.902285302487)*x_ref[2]
        ref[0,4,0,0]=(-0.633106625232)*x_ref[0]+(-0.430651917485)*x_ref[1]+(0.583149841952)*x_ref[2]
        ref[0,4,0,1]=(0.579035528956)*x_ref[0]+(0.159439682519)*x_ref[1]+(-0.0926277063021)*x_ref[2]
        ref[0,4,1,0]=(0.00484756366973)*x_ref[0]+(0.0324822270755)*x_ref[1]+(1.20052413378)*x_ref[2]
        ref[0,4,1,1]=(-0.337383994688)*x_ref[0]+(0.902076367981)*x_ref[1]+(-0.1956377395)*x_ref[2]
        ref[0,4,2,0]=(0.928434343456)*x_ref[0]+(0.0376569884355)*x_ref[1]+(-0.155274118568)*x_ref[2]
        ref[0,4,2,1]=(-0.541990213238)*x_ref[0]+(-0.0165076697829)*x_ref[1]+(-0.210319224859)*x_ref[2]
        ref[1,0,0,0]=(-1.67762025181)*x_ref[0]+(1.61434755579)*x_ref[1]+(0.357272049999)*x_ref[2]
        ref[1,0,0,1]=(-1.62532544549)*x_ref[0]+(-0.271426231027)*x_ref[1]+(0.338680991778)*x_ref[2]
        ref[1,0,1,0]=(1.73523408033)*x_ref[0]+(1.11987092368)*x_ref[1]+(0.126769364273)*x_ref[2]
        ref[1,0,1,1]=(-1.49220786558)*x_ref[0]+(1.51915566092)*x_ref[1]+(1.24788192157)*x_ref[2]
        ref[1,0,2,0]=(-0.687808766765)*x_ref[0]+(-1.02318754737)*x_ref[1]+(0.590653972416)*x_ref[2]
        ref[1,0,2,1]=(0.756067354802)*x_ref[0]+(-0.66964706965)*x_ref[1]+(-1.18986499158)*x_ref[2]
        ref[1,1,0,0]=(0.371345732177)*x_ref[0]+(0.212739617115)*x_ref[1]+(-1.13439047405)*x_ref[2]
        ref[1,1,0,1]=(0.525761804111)*x_ref[0]+(0.144130399691)*x_ref[1]+(0.715517587677)*x_ref[2]
        ref[1,1,1,0]=(-0.376160293215)*x_ref[0]+(-0.469964935263)*x_ref[1]+(0.413518194707)*x_ref[2]
        ref[1,1,1,1]=(-0.787748398144)*x_ref[0]+(-0.810090402237)*x_ref[1]+(-0.50127109256)*x_ref[2]
        ref[1,1,2,0]=(0.654555083692)*x_ref[0]+(-0.142134700515)*x_ref[1]+(-1.55891930521)*x_ref[2]
        ref[1,1,2,1]=(-0.423058885456)*x_ref[0]+(-0.65930158641)*x_ref[1]+(0.94440057744)*x_ref[2]
        ref[1,2,0,0]=(-0.924832007053)*x_ref[0]+(1.0154890382)*x_ref[1]+(0.11938079748)*x_ref[2]
        ref[1,2,0,1]=(0.134642686931)*x_ref[0]+(-1.30908067998)*x_ref[1]+(0.367206139074)*x_ref[2]
        ref[1,2,1,0]=(1.55524055201)*x_ref[0]+(0.727208660637)*x_ref[1]+(0.809810917377)*x_ref[2]
        ref[1,2,1,1]=(0.94813391462)*x_ref[0]+(-1.49402524308)*x_ref[1]+(0.367824542434)*x_ref[2]
        ref[1,2,2,0]=(-0.608295798731)*x_ref[0]+(-0.142302252814)*x_ref[1]+(-0.986529251308)*x_ref[2]
        ref[1,2,2,1]=(0.0259394619607)*x_ref[0]+(-1.60728202467)*x_ref[1]+(0.312534656619)*x_ref[2]
        ref[1,3,0,0]=(-1.34063373729)*x_ref[0]+(-0.268499839329)*x_ref[1]+(-0.335125886299)*x_ref[2]
        ref[1,3,0,1]=(1.53280698072)*x_ref[0]+(1.24369684711)*x_ref[1]+(0.766369156132)*x_ref[2]
        ref[1,3,1,0]=(0.902067176876)*x_ref[0]+(0.791102994814)*x_ref[1]+(-0.89266430493)*x_ref[2]
        ref[1,3,1,1]=(0.704997629232)*x_ref[0]+(-0.551611360332)*x_ref[1]+(1.39134537491)*x_ref[2]
        ref[1,3,2,0]=(-0.151690675211)*x_ref[0]+(-1.01762975893)*x_ref[1]+(-0.627493214284)*x_ref[2]
        ref[1,3,2,1]=(1.00936645922)*x_ref[0]+(0.0322458997248)*x_ref[1]+(1.45039938194)*x_ref[2]
        ref[1,4,0,0]=(-0.0916390388904)*x_ref[0]+(-0.783111746469)*x_ref[1]+(0.389725342225)*x_ref[2]
        ref[1,4,0,1]=(1.18859328243)*x_ref[0]+(0.871689661558)*x_ref[1]+(-0.775464713448)*x_ref[2]
        ref[1,4,1,0]=(0.913118342488)*x_ref[0]+(1.53285648884)*x_ref[1]+(0.0947365101243)*x_ref[2]
        ref[1,4,1,1]=(-1.00735841224)*x_ref[0]+(-1.12544193607)*x_ref[1]+(0.267347229372)*x_ref[2]
        ref[1,4,2,0]=(0.285943438284)*x_ref[0]+(-1.0889389244)*x_ref[1]+(-1.03054842327)*x_ref[2]
        ref[1,4,2,1]=(-1.5768940139)*x_ref[0]+(0.91237754587)*x_ref[1]+(1.30220394199)*x_ref[2]
        ref[2,0,0,0]=(-0.0634935883277)*x_ref[0]+(0.157304131712)*x_ref[1]+(-1.2995042074)*x_ref[2]
        ref[2,0,0,1]=(-0.802705266871)*x_ref[0]+(0.238233162774)*x_ref[1]+(1.29463414013)*x_ref[2]
        ref[2,0,1,0]=(-0.925468462671)*x_ref[0]+(0.870960795966)*x_ref[1]+(0.156522765713)*x_ref[2]
        ref[2,0,1,1]=(-0.546666833309)*x_ref[0]+(-1.14649607995)*x_ref[1]+(-1.53906505048)*x_ref[2]
        ref[2,0,2,0]=(0.464952910857)*x_ref[0]+(0.408605786027)*x_ref[1]+(-0.886974881114)*x_ref[2]
        ref[2,0,2,1]=(0.491723247948)*x_ref[0]+(0.194616420077)*x_ref[1]+(0.409049782749)*x_ref[2]
        ref[2,1,0,0]=(-1.19836883114)*x_ref[0]+(1.0622722337)*x_ref[1]+(-0.109193867727)*x_ref[2]
        ref[2,1,0,1]=(1.01973600383)*x_ref[0]+(-0.154482220539)*x_ref[1]+(0.378696371956)*x_ref[2]
        ref[2,1,1,0]=(-0.0812015296246)*x_ref[0]+(-1.10252794568)*x_ref[1]+(-0.00585426581857)*x_ref[2]
        ref[2,1,1,1]=(0.536042665272)*x_ref[0]+(0.685507117255)*x_ref[1]+(-0.179652683139)*x_ref[2]
        ref[2,1,2,0]=(-0.36750467811)*x_ref[0]+(-1.23674550869)*x_ref[1]+(-0.762905702801)*x_ref[2]
        ref[2,1,2,1]=(0.329665965651)*x_ref[0]+(0.101195810737)*x_ref[1]+(-0.610015052375)*x_ref[2]
        ref[2,2,0,0]=(-0.81644635185)*x_ref[0]+(-0.604578476306)*x_ref[1]+(0.0440925706991)*x_ref[2]
        ref[2,2,0,1]=(-0.958241437971)*x_ref[0]+(0.38563498206)*x_ref[1]+(0.913406762318)*x_ref[2]
        ref[2,2,1,0]=(1.22027492224)*x_ref[0]+(-1.00726209939)*x_ref[1]+(-1.05083068303)*x_ref[2]
        ref[2,2,1,1]=(-1.47561192645)*x_ref[0]+(-0.903549355608)*x_ref[1]+(0.236771324671)*x_ref[2]
        ref[2,2,2,0]=(0.284995864449)*x_ref[0]+(-0.175760829921)*x_ref[1]+(-0.720607220001)*x_ref[2]
        ref[2,2,2,1]=(-0.535894515963)*x_ref[0]+(-0.821302957162)*x_ref[1]+(-1.75141540241)*x_ref[2]
        ref[2,3,0,0]=(-0.333082106674)*x_ref[0]+(0.579471826412)*x_ref[1]+(0.561184138108)*x_ref[2]
        ref[2,3,0,1]=(1.34803519557)*x_ref[0]+(-0.441299088849)*x_ref[1]+(-1.04945951616)*x_ref[2]
        ref[2,3,1,0]=(0.299011096602)*x_ref[0]+(1.32726458235)*x_ref[1]+(-0.196298687359)*x_ref[2]
        ref[2,3,1,1]=(1.48698834204)*x_ref[0]+(0.527221496971)*x_ref[1]+(0.0819813519422)*x_ref[2]
        ref[2,3,2,0]=(0.681680955153)*x_ref[0]+(0.503653972357)*x_ref[1]+(0.434669347891)*x_ref[2]
        ref[2,3,2,1]=(-0.152586789311)*x_ref[0]+(-0.765242635579)*x_ref[1]+(0.712663994387)*x_ref[2]
        ref[2,4,0,0]=(0.630152300463)*x_ref[0]+(1.12366733588)*x_ref[1]+(-0.662966700638)*x_ref[2]
        ref[2,4,0,1]=(1.02796709755)*x_ref[0]+(-0.24093881584)*x_ref[1]+(1.79445417019)*x_ref[2]
        ref[2,4,1,0]=(-0.954575760758)*x_ref[0]+(0.16331782608)*x_ref[1]+(1.38309921277)*x_ref[2]
        ref[2,4,1,1]=(0.536092514281)*x_ref[0]+(-0.276364952673)*x_ref[1]+(0.779619701528)*x_ref[2]
        ref[2,4,2,0]=(0.267270444141)*x_ref[0]+(-0.810746424331)*x_ref[1]+(-0.733877853264)*x_ref[2]
        ref[2,4,2,1]=(-0.0641688847923)*x_ref[0]+(-0.386279280667)*x_ref[1]+(0.0637571595255)*x_ref[2]
        ref[3,0,0,0]=(0.469903180687)*x_ref[0]+(-0.793812397342)*x_ref[1]+(1.60841921779)*x_ref[2]
        ref[3,0,0,1]=(0.720057600102)*x_ref[0]+(0.502582253261)*x_ref[1]+(0.460147791003)*x_ref[2]
        ref[3,0,1,0]=(0.0327862564137)*x_ref[0]+(-1.142962965)*x_ref[1]+(0.114799131894)*x_ref[2]
        ref[3,0,1,1]=(-0.997412656978)*x_ref[0]+(-0.678175468576)*x_ref[1]+(-0.152663147682)*x_ref[2]
        ref[3,0,2,0]=(0.810424599078)*x_ref[0]+(0.75522727786)*x_ref[1]+(-0.205880941501)*x_ref[2]
        ref[3,0,2,1]=(0.353601949476)*x_ref[0]+(0.0963200813379)*x_ref[1]+(0.0578829287146)*x_ref[2]
        ref[3,1,0,0]=(0.346590730062)*x_ref[0]+(1.07953790092)*x_ref[1]+(0.647586851217)*x_ref[2]
        ref[3,1,0,1]=(0.889879195334)*x_ref[0]+(-0.352788186249)*x_ref[1]+(0.434900135803)*x_ref[2]
        ref[3,1,1,0]=(-0.721742734959)*x_ref[0]+(0.166066672134)*x_ref[1]+(-0.489266924809)*x_ref[2]
        ref[3,1,1,1]=(-0.810487755181)*x_ref[0]+(0.0354916979253)*x_ref[1]+(0.0735520654395)*x_ref[2]
        ref[3,1,2,0]=(1.02414338939)*x_ref[0]+(-0.851076284556)*x_ref[1]+(-0.232174494405)*x_ref[2]
        ref[3,1,2,1]=(-0.327246566026)*x_ref[0]+(-0.0499396023605)*x_ref[1]+(0.611585402379)*x_ref[2]
        ref[3,2,0,0]=(-0.365757628557)*x_ref[0]+(0.868697342115)*x_ref[1]+(-1.102545972)*x_ref[2]
        ref[3,2,0,1]=(0.628592129227)*x_ref[0]+(1.25545515311)*x_ref[1]+(-1.0577476264)*x_ref[2]
        ref[3,2,1,0]=(0.160995978044)*x_ref[0]+(-0.947389499556)*x_ref[1]+(-0.870476160656)*x_ref[2]
        ref[3,2,1,1]=(0.910787139938)*x_ref[0]+(-0.0681493345139)*x_ref[1]+(-0.561910153781)*x_ref[2]
        ref[3,2,2,0]=(0.0761644634428)*x_ref[0]+(-0.700940967343)*x_ref[1]+(1.19407210904)*x_ref[2]
        ref[3,2,2,1]=(1.75717437065)*x_ref[0]+(-0.144457317494)*x_ref[1]+(-0.523233453079)*x_ref[2]
        ref[3,3,0,0]=(0.448782872606)*x_ref[0]+(0.650464256547)*x_ref[1]+(1.19476547456)*x_ref[2]
        ref[3,3,0,1]=(0.118586027201)*x_ref[0]+(0.449971816918)*x_ref[1]+(-0.510812733389)*x_ref[2]
        ref[3,3,1,0]=(-0.330032265081)*x_ref[0]+(-0.13657534891)*x_ref[1]+(0.456365289556)*x_ref[2]
        ref[3,3,1,1]=(0.697584212635)*x_ref[0]+(0.897604241724)*x_ref[1]+(1.31517116663)*x_ref[2]
        ref[3,3,2,0]=(-0.940538707365)*x_ref[0]+(-0.334862221665)*x_ref[1]+(0.0969742096754)*x_ref[2]
        ref[3,3,2,1]=(-0.985788813516)*x_ref[0]+(0.348389725548)*x_ref[1]+(-0.260816590361)*x_ref[2]
        ref[3,4,0,0]=(0.754692421417)*x_ref[0]+(0.775985781769)*x_ref[1]+(-0.0866159222457)*x_ref[2]
        ref[3,4,0,1]=(-1.17670502429)*x_ref[0]+(1.12268279393)*x_ref[1]+(-1.1279822095)*x_ref[2]
        ref[3,4,1,0]=(-1.48674370666)*x_ref[0]+(1.26685445713)*x_ref[1]+(-0.0522828809306)*x_ref[2]
        ref[3,4,1,1]=(0.401760689843)*x_ref[0]+(-1.35707797971)*x_ref[1]+(-0.0580654273828)*x_ref[2]
        ref[3,4,2,0]=(-1.40697830852)*x_ref[0]+(0.933923369958)*x_ref[1]+(0.230152169159)*x_ref[2]
        ref[3,4,2,1]=(-1.21242130555)*x_ref[0]+(0.858316339663)*x_ref[1]+(-0.839747078526)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.485106931655)*x[0]**o+(-0.45911691249)*x[0]+(-0.780654139819)*x[1]**o+(-0.496275334303)*x[1]
        ref=(-0.485106931655)*x_ref[0]**o+(-0.45911691249)*x_ref[0]+(-0.780654139819)*x_ref[1]**o+(-0.496275334303)*x_ref[1]
      else:
        arg=(-0.312233545328)*x[0]**o+(0.386652391857)*x[0]+(0.893938213685)*x[1]**o+(-0.109836308832)*x[1]+(0.293419600866)*x[2]**o+(-0.802363407879)*x[2]
        ref=(-0.312233545328)*x_ref[0]**o+(0.386652391857)*x_ref[0]+(0.893938213685)*x_ref[1]**o+(-0.109836308832)*x_ref[1]+(0.293419600866)*x_ref[2]**o+(-0.802363407879)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.383147346126)*x[0]**o+(-0.103098043084)*x[0]+(0.389581571033)*x[1]**o+(0.647441523744)*x[1]
        arg[1]=(-0.0718078186662)*x[0]**o+(-0.975209175853)*x[0]+(-0.62913306056)*x[1]**o+(0.648038487393)*x[1]
        ref[0]=(0.383147346126)*x_ref[0]**o+(-0.103098043084)*x_ref[0]+(0.389581571033)*x_ref[1]**o+(0.647441523744)*x_ref[1]
        ref[1]=(-0.0718078186662)*x_ref[0]**o+(-0.975209175853)*x_ref[0]+(-0.62913306056)*x_ref[1]**o+(0.648038487393)*x_ref[1]
      else:
        arg[0]=(-0.392958184101)*x[0]**o+(-0.529709692826)*x[0]+(-0.834422900831)*x[1]**o+(0.52562214773)*x[1]+(-0.726762801989)*x[2]**o+(-0.329225625652)*x[2]
        arg[1]=(-0.388486489727)*x[0]**o+(0.62353995706)*x[0]+(0.0198506797091)*x[1]**o+(0.744336375443)*x[1]+(0.115966573066)*x[2]**o+(-0.195838697299)*x[2]
        ref[0]=(-0.392958184101)*x_ref[0]**o+(-0.529709692826)*x_ref[0]+(-0.834422900831)*x_ref[1]**o+(0.52562214773)*x_ref[1]+(-0.726762801989)*x_ref[2]**o+(-0.329225625652)*x_ref[2]
        ref[1]=(-0.388486489727)*x_ref[0]**o+(0.62353995706)*x_ref[0]+(0.0198506797091)*x_ref[1]**o+(0.744336375443)*x_ref[1]+(0.115966573066)*x_ref[2]**o+(-0.195838697299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.657915437562)*x[0]**o+(-0.545113889022)*x[0]+(0.531900702747)*x[1]**o+(0.593791195717)*x[1]
        arg[0,1]=(0.051080053648)*x[0]**o+(-0.693135614033)*x[0]+(0.132124625293)*x[1]**o+(0.790120778986)*x[1]
        arg[0,2]=(-0.474326324304)*x[0]**o+(0.844119715746)*x[0]+(-0.107139623866)*x[1]**o+(0.456623369855)*x[1]
        arg[0,3]=(-0.582421376463)*x[0]**o+(0.972810074845)*x[0]+(0.875278994526)*x[1]**o+(-0.382656553829)*x[1]
        arg[0,4]=(-0.904956623057)*x[0]**o+(0.701968065151)*x[0]+(-0.710231782855)*x[1]**o+(-0.123456684979)*x[1]
        arg[1,0]=(0.361017539866)*x[0]**o+(0.865249575899)*x[0]+(0.529515655309)*x[1]**o+(0.161402780958)*x[1]
        arg[1,1]=(0.813945406612)*x[0]**o+(0.801153720706)*x[0]+(-0.254227034266)*x[1]**o+(-0.422848409284)*x[1]
        arg[1,2]=(0.56207473833)*x[0]**o+(-0.831638855484)*x[0]+(-0.584619648965)*x[1]**o+(0.453944553892)*x[1]
        arg[1,3]=(0.25526821881)*x[0]**o+(-0.436135120045)*x[0]+(0.524842439137)*x[1]**o+(0.568007117728)*x[1]
        arg[1,4]=(-0.0587551077421)*x[0]**o+(-0.434635762185)*x[0]+(-0.592094375986)*x[1]**o+(0.873123434395)*x[1]
        arg[2,0]=(0.266819439236)*x[0]**o+(-0.45156140915)*x[0]+(0.436756372775)*x[1]**o+(0.12295128383)*x[1]
        arg[2,1]=(0.347128677451)*x[0]**o+(-0.0983169467916)*x[0]+(-0.414573209624)*x[1]**o+(-0.171300108487)*x[1]
        arg[2,2]=(-0.279695154296)*x[0]**o+(-0.911450144779)*x[0]+(0.5055340448)*x[1]**o+(-0.858058864152)*x[1]
        arg[2,3]=(-0.53768906863)*x[0]**o+(-0.0626923004303)*x[0]+(0.72295545269)*x[1]**o+(-0.77070425804)*x[1]
        arg[2,4]=(0.18215032008)*x[0]**o+(-0.965014882526)*x[0]+(0.596564813119)*x[1]**o+(0.0817024840554)*x[1]
        arg[3,0]=(-0.572404596368)*x[0]**o+(0.488365635829)*x[0]+(0.145439800826)*x[1]**o+(-0.337589667899)*x[1]
        arg[3,1]=(-0.628977426654)*x[0]**o+(0.22731945375)*x[0]+(0.989224773975)*x[1]**o+(0.0691553279499)*x[1]
        arg[3,2]=(-0.43609079916)*x[0]**o+(0.966113801416)*x[0]+(-0.853136416278)*x[1]**o+(-0.0125151911906)*x[1]
        arg[3,3]=(0.340159347855)*x[0]**o+(-0.254775013843)*x[0]+(0.795261224144)*x[1]**o+(-0.692255020659)*x[1]
        arg[3,4]=(-0.153521724504)*x[0]**o+(-0.681779299881)*x[0]+(0.693860779041)*x[1]**o+(0.882012604794)*x[1]
        ref[0,0]=(-0.657915437562)*x_ref[0]**o+(-0.545113889022)*x_ref[0]+(0.531900702747)*x_ref[1]**o+(0.593791195717)*x_ref[1]
        ref[0,1]=(0.051080053648)*x_ref[0]**o+(-0.693135614033)*x_ref[0]+(0.132124625293)*x_ref[1]**o+(0.790120778986)*x_ref[1]
        ref[0,2]=(-0.474326324304)*x_ref[0]**o+(0.844119715746)*x_ref[0]+(-0.107139623866)*x_ref[1]**o+(0.456623369855)*x_ref[1]
        ref[0,3]=(-0.582421376463)*x_ref[0]**o+(0.972810074845)*x_ref[0]+(0.875278994526)*x_ref[1]**o+(-0.382656553829)*x_ref[1]
        ref[0,4]=(-0.904956623057)*x_ref[0]**o+(0.701968065151)*x_ref[0]+(-0.710231782855)*x_ref[1]**o+(-0.123456684979)*x_ref[1]
        ref[1,0]=(0.361017539866)*x_ref[0]**o+(0.865249575899)*x_ref[0]+(0.529515655309)*x_ref[1]**o+(0.161402780958)*x_ref[1]
        ref[1,1]=(0.813945406612)*x_ref[0]**o+(0.801153720706)*x_ref[0]+(-0.254227034266)*x_ref[1]**o+(-0.422848409284)*x_ref[1]
        ref[1,2]=(0.56207473833)*x_ref[0]**o+(-0.831638855484)*x_ref[0]+(-0.584619648965)*x_ref[1]**o+(0.453944553892)*x_ref[1]
        ref[1,3]=(0.25526821881)*x_ref[0]**o+(-0.436135120045)*x_ref[0]+(0.524842439137)*x_ref[1]**o+(0.568007117728)*x_ref[1]
        ref[1,4]=(-0.0587551077421)*x_ref[0]**o+(-0.434635762185)*x_ref[0]+(-0.592094375986)*x_ref[1]**o+(0.873123434395)*x_ref[1]
        ref[2,0]=(0.266819439236)*x_ref[0]**o+(-0.45156140915)*x_ref[0]+(0.436756372775)*x_ref[1]**o+(0.12295128383)*x_ref[1]
        ref[2,1]=(0.347128677451)*x_ref[0]**o+(-0.0983169467916)*x_ref[0]+(-0.414573209624)*x_ref[1]**o+(-0.171300108487)*x_ref[1]
        ref[2,2]=(-0.279695154296)*x_ref[0]**o+(-0.911450144779)*x_ref[0]+(0.5055340448)*x_ref[1]**o+(-0.858058864152)*x_ref[1]
        ref[2,3]=(-0.53768906863)*x_ref[0]**o+(-0.0626923004303)*x_ref[0]+(0.72295545269)*x_ref[1]**o+(-0.77070425804)*x_ref[1]
        ref[2,4]=(0.18215032008)*x_ref[0]**o+(-0.965014882526)*x_ref[0]+(0.596564813119)*x_ref[1]**o+(0.0817024840554)*x_ref[1]
        ref[3,0]=(-0.572404596368)*x_ref[0]**o+(0.488365635829)*x_ref[0]+(0.145439800826)*x_ref[1]**o+(-0.337589667899)*x_ref[1]
        ref[3,1]=(-0.628977426654)*x_ref[0]**o+(0.22731945375)*x_ref[0]+(0.989224773975)*x_ref[1]**o+(0.0691553279499)*x_ref[1]
        ref[3,2]=(-0.43609079916)*x_ref[0]**o+(0.966113801416)*x_ref[0]+(-0.853136416278)*x_ref[1]**o+(-0.0125151911906)*x_ref[1]
        ref[3,3]=(0.340159347855)*x_ref[0]**o+(-0.254775013843)*x_ref[0]+(0.795261224144)*x_ref[1]**o+(-0.692255020659)*x_ref[1]
        ref[3,4]=(-0.153521724504)*x_ref[0]**o+(-0.681779299881)*x_ref[0]+(0.693860779041)*x_ref[1]**o+(0.882012604794)*x_ref[1]
      else:
        arg[0,0]=(-0.0201751669244)*x[0]**o+(0.337967826628)*x[0]+(0.262482967798)*x[1]**o+(-0.192863458488)*x[1]+(-0.696822606165)*x[2]**o+(-0.476556518002)*x[2]
        arg[0,1]=(-0.546277895941)*x[0]**o+(-0.118793125394)*x[0]+(0.987786322041)*x[1]**o+(-0.241472898602)*x[1]+(-0.454308453322)*x[2]**o+(-0.27057260464)*x[2]
        arg[0,2]=(0.383631846628)*x[0]**o+(0.979909217026)*x[0]+(-0.0503694684178)*x[1]**o+(-0.748153155067)*x[1]+(-0.721832849635)*x[2]**o+(0.144937710994)*x[2]
        arg[0,3]=(-0.540144273345)*x[0]**o+(-0.986853732479)*x[0]+(-0.14527387031)*x[1]**o+(0.329815126012)*x[1]+(0.566431907689)*x[2]**o+(-0.323179211167)*x[2]
        arg[0,4]=(-0.448624394254)*x[0]**o+(-0.272742080619)*x[0]+(-0.451405020267)*x[1]**o+(0.173204890481)*x[1]+(-0.0122463571841)*x[2]**o+(-0.595932249794)*x[2]
        arg[1,0]=(0.984160677776)*x[0]**o+(-0.243512460216)*x[0]+(0.82590496761)*x[1]**o+(-0.448270326787)*x[1]+(0.286008153265)*x[2]**o+(0.804229889769)*x[2]
        arg[1,1]=(0.531762548093)*x[0]**o+(0.211455783615)*x[0]+(0.836058347792)*x[1]**o+(0.806904570574)*x[1]+(-0.052512287992)*x[2]**o+(0.998679219355)*x[2]
        arg[1,2]=(0.462615968956)*x[0]**o+(0.960862689218)*x[0]+(-0.693932866844)*x[1]**o+(0.881110014782)*x[1]+(0.976329091354)*x[2]**o+(-0.647483316997)*x[2]
        arg[1,3]=(-0.764181748932)*x[0]**o+(-0.962197871121)*x[0]+(-0.143747029438)*x[1]**o+(0.913962833322)*x[1]+(0.165013597835)*x[2]**o+(0.166877976184)*x[2]
        arg[1,4]=(0.687688905493)*x[0]**o+(-0.152816071208)*x[0]+(-0.277974702822)*x[1]**o+(-0.389918839086)*x[1]+(0.229658530129)*x[2]**o+(-0.69924627348)*x[2]
        arg[2,0]=(0.0598215047949)*x[0]**o+(0.592389000293)*x[0]+(0.00886573296364)*x[1]**o+(0.339086619934)*x[1]+(0.476600446349)*x[2]**o+(0.767524427684)*x[2]
        arg[2,1]=(0.525936212445)*x[0]**o+(0.44607860423)*x[0]+(0.0397184541609)*x[1]**o+(-0.424118244789)*x[1]+(-0.273077104391)*x[2]**o+(0.435027185784)*x[2]
        arg[2,2]=(0.672992314995)*x[0]**o+(0.993339100191)*x[0]+(0.488983819898)*x[1]**o+(0.568579529177)*x[1]+(-0.321282052211)*x[2]**o+(-0.268847753649)*x[2]
        arg[2,3]=(-0.0536748741467)*x[0]**o+(-0.195658514433)*x[0]+(-0.931652264383)*x[1]**o+(-0.89957282695)*x[1]+(-0.811088135231)*x[2]**o+(0.0699177832759)*x[2]
        arg[2,4]=(-0.707875141752)*x[0]**o+(-0.00297935832343)*x[0]+(-0.959580991547)*x[1]**o+(0.0787110832272)*x[1]+(-0.754646316678)*x[2]**o+(-0.432731727467)*x[2]
        arg[3,0]=(0.728853463111)*x[0]**o+(-0.681375546727)*x[0]+(-0.283722278591)*x[1]**o+(0.950522890527)*x[1]+(0.608194817491)*x[2]**o+(0.571156129477)*x[2]
        arg[3,1]=(-0.0208969501228)*x[0]**o+(0.572596028151)*x[0]+(0.811179009214)*x[1]**o+(0.519098352581)*x[1]+(0.154527510954)*x[2]**o+(-0.21413742337)*x[2]
        arg[3,2]=(0.347209642455)*x[0]**o+(-0.114911324326)*x[0]+(-0.104561779357)*x[1]**o+(0.848703366841)*x[1]+(0.680312939858)*x[2]**o+(-0.680566404979)*x[2]
        arg[3,3]=(0.342687985669)*x[0]**o+(0.443470953221)*x[0]+(-0.849668173915)*x[1]**o+(0.337689225078)*x[1]+(-0.35083505383)*x[2]**o+(-0.284627678739)*x[2]
        arg[3,4]=(0.661690111131)*x[0]**o+(-0.0918926979359)*x[0]+(-0.587753600932)*x[1]**o+(-0.383115718043)*x[1]+(-0.0696588695623)*x[2]**o+(-0.799630628972)*x[2]
        ref[0,0]=(-0.0201751669244)*x_ref[0]**o+(0.337967826628)*x_ref[0]+(0.262482967798)*x_ref[1]**o+(-0.192863458488)*x_ref[1]+(-0.696822606165)*x_ref[2]**o+(-0.476556518002)*x_ref[2]
        ref[0,1]=(-0.546277895941)*x_ref[0]**o+(-0.118793125394)*x_ref[0]+(0.987786322041)*x_ref[1]**o+(-0.241472898602)*x_ref[1]+(-0.454308453322)*x_ref[2]**o+(-0.27057260464)*x_ref[2]
        ref[0,2]=(0.383631846628)*x_ref[0]**o+(0.979909217026)*x_ref[0]+(-0.0503694684178)*x_ref[1]**o+(-0.748153155067)*x_ref[1]+(-0.721832849635)*x_ref[2]**o+(0.144937710994)*x_ref[2]
        ref[0,3]=(-0.540144273345)*x_ref[0]**o+(-0.986853732479)*x_ref[0]+(-0.14527387031)*x_ref[1]**o+(0.329815126012)*x_ref[1]+(0.566431907689)*x_ref[2]**o+(-0.323179211167)*x_ref[2]
        ref[0,4]=(-0.448624394254)*x_ref[0]**o+(-0.272742080619)*x_ref[0]+(-0.451405020267)*x_ref[1]**o+(0.173204890481)*x_ref[1]+(-0.0122463571841)*x_ref[2]**o+(-0.595932249794)*x_ref[2]
        ref[1,0]=(0.984160677776)*x_ref[0]**o+(-0.243512460216)*x_ref[0]+(0.82590496761)*x_ref[1]**o+(-0.448270326787)*x_ref[1]+(0.286008153265)*x_ref[2]**o+(0.804229889769)*x_ref[2]
        ref[1,1]=(0.531762548093)*x_ref[0]**o+(0.211455783615)*x_ref[0]+(0.836058347792)*x_ref[1]**o+(0.806904570574)*x_ref[1]+(-0.052512287992)*x_ref[2]**o+(0.998679219355)*x_ref[2]
        ref[1,2]=(0.462615968956)*x_ref[0]**o+(0.960862689218)*x_ref[0]+(-0.693932866844)*x_ref[1]**o+(0.881110014782)*x_ref[1]+(0.976329091354)*x_ref[2]**o+(-0.647483316997)*x_ref[2]
        ref[1,3]=(-0.764181748932)*x_ref[0]**o+(-0.962197871121)*x_ref[0]+(-0.143747029438)*x_ref[1]**o+(0.913962833322)*x_ref[1]+(0.165013597835)*x_ref[2]**o+(0.166877976184)*x_ref[2]
        ref[1,4]=(0.687688905493)*x_ref[0]**o+(-0.152816071208)*x_ref[0]+(-0.277974702822)*x_ref[1]**o+(-0.389918839086)*x_ref[1]+(0.229658530129)*x_ref[2]**o+(-0.69924627348)*x_ref[2]
        ref[2,0]=(0.0598215047949)*x_ref[0]**o+(0.592389000293)*x_ref[0]+(0.00886573296364)*x_ref[1]**o+(0.339086619934)*x_ref[1]+(0.476600446349)*x_ref[2]**o+(0.767524427684)*x_ref[2]
        ref[2,1]=(0.525936212445)*x_ref[0]**o+(0.44607860423)*x_ref[0]+(0.0397184541609)*x_ref[1]**o+(-0.424118244789)*x_ref[1]+(-0.273077104391)*x_ref[2]**o+(0.435027185784)*x_ref[2]
        ref[2,2]=(0.672992314995)*x_ref[0]**o+(0.993339100191)*x_ref[0]+(0.488983819898)*x_ref[1]**o+(0.568579529177)*x_ref[1]+(-0.321282052211)*x_ref[2]**o+(-0.268847753649)*x_ref[2]
        ref[2,3]=(-0.0536748741467)*x_ref[0]**o+(-0.195658514433)*x_ref[0]+(-0.931652264383)*x_ref[1]**o+(-0.89957282695)*x_ref[1]+(-0.811088135231)*x_ref[2]**o+(0.0699177832759)*x_ref[2]
        ref[2,4]=(-0.707875141752)*x_ref[0]**o+(-0.00297935832343)*x_ref[0]+(-0.959580991547)*x_ref[1]**o+(0.0787110832272)*x_ref[1]+(-0.754646316678)*x_ref[2]**o+(-0.432731727467)*x_ref[2]
        ref[3,0]=(0.728853463111)*x_ref[0]**o+(-0.681375546727)*x_ref[0]+(-0.283722278591)*x_ref[1]**o+(0.950522890527)*x_ref[1]+(0.608194817491)*x_ref[2]**o+(0.571156129477)*x_ref[2]
        ref[3,1]=(-0.0208969501228)*x_ref[0]**o+(0.572596028151)*x_ref[0]+(0.811179009214)*x_ref[1]**o+(0.519098352581)*x_ref[1]+(0.154527510954)*x_ref[2]**o+(-0.21413742337)*x_ref[2]
        ref[3,2]=(0.347209642455)*x_ref[0]**o+(-0.114911324326)*x_ref[0]+(-0.104561779357)*x_ref[1]**o+(0.848703366841)*x_ref[1]+(0.680312939858)*x_ref[2]**o+(-0.680566404979)*x_ref[2]
        ref[3,3]=(0.342687985669)*x_ref[0]**o+(0.443470953221)*x_ref[0]+(-0.849668173915)*x_ref[1]**o+(0.337689225078)*x_ref[1]+(-0.35083505383)*x_ref[2]**o+(-0.284627678739)*x_ref[2]
        ref[3,4]=(0.661690111131)*x_ref[0]**o+(-0.0918926979359)*x_ref[0]+(-0.587753600932)*x_ref[1]**o+(-0.383115718043)*x_ref[1]+(-0.0696588695623)*x_ref[2]**o+(-0.799630628972)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.906705618142)*x[0]**o+(0.458982439103)*x[0]+(0.556464353903)*x[1]**o+(0.733236686234)*x[1]
        arg[0,0,1]=(-0.192889954523)*x[0]**o+(-0.814332539133)*x[0]+(0.446164214282)*x[1]**o+(0.032291315446)*x[1]
        arg[0,1,0]=(0.757081929199)*x[0]**o+(-0.3938524931)*x[0]+(-0.458060132115)*x[1]**o+(0.471831491182)*x[1]
        arg[0,1,1]=(0.459868785407)*x[0]**o+(0.513707183316)*x[0]+(-0.9682413377)*x[1]**o+(-0.49496265187)*x[1]
        arg[1,0,0]=(0.546655212697)*x[0]**o+(-0.296125788768)*x[0]+(0.570033305927)*x[1]**o+(-0.818919094317)*x[1]
        arg[1,0,1]=(-0.954422252892)*x[0]**o+(-0.925035286684)*x[0]+(0.171311127546)*x[1]**o+(-0.584606597796)*x[1]
        arg[1,1,0]=(-0.203956024883)*x[0]**o+(-0.933117428128)*x[0]+(-0.745101180472)*x[1]**o+(0.0390208509919)*x[1]
        arg[1,1,1]=(-0.67467593628)*x[0]**o+(0.42281003355)*x[0]+(-0.495401176467)*x[1]**o+(0.993257306842)*x[1]
        arg[2,0,0]=(-0.204011890956)*x[0]**o+(0.730452092671)*x[0]+(-0.0486718200506)*x[1]**o+(0.691928726795)*x[1]
        arg[2,0,1]=(0.381571380631)*x[0]**o+(0.504896621541)*x[0]+(-0.229805983176)*x[1]**o+(0.281401365705)*x[1]
        arg[2,1,0]=(0.578852751326)*x[0]**o+(-0.38588684985)*x[0]+(-0.630908284751)*x[1]**o+(0.237972319424)*x[1]
        arg[2,1,1]=(-0.928366568501)*x[0]**o+(0.254484026833)*x[0]+(0.903931940538)*x[1]**o+(-0.370818194785)*x[1]
        arg[3,0,0]=(-0.216012465475)*x[0]**o+(-0.938274177001)*x[0]+(0.0674408710328)*x[1]**o+(-0.731594784098)*x[1]
        arg[3,0,1]=(-0.156790905207)*x[0]**o+(0.284403081253)*x[0]+(0.3754228052)*x[1]**o+(-0.772758808958)*x[1]
        arg[3,1,0]=(0.89987768188)*x[0]**o+(0.0177886564195)*x[0]+(0.102329016274)*x[1]**o+(0.474039949682)*x[1]
        arg[3,1,1]=(0.0416243959618)*x[0]**o+(0.739593496294)*x[0]+(-0.0215969068381)*x[1]**o+(0.203954403681)*x[1]
        arg[4,0,0]=(-0.479380091881)*x[0]**o+(0.973630227471)*x[0]+(0.372617682727)*x[1]**o+(0.721993565094)*x[1]
        arg[4,0,1]=(0.360211306746)*x[0]**o+(0.260993313325)*x[0]+(0.252635752916)*x[1]**o+(0.534875843061)*x[1]
        arg[4,1,0]=(-0.609695114069)*x[0]**o+(0.283865161572)*x[0]+(-0.607441168016)*x[1]**o+(-0.838279309054)*x[1]
        arg[4,1,1]=(0.984921561958)*x[0]**o+(-0.352255390774)*x[0]+(-0.770706841576)*x[1]**o+(0.456415771191)*x[1]
        arg[5,0,0]=(0.903410389973)*x[0]**o+(0.609651342058)*x[0]+(-0.0890705788805)*x[1]**o+(0.613493629434)*x[1]
        arg[5,0,1]=(-0.977979577058)*x[0]**o+(-0.121665631763)*x[0]+(0.466897174395)*x[1]**o+(-0.438819940752)*x[1]
        arg[5,1,0]=(-0.537313389905)*x[0]**o+(-0.0401503571218)*x[0]+(-0.454563705252)*x[1]**o+(-0.728276295145)*x[1]
        arg[5,1,1]=(0.858855332454)*x[0]**o+(0.378938106473)*x[0]+(0.908957690039)*x[1]**o+(0.989040980713)*x[1]
        ref[0,0,0]=(-0.906705618142)*x_ref[0]**o+(0.458982439103)*x_ref[0]+(0.556464353903)*x_ref[1]**o+(0.733236686234)*x_ref[1]
        ref[0,0,1]=(-0.192889954523)*x_ref[0]**o+(-0.814332539133)*x_ref[0]+(0.446164214282)*x_ref[1]**o+(0.032291315446)*x_ref[1]
        ref[0,1,0]=(0.757081929199)*x_ref[0]**o+(-0.3938524931)*x_ref[0]+(-0.458060132115)*x_ref[1]**o+(0.471831491182)*x_ref[1]
        ref[0,1,1]=(0.459868785407)*x_ref[0]**o+(0.513707183316)*x_ref[0]+(-0.9682413377)*x_ref[1]**o+(-0.49496265187)*x_ref[1]
        ref[1,0,0]=(0.546655212697)*x_ref[0]**o+(-0.296125788768)*x_ref[0]+(0.570033305927)*x_ref[1]**o+(-0.818919094317)*x_ref[1]
        ref[1,0,1]=(-0.954422252892)*x_ref[0]**o+(-0.925035286684)*x_ref[0]+(0.171311127546)*x_ref[1]**o+(-0.584606597796)*x_ref[1]
        ref[1,1,0]=(-0.203956024883)*x_ref[0]**o+(-0.933117428128)*x_ref[0]+(-0.745101180472)*x_ref[1]**o+(0.0390208509919)*x_ref[1]
        ref[1,1,1]=(-0.67467593628)*x_ref[0]**o+(0.42281003355)*x_ref[0]+(-0.495401176467)*x_ref[1]**o+(0.993257306842)*x_ref[1]
        ref[2,0,0]=(-0.204011890956)*x_ref[0]**o+(0.730452092671)*x_ref[0]+(-0.0486718200506)*x_ref[1]**o+(0.691928726795)*x_ref[1]
        ref[2,0,1]=(0.381571380631)*x_ref[0]**o+(0.504896621541)*x_ref[0]+(-0.229805983176)*x_ref[1]**o+(0.281401365705)*x_ref[1]
        ref[2,1,0]=(0.578852751326)*x_ref[0]**o+(-0.38588684985)*x_ref[0]+(-0.630908284751)*x_ref[1]**o+(0.237972319424)*x_ref[1]
        ref[2,1,1]=(-0.928366568501)*x_ref[0]**o+(0.254484026833)*x_ref[0]+(0.903931940538)*x_ref[1]**o+(-0.370818194785)*x_ref[1]
        ref[3,0,0]=(-0.216012465475)*x_ref[0]**o+(-0.938274177001)*x_ref[0]+(0.0674408710328)*x_ref[1]**o+(-0.731594784098)*x_ref[1]
        ref[3,0,1]=(-0.156790905207)*x_ref[0]**o+(0.284403081253)*x_ref[0]+(0.3754228052)*x_ref[1]**o+(-0.772758808958)*x_ref[1]
        ref[3,1,0]=(0.89987768188)*x_ref[0]**o+(0.0177886564195)*x_ref[0]+(0.102329016274)*x_ref[1]**o+(0.474039949682)*x_ref[1]
        ref[3,1,1]=(0.0416243959618)*x_ref[0]**o+(0.739593496294)*x_ref[0]+(-0.0215969068381)*x_ref[1]**o+(0.203954403681)*x_ref[1]
        ref[4,0,0]=(-0.479380091881)*x_ref[0]**o+(0.973630227471)*x_ref[0]+(0.372617682727)*x_ref[1]**o+(0.721993565094)*x_ref[1]
        ref[4,0,1]=(0.360211306746)*x_ref[0]**o+(0.260993313325)*x_ref[0]+(0.252635752916)*x_ref[1]**o+(0.534875843061)*x_ref[1]
        ref[4,1,0]=(-0.609695114069)*x_ref[0]**o+(0.283865161572)*x_ref[0]+(-0.607441168016)*x_ref[1]**o+(-0.838279309054)*x_ref[1]
        ref[4,1,1]=(0.984921561958)*x_ref[0]**o+(-0.352255390774)*x_ref[0]+(-0.770706841576)*x_ref[1]**o+(0.456415771191)*x_ref[1]
        ref[5,0,0]=(0.903410389973)*x_ref[0]**o+(0.609651342058)*x_ref[0]+(-0.0890705788805)*x_ref[1]**o+(0.613493629434)*x_ref[1]
        ref[5,0,1]=(-0.977979577058)*x_ref[0]**o+(-0.121665631763)*x_ref[0]+(0.466897174395)*x_ref[1]**o+(-0.438819940752)*x_ref[1]
        ref[5,1,0]=(-0.537313389905)*x_ref[0]**o+(-0.0401503571218)*x_ref[0]+(-0.454563705252)*x_ref[1]**o+(-0.728276295145)*x_ref[1]
        ref[5,1,1]=(0.858855332454)*x_ref[0]**o+(0.378938106473)*x_ref[0]+(0.908957690039)*x_ref[1]**o+(0.989040980713)*x_ref[1]
      else:
        arg[0,0,0]=(-0.550852990025)*x[0]**o+(0.578705971531)*x[0]+(0.9533401901)*x[1]**o+(0.233181910221)*x[1]+(-0.143544272432)*x[2]**o+(-0.986513253736)*x[2]
        arg[0,0,1]=(-0.326417636931)*x[0]**o+(0.0910236969351)*x[0]+(-0.8240642668)*x[1]**o+(-0.242359389714)*x[1]+(0.266813299647)*x[2]**o+(-0.364956839435)*x[2]
        arg[0,1,0]=(-0.777107068999)*x[0]**o+(0.201120303703)*x[0]+(-0.684906707463)*x[1]**o+(-0.9620762843)*x[1]+(-0.314789913974)*x[2]**o+(-0.70682592319)*x[2]
        arg[0,1,1]=(-0.980292769796)*x[0]**o+(-0.827763529178)*x[0]+(0.256770279946)*x[1]**o+(0.619665292369)*x[1]+(0.014752087873)*x[2]**o+(-0.810044916486)*x[2]
        arg[1,0,0]=(-0.51164541516)*x[0]**o+(-0.429646605692)*x[0]+(-0.844516540906)*x[1]**o+(-0.0334629941654)*x[1]+(0.942164301071)*x[2]**o+(0.129008643344)*x[2]
        arg[1,0,1]=(-0.545329352463)*x[0]**o+(-0.0601985885543)*x[0]+(0.959752895562)*x[1]**o+(-0.617790988468)*x[1]+(-0.352323774737)*x[2]**o+(0.0854549755648)*x[2]
        arg[1,1,0]=(0.163270290732)*x[0]**o+(0.715418689409)*x[0]+(-0.892736094727)*x[1]**o+(-0.250085969902)*x[1]+(-0.323686607391)*x[2]**o+(-0.808822044566)*x[2]
        arg[1,1,1]=(-0.582858228848)*x[0]**o+(0.245356494646)*x[0]+(0.253867860186)*x[1]**o+(0.294533212339)*x[1]+(0.802766375942)*x[2]**o+(-0.142246487978)*x[2]
        arg[2,0,0]=(0.559682931749)*x[0]**o+(0.349781768104)*x[0]+(-0.903942828138)*x[1]**o+(0.285790428402)*x[1]+(0.334765408827)*x[2]**o+(-0.319553167552)*x[2]
        arg[2,0,1]=(-0.33958138616)*x[0]**o+(-0.476356251551)*x[0]+(0.222685919591)*x[1]**o+(0.947677379274)*x[1]+(-0.238198083505)*x[2]**o+(-0.158835521934)*x[2]
        arg[2,1,0]=(0.233005925554)*x[0]**o+(-0.0885827512025)*x[0]+(0.99298235997)*x[1]**o+(-0.972864413346)*x[1]+(-0.328337965331)*x[2]**o+(0.548449226427)*x[2]
        arg[2,1,1]=(-0.682434615287)*x[0]**o+(-0.681560497743)*x[0]+(-0.0427998840708)*x[1]**o+(-0.640532507401)*x[1]+(0.0600572641903)*x[2]**o+(0.580127097835)*x[2]
        arg[3,0,0]=(0.627556962303)*x[0]**o+(-0.2282646452)*x[0]+(0.201159933134)*x[1]**o+(-0.0874921144181)*x[1]+(0.487026890349)*x[2]**o+(-0.044256665504)*x[2]
        arg[3,0,1]=(-0.202969525043)*x[0]**o+(0.0445621391034)*x[0]+(-0.763479658456)*x[1]**o+(-0.600818445386)*x[1]+(-0.442800171339)*x[2]**o+(0.0422443400421)*x[2]
        arg[3,1,0]=(-0.422348404642)*x[0]**o+(0.723277196982)*x[0]+(0.882658191301)*x[1]**o+(-0.415218341669)*x[1]+(0.834911332561)*x[2]**o+(-0.843638850779)*x[2]
        arg[3,1,1]=(-0.685037113836)*x[0]**o+(0.119526696821)*x[0]+(0.732784440424)*x[1]**o+(0.156993935952)*x[1]+(-0.565837827825)*x[2]**o+(0.667635603478)*x[2]
        arg[4,0,0]=(-0.18210082527)*x[0]**o+(-0.00639006601355)*x[0]+(0.633771301081)*x[1]**o+(0.765450011643)*x[1]+(-0.323701472535)*x[2]**o+(0.841246809469)*x[2]
        arg[4,0,1]=(0.345847635107)*x[0]**o+(0.910917657816)*x[0]+(0.329667803527)*x[1]**o+(-0.0638797846725)*x[1]+(-0.705758696202)*x[2]**o+(0.39567870447)*x[2]
        arg[4,1,0]=(0.700805557734)*x[0]**o+(0.215376719066)*x[0]+(0.057212023986)*x[1]**o+(-0.601972647211)*x[1]+(-0.0163323501628)*x[2]**o+(-0.182314452762)*x[2]
        arg[4,1,1]=(-0.25633784052)*x[0]**o+(-0.36128209808)*x[0]+(-0.580660119759)*x[1]**o+(-0.718361413439)*x[1]+(0.294752937469)*x[2]**o+(0.479425008353)*x[2]
        arg[5,0,0]=(0.0250768411622)*x[0]**o+(-0.954437161073)*x[0]+(0.314623130846)*x[1]**o+(-0.915911802082)*x[1]+(-0.293481557155)*x[2]**o+(0.193040417512)*x[2]
        arg[5,0,1]=(-0.260552760825)*x[0]**o+(-0.993318746791)*x[0]+(-0.0888187740235)*x[1]**o+(0.228899511722)*x[1]+(0.667467747688)*x[2]**o+(0.210051895803)*x[2]
        arg[5,1,0]=(-0.925551425372)*x[0]**o+(-0.051063099903)*x[0]+(0.0607054862843)*x[1]**o+(0.778043999849)*x[1]+(-0.393429559505)*x[2]**o+(0.235480458425)*x[2]
        arg[5,1,1]=(0.00188621198193)*x[0]**o+(-0.529444982511)*x[0]+(0.766032106394)*x[1]**o+(-0.554844276584)*x[1]+(0.793956883291)*x[2]**o+(0.221253426455)*x[2]
        ref[0,0,0]=(-0.550852990025)*x_ref[0]**o+(0.578705971531)*x_ref[0]+(0.9533401901)*x_ref[1]**o+(0.233181910221)*x_ref[1]+(-0.143544272432)*x_ref[2]**o+(-0.986513253736)*x_ref[2]
        ref[0,0,1]=(-0.326417636931)*x_ref[0]**o+(0.0910236969351)*x_ref[0]+(-0.8240642668)*x_ref[1]**o+(-0.242359389714)*x_ref[1]+(0.266813299647)*x_ref[2]**o+(-0.364956839435)*x_ref[2]
        ref[0,1,0]=(-0.777107068999)*x_ref[0]**o+(0.201120303703)*x_ref[0]+(-0.684906707463)*x_ref[1]**o+(-0.9620762843)*x_ref[1]+(-0.314789913974)*x_ref[2]**o+(-0.70682592319)*x_ref[2]
        ref[0,1,1]=(-0.980292769796)*x_ref[0]**o+(-0.827763529178)*x_ref[0]+(0.256770279946)*x_ref[1]**o+(0.619665292369)*x_ref[1]+(0.014752087873)*x_ref[2]**o+(-0.810044916486)*x_ref[2]
        ref[1,0,0]=(-0.51164541516)*x_ref[0]**o+(-0.429646605692)*x_ref[0]+(-0.844516540906)*x_ref[1]**o+(-0.0334629941654)*x_ref[1]+(0.942164301071)*x_ref[2]**o+(0.129008643344)*x_ref[2]
        ref[1,0,1]=(-0.545329352463)*x_ref[0]**o+(-0.0601985885543)*x_ref[0]+(0.959752895562)*x_ref[1]**o+(-0.617790988468)*x_ref[1]+(-0.352323774737)*x_ref[2]**o+(0.0854549755648)*x_ref[2]
        ref[1,1,0]=(0.163270290732)*x_ref[0]**o+(0.715418689409)*x_ref[0]+(-0.892736094727)*x_ref[1]**o+(-0.250085969902)*x_ref[1]+(-0.323686607391)*x_ref[2]**o+(-0.808822044566)*x_ref[2]
        ref[1,1,1]=(-0.582858228848)*x_ref[0]**o+(0.245356494646)*x_ref[0]+(0.253867860186)*x_ref[1]**o+(0.294533212339)*x_ref[1]+(0.802766375942)*x_ref[2]**o+(-0.142246487978)*x_ref[2]
        ref[2,0,0]=(0.559682931749)*x_ref[0]**o+(0.349781768104)*x_ref[0]+(-0.903942828138)*x_ref[1]**o+(0.285790428402)*x_ref[1]+(0.334765408827)*x_ref[2]**o+(-0.319553167552)*x_ref[2]
        ref[2,0,1]=(-0.33958138616)*x_ref[0]**o+(-0.476356251551)*x_ref[0]+(0.222685919591)*x_ref[1]**o+(0.947677379274)*x_ref[1]+(-0.238198083505)*x_ref[2]**o+(-0.158835521934)*x_ref[2]
        ref[2,1,0]=(0.233005925554)*x_ref[0]**o+(-0.0885827512025)*x_ref[0]+(0.99298235997)*x_ref[1]**o+(-0.972864413346)*x_ref[1]+(-0.328337965331)*x_ref[2]**o+(0.548449226427)*x_ref[2]
        ref[2,1,1]=(-0.682434615287)*x_ref[0]**o+(-0.681560497743)*x_ref[0]+(-0.0427998840708)*x_ref[1]**o+(-0.640532507401)*x_ref[1]+(0.0600572641903)*x_ref[2]**o+(0.580127097835)*x_ref[2]
        ref[3,0,0]=(0.627556962303)*x_ref[0]**o+(-0.2282646452)*x_ref[0]+(0.201159933134)*x_ref[1]**o+(-0.0874921144181)*x_ref[1]+(0.487026890349)*x_ref[2]**o+(-0.044256665504)*x_ref[2]
        ref[3,0,1]=(-0.202969525043)*x_ref[0]**o+(0.0445621391034)*x_ref[0]+(-0.763479658456)*x_ref[1]**o+(-0.600818445386)*x_ref[1]+(-0.442800171339)*x_ref[2]**o+(0.0422443400421)*x_ref[2]
        ref[3,1,0]=(-0.422348404642)*x_ref[0]**o+(0.723277196982)*x_ref[0]+(0.882658191301)*x_ref[1]**o+(-0.415218341669)*x_ref[1]+(0.834911332561)*x_ref[2]**o+(-0.843638850779)*x_ref[2]
        ref[3,1,1]=(-0.685037113836)*x_ref[0]**o+(0.119526696821)*x_ref[0]+(0.732784440424)*x_ref[1]**o+(0.156993935952)*x_ref[1]+(-0.565837827825)*x_ref[2]**o+(0.667635603478)*x_ref[2]
        ref[4,0,0]=(-0.18210082527)*x_ref[0]**o+(-0.00639006601355)*x_ref[0]+(0.633771301081)*x_ref[1]**o+(0.765450011643)*x_ref[1]+(-0.323701472535)*x_ref[2]**o+(0.841246809469)*x_ref[2]
        ref[4,0,1]=(0.345847635107)*x_ref[0]**o+(0.910917657816)*x_ref[0]+(0.329667803527)*x_ref[1]**o+(-0.0638797846725)*x_ref[1]+(-0.705758696202)*x_ref[2]**o+(0.39567870447)*x_ref[2]
        ref[4,1,0]=(0.700805557734)*x_ref[0]**o+(0.215376719066)*x_ref[0]+(0.057212023986)*x_ref[1]**o+(-0.601972647211)*x_ref[1]+(-0.0163323501628)*x_ref[2]**o+(-0.182314452762)*x_ref[2]
        ref[4,1,1]=(-0.25633784052)*x_ref[0]**o+(-0.36128209808)*x_ref[0]+(-0.580660119759)*x_ref[1]**o+(-0.718361413439)*x_ref[1]+(0.294752937469)*x_ref[2]**o+(0.479425008353)*x_ref[2]
        ref[5,0,0]=(0.0250768411622)*x_ref[0]**o+(-0.954437161073)*x_ref[0]+(0.314623130846)*x_ref[1]**o+(-0.915911802082)*x_ref[1]+(-0.293481557155)*x_ref[2]**o+(0.193040417512)*x_ref[2]
        ref[5,0,1]=(-0.260552760825)*x_ref[0]**o+(-0.993318746791)*x_ref[0]+(-0.0888187740235)*x_ref[1]**o+(0.228899511722)*x_ref[1]+(0.667467747688)*x_ref[2]**o+(0.210051895803)*x_ref[2]
        ref[5,1,0]=(-0.925551425372)*x_ref[0]**o+(-0.051063099903)*x_ref[0]+(0.0607054862843)*x_ref[1]**o+(0.778043999849)*x_ref[1]+(-0.393429559505)*x_ref[2]**o+(0.235480458425)*x_ref[2]
        ref[5,1,1]=(0.00188621198193)*x_ref[0]**o+(-0.529444982511)*x_ref[0]+(0.766032106394)*x_ref[1]**o+(-0.554844276584)*x_ref[1]+(0.793956883291)*x_ref[2]**o+(0.221253426455)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.358768976574)*x[0]**o+(-0.231637053165)*x[0]+(0.798891506345)*x[1]**o+(0.292325388353)*x[1]
        arg[0,0,0,1]=(0.53301689922)*x[0]**o+(-0.761786870664)*x[0]+(0.284067900826)*x[1]**o+(-0.688662082205)*x[1]
        arg[0,0,1,0]=(-0.721988036113)*x[0]**o+(-0.743892757815)*x[0]+(-0.796787022375)*x[1]**o+(-0.499468412523)*x[1]
        arg[0,0,1,1]=(0.0710691488306)*x[0]**o+(0.0210385141456)*x[0]+(-0.709342417683)*x[1]**o+(0.901794071115)*x[1]
        arg[0,0,2,0]=(-0.800537263339)*x[0]**o+(0.583114786539)*x[0]+(0.422752622522)*x[1]**o+(-0.0954981460748)*x[1]
        arg[0,0,2,1]=(-0.0805357065445)*x[0]**o+(0.476377895547)*x[0]+(0.242203912566)*x[1]**o+(-0.479971540043)*x[1]
        arg[0,1,0,0]=(-0.476356823162)*x[0]**o+(-0.438853622252)*x[0]+(-0.937228141746)*x[1]**o+(0.366496742588)*x[1]
        arg[0,1,0,1]=(-0.496918253956)*x[0]**o+(-0.0227549061268)*x[0]+(0.830195584768)*x[1]**o+(-0.114624902423)*x[1]
        arg[0,1,1,0]=(0.431291036997)*x[0]**o+(0.807713720258)*x[0]+(0.730057543333)*x[1]**o+(0.635171725232)*x[1]
        arg[0,1,1,1]=(0.784425995411)*x[0]**o+(-0.397510494899)*x[0]+(-0.579911740115)*x[1]**o+(0.130426014607)*x[1]
        arg[0,1,2,0]=(0.473613650641)*x[0]**o+(-0.155617741177)*x[0]+(0.873237852103)*x[1]**o+(-0.749395152414)*x[1]
        arg[0,1,2,1]=(0.367649559431)*x[0]**o+(0.830181309644)*x[0]+(0.87726968147)*x[1]**o+(-0.450304793682)*x[1]
        arg[0,2,0,0]=(-0.700522876654)*x[0]**o+(-0.998187126003)*x[0]+(0.732491610173)*x[1]**o+(0.648679267)*x[1]
        arg[0,2,0,1]=(0.499746141285)*x[0]**o+(-0.730090975806)*x[0]+(-0.525566438654)*x[1]**o+(0.212469850421)*x[1]
        arg[0,2,1,0]=(-0.847752916298)*x[0]**o+(0.43510472619)*x[0]+(0.815305936548)*x[1]**o+(-0.143802676552)*x[1]
        arg[0,2,1,1]=(-0.453572376958)*x[0]**o+(-0.913023531929)*x[0]+(0.743328127665)*x[1]**o+(0.152586122591)*x[1]
        arg[0,2,2,0]=(0.526939334508)*x[0]**o+(0.247268722057)*x[0]+(-0.176432621481)*x[1]**o+(-0.510110202196)*x[1]
        arg[0,2,2,1]=(0.139685732003)*x[0]**o+(0.278304455867)*x[0]+(0.517953769616)*x[1]**o+(-0.272384261701)*x[1]
        arg[0,3,0,0]=(-0.38867907688)*x[0]**o+(-0.73043975055)*x[0]+(-0.14293243901)*x[1]**o+(0.0601042854043)*x[1]
        arg[0,3,0,1]=(-0.163055949188)*x[0]**o+(-0.035877062482)*x[0]+(0.0396615423106)*x[1]**o+(-0.870062303275)*x[1]
        arg[0,3,1,0]=(0.283839779914)*x[0]**o+(-0.519316925197)*x[0]+(-0.437401973959)*x[1]**o+(-0.823707232663)*x[1]
        arg[0,3,1,1]=(0.700305425892)*x[0]**o+(0.0299761521094)*x[0]+(-0.993032151025)*x[1]**o+(0.633293413221)*x[1]
        arg[0,3,2,0]=(-0.745357256365)*x[0]**o+(-0.451553456961)*x[0]+(0.05789646302)*x[1]**o+(-0.755901155251)*x[1]
        arg[0,3,2,1]=(-0.57330840403)*x[0]**o+(-0.589361483751)*x[0]+(-0.038677123513)*x[1]**o+(0.417332032168)*x[1]
        arg[0,4,0,0]=(0.759655161531)*x[0]**o+(0.737483461328)*x[0]+(0.573158629937)*x[1]**o+(-0.454191109106)*x[1]
        arg[0,4,0,1]=(-0.232982406645)*x[0]**o+(-0.926030038179)*x[0]+(-0.504256834465)*x[1]**o+(-0.107703527058)*x[1]
        arg[0,4,1,0]=(0.898411697016)*x[0]**o+(0.642045536742)*x[0]+(-0.953059791895)*x[1]**o+(-0.67501710093)*x[1]
        arg[0,4,1,1]=(0.300380574165)*x[0]**o+(0.204591960115)*x[0]+(-0.308849697562)*x[1]**o+(-0.489495887357)*x[1]
        arg[0,4,2,0]=(-0.814560286716)*x[0]**o+(0.154235599953)*x[0]+(0.279329273498)*x[1]**o+(-0.674063978926)*x[1]
        arg[0,4,2,1]=(-0.718018870365)*x[0]**o+(-0.735982858424)*x[0]+(-0.0132197132926)*x[1]**o+(0.398939587543)*x[1]
        arg[1,0,0,0]=(0.0559995351725)*x[0]**o+(-0.164549619892)*x[0]+(-0.794242529408)*x[1]**o+(0.0927822314785)*x[1]
        arg[1,0,0,1]=(-0.894238542341)*x[0]**o+(0.378559247473)*x[0]+(-0.183957272639)*x[1]**o+(0.486358968942)*x[1]
        arg[1,0,1,0]=(0.931288801001)*x[0]**o+(0.223815769968)*x[0]+(-0.719591906328)*x[1]**o+(-0.743644741105)*x[1]
        arg[1,0,1,1]=(0.156646029123)*x[0]**o+(0.306942879419)*x[0]+(0.747006830674)*x[1]**o+(-0.876534912274)*x[1]
        arg[1,0,2,0]=(-0.496819081999)*x[0]**o+(-0.939227402425)*x[0]+(0.0566331712182)*x[1]**o+(0.565724364661)*x[1]
        arg[1,0,2,1]=(0.903706429954)*x[0]**o+(0.667947320307)*x[0]+(0.97140855373)*x[1]**o+(0.924095235752)*x[1]
        arg[1,1,0,0]=(0.526180197299)*x[0]**o+(0.435493351674)*x[0]+(0.400107077848)*x[1]**o+(-0.704760833758)*x[1]
        arg[1,1,0,1]=(0.0723740015004)*x[0]**o+(-0.405859051371)*x[0]+(0.200787757511)*x[1]**o+(0.41819184563)*x[1]
        arg[1,1,1,0]=(-0.0320951655627)*x[0]**o+(0.235068445821)*x[0]+(0.450383349599)*x[1]**o+(-0.801487424765)*x[1]
        arg[1,1,1,1]=(-0.52478892744)*x[0]**o+(-0.449111297125)*x[0]+(0.731062182014)*x[1]**o+(0.533619608577)*x[1]
        arg[1,1,2,0]=(0.0889794057057)*x[0]**o+(-0.971856981571)*x[0]+(0.662216408922)*x[1]**o+(0.164089024468)*x[1]
        arg[1,1,2,1]=(-0.196374963647)*x[0]**o+(-0.739360211802)*x[0]+(0.916611650753)*x[1]**o+(-0.260102318115)*x[1]
        arg[1,2,0,0]=(0.627763572103)*x[0]**o+(-0.893900847319)*x[0]+(0.907548446417)*x[1]**o+(0.0687906203663)*x[1]
        arg[1,2,0,1]=(0.59755498682)*x[0]**o+(-0.457108009787)*x[0]+(0.71562139654)*x[1]**o+(-0.300872832559)*x[1]
        arg[1,2,1,0]=(-0.479641829834)*x[0]**o+(0.297188513722)*x[0]+(0.0227730978493)*x[1]**o+(0.199329665406)*x[1]
        arg[1,2,1,1]=(0.918974940101)*x[0]**o+(0.595194981816)*x[0]+(0.760030152554)*x[1]**o+(-0.798899878486)*x[1]
        arg[1,2,2,0]=(-0.878334511842)*x[0]**o+(0.910857236287)*x[0]+(0.0164747567997)*x[1]**o+(-0.876970888445)*x[1]
        arg[1,2,2,1]=(0.29445809181)*x[0]**o+(0.837849454931)*x[0]+(0.0759054371211)*x[1]**o+(0.501243080372)*x[1]
        arg[1,3,0,0]=(-0.529006436258)*x[0]**o+(-0.932325855538)*x[0]+(-0.561472972374)*x[1]**o+(0.845125736656)*x[1]
        arg[1,3,0,1]=(-0.843008257488)*x[0]**o+(0.614118623012)*x[0]+(-0.167015137447)*x[1]**o+(0.158701952625)*x[1]
        arg[1,3,1,0]=(-0.662966916148)*x[0]**o+(-0.443297576512)*x[0]+(0.295988608077)*x[1]**o+(0.059506908615)*x[1]
        arg[1,3,1,1]=(0.864199978014)*x[0]**o+(0.138723519602)*x[0]+(0.144964839302)*x[1]**o+(0.239863709389)*x[1]
        arg[1,3,2,0]=(0.479006097431)*x[0]**o+(0.542168797821)*x[0]+(0.561074562163)*x[1]**o+(-0.133365787275)*x[1]
        arg[1,3,2,1]=(-0.901033449452)*x[0]**o+(-0.620245150074)*x[0]+(0.598461569697)*x[1]**o+(-0.2213246854)*x[1]
        arg[1,4,0,0]=(-0.577067401348)*x[0]**o+(0.486154324874)*x[0]+(0.809205150442)*x[1]**o+(-0.218206418459)*x[1]
        arg[1,4,0,1]=(0.726556862066)*x[0]**o+(0.0293695509867)*x[0]+(0.601938348193)*x[1]**o+(0.967332697531)*x[1]
        arg[1,4,1,0]=(-0.675624388054)*x[0]**o+(-0.871613723214)*x[0]+(0.605519387693)*x[1]**o+(0.287186894411)*x[1]
        arg[1,4,1,1]=(-0.689904898214)*x[0]**o+(-0.918236705665)*x[0]+(-0.892168622177)*x[1]**o+(0.507441662612)*x[1]
        arg[1,4,2,0]=(0.845768980712)*x[0]**o+(-0.458467508114)*x[0]+(-0.518185249391)*x[1]**o+(0.829945330327)*x[1]
        arg[1,4,2,1]=(0.776852312178)*x[0]**o+(-0.268810999469)*x[0]+(-0.576757148841)*x[1]**o+(-0.50628294802)*x[1]
        arg[2,0,0,0]=(-0.686840219939)*x[0]**o+(0.6861172847)*x[0]+(0.240362263052)*x[1]**o+(0.962361523884)*x[1]
        arg[2,0,0,1]=(-0.943837188119)*x[0]**o+(0.829832768268)*x[0]+(0.377590124133)*x[1]**o+(-0.171714144962)*x[1]
        arg[2,0,1,0]=(0.862809148447)*x[0]**o+(0.0573402136489)*x[0]+(-0.743311826013)*x[1]**o+(0.427809313437)*x[1]
        arg[2,0,1,1]=(0.377380819576)*x[0]**o+(-0.223280285041)*x[0]+(0.178628658562)*x[1]**o+(0.121964611877)*x[1]
        arg[2,0,2,0]=(0.41723758689)*x[0]**o+(0.256039203883)*x[0]+(0.255358776676)*x[1]**o+(-0.881665964974)*x[1]
        arg[2,0,2,1]=(-0.0797103024499)*x[0]**o+(0.128068931414)*x[0]+(-0.758317834409)*x[1]**o+(0.654657809183)*x[1]
        arg[2,1,0,0]=(0.603648218133)*x[0]**o+(-0.286780270729)*x[0]+(-0.341108067214)*x[1]**o+(-0.769667356087)*x[1]
        arg[2,1,0,1]=(0.844044270035)*x[0]**o+(0.0585907857764)*x[0]+(-0.578232785252)*x[1]**o+(-0.40959017859)*x[1]
        arg[2,1,1,0]=(0.599709567393)*x[0]**o+(0.307092449739)*x[0]+(-0.402950586215)*x[1]**o+(0.0674239470451)*x[1]
        arg[2,1,1,1]=(0.182608249961)*x[0]**o+(0.373533584601)*x[0]+(0.40949869082)*x[1]**o+(0.696333617699)*x[1]
        arg[2,1,2,0]=(0.0462891111054)*x[0]**o+(0.371258293292)*x[0]+(0.782976187716)*x[1]**o+(0.117976017999)*x[1]
        arg[2,1,2,1]=(0.891841121526)*x[0]**o+(-0.998971173204)*x[0]+(0.0286712507928)*x[1]**o+(-0.0198507022516)*x[1]
        arg[2,2,0,0]=(0.284620902703)*x[0]**o+(0.936231203073)*x[0]+(0.829209752046)*x[1]**o+(-0.94940982125)*x[1]
        arg[2,2,0,1]=(-0.647956027376)*x[0]**o+(0.923690718922)*x[0]+(-0.0322423837346)*x[1]**o+(0.124962508606)*x[1]
        arg[2,2,1,0]=(-0.0696010411705)*x[0]**o+(-0.161672936739)*x[0]+(-0.997445762186)*x[1]**o+(0.177962302014)*x[1]
        arg[2,2,1,1]=(-0.897657700296)*x[0]**o+(-0.467866721175)*x[0]+(-0.560034332007)*x[1]**o+(-0.861509898295)*x[1]
        arg[2,2,2,0]=(-0.100880174048)*x[0]**o+(0.678568254695)*x[0]+(-0.83289411684)*x[1]**o+(0.680564393149)*x[1]
        arg[2,2,2,1]=(0.546159181491)*x[0]**o+(0.720296622966)*x[0]+(0.939046941118)*x[1]**o+(-0.716147554733)*x[1]
        arg[2,3,0,0]=(-0.921073162235)*x[0]**o+(0.979925882289)*x[0]+(-0.0938642966949)*x[1]**o+(0.482093047223)*x[1]
        arg[2,3,0,1]=(0.0709940956822)*x[0]**o+(-0.691401017243)*x[0]+(0.447000814029)*x[1]**o+(0.0575856240533)*x[1]
        arg[2,3,1,0]=(0.562800393365)*x[0]**o+(0.657646549915)*x[0]+(-0.835326361689)*x[1]**o+(0.3487960687)*x[1]
        arg[2,3,1,1]=(0.106668867409)*x[0]**o+(0.436327240765)*x[0]+(-0.585051261259)*x[1]**o+(0.616655338667)*x[1]
        arg[2,3,2,0]=(-0.501849628176)*x[0]**o+(-0.284722503916)*x[0]+(-0.892470199837)*x[1]**o+(-0.979942541795)*x[1]
        arg[2,3,2,1]=(0.236759540073)*x[0]**o+(-0.806931113769)*x[0]+(-0.225220003475)*x[1]**o+(-0.95183239844)*x[1]
        arg[2,4,0,0]=(-0.684823909581)*x[0]**o+(0.445615228059)*x[0]+(-0.638879895767)*x[1]**o+(-0.335505874088)*x[1]
        arg[2,4,0,1]=(-0.196708460612)*x[0]**o+(0.981592996809)*x[0]+(0.435774317028)*x[1]**o+(-0.0952200663109)*x[1]
        arg[2,4,1,0]=(-0.552759474764)*x[0]**o+(-0.614031173545)*x[0]+(0.304375542871)*x[1]**o+(-0.285131901533)*x[1]
        arg[2,4,1,1]=(0.147812951421)*x[0]**o+(-0.78533487681)*x[0]+(0.942205524401)*x[1]**o+(0.166280215075)*x[1]
        arg[2,4,2,0]=(0.843067982563)*x[0]**o+(0.0116520710289)*x[0]+(0.927948889059)*x[1]**o+(-0.395251919577)*x[1]
        arg[2,4,2,1]=(-0.316538010007)*x[0]**o+(-0.139262497126)*x[0]+(0.645409249217)*x[1]**o+(0.192949521992)*x[1]
        arg[3,0,0,0]=(0.88179290666)*x[0]**o+(0.337853287365)*x[0]+(-0.590219587193)*x[1]**o+(-0.827051504196)*x[1]
        arg[3,0,0,1]=(-0.104526336997)*x[0]**o+(0.299097012282)*x[0]+(-0.348493995853)*x[1]**o+(-0.780034322304)*x[1]
        arg[3,0,1,0]=(0.358724159445)*x[0]**o+(-0.0338089482488)*x[0]+(-0.0423054506456)*x[1]**o+(0.67180489557)*x[1]
        arg[3,0,1,1]=(-0.778880095665)*x[0]**o+(-0.448812173928)*x[0]+(-0.823686122382)*x[1]**o+(0.636028922656)*x[1]
        arg[3,0,2,0]=(0.146891002069)*x[0]**o+(-0.368903199628)*x[0]+(-0.596272077557)*x[1]**o+(0.703510683225)*x[1]
        arg[3,0,2,1]=(0.357416423547)*x[0]**o+(-0.103754905142)*x[0]+(0.32026011189)*x[1]**o+(0.755648597627)*x[1]
        arg[3,1,0,0]=(0.399082464238)*x[0]**o+(-0.12611921797)*x[0]+(0.426552225948)*x[1]**o+(-0.228714297286)*x[1]
        arg[3,1,0,1]=(-0.638411026717)*x[0]**o+(-0.342143922306)*x[0]+(-0.502107802626)*x[1]**o+(-0.881045304848)*x[1]
        arg[3,1,1,0]=(0.411134247295)*x[0]**o+(-0.216585727086)*x[0]+(0.270013339156)*x[1]**o+(0.251581243494)*x[1]
        arg[3,1,1,1]=(0.788758848294)*x[0]**o+(0.345181621902)*x[0]+(0.102843710598)*x[1]**o+(0.0561071077516)*x[1]
        arg[3,1,2,0]=(0.232441162347)*x[0]**o+(-0.687544250386)*x[0]+(0.369065746372)*x[1]**o+(0.628109564412)*x[1]
        arg[3,1,2,1]=(-0.912295044312)*x[0]**o+(0.847830497468)*x[0]+(0.776648618003)*x[1]**o+(0.522638388084)*x[1]
        arg[3,2,0,0]=(0.779333120942)*x[0]**o+(-0.0747119967046)*x[0]+(-0.594363282198)*x[1]**o+(0.513922159444)*x[1]
        arg[3,2,0,1]=(0.877492427404)*x[0]**o+(-0.191727825399)*x[0]+(0.0881296554818)*x[1]**o+(-0.408407503067)*x[1]
        arg[3,2,1,0]=(0.29696771215)*x[0]**o+(-0.914882727138)*x[0]+(0.684156831385)*x[1]**o+(0.142293411327)*x[1]
        arg[3,2,1,1]=(0.105927449304)*x[0]**o+(-0.863339661598)*x[0]+(-0.815550662607)*x[1]**o+(-0.261034715274)*x[1]
        arg[3,2,2,0]=(0.944082529871)*x[0]**o+(-0.449259627627)*x[0]+(-0.560526159016)*x[1]**o+(0.609050500798)*x[1]
        arg[3,2,2,1]=(0.529508870267)*x[0]**o+(0.627216712239)*x[0]+(-0.0148122228382)*x[1]**o+(0.0743274653452)*x[1]
        arg[3,3,0,0]=(-0.30609697448)*x[0]**o+(0.910972095956)*x[0]+(0.256274444924)*x[1]**o+(0.229096821635)*x[1]
        arg[3,3,0,1]=(-0.510323627078)*x[0]**o+(0.308418491165)*x[0]+(0.786418296678)*x[1]**o+(-0.248788086479)*x[1]
        arg[3,3,1,0]=(-0.981107944688)*x[0]**o+(-0.331637054173)*x[0]+(-0.972981902152)*x[1]**o+(-0.389364980465)*x[1]
        arg[3,3,1,1]=(-0.839880147367)*x[0]**o+(-0.911569972467)*x[0]+(0.209642332468)*x[1]**o+(0.0926875442137)*x[1]
        arg[3,3,2,0]=(-0.22724314378)*x[0]**o+(-0.0396604383619)*x[0]+(0.906829234649)*x[1]**o+(-0.862410550066)*x[1]
        arg[3,3,2,1]=(-0.412816401935)*x[0]**o+(-0.848442294241)*x[0]+(-0.599302206188)*x[1]**o+(0.719403009819)*x[1]
        arg[3,4,0,0]=(0.259283054664)*x[0]**o+(0.0983272120339)*x[0]+(-0.538373082787)*x[1]**o+(0.950164435617)*x[1]
        arg[3,4,0,1]=(0.633594236794)*x[0]**o+(0.120984469179)*x[0]+(0.472771297982)*x[1]**o+(0.873015785724)*x[1]
        arg[3,4,1,0]=(0.913524970057)*x[0]**o+(0.106590018424)*x[0]+(0.124026494061)*x[1]**o+(-0.254989594971)*x[1]
        arg[3,4,1,1]=(0.945528885492)*x[0]**o+(-0.323205519576)*x[0]+(-0.467294290307)*x[1]**o+(-0.557033486306)*x[1]
        arg[3,4,2,0]=(-0.268364682764)*x[0]**o+(0.0289685280025)*x[0]+(0.0493964636013)*x[1]**o+(0.872120137365)*x[1]
        arg[3,4,2,1]=(-0.266698796897)*x[0]**o+(-0.679148300702)*x[0]+(0.337653394795)*x[1]**o+(0.186866149716)*x[1]
        ref[0,0,0,0]=(0.358768976574)*x_ref[0]**o+(-0.231637053165)*x_ref[0]+(0.798891506345)*x_ref[1]**o+(0.292325388353)*x_ref[1]
        ref[0,0,0,1]=(0.53301689922)*x_ref[0]**o+(-0.761786870664)*x_ref[0]+(0.284067900826)*x_ref[1]**o+(-0.688662082205)*x_ref[1]
        ref[0,0,1,0]=(-0.721988036113)*x_ref[0]**o+(-0.743892757815)*x_ref[0]+(-0.796787022375)*x_ref[1]**o+(-0.499468412523)*x_ref[1]
        ref[0,0,1,1]=(0.0710691488306)*x_ref[0]**o+(0.0210385141456)*x_ref[0]+(-0.709342417683)*x_ref[1]**o+(0.901794071115)*x_ref[1]
        ref[0,0,2,0]=(-0.800537263339)*x_ref[0]**o+(0.583114786539)*x_ref[0]+(0.422752622522)*x_ref[1]**o+(-0.0954981460748)*x_ref[1]
        ref[0,0,2,1]=(-0.0805357065445)*x_ref[0]**o+(0.476377895547)*x_ref[0]+(0.242203912566)*x_ref[1]**o+(-0.479971540043)*x_ref[1]
        ref[0,1,0,0]=(-0.476356823162)*x_ref[0]**o+(-0.438853622252)*x_ref[0]+(-0.937228141746)*x_ref[1]**o+(0.366496742588)*x_ref[1]
        ref[0,1,0,1]=(-0.496918253956)*x_ref[0]**o+(-0.0227549061268)*x_ref[0]+(0.830195584768)*x_ref[1]**o+(-0.114624902423)*x_ref[1]
        ref[0,1,1,0]=(0.431291036997)*x_ref[0]**o+(0.807713720258)*x_ref[0]+(0.730057543333)*x_ref[1]**o+(0.635171725232)*x_ref[1]
        ref[0,1,1,1]=(0.784425995411)*x_ref[0]**o+(-0.397510494899)*x_ref[0]+(-0.579911740115)*x_ref[1]**o+(0.130426014607)*x_ref[1]
        ref[0,1,2,0]=(0.473613650641)*x_ref[0]**o+(-0.155617741177)*x_ref[0]+(0.873237852103)*x_ref[1]**o+(-0.749395152414)*x_ref[1]
        ref[0,1,2,1]=(0.367649559431)*x_ref[0]**o+(0.830181309644)*x_ref[0]+(0.87726968147)*x_ref[1]**o+(-0.450304793682)*x_ref[1]
        ref[0,2,0,0]=(-0.700522876654)*x_ref[0]**o+(-0.998187126003)*x_ref[0]+(0.732491610173)*x_ref[1]**o+(0.648679267)*x_ref[1]
        ref[0,2,0,1]=(0.499746141285)*x_ref[0]**o+(-0.730090975806)*x_ref[0]+(-0.525566438654)*x_ref[1]**o+(0.212469850421)*x_ref[1]
        ref[0,2,1,0]=(-0.847752916298)*x_ref[0]**o+(0.43510472619)*x_ref[0]+(0.815305936548)*x_ref[1]**o+(-0.143802676552)*x_ref[1]
        ref[0,2,1,1]=(-0.453572376958)*x_ref[0]**o+(-0.913023531929)*x_ref[0]+(0.743328127665)*x_ref[1]**o+(0.152586122591)*x_ref[1]
        ref[0,2,2,0]=(0.526939334508)*x_ref[0]**o+(0.247268722057)*x_ref[0]+(-0.176432621481)*x_ref[1]**o+(-0.510110202196)*x_ref[1]
        ref[0,2,2,1]=(0.139685732003)*x_ref[0]**o+(0.278304455867)*x_ref[0]+(0.517953769616)*x_ref[1]**o+(-0.272384261701)*x_ref[1]
        ref[0,3,0,0]=(-0.38867907688)*x_ref[0]**o+(-0.73043975055)*x_ref[0]+(-0.14293243901)*x_ref[1]**o+(0.0601042854043)*x_ref[1]
        ref[0,3,0,1]=(-0.163055949188)*x_ref[0]**o+(-0.035877062482)*x_ref[0]+(0.0396615423106)*x_ref[1]**o+(-0.870062303275)*x_ref[1]
        ref[0,3,1,0]=(0.283839779914)*x_ref[0]**o+(-0.519316925197)*x_ref[0]+(-0.437401973959)*x_ref[1]**o+(-0.823707232663)*x_ref[1]
        ref[0,3,1,1]=(0.700305425892)*x_ref[0]**o+(0.0299761521094)*x_ref[0]+(-0.993032151025)*x_ref[1]**o+(0.633293413221)*x_ref[1]
        ref[0,3,2,0]=(-0.745357256365)*x_ref[0]**o+(-0.451553456961)*x_ref[0]+(0.05789646302)*x_ref[1]**o+(-0.755901155251)*x_ref[1]
        ref[0,3,2,1]=(-0.57330840403)*x_ref[0]**o+(-0.589361483751)*x_ref[0]+(-0.038677123513)*x_ref[1]**o+(0.417332032168)*x_ref[1]
        ref[0,4,0,0]=(0.759655161531)*x_ref[0]**o+(0.737483461328)*x_ref[0]+(0.573158629937)*x_ref[1]**o+(-0.454191109106)*x_ref[1]
        ref[0,4,0,1]=(-0.232982406645)*x_ref[0]**o+(-0.926030038179)*x_ref[0]+(-0.504256834465)*x_ref[1]**o+(-0.107703527058)*x_ref[1]
        ref[0,4,1,0]=(0.898411697016)*x_ref[0]**o+(0.642045536742)*x_ref[0]+(-0.953059791895)*x_ref[1]**o+(-0.67501710093)*x_ref[1]
        ref[0,4,1,1]=(0.300380574165)*x_ref[0]**o+(0.204591960115)*x_ref[0]+(-0.308849697562)*x_ref[1]**o+(-0.489495887357)*x_ref[1]
        ref[0,4,2,0]=(-0.814560286716)*x_ref[0]**o+(0.154235599953)*x_ref[0]+(0.279329273498)*x_ref[1]**o+(-0.674063978926)*x_ref[1]
        ref[0,4,2,1]=(-0.718018870365)*x_ref[0]**o+(-0.735982858424)*x_ref[0]+(-0.0132197132926)*x_ref[1]**o+(0.398939587543)*x_ref[1]
        ref[1,0,0,0]=(0.0559995351725)*x_ref[0]**o+(-0.164549619892)*x_ref[0]+(-0.794242529408)*x_ref[1]**o+(0.0927822314785)*x_ref[1]
        ref[1,0,0,1]=(-0.894238542341)*x_ref[0]**o+(0.378559247473)*x_ref[0]+(-0.183957272639)*x_ref[1]**o+(0.486358968942)*x_ref[1]
        ref[1,0,1,0]=(0.931288801001)*x_ref[0]**o+(0.223815769968)*x_ref[0]+(-0.719591906328)*x_ref[1]**o+(-0.743644741105)*x_ref[1]
        ref[1,0,1,1]=(0.156646029123)*x_ref[0]**o+(0.306942879419)*x_ref[0]+(0.747006830674)*x_ref[1]**o+(-0.876534912274)*x_ref[1]
        ref[1,0,2,0]=(-0.496819081999)*x_ref[0]**o+(-0.939227402425)*x_ref[0]+(0.0566331712182)*x_ref[1]**o+(0.565724364661)*x_ref[1]
        ref[1,0,2,1]=(0.903706429954)*x_ref[0]**o+(0.667947320307)*x_ref[0]+(0.97140855373)*x_ref[1]**o+(0.924095235752)*x_ref[1]
        ref[1,1,0,0]=(0.526180197299)*x_ref[0]**o+(0.435493351674)*x_ref[0]+(0.400107077848)*x_ref[1]**o+(-0.704760833758)*x_ref[1]
        ref[1,1,0,1]=(0.0723740015004)*x_ref[0]**o+(-0.405859051371)*x_ref[0]+(0.200787757511)*x_ref[1]**o+(0.41819184563)*x_ref[1]
        ref[1,1,1,0]=(-0.0320951655627)*x_ref[0]**o+(0.235068445821)*x_ref[0]+(0.450383349599)*x_ref[1]**o+(-0.801487424765)*x_ref[1]
        ref[1,1,1,1]=(-0.52478892744)*x_ref[0]**o+(-0.449111297125)*x_ref[0]+(0.731062182014)*x_ref[1]**o+(0.533619608577)*x_ref[1]
        ref[1,1,2,0]=(0.0889794057057)*x_ref[0]**o+(-0.971856981571)*x_ref[0]+(0.662216408922)*x_ref[1]**o+(0.164089024468)*x_ref[1]
        ref[1,1,2,1]=(-0.196374963647)*x_ref[0]**o+(-0.739360211802)*x_ref[0]+(0.916611650753)*x_ref[1]**o+(-0.260102318115)*x_ref[1]
        ref[1,2,0,0]=(0.627763572103)*x_ref[0]**o+(-0.893900847319)*x_ref[0]+(0.907548446417)*x_ref[1]**o+(0.0687906203663)*x_ref[1]
        ref[1,2,0,1]=(0.59755498682)*x_ref[0]**o+(-0.457108009787)*x_ref[0]+(0.71562139654)*x_ref[1]**o+(-0.300872832559)*x_ref[1]
        ref[1,2,1,0]=(-0.479641829834)*x_ref[0]**o+(0.297188513722)*x_ref[0]+(0.0227730978493)*x_ref[1]**o+(0.199329665406)*x_ref[1]
        ref[1,2,1,1]=(0.918974940101)*x_ref[0]**o+(0.595194981816)*x_ref[0]+(0.760030152554)*x_ref[1]**o+(-0.798899878486)*x_ref[1]
        ref[1,2,2,0]=(-0.878334511842)*x_ref[0]**o+(0.910857236287)*x_ref[0]+(0.0164747567997)*x_ref[1]**o+(-0.876970888445)*x_ref[1]
        ref[1,2,2,1]=(0.29445809181)*x_ref[0]**o+(0.837849454931)*x_ref[0]+(0.0759054371211)*x_ref[1]**o+(0.501243080372)*x_ref[1]
        ref[1,3,0,0]=(-0.529006436258)*x_ref[0]**o+(-0.932325855538)*x_ref[0]+(-0.561472972374)*x_ref[1]**o+(0.845125736656)*x_ref[1]
        ref[1,3,0,1]=(-0.843008257488)*x_ref[0]**o+(0.614118623012)*x_ref[0]+(-0.167015137447)*x_ref[1]**o+(0.158701952625)*x_ref[1]
        ref[1,3,1,0]=(-0.662966916148)*x_ref[0]**o+(-0.443297576512)*x_ref[0]+(0.295988608077)*x_ref[1]**o+(0.059506908615)*x_ref[1]
        ref[1,3,1,1]=(0.864199978014)*x_ref[0]**o+(0.138723519602)*x_ref[0]+(0.144964839302)*x_ref[1]**o+(0.239863709389)*x_ref[1]
        ref[1,3,2,0]=(0.479006097431)*x_ref[0]**o+(0.542168797821)*x_ref[0]+(0.561074562163)*x_ref[1]**o+(-0.133365787275)*x_ref[1]
        ref[1,3,2,1]=(-0.901033449452)*x_ref[0]**o+(-0.620245150074)*x_ref[0]+(0.598461569697)*x_ref[1]**o+(-0.2213246854)*x_ref[1]
        ref[1,4,0,0]=(-0.577067401348)*x_ref[0]**o+(0.486154324874)*x_ref[0]+(0.809205150442)*x_ref[1]**o+(-0.218206418459)*x_ref[1]
        ref[1,4,0,1]=(0.726556862066)*x_ref[0]**o+(0.0293695509867)*x_ref[0]+(0.601938348193)*x_ref[1]**o+(0.967332697531)*x_ref[1]
        ref[1,4,1,0]=(-0.675624388054)*x_ref[0]**o+(-0.871613723214)*x_ref[0]+(0.605519387693)*x_ref[1]**o+(0.287186894411)*x_ref[1]
        ref[1,4,1,1]=(-0.689904898214)*x_ref[0]**o+(-0.918236705665)*x_ref[0]+(-0.892168622177)*x_ref[1]**o+(0.507441662612)*x_ref[1]
        ref[1,4,2,0]=(0.845768980712)*x_ref[0]**o+(-0.458467508114)*x_ref[0]+(-0.518185249391)*x_ref[1]**o+(0.829945330327)*x_ref[1]
        ref[1,4,2,1]=(0.776852312178)*x_ref[0]**o+(-0.268810999469)*x_ref[0]+(-0.576757148841)*x_ref[1]**o+(-0.50628294802)*x_ref[1]
        ref[2,0,0,0]=(-0.686840219939)*x_ref[0]**o+(0.6861172847)*x_ref[0]+(0.240362263052)*x_ref[1]**o+(0.962361523884)*x_ref[1]
        ref[2,0,0,1]=(-0.943837188119)*x_ref[0]**o+(0.829832768268)*x_ref[0]+(0.377590124133)*x_ref[1]**o+(-0.171714144962)*x_ref[1]
        ref[2,0,1,0]=(0.862809148447)*x_ref[0]**o+(0.0573402136489)*x_ref[0]+(-0.743311826013)*x_ref[1]**o+(0.427809313437)*x_ref[1]
        ref[2,0,1,1]=(0.377380819576)*x_ref[0]**o+(-0.223280285041)*x_ref[0]+(0.178628658562)*x_ref[1]**o+(0.121964611877)*x_ref[1]
        ref[2,0,2,0]=(0.41723758689)*x_ref[0]**o+(0.256039203883)*x_ref[0]+(0.255358776676)*x_ref[1]**o+(-0.881665964974)*x_ref[1]
        ref[2,0,2,1]=(-0.0797103024499)*x_ref[0]**o+(0.128068931414)*x_ref[0]+(-0.758317834409)*x_ref[1]**o+(0.654657809183)*x_ref[1]
        ref[2,1,0,0]=(0.603648218133)*x_ref[0]**o+(-0.286780270729)*x_ref[0]+(-0.341108067214)*x_ref[1]**o+(-0.769667356087)*x_ref[1]
        ref[2,1,0,1]=(0.844044270035)*x_ref[0]**o+(0.0585907857764)*x_ref[0]+(-0.578232785252)*x_ref[1]**o+(-0.40959017859)*x_ref[1]
        ref[2,1,1,0]=(0.599709567393)*x_ref[0]**o+(0.307092449739)*x_ref[0]+(-0.402950586215)*x_ref[1]**o+(0.0674239470451)*x_ref[1]
        ref[2,1,1,1]=(0.182608249961)*x_ref[0]**o+(0.373533584601)*x_ref[0]+(0.40949869082)*x_ref[1]**o+(0.696333617699)*x_ref[1]
        ref[2,1,2,0]=(0.0462891111054)*x_ref[0]**o+(0.371258293292)*x_ref[0]+(0.782976187716)*x_ref[1]**o+(0.117976017999)*x_ref[1]
        ref[2,1,2,1]=(0.891841121526)*x_ref[0]**o+(-0.998971173204)*x_ref[0]+(0.0286712507928)*x_ref[1]**o+(-0.0198507022516)*x_ref[1]
        ref[2,2,0,0]=(0.284620902703)*x_ref[0]**o+(0.936231203073)*x_ref[0]+(0.829209752046)*x_ref[1]**o+(-0.94940982125)*x_ref[1]
        ref[2,2,0,1]=(-0.647956027376)*x_ref[0]**o+(0.923690718922)*x_ref[0]+(-0.0322423837346)*x_ref[1]**o+(0.124962508606)*x_ref[1]
        ref[2,2,1,0]=(-0.0696010411705)*x_ref[0]**o+(-0.161672936739)*x_ref[0]+(-0.997445762186)*x_ref[1]**o+(0.177962302014)*x_ref[1]
        ref[2,2,1,1]=(-0.897657700296)*x_ref[0]**o+(-0.467866721175)*x_ref[0]+(-0.560034332007)*x_ref[1]**o+(-0.861509898295)*x_ref[1]
        ref[2,2,2,0]=(-0.100880174048)*x_ref[0]**o+(0.678568254695)*x_ref[0]+(-0.83289411684)*x_ref[1]**o+(0.680564393149)*x_ref[1]
        ref[2,2,2,1]=(0.546159181491)*x_ref[0]**o+(0.720296622966)*x_ref[0]+(0.939046941118)*x_ref[1]**o+(-0.716147554733)*x_ref[1]
        ref[2,3,0,0]=(-0.921073162235)*x_ref[0]**o+(0.979925882289)*x_ref[0]+(-0.0938642966949)*x_ref[1]**o+(0.482093047223)*x_ref[1]
        ref[2,3,0,1]=(0.0709940956822)*x_ref[0]**o+(-0.691401017243)*x_ref[0]+(0.447000814029)*x_ref[1]**o+(0.0575856240533)*x_ref[1]
        ref[2,3,1,0]=(0.562800393365)*x_ref[0]**o+(0.657646549915)*x_ref[0]+(-0.835326361689)*x_ref[1]**o+(0.3487960687)*x_ref[1]
        ref[2,3,1,1]=(0.106668867409)*x_ref[0]**o+(0.436327240765)*x_ref[0]+(-0.585051261259)*x_ref[1]**o+(0.616655338667)*x_ref[1]
        ref[2,3,2,0]=(-0.501849628176)*x_ref[0]**o+(-0.284722503916)*x_ref[0]+(-0.892470199837)*x_ref[1]**o+(-0.979942541795)*x_ref[1]
        ref[2,3,2,1]=(0.236759540073)*x_ref[0]**o+(-0.806931113769)*x_ref[0]+(-0.225220003475)*x_ref[1]**o+(-0.95183239844)*x_ref[1]
        ref[2,4,0,0]=(-0.684823909581)*x_ref[0]**o+(0.445615228059)*x_ref[0]+(-0.638879895767)*x_ref[1]**o+(-0.335505874088)*x_ref[1]
        ref[2,4,0,1]=(-0.196708460612)*x_ref[0]**o+(0.981592996809)*x_ref[0]+(0.435774317028)*x_ref[1]**o+(-0.0952200663109)*x_ref[1]
        ref[2,4,1,0]=(-0.552759474764)*x_ref[0]**o+(-0.614031173545)*x_ref[0]+(0.304375542871)*x_ref[1]**o+(-0.285131901533)*x_ref[1]
        ref[2,4,1,1]=(0.147812951421)*x_ref[0]**o+(-0.78533487681)*x_ref[0]+(0.942205524401)*x_ref[1]**o+(0.166280215075)*x_ref[1]
        ref[2,4,2,0]=(0.843067982563)*x_ref[0]**o+(0.0116520710289)*x_ref[0]+(0.927948889059)*x_ref[1]**o+(-0.395251919577)*x_ref[1]
        ref[2,4,2,1]=(-0.316538010007)*x_ref[0]**o+(-0.139262497126)*x_ref[0]+(0.645409249217)*x_ref[1]**o+(0.192949521992)*x_ref[1]
        ref[3,0,0,0]=(0.88179290666)*x_ref[0]**o+(0.337853287365)*x_ref[0]+(-0.590219587193)*x_ref[1]**o+(-0.827051504196)*x_ref[1]
        ref[3,0,0,1]=(-0.104526336997)*x_ref[0]**o+(0.299097012282)*x_ref[0]+(-0.348493995853)*x_ref[1]**o+(-0.780034322304)*x_ref[1]
        ref[3,0,1,0]=(0.358724159445)*x_ref[0]**o+(-0.0338089482488)*x_ref[0]+(-0.0423054506456)*x_ref[1]**o+(0.67180489557)*x_ref[1]
        ref[3,0,1,1]=(-0.778880095665)*x_ref[0]**o+(-0.448812173928)*x_ref[0]+(-0.823686122382)*x_ref[1]**o+(0.636028922656)*x_ref[1]
        ref[3,0,2,0]=(0.146891002069)*x_ref[0]**o+(-0.368903199628)*x_ref[0]+(-0.596272077557)*x_ref[1]**o+(0.703510683225)*x_ref[1]
        ref[3,0,2,1]=(0.357416423547)*x_ref[0]**o+(-0.103754905142)*x_ref[0]+(0.32026011189)*x_ref[1]**o+(0.755648597627)*x_ref[1]
        ref[3,1,0,0]=(0.399082464238)*x_ref[0]**o+(-0.12611921797)*x_ref[0]+(0.426552225948)*x_ref[1]**o+(-0.228714297286)*x_ref[1]
        ref[3,1,0,1]=(-0.638411026717)*x_ref[0]**o+(-0.342143922306)*x_ref[0]+(-0.502107802626)*x_ref[1]**o+(-0.881045304848)*x_ref[1]
        ref[3,1,1,0]=(0.411134247295)*x_ref[0]**o+(-0.216585727086)*x_ref[0]+(0.270013339156)*x_ref[1]**o+(0.251581243494)*x_ref[1]
        ref[3,1,1,1]=(0.788758848294)*x_ref[0]**o+(0.345181621902)*x_ref[0]+(0.102843710598)*x_ref[1]**o+(0.0561071077516)*x_ref[1]
        ref[3,1,2,0]=(0.232441162347)*x_ref[0]**o+(-0.687544250386)*x_ref[0]+(0.369065746372)*x_ref[1]**o+(0.628109564412)*x_ref[1]
        ref[3,1,2,1]=(-0.912295044312)*x_ref[0]**o+(0.847830497468)*x_ref[0]+(0.776648618003)*x_ref[1]**o+(0.522638388084)*x_ref[1]
        ref[3,2,0,0]=(0.779333120942)*x_ref[0]**o+(-0.0747119967046)*x_ref[0]+(-0.594363282198)*x_ref[1]**o+(0.513922159444)*x_ref[1]
        ref[3,2,0,1]=(0.877492427404)*x_ref[0]**o+(-0.191727825399)*x_ref[0]+(0.0881296554818)*x_ref[1]**o+(-0.408407503067)*x_ref[1]
        ref[3,2,1,0]=(0.29696771215)*x_ref[0]**o+(-0.914882727138)*x_ref[0]+(0.684156831385)*x_ref[1]**o+(0.142293411327)*x_ref[1]
        ref[3,2,1,1]=(0.105927449304)*x_ref[0]**o+(-0.863339661598)*x_ref[0]+(-0.815550662607)*x_ref[1]**o+(-0.261034715274)*x_ref[1]
        ref[3,2,2,0]=(0.944082529871)*x_ref[0]**o+(-0.449259627627)*x_ref[0]+(-0.560526159016)*x_ref[1]**o+(0.609050500798)*x_ref[1]
        ref[3,2,2,1]=(0.529508870267)*x_ref[0]**o+(0.627216712239)*x_ref[0]+(-0.0148122228382)*x_ref[1]**o+(0.0743274653452)*x_ref[1]
        ref[3,3,0,0]=(-0.30609697448)*x_ref[0]**o+(0.910972095956)*x_ref[0]+(0.256274444924)*x_ref[1]**o+(0.229096821635)*x_ref[1]
        ref[3,3,0,1]=(-0.510323627078)*x_ref[0]**o+(0.308418491165)*x_ref[0]+(0.786418296678)*x_ref[1]**o+(-0.248788086479)*x_ref[1]
        ref[3,3,1,0]=(-0.981107944688)*x_ref[0]**o+(-0.331637054173)*x_ref[0]+(-0.972981902152)*x_ref[1]**o+(-0.389364980465)*x_ref[1]
        ref[3,3,1,1]=(-0.839880147367)*x_ref[0]**o+(-0.911569972467)*x_ref[0]+(0.209642332468)*x_ref[1]**o+(0.0926875442137)*x_ref[1]
        ref[3,3,2,0]=(-0.22724314378)*x_ref[0]**o+(-0.0396604383619)*x_ref[0]+(0.906829234649)*x_ref[1]**o+(-0.862410550066)*x_ref[1]
        ref[3,3,2,1]=(-0.412816401935)*x_ref[0]**o+(-0.848442294241)*x_ref[0]+(-0.599302206188)*x_ref[1]**o+(0.719403009819)*x_ref[1]
        ref[3,4,0,0]=(0.259283054664)*x_ref[0]**o+(0.0983272120339)*x_ref[0]+(-0.538373082787)*x_ref[1]**o+(0.950164435617)*x_ref[1]
        ref[3,4,0,1]=(0.633594236794)*x_ref[0]**o+(0.120984469179)*x_ref[0]+(0.472771297982)*x_ref[1]**o+(0.873015785724)*x_ref[1]
        ref[3,4,1,0]=(0.913524970057)*x_ref[0]**o+(0.106590018424)*x_ref[0]+(0.124026494061)*x_ref[1]**o+(-0.254989594971)*x_ref[1]
        ref[3,4,1,1]=(0.945528885492)*x_ref[0]**o+(-0.323205519576)*x_ref[0]+(-0.467294290307)*x_ref[1]**o+(-0.557033486306)*x_ref[1]
        ref[3,4,2,0]=(-0.268364682764)*x_ref[0]**o+(0.0289685280025)*x_ref[0]+(0.0493964636013)*x_ref[1]**o+(0.872120137365)*x_ref[1]
        ref[3,4,2,1]=(-0.266698796897)*x_ref[0]**o+(-0.679148300702)*x_ref[0]+(0.337653394795)*x_ref[1]**o+(0.186866149716)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.928996629952)*x[0]**o+(-0.171383074114)*x[0]+(-0.324423798901)*x[1]**o+(-0.801354053962)*x[1]+(0.365256142808)*x[2]**o+(0.126262208857)*x[2]
        arg[0,0,0,1]=(0.326024843626)*x[0]**o+(-0.0091566240426)*x[0]+(-0.0396235158589)*x[1]**o+(-0.753302577837)*x[1]+(-0.38709916587)*x[2]**o+(-0.28660316595)*x[2]
        arg[0,0,1,0]=(0.755338255506)*x[0]**o+(0.851895434343)*x[0]+(0.419119464402)*x[1]**o+(0.670733085023)*x[1]+(0.819398310958)*x[2]**o+(0.476185335773)*x[2]
        arg[0,0,1,1]=(-0.981416259131)*x[0]**o+(-0.97000252267)*x[0]+(-0.132624451699)*x[1]**o+(-0.261428370649)*x[1]+(-0.349245951727)*x[2]**o+(-0.717343195877)*x[2]
        arg[0,0,2,0]=(-0.874826075266)*x[0]**o+(-0.399307905707)*x[0]+(0.350873599317)*x[1]**o+(0.15861651992)*x[1]+(0.396279933216)*x[2]**o+(0.92332515708)*x[2]
        arg[0,0,2,1]=(-0.183596683527)*x[0]**o+(-0.115194639348)*x[0]+(0.616111723618)*x[1]**o+(-0.402157657669)*x[1]+(0.44145594193)*x[2]**o+(0.686052407784)*x[2]
        arg[0,1,0,0]=(-0.669169178244)*x[0]**o+(0.835937801609)*x[0]+(0.468805302181)*x[1]**o+(-0.627292634023)*x[1]+(-0.674361231979)*x[2]**o+(-0.00929181785129)*x[2]
        arg[0,1,0,1]=(-0.568131655705)*x[0]**o+(0.690366535852)*x[0]+(-0.212524396914)*x[1]**o+(-0.767983769407)*x[1]+(0.477500128569)*x[2]**o+(-0.0190887598644)*x[2]
        arg[0,1,1,0]=(-0.374170635822)*x[0]**o+(0.0169009974558)*x[0]+(-0.162519103357)*x[1]**o+(0.786991144449)*x[1]+(0.0590294918131)*x[2]**o+(-0.213442509309)*x[2]
        arg[0,1,1,1]=(-0.558268657439)*x[0]**o+(0.584734933669)*x[0]+(-0.650790013689)*x[1]**o+(-0.192849048407)*x[1]+(0.505458636714)*x[2]**o+(-0.585174281371)*x[2]
        arg[0,1,2,0]=(-0.933243590775)*x[0]**o+(0.528060249767)*x[0]+(-0.763255746252)*x[1]**o+(0.771589118163)*x[1]+(-0.641287604338)*x[2]**o+(0.379566401585)*x[2]
        arg[0,1,2,1]=(-0.836976373935)*x[0]**o+(-0.969434110974)*x[0]+(0.758148372888)*x[1]**o+(-0.89212062535)*x[1]+(0.788693820015)*x[2]**o+(-0.620704367414)*x[2]
        arg[0,2,0,0]=(0.0163117360814)*x[0]**o+(-0.390867752068)*x[0]+(-0.725068297982)*x[1]**o+(-0.680312496646)*x[1]+(0.069273494969)*x[2]**o+(-0.576122275696)*x[2]
        arg[0,2,0,1]=(0.461771197614)*x[0]**o+(-0.767541404109)*x[0]+(-0.694279263622)*x[1]**o+(-0.721448812749)*x[1]+(0.248787896584)*x[2]**o+(0.237075401327)*x[2]
        arg[0,2,1,0]=(0.381987012965)*x[0]**o+(-0.288236233238)*x[0]+(-0.519012025084)*x[1]**o+(-0.256658293082)*x[1]+(0.807322167368)*x[2]**o+(-0.261916179859)*x[2]
        arg[0,2,1,1]=(0.257241267407)*x[0]**o+(-0.470612990721)*x[0]+(-0.208612656448)*x[1]**o+(0.839646962914)*x[1]+(0.916675150494)*x[2]**o+(-0.165468351005)*x[2]
        arg[0,2,2,0]=(0.92569232521)*x[0]**o+(0.809982458308)*x[0]+(-0.860530305336)*x[1]**o+(0.261687622817)*x[1]+(0.948003684651)*x[2]**o+(-0.473133742361)*x[2]
        arg[0,2,2,1]=(-0.113846814874)*x[0]**o+(0.680846459377)*x[0]+(0.920219149741)*x[1]**o+(0.172027554503)*x[1]+(0.475330041115)*x[2]**o+(-0.176218891429)*x[2]
        arg[0,3,0,0]=(-0.948638294297)*x[0]**o+(-0.0411621070703)*x[0]+(-0.294759271027)*x[1]**o+(-0.903733455086)*x[1]+(-0.995253733593)*x[2]**o+(0.106726960462)*x[2]
        arg[0,3,0,1]=(0.1592709614)*x[0]**o+(0.0683266445293)*x[0]+(0.336092506745)*x[1]**o+(-0.209047739881)*x[1]+(-0.330417243459)*x[2]**o+(-0.809242647755)*x[2]
        arg[0,3,1,0]=(-0.680637516359)*x[0]**o+(-0.713405815404)*x[0]+(-0.484624435043)*x[1]**o+(-0.279314462332)*x[1]+(-0.458856066504)*x[2]**o+(0.0588111305227)*x[2]
        arg[0,3,1,1]=(-0.853676338655)*x[0]**o+(-0.651539814195)*x[0]+(-0.109727277725)*x[1]**o+(0.0139006407942)*x[1]+(0.231943245426)*x[2]**o+(0.42972517233)*x[2]
        arg[0,3,2,0]=(-0.60799269765)*x[0]**o+(-0.802413269128)*x[0]+(0.151362498863)*x[1]**o+(-0.678226252875)*x[1]+(-0.867117685659)*x[2]**o+(-0.711982116071)*x[2]
        arg[0,3,2,1]=(-0.243012743941)*x[0]**o+(-0.692614579016)*x[0]+(0.0986817641914)*x[1]**o+(-0.495735862695)*x[1]+(-0.430040681653)*x[2]**o+(0.0108928722284)*x[2]
        arg[0,4,0,0]=(0.0989402181558)*x[0]**o+(-0.60096719882)*x[0]+(-0.324253928432)*x[1]**o+(-0.862723687559)*x[1]+(-0.34745416427)*x[2]**o+(0.936559074873)*x[2]
        arg[0,4,0,1]=(0.711021967566)*x[0]**o+(-0.939408401113)*x[0]+(-0.218100921633)*x[1]**o+(0.92835194875)*x[1]+(-0.353936913865)*x[2]**o+(-0.547266090928)*x[2]
        arg[0,4,1,0]=(0.172231614188)*x[0]**o+(-0.734622090671)*x[0]+(0.432686916647)*x[1]**o+(0.552595350492)*x[1]+(0.779662641076)*x[2]**o+(-0.0933220492779)*x[2]
        arg[0,4,1,1]=(0.996776716866)*x[0]**o+(-0.340786520686)*x[0]+(0.260075214171)*x[1]**o+(0.00385504852496)*x[1]+(0.377516772127)*x[2]**o+(-0.341751634181)*x[2]
        arg[0,4,2,0]=(-0.655642896664)*x[0]**o+(-0.840525173737)*x[0]+(-0.855343721564)*x[1]**o+(0.253558571524)*x[1]+(0.0463733059925)*x[2]**o+(-0.116359866912)*x[2]
        arg[0,4,2,1]=(0.131619661414)*x[0]**o+(-0.162067255954)*x[0]+(0.562536831367)*x[1]**o+(-0.940412472035)*x[1]+(-0.282265762783)*x[2]**o+(0.634711607923)*x[2]
        arg[1,0,0,0]=(0.484071808097)*x[0]**o+(0.566763524691)*x[0]+(-0.658614173653)*x[1]**o+(-0.140192232895)*x[1]+(-0.310114236409)*x[2]**o+(-0.918532401595)*x[2]
        arg[1,0,0,1]=(-0.81190380441)*x[0]**o+(0.644439878497)*x[0]+(0.479583403289)*x[1]**o+(-0.493318882779)*x[1]+(-0.424704492175)*x[2]**o+(0.700624317732)*x[2]
        arg[1,0,1,0]=(0.766650182435)*x[0]**o+(0.226365545129)*x[0]+(0.510828945971)*x[1]**o+(0.283221954498)*x[1]+(0.817303877886)*x[2]**o+(-0.245709044628)*x[2]
        arg[1,0,1,1]=(0.194787112664)*x[0]**o+(-0.879306736922)*x[0]+(0.185381291627)*x[1]**o+(-0.510692743975)*x[1]+(-0.422555014055)*x[2]**o+(0.556629919959)*x[2]
        arg[1,0,2,0]=(-0.958345828895)*x[0]**o+(-0.667735551013)*x[0]+(0.278271284225)*x[1]**o+(-0.0847995194757)*x[1]+(0.902184171441)*x[2]**o+(0.763125168854)*x[2]
        arg[1,0,2,1]=(-0.0766429670195)*x[0]**o+(0.894421139029)*x[0]+(-0.331560314101)*x[1]**o+(-0.0253297623997)*x[1]+(0.13406045967)*x[2]**o+(-0.326681111625)*x[2]
        arg[1,1,0,0]=(-0.979406855838)*x[0]**o+(0.528849909402)*x[0]+(-0.0486159476737)*x[1]**o+(-0.898392512695)*x[1]+(-0.678328327407)*x[2]**o+(0.644201802057)*x[2]
        arg[1,1,0,1]=(-0.977485554605)*x[0]**o+(0.794799941538)*x[0]+(-0.00237183848504)*x[1]**o+(-0.633497483149)*x[1]+(0.746593853568)*x[2]**o+(0.197741126365)*x[2]
        arg[1,1,1,0]=(0.379985506182)*x[0]**o+(0.69889144234)*x[0]+(-0.507293887409)*x[1]**o+(-0.727675545291)*x[1]+(-0.467064531693)*x[2]**o+(0.130273930089)*x[2]
        arg[1,1,1,1]=(-0.330627830819)*x[0]**o+(-0.0311542935889)*x[0]+(-0.0671494148749)*x[1]**o+(0.583605469163)*x[1]+(-0.172083996069)*x[2]**o+(0.666718011074)*x[2]
        arg[1,1,2,0]=(0.539618381063)*x[0]**o+(-0.546796237888)*x[0]+(0.857911750392)*x[1]**o+(0.119552731463)*x[1]+(0.329286820109)*x[2]**o+(0.637240649087)*x[2]
        arg[1,1,2,1]=(-0.828603071334)*x[0]**o+(0.861799149065)*x[0]+(0.615224040538)*x[1]**o+(-0.0868985543263)*x[1]+(-0.206168997809)*x[2]**o+(0.311103719286)*x[2]
        arg[1,2,0,0]=(0.695106867715)*x[0]**o+(-0.533303418727)*x[0]+(-0.393936955207)*x[1]**o+(-0.407635261686)*x[1]+(0.604346269147)*x[2]**o+(-0.10213061533)*x[2]
        arg[1,2,0,1]=(0.338035309929)*x[0]**o+(0.517135412671)*x[0]+(0.277863735829)*x[1]**o+(-0.703417079398)*x[1]+(0.369398635984)*x[2]**o+(-0.952905070601)*x[2]
        arg[1,2,1,0]=(0.937866603297)*x[0]**o+(0.436121631708)*x[0]+(0.568362098461)*x[1]**o+(0.223837722984)*x[1]+(-0.236896910941)*x[2]**o+(0.316114408308)*x[2]
        arg[1,2,1,1]=(0.285714491932)*x[0]**o+(0.633321537987)*x[0]+(0.953737237202)*x[1]**o+(0.732065555924)*x[1]+(-0.515759068197)*x[2]**o+(-0.519775334889)*x[2]
        arg[1,2,2,0]=(-0.0610222303479)*x[0]**o+(0.135974983944)*x[0]+(0.450213393447)*x[1]**o+(-0.44747541715)*x[1]+(-0.640660207187)*x[2]**o+(0.348457261119)*x[2]
        arg[1,2,2,1]=(0.131753033369)*x[0]**o+(-0.311359549998)*x[0]+(0.445707767235)*x[1]**o+(-0.710795569938)*x[1]+(-0.746038648477)*x[2]**o+(-0.210414773188)*x[2]
        arg[1,3,0,0]=(-0.323684061289)*x[0]**o+(-0.0937091220282)*x[0]+(-0.827884898323)*x[1]**o+(0.604442907568)*x[1]+(0.620923596379)*x[2]**o+(0.851758170763)*x[2]
        arg[1,3,0,1]=(-0.269239872573)*x[0]**o+(-0.603248280988)*x[0]+(0.603804530326)*x[1]**o+(-0.331910095046)*x[1]+(-0.653877236308)*x[2]**o+(0.115434865132)*x[2]
        arg[1,3,1,0]=(0.0531798412869)*x[0]**o+(0.292950346899)*x[0]+(-0.634376081823)*x[1]**o+(-0.775427771578)*x[1]+(-0.987113467149)*x[2]**o+(0.757980444447)*x[2]
        arg[1,3,1,1]=(0.757807292214)*x[0]**o+(0.81778478971)*x[0]+(-0.619873148405)*x[1]**o+(0.0129855659138)*x[1]+(-0.282861093681)*x[2]**o+(0.667211104045)*x[2]
        arg[1,3,2,0]=(-0.915205606853)*x[0]**o+(0.918704643459)*x[0]+(0.143238155504)*x[1]**o+(0.0285686377214)*x[1]+(-0.848712843243)*x[2]**o+(0.058647124862)*x[2]
        arg[1,3,2,1]=(0.441487712831)*x[0]**o+(-0.626878272507)*x[0]+(-0.862579662295)*x[1]**o+(0.204629243832)*x[1]+(0.964143347265)*x[2]**o+(-0.399125844282)*x[2]
        arg[1,4,0,0]=(-0.349791275126)*x[0]**o+(-0.140293844776)*x[0]+(-0.693879671138)*x[1]**o+(0.429290435637)*x[1]+(-0.892670314048)*x[2]**o+(-0.380777380409)*x[2]
        arg[1,4,0,1]=(0.528712424647)*x[0]**o+(0.449268002671)*x[0]+(-0.202472214961)*x[1]**o+(-0.374762208851)*x[1]+(-0.625590920648)*x[2]**o+(-0.000776205555531)*x[2]
        arg[1,4,1,0]=(0.00425881397184)*x[0]**o+(0.265831686248)*x[0]+(0.757626547866)*x[1]**o+(-0.934207064756)*x[1]+(0.541356932538)*x[2]**o+(0.217841597089)*x[2]
        arg[1,4,1,1]=(-0.02422207106)*x[0]**o+(-0.807018149115)*x[0]+(-0.402347046418)*x[1]**o+(-0.0548851919048)*x[1]+(0.850202044077)*x[2]**o+(-0.889576261979)*x[2]
        arg[1,4,2,0]=(0.487059068561)*x[0]**o+(0.338254525003)*x[0]+(-0.387019663154)*x[1]**o+(-0.436969552236)*x[1]+(-0.747690766603)*x[2]**o+(0.265854907003)*x[2]
        arg[1,4,2,1]=(0.687562221132)*x[0]**o+(0.4755782976)*x[0]+(-0.456287111804)*x[1]**o+(0.412138123928)*x[1]+(0.637920334647)*x[2]**o+(0.817007725525)*x[2]
        arg[2,0,0,0]=(-0.728166422776)*x[0]**o+(0.108904177195)*x[0]+(0.383848108635)*x[1]**o+(-0.19252707292)*x[1]+(-0.279319122495)*x[2]**o+(-0.0530727745585)*x[2]
        arg[2,0,0,1]=(0.380827452272)*x[0]**o+(-0.133776280687)*x[0]+(-0.174731551283)*x[1]**o+(-0.369231086652)*x[1]+(-0.0190771818046)*x[2]**o+(-0.331873068919)*x[2]
        arg[2,0,1,0]=(0.653644298826)*x[0]**o+(-0.916110182689)*x[0]+(0.0510855252492)*x[1]**o+(-0.934192949208)*x[1]+(0.572048462784)*x[2]**o+(-0.451401856863)*x[2]
        arg[2,0,1,1]=(-0.394958497502)*x[0]**o+(-0.690752972227)*x[0]+(-0.436352069349)*x[1]**o+(0.933524576573)*x[1]+(-0.524344161699)*x[2]**o+(0.526748792776)*x[2]
        arg[2,0,2,0]=(0.845649523782)*x[0]**o+(0.63927747558)*x[0]+(-0.534192218216)*x[1]**o+(0.714500925459)*x[1]+(0.156077814898)*x[2]**o+(-0.193712778792)*x[2]
        arg[2,0,2,1]=(0.809636640413)*x[0]**o+(-0.930654037481)*x[0]+(0.509441692942)*x[1]**o+(0.881563824585)*x[1]+(0.101333749057)*x[2]**o+(0.418791722704)*x[2]
        arg[2,1,0,0]=(0.289481281745)*x[0]**o+(-0.98279193802)*x[0]+(-0.443382747319)*x[1]**o+(-0.281710992472)*x[1]+(0.426349350391)*x[2]**o+(-0.870192121251)*x[2]
        arg[2,1,0,1]=(0.837046002753)*x[0]**o+(-0.0254166983371)*x[0]+(0.30178635814)*x[1]**o+(0.973915444924)*x[1]+(0.298631524734)*x[2]**o+(-0.422490970248)*x[2]
        arg[2,1,1,0]=(0.869799082394)*x[0]**o+(0.376436513628)*x[0]+(-0.74735679118)*x[1]**o+(0.477184335901)*x[1]+(-0.797472924916)*x[2]**o+(-0.568827873964)*x[2]
        arg[2,1,1,1]=(0.993441009779)*x[0]**o+(-0.264888299493)*x[0]+(0.861197560367)*x[1]**o+(-0.769058084144)*x[1]+(-0.86221947415)*x[2]**o+(-0.617024169672)*x[2]
        arg[2,1,2,0]=(-0.89303691749)*x[0]**o+(-0.390287412175)*x[0]+(-0.312883084291)*x[1]**o+(-0.432368502641)*x[1]+(-0.121476666357)*x[2]**o+(0.416361314288)*x[2]
        arg[2,1,2,1]=(-0.622214426504)*x[0]**o+(-0.595279134338)*x[0]+(0.000927173498715)*x[1]**o+(0.783001179117)*x[1]+(-0.956003738061)*x[2]**o+(0.482321214492)*x[2]
        arg[2,2,0,0]=(-0.0047994639068)*x[0]**o+(0.478805093492)*x[0]+(0.747825846519)*x[1]**o+(0.0756064430807)*x[1]+(-0.470534826525)*x[2]**o+(-0.421524571831)*x[2]
        arg[2,2,0,1]=(-0.0569198078126)*x[0]**o+(-0.36284856463)*x[0]+(-0.374343856016)*x[1]**o+(0.193470734451)*x[1]+(0.425350990791)*x[2]**o+(-0.903309757487)*x[2]
        arg[2,2,1,0]=(0.602075251229)*x[0]**o+(-0.558163518701)*x[0]+(-0.00146699467501)*x[1]**o+(0.437116756605)*x[1]+(-0.252696658318)*x[2]**o+(0.154017654456)*x[2]
        arg[2,2,1,1]=(0.226970288166)*x[0]**o+(0.844076458869)*x[0]+(0.231874137232)*x[1]**o+(-0.270233945335)*x[1]+(0.365979082276)*x[2]**o+(-0.937144802776)*x[2]
        arg[2,2,2,0]=(0.662440813722)*x[0]**o+(-0.600830429698)*x[0]+(0.248436486724)*x[1]**o+(0.73439963812)*x[1]+(-0.406317484259)*x[2]**o+(-0.406641354653)*x[2]
        arg[2,2,2,1]=(0.948630940404)*x[0]**o+(-0.695098513788)*x[0]+(-0.819843883615)*x[1]**o+(0.0479609386435)*x[1]+(-0.968033045837)*x[2]**o+(0.199490071238)*x[2]
        arg[2,3,0,0]=(0.967996364186)*x[0]**o+(0.562562580668)*x[0]+(0.660658005197)*x[1]**o+(-0.00126241998808)*x[1]+(-0.0188273819622)*x[2]**o+(-0.798370566647)*x[2]
        arg[2,3,0,1]=(0.108137383906)*x[0]**o+(0.230815090379)*x[0]+(-0.0817434914408)*x[1]**o+(0.256977068172)*x[1]+(0.323454379512)*x[2]**o+(0.0382049771532)*x[2]
        arg[2,3,1,0]=(-0.279443973124)*x[0]**o+(-0.948014830608)*x[0]+(-0.237787814838)*x[1]**o+(-0.880632864876)*x[1]+(0.305230465688)*x[2]**o+(0.722139326346)*x[2]
        arg[2,3,1,1]=(-0.192752669063)*x[0]**o+(0.763575125553)*x[0]+(-0.317096602795)*x[1]**o+(-0.241479822867)*x[1]+(0.643154065804)*x[2]**o+(0.903250562165)*x[2]
        arg[2,3,2,0]=(0.58153852366)*x[0]**o+(-0.131284078682)*x[0]+(0.803022288461)*x[1]**o+(0.862074208938)*x[1]+(0.934648335783)*x[2]**o+(0.720337200812)*x[2]
        arg[2,3,2,1]=(-0.866755142782)*x[0]**o+(0.299995782189)*x[0]+(0.830022439718)*x[1]**o+(0.787775630269)*x[1]+(-0.436980900353)*x[2]**o+(0.436719598692)*x[2]
        arg[2,4,0,0]=(-0.416946111972)*x[0]**o+(-0.342692819134)*x[0]+(-0.265712927184)*x[1]**o+(-0.503154497528)*x[1]+(0.436447837569)*x[2]**o+(-0.514226068117)*x[2]
        arg[2,4,0,1]=(-0.915995315837)*x[0]**o+(-0.562870475868)*x[0]+(0.158881525561)*x[1]**o+(-0.458461594412)*x[1]+(-0.117927681987)*x[2]**o+(-0.697751714189)*x[2]
        arg[2,4,1,0]=(-0.961291698815)*x[0]**o+(0.457487466485)*x[0]+(0.170987701633)*x[1]**o+(-0.14226156627)*x[1]+(-0.663462703057)*x[2]**o+(-0.0307934454283)*x[2]
        arg[2,4,1,1]=(0.588162542138)*x[0]**o+(-0.761783611043)*x[0]+(0.510241246326)*x[1]**o+(0.805053446677)*x[1]+(-0.851391511147)*x[2]**o+(-0.0976037736598)*x[2]
        arg[2,4,2,0]=(0.649047985858)*x[0]**o+(0.677435498282)*x[0]+(-0.620777977748)*x[1]**o+(-0.209440794016)*x[1]+(0.58009322661)*x[2]**o+(0.56317180284)*x[2]
        arg[2,4,2,1]=(-0.975596464331)*x[0]**o+(0.759364271658)*x[0]+(0.43142780149)*x[1]**o+(-0.701043203786)*x[1]+(-0.648159388134)*x[2]**o+(-0.129618261144)*x[2]
        arg[3,0,0,0]=(-0.0207100058532)*x[0]**o+(-0.829637327952)*x[0]+(-0.0326040806041)*x[1]**o+(0.652320712227)*x[1]+(-0.509435869246)*x[2]**o+(-0.66044392521)*x[2]
        arg[3,0,0,1]=(-0.312677151974)*x[0]**o+(0.0204494340128)*x[0]+(-0.989457740146)*x[1]**o+(-0.0606077687298)*x[1]+(-0.527267756606)*x[2]**o+(0.496006274201)*x[2]
        arg[3,0,1,0]=(0.842871894549)*x[0]**o+(-0.577687338895)*x[0]+(0.414019506131)*x[1]**o+(0.0410521331015)*x[1]+(-0.830482660184)*x[2]**o+(0.688692187619)*x[2]
        arg[3,0,1,1]=(-0.920070939376)*x[0]**o+(0.632972952755)*x[0]+(0.944186412149)*x[1]**o+(0.620319712728)*x[1]+(-0.171433727161)*x[2]**o+(0.267125744191)*x[2]
        arg[3,0,2,0]=(-0.0669535338492)*x[0]**o+(0.249181230746)*x[0]+(-0.147448553328)*x[1]**o+(-0.459701695839)*x[1]+(0.378458172384)*x[2]**o+(0.916594147889)*x[2]
        arg[3,0,2,1]=(0.0720111269432)*x[0]**o+(-0.390032127319)*x[0]+(-0.115565740734)*x[1]**o+(0.501210832815)*x[1]+(0.568129023658)*x[2]**o+(-0.303909080042)*x[2]
        arg[3,1,0,0]=(0.205675480731)*x[0]**o+(-0.843093328795)*x[0]+(-0.812415573635)*x[1]**o+(-0.706806177394)*x[1]+(-0.158442718286)*x[2]**o+(0.15187176613)*x[2]
        arg[3,1,0,1]=(0.149669308017)*x[0]**o+(0.487445503149)*x[0]+(-0.27973905833)*x[1]**o+(-0.504992462225)*x[1]+(-0.728166691522)*x[2]**o+(0.734846267391)*x[2]
        arg[3,1,1,0]=(0.0410578015251)*x[0]**o+(-0.525001495676)*x[0]+(-0.512685697472)*x[1]**o+(-0.284314686987)*x[1]+(0.888112049081)*x[2]**o+(0.425466793802)*x[2]
        arg[3,1,1,1]=(-0.374237047043)*x[0]**o+(-0.00078724690461)*x[0]+(0.849952487348)*x[1]**o+(-0.654945893799)*x[1]+(-0.536739132177)*x[2]**o+(0.20919294066)*x[2]
        arg[3,1,2,0]=(0.685749084068)*x[0]**o+(-0.730378478543)*x[0]+(-0.918072746761)*x[1]**o+(-0.706713872062)*x[1]+(-0.486100957526)*x[2]**o+(0.95736939257)*x[2]
        arg[3,1,2,1]=(-0.287859736321)*x[0]**o+(-0.271882215207)*x[0]+(-0.673357024623)*x[1]**o+(0.702867457627)*x[1]+(0.78445294416)*x[2]**o+(0.613888365259)*x[2]
        arg[3,2,0,0]=(0.568506008135)*x[0]**o+(-0.825947283734)*x[0]+(0.687369924271)*x[1]**o+(-0.866597782569)*x[1]+(-0.332810902355)*x[2]**o+(0.499502049139)*x[2]
        arg[3,2,0,1]=(0.757251773732)*x[0]**o+(0.293337812465)*x[0]+(0.686709285382)*x[1]**o+(-0.084997189402)*x[1]+(-0.338903817381)*x[2]**o+(-0.787646799698)*x[2]
        arg[3,2,1,0]=(-0.580327419983)*x[0]**o+(-0.603952498996)*x[0]+(0.523057988618)*x[1]**o+(-0.0835187394107)*x[1]+(0.459827887147)*x[2]**o+(0.689021201042)*x[2]
        arg[3,2,1,1]=(-0.41923416619)*x[0]**o+(0.0406426957237)*x[0]+(-0.908041008186)*x[1]**o+(0.382097138462)*x[1]+(0.66835706849)*x[2]**o+(0.355849275378)*x[2]
        arg[3,2,2,0]=(-0.133756767108)*x[0]**o+(-0.233743496333)*x[0]+(-0.661420120856)*x[1]**o+(0.788722576584)*x[1]+(-0.131764585707)*x[2]**o+(-0.781004479871)*x[2]
        arg[3,2,2,1]=(0.0308645139784)*x[0]**o+(-0.194986246344)*x[0]+(-0.267418187303)*x[1]**o+(0.332143733168)*x[1]+(0.513266431894)*x[2]**o+(0.594016981473)*x[2]
        arg[3,3,0,0]=(0.644314102613)*x[0]**o+(-0.977355279476)*x[0]+(0.516452323667)*x[1]**o+(0.172873579919)*x[1]+(0.778904819969)*x[2]**o+(0.559689750425)*x[2]
        arg[3,3,0,1]=(0.736135986201)*x[0]**o+(0.366918250826)*x[0]+(-0.0547007988818)*x[1]**o+(-0.830174436433)*x[1]+(0.613440124338)*x[2]**o+(0.574992469591)*x[2]
        arg[3,3,1,0]=(0.324504938146)*x[0]**o+(-0.915240176689)*x[0]+(-0.398105336249)*x[1]**o+(-0.526498343533)*x[1]+(0.43264573336)*x[2]**o+(0.393008086529)*x[2]
        arg[3,3,1,1]=(0.926367438323)*x[0]**o+(0.526364305462)*x[0]+(0.15596582236)*x[1]**o+(0.165158439711)*x[1]+(-0.508505348973)*x[2]**o+(-0.603637833992)*x[2]
        arg[3,3,2,0]=(0.661333046004)*x[0]**o+(0.425320373512)*x[0]+(-0.810809051016)*x[1]**o+(0.20446843023)*x[1]+(0.685136412413)*x[2]**o+(-0.0281127019894)*x[2]
        arg[3,3,2,1]=(0.851742499007)*x[0]**o+(-0.405506974975)*x[0]+(-0.296768983731)*x[1]**o+(0.389557354929)*x[1]+(0.335108341463)*x[2]**o+(0.0313094433351)*x[2]
        arg[3,4,0,0]=(0.0415035498649)*x[0]**o+(0.827106496882)*x[0]+(-0.481831487377)*x[1]**o+(-0.14935685817)*x[1]+(0.354765095566)*x[2]**o+(0.521898910137)*x[2]
        arg[3,4,0,1]=(-0.36328099035)*x[0]**o+(-0.243342386595)*x[0]+(0.0475935188475)*x[1]**o+(0.845308417697)*x[1]+(0.315259620082)*x[2]**o+(0.541313742759)*x[2]
        arg[3,4,1,0]=(-0.66364149672)*x[0]**o+(0.987505978353)*x[0]+(-0.337659775957)*x[1]**o+(0.197277151989)*x[1]+(-0.431260870605)*x[2]**o+(-0.803375909471)*x[2]
        arg[3,4,1,1]=(0.722104365428)*x[0]**o+(0.168416682879)*x[0]+(-0.415740544748)*x[1]**o+(-0.437836696476)*x[1]+(-0.280581752598)*x[2]**o+(0.434326371762)*x[2]
        arg[3,4,2,0]=(0.673144889684)*x[0]**o+(-0.0628529591764)*x[0]+(0.0161458704408)*x[1]**o+(-0.439966404278)*x[1]+(-0.963402070447)*x[2]**o+(0.308312287021)*x[2]
        arg[3,4,2,1]=(0.649131384887)*x[0]**o+(-0.825086043001)*x[0]+(0.204615530143)*x[1]**o+(0.00801105023644)*x[1]+(-0.234754758061)*x[2]**o+(0.659650083864)*x[2]
        ref[0,0,0,0]=(-0.928996629952)*x_ref[0]**o+(-0.171383074114)*x_ref[0]+(-0.324423798901)*x_ref[1]**o+(-0.801354053962)*x_ref[1]+(0.365256142808)*x_ref[2]**o+(0.126262208857)*x_ref[2]
        ref[0,0,0,1]=(0.326024843626)*x_ref[0]**o+(-0.0091566240426)*x_ref[0]+(-0.0396235158589)*x_ref[1]**o+(-0.753302577837)*x_ref[1]+(-0.38709916587)*x_ref[2]**o+(-0.28660316595)*x_ref[2]
        ref[0,0,1,0]=(0.755338255506)*x_ref[0]**o+(0.851895434343)*x_ref[0]+(0.419119464402)*x_ref[1]**o+(0.670733085023)*x_ref[1]+(0.819398310958)*x_ref[2]**o+(0.476185335773)*x_ref[2]
        ref[0,0,1,1]=(-0.981416259131)*x_ref[0]**o+(-0.97000252267)*x_ref[0]+(-0.132624451699)*x_ref[1]**o+(-0.261428370649)*x_ref[1]+(-0.349245951727)*x_ref[2]**o+(-0.717343195877)*x_ref[2]
        ref[0,0,2,0]=(-0.874826075266)*x_ref[0]**o+(-0.399307905707)*x_ref[0]+(0.350873599317)*x_ref[1]**o+(0.15861651992)*x_ref[1]+(0.396279933216)*x_ref[2]**o+(0.92332515708)*x_ref[2]
        ref[0,0,2,1]=(-0.183596683527)*x_ref[0]**o+(-0.115194639348)*x_ref[0]+(0.616111723618)*x_ref[1]**o+(-0.402157657669)*x_ref[1]+(0.44145594193)*x_ref[2]**o+(0.686052407784)*x_ref[2]
        ref[0,1,0,0]=(-0.669169178244)*x_ref[0]**o+(0.835937801609)*x_ref[0]+(0.468805302181)*x_ref[1]**o+(-0.627292634023)*x_ref[1]+(-0.674361231979)*x_ref[2]**o+(-0.00929181785129)*x_ref[2]
        ref[0,1,0,1]=(-0.568131655705)*x_ref[0]**o+(0.690366535852)*x_ref[0]+(-0.212524396914)*x_ref[1]**o+(-0.767983769407)*x_ref[1]+(0.477500128569)*x_ref[2]**o+(-0.0190887598644)*x_ref[2]
        ref[0,1,1,0]=(-0.374170635822)*x_ref[0]**o+(0.0169009974558)*x_ref[0]+(-0.162519103357)*x_ref[1]**o+(0.786991144449)*x_ref[1]+(0.0590294918131)*x_ref[2]**o+(-0.213442509309)*x_ref[2]
        ref[0,1,1,1]=(-0.558268657439)*x_ref[0]**o+(0.584734933669)*x_ref[0]+(-0.650790013689)*x_ref[1]**o+(-0.192849048407)*x_ref[1]+(0.505458636714)*x_ref[2]**o+(-0.585174281371)*x_ref[2]
        ref[0,1,2,0]=(-0.933243590775)*x_ref[0]**o+(0.528060249767)*x_ref[0]+(-0.763255746252)*x_ref[1]**o+(0.771589118163)*x_ref[1]+(-0.641287604338)*x_ref[2]**o+(0.379566401585)*x_ref[2]
        ref[0,1,2,1]=(-0.836976373935)*x_ref[0]**o+(-0.969434110974)*x_ref[0]+(0.758148372888)*x_ref[1]**o+(-0.89212062535)*x_ref[1]+(0.788693820015)*x_ref[2]**o+(-0.620704367414)*x_ref[2]
        ref[0,2,0,0]=(0.0163117360814)*x_ref[0]**o+(-0.390867752068)*x_ref[0]+(-0.725068297982)*x_ref[1]**o+(-0.680312496646)*x_ref[1]+(0.069273494969)*x_ref[2]**o+(-0.576122275696)*x_ref[2]
        ref[0,2,0,1]=(0.461771197614)*x_ref[0]**o+(-0.767541404109)*x_ref[0]+(-0.694279263622)*x_ref[1]**o+(-0.721448812749)*x_ref[1]+(0.248787896584)*x_ref[2]**o+(0.237075401327)*x_ref[2]
        ref[0,2,1,0]=(0.381987012965)*x_ref[0]**o+(-0.288236233238)*x_ref[0]+(-0.519012025084)*x_ref[1]**o+(-0.256658293082)*x_ref[1]+(0.807322167368)*x_ref[2]**o+(-0.261916179859)*x_ref[2]
        ref[0,2,1,1]=(0.257241267407)*x_ref[0]**o+(-0.470612990721)*x_ref[0]+(-0.208612656448)*x_ref[1]**o+(0.839646962914)*x_ref[1]+(0.916675150494)*x_ref[2]**o+(-0.165468351005)*x_ref[2]
        ref[0,2,2,0]=(0.92569232521)*x_ref[0]**o+(0.809982458308)*x_ref[0]+(-0.860530305336)*x_ref[1]**o+(0.261687622817)*x_ref[1]+(0.948003684651)*x_ref[2]**o+(-0.473133742361)*x_ref[2]
        ref[0,2,2,1]=(-0.113846814874)*x_ref[0]**o+(0.680846459377)*x_ref[0]+(0.920219149741)*x_ref[1]**o+(0.172027554503)*x_ref[1]+(0.475330041115)*x_ref[2]**o+(-0.176218891429)*x_ref[2]
        ref[0,3,0,0]=(-0.948638294297)*x_ref[0]**o+(-0.0411621070703)*x_ref[0]+(-0.294759271027)*x_ref[1]**o+(-0.903733455086)*x_ref[1]+(-0.995253733593)*x_ref[2]**o+(0.106726960462)*x_ref[2]
        ref[0,3,0,1]=(0.1592709614)*x_ref[0]**o+(0.0683266445293)*x_ref[0]+(0.336092506745)*x_ref[1]**o+(-0.209047739881)*x_ref[1]+(-0.330417243459)*x_ref[2]**o+(-0.809242647755)*x_ref[2]
        ref[0,3,1,0]=(-0.680637516359)*x_ref[0]**o+(-0.713405815404)*x_ref[0]+(-0.484624435043)*x_ref[1]**o+(-0.279314462332)*x_ref[1]+(-0.458856066504)*x_ref[2]**o+(0.0588111305227)*x_ref[2]
        ref[0,3,1,1]=(-0.853676338655)*x_ref[0]**o+(-0.651539814195)*x_ref[0]+(-0.109727277725)*x_ref[1]**o+(0.0139006407942)*x_ref[1]+(0.231943245426)*x_ref[2]**o+(0.42972517233)*x_ref[2]
        ref[0,3,2,0]=(-0.60799269765)*x_ref[0]**o+(-0.802413269128)*x_ref[0]+(0.151362498863)*x_ref[1]**o+(-0.678226252875)*x_ref[1]+(-0.867117685659)*x_ref[2]**o+(-0.711982116071)*x_ref[2]
        ref[0,3,2,1]=(-0.243012743941)*x_ref[0]**o+(-0.692614579016)*x_ref[0]+(0.0986817641914)*x_ref[1]**o+(-0.495735862695)*x_ref[1]+(-0.430040681653)*x_ref[2]**o+(0.0108928722284)*x_ref[2]
        ref[0,4,0,0]=(0.0989402181558)*x_ref[0]**o+(-0.60096719882)*x_ref[0]+(-0.324253928432)*x_ref[1]**o+(-0.862723687559)*x_ref[1]+(-0.34745416427)*x_ref[2]**o+(0.936559074873)*x_ref[2]
        ref[0,4,0,1]=(0.711021967566)*x_ref[0]**o+(-0.939408401113)*x_ref[0]+(-0.218100921633)*x_ref[1]**o+(0.92835194875)*x_ref[1]+(-0.353936913865)*x_ref[2]**o+(-0.547266090928)*x_ref[2]
        ref[0,4,1,0]=(0.172231614188)*x_ref[0]**o+(-0.734622090671)*x_ref[0]+(0.432686916647)*x_ref[1]**o+(0.552595350492)*x_ref[1]+(0.779662641076)*x_ref[2]**o+(-0.0933220492779)*x_ref[2]
        ref[0,4,1,1]=(0.996776716866)*x_ref[0]**o+(-0.340786520686)*x_ref[0]+(0.260075214171)*x_ref[1]**o+(0.00385504852496)*x_ref[1]+(0.377516772127)*x_ref[2]**o+(-0.341751634181)*x_ref[2]
        ref[0,4,2,0]=(-0.655642896664)*x_ref[0]**o+(-0.840525173737)*x_ref[0]+(-0.855343721564)*x_ref[1]**o+(0.253558571524)*x_ref[1]+(0.0463733059925)*x_ref[2]**o+(-0.116359866912)*x_ref[2]
        ref[0,4,2,1]=(0.131619661414)*x_ref[0]**o+(-0.162067255954)*x_ref[0]+(0.562536831367)*x_ref[1]**o+(-0.940412472035)*x_ref[1]+(-0.282265762783)*x_ref[2]**o+(0.634711607923)*x_ref[2]
        ref[1,0,0,0]=(0.484071808097)*x_ref[0]**o+(0.566763524691)*x_ref[0]+(-0.658614173653)*x_ref[1]**o+(-0.140192232895)*x_ref[1]+(-0.310114236409)*x_ref[2]**o+(-0.918532401595)*x_ref[2]
        ref[1,0,0,1]=(-0.81190380441)*x_ref[0]**o+(0.644439878497)*x_ref[0]+(0.479583403289)*x_ref[1]**o+(-0.493318882779)*x_ref[1]+(-0.424704492175)*x_ref[2]**o+(0.700624317732)*x_ref[2]
        ref[1,0,1,0]=(0.766650182435)*x_ref[0]**o+(0.226365545129)*x_ref[0]+(0.510828945971)*x_ref[1]**o+(0.283221954498)*x_ref[1]+(0.817303877886)*x_ref[2]**o+(-0.245709044628)*x_ref[2]
        ref[1,0,1,1]=(0.194787112664)*x_ref[0]**o+(-0.879306736922)*x_ref[0]+(0.185381291627)*x_ref[1]**o+(-0.510692743975)*x_ref[1]+(-0.422555014055)*x_ref[2]**o+(0.556629919959)*x_ref[2]
        ref[1,0,2,0]=(-0.958345828895)*x_ref[0]**o+(-0.667735551013)*x_ref[0]+(0.278271284225)*x_ref[1]**o+(-0.0847995194757)*x_ref[1]+(0.902184171441)*x_ref[2]**o+(0.763125168854)*x_ref[2]
        ref[1,0,2,1]=(-0.0766429670195)*x_ref[0]**o+(0.894421139029)*x_ref[0]+(-0.331560314101)*x_ref[1]**o+(-0.0253297623997)*x_ref[1]+(0.13406045967)*x_ref[2]**o+(-0.326681111625)*x_ref[2]
        ref[1,1,0,0]=(-0.979406855838)*x_ref[0]**o+(0.528849909402)*x_ref[0]+(-0.0486159476737)*x_ref[1]**o+(-0.898392512695)*x_ref[1]+(-0.678328327407)*x_ref[2]**o+(0.644201802057)*x_ref[2]
        ref[1,1,0,1]=(-0.977485554605)*x_ref[0]**o+(0.794799941538)*x_ref[0]+(-0.00237183848504)*x_ref[1]**o+(-0.633497483149)*x_ref[1]+(0.746593853568)*x_ref[2]**o+(0.197741126365)*x_ref[2]
        ref[1,1,1,0]=(0.379985506182)*x_ref[0]**o+(0.69889144234)*x_ref[0]+(-0.507293887409)*x_ref[1]**o+(-0.727675545291)*x_ref[1]+(-0.467064531693)*x_ref[2]**o+(0.130273930089)*x_ref[2]
        ref[1,1,1,1]=(-0.330627830819)*x_ref[0]**o+(-0.0311542935889)*x_ref[0]+(-0.0671494148749)*x_ref[1]**o+(0.583605469163)*x_ref[1]+(-0.172083996069)*x_ref[2]**o+(0.666718011074)*x_ref[2]
        ref[1,1,2,0]=(0.539618381063)*x_ref[0]**o+(-0.546796237888)*x_ref[0]+(0.857911750392)*x_ref[1]**o+(0.119552731463)*x_ref[1]+(0.329286820109)*x_ref[2]**o+(0.637240649087)*x_ref[2]
        ref[1,1,2,1]=(-0.828603071334)*x_ref[0]**o+(0.861799149065)*x_ref[0]+(0.615224040538)*x_ref[1]**o+(-0.0868985543263)*x_ref[1]+(-0.206168997809)*x_ref[2]**o+(0.311103719286)*x_ref[2]
        ref[1,2,0,0]=(0.695106867715)*x_ref[0]**o+(-0.533303418727)*x_ref[0]+(-0.393936955207)*x_ref[1]**o+(-0.407635261686)*x_ref[1]+(0.604346269147)*x_ref[2]**o+(-0.10213061533)*x_ref[2]
        ref[1,2,0,1]=(0.338035309929)*x_ref[0]**o+(0.517135412671)*x_ref[0]+(0.277863735829)*x_ref[1]**o+(-0.703417079398)*x_ref[1]+(0.369398635984)*x_ref[2]**o+(-0.952905070601)*x_ref[2]
        ref[1,2,1,0]=(0.937866603297)*x_ref[0]**o+(0.436121631708)*x_ref[0]+(0.568362098461)*x_ref[1]**o+(0.223837722984)*x_ref[1]+(-0.236896910941)*x_ref[2]**o+(0.316114408308)*x_ref[2]
        ref[1,2,1,1]=(0.285714491932)*x_ref[0]**o+(0.633321537987)*x_ref[0]+(0.953737237202)*x_ref[1]**o+(0.732065555924)*x_ref[1]+(-0.515759068197)*x_ref[2]**o+(-0.519775334889)*x_ref[2]
        ref[1,2,2,0]=(-0.0610222303479)*x_ref[0]**o+(0.135974983944)*x_ref[0]+(0.450213393447)*x_ref[1]**o+(-0.44747541715)*x_ref[1]+(-0.640660207187)*x_ref[2]**o+(0.348457261119)*x_ref[2]
        ref[1,2,2,1]=(0.131753033369)*x_ref[0]**o+(-0.311359549998)*x_ref[0]+(0.445707767235)*x_ref[1]**o+(-0.710795569938)*x_ref[1]+(-0.746038648477)*x_ref[2]**o+(-0.210414773188)*x_ref[2]
        ref[1,3,0,0]=(-0.323684061289)*x_ref[0]**o+(-0.0937091220282)*x_ref[0]+(-0.827884898323)*x_ref[1]**o+(0.604442907568)*x_ref[1]+(0.620923596379)*x_ref[2]**o+(0.851758170763)*x_ref[2]
        ref[1,3,0,1]=(-0.269239872573)*x_ref[0]**o+(-0.603248280988)*x_ref[0]+(0.603804530326)*x_ref[1]**o+(-0.331910095046)*x_ref[1]+(-0.653877236308)*x_ref[2]**o+(0.115434865132)*x_ref[2]
        ref[1,3,1,0]=(0.0531798412869)*x_ref[0]**o+(0.292950346899)*x_ref[0]+(-0.634376081823)*x_ref[1]**o+(-0.775427771578)*x_ref[1]+(-0.987113467149)*x_ref[2]**o+(0.757980444447)*x_ref[2]
        ref[1,3,1,1]=(0.757807292214)*x_ref[0]**o+(0.81778478971)*x_ref[0]+(-0.619873148405)*x_ref[1]**o+(0.0129855659138)*x_ref[1]+(-0.282861093681)*x_ref[2]**o+(0.667211104045)*x_ref[2]
        ref[1,3,2,0]=(-0.915205606853)*x_ref[0]**o+(0.918704643459)*x_ref[0]+(0.143238155504)*x_ref[1]**o+(0.0285686377214)*x_ref[1]+(-0.848712843243)*x_ref[2]**o+(0.058647124862)*x_ref[2]
        ref[1,3,2,1]=(0.441487712831)*x_ref[0]**o+(-0.626878272507)*x_ref[0]+(-0.862579662295)*x_ref[1]**o+(0.204629243832)*x_ref[1]+(0.964143347265)*x_ref[2]**o+(-0.399125844282)*x_ref[2]
        ref[1,4,0,0]=(-0.349791275126)*x_ref[0]**o+(-0.140293844776)*x_ref[0]+(-0.693879671138)*x_ref[1]**o+(0.429290435637)*x_ref[1]+(-0.892670314048)*x_ref[2]**o+(-0.380777380409)*x_ref[2]
        ref[1,4,0,1]=(0.528712424647)*x_ref[0]**o+(0.449268002671)*x_ref[0]+(-0.202472214961)*x_ref[1]**o+(-0.374762208851)*x_ref[1]+(-0.625590920648)*x_ref[2]**o+(-0.000776205555531)*x_ref[2]
        ref[1,4,1,0]=(0.00425881397184)*x_ref[0]**o+(0.265831686248)*x_ref[0]+(0.757626547866)*x_ref[1]**o+(-0.934207064756)*x_ref[1]+(0.541356932538)*x_ref[2]**o+(0.217841597089)*x_ref[2]
        ref[1,4,1,1]=(-0.02422207106)*x_ref[0]**o+(-0.807018149115)*x_ref[0]+(-0.402347046418)*x_ref[1]**o+(-0.0548851919048)*x_ref[1]+(0.850202044077)*x_ref[2]**o+(-0.889576261979)*x_ref[2]
        ref[1,4,2,0]=(0.487059068561)*x_ref[0]**o+(0.338254525003)*x_ref[0]+(-0.387019663154)*x_ref[1]**o+(-0.436969552236)*x_ref[1]+(-0.747690766603)*x_ref[2]**o+(0.265854907003)*x_ref[2]
        ref[1,4,2,1]=(0.687562221132)*x_ref[0]**o+(0.4755782976)*x_ref[0]+(-0.456287111804)*x_ref[1]**o+(0.412138123928)*x_ref[1]+(0.637920334647)*x_ref[2]**o+(0.817007725525)*x_ref[2]
        ref[2,0,0,0]=(-0.728166422776)*x_ref[0]**o+(0.108904177195)*x_ref[0]+(0.383848108635)*x_ref[1]**o+(-0.19252707292)*x_ref[1]+(-0.279319122495)*x_ref[2]**o+(-0.0530727745585)*x_ref[2]
        ref[2,0,0,1]=(0.380827452272)*x_ref[0]**o+(-0.133776280687)*x_ref[0]+(-0.174731551283)*x_ref[1]**o+(-0.369231086652)*x_ref[1]+(-0.0190771818046)*x_ref[2]**o+(-0.331873068919)*x_ref[2]
        ref[2,0,1,0]=(0.653644298826)*x_ref[0]**o+(-0.916110182689)*x_ref[0]+(0.0510855252492)*x_ref[1]**o+(-0.934192949208)*x_ref[1]+(0.572048462784)*x_ref[2]**o+(-0.451401856863)*x_ref[2]
        ref[2,0,1,1]=(-0.394958497502)*x_ref[0]**o+(-0.690752972227)*x_ref[0]+(-0.436352069349)*x_ref[1]**o+(0.933524576573)*x_ref[1]+(-0.524344161699)*x_ref[2]**o+(0.526748792776)*x_ref[2]
        ref[2,0,2,0]=(0.845649523782)*x_ref[0]**o+(0.63927747558)*x_ref[0]+(-0.534192218216)*x_ref[1]**o+(0.714500925459)*x_ref[1]+(0.156077814898)*x_ref[2]**o+(-0.193712778792)*x_ref[2]
        ref[2,0,2,1]=(0.809636640413)*x_ref[0]**o+(-0.930654037481)*x_ref[0]+(0.509441692942)*x_ref[1]**o+(0.881563824585)*x_ref[1]+(0.101333749057)*x_ref[2]**o+(0.418791722704)*x_ref[2]
        ref[2,1,0,0]=(0.289481281745)*x_ref[0]**o+(-0.98279193802)*x_ref[0]+(-0.443382747319)*x_ref[1]**o+(-0.281710992472)*x_ref[1]+(0.426349350391)*x_ref[2]**o+(-0.870192121251)*x_ref[2]
        ref[2,1,0,1]=(0.837046002753)*x_ref[0]**o+(-0.0254166983371)*x_ref[0]+(0.30178635814)*x_ref[1]**o+(0.973915444924)*x_ref[1]+(0.298631524734)*x_ref[2]**o+(-0.422490970248)*x_ref[2]
        ref[2,1,1,0]=(0.869799082394)*x_ref[0]**o+(0.376436513628)*x_ref[0]+(-0.74735679118)*x_ref[1]**o+(0.477184335901)*x_ref[1]+(-0.797472924916)*x_ref[2]**o+(-0.568827873964)*x_ref[2]
        ref[2,1,1,1]=(0.993441009779)*x_ref[0]**o+(-0.264888299493)*x_ref[0]+(0.861197560367)*x_ref[1]**o+(-0.769058084144)*x_ref[1]+(-0.86221947415)*x_ref[2]**o+(-0.617024169672)*x_ref[2]
        ref[2,1,2,0]=(-0.89303691749)*x_ref[0]**o+(-0.390287412175)*x_ref[0]+(-0.312883084291)*x_ref[1]**o+(-0.432368502641)*x_ref[1]+(-0.121476666357)*x_ref[2]**o+(0.416361314288)*x_ref[2]
        ref[2,1,2,1]=(-0.622214426504)*x_ref[0]**o+(-0.595279134338)*x_ref[0]+(0.000927173498715)*x_ref[1]**o+(0.783001179117)*x_ref[1]+(-0.956003738061)*x_ref[2]**o+(0.482321214492)*x_ref[2]
        ref[2,2,0,0]=(-0.0047994639068)*x_ref[0]**o+(0.478805093492)*x_ref[0]+(0.747825846519)*x_ref[1]**o+(0.0756064430807)*x_ref[1]+(-0.470534826525)*x_ref[2]**o+(-0.421524571831)*x_ref[2]
        ref[2,2,0,1]=(-0.0569198078126)*x_ref[0]**o+(-0.36284856463)*x_ref[0]+(-0.374343856016)*x_ref[1]**o+(0.193470734451)*x_ref[1]+(0.425350990791)*x_ref[2]**o+(-0.903309757487)*x_ref[2]
        ref[2,2,1,0]=(0.602075251229)*x_ref[0]**o+(-0.558163518701)*x_ref[0]+(-0.00146699467501)*x_ref[1]**o+(0.437116756605)*x_ref[1]+(-0.252696658318)*x_ref[2]**o+(0.154017654456)*x_ref[2]
        ref[2,2,1,1]=(0.226970288166)*x_ref[0]**o+(0.844076458869)*x_ref[0]+(0.231874137232)*x_ref[1]**o+(-0.270233945335)*x_ref[1]+(0.365979082276)*x_ref[2]**o+(-0.937144802776)*x_ref[2]
        ref[2,2,2,0]=(0.662440813722)*x_ref[0]**o+(-0.600830429698)*x_ref[0]+(0.248436486724)*x_ref[1]**o+(0.73439963812)*x_ref[1]+(-0.406317484259)*x_ref[2]**o+(-0.406641354653)*x_ref[2]
        ref[2,2,2,1]=(0.948630940404)*x_ref[0]**o+(-0.695098513788)*x_ref[0]+(-0.819843883615)*x_ref[1]**o+(0.0479609386435)*x_ref[1]+(-0.968033045837)*x_ref[2]**o+(0.199490071238)*x_ref[2]
        ref[2,3,0,0]=(0.967996364186)*x_ref[0]**o+(0.562562580668)*x_ref[0]+(0.660658005197)*x_ref[1]**o+(-0.00126241998808)*x_ref[1]+(-0.0188273819622)*x_ref[2]**o+(-0.798370566647)*x_ref[2]
        ref[2,3,0,1]=(0.108137383906)*x_ref[0]**o+(0.230815090379)*x_ref[0]+(-0.0817434914408)*x_ref[1]**o+(0.256977068172)*x_ref[1]+(0.323454379512)*x_ref[2]**o+(0.0382049771532)*x_ref[2]
        ref[2,3,1,0]=(-0.279443973124)*x_ref[0]**o+(-0.948014830608)*x_ref[0]+(-0.237787814838)*x_ref[1]**o+(-0.880632864876)*x_ref[1]+(0.305230465688)*x_ref[2]**o+(0.722139326346)*x_ref[2]
        ref[2,3,1,1]=(-0.192752669063)*x_ref[0]**o+(0.763575125553)*x_ref[0]+(-0.317096602795)*x_ref[1]**o+(-0.241479822867)*x_ref[1]+(0.643154065804)*x_ref[2]**o+(0.903250562165)*x_ref[2]
        ref[2,3,2,0]=(0.58153852366)*x_ref[0]**o+(-0.131284078682)*x_ref[0]+(0.803022288461)*x_ref[1]**o+(0.862074208938)*x_ref[1]+(0.934648335783)*x_ref[2]**o+(0.720337200812)*x_ref[2]
        ref[2,3,2,1]=(-0.866755142782)*x_ref[0]**o+(0.299995782189)*x_ref[0]+(0.830022439718)*x_ref[1]**o+(0.787775630269)*x_ref[1]+(-0.436980900353)*x_ref[2]**o+(0.436719598692)*x_ref[2]
        ref[2,4,0,0]=(-0.416946111972)*x_ref[0]**o+(-0.342692819134)*x_ref[0]+(-0.265712927184)*x_ref[1]**o+(-0.503154497528)*x_ref[1]+(0.436447837569)*x_ref[2]**o+(-0.514226068117)*x_ref[2]
        ref[2,4,0,1]=(-0.915995315837)*x_ref[0]**o+(-0.562870475868)*x_ref[0]+(0.158881525561)*x_ref[1]**o+(-0.458461594412)*x_ref[1]+(-0.117927681987)*x_ref[2]**o+(-0.697751714189)*x_ref[2]
        ref[2,4,1,0]=(-0.961291698815)*x_ref[0]**o+(0.457487466485)*x_ref[0]+(0.170987701633)*x_ref[1]**o+(-0.14226156627)*x_ref[1]+(-0.663462703057)*x_ref[2]**o+(-0.0307934454283)*x_ref[2]
        ref[2,4,1,1]=(0.588162542138)*x_ref[0]**o+(-0.761783611043)*x_ref[0]+(0.510241246326)*x_ref[1]**o+(0.805053446677)*x_ref[1]+(-0.851391511147)*x_ref[2]**o+(-0.0976037736598)*x_ref[2]
        ref[2,4,2,0]=(0.649047985858)*x_ref[0]**o+(0.677435498282)*x_ref[0]+(-0.620777977748)*x_ref[1]**o+(-0.209440794016)*x_ref[1]+(0.58009322661)*x_ref[2]**o+(0.56317180284)*x_ref[2]
        ref[2,4,2,1]=(-0.975596464331)*x_ref[0]**o+(0.759364271658)*x_ref[0]+(0.43142780149)*x_ref[1]**o+(-0.701043203786)*x_ref[1]+(-0.648159388134)*x_ref[2]**o+(-0.129618261144)*x_ref[2]
        ref[3,0,0,0]=(-0.0207100058532)*x_ref[0]**o+(-0.829637327952)*x_ref[0]+(-0.0326040806041)*x_ref[1]**o+(0.652320712227)*x_ref[1]+(-0.509435869246)*x_ref[2]**o+(-0.66044392521)*x_ref[2]
        ref[3,0,0,1]=(-0.312677151974)*x_ref[0]**o+(0.0204494340128)*x_ref[0]+(-0.989457740146)*x_ref[1]**o+(-0.0606077687298)*x_ref[1]+(-0.527267756606)*x_ref[2]**o+(0.496006274201)*x_ref[2]
        ref[3,0,1,0]=(0.842871894549)*x_ref[0]**o+(-0.577687338895)*x_ref[0]+(0.414019506131)*x_ref[1]**o+(0.0410521331015)*x_ref[1]+(-0.830482660184)*x_ref[2]**o+(0.688692187619)*x_ref[2]
        ref[3,0,1,1]=(-0.920070939376)*x_ref[0]**o+(0.632972952755)*x_ref[0]+(0.944186412149)*x_ref[1]**o+(0.620319712728)*x_ref[1]+(-0.171433727161)*x_ref[2]**o+(0.267125744191)*x_ref[2]
        ref[3,0,2,0]=(-0.0669535338492)*x_ref[0]**o+(0.249181230746)*x_ref[0]+(-0.147448553328)*x_ref[1]**o+(-0.459701695839)*x_ref[1]+(0.378458172384)*x_ref[2]**o+(0.916594147889)*x_ref[2]
        ref[3,0,2,1]=(0.0720111269432)*x_ref[0]**o+(-0.390032127319)*x_ref[0]+(-0.115565740734)*x_ref[1]**o+(0.501210832815)*x_ref[1]+(0.568129023658)*x_ref[2]**o+(-0.303909080042)*x_ref[2]
        ref[3,1,0,0]=(0.205675480731)*x_ref[0]**o+(-0.843093328795)*x_ref[0]+(-0.812415573635)*x_ref[1]**o+(-0.706806177394)*x_ref[1]+(-0.158442718286)*x_ref[2]**o+(0.15187176613)*x_ref[2]
        ref[3,1,0,1]=(0.149669308017)*x_ref[0]**o+(0.487445503149)*x_ref[0]+(-0.27973905833)*x_ref[1]**o+(-0.504992462225)*x_ref[1]+(-0.728166691522)*x_ref[2]**o+(0.734846267391)*x_ref[2]
        ref[3,1,1,0]=(0.0410578015251)*x_ref[0]**o+(-0.525001495676)*x_ref[0]+(-0.512685697472)*x_ref[1]**o+(-0.284314686987)*x_ref[1]+(0.888112049081)*x_ref[2]**o+(0.425466793802)*x_ref[2]
        ref[3,1,1,1]=(-0.374237047043)*x_ref[0]**o+(-0.00078724690461)*x_ref[0]+(0.849952487348)*x_ref[1]**o+(-0.654945893799)*x_ref[1]+(-0.536739132177)*x_ref[2]**o+(0.20919294066)*x_ref[2]
        ref[3,1,2,0]=(0.685749084068)*x_ref[0]**o+(-0.730378478543)*x_ref[0]+(-0.918072746761)*x_ref[1]**o+(-0.706713872062)*x_ref[1]+(-0.486100957526)*x_ref[2]**o+(0.95736939257)*x_ref[2]
        ref[3,1,2,1]=(-0.287859736321)*x_ref[0]**o+(-0.271882215207)*x_ref[0]+(-0.673357024623)*x_ref[1]**o+(0.702867457627)*x_ref[1]+(0.78445294416)*x_ref[2]**o+(0.613888365259)*x_ref[2]
        ref[3,2,0,0]=(0.568506008135)*x_ref[0]**o+(-0.825947283734)*x_ref[0]+(0.687369924271)*x_ref[1]**o+(-0.866597782569)*x_ref[1]+(-0.332810902355)*x_ref[2]**o+(0.499502049139)*x_ref[2]
        ref[3,2,0,1]=(0.757251773732)*x_ref[0]**o+(0.293337812465)*x_ref[0]+(0.686709285382)*x_ref[1]**o+(-0.084997189402)*x_ref[1]+(-0.338903817381)*x_ref[2]**o+(-0.787646799698)*x_ref[2]
        ref[3,2,1,0]=(-0.580327419983)*x_ref[0]**o+(-0.603952498996)*x_ref[0]+(0.523057988618)*x_ref[1]**o+(-0.0835187394107)*x_ref[1]+(0.459827887147)*x_ref[2]**o+(0.689021201042)*x_ref[2]
        ref[3,2,1,1]=(-0.41923416619)*x_ref[0]**o+(0.0406426957237)*x_ref[0]+(-0.908041008186)*x_ref[1]**o+(0.382097138462)*x_ref[1]+(0.66835706849)*x_ref[2]**o+(0.355849275378)*x_ref[2]
        ref[3,2,2,0]=(-0.133756767108)*x_ref[0]**o+(-0.233743496333)*x_ref[0]+(-0.661420120856)*x_ref[1]**o+(0.788722576584)*x_ref[1]+(-0.131764585707)*x_ref[2]**o+(-0.781004479871)*x_ref[2]
        ref[3,2,2,1]=(0.0308645139784)*x_ref[0]**o+(-0.194986246344)*x_ref[0]+(-0.267418187303)*x_ref[1]**o+(0.332143733168)*x_ref[1]+(0.513266431894)*x_ref[2]**o+(0.594016981473)*x_ref[2]
        ref[3,3,0,0]=(0.644314102613)*x_ref[0]**o+(-0.977355279476)*x_ref[0]+(0.516452323667)*x_ref[1]**o+(0.172873579919)*x_ref[1]+(0.778904819969)*x_ref[2]**o+(0.559689750425)*x_ref[2]
        ref[3,3,0,1]=(0.736135986201)*x_ref[0]**o+(0.366918250826)*x_ref[0]+(-0.0547007988818)*x_ref[1]**o+(-0.830174436433)*x_ref[1]+(0.613440124338)*x_ref[2]**o+(0.574992469591)*x_ref[2]
        ref[3,3,1,0]=(0.324504938146)*x_ref[0]**o+(-0.915240176689)*x_ref[0]+(-0.398105336249)*x_ref[1]**o+(-0.526498343533)*x_ref[1]+(0.43264573336)*x_ref[2]**o+(0.393008086529)*x_ref[2]
        ref[3,3,1,1]=(0.926367438323)*x_ref[0]**o+(0.526364305462)*x_ref[0]+(0.15596582236)*x_ref[1]**o+(0.165158439711)*x_ref[1]+(-0.508505348973)*x_ref[2]**o+(-0.603637833992)*x_ref[2]
        ref[3,3,2,0]=(0.661333046004)*x_ref[0]**o+(0.425320373512)*x_ref[0]+(-0.810809051016)*x_ref[1]**o+(0.20446843023)*x_ref[1]+(0.685136412413)*x_ref[2]**o+(-0.0281127019894)*x_ref[2]
        ref[3,3,2,1]=(0.851742499007)*x_ref[0]**o+(-0.405506974975)*x_ref[0]+(-0.296768983731)*x_ref[1]**o+(0.389557354929)*x_ref[1]+(0.335108341463)*x_ref[2]**o+(0.0313094433351)*x_ref[2]
        ref[3,4,0,0]=(0.0415035498649)*x_ref[0]**o+(0.827106496882)*x_ref[0]+(-0.481831487377)*x_ref[1]**o+(-0.14935685817)*x_ref[1]+(0.354765095566)*x_ref[2]**o+(0.521898910137)*x_ref[2]
        ref[3,4,0,1]=(-0.36328099035)*x_ref[0]**o+(-0.243342386595)*x_ref[0]+(0.0475935188475)*x_ref[1]**o+(0.845308417697)*x_ref[1]+(0.315259620082)*x_ref[2]**o+(0.541313742759)*x_ref[2]
        ref[3,4,1,0]=(-0.66364149672)*x_ref[0]**o+(0.987505978353)*x_ref[0]+(-0.337659775957)*x_ref[1]**o+(0.197277151989)*x_ref[1]+(-0.431260870605)*x_ref[2]**o+(-0.803375909471)*x_ref[2]
        ref[3,4,1,1]=(0.722104365428)*x_ref[0]**o+(0.168416682879)*x_ref[0]+(-0.415740544748)*x_ref[1]**o+(-0.437836696476)*x_ref[1]+(-0.280581752598)*x_ref[2]**o+(0.434326371762)*x_ref[2]
        ref[3,4,2,0]=(0.673144889684)*x_ref[0]**o+(-0.0628529591764)*x_ref[0]+(0.0161458704408)*x_ref[1]**o+(-0.439966404278)*x_ref[1]+(-0.963402070447)*x_ref[2]**o+(0.308312287021)*x_ref[2]
        ref[3,4,2,1]=(0.649131384887)*x_ref[0]**o+(-0.825086043001)*x_ref[0]+(0.204615530143)*x_ref[1]**o+(0.00801105023644)*x_ref[1]+(-0.234754758061)*x_ref[2]**o+(0.659650083864)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.935068705503)*x[0]**o+(-0.843768568975)*x[0]+(0.220151879481)*x[1]**o+(0.629247769917)*x[1]
        ref=(-0.935068705503)*x_ref[0]**o+(-0.843768568975)*x_ref[0]+(0.220151879481)*x_ref[1]**o+(0.629247769917)*x_ref[1]
      else:
        arg=(0.204715133488)*x[0]**o+(-0.933589384176)*x[0]+(-0.651477613601)*x[1]**o+(-0.128099401086)*x[1]+(0.0507915334433)*x[2]**o+(0.63225604515)*x[2]
        ref=(0.204715133488)*x_ref[0]**o+(-0.933589384176)*x_ref[0]+(-0.651477613601)*x_ref[1]**o+(-0.128099401086)*x_ref[1]+(0.0507915334433)*x_ref[2]**o+(0.63225604515)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.693671772608)*x[0]**o+(0.260200526395)*x[0]+(-0.99202032752)*x[1]**o+(0.698503413133)*x[1]
        arg[1]=(-0.549448913905)*x[0]**o+(-0.728769679965)*x[0]+(0.415132041969)*x[1]**o+(-0.106391482954)*x[1]
        ref[0]=(0.693671772608)*x_ref[0]**o+(0.260200526395)*x_ref[0]+(-0.99202032752)*x_ref[1]**o+(0.698503413133)*x_ref[1]
        ref[1]=(-0.549448913905)*x_ref[0]**o+(-0.728769679965)*x_ref[0]+(0.415132041969)*x_ref[1]**o+(-0.106391482954)*x_ref[1]
      else:
        arg[0]=(0.479813279739)*x[0]**o+(0.940795516779)*x[0]+(-0.671143421639)*x[1]**o+(-0.485412476854)*x[1]+(-0.962114397463)*x[2]**o+(0.0250734884375)*x[2]
        arg[1]=(0.703481492279)*x[0]**o+(0.298972367312)*x[0]+(0.22386003834)*x[1]**o+(-0.726486516907)*x[1]+(0.0556568111911)*x[2]**o+(-0.603711754869)*x[2]
        ref[0]=(0.479813279739)*x_ref[0]**o+(0.940795516779)*x_ref[0]+(-0.671143421639)*x_ref[1]**o+(-0.485412476854)*x_ref[1]+(-0.962114397463)*x_ref[2]**o+(0.0250734884375)*x_ref[2]
        ref[1]=(0.703481492279)*x_ref[0]**o+(0.298972367312)*x_ref[0]+(0.22386003834)*x_ref[1]**o+(-0.726486516907)*x_ref[1]+(0.0556568111911)*x_ref[2]**o+(-0.603711754869)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.436439283545)*x[0]**o+(0.705849609672)*x[0]+(-0.261476237133)*x[1]**o+(-0.10504570237)*x[1]
        arg[0,1]=(0.035208834696)*x[0]**o+(0.590058421274)*x[0]+(-0.532905126944)*x[1]**o+(0.126572388319)*x[1]
        arg[0,2]=(0.985266120604)*x[0]**o+(-0.465951670003)*x[0]+(0.855881157118)*x[1]**o+(0.00932326399668)*x[1]
        arg[0,3]=(0.867525798771)*x[0]**o+(-0.480556561488)*x[0]+(-0.958556900472)*x[1]**o+(-0.764939529718)*x[1]
        arg[0,4]=(-0.51062937871)*x[0]**o+(-0.135415437913)*x[0]+(0.340077694039)*x[1]**o+(0.874347806916)*x[1]
        arg[1,0]=(0.636810684661)*x[0]**o+(0.774651780754)*x[0]+(-0.314130997812)*x[1]**o+(-0.422706240721)*x[1]
        arg[1,1]=(0.495122166216)*x[0]**o+(0.757099858142)*x[0]+(-0.25861054515)*x[1]**o+(0.612180732826)*x[1]
        arg[1,2]=(-0.909636123949)*x[0]**o+(-0.0434375137409)*x[0]+(0.100612174358)*x[1]**o+(0.787104124194)*x[1]
        arg[1,3]=(-0.499307726848)*x[0]**o+(0.013205984499)*x[0]+(0.858074302957)*x[1]**o+(-0.630926261217)*x[1]
        arg[1,4]=(-0.410815849264)*x[0]**o+(0.820939326325)*x[0]+(-0.415370205615)*x[1]**o+(0.0602438669371)*x[1]
        arg[2,0]=(0.642545275281)*x[0]**o+(0.872239548588)*x[0]+(0.446294659843)*x[1]**o+(0.929350720888)*x[1]
        arg[2,1]=(0.182218495486)*x[0]**o+(-0.393184556589)*x[0]+(0.42311545961)*x[1]**o+(-0.479636250878)*x[1]
        arg[2,2]=(0.874574618473)*x[0]**o+(0.879031610251)*x[0]+(0.686157301617)*x[1]**o+(0.706917527697)*x[1]
        arg[2,3]=(0.614189029029)*x[0]**o+(0.0973950074653)*x[0]+(0.626052996278)*x[1]**o+(0.75043535006)*x[1]
        arg[2,4]=(0.386209196978)*x[0]**o+(0.644169312279)*x[0]+(-0.554562346917)*x[1]**o+(-0.474292025255)*x[1]
        arg[3,0]=(0.015012073686)*x[0]**o+(-0.665804159938)*x[0]+(-0.849030247328)*x[1]**o+(-0.936779648988)*x[1]
        arg[3,1]=(-0.818194334698)*x[0]**o+(-0.569993362998)*x[0]+(-0.806235611766)*x[1]**o+(0.35347475408)*x[1]
        arg[3,2]=(-0.227117779134)*x[0]**o+(-0.981285175536)*x[0]+(-0.226804205345)*x[1]**o+(-0.0134258740334)*x[1]
        arg[3,3]=(0.126716245369)*x[0]**o+(0.135476076345)*x[0]+(-0.641421845681)*x[1]**o+(0.0595388928019)*x[1]
        arg[3,4]=(-0.658260624128)*x[0]**o+(0.124989215683)*x[0]+(-0.949146047463)*x[1]**o+(0.805277938593)*x[1]
        ref[0,0]=(0.436439283545)*x_ref[0]**o+(0.705849609672)*x_ref[0]+(-0.261476237133)*x_ref[1]**o+(-0.10504570237)*x_ref[1]
        ref[0,1]=(0.035208834696)*x_ref[0]**o+(0.590058421274)*x_ref[0]+(-0.532905126944)*x_ref[1]**o+(0.126572388319)*x_ref[1]
        ref[0,2]=(0.985266120604)*x_ref[0]**o+(-0.465951670003)*x_ref[0]+(0.855881157118)*x_ref[1]**o+(0.00932326399668)*x_ref[1]
        ref[0,3]=(0.867525798771)*x_ref[0]**o+(-0.480556561488)*x_ref[0]+(-0.958556900472)*x_ref[1]**o+(-0.764939529718)*x_ref[1]
        ref[0,4]=(-0.51062937871)*x_ref[0]**o+(-0.135415437913)*x_ref[0]+(0.340077694039)*x_ref[1]**o+(0.874347806916)*x_ref[1]
        ref[1,0]=(0.636810684661)*x_ref[0]**o+(0.774651780754)*x_ref[0]+(-0.314130997812)*x_ref[1]**o+(-0.422706240721)*x_ref[1]
        ref[1,1]=(0.495122166216)*x_ref[0]**o+(0.757099858142)*x_ref[0]+(-0.25861054515)*x_ref[1]**o+(0.612180732826)*x_ref[1]
        ref[1,2]=(-0.909636123949)*x_ref[0]**o+(-0.0434375137409)*x_ref[0]+(0.100612174358)*x_ref[1]**o+(0.787104124194)*x_ref[1]
        ref[1,3]=(-0.499307726848)*x_ref[0]**o+(0.013205984499)*x_ref[0]+(0.858074302957)*x_ref[1]**o+(-0.630926261217)*x_ref[1]
        ref[1,4]=(-0.410815849264)*x_ref[0]**o+(0.820939326325)*x_ref[0]+(-0.415370205615)*x_ref[1]**o+(0.0602438669371)*x_ref[1]
        ref[2,0]=(0.642545275281)*x_ref[0]**o+(0.872239548588)*x_ref[0]+(0.446294659843)*x_ref[1]**o+(0.929350720888)*x_ref[1]
        ref[2,1]=(0.182218495486)*x_ref[0]**o+(-0.393184556589)*x_ref[0]+(0.42311545961)*x_ref[1]**o+(-0.479636250878)*x_ref[1]
        ref[2,2]=(0.874574618473)*x_ref[0]**o+(0.879031610251)*x_ref[0]+(0.686157301617)*x_ref[1]**o+(0.706917527697)*x_ref[1]
        ref[2,3]=(0.614189029029)*x_ref[0]**o+(0.0973950074653)*x_ref[0]+(0.626052996278)*x_ref[1]**o+(0.75043535006)*x_ref[1]
        ref[2,4]=(0.386209196978)*x_ref[0]**o+(0.644169312279)*x_ref[0]+(-0.554562346917)*x_ref[1]**o+(-0.474292025255)*x_ref[1]
        ref[3,0]=(0.015012073686)*x_ref[0]**o+(-0.665804159938)*x_ref[0]+(-0.849030247328)*x_ref[1]**o+(-0.936779648988)*x_ref[1]
        ref[3,1]=(-0.818194334698)*x_ref[0]**o+(-0.569993362998)*x_ref[0]+(-0.806235611766)*x_ref[1]**o+(0.35347475408)*x_ref[1]
        ref[3,2]=(-0.227117779134)*x_ref[0]**o+(-0.981285175536)*x_ref[0]+(-0.226804205345)*x_ref[1]**o+(-0.0134258740334)*x_ref[1]
        ref[3,3]=(0.126716245369)*x_ref[0]**o+(0.135476076345)*x_ref[0]+(-0.641421845681)*x_ref[1]**o+(0.0595388928019)*x_ref[1]
        ref[3,4]=(-0.658260624128)*x_ref[0]**o+(0.124989215683)*x_ref[0]+(-0.949146047463)*x_ref[1]**o+(0.805277938593)*x_ref[1]
      else:
        arg[0,0]=(-0.859332605724)*x[0]**o+(-0.239005430756)*x[0]+(-0.52959833033)*x[1]**o+(0.725877930655)*x[1]+(-0.0348436114161)*x[2]**o+(0.47287744745)*x[2]
        arg[0,1]=(-0.0816886854463)*x[0]**o+(-0.626507446322)*x[0]+(-0.93804868959)*x[1]**o+(-0.450134311929)*x[1]+(0.0957946938715)*x[2]**o+(-0.236994496198)*x[2]
        arg[0,2]=(-0.669441310039)*x[0]**o+(-0.406451428536)*x[0]+(-0.292718530458)*x[1]**o+(-0.0948016380336)*x[1]+(-0.0943083566957)*x[2]**o+(0.747978324024)*x[2]
        arg[0,3]=(-0.0350228116485)*x[0]**o+(-0.967728170529)*x[0]+(-0.16425180976)*x[1]**o+(-0.141334783297)*x[1]+(0.179458630309)*x[2]**o+(0.45080506919)*x[2]
        arg[0,4]=(0.909578721407)*x[0]**o+(-0.341028350711)*x[0]+(-0.937703389458)*x[1]**o+(-0.845025015446)*x[1]+(-0.762836324997)*x[2]**o+(0.772817997464)*x[2]
        arg[1,0]=(0.75675349915)*x[0]**o+(-0.208827138169)*x[0]+(-0.0146742840091)*x[1]**o+(-0.1486215734)*x[1]+(0.205344647252)*x[2]**o+(-0.409293867732)*x[2]
        arg[1,1]=(-0.215359027624)*x[0]**o+(0.220661567012)*x[0]+(0.122954614324)*x[1]**o+(-0.888828539035)*x[1]+(-0.860535003713)*x[2]**o+(-0.582396229154)*x[2]
        arg[1,2]=(-0.252703089432)*x[0]**o+(0.294168447756)*x[0]+(0.323290819691)*x[1]**o+(-0.0440370164755)*x[1]+(0.487110700327)*x[2]**o+(-0.269109129573)*x[2]
        arg[1,3]=(0.0653521784751)*x[0]**o+(0.328048279094)*x[0]+(-0.414467594535)*x[1]**o+(0.436061182815)*x[1]+(0.483513283452)*x[2]**o+(-0.744560862926)*x[2]
        arg[1,4]=(0.0102955000926)*x[0]**o+(-0.636994970624)*x[0]+(0.83732393089)*x[1]**o+(0.914683806871)*x[1]+(0.574439856221)*x[2]**o+(-0.096821482591)*x[2]
        arg[2,0]=(-0.129692996812)*x[0]**o+(-0.617534967242)*x[0]+(0.299770969796)*x[1]**o+(0.93870530376)*x[1]+(-0.864448551767)*x[2]**o+(0.755129965839)*x[2]
        arg[2,1]=(-0.354122778719)*x[0]**o+(0.576654786045)*x[0]+(-0.965116346607)*x[1]**o+(0.0705856628847)*x[1]+(0.817059960545)*x[2]**o+(-0.597731384232)*x[2]
        arg[2,2]=(-0.705190107591)*x[0]**o+(0.876019062019)*x[0]+(-0.732733986501)*x[1]**o+(0.505334922081)*x[1]+(-0.616601928994)*x[2]**o+(-0.332268983863)*x[2]
        arg[2,3]=(0.663768017414)*x[0]**o+(-0.431682444208)*x[0]+(0.186159158088)*x[1]**o+(-0.689766933274)*x[1]+(-0.765854811539)*x[2]**o+(-0.402774599632)*x[2]
        arg[2,4]=(-0.698031877485)*x[0]**o+(0.914418269242)*x[0]+(-0.693564630363)*x[1]**o+(-0.18289238868)*x[1]+(0.988158500287)*x[2]**o+(-0.33952288549)*x[2]
        arg[3,0]=(-0.188499634873)*x[0]**o+(-0.10427816897)*x[0]+(-0.772866404794)*x[1]**o+(0.841928986479)*x[1]+(-0.866955639695)*x[2]**o+(-0.152515560713)*x[2]
        arg[3,1]=(0.937695460564)*x[0]**o+(-0.551079526492)*x[0]+(0.353043727721)*x[1]**o+(-0.604814667118)*x[1]+(0.200913251642)*x[2]**o+(-0.0497427852682)*x[2]
        arg[3,2]=(-0.49130784927)*x[0]**o+(0.537445967432)*x[0]+(-0.408371200391)*x[1]**o+(0.945593173362)*x[1]+(-0.0956969348675)*x[2]**o+(-0.174133549441)*x[2]
        arg[3,3]=(0.847790369559)*x[0]**o+(0.743993523287)*x[0]+(-0.810126540693)*x[1]**o+(-0.775699501886)*x[1]+(-0.479687101313)*x[2]**o+(0.627116825116)*x[2]
        arg[3,4]=(0.352455163407)*x[0]**o+(0.86343850667)*x[0]+(0.126742974996)*x[1]**o+(0.0175699888323)*x[1]+(0.964899127449)*x[2]**o+(-0.206749046822)*x[2]
        ref[0,0]=(-0.859332605724)*x_ref[0]**o+(-0.239005430756)*x_ref[0]+(-0.52959833033)*x_ref[1]**o+(0.725877930655)*x_ref[1]+(-0.0348436114161)*x_ref[2]**o+(0.47287744745)*x_ref[2]
        ref[0,1]=(-0.0816886854463)*x_ref[0]**o+(-0.626507446322)*x_ref[0]+(-0.93804868959)*x_ref[1]**o+(-0.450134311929)*x_ref[1]+(0.0957946938715)*x_ref[2]**o+(-0.236994496198)*x_ref[2]
        ref[0,2]=(-0.669441310039)*x_ref[0]**o+(-0.406451428536)*x_ref[0]+(-0.292718530458)*x_ref[1]**o+(-0.0948016380336)*x_ref[1]+(-0.0943083566957)*x_ref[2]**o+(0.747978324024)*x_ref[2]
        ref[0,3]=(-0.0350228116485)*x_ref[0]**o+(-0.967728170529)*x_ref[0]+(-0.16425180976)*x_ref[1]**o+(-0.141334783297)*x_ref[1]+(0.179458630309)*x_ref[2]**o+(0.45080506919)*x_ref[2]
        ref[0,4]=(0.909578721407)*x_ref[0]**o+(-0.341028350711)*x_ref[0]+(-0.937703389458)*x_ref[1]**o+(-0.845025015446)*x_ref[1]+(-0.762836324997)*x_ref[2]**o+(0.772817997464)*x_ref[2]
        ref[1,0]=(0.75675349915)*x_ref[0]**o+(-0.208827138169)*x_ref[0]+(-0.0146742840091)*x_ref[1]**o+(-0.1486215734)*x_ref[1]+(0.205344647252)*x_ref[2]**o+(-0.409293867732)*x_ref[2]
        ref[1,1]=(-0.215359027624)*x_ref[0]**o+(0.220661567012)*x_ref[0]+(0.122954614324)*x_ref[1]**o+(-0.888828539035)*x_ref[1]+(-0.860535003713)*x_ref[2]**o+(-0.582396229154)*x_ref[2]
        ref[1,2]=(-0.252703089432)*x_ref[0]**o+(0.294168447756)*x_ref[0]+(0.323290819691)*x_ref[1]**o+(-0.0440370164755)*x_ref[1]+(0.487110700327)*x_ref[2]**o+(-0.269109129573)*x_ref[2]
        ref[1,3]=(0.0653521784751)*x_ref[0]**o+(0.328048279094)*x_ref[0]+(-0.414467594535)*x_ref[1]**o+(0.436061182815)*x_ref[1]+(0.483513283452)*x_ref[2]**o+(-0.744560862926)*x_ref[2]
        ref[1,4]=(0.0102955000926)*x_ref[0]**o+(-0.636994970624)*x_ref[0]+(0.83732393089)*x_ref[1]**o+(0.914683806871)*x_ref[1]+(0.574439856221)*x_ref[2]**o+(-0.096821482591)*x_ref[2]
        ref[2,0]=(-0.129692996812)*x_ref[0]**o+(-0.617534967242)*x_ref[0]+(0.299770969796)*x_ref[1]**o+(0.93870530376)*x_ref[1]+(-0.864448551767)*x_ref[2]**o+(0.755129965839)*x_ref[2]
        ref[2,1]=(-0.354122778719)*x_ref[0]**o+(0.576654786045)*x_ref[0]+(-0.965116346607)*x_ref[1]**o+(0.0705856628847)*x_ref[1]+(0.817059960545)*x_ref[2]**o+(-0.597731384232)*x_ref[2]
        ref[2,2]=(-0.705190107591)*x_ref[0]**o+(0.876019062019)*x_ref[0]+(-0.732733986501)*x_ref[1]**o+(0.505334922081)*x_ref[1]+(-0.616601928994)*x_ref[2]**o+(-0.332268983863)*x_ref[2]
        ref[2,3]=(0.663768017414)*x_ref[0]**o+(-0.431682444208)*x_ref[0]+(0.186159158088)*x_ref[1]**o+(-0.689766933274)*x_ref[1]+(-0.765854811539)*x_ref[2]**o+(-0.402774599632)*x_ref[2]
        ref[2,4]=(-0.698031877485)*x_ref[0]**o+(0.914418269242)*x_ref[0]+(-0.693564630363)*x_ref[1]**o+(-0.18289238868)*x_ref[1]+(0.988158500287)*x_ref[2]**o+(-0.33952288549)*x_ref[2]
        ref[3,0]=(-0.188499634873)*x_ref[0]**o+(-0.10427816897)*x_ref[0]+(-0.772866404794)*x_ref[1]**o+(0.841928986479)*x_ref[1]+(-0.866955639695)*x_ref[2]**o+(-0.152515560713)*x_ref[2]
        ref[3,1]=(0.937695460564)*x_ref[0]**o+(-0.551079526492)*x_ref[0]+(0.353043727721)*x_ref[1]**o+(-0.604814667118)*x_ref[1]+(0.200913251642)*x_ref[2]**o+(-0.0497427852682)*x_ref[2]
        ref[3,2]=(-0.49130784927)*x_ref[0]**o+(0.537445967432)*x_ref[0]+(-0.408371200391)*x_ref[1]**o+(0.945593173362)*x_ref[1]+(-0.0956969348675)*x_ref[2]**o+(-0.174133549441)*x_ref[2]
        ref[3,3]=(0.847790369559)*x_ref[0]**o+(0.743993523287)*x_ref[0]+(-0.810126540693)*x_ref[1]**o+(-0.775699501886)*x_ref[1]+(-0.479687101313)*x_ref[2]**o+(0.627116825116)*x_ref[2]
        ref[3,4]=(0.352455163407)*x_ref[0]**o+(0.86343850667)*x_ref[0]+(0.126742974996)*x_ref[1]**o+(0.0175699888323)*x_ref[1]+(0.964899127449)*x_ref[2]**o+(-0.206749046822)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.396745847367)*x[0]**o+(-0.700719490691)*x[0]+(0.117558242545)*x[1]**o+(-0.161976142849)*x[1]
        arg[0,0,1]=(-0.490732277107)*x[0]**o+(0.526833149216)*x[0]+(-0.136309303122)*x[1]**o+(0.629465649533)*x[1]
        arg[0,1,0]=(-0.242474968121)*x[0]**o+(-0.76270698332)*x[0]+(-0.0979504547153)*x[1]**o+(-0.0743521177641)*x[1]
        arg[0,1,1]=(0.641246476843)*x[0]**o+(0.0158147797433)*x[0]+(-0.213898399703)*x[1]**o+(-0.378401519675)*x[1]
        arg[1,0,0]=(0.603700544573)*x[0]**o+(-0.993751144334)*x[0]+(0.466222025692)*x[1]**o+(0.457689495241)*x[1]
        arg[1,0,1]=(-0.727287858841)*x[0]**o+(-0.177841994955)*x[0]+(0.720237448692)*x[1]**o+(0.716175629178)*x[1]
        arg[1,1,0]=(0.784514682242)*x[0]**o+(0.681214612353)*x[0]+(0.404266860166)*x[1]**o+(0.871218157219)*x[1]
        arg[1,1,1]=(-0.918634586367)*x[0]**o+(-0.722365105889)*x[0]+(-0.0337202979587)*x[1]**o+(-0.487790376172)*x[1]
        arg[2,0,0]=(0.159515902499)*x[0]**o+(-0.0650417831188)*x[0]+(0.107260481333)*x[1]**o+(0.469096212511)*x[1]
        arg[2,0,1]=(0.0436003865409)*x[0]**o+(-0.67488837592)*x[0]+(0.547761443272)*x[1]**o+(-0.69222031017)*x[1]
        arg[2,1,0]=(-0.153636036391)*x[0]**o+(0.0415911821723)*x[0]+(0.421990370678)*x[1]**o+(-0.128883832414)*x[1]
        arg[2,1,1]=(0.513420393955)*x[0]**o+(-0.769636964711)*x[0]+(0.874383588901)*x[1]**o+(-0.357923227767)*x[1]
        arg[3,0,0]=(0.943275907357)*x[0]**o+(0.372378913231)*x[0]+(0.559656674007)*x[1]**o+(-0.878007664293)*x[1]
        arg[3,0,1]=(-0.541535388034)*x[0]**o+(-0.0510200932939)*x[0]+(-0.448618747155)*x[1]**o+(0.693514427768)*x[1]
        arg[3,1,0]=(-0.287622342268)*x[0]**o+(0.8333134988)*x[0]+(-0.691930501782)*x[1]**o+(-0.659618596505)*x[1]
        arg[3,1,1]=(0.44446133556)*x[0]**o+(0.697704081094)*x[0]+(0.0679863572322)*x[1]**o+(-0.949443102916)*x[1]
        arg[4,0,0]=(0.546245770076)*x[0]**o+(-0.466891879347)*x[0]+(-0.608261901126)*x[1]**o+(0.669561643667)*x[1]
        arg[4,0,1]=(-0.522497485476)*x[0]**o+(0.70886716903)*x[0]+(-0.296815361132)*x[1]**o+(0.500540583082)*x[1]
        arg[4,1,0]=(-0.841653490079)*x[0]**o+(0.400589086976)*x[0]+(0.865994660457)*x[1]**o+(0.0205817210969)*x[1]
        arg[4,1,1]=(-0.64369513787)*x[0]**o+(0.883032618928)*x[0]+(-0.178766897224)*x[1]**o+(0.748041632322)*x[1]
        arg[5,0,0]=(0.471810615096)*x[0]**o+(-0.140340964274)*x[0]+(-0.7063314968)*x[1]**o+(0.0524231879837)*x[1]
        arg[5,0,1]=(-0.51339482213)*x[0]**o+(-0.81447661092)*x[0]+(0.437085868301)*x[1]**o+(0.862678017842)*x[1]
        arg[5,1,0]=(-0.752858172473)*x[0]**o+(0.325827525243)*x[0]+(-0.702286757187)*x[1]**o+(-0.467810951898)*x[1]
        arg[5,1,1]=(-0.256879797442)*x[0]**o+(0.257067666229)*x[0]+(0.912308244411)*x[1]**o+(0.0620337856352)*x[1]
        ref[0,0,0]=(-0.396745847367)*x_ref[0]**o+(-0.700719490691)*x_ref[0]+(0.117558242545)*x_ref[1]**o+(-0.161976142849)*x_ref[1]
        ref[0,0,1]=(-0.490732277107)*x_ref[0]**o+(0.526833149216)*x_ref[0]+(-0.136309303122)*x_ref[1]**o+(0.629465649533)*x_ref[1]
        ref[0,1,0]=(-0.242474968121)*x_ref[0]**o+(-0.76270698332)*x_ref[0]+(-0.0979504547153)*x_ref[1]**o+(-0.0743521177641)*x_ref[1]
        ref[0,1,1]=(0.641246476843)*x_ref[0]**o+(0.0158147797433)*x_ref[0]+(-0.213898399703)*x_ref[1]**o+(-0.378401519675)*x_ref[1]
        ref[1,0,0]=(0.603700544573)*x_ref[0]**o+(-0.993751144334)*x_ref[0]+(0.466222025692)*x_ref[1]**o+(0.457689495241)*x_ref[1]
        ref[1,0,1]=(-0.727287858841)*x_ref[0]**o+(-0.177841994955)*x_ref[0]+(0.720237448692)*x_ref[1]**o+(0.716175629178)*x_ref[1]
        ref[1,1,0]=(0.784514682242)*x_ref[0]**o+(0.681214612353)*x_ref[0]+(0.404266860166)*x_ref[1]**o+(0.871218157219)*x_ref[1]
        ref[1,1,1]=(-0.918634586367)*x_ref[0]**o+(-0.722365105889)*x_ref[0]+(-0.0337202979587)*x_ref[1]**o+(-0.487790376172)*x_ref[1]
        ref[2,0,0]=(0.159515902499)*x_ref[0]**o+(-0.0650417831188)*x_ref[0]+(0.107260481333)*x_ref[1]**o+(0.469096212511)*x_ref[1]
        ref[2,0,1]=(0.0436003865409)*x_ref[0]**o+(-0.67488837592)*x_ref[0]+(0.547761443272)*x_ref[1]**o+(-0.69222031017)*x_ref[1]
        ref[2,1,0]=(-0.153636036391)*x_ref[0]**o+(0.0415911821723)*x_ref[0]+(0.421990370678)*x_ref[1]**o+(-0.128883832414)*x_ref[1]
        ref[2,1,1]=(0.513420393955)*x_ref[0]**o+(-0.769636964711)*x_ref[0]+(0.874383588901)*x_ref[1]**o+(-0.357923227767)*x_ref[1]
        ref[3,0,0]=(0.943275907357)*x_ref[0]**o+(0.372378913231)*x_ref[0]+(0.559656674007)*x_ref[1]**o+(-0.878007664293)*x_ref[1]
        ref[3,0,1]=(-0.541535388034)*x_ref[0]**o+(-0.0510200932939)*x_ref[0]+(-0.448618747155)*x_ref[1]**o+(0.693514427768)*x_ref[1]
        ref[3,1,0]=(-0.287622342268)*x_ref[0]**o+(0.8333134988)*x_ref[0]+(-0.691930501782)*x_ref[1]**o+(-0.659618596505)*x_ref[1]
        ref[3,1,1]=(0.44446133556)*x_ref[0]**o+(0.697704081094)*x_ref[0]+(0.0679863572322)*x_ref[1]**o+(-0.949443102916)*x_ref[1]
        ref[4,0,0]=(0.546245770076)*x_ref[0]**o+(-0.466891879347)*x_ref[0]+(-0.608261901126)*x_ref[1]**o+(0.669561643667)*x_ref[1]
        ref[4,0,1]=(-0.522497485476)*x_ref[0]**o+(0.70886716903)*x_ref[0]+(-0.296815361132)*x_ref[1]**o+(0.500540583082)*x_ref[1]
        ref[4,1,0]=(-0.841653490079)*x_ref[0]**o+(0.400589086976)*x_ref[0]+(0.865994660457)*x_ref[1]**o+(0.0205817210969)*x_ref[1]
        ref[4,1,1]=(-0.64369513787)*x_ref[0]**o+(0.883032618928)*x_ref[0]+(-0.178766897224)*x_ref[1]**o+(0.748041632322)*x_ref[1]
        ref[5,0,0]=(0.471810615096)*x_ref[0]**o+(-0.140340964274)*x_ref[0]+(-0.7063314968)*x_ref[1]**o+(0.0524231879837)*x_ref[1]
        ref[5,0,1]=(-0.51339482213)*x_ref[0]**o+(-0.81447661092)*x_ref[0]+(0.437085868301)*x_ref[1]**o+(0.862678017842)*x_ref[1]
        ref[5,1,0]=(-0.752858172473)*x_ref[0]**o+(0.325827525243)*x_ref[0]+(-0.702286757187)*x_ref[1]**o+(-0.467810951898)*x_ref[1]
        ref[5,1,1]=(-0.256879797442)*x_ref[0]**o+(0.257067666229)*x_ref[0]+(0.912308244411)*x_ref[1]**o+(0.0620337856352)*x_ref[1]
      else:
        arg[0,0,0]=(0.564502533309)*x[0]**o+(-0.237721559752)*x[0]+(0.225601658644)*x[1]**o+(0.967190188292)*x[1]+(0.471760897542)*x[2]**o+(0.79404420728)*x[2]
        arg[0,0,1]=(-0.373762346136)*x[0]**o+(-0.643766021731)*x[0]+(0.088028383792)*x[1]**o+(-0.541163410313)*x[1]+(0.463315911637)*x[2]**o+(-0.596740807595)*x[2]
        arg[0,1,0]=(-0.526694453842)*x[0]**o+(0.405010889931)*x[0]+(-0.0153807962972)*x[1]**o+(-0.0446620712949)*x[1]+(0.00553918668197)*x[2]**o+(0.72170514564)*x[2]
        arg[0,1,1]=(-0.00750073616089)*x[0]**o+(-0.60431903208)*x[0]+(0.790032415931)*x[1]**o+(0.272709398336)*x[1]+(0.228539600712)*x[2]**o+(0.62284140325)*x[2]
        arg[1,0,0]=(-0.892797595418)*x[0]**o+(-0.151608141828)*x[0]+(0.313142074823)*x[1]**o+(0.80280439466)*x[1]+(-0.593056273594)*x[2]**o+(-0.41551679707)*x[2]
        arg[1,0,1]=(-0.385715092974)*x[0]**o+(-0.77884107065)*x[0]+(-0.0804019137225)*x[1]**o+(-0.240129932101)*x[1]+(0.238145788042)*x[2]**o+(-0.806337666628)*x[2]
        arg[1,1,0]=(0.940492459417)*x[0]**o+(-0.0054460073194)*x[0]+(-0.697916565168)*x[1]**o+(-0.949456184725)*x[1]+(-0.115461214789)*x[2]**o+(-0.403742409948)*x[2]
        arg[1,1,1]=(-0.711716124415)*x[0]**o+(0.586131021871)*x[0]+(0.625988666837)*x[1]**o+(0.638433219635)*x[1]+(0.971169174015)*x[2]**o+(-0.064949626884)*x[2]
        arg[2,0,0]=(-0.891973333325)*x[0]**o+(0.76090478364)*x[0]+(0.439956136239)*x[1]**o+(-0.74807071296)*x[1]+(-0.209999803467)*x[2]**o+(0.404396254503)*x[2]
        arg[2,0,1]=(0.559776288735)*x[0]**o+(0.284643618707)*x[0]+(0.0527799359967)*x[1]**o+(-0.201660204363)*x[1]+(0.590722655898)*x[2]**o+(-0.24940757474)*x[2]
        arg[2,1,0]=(-0.747588234432)*x[0]**o+(0.0196702046746)*x[0]+(0.0454238912973)*x[1]**o+(0.174053915643)*x[1]+(-0.868554688657)*x[2]**o+(-0.127425849813)*x[2]
        arg[2,1,1]=(-0.324276984008)*x[0]**o+(-0.494912096003)*x[0]+(-0.780953487415)*x[1]**o+(0.225787179491)*x[1]+(-0.760330660035)*x[2]**o+(0.0440722700759)*x[2]
        arg[3,0,0]=(0.37479028021)*x[0]**o+(-0.869009335783)*x[0]+(0.811673580204)*x[1]**o+(0.521877505351)*x[1]+(-0.508710979256)*x[2]**o+(0.558085136638)*x[2]
        arg[3,0,1]=(0.815143535066)*x[0]**o+(0.867977183071)*x[0]+(0.0315543942614)*x[1]**o+(0.786379597601)*x[1]+(0.0433279381339)*x[2]**o+(-0.56554445624)*x[2]
        arg[3,1,0]=(-0.948224238625)*x[0]**o+(-0.759282528831)*x[0]+(0.511802336753)*x[1]**o+(0.331073342499)*x[1]+(0.295241697427)*x[2]**o+(0.154864664771)*x[2]
        arg[3,1,1]=(-0.982634132939)*x[0]**o+(-0.0762014431695)*x[0]+(-0.0797462613351)*x[1]**o+(-0.512305271957)*x[1]+(-0.348883505418)*x[2]**o+(0.00580412526927)*x[2]
        arg[4,0,0]=(0.186597109908)*x[0]**o+(0.051957206238)*x[0]+(0.572882754553)*x[1]**o+(-0.114338954084)*x[1]+(-0.517239439578)*x[2]**o+(-0.298287404607)*x[2]
        arg[4,0,1]=(0.637628569107)*x[0]**o+(0.228562669958)*x[0]+(0.053657321561)*x[1]**o+(0.762909608846)*x[1]+(0.873936819574)*x[2]**o+(-0.20000969514)*x[2]
        arg[4,1,0]=(-0.717459489066)*x[0]**o+(-0.462174191129)*x[0]+(-0.70491180102)*x[1]**o+(-0.406072010622)*x[1]+(-0.903872259447)*x[2]**o+(0.487491537826)*x[2]
        arg[4,1,1]=(0.947654286923)*x[0]**o+(-0.228830597313)*x[0]+(-0.501442713655)*x[1]**o+(0.282437368749)*x[1]+(0.543592950002)*x[2]**o+(0.5400198247)*x[2]
        arg[5,0,0]=(0.144349494033)*x[0]**o+(-0.635328703132)*x[0]+(-0.797851162499)*x[1]**o+(0.356874780185)*x[1]+(-0.205725457569)*x[2]**o+(-0.674962512018)*x[2]
        arg[5,0,1]=(-0.728948504419)*x[0]**o+(0.944337871242)*x[0]+(-0.790388944678)*x[1]**o+(0.59119693076)*x[1]+(0.9569152246)*x[2]**o+(-0.205404245177)*x[2]
        arg[5,1,0]=(0.477597678348)*x[0]**o+(-0.318993929605)*x[0]+(0.481029970286)*x[1]**o+(0.708696918395)*x[1]+(0.0531716082887)*x[2]**o+(-0.847697491061)*x[2]
        arg[5,1,1]=(-0.458572001979)*x[0]**o+(0.582048244354)*x[0]+(-0.040756132547)*x[1]**o+(-0.953201198076)*x[1]+(-0.971859723822)*x[2]**o+(-0.796426143181)*x[2]
        ref[0,0,0]=(0.564502533309)*x_ref[0]**o+(-0.237721559752)*x_ref[0]+(0.225601658644)*x_ref[1]**o+(0.967190188292)*x_ref[1]+(0.471760897542)*x_ref[2]**o+(0.79404420728)*x_ref[2]
        ref[0,0,1]=(-0.373762346136)*x_ref[0]**o+(-0.643766021731)*x_ref[0]+(0.088028383792)*x_ref[1]**o+(-0.541163410313)*x_ref[1]+(0.463315911637)*x_ref[2]**o+(-0.596740807595)*x_ref[2]
        ref[0,1,0]=(-0.526694453842)*x_ref[0]**o+(0.405010889931)*x_ref[0]+(-0.0153807962972)*x_ref[1]**o+(-0.0446620712949)*x_ref[1]+(0.00553918668197)*x_ref[2]**o+(0.72170514564)*x_ref[2]
        ref[0,1,1]=(-0.00750073616089)*x_ref[0]**o+(-0.60431903208)*x_ref[0]+(0.790032415931)*x_ref[1]**o+(0.272709398336)*x_ref[1]+(0.228539600712)*x_ref[2]**o+(0.62284140325)*x_ref[2]
        ref[1,0,0]=(-0.892797595418)*x_ref[0]**o+(-0.151608141828)*x_ref[0]+(0.313142074823)*x_ref[1]**o+(0.80280439466)*x_ref[1]+(-0.593056273594)*x_ref[2]**o+(-0.41551679707)*x_ref[2]
        ref[1,0,1]=(-0.385715092974)*x_ref[0]**o+(-0.77884107065)*x_ref[0]+(-0.0804019137225)*x_ref[1]**o+(-0.240129932101)*x_ref[1]+(0.238145788042)*x_ref[2]**o+(-0.806337666628)*x_ref[2]
        ref[1,1,0]=(0.940492459417)*x_ref[0]**o+(-0.0054460073194)*x_ref[0]+(-0.697916565168)*x_ref[1]**o+(-0.949456184725)*x_ref[1]+(-0.115461214789)*x_ref[2]**o+(-0.403742409948)*x_ref[2]
        ref[1,1,1]=(-0.711716124415)*x_ref[0]**o+(0.586131021871)*x_ref[0]+(0.625988666837)*x_ref[1]**o+(0.638433219635)*x_ref[1]+(0.971169174015)*x_ref[2]**o+(-0.064949626884)*x_ref[2]
        ref[2,0,0]=(-0.891973333325)*x_ref[0]**o+(0.76090478364)*x_ref[0]+(0.439956136239)*x_ref[1]**o+(-0.74807071296)*x_ref[1]+(-0.209999803467)*x_ref[2]**o+(0.404396254503)*x_ref[2]
        ref[2,0,1]=(0.559776288735)*x_ref[0]**o+(0.284643618707)*x_ref[0]+(0.0527799359967)*x_ref[1]**o+(-0.201660204363)*x_ref[1]+(0.590722655898)*x_ref[2]**o+(-0.24940757474)*x_ref[2]
        ref[2,1,0]=(-0.747588234432)*x_ref[0]**o+(0.0196702046746)*x_ref[0]+(0.0454238912973)*x_ref[1]**o+(0.174053915643)*x_ref[1]+(-0.868554688657)*x_ref[2]**o+(-0.127425849813)*x_ref[2]
        ref[2,1,1]=(-0.324276984008)*x_ref[0]**o+(-0.494912096003)*x_ref[0]+(-0.780953487415)*x_ref[1]**o+(0.225787179491)*x_ref[1]+(-0.760330660035)*x_ref[2]**o+(0.0440722700759)*x_ref[2]
        ref[3,0,0]=(0.37479028021)*x_ref[0]**o+(-0.869009335783)*x_ref[0]+(0.811673580204)*x_ref[1]**o+(0.521877505351)*x_ref[1]+(-0.508710979256)*x_ref[2]**o+(0.558085136638)*x_ref[2]
        ref[3,0,1]=(0.815143535066)*x_ref[0]**o+(0.867977183071)*x_ref[0]+(0.0315543942614)*x_ref[1]**o+(0.786379597601)*x_ref[1]+(0.0433279381339)*x_ref[2]**o+(-0.56554445624)*x_ref[2]
        ref[3,1,0]=(-0.948224238625)*x_ref[0]**o+(-0.759282528831)*x_ref[0]+(0.511802336753)*x_ref[1]**o+(0.331073342499)*x_ref[1]+(0.295241697427)*x_ref[2]**o+(0.154864664771)*x_ref[2]
        ref[3,1,1]=(-0.982634132939)*x_ref[0]**o+(-0.0762014431695)*x_ref[0]+(-0.0797462613351)*x_ref[1]**o+(-0.512305271957)*x_ref[1]+(-0.348883505418)*x_ref[2]**o+(0.00580412526927)*x_ref[2]
        ref[4,0,0]=(0.186597109908)*x_ref[0]**o+(0.051957206238)*x_ref[0]+(0.572882754553)*x_ref[1]**o+(-0.114338954084)*x_ref[1]+(-0.517239439578)*x_ref[2]**o+(-0.298287404607)*x_ref[2]
        ref[4,0,1]=(0.637628569107)*x_ref[0]**o+(0.228562669958)*x_ref[0]+(0.053657321561)*x_ref[1]**o+(0.762909608846)*x_ref[1]+(0.873936819574)*x_ref[2]**o+(-0.20000969514)*x_ref[2]
        ref[4,1,0]=(-0.717459489066)*x_ref[0]**o+(-0.462174191129)*x_ref[0]+(-0.70491180102)*x_ref[1]**o+(-0.406072010622)*x_ref[1]+(-0.903872259447)*x_ref[2]**o+(0.487491537826)*x_ref[2]
        ref[4,1,1]=(0.947654286923)*x_ref[0]**o+(-0.228830597313)*x_ref[0]+(-0.501442713655)*x_ref[1]**o+(0.282437368749)*x_ref[1]+(0.543592950002)*x_ref[2]**o+(0.5400198247)*x_ref[2]
        ref[5,0,0]=(0.144349494033)*x_ref[0]**o+(-0.635328703132)*x_ref[0]+(-0.797851162499)*x_ref[1]**o+(0.356874780185)*x_ref[1]+(-0.205725457569)*x_ref[2]**o+(-0.674962512018)*x_ref[2]
        ref[5,0,1]=(-0.728948504419)*x_ref[0]**o+(0.944337871242)*x_ref[0]+(-0.790388944678)*x_ref[1]**o+(0.59119693076)*x_ref[1]+(0.9569152246)*x_ref[2]**o+(-0.205404245177)*x_ref[2]
        ref[5,1,0]=(0.477597678348)*x_ref[0]**o+(-0.318993929605)*x_ref[0]+(0.481029970286)*x_ref[1]**o+(0.708696918395)*x_ref[1]+(0.0531716082887)*x_ref[2]**o+(-0.847697491061)*x_ref[2]
        ref[5,1,1]=(-0.458572001979)*x_ref[0]**o+(0.582048244354)*x_ref[0]+(-0.040756132547)*x_ref[1]**o+(-0.953201198076)*x_ref[1]+(-0.971859723822)*x_ref[2]**o+(-0.796426143181)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.937777610663)*x[0]**o+(-0.0228520946632)*x[0]+(0.642385396583)*x[1]**o+(-0.803821945812)*x[1]
        arg[0,0,0,1]=(-0.398785307492)*x[0]**o+(0.0450288952883)*x[0]+(-0.777780548133)*x[1]**o+(0.0273772099414)*x[1]
        arg[0,0,1,0]=(-0.667280086925)*x[0]**o+(-0.934642710274)*x[0]+(0.565327494491)*x[1]**o+(0.988521769056)*x[1]
        arg[0,0,1,1]=(-0.225562383498)*x[0]**o+(-0.280566031134)*x[0]+(-0.295768954831)*x[1]**o+(0.102759013802)*x[1]
        arg[0,0,2,0]=(-0.179577502546)*x[0]**o+(-0.156932777074)*x[0]+(-0.145616153633)*x[1]**o+(0.971876899135)*x[1]
        arg[0,0,2,1]=(-0.840266409118)*x[0]**o+(0.704356785022)*x[0]+(0.740091470711)*x[1]**o+(0.708478499023)*x[1]
        arg[0,1,0,0]=(-0.487161662806)*x[0]**o+(0.565313908399)*x[0]+(-0.33708697293)*x[1]**o+(0.731675966546)*x[1]
        arg[0,1,0,1]=(-0.135615857963)*x[0]**o+(0.265408777025)*x[0]+(-0.168207260227)*x[1]**o+(-0.598974659317)*x[1]
        arg[0,1,1,0]=(0.0533944010258)*x[0]**o+(0.819248906738)*x[0]+(-0.837005251524)*x[1]**o+(0.119613618595)*x[1]
        arg[0,1,1,1]=(-0.257520627047)*x[0]**o+(0.339264545713)*x[0]+(0.137721228188)*x[1]**o+(-0.512251030979)*x[1]
        arg[0,1,2,0]=(-0.797070396749)*x[0]**o+(0.690173285077)*x[0]+(-0.213206187271)*x[1]**o+(-0.873999125945)*x[1]
        arg[0,1,2,1]=(-0.847662300367)*x[0]**o+(0.785781114266)*x[0]+(-0.456918398811)*x[1]**o+(0.458881079947)*x[1]
        arg[0,2,0,0]=(-0.385269163334)*x[0]**o+(0.135941228964)*x[0]+(0.233245620648)*x[1]**o+(-0.246282209626)*x[1]
        arg[0,2,0,1]=(0.372914636218)*x[0]**o+(-0.160283657003)*x[0]+(0.0565043889781)*x[1]**o+(0.199715522021)*x[1]
        arg[0,2,1,0]=(0.676198390598)*x[0]**o+(0.938025122014)*x[0]+(0.161074696651)*x[1]**o+(-0.155495646964)*x[1]
        arg[0,2,1,1]=(-0.709286924155)*x[0]**o+(-0.173716284776)*x[0]+(0.0576435484694)*x[1]**o+(0.673924697832)*x[1]
        arg[0,2,2,0]=(-0.436468652969)*x[0]**o+(0.117279009082)*x[0]+(-0.731436369191)*x[1]**o+(0.919578503469)*x[1]
        arg[0,2,2,1]=(-0.691042646151)*x[0]**o+(-0.606293725574)*x[0]+(0.657291692759)*x[1]**o+(-0.0146398994729)*x[1]
        arg[0,3,0,0]=(0.365939313719)*x[0]**o+(-0.221236913792)*x[0]+(-0.729594967883)*x[1]**o+(0.0179024719114)*x[1]
        arg[0,3,0,1]=(0.135892133258)*x[0]**o+(0.728069461936)*x[0]+(-0.365691795279)*x[1]**o+(0.877708599922)*x[1]
        arg[0,3,1,0]=(-0.853312456817)*x[0]**o+(-0.713031969049)*x[0]+(0.274755827464)*x[1]**o+(-0.527330699033)*x[1]
        arg[0,3,1,1]=(-0.30358584687)*x[0]**o+(0.871629391968)*x[0]+(0.906915162645)*x[1]**o+(0.499955429041)*x[1]
        arg[0,3,2,0]=(-0.880532587427)*x[0]**o+(0.265729672258)*x[0]+(0.528677136349)*x[1]**o+(-0.224602964925)*x[1]
        arg[0,3,2,1]=(0.158632463532)*x[0]**o+(-0.665418329643)*x[0]+(0.957661069067)*x[1]**o+(-0.23771309434)*x[1]
        arg[0,4,0,0]=(-0.128073514905)*x[0]**o+(-0.452543798055)*x[0]+(0.303808626277)*x[1]**o+(0.841041779759)*x[1]
        arg[0,4,0,1]=(0.0219309964781)*x[0]**o+(-0.949488393443)*x[0]+(0.0436905404663)*x[1]**o+(-0.304527242795)*x[1]
        arg[0,4,1,0]=(-0.792275137775)*x[0]**o+(0.450079829644)*x[0]+(0.162879981637)*x[1]**o+(-0.185839015805)*x[1]
        arg[0,4,1,1]=(0.761932713893)*x[0]**o+(0.932601232945)*x[0]+(-0.606328305783)*x[1]**o+(-0.123973545682)*x[1]
        arg[0,4,2,0]=(-0.514088304908)*x[0]**o+(0.987668774807)*x[0]+(0.866078887168)*x[1]**o+(-0.608518224484)*x[1]
        arg[0,4,2,1]=(-0.427564608028)*x[0]**o+(0.231599614445)*x[0]+(-0.406862229819)*x[1]**o+(0.358456864265)*x[1]
        arg[1,0,0,0]=(0.342092171801)*x[0]**o+(-0.264522675185)*x[0]+(-0.37976917867)*x[1]**o+(0.778596602561)*x[1]
        arg[1,0,0,1]=(0.672156473235)*x[0]**o+(0.320317500856)*x[0]+(0.514206881296)*x[1]**o+(0.602594631705)*x[1]
        arg[1,0,1,0]=(0.914472920293)*x[0]**o+(0.525647943447)*x[0]+(0.685613621597)*x[1]**o+(0.482308039231)*x[1]
        arg[1,0,1,1]=(-0.727982874425)*x[0]**o+(-0.417964463065)*x[0]+(0.825853988063)*x[1]**o+(0.236134760252)*x[1]
        arg[1,0,2,0]=(0.922186381756)*x[0]**o+(-0.140522762044)*x[0]+(-0.282169431666)*x[1]**o+(0.778383988397)*x[1]
        arg[1,0,2,1]=(0.307296736289)*x[0]**o+(0.589844868569)*x[0]+(0.64571457752)*x[1]**o+(0.749333079485)*x[1]
        arg[1,1,0,0]=(0.908345176703)*x[0]**o+(-0.476053147164)*x[0]+(0.0828501189289)*x[1]**o+(0.280657117705)*x[1]
        arg[1,1,0,1]=(0.396128814097)*x[0]**o+(0.368821575482)*x[0]+(-0.916584602334)*x[1]**o+(-0.149377992159)*x[1]
        arg[1,1,1,0]=(-0.38468474313)*x[0]**o+(0.304742499301)*x[0]+(0.638083865846)*x[1]**o+(-0.332046460663)*x[1]
        arg[1,1,1,1]=(0.845097787138)*x[0]**o+(-0.34949498164)*x[0]+(0.229157202572)*x[1]**o+(0.677584672856)*x[1]
        arg[1,1,2,0]=(0.0280692825358)*x[0]**o+(0.567451370107)*x[0]+(0.66934188305)*x[1]**o+(-0.477197213188)*x[1]
        arg[1,1,2,1]=(0.375411487264)*x[0]**o+(0.55086548839)*x[0]+(-0.392135351366)*x[1]**o+(0.91971334589)*x[1]
        arg[1,2,0,0]=(-0.7659270368)*x[0]**o+(-0.338082792509)*x[0]+(-0.26272975232)*x[1]**o+(0.267936454568)*x[1]
        arg[1,2,0,1]=(0.918125455771)*x[0]**o+(0.182896655906)*x[0]+(0.525396835337)*x[1]**o+(0.201918789801)*x[1]
        arg[1,2,1,0]=(-0.54472490266)*x[0]**o+(0.693714832775)*x[0]+(0.898278265794)*x[1]**o+(0.345095187159)*x[1]
        arg[1,2,1,1]=(-0.498607752195)*x[0]**o+(0.84678344253)*x[0]+(0.678992455984)*x[1]**o+(-0.597396262326)*x[1]
        arg[1,2,2,0]=(0.215267998554)*x[0]**o+(0.384479388838)*x[0]+(-0.09078862662)*x[1]**o+(-0.0162185695524)*x[1]
        arg[1,2,2,1]=(-0.335574750173)*x[0]**o+(-0.439933690512)*x[0]+(0.194601817488)*x[1]**o+(0.795181445633)*x[1]
        arg[1,3,0,0]=(0.862180273604)*x[0]**o+(0.276662657248)*x[0]+(-0.908827123328)*x[1]**o+(0.290406540669)*x[1]
        arg[1,3,0,1]=(-0.620441636176)*x[0]**o+(0.448236189204)*x[0]+(-0.503817375869)*x[1]**o+(0.689447469467)*x[1]
        arg[1,3,1,0]=(0.126327572417)*x[0]**o+(-0.476552913848)*x[0]+(0.362273273226)*x[1]**o+(-0.816228830372)*x[1]
        arg[1,3,1,1]=(-0.65091977023)*x[0]**o+(0.0316560718061)*x[0]+(-0.922906273452)*x[1]**o+(-0.584811809954)*x[1]
        arg[1,3,2,0]=(-0.675982262546)*x[0]**o+(0.23518662745)*x[0]+(-0.270229269162)*x[1]**o+(-0.00300656227686)*x[1]
        arg[1,3,2,1]=(0.55691633063)*x[0]**o+(-0.0643123929496)*x[0]+(0.212825067376)*x[1]**o+(-0.0807098593265)*x[1]
        arg[1,4,0,0]=(0.199100268435)*x[0]**o+(-0.358028564112)*x[0]+(0.130301315926)*x[1]**o+(-0.255752658178)*x[1]
        arg[1,4,0,1]=(0.821720568968)*x[0]**o+(0.193634063182)*x[0]+(0.862353589302)*x[1]**o+(-0.42971617129)*x[1]
        arg[1,4,1,0]=(0.51160977767)*x[0]**o+(0.326743318954)*x[0]+(-0.866878934278)*x[1]**o+(-0.927585776234)*x[1]
        arg[1,4,1,1]=(0.299835915861)*x[0]**o+(-0.969027084981)*x[0]+(-0.589772230832)*x[1]**o+(0.382433302029)*x[1]
        arg[1,4,2,0]=(0.473101160384)*x[0]**o+(-0.591124830283)*x[0]+(0.0929141413982)*x[1]**o+(-0.529811769054)*x[1]
        arg[1,4,2,1]=(-0.712363897962)*x[0]**o+(-0.259229026597)*x[0]+(0.00983651162685)*x[1]**o+(-0.329312308231)*x[1]
        arg[2,0,0,0]=(-0.507414918232)*x[0]**o+(0.668167664289)*x[0]+(0.654870826953)*x[1]**o+(-0.411963068213)*x[1]
        arg[2,0,0,1]=(-0.894915678458)*x[0]**o+(0.647579368695)*x[0]+(0.222274441569)*x[1]**o+(0.737972296743)*x[1]
        arg[2,0,1,0]=(-0.457176106717)*x[0]**o+(0.99997698232)*x[0]+(0.524381951875)*x[1]**o+(0.796302331418)*x[1]
        arg[2,0,1,1]=(-0.761912711581)*x[0]**o+(-0.995235985398)*x[0]+(-0.371926591709)*x[1]**o+(-0.37949019966)*x[1]
        arg[2,0,2,0]=(0.402826798397)*x[0]**o+(0.473547871218)*x[0]+(0.801636706721)*x[1]**o+(0.314221308918)*x[1]
        arg[2,0,2,1]=(0.0416846902183)*x[0]**o+(0.815306853966)*x[0]+(-0.526315166764)*x[1]**o+(-0.327935702076)*x[1]
        arg[2,1,0,0]=(-0.53829655239)*x[0]**o+(-0.467914397053)*x[0]+(0.392659650792)*x[1]**o+(0.309531884089)*x[1]
        arg[2,1,0,1]=(-0.768324371377)*x[0]**o+(0.136336455054)*x[0]+(0.0653485752909)*x[1]**o+(-0.0181923467911)*x[1]
        arg[2,1,1,0]=(0.121561771322)*x[0]**o+(-0.794575616334)*x[0]+(0.038194779569)*x[1]**o+(-0.695864614551)*x[1]
        arg[2,1,1,1]=(0.941872575674)*x[0]**o+(0.322067879395)*x[0]+(0.176873583908)*x[1]**o+(0.266964433323)*x[1]
        arg[2,1,2,0]=(0.324335817101)*x[0]**o+(0.609048475305)*x[0]+(0.350973789003)*x[1]**o+(-0.284995830372)*x[1]
        arg[2,1,2,1]=(0.939941132515)*x[0]**o+(0.782468124836)*x[0]+(-0.238539378134)*x[1]**o+(-0.515203802574)*x[1]
        arg[2,2,0,0]=(0.080179317168)*x[0]**o+(-0.555769741953)*x[0]+(0.283304726007)*x[1]**o+(0.844935300942)*x[1]
        arg[2,2,0,1]=(0.52001819752)*x[0]**o+(-0.105912273152)*x[0]+(-0.360523424917)*x[1]**o+(-0.179135541646)*x[1]
        arg[2,2,1,0]=(0.719593478448)*x[0]**o+(-0.722945454687)*x[0]+(0.0316404625439)*x[1]**o+(0.154290328891)*x[1]
        arg[2,2,1,1]=(0.906382302919)*x[0]**o+(-0.137753840588)*x[0]+(-0.626417865101)*x[1]**o+(-0.890642745149)*x[1]
        arg[2,2,2,0]=(0.618763086671)*x[0]**o+(-0.267876013516)*x[0]+(0.582353922702)*x[1]**o+(-0.190278788501)*x[1]
        arg[2,2,2,1]=(0.778824546516)*x[0]**o+(-0.00837218512008)*x[0]+(0.576376019084)*x[1]**o+(-0.173470710629)*x[1]
        arg[2,3,0,0]=(0.351742111948)*x[0]**o+(-0.641801256932)*x[0]+(-0.979232109277)*x[1]**o+(0.973197715414)*x[1]
        arg[2,3,0,1]=(-0.194493570182)*x[0]**o+(-0.24088269914)*x[0]+(-0.594521666623)*x[1]**o+(0.442357460052)*x[1]
        arg[2,3,1,0]=(-0.0677209625916)*x[0]**o+(-0.329246651504)*x[0]+(-0.200109031557)*x[1]**o+(0.0666064469254)*x[1]
        arg[2,3,1,1]=(-0.31026797404)*x[0]**o+(-0.0420778789656)*x[0]+(-0.534104800904)*x[1]**o+(0.940011883171)*x[1]
        arg[2,3,2,0]=(0.572825859789)*x[0]**o+(-0.439819027899)*x[0]+(-0.641214796724)*x[1]**o+(0.281317897435)*x[1]
        arg[2,3,2,1]=(0.410920070755)*x[0]**o+(-0.517259918184)*x[0]+(0.436420469387)*x[1]**o+(0.212276585342)*x[1]
        arg[2,4,0,0]=(0.0537540090315)*x[0]**o+(-0.362577123782)*x[0]+(-0.0134992480023)*x[1]**o+(0.59055317928)*x[1]
        arg[2,4,0,1]=(0.894896469792)*x[0]**o+(-0.594219811149)*x[0]+(0.755490579022)*x[1]**o+(-0.222913251691)*x[1]
        arg[2,4,1,0]=(-0.95471806813)*x[0]**o+(0.364260591437)*x[0]+(-0.51232745259)*x[1]**o+(-0.874795288001)*x[1]
        arg[2,4,1,1]=(-0.146483465086)*x[0]**o+(0.182495641134)*x[0]+(-0.132211746968)*x[1]**o+(0.39915899917)*x[1]
        arg[2,4,2,0]=(0.630856642463)*x[0]**o+(0.568438094437)*x[0]+(0.283670966088)*x[1]**o+(0.200206672418)*x[1]
        arg[2,4,2,1]=(-0.181938802717)*x[0]**o+(0.850684926475)*x[0]+(-0.765889361148)*x[1]**o+(0.750762858672)*x[1]
        arg[3,0,0,0]=(0.618955933151)*x[0]**o+(0.0280502885803)*x[0]+(-0.518282623282)*x[1]**o+(-0.741257199588)*x[1]
        arg[3,0,0,1]=(0.122462861306)*x[0]**o+(-0.103190806023)*x[0]+(0.647959918767)*x[1]**o+(0.531727405032)*x[1]
        arg[3,0,1,0]=(-0.129468664704)*x[0]**o+(0.342288003693)*x[0]+(-0.0268246756133)*x[1]**o+(0.671147920273)*x[1]
        arg[3,0,1,1]=(-0.99153395274)*x[0]**o+(0.584874556271)*x[0]+(-0.29273067599)*x[1]**o+(-0.802100630404)*x[1]
        arg[3,0,2,0]=(0.406033817569)*x[0]**o+(-0.664112618189)*x[0]+(0.970253482906)*x[1]**o+(0.979709129918)*x[1]
        arg[3,0,2,1]=(-0.708798148641)*x[0]**o+(-0.574259437301)*x[0]+(0.305353886124)*x[1]**o+(0.90520799029)*x[1]
        arg[3,1,0,0]=(-0.296777292529)*x[0]**o+(0.788817860427)*x[0]+(-0.817934426475)*x[1]**o+(-0.480405944642)*x[1]
        arg[3,1,0,1]=(0.382866653892)*x[0]**o+(-0.0911340124698)*x[0]+(0.449555911305)*x[1]**o+(0.972590525904)*x[1]
        arg[3,1,1,0]=(0.150349675921)*x[0]**o+(0.316401916978)*x[0]+(0.176755455021)*x[1]**o+(-0.895165552576)*x[1]
        arg[3,1,1,1]=(-0.334194603044)*x[0]**o+(0.40353334527)*x[0]+(0.510619331296)*x[1]**o+(-0.423857458769)*x[1]
        arg[3,1,2,0]=(0.876266380991)*x[0]**o+(-0.1146693887)*x[0]+(0.0388933985336)*x[1]**o+(0.537276050219)*x[1]
        arg[3,1,2,1]=(0.953489127278)*x[0]**o+(-0.40088917811)*x[0]+(0.322112015322)*x[1]**o+(0.570451651396)*x[1]
        arg[3,2,0,0]=(-0.392674815471)*x[0]**o+(0.805725364633)*x[0]+(-0.550131544344)*x[1]**o+(-0.902042986324)*x[1]
        arg[3,2,0,1]=(0.996009652754)*x[0]**o+(0.534022968708)*x[0]+(-0.251810035763)*x[1]**o+(-0.324231235105)*x[1]
        arg[3,2,1,0]=(-0.280691594637)*x[0]**o+(0.716441336919)*x[0]+(0.444717298591)*x[1]**o+(-0.417682576675)*x[1]
        arg[3,2,1,1]=(-0.528033696021)*x[0]**o+(0.105118392268)*x[0]+(0.219345332231)*x[1]**o+(-0.554011705858)*x[1]
        arg[3,2,2,0]=(0.869967952526)*x[0]**o+(-0.285204973916)*x[0]+(0.676958471189)*x[1]**o+(-0.109813007662)*x[1]
        arg[3,2,2,1]=(-0.183338615818)*x[0]**o+(0.124214512553)*x[0]+(-0.00262420395052)*x[1]**o+(0.505099322191)*x[1]
        arg[3,3,0,0]=(-0.77019083751)*x[0]**o+(-0.288304317474)*x[0]+(-0.674724525333)*x[1]**o+(0.6230658909)*x[1]
        arg[3,3,0,1]=(0.4027332007)*x[0]**o+(0.0181383553658)*x[0]+(0.0346834190431)*x[1]**o+(-0.343302420746)*x[1]
        arg[3,3,1,0]=(0.0907452191096)*x[0]**o+(-0.734197819823)*x[0]+(-0.262536956938)*x[1]**o+(-0.838964523281)*x[1]
        arg[3,3,1,1]=(0.362001847336)*x[0]**o+(-0.99196319473)*x[0]+(-0.544565811529)*x[1]**o+(0.0460139495686)*x[1]
        arg[3,3,2,0]=(0.43659913886)*x[0]**o+(-0.42986865456)*x[0]+(-0.298749218949)*x[1]**o+(0.927334871957)*x[1]
        arg[3,3,2,1]=(0.400272592051)*x[0]**o+(0.332756541456)*x[0]+(0.971127734155)*x[1]**o+(0.653965425034)*x[1]
        arg[3,4,0,0]=(0.663781122153)*x[0]**o+(0.697898766724)*x[0]+(0.157432061303)*x[1]**o+(-0.0526003145497)*x[1]
        arg[3,4,0,1]=(-0.463791084222)*x[0]**o+(0.384914785855)*x[0]+(-0.996485418809)*x[1]**o+(-0.519106290889)*x[1]
        arg[3,4,1,0]=(-0.598068247392)*x[0]**o+(0.553482523952)*x[0]+(0.202485340343)*x[1]**o+(-0.540515409196)*x[1]
        arg[3,4,1,1]=(0.134764979841)*x[0]**o+(0.0141136152488)*x[0]+(-0.536476467366)*x[1]**o+(0.323150057577)*x[1]
        arg[3,4,2,0]=(-0.73096888708)*x[0]**o+(0.485529166207)*x[0]+(-0.541219829073)*x[1]**o+(0.471443044424)*x[1]
        arg[3,4,2,1]=(-0.815808733094)*x[0]**o+(0.694490460113)*x[0]+(0.292301816313)*x[1]**o+(0.700014734893)*x[1]
        ref[0,0,0,0]=(-0.937777610663)*x_ref[0]**o+(-0.0228520946632)*x_ref[0]+(0.642385396583)*x_ref[1]**o+(-0.803821945812)*x_ref[1]
        ref[0,0,0,1]=(-0.398785307492)*x_ref[0]**o+(0.0450288952883)*x_ref[0]+(-0.777780548133)*x_ref[1]**o+(0.0273772099414)*x_ref[1]
        ref[0,0,1,0]=(-0.667280086925)*x_ref[0]**o+(-0.934642710274)*x_ref[0]+(0.565327494491)*x_ref[1]**o+(0.988521769056)*x_ref[1]
        ref[0,0,1,1]=(-0.225562383498)*x_ref[0]**o+(-0.280566031134)*x_ref[0]+(-0.295768954831)*x_ref[1]**o+(0.102759013802)*x_ref[1]
        ref[0,0,2,0]=(-0.179577502546)*x_ref[0]**o+(-0.156932777074)*x_ref[0]+(-0.145616153633)*x_ref[1]**o+(0.971876899135)*x_ref[1]
        ref[0,0,2,1]=(-0.840266409118)*x_ref[0]**o+(0.704356785022)*x_ref[0]+(0.740091470711)*x_ref[1]**o+(0.708478499023)*x_ref[1]
        ref[0,1,0,0]=(-0.487161662806)*x_ref[0]**o+(0.565313908399)*x_ref[0]+(-0.33708697293)*x_ref[1]**o+(0.731675966546)*x_ref[1]
        ref[0,1,0,1]=(-0.135615857963)*x_ref[0]**o+(0.265408777025)*x_ref[0]+(-0.168207260227)*x_ref[1]**o+(-0.598974659317)*x_ref[1]
        ref[0,1,1,0]=(0.0533944010258)*x_ref[0]**o+(0.819248906738)*x_ref[0]+(-0.837005251524)*x_ref[1]**o+(0.119613618595)*x_ref[1]
        ref[0,1,1,1]=(-0.257520627047)*x_ref[0]**o+(0.339264545713)*x_ref[0]+(0.137721228188)*x_ref[1]**o+(-0.512251030979)*x_ref[1]
        ref[0,1,2,0]=(-0.797070396749)*x_ref[0]**o+(0.690173285077)*x_ref[0]+(-0.213206187271)*x_ref[1]**o+(-0.873999125945)*x_ref[1]
        ref[0,1,2,1]=(-0.847662300367)*x_ref[0]**o+(0.785781114266)*x_ref[0]+(-0.456918398811)*x_ref[1]**o+(0.458881079947)*x_ref[1]
        ref[0,2,0,0]=(-0.385269163334)*x_ref[0]**o+(0.135941228964)*x_ref[0]+(0.233245620648)*x_ref[1]**o+(-0.246282209626)*x_ref[1]
        ref[0,2,0,1]=(0.372914636218)*x_ref[0]**o+(-0.160283657003)*x_ref[0]+(0.0565043889781)*x_ref[1]**o+(0.199715522021)*x_ref[1]
        ref[0,2,1,0]=(0.676198390598)*x_ref[0]**o+(0.938025122014)*x_ref[0]+(0.161074696651)*x_ref[1]**o+(-0.155495646964)*x_ref[1]
        ref[0,2,1,1]=(-0.709286924155)*x_ref[0]**o+(-0.173716284776)*x_ref[0]+(0.0576435484694)*x_ref[1]**o+(0.673924697832)*x_ref[1]
        ref[0,2,2,0]=(-0.436468652969)*x_ref[0]**o+(0.117279009082)*x_ref[0]+(-0.731436369191)*x_ref[1]**o+(0.919578503469)*x_ref[1]
        ref[0,2,2,1]=(-0.691042646151)*x_ref[0]**o+(-0.606293725574)*x_ref[0]+(0.657291692759)*x_ref[1]**o+(-0.0146398994729)*x_ref[1]
        ref[0,3,0,0]=(0.365939313719)*x_ref[0]**o+(-0.221236913792)*x_ref[0]+(-0.729594967883)*x_ref[1]**o+(0.0179024719114)*x_ref[1]
        ref[0,3,0,1]=(0.135892133258)*x_ref[0]**o+(0.728069461936)*x_ref[0]+(-0.365691795279)*x_ref[1]**o+(0.877708599922)*x_ref[1]
        ref[0,3,1,0]=(-0.853312456817)*x_ref[0]**o+(-0.713031969049)*x_ref[0]+(0.274755827464)*x_ref[1]**o+(-0.527330699033)*x_ref[1]
        ref[0,3,1,1]=(-0.30358584687)*x_ref[0]**o+(0.871629391968)*x_ref[0]+(0.906915162645)*x_ref[1]**o+(0.499955429041)*x_ref[1]
        ref[0,3,2,0]=(-0.880532587427)*x_ref[0]**o+(0.265729672258)*x_ref[0]+(0.528677136349)*x_ref[1]**o+(-0.224602964925)*x_ref[1]
        ref[0,3,2,1]=(0.158632463532)*x_ref[0]**o+(-0.665418329643)*x_ref[0]+(0.957661069067)*x_ref[1]**o+(-0.23771309434)*x_ref[1]
        ref[0,4,0,0]=(-0.128073514905)*x_ref[0]**o+(-0.452543798055)*x_ref[0]+(0.303808626277)*x_ref[1]**o+(0.841041779759)*x_ref[1]
        ref[0,4,0,1]=(0.0219309964781)*x_ref[0]**o+(-0.949488393443)*x_ref[0]+(0.0436905404663)*x_ref[1]**o+(-0.304527242795)*x_ref[1]
        ref[0,4,1,0]=(-0.792275137775)*x_ref[0]**o+(0.450079829644)*x_ref[0]+(0.162879981637)*x_ref[1]**o+(-0.185839015805)*x_ref[1]
        ref[0,4,1,1]=(0.761932713893)*x_ref[0]**o+(0.932601232945)*x_ref[0]+(-0.606328305783)*x_ref[1]**o+(-0.123973545682)*x_ref[1]
        ref[0,4,2,0]=(-0.514088304908)*x_ref[0]**o+(0.987668774807)*x_ref[0]+(0.866078887168)*x_ref[1]**o+(-0.608518224484)*x_ref[1]
        ref[0,4,2,1]=(-0.427564608028)*x_ref[0]**o+(0.231599614445)*x_ref[0]+(-0.406862229819)*x_ref[1]**o+(0.358456864265)*x_ref[1]
        ref[1,0,0,0]=(0.342092171801)*x_ref[0]**o+(-0.264522675185)*x_ref[0]+(-0.37976917867)*x_ref[1]**o+(0.778596602561)*x_ref[1]
        ref[1,0,0,1]=(0.672156473235)*x_ref[0]**o+(0.320317500856)*x_ref[0]+(0.514206881296)*x_ref[1]**o+(0.602594631705)*x_ref[1]
        ref[1,0,1,0]=(0.914472920293)*x_ref[0]**o+(0.525647943447)*x_ref[0]+(0.685613621597)*x_ref[1]**o+(0.482308039231)*x_ref[1]
        ref[1,0,1,1]=(-0.727982874425)*x_ref[0]**o+(-0.417964463065)*x_ref[0]+(0.825853988063)*x_ref[1]**o+(0.236134760252)*x_ref[1]
        ref[1,0,2,0]=(0.922186381756)*x_ref[0]**o+(-0.140522762044)*x_ref[0]+(-0.282169431666)*x_ref[1]**o+(0.778383988397)*x_ref[1]
        ref[1,0,2,1]=(0.307296736289)*x_ref[0]**o+(0.589844868569)*x_ref[0]+(0.64571457752)*x_ref[1]**o+(0.749333079485)*x_ref[1]
        ref[1,1,0,0]=(0.908345176703)*x_ref[0]**o+(-0.476053147164)*x_ref[0]+(0.0828501189289)*x_ref[1]**o+(0.280657117705)*x_ref[1]
        ref[1,1,0,1]=(0.396128814097)*x_ref[0]**o+(0.368821575482)*x_ref[0]+(-0.916584602334)*x_ref[1]**o+(-0.149377992159)*x_ref[1]
        ref[1,1,1,0]=(-0.38468474313)*x_ref[0]**o+(0.304742499301)*x_ref[0]+(0.638083865846)*x_ref[1]**o+(-0.332046460663)*x_ref[1]
        ref[1,1,1,1]=(0.845097787138)*x_ref[0]**o+(-0.34949498164)*x_ref[0]+(0.229157202572)*x_ref[1]**o+(0.677584672856)*x_ref[1]
        ref[1,1,2,0]=(0.0280692825358)*x_ref[0]**o+(0.567451370107)*x_ref[0]+(0.66934188305)*x_ref[1]**o+(-0.477197213188)*x_ref[1]
        ref[1,1,2,1]=(0.375411487264)*x_ref[0]**o+(0.55086548839)*x_ref[0]+(-0.392135351366)*x_ref[1]**o+(0.91971334589)*x_ref[1]
        ref[1,2,0,0]=(-0.7659270368)*x_ref[0]**o+(-0.338082792509)*x_ref[0]+(-0.26272975232)*x_ref[1]**o+(0.267936454568)*x_ref[1]
        ref[1,2,0,1]=(0.918125455771)*x_ref[0]**o+(0.182896655906)*x_ref[0]+(0.525396835337)*x_ref[1]**o+(0.201918789801)*x_ref[1]
        ref[1,2,1,0]=(-0.54472490266)*x_ref[0]**o+(0.693714832775)*x_ref[0]+(0.898278265794)*x_ref[1]**o+(0.345095187159)*x_ref[1]
        ref[1,2,1,1]=(-0.498607752195)*x_ref[0]**o+(0.84678344253)*x_ref[0]+(0.678992455984)*x_ref[1]**o+(-0.597396262326)*x_ref[1]
        ref[1,2,2,0]=(0.215267998554)*x_ref[0]**o+(0.384479388838)*x_ref[0]+(-0.09078862662)*x_ref[1]**o+(-0.0162185695524)*x_ref[1]
        ref[1,2,2,1]=(-0.335574750173)*x_ref[0]**o+(-0.439933690512)*x_ref[0]+(0.194601817488)*x_ref[1]**o+(0.795181445633)*x_ref[1]
        ref[1,3,0,0]=(0.862180273604)*x_ref[0]**o+(0.276662657248)*x_ref[0]+(-0.908827123328)*x_ref[1]**o+(0.290406540669)*x_ref[1]
        ref[1,3,0,1]=(-0.620441636176)*x_ref[0]**o+(0.448236189204)*x_ref[0]+(-0.503817375869)*x_ref[1]**o+(0.689447469467)*x_ref[1]
        ref[1,3,1,0]=(0.126327572417)*x_ref[0]**o+(-0.476552913848)*x_ref[0]+(0.362273273226)*x_ref[1]**o+(-0.816228830372)*x_ref[1]
        ref[1,3,1,1]=(-0.65091977023)*x_ref[0]**o+(0.0316560718061)*x_ref[0]+(-0.922906273452)*x_ref[1]**o+(-0.584811809954)*x_ref[1]
        ref[1,3,2,0]=(-0.675982262546)*x_ref[0]**o+(0.23518662745)*x_ref[0]+(-0.270229269162)*x_ref[1]**o+(-0.00300656227686)*x_ref[1]
        ref[1,3,2,1]=(0.55691633063)*x_ref[0]**o+(-0.0643123929496)*x_ref[0]+(0.212825067376)*x_ref[1]**o+(-0.0807098593265)*x_ref[1]
        ref[1,4,0,0]=(0.199100268435)*x_ref[0]**o+(-0.358028564112)*x_ref[0]+(0.130301315926)*x_ref[1]**o+(-0.255752658178)*x_ref[1]
        ref[1,4,0,1]=(0.821720568968)*x_ref[0]**o+(0.193634063182)*x_ref[0]+(0.862353589302)*x_ref[1]**o+(-0.42971617129)*x_ref[1]
        ref[1,4,1,0]=(0.51160977767)*x_ref[0]**o+(0.326743318954)*x_ref[0]+(-0.866878934278)*x_ref[1]**o+(-0.927585776234)*x_ref[1]
        ref[1,4,1,1]=(0.299835915861)*x_ref[0]**o+(-0.969027084981)*x_ref[0]+(-0.589772230832)*x_ref[1]**o+(0.382433302029)*x_ref[1]
        ref[1,4,2,0]=(0.473101160384)*x_ref[0]**o+(-0.591124830283)*x_ref[0]+(0.0929141413982)*x_ref[1]**o+(-0.529811769054)*x_ref[1]
        ref[1,4,2,1]=(-0.712363897962)*x_ref[0]**o+(-0.259229026597)*x_ref[0]+(0.00983651162685)*x_ref[1]**o+(-0.329312308231)*x_ref[1]
        ref[2,0,0,0]=(-0.507414918232)*x_ref[0]**o+(0.668167664289)*x_ref[0]+(0.654870826953)*x_ref[1]**o+(-0.411963068213)*x_ref[1]
        ref[2,0,0,1]=(-0.894915678458)*x_ref[0]**o+(0.647579368695)*x_ref[0]+(0.222274441569)*x_ref[1]**o+(0.737972296743)*x_ref[1]
        ref[2,0,1,0]=(-0.457176106717)*x_ref[0]**o+(0.99997698232)*x_ref[0]+(0.524381951875)*x_ref[1]**o+(0.796302331418)*x_ref[1]
        ref[2,0,1,1]=(-0.761912711581)*x_ref[0]**o+(-0.995235985398)*x_ref[0]+(-0.371926591709)*x_ref[1]**o+(-0.37949019966)*x_ref[1]
        ref[2,0,2,0]=(0.402826798397)*x_ref[0]**o+(0.473547871218)*x_ref[0]+(0.801636706721)*x_ref[1]**o+(0.314221308918)*x_ref[1]
        ref[2,0,2,1]=(0.0416846902183)*x_ref[0]**o+(0.815306853966)*x_ref[0]+(-0.526315166764)*x_ref[1]**o+(-0.327935702076)*x_ref[1]
        ref[2,1,0,0]=(-0.53829655239)*x_ref[0]**o+(-0.467914397053)*x_ref[0]+(0.392659650792)*x_ref[1]**o+(0.309531884089)*x_ref[1]
        ref[2,1,0,1]=(-0.768324371377)*x_ref[0]**o+(0.136336455054)*x_ref[0]+(0.0653485752909)*x_ref[1]**o+(-0.0181923467911)*x_ref[1]
        ref[2,1,1,0]=(0.121561771322)*x_ref[0]**o+(-0.794575616334)*x_ref[0]+(0.038194779569)*x_ref[1]**o+(-0.695864614551)*x_ref[1]
        ref[2,1,1,1]=(0.941872575674)*x_ref[0]**o+(0.322067879395)*x_ref[0]+(0.176873583908)*x_ref[1]**o+(0.266964433323)*x_ref[1]
        ref[2,1,2,0]=(0.324335817101)*x_ref[0]**o+(0.609048475305)*x_ref[0]+(0.350973789003)*x_ref[1]**o+(-0.284995830372)*x_ref[1]
        ref[2,1,2,1]=(0.939941132515)*x_ref[0]**o+(0.782468124836)*x_ref[0]+(-0.238539378134)*x_ref[1]**o+(-0.515203802574)*x_ref[1]
        ref[2,2,0,0]=(0.080179317168)*x_ref[0]**o+(-0.555769741953)*x_ref[0]+(0.283304726007)*x_ref[1]**o+(0.844935300942)*x_ref[1]
        ref[2,2,0,1]=(0.52001819752)*x_ref[0]**o+(-0.105912273152)*x_ref[0]+(-0.360523424917)*x_ref[1]**o+(-0.179135541646)*x_ref[1]
        ref[2,2,1,0]=(0.719593478448)*x_ref[0]**o+(-0.722945454687)*x_ref[0]+(0.0316404625439)*x_ref[1]**o+(0.154290328891)*x_ref[1]
        ref[2,2,1,1]=(0.906382302919)*x_ref[0]**o+(-0.137753840588)*x_ref[0]+(-0.626417865101)*x_ref[1]**o+(-0.890642745149)*x_ref[1]
        ref[2,2,2,0]=(0.618763086671)*x_ref[0]**o+(-0.267876013516)*x_ref[0]+(0.582353922702)*x_ref[1]**o+(-0.190278788501)*x_ref[1]
        ref[2,2,2,1]=(0.778824546516)*x_ref[0]**o+(-0.00837218512008)*x_ref[0]+(0.576376019084)*x_ref[1]**o+(-0.173470710629)*x_ref[1]
        ref[2,3,0,0]=(0.351742111948)*x_ref[0]**o+(-0.641801256932)*x_ref[0]+(-0.979232109277)*x_ref[1]**o+(0.973197715414)*x_ref[1]
        ref[2,3,0,1]=(-0.194493570182)*x_ref[0]**o+(-0.24088269914)*x_ref[0]+(-0.594521666623)*x_ref[1]**o+(0.442357460052)*x_ref[1]
        ref[2,3,1,0]=(-0.0677209625916)*x_ref[0]**o+(-0.329246651504)*x_ref[0]+(-0.200109031557)*x_ref[1]**o+(0.0666064469254)*x_ref[1]
        ref[2,3,1,1]=(-0.31026797404)*x_ref[0]**o+(-0.0420778789656)*x_ref[0]+(-0.534104800904)*x_ref[1]**o+(0.940011883171)*x_ref[1]
        ref[2,3,2,0]=(0.572825859789)*x_ref[0]**o+(-0.439819027899)*x_ref[0]+(-0.641214796724)*x_ref[1]**o+(0.281317897435)*x_ref[1]
        ref[2,3,2,1]=(0.410920070755)*x_ref[0]**o+(-0.517259918184)*x_ref[0]+(0.436420469387)*x_ref[1]**o+(0.212276585342)*x_ref[1]
        ref[2,4,0,0]=(0.0537540090315)*x_ref[0]**o+(-0.362577123782)*x_ref[0]+(-0.0134992480023)*x_ref[1]**o+(0.59055317928)*x_ref[1]
        ref[2,4,0,1]=(0.894896469792)*x_ref[0]**o+(-0.594219811149)*x_ref[0]+(0.755490579022)*x_ref[1]**o+(-0.222913251691)*x_ref[1]
        ref[2,4,1,0]=(-0.95471806813)*x_ref[0]**o+(0.364260591437)*x_ref[0]+(-0.51232745259)*x_ref[1]**o+(-0.874795288001)*x_ref[1]
        ref[2,4,1,1]=(-0.146483465086)*x_ref[0]**o+(0.182495641134)*x_ref[0]+(-0.132211746968)*x_ref[1]**o+(0.39915899917)*x_ref[1]
        ref[2,4,2,0]=(0.630856642463)*x_ref[0]**o+(0.568438094437)*x_ref[0]+(0.283670966088)*x_ref[1]**o+(0.200206672418)*x_ref[1]
        ref[2,4,2,1]=(-0.181938802717)*x_ref[0]**o+(0.850684926475)*x_ref[0]+(-0.765889361148)*x_ref[1]**o+(0.750762858672)*x_ref[1]
        ref[3,0,0,0]=(0.618955933151)*x_ref[0]**o+(0.0280502885803)*x_ref[0]+(-0.518282623282)*x_ref[1]**o+(-0.741257199588)*x_ref[1]
        ref[3,0,0,1]=(0.122462861306)*x_ref[0]**o+(-0.103190806023)*x_ref[0]+(0.647959918767)*x_ref[1]**o+(0.531727405032)*x_ref[1]
        ref[3,0,1,0]=(-0.129468664704)*x_ref[0]**o+(0.342288003693)*x_ref[0]+(-0.0268246756133)*x_ref[1]**o+(0.671147920273)*x_ref[1]
        ref[3,0,1,1]=(-0.99153395274)*x_ref[0]**o+(0.584874556271)*x_ref[0]+(-0.29273067599)*x_ref[1]**o+(-0.802100630404)*x_ref[1]
        ref[3,0,2,0]=(0.406033817569)*x_ref[0]**o+(-0.664112618189)*x_ref[0]+(0.970253482906)*x_ref[1]**o+(0.979709129918)*x_ref[1]
        ref[3,0,2,1]=(-0.708798148641)*x_ref[0]**o+(-0.574259437301)*x_ref[0]+(0.305353886124)*x_ref[1]**o+(0.90520799029)*x_ref[1]
        ref[3,1,0,0]=(-0.296777292529)*x_ref[0]**o+(0.788817860427)*x_ref[0]+(-0.817934426475)*x_ref[1]**o+(-0.480405944642)*x_ref[1]
        ref[3,1,0,1]=(0.382866653892)*x_ref[0]**o+(-0.0911340124698)*x_ref[0]+(0.449555911305)*x_ref[1]**o+(0.972590525904)*x_ref[1]
        ref[3,1,1,0]=(0.150349675921)*x_ref[0]**o+(0.316401916978)*x_ref[0]+(0.176755455021)*x_ref[1]**o+(-0.895165552576)*x_ref[1]
        ref[3,1,1,1]=(-0.334194603044)*x_ref[0]**o+(0.40353334527)*x_ref[0]+(0.510619331296)*x_ref[1]**o+(-0.423857458769)*x_ref[1]
        ref[3,1,2,0]=(0.876266380991)*x_ref[0]**o+(-0.1146693887)*x_ref[0]+(0.0388933985336)*x_ref[1]**o+(0.537276050219)*x_ref[1]
        ref[3,1,2,1]=(0.953489127278)*x_ref[0]**o+(-0.40088917811)*x_ref[0]+(0.322112015322)*x_ref[1]**o+(0.570451651396)*x_ref[1]
        ref[3,2,0,0]=(-0.392674815471)*x_ref[0]**o+(0.805725364633)*x_ref[0]+(-0.550131544344)*x_ref[1]**o+(-0.902042986324)*x_ref[1]
        ref[3,2,0,1]=(0.996009652754)*x_ref[0]**o+(0.534022968708)*x_ref[0]+(-0.251810035763)*x_ref[1]**o+(-0.324231235105)*x_ref[1]
        ref[3,2,1,0]=(-0.280691594637)*x_ref[0]**o+(0.716441336919)*x_ref[0]+(0.444717298591)*x_ref[1]**o+(-0.417682576675)*x_ref[1]
        ref[3,2,1,1]=(-0.528033696021)*x_ref[0]**o+(0.105118392268)*x_ref[0]+(0.219345332231)*x_ref[1]**o+(-0.554011705858)*x_ref[1]
        ref[3,2,2,0]=(0.869967952526)*x_ref[0]**o+(-0.285204973916)*x_ref[0]+(0.676958471189)*x_ref[1]**o+(-0.109813007662)*x_ref[1]
        ref[3,2,2,1]=(-0.183338615818)*x_ref[0]**o+(0.124214512553)*x_ref[0]+(-0.00262420395052)*x_ref[1]**o+(0.505099322191)*x_ref[1]
        ref[3,3,0,0]=(-0.77019083751)*x_ref[0]**o+(-0.288304317474)*x_ref[0]+(-0.674724525333)*x_ref[1]**o+(0.6230658909)*x_ref[1]
        ref[3,3,0,1]=(0.4027332007)*x_ref[0]**o+(0.0181383553658)*x_ref[0]+(0.0346834190431)*x_ref[1]**o+(-0.343302420746)*x_ref[1]
        ref[3,3,1,0]=(0.0907452191096)*x_ref[0]**o+(-0.734197819823)*x_ref[0]+(-0.262536956938)*x_ref[1]**o+(-0.838964523281)*x_ref[1]
        ref[3,3,1,1]=(0.362001847336)*x_ref[0]**o+(-0.99196319473)*x_ref[0]+(-0.544565811529)*x_ref[1]**o+(0.0460139495686)*x_ref[1]
        ref[3,3,2,0]=(0.43659913886)*x_ref[0]**o+(-0.42986865456)*x_ref[0]+(-0.298749218949)*x_ref[1]**o+(0.927334871957)*x_ref[1]
        ref[3,3,2,1]=(0.400272592051)*x_ref[0]**o+(0.332756541456)*x_ref[0]+(0.971127734155)*x_ref[1]**o+(0.653965425034)*x_ref[1]
        ref[3,4,0,0]=(0.663781122153)*x_ref[0]**o+(0.697898766724)*x_ref[0]+(0.157432061303)*x_ref[1]**o+(-0.0526003145497)*x_ref[1]
        ref[3,4,0,1]=(-0.463791084222)*x_ref[0]**o+(0.384914785855)*x_ref[0]+(-0.996485418809)*x_ref[1]**o+(-0.519106290889)*x_ref[1]
        ref[3,4,1,0]=(-0.598068247392)*x_ref[0]**o+(0.553482523952)*x_ref[0]+(0.202485340343)*x_ref[1]**o+(-0.540515409196)*x_ref[1]
        ref[3,4,1,1]=(0.134764979841)*x_ref[0]**o+(0.0141136152488)*x_ref[0]+(-0.536476467366)*x_ref[1]**o+(0.323150057577)*x_ref[1]
        ref[3,4,2,0]=(-0.73096888708)*x_ref[0]**o+(0.485529166207)*x_ref[0]+(-0.541219829073)*x_ref[1]**o+(0.471443044424)*x_ref[1]
        ref[3,4,2,1]=(-0.815808733094)*x_ref[0]**o+(0.694490460113)*x_ref[0]+(0.292301816313)*x_ref[1]**o+(0.700014734893)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.244520699379)*x[0]**o+(-0.125771252842)*x[0]+(0.654557798928)*x[1]**o+(0.438110948503)*x[1]+(0.399996169353)*x[2]**o+(0.37037141324)*x[2]
        arg[0,0,0,1]=(0.019693348021)*x[0]**o+(-0.509938412509)*x[0]+(-0.425852907243)*x[1]**o+(0.351462945072)*x[1]+(0.628410872426)*x[2]**o+(0.4769252017)*x[2]
        arg[0,0,1,0]=(0.623813767602)*x[0]**o+(-0.940818977456)*x[0]+(-0.00111063869285)*x[1]**o+(-0.577558081394)*x[1]+(-0.30810971068)*x[2]**o+(-0.660149007373)*x[2]
        arg[0,0,1,1]=(0.994105796371)*x[0]**o+(0.199970906966)*x[0]+(0.961509001434)*x[1]**o+(-0.783069602127)*x[1]+(-0.179922863021)*x[2]**o+(-0.386974503089)*x[2]
        arg[0,0,2,0]=(0.932745137122)*x[0]**o+(0.627995830114)*x[0]+(-0.334344490736)*x[1]**o+(0.733210590815)*x[1]+(-0.23424959051)*x[2]**o+(0.140644214958)*x[2]
        arg[0,0,2,1]=(-0.282836519247)*x[0]**o+(-0.744077669768)*x[0]+(0.123898574243)*x[1]**o+(0.562384289004)*x[1]+(0.0757927394637)*x[2]**o+(0.767684067921)*x[2]
        arg[0,1,0,0]=(0.212366202901)*x[0]**o+(-0.679228304269)*x[0]+(0.829706652419)*x[1]**o+(0.24389996073)*x[1]+(0.909367025978)*x[2]**o+(0.280617790407)*x[2]
        arg[0,1,0,1]=(0.760070952632)*x[0]**o+(-0.678309252372)*x[0]+(0.0372281910363)*x[1]**o+(0.0284582814297)*x[1]+(-0.092349250941)*x[2]**o+(0.347713645921)*x[2]
        arg[0,1,1,0]=(0.248410230191)*x[0]**o+(-0.833983202295)*x[0]+(-0.759357278899)*x[1]**o+(0.880038285893)*x[1]+(-0.734799540652)*x[2]**o+(0.493330231104)*x[2]
        arg[0,1,1,1]=(0.0359919724228)*x[0]**o+(-0.0393899014057)*x[0]+(0.185155782488)*x[1]**o+(-0.92692925719)*x[1]+(0.74718037404)*x[2]**o+(-0.283900681763)*x[2]
        arg[0,1,2,0]=(-0.208643544667)*x[0]**o+(0.368200149096)*x[0]+(-0.327032725172)*x[1]**o+(-0.245547905792)*x[1]+(0.0886488214242)*x[2]**o+(0.179139897691)*x[2]
        arg[0,1,2,1]=(-0.0618762861459)*x[0]**o+(0.253419217897)*x[0]+(0.840270125508)*x[1]**o+(0.0223556878387)*x[1]+(0.0560904700176)*x[2]**o+(-0.646231049424)*x[2]
        arg[0,2,0,0]=(0.568966969865)*x[0]**o+(0.786776511348)*x[0]+(0.150595926116)*x[1]**o+(-0.780164417762)*x[1]+(0.510149831242)*x[2]**o+(-0.902871929903)*x[2]
        arg[0,2,0,1]=(-0.817062669957)*x[0]**o+(0.626805486868)*x[0]+(-0.566129786829)*x[1]**o+(0.248418548905)*x[1]+(-0.273995326019)*x[2]**o+(0.402023521018)*x[2]
        arg[0,2,1,0]=(-0.160276235113)*x[0]**o+(-0.0732543937053)*x[0]+(-0.0651595982698)*x[1]**o+(-0.529463476237)*x[1]+(0.346098812945)*x[2]**o+(0.418367016439)*x[2]
        arg[0,2,1,1]=(0.736907408498)*x[0]**o+(-0.123612748157)*x[0]+(0.000834659108651)*x[1]**o+(0.568833761363)*x[1]+(-0.542676953582)*x[2]**o+(0.026182553484)*x[2]
        arg[0,2,2,0]=(0.320985918151)*x[0]**o+(0.0512654311644)*x[0]+(-0.0313412913713)*x[1]**o+(0.295659930913)*x[1]+(-0.989560405329)*x[2]**o+(-0.107344246298)*x[2]
        arg[0,2,2,1]=(-0.957793939949)*x[0]**o+(-0.177591880789)*x[0]+(0.0785829772689)*x[1]**o+(0.121445956612)*x[1]+(-0.384134176508)*x[2]**o+(-0.775433914622)*x[2]
        arg[0,3,0,0]=(-0.130620609021)*x[0]**o+(0.268283814564)*x[0]+(-0.192835222037)*x[1]**o+(0.235341688541)*x[1]+(0.957069315916)*x[2]**o+(-0.708853567529)*x[2]
        arg[0,3,0,1]=(-0.377028600614)*x[0]**o+(-0.858446162753)*x[0]+(-0.850630580491)*x[1]**o+(-0.810676898206)*x[1]+(-0.613330062817)*x[2]**o+(0.714335738947)*x[2]
        arg[0,3,1,0]=(-0.602336649415)*x[0]**o+(-0.166667559558)*x[0]+(0.962948866081)*x[1]**o+(0.939531483467)*x[1]+(0.267767294244)*x[2]**o+(0.247155872428)*x[2]
        arg[0,3,1,1]=(0.838505391049)*x[0]**o+(0.115204438919)*x[0]+(0.989010581013)*x[1]**o+(-0.649789818418)*x[1]+(0.757328444925)*x[2]**o+(0.825158129449)*x[2]
        arg[0,3,2,0]=(0.724488034506)*x[0]**o+(0.0333093325925)*x[0]+(0.864643652652)*x[1]**o+(-0.557517585925)*x[1]+(-0.340534672063)*x[2]**o+(-0.00425705176271)*x[2]
        arg[0,3,2,1]=(-0.366459483792)*x[0]**o+(0.257706980646)*x[0]+(0.821731531023)*x[1]**o+(-0.212501303617)*x[1]+(-0.128977067795)*x[2]**o+(-0.931439847041)*x[2]
        arg[0,4,0,0]=(0.398430239547)*x[0]**o+(-0.00991917699933)*x[0]+(-0.182471763172)*x[1]**o+(0.218327764946)*x[1]+(-0.484510127696)*x[2]**o+(-0.360678193996)*x[2]
        arg[0,4,0,1]=(0.0640680415135)*x[0]**o+(0.671150063087)*x[0]+(0.208610130381)*x[1]**o+(-0.441252393083)*x[1]+(0.365150146582)*x[2]**o+(-0.228346558142)*x[2]
        arg[0,4,1,0]=(-0.852694824303)*x[0]**o+(0.181041494463)*x[0]+(0.898413617086)*x[1]**o+(-0.918988862595)*x[1]+(0.169010143324)*x[2]**o+(0.540458809834)*x[2]
        arg[0,4,1,1]=(-0.213782550366)*x[0]**o+(-0.180557891041)*x[0]+(0.23144059382)*x[1]**o+(0.0609482143429)*x[1]+(0.89256221946)*x[2]**o+(0.977857158982)*x[2]
        arg[0,4,2,0]=(0.320772898841)*x[0]**o+(0.371610189518)*x[0]+(0.692264150983)*x[1]**o+(-0.751154935415)*x[1]+(-0.163272902772)*x[2]**o+(0.0635386628764)*x[2]
        arg[0,4,2,1]=(-0.316885908577)*x[0]**o+(0.476104431241)*x[0]+(-0.647713072489)*x[1]**o+(0.683813449908)*x[1]+(0.699293342945)*x[2]**o+(-0.0373906770097)*x[2]
        arg[1,0,0,0]=(-0.030329265733)*x[0]**o+(-0.0115761157375)*x[0]+(-0.30993823088)*x[1]**o+(0.628877975389)*x[1]+(-0.582940961417)*x[2]**o+(0.154786570457)*x[2]
        arg[1,0,0,1]=(-0.593610753495)*x[0]**o+(-0.822232097764)*x[0]+(-0.465144000698)*x[1]**o+(-0.62576868204)*x[1]+(-0.53588531407)*x[2]**o+(0.119119578421)*x[2]
        arg[1,0,1,0]=(-0.15668734059)*x[0]**o+(-0.950658811047)*x[0]+(-0.267969935748)*x[1]**o+(0.0919415340304)*x[1]+(-0.0445382452662)*x[2]**o+(0.949984936459)*x[2]
        arg[1,0,1,1]=(-0.762474809436)*x[0]**o+(0.794314442281)*x[0]+(0.617742930372)*x[1]**o+(-0.502057530669)*x[1]+(0.42532154189)*x[2]**o+(0.343417687179)*x[2]
        arg[1,0,2,0]=(0.259583418084)*x[0]**o+(0.460122326434)*x[0]+(0.875185242587)*x[1]**o+(-0.621232091699)*x[1]+(0.922536397008)*x[2]**o+(-0.494729334323)*x[2]
        arg[1,0,2,1]=(0.79625863156)*x[0]**o+(-0.980360881506)*x[0]+(0.947721778185)*x[1]**o+(-0.234201025597)*x[1]+(-0.974251961452)*x[2]**o+(-0.300484605111)*x[2]
        arg[1,1,0,0]=(0.719927358834)*x[0]**o+(-0.296944527139)*x[0]+(0.970290716632)*x[1]**o+(0.45656092862)*x[1]+(0.657023751979)*x[2]**o+(-0.703898365316)*x[2]
        arg[1,1,0,1]=(-0.0600633107738)*x[0]**o+(0.525293730667)*x[0]+(0.17955637601)*x[1]**o+(-0.625782540859)*x[1]+(0.205255859488)*x[2]**o+(-0.963128832741)*x[2]
        arg[1,1,1,0]=(0.119721921725)*x[0]**o+(0.957918016103)*x[0]+(-0.574883179285)*x[1]**o+(0.915580622743)*x[1]+(-0.432538225757)*x[2]**o+(-0.937539574986)*x[2]
        arg[1,1,1,1]=(-0.295639299942)*x[0]**o+(-0.703954000058)*x[0]+(-0.185001166055)*x[1]**o+(-0.943933820897)*x[1]+(-0.12692007135)*x[2]**o+(0.476481616585)*x[2]
        arg[1,1,2,0]=(0.247892010246)*x[0]**o+(0.901198196531)*x[0]+(-0.586162224974)*x[1]**o+(-0.934177552379)*x[1]+(0.704265144051)*x[2]**o+(-0.688632953989)*x[2]
        arg[1,1,2,1]=(-0.116039842334)*x[0]**o+(-0.300373025101)*x[0]+(-0.286684372551)*x[1]**o+(-0.452337760059)*x[1]+(0.883210026596)*x[2]**o+(-0.319739612634)*x[2]
        arg[1,2,0,0]=(-0.0326555626249)*x[0]**o+(-0.0933559852573)*x[0]+(0.807443822382)*x[1]**o+(-0.148842350195)*x[1]+(-0.118904471983)*x[2]**o+(-0.740440145747)*x[2]
        arg[1,2,0,1]=(0.313623768488)*x[0]**o+(0.524958399998)*x[0]+(0.876747001712)*x[1]**o+(-0.468458894694)*x[1]+(0.52118935613)*x[2]**o+(-0.776405429712)*x[2]
        arg[1,2,1,0]=(0.1328367779)*x[0]**o+(-0.959628068869)*x[0]+(0.746000325891)*x[1]**o+(-0.454705589649)*x[1]+(0.08534697763)*x[2]**o+(0.174892567303)*x[2]
        arg[1,2,1,1]=(-0.482594997843)*x[0]**o+(0.765628170342)*x[0]+(0.103333249513)*x[1]**o+(0.840778499017)*x[1]+(0.136748193255)*x[2]**o+(0.771613240623)*x[2]
        arg[1,2,2,0]=(-0.305745376076)*x[0]**o+(-0.0464234361529)*x[0]+(-0.163528204043)*x[1]**o+(0.80782719974)*x[1]+(0.131046315567)*x[2]**o+(0.0849217124905)*x[2]
        arg[1,2,2,1]=(0.564915635291)*x[0]**o+(-0.162467524139)*x[0]+(-0.696721556806)*x[1]**o+(0.0172839552576)*x[1]+(-0.630040402586)*x[2]**o+(-0.181232689117)*x[2]
        arg[1,3,0,0]=(-0.768631958549)*x[0]**o+(0.537076196504)*x[0]+(-0.403489871013)*x[1]**o+(0.831966318632)*x[1]+(0.602387061638)*x[2]**o+(0.330495470463)*x[2]
        arg[1,3,0,1]=(0.770878177942)*x[0]**o+(-0.678609489017)*x[0]+(-0.895200519099)*x[1]**o+(0.935093561205)*x[1]+(0.0497551733469)*x[2]**o+(0.820432808383)*x[2]
        arg[1,3,1,0]=(0.052867263447)*x[0]**o+(0.566167926663)*x[0]+(-0.00397870350455)*x[1]**o+(-0.780374460006)*x[1]+(-0.166645199933)*x[2]**o+(-0.867274057327)*x[2]
        arg[1,3,1,1]=(0.0573446476183)*x[0]**o+(-0.699343104682)*x[0]+(0.428624411932)*x[1]**o+(-0.400834963806)*x[1]+(-0.722750021776)*x[2]**o+(0.386821944027)*x[2]
        arg[1,3,2,0]=(0.390777333115)*x[0]**o+(0.191564882434)*x[0]+(-0.774547811812)*x[1]**o+(0.747678394063)*x[1]+(0.662273739757)*x[2]**o+(-0.295629568405)*x[2]
        arg[1,3,2,1]=(-0.160077661734)*x[0]**o+(-0.512286879466)*x[0]+(0.616824388099)*x[1]**o+(0.0257995114324)*x[1]+(0.644270949659)*x[2]**o+(0.243520087896)*x[2]
        arg[1,4,0,0]=(-0.980501571755)*x[0]**o+(0.844184850883)*x[0]+(-0.608509998911)*x[1]**o+(-0.347687895783)*x[1]+(-0.536534228926)*x[2]**o+(0.632821444072)*x[2]
        arg[1,4,0,1]=(-0.881676821515)*x[0]**o+(-0.858728104026)*x[0]+(-0.576516592622)*x[1]**o+(0.695748958418)*x[1]+(0.856547670877)*x[2]**o+(-0.932960091743)*x[2]
        arg[1,4,1,0]=(-0.451435812217)*x[0]**o+(0.963886330308)*x[0]+(0.888532771273)*x[1]**o+(0.807924978471)*x[1]+(0.789117121354)*x[2]**o+(0.577647188972)*x[2]
        arg[1,4,1,1]=(-0.315320308371)*x[0]**o+(0.182798184972)*x[0]+(0.936963228212)*x[1]**o+(-0.634429372225)*x[1]+(0.634343556662)*x[2]**o+(0.975851394002)*x[2]
        arg[1,4,2,0]=(0.858246326961)*x[0]**o+(0.0955812000864)*x[0]+(0.16143064785)*x[1]**o+(0.397013086552)*x[1]+(-0.507188929963)*x[2]**o+(0.49789827039)*x[2]
        arg[1,4,2,1]=(-0.524634529078)*x[0]**o+(0.675491034226)*x[0]+(0.858857975014)*x[1]**o+(-0.57874207047)*x[1]+(0.476551353232)*x[2]**o+(-0.622601386038)*x[2]
        arg[2,0,0,0]=(0.976563609)*x[0]**o+(0.127133211238)*x[0]+(-0.410155025204)*x[1]**o+(-0.909032700787)*x[1]+(-0.197014506331)*x[2]**o+(0.557171133933)*x[2]
        arg[2,0,0,1]=(0.434874625301)*x[0]**o+(-0.576955025374)*x[0]+(-0.466686136319)*x[1]**o+(-0.359616702831)*x[1]+(-0.916050768111)*x[2]**o+(0.900510057507)*x[2]
        arg[2,0,1,0]=(0.615259852986)*x[0]**o+(-0.685399646912)*x[0]+(-0.222742007887)*x[1]**o+(-0.487739486778)*x[1]+(-0.675225266903)*x[2]**o+(0.494623157154)*x[2]
        arg[2,0,1,1]=(-0.89129010918)*x[0]**o+(0.0051571255592)*x[0]+(0.681003954208)*x[1]**o+(0.634041720719)*x[1]+(-0.308587458722)*x[2]**o+(-0.107801539193)*x[2]
        arg[2,0,2,0]=(-0.660496698398)*x[0]**o+(-0.889315862375)*x[0]+(0.64248433679)*x[1]**o+(-0.941990511222)*x[1]+(-0.173668816201)*x[2]**o+(0.237812802184)*x[2]
        arg[2,0,2,1]=(0.847365004226)*x[0]**o+(0.802449170085)*x[0]+(0.504737931993)*x[1]**o+(-0.695546680342)*x[1]+(0.0742668699455)*x[2]**o+(0.396565506083)*x[2]
        arg[2,1,0,0]=(0.812115742561)*x[0]**o+(-0.528091349418)*x[0]+(-0.442834134691)*x[1]**o+(-0.574414379092)*x[1]+(-0.734077338221)*x[2]**o+(0.159306515061)*x[2]
        arg[2,1,0,1]=(-0.259422255041)*x[0]**o+(-0.104016445541)*x[0]+(0.3560854635)*x[1]**o+(0.997015276785)*x[1]+(-0.707278312293)*x[2]**o+(-0.975235879836)*x[2]
        arg[2,1,1,0]=(0.402099188227)*x[0]**o+(0.0320354114274)*x[0]+(0.203384927114)*x[1]**o+(-0.847804170327)*x[1]+(-0.898459903546)*x[2]**o+(0.639388498254)*x[2]
        arg[2,1,1,1]=(-0.0477209380768)*x[0]**o+(0.531574667575)*x[0]+(0.280889682833)*x[1]**o+(-0.444963671116)*x[1]+(0.33935324652)*x[2]**o+(0.943089036339)*x[2]
        arg[2,1,2,0]=(0.396038211375)*x[0]**o+(-0.54994373568)*x[0]+(0.851017234687)*x[1]**o+(-0.739787661885)*x[1]+(-0.368325897478)*x[2]**o+(-0.449096507837)*x[2]
        arg[2,1,2,1]=(0.857714720302)*x[0]**o+(-0.750668366832)*x[0]+(-0.587901076381)*x[1]**o+(-0.24906126763)*x[1]+(0.131707530536)*x[2]**o+(0.900975507373)*x[2]
        arg[2,2,0,0]=(0.783919932392)*x[0]**o+(-0.967669966647)*x[0]+(0.82052580972)*x[1]**o+(-0.711919131449)*x[1]+(-0.396059626351)*x[2]**o+(-0.265805391355)*x[2]
        arg[2,2,0,1]=(-0.0398160197307)*x[0]**o+(-0.238193541)*x[0]+(0.282277679599)*x[1]**o+(-0.0526835304963)*x[1]+(0.590871489195)*x[2]**o+(-0.925835585742)*x[2]
        arg[2,2,1,0]=(-0.825757767738)*x[0]**o+(0.238488063696)*x[0]+(0.109616380527)*x[1]**o+(-0.942941715764)*x[1]+(0.632013079754)*x[2]**o+(0.0517997920543)*x[2]
        arg[2,2,1,1]=(-0.85343230326)*x[0]**o+(0.225247381624)*x[0]+(0.43171428489)*x[1]**o+(0.0937487532231)*x[1]+(0.894064129869)*x[2]**o+(-0.0876265498234)*x[2]
        arg[2,2,2,0]=(0.349294373111)*x[0]**o+(-0.734159680014)*x[0]+(0.0489227393033)*x[1]**o+(-0.245284468757)*x[1]+(-0.346469825649)*x[2]**o+(0.547054093324)*x[2]
        arg[2,2,2,1]=(-0.0478095998047)*x[0]**o+(0.440261344405)*x[0]+(-0.590330499679)*x[1]**o+(-0.961009986328)*x[1]+(0.546190002905)*x[2]**o+(-0.027965996519)*x[2]
        arg[2,3,0,0]=(0.148563550644)*x[0]**o+(0.576704219545)*x[0]+(-0.782168581109)*x[1]**o+(-0.8301245678)*x[1]+(0.937936245682)*x[2]**o+(0.846740948371)*x[2]
        arg[2,3,0,1]=(-0.447832058524)*x[0]**o+(0.385431388609)*x[0]+(0.193985676964)*x[1]**o+(-0.91717180275)*x[1]+(-0.408732230908)*x[2]**o+(0.548173152118)*x[2]
        arg[2,3,1,0]=(0.197592567864)*x[0]**o+(-0.77657127384)*x[0]+(0.266355135001)*x[1]**o+(0.604009981584)*x[1]+(-0.379930450459)*x[2]**o+(0.46952844479)*x[2]
        arg[2,3,1,1]=(0.996803794962)*x[0]**o+(0.319115676981)*x[0]+(0.255552997875)*x[1]**o+(0.21834578367)*x[1]+(0.389966036243)*x[2]**o+(-0.115308080726)*x[2]
        arg[2,3,2,0]=(-0.0244303880675)*x[0]**o+(-0.128929049017)*x[0]+(-0.350933414406)*x[1]**o+(-0.341716440852)*x[1]+(-0.724579076377)*x[2]**o+(0.21963465519)*x[2]
        arg[2,3,2,1]=(-0.92899971478)*x[0]**o+(-0.304496074717)*x[0]+(0.568151427024)*x[1]**o+(0.504629517922)*x[1]+(0.846158711477)*x[2]**o+(0.35936792568)*x[2]
        arg[2,4,0,0]=(-0.350964834352)*x[0]**o+(0.387667833884)*x[0]+(0.547101274223)*x[1]**o+(0.257218757373)*x[1]+(-0.00316748527841)*x[2]**o+(0.314561238615)*x[2]
        arg[2,4,0,1]=(0.408941739151)*x[0]**o+(0.819113792537)*x[0]+(0.532891848781)*x[1]**o+(-0.30688659725)*x[1]+(-0.736611937618)*x[2]**o+(-0.875783706051)*x[2]
        arg[2,4,1,0]=(-0.730885165985)*x[0]**o+(0.453908506302)*x[0]+(-0.617385752505)*x[1]**o+(0.357459789592)*x[1]+(0.286323175805)*x[2]**o+(-0.448879127031)*x[2]
        arg[2,4,1,1]=(-0.295377341281)*x[0]**o+(0.423424371051)*x[0]+(0.524695076927)*x[1]**o+(-0.698499625038)*x[1]+(0.617308008395)*x[2]**o+(-0.423469866581)*x[2]
        arg[2,4,2,0]=(0.104251379635)*x[0]**o+(0.989709098795)*x[0]+(-0.422075396624)*x[1]**o+(-0.254328561231)*x[1]+(-0.502052795074)*x[2]**o+(-0.0851411652168)*x[2]
        arg[2,4,2,1]=(0.379332842082)*x[0]**o+(-0.707504861783)*x[0]+(0.00855075647146)*x[1]**o+(0.792851159033)*x[1]+(-0.692042160791)*x[2]**o+(-0.18327165243)*x[2]
        arg[3,0,0,0]=(0.310700158718)*x[0]**o+(0.285647240187)*x[0]+(0.817947571838)*x[1]**o+(0.808698040916)*x[1]+(0.844184006443)*x[2]**o+(-0.310383215636)*x[2]
        arg[3,0,0,1]=(-0.0173109952616)*x[0]**o+(0.672034452037)*x[0]+(-0.464572256298)*x[1]**o+(0.205799523299)*x[1]+(-0.971979841338)*x[2]**o+(0.261515966001)*x[2]
        arg[3,0,1,0]=(0.0329384377028)*x[0]**o+(0.571501185072)*x[0]+(-0.902222556642)*x[1]**o+(-0.528317784373)*x[1]+(0.331989640881)*x[2]**o+(0.0427505624318)*x[2]
        arg[3,0,1,1]=(0.964350410761)*x[0]**o+(-0.327198989999)*x[0]+(-0.195271523191)*x[1]**o+(0.534663014601)*x[1]+(-0.0208261208706)*x[2]**o+(-0.158019249052)*x[2]
        arg[3,0,2,0]=(-0.622082726008)*x[0]**o+(0.559463883442)*x[0]+(0.841439693313)*x[1]**o+(-0.196363987516)*x[1]+(-0.0143926547396)*x[2]**o+(0.144732603689)*x[2]
        arg[3,0,2,1]=(-0.149531653278)*x[0]**o+(0.354229138322)*x[0]+(-0.507916088829)*x[1]**o+(0.221251580047)*x[1]+(0.197213706274)*x[2]**o+(-0.555645149276)*x[2]
        arg[3,1,0,0]=(0.331020785374)*x[0]**o+(-0.377647773513)*x[0]+(0.563355928949)*x[1]**o+(0.789309216711)*x[1]+(0.203663383875)*x[2]**o+(-0.0333894540645)*x[2]
        arg[3,1,0,1]=(-0.0965613913882)*x[0]**o+(0.18656484323)*x[0]+(-0.138245941339)*x[1]**o+(-0.34549538745)*x[1]+(0.827173403167)*x[2]**o+(-0.873380503939)*x[2]
        arg[3,1,1,0]=(0.932940890033)*x[0]**o+(-0.564883830288)*x[0]+(-0.505052725295)*x[1]**o+(-0.669420210206)*x[1]+(-0.799638096039)*x[2]**o+(0.469400233749)*x[2]
        arg[3,1,1,1]=(-0.9741761069)*x[0]**o+(-0.748125328781)*x[0]+(0.664448241931)*x[1]**o+(0.0957643296979)*x[1]+(0.112211322019)*x[2]**o+(-0.131870290434)*x[2]
        arg[3,1,2,0]=(-0.438609127704)*x[0]**o+(-0.36618286515)*x[0]+(-0.376420011115)*x[1]**o+(-0.67084624599)*x[1]+(0.821679796644)*x[2]**o+(0.80111319897)*x[2]
        arg[3,1,2,1]=(-0.409859432085)*x[0]**o+(0.848767816929)*x[0]+(0.382358459835)*x[1]**o+(-0.750139843631)*x[1]+(0.587152586559)*x[2]**o+(-0.26162952001)*x[2]
        arg[3,2,0,0]=(-0.506329785074)*x[0]**o+(-0.0967745752004)*x[0]+(-0.163316795268)*x[1]**o+(0.0628481921732)*x[1]+(0.395355494367)*x[2]**o+(0.0978901916332)*x[2]
        arg[3,2,0,1]=(0.803590309587)*x[0]**o+(0.470388053389)*x[0]+(0.674942922168)*x[1]**o+(0.223472834762)*x[1]+(0.552332209028)*x[2]**o+(-0.396443498843)*x[2]
        arg[3,2,1,0]=(-0.312384723839)*x[0]**o+(0.392213828608)*x[0]+(0.449465316671)*x[1]**o+(-0.249103089401)*x[1]+(0.601412913458)*x[2]**o+(0.209899368039)*x[2]
        arg[3,2,1,1]=(-0.583776179557)*x[0]**o+(0.485007047754)*x[0]+(-0.0354153633589)*x[1]**o+(-0.898630969916)*x[1]+(-0.266581492795)*x[2]**o+(0.822086680281)*x[2]
        arg[3,2,2,0]=(0.944617094397)*x[0]**o+(-0.173057994833)*x[0]+(0.927832160048)*x[1]**o+(-0.218599594252)*x[1]+(0.349957755011)*x[2]**o+(0.392602744545)*x[2]
        arg[3,2,2,1]=(0.461648084674)*x[0]**o+(0.826213376949)*x[0]+(0.143429054731)*x[1]**o+(-0.214265128007)*x[1]+(0.0317208127607)*x[2]**o+(-0.858361078381)*x[2]
        arg[3,3,0,0]=(0.306607920368)*x[0]**o+(0.147692891688)*x[0]+(-0.562661766456)*x[1]**o+(0.448595564727)*x[1]+(-0.707178218373)*x[2]**o+(0.0655554587866)*x[2]
        arg[3,3,0,1]=(0.135475820362)*x[0]**o+(0.51420718793)*x[0]+(0.343291220736)*x[1]**o+(-0.797532312602)*x[1]+(-0.273431011273)*x[2]**o+(0.135307779253)*x[2]
        arg[3,3,1,0]=(0.488028231073)*x[0]**o+(0.542021806829)*x[0]+(0.91916268609)*x[1]**o+(0.317368273365)*x[1]+(0.314807039395)*x[2]**o+(-0.841583308025)*x[2]
        arg[3,3,1,1]=(-0.182783814917)*x[0]**o+(0.250863481201)*x[0]+(0.755551457003)*x[1]**o+(0.946171548529)*x[1]+(0.0350225436682)*x[2]**o+(0.776059830052)*x[2]
        arg[3,3,2,0]=(-0.116707124994)*x[0]**o+(-0.727010280588)*x[0]+(0.172065389268)*x[1]**o+(0.754164564639)*x[1]+(0.803777635621)*x[2]**o+(0.293490264417)*x[2]
        arg[3,3,2,1]=(0.154235982797)*x[0]**o+(-0.124286021486)*x[0]+(-0.737324748116)*x[1]**o+(-0.136803541581)*x[1]+(-0.608370191604)*x[2]**o+(-0.933599034128)*x[2]
        arg[3,4,0,0]=(0.818061677328)*x[0]**o+(-0.771719613166)*x[0]+(0.742772795251)*x[1]**o+(0.0359285091312)*x[1]+(-0.985656851837)*x[2]**o+(-0.884598206512)*x[2]
        arg[3,4,0,1]=(0.198371732741)*x[0]**o+(0.646496618681)*x[0]+(0.232912446494)*x[1]**o+(0.945619492023)*x[1]+(-0.344370341731)*x[2]**o+(0.31156840357)*x[2]
        arg[3,4,1,0]=(-0.609298834606)*x[0]**o+(0.784697392412)*x[0]+(0.996065619079)*x[1]**o+(-0.284498300043)*x[1]+(-0.652795706803)*x[2]**o+(-0.254746840396)*x[2]
        arg[3,4,1,1]=(-0.0083632822507)*x[0]**o+(0.200923618933)*x[0]+(0.353883930064)*x[1]**o+(0.153257380179)*x[1]+(0.843965464205)*x[2]**o+(0.0418501400703)*x[2]
        arg[3,4,2,0]=(-0.226055622191)*x[0]**o+(-0.701789166696)*x[0]+(0.274080957143)*x[1]**o+(0.0596744660556)*x[1]+(-0.342083152987)*x[2]**o+(0.593518464697)*x[2]
        arg[3,4,2,1]=(-0.288168945023)*x[0]**o+(-0.742841678774)*x[0]+(0.972629532424)*x[1]**o+(0.892100560733)*x[1]+(0.803880474721)*x[2]**o+(-0.0638336211224)*x[2]
        ref[0,0,0,0]=(0.244520699379)*x_ref[0]**o+(-0.125771252842)*x_ref[0]+(0.654557798928)*x_ref[1]**o+(0.438110948503)*x_ref[1]+(0.399996169353)*x_ref[2]**o+(0.37037141324)*x_ref[2]
        ref[0,0,0,1]=(0.019693348021)*x_ref[0]**o+(-0.509938412509)*x_ref[0]+(-0.425852907243)*x_ref[1]**o+(0.351462945072)*x_ref[1]+(0.628410872426)*x_ref[2]**o+(0.4769252017)*x_ref[2]
        ref[0,0,1,0]=(0.623813767602)*x_ref[0]**o+(-0.940818977456)*x_ref[0]+(-0.00111063869285)*x_ref[1]**o+(-0.577558081394)*x_ref[1]+(-0.30810971068)*x_ref[2]**o+(-0.660149007373)*x_ref[2]
        ref[0,0,1,1]=(0.994105796371)*x_ref[0]**o+(0.199970906966)*x_ref[0]+(0.961509001434)*x_ref[1]**o+(-0.783069602127)*x_ref[1]+(-0.179922863021)*x_ref[2]**o+(-0.386974503089)*x_ref[2]
        ref[0,0,2,0]=(0.932745137122)*x_ref[0]**o+(0.627995830114)*x_ref[0]+(-0.334344490736)*x_ref[1]**o+(0.733210590815)*x_ref[1]+(-0.23424959051)*x_ref[2]**o+(0.140644214958)*x_ref[2]
        ref[0,0,2,1]=(-0.282836519247)*x_ref[0]**o+(-0.744077669768)*x_ref[0]+(0.123898574243)*x_ref[1]**o+(0.562384289004)*x_ref[1]+(0.0757927394637)*x_ref[2]**o+(0.767684067921)*x_ref[2]
        ref[0,1,0,0]=(0.212366202901)*x_ref[0]**o+(-0.679228304269)*x_ref[0]+(0.829706652419)*x_ref[1]**o+(0.24389996073)*x_ref[1]+(0.909367025978)*x_ref[2]**o+(0.280617790407)*x_ref[2]
        ref[0,1,0,1]=(0.760070952632)*x_ref[0]**o+(-0.678309252372)*x_ref[0]+(0.0372281910363)*x_ref[1]**o+(0.0284582814297)*x_ref[1]+(-0.092349250941)*x_ref[2]**o+(0.347713645921)*x_ref[2]
        ref[0,1,1,0]=(0.248410230191)*x_ref[0]**o+(-0.833983202295)*x_ref[0]+(-0.759357278899)*x_ref[1]**o+(0.880038285893)*x_ref[1]+(-0.734799540652)*x_ref[2]**o+(0.493330231104)*x_ref[2]
        ref[0,1,1,1]=(0.0359919724228)*x_ref[0]**o+(-0.0393899014057)*x_ref[0]+(0.185155782488)*x_ref[1]**o+(-0.92692925719)*x_ref[1]+(0.74718037404)*x_ref[2]**o+(-0.283900681763)*x_ref[2]
        ref[0,1,2,0]=(-0.208643544667)*x_ref[0]**o+(0.368200149096)*x_ref[0]+(-0.327032725172)*x_ref[1]**o+(-0.245547905792)*x_ref[1]+(0.0886488214242)*x_ref[2]**o+(0.179139897691)*x_ref[2]
        ref[0,1,2,1]=(-0.0618762861459)*x_ref[0]**o+(0.253419217897)*x_ref[0]+(0.840270125508)*x_ref[1]**o+(0.0223556878387)*x_ref[1]+(0.0560904700176)*x_ref[2]**o+(-0.646231049424)*x_ref[2]
        ref[0,2,0,0]=(0.568966969865)*x_ref[0]**o+(0.786776511348)*x_ref[0]+(0.150595926116)*x_ref[1]**o+(-0.780164417762)*x_ref[1]+(0.510149831242)*x_ref[2]**o+(-0.902871929903)*x_ref[2]
        ref[0,2,0,1]=(-0.817062669957)*x_ref[0]**o+(0.626805486868)*x_ref[0]+(-0.566129786829)*x_ref[1]**o+(0.248418548905)*x_ref[1]+(-0.273995326019)*x_ref[2]**o+(0.402023521018)*x_ref[2]
        ref[0,2,1,0]=(-0.160276235113)*x_ref[0]**o+(-0.0732543937053)*x_ref[0]+(-0.0651595982698)*x_ref[1]**o+(-0.529463476237)*x_ref[1]+(0.346098812945)*x_ref[2]**o+(0.418367016439)*x_ref[2]
        ref[0,2,1,1]=(0.736907408498)*x_ref[0]**o+(-0.123612748157)*x_ref[0]+(0.000834659108651)*x_ref[1]**o+(0.568833761363)*x_ref[1]+(-0.542676953582)*x_ref[2]**o+(0.026182553484)*x_ref[2]
        ref[0,2,2,0]=(0.320985918151)*x_ref[0]**o+(0.0512654311644)*x_ref[0]+(-0.0313412913713)*x_ref[1]**o+(0.295659930913)*x_ref[1]+(-0.989560405329)*x_ref[2]**o+(-0.107344246298)*x_ref[2]
        ref[0,2,2,1]=(-0.957793939949)*x_ref[0]**o+(-0.177591880789)*x_ref[0]+(0.0785829772689)*x_ref[1]**o+(0.121445956612)*x_ref[1]+(-0.384134176508)*x_ref[2]**o+(-0.775433914622)*x_ref[2]
        ref[0,3,0,0]=(-0.130620609021)*x_ref[0]**o+(0.268283814564)*x_ref[0]+(-0.192835222037)*x_ref[1]**o+(0.235341688541)*x_ref[1]+(0.957069315916)*x_ref[2]**o+(-0.708853567529)*x_ref[2]
        ref[0,3,0,1]=(-0.377028600614)*x_ref[0]**o+(-0.858446162753)*x_ref[0]+(-0.850630580491)*x_ref[1]**o+(-0.810676898206)*x_ref[1]+(-0.613330062817)*x_ref[2]**o+(0.714335738947)*x_ref[2]
        ref[0,3,1,0]=(-0.602336649415)*x_ref[0]**o+(-0.166667559558)*x_ref[0]+(0.962948866081)*x_ref[1]**o+(0.939531483467)*x_ref[1]+(0.267767294244)*x_ref[2]**o+(0.247155872428)*x_ref[2]
        ref[0,3,1,1]=(0.838505391049)*x_ref[0]**o+(0.115204438919)*x_ref[0]+(0.989010581013)*x_ref[1]**o+(-0.649789818418)*x_ref[1]+(0.757328444925)*x_ref[2]**o+(0.825158129449)*x_ref[2]
        ref[0,3,2,0]=(0.724488034506)*x_ref[0]**o+(0.0333093325925)*x_ref[0]+(0.864643652652)*x_ref[1]**o+(-0.557517585925)*x_ref[1]+(-0.340534672063)*x_ref[2]**o+(-0.00425705176271)*x_ref[2]
        ref[0,3,2,1]=(-0.366459483792)*x_ref[0]**o+(0.257706980646)*x_ref[0]+(0.821731531023)*x_ref[1]**o+(-0.212501303617)*x_ref[1]+(-0.128977067795)*x_ref[2]**o+(-0.931439847041)*x_ref[2]
        ref[0,4,0,0]=(0.398430239547)*x_ref[0]**o+(-0.00991917699933)*x_ref[0]+(-0.182471763172)*x_ref[1]**o+(0.218327764946)*x_ref[1]+(-0.484510127696)*x_ref[2]**o+(-0.360678193996)*x_ref[2]
        ref[0,4,0,1]=(0.0640680415135)*x_ref[0]**o+(0.671150063087)*x_ref[0]+(0.208610130381)*x_ref[1]**o+(-0.441252393083)*x_ref[1]+(0.365150146582)*x_ref[2]**o+(-0.228346558142)*x_ref[2]
        ref[0,4,1,0]=(-0.852694824303)*x_ref[0]**o+(0.181041494463)*x_ref[0]+(0.898413617086)*x_ref[1]**o+(-0.918988862595)*x_ref[1]+(0.169010143324)*x_ref[2]**o+(0.540458809834)*x_ref[2]
        ref[0,4,1,1]=(-0.213782550366)*x_ref[0]**o+(-0.180557891041)*x_ref[0]+(0.23144059382)*x_ref[1]**o+(0.0609482143429)*x_ref[1]+(0.89256221946)*x_ref[2]**o+(0.977857158982)*x_ref[2]
        ref[0,4,2,0]=(0.320772898841)*x_ref[0]**o+(0.371610189518)*x_ref[0]+(0.692264150983)*x_ref[1]**o+(-0.751154935415)*x_ref[1]+(-0.163272902772)*x_ref[2]**o+(0.0635386628764)*x_ref[2]
        ref[0,4,2,1]=(-0.316885908577)*x_ref[0]**o+(0.476104431241)*x_ref[0]+(-0.647713072489)*x_ref[1]**o+(0.683813449908)*x_ref[1]+(0.699293342945)*x_ref[2]**o+(-0.0373906770097)*x_ref[2]
        ref[1,0,0,0]=(-0.030329265733)*x_ref[0]**o+(-0.0115761157375)*x_ref[0]+(-0.30993823088)*x_ref[1]**o+(0.628877975389)*x_ref[1]+(-0.582940961417)*x_ref[2]**o+(0.154786570457)*x_ref[2]
        ref[1,0,0,1]=(-0.593610753495)*x_ref[0]**o+(-0.822232097764)*x_ref[0]+(-0.465144000698)*x_ref[1]**o+(-0.62576868204)*x_ref[1]+(-0.53588531407)*x_ref[2]**o+(0.119119578421)*x_ref[2]
        ref[1,0,1,0]=(-0.15668734059)*x_ref[0]**o+(-0.950658811047)*x_ref[0]+(-0.267969935748)*x_ref[1]**o+(0.0919415340304)*x_ref[1]+(-0.0445382452662)*x_ref[2]**o+(0.949984936459)*x_ref[2]
        ref[1,0,1,1]=(-0.762474809436)*x_ref[0]**o+(0.794314442281)*x_ref[0]+(0.617742930372)*x_ref[1]**o+(-0.502057530669)*x_ref[1]+(0.42532154189)*x_ref[2]**o+(0.343417687179)*x_ref[2]
        ref[1,0,2,0]=(0.259583418084)*x_ref[0]**o+(0.460122326434)*x_ref[0]+(0.875185242587)*x_ref[1]**o+(-0.621232091699)*x_ref[1]+(0.922536397008)*x_ref[2]**o+(-0.494729334323)*x_ref[2]
        ref[1,0,2,1]=(0.79625863156)*x_ref[0]**o+(-0.980360881506)*x_ref[0]+(0.947721778185)*x_ref[1]**o+(-0.234201025597)*x_ref[1]+(-0.974251961452)*x_ref[2]**o+(-0.300484605111)*x_ref[2]
        ref[1,1,0,0]=(0.719927358834)*x_ref[0]**o+(-0.296944527139)*x_ref[0]+(0.970290716632)*x_ref[1]**o+(0.45656092862)*x_ref[1]+(0.657023751979)*x_ref[2]**o+(-0.703898365316)*x_ref[2]
        ref[1,1,0,1]=(-0.0600633107738)*x_ref[0]**o+(0.525293730667)*x_ref[0]+(0.17955637601)*x_ref[1]**o+(-0.625782540859)*x_ref[1]+(0.205255859488)*x_ref[2]**o+(-0.963128832741)*x_ref[2]
        ref[1,1,1,0]=(0.119721921725)*x_ref[0]**o+(0.957918016103)*x_ref[0]+(-0.574883179285)*x_ref[1]**o+(0.915580622743)*x_ref[1]+(-0.432538225757)*x_ref[2]**o+(-0.937539574986)*x_ref[2]
        ref[1,1,1,1]=(-0.295639299942)*x_ref[0]**o+(-0.703954000058)*x_ref[0]+(-0.185001166055)*x_ref[1]**o+(-0.943933820897)*x_ref[1]+(-0.12692007135)*x_ref[2]**o+(0.476481616585)*x_ref[2]
        ref[1,1,2,0]=(0.247892010246)*x_ref[0]**o+(0.901198196531)*x_ref[0]+(-0.586162224974)*x_ref[1]**o+(-0.934177552379)*x_ref[1]+(0.704265144051)*x_ref[2]**o+(-0.688632953989)*x_ref[2]
        ref[1,1,2,1]=(-0.116039842334)*x_ref[0]**o+(-0.300373025101)*x_ref[0]+(-0.286684372551)*x_ref[1]**o+(-0.452337760059)*x_ref[1]+(0.883210026596)*x_ref[2]**o+(-0.319739612634)*x_ref[2]
        ref[1,2,0,0]=(-0.0326555626249)*x_ref[0]**o+(-0.0933559852573)*x_ref[0]+(0.807443822382)*x_ref[1]**o+(-0.148842350195)*x_ref[1]+(-0.118904471983)*x_ref[2]**o+(-0.740440145747)*x_ref[2]
        ref[1,2,0,1]=(0.313623768488)*x_ref[0]**o+(0.524958399998)*x_ref[0]+(0.876747001712)*x_ref[1]**o+(-0.468458894694)*x_ref[1]+(0.52118935613)*x_ref[2]**o+(-0.776405429712)*x_ref[2]
        ref[1,2,1,0]=(0.1328367779)*x_ref[0]**o+(-0.959628068869)*x_ref[0]+(0.746000325891)*x_ref[1]**o+(-0.454705589649)*x_ref[1]+(0.08534697763)*x_ref[2]**o+(0.174892567303)*x_ref[2]
        ref[1,2,1,1]=(-0.482594997843)*x_ref[0]**o+(0.765628170342)*x_ref[0]+(0.103333249513)*x_ref[1]**o+(0.840778499017)*x_ref[1]+(0.136748193255)*x_ref[2]**o+(0.771613240623)*x_ref[2]
        ref[1,2,2,0]=(-0.305745376076)*x_ref[0]**o+(-0.0464234361529)*x_ref[0]+(-0.163528204043)*x_ref[1]**o+(0.80782719974)*x_ref[1]+(0.131046315567)*x_ref[2]**o+(0.0849217124905)*x_ref[2]
        ref[1,2,2,1]=(0.564915635291)*x_ref[0]**o+(-0.162467524139)*x_ref[0]+(-0.696721556806)*x_ref[1]**o+(0.0172839552576)*x_ref[1]+(-0.630040402586)*x_ref[2]**o+(-0.181232689117)*x_ref[2]
        ref[1,3,0,0]=(-0.768631958549)*x_ref[0]**o+(0.537076196504)*x_ref[0]+(-0.403489871013)*x_ref[1]**o+(0.831966318632)*x_ref[1]+(0.602387061638)*x_ref[2]**o+(0.330495470463)*x_ref[2]
        ref[1,3,0,1]=(0.770878177942)*x_ref[0]**o+(-0.678609489017)*x_ref[0]+(-0.895200519099)*x_ref[1]**o+(0.935093561205)*x_ref[1]+(0.0497551733469)*x_ref[2]**o+(0.820432808383)*x_ref[2]
        ref[1,3,1,0]=(0.052867263447)*x_ref[0]**o+(0.566167926663)*x_ref[0]+(-0.00397870350455)*x_ref[1]**o+(-0.780374460006)*x_ref[1]+(-0.166645199933)*x_ref[2]**o+(-0.867274057327)*x_ref[2]
        ref[1,3,1,1]=(0.0573446476183)*x_ref[0]**o+(-0.699343104682)*x_ref[0]+(0.428624411932)*x_ref[1]**o+(-0.400834963806)*x_ref[1]+(-0.722750021776)*x_ref[2]**o+(0.386821944027)*x_ref[2]
        ref[1,3,2,0]=(0.390777333115)*x_ref[0]**o+(0.191564882434)*x_ref[0]+(-0.774547811812)*x_ref[1]**o+(0.747678394063)*x_ref[1]+(0.662273739757)*x_ref[2]**o+(-0.295629568405)*x_ref[2]
        ref[1,3,2,1]=(-0.160077661734)*x_ref[0]**o+(-0.512286879466)*x_ref[0]+(0.616824388099)*x_ref[1]**o+(0.0257995114324)*x_ref[1]+(0.644270949659)*x_ref[2]**o+(0.243520087896)*x_ref[2]
        ref[1,4,0,0]=(-0.980501571755)*x_ref[0]**o+(0.844184850883)*x_ref[0]+(-0.608509998911)*x_ref[1]**o+(-0.347687895783)*x_ref[1]+(-0.536534228926)*x_ref[2]**o+(0.632821444072)*x_ref[2]
        ref[1,4,0,1]=(-0.881676821515)*x_ref[0]**o+(-0.858728104026)*x_ref[0]+(-0.576516592622)*x_ref[1]**o+(0.695748958418)*x_ref[1]+(0.856547670877)*x_ref[2]**o+(-0.932960091743)*x_ref[2]
        ref[1,4,1,0]=(-0.451435812217)*x_ref[0]**o+(0.963886330308)*x_ref[0]+(0.888532771273)*x_ref[1]**o+(0.807924978471)*x_ref[1]+(0.789117121354)*x_ref[2]**o+(0.577647188972)*x_ref[2]
        ref[1,4,1,1]=(-0.315320308371)*x_ref[0]**o+(0.182798184972)*x_ref[0]+(0.936963228212)*x_ref[1]**o+(-0.634429372225)*x_ref[1]+(0.634343556662)*x_ref[2]**o+(0.975851394002)*x_ref[2]
        ref[1,4,2,0]=(0.858246326961)*x_ref[0]**o+(0.0955812000864)*x_ref[0]+(0.16143064785)*x_ref[1]**o+(0.397013086552)*x_ref[1]+(-0.507188929963)*x_ref[2]**o+(0.49789827039)*x_ref[2]
        ref[1,4,2,1]=(-0.524634529078)*x_ref[0]**o+(0.675491034226)*x_ref[0]+(0.858857975014)*x_ref[1]**o+(-0.57874207047)*x_ref[1]+(0.476551353232)*x_ref[2]**o+(-0.622601386038)*x_ref[2]
        ref[2,0,0,0]=(0.976563609)*x_ref[0]**o+(0.127133211238)*x_ref[0]+(-0.410155025204)*x_ref[1]**o+(-0.909032700787)*x_ref[1]+(-0.197014506331)*x_ref[2]**o+(0.557171133933)*x_ref[2]
        ref[2,0,0,1]=(0.434874625301)*x_ref[0]**o+(-0.576955025374)*x_ref[0]+(-0.466686136319)*x_ref[1]**o+(-0.359616702831)*x_ref[1]+(-0.916050768111)*x_ref[2]**o+(0.900510057507)*x_ref[2]
        ref[2,0,1,0]=(0.615259852986)*x_ref[0]**o+(-0.685399646912)*x_ref[0]+(-0.222742007887)*x_ref[1]**o+(-0.487739486778)*x_ref[1]+(-0.675225266903)*x_ref[2]**o+(0.494623157154)*x_ref[2]
        ref[2,0,1,1]=(-0.89129010918)*x_ref[0]**o+(0.0051571255592)*x_ref[0]+(0.681003954208)*x_ref[1]**o+(0.634041720719)*x_ref[1]+(-0.308587458722)*x_ref[2]**o+(-0.107801539193)*x_ref[2]
        ref[2,0,2,0]=(-0.660496698398)*x_ref[0]**o+(-0.889315862375)*x_ref[0]+(0.64248433679)*x_ref[1]**o+(-0.941990511222)*x_ref[1]+(-0.173668816201)*x_ref[2]**o+(0.237812802184)*x_ref[2]
        ref[2,0,2,1]=(0.847365004226)*x_ref[0]**o+(0.802449170085)*x_ref[0]+(0.504737931993)*x_ref[1]**o+(-0.695546680342)*x_ref[1]+(0.0742668699455)*x_ref[2]**o+(0.396565506083)*x_ref[2]
        ref[2,1,0,0]=(0.812115742561)*x_ref[0]**o+(-0.528091349418)*x_ref[0]+(-0.442834134691)*x_ref[1]**o+(-0.574414379092)*x_ref[1]+(-0.734077338221)*x_ref[2]**o+(0.159306515061)*x_ref[2]
        ref[2,1,0,1]=(-0.259422255041)*x_ref[0]**o+(-0.104016445541)*x_ref[0]+(0.3560854635)*x_ref[1]**o+(0.997015276785)*x_ref[1]+(-0.707278312293)*x_ref[2]**o+(-0.975235879836)*x_ref[2]
        ref[2,1,1,0]=(0.402099188227)*x_ref[0]**o+(0.0320354114274)*x_ref[0]+(0.203384927114)*x_ref[1]**o+(-0.847804170327)*x_ref[1]+(-0.898459903546)*x_ref[2]**o+(0.639388498254)*x_ref[2]
        ref[2,1,1,1]=(-0.0477209380768)*x_ref[0]**o+(0.531574667575)*x_ref[0]+(0.280889682833)*x_ref[1]**o+(-0.444963671116)*x_ref[1]+(0.33935324652)*x_ref[2]**o+(0.943089036339)*x_ref[2]
        ref[2,1,2,0]=(0.396038211375)*x_ref[0]**o+(-0.54994373568)*x_ref[0]+(0.851017234687)*x_ref[1]**o+(-0.739787661885)*x_ref[1]+(-0.368325897478)*x_ref[2]**o+(-0.449096507837)*x_ref[2]
        ref[2,1,2,1]=(0.857714720302)*x_ref[0]**o+(-0.750668366832)*x_ref[0]+(-0.587901076381)*x_ref[1]**o+(-0.24906126763)*x_ref[1]+(0.131707530536)*x_ref[2]**o+(0.900975507373)*x_ref[2]
        ref[2,2,0,0]=(0.783919932392)*x_ref[0]**o+(-0.967669966647)*x_ref[0]+(0.82052580972)*x_ref[1]**o+(-0.711919131449)*x_ref[1]+(-0.396059626351)*x_ref[2]**o+(-0.265805391355)*x_ref[2]
        ref[2,2,0,1]=(-0.0398160197307)*x_ref[0]**o+(-0.238193541)*x_ref[0]+(0.282277679599)*x_ref[1]**o+(-0.0526835304963)*x_ref[1]+(0.590871489195)*x_ref[2]**o+(-0.925835585742)*x_ref[2]
        ref[2,2,1,0]=(-0.825757767738)*x_ref[0]**o+(0.238488063696)*x_ref[0]+(0.109616380527)*x_ref[1]**o+(-0.942941715764)*x_ref[1]+(0.632013079754)*x_ref[2]**o+(0.0517997920543)*x_ref[2]
        ref[2,2,1,1]=(-0.85343230326)*x_ref[0]**o+(0.225247381624)*x_ref[0]+(0.43171428489)*x_ref[1]**o+(0.0937487532231)*x_ref[1]+(0.894064129869)*x_ref[2]**o+(-0.0876265498234)*x_ref[2]
        ref[2,2,2,0]=(0.349294373111)*x_ref[0]**o+(-0.734159680014)*x_ref[0]+(0.0489227393033)*x_ref[1]**o+(-0.245284468757)*x_ref[1]+(-0.346469825649)*x_ref[2]**o+(0.547054093324)*x_ref[2]
        ref[2,2,2,1]=(-0.0478095998047)*x_ref[0]**o+(0.440261344405)*x_ref[0]+(-0.590330499679)*x_ref[1]**o+(-0.961009986328)*x_ref[1]+(0.546190002905)*x_ref[2]**o+(-0.027965996519)*x_ref[2]
        ref[2,3,0,0]=(0.148563550644)*x_ref[0]**o+(0.576704219545)*x_ref[0]+(-0.782168581109)*x_ref[1]**o+(-0.8301245678)*x_ref[1]+(0.937936245682)*x_ref[2]**o+(0.846740948371)*x_ref[2]
        ref[2,3,0,1]=(-0.447832058524)*x_ref[0]**o+(0.385431388609)*x_ref[0]+(0.193985676964)*x_ref[1]**o+(-0.91717180275)*x_ref[1]+(-0.408732230908)*x_ref[2]**o+(0.548173152118)*x_ref[2]
        ref[2,3,1,0]=(0.197592567864)*x_ref[0]**o+(-0.77657127384)*x_ref[0]+(0.266355135001)*x_ref[1]**o+(0.604009981584)*x_ref[1]+(-0.379930450459)*x_ref[2]**o+(0.46952844479)*x_ref[2]
        ref[2,3,1,1]=(0.996803794962)*x_ref[0]**o+(0.319115676981)*x_ref[0]+(0.255552997875)*x_ref[1]**o+(0.21834578367)*x_ref[1]+(0.389966036243)*x_ref[2]**o+(-0.115308080726)*x_ref[2]
        ref[2,3,2,0]=(-0.0244303880675)*x_ref[0]**o+(-0.128929049017)*x_ref[0]+(-0.350933414406)*x_ref[1]**o+(-0.341716440852)*x_ref[1]+(-0.724579076377)*x_ref[2]**o+(0.21963465519)*x_ref[2]
        ref[2,3,2,1]=(-0.92899971478)*x_ref[0]**o+(-0.304496074717)*x_ref[0]+(0.568151427024)*x_ref[1]**o+(0.504629517922)*x_ref[1]+(0.846158711477)*x_ref[2]**o+(0.35936792568)*x_ref[2]
        ref[2,4,0,0]=(-0.350964834352)*x_ref[0]**o+(0.387667833884)*x_ref[0]+(0.547101274223)*x_ref[1]**o+(0.257218757373)*x_ref[1]+(-0.00316748527841)*x_ref[2]**o+(0.314561238615)*x_ref[2]
        ref[2,4,0,1]=(0.408941739151)*x_ref[0]**o+(0.819113792537)*x_ref[0]+(0.532891848781)*x_ref[1]**o+(-0.30688659725)*x_ref[1]+(-0.736611937618)*x_ref[2]**o+(-0.875783706051)*x_ref[2]
        ref[2,4,1,0]=(-0.730885165985)*x_ref[0]**o+(0.453908506302)*x_ref[0]+(-0.617385752505)*x_ref[1]**o+(0.357459789592)*x_ref[1]+(0.286323175805)*x_ref[2]**o+(-0.448879127031)*x_ref[2]
        ref[2,4,1,1]=(-0.295377341281)*x_ref[0]**o+(0.423424371051)*x_ref[0]+(0.524695076927)*x_ref[1]**o+(-0.698499625038)*x_ref[1]+(0.617308008395)*x_ref[2]**o+(-0.423469866581)*x_ref[2]
        ref[2,4,2,0]=(0.104251379635)*x_ref[0]**o+(0.989709098795)*x_ref[0]+(-0.422075396624)*x_ref[1]**o+(-0.254328561231)*x_ref[1]+(-0.502052795074)*x_ref[2]**o+(-0.0851411652168)*x_ref[2]
        ref[2,4,2,1]=(0.379332842082)*x_ref[0]**o+(-0.707504861783)*x_ref[0]+(0.00855075647146)*x_ref[1]**o+(0.792851159033)*x_ref[1]+(-0.692042160791)*x_ref[2]**o+(-0.18327165243)*x_ref[2]
        ref[3,0,0,0]=(0.310700158718)*x_ref[0]**o+(0.285647240187)*x_ref[0]+(0.817947571838)*x_ref[1]**o+(0.808698040916)*x_ref[1]+(0.844184006443)*x_ref[2]**o+(-0.310383215636)*x_ref[2]
        ref[3,0,0,1]=(-0.0173109952616)*x_ref[0]**o+(0.672034452037)*x_ref[0]+(-0.464572256298)*x_ref[1]**o+(0.205799523299)*x_ref[1]+(-0.971979841338)*x_ref[2]**o+(0.261515966001)*x_ref[2]
        ref[3,0,1,0]=(0.0329384377028)*x_ref[0]**o+(0.571501185072)*x_ref[0]+(-0.902222556642)*x_ref[1]**o+(-0.528317784373)*x_ref[1]+(0.331989640881)*x_ref[2]**o+(0.0427505624318)*x_ref[2]
        ref[3,0,1,1]=(0.964350410761)*x_ref[0]**o+(-0.327198989999)*x_ref[0]+(-0.195271523191)*x_ref[1]**o+(0.534663014601)*x_ref[1]+(-0.0208261208706)*x_ref[2]**o+(-0.158019249052)*x_ref[2]
        ref[3,0,2,0]=(-0.622082726008)*x_ref[0]**o+(0.559463883442)*x_ref[0]+(0.841439693313)*x_ref[1]**o+(-0.196363987516)*x_ref[1]+(-0.0143926547396)*x_ref[2]**o+(0.144732603689)*x_ref[2]
        ref[3,0,2,1]=(-0.149531653278)*x_ref[0]**o+(0.354229138322)*x_ref[0]+(-0.507916088829)*x_ref[1]**o+(0.221251580047)*x_ref[1]+(0.197213706274)*x_ref[2]**o+(-0.555645149276)*x_ref[2]
        ref[3,1,0,0]=(0.331020785374)*x_ref[0]**o+(-0.377647773513)*x_ref[0]+(0.563355928949)*x_ref[1]**o+(0.789309216711)*x_ref[1]+(0.203663383875)*x_ref[2]**o+(-0.0333894540645)*x_ref[2]
        ref[3,1,0,1]=(-0.0965613913882)*x_ref[0]**o+(0.18656484323)*x_ref[0]+(-0.138245941339)*x_ref[1]**o+(-0.34549538745)*x_ref[1]+(0.827173403167)*x_ref[2]**o+(-0.873380503939)*x_ref[2]
        ref[3,1,1,0]=(0.932940890033)*x_ref[0]**o+(-0.564883830288)*x_ref[0]+(-0.505052725295)*x_ref[1]**o+(-0.669420210206)*x_ref[1]+(-0.799638096039)*x_ref[2]**o+(0.469400233749)*x_ref[2]
        ref[3,1,1,1]=(-0.9741761069)*x_ref[0]**o+(-0.748125328781)*x_ref[0]+(0.664448241931)*x_ref[1]**o+(0.0957643296979)*x_ref[1]+(0.112211322019)*x_ref[2]**o+(-0.131870290434)*x_ref[2]
        ref[3,1,2,0]=(-0.438609127704)*x_ref[0]**o+(-0.36618286515)*x_ref[0]+(-0.376420011115)*x_ref[1]**o+(-0.67084624599)*x_ref[1]+(0.821679796644)*x_ref[2]**o+(0.80111319897)*x_ref[2]
        ref[3,1,2,1]=(-0.409859432085)*x_ref[0]**o+(0.848767816929)*x_ref[0]+(0.382358459835)*x_ref[1]**o+(-0.750139843631)*x_ref[1]+(0.587152586559)*x_ref[2]**o+(-0.26162952001)*x_ref[2]
        ref[3,2,0,0]=(-0.506329785074)*x_ref[0]**o+(-0.0967745752004)*x_ref[0]+(-0.163316795268)*x_ref[1]**o+(0.0628481921732)*x_ref[1]+(0.395355494367)*x_ref[2]**o+(0.0978901916332)*x_ref[2]
        ref[3,2,0,1]=(0.803590309587)*x_ref[0]**o+(0.470388053389)*x_ref[0]+(0.674942922168)*x_ref[1]**o+(0.223472834762)*x_ref[1]+(0.552332209028)*x_ref[2]**o+(-0.396443498843)*x_ref[2]
        ref[3,2,1,0]=(-0.312384723839)*x_ref[0]**o+(0.392213828608)*x_ref[0]+(0.449465316671)*x_ref[1]**o+(-0.249103089401)*x_ref[1]+(0.601412913458)*x_ref[2]**o+(0.209899368039)*x_ref[2]
        ref[3,2,1,1]=(-0.583776179557)*x_ref[0]**o+(0.485007047754)*x_ref[0]+(-0.0354153633589)*x_ref[1]**o+(-0.898630969916)*x_ref[1]+(-0.266581492795)*x_ref[2]**o+(0.822086680281)*x_ref[2]
        ref[3,2,2,0]=(0.944617094397)*x_ref[0]**o+(-0.173057994833)*x_ref[0]+(0.927832160048)*x_ref[1]**o+(-0.218599594252)*x_ref[1]+(0.349957755011)*x_ref[2]**o+(0.392602744545)*x_ref[2]
        ref[3,2,2,1]=(0.461648084674)*x_ref[0]**o+(0.826213376949)*x_ref[0]+(0.143429054731)*x_ref[1]**o+(-0.214265128007)*x_ref[1]+(0.0317208127607)*x_ref[2]**o+(-0.858361078381)*x_ref[2]
        ref[3,3,0,0]=(0.306607920368)*x_ref[0]**o+(0.147692891688)*x_ref[0]+(-0.562661766456)*x_ref[1]**o+(0.448595564727)*x_ref[1]+(-0.707178218373)*x_ref[2]**o+(0.0655554587866)*x_ref[2]
        ref[3,3,0,1]=(0.135475820362)*x_ref[0]**o+(0.51420718793)*x_ref[0]+(0.343291220736)*x_ref[1]**o+(-0.797532312602)*x_ref[1]+(-0.273431011273)*x_ref[2]**o+(0.135307779253)*x_ref[2]
        ref[3,3,1,0]=(0.488028231073)*x_ref[0]**o+(0.542021806829)*x_ref[0]+(0.91916268609)*x_ref[1]**o+(0.317368273365)*x_ref[1]+(0.314807039395)*x_ref[2]**o+(-0.841583308025)*x_ref[2]
        ref[3,3,1,1]=(-0.182783814917)*x_ref[0]**o+(0.250863481201)*x_ref[0]+(0.755551457003)*x_ref[1]**o+(0.946171548529)*x_ref[1]+(0.0350225436682)*x_ref[2]**o+(0.776059830052)*x_ref[2]
        ref[3,3,2,0]=(-0.116707124994)*x_ref[0]**o+(-0.727010280588)*x_ref[0]+(0.172065389268)*x_ref[1]**o+(0.754164564639)*x_ref[1]+(0.803777635621)*x_ref[2]**o+(0.293490264417)*x_ref[2]
        ref[3,3,2,1]=(0.154235982797)*x_ref[0]**o+(-0.124286021486)*x_ref[0]+(-0.737324748116)*x_ref[1]**o+(-0.136803541581)*x_ref[1]+(-0.608370191604)*x_ref[2]**o+(-0.933599034128)*x_ref[2]
        ref[3,4,0,0]=(0.818061677328)*x_ref[0]**o+(-0.771719613166)*x_ref[0]+(0.742772795251)*x_ref[1]**o+(0.0359285091312)*x_ref[1]+(-0.985656851837)*x_ref[2]**o+(-0.884598206512)*x_ref[2]
        ref[3,4,0,1]=(0.198371732741)*x_ref[0]**o+(0.646496618681)*x_ref[0]+(0.232912446494)*x_ref[1]**o+(0.945619492023)*x_ref[1]+(-0.344370341731)*x_ref[2]**o+(0.31156840357)*x_ref[2]
        ref[3,4,1,0]=(-0.609298834606)*x_ref[0]**o+(0.784697392412)*x_ref[0]+(0.996065619079)*x_ref[1]**o+(-0.284498300043)*x_ref[1]+(-0.652795706803)*x_ref[2]**o+(-0.254746840396)*x_ref[2]
        ref[3,4,1,1]=(-0.0083632822507)*x_ref[0]**o+(0.200923618933)*x_ref[0]+(0.353883930064)*x_ref[1]**o+(0.153257380179)*x_ref[1]+(0.843965464205)*x_ref[2]**o+(0.0418501400703)*x_ref[2]
        ref[3,4,2,0]=(-0.226055622191)*x_ref[0]**o+(-0.701789166696)*x_ref[0]+(0.274080957143)*x_ref[1]**o+(0.0596744660556)*x_ref[1]+(-0.342083152987)*x_ref[2]**o+(0.593518464697)*x_ref[2]
        ref[3,4,2,1]=(-0.288168945023)*x_ref[0]**o+(-0.742841678774)*x_ref[0]+(0.972629532424)*x_ref[1]**o+(0.892100560733)*x_ref[1]+(0.803880474721)*x_ref[2]**o+(-0.0638336211224)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.492060381705)*x[0]**o+(0.578791868713)*x[0]+(-0.439849757114)*x[1]**o+(-0.52629328967)*x[1]
        ref=(0.492060381705)*x_ref[0]**o+(0.578791868713)*x_ref[0]+(-0.439849757114)*x_ref[1]**o+(-0.52629328967)*x_ref[1]
      else:
        arg=(0.954737963273)*x[0]**o+(0.452578946878)*x[0]+(0.0471085060698)*x[1]**o+(-0.640169119832)*x[1]+(0.710401490627)*x[2]**o+(-0.382589884739)*x[2]
        ref=(0.954737963273)*x_ref[0]**o+(0.452578946878)*x_ref[0]+(0.0471085060698)*x_ref[1]**o+(-0.640169119832)*x_ref[1]+(0.710401490627)*x_ref[2]**o+(-0.382589884739)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.59604585527)*x[0]**o+(0.826981169401)*x[0]+(0.227617349484)*x[1]**o+(0.184907819029)*x[1]
        arg[1]=(-0.424684036731)*x[0]**o+(0.346537997905)*x[0]+(0.891921272597)*x[1]**o+(-0.145924182872)*x[1]
        ref[0]=(-0.59604585527)*x_ref[0]**o+(0.826981169401)*x_ref[0]+(0.227617349484)*x_ref[1]**o+(0.184907819029)*x_ref[1]
        ref[1]=(-0.424684036731)*x_ref[0]**o+(0.346537997905)*x_ref[0]+(0.891921272597)*x_ref[1]**o+(-0.145924182872)*x_ref[1]
      else:
        arg[0]=(0.777228173978)*x[0]**o+(-0.812215827155)*x[0]+(-0.618922680968)*x[1]**o+(0.0660312103206)*x[1]+(-0.917545504466)*x[2]**o+(0.988524619414)*x[2]
        arg[1]=(0.110219469565)*x[0]**o+(-0.441866866062)*x[0]+(-0.654447764335)*x[1]**o+(0.230716133289)*x[1]+(0.347487317165)*x[2]**o+(-0.948877903043)*x[2]
        ref[0]=(0.777228173978)*x_ref[0]**o+(-0.812215827155)*x_ref[0]+(-0.618922680968)*x_ref[1]**o+(0.0660312103206)*x_ref[1]+(-0.917545504466)*x_ref[2]**o+(0.988524619414)*x_ref[2]
        ref[1]=(0.110219469565)*x_ref[0]**o+(-0.441866866062)*x_ref[0]+(-0.654447764335)*x_ref[1]**o+(0.230716133289)*x_ref[1]+(0.347487317165)*x_ref[2]**o+(-0.948877903043)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.958809754913)*x[0]**o+(-0.844717049041)*x[0]+(0.462564514427)*x[1]**o+(-0.247678335594)*x[1]
        arg[0,1]=(0.700700054189)*x[0]**o+(-0.653724611457)*x[0]+(0.52659177803)*x[1]**o+(-0.387559446418)*x[1]
        arg[0,2]=(-0.326124300134)*x[0]**o+(0.115197897152)*x[0]+(-0.876340641846)*x[1]**o+(0.433060472581)*x[1]
        arg[0,3]=(-0.101215736407)*x[0]**o+(-0.483684510661)*x[0]+(0.0418770907036)*x[1]**o+(-0.845564414421)*x[1]
        arg[0,4]=(-0.293094855718)*x[0]**o+(0.710954930773)*x[0]+(0.238671713869)*x[1]**o+(0.576638343412)*x[1]
        arg[1,0]=(0.529861280122)*x[0]**o+(-0.259437904048)*x[0]+(-0.620725288158)*x[1]**o+(0.326787999382)*x[1]
        arg[1,1]=(0.421245543012)*x[0]**o+(-0.954139636698)*x[0]+(0.0498199734926)*x[1]**o+(0.24010238499)*x[1]
        arg[1,2]=(0.333362362373)*x[0]**o+(-0.999953175952)*x[0]+(0.866193849246)*x[1]**o+(0.764287410177)*x[1]
        arg[1,3]=(0.470415957539)*x[0]**o+(-0.133477966871)*x[0]+(-0.0416970809832)*x[1]**o+(-0.416352740444)*x[1]
        arg[1,4]=(-0.178613018492)*x[0]**o+(-0.967264985978)*x[0]+(-0.297102423687)*x[1]**o+(0.884184973413)*x[1]
        arg[2,0]=(-0.29363453294)*x[0]**o+(-0.0237360594862)*x[0]+(-0.300776450747)*x[1]**o+(-0.493932721521)*x[1]
        arg[2,1]=(0.795045745699)*x[0]**o+(0.984956851154)*x[0]+(-0.724208405235)*x[1]**o+(0.59029067239)*x[1]
        arg[2,2]=(-0.803012112531)*x[0]**o+(-0.831727785428)*x[0]+(0.330295112332)*x[1]**o+(0.0966317908147)*x[1]
        arg[2,3]=(-0.652332429144)*x[0]**o+(-0.38689457016)*x[0]+(-0.391780825295)*x[1]**o+(-0.444284790848)*x[1]
        arg[2,4]=(0.416775161398)*x[0]**o+(-0.16281006536)*x[0]+(-0.742938932867)*x[1]**o+(0.456486183714)*x[1]
        arg[3,0]=(0.506731632086)*x[0]**o+(-0.222468320035)*x[0]+(-0.906398223556)*x[1]**o+(-0.0250630448611)*x[1]
        arg[3,1]=(-0.39207398783)*x[0]**o+(-0.574739374252)*x[0]+(0.262101584569)*x[1]**o+(0.249587535087)*x[1]
        arg[3,2]=(-0.252593913503)*x[0]**o+(-0.864294052989)*x[0]+(0.0110434296336)*x[1]**o+(-0.224585040999)*x[1]
        arg[3,3]=(-0.631704469212)*x[0]**o+(-0.495153785048)*x[0]+(-0.304420746672)*x[1]**o+(-0.254005222021)*x[1]
        arg[3,4]=(0.775631055421)*x[0]**o+(0.370095409961)*x[0]+(0.894627612994)*x[1]**o+(-0.329716708629)*x[1]
        ref[0,0]=(-0.958809754913)*x_ref[0]**o+(-0.844717049041)*x_ref[0]+(0.462564514427)*x_ref[1]**o+(-0.247678335594)*x_ref[1]
        ref[0,1]=(0.700700054189)*x_ref[0]**o+(-0.653724611457)*x_ref[0]+(0.52659177803)*x_ref[1]**o+(-0.387559446418)*x_ref[1]
        ref[0,2]=(-0.326124300134)*x_ref[0]**o+(0.115197897152)*x_ref[0]+(-0.876340641846)*x_ref[1]**o+(0.433060472581)*x_ref[1]
        ref[0,3]=(-0.101215736407)*x_ref[0]**o+(-0.483684510661)*x_ref[0]+(0.0418770907036)*x_ref[1]**o+(-0.845564414421)*x_ref[1]
        ref[0,4]=(-0.293094855718)*x_ref[0]**o+(0.710954930773)*x_ref[0]+(0.238671713869)*x_ref[1]**o+(0.576638343412)*x_ref[1]
        ref[1,0]=(0.529861280122)*x_ref[0]**o+(-0.259437904048)*x_ref[0]+(-0.620725288158)*x_ref[1]**o+(0.326787999382)*x_ref[1]
        ref[1,1]=(0.421245543012)*x_ref[0]**o+(-0.954139636698)*x_ref[0]+(0.0498199734926)*x_ref[1]**o+(0.24010238499)*x_ref[1]
        ref[1,2]=(0.333362362373)*x_ref[0]**o+(-0.999953175952)*x_ref[0]+(0.866193849246)*x_ref[1]**o+(0.764287410177)*x_ref[1]
        ref[1,3]=(0.470415957539)*x_ref[0]**o+(-0.133477966871)*x_ref[0]+(-0.0416970809832)*x_ref[1]**o+(-0.416352740444)*x_ref[1]
        ref[1,4]=(-0.178613018492)*x_ref[0]**o+(-0.967264985978)*x_ref[0]+(-0.297102423687)*x_ref[1]**o+(0.884184973413)*x_ref[1]
        ref[2,0]=(-0.29363453294)*x_ref[0]**o+(-0.0237360594862)*x_ref[0]+(-0.300776450747)*x_ref[1]**o+(-0.493932721521)*x_ref[1]
        ref[2,1]=(0.795045745699)*x_ref[0]**o+(0.984956851154)*x_ref[0]+(-0.724208405235)*x_ref[1]**o+(0.59029067239)*x_ref[1]
        ref[2,2]=(-0.803012112531)*x_ref[0]**o+(-0.831727785428)*x_ref[0]+(0.330295112332)*x_ref[1]**o+(0.0966317908147)*x_ref[1]
        ref[2,3]=(-0.652332429144)*x_ref[0]**o+(-0.38689457016)*x_ref[0]+(-0.391780825295)*x_ref[1]**o+(-0.444284790848)*x_ref[1]
        ref[2,4]=(0.416775161398)*x_ref[0]**o+(-0.16281006536)*x_ref[0]+(-0.742938932867)*x_ref[1]**o+(0.456486183714)*x_ref[1]
        ref[3,0]=(0.506731632086)*x_ref[0]**o+(-0.222468320035)*x_ref[0]+(-0.906398223556)*x_ref[1]**o+(-0.0250630448611)*x_ref[1]
        ref[3,1]=(-0.39207398783)*x_ref[0]**o+(-0.574739374252)*x_ref[0]+(0.262101584569)*x_ref[1]**o+(0.249587535087)*x_ref[1]
        ref[3,2]=(-0.252593913503)*x_ref[0]**o+(-0.864294052989)*x_ref[0]+(0.0110434296336)*x_ref[1]**o+(-0.224585040999)*x_ref[1]
        ref[3,3]=(-0.631704469212)*x_ref[0]**o+(-0.495153785048)*x_ref[0]+(-0.304420746672)*x_ref[1]**o+(-0.254005222021)*x_ref[1]
        ref[3,4]=(0.775631055421)*x_ref[0]**o+(0.370095409961)*x_ref[0]+(0.894627612994)*x_ref[1]**o+(-0.329716708629)*x_ref[1]
      else:
        arg[0,0]=(0.392450639899)*x[0]**o+(-0.685488230002)*x[0]+(-0.41545928428)*x[1]**o+(0.244144392379)*x[1]+(0.602642156528)*x[2]**o+(-0.348717378461)*x[2]
        arg[0,1]=(-0.341243165541)*x[0]**o+(-0.616870164204)*x[0]+(-0.510613066362)*x[1]**o+(-0.81387010212)*x[1]+(0.856239691119)*x[2]**o+(0.0744651854239)*x[2]
        arg[0,2]=(-0.765983413604)*x[0]**o+(0.474551394174)*x[0]+(-0.939524929765)*x[1]**o+(0.418297898065)*x[1]+(0.00589909539531)*x[2]**o+(0.0889012994897)*x[2]
        arg[0,3]=(-0.521217563352)*x[0]**o+(-0.256316459647)*x[0]+(-0.0265965098379)*x[1]**o+(-0.597562278966)*x[1]+(0.0389151661339)*x[2]**o+(-0.572109451352)*x[2]
        arg[0,4]=(-0.853801268491)*x[0]**o+(-0.691274812057)*x[0]+(-0.682435618581)*x[1]**o+(-0.974334375359)*x[1]+(-0.642333316794)*x[2]**o+(0.353816751252)*x[2]
        arg[1,0]=(-0.657545394322)*x[0]**o+(-0.334172338096)*x[0]+(0.016748279459)*x[1]**o+(-0.294858621125)*x[1]+(-0.49403880046)*x[2]**o+(-0.152416774822)*x[2]
        arg[1,1]=(-0.602731976042)*x[0]**o+(0.685046816545)*x[0]+(0.750001644716)*x[1]**o+(-0.202076972289)*x[1]+(-0.0232888078434)*x[2]**o+(-0.0774477649878)*x[2]
        arg[1,2]=(-0.707321785527)*x[0]**o+(0.618526464205)*x[0]+(-0.578548442468)*x[1]**o+(-0.411374744442)*x[1]+(-0.944277708787)*x[2]**o+(0.647856975707)*x[2]
        arg[1,3]=(-0.222503271022)*x[0]**o+(-0.929428092728)*x[0]+(-0.976553212424)*x[1]**o+(0.0376439617084)*x[1]+(0.0428078568512)*x[2]**o+(-0.441850944272)*x[2]
        arg[1,4]=(-0.862164898455)*x[0]**o+(-0.989900221991)*x[0]+(0.110610844567)*x[1]**o+(0.419384612236)*x[1]+(-0.00271948841262)*x[2]**o+(0.301635101629)*x[2]
        arg[2,0]=(-0.0950421881942)*x[0]**o+(-0.741293053909)*x[0]+(-0.461538879874)*x[1]**o+(-0.8199842553)*x[1]+(0.0590344916829)*x[2]**o+(0.352541518434)*x[2]
        arg[2,1]=(-0.734407995647)*x[0]**o+(-0.726167721992)*x[0]+(0.700521229916)*x[1]**o+(0.753555863428)*x[1]+(-0.891638139696)*x[2]**o+(-0.996550711213)*x[2]
        arg[2,2]=(0.777110352629)*x[0]**o+(0.309719201157)*x[0]+(0.026816547522)*x[1]**o+(0.33161010287)*x[1]+(0.257929504352)*x[2]**o+(0.71871314654)*x[2]
        arg[2,3]=(-0.0982522143825)*x[0]**o+(0.0532144836415)*x[0]+(0.155895442381)*x[1]**o+(-0.872222127033)*x[1]+(0.810538165596)*x[2]**o+(-0.917071363009)*x[2]
        arg[2,4]=(0.882674783877)*x[0]**o+(-0.246439463648)*x[0]+(-0.478821457401)*x[1]**o+(-0.629183697296)*x[1]+(0.833283185393)*x[2]**o+(-0.895796651184)*x[2]
        arg[3,0]=(0.949019461395)*x[0]**o+(0.437511668107)*x[0]+(0.46679355983)*x[1]**o+(0.375391913775)*x[1]+(0.992336073925)*x[2]**o+(0.784847856608)*x[2]
        arg[3,1]=(0.10730061928)*x[0]**o+(0.624678931808)*x[0]+(-0.105084917663)*x[1]**o+(-0.0650034223387)*x[1]+(-0.271922353868)*x[2]**o+(-0.182967296038)*x[2]
        arg[3,2]=(0.115856234055)*x[0]**o+(0.711912354922)*x[0]+(-0.317962972122)*x[1]**o+(0.577856092951)*x[1]+(-0.246833922867)*x[2]**o+(0.354116462218)*x[2]
        arg[3,3]=(0.40937236699)*x[0]**o+(0.639252679403)*x[0]+(0.540206943219)*x[1]**o+(0.287006139758)*x[1]+(0.123311433598)*x[2]**o+(-0.0300672780213)*x[2]
        arg[3,4]=(-0.190299257713)*x[0]**o+(-0.022055675261)*x[0]+(-0.640751994447)*x[1]**o+(-0.411801947317)*x[1]+(0.0675008125057)*x[2]**o+(-0.659424255203)*x[2]
        ref[0,0]=(0.392450639899)*x_ref[0]**o+(-0.685488230002)*x_ref[0]+(-0.41545928428)*x_ref[1]**o+(0.244144392379)*x_ref[1]+(0.602642156528)*x_ref[2]**o+(-0.348717378461)*x_ref[2]
        ref[0,1]=(-0.341243165541)*x_ref[0]**o+(-0.616870164204)*x_ref[0]+(-0.510613066362)*x_ref[1]**o+(-0.81387010212)*x_ref[1]+(0.856239691119)*x_ref[2]**o+(0.0744651854239)*x_ref[2]
        ref[0,2]=(-0.765983413604)*x_ref[0]**o+(0.474551394174)*x_ref[0]+(-0.939524929765)*x_ref[1]**o+(0.418297898065)*x_ref[1]+(0.00589909539531)*x_ref[2]**o+(0.0889012994897)*x_ref[2]
        ref[0,3]=(-0.521217563352)*x_ref[0]**o+(-0.256316459647)*x_ref[0]+(-0.0265965098379)*x_ref[1]**o+(-0.597562278966)*x_ref[1]+(0.0389151661339)*x_ref[2]**o+(-0.572109451352)*x_ref[2]
        ref[0,4]=(-0.853801268491)*x_ref[0]**o+(-0.691274812057)*x_ref[0]+(-0.682435618581)*x_ref[1]**o+(-0.974334375359)*x_ref[1]+(-0.642333316794)*x_ref[2]**o+(0.353816751252)*x_ref[2]
        ref[1,0]=(-0.657545394322)*x_ref[0]**o+(-0.334172338096)*x_ref[0]+(0.016748279459)*x_ref[1]**o+(-0.294858621125)*x_ref[1]+(-0.49403880046)*x_ref[2]**o+(-0.152416774822)*x_ref[2]
        ref[1,1]=(-0.602731976042)*x_ref[0]**o+(0.685046816545)*x_ref[0]+(0.750001644716)*x_ref[1]**o+(-0.202076972289)*x_ref[1]+(-0.0232888078434)*x_ref[2]**o+(-0.0774477649878)*x_ref[2]
        ref[1,2]=(-0.707321785527)*x_ref[0]**o+(0.618526464205)*x_ref[0]+(-0.578548442468)*x_ref[1]**o+(-0.411374744442)*x_ref[1]+(-0.944277708787)*x_ref[2]**o+(0.647856975707)*x_ref[2]
        ref[1,3]=(-0.222503271022)*x_ref[0]**o+(-0.929428092728)*x_ref[0]+(-0.976553212424)*x_ref[1]**o+(0.0376439617084)*x_ref[1]+(0.0428078568512)*x_ref[2]**o+(-0.441850944272)*x_ref[2]
        ref[1,4]=(-0.862164898455)*x_ref[0]**o+(-0.989900221991)*x_ref[0]+(0.110610844567)*x_ref[1]**o+(0.419384612236)*x_ref[1]+(-0.00271948841262)*x_ref[2]**o+(0.301635101629)*x_ref[2]
        ref[2,0]=(-0.0950421881942)*x_ref[0]**o+(-0.741293053909)*x_ref[0]+(-0.461538879874)*x_ref[1]**o+(-0.8199842553)*x_ref[1]+(0.0590344916829)*x_ref[2]**o+(0.352541518434)*x_ref[2]
        ref[2,1]=(-0.734407995647)*x_ref[0]**o+(-0.726167721992)*x_ref[0]+(0.700521229916)*x_ref[1]**o+(0.753555863428)*x_ref[1]+(-0.891638139696)*x_ref[2]**o+(-0.996550711213)*x_ref[2]
        ref[2,2]=(0.777110352629)*x_ref[0]**o+(0.309719201157)*x_ref[0]+(0.026816547522)*x_ref[1]**o+(0.33161010287)*x_ref[1]+(0.257929504352)*x_ref[2]**o+(0.71871314654)*x_ref[2]
        ref[2,3]=(-0.0982522143825)*x_ref[0]**o+(0.0532144836415)*x_ref[0]+(0.155895442381)*x_ref[1]**o+(-0.872222127033)*x_ref[1]+(0.810538165596)*x_ref[2]**o+(-0.917071363009)*x_ref[2]
        ref[2,4]=(0.882674783877)*x_ref[0]**o+(-0.246439463648)*x_ref[0]+(-0.478821457401)*x_ref[1]**o+(-0.629183697296)*x_ref[1]+(0.833283185393)*x_ref[2]**o+(-0.895796651184)*x_ref[2]
        ref[3,0]=(0.949019461395)*x_ref[0]**o+(0.437511668107)*x_ref[0]+(0.46679355983)*x_ref[1]**o+(0.375391913775)*x_ref[1]+(0.992336073925)*x_ref[2]**o+(0.784847856608)*x_ref[2]
        ref[3,1]=(0.10730061928)*x_ref[0]**o+(0.624678931808)*x_ref[0]+(-0.105084917663)*x_ref[1]**o+(-0.0650034223387)*x_ref[1]+(-0.271922353868)*x_ref[2]**o+(-0.182967296038)*x_ref[2]
        ref[3,2]=(0.115856234055)*x_ref[0]**o+(0.711912354922)*x_ref[0]+(-0.317962972122)*x_ref[1]**o+(0.577856092951)*x_ref[1]+(-0.246833922867)*x_ref[2]**o+(0.354116462218)*x_ref[2]
        ref[3,3]=(0.40937236699)*x_ref[0]**o+(0.639252679403)*x_ref[0]+(0.540206943219)*x_ref[1]**o+(0.287006139758)*x_ref[1]+(0.123311433598)*x_ref[2]**o+(-0.0300672780213)*x_ref[2]
        ref[3,4]=(-0.190299257713)*x_ref[0]**o+(-0.022055675261)*x_ref[0]+(-0.640751994447)*x_ref[1]**o+(-0.411801947317)*x_ref[1]+(0.0675008125057)*x_ref[2]**o+(-0.659424255203)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.697323110242)*x[0]**o+(-0.806494185218)*x[0]+(0.387914602484)*x[1]**o+(0.841835487965)*x[1]
        arg[0,0,1]=(-0.77670653908)*x[0]**o+(0.311710212706)*x[0]+(-0.0951864020181)*x[1]**o+(-0.14909139946)*x[1]
        arg[0,1,0]=(0.80933791543)*x[0]**o+(0.72441968276)*x[0]+(-0.446054850085)*x[1]**o+(0.419303534988)*x[1]
        arg[0,1,1]=(0.174464190017)*x[0]**o+(0.330609661878)*x[0]+(0.798365390058)*x[1]**o+(0.354899223817)*x[1]
        arg[1,0,0]=(0.36073672839)*x[0]**o+(0.831900846214)*x[0]+(-0.656354645044)*x[1]**o+(0.200645044646)*x[1]
        arg[1,0,1]=(0.817889801065)*x[0]**o+(0.34052889294)*x[0]+(-0.2416944924)*x[1]**o+(0.0922342148934)*x[1]
        arg[1,1,0]=(-0.962116509558)*x[0]**o+(-0.0785232077542)*x[0]+(-0.981399563356)*x[1]**o+(0.785509095614)*x[1]
        arg[1,1,1]=(-0.688444030508)*x[0]**o+(0.651617638972)*x[0]+(0.160584046401)*x[1]**o+(0.253258991923)*x[1]
        arg[2,0,0]=(0.0728970412574)*x[0]**o+(-0.647328240217)*x[0]+(0.517665494535)*x[1]**o+(-0.416383753573)*x[1]
        arg[2,0,1]=(0.148104739733)*x[0]**o+(-0.378371639837)*x[0]+(0.848608631336)*x[1]**o+(-0.881516167852)*x[1]
        arg[2,1,0]=(-0.506376676256)*x[0]**o+(0.234090340446)*x[0]+(-0.737828959566)*x[1]**o+(-0.74093763135)*x[1]
        arg[2,1,1]=(0.259605354135)*x[0]**o+(0.537424356513)*x[0]+(-0.405718119693)*x[1]**o+(0.120663712044)*x[1]
        arg[3,0,0]=(-0.609659975269)*x[0]**o+(-0.455923293874)*x[0]+(-0.00765676965971)*x[1]**o+(-0.710899349259)*x[1]
        arg[3,0,1]=(-0.731011612904)*x[0]**o+(-0.355060765358)*x[0]+(-0.127369372538)*x[1]**o+(0.694265371908)*x[1]
        arg[3,1,0]=(0.188766337132)*x[0]**o+(-0.194878304215)*x[0]+(-0.963356218109)*x[1]**o+(0.327053372293)*x[1]
        arg[3,1,1]=(-0.566712779324)*x[0]**o+(-0.251473305071)*x[0]+(0.0181297146138)*x[1]**o+(0.257664697754)*x[1]
        arg[4,0,0]=(0.574100764569)*x[0]**o+(-0.717776999574)*x[0]+(0.880998124155)*x[1]**o+(-0.810615928848)*x[1]
        arg[4,0,1]=(0.42552502518)*x[0]**o+(0.456751341147)*x[0]+(-0.484998619005)*x[1]**o+(0.932151187696)*x[1]
        arg[4,1,0]=(0.56865980485)*x[0]**o+(0.593928872734)*x[0]+(-0.475639894836)*x[1]**o+(-0.990718402693)*x[1]
        arg[4,1,1]=(0.129586342261)*x[0]**o+(-0.408951830619)*x[0]+(0.132518475289)*x[1]**o+(0.708612820692)*x[1]
        arg[5,0,0]=(0.0708038739082)*x[0]**o+(0.524270558649)*x[0]+(0.368504533482)*x[1]**o+(0.935828810393)*x[1]
        arg[5,0,1]=(-0.383650306829)*x[0]**o+(-0.0439362943961)*x[0]+(0.80939275198)*x[1]**o+(-0.61335508023)*x[1]
        arg[5,1,0]=(-0.621571339247)*x[0]**o+(-0.185069293199)*x[0]+(0.368472011431)*x[1]**o+(-0.388752495932)*x[1]
        arg[5,1,1]=(-0.263310763174)*x[0]**o+(-0.694616630015)*x[0]+(0.271238274203)*x[1]**o+(-0.477861402347)*x[1]
        ref[0,0,0]=(-0.697323110242)*x_ref[0]**o+(-0.806494185218)*x_ref[0]+(0.387914602484)*x_ref[1]**o+(0.841835487965)*x_ref[1]
        ref[0,0,1]=(-0.77670653908)*x_ref[0]**o+(0.311710212706)*x_ref[0]+(-0.0951864020181)*x_ref[1]**o+(-0.14909139946)*x_ref[1]
        ref[0,1,0]=(0.80933791543)*x_ref[0]**o+(0.72441968276)*x_ref[0]+(-0.446054850085)*x_ref[1]**o+(0.419303534988)*x_ref[1]
        ref[0,1,1]=(0.174464190017)*x_ref[0]**o+(0.330609661878)*x_ref[0]+(0.798365390058)*x_ref[1]**o+(0.354899223817)*x_ref[1]
        ref[1,0,0]=(0.36073672839)*x_ref[0]**o+(0.831900846214)*x_ref[0]+(-0.656354645044)*x_ref[1]**o+(0.200645044646)*x_ref[1]
        ref[1,0,1]=(0.817889801065)*x_ref[0]**o+(0.34052889294)*x_ref[0]+(-0.2416944924)*x_ref[1]**o+(0.0922342148934)*x_ref[1]
        ref[1,1,0]=(-0.962116509558)*x_ref[0]**o+(-0.0785232077542)*x_ref[0]+(-0.981399563356)*x_ref[1]**o+(0.785509095614)*x_ref[1]
        ref[1,1,1]=(-0.688444030508)*x_ref[0]**o+(0.651617638972)*x_ref[0]+(0.160584046401)*x_ref[1]**o+(0.253258991923)*x_ref[1]
        ref[2,0,0]=(0.0728970412574)*x_ref[0]**o+(-0.647328240217)*x_ref[0]+(0.517665494535)*x_ref[1]**o+(-0.416383753573)*x_ref[1]
        ref[2,0,1]=(0.148104739733)*x_ref[0]**o+(-0.378371639837)*x_ref[0]+(0.848608631336)*x_ref[1]**o+(-0.881516167852)*x_ref[1]
        ref[2,1,0]=(-0.506376676256)*x_ref[0]**o+(0.234090340446)*x_ref[0]+(-0.737828959566)*x_ref[1]**o+(-0.74093763135)*x_ref[1]
        ref[2,1,1]=(0.259605354135)*x_ref[0]**o+(0.537424356513)*x_ref[0]+(-0.405718119693)*x_ref[1]**o+(0.120663712044)*x_ref[1]
        ref[3,0,0]=(-0.609659975269)*x_ref[0]**o+(-0.455923293874)*x_ref[0]+(-0.00765676965971)*x_ref[1]**o+(-0.710899349259)*x_ref[1]
        ref[3,0,1]=(-0.731011612904)*x_ref[0]**o+(-0.355060765358)*x_ref[0]+(-0.127369372538)*x_ref[1]**o+(0.694265371908)*x_ref[1]
        ref[3,1,0]=(0.188766337132)*x_ref[0]**o+(-0.194878304215)*x_ref[0]+(-0.963356218109)*x_ref[1]**o+(0.327053372293)*x_ref[1]
        ref[3,1,1]=(-0.566712779324)*x_ref[0]**o+(-0.251473305071)*x_ref[0]+(0.0181297146138)*x_ref[1]**o+(0.257664697754)*x_ref[1]
        ref[4,0,0]=(0.574100764569)*x_ref[0]**o+(-0.717776999574)*x_ref[0]+(0.880998124155)*x_ref[1]**o+(-0.810615928848)*x_ref[1]
        ref[4,0,1]=(0.42552502518)*x_ref[0]**o+(0.456751341147)*x_ref[0]+(-0.484998619005)*x_ref[1]**o+(0.932151187696)*x_ref[1]
        ref[4,1,0]=(0.56865980485)*x_ref[0]**o+(0.593928872734)*x_ref[0]+(-0.475639894836)*x_ref[1]**o+(-0.990718402693)*x_ref[1]
        ref[4,1,1]=(0.129586342261)*x_ref[0]**o+(-0.408951830619)*x_ref[0]+(0.132518475289)*x_ref[1]**o+(0.708612820692)*x_ref[1]
        ref[5,0,0]=(0.0708038739082)*x_ref[0]**o+(0.524270558649)*x_ref[0]+(0.368504533482)*x_ref[1]**o+(0.935828810393)*x_ref[1]
        ref[5,0,1]=(-0.383650306829)*x_ref[0]**o+(-0.0439362943961)*x_ref[0]+(0.80939275198)*x_ref[1]**o+(-0.61335508023)*x_ref[1]
        ref[5,1,0]=(-0.621571339247)*x_ref[0]**o+(-0.185069293199)*x_ref[0]+(0.368472011431)*x_ref[1]**o+(-0.388752495932)*x_ref[1]
        ref[5,1,1]=(-0.263310763174)*x_ref[0]**o+(-0.694616630015)*x_ref[0]+(0.271238274203)*x_ref[1]**o+(-0.477861402347)*x_ref[1]
      else:
        arg[0,0,0]=(0.287534645778)*x[0]**o+(0.94865229379)*x[0]+(0.655299232864)*x[1]**o+(0.350034924637)*x[1]+(0.556341391183)*x[2]**o+(0.840643265765)*x[2]
        arg[0,0,1]=(-0.207213781549)*x[0]**o+(-0.0394497047797)*x[0]+(-0.0236016308575)*x[1]**o+(0.344122299199)*x[1]+(0.389588325996)*x[2]**o+(0.941980468657)*x[2]
        arg[0,1,0]=(0.00980412235862)*x[0]**o+(0.74862549447)*x[0]+(0.630165284383)*x[1]**o+(0.234711268989)*x[1]+(-0.22509981272)*x[2]**o+(-0.0153888168649)*x[2]
        arg[0,1,1]=(-0.87309186262)*x[0]**o+(0.833526923917)*x[0]+(0.78785479415)*x[1]**o+(0.823967259171)*x[1]+(0.320244316109)*x[2]**o+(-0.735225854417)*x[2]
        arg[1,0,0]=(-0.0892232551051)*x[0]**o+(-0.346728833563)*x[0]+(0.422774634671)*x[1]**o+(0.793222381957)*x[1]+(-0.670476677006)*x[2]**o+(0.896884540116)*x[2]
        arg[1,0,1]=(-0.0023382136345)*x[0]**o+(0.73058281545)*x[0]+(0.11828880374)*x[1]**o+(-0.691839091619)*x[1]+(-0.39086463392)*x[2]**o+(0.66200209782)*x[2]
        arg[1,1,0]=(-0.924238055558)*x[0]**o+(-0.334208271705)*x[0]+(-0.57106045555)*x[1]**o+(-0.765650999193)*x[1]+(0.0404533468054)*x[2]**o+(-0.225975260777)*x[2]
        arg[1,1,1]=(-0.351110999108)*x[0]**o+(0.626166346251)*x[0]+(0.385737198467)*x[1]**o+(0.577561466908)*x[1]+(0.708325217242)*x[2]**o+(0.193727923904)*x[2]
        arg[2,0,0]=(0.278942080997)*x[0]**o+(-0.569761653572)*x[0]+(-0.624879270918)*x[1]**o+(-0.0151754112896)*x[1]+(0.411399083825)*x[2]**o+(-0.846669568641)*x[2]
        arg[2,0,1]=(-0.543586963879)*x[0]**o+(-0.383959956625)*x[0]+(0.184375944938)*x[1]**o+(-0.104178233562)*x[1]+(-0.223101894662)*x[2]**o+(-0.17221835595)*x[2]
        arg[2,1,0]=(0.317103921207)*x[0]**o+(-0.919542461354)*x[0]+(0.56666554454)*x[1]**o+(0.895772298847)*x[1]+(0.714550707822)*x[2]**o+(0.634702089556)*x[2]
        arg[2,1,1]=(0.986576254164)*x[0]**o+(0.544493787633)*x[0]+(0.964620027203)*x[1]**o+(0.259991384678)*x[1]+(-0.577361835739)*x[2]**o+(0.651593456822)*x[2]
        arg[3,0,0]=(0.362841472614)*x[0]**o+(-0.521771228041)*x[0]+(-0.594530755517)*x[1]**o+(0.700647681241)*x[1]+(-0.131683781311)*x[2]**o+(0.636310189834)*x[2]
        arg[3,0,1]=(0.885891862321)*x[0]**o+(0.397438320764)*x[0]+(0.711380452937)*x[1]**o+(0.962571325144)*x[1]+(-0.300114113339)*x[2]**o+(0.34700388868)*x[2]
        arg[3,1,0]=(0.52140726408)*x[0]**o+(-0.0921295295522)*x[0]+(-0.823076091446)*x[1]**o+(0.477401876125)*x[1]+(-0.973057905067)*x[2]**o+(-0.975526794128)*x[2]
        arg[3,1,1]=(-0.241503887113)*x[0]**o+(0.133988088285)*x[0]+(0.243026393121)*x[1]**o+(0.0848716779473)*x[1]+(0.140349614791)*x[2]**o+(-0.217182025695)*x[2]
        arg[4,0,0]=(0.680080194053)*x[0]**o+(0.512009633844)*x[0]+(-0.640214066363)*x[1]**o+(0.571800633214)*x[1]+(-0.749395617545)*x[2]**o+(0.448487668786)*x[2]
        arg[4,0,1]=(-0.792959916349)*x[0]**o+(-0.433770516776)*x[0]+(-0.380980758215)*x[1]**o+(-0.831154769101)*x[1]+(0.277551460651)*x[2]**o+(0.626476984233)*x[2]
        arg[4,1,0]=(-0.285102566546)*x[0]**o+(0.745636275253)*x[0]+(0.4999427836)*x[1]**o+(0.561343302884)*x[1]+(0.796220302361)*x[2]**o+(-0.830020520143)*x[2]
        arg[4,1,1]=(0.237420507314)*x[0]**o+(0.327181858092)*x[0]+(-0.895743235593)*x[1]**o+(-0.496964477443)*x[1]+(0.57336428204)*x[2]**o+(-0.154267071287)*x[2]
        arg[5,0,0]=(0.0401925997607)*x[0]**o+(-0.0155957424748)*x[0]+(-0.658363451239)*x[1]**o+(-0.175223211879)*x[1]+(-0.743729768327)*x[2]**o+(0.391531327128)*x[2]
        arg[5,0,1]=(0.836239801824)*x[0]**o+(-0.72288085675)*x[0]+(0.177981042887)*x[1]**o+(0.354505260098)*x[1]+(-0.180520504267)*x[2]**o+(-0.380306239974)*x[2]
        arg[5,1,0]=(-0.65536853024)*x[0]**o+(-0.00573854242476)*x[0]+(-0.261460514314)*x[1]**o+(-0.289771405693)*x[1]+(0.750775397523)*x[2]**o+(-0.333215608562)*x[2]
        arg[5,1,1]=(-0.579532884024)*x[0]**o+(-0.208677373724)*x[0]+(-0.379612024024)*x[1]**o+(-0.0565813995114)*x[1]+(-0.867913562592)*x[2]**o+(-0.254654298293)*x[2]
        ref[0,0,0]=(0.287534645778)*x_ref[0]**o+(0.94865229379)*x_ref[0]+(0.655299232864)*x_ref[1]**o+(0.350034924637)*x_ref[1]+(0.556341391183)*x_ref[2]**o+(0.840643265765)*x_ref[2]
        ref[0,0,1]=(-0.207213781549)*x_ref[0]**o+(-0.0394497047797)*x_ref[0]+(-0.0236016308575)*x_ref[1]**o+(0.344122299199)*x_ref[1]+(0.389588325996)*x_ref[2]**o+(0.941980468657)*x_ref[2]
        ref[0,1,0]=(0.00980412235862)*x_ref[0]**o+(0.74862549447)*x_ref[0]+(0.630165284383)*x_ref[1]**o+(0.234711268989)*x_ref[1]+(-0.22509981272)*x_ref[2]**o+(-0.0153888168649)*x_ref[2]
        ref[0,1,1]=(-0.87309186262)*x_ref[0]**o+(0.833526923917)*x_ref[0]+(0.78785479415)*x_ref[1]**o+(0.823967259171)*x_ref[1]+(0.320244316109)*x_ref[2]**o+(-0.735225854417)*x_ref[2]
        ref[1,0,0]=(-0.0892232551051)*x_ref[0]**o+(-0.346728833563)*x_ref[0]+(0.422774634671)*x_ref[1]**o+(0.793222381957)*x_ref[1]+(-0.670476677006)*x_ref[2]**o+(0.896884540116)*x_ref[2]
        ref[1,0,1]=(-0.0023382136345)*x_ref[0]**o+(0.73058281545)*x_ref[0]+(0.11828880374)*x_ref[1]**o+(-0.691839091619)*x_ref[1]+(-0.39086463392)*x_ref[2]**o+(0.66200209782)*x_ref[2]
        ref[1,1,0]=(-0.924238055558)*x_ref[0]**o+(-0.334208271705)*x_ref[0]+(-0.57106045555)*x_ref[1]**o+(-0.765650999193)*x_ref[1]+(0.0404533468054)*x_ref[2]**o+(-0.225975260777)*x_ref[2]
        ref[1,1,1]=(-0.351110999108)*x_ref[0]**o+(0.626166346251)*x_ref[0]+(0.385737198467)*x_ref[1]**o+(0.577561466908)*x_ref[1]+(0.708325217242)*x_ref[2]**o+(0.193727923904)*x_ref[2]
        ref[2,0,0]=(0.278942080997)*x_ref[0]**o+(-0.569761653572)*x_ref[0]+(-0.624879270918)*x_ref[1]**o+(-0.0151754112896)*x_ref[1]+(0.411399083825)*x_ref[2]**o+(-0.846669568641)*x_ref[2]
        ref[2,0,1]=(-0.543586963879)*x_ref[0]**o+(-0.383959956625)*x_ref[0]+(0.184375944938)*x_ref[1]**o+(-0.104178233562)*x_ref[1]+(-0.223101894662)*x_ref[2]**o+(-0.17221835595)*x_ref[2]
        ref[2,1,0]=(0.317103921207)*x_ref[0]**o+(-0.919542461354)*x_ref[0]+(0.56666554454)*x_ref[1]**o+(0.895772298847)*x_ref[1]+(0.714550707822)*x_ref[2]**o+(0.634702089556)*x_ref[2]
        ref[2,1,1]=(0.986576254164)*x_ref[0]**o+(0.544493787633)*x_ref[0]+(0.964620027203)*x_ref[1]**o+(0.259991384678)*x_ref[1]+(-0.577361835739)*x_ref[2]**o+(0.651593456822)*x_ref[2]
        ref[3,0,0]=(0.362841472614)*x_ref[0]**o+(-0.521771228041)*x_ref[0]+(-0.594530755517)*x_ref[1]**o+(0.700647681241)*x_ref[1]+(-0.131683781311)*x_ref[2]**o+(0.636310189834)*x_ref[2]
        ref[3,0,1]=(0.885891862321)*x_ref[0]**o+(0.397438320764)*x_ref[0]+(0.711380452937)*x_ref[1]**o+(0.962571325144)*x_ref[1]+(-0.300114113339)*x_ref[2]**o+(0.34700388868)*x_ref[2]
        ref[3,1,0]=(0.52140726408)*x_ref[0]**o+(-0.0921295295522)*x_ref[0]+(-0.823076091446)*x_ref[1]**o+(0.477401876125)*x_ref[1]+(-0.973057905067)*x_ref[2]**o+(-0.975526794128)*x_ref[2]
        ref[3,1,1]=(-0.241503887113)*x_ref[0]**o+(0.133988088285)*x_ref[0]+(0.243026393121)*x_ref[1]**o+(0.0848716779473)*x_ref[1]+(0.140349614791)*x_ref[2]**o+(-0.217182025695)*x_ref[2]
        ref[4,0,0]=(0.680080194053)*x_ref[0]**o+(0.512009633844)*x_ref[0]+(-0.640214066363)*x_ref[1]**o+(0.571800633214)*x_ref[1]+(-0.749395617545)*x_ref[2]**o+(0.448487668786)*x_ref[2]
        ref[4,0,1]=(-0.792959916349)*x_ref[0]**o+(-0.433770516776)*x_ref[0]+(-0.380980758215)*x_ref[1]**o+(-0.831154769101)*x_ref[1]+(0.277551460651)*x_ref[2]**o+(0.626476984233)*x_ref[2]
        ref[4,1,0]=(-0.285102566546)*x_ref[0]**o+(0.745636275253)*x_ref[0]+(0.4999427836)*x_ref[1]**o+(0.561343302884)*x_ref[1]+(0.796220302361)*x_ref[2]**o+(-0.830020520143)*x_ref[2]
        ref[4,1,1]=(0.237420507314)*x_ref[0]**o+(0.327181858092)*x_ref[0]+(-0.895743235593)*x_ref[1]**o+(-0.496964477443)*x_ref[1]+(0.57336428204)*x_ref[2]**o+(-0.154267071287)*x_ref[2]
        ref[5,0,0]=(0.0401925997607)*x_ref[0]**o+(-0.0155957424748)*x_ref[0]+(-0.658363451239)*x_ref[1]**o+(-0.175223211879)*x_ref[1]+(-0.743729768327)*x_ref[2]**o+(0.391531327128)*x_ref[2]
        ref[5,0,1]=(0.836239801824)*x_ref[0]**o+(-0.72288085675)*x_ref[0]+(0.177981042887)*x_ref[1]**o+(0.354505260098)*x_ref[1]+(-0.180520504267)*x_ref[2]**o+(-0.380306239974)*x_ref[2]
        ref[5,1,0]=(-0.65536853024)*x_ref[0]**o+(-0.00573854242476)*x_ref[0]+(-0.261460514314)*x_ref[1]**o+(-0.289771405693)*x_ref[1]+(0.750775397523)*x_ref[2]**o+(-0.333215608562)*x_ref[2]
        ref[5,1,1]=(-0.579532884024)*x_ref[0]**o+(-0.208677373724)*x_ref[0]+(-0.379612024024)*x_ref[1]**o+(-0.0565813995114)*x_ref[1]+(-0.867913562592)*x_ref[2]**o+(-0.254654298293)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.274472293956)*x[0]**o+(-0.293347130929)*x[0]+(0.392586448254)*x[1]**o+(-0.403066854054)*x[1]
        arg[0,0,0,1]=(0.385933192907)*x[0]**o+(0.484240065867)*x[0]+(0.144565631551)*x[1]**o+(0.882508492739)*x[1]
        arg[0,0,1,0]=(0.993368905137)*x[0]**o+(0.340349919343)*x[0]+(-0.956806746524)*x[1]**o+(0.990607958001)*x[1]
        arg[0,0,1,1]=(-0.88610250102)*x[0]**o+(-0.543705587259)*x[0]+(0.170798164924)*x[1]**o+(0.685331195754)*x[1]
        arg[0,0,2,0]=(-0.922211611143)*x[0]**o+(-0.103640037893)*x[0]+(0.401354703434)*x[1]**o+(-0.117826261635)*x[1]
        arg[0,0,2,1]=(-0.391601387132)*x[0]**o+(-0.472753344804)*x[0]+(-0.473433838993)*x[1]**o+(-0.173501635989)*x[1]
        arg[0,1,0,0]=(0.747353274916)*x[0]**o+(-0.960856612086)*x[0]+(-0.54809012978)*x[1]**o+(0.862023893332)*x[1]
        arg[0,1,0,1]=(-0.256359001149)*x[0]**o+(-0.868975044191)*x[0]+(0.494314269744)*x[1]**o+(0.949957481172)*x[1]
        arg[0,1,1,0]=(-0.0256658750601)*x[0]**o+(0.145499104203)*x[0]+(-0.68669791629)*x[1]**o+(0.447326207295)*x[1]
        arg[0,1,1,1]=(0.150617284467)*x[0]**o+(-0.350046088196)*x[0]+(-0.367671302053)*x[1]**o+(0.633220122068)*x[1]
        arg[0,1,2,0]=(-0.714145362271)*x[0]**o+(-0.957661490758)*x[0]+(0.386253770747)*x[1]**o+(0.982634262634)*x[1]
        arg[0,1,2,1]=(0.707565096824)*x[0]**o+(0.281758058573)*x[0]+(-0.407791315587)*x[1]**o+(0.86069036983)*x[1]
        arg[0,2,0,0]=(-0.453860220598)*x[0]**o+(0.944457388061)*x[0]+(-0.903818960162)*x[1]**o+(-0.799423205297)*x[1]
        arg[0,2,0,1]=(-0.621257500941)*x[0]**o+(0.748963737937)*x[0]+(-0.12660934458)*x[1]**o+(-0.362830639592)*x[1]
        arg[0,2,1,0]=(-0.702485820037)*x[0]**o+(0.355584898593)*x[0]+(0.657127357751)*x[1]**o+(0.535533390762)*x[1]
        arg[0,2,1,1]=(0.795733125008)*x[0]**o+(0.983607327249)*x[0]+(-0.463818780644)*x[1]**o+(-0.681831588244)*x[1]
        arg[0,2,2,0]=(0.335783204138)*x[0]**o+(-0.712225782679)*x[0]+(0.853426644096)*x[1]**o+(-0.143845263686)*x[1]
        arg[0,2,2,1]=(0.389724375144)*x[0]**o+(0.715176227705)*x[0]+(-0.660690893888)*x[1]**o+(0.428289724709)*x[1]
        arg[0,3,0,0]=(-0.363836203724)*x[0]**o+(-0.257775427457)*x[0]+(0.407904889098)*x[1]**o+(-0.731273814284)*x[1]
        arg[0,3,0,1]=(0.402038262494)*x[0]**o+(0.33289103638)*x[0]+(0.444957562329)*x[1]**o+(0.0657971935664)*x[1]
        arg[0,3,1,0]=(0.979673814602)*x[0]**o+(-0.133405332848)*x[0]+(0.23884689304)*x[1]**o+(0.728464997681)*x[1]
        arg[0,3,1,1]=(0.724603429246)*x[0]**o+(-0.391040898144)*x[0]+(-0.966816499003)*x[1]**o+(0.0984302578905)*x[1]
        arg[0,3,2,0]=(-0.986139645754)*x[0]**o+(0.60807351165)*x[0]+(0.799807413231)*x[1]**o+(0.12091479711)*x[1]
        arg[0,3,2,1]=(0.387453999377)*x[0]**o+(0.212725006628)*x[0]+(-0.120374039139)*x[1]**o+(-0.262304757951)*x[1]
        arg[0,4,0,0]=(-0.260545986276)*x[0]**o+(0.251085312932)*x[0]+(0.738082708518)*x[1]**o+(-0.0804883797938)*x[1]
        arg[0,4,0,1]=(-0.190701356138)*x[0]**o+(-0.0205295935522)*x[0]+(0.904261681819)*x[1]**o+(-0.859647709957)*x[1]
        arg[0,4,1,0]=(-0.945357492849)*x[0]**o+(-0.85298463389)*x[0]+(0.688964336673)*x[1]**o+(-0.672881696932)*x[1]
        arg[0,4,1,1]=(0.877720258045)*x[0]**o+(-0.361228026805)*x[0]+(-0.326626527006)*x[1]**o+(-0.274100421232)*x[1]
        arg[0,4,2,0]=(0.202358729084)*x[0]**o+(-0.997571481837)*x[0]+(0.431193711862)*x[1]**o+(0.513561123915)*x[1]
        arg[0,4,2,1]=(-0.136602954592)*x[0]**o+(-0.40864934437)*x[0]+(0.675945574351)*x[1]**o+(0.728757793536)*x[1]
        arg[1,0,0,0]=(0.835464261851)*x[0]**o+(-0.131447020601)*x[0]+(0.0810352521446)*x[1]**o+(0.664744750899)*x[1]
        arg[1,0,0,1]=(0.239716928689)*x[0]**o+(-0.192258999531)*x[0]+(0.615877054036)*x[1]**o+(0.835274774437)*x[1]
        arg[1,0,1,0]=(0.509342803767)*x[0]**o+(-0.633944481882)*x[0]+(0.262098006086)*x[1]**o+(0.159902070799)*x[1]
        arg[1,0,1,1]=(0.894432440588)*x[0]**o+(-0.167778431319)*x[0]+(-0.984183148477)*x[1]**o+(-0.554981658632)*x[1]
        arg[1,0,2,0]=(-0.762472607231)*x[0]**o+(0.405550139256)*x[0]+(0.4051290641)*x[1]**o+(-0.27180186672)*x[1]
        arg[1,0,2,1]=(0.663869273816)*x[0]**o+(0.867453974628)*x[0]+(-0.279659969132)*x[1]**o+(-0.370539101204)*x[1]
        arg[1,1,0,0]=(-0.969326778164)*x[0]**o+(0.965471349497)*x[0]+(-0.529983232946)*x[1]**o+(0.714091610977)*x[1]
        arg[1,1,0,1]=(-0.840360667093)*x[0]**o+(-0.399319808256)*x[0]+(0.568769915582)*x[1]**o+(0.872427494442)*x[1]
        arg[1,1,1,0]=(0.46099116941)*x[0]**o+(-0.986008874142)*x[0]+(-0.53999023556)*x[1]**o+(-0.549695567855)*x[1]
        arg[1,1,1,1]=(0.485944918559)*x[0]**o+(0.189773828651)*x[0]+(-0.611654980493)*x[1]**o+(-0.337071911657)*x[1]
        arg[1,1,2,0]=(0.191663603248)*x[0]**o+(0.0906374965004)*x[0]+(-0.271051598251)*x[1]**o+(0.880507788825)*x[1]
        arg[1,1,2,1]=(-0.439514997001)*x[0]**o+(0.701230426224)*x[0]+(-0.193231641029)*x[1]**o+(0.950584054332)*x[1]
        arg[1,2,0,0]=(0.439456565259)*x[0]**o+(-0.25664171484)*x[0]+(-0.653059819329)*x[1]**o+(-0.680544747777)*x[1]
        arg[1,2,0,1]=(0.34890454443)*x[0]**o+(-0.69851200963)*x[0]+(-0.00159381688549)*x[1]**o+(-0.612359675499)*x[1]
        arg[1,2,1,0]=(0.132759671375)*x[0]**o+(-0.235126716401)*x[0]+(0.652068885826)*x[1]**o+(0.749698645371)*x[1]
        arg[1,2,1,1]=(0.934980403627)*x[0]**o+(0.321409273437)*x[0]+(-0.605909929553)*x[1]**o+(0.135554927418)*x[1]
        arg[1,2,2,0]=(-0.68327709329)*x[0]**o+(-0.0381292756005)*x[0]+(0.710953191114)*x[1]**o+(-0.344522256689)*x[1]
        arg[1,2,2,1]=(-0.813976235652)*x[0]**o+(0.613879708123)*x[0]+(0.451599277293)*x[1]**o+(-0.6134026375)*x[1]
        arg[1,3,0,0]=(0.595310128627)*x[0]**o+(-0.247621573131)*x[0]+(0.165607783104)*x[1]**o+(0.578132180805)*x[1]
        arg[1,3,0,1]=(0.848827238882)*x[0]**o+(0.173154067999)*x[0]+(0.340662419437)*x[1]**o+(-0.389656447251)*x[1]
        arg[1,3,1,0]=(0.250888415228)*x[0]**o+(-0.278203136198)*x[0]+(-0.693211186734)*x[1]**o+(0.613812194735)*x[1]
        arg[1,3,1,1]=(0.634533003162)*x[0]**o+(-0.0924665762471)*x[0]+(0.749575894384)*x[1]**o+(-0.737205017415)*x[1]
        arg[1,3,2,0]=(0.607750966973)*x[0]**o+(0.905016928404)*x[0]+(-0.171049542163)*x[1]**o+(-0.20526244601)*x[1]
        arg[1,3,2,1]=(0.723946025405)*x[0]**o+(0.0157576743987)*x[0]+(0.190875428312)*x[1]**o+(0.407925899509)*x[1]
        arg[1,4,0,0]=(-0.561582695803)*x[0]**o+(-0.350262461371)*x[0]+(0.57610409872)*x[1]**o+(0.881780963877)*x[1]
        arg[1,4,0,1]=(0.077066498695)*x[0]**o+(0.515597802516)*x[0]+(0.332011413016)*x[1]**o+(0.215491334787)*x[1]
        arg[1,4,1,0]=(0.33858727737)*x[0]**o+(-0.573231800768)*x[0]+(-0.559915809132)*x[1]**o+(0.0240869142176)*x[1]
        arg[1,4,1,1]=(-0.684581727952)*x[0]**o+(0.365541904201)*x[0]+(0.204252397376)*x[1]**o+(0.149300766054)*x[1]
        arg[1,4,2,0]=(-0.574225088383)*x[0]**o+(0.0286339776067)*x[0]+(-0.358732122607)*x[1]**o+(-0.416932738957)*x[1]
        arg[1,4,2,1]=(-0.308460184932)*x[0]**o+(0.389214987836)*x[0]+(-0.358303257031)*x[1]**o+(0.348109483688)*x[1]
        arg[2,0,0,0]=(-0.811318668804)*x[0]**o+(-0.298028116982)*x[0]+(0.679871523483)*x[1]**o+(0.958042925287)*x[1]
        arg[2,0,0,1]=(-0.616521818338)*x[0]**o+(-0.0817124854496)*x[0]+(0.403257366893)*x[1]**o+(0.740689710435)*x[1]
        arg[2,0,1,0]=(0.271046179127)*x[0]**o+(0.814928093947)*x[0]+(0.423722781579)*x[1]**o+(0.925333209811)*x[1]
        arg[2,0,1,1]=(0.0596651806816)*x[0]**o+(0.493854605359)*x[0]+(0.539998537439)*x[1]**o+(-0.910865639375)*x[1]
        arg[2,0,2,0]=(0.503267602327)*x[0]**o+(-0.270966471743)*x[0]+(-0.342063366676)*x[1]**o+(-0.975700224705)*x[1]
        arg[2,0,2,1]=(-0.0362358939195)*x[0]**o+(-0.766033869339)*x[0]+(0.891365100088)*x[1]**o+(0.815025898929)*x[1]
        arg[2,1,0,0]=(-0.667240459984)*x[0]**o+(0.8659076701)*x[0]+(0.215089397562)*x[1]**o+(0.464769182307)*x[1]
        arg[2,1,0,1]=(-0.298625221931)*x[0]**o+(-0.504329743472)*x[0]+(0.412472510208)*x[1]**o+(0.241693152889)*x[1]
        arg[2,1,1,0]=(-0.670730179133)*x[0]**o+(0.0201070113832)*x[0]+(-0.572725898919)*x[1]**o+(0.665609259014)*x[1]
        arg[2,1,1,1]=(-0.319080984089)*x[0]**o+(0.0299423045224)*x[0]+(0.0152311810613)*x[1]**o+(-0.784646189026)*x[1]
        arg[2,1,2,0]=(0.153308789716)*x[0]**o+(0.725808150613)*x[0]+(0.62876493338)*x[1]**o+(0.80275284252)*x[1]
        arg[2,1,2,1]=(0.804698870013)*x[0]**o+(-0.631078773061)*x[0]+(-0.015221975698)*x[1]**o+(-0.954169686746)*x[1]
        arg[2,2,0,0]=(0.169022797498)*x[0]**o+(0.733025791099)*x[0]+(0.321554996633)*x[1]**o+(0.292173365364)*x[1]
        arg[2,2,0,1]=(0.327345955109)*x[0]**o+(0.21757823896)*x[0]+(0.873027160446)*x[1]**o+(-0.250192861579)*x[1]
        arg[2,2,1,0]=(0.394387190832)*x[0]**o+(-0.613661883482)*x[0]+(0.806965291386)*x[1]**o+(0.244513244016)*x[1]
        arg[2,2,1,1]=(0.043913907536)*x[0]**o+(-0.303567806424)*x[0]+(0.547037093323)*x[1]**o+(-0.598687385328)*x[1]
        arg[2,2,2,0]=(0.875663154787)*x[0]**o+(-0.34984832772)*x[0]+(0.879143174396)*x[1]**o+(-0.193154051109)*x[1]
        arg[2,2,2,1]=(0.940311132918)*x[0]**o+(-0.358708743382)*x[0]+(0.117370945146)*x[1]**o+(-0.664208034757)*x[1]
        arg[2,3,0,0]=(0.0334525660588)*x[0]**o+(0.592805915255)*x[0]+(-0.278216157971)*x[1]**o+(0.894369967671)*x[1]
        arg[2,3,0,1]=(-0.966591106326)*x[0]**o+(-0.882160390768)*x[0]+(0.843982702732)*x[1]**o+(0.363868921656)*x[1]
        arg[2,3,1,0]=(-0.246392342771)*x[0]**o+(-0.934459096242)*x[0]+(-0.472771368537)*x[1]**o+(0.148854304706)*x[1]
        arg[2,3,1,1]=(0.248218424051)*x[0]**o+(0.0980396576111)*x[0]+(0.413322166489)*x[1]**o+(0.429747261346)*x[1]
        arg[2,3,2,0]=(0.742987565733)*x[0]**o+(-0.213736450136)*x[0]+(0.996975741314)*x[1]**o+(0.960709958665)*x[1]
        arg[2,3,2,1]=(0.436149490977)*x[0]**o+(0.87632181197)*x[0]+(-0.881934669912)*x[1]**o+(-0.457520504175)*x[1]
        arg[2,4,0,0]=(-0.519014330693)*x[0]**o+(0.938619608407)*x[0]+(-0.847954563609)*x[1]**o+(0.791774379621)*x[1]
        arg[2,4,0,1]=(-0.599487299453)*x[0]**o+(0.977867555132)*x[0]+(0.693105630579)*x[1]**o+(-0.688827812942)*x[1]
        arg[2,4,1,0]=(0.955907964163)*x[0]**o+(0.0724056986172)*x[0]+(0.254773619772)*x[1]**o+(0.0428861146964)*x[1]
        arg[2,4,1,1]=(-0.422202101998)*x[0]**o+(0.803263223544)*x[0]+(0.456048140785)*x[1]**o+(0.0419133126183)*x[1]
        arg[2,4,2,0]=(-0.182308704781)*x[0]**o+(0.862282443488)*x[0]+(-0.844027007955)*x[1]**o+(0.586220526376)*x[1]
        arg[2,4,2,1]=(-0.801542666229)*x[0]**o+(0.903685299455)*x[0]+(0.428926544377)*x[1]**o+(-0.891220471549)*x[1]
        arg[3,0,0,0]=(0.602917787713)*x[0]**o+(-0.592239549743)*x[0]+(0.761830006718)*x[1]**o+(-0.784264167777)*x[1]
        arg[3,0,0,1]=(0.819962957732)*x[0]**o+(-0.455126790372)*x[0]+(-0.144898728525)*x[1]**o+(0.121566345339)*x[1]
        arg[3,0,1,0]=(0.570694941559)*x[0]**o+(-0.505678656386)*x[0]+(0.295536587355)*x[1]**o+(-0.852779737814)*x[1]
        arg[3,0,1,1]=(-0.226246858753)*x[0]**o+(-0.513900306228)*x[0]+(-0.389483591407)*x[1]**o+(-0.735105335539)*x[1]
        arg[3,0,2,0]=(0.941877610938)*x[0]**o+(0.93808330298)*x[0]+(0.189785162182)*x[1]**o+(0.546525366304)*x[1]
        arg[3,0,2,1]=(0.10639100485)*x[0]**o+(-0.773361007972)*x[0]+(-0.470531427579)*x[1]**o+(0.467359342808)*x[1]
        arg[3,1,0,0]=(0.857725599589)*x[0]**o+(0.0881746413132)*x[0]+(0.040713151152)*x[1]**o+(-0.52070729883)*x[1]
        arg[3,1,0,1]=(0.266807652146)*x[0]**o+(0.337496673892)*x[0]+(-0.562266297445)*x[1]**o+(0.326015195113)*x[1]
        arg[3,1,1,0]=(0.177275509465)*x[0]**o+(-0.647855761301)*x[0]+(-0.0963373631995)*x[1]**o+(-0.535256252994)*x[1]
        arg[3,1,1,1]=(0.430309295606)*x[0]**o+(-0.0894749557428)*x[0]+(0.503470724483)*x[1]**o+(-0.202361122248)*x[1]
        arg[3,1,2,0]=(0.283995014194)*x[0]**o+(-0.544419907194)*x[0]+(0.78478571712)*x[1]**o+(0.341540664147)*x[1]
        arg[3,1,2,1]=(-0.281042869936)*x[0]**o+(0.728771766492)*x[0]+(-0.298974388557)*x[1]**o+(-0.836789906985)*x[1]
        arg[3,2,0,0]=(-0.415323371561)*x[0]**o+(0.583971137371)*x[0]+(-0.265658679489)*x[1]**o+(-0.432447373809)*x[1]
        arg[3,2,0,1]=(0.690486298855)*x[0]**o+(0.552521632226)*x[0]+(0.237622739042)*x[1]**o+(-0.254513693924)*x[1]
        arg[3,2,1,0]=(0.453985687207)*x[0]**o+(-0.785385580962)*x[0]+(0.0190987841634)*x[1]**o+(-0.7923288861)*x[1]
        arg[3,2,1,1]=(-0.964082656379)*x[0]**o+(-0.806315937371)*x[0]+(0.271145420366)*x[1]**o+(-0.839014535945)*x[1]
        arg[3,2,2,0]=(-0.372062773997)*x[0]**o+(-0.069749560567)*x[0]+(-0.126310395383)*x[1]**o+(0.972025865285)*x[1]
        arg[3,2,2,1]=(-0.580353020086)*x[0]**o+(-0.152746821477)*x[0]+(0.297553660535)*x[1]**o+(0.444791568145)*x[1]
        arg[3,3,0,0]=(-0.361520029439)*x[0]**o+(-0.810391070499)*x[0]+(0.462123882642)*x[1]**o+(-0.974534227742)*x[1]
        arg[3,3,0,1]=(0.113370617837)*x[0]**o+(0.428257355435)*x[0]+(-0.18656623758)*x[1]**o+(0.206503327394)*x[1]
        arg[3,3,1,0]=(0.341328445689)*x[0]**o+(0.34002947054)*x[0]+(0.997677286362)*x[1]**o+(0.43300073225)*x[1]
        arg[3,3,1,1]=(-0.513912982823)*x[0]**o+(0.826573827064)*x[0]+(-0.701417864085)*x[1]**o+(0.567078883219)*x[1]
        arg[3,3,2,0]=(-0.325536866965)*x[0]**o+(0.924405767451)*x[0]+(-0.0481381954346)*x[1]**o+(0.912210617073)*x[1]
        arg[3,3,2,1]=(0.491532434302)*x[0]**o+(-0.584980848036)*x[0]+(-0.788317530858)*x[1]**o+(-0.900681893172)*x[1]
        arg[3,4,0,0]=(0.432249855148)*x[0]**o+(-0.673369312552)*x[0]+(0.244704845849)*x[1]**o+(-0.300995595367)*x[1]
        arg[3,4,0,1]=(0.383037211984)*x[0]**o+(0.299711400488)*x[0]+(0.870584255454)*x[1]**o+(-0.33571823329)*x[1]
        arg[3,4,1,0]=(0.737134180177)*x[0]**o+(-0.101307090628)*x[0]+(-0.970632061143)*x[1]**o+(-0.449042684137)*x[1]
        arg[3,4,1,1]=(0.00684250643632)*x[0]**o+(0.676964654022)*x[0]+(0.789069834858)*x[1]**o+(-0.829739508197)*x[1]
        arg[3,4,2,0]=(0.7279044944)*x[0]**o+(0.215711122038)*x[0]+(0.385631252632)*x[1]**o+(0.629390433529)*x[1]
        arg[3,4,2,1]=(-0.380750716629)*x[0]**o+(-0.413040692437)*x[0]+(0.26795244639)*x[1]**o+(0.492475504659)*x[1]
        ref[0,0,0,0]=(-0.274472293956)*x_ref[0]**o+(-0.293347130929)*x_ref[0]+(0.392586448254)*x_ref[1]**o+(-0.403066854054)*x_ref[1]
        ref[0,0,0,1]=(0.385933192907)*x_ref[0]**o+(0.484240065867)*x_ref[0]+(0.144565631551)*x_ref[1]**o+(0.882508492739)*x_ref[1]
        ref[0,0,1,0]=(0.993368905137)*x_ref[0]**o+(0.340349919343)*x_ref[0]+(-0.956806746524)*x_ref[1]**o+(0.990607958001)*x_ref[1]
        ref[0,0,1,1]=(-0.88610250102)*x_ref[0]**o+(-0.543705587259)*x_ref[0]+(0.170798164924)*x_ref[1]**o+(0.685331195754)*x_ref[1]
        ref[0,0,2,0]=(-0.922211611143)*x_ref[0]**o+(-0.103640037893)*x_ref[0]+(0.401354703434)*x_ref[1]**o+(-0.117826261635)*x_ref[1]
        ref[0,0,2,1]=(-0.391601387132)*x_ref[0]**o+(-0.472753344804)*x_ref[0]+(-0.473433838993)*x_ref[1]**o+(-0.173501635989)*x_ref[1]
        ref[0,1,0,0]=(0.747353274916)*x_ref[0]**o+(-0.960856612086)*x_ref[0]+(-0.54809012978)*x_ref[1]**o+(0.862023893332)*x_ref[1]
        ref[0,1,0,1]=(-0.256359001149)*x_ref[0]**o+(-0.868975044191)*x_ref[0]+(0.494314269744)*x_ref[1]**o+(0.949957481172)*x_ref[1]
        ref[0,1,1,0]=(-0.0256658750601)*x_ref[0]**o+(0.145499104203)*x_ref[0]+(-0.68669791629)*x_ref[1]**o+(0.447326207295)*x_ref[1]
        ref[0,1,1,1]=(0.150617284467)*x_ref[0]**o+(-0.350046088196)*x_ref[0]+(-0.367671302053)*x_ref[1]**o+(0.633220122068)*x_ref[1]
        ref[0,1,2,0]=(-0.714145362271)*x_ref[0]**o+(-0.957661490758)*x_ref[0]+(0.386253770747)*x_ref[1]**o+(0.982634262634)*x_ref[1]
        ref[0,1,2,1]=(0.707565096824)*x_ref[0]**o+(0.281758058573)*x_ref[0]+(-0.407791315587)*x_ref[1]**o+(0.86069036983)*x_ref[1]
        ref[0,2,0,0]=(-0.453860220598)*x_ref[0]**o+(0.944457388061)*x_ref[0]+(-0.903818960162)*x_ref[1]**o+(-0.799423205297)*x_ref[1]
        ref[0,2,0,1]=(-0.621257500941)*x_ref[0]**o+(0.748963737937)*x_ref[0]+(-0.12660934458)*x_ref[1]**o+(-0.362830639592)*x_ref[1]
        ref[0,2,1,0]=(-0.702485820037)*x_ref[0]**o+(0.355584898593)*x_ref[0]+(0.657127357751)*x_ref[1]**o+(0.535533390762)*x_ref[1]
        ref[0,2,1,1]=(0.795733125008)*x_ref[0]**o+(0.983607327249)*x_ref[0]+(-0.463818780644)*x_ref[1]**o+(-0.681831588244)*x_ref[1]
        ref[0,2,2,0]=(0.335783204138)*x_ref[0]**o+(-0.712225782679)*x_ref[0]+(0.853426644096)*x_ref[1]**o+(-0.143845263686)*x_ref[1]
        ref[0,2,2,1]=(0.389724375144)*x_ref[0]**o+(0.715176227705)*x_ref[0]+(-0.660690893888)*x_ref[1]**o+(0.428289724709)*x_ref[1]
        ref[0,3,0,0]=(-0.363836203724)*x_ref[0]**o+(-0.257775427457)*x_ref[0]+(0.407904889098)*x_ref[1]**o+(-0.731273814284)*x_ref[1]
        ref[0,3,0,1]=(0.402038262494)*x_ref[0]**o+(0.33289103638)*x_ref[0]+(0.444957562329)*x_ref[1]**o+(0.0657971935664)*x_ref[1]
        ref[0,3,1,0]=(0.979673814602)*x_ref[0]**o+(-0.133405332848)*x_ref[0]+(0.23884689304)*x_ref[1]**o+(0.728464997681)*x_ref[1]
        ref[0,3,1,1]=(0.724603429246)*x_ref[0]**o+(-0.391040898144)*x_ref[0]+(-0.966816499003)*x_ref[1]**o+(0.0984302578905)*x_ref[1]
        ref[0,3,2,0]=(-0.986139645754)*x_ref[0]**o+(0.60807351165)*x_ref[0]+(0.799807413231)*x_ref[1]**o+(0.12091479711)*x_ref[1]
        ref[0,3,2,1]=(0.387453999377)*x_ref[0]**o+(0.212725006628)*x_ref[0]+(-0.120374039139)*x_ref[1]**o+(-0.262304757951)*x_ref[1]
        ref[0,4,0,0]=(-0.260545986276)*x_ref[0]**o+(0.251085312932)*x_ref[0]+(0.738082708518)*x_ref[1]**o+(-0.0804883797938)*x_ref[1]
        ref[0,4,0,1]=(-0.190701356138)*x_ref[0]**o+(-0.0205295935522)*x_ref[0]+(0.904261681819)*x_ref[1]**o+(-0.859647709957)*x_ref[1]
        ref[0,4,1,0]=(-0.945357492849)*x_ref[0]**o+(-0.85298463389)*x_ref[0]+(0.688964336673)*x_ref[1]**o+(-0.672881696932)*x_ref[1]
        ref[0,4,1,1]=(0.877720258045)*x_ref[0]**o+(-0.361228026805)*x_ref[0]+(-0.326626527006)*x_ref[1]**o+(-0.274100421232)*x_ref[1]
        ref[0,4,2,0]=(0.202358729084)*x_ref[0]**o+(-0.997571481837)*x_ref[0]+(0.431193711862)*x_ref[1]**o+(0.513561123915)*x_ref[1]
        ref[0,4,2,1]=(-0.136602954592)*x_ref[0]**o+(-0.40864934437)*x_ref[0]+(0.675945574351)*x_ref[1]**o+(0.728757793536)*x_ref[1]
        ref[1,0,0,0]=(0.835464261851)*x_ref[0]**o+(-0.131447020601)*x_ref[0]+(0.0810352521446)*x_ref[1]**o+(0.664744750899)*x_ref[1]
        ref[1,0,0,1]=(0.239716928689)*x_ref[0]**o+(-0.192258999531)*x_ref[0]+(0.615877054036)*x_ref[1]**o+(0.835274774437)*x_ref[1]
        ref[1,0,1,0]=(0.509342803767)*x_ref[0]**o+(-0.633944481882)*x_ref[0]+(0.262098006086)*x_ref[1]**o+(0.159902070799)*x_ref[1]
        ref[1,0,1,1]=(0.894432440588)*x_ref[0]**o+(-0.167778431319)*x_ref[0]+(-0.984183148477)*x_ref[1]**o+(-0.554981658632)*x_ref[1]
        ref[1,0,2,0]=(-0.762472607231)*x_ref[0]**o+(0.405550139256)*x_ref[0]+(0.4051290641)*x_ref[1]**o+(-0.27180186672)*x_ref[1]
        ref[1,0,2,1]=(0.663869273816)*x_ref[0]**o+(0.867453974628)*x_ref[0]+(-0.279659969132)*x_ref[1]**o+(-0.370539101204)*x_ref[1]
        ref[1,1,0,0]=(-0.969326778164)*x_ref[0]**o+(0.965471349497)*x_ref[0]+(-0.529983232946)*x_ref[1]**o+(0.714091610977)*x_ref[1]
        ref[1,1,0,1]=(-0.840360667093)*x_ref[0]**o+(-0.399319808256)*x_ref[0]+(0.568769915582)*x_ref[1]**o+(0.872427494442)*x_ref[1]
        ref[1,1,1,0]=(0.46099116941)*x_ref[0]**o+(-0.986008874142)*x_ref[0]+(-0.53999023556)*x_ref[1]**o+(-0.549695567855)*x_ref[1]
        ref[1,1,1,1]=(0.485944918559)*x_ref[0]**o+(0.189773828651)*x_ref[0]+(-0.611654980493)*x_ref[1]**o+(-0.337071911657)*x_ref[1]
        ref[1,1,2,0]=(0.191663603248)*x_ref[0]**o+(0.0906374965004)*x_ref[0]+(-0.271051598251)*x_ref[1]**o+(0.880507788825)*x_ref[1]
        ref[1,1,2,1]=(-0.439514997001)*x_ref[0]**o+(0.701230426224)*x_ref[0]+(-0.193231641029)*x_ref[1]**o+(0.950584054332)*x_ref[1]
        ref[1,2,0,0]=(0.439456565259)*x_ref[0]**o+(-0.25664171484)*x_ref[0]+(-0.653059819329)*x_ref[1]**o+(-0.680544747777)*x_ref[1]
        ref[1,2,0,1]=(0.34890454443)*x_ref[0]**o+(-0.69851200963)*x_ref[0]+(-0.00159381688549)*x_ref[1]**o+(-0.612359675499)*x_ref[1]
        ref[1,2,1,0]=(0.132759671375)*x_ref[0]**o+(-0.235126716401)*x_ref[0]+(0.652068885826)*x_ref[1]**o+(0.749698645371)*x_ref[1]
        ref[1,2,1,1]=(0.934980403627)*x_ref[0]**o+(0.321409273437)*x_ref[0]+(-0.605909929553)*x_ref[1]**o+(0.135554927418)*x_ref[1]
        ref[1,2,2,0]=(-0.68327709329)*x_ref[0]**o+(-0.0381292756005)*x_ref[0]+(0.710953191114)*x_ref[1]**o+(-0.344522256689)*x_ref[1]
        ref[1,2,2,1]=(-0.813976235652)*x_ref[0]**o+(0.613879708123)*x_ref[0]+(0.451599277293)*x_ref[1]**o+(-0.6134026375)*x_ref[1]
        ref[1,3,0,0]=(0.595310128627)*x_ref[0]**o+(-0.247621573131)*x_ref[0]+(0.165607783104)*x_ref[1]**o+(0.578132180805)*x_ref[1]
        ref[1,3,0,1]=(0.848827238882)*x_ref[0]**o+(0.173154067999)*x_ref[0]+(0.340662419437)*x_ref[1]**o+(-0.389656447251)*x_ref[1]
        ref[1,3,1,0]=(0.250888415228)*x_ref[0]**o+(-0.278203136198)*x_ref[0]+(-0.693211186734)*x_ref[1]**o+(0.613812194735)*x_ref[1]
        ref[1,3,1,1]=(0.634533003162)*x_ref[0]**o+(-0.0924665762471)*x_ref[0]+(0.749575894384)*x_ref[1]**o+(-0.737205017415)*x_ref[1]
        ref[1,3,2,0]=(0.607750966973)*x_ref[0]**o+(0.905016928404)*x_ref[0]+(-0.171049542163)*x_ref[1]**o+(-0.20526244601)*x_ref[1]
        ref[1,3,2,1]=(0.723946025405)*x_ref[0]**o+(0.0157576743987)*x_ref[0]+(0.190875428312)*x_ref[1]**o+(0.407925899509)*x_ref[1]
        ref[1,4,0,0]=(-0.561582695803)*x_ref[0]**o+(-0.350262461371)*x_ref[0]+(0.57610409872)*x_ref[1]**o+(0.881780963877)*x_ref[1]
        ref[1,4,0,1]=(0.077066498695)*x_ref[0]**o+(0.515597802516)*x_ref[0]+(0.332011413016)*x_ref[1]**o+(0.215491334787)*x_ref[1]
        ref[1,4,1,0]=(0.33858727737)*x_ref[0]**o+(-0.573231800768)*x_ref[0]+(-0.559915809132)*x_ref[1]**o+(0.0240869142176)*x_ref[1]
        ref[1,4,1,1]=(-0.684581727952)*x_ref[0]**o+(0.365541904201)*x_ref[0]+(0.204252397376)*x_ref[1]**o+(0.149300766054)*x_ref[1]
        ref[1,4,2,0]=(-0.574225088383)*x_ref[0]**o+(0.0286339776067)*x_ref[0]+(-0.358732122607)*x_ref[1]**o+(-0.416932738957)*x_ref[1]
        ref[1,4,2,1]=(-0.308460184932)*x_ref[0]**o+(0.389214987836)*x_ref[0]+(-0.358303257031)*x_ref[1]**o+(0.348109483688)*x_ref[1]
        ref[2,0,0,0]=(-0.811318668804)*x_ref[0]**o+(-0.298028116982)*x_ref[0]+(0.679871523483)*x_ref[1]**o+(0.958042925287)*x_ref[1]
        ref[2,0,0,1]=(-0.616521818338)*x_ref[0]**o+(-0.0817124854496)*x_ref[0]+(0.403257366893)*x_ref[1]**o+(0.740689710435)*x_ref[1]
        ref[2,0,1,0]=(0.271046179127)*x_ref[0]**o+(0.814928093947)*x_ref[0]+(0.423722781579)*x_ref[1]**o+(0.925333209811)*x_ref[1]
        ref[2,0,1,1]=(0.0596651806816)*x_ref[0]**o+(0.493854605359)*x_ref[0]+(0.539998537439)*x_ref[1]**o+(-0.910865639375)*x_ref[1]
        ref[2,0,2,0]=(0.503267602327)*x_ref[0]**o+(-0.270966471743)*x_ref[0]+(-0.342063366676)*x_ref[1]**o+(-0.975700224705)*x_ref[1]
        ref[2,0,2,1]=(-0.0362358939195)*x_ref[0]**o+(-0.766033869339)*x_ref[0]+(0.891365100088)*x_ref[1]**o+(0.815025898929)*x_ref[1]
        ref[2,1,0,0]=(-0.667240459984)*x_ref[0]**o+(0.8659076701)*x_ref[0]+(0.215089397562)*x_ref[1]**o+(0.464769182307)*x_ref[1]
        ref[2,1,0,1]=(-0.298625221931)*x_ref[0]**o+(-0.504329743472)*x_ref[0]+(0.412472510208)*x_ref[1]**o+(0.241693152889)*x_ref[1]
        ref[2,1,1,0]=(-0.670730179133)*x_ref[0]**o+(0.0201070113832)*x_ref[0]+(-0.572725898919)*x_ref[1]**o+(0.665609259014)*x_ref[1]
        ref[2,1,1,1]=(-0.319080984089)*x_ref[0]**o+(0.0299423045224)*x_ref[0]+(0.0152311810613)*x_ref[1]**o+(-0.784646189026)*x_ref[1]
        ref[2,1,2,0]=(0.153308789716)*x_ref[0]**o+(0.725808150613)*x_ref[0]+(0.62876493338)*x_ref[1]**o+(0.80275284252)*x_ref[1]
        ref[2,1,2,1]=(0.804698870013)*x_ref[0]**o+(-0.631078773061)*x_ref[0]+(-0.015221975698)*x_ref[1]**o+(-0.954169686746)*x_ref[1]
        ref[2,2,0,0]=(0.169022797498)*x_ref[0]**o+(0.733025791099)*x_ref[0]+(0.321554996633)*x_ref[1]**o+(0.292173365364)*x_ref[1]
        ref[2,2,0,1]=(0.327345955109)*x_ref[0]**o+(0.21757823896)*x_ref[0]+(0.873027160446)*x_ref[1]**o+(-0.250192861579)*x_ref[1]
        ref[2,2,1,0]=(0.394387190832)*x_ref[0]**o+(-0.613661883482)*x_ref[0]+(0.806965291386)*x_ref[1]**o+(0.244513244016)*x_ref[1]
        ref[2,2,1,1]=(0.043913907536)*x_ref[0]**o+(-0.303567806424)*x_ref[0]+(0.547037093323)*x_ref[1]**o+(-0.598687385328)*x_ref[1]
        ref[2,2,2,0]=(0.875663154787)*x_ref[0]**o+(-0.34984832772)*x_ref[0]+(0.879143174396)*x_ref[1]**o+(-0.193154051109)*x_ref[1]
        ref[2,2,2,1]=(0.940311132918)*x_ref[0]**o+(-0.358708743382)*x_ref[0]+(0.117370945146)*x_ref[1]**o+(-0.664208034757)*x_ref[1]
        ref[2,3,0,0]=(0.0334525660588)*x_ref[0]**o+(0.592805915255)*x_ref[0]+(-0.278216157971)*x_ref[1]**o+(0.894369967671)*x_ref[1]
        ref[2,3,0,1]=(-0.966591106326)*x_ref[0]**o+(-0.882160390768)*x_ref[0]+(0.843982702732)*x_ref[1]**o+(0.363868921656)*x_ref[1]
        ref[2,3,1,0]=(-0.246392342771)*x_ref[0]**o+(-0.934459096242)*x_ref[0]+(-0.472771368537)*x_ref[1]**o+(0.148854304706)*x_ref[1]
        ref[2,3,1,1]=(0.248218424051)*x_ref[0]**o+(0.0980396576111)*x_ref[0]+(0.413322166489)*x_ref[1]**o+(0.429747261346)*x_ref[1]
        ref[2,3,2,0]=(0.742987565733)*x_ref[0]**o+(-0.213736450136)*x_ref[0]+(0.996975741314)*x_ref[1]**o+(0.960709958665)*x_ref[1]
        ref[2,3,2,1]=(0.436149490977)*x_ref[0]**o+(0.87632181197)*x_ref[0]+(-0.881934669912)*x_ref[1]**o+(-0.457520504175)*x_ref[1]
        ref[2,4,0,0]=(-0.519014330693)*x_ref[0]**o+(0.938619608407)*x_ref[0]+(-0.847954563609)*x_ref[1]**o+(0.791774379621)*x_ref[1]
        ref[2,4,0,1]=(-0.599487299453)*x_ref[0]**o+(0.977867555132)*x_ref[0]+(0.693105630579)*x_ref[1]**o+(-0.688827812942)*x_ref[1]
        ref[2,4,1,0]=(0.955907964163)*x_ref[0]**o+(0.0724056986172)*x_ref[0]+(0.254773619772)*x_ref[1]**o+(0.0428861146964)*x_ref[1]
        ref[2,4,1,1]=(-0.422202101998)*x_ref[0]**o+(0.803263223544)*x_ref[0]+(0.456048140785)*x_ref[1]**o+(0.0419133126183)*x_ref[1]
        ref[2,4,2,0]=(-0.182308704781)*x_ref[0]**o+(0.862282443488)*x_ref[0]+(-0.844027007955)*x_ref[1]**o+(0.586220526376)*x_ref[1]
        ref[2,4,2,1]=(-0.801542666229)*x_ref[0]**o+(0.903685299455)*x_ref[0]+(0.428926544377)*x_ref[1]**o+(-0.891220471549)*x_ref[1]
        ref[3,0,0,0]=(0.602917787713)*x_ref[0]**o+(-0.592239549743)*x_ref[0]+(0.761830006718)*x_ref[1]**o+(-0.784264167777)*x_ref[1]
        ref[3,0,0,1]=(0.819962957732)*x_ref[0]**o+(-0.455126790372)*x_ref[0]+(-0.144898728525)*x_ref[1]**o+(0.121566345339)*x_ref[1]
        ref[3,0,1,0]=(0.570694941559)*x_ref[0]**o+(-0.505678656386)*x_ref[0]+(0.295536587355)*x_ref[1]**o+(-0.852779737814)*x_ref[1]
        ref[3,0,1,1]=(-0.226246858753)*x_ref[0]**o+(-0.513900306228)*x_ref[0]+(-0.389483591407)*x_ref[1]**o+(-0.735105335539)*x_ref[1]
        ref[3,0,2,0]=(0.941877610938)*x_ref[0]**o+(0.93808330298)*x_ref[0]+(0.189785162182)*x_ref[1]**o+(0.546525366304)*x_ref[1]
        ref[3,0,2,1]=(0.10639100485)*x_ref[0]**o+(-0.773361007972)*x_ref[0]+(-0.470531427579)*x_ref[1]**o+(0.467359342808)*x_ref[1]
        ref[3,1,0,0]=(0.857725599589)*x_ref[0]**o+(0.0881746413132)*x_ref[0]+(0.040713151152)*x_ref[1]**o+(-0.52070729883)*x_ref[1]
        ref[3,1,0,1]=(0.266807652146)*x_ref[0]**o+(0.337496673892)*x_ref[0]+(-0.562266297445)*x_ref[1]**o+(0.326015195113)*x_ref[1]
        ref[3,1,1,0]=(0.177275509465)*x_ref[0]**o+(-0.647855761301)*x_ref[0]+(-0.0963373631995)*x_ref[1]**o+(-0.535256252994)*x_ref[1]
        ref[3,1,1,1]=(0.430309295606)*x_ref[0]**o+(-0.0894749557428)*x_ref[0]+(0.503470724483)*x_ref[1]**o+(-0.202361122248)*x_ref[1]
        ref[3,1,2,0]=(0.283995014194)*x_ref[0]**o+(-0.544419907194)*x_ref[0]+(0.78478571712)*x_ref[1]**o+(0.341540664147)*x_ref[1]
        ref[3,1,2,1]=(-0.281042869936)*x_ref[0]**o+(0.728771766492)*x_ref[0]+(-0.298974388557)*x_ref[1]**o+(-0.836789906985)*x_ref[1]
        ref[3,2,0,0]=(-0.415323371561)*x_ref[0]**o+(0.583971137371)*x_ref[0]+(-0.265658679489)*x_ref[1]**o+(-0.432447373809)*x_ref[1]
        ref[3,2,0,1]=(0.690486298855)*x_ref[0]**o+(0.552521632226)*x_ref[0]+(0.237622739042)*x_ref[1]**o+(-0.254513693924)*x_ref[1]
        ref[3,2,1,0]=(0.453985687207)*x_ref[0]**o+(-0.785385580962)*x_ref[0]+(0.0190987841634)*x_ref[1]**o+(-0.7923288861)*x_ref[1]
        ref[3,2,1,1]=(-0.964082656379)*x_ref[0]**o+(-0.806315937371)*x_ref[0]+(0.271145420366)*x_ref[1]**o+(-0.839014535945)*x_ref[1]
        ref[3,2,2,0]=(-0.372062773997)*x_ref[0]**o+(-0.069749560567)*x_ref[0]+(-0.126310395383)*x_ref[1]**o+(0.972025865285)*x_ref[1]
        ref[3,2,2,1]=(-0.580353020086)*x_ref[0]**o+(-0.152746821477)*x_ref[0]+(0.297553660535)*x_ref[1]**o+(0.444791568145)*x_ref[1]
        ref[3,3,0,0]=(-0.361520029439)*x_ref[0]**o+(-0.810391070499)*x_ref[0]+(0.462123882642)*x_ref[1]**o+(-0.974534227742)*x_ref[1]
        ref[3,3,0,1]=(0.113370617837)*x_ref[0]**o+(0.428257355435)*x_ref[0]+(-0.18656623758)*x_ref[1]**o+(0.206503327394)*x_ref[1]
        ref[3,3,1,0]=(0.341328445689)*x_ref[0]**o+(0.34002947054)*x_ref[0]+(0.997677286362)*x_ref[1]**o+(0.43300073225)*x_ref[1]
        ref[3,3,1,1]=(-0.513912982823)*x_ref[0]**o+(0.826573827064)*x_ref[0]+(-0.701417864085)*x_ref[1]**o+(0.567078883219)*x_ref[1]
        ref[3,3,2,0]=(-0.325536866965)*x_ref[0]**o+(0.924405767451)*x_ref[0]+(-0.0481381954346)*x_ref[1]**o+(0.912210617073)*x_ref[1]
        ref[3,3,2,1]=(0.491532434302)*x_ref[0]**o+(-0.584980848036)*x_ref[0]+(-0.788317530858)*x_ref[1]**o+(-0.900681893172)*x_ref[1]
        ref[3,4,0,0]=(0.432249855148)*x_ref[0]**o+(-0.673369312552)*x_ref[0]+(0.244704845849)*x_ref[1]**o+(-0.300995595367)*x_ref[1]
        ref[3,4,0,1]=(0.383037211984)*x_ref[0]**o+(0.299711400488)*x_ref[0]+(0.870584255454)*x_ref[1]**o+(-0.33571823329)*x_ref[1]
        ref[3,4,1,0]=(0.737134180177)*x_ref[0]**o+(-0.101307090628)*x_ref[0]+(-0.970632061143)*x_ref[1]**o+(-0.449042684137)*x_ref[1]
        ref[3,4,1,1]=(0.00684250643632)*x_ref[0]**o+(0.676964654022)*x_ref[0]+(0.789069834858)*x_ref[1]**o+(-0.829739508197)*x_ref[1]
        ref[3,4,2,0]=(0.7279044944)*x_ref[0]**o+(0.215711122038)*x_ref[0]+(0.385631252632)*x_ref[1]**o+(0.629390433529)*x_ref[1]
        ref[3,4,2,1]=(-0.380750716629)*x_ref[0]**o+(-0.413040692437)*x_ref[0]+(0.26795244639)*x_ref[1]**o+(0.492475504659)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.287160004495)*x[0]**o+(0.35112250158)*x[0]+(0.061693427026)*x[1]**o+(-0.752505628029)*x[1]+(-0.57137756028)*x[2]**o+(-0.0397598892504)*x[2]
        arg[0,0,0,1]=(0.272799584607)*x[0]**o+(-0.569947956933)*x[0]+(0.322189261657)*x[1]**o+(-0.440552272224)*x[1]+(-0.10460241481)*x[2]**o+(-0.884848984752)*x[2]
        arg[0,0,1,0]=(0.147766301604)*x[0]**o+(0.405651654428)*x[0]+(-0.440827209529)*x[1]**o+(-0.399952864271)*x[1]+(-0.770191009459)*x[2]**o+(-0.789853968049)*x[2]
        arg[0,0,1,1]=(0.688053165375)*x[0]**o+(-0.154890421924)*x[0]+(0.583684164319)*x[1]**o+(-0.398082068217)*x[1]+(-0.573142923099)*x[2]**o+(-0.205271877337)*x[2]
        arg[0,0,2,0]=(0.807043275877)*x[0]**o+(0.843046383231)*x[0]+(-0.588146471402)*x[1]**o+(0.157787304557)*x[1]+(-0.771009748568)*x[2]**o+(0.162090791609)*x[2]
        arg[0,0,2,1]=(-0.0706800489392)*x[0]**o+(0.499961344665)*x[0]+(0.208926178797)*x[1]**o+(-0.7462472769)*x[1]+(-0.105475405648)*x[2]**o+(-0.18670229948)*x[2]
        arg[0,1,0,0]=(-0.696657281943)*x[0]**o+(-0.0251785345759)*x[0]+(-0.579575530451)*x[1]**o+(-0.0628378665725)*x[1]+(0.668395441302)*x[2]**o+(0.813478105092)*x[2]
        arg[0,1,0,1]=(-0.574784314876)*x[0]**o+(-0.321268045204)*x[0]+(0.160704464213)*x[1]**o+(-0.676016748103)*x[1]+(0.918971102865)*x[2]**o+(0.252082410081)*x[2]
        arg[0,1,1,0]=(-0.905062846817)*x[0]**o+(0.788906658477)*x[0]+(-0.408807611936)*x[1]**o+(0.114686186437)*x[1]+(-0.654685568493)*x[2]**o+(0.0879202139254)*x[2]
        arg[0,1,1,1]=(-0.68020231443)*x[0]**o+(-0.82049202902)*x[0]+(-0.958352740187)*x[1]**o+(-0.33201745802)*x[1]+(0.539810363682)*x[2]**o+(0.744988573087)*x[2]
        arg[0,1,2,0]=(-0.960884744997)*x[0]**o+(-0.704157132503)*x[0]+(0.282236905414)*x[1]**o+(0.756363456433)*x[1]+(-0.365400101792)*x[2]**o+(0.975735846554)*x[2]
        arg[0,1,2,1]=(-0.691780076951)*x[0]**o+(-0.952575440059)*x[0]+(-0.921247808098)*x[1]**o+(-0.88872834095)*x[1]+(0.534103013833)*x[2]**o+(-0.602295186104)*x[2]
        arg[0,2,0,0]=(0.460698415704)*x[0]**o+(0.770243814897)*x[0]+(-0.380381727574)*x[1]**o+(0.476360845668)*x[1]+(-0.681040425102)*x[2]**o+(-0.244369435116)*x[2]
        arg[0,2,0,1]=(0.251591652281)*x[0]**o+(0.748234919686)*x[0]+(-0.698189037307)*x[1]**o+(-0.292400412508)*x[1]+(0.613850696577)*x[2]**o+(-0.942243226834)*x[2]
        arg[0,2,1,0]=(-0.718090066328)*x[0]**o+(-0.322499511794)*x[0]+(-0.75105623397)*x[1]**o+(0.526680271832)*x[1]+(-0.186671531793)*x[2]**o+(-0.449273875991)*x[2]
        arg[0,2,1,1]=(-0.52165552452)*x[0]**o+(0.902446006085)*x[0]+(-0.915575172702)*x[1]**o+(-0.150816194662)*x[1]+(0.596782785689)*x[2]**o+(0.546880865642)*x[2]
        arg[0,2,2,0]=(-0.104486704816)*x[0]**o+(0.723863699457)*x[0]+(0.758622117279)*x[1]**o+(0.372271580145)*x[1]+(-0.0820570383792)*x[2]**o+(-0.63539781104)*x[2]
        arg[0,2,2,1]=(0.0355930823603)*x[0]**o+(0.0931942600432)*x[0]+(-0.118096985978)*x[1]**o+(0.296983782212)*x[1]+(-0.864491763712)*x[2]**o+(0.942230663978)*x[2]
        arg[0,3,0,0]=(0.371153709815)*x[0]**o+(0.0427309083893)*x[0]+(-0.886563919084)*x[1]**o+(-0.14946513404)*x[1]+(0.80903804519)*x[2]**o+(0.69728533073)*x[2]
        arg[0,3,0,1]=(0.0593930396142)*x[0]**o+(-0.370327974089)*x[0]+(-0.388354464683)*x[1]**o+(-0.367648771812)*x[1]+(-0.573754821458)*x[2]**o+(-0.947499155479)*x[2]
        arg[0,3,1,0]=(-0.869817317)*x[0]**o+(0.874317772678)*x[0]+(-0.870060178375)*x[1]**o+(0.66571074524)*x[1]+(-0.669052377378)*x[2]**o+(0.940212680575)*x[2]
        arg[0,3,1,1]=(-0.417445606772)*x[0]**o+(0.942606057619)*x[0]+(-0.133936034409)*x[1]**o+(-0.363274280866)*x[1]+(0.83487439935)*x[2]**o+(-0.108398981461)*x[2]
        arg[0,3,2,0]=(0.363623383293)*x[0]**o+(0.624267309217)*x[0]+(0.801719984899)*x[1]**o+(0.189212370122)*x[1]+(0.464633630928)*x[2]**o+(-0.979808318438)*x[2]
        arg[0,3,2,1]=(0.85384408805)*x[0]**o+(-0.889214754116)*x[0]+(0.164915391224)*x[1]**o+(0.0168991405201)*x[1]+(0.097171994236)*x[2]**o+(-0.146094197339)*x[2]
        arg[0,4,0,0]=(0.778356158972)*x[0]**o+(0.910241265149)*x[0]+(0.123690863272)*x[1]**o+(0.87802121779)*x[1]+(-0.714615668595)*x[2]**o+(-0.437472831583)*x[2]
        arg[0,4,0,1]=(-0.157185481783)*x[0]**o+(0.352802163968)*x[0]+(0.390184123612)*x[1]**o+(-0.680415093098)*x[1]+(-0.121195042454)*x[2]**o+(-0.713693712117)*x[2]
        arg[0,4,1,0]=(0.778115668646)*x[0]**o+(-0.111729042892)*x[0]+(0.451518002735)*x[1]**o+(0.232731183374)*x[1]+(-0.764246158086)*x[2]**o+(-0.795525155328)*x[2]
        arg[0,4,1,1]=(-0.880652367027)*x[0]**o+(-0.314966079126)*x[0]+(-0.640720373065)*x[1]**o+(0.336230541901)*x[1]+(0.566956999526)*x[2]**o+(0.74098258811)*x[2]
        arg[0,4,2,0]=(0.908657274301)*x[0]**o+(0.636427823614)*x[0]+(0.650137414386)*x[1]**o+(0.695558379005)*x[1]+(-0.489852401087)*x[2]**o+(-0.479227108044)*x[2]
        arg[0,4,2,1]=(0.015734241525)*x[0]**o+(0.0719884727088)*x[0]+(0.811864895261)*x[1]**o+(0.959127986784)*x[1]+(-0.238146663368)*x[2]**o+(-0.454939322571)*x[2]
        arg[1,0,0,0]=(-0.918218246559)*x[0]**o+(-0.662787169269)*x[0]+(-0.954035144111)*x[1]**o+(-0.647264685228)*x[1]+(-0.834085960213)*x[2]**o+(-0.167541951024)*x[2]
        arg[1,0,0,1]=(-0.670950869382)*x[0]**o+(0.826472159989)*x[0]+(0.180410872786)*x[1]**o+(-0.715162871594)*x[1]+(0.653287853169)*x[2]**o+(0.892060529625)*x[2]
        arg[1,0,1,0]=(0.398789786407)*x[0]**o+(0.689408033554)*x[0]+(0.124338245128)*x[1]**o+(0.748280808587)*x[1]+(0.418357122274)*x[2]**o+(0.230707440075)*x[2]
        arg[1,0,1,1]=(-0.912778964055)*x[0]**o+(0.904288708694)*x[0]+(-0.874744856901)*x[1]**o+(0.550162195719)*x[1]+(-0.7526423373)*x[2]**o+(0.820993231654)*x[2]
        arg[1,0,2,0]=(-0.843459684648)*x[0]**o+(-0.579708764091)*x[0]+(-0.876072131849)*x[1]**o+(0.933694599696)*x[1]+(0.507781209021)*x[2]**o+(-0.719729482133)*x[2]
        arg[1,0,2,1]=(0.594566925796)*x[0]**o+(-0.994301292358)*x[0]+(0.808114389564)*x[1]**o+(-0.970803712864)*x[1]+(0.433544910389)*x[2]**o+(0.260659440299)*x[2]
        arg[1,1,0,0]=(0.997377553172)*x[0]**o+(0.458428616383)*x[0]+(0.357280778493)*x[1]**o+(0.349837110741)*x[1]+(-0.709592214445)*x[2]**o+(-0.630000739033)*x[2]
        arg[1,1,0,1]=(0.121595650501)*x[0]**o+(0.40870120924)*x[0]+(0.576917123807)*x[1]**o+(-0.932268308325)*x[1]+(0.710043472116)*x[2]**o+(-0.716528412077)*x[2]
        arg[1,1,1,0]=(-0.151416035033)*x[0]**o+(0.0255347649787)*x[0]+(-0.752987701889)*x[1]**o+(-0.480969875813)*x[1]+(-0.610514504179)*x[2]**o+(0.959274757402)*x[2]
        arg[1,1,1,1]=(0.76695706106)*x[0]**o+(-0.0750965975512)*x[0]+(0.846833900735)*x[1]**o+(-0.105838465439)*x[1]+(0.541027486457)*x[2]**o+(0.152543017589)*x[2]
        arg[1,1,2,0]=(0.0856977181865)*x[0]**o+(-0.188139369706)*x[0]+(0.936071565289)*x[1]**o+(-0.3958713036)*x[1]+(-0.350945878893)*x[2]**o+(0.382302674883)*x[2]
        arg[1,1,2,1]=(0.873054234789)*x[0]**o+(-0.490869716146)*x[0]+(0.46720526212)*x[1]**o+(-0.334203412334)*x[1]+(-0.81140895636)*x[2]**o+(-0.973858971749)*x[2]
        arg[1,2,0,0]=(-0.716676352443)*x[0]**o+(0.659821396277)*x[0]+(0.335964292954)*x[1]**o+(0.11609664773)*x[1]+(-0.889186749466)*x[2]**o+(0.454231973946)*x[2]
        arg[1,2,0,1]=(0.367048872648)*x[0]**o+(0.276083713126)*x[0]+(0.686639128043)*x[1]**o+(-0.991414415293)*x[1]+(0.25360515964)*x[2]**o+(-0.377623315093)*x[2]
        arg[1,2,1,0]=(-0.723089358777)*x[0]**o+(-0.516690112406)*x[0]+(0.262334273979)*x[1]**o+(0.140773502537)*x[1]+(0.447333972396)*x[2]**o+(-0.551395692966)*x[2]
        arg[1,2,1,1]=(0.738791182491)*x[0]**o+(-0.978120626668)*x[0]+(-0.217970433713)*x[1]**o+(0.0181011067492)*x[1]+(0.756322867523)*x[2]**o+(0.495269195978)*x[2]
        arg[1,2,2,0]=(-0.630506449854)*x[0]**o+(-0.0865482364231)*x[0]+(-0.415896403157)*x[1]**o+(-0.712833149895)*x[1]+(-0.180482789777)*x[2]**o+(0.00788290661235)*x[2]
        arg[1,2,2,1]=(0.0946913865377)*x[0]**o+(0.752854827716)*x[0]+(0.257846220843)*x[1]**o+(-0.953926983578)*x[1]+(-0.388505920682)*x[2]**o+(-0.10751401251)*x[2]
        arg[1,3,0,0]=(0.694014030611)*x[0]**o+(0.0433103547797)*x[0]+(-0.0805025225569)*x[1]**o+(0.389052254754)*x[1]+(-0.813210939614)*x[2]**o+(-0.593788787896)*x[2]
        arg[1,3,0,1]=(0.0101606353406)*x[0]**o+(-0.849874258855)*x[0]+(-0.0578973959381)*x[1]**o+(0.942113140424)*x[1]+(0.613716459216)*x[2]**o+(0.72631176833)*x[2]
        arg[1,3,1,0]=(0.264703608811)*x[0]**o+(0.457962548326)*x[0]+(-0.433795242673)*x[1]**o+(0.650995712848)*x[1]+(0.372471807967)*x[2]**o+(0.00416636841047)*x[2]
        arg[1,3,1,1]=(-0.73686451669)*x[0]**o+(-0.0425614910776)*x[0]+(-0.587820431311)*x[1]**o+(-0.377628617833)*x[1]+(-0.463964014406)*x[2]**o+(0.283273874989)*x[2]
        arg[1,3,2,0]=(-0.689341439143)*x[0]**o+(-0.675861918889)*x[0]+(-0.918622662352)*x[1]**o+(0.740990261888)*x[1]+(0.0454007677041)*x[2]**o+(0.513623361074)*x[2]
        arg[1,3,2,1]=(0.13229101848)*x[0]**o+(0.171873799817)*x[0]+(-0.154491174434)*x[1]**o+(0.897976884089)*x[1]+(-0.598886311821)*x[2]**o+(0.738696710918)*x[2]
        arg[1,4,0,0]=(0.59403636966)*x[0]**o+(0.93856848969)*x[0]+(0.80164849167)*x[1]**o+(-0.753322979568)*x[1]+(-0.107098073916)*x[2]**o+(-0.623047487973)*x[2]
        arg[1,4,0,1]=(0.291046099082)*x[0]**o+(-0.437982471049)*x[0]+(0.017395762947)*x[1]**o+(0.852440593923)*x[1]+(0.630825424359)*x[2]**o+(0.419347491727)*x[2]
        arg[1,4,1,0]=(-0.747962012701)*x[0]**o+(-0.11622032094)*x[0]+(0.324950549781)*x[1]**o+(0.950911398327)*x[1]+(-0.18098990205)*x[2]**o+(-0.676060633522)*x[2]
        arg[1,4,1,1]=(0.545224363128)*x[0]**o+(0.753376367121)*x[0]+(0.271426695753)*x[1]**o+(-0.139309844675)*x[1]+(-0.923434716375)*x[2]**o+(-0.985988324389)*x[2]
        arg[1,4,2,0]=(-0.532260822047)*x[0]**o+(0.901407135724)*x[0]+(0.937451330141)*x[1]**o+(0.285217666559)*x[1]+(-0.133028984824)*x[2]**o+(0.993965133696)*x[2]
        arg[1,4,2,1]=(0.2231815607)*x[0]**o+(0.994845537582)*x[0]+(0.0368727378195)*x[1]**o+(0.511772933777)*x[1]+(-0.871580215976)*x[2]**o+(-0.587010057329)*x[2]
        arg[2,0,0,0]=(-0.55149411913)*x[0]**o+(-0.905541278559)*x[0]+(0.461217373096)*x[1]**o+(-0.53359853241)*x[1]+(-0.666643290883)*x[2]**o+(0.151495826926)*x[2]
        arg[2,0,0,1]=(-0.980485062148)*x[0]**o+(0.150165616162)*x[0]+(0.59165725487)*x[1]**o+(-0.243853536017)*x[1]+(0.66816261686)*x[2]**o+(0.2425696131)*x[2]
        arg[2,0,1,0]=(-0.509337213758)*x[0]**o+(0.652936706687)*x[0]+(-0.849214338091)*x[1]**o+(-0.523354123549)*x[1]+(0.371515689887)*x[2]**o+(0.799989815414)*x[2]
        arg[2,0,1,1]=(0.461584544541)*x[0]**o+(0.662931355954)*x[0]+(0.744415634786)*x[1]**o+(-0.016743801256)*x[1]+(0.172945846325)*x[2]**o+(-0.0889827089341)*x[2]
        arg[2,0,2,0]=(0.506514791562)*x[0]**o+(0.974477160694)*x[0]+(-0.968485343167)*x[1]**o+(0.427480013599)*x[1]+(0.0460755879358)*x[2]**o+(0.186500994948)*x[2]
        arg[2,0,2,1]=(0.259831607993)*x[0]**o+(-0.0141290444125)*x[0]+(-0.968452177682)*x[1]**o+(0.635619125009)*x[1]+(-0.958848431827)*x[2]**o+(-0.0977241523308)*x[2]
        arg[2,1,0,0]=(0.204233222933)*x[0]**o+(0.834739381877)*x[0]+(0.19961185048)*x[1]**o+(-0.238470380578)*x[1]+(-0.749979713773)*x[2]**o+(-0.268168886251)*x[2]
        arg[2,1,0,1]=(0.998123643088)*x[0]**o+(0.635823733041)*x[0]+(-0.265447124419)*x[1]**o+(-0.225582751946)*x[1]+(0.91705937439)*x[2]**o+(-0.739461510792)*x[2]
        arg[2,1,1,0]=(0.0800715288646)*x[0]**o+(-0.37677196659)*x[0]+(-0.0737286350219)*x[1]**o+(0.032076062933)*x[1]+(0.245421912709)*x[2]**o+(-0.8378023935)*x[2]
        arg[2,1,1,1]=(-0.572376055603)*x[0]**o+(0.8675468469)*x[0]+(-0.533012889867)*x[1]**o+(-0.247215658297)*x[1]+(0.673144564693)*x[2]**o+(-0.749136108985)*x[2]
        arg[2,1,2,0]=(-0.661188192488)*x[0]**o+(-0.924435593008)*x[0]+(-0.791845808916)*x[1]**o+(0.00133765351092)*x[1]+(0.233676862873)*x[2]**o+(0.981812737962)*x[2]
        arg[2,1,2,1]=(0.332072445986)*x[0]**o+(-0.909373798588)*x[0]+(0.927025647819)*x[1]**o+(-0.936348942617)*x[1]+(0.251508592152)*x[2]**o+(-0.468147973987)*x[2]
        arg[2,2,0,0]=(0.892377893841)*x[0]**o+(-0.147933293091)*x[0]+(0.950491102076)*x[1]**o+(0.595250355926)*x[1]+(0.0169764649992)*x[2]**o+(-0.683137752586)*x[2]
        arg[2,2,0,1]=(0.00579397005645)*x[0]**o+(0.627197373917)*x[0]+(0.663574756196)*x[1]**o+(0.934106767513)*x[1]+(0.105844878022)*x[2]**o+(0.513322917688)*x[2]
        arg[2,2,1,0]=(-0.168776016133)*x[0]**o+(-0.940071612944)*x[0]+(0.0366053554566)*x[1]**o+(-0.628010909771)*x[1]+(0.250023045906)*x[2]**o+(-0.00327322332166)*x[2]
        arg[2,2,1,1]=(0.904155764376)*x[0]**o+(-0.258018307306)*x[0]+(0.65862820521)*x[1]**o+(0.434480798515)*x[1]+(0.4089419173)*x[2]**o+(-0.518032695598)*x[2]
        arg[2,2,2,0]=(-0.41470070958)*x[0]**o+(-0.783186751952)*x[0]+(0.548309561942)*x[1]**o+(0.837130592803)*x[1]+(-0.346808491848)*x[2]**o+(0.165416040985)*x[2]
        arg[2,2,2,1]=(0.997845909195)*x[0]**o+(0.448428042158)*x[0]+(-0.657050004311)*x[1]**o+(0.493104851747)*x[1]+(0.558315071197)*x[2]**o+(0.206659432923)*x[2]
        arg[2,3,0,0]=(0.433874360232)*x[0]**o+(-0.0635646197131)*x[0]+(-0.521690521523)*x[1]**o+(-0.298130684888)*x[1]+(-0.131905313644)*x[2]**o+(0.64389451112)*x[2]
        arg[2,3,0,1]=(0.457729315216)*x[0]**o+(-0.419323242135)*x[0]+(0.614544051135)*x[1]**o+(-0.701671999275)*x[1]+(0.905494300727)*x[2]**o+(-0.565716985337)*x[2]
        arg[2,3,1,0]=(0.256570728097)*x[0]**o+(0.866349285103)*x[0]+(-0.115768354896)*x[1]**o+(-0.613044456549)*x[1]+(-0.20074185156)*x[2]**o+(-0.188403571776)*x[2]
        arg[2,3,1,1]=(0.767479943414)*x[0]**o+(0.884066215007)*x[0]+(0.363855420056)*x[1]**o+(0.919698981104)*x[1]+(0.225747762317)*x[2]**o+(-0.220362271104)*x[2]
        arg[2,3,2,0]=(-0.746469501382)*x[0]**o+(0.680520398001)*x[0]+(0.926950996325)*x[1]**o+(-0.122831303044)*x[1]+(-0.27226561675)*x[2]**o+(0.933304087951)*x[2]
        arg[2,3,2,1]=(-0.0961434747216)*x[0]**o+(0.987660516767)*x[0]+(0.288175205045)*x[1]**o+(0.683720820403)*x[1]+(0.354403143642)*x[2]**o+(-0.745694089132)*x[2]
        arg[2,4,0,0]=(0.921750835986)*x[0]**o+(-0.0358861598394)*x[0]+(0.190129645259)*x[1]**o+(0.377067803331)*x[1]+(-0.0188615236582)*x[2]**o+(0.804533888824)*x[2]
        arg[2,4,0,1]=(-0.256132234484)*x[0]**o+(-0.102661411017)*x[0]+(-0.903465157009)*x[1]**o+(0.50133310168)*x[1]+(-0.0889102598619)*x[2]**o+(-0.0310174238365)*x[2]
        arg[2,4,1,0]=(0.58483798855)*x[0]**o+(0.967943344305)*x[0]+(-0.0701080005438)*x[1]**o+(0.489476104406)*x[1]+(0.615308427038)*x[2]**o+(-0.38065505709)*x[2]
        arg[2,4,1,1]=(0.647734130724)*x[0]**o+(-0.930934392719)*x[0]+(0.267370540491)*x[1]**o+(-0.0175019424556)*x[1]+(0.38593657129)*x[2]**o+(-0.985922292359)*x[2]
        arg[2,4,2,0]=(-0.289971479933)*x[0]**o+(0.882014313157)*x[0]+(0.147321067579)*x[1]**o+(-0.176771838489)*x[1]+(0.710916531544)*x[2]**o+(-0.0856290170401)*x[2]
        arg[2,4,2,1]=(-0.465930997733)*x[0]**o+(-0.305563942094)*x[0]+(-0.824535262061)*x[1]**o+(-0.917892226972)*x[1]+(0.624462871955)*x[2]**o+(0.544736500065)*x[2]
        arg[3,0,0,0]=(-0.673230815097)*x[0]**o+(-0.726292712917)*x[0]+(0.0531037021688)*x[1]**o+(-0.467219030159)*x[1]+(-0.0729861369718)*x[2]**o+(-0.440394335015)*x[2]
        arg[3,0,0,1]=(-0.358235295558)*x[0]**o+(-0.301184713671)*x[0]+(-0.496030649086)*x[1]**o+(-0.125922184777)*x[1]+(0.946715036973)*x[2]**o+(0.846370337893)*x[2]
        arg[3,0,1,0]=(-0.394175109612)*x[0]**o+(-0.85137582307)*x[0]+(-0.667804785044)*x[1]**o+(0.156762109891)*x[1]+(-0.616444700525)*x[2]**o+(0.0738729399356)*x[2]
        arg[3,0,1,1]=(-0.652875462534)*x[0]**o+(-0.0605650192541)*x[0]+(0.840539933823)*x[1]**o+(-0.422819423301)*x[1]+(0.113440226819)*x[2]**o+(-0.915154532306)*x[2]
        arg[3,0,2,0]=(-0.162863352328)*x[0]**o+(-0.364461332292)*x[0]+(-0.222181741398)*x[1]**o+(-0.563668008365)*x[1]+(0.118491770105)*x[2]**o+(0.0809155053022)*x[2]
        arg[3,0,2,1]=(-0.985278929437)*x[0]**o+(-0.919440760157)*x[0]+(-0.776910686)*x[1]**o+(-0.884270488943)*x[1]+(0.984279010363)*x[2]**o+(-0.489503278582)*x[2]
        arg[3,1,0,0]=(-0.806565449823)*x[0]**o+(-0.0851385375281)*x[0]+(-0.88851381773)*x[1]**o+(-0.666539867528)*x[1]+(0.154825531411)*x[2]**o+(0.764069849144)*x[2]
        arg[3,1,0,1]=(-0.222125783918)*x[0]**o+(0.568035675467)*x[0]+(0.470171778932)*x[1]**o+(-0.307608710881)*x[1]+(-0.836083742471)*x[2]**o+(0.553597553381)*x[2]
        arg[3,1,1,0]=(0.431237389847)*x[0]**o+(-0.141391012471)*x[0]+(0.726895025519)*x[1]**o+(-0.527868751679)*x[1]+(0.711761083137)*x[2]**o+(0.996992864103)*x[2]
        arg[3,1,1,1]=(-0.123885959924)*x[0]**o+(0.877034475529)*x[0]+(-0.0432492682928)*x[1]**o+(-0.370315494618)*x[1]+(0.243139719071)*x[2]**o+(-0.684318634902)*x[2]
        arg[3,1,2,0]=(-0.477214489355)*x[0]**o+(0.340808576619)*x[0]+(-0.0991802306558)*x[1]**o+(0.462599821908)*x[1]+(0.091768281564)*x[2]**o+(0.0315552838839)*x[2]
        arg[3,1,2,1]=(0.676189142894)*x[0]**o+(0.510136386913)*x[0]+(-0.558579041952)*x[1]**o+(0.961615047877)*x[1]+(0.510717386124)*x[2]**o+(0.774430183432)*x[2]
        arg[3,2,0,0]=(0.158370232444)*x[0]**o+(0.209424855992)*x[0]+(0.691438753013)*x[1]**o+(-0.781594298738)*x[1]+(-0.0188536627517)*x[2]**o+(0.136400798995)*x[2]
        arg[3,2,0,1]=(-0.440674414328)*x[0]**o+(0.118228612687)*x[0]+(-0.700026492613)*x[1]**o+(-0.981341215714)*x[1]+(-0.172847519578)*x[2]**o+(-0.423095202172)*x[2]
        arg[3,2,1,0]=(-0.0694960391695)*x[0]**o+(-0.66448610135)*x[0]+(-0.177572703748)*x[1]**o+(-0.060488161727)*x[1]+(0.306021853885)*x[2]**o+(0.0730338716187)*x[2]
        arg[3,2,1,1]=(-0.360662154547)*x[0]**o+(0.577679359648)*x[0]+(-0.515313534966)*x[1]**o+(0.217352687358)*x[1]+(0.728367577882)*x[2]**o+(0.38505849834)*x[2]
        arg[3,2,2,0]=(0.568671861232)*x[0]**o+(-0.0792490917882)*x[0]+(0.0188002885229)*x[1]**o+(-0.956862575759)*x[1]+(-0.844332442983)*x[2]**o+(0.670792928042)*x[2]
        arg[3,2,2,1]=(0.833099204713)*x[0]**o+(0.368482169933)*x[0]+(-0.926002498892)*x[1]**o+(-0.130646446568)*x[1]+(0.81164928914)*x[2]**o+(-0.641079919927)*x[2]
        arg[3,3,0,0]=(-0.277769350989)*x[0]**o+(0.753452018649)*x[0]+(-0.766322344505)*x[1]**o+(-0.679108224612)*x[1]+(0.0799499795424)*x[2]**o+(0.255751308975)*x[2]
        arg[3,3,0,1]=(0.429176753733)*x[0]**o+(0.123106050259)*x[0]+(-0.169037728992)*x[1]**o+(0.749245933687)*x[1]+(0.560140859355)*x[2]**o+(-0.937091840021)*x[2]
        arg[3,3,1,0]=(0.648792557593)*x[0]**o+(-0.057555187513)*x[0]+(0.883290331726)*x[1]**o+(0.539276905995)*x[1]+(0.473614408583)*x[2]**o+(0.723258340333)*x[2]
        arg[3,3,1,1]=(0.527298833572)*x[0]**o+(0.0753824058279)*x[0]+(0.719196099257)*x[1]**o+(-0.0124603350035)*x[1]+(-0.964484650253)*x[2]**o+(0.404444269694)*x[2]
        arg[3,3,2,0]=(-0.403640677462)*x[0]**o+(0.906195964766)*x[0]+(-0.561419078404)*x[1]**o+(0.645408717838)*x[1]+(0.671624799609)*x[2]**o+(-0.573901868621)*x[2]
        arg[3,3,2,1]=(-0.288117182165)*x[0]**o+(-0.391003584945)*x[0]+(-0.555037182038)*x[1]**o+(0.299362209017)*x[1]+(-0.552429724326)*x[2]**o+(-0.416923306774)*x[2]
        arg[3,4,0,0]=(-0.19594719192)*x[0]**o+(0.150898934692)*x[0]+(0.210935232044)*x[1]**o+(-0.319764293347)*x[1]+(-0.642450751195)*x[2]**o+(0.22231085168)*x[2]
        arg[3,4,0,1]=(-0.0514589460225)*x[0]**o+(0.0366025646754)*x[0]+(0.416280576691)*x[1]**o+(0.609178063951)*x[1]+(0.835611677062)*x[2]**o+(-0.131630571478)*x[2]
        arg[3,4,1,0]=(0.171509152214)*x[0]**o+(-0.761188271093)*x[0]+(-0.915951802896)*x[1]**o+(-0.451796709915)*x[1]+(-0.979847828162)*x[2]**o+(0.0566453393315)*x[2]
        arg[3,4,1,1]=(0.690427998102)*x[0]**o+(-0.358318625695)*x[0]+(0.330270871613)*x[1]**o+(-0.724392522137)*x[1]+(-0.535474623489)*x[2]**o+(-0.12341743885)*x[2]
        arg[3,4,2,0]=(0.488090753426)*x[0]**o+(-0.352417078372)*x[0]+(0.971988099319)*x[1]**o+(-0.288929516009)*x[1]+(0.612576734094)*x[2]**o+(-0.43799514603)*x[2]
        arg[3,4,2,1]=(-0.098768101356)*x[0]**o+(0.123300661895)*x[0]+(-0.0489724280789)*x[1]**o+(-0.422508669137)*x[1]+(-0.861483104139)*x[2]**o+(0.950198316001)*x[2]
        ref[0,0,0,0]=(0.287160004495)*x_ref[0]**o+(0.35112250158)*x_ref[0]+(0.061693427026)*x_ref[1]**o+(-0.752505628029)*x_ref[1]+(-0.57137756028)*x_ref[2]**o+(-0.0397598892504)*x_ref[2]
        ref[0,0,0,1]=(0.272799584607)*x_ref[0]**o+(-0.569947956933)*x_ref[0]+(0.322189261657)*x_ref[1]**o+(-0.440552272224)*x_ref[1]+(-0.10460241481)*x_ref[2]**o+(-0.884848984752)*x_ref[2]
        ref[0,0,1,0]=(0.147766301604)*x_ref[0]**o+(0.405651654428)*x_ref[0]+(-0.440827209529)*x_ref[1]**o+(-0.399952864271)*x_ref[1]+(-0.770191009459)*x_ref[2]**o+(-0.789853968049)*x_ref[2]
        ref[0,0,1,1]=(0.688053165375)*x_ref[0]**o+(-0.154890421924)*x_ref[0]+(0.583684164319)*x_ref[1]**o+(-0.398082068217)*x_ref[1]+(-0.573142923099)*x_ref[2]**o+(-0.205271877337)*x_ref[2]
        ref[0,0,2,0]=(0.807043275877)*x_ref[0]**o+(0.843046383231)*x_ref[0]+(-0.588146471402)*x_ref[1]**o+(0.157787304557)*x_ref[1]+(-0.771009748568)*x_ref[2]**o+(0.162090791609)*x_ref[2]
        ref[0,0,2,1]=(-0.0706800489392)*x_ref[0]**o+(0.499961344665)*x_ref[0]+(0.208926178797)*x_ref[1]**o+(-0.7462472769)*x_ref[1]+(-0.105475405648)*x_ref[2]**o+(-0.18670229948)*x_ref[2]
        ref[0,1,0,0]=(-0.696657281943)*x_ref[0]**o+(-0.0251785345759)*x_ref[0]+(-0.579575530451)*x_ref[1]**o+(-0.0628378665725)*x_ref[1]+(0.668395441302)*x_ref[2]**o+(0.813478105092)*x_ref[2]
        ref[0,1,0,1]=(-0.574784314876)*x_ref[0]**o+(-0.321268045204)*x_ref[0]+(0.160704464213)*x_ref[1]**o+(-0.676016748103)*x_ref[1]+(0.918971102865)*x_ref[2]**o+(0.252082410081)*x_ref[2]
        ref[0,1,1,0]=(-0.905062846817)*x_ref[0]**o+(0.788906658477)*x_ref[0]+(-0.408807611936)*x_ref[1]**o+(0.114686186437)*x_ref[1]+(-0.654685568493)*x_ref[2]**o+(0.0879202139254)*x_ref[2]
        ref[0,1,1,1]=(-0.68020231443)*x_ref[0]**o+(-0.82049202902)*x_ref[0]+(-0.958352740187)*x_ref[1]**o+(-0.33201745802)*x_ref[1]+(0.539810363682)*x_ref[2]**o+(0.744988573087)*x_ref[2]
        ref[0,1,2,0]=(-0.960884744997)*x_ref[0]**o+(-0.704157132503)*x_ref[0]+(0.282236905414)*x_ref[1]**o+(0.756363456433)*x_ref[1]+(-0.365400101792)*x_ref[2]**o+(0.975735846554)*x_ref[2]
        ref[0,1,2,1]=(-0.691780076951)*x_ref[0]**o+(-0.952575440059)*x_ref[0]+(-0.921247808098)*x_ref[1]**o+(-0.88872834095)*x_ref[1]+(0.534103013833)*x_ref[2]**o+(-0.602295186104)*x_ref[2]
        ref[0,2,0,0]=(0.460698415704)*x_ref[0]**o+(0.770243814897)*x_ref[0]+(-0.380381727574)*x_ref[1]**o+(0.476360845668)*x_ref[1]+(-0.681040425102)*x_ref[2]**o+(-0.244369435116)*x_ref[2]
        ref[0,2,0,1]=(0.251591652281)*x_ref[0]**o+(0.748234919686)*x_ref[0]+(-0.698189037307)*x_ref[1]**o+(-0.292400412508)*x_ref[1]+(0.613850696577)*x_ref[2]**o+(-0.942243226834)*x_ref[2]
        ref[0,2,1,0]=(-0.718090066328)*x_ref[0]**o+(-0.322499511794)*x_ref[0]+(-0.75105623397)*x_ref[1]**o+(0.526680271832)*x_ref[1]+(-0.186671531793)*x_ref[2]**o+(-0.449273875991)*x_ref[2]
        ref[0,2,1,1]=(-0.52165552452)*x_ref[0]**o+(0.902446006085)*x_ref[0]+(-0.915575172702)*x_ref[1]**o+(-0.150816194662)*x_ref[1]+(0.596782785689)*x_ref[2]**o+(0.546880865642)*x_ref[2]
        ref[0,2,2,0]=(-0.104486704816)*x_ref[0]**o+(0.723863699457)*x_ref[0]+(0.758622117279)*x_ref[1]**o+(0.372271580145)*x_ref[1]+(-0.0820570383792)*x_ref[2]**o+(-0.63539781104)*x_ref[2]
        ref[0,2,2,1]=(0.0355930823603)*x_ref[0]**o+(0.0931942600432)*x_ref[0]+(-0.118096985978)*x_ref[1]**o+(0.296983782212)*x_ref[1]+(-0.864491763712)*x_ref[2]**o+(0.942230663978)*x_ref[2]
        ref[0,3,0,0]=(0.371153709815)*x_ref[0]**o+(0.0427309083893)*x_ref[0]+(-0.886563919084)*x_ref[1]**o+(-0.14946513404)*x_ref[1]+(0.80903804519)*x_ref[2]**o+(0.69728533073)*x_ref[2]
        ref[0,3,0,1]=(0.0593930396142)*x_ref[0]**o+(-0.370327974089)*x_ref[0]+(-0.388354464683)*x_ref[1]**o+(-0.367648771812)*x_ref[1]+(-0.573754821458)*x_ref[2]**o+(-0.947499155479)*x_ref[2]
        ref[0,3,1,0]=(-0.869817317)*x_ref[0]**o+(0.874317772678)*x_ref[0]+(-0.870060178375)*x_ref[1]**o+(0.66571074524)*x_ref[1]+(-0.669052377378)*x_ref[2]**o+(0.940212680575)*x_ref[2]
        ref[0,3,1,1]=(-0.417445606772)*x_ref[0]**o+(0.942606057619)*x_ref[0]+(-0.133936034409)*x_ref[1]**o+(-0.363274280866)*x_ref[1]+(0.83487439935)*x_ref[2]**o+(-0.108398981461)*x_ref[2]
        ref[0,3,2,0]=(0.363623383293)*x_ref[0]**o+(0.624267309217)*x_ref[0]+(0.801719984899)*x_ref[1]**o+(0.189212370122)*x_ref[1]+(0.464633630928)*x_ref[2]**o+(-0.979808318438)*x_ref[2]
        ref[0,3,2,1]=(0.85384408805)*x_ref[0]**o+(-0.889214754116)*x_ref[0]+(0.164915391224)*x_ref[1]**o+(0.0168991405201)*x_ref[1]+(0.097171994236)*x_ref[2]**o+(-0.146094197339)*x_ref[2]
        ref[0,4,0,0]=(0.778356158972)*x_ref[0]**o+(0.910241265149)*x_ref[0]+(0.123690863272)*x_ref[1]**o+(0.87802121779)*x_ref[1]+(-0.714615668595)*x_ref[2]**o+(-0.437472831583)*x_ref[2]
        ref[0,4,0,1]=(-0.157185481783)*x_ref[0]**o+(0.352802163968)*x_ref[0]+(0.390184123612)*x_ref[1]**o+(-0.680415093098)*x_ref[1]+(-0.121195042454)*x_ref[2]**o+(-0.713693712117)*x_ref[2]
        ref[0,4,1,0]=(0.778115668646)*x_ref[0]**o+(-0.111729042892)*x_ref[0]+(0.451518002735)*x_ref[1]**o+(0.232731183374)*x_ref[1]+(-0.764246158086)*x_ref[2]**o+(-0.795525155328)*x_ref[2]
        ref[0,4,1,1]=(-0.880652367027)*x_ref[0]**o+(-0.314966079126)*x_ref[0]+(-0.640720373065)*x_ref[1]**o+(0.336230541901)*x_ref[1]+(0.566956999526)*x_ref[2]**o+(0.74098258811)*x_ref[2]
        ref[0,4,2,0]=(0.908657274301)*x_ref[0]**o+(0.636427823614)*x_ref[0]+(0.650137414386)*x_ref[1]**o+(0.695558379005)*x_ref[1]+(-0.489852401087)*x_ref[2]**o+(-0.479227108044)*x_ref[2]
        ref[0,4,2,1]=(0.015734241525)*x_ref[0]**o+(0.0719884727088)*x_ref[0]+(0.811864895261)*x_ref[1]**o+(0.959127986784)*x_ref[1]+(-0.238146663368)*x_ref[2]**o+(-0.454939322571)*x_ref[2]
        ref[1,0,0,0]=(-0.918218246559)*x_ref[0]**o+(-0.662787169269)*x_ref[0]+(-0.954035144111)*x_ref[1]**o+(-0.647264685228)*x_ref[1]+(-0.834085960213)*x_ref[2]**o+(-0.167541951024)*x_ref[2]
        ref[1,0,0,1]=(-0.670950869382)*x_ref[0]**o+(0.826472159989)*x_ref[0]+(0.180410872786)*x_ref[1]**o+(-0.715162871594)*x_ref[1]+(0.653287853169)*x_ref[2]**o+(0.892060529625)*x_ref[2]
        ref[1,0,1,0]=(0.398789786407)*x_ref[0]**o+(0.689408033554)*x_ref[0]+(0.124338245128)*x_ref[1]**o+(0.748280808587)*x_ref[1]+(0.418357122274)*x_ref[2]**o+(0.230707440075)*x_ref[2]
        ref[1,0,1,1]=(-0.912778964055)*x_ref[0]**o+(0.904288708694)*x_ref[0]+(-0.874744856901)*x_ref[1]**o+(0.550162195719)*x_ref[1]+(-0.7526423373)*x_ref[2]**o+(0.820993231654)*x_ref[2]
        ref[1,0,2,0]=(-0.843459684648)*x_ref[0]**o+(-0.579708764091)*x_ref[0]+(-0.876072131849)*x_ref[1]**o+(0.933694599696)*x_ref[1]+(0.507781209021)*x_ref[2]**o+(-0.719729482133)*x_ref[2]
        ref[1,0,2,1]=(0.594566925796)*x_ref[0]**o+(-0.994301292358)*x_ref[0]+(0.808114389564)*x_ref[1]**o+(-0.970803712864)*x_ref[1]+(0.433544910389)*x_ref[2]**o+(0.260659440299)*x_ref[2]
        ref[1,1,0,0]=(0.997377553172)*x_ref[0]**o+(0.458428616383)*x_ref[0]+(0.357280778493)*x_ref[1]**o+(0.349837110741)*x_ref[1]+(-0.709592214445)*x_ref[2]**o+(-0.630000739033)*x_ref[2]
        ref[1,1,0,1]=(0.121595650501)*x_ref[0]**o+(0.40870120924)*x_ref[0]+(0.576917123807)*x_ref[1]**o+(-0.932268308325)*x_ref[1]+(0.710043472116)*x_ref[2]**o+(-0.716528412077)*x_ref[2]
        ref[1,1,1,0]=(-0.151416035033)*x_ref[0]**o+(0.0255347649787)*x_ref[0]+(-0.752987701889)*x_ref[1]**o+(-0.480969875813)*x_ref[1]+(-0.610514504179)*x_ref[2]**o+(0.959274757402)*x_ref[2]
        ref[1,1,1,1]=(0.76695706106)*x_ref[0]**o+(-0.0750965975512)*x_ref[0]+(0.846833900735)*x_ref[1]**o+(-0.105838465439)*x_ref[1]+(0.541027486457)*x_ref[2]**o+(0.152543017589)*x_ref[2]
        ref[1,1,2,0]=(0.0856977181865)*x_ref[0]**o+(-0.188139369706)*x_ref[0]+(0.936071565289)*x_ref[1]**o+(-0.3958713036)*x_ref[1]+(-0.350945878893)*x_ref[2]**o+(0.382302674883)*x_ref[2]
        ref[1,1,2,1]=(0.873054234789)*x_ref[0]**o+(-0.490869716146)*x_ref[0]+(0.46720526212)*x_ref[1]**o+(-0.334203412334)*x_ref[1]+(-0.81140895636)*x_ref[2]**o+(-0.973858971749)*x_ref[2]
        ref[1,2,0,0]=(-0.716676352443)*x_ref[0]**o+(0.659821396277)*x_ref[0]+(0.335964292954)*x_ref[1]**o+(0.11609664773)*x_ref[1]+(-0.889186749466)*x_ref[2]**o+(0.454231973946)*x_ref[2]
        ref[1,2,0,1]=(0.367048872648)*x_ref[0]**o+(0.276083713126)*x_ref[0]+(0.686639128043)*x_ref[1]**o+(-0.991414415293)*x_ref[1]+(0.25360515964)*x_ref[2]**o+(-0.377623315093)*x_ref[2]
        ref[1,2,1,0]=(-0.723089358777)*x_ref[0]**o+(-0.516690112406)*x_ref[0]+(0.262334273979)*x_ref[1]**o+(0.140773502537)*x_ref[1]+(0.447333972396)*x_ref[2]**o+(-0.551395692966)*x_ref[2]
        ref[1,2,1,1]=(0.738791182491)*x_ref[0]**o+(-0.978120626668)*x_ref[0]+(-0.217970433713)*x_ref[1]**o+(0.0181011067492)*x_ref[1]+(0.756322867523)*x_ref[2]**o+(0.495269195978)*x_ref[2]
        ref[1,2,2,0]=(-0.630506449854)*x_ref[0]**o+(-0.0865482364231)*x_ref[0]+(-0.415896403157)*x_ref[1]**o+(-0.712833149895)*x_ref[1]+(-0.180482789777)*x_ref[2]**o+(0.00788290661235)*x_ref[2]
        ref[1,2,2,1]=(0.0946913865377)*x_ref[0]**o+(0.752854827716)*x_ref[0]+(0.257846220843)*x_ref[1]**o+(-0.953926983578)*x_ref[1]+(-0.388505920682)*x_ref[2]**o+(-0.10751401251)*x_ref[2]
        ref[1,3,0,0]=(0.694014030611)*x_ref[0]**o+(0.0433103547797)*x_ref[0]+(-0.0805025225569)*x_ref[1]**o+(0.389052254754)*x_ref[1]+(-0.813210939614)*x_ref[2]**o+(-0.593788787896)*x_ref[2]
        ref[1,3,0,1]=(0.0101606353406)*x_ref[0]**o+(-0.849874258855)*x_ref[0]+(-0.0578973959381)*x_ref[1]**o+(0.942113140424)*x_ref[1]+(0.613716459216)*x_ref[2]**o+(0.72631176833)*x_ref[2]
        ref[1,3,1,0]=(0.264703608811)*x_ref[0]**o+(0.457962548326)*x_ref[0]+(-0.433795242673)*x_ref[1]**o+(0.650995712848)*x_ref[1]+(0.372471807967)*x_ref[2]**o+(0.00416636841047)*x_ref[2]
        ref[1,3,1,1]=(-0.73686451669)*x_ref[0]**o+(-0.0425614910776)*x_ref[0]+(-0.587820431311)*x_ref[1]**o+(-0.377628617833)*x_ref[1]+(-0.463964014406)*x_ref[2]**o+(0.283273874989)*x_ref[2]
        ref[1,3,2,0]=(-0.689341439143)*x_ref[0]**o+(-0.675861918889)*x_ref[0]+(-0.918622662352)*x_ref[1]**o+(0.740990261888)*x_ref[1]+(0.0454007677041)*x_ref[2]**o+(0.513623361074)*x_ref[2]
        ref[1,3,2,1]=(0.13229101848)*x_ref[0]**o+(0.171873799817)*x_ref[0]+(-0.154491174434)*x_ref[1]**o+(0.897976884089)*x_ref[1]+(-0.598886311821)*x_ref[2]**o+(0.738696710918)*x_ref[2]
        ref[1,4,0,0]=(0.59403636966)*x_ref[0]**o+(0.93856848969)*x_ref[0]+(0.80164849167)*x_ref[1]**o+(-0.753322979568)*x_ref[1]+(-0.107098073916)*x_ref[2]**o+(-0.623047487973)*x_ref[2]
        ref[1,4,0,1]=(0.291046099082)*x_ref[0]**o+(-0.437982471049)*x_ref[0]+(0.017395762947)*x_ref[1]**o+(0.852440593923)*x_ref[1]+(0.630825424359)*x_ref[2]**o+(0.419347491727)*x_ref[2]
        ref[1,4,1,0]=(-0.747962012701)*x_ref[0]**o+(-0.11622032094)*x_ref[0]+(0.324950549781)*x_ref[1]**o+(0.950911398327)*x_ref[1]+(-0.18098990205)*x_ref[2]**o+(-0.676060633522)*x_ref[2]
        ref[1,4,1,1]=(0.545224363128)*x_ref[0]**o+(0.753376367121)*x_ref[0]+(0.271426695753)*x_ref[1]**o+(-0.139309844675)*x_ref[1]+(-0.923434716375)*x_ref[2]**o+(-0.985988324389)*x_ref[2]
        ref[1,4,2,0]=(-0.532260822047)*x_ref[0]**o+(0.901407135724)*x_ref[0]+(0.937451330141)*x_ref[1]**o+(0.285217666559)*x_ref[1]+(-0.133028984824)*x_ref[2]**o+(0.993965133696)*x_ref[2]
        ref[1,4,2,1]=(0.2231815607)*x_ref[0]**o+(0.994845537582)*x_ref[0]+(0.0368727378195)*x_ref[1]**o+(0.511772933777)*x_ref[1]+(-0.871580215976)*x_ref[2]**o+(-0.587010057329)*x_ref[2]
        ref[2,0,0,0]=(-0.55149411913)*x_ref[0]**o+(-0.905541278559)*x_ref[0]+(0.461217373096)*x_ref[1]**o+(-0.53359853241)*x_ref[1]+(-0.666643290883)*x_ref[2]**o+(0.151495826926)*x_ref[2]
        ref[2,0,0,1]=(-0.980485062148)*x_ref[0]**o+(0.150165616162)*x_ref[0]+(0.59165725487)*x_ref[1]**o+(-0.243853536017)*x_ref[1]+(0.66816261686)*x_ref[2]**o+(0.2425696131)*x_ref[2]
        ref[2,0,1,0]=(-0.509337213758)*x_ref[0]**o+(0.652936706687)*x_ref[0]+(-0.849214338091)*x_ref[1]**o+(-0.523354123549)*x_ref[1]+(0.371515689887)*x_ref[2]**o+(0.799989815414)*x_ref[2]
        ref[2,0,1,1]=(0.461584544541)*x_ref[0]**o+(0.662931355954)*x_ref[0]+(0.744415634786)*x_ref[1]**o+(-0.016743801256)*x_ref[1]+(0.172945846325)*x_ref[2]**o+(-0.0889827089341)*x_ref[2]
        ref[2,0,2,0]=(0.506514791562)*x_ref[0]**o+(0.974477160694)*x_ref[0]+(-0.968485343167)*x_ref[1]**o+(0.427480013599)*x_ref[1]+(0.0460755879358)*x_ref[2]**o+(0.186500994948)*x_ref[2]
        ref[2,0,2,1]=(0.259831607993)*x_ref[0]**o+(-0.0141290444125)*x_ref[0]+(-0.968452177682)*x_ref[1]**o+(0.635619125009)*x_ref[1]+(-0.958848431827)*x_ref[2]**o+(-0.0977241523308)*x_ref[2]
        ref[2,1,0,0]=(0.204233222933)*x_ref[0]**o+(0.834739381877)*x_ref[0]+(0.19961185048)*x_ref[1]**o+(-0.238470380578)*x_ref[1]+(-0.749979713773)*x_ref[2]**o+(-0.268168886251)*x_ref[2]
        ref[2,1,0,1]=(0.998123643088)*x_ref[0]**o+(0.635823733041)*x_ref[0]+(-0.265447124419)*x_ref[1]**o+(-0.225582751946)*x_ref[1]+(0.91705937439)*x_ref[2]**o+(-0.739461510792)*x_ref[2]
        ref[2,1,1,0]=(0.0800715288646)*x_ref[0]**o+(-0.37677196659)*x_ref[0]+(-0.0737286350219)*x_ref[1]**o+(0.032076062933)*x_ref[1]+(0.245421912709)*x_ref[2]**o+(-0.8378023935)*x_ref[2]
        ref[2,1,1,1]=(-0.572376055603)*x_ref[0]**o+(0.8675468469)*x_ref[0]+(-0.533012889867)*x_ref[1]**o+(-0.247215658297)*x_ref[1]+(0.673144564693)*x_ref[2]**o+(-0.749136108985)*x_ref[2]
        ref[2,1,2,0]=(-0.661188192488)*x_ref[0]**o+(-0.924435593008)*x_ref[0]+(-0.791845808916)*x_ref[1]**o+(0.00133765351092)*x_ref[1]+(0.233676862873)*x_ref[2]**o+(0.981812737962)*x_ref[2]
        ref[2,1,2,1]=(0.332072445986)*x_ref[0]**o+(-0.909373798588)*x_ref[0]+(0.927025647819)*x_ref[1]**o+(-0.936348942617)*x_ref[1]+(0.251508592152)*x_ref[2]**o+(-0.468147973987)*x_ref[2]
        ref[2,2,0,0]=(0.892377893841)*x_ref[0]**o+(-0.147933293091)*x_ref[0]+(0.950491102076)*x_ref[1]**o+(0.595250355926)*x_ref[1]+(0.0169764649992)*x_ref[2]**o+(-0.683137752586)*x_ref[2]
        ref[2,2,0,1]=(0.00579397005645)*x_ref[0]**o+(0.627197373917)*x_ref[0]+(0.663574756196)*x_ref[1]**o+(0.934106767513)*x_ref[1]+(0.105844878022)*x_ref[2]**o+(0.513322917688)*x_ref[2]
        ref[2,2,1,0]=(-0.168776016133)*x_ref[0]**o+(-0.940071612944)*x_ref[0]+(0.0366053554566)*x_ref[1]**o+(-0.628010909771)*x_ref[1]+(0.250023045906)*x_ref[2]**o+(-0.00327322332166)*x_ref[2]
        ref[2,2,1,1]=(0.904155764376)*x_ref[0]**o+(-0.258018307306)*x_ref[0]+(0.65862820521)*x_ref[1]**o+(0.434480798515)*x_ref[1]+(0.4089419173)*x_ref[2]**o+(-0.518032695598)*x_ref[2]
        ref[2,2,2,0]=(-0.41470070958)*x_ref[0]**o+(-0.783186751952)*x_ref[0]+(0.548309561942)*x_ref[1]**o+(0.837130592803)*x_ref[1]+(-0.346808491848)*x_ref[2]**o+(0.165416040985)*x_ref[2]
        ref[2,2,2,1]=(0.997845909195)*x_ref[0]**o+(0.448428042158)*x_ref[0]+(-0.657050004311)*x_ref[1]**o+(0.493104851747)*x_ref[1]+(0.558315071197)*x_ref[2]**o+(0.206659432923)*x_ref[2]
        ref[2,3,0,0]=(0.433874360232)*x_ref[0]**o+(-0.0635646197131)*x_ref[0]+(-0.521690521523)*x_ref[1]**o+(-0.298130684888)*x_ref[1]+(-0.131905313644)*x_ref[2]**o+(0.64389451112)*x_ref[2]
        ref[2,3,0,1]=(0.457729315216)*x_ref[0]**o+(-0.419323242135)*x_ref[0]+(0.614544051135)*x_ref[1]**o+(-0.701671999275)*x_ref[1]+(0.905494300727)*x_ref[2]**o+(-0.565716985337)*x_ref[2]
        ref[2,3,1,0]=(0.256570728097)*x_ref[0]**o+(0.866349285103)*x_ref[0]+(-0.115768354896)*x_ref[1]**o+(-0.613044456549)*x_ref[1]+(-0.20074185156)*x_ref[2]**o+(-0.188403571776)*x_ref[2]
        ref[2,3,1,1]=(0.767479943414)*x_ref[0]**o+(0.884066215007)*x_ref[0]+(0.363855420056)*x_ref[1]**o+(0.919698981104)*x_ref[1]+(0.225747762317)*x_ref[2]**o+(-0.220362271104)*x_ref[2]
        ref[2,3,2,0]=(-0.746469501382)*x_ref[0]**o+(0.680520398001)*x_ref[0]+(0.926950996325)*x_ref[1]**o+(-0.122831303044)*x_ref[1]+(-0.27226561675)*x_ref[2]**o+(0.933304087951)*x_ref[2]
        ref[2,3,2,1]=(-0.0961434747216)*x_ref[0]**o+(0.987660516767)*x_ref[0]+(0.288175205045)*x_ref[1]**o+(0.683720820403)*x_ref[1]+(0.354403143642)*x_ref[2]**o+(-0.745694089132)*x_ref[2]
        ref[2,4,0,0]=(0.921750835986)*x_ref[0]**o+(-0.0358861598394)*x_ref[0]+(0.190129645259)*x_ref[1]**o+(0.377067803331)*x_ref[1]+(-0.0188615236582)*x_ref[2]**o+(0.804533888824)*x_ref[2]
        ref[2,4,0,1]=(-0.256132234484)*x_ref[0]**o+(-0.102661411017)*x_ref[0]+(-0.903465157009)*x_ref[1]**o+(0.50133310168)*x_ref[1]+(-0.0889102598619)*x_ref[2]**o+(-0.0310174238365)*x_ref[2]
        ref[2,4,1,0]=(0.58483798855)*x_ref[0]**o+(0.967943344305)*x_ref[0]+(-0.0701080005438)*x_ref[1]**o+(0.489476104406)*x_ref[1]+(0.615308427038)*x_ref[2]**o+(-0.38065505709)*x_ref[2]
        ref[2,4,1,1]=(0.647734130724)*x_ref[0]**o+(-0.930934392719)*x_ref[0]+(0.267370540491)*x_ref[1]**o+(-0.0175019424556)*x_ref[1]+(0.38593657129)*x_ref[2]**o+(-0.985922292359)*x_ref[2]
        ref[2,4,2,0]=(-0.289971479933)*x_ref[0]**o+(0.882014313157)*x_ref[0]+(0.147321067579)*x_ref[1]**o+(-0.176771838489)*x_ref[1]+(0.710916531544)*x_ref[2]**o+(-0.0856290170401)*x_ref[2]
        ref[2,4,2,1]=(-0.465930997733)*x_ref[0]**o+(-0.305563942094)*x_ref[0]+(-0.824535262061)*x_ref[1]**o+(-0.917892226972)*x_ref[1]+(0.624462871955)*x_ref[2]**o+(0.544736500065)*x_ref[2]
        ref[3,0,0,0]=(-0.673230815097)*x_ref[0]**o+(-0.726292712917)*x_ref[0]+(0.0531037021688)*x_ref[1]**o+(-0.467219030159)*x_ref[1]+(-0.0729861369718)*x_ref[2]**o+(-0.440394335015)*x_ref[2]
        ref[3,0,0,1]=(-0.358235295558)*x_ref[0]**o+(-0.301184713671)*x_ref[0]+(-0.496030649086)*x_ref[1]**o+(-0.125922184777)*x_ref[1]+(0.946715036973)*x_ref[2]**o+(0.846370337893)*x_ref[2]
        ref[3,0,1,0]=(-0.394175109612)*x_ref[0]**o+(-0.85137582307)*x_ref[0]+(-0.667804785044)*x_ref[1]**o+(0.156762109891)*x_ref[1]+(-0.616444700525)*x_ref[2]**o+(0.0738729399356)*x_ref[2]
        ref[3,0,1,1]=(-0.652875462534)*x_ref[0]**o+(-0.0605650192541)*x_ref[0]+(0.840539933823)*x_ref[1]**o+(-0.422819423301)*x_ref[1]+(0.113440226819)*x_ref[2]**o+(-0.915154532306)*x_ref[2]
        ref[3,0,2,0]=(-0.162863352328)*x_ref[0]**o+(-0.364461332292)*x_ref[0]+(-0.222181741398)*x_ref[1]**o+(-0.563668008365)*x_ref[1]+(0.118491770105)*x_ref[2]**o+(0.0809155053022)*x_ref[2]
        ref[3,0,2,1]=(-0.985278929437)*x_ref[0]**o+(-0.919440760157)*x_ref[0]+(-0.776910686)*x_ref[1]**o+(-0.884270488943)*x_ref[1]+(0.984279010363)*x_ref[2]**o+(-0.489503278582)*x_ref[2]
        ref[3,1,0,0]=(-0.806565449823)*x_ref[0]**o+(-0.0851385375281)*x_ref[0]+(-0.88851381773)*x_ref[1]**o+(-0.666539867528)*x_ref[1]+(0.154825531411)*x_ref[2]**o+(0.764069849144)*x_ref[2]
        ref[3,1,0,1]=(-0.222125783918)*x_ref[0]**o+(0.568035675467)*x_ref[0]+(0.470171778932)*x_ref[1]**o+(-0.307608710881)*x_ref[1]+(-0.836083742471)*x_ref[2]**o+(0.553597553381)*x_ref[2]
        ref[3,1,1,0]=(0.431237389847)*x_ref[0]**o+(-0.141391012471)*x_ref[0]+(0.726895025519)*x_ref[1]**o+(-0.527868751679)*x_ref[1]+(0.711761083137)*x_ref[2]**o+(0.996992864103)*x_ref[2]
        ref[3,1,1,1]=(-0.123885959924)*x_ref[0]**o+(0.877034475529)*x_ref[0]+(-0.0432492682928)*x_ref[1]**o+(-0.370315494618)*x_ref[1]+(0.243139719071)*x_ref[2]**o+(-0.684318634902)*x_ref[2]
        ref[3,1,2,0]=(-0.477214489355)*x_ref[0]**o+(0.340808576619)*x_ref[0]+(-0.0991802306558)*x_ref[1]**o+(0.462599821908)*x_ref[1]+(0.091768281564)*x_ref[2]**o+(0.0315552838839)*x_ref[2]
        ref[3,1,2,1]=(0.676189142894)*x_ref[0]**o+(0.510136386913)*x_ref[0]+(-0.558579041952)*x_ref[1]**o+(0.961615047877)*x_ref[1]+(0.510717386124)*x_ref[2]**o+(0.774430183432)*x_ref[2]
        ref[3,2,0,0]=(0.158370232444)*x_ref[0]**o+(0.209424855992)*x_ref[0]+(0.691438753013)*x_ref[1]**o+(-0.781594298738)*x_ref[1]+(-0.0188536627517)*x_ref[2]**o+(0.136400798995)*x_ref[2]
        ref[3,2,0,1]=(-0.440674414328)*x_ref[0]**o+(0.118228612687)*x_ref[0]+(-0.700026492613)*x_ref[1]**o+(-0.981341215714)*x_ref[1]+(-0.172847519578)*x_ref[2]**o+(-0.423095202172)*x_ref[2]
        ref[3,2,1,0]=(-0.0694960391695)*x_ref[0]**o+(-0.66448610135)*x_ref[0]+(-0.177572703748)*x_ref[1]**o+(-0.060488161727)*x_ref[1]+(0.306021853885)*x_ref[2]**o+(0.0730338716187)*x_ref[2]
        ref[3,2,1,1]=(-0.360662154547)*x_ref[0]**o+(0.577679359648)*x_ref[0]+(-0.515313534966)*x_ref[1]**o+(0.217352687358)*x_ref[1]+(0.728367577882)*x_ref[2]**o+(0.38505849834)*x_ref[2]
        ref[3,2,2,0]=(0.568671861232)*x_ref[0]**o+(-0.0792490917882)*x_ref[0]+(0.0188002885229)*x_ref[1]**o+(-0.956862575759)*x_ref[1]+(-0.844332442983)*x_ref[2]**o+(0.670792928042)*x_ref[2]
        ref[3,2,2,1]=(0.833099204713)*x_ref[0]**o+(0.368482169933)*x_ref[0]+(-0.926002498892)*x_ref[1]**o+(-0.130646446568)*x_ref[1]+(0.81164928914)*x_ref[2]**o+(-0.641079919927)*x_ref[2]
        ref[3,3,0,0]=(-0.277769350989)*x_ref[0]**o+(0.753452018649)*x_ref[0]+(-0.766322344505)*x_ref[1]**o+(-0.679108224612)*x_ref[1]+(0.0799499795424)*x_ref[2]**o+(0.255751308975)*x_ref[2]
        ref[3,3,0,1]=(0.429176753733)*x_ref[0]**o+(0.123106050259)*x_ref[0]+(-0.169037728992)*x_ref[1]**o+(0.749245933687)*x_ref[1]+(0.560140859355)*x_ref[2]**o+(-0.937091840021)*x_ref[2]
        ref[3,3,1,0]=(0.648792557593)*x_ref[0]**o+(-0.057555187513)*x_ref[0]+(0.883290331726)*x_ref[1]**o+(0.539276905995)*x_ref[1]+(0.473614408583)*x_ref[2]**o+(0.723258340333)*x_ref[2]
        ref[3,3,1,1]=(0.527298833572)*x_ref[0]**o+(0.0753824058279)*x_ref[0]+(0.719196099257)*x_ref[1]**o+(-0.0124603350035)*x_ref[1]+(-0.964484650253)*x_ref[2]**o+(0.404444269694)*x_ref[2]
        ref[3,3,2,0]=(-0.403640677462)*x_ref[0]**o+(0.906195964766)*x_ref[0]+(-0.561419078404)*x_ref[1]**o+(0.645408717838)*x_ref[1]+(0.671624799609)*x_ref[2]**o+(-0.573901868621)*x_ref[2]
        ref[3,3,2,1]=(-0.288117182165)*x_ref[0]**o+(-0.391003584945)*x_ref[0]+(-0.555037182038)*x_ref[1]**o+(0.299362209017)*x_ref[1]+(-0.552429724326)*x_ref[2]**o+(-0.416923306774)*x_ref[2]
        ref[3,4,0,0]=(-0.19594719192)*x_ref[0]**o+(0.150898934692)*x_ref[0]+(0.210935232044)*x_ref[1]**o+(-0.319764293347)*x_ref[1]+(-0.642450751195)*x_ref[2]**o+(0.22231085168)*x_ref[2]
        ref[3,4,0,1]=(-0.0514589460225)*x_ref[0]**o+(0.0366025646754)*x_ref[0]+(0.416280576691)*x_ref[1]**o+(0.609178063951)*x_ref[1]+(0.835611677062)*x_ref[2]**o+(-0.131630571478)*x_ref[2]
        ref[3,4,1,0]=(0.171509152214)*x_ref[0]**o+(-0.761188271093)*x_ref[0]+(-0.915951802896)*x_ref[1]**o+(-0.451796709915)*x_ref[1]+(-0.979847828162)*x_ref[2]**o+(0.0566453393315)*x_ref[2]
        ref[3,4,1,1]=(0.690427998102)*x_ref[0]**o+(-0.358318625695)*x_ref[0]+(0.330270871613)*x_ref[1]**o+(-0.724392522137)*x_ref[1]+(-0.535474623489)*x_ref[2]**o+(-0.12341743885)*x_ref[2]
        ref[3,4,2,0]=(0.488090753426)*x_ref[0]**o+(-0.352417078372)*x_ref[0]+(0.971988099319)*x_ref[1]**o+(-0.288929516009)*x_ref[1]+(0.612576734094)*x_ref[2]**o+(-0.43799514603)*x_ref[2]
        ref[3,4,2,1]=(-0.098768101356)*x_ref[0]**o+(0.123300661895)*x_ref[0]+(-0.0489724280789)*x_ref[1]**o+(-0.422508669137)*x_ref[1]+(-0.861483104139)*x_ref[2]**o+(0.950198316001)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.724373964817)*x[0]**o+(-0.0560850774901)*x[0]+(0.201815078281)*x[1]**o+(-0.721683697048)*x[1]
        ref=(0.724373964817)*x_ref[0]**o+(-0.0560850774901)*x_ref[0]+(0.201815078281)*x_ref[1]**o+(-0.721683697048)*x_ref[1]
      else:
        arg=(0.390296010052)*x[0]**o+(-0.279464234123)*x[0]+(-0.791334947488)*x[1]**o+(-0.0210629346704)*x[1]+(0.0816556526012)*x[2]**o+(0.77326269137)*x[2]
        ref=(0.390296010052)*x_ref[0]**o+(-0.279464234123)*x_ref[0]+(-0.791334947488)*x_ref[1]**o+(-0.0210629346704)*x_ref[1]+(0.0816556526012)*x_ref[2]**o+(0.77326269137)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.645585823917)*x[0]**o+(-0.203443728436)*x[0]+(0.278126891692)*x[1]**o+(0.922093564911)*x[1]
        arg[1]=(0.23370137029)*x[0]**o+(0.84371173541)*x[0]+(0.99297574548)*x[1]**o+(0.72481785931)*x[1]
        ref[0]=(0.645585823917)*x_ref[0]**o+(-0.203443728436)*x_ref[0]+(0.278126891692)*x_ref[1]**o+(0.922093564911)*x_ref[1]
        ref[1]=(0.23370137029)*x_ref[0]**o+(0.84371173541)*x_ref[0]+(0.99297574548)*x_ref[1]**o+(0.72481785931)*x_ref[1]
      else:
        arg[0]=(0.159998049951)*x[0]**o+(-0.863367327023)*x[0]+(-0.0628498246638)*x[1]**o+(0.685139358273)*x[1]+(0.783213360641)*x[2]**o+(0.397526421338)*x[2]
        arg[1]=(0.165519279053)*x[0]**o+(-0.0936018967661)*x[0]+(0.863047696619)*x[1]**o+(0.653118186434)*x[1]+(-0.408224131944)*x[2]**o+(0.82400132854)*x[2]
        ref[0]=(0.159998049951)*x_ref[0]**o+(-0.863367327023)*x_ref[0]+(-0.0628498246638)*x_ref[1]**o+(0.685139358273)*x_ref[1]+(0.783213360641)*x_ref[2]**o+(0.397526421338)*x_ref[2]
        ref[1]=(0.165519279053)*x_ref[0]**o+(-0.0936018967661)*x_ref[0]+(0.863047696619)*x_ref[1]**o+(0.653118186434)*x_ref[1]+(-0.408224131944)*x_ref[2]**o+(0.82400132854)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.490709482882)*x[0]**o+(0.518382584887)*x[0]+(-0.940707425885)*x[1]**o+(0.224578726673)*x[1]
        arg[0,1]=(-0.747621141516)*x[0]**o+(-0.217452489097)*x[0]+(0.929120128008)*x[1]**o+(0.55263700209)*x[1]
        arg[0,2]=(-0.183545119672)*x[0]**o+(0.148192386895)*x[0]+(-0.22046309708)*x[1]**o+(0.433620446668)*x[1]
        arg[0,3]=(-0.0724840266959)*x[0]**o+(0.835144620602)*x[0]+(-0.962971534144)*x[1]**o+(-0.149681811534)*x[1]
        arg[0,4]=(-0.380150192228)*x[0]**o+(-0.309494603814)*x[0]+(-0.217425635182)*x[1]**o+(0.93567286482)*x[1]
        arg[1,0]=(0.0367921563947)*x[0]**o+(0.220217625576)*x[0]+(-0.120595819238)*x[1]**o+(0.308108793279)*x[1]
        arg[1,1]=(-0.142860237109)*x[0]**o+(-0.789758888253)*x[0]+(0.748085258229)*x[1]**o+(0.784767485883)*x[1]
        arg[1,2]=(-0.154690818845)*x[0]**o+(-0.678458370784)*x[0]+(0.750742996193)*x[1]**o+(-0.350199095531)*x[1]
        arg[1,3]=(-0.893295215013)*x[0]**o+(0.0295805814985)*x[0]+(0.265893730887)*x[1]**o+(0.435487860418)*x[1]
        arg[1,4]=(-0.302701704727)*x[0]**o+(0.618274652266)*x[0]+(-0.688386407308)*x[1]**o+(-0.832515064466)*x[1]
        arg[2,0]=(-0.926658745469)*x[0]**o+(0.548174350496)*x[0]+(-0.350700366641)*x[1]**o+(0.530188602011)*x[1]
        arg[2,1]=(0.900965385167)*x[0]**o+(0.0234051845124)*x[0]+(-0.00783834941699)*x[1]**o+(-0.766352156598)*x[1]
        arg[2,2]=(0.694733282444)*x[0]**o+(-0.0644519234093)*x[0]+(0.0493475528391)*x[1]**o+(-0.589101546403)*x[1]
        arg[2,3]=(-0.100333102737)*x[0]**o+(-0.633740106784)*x[0]+(0.408697826034)*x[1]**o+(-0.429059439472)*x[1]
        arg[2,4]=(0.635563324404)*x[0]**o+(0.955881346028)*x[0]+(0.0755013269452)*x[1]**o+(-0.743681282337)*x[1]
        arg[3,0]=(0.604891134834)*x[0]**o+(0.942047409702)*x[0]+(-0.00544373299112)*x[1]**o+(0.447523533166)*x[1]
        arg[3,1]=(-0.980967446837)*x[0]**o+(0.019824933239)*x[0]+(-0.826876930392)*x[1]**o+(0.113548968693)*x[1]
        arg[3,2]=(-0.731417395373)*x[0]**o+(-0.275072472108)*x[0]+(0.933867711492)*x[1]**o+(-0.444847972819)*x[1]
        arg[3,3]=(0.00121172413626)*x[0]**o+(-0.171073908218)*x[0]+(0.913182873344)*x[1]**o+(0.960256570926)*x[1]
        arg[3,4]=(-0.395134372727)*x[0]**o+(0.646385082409)*x[0]+(0.17387019219)*x[1]**o+(-0.287550462397)*x[1]
        ref[0,0]=(0.490709482882)*x_ref[0]**o+(0.518382584887)*x_ref[0]+(-0.940707425885)*x_ref[1]**o+(0.224578726673)*x_ref[1]
        ref[0,1]=(-0.747621141516)*x_ref[0]**o+(-0.217452489097)*x_ref[0]+(0.929120128008)*x_ref[1]**o+(0.55263700209)*x_ref[1]
        ref[0,2]=(-0.183545119672)*x_ref[0]**o+(0.148192386895)*x_ref[0]+(-0.22046309708)*x_ref[1]**o+(0.433620446668)*x_ref[1]
        ref[0,3]=(-0.0724840266959)*x_ref[0]**o+(0.835144620602)*x_ref[0]+(-0.962971534144)*x_ref[1]**o+(-0.149681811534)*x_ref[1]
        ref[0,4]=(-0.380150192228)*x_ref[0]**o+(-0.309494603814)*x_ref[0]+(-0.217425635182)*x_ref[1]**o+(0.93567286482)*x_ref[1]
        ref[1,0]=(0.0367921563947)*x_ref[0]**o+(0.220217625576)*x_ref[0]+(-0.120595819238)*x_ref[1]**o+(0.308108793279)*x_ref[1]
        ref[1,1]=(-0.142860237109)*x_ref[0]**o+(-0.789758888253)*x_ref[0]+(0.748085258229)*x_ref[1]**o+(0.784767485883)*x_ref[1]
        ref[1,2]=(-0.154690818845)*x_ref[0]**o+(-0.678458370784)*x_ref[0]+(0.750742996193)*x_ref[1]**o+(-0.350199095531)*x_ref[1]
        ref[1,3]=(-0.893295215013)*x_ref[0]**o+(0.0295805814985)*x_ref[0]+(0.265893730887)*x_ref[1]**o+(0.435487860418)*x_ref[1]
        ref[1,4]=(-0.302701704727)*x_ref[0]**o+(0.618274652266)*x_ref[0]+(-0.688386407308)*x_ref[1]**o+(-0.832515064466)*x_ref[1]
        ref[2,0]=(-0.926658745469)*x_ref[0]**o+(0.548174350496)*x_ref[0]+(-0.350700366641)*x_ref[1]**o+(0.530188602011)*x_ref[1]
        ref[2,1]=(0.900965385167)*x_ref[0]**o+(0.0234051845124)*x_ref[0]+(-0.00783834941699)*x_ref[1]**o+(-0.766352156598)*x_ref[1]
        ref[2,2]=(0.694733282444)*x_ref[0]**o+(-0.0644519234093)*x_ref[0]+(0.0493475528391)*x_ref[1]**o+(-0.589101546403)*x_ref[1]
        ref[2,3]=(-0.100333102737)*x_ref[0]**o+(-0.633740106784)*x_ref[0]+(0.408697826034)*x_ref[1]**o+(-0.429059439472)*x_ref[1]
        ref[2,4]=(0.635563324404)*x_ref[0]**o+(0.955881346028)*x_ref[0]+(0.0755013269452)*x_ref[1]**o+(-0.743681282337)*x_ref[1]
        ref[3,0]=(0.604891134834)*x_ref[0]**o+(0.942047409702)*x_ref[0]+(-0.00544373299112)*x_ref[1]**o+(0.447523533166)*x_ref[1]
        ref[3,1]=(-0.980967446837)*x_ref[0]**o+(0.019824933239)*x_ref[0]+(-0.826876930392)*x_ref[1]**o+(0.113548968693)*x_ref[1]
        ref[3,2]=(-0.731417395373)*x_ref[0]**o+(-0.275072472108)*x_ref[0]+(0.933867711492)*x_ref[1]**o+(-0.444847972819)*x_ref[1]
        ref[3,3]=(0.00121172413626)*x_ref[0]**o+(-0.171073908218)*x_ref[0]+(0.913182873344)*x_ref[1]**o+(0.960256570926)*x_ref[1]
        ref[3,4]=(-0.395134372727)*x_ref[0]**o+(0.646385082409)*x_ref[0]+(0.17387019219)*x_ref[1]**o+(-0.287550462397)*x_ref[1]
      else:
        arg[0,0]=(0.386545340192)*x[0]**o+(-0.397636296285)*x[0]+(-0.849437816516)*x[1]**o+(-0.83083406493)*x[1]+(0.479216612031)*x[2]**o+(0.0903787541308)*x[2]
        arg[0,1]=(-0.331651855726)*x[0]**o+(0.849421626918)*x[0]+(0.528368477362)*x[1]**o+(-0.015122214824)*x[1]+(0.494310578452)*x[2]**o+(0.190239928161)*x[2]
        arg[0,2]=(-0.662389759646)*x[0]**o+(0.204378343962)*x[0]+(0.486802744897)*x[1]**o+(0.822174784922)*x[1]+(-0.309203812647)*x[2]**o+(-0.101665040799)*x[2]
        arg[0,3]=(0.609106341828)*x[0]**o+(0.658404088333)*x[0]+(-0.0092991120732)*x[1]**o+(-0.73794934348)*x[1]+(0.270245135914)*x[2]**o+(-0.594067838407)*x[2]
        arg[0,4]=(-0.291752213357)*x[0]**o+(-0.596751343066)*x[0]+(-0.109196903078)*x[1]**o+(-0.868610307708)*x[1]+(0.622541308544)*x[2]**o+(0.482202762224)*x[2]
        arg[1,0]=(-0.113988933227)*x[0]**o+(-0.846328804695)*x[0]+(0.691769553626)*x[1]**o+(-0.377441645148)*x[1]+(-0.769857970682)*x[2]**o+(-0.583083729037)*x[2]
        arg[1,1]=(-0.635333757145)*x[0]**o+(0.194761117849)*x[0]+(-0.268287296731)*x[1]**o+(0.463906930647)*x[1]+(0.768293821029)*x[2]**o+(-0.532824906687)*x[2]
        arg[1,2]=(0.355380090138)*x[0]**o+(-0.181191575848)*x[0]+(-0.558667766467)*x[1]**o+(0.561707671872)*x[1]+(-0.427004232725)*x[2]**o+(-0.996564299989)*x[2]
        arg[1,3]=(-0.891638892893)*x[0]**o+(0.0867350850512)*x[0]+(0.880258390584)*x[1]**o+(-0.798123407525)*x[1]+(-0.571505202201)*x[2]**o+(0.35374245063)*x[2]
        arg[1,4]=(0.543632496941)*x[0]**o+(-0.630363803582)*x[0]+(0.38107312058)*x[1]**o+(0.222401831883)*x[1]+(-0.932212080175)*x[2]**o+(-0.0482899826174)*x[2]
        arg[2,0]=(0.46514308325)*x[0]**o+(-0.907464652384)*x[0]+(-0.706105595148)*x[1]**o+(0.486989958675)*x[1]+(0.422702683485)*x[2]**o+(0.640916925932)*x[2]
        arg[2,1]=(0.912808317479)*x[0]**o+(-0.860779312763)*x[0]+(-0.915364956133)*x[1]**o+(-0.314186117224)*x[1]+(-0.860814410035)*x[2]**o+(0.837570442722)*x[2]
        arg[2,2]=(0.00932917919464)*x[0]**o+(0.393773283049)*x[0]+(0.699998572668)*x[1]**o+(0.153407885795)*x[1]+(-0.166245189281)*x[2]**o+(-0.430541214236)*x[2]
        arg[2,3]=(-0.00127511229816)*x[0]**o+(-0.327443174924)*x[0]+(0.61881322537)*x[1]**o+(-0.725854102962)*x[1]+(0.650649062427)*x[2]**o+(0.246373184041)*x[2]
        arg[2,4]=(-0.530580898049)*x[0]**o+(-0.208506132901)*x[0]+(-0.102049332015)*x[1]**o+(-0.537037644189)*x[1]+(0.740113470729)*x[2]**o+(0.12094405264)*x[2]
        arg[3,0]=(-0.103423602042)*x[0]**o+(-0.535742323666)*x[0]+(0.678663019405)*x[1]**o+(0.579375950057)*x[1]+(0.323627490052)*x[2]**o+(-0.681323301356)*x[2]
        arg[3,1]=(0.670239087519)*x[0]**o+(0.209657968128)*x[0]+(-0.357490300727)*x[1]**o+(-0.441776782321)*x[1]+(0.00927389037562)*x[2]**o+(0.633527599723)*x[2]
        arg[3,2]=(0.415569606138)*x[0]**o+(-0.422211616068)*x[0]+(-0.695086235645)*x[1]**o+(0.709158289848)*x[1]+(-0.523450845183)*x[2]**o+(0.555778994398)*x[2]
        arg[3,3]=(0.0188993892075)*x[0]**o+(0.611088737828)*x[0]+(0.0313795167921)*x[1]**o+(0.0842651755016)*x[1]+(0.288414199415)*x[2]**o+(0.695910780235)*x[2]
        arg[3,4]=(-0.500338800879)*x[0]**o+(0.0984192369328)*x[0]+(0.71163172992)*x[1]**o+(-0.102123116139)*x[1]+(0.954613955678)*x[2]**o+(0.384872409422)*x[2]
        ref[0,0]=(0.386545340192)*x_ref[0]**o+(-0.397636296285)*x_ref[0]+(-0.849437816516)*x_ref[1]**o+(-0.83083406493)*x_ref[1]+(0.479216612031)*x_ref[2]**o+(0.0903787541308)*x_ref[2]
        ref[0,1]=(-0.331651855726)*x_ref[0]**o+(0.849421626918)*x_ref[0]+(0.528368477362)*x_ref[1]**o+(-0.015122214824)*x_ref[1]+(0.494310578452)*x_ref[2]**o+(0.190239928161)*x_ref[2]
        ref[0,2]=(-0.662389759646)*x_ref[0]**o+(0.204378343962)*x_ref[0]+(0.486802744897)*x_ref[1]**o+(0.822174784922)*x_ref[1]+(-0.309203812647)*x_ref[2]**o+(-0.101665040799)*x_ref[2]
        ref[0,3]=(0.609106341828)*x_ref[0]**o+(0.658404088333)*x_ref[0]+(-0.0092991120732)*x_ref[1]**o+(-0.73794934348)*x_ref[1]+(0.270245135914)*x_ref[2]**o+(-0.594067838407)*x_ref[2]
        ref[0,4]=(-0.291752213357)*x_ref[0]**o+(-0.596751343066)*x_ref[0]+(-0.109196903078)*x_ref[1]**o+(-0.868610307708)*x_ref[1]+(0.622541308544)*x_ref[2]**o+(0.482202762224)*x_ref[2]
        ref[1,0]=(-0.113988933227)*x_ref[0]**o+(-0.846328804695)*x_ref[0]+(0.691769553626)*x_ref[1]**o+(-0.377441645148)*x_ref[1]+(-0.769857970682)*x_ref[2]**o+(-0.583083729037)*x_ref[2]
        ref[1,1]=(-0.635333757145)*x_ref[0]**o+(0.194761117849)*x_ref[0]+(-0.268287296731)*x_ref[1]**o+(0.463906930647)*x_ref[1]+(0.768293821029)*x_ref[2]**o+(-0.532824906687)*x_ref[2]
        ref[1,2]=(0.355380090138)*x_ref[0]**o+(-0.181191575848)*x_ref[0]+(-0.558667766467)*x_ref[1]**o+(0.561707671872)*x_ref[1]+(-0.427004232725)*x_ref[2]**o+(-0.996564299989)*x_ref[2]
        ref[1,3]=(-0.891638892893)*x_ref[0]**o+(0.0867350850512)*x_ref[0]+(0.880258390584)*x_ref[1]**o+(-0.798123407525)*x_ref[1]+(-0.571505202201)*x_ref[2]**o+(0.35374245063)*x_ref[2]
        ref[1,4]=(0.543632496941)*x_ref[0]**o+(-0.630363803582)*x_ref[0]+(0.38107312058)*x_ref[1]**o+(0.222401831883)*x_ref[1]+(-0.932212080175)*x_ref[2]**o+(-0.0482899826174)*x_ref[2]
        ref[2,0]=(0.46514308325)*x_ref[0]**o+(-0.907464652384)*x_ref[0]+(-0.706105595148)*x_ref[1]**o+(0.486989958675)*x_ref[1]+(0.422702683485)*x_ref[2]**o+(0.640916925932)*x_ref[2]
        ref[2,1]=(0.912808317479)*x_ref[0]**o+(-0.860779312763)*x_ref[0]+(-0.915364956133)*x_ref[1]**o+(-0.314186117224)*x_ref[1]+(-0.860814410035)*x_ref[2]**o+(0.837570442722)*x_ref[2]
        ref[2,2]=(0.00932917919464)*x_ref[0]**o+(0.393773283049)*x_ref[0]+(0.699998572668)*x_ref[1]**o+(0.153407885795)*x_ref[1]+(-0.166245189281)*x_ref[2]**o+(-0.430541214236)*x_ref[2]
        ref[2,3]=(-0.00127511229816)*x_ref[0]**o+(-0.327443174924)*x_ref[0]+(0.61881322537)*x_ref[1]**o+(-0.725854102962)*x_ref[1]+(0.650649062427)*x_ref[2]**o+(0.246373184041)*x_ref[2]
        ref[2,4]=(-0.530580898049)*x_ref[0]**o+(-0.208506132901)*x_ref[0]+(-0.102049332015)*x_ref[1]**o+(-0.537037644189)*x_ref[1]+(0.740113470729)*x_ref[2]**o+(0.12094405264)*x_ref[2]
        ref[3,0]=(-0.103423602042)*x_ref[0]**o+(-0.535742323666)*x_ref[0]+(0.678663019405)*x_ref[1]**o+(0.579375950057)*x_ref[1]+(0.323627490052)*x_ref[2]**o+(-0.681323301356)*x_ref[2]
        ref[3,1]=(0.670239087519)*x_ref[0]**o+(0.209657968128)*x_ref[0]+(-0.357490300727)*x_ref[1]**o+(-0.441776782321)*x_ref[1]+(0.00927389037562)*x_ref[2]**o+(0.633527599723)*x_ref[2]
        ref[3,2]=(0.415569606138)*x_ref[0]**o+(-0.422211616068)*x_ref[0]+(-0.695086235645)*x_ref[1]**o+(0.709158289848)*x_ref[1]+(-0.523450845183)*x_ref[2]**o+(0.555778994398)*x_ref[2]
        ref[3,3]=(0.0188993892075)*x_ref[0]**o+(0.611088737828)*x_ref[0]+(0.0313795167921)*x_ref[1]**o+(0.0842651755016)*x_ref[1]+(0.288414199415)*x_ref[2]**o+(0.695910780235)*x_ref[2]
        ref[3,4]=(-0.500338800879)*x_ref[0]**o+(0.0984192369328)*x_ref[0]+(0.71163172992)*x_ref[1]**o+(-0.102123116139)*x_ref[1]+(0.954613955678)*x_ref[2]**o+(0.384872409422)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.317274364174)*x[0]**o+(-0.565307843766)*x[0]+(-0.367874638208)*x[1]**o+(0.0029581033546)*x[1]
        arg[0,0,1]=(0.832728755858)*x[0]**o+(0.245271193641)*x[0]+(-0.727130029377)*x[1]**o+(0.137912975829)*x[1]
        arg[0,1,0]=(-0.634508968932)*x[0]**o+(-0.864200861637)*x[0]+(-0.566985971867)*x[1]**o+(0.458187145519)*x[1]
        arg[0,1,1]=(0.133772683885)*x[0]**o+(-0.630800341215)*x[0]+(0.958934971335)*x[1]**o+(-0.15839182257)*x[1]
        arg[1,0,0]=(-0.0948724752087)*x[0]**o+(-0.25589337469)*x[0]+(-0.315713312867)*x[1]**o+(0.223487524432)*x[1]
        arg[1,0,1]=(0.794634271698)*x[0]**o+(-0.505797079712)*x[0]+(0.54367349338)*x[1]**o+(0.610149581933)*x[1]
        arg[1,1,0]=(-0.447039667341)*x[0]**o+(-0.917009838715)*x[0]+(0.525331113522)*x[1]**o+(-0.626176620973)*x[1]
        arg[1,1,1]=(0.0749028334839)*x[0]**o+(0.106950341725)*x[0]+(-0.375321288987)*x[1]**o+(0.510260762626)*x[1]
        arg[2,0,0]=(-0.286661970669)*x[0]**o+(0.584169107201)*x[0]+(0.189852415568)*x[1]**o+(-0.685695910848)*x[1]
        arg[2,0,1]=(-0.689509796724)*x[0]**o+(-0.316067007053)*x[0]+(0.93758916753)*x[1]**o+(0.987368705364)*x[1]
        arg[2,1,0]=(-0.411695937554)*x[0]**o+(0.858528628791)*x[0]+(-0.579361651418)*x[1]**o+(-0.179617869494)*x[1]
        arg[2,1,1]=(0.871441175049)*x[0]**o+(0.445176827889)*x[0]+(0.282613647019)*x[1]**o+(-0.0414230375843)*x[1]
        arg[3,0,0]=(-0.242136375691)*x[0]**o+(-0.91917484201)*x[0]+(-0.88026838607)*x[1]**o+(0.363309786006)*x[1]
        arg[3,0,1]=(-0.991221762386)*x[0]**o+(-0.0149280738883)*x[0]+(0.420322977563)*x[1]**o+(-0.0533076687981)*x[1]
        arg[3,1,0]=(-0.889003524303)*x[0]**o+(-0.023519502168)*x[0]+(0.826015855889)*x[1]**o+(0.0568467842617)*x[1]
        arg[3,1,1]=(0.608567002824)*x[0]**o+(-0.531288593441)*x[0]+(0.0886678034719)*x[1]**o+(-0.732353220319)*x[1]
        arg[4,0,0]=(-0.0537755786538)*x[0]**o+(-0.475886396481)*x[0]+(-0.153787571145)*x[1]**o+(0.689850737569)*x[1]
        arg[4,0,1]=(0.413017674549)*x[0]**o+(-0.993353261494)*x[0]+(-0.944597571268)*x[1]**o+(0.566423551758)*x[1]
        arg[4,1,0]=(-0.716864417268)*x[0]**o+(-0.871574272865)*x[0]+(0.394071572628)*x[1]**o+(0.452903048597)*x[1]
        arg[4,1,1]=(0.861790785883)*x[0]**o+(-0.511789401943)*x[0]+(-0.272452527007)*x[1]**o+(-0.379021521619)*x[1]
        arg[5,0,0]=(0.509668147054)*x[0]**o+(0.919355584411)*x[0]+(-0.673215166917)*x[1]**o+(0.766459631931)*x[1]
        arg[5,0,1]=(0.177006227916)*x[0]**o+(-0.275118477)*x[0]+(0.0302367843701)*x[1]**o+(-0.761226360328)*x[1]
        arg[5,1,0]=(0.013432534638)*x[0]**o+(-0.911626567027)*x[0]+(-0.710520442529)*x[1]**o+(-0.882317448423)*x[1]
        arg[5,1,1]=(0.595583093615)*x[0]**o+(0.163527905156)*x[0]+(-0.122165051333)*x[1]**o+(-0.526359288181)*x[1]
        ref[0,0,0]=(0.317274364174)*x_ref[0]**o+(-0.565307843766)*x_ref[0]+(-0.367874638208)*x_ref[1]**o+(0.0029581033546)*x_ref[1]
        ref[0,0,1]=(0.832728755858)*x_ref[0]**o+(0.245271193641)*x_ref[0]+(-0.727130029377)*x_ref[1]**o+(0.137912975829)*x_ref[1]
        ref[0,1,0]=(-0.634508968932)*x_ref[0]**o+(-0.864200861637)*x_ref[0]+(-0.566985971867)*x_ref[1]**o+(0.458187145519)*x_ref[1]
        ref[0,1,1]=(0.133772683885)*x_ref[0]**o+(-0.630800341215)*x_ref[0]+(0.958934971335)*x_ref[1]**o+(-0.15839182257)*x_ref[1]
        ref[1,0,0]=(-0.0948724752087)*x_ref[0]**o+(-0.25589337469)*x_ref[0]+(-0.315713312867)*x_ref[1]**o+(0.223487524432)*x_ref[1]
        ref[1,0,1]=(0.794634271698)*x_ref[0]**o+(-0.505797079712)*x_ref[0]+(0.54367349338)*x_ref[1]**o+(0.610149581933)*x_ref[1]
        ref[1,1,0]=(-0.447039667341)*x_ref[0]**o+(-0.917009838715)*x_ref[0]+(0.525331113522)*x_ref[1]**o+(-0.626176620973)*x_ref[1]
        ref[1,1,1]=(0.0749028334839)*x_ref[0]**o+(0.106950341725)*x_ref[0]+(-0.375321288987)*x_ref[1]**o+(0.510260762626)*x_ref[1]
        ref[2,0,0]=(-0.286661970669)*x_ref[0]**o+(0.584169107201)*x_ref[0]+(0.189852415568)*x_ref[1]**o+(-0.685695910848)*x_ref[1]
        ref[2,0,1]=(-0.689509796724)*x_ref[0]**o+(-0.316067007053)*x_ref[0]+(0.93758916753)*x_ref[1]**o+(0.987368705364)*x_ref[1]
        ref[2,1,0]=(-0.411695937554)*x_ref[0]**o+(0.858528628791)*x_ref[0]+(-0.579361651418)*x_ref[1]**o+(-0.179617869494)*x_ref[1]
        ref[2,1,1]=(0.871441175049)*x_ref[0]**o+(0.445176827889)*x_ref[0]+(0.282613647019)*x_ref[1]**o+(-0.0414230375843)*x_ref[1]
        ref[3,0,0]=(-0.242136375691)*x_ref[0]**o+(-0.91917484201)*x_ref[0]+(-0.88026838607)*x_ref[1]**o+(0.363309786006)*x_ref[1]
        ref[3,0,1]=(-0.991221762386)*x_ref[0]**o+(-0.0149280738883)*x_ref[0]+(0.420322977563)*x_ref[1]**o+(-0.0533076687981)*x_ref[1]
        ref[3,1,0]=(-0.889003524303)*x_ref[0]**o+(-0.023519502168)*x_ref[0]+(0.826015855889)*x_ref[1]**o+(0.0568467842617)*x_ref[1]
        ref[3,1,1]=(0.608567002824)*x_ref[0]**o+(-0.531288593441)*x_ref[0]+(0.0886678034719)*x_ref[1]**o+(-0.732353220319)*x_ref[1]
        ref[4,0,0]=(-0.0537755786538)*x_ref[0]**o+(-0.475886396481)*x_ref[0]+(-0.153787571145)*x_ref[1]**o+(0.689850737569)*x_ref[1]
        ref[4,0,1]=(0.413017674549)*x_ref[0]**o+(-0.993353261494)*x_ref[0]+(-0.944597571268)*x_ref[1]**o+(0.566423551758)*x_ref[1]
        ref[4,1,0]=(-0.716864417268)*x_ref[0]**o+(-0.871574272865)*x_ref[0]+(0.394071572628)*x_ref[1]**o+(0.452903048597)*x_ref[1]
        ref[4,1,1]=(0.861790785883)*x_ref[0]**o+(-0.511789401943)*x_ref[0]+(-0.272452527007)*x_ref[1]**o+(-0.379021521619)*x_ref[1]
        ref[5,0,0]=(0.509668147054)*x_ref[0]**o+(0.919355584411)*x_ref[0]+(-0.673215166917)*x_ref[1]**o+(0.766459631931)*x_ref[1]
        ref[5,0,1]=(0.177006227916)*x_ref[0]**o+(-0.275118477)*x_ref[0]+(0.0302367843701)*x_ref[1]**o+(-0.761226360328)*x_ref[1]
        ref[5,1,0]=(0.013432534638)*x_ref[0]**o+(-0.911626567027)*x_ref[0]+(-0.710520442529)*x_ref[1]**o+(-0.882317448423)*x_ref[1]
        ref[5,1,1]=(0.595583093615)*x_ref[0]**o+(0.163527905156)*x_ref[0]+(-0.122165051333)*x_ref[1]**o+(-0.526359288181)*x_ref[1]
      else:
        arg[0,0,0]=(0.230695674262)*x[0]**o+(-0.987212801716)*x[0]+(-0.0599001434987)*x[1]**o+(0.719591259555)*x[1]+(-0.665186369759)*x[2]**o+(-0.563552602515)*x[2]
        arg[0,0,1]=(-0.775441966131)*x[0]**o+(-0.152166849702)*x[0]+(0.626466789363)*x[1]**o+(0.179356346581)*x[1]+(-0.687406927434)*x[2]**o+(-0.563492313114)*x[2]
        arg[0,1,0]=(0.749522748133)*x[0]**o+(-0.582456308337)*x[0]+(0.676882643058)*x[1]**o+(0.365363182598)*x[1]+(-0.550867708812)*x[2]**o+(0.777938151183)*x[2]
        arg[0,1,1]=(0.389788552775)*x[0]**o+(-0.298877489997)*x[0]+(0.560326209269)*x[1]**o+(-0.113249832155)*x[1]+(0.588087510254)*x[2]**o+(-0.593503776461)*x[2]
        arg[1,0,0]=(-0.354661103909)*x[0]**o+(-0.195375369549)*x[0]+(-0.798795069895)*x[1]**o+(0.0968631048579)*x[1]+(-0.016118190391)*x[2]**o+(-0.903745407705)*x[2]
        arg[1,0,1]=(-0.383688834784)*x[0]**o+(0.861256231933)*x[0]+(-0.767204834667)*x[1]**o+(0.338875482372)*x[1]+(0.117146310891)*x[2]**o+(-0.302651846083)*x[2]
        arg[1,1,0]=(0.646263670969)*x[0]**o+(-0.447984008871)*x[0]+(-0.169322492785)*x[1]**o+(-0.570614724481)*x[1]+(-0.848518457618)*x[2]**o+(-0.0750005539216)*x[2]
        arg[1,1,1]=(-0.232496745304)*x[0]**o+(0.182581173413)*x[0]+(0.8032436357)*x[1]**o+(-0.391266449476)*x[1]+(0.491677629766)*x[2]**o+(-0.188361708708)*x[2]
        arg[2,0,0]=(-0.571145099091)*x[0]**o+(-0.486855242188)*x[0]+(0.52833582073)*x[1]**o+(0.326716348242)*x[1]+(0.528475192869)*x[2]**o+(0.163280575137)*x[2]
        arg[2,0,1]=(-0.368889674601)*x[0]**o+(0.601775816775)*x[0]+(0.0281742455403)*x[1]**o+(-0.198869716243)*x[1]+(0.193343065506)*x[2]**o+(-0.0614451571108)*x[2]
        arg[2,1,0]=(0.724595792492)*x[0]**o+(-0.206983407651)*x[0]+(-0.33966393719)*x[1]**o+(0.929465830796)*x[1]+(-0.938517195863)*x[2]**o+(0.70874077222)*x[2]
        arg[2,1,1]=(-0.536303362606)*x[0]**o+(-0.430663025812)*x[0]+(-0.0845923469062)*x[1]**o+(0.866655631144)*x[1]+(-0.2378102403)*x[2]**o+(0.969070797268)*x[2]
        arg[3,0,0]=(0.345137352559)*x[0]**o+(-0.696652518382)*x[0]+(0.914998974008)*x[1]**o+(-0.874935163156)*x[1]+(0.947812661283)*x[2]**o+(0.868611662789)*x[2]
        arg[3,0,1]=(-0.421640746358)*x[0]**o+(-0.37925991208)*x[0]+(0.468281241452)*x[1]**o+(0.439086689805)*x[1]+(-0.607986317001)*x[2]**o+(0.269488168937)*x[2]
        arg[3,1,0]=(-0.550485456473)*x[0]**o+(-0.520018714495)*x[0]+(0.439679309654)*x[1]**o+(-0.0652434084668)*x[1]+(-0.584838668708)*x[2]**o+(-0.46047281034)*x[2]
        arg[3,1,1]=(-0.421746456038)*x[0]**o+(0.370489549278)*x[0]+(-0.683682295105)*x[1]**o+(0.324051718035)*x[1]+(-0.579046430636)*x[2]**o+(0.587623604941)*x[2]
        arg[4,0,0]=(-0.167582584882)*x[0]**o+(0.524904122243)*x[0]+(0.665109348362)*x[1]**o+(-0.093655366509)*x[1]+(0.704404656254)*x[2]**o+(-0.402602441741)*x[2]
        arg[4,0,1]=(0.633854964824)*x[0]**o+(-0.836385840194)*x[0]+(-0.0995437996628)*x[1]**o+(-0.200285338869)*x[1]+(0.535023783881)*x[2]**o+(0.884702543593)*x[2]
        arg[4,1,0]=(-0.953619378208)*x[0]**o+(-0.390518815669)*x[0]+(-0.0590889178236)*x[1]**o+(-0.997225825019)*x[1]+(0.0520418736402)*x[2]**o+(0.484749465262)*x[2]
        arg[4,1,1]=(-0.440389854756)*x[0]**o+(0.15520431778)*x[0]+(0.346839174513)*x[1]**o+(0.0172676873159)*x[1]+(0.86275919224)*x[2]**o+(0.568069148992)*x[2]
        arg[5,0,0]=(-0.303687941217)*x[0]**o+(0.516115915687)*x[0]+(-0.750276597519)*x[1]**o+(0.734039058778)*x[1]+(-0.624499210252)*x[2]**o+(-0.926338135421)*x[2]
        arg[5,0,1]=(-0.472230777895)*x[0]**o+(-0.557827728373)*x[0]+(-0.403783023157)*x[1]**o+(0.70433336955)*x[1]+(-0.0211737454531)*x[2]**o+(0.0774406125836)*x[2]
        arg[5,1,0]=(0.718149875204)*x[0]**o+(-0.183000739787)*x[0]+(-0.655724252373)*x[1]**o+(-0.634941710776)*x[1]+(-0.26300163323)*x[2]**o+(-0.25105424715)*x[2]
        arg[5,1,1]=(0.109601904441)*x[0]**o+(-0.00640764631554)*x[0]+(-0.488035610009)*x[1]**o+(0.909668323769)*x[1]+(0.229692168501)*x[2]**o+(-0.739241309497)*x[2]
        ref[0,0,0]=(0.230695674262)*x_ref[0]**o+(-0.987212801716)*x_ref[0]+(-0.0599001434987)*x_ref[1]**o+(0.719591259555)*x_ref[1]+(-0.665186369759)*x_ref[2]**o+(-0.563552602515)*x_ref[2]
        ref[0,0,1]=(-0.775441966131)*x_ref[0]**o+(-0.152166849702)*x_ref[0]+(0.626466789363)*x_ref[1]**o+(0.179356346581)*x_ref[1]+(-0.687406927434)*x_ref[2]**o+(-0.563492313114)*x_ref[2]
        ref[0,1,0]=(0.749522748133)*x_ref[0]**o+(-0.582456308337)*x_ref[0]+(0.676882643058)*x_ref[1]**o+(0.365363182598)*x_ref[1]+(-0.550867708812)*x_ref[2]**o+(0.777938151183)*x_ref[2]
        ref[0,1,1]=(0.389788552775)*x_ref[0]**o+(-0.298877489997)*x_ref[0]+(0.560326209269)*x_ref[1]**o+(-0.113249832155)*x_ref[1]+(0.588087510254)*x_ref[2]**o+(-0.593503776461)*x_ref[2]
        ref[1,0,0]=(-0.354661103909)*x_ref[0]**o+(-0.195375369549)*x_ref[0]+(-0.798795069895)*x_ref[1]**o+(0.0968631048579)*x_ref[1]+(-0.016118190391)*x_ref[2]**o+(-0.903745407705)*x_ref[2]
        ref[1,0,1]=(-0.383688834784)*x_ref[0]**o+(0.861256231933)*x_ref[0]+(-0.767204834667)*x_ref[1]**o+(0.338875482372)*x_ref[1]+(0.117146310891)*x_ref[2]**o+(-0.302651846083)*x_ref[2]
        ref[1,1,0]=(0.646263670969)*x_ref[0]**o+(-0.447984008871)*x_ref[0]+(-0.169322492785)*x_ref[1]**o+(-0.570614724481)*x_ref[1]+(-0.848518457618)*x_ref[2]**o+(-0.0750005539216)*x_ref[2]
        ref[1,1,1]=(-0.232496745304)*x_ref[0]**o+(0.182581173413)*x_ref[0]+(0.8032436357)*x_ref[1]**o+(-0.391266449476)*x_ref[1]+(0.491677629766)*x_ref[2]**o+(-0.188361708708)*x_ref[2]
        ref[2,0,0]=(-0.571145099091)*x_ref[0]**o+(-0.486855242188)*x_ref[0]+(0.52833582073)*x_ref[1]**o+(0.326716348242)*x_ref[1]+(0.528475192869)*x_ref[2]**o+(0.163280575137)*x_ref[2]
        ref[2,0,1]=(-0.368889674601)*x_ref[0]**o+(0.601775816775)*x_ref[0]+(0.0281742455403)*x_ref[1]**o+(-0.198869716243)*x_ref[1]+(0.193343065506)*x_ref[2]**o+(-0.0614451571108)*x_ref[2]
        ref[2,1,0]=(0.724595792492)*x_ref[0]**o+(-0.206983407651)*x_ref[0]+(-0.33966393719)*x_ref[1]**o+(0.929465830796)*x_ref[1]+(-0.938517195863)*x_ref[2]**o+(0.70874077222)*x_ref[2]
        ref[2,1,1]=(-0.536303362606)*x_ref[0]**o+(-0.430663025812)*x_ref[0]+(-0.0845923469062)*x_ref[1]**o+(0.866655631144)*x_ref[1]+(-0.2378102403)*x_ref[2]**o+(0.969070797268)*x_ref[2]
        ref[3,0,0]=(0.345137352559)*x_ref[0]**o+(-0.696652518382)*x_ref[0]+(0.914998974008)*x_ref[1]**o+(-0.874935163156)*x_ref[1]+(0.947812661283)*x_ref[2]**o+(0.868611662789)*x_ref[2]
        ref[3,0,1]=(-0.421640746358)*x_ref[0]**o+(-0.37925991208)*x_ref[0]+(0.468281241452)*x_ref[1]**o+(0.439086689805)*x_ref[1]+(-0.607986317001)*x_ref[2]**o+(0.269488168937)*x_ref[2]
        ref[3,1,0]=(-0.550485456473)*x_ref[0]**o+(-0.520018714495)*x_ref[0]+(0.439679309654)*x_ref[1]**o+(-0.0652434084668)*x_ref[1]+(-0.584838668708)*x_ref[2]**o+(-0.46047281034)*x_ref[2]
        ref[3,1,1]=(-0.421746456038)*x_ref[0]**o+(0.370489549278)*x_ref[0]+(-0.683682295105)*x_ref[1]**o+(0.324051718035)*x_ref[1]+(-0.579046430636)*x_ref[2]**o+(0.587623604941)*x_ref[2]
        ref[4,0,0]=(-0.167582584882)*x_ref[0]**o+(0.524904122243)*x_ref[0]+(0.665109348362)*x_ref[1]**o+(-0.093655366509)*x_ref[1]+(0.704404656254)*x_ref[2]**o+(-0.402602441741)*x_ref[2]
        ref[4,0,1]=(0.633854964824)*x_ref[0]**o+(-0.836385840194)*x_ref[0]+(-0.0995437996628)*x_ref[1]**o+(-0.200285338869)*x_ref[1]+(0.535023783881)*x_ref[2]**o+(0.884702543593)*x_ref[2]
        ref[4,1,0]=(-0.953619378208)*x_ref[0]**o+(-0.390518815669)*x_ref[0]+(-0.0590889178236)*x_ref[1]**o+(-0.997225825019)*x_ref[1]+(0.0520418736402)*x_ref[2]**o+(0.484749465262)*x_ref[2]
        ref[4,1,1]=(-0.440389854756)*x_ref[0]**o+(0.15520431778)*x_ref[0]+(0.346839174513)*x_ref[1]**o+(0.0172676873159)*x_ref[1]+(0.86275919224)*x_ref[2]**o+(0.568069148992)*x_ref[2]
        ref[5,0,0]=(-0.303687941217)*x_ref[0]**o+(0.516115915687)*x_ref[0]+(-0.750276597519)*x_ref[1]**o+(0.734039058778)*x_ref[1]+(-0.624499210252)*x_ref[2]**o+(-0.926338135421)*x_ref[2]
        ref[5,0,1]=(-0.472230777895)*x_ref[0]**o+(-0.557827728373)*x_ref[0]+(-0.403783023157)*x_ref[1]**o+(0.70433336955)*x_ref[1]+(-0.0211737454531)*x_ref[2]**o+(0.0774406125836)*x_ref[2]
        ref[5,1,0]=(0.718149875204)*x_ref[0]**o+(-0.183000739787)*x_ref[0]+(-0.655724252373)*x_ref[1]**o+(-0.634941710776)*x_ref[1]+(-0.26300163323)*x_ref[2]**o+(-0.25105424715)*x_ref[2]
        ref[5,1,1]=(0.109601904441)*x_ref[0]**o+(-0.00640764631554)*x_ref[0]+(-0.488035610009)*x_ref[1]**o+(0.909668323769)*x_ref[1]+(0.229692168501)*x_ref[2]**o+(-0.739241309497)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.912518456494)*x[0]**o+(-0.828176151903)*x[0]+(-0.313940298121)*x[1]**o+(-0.850120336531)*x[1]
        arg[0,0,0,1]=(0.14415852783)*x[0]**o+(-0.294155246668)*x[0]+(0.562517248232)*x[1]**o+(-0.204505938286)*x[1]
        arg[0,0,1,0]=(0.555455404368)*x[0]**o+(0.916425343576)*x[0]+(0.547252315693)*x[1]**o+(-0.0992842367479)*x[1]
        arg[0,0,1,1]=(0.603632473928)*x[0]**o+(0.505571549062)*x[0]+(-0.490402921542)*x[1]**o+(0.908624427316)*x[1]
        arg[0,0,2,0]=(0.729542308985)*x[0]**o+(-0.654310868638)*x[0]+(0.0845981276245)*x[1]**o+(-0.95231604686)*x[1]
        arg[0,0,2,1]=(0.0511615606745)*x[0]**o+(-0.939469887975)*x[0]+(-0.206229418155)*x[1]**o+(0.254879844701)*x[1]
        arg[0,1,0,0]=(0.257917344507)*x[0]**o+(0.860729621288)*x[0]+(0.489966193348)*x[1]**o+(0.174897233858)*x[1]
        arg[0,1,0,1]=(0.336588362051)*x[0]**o+(-0.690598635114)*x[0]+(-0.0913735300483)*x[1]**o+(-0.932914439004)*x[1]
        arg[0,1,1,0]=(0.244583093108)*x[0]**o+(-0.827768167758)*x[0]+(0.573065359412)*x[1]**o+(0.176685515596)*x[1]
        arg[0,1,1,1]=(-0.520550726127)*x[0]**o+(-0.605243177286)*x[0]+(0.130549556226)*x[1]**o+(0.750019884067)*x[1]
        arg[0,1,2,0]=(-0.680201627614)*x[0]**o+(0.679925324882)*x[0]+(-0.060408604263)*x[1]**o+(0.568125689359)*x[1]
        arg[0,1,2,1]=(-0.111908859432)*x[0]**o+(0.256521037018)*x[0]+(0.332050616611)*x[1]**o+(-0.0500016397683)*x[1]
        arg[0,2,0,0]=(-0.484953933663)*x[0]**o+(0.628032211831)*x[0]+(-0.63297141454)*x[1]**o+(0.141016662755)*x[1]
        arg[0,2,0,1]=(-0.463920680465)*x[0]**o+(-0.868825520511)*x[0]+(-0.36594425006)*x[1]**o+(-0.823052216578)*x[1]
        arg[0,2,1,0]=(0.972337802357)*x[0]**o+(0.943605157343)*x[0]+(-0.113950717844)*x[1]**o+(-0.839596452897)*x[1]
        arg[0,2,1,1]=(-0.295562401302)*x[0]**o+(0.735770621017)*x[0]+(-0.183838912348)*x[1]**o+(-0.929365380252)*x[1]
        arg[0,2,2,0]=(0.912886298388)*x[0]**o+(0.847507982065)*x[0]+(-0.341306036323)*x[1]**o+(0.393627025978)*x[1]
        arg[0,2,2,1]=(-0.769650638386)*x[0]**o+(-0.619672356488)*x[0]+(-0.126894288906)*x[1]**o+(0.231547632139)*x[1]
        arg[0,3,0,0]=(-0.884442976834)*x[0]**o+(-0.0194505760676)*x[0]+(-0.338764675072)*x[1]**o+(0.723481600171)*x[1]
        arg[0,3,0,1]=(0.655973555091)*x[0]**o+(-0.556230649412)*x[0]+(-0.850356832441)*x[1]**o+(0.115267314853)*x[1]
        arg[0,3,1,0]=(0.659628443559)*x[0]**o+(-0.586677822491)*x[0]+(0.135930588518)*x[1]**o+(0.206866895948)*x[1]
        arg[0,3,1,1]=(-0.700258957557)*x[0]**o+(-0.699774198764)*x[0]+(-0.294009479186)*x[1]**o+(0.848159666503)*x[1]
        arg[0,3,2,0]=(-0.927501754488)*x[0]**o+(-0.981178071331)*x[0]+(0.614074194451)*x[1]**o+(0.711549573086)*x[1]
        arg[0,3,2,1]=(0.312434473198)*x[0]**o+(-0.241506125991)*x[0]+(-0.0840979088912)*x[1]**o+(0.902948463117)*x[1]
        arg[0,4,0,0]=(-0.892269860242)*x[0]**o+(0.823434435855)*x[0]+(0.50499383365)*x[1]**o+(0.763736541034)*x[1]
        arg[0,4,0,1]=(-0.606959691191)*x[0]**o+(0.945562536791)*x[0]+(-0.46271929481)*x[1]**o+(0.929514432096)*x[1]
        arg[0,4,1,0]=(0.953382685939)*x[0]**o+(0.553766359805)*x[0]+(-0.191004010293)*x[1]**o+(0.556369491698)*x[1]
        arg[0,4,1,1]=(-0.269402649165)*x[0]**o+(-0.158926232076)*x[0]+(-0.390038749681)*x[1]**o+(0.553655192666)*x[1]
        arg[0,4,2,0]=(0.0785586372006)*x[0]**o+(-0.778452221862)*x[0]+(-0.927627273292)*x[1]**o+(-0.187742405246)*x[1]
        arg[0,4,2,1]=(-0.965329945447)*x[0]**o+(0.304539006089)*x[0]+(0.213031689568)*x[1]**o+(-0.2807790149)*x[1]
        arg[1,0,0,0]=(0.981534698569)*x[0]**o+(0.466451685806)*x[0]+(0.845322931856)*x[1]**o+(0.582422032334)*x[1]
        arg[1,0,0,1]=(-0.0645798740394)*x[0]**o+(0.0475214206619)*x[0]+(-0.687759433076)*x[1]**o+(0.403440938285)*x[1]
        arg[1,0,1,0]=(-0.317887773392)*x[0]**o+(0.59101163416)*x[0]+(0.658626263291)*x[1]**o+(-0.618167594137)*x[1]
        arg[1,0,1,1]=(-0.314093092813)*x[0]**o+(0.699933899942)*x[0]+(0.824195260822)*x[1]**o+(-0.42875327345)*x[1]
        arg[1,0,2,0]=(-0.254254956887)*x[0]**o+(-0.912137080163)*x[0]+(0.360457121452)*x[1]**o+(-0.902596912588)*x[1]
        arg[1,0,2,1]=(-0.577214593683)*x[0]**o+(-0.419495663008)*x[0]+(-0.803459783331)*x[1]**o+(0.490885513568)*x[1]
        arg[1,1,0,0]=(0.477947766892)*x[0]**o+(-0.853046971581)*x[0]+(0.576044041601)*x[1]**o+(-0.447007701252)*x[1]
        arg[1,1,0,1]=(0.5886933409)*x[0]**o+(0.253650778351)*x[0]+(-0.198070684431)*x[1]**o+(0.189479364528)*x[1]
        arg[1,1,1,0]=(-0.528744478872)*x[0]**o+(-0.504072059557)*x[0]+(-0.0161832474015)*x[1]**o+(0.639592055966)*x[1]
        arg[1,1,1,1]=(-0.581844827169)*x[0]**o+(0.997883224538)*x[0]+(0.9354364489)*x[1]**o+(0.884060785136)*x[1]
        arg[1,1,2,0]=(0.858226955817)*x[0]**o+(0.707261698312)*x[0]+(-0.738309111076)*x[1]**o+(-0.644655976448)*x[1]
        arg[1,1,2,1]=(-0.254976445964)*x[0]**o+(-0.624982185158)*x[0]+(0.304854486996)*x[1]**o+(0.597089811682)*x[1]
        arg[1,2,0,0]=(-0.625190759215)*x[0]**o+(-0.267820197905)*x[0]+(-0.721452024377)*x[1]**o+(-0.830984245819)*x[1]
        arg[1,2,0,1]=(-0.972257608766)*x[0]**o+(0.981368890544)*x[0]+(-0.925886510064)*x[1]**o+(-0.337107356862)*x[1]
        arg[1,2,1,0]=(-0.178932954207)*x[0]**o+(-0.941789507374)*x[0]+(0.360726744794)*x[1]**o+(0.347621863123)*x[1]
        arg[1,2,1,1]=(-0.59790695611)*x[0]**o+(0.552509013166)*x[0]+(0.592412443998)*x[1]**o+(0.105369637971)*x[1]
        arg[1,2,2,0]=(0.300713195225)*x[0]**o+(0.814257503574)*x[0]+(-0.281407483492)*x[1]**o+(-0.803431681095)*x[1]
        arg[1,2,2,1]=(0.80113022942)*x[0]**o+(0.749336171665)*x[0]+(-0.0842928982283)*x[1]**o+(-0.974041792058)*x[1]
        arg[1,3,0,0]=(0.533118876666)*x[0]**o+(-0.00888178987567)*x[0]+(0.372869213762)*x[1]**o+(-0.652513683144)*x[1]
        arg[1,3,0,1]=(-0.295090222537)*x[0]**o+(0.772749124303)*x[0]+(-0.00857029468729)*x[1]**o+(0.927794431125)*x[1]
        arg[1,3,1,0]=(-0.152537465441)*x[0]**o+(0.154900534979)*x[0]+(-0.488743959543)*x[1]**o+(-0.0318497199232)*x[1]
        arg[1,3,1,1]=(-0.754333787921)*x[0]**o+(-0.0583007358701)*x[0]+(-0.050921106891)*x[1]**o+(0.161426452156)*x[1]
        arg[1,3,2,0]=(0.387357469451)*x[0]**o+(-0.213640305885)*x[0]+(-0.844609663023)*x[1]**o+(0.0761914594192)*x[1]
        arg[1,3,2,1]=(0.36964862459)*x[0]**o+(0.887888401952)*x[0]+(-0.933930511093)*x[1]**o+(0.769046614932)*x[1]
        arg[1,4,0,0]=(0.680878887266)*x[0]**o+(0.62930033679)*x[0]+(0.894793542585)*x[1]**o+(-0.636881609067)*x[1]
        arg[1,4,0,1]=(-0.346053756598)*x[0]**o+(0.225254433923)*x[0]+(-0.861052985621)*x[1]**o+(0.415123108136)*x[1]
        arg[1,4,1,0]=(-0.997758408183)*x[0]**o+(-0.265055614341)*x[0]+(-0.7568881617)*x[1]**o+(0.95144584777)*x[1]
        arg[1,4,1,1]=(0.227777912008)*x[0]**o+(-0.83647660817)*x[0]+(-0.275733410165)*x[1]**o+(-0.971560248343)*x[1]
        arg[1,4,2,0]=(0.718803859114)*x[0]**o+(-0.978966510682)*x[0]+(0.463102438649)*x[1]**o+(0.987805610534)*x[1]
        arg[1,4,2,1]=(-0.107543838917)*x[0]**o+(-0.362684018694)*x[0]+(-0.310926939373)*x[1]**o+(-0.571571044712)*x[1]
        arg[2,0,0,0]=(-0.475977416182)*x[0]**o+(-0.627173002609)*x[0]+(-0.369962970984)*x[1]**o+(-0.245077302537)*x[1]
        arg[2,0,0,1]=(0.913949611751)*x[0]**o+(0.645611424617)*x[0]+(0.639304734076)*x[1]**o+(-0.12171441998)*x[1]
        arg[2,0,1,0]=(0.251434098263)*x[0]**o+(-0.119898861407)*x[0]+(-0.365614523709)*x[1]**o+(-0.802928738296)*x[1]
        arg[2,0,1,1]=(0.756390927006)*x[0]**o+(0.246578936199)*x[0]+(0.714968085461)*x[1]**o+(-0.929428808096)*x[1]
        arg[2,0,2,0]=(-0.516888361447)*x[0]**o+(0.0344339250886)*x[0]+(-0.383709788774)*x[1]**o+(-0.437344417623)*x[1]
        arg[2,0,2,1]=(-0.219307162317)*x[0]**o+(0.704909718842)*x[0]+(-0.340454491855)*x[1]**o+(-0.933866231773)*x[1]
        arg[2,1,0,0]=(-0.0618967891627)*x[0]**o+(-0.0590712030789)*x[0]+(0.116776433077)*x[1]**o+(0.0241239764104)*x[1]
        arg[2,1,0,1]=(-0.802207645638)*x[0]**o+(0.590732740531)*x[0]+(-0.25990272933)*x[1]**o+(-0.257633805416)*x[1]
        arg[2,1,1,0]=(0.827623274379)*x[0]**o+(-0.450310573599)*x[0]+(0.213809775826)*x[1]**o+(-0.905410777787)*x[1]
        arg[2,1,1,1]=(-0.624510830356)*x[0]**o+(-0.705682607787)*x[0]+(0.430105716993)*x[1]**o+(0.975510832521)*x[1]
        arg[2,1,2,0]=(-0.297562223309)*x[0]**o+(-0.99090348994)*x[0]+(0.548589416291)*x[1]**o+(0.591718478583)*x[1]
        arg[2,1,2,1]=(-0.531899919376)*x[0]**o+(-0.227453719581)*x[0]+(-0.131664806128)*x[1]**o+(0.0521895817509)*x[1]
        arg[2,2,0,0]=(-0.592291745786)*x[0]**o+(-0.586879961972)*x[0]+(0.563196552327)*x[1]**o+(0.876989321187)*x[1]
        arg[2,2,0,1]=(-0.273360445369)*x[0]**o+(-0.714740527775)*x[0]+(-0.647699497837)*x[1]**o+(-0.0405089612762)*x[1]
        arg[2,2,1,0]=(-0.536629732224)*x[0]**o+(0.72656373966)*x[0]+(0.986630594165)*x[1]**o+(0.0741630711667)*x[1]
        arg[2,2,1,1]=(0.770746088838)*x[0]**o+(-0.219279833426)*x[0]+(0.875445174779)*x[1]**o+(-0.947889926444)*x[1]
        arg[2,2,2,0]=(-0.41428815119)*x[0]**o+(-0.696166268903)*x[0]+(-0.96838062553)*x[1]**o+(0.781394205899)*x[1]
        arg[2,2,2,1]=(-0.810267960752)*x[0]**o+(-0.873632687702)*x[0]+(0.0131971321611)*x[1]**o+(-0.552767192193)*x[1]
        arg[2,3,0,0]=(-0.216091366291)*x[0]**o+(-0.17824047472)*x[0]+(-0.4644515725)*x[1]**o+(0.135819458957)*x[1]
        arg[2,3,0,1]=(-0.0823616523105)*x[0]**o+(0.33121362151)*x[0]+(0.926719919839)*x[1]**o+(0.663475641636)*x[1]
        arg[2,3,1,0]=(-0.278303066047)*x[0]**o+(-0.457286378314)*x[0]+(0.186792859092)*x[1]**o+(-0.734245612464)*x[1]
        arg[2,3,1,1]=(0.810871459779)*x[0]**o+(-0.706641995324)*x[0]+(-0.287827176091)*x[1]**o+(0.453216090282)*x[1]
        arg[2,3,2,0]=(0.83826221049)*x[0]**o+(0.556887792111)*x[0]+(-0.000231102245991)*x[1]**o+(-0.440730958954)*x[1]
        arg[2,3,2,1]=(-0.455087877907)*x[0]**o+(-0.132259634036)*x[0]+(-0.492728335525)*x[1]**o+(-0.880311317777)*x[1]
        arg[2,4,0,0]=(-0.303141352388)*x[0]**o+(0.426137046804)*x[0]+(0.788934776738)*x[1]**o+(0.157477785179)*x[1]
        arg[2,4,0,1]=(0.603411288614)*x[0]**o+(-0.533561451987)*x[0]+(-0.901259210942)*x[1]**o+(0.547372624746)*x[1]
        arg[2,4,1,0]=(0.69517961739)*x[0]**o+(-0.106282888246)*x[0]+(0.257484565681)*x[1]**o+(0.655749633294)*x[1]
        arg[2,4,1,1]=(-0.072748573118)*x[0]**o+(0.905966560113)*x[0]+(0.433145718123)*x[1]**o+(0.187620401615)*x[1]
        arg[2,4,2,0]=(-0.0280176325954)*x[0]**o+(0.489610131347)*x[0]+(-0.740554922134)*x[1]**o+(-0.85045302096)*x[1]
        arg[2,4,2,1]=(-0.158503180709)*x[0]**o+(0.525929643075)*x[0]+(-0.352022652515)*x[1]**o+(-0.706553104932)*x[1]
        arg[3,0,0,0]=(-0.837316508127)*x[0]**o+(-0.89490181779)*x[0]+(0.510084075291)*x[1]**o+(0.321206310024)*x[1]
        arg[3,0,0,1]=(-0.325038591079)*x[0]**o+(0.140342922753)*x[0]+(-0.937466403004)*x[1]**o+(0.580557383497)*x[1]
        arg[3,0,1,0]=(-0.939713654062)*x[0]**o+(-0.394660244166)*x[0]+(0.762376614557)*x[1]**o+(-0.135565547532)*x[1]
        arg[3,0,1,1]=(0.0275332524464)*x[0]**o+(0.368737701882)*x[0]+(0.983641372218)*x[1]**o+(0.573815516366)*x[1]
        arg[3,0,2,0]=(0.816957570314)*x[0]**o+(-0.357626892562)*x[0]+(-0.368933149642)*x[1]**o+(-0.692502486914)*x[1]
        arg[3,0,2,1]=(-0.372408810926)*x[0]**o+(-0.113036390416)*x[0]+(-0.355880297003)*x[1]**o+(0.755478191481)*x[1]
        arg[3,1,0,0]=(0.0399012900832)*x[0]**o+(-0.879576900086)*x[0]+(-0.660858776134)*x[1]**o+(-0.59837040429)*x[1]
        arg[3,1,0,1]=(0.569397128496)*x[0]**o+(-0.138333417746)*x[0]+(0.257769147727)*x[1]**o+(0.174737141027)*x[1]
        arg[3,1,1,0]=(-0.67938408823)*x[0]**o+(0.862731161981)*x[0]+(-0.792538485264)*x[1]**o+(-0.0805197796157)*x[1]
        arg[3,1,1,1]=(0.0113523229279)*x[0]**o+(0.475393597561)*x[0]+(-0.705748021886)*x[1]**o+(0.00836460135285)*x[1]
        arg[3,1,2,0]=(0.463839509799)*x[0]**o+(-0.12663544643)*x[0]+(0.353158399738)*x[1]**o+(0.674929537597)*x[1]
        arg[3,1,2,1]=(-0.6019896019)*x[0]**o+(-0.638280685565)*x[0]+(0.280047072988)*x[1]**o+(0.569239958138)*x[1]
        arg[3,2,0,0]=(-0.113906500346)*x[0]**o+(-0.493494498501)*x[0]+(-0.686419229162)*x[1]**o+(-0.0727046763516)*x[1]
        arg[3,2,0,1]=(-0.302092112795)*x[0]**o+(-0.74819167)*x[0]+(-0.478908975282)*x[1]**o+(0.430024789627)*x[1]
        arg[3,2,1,0]=(0.353535608984)*x[0]**o+(-0.743408549264)*x[0]+(-0.069122881538)*x[1]**o+(-0.0767595089529)*x[1]
        arg[3,2,1,1]=(-0.629946386234)*x[0]**o+(0.851779590894)*x[0]+(0.728577108695)*x[1]**o+(-0.366983040147)*x[1]
        arg[3,2,2,0]=(0.1685167508)*x[0]**o+(0.149999747637)*x[0]+(0.709320430374)*x[1]**o+(-0.280950677222)*x[1]
        arg[3,2,2,1]=(-0.956164974578)*x[0]**o+(0.240460380901)*x[0]+(-0.2820176577)*x[1]**o+(-0.595667156955)*x[1]
        arg[3,3,0,0]=(-0.0711295055938)*x[0]**o+(-0.314513251791)*x[0]+(-0.41924591293)*x[1]**o+(0.0813197780959)*x[1]
        arg[3,3,0,1]=(-0.230592584877)*x[0]**o+(-0.488005144715)*x[0]+(-0.450611759976)*x[1]**o+(-0.0571152769768)*x[1]
        arg[3,3,1,0]=(-0.608255699014)*x[0]**o+(-0.697324108575)*x[0]+(-0.885935900746)*x[1]**o+(-0.884497550016)*x[1]
        arg[3,3,1,1]=(0.857286225292)*x[0]**o+(0.509954550626)*x[0]+(0.182030821433)*x[1]**o+(-0.268076060316)*x[1]
        arg[3,3,2,0]=(-0.77322653479)*x[0]**o+(-0.583013210018)*x[0]+(-0.683867300277)*x[1]**o+(0.754530541251)*x[1]
        arg[3,3,2,1]=(-0.744667985181)*x[0]**o+(0.164165128978)*x[0]+(0.412340155958)*x[1]**o+(-0.38451107118)*x[1]
        arg[3,4,0,0]=(0.216720755049)*x[0]**o+(-0.347803433722)*x[0]+(0.417376268797)*x[1]**o+(0.475787304652)*x[1]
        arg[3,4,0,1]=(-0.737038295974)*x[0]**o+(0.579172519178)*x[0]+(-0.900272993672)*x[1]**o+(0.407018603625)*x[1]
        arg[3,4,1,0]=(-0.898959236474)*x[0]**o+(-0.886636008121)*x[0]+(-0.901972610044)*x[1]**o+(-0.518393782318)*x[1]
        arg[3,4,1,1]=(0.976342884921)*x[0]**o+(-0.273568319493)*x[0]+(-0.332424980846)*x[1]**o+(-0.0146661731139)*x[1]
        arg[3,4,2,0]=(0.794792214027)*x[0]**o+(0.846927984413)*x[0]+(0.535818809709)*x[1]**o+(-0.49324400033)*x[1]
        arg[3,4,2,1]=(0.695766308812)*x[0]**o+(-0.37923153689)*x[0]+(0.0633513026782)*x[1]**o+(0.12334892105)*x[1]
        ref[0,0,0,0]=(0.912518456494)*x_ref[0]**o+(-0.828176151903)*x_ref[0]+(-0.313940298121)*x_ref[1]**o+(-0.850120336531)*x_ref[1]
        ref[0,0,0,1]=(0.14415852783)*x_ref[0]**o+(-0.294155246668)*x_ref[0]+(0.562517248232)*x_ref[1]**o+(-0.204505938286)*x_ref[1]
        ref[0,0,1,0]=(0.555455404368)*x_ref[0]**o+(0.916425343576)*x_ref[0]+(0.547252315693)*x_ref[1]**o+(-0.0992842367479)*x_ref[1]
        ref[0,0,1,1]=(0.603632473928)*x_ref[0]**o+(0.505571549062)*x_ref[0]+(-0.490402921542)*x_ref[1]**o+(0.908624427316)*x_ref[1]
        ref[0,0,2,0]=(0.729542308985)*x_ref[0]**o+(-0.654310868638)*x_ref[0]+(0.0845981276245)*x_ref[1]**o+(-0.95231604686)*x_ref[1]
        ref[0,0,2,1]=(0.0511615606745)*x_ref[0]**o+(-0.939469887975)*x_ref[0]+(-0.206229418155)*x_ref[1]**o+(0.254879844701)*x_ref[1]
        ref[0,1,0,0]=(0.257917344507)*x_ref[0]**o+(0.860729621288)*x_ref[0]+(0.489966193348)*x_ref[1]**o+(0.174897233858)*x_ref[1]
        ref[0,1,0,1]=(0.336588362051)*x_ref[0]**o+(-0.690598635114)*x_ref[0]+(-0.0913735300483)*x_ref[1]**o+(-0.932914439004)*x_ref[1]
        ref[0,1,1,0]=(0.244583093108)*x_ref[0]**o+(-0.827768167758)*x_ref[0]+(0.573065359412)*x_ref[1]**o+(0.176685515596)*x_ref[1]
        ref[0,1,1,1]=(-0.520550726127)*x_ref[0]**o+(-0.605243177286)*x_ref[0]+(0.130549556226)*x_ref[1]**o+(0.750019884067)*x_ref[1]
        ref[0,1,2,0]=(-0.680201627614)*x_ref[0]**o+(0.679925324882)*x_ref[0]+(-0.060408604263)*x_ref[1]**o+(0.568125689359)*x_ref[1]
        ref[0,1,2,1]=(-0.111908859432)*x_ref[0]**o+(0.256521037018)*x_ref[0]+(0.332050616611)*x_ref[1]**o+(-0.0500016397683)*x_ref[1]
        ref[0,2,0,0]=(-0.484953933663)*x_ref[0]**o+(0.628032211831)*x_ref[0]+(-0.63297141454)*x_ref[1]**o+(0.141016662755)*x_ref[1]
        ref[0,2,0,1]=(-0.463920680465)*x_ref[0]**o+(-0.868825520511)*x_ref[0]+(-0.36594425006)*x_ref[1]**o+(-0.823052216578)*x_ref[1]
        ref[0,2,1,0]=(0.972337802357)*x_ref[0]**o+(0.943605157343)*x_ref[0]+(-0.113950717844)*x_ref[1]**o+(-0.839596452897)*x_ref[1]
        ref[0,2,1,1]=(-0.295562401302)*x_ref[0]**o+(0.735770621017)*x_ref[0]+(-0.183838912348)*x_ref[1]**o+(-0.929365380252)*x_ref[1]
        ref[0,2,2,0]=(0.912886298388)*x_ref[0]**o+(0.847507982065)*x_ref[0]+(-0.341306036323)*x_ref[1]**o+(0.393627025978)*x_ref[1]
        ref[0,2,2,1]=(-0.769650638386)*x_ref[0]**o+(-0.619672356488)*x_ref[0]+(-0.126894288906)*x_ref[1]**o+(0.231547632139)*x_ref[1]
        ref[0,3,0,0]=(-0.884442976834)*x_ref[0]**o+(-0.0194505760676)*x_ref[0]+(-0.338764675072)*x_ref[1]**o+(0.723481600171)*x_ref[1]
        ref[0,3,0,1]=(0.655973555091)*x_ref[0]**o+(-0.556230649412)*x_ref[0]+(-0.850356832441)*x_ref[1]**o+(0.115267314853)*x_ref[1]
        ref[0,3,1,0]=(0.659628443559)*x_ref[0]**o+(-0.586677822491)*x_ref[0]+(0.135930588518)*x_ref[1]**o+(0.206866895948)*x_ref[1]
        ref[0,3,1,1]=(-0.700258957557)*x_ref[0]**o+(-0.699774198764)*x_ref[0]+(-0.294009479186)*x_ref[1]**o+(0.848159666503)*x_ref[1]
        ref[0,3,2,0]=(-0.927501754488)*x_ref[0]**o+(-0.981178071331)*x_ref[0]+(0.614074194451)*x_ref[1]**o+(0.711549573086)*x_ref[1]
        ref[0,3,2,1]=(0.312434473198)*x_ref[0]**o+(-0.241506125991)*x_ref[0]+(-0.0840979088912)*x_ref[1]**o+(0.902948463117)*x_ref[1]
        ref[0,4,0,0]=(-0.892269860242)*x_ref[0]**o+(0.823434435855)*x_ref[0]+(0.50499383365)*x_ref[1]**o+(0.763736541034)*x_ref[1]
        ref[0,4,0,1]=(-0.606959691191)*x_ref[0]**o+(0.945562536791)*x_ref[0]+(-0.46271929481)*x_ref[1]**o+(0.929514432096)*x_ref[1]
        ref[0,4,1,0]=(0.953382685939)*x_ref[0]**o+(0.553766359805)*x_ref[0]+(-0.191004010293)*x_ref[1]**o+(0.556369491698)*x_ref[1]
        ref[0,4,1,1]=(-0.269402649165)*x_ref[0]**o+(-0.158926232076)*x_ref[0]+(-0.390038749681)*x_ref[1]**o+(0.553655192666)*x_ref[1]
        ref[0,4,2,0]=(0.0785586372006)*x_ref[0]**o+(-0.778452221862)*x_ref[0]+(-0.927627273292)*x_ref[1]**o+(-0.187742405246)*x_ref[1]
        ref[0,4,2,1]=(-0.965329945447)*x_ref[0]**o+(0.304539006089)*x_ref[0]+(0.213031689568)*x_ref[1]**o+(-0.2807790149)*x_ref[1]
        ref[1,0,0,0]=(0.981534698569)*x_ref[0]**o+(0.466451685806)*x_ref[0]+(0.845322931856)*x_ref[1]**o+(0.582422032334)*x_ref[1]
        ref[1,0,0,1]=(-0.0645798740394)*x_ref[0]**o+(0.0475214206619)*x_ref[0]+(-0.687759433076)*x_ref[1]**o+(0.403440938285)*x_ref[1]
        ref[1,0,1,0]=(-0.317887773392)*x_ref[0]**o+(0.59101163416)*x_ref[0]+(0.658626263291)*x_ref[1]**o+(-0.618167594137)*x_ref[1]
        ref[1,0,1,1]=(-0.314093092813)*x_ref[0]**o+(0.699933899942)*x_ref[0]+(0.824195260822)*x_ref[1]**o+(-0.42875327345)*x_ref[1]
        ref[1,0,2,0]=(-0.254254956887)*x_ref[0]**o+(-0.912137080163)*x_ref[0]+(0.360457121452)*x_ref[1]**o+(-0.902596912588)*x_ref[1]
        ref[1,0,2,1]=(-0.577214593683)*x_ref[0]**o+(-0.419495663008)*x_ref[0]+(-0.803459783331)*x_ref[1]**o+(0.490885513568)*x_ref[1]
        ref[1,1,0,0]=(0.477947766892)*x_ref[0]**o+(-0.853046971581)*x_ref[0]+(0.576044041601)*x_ref[1]**o+(-0.447007701252)*x_ref[1]
        ref[1,1,0,1]=(0.5886933409)*x_ref[0]**o+(0.253650778351)*x_ref[0]+(-0.198070684431)*x_ref[1]**o+(0.189479364528)*x_ref[1]
        ref[1,1,1,0]=(-0.528744478872)*x_ref[0]**o+(-0.504072059557)*x_ref[0]+(-0.0161832474015)*x_ref[1]**o+(0.639592055966)*x_ref[1]
        ref[1,1,1,1]=(-0.581844827169)*x_ref[0]**o+(0.997883224538)*x_ref[0]+(0.9354364489)*x_ref[1]**o+(0.884060785136)*x_ref[1]
        ref[1,1,2,0]=(0.858226955817)*x_ref[0]**o+(0.707261698312)*x_ref[0]+(-0.738309111076)*x_ref[1]**o+(-0.644655976448)*x_ref[1]
        ref[1,1,2,1]=(-0.254976445964)*x_ref[0]**o+(-0.624982185158)*x_ref[0]+(0.304854486996)*x_ref[1]**o+(0.597089811682)*x_ref[1]
        ref[1,2,0,0]=(-0.625190759215)*x_ref[0]**o+(-0.267820197905)*x_ref[0]+(-0.721452024377)*x_ref[1]**o+(-0.830984245819)*x_ref[1]
        ref[1,2,0,1]=(-0.972257608766)*x_ref[0]**o+(0.981368890544)*x_ref[0]+(-0.925886510064)*x_ref[1]**o+(-0.337107356862)*x_ref[1]
        ref[1,2,1,0]=(-0.178932954207)*x_ref[0]**o+(-0.941789507374)*x_ref[0]+(0.360726744794)*x_ref[1]**o+(0.347621863123)*x_ref[1]
        ref[1,2,1,1]=(-0.59790695611)*x_ref[0]**o+(0.552509013166)*x_ref[0]+(0.592412443998)*x_ref[1]**o+(0.105369637971)*x_ref[1]
        ref[1,2,2,0]=(0.300713195225)*x_ref[0]**o+(0.814257503574)*x_ref[0]+(-0.281407483492)*x_ref[1]**o+(-0.803431681095)*x_ref[1]
        ref[1,2,2,1]=(0.80113022942)*x_ref[0]**o+(0.749336171665)*x_ref[0]+(-0.0842928982283)*x_ref[1]**o+(-0.974041792058)*x_ref[1]
        ref[1,3,0,0]=(0.533118876666)*x_ref[0]**o+(-0.00888178987567)*x_ref[0]+(0.372869213762)*x_ref[1]**o+(-0.652513683144)*x_ref[1]
        ref[1,3,0,1]=(-0.295090222537)*x_ref[0]**o+(0.772749124303)*x_ref[0]+(-0.00857029468729)*x_ref[1]**o+(0.927794431125)*x_ref[1]
        ref[1,3,1,0]=(-0.152537465441)*x_ref[0]**o+(0.154900534979)*x_ref[0]+(-0.488743959543)*x_ref[1]**o+(-0.0318497199232)*x_ref[1]
        ref[1,3,1,1]=(-0.754333787921)*x_ref[0]**o+(-0.0583007358701)*x_ref[0]+(-0.050921106891)*x_ref[1]**o+(0.161426452156)*x_ref[1]
        ref[1,3,2,0]=(0.387357469451)*x_ref[0]**o+(-0.213640305885)*x_ref[0]+(-0.844609663023)*x_ref[1]**o+(0.0761914594192)*x_ref[1]
        ref[1,3,2,1]=(0.36964862459)*x_ref[0]**o+(0.887888401952)*x_ref[0]+(-0.933930511093)*x_ref[1]**o+(0.769046614932)*x_ref[1]
        ref[1,4,0,0]=(0.680878887266)*x_ref[0]**o+(0.62930033679)*x_ref[0]+(0.894793542585)*x_ref[1]**o+(-0.636881609067)*x_ref[1]
        ref[1,4,0,1]=(-0.346053756598)*x_ref[0]**o+(0.225254433923)*x_ref[0]+(-0.861052985621)*x_ref[1]**o+(0.415123108136)*x_ref[1]
        ref[1,4,1,0]=(-0.997758408183)*x_ref[0]**o+(-0.265055614341)*x_ref[0]+(-0.7568881617)*x_ref[1]**o+(0.95144584777)*x_ref[1]
        ref[1,4,1,1]=(0.227777912008)*x_ref[0]**o+(-0.83647660817)*x_ref[0]+(-0.275733410165)*x_ref[1]**o+(-0.971560248343)*x_ref[1]
        ref[1,4,2,0]=(0.718803859114)*x_ref[0]**o+(-0.978966510682)*x_ref[0]+(0.463102438649)*x_ref[1]**o+(0.987805610534)*x_ref[1]
        ref[1,4,2,1]=(-0.107543838917)*x_ref[0]**o+(-0.362684018694)*x_ref[0]+(-0.310926939373)*x_ref[1]**o+(-0.571571044712)*x_ref[1]
        ref[2,0,0,0]=(-0.475977416182)*x_ref[0]**o+(-0.627173002609)*x_ref[0]+(-0.369962970984)*x_ref[1]**o+(-0.245077302537)*x_ref[1]
        ref[2,0,0,1]=(0.913949611751)*x_ref[0]**o+(0.645611424617)*x_ref[0]+(0.639304734076)*x_ref[1]**o+(-0.12171441998)*x_ref[1]
        ref[2,0,1,0]=(0.251434098263)*x_ref[0]**o+(-0.119898861407)*x_ref[0]+(-0.365614523709)*x_ref[1]**o+(-0.802928738296)*x_ref[1]
        ref[2,0,1,1]=(0.756390927006)*x_ref[0]**o+(0.246578936199)*x_ref[0]+(0.714968085461)*x_ref[1]**o+(-0.929428808096)*x_ref[1]
        ref[2,0,2,0]=(-0.516888361447)*x_ref[0]**o+(0.0344339250886)*x_ref[0]+(-0.383709788774)*x_ref[1]**o+(-0.437344417623)*x_ref[1]
        ref[2,0,2,1]=(-0.219307162317)*x_ref[0]**o+(0.704909718842)*x_ref[0]+(-0.340454491855)*x_ref[1]**o+(-0.933866231773)*x_ref[1]
        ref[2,1,0,0]=(-0.0618967891627)*x_ref[0]**o+(-0.0590712030789)*x_ref[0]+(0.116776433077)*x_ref[1]**o+(0.0241239764104)*x_ref[1]
        ref[2,1,0,1]=(-0.802207645638)*x_ref[0]**o+(0.590732740531)*x_ref[0]+(-0.25990272933)*x_ref[1]**o+(-0.257633805416)*x_ref[1]
        ref[2,1,1,0]=(0.827623274379)*x_ref[0]**o+(-0.450310573599)*x_ref[0]+(0.213809775826)*x_ref[1]**o+(-0.905410777787)*x_ref[1]
        ref[2,1,1,1]=(-0.624510830356)*x_ref[0]**o+(-0.705682607787)*x_ref[0]+(0.430105716993)*x_ref[1]**o+(0.975510832521)*x_ref[1]
        ref[2,1,2,0]=(-0.297562223309)*x_ref[0]**o+(-0.99090348994)*x_ref[0]+(0.548589416291)*x_ref[1]**o+(0.591718478583)*x_ref[1]
        ref[2,1,2,1]=(-0.531899919376)*x_ref[0]**o+(-0.227453719581)*x_ref[0]+(-0.131664806128)*x_ref[1]**o+(0.0521895817509)*x_ref[1]
        ref[2,2,0,0]=(-0.592291745786)*x_ref[0]**o+(-0.586879961972)*x_ref[0]+(0.563196552327)*x_ref[1]**o+(0.876989321187)*x_ref[1]
        ref[2,2,0,1]=(-0.273360445369)*x_ref[0]**o+(-0.714740527775)*x_ref[0]+(-0.647699497837)*x_ref[1]**o+(-0.0405089612762)*x_ref[1]
        ref[2,2,1,0]=(-0.536629732224)*x_ref[0]**o+(0.72656373966)*x_ref[0]+(0.986630594165)*x_ref[1]**o+(0.0741630711667)*x_ref[1]
        ref[2,2,1,1]=(0.770746088838)*x_ref[0]**o+(-0.219279833426)*x_ref[0]+(0.875445174779)*x_ref[1]**o+(-0.947889926444)*x_ref[1]
        ref[2,2,2,0]=(-0.41428815119)*x_ref[0]**o+(-0.696166268903)*x_ref[0]+(-0.96838062553)*x_ref[1]**o+(0.781394205899)*x_ref[1]
        ref[2,2,2,1]=(-0.810267960752)*x_ref[0]**o+(-0.873632687702)*x_ref[0]+(0.0131971321611)*x_ref[1]**o+(-0.552767192193)*x_ref[1]
        ref[2,3,0,0]=(-0.216091366291)*x_ref[0]**o+(-0.17824047472)*x_ref[0]+(-0.4644515725)*x_ref[1]**o+(0.135819458957)*x_ref[1]
        ref[2,3,0,1]=(-0.0823616523105)*x_ref[0]**o+(0.33121362151)*x_ref[0]+(0.926719919839)*x_ref[1]**o+(0.663475641636)*x_ref[1]
        ref[2,3,1,0]=(-0.278303066047)*x_ref[0]**o+(-0.457286378314)*x_ref[0]+(0.186792859092)*x_ref[1]**o+(-0.734245612464)*x_ref[1]
        ref[2,3,1,1]=(0.810871459779)*x_ref[0]**o+(-0.706641995324)*x_ref[0]+(-0.287827176091)*x_ref[1]**o+(0.453216090282)*x_ref[1]
        ref[2,3,2,0]=(0.83826221049)*x_ref[0]**o+(0.556887792111)*x_ref[0]+(-0.000231102245991)*x_ref[1]**o+(-0.440730958954)*x_ref[1]
        ref[2,3,2,1]=(-0.455087877907)*x_ref[0]**o+(-0.132259634036)*x_ref[0]+(-0.492728335525)*x_ref[1]**o+(-0.880311317777)*x_ref[1]
        ref[2,4,0,0]=(-0.303141352388)*x_ref[0]**o+(0.426137046804)*x_ref[0]+(0.788934776738)*x_ref[1]**o+(0.157477785179)*x_ref[1]
        ref[2,4,0,1]=(0.603411288614)*x_ref[0]**o+(-0.533561451987)*x_ref[0]+(-0.901259210942)*x_ref[1]**o+(0.547372624746)*x_ref[1]
        ref[2,4,1,0]=(0.69517961739)*x_ref[0]**o+(-0.106282888246)*x_ref[0]+(0.257484565681)*x_ref[1]**o+(0.655749633294)*x_ref[1]
        ref[2,4,1,1]=(-0.072748573118)*x_ref[0]**o+(0.905966560113)*x_ref[0]+(0.433145718123)*x_ref[1]**o+(0.187620401615)*x_ref[1]
        ref[2,4,2,0]=(-0.0280176325954)*x_ref[0]**o+(0.489610131347)*x_ref[0]+(-0.740554922134)*x_ref[1]**o+(-0.85045302096)*x_ref[1]
        ref[2,4,2,1]=(-0.158503180709)*x_ref[0]**o+(0.525929643075)*x_ref[0]+(-0.352022652515)*x_ref[1]**o+(-0.706553104932)*x_ref[1]
        ref[3,0,0,0]=(-0.837316508127)*x_ref[0]**o+(-0.89490181779)*x_ref[0]+(0.510084075291)*x_ref[1]**o+(0.321206310024)*x_ref[1]
        ref[3,0,0,1]=(-0.325038591079)*x_ref[0]**o+(0.140342922753)*x_ref[0]+(-0.937466403004)*x_ref[1]**o+(0.580557383497)*x_ref[1]
        ref[3,0,1,0]=(-0.939713654062)*x_ref[0]**o+(-0.394660244166)*x_ref[0]+(0.762376614557)*x_ref[1]**o+(-0.135565547532)*x_ref[1]
        ref[3,0,1,1]=(0.0275332524464)*x_ref[0]**o+(0.368737701882)*x_ref[0]+(0.983641372218)*x_ref[1]**o+(0.573815516366)*x_ref[1]
        ref[3,0,2,0]=(0.816957570314)*x_ref[0]**o+(-0.357626892562)*x_ref[0]+(-0.368933149642)*x_ref[1]**o+(-0.692502486914)*x_ref[1]
        ref[3,0,2,1]=(-0.372408810926)*x_ref[0]**o+(-0.113036390416)*x_ref[0]+(-0.355880297003)*x_ref[1]**o+(0.755478191481)*x_ref[1]
        ref[3,1,0,0]=(0.0399012900832)*x_ref[0]**o+(-0.879576900086)*x_ref[0]+(-0.660858776134)*x_ref[1]**o+(-0.59837040429)*x_ref[1]
        ref[3,1,0,1]=(0.569397128496)*x_ref[0]**o+(-0.138333417746)*x_ref[0]+(0.257769147727)*x_ref[1]**o+(0.174737141027)*x_ref[1]
        ref[3,1,1,0]=(-0.67938408823)*x_ref[0]**o+(0.862731161981)*x_ref[0]+(-0.792538485264)*x_ref[1]**o+(-0.0805197796157)*x_ref[1]
        ref[3,1,1,1]=(0.0113523229279)*x_ref[0]**o+(0.475393597561)*x_ref[0]+(-0.705748021886)*x_ref[1]**o+(0.00836460135285)*x_ref[1]
        ref[3,1,2,0]=(0.463839509799)*x_ref[0]**o+(-0.12663544643)*x_ref[0]+(0.353158399738)*x_ref[1]**o+(0.674929537597)*x_ref[1]
        ref[3,1,2,1]=(-0.6019896019)*x_ref[0]**o+(-0.638280685565)*x_ref[0]+(0.280047072988)*x_ref[1]**o+(0.569239958138)*x_ref[1]
        ref[3,2,0,0]=(-0.113906500346)*x_ref[0]**o+(-0.493494498501)*x_ref[0]+(-0.686419229162)*x_ref[1]**o+(-0.0727046763516)*x_ref[1]
        ref[3,2,0,1]=(-0.302092112795)*x_ref[0]**o+(-0.74819167)*x_ref[0]+(-0.478908975282)*x_ref[1]**o+(0.430024789627)*x_ref[1]
        ref[3,2,1,0]=(0.353535608984)*x_ref[0]**o+(-0.743408549264)*x_ref[0]+(-0.069122881538)*x_ref[1]**o+(-0.0767595089529)*x_ref[1]
        ref[3,2,1,1]=(-0.629946386234)*x_ref[0]**o+(0.851779590894)*x_ref[0]+(0.728577108695)*x_ref[1]**o+(-0.366983040147)*x_ref[1]
        ref[3,2,2,0]=(0.1685167508)*x_ref[0]**o+(0.149999747637)*x_ref[0]+(0.709320430374)*x_ref[1]**o+(-0.280950677222)*x_ref[1]
        ref[3,2,2,1]=(-0.956164974578)*x_ref[0]**o+(0.240460380901)*x_ref[0]+(-0.2820176577)*x_ref[1]**o+(-0.595667156955)*x_ref[1]
        ref[3,3,0,0]=(-0.0711295055938)*x_ref[0]**o+(-0.314513251791)*x_ref[0]+(-0.41924591293)*x_ref[1]**o+(0.0813197780959)*x_ref[1]
        ref[3,3,0,1]=(-0.230592584877)*x_ref[0]**o+(-0.488005144715)*x_ref[0]+(-0.450611759976)*x_ref[1]**o+(-0.0571152769768)*x_ref[1]
        ref[3,3,1,0]=(-0.608255699014)*x_ref[0]**o+(-0.697324108575)*x_ref[0]+(-0.885935900746)*x_ref[1]**o+(-0.884497550016)*x_ref[1]
        ref[3,3,1,1]=(0.857286225292)*x_ref[0]**o+(0.509954550626)*x_ref[0]+(0.182030821433)*x_ref[1]**o+(-0.268076060316)*x_ref[1]
        ref[3,3,2,0]=(-0.77322653479)*x_ref[0]**o+(-0.583013210018)*x_ref[0]+(-0.683867300277)*x_ref[1]**o+(0.754530541251)*x_ref[1]
        ref[3,3,2,1]=(-0.744667985181)*x_ref[0]**o+(0.164165128978)*x_ref[0]+(0.412340155958)*x_ref[1]**o+(-0.38451107118)*x_ref[1]
        ref[3,4,0,0]=(0.216720755049)*x_ref[0]**o+(-0.347803433722)*x_ref[0]+(0.417376268797)*x_ref[1]**o+(0.475787304652)*x_ref[1]
        ref[3,4,0,1]=(-0.737038295974)*x_ref[0]**o+(0.579172519178)*x_ref[0]+(-0.900272993672)*x_ref[1]**o+(0.407018603625)*x_ref[1]
        ref[3,4,1,0]=(-0.898959236474)*x_ref[0]**o+(-0.886636008121)*x_ref[0]+(-0.901972610044)*x_ref[1]**o+(-0.518393782318)*x_ref[1]
        ref[3,4,1,1]=(0.976342884921)*x_ref[0]**o+(-0.273568319493)*x_ref[0]+(-0.332424980846)*x_ref[1]**o+(-0.0146661731139)*x_ref[1]
        ref[3,4,2,0]=(0.794792214027)*x_ref[0]**o+(0.846927984413)*x_ref[0]+(0.535818809709)*x_ref[1]**o+(-0.49324400033)*x_ref[1]
        ref[3,4,2,1]=(0.695766308812)*x_ref[0]**o+(-0.37923153689)*x_ref[0]+(0.0633513026782)*x_ref[1]**o+(0.12334892105)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.401901998082)*x[0]**o+(-0.805091116808)*x[0]+(0.997659882792)*x[1]**o+(-0.953957389492)*x[1]+(-0.450185408289)*x[2]**o+(0.636472207773)*x[2]
        arg[0,0,0,1]=(-0.786386817128)*x[0]**o+(-0.882559775393)*x[0]+(0.117871400819)*x[1]**o+(-0.625922274774)*x[1]+(0.410431886931)*x[2]**o+(-0.979356501067)*x[2]
        arg[0,0,1,0]=(0.686025409275)*x[0]**o+(-0.19806907659)*x[0]+(-0.710420224449)*x[1]**o+(-0.839236953983)*x[1]+(-0.79971896093)*x[2]**o+(0.378944111319)*x[2]
        arg[0,0,1,1]=(0.771205112484)*x[0]**o+(-0.007083192859)*x[0]+(0.672918588001)*x[1]**o+(-0.351652495793)*x[1]+(-0.841767161187)*x[2]**o+(-0.345642373014)*x[2]
        arg[0,0,2,0]=(0.266930729015)*x[0]**o+(-0.873808152307)*x[0]+(0.206466759135)*x[1]**o+(0.539888649948)*x[1]+(0.17985862191)*x[2]**o+(0.240651460153)*x[2]
        arg[0,0,2,1]=(-0.232569768041)*x[0]**o+(-0.579841454072)*x[0]+(-0.706956104891)*x[1]**o+(0.822863815698)*x[1]+(0.345774991358)*x[2]**o+(-0.0848567668311)*x[2]
        arg[0,1,0,0]=(0.51593851507)*x[0]**o+(-0.233342434349)*x[0]+(-0.92498298026)*x[1]**o+(0.164540234829)*x[1]+(0.356559744499)*x[2]**o+(0.483902511717)*x[2]
        arg[0,1,0,1]=(-0.811509654443)*x[0]**o+(0.73929453189)*x[0]+(0.256205706418)*x[1]**o+(-0.667924913793)*x[1]+(0.266065641257)*x[2]**o+(0.856490258029)*x[2]
        arg[0,1,1,0]=(0.502051590284)*x[0]**o+(0.927283797128)*x[0]+(0.500854521411)*x[1]**o+(-0.486326996369)*x[1]+(-0.327578523685)*x[2]**o+(-0.28106070397)*x[2]
        arg[0,1,1,1]=(0.201740014513)*x[0]**o+(-0.166988241759)*x[0]+(-0.513072955566)*x[1]**o+(0.67179532417)*x[1]+(-0.72135980627)*x[2]**o+(0.763548886467)*x[2]
        arg[0,1,2,0]=(0.3995590338)*x[0]**o+(0.433286929621)*x[0]+(-0.287309453792)*x[1]**o+(-0.0671679443074)*x[1]+(-0.80815399993)*x[2]**o+(-0.398740362114)*x[2]
        arg[0,1,2,1]=(-0.16314640684)*x[0]**o+(-0.698681697367)*x[0]+(-0.11609738418)*x[1]**o+(-0.798465154739)*x[1]+(-0.169052628971)*x[2]**o+(-0.880316962539)*x[2]
        arg[0,2,0,0]=(-0.219731279442)*x[0]**o+(0.731464233112)*x[0]+(0.637811533993)*x[1]**o+(0.381218512692)*x[1]+(-0.181357790048)*x[2]**o+(-0.296793024723)*x[2]
        arg[0,2,0,1]=(0.566451624572)*x[0]**o+(0.302904935559)*x[0]+(-0.771526665232)*x[1]**o+(0.460577683302)*x[1]+(-0.921312455507)*x[2]**o+(0.604299765394)*x[2]
        arg[0,2,1,0]=(-0.0887516895197)*x[0]**o+(0.872208558285)*x[0]+(-0.306564750999)*x[1]**o+(0.841982500263)*x[1]+(-0.550398737046)*x[2]**o+(-0.869697162427)*x[2]
        arg[0,2,1,1]=(-0.652471383912)*x[0]**o+(0.634592304393)*x[0]+(0.0709212473167)*x[1]**o+(-0.671330082579)*x[1]+(-0.232895004478)*x[2]**o+(0.479008351165)*x[2]
        arg[0,2,2,0]=(-0.0764115046136)*x[0]**o+(0.649230346025)*x[0]+(0.544503480999)*x[1]**o+(0.788838981797)*x[1]+(0.416236148267)*x[2]**o+(0.722979541918)*x[2]
        arg[0,2,2,1]=(-0.0776368628837)*x[0]**o+(0.801984674386)*x[0]+(0.443670313486)*x[1]**o+(0.424676792255)*x[1]+(-0.895989937982)*x[2]**o+(0.0774648143589)*x[2]
        arg[0,3,0,0]=(-0.39041318438)*x[0]**o+(-0.174335409262)*x[0]+(0.658666482337)*x[1]**o+(0.23060605921)*x[1]+(-0.336603364419)*x[2]**o+(0.871530192198)*x[2]
        arg[0,3,0,1]=(-0.493104661363)*x[0]**o+(0.590779697499)*x[0]+(-0.903897457443)*x[1]**o+(-0.377640520704)*x[1]+(0.227174544833)*x[2]**o+(-0.860150483539)*x[2]
        arg[0,3,1,0]=(-0.093367253954)*x[0]**o+(0.159230499811)*x[0]+(-0.731396509936)*x[1]**o+(0.600088148838)*x[1]+(0.829175660147)*x[2]**o+(-0.435878280919)*x[2]
        arg[0,3,1,1]=(0.665125540711)*x[0]**o+(-0.92128737118)*x[0]+(-0.33260297543)*x[1]**o+(0.682569476816)*x[1]+(0.999761405058)*x[2]**o+(-0.748303702618)*x[2]
        arg[0,3,2,0]=(-0.430442277525)*x[0]**o+(-0.841437664378)*x[0]+(-0.374267371055)*x[1]**o+(0.667860327373)*x[1]+(0.694185699238)*x[2]**o+(-0.673629629622)*x[2]
        arg[0,3,2,1]=(0.304808617243)*x[0]**o+(-0.894407776037)*x[0]+(-0.758585281699)*x[1]**o+(0.20519963241)*x[1]+(0.538222015181)*x[2]**o+(-0.792962435573)*x[2]
        arg[0,4,0,0]=(0.568300300269)*x[0]**o+(0.407455936625)*x[0]+(0.980898344535)*x[1]**o+(0.988220474461)*x[1]+(0.0598892620089)*x[2]**o+(-0.279555207723)*x[2]
        arg[0,4,0,1]=(-0.208391854989)*x[0]**o+(0.9428019078)*x[0]+(-0.540916877275)*x[1]**o+(-0.427185538555)*x[1]+(-0.879268085857)*x[2]**o+(0.025102321099)*x[2]
        arg[0,4,1,0]=(0.412690112527)*x[0]**o+(-0.957376863734)*x[0]+(-0.913623139719)*x[1]**o+(-0.462418849979)*x[1]+(0.463336747019)*x[2]**o+(0.330747531293)*x[2]
        arg[0,4,1,1]=(-0.632140552118)*x[0]**o+(0.830769577671)*x[0]+(-0.888012251479)*x[1]**o+(-0.0560181569699)*x[1]+(-0.70068099248)*x[2]**o+(0.834483253039)*x[2]
        arg[0,4,2,0]=(0.0429717567385)*x[0]**o+(-0.691784042595)*x[0]+(-0.264988764713)*x[1]**o+(0.359873170931)*x[1]+(-0.116275128437)*x[2]**o+(0.0348873917345)*x[2]
        arg[0,4,2,1]=(-0.203754358311)*x[0]**o+(0.177896174646)*x[0]+(-0.0785920580727)*x[1]**o+(-0.124905949184)*x[1]+(0.968871355367)*x[2]**o+(-0.0369722247589)*x[2]
        arg[1,0,0,0]=(0.594851756786)*x[0]**o+(-0.594400822936)*x[0]+(0.868401350949)*x[1]**o+(0.186739616297)*x[1]+(-0.937219946754)*x[2]**o+(-0.248916422131)*x[2]
        arg[1,0,0,1]=(-0.639865978203)*x[0]**o+(-0.00990375501714)*x[0]+(-0.038266357411)*x[1]**o+(-0.741522380136)*x[1]+(0.460572922772)*x[2]**o+(0.4314205896)*x[2]
        arg[1,0,1,0]=(-0.460110413486)*x[0]**o+(-0.0964934735182)*x[0]+(0.474209024352)*x[1]**o+(-0.928476110805)*x[1]+(0.598621411487)*x[2]**o+(-0.00858812754345)*x[2]
        arg[1,0,1,1]=(0.936466172807)*x[0]**o+(-0.363937334197)*x[0]+(-0.540224651799)*x[1]**o+(0.484530722263)*x[1]+(0.675416993296)*x[2]**o+(0.89657344343)*x[2]
        arg[1,0,2,0]=(-0.713050692555)*x[0]**o+(-0.00851892179229)*x[0]+(0.689494113809)*x[1]**o+(0.633341028044)*x[1]+(0.272726911113)*x[2]**o+(-0.122017566819)*x[2]
        arg[1,0,2,1]=(0.793152080981)*x[0]**o+(0.287379739639)*x[0]+(0.304047372771)*x[1]**o+(-0.509079361053)*x[1]+(0.888302006985)*x[2]**o+(-0.297023045775)*x[2]
        arg[1,1,0,0]=(-0.69695118901)*x[0]**o+(0.505182668512)*x[0]+(0.827555590952)*x[1]**o+(-0.777232447056)*x[1]+(-0.460918138036)*x[2]**o+(0.782090206922)*x[2]
        arg[1,1,0,1]=(-0.435398535566)*x[0]**o+(-0.0997593681984)*x[0]+(0.140250252691)*x[1]**o+(-0.0204469256488)*x[1]+(-0.445836771028)*x[2]**o+(-0.775366037861)*x[2]
        arg[1,1,1,0]=(0.918015347957)*x[0]**o+(-0.491940111489)*x[0]+(0.793071617663)*x[1]**o+(-0.607035294859)*x[1]+(0.610076618566)*x[2]**o+(0.148116733513)*x[2]
        arg[1,1,1,1]=(0.224062518811)*x[0]**o+(-0.145897156735)*x[0]+(0.165380866901)*x[1]**o+(0.910562981814)*x[1]+(-0.406620561548)*x[2]**o+(0.528862812468)*x[2]
        arg[1,1,2,0]=(0.85148444093)*x[0]**o+(0.13210280653)*x[0]+(0.364136256712)*x[1]**o+(0.520388575386)*x[1]+(0.0145641287346)*x[2]**o+(-0.7849036351)*x[2]
        arg[1,1,2,1]=(0.493837193832)*x[0]**o+(-0.193326176569)*x[0]+(-0.665920972154)*x[1]**o+(-0.321645070095)*x[1]+(0.0209354853869)*x[2]**o+(-0.328315981752)*x[2]
        arg[1,2,0,0]=(-0.349281164766)*x[0]**o+(-0.318415198113)*x[0]+(-0.687972781234)*x[1]**o+(0.91825998605)*x[1]+(0.396868773798)*x[2]**o+(0.998204458027)*x[2]
        arg[1,2,0,1]=(-0.967493222935)*x[0]**o+(-0.721128465653)*x[0]+(-0.672178802974)*x[1]**o+(0.0924145224431)*x[1]+(0.114040587594)*x[2]**o+(0.707623895537)*x[2]
        arg[1,2,1,0]=(-0.253327088773)*x[0]**o+(0.206063827625)*x[0]+(0.355152654895)*x[1]**o+(-0.27339285621)*x[1]+(-0.0135567122291)*x[2]**o+(-0.983990446362)*x[2]
        arg[1,2,1,1]=(0.98471687132)*x[0]**o+(0.262902745655)*x[0]+(-0.589359138227)*x[1]**o+(-0.0982548958556)*x[1]+(-0.526161200133)*x[2]**o+(-0.850939484752)*x[2]
        arg[1,2,2,0]=(0.37583596705)*x[0]**o+(0.274845981222)*x[0]+(0.601331110943)*x[1]**o+(0.788294448218)*x[1]+(0.366445166459)*x[2]**o+(-0.554935072394)*x[2]
        arg[1,2,2,1]=(-0.986387850237)*x[0]**o+(0.68772989781)*x[0]+(0.345166604933)*x[1]**o+(0.930376726384)*x[1]+(0.760974421701)*x[2]**o+(0.261797396808)*x[2]
        arg[1,3,0,0]=(-0.171005013523)*x[0]**o+(-0.339229596812)*x[0]+(-0.150317419556)*x[1]**o+(-0.0146533427463)*x[1]+(0.957020663875)*x[2]**o+(0.616513940347)*x[2]
        arg[1,3,0,1]=(-0.0727328269499)*x[0]**o+(-0.635591748073)*x[0]+(-0.502426989774)*x[1]**o+(0.772595119619)*x[1]+(-0.76006511271)*x[2]**o+(-0.665920288084)*x[2]
        arg[1,3,1,0]=(-0.394284205622)*x[0]**o+(-0.105959165811)*x[0]+(0.72649045701)*x[1]**o+(-0.970276379473)*x[1]+(-0.236520349415)*x[2]**o+(0.994973267783)*x[2]
        arg[1,3,1,1]=(-0.444149867906)*x[0]**o+(0.748449201475)*x[0]+(0.419604170756)*x[1]**o+(-0.0291702705178)*x[1]+(-0.109580142628)*x[2]**o+(-0.0791870699838)*x[2]
        arg[1,3,2,0]=(-0.629657820523)*x[0]**o+(-0.995576552624)*x[0]+(0.530135462206)*x[1]**o+(0.197336335845)*x[1]+(0.0224815558807)*x[2]**o+(0.240498463912)*x[2]
        arg[1,3,2,1]=(0.52933948346)*x[0]**o+(0.98353913011)*x[0]+(0.0217614943138)*x[1]**o+(0.0157778453682)*x[1]+(0.177956932986)*x[2]**o+(0.234629778706)*x[2]
        arg[1,4,0,0]=(0.593028367137)*x[0]**o+(0.0340892538994)*x[0]+(-0.0295133170051)*x[1]**o+(0.0865921041963)*x[1]+(0.641131718656)*x[2]**o+(0.788858018609)*x[2]
        arg[1,4,0,1]=(-0.283333820634)*x[0]**o+(0.864482725099)*x[0]+(-0.560218794286)*x[1]**o+(-0.0083115362546)*x[1]+(0.726423285129)*x[2]**o+(0.031531339166)*x[2]
        arg[1,4,1,0]=(0.801452411554)*x[0]**o+(-0.59916393038)*x[0]+(-0.106568863591)*x[1]**o+(0.357442125133)*x[1]+(-0.203380280132)*x[2]**o+(-0.129031606441)*x[2]
        arg[1,4,1,1]=(-0.589220928637)*x[0]**o+(-0.308949389428)*x[0]+(0.785549858338)*x[1]**o+(-0.131153853015)*x[1]+(-0.30571855555)*x[2]**o+(-0.329634802588)*x[2]
        arg[1,4,2,0]=(-0.479686944)*x[0]**o+(-0.707702026873)*x[0]+(0.635894172352)*x[1]**o+(0.370673635069)*x[1]+(0.942862875991)*x[2]**o+(-0.846943456738)*x[2]
        arg[1,4,2,1]=(-0.403737324968)*x[0]**o+(0.101664426734)*x[0]+(-0.677093482354)*x[1]**o+(-0.36526066745)*x[1]+(-0.0889619136182)*x[2]**o+(-0.772491940844)*x[2]
        arg[2,0,0,0]=(0.753148325987)*x[0]**o+(-0.342843083128)*x[0]+(0.812661029553)*x[1]**o+(-0.112683035459)*x[1]+(0.900976470275)*x[2]**o+(0.00934086955102)*x[2]
        arg[2,0,0,1]=(-0.682315007143)*x[0]**o+(-0.694069900861)*x[0]+(-0.469082247747)*x[1]**o+(-0.522165903301)*x[1]+(-0.0997886221924)*x[2]**o+(-0.954161552419)*x[2]
        arg[2,0,1,0]=(0.780125539705)*x[0]**o+(-0.445331450225)*x[0]+(-0.282765510424)*x[1]**o+(-0.0131898361592)*x[1]+(-0.200186433923)*x[2]**o+(0.715139435696)*x[2]
        arg[2,0,1,1]=(-0.132310660786)*x[0]**o+(-0.0981344234369)*x[0]+(0.959297529475)*x[1]**o+(-0.85944894515)*x[1]+(-0.396258659619)*x[2]**o+(-0.09672066185)*x[2]
        arg[2,0,2,0]=(-0.344134014191)*x[0]**o+(0.534345864287)*x[0]+(0.797209509484)*x[1]**o+(0.0479050101499)*x[1]+(0.13620476987)*x[2]**o+(0.82708115383)*x[2]
        arg[2,0,2,1]=(-0.342509328238)*x[0]**o+(-0.833164493006)*x[0]+(0.472047721532)*x[1]**o+(0.301613158637)*x[1]+(0.858691679638)*x[2]**o+(-0.691744164006)*x[2]
        arg[2,1,0,0]=(0.498979390299)*x[0]**o+(0.371513881824)*x[0]+(-0.243506167971)*x[1]**o+(-0.245423890439)*x[1]+(-0.0494707840274)*x[2]**o+(0.0965573709928)*x[2]
        arg[2,1,0,1]=(0.665982455862)*x[0]**o+(-0.409710866336)*x[0]+(0.669257568046)*x[1]**o+(0.711947704788)*x[1]+(-0.0362568635785)*x[2]**o+(-0.30394278188)*x[2]
        arg[2,1,1,0]=(0.296245249928)*x[0]**o+(0.132631019281)*x[0]+(-0.376880152158)*x[1]**o+(0.303184156241)*x[1]+(0.836046560882)*x[2]**o+(0.195205645106)*x[2]
        arg[2,1,1,1]=(-0.971045820953)*x[0]**o+(0.154708896467)*x[0]+(-0.663844656602)*x[1]**o+(-0.0873797588245)*x[1]+(0.985123372542)*x[2]**o+(0.349554922131)*x[2]
        arg[2,1,2,0]=(-0.0117443431348)*x[0]**o+(-0.71980503985)*x[0]+(0.175348078994)*x[1]**o+(-0.883987892161)*x[1]+(0.0587275450458)*x[2]**o+(0.464811715548)*x[2]
        arg[2,1,2,1]=(-0.309202575765)*x[0]**o+(-0.528513248453)*x[0]+(-0.509298303972)*x[1]**o+(-0.228278863496)*x[1]+(0.933917053676)*x[2]**o+(-0.823678942264)*x[2]
        arg[2,2,0,0]=(0.374387386929)*x[0]**o+(0.64969993847)*x[0]+(-0.939822635641)*x[1]**o+(0.578249430422)*x[1]+(-0.973412660635)*x[2]**o+(0.484052718386)*x[2]
        arg[2,2,0,1]=(-0.972635162618)*x[0]**o+(-0.255007433802)*x[0]+(0.134165187097)*x[1]**o+(0.313213615976)*x[1]+(0.143298148686)*x[2]**o+(0.901819440229)*x[2]
        arg[2,2,1,0]=(0.868173256315)*x[0]**o+(-0.532325343453)*x[0]+(-0.712500880845)*x[1]**o+(0.216943462308)*x[1]+(0.733920091212)*x[2]**o+(0.0992400603383)*x[2]
        arg[2,2,1,1]=(0.642839350527)*x[0]**o+(0.127721543559)*x[0]+(-0.152686523032)*x[1]**o+(-0.399032361234)*x[1]+(0.317014518842)*x[2]**o+(0.905538790115)*x[2]
        arg[2,2,2,0]=(0.685267932413)*x[0]**o+(-0.167571547782)*x[0]+(-0.213446144405)*x[1]**o+(0.420270456377)*x[1]+(0.827725350621)*x[2]**o+(-0.920242274399)*x[2]
        arg[2,2,2,1]=(-0.507424593437)*x[0]**o+(-0.694222840019)*x[0]+(-0.904108970188)*x[1]**o+(-0.750899614966)*x[1]+(0.783997553302)*x[2]**o+(0.196327513447)*x[2]
        arg[2,3,0,0]=(-0.473900600479)*x[0]**o+(-0.464347097789)*x[0]+(-0.472423358671)*x[1]**o+(0.373786456364)*x[1]+(-0.989371300414)*x[2]**o+(-0.959508065239)*x[2]
        arg[2,3,0,1]=(0.174375189399)*x[0]**o+(0.925503059886)*x[0]+(-0.647274384723)*x[1]**o+(-0.577149479679)*x[1]+(0.171298877768)*x[2]**o+(-0.936612100882)*x[2]
        arg[2,3,1,0]=(0.482959343396)*x[0]**o+(0.842243554011)*x[0]+(0.814423633314)*x[1]**o+(-0.989445050315)*x[1]+(0.817933433065)*x[2]**o+(0.552396881404)*x[2]
        arg[2,3,1,1]=(0.55641736229)*x[0]**o+(0.498094640058)*x[0]+(0.0890113606037)*x[1]**o+(0.36592723553)*x[1]+(0.458607593386)*x[2]**o+(0.603047420063)*x[2]
        arg[2,3,2,0]=(0.884515080124)*x[0]**o+(-0.561898381022)*x[0]+(-0.292475503834)*x[1]**o+(0.425634423478)*x[1]+(0.65813323252)*x[2]**o+(-0.548850399842)*x[2]
        arg[2,3,2,1]=(-0.72895566982)*x[0]**o+(-0.569405588005)*x[0]+(0.117770875594)*x[1]**o+(-0.33515735969)*x[1]+(0.102062736382)*x[2]**o+(0.611532417683)*x[2]
        arg[2,4,0,0]=(-0.669208630424)*x[0]**o+(0.944393014177)*x[0]+(-0.00345511889683)*x[1]**o+(0.0695279595783)*x[1]+(0.496948250833)*x[2]**o+(-0.899750463971)*x[2]
        arg[2,4,0,1]=(-0.387828717292)*x[0]**o+(0.621172267004)*x[0]+(-0.280632674111)*x[1]**o+(0.401702567575)*x[1]+(-0.65554489378)*x[2]**o+(0.403018047007)*x[2]
        arg[2,4,1,0]=(-0.0770246925803)*x[0]**o+(0.52913302745)*x[0]+(0.922468724357)*x[1]**o+(-0.304442750439)*x[1]+(-0.556575079059)*x[2]**o+(-0.725790672251)*x[2]
        arg[2,4,1,1]=(-0.828378342911)*x[0]**o+(-0.0696370821776)*x[0]+(0.394018581667)*x[1]**o+(-0.251099848269)*x[1]+(-0.29878985656)*x[2]**o+(0.925753371056)*x[2]
        arg[2,4,2,0]=(-0.818277786029)*x[0]**o+(0.225862354706)*x[0]+(-0.629050212595)*x[1]**o+(0.139186107062)*x[1]+(-0.0134835519238)*x[2]**o+(0.301899528641)*x[2]
        arg[2,4,2,1]=(-0.783053684767)*x[0]**o+(-0.0926938783599)*x[0]+(0.519034210352)*x[1]**o+(-0.494564657126)*x[1]+(0.536117310622)*x[2]**o+(0.6278756634)*x[2]
        arg[3,0,0,0]=(-0.694353957685)*x[0]**o+(-0.361037363965)*x[0]+(-0.327004091885)*x[1]**o+(-0.765735142884)*x[1]+(-0.810915729713)*x[2]**o+(0.0640539311609)*x[2]
        arg[3,0,0,1]=(-0.131601548312)*x[0]**o+(0.271587756708)*x[0]+(0.322486168005)*x[1]**o+(0.238086581801)*x[1]+(-0.488081720555)*x[2]**o+(0.607278995635)*x[2]
        arg[3,0,1,0]=(0.903075418102)*x[0]**o+(-0.738609834848)*x[0]+(-0.165097989126)*x[1]**o+(-0.25967140914)*x[1]+(0.271649887428)*x[2]**o+(-0.786131648727)*x[2]
        arg[3,0,1,1]=(-0.963442989157)*x[0]**o+(-0.240205602334)*x[0]+(0.102153389731)*x[1]**o+(-0.106402434995)*x[1]+(-0.383097724211)*x[2]**o+(0.680290963195)*x[2]
        arg[3,0,2,0]=(-0.373162835796)*x[0]**o+(-0.186323761814)*x[0]+(0.661494315705)*x[1]**o+(0.320072189494)*x[1]+(0.0964443431067)*x[2]**o+(0.829015838241)*x[2]
        arg[3,0,2,1]=(0.527719514862)*x[0]**o+(0.0819948261665)*x[0]+(0.826598712178)*x[1]**o+(-0.588102975125)*x[1]+(-0.632584402292)*x[2]**o+(0.0826992233702)*x[2]
        arg[3,1,0,0]=(-0.102243686287)*x[0]**o+(-0.497336112929)*x[0]+(0.941827592628)*x[1]**o+(0.68678991944)*x[1]+(-0.508684841667)*x[2]**o+(-0.752939965161)*x[2]
        arg[3,1,0,1]=(-0.896417523183)*x[0]**o+(0.759224781685)*x[0]+(0.138537695823)*x[1]**o+(0.466727480471)*x[1]+(-0.838380824412)*x[2]**o+(0.285761157398)*x[2]
        arg[3,1,1,0]=(-0.547853182919)*x[0]**o+(-0.41595413414)*x[0]+(0.76894221584)*x[1]**o+(-0.142394275192)*x[1]+(-0.96349294581)*x[2]**o+(-0.873639964693)*x[2]
        arg[3,1,1,1]=(-0.151299403888)*x[0]**o+(0.789406063878)*x[0]+(0.163280670253)*x[1]**o+(-0.287315494745)*x[1]+(-0.627764670218)*x[2]**o+(0.842131632598)*x[2]
        arg[3,1,2,0]=(-0.957096489189)*x[0]**o+(0.0616037805677)*x[0]+(0.255858200166)*x[1]**o+(-0.430555450798)*x[1]+(0.730436930737)*x[2]**o+(0.405094271653)*x[2]
        arg[3,1,2,1]=(-0.924756832665)*x[0]**o+(0.525754162391)*x[0]+(0.281303228038)*x[1]**o+(-0.479500161878)*x[1]+(0.658416209181)*x[2]**o+(-0.0413881595557)*x[2]
        arg[3,2,0,0]=(0.516429381997)*x[0]**o+(-0.8387318467)*x[0]+(0.329773978227)*x[1]**o+(0.599641753085)*x[1]+(0.939937743271)*x[2]**o+(-0.247331169626)*x[2]
        arg[3,2,0,1]=(-0.644982044558)*x[0]**o+(-0.242033352774)*x[0]+(0.531835185779)*x[1]**o+(-0.851565064183)*x[1]+(-0.79213624915)*x[2]**o+(0.0607365690732)*x[2]
        arg[3,2,1,0]=(0.359781383537)*x[0]**o+(-0.0529212842074)*x[0]+(0.845915169452)*x[1]**o+(-0.323979570458)*x[1]+(0.828909997631)*x[2]**o+(-0.839886339138)*x[2]
        arg[3,2,1,1]=(0.559476630421)*x[0]**o+(-0.189725214615)*x[0]+(-0.00927446700413)*x[1]**o+(-0.234703720782)*x[1]+(-0.608362806755)*x[2]**o+(0.968571765534)*x[2]
        arg[3,2,2,0]=(0.924837507913)*x[0]**o+(0.846397611121)*x[0]+(-0.837765472403)*x[1]**o+(0.0225105985057)*x[1]+(-0.834187513844)*x[2]**o+(0.194311845578)*x[2]
        arg[3,2,2,1]=(-0.183561089961)*x[0]**o+(0.0745071729471)*x[0]+(-0.660415615202)*x[1]**o+(-0.528184997345)*x[1]+(0.205554094557)*x[2]**o+(-0.0309246318944)*x[2]
        arg[3,3,0,0]=(-0.67124095974)*x[0]**o+(-0.48501497083)*x[0]+(-0.0394345869094)*x[1]**o+(0.586692270429)*x[1]+(-0.593760749687)*x[2]**o+(-0.124979751327)*x[2]
        arg[3,3,0,1]=(0.778544173855)*x[0]**o+(-0.931466478494)*x[0]+(-0.746894777998)*x[1]**o+(0.806417936297)*x[1]+(0.989460929848)*x[2]**o+(0.666534366369)*x[2]
        arg[3,3,1,0]=(-0.798870838602)*x[0]**o+(-0.294618868937)*x[0]+(0.275029092968)*x[1]**o+(0.45390580364)*x[1]+(-0.311797818508)*x[2]**o+(0.751606624393)*x[2]
        arg[3,3,1,1]=(-0.924260987477)*x[0]**o+(0.771970726902)*x[0]+(-0.909099224213)*x[1]**o+(-0.45204206766)*x[1]+(0.630042614017)*x[2]**o+(-0.990915130464)*x[2]
        arg[3,3,2,0]=(0.734962154852)*x[0]**o+(-0.00621898011475)*x[0]+(-0.786362527189)*x[1]**o+(0.623507947406)*x[1]+(0.867175567104)*x[2]**o+(-0.383283255448)*x[2]
        arg[3,3,2,1]=(-0.955613156717)*x[0]**o+(0.688013754841)*x[0]+(-0.600559375208)*x[1]**o+(0.453342626645)*x[1]+(-0.864890477961)*x[2]**o+(-0.458029873063)*x[2]
        arg[3,4,0,0]=(-0.897010884834)*x[0]**o+(-0.251223576458)*x[0]+(-0.0338052047964)*x[1]**o+(-0.903421466132)*x[1]+(0.836864217874)*x[2]**o+(0.660805905175)*x[2]
        arg[3,4,0,1]=(-0.286025690215)*x[0]**o+(0.746732975604)*x[0]+(0.375810853218)*x[1]**o+(-0.985192974869)*x[1]+(-0.731463214455)*x[2]**o+(-0.465450352444)*x[2]
        arg[3,4,1,0]=(-0.500910000652)*x[0]**o+(0.387502419727)*x[0]+(0.962369164802)*x[1]**o+(-0.10933824251)*x[1]+(-0.33642703989)*x[2]**o+(0.546851133256)*x[2]
        arg[3,4,1,1]=(-0.117959459645)*x[0]**o+(0.867699323392)*x[0]+(0.950506403861)*x[1]**o+(0.613806207745)*x[1]+(0.425723173807)*x[2]**o+(-0.906937548619)*x[2]
        arg[3,4,2,0]=(0.331114892194)*x[0]**o+(-0.168829643424)*x[0]+(-0.855962311982)*x[1]**o+(-0.144202406673)*x[1]+(0.0769985713088)*x[2]**o+(0.856003746423)*x[2]
        arg[3,4,2,1]=(-0.439397439364)*x[0]**o+(0.686116343481)*x[0]+(0.519883854523)*x[1]**o+(0.716231469289)*x[1]+(0.270098231149)*x[2]**o+(-0.448525436427)*x[2]
        ref[0,0,0,0]=(-0.401901998082)*x_ref[0]**o+(-0.805091116808)*x_ref[0]+(0.997659882792)*x_ref[1]**o+(-0.953957389492)*x_ref[1]+(-0.450185408289)*x_ref[2]**o+(0.636472207773)*x_ref[2]
        ref[0,0,0,1]=(-0.786386817128)*x_ref[0]**o+(-0.882559775393)*x_ref[0]+(0.117871400819)*x_ref[1]**o+(-0.625922274774)*x_ref[1]+(0.410431886931)*x_ref[2]**o+(-0.979356501067)*x_ref[2]
        ref[0,0,1,0]=(0.686025409275)*x_ref[0]**o+(-0.19806907659)*x_ref[0]+(-0.710420224449)*x_ref[1]**o+(-0.839236953983)*x_ref[1]+(-0.79971896093)*x_ref[2]**o+(0.378944111319)*x_ref[2]
        ref[0,0,1,1]=(0.771205112484)*x_ref[0]**o+(-0.007083192859)*x_ref[0]+(0.672918588001)*x_ref[1]**o+(-0.351652495793)*x_ref[1]+(-0.841767161187)*x_ref[2]**o+(-0.345642373014)*x_ref[2]
        ref[0,0,2,0]=(0.266930729015)*x_ref[0]**o+(-0.873808152307)*x_ref[0]+(0.206466759135)*x_ref[1]**o+(0.539888649948)*x_ref[1]+(0.17985862191)*x_ref[2]**o+(0.240651460153)*x_ref[2]
        ref[0,0,2,1]=(-0.232569768041)*x_ref[0]**o+(-0.579841454072)*x_ref[0]+(-0.706956104891)*x_ref[1]**o+(0.822863815698)*x_ref[1]+(0.345774991358)*x_ref[2]**o+(-0.0848567668311)*x_ref[2]
        ref[0,1,0,0]=(0.51593851507)*x_ref[0]**o+(-0.233342434349)*x_ref[0]+(-0.92498298026)*x_ref[1]**o+(0.164540234829)*x_ref[1]+(0.356559744499)*x_ref[2]**o+(0.483902511717)*x_ref[2]
        ref[0,1,0,1]=(-0.811509654443)*x_ref[0]**o+(0.73929453189)*x_ref[0]+(0.256205706418)*x_ref[1]**o+(-0.667924913793)*x_ref[1]+(0.266065641257)*x_ref[2]**o+(0.856490258029)*x_ref[2]
        ref[0,1,1,0]=(0.502051590284)*x_ref[0]**o+(0.927283797128)*x_ref[0]+(0.500854521411)*x_ref[1]**o+(-0.486326996369)*x_ref[1]+(-0.327578523685)*x_ref[2]**o+(-0.28106070397)*x_ref[2]
        ref[0,1,1,1]=(0.201740014513)*x_ref[0]**o+(-0.166988241759)*x_ref[0]+(-0.513072955566)*x_ref[1]**o+(0.67179532417)*x_ref[1]+(-0.72135980627)*x_ref[2]**o+(0.763548886467)*x_ref[2]
        ref[0,1,2,0]=(0.3995590338)*x_ref[0]**o+(0.433286929621)*x_ref[0]+(-0.287309453792)*x_ref[1]**o+(-0.0671679443074)*x_ref[1]+(-0.80815399993)*x_ref[2]**o+(-0.398740362114)*x_ref[2]
        ref[0,1,2,1]=(-0.16314640684)*x_ref[0]**o+(-0.698681697367)*x_ref[0]+(-0.11609738418)*x_ref[1]**o+(-0.798465154739)*x_ref[1]+(-0.169052628971)*x_ref[2]**o+(-0.880316962539)*x_ref[2]
        ref[0,2,0,0]=(-0.219731279442)*x_ref[0]**o+(0.731464233112)*x_ref[0]+(0.637811533993)*x_ref[1]**o+(0.381218512692)*x_ref[1]+(-0.181357790048)*x_ref[2]**o+(-0.296793024723)*x_ref[2]
        ref[0,2,0,1]=(0.566451624572)*x_ref[0]**o+(0.302904935559)*x_ref[0]+(-0.771526665232)*x_ref[1]**o+(0.460577683302)*x_ref[1]+(-0.921312455507)*x_ref[2]**o+(0.604299765394)*x_ref[2]
        ref[0,2,1,0]=(-0.0887516895197)*x_ref[0]**o+(0.872208558285)*x_ref[0]+(-0.306564750999)*x_ref[1]**o+(0.841982500263)*x_ref[1]+(-0.550398737046)*x_ref[2]**o+(-0.869697162427)*x_ref[2]
        ref[0,2,1,1]=(-0.652471383912)*x_ref[0]**o+(0.634592304393)*x_ref[0]+(0.0709212473167)*x_ref[1]**o+(-0.671330082579)*x_ref[1]+(-0.232895004478)*x_ref[2]**o+(0.479008351165)*x_ref[2]
        ref[0,2,2,0]=(-0.0764115046136)*x_ref[0]**o+(0.649230346025)*x_ref[0]+(0.544503480999)*x_ref[1]**o+(0.788838981797)*x_ref[1]+(0.416236148267)*x_ref[2]**o+(0.722979541918)*x_ref[2]
        ref[0,2,2,1]=(-0.0776368628837)*x_ref[0]**o+(0.801984674386)*x_ref[0]+(0.443670313486)*x_ref[1]**o+(0.424676792255)*x_ref[1]+(-0.895989937982)*x_ref[2]**o+(0.0774648143589)*x_ref[2]
        ref[0,3,0,0]=(-0.39041318438)*x_ref[0]**o+(-0.174335409262)*x_ref[0]+(0.658666482337)*x_ref[1]**o+(0.23060605921)*x_ref[1]+(-0.336603364419)*x_ref[2]**o+(0.871530192198)*x_ref[2]
        ref[0,3,0,1]=(-0.493104661363)*x_ref[0]**o+(0.590779697499)*x_ref[0]+(-0.903897457443)*x_ref[1]**o+(-0.377640520704)*x_ref[1]+(0.227174544833)*x_ref[2]**o+(-0.860150483539)*x_ref[2]
        ref[0,3,1,0]=(-0.093367253954)*x_ref[0]**o+(0.159230499811)*x_ref[0]+(-0.731396509936)*x_ref[1]**o+(0.600088148838)*x_ref[1]+(0.829175660147)*x_ref[2]**o+(-0.435878280919)*x_ref[2]
        ref[0,3,1,1]=(0.665125540711)*x_ref[0]**o+(-0.92128737118)*x_ref[0]+(-0.33260297543)*x_ref[1]**o+(0.682569476816)*x_ref[1]+(0.999761405058)*x_ref[2]**o+(-0.748303702618)*x_ref[2]
        ref[0,3,2,0]=(-0.430442277525)*x_ref[0]**o+(-0.841437664378)*x_ref[0]+(-0.374267371055)*x_ref[1]**o+(0.667860327373)*x_ref[1]+(0.694185699238)*x_ref[2]**o+(-0.673629629622)*x_ref[2]
        ref[0,3,2,1]=(0.304808617243)*x_ref[0]**o+(-0.894407776037)*x_ref[0]+(-0.758585281699)*x_ref[1]**o+(0.20519963241)*x_ref[1]+(0.538222015181)*x_ref[2]**o+(-0.792962435573)*x_ref[2]
        ref[0,4,0,0]=(0.568300300269)*x_ref[0]**o+(0.407455936625)*x_ref[0]+(0.980898344535)*x_ref[1]**o+(0.988220474461)*x_ref[1]+(0.0598892620089)*x_ref[2]**o+(-0.279555207723)*x_ref[2]
        ref[0,4,0,1]=(-0.208391854989)*x_ref[0]**o+(0.9428019078)*x_ref[0]+(-0.540916877275)*x_ref[1]**o+(-0.427185538555)*x_ref[1]+(-0.879268085857)*x_ref[2]**o+(0.025102321099)*x_ref[2]
        ref[0,4,1,0]=(0.412690112527)*x_ref[0]**o+(-0.957376863734)*x_ref[0]+(-0.913623139719)*x_ref[1]**o+(-0.462418849979)*x_ref[1]+(0.463336747019)*x_ref[2]**o+(0.330747531293)*x_ref[2]
        ref[0,4,1,1]=(-0.632140552118)*x_ref[0]**o+(0.830769577671)*x_ref[0]+(-0.888012251479)*x_ref[1]**o+(-0.0560181569699)*x_ref[1]+(-0.70068099248)*x_ref[2]**o+(0.834483253039)*x_ref[2]
        ref[0,4,2,0]=(0.0429717567385)*x_ref[0]**o+(-0.691784042595)*x_ref[0]+(-0.264988764713)*x_ref[1]**o+(0.359873170931)*x_ref[1]+(-0.116275128437)*x_ref[2]**o+(0.0348873917345)*x_ref[2]
        ref[0,4,2,1]=(-0.203754358311)*x_ref[0]**o+(0.177896174646)*x_ref[0]+(-0.0785920580727)*x_ref[1]**o+(-0.124905949184)*x_ref[1]+(0.968871355367)*x_ref[2]**o+(-0.0369722247589)*x_ref[2]
        ref[1,0,0,0]=(0.594851756786)*x_ref[0]**o+(-0.594400822936)*x_ref[0]+(0.868401350949)*x_ref[1]**o+(0.186739616297)*x_ref[1]+(-0.937219946754)*x_ref[2]**o+(-0.248916422131)*x_ref[2]
        ref[1,0,0,1]=(-0.639865978203)*x_ref[0]**o+(-0.00990375501714)*x_ref[0]+(-0.038266357411)*x_ref[1]**o+(-0.741522380136)*x_ref[1]+(0.460572922772)*x_ref[2]**o+(0.4314205896)*x_ref[2]
        ref[1,0,1,0]=(-0.460110413486)*x_ref[0]**o+(-0.0964934735182)*x_ref[0]+(0.474209024352)*x_ref[1]**o+(-0.928476110805)*x_ref[1]+(0.598621411487)*x_ref[2]**o+(-0.00858812754345)*x_ref[2]
        ref[1,0,1,1]=(0.936466172807)*x_ref[0]**o+(-0.363937334197)*x_ref[0]+(-0.540224651799)*x_ref[1]**o+(0.484530722263)*x_ref[1]+(0.675416993296)*x_ref[2]**o+(0.89657344343)*x_ref[2]
        ref[1,0,2,0]=(-0.713050692555)*x_ref[0]**o+(-0.00851892179229)*x_ref[0]+(0.689494113809)*x_ref[1]**o+(0.633341028044)*x_ref[1]+(0.272726911113)*x_ref[2]**o+(-0.122017566819)*x_ref[2]
        ref[1,0,2,1]=(0.793152080981)*x_ref[0]**o+(0.287379739639)*x_ref[0]+(0.304047372771)*x_ref[1]**o+(-0.509079361053)*x_ref[1]+(0.888302006985)*x_ref[2]**o+(-0.297023045775)*x_ref[2]
        ref[1,1,0,0]=(-0.69695118901)*x_ref[0]**o+(0.505182668512)*x_ref[0]+(0.827555590952)*x_ref[1]**o+(-0.777232447056)*x_ref[1]+(-0.460918138036)*x_ref[2]**o+(0.782090206922)*x_ref[2]
        ref[1,1,0,1]=(-0.435398535566)*x_ref[0]**o+(-0.0997593681984)*x_ref[0]+(0.140250252691)*x_ref[1]**o+(-0.0204469256488)*x_ref[1]+(-0.445836771028)*x_ref[2]**o+(-0.775366037861)*x_ref[2]
        ref[1,1,1,0]=(0.918015347957)*x_ref[0]**o+(-0.491940111489)*x_ref[0]+(0.793071617663)*x_ref[1]**o+(-0.607035294859)*x_ref[1]+(0.610076618566)*x_ref[2]**o+(0.148116733513)*x_ref[2]
        ref[1,1,1,1]=(0.224062518811)*x_ref[0]**o+(-0.145897156735)*x_ref[0]+(0.165380866901)*x_ref[1]**o+(0.910562981814)*x_ref[1]+(-0.406620561548)*x_ref[2]**o+(0.528862812468)*x_ref[2]
        ref[1,1,2,0]=(0.85148444093)*x_ref[0]**o+(0.13210280653)*x_ref[0]+(0.364136256712)*x_ref[1]**o+(0.520388575386)*x_ref[1]+(0.0145641287346)*x_ref[2]**o+(-0.7849036351)*x_ref[2]
        ref[1,1,2,1]=(0.493837193832)*x_ref[0]**o+(-0.193326176569)*x_ref[0]+(-0.665920972154)*x_ref[1]**o+(-0.321645070095)*x_ref[1]+(0.0209354853869)*x_ref[2]**o+(-0.328315981752)*x_ref[2]
        ref[1,2,0,0]=(-0.349281164766)*x_ref[0]**o+(-0.318415198113)*x_ref[0]+(-0.687972781234)*x_ref[1]**o+(0.91825998605)*x_ref[1]+(0.396868773798)*x_ref[2]**o+(0.998204458027)*x_ref[2]
        ref[1,2,0,1]=(-0.967493222935)*x_ref[0]**o+(-0.721128465653)*x_ref[0]+(-0.672178802974)*x_ref[1]**o+(0.0924145224431)*x_ref[1]+(0.114040587594)*x_ref[2]**o+(0.707623895537)*x_ref[2]
        ref[1,2,1,0]=(-0.253327088773)*x_ref[0]**o+(0.206063827625)*x_ref[0]+(0.355152654895)*x_ref[1]**o+(-0.27339285621)*x_ref[1]+(-0.0135567122291)*x_ref[2]**o+(-0.983990446362)*x_ref[2]
        ref[1,2,1,1]=(0.98471687132)*x_ref[0]**o+(0.262902745655)*x_ref[0]+(-0.589359138227)*x_ref[1]**o+(-0.0982548958556)*x_ref[1]+(-0.526161200133)*x_ref[2]**o+(-0.850939484752)*x_ref[2]
        ref[1,2,2,0]=(0.37583596705)*x_ref[0]**o+(0.274845981222)*x_ref[0]+(0.601331110943)*x_ref[1]**o+(0.788294448218)*x_ref[1]+(0.366445166459)*x_ref[2]**o+(-0.554935072394)*x_ref[2]
        ref[1,2,2,1]=(-0.986387850237)*x_ref[0]**o+(0.68772989781)*x_ref[0]+(0.345166604933)*x_ref[1]**o+(0.930376726384)*x_ref[1]+(0.760974421701)*x_ref[2]**o+(0.261797396808)*x_ref[2]
        ref[1,3,0,0]=(-0.171005013523)*x_ref[0]**o+(-0.339229596812)*x_ref[0]+(-0.150317419556)*x_ref[1]**o+(-0.0146533427463)*x_ref[1]+(0.957020663875)*x_ref[2]**o+(0.616513940347)*x_ref[2]
        ref[1,3,0,1]=(-0.0727328269499)*x_ref[0]**o+(-0.635591748073)*x_ref[0]+(-0.502426989774)*x_ref[1]**o+(0.772595119619)*x_ref[1]+(-0.76006511271)*x_ref[2]**o+(-0.665920288084)*x_ref[2]
        ref[1,3,1,0]=(-0.394284205622)*x_ref[0]**o+(-0.105959165811)*x_ref[0]+(0.72649045701)*x_ref[1]**o+(-0.970276379473)*x_ref[1]+(-0.236520349415)*x_ref[2]**o+(0.994973267783)*x_ref[2]
        ref[1,3,1,1]=(-0.444149867906)*x_ref[0]**o+(0.748449201475)*x_ref[0]+(0.419604170756)*x_ref[1]**o+(-0.0291702705178)*x_ref[1]+(-0.109580142628)*x_ref[2]**o+(-0.0791870699838)*x_ref[2]
        ref[1,3,2,0]=(-0.629657820523)*x_ref[0]**o+(-0.995576552624)*x_ref[0]+(0.530135462206)*x_ref[1]**o+(0.197336335845)*x_ref[1]+(0.0224815558807)*x_ref[2]**o+(0.240498463912)*x_ref[2]
        ref[1,3,2,1]=(0.52933948346)*x_ref[0]**o+(0.98353913011)*x_ref[0]+(0.0217614943138)*x_ref[1]**o+(0.0157778453682)*x_ref[1]+(0.177956932986)*x_ref[2]**o+(0.234629778706)*x_ref[2]
        ref[1,4,0,0]=(0.593028367137)*x_ref[0]**o+(0.0340892538994)*x_ref[0]+(-0.0295133170051)*x_ref[1]**o+(0.0865921041963)*x_ref[1]+(0.641131718656)*x_ref[2]**o+(0.788858018609)*x_ref[2]
        ref[1,4,0,1]=(-0.283333820634)*x_ref[0]**o+(0.864482725099)*x_ref[0]+(-0.560218794286)*x_ref[1]**o+(-0.0083115362546)*x_ref[1]+(0.726423285129)*x_ref[2]**o+(0.031531339166)*x_ref[2]
        ref[1,4,1,0]=(0.801452411554)*x_ref[0]**o+(-0.59916393038)*x_ref[0]+(-0.106568863591)*x_ref[1]**o+(0.357442125133)*x_ref[1]+(-0.203380280132)*x_ref[2]**o+(-0.129031606441)*x_ref[2]
        ref[1,4,1,1]=(-0.589220928637)*x_ref[0]**o+(-0.308949389428)*x_ref[0]+(0.785549858338)*x_ref[1]**o+(-0.131153853015)*x_ref[1]+(-0.30571855555)*x_ref[2]**o+(-0.329634802588)*x_ref[2]
        ref[1,4,2,0]=(-0.479686944)*x_ref[0]**o+(-0.707702026873)*x_ref[0]+(0.635894172352)*x_ref[1]**o+(0.370673635069)*x_ref[1]+(0.942862875991)*x_ref[2]**o+(-0.846943456738)*x_ref[2]
        ref[1,4,2,1]=(-0.403737324968)*x_ref[0]**o+(0.101664426734)*x_ref[0]+(-0.677093482354)*x_ref[1]**o+(-0.36526066745)*x_ref[1]+(-0.0889619136182)*x_ref[2]**o+(-0.772491940844)*x_ref[2]
        ref[2,0,0,0]=(0.753148325987)*x_ref[0]**o+(-0.342843083128)*x_ref[0]+(0.812661029553)*x_ref[1]**o+(-0.112683035459)*x_ref[1]+(0.900976470275)*x_ref[2]**o+(0.00934086955102)*x_ref[2]
        ref[2,0,0,1]=(-0.682315007143)*x_ref[0]**o+(-0.694069900861)*x_ref[0]+(-0.469082247747)*x_ref[1]**o+(-0.522165903301)*x_ref[1]+(-0.0997886221924)*x_ref[2]**o+(-0.954161552419)*x_ref[2]
        ref[2,0,1,0]=(0.780125539705)*x_ref[0]**o+(-0.445331450225)*x_ref[0]+(-0.282765510424)*x_ref[1]**o+(-0.0131898361592)*x_ref[1]+(-0.200186433923)*x_ref[2]**o+(0.715139435696)*x_ref[2]
        ref[2,0,1,1]=(-0.132310660786)*x_ref[0]**o+(-0.0981344234369)*x_ref[0]+(0.959297529475)*x_ref[1]**o+(-0.85944894515)*x_ref[1]+(-0.396258659619)*x_ref[2]**o+(-0.09672066185)*x_ref[2]
        ref[2,0,2,0]=(-0.344134014191)*x_ref[0]**o+(0.534345864287)*x_ref[0]+(0.797209509484)*x_ref[1]**o+(0.0479050101499)*x_ref[1]+(0.13620476987)*x_ref[2]**o+(0.82708115383)*x_ref[2]
        ref[2,0,2,1]=(-0.342509328238)*x_ref[0]**o+(-0.833164493006)*x_ref[0]+(0.472047721532)*x_ref[1]**o+(0.301613158637)*x_ref[1]+(0.858691679638)*x_ref[2]**o+(-0.691744164006)*x_ref[2]
        ref[2,1,0,0]=(0.498979390299)*x_ref[0]**o+(0.371513881824)*x_ref[0]+(-0.243506167971)*x_ref[1]**o+(-0.245423890439)*x_ref[1]+(-0.0494707840274)*x_ref[2]**o+(0.0965573709928)*x_ref[2]
        ref[2,1,0,1]=(0.665982455862)*x_ref[0]**o+(-0.409710866336)*x_ref[0]+(0.669257568046)*x_ref[1]**o+(0.711947704788)*x_ref[1]+(-0.0362568635785)*x_ref[2]**o+(-0.30394278188)*x_ref[2]
        ref[2,1,1,0]=(0.296245249928)*x_ref[0]**o+(0.132631019281)*x_ref[0]+(-0.376880152158)*x_ref[1]**o+(0.303184156241)*x_ref[1]+(0.836046560882)*x_ref[2]**o+(0.195205645106)*x_ref[2]
        ref[2,1,1,1]=(-0.971045820953)*x_ref[0]**o+(0.154708896467)*x_ref[0]+(-0.663844656602)*x_ref[1]**o+(-0.0873797588245)*x_ref[1]+(0.985123372542)*x_ref[2]**o+(0.349554922131)*x_ref[2]
        ref[2,1,2,0]=(-0.0117443431348)*x_ref[0]**o+(-0.71980503985)*x_ref[0]+(0.175348078994)*x_ref[1]**o+(-0.883987892161)*x_ref[1]+(0.0587275450458)*x_ref[2]**o+(0.464811715548)*x_ref[2]
        ref[2,1,2,1]=(-0.309202575765)*x_ref[0]**o+(-0.528513248453)*x_ref[0]+(-0.509298303972)*x_ref[1]**o+(-0.228278863496)*x_ref[1]+(0.933917053676)*x_ref[2]**o+(-0.823678942264)*x_ref[2]
        ref[2,2,0,0]=(0.374387386929)*x_ref[0]**o+(0.64969993847)*x_ref[0]+(-0.939822635641)*x_ref[1]**o+(0.578249430422)*x_ref[1]+(-0.973412660635)*x_ref[2]**o+(0.484052718386)*x_ref[2]
        ref[2,2,0,1]=(-0.972635162618)*x_ref[0]**o+(-0.255007433802)*x_ref[0]+(0.134165187097)*x_ref[1]**o+(0.313213615976)*x_ref[1]+(0.143298148686)*x_ref[2]**o+(0.901819440229)*x_ref[2]
        ref[2,2,1,0]=(0.868173256315)*x_ref[0]**o+(-0.532325343453)*x_ref[0]+(-0.712500880845)*x_ref[1]**o+(0.216943462308)*x_ref[1]+(0.733920091212)*x_ref[2]**o+(0.0992400603383)*x_ref[2]
        ref[2,2,1,1]=(0.642839350527)*x_ref[0]**o+(0.127721543559)*x_ref[0]+(-0.152686523032)*x_ref[1]**o+(-0.399032361234)*x_ref[1]+(0.317014518842)*x_ref[2]**o+(0.905538790115)*x_ref[2]
        ref[2,2,2,0]=(0.685267932413)*x_ref[0]**o+(-0.167571547782)*x_ref[0]+(-0.213446144405)*x_ref[1]**o+(0.420270456377)*x_ref[1]+(0.827725350621)*x_ref[2]**o+(-0.920242274399)*x_ref[2]
        ref[2,2,2,1]=(-0.507424593437)*x_ref[0]**o+(-0.694222840019)*x_ref[0]+(-0.904108970188)*x_ref[1]**o+(-0.750899614966)*x_ref[1]+(0.783997553302)*x_ref[2]**o+(0.196327513447)*x_ref[2]
        ref[2,3,0,0]=(-0.473900600479)*x_ref[0]**o+(-0.464347097789)*x_ref[0]+(-0.472423358671)*x_ref[1]**o+(0.373786456364)*x_ref[1]+(-0.989371300414)*x_ref[2]**o+(-0.959508065239)*x_ref[2]
        ref[2,3,0,1]=(0.174375189399)*x_ref[0]**o+(0.925503059886)*x_ref[0]+(-0.647274384723)*x_ref[1]**o+(-0.577149479679)*x_ref[1]+(0.171298877768)*x_ref[2]**o+(-0.936612100882)*x_ref[2]
        ref[2,3,1,0]=(0.482959343396)*x_ref[0]**o+(0.842243554011)*x_ref[0]+(0.814423633314)*x_ref[1]**o+(-0.989445050315)*x_ref[1]+(0.817933433065)*x_ref[2]**o+(0.552396881404)*x_ref[2]
        ref[2,3,1,1]=(0.55641736229)*x_ref[0]**o+(0.498094640058)*x_ref[0]+(0.0890113606037)*x_ref[1]**o+(0.36592723553)*x_ref[1]+(0.458607593386)*x_ref[2]**o+(0.603047420063)*x_ref[2]
        ref[2,3,2,0]=(0.884515080124)*x_ref[0]**o+(-0.561898381022)*x_ref[0]+(-0.292475503834)*x_ref[1]**o+(0.425634423478)*x_ref[1]+(0.65813323252)*x_ref[2]**o+(-0.548850399842)*x_ref[2]
        ref[2,3,2,1]=(-0.72895566982)*x_ref[0]**o+(-0.569405588005)*x_ref[0]+(0.117770875594)*x_ref[1]**o+(-0.33515735969)*x_ref[1]+(0.102062736382)*x_ref[2]**o+(0.611532417683)*x_ref[2]
        ref[2,4,0,0]=(-0.669208630424)*x_ref[0]**o+(0.944393014177)*x_ref[0]+(-0.00345511889683)*x_ref[1]**o+(0.0695279595783)*x_ref[1]+(0.496948250833)*x_ref[2]**o+(-0.899750463971)*x_ref[2]
        ref[2,4,0,1]=(-0.387828717292)*x_ref[0]**o+(0.621172267004)*x_ref[0]+(-0.280632674111)*x_ref[1]**o+(0.401702567575)*x_ref[1]+(-0.65554489378)*x_ref[2]**o+(0.403018047007)*x_ref[2]
        ref[2,4,1,0]=(-0.0770246925803)*x_ref[0]**o+(0.52913302745)*x_ref[0]+(0.922468724357)*x_ref[1]**o+(-0.304442750439)*x_ref[1]+(-0.556575079059)*x_ref[2]**o+(-0.725790672251)*x_ref[2]
        ref[2,4,1,1]=(-0.828378342911)*x_ref[0]**o+(-0.0696370821776)*x_ref[0]+(0.394018581667)*x_ref[1]**o+(-0.251099848269)*x_ref[1]+(-0.29878985656)*x_ref[2]**o+(0.925753371056)*x_ref[2]
        ref[2,4,2,0]=(-0.818277786029)*x_ref[0]**o+(0.225862354706)*x_ref[0]+(-0.629050212595)*x_ref[1]**o+(0.139186107062)*x_ref[1]+(-0.0134835519238)*x_ref[2]**o+(0.301899528641)*x_ref[2]
        ref[2,4,2,1]=(-0.783053684767)*x_ref[0]**o+(-0.0926938783599)*x_ref[0]+(0.519034210352)*x_ref[1]**o+(-0.494564657126)*x_ref[1]+(0.536117310622)*x_ref[2]**o+(0.6278756634)*x_ref[2]
        ref[3,0,0,0]=(-0.694353957685)*x_ref[0]**o+(-0.361037363965)*x_ref[0]+(-0.327004091885)*x_ref[1]**o+(-0.765735142884)*x_ref[1]+(-0.810915729713)*x_ref[2]**o+(0.0640539311609)*x_ref[2]
        ref[3,0,0,1]=(-0.131601548312)*x_ref[0]**o+(0.271587756708)*x_ref[0]+(0.322486168005)*x_ref[1]**o+(0.238086581801)*x_ref[1]+(-0.488081720555)*x_ref[2]**o+(0.607278995635)*x_ref[2]
        ref[3,0,1,0]=(0.903075418102)*x_ref[0]**o+(-0.738609834848)*x_ref[0]+(-0.165097989126)*x_ref[1]**o+(-0.25967140914)*x_ref[1]+(0.271649887428)*x_ref[2]**o+(-0.786131648727)*x_ref[2]
        ref[3,0,1,1]=(-0.963442989157)*x_ref[0]**o+(-0.240205602334)*x_ref[0]+(0.102153389731)*x_ref[1]**o+(-0.106402434995)*x_ref[1]+(-0.383097724211)*x_ref[2]**o+(0.680290963195)*x_ref[2]
        ref[3,0,2,0]=(-0.373162835796)*x_ref[0]**o+(-0.186323761814)*x_ref[0]+(0.661494315705)*x_ref[1]**o+(0.320072189494)*x_ref[1]+(0.0964443431067)*x_ref[2]**o+(0.829015838241)*x_ref[2]
        ref[3,0,2,1]=(0.527719514862)*x_ref[0]**o+(0.0819948261665)*x_ref[0]+(0.826598712178)*x_ref[1]**o+(-0.588102975125)*x_ref[1]+(-0.632584402292)*x_ref[2]**o+(0.0826992233702)*x_ref[2]
        ref[3,1,0,0]=(-0.102243686287)*x_ref[0]**o+(-0.497336112929)*x_ref[0]+(0.941827592628)*x_ref[1]**o+(0.68678991944)*x_ref[1]+(-0.508684841667)*x_ref[2]**o+(-0.752939965161)*x_ref[2]
        ref[3,1,0,1]=(-0.896417523183)*x_ref[0]**o+(0.759224781685)*x_ref[0]+(0.138537695823)*x_ref[1]**o+(0.466727480471)*x_ref[1]+(-0.838380824412)*x_ref[2]**o+(0.285761157398)*x_ref[2]
        ref[3,1,1,0]=(-0.547853182919)*x_ref[0]**o+(-0.41595413414)*x_ref[0]+(0.76894221584)*x_ref[1]**o+(-0.142394275192)*x_ref[1]+(-0.96349294581)*x_ref[2]**o+(-0.873639964693)*x_ref[2]
        ref[3,1,1,1]=(-0.151299403888)*x_ref[0]**o+(0.789406063878)*x_ref[0]+(0.163280670253)*x_ref[1]**o+(-0.287315494745)*x_ref[1]+(-0.627764670218)*x_ref[2]**o+(0.842131632598)*x_ref[2]
        ref[3,1,2,0]=(-0.957096489189)*x_ref[0]**o+(0.0616037805677)*x_ref[0]+(0.255858200166)*x_ref[1]**o+(-0.430555450798)*x_ref[1]+(0.730436930737)*x_ref[2]**o+(0.405094271653)*x_ref[2]
        ref[3,1,2,1]=(-0.924756832665)*x_ref[0]**o+(0.525754162391)*x_ref[0]+(0.281303228038)*x_ref[1]**o+(-0.479500161878)*x_ref[1]+(0.658416209181)*x_ref[2]**o+(-0.0413881595557)*x_ref[2]
        ref[3,2,0,0]=(0.516429381997)*x_ref[0]**o+(-0.8387318467)*x_ref[0]+(0.329773978227)*x_ref[1]**o+(0.599641753085)*x_ref[1]+(0.939937743271)*x_ref[2]**o+(-0.247331169626)*x_ref[2]
        ref[3,2,0,1]=(-0.644982044558)*x_ref[0]**o+(-0.242033352774)*x_ref[0]+(0.531835185779)*x_ref[1]**o+(-0.851565064183)*x_ref[1]+(-0.79213624915)*x_ref[2]**o+(0.0607365690732)*x_ref[2]
        ref[3,2,1,0]=(0.359781383537)*x_ref[0]**o+(-0.0529212842074)*x_ref[0]+(0.845915169452)*x_ref[1]**o+(-0.323979570458)*x_ref[1]+(0.828909997631)*x_ref[2]**o+(-0.839886339138)*x_ref[2]
        ref[3,2,1,1]=(0.559476630421)*x_ref[0]**o+(-0.189725214615)*x_ref[0]+(-0.00927446700413)*x_ref[1]**o+(-0.234703720782)*x_ref[1]+(-0.608362806755)*x_ref[2]**o+(0.968571765534)*x_ref[2]
        ref[3,2,2,0]=(0.924837507913)*x_ref[0]**o+(0.846397611121)*x_ref[0]+(-0.837765472403)*x_ref[1]**o+(0.0225105985057)*x_ref[1]+(-0.834187513844)*x_ref[2]**o+(0.194311845578)*x_ref[2]
        ref[3,2,2,1]=(-0.183561089961)*x_ref[0]**o+(0.0745071729471)*x_ref[0]+(-0.660415615202)*x_ref[1]**o+(-0.528184997345)*x_ref[1]+(0.205554094557)*x_ref[2]**o+(-0.0309246318944)*x_ref[2]
        ref[3,3,0,0]=(-0.67124095974)*x_ref[0]**o+(-0.48501497083)*x_ref[0]+(-0.0394345869094)*x_ref[1]**o+(0.586692270429)*x_ref[1]+(-0.593760749687)*x_ref[2]**o+(-0.124979751327)*x_ref[2]
        ref[3,3,0,1]=(0.778544173855)*x_ref[0]**o+(-0.931466478494)*x_ref[0]+(-0.746894777998)*x_ref[1]**o+(0.806417936297)*x_ref[1]+(0.989460929848)*x_ref[2]**o+(0.666534366369)*x_ref[2]
        ref[3,3,1,0]=(-0.798870838602)*x_ref[0]**o+(-0.294618868937)*x_ref[0]+(0.275029092968)*x_ref[1]**o+(0.45390580364)*x_ref[1]+(-0.311797818508)*x_ref[2]**o+(0.751606624393)*x_ref[2]
        ref[3,3,1,1]=(-0.924260987477)*x_ref[0]**o+(0.771970726902)*x_ref[0]+(-0.909099224213)*x_ref[1]**o+(-0.45204206766)*x_ref[1]+(0.630042614017)*x_ref[2]**o+(-0.990915130464)*x_ref[2]
        ref[3,3,2,0]=(0.734962154852)*x_ref[0]**o+(-0.00621898011475)*x_ref[0]+(-0.786362527189)*x_ref[1]**o+(0.623507947406)*x_ref[1]+(0.867175567104)*x_ref[2]**o+(-0.383283255448)*x_ref[2]
        ref[3,3,2,1]=(-0.955613156717)*x_ref[0]**o+(0.688013754841)*x_ref[0]+(-0.600559375208)*x_ref[1]**o+(0.453342626645)*x_ref[1]+(-0.864890477961)*x_ref[2]**o+(-0.458029873063)*x_ref[2]
        ref[3,4,0,0]=(-0.897010884834)*x_ref[0]**o+(-0.251223576458)*x_ref[0]+(-0.0338052047964)*x_ref[1]**o+(-0.903421466132)*x_ref[1]+(0.836864217874)*x_ref[2]**o+(0.660805905175)*x_ref[2]
        ref[3,4,0,1]=(-0.286025690215)*x_ref[0]**o+(0.746732975604)*x_ref[0]+(0.375810853218)*x_ref[1]**o+(-0.985192974869)*x_ref[1]+(-0.731463214455)*x_ref[2]**o+(-0.465450352444)*x_ref[2]
        ref[3,4,1,0]=(-0.500910000652)*x_ref[0]**o+(0.387502419727)*x_ref[0]+(0.962369164802)*x_ref[1]**o+(-0.10933824251)*x_ref[1]+(-0.33642703989)*x_ref[2]**o+(0.546851133256)*x_ref[2]
        ref[3,4,1,1]=(-0.117959459645)*x_ref[0]**o+(0.867699323392)*x_ref[0]+(0.950506403861)*x_ref[1]**o+(0.613806207745)*x_ref[1]+(0.425723173807)*x_ref[2]**o+(-0.906937548619)*x_ref[2]
        ref[3,4,2,0]=(0.331114892194)*x_ref[0]**o+(-0.168829643424)*x_ref[0]+(-0.855962311982)*x_ref[1]**o+(-0.144202406673)*x_ref[1]+(0.0769985713088)*x_ref[2]**o+(0.856003746423)*x_ref[2]
        ref[3,4,2,1]=(-0.439397439364)*x_ref[0]**o+(0.686116343481)*x_ref[0]+(0.519883854523)*x_ref[1]**o+(0.716231469289)*x_ref[1]+(0.270098231149)*x_ref[2]**o+(-0.448525436427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.787646538574)*x[0]+(0.912422904218)*x[1]
        ref=(-0.787646538574)*x_ref[0]+(0.912422904218)*x_ref[1]
      else:
        arg=(0.934044939024)*x[0]+(0.628477728603)*x[1]+(-0.786339405837)*x[2]
        ref=(0.934044939024)*x_ref[0]+(0.628477728603)*x_ref[1]+(-0.786339405837)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        arg[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
        ref[0]=(-0.588326094839)*x[0]+(1.56296941343)*x[1]
        ref[1]=(-0.802191222185)*x[0]+(-1.13471016062)*x[1]
      else:
        arg[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        arg[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
        ref[0]=(0.160702771321)*x[0]+(-0.0784425374814)*x[1]+(1.65226639722)*x[2]
        ref[1]=(0.745050820919)*x[0]+(-0.183859360073)*x[1]+(0.399813139488)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.797066914479)*x[0]+(-0.0269853164939)*x[1]
        arg[0,1]=(-1.07443661798)*x[0]+(0.818510216043)*x[1]
        arg[0,2]=(0.398735262179)*x[0]+(-0.580387872083)*x[1]
        arg[0,3]=(-0.003185613959)*x[0]+(-0.670506330615)*x[1]
        arg[0,4]=(0.385009570186)*x[0]+(0.858570512936)*x[1]
        arg[1,0]=(1.60917176753)*x[0]+(-0.0866362449992)*x[1]
        arg[1,1]=(-0.883782122452)*x[0]+(-0.402187705578)*x[1]
        arg[1,2]=(0.153893095122)*x[0]+(-1.19636362644)*x[1]
        arg[1,3]=(1.92041961031)*x[0]+(0.325857877636)*x[1]
        arg[1,4]=(-0.0947698843729)*x[0]+(-0.0663785758627)*x[1]
        arg[2,0]=(0.0232949467051)*x[0]+(1.60134210326)*x[1]
        arg[2,1]=(-0.521182882924)*x[0]+(-0.507401571343)*x[1]
        arg[2,2]=(0.833797117616)*x[0]+(-0.268886477194)*x[1]
        arg[2,3]=(-0.867762611227)*x[0]+(1.03727088566)*x[1]
        arg[2,4]=(0.380510651918)*x[0]+(0.350246888456)*x[1]
        arg[3,0]=(-1.10155500839)*x[0]+(0.414683749118)*x[1]
        arg[3,1]=(-0.0306733233552)*x[0]+(-0.351251014237)*x[1]
        arg[3,2]=(-0.0708773001498)*x[0]+(-0.326636587948)*x[1]
        arg[3,3]=(-1.32553119895)*x[0]+(-1.5852157477)*x[1]
        arg[3,4]=(1.40177154802)*x[0]+(-1.24804787994)*x[1]
        ref[0,0]=(-0.797066914479)*x_ref[0]+(-0.0269853164939)*x_ref[1]
        ref[0,1]=(-1.07443661798)*x_ref[0]+(0.818510216043)*x_ref[1]
        ref[0,2]=(0.398735262179)*x_ref[0]+(-0.580387872083)*x_ref[1]
        ref[0,3]=(-0.003185613959)*x_ref[0]+(-0.670506330615)*x_ref[1]
        ref[0,4]=(0.385009570186)*x_ref[0]+(0.858570512936)*x_ref[1]
        ref[1,0]=(1.60917176753)*x_ref[0]+(-0.0866362449992)*x_ref[1]
        ref[1,1]=(-0.883782122452)*x_ref[0]+(-0.402187705578)*x_ref[1]
        ref[1,2]=(0.153893095122)*x_ref[0]+(-1.19636362644)*x_ref[1]
        ref[1,3]=(1.92041961031)*x_ref[0]+(0.325857877636)*x_ref[1]
        ref[1,4]=(-0.0947698843729)*x_ref[0]+(-0.0663785758627)*x_ref[1]
        ref[2,0]=(0.0232949467051)*x_ref[0]+(1.60134210326)*x_ref[1]
        ref[2,1]=(-0.521182882924)*x_ref[0]+(-0.507401571343)*x_ref[1]
        ref[2,2]=(0.833797117616)*x_ref[0]+(-0.268886477194)*x_ref[1]
        ref[2,3]=(-0.867762611227)*x_ref[0]+(1.03727088566)*x_ref[1]
        ref[2,4]=(0.380510651918)*x_ref[0]+(0.350246888456)*x_ref[1]
        ref[3,0]=(-1.10155500839)*x_ref[0]+(0.414683749118)*x_ref[1]
        ref[3,1]=(-0.0306733233552)*x_ref[0]+(-0.351251014237)*x_ref[1]
        ref[3,2]=(-0.0708773001498)*x_ref[0]+(-0.326636587948)*x_ref[1]
        ref[3,3]=(-1.32553119895)*x_ref[0]+(-1.5852157477)*x_ref[1]
        ref[3,4]=(1.40177154802)*x_ref[0]+(-1.24804787994)*x_ref[1]
      else:
        arg[0,0]=(0.000362194678523)*x[0]+(-1.63949220671)*x[1]+(-0.176997156562)*x[2]
        arg[0,1]=(-0.357589049408)*x[0]+(0.828191515325)*x[1]+(-0.269706780136)*x[2]
        arg[0,2]=(0.886395070544)*x[0]+(0.815688451839)*x[1]+(-0.658412270558)*x[2]
        arg[0,3]=(-0.222573318026)*x[0]+(0.137763803092)*x[1]+(0.171902150187)*x[2]
        arg[0,4]=(1.05159741251)*x[0]+(0.923836841038)*x[1]+(-0.426590801731)*x[2]
        arg[1,0]=(-0.00718419057788)*x[0]+(1.55809964998)*x[1]+(-0.912766657496)*x[2]
        arg[1,1]=(1.30769045785)*x[0]+(-0.0356609338389)*x[1]+(0.214543084134)*x[2]
        arg[1,2]=(-1.56763014241)*x[0]+(-0.582970995722)*x[1]+(-0.513972924752)*x[2]
        arg[1,3]=(0.118027736928)*x[0]+(-0.28795014876)*x[1]+(0.620718623034)*x[2]
        arg[1,4]=(-0.784496170193)*x[0]+(0.233037866924)*x[1]+(-0.0500718134572)*x[2]
        arg[2,0]=(0.800851281806)*x[0]+(0.219346484846)*x[1]+(1.19125787701)*x[2]
        arg[2,1]=(-1.45025879429)*x[0]+(-1.32011538154)*x[1]+(0.47946669308)*x[2]
        arg[2,2]=(-0.231377110469)*x[0]+(-0.921551695999)*x[1]+(-0.298964635841)*x[2]
        arg[2,3]=(0.965766335546)*x[0]+(0.424791610089)*x[1]+(1.01860652092)*x[2]
        arg[2,4]=(0.676247093636)*x[0]+(-1.50802741511)*x[1]+(0.0385960918391)*x[2]
        arg[3,0]=(0.322378151756)*x[0]+(-0.483633005509)*x[1]+(-0.852371707865)*x[2]
        arg[3,1]=(-0.241998307153)*x[0]+(-1.08134340082)*x[1]+(-0.849574532537)*x[2]
        arg[3,2]=(-0.0185751405467)*x[0]+(0.179558416001)*x[1]+(-0.488727308548)*x[2]
        arg[3,3]=(0.0984555057845)*x[0]+(1.03085967788)*x[1]+(-0.357009094371)*x[2]
        arg[3,4]=(0.430593980121)*x[0]+(0.689565614489)*x[1]+(-0.0916541808491)*x[2]
        ref[0,0]=(0.000362194678523)*x_ref[0]+(-1.63949220671)*x_ref[1]+(-0.176997156562)*x_ref[2]
        ref[0,1]=(-0.357589049408)*x_ref[0]+(0.828191515325)*x_ref[1]+(-0.269706780136)*x_ref[2]
        ref[0,2]=(0.886395070544)*x_ref[0]+(0.815688451839)*x_ref[1]+(-0.658412270558)*x_ref[2]
        ref[0,3]=(-0.222573318026)*x_ref[0]+(0.137763803092)*x_ref[1]+(0.171902150187)*x_ref[2]
        ref[0,4]=(1.05159741251)*x_ref[0]+(0.923836841038)*x_ref[1]+(-0.426590801731)*x_ref[2]
        ref[1,0]=(-0.00718419057788)*x_ref[0]+(1.55809964998)*x_ref[1]+(-0.912766657496)*x_ref[2]
        ref[1,1]=(1.30769045785)*x_ref[0]+(-0.0356609338389)*x_ref[1]+(0.214543084134)*x_ref[2]
        ref[1,2]=(-1.56763014241)*x_ref[0]+(-0.582970995722)*x_ref[1]+(-0.513972924752)*x_ref[2]
        ref[1,3]=(0.118027736928)*x_ref[0]+(-0.28795014876)*x_ref[1]+(0.620718623034)*x_ref[2]
        ref[1,4]=(-0.784496170193)*x_ref[0]+(0.233037866924)*x_ref[1]+(-0.0500718134572)*x_ref[2]
        ref[2,0]=(0.800851281806)*x_ref[0]+(0.219346484846)*x_ref[1]+(1.19125787701)*x_ref[2]
        ref[2,1]=(-1.45025879429)*x_ref[0]+(-1.32011538154)*x_ref[1]+(0.47946669308)*x_ref[2]
        ref[2,2]=(-0.231377110469)*x_ref[0]+(-0.921551695999)*x_ref[1]+(-0.298964635841)*x_ref[2]
        ref[2,3]=(0.965766335546)*x_ref[0]+(0.424791610089)*x_ref[1]+(1.01860652092)*x_ref[2]
        ref[2,4]=(0.676247093636)*x_ref[0]+(-1.50802741511)*x_ref[1]+(0.0385960918391)*x_ref[2]
        ref[3,0]=(0.322378151756)*x_ref[0]+(-0.483633005509)*x_ref[1]+(-0.852371707865)*x_ref[2]
        ref[3,1]=(-0.241998307153)*x_ref[0]+(-1.08134340082)*x_ref[1]+(-0.849574532537)*x_ref[2]
        ref[3,2]=(-0.0185751405467)*x_ref[0]+(0.179558416001)*x_ref[1]+(-0.488727308548)*x_ref[2]
        ref[3,3]=(0.0984555057845)*x_ref[0]+(1.03085967788)*x_ref[1]+(-0.357009094371)*x_ref[2]
        ref[3,4]=(0.430593980121)*x_ref[0]+(0.689565614489)*x_ref[1]+(-0.0916541808491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0641544023301)*x[0]+(0.364286357763)*x[1]
        arg[0,0,1]=(-0.858288240411)*x[0]+(-0.847764393559)*x[1]
        arg[0,1,0]=(-0.656948775029)*x[0]+(1.69748608992)*x[1]
        arg[0,1,1]=(0.579399191549)*x[0]+(1.48407368783)*x[1]
        arg[1,0,0]=(-0.35948657403)*x[0]+(0.795436090765)*x[1]
        arg[1,0,1]=(-1.32807156246)*x[0]+(0.921960610294)*x[1]
        arg[1,1,0]=(1.0904085219)*x[0]+(-1.41870997296)*x[1]
        arg[1,1,1]=(-0.677691596793)*x[0]+(-0.395373706333)*x[1]
        arg[2,0,0]=(0.295591917216)*x[0]+(-1.75550830457)*x[1]
        arg[2,0,1]=(0.506665326626)*x[0]+(0.273265713965)*x[1]
        arg[2,1,0]=(0.543899865352)*x[0]+(-0.0678457091657)*x[1]
        arg[2,1,1]=(-0.887824769535)*x[0]+(0.527475538055)*x[1]
        arg[3,0,0]=(-0.788511557365)*x[0]+(-0.635972100675)*x[1]
        arg[3,0,1]=(-1.02799066459)*x[0]+(-0.416241247725)*x[1]
        arg[3,1,0]=(-0.310072483091)*x[0]+(1.2445160445)*x[1]
        arg[3,1,1]=(0.912462196567)*x[0]+(0.631855269783)*x[1]
        arg[4,0,0]=(0.700426861807)*x[0]+(-0.491365913713)*x[1]
        arg[4,0,1]=(0.628791538338)*x[0]+(0.640767645586)*x[1]
        arg[4,1,0]=(0.370886720329)*x[0]+(0.823663448834)*x[1]
        arg[4,1,1]=(1.75857310774)*x[0]+(0.445439424641)*x[1]
        arg[5,0,0]=(1.39637385066)*x[0]+(-0.812211860169)*x[1]
        arg[5,0,1]=(0.865050350195)*x[0]+(-1.25983380375)*x[1]
        arg[5,1,0]=(-0.0429057645445)*x[0]+(-1.08490578934)*x[1]
        arg[5,1,1]=(1.42851264594)*x[0]+(1.52192971749)*x[1]
        ref[0,0,0]=(-0.0641544023301)*x_ref[0]+(0.364286357763)*x_ref[1]
        ref[0,0,1]=(-0.858288240411)*x_ref[0]+(-0.847764393559)*x_ref[1]
        ref[0,1,0]=(-0.656948775029)*x_ref[0]+(1.69748608992)*x_ref[1]
        ref[0,1,1]=(0.579399191549)*x_ref[0]+(1.48407368783)*x_ref[1]
        ref[1,0,0]=(-0.35948657403)*x_ref[0]+(0.795436090765)*x_ref[1]
        ref[1,0,1]=(-1.32807156246)*x_ref[0]+(0.921960610294)*x_ref[1]
        ref[1,1,0]=(1.0904085219)*x_ref[0]+(-1.41870997296)*x_ref[1]
        ref[1,1,1]=(-0.677691596793)*x_ref[0]+(-0.395373706333)*x_ref[1]
        ref[2,0,0]=(0.295591917216)*x_ref[0]+(-1.75550830457)*x_ref[1]
        ref[2,0,1]=(0.506665326626)*x_ref[0]+(0.273265713965)*x_ref[1]
        ref[2,1,0]=(0.543899865352)*x_ref[0]+(-0.0678457091657)*x_ref[1]
        ref[2,1,1]=(-0.887824769535)*x_ref[0]+(0.527475538055)*x_ref[1]
        ref[3,0,0]=(-0.788511557365)*x_ref[0]+(-0.635972100675)*x_ref[1]
        ref[3,0,1]=(-1.02799066459)*x_ref[0]+(-0.416241247725)*x_ref[1]
        ref[3,1,0]=(-0.310072483091)*x_ref[0]+(1.2445160445)*x_ref[1]
        ref[3,1,1]=(0.912462196567)*x_ref[0]+(0.631855269783)*x_ref[1]
        ref[4,0,0]=(0.700426861807)*x_ref[0]+(-0.491365913713)*x_ref[1]
        ref[4,0,1]=(0.628791538338)*x_ref[0]+(0.640767645586)*x_ref[1]
        ref[4,1,0]=(0.370886720329)*x_ref[0]+(0.823663448834)*x_ref[1]
        ref[4,1,1]=(1.75857310774)*x_ref[0]+(0.445439424641)*x_ref[1]
        ref[5,0,0]=(1.39637385066)*x_ref[0]+(-0.812211860169)*x_ref[1]
        ref[5,0,1]=(0.865050350195)*x_ref[0]+(-1.25983380375)*x_ref[1]
        ref[5,1,0]=(-0.0429057645445)*x_ref[0]+(-1.08490578934)*x_ref[1]
        ref[5,1,1]=(1.42851264594)*x_ref[0]+(1.52192971749)*x_ref[1]
      else:
        arg[0,0,0]=(0.731194732468)*x[0]+(-0.0822115660022)*x[1]+(-0.403175091289)*x[2]
        arg[0,0,1]=(0.39292230307)*x[0]+(1.48598681615)*x[1]+(1.05545374107)*x[2]
        arg[0,1,0]=(0.306844260706)*x[0]+(0.836496026763)*x[1]+(-0.121120983155)*x[2]
        arg[0,1,1]=(-0.155245802062)*x[0]+(0.800232934864)*x[1]+(0.373397160636)*x[2]
        arg[1,0,0]=(0.155941790359)*x[0]+(0.826821969281)*x[1]+(0.849596807425)*x[2]
        arg[1,0,1]=(-0.0988854656187)*x[0]+(-0.855032499842)*x[1]+(0.141988288566)*x[2]
        arg[1,1,0]=(-0.761032539296)*x[0]+(-0.0892979073234)*x[1]+(1.93034385132)*x[2]
        arg[1,1,1]=(-1.08303603558)*x[0]+(-0.487250117659)*x[1]+(0.0358501036173)*x[2]
        arg[2,0,0]=(0.148391465413)*x[0]+(-0.731963025625)*x[1]+(0.484236965689)*x[2]
        arg[2,0,1]=(1.06691217722)*x[0]+(-0.638095266244)*x[1]+(-0.971838410779)*x[2]
        arg[2,1,0]=(-0.377750829061)*x[0]+(-0.101467967263)*x[1]+(-1.54241387349)*x[2]
        arg[2,1,1]=(0.62210350231)*x[0]+(0.0435334873911)*x[1]+(-0.299303646623)*x[2]
        arg[3,0,0]=(0.523811147108)*x[0]+(-1.09709839108)*x[1]+(1.83592280935)*x[2]
        arg[3,0,1]=(0.957738143475)*x[0]+(0.282158440867)*x[1]+(0.621694180522)*x[2]
        arg[3,1,0]=(0.366557891822)*x[0]+(0.812194047631)*x[1]+(-0.245067876349)*x[2]
        arg[3,1,1]=(0.604859464982)*x[0]+(1.08415514854)*x[1]+(-1.470128848)*x[2]
        arg[4,0,0]=(-0.13031502499)*x[0]+(-1.20707238733)*x[1]+(0.108382497552)*x[2]
        arg[4,0,1]=(1.30453246317)*x[0]+(1.59696612588)*x[1]+(-0.234259043697)*x[2]
        arg[4,1,0]=(-0.233494640836)*x[0]+(1.10701061816)*x[1]+(-0.409927013674)*x[2]
        arg[4,1,1]=(0.102708240266)*x[0]+(-0.381598345002)*x[1]+(-1.39077748656)*x[2]
        arg[5,0,0]=(1.06217404493)*x[0]+(1.1600791444)*x[1]+(-1.45242735389)*x[2]
        arg[5,0,1]=(-0.423024189506)*x[0]+(-0.0868948651641)*x[1]+(0.0894539043522)*x[2]
        arg[5,1,0]=(1.17650613067)*x[0]+(0.993832919965)*x[1]+(1.16546768569)*x[2]
        arg[5,1,1]=(-0.61549132897)*x[0]+(-0.499619662734)*x[1]+(-0.643422281669)*x[2]
        ref[0,0,0]=(0.731194732468)*x_ref[0]+(-0.0822115660022)*x_ref[1]+(-0.403175091289)*x_ref[2]
        ref[0,0,1]=(0.39292230307)*x_ref[0]+(1.48598681615)*x_ref[1]+(1.05545374107)*x_ref[2]
        ref[0,1,0]=(0.306844260706)*x_ref[0]+(0.836496026763)*x_ref[1]+(-0.121120983155)*x_ref[2]
        ref[0,1,1]=(-0.155245802062)*x_ref[0]+(0.800232934864)*x_ref[1]+(0.373397160636)*x_ref[2]
        ref[1,0,0]=(0.155941790359)*x_ref[0]+(0.826821969281)*x_ref[1]+(0.849596807425)*x_ref[2]
        ref[1,0,1]=(-0.0988854656187)*x_ref[0]+(-0.855032499842)*x_ref[1]+(0.141988288566)*x_ref[2]
        ref[1,1,0]=(-0.761032539296)*x_ref[0]+(-0.0892979073234)*x_ref[1]+(1.93034385132)*x_ref[2]
        ref[1,1,1]=(-1.08303603558)*x_ref[0]+(-0.487250117659)*x_ref[1]+(0.0358501036173)*x_ref[2]
        ref[2,0,0]=(0.148391465413)*x_ref[0]+(-0.731963025625)*x_ref[1]+(0.484236965689)*x_ref[2]
        ref[2,0,1]=(1.06691217722)*x_ref[0]+(-0.638095266244)*x_ref[1]+(-0.971838410779)*x_ref[2]
        ref[2,1,0]=(-0.377750829061)*x_ref[0]+(-0.101467967263)*x_ref[1]+(-1.54241387349)*x_ref[2]
        ref[2,1,1]=(0.62210350231)*x_ref[0]+(0.0435334873911)*x_ref[1]+(-0.299303646623)*x_ref[2]
        ref[3,0,0]=(0.523811147108)*x_ref[0]+(-1.09709839108)*x_ref[1]+(1.83592280935)*x_ref[2]
        ref[3,0,1]=(0.957738143475)*x_ref[0]+(0.282158440867)*x_ref[1]+(0.621694180522)*x_ref[2]
        ref[3,1,0]=(0.366557891822)*x_ref[0]+(0.812194047631)*x_ref[1]+(-0.245067876349)*x_ref[2]
        ref[3,1,1]=(0.604859464982)*x_ref[0]+(1.08415514854)*x_ref[1]+(-1.470128848)*x_ref[2]
        ref[4,0,0]=(-0.13031502499)*x_ref[0]+(-1.20707238733)*x_ref[1]+(0.108382497552)*x_ref[2]
        ref[4,0,1]=(1.30453246317)*x_ref[0]+(1.59696612588)*x_ref[1]+(-0.234259043697)*x_ref[2]
        ref[4,1,0]=(-0.233494640836)*x_ref[0]+(1.10701061816)*x_ref[1]+(-0.409927013674)*x_ref[2]
        ref[4,1,1]=(0.102708240266)*x_ref[0]+(-0.381598345002)*x_ref[1]+(-1.39077748656)*x_ref[2]
        ref[5,0,0]=(1.06217404493)*x_ref[0]+(1.1600791444)*x_ref[1]+(-1.45242735389)*x_ref[2]
        ref[5,0,1]=(-0.423024189506)*x_ref[0]+(-0.0868948651641)*x_ref[1]+(0.0894539043522)*x_ref[2]
        ref[5,1,0]=(1.17650613067)*x_ref[0]+(0.993832919965)*x_ref[1]+(1.16546768569)*x_ref[2]
        ref[5,1,1]=(-0.61549132897)*x_ref[0]+(-0.499619662734)*x_ref[1]+(-0.643422281669)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactZero
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.227969627252)*x[0]+(0.778177350248)*x[1]
        arg[0,0,0,1]=(-0.684883711494)*x[0]+(-0.574285957132)*x[1]
        arg[0,0,1,0]=(0.483090608974)*x[0]+(0.0779929087959)*x[1]
        arg[0,0,1,1]=(-0.293150814932)*x[0]+(1.35368931993)*x[1]
        arg[0,0,2,0]=(-0.190664734662)*x[0]+(0.910883386754)*x[1]
        arg[0,0,2,1]=(-1.51255575457)*x[0]+(0.421560181501)*x[1]
        arg[0,1,0,0]=(-0.833402828032)*x[0]+(-0.400466806781)*x[1]
        arg[0,1,0,1]=(-0.212099323605)*x[0]+(-0.112478697221)*x[1]
        arg[0,1,1,0]=(0.57687662926)*x[0]+(0.869758954551)*x[1]
        arg[0,1,1,1]=(1.37261769917)*x[0]+(1.37855789907)*x[1]
        arg[0,1,2,0]=(-1.02838589722)*x[0]+(-0.280409246204)*x[1]
        arg[0,1,2,1]=(1.2333373468)*x[0]+(0.337631183798)*x[1]
        arg[0,2,0,0]=(-0.249065448772)*x[0]+(0.834483646317)*x[1]
        arg[0,2,0,1]=(-1.30860297683)*x[0]+(-0.653929967017)*x[1]
        arg[0,2,1,0]=(-0.0155994484453)*x[0]+(-0.274023166937)*x[1]
        arg[0,2,1,1]=(0.31458360335)*x[0]+(1.59507700673)*x[1]
        arg[0,2,2,0]=(-0.11442619553)*x[0]+(1.6577028576)*x[1]
        arg[0,2,2,1]=(0.274261132032)*x[0]+(1.15236962479)*x[1]
        arg[0,3,0,0]=(1.91433860239)*x[0]+(-0.316349747811)*x[1]
        arg[0,3,0,1]=(-1.42457591054)*x[0]+(-0.362812690805)*x[1]
        arg[0,3,1,0]=(0.572734883096)*x[0]+(0.424724787283)*x[1]
        arg[0,3,1,1]=(-0.118538376813)*x[0]+(0.25190312527)*x[1]
        arg[0,3,2,0]=(-0.533409510081)*x[0]+(-0.804349107915)*x[1]
        arg[0,3,2,1]=(-0.290555363666)*x[0]+(-0.912641477145)*x[1]
        arg[0,4,0,0]=(-0.469612775918)*x[0]+(-0.0447411310793)*x[1]
        arg[0,4,0,1]=(0.108538258831)*x[0]+(0.0453962016662)*x[1]
        arg[0,4,1,0]=(-0.130785944028)*x[0]+(0.0887280914032)*x[1]
        arg[0,4,1,1]=(0.166412309585)*x[0]+(-0.75130797706)*x[1]
        arg[0,4,2,0]=(1.22538232455)*x[0]+(0.112925588343)*x[1]
        arg[0,4,2,1]=(-1.02455171854)*x[0]+(-0.725754942749)*x[1]
        arg[1,0,0,0]=(-0.869171086952)*x[0]+(-0.948876783682)*x[1]
        arg[1,0,0,1]=(0.17730673689)*x[0]+(-0.00166180855579)*x[1]
        arg[1,0,1,0]=(0.482072704433)*x[0]+(0.895847096783)*x[1]
        arg[1,0,1,1]=(0.695736834617)*x[0]+(0.358082704558)*x[1]
        arg[1,0,2,0]=(-0.501799636565)*x[0]+(0.363612558025)*x[1]
        arg[1,0,2,1]=(-1.47204447226)*x[0]+(1.91281426121)*x[1]
        arg[1,1,0,0]=(0.241893182978)*x[0]+(0.95199758669)*x[1]
        arg[1,1,0,1]=(-0.68112361827)*x[0]+(-0.241935141301)*x[1]
        arg[1,1,1,0]=(-0.553030972306)*x[0]+(0.0244503690439)*x[1]
        arg[1,1,1,1]=(0.278177176896)*x[0]+(1.58693846427)*x[1]
        arg[1,1,2,0]=(-0.87980015517)*x[0]+(0.315618837123)*x[1]
        arg[1,1,2,1]=(-0.335783961379)*x[0]+(0.368650136138)*x[1]
        arg[1,2,0,0]=(-0.903535146802)*x[0]+(1.68182923194)*x[1]
        arg[1,2,0,1]=(1.06511511382)*x[0]+(-1.08524762302)*x[1]
        arg[1,2,1,0]=(-0.0455274784681)*x[0]+(-0.519576710661)*x[1]
        arg[1,2,1,1]=(0.735100217062)*x[0]+(-0.986268682671)*x[1]
        arg[1,2,2,0]=(-0.544610797462)*x[0]+(0.447237114882)*x[1]
        arg[1,2,2,1]=(0.385242031153)*x[0]+(-0.960538844247)*x[1]
        arg[1,3,0,0]=(-0.354927072897)*x[0]+(0.842574324062)*x[1]
        arg[1,3,0,1]=(0.319036427919)*x[0]+(-0.156282262536)*x[1]
        arg[1,3,1,0]=(0.587123807717)*x[0]+(1.65277990726)*x[1]
        arg[1,3,1,1]=(0.791697370298)*x[0]+(-1.5681680344)*x[1]
        arg[1,3,2,0]=(-1.41931242045)*x[0]+(0.351285136557)*x[1]
        arg[1,3,2,1]=(0.53175704402)*x[0]+(0.987523241324)*x[1]
        arg[1,4,0,0]=(0.00533504818465)*x[0]+(-0.158886930817)*x[1]
        arg[1,4,0,1]=(1.23725297303)*x[0]+(-0.302896452172)*x[1]
        arg[1,4,1,0]=(0.433953146329)*x[0]+(-0.0721602234887)*x[1]
        arg[1,4,1,1]=(-0.383499726865)*x[0]+(-0.253834671235)*x[1]
        arg[1,4,2,0]=(-0.480776695247)*x[0]+(-0.302145524759)*x[1]
        arg[1,4,2,1]=(1.14113180903)*x[0]+(-1.24304726003)*x[1]
        arg[2,0,0,0]=(0.513776013893)*x[0]+(-0.16660259885)*x[1]
        arg[2,0,0,1]=(1.55739209698)*x[0]+(0.0524084654531)*x[1]
        arg[2,0,1,0]=(0.222567881446)*x[0]+(-0.357158470388)*x[1]
        arg[2,0,1,1]=(0.0179816596229)*x[0]+(0.231534460784)*x[1]
        arg[2,0,2,0]=(1.58460345297)*x[0]+(-0.383588433888)*x[1]
        arg[2,0,2,1]=(-1.73386465536)*x[0]+(-0.743326657544)*x[1]
        arg[2,1,0,0]=(-0.737976950246)*x[0]+(-1.01053396224)*x[1]
        arg[2,1,0,1]=(0.926911190393)*x[0]+(0.620251783925)*x[1]
        arg[2,1,1,0]=(-1.34289630129)*x[0]+(-0.429933681523)*x[1]
        arg[2,1,1,1]=(1.79530806415)*x[0]+(-0.539783957021)*x[1]
        arg[2,1,2,0]=(0.237950974938)*x[0]+(1.93783427703)*x[1]
        arg[2,1,2,1]=(-0.439640741973)*x[0]+(-0.61543984673)*x[1]
        arg[2,2,0,0]=(-1.66403612438)*x[0]+(1.00839566441)*x[1]
        arg[2,2,0,1]=(0.249635983351)*x[0]+(-0.722261363675)*x[1]
        arg[2,2,1,0]=(1.34179674459)*x[0]+(0.502542144188)*x[1]
        arg[2,2,1,1]=(-0.815862586074)*x[0]+(-0.110688273707)*x[1]
        arg[2,2,2,0]=(-0.686810719526)*x[0]+(0.500406563421)*x[1]
        arg[2,2,2,1]=(0.681231089598)*x[0]+(0.605219505613)*x[1]
        arg[2,3,0,0]=(-0.953299029956)*x[0]+(-0.223725858372)*x[1]
        arg[2,3,0,1]=(0.831380756684)*x[0]+(-0.0379777076335)*x[1]
        arg[2,3,1,0]=(-1.63804393157)*x[0]+(0.148235731301)*x[1]
        arg[2,3,1,1]=(-0.17786458643)*x[0]+(-1.77342152699)*x[1]
        arg[2,3,2,0]=(0.0210446244192)*x[0]+(1.59778781524)*x[1]
        arg[2,3,2,1]=(0.877162721214)*x[0]+(0.262990448614)*x[1]
        arg[2,4,0,0]=(-0.650265970331)*x[0]+(-1.07003604369)*x[1]
        arg[2,4,0,1]=(1.12940101029)*x[0]+(0.860589290891)*x[1]
        arg[2,4,1,0]=(0.520056210056)*x[0]+(-1.47691193979)*x[1]
        arg[2,4,1,1]=(1.26823535235)*x[0]+(0.496811720647)*x[1]
        arg[2,4,2,0]=(-1.388913443)*x[0]+(0.585343197943)*x[1]
        arg[2,4,2,1]=(-0.762904547594)*x[0]+(0.503438815917)*x[1]
        arg[3,0,0,0]=(0.382862686993)*x[0]+(-0.923525513303)*x[1]
        arg[3,0,0,1]=(-0.54305794485)*x[0]+(0.467513531444)*x[1]
        arg[3,0,1,0]=(1.24228759776)*x[0]+(-1.65740117421)*x[1]
        arg[3,0,1,1]=(-0.463045362742)*x[0]+(0.186438587087)*x[1]
        arg[3,0,2,0]=(0.296881980828)*x[0]+(-0.385878814949)*x[1]
        arg[3,0,2,1]=(-0.299705899564)*x[0]+(0.490736038538)*x[1]
        arg[3,1,0,0]=(-0.786851428658)*x[0]+(-0.760593794308)*x[1]
        arg[3,1,0,1]=(0.506851336483)*x[0]+(0.995163951197)*x[1]
        arg[3,1,1,0]=(-0.678883937341)*x[0]+(0.56996778248)*x[1]
        arg[3,1,1,1]=(-0.401483642528)*x[0]+(0.0476797332537)*x[1]
        arg[3,1,2,0]=(-0.585357955375)*x[0]+(0.330103647539)*x[1]
        arg[3,1,2,1]=(-0.355539976825)*x[0]+(0.416287924349)*x[1]
        arg[3,2,0,0]=(-0.662837130286)*x[0]+(-0.26640899674)*x[1]
        arg[3,2,0,1]=(0.397553900588)*x[0]+(-1.21708620294)*x[1]
        arg[3,2,1,0]=(0.464963326429)*x[0]+(-0.774515114142)*x[1]
        arg[3,2,1,1]=(0.834351469791)*x[0]+(0.487629540715)*x[1]
        arg[3,2,2,0]=(0.27622693716)*x[0]+(0.811338337719)*x[1]
        arg[3,2,2,1]=(0.0384997772653)*x[0]+(1.03724996578)*x[1]
        arg[3,3,0,0]=(-0.805491987874)*x[0]+(-0.476384547853)*x[1]
        arg[3,3,0,1]=(0.615551780195)*x[0]+(-0.312805967691)*x[1]
        arg[3,3,1,0]=(0.348666798955)*x[0]+(-1.31271156524)*x[1]
        arg[3,3,1,1]=(-0.423569975374)*x[0]+(0.930713125389)*x[1]
        arg[3,3,2,0]=(0.325632686234)*x[0]+(-0.166401033259)*x[1]
        arg[3,3,2,1]=(0.126924861797)*x[0]+(-0.425380897969)*x[1]
        arg[3,4,0,0]=(0.169532204538)*x[0]+(0.614632525765)*x[1]
        arg[3,4,0,1]=(-0.938098696851)*x[0]+(0.648854147587)*x[1]
        arg[3,4,1,0]=(-0.545143970434)*x[0]+(-1.59787515966)*x[1]
        arg[3,4,1,1]=(-0.711190914978)*x[0]+(0.897331522911)*x[1]
        arg[3,4,2,0]=(-0.386498124525)*x[0]+(0.299740420711)*x[1]
        arg[3,4,2,1]=(0.617683032062)*x[0]+(-0.449587703061)*x[1]
        ref[0,0,0,0]=(0.227969627252)*x_ref[0]+(0.778177350248)*x_ref[1]
        ref[0,0,0,1]=(-0.684883711494)*x_ref[0]+(-0.574285957132)*x_ref[1]
        ref[0,0,1,0]=(0.483090608974)*x_ref[0]+(0.0779929087959)*x_ref[1]
        ref[0,0,1,1]=(-0.293150814932)*x_ref[0]+(1.35368931993)*x_ref[1]
        ref[0,0,2,0]=(-0.190664734662)*x_ref[0]+(0.910883386754)*x_ref[1]
        ref[0,0,2,1]=(-1.51255575457)*x_ref[0]+(0.421560181501)*x_ref[1]
        ref[0,1,0,0]=(-0.833402828032)*x_ref[0]+(-0.400466806781)*x_ref[1]
        ref[0,1,0,1]=(-0.212099323605)*x_ref[0]+(-0.112478697221)*x_ref[1]
        ref[0,1,1,0]=(0.57687662926)*x_ref[0]+(0.869758954551)*x_ref[1]
        ref[0,1,1,1]=(1.37261769917)*x_ref[0]+(1.37855789907)*x_ref[1]
        ref[0,1,2,0]=(-1.02838589722)*x_ref[0]+(-0.280409246204)*x_ref[1]
        ref[0,1,2,1]=(1.2333373468)*x_ref[0]+(0.337631183798)*x_ref[1]
        ref[0,2,0,0]=(-0.249065448772)*x_ref[0]+(0.834483646317)*x_ref[1]
        ref[0,2,0,1]=(-1.30860297683)*x_ref[0]+(-0.653929967017)*x_ref[1]
        ref[0,2,1,0]=(-0.0155994484453)*x_ref[0]+(-0.274023166937)*x_ref[1]
        ref[0,2,1,1]=(0.31458360335)*x_ref[0]+(1.59507700673)*x_ref[1]
        ref[0,2,2,0]=(-0.11442619553)*x_ref[0]+(1.6577028576)*x_ref[1]
        ref[0,2,2,1]=(0.274261132032)*x_ref[0]+(1.15236962479)*x_ref[1]
        ref[0,3,0,0]=(1.91433860239)*x_ref[0]+(-0.316349747811)*x_ref[1]
        ref[0,3,0,1]=(-1.42457591054)*x_ref[0]+(-0.362812690805)*x_ref[1]
        ref[0,3,1,0]=(0.572734883096)*x_ref[0]+(0.424724787283)*x_ref[1]
        ref[0,3,1,1]=(-0.118538376813)*x_ref[0]+(0.25190312527)*x_ref[1]
        ref[0,3,2,0]=(-0.533409510081)*x_ref[0]+(-0.804349107915)*x_ref[1]
        ref[0,3,2,1]=(-0.290555363666)*x_ref[0]+(-0.912641477145)*x_ref[1]
        ref[0,4,0,0]=(-0.469612775918)*x_ref[0]+(-0.0447411310793)*x_ref[1]
        ref[0,4,0,1]=(0.108538258831)*x_ref[0]+(0.0453962016662)*x_ref[1]
        ref[0,4,1,0]=(-0.130785944028)*x_ref[0]+(0.0887280914032)*x_ref[1]
        ref[0,4,1,1]=(0.166412309585)*x_ref[0]+(-0.75130797706)*x_ref[1]
        ref[0,4,2,0]=(1.22538232455)*x_ref[0]+(0.112925588343)*x_ref[1]
        ref[0,4,2,1]=(-1.02455171854)*x_ref[0]+(-0.725754942749)*x_ref[1]
        ref[1,0,0,0]=(-0.869171086952)*x_ref[0]+(-0.948876783682)*x_ref[1]
        ref[1,0,0,1]=(0.17730673689)*x_ref[0]+(-0.00166180855579)*x_ref[1]
        ref[1,0,1,0]=(0.482072704433)*x_ref[0]+(0.895847096783)*x_ref[1]
        ref[1,0,1,1]=(0.695736834617)*x_ref[0]+(0.358082704558)*x_ref[1]
        ref[1,0,2,0]=(-0.501799636565)*x_ref[0]+(0.363612558025)*x_ref[1]
        ref[1,0,2,1]=(-1.47204447226)*x_ref[0]+(1.91281426121)*x_ref[1]
        ref[1,1,0,0]=(0.241893182978)*x_ref[0]+(0.95199758669)*x_ref[1]
        ref[1,1,0,1]=(-0.68112361827)*x_ref[0]+(-0.241935141301)*x_ref[1]
        ref[1,1,1,0]=(-0.553030972306)*x_ref[0]+(0.0244503690439)*x_ref[1]
        ref[1,1,1,1]=(0.278177176896)*x_ref[0]+(1.58693846427)*x_ref[1]
        ref[1,1,2,0]=(-0.87980015517)*x_ref[0]+(0.315618837123)*x_ref[1]
        ref[1,1,2,1]=(-0.335783961379)*x_ref[0]+(0.368650136138)*x_ref[1]
        ref[1,2,0,0]=(-0.903535146802)*x_ref[0]+(1.68182923194)*x_ref[1]
        ref[1,2,0,1]=(1.06511511382)*x_ref[0]+(-1.08524762302)*x_ref[1]
        ref[1,2,1,0]=(-0.0455274784681)*x_ref[0]+(-0.519576710661)*x_ref[1]
        ref[1,2,1,1]=(0.735100217062)*x_ref[0]+(-0.986268682671)*x_ref[1]
        ref[1,2,2,0]=(-0.544610797462)*x_ref[0]+(0.447237114882)*x_ref[1]
        ref[1,2,2,1]=(0.385242031153)*x_ref[0]+(-0.960538844247)*x_ref[1]
        ref[1,3,0,0]=(-0.354927072897)*x_ref[0]+(0.842574324062)*x_ref[1]
        ref[1,3,0,1]=(0.319036427919)*x_ref[0]+(-0.156282262536)*x_ref[1]
        ref[1,3,1,0]=(0.587123807717)*x_ref[0]+(1.65277990726)*x_ref[1]
        ref[1,3,1,1]=(0.791697370298)*x_ref[0]+(-1.5681680344)*x_ref[1]
        ref[1,3,2,0]=(-1.41931242045)*x_ref[0]+(0.351285136557)*x_ref[1]
        ref[1,3,2,1]=(0.53175704402)*x_ref[0]+(0.987523241324)*x_ref[1]
        ref[1,4,0,0]=(0.00533504818465)*x_ref[0]+(-0.158886930817)*x_ref[1]
        ref[1,4,0,1]=(1.23725297303)*x_ref[0]+(-0.302896452172)*x_ref[1]
        ref[1,4,1,0]=(0.433953146329)*x_ref[0]+(-0.0721602234887)*x_ref[1]
        ref[1,4,1,1]=(-0.383499726865)*x_ref[0]+(-0.253834671235)*x_ref[1]
        ref[1,4,2,0]=(-0.480776695247)*x_ref[0]+(-0.302145524759)*x_ref[1]
        ref[1,4,2,1]=(1.14113180903)*x_ref[0]+(-1.24304726003)*x_ref[1]
        ref[2,0,0,0]=(0.513776013893)*x_ref[0]+(-0.16660259885)*x_ref[1]
        ref[2,0,0,1]=(1.55739209698)*x_ref[0]+(0.0524084654531)*x_ref[1]
        ref[2,0,1,0]=(0.222567881446)*x_ref[0]+(-0.357158470388)*x_ref[1]
        ref[2,0,1,1]=(0.0179816596229)*x_ref[0]+(0.231534460784)*x_ref[1]
        ref[2,0,2,0]=(1.58460345297)*x_ref[0]+(-0.383588433888)*x_ref[1]
        ref[2,0,2,1]=(-1.73386465536)*x_ref[0]+(-0.743326657544)*x_ref[1]
        ref[2,1,0,0]=(-0.737976950246)*x_ref[0]+(-1.01053396224)*x_ref[1]
        ref[2,1,0,1]=(0.926911190393)*x_ref[0]+(0.620251783925)*x_ref[1]
        ref[2,1,1,0]=(-1.34289630129)*x_ref[0]+(-0.429933681523)*x_ref[1]
        ref[2,1,1,1]=(1.79530806415)*x_ref[0]+(-0.539783957021)*x_ref[1]
        ref[2,1,2,0]=(0.237950974938)*x_ref[0]+(1.93783427703)*x_ref[1]
        ref[2,1,2,1]=(-0.439640741973)*x_ref[0]+(-0.61543984673)*x_ref[1]
        ref[2,2,0,0]=(-1.66403612438)*x_ref[0]+(1.00839566441)*x_ref[1]
        ref[2,2,0,1]=(0.249635983351)*x_ref[0]+(-0.722261363675)*x_ref[1]
        ref[2,2,1,0]=(1.34179674459)*x_ref[0]+(0.502542144188)*x_ref[1]
        ref[2,2,1,1]=(-0.815862586074)*x_ref[0]+(-0.110688273707)*x_ref[1]
        ref[2,2,2,0]=(-0.686810719526)*x_ref[0]+(0.500406563421)*x_ref[1]
        ref[2,2,2,1]=(0.681231089598)*x_ref[0]+(0.605219505613)*x_ref[1]
        ref[2,3,0,0]=(-0.953299029956)*x_ref[0]+(-0.223725858372)*x_ref[1]
        ref[2,3,0,1]=(0.831380756684)*x_ref[0]+(-0.0379777076335)*x_ref[1]
        ref[2,3,1,0]=(-1.63804393157)*x_ref[0]+(0.148235731301)*x_ref[1]
        ref[2,3,1,1]=(-0.17786458643)*x_ref[0]+(-1.77342152699)*x_ref[1]
        ref[2,3,2,0]=(0.0210446244192)*x_ref[0]+(1.59778781524)*x_ref[1]
        ref[2,3,2,1]=(0.877162721214)*x_ref[0]+(0.262990448614)*x_ref[1]
        ref[2,4,0,0]=(-0.650265970331)*x_ref[0]+(-1.07003604369)*x_ref[1]
        ref[2,4,0,1]=(1.12940101029)*x_ref[0]+(0.860589290891)*x_ref[1]
        ref[2,4,1,0]=(0.520056210056)*x_ref[0]+(-1.47691193979)*x_ref[1]
        ref[2,4,1,1]=(1.26823535235)*x_ref[0]+(0.496811720647)*x_ref[1]
        ref[2,4,2,0]=(-1.388913443)*x_ref[0]+(0.585343197943)*x_ref[1]
        ref[2,4,2,1]=(-0.762904547594)*x_ref[0]+(0.503438815917)*x_ref[1]
        ref[3,0,0,0]=(0.382862686993)*x_ref[0]+(-0.923525513303)*x_ref[1]
        ref[3,0,0,1]=(-0.54305794485)*x_ref[0]+(0.467513531444)*x_ref[1]
        ref[3,0,1,0]=(1.24228759776)*x_ref[0]+(-1.65740117421)*x_ref[1]
        ref[3,0,1,1]=(-0.463045362742)*x_ref[0]+(0.186438587087)*x_ref[1]
        ref[3,0,2,0]=(0.296881980828)*x_ref[0]+(-0.385878814949)*x_ref[1]
        ref[3,0,2,1]=(-0.299705899564)*x_ref[0]+(0.490736038538)*x_ref[1]
        ref[3,1,0,0]=(-0.786851428658)*x_ref[0]+(-0.760593794308)*x_ref[1]
        ref[3,1,0,1]=(0.506851336483)*x_ref[0]+(0.995163951197)*x_ref[1]
        ref[3,1,1,0]=(-0.678883937341)*x_ref[0]+(0.56996778248)*x_ref[1]
        ref[3,1,1,1]=(-0.401483642528)*x_ref[0]+(0.0476797332537)*x_ref[1]
        ref[3,1,2,0]=(-0.585357955375)*x_ref[0]+(0.330103647539)*x_ref[1]
        ref[3,1,2,1]=(-0.355539976825)*x_ref[0]+(0.416287924349)*x_ref[1]
        ref[3,2,0,0]=(-0.662837130286)*x_ref[0]+(-0.26640899674)*x_ref[1]
        ref[3,2,0,1]=(0.397553900588)*x_ref[0]+(-1.21708620294)*x_ref[1]
        ref[3,2,1,0]=(0.464963326429)*x_ref[0]+(-0.774515114142)*x_ref[1]
        ref[3,2,1,1]=(0.834351469791)*x_ref[0]+(0.487629540715)*x_ref[1]
        ref[3,2,2,0]=(0.27622693716)*x_ref[0]+(0.811338337719)*x_ref[1]
        ref[3,2,2,1]=(0.0384997772653)*x_ref[0]+(1.03724996578)*x_ref[1]
        ref[3,3,0,0]=(-0.805491987874)*x_ref[0]+(-0.476384547853)*x_ref[1]
        ref[3,3,0,1]=(0.615551780195)*x_ref[0]+(-0.312805967691)*x_ref[1]
        ref[3,3,1,0]=(0.348666798955)*x_ref[0]+(-1.31271156524)*x_ref[1]
        ref[3,3,1,1]=(-0.423569975374)*x_ref[0]+(0.930713125389)*x_ref[1]
        ref[3,3,2,0]=(0.325632686234)*x_ref[0]+(-0.166401033259)*x_ref[1]
        ref[3,3,2,1]=(0.126924861797)*x_ref[0]+(-0.425380897969)*x_ref[1]
        ref[3,4,0,0]=(0.169532204538)*x_ref[0]+(0.614632525765)*x_ref[1]
        ref[3,4,0,1]=(-0.938098696851)*x_ref[0]+(0.648854147587)*x_ref[1]
        ref[3,4,1,0]=(-0.545143970434)*x_ref[0]+(-1.59787515966)*x_ref[1]
        ref[3,4,1,1]=(-0.711190914978)*x_ref[0]+(0.897331522911)*x_ref[1]
        ref[3,4,2,0]=(-0.386498124525)*x_ref[0]+(0.299740420711)*x_ref[1]
        ref[3,4,2,1]=(0.617683032062)*x_ref[0]+(-0.449587703061)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.791620170192)*x[0]+(1.03871460301)*x[1]+(-0.225035657687)*x[2]
        arg[0,0,0,1]=(-0.464422319203)*x[0]+(0.769454931053)*x[1]+(-0.783260637771)*x[2]
        arg[0,0,1,0]=(0.72655538215)*x[0]+(-1.65736312015)*x[1]+(1.08402505927)*x[2]
        arg[0,0,1,1]=(-0.207576709573)*x[0]+(-0.55250594761)*x[1]+(-0.923702938288)*x[2]
        arg[0,0,2,0]=(-0.701991085516)*x[0]+(-1.76863643107)*x[1]+(0.831382933104)*x[2]
        arg[0,0,2,1]=(0.0640035527319)*x[0]+(0.425446342976)*x[1]+(1.64626766696)*x[2]
        arg[0,1,0,0]=(1.01022963801)*x[0]+(-0.0215015059716)*x[1]+(0.849862425592)*x[2]
        arg[0,1,0,1]=(-0.45776932943)*x[0]+(0.7621342151)*x[1]+(-1.40123018391)*x[2]
        arg[0,1,1,0]=(0.737485054127)*x[0]+(-0.147513905858)*x[1]+(-0.616928595213)*x[2]
        arg[0,1,1,1]=(-0.0608005391894)*x[0]+(-0.154978210573)*x[1]+(-1.40322011577)*x[2]
        arg[0,1,2,0]=(1.59908177714)*x[0]+(-0.378738573932)*x[1]+(1.45285315247)*x[2]
        arg[0,1,2,1]=(-0.622552157046)*x[0]+(1.19996417516)*x[1]+(0.728842900843)*x[2]
        arg[0,2,0,0]=(-0.46942034691)*x[0]+(0.333309721372)*x[1]+(-0.280846553832)*x[2]
        arg[0,2,0,1]=(1.22208788109)*x[0]+(0.369339594285)*x[1]+(1.31078425589)*x[2]
        arg[0,2,1,0]=(0.0794165004114)*x[0]+(0.44665451846)*x[1]+(-0.814625708916)*x[2]
        arg[0,2,1,1]=(0.594200807165)*x[0]+(-0.00616720333193)*x[1]+(0.382706175172)*x[2]
        arg[0,2,2,0]=(0.0372230170459)*x[0]+(0.440339923291)*x[1]+(0.137641070028)*x[2]
        arg[0,2,2,1]=(-1.29863078274)*x[0]+(0.283142455171)*x[1]+(0.0844540074659)*x[2]
        arg[0,3,0,0]=(1.72811808036)*x[0]+(0.536456781522)*x[1]+(0.0266737704706)*x[2]
        arg[0,3,0,1]=(0.148219555101)*x[0]+(-0.6887500275)*x[1]+(-0.405613136632)*x[2]
        arg[0,3,1,0]=(0.708262537279)*x[0]+(1.16319635773)*x[1]+(-0.030808721674)*x[2]
        arg[0,3,1,1]=(0.80834615863)*x[0]+(-0.176789000925)*x[1]+(-1.72576120963)*x[2]
        arg[0,3,2,0]=(0.21128315299)*x[0]+(0.770037046126)*x[1]+(0.988028934422)*x[2]
        arg[0,3,2,1]=(1.10522136317)*x[0]+(1.6319351743)*x[1]+(-0.784693713387)*x[2]
        arg[0,4,0,0]=(0.121088121064)*x[0]+(-1.79718067791)*x[1]+(-0.907857173318)*x[2]
        arg[0,4,0,1]=(-0.695887294217)*x[0]+(-1.24166787807)*x[1]+(-1.24968023856)*x[2]
        arg[0,4,1,0]=(-0.202510797291)*x[0]+(0.628310156105)*x[1]+(0.605146447822)*x[2]
        arg[0,4,1,1]=(0.149227458424)*x[0]+(-1.5650392775)*x[1]+(-0.763944892663)*x[2]
        arg[0,4,2,0]=(-0.393218625872)*x[0]+(0.479145914948)*x[1]+(-0.313960447492)*x[2]
        arg[0,4,2,1]=(-0.0191807443799)*x[0]+(0.592992659119)*x[1]+(-0.475633115484)*x[2]
        arg[1,0,0,0]=(-0.964829354261)*x[0]+(0.453014802906)*x[1]+(-0.779502566938)*x[2]
        arg[1,0,0,1]=(0.0151205739508)*x[0]+(-0.153706429594)*x[1]+(0.00866460475724)*x[2]
        arg[1,0,1,0]=(0.92592276366)*x[0]+(0.530787400525)*x[1]+(-0.0577912606876)*x[2]
        arg[1,0,1,1]=(0.948404467077)*x[0]+(-0.265905470751)*x[1]+(0.472349192005)*x[2]
        arg[1,0,2,0]=(0.0176615383702)*x[0]+(0.679605821761)*x[1]+(-0.0333590911941)*x[2]
        arg[1,0,2,1]=(-1.15594952629)*x[0]+(0.264868376508)*x[1]+(-0.883690535499)*x[2]
        arg[1,1,0,0]=(-0.738712703994)*x[0]+(0.545839232808)*x[1]+(0.592474501134)*x[2]
        arg[1,1,0,1]=(-1.3468960453)*x[0]+(-0.00848202443614)*x[1]+(-0.0124138286126)*x[2]
        arg[1,1,1,0]=(0.900228409707)*x[0]+(-1.86848082357)*x[1]+(0.421539431142)*x[2]
        arg[1,1,1,1]=(1.42754013963)*x[0]+(-0.861376440449)*x[1]+(0.0965520362291)*x[2]
        arg[1,1,2,0]=(-0.35873674515)*x[0]+(0.0862592617941)*x[1]+(-0.536106207025)*x[2]
        arg[1,1,2,1]=(0.90712910097)*x[0]+(0.974545253055)*x[1]+(0.570708510858)*x[2]
        arg[1,2,0,0]=(1.19310775669)*x[0]+(0.309414331679)*x[1]+(-0.271298061586)*x[2]
        arg[1,2,0,1]=(0.551778129545)*x[0]+(0.539028430134)*x[1]+(0.239406239262)*x[2]
        arg[1,2,1,0]=(-0.810625054344)*x[0]+(-0.241072190233)*x[1]+(-1.58590369097)*x[2]
        arg[1,2,1,1]=(0.529669075734)*x[0]+(-0.890365024163)*x[1]+(-1.69798509084)*x[2]
        arg[1,2,2,0]=(0.0437087707448)*x[0]+(0.88372190198)*x[1]+(1.86547747806)*x[2]
        arg[1,2,2,1]=(1.30739438898)*x[0]+(0.315850112101)*x[1]+(-0.262001133683)*x[2]
        arg[1,3,0,0]=(-0.143426236837)*x[0]+(-1.1703384826)*x[1]+(0.50608348646)*x[2]
        arg[1,3,0,1]=(-0.0444408776922)*x[0]+(1.26952782036)*x[1]+(1.67192602934)*x[2]
        arg[1,3,1,0]=(0.672742824368)*x[0]+(-0.785246676537)*x[1]+(0.716130664056)*x[2]
        arg[1,3,1,1]=(-0.295182508335)*x[0]+(-1.53658860897)*x[1]+(-1.87133995422)*x[2]
        arg[1,3,2,0]=(-0.440859530081)*x[0]+(0.0730120015821)*x[1]+(-1.02465537353)*x[2]
        arg[1,3,2,1]=(0.838320612003)*x[0]+(1.81233688454)*x[1]+(-1.0633876685)*x[2]
        arg[1,4,0,0]=(0.52440453051)*x[0]+(-0.0182588999022)*x[1]+(1.05556310351)*x[2]
        arg[1,4,0,1]=(0.766612727495)*x[0]+(1.58987590746)*x[1]+(-0.703801484569)*x[2]
        arg[1,4,1,0]=(0.373439757535)*x[0]+(-0.0676963704305)*x[1]+(0.0948034094293)*x[2]
        arg[1,4,1,1]=(-0.542283543375)*x[0]+(-0.454408605761)*x[1]+(-0.398904635316)*x[2]
        arg[1,4,2,0]=(1.3374871143)*x[0]+(0.0067849394587)*x[1]+(-0.319452094882)*x[2]
        arg[1,4,2,1]=(0.38530755393)*x[0]+(1.05356462429)*x[1]+(-0.302179447719)*x[2]
        arg[2,0,0,0]=(-0.123109761583)*x[0]+(0.69223848143)*x[1]+(-1.36903487092)*x[2]
        arg[2,0,0,1]=(0.121957328459)*x[0]+(0.102229685387)*x[1]+(0.962499974737)*x[2]
        arg[2,0,1,0]=(0.478614188653)*x[0]+(-1.23058817023)*x[1]+(1.30950997478)*x[2]
        arg[2,0,1,1]=(-0.529033562461)*x[0]+(0.563030509591)*x[1]+(-0.564939226654)*x[2]
        arg[2,0,2,0]=(-0.377740282406)*x[0]+(1.20532110343)*x[1]+(1.55615174947)*x[2]
        arg[2,0,2,1]=(0.539119710703)*x[0]+(1.91771865857)*x[1]+(0.0809882488048)*x[2]
        arg[2,1,0,0]=(0.00548664737796)*x[0]+(-0.157716099068)*x[1]+(-1.40009044235)*x[2]
        arg[2,1,0,1]=(0.0667557890666)*x[0]+(0.742484448131)*x[1]+(-0.325246111346)*x[2]
        arg[2,1,1,0]=(-0.214965474071)*x[0]+(-0.215055151964)*x[1]+(1.15673960537)*x[2]
        arg[2,1,1,1]=(0.605764831789)*x[0]+(-1.0432964832)*x[1]+(0.645482523161)*x[2]
        arg[2,1,2,0]=(-1.13098364088)*x[0]+(-0.932044089041)*x[1]+(-0.308053828241)*x[2]
        arg[2,1,2,1]=(0.481961305459)*x[0]+(-0.101277472029)*x[1]+(-0.17520207974)*x[2]
        arg[2,2,0,0]=(0.579831852303)*x[0]+(0.538731990254)*x[1]+(-0.0897460374205)*x[2]
        arg[2,2,0,1]=(-1.8763181434)*x[0]+(-0.132710928569)*x[1]+(1.03169751651)*x[2]
        arg[2,2,1,0]=(0.56871511431)*x[0]+(0.075333481656)*x[1]+(1.10238818953)*x[2]
        arg[2,2,1,1]=(1.49347755246)*x[0]+(-1.00500941047)*x[1]+(-0.181733196359)*x[2]
        arg[2,2,2,0]=(0.532881509971)*x[0]+(-0.364333964659)*x[1]+(1.09633720883)*x[2]
        arg[2,2,2,1]=(-0.518419048615)*x[0]+(-0.48091100593)*x[1]+(0.724628840153)*x[2]
        arg[2,3,0,0]=(0.353475859331)*x[0]+(-0.0239879269825)*x[1]+(1.32068893718)*x[2]
        arg[2,3,0,1]=(0.553999122102)*x[0]+(-0.128530869239)*x[1]+(0.134663336287)*x[2]
        arg[2,3,1,0]=(-1.59775882569)*x[0]+(-1.07335424785)*x[1]+(1.39223587988)*x[2]
        arg[2,3,1,1]=(-0.921252080095)*x[0]+(1.46226299067)*x[1]+(-1.13412499932)*x[2]
        arg[2,3,2,0]=(0.417572910178)*x[0]+(0.0814889599489)*x[1]+(0.292159249719)*x[2]
        arg[2,3,2,1]=(1.44620228921)*x[0]+(0.767854577367)*x[1]+(-0.0113709806383)*x[2]
        arg[2,4,0,0]=(-0.156987001681)*x[0]+(-0.922524257833)*x[1]+(0.521128360838)*x[2]
        arg[2,4,0,1]=(0.154198133354)*x[0]+(0.462020370175)*x[1]+(0.966768653665)*x[2]
        arg[2,4,1,0]=(0.507836849242)*x[0]+(1.16701911244)*x[1]+(0.264543038211)*x[2]
        arg[2,4,1,1]=(0.452059436879)*x[0]+(0.409662432778)*x[1]+(-0.403800872595)*x[2]
        arg[2,4,2,0]=(0.521175521008)*x[0]+(-0.362551578816)*x[1]+(0.0922727331623)*x[2]
        arg[2,4,2,1]=(-1.0990361823)*x[0]+(1.73655146851)*x[1]+(-0.867917402922)*x[2]
        arg[3,0,0,0]=(-0.0527010882236)*x[0]+(-1.11528215488)*x[1]+(1.3141906393)*x[2]
        arg[3,0,0,1]=(0.151284271917)*x[0]+(-0.0159077450668)*x[1]+(1.1798338931)*x[2]
        arg[3,0,1,0]=(-1.7648963435)*x[0]+(0.229511437402)*x[1]+(0.924415900312)*x[2]
        arg[3,0,1,1]=(-0.498271821094)*x[0]+(1.58363063019)*x[1]+(1.64871750388)*x[2]
        arg[3,0,2,0]=(-1.11245170725)*x[0]+(1.27662612569)*x[1]+(0.145950372369)*x[2]
        arg[3,0,2,1]=(-0.691643924238)*x[0]+(-0.178211874279)*x[1]+(-0.341071050366)*x[2]
        arg[3,1,0,0]=(-0.929804889461)*x[0]+(0.199087067309)*x[1]+(1.00150168883)*x[2]
        arg[3,1,0,1]=(-1.25152265093)*x[0]+(0.517076294911)*x[1]+(0.103798312362)*x[2]
        arg[3,1,1,0]=(0.797316290623)*x[0]+(0.0239285379293)*x[1]+(-0.836519299418)*x[2]
        arg[3,1,1,1]=(-0.470084592176)*x[0]+(0.844364701728)*x[1]+(-0.027373275093)*x[2]
        arg[3,1,2,0]=(1.0340279373)*x[0]+(-0.0133951276315)*x[1]+(1.17710546194)*x[2]
        arg[3,1,2,1]=(0.0325929541496)*x[0]+(1.13784867184)*x[1]+(-0.0529194266769)*x[2]
        arg[3,2,0,0]=(1.92837561772)*x[0]+(-1.63633468831)*x[1]+(1.4212324281)*x[2]
        arg[3,2,0,1]=(1.09556054408)*x[0]+(1.43286059545)*x[1]+(1.20005604593)*x[2]
        arg[3,2,1,0]=(0.308887169169)*x[0]+(-0.149214224208)*x[1]+(-1.13559614877)*x[2]
        arg[3,2,1,1]=(-0.0648352909004)*x[0]+(-0.5645470596)*x[1]+(-0.525131730514)*x[2]
        arg[3,2,2,0]=(0.21826829974)*x[0]+(1.5950239996)*x[1]+(0.333592425438)*x[2]
        arg[3,2,2,1]=(-1.25613956172)*x[0]+(-0.242338679844)*x[1]+(-0.958875577074)*x[2]
        arg[3,3,0,0]=(1.65005362159)*x[0]+(-1.16713867165)*x[1]+(-0.282495293685)*x[2]
        arg[3,3,0,1]=(1.04854406427)*x[0]+(0.876270949531)*x[1]+(0.291036304743)*x[2]
        arg[3,3,1,0]=(-1.66619451175)*x[0]+(0.656275192484)*x[1]+(-1.27104499942)*x[2]
        arg[3,3,1,1]=(1.55407146907)*x[0]+(-0.322950943159)*x[1]+(-0.287565403896)*x[2]
        arg[3,3,2,0]=(-1.54760268583)*x[0]+(-1.45847761887)*x[1]+(1.13939567009)*x[2]
        arg[3,3,2,1]=(1.24632351699)*x[0]+(0.0265500081594)*x[1]+(1.61925095159)*x[2]
        arg[3,4,0,0]=(-0.0430889433549)*x[0]+(-1.46995559463)*x[1]+(-1.14915159823)*x[2]
        arg[3,4,0,1]=(0.0392641403409)*x[0]+(0.0603053049639)*x[1]+(0.647321169934)*x[2]
        arg[3,4,1,0]=(-0.156300985227)*x[0]+(0.79369911928)*x[1]+(-0.413124762507)*x[2]
        arg[3,4,1,1]=(0.258402059015)*x[0]+(0.00227307988585)*x[1]+(0.70067141295)*x[2]
        arg[3,4,2,0]=(-1.15131298872)*x[0]+(-0.627449043157)*x[1]+(-0.402268345861)*x[2]
        arg[3,4,2,1]=(0.455934238179)*x[0]+(-0.951120759108)*x[1]+(-0.280459367131)*x[2]
        ref[0,0,0,0]=(-0.791620170192)*x_ref[0]+(1.03871460301)*x_ref[1]+(-0.225035657687)*x_ref[2]
        ref[0,0,0,1]=(-0.464422319203)*x_ref[0]+(0.769454931053)*x_ref[1]+(-0.783260637771)*x_ref[2]
        ref[0,0,1,0]=(0.72655538215)*x_ref[0]+(-1.65736312015)*x_ref[1]+(1.08402505927)*x_ref[2]
        ref[0,0,1,1]=(-0.207576709573)*x_ref[0]+(-0.55250594761)*x_ref[1]+(-0.923702938288)*x_ref[2]
        ref[0,0,2,0]=(-0.701991085516)*x_ref[0]+(-1.76863643107)*x_ref[1]+(0.831382933104)*x_ref[2]
        ref[0,0,2,1]=(0.0640035527319)*x_ref[0]+(0.425446342976)*x_ref[1]+(1.64626766696)*x_ref[2]
        ref[0,1,0,0]=(1.01022963801)*x_ref[0]+(-0.0215015059716)*x_ref[1]+(0.849862425592)*x_ref[2]
        ref[0,1,0,1]=(-0.45776932943)*x_ref[0]+(0.7621342151)*x_ref[1]+(-1.40123018391)*x_ref[2]
        ref[0,1,1,0]=(0.737485054127)*x_ref[0]+(-0.147513905858)*x_ref[1]+(-0.616928595213)*x_ref[2]
        ref[0,1,1,1]=(-0.0608005391894)*x_ref[0]+(-0.154978210573)*x_ref[1]+(-1.40322011577)*x_ref[2]
        ref[0,1,2,0]=(1.59908177714)*x_ref[0]+(-0.378738573932)*x_ref[1]+(1.45285315247)*x_ref[2]
        ref[0,1,2,1]=(-0.622552157046)*x_ref[0]+(1.19996417516)*x_ref[1]+(0.728842900843)*x_ref[2]
        ref[0,2,0,0]=(-0.46942034691)*x_ref[0]+(0.333309721372)*x_ref[1]+(-0.280846553832)*x_ref[2]
        ref[0,2,0,1]=(1.22208788109)*x_ref[0]+(0.369339594285)*x_ref[1]+(1.31078425589)*x_ref[2]
        ref[0,2,1,0]=(0.0794165004114)*x_ref[0]+(0.44665451846)*x_ref[1]+(-0.814625708916)*x_ref[2]
        ref[0,2,1,1]=(0.594200807165)*x_ref[0]+(-0.00616720333193)*x_ref[1]+(0.382706175172)*x_ref[2]
        ref[0,2,2,0]=(0.0372230170459)*x_ref[0]+(0.440339923291)*x_ref[1]+(0.137641070028)*x_ref[2]
        ref[0,2,2,1]=(-1.29863078274)*x_ref[0]+(0.283142455171)*x_ref[1]+(0.0844540074659)*x_ref[2]
        ref[0,3,0,0]=(1.72811808036)*x_ref[0]+(0.536456781522)*x_ref[1]+(0.0266737704706)*x_ref[2]
        ref[0,3,0,1]=(0.148219555101)*x_ref[0]+(-0.6887500275)*x_ref[1]+(-0.405613136632)*x_ref[2]
        ref[0,3,1,0]=(0.708262537279)*x_ref[0]+(1.16319635773)*x_ref[1]+(-0.030808721674)*x_ref[2]
        ref[0,3,1,1]=(0.80834615863)*x_ref[0]+(-0.176789000925)*x_ref[1]+(-1.72576120963)*x_ref[2]
        ref[0,3,2,0]=(0.21128315299)*x_ref[0]+(0.770037046126)*x_ref[1]+(0.988028934422)*x_ref[2]
        ref[0,3,2,1]=(1.10522136317)*x_ref[0]+(1.6319351743)*x_ref[1]+(-0.784693713387)*x_ref[2]
        ref[0,4,0,0]=(0.121088121064)*x_ref[0]+(-1.79718067791)*x_ref[1]+(-0.907857173318)*x_ref[2]
        ref[0,4,0,1]=(-0.695887294217)*x_ref[0]+(-1.24166787807)*x_ref[1]+(-1.24968023856)*x_ref[2]
        ref[0,4,1,0]=(-0.202510797291)*x_ref[0]+(0.628310156105)*x_ref[1]+(0.605146447822)*x_ref[2]
        ref[0,4,1,1]=(0.149227458424)*x_ref[0]+(-1.5650392775)*x_ref[1]+(-0.763944892663)*x_ref[2]
        ref[0,4,2,0]=(-0.393218625872)*x_ref[0]+(0.479145914948)*x_ref[1]+(-0.313960447492)*x_ref[2]
        ref[0,4,2,1]=(-0.0191807443799)*x_ref[0]+(0.592992659119)*x_ref[1]+(-0.475633115484)*x_ref[2]
        ref[1,0,0,0]=(-0.964829354261)*x_ref[0]+(0.453014802906)*x_ref[1]+(-0.779502566938)*x_ref[2]
        ref[1,0,0,1]=(0.0151205739508)*x_ref[0]+(-0.153706429594)*x_ref[1]+(0.00866460475724)*x_ref[2]
        ref[1,0,1,0]=(0.92592276366)*x_ref[0]+(0.530787400525)*x_ref[1]+(-0.0577912606876)*x_ref[2]
        ref[1,0,1,1]=(0.948404467077)*x_ref[0]+(-0.265905470751)*x_ref[1]+(0.472349192005)*x_ref[2]
        ref[1,0,2,0]=(0.0176615383702)*x_ref[0]+(0.679605821761)*x_ref[1]+(-0.0333590911941)*x_ref[2]
        ref[1,0,2,1]=(-1.15594952629)*x_ref[0]+(0.264868376508)*x_ref[1]+(-0.883690535499)*x_ref[2]
        ref[1,1,0,0]=(-0.738712703994)*x_ref[0]+(0.545839232808)*x_ref[1]+(0.592474501134)*x_ref[2]
        ref[1,1,0,1]=(-1.3468960453)*x_ref[0]+(-0.00848202443614)*x_ref[1]+(-0.0124138286126)*x_ref[2]
        ref[1,1,1,0]=(0.900228409707)*x_ref[0]+(-1.86848082357)*x_ref[1]+(0.421539431142)*x_ref[2]
        ref[1,1,1,1]=(1.42754013963)*x_ref[0]+(-0.861376440449)*x_ref[1]+(0.0965520362291)*x_ref[2]
        ref[1,1,2,0]=(-0.35873674515)*x_ref[0]+(0.0862592617941)*x_ref[1]+(-0.536106207025)*x_ref[2]
        ref[1,1,2,1]=(0.90712910097)*x_ref[0]+(0.974545253055)*x_ref[1]+(0.570708510858)*x_ref[2]
        ref[1,2,0,0]=(1.19310775669)*x_ref[0]+(0.309414331679)*x_ref[1]+(-0.271298061586)*x_ref[2]
        ref[1,2,0,1]=(0.551778129545)*x_ref[0]+(0.539028430134)*x_ref[1]+(0.239406239262)*x_ref[2]
        ref[1,2,1,0]=(-0.810625054344)*x_ref[0]+(-0.241072190233)*x_ref[1]+(-1.58590369097)*x_ref[2]
        ref[1,2,1,1]=(0.529669075734)*x_ref[0]+(-0.890365024163)*x_ref[1]+(-1.69798509084)*x_ref[2]
        ref[1,2,2,0]=(0.0437087707448)*x_ref[0]+(0.88372190198)*x_ref[1]+(1.86547747806)*x_ref[2]
        ref[1,2,2,1]=(1.30739438898)*x_ref[0]+(0.315850112101)*x_ref[1]+(-0.262001133683)*x_ref[2]
        ref[1,3,0,0]=(-0.143426236837)*x_ref[0]+(-1.1703384826)*x_ref[1]+(0.50608348646)*x_ref[2]
        ref[1,3,0,1]=(-0.0444408776922)*x_ref[0]+(1.26952782036)*x_ref[1]+(1.67192602934)*x_ref[2]
        ref[1,3,1,0]=(0.672742824368)*x_ref[0]+(-0.785246676537)*x_ref[1]+(0.716130664056)*x_ref[2]
        ref[1,3,1,1]=(-0.295182508335)*x_ref[0]+(-1.53658860897)*x_ref[1]+(-1.87133995422)*x_ref[2]
        ref[1,3,2,0]=(-0.440859530081)*x_ref[0]+(0.0730120015821)*x_ref[1]+(-1.02465537353)*x_ref[2]
        ref[1,3,2,1]=(0.838320612003)*x_ref[0]+(1.81233688454)*x_ref[1]+(-1.0633876685)*x_ref[2]
        ref[1,4,0,0]=(0.52440453051)*x_ref[0]+(-0.0182588999022)*x_ref[1]+(1.05556310351)*x_ref[2]
        ref[1,4,0,1]=(0.766612727495)*x_ref[0]+(1.58987590746)*x_ref[1]+(-0.703801484569)*x_ref[2]
        ref[1,4,1,0]=(0.373439757535)*x_ref[0]+(-0.0676963704305)*x_ref[1]+(0.0948034094293)*x_ref[2]
        ref[1,4,1,1]=(-0.542283543375)*x_ref[0]+(-0.454408605761)*x_ref[1]+(-0.398904635316)*x_ref[2]
        ref[1,4,2,0]=(1.3374871143)*x_ref[0]+(0.0067849394587)*x_ref[1]+(-0.319452094882)*x_ref[2]
        ref[1,4,2,1]=(0.38530755393)*x_ref[0]+(1.05356462429)*x_ref[1]+(-0.302179447719)*x_ref[2]
        ref[2,0,0,0]=(-0.123109761583)*x_ref[0]+(0.69223848143)*x_ref[1]+(-1.36903487092)*x_ref[2]
        ref[2,0,0,1]=(0.121957328459)*x_ref[0]+(0.102229685387)*x_ref[1]+(0.962499974737)*x_ref[2]
        ref[2,0,1,0]=(0.478614188653)*x_ref[0]+(-1.23058817023)*x_ref[1]+(1.30950997478)*x_ref[2]
        ref[2,0,1,1]=(-0.529033562461)*x_ref[0]+(0.563030509591)*x_ref[1]+(-0.564939226654)*x_ref[2]
        ref[2,0,2,0]=(-0.377740282406)*x_ref[0]+(1.20532110343)*x_ref[1]+(1.55615174947)*x_ref[2]
        ref[2,0,2,1]=(0.539119710703)*x_ref[0]+(1.91771865857)*x_ref[1]+(0.0809882488048)*x_ref[2]
        ref[2,1,0,0]=(0.00548664737796)*x_ref[0]+(-0.157716099068)*x_ref[1]+(-1.40009044235)*x_ref[2]
        ref[2,1,0,1]=(0.0667557890666)*x_ref[0]+(0.742484448131)*x_ref[1]+(-0.325246111346)*x_ref[2]
        ref[2,1,1,0]=(-0.214965474071)*x_ref[0]+(-0.215055151964)*x_ref[1]+(1.15673960537)*x_ref[2]
        ref[2,1,1,1]=(0.605764831789)*x_ref[0]+(-1.0432964832)*x_ref[1]+(0.645482523161)*x_ref[2]
        ref[2,1,2,0]=(-1.13098364088)*x_ref[0]+(-0.932044089041)*x_ref[1]+(-0.308053828241)*x_ref[2]
        ref[2,1,2,1]=(0.481961305459)*x_ref[0]+(-0.101277472029)*x_ref[1]+(-0.17520207974)*x_ref[2]
        ref[2,2,0,0]=(0.579831852303)*x_ref[0]+(0.538731990254)*x_ref[1]+(-0.0897460374205)*x_ref[2]
        ref[2,2,0,1]=(-1.8763181434)*x_ref[0]+(-0.132710928569)*x_ref[1]+(1.03169751651)*x_ref[2]
        ref[2,2,1,0]=(0.56871511431)*x_ref[0]+(0.075333481656)*x_ref[1]+(1.10238818953)*x_ref[2]
        ref[2,2,1,1]=(1.49347755246)*x_ref[0]+(-1.00500941047)*x_ref[1]+(-0.181733196359)*x_ref[2]
        ref[2,2,2,0]=(0.532881509971)*x_ref[0]+(-0.364333964659)*x_ref[1]+(1.09633720883)*x_ref[2]
        ref[2,2,2,1]=(-0.518419048615)*x_ref[0]+(-0.48091100593)*x_ref[1]+(0.724628840153)*x_ref[2]
        ref[2,3,0,0]=(0.353475859331)*x_ref[0]+(-0.0239879269825)*x_ref[1]+(1.32068893718)*x_ref[2]
        ref[2,3,0,1]=(0.553999122102)*x_ref[0]+(-0.128530869239)*x_ref[1]+(0.134663336287)*x_ref[2]
        ref[2,3,1,0]=(-1.59775882569)*x_ref[0]+(-1.07335424785)*x_ref[1]+(1.39223587988)*x_ref[2]
        ref[2,3,1,1]=(-0.921252080095)*x_ref[0]+(1.46226299067)*x_ref[1]+(-1.13412499932)*x_ref[2]
        ref[2,3,2,0]=(0.417572910178)*x_ref[0]+(0.0814889599489)*x_ref[1]+(0.292159249719)*x_ref[2]
        ref[2,3,2,1]=(1.44620228921)*x_ref[0]+(0.767854577367)*x_ref[1]+(-0.0113709806383)*x_ref[2]
        ref[2,4,0,0]=(-0.156987001681)*x_ref[0]+(-0.922524257833)*x_ref[1]+(0.521128360838)*x_ref[2]
        ref[2,4,0,1]=(0.154198133354)*x_ref[0]+(0.462020370175)*x_ref[1]+(0.966768653665)*x_ref[2]
        ref[2,4,1,0]=(0.507836849242)*x_ref[0]+(1.16701911244)*x_ref[1]+(0.264543038211)*x_ref[2]
        ref[2,4,1,1]=(0.452059436879)*x_ref[0]+(0.409662432778)*x_ref[1]+(-0.403800872595)*x_ref[2]
        ref[2,4,2,0]=(0.521175521008)*x_ref[0]+(-0.362551578816)*x_ref[1]+(0.0922727331623)*x_ref[2]
        ref[2,4,2,1]=(-1.0990361823)*x_ref[0]+(1.73655146851)*x_ref[1]+(-0.867917402922)*x_ref[2]
        ref[3,0,0,0]=(-0.0527010882236)*x_ref[0]+(-1.11528215488)*x_ref[1]+(1.3141906393)*x_ref[2]
        ref[3,0,0,1]=(0.151284271917)*x_ref[0]+(-0.0159077450668)*x_ref[1]+(1.1798338931)*x_ref[2]
        ref[3,0,1,0]=(-1.7648963435)*x_ref[0]+(0.229511437402)*x_ref[1]+(0.924415900312)*x_ref[2]
        ref[3,0,1,1]=(-0.498271821094)*x_ref[0]+(1.58363063019)*x_ref[1]+(1.64871750388)*x_ref[2]
        ref[3,0,2,0]=(-1.11245170725)*x_ref[0]+(1.27662612569)*x_ref[1]+(0.145950372369)*x_ref[2]
        ref[3,0,2,1]=(-0.691643924238)*x_ref[0]+(-0.178211874279)*x_ref[1]+(-0.341071050366)*x_ref[2]
        ref[3,1,0,0]=(-0.929804889461)*x_ref[0]+(0.199087067309)*x_ref[1]+(1.00150168883)*x_ref[2]
        ref[3,1,0,1]=(-1.25152265093)*x_ref[0]+(0.517076294911)*x_ref[1]+(0.103798312362)*x_ref[2]
        ref[3,1,1,0]=(0.797316290623)*x_ref[0]+(0.0239285379293)*x_ref[1]+(-0.836519299418)*x_ref[2]
        ref[3,1,1,1]=(-0.470084592176)*x_ref[0]+(0.844364701728)*x_ref[1]+(-0.027373275093)*x_ref[2]
        ref[3,1,2,0]=(1.0340279373)*x_ref[0]+(-0.0133951276315)*x_ref[1]+(1.17710546194)*x_ref[2]
        ref[3,1,2,1]=(0.0325929541496)*x_ref[0]+(1.13784867184)*x_ref[1]+(-0.0529194266769)*x_ref[2]
        ref[3,2,0,0]=(1.92837561772)*x_ref[0]+(-1.63633468831)*x_ref[1]+(1.4212324281)*x_ref[2]
        ref[3,2,0,1]=(1.09556054408)*x_ref[0]+(1.43286059545)*x_ref[1]+(1.20005604593)*x_ref[2]
        ref[3,2,1,0]=(0.308887169169)*x_ref[0]+(-0.149214224208)*x_ref[1]+(-1.13559614877)*x_ref[2]
        ref[3,2,1,1]=(-0.0648352909004)*x_ref[0]+(-0.5645470596)*x_ref[1]+(-0.525131730514)*x_ref[2]
        ref[3,2,2,0]=(0.21826829974)*x_ref[0]+(1.5950239996)*x_ref[1]+(0.333592425438)*x_ref[2]
        ref[3,2,2,1]=(-1.25613956172)*x_ref[0]+(-0.242338679844)*x_ref[1]+(-0.958875577074)*x_ref[2]
        ref[3,3,0,0]=(1.65005362159)*x_ref[0]+(-1.16713867165)*x_ref[1]+(-0.282495293685)*x_ref[2]
        ref[3,3,0,1]=(1.04854406427)*x_ref[0]+(0.876270949531)*x_ref[1]+(0.291036304743)*x_ref[2]
        ref[3,3,1,0]=(-1.66619451175)*x_ref[0]+(0.656275192484)*x_ref[1]+(-1.27104499942)*x_ref[2]
        ref[3,3,1,1]=(1.55407146907)*x_ref[0]+(-0.322950943159)*x_ref[1]+(-0.287565403896)*x_ref[2]
        ref[3,3,2,0]=(-1.54760268583)*x_ref[0]+(-1.45847761887)*x_ref[1]+(1.13939567009)*x_ref[2]
        ref[3,3,2,1]=(1.24632351699)*x_ref[0]+(0.0265500081594)*x_ref[1]+(1.61925095159)*x_ref[2]
        ref[3,4,0,0]=(-0.0430889433549)*x_ref[0]+(-1.46995559463)*x_ref[1]+(-1.14915159823)*x_ref[2]
        ref[3,4,0,1]=(0.0392641403409)*x_ref[0]+(0.0603053049639)*x_ref[1]+(0.647321169934)*x_ref[2]
        ref[3,4,1,0]=(-0.156300985227)*x_ref[0]+(0.79369911928)*x_ref[1]+(-0.413124762507)*x_ref[2]
        ref[3,4,1,1]=(0.258402059015)*x_ref[0]+(0.00227307988585)*x_ref[1]+(0.70067141295)*x_ref[2]
        ref[3,4,2,0]=(-1.15131298872)*x_ref[0]+(-0.627449043157)*x_ref[1]+(-0.402268345861)*x_ref[2]
        ref[3,4,2,1]=(0.455934238179)*x_ref[0]+(-0.951120759108)*x_ref[1]+(-0.280459367131)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.203238386406)*x[0]**o+(0.734542238717)*x[0]+(-0.627248060056)*x[1]**o+(0.347224669854)*x[1]
        ref=(0.203238386406)*x_ref[0]**o+(0.734542238717)*x_ref[0]+(-0.627248060056)*x_ref[1]**o+(0.347224669854)*x_ref[1]
      else:
        arg=(-0.387501489724)*x[0]**o+(0.752554332333)*x[0]+(-0.0831588271355)*x[1]**o+(0.615191870812)*x[1]+(-0.264702923622)*x[2]**o+(0.404594629993)*x[2]
        ref=(-0.387501489724)*x_ref[0]**o+(0.752554332333)*x_ref[0]+(-0.0831588271355)*x_ref[1]**o+(0.615191870812)*x_ref[1]+(-0.264702923622)*x_ref[2]**o+(0.404594629993)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.938444792136)*x[0]**o+(-0.809407040724)*x[0]+(-0.963148615762)*x[1]**o+(-0.257000244129)*x[1]
        arg[1]=(-0.791280640039)*x[0]**o+(0.293858623095)*x[0]+(0.964942589784)*x[1]**o+(0.522071353081)*x[1]
        ref[0]=(0.938444792136)*x_ref[0]**o+(-0.809407040724)*x_ref[0]+(-0.963148615762)*x_ref[1]**o+(-0.257000244129)*x_ref[1]
        ref[1]=(-0.791280640039)*x_ref[0]**o+(0.293858623095)*x_ref[0]+(0.964942589784)*x_ref[1]**o+(0.522071353081)*x_ref[1]
      else:
        arg[0]=(0.566110580666)*x[0]**o+(-0.89723913213)*x[0]+(-0.54781629706)*x[1]**o+(-0.0405507445628)*x[1]+(0.435477036501)*x[2]**o+(-0.296404435183)*x[2]
        arg[1]=(-0.436126359475)*x[0]**o+(0.863706244241)*x[0]+(-0.149818409243)*x[1]**o+(-0.654801920098)*x[1]+(-0.281005092809)*x[2]**o+(-0.0257179420057)*x[2]
        ref[0]=(0.566110580666)*x_ref[0]**o+(-0.89723913213)*x_ref[0]+(-0.54781629706)*x_ref[1]**o+(-0.0405507445628)*x_ref[1]+(0.435477036501)*x_ref[2]**o+(-0.296404435183)*x_ref[2]
        ref[1]=(-0.436126359475)*x_ref[0]**o+(0.863706244241)*x_ref[0]+(-0.149818409243)*x_ref[1]**o+(-0.654801920098)*x_ref[1]+(-0.281005092809)*x_ref[2]**o+(-0.0257179420057)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0615713819144)*x[0]**o+(0.325133297443)*x[0]+(-0.529769594257)*x[1]**o+(-0.206126939263)*x[1]
        arg[0,1]=(0.090956391947)*x[0]**o+(0.105841035331)*x[0]+(-0.537033610998)*x[1]**o+(0.429033781495)*x[1]
        arg[0,2]=(-0.568216263139)*x[0]**o+(0.698057796806)*x[0]+(-0.0981614063428)*x[1]**o+(-0.952212332523)*x[1]
        arg[0,3]=(0.128505813477)*x[0]**o+(0.206828305852)*x[0]+(0.938851873532)*x[1]**o+(-0.339234589281)*x[1]
        arg[0,4]=(0.581805477855)*x[0]**o+(-0.621122347153)*x[0]+(-0.601029625565)*x[1]**o+(0.243084366725)*x[1]
        arg[1,0]=(-0.342411991491)*x[0]**o+(0.642434320731)*x[0]+(-0.452657094515)*x[1]**o+(-0.899779109444)*x[1]
        arg[1,1]=(-0.423230603077)*x[0]**o+(0.739600521332)*x[0]+(0.916226480419)*x[1]**o+(0.755264173162)*x[1]
        arg[1,2]=(0.381864221314)*x[0]**o+(0.806402187902)*x[0]+(0.410055116458)*x[1]**o+(-0.264405213472)*x[1]
        arg[1,3]=(0.293401427429)*x[0]**o+(-0.439209675154)*x[0]+(0.299588226145)*x[1]**o+(-0.289652201976)*x[1]
        arg[1,4]=(0.157626502865)*x[0]**o+(0.32264788209)*x[0]+(0.284522039757)*x[1]**o+(-0.460413671191)*x[1]
        arg[2,0]=(0.486148207)*x[0]**o+(0.0366950072658)*x[0]+(0.315299830901)*x[1]**o+(-0.320405042262)*x[1]
        arg[2,1]=(0.207440897542)*x[0]**o+(-0.604659890692)*x[0]+(-0.64841431916)*x[1]**o+(0.0881921209641)*x[1]
        arg[2,2]=(-0.61376840347)*x[0]**o+(0.421837679185)*x[0]+(-0.0445850531916)*x[1]**o+(0.771892018903)*x[1]
        arg[2,3]=(0.632443182263)*x[0]**o+(-0.965297013012)*x[0]+(-0.029677959681)*x[1]**o+(0.598431910551)*x[1]
        arg[2,4]=(-0.457656411094)*x[0]**o+(0.122435925199)*x[0]+(0.772690251937)*x[1]**o+(0.954829922926)*x[1]
        arg[3,0]=(0.929112925461)*x[0]**o+(-0.35689901655)*x[0]+(-0.139192866782)*x[1]**o+(-0.677883436952)*x[1]
        arg[3,1]=(-0.410318721103)*x[0]**o+(0.370830312699)*x[0]+(0.429011619434)*x[1]**o+(0.81780420854)*x[1]
        arg[3,2]=(0.312456682985)*x[0]**o+(0.526447664999)*x[0]+(0.763610570932)*x[1]**o+(-0.631887901017)*x[1]
        arg[3,3]=(0.749121567955)*x[0]**o+(0.168525340464)*x[0]+(-0.838409619895)*x[1]**o+(0.912745137098)*x[1]
        arg[3,4]=(-0.665336604643)*x[0]**o+(0.638293226479)*x[0]+(0.515957084712)*x[1]**o+(-0.261426720977)*x[1]
        ref[0,0]=(0.0615713819144)*x_ref[0]**o+(0.325133297443)*x_ref[0]+(-0.529769594257)*x_ref[1]**o+(-0.206126939263)*x_ref[1]
        ref[0,1]=(0.090956391947)*x_ref[0]**o+(0.105841035331)*x_ref[0]+(-0.537033610998)*x_ref[1]**o+(0.429033781495)*x_ref[1]
        ref[0,2]=(-0.568216263139)*x_ref[0]**o+(0.698057796806)*x_ref[0]+(-0.0981614063428)*x_ref[1]**o+(-0.952212332523)*x_ref[1]
        ref[0,3]=(0.128505813477)*x_ref[0]**o+(0.206828305852)*x_ref[0]+(0.938851873532)*x_ref[1]**o+(-0.339234589281)*x_ref[1]
        ref[0,4]=(0.581805477855)*x_ref[0]**o+(-0.621122347153)*x_ref[0]+(-0.601029625565)*x_ref[1]**o+(0.243084366725)*x_ref[1]
        ref[1,0]=(-0.342411991491)*x_ref[0]**o+(0.642434320731)*x_ref[0]+(-0.452657094515)*x_ref[1]**o+(-0.899779109444)*x_ref[1]
        ref[1,1]=(-0.423230603077)*x_ref[0]**o+(0.739600521332)*x_ref[0]+(0.916226480419)*x_ref[1]**o+(0.755264173162)*x_ref[1]
        ref[1,2]=(0.381864221314)*x_ref[0]**o+(0.806402187902)*x_ref[0]+(0.410055116458)*x_ref[1]**o+(-0.264405213472)*x_ref[1]
        ref[1,3]=(0.293401427429)*x_ref[0]**o+(-0.439209675154)*x_ref[0]+(0.299588226145)*x_ref[1]**o+(-0.289652201976)*x_ref[1]
        ref[1,4]=(0.157626502865)*x_ref[0]**o+(0.32264788209)*x_ref[0]+(0.284522039757)*x_ref[1]**o+(-0.460413671191)*x_ref[1]
        ref[2,0]=(0.486148207)*x_ref[0]**o+(0.0366950072658)*x_ref[0]+(0.315299830901)*x_ref[1]**o+(-0.320405042262)*x_ref[1]
        ref[2,1]=(0.207440897542)*x_ref[0]**o+(-0.604659890692)*x_ref[0]+(-0.64841431916)*x_ref[1]**o+(0.0881921209641)*x_ref[1]
        ref[2,2]=(-0.61376840347)*x_ref[0]**o+(0.421837679185)*x_ref[0]+(-0.0445850531916)*x_ref[1]**o+(0.771892018903)*x_ref[1]
        ref[2,3]=(0.632443182263)*x_ref[0]**o+(-0.965297013012)*x_ref[0]+(-0.029677959681)*x_ref[1]**o+(0.598431910551)*x_ref[1]
        ref[2,4]=(-0.457656411094)*x_ref[0]**o+(0.122435925199)*x_ref[0]+(0.772690251937)*x_ref[1]**o+(0.954829922926)*x_ref[1]
        ref[3,0]=(0.929112925461)*x_ref[0]**o+(-0.35689901655)*x_ref[0]+(-0.139192866782)*x_ref[1]**o+(-0.677883436952)*x_ref[1]
        ref[3,1]=(-0.410318721103)*x_ref[0]**o+(0.370830312699)*x_ref[0]+(0.429011619434)*x_ref[1]**o+(0.81780420854)*x_ref[1]
        ref[3,2]=(0.312456682985)*x_ref[0]**o+(0.526447664999)*x_ref[0]+(0.763610570932)*x_ref[1]**o+(-0.631887901017)*x_ref[1]
        ref[3,3]=(0.749121567955)*x_ref[0]**o+(0.168525340464)*x_ref[0]+(-0.838409619895)*x_ref[1]**o+(0.912745137098)*x_ref[1]
        ref[3,4]=(-0.665336604643)*x_ref[0]**o+(0.638293226479)*x_ref[0]+(0.515957084712)*x_ref[1]**o+(-0.261426720977)*x_ref[1]
      else:
        arg[0,0]=(-0.330165086386)*x[0]**o+(0.60381119065)*x[0]+(0.981018763011)*x[1]**o+(0.0487965999608)*x[1]+(-0.389835388349)*x[2]**o+(0.111199857376)*x[2]
        arg[0,1]=(0.518239033178)*x[0]**o+(0.60358784684)*x[0]+(-0.476990159655)*x[1]**o+(-0.93871565843)*x[1]+(-0.869845295741)*x[2]**o+(0.391409745802)*x[2]
        arg[0,2]=(0.0927396115951)*x[0]**o+(0.546122822928)*x[0]+(-0.557424056733)*x[1]**o+(0.0623374966436)*x[1]+(-0.414595153536)*x[2]**o+(-0.196069073366)*x[2]
        arg[0,3]=(-0.678432580882)*x[0]**o+(0.281850265141)*x[0]+(-0.323978960602)*x[1]**o+(0.655558769921)*x[1]+(-0.295747032089)*x[2]**o+(-0.152719838982)*x[2]
        arg[0,4]=(-0.904177669203)*x[0]**o+(0.750395234205)*x[0]+(0.267431783461)*x[1]**o+(-0.781406235306)*x[1]+(0.856906655757)*x[2]**o+(-0.875554940224)*x[2]
        arg[1,0]=(-0.243144846569)*x[0]**o+(0.970584922467)*x[0]+(-0.977250652473)*x[1]**o+(0.0239563255869)*x[1]+(-0.769423602851)*x[2]**o+(-0.433896361294)*x[2]
        arg[1,1]=(0.31104031046)*x[0]**o+(0.899859709831)*x[0]+(0.692351561919)*x[1]**o+(-0.0838870550002)*x[1]+(-0.276105616798)*x[2]**o+(0.115501348248)*x[2]
        arg[1,2]=(0.56702114595)*x[0]**o+(0.00240284646131)*x[0]+(-0.676314888884)*x[1]**o+(-0.678204346328)*x[1]+(0.472697088984)*x[2]**o+(-0.493949191712)*x[2]
        arg[1,3]=(-0.242548554067)*x[0]**o+(0.980927507053)*x[0]+(-0.955706390436)*x[1]**o+(0.347031550126)*x[1]+(0.234352268238)*x[2]**o+(0.321126345333)*x[2]
        arg[1,4]=(-0.390890566228)*x[0]**o+(-0.958497753144)*x[0]+(0.316397716043)*x[1]**o+(0.952289821769)*x[1]+(-0.027182029514)*x[2]**o+(-0.454422377872)*x[2]
        arg[2,0]=(0.295865166318)*x[0]**o+(-0.301129718468)*x[0]+(-0.748135755269)*x[1]**o+(-0.762413814085)*x[1]+(0.572270067591)*x[2]**o+(-0.393847032684)*x[2]
        arg[2,1]=(0.20229063622)*x[0]**o+(-0.768444028163)*x[0]+(0.378669775381)*x[1]**o+(-0.703384410944)*x[1]+(0.355653859396)*x[2]**o+(-0.228168949639)*x[2]
        arg[2,2]=(-0.31725134742)*x[0]**o+(0.525802867508)*x[0]+(0.13375126758)*x[1]**o+(-0.673551491186)*x[1]+(0.498905883728)*x[2]**o+(-0.102794461221)*x[2]
        arg[2,3]=(-0.396813838168)*x[0]**o+(0.0892605807843)*x[0]+(0.244133298111)*x[1]**o+(0.310900636168)*x[1]+(-0.256807063422)*x[2]**o+(0.565370697688)*x[2]
        arg[2,4]=(0.42011835881)*x[0]**o+(0.439847594375)*x[0]+(0.0731897272006)*x[1]**o+(-0.373911054753)*x[1]+(-0.50313315963)*x[2]**o+(0.356362677601)*x[2]
        arg[3,0]=(-0.444672845607)*x[0]**o+(-0.978424347535)*x[0]+(0.421699748442)*x[1]**o+(-0.0111798084161)*x[1]+(0.206042685881)*x[2]**o+(0.990523831215)*x[2]
        arg[3,1]=(0.942429627068)*x[0]**o+(-0.769989484282)*x[0]+(-0.879916738075)*x[1]**o+(-0.746448112871)*x[1]+(0.280824650861)*x[2]**o+(0.0632084823022)*x[2]
        arg[3,2]=(0.281070481177)*x[0]**o+(0.628020001876)*x[0]+(0.347767303962)*x[1]**o+(-0.276761252142)*x[1]+(0.950345260734)*x[2]**o+(-0.00148186335195)*x[2]
        arg[3,3]=(-0.37590210465)*x[0]**o+(0.034147350068)*x[0]+(0.245173418336)*x[1]**o+(0.0564051241759)*x[1]+(-0.157928106127)*x[2]**o+(-0.611677441847)*x[2]
        arg[3,4]=(-0.279066532276)*x[0]**o+(0.316033119063)*x[0]+(0.655948999472)*x[1]**o+(-0.520504180943)*x[1]+(0.531738158819)*x[2]**o+(0.899490827443)*x[2]
        ref[0,0]=(-0.330165086386)*x_ref[0]**o+(0.60381119065)*x_ref[0]+(0.981018763011)*x_ref[1]**o+(0.0487965999608)*x_ref[1]+(-0.389835388349)*x_ref[2]**o+(0.111199857376)*x_ref[2]
        ref[0,1]=(0.518239033178)*x_ref[0]**o+(0.60358784684)*x_ref[0]+(-0.476990159655)*x_ref[1]**o+(-0.93871565843)*x_ref[1]+(-0.869845295741)*x_ref[2]**o+(0.391409745802)*x_ref[2]
        ref[0,2]=(0.0927396115951)*x_ref[0]**o+(0.546122822928)*x_ref[0]+(-0.557424056733)*x_ref[1]**o+(0.0623374966436)*x_ref[1]+(-0.414595153536)*x_ref[2]**o+(-0.196069073366)*x_ref[2]
        ref[0,3]=(-0.678432580882)*x_ref[0]**o+(0.281850265141)*x_ref[0]+(-0.323978960602)*x_ref[1]**o+(0.655558769921)*x_ref[1]+(-0.295747032089)*x_ref[2]**o+(-0.152719838982)*x_ref[2]
        ref[0,4]=(-0.904177669203)*x_ref[0]**o+(0.750395234205)*x_ref[0]+(0.267431783461)*x_ref[1]**o+(-0.781406235306)*x_ref[1]+(0.856906655757)*x_ref[2]**o+(-0.875554940224)*x_ref[2]
        ref[1,0]=(-0.243144846569)*x_ref[0]**o+(0.970584922467)*x_ref[0]+(-0.977250652473)*x_ref[1]**o+(0.0239563255869)*x_ref[1]+(-0.769423602851)*x_ref[2]**o+(-0.433896361294)*x_ref[2]
        ref[1,1]=(0.31104031046)*x_ref[0]**o+(0.899859709831)*x_ref[0]+(0.692351561919)*x_ref[1]**o+(-0.0838870550002)*x_ref[1]+(-0.276105616798)*x_ref[2]**o+(0.115501348248)*x_ref[2]
        ref[1,2]=(0.56702114595)*x_ref[0]**o+(0.00240284646131)*x_ref[0]+(-0.676314888884)*x_ref[1]**o+(-0.678204346328)*x_ref[1]+(0.472697088984)*x_ref[2]**o+(-0.493949191712)*x_ref[2]
        ref[1,3]=(-0.242548554067)*x_ref[0]**o+(0.980927507053)*x_ref[0]+(-0.955706390436)*x_ref[1]**o+(0.347031550126)*x_ref[1]+(0.234352268238)*x_ref[2]**o+(0.321126345333)*x_ref[2]
        ref[1,4]=(-0.390890566228)*x_ref[0]**o+(-0.958497753144)*x_ref[0]+(0.316397716043)*x_ref[1]**o+(0.952289821769)*x_ref[1]+(-0.027182029514)*x_ref[2]**o+(-0.454422377872)*x_ref[2]
        ref[2,0]=(0.295865166318)*x_ref[0]**o+(-0.301129718468)*x_ref[0]+(-0.748135755269)*x_ref[1]**o+(-0.762413814085)*x_ref[1]+(0.572270067591)*x_ref[2]**o+(-0.393847032684)*x_ref[2]
        ref[2,1]=(0.20229063622)*x_ref[0]**o+(-0.768444028163)*x_ref[0]+(0.378669775381)*x_ref[1]**o+(-0.703384410944)*x_ref[1]+(0.355653859396)*x_ref[2]**o+(-0.228168949639)*x_ref[2]
        ref[2,2]=(-0.31725134742)*x_ref[0]**o+(0.525802867508)*x_ref[0]+(0.13375126758)*x_ref[1]**o+(-0.673551491186)*x_ref[1]+(0.498905883728)*x_ref[2]**o+(-0.102794461221)*x_ref[2]
        ref[2,3]=(-0.396813838168)*x_ref[0]**o+(0.0892605807843)*x_ref[0]+(0.244133298111)*x_ref[1]**o+(0.310900636168)*x_ref[1]+(-0.256807063422)*x_ref[2]**o+(0.565370697688)*x_ref[2]
        ref[2,4]=(0.42011835881)*x_ref[0]**o+(0.439847594375)*x_ref[0]+(0.0731897272006)*x_ref[1]**o+(-0.373911054753)*x_ref[1]+(-0.50313315963)*x_ref[2]**o+(0.356362677601)*x_ref[2]
        ref[3,0]=(-0.444672845607)*x_ref[0]**o+(-0.978424347535)*x_ref[0]+(0.421699748442)*x_ref[1]**o+(-0.0111798084161)*x_ref[1]+(0.206042685881)*x_ref[2]**o+(0.990523831215)*x_ref[2]
        ref[3,1]=(0.942429627068)*x_ref[0]**o+(-0.769989484282)*x_ref[0]+(-0.879916738075)*x_ref[1]**o+(-0.746448112871)*x_ref[1]+(0.280824650861)*x_ref[2]**o+(0.0632084823022)*x_ref[2]
        ref[3,2]=(0.281070481177)*x_ref[0]**o+(0.628020001876)*x_ref[0]+(0.347767303962)*x_ref[1]**o+(-0.276761252142)*x_ref[1]+(0.950345260734)*x_ref[2]**o+(-0.00148186335195)*x_ref[2]
        ref[3,3]=(-0.37590210465)*x_ref[0]**o+(0.034147350068)*x_ref[0]+(0.245173418336)*x_ref[1]**o+(0.0564051241759)*x_ref[1]+(-0.157928106127)*x_ref[2]**o+(-0.611677441847)*x_ref[2]
        ref[3,4]=(-0.279066532276)*x_ref[0]**o+(0.316033119063)*x_ref[0]+(0.655948999472)*x_ref[1]**o+(-0.520504180943)*x_ref[1]+(0.531738158819)*x_ref[2]**o+(0.899490827443)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.415438516191)*x[0]**o+(0.374659323173)*x[0]+(0.688470746611)*x[1]**o+(0.498183692176)*x[1]
        arg[0,0,1]=(-0.846725329522)*x[0]**o+(0.414376816326)*x[0]+(0.651609661615)*x[1]**o+(0.357606396667)*x[1]
        arg[0,1,0]=(0.728029172023)*x[0]**o+(0.975728459112)*x[0]+(-0.658211371639)*x[1]**o+(0.411894140351)*x[1]
        arg[0,1,1]=(0.807350281817)*x[0]**o+(-0.369755733522)*x[0]+(-0.519483139314)*x[1]**o+(-0.709353537648)*x[1]
        arg[1,0,0]=(0.0588259329479)*x[0]**o+(0.218362870441)*x[0]+(-0.0279239318934)*x[1]**o+(-0.75783635329)*x[1]
        arg[1,0,1]=(-0.429573558505)*x[0]**o+(-0.339046134618)*x[0]+(0.412777825743)*x[1]**o+(-0.306607333437)*x[1]
        arg[1,1,0]=(0.0222661407543)*x[0]**o+(0.927923674781)*x[0]+(-0.767468570033)*x[1]**o+(-0.0659150590367)*x[1]
        arg[1,1,1]=(0.372811733787)*x[0]**o+(-0.286500260061)*x[0]+(-0.807504375569)*x[1]**o+(-0.992850148234)*x[1]
        arg[2,0,0]=(0.368229828596)*x[0]**o+(0.672540290379)*x[0]+(-0.760462595095)*x[1]**o+(-0.761242492425)*x[1]
        arg[2,0,1]=(0.306249523133)*x[0]**o+(0.769928924717)*x[0]+(0.610293800742)*x[1]**o+(-0.221996271153)*x[1]
        arg[2,1,0]=(-0.909148832991)*x[0]**o+(-0.0293038712703)*x[0]+(-0.161707473269)*x[1]**o+(-0.534345319122)*x[1]
        arg[2,1,1]=(-0.146432487737)*x[0]**o+(0.93729266958)*x[0]+(-0.561919238789)*x[1]**o+(-0.449440501171)*x[1]
        arg[3,0,0]=(0.411711661035)*x[0]**o+(-0.0699878047274)*x[0]+(-0.0227688826645)*x[1]**o+(-0.939787633547)*x[1]
        arg[3,0,1]=(-0.82108820245)*x[0]**o+(0.225501168847)*x[0]+(0.622921674682)*x[1]**o+(0.61092413094)*x[1]
        arg[3,1,0]=(0.758116155912)*x[0]**o+(-0.117316615447)*x[0]+(0.0702245119124)*x[1]**o+(-0.275114508561)*x[1]
        arg[3,1,1]=(-0.32568550821)*x[0]**o+(-0.280144846326)*x[0]+(0.194971187876)*x[1]**o+(-0.728447244746)*x[1]
        arg[4,0,0]=(0.880042622981)*x[0]**o+(-0.395684607563)*x[0]+(-0.563304061265)*x[1]**o+(0.968663606772)*x[1]
        arg[4,0,1]=(-0.515392712458)*x[0]**o+(0.820857647342)*x[0]+(-0.476289639495)*x[1]**o+(-0.478062589609)*x[1]
        arg[4,1,0]=(0.534322425269)*x[0]**o+(0.975618695265)*x[0]+(-0.195764005045)*x[1]**o+(0.561996200643)*x[1]
        arg[4,1,1]=(-0.906243138672)*x[0]**o+(-0.580985047294)*x[0]+(-0.094609405051)*x[1]**o+(0.734326711649)*x[1]
        arg[5,0,0]=(-0.0655516121632)*x[0]**o+(0.785493123174)*x[0]+(-0.760068333675)*x[1]**o+(0.693869247554)*x[1]
        arg[5,0,1]=(-0.931019838892)*x[0]**o+(0.80958291621)*x[0]+(-0.590591206111)*x[1]**o+(-0.800900346601)*x[1]
        arg[5,1,0]=(0.448779048952)*x[0]**o+(-0.793450503268)*x[0]+(0.161067403301)*x[1]**o+(0.432054043544)*x[1]
        arg[5,1,1]=(0.0423607361972)*x[0]**o+(-0.664958045702)*x[0]+(-0.174728816918)*x[1]**o+(-0.394390068699)*x[1]
        ref[0,0,0]=(0.415438516191)*x_ref[0]**o+(0.374659323173)*x_ref[0]+(0.688470746611)*x_ref[1]**o+(0.498183692176)*x_ref[1]
        ref[0,0,1]=(-0.846725329522)*x_ref[0]**o+(0.414376816326)*x_ref[0]+(0.651609661615)*x_ref[1]**o+(0.357606396667)*x_ref[1]
        ref[0,1,0]=(0.728029172023)*x_ref[0]**o+(0.975728459112)*x_ref[0]+(-0.658211371639)*x_ref[1]**o+(0.411894140351)*x_ref[1]
        ref[0,1,1]=(0.807350281817)*x_ref[0]**o+(-0.369755733522)*x_ref[0]+(-0.519483139314)*x_ref[1]**o+(-0.709353537648)*x_ref[1]
        ref[1,0,0]=(0.0588259329479)*x_ref[0]**o+(0.218362870441)*x_ref[0]+(-0.0279239318934)*x_ref[1]**o+(-0.75783635329)*x_ref[1]
        ref[1,0,1]=(-0.429573558505)*x_ref[0]**o+(-0.339046134618)*x_ref[0]+(0.412777825743)*x_ref[1]**o+(-0.306607333437)*x_ref[1]
        ref[1,1,0]=(0.0222661407543)*x_ref[0]**o+(0.927923674781)*x_ref[0]+(-0.767468570033)*x_ref[1]**o+(-0.0659150590367)*x_ref[1]
        ref[1,1,1]=(0.372811733787)*x_ref[0]**o+(-0.286500260061)*x_ref[0]+(-0.807504375569)*x_ref[1]**o+(-0.992850148234)*x_ref[1]
        ref[2,0,0]=(0.368229828596)*x_ref[0]**o+(0.672540290379)*x_ref[0]+(-0.760462595095)*x_ref[1]**o+(-0.761242492425)*x_ref[1]
        ref[2,0,1]=(0.306249523133)*x_ref[0]**o+(0.769928924717)*x_ref[0]+(0.610293800742)*x_ref[1]**o+(-0.221996271153)*x_ref[1]
        ref[2,1,0]=(-0.909148832991)*x_ref[0]**o+(-0.0293038712703)*x_ref[0]+(-0.161707473269)*x_ref[1]**o+(-0.534345319122)*x_ref[1]
        ref[2,1,1]=(-0.146432487737)*x_ref[0]**o+(0.93729266958)*x_ref[0]+(-0.561919238789)*x_ref[1]**o+(-0.449440501171)*x_ref[1]
        ref[3,0,0]=(0.411711661035)*x_ref[0]**o+(-0.0699878047274)*x_ref[0]+(-0.0227688826645)*x_ref[1]**o+(-0.939787633547)*x_ref[1]
        ref[3,0,1]=(-0.82108820245)*x_ref[0]**o+(0.225501168847)*x_ref[0]+(0.622921674682)*x_ref[1]**o+(0.61092413094)*x_ref[1]
        ref[3,1,0]=(0.758116155912)*x_ref[0]**o+(-0.117316615447)*x_ref[0]+(0.0702245119124)*x_ref[1]**o+(-0.275114508561)*x_ref[1]
        ref[3,1,1]=(-0.32568550821)*x_ref[0]**o+(-0.280144846326)*x_ref[0]+(0.194971187876)*x_ref[1]**o+(-0.728447244746)*x_ref[1]
        ref[4,0,0]=(0.880042622981)*x_ref[0]**o+(-0.395684607563)*x_ref[0]+(-0.563304061265)*x_ref[1]**o+(0.968663606772)*x_ref[1]
        ref[4,0,1]=(-0.515392712458)*x_ref[0]**o+(0.820857647342)*x_ref[0]+(-0.476289639495)*x_ref[1]**o+(-0.478062589609)*x_ref[1]
        ref[4,1,0]=(0.534322425269)*x_ref[0]**o+(0.975618695265)*x_ref[0]+(-0.195764005045)*x_ref[1]**o+(0.561996200643)*x_ref[1]
        ref[4,1,1]=(-0.906243138672)*x_ref[0]**o+(-0.580985047294)*x_ref[0]+(-0.094609405051)*x_ref[1]**o+(0.734326711649)*x_ref[1]
        ref[5,0,0]=(-0.0655516121632)*x_ref[0]**o+(0.785493123174)*x_ref[0]+(-0.760068333675)*x_ref[1]**o+(0.693869247554)*x_ref[1]
        ref[5,0,1]=(-0.931019838892)*x_ref[0]**o+(0.80958291621)*x_ref[0]+(-0.590591206111)*x_ref[1]**o+(-0.800900346601)*x_ref[1]
        ref[5,1,0]=(0.448779048952)*x_ref[0]**o+(-0.793450503268)*x_ref[0]+(0.161067403301)*x_ref[1]**o+(0.432054043544)*x_ref[1]
        ref[5,1,1]=(0.0423607361972)*x_ref[0]**o+(-0.664958045702)*x_ref[0]+(-0.174728816918)*x_ref[1]**o+(-0.394390068699)*x_ref[1]
      else:
        arg[0,0,0]=(0.34355250678)*x[0]**o+(0.554590991175)*x[0]+(0.00673862595967)*x[1]**o+(-0.405764224977)*x[1]+(0.0865564883661)*x[2]**o+(0.102294137607)*x[2]
        arg[0,0,1]=(0.855234460708)*x[0]**o+(-0.710901486015)*x[0]+(0.546393138087)*x[1]**o+(0.847113479239)*x[1]+(-0.650701604407)*x[2]**o+(-0.914755510262)*x[2]
        arg[0,1,0]=(0.252544568503)*x[0]**o+(-0.473471976093)*x[0]+(-0.587315274764)*x[1]**o+(-0.269940894375)*x[1]+(0.360340735507)*x[2]**o+(-0.745440004827)*x[2]
        arg[0,1,1]=(0.11967506555)*x[0]**o+(0.696836453942)*x[0]+(-0.871162187396)*x[1]**o+(-0.315170111067)*x[1]+(0.232066109266)*x[2]**o+(-0.464730524838)*x[2]
        arg[1,0,0]=(-0.790535662997)*x[0]**o+(-0.964779159884)*x[0]+(-0.689937881374)*x[1]**o+(0.422799231262)*x[1]+(-0.97315874804)*x[2]**o+(-0.532103115205)*x[2]
        arg[1,0,1]=(-0.0487589554945)*x[0]**o+(-0.622735713658)*x[0]+(-0.520842613265)*x[1]**o+(0.803711422599)*x[1]+(0.440648875556)*x[2]**o+(0.98786672574)*x[2]
        arg[1,1,0]=(0.811189977538)*x[0]**o+(0.0175150486788)*x[0]+(-0.124328899976)*x[1]**o+(-0.942398680503)*x[1]+(-0.149270209895)*x[2]**o+(-0.853084407696)*x[2]
        arg[1,1,1]=(0.688378159744)*x[0]**o+(0.408766927878)*x[0]+(0.834128919609)*x[1]**o+(0.253605923886)*x[1]+(0.424992121054)*x[2]**o+(0.21016117963)*x[2]
        arg[2,0,0]=(0.277469067087)*x[0]**o+(0.539821557611)*x[0]+(0.348679402521)*x[1]**o+(-0.469772506537)*x[1]+(0.454583229137)*x[2]**o+(0.653681481133)*x[2]
        arg[2,0,1]=(-0.934067178483)*x[0]**o+(-0.172735098497)*x[0]+(-0.0347256850948)*x[1]**o+(-0.459971424944)*x[1]+(0.932661853019)*x[2]**o+(-0.186822590955)*x[2]
        arg[2,1,0]=(0.71698574321)*x[0]**o+(-0.197244831261)*x[0]+(0.183419576182)*x[1]**o+(-0.871451480853)*x[1]+(-0.349396269955)*x[2]**o+(0.095909246936)*x[2]
        arg[2,1,1]=(-0.00972985962106)*x[0]**o+(-0.919204152811)*x[0]+(-0.751744404365)*x[1]**o+(0.377096436852)*x[1]+(-0.924228885362)*x[2]**o+(0.451053813959)*x[2]
        arg[3,0,0]=(0.951119968373)*x[0]**o+(-0.0680027504528)*x[0]+(-0.113660990287)*x[1]**o+(-0.428999453003)*x[1]+(0.171804544103)*x[2]**o+(0.677635036297)*x[2]
        arg[3,0,1]=(0.62534864581)*x[0]**o+(0.389496246771)*x[0]+(0.621295327948)*x[1]**o+(-0.684165739023)*x[1]+(0.258637161492)*x[2]**o+(0.0441460764076)*x[2]
        arg[3,1,0]=(0.0348922721145)*x[0]**o+(0.091201286767)*x[0]+(0.421079169132)*x[1]**o+(-0.616504862591)*x[1]+(0.563927673176)*x[2]**o+(-0.183993394822)*x[2]
        arg[3,1,1]=(0.657959811783)*x[0]**o+(0.229729671382)*x[0]+(0.549853155207)*x[1]**o+(-0.847412930565)*x[1]+(-0.0780682500337)*x[2]**o+(0.574435563406)*x[2]
        arg[4,0,0]=(-0.679988515344)*x[0]**o+(0.0517331630343)*x[0]+(-0.160281064879)*x[1]**o+(-0.0672405944261)*x[1]+(-0.118968444607)*x[2]**o+(0.00980065834148)*x[2]
        arg[4,0,1]=(-0.150457972951)*x[0]**o+(0.671586511406)*x[0]+(-0.844112675259)*x[1]**o+(0.42766255101)*x[1]+(-0.22449721018)*x[2]**o+(-0.150416151059)*x[2]
        arg[4,1,0]=(0.0768992597753)*x[0]**o+(0.0139605476692)*x[0]+(-0.831269554863)*x[1]**o+(-0.279339344511)*x[1]+(0.387318821185)*x[2]**o+(0.967687121678)*x[2]
        arg[4,1,1]=(-0.670585948598)*x[0]**o+(0.422715818654)*x[0]+(0.623702390878)*x[1]**o+(0.922646081646)*x[1]+(-0.714191287039)*x[2]**o+(-0.99018084869)*x[2]
        arg[5,0,0]=(0.100273819257)*x[0]**o+(-0.769422291863)*x[0]+(-0.152611272068)*x[1]**o+(-0.396798631305)*x[1]+(0.502251047462)*x[2]**o+(0.993791216449)*x[2]
        arg[5,0,1]=(-0.0222540900657)*x[0]**o+(-0.339810353987)*x[0]+(0.241665078696)*x[1]**o+(0.111020568564)*x[1]+(0.298147316931)*x[2]**o+(0.391278660594)*x[2]
        arg[5,1,0]=(0.828297599555)*x[0]**o+(0.237678875371)*x[0]+(0.924551126274)*x[1]**o+(-0.663695049485)*x[1]+(-0.547123493132)*x[2]**o+(-0.788635745118)*x[2]
        arg[5,1,1]=(0.0726334370982)*x[0]**o+(-0.875688492957)*x[0]+(0.473726235108)*x[1]**o+(0.023025596151)*x[1]+(0.827449396402)*x[2]**o+(-0.154698941604)*x[2]
        ref[0,0,0]=(0.34355250678)*x_ref[0]**o+(0.554590991175)*x_ref[0]+(0.00673862595967)*x_ref[1]**o+(-0.405764224977)*x_ref[1]+(0.0865564883661)*x_ref[2]**o+(0.102294137607)*x_ref[2]
        ref[0,0,1]=(0.855234460708)*x_ref[0]**o+(-0.710901486015)*x_ref[0]+(0.546393138087)*x_ref[1]**o+(0.847113479239)*x_ref[1]+(-0.650701604407)*x_ref[2]**o+(-0.914755510262)*x_ref[2]
        ref[0,1,0]=(0.252544568503)*x_ref[0]**o+(-0.473471976093)*x_ref[0]+(-0.587315274764)*x_ref[1]**o+(-0.269940894375)*x_ref[1]+(0.360340735507)*x_ref[2]**o+(-0.745440004827)*x_ref[2]
        ref[0,1,1]=(0.11967506555)*x_ref[0]**o+(0.696836453942)*x_ref[0]+(-0.871162187396)*x_ref[1]**o+(-0.315170111067)*x_ref[1]+(0.232066109266)*x_ref[2]**o+(-0.464730524838)*x_ref[2]
        ref[1,0,0]=(-0.790535662997)*x_ref[0]**o+(-0.964779159884)*x_ref[0]+(-0.689937881374)*x_ref[1]**o+(0.422799231262)*x_ref[1]+(-0.97315874804)*x_ref[2]**o+(-0.532103115205)*x_ref[2]
        ref[1,0,1]=(-0.0487589554945)*x_ref[0]**o+(-0.622735713658)*x_ref[0]+(-0.520842613265)*x_ref[1]**o+(0.803711422599)*x_ref[1]+(0.440648875556)*x_ref[2]**o+(0.98786672574)*x_ref[2]
        ref[1,1,0]=(0.811189977538)*x_ref[0]**o+(0.0175150486788)*x_ref[0]+(-0.124328899976)*x_ref[1]**o+(-0.942398680503)*x_ref[1]+(-0.149270209895)*x_ref[2]**o+(-0.853084407696)*x_ref[2]
        ref[1,1,1]=(0.688378159744)*x_ref[0]**o+(0.408766927878)*x_ref[0]+(0.834128919609)*x_ref[1]**o+(0.253605923886)*x_ref[1]+(0.424992121054)*x_ref[2]**o+(0.21016117963)*x_ref[2]
        ref[2,0,0]=(0.277469067087)*x_ref[0]**o+(0.539821557611)*x_ref[0]+(0.348679402521)*x_ref[1]**o+(-0.469772506537)*x_ref[1]+(0.454583229137)*x_ref[2]**o+(0.653681481133)*x_ref[2]
        ref[2,0,1]=(-0.934067178483)*x_ref[0]**o+(-0.172735098497)*x_ref[0]+(-0.0347256850948)*x_ref[1]**o+(-0.459971424944)*x_ref[1]+(0.932661853019)*x_ref[2]**o+(-0.186822590955)*x_ref[2]
        ref[2,1,0]=(0.71698574321)*x_ref[0]**o+(-0.197244831261)*x_ref[0]+(0.183419576182)*x_ref[1]**o+(-0.871451480853)*x_ref[1]+(-0.349396269955)*x_ref[2]**o+(0.095909246936)*x_ref[2]
        ref[2,1,1]=(-0.00972985962106)*x_ref[0]**o+(-0.919204152811)*x_ref[0]+(-0.751744404365)*x_ref[1]**o+(0.377096436852)*x_ref[1]+(-0.924228885362)*x_ref[2]**o+(0.451053813959)*x_ref[2]
        ref[3,0,0]=(0.951119968373)*x_ref[0]**o+(-0.0680027504528)*x_ref[0]+(-0.113660990287)*x_ref[1]**o+(-0.428999453003)*x_ref[1]+(0.171804544103)*x_ref[2]**o+(0.677635036297)*x_ref[2]
        ref[3,0,1]=(0.62534864581)*x_ref[0]**o+(0.389496246771)*x_ref[0]+(0.621295327948)*x_ref[1]**o+(-0.684165739023)*x_ref[1]+(0.258637161492)*x_ref[2]**o+(0.0441460764076)*x_ref[2]
        ref[3,1,0]=(0.0348922721145)*x_ref[0]**o+(0.091201286767)*x_ref[0]+(0.421079169132)*x_ref[1]**o+(-0.616504862591)*x_ref[1]+(0.563927673176)*x_ref[2]**o+(-0.183993394822)*x_ref[2]
        ref[3,1,1]=(0.657959811783)*x_ref[0]**o+(0.229729671382)*x_ref[0]+(0.549853155207)*x_ref[1]**o+(-0.847412930565)*x_ref[1]+(-0.0780682500337)*x_ref[2]**o+(0.574435563406)*x_ref[2]
        ref[4,0,0]=(-0.679988515344)*x_ref[0]**o+(0.0517331630343)*x_ref[0]+(-0.160281064879)*x_ref[1]**o+(-0.0672405944261)*x_ref[1]+(-0.118968444607)*x_ref[2]**o+(0.00980065834148)*x_ref[2]
        ref[4,0,1]=(-0.150457972951)*x_ref[0]**o+(0.671586511406)*x_ref[0]+(-0.844112675259)*x_ref[1]**o+(0.42766255101)*x_ref[1]+(-0.22449721018)*x_ref[2]**o+(-0.150416151059)*x_ref[2]
        ref[4,1,0]=(0.0768992597753)*x_ref[0]**o+(0.0139605476692)*x_ref[0]+(-0.831269554863)*x_ref[1]**o+(-0.279339344511)*x_ref[1]+(0.387318821185)*x_ref[2]**o+(0.967687121678)*x_ref[2]
        ref[4,1,1]=(-0.670585948598)*x_ref[0]**o+(0.422715818654)*x_ref[0]+(0.623702390878)*x_ref[1]**o+(0.922646081646)*x_ref[1]+(-0.714191287039)*x_ref[2]**o+(-0.99018084869)*x_ref[2]
        ref[5,0,0]=(0.100273819257)*x_ref[0]**o+(-0.769422291863)*x_ref[0]+(-0.152611272068)*x_ref[1]**o+(-0.396798631305)*x_ref[1]+(0.502251047462)*x_ref[2]**o+(0.993791216449)*x_ref[2]
        ref[5,0,1]=(-0.0222540900657)*x_ref[0]**o+(-0.339810353987)*x_ref[0]+(0.241665078696)*x_ref[1]**o+(0.111020568564)*x_ref[1]+(0.298147316931)*x_ref[2]**o+(0.391278660594)*x_ref[2]
        ref[5,1,0]=(0.828297599555)*x_ref[0]**o+(0.237678875371)*x_ref[0]+(0.924551126274)*x_ref[1]**o+(-0.663695049485)*x_ref[1]+(-0.547123493132)*x_ref[2]**o+(-0.788635745118)*x_ref[2]
        ref[5,1,1]=(0.0726334370982)*x_ref[0]**o+(-0.875688492957)*x_ref[0]+(0.473726235108)*x_ref[1]**o+(0.023025596151)*x_ref[1]+(0.827449396402)*x_ref[2]**o+(-0.154698941604)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0338299192888)*x[0]**o+(-0.114906274359)*x[0]+(0.15010339398)*x[1]**o+(0.625295991328)*x[1]
        arg[0,0,0,1]=(-0.1311901039)*x[0]**o+(-0.551712861217)*x[0]+(-0.287411404506)*x[1]**o+(0.0242820629572)*x[1]
        arg[0,0,1,0]=(0.121962094111)*x[0]**o+(0.203169622345)*x[0]+(0.241504218053)*x[1]**o+(-0.0696356239498)*x[1]
        arg[0,0,1,1]=(0.0461537352329)*x[0]**o+(-0.275908938567)*x[0]+(-0.271184121027)*x[1]**o+(-0.364024729567)*x[1]
        arg[0,0,2,0]=(0.199272234315)*x[0]**o+(-0.906364258855)*x[0]+(-0.92102930037)*x[1]**o+(0.0718981903454)*x[1]
        arg[0,0,2,1]=(-0.742950958383)*x[0]**o+(0.0301464847722)*x[0]+(0.466993477687)*x[1]**o+(-0.402131750698)*x[1]
        arg[0,1,0,0]=(-0.0653776546375)*x[0]**o+(-0.679360612372)*x[0]+(-0.32783837768)*x[1]**o+(-0.113753016428)*x[1]
        arg[0,1,0,1]=(0.58610119414)*x[0]**o+(0.142613166231)*x[0]+(0.0409105619986)*x[1]**o+(-0.660641663558)*x[1]
        arg[0,1,1,0]=(0.212172308585)*x[0]**o+(-0.452767768344)*x[0]+(0.963053969889)*x[1]**o+(0.752691773953)*x[1]
        arg[0,1,1,1]=(-0.514998377887)*x[0]**o+(0.0872464329801)*x[0]+(0.265590295505)*x[1]**o+(0.557238826549)*x[1]
        arg[0,1,2,0]=(0.360523954724)*x[0]**o+(0.550238402994)*x[0]+(-0.206349415539)*x[1]**o+(0.357304730511)*x[1]
        arg[0,1,2,1]=(-0.872755217435)*x[0]**o+(0.190943683664)*x[0]+(-0.245454292354)*x[1]**o+(0.507229578506)*x[1]
        arg[0,2,0,0]=(0.994619505257)*x[0]**o+(0.33482820568)*x[0]+(-0.12721791236)*x[1]**o+(-0.505873171195)*x[1]
        arg[0,2,0,1]=(0.224805131522)*x[0]**o+(0.332022434603)*x[0]+(0.0583494507268)*x[1]**o+(0.857185410994)*x[1]
        arg[0,2,1,0]=(-0.425367077516)*x[0]**o+(-0.25207101061)*x[0]+(-0.11164491753)*x[1]**o+(0.44752084285)*x[1]
        arg[0,2,1,1]=(-0.158725210751)*x[0]**o+(-0.0692363138932)*x[0]+(0.214803128256)*x[1]**o+(0.886699106748)*x[1]
        arg[0,2,2,0]=(-0.332146331257)*x[0]**o+(0.114845050128)*x[0]+(-0.0781411611124)*x[1]**o+(0.474972057387)*x[1]
        arg[0,2,2,1]=(-0.481949630694)*x[0]**o+(0.644984972151)*x[0]+(0.171146519527)*x[1]**o+(0.57534803966)*x[1]
        arg[0,3,0,0]=(0.7580633024)*x[0]**o+(0.0280511776577)*x[0]+(0.375583652078)*x[1]**o+(-0.809112787284)*x[1]
        arg[0,3,0,1]=(-0.466737159934)*x[0]**o+(0.410032775686)*x[0]+(0.176141189584)*x[1]**o+(-0.228288309456)*x[1]
        arg[0,3,1,0]=(0.427129582196)*x[0]**o+(-0.718385859483)*x[0]+(-0.172261212989)*x[1]**o+(0.913157476188)*x[1]
        arg[0,3,1,1]=(0.952506706487)*x[0]**o+(-0.497478163899)*x[0]+(-0.930273038716)*x[1]**o+(0.673573149025)*x[1]
        arg[0,3,2,0]=(-0.0074302930747)*x[0]**o+(0.226424842035)*x[0]+(0.871190620671)*x[1]**o+(0.277695333679)*x[1]
        arg[0,3,2,1]=(-0.540820239946)*x[0]**o+(0.33044763211)*x[0]+(-0.772648820924)*x[1]**o+(-0.362410088021)*x[1]
        arg[0,4,0,0]=(-0.748396039111)*x[0]**o+(0.215131518212)*x[0]+(0.514087824403)*x[1]**o+(0.335563854205)*x[1]
        arg[0,4,0,1]=(-0.443263590391)*x[0]**o+(-0.523167282542)*x[0]+(0.122612452799)*x[1]**o+(0.216527857615)*x[1]
        arg[0,4,1,0]=(0.448176296777)*x[0]**o+(0.478292146131)*x[0]+(-0.558921083502)*x[1]**o+(-0.17018258654)*x[1]
        arg[0,4,1,1]=(-0.912765673953)*x[0]**o+(0.751746623223)*x[0]+(-0.104811579057)*x[1]**o+(0.456436870323)*x[1]
        arg[0,4,2,0]=(0.157946450932)*x[0]**o+(0.147858587857)*x[0]+(0.775747362874)*x[1]**o+(0.478478862463)*x[1]
        arg[0,4,2,1]=(-0.819625058185)*x[0]**o+(-0.0231980393919)*x[0]+(-0.186647585522)*x[1]**o+(0.513350371371)*x[1]
        arg[1,0,0,0]=(-0.942373946995)*x[0]**o+(-0.761326160931)*x[0]+(0.192511867167)*x[1]**o+(-0.50895969579)*x[1]
        arg[1,0,0,1]=(0.0670813349951)*x[0]**o+(0.55602953512)*x[0]+(0.757621178635)*x[1]**o+(0.741550201759)*x[1]
        arg[1,0,1,0]=(-0.114642609493)*x[0]**o+(-0.639500048447)*x[0]+(-0.861813039243)*x[1]**o+(0.36413406744)*x[1]
        arg[1,0,1,1]=(0.126555451594)*x[0]**o+(-0.885172557082)*x[0]+(-0.376073252921)*x[1]**o+(0.166102850653)*x[1]
        arg[1,0,2,0]=(-0.381517068219)*x[0]**o+(0.94870328267)*x[0]+(-0.129934777755)*x[1]**o+(0.78090715758)*x[1]
        arg[1,0,2,1]=(-0.535852302247)*x[0]**o+(0.63751039318)*x[0]+(-0.537556159925)*x[1]**o+(0.147766714221)*x[1]
        arg[1,1,0,0]=(-0.605543047533)*x[0]**o+(-0.366612785372)*x[0]+(0.544900149212)*x[1]**o+(0.722108900447)*x[1]
        arg[1,1,0,1]=(0.133378561142)*x[0]**o+(-0.227808772494)*x[0]+(-0.729399624175)*x[1]**o+(0.307571743455)*x[1]
        arg[1,1,1,0]=(0.0458830538139)*x[0]**o+(0.0585044493026)*x[0]+(-0.136630547605)*x[1]**o+(-0.0823288389309)*x[1]
        arg[1,1,1,1]=(0.339263808784)*x[0]**o+(0.00988917109076)*x[0]+(-0.786591838083)*x[1]**o+(-0.361686799179)*x[1]
        arg[1,1,2,0]=(0.789942037592)*x[0]**o+(-0.266689758676)*x[0]+(0.464856327495)*x[1]**o+(0.138022774293)*x[1]
        arg[1,1,2,1]=(-0.723544867217)*x[0]**o+(-0.523023674205)*x[0]+(0.629726026331)*x[1]**o+(-0.867808120997)*x[1]
        arg[1,2,0,0]=(-0.495589518787)*x[0]**o+(-0.561664880262)*x[0]+(0.0319320417078)*x[1]**o+(-0.153720661498)*x[1]
        arg[1,2,0,1]=(0.293168114024)*x[0]**o+(-0.344107163981)*x[0]+(-0.275629670068)*x[1]**o+(0.06103653487)*x[1]
        arg[1,2,1,0]=(0.495640927339)*x[0]**o+(0.596540095043)*x[0]+(0.208792567756)*x[1]**o+(0.886040724145)*x[1]
        arg[1,2,1,1]=(0.331476398093)*x[0]**o+(-0.472588572455)*x[0]+(-0.455670457285)*x[1]**o+(-0.731223615642)*x[1]
        arg[1,2,2,0]=(-0.127107829674)*x[0]**o+(-0.21113128456)*x[0]+(-0.733928194168)*x[1]**o+(-0.174127696853)*x[1]
        arg[1,2,2,1]=(0.521879316151)*x[0]**o+(0.983424557865)*x[0]+(-0.288659660182)*x[1]**o+(-0.431045203113)*x[1]
        arg[1,3,0,0]=(0.292348765746)*x[0]**o+(0.0730741240915)*x[0]+(0.079624022683)*x[1]**o+(0.237483071633)*x[1]
        arg[1,3,0,1]=(0.291035771848)*x[0]**o+(-0.482909477855)*x[0]+(-0.999142129615)*x[1]**o+(-0.559723367197)*x[1]
        arg[1,3,1,0]=(0.747096275653)*x[0]**o+(0.762757649624)*x[0]+(-0.821650889507)*x[1]**o+(0.301128670189)*x[1]
        arg[1,3,1,1]=(0.82703681414)*x[0]**o+(-0.336570070436)*x[0]+(0.798131952979)*x[1]**o+(0.603562737956)*x[1]
        arg[1,3,2,0]=(-0.164875914126)*x[0]**o+(0.219409873902)*x[0]+(0.784253160678)*x[1]**o+(0.0219597460707)*x[1]
        arg[1,3,2,1]=(0.124778899794)*x[0]**o+(0.785367842145)*x[0]+(0.948111125908)*x[1]**o+(-0.887427847089)*x[1]
        arg[1,4,0,0]=(0.0911816002883)*x[0]**o+(0.645410022441)*x[0]+(0.720551531965)*x[1]**o+(-0.388407683413)*x[1]
        arg[1,4,0,1]=(-0.0416762613062)*x[0]**o+(-0.548874829834)*x[0]+(-0.377952555557)*x[1]**o+(0.742795234361)*x[1]
        arg[1,4,1,0]=(-0.896929017971)*x[0]**o+(0.678343686516)*x[0]+(0.711572763082)*x[1]**o+(-0.43397331651)*x[1]
        arg[1,4,1,1]=(-0.190804421566)*x[0]**o+(-0.0848405124789)*x[0]+(0.36246957905)*x[1]**o+(0.963599086227)*x[1]
        arg[1,4,2,0]=(-0.0210248567982)*x[0]**o+(0.545977814581)*x[0]+(0.858962864401)*x[1]**o+(0.113792851997)*x[1]
        arg[1,4,2,1]=(0.229641892622)*x[0]**o+(-0.739561617198)*x[0]+(-0.164372740617)*x[1]**o+(0.537097868843)*x[1]
        arg[2,0,0,0]=(-0.13321389343)*x[0]**o+(-0.668403350503)*x[0]+(-0.924879105983)*x[1]**o+(-0.5161790717)*x[1]
        arg[2,0,0,1]=(0.26582844685)*x[0]**o+(0.89048660684)*x[0]+(-0.377225981719)*x[1]**o+(-0.303790710194)*x[1]
        arg[2,0,1,0]=(0.0997470860283)*x[0]**o+(0.0385427264217)*x[0]+(-0.71387347273)*x[1]**o+(-0.621376499192)*x[1]
        arg[2,0,1,1]=(0.595544616326)*x[0]**o+(-0.850529503975)*x[0]+(0.602203957938)*x[1]**o+(-0.329484843907)*x[1]
        arg[2,0,2,0]=(-0.106031967029)*x[0]**o+(-0.165122821622)*x[0]+(0.647473815919)*x[1]**o+(0.28942226062)*x[1]
        arg[2,0,2,1]=(0.585955875472)*x[0]**o+(0.922667710403)*x[0]+(-0.0710905957841)*x[1]**o+(-0.669138163697)*x[1]
        arg[2,1,0,0]=(0.461941211081)*x[0]**o+(0.723022466263)*x[0]+(-0.118551893472)*x[1]**o+(0.00445187353637)*x[1]
        arg[2,1,0,1]=(-0.253168248108)*x[0]**o+(0.476938842605)*x[0]+(0.476641658034)*x[1]**o+(-0.916886387595)*x[1]
        arg[2,1,1,0]=(0.722029309503)*x[0]**o+(0.0588848593425)*x[0]+(0.84737370043)*x[1]**o+(-0.674604415213)*x[1]
        arg[2,1,1,1]=(-0.644736759298)*x[0]**o+(-0.212847303893)*x[0]+(0.25394587552)*x[1]**o+(-0.366955175081)*x[1]
        arg[2,1,2,0]=(-0.657395296801)*x[0]**o+(-0.967570027599)*x[0]+(0.180926628541)*x[1]**o+(-0.940599897942)*x[1]
        arg[2,1,2,1]=(0.849897328801)*x[0]**o+(-0.299199474051)*x[0]+(0.427221207827)*x[1]**o+(0.212027985441)*x[1]
        arg[2,2,0,0]=(-0.561573507381)*x[0]**o+(-0.446925224908)*x[0]+(0.132091224948)*x[1]**o+(-0.673375661849)*x[1]
        arg[2,2,0,1]=(-0.386657291574)*x[0]**o+(0.356784637112)*x[0]+(0.644292006233)*x[1]**o+(0.922937167842)*x[1]
        arg[2,2,1,0]=(0.484446757285)*x[0]**o+(-0.822469870455)*x[0]+(-0.89605890123)*x[1]**o+(-0.895891500031)*x[1]
        arg[2,2,1,1]=(-0.0930740520192)*x[0]**o+(0.953356473678)*x[0]+(-0.714418458806)*x[1]**o+(-0.100584950946)*x[1]
        arg[2,2,2,0]=(0.945572242667)*x[0]**o+(-0.259280849605)*x[0]+(-0.925459846341)*x[1]**o+(0.906002758873)*x[1]
        arg[2,2,2,1]=(-0.638993332975)*x[0]**o+(-0.668594641703)*x[0]+(0.012666231369)*x[1]**o+(0.89732685856)*x[1]
        arg[2,3,0,0]=(-0.885061116144)*x[0]**o+(-0.576097536483)*x[0]+(-0.43166075788)*x[1]**o+(0.538821620917)*x[1]
        arg[2,3,0,1]=(-0.0743638381797)*x[0]**o+(0.251037500035)*x[0]+(-0.905989928076)*x[1]**o+(0.243771000867)*x[1]
        arg[2,3,1,0]=(-0.016045622351)*x[0]**o+(-0.632944072174)*x[0]+(-0.296175918793)*x[1]**o+(-0.00120123641365)*x[1]
        arg[2,3,1,1]=(-0.896354921812)*x[0]**o+(-0.331348989027)*x[0]+(-0.450865285201)*x[1]**o+(-0.243268582009)*x[1]
        arg[2,3,2,0]=(0.609332927431)*x[0]**o+(-0.0815856802115)*x[0]+(0.201924792273)*x[1]**o+(-0.890424451348)*x[1]
        arg[2,3,2,1]=(0.395090137264)*x[0]**o+(0.294230442573)*x[0]+(-0.345371510364)*x[1]**o+(0.181229663241)*x[1]
        arg[2,4,0,0]=(-0.823684509034)*x[0]**o+(0.431019301942)*x[0]+(0.554707015054)*x[1]**o+(-0.846322903003)*x[1]
        arg[2,4,0,1]=(0.862984476658)*x[0]**o+(0.20512500102)*x[0]+(0.0247756159287)*x[1]**o+(-0.374795100946)*x[1]
        arg[2,4,1,0]=(-0.0849914269517)*x[0]**o+(0.754908200339)*x[0]+(0.593363754807)*x[1]**o+(0.124923961126)*x[1]
        arg[2,4,1,1]=(-0.691440264234)*x[0]**o+(0.825739298015)*x[0]+(0.426452231478)*x[1]**o+(0.0995610721341)*x[1]
        arg[2,4,2,0]=(-0.175322750567)*x[0]**o+(0.419750865569)*x[0]+(0.208448987712)*x[1]**o+(0.150584798952)*x[1]
        arg[2,4,2,1]=(0.931680152344)*x[0]**o+(0.390947206307)*x[0]+(0.388437348378)*x[1]**o+(-0.808413079032)*x[1]
        arg[3,0,0,0]=(0.58855431402)*x[0]**o+(0.383865913364)*x[0]+(-0.691445323219)*x[1]**o+(-0.443370658523)*x[1]
        arg[3,0,0,1]=(-0.657371481478)*x[0]**o+(0.746891082673)*x[0]+(-0.829229947773)*x[1]**o+(0.761915670404)*x[1]
        arg[3,0,1,0]=(-0.431502932714)*x[0]**o+(-0.961472285716)*x[0]+(0.0423384388031)*x[1]**o+(0.339373458118)*x[1]
        arg[3,0,1,1]=(-0.272831163259)*x[0]**o+(-0.533438904494)*x[0]+(-0.595625384875)*x[1]**o+(0.0366857022323)*x[1]
        arg[3,0,2,0]=(-0.319684662306)*x[0]**o+(0.672185828912)*x[0]+(0.483126548994)*x[1]**o+(0.745872629604)*x[1]
        arg[3,0,2,1]=(0.437895277404)*x[0]**o+(-0.337209416963)*x[0]+(0.101080522788)*x[1]**o+(0.892022488644)*x[1]
        arg[3,1,0,0]=(-0.582325078455)*x[0]**o+(-0.854689347157)*x[0]+(0.0308638312259)*x[1]**o+(0.633382365704)*x[1]
        arg[3,1,0,1]=(0.319735178376)*x[0]**o+(-0.782764593952)*x[0]+(0.082230132244)*x[1]**o+(-0.946010937852)*x[1]
        arg[3,1,1,0]=(-0.698490410666)*x[0]**o+(-0.801358802689)*x[0]+(0.20695648235)*x[1]**o+(0.75257523471)*x[1]
        arg[3,1,1,1]=(0.66321696527)*x[0]**o+(0.704572918868)*x[0]+(0.528208879002)*x[1]**o+(-0.408378325513)*x[1]
        arg[3,1,2,0]=(-0.925177198417)*x[0]**o+(-0.152427811549)*x[0]+(0.341781993583)*x[1]**o+(-0.589976456377)*x[1]
        arg[3,1,2,1]=(0.206706628353)*x[0]**o+(0.673777173329)*x[0]+(-0.994138314254)*x[1]**o+(-0.843696565863)*x[1]
        arg[3,2,0,0]=(-0.588881067333)*x[0]**o+(-0.537572388341)*x[0]+(-0.206570371675)*x[1]**o+(-0.127468554294)*x[1]
        arg[3,2,0,1]=(0.498638284096)*x[0]**o+(-0.488278831086)*x[0]+(-0.28320874594)*x[1]**o+(-0.638241576615)*x[1]
        arg[3,2,1,0]=(-0.445780790822)*x[0]**o+(-0.347493953089)*x[0]+(-0.558010086232)*x[1]**o+(-0.0287339307689)*x[1]
        arg[3,2,1,1]=(-0.361727811272)*x[0]**o+(-0.559565378007)*x[0]+(-0.211685558972)*x[1]**o+(0.330088396811)*x[1]
        arg[3,2,2,0]=(-0.42600117831)*x[0]**o+(-0.488997111988)*x[0]+(0.127421278051)*x[1]**o+(-0.48233167806)*x[1]
        arg[3,2,2,1]=(0.120468829686)*x[0]**o+(0.0675603906118)*x[0]+(-0.60674361864)*x[1]**o+(0.639115708258)*x[1]
        arg[3,3,0,0]=(0.369343033015)*x[0]**o+(0.707735743593)*x[0]+(-0.426275591836)*x[1]**o+(-0.733377765625)*x[1]
        arg[3,3,0,1]=(-0.931245338156)*x[0]**o+(-0.465861794185)*x[0]+(-0.650664500687)*x[1]**o+(0.103077662402)*x[1]
        arg[3,3,1,0]=(0.731834744512)*x[0]**o+(0.267840571071)*x[0]+(0.952106457495)*x[1]**o+(-0.943845364366)*x[1]
        arg[3,3,1,1]=(-0.21692720916)*x[0]**o+(0.580841374997)*x[0]+(-0.412129341223)*x[1]**o+(0.40414445304)*x[1]
        arg[3,3,2,0]=(-0.447758598726)*x[0]**o+(0.0616926352912)*x[0]+(-0.824624226122)*x[1]**o+(0.19099833104)*x[1]
        arg[3,3,2,1]=(-0.615519138165)*x[0]**o+(0.233926822164)*x[0]+(-0.969014060236)*x[1]**o+(0.835103627893)*x[1]
        arg[3,4,0,0]=(0.550778983301)*x[0]**o+(0.0198253530882)*x[0]+(-0.39383849187)*x[1]**o+(0.367721826304)*x[1]
        arg[3,4,0,1]=(0.444877923149)*x[0]**o+(-0.899159835223)*x[0]+(0.923392540452)*x[1]**o+(0.512905141718)*x[1]
        arg[3,4,1,0]=(-0.563807675145)*x[0]**o+(0.994405383899)*x[0]+(-0.777352948218)*x[1]**o+(0.865994906227)*x[1]
        arg[3,4,1,1]=(-0.593800452139)*x[0]**o+(0.613778803237)*x[0]+(0.582200979718)*x[1]**o+(0.448026243271)*x[1]
        arg[3,4,2,0]=(-0.900331125416)*x[0]**o+(0.234613965424)*x[0]+(0.510400273805)*x[1]**o+(0.0836773641661)*x[1]
        arg[3,4,2,1]=(0.710999097307)*x[0]**o+(0.886603988265)*x[0]+(-0.405644887679)*x[1]**o+(0.362257258809)*x[1]
        ref[0,0,0,0]=(-0.0338299192888)*x_ref[0]**o+(-0.114906274359)*x_ref[0]+(0.15010339398)*x_ref[1]**o+(0.625295991328)*x_ref[1]
        ref[0,0,0,1]=(-0.1311901039)*x_ref[0]**o+(-0.551712861217)*x_ref[0]+(-0.287411404506)*x_ref[1]**o+(0.0242820629572)*x_ref[1]
        ref[0,0,1,0]=(0.121962094111)*x_ref[0]**o+(0.203169622345)*x_ref[0]+(0.241504218053)*x_ref[1]**o+(-0.0696356239498)*x_ref[1]
        ref[0,0,1,1]=(0.0461537352329)*x_ref[0]**o+(-0.275908938567)*x_ref[0]+(-0.271184121027)*x_ref[1]**o+(-0.364024729567)*x_ref[1]
        ref[0,0,2,0]=(0.199272234315)*x_ref[0]**o+(-0.906364258855)*x_ref[0]+(-0.92102930037)*x_ref[1]**o+(0.0718981903454)*x_ref[1]
        ref[0,0,2,1]=(-0.742950958383)*x_ref[0]**o+(0.0301464847722)*x_ref[0]+(0.466993477687)*x_ref[1]**o+(-0.402131750698)*x_ref[1]
        ref[0,1,0,0]=(-0.0653776546375)*x_ref[0]**o+(-0.679360612372)*x_ref[0]+(-0.32783837768)*x_ref[1]**o+(-0.113753016428)*x_ref[1]
        ref[0,1,0,1]=(0.58610119414)*x_ref[0]**o+(0.142613166231)*x_ref[0]+(0.0409105619986)*x_ref[1]**o+(-0.660641663558)*x_ref[1]
        ref[0,1,1,0]=(0.212172308585)*x_ref[0]**o+(-0.452767768344)*x_ref[0]+(0.963053969889)*x_ref[1]**o+(0.752691773953)*x_ref[1]
        ref[0,1,1,1]=(-0.514998377887)*x_ref[0]**o+(0.0872464329801)*x_ref[0]+(0.265590295505)*x_ref[1]**o+(0.557238826549)*x_ref[1]
        ref[0,1,2,0]=(0.360523954724)*x_ref[0]**o+(0.550238402994)*x_ref[0]+(-0.206349415539)*x_ref[1]**o+(0.357304730511)*x_ref[1]
        ref[0,1,2,1]=(-0.872755217435)*x_ref[0]**o+(0.190943683664)*x_ref[0]+(-0.245454292354)*x_ref[1]**o+(0.507229578506)*x_ref[1]
        ref[0,2,0,0]=(0.994619505257)*x_ref[0]**o+(0.33482820568)*x_ref[0]+(-0.12721791236)*x_ref[1]**o+(-0.505873171195)*x_ref[1]
        ref[0,2,0,1]=(0.224805131522)*x_ref[0]**o+(0.332022434603)*x_ref[0]+(0.0583494507268)*x_ref[1]**o+(0.857185410994)*x_ref[1]
        ref[0,2,1,0]=(-0.425367077516)*x_ref[0]**o+(-0.25207101061)*x_ref[0]+(-0.11164491753)*x_ref[1]**o+(0.44752084285)*x_ref[1]
        ref[0,2,1,1]=(-0.158725210751)*x_ref[0]**o+(-0.0692363138932)*x_ref[0]+(0.214803128256)*x_ref[1]**o+(0.886699106748)*x_ref[1]
        ref[0,2,2,0]=(-0.332146331257)*x_ref[0]**o+(0.114845050128)*x_ref[0]+(-0.0781411611124)*x_ref[1]**o+(0.474972057387)*x_ref[1]
        ref[0,2,2,1]=(-0.481949630694)*x_ref[0]**o+(0.644984972151)*x_ref[0]+(0.171146519527)*x_ref[1]**o+(0.57534803966)*x_ref[1]
        ref[0,3,0,0]=(0.7580633024)*x_ref[0]**o+(0.0280511776577)*x_ref[0]+(0.375583652078)*x_ref[1]**o+(-0.809112787284)*x_ref[1]
        ref[0,3,0,1]=(-0.466737159934)*x_ref[0]**o+(0.410032775686)*x_ref[0]+(0.176141189584)*x_ref[1]**o+(-0.228288309456)*x_ref[1]
        ref[0,3,1,0]=(0.427129582196)*x_ref[0]**o+(-0.718385859483)*x_ref[0]+(-0.172261212989)*x_ref[1]**o+(0.913157476188)*x_ref[1]
        ref[0,3,1,1]=(0.952506706487)*x_ref[0]**o+(-0.497478163899)*x_ref[0]+(-0.930273038716)*x_ref[1]**o+(0.673573149025)*x_ref[1]
        ref[0,3,2,0]=(-0.0074302930747)*x_ref[0]**o+(0.226424842035)*x_ref[0]+(0.871190620671)*x_ref[1]**o+(0.277695333679)*x_ref[1]
        ref[0,3,2,1]=(-0.540820239946)*x_ref[0]**o+(0.33044763211)*x_ref[0]+(-0.772648820924)*x_ref[1]**o+(-0.362410088021)*x_ref[1]
        ref[0,4,0,0]=(-0.748396039111)*x_ref[0]**o+(0.215131518212)*x_ref[0]+(0.514087824403)*x_ref[1]**o+(0.335563854205)*x_ref[1]
        ref[0,4,0,1]=(-0.443263590391)*x_ref[0]**o+(-0.523167282542)*x_ref[0]+(0.122612452799)*x_ref[1]**o+(0.216527857615)*x_ref[1]
        ref[0,4,1,0]=(0.448176296777)*x_ref[0]**o+(0.478292146131)*x_ref[0]+(-0.558921083502)*x_ref[1]**o+(-0.17018258654)*x_ref[1]
        ref[0,4,1,1]=(-0.912765673953)*x_ref[0]**o+(0.751746623223)*x_ref[0]+(-0.104811579057)*x_ref[1]**o+(0.456436870323)*x_ref[1]
        ref[0,4,2,0]=(0.157946450932)*x_ref[0]**o+(0.147858587857)*x_ref[0]+(0.775747362874)*x_ref[1]**o+(0.478478862463)*x_ref[1]
        ref[0,4,2,1]=(-0.819625058185)*x_ref[0]**o+(-0.0231980393919)*x_ref[0]+(-0.186647585522)*x_ref[1]**o+(0.513350371371)*x_ref[1]
        ref[1,0,0,0]=(-0.942373946995)*x_ref[0]**o+(-0.761326160931)*x_ref[0]+(0.192511867167)*x_ref[1]**o+(-0.50895969579)*x_ref[1]
        ref[1,0,0,1]=(0.0670813349951)*x_ref[0]**o+(0.55602953512)*x_ref[0]+(0.757621178635)*x_ref[1]**o+(0.741550201759)*x_ref[1]
        ref[1,0,1,0]=(-0.114642609493)*x_ref[0]**o+(-0.639500048447)*x_ref[0]+(-0.861813039243)*x_ref[1]**o+(0.36413406744)*x_ref[1]
        ref[1,0,1,1]=(0.126555451594)*x_ref[0]**o+(-0.885172557082)*x_ref[0]+(-0.376073252921)*x_ref[1]**o+(0.166102850653)*x_ref[1]
        ref[1,0,2,0]=(-0.381517068219)*x_ref[0]**o+(0.94870328267)*x_ref[0]+(-0.129934777755)*x_ref[1]**o+(0.78090715758)*x_ref[1]
        ref[1,0,2,1]=(-0.535852302247)*x_ref[0]**o+(0.63751039318)*x_ref[0]+(-0.537556159925)*x_ref[1]**o+(0.147766714221)*x_ref[1]
        ref[1,1,0,0]=(-0.605543047533)*x_ref[0]**o+(-0.366612785372)*x_ref[0]+(0.544900149212)*x_ref[1]**o+(0.722108900447)*x_ref[1]
        ref[1,1,0,1]=(0.133378561142)*x_ref[0]**o+(-0.227808772494)*x_ref[0]+(-0.729399624175)*x_ref[1]**o+(0.307571743455)*x_ref[1]
        ref[1,1,1,0]=(0.0458830538139)*x_ref[0]**o+(0.0585044493026)*x_ref[0]+(-0.136630547605)*x_ref[1]**o+(-0.0823288389309)*x_ref[1]
        ref[1,1,1,1]=(0.339263808784)*x_ref[0]**o+(0.00988917109076)*x_ref[0]+(-0.786591838083)*x_ref[1]**o+(-0.361686799179)*x_ref[1]
        ref[1,1,2,0]=(0.789942037592)*x_ref[0]**o+(-0.266689758676)*x_ref[0]+(0.464856327495)*x_ref[1]**o+(0.138022774293)*x_ref[1]
        ref[1,1,2,1]=(-0.723544867217)*x_ref[0]**o+(-0.523023674205)*x_ref[0]+(0.629726026331)*x_ref[1]**o+(-0.867808120997)*x_ref[1]
        ref[1,2,0,0]=(-0.495589518787)*x_ref[0]**o+(-0.561664880262)*x_ref[0]+(0.0319320417078)*x_ref[1]**o+(-0.153720661498)*x_ref[1]
        ref[1,2,0,1]=(0.293168114024)*x_ref[0]**o+(-0.344107163981)*x_ref[0]+(-0.275629670068)*x_ref[1]**o+(0.06103653487)*x_ref[1]
        ref[1,2,1,0]=(0.495640927339)*x_ref[0]**o+(0.596540095043)*x_ref[0]+(0.208792567756)*x_ref[1]**o+(0.886040724145)*x_ref[1]
        ref[1,2,1,1]=(0.331476398093)*x_ref[0]**o+(-0.472588572455)*x_ref[0]+(-0.455670457285)*x_ref[1]**o+(-0.731223615642)*x_ref[1]
        ref[1,2,2,0]=(-0.127107829674)*x_ref[0]**o+(-0.21113128456)*x_ref[0]+(-0.733928194168)*x_ref[1]**o+(-0.174127696853)*x_ref[1]
        ref[1,2,2,1]=(0.521879316151)*x_ref[0]**o+(0.983424557865)*x_ref[0]+(-0.288659660182)*x_ref[1]**o+(-0.431045203113)*x_ref[1]
        ref[1,3,0,0]=(0.292348765746)*x_ref[0]**o+(0.0730741240915)*x_ref[0]+(0.079624022683)*x_ref[1]**o+(0.237483071633)*x_ref[1]
        ref[1,3,0,1]=(0.291035771848)*x_ref[0]**o+(-0.482909477855)*x_ref[0]+(-0.999142129615)*x_ref[1]**o+(-0.559723367197)*x_ref[1]
        ref[1,3,1,0]=(0.747096275653)*x_ref[0]**o+(0.762757649624)*x_ref[0]+(-0.821650889507)*x_ref[1]**o+(0.301128670189)*x_ref[1]
        ref[1,3,1,1]=(0.82703681414)*x_ref[0]**o+(-0.336570070436)*x_ref[0]+(0.798131952979)*x_ref[1]**o+(0.603562737956)*x_ref[1]
        ref[1,3,2,0]=(-0.164875914126)*x_ref[0]**o+(0.219409873902)*x_ref[0]+(0.784253160678)*x_ref[1]**o+(0.0219597460707)*x_ref[1]
        ref[1,3,2,1]=(0.124778899794)*x_ref[0]**o+(0.785367842145)*x_ref[0]+(0.948111125908)*x_ref[1]**o+(-0.887427847089)*x_ref[1]
        ref[1,4,0,0]=(0.0911816002883)*x_ref[0]**o+(0.645410022441)*x_ref[0]+(0.720551531965)*x_ref[1]**o+(-0.388407683413)*x_ref[1]
        ref[1,4,0,1]=(-0.0416762613062)*x_ref[0]**o+(-0.548874829834)*x_ref[0]+(-0.377952555557)*x_ref[1]**o+(0.742795234361)*x_ref[1]
        ref[1,4,1,0]=(-0.896929017971)*x_ref[0]**o+(0.678343686516)*x_ref[0]+(0.711572763082)*x_ref[1]**o+(-0.43397331651)*x_ref[1]
        ref[1,4,1,1]=(-0.190804421566)*x_ref[0]**o+(-0.0848405124789)*x_ref[0]+(0.36246957905)*x_ref[1]**o+(0.963599086227)*x_ref[1]
        ref[1,4,2,0]=(-0.0210248567982)*x_ref[0]**o+(0.545977814581)*x_ref[0]+(0.858962864401)*x_ref[1]**o+(0.113792851997)*x_ref[1]
        ref[1,4,2,1]=(0.229641892622)*x_ref[0]**o+(-0.739561617198)*x_ref[0]+(-0.164372740617)*x_ref[1]**o+(0.537097868843)*x_ref[1]
        ref[2,0,0,0]=(-0.13321389343)*x_ref[0]**o+(-0.668403350503)*x_ref[0]+(-0.924879105983)*x_ref[1]**o+(-0.5161790717)*x_ref[1]
        ref[2,0,0,1]=(0.26582844685)*x_ref[0]**o+(0.89048660684)*x_ref[0]+(-0.377225981719)*x_ref[1]**o+(-0.303790710194)*x_ref[1]
        ref[2,0,1,0]=(0.0997470860283)*x_ref[0]**o+(0.0385427264217)*x_ref[0]+(-0.71387347273)*x_ref[1]**o+(-0.621376499192)*x_ref[1]
        ref[2,0,1,1]=(0.595544616326)*x_ref[0]**o+(-0.850529503975)*x_ref[0]+(0.602203957938)*x_ref[1]**o+(-0.329484843907)*x_ref[1]
        ref[2,0,2,0]=(-0.106031967029)*x_ref[0]**o+(-0.165122821622)*x_ref[0]+(0.647473815919)*x_ref[1]**o+(0.28942226062)*x_ref[1]
        ref[2,0,2,1]=(0.585955875472)*x_ref[0]**o+(0.922667710403)*x_ref[0]+(-0.0710905957841)*x_ref[1]**o+(-0.669138163697)*x_ref[1]
        ref[2,1,0,0]=(0.461941211081)*x_ref[0]**o+(0.723022466263)*x_ref[0]+(-0.118551893472)*x_ref[1]**o+(0.00445187353637)*x_ref[1]
        ref[2,1,0,1]=(-0.253168248108)*x_ref[0]**o+(0.476938842605)*x_ref[0]+(0.476641658034)*x_ref[1]**o+(-0.916886387595)*x_ref[1]
        ref[2,1,1,0]=(0.722029309503)*x_ref[0]**o+(0.0588848593425)*x_ref[0]+(0.84737370043)*x_ref[1]**o+(-0.674604415213)*x_ref[1]
        ref[2,1,1,1]=(-0.644736759298)*x_ref[0]**o+(-0.212847303893)*x_ref[0]+(0.25394587552)*x_ref[1]**o+(-0.366955175081)*x_ref[1]
        ref[2,1,2,0]=(-0.657395296801)*x_ref[0]**o+(-0.967570027599)*x_ref[0]+(0.180926628541)*x_ref[1]**o+(-0.940599897942)*x_ref[1]
        ref[2,1,2,1]=(0.849897328801)*x_ref[0]**o+(-0.299199474051)*x_ref[0]+(0.427221207827)*x_ref[1]**o+(0.212027985441)*x_ref[1]
        ref[2,2,0,0]=(-0.561573507381)*x_ref[0]**o+(-0.446925224908)*x_ref[0]+(0.132091224948)*x_ref[1]**o+(-0.673375661849)*x_ref[1]
        ref[2,2,0,1]=(-0.386657291574)*x_ref[0]**o+(0.356784637112)*x_ref[0]+(0.644292006233)*x_ref[1]**o+(0.922937167842)*x_ref[1]
        ref[2,2,1,0]=(0.484446757285)*x_ref[0]**o+(-0.822469870455)*x_ref[0]+(-0.89605890123)*x_ref[1]**o+(-0.895891500031)*x_ref[1]
        ref[2,2,1,1]=(-0.0930740520192)*x_ref[0]**o+(0.953356473678)*x_ref[0]+(-0.714418458806)*x_ref[1]**o+(-0.100584950946)*x_ref[1]
        ref[2,2,2,0]=(0.945572242667)*x_ref[0]**o+(-0.259280849605)*x_ref[0]+(-0.925459846341)*x_ref[1]**o+(0.906002758873)*x_ref[1]
        ref[2,2,2,1]=(-0.638993332975)*x_ref[0]**o+(-0.668594641703)*x_ref[0]+(0.012666231369)*x_ref[1]**o+(0.89732685856)*x_ref[1]
        ref[2,3,0,0]=(-0.885061116144)*x_ref[0]**o+(-0.576097536483)*x_ref[0]+(-0.43166075788)*x_ref[1]**o+(0.538821620917)*x_ref[1]
        ref[2,3,0,1]=(-0.0743638381797)*x_ref[0]**o+(0.251037500035)*x_ref[0]+(-0.905989928076)*x_ref[1]**o+(0.243771000867)*x_ref[1]
        ref[2,3,1,0]=(-0.016045622351)*x_ref[0]**o+(-0.632944072174)*x_ref[0]+(-0.296175918793)*x_ref[1]**o+(-0.00120123641365)*x_ref[1]
        ref[2,3,1,1]=(-0.896354921812)*x_ref[0]**o+(-0.331348989027)*x_ref[0]+(-0.450865285201)*x_ref[1]**o+(-0.243268582009)*x_ref[1]
        ref[2,3,2,0]=(0.609332927431)*x_ref[0]**o+(-0.0815856802115)*x_ref[0]+(0.201924792273)*x_ref[1]**o+(-0.890424451348)*x_ref[1]
        ref[2,3,2,1]=(0.395090137264)*x_ref[0]**o+(0.294230442573)*x_ref[0]+(-0.345371510364)*x_ref[1]**o+(0.181229663241)*x_ref[1]
        ref[2,4,0,0]=(-0.823684509034)*x_ref[0]**o+(0.431019301942)*x_ref[0]+(0.554707015054)*x_ref[1]**o+(-0.846322903003)*x_ref[1]
        ref[2,4,0,1]=(0.862984476658)*x_ref[0]**o+(0.20512500102)*x_ref[0]+(0.0247756159287)*x_ref[1]**o+(-0.374795100946)*x_ref[1]
        ref[2,4,1,0]=(-0.0849914269517)*x_ref[0]**o+(0.754908200339)*x_ref[0]+(0.593363754807)*x_ref[1]**o+(0.124923961126)*x_ref[1]
        ref[2,4,1,1]=(-0.691440264234)*x_ref[0]**o+(0.825739298015)*x_ref[0]+(0.426452231478)*x_ref[1]**o+(0.0995610721341)*x_ref[1]
        ref[2,4,2,0]=(-0.175322750567)*x_ref[0]**o+(0.419750865569)*x_ref[0]+(0.208448987712)*x_ref[1]**o+(0.150584798952)*x_ref[1]
        ref[2,4,2,1]=(0.931680152344)*x_ref[0]**o+(0.390947206307)*x_ref[0]+(0.388437348378)*x_ref[1]**o+(-0.808413079032)*x_ref[1]
        ref[3,0,0,0]=(0.58855431402)*x_ref[0]**o+(0.383865913364)*x_ref[0]+(-0.691445323219)*x_ref[1]**o+(-0.443370658523)*x_ref[1]
        ref[3,0,0,1]=(-0.657371481478)*x_ref[0]**o+(0.746891082673)*x_ref[0]+(-0.829229947773)*x_ref[1]**o+(0.761915670404)*x_ref[1]
        ref[3,0,1,0]=(-0.431502932714)*x_ref[0]**o+(-0.961472285716)*x_ref[0]+(0.0423384388031)*x_ref[1]**o+(0.339373458118)*x_ref[1]
        ref[3,0,1,1]=(-0.272831163259)*x_ref[0]**o+(-0.533438904494)*x_ref[0]+(-0.595625384875)*x_ref[1]**o+(0.0366857022323)*x_ref[1]
        ref[3,0,2,0]=(-0.319684662306)*x_ref[0]**o+(0.672185828912)*x_ref[0]+(0.483126548994)*x_ref[1]**o+(0.745872629604)*x_ref[1]
        ref[3,0,2,1]=(0.437895277404)*x_ref[0]**o+(-0.337209416963)*x_ref[0]+(0.101080522788)*x_ref[1]**o+(0.892022488644)*x_ref[1]
        ref[3,1,0,0]=(-0.582325078455)*x_ref[0]**o+(-0.854689347157)*x_ref[0]+(0.0308638312259)*x_ref[1]**o+(0.633382365704)*x_ref[1]
        ref[3,1,0,1]=(0.319735178376)*x_ref[0]**o+(-0.782764593952)*x_ref[0]+(0.082230132244)*x_ref[1]**o+(-0.946010937852)*x_ref[1]
        ref[3,1,1,0]=(-0.698490410666)*x_ref[0]**o+(-0.801358802689)*x_ref[0]+(0.20695648235)*x_ref[1]**o+(0.75257523471)*x_ref[1]
        ref[3,1,1,1]=(0.66321696527)*x_ref[0]**o+(0.704572918868)*x_ref[0]+(0.528208879002)*x_ref[1]**o+(-0.408378325513)*x_ref[1]
        ref[3,1,2,0]=(-0.925177198417)*x_ref[0]**o+(-0.152427811549)*x_ref[0]+(0.341781993583)*x_ref[1]**o+(-0.589976456377)*x_ref[1]
        ref[3,1,2,1]=(0.206706628353)*x_ref[0]**o+(0.673777173329)*x_ref[0]+(-0.994138314254)*x_ref[1]**o+(-0.843696565863)*x_ref[1]
        ref[3,2,0,0]=(-0.588881067333)*x_ref[0]**o+(-0.537572388341)*x_ref[0]+(-0.206570371675)*x_ref[1]**o+(-0.127468554294)*x_ref[1]
        ref[3,2,0,1]=(0.498638284096)*x_ref[0]**o+(-0.488278831086)*x_ref[0]+(-0.28320874594)*x_ref[1]**o+(-0.638241576615)*x_ref[1]
        ref[3,2,1,0]=(-0.445780790822)*x_ref[0]**o+(-0.347493953089)*x_ref[0]+(-0.558010086232)*x_ref[1]**o+(-0.0287339307689)*x_ref[1]
        ref[3,2,1,1]=(-0.361727811272)*x_ref[0]**o+(-0.559565378007)*x_ref[0]+(-0.211685558972)*x_ref[1]**o+(0.330088396811)*x_ref[1]
        ref[3,2,2,0]=(-0.42600117831)*x_ref[0]**o+(-0.488997111988)*x_ref[0]+(0.127421278051)*x_ref[1]**o+(-0.48233167806)*x_ref[1]
        ref[3,2,2,1]=(0.120468829686)*x_ref[0]**o+(0.0675603906118)*x_ref[0]+(-0.60674361864)*x_ref[1]**o+(0.639115708258)*x_ref[1]
        ref[3,3,0,0]=(0.369343033015)*x_ref[0]**o+(0.707735743593)*x_ref[0]+(-0.426275591836)*x_ref[1]**o+(-0.733377765625)*x_ref[1]
        ref[3,3,0,1]=(-0.931245338156)*x_ref[0]**o+(-0.465861794185)*x_ref[0]+(-0.650664500687)*x_ref[1]**o+(0.103077662402)*x_ref[1]
        ref[3,3,1,0]=(0.731834744512)*x_ref[0]**o+(0.267840571071)*x_ref[0]+(0.952106457495)*x_ref[1]**o+(-0.943845364366)*x_ref[1]
        ref[3,3,1,1]=(-0.21692720916)*x_ref[0]**o+(0.580841374997)*x_ref[0]+(-0.412129341223)*x_ref[1]**o+(0.40414445304)*x_ref[1]
        ref[3,3,2,0]=(-0.447758598726)*x_ref[0]**o+(0.0616926352912)*x_ref[0]+(-0.824624226122)*x_ref[1]**o+(0.19099833104)*x_ref[1]
        ref[3,3,2,1]=(-0.615519138165)*x_ref[0]**o+(0.233926822164)*x_ref[0]+(-0.969014060236)*x_ref[1]**o+(0.835103627893)*x_ref[1]
        ref[3,4,0,0]=(0.550778983301)*x_ref[0]**o+(0.0198253530882)*x_ref[0]+(-0.39383849187)*x_ref[1]**o+(0.367721826304)*x_ref[1]
        ref[3,4,0,1]=(0.444877923149)*x_ref[0]**o+(-0.899159835223)*x_ref[0]+(0.923392540452)*x_ref[1]**o+(0.512905141718)*x_ref[1]
        ref[3,4,1,0]=(-0.563807675145)*x_ref[0]**o+(0.994405383899)*x_ref[0]+(-0.777352948218)*x_ref[1]**o+(0.865994906227)*x_ref[1]
        ref[3,4,1,1]=(-0.593800452139)*x_ref[0]**o+(0.613778803237)*x_ref[0]+(0.582200979718)*x_ref[1]**o+(0.448026243271)*x_ref[1]
        ref[3,4,2,0]=(-0.900331125416)*x_ref[0]**o+(0.234613965424)*x_ref[0]+(0.510400273805)*x_ref[1]**o+(0.0836773641661)*x_ref[1]
        ref[3,4,2,1]=(0.710999097307)*x_ref[0]**o+(0.886603988265)*x_ref[0]+(-0.405644887679)*x_ref[1]**o+(0.362257258809)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.310272924207)*x[0]**o+(-0.664497806351)*x[0]+(0.741942132609)*x[1]**o+(0.636051298779)*x[1]+(0.76627159146)*x[2]**o+(0.492960732588)*x[2]
        arg[0,0,0,1]=(-0.513740801773)*x[0]**o+(-0.633330750909)*x[0]+(0.177934618983)*x[1]**o+(-0.323745723166)*x[1]+(0.694167180904)*x[2]**o+(0.441239251435)*x[2]
        arg[0,0,1,0]=(0.417477695848)*x[0]**o+(-0.923415976002)*x[0]+(0.0765555276662)*x[1]**o+(-0.277290890834)*x[1]+(0.0177191326614)*x[2]**o+(-0.0842520100387)*x[2]
        arg[0,0,1,1]=(-0.761916249894)*x[0]**o+(0.169673139552)*x[0]+(0.0306290828795)*x[1]**o+(0.842234592613)*x[1]+(-0.350342692669)*x[2]**o+(0.321292270658)*x[2]
        arg[0,0,2,0]=(-0.707494137624)*x[0]**o+(0.346873490395)*x[0]+(-0.68281925138)*x[1]**o+(-0.830538083274)*x[1]+(0.329372141613)*x[2]**o+(-0.121366157163)*x[2]
        arg[0,0,2,1]=(-0.708018186507)*x[0]**o+(-0.778051239314)*x[0]+(-0.43106291563)*x[1]**o+(-0.0115914251143)*x[1]+(0.487334215892)*x[2]**o+(-0.289232284658)*x[2]
        arg[0,1,0,0]=(0.530614704199)*x[0]**o+(0.518622028317)*x[0]+(-0.0143129589551)*x[1]**o+(0.808469352257)*x[1]+(-0.156787373225)*x[2]**o+(0.668887813143)*x[2]
        arg[0,1,0,1]=(0.953223233706)*x[0]**o+(-0.164323838353)*x[0]+(0.336717320858)*x[1]**o+(-0.520230488172)*x[1]+(0.689960738002)*x[2]**o+(0.487630947578)*x[2]
        arg[0,1,1,0]=(-0.418095373429)*x[0]**o+(-0.941817844939)*x[0]+(0.88105115607)*x[1]**o+(-0.820502501439)*x[1]+(-0.693858267667)*x[2]**o+(0.530567713127)*x[2]
        arg[0,1,1,1]=(-0.716861320218)*x[0]**o+(-0.736619859467)*x[0]+(-0.885478605601)*x[1]**o+(-0.129375578967)*x[1]+(-0.0181433907318)*x[2]**o+(0.373255537959)*x[2]
        arg[0,1,2,0]=(-0.174742583167)*x[0]**o+(-0.0887967480112)*x[0]+(-0.291168698662)*x[1]**o+(-0.188608506769)*x[1]+(0.22479613454)*x[2]**o+(0.34173928582)*x[2]
        arg[0,1,2,1]=(0.117208929504)*x[0]**o+(-0.420969282079)*x[0]+(0.585298109836)*x[1]**o+(0.771355162745)*x[1]+(-0.51374493747)*x[2]**o+(0.785016548855)*x[2]
        arg[0,2,0,0]=(0.737394887886)*x[0]**o+(0.966080271342)*x[0]+(-0.373178384492)*x[1]**o+(-0.845505386142)*x[1]+(-0.274654519371)*x[2]**o+(0.709356557887)*x[2]
        arg[0,2,0,1]=(0.0365864090636)*x[0]**o+(-0.393382479787)*x[0]+(0.018273360977)*x[1]**o+(-0.509201909328)*x[1]+(-0.777852392534)*x[2]**o+(0.0417280247507)*x[2]
        arg[0,2,1,0]=(0.637829302751)*x[0]**o+(-0.767824790186)*x[0]+(0.843382009511)*x[1]**o+(-0.599633234403)*x[1]+(0.14283138941)*x[2]**o+(-0.354404633991)*x[2]
        arg[0,2,1,1]=(-0.465258613599)*x[0]**o+(0.09341154383)*x[0]+(-0.629114782988)*x[1]**o+(0.91608655383)*x[1]+(-0.692901135848)*x[2]**o+(-0.707225310667)*x[2]
        arg[0,2,2,0]=(0.9935958759)*x[0]**o+(-0.385008742466)*x[0]+(0.153885356854)*x[1]**o+(0.146270680656)*x[1]+(0.305838410343)*x[2]**o+(0.141193552717)*x[2]
        arg[0,2,2,1]=(-0.487538023894)*x[0]**o+(-0.814283442647)*x[0]+(0.166008520199)*x[1]**o+(0.313655774416)*x[1]+(0.282308782973)*x[2]**o+(0.309176062484)*x[2]
        arg[0,3,0,0]=(0.00203507798913)*x[0]**o+(-0.19902542604)*x[0]+(0.88110918624)*x[1]**o+(0.223549219288)*x[1]+(0.135758952369)*x[2]**o+(0.59116783639)*x[2]
        arg[0,3,0,1]=(-0.185063700767)*x[0]**o+(0.833862155173)*x[0]+(-0.0939117509645)*x[1]**o+(0.288003692231)*x[1]+(0.580196157839)*x[2]**o+(0.299320177754)*x[2]
        arg[0,3,1,0]=(-0.039100718826)*x[0]**o+(0.423179869395)*x[0]+(-0.013669578975)*x[1]**o+(-0.355053343085)*x[1]+(0.89285158975)*x[2]**o+(0.392984367877)*x[2]
        arg[0,3,1,1]=(0.0988989416429)*x[0]**o+(-0.446243235174)*x[0]+(-0.691253313617)*x[1]**o+(0.338823693086)*x[1]+(0.707054126334)*x[2]**o+(0.601121388278)*x[2]
        arg[0,3,2,0]=(-0.191366770484)*x[0]**o+(0.961791275903)*x[0]+(-0.103547475035)*x[1]**o+(-0.745236510079)*x[1]+(-0.157506981712)*x[2]**o+(0.986229506787)*x[2]
        arg[0,3,2,1]=(-0.886091883647)*x[0]**o+(0.6979804248)*x[0]+(-0.366575419428)*x[1]**o+(-0.958601153259)*x[1]+(-0.0119724511494)*x[2]**o+(-0.76101132076)*x[2]
        arg[0,4,0,0]=(0.923290554434)*x[0]**o+(0.411359492198)*x[0]+(-0.977999562875)*x[1]**o+(0.352085161527)*x[1]+(0.955745235349)*x[2]**o+(-0.978885609257)*x[2]
        arg[0,4,0,1]=(0.226973479105)*x[0]**o+(-0.423881897251)*x[0]+(0.889101969481)*x[1]**o+(-0.994236389028)*x[1]+(0.0159169033361)*x[2]**o+(-0.559935154203)*x[2]
        arg[0,4,1,0]=(-0.225427291797)*x[0]**o+(-0.557334621848)*x[0]+(-0.486252586142)*x[1]**o+(0.0497120905216)*x[1]+(0.84165736691)*x[2]**o+(-0.185010156797)*x[2]
        arg[0,4,1,1]=(0.762694602609)*x[0]**o+(-0.447367170727)*x[0]+(-0.61103181818)*x[1]**o+(0.689761057656)*x[1]+(-0.631855120122)*x[2]**o+(0.33274655602)*x[2]
        arg[0,4,2,0]=(-0.198884868914)*x[0]**o+(0.716371959278)*x[0]+(0.260871768919)*x[1]**o+(0.622241740082)*x[1]+(0.774103524035)*x[2]**o+(0.186132703676)*x[2]
        arg[0,4,2,1]=(-0.309450645317)*x[0]**o+(0.204754571336)*x[0]+(0.483402810171)*x[1]**o+(0.404620744188)*x[1]+(-0.699137246959)*x[2]**o+(0.467428706196)*x[2]
        arg[1,0,0,0]=(0.642289506448)*x[0]**o+(0.383990304394)*x[0]+(0.556940973318)*x[1]**o+(0.135219729041)*x[1]+(0.577195836358)*x[2]**o+(-0.338924419674)*x[2]
        arg[1,0,0,1]=(-0.920873911295)*x[0]**o+(-0.330610192304)*x[0]+(-0.117226803856)*x[1]**o+(0.172482867684)*x[1]+(0.909140902709)*x[2]**o+(-0.590494167913)*x[2]
        arg[1,0,1,0]=(0.210161538861)*x[0]**o+(-0.83860241761)*x[0]+(0.0257180382661)*x[1]**o+(0.877633854348)*x[1]+(0.26045311754)*x[2]**o+(-0.971349902714)*x[2]
        arg[1,0,1,1]=(0.70736843125)*x[0]**o+(-0.765121985431)*x[0]+(-0.196525225943)*x[1]**o+(-0.175928729346)*x[1]+(0.945523028748)*x[2]**o+(0.877830322823)*x[2]
        arg[1,0,2,0]=(-0.963513016065)*x[0]**o+(0.0820537856351)*x[0]+(0.0122070196118)*x[1]**o+(-0.154147449592)*x[1]+(0.323844633268)*x[2]**o+(-0.124897698214)*x[2]
        arg[1,0,2,1]=(0.656267566405)*x[0]**o+(0.0174786620768)*x[0]+(0.763902435618)*x[1]**o+(-0.986021506043)*x[1]+(0.632625555994)*x[2]**o+(-0.110657743523)*x[2]
        arg[1,1,0,0]=(0.00354144688732)*x[0]**o+(-0.813515808171)*x[0]+(-0.100717717446)*x[1]**o+(0.0543708163629)*x[1]+(0.785927880762)*x[2]**o+(0.0109166652227)*x[2]
        arg[1,1,0,1]=(-0.239705688631)*x[0]**o+(0.402915642509)*x[0]+(0.463371068163)*x[1]**o+(-0.507302462928)*x[1]+(-0.0789844128295)*x[2]**o+(0.669144891499)*x[2]
        arg[1,1,1,0]=(-0.491891367001)*x[0]**o+(-0.524377636928)*x[0]+(0.903091863917)*x[1]**o+(-0.554368807183)*x[1]+(0.332899195339)*x[2]**o+(-0.645216147049)*x[2]
        arg[1,1,1,1]=(-0.915843941243)*x[0]**o+(-0.233175654748)*x[0]+(0.467730425037)*x[1]**o+(-0.765957936683)*x[1]+(0.726678877276)*x[2]**o+(0.770533947942)*x[2]
        arg[1,1,2,0]=(0.678071852032)*x[0]**o+(0.266537844237)*x[0]+(0.324335300151)*x[1]**o+(-0.286713971373)*x[1]+(-0.702384414014)*x[2]**o+(0.887198276519)*x[2]
        arg[1,1,2,1]=(0.0466444959648)*x[0]**o+(0.689162912677)*x[0]+(-0.35033923676)*x[1]**o+(-0.0536380766876)*x[1]+(-0.511833442517)*x[2]**o+(0.125566299829)*x[2]
        arg[1,2,0,0]=(-0.803170431622)*x[0]**o+(0.476211169752)*x[0]+(-0.170510267953)*x[1]**o+(-0.765945735317)*x[1]+(-0.169966329291)*x[2]**o+(-0.988525918724)*x[2]
        arg[1,2,0,1]=(0.289618751212)*x[0]**o+(0.294379040976)*x[0]+(-0.922478416838)*x[1]**o+(0.210546495537)*x[1]+(-0.87266381496)*x[2]**o+(-0.865571669577)*x[2]
        arg[1,2,1,0]=(-0.292811054005)*x[0]**o+(0.651280833746)*x[0]+(-0.625004203736)*x[1]**o+(0.372078169172)*x[1]+(-0.0539099055169)*x[2]**o+(0.0911869270008)*x[2]
        arg[1,2,1,1]=(0.496217698294)*x[0]**o+(-0.986021787133)*x[0]+(0.246417059532)*x[1]**o+(0.610226802401)*x[1]+(0.991534665298)*x[2]**o+(-0.507441185982)*x[2]
        arg[1,2,2,0]=(-0.839902565947)*x[0]**o+(-0.279607147278)*x[0]+(0.278212307963)*x[1]**o+(0.609793640036)*x[1]+(0.940471913774)*x[2]**o+(0.21860828705)*x[2]
        arg[1,2,2,1]=(0.789578112278)*x[0]**o+(-0.0226697818608)*x[0]+(0.216321340291)*x[1]**o+(-0.701818463454)*x[1]+(-0.166689030646)*x[2]**o+(0.966131542292)*x[2]
        arg[1,3,0,0]=(0.149337042551)*x[0]**o+(-0.785050579674)*x[0]+(0.530219924721)*x[1]**o+(-0.971927239398)*x[1]+(0.950917908603)*x[2]**o+(0.132668077447)*x[2]
        arg[1,3,0,1]=(0.983081559724)*x[0]**o+(-0.358694498455)*x[0]+(0.181080099281)*x[1]**o+(-0.699756291643)*x[1]+(-0.744335953304)*x[2]**o+(0.00262043221382)*x[2]
        arg[1,3,1,0]=(0.31776032045)*x[0]**o+(-0.0726184220029)*x[0]+(-0.759883735971)*x[1]**o+(-0.6389710873)*x[1]+(-0.607858499502)*x[2]**o+(-0.397169895784)*x[2]
        arg[1,3,1,1]=(-0.950468389728)*x[0]**o+(0.831121915521)*x[0]+(0.552857011083)*x[1]**o+(0.848599672594)*x[1]+(0.432183564231)*x[2]**o+(0.57537721362)*x[2]
        arg[1,3,2,0]=(0.736080143601)*x[0]**o+(-0.804140307747)*x[0]+(-0.163864867884)*x[1]**o+(0.950874042315)*x[1]+(-0.99676139412)*x[2]**o+(0.332320294246)*x[2]
        arg[1,3,2,1]=(0.119082582041)*x[0]**o+(-0.969740778852)*x[0]+(0.828432583532)*x[1]**o+(0.190324050395)*x[1]+(-0.42639142674)*x[2]**o+(0.315519578192)*x[2]
        arg[1,4,0,0]=(-0.798465141775)*x[0]**o+(-0.717282977908)*x[0]+(0.935655148713)*x[1]**o+(-0.655752964484)*x[1]+(0.159990149099)*x[2]**o+(-0.537115062687)*x[2]
        arg[1,4,0,1]=(0.445151775007)*x[0]**o+(-0.867956302051)*x[0]+(-0.809856879825)*x[1]**o+(-0.326897750981)*x[1]+(0.756352311987)*x[2]**o+(0.577585513459)*x[2]
        arg[1,4,1,0]=(0.294243347591)*x[0]**o+(0.774741656931)*x[0]+(-0.0799985737984)*x[1]**o+(0.179789107412)*x[1]+(-0.612198025189)*x[2]**o+(-0.07086009611)*x[2]
        arg[1,4,1,1]=(0.0692586287045)*x[0]**o+(-0.671956300122)*x[0]+(-0.948667104683)*x[1]**o+(0.68656246801)*x[1]+(-0.352975297696)*x[2]**o+(0.277952113716)*x[2]
        arg[1,4,2,0]=(-0.491316095517)*x[0]**o+(0.725899087239)*x[0]+(0.634113963431)*x[1]**o+(0.235694393453)*x[1]+(-0.15600720713)*x[2]**o+(0.954814976539)*x[2]
        arg[1,4,2,1]=(0.120206996339)*x[0]**o+(-0.86151411823)*x[0]+(-0.137093650097)*x[1]**o+(0.633402978246)*x[1]+(-0.42515364728)*x[2]**o+(-0.987634210617)*x[2]
        arg[2,0,0,0]=(0.539463971222)*x[0]**o+(-0.03435650579)*x[0]+(-0.884395295544)*x[1]**o+(0.215454879022)*x[1]+(0.842377328246)*x[2]**o+(0.370920136871)*x[2]
        arg[2,0,0,1]=(-0.411505463039)*x[0]**o+(0.446494039439)*x[0]+(0.0313771969556)*x[1]**o+(0.602893223061)*x[1]+(-0.30383490021)*x[2]**o+(-0.788028756758)*x[2]
        arg[2,0,1,0]=(0.489009615999)*x[0]**o+(-0.95508248016)*x[0]+(-0.21489475022)*x[1]**o+(-0.857263478431)*x[1]+(-0.123703409045)*x[2]**o+(0.663389562431)*x[2]
        arg[2,0,1,1]=(-0.0224763310929)*x[0]**o+(-0.315886249237)*x[0]+(-0.318201849695)*x[1]**o+(0.423732682462)*x[1]+(0.396065636381)*x[2]**o+(0.0816717878865)*x[2]
        arg[2,0,2,0]=(-0.191432618224)*x[0]**o+(0.688389390795)*x[0]+(-0.382948112159)*x[1]**o+(0.562905173818)*x[1]+(-0.118396714029)*x[2]**o+(0.234478642577)*x[2]
        arg[2,0,2,1]=(-0.348008173171)*x[0]**o+(0.191994166636)*x[0]+(0.450556057147)*x[1]**o+(-0.710821693343)*x[1]+(-0.245119333857)*x[2]**o+(0.262869525239)*x[2]
        arg[2,1,0,0]=(-0.160444506203)*x[0]**o+(-0.816142837432)*x[0]+(-0.619722480492)*x[1]**o+(-0.989077039675)*x[1]+(-0.0282349103977)*x[2]**o+(-0.629849819691)*x[2]
        arg[2,1,0,1]=(0.379513765297)*x[0]**o+(0.934652105586)*x[0]+(-0.632636427868)*x[1]**o+(-0.0492550496553)*x[1]+(0.137733572307)*x[2]**o+(0.779782637942)*x[2]
        arg[2,1,1,0]=(0.469499320417)*x[0]**o+(0.570880924125)*x[0]+(-0.0329833251419)*x[1]**o+(-0.0388751034604)*x[1]+(-0.878520094389)*x[2]**o+(0.992656649048)*x[2]
        arg[2,1,1,1]=(0.592005181995)*x[0]**o+(-0.553980493526)*x[0]+(0.359837665019)*x[1]**o+(0.811107941179)*x[1]+(-0.441929650912)*x[2]**o+(-0.768618378636)*x[2]
        arg[2,1,2,0]=(0.741376258816)*x[0]**o+(-0.238909756717)*x[0]+(0.204631082449)*x[1]**o+(0.970501662949)*x[1]+(-0.615638092239)*x[2]**o+(-0.575948750844)*x[2]
        arg[2,1,2,1]=(0.774050933112)*x[0]**o+(-0.872234076721)*x[0]+(-0.0936189240061)*x[1]**o+(-0.773608756961)*x[1]+(-0.119171907808)*x[2]**o+(-0.256124796845)*x[2]
        arg[2,2,0,0]=(0.459481754425)*x[0]**o+(0.478364929493)*x[0]+(0.460199733287)*x[1]**o+(0.66880294659)*x[1]+(-0.859525058389)*x[2]**o+(0.77960284414)*x[2]
        arg[2,2,0,1]=(0.690557835207)*x[0]**o+(-0.812305038592)*x[0]+(0.837924124791)*x[1]**o+(-0.785979206364)*x[1]+(-0.295553949819)*x[2]**o+(0.575096295897)*x[2]
        arg[2,2,1,0]=(0.443193683205)*x[0]**o+(0.0355033222253)*x[0]+(0.497641002766)*x[1]**o+(-0.636955142451)*x[1]+(-0.265324357454)*x[2]**o+(0.734148569078)*x[2]
        arg[2,2,1,1]=(-0.088707242599)*x[0]**o+(0.513403518669)*x[0]+(0.337741737651)*x[1]**o+(-0.174843924478)*x[1]+(0.687291255986)*x[2]**o+(0.473850502258)*x[2]
        arg[2,2,2,0]=(0.5305624642)*x[0]**o+(-0.972997835953)*x[0]+(0.00197107287888)*x[1]**o+(-0.75490872634)*x[1]+(-0.0569624569435)*x[2]**o+(-0.0780532868631)*x[2]
        arg[2,2,2,1]=(-0.744584471866)*x[0]**o+(0.915494909272)*x[0]+(0.239235050038)*x[1]**o+(-0.192466753187)*x[1]+(0.933556550824)*x[2]**o+(-0.786420251487)*x[2]
        arg[2,3,0,0]=(0.472114530627)*x[0]**o+(0.0693211807557)*x[0]+(-0.274273425713)*x[1]**o+(0.136242554894)*x[1]+(-0.329240675785)*x[2]**o+(0.0212539247343)*x[2]
        arg[2,3,0,1]=(-0.780426108099)*x[0]**o+(-0.632382551025)*x[0]+(0.554926949239)*x[1]**o+(0.264204009059)*x[1]+(-0.868007525824)*x[2]**o+(-0.937631120201)*x[2]
        arg[2,3,1,0]=(-0.826692080947)*x[0]**o+(0.915241858538)*x[0]+(-0.49986160859)*x[1]**o+(0.921686464658)*x[1]+(0.71996666304)*x[2]**o+(0.297609501096)*x[2]
        arg[2,3,1,1]=(0.991896994614)*x[0]**o+(-0.543310008103)*x[0]+(0.0791899631111)*x[1]**o+(-0.914088200145)*x[1]+(-0.496337624951)*x[2]**o+(0.708950898614)*x[2]
        arg[2,3,2,0]=(-0.235116966432)*x[0]**o+(-0.492848730207)*x[0]+(0.428728646789)*x[1]**o+(-0.410747634691)*x[1]+(0.705059815458)*x[2]**o+(-0.185604633831)*x[2]
        arg[2,3,2,1]=(0.838829037117)*x[0]**o+(0.670561147563)*x[0]+(0.621407965075)*x[1]**o+(-0.10624946305)*x[1]+(0.893703402687)*x[2]**o+(0.985504995821)*x[2]
        arg[2,4,0,0]=(0.635072504682)*x[0]**o+(-0.681962427555)*x[0]+(-0.00774518011081)*x[1]**o+(-0.264996228668)*x[1]+(0.690668403198)*x[2]**o+(0.883573761062)*x[2]
        arg[2,4,0,1]=(0.545632268651)*x[0]**o+(0.439668519394)*x[0]+(-0.482479808491)*x[1]**o+(0.752028962391)*x[1]+(0.0596773623939)*x[2]**o+(-0.369987181127)*x[2]
        arg[2,4,1,0]=(0.659005322796)*x[0]**o+(0.503940833428)*x[0]+(0.643633321586)*x[1]**o+(0.927082323252)*x[1]+(0.354072146184)*x[2]**o+(-0.638737560608)*x[2]
        arg[2,4,1,1]=(-0.443548860913)*x[0]**o+(0.707607048458)*x[0]+(-0.955225746402)*x[1]**o+(0.515995563585)*x[1]+(-0.288406637165)*x[2]**o+(-0.760864937318)*x[2]
        arg[2,4,2,0]=(0.898146580355)*x[0]**o+(-0.590650787106)*x[0]+(0.036095789753)*x[1]**o+(0.326040260581)*x[1]+(0.680069361306)*x[2]**o+(0.333839099341)*x[2]
        arg[2,4,2,1]=(-0.151364260271)*x[0]**o+(-0.675198183832)*x[0]+(-0.183626298978)*x[1]**o+(-0.182565217146)*x[1]+(0.0345632129814)*x[2]**o+(-0.581534100763)*x[2]
        arg[3,0,0,0]=(-0.942351703246)*x[0]**o+(-0.699567649453)*x[0]+(-0.832068665702)*x[1]**o+(0.163439416504)*x[1]+(0.742500339519)*x[2]**o+(-0.492882887489)*x[2]
        arg[3,0,0,1]=(0.990078296484)*x[0]**o+(-0.999837336587)*x[0]+(0.197589456459)*x[1]**o+(0.431403059019)*x[1]+(0.859864015871)*x[2]**o+(-0.267474829282)*x[2]
        arg[3,0,1,0]=(-0.776773660723)*x[0]**o+(-0.189048370486)*x[0]+(0.926953856802)*x[1]**o+(-0.72314744652)*x[1]+(0.330064363809)*x[2]**o+(0.188533334359)*x[2]
        arg[3,0,1,1]=(-0.854830058762)*x[0]**o+(0.832073883276)*x[0]+(-0.425691165124)*x[1]**o+(0.162351500316)*x[1]+(0.370604864512)*x[2]**o+(-0.581907071484)*x[2]
        arg[3,0,2,0]=(-0.316716909548)*x[0]**o+(0.335417284383)*x[0]+(-0.783061160352)*x[1]**o+(-0.253972519239)*x[1]+(-0.826432933721)*x[2]**o+(0.267368378363)*x[2]
        arg[3,0,2,1]=(-0.765899705462)*x[0]**o+(-0.0993875861476)*x[0]+(-0.211785004352)*x[1]**o+(-0.635330151497)*x[1]+(0.414575412599)*x[2]**o+(0.586882274623)*x[2]
        arg[3,1,0,0]=(0.987776377237)*x[0]**o+(0.796660669191)*x[0]+(0.580724891862)*x[1]**o+(-0.24149253995)*x[1]+(-0.633421016909)*x[2]**o+(-0.220110913711)*x[2]
        arg[3,1,0,1]=(-0.182240918081)*x[0]**o+(-0.290526050572)*x[0]+(0.00283230959976)*x[1]**o+(-0.933282166805)*x[1]+(-0.857893235883)*x[2]**o+(0.551414074127)*x[2]
        arg[3,1,1,0]=(-0.854435997816)*x[0]**o+(0.706455144263)*x[0]+(0.406237965254)*x[1]**o+(0.192071142415)*x[1]+(0.357966796834)*x[2]**o+(0.887011087987)*x[2]
        arg[3,1,1,1]=(0.832013396996)*x[0]**o+(0.696133072321)*x[0]+(0.682726968712)*x[1]**o+(0.613316240514)*x[1]+(0.12600339887)*x[2]**o+(-0.620088231382)*x[2]
        arg[3,1,2,0]=(-0.178681615221)*x[0]**o+(-0.952835986168)*x[0]+(0.299692693746)*x[1]**o+(0.892451908416)*x[1]+(0.202403273634)*x[2]**o+(0.677212897943)*x[2]
        arg[3,1,2,1]=(0.0592594053342)*x[0]**o+(0.56391153959)*x[0]+(-0.874866627667)*x[1]**o+(0.432102900769)*x[1]+(-0.49073126285)*x[2]**o+(-0.010495860028)*x[2]
        arg[3,2,0,0]=(0.548006445814)*x[0]**o+(0.667313392202)*x[0]+(0.397805355499)*x[1]**o+(-0.788100267808)*x[1]+(0.869234055258)*x[2]**o+(-0.830241520401)*x[2]
        arg[3,2,0,1]=(0.72586139216)*x[0]**o+(-0.931878957878)*x[0]+(-0.94119555858)*x[1]**o+(-0.133544806169)*x[1]+(0.754263323382)*x[2]**o+(-0.446604740157)*x[2]
        arg[3,2,1,0]=(-0.28135830707)*x[0]**o+(-0.478744058518)*x[0]+(0.88219033731)*x[1]**o+(0.460587286945)*x[1]+(-0.849692146096)*x[2]**o+(0.0927701345107)*x[2]
        arg[3,2,1,1]=(-0.681859641611)*x[0]**o+(-0.670164293088)*x[0]+(0.796366374435)*x[1]**o+(0.837230317054)*x[1]+(-0.413283915904)*x[2]**o+(0.779624227187)*x[2]
        arg[3,2,2,0]=(-0.803242771255)*x[0]**o+(-0.0687616271776)*x[0]+(-0.948856027999)*x[1]**o+(0.762568790931)*x[1]+(0.937588292604)*x[2]**o+(-0.0224232611283)*x[2]
        arg[3,2,2,1]=(-0.215752944937)*x[0]**o+(-0.807114296075)*x[0]+(0.970625375964)*x[1]**o+(0.68578124692)*x[1]+(-0.165007047419)*x[2]**o+(0.942718439981)*x[2]
        arg[3,3,0,0]=(0.152001104716)*x[0]**o+(-0.657242342357)*x[0]+(0.559608661705)*x[1]**o+(0.0764403273251)*x[1]+(0.633414754695)*x[2]**o+(0.430777868344)*x[2]
        arg[3,3,0,1]=(-0.898109497759)*x[0]**o+(0.736574249406)*x[0]+(-0.569584184574)*x[1]**o+(-0.145501395506)*x[1]+(-0.412424116475)*x[2]**o+(-0.416350923096)*x[2]
        arg[3,3,1,0]=(0.602022795614)*x[0]**o+(-0.848525569825)*x[0]+(-0.713426287505)*x[1]**o+(-0.829278452687)*x[1]+(-0.491257659316)*x[2]**o+(0.90219908999)*x[2]
        arg[3,3,1,1]=(-0.769110216066)*x[0]**o+(-0.789464037338)*x[0]+(-0.363686172584)*x[1]**o+(-0.86190787068)*x[1]+(0.275438119232)*x[2]**o+(0.285209834248)*x[2]
        arg[3,3,2,0]=(0.351662405853)*x[0]**o+(0.8403302307)*x[0]+(-0.449869332519)*x[1]**o+(0.936043614838)*x[1]+(-0.656908777335)*x[2]**o+(-0.231923872531)*x[2]
        arg[3,3,2,1]=(0.873848945653)*x[0]**o+(0.551429883959)*x[0]+(-0.577942548701)*x[1]**o+(-0.254317163049)*x[1]+(0.64878089878)*x[2]**o+(0.7135339515)*x[2]
        arg[3,4,0,0]=(0.891435610494)*x[0]**o+(0.598893704498)*x[0]+(0.24682094756)*x[1]**o+(0.0895612710338)*x[1]+(-0.513686244953)*x[2]**o+(0.307205505502)*x[2]
        arg[3,4,0,1]=(-0.577752944196)*x[0]**o+(-0.458383751623)*x[0]+(-0.913956988143)*x[1]**o+(0.470857602235)*x[1]+(0.631726850665)*x[2]**o+(-0.112962289904)*x[2]
        arg[3,4,1,0]=(-0.65648177489)*x[0]**o+(-0.843530972747)*x[0]+(-0.0332153723809)*x[1]**o+(0.230484263355)*x[1]+(-0.865687508592)*x[2]**o+(0.122751735082)*x[2]
        arg[3,4,1,1]=(-0.353502640734)*x[0]**o+(-0.816020900282)*x[0]+(0.943688796868)*x[1]**o+(-0.526707881958)*x[1]+(-0.790258298117)*x[2]**o+(-0.748706809042)*x[2]
        arg[3,4,2,0]=(-0.596345912175)*x[0]**o+(0.872771121837)*x[0]+(0.100107395575)*x[1]**o+(0.950482249217)*x[1]+(0.0487634582783)*x[2]**o+(-0.0285218097887)*x[2]
        arg[3,4,2,1]=(-0.827443553933)*x[0]**o+(-0.922564800149)*x[0]+(-0.810354385287)*x[1]**o+(0.630079217293)*x[1]+(-0.316280559723)*x[2]**o+(0.977830435871)*x[2]
        ref[0,0,0,0]=(0.310272924207)*x_ref[0]**o+(-0.664497806351)*x_ref[0]+(0.741942132609)*x_ref[1]**o+(0.636051298779)*x_ref[1]+(0.76627159146)*x_ref[2]**o+(0.492960732588)*x_ref[2]
        ref[0,0,0,1]=(-0.513740801773)*x_ref[0]**o+(-0.633330750909)*x_ref[0]+(0.177934618983)*x_ref[1]**o+(-0.323745723166)*x_ref[1]+(0.694167180904)*x_ref[2]**o+(0.441239251435)*x_ref[2]
        ref[0,0,1,0]=(0.417477695848)*x_ref[0]**o+(-0.923415976002)*x_ref[0]+(0.0765555276662)*x_ref[1]**o+(-0.277290890834)*x_ref[1]+(0.0177191326614)*x_ref[2]**o+(-0.0842520100387)*x_ref[2]
        ref[0,0,1,1]=(-0.761916249894)*x_ref[0]**o+(0.169673139552)*x_ref[0]+(0.0306290828795)*x_ref[1]**o+(0.842234592613)*x_ref[1]+(-0.350342692669)*x_ref[2]**o+(0.321292270658)*x_ref[2]
        ref[0,0,2,0]=(-0.707494137624)*x_ref[0]**o+(0.346873490395)*x_ref[0]+(-0.68281925138)*x_ref[1]**o+(-0.830538083274)*x_ref[1]+(0.329372141613)*x_ref[2]**o+(-0.121366157163)*x_ref[2]
        ref[0,0,2,1]=(-0.708018186507)*x_ref[0]**o+(-0.778051239314)*x_ref[0]+(-0.43106291563)*x_ref[1]**o+(-0.0115914251143)*x_ref[1]+(0.487334215892)*x_ref[2]**o+(-0.289232284658)*x_ref[2]
        ref[0,1,0,0]=(0.530614704199)*x_ref[0]**o+(0.518622028317)*x_ref[0]+(-0.0143129589551)*x_ref[1]**o+(0.808469352257)*x_ref[1]+(-0.156787373225)*x_ref[2]**o+(0.668887813143)*x_ref[2]
        ref[0,1,0,1]=(0.953223233706)*x_ref[0]**o+(-0.164323838353)*x_ref[0]+(0.336717320858)*x_ref[1]**o+(-0.520230488172)*x_ref[1]+(0.689960738002)*x_ref[2]**o+(0.487630947578)*x_ref[2]
        ref[0,1,1,0]=(-0.418095373429)*x_ref[0]**o+(-0.941817844939)*x_ref[0]+(0.88105115607)*x_ref[1]**o+(-0.820502501439)*x_ref[1]+(-0.693858267667)*x_ref[2]**o+(0.530567713127)*x_ref[2]
        ref[0,1,1,1]=(-0.716861320218)*x_ref[0]**o+(-0.736619859467)*x_ref[0]+(-0.885478605601)*x_ref[1]**o+(-0.129375578967)*x_ref[1]+(-0.0181433907318)*x_ref[2]**o+(0.373255537959)*x_ref[2]
        ref[0,1,2,0]=(-0.174742583167)*x_ref[0]**o+(-0.0887967480112)*x_ref[0]+(-0.291168698662)*x_ref[1]**o+(-0.188608506769)*x_ref[1]+(0.22479613454)*x_ref[2]**o+(0.34173928582)*x_ref[2]
        ref[0,1,2,1]=(0.117208929504)*x_ref[0]**o+(-0.420969282079)*x_ref[0]+(0.585298109836)*x_ref[1]**o+(0.771355162745)*x_ref[1]+(-0.51374493747)*x_ref[2]**o+(0.785016548855)*x_ref[2]
        ref[0,2,0,0]=(0.737394887886)*x_ref[0]**o+(0.966080271342)*x_ref[0]+(-0.373178384492)*x_ref[1]**o+(-0.845505386142)*x_ref[1]+(-0.274654519371)*x_ref[2]**o+(0.709356557887)*x_ref[2]
        ref[0,2,0,1]=(0.0365864090636)*x_ref[0]**o+(-0.393382479787)*x_ref[0]+(0.018273360977)*x_ref[1]**o+(-0.509201909328)*x_ref[1]+(-0.777852392534)*x_ref[2]**o+(0.0417280247507)*x_ref[2]
        ref[0,2,1,0]=(0.637829302751)*x_ref[0]**o+(-0.767824790186)*x_ref[0]+(0.843382009511)*x_ref[1]**o+(-0.599633234403)*x_ref[1]+(0.14283138941)*x_ref[2]**o+(-0.354404633991)*x_ref[2]
        ref[0,2,1,1]=(-0.465258613599)*x_ref[0]**o+(0.09341154383)*x_ref[0]+(-0.629114782988)*x_ref[1]**o+(0.91608655383)*x_ref[1]+(-0.692901135848)*x_ref[2]**o+(-0.707225310667)*x_ref[2]
        ref[0,2,2,0]=(0.9935958759)*x_ref[0]**o+(-0.385008742466)*x_ref[0]+(0.153885356854)*x_ref[1]**o+(0.146270680656)*x_ref[1]+(0.305838410343)*x_ref[2]**o+(0.141193552717)*x_ref[2]
        ref[0,2,2,1]=(-0.487538023894)*x_ref[0]**o+(-0.814283442647)*x_ref[0]+(0.166008520199)*x_ref[1]**o+(0.313655774416)*x_ref[1]+(0.282308782973)*x_ref[2]**o+(0.309176062484)*x_ref[2]
        ref[0,3,0,0]=(0.00203507798913)*x_ref[0]**o+(-0.19902542604)*x_ref[0]+(0.88110918624)*x_ref[1]**o+(0.223549219288)*x_ref[1]+(0.135758952369)*x_ref[2]**o+(0.59116783639)*x_ref[2]
        ref[0,3,0,1]=(-0.185063700767)*x_ref[0]**o+(0.833862155173)*x_ref[0]+(-0.0939117509645)*x_ref[1]**o+(0.288003692231)*x_ref[1]+(0.580196157839)*x_ref[2]**o+(0.299320177754)*x_ref[2]
        ref[0,3,1,0]=(-0.039100718826)*x_ref[0]**o+(0.423179869395)*x_ref[0]+(-0.013669578975)*x_ref[1]**o+(-0.355053343085)*x_ref[1]+(0.89285158975)*x_ref[2]**o+(0.392984367877)*x_ref[2]
        ref[0,3,1,1]=(0.0988989416429)*x_ref[0]**o+(-0.446243235174)*x_ref[0]+(-0.691253313617)*x_ref[1]**o+(0.338823693086)*x_ref[1]+(0.707054126334)*x_ref[2]**o+(0.601121388278)*x_ref[2]
        ref[0,3,2,0]=(-0.191366770484)*x_ref[0]**o+(0.961791275903)*x_ref[0]+(-0.103547475035)*x_ref[1]**o+(-0.745236510079)*x_ref[1]+(-0.157506981712)*x_ref[2]**o+(0.986229506787)*x_ref[2]
        ref[0,3,2,1]=(-0.886091883647)*x_ref[0]**o+(0.6979804248)*x_ref[0]+(-0.366575419428)*x_ref[1]**o+(-0.958601153259)*x_ref[1]+(-0.0119724511494)*x_ref[2]**o+(-0.76101132076)*x_ref[2]
        ref[0,4,0,0]=(0.923290554434)*x_ref[0]**o+(0.411359492198)*x_ref[0]+(-0.977999562875)*x_ref[1]**o+(0.352085161527)*x_ref[1]+(0.955745235349)*x_ref[2]**o+(-0.978885609257)*x_ref[2]
        ref[0,4,0,1]=(0.226973479105)*x_ref[0]**o+(-0.423881897251)*x_ref[0]+(0.889101969481)*x_ref[1]**o+(-0.994236389028)*x_ref[1]+(0.0159169033361)*x_ref[2]**o+(-0.559935154203)*x_ref[2]
        ref[0,4,1,0]=(-0.225427291797)*x_ref[0]**o+(-0.557334621848)*x_ref[0]+(-0.486252586142)*x_ref[1]**o+(0.0497120905216)*x_ref[1]+(0.84165736691)*x_ref[2]**o+(-0.185010156797)*x_ref[2]
        ref[0,4,1,1]=(0.762694602609)*x_ref[0]**o+(-0.447367170727)*x_ref[0]+(-0.61103181818)*x_ref[1]**o+(0.689761057656)*x_ref[1]+(-0.631855120122)*x_ref[2]**o+(0.33274655602)*x_ref[2]
        ref[0,4,2,0]=(-0.198884868914)*x_ref[0]**o+(0.716371959278)*x_ref[0]+(0.260871768919)*x_ref[1]**o+(0.622241740082)*x_ref[1]+(0.774103524035)*x_ref[2]**o+(0.186132703676)*x_ref[2]
        ref[0,4,2,1]=(-0.309450645317)*x_ref[0]**o+(0.204754571336)*x_ref[0]+(0.483402810171)*x_ref[1]**o+(0.404620744188)*x_ref[1]+(-0.699137246959)*x_ref[2]**o+(0.467428706196)*x_ref[2]
        ref[1,0,0,0]=(0.642289506448)*x_ref[0]**o+(0.383990304394)*x_ref[0]+(0.556940973318)*x_ref[1]**o+(0.135219729041)*x_ref[1]+(0.577195836358)*x_ref[2]**o+(-0.338924419674)*x_ref[2]
        ref[1,0,0,1]=(-0.920873911295)*x_ref[0]**o+(-0.330610192304)*x_ref[0]+(-0.117226803856)*x_ref[1]**o+(0.172482867684)*x_ref[1]+(0.909140902709)*x_ref[2]**o+(-0.590494167913)*x_ref[2]
        ref[1,0,1,0]=(0.210161538861)*x_ref[0]**o+(-0.83860241761)*x_ref[0]+(0.0257180382661)*x_ref[1]**o+(0.877633854348)*x_ref[1]+(0.26045311754)*x_ref[2]**o+(-0.971349902714)*x_ref[2]
        ref[1,0,1,1]=(0.70736843125)*x_ref[0]**o+(-0.765121985431)*x_ref[0]+(-0.196525225943)*x_ref[1]**o+(-0.175928729346)*x_ref[1]+(0.945523028748)*x_ref[2]**o+(0.877830322823)*x_ref[2]
        ref[1,0,2,0]=(-0.963513016065)*x_ref[0]**o+(0.0820537856351)*x_ref[0]+(0.0122070196118)*x_ref[1]**o+(-0.154147449592)*x_ref[1]+(0.323844633268)*x_ref[2]**o+(-0.124897698214)*x_ref[2]
        ref[1,0,2,1]=(0.656267566405)*x_ref[0]**o+(0.0174786620768)*x_ref[0]+(0.763902435618)*x_ref[1]**o+(-0.986021506043)*x_ref[1]+(0.632625555994)*x_ref[2]**o+(-0.110657743523)*x_ref[2]
        ref[1,1,0,0]=(0.00354144688732)*x_ref[0]**o+(-0.813515808171)*x_ref[0]+(-0.100717717446)*x_ref[1]**o+(0.0543708163629)*x_ref[1]+(0.785927880762)*x_ref[2]**o+(0.0109166652227)*x_ref[2]
        ref[1,1,0,1]=(-0.239705688631)*x_ref[0]**o+(0.402915642509)*x_ref[0]+(0.463371068163)*x_ref[1]**o+(-0.507302462928)*x_ref[1]+(-0.0789844128295)*x_ref[2]**o+(0.669144891499)*x_ref[2]
        ref[1,1,1,0]=(-0.491891367001)*x_ref[0]**o+(-0.524377636928)*x_ref[0]+(0.903091863917)*x_ref[1]**o+(-0.554368807183)*x_ref[1]+(0.332899195339)*x_ref[2]**o+(-0.645216147049)*x_ref[2]
        ref[1,1,1,1]=(-0.915843941243)*x_ref[0]**o+(-0.233175654748)*x_ref[0]+(0.467730425037)*x_ref[1]**o+(-0.765957936683)*x_ref[1]+(0.726678877276)*x_ref[2]**o+(0.770533947942)*x_ref[2]
        ref[1,1,2,0]=(0.678071852032)*x_ref[0]**o+(0.266537844237)*x_ref[0]+(0.324335300151)*x_ref[1]**o+(-0.286713971373)*x_ref[1]+(-0.702384414014)*x_ref[2]**o+(0.887198276519)*x_ref[2]
        ref[1,1,2,1]=(0.0466444959648)*x_ref[0]**o+(0.689162912677)*x_ref[0]+(-0.35033923676)*x_ref[1]**o+(-0.0536380766876)*x_ref[1]+(-0.511833442517)*x_ref[2]**o+(0.125566299829)*x_ref[2]
        ref[1,2,0,0]=(-0.803170431622)*x_ref[0]**o+(0.476211169752)*x_ref[0]+(-0.170510267953)*x_ref[1]**o+(-0.765945735317)*x_ref[1]+(-0.169966329291)*x_ref[2]**o+(-0.988525918724)*x_ref[2]
        ref[1,2,0,1]=(0.289618751212)*x_ref[0]**o+(0.294379040976)*x_ref[0]+(-0.922478416838)*x_ref[1]**o+(0.210546495537)*x_ref[1]+(-0.87266381496)*x_ref[2]**o+(-0.865571669577)*x_ref[2]
        ref[1,2,1,0]=(-0.292811054005)*x_ref[0]**o+(0.651280833746)*x_ref[0]+(-0.625004203736)*x_ref[1]**o+(0.372078169172)*x_ref[1]+(-0.0539099055169)*x_ref[2]**o+(0.0911869270008)*x_ref[2]
        ref[1,2,1,1]=(0.496217698294)*x_ref[0]**o+(-0.986021787133)*x_ref[0]+(0.246417059532)*x_ref[1]**o+(0.610226802401)*x_ref[1]+(0.991534665298)*x_ref[2]**o+(-0.507441185982)*x_ref[2]
        ref[1,2,2,0]=(-0.839902565947)*x_ref[0]**o+(-0.279607147278)*x_ref[0]+(0.278212307963)*x_ref[1]**o+(0.609793640036)*x_ref[1]+(0.940471913774)*x_ref[2]**o+(0.21860828705)*x_ref[2]
        ref[1,2,2,1]=(0.789578112278)*x_ref[0]**o+(-0.0226697818608)*x_ref[0]+(0.216321340291)*x_ref[1]**o+(-0.701818463454)*x_ref[1]+(-0.166689030646)*x_ref[2]**o+(0.966131542292)*x_ref[2]
        ref[1,3,0,0]=(0.149337042551)*x_ref[0]**o+(-0.785050579674)*x_ref[0]+(0.530219924721)*x_ref[1]**o+(-0.971927239398)*x_ref[1]+(0.950917908603)*x_ref[2]**o+(0.132668077447)*x_ref[2]
        ref[1,3,0,1]=(0.983081559724)*x_ref[0]**o+(-0.358694498455)*x_ref[0]+(0.181080099281)*x_ref[1]**o+(-0.699756291643)*x_ref[1]+(-0.744335953304)*x_ref[2]**o+(0.00262043221382)*x_ref[2]
        ref[1,3,1,0]=(0.31776032045)*x_ref[0]**o+(-0.0726184220029)*x_ref[0]+(-0.759883735971)*x_ref[1]**o+(-0.6389710873)*x_ref[1]+(-0.607858499502)*x_ref[2]**o+(-0.397169895784)*x_ref[2]
        ref[1,3,1,1]=(-0.950468389728)*x_ref[0]**o+(0.831121915521)*x_ref[0]+(0.552857011083)*x_ref[1]**o+(0.848599672594)*x_ref[1]+(0.432183564231)*x_ref[2]**o+(0.57537721362)*x_ref[2]
        ref[1,3,2,0]=(0.736080143601)*x_ref[0]**o+(-0.804140307747)*x_ref[0]+(-0.163864867884)*x_ref[1]**o+(0.950874042315)*x_ref[1]+(-0.99676139412)*x_ref[2]**o+(0.332320294246)*x_ref[2]
        ref[1,3,2,1]=(0.119082582041)*x_ref[0]**o+(-0.969740778852)*x_ref[0]+(0.828432583532)*x_ref[1]**o+(0.190324050395)*x_ref[1]+(-0.42639142674)*x_ref[2]**o+(0.315519578192)*x_ref[2]
        ref[1,4,0,0]=(-0.798465141775)*x_ref[0]**o+(-0.717282977908)*x_ref[0]+(0.935655148713)*x_ref[1]**o+(-0.655752964484)*x_ref[1]+(0.159990149099)*x_ref[2]**o+(-0.537115062687)*x_ref[2]
        ref[1,4,0,1]=(0.445151775007)*x_ref[0]**o+(-0.867956302051)*x_ref[0]+(-0.809856879825)*x_ref[1]**o+(-0.326897750981)*x_ref[1]+(0.756352311987)*x_ref[2]**o+(0.577585513459)*x_ref[2]
        ref[1,4,1,0]=(0.294243347591)*x_ref[0]**o+(0.774741656931)*x_ref[0]+(-0.0799985737984)*x_ref[1]**o+(0.179789107412)*x_ref[1]+(-0.612198025189)*x_ref[2]**o+(-0.07086009611)*x_ref[2]
        ref[1,4,1,1]=(0.0692586287045)*x_ref[0]**o+(-0.671956300122)*x_ref[0]+(-0.948667104683)*x_ref[1]**o+(0.68656246801)*x_ref[1]+(-0.352975297696)*x_ref[2]**o+(0.277952113716)*x_ref[2]
        ref[1,4,2,0]=(-0.491316095517)*x_ref[0]**o+(0.725899087239)*x_ref[0]+(0.634113963431)*x_ref[1]**o+(0.235694393453)*x_ref[1]+(-0.15600720713)*x_ref[2]**o+(0.954814976539)*x_ref[2]
        ref[1,4,2,1]=(0.120206996339)*x_ref[0]**o+(-0.86151411823)*x_ref[0]+(-0.137093650097)*x_ref[1]**o+(0.633402978246)*x_ref[1]+(-0.42515364728)*x_ref[2]**o+(-0.987634210617)*x_ref[2]
        ref[2,0,0,0]=(0.539463971222)*x_ref[0]**o+(-0.03435650579)*x_ref[0]+(-0.884395295544)*x_ref[1]**o+(0.215454879022)*x_ref[1]+(0.842377328246)*x_ref[2]**o+(0.370920136871)*x_ref[2]
        ref[2,0,0,1]=(-0.411505463039)*x_ref[0]**o+(0.446494039439)*x_ref[0]+(0.0313771969556)*x_ref[1]**o+(0.602893223061)*x_ref[1]+(-0.30383490021)*x_ref[2]**o+(-0.788028756758)*x_ref[2]
        ref[2,0,1,0]=(0.489009615999)*x_ref[0]**o+(-0.95508248016)*x_ref[0]+(-0.21489475022)*x_ref[1]**o+(-0.857263478431)*x_ref[1]+(-0.123703409045)*x_ref[2]**o+(0.663389562431)*x_ref[2]
        ref[2,0,1,1]=(-0.0224763310929)*x_ref[0]**o+(-0.315886249237)*x_ref[0]+(-0.318201849695)*x_ref[1]**o+(0.423732682462)*x_ref[1]+(0.396065636381)*x_ref[2]**o+(0.0816717878865)*x_ref[2]
        ref[2,0,2,0]=(-0.191432618224)*x_ref[0]**o+(0.688389390795)*x_ref[0]+(-0.382948112159)*x_ref[1]**o+(0.562905173818)*x_ref[1]+(-0.118396714029)*x_ref[2]**o+(0.234478642577)*x_ref[2]
        ref[2,0,2,1]=(-0.348008173171)*x_ref[0]**o+(0.191994166636)*x_ref[0]+(0.450556057147)*x_ref[1]**o+(-0.710821693343)*x_ref[1]+(-0.245119333857)*x_ref[2]**o+(0.262869525239)*x_ref[2]
        ref[2,1,0,0]=(-0.160444506203)*x_ref[0]**o+(-0.816142837432)*x_ref[0]+(-0.619722480492)*x_ref[1]**o+(-0.989077039675)*x_ref[1]+(-0.0282349103977)*x_ref[2]**o+(-0.629849819691)*x_ref[2]
        ref[2,1,0,1]=(0.379513765297)*x_ref[0]**o+(0.934652105586)*x_ref[0]+(-0.632636427868)*x_ref[1]**o+(-0.0492550496553)*x_ref[1]+(0.137733572307)*x_ref[2]**o+(0.779782637942)*x_ref[2]
        ref[2,1,1,0]=(0.469499320417)*x_ref[0]**o+(0.570880924125)*x_ref[0]+(-0.0329833251419)*x_ref[1]**o+(-0.0388751034604)*x_ref[1]+(-0.878520094389)*x_ref[2]**o+(0.992656649048)*x_ref[2]
        ref[2,1,1,1]=(0.592005181995)*x_ref[0]**o+(-0.553980493526)*x_ref[0]+(0.359837665019)*x_ref[1]**o+(0.811107941179)*x_ref[1]+(-0.441929650912)*x_ref[2]**o+(-0.768618378636)*x_ref[2]
        ref[2,1,2,0]=(0.741376258816)*x_ref[0]**o+(-0.238909756717)*x_ref[0]+(0.204631082449)*x_ref[1]**o+(0.970501662949)*x_ref[1]+(-0.615638092239)*x_ref[2]**o+(-0.575948750844)*x_ref[2]
        ref[2,1,2,1]=(0.774050933112)*x_ref[0]**o+(-0.872234076721)*x_ref[0]+(-0.0936189240061)*x_ref[1]**o+(-0.773608756961)*x_ref[1]+(-0.119171907808)*x_ref[2]**o+(-0.256124796845)*x_ref[2]
        ref[2,2,0,0]=(0.459481754425)*x_ref[0]**o+(0.478364929493)*x_ref[0]+(0.460199733287)*x_ref[1]**o+(0.66880294659)*x_ref[1]+(-0.859525058389)*x_ref[2]**o+(0.77960284414)*x_ref[2]
        ref[2,2,0,1]=(0.690557835207)*x_ref[0]**o+(-0.812305038592)*x_ref[0]+(0.837924124791)*x_ref[1]**o+(-0.785979206364)*x_ref[1]+(-0.295553949819)*x_ref[2]**o+(0.575096295897)*x_ref[2]
        ref[2,2,1,0]=(0.443193683205)*x_ref[0]**o+(0.0355033222253)*x_ref[0]+(0.497641002766)*x_ref[1]**o+(-0.636955142451)*x_ref[1]+(-0.265324357454)*x_ref[2]**o+(0.734148569078)*x_ref[2]
        ref[2,2,1,1]=(-0.088707242599)*x_ref[0]**o+(0.513403518669)*x_ref[0]+(0.337741737651)*x_ref[1]**o+(-0.174843924478)*x_ref[1]+(0.687291255986)*x_ref[2]**o+(0.473850502258)*x_ref[2]
        ref[2,2,2,0]=(0.5305624642)*x_ref[0]**o+(-0.972997835953)*x_ref[0]+(0.00197107287888)*x_ref[1]**o+(-0.75490872634)*x_ref[1]+(-0.0569624569435)*x_ref[2]**o+(-0.0780532868631)*x_ref[2]
        ref[2,2,2,1]=(-0.744584471866)*x_ref[0]**o+(0.915494909272)*x_ref[0]+(0.239235050038)*x_ref[1]**o+(-0.192466753187)*x_ref[1]+(0.933556550824)*x_ref[2]**o+(-0.786420251487)*x_ref[2]
        ref[2,3,0,0]=(0.472114530627)*x_ref[0]**o+(0.0693211807557)*x_ref[0]+(-0.274273425713)*x_ref[1]**o+(0.136242554894)*x_ref[1]+(-0.329240675785)*x_ref[2]**o+(0.0212539247343)*x_ref[2]
        ref[2,3,0,1]=(-0.780426108099)*x_ref[0]**o+(-0.632382551025)*x_ref[0]+(0.554926949239)*x_ref[1]**o+(0.264204009059)*x_ref[1]+(-0.868007525824)*x_ref[2]**o+(-0.937631120201)*x_ref[2]
        ref[2,3,1,0]=(-0.826692080947)*x_ref[0]**o+(0.915241858538)*x_ref[0]+(-0.49986160859)*x_ref[1]**o+(0.921686464658)*x_ref[1]+(0.71996666304)*x_ref[2]**o+(0.297609501096)*x_ref[2]
        ref[2,3,1,1]=(0.991896994614)*x_ref[0]**o+(-0.543310008103)*x_ref[0]+(0.0791899631111)*x_ref[1]**o+(-0.914088200145)*x_ref[1]+(-0.496337624951)*x_ref[2]**o+(0.708950898614)*x_ref[2]
        ref[2,3,2,0]=(-0.235116966432)*x_ref[0]**o+(-0.492848730207)*x_ref[0]+(0.428728646789)*x_ref[1]**o+(-0.410747634691)*x_ref[1]+(0.705059815458)*x_ref[2]**o+(-0.185604633831)*x_ref[2]
        ref[2,3,2,1]=(0.838829037117)*x_ref[0]**o+(0.670561147563)*x_ref[0]+(0.621407965075)*x_ref[1]**o+(-0.10624946305)*x_ref[1]+(0.893703402687)*x_ref[2]**o+(0.985504995821)*x_ref[2]
        ref[2,4,0,0]=(0.635072504682)*x_ref[0]**o+(-0.681962427555)*x_ref[0]+(-0.00774518011081)*x_ref[1]**o+(-0.264996228668)*x_ref[1]+(0.690668403198)*x_ref[2]**o+(0.883573761062)*x_ref[2]
        ref[2,4,0,1]=(0.545632268651)*x_ref[0]**o+(0.439668519394)*x_ref[0]+(-0.482479808491)*x_ref[1]**o+(0.752028962391)*x_ref[1]+(0.0596773623939)*x_ref[2]**o+(-0.369987181127)*x_ref[2]
        ref[2,4,1,0]=(0.659005322796)*x_ref[0]**o+(0.503940833428)*x_ref[0]+(0.643633321586)*x_ref[1]**o+(0.927082323252)*x_ref[1]+(0.354072146184)*x_ref[2]**o+(-0.638737560608)*x_ref[2]
        ref[2,4,1,1]=(-0.443548860913)*x_ref[0]**o+(0.707607048458)*x_ref[0]+(-0.955225746402)*x_ref[1]**o+(0.515995563585)*x_ref[1]+(-0.288406637165)*x_ref[2]**o+(-0.760864937318)*x_ref[2]
        ref[2,4,2,0]=(0.898146580355)*x_ref[0]**o+(-0.590650787106)*x_ref[0]+(0.036095789753)*x_ref[1]**o+(0.326040260581)*x_ref[1]+(0.680069361306)*x_ref[2]**o+(0.333839099341)*x_ref[2]
        ref[2,4,2,1]=(-0.151364260271)*x_ref[0]**o+(-0.675198183832)*x_ref[0]+(-0.183626298978)*x_ref[1]**o+(-0.182565217146)*x_ref[1]+(0.0345632129814)*x_ref[2]**o+(-0.581534100763)*x_ref[2]
        ref[3,0,0,0]=(-0.942351703246)*x_ref[0]**o+(-0.699567649453)*x_ref[0]+(-0.832068665702)*x_ref[1]**o+(0.163439416504)*x_ref[1]+(0.742500339519)*x_ref[2]**o+(-0.492882887489)*x_ref[2]
        ref[3,0,0,1]=(0.990078296484)*x_ref[0]**o+(-0.999837336587)*x_ref[0]+(0.197589456459)*x_ref[1]**o+(0.431403059019)*x_ref[1]+(0.859864015871)*x_ref[2]**o+(-0.267474829282)*x_ref[2]
        ref[3,0,1,0]=(-0.776773660723)*x_ref[0]**o+(-0.189048370486)*x_ref[0]+(0.926953856802)*x_ref[1]**o+(-0.72314744652)*x_ref[1]+(0.330064363809)*x_ref[2]**o+(0.188533334359)*x_ref[2]
        ref[3,0,1,1]=(-0.854830058762)*x_ref[0]**o+(0.832073883276)*x_ref[0]+(-0.425691165124)*x_ref[1]**o+(0.162351500316)*x_ref[1]+(0.370604864512)*x_ref[2]**o+(-0.581907071484)*x_ref[2]
        ref[3,0,2,0]=(-0.316716909548)*x_ref[0]**o+(0.335417284383)*x_ref[0]+(-0.783061160352)*x_ref[1]**o+(-0.253972519239)*x_ref[1]+(-0.826432933721)*x_ref[2]**o+(0.267368378363)*x_ref[2]
        ref[3,0,2,1]=(-0.765899705462)*x_ref[0]**o+(-0.0993875861476)*x_ref[0]+(-0.211785004352)*x_ref[1]**o+(-0.635330151497)*x_ref[1]+(0.414575412599)*x_ref[2]**o+(0.586882274623)*x_ref[2]
        ref[3,1,0,0]=(0.987776377237)*x_ref[0]**o+(0.796660669191)*x_ref[0]+(0.580724891862)*x_ref[1]**o+(-0.24149253995)*x_ref[1]+(-0.633421016909)*x_ref[2]**o+(-0.220110913711)*x_ref[2]
        ref[3,1,0,1]=(-0.182240918081)*x_ref[0]**o+(-0.290526050572)*x_ref[0]+(0.00283230959976)*x_ref[1]**o+(-0.933282166805)*x_ref[1]+(-0.857893235883)*x_ref[2]**o+(0.551414074127)*x_ref[2]
        ref[3,1,1,0]=(-0.854435997816)*x_ref[0]**o+(0.706455144263)*x_ref[0]+(0.406237965254)*x_ref[1]**o+(0.192071142415)*x_ref[1]+(0.357966796834)*x_ref[2]**o+(0.887011087987)*x_ref[2]
        ref[3,1,1,1]=(0.832013396996)*x_ref[0]**o+(0.696133072321)*x_ref[0]+(0.682726968712)*x_ref[1]**o+(0.613316240514)*x_ref[1]+(0.12600339887)*x_ref[2]**o+(-0.620088231382)*x_ref[2]
        ref[3,1,2,0]=(-0.178681615221)*x_ref[0]**o+(-0.952835986168)*x_ref[0]+(0.299692693746)*x_ref[1]**o+(0.892451908416)*x_ref[1]+(0.202403273634)*x_ref[2]**o+(0.677212897943)*x_ref[2]
        ref[3,1,2,1]=(0.0592594053342)*x_ref[0]**o+(0.56391153959)*x_ref[0]+(-0.874866627667)*x_ref[1]**o+(0.432102900769)*x_ref[1]+(-0.49073126285)*x_ref[2]**o+(-0.010495860028)*x_ref[2]
        ref[3,2,0,0]=(0.548006445814)*x_ref[0]**o+(0.667313392202)*x_ref[0]+(0.397805355499)*x_ref[1]**o+(-0.788100267808)*x_ref[1]+(0.869234055258)*x_ref[2]**o+(-0.830241520401)*x_ref[2]
        ref[3,2,0,1]=(0.72586139216)*x_ref[0]**o+(-0.931878957878)*x_ref[0]+(-0.94119555858)*x_ref[1]**o+(-0.133544806169)*x_ref[1]+(0.754263323382)*x_ref[2]**o+(-0.446604740157)*x_ref[2]
        ref[3,2,1,0]=(-0.28135830707)*x_ref[0]**o+(-0.478744058518)*x_ref[0]+(0.88219033731)*x_ref[1]**o+(0.460587286945)*x_ref[1]+(-0.849692146096)*x_ref[2]**o+(0.0927701345107)*x_ref[2]
        ref[3,2,1,1]=(-0.681859641611)*x_ref[0]**o+(-0.670164293088)*x_ref[0]+(0.796366374435)*x_ref[1]**o+(0.837230317054)*x_ref[1]+(-0.413283915904)*x_ref[2]**o+(0.779624227187)*x_ref[2]
        ref[3,2,2,0]=(-0.803242771255)*x_ref[0]**o+(-0.0687616271776)*x_ref[0]+(-0.948856027999)*x_ref[1]**o+(0.762568790931)*x_ref[1]+(0.937588292604)*x_ref[2]**o+(-0.0224232611283)*x_ref[2]
        ref[3,2,2,1]=(-0.215752944937)*x_ref[0]**o+(-0.807114296075)*x_ref[0]+(0.970625375964)*x_ref[1]**o+(0.68578124692)*x_ref[1]+(-0.165007047419)*x_ref[2]**o+(0.942718439981)*x_ref[2]
        ref[3,3,0,0]=(0.152001104716)*x_ref[0]**o+(-0.657242342357)*x_ref[0]+(0.559608661705)*x_ref[1]**o+(0.0764403273251)*x_ref[1]+(0.633414754695)*x_ref[2]**o+(0.430777868344)*x_ref[2]
        ref[3,3,0,1]=(-0.898109497759)*x_ref[0]**o+(0.736574249406)*x_ref[0]+(-0.569584184574)*x_ref[1]**o+(-0.145501395506)*x_ref[1]+(-0.412424116475)*x_ref[2]**o+(-0.416350923096)*x_ref[2]
        ref[3,3,1,0]=(0.602022795614)*x_ref[0]**o+(-0.848525569825)*x_ref[0]+(-0.713426287505)*x_ref[1]**o+(-0.829278452687)*x_ref[1]+(-0.491257659316)*x_ref[2]**o+(0.90219908999)*x_ref[2]
        ref[3,3,1,1]=(-0.769110216066)*x_ref[0]**o+(-0.789464037338)*x_ref[0]+(-0.363686172584)*x_ref[1]**o+(-0.86190787068)*x_ref[1]+(0.275438119232)*x_ref[2]**o+(0.285209834248)*x_ref[2]
        ref[3,3,2,0]=(0.351662405853)*x_ref[0]**o+(0.8403302307)*x_ref[0]+(-0.449869332519)*x_ref[1]**o+(0.936043614838)*x_ref[1]+(-0.656908777335)*x_ref[2]**o+(-0.231923872531)*x_ref[2]
        ref[3,3,2,1]=(0.873848945653)*x_ref[0]**o+(0.551429883959)*x_ref[0]+(-0.577942548701)*x_ref[1]**o+(-0.254317163049)*x_ref[1]+(0.64878089878)*x_ref[2]**o+(0.7135339515)*x_ref[2]
        ref[3,4,0,0]=(0.891435610494)*x_ref[0]**o+(0.598893704498)*x_ref[0]+(0.24682094756)*x_ref[1]**o+(0.0895612710338)*x_ref[1]+(-0.513686244953)*x_ref[2]**o+(0.307205505502)*x_ref[2]
        ref[3,4,0,1]=(-0.577752944196)*x_ref[0]**o+(-0.458383751623)*x_ref[0]+(-0.913956988143)*x_ref[1]**o+(0.470857602235)*x_ref[1]+(0.631726850665)*x_ref[2]**o+(-0.112962289904)*x_ref[2]
        ref[3,4,1,0]=(-0.65648177489)*x_ref[0]**o+(-0.843530972747)*x_ref[0]+(-0.0332153723809)*x_ref[1]**o+(0.230484263355)*x_ref[1]+(-0.865687508592)*x_ref[2]**o+(0.122751735082)*x_ref[2]
        ref[3,4,1,1]=(-0.353502640734)*x_ref[0]**o+(-0.816020900282)*x_ref[0]+(0.943688796868)*x_ref[1]**o+(-0.526707881958)*x_ref[1]+(-0.790258298117)*x_ref[2]**o+(-0.748706809042)*x_ref[2]
        ref[3,4,2,0]=(-0.596345912175)*x_ref[0]**o+(0.872771121837)*x_ref[0]+(0.100107395575)*x_ref[1]**o+(0.950482249217)*x_ref[1]+(0.0487634582783)*x_ref[2]**o+(-0.0285218097887)*x_ref[2]
        ref[3,4,2,1]=(-0.827443553933)*x_ref[0]**o+(-0.922564800149)*x_ref[0]+(-0.810354385287)*x_ref[1]**o+(0.630079217293)*x_ref[1]+(-0.316280559723)*x_ref[2]**o+(0.977830435871)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.142445811229)*x[0]**o+(-0.170111870439)*x[0]+(-0.571150729608)*x[1]**o+(-0.799596368319)*x[1]
        ref=(0.142445811229)*x_ref[0]**o+(-0.170111870439)*x_ref[0]+(-0.571150729608)*x_ref[1]**o+(-0.799596368319)*x_ref[1]
      else:
        arg=(0.131269935442)*x[0]**o+(0.67764849927)*x[0]+(0.199946665)*x[1]**o+(0.775678673404)*x[1]+(-0.818505707274)*x[2]**o+(-0.141160005914)*x[2]
        ref=(0.131269935442)*x_ref[0]**o+(0.67764849927)*x_ref[0]+(0.199946665)*x_ref[1]**o+(0.775678673404)*x_ref[1]+(-0.818505707274)*x_ref[2]**o+(-0.141160005914)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.271030346059)*x[0]**o+(-0.604054986035)*x[0]+(-0.856492962015)*x[1]**o+(0.229582187423)*x[1]
        arg[1]=(-0.152539485929)*x[0]**o+(0.936634973224)*x[0]+(0.805016132937)*x[1]**o+(-0.6832854412)*x[1]
        ref[0]=(-0.271030346059)*x_ref[0]**o+(-0.604054986035)*x_ref[0]+(-0.856492962015)*x_ref[1]**o+(0.229582187423)*x_ref[1]
        ref[1]=(-0.152539485929)*x_ref[0]**o+(0.936634973224)*x_ref[0]+(0.805016132937)*x_ref[1]**o+(-0.6832854412)*x_ref[1]
      else:
        arg[0]=(-0.484580515247)*x[0]**o+(-0.134539997002)*x[0]+(0.589742989514)*x[1]**o+(-0.950803827747)*x[1]+(-0.852976687869)*x[2]**o+(0.0999526415345)*x[2]
        arg[1]=(-0.863620642149)*x[0]**o+(-0.16255544441)*x[0]+(-0.4963232161)*x[1]**o+(0.180988811217)*x[1]+(-0.787186716126)*x[2]**o+(-0.359825851725)*x[2]
        ref[0]=(-0.484580515247)*x_ref[0]**o+(-0.134539997002)*x_ref[0]+(0.589742989514)*x_ref[1]**o+(-0.950803827747)*x_ref[1]+(-0.852976687869)*x_ref[2]**o+(0.0999526415345)*x_ref[2]
        ref[1]=(-0.863620642149)*x_ref[0]**o+(-0.16255544441)*x_ref[0]+(-0.4963232161)*x_ref[1]**o+(0.180988811217)*x_ref[1]+(-0.787186716126)*x_ref[2]**o+(-0.359825851725)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.88064835903)*x[0]**o+(-0.140271048585)*x[0]+(-0.363467917345)*x[1]**o+(0.382530733181)*x[1]
        arg[0,1]=(0.835096280471)*x[0]**o+(0.132077921375)*x[0]+(-0.814392115734)*x[1]**o+(-0.379227781873)*x[1]
        arg[0,2]=(0.257391174596)*x[0]**o+(-0.135017751364)*x[0]+(-0.462760793484)*x[1]**o+(-0.332510860616)*x[1]
        arg[0,3]=(0.303541666957)*x[0]**o+(0.729054207976)*x[0]+(-0.0474210800089)*x[1]**o+(0.213859363809)*x[1]
        arg[0,4]=(0.564466740909)*x[0]**o+(0.684761789227)*x[0]+(0.432934636771)*x[1]**o+(0.261269614075)*x[1]
        arg[1,0]=(0.83080596334)*x[0]**o+(0.968795153198)*x[0]+(0.835115581928)*x[1]**o+(-0.134688885728)*x[1]
        arg[1,1]=(-0.876235093042)*x[0]**o+(-0.43923440557)*x[0]+(-0.641186633503)*x[1]**o+(-0.987129244994)*x[1]
        arg[1,2]=(-0.994007922096)*x[0]**o+(0.534914510762)*x[0]+(0.754420969331)*x[1]**o+(0.241533891038)*x[1]
        arg[1,3]=(-0.196027247194)*x[0]**o+(0.962231147494)*x[0]+(-0.819146467532)*x[1]**o+(0.966024528731)*x[1]
        arg[1,4]=(-0.577014723723)*x[0]**o+(0.350310620855)*x[0]+(0.16266475329)*x[1]**o+(-0.926739533525)*x[1]
        arg[2,0]=(-0.262964780787)*x[0]**o+(-0.824445469277)*x[0]+(-0.982497605841)*x[1]**o+(0.900080022932)*x[1]
        arg[2,1]=(0.54409272334)*x[0]**o+(0.789421020125)*x[0]+(0.672239490644)*x[1]**o+(-0.0306885937363)*x[1]
        arg[2,2]=(-0.130272075526)*x[0]**o+(0.51213564573)*x[0]+(-0.819817699321)*x[1]**o+(-0.259068167835)*x[1]
        arg[2,3]=(-0.61526267092)*x[0]**o+(-0.504362726269)*x[0]+(0.362429515671)*x[1]**o+(-0.0921852465615)*x[1]
        arg[2,4]=(-0.344772989987)*x[0]**o+(0.260747743828)*x[0]+(0.474457474131)*x[1]**o+(0.436008642452)*x[1]
        arg[3,0]=(-0.72740499107)*x[0]**o+(-0.297524430077)*x[0]+(-0.809254720524)*x[1]**o+(-0.230695822532)*x[1]
        arg[3,1]=(-0.0817042383885)*x[0]**o+(-0.324890823023)*x[0]+(0.923635342281)*x[1]**o+(-0.336825317481)*x[1]
        arg[3,2]=(-0.872954045409)*x[0]**o+(0.242165239629)*x[0]+(0.919264824591)*x[1]**o+(0.424247774945)*x[1]
        arg[3,3]=(0.146344350418)*x[0]**o+(-0.705285807962)*x[0]+(-0.0735095471635)*x[1]**o+(-0.464180824267)*x[1]
        arg[3,4]=(0.624085113245)*x[0]**o+(0.719452515121)*x[0]+(0.28331108038)*x[1]**o+(-0.897435702442)*x[1]
        ref[0,0]=(-0.88064835903)*x_ref[0]**o+(-0.140271048585)*x_ref[0]+(-0.363467917345)*x_ref[1]**o+(0.382530733181)*x_ref[1]
        ref[0,1]=(0.835096280471)*x_ref[0]**o+(0.132077921375)*x_ref[0]+(-0.814392115734)*x_ref[1]**o+(-0.379227781873)*x_ref[1]
        ref[0,2]=(0.257391174596)*x_ref[0]**o+(-0.135017751364)*x_ref[0]+(-0.462760793484)*x_ref[1]**o+(-0.332510860616)*x_ref[1]
        ref[0,3]=(0.303541666957)*x_ref[0]**o+(0.729054207976)*x_ref[0]+(-0.0474210800089)*x_ref[1]**o+(0.213859363809)*x_ref[1]
        ref[0,4]=(0.564466740909)*x_ref[0]**o+(0.684761789227)*x_ref[0]+(0.432934636771)*x_ref[1]**o+(0.261269614075)*x_ref[1]
        ref[1,0]=(0.83080596334)*x_ref[0]**o+(0.968795153198)*x_ref[0]+(0.835115581928)*x_ref[1]**o+(-0.134688885728)*x_ref[1]
        ref[1,1]=(-0.876235093042)*x_ref[0]**o+(-0.43923440557)*x_ref[0]+(-0.641186633503)*x_ref[1]**o+(-0.987129244994)*x_ref[1]
        ref[1,2]=(-0.994007922096)*x_ref[0]**o+(0.534914510762)*x_ref[0]+(0.754420969331)*x_ref[1]**o+(0.241533891038)*x_ref[1]
        ref[1,3]=(-0.196027247194)*x_ref[0]**o+(0.962231147494)*x_ref[0]+(-0.819146467532)*x_ref[1]**o+(0.966024528731)*x_ref[1]
        ref[1,4]=(-0.577014723723)*x_ref[0]**o+(0.350310620855)*x_ref[0]+(0.16266475329)*x_ref[1]**o+(-0.926739533525)*x_ref[1]
        ref[2,0]=(-0.262964780787)*x_ref[0]**o+(-0.824445469277)*x_ref[0]+(-0.982497605841)*x_ref[1]**o+(0.900080022932)*x_ref[1]
        ref[2,1]=(0.54409272334)*x_ref[0]**o+(0.789421020125)*x_ref[0]+(0.672239490644)*x_ref[1]**o+(-0.0306885937363)*x_ref[1]
        ref[2,2]=(-0.130272075526)*x_ref[0]**o+(0.51213564573)*x_ref[0]+(-0.819817699321)*x_ref[1]**o+(-0.259068167835)*x_ref[1]
        ref[2,3]=(-0.61526267092)*x_ref[0]**o+(-0.504362726269)*x_ref[0]+(0.362429515671)*x_ref[1]**o+(-0.0921852465615)*x_ref[1]
        ref[2,4]=(-0.344772989987)*x_ref[0]**o+(0.260747743828)*x_ref[0]+(0.474457474131)*x_ref[1]**o+(0.436008642452)*x_ref[1]
        ref[3,0]=(-0.72740499107)*x_ref[0]**o+(-0.297524430077)*x_ref[0]+(-0.809254720524)*x_ref[1]**o+(-0.230695822532)*x_ref[1]
        ref[3,1]=(-0.0817042383885)*x_ref[0]**o+(-0.324890823023)*x_ref[0]+(0.923635342281)*x_ref[1]**o+(-0.336825317481)*x_ref[1]
        ref[3,2]=(-0.872954045409)*x_ref[0]**o+(0.242165239629)*x_ref[0]+(0.919264824591)*x_ref[1]**o+(0.424247774945)*x_ref[1]
        ref[3,3]=(0.146344350418)*x_ref[0]**o+(-0.705285807962)*x_ref[0]+(-0.0735095471635)*x_ref[1]**o+(-0.464180824267)*x_ref[1]
        ref[3,4]=(0.624085113245)*x_ref[0]**o+(0.719452515121)*x_ref[0]+(0.28331108038)*x_ref[1]**o+(-0.897435702442)*x_ref[1]
      else:
        arg[0,0]=(0.47332206873)*x[0]**o+(0.20564453167)*x[0]+(-0.175485469473)*x[1]**o+(-0.25166656853)*x[1]+(-0.242693318919)*x[2]**o+(-0.773417228526)*x[2]
        arg[0,1]=(0.788563488587)*x[0]**o+(0.598162739778)*x[0]+(-0.736463777722)*x[1]**o+(-0.161555683484)*x[1]+(0.480782882175)*x[2]**o+(-0.500552084345)*x[2]
        arg[0,2]=(-0.207893050318)*x[0]**o+(0.206221241577)*x[0]+(0.306329036151)*x[1]**o+(0.585237890094)*x[1]+(0.0276469766188)*x[2]**o+(0.0502503670529)*x[2]
        arg[0,3]=(0.914950534537)*x[0]**o+(0.401315333176)*x[0]+(0.245224334898)*x[1]**o+(0.526049350134)*x[1]+(-0.735617023846)*x[2]**o+(-0.674692140531)*x[2]
        arg[0,4]=(-0.359566330106)*x[0]**o+(-0.279627730979)*x[0]+(-0.08398829656)*x[1]**o+(-0.789678621983)*x[1]+(-0.357374658443)*x[2]**o+(0.795724504883)*x[2]
        arg[1,0]=(-0.0715278193161)*x[0]**o+(-0.562120168502)*x[0]+(0.428614380101)*x[1]**o+(0.115408373591)*x[1]+(-0.154674397872)*x[2]**o+(0.233468972682)*x[2]
        arg[1,1]=(0.707445640039)*x[0]**o+(-0.719649334705)*x[0]+(-0.349743923514)*x[1]**o+(0.369262203159)*x[1]+(0.712786695739)*x[2]**o+(-0.962450154391)*x[2]
        arg[1,2]=(-0.306983094408)*x[0]**o+(-0.933322822964)*x[0]+(-0.525515899443)*x[1]**o+(0.642320930433)*x[1]+(-0.950421802297)*x[2]**o+(-0.392142385683)*x[2]
        arg[1,3]=(-0.765158324235)*x[0]**o+(-0.362010493356)*x[0]+(-0.918673930909)*x[1]**o+(0.72772784184)*x[1]+(-0.392848438408)*x[2]**o+(0.566119231816)*x[2]
        arg[1,4]=(0.429984662315)*x[0]**o+(-0.165476415697)*x[0]+(0.778539911928)*x[1]**o+(0.411518911472)*x[1]+(0.418323432603)*x[2]**o+(-0.846495913126)*x[2]
        arg[2,0]=(-0.836556750858)*x[0]**o+(0.976027316021)*x[0]+(-0.556499096672)*x[1]**o+(0.320038395915)*x[1]+(0.377663884516)*x[2]**o+(0.491202727923)*x[2]
        arg[2,1]=(-0.329597826951)*x[0]**o+(-0.555819984044)*x[0]+(0.110029358927)*x[1]**o+(-0.737070187836)*x[1]+(-0.587142473737)*x[2]**o+(-0.0388834177112)*x[2]
        arg[2,2]=(0.442251030646)*x[0]**o+(0.285010574287)*x[0]+(0.603267537331)*x[1]**o+(-0.0776884302384)*x[1]+(-0.123541484708)*x[2]**o+(0.44717125529)*x[2]
        arg[2,3]=(0.474458930313)*x[0]**o+(0.415947401939)*x[0]+(0.381539850696)*x[1]**o+(-0.307590909422)*x[1]+(0.488348451586)*x[2]**o+(0.336883810306)*x[2]
        arg[2,4]=(0.804466542611)*x[0]**o+(0.891337158669)*x[0]+(-0.599119487307)*x[1]**o+(-0.92047732727)*x[1]+(-0.251489480657)*x[2]**o+(-0.266483991526)*x[2]
        arg[3,0]=(-0.733633520827)*x[0]**o+(0.494805537624)*x[0]+(-0.117702558274)*x[1]**o+(-0.000461693303442)*x[1]+(-0.102033981202)*x[2]**o+(-0.772865985272)*x[2]
        arg[3,1]=(-0.636901276465)*x[0]**o+(-0.436330072041)*x[0]+(0.403048850222)*x[1]**o+(-0.94737100009)*x[1]+(0.781059874101)*x[2]**o+(-0.952781516037)*x[2]
        arg[3,2]=(-0.489372336164)*x[0]**o+(0.221211043152)*x[0]+(0.793975373906)*x[1]**o+(0.424349965288)*x[1]+(0.453332885662)*x[2]**o+(-0.513541753005)*x[2]
        arg[3,3]=(-0.700899677836)*x[0]**o+(-0.984271896306)*x[0]+(-0.663909486898)*x[1]**o+(0.461302174316)*x[1]+(-0.906457707724)*x[2]**o+(-0.0440003164038)*x[2]
        arg[3,4]=(0.838326295366)*x[0]**o+(-0.483407593962)*x[0]+(-0.122458752397)*x[1]**o+(0.607337951095)*x[1]+(0.750481384434)*x[2]**o+(-0.677155633204)*x[2]
        ref[0,0]=(0.47332206873)*x_ref[0]**o+(0.20564453167)*x_ref[0]+(-0.175485469473)*x_ref[1]**o+(-0.25166656853)*x_ref[1]+(-0.242693318919)*x_ref[2]**o+(-0.773417228526)*x_ref[2]
        ref[0,1]=(0.788563488587)*x_ref[0]**o+(0.598162739778)*x_ref[0]+(-0.736463777722)*x_ref[1]**o+(-0.161555683484)*x_ref[1]+(0.480782882175)*x_ref[2]**o+(-0.500552084345)*x_ref[2]
        ref[0,2]=(-0.207893050318)*x_ref[0]**o+(0.206221241577)*x_ref[0]+(0.306329036151)*x_ref[1]**o+(0.585237890094)*x_ref[1]+(0.0276469766188)*x_ref[2]**o+(0.0502503670529)*x_ref[2]
        ref[0,3]=(0.914950534537)*x_ref[0]**o+(0.401315333176)*x_ref[0]+(0.245224334898)*x_ref[1]**o+(0.526049350134)*x_ref[1]+(-0.735617023846)*x_ref[2]**o+(-0.674692140531)*x_ref[2]
        ref[0,4]=(-0.359566330106)*x_ref[0]**o+(-0.279627730979)*x_ref[0]+(-0.08398829656)*x_ref[1]**o+(-0.789678621983)*x_ref[1]+(-0.357374658443)*x_ref[2]**o+(0.795724504883)*x_ref[2]
        ref[1,0]=(-0.0715278193161)*x_ref[0]**o+(-0.562120168502)*x_ref[0]+(0.428614380101)*x_ref[1]**o+(0.115408373591)*x_ref[1]+(-0.154674397872)*x_ref[2]**o+(0.233468972682)*x_ref[2]
        ref[1,1]=(0.707445640039)*x_ref[0]**o+(-0.719649334705)*x_ref[0]+(-0.349743923514)*x_ref[1]**o+(0.369262203159)*x_ref[1]+(0.712786695739)*x_ref[2]**o+(-0.962450154391)*x_ref[2]
        ref[1,2]=(-0.306983094408)*x_ref[0]**o+(-0.933322822964)*x_ref[0]+(-0.525515899443)*x_ref[1]**o+(0.642320930433)*x_ref[1]+(-0.950421802297)*x_ref[2]**o+(-0.392142385683)*x_ref[2]
        ref[1,3]=(-0.765158324235)*x_ref[0]**o+(-0.362010493356)*x_ref[0]+(-0.918673930909)*x_ref[1]**o+(0.72772784184)*x_ref[1]+(-0.392848438408)*x_ref[2]**o+(0.566119231816)*x_ref[2]
        ref[1,4]=(0.429984662315)*x_ref[0]**o+(-0.165476415697)*x_ref[0]+(0.778539911928)*x_ref[1]**o+(0.411518911472)*x_ref[1]+(0.418323432603)*x_ref[2]**o+(-0.846495913126)*x_ref[2]
        ref[2,0]=(-0.836556750858)*x_ref[0]**o+(0.976027316021)*x_ref[0]+(-0.556499096672)*x_ref[1]**o+(0.320038395915)*x_ref[1]+(0.377663884516)*x_ref[2]**o+(0.491202727923)*x_ref[2]
        ref[2,1]=(-0.329597826951)*x_ref[0]**o+(-0.555819984044)*x_ref[0]+(0.110029358927)*x_ref[1]**o+(-0.737070187836)*x_ref[1]+(-0.587142473737)*x_ref[2]**o+(-0.0388834177112)*x_ref[2]
        ref[2,2]=(0.442251030646)*x_ref[0]**o+(0.285010574287)*x_ref[0]+(0.603267537331)*x_ref[1]**o+(-0.0776884302384)*x_ref[1]+(-0.123541484708)*x_ref[2]**o+(0.44717125529)*x_ref[2]
        ref[2,3]=(0.474458930313)*x_ref[0]**o+(0.415947401939)*x_ref[0]+(0.381539850696)*x_ref[1]**o+(-0.307590909422)*x_ref[1]+(0.488348451586)*x_ref[2]**o+(0.336883810306)*x_ref[2]
        ref[2,4]=(0.804466542611)*x_ref[0]**o+(0.891337158669)*x_ref[0]+(-0.599119487307)*x_ref[1]**o+(-0.92047732727)*x_ref[1]+(-0.251489480657)*x_ref[2]**o+(-0.266483991526)*x_ref[2]
        ref[3,0]=(-0.733633520827)*x_ref[0]**o+(0.494805537624)*x_ref[0]+(-0.117702558274)*x_ref[1]**o+(-0.000461693303442)*x_ref[1]+(-0.102033981202)*x_ref[2]**o+(-0.772865985272)*x_ref[2]
        ref[3,1]=(-0.636901276465)*x_ref[0]**o+(-0.436330072041)*x_ref[0]+(0.403048850222)*x_ref[1]**o+(-0.94737100009)*x_ref[1]+(0.781059874101)*x_ref[2]**o+(-0.952781516037)*x_ref[2]
        ref[3,2]=(-0.489372336164)*x_ref[0]**o+(0.221211043152)*x_ref[0]+(0.793975373906)*x_ref[1]**o+(0.424349965288)*x_ref[1]+(0.453332885662)*x_ref[2]**o+(-0.513541753005)*x_ref[2]
        ref[3,3]=(-0.700899677836)*x_ref[0]**o+(-0.984271896306)*x_ref[0]+(-0.663909486898)*x_ref[1]**o+(0.461302174316)*x_ref[1]+(-0.906457707724)*x_ref[2]**o+(-0.0440003164038)*x_ref[2]
        ref[3,4]=(0.838326295366)*x_ref[0]**o+(-0.483407593962)*x_ref[0]+(-0.122458752397)*x_ref[1]**o+(0.607337951095)*x_ref[1]+(0.750481384434)*x_ref[2]**o+(-0.677155633204)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.111473275114)*x[0]**o+(-0.552264448532)*x[0]+(0.67142534913)*x[1]**o+(-0.642305751076)*x[1]
        arg[0,0,1]=(-0.431075800974)*x[0]**o+(0.911382078454)*x[0]+(0.546541644612)*x[1]**o+(-0.67679652536)*x[1]
        arg[0,1,0]=(0.103056292292)*x[0]**o+(-0.640858804416)*x[0]+(0.159596964992)*x[1]**o+(-0.237782007316)*x[1]
        arg[0,1,1]=(0.975898170389)*x[0]**o+(0.255916326383)*x[0]+(-0.692770462322)*x[1]**o+(0.786355114846)*x[1]
        arg[1,0,0]=(-0.669378370759)*x[0]**o+(-0.0153788709146)*x[0]+(0.924383188557)*x[1]**o+(0.885037556829)*x[1]
        arg[1,0,1]=(-0.509891819881)*x[0]**o+(-0.377779514697)*x[0]+(-0.343176473036)*x[1]**o+(-0.39978416725)*x[1]
        arg[1,1,0]=(0.457959467665)*x[0]**o+(0.786511967341)*x[0]+(0.256561665146)*x[1]**o+(-0.261300099643)*x[1]
        arg[1,1,1]=(0.219754515189)*x[0]**o+(-0.131747762867)*x[0]+(0.736017818893)*x[1]**o+(-0.450963700533)*x[1]
        arg[2,0,0]=(-0.182946832789)*x[0]**o+(-0.103064731935)*x[0]+(-0.223584334222)*x[1]**o+(-0.578088201628)*x[1]
        arg[2,0,1]=(-0.830744157425)*x[0]**o+(0.590627167207)*x[0]+(-0.538048378049)*x[1]**o+(-0.0529793352741)*x[1]
        arg[2,1,0]=(-0.307908670388)*x[0]**o+(-0.929987734045)*x[0]+(-0.167190563092)*x[1]**o+(0.126466628227)*x[1]
        arg[2,1,1]=(-0.812045378916)*x[0]**o+(0.676973938072)*x[0]+(0.528243909204)*x[1]**o+(-0.594241387691)*x[1]
        arg[3,0,0]=(-0.0198606164114)*x[0]**o+(-0.00574708685554)*x[0]+(0.908946312745)*x[1]**o+(0.775054099788)*x[1]
        arg[3,0,1]=(-0.810078120518)*x[0]**o+(-0.960585888485)*x[0]+(-0.0747272427198)*x[1]**o+(0.164107641817)*x[1]
        arg[3,1,0]=(-0.535791917139)*x[0]**o+(0.0424226440074)*x[0]+(0.701479263902)*x[1]**o+(0.787745423088)*x[1]
        arg[3,1,1]=(-0.896532152793)*x[0]**o+(-0.0224566920251)*x[0]+(-0.554655626689)*x[1]**o+(0.41012867952)*x[1]
        arg[4,0,0]=(0.972433817496)*x[0]**o+(-0.440662536033)*x[0]+(-0.910456530943)*x[1]**o+(-0.333429535028)*x[1]
        arg[4,0,1]=(-0.415824313942)*x[0]**o+(0.00255557712922)*x[0]+(0.304897639195)*x[1]**o+(-0.866600848052)*x[1]
        arg[4,1,0]=(-0.866277739076)*x[0]**o+(0.282976169776)*x[0]+(0.477528278501)*x[1]**o+(-0.233506497695)*x[1]
        arg[4,1,1]=(-0.12729626381)*x[0]**o+(-0.723204796785)*x[0]+(0.0856495713442)*x[1]**o+(-0.214819437731)*x[1]
        arg[5,0,0]=(0.208133503644)*x[0]**o+(-0.955529202895)*x[0]+(0.0291569127831)*x[1]**o+(0.489913918305)*x[1]
        arg[5,0,1]=(0.96923667477)*x[0]**o+(0.543776525315)*x[0]+(0.38666450029)*x[1]**o+(-0.193493550272)*x[1]
        arg[5,1,0]=(-0.0912940016888)*x[0]**o+(-0.477561072681)*x[0]+(0.474816021613)*x[1]**o+(-0.252664229253)*x[1]
        arg[5,1,1]=(0.748337265159)*x[0]**o+(-0.706194454618)*x[0]+(-0.504466281507)*x[1]**o+(0.582500817679)*x[1]
        ref[0,0,0]=(0.111473275114)*x_ref[0]**o+(-0.552264448532)*x_ref[0]+(0.67142534913)*x_ref[1]**o+(-0.642305751076)*x_ref[1]
        ref[0,0,1]=(-0.431075800974)*x_ref[0]**o+(0.911382078454)*x_ref[0]+(0.546541644612)*x_ref[1]**o+(-0.67679652536)*x_ref[1]
        ref[0,1,0]=(0.103056292292)*x_ref[0]**o+(-0.640858804416)*x_ref[0]+(0.159596964992)*x_ref[1]**o+(-0.237782007316)*x_ref[1]
        ref[0,1,1]=(0.975898170389)*x_ref[0]**o+(0.255916326383)*x_ref[0]+(-0.692770462322)*x_ref[1]**o+(0.786355114846)*x_ref[1]
        ref[1,0,0]=(-0.669378370759)*x_ref[0]**o+(-0.0153788709146)*x_ref[0]+(0.924383188557)*x_ref[1]**o+(0.885037556829)*x_ref[1]
        ref[1,0,1]=(-0.509891819881)*x_ref[0]**o+(-0.377779514697)*x_ref[0]+(-0.343176473036)*x_ref[1]**o+(-0.39978416725)*x_ref[1]
        ref[1,1,0]=(0.457959467665)*x_ref[0]**o+(0.786511967341)*x_ref[0]+(0.256561665146)*x_ref[1]**o+(-0.261300099643)*x_ref[1]
        ref[1,1,1]=(0.219754515189)*x_ref[0]**o+(-0.131747762867)*x_ref[0]+(0.736017818893)*x_ref[1]**o+(-0.450963700533)*x_ref[1]
        ref[2,0,0]=(-0.182946832789)*x_ref[0]**o+(-0.103064731935)*x_ref[0]+(-0.223584334222)*x_ref[1]**o+(-0.578088201628)*x_ref[1]
        ref[2,0,1]=(-0.830744157425)*x_ref[0]**o+(0.590627167207)*x_ref[0]+(-0.538048378049)*x_ref[1]**o+(-0.0529793352741)*x_ref[1]
        ref[2,1,0]=(-0.307908670388)*x_ref[0]**o+(-0.929987734045)*x_ref[0]+(-0.167190563092)*x_ref[1]**o+(0.126466628227)*x_ref[1]
        ref[2,1,1]=(-0.812045378916)*x_ref[0]**o+(0.676973938072)*x_ref[0]+(0.528243909204)*x_ref[1]**o+(-0.594241387691)*x_ref[1]
        ref[3,0,0]=(-0.0198606164114)*x_ref[0]**o+(-0.00574708685554)*x_ref[0]+(0.908946312745)*x_ref[1]**o+(0.775054099788)*x_ref[1]
        ref[3,0,1]=(-0.810078120518)*x_ref[0]**o+(-0.960585888485)*x_ref[0]+(-0.0747272427198)*x_ref[1]**o+(0.164107641817)*x_ref[1]
        ref[3,1,0]=(-0.535791917139)*x_ref[0]**o+(0.0424226440074)*x_ref[0]+(0.701479263902)*x_ref[1]**o+(0.787745423088)*x_ref[1]
        ref[3,1,1]=(-0.896532152793)*x_ref[0]**o+(-0.0224566920251)*x_ref[0]+(-0.554655626689)*x_ref[1]**o+(0.41012867952)*x_ref[1]
        ref[4,0,0]=(0.972433817496)*x_ref[0]**o+(-0.440662536033)*x_ref[0]+(-0.910456530943)*x_ref[1]**o+(-0.333429535028)*x_ref[1]
        ref[4,0,1]=(-0.415824313942)*x_ref[0]**o+(0.00255557712922)*x_ref[0]+(0.304897639195)*x_ref[1]**o+(-0.866600848052)*x_ref[1]
        ref[4,1,0]=(-0.866277739076)*x_ref[0]**o+(0.282976169776)*x_ref[0]+(0.477528278501)*x_ref[1]**o+(-0.233506497695)*x_ref[1]
        ref[4,1,1]=(-0.12729626381)*x_ref[0]**o+(-0.723204796785)*x_ref[0]+(0.0856495713442)*x_ref[1]**o+(-0.214819437731)*x_ref[1]
        ref[5,0,0]=(0.208133503644)*x_ref[0]**o+(-0.955529202895)*x_ref[0]+(0.0291569127831)*x_ref[1]**o+(0.489913918305)*x_ref[1]
        ref[5,0,1]=(0.96923667477)*x_ref[0]**o+(0.543776525315)*x_ref[0]+(0.38666450029)*x_ref[1]**o+(-0.193493550272)*x_ref[1]
        ref[5,1,0]=(-0.0912940016888)*x_ref[0]**o+(-0.477561072681)*x_ref[0]+(0.474816021613)*x_ref[1]**o+(-0.252664229253)*x_ref[1]
        ref[5,1,1]=(0.748337265159)*x_ref[0]**o+(-0.706194454618)*x_ref[0]+(-0.504466281507)*x_ref[1]**o+(0.582500817679)*x_ref[1]
      else:
        arg[0,0,0]=(-0.179972720095)*x[0]**o+(-0.546715268523)*x[0]+(0.835867431319)*x[1]**o+(0.626798499259)*x[1]+(0.397356013278)*x[2]**o+(0.0771586784158)*x[2]
        arg[0,0,1]=(0.625676031039)*x[0]**o+(0.65843849475)*x[0]+(0.456469199307)*x[1]**o+(-0.970709759479)*x[1]+(-0.699606143992)*x[2]**o+(0.76429400724)*x[2]
        arg[0,1,0]=(0.613916424165)*x[0]**o+(0.556778588303)*x[0]+(-0.509658743529)*x[1]**o+(-0.978718968218)*x[1]+(-0.603928369428)*x[2]**o+(-0.550813953692)*x[2]
        arg[0,1,1]=(-0.935214755436)*x[0]**o+(0.44745901813)*x[0]+(-0.582828846453)*x[1]**o+(-0.181547183233)*x[1]+(-0.183912053485)*x[2]**o+(0.446152494296)*x[2]
        arg[1,0,0]=(-0.486345219779)*x[0]**o+(0.753529436833)*x[0]+(0.138157856552)*x[1]**o+(-0.874849020539)*x[1]+(0.374396543792)*x[2]**o+(0.26548456046)*x[2]
        arg[1,0,1]=(0.492908753036)*x[0]**o+(-0.418118685105)*x[0]+(0.880817569823)*x[1]**o+(-0.827691145524)*x[1]+(0.238633156133)*x[2]**o+(-0.463687731341)*x[2]
        arg[1,1,0]=(-0.397373578442)*x[0]**o+(-0.316321105607)*x[0]+(0.647511225459)*x[1]**o+(-0.82553752195)*x[1]+(-0.715549303171)*x[2]**o+(0.804623957459)*x[2]
        arg[1,1,1]=(0.414743130234)*x[0]**o+(0.0201794413683)*x[0]+(-0.636337599584)*x[1]**o+(-0.230053616026)*x[1]+(-0.4302153311)*x[2]**o+(-0.387468029626)*x[2]
        arg[2,0,0]=(0.715100517336)*x[0]**o+(-0.4286120178)*x[0]+(0.0987858542117)*x[1]**o+(0.815093470301)*x[1]+(0.293455353257)*x[2]**o+(0.200576090379)*x[2]
        arg[2,0,1]=(-0.158330015822)*x[0]**o+(0.363339023457)*x[0]+(-0.796055594971)*x[1]**o+(-0.213299205306)*x[1]+(-0.136620988812)*x[2]**o+(0.10733886947)*x[2]
        arg[2,1,0]=(-0.258729405118)*x[0]**o+(0.0669553480038)*x[0]+(0.997663166613)*x[1]**o+(0.00398861063931)*x[1]+(-0.989459856803)*x[2]**o+(0.932687384325)*x[2]
        arg[2,1,1]=(-0.974431322319)*x[0]**o+(-0.486447431804)*x[0]+(0.871784873347)*x[1]**o+(-0.978904038765)*x[1]+(0.647341201506)*x[2]**o+(-0.269683556528)*x[2]
        arg[3,0,0]=(-0.369507398073)*x[0]**o+(-0.772085607074)*x[0]+(0.417550881937)*x[1]**o+(0.271743719652)*x[1]+(0.881807154737)*x[2]**o+(0.54725428039)*x[2]
        arg[3,0,1]=(0.440089335941)*x[0]**o+(0.0844171421294)*x[0]+(0.926389704362)*x[1]**o+(-0.0709182998297)*x[1]+(0.903978660317)*x[2]**o+(-0.160556206716)*x[2]
        arg[3,1,0]=(-0.49936391282)*x[0]**o+(-0.632753931881)*x[0]+(-0.604686084554)*x[1]**o+(-0.954205199531)*x[1]+(-0.535102184845)*x[2]**o+(0.621683814123)*x[2]
        arg[3,1,1]=(-0.117790973132)*x[0]**o+(-0.649091727896)*x[0]+(0.925682916785)*x[1]**o+(0.852053515469)*x[1]+(-0.479218429897)*x[2]**o+(0.426823478575)*x[2]
        arg[4,0,0]=(0.454174564319)*x[0]**o+(-0.855975647406)*x[0]+(0.775972969288)*x[1]**o+(0.267900999893)*x[1]+(0.221387501995)*x[2]**o+(-0.26413791201)*x[2]
        arg[4,0,1]=(0.875191302431)*x[0]**o+(-0.362782502152)*x[0]+(0.24600070339)*x[1]**o+(-0.566709533298)*x[1]+(-0.82187455531)*x[2]**o+(-0.943929640301)*x[2]
        arg[4,1,0]=(-0.409313259973)*x[0]**o+(-0.784833509001)*x[0]+(-0.939690505056)*x[1]**o+(-0.0736280226974)*x[1]+(0.939072745486)*x[2]**o+(0.611697159959)*x[2]
        arg[4,1,1]=(0.893374246808)*x[0]**o+(0.36113513561)*x[0]+(0.227899271271)*x[1]**o+(-0.847221657012)*x[1]+(-0.0106368947417)*x[2]**o+(0.0358896817595)*x[2]
        arg[5,0,0]=(0.0329230524924)*x[0]**o+(0.0565192084014)*x[0]+(-0.68398448658)*x[1]**o+(0.909035618076)*x[1]+(-0.900060440872)*x[2]**o+(-0.481014787833)*x[2]
        arg[5,0,1]=(0.404966692391)*x[0]**o+(0.647468764373)*x[0]+(-0.723847499346)*x[1]**o+(-0.0740896957606)*x[1]+(-0.00628515079923)*x[2]**o+(0.906667863539)*x[2]
        arg[5,1,0]=(0.983181781531)*x[0]**o+(-0.297325990236)*x[0]+(0.00489064543352)*x[1]**o+(0.667560767038)*x[1]+(-0.665768141301)*x[2]**o+(0.772761393608)*x[2]
        arg[5,1,1]=(-0.496581691354)*x[0]**o+(0.0121337979505)*x[0]+(0.492281382726)*x[1]**o+(0.545232714025)*x[1]+(0.418321481532)*x[2]**o+(0.489635969764)*x[2]
        ref[0,0,0]=(-0.179972720095)*x_ref[0]**o+(-0.546715268523)*x_ref[0]+(0.835867431319)*x_ref[1]**o+(0.626798499259)*x_ref[1]+(0.397356013278)*x_ref[2]**o+(0.0771586784158)*x_ref[2]
        ref[0,0,1]=(0.625676031039)*x_ref[0]**o+(0.65843849475)*x_ref[0]+(0.456469199307)*x_ref[1]**o+(-0.970709759479)*x_ref[1]+(-0.699606143992)*x_ref[2]**o+(0.76429400724)*x_ref[2]
        ref[0,1,0]=(0.613916424165)*x_ref[0]**o+(0.556778588303)*x_ref[0]+(-0.509658743529)*x_ref[1]**o+(-0.978718968218)*x_ref[1]+(-0.603928369428)*x_ref[2]**o+(-0.550813953692)*x_ref[2]
        ref[0,1,1]=(-0.935214755436)*x_ref[0]**o+(0.44745901813)*x_ref[0]+(-0.582828846453)*x_ref[1]**o+(-0.181547183233)*x_ref[1]+(-0.183912053485)*x_ref[2]**o+(0.446152494296)*x_ref[2]
        ref[1,0,0]=(-0.486345219779)*x_ref[0]**o+(0.753529436833)*x_ref[0]+(0.138157856552)*x_ref[1]**o+(-0.874849020539)*x_ref[1]+(0.374396543792)*x_ref[2]**o+(0.26548456046)*x_ref[2]
        ref[1,0,1]=(0.492908753036)*x_ref[0]**o+(-0.418118685105)*x_ref[0]+(0.880817569823)*x_ref[1]**o+(-0.827691145524)*x_ref[1]+(0.238633156133)*x_ref[2]**o+(-0.463687731341)*x_ref[2]
        ref[1,1,0]=(-0.397373578442)*x_ref[0]**o+(-0.316321105607)*x_ref[0]+(0.647511225459)*x_ref[1]**o+(-0.82553752195)*x_ref[1]+(-0.715549303171)*x_ref[2]**o+(0.804623957459)*x_ref[2]
        ref[1,1,1]=(0.414743130234)*x_ref[0]**o+(0.0201794413683)*x_ref[0]+(-0.636337599584)*x_ref[1]**o+(-0.230053616026)*x_ref[1]+(-0.4302153311)*x_ref[2]**o+(-0.387468029626)*x_ref[2]
        ref[2,0,0]=(0.715100517336)*x_ref[0]**o+(-0.4286120178)*x_ref[0]+(0.0987858542117)*x_ref[1]**o+(0.815093470301)*x_ref[1]+(0.293455353257)*x_ref[2]**o+(0.200576090379)*x_ref[2]
        ref[2,0,1]=(-0.158330015822)*x_ref[0]**o+(0.363339023457)*x_ref[0]+(-0.796055594971)*x_ref[1]**o+(-0.213299205306)*x_ref[1]+(-0.136620988812)*x_ref[2]**o+(0.10733886947)*x_ref[2]
        ref[2,1,0]=(-0.258729405118)*x_ref[0]**o+(0.0669553480038)*x_ref[0]+(0.997663166613)*x_ref[1]**o+(0.00398861063931)*x_ref[1]+(-0.989459856803)*x_ref[2]**o+(0.932687384325)*x_ref[2]
        ref[2,1,1]=(-0.974431322319)*x_ref[0]**o+(-0.486447431804)*x_ref[0]+(0.871784873347)*x_ref[1]**o+(-0.978904038765)*x_ref[1]+(0.647341201506)*x_ref[2]**o+(-0.269683556528)*x_ref[2]
        ref[3,0,0]=(-0.369507398073)*x_ref[0]**o+(-0.772085607074)*x_ref[0]+(0.417550881937)*x_ref[1]**o+(0.271743719652)*x_ref[1]+(0.881807154737)*x_ref[2]**o+(0.54725428039)*x_ref[2]
        ref[3,0,1]=(0.440089335941)*x_ref[0]**o+(0.0844171421294)*x_ref[0]+(0.926389704362)*x_ref[1]**o+(-0.0709182998297)*x_ref[1]+(0.903978660317)*x_ref[2]**o+(-0.160556206716)*x_ref[2]
        ref[3,1,0]=(-0.49936391282)*x_ref[0]**o+(-0.632753931881)*x_ref[0]+(-0.604686084554)*x_ref[1]**o+(-0.954205199531)*x_ref[1]+(-0.535102184845)*x_ref[2]**o+(0.621683814123)*x_ref[2]
        ref[3,1,1]=(-0.117790973132)*x_ref[0]**o+(-0.649091727896)*x_ref[0]+(0.925682916785)*x_ref[1]**o+(0.852053515469)*x_ref[1]+(-0.479218429897)*x_ref[2]**o+(0.426823478575)*x_ref[2]
        ref[4,0,0]=(0.454174564319)*x_ref[0]**o+(-0.855975647406)*x_ref[0]+(0.775972969288)*x_ref[1]**o+(0.267900999893)*x_ref[1]+(0.221387501995)*x_ref[2]**o+(-0.26413791201)*x_ref[2]
        ref[4,0,1]=(0.875191302431)*x_ref[0]**o+(-0.362782502152)*x_ref[0]+(0.24600070339)*x_ref[1]**o+(-0.566709533298)*x_ref[1]+(-0.82187455531)*x_ref[2]**o+(-0.943929640301)*x_ref[2]
        ref[4,1,0]=(-0.409313259973)*x_ref[0]**o+(-0.784833509001)*x_ref[0]+(-0.939690505056)*x_ref[1]**o+(-0.0736280226974)*x_ref[1]+(0.939072745486)*x_ref[2]**o+(0.611697159959)*x_ref[2]
        ref[4,1,1]=(0.893374246808)*x_ref[0]**o+(0.36113513561)*x_ref[0]+(0.227899271271)*x_ref[1]**o+(-0.847221657012)*x_ref[1]+(-0.0106368947417)*x_ref[2]**o+(0.0358896817595)*x_ref[2]
        ref[5,0,0]=(0.0329230524924)*x_ref[0]**o+(0.0565192084014)*x_ref[0]+(-0.68398448658)*x_ref[1]**o+(0.909035618076)*x_ref[1]+(-0.900060440872)*x_ref[2]**o+(-0.481014787833)*x_ref[2]
        ref[5,0,1]=(0.404966692391)*x_ref[0]**o+(0.647468764373)*x_ref[0]+(-0.723847499346)*x_ref[1]**o+(-0.0740896957606)*x_ref[1]+(-0.00628515079923)*x_ref[2]**o+(0.906667863539)*x_ref[2]
        ref[5,1,0]=(0.983181781531)*x_ref[0]**o+(-0.297325990236)*x_ref[0]+(0.00489064543352)*x_ref[1]**o+(0.667560767038)*x_ref[1]+(-0.665768141301)*x_ref[2]**o+(0.772761393608)*x_ref[2]
        ref[5,1,1]=(-0.496581691354)*x_ref[0]**o+(0.0121337979505)*x_ref[0]+(0.492281382726)*x_ref[1]**o+(0.545232714025)*x_ref[1]+(0.418321481532)*x_ref[2]**o+(0.489635969764)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactZero_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactZero

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactZero
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactZero(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.282394950702)*x[0]**o+(0.836580157584)*x[0]+(0.337652262648)*x[1]**o+(-0.442694263325)*x[1]
        arg[0,0,0,1]=(0.85150774028)*x[0]**o+(0.998297296593)*x[0]+(-0.710029954342)*x[1]**o+(0.0816168287837)*x[1]
        arg[0,0,1,0]=(-0.636435102838)*x[0]**o+(-0.613479863553)*x[0]+(-0.861506513168)*x[1]**o+(-0.242998961478)*x[1]
        arg[0,0,1,1]=(-0.00671863978521)*x[0]**o+(0.909404084068)*x[0]+(-0.734956067026)*x[1]**o+(0.256530538264)*x[1]
        arg[0,0,2,0]=(-0.646631608999)*x[0]**o+(-0.521130455843)*x[0]+(-0.44328404557)*x[1]**o+(-0.156860325817)*x[1]
        arg[0,0,2,1]=(-0.817550008949)*x[0]**o+(0.702414497795)*x[0]+(0.946449680763)*x[1]**o+(-0.61576361925)*x[1]
        arg[0,1,0,0]=(-0.527206691595)*x[0]**o+(-0.43469216984)*x[0]+(-0.108455901154)*x[1]**o+(0.435646962156)*x[1]
        arg[0,1,0,1]=(-0.148522219535)*x[0]**o+(0.146912249755)*x[0]+(-0.414934945665)*x[1]**o+(-0.422754988714)*x[1]
        arg[0,1,1,0]=(0.177760972947)*x[0]**o+(-0.938522530862)*x[0]+(0.566229652264)*x[1]**o+(0.255269743628)*x[1]
        arg[0,1,1,1]=(-0.317807933792)*x[0]**o+(0.432181451246)*x[0]+(-0.358821513362)*x[1]**o+(-0.470517232585)*x[1]
        arg[0,1,2,0]=(0.486536430227)*x[0]**o+(-0.533653809349)*x[0]+(-0.126612514045)*x[1]**o+(0.103088115487)*x[1]
        arg[0,1,2,1]=(-0.995976559589)*x[0]**o+(0.965537167421)*x[0]+(-0.351307699531)*x[1]**o+(-0.684367434666)*x[1]
        arg[0,2,0,0]=(0.976008613768)*x[0]**o+(-0.890573007002)*x[0]+(-0.440709423944)*x[1]**o+(0.315724459367)*x[1]
        arg[0,2,0,1]=(-0.595275811962)*x[0]**o+(-0.599634144509)*x[0]+(0.646466002346)*x[1]**o+(-0.94529568651)*x[1]
        arg[0,2,1,0]=(0.977691521821)*x[0]**o+(-0.0730997643682)*x[0]+(-0.338897498737)*x[1]**o+(0.246308678824)*x[1]
        arg[0,2,1,1]=(-0.111317410732)*x[0]**o+(-0.574381793398)*x[0]+(-0.0314431899843)*x[1]**o+(0.393105047096)*x[1]
        arg[0,2,2,0]=(0.886320020757)*x[0]**o+(-0.785042003544)*x[0]+(0.82609269979)*x[1]**o+(-0.677169484337)*x[1]
        arg[0,2,2,1]=(0.811027138079)*x[0]**o+(-0.375108470948)*x[0]+(0.781472464253)*x[1]**o+(0.217640794143)*x[1]
        arg[0,3,0,0]=(0.264884117091)*x[0]**o+(0.116591319557)*x[0]+(0.282667679763)*x[1]**o+(0.111123454337)*x[1]
        arg[0,3,0,1]=(0.0451626985922)*x[0]**o+(-0.93771279728)*x[0]+(0.129171764072)*x[1]**o+(0.598367954162)*x[1]
        arg[0,3,1,0]=(0.905243511086)*x[0]**o+(0.0837062573586)*x[0]+(0.0657223248433)*x[1]**o+(-0.746644690072)*x[1]
        arg[0,3,1,1]=(-0.0547856755485)*x[0]**o+(-0.22157806755)*x[0]+(0.124623625108)*x[1]**o+(0.0436590157081)*x[1]
        arg[0,3,2,0]=(-0.558323744577)*x[0]**o+(-0.646043519007)*x[0]+(0.991885458004)*x[1]**o+(-0.634140808163)*x[1]
        arg[0,3,2,1]=(-0.27247983028)*x[0]**o+(0.803906404485)*x[0]+(-0.816826609924)*x[1]**o+(-0.969532783336)*x[1]
        arg[0,4,0,0]=(0.638733842891)*x[0]**o+(-0.972793270208)*x[0]+(0.714611008099)*x[1]**o+(0.586988595847)*x[1]
        arg[0,4,0,1]=(0.63584760734)*x[0]**o+(-0.653740236747)*x[0]+(-0.596304731446)*x[1]**o+(0.42519786429)*x[1]
        arg[0,4,1,0]=(0.0665562973529)*x[0]**o+(0.362152092093)*x[0]+(0.732530335739)*x[1]**o+(0.618536326941)*x[1]
        arg[0,4,1,1]=(0.486060667608)*x[0]**o+(0.59489013279)*x[0]+(-0.547237017498)*x[1]**o+(0.0371894705857)*x[1]
        arg[0,4,2,0]=(0.395505844397)*x[0]**o+(-0.869090434535)*x[0]+(-0.606640827437)*x[1]**o+(0.269995459083)*x[1]
        arg[0,4,2,1]=(0.304671547176)*x[0]**o+(-0.515772523849)*x[0]+(0.546114300876)*x[1]**o+(0.530247404341)*x[1]
        arg[1,0,0,0]=(0.408332677433)*x[0]**o+(-0.0778807359918)*x[0]+(-0.803337489297)*x[1]**o+(-0.123835718422)*x[1]
        arg[1,0,0,1]=(-0.00441620254944)*x[0]**o+(0.0321259598831)*x[0]+(-0.998381508472)*x[1]**o+(0.800767747511)*x[1]
        arg[1,0,1,0]=(-0.923683057418)*x[0]**o+(0.579405114902)*x[0]+(-0.939639823585)*x[1]**o+(-0.138423358134)*x[1]
        arg[1,0,1,1]=(-0.558420234848)*x[0]**o+(0.133316569813)*x[0]+(-0.294134717329)*x[1]**o+(-0.750897749044)*x[1]
        arg[1,0,2,0]=(0.478951298545)*x[0]**o+(-0.807173978799)*x[0]+(0.852438146763)*x[1]**o+(-0.103814229563)*x[1]
        arg[1,0,2,1]=(-0.875250726603)*x[0]**o+(-0.934803428491)*x[0]+(0.143352873672)*x[1]**o+(0.600934881206)*x[1]
        arg[1,1,0,0]=(-0.0539476321018)*x[0]**o+(0.194177663417)*x[0]+(-0.862877991009)*x[1]**o+(-0.438337993664)*x[1]
        arg[1,1,0,1]=(-0.623379355474)*x[0]**o+(-0.0746994739748)*x[0]+(0.365005438956)*x[1]**o+(-0.431737688205)*x[1]
        arg[1,1,1,0]=(0.0936054461735)*x[0]**o+(-0.974594269446)*x[0]+(-0.121917598428)*x[1]**o+(-0.84429083727)*x[1]
        arg[1,1,1,1]=(0.967763348672)*x[0]**o+(0.331698826475)*x[0]+(0.66448515877)*x[1]**o+(0.00214985930661)*x[1]
        arg[1,1,2,0]=(-0.312146580762)*x[0]**o+(0.998152287706)*x[0]+(-0.474242942777)*x[1]**o+(0.637827347846)*x[1]
        arg[1,1,2,1]=(-0.556765443436)*x[0]**o+(0.227327929774)*x[0]+(0.739457845569)*x[1]**o+(-0.615070017663)*x[1]
        arg[1,2,0,0]=(-0.23030660552)*x[0]**o+(0.750479465953)*x[0]+(0.278215408392)*x[1]**o+(-0.640618698224)*x[1]
        arg[1,2,0,1]=(-0.0713217101129)*x[0]**o+(0.500737477887)*x[0]+(-0.972834770005)*x[1]**o+(-0.420386056033)*x[1]
        arg[1,2,1,0]=(-0.497322989847)*x[0]**o+(-0.528685124841)*x[0]+(-0.934386348496)*x[1]**o+(0.191930553839)*x[1]
        arg[1,2,1,1]=(0.124362419781)*x[0]**o+(-0.0457849787456)*x[0]+(-0.360596154712)*x[1]**o+(0.799736253715)*x[1]
        arg[1,2,2,0]=(-0.346294878893)*x[0]**o+(0.427772309361)*x[0]+(-0.0642666791012)*x[1]**o+(-0.214442238311)*x[1]
        arg[1,2,2,1]=(0.24434837411)*x[0]**o+(-0.913373159786)*x[0]+(0.158750862391)*x[1]**o+(-0.544328807576)*x[1]
        arg[1,3,0,0]=(0.870847002687)*x[0]**o+(-0.466884462301)*x[0]+(-0.343358829445)*x[1]**o+(-0.180956419766)*x[1]
        arg[1,3,0,1]=(-0.813518850863)*x[0]**o+(0.960810467501)*x[0]+(0.35091416904)*x[1]**o+(0.84534544948)*x[1]
        arg[1,3,1,0]=(-0.284687751218)*x[0]**o+(-0.274583519355)*x[0]+(0.338892764102)*x[1]**o+(-0.145142443753)*x[1]
        arg[1,3,1,1]=(0.921064262137)*x[0]**o+(0.152552421678)*x[0]+(0.827247693375)*x[1]**o+(-0.569529848619)*x[1]
        arg[1,3,2,0]=(0.133567630356)*x[0]**o+(-0.0429397358364)*x[0]+(-0.175383170663)*x[1]**o+(0.584407631334)*x[1]
        arg[1,3,2,1]=(0.203880167371)*x[0]**o+(-0.422978073825)*x[0]+(-0.210109267588)*x[1]**o+(-0.902327373219)*x[1]
        arg[1,4,0,0]=(-0.566386864876)*x[0]**o+(-0.139795085483)*x[0]+(-0.00852913464238)*x[1]**o+(0.569151381884)*x[1]
        arg[1,4,0,1]=(-0.937256625212)*x[0]**o+(0.563438253128)*x[0]+(-0.586182238715)*x[1]**o+(-0.566169678723)*x[1]
        arg[1,4,1,0]=(-0.73296890712)*x[0]**o+(0.359249594406)*x[0]+(0.66050790616)*x[1]**o+(-0.367919797426)*x[1]
        arg[1,4,1,1]=(-0.81165734248)*x[0]**o+(0.460592817838)*x[0]+(0.953573909492)*x[1]**o+(0.329659021274)*x[1]
        arg[1,4,2,0]=(0.345994483595)*x[0]**o+(-0.18547324343)*x[0]+(0.669452926238)*x[1]**o+(0.42513249272)*x[1]
        arg[1,4,2,1]=(0.250677642616)*x[0]**o+(0.466118496803)*x[0]+(-0.776608543595)*x[1]**o+(0.449144243682)*x[1]
        arg[2,0,0,0]=(0.244172909021)*x[0]**o+(0.0311068322898)*x[0]+(-0.909348039508)*x[1]**o+(0.376500719414)*x[1]
        arg[2,0,0,1]=(-0.033516640667)*x[0]**o+(0.698424718328)*x[0]+(-0.282814785542)*x[1]**o+(0.61118022947)*x[1]
        arg[2,0,1,0]=(-0.946672649538)*x[0]**o+(0.814325244591)*x[0]+(-0.353215562825)*x[1]**o+(-0.138150257429)*x[1]
        arg[2,0,1,1]=(-0.886509051692)*x[0]**o+(-0.0468334385819)*x[0]+(0.71984719189)*x[1]**o+(-0.324744971493)*x[1]
        arg[2,0,2,0]=(-0.780921647053)*x[0]**o+(-0.401418496932)*x[0]+(0.647847274332)*x[1]**o+(-0.831834322278)*x[1]
        arg[2,0,2,1]=(0.835155004862)*x[0]**o+(0.684459986902)*x[0]+(-0.301244240677)*x[1]**o+(-0.161292780187)*x[1]
        arg[2,1,0,0]=(-0.652199398394)*x[0]**o+(0.62480722178)*x[0]+(-0.474535712914)*x[1]**o+(0.926061797766)*x[1]
        arg[2,1,0,1]=(-0.813365478568)*x[0]**o+(-0.38795792417)*x[0]+(0.378161353697)*x[1]**o+(0.70581806211)*x[1]
        arg[2,1,1,0]=(0.729265594934)*x[0]**o+(-0.469397842359)*x[0]+(0.642472199926)*x[1]**o+(-0.779527283011)*x[1]
        arg[2,1,1,1]=(0.0517342917574)*x[0]**o+(-0.414111007917)*x[0]+(-0.89437375241)*x[1]**o+(-0.830119585842)*x[1]
        arg[2,1,2,0]=(0.769898486257)*x[0]**o+(0.791725624832)*x[0]+(0.297519329396)*x[1]**o+(-0.959462006611)*x[1]
        arg[2,1,2,1]=(0.817802144251)*x[0]**o+(0.69851258911)*x[0]+(-0.174401735228)*x[1]**o+(0.13449696498)*x[1]
        arg[2,2,0,0]=(0.547972662256)*x[0]**o+(-0.66491929979)*x[0]+(0.330783041672)*x[1]**o+(0.816196730606)*x[1]
        arg[2,2,0,1]=(0.245971097763)*x[0]**o+(0.213087481481)*x[0]+(-0.840766291392)*x[1]**o+(0.388414557376)*x[1]
        arg[2,2,1,0]=(0.523151676444)*x[0]**o+(0.854817486027)*x[0]+(-0.53570188528)*x[1]**o+(-0.68149362242)*x[1]
        arg[2,2,1,1]=(-0.984200247787)*x[0]**o+(0.83324365092)*x[0]+(-0.896184408816)*x[1]**o+(0.102276243487)*x[1]
        arg[2,2,2,0]=(-0.216357164308)*x[0]**o+(-0.597984882107)*x[0]+(-0.205435921511)*x[1]**o+(0.484947652759)*x[1]
        arg[2,2,2,1]=(-0.650046253868)*x[0]**o+(-0.624878186105)*x[0]+(0.786519914406)*x[1]**o+(0.28027273095)*x[1]
        arg[2,3,0,0]=(-0.0821718346579)*x[0]**o+(-0.687049917467)*x[0]+(0.380792333736)*x[1]**o+(0.151185837715)*x[1]
        arg[2,3,0,1]=(0.61796113854)*x[0]**o+(0.844624413065)*x[0]+(-0.980619401098)*x[1]**o+(-0.854924862641)*x[1]
        arg[2,3,1,0]=(0.812293906883)*x[0]**o+(-0.225593208768)*x[0]+(0.964175374163)*x[1]**o+(0.848850064926)*x[1]
        arg[2,3,1,1]=(0.376453572937)*x[0]**o+(0.994818812059)*x[0]+(0.58872018303)*x[1]**o+(0.68332473361)*x[1]
        arg[2,3,2,0]=(-0.00948421972234)*x[0]**o+(-0.680685462738)*x[0]+(-0.767076278167)*x[1]**o+(0.210165437284)*x[1]
        arg[2,3,2,1]=(-0.455199698263)*x[0]**o+(-0.763430401419)*x[0]+(0.183773350212)*x[1]**o+(0.43105940923)*x[1]
        arg[2,4,0,0]=(-0.673962025988)*x[0]**o+(-0.679073069725)*x[0]+(0.11018431715)*x[1]**o+(0.349020527482)*x[1]
        arg[2,4,0,1]=(-0.876691282076)*x[0]**o+(0.920902226631)*x[0]+(-0.477417470246)*x[1]**o+(0.836417982216)*x[1]
        arg[2,4,1,0]=(0.567966851024)*x[0]**o+(0.583241431716)*x[0]+(-0.271851271008)*x[1]**o+(0.678727902863)*x[1]
        arg[2,4,1,1]=(0.707400094)*x[0]**o+(0.802820856444)*x[0]+(0.84518937616)*x[1]**o+(-0.0267193533403)*x[1]
        arg[2,4,2,0]=(0.213268920036)*x[0]**o+(0.780721415328)*x[0]+(0.839292948154)*x[1]**o+(-0.0376334744443)*x[1]
        arg[2,4,2,1]=(0.136255687031)*x[0]**o+(-0.582766278693)*x[0]+(0.257662521321)*x[1]**o+(0.808337521979)*x[1]
        arg[3,0,0,0]=(-0.524068770822)*x[0]**o+(0.942236878842)*x[0]+(0.889496255793)*x[1]**o+(-0.904810742393)*x[1]
        arg[3,0,0,1]=(0.780157465475)*x[0]**o+(-0.763811668832)*x[0]+(-0.517482070639)*x[1]**o+(0.662770780676)*x[1]
        arg[3,0,1,0]=(-0.693227564812)*x[0]**o+(-0.763506211841)*x[0]+(-0.427709881126)*x[1]**o+(0.159109827204)*x[1]
        arg[3,0,1,1]=(0.644975629084)*x[0]**o+(-0.928112761899)*x[0]+(-0.530772740083)*x[1]**o+(-0.723174790087)*x[1]
        arg[3,0,2,0]=(-0.923467274097)*x[0]**o+(0.806128117288)*x[0]+(0.177779363207)*x[1]**o+(0.862243885542)*x[1]
        arg[3,0,2,1]=(-0.0823252901358)*x[0]**o+(0.289022916284)*x[0]+(-0.518508771021)*x[1]**o+(0.174564769845)*x[1]
        arg[3,1,0,0]=(0.608837578444)*x[0]**o+(0.898216758361)*x[0]+(-0.844730109761)*x[1]**o+(0.745974173207)*x[1]
        arg[3,1,0,1]=(-0.11776411078)*x[0]**o+(0.186256502356)*x[0]+(0.565289400333)*x[1]**o+(-0.776173716206)*x[1]
        arg[3,1,1,0]=(-0.170736580152)*x[0]**o+(0.829581661683)*x[0]+(0.310135833859)*x[1]**o+(-0.509629545687)*x[1]
        arg[3,1,1,1]=(-0.225246122751)*x[0]**o+(0.586754377294)*x[0]+(-0.726064725635)*x[1]**o+(0.136952605704)*x[1]
        arg[3,1,2,0]=(-0.287932115478)*x[0]**o+(0.858031396153)*x[0]+(-0.493441745353)*x[1]**o+(0.601827664994)*x[1]
        arg[3,1,2,1]=(0.665503842862)*x[0]**o+(-0.515998970452)*x[0]+(-0.890055558624)*x[1]**o+(-0.0446371451457)*x[1]
        arg[3,2,0,0]=(-0.33741557585)*x[0]**o+(-0.416183752333)*x[0]+(-0.92430073538)*x[1]**o+(-0.486265090904)*x[1]
        arg[3,2,0,1]=(0.910860925994)*x[0]**o+(-0.896869755389)*x[0]+(0.0634658185324)*x[1]**o+(-0.383136455739)*x[1]
        arg[3,2,1,0]=(0.386839297976)*x[0]**o+(-0.190659927071)*x[0]+(-0.304846978185)*x[1]**o+(-0.92705511899)*x[1]
        arg[3,2,1,1]=(-0.0197812018683)*x[0]**o+(-0.479083095775)*x[0]+(0.502489976089)*x[1]**o+(-0.859440665475)*x[1]
        arg[3,2,2,0]=(-0.507672045299)*x[0]**o+(0.498689676072)*x[0]+(-0.719735804645)*x[1]**o+(-0.522053306674)*x[1]
        arg[3,2,2,1]=(0.46207629211)*x[0]**o+(-0.496697066039)*x[0]+(0.994069339904)*x[1]**o+(0.873222754449)*x[1]
        arg[3,3,0,0]=(0.126410029692)*x[0]**o+(0.51067050196)*x[0]+(-0.978384824088)*x[1]**o+(-0.68606930503)*x[1]
        arg[3,3,0,1]=(0.376888157008)*x[0]**o+(-0.14015764025)*x[0]+(0.523667700758)*x[1]**o+(0.0277726855258)*x[1]
        arg[3,3,1,0]=(-0.105865233294)*x[0]**o+(-0.0515230741203)*x[0]+(0.203766525036)*x[1]**o+(-0.0431893014387)*x[1]
        arg[3,3,1,1]=(-0.404351448486)*x[0]**o+(-0.639161429157)*x[0]+(-0.957341091708)*x[1]**o+(-0.103346466984)*x[1]
        arg[3,3,2,0]=(-0.0523508780623)*x[0]**o+(-0.563589575544)*x[0]+(0.158487727325)*x[1]**o+(-0.519455492178)*x[1]
        arg[3,3,2,1]=(-0.823286342218)*x[0]**o+(-0.493663499535)*x[0]+(-0.409825574064)*x[1]**o+(-0.0406101279806)*x[1]
        arg[3,4,0,0]=(0.926212910387)*x[0]**o+(-0.0944507963)*x[0]+(0.640740499391)*x[1]**o+(-0.106969076047)*x[1]
        arg[3,4,0,1]=(-0.912044993941)*x[0]**o+(-0.495323777918)*x[0]+(-0.595831845108)*x[1]**o+(-0.313723426318)*x[1]
        arg[3,4,1,0]=(-0.0730898742548)*x[0]**o+(0.0524008484492)*x[0]+(-0.948232935497)*x[1]**o+(0.952670824572)*x[1]
        arg[3,4,1,1]=(-0.81869438138)*x[0]**o+(-0.894654266952)*x[0]+(0.414078621872)*x[1]**o+(-0.876690817847)*x[1]
        arg[3,4,2,0]=(0.00688155078441)*x[0]**o+(0.845964586551)*x[0]+(0.0183149371892)*x[1]**o+(-0.882931394537)*x[1]
        arg[3,4,2,1]=(-0.734830097149)*x[0]**o+(-0.37824588859)*x[0]+(-0.815402620311)*x[1]**o+(-0.357905685517)*x[1]
        ref[0,0,0,0]=(-0.282394950702)*x_ref[0]**o+(0.836580157584)*x_ref[0]+(0.337652262648)*x_ref[1]**o+(-0.442694263325)*x_ref[1]
        ref[0,0,0,1]=(0.85150774028)*x_ref[0]**o+(0.998297296593)*x_ref[0]+(-0.710029954342)*x_ref[1]**o+(0.0816168287837)*x_ref[1]
        ref[0,0,1,0]=(-0.636435102838)*x_ref[0]**o+(-0.613479863553)*x_ref[0]+(-0.861506513168)*x_ref[1]**o+(-0.242998961478)*x_ref[1]
        ref[0,0,1,1]=(-0.00671863978521)*x_ref[0]**o+(0.909404084068)*x_ref[0]+(-0.734956067026)*x_ref[1]**o+(0.256530538264)*x_ref[1]
        ref[0,0,2,0]=(-0.646631608999)*x_ref[0]**o+(-0.521130455843)*x_ref[0]+(-0.44328404557)*x_ref[1]**o+(-0.156860325817)*x_ref[1]
        ref[0,0,2,1]=(-0.817550008949)*x_ref[0]**o+(0.702414497795)*x_ref[0]+(0.946449680763)*x_ref[1]**o+(-0.61576361925)*x_ref[1]
        ref[0,1,0,0]=(-0.527206691595)*x_ref[0]**o+(-0.43469216984)*x_ref[0]+(-0.108455901154)*x_ref[1]**o+(0.435646962156)*x_ref[1]
        ref[0,1,0,1]=(-0.148522219535)*x_ref[0]**o+(0.146912249755)*x_ref[0]+(-0.414934945665)*x_ref[1]**o+(-0.422754988714)*x_ref[1]
        ref[0,1,1,0]=(0.177760972947)*x_ref[0]**o+(-0.938522530862)*x_ref[0]+(0.566229652264)*x_ref[1]**o+(0.255269743628)*x_ref[1]
        ref[0,1,1,1]=(-0.317807933792)*x_ref[0]**o+(0.432181451246)*x_ref[0]+(-0.358821513362)*x_ref[1]**o+(-0.470517232585)*x_ref[1]
        ref[0,1,2,0]=(0.486536430227)*x_ref[0]**o+(-0.533653809349)*x_ref[0]+(-0.126612514045)*x_ref[1]**o+(0.103088115487)*x_ref[1]
        ref[0,1,2,1]=(-0.995976559589)*x_ref[0]**o+(0.965537167421)*x_ref[0]+(-0.351307699531)*x_ref[1]**o+(-0.684367434666)*x_ref[1]
        ref[0,2,0,0]=(0.976008613768)*x_ref[0]**o+(-0.890573007002)*x_ref[0]+(-0.440709423944)*x_ref[1]**o+(0.315724459367)*x_ref[1]
        ref[0,2,0,1]=(-0.595275811962)*x_ref[0]**o+(-0.599634144509)*x_ref[0]+(0.646466002346)*x_ref[1]**o+(-0.94529568651)*x_ref[1]
        ref[0,2,1,0]=(0.977691521821)*x_ref[0]**o+(-0.0730997643682)*x_ref[0]+(-0.338897498737)*x_ref[1]**o+(0.246308678824)*x_ref[1]
        ref[0,2,1,1]=(-0.111317410732)*x_ref[0]**o+(-0.574381793398)*x_ref[0]+(-0.0314431899843)*x_ref[1]**o+(0.393105047096)*x_ref[1]
        ref[0,2,2,0]=(0.886320020757)*x_ref[0]**o+(-0.785042003544)*x_ref[0]+(0.82609269979)*x_ref[1]**o+(-0.677169484337)*x_ref[1]
        ref[0,2,2,1]=(0.811027138079)*x_ref[0]**o+(-0.375108470948)*x_ref[0]+(0.781472464253)*x_ref[1]**o+(0.217640794143)*x_ref[1]
        ref[0,3,0,0]=(0.264884117091)*x_ref[0]**o+(0.116591319557)*x_ref[0]+(0.282667679763)*x_ref[1]**o+(0.111123454337)*x_ref[1]
        ref[0,3,0,1]=(0.0451626985922)*x_ref[0]**o+(-0.93771279728)*x_ref[0]+(0.129171764072)*x_ref[1]**o+(0.598367954162)*x_ref[1]
        ref[0,3,1,0]=(0.905243511086)*x_ref[0]**o+(0.0837062573586)*x_ref[0]+(0.0657223248433)*x_ref[1]**o+(-0.746644690072)*x_ref[1]
        ref[0,3,1,1]=(-0.0547856755485)*x_ref[0]**o+(-0.22157806755)*x_ref[0]+(0.124623625108)*x_ref[1]**o+(0.0436590157081)*x_ref[1]
        ref[0,3,2,0]=(-0.558323744577)*x_ref[0]**o+(-0.646043519007)*x_ref[0]+(0.991885458004)*x_ref[1]**o+(-0.634140808163)*x_ref[1]
        ref[0,3,2,1]=(-0.27247983028)*x_ref[0]**o+(0.803906404485)*x_ref[0]+(-0.816826609924)*x_ref[1]**o+(-0.969532783336)*x_ref[1]
        ref[0,4,0,0]=(0.638733842891)*x_ref[0]**o+(-0.972793270208)*x_ref[0]+(0.714611008099)*x_ref[1]**o+(0.586988595847)*x_ref[1]
        ref[0,4,0,1]=(0.63584760734)*x_ref[0]**o+(-0.653740236747)*x_ref[0]+(-0.596304731446)*x_ref[1]**o+(0.42519786429)*x_ref[1]
        ref[0,4,1,0]=(0.0665562973529)*x_ref[0]**o+(0.362152092093)*x_ref[0]+(0.732530335739)*x_ref[1]**o+(0.618536326941)*x_ref[1]
        ref[0,4,1,1]=(0.486060667608)*x_ref[0]**o+(0.59489013279)*x_ref[0]+(-0.547237017498)*x_ref[1]**o+(0.0371894705857)*x_ref[1]
        ref[0,4,2,0]=(0.395505844397)*x_ref[0]**o+(-0.869090434535)*x_ref[0]+(-0.606640827437)*x_ref[1]**o+(0.269995459083)*x_ref[1]
        ref[0,4,2,1]=(0.304671547176)*x_ref[0]**o+(-0.515772523849)*x_ref[0]+(0.546114300876)*x_ref[1]**o+(0.530247404341)*x_ref[1]
        ref[1,0,0,0]=(0.408332677433)*x_ref[0]**o+(-0.0778807359918)*x_ref[0]+(-0.803337489297)*x_ref[1]**o+(-0.123835718422)*x_ref[1]
        ref[1,0,0,1]=(-0.00441620254944)*x_ref[0]**o+(0.0321259598831)*x_ref[0]+(-0.998381508472)*x_ref[1]**o+(0.800767747511)*x_ref[1]
        ref[1,0,1,0]=(-0.923683057418)*x_ref[0]**o+(0.579405114902)*x_ref[0]+(-0.939639823585)*x_ref[1]**o+(-0.138423358134)*x_ref[1]
        ref[1,0,1,1]=(-0.558420234848)*x_ref[0]**o+(0.133316569813)*x_ref[0]+(-0.294134717329)*x_ref[1]**o+(-0.750897749044)*x_ref[1]
        ref[1,0,2,0]=(0.478951298545)*x_ref[0]**o+(-0.807173978799)*x_ref[0]+(0.852438146763)*x_ref[1]**o+(-0.103814229563)*x_ref[1]
        ref[1,0,2,1]=(-0.875250726603)*x_ref[0]**o+(-0.934803428491)*x_ref[0]+(0.143352873672)*x_ref[1]**o+(0.600934881206)*x_ref[1]
        ref[1,1,0,0]=(-0.0539476321018)*x_ref[0]**o+(0.194177663417)*x_ref[0]+(-0.862877991009)*x_ref[1]**o+(-0.438337993664)*x_ref[1]
        ref[1,1,0,1]=(-0.623379355474)*x_ref[0]**o+(-0.0746994739748)*x_ref[0]+(0.365005438956)*x_ref[1]**o+(-0.431737688205)*x_ref[1]
        ref[1,1,1,0]=(0.0936054461735)*x_ref[0]**o+(-0.974594269446)*x_ref[0]+(-0.121917598428)*x_ref[1]**o+(-0.84429083727)*x_ref[1]
        ref[1,1,1,1]=(0.967763348672)*x_ref[0]**o+(0.331698826475)*x_ref[0]+(0.66448515877)*x_ref[1]**o+(0.00214985930661)*x_ref[1]
        ref[1,1,2,0]=(-0.312146580762)*x_ref[0]**o+(0.998152287706)*x_ref[0]+(-0.474242942777)*x_ref[1]**o+(0.637827347846)*x_ref[1]
        ref[1,1,2,1]=(-0.556765443436)*x_ref[0]**o+(0.227327929774)*x_ref[0]+(0.739457845569)*x_ref[1]**o+(-0.615070017663)*x_ref[1]
        ref[1,2,0,0]=(-0.23030660552)*x_ref[0]**o+(0.750479465953)*x_ref[0]+(0.278215408392)*x_ref[1]**o+(-0.640618698224)*x_ref[1]
        ref[1,2,0,1]=(-0.0713217101129)*x_ref[0]**o+(0.500737477887)*x_ref[0]+(-0.972834770005)*x_ref[1]**o+(-0.420386056033)*x_ref[1]
        ref[1,2,1,0]=(-0.497322989847)*x_ref[0]**o+(-0.528685124841)*x_ref[0]+(-0.934386348496)*x_ref[1]**o+(0.191930553839)*x_ref[1]
        ref[1,2,1,1]=(0.124362419781)*x_ref[0]**o+(-0.0457849787456)*x_ref[0]+(-0.360596154712)*x_ref[1]**o+(0.799736253715)*x_ref[1]
        ref[1,2,2,0]=(-0.346294878893)*x_ref[0]**o+(0.427772309361)*x_ref[0]+(-0.0642666791012)*x_ref[1]**o+(-0.214442238311)*x_ref[1]
        ref[1,2,2,1]=(0.24434837411)*x_ref[0]**o+(-0.913373159786)*x_ref[0]+(0.158750862391)*x_ref[1]**o+(-0.544328807576)*x_ref[1]
        ref[1,3,0,0]=(0.870847002687)*x_ref[0]**o+(-0.466884462301)*x_ref[0]+(-0.343358829445)*x_ref[1]**o+(-0.180956419766)*x_ref[1]
        ref[1,3,0,1]=(-0.813518850863)*x_ref[0]**o+(0.960810467501)*x_ref[0]+(0.35091416904)*x_ref[1]**o+(0.84534544948)*x_ref[1]
        ref[1,3,1,0]=(-0.284687751218)*x_ref[0]**o+(-0.274583519355)*x_ref[0]+(0.338892764102)*x_ref[1]**o+(-0.145142443753)*x_ref[1]
        ref[1,3,1,1]=(0.921064262137)*x_ref[0]**o+(0.152552421678)*x_ref[0]+(0.827247693375)*x_ref[1]**o+(-0.569529848619)*x_ref[1]
        ref[1,3,2,0]=(0.133567630356)*x_ref[0]**o+(-0.0429397358364)*x_ref[0]+(-0.175383170663)*x_ref[1]**o+(0.584407631334)*x_ref[1]
        ref[1,3,2,1]=(0.203880167371)*x_ref[0]**o+(-0.422978073825)*x_ref[0]+(-0.210109267588)*x_ref[1]**o+(-0.902327373219)*x_ref[1]
        ref[1,4,0,0]=(-0.566386864876)*x_ref[0]**o+(-0.139795085483)*x_ref[0]+(-0.00852913464238)*x_ref[1]**o+(0.569151381884)*x_ref[1]
        ref[1,4,0,1]=(-0.937256625212)*x_ref[0]**o+(0.563438253128)*x_ref[0]+(-0.586182238715)*x_ref[1]**o+(-0.566169678723)*x_ref[1]
        ref[1,4,1,0]=(-0.73296890712)*x_ref[0]**o+(0.359249594406)*x_ref[0]+(0.66050790616)*x_ref[1]**o+(-0.367919797426)*x_ref[1]
        ref[1,4,1,1]=(-0.81165734248)*x_ref[0]**o+(0.460592817838)*x_ref[0]+(0.953573909492)*x_ref[1]**o+(0.329659021274)*x_ref[1]
        ref[1,4,2,0]=(0.345994483595)*x_ref[0]**o+(-0.18547324343)*x_ref[0]+(0.669452926238)*x_ref[1]**o+(0.42513249272)*x_ref[1]
        ref[1,4,2,1]=(0.250677642616)*x_ref[0]**o+(0.466118496803)*x_ref[0]+(-0.776608543595)*x_ref[1]**o+(0.449144243682)*x_ref[1]
        ref[2,0,0,0]=(0.244172909021)*x_ref[0]**o+(0.0311068322898)*x_ref[0]+(-0.909348039508)*x_ref[1]**o+(0.376500719414)*x_ref[1]
        ref[2,0,0,1]=(-0.033516640667)*x_ref[0]**o+(0.698424718328)*x_ref[0]+(-0.282814785542)*x_ref[1]**o+(0.61118022947)*x_ref[1]
        ref[2,0,1,0]=(-0.946672649538)*x_ref[0]**o+(0.814325244591)*x_ref[0]+(-0.353215562825)*x_ref[1]**o+(-0.138150257429)*x_ref[1]
        ref[2,0,1,1]=(-0.886509051692)*x_ref[0]**o+(-0.0468334385819)*x_ref[0]+(0.71984719189)*x_ref[1]**o+(-0.324744971493)*x_ref[1]
        ref[2,0,2,0]=(-0.780921647053)*x_ref[0]**o+(-0.401418496932)*x_ref[0]+(0.647847274332)*x_ref[1]**o+(-0.831834322278)*x_ref[1]
        ref[2,0,2,1]=(0.835155004862)*x_ref[0]**o+(0.684459986902)*x_ref[0]+(-0.301244240677)*x_ref[1]**o+(-0.161292780187)*x_ref[1]
        ref[2,1,0,0]=(-0.652199398394)*x_ref[0]**o+(0.62480722178)*x_ref[0]+(-0.474535712914)*x_ref[1]**o+(0.926061797766)*x_ref[1]
        ref[2,1,0,1]=(-0.813365478568)*x_ref[0]**o+(-0.38795792417)*x_ref[0]+(0.378161353697)*x_ref[1]**o+(0.70581806211)*x_ref[1]
        ref[2,1,1,0]=(0.729265594934)*x_ref[0]**o+(-0.469397842359)*x_ref[0]+(0.642472199926)*x_ref[1]**o+(-0.779527283011)*x_ref[1]
        ref[2,1,1,1]=(0.0517342917574)*x_ref[0]**o+(-0.414111007917)*x_ref[0]+(-0.89437375241)*x_ref[1]**o+(-0.830119585842)*x_ref[1]
        ref[2,1,2,0]=(0.769898486257)*x_ref[0]**o+(0.791725624832)*x_ref[0]+(0.297519329396)*x_ref[1]**o+(-0.959462006611)*x_ref[1]
        ref[2,1,2,1]=(0.817802144251)*x_ref[0]**o+(0.69851258911)*x_ref[0]+(-0.174401735228)*x_ref[1]**o+(0.13449696498)*x_ref[1]
        ref[2,2,0,0]=(0.547972662256)*x_ref[0]**o+(-0.66491929979)*x_ref[0]+(0.330783041672)*x_ref[1]**o+(0.816196730606)*x_ref[1]
        ref[2,2,0,1]=(0.245971097763)*x_ref[0]**o+(0.213087481481)*x_ref[0]+(-0.840766291392)*x_ref[1]**o+(0.388414557376)*x_ref[1]
        ref[2,2,1,0]=(0.523151676444)*x_ref[0]**o+(0.854817486027)*x_ref[0]+(-0.53570188528)*x_ref[1]**o+(-0.68149362242)*x_ref[1]
        ref[2,2,1,1]=(-0.984200247787)*x_ref[0]**o+(0.83324365092)*x_ref[0]+(-0.896184408816)*x_ref[1]**o+(0.102276243487)*x_ref[1]
        ref[2,2,2,0]=(-0.216357164308)*x_ref[0]**o+(-0.597984882107)*x_ref[0]+(-0.205435921511)*x_ref[1]**o+(0.484947652759)*x_ref[1]
        ref[2,2,2,1]=(-0.650046253868)*x_ref[0]**o+(-0.624878186105)*x_ref[0]+(0.786519914406)*x_ref[1]**o+(0.28027273095)*x_ref[1]
        ref[2,3,0,0]=(-0.0821718346579)*x_ref[0]**o+(-0.687049917467)*x_ref[0]+(0.380792333736)*x_ref[1]**o+(0.151185837715)*x_ref[1]
        ref[2,3,0,1]=(0.61796113854)*x_ref[0]**o+(0.844624413065)*x_ref[0]+(-0.980619401098)*x_ref[1]**o+(-0.854924862641)*x_ref[1]
        ref[2,3,1,0]=(0.812293906883)*x_ref[0]**o+(-0.225593208768)*x_ref[0]+(0.964175374163)*x_ref[1]**o+(0.848850064926)*x_ref[1]
        ref[2,3,1,1]=(0.376453572937)*x_ref[0]**o+(0.994818812059)*x_ref[0]+(0.58872018303)*x_ref[1]**o+(0.68332473361)*x_ref[1]
        ref[2,3,2,0]=(-0.00948421972234)*x_ref[0]**o+(-0.680685462738)*x_ref[0]+(-0.767076278167)*x_ref[1]**o+(0.210165437284)*x_ref[1]
        ref[2,3,2,1]=(-0.455199698263)*x_ref[0]**o+(-0.763430401419)*x_ref[0]+(0.183773350212)*x_ref[1]**o+(0.43105940923)*x_ref[1]
        ref[2,4,0,0]=(-0.673962025988)*x_ref[0]**o+(-0.679073069725)*x_ref[0]+(0.11018431715)*x_ref[1]**o+(0.349020527482)*x_ref[1]
        ref[2,4,0,1]=(-0.876691282076)*x_ref[0]**o+(0.920902226631)*x_ref[0]+(-0.477417470246)*x_ref[1]**o+(0.836417982216)*x_ref[1]
        ref[2,4,1,0]=(0.567966851024)*x_ref[0]**o+(0.583241431716)*x_ref[0]+(-0.271851271008)*x_ref[1]**o+(0.678727902863)*x_ref[1]
        ref[2,4,1,1]=(0.707400094)*x_ref[0]**o+(0.802820856444)*x_ref[0]+(0.84518937616)*x_ref[1]**o+(-0.0267193533403)*x_ref[1]
        ref[2,4,2,0]=(0.213268920036)*x_ref[0]**o+(0.780721415328)*x_ref[0]+(0.839292948154)*x_ref[1]**o+(-0.0376334744443)*x_ref[1]
        ref[2,4,2,1]=(0.136255687031)*x_ref[0]**o+(-0.582766278693)*x_ref[0]+(0.257662521321)*x_ref[1]**o+(0.808337521979)*x_ref[1]
        ref[3,0,0,0]=(-0.524068770822)*x_ref[0]**o+(0.942236878842)*x_ref[0]+(0.889496255793)*x_ref[1]**o+(-0.904810742393)*x_ref[1]
        ref[3,0,0,1]=(0.780157465475)*x_ref[0]**o+(-0.763811668832)*x_ref[0]+(-0.517482070639)*x_ref[1]**o+(0.662770780676)*x_ref[1]
        ref[3,0,1,0]=(-0.693227564812)*x_ref[0]**o+(-0.763506211841)*x_ref[0]+(-0.427709881126)*x_ref[1]**o+(0.159109827204)*x_ref[1]
        ref[3,0,1,1]=(0.644975629084)*x_ref[0]**o+(-0.928112761899)*x_ref[0]+(-0.530772740083)*x_ref[1]**o+(-0.723174790087)*x_ref[1]
        ref[3,0,2,0]=(-0.923467274097)*x_ref[0]**o+(0.806128117288)*x_ref[0]+(0.177779363207)*x_ref[1]**o+(0.862243885542)*x_ref[1]
        ref[3,0,2,1]=(-0.0823252901358)*x_ref[0]**o+(0.289022916284)*x_ref[0]+(-0.518508771021)*x_ref[1]**o+(0.174564769845)*x_ref[1]
        ref[3,1,0,0]=(0.608837578444)*x_ref[0]**o+(0.898216758361)*x_ref[0]+(-0.844730109761)*x_ref[1]**o+(0.745974173207)*x_ref[1]
        ref[3,1,0,1]=(-0.11776411078)*x_ref[0]**o+(0.186256502356)*x_ref[0]+(0.565289400333)*x_ref[1]**o+(-0.776173716206)*x_ref[1]
        ref[3,1,1,0]=(-0.170736580152)*x_ref[0]**o+(0.829581661683)*x_ref[0]+(0.310135833859)*x_ref[1]**o+(-0.509629545687)*x_ref[1]
        ref[3,1,1,1]=(-0.225246122751)*x_ref[0]**o+(0.586754377294)*x_ref[0]+(-0.726064725635)*x_ref[1]**o+(0.136952605704)*x_ref[1]
        ref[3,1,2,0]=(-0.287932115478)*x_ref[0]**o+(0.858031396153)*x_ref[0]+(-0.493441745353)*x_ref[1]**o+(0.601827664994)*x_ref[1]
        ref[3,1,2,1]=(0.665503842862)*x_ref[0]**o+(-0.515998970452)*x_ref[0]+(-0.890055558624)*x_ref[1]**o+(-0.0446371451457)*x_ref[1]
        ref[3,2,0,0]=(-0.33741557585)*x_ref[0]**o+(-0.416183752333)*x_ref[0]+(-0.92430073538)*x_ref[1]**o+(-0.486265090904)*x_ref[1]
        ref[3,2,0,1]=(0.910860925994)*x_ref[0]**o+(-0.896869755389)*x_ref[0]+(0.0634658185324)*x_ref[1]**o+(-0.383136455739)*x_ref[1]
        ref[3,2,1,0]=(0.386839297976)*x_ref[0]**o+(-0.190659927071)*x_ref[0]+(-0.304846978185)*x_ref[1]**o+(-0.92705511899)*x_ref[1]
        ref[3,2,1,1]=(-0.0197812018683)*x_ref[0]**o+(-0.479083095775)*x_ref[0]+(0.502489976089)*x_ref[1]**o+(-0.859440665475)*x_ref[1]
        ref[3,2,2,0]=(-0.507672045299)*x_ref[0]**o+(0.498689676072)*x_ref[0]+(-0.719735804645)*x_ref[1]**o+(-0.522053306674)*x_ref[1]
        ref[3,2,2,1]=(0.46207629211)*x_ref[0]**o+(-0.496697066039)*x_ref[0]+(0.994069339904)*x_ref[1]**o+(0.873222754449)*x_ref[1]
        ref[3,3,0,0]=(0.126410029692)*x_ref[0]**o+(0.51067050196)*x_ref[0]+(-0.978384824088)*x_ref[1]**o+(-0.68606930503)*x_ref[1]
        ref[3,3,0,1]=(0.376888157008)*x_ref[0]**o+(-0.14015764025)*x_ref[0]+(0.523667700758)*x_ref[1]**o+(0.0277726855258)*x_ref[1]
        ref[3,3,1,0]=(-0.105865233294)*x_ref[0]**o+(-0.0515230741203)*x_ref[0]+(0.203766525036)*x_ref[1]**o+(-0.0431893014387)*x_ref[1]
        ref[3,3,1,1]=(-0.404351448486)*x_ref[0]**o+(-0.639161429157)*x_ref[0]+(-0.957341091708)*x_ref[1]**o+(-0.103346466984)*x_ref[1]
        ref[3,3,2,0]=(-0.0523508780623)*x_ref[0]**o+(-0.563589575544)*x_ref[0]+(0.158487727325)*x_ref[1]**o+(-0.519455492178)*x_ref[1]
        ref[3,3,2,1]=(-0.823286342218)*x_ref[0]**o+(-0.493663499535)*x_ref[0]+(-0.409825574064)*x_ref[1]**o+(-0.0406101279806)*x_ref[1]
        ref[3,4,0,0]=(0.926212910387)*x_ref[0]**o+(-0.0944507963)*x_ref[0]+(0.640740499391)*x_ref[1]**o+(-0.106969076047)*x_ref[1]
        ref[3,4,0,1]=(-0.912044993941)*x_ref[0]**o+(-0.495323777918)*x_ref[0]+(-0.595831845108)*x_ref[1]**o+(-0.313723426318)*x_ref[1]
        ref[3,4,1,0]=(-0.0730898742548)*x_ref[0]**o+(0.0524008484492)*x_ref[0]+(-0.948232935497)*x_ref[1]**o+(0.952670824572)*x_ref[1]
        ref[3,4,1,1]=(-0.81869438138)*x_ref[0]**o+(-0.894654266952)*x_ref[0]+(0.414078621872)*x_ref[1]**o+(-0.876690817847)*x_ref[1]
        ref[3,4,2,0]=(0.00688155078441)*x_ref[0]**o+(0.845964586551)*x_ref[0]+(0.0183149371892)*x_ref[1]**o+(-0.882931394537)*x_ref[1]
        ref[3,4,2,1]=(-0.734830097149)*x_ref[0]**o+(-0.37824588859)*x_ref[0]+(-0.815402620311)*x_ref[1]**o+(-0.357905685517)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.110519315354)*x[0]**o+(0.904227374292)*x[0]+(0.459005710707)*x[1]**o+(-0.980231422256)*x[1]+(-0.819713355475)*x[2]**o+(-0.93459517989)*x[2]
        arg[0,0,0,1]=(-0.326902946413)*x[0]**o+(0.489327766219)*x[0]+(-0.965308741061)*x[1]**o+(0.0351399111471)*x[1]+(0.257760096595)*x[2]**o+(0.125123023244)*x[2]
        arg[0,0,1,0]=(-0.584360710004)*x[0]**o+(0.681892732544)*x[0]+(-0.118257741614)*x[1]**o+(0.839875738892)*x[1]+(0.107934154241)*x[2]**o+(0.279757054665)*x[2]
        arg[0,0,1,1]=(0.90973214794)*x[0]**o+(-0.776804618826)*x[0]+(-0.910633853381)*x[1]**o+(-0.25678788529)*x[1]+(0.783709363646)*x[2]**o+(-0.922660855915)*x[2]
        arg[0,0,2,0]=(-0.802273263394)*x[0]**o+(-0.917749665952)*x[0]+(0.796206817226)*x[1]**o+(0.917284637885)*x[1]+(0.219841508893)*x[2]**o+(0.0655848200083)*x[2]
        arg[0,0,2,1]=(0.550596792343)*x[0]**o+(-0.760207057617)*x[0]+(0.0368469683098)*x[1]**o+(0.509347717331)*x[1]+(0.813780778487)*x[2]**o+(-0.432656298451)*x[2]
        arg[0,1,0,0]=(-0.238502042941)*x[0]**o+(-0.966632221032)*x[0]+(-0.373745652933)*x[1]**o+(-0.436176336389)*x[1]+(0.124280197699)*x[2]**o+(0.552700562803)*x[2]
        arg[0,1,0,1]=(-0.19848053276)*x[0]**o+(0.110167634957)*x[0]+(-0.748553857473)*x[1]**o+(0.355932508959)*x[1]+(-0.226987914753)*x[2]**o+(-0.68870238197)*x[2]
        arg[0,1,1,0]=(0.365822518882)*x[0]**o+(0.327285102897)*x[0]+(0.549212504635)*x[1]**o+(0.918179946504)*x[1]+(0.20644004679)*x[2]**o+(0.821299409756)*x[2]
        arg[0,1,1,1]=(-0.366167489101)*x[0]**o+(0.529933186572)*x[0]+(-0.604965310967)*x[1]**o+(-0.833093092352)*x[1]+(-0.73593913912)*x[2]**o+(0.872325614453)*x[2]
        arg[0,1,2,0]=(0.289123646902)*x[0]**o+(0.705862153789)*x[0]+(-0.319756886822)*x[1]**o+(-0.453220232524)*x[1]+(-0.192336374167)*x[2]**o+(0.363575495006)*x[2]
        arg[0,1,2,1]=(-0.213014833353)*x[0]**o+(-0.0706784230627)*x[0]+(-0.941272982229)*x[1]**o+(-0.96641361036)*x[1]+(0.289919014964)*x[2]**o+(0.57658772066)*x[2]
        arg[0,2,0,0]=(-0.406683991704)*x[0]**o+(-0.0110042213592)*x[0]+(-0.273148083258)*x[1]**o+(0.054257030447)*x[1]+(0.387598147681)*x[2]**o+(0.480398109069)*x[2]
        arg[0,2,0,1]=(0.318713057562)*x[0]**o+(-0.449467749585)*x[0]+(0.669991267829)*x[1]**o+(0.237986860249)*x[1]+(0.894370523471)*x[2]**o+(-0.397832436233)*x[2]
        arg[0,2,1,0]=(-0.858603253642)*x[0]**o+(-0.713122015919)*x[0]+(0.870251906188)*x[1]**o+(-0.781601254578)*x[1]+(-0.504531525822)*x[2]**o+(-0.699819349556)*x[2]
        arg[0,2,1,1]=(-0.825591116361)*x[0]**o+(-0.552452876394)*x[0]+(-0.262627261618)*x[1]**o+(-0.147127256184)*x[1]+(-0.171485665103)*x[2]**o+(0.823279741501)*x[2]
        arg[0,2,2,0]=(-0.0416279689068)*x[0]**o+(-0.0903338000001)*x[0]+(0.911673780635)*x[1]**o+(0.86298673049)*x[1]+(-0.60251179847)*x[2]**o+(0.378303877746)*x[2]
        arg[0,2,2,1]=(0.261384106387)*x[0]**o+(0.0326416199811)*x[0]+(-0.426740102269)*x[1]**o+(-0.445401626921)*x[1]+(-0.388668254526)*x[2]**o+(-0.977942913284)*x[2]
        arg[0,3,0,0]=(-0.0195053813483)*x[0]**o+(-0.294061998664)*x[0]+(0.790461098426)*x[1]**o+(-0.506334268838)*x[1]+(0.394015541605)*x[2]**o+(-0.562248105432)*x[2]
        arg[0,3,0,1]=(0.390726516735)*x[0]**o+(-0.757597437859)*x[0]+(0.861882149861)*x[1]**o+(0.621870451314)*x[1]+(0.900777666162)*x[2]**o+(0.271507131938)*x[2]
        arg[0,3,1,0]=(-0.725635966475)*x[0]**o+(-0.695837548737)*x[0]+(-0.163078176328)*x[1]**o+(-0.0435808418729)*x[1]+(0.0890756544611)*x[2]**o+(-0.270422627949)*x[2]
        arg[0,3,1,1]=(-0.175498720647)*x[0]**o+(-0.213605441632)*x[0]+(-0.358632620362)*x[1]**o+(-0.193273004152)*x[1]+(0.937704823805)*x[2]**o+(0.861707832291)*x[2]
        arg[0,3,2,0]=(0.0976010265611)*x[0]**o+(0.995135925341)*x[0]+(0.755339287387)*x[1]**o+(-0.924722068193)*x[1]+(-0.480212181904)*x[2]**o+(-0.40237760439)*x[2]
        arg[0,3,2,1]=(-0.679215392952)*x[0]**o+(-0.00796096253297)*x[0]+(-0.964972599421)*x[1]**o+(-0.410415435656)*x[1]+(-0.292638067785)*x[2]**o+(-0.950383976561)*x[2]
        arg[0,4,0,0]=(0.245586594928)*x[0]**o+(-0.593420401896)*x[0]+(0.553450486534)*x[1]**o+(0.173814965689)*x[1]+(0.0719392610563)*x[2]**o+(0.676231437486)*x[2]
        arg[0,4,0,1]=(0.154382756425)*x[0]**o+(-0.0655740293432)*x[0]+(0.924689745142)*x[1]**o+(-0.246876531175)*x[1]+(-0.241460713543)*x[2]**o+(-0.489840108568)*x[2]
        arg[0,4,1,0]=(-0.482705840359)*x[0]**o+(-0.392960380114)*x[0]+(-0.503463823972)*x[1]**o+(0.317163204738)*x[1]+(0.775048877356)*x[2]**o+(0.599889732326)*x[2]
        arg[0,4,1,1]=(-0.122555861716)*x[0]**o+(0.930175770449)*x[0]+(0.483503764941)*x[1]**o+(0.340001033217)*x[1]+(0.672924280979)*x[2]**o+(0.746489359683)*x[2]
        arg[0,4,2,0]=(-0.339421141033)*x[0]**o+(-0.604966575169)*x[0]+(0.862413017648)*x[1]**o+(0.936752078454)*x[1]+(-0.0549628353285)*x[2]**o+(-0.0858885835314)*x[2]
        arg[0,4,2,1]=(-0.581767357421)*x[0]**o+(-0.644286027548)*x[0]+(0.76948396865)*x[1]**o+(0.692265904744)*x[1]+(0.955871091531)*x[2]**o+(-0.378888503457)*x[2]
        arg[1,0,0,0]=(-0.0407846595046)*x[0]**o+(-0.293709943251)*x[0]+(-0.310552931582)*x[1]**o+(-0.696244178466)*x[1]+(0.962985284252)*x[2]**o+(-0.0266457621404)*x[2]
        arg[1,0,0,1]=(-0.480942811584)*x[0]**o+(-0.400158759679)*x[0]+(0.311171869807)*x[1]**o+(-0.79518809427)*x[1]+(-0.497004450478)*x[2]**o+(-0.639822005752)*x[2]
        arg[1,0,1,0]=(0.158738648483)*x[0]**o+(0.274859816468)*x[0]+(-0.208677612654)*x[1]**o+(0.911107598366)*x[1]+(-0.400163007778)*x[2]**o+(0.276296890267)*x[2]
        arg[1,0,1,1]=(0.613312276864)*x[0]**o+(-0.644237218776)*x[0]+(0.81927272895)*x[1]**o+(-0.12971860556)*x[1]+(0.214379406799)*x[2]**o+(0.856335675043)*x[2]
        arg[1,0,2,0]=(-0.699925173766)*x[0]**o+(-0.560613547952)*x[0]+(-0.355325107637)*x[1]**o+(-0.598747822041)*x[1]+(0.243087037553)*x[2]**o+(-0.484176993015)*x[2]
        arg[1,0,2,1]=(-0.218912300848)*x[0]**o+(-0.259483312969)*x[0]+(-0.863788186295)*x[1]**o+(0.924423774757)*x[1]+(-0.0823752192204)*x[2]**o+(-0.928377900602)*x[2]
        arg[1,1,0,0]=(-0.368901144425)*x[0]**o+(0.544104655592)*x[0]+(0.321295151176)*x[1]**o+(0.395692754134)*x[1]+(-0.721319111754)*x[2]**o+(-0.698518442137)*x[2]
        arg[1,1,0,1]=(-0.592867336047)*x[0]**o+(0.145120231038)*x[0]+(0.110390771366)*x[1]**o+(0.168333368086)*x[1]+(-0.577866419921)*x[2]**o+(0.705027708469)*x[2]
        arg[1,1,1,0]=(-0.210603428235)*x[0]**o+(-0.250156447116)*x[0]+(-0.397803248976)*x[1]**o+(0.332163380867)*x[1]+(-0.250648216541)*x[2]**o+(-0.0693729055697)*x[2]
        arg[1,1,1,1]=(0.324927558096)*x[0]**o+(-0.262726116223)*x[0]+(-0.555439758488)*x[1]**o+(-0.865019248609)*x[1]+(-0.0971907425857)*x[2]**o+(-0.742359236711)*x[2]
        arg[1,1,2,0]=(0.161058869414)*x[0]**o+(0.349191500791)*x[0]+(0.468336473315)*x[1]**o+(-0.856364775995)*x[1]+(0.491736696689)*x[2]**o+(0.352984104173)*x[2]
        arg[1,1,2,1]=(-0.243275976508)*x[0]**o+(-0.785605388878)*x[0]+(-0.282478919296)*x[1]**o+(-0.728851522341)*x[1]+(0.375134848237)*x[2]**o+(-0.486200561256)*x[2]
        arg[1,2,0,0]=(-0.172241451405)*x[0]**o+(0.422361498196)*x[0]+(0.464862755164)*x[1]**o+(0.466661717105)*x[1]+(-0.886209677512)*x[2]**o+(-0.43529076789)*x[2]
        arg[1,2,0,1]=(0.946582222405)*x[0]**o+(0.513640973894)*x[0]+(-0.923788000633)*x[1]**o+(-0.358163567195)*x[1]+(-0.709999132129)*x[2]**o+(-0.761163299694)*x[2]
        arg[1,2,1,0]=(0.236339981889)*x[0]**o+(0.259874866134)*x[0]+(-0.32460762691)*x[1]**o+(-0.629786950313)*x[1]+(-0.679261685402)*x[2]**o+(0.593069192918)*x[2]
        arg[1,2,1,1]=(0.11006800009)*x[0]**o+(0.970532991787)*x[0]+(-0.467608188361)*x[1]**o+(-0.959281984552)*x[1]+(0.610396622519)*x[2]**o+(0.282468600058)*x[2]
        arg[1,2,2,0]=(-0.656419157742)*x[0]**o+(0.234027308171)*x[0]+(-0.67027206978)*x[1]**o+(-0.746034436817)*x[1]+(-0.771894073652)*x[2]**o+(0.949545830287)*x[2]
        arg[1,2,2,1]=(0.615884236811)*x[0]**o+(-0.670156612555)*x[0]+(0.543165493345)*x[1]**o+(-0.618294277594)*x[1]+(-0.974114255061)*x[2]**o+(-0.762320542298)*x[2]
        arg[1,3,0,0]=(0.096201957648)*x[0]**o+(0.440870816208)*x[0]+(0.124448859607)*x[1]**o+(0.507919407561)*x[1]+(-0.401666263687)*x[2]**o+(-0.437653713723)*x[2]
        arg[1,3,0,1]=(0.165129229294)*x[0]**o+(-0.353217428955)*x[0]+(0.473666065504)*x[1]**o+(0.907667166988)*x[1]+(-0.864298533288)*x[2]**o+(0.641505117281)*x[2]
        arg[1,3,1,0]=(0.381900624425)*x[0]**o+(0.427520178597)*x[0]+(-0.560279540293)*x[1]**o+(0.190814532271)*x[1]+(-0.187947688629)*x[2]**o+(-0.127694148593)*x[2]
        arg[1,3,1,1]=(-0.903921363895)*x[0]**o+(-0.659942040828)*x[0]+(-0.549460213492)*x[1]**o+(0.869909651895)*x[1]+(-0.962273439484)*x[2]**o+(-0.046365857751)*x[2]
        arg[1,3,2,0]=(-0.021688482187)*x[0]**o+(-0.209531000717)*x[0]+(0.668564395133)*x[1]**o+(-0.0347978700975)*x[1]+(0.947677147488)*x[2]**o+(-0.0152385052475)*x[2]
        arg[1,3,2,1]=(0.880839485817)*x[0]**o+(-0.319275929312)*x[0]+(-0.940056657485)*x[1]**o+(0.570855397726)*x[1]+(0.279710812074)*x[2]**o+(0.496622215051)*x[2]
        arg[1,4,0,0]=(-0.139033645615)*x[0]**o+(-0.271453722366)*x[0]+(-0.268516720244)*x[1]**o+(-0.0467555017438)*x[1]+(-0.41618262265)*x[2]**o+(0.315320363813)*x[2]
        arg[1,4,0,1]=(0.296108607204)*x[0]**o+(-0.184582165661)*x[0]+(0.121416543744)*x[1]**o+(-0.922284926751)*x[1]+(-0.437600089668)*x[2]**o+(-0.608800626513)*x[2]
        arg[1,4,1,0]=(0.8942028423)*x[0]**o+(0.673402831188)*x[0]+(-0.622529195562)*x[1]**o+(-0.266052338798)*x[1]+(-0.273167851702)*x[2]**o+(-0.312211982747)*x[2]
        arg[1,4,1,1]=(-0.102315711178)*x[0]**o+(-0.885515719371)*x[0]+(-0.532910101204)*x[1]**o+(0.798525944348)*x[1]+(0.640586802303)*x[2]**o+(0.36308806536)*x[2]
        arg[1,4,2,0]=(0.376793506541)*x[0]**o+(-0.667063564875)*x[0]+(0.609591985816)*x[1]**o+(0.28347745142)*x[1]+(-0.751832936302)*x[2]**o+(-0.683738856965)*x[2]
        arg[1,4,2,1]=(-0.51412252819)*x[0]**o+(0.559560297206)*x[0]+(0.751134603429)*x[1]**o+(-0.0031812310797)*x[1]+(-0.525776122178)*x[2]**o+(0.338471942848)*x[2]
        arg[2,0,0,0]=(0.937572244905)*x[0]**o+(-0.293020870379)*x[0]+(-0.918591520817)*x[1]**o+(0.622194477261)*x[1]+(0.31641482851)*x[2]**o+(0.276912266864)*x[2]
        arg[2,0,0,1]=(0.853041599998)*x[0]**o+(-0.0220623724171)*x[0]+(0.89909044296)*x[1]**o+(-0.838782642732)*x[1]+(0.0133197592042)*x[2]**o+(-0.868977618993)*x[2]
        arg[2,0,1,0]=(-0.325799042976)*x[0]**o+(0.746269971994)*x[0]+(0.882671061757)*x[1]**o+(-0.284133373252)*x[1]+(0.120086549521)*x[2]**o+(-0.397352610969)*x[2]
        arg[2,0,1,1]=(0.282336934951)*x[0]**o+(-0.339893978547)*x[0]+(-0.0861660125386)*x[1]**o+(0.283063579798)*x[1]+(0.714762478569)*x[2]**o+(-0.497021718549)*x[2]
        arg[2,0,2,0]=(0.891782779491)*x[0]**o+(-0.817647577644)*x[0]+(0.364088142085)*x[1]**o+(-0.950577637048)*x[1]+(0.4070663557)*x[2]**o+(0.270753179246)*x[2]
        arg[2,0,2,1]=(0.316594913808)*x[0]**o+(0.310731486851)*x[0]+(0.250421828765)*x[1]**o+(0.772786141458)*x[1]+(0.64843040679)*x[2]**o+(-0.691903332511)*x[2]
        arg[2,1,0,0]=(0.87180715028)*x[0]**o+(-0.953173230375)*x[0]+(-0.823516344548)*x[1]**o+(0.934653756548)*x[1]+(-0.537124962489)*x[2]**o+(-0.233365939659)*x[2]
        arg[2,1,0,1]=(-0.407189744631)*x[0]**o+(-0.754425720288)*x[0]+(0.737526767897)*x[1]**o+(0.745393943502)*x[1]+(0.811744293685)*x[2]**o+(-0.459556659177)*x[2]
        arg[2,1,1,0]=(-0.693529259877)*x[0]**o+(0.580197938048)*x[0]+(0.506055440405)*x[1]**o+(0.122531014759)*x[1]+(0.427736137775)*x[2]**o+(-0.83260140401)*x[2]
        arg[2,1,1,1]=(0.542621975026)*x[0]**o+(-0.568425713106)*x[0]+(0.303002043446)*x[1]**o+(0.363286261651)*x[1]+(0.2729476289)*x[2]**o+(-0.0113629744659)*x[2]
        arg[2,1,2,0]=(-0.266750235694)*x[0]**o+(0.10116400306)*x[0]+(0.315977962897)*x[1]**o+(-0.406236201454)*x[1]+(-0.586845072374)*x[2]**o+(0.513937047282)*x[2]
        arg[2,1,2,1]=(-0.518780811975)*x[0]**o+(0.275053640829)*x[0]+(0.4234970646)*x[1]**o+(0.389585163376)*x[1]+(-0.609771068778)*x[2]**o+(-0.121467061755)*x[2]
        arg[2,2,0,0]=(0.138823207466)*x[0]**o+(-0.779837455014)*x[0]+(0.70167049555)*x[1]**o+(0.766428813369)*x[1]+(0.114023530366)*x[2]**o+(0.650098536187)*x[2]
        arg[2,2,0,1]=(0.974998667633)*x[0]**o+(0.958471050654)*x[0]+(0.193515162503)*x[1]**o+(-0.252148167253)*x[1]+(0.495475460125)*x[2]**o+(0.562693347983)*x[2]
        arg[2,2,1,0]=(-0.552379548988)*x[0]**o+(-0.831399971618)*x[0]+(0.437581216647)*x[1]**o+(-0.775331267696)*x[1]+(0.729041628844)*x[2]**o+(0.633113359893)*x[2]
        arg[2,2,1,1]=(0.379858437049)*x[0]**o+(0.923586504203)*x[0]+(0.707244783969)*x[1]**o+(-0.293157375059)*x[1]+(-0.368717941876)*x[2]**o+(-0.202610288616)*x[2]
        arg[2,2,2,0]=(-0.804826108231)*x[0]**o+(-0.0208712096312)*x[0]+(0.0482547072498)*x[1]**o+(0.852366543609)*x[1]+(0.328962500136)*x[2]**o+(-0.132777835512)*x[2]
        arg[2,2,2,1]=(0.91373752217)*x[0]**o+(-0.792112120729)*x[0]+(-0.0457564087072)*x[1]**o+(0.309694852949)*x[1]+(0.089219285337)*x[2]**o+(-0.516000514614)*x[2]
        arg[2,3,0,0]=(0.343215129115)*x[0]**o+(0.40804099837)*x[0]+(-0.865942442914)*x[1]**o+(0.842241683043)*x[1]+(0.0727803297601)*x[2]**o+(-0.395225682949)*x[2]
        arg[2,3,0,1]=(-0.965340522771)*x[0]**o+(-0.897056478677)*x[0]+(-0.305205944062)*x[1]**o+(0.393941789685)*x[1]+(-0.253328379131)*x[2]**o+(0.0951510469271)*x[2]
        arg[2,3,1,0]=(0.536251514752)*x[0]**o+(0.0629776386779)*x[0]+(-0.706311222379)*x[1]**o+(-0.84362242029)*x[1]+(0.00110678760539)*x[2]**o+(-0.30286059941)*x[2]
        arg[2,3,1,1]=(0.114584737295)*x[0]**o+(0.0708763738067)*x[0]+(-0.454760062397)*x[1]**o+(0.228742539922)*x[1]+(0.514960810352)*x[2]**o+(-0.815749211085)*x[2]
        arg[2,3,2,0]=(0.246335688008)*x[0]**o+(-0.239199299885)*x[0]+(-0.447390151659)*x[1]**o+(0.918949356391)*x[1]+(-0.307864840175)*x[2]**o+(0.560139080481)*x[2]
        arg[2,3,2,1]=(0.118766035723)*x[0]**o+(0.610711164899)*x[0]+(0.336628771776)*x[1]**o+(0.833699268751)*x[1]+(-0.424694647429)*x[2]**o+(-0.639929688104)*x[2]
        arg[2,4,0,0]=(0.814543506987)*x[0]**o+(-0.797684321368)*x[0]+(0.337242635783)*x[1]**o+(0.278139837072)*x[1]+(0.633511526441)*x[2]**o+(0.372175615028)*x[2]
        arg[2,4,0,1]=(-0.0590668694094)*x[0]**o+(0.256732067267)*x[0]+(0.519737187577)*x[1]**o+(-0.376391855948)*x[1]+(-0.810934658651)*x[2]**o+(0.826899476928)*x[2]
        arg[2,4,1,0]=(-0.428976843661)*x[0]**o+(0.672667833313)*x[0]+(0.263194890946)*x[1]**o+(-0.894790773661)*x[1]+(-0.550298557295)*x[2]**o+(-0.00560384035502)*x[2]
        arg[2,4,1,1]=(-0.818372430144)*x[0]**o+(-0.829920833073)*x[0]+(-0.359560636119)*x[1]**o+(-0.113484563205)*x[1]+(0.586796110365)*x[2]**o+(0.296820898426)*x[2]
        arg[2,4,2,0]=(-0.691597969566)*x[0]**o+(0.715813153538)*x[0]+(0.954093612503)*x[1]**o+(0.88803420754)*x[1]+(0.697447667765)*x[2]**o+(0.793730496625)*x[2]
        arg[2,4,2,1]=(0.351099833336)*x[0]**o+(-0.716419272807)*x[0]+(0.318726674998)*x[1]**o+(0.246151914414)*x[1]+(-0.945520673884)*x[2]**o+(-0.489026233954)*x[2]
        arg[3,0,0,0]=(0.511598022804)*x[0]**o+(0.567617269582)*x[0]+(-0.526293319137)*x[1]**o+(0.333750098146)*x[1]+(-0.0869228849144)*x[2]**o+(-0.204072834867)*x[2]
        arg[3,0,0,1]=(0.905384007172)*x[0]**o+(-0.661067390314)*x[0]+(-0.470790632295)*x[1]**o+(-0.149608017616)*x[1]+(0.427879278267)*x[2]**o+(0.987040179184)*x[2]
        arg[3,0,1,0]=(-0.946562006809)*x[0]**o+(0.64141909633)*x[0]+(-0.762076826276)*x[1]**o+(-0.642912850794)*x[1]+(0.712567898559)*x[2]**o+(-0.292128944205)*x[2]
        arg[3,0,1,1]=(-0.955087949361)*x[0]**o+(0.426588041133)*x[0]+(0.339313104127)*x[1]**o+(-0.375147344045)*x[1]+(0.967259014679)*x[2]**o+(0.504521081858)*x[2]
        arg[3,0,2,0]=(0.141876598344)*x[0]**o+(-0.335579252825)*x[0]+(0.514552993602)*x[1]**o+(-0.531116948499)*x[1]+(-0.369319413141)*x[2]**o+(0.0374413262362)*x[2]
        arg[3,0,2,1]=(0.678360463658)*x[0]**o+(-0.357900119646)*x[0]+(0.300750717059)*x[1]**o+(0.516196189448)*x[1]+(0.0535466018203)*x[2]**o+(-0.149961940786)*x[2]
        arg[3,1,0,0]=(0.0975994351079)*x[0]**o+(0.634544068499)*x[0]+(-0.963217046352)*x[1]**o+(-0.804658091456)*x[1]+(-0.620132720823)*x[2]**o+(0.862250619369)*x[2]
        arg[3,1,0,1]=(0.210306328879)*x[0]**o+(-0.457341187966)*x[0]+(0.957556084801)*x[1]**o+(0.704073170928)*x[1]+(0.423134492133)*x[2]**o+(-0.167268375412)*x[2]
        arg[3,1,1,0]=(-0.909228817744)*x[0]**o+(0.110665018921)*x[0]+(0.36275630147)*x[1]**o+(0.346866824763)*x[1]+(-0.535600547489)*x[2]**o+(0.627638814375)*x[2]
        arg[3,1,1,1]=(0.328780933823)*x[0]**o+(-0.249563994161)*x[0]+(-0.555789062892)*x[1]**o+(-0.917287525036)*x[1]+(0.386191175017)*x[2]**o+(0.123904319061)*x[2]
        arg[3,1,2,0]=(-0.838409901623)*x[0]**o+(0.673479705393)*x[0]+(0.482921104917)*x[1]**o+(0.136411388588)*x[1]+(0.583674116225)*x[2]**o+(0.305100836877)*x[2]
        arg[3,1,2,1]=(0.952961008312)*x[0]**o+(-0.0501981079532)*x[0]+(0.661434464642)*x[1]**o+(0.65030469399)*x[1]+(-0.157562001295)*x[2]**o+(-0.249551892917)*x[2]
        arg[3,2,0,0]=(-0.0232804500286)*x[0]**o+(-0.559229392203)*x[0]+(0.127967930273)*x[1]**o+(-0.635827181645)*x[1]+(-0.0426703224383)*x[2]**o+(0.47720230258)*x[2]
        arg[3,2,0,1]=(-0.421658996106)*x[0]**o+(-0.468260038077)*x[0]+(0.479083572554)*x[1]**o+(-0.214417881841)*x[1]+(0.372430534261)*x[2]**o+(0.133063789546)*x[2]
        arg[3,2,1,0]=(0.905811586567)*x[0]**o+(0.245447340384)*x[0]+(-0.337650999146)*x[1]**o+(0.908188235976)*x[1]+(-0.804029751277)*x[2]**o+(0.0512676456643)*x[2]
        arg[3,2,1,1]=(-0.891486839409)*x[0]**o+(0.20753794329)*x[0]+(0.879183464252)*x[1]**o+(0.902028906992)*x[1]+(0.416776813598)*x[2]**o+(0.491964374373)*x[2]
        arg[3,2,2,0]=(0.653067027119)*x[0]**o+(0.765293890745)*x[0]+(-0.348902169508)*x[1]**o+(0.324429035849)*x[1]+(-0.363026365065)*x[2]**o+(0.0816186405511)*x[2]
        arg[3,2,2,1]=(-0.0415375052325)*x[0]**o+(-0.480779075784)*x[0]+(-0.137223552938)*x[1]**o+(-0.915733625512)*x[1]+(0.750199445159)*x[2]**o+(-0.789818945171)*x[2]
        arg[3,3,0,0]=(0.0744921239083)*x[0]**o+(0.0700589144159)*x[0]+(-0.327105425297)*x[1]**o+(0.469843257765)*x[1]+(-0.170191600844)*x[2]**o+(0.732071534155)*x[2]
        arg[3,3,0,1]=(0.172643927441)*x[0]**o+(0.848924114383)*x[0]+(-0.79125762228)*x[1]**o+(-0.694725775976)*x[1]+(-0.309710503985)*x[2]**o+(-0.405310234927)*x[2]
        arg[3,3,1,0]=(-0.465911787595)*x[0]**o+(-0.71080474709)*x[0]+(0.959925934203)*x[1]**o+(-0.111110644825)*x[1]+(-0.145349925218)*x[2]**o+(0.122955165811)*x[2]
        arg[3,3,1,1]=(-0.769111863754)*x[0]**o+(0.348036938532)*x[0]+(0.485890856474)*x[1]**o+(0.382595218985)*x[1]+(-0.897488692981)*x[2]**o+(0.830822333819)*x[2]
        arg[3,3,2,0]=(0.341890793681)*x[0]**o+(-0.667145936571)*x[0]+(-0.145094108268)*x[1]**o+(0.771433911212)*x[1]+(-0.299752164325)*x[2]**o+(-0.357856549677)*x[2]
        arg[3,3,2,1]=(0.26381838796)*x[0]**o+(-0.44474724998)*x[0]+(0.310283988195)*x[1]**o+(0.698675181314)*x[1]+(0.437624133238)*x[2]**o+(0.394916054445)*x[2]
        arg[3,4,0,0]=(-0.576499406674)*x[0]**o+(0.770936634959)*x[0]+(0.707441735783)*x[1]**o+(-0.794953257043)*x[1]+(-0.535342943209)*x[2]**o+(-0.509446811787)*x[2]
        arg[3,4,0,1]=(0.161115012113)*x[0]**o+(0.11079092252)*x[0]+(0.0633804181164)*x[1]**o+(-0.406049293019)*x[1]+(-0.824283994336)*x[2]**o+(-0.481107560499)*x[2]
        arg[3,4,1,0]=(-0.411133810295)*x[0]**o+(0.988486625099)*x[0]+(-0.36050191367)*x[1]**o+(0.469479821965)*x[1]+(-0.431175357325)*x[2]**o+(-0.966295015263)*x[2]
        arg[3,4,1,1]=(0.233731410367)*x[0]**o+(-0.821900740766)*x[0]+(-0.823123461385)*x[1]**o+(-0.380537914421)*x[1]+(-0.896146300355)*x[2]**o+(-0.775172990257)*x[2]
        arg[3,4,2,0]=(-0.96323363586)*x[0]**o+(0.0935412405245)*x[0]+(0.929333026259)*x[1]**o+(-0.322236627348)*x[1]+(-0.390898135964)*x[2]**o+(-0.766178290162)*x[2]
        arg[3,4,2,1]=(-0.773670533268)*x[0]**o+(-0.156975157098)*x[0]+(-0.91512996795)*x[1]**o+(0.0247803682126)*x[1]+(0.995159733782)*x[2]**o+(-0.669224144225)*x[2]
        ref[0,0,0,0]=(-0.110519315354)*x_ref[0]**o+(0.904227374292)*x_ref[0]+(0.459005710707)*x_ref[1]**o+(-0.980231422256)*x_ref[1]+(-0.819713355475)*x_ref[2]**o+(-0.93459517989)*x_ref[2]
        ref[0,0,0,1]=(-0.326902946413)*x_ref[0]**o+(0.489327766219)*x_ref[0]+(-0.965308741061)*x_ref[1]**o+(0.0351399111471)*x_ref[1]+(0.257760096595)*x_ref[2]**o+(0.125123023244)*x_ref[2]
        ref[0,0,1,0]=(-0.584360710004)*x_ref[0]**o+(0.681892732544)*x_ref[0]+(-0.118257741614)*x_ref[1]**o+(0.839875738892)*x_ref[1]+(0.107934154241)*x_ref[2]**o+(0.279757054665)*x_ref[2]
        ref[0,0,1,1]=(0.90973214794)*x_ref[0]**o+(-0.776804618826)*x_ref[0]+(-0.910633853381)*x_ref[1]**o+(-0.25678788529)*x_ref[1]+(0.783709363646)*x_ref[2]**o+(-0.922660855915)*x_ref[2]
        ref[0,0,2,0]=(-0.802273263394)*x_ref[0]**o+(-0.917749665952)*x_ref[0]+(0.796206817226)*x_ref[1]**o+(0.917284637885)*x_ref[1]+(0.219841508893)*x_ref[2]**o+(0.0655848200083)*x_ref[2]
        ref[0,0,2,1]=(0.550596792343)*x_ref[0]**o+(-0.760207057617)*x_ref[0]+(0.0368469683098)*x_ref[1]**o+(0.509347717331)*x_ref[1]+(0.813780778487)*x_ref[2]**o+(-0.432656298451)*x_ref[2]
        ref[0,1,0,0]=(-0.238502042941)*x_ref[0]**o+(-0.966632221032)*x_ref[0]+(-0.373745652933)*x_ref[1]**o+(-0.436176336389)*x_ref[1]+(0.124280197699)*x_ref[2]**o+(0.552700562803)*x_ref[2]
        ref[0,1,0,1]=(-0.19848053276)*x_ref[0]**o+(0.110167634957)*x_ref[0]+(-0.748553857473)*x_ref[1]**o+(0.355932508959)*x_ref[1]+(-0.226987914753)*x_ref[2]**o+(-0.68870238197)*x_ref[2]
        ref[0,1,1,0]=(0.365822518882)*x_ref[0]**o+(0.327285102897)*x_ref[0]+(0.549212504635)*x_ref[1]**o+(0.918179946504)*x_ref[1]+(0.20644004679)*x_ref[2]**o+(0.821299409756)*x_ref[2]
        ref[0,1,1,1]=(-0.366167489101)*x_ref[0]**o+(0.529933186572)*x_ref[0]+(-0.604965310967)*x_ref[1]**o+(-0.833093092352)*x_ref[1]+(-0.73593913912)*x_ref[2]**o+(0.872325614453)*x_ref[2]
        ref[0,1,2,0]=(0.289123646902)*x_ref[0]**o+(0.705862153789)*x_ref[0]+(-0.319756886822)*x_ref[1]**o+(-0.453220232524)*x_ref[1]+(-0.192336374167)*x_ref[2]**o+(0.363575495006)*x_ref[2]
        ref[0,1,2,1]=(-0.213014833353)*x_ref[0]**o+(-0.0706784230627)*x_ref[0]+(-0.941272982229)*x_ref[1]**o+(-0.96641361036)*x_ref[1]+(0.289919014964)*x_ref[2]**o+(0.57658772066)*x_ref[2]
        ref[0,2,0,0]=(-0.406683991704)*x_ref[0]**o+(-0.0110042213592)*x_ref[0]+(-0.273148083258)*x_ref[1]**o+(0.054257030447)*x_ref[1]+(0.387598147681)*x_ref[2]**o+(0.480398109069)*x_ref[2]
        ref[0,2,0,1]=(0.318713057562)*x_ref[0]**o+(-0.449467749585)*x_ref[0]+(0.669991267829)*x_ref[1]**o+(0.237986860249)*x_ref[1]+(0.894370523471)*x_ref[2]**o+(-0.397832436233)*x_ref[2]
        ref[0,2,1,0]=(-0.858603253642)*x_ref[0]**o+(-0.713122015919)*x_ref[0]+(0.870251906188)*x_ref[1]**o+(-0.781601254578)*x_ref[1]+(-0.504531525822)*x_ref[2]**o+(-0.699819349556)*x_ref[2]
        ref[0,2,1,1]=(-0.825591116361)*x_ref[0]**o+(-0.552452876394)*x_ref[0]+(-0.262627261618)*x_ref[1]**o+(-0.147127256184)*x_ref[1]+(-0.171485665103)*x_ref[2]**o+(0.823279741501)*x_ref[2]
        ref[0,2,2,0]=(-0.0416279689068)*x_ref[0]**o+(-0.0903338000001)*x_ref[0]+(0.911673780635)*x_ref[1]**o+(0.86298673049)*x_ref[1]+(-0.60251179847)*x_ref[2]**o+(0.378303877746)*x_ref[2]
        ref[0,2,2,1]=(0.261384106387)*x_ref[0]**o+(0.0326416199811)*x_ref[0]+(-0.426740102269)*x_ref[1]**o+(-0.445401626921)*x_ref[1]+(-0.388668254526)*x_ref[2]**o+(-0.977942913284)*x_ref[2]
        ref[0,3,0,0]=(-0.0195053813483)*x_ref[0]**o+(-0.294061998664)*x_ref[0]+(0.790461098426)*x_ref[1]**o+(-0.506334268838)*x_ref[1]+(0.394015541605)*x_ref[2]**o+(-0.562248105432)*x_ref[2]
        ref[0,3,0,1]=(0.390726516735)*x_ref[0]**o+(-0.757597437859)*x_ref[0]+(0.861882149861)*x_ref[1]**o+(0.621870451314)*x_ref[1]+(0.900777666162)*x_ref[2]**o+(0.271507131938)*x_ref[2]
        ref[0,3,1,0]=(-0.725635966475)*x_ref[0]**o+(-0.695837548737)*x_ref[0]+(-0.163078176328)*x_ref[1]**o+(-0.0435808418729)*x_ref[1]+(0.0890756544611)*x_ref[2]**o+(-0.270422627949)*x_ref[2]
        ref[0,3,1,1]=(-0.175498720647)*x_ref[0]**o+(-0.213605441632)*x_ref[0]+(-0.358632620362)*x_ref[1]**o+(-0.193273004152)*x_ref[1]+(0.937704823805)*x_ref[2]**o+(0.861707832291)*x_ref[2]
        ref[0,3,2,0]=(0.0976010265611)*x_ref[0]**o+(0.995135925341)*x_ref[0]+(0.755339287387)*x_ref[1]**o+(-0.924722068193)*x_ref[1]+(-0.480212181904)*x_ref[2]**o+(-0.40237760439)*x_ref[2]
        ref[0,3,2,1]=(-0.679215392952)*x_ref[0]**o+(-0.00796096253297)*x_ref[0]+(-0.964972599421)*x_ref[1]**o+(-0.410415435656)*x_ref[1]+(-0.292638067785)*x_ref[2]**o+(-0.950383976561)*x_ref[2]
        ref[0,4,0,0]=(0.245586594928)*x_ref[0]**o+(-0.593420401896)*x_ref[0]+(0.553450486534)*x_ref[1]**o+(0.173814965689)*x_ref[1]+(0.0719392610563)*x_ref[2]**o+(0.676231437486)*x_ref[2]
        ref[0,4,0,1]=(0.154382756425)*x_ref[0]**o+(-0.0655740293432)*x_ref[0]+(0.924689745142)*x_ref[1]**o+(-0.246876531175)*x_ref[1]+(-0.241460713543)*x_ref[2]**o+(-0.489840108568)*x_ref[2]
        ref[0,4,1,0]=(-0.482705840359)*x_ref[0]**o+(-0.392960380114)*x_ref[0]+(-0.503463823972)*x_ref[1]**o+(0.317163204738)*x_ref[1]+(0.775048877356)*x_ref[2]**o+(0.599889732326)*x_ref[2]
        ref[0,4,1,1]=(-0.122555861716)*x_ref[0]**o+(0.930175770449)*x_ref[0]+(0.483503764941)*x_ref[1]**o+(0.340001033217)*x_ref[1]+(0.672924280979)*x_ref[2]**o+(0.746489359683)*x_ref[2]
        ref[0,4,2,0]=(-0.339421141033)*x_ref[0]**o+(-0.604966575169)*x_ref[0]+(0.862413017648)*x_ref[1]**o+(0.936752078454)*x_ref[1]+(-0.0549628353285)*x_ref[2]**o+(-0.0858885835314)*x_ref[2]
        ref[0,4,2,1]=(-0.581767357421)*x_ref[0]**o+(-0.644286027548)*x_ref[0]+(0.76948396865)*x_ref[1]**o+(0.692265904744)*x_ref[1]+(0.955871091531)*x_ref[2]**o+(-0.378888503457)*x_ref[2]
        ref[1,0,0,0]=(-0.0407846595046)*x_ref[0]**o+(-0.293709943251)*x_ref[0]+(-0.310552931582)*x_ref[1]**o+(-0.696244178466)*x_ref[1]+(0.962985284252)*x_ref[2]**o+(-0.0266457621404)*x_ref[2]
        ref[1,0,0,1]=(-0.480942811584)*x_ref[0]**o+(-0.400158759679)*x_ref[0]+(0.311171869807)*x_ref[1]**o+(-0.79518809427)*x_ref[1]+(-0.497004450478)*x_ref[2]**o+(-0.639822005752)*x_ref[2]
        ref[1,0,1,0]=(0.158738648483)*x_ref[0]**o+(0.274859816468)*x_ref[0]+(-0.208677612654)*x_ref[1]**o+(0.911107598366)*x_ref[1]+(-0.400163007778)*x_ref[2]**o+(0.276296890267)*x_ref[2]
        ref[1,0,1,1]=(0.613312276864)*x_ref[0]**o+(-0.644237218776)*x_ref[0]+(0.81927272895)*x_ref[1]**o+(-0.12971860556)*x_ref[1]+(0.214379406799)*x_ref[2]**o+(0.856335675043)*x_ref[2]
        ref[1,0,2,0]=(-0.699925173766)*x_ref[0]**o+(-0.560613547952)*x_ref[0]+(-0.355325107637)*x_ref[1]**o+(-0.598747822041)*x_ref[1]+(0.243087037553)*x_ref[2]**o+(-0.484176993015)*x_ref[2]
        ref[1,0,2,1]=(-0.218912300848)*x_ref[0]**o+(-0.259483312969)*x_ref[0]+(-0.863788186295)*x_ref[1]**o+(0.924423774757)*x_ref[1]+(-0.0823752192204)*x_ref[2]**o+(-0.928377900602)*x_ref[2]
        ref[1,1,0,0]=(-0.368901144425)*x_ref[0]**o+(0.544104655592)*x_ref[0]+(0.321295151176)*x_ref[1]**o+(0.395692754134)*x_ref[1]+(-0.721319111754)*x_ref[2]**o+(-0.698518442137)*x_ref[2]
        ref[1,1,0,1]=(-0.592867336047)*x_ref[0]**o+(0.145120231038)*x_ref[0]+(0.110390771366)*x_ref[1]**o+(0.168333368086)*x_ref[1]+(-0.577866419921)*x_ref[2]**o+(0.705027708469)*x_ref[2]
        ref[1,1,1,0]=(-0.210603428235)*x_ref[0]**o+(-0.250156447116)*x_ref[0]+(-0.397803248976)*x_ref[1]**o+(0.332163380867)*x_ref[1]+(-0.250648216541)*x_ref[2]**o+(-0.0693729055697)*x_ref[2]
        ref[1,1,1,1]=(0.324927558096)*x_ref[0]**o+(-0.262726116223)*x_ref[0]+(-0.555439758488)*x_ref[1]**o+(-0.865019248609)*x_ref[1]+(-0.0971907425857)*x_ref[2]**o+(-0.742359236711)*x_ref[2]
        ref[1,1,2,0]=(0.161058869414)*x_ref[0]**o+(0.349191500791)*x_ref[0]+(0.468336473315)*x_ref[1]**o+(-0.856364775995)*x_ref[1]+(0.491736696689)*x_ref[2]**o+(0.352984104173)*x_ref[2]
        ref[1,1,2,1]=(-0.243275976508)*x_ref[0]**o+(-0.785605388878)*x_ref[0]+(-0.282478919296)*x_ref[1]**o+(-0.728851522341)*x_ref[1]+(0.375134848237)*x_ref[2]**o+(-0.486200561256)*x_ref[2]
        ref[1,2,0,0]=(-0.172241451405)*x_ref[0]**o+(0.422361498196)*x_ref[0]+(0.464862755164)*x_ref[1]**o+(0.466661717105)*x_ref[1]+(-0.886209677512)*x_ref[2]**o+(-0.43529076789)*x_ref[2]
        ref[1,2,0,1]=(0.946582222405)*x_ref[0]**o+(0.513640973894)*x_ref[0]+(-0.923788000633)*x_ref[1]**o+(-0.358163567195)*x_ref[1]+(-0.709999132129)*x_ref[2]**o+(-0.761163299694)*x_ref[2]
        ref[1,2,1,0]=(0.236339981889)*x_ref[0]**o+(0.259874866134)*x_ref[0]+(-0.32460762691)*x_ref[1]**o+(-0.629786950313)*x_ref[1]+(-0.679261685402)*x_ref[2]**o+(0.593069192918)*x_ref[2]
        ref[1,2,1,1]=(0.11006800009)*x_ref[0]**o+(0.970532991787)*x_ref[0]+(-0.467608188361)*x_ref[1]**o+(-0.959281984552)*x_ref[1]+(0.610396622519)*x_ref[2]**o+(0.282468600058)*x_ref[2]
        ref[1,2,2,0]=(-0.656419157742)*x_ref[0]**o+(0.234027308171)*x_ref[0]+(-0.67027206978)*x_ref[1]**o+(-0.746034436817)*x_ref[1]+(-0.771894073652)*x_ref[2]**o+(0.949545830287)*x_ref[2]
        ref[1,2,2,1]=(0.615884236811)*x_ref[0]**o+(-0.670156612555)*x_ref[0]+(0.543165493345)*x_ref[1]**o+(-0.618294277594)*x_ref[1]+(-0.974114255061)*x_ref[2]**o+(-0.762320542298)*x_ref[2]
        ref[1,3,0,0]=(0.096201957648)*x_ref[0]**o+(0.440870816208)*x_ref[0]+(0.124448859607)*x_ref[1]**o+(0.507919407561)*x_ref[1]+(-0.401666263687)*x_ref[2]**o+(-0.437653713723)*x_ref[2]
        ref[1,3,0,1]=(0.165129229294)*x_ref[0]**o+(-0.353217428955)*x_ref[0]+(0.473666065504)*x_ref[1]**o+(0.907667166988)*x_ref[1]+(-0.864298533288)*x_ref[2]**o+(0.641505117281)*x_ref[2]
        ref[1,3,1,0]=(0.381900624425)*x_ref[0]**o+(0.427520178597)*x_ref[0]+(-0.560279540293)*x_ref[1]**o+(0.190814532271)*x_ref[1]+(-0.187947688629)*x_ref[2]**o+(-0.127694148593)*x_ref[2]
        ref[1,3,1,1]=(-0.903921363895)*x_ref[0]**o+(-0.659942040828)*x_ref[0]+(-0.549460213492)*x_ref[1]**o+(0.869909651895)*x_ref[1]+(-0.962273439484)*x_ref[2]**o+(-0.046365857751)*x_ref[2]
        ref[1,3,2,0]=(-0.021688482187)*x_ref[0]**o+(-0.209531000717)*x_ref[0]+(0.668564395133)*x_ref[1]**o+(-0.0347978700975)*x_ref[1]+(0.947677147488)*x_ref[2]**o+(-0.0152385052475)*x_ref[2]
        ref[1,3,2,1]=(0.880839485817)*x_ref[0]**o+(-0.319275929312)*x_ref[0]+(-0.940056657485)*x_ref[1]**o+(0.570855397726)*x_ref[1]+(0.279710812074)*x_ref[2]**o+(0.496622215051)*x_ref[2]
        ref[1,4,0,0]=(-0.139033645615)*x_ref[0]**o+(-0.271453722366)*x_ref[0]+(-0.268516720244)*x_ref[1]**o+(-0.0467555017438)*x_ref[1]+(-0.41618262265)*x_ref[2]**o+(0.315320363813)*x_ref[2]
        ref[1,4,0,1]=(0.296108607204)*x_ref[0]**o+(-0.184582165661)*x_ref[0]+(0.121416543744)*x_ref[1]**o+(-0.922284926751)*x_ref[1]+(-0.437600089668)*x_ref[2]**o+(-0.608800626513)*x_ref[2]
        ref[1,4,1,0]=(0.8942028423)*x_ref[0]**o+(0.673402831188)*x_ref[0]+(-0.622529195562)*x_ref[1]**o+(-0.266052338798)*x_ref[1]+(-0.273167851702)*x_ref[2]**o+(-0.312211982747)*x_ref[2]
        ref[1,4,1,1]=(-0.102315711178)*x_ref[0]**o+(-0.885515719371)*x_ref[0]+(-0.532910101204)*x_ref[1]**o+(0.798525944348)*x_ref[1]+(0.640586802303)*x_ref[2]**o+(0.36308806536)*x_ref[2]
        ref[1,4,2,0]=(0.376793506541)*x_ref[0]**o+(-0.667063564875)*x_ref[0]+(0.609591985816)*x_ref[1]**o+(0.28347745142)*x_ref[1]+(-0.751832936302)*x_ref[2]**o+(-0.683738856965)*x_ref[2]
        ref[1,4,2,1]=(-0.51412252819)*x_ref[0]**o+(0.559560297206)*x_ref[0]+(0.751134603429)*x_ref[1]**o+(-0.0031812310797)*x_ref[1]+(-0.525776122178)*x_ref[2]**o+(0.338471942848)*x_ref[2]
        ref[2,0,0,0]=(0.937572244905)*x_ref[0]**o+(-0.293020870379)*x_ref[0]+(-0.918591520817)*x_ref[1]**o+(0.622194477261)*x_ref[1]+(0.31641482851)*x_ref[2]**o+(0.276912266864)*x_ref[2]
        ref[2,0,0,1]=(0.853041599998)*x_ref[0]**o+(-0.0220623724171)*x_ref[0]+(0.89909044296)*x_ref[1]**o+(-0.838782642732)*x_ref[1]+(0.0133197592042)*x_ref[2]**o+(-0.868977618993)*x_ref[2]
        ref[2,0,1,0]=(-0.325799042976)*x_ref[0]**o+(0.746269971994)*x_ref[0]+(0.882671061757)*x_ref[1]**o+(-0.284133373252)*x_ref[1]+(0.120086549521)*x_ref[2]**o+(-0.397352610969)*x_ref[2]
        ref[2,0,1,1]=(0.282336934951)*x_ref[0]**o+(-0.339893978547)*x_ref[0]+(-0.0861660125386)*x_ref[1]**o+(0.283063579798)*x_ref[1]+(0.714762478569)*x_ref[2]**o+(-0.497021718549)*x_ref[2]
        ref[2,0,2,0]=(0.891782779491)*x_ref[0]**o+(-0.817647577644)*x_ref[0]+(0.364088142085)*x_ref[1]**o+(-0.950577637048)*x_ref[1]+(0.4070663557)*x_ref[2]**o+(0.270753179246)*x_ref[2]
        ref[2,0,2,1]=(0.316594913808)*x_ref[0]**o+(0.310731486851)*x_ref[0]+(0.250421828765)*x_ref[1]**o+(0.772786141458)*x_ref[1]+(0.64843040679)*x_ref[2]**o+(-0.691903332511)*x_ref[2]
        ref[2,1,0,0]=(0.87180715028)*x_ref[0]**o+(-0.953173230375)*x_ref[0]+(-0.823516344548)*x_ref[1]**o+(0.934653756548)*x_ref[1]+(-0.537124962489)*x_ref[2]**o+(-0.233365939659)*x_ref[2]
        ref[2,1,0,1]=(-0.407189744631)*x_ref[0]**o+(-0.754425720288)*x_ref[0]+(0.737526767897)*x_ref[1]**o+(0.745393943502)*x_ref[1]+(0.811744293685)*x_ref[2]**o+(-0.459556659177)*x_ref[2]
        ref[2,1,1,0]=(-0.693529259877)*x_ref[0]**o+(0.580197938048)*x_ref[0]+(0.506055440405)*x_ref[1]**o+(0.122531014759)*x_ref[1]+(0.427736137775)*x_ref[2]**o+(-0.83260140401)*x_ref[2]
        ref[2,1,1,1]=(0.542621975026)*x_ref[0]**o+(-0.568425713106)*x_ref[0]+(0.303002043446)*x_ref[1]**o+(0.363286261651)*x_ref[1]+(0.2729476289)*x_ref[2]**o+(-0.0113629744659)*x_ref[2]
        ref[2,1,2,0]=(-0.266750235694)*x_ref[0]**o+(0.10116400306)*x_ref[0]+(0.315977962897)*x_ref[1]**o+(-0.406236201454)*x_ref[1]+(-0.586845072374)*x_ref[2]**o+(0.513937047282)*x_ref[2]
        ref[2,1,2,1]=(-0.518780811975)*x_ref[0]**o+(0.275053640829)*x_ref[0]+(0.4234970646)*x_ref[1]**o+(0.389585163376)*x_ref[1]+(-0.609771068778)*x_ref[2]**o+(-0.121467061755)*x_ref[2]
        ref[2,2,0,0]=(0.138823207466)*x_ref[0]**o+(-0.779837455014)*x_ref[0]+(0.70167049555)*x_ref[1]**o+(0.766428813369)*x_ref[1]+(0.114023530366)*x_ref[2]**o+(0.650098536187)*x_ref[2]
        ref[2,2,0,1]=(0.974998667633)*x_ref[0]**o+(0.958471050654)*x_ref[0]+(0.193515162503)*x_ref[1]**o+(-0.252148167253)*x_ref[1]+(0.495475460125)*x_ref[2]**o+(0.562693347983)*x_ref[2]
        ref[2,2,1,0]=(-0.552379548988)*x_ref[0]**o+(-0.831399971618)*x_ref[0]+(0.437581216647)*x_ref[1]**o+(-0.775331267696)*x_ref[1]+(0.729041628844)*x_ref[2]**o+(0.633113359893)*x_ref[2]
        ref[2,2,1,1]=(0.379858437049)*x_ref[0]**o+(0.923586504203)*x_ref[0]+(0.707244783969)*x_ref[1]**o+(-0.293157375059)*x_ref[1]+(-0.368717941876)*x_ref[2]**o+(-0.202610288616)*x_ref[2]
        ref[2,2,2,0]=(-0.804826108231)*x_ref[0]**o+(-0.0208712096312)*x_ref[0]+(0.0482547072498)*x_ref[1]**o+(0.852366543609)*x_ref[1]+(0.328962500136)*x_ref[2]**o+(-0.132777835512)*x_ref[2]
        ref[2,2,2,1]=(0.91373752217)*x_ref[0]**o+(-0.792112120729)*x_ref[0]+(-0.0457564087072)*x_ref[1]**o+(0.309694852949)*x_ref[1]+(0.089219285337)*x_ref[2]**o+(-0.516000514614)*x_ref[2]
        ref[2,3,0,0]=(0.343215129115)*x_ref[0]**o+(0.40804099837)*x_ref[0]+(-0.865942442914)*x_ref[1]**o+(0.842241683043)*x_ref[1]+(0.0727803297601)*x_ref[2]**o+(-0.395225682949)*x_ref[2]
        ref[2,3,0,1]=(-0.965340522771)*x_ref[0]**o+(-0.897056478677)*x_ref[0]+(-0.305205944062)*x_ref[1]**o+(0.393941789685)*x_ref[1]+(-0.253328379131)*x_ref[2]**o+(0.0951510469271)*x_ref[2]
        ref[2,3,1,0]=(0.536251514752)*x_ref[0]**o+(0.0629776386779)*x_ref[0]+(-0.706311222379)*x_ref[1]**o+(-0.84362242029)*x_ref[1]+(0.00110678760539)*x_ref[2]**o+(-0.30286059941)*x_ref[2]
        ref[2,3,1,1]=(0.114584737295)*x_ref[0]**o+(0.0708763738067)*x_ref[0]+(-0.454760062397)*x_ref[1]**o+(0.228742539922)*x_ref[1]+(0.514960810352)*x_ref[2]**o+(-0.815749211085)*x_ref[2]
        ref[2,3,2,0]=(0.246335688008)*x_ref[0]**o+(-0.239199299885)*x_ref[0]+(-0.447390151659)*x_ref[1]**o+(0.918949356391)*x_ref[1]+(-0.307864840175)*x_ref[2]**o+(0.560139080481)*x_ref[2]
        ref[2,3,2,1]=(0.118766035723)*x_ref[0]**o+(0.610711164899)*x_ref[0]+(0.336628771776)*x_ref[1]**o+(0.833699268751)*x_ref[1]+(-0.424694647429)*x_ref[2]**o+(-0.639929688104)*x_ref[2]
        ref[2,4,0,0]=(0.814543506987)*x_ref[0]**o+(-0.797684321368)*x_ref[0]+(0.337242635783)*x_ref[1]**o+(0.278139837072)*x_ref[1]+(0.633511526441)*x_ref[2]**o+(0.372175615028)*x_ref[2]
        ref[2,4,0,1]=(-0.0590668694094)*x_ref[0]**o+(0.256732067267)*x_ref[0]+(0.519737187577)*x_ref[1]**o+(-0.376391855948)*x_ref[1]+(-0.810934658651)*x_ref[2]**o+(0.826899476928)*x_ref[2]
        ref[2,4,1,0]=(-0.428976843661)*x_ref[0]**o+(0.672667833313)*x_ref[0]+(0.263194890946)*x_ref[1]**o+(-0.894790773661)*x_ref[1]+(-0.550298557295)*x_ref[2]**o+(-0.00560384035502)*x_ref[2]
        ref[2,4,1,1]=(-0.818372430144)*x_ref[0]**o+(-0.829920833073)*x_ref[0]+(-0.359560636119)*x_ref[1]**o+(-0.113484563205)*x_ref[1]+(0.586796110365)*x_ref[2]**o+(0.296820898426)*x_ref[2]
        ref[2,4,2,0]=(-0.691597969566)*x_ref[0]**o+(0.715813153538)*x_ref[0]+(0.954093612503)*x_ref[1]**o+(0.88803420754)*x_ref[1]+(0.697447667765)*x_ref[2]**o+(0.793730496625)*x_ref[2]
        ref[2,4,2,1]=(0.351099833336)*x_ref[0]**o+(-0.716419272807)*x_ref[0]+(0.318726674998)*x_ref[1]**o+(0.246151914414)*x_ref[1]+(-0.945520673884)*x_ref[2]**o+(-0.489026233954)*x_ref[2]
        ref[3,0,0,0]=(0.511598022804)*x_ref[0]**o+(0.567617269582)*x_ref[0]+(-0.526293319137)*x_ref[1]**o+(0.333750098146)*x_ref[1]+(-0.0869228849144)*x_ref[2]**o+(-0.204072834867)*x_ref[2]
        ref[3,0,0,1]=(0.905384007172)*x_ref[0]**o+(-0.661067390314)*x_ref[0]+(-0.470790632295)*x_ref[1]**o+(-0.149608017616)*x_ref[1]+(0.427879278267)*x_ref[2]**o+(0.987040179184)*x_ref[2]
        ref[3,0,1,0]=(-0.946562006809)*x_ref[0]**o+(0.64141909633)*x_ref[0]+(-0.762076826276)*x_ref[1]**o+(-0.642912850794)*x_ref[1]+(0.712567898559)*x_ref[2]**o+(-0.292128944205)*x_ref[2]
        ref[3,0,1,1]=(-0.955087949361)*x_ref[0]**o+(0.426588041133)*x_ref[0]+(0.339313104127)*x_ref[1]**o+(-0.375147344045)*x_ref[1]+(0.967259014679)*x_ref[2]**o+(0.504521081858)*x_ref[2]
        ref[3,0,2,0]=(0.141876598344)*x_ref[0]**o+(-0.335579252825)*x_ref[0]+(0.514552993602)*x_ref[1]**o+(-0.531116948499)*x_ref[1]+(-0.369319413141)*x_ref[2]**o+(0.0374413262362)*x_ref[2]
        ref[3,0,2,1]=(0.678360463658)*x_ref[0]**o+(-0.357900119646)*x_ref[0]+(0.300750717059)*x_ref[1]**o+(0.516196189448)*x_ref[1]+(0.0535466018203)*x_ref[2]**o+(-0.149961940786)*x_ref[2]
        ref[3,1,0,0]=(0.0975994351079)*x_ref[0]**o+(0.634544068499)*x_ref[0]+(-0.963217046352)*x_ref[1]**o+(-0.804658091456)*x_ref[1]+(-0.620132720823)*x_ref[2]**o+(0.862250619369)*x_ref[2]
        ref[3,1,0,1]=(0.210306328879)*x_ref[0]**o+(-0.457341187966)*x_ref[0]+(0.957556084801)*x_ref[1]**o+(0.704073170928)*x_ref[1]+(0.423134492133)*x_ref[2]**o+(-0.167268375412)*x_ref[2]
        ref[3,1,1,0]=(-0.909228817744)*x_ref[0]**o+(0.110665018921)*x_ref[0]+(0.36275630147)*x_ref[1]**o+(0.346866824763)*x_ref[1]+(-0.535600547489)*x_ref[2]**o+(0.627638814375)*x_ref[2]
        ref[3,1,1,1]=(0.328780933823)*x_ref[0]**o+(-0.249563994161)*x_ref[0]+(-0.555789062892)*x_ref[1]**o+(-0.917287525036)*x_ref[1]+(0.386191175017)*x_ref[2]**o+(0.123904319061)*x_ref[2]
        ref[3,1,2,0]=(-0.838409901623)*x_ref[0]**o+(0.673479705393)*x_ref[0]+(0.482921104917)*x_ref[1]**o+(0.136411388588)*x_ref[1]+(0.583674116225)*x_ref[2]**o+(0.305100836877)*x_ref[2]
        ref[3,1,2,1]=(0.952961008312)*x_ref[0]**o+(-0.0501981079532)*x_ref[0]+(0.661434464642)*x_ref[1]**o+(0.65030469399)*x_ref[1]+(-0.157562001295)*x_ref[2]**o+(-0.249551892917)*x_ref[2]
        ref[3,2,0,0]=(-0.0232804500286)*x_ref[0]**o+(-0.559229392203)*x_ref[0]+(0.127967930273)*x_ref[1]**o+(-0.635827181645)*x_ref[1]+(-0.0426703224383)*x_ref[2]**o+(0.47720230258)*x_ref[2]
        ref[3,2,0,1]=(-0.421658996106)*x_ref[0]**o+(-0.468260038077)*x_ref[0]+(0.479083572554)*x_ref[1]**o+(-0.214417881841)*x_ref[1]+(0.372430534261)*x_ref[2]**o+(0.133063789546)*x_ref[2]
        ref[3,2,1,0]=(0.905811586567)*x_ref[0]**o+(0.245447340384)*x_ref[0]+(-0.337650999146)*x_ref[1]**o+(0.908188235976)*x_ref[1]+(-0.804029751277)*x_ref[2]**o+(0.0512676456643)*x_ref[2]
        ref[3,2,1,1]=(-0.891486839409)*x_ref[0]**o+(0.20753794329)*x_ref[0]+(0.879183464252)*x_ref[1]**o+(0.902028906992)*x_ref[1]+(0.416776813598)*x_ref[2]**o+(0.491964374373)*x_ref[2]
        ref[3,2,2,0]=(0.653067027119)*x_ref[0]**o+(0.765293890745)*x_ref[0]+(-0.348902169508)*x_ref[1]**o+(0.324429035849)*x_ref[1]+(-0.363026365065)*x_ref[2]**o+(0.0816186405511)*x_ref[2]
        ref[3,2,2,1]=(-0.0415375052325)*x_ref[0]**o+(-0.480779075784)*x_ref[0]+(-0.137223552938)*x_ref[1]**o+(-0.915733625512)*x_ref[1]+(0.750199445159)*x_ref[2]**o+(-0.789818945171)*x_ref[2]
        ref[3,3,0,0]=(0.0744921239083)*x_ref[0]**o+(0.0700589144159)*x_ref[0]+(-0.327105425297)*x_ref[1]**o+(0.469843257765)*x_ref[1]+(-0.170191600844)*x_ref[2]**o+(0.732071534155)*x_ref[2]
        ref[3,3,0,1]=(0.172643927441)*x_ref[0]**o+(0.848924114383)*x_ref[0]+(-0.79125762228)*x_ref[1]**o+(-0.694725775976)*x_ref[1]+(-0.309710503985)*x_ref[2]**o+(-0.405310234927)*x_ref[2]
        ref[3,3,1,0]=(-0.465911787595)*x_ref[0]**o+(-0.71080474709)*x_ref[0]+(0.959925934203)*x_ref[1]**o+(-0.111110644825)*x_ref[1]+(-0.145349925218)*x_ref[2]**o+(0.122955165811)*x_ref[2]
        ref[3,3,1,1]=(-0.769111863754)*x_ref[0]**o+(0.348036938532)*x_ref[0]+(0.485890856474)*x_ref[1]**o+(0.382595218985)*x_ref[1]+(-0.897488692981)*x_ref[2]**o+(0.830822333819)*x_ref[2]
        ref[3,3,2,0]=(0.341890793681)*x_ref[0]**o+(-0.667145936571)*x_ref[0]+(-0.145094108268)*x_ref[1]**o+(0.771433911212)*x_ref[1]+(-0.299752164325)*x_ref[2]**o+(-0.357856549677)*x_ref[2]
        ref[3,3,2,1]=(0.26381838796)*x_ref[0]**o+(-0.44474724998)*x_ref[0]+(0.310283988195)*x_ref[1]**o+(0.698675181314)*x_ref[1]+(0.437624133238)*x_ref[2]**o+(0.394916054445)*x_ref[2]
        ref[3,4,0,0]=(-0.576499406674)*x_ref[0]**o+(0.770936634959)*x_ref[0]+(0.707441735783)*x_ref[1]**o+(-0.794953257043)*x_ref[1]+(-0.535342943209)*x_ref[2]**o+(-0.509446811787)*x_ref[2]
        ref[3,4,0,1]=(0.161115012113)*x_ref[0]**o+(0.11079092252)*x_ref[0]+(0.0633804181164)*x_ref[1]**o+(-0.406049293019)*x_ref[1]+(-0.824283994336)*x_ref[2]**o+(-0.481107560499)*x_ref[2]
        ref[3,4,1,0]=(-0.411133810295)*x_ref[0]**o+(0.988486625099)*x_ref[0]+(-0.36050191367)*x_ref[1]**o+(0.469479821965)*x_ref[1]+(-0.431175357325)*x_ref[2]**o+(-0.966295015263)*x_ref[2]
        ref[3,4,1,1]=(0.233731410367)*x_ref[0]**o+(-0.821900740766)*x_ref[0]+(-0.823123461385)*x_ref[1]**o+(-0.380537914421)*x_ref[1]+(-0.896146300355)*x_ref[2]**o+(-0.775172990257)*x_ref[2]
        ref[3,4,2,0]=(-0.96323363586)*x_ref[0]**o+(0.0935412405245)*x_ref[0]+(0.929333026259)*x_ref[1]**o+(-0.322236627348)*x_ref[1]+(-0.390898135964)*x_ref[2]**o+(-0.766178290162)*x_ref[2]
        ref[3,4,2,1]=(-0.773670533268)*x_ref[0]**o+(-0.156975157098)*x_ref[0]+(-0.91512996795)*x_ref[1]**o+(0.0247803682126)*x_ref[1]+(0.995159733782)*x_ref[2]**o+(-0.669224144225)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.0071553637759)*x[0]**o+(0.357615094637)*x[0]+(0.980385970972)*x[1]**o+(0.423505273491)*x[1]
        ref=(-0.0071553637759)*x_ref[0]**o+(0.357615094637)*x_ref[0]+(0.980385970972)*x_ref[1]**o+(0.423505273491)*x_ref[1]
      else:
        arg=(0.546251547514)*x[0]**o+(-0.530405447164)*x[0]+(-0.901465213708)*x[1]**o+(-0.986019835171)*x[1]+(-0.731836172089)*x[2]**o+(0.377317674082)*x[2]
        ref=(0.546251547514)*x_ref[0]**o+(-0.530405447164)*x_ref[0]+(-0.901465213708)*x_ref[1]**o+(-0.986019835171)*x_ref[1]+(-0.731836172089)*x_ref[2]**o+(0.377317674082)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.138216822552)*x[0]**o+(0.908302047805)*x[0]+(-0.247600939209)*x[1]**o+(0.776239000167)*x[1]
        arg[1]=(-0.577926899073)*x[0]**o+(0.714263102346)*x[0]+(-0.343808625475)*x[1]**o+(0.786295563423)*x[1]
        ref[0]=(-0.138216822552)*x_ref[0]**o+(0.908302047805)*x_ref[0]+(-0.247600939209)*x_ref[1]**o+(0.776239000167)*x_ref[1]
        ref[1]=(-0.577926899073)*x_ref[0]**o+(0.714263102346)*x_ref[0]+(-0.343808625475)*x_ref[1]**o+(0.786295563423)*x_ref[1]
      else:
        arg[0]=(0.020232467504)*x[0]**o+(-0.0449727381653)*x[0]+(0.00132143537494)*x[1]**o+(-0.870292414095)*x[1]+(-0.926731511382)*x[2]**o+(-0.398915622332)*x[2]
        arg[1]=(0.203864545604)*x[0]**o+(0.150501673382)*x[0]+(0.132132290114)*x[1]**o+(-0.720617456803)*x[1]+(0.673909788445)*x[2]**o+(0.145376368499)*x[2]
        ref[0]=(0.020232467504)*x_ref[0]**o+(-0.0449727381653)*x_ref[0]+(0.00132143537494)*x_ref[1]**o+(-0.870292414095)*x_ref[1]+(-0.926731511382)*x_ref[2]**o+(-0.398915622332)*x_ref[2]
        ref[1]=(0.203864545604)*x_ref[0]**o+(0.150501673382)*x_ref[0]+(0.132132290114)*x_ref[1]**o+(-0.720617456803)*x_ref[1]+(0.673909788445)*x_ref[2]**o+(0.145376368499)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.344685653483)*x[0]**o+(-0.618443209029)*x[0]+(-0.338246837583)*x[1]**o+(-0.709311881709)*x[1]
        arg[0,1]=(-0.176341370817)*x[0]**o+(-0.230030832268)*x[0]+(-0.335163038431)*x[1]**o+(0.278627093006)*x[1]
        arg[0,2]=(0.376270607923)*x[0]**o+(0.895822309578)*x[0]+(0.418316736671)*x[1]**o+(-0.951461687587)*x[1]
        arg[0,3]=(0.69925470532)*x[0]**o+(-0.829917904706)*x[0]+(0.60851481091)*x[1]**o+(-0.880924691692)*x[1]
        arg[0,4]=(0.537570501499)*x[0]**o+(0.584091505535)*x[0]+(-0.0045669307163)*x[1]**o+(0.162714723047)*x[1]
        arg[1,0]=(-0.244615538677)*x[0]**o+(-0.0168359545772)*x[0]+(-0.511097014251)*x[1]**o+(-0.476830089968)*x[1]
        arg[1,1]=(0.350953788855)*x[0]**o+(0.0859931352733)*x[0]+(-0.565664846934)*x[1]**o+(-0.230383210864)*x[1]
        arg[1,2]=(0.787537227138)*x[0]**o+(-0.97589028834)*x[0]+(0.978870857538)*x[1]**o+(0.436540016432)*x[1]
        arg[1,3]=(-0.806777674064)*x[0]**o+(-0.384971723467)*x[0]+(-0.423836322283)*x[1]**o+(0.470609842503)*x[1]
        arg[1,4]=(0.288422170494)*x[0]**o+(-0.467670464119)*x[0]+(-0.616952048135)*x[1]**o+(0.604831335817)*x[1]
        arg[2,0]=(0.0461969205986)*x[0]**o+(0.663083190578)*x[0]+(-0.673597938582)*x[1]**o+(0.217389040208)*x[1]
        arg[2,1]=(-0.32008443048)*x[0]**o+(-0.970854886309)*x[0]+(0.630220968872)*x[1]**o+(0.827824011117)*x[1]
        arg[2,2]=(-0.232867522682)*x[0]**o+(-0.17921445933)*x[0]+(-0.565289194319)*x[1]**o+(0.0251729876325)*x[1]
        arg[2,3]=(-0.862002880133)*x[0]**o+(0.64790108714)*x[0]+(-0.469428882373)*x[1]**o+(0.50581485158)*x[1]
        arg[2,4]=(-0.957389923449)*x[0]**o+(-0.716972488151)*x[0]+(0.848168933491)*x[1]**o+(0.78614576485)*x[1]
        arg[3,0]=(0.417890378265)*x[0]**o+(-0.128884271967)*x[0]+(0.981251819451)*x[1]**o+(0.837945499182)*x[1]
        arg[3,1]=(-0.296791907211)*x[0]**o+(0.913966659909)*x[0]+(0.0171079547828)*x[1]**o+(0.472922286638)*x[1]
        arg[3,2]=(-0.549504779354)*x[0]**o+(0.356989069435)*x[0]+(-0.0978494385269)*x[1]**o+(0.59354694599)*x[1]
        arg[3,3]=(0.0272055308458)*x[0]**o+(0.560184544705)*x[0]+(-0.825452907104)*x[1]**o+(0.701919616153)*x[1]
        arg[3,4]=(-0.33094753848)*x[0]**o+(0.877723269108)*x[0]+(-0.491915661514)*x[1]**o+(-0.98784807255)*x[1]
        ref[0,0]=(-0.344685653483)*x_ref[0]**o+(-0.618443209029)*x_ref[0]+(-0.338246837583)*x_ref[1]**o+(-0.709311881709)*x_ref[1]
        ref[0,1]=(-0.176341370817)*x_ref[0]**o+(-0.230030832268)*x_ref[0]+(-0.335163038431)*x_ref[1]**o+(0.278627093006)*x_ref[1]
        ref[0,2]=(0.376270607923)*x_ref[0]**o+(0.895822309578)*x_ref[0]+(0.418316736671)*x_ref[1]**o+(-0.951461687587)*x_ref[1]
        ref[0,3]=(0.69925470532)*x_ref[0]**o+(-0.829917904706)*x_ref[0]+(0.60851481091)*x_ref[1]**o+(-0.880924691692)*x_ref[1]
        ref[0,4]=(0.537570501499)*x_ref[0]**o+(0.584091505535)*x_ref[0]+(-0.0045669307163)*x_ref[1]**o+(0.162714723047)*x_ref[1]
        ref[1,0]=(-0.244615538677)*x_ref[0]**o+(-0.0168359545772)*x_ref[0]+(-0.511097014251)*x_ref[1]**o+(-0.476830089968)*x_ref[1]
        ref[1,1]=(0.350953788855)*x_ref[0]**o+(0.0859931352733)*x_ref[0]+(-0.565664846934)*x_ref[1]**o+(-0.230383210864)*x_ref[1]
        ref[1,2]=(0.787537227138)*x_ref[0]**o+(-0.97589028834)*x_ref[0]+(0.978870857538)*x_ref[1]**o+(0.436540016432)*x_ref[1]
        ref[1,3]=(-0.806777674064)*x_ref[0]**o+(-0.384971723467)*x_ref[0]+(-0.423836322283)*x_ref[1]**o+(0.470609842503)*x_ref[1]
        ref[1,4]=(0.288422170494)*x_ref[0]**o+(-0.467670464119)*x_ref[0]+(-0.616952048135)*x_ref[1]**o+(0.604831335817)*x_ref[1]
        ref[2,0]=(0.0461969205986)*x_ref[0]**o+(0.663083190578)*x_ref[0]+(-0.673597938582)*x_ref[1]**o+(0.217389040208)*x_ref[1]
        ref[2,1]=(-0.32008443048)*x_ref[0]**o+(-0.970854886309)*x_ref[0]+(0.630220968872)*x_ref[1]**o+(0.827824011117)*x_ref[1]
        ref[2,2]=(-0.232867522682)*x_ref[0]**o+(-0.17921445933)*x_ref[0]+(-0.565289194319)*x_ref[1]**o+(0.0251729876325)*x_ref[1]
        ref[2,3]=(-0.862002880133)*x_ref[0]**o+(0.64790108714)*x_ref[0]+(-0.469428882373)*x_ref[1]**o+(0.50581485158)*x_ref[1]
        ref[2,4]=(-0.957389923449)*x_ref[0]**o+(-0.716972488151)*x_ref[0]+(0.848168933491)*x_ref[1]**o+(0.78614576485)*x_ref[1]
        ref[3,0]=(0.417890378265)*x_ref[0]**o+(-0.128884271967)*x_ref[0]+(0.981251819451)*x_ref[1]**o+(0.837945499182)*x_ref[1]
        ref[3,1]=(-0.296791907211)*x_ref[0]**o+(0.913966659909)*x_ref[0]+(0.0171079547828)*x_ref[1]**o+(0.472922286638)*x_ref[1]
        ref[3,2]=(-0.549504779354)*x_ref[0]**o+(0.356989069435)*x_ref[0]+(-0.0978494385269)*x_ref[1]**o+(0.59354694599)*x_ref[1]
        ref[3,3]=(0.0272055308458)*x_ref[0]**o+(0.560184544705)*x_ref[0]+(-0.825452907104)*x_ref[1]**o+(0.701919616153)*x_ref[1]
        ref[3,4]=(-0.33094753848)*x_ref[0]**o+(0.877723269108)*x_ref[0]+(-0.491915661514)*x_ref[1]**o+(-0.98784807255)*x_ref[1]
      else:
        arg[0,0]=(-0.434721514594)*x[0]**o+(-0.28074158231)*x[0]+(0.102692621316)*x[1]**o+(0.399677492352)*x[1]+(0.39799987301)*x[2]**o+(-0.134547332796)*x[2]
        arg[0,1]=(-0.814630031471)*x[0]**o+(-0.0722898377573)*x[0]+(0.356831166041)*x[1]**o+(0.0556663518903)*x[1]+(-0.335764152958)*x[2]**o+(0.344564519072)*x[2]
        arg[0,2]=(-0.881698115197)*x[0]**o+(-0.030759288922)*x[0]+(-0.23126177081)*x[1]**o+(0.865011678258)*x[1]+(0.961867471932)*x[2]**o+(-0.04809978604)*x[2]
        arg[0,3]=(-0.931319970946)*x[0]**o+(0.807195614255)*x[0]+(0.116305510032)*x[1]**o+(0.337873586448)*x[1]+(0.523766276637)*x[2]**o+(-0.890502808117)*x[2]
        arg[0,4]=(0.952369561411)*x[0]**o+(0.518186569223)*x[0]+(0.184454818734)*x[1]**o+(-0.198618411156)*x[1]+(0.891317797589)*x[2]**o+(-0.312268228065)*x[2]
        arg[1,0]=(-0.770379397538)*x[0]**o+(0.11068893635)*x[0]+(0.435574518842)*x[1]**o+(-0.895094805657)*x[1]+(0.719886389608)*x[2]**o+(0.617254266829)*x[2]
        arg[1,1]=(-0.968566082536)*x[0]**o+(0.70331821608)*x[0]+(-0.256092697223)*x[1]**o+(0.214482838048)*x[1]+(-0.20817108493)*x[2]**o+(-0.741961694045)*x[2]
        arg[1,2]=(0.928194134464)*x[0]**o+(0.748278982508)*x[0]+(0.459001496702)*x[1]**o+(-0.333954496033)*x[1]+(0.0451659439418)*x[2]**o+(0.0355669781059)*x[2]
        arg[1,3]=(-0.285564661682)*x[0]**o+(-0.780972389596)*x[0]+(0.20442278849)*x[1]**o+(-0.425673038027)*x[1]+(-0.886021022849)*x[2]**o+(0.2231241391)*x[2]
        arg[1,4]=(0.48756511391)*x[0]**o+(0.904219456651)*x[0]+(-0.723245954496)*x[1]**o+(-0.876483077689)*x[1]+(0.454913273237)*x[2]**o+(0.907651543551)*x[2]
        arg[2,0]=(-0.490623674798)*x[0]**o+(0.179561344392)*x[0]+(-0.257796709265)*x[1]**o+(0.470216680739)*x[1]+(-0.334946805819)*x[2]**o+(-0.295308708488)*x[2]
        arg[2,1]=(-0.0753731003603)*x[0]**o+(0.848622766023)*x[0]+(-0.168762035291)*x[1]**o+(-0.132048839333)*x[1]+(0.229089231051)*x[2]**o+(0.280032186468)*x[2]
        arg[2,2]=(0.601432167354)*x[0]**o+(-0.0857688429764)*x[0]+(0.556763499853)*x[1]**o+(-0.0953568285359)*x[1]+(-0.396589175262)*x[2]**o+(0.437257052414)*x[2]
        arg[2,3]=(0.239710792478)*x[0]**o+(0.255863660922)*x[0]+(-0.860831814227)*x[1]**o+(0.416057968294)*x[1]+(0.284515440811)*x[2]**o+(-0.0737992548315)*x[2]
        arg[2,4]=(-0.86884373286)*x[0]**o+(0.306811019609)*x[0]+(0.955007347168)*x[1]**o+(0.731959544948)*x[1]+(0.326947842619)*x[2]**o+(0.430618227986)*x[2]
        arg[3,0]=(0.450341287007)*x[0]**o+(-0.0318463261779)*x[0]+(-0.755083793007)*x[1]**o+(0.427706480425)*x[1]+(0.224383931841)*x[2]**o+(-0.801200151542)*x[2]
        arg[3,1]=(-0.750405028813)*x[0]**o+(-0.531057162102)*x[0]+(0.120116357521)*x[1]**o+(-0.247640504917)*x[1]+(-0.360494371461)*x[2]**o+(-0.496404068337)*x[2]
        arg[3,2]=(-0.202775617449)*x[0]**o+(0.833337604045)*x[0]+(0.551285270602)*x[1]**o+(0.266630410234)*x[1]+(0.903041333927)*x[2]**o+(0.527233939028)*x[2]
        arg[3,3]=(0.12892683791)*x[0]**o+(-0.486726243104)*x[0]+(-0.340728397711)*x[1]**o+(-0.670692588507)*x[1]+(0.832250132781)*x[2]**o+(0.276897913372)*x[2]
        arg[3,4]=(0.839441639419)*x[0]**o+(0.362150671815)*x[0]+(0.728370029046)*x[1]**o+(0.149540296968)*x[1]+(-0.0418719598944)*x[2]**o+(-0.331805297414)*x[2]
        ref[0,0]=(-0.434721514594)*x_ref[0]**o+(-0.28074158231)*x_ref[0]+(0.102692621316)*x_ref[1]**o+(0.399677492352)*x_ref[1]+(0.39799987301)*x_ref[2]**o+(-0.134547332796)*x_ref[2]
        ref[0,1]=(-0.814630031471)*x_ref[0]**o+(-0.0722898377573)*x_ref[0]+(0.356831166041)*x_ref[1]**o+(0.0556663518903)*x_ref[1]+(-0.335764152958)*x_ref[2]**o+(0.344564519072)*x_ref[2]
        ref[0,2]=(-0.881698115197)*x_ref[0]**o+(-0.030759288922)*x_ref[0]+(-0.23126177081)*x_ref[1]**o+(0.865011678258)*x_ref[1]+(0.961867471932)*x_ref[2]**o+(-0.04809978604)*x_ref[2]
        ref[0,3]=(-0.931319970946)*x_ref[0]**o+(0.807195614255)*x_ref[0]+(0.116305510032)*x_ref[1]**o+(0.337873586448)*x_ref[1]+(0.523766276637)*x_ref[2]**o+(-0.890502808117)*x_ref[2]
        ref[0,4]=(0.952369561411)*x_ref[0]**o+(0.518186569223)*x_ref[0]+(0.184454818734)*x_ref[1]**o+(-0.198618411156)*x_ref[1]+(0.891317797589)*x_ref[2]**o+(-0.312268228065)*x_ref[2]
        ref[1,0]=(-0.770379397538)*x_ref[0]**o+(0.11068893635)*x_ref[0]+(0.435574518842)*x_ref[1]**o+(-0.895094805657)*x_ref[1]+(0.719886389608)*x_ref[2]**o+(0.617254266829)*x_ref[2]
        ref[1,1]=(-0.968566082536)*x_ref[0]**o+(0.70331821608)*x_ref[0]+(-0.256092697223)*x_ref[1]**o+(0.214482838048)*x_ref[1]+(-0.20817108493)*x_ref[2]**o+(-0.741961694045)*x_ref[2]
        ref[1,2]=(0.928194134464)*x_ref[0]**o+(0.748278982508)*x_ref[0]+(0.459001496702)*x_ref[1]**o+(-0.333954496033)*x_ref[1]+(0.0451659439418)*x_ref[2]**o+(0.0355669781059)*x_ref[2]
        ref[1,3]=(-0.285564661682)*x_ref[0]**o+(-0.780972389596)*x_ref[0]+(0.20442278849)*x_ref[1]**o+(-0.425673038027)*x_ref[1]+(-0.886021022849)*x_ref[2]**o+(0.2231241391)*x_ref[2]
        ref[1,4]=(0.48756511391)*x_ref[0]**o+(0.904219456651)*x_ref[0]+(-0.723245954496)*x_ref[1]**o+(-0.876483077689)*x_ref[1]+(0.454913273237)*x_ref[2]**o+(0.907651543551)*x_ref[2]
        ref[2,0]=(-0.490623674798)*x_ref[0]**o+(0.179561344392)*x_ref[0]+(-0.257796709265)*x_ref[1]**o+(0.470216680739)*x_ref[1]+(-0.334946805819)*x_ref[2]**o+(-0.295308708488)*x_ref[2]
        ref[2,1]=(-0.0753731003603)*x_ref[0]**o+(0.848622766023)*x_ref[0]+(-0.168762035291)*x_ref[1]**o+(-0.132048839333)*x_ref[1]+(0.229089231051)*x_ref[2]**o+(0.280032186468)*x_ref[2]
        ref[2,2]=(0.601432167354)*x_ref[0]**o+(-0.0857688429764)*x_ref[0]+(0.556763499853)*x_ref[1]**o+(-0.0953568285359)*x_ref[1]+(-0.396589175262)*x_ref[2]**o+(0.437257052414)*x_ref[2]
        ref[2,3]=(0.239710792478)*x_ref[0]**o+(0.255863660922)*x_ref[0]+(-0.860831814227)*x_ref[1]**o+(0.416057968294)*x_ref[1]+(0.284515440811)*x_ref[2]**o+(-0.0737992548315)*x_ref[2]
        ref[2,4]=(-0.86884373286)*x_ref[0]**o+(0.306811019609)*x_ref[0]+(0.955007347168)*x_ref[1]**o+(0.731959544948)*x_ref[1]+(0.326947842619)*x_ref[2]**o+(0.430618227986)*x_ref[2]
        ref[3,0]=(0.450341287007)*x_ref[0]**o+(-0.0318463261779)*x_ref[0]+(-0.755083793007)*x_ref[1]**o+(0.427706480425)*x_ref[1]+(0.224383931841)*x_ref[2]**o+(-0.801200151542)*x_ref[2]
        ref[3,1]=(-0.750405028813)*x_ref[0]**o+(-0.531057162102)*x_ref[0]+(0.120116357521)*x_ref[1]**o+(-0.247640504917)*x_ref[1]+(-0.360494371461)*x_ref[2]**o+(-0.496404068337)*x_ref[2]
        ref[3,2]=(-0.202775617449)*x_ref[0]**o+(0.833337604045)*x_ref[0]+(0.551285270602)*x_ref[1]**o+(0.266630410234)*x_ref[1]+(0.903041333927)*x_ref[2]**o+(0.527233939028)*x_ref[2]
        ref[3,3]=(0.12892683791)*x_ref[0]**o+(-0.486726243104)*x_ref[0]+(-0.340728397711)*x_ref[1]**o+(-0.670692588507)*x_ref[1]+(0.832250132781)*x_ref[2]**o+(0.276897913372)*x_ref[2]
        ref[3,4]=(0.839441639419)*x_ref[0]**o+(0.362150671815)*x_ref[0]+(0.728370029046)*x_ref[1]**o+(0.149540296968)*x_ref[1]+(-0.0418719598944)*x_ref[2]**o+(-0.331805297414)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.475615092269)*x[0]**o+(-0.0893751722575)*x[0]+(0.595881277166)*x[1]**o+(0.397900173915)*x[1]
        arg[0,0,1]=(-0.552614805712)*x[0]**o+(0.14072435445)*x[0]+(0.290231845094)*x[1]**o+(0.573878059041)*x[1]
        arg[0,1,0]=(-0.226148312038)*x[0]**o+(-0.737524507359)*x[0]+(-0.704362709293)*x[1]**o+(-0.197223346266)*x[1]
        arg[0,1,1]=(0.844313529642)*x[0]**o+(-0.948514286846)*x[0]+(0.383754990882)*x[1]**o+(0.629819148632)*x[1]
        arg[1,0,0]=(-0.519122333314)*x[0]**o+(-0.895333584981)*x[0]+(-0.941503934561)*x[1]**o+(0.574440495425)*x[1]
        arg[1,0,1]=(0.34967649382)*x[0]**o+(-0.332598318229)*x[0]+(0.0587818420974)*x[1]**o+(0.0848419508498)*x[1]
        arg[1,1,0]=(0.126943263196)*x[0]**o+(0.20287950383)*x[0]+(-0.305454685691)*x[1]**o+(-0.357695790052)*x[1]
        arg[1,1,1]=(0.104154708153)*x[0]**o+(0.867435765411)*x[0]+(0.656946071961)*x[1]**o+(0.96433811362)*x[1]
        arg[2,0,0]=(-0.407362949457)*x[0]**o+(-0.171021521273)*x[0]+(0.901342300398)*x[1]**o+(-0.0360476097399)*x[1]
        arg[2,0,1]=(-0.709421035652)*x[0]**o+(0.949318093375)*x[0]+(0.155061840049)*x[1]**o+(0.828999317354)*x[1]
        arg[2,1,0]=(-0.573031939083)*x[0]**o+(-0.766186602896)*x[0]+(-0.0483071013367)*x[1]**o+(0.741988754559)*x[1]
        arg[2,1,1]=(-0.826747727748)*x[0]**o+(0.481970220735)*x[0]+(0.735617076451)*x[1]**o+(-0.64482586429)*x[1]
        arg[3,0,0]=(0.0834619008603)*x[0]**o+(-0.122260004323)*x[0]+(-0.193792114997)*x[1]**o+(0.635792828013)*x[1]
        arg[3,0,1]=(0.868630978759)*x[0]**o+(0.54474614716)*x[0]+(-0.514377753448)*x[1]**o+(0.58889412844)*x[1]
        arg[3,1,0]=(-0.233744744882)*x[0]**o+(0.883601383004)*x[0]+(0.575012113003)*x[1]**o+(0.185704140149)*x[1]
        arg[3,1,1]=(-0.948893544046)*x[0]**o+(0.851870038501)*x[0]+(-0.591835259957)*x[1]**o+(-0.958375628808)*x[1]
        arg[4,0,0]=(0.981685480559)*x[0]**o+(0.718249744763)*x[0]+(0.589534430546)*x[1]**o+(0.72294148429)*x[1]
        arg[4,0,1]=(-0.603601154465)*x[0]**o+(0.293688738803)*x[0]+(0.151726724518)*x[1]**o+(0.56316377967)*x[1]
        arg[4,1,0]=(0.539033109011)*x[0]**o+(0.791686885872)*x[0]+(0.878973377358)*x[1]**o+(0.869606341625)*x[1]
        arg[4,1,1]=(-0.0920873290752)*x[0]**o+(-0.132682060309)*x[0]+(0.34023826981)*x[1]**o+(0.585476816741)*x[1]
        arg[5,0,0]=(-0.430989548175)*x[0]**o+(-0.525387682044)*x[0]+(-0.996483829778)*x[1]**o+(-0.564503119535)*x[1]
        arg[5,0,1]=(0.602911018749)*x[0]**o+(-0.348972500704)*x[0]+(0.174681048071)*x[1]**o+(-0.94551825287)*x[1]
        arg[5,1,0]=(0.780926392259)*x[0]**o+(-0.164977652547)*x[0]+(0.224726701017)*x[1]**o+(-0.779109318124)*x[1]
        arg[5,1,1]=(0.470928696402)*x[0]**o+(-0.505871268699)*x[0]+(0.407158974816)*x[1]**o+(0.392856659954)*x[1]
        ref[0,0,0]=(-0.475615092269)*x_ref[0]**o+(-0.0893751722575)*x_ref[0]+(0.595881277166)*x_ref[1]**o+(0.397900173915)*x_ref[1]
        ref[0,0,1]=(-0.552614805712)*x_ref[0]**o+(0.14072435445)*x_ref[0]+(0.290231845094)*x_ref[1]**o+(0.573878059041)*x_ref[1]
        ref[0,1,0]=(-0.226148312038)*x_ref[0]**o+(-0.737524507359)*x_ref[0]+(-0.704362709293)*x_ref[1]**o+(-0.197223346266)*x_ref[1]
        ref[0,1,1]=(0.844313529642)*x_ref[0]**o+(-0.948514286846)*x_ref[0]+(0.383754990882)*x_ref[1]**o+(0.629819148632)*x_ref[1]
        ref[1,0,0]=(-0.519122333314)*x_ref[0]**o+(-0.895333584981)*x_ref[0]+(-0.941503934561)*x_ref[1]**o+(0.574440495425)*x_ref[1]
        ref[1,0,1]=(0.34967649382)*x_ref[0]**o+(-0.332598318229)*x_ref[0]+(0.0587818420974)*x_ref[1]**o+(0.0848419508498)*x_ref[1]
        ref[1,1,0]=(0.126943263196)*x_ref[0]**o+(0.20287950383)*x_ref[0]+(-0.305454685691)*x_ref[1]**o+(-0.357695790052)*x_ref[1]
        ref[1,1,1]=(0.104154708153)*x_ref[0]**o+(0.867435765411)*x_ref[0]+(0.656946071961)*x_ref[1]**o+(0.96433811362)*x_ref[1]
        ref[2,0,0]=(-0.407362949457)*x_ref[0]**o+(-0.171021521273)*x_ref[0]+(0.901342300398)*x_ref[1]**o+(-0.0360476097399)*x_ref[1]
        ref[2,0,1]=(-0.709421035652)*x_ref[0]**o+(0.949318093375)*x_ref[0]+(0.155061840049)*x_ref[1]**o+(0.828999317354)*x_ref[1]
        ref[2,1,0]=(-0.573031939083)*x_ref[0]**o+(-0.766186602896)*x_ref[0]+(-0.0483071013367)*x_ref[1]**o+(0.741988754559)*x_ref[1]
        ref[2,1,1]=(-0.826747727748)*x_ref[0]**o+(0.481970220735)*x_ref[0]+(0.735617076451)*x_ref[1]**o+(-0.64482586429)*x_ref[1]
        ref[3,0,0]=(0.0834619008603)*x_ref[0]**o+(-0.122260004323)*x_ref[0]+(-0.193792114997)*x_ref[1]**o+(0.635792828013)*x_ref[1]
        ref[3,0,1]=(0.868630978759)*x_ref[0]**o+(0.54474614716)*x_ref[0]+(-0.514377753448)*x_ref[1]**o+(0.58889412844)*x_ref[1]
        ref[3,1,0]=(-0.233744744882)*x_ref[0]**o+(0.883601383004)*x_ref[0]+(0.575012113003)*x_ref[1]**o+(0.185704140149)*x_ref[1]
        ref[3,1,1]=(-0.948893544046)*x_ref[0]**o+(0.851870038501)*x_ref[0]+(-0.591835259957)*x_ref[1]**o+(-0.958375628808)*x_ref[1]
        ref[4,0,0]=(0.981685480559)*x_ref[0]**o+(0.718249744763)*x_ref[0]+(0.589534430546)*x_ref[1]**o+(0.72294148429)*x_ref[1]
        ref[4,0,1]=(-0.603601154465)*x_ref[0]**o+(0.293688738803)*x_ref[0]+(0.151726724518)*x_ref[1]**o+(0.56316377967)*x_ref[1]
        ref[4,1,0]=(0.539033109011)*x_ref[0]**o+(0.791686885872)*x_ref[0]+(0.878973377358)*x_ref[1]**o+(0.869606341625)*x_ref[1]
        ref[4,1,1]=(-0.0920873290752)*x_ref[0]**o+(-0.132682060309)*x_ref[0]+(0.34023826981)*x_ref[1]**o+(0.585476816741)*x_ref[1]
        ref[5,0,0]=(-0.430989548175)*x_ref[0]**o+(-0.525387682044)*x_ref[0]+(-0.996483829778)*x_ref[1]**o+(-0.564503119535)*x_ref[1]
        ref[5,0,1]=(0.602911018749)*x_ref[0]**o+(-0.348972500704)*x_ref[0]+(0.174681048071)*x_ref[1]**o+(-0.94551825287)*x_ref[1]
        ref[5,1,0]=(0.780926392259)*x_ref[0]**o+(-0.164977652547)*x_ref[0]+(0.224726701017)*x_ref[1]**o+(-0.779109318124)*x_ref[1]
        ref[5,1,1]=(0.470928696402)*x_ref[0]**o+(-0.505871268699)*x_ref[0]+(0.407158974816)*x_ref[1]**o+(0.392856659954)*x_ref[1]
      else:
        arg[0,0,0]=(0.759590374826)*x[0]**o+(-0.776182110267)*x[0]+(0.713871101133)*x[1]**o+(-0.0991655752543)*x[1]+(0.339491998788)*x[2]**o+(-0.624141085669)*x[2]
        arg[0,0,1]=(-0.948679111098)*x[0]**o+(0.0918139408797)*x[0]+(-0.519481784001)*x[1]**o+(-0.480754367303)*x[1]+(0.996421472879)*x[2]**o+(-0.933032853651)*x[2]
        arg[0,1,0]=(-0.824750788431)*x[0]**o+(0.846735099726)*x[0]+(-0.436835905895)*x[1]**o+(-0.732363323611)*x[1]+(-0.865610972073)*x[2]**o+(0.754759759319)*x[2]
        arg[0,1,1]=(-0.403783782613)*x[0]**o+(0.918875237045)*x[0]+(0.0934155750366)*x[1]**o+(0.832477256884)*x[1]+(0.615080190605)*x[2]**o+(0.000920074523901)*x[2]
        arg[1,0,0]=(0.260531941775)*x[0]**o+(-0.514647294043)*x[0]+(0.237113978946)*x[1]**o+(-0.116626766689)*x[1]+(-0.670422853592)*x[2]**o+(0.574941151693)*x[2]
        arg[1,0,1]=(-0.0133773984155)*x[0]**o+(0.286273972547)*x[0]+(0.0493985332233)*x[1]**o+(-0.567835671454)*x[1]+(0.508314781519)*x[2]**o+(-0.603506084784)*x[2]
        arg[1,1,0]=(0.0237413792908)*x[0]**o+(-0.758893281773)*x[0]+(0.150670689461)*x[1]**o+(0.136386516616)*x[1]+(-0.440203817756)*x[2]**o+(0.796610908529)*x[2]
        arg[1,1,1]=(-0.223549634917)*x[0]**o+(0.396472946824)*x[0]+(0.47830083818)*x[1]**o+(0.059933592631)*x[1]+(0.709371074379)*x[2]**o+(-0.129442424949)*x[2]
        arg[2,0,0]=(0.686170679191)*x[0]**o+(-0.118668021586)*x[0]+(0.033954449978)*x[1]**o+(-0.0508661210104)*x[1]+(0.352824724076)*x[2]**o+(-0.772874133692)*x[2]
        arg[2,0,1]=(0.434879772789)*x[0]**o+(-0.148309931356)*x[0]+(-0.013792688421)*x[1]**o+(0.597899709109)*x[1]+(-0.280336465565)*x[2]**o+(-0.177134087728)*x[2]
        arg[2,1,0]=(-0.636844238715)*x[0]**o+(-0.646269255495)*x[0]+(0.243236542558)*x[1]**o+(0.203480511801)*x[1]+(0.693290035164)*x[2]**o+(0.773354114662)*x[2]
        arg[2,1,1]=(-0.161998867128)*x[0]**o+(0.193608078775)*x[0]+(0.607954851405)*x[1]**o+(0.648078354644)*x[1]+(-0.488891750936)*x[2]**o+(0.912860597431)*x[2]
        arg[3,0,0]=(-0.696909279028)*x[0]**o+(0.039558795649)*x[0]+(0.441121646773)*x[1]**o+(-0.0681790925087)*x[1]+(-0.792437910441)*x[2]**o+(-0.512942395878)*x[2]
        arg[3,0,1]=(-0.264832490825)*x[0]**o+(-0.955976365474)*x[0]+(-0.884641387524)*x[1]**o+(-0.736970630699)*x[1]+(-0.0525378155791)*x[2]**o+(0.149164361449)*x[2]
        arg[3,1,0]=(0.434888951175)*x[0]**o+(-0.218511432369)*x[0]+(0.564377624401)*x[1]**o+(-0.0978716780106)*x[1]+(0.447857191989)*x[2]**o+(-0.466572586856)*x[2]
        arg[3,1,1]=(0.457689985738)*x[0]**o+(0.00322996620178)*x[0]+(0.307787601144)*x[1]**o+(-0.119455639704)*x[1]+(0.70367723461)*x[2]**o+(0.561913096313)*x[2]
        arg[4,0,0]=(-0.759802700221)*x[0]**o+(0.469649690329)*x[0]+(0.0913972297116)*x[1]**o+(0.310277900593)*x[1]+(-0.590266889319)*x[2]**o+(-0.89711266859)*x[2]
        arg[4,0,1]=(0.85506147544)*x[0]**o+(-0.357476737069)*x[0]+(-0.646898911915)*x[1]**o+(-0.674811608049)*x[1]+(0.865265982147)*x[2]**o+(0.720153126392)*x[2]
        arg[4,1,0]=(0.848146446583)*x[0]**o+(0.868166234685)*x[0]+(0.0477110054774)*x[1]**o+(-0.0338098833736)*x[1]+(-0.360024439914)*x[2]**o+(-0.634359083766)*x[2]
        arg[4,1,1]=(-0.672530624126)*x[0]**o+(0.354170615043)*x[0]+(0.486028226561)*x[1]**o+(0.857532578327)*x[1]+(0.15798171243)*x[2]**o+(0.0579857791862)*x[2]
        arg[5,0,0]=(0.653171799956)*x[0]**o+(0.605427034155)*x[0]+(0.7414424011)*x[1]**o+(-0.807505211605)*x[1]+(-0.0411776868808)*x[2]**o+(0.24301961063)*x[2]
        arg[5,0,1]=(-0.336860102074)*x[0]**o+(0.235697646908)*x[0]+(-0.173217953629)*x[1]**o+(0.989643649714)*x[1]+(0.977460054831)*x[2]**o+(-0.743825027125)*x[2]
        arg[5,1,0]=(0.794904875581)*x[0]**o+(0.607548574046)*x[0]+(0.603180304387)*x[1]**o+(0.276995429843)*x[1]+(0.698731315308)*x[2]**o+(-0.966744108727)*x[2]
        arg[5,1,1]=(-0.0230004766002)*x[0]**o+(0.489585418104)*x[0]+(0.382764947116)*x[1]**o+(0.143873532935)*x[1]+(0.932000957196)*x[2]**o+(0.739975795806)*x[2]
        ref[0,0,0]=(0.759590374826)*x_ref[0]**o+(-0.776182110267)*x_ref[0]+(0.713871101133)*x_ref[1]**o+(-0.0991655752543)*x_ref[1]+(0.339491998788)*x_ref[2]**o+(-0.624141085669)*x_ref[2]
        ref[0,0,1]=(-0.948679111098)*x_ref[0]**o+(0.0918139408797)*x_ref[0]+(-0.519481784001)*x_ref[1]**o+(-0.480754367303)*x_ref[1]+(0.996421472879)*x_ref[2]**o+(-0.933032853651)*x_ref[2]
        ref[0,1,0]=(-0.824750788431)*x_ref[0]**o+(0.846735099726)*x_ref[0]+(-0.436835905895)*x_ref[1]**o+(-0.732363323611)*x_ref[1]+(-0.865610972073)*x_ref[2]**o+(0.754759759319)*x_ref[2]
        ref[0,1,1]=(-0.403783782613)*x_ref[0]**o+(0.918875237045)*x_ref[0]+(0.0934155750366)*x_ref[1]**o+(0.832477256884)*x_ref[1]+(0.615080190605)*x_ref[2]**o+(0.000920074523901)*x_ref[2]
        ref[1,0,0]=(0.260531941775)*x_ref[0]**o+(-0.514647294043)*x_ref[0]+(0.237113978946)*x_ref[1]**o+(-0.116626766689)*x_ref[1]+(-0.670422853592)*x_ref[2]**o+(0.574941151693)*x_ref[2]
        ref[1,0,1]=(-0.0133773984155)*x_ref[0]**o+(0.286273972547)*x_ref[0]+(0.0493985332233)*x_ref[1]**o+(-0.567835671454)*x_ref[1]+(0.508314781519)*x_ref[2]**o+(-0.603506084784)*x_ref[2]
        ref[1,1,0]=(0.0237413792908)*x_ref[0]**o+(-0.758893281773)*x_ref[0]+(0.150670689461)*x_ref[1]**o+(0.136386516616)*x_ref[1]+(-0.440203817756)*x_ref[2]**o+(0.796610908529)*x_ref[2]
        ref[1,1,1]=(-0.223549634917)*x_ref[0]**o+(0.396472946824)*x_ref[0]+(0.47830083818)*x_ref[1]**o+(0.059933592631)*x_ref[1]+(0.709371074379)*x_ref[2]**o+(-0.129442424949)*x_ref[2]
        ref[2,0,0]=(0.686170679191)*x_ref[0]**o+(-0.118668021586)*x_ref[0]+(0.033954449978)*x_ref[1]**o+(-0.0508661210104)*x_ref[1]+(0.352824724076)*x_ref[2]**o+(-0.772874133692)*x_ref[2]
        ref[2,0,1]=(0.434879772789)*x_ref[0]**o+(-0.148309931356)*x_ref[0]+(-0.013792688421)*x_ref[1]**o+(0.597899709109)*x_ref[1]+(-0.280336465565)*x_ref[2]**o+(-0.177134087728)*x_ref[2]
        ref[2,1,0]=(-0.636844238715)*x_ref[0]**o+(-0.646269255495)*x_ref[0]+(0.243236542558)*x_ref[1]**o+(0.203480511801)*x_ref[1]+(0.693290035164)*x_ref[2]**o+(0.773354114662)*x_ref[2]
        ref[2,1,1]=(-0.161998867128)*x_ref[0]**o+(0.193608078775)*x_ref[0]+(0.607954851405)*x_ref[1]**o+(0.648078354644)*x_ref[1]+(-0.488891750936)*x_ref[2]**o+(0.912860597431)*x_ref[2]
        ref[3,0,0]=(-0.696909279028)*x_ref[0]**o+(0.039558795649)*x_ref[0]+(0.441121646773)*x_ref[1]**o+(-0.0681790925087)*x_ref[1]+(-0.792437910441)*x_ref[2]**o+(-0.512942395878)*x_ref[2]
        ref[3,0,1]=(-0.264832490825)*x_ref[0]**o+(-0.955976365474)*x_ref[0]+(-0.884641387524)*x_ref[1]**o+(-0.736970630699)*x_ref[1]+(-0.0525378155791)*x_ref[2]**o+(0.149164361449)*x_ref[2]
        ref[3,1,0]=(0.434888951175)*x_ref[0]**o+(-0.218511432369)*x_ref[0]+(0.564377624401)*x_ref[1]**o+(-0.0978716780106)*x_ref[1]+(0.447857191989)*x_ref[2]**o+(-0.466572586856)*x_ref[2]
        ref[3,1,1]=(0.457689985738)*x_ref[0]**o+(0.00322996620178)*x_ref[0]+(0.307787601144)*x_ref[1]**o+(-0.119455639704)*x_ref[1]+(0.70367723461)*x_ref[2]**o+(0.561913096313)*x_ref[2]
        ref[4,0,0]=(-0.759802700221)*x_ref[0]**o+(0.469649690329)*x_ref[0]+(0.0913972297116)*x_ref[1]**o+(0.310277900593)*x_ref[1]+(-0.590266889319)*x_ref[2]**o+(-0.89711266859)*x_ref[2]
        ref[4,0,1]=(0.85506147544)*x_ref[0]**o+(-0.357476737069)*x_ref[0]+(-0.646898911915)*x_ref[1]**o+(-0.674811608049)*x_ref[1]+(0.865265982147)*x_ref[2]**o+(0.720153126392)*x_ref[2]
        ref[4,1,0]=(0.848146446583)*x_ref[0]**o+(0.868166234685)*x_ref[0]+(0.0477110054774)*x_ref[1]**o+(-0.0338098833736)*x_ref[1]+(-0.360024439914)*x_ref[2]**o+(-0.634359083766)*x_ref[2]
        ref[4,1,1]=(-0.672530624126)*x_ref[0]**o+(0.354170615043)*x_ref[0]+(0.486028226561)*x_ref[1]**o+(0.857532578327)*x_ref[1]+(0.15798171243)*x_ref[2]**o+(0.0579857791862)*x_ref[2]
        ref[5,0,0]=(0.653171799956)*x_ref[0]**o+(0.605427034155)*x_ref[0]+(0.7414424011)*x_ref[1]**o+(-0.807505211605)*x_ref[1]+(-0.0411776868808)*x_ref[2]**o+(0.24301961063)*x_ref[2]
        ref[5,0,1]=(-0.336860102074)*x_ref[0]**o+(0.235697646908)*x_ref[0]+(-0.173217953629)*x_ref[1]**o+(0.989643649714)*x_ref[1]+(0.977460054831)*x_ref[2]**o+(-0.743825027125)*x_ref[2]
        ref[5,1,0]=(0.794904875581)*x_ref[0]**o+(0.607548574046)*x_ref[0]+(0.603180304387)*x_ref[1]**o+(0.276995429843)*x_ref[1]+(0.698731315308)*x_ref[2]**o+(-0.966744108727)*x_ref[2]
        ref[5,1,1]=(-0.0230004766002)*x_ref[0]**o+(0.489585418104)*x_ref[0]+(0.382764947116)*x_ref[1]**o+(0.143873532935)*x_ref[1]+(0.932000957196)*x_ref[2]**o+(0.739975795806)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.922177773381)*x[0]**o+(0.37324175781)*x[0]+(-0.169208776011)*x[1]**o+(0.72134249236)*x[1]
        arg[0,0,0,1]=(0.85312270129)*x[0]**o+(0.0512919651836)*x[0]+(0.968880249675)*x[1]**o+(0.740058819096)*x[1]
        arg[0,0,1,0]=(-0.018205657901)*x[0]**o+(0.601838710779)*x[0]+(0.211002282731)*x[1]**o+(0.801382123792)*x[1]
        arg[0,0,1,1]=(-0.249128714489)*x[0]**o+(-0.408809542089)*x[0]+(-0.684434920261)*x[1]**o+(0.223674849818)*x[1]
        arg[0,0,2,0]=(0.396627551183)*x[0]**o+(-0.223592778052)*x[0]+(0.227333834259)*x[1]**o+(0.66580805616)*x[1]
        arg[0,0,2,1]=(0.365216692731)*x[0]**o+(-0.670947594503)*x[0]+(0.400260365509)*x[1]**o+(0.10903299253)*x[1]
        arg[0,1,0,0]=(0.823297765985)*x[0]**o+(0.60337946486)*x[0]+(-0.314265083693)*x[1]**o+(-0.00636816227929)*x[1]
        arg[0,1,0,1]=(-0.26216625982)*x[0]**o+(0.767412606588)*x[0]+(-0.284917238441)*x[1]**o+(0.56385025143)*x[1]
        arg[0,1,1,0]=(-0.732326448281)*x[0]**o+(-0.669197323856)*x[0]+(-0.608751752289)*x[1]**o+(-0.853458985188)*x[1]
        arg[0,1,1,1]=(0.637275594334)*x[0]**o+(0.696570850488)*x[0]+(0.012026820657)*x[1]**o+(0.558317439186)*x[1]
        arg[0,1,2,0]=(0.925899092402)*x[0]**o+(0.568045453255)*x[0]+(-0.766701931715)*x[1]**o+(-0.95082586036)*x[1]
        arg[0,1,2,1]=(0.127524211574)*x[0]**o+(-0.65150542164)*x[0]+(-0.365773940068)*x[1]**o+(0.990465711754)*x[1]
        arg[0,2,0,0]=(0.456469052217)*x[0]**o+(0.403391798397)*x[0]+(-0.21921915949)*x[1]**o+(0.171493695945)*x[1]
        arg[0,2,0,1]=(-0.980242506743)*x[0]**o+(0.880225959645)*x[0]+(0.682653186307)*x[1]**o+(-0.561867916377)*x[1]
        arg[0,2,1,0]=(-0.329881998938)*x[0]**o+(0.464976480407)*x[0]+(-0.812583693347)*x[1]**o+(0.0103356333568)*x[1]
        arg[0,2,1,1]=(-0.710350374606)*x[0]**o+(-0.445981415127)*x[0]+(-0.159054554695)*x[1]**o+(-0.0306514318922)*x[1]
        arg[0,2,2,0]=(0.0903142465124)*x[0]**o+(0.872680271888)*x[0]+(0.230242338668)*x[1]**o+(0.557126778217)*x[1]
        arg[0,2,2,1]=(-0.842046475241)*x[0]**o+(0.350544067128)*x[0]+(0.486186516633)*x[1]**o+(-0.550801066658)*x[1]
        arg[0,3,0,0]=(0.740313870334)*x[0]**o+(-0.177105021967)*x[0]+(0.00114640526837)*x[1]**o+(0.349602955151)*x[1]
        arg[0,3,0,1]=(0.898906724506)*x[0]**o+(-0.882729119203)*x[0]+(0.000278645863091)*x[1]**o+(-0.450247260284)*x[1]
        arg[0,3,1,0]=(0.411917691981)*x[0]**o+(0.0461900014479)*x[0]+(0.1608400364)*x[1]**o+(0.130550626888)*x[1]
        arg[0,3,1,1]=(0.610378099059)*x[0]**o+(-0.903971943528)*x[0]+(-0.0170960427518)*x[1]**o+(0.765729630725)*x[1]
        arg[0,3,2,0]=(-0.811421756909)*x[0]**o+(-0.220065414196)*x[0]+(0.529467921065)*x[1]**o+(0.13896086667)*x[1]
        arg[0,3,2,1]=(-0.621021589728)*x[0]**o+(0.181903892108)*x[0]+(0.724296638902)*x[1]**o+(0.651705107858)*x[1]
        arg[0,4,0,0]=(0.323254788695)*x[0]**o+(0.758728125147)*x[0]+(0.0719056895797)*x[1]**o+(-0.486299017994)*x[1]
        arg[0,4,0,1]=(-0.426525215977)*x[0]**o+(-0.70344592953)*x[0]+(-0.467799544116)*x[1]**o+(0.0394017649712)*x[1]
        arg[0,4,1,0]=(-0.559984242808)*x[0]**o+(-0.963639424043)*x[0]+(0.884232194806)*x[1]**o+(0.948111505594)*x[1]
        arg[0,4,1,1]=(-0.825627656302)*x[0]**o+(0.167871540452)*x[0]+(0.0969950885663)*x[1]**o+(-0.477024385485)*x[1]
        arg[0,4,2,0]=(-0.890191205005)*x[0]**o+(-0.291891624031)*x[0]+(-0.924812934321)*x[1]**o+(-0.898567598264)*x[1]
        arg[0,4,2,1]=(0.0552241756422)*x[0]**o+(-0.47878688784)*x[0]+(-0.208893161197)*x[1]**o+(-0.671733441515)*x[1]
        arg[1,0,0,0]=(0.609961055126)*x[0]**o+(0.389225845794)*x[0]+(0.147977819654)*x[1]**o+(-0.75096024248)*x[1]
        arg[1,0,0,1]=(-0.407376527411)*x[0]**o+(-0.696660373977)*x[0]+(-0.803862178008)*x[1]**o+(-0.367087112506)*x[1]
        arg[1,0,1,0]=(0.89146134332)*x[0]**o+(0.388607608477)*x[0]+(0.763190367241)*x[1]**o+(0.169275499463)*x[1]
        arg[1,0,1,1]=(-0.870871474063)*x[0]**o+(-0.21526798243)*x[0]+(0.159168404985)*x[1]**o+(0.715049896538)*x[1]
        arg[1,0,2,0]=(-0.842978554795)*x[0]**o+(0.209649486001)*x[0]+(-0.767629750054)*x[1]**o+(-0.126634026874)*x[1]
        arg[1,0,2,1]=(-0.728571756044)*x[0]**o+(-0.247852313042)*x[0]+(0.818809803616)*x[1]**o+(0.736163163331)*x[1]
        arg[1,1,0,0]=(-0.670608139072)*x[0]**o+(0.330167259903)*x[0]+(0.275974484086)*x[1]**o+(0.492110644529)*x[1]
        arg[1,1,0,1]=(0.724664287226)*x[0]**o+(0.738855592065)*x[0]+(0.511369798951)*x[1]**o+(0.455794112842)*x[1]
        arg[1,1,1,0]=(0.783504599076)*x[0]**o+(-0.558579246854)*x[0]+(0.608044912637)*x[1]**o+(-0.401058633268)*x[1]
        arg[1,1,1,1]=(0.406697861117)*x[0]**o+(-0.31302718156)*x[0]+(-0.999553049864)*x[1]**o+(0.647698243869)*x[1]
        arg[1,1,2,0]=(-0.788346252365)*x[0]**o+(-0.809895265253)*x[0]+(0.441738551485)*x[1]**o+(-0.686785865679)*x[1]
        arg[1,1,2,1]=(0.794571346519)*x[0]**o+(-0.552071697712)*x[0]+(0.77260218173)*x[1]**o+(-0.969584549528)*x[1]
        arg[1,2,0,0]=(-0.867948229325)*x[0]**o+(-0.582387320474)*x[0]+(0.0387726756335)*x[1]**o+(0.4143770831)*x[1]
        arg[1,2,0,1]=(-0.913557875303)*x[0]**o+(-0.517970834439)*x[0]+(-0.922517499063)*x[1]**o+(0.858552054359)*x[1]
        arg[1,2,1,0]=(0.409302016181)*x[0]**o+(-0.684529177441)*x[0]+(0.818122023218)*x[1]**o+(-0.583589419623)*x[1]
        arg[1,2,1,1]=(0.750795883056)*x[0]**o+(-0.710882726624)*x[0]+(0.915210035315)*x[1]**o+(0.802588664027)*x[1]
        arg[1,2,2,0]=(0.639216566804)*x[0]**o+(-0.300954214125)*x[0]+(-0.26724393342)*x[1]**o+(0.426742851293)*x[1]
        arg[1,2,2,1]=(-0.00337538020121)*x[0]**o+(0.283012169215)*x[0]+(0.725040731001)*x[1]**o+(-0.708238265029)*x[1]
        arg[1,3,0,0]=(0.217404922432)*x[0]**o+(-0.590883003502)*x[0]+(0.451142849723)*x[1]**o+(0.624698767392)*x[1]
        arg[1,3,0,1]=(0.0351258739828)*x[0]**o+(-0.784068492686)*x[0]+(0.469956184167)*x[1]**o+(0.634124784387)*x[1]
        arg[1,3,1,0]=(0.998414775431)*x[0]**o+(0.388202946918)*x[0]+(-0.849223443255)*x[1]**o+(-0.804590028218)*x[1]
        arg[1,3,1,1]=(0.0219748851856)*x[0]**o+(0.464069403336)*x[0]+(-0.470759541095)*x[1]**o+(-0.21773573235)*x[1]
        arg[1,3,2,0]=(-0.879175905661)*x[0]**o+(0.100238280861)*x[0]+(0.331472085056)*x[1]**o+(0.0310947651268)*x[1]
        arg[1,3,2,1]=(-0.195155740038)*x[0]**o+(0.532078676231)*x[0]+(-0.863288082011)*x[1]**o+(-0.0905878788265)*x[1]
        arg[1,4,0,0]=(0.470883637487)*x[0]**o+(0.0291028245963)*x[0]+(-0.152462235474)*x[1]**o+(0.890883963301)*x[1]
        arg[1,4,0,1]=(-0.0780797534409)*x[0]**o+(-0.999649003413)*x[0]+(-0.79446765004)*x[1]**o+(-0.820520881428)*x[1]
        arg[1,4,1,0]=(-0.82479629638)*x[0]**o+(-0.28924829922)*x[0]+(-0.960430867949)*x[1]**o+(-0.514730558673)*x[1]
        arg[1,4,1,1]=(-0.0853911110391)*x[0]**o+(0.726747270178)*x[0]+(-0.70925711586)*x[1]**o+(-0.331433386035)*x[1]
        arg[1,4,2,0]=(-0.434350343402)*x[0]**o+(0.335341284415)*x[0]+(0.496818240476)*x[1]**o+(-0.925006191184)*x[1]
        arg[1,4,2,1]=(0.0753516047957)*x[0]**o+(0.171090641529)*x[0]+(-0.836002762987)*x[1]**o+(-0.753836171615)*x[1]
        arg[2,0,0,0]=(-0.331008073553)*x[0]**o+(-0.299051880715)*x[0]+(-0.84520124413)*x[1]**o+(0.780471473975)*x[1]
        arg[2,0,0,1]=(-0.947973851318)*x[0]**o+(0.579142212501)*x[0]+(0.137317760268)*x[1]**o+(-0.379509530229)*x[1]
        arg[2,0,1,0]=(-0.741320730287)*x[0]**o+(0.642110710824)*x[0]+(-0.632595944259)*x[1]**o+(0.944980046528)*x[1]
        arg[2,0,1,1]=(0.616943682269)*x[0]**o+(-0.348189507611)*x[0]+(-0.414080070092)*x[1]**o+(-0.733041081362)*x[1]
        arg[2,0,2,0]=(-0.021261659709)*x[0]**o+(0.790041129031)*x[0]+(0.788378634137)*x[1]**o+(0.391057560074)*x[1]
        arg[2,0,2,1]=(0.563737321295)*x[0]**o+(-0.110516278477)*x[0]+(0.383563652425)*x[1]**o+(-0.857100387101)*x[1]
        arg[2,1,0,0]=(-0.0839346911978)*x[0]**o+(0.699255867073)*x[0]+(-0.698488781835)*x[1]**o+(0.225940657322)*x[1]
        arg[2,1,0,1]=(0.93638601214)*x[0]**o+(-0.593516691147)*x[0]+(-0.445863239521)*x[1]**o+(0.601067357746)*x[1]
        arg[2,1,1,0]=(0.114814804735)*x[0]**o+(0.717600499991)*x[0]+(0.7371911818)*x[1]**o+(0.136012416552)*x[1]
        arg[2,1,1,1]=(0.794499830148)*x[0]**o+(0.0581947415855)*x[0]+(0.626382859976)*x[1]**o+(-0.668272098674)*x[1]
        arg[2,1,2,0]=(0.564628771073)*x[0]**o+(0.102820241267)*x[0]+(0.397960234363)*x[1]**o+(0.379632827896)*x[1]
        arg[2,1,2,1]=(-0.0880639775913)*x[0]**o+(0.325978376801)*x[0]+(-0.505498229733)*x[1]**o+(-0.595922646655)*x[1]
        arg[2,2,0,0]=(-0.132672737928)*x[0]**o+(-0.0291240195523)*x[0]+(-0.00825944211706)*x[1]**o+(0.985993963145)*x[1]
        arg[2,2,0,1]=(0.244321497515)*x[0]**o+(-0.111329005236)*x[0]+(0.769353471123)*x[1]**o+(-0.174257027136)*x[1]
        arg[2,2,1,0]=(0.252286330282)*x[0]**o+(0.450975615825)*x[0]+(-0.515016578606)*x[1]**o+(0.834736757555)*x[1]
        arg[2,2,1,1]=(0.0588273338962)*x[0]**o+(-0.865406338788)*x[0]+(0.482095645149)*x[1]**o+(-0.595126490156)*x[1]
        arg[2,2,2,0]=(0.244580063555)*x[0]**o+(0.576709884892)*x[0]+(-0.212616562122)*x[1]**o+(-0.407508050723)*x[1]
        arg[2,2,2,1]=(0.416825783755)*x[0]**o+(0.430959078988)*x[0]+(-0.398840301481)*x[1]**o+(0.889105964295)*x[1]
        arg[2,3,0,0]=(-0.731872136283)*x[0]**o+(0.237034931745)*x[0]+(0.547737117698)*x[1]**o+(0.364733861661)*x[1]
        arg[2,3,0,1]=(0.971716549689)*x[0]**o+(0.466300970081)*x[0]+(-0.51802162747)*x[1]**o+(-0.859910778018)*x[1]
        arg[2,3,1,0]=(-0.562918219893)*x[0]**o+(-0.0532549959391)*x[0]+(0.842033495359)*x[1]**o+(-0.708554283815)*x[1]
        arg[2,3,1,1]=(0.437125534104)*x[0]**o+(-0.376751615116)*x[0]+(-0.330889383623)*x[1]**o+(0.958884035289)*x[1]
        arg[2,3,2,0]=(-0.194290501566)*x[0]**o+(-0.368131904292)*x[0]+(0.802917650409)*x[1]**o+(0.628603987906)*x[1]
        arg[2,3,2,1]=(0.70219875881)*x[0]**o+(0.822181339005)*x[0]+(0.768635447234)*x[1]**o+(0.871676010439)*x[1]
        arg[2,4,0,0]=(-0.654498700662)*x[0]**o+(0.851143676147)*x[0]+(0.468468395415)*x[1]**o+(-0.447240713576)*x[1]
        arg[2,4,0,1]=(0.893177052112)*x[0]**o+(0.652743480988)*x[0]+(0.726646163682)*x[1]**o+(-0.854895731892)*x[1]
        arg[2,4,1,0]=(0.0540542301334)*x[0]**o+(-0.208654443984)*x[0]+(0.925127576853)*x[1]**o+(-0.932489814868)*x[1]
        arg[2,4,1,1]=(-0.460268214104)*x[0]**o+(-0.303569308977)*x[0]+(-0.381458880316)*x[1]**o+(0.690537661673)*x[1]
        arg[2,4,2,0]=(0.401012366366)*x[0]**o+(-0.970993441679)*x[0]+(0.0249535421254)*x[1]**o+(-0.499737254014)*x[1]
        arg[2,4,2,1]=(-0.692604790859)*x[0]**o+(-0.685292098725)*x[0]+(0.22550357354)*x[1]**o+(0.908152783462)*x[1]
        arg[3,0,0,0]=(-0.634246407254)*x[0]**o+(-0.571223033336)*x[0]+(0.0574319920181)*x[1]**o+(0.123763929908)*x[1]
        arg[3,0,0,1]=(-0.569358219213)*x[0]**o+(-0.107237509632)*x[0]+(0.97672999497)*x[1]**o+(-0.114939366813)*x[1]
        arg[3,0,1,0]=(-0.767885152479)*x[0]**o+(-0.105924102555)*x[0]+(0.539361607398)*x[1]**o+(0.546186446469)*x[1]
        arg[3,0,1,1]=(-0.58818696897)*x[0]**o+(0.629371516566)*x[0]+(-0.598173904376)*x[1]**o+(0.959043112088)*x[1]
        arg[3,0,2,0]=(0.751165393927)*x[0]**o+(0.732151619397)*x[0]+(-0.435276735735)*x[1]**o+(0.947142140722)*x[1]
        arg[3,0,2,1]=(0.463904349547)*x[0]**o+(-0.947270841014)*x[0]+(0.304239969784)*x[1]**o+(-0.233566944292)*x[1]
        arg[3,1,0,0]=(0.961122232395)*x[0]**o+(-0.989040215156)*x[0]+(0.833186746232)*x[1]**o+(0.226673105305)*x[1]
        arg[3,1,0,1]=(-0.156957616578)*x[0]**o+(-0.361665587314)*x[0]+(0.987490273947)*x[1]**o+(-0.890271058746)*x[1]
        arg[3,1,1,0]=(-0.259959222953)*x[0]**o+(-0.696135598749)*x[0]+(0.234411151165)*x[1]**o+(0.783105897256)*x[1]
        arg[3,1,1,1]=(0.020886322199)*x[0]**o+(0.887445555864)*x[0]+(0.709489180254)*x[1]**o+(0.990590112507)*x[1]
        arg[3,1,2,0]=(-0.963790247866)*x[0]**o+(-0.417180971158)*x[0]+(0.303859261646)*x[1]**o+(0.745552024399)*x[1]
        arg[3,1,2,1]=(-0.364448135641)*x[0]**o+(0.760841571414)*x[0]+(-0.0283374186763)*x[1]**o+(-0.491433855748)*x[1]
        arg[3,2,0,0]=(-0.649766142291)*x[0]**o+(-0.177410238297)*x[0]+(0.137316641289)*x[1]**o+(-0.40890299152)*x[1]
        arg[3,2,0,1]=(0.282922984657)*x[0]**o+(-0.660238152943)*x[0]+(0.481726052488)*x[1]**o+(0.976011254218)*x[1]
        arg[3,2,1,0]=(0.989761520473)*x[0]**o+(0.194082255647)*x[0]+(-0.648435682551)*x[1]**o+(-0.721569599311)*x[1]
        arg[3,2,1,1]=(0.734634540883)*x[0]**o+(0.524638529049)*x[0]+(-0.737636791717)*x[1]**o+(-0.995904525822)*x[1]
        arg[3,2,2,0]=(0.422396808663)*x[0]**o+(-0.820639154105)*x[0]+(-0.566207315373)*x[1]**o+(-0.224174711984)*x[1]
        arg[3,2,2,1]=(0.719145374997)*x[0]**o+(-0.885860499468)*x[0]+(0.184062878609)*x[1]**o+(0.47660543007)*x[1]
        arg[3,3,0,0]=(-0.0122290849915)*x[0]**o+(-0.0314298339876)*x[0]+(0.937255407134)*x[1]**o+(-0.754858547072)*x[1]
        arg[3,3,0,1]=(-0.507826956669)*x[0]**o+(0.951205020689)*x[0]+(0.837096245109)*x[1]**o+(-0.576692399518)*x[1]
        arg[3,3,1,0]=(0.602119808238)*x[0]**o+(-0.943937134298)*x[0]+(-0.244250254626)*x[1]**o+(-0.814103542727)*x[1]
        arg[3,3,1,1]=(-0.51811661822)*x[0]**o+(-0.188969078762)*x[0]+(-0.697470816837)*x[1]**o+(-0.741593014841)*x[1]
        arg[3,3,2,0]=(-0.673948933786)*x[0]**o+(0.816404336566)*x[0]+(0.339239090092)*x[1]**o+(-0.0173550947113)*x[1]
        arg[3,3,2,1]=(0.499109852716)*x[0]**o+(0.515501552893)*x[0]+(0.926442939974)*x[1]**o+(0.157227173254)*x[1]
        arg[3,4,0,0]=(-0.1837389273)*x[0]**o+(0.78490229721)*x[0]+(-0.248033788374)*x[1]**o+(0.735136386405)*x[1]
        arg[3,4,0,1]=(0.121169550194)*x[0]**o+(-0.854027732077)*x[0]+(0.937680848475)*x[1]**o+(-0.447184197264)*x[1]
        arg[3,4,1,0]=(-0.788610934694)*x[0]**o+(-0.397825388144)*x[0]+(0.604421361449)*x[1]**o+(0.937184552821)*x[1]
        arg[3,4,1,1]=(-0.493380143496)*x[0]**o+(0.958197000398)*x[0]+(0.282372561751)*x[1]**o+(-0.55322669826)*x[1]
        arg[3,4,2,0]=(0.11333865923)*x[0]**o+(-0.542805511569)*x[0]+(-0.640646701928)*x[1]**o+(-0.342030716597)*x[1]
        arg[3,4,2,1]=(0.505356266297)*x[0]**o+(-0.107422782294)*x[0]+(0.542833447968)*x[1]**o+(-0.615275716592)*x[1]
        ref[0,0,0,0]=(0.922177773381)*x_ref[0]**o+(0.37324175781)*x_ref[0]+(-0.169208776011)*x_ref[1]**o+(0.72134249236)*x_ref[1]
        ref[0,0,0,1]=(0.85312270129)*x_ref[0]**o+(0.0512919651836)*x_ref[0]+(0.968880249675)*x_ref[1]**o+(0.740058819096)*x_ref[1]
        ref[0,0,1,0]=(-0.018205657901)*x_ref[0]**o+(0.601838710779)*x_ref[0]+(0.211002282731)*x_ref[1]**o+(0.801382123792)*x_ref[1]
        ref[0,0,1,1]=(-0.249128714489)*x_ref[0]**o+(-0.408809542089)*x_ref[0]+(-0.684434920261)*x_ref[1]**o+(0.223674849818)*x_ref[1]
        ref[0,0,2,0]=(0.396627551183)*x_ref[0]**o+(-0.223592778052)*x_ref[0]+(0.227333834259)*x_ref[1]**o+(0.66580805616)*x_ref[1]
        ref[0,0,2,1]=(0.365216692731)*x_ref[0]**o+(-0.670947594503)*x_ref[0]+(0.400260365509)*x_ref[1]**o+(0.10903299253)*x_ref[1]
        ref[0,1,0,0]=(0.823297765985)*x_ref[0]**o+(0.60337946486)*x_ref[0]+(-0.314265083693)*x_ref[1]**o+(-0.00636816227929)*x_ref[1]
        ref[0,1,0,1]=(-0.26216625982)*x_ref[0]**o+(0.767412606588)*x_ref[0]+(-0.284917238441)*x_ref[1]**o+(0.56385025143)*x_ref[1]
        ref[0,1,1,0]=(-0.732326448281)*x_ref[0]**o+(-0.669197323856)*x_ref[0]+(-0.608751752289)*x_ref[1]**o+(-0.853458985188)*x_ref[1]
        ref[0,1,1,1]=(0.637275594334)*x_ref[0]**o+(0.696570850488)*x_ref[0]+(0.012026820657)*x_ref[1]**o+(0.558317439186)*x_ref[1]
        ref[0,1,2,0]=(0.925899092402)*x_ref[0]**o+(0.568045453255)*x_ref[0]+(-0.766701931715)*x_ref[1]**o+(-0.95082586036)*x_ref[1]
        ref[0,1,2,1]=(0.127524211574)*x_ref[0]**o+(-0.65150542164)*x_ref[0]+(-0.365773940068)*x_ref[1]**o+(0.990465711754)*x_ref[1]
        ref[0,2,0,0]=(0.456469052217)*x_ref[0]**o+(0.403391798397)*x_ref[0]+(-0.21921915949)*x_ref[1]**o+(0.171493695945)*x_ref[1]
        ref[0,2,0,1]=(-0.980242506743)*x_ref[0]**o+(0.880225959645)*x_ref[0]+(0.682653186307)*x_ref[1]**o+(-0.561867916377)*x_ref[1]
        ref[0,2,1,0]=(-0.329881998938)*x_ref[0]**o+(0.464976480407)*x_ref[0]+(-0.812583693347)*x_ref[1]**o+(0.0103356333568)*x_ref[1]
        ref[0,2,1,1]=(-0.710350374606)*x_ref[0]**o+(-0.445981415127)*x_ref[0]+(-0.159054554695)*x_ref[1]**o+(-0.0306514318922)*x_ref[1]
        ref[0,2,2,0]=(0.0903142465124)*x_ref[0]**o+(0.872680271888)*x_ref[0]+(0.230242338668)*x_ref[1]**o+(0.557126778217)*x_ref[1]
        ref[0,2,2,1]=(-0.842046475241)*x_ref[0]**o+(0.350544067128)*x_ref[0]+(0.486186516633)*x_ref[1]**o+(-0.550801066658)*x_ref[1]
        ref[0,3,0,0]=(0.740313870334)*x_ref[0]**o+(-0.177105021967)*x_ref[0]+(0.00114640526837)*x_ref[1]**o+(0.349602955151)*x_ref[1]
        ref[0,3,0,1]=(0.898906724506)*x_ref[0]**o+(-0.882729119203)*x_ref[0]+(0.000278645863091)*x_ref[1]**o+(-0.450247260284)*x_ref[1]
        ref[0,3,1,0]=(0.411917691981)*x_ref[0]**o+(0.0461900014479)*x_ref[0]+(0.1608400364)*x_ref[1]**o+(0.130550626888)*x_ref[1]
        ref[0,3,1,1]=(0.610378099059)*x_ref[0]**o+(-0.903971943528)*x_ref[0]+(-0.0170960427518)*x_ref[1]**o+(0.765729630725)*x_ref[1]
        ref[0,3,2,0]=(-0.811421756909)*x_ref[0]**o+(-0.220065414196)*x_ref[0]+(0.529467921065)*x_ref[1]**o+(0.13896086667)*x_ref[1]
        ref[0,3,2,1]=(-0.621021589728)*x_ref[0]**o+(0.181903892108)*x_ref[0]+(0.724296638902)*x_ref[1]**o+(0.651705107858)*x_ref[1]
        ref[0,4,0,0]=(0.323254788695)*x_ref[0]**o+(0.758728125147)*x_ref[0]+(0.0719056895797)*x_ref[1]**o+(-0.486299017994)*x_ref[1]
        ref[0,4,0,1]=(-0.426525215977)*x_ref[0]**o+(-0.70344592953)*x_ref[0]+(-0.467799544116)*x_ref[1]**o+(0.0394017649712)*x_ref[1]
        ref[0,4,1,0]=(-0.559984242808)*x_ref[0]**o+(-0.963639424043)*x_ref[0]+(0.884232194806)*x_ref[1]**o+(0.948111505594)*x_ref[1]
        ref[0,4,1,1]=(-0.825627656302)*x_ref[0]**o+(0.167871540452)*x_ref[0]+(0.0969950885663)*x_ref[1]**o+(-0.477024385485)*x_ref[1]
        ref[0,4,2,0]=(-0.890191205005)*x_ref[0]**o+(-0.291891624031)*x_ref[0]+(-0.924812934321)*x_ref[1]**o+(-0.898567598264)*x_ref[1]
        ref[0,4,2,1]=(0.0552241756422)*x_ref[0]**o+(-0.47878688784)*x_ref[0]+(-0.208893161197)*x_ref[1]**o+(-0.671733441515)*x_ref[1]
        ref[1,0,0,0]=(0.609961055126)*x_ref[0]**o+(0.389225845794)*x_ref[0]+(0.147977819654)*x_ref[1]**o+(-0.75096024248)*x_ref[1]
        ref[1,0,0,1]=(-0.407376527411)*x_ref[0]**o+(-0.696660373977)*x_ref[0]+(-0.803862178008)*x_ref[1]**o+(-0.367087112506)*x_ref[1]
        ref[1,0,1,0]=(0.89146134332)*x_ref[0]**o+(0.388607608477)*x_ref[0]+(0.763190367241)*x_ref[1]**o+(0.169275499463)*x_ref[1]
        ref[1,0,1,1]=(-0.870871474063)*x_ref[0]**o+(-0.21526798243)*x_ref[0]+(0.159168404985)*x_ref[1]**o+(0.715049896538)*x_ref[1]
        ref[1,0,2,0]=(-0.842978554795)*x_ref[0]**o+(0.209649486001)*x_ref[0]+(-0.767629750054)*x_ref[1]**o+(-0.126634026874)*x_ref[1]
        ref[1,0,2,1]=(-0.728571756044)*x_ref[0]**o+(-0.247852313042)*x_ref[0]+(0.818809803616)*x_ref[1]**o+(0.736163163331)*x_ref[1]
        ref[1,1,0,0]=(-0.670608139072)*x_ref[0]**o+(0.330167259903)*x_ref[0]+(0.275974484086)*x_ref[1]**o+(0.492110644529)*x_ref[1]
        ref[1,1,0,1]=(0.724664287226)*x_ref[0]**o+(0.738855592065)*x_ref[0]+(0.511369798951)*x_ref[1]**o+(0.455794112842)*x_ref[1]
        ref[1,1,1,0]=(0.783504599076)*x_ref[0]**o+(-0.558579246854)*x_ref[0]+(0.608044912637)*x_ref[1]**o+(-0.401058633268)*x_ref[1]
        ref[1,1,1,1]=(0.406697861117)*x_ref[0]**o+(-0.31302718156)*x_ref[0]+(-0.999553049864)*x_ref[1]**o+(0.647698243869)*x_ref[1]
        ref[1,1,2,0]=(-0.788346252365)*x_ref[0]**o+(-0.809895265253)*x_ref[0]+(0.441738551485)*x_ref[1]**o+(-0.686785865679)*x_ref[1]
        ref[1,1,2,1]=(0.794571346519)*x_ref[0]**o+(-0.552071697712)*x_ref[0]+(0.77260218173)*x_ref[1]**o+(-0.969584549528)*x_ref[1]
        ref[1,2,0,0]=(-0.867948229325)*x_ref[0]**o+(-0.582387320474)*x_ref[0]+(0.0387726756335)*x_ref[1]**o+(0.4143770831)*x_ref[1]
        ref[1,2,0,1]=(-0.913557875303)*x_ref[0]**o+(-0.517970834439)*x_ref[0]+(-0.922517499063)*x_ref[1]**o+(0.858552054359)*x_ref[1]
        ref[1,2,1,0]=(0.409302016181)*x_ref[0]**o+(-0.684529177441)*x_ref[0]+(0.818122023218)*x_ref[1]**o+(-0.583589419623)*x_ref[1]
        ref[1,2,1,1]=(0.750795883056)*x_ref[0]**o+(-0.710882726624)*x_ref[0]+(0.915210035315)*x_ref[1]**o+(0.802588664027)*x_ref[1]
        ref[1,2,2,0]=(0.639216566804)*x_ref[0]**o+(-0.300954214125)*x_ref[0]+(-0.26724393342)*x_ref[1]**o+(0.426742851293)*x_ref[1]
        ref[1,2,2,1]=(-0.00337538020121)*x_ref[0]**o+(0.283012169215)*x_ref[0]+(0.725040731001)*x_ref[1]**o+(-0.708238265029)*x_ref[1]
        ref[1,3,0,0]=(0.217404922432)*x_ref[0]**o+(-0.590883003502)*x_ref[0]+(0.451142849723)*x_ref[1]**o+(0.624698767392)*x_ref[1]
        ref[1,3,0,1]=(0.0351258739828)*x_ref[0]**o+(-0.784068492686)*x_ref[0]+(0.469956184167)*x_ref[1]**o+(0.634124784387)*x_ref[1]
        ref[1,3,1,0]=(0.998414775431)*x_ref[0]**o+(0.388202946918)*x_ref[0]+(-0.849223443255)*x_ref[1]**o+(-0.804590028218)*x_ref[1]
        ref[1,3,1,1]=(0.0219748851856)*x_ref[0]**o+(0.464069403336)*x_ref[0]+(-0.470759541095)*x_ref[1]**o+(-0.21773573235)*x_ref[1]
        ref[1,3,2,0]=(-0.879175905661)*x_ref[0]**o+(0.100238280861)*x_ref[0]+(0.331472085056)*x_ref[1]**o+(0.0310947651268)*x_ref[1]
        ref[1,3,2,1]=(-0.195155740038)*x_ref[0]**o+(0.532078676231)*x_ref[0]+(-0.863288082011)*x_ref[1]**o+(-0.0905878788265)*x_ref[1]
        ref[1,4,0,0]=(0.470883637487)*x_ref[0]**o+(0.0291028245963)*x_ref[0]+(-0.152462235474)*x_ref[1]**o+(0.890883963301)*x_ref[1]
        ref[1,4,0,1]=(-0.0780797534409)*x_ref[0]**o+(-0.999649003413)*x_ref[0]+(-0.79446765004)*x_ref[1]**o+(-0.820520881428)*x_ref[1]
        ref[1,4,1,0]=(-0.82479629638)*x_ref[0]**o+(-0.28924829922)*x_ref[0]+(-0.960430867949)*x_ref[1]**o+(-0.514730558673)*x_ref[1]
        ref[1,4,1,1]=(-0.0853911110391)*x_ref[0]**o+(0.726747270178)*x_ref[0]+(-0.70925711586)*x_ref[1]**o+(-0.331433386035)*x_ref[1]
        ref[1,4,2,0]=(-0.434350343402)*x_ref[0]**o+(0.335341284415)*x_ref[0]+(0.496818240476)*x_ref[1]**o+(-0.925006191184)*x_ref[1]
        ref[1,4,2,1]=(0.0753516047957)*x_ref[0]**o+(0.171090641529)*x_ref[0]+(-0.836002762987)*x_ref[1]**o+(-0.753836171615)*x_ref[1]
        ref[2,0,0,0]=(-0.331008073553)*x_ref[0]**o+(-0.299051880715)*x_ref[0]+(-0.84520124413)*x_ref[1]**o+(0.780471473975)*x_ref[1]
        ref[2,0,0,1]=(-0.947973851318)*x_ref[0]**o+(0.579142212501)*x_ref[0]+(0.137317760268)*x_ref[1]**o+(-0.379509530229)*x_ref[1]
        ref[2,0,1,0]=(-0.741320730287)*x_ref[0]**o+(0.642110710824)*x_ref[0]+(-0.632595944259)*x_ref[1]**o+(0.944980046528)*x_ref[1]
        ref[2,0,1,1]=(0.616943682269)*x_ref[0]**o+(-0.348189507611)*x_ref[0]+(-0.414080070092)*x_ref[1]**o+(-0.733041081362)*x_ref[1]
        ref[2,0,2,0]=(-0.021261659709)*x_ref[0]**o+(0.790041129031)*x_ref[0]+(0.788378634137)*x_ref[1]**o+(0.391057560074)*x_ref[1]
        ref[2,0,2,1]=(0.563737321295)*x_ref[0]**o+(-0.110516278477)*x_ref[0]+(0.383563652425)*x_ref[1]**o+(-0.857100387101)*x_ref[1]
        ref[2,1,0,0]=(-0.0839346911978)*x_ref[0]**o+(0.699255867073)*x_ref[0]+(-0.698488781835)*x_ref[1]**o+(0.225940657322)*x_ref[1]
        ref[2,1,0,1]=(0.93638601214)*x_ref[0]**o+(-0.593516691147)*x_ref[0]+(-0.445863239521)*x_ref[1]**o+(0.601067357746)*x_ref[1]
        ref[2,1,1,0]=(0.114814804735)*x_ref[0]**o+(0.717600499991)*x_ref[0]+(0.7371911818)*x_ref[1]**o+(0.136012416552)*x_ref[1]
        ref[2,1,1,1]=(0.794499830148)*x_ref[0]**o+(0.0581947415855)*x_ref[0]+(0.626382859976)*x_ref[1]**o+(-0.668272098674)*x_ref[1]
        ref[2,1,2,0]=(0.564628771073)*x_ref[0]**o+(0.102820241267)*x_ref[0]+(0.397960234363)*x_ref[1]**o+(0.379632827896)*x_ref[1]
        ref[2,1,2,1]=(-0.0880639775913)*x_ref[0]**o+(0.325978376801)*x_ref[0]+(-0.505498229733)*x_ref[1]**o+(-0.595922646655)*x_ref[1]
        ref[2,2,0,0]=(-0.132672737928)*x_ref[0]**o+(-0.0291240195523)*x_ref[0]+(-0.00825944211706)*x_ref[1]**o+(0.985993963145)*x_ref[1]
        ref[2,2,0,1]=(0.244321497515)*x_ref[0]**o+(-0.111329005236)*x_ref[0]+(0.769353471123)*x_ref[1]**o+(-0.174257027136)*x_ref[1]
        ref[2,2,1,0]=(0.252286330282)*x_ref[0]**o+(0.450975615825)*x_ref[0]+(-0.515016578606)*x_ref[1]**o+(0.834736757555)*x_ref[1]
        ref[2,2,1,1]=(0.0588273338962)*x_ref[0]**o+(-0.865406338788)*x_ref[0]+(0.482095645149)*x_ref[1]**o+(-0.595126490156)*x_ref[1]
        ref[2,2,2,0]=(0.244580063555)*x_ref[0]**o+(0.576709884892)*x_ref[0]+(-0.212616562122)*x_ref[1]**o+(-0.407508050723)*x_ref[1]
        ref[2,2,2,1]=(0.416825783755)*x_ref[0]**o+(0.430959078988)*x_ref[0]+(-0.398840301481)*x_ref[1]**o+(0.889105964295)*x_ref[1]
        ref[2,3,0,0]=(-0.731872136283)*x_ref[0]**o+(0.237034931745)*x_ref[0]+(0.547737117698)*x_ref[1]**o+(0.364733861661)*x_ref[1]
        ref[2,3,0,1]=(0.971716549689)*x_ref[0]**o+(0.466300970081)*x_ref[0]+(-0.51802162747)*x_ref[1]**o+(-0.859910778018)*x_ref[1]
        ref[2,3,1,0]=(-0.562918219893)*x_ref[0]**o+(-0.0532549959391)*x_ref[0]+(0.842033495359)*x_ref[1]**o+(-0.708554283815)*x_ref[1]
        ref[2,3,1,1]=(0.437125534104)*x_ref[0]**o+(-0.376751615116)*x_ref[0]+(-0.330889383623)*x_ref[1]**o+(0.958884035289)*x_ref[1]
        ref[2,3,2,0]=(-0.194290501566)*x_ref[0]**o+(-0.368131904292)*x_ref[0]+(0.802917650409)*x_ref[1]**o+(0.628603987906)*x_ref[1]
        ref[2,3,2,1]=(0.70219875881)*x_ref[0]**o+(0.822181339005)*x_ref[0]+(0.768635447234)*x_ref[1]**o+(0.871676010439)*x_ref[1]
        ref[2,4,0,0]=(-0.654498700662)*x_ref[0]**o+(0.851143676147)*x_ref[0]+(0.468468395415)*x_ref[1]**o+(-0.447240713576)*x_ref[1]
        ref[2,4,0,1]=(0.893177052112)*x_ref[0]**o+(0.652743480988)*x_ref[0]+(0.726646163682)*x_ref[1]**o+(-0.854895731892)*x_ref[1]
        ref[2,4,1,0]=(0.0540542301334)*x_ref[0]**o+(-0.208654443984)*x_ref[0]+(0.925127576853)*x_ref[1]**o+(-0.932489814868)*x_ref[1]
        ref[2,4,1,1]=(-0.460268214104)*x_ref[0]**o+(-0.303569308977)*x_ref[0]+(-0.381458880316)*x_ref[1]**o+(0.690537661673)*x_ref[1]
        ref[2,4,2,0]=(0.401012366366)*x_ref[0]**o+(-0.970993441679)*x_ref[0]+(0.0249535421254)*x_ref[1]**o+(-0.499737254014)*x_ref[1]
        ref[2,4,2,1]=(-0.692604790859)*x_ref[0]**o+(-0.685292098725)*x_ref[0]+(0.22550357354)*x_ref[1]**o+(0.908152783462)*x_ref[1]
        ref[3,0,0,0]=(-0.634246407254)*x_ref[0]**o+(-0.571223033336)*x_ref[0]+(0.0574319920181)*x_ref[1]**o+(0.123763929908)*x_ref[1]
        ref[3,0,0,1]=(-0.569358219213)*x_ref[0]**o+(-0.107237509632)*x_ref[0]+(0.97672999497)*x_ref[1]**o+(-0.114939366813)*x_ref[1]
        ref[3,0,1,0]=(-0.767885152479)*x_ref[0]**o+(-0.105924102555)*x_ref[0]+(0.539361607398)*x_ref[1]**o+(0.546186446469)*x_ref[1]
        ref[3,0,1,1]=(-0.58818696897)*x_ref[0]**o+(0.629371516566)*x_ref[0]+(-0.598173904376)*x_ref[1]**o+(0.959043112088)*x_ref[1]
        ref[3,0,2,0]=(0.751165393927)*x_ref[0]**o+(0.732151619397)*x_ref[0]+(-0.435276735735)*x_ref[1]**o+(0.947142140722)*x_ref[1]
        ref[3,0,2,1]=(0.463904349547)*x_ref[0]**o+(-0.947270841014)*x_ref[0]+(0.304239969784)*x_ref[1]**o+(-0.233566944292)*x_ref[1]
        ref[3,1,0,0]=(0.961122232395)*x_ref[0]**o+(-0.989040215156)*x_ref[0]+(0.833186746232)*x_ref[1]**o+(0.226673105305)*x_ref[1]
        ref[3,1,0,1]=(-0.156957616578)*x_ref[0]**o+(-0.361665587314)*x_ref[0]+(0.987490273947)*x_ref[1]**o+(-0.890271058746)*x_ref[1]
        ref[3,1,1,0]=(-0.259959222953)*x_ref[0]**o+(-0.696135598749)*x_ref[0]+(0.234411151165)*x_ref[1]**o+(0.783105897256)*x_ref[1]
        ref[3,1,1,1]=(0.020886322199)*x_ref[0]**o+(0.887445555864)*x_ref[0]+(0.709489180254)*x_ref[1]**o+(0.990590112507)*x_ref[1]
        ref[3,1,2,0]=(-0.963790247866)*x_ref[0]**o+(-0.417180971158)*x_ref[0]+(0.303859261646)*x_ref[1]**o+(0.745552024399)*x_ref[1]
        ref[3,1,2,1]=(-0.364448135641)*x_ref[0]**o+(0.760841571414)*x_ref[0]+(-0.0283374186763)*x_ref[1]**o+(-0.491433855748)*x_ref[1]
        ref[3,2,0,0]=(-0.649766142291)*x_ref[0]**o+(-0.177410238297)*x_ref[0]+(0.137316641289)*x_ref[1]**o+(-0.40890299152)*x_ref[1]
        ref[3,2,0,1]=(0.282922984657)*x_ref[0]**o+(-0.660238152943)*x_ref[0]+(0.481726052488)*x_ref[1]**o+(0.976011254218)*x_ref[1]
        ref[3,2,1,0]=(0.989761520473)*x_ref[0]**o+(0.194082255647)*x_ref[0]+(-0.648435682551)*x_ref[1]**o+(-0.721569599311)*x_ref[1]
        ref[3,2,1,1]=(0.734634540883)*x_ref[0]**o+(0.524638529049)*x_ref[0]+(-0.737636791717)*x_ref[1]**o+(-0.995904525822)*x_ref[1]
        ref[3,2,2,0]=(0.422396808663)*x_ref[0]**o+(-0.820639154105)*x_ref[0]+(-0.566207315373)*x_ref[1]**o+(-0.224174711984)*x_ref[1]
        ref[3,2,2,1]=(0.719145374997)*x_ref[0]**o+(-0.885860499468)*x_ref[0]+(0.184062878609)*x_ref[1]**o+(0.47660543007)*x_ref[1]
        ref[3,3,0,0]=(-0.0122290849915)*x_ref[0]**o+(-0.0314298339876)*x_ref[0]+(0.937255407134)*x_ref[1]**o+(-0.754858547072)*x_ref[1]
        ref[3,3,0,1]=(-0.507826956669)*x_ref[0]**o+(0.951205020689)*x_ref[0]+(0.837096245109)*x_ref[1]**o+(-0.576692399518)*x_ref[1]
        ref[3,3,1,0]=(0.602119808238)*x_ref[0]**o+(-0.943937134298)*x_ref[0]+(-0.244250254626)*x_ref[1]**o+(-0.814103542727)*x_ref[1]
        ref[3,3,1,1]=(-0.51811661822)*x_ref[0]**o+(-0.188969078762)*x_ref[0]+(-0.697470816837)*x_ref[1]**o+(-0.741593014841)*x_ref[1]
        ref[3,3,2,0]=(-0.673948933786)*x_ref[0]**o+(0.816404336566)*x_ref[0]+(0.339239090092)*x_ref[1]**o+(-0.0173550947113)*x_ref[1]
        ref[3,3,2,1]=(0.499109852716)*x_ref[0]**o+(0.515501552893)*x_ref[0]+(0.926442939974)*x_ref[1]**o+(0.157227173254)*x_ref[1]
        ref[3,4,0,0]=(-0.1837389273)*x_ref[0]**o+(0.78490229721)*x_ref[0]+(-0.248033788374)*x_ref[1]**o+(0.735136386405)*x_ref[1]
        ref[3,4,0,1]=(0.121169550194)*x_ref[0]**o+(-0.854027732077)*x_ref[0]+(0.937680848475)*x_ref[1]**o+(-0.447184197264)*x_ref[1]
        ref[3,4,1,0]=(-0.788610934694)*x_ref[0]**o+(-0.397825388144)*x_ref[0]+(0.604421361449)*x_ref[1]**o+(0.937184552821)*x_ref[1]
        ref[3,4,1,1]=(-0.493380143496)*x_ref[0]**o+(0.958197000398)*x_ref[0]+(0.282372561751)*x_ref[1]**o+(-0.55322669826)*x_ref[1]
        ref[3,4,2,0]=(0.11333865923)*x_ref[0]**o+(-0.542805511569)*x_ref[0]+(-0.640646701928)*x_ref[1]**o+(-0.342030716597)*x_ref[1]
        ref[3,4,2,1]=(0.505356266297)*x_ref[0]**o+(-0.107422782294)*x_ref[0]+(0.542833447968)*x_ref[1]**o+(-0.615275716592)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.507422888369)*x[0]**o+(0.808798513305)*x[0]+(0.51493073805)*x[1]**o+(-0.928527686947)*x[1]+(-0.788010447498)*x[2]**o+(-0.378389833131)*x[2]
        arg[0,0,0,1]=(-0.475851243177)*x[0]**o+(0.746578446988)*x[0]+(0.504778622494)*x[1]**o+(-0.677770398899)*x[1]+(-0.862535648403)*x[2]**o+(0.620414141691)*x[2]
        arg[0,0,1,0]=(-0.361726879582)*x[0]**o+(0.554584127587)*x[0]+(0.436610414772)*x[1]**o+(-0.686534492717)*x[1]+(0.757226061419)*x[2]**o+(0.0948537725389)*x[2]
        arg[0,0,1,1]=(-0.362160696355)*x[0]**o+(0.766524051703)*x[0]+(0.209980857795)*x[1]**o+(-0.478398061437)*x[1]+(-0.410960179122)*x[2]**o+(-0.213272138462)*x[2]
        arg[0,0,2,0]=(0.87154872554)*x[0]**o+(-0.727915301904)*x[0]+(0.683580698056)*x[1]**o+(0.215052521277)*x[1]+(0.267231507898)*x[2]**o+(-0.0124189912449)*x[2]
        arg[0,0,2,1]=(0.631400944876)*x[0]**o+(0.537293628606)*x[0]+(0.944754350994)*x[1]**o+(0.108068908012)*x[1]+(0.890117096365)*x[2]**o+(0.164706867235)*x[2]
        arg[0,1,0,0]=(-0.935871759207)*x[0]**o+(0.350133758077)*x[0]+(-0.966482181015)*x[1]**o+(0.440902701528)*x[1]+(-0.893329541322)*x[2]**o+(0.350317827938)*x[2]
        arg[0,1,0,1]=(0.465305203601)*x[0]**o+(0.711131797928)*x[0]+(0.104091536498)*x[1]**o+(0.484811628127)*x[1]+(0.676878954516)*x[2]**o+(-0.99327016995)*x[2]
        arg[0,1,1,0]=(-0.361189016151)*x[0]**o+(0.962468573492)*x[0]+(-0.158571122564)*x[1]**o+(-0.293583660912)*x[1]+(-0.713699173608)*x[2]**o+(-0.507430163361)*x[2]
        arg[0,1,1,1]=(0.172564527054)*x[0]**o+(0.273438145027)*x[0]+(0.137378266239)*x[1]**o+(-0.675788606841)*x[1]+(-0.362143416121)*x[2]**o+(0.0703897749775)*x[2]
        arg[0,1,2,0]=(-0.941374717271)*x[0]**o+(0.289433165458)*x[0]+(0.318615227028)*x[1]**o+(0.524362568204)*x[1]+(-0.612590305095)*x[2]**o+(-0.117555538957)*x[2]
        arg[0,1,2,1]=(-0.442920825551)*x[0]**o+(-0.454982146579)*x[0]+(0.355353737589)*x[1]**o+(-0.729623917004)*x[1]+(-0.567220335937)*x[2]**o+(0.977652639976)*x[2]
        arg[0,2,0,0]=(0.942485930373)*x[0]**o+(-0.605823961033)*x[0]+(-0.143465779338)*x[1]**o+(0.327354308987)*x[1]+(0.691900417542)*x[2]**o+(-0.514165213849)*x[2]
        arg[0,2,0,1]=(0.018029272977)*x[0]**o+(-0.209992705469)*x[0]+(-0.224614727554)*x[1]**o+(-0.862697462048)*x[1]+(0.274791387928)*x[2]**o+(-0.907445869506)*x[2]
        arg[0,2,1,0]=(0.862755248265)*x[0]**o+(0.020831641939)*x[0]+(0.496662599261)*x[1]**o+(0.79318975953)*x[1]+(0.0912166895541)*x[2]**o+(0.556337610316)*x[2]
        arg[0,2,1,1]=(-0.09652997071)*x[0]**o+(-0.954279421394)*x[0]+(-0.601053524018)*x[1]**o+(0.642528012068)*x[1]+(0.295748113888)*x[2]**o+(-0.318166346186)*x[2]
        arg[0,2,2,0]=(-0.914878854129)*x[0]**o+(0.159201123642)*x[0]+(0.589819056368)*x[1]**o+(-0.715763226823)*x[1]+(-0.444572583567)*x[2]**o+(-0.487866651367)*x[2]
        arg[0,2,2,1]=(-0.71271969175)*x[0]**o+(0.454616308594)*x[0]+(0.510270095021)*x[1]**o+(-0.0385822967955)*x[1]+(-0.519736540313)*x[2]**o+(0.0556733091556)*x[2]
        arg[0,3,0,0]=(0.355768034846)*x[0]**o+(0.152240594627)*x[0]+(-0.81081131843)*x[1]**o+(-0.913373483156)*x[1]+(-0.319859744799)*x[2]**o+(-0.306218435393)*x[2]
        arg[0,3,0,1]=(-0.656932063693)*x[0]**o+(0.832886861783)*x[0]+(-0.417112750683)*x[1]**o+(-0.388190265957)*x[1]+(0.246534947096)*x[2]**o+(-0.896822839189)*x[2]
        arg[0,3,1,0]=(0.204713418891)*x[0]**o+(-0.150097488817)*x[0]+(0.806578426941)*x[1]**o+(-0.590052708995)*x[1]+(-0.758958608419)*x[2]**o+(-0.221973528232)*x[2]
        arg[0,3,1,1]=(-0.854308184618)*x[0]**o+(-0.292526203646)*x[0]+(-0.662403504833)*x[1]**o+(0.0955657982227)*x[1]+(0.69471076173)*x[2]**o+(0.384077309264)*x[2]
        arg[0,3,2,0]=(0.982057200864)*x[0]**o+(-0.64758727704)*x[0]+(0.15941493222)*x[1]**o+(-0.17745437115)*x[1]+(-0.72953252955)*x[2]**o+(0.986979270802)*x[2]
        arg[0,3,2,1]=(0.764526698779)*x[0]**o+(0.22809521115)*x[0]+(-0.0762092087998)*x[1]**o+(0.415212410471)*x[1]+(-0.263638988463)*x[2]**o+(-0.934669942336)*x[2]
        arg[0,4,0,0]=(0.0812656932707)*x[0]**o+(0.0397235062518)*x[0]+(-0.0496912823983)*x[1]**o+(0.189891585189)*x[1]+(0.243537823293)*x[2]**o+(0.461256134208)*x[2]
        arg[0,4,0,1]=(0.167596358861)*x[0]**o+(0.441650690481)*x[0]+(0.638100632205)*x[1]**o+(-0.69240615765)*x[1]+(-0.501855369512)*x[2]**o+(0.769465683782)*x[2]
        arg[0,4,1,0]=(-0.137149398927)*x[0]**o+(-0.186548325285)*x[0]+(-0.504656948352)*x[1]**o+(0.228896656134)*x[1]+(-0.472425358061)*x[2]**o+(-0.0769834754996)*x[2]
        arg[0,4,1,1]=(0.311655648626)*x[0]**o+(0.856513683959)*x[0]+(0.47173626537)*x[1]**o+(0.245497333467)*x[1]+(-0.513727691305)*x[2]**o+(-0.588614785682)*x[2]
        arg[0,4,2,0]=(-0.152907854625)*x[0]**o+(-0.86229950575)*x[0]+(-0.931356414564)*x[1]**o+(-0.915667991864)*x[1]+(0.322316118527)*x[2]**o+(-0.987559462539)*x[2]
        arg[0,4,2,1]=(-0.705947769164)*x[0]**o+(0.104081366627)*x[0]+(-0.540454644392)*x[1]**o+(0.404522880836)*x[1]+(0.146464264141)*x[2]**o+(0.586755030006)*x[2]
        arg[1,0,0,0]=(-0.466614768106)*x[0]**o+(0.482681414943)*x[0]+(-0.176553738108)*x[1]**o+(-0.39162272813)*x[1]+(0.74312023324)*x[2]**o+(0.824538905894)*x[2]
        arg[1,0,0,1]=(-0.9351073631)*x[0]**o+(-0.640226640449)*x[0]+(0.966208762054)*x[1]**o+(0.101792167944)*x[1]+(-0.732983657038)*x[2]**o+(-0.112313375937)*x[2]
        arg[1,0,1,0]=(-0.305403470434)*x[0]**o+(-0.414397930347)*x[0]+(0.356093254322)*x[1]**o+(-0.52607560749)*x[1]+(-0.70070674473)*x[2]**o+(-0.0192901616015)*x[2]
        arg[1,0,1,1]=(0.450046989211)*x[0]**o+(0.602770057268)*x[0]+(0.0193812509113)*x[1]**o+(-0.275615149191)*x[1]+(0.0180468267836)*x[2]**o+(-0.81388255113)*x[2]
        arg[1,0,2,0]=(0.913651483249)*x[0]**o+(-0.92188313743)*x[0]+(0.680007150195)*x[1]**o+(-0.300145584708)*x[1]+(-0.398883520002)*x[2]**o+(-0.752781496463)*x[2]
        arg[1,0,2,1]=(-0.265749827845)*x[0]**o+(0.594379728352)*x[0]+(0.463409943587)*x[1]**o+(-0.85840420426)*x[1]+(-0.705963756674)*x[2]**o+(0.251711112887)*x[2]
        arg[1,1,0,0]=(-0.28895690781)*x[0]**o+(0.476461656166)*x[0]+(0.0692567954597)*x[1]**o+(0.88763437878)*x[1]+(0.0872256197554)*x[2]**o+(0.19467039694)*x[2]
        arg[1,1,0,1]=(0.741045418981)*x[0]**o+(-0.726935151332)*x[0]+(0.899418662131)*x[1]**o+(0.27516497286)*x[1]+(-0.715927474971)*x[2]**o+(-0.709283098732)*x[2]
        arg[1,1,1,0]=(0.991407130174)*x[0]**o+(-0.163484265852)*x[0]+(-0.983254548659)*x[1]**o+(0.0568443855477)*x[1]+(0.309664644588)*x[2]**o+(0.904334718845)*x[2]
        arg[1,1,1,1]=(0.237261672833)*x[0]**o+(0.0861047746345)*x[0]+(-0.494898018662)*x[1]**o+(-0.433075334646)*x[1]+(0.588597346505)*x[2]**o+(-0.185559567553)*x[2]
        arg[1,1,2,0]=(-0.969384091709)*x[0]**o+(-0.0229777223491)*x[0]+(0.0945623059666)*x[1]**o+(0.303543647034)*x[1]+(0.659000135381)*x[2]**o+(0.00873791711254)*x[2]
        arg[1,1,2,1]=(-0.421406731882)*x[0]**o+(-0.5017394377)*x[0]+(0.027747130694)*x[1]**o+(0.327278873996)*x[1]+(-0.754133439304)*x[2]**o+(0.673525803845)*x[2]
        arg[1,2,0,0]=(-0.104620319685)*x[0]**o+(-0.685602393278)*x[0]+(0.972173717014)*x[1]**o+(-0.440573017063)*x[1]+(-0.709529395743)*x[2]**o+(0.634170741457)*x[2]
        arg[1,2,0,1]=(-0.560386754264)*x[0]**o+(0.570894623654)*x[0]+(-0.455122833362)*x[1]**o+(-0.00322505937935)*x[1]+(0.959400336149)*x[2]**o+(0.64014859954)*x[2]
        arg[1,2,1,0]=(0.100585233221)*x[0]**o+(0.939532360957)*x[0]+(-0.900301955112)*x[1]**o+(-0.776375896786)*x[1]+(-0.779373294746)*x[2]**o+(0.546157461116)*x[2]
        arg[1,2,1,1]=(-0.680457315071)*x[0]**o+(-0.0278844539208)*x[0]+(-0.629673590044)*x[1]**o+(0.813218376348)*x[1]+(-0.449484131841)*x[2]**o+(0.641682150048)*x[2]
        arg[1,2,2,0]=(0.60056954386)*x[0]**o+(-0.763169434657)*x[0]+(-0.672890915035)*x[1]**o+(-0.368711154571)*x[1]+(0.476810195318)*x[2]**o+(-0.935071124382)*x[2]
        arg[1,2,2,1]=(0.994100148333)*x[0]**o+(0.471647033381)*x[0]+(0.341677296654)*x[1]**o+(0.466442262068)*x[1]+(-0.0634919249117)*x[2]**o+(0.0637348830912)*x[2]
        arg[1,3,0,0]=(-0.120744286726)*x[0]**o+(0.630378984537)*x[0]+(0.0837817796507)*x[1]**o+(-0.850115274881)*x[1]+(0.574342626561)*x[2]**o+(-0.705409915626)*x[2]
        arg[1,3,0,1]=(-0.51181061391)*x[0]**o+(0.711479747616)*x[0]+(-0.00800928076734)*x[1]**o+(-0.703424580277)*x[1]+(-0.925935026157)*x[2]**o+(0.488175992827)*x[2]
        arg[1,3,1,0]=(0.531879615751)*x[0]**o+(-0.214270521934)*x[0]+(-0.432344726015)*x[1]**o+(-0.107660826449)*x[1]+(0.763406427398)*x[2]**o+(-0.511000150429)*x[2]
        arg[1,3,1,1]=(-0.867467568206)*x[0]**o+(-0.738425961596)*x[0]+(-0.925337667706)*x[1]**o+(-0.586227765107)*x[1]+(0.909987211462)*x[2]**o+(0.533347022582)*x[2]
        arg[1,3,2,0]=(-0.47713113862)*x[0]**o+(-0.130876138883)*x[0]+(-0.607996835981)*x[1]**o+(-0.0306261551172)*x[1]+(0.956207500475)*x[2]**o+(-0.287731719382)*x[2]
        arg[1,3,2,1]=(0.179479486943)*x[0]**o+(0.361266562479)*x[0]+(0.374904472267)*x[1]**o+(0.00788379106301)*x[1]+(-0.34090684265)*x[2]**o+(-0.190875910791)*x[2]
        arg[1,4,0,0]=(0.685464203729)*x[0]**o+(0.0634561930987)*x[0]+(-0.662534110385)*x[1]**o+(0.383849898086)*x[1]+(0.948230524205)*x[2]**o+(-0.347042281976)*x[2]
        arg[1,4,0,1]=(-0.484499755322)*x[0]**o+(-0.765655349243)*x[0]+(0.805331514606)*x[1]**o+(-0.0635912547496)*x[1]+(-0.139457134772)*x[2]**o+(0.227446734221)*x[2]
        arg[1,4,1,0]=(-0.913971229647)*x[0]**o+(0.042531889985)*x[0]+(0.143956410289)*x[1]**o+(0.597990264168)*x[1]+(-0.85926903006)*x[2]**o+(0.205900361792)*x[2]
        arg[1,4,1,1]=(-0.955540264653)*x[0]**o+(0.983439073814)*x[0]+(0.246870351555)*x[1]**o+(-0.610917180843)*x[1]+(-0.0418134447092)*x[2]**o+(-0.58407787472)*x[2]
        arg[1,4,2,0]=(0.787274711453)*x[0]**o+(-0.899651339467)*x[0]+(-0.62195254225)*x[1]**o+(-0.516447289251)*x[1]+(0.267082643601)*x[2]**o+(0.0408186039688)*x[2]
        arg[1,4,2,1]=(-0.485773801895)*x[0]**o+(0.663478067655)*x[0]+(0.928973237869)*x[1]**o+(0.353735145156)*x[1]+(0.128409748978)*x[2]**o+(-0.476479648592)*x[2]
        arg[2,0,0,0]=(0.491574821811)*x[0]**o+(0.455303820526)*x[0]+(-0.747773112075)*x[1]**o+(-0.653255987137)*x[1]+(-0.738293040346)*x[2]**o+(0.948070477272)*x[2]
        arg[2,0,0,1]=(-0.697412782403)*x[0]**o+(-0.738389544586)*x[0]+(0.231728579935)*x[1]**o+(0.248503931032)*x[1]+(-0.561155590703)*x[2]**o+(0.324682113147)*x[2]
        arg[2,0,1,0]=(-0.00507638382064)*x[0]**o+(-0.983105866448)*x[0]+(-0.253141627484)*x[1]**o+(-0.0664178091069)*x[1]+(0.194460220144)*x[2]**o+(-0.174257469198)*x[2]
        arg[2,0,1,1]=(-0.644447560463)*x[0]**o+(0.658703761133)*x[0]+(-0.685392825104)*x[1]**o+(-0.766295260796)*x[1]+(-0.922004388493)*x[2]**o+(-0.0745707720923)*x[2]
        arg[2,0,2,0]=(-0.69924009606)*x[0]**o+(-0.398212597328)*x[0]+(-0.364991550186)*x[1]**o+(0.243474288977)*x[1]+(0.129545162767)*x[2]**o+(0.920611954845)*x[2]
        arg[2,0,2,1]=(-0.494476126875)*x[0]**o+(-0.131503103148)*x[0]+(-0.241670500874)*x[1]**o+(0.40439125905)*x[1]+(0.134166779399)*x[2]**o+(-0.507385751809)*x[2]
        arg[2,1,0,0]=(0.301148115913)*x[0]**o+(0.189635133395)*x[0]+(0.744167404774)*x[1]**o+(0.995754344203)*x[1]+(-0.334760662218)*x[2]**o+(-0.182462923741)*x[2]
        arg[2,1,0,1]=(0.523500022508)*x[0]**o+(-0.257042661558)*x[0]+(-0.940233100381)*x[1]**o+(-0.202431731891)*x[1]+(-0.653480190287)*x[2]**o+(0.993156915081)*x[2]
        arg[2,1,1,0]=(-0.502452234661)*x[0]**o+(-0.711581805303)*x[0]+(0.696450091152)*x[1]**o+(0.436746953802)*x[1]+(0.695969418438)*x[2]**o+(-0.671400022595)*x[2]
        arg[2,1,1,1]=(0.0705944125182)*x[0]**o+(-0.865755834485)*x[0]+(0.921555996641)*x[1]**o+(-0.567855814858)*x[1]+(-0.213455192388)*x[2]**o+(0.820868399259)*x[2]
        arg[2,1,2,0]=(0.744232126707)*x[0]**o+(0.336969176485)*x[0]+(0.339407729288)*x[1]**o+(-0.465154968917)*x[1]+(0.389524167202)*x[2]**o+(0.645394954477)*x[2]
        arg[2,1,2,1]=(-0.299085687516)*x[0]**o+(0.220795864743)*x[0]+(-0.19353913538)*x[1]**o+(0.445131396297)*x[1]+(-0.779605318636)*x[2]**o+(-0.76707264371)*x[2]
        arg[2,2,0,0]=(-0.357585997171)*x[0]**o+(-0.248795296406)*x[0]+(0.0876364714478)*x[1]**o+(-0.682833482179)*x[1]+(0.497052575986)*x[2]**o+(0.893378007889)*x[2]
        arg[2,2,0,1]=(0.38067999218)*x[0]**o+(-0.857246232026)*x[0]+(0.0785818284359)*x[1]**o+(-0.473287028123)*x[1]+(-0.763041318956)*x[2]**o+(-0.1119710251)*x[2]
        arg[2,2,1,0]=(-0.0278856926369)*x[0]**o+(0.435955626553)*x[0]+(-0.676274741782)*x[1]**o+(-0.719721825467)*x[1]+(0.195007734703)*x[2]**o+(0.504771278902)*x[2]
        arg[2,2,1,1]=(0.459547475772)*x[0]**o+(-0.669041681613)*x[0]+(-0.741957874364)*x[1]**o+(0.604585337407)*x[1]+(-0.886718742046)*x[2]**o+(0.338113237761)*x[2]
        arg[2,2,2,0]=(0.392926733678)*x[0]**o+(-0.923680589861)*x[0]+(-0.425503328457)*x[1]**o+(-0.0902669250117)*x[1]+(-0.346726072703)*x[2]**o+(0.541482976709)*x[2]
        arg[2,2,2,1]=(0.134867023616)*x[0]**o+(0.478349607348)*x[0]+(-0.334617836366)*x[1]**o+(-0.74587411744)*x[1]+(-0.880365937567)*x[2]**o+(0.820743064452)*x[2]
        arg[2,3,0,0]=(0.081184173895)*x[0]**o+(0.315011929905)*x[0]+(-0.137083926616)*x[1]**o+(0.103926223299)*x[1]+(0.363270989968)*x[2]**o+(-0.454075563606)*x[2]
        arg[2,3,0,1]=(0.457730701321)*x[0]**o+(0.922345625691)*x[0]+(-0.412933933474)*x[1]**o+(-0.546359655103)*x[1]+(0.719101029933)*x[2]**o+(-0.790522859322)*x[2]
        arg[2,3,1,0]=(-0.811816419841)*x[0]**o+(0.0379298408788)*x[0]+(-0.852643125131)*x[1]**o+(0.76881274254)*x[1]+(0.75214553496)*x[2]**o+(0.800382814403)*x[2]
        arg[2,3,1,1]=(0.524802143618)*x[0]**o+(-0.717457078371)*x[0]+(0.389026887965)*x[1]**o+(-0.231912558721)*x[1]+(-0.394705562333)*x[2]**o+(-0.794723802655)*x[2]
        arg[2,3,2,0]=(-0.232215520303)*x[0]**o+(0.994932250617)*x[0]+(-0.760425274222)*x[1]**o+(0.0910727760744)*x[1]+(0.621642024132)*x[2]**o+(0.263978719304)*x[2]
        arg[2,3,2,1]=(0.515269404424)*x[0]**o+(-0.998266958288)*x[0]+(-0.485078033534)*x[1]**o+(-0.960043247553)*x[1]+(-0.926562413314)*x[2]**o+(0.612710762712)*x[2]
        arg[2,4,0,0]=(0.419808446591)*x[0]**o+(0.132792190191)*x[0]+(-0.690603562736)*x[1]**o+(0.45137757885)*x[1]+(-0.262249041567)*x[2]**o+(-0.432995254783)*x[2]
        arg[2,4,0,1]=(0.62413323935)*x[0]**o+(0.963559148607)*x[0]+(-0.879044261015)*x[1]**o+(-0.504981574957)*x[1]+(-0.575204029973)*x[2]**o+(0.066855791448)*x[2]
        arg[2,4,1,0]=(0.321218662791)*x[0]**o+(0.975103526588)*x[0]+(0.783716185844)*x[1]**o+(-0.816368139389)*x[1]+(0.778169798128)*x[2]**o+(-0.527443068085)*x[2]
        arg[2,4,1,1]=(-0.8801082648)*x[0]**o+(-0.0313845929272)*x[0]+(0.485262237838)*x[1]**o+(0.0778888535573)*x[1]+(-0.900948214929)*x[2]**o+(-0.98830300089)*x[2]
        arg[2,4,2,0]=(-0.996395042226)*x[0]**o+(0.253635852642)*x[0]+(0.180965173659)*x[1]**o+(-0.488294662459)*x[1]+(-0.144583511741)*x[2]**o+(0.285854081388)*x[2]
        arg[2,4,2,1]=(0.00988506292061)*x[0]**o+(-0.00192786315452)*x[0]+(0.720775470647)*x[1]**o+(-0.587673311506)*x[1]+(0.313836238675)*x[2]**o+(0.45975125457)*x[2]
        arg[3,0,0,0]=(-0.822049489749)*x[0]**o+(-0.989589435175)*x[0]+(-0.0943830434118)*x[1]**o+(-0.789229900035)*x[1]+(0.311382619317)*x[2]**o+(-0.316646938685)*x[2]
        arg[3,0,0,1]=(0.501517916677)*x[0]**o+(-0.704502365436)*x[0]+(0.913184728006)*x[1]**o+(-0.843712848791)*x[1]+(0.901479696878)*x[2]**o+(-0.713893336542)*x[2]
        arg[3,0,1,0]=(0.0771592811086)*x[0]**o+(0.796038032206)*x[0]+(0.566256212871)*x[1]**o+(0.394136659986)*x[1]+(-0.794422065388)*x[2]**o+(0.662174992848)*x[2]
        arg[3,0,1,1]=(0.180225115167)*x[0]**o+(0.366965867971)*x[0]+(-0.244015756153)*x[1]**o+(0.433571067335)*x[1]+(0.129075150185)*x[2]**o+(-0.216304414605)*x[2]
        arg[3,0,2,0]=(0.100277967034)*x[0]**o+(-0.590583546786)*x[0]+(-0.595553697152)*x[1]**o+(0.923573252239)*x[1]+(0.578966663184)*x[2]**o+(-0.411082913097)*x[2]
        arg[3,0,2,1]=(-0.575315696811)*x[0]**o+(0.608760285918)*x[0]+(-0.3592074605)*x[1]**o+(0.515119945308)*x[1]+(-0.4000138306)*x[2]**o+(0.477685324937)*x[2]
        arg[3,1,0,0]=(0.0780014276919)*x[0]**o+(-0.0932830671471)*x[0]+(0.39477236642)*x[1]**o+(0.406673135556)*x[1]+(-0.235480707568)*x[2]**o+(-0.872302257144)*x[2]
        arg[3,1,0,1]=(0.170573072646)*x[0]**o+(0.868146561326)*x[0]+(0.262136343896)*x[1]**o+(0.192343529912)*x[1]+(0.157400965693)*x[2]**o+(-0.932981619473)*x[2]
        arg[3,1,1,0]=(0.891433179291)*x[0]**o+(-0.281396449997)*x[0]+(-0.0803173619772)*x[1]**o+(-0.378460054356)*x[1]+(-0.581578656014)*x[2]**o+(-0.65856067509)*x[2]
        arg[3,1,1,1]=(0.343599956626)*x[0]**o+(-0.92063694125)*x[0]+(0.440698102072)*x[1]**o+(0.767124166388)*x[1]+(0.050964317543)*x[2]**o+(1.35605562239e-05)*x[2]
        arg[3,1,2,0]=(-0.614268150987)*x[0]**o+(0.363806471556)*x[0]+(0.92193457905)*x[1]**o+(0.157919580642)*x[1]+(-0.955575321124)*x[2]**o+(-0.984808437666)*x[2]
        arg[3,1,2,1]=(0.030099648966)*x[0]**o+(-0.375999369491)*x[0]+(-0.0992643926008)*x[1]**o+(0.109209161093)*x[1]+(-0.974043780088)*x[2]**o+(0.908049658745)*x[2]
        arg[3,2,0,0]=(0.297771341106)*x[0]**o+(-0.637802228169)*x[0]+(0.398100441783)*x[1]**o+(-0.0643425181651)*x[1]+(0.662775339033)*x[2]**o+(0.744835182052)*x[2]
        arg[3,2,0,1]=(0.312026420583)*x[0]**o+(0.865855052574)*x[0]+(0.0195001783165)*x[1]**o+(0.537899238473)*x[1]+(0.554893357481)*x[2]**o+(0.71555610145)*x[2]
        arg[3,2,1,0]=(-0.199138146186)*x[0]**o+(0.827546489855)*x[0]+(-0.480965464736)*x[1]**o+(-0.473796093187)*x[1]+(0.880264675049)*x[2]**o+(0.544506670844)*x[2]
        arg[3,2,1,1]=(0.773798005008)*x[0]**o+(0.744945241634)*x[0]+(0.548170564608)*x[1]**o+(-0.560503692823)*x[1]+(-0.435630215427)*x[2]**o+(0.080015730707)*x[2]
        arg[3,2,2,0]=(0.36735714937)*x[0]**o+(-0.867071557873)*x[0]+(0.936135650431)*x[1]**o+(-0.432533600524)*x[1]+(-0.0224430204069)*x[2]**o+(0.218590302621)*x[2]
        arg[3,2,2,1]=(0.932808382499)*x[0]**o+(-0.0970016713319)*x[0]+(0.458943467579)*x[1]**o+(-0.134338789898)*x[1]+(-0.426892557314)*x[2]**o+(0.627071826863)*x[2]
        arg[3,3,0,0]=(-0.623804149075)*x[0]**o+(0.910423590172)*x[0]+(0.465938181097)*x[1]**o+(0.495479245391)*x[1]+(0.693745442185)*x[2]**o+(-0.437014226757)*x[2]
        arg[3,3,0,1]=(0.598807017496)*x[0]**o+(-0.112361840211)*x[0]+(-0.550899685213)*x[1]**o+(-0.0478222019564)*x[1]+(-0.0110129517693)*x[2]**o+(-0.059534786374)*x[2]
        arg[3,3,1,0]=(-0.561155633409)*x[0]**o+(-0.463563057636)*x[0]+(-0.170463996009)*x[1]**o+(-0.214509204942)*x[1]+(0.851200198666)*x[2]**o+(-0.721036329689)*x[2]
        arg[3,3,1,1]=(0.608666240405)*x[0]**o+(0.814232455115)*x[0]+(-0.604645629689)*x[1]**o+(0.960976509673)*x[1]+(0.365807996219)*x[2]**o+(-0.87491909238)*x[2]
        arg[3,3,2,0]=(0.120126988418)*x[0]**o+(-0.745687314825)*x[0]+(-0.332697826984)*x[1]**o+(-0.823683406516)*x[1]+(0.305444079613)*x[2]**o+(0.901972115441)*x[2]
        arg[3,3,2,1]=(0.0470293848904)*x[0]**o+(-0.195496754814)*x[0]+(0.671870114105)*x[1]**o+(0.170273146371)*x[1]+(-0.851927986188)*x[2]**o+(0.772192863973)*x[2]
        arg[3,4,0,0]=(-0.95392341727)*x[0]**o+(-0.665694215521)*x[0]+(-0.0247396499378)*x[1]**o+(0.441352857387)*x[1]+(0.767156547791)*x[2]**o+(-0.14696372018)*x[2]
        arg[3,4,0,1]=(0.209761166815)*x[0]**o+(0.978934473425)*x[0]+(0.943806891053)*x[1]**o+(-0.756619917601)*x[1]+(0.635699949764)*x[2]**o+(-0.345248269568)*x[2]
        arg[3,4,1,0]=(0.748020785785)*x[0]**o+(0.311104549542)*x[0]+(0.739127551248)*x[1]**o+(-0.635238575859)*x[1]+(0.623898875653)*x[2]**o+(0.343407392068)*x[2]
        arg[3,4,1,1]=(-0.422500892471)*x[0]**o+(0.0688495013346)*x[0]+(-0.919653906232)*x[1]**o+(0.531036586396)*x[1]+(-0.22487910672)*x[2]**o+(-0.848225994541)*x[2]
        arg[3,4,2,0]=(0.723965744673)*x[0]**o+(0.558602959224)*x[0]+(0.699539243308)*x[1]**o+(-0.597426541715)*x[1]+(-0.0128675070522)*x[2]**o+(-0.874796626242)*x[2]
        arg[3,4,2,1]=(-0.962970021216)*x[0]**o+(0.532886481874)*x[0]+(-0.523159705965)*x[1]**o+(-0.82887407638)*x[1]+(0.874288116843)*x[2]**o+(0.520024515804)*x[2]
        ref[0,0,0,0]=(0.507422888369)*x_ref[0]**o+(0.808798513305)*x_ref[0]+(0.51493073805)*x_ref[1]**o+(-0.928527686947)*x_ref[1]+(-0.788010447498)*x_ref[2]**o+(-0.378389833131)*x_ref[2]
        ref[0,0,0,1]=(-0.475851243177)*x_ref[0]**o+(0.746578446988)*x_ref[0]+(0.504778622494)*x_ref[1]**o+(-0.677770398899)*x_ref[1]+(-0.862535648403)*x_ref[2]**o+(0.620414141691)*x_ref[2]
        ref[0,0,1,0]=(-0.361726879582)*x_ref[0]**o+(0.554584127587)*x_ref[0]+(0.436610414772)*x_ref[1]**o+(-0.686534492717)*x_ref[1]+(0.757226061419)*x_ref[2]**o+(0.0948537725389)*x_ref[2]
        ref[0,0,1,1]=(-0.362160696355)*x_ref[0]**o+(0.766524051703)*x_ref[0]+(0.209980857795)*x_ref[1]**o+(-0.478398061437)*x_ref[1]+(-0.410960179122)*x_ref[2]**o+(-0.213272138462)*x_ref[2]
        ref[0,0,2,0]=(0.87154872554)*x_ref[0]**o+(-0.727915301904)*x_ref[0]+(0.683580698056)*x_ref[1]**o+(0.215052521277)*x_ref[1]+(0.267231507898)*x_ref[2]**o+(-0.0124189912449)*x_ref[2]
        ref[0,0,2,1]=(0.631400944876)*x_ref[0]**o+(0.537293628606)*x_ref[0]+(0.944754350994)*x_ref[1]**o+(0.108068908012)*x_ref[1]+(0.890117096365)*x_ref[2]**o+(0.164706867235)*x_ref[2]
        ref[0,1,0,0]=(-0.935871759207)*x_ref[0]**o+(0.350133758077)*x_ref[0]+(-0.966482181015)*x_ref[1]**o+(0.440902701528)*x_ref[1]+(-0.893329541322)*x_ref[2]**o+(0.350317827938)*x_ref[2]
        ref[0,1,0,1]=(0.465305203601)*x_ref[0]**o+(0.711131797928)*x_ref[0]+(0.104091536498)*x_ref[1]**o+(0.484811628127)*x_ref[1]+(0.676878954516)*x_ref[2]**o+(-0.99327016995)*x_ref[2]
        ref[0,1,1,0]=(-0.361189016151)*x_ref[0]**o+(0.962468573492)*x_ref[0]+(-0.158571122564)*x_ref[1]**o+(-0.293583660912)*x_ref[1]+(-0.713699173608)*x_ref[2]**o+(-0.507430163361)*x_ref[2]
        ref[0,1,1,1]=(0.172564527054)*x_ref[0]**o+(0.273438145027)*x_ref[0]+(0.137378266239)*x_ref[1]**o+(-0.675788606841)*x_ref[1]+(-0.362143416121)*x_ref[2]**o+(0.0703897749775)*x_ref[2]
        ref[0,1,2,0]=(-0.941374717271)*x_ref[0]**o+(0.289433165458)*x_ref[0]+(0.318615227028)*x_ref[1]**o+(0.524362568204)*x_ref[1]+(-0.612590305095)*x_ref[2]**o+(-0.117555538957)*x_ref[2]
        ref[0,1,2,1]=(-0.442920825551)*x_ref[0]**o+(-0.454982146579)*x_ref[0]+(0.355353737589)*x_ref[1]**o+(-0.729623917004)*x_ref[1]+(-0.567220335937)*x_ref[2]**o+(0.977652639976)*x_ref[2]
        ref[0,2,0,0]=(0.942485930373)*x_ref[0]**o+(-0.605823961033)*x_ref[0]+(-0.143465779338)*x_ref[1]**o+(0.327354308987)*x_ref[1]+(0.691900417542)*x_ref[2]**o+(-0.514165213849)*x_ref[2]
        ref[0,2,0,1]=(0.018029272977)*x_ref[0]**o+(-0.209992705469)*x_ref[0]+(-0.224614727554)*x_ref[1]**o+(-0.862697462048)*x_ref[1]+(0.274791387928)*x_ref[2]**o+(-0.907445869506)*x_ref[2]
        ref[0,2,1,0]=(0.862755248265)*x_ref[0]**o+(0.020831641939)*x_ref[0]+(0.496662599261)*x_ref[1]**o+(0.79318975953)*x_ref[1]+(0.0912166895541)*x_ref[2]**o+(0.556337610316)*x_ref[2]
        ref[0,2,1,1]=(-0.09652997071)*x_ref[0]**o+(-0.954279421394)*x_ref[0]+(-0.601053524018)*x_ref[1]**o+(0.642528012068)*x_ref[1]+(0.295748113888)*x_ref[2]**o+(-0.318166346186)*x_ref[2]
        ref[0,2,2,0]=(-0.914878854129)*x_ref[0]**o+(0.159201123642)*x_ref[0]+(0.589819056368)*x_ref[1]**o+(-0.715763226823)*x_ref[1]+(-0.444572583567)*x_ref[2]**o+(-0.487866651367)*x_ref[2]
        ref[0,2,2,1]=(-0.71271969175)*x_ref[0]**o+(0.454616308594)*x_ref[0]+(0.510270095021)*x_ref[1]**o+(-0.0385822967955)*x_ref[1]+(-0.519736540313)*x_ref[2]**o+(0.0556733091556)*x_ref[2]
        ref[0,3,0,0]=(0.355768034846)*x_ref[0]**o+(0.152240594627)*x_ref[0]+(-0.81081131843)*x_ref[1]**o+(-0.913373483156)*x_ref[1]+(-0.319859744799)*x_ref[2]**o+(-0.306218435393)*x_ref[2]
        ref[0,3,0,1]=(-0.656932063693)*x_ref[0]**o+(0.832886861783)*x_ref[0]+(-0.417112750683)*x_ref[1]**o+(-0.388190265957)*x_ref[1]+(0.246534947096)*x_ref[2]**o+(-0.896822839189)*x_ref[2]
        ref[0,3,1,0]=(0.204713418891)*x_ref[0]**o+(-0.150097488817)*x_ref[0]+(0.806578426941)*x_ref[1]**o+(-0.590052708995)*x_ref[1]+(-0.758958608419)*x_ref[2]**o+(-0.221973528232)*x_ref[2]
        ref[0,3,1,1]=(-0.854308184618)*x_ref[0]**o+(-0.292526203646)*x_ref[0]+(-0.662403504833)*x_ref[1]**o+(0.0955657982227)*x_ref[1]+(0.69471076173)*x_ref[2]**o+(0.384077309264)*x_ref[2]
        ref[0,3,2,0]=(0.982057200864)*x_ref[0]**o+(-0.64758727704)*x_ref[0]+(0.15941493222)*x_ref[1]**o+(-0.17745437115)*x_ref[1]+(-0.72953252955)*x_ref[2]**o+(0.986979270802)*x_ref[2]
        ref[0,3,2,1]=(0.764526698779)*x_ref[0]**o+(0.22809521115)*x_ref[0]+(-0.0762092087998)*x_ref[1]**o+(0.415212410471)*x_ref[1]+(-0.263638988463)*x_ref[2]**o+(-0.934669942336)*x_ref[2]
        ref[0,4,0,0]=(0.0812656932707)*x_ref[0]**o+(0.0397235062518)*x_ref[0]+(-0.0496912823983)*x_ref[1]**o+(0.189891585189)*x_ref[1]+(0.243537823293)*x_ref[2]**o+(0.461256134208)*x_ref[2]
        ref[0,4,0,1]=(0.167596358861)*x_ref[0]**o+(0.441650690481)*x_ref[0]+(0.638100632205)*x_ref[1]**o+(-0.69240615765)*x_ref[1]+(-0.501855369512)*x_ref[2]**o+(0.769465683782)*x_ref[2]
        ref[0,4,1,0]=(-0.137149398927)*x_ref[0]**o+(-0.186548325285)*x_ref[0]+(-0.504656948352)*x_ref[1]**o+(0.228896656134)*x_ref[1]+(-0.472425358061)*x_ref[2]**o+(-0.0769834754996)*x_ref[2]
        ref[0,4,1,1]=(0.311655648626)*x_ref[0]**o+(0.856513683959)*x_ref[0]+(0.47173626537)*x_ref[1]**o+(0.245497333467)*x_ref[1]+(-0.513727691305)*x_ref[2]**o+(-0.588614785682)*x_ref[2]
        ref[0,4,2,0]=(-0.152907854625)*x_ref[0]**o+(-0.86229950575)*x_ref[0]+(-0.931356414564)*x_ref[1]**o+(-0.915667991864)*x_ref[1]+(0.322316118527)*x_ref[2]**o+(-0.987559462539)*x_ref[2]
        ref[0,4,2,1]=(-0.705947769164)*x_ref[0]**o+(0.104081366627)*x_ref[0]+(-0.540454644392)*x_ref[1]**o+(0.404522880836)*x_ref[1]+(0.146464264141)*x_ref[2]**o+(0.586755030006)*x_ref[2]
        ref[1,0,0,0]=(-0.466614768106)*x_ref[0]**o+(0.482681414943)*x_ref[0]+(-0.176553738108)*x_ref[1]**o+(-0.39162272813)*x_ref[1]+(0.74312023324)*x_ref[2]**o+(0.824538905894)*x_ref[2]
        ref[1,0,0,1]=(-0.9351073631)*x_ref[0]**o+(-0.640226640449)*x_ref[0]+(0.966208762054)*x_ref[1]**o+(0.101792167944)*x_ref[1]+(-0.732983657038)*x_ref[2]**o+(-0.112313375937)*x_ref[2]
        ref[1,0,1,0]=(-0.305403470434)*x_ref[0]**o+(-0.414397930347)*x_ref[0]+(0.356093254322)*x_ref[1]**o+(-0.52607560749)*x_ref[1]+(-0.70070674473)*x_ref[2]**o+(-0.0192901616015)*x_ref[2]
        ref[1,0,1,1]=(0.450046989211)*x_ref[0]**o+(0.602770057268)*x_ref[0]+(0.0193812509113)*x_ref[1]**o+(-0.275615149191)*x_ref[1]+(0.0180468267836)*x_ref[2]**o+(-0.81388255113)*x_ref[2]
        ref[1,0,2,0]=(0.913651483249)*x_ref[0]**o+(-0.92188313743)*x_ref[0]+(0.680007150195)*x_ref[1]**o+(-0.300145584708)*x_ref[1]+(-0.398883520002)*x_ref[2]**o+(-0.752781496463)*x_ref[2]
        ref[1,0,2,1]=(-0.265749827845)*x_ref[0]**o+(0.594379728352)*x_ref[0]+(0.463409943587)*x_ref[1]**o+(-0.85840420426)*x_ref[1]+(-0.705963756674)*x_ref[2]**o+(0.251711112887)*x_ref[2]
        ref[1,1,0,0]=(-0.28895690781)*x_ref[0]**o+(0.476461656166)*x_ref[0]+(0.0692567954597)*x_ref[1]**o+(0.88763437878)*x_ref[1]+(0.0872256197554)*x_ref[2]**o+(0.19467039694)*x_ref[2]
        ref[1,1,0,1]=(0.741045418981)*x_ref[0]**o+(-0.726935151332)*x_ref[0]+(0.899418662131)*x_ref[1]**o+(0.27516497286)*x_ref[1]+(-0.715927474971)*x_ref[2]**o+(-0.709283098732)*x_ref[2]
        ref[1,1,1,0]=(0.991407130174)*x_ref[0]**o+(-0.163484265852)*x_ref[0]+(-0.983254548659)*x_ref[1]**o+(0.0568443855477)*x_ref[1]+(0.309664644588)*x_ref[2]**o+(0.904334718845)*x_ref[2]
        ref[1,1,1,1]=(0.237261672833)*x_ref[0]**o+(0.0861047746345)*x_ref[0]+(-0.494898018662)*x_ref[1]**o+(-0.433075334646)*x_ref[1]+(0.588597346505)*x_ref[2]**o+(-0.185559567553)*x_ref[2]
        ref[1,1,2,0]=(-0.969384091709)*x_ref[0]**o+(-0.0229777223491)*x_ref[0]+(0.0945623059666)*x_ref[1]**o+(0.303543647034)*x_ref[1]+(0.659000135381)*x_ref[2]**o+(0.00873791711254)*x_ref[2]
        ref[1,1,2,1]=(-0.421406731882)*x_ref[0]**o+(-0.5017394377)*x_ref[0]+(0.027747130694)*x_ref[1]**o+(0.327278873996)*x_ref[1]+(-0.754133439304)*x_ref[2]**o+(0.673525803845)*x_ref[2]
        ref[1,2,0,0]=(-0.104620319685)*x_ref[0]**o+(-0.685602393278)*x_ref[0]+(0.972173717014)*x_ref[1]**o+(-0.440573017063)*x_ref[1]+(-0.709529395743)*x_ref[2]**o+(0.634170741457)*x_ref[2]
        ref[1,2,0,1]=(-0.560386754264)*x_ref[0]**o+(0.570894623654)*x_ref[0]+(-0.455122833362)*x_ref[1]**o+(-0.00322505937935)*x_ref[1]+(0.959400336149)*x_ref[2]**o+(0.64014859954)*x_ref[2]
        ref[1,2,1,0]=(0.100585233221)*x_ref[0]**o+(0.939532360957)*x_ref[0]+(-0.900301955112)*x_ref[1]**o+(-0.776375896786)*x_ref[1]+(-0.779373294746)*x_ref[2]**o+(0.546157461116)*x_ref[2]
        ref[1,2,1,1]=(-0.680457315071)*x_ref[0]**o+(-0.0278844539208)*x_ref[0]+(-0.629673590044)*x_ref[1]**o+(0.813218376348)*x_ref[1]+(-0.449484131841)*x_ref[2]**o+(0.641682150048)*x_ref[2]
        ref[1,2,2,0]=(0.60056954386)*x_ref[0]**o+(-0.763169434657)*x_ref[0]+(-0.672890915035)*x_ref[1]**o+(-0.368711154571)*x_ref[1]+(0.476810195318)*x_ref[2]**o+(-0.935071124382)*x_ref[2]
        ref[1,2,2,1]=(0.994100148333)*x_ref[0]**o+(0.471647033381)*x_ref[0]+(0.341677296654)*x_ref[1]**o+(0.466442262068)*x_ref[1]+(-0.0634919249117)*x_ref[2]**o+(0.0637348830912)*x_ref[2]
        ref[1,3,0,0]=(-0.120744286726)*x_ref[0]**o+(0.630378984537)*x_ref[0]+(0.0837817796507)*x_ref[1]**o+(-0.850115274881)*x_ref[1]+(0.574342626561)*x_ref[2]**o+(-0.705409915626)*x_ref[2]
        ref[1,3,0,1]=(-0.51181061391)*x_ref[0]**o+(0.711479747616)*x_ref[0]+(-0.00800928076734)*x_ref[1]**o+(-0.703424580277)*x_ref[1]+(-0.925935026157)*x_ref[2]**o+(0.488175992827)*x_ref[2]
        ref[1,3,1,0]=(0.531879615751)*x_ref[0]**o+(-0.214270521934)*x_ref[0]+(-0.432344726015)*x_ref[1]**o+(-0.107660826449)*x_ref[1]+(0.763406427398)*x_ref[2]**o+(-0.511000150429)*x_ref[2]
        ref[1,3,1,1]=(-0.867467568206)*x_ref[0]**o+(-0.738425961596)*x_ref[0]+(-0.925337667706)*x_ref[1]**o+(-0.586227765107)*x_ref[1]+(0.909987211462)*x_ref[2]**o+(0.533347022582)*x_ref[2]
        ref[1,3,2,0]=(-0.47713113862)*x_ref[0]**o+(-0.130876138883)*x_ref[0]+(-0.607996835981)*x_ref[1]**o+(-0.0306261551172)*x_ref[1]+(0.956207500475)*x_ref[2]**o+(-0.287731719382)*x_ref[2]
        ref[1,3,2,1]=(0.179479486943)*x_ref[0]**o+(0.361266562479)*x_ref[0]+(0.374904472267)*x_ref[1]**o+(0.00788379106301)*x_ref[1]+(-0.34090684265)*x_ref[2]**o+(-0.190875910791)*x_ref[2]
        ref[1,4,0,0]=(0.685464203729)*x_ref[0]**o+(0.0634561930987)*x_ref[0]+(-0.662534110385)*x_ref[1]**o+(0.383849898086)*x_ref[1]+(0.948230524205)*x_ref[2]**o+(-0.347042281976)*x_ref[2]
        ref[1,4,0,1]=(-0.484499755322)*x_ref[0]**o+(-0.765655349243)*x_ref[0]+(0.805331514606)*x_ref[1]**o+(-0.0635912547496)*x_ref[1]+(-0.139457134772)*x_ref[2]**o+(0.227446734221)*x_ref[2]
        ref[1,4,1,0]=(-0.913971229647)*x_ref[0]**o+(0.042531889985)*x_ref[0]+(0.143956410289)*x_ref[1]**o+(0.597990264168)*x_ref[1]+(-0.85926903006)*x_ref[2]**o+(0.205900361792)*x_ref[2]
        ref[1,4,1,1]=(-0.955540264653)*x_ref[0]**o+(0.983439073814)*x_ref[0]+(0.246870351555)*x_ref[1]**o+(-0.610917180843)*x_ref[1]+(-0.0418134447092)*x_ref[2]**o+(-0.58407787472)*x_ref[2]
        ref[1,4,2,0]=(0.787274711453)*x_ref[0]**o+(-0.899651339467)*x_ref[0]+(-0.62195254225)*x_ref[1]**o+(-0.516447289251)*x_ref[1]+(0.267082643601)*x_ref[2]**o+(0.0408186039688)*x_ref[2]
        ref[1,4,2,1]=(-0.485773801895)*x_ref[0]**o+(0.663478067655)*x_ref[0]+(0.928973237869)*x_ref[1]**o+(0.353735145156)*x_ref[1]+(0.128409748978)*x_ref[2]**o+(-0.476479648592)*x_ref[2]
        ref[2,0,0,0]=(0.491574821811)*x_ref[0]**o+(0.455303820526)*x_ref[0]+(-0.747773112075)*x_ref[1]**o+(-0.653255987137)*x_ref[1]+(-0.738293040346)*x_ref[2]**o+(0.948070477272)*x_ref[2]
        ref[2,0,0,1]=(-0.697412782403)*x_ref[0]**o+(-0.738389544586)*x_ref[0]+(0.231728579935)*x_ref[1]**o+(0.248503931032)*x_ref[1]+(-0.561155590703)*x_ref[2]**o+(0.324682113147)*x_ref[2]
        ref[2,0,1,0]=(-0.00507638382064)*x_ref[0]**o+(-0.983105866448)*x_ref[0]+(-0.253141627484)*x_ref[1]**o+(-0.0664178091069)*x_ref[1]+(0.194460220144)*x_ref[2]**o+(-0.174257469198)*x_ref[2]
        ref[2,0,1,1]=(-0.644447560463)*x_ref[0]**o+(0.658703761133)*x_ref[0]+(-0.685392825104)*x_ref[1]**o+(-0.766295260796)*x_ref[1]+(-0.922004388493)*x_ref[2]**o+(-0.0745707720923)*x_ref[2]
        ref[2,0,2,0]=(-0.69924009606)*x_ref[0]**o+(-0.398212597328)*x_ref[0]+(-0.364991550186)*x_ref[1]**o+(0.243474288977)*x_ref[1]+(0.129545162767)*x_ref[2]**o+(0.920611954845)*x_ref[2]
        ref[2,0,2,1]=(-0.494476126875)*x_ref[0]**o+(-0.131503103148)*x_ref[0]+(-0.241670500874)*x_ref[1]**o+(0.40439125905)*x_ref[1]+(0.134166779399)*x_ref[2]**o+(-0.507385751809)*x_ref[2]
        ref[2,1,0,0]=(0.301148115913)*x_ref[0]**o+(0.189635133395)*x_ref[0]+(0.744167404774)*x_ref[1]**o+(0.995754344203)*x_ref[1]+(-0.334760662218)*x_ref[2]**o+(-0.182462923741)*x_ref[2]
        ref[2,1,0,1]=(0.523500022508)*x_ref[0]**o+(-0.257042661558)*x_ref[0]+(-0.940233100381)*x_ref[1]**o+(-0.202431731891)*x_ref[1]+(-0.653480190287)*x_ref[2]**o+(0.993156915081)*x_ref[2]
        ref[2,1,1,0]=(-0.502452234661)*x_ref[0]**o+(-0.711581805303)*x_ref[0]+(0.696450091152)*x_ref[1]**o+(0.436746953802)*x_ref[1]+(0.695969418438)*x_ref[2]**o+(-0.671400022595)*x_ref[2]
        ref[2,1,1,1]=(0.0705944125182)*x_ref[0]**o+(-0.865755834485)*x_ref[0]+(0.921555996641)*x_ref[1]**o+(-0.567855814858)*x_ref[1]+(-0.213455192388)*x_ref[2]**o+(0.820868399259)*x_ref[2]
        ref[2,1,2,0]=(0.744232126707)*x_ref[0]**o+(0.336969176485)*x_ref[0]+(0.339407729288)*x_ref[1]**o+(-0.465154968917)*x_ref[1]+(0.389524167202)*x_ref[2]**o+(0.645394954477)*x_ref[2]
        ref[2,1,2,1]=(-0.299085687516)*x_ref[0]**o+(0.220795864743)*x_ref[0]+(-0.19353913538)*x_ref[1]**o+(0.445131396297)*x_ref[1]+(-0.779605318636)*x_ref[2]**o+(-0.76707264371)*x_ref[2]
        ref[2,2,0,0]=(-0.357585997171)*x_ref[0]**o+(-0.248795296406)*x_ref[0]+(0.0876364714478)*x_ref[1]**o+(-0.682833482179)*x_ref[1]+(0.497052575986)*x_ref[2]**o+(0.893378007889)*x_ref[2]
        ref[2,2,0,1]=(0.38067999218)*x_ref[0]**o+(-0.857246232026)*x_ref[0]+(0.0785818284359)*x_ref[1]**o+(-0.473287028123)*x_ref[1]+(-0.763041318956)*x_ref[2]**o+(-0.1119710251)*x_ref[2]
        ref[2,2,1,0]=(-0.0278856926369)*x_ref[0]**o+(0.435955626553)*x_ref[0]+(-0.676274741782)*x_ref[1]**o+(-0.719721825467)*x_ref[1]+(0.195007734703)*x_ref[2]**o+(0.504771278902)*x_ref[2]
        ref[2,2,1,1]=(0.459547475772)*x_ref[0]**o+(-0.669041681613)*x_ref[0]+(-0.741957874364)*x_ref[1]**o+(0.604585337407)*x_ref[1]+(-0.886718742046)*x_ref[2]**o+(0.338113237761)*x_ref[2]
        ref[2,2,2,0]=(0.392926733678)*x_ref[0]**o+(-0.923680589861)*x_ref[0]+(-0.425503328457)*x_ref[1]**o+(-0.0902669250117)*x_ref[1]+(-0.346726072703)*x_ref[2]**o+(0.541482976709)*x_ref[2]
        ref[2,2,2,1]=(0.134867023616)*x_ref[0]**o+(0.478349607348)*x_ref[0]+(-0.334617836366)*x_ref[1]**o+(-0.74587411744)*x_ref[1]+(-0.880365937567)*x_ref[2]**o+(0.820743064452)*x_ref[2]
        ref[2,3,0,0]=(0.081184173895)*x_ref[0]**o+(0.315011929905)*x_ref[0]+(-0.137083926616)*x_ref[1]**o+(0.103926223299)*x_ref[1]+(0.363270989968)*x_ref[2]**o+(-0.454075563606)*x_ref[2]
        ref[2,3,0,1]=(0.457730701321)*x_ref[0]**o+(0.922345625691)*x_ref[0]+(-0.412933933474)*x_ref[1]**o+(-0.546359655103)*x_ref[1]+(0.719101029933)*x_ref[2]**o+(-0.790522859322)*x_ref[2]
        ref[2,3,1,0]=(-0.811816419841)*x_ref[0]**o+(0.0379298408788)*x_ref[0]+(-0.852643125131)*x_ref[1]**o+(0.76881274254)*x_ref[1]+(0.75214553496)*x_ref[2]**o+(0.800382814403)*x_ref[2]
        ref[2,3,1,1]=(0.524802143618)*x_ref[0]**o+(-0.717457078371)*x_ref[0]+(0.389026887965)*x_ref[1]**o+(-0.231912558721)*x_ref[1]+(-0.394705562333)*x_ref[2]**o+(-0.794723802655)*x_ref[2]
        ref[2,3,2,0]=(-0.232215520303)*x_ref[0]**o+(0.994932250617)*x_ref[0]+(-0.760425274222)*x_ref[1]**o+(0.0910727760744)*x_ref[1]+(0.621642024132)*x_ref[2]**o+(0.263978719304)*x_ref[2]
        ref[2,3,2,1]=(0.515269404424)*x_ref[0]**o+(-0.998266958288)*x_ref[0]+(-0.485078033534)*x_ref[1]**o+(-0.960043247553)*x_ref[1]+(-0.926562413314)*x_ref[2]**o+(0.612710762712)*x_ref[2]
        ref[2,4,0,0]=(0.419808446591)*x_ref[0]**o+(0.132792190191)*x_ref[0]+(-0.690603562736)*x_ref[1]**o+(0.45137757885)*x_ref[1]+(-0.262249041567)*x_ref[2]**o+(-0.432995254783)*x_ref[2]
        ref[2,4,0,1]=(0.62413323935)*x_ref[0]**o+(0.963559148607)*x_ref[0]+(-0.879044261015)*x_ref[1]**o+(-0.504981574957)*x_ref[1]+(-0.575204029973)*x_ref[2]**o+(0.066855791448)*x_ref[2]
        ref[2,4,1,0]=(0.321218662791)*x_ref[0]**o+(0.975103526588)*x_ref[0]+(0.783716185844)*x_ref[1]**o+(-0.816368139389)*x_ref[1]+(0.778169798128)*x_ref[2]**o+(-0.527443068085)*x_ref[2]
        ref[2,4,1,1]=(-0.8801082648)*x_ref[0]**o+(-0.0313845929272)*x_ref[0]+(0.485262237838)*x_ref[1]**o+(0.0778888535573)*x_ref[1]+(-0.900948214929)*x_ref[2]**o+(-0.98830300089)*x_ref[2]
        ref[2,4,2,0]=(-0.996395042226)*x_ref[0]**o+(0.253635852642)*x_ref[0]+(0.180965173659)*x_ref[1]**o+(-0.488294662459)*x_ref[1]+(-0.144583511741)*x_ref[2]**o+(0.285854081388)*x_ref[2]
        ref[2,4,2,1]=(0.00988506292061)*x_ref[0]**o+(-0.00192786315452)*x_ref[0]+(0.720775470647)*x_ref[1]**o+(-0.587673311506)*x_ref[1]+(0.313836238675)*x_ref[2]**o+(0.45975125457)*x_ref[2]
        ref[3,0,0,0]=(-0.822049489749)*x_ref[0]**o+(-0.989589435175)*x_ref[0]+(-0.0943830434118)*x_ref[1]**o+(-0.789229900035)*x_ref[1]+(0.311382619317)*x_ref[2]**o+(-0.316646938685)*x_ref[2]
        ref[3,0,0,1]=(0.501517916677)*x_ref[0]**o+(-0.704502365436)*x_ref[0]+(0.913184728006)*x_ref[1]**o+(-0.843712848791)*x_ref[1]+(0.901479696878)*x_ref[2]**o+(-0.713893336542)*x_ref[2]
        ref[3,0,1,0]=(0.0771592811086)*x_ref[0]**o+(0.796038032206)*x_ref[0]+(0.566256212871)*x_ref[1]**o+(0.394136659986)*x_ref[1]+(-0.794422065388)*x_ref[2]**o+(0.662174992848)*x_ref[2]
        ref[3,0,1,1]=(0.180225115167)*x_ref[0]**o+(0.366965867971)*x_ref[0]+(-0.244015756153)*x_ref[1]**o+(0.433571067335)*x_ref[1]+(0.129075150185)*x_ref[2]**o+(-0.216304414605)*x_ref[2]
        ref[3,0,2,0]=(0.100277967034)*x_ref[0]**o+(-0.590583546786)*x_ref[0]+(-0.595553697152)*x_ref[1]**o+(0.923573252239)*x_ref[1]+(0.578966663184)*x_ref[2]**o+(-0.411082913097)*x_ref[2]
        ref[3,0,2,1]=(-0.575315696811)*x_ref[0]**o+(0.608760285918)*x_ref[0]+(-0.3592074605)*x_ref[1]**o+(0.515119945308)*x_ref[1]+(-0.4000138306)*x_ref[2]**o+(0.477685324937)*x_ref[2]
        ref[3,1,0,0]=(0.0780014276919)*x_ref[0]**o+(-0.0932830671471)*x_ref[0]+(0.39477236642)*x_ref[1]**o+(0.406673135556)*x_ref[1]+(-0.235480707568)*x_ref[2]**o+(-0.872302257144)*x_ref[2]
        ref[3,1,0,1]=(0.170573072646)*x_ref[0]**o+(0.868146561326)*x_ref[0]+(0.262136343896)*x_ref[1]**o+(0.192343529912)*x_ref[1]+(0.157400965693)*x_ref[2]**o+(-0.932981619473)*x_ref[2]
        ref[3,1,1,0]=(0.891433179291)*x_ref[0]**o+(-0.281396449997)*x_ref[0]+(-0.0803173619772)*x_ref[1]**o+(-0.378460054356)*x_ref[1]+(-0.581578656014)*x_ref[2]**o+(-0.65856067509)*x_ref[2]
        ref[3,1,1,1]=(0.343599956626)*x_ref[0]**o+(-0.92063694125)*x_ref[0]+(0.440698102072)*x_ref[1]**o+(0.767124166388)*x_ref[1]+(0.050964317543)*x_ref[2]**o+(1.35605562239e-05)*x_ref[2]
        ref[3,1,2,0]=(-0.614268150987)*x_ref[0]**o+(0.363806471556)*x_ref[0]+(0.92193457905)*x_ref[1]**o+(0.157919580642)*x_ref[1]+(-0.955575321124)*x_ref[2]**o+(-0.984808437666)*x_ref[2]
        ref[3,1,2,1]=(0.030099648966)*x_ref[0]**o+(-0.375999369491)*x_ref[0]+(-0.0992643926008)*x_ref[1]**o+(0.109209161093)*x_ref[1]+(-0.974043780088)*x_ref[2]**o+(0.908049658745)*x_ref[2]
        ref[3,2,0,0]=(0.297771341106)*x_ref[0]**o+(-0.637802228169)*x_ref[0]+(0.398100441783)*x_ref[1]**o+(-0.0643425181651)*x_ref[1]+(0.662775339033)*x_ref[2]**o+(0.744835182052)*x_ref[2]
        ref[3,2,0,1]=(0.312026420583)*x_ref[0]**o+(0.865855052574)*x_ref[0]+(0.0195001783165)*x_ref[1]**o+(0.537899238473)*x_ref[1]+(0.554893357481)*x_ref[2]**o+(0.71555610145)*x_ref[2]
        ref[3,2,1,0]=(-0.199138146186)*x_ref[0]**o+(0.827546489855)*x_ref[0]+(-0.480965464736)*x_ref[1]**o+(-0.473796093187)*x_ref[1]+(0.880264675049)*x_ref[2]**o+(0.544506670844)*x_ref[2]
        ref[3,2,1,1]=(0.773798005008)*x_ref[0]**o+(0.744945241634)*x_ref[0]+(0.548170564608)*x_ref[1]**o+(-0.560503692823)*x_ref[1]+(-0.435630215427)*x_ref[2]**o+(0.080015730707)*x_ref[2]
        ref[3,2,2,0]=(0.36735714937)*x_ref[0]**o+(-0.867071557873)*x_ref[0]+(0.936135650431)*x_ref[1]**o+(-0.432533600524)*x_ref[1]+(-0.0224430204069)*x_ref[2]**o+(0.218590302621)*x_ref[2]
        ref[3,2,2,1]=(0.932808382499)*x_ref[0]**o+(-0.0970016713319)*x_ref[0]+(0.458943467579)*x_ref[1]**o+(-0.134338789898)*x_ref[1]+(-0.426892557314)*x_ref[2]**o+(0.627071826863)*x_ref[2]
        ref[3,3,0,0]=(-0.623804149075)*x_ref[0]**o+(0.910423590172)*x_ref[0]+(0.465938181097)*x_ref[1]**o+(0.495479245391)*x_ref[1]+(0.693745442185)*x_ref[2]**o+(-0.437014226757)*x_ref[2]
        ref[3,3,0,1]=(0.598807017496)*x_ref[0]**o+(-0.112361840211)*x_ref[0]+(-0.550899685213)*x_ref[1]**o+(-0.0478222019564)*x_ref[1]+(-0.0110129517693)*x_ref[2]**o+(-0.059534786374)*x_ref[2]
        ref[3,3,1,0]=(-0.561155633409)*x_ref[0]**o+(-0.463563057636)*x_ref[0]+(-0.170463996009)*x_ref[1]**o+(-0.214509204942)*x_ref[1]+(0.851200198666)*x_ref[2]**o+(-0.721036329689)*x_ref[2]
        ref[3,3,1,1]=(0.608666240405)*x_ref[0]**o+(0.814232455115)*x_ref[0]+(-0.604645629689)*x_ref[1]**o+(0.960976509673)*x_ref[1]+(0.365807996219)*x_ref[2]**o+(-0.87491909238)*x_ref[2]
        ref[3,3,2,0]=(0.120126988418)*x_ref[0]**o+(-0.745687314825)*x_ref[0]+(-0.332697826984)*x_ref[1]**o+(-0.823683406516)*x_ref[1]+(0.305444079613)*x_ref[2]**o+(0.901972115441)*x_ref[2]
        ref[3,3,2,1]=(0.0470293848904)*x_ref[0]**o+(-0.195496754814)*x_ref[0]+(0.671870114105)*x_ref[1]**o+(0.170273146371)*x_ref[1]+(-0.851927986188)*x_ref[2]**o+(0.772192863973)*x_ref[2]
        ref[3,4,0,0]=(-0.95392341727)*x_ref[0]**o+(-0.665694215521)*x_ref[0]+(-0.0247396499378)*x_ref[1]**o+(0.441352857387)*x_ref[1]+(0.767156547791)*x_ref[2]**o+(-0.14696372018)*x_ref[2]
        ref[3,4,0,1]=(0.209761166815)*x_ref[0]**o+(0.978934473425)*x_ref[0]+(0.943806891053)*x_ref[1]**o+(-0.756619917601)*x_ref[1]+(0.635699949764)*x_ref[2]**o+(-0.345248269568)*x_ref[2]
        ref[3,4,1,0]=(0.748020785785)*x_ref[0]**o+(0.311104549542)*x_ref[0]+(0.739127551248)*x_ref[1]**o+(-0.635238575859)*x_ref[1]+(0.623898875653)*x_ref[2]**o+(0.343407392068)*x_ref[2]
        ref[3,4,1,1]=(-0.422500892471)*x_ref[0]**o+(0.0688495013346)*x_ref[0]+(-0.919653906232)*x_ref[1]**o+(0.531036586396)*x_ref[1]+(-0.22487910672)*x_ref[2]**o+(-0.848225994541)*x_ref[2]
        ref[3,4,2,0]=(0.723965744673)*x_ref[0]**o+(0.558602959224)*x_ref[0]+(0.699539243308)*x_ref[1]**o+(-0.597426541715)*x_ref[1]+(-0.0128675070522)*x_ref[2]**o+(-0.874796626242)*x_ref[2]
        ref[3,4,2,1]=(-0.962970021216)*x_ref[0]**o+(0.532886481874)*x_ref[0]+(-0.523159705965)*x_ref[1]**o+(-0.82887407638)*x_ref[1]+(0.874288116843)*x_ref[2]**o+(0.520024515804)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.0603413935246)*x[0]**o+(-0.70184894912)*x[0]+(-0.206341392562)*x[1]**o+(0.892585857385)*x[1]
        ref=(-0.0603413935246)*x_ref[0]**o+(-0.70184894912)*x_ref[0]+(-0.206341392562)*x_ref[1]**o+(0.892585857385)*x_ref[1]
      else:
        arg=(0.120763612793)*x[0]**o+(-0.298506957699)*x[0]+(0.0620633729217)*x[1]**o+(0.807000465544)*x[1]+(-0.564443939995)*x[2]**o+(0.820783035273)*x[2]
        ref=(0.120763612793)*x_ref[0]**o+(-0.298506957699)*x_ref[0]+(0.0620633729217)*x_ref[1]**o+(0.807000465544)*x_ref[1]+(-0.564443939995)*x_ref[2]**o+(0.820783035273)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.69193390973)*x[0]**o+(-0.768888031033)*x[0]+(-0.813493496124)*x[1]**o+(-0.767059499773)*x[1]
        arg[1]=(0.111870073987)*x[0]**o+(-0.196406293238)*x[0]+(-0.10683601134)*x[1]**o+(0.69972228684)*x[1]
        ref[0]=(-0.69193390973)*x_ref[0]**o+(-0.768888031033)*x_ref[0]+(-0.813493496124)*x_ref[1]**o+(-0.767059499773)*x_ref[1]
        ref[1]=(0.111870073987)*x_ref[0]**o+(-0.196406293238)*x_ref[0]+(-0.10683601134)*x_ref[1]**o+(0.69972228684)*x_ref[1]
      else:
        arg[0]=(-0.20606732369)*x[0]**o+(0.773916734719)*x[0]+(-0.915594370888)*x[1]**o+(-0.684202475504)*x[1]+(0.161644377631)*x[2]**o+(-0.834568364942)*x[2]
        arg[1]=(-0.475848931399)*x[0]**o+(-0.967325026091)*x[0]+(0.371032251427)*x[1]**o+(0.976933923378)*x[1]+(0.326121409381)*x[2]**o+(0.880666151318)*x[2]
        ref[0]=(-0.20606732369)*x_ref[0]**o+(0.773916734719)*x_ref[0]+(-0.915594370888)*x_ref[1]**o+(-0.684202475504)*x_ref[1]+(0.161644377631)*x_ref[2]**o+(-0.834568364942)*x_ref[2]
        ref[1]=(-0.475848931399)*x_ref[0]**o+(-0.967325026091)*x_ref[0]+(0.371032251427)*x_ref[1]**o+(0.976933923378)*x_ref[1]+(0.326121409381)*x_ref[2]**o+(0.880666151318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.664616598004)*x[0]**o+(0.159791014359)*x[0]+(-0.3747223278)*x[1]**o+(0.302802720584)*x[1]
        arg[0,1]=(-0.334311194265)*x[0]**o+(0.865855840771)*x[0]+(-0.0184933229609)*x[1]**o+(0.589277746683)*x[1]
        arg[0,2]=(0.685266031246)*x[0]**o+(0.846357835994)*x[0]+(0.233422031025)*x[1]**o+(0.391225449794)*x[1]
        arg[0,3]=(-0.889412463645)*x[0]**o+(0.336985584587)*x[0]+(-0.311629085429)*x[1]**o+(-0.574756125055)*x[1]
        arg[0,4]=(-0.786434255596)*x[0]**o+(0.897529643329)*x[0]+(0.112631455142)*x[1]**o+(0.501957817255)*x[1]
        arg[1,0]=(-0.841995669352)*x[0]**o+(-0.372241861224)*x[0]+(0.669429619157)*x[1]**o+(0.913007793035)*x[1]
        arg[1,1]=(-0.965486931849)*x[0]**o+(-0.163330142621)*x[0]+(0.302996628649)*x[1]**o+(-0.214176383433)*x[1]
        arg[1,2]=(-0.717841677595)*x[0]**o+(-0.560453345617)*x[0]+(0.885221610478)*x[1]**o+(0.0925933897521)*x[1]
        arg[1,3]=(0.827380987548)*x[0]**o+(-0.185460343082)*x[0]+(-0.618848726331)*x[1]**o+(-0.25427152262)*x[1]
        arg[1,4]=(-0.232848766689)*x[0]**o+(-0.812427149527)*x[0]+(-0.6328125174)*x[1]**o+(0.377849173478)*x[1]
        arg[2,0]=(0.725568000689)*x[0]**o+(-0.0119648722611)*x[0]+(0.0551179058345)*x[1]**o+(-0.313601095535)*x[1]
        arg[2,1]=(-0.371331280463)*x[0]**o+(0.164812693414)*x[0]+(0.140689113041)*x[1]**o+(0.866863122864)*x[1]
        arg[2,2]=(0.445443472696)*x[0]**o+(0.885420986537)*x[0]+(0.434326737693)*x[1]**o+(0.013677800309)*x[1]
        arg[2,3]=(0.186221146816)*x[0]**o+(0.919506151782)*x[0]+(-0.163957717851)*x[1]**o+(-0.533309231769)*x[1]
        arg[2,4]=(0.137164729623)*x[0]**o+(0.561715777859)*x[0]+(0.33724606372)*x[1]**o+(-0.554531877504)*x[1]
        arg[3,0]=(-0.0632399630146)*x[0]**o+(0.712014652956)*x[0]+(0.107109261787)*x[1]**o+(0.228670262705)*x[1]
        arg[3,1]=(-0.973793135038)*x[0]**o+(0.923465546039)*x[0]+(0.189825544105)*x[1]**o+(-0.33784496447)*x[1]
        arg[3,2]=(-0.12686258826)*x[0]**o+(-0.582653698893)*x[0]+(0.860503162549)*x[1]**o+(-0.477305983221)*x[1]
        arg[3,3]=(-0.407988306429)*x[0]**o+(-0.174396602789)*x[0]+(0.646815890222)*x[1]**o+(-0.450974163079)*x[1]
        arg[3,4]=(0.162458379489)*x[0]**o+(-0.0569354582388)*x[0]+(0.0463216806125)*x[1]**o+(-0.381777259635)*x[1]
        ref[0,0]=(-0.664616598004)*x_ref[0]**o+(0.159791014359)*x_ref[0]+(-0.3747223278)*x_ref[1]**o+(0.302802720584)*x_ref[1]
        ref[0,1]=(-0.334311194265)*x_ref[0]**o+(0.865855840771)*x_ref[0]+(-0.0184933229609)*x_ref[1]**o+(0.589277746683)*x_ref[1]
        ref[0,2]=(0.685266031246)*x_ref[0]**o+(0.846357835994)*x_ref[0]+(0.233422031025)*x_ref[1]**o+(0.391225449794)*x_ref[1]
        ref[0,3]=(-0.889412463645)*x_ref[0]**o+(0.336985584587)*x_ref[0]+(-0.311629085429)*x_ref[1]**o+(-0.574756125055)*x_ref[1]
        ref[0,4]=(-0.786434255596)*x_ref[0]**o+(0.897529643329)*x_ref[0]+(0.112631455142)*x_ref[1]**o+(0.501957817255)*x_ref[1]
        ref[1,0]=(-0.841995669352)*x_ref[0]**o+(-0.372241861224)*x_ref[0]+(0.669429619157)*x_ref[1]**o+(0.913007793035)*x_ref[1]
        ref[1,1]=(-0.965486931849)*x_ref[0]**o+(-0.163330142621)*x_ref[0]+(0.302996628649)*x_ref[1]**o+(-0.214176383433)*x_ref[1]
        ref[1,2]=(-0.717841677595)*x_ref[0]**o+(-0.560453345617)*x_ref[0]+(0.885221610478)*x_ref[1]**o+(0.0925933897521)*x_ref[1]
        ref[1,3]=(0.827380987548)*x_ref[0]**o+(-0.185460343082)*x_ref[0]+(-0.618848726331)*x_ref[1]**o+(-0.25427152262)*x_ref[1]
        ref[1,4]=(-0.232848766689)*x_ref[0]**o+(-0.812427149527)*x_ref[0]+(-0.6328125174)*x_ref[1]**o+(0.377849173478)*x_ref[1]
        ref[2,0]=(0.725568000689)*x_ref[0]**o+(-0.0119648722611)*x_ref[0]+(0.0551179058345)*x_ref[1]**o+(-0.313601095535)*x_ref[1]
        ref[2,1]=(-0.371331280463)*x_ref[0]**o+(0.164812693414)*x_ref[0]+(0.140689113041)*x_ref[1]**o+(0.866863122864)*x_ref[1]
        ref[2,2]=(0.445443472696)*x_ref[0]**o+(0.885420986537)*x_ref[0]+(0.434326737693)*x_ref[1]**o+(0.013677800309)*x_ref[1]
        ref[2,3]=(0.186221146816)*x_ref[0]**o+(0.919506151782)*x_ref[0]+(-0.163957717851)*x_ref[1]**o+(-0.533309231769)*x_ref[1]
        ref[2,4]=(0.137164729623)*x_ref[0]**o+(0.561715777859)*x_ref[0]+(0.33724606372)*x_ref[1]**o+(-0.554531877504)*x_ref[1]
        ref[3,0]=(-0.0632399630146)*x_ref[0]**o+(0.712014652956)*x_ref[0]+(0.107109261787)*x_ref[1]**o+(0.228670262705)*x_ref[1]
        ref[3,1]=(-0.973793135038)*x_ref[0]**o+(0.923465546039)*x_ref[0]+(0.189825544105)*x_ref[1]**o+(-0.33784496447)*x_ref[1]
        ref[3,2]=(-0.12686258826)*x_ref[0]**o+(-0.582653698893)*x_ref[0]+(0.860503162549)*x_ref[1]**o+(-0.477305983221)*x_ref[1]
        ref[3,3]=(-0.407988306429)*x_ref[0]**o+(-0.174396602789)*x_ref[0]+(0.646815890222)*x_ref[1]**o+(-0.450974163079)*x_ref[1]
        ref[3,4]=(0.162458379489)*x_ref[0]**o+(-0.0569354582388)*x_ref[0]+(0.0463216806125)*x_ref[1]**o+(-0.381777259635)*x_ref[1]
      else:
        arg[0,0]=(0.865079887449)*x[0]**o+(-0.833757275547)*x[0]+(-0.967658088049)*x[1]**o+(-0.168314490396)*x[1]+(0.151618201772)*x[2]**o+(-0.251062354355)*x[2]
        arg[0,1]=(0.110207997589)*x[0]**o+(-0.510994295641)*x[0]+(0.855215902178)*x[1]**o+(0.941502315255)*x[1]+(-0.794874945971)*x[2]**o+(-0.309190709938)*x[2]
        arg[0,2]=(0.71701217173)*x[0]**o+(-0.221333357165)*x[0]+(0.689963224026)*x[1]**o+(0.221517236085)*x[1]+(-0.52608140393)*x[2]**o+(0.371488496481)*x[2]
        arg[0,3]=(0.618194041415)*x[0]**o+(0.454387091678)*x[0]+(-0.707329629946)*x[1]**o+(0.41972715681)*x[1]+(0.0180789814415)*x[2]**o+(-0.176487980225)*x[2]
        arg[0,4]=(0.329275243989)*x[0]**o+(-0.679005993002)*x[0]+(-0.261017436418)*x[1]**o+(0.151780687062)*x[1]+(-0.480936330475)*x[2]**o+(-0.697077234587)*x[2]
        arg[1,0]=(0.720913763906)*x[0]**o+(0.283486705277)*x[0]+(0.498289077088)*x[1]**o+(0.657543220863)*x[1]+(-0.666293219679)*x[2]**o+(0.901456618353)*x[2]
        arg[1,1]=(-0.984526350563)*x[0]**o+(0.677118292603)*x[0]+(0.614707818107)*x[1]**o+(0.272173514554)*x[1]+(0.41492971828)*x[2]**o+(0.274511869871)*x[2]
        arg[1,2]=(0.322224985373)*x[0]**o+(0.567272204545)*x[0]+(-0.437313912332)*x[1]**o+(0.847486025896)*x[1]+(0.769434908781)*x[2]**o+(0.325917607497)*x[2]
        arg[1,3]=(0.709357837264)*x[0]**o+(0.927425103914)*x[0]+(-0.622876388501)*x[1]**o+(-0.296141603555)*x[1]+(-0.038448878923)*x[2]**o+(-0.320692580169)*x[2]
        arg[1,4]=(-0.27890333885)*x[0]**o+(0.803346569394)*x[0]+(0.00665934218255)*x[1]**o+(-0.89057979428)*x[1]+(-0.727135553706)*x[2]**o+(0.892946022497)*x[2]
        arg[2,0]=(0.579951394388)*x[0]**o+(0.0406506913385)*x[0]+(0.363944368798)*x[1]**o+(0.582732756951)*x[1]+(0.542548838027)*x[2]**o+(0.365742797343)*x[2]
        arg[2,1]=(0.253260777874)*x[0]**o+(-0.69077350875)*x[0]+(0.458080700492)*x[1]**o+(-0.79224006672)*x[1]+(-0.504865679524)*x[2]**o+(0.681425632809)*x[2]
        arg[2,2]=(0.600427718261)*x[0]**o+(0.91761242151)*x[0]+(0.602316047463)*x[1]**o+(0.890383614185)*x[1]+(-0.386691039525)*x[2]**o+(0.688659949109)*x[2]
        arg[2,3]=(-0.181338810914)*x[0]**o+(0.933744241999)*x[0]+(0.0212774093137)*x[1]**o+(0.284972854884)*x[1]+(-0.150219922419)*x[2]**o+(0.0878011402107)*x[2]
        arg[2,4]=(-0.848985879538)*x[0]**o+(0.298962025531)*x[0]+(0.593300576179)*x[1]**o+(0.301509253541)*x[1]+(0.971171084219)*x[2]**o+(-0.973386482857)*x[2]
        arg[3,0]=(-0.735608389534)*x[0]**o+(0.44902436426)*x[0]+(0.0536456095512)*x[1]**o+(0.208969935548)*x[1]+(-0.787609300211)*x[2]**o+(0.726024812477)*x[2]
        arg[3,1]=(0.200882171943)*x[0]**o+(-0.644461780745)*x[0]+(-0.0995704510264)*x[1]**o+(-0.857730241944)*x[1]+(-0.372302969302)*x[2]**o+(-0.171367954739)*x[2]
        arg[3,2]=(-0.673606332627)*x[0]**o+(-0.828183678442)*x[0]+(-0.236856099365)*x[1]**o+(-0.266339053288)*x[1]+(-0.972034740204)*x[2]**o+(-0.33151750348)*x[2]
        arg[3,3]=(-0.312470720017)*x[0]**o+(-0.657433718629)*x[0]+(-0.984940036612)*x[1]**o+(0.197707053145)*x[1]+(0.512009064643)*x[2]**o+(0.855601033248)*x[2]
        arg[3,4]=(-0.674854508228)*x[0]**o+(-0.394035682831)*x[0]+(-0.749446232511)*x[1]**o+(0.866992763518)*x[1]+(-0.945837361158)*x[2]**o+(0.593490782481)*x[2]
        ref[0,0]=(0.865079887449)*x_ref[0]**o+(-0.833757275547)*x_ref[0]+(-0.967658088049)*x_ref[1]**o+(-0.168314490396)*x_ref[1]+(0.151618201772)*x_ref[2]**o+(-0.251062354355)*x_ref[2]
        ref[0,1]=(0.110207997589)*x_ref[0]**o+(-0.510994295641)*x_ref[0]+(0.855215902178)*x_ref[1]**o+(0.941502315255)*x_ref[1]+(-0.794874945971)*x_ref[2]**o+(-0.309190709938)*x_ref[2]
        ref[0,2]=(0.71701217173)*x_ref[0]**o+(-0.221333357165)*x_ref[0]+(0.689963224026)*x_ref[1]**o+(0.221517236085)*x_ref[1]+(-0.52608140393)*x_ref[2]**o+(0.371488496481)*x_ref[2]
        ref[0,3]=(0.618194041415)*x_ref[0]**o+(0.454387091678)*x_ref[0]+(-0.707329629946)*x_ref[1]**o+(0.41972715681)*x_ref[1]+(0.0180789814415)*x_ref[2]**o+(-0.176487980225)*x_ref[2]
        ref[0,4]=(0.329275243989)*x_ref[0]**o+(-0.679005993002)*x_ref[0]+(-0.261017436418)*x_ref[1]**o+(0.151780687062)*x_ref[1]+(-0.480936330475)*x_ref[2]**o+(-0.697077234587)*x_ref[2]
        ref[1,0]=(0.720913763906)*x_ref[0]**o+(0.283486705277)*x_ref[0]+(0.498289077088)*x_ref[1]**o+(0.657543220863)*x_ref[1]+(-0.666293219679)*x_ref[2]**o+(0.901456618353)*x_ref[2]
        ref[1,1]=(-0.984526350563)*x_ref[0]**o+(0.677118292603)*x_ref[0]+(0.614707818107)*x_ref[1]**o+(0.272173514554)*x_ref[1]+(0.41492971828)*x_ref[2]**o+(0.274511869871)*x_ref[2]
        ref[1,2]=(0.322224985373)*x_ref[0]**o+(0.567272204545)*x_ref[0]+(-0.437313912332)*x_ref[1]**o+(0.847486025896)*x_ref[1]+(0.769434908781)*x_ref[2]**o+(0.325917607497)*x_ref[2]
        ref[1,3]=(0.709357837264)*x_ref[0]**o+(0.927425103914)*x_ref[0]+(-0.622876388501)*x_ref[1]**o+(-0.296141603555)*x_ref[1]+(-0.038448878923)*x_ref[2]**o+(-0.320692580169)*x_ref[2]
        ref[1,4]=(-0.27890333885)*x_ref[0]**o+(0.803346569394)*x_ref[0]+(0.00665934218255)*x_ref[1]**o+(-0.89057979428)*x_ref[1]+(-0.727135553706)*x_ref[2]**o+(0.892946022497)*x_ref[2]
        ref[2,0]=(0.579951394388)*x_ref[0]**o+(0.0406506913385)*x_ref[0]+(0.363944368798)*x_ref[1]**o+(0.582732756951)*x_ref[1]+(0.542548838027)*x_ref[2]**o+(0.365742797343)*x_ref[2]
        ref[2,1]=(0.253260777874)*x_ref[0]**o+(-0.69077350875)*x_ref[0]+(0.458080700492)*x_ref[1]**o+(-0.79224006672)*x_ref[1]+(-0.504865679524)*x_ref[2]**o+(0.681425632809)*x_ref[2]
        ref[2,2]=(0.600427718261)*x_ref[0]**o+(0.91761242151)*x_ref[0]+(0.602316047463)*x_ref[1]**o+(0.890383614185)*x_ref[1]+(-0.386691039525)*x_ref[2]**o+(0.688659949109)*x_ref[2]
        ref[2,3]=(-0.181338810914)*x_ref[0]**o+(0.933744241999)*x_ref[0]+(0.0212774093137)*x_ref[1]**o+(0.284972854884)*x_ref[1]+(-0.150219922419)*x_ref[2]**o+(0.0878011402107)*x_ref[2]
        ref[2,4]=(-0.848985879538)*x_ref[0]**o+(0.298962025531)*x_ref[0]+(0.593300576179)*x_ref[1]**o+(0.301509253541)*x_ref[1]+(0.971171084219)*x_ref[2]**o+(-0.973386482857)*x_ref[2]
        ref[3,0]=(-0.735608389534)*x_ref[0]**o+(0.44902436426)*x_ref[0]+(0.0536456095512)*x_ref[1]**o+(0.208969935548)*x_ref[1]+(-0.787609300211)*x_ref[2]**o+(0.726024812477)*x_ref[2]
        ref[3,1]=(0.200882171943)*x_ref[0]**o+(-0.644461780745)*x_ref[0]+(-0.0995704510264)*x_ref[1]**o+(-0.857730241944)*x_ref[1]+(-0.372302969302)*x_ref[2]**o+(-0.171367954739)*x_ref[2]
        ref[3,2]=(-0.673606332627)*x_ref[0]**o+(-0.828183678442)*x_ref[0]+(-0.236856099365)*x_ref[1]**o+(-0.266339053288)*x_ref[1]+(-0.972034740204)*x_ref[2]**o+(-0.33151750348)*x_ref[2]
        ref[3,3]=(-0.312470720017)*x_ref[0]**o+(-0.657433718629)*x_ref[0]+(-0.984940036612)*x_ref[1]**o+(0.197707053145)*x_ref[1]+(0.512009064643)*x_ref[2]**o+(0.855601033248)*x_ref[2]
        ref[3,4]=(-0.674854508228)*x_ref[0]**o+(-0.394035682831)*x_ref[0]+(-0.749446232511)*x_ref[1]**o+(0.866992763518)*x_ref[1]+(-0.945837361158)*x_ref[2]**o+(0.593490782481)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.511844892619)*x[0]**o+(0.72447480361)*x[0]+(-0.415725587152)*x[1]**o+(-0.495233818217)*x[1]
        arg[0,0,1]=(0.513861713949)*x[0]**o+(-0.918192893835)*x[0]+(0.563871531176)*x[1]**o+(0.885768393284)*x[1]
        arg[0,1,0]=(-0.412320762031)*x[0]**o+(0.576776422795)*x[0]+(0.0103535112247)*x[1]**o+(-0.738144219037)*x[1]
        arg[0,1,1]=(0.69941105604)*x[0]**o+(0.923470022558)*x[0]+(0.568449180547)*x[1]**o+(0.361429609343)*x[1]
        arg[1,0,0]=(-0.574139432427)*x[0]**o+(0.968980609212)*x[0]+(-0.198809151818)*x[1]**o+(0.751623099192)*x[1]
        arg[1,0,1]=(-0.743433955873)*x[0]**o+(-0.608013190658)*x[0]+(-0.0625473402572)*x[1]**o+(0.168418241981)*x[1]
        arg[1,1,0]=(-0.632180449517)*x[0]**o+(-0.892666803222)*x[0]+(-0.558127678438)*x[1]**o+(0.212462582041)*x[1]
        arg[1,1,1]=(-0.666381769622)*x[0]**o+(-0.0299308846938)*x[0]+(-0.450221481906)*x[1]**o+(0.701099654576)*x[1]
        arg[2,0,0]=(-0.927832262778)*x[0]**o+(0.901665098705)*x[0]+(-0.0462955000541)*x[1]**o+(-0.92139600993)*x[1]
        arg[2,0,1]=(-0.813732112672)*x[0]**o+(0.569194058283)*x[0]+(-0.527298120198)*x[1]**o+(-0.531607593369)*x[1]
        arg[2,1,0]=(-0.760238058111)*x[0]**o+(0.628718839289)*x[0]+(-0.420066582474)*x[1]**o+(-0.205537665056)*x[1]
        arg[2,1,1]=(0.421784629744)*x[0]**o+(-0.781451031751)*x[0]+(-0.0401156726902)*x[1]**o+(0.437044655245)*x[1]
        arg[3,0,0]=(0.324136561188)*x[0]**o+(-0.920235828584)*x[0]+(0.677675509043)*x[1]**o+(0.322337884254)*x[1]
        arg[3,0,1]=(-0.0711813236306)*x[0]**o+(0.0770312838088)*x[0]+(0.259613502243)*x[1]**o+(0.508154611576)*x[1]
        arg[3,1,0]=(0.944048224599)*x[0]**o+(0.039954080407)*x[0]+(-0.503309434151)*x[1]**o+(0.747109874504)*x[1]
        arg[3,1,1]=(0.692616323513)*x[0]**o+(0.0114568310884)*x[0]+(0.0708318026263)*x[1]**o+(0.822167043168)*x[1]
        arg[4,0,0]=(0.392645136008)*x[0]**o+(0.160054219263)*x[0]+(-0.32320145108)*x[1]**o+(0.598963113077)*x[1]
        arg[4,0,1]=(0.124778736846)*x[0]**o+(0.949589553291)*x[0]+(0.0985585514202)*x[1]**o+(-0.536684039884)*x[1]
        arg[4,1,0]=(-0.983690591076)*x[0]**o+(-0.115321512368)*x[0]+(0.0875590878092)*x[1]**o+(0.992162964289)*x[1]
        arg[4,1,1]=(-0.89264628952)*x[0]**o+(0.414589880453)*x[0]+(-0.40959764673)*x[1]**o+(-0.261016337239)*x[1]
        arg[5,0,0]=(-0.613553175002)*x[0]**o+(0.320685999723)*x[0]+(-0.523099455238)*x[1]**o+(0.855780524769)*x[1]
        arg[5,0,1]=(-0.45911360029)*x[0]**o+(-0.285942390871)*x[0]+(0.7279702976)*x[1]**o+(0.981483922949)*x[1]
        arg[5,1,0]=(0.577254626037)*x[0]**o+(0.146911866492)*x[0]+(-0.414545328014)*x[1]**o+(-0.979179708006)*x[1]
        arg[5,1,1]=(-0.838392720256)*x[0]**o+(-0.51717223091)*x[0]+(0.635616756614)*x[1]**o+(0.112711970543)*x[1]
        ref[0,0,0]=(-0.511844892619)*x_ref[0]**o+(0.72447480361)*x_ref[0]+(-0.415725587152)*x_ref[1]**o+(-0.495233818217)*x_ref[1]
        ref[0,0,1]=(0.513861713949)*x_ref[0]**o+(-0.918192893835)*x_ref[0]+(0.563871531176)*x_ref[1]**o+(0.885768393284)*x_ref[1]
        ref[0,1,0]=(-0.412320762031)*x_ref[0]**o+(0.576776422795)*x_ref[0]+(0.0103535112247)*x_ref[1]**o+(-0.738144219037)*x_ref[1]
        ref[0,1,1]=(0.69941105604)*x_ref[0]**o+(0.923470022558)*x_ref[0]+(0.568449180547)*x_ref[1]**o+(0.361429609343)*x_ref[1]
        ref[1,0,0]=(-0.574139432427)*x_ref[0]**o+(0.968980609212)*x_ref[0]+(-0.198809151818)*x_ref[1]**o+(0.751623099192)*x_ref[1]
        ref[1,0,1]=(-0.743433955873)*x_ref[0]**o+(-0.608013190658)*x_ref[0]+(-0.0625473402572)*x_ref[1]**o+(0.168418241981)*x_ref[1]
        ref[1,1,0]=(-0.632180449517)*x_ref[0]**o+(-0.892666803222)*x_ref[0]+(-0.558127678438)*x_ref[1]**o+(0.212462582041)*x_ref[1]
        ref[1,1,1]=(-0.666381769622)*x_ref[0]**o+(-0.0299308846938)*x_ref[0]+(-0.450221481906)*x_ref[1]**o+(0.701099654576)*x_ref[1]
        ref[2,0,0]=(-0.927832262778)*x_ref[0]**o+(0.901665098705)*x_ref[0]+(-0.0462955000541)*x_ref[1]**o+(-0.92139600993)*x_ref[1]
        ref[2,0,1]=(-0.813732112672)*x_ref[0]**o+(0.569194058283)*x_ref[0]+(-0.527298120198)*x_ref[1]**o+(-0.531607593369)*x_ref[1]
        ref[2,1,0]=(-0.760238058111)*x_ref[0]**o+(0.628718839289)*x_ref[0]+(-0.420066582474)*x_ref[1]**o+(-0.205537665056)*x_ref[1]
        ref[2,1,1]=(0.421784629744)*x_ref[0]**o+(-0.781451031751)*x_ref[0]+(-0.0401156726902)*x_ref[1]**o+(0.437044655245)*x_ref[1]
        ref[3,0,0]=(0.324136561188)*x_ref[0]**o+(-0.920235828584)*x_ref[0]+(0.677675509043)*x_ref[1]**o+(0.322337884254)*x_ref[1]
        ref[3,0,1]=(-0.0711813236306)*x_ref[0]**o+(0.0770312838088)*x_ref[0]+(0.259613502243)*x_ref[1]**o+(0.508154611576)*x_ref[1]
        ref[3,1,0]=(0.944048224599)*x_ref[0]**o+(0.039954080407)*x_ref[0]+(-0.503309434151)*x_ref[1]**o+(0.747109874504)*x_ref[1]
        ref[3,1,1]=(0.692616323513)*x_ref[0]**o+(0.0114568310884)*x_ref[0]+(0.0708318026263)*x_ref[1]**o+(0.822167043168)*x_ref[1]
        ref[4,0,0]=(0.392645136008)*x_ref[0]**o+(0.160054219263)*x_ref[0]+(-0.32320145108)*x_ref[1]**o+(0.598963113077)*x_ref[1]
        ref[4,0,1]=(0.124778736846)*x_ref[0]**o+(0.949589553291)*x_ref[0]+(0.0985585514202)*x_ref[1]**o+(-0.536684039884)*x_ref[1]
        ref[4,1,0]=(-0.983690591076)*x_ref[0]**o+(-0.115321512368)*x_ref[0]+(0.0875590878092)*x_ref[1]**o+(0.992162964289)*x_ref[1]
        ref[4,1,1]=(-0.89264628952)*x_ref[0]**o+(0.414589880453)*x_ref[0]+(-0.40959764673)*x_ref[1]**o+(-0.261016337239)*x_ref[1]
        ref[5,0,0]=(-0.613553175002)*x_ref[0]**o+(0.320685999723)*x_ref[0]+(-0.523099455238)*x_ref[1]**o+(0.855780524769)*x_ref[1]
        ref[5,0,1]=(-0.45911360029)*x_ref[0]**o+(-0.285942390871)*x_ref[0]+(0.7279702976)*x_ref[1]**o+(0.981483922949)*x_ref[1]
        ref[5,1,0]=(0.577254626037)*x_ref[0]**o+(0.146911866492)*x_ref[0]+(-0.414545328014)*x_ref[1]**o+(-0.979179708006)*x_ref[1]
        ref[5,1,1]=(-0.838392720256)*x_ref[0]**o+(-0.51717223091)*x_ref[0]+(0.635616756614)*x_ref[1]**o+(0.112711970543)*x_ref[1]
      else:
        arg[0,0,0]=(0.807360140818)*x[0]**o+(-0.399141641865)*x[0]+(0.752910658124)*x[1]**o+(0.0998734354996)*x[1]+(-0.146785318133)*x[2]**o+(-0.393796605729)*x[2]
        arg[0,0,1]=(-0.502209573844)*x[0]**o+(0.501684846477)*x[0]+(0.634537290794)*x[1]**o+(0.481836589517)*x[1]+(-0.0221536339712)*x[2]**o+(-0.868693970024)*x[2]
        arg[0,1,0]=(-0.268854282014)*x[0]**o+(0.0516535086968)*x[0]+(-0.505997013698)*x[1]**o+(0.414205349613)*x[1]+(0.0119995662599)*x[2]**o+(0.584354222675)*x[2]
        arg[0,1,1]=(0.81923283939)*x[0]**o+(-0.174637610292)*x[0]+(0.283785251937)*x[1]**o+(0.716260504685)*x[1]+(-0.957279283093)*x[2]**o+(0.632750238692)*x[2]
        arg[1,0,0]=(0.601003888536)*x[0]**o+(-0.575011763358)*x[0]+(0.095159775614)*x[1]**o+(0.141200996742)*x[1]+(0.0173869440863)*x[2]**o+(0.478733688259)*x[2]
        arg[1,0,1]=(-0.783826426764)*x[0]**o+(-0.40604430175)*x[0]+(-0.4574879786)*x[1]**o+(0.020222886424)*x[1]+(0.225419750344)*x[2]**o+(-0.530026190372)*x[2]
        arg[1,1,0]=(0.542744341911)*x[0]**o+(-0.501169367089)*x[0]+(0.447661946532)*x[1]**o+(0.384528043888)*x[1]+(-0.599271261928)*x[2]**o+(-0.535447844217)*x[2]
        arg[1,1,1]=(-0.457163465824)*x[0]**o+(0.0475382663524)*x[0]+(0.369626614251)*x[1]**o+(0.1125159395)*x[1]+(-0.252442923096)*x[2]**o+(0.0717205095822)*x[2]
        arg[2,0,0]=(0.341536174626)*x[0]**o+(0.4408059057)*x[0]+(0.196877383546)*x[1]**o+(0.903072917866)*x[1]+(0.423403716275)*x[2]**o+(-0.521563534689)*x[2]
        arg[2,0,1]=(0.553187361262)*x[0]**o+(-0.539057536895)*x[0]+(-0.814990210449)*x[1]**o+(-0.418614789609)*x[1]+(0.857298489434)*x[2]**o+(0.629708672)*x[2]
        arg[2,1,0]=(0.494803348038)*x[0]**o+(0.143644562028)*x[0]+(-0.0425813919235)*x[1]**o+(-0.818562153097)*x[1]+(-0.730690839032)*x[2]**o+(-0.569426137866)*x[2]
        arg[2,1,1]=(-0.573254727715)*x[0]**o+(-0.536459946225)*x[0]+(-0.244717967781)*x[1]**o+(-0.67463285343)*x[1]+(-0.383959754049)*x[2]**o+(-0.87595126443)*x[2]
        arg[3,0,0]=(0.758065154533)*x[0]**o+(0.634049823109)*x[0]+(0.851352543082)*x[1]**o+(0.302790462394)*x[1]+(-0.920790730269)*x[2]**o+(0.460419670912)*x[2]
        arg[3,0,1]=(0.415682634939)*x[0]**o+(-0.664821834506)*x[0]+(-0.314068887264)*x[1]**o+(-0.322413102029)*x[1]+(-0.498768865759)*x[2]**o+(-0.165338012626)*x[2]
        arg[3,1,0]=(-0.398599595126)*x[0]**o+(0.744397086526)*x[0]+(-0.02321839482)*x[1]**o+(0.989743131098)*x[1]+(0.0713981107777)*x[2]**o+(-0.7862942966)*x[2]
        arg[3,1,1]=(0.90835870295)*x[0]**o+(0.285965355995)*x[0]+(-0.302780058578)*x[1]**o+(0.396923771953)*x[1]+(0.717312324488)*x[2]**o+(0.967488138744)*x[2]
        arg[4,0,0]=(-0.286827561624)*x[0]**o+(0.67973673275)*x[0]+(0.325169520671)*x[1]**o+(0.482002981846)*x[1]+(-0.826037251282)*x[2]**o+(-0.0567054906659)*x[2]
        arg[4,0,1]=(-0.667297458554)*x[0]**o+(-0.207369757609)*x[0]+(-0.982746021703)*x[1]**o+(0.963677089363)*x[1]+(0.917328237047)*x[2]**o+(-0.684386577728)*x[2]
        arg[4,1,0]=(-0.835525763762)*x[0]**o+(-0.113324184435)*x[0]+(0.465855661969)*x[1]**o+(0.656047307997)*x[1]+(0.331675778117)*x[2]**o+(-0.686067177565)*x[2]
        arg[4,1,1]=(-0.468668496499)*x[0]**o+(0.295197251458)*x[0]+(0.065102699741)*x[1]**o+(-0.378381176309)*x[1]+(0.933627973066)*x[2]**o+(-0.91462136931)*x[2]
        arg[5,0,0]=(-0.161916994488)*x[0]**o+(0.3070005414)*x[0]+(0.195946524325)*x[1]**o+(-0.453290146527)*x[1]+(0.751532065651)*x[2]**o+(0.82676125654)*x[2]
        arg[5,0,1]=(-0.925362363492)*x[0]**o+(-0.0295351854097)*x[0]+(0.533980764998)*x[1]**o+(0.128080489262)*x[1]+(0.502858052015)*x[2]**o+(0.813196540163)*x[2]
        arg[5,1,0]=(-0.740146099318)*x[0]**o+(0.27623528724)*x[0]+(-0.361620073401)*x[1]**o+(-0.310534085108)*x[1]+(-0.0363428185757)*x[2]**o+(-0.145397849357)*x[2]
        arg[5,1,1]=(0.362089087335)*x[0]**o+(0.539359659929)*x[0]+(0.00904990097799)*x[1]**o+(-0.150401175722)*x[1]+(-0.20465882105)*x[2]**o+(-0.519331232192)*x[2]
        ref[0,0,0]=(0.807360140818)*x_ref[0]**o+(-0.399141641865)*x_ref[0]+(0.752910658124)*x_ref[1]**o+(0.0998734354996)*x_ref[1]+(-0.146785318133)*x_ref[2]**o+(-0.393796605729)*x_ref[2]
        ref[0,0,1]=(-0.502209573844)*x_ref[0]**o+(0.501684846477)*x_ref[0]+(0.634537290794)*x_ref[1]**o+(0.481836589517)*x_ref[1]+(-0.0221536339712)*x_ref[2]**o+(-0.868693970024)*x_ref[2]
        ref[0,1,0]=(-0.268854282014)*x_ref[0]**o+(0.0516535086968)*x_ref[0]+(-0.505997013698)*x_ref[1]**o+(0.414205349613)*x_ref[1]+(0.0119995662599)*x_ref[2]**o+(0.584354222675)*x_ref[2]
        ref[0,1,1]=(0.81923283939)*x_ref[0]**o+(-0.174637610292)*x_ref[0]+(0.283785251937)*x_ref[1]**o+(0.716260504685)*x_ref[1]+(-0.957279283093)*x_ref[2]**o+(0.632750238692)*x_ref[2]
        ref[1,0,0]=(0.601003888536)*x_ref[0]**o+(-0.575011763358)*x_ref[0]+(0.095159775614)*x_ref[1]**o+(0.141200996742)*x_ref[1]+(0.0173869440863)*x_ref[2]**o+(0.478733688259)*x_ref[2]
        ref[1,0,1]=(-0.783826426764)*x_ref[0]**o+(-0.40604430175)*x_ref[0]+(-0.4574879786)*x_ref[1]**o+(0.020222886424)*x_ref[1]+(0.225419750344)*x_ref[2]**o+(-0.530026190372)*x_ref[2]
        ref[1,1,0]=(0.542744341911)*x_ref[0]**o+(-0.501169367089)*x_ref[0]+(0.447661946532)*x_ref[1]**o+(0.384528043888)*x_ref[1]+(-0.599271261928)*x_ref[2]**o+(-0.535447844217)*x_ref[2]
        ref[1,1,1]=(-0.457163465824)*x_ref[0]**o+(0.0475382663524)*x_ref[0]+(0.369626614251)*x_ref[1]**o+(0.1125159395)*x_ref[1]+(-0.252442923096)*x_ref[2]**o+(0.0717205095822)*x_ref[2]
        ref[2,0,0]=(0.341536174626)*x_ref[0]**o+(0.4408059057)*x_ref[0]+(0.196877383546)*x_ref[1]**o+(0.903072917866)*x_ref[1]+(0.423403716275)*x_ref[2]**o+(-0.521563534689)*x_ref[2]
        ref[2,0,1]=(0.553187361262)*x_ref[0]**o+(-0.539057536895)*x_ref[0]+(-0.814990210449)*x_ref[1]**o+(-0.418614789609)*x_ref[1]+(0.857298489434)*x_ref[2]**o+(0.629708672)*x_ref[2]
        ref[2,1,0]=(0.494803348038)*x_ref[0]**o+(0.143644562028)*x_ref[0]+(-0.0425813919235)*x_ref[1]**o+(-0.818562153097)*x_ref[1]+(-0.730690839032)*x_ref[2]**o+(-0.569426137866)*x_ref[2]
        ref[2,1,1]=(-0.573254727715)*x_ref[0]**o+(-0.536459946225)*x_ref[0]+(-0.244717967781)*x_ref[1]**o+(-0.67463285343)*x_ref[1]+(-0.383959754049)*x_ref[2]**o+(-0.87595126443)*x_ref[2]
        ref[3,0,0]=(0.758065154533)*x_ref[0]**o+(0.634049823109)*x_ref[0]+(0.851352543082)*x_ref[1]**o+(0.302790462394)*x_ref[1]+(-0.920790730269)*x_ref[2]**o+(0.460419670912)*x_ref[2]
        ref[3,0,1]=(0.415682634939)*x_ref[0]**o+(-0.664821834506)*x_ref[0]+(-0.314068887264)*x_ref[1]**o+(-0.322413102029)*x_ref[1]+(-0.498768865759)*x_ref[2]**o+(-0.165338012626)*x_ref[2]
        ref[3,1,0]=(-0.398599595126)*x_ref[0]**o+(0.744397086526)*x_ref[0]+(-0.02321839482)*x_ref[1]**o+(0.989743131098)*x_ref[1]+(0.0713981107777)*x_ref[2]**o+(-0.7862942966)*x_ref[2]
        ref[3,1,1]=(0.90835870295)*x_ref[0]**o+(0.285965355995)*x_ref[0]+(-0.302780058578)*x_ref[1]**o+(0.396923771953)*x_ref[1]+(0.717312324488)*x_ref[2]**o+(0.967488138744)*x_ref[2]
        ref[4,0,0]=(-0.286827561624)*x_ref[0]**o+(0.67973673275)*x_ref[0]+(0.325169520671)*x_ref[1]**o+(0.482002981846)*x_ref[1]+(-0.826037251282)*x_ref[2]**o+(-0.0567054906659)*x_ref[2]
        ref[4,0,1]=(-0.667297458554)*x_ref[0]**o+(-0.207369757609)*x_ref[0]+(-0.982746021703)*x_ref[1]**o+(0.963677089363)*x_ref[1]+(0.917328237047)*x_ref[2]**o+(-0.684386577728)*x_ref[2]
        ref[4,1,0]=(-0.835525763762)*x_ref[0]**o+(-0.113324184435)*x_ref[0]+(0.465855661969)*x_ref[1]**o+(0.656047307997)*x_ref[1]+(0.331675778117)*x_ref[2]**o+(-0.686067177565)*x_ref[2]
        ref[4,1,1]=(-0.468668496499)*x_ref[0]**o+(0.295197251458)*x_ref[0]+(0.065102699741)*x_ref[1]**o+(-0.378381176309)*x_ref[1]+(0.933627973066)*x_ref[2]**o+(-0.91462136931)*x_ref[2]
        ref[5,0,0]=(-0.161916994488)*x_ref[0]**o+(0.3070005414)*x_ref[0]+(0.195946524325)*x_ref[1]**o+(-0.453290146527)*x_ref[1]+(0.751532065651)*x_ref[2]**o+(0.82676125654)*x_ref[2]
        ref[5,0,1]=(-0.925362363492)*x_ref[0]**o+(-0.0295351854097)*x_ref[0]+(0.533980764998)*x_ref[1]**o+(0.128080489262)*x_ref[1]+(0.502858052015)*x_ref[2]**o+(0.813196540163)*x_ref[2]
        ref[5,1,0]=(-0.740146099318)*x_ref[0]**o+(0.27623528724)*x_ref[0]+(-0.361620073401)*x_ref[1]**o+(-0.310534085108)*x_ref[1]+(-0.0363428185757)*x_ref[2]**o+(-0.145397849357)*x_ref[2]
        ref[5,1,1]=(0.362089087335)*x_ref[0]**o+(0.539359659929)*x_ref[0]+(0.00904990097799)*x_ref[1]**o+(-0.150401175722)*x_ref[1]+(-0.20465882105)*x_ref[2]**o+(-0.519331232192)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.14016992168)*x[0]**o+(-0.396856222091)*x[0]+(0.747446299477)*x[1]**o+(-0.299063042467)*x[1]
        arg[0,0,0,1]=(0.126297805139)*x[0]**o+(0.317089000483)*x[0]+(0.0418478916266)*x[1]**o+(-0.892825689806)*x[1]
        arg[0,0,1,0]=(0.127263485532)*x[0]**o+(0.400880552184)*x[0]+(-0.858713979216)*x[1]**o+(0.238693631738)*x[1]
        arg[0,0,1,1]=(-0.985703521737)*x[0]**o+(-0.0192845851979)*x[0]+(-0.442493677109)*x[1]**o+(0.188038174612)*x[1]
        arg[0,0,2,0]=(0.814303535641)*x[0]**o+(-0.201938227136)*x[0]+(-0.134135136581)*x[1]**o+(0.939272592089)*x[1]
        arg[0,0,2,1]=(-0.609570412575)*x[0]**o+(0.181669701944)*x[0]+(0.882708213835)*x[1]**o+(0.315561722535)*x[1]
        arg[0,1,0,0]=(0.180272217085)*x[0]**o+(-0.921231698951)*x[0]+(-0.670170775864)*x[1]**o+(0.36959006719)*x[1]
        arg[0,1,0,1]=(-0.907447754449)*x[0]**o+(0.984356686322)*x[0]+(0.300621254386)*x[1]**o+(0.491995830235)*x[1]
        arg[0,1,1,0]=(-0.535808887861)*x[0]**o+(0.991647965865)*x[0]+(-0.505667913884)*x[1]**o+(-0.403245981754)*x[1]
        arg[0,1,1,1]=(-0.315687983887)*x[0]**o+(-0.570118136173)*x[0]+(0.327842038508)*x[1]**o+(-0.475833876716)*x[1]
        arg[0,1,2,0]=(0.598143043008)*x[0]**o+(0.119789453354)*x[0]+(-0.307994361027)*x[1]**o+(-0.747443620209)*x[1]
        arg[0,1,2,1]=(-0.384576138426)*x[0]**o+(-0.21431195392)*x[0]+(0.622157923623)*x[1]**o+(-0.535976986322)*x[1]
        arg[0,2,0,0]=(-0.823640169824)*x[0]**o+(-0.946436729867)*x[0]+(0.158819942161)*x[1]**o+(-0.175783146941)*x[1]
        arg[0,2,0,1]=(0.31586435609)*x[0]**o+(0.480823809413)*x[0]+(-0.155590673671)*x[1]**o+(-0.283774897002)*x[1]
        arg[0,2,1,0]=(0.973677960338)*x[0]**o+(-0.42119441385)*x[0]+(0.936434159723)*x[1]**o+(-0.271328279678)*x[1]
        arg[0,2,1,1]=(0.995977746066)*x[0]**o+(0.115566554312)*x[0]+(0.6023712777)*x[1]**o+(0.523708594385)*x[1]
        arg[0,2,2,0]=(0.690331991941)*x[0]**o+(0.708864222789)*x[0]+(-0.0209681974418)*x[1]**o+(-0.402664547241)*x[1]
        arg[0,2,2,1]=(0.793403379571)*x[0]**o+(-0.44195907862)*x[0]+(-0.750943552401)*x[1]**o+(0.87817105061)*x[1]
        arg[0,3,0,0]=(-0.960650344452)*x[0]**o+(0.918969286865)*x[0]+(-0.788611475275)*x[1]**o+(-0.917265819096)*x[1]
        arg[0,3,0,1]=(0.379234695969)*x[0]**o+(0.288784268667)*x[0]+(-0.231788371542)*x[1]**o+(-0.521488838978)*x[1]
        arg[0,3,1,0]=(-0.482353410226)*x[0]**o+(0.666269361063)*x[0]+(0.207628082266)*x[1]**o+(0.890536697637)*x[1]
        arg[0,3,1,1]=(-0.256826971067)*x[0]**o+(0.244573109457)*x[0]+(0.435182409565)*x[1]**o+(-0.98882467576)*x[1]
        arg[0,3,2,0]=(0.727681501266)*x[0]**o+(0.174117939756)*x[0]+(0.626998345198)*x[1]**o+(0.666518756047)*x[1]
        arg[0,3,2,1]=(-0.125855300157)*x[0]**o+(-0.750291117974)*x[0]+(0.840722993443)*x[1]**o+(0.0798848819291)*x[1]
        arg[0,4,0,0]=(-0.156808480304)*x[0]**o+(-0.864855000875)*x[0]+(-0.825786767174)*x[1]**o+(-0.563999207624)*x[1]
        arg[0,4,0,1]=(-0.890116284779)*x[0]**o+(0.795471417269)*x[0]+(0.326454331959)*x[1]**o+(-0.376644334859)*x[1]
        arg[0,4,1,0]=(-0.453181887597)*x[0]**o+(0.488149970589)*x[0]+(0.506352954527)*x[1]**o+(0.971016325918)*x[1]
        arg[0,4,1,1]=(0.156976484311)*x[0]**o+(0.988818166411)*x[0]+(0.948579296281)*x[1]**o+(-0.80303322078)*x[1]
        arg[0,4,2,0]=(-0.0625333956516)*x[0]**o+(0.357609677642)*x[0]+(0.00245182913993)*x[1]**o+(-0.162058851234)*x[1]
        arg[0,4,2,1]=(0.645504631508)*x[0]**o+(-0.0107543576265)*x[0]+(0.15490632437)*x[1]**o+(0.841985786269)*x[1]
        arg[1,0,0,0]=(0.703923364197)*x[0]**o+(-0.424818178123)*x[0]+(-0.744399116071)*x[1]**o+(0.303429987787)*x[1]
        arg[1,0,0,1]=(-0.92849671168)*x[0]**o+(0.050597602551)*x[0]+(0.91000243103)*x[1]**o+(-0.742775174195)*x[1]
        arg[1,0,1,0]=(0.281218131435)*x[0]**o+(0.171029842909)*x[0]+(-0.65806576709)*x[1]**o+(-0.718069456213)*x[1]
        arg[1,0,1,1]=(0.312255067232)*x[0]**o+(0.566401030665)*x[0]+(-0.941709776271)*x[1]**o+(0.0111896041808)*x[1]
        arg[1,0,2,0]=(-0.886019321094)*x[0]**o+(-0.688815756004)*x[0]+(-0.595491288141)*x[1]**o+(0.343402495621)*x[1]
        arg[1,0,2,1]=(-0.681063845863)*x[0]**o+(-0.676550403066)*x[0]+(0.186106532969)*x[1]**o+(0.181779537727)*x[1]
        arg[1,1,0,0]=(0.128588939989)*x[0]**o+(0.477711214734)*x[0]+(-0.17697460919)*x[1]**o+(0.0217208670244)*x[1]
        arg[1,1,0,1]=(-0.652028522296)*x[0]**o+(-0.503146425209)*x[0]+(-0.427973128189)*x[1]**o+(-0.182775422425)*x[1]
        arg[1,1,1,0]=(-0.0499349942904)*x[0]**o+(-0.0183804745613)*x[0]+(-0.912604094211)*x[1]**o+(-0.949181938689)*x[1]
        arg[1,1,1,1]=(-0.472737626621)*x[0]**o+(-0.142195314687)*x[0]+(0.580775749383)*x[1]**o+(-0.832549796746)*x[1]
        arg[1,1,2,0]=(-0.288756468146)*x[0]**o+(-0.50657056474)*x[0]+(-0.809713268333)*x[1]**o+(0.598450102365)*x[1]
        arg[1,1,2,1]=(-0.259779636241)*x[0]**o+(-0.227925001879)*x[0]+(-0.0752800702413)*x[1]**o+(-0.438841969161)*x[1]
        arg[1,2,0,0]=(-0.617835522076)*x[0]**o+(-0.969275786274)*x[0]+(0.899717033555)*x[1]**o+(0.77648944909)*x[1]
        arg[1,2,0,1]=(-0.0619395581419)*x[0]**o+(0.926394874574)*x[0]+(0.984605961354)*x[1]**o+(0.253347734181)*x[1]
        arg[1,2,1,0]=(0.748474342269)*x[0]**o+(0.0794003341104)*x[0]+(-0.244607682532)*x[1]**o+(-0.228787873564)*x[1]
        arg[1,2,1,1]=(-0.266853205536)*x[0]**o+(0.529714058158)*x[0]+(-0.421599080872)*x[1]**o+(-0.595557599245)*x[1]
        arg[1,2,2,0]=(-0.377962443753)*x[0]**o+(0.609550409525)*x[0]+(0.951904771097)*x[1]**o+(-0.123551801933)*x[1]
        arg[1,2,2,1]=(0.3262671136)*x[0]**o+(-0.574281078833)*x[0]+(0.308261022622)*x[1]**o+(-0.784762723843)*x[1]
        arg[1,3,0,0]=(-0.419601410829)*x[0]**o+(-0.576064062071)*x[0]+(-0.431574278091)*x[1]**o+(-0.264303418881)*x[1]
        arg[1,3,0,1]=(0.176058414693)*x[0]**o+(0.672752046411)*x[0]+(-0.8579992002)*x[1]**o+(-0.135904959185)*x[1]
        arg[1,3,1,0]=(0.84480795105)*x[0]**o+(0.846846799225)*x[0]+(0.134417432017)*x[1]**o+(-0.725316162128)*x[1]
        arg[1,3,1,1]=(-0.102162279538)*x[0]**o+(0.451180031789)*x[0]+(0.490175190736)*x[1]**o+(0.811823281177)*x[1]
        arg[1,3,2,0]=(-0.582177348293)*x[0]**o+(-0.74631626712)*x[0]+(-0.275524221012)*x[1]**o+(-0.658276720144)*x[1]
        arg[1,3,2,1]=(0.304511057297)*x[0]**o+(-0.354531411968)*x[0]+(0.0944491594546)*x[1]**o+(0.619039121597)*x[1]
        arg[1,4,0,0]=(0.754898520324)*x[0]**o+(-0.28914218119)*x[0]+(0.8660675945)*x[1]**o+(0.526777161228)*x[1]
        arg[1,4,0,1]=(-0.820150637429)*x[0]**o+(-0.399203443857)*x[0]+(0.118863382167)*x[1]**o+(0.269335037373)*x[1]
        arg[1,4,1,0]=(-0.669389232967)*x[0]**o+(-0.119080399209)*x[0]+(0.999919873616)*x[1]**o+(0.733390578378)*x[1]
        arg[1,4,1,1]=(-0.85147101348)*x[0]**o+(-0.0453996523642)*x[0]+(0.00891698330334)*x[1]**o+(-0.27128670805)*x[1]
        arg[1,4,2,0]=(-0.946558097166)*x[0]**o+(0.0141300604882)*x[0]+(0.04075090617)*x[1]**o+(-0.141095386138)*x[1]
        arg[1,4,2,1]=(0.613700159605)*x[0]**o+(0.115107829201)*x[0]+(0.907429968641)*x[1]**o+(0.901820725424)*x[1]
        arg[2,0,0,0]=(-0.629486588984)*x[0]**o+(-0.590908868493)*x[0]+(0.689333052519)*x[1]**o+(0.22096464697)*x[1]
        arg[2,0,0,1]=(-0.475364708492)*x[0]**o+(-0.3195126467)*x[0]+(-0.688082519853)*x[1]**o+(-0.889722517035)*x[1]
        arg[2,0,1,0]=(0.931261765284)*x[0]**o+(0.399101548845)*x[0]+(-0.185342345742)*x[1]**o+(0.188927724838)*x[1]
        arg[2,0,1,1]=(-0.195889209768)*x[0]**o+(-0.0878952476483)*x[0]+(0.636980893598)*x[1]**o+(0.827709469156)*x[1]
        arg[2,0,2,0]=(0.644658731705)*x[0]**o+(0.275771474382)*x[0]+(0.0802903363405)*x[1]**o+(-0.157785300253)*x[1]
        arg[2,0,2,1]=(0.324153739483)*x[0]**o+(-0.978043993375)*x[0]+(0.440060321393)*x[1]**o+(0.103952166755)*x[1]
        arg[2,1,0,0]=(-0.233567312787)*x[0]**o+(0.139529308019)*x[0]+(-0.200168759944)*x[1]**o+(-0.629057762953)*x[1]
        arg[2,1,0,1]=(-0.407808993281)*x[0]**o+(-0.208049914815)*x[0]+(0.676817990818)*x[1]**o+(-0.27174220248)*x[1]
        arg[2,1,1,0]=(-0.159113456499)*x[0]**o+(0.586749846436)*x[0]+(0.112658884741)*x[1]**o+(0.521917298039)*x[1]
        arg[2,1,1,1]=(-0.377986619436)*x[0]**o+(0.920912593255)*x[0]+(0.61285598265)*x[1]**o+(0.491316381102)*x[1]
        arg[2,1,2,0]=(-0.594573349393)*x[0]**o+(-0.784940398408)*x[0]+(0.0602427709094)*x[1]**o+(0.381334187216)*x[1]
        arg[2,1,2,1]=(-0.359402310466)*x[0]**o+(0.243014104221)*x[0]+(0.960587701667)*x[1]**o+(0.653124687952)*x[1]
        arg[2,2,0,0]=(0.667031182369)*x[0]**o+(0.099374903559)*x[0]+(0.00279841725989)*x[1]**o+(-0.413751543412)*x[1]
        arg[2,2,0,1]=(0.793557898241)*x[0]**o+(0.146572317359)*x[0]+(0.812964170008)*x[1]**o+(-0.988110793397)*x[1]
        arg[2,2,1,0]=(0.900661246137)*x[0]**o+(-0.997975978586)*x[0]+(0.63300171387)*x[1]**o+(-0.103878548778)*x[1]
        arg[2,2,1,1]=(-0.456464050469)*x[0]**o+(0.830699811714)*x[0]+(-0.186689927425)*x[1]**o+(-0.100170151164)*x[1]
        arg[2,2,2,0]=(0.732282428164)*x[0]**o+(0.928947589228)*x[0]+(-0.282773641579)*x[1]**o+(-0.927533112027)*x[1]
        arg[2,2,2,1]=(0.768111868708)*x[0]**o+(-0.308231472858)*x[0]+(0.890807565689)*x[1]**o+(-0.731196891492)*x[1]
        arg[2,3,0,0]=(0.559658559574)*x[0]**o+(0.104536615096)*x[0]+(-0.162594824125)*x[1]**o+(-0.172859200721)*x[1]
        arg[2,3,0,1]=(0.769297853152)*x[0]**o+(0.346598673904)*x[0]+(0.0313124779153)*x[1]**o+(-0.751767979251)*x[1]
        arg[2,3,1,0]=(0.870887050816)*x[0]**o+(0.593904859463)*x[0]+(-0.108083837899)*x[1]**o+(0.5467383682)*x[1]
        arg[2,3,1,1]=(0.244253227645)*x[0]**o+(0.911176738654)*x[0]+(-0.497855487714)*x[1]**o+(0.226443212585)*x[1]
        arg[2,3,2,0]=(-0.599893560596)*x[0]**o+(-0.893030581717)*x[0]+(-0.745249903413)*x[1]**o+(0.563740068619)*x[1]
        arg[2,3,2,1]=(-0.198088241103)*x[0]**o+(-0.171306730647)*x[0]+(-0.917417138299)*x[1]**o+(0.233920617125)*x[1]
        arg[2,4,0,0]=(0.804585270618)*x[0]**o+(0.3784879961)*x[0]+(0.621076476928)*x[1]**o+(-0.534671153952)*x[1]
        arg[2,4,0,1]=(0.380555159947)*x[0]**o+(-0.630487045841)*x[0]+(0.917413683072)*x[1]**o+(0.491925783834)*x[1]
        arg[2,4,1,0]=(0.946642644964)*x[0]**o+(-0.236438579482)*x[0]+(0.525585953739)*x[1]**o+(0.499940951133)*x[1]
        arg[2,4,1,1]=(0.827450889548)*x[0]**o+(-0.73264543626)*x[0]+(0.0626492901072)*x[1]**o+(-0.882759632862)*x[1]
        arg[2,4,2,0]=(-0.0671814496498)*x[0]**o+(0.108634446052)*x[0]+(-0.421081640868)*x[1]**o+(-0.546455336395)*x[1]
        arg[2,4,2,1]=(-0.601339784479)*x[0]**o+(-0.206937965916)*x[0]+(-0.849120792543)*x[1]**o+(0.615842037441)*x[1]
        arg[3,0,0,0]=(0.640466650913)*x[0]**o+(-0.936176165762)*x[0]+(0.12758686525)*x[1]**o+(-0.104771286099)*x[1]
        arg[3,0,0,1]=(-0.868109993006)*x[0]**o+(0.470815003671)*x[0]+(-0.301208743474)*x[1]**o+(-0.296271408095)*x[1]
        arg[3,0,1,0]=(0.66584252787)*x[0]**o+(-0.889549573986)*x[0]+(-0.0876076820512)*x[1]**o+(-0.249000029178)*x[1]
        arg[3,0,1,1]=(0.289388044606)*x[0]**o+(0.674519856916)*x[0]+(-0.721031236925)*x[1]**o+(0.093281544052)*x[1]
        arg[3,0,2,0]=(-0.0558863543119)*x[0]**o+(-0.367544430651)*x[0]+(-0.85580152565)*x[1]**o+(-0.140166367203)*x[1]
        arg[3,0,2,1]=(0.600709511675)*x[0]**o+(0.286322916564)*x[0]+(-0.544094007372)*x[1]**o+(0.470305833347)*x[1]
        arg[3,1,0,0]=(-0.828780197168)*x[0]**o+(-0.568096532278)*x[0]+(0.435127414299)*x[1]**o+(0.96890877255)*x[1]
        arg[3,1,0,1]=(0.191455614175)*x[0]**o+(-0.99507613827)*x[0]+(0.0735829936134)*x[1]**o+(0.635849313153)*x[1]
        arg[3,1,1,0]=(-0.839049013591)*x[0]**o+(-0.718731394184)*x[0]+(0.575853740085)*x[1]**o+(-0.57924857477)*x[1]
        arg[3,1,1,1]=(0.116876810453)*x[0]**o+(0.145727923885)*x[0]+(0.0155069936368)*x[1]**o+(0.223214885433)*x[1]
        arg[3,1,2,0]=(-0.674811720223)*x[0]**o+(0.651640807063)*x[0]+(-0.652022625703)*x[1]**o+(-0.119110788344)*x[1]
        arg[3,1,2,1]=(-0.544748794193)*x[0]**o+(-0.689440165483)*x[0]+(0.936900319026)*x[1]**o+(-0.337287311564)*x[1]
        arg[3,2,0,0]=(-0.355059804554)*x[0]**o+(0.458306329535)*x[0]+(0.172261907494)*x[1]**o+(0.498548138594)*x[1]
        arg[3,2,0,1]=(0.624420187129)*x[0]**o+(0.808931844722)*x[0]+(-0.690635894194)*x[1]**o+(-0.156122444554)*x[1]
        arg[3,2,1,0]=(-0.190029529776)*x[0]**o+(0.925600418137)*x[0]+(0.601040324953)*x[1]**o+(0.615485749173)*x[1]
        arg[3,2,1,1]=(0.679796674662)*x[0]**o+(0.879285148869)*x[0]+(0.0476373514288)*x[1]**o+(0.37655760501)*x[1]
        arg[3,2,2,0]=(0.231748550393)*x[0]**o+(0.209858138221)*x[0]+(-0.88372579866)*x[1]**o+(0.30534744967)*x[1]
        arg[3,2,2,1]=(-0.772177697671)*x[0]**o+(0.012439044435)*x[0]+(-0.293218976284)*x[1]**o+(0.306955066595)*x[1]
        arg[3,3,0,0]=(0.387256929042)*x[0]**o+(-0.434015953267)*x[0]+(-0.098736946868)*x[1]**o+(-0.0837712534186)*x[1]
        arg[3,3,0,1]=(0.396246252967)*x[0]**o+(-0.338955181605)*x[0]+(-0.353504371554)*x[1]**o+(-0.969811402293)*x[1]
        arg[3,3,1,0]=(-0.336508964723)*x[0]**o+(-0.411882233342)*x[0]+(-0.603965321132)*x[1]**o+(0.177221705416)*x[1]
        arg[3,3,1,1]=(0.786971174387)*x[0]**o+(0.759003522166)*x[0]+(-0.0961289050409)*x[1]**o+(-0.0811258682982)*x[1]
        arg[3,3,2,0]=(0.526266925685)*x[0]**o+(-0.113145219611)*x[0]+(-0.208427816884)*x[1]**o+(0.291484710079)*x[1]
        arg[3,3,2,1]=(-0.591153907468)*x[0]**o+(-0.342892047641)*x[0]+(0.865954283634)*x[1]**o+(-0.22112030903)*x[1]
        arg[3,4,0,0]=(0.755162128186)*x[0]**o+(0.494457469003)*x[0]+(0.889254517137)*x[1]**o+(0.880654056969)*x[1]
        arg[3,4,0,1]=(0.450983401629)*x[0]**o+(-0.0855489110466)*x[0]+(0.0601625202676)*x[1]**o+(-0.0730383237708)*x[1]
        arg[3,4,1,0]=(0.495514588456)*x[0]**o+(-0.0895269876334)*x[0]+(-0.89729630673)*x[1]**o+(0.759465768614)*x[1]
        arg[3,4,1,1]=(-0.666472982468)*x[0]**o+(-0.251715028171)*x[0]+(-0.0623287375223)*x[1]**o+(-0.895663326995)*x[1]
        arg[3,4,2,0]=(0.489737923112)*x[0]**o+(0.868066981889)*x[0]+(0.0371747456704)*x[1]**o+(-0.528951636028)*x[1]
        arg[3,4,2,1]=(0.556312598455)*x[0]**o+(-0.0979773350236)*x[0]+(-0.82785928691)*x[1]**o+(-0.567596765858)*x[1]
        ref[0,0,0,0]=(0.14016992168)*x_ref[0]**o+(-0.396856222091)*x_ref[0]+(0.747446299477)*x_ref[1]**o+(-0.299063042467)*x_ref[1]
        ref[0,0,0,1]=(0.126297805139)*x_ref[0]**o+(0.317089000483)*x_ref[0]+(0.0418478916266)*x_ref[1]**o+(-0.892825689806)*x_ref[1]
        ref[0,0,1,0]=(0.127263485532)*x_ref[0]**o+(0.400880552184)*x_ref[0]+(-0.858713979216)*x_ref[1]**o+(0.238693631738)*x_ref[1]
        ref[0,0,1,1]=(-0.985703521737)*x_ref[0]**o+(-0.0192845851979)*x_ref[0]+(-0.442493677109)*x_ref[1]**o+(0.188038174612)*x_ref[1]
        ref[0,0,2,0]=(0.814303535641)*x_ref[0]**o+(-0.201938227136)*x_ref[0]+(-0.134135136581)*x_ref[1]**o+(0.939272592089)*x_ref[1]
        ref[0,0,2,1]=(-0.609570412575)*x_ref[0]**o+(0.181669701944)*x_ref[0]+(0.882708213835)*x_ref[1]**o+(0.315561722535)*x_ref[1]
        ref[0,1,0,0]=(0.180272217085)*x_ref[0]**o+(-0.921231698951)*x_ref[0]+(-0.670170775864)*x_ref[1]**o+(0.36959006719)*x_ref[1]
        ref[0,1,0,1]=(-0.907447754449)*x_ref[0]**o+(0.984356686322)*x_ref[0]+(0.300621254386)*x_ref[1]**o+(0.491995830235)*x_ref[1]
        ref[0,1,1,0]=(-0.535808887861)*x_ref[0]**o+(0.991647965865)*x_ref[0]+(-0.505667913884)*x_ref[1]**o+(-0.403245981754)*x_ref[1]
        ref[0,1,1,1]=(-0.315687983887)*x_ref[0]**o+(-0.570118136173)*x_ref[0]+(0.327842038508)*x_ref[1]**o+(-0.475833876716)*x_ref[1]
        ref[0,1,2,0]=(0.598143043008)*x_ref[0]**o+(0.119789453354)*x_ref[0]+(-0.307994361027)*x_ref[1]**o+(-0.747443620209)*x_ref[1]
        ref[0,1,2,1]=(-0.384576138426)*x_ref[0]**o+(-0.21431195392)*x_ref[0]+(0.622157923623)*x_ref[1]**o+(-0.535976986322)*x_ref[1]
        ref[0,2,0,0]=(-0.823640169824)*x_ref[0]**o+(-0.946436729867)*x_ref[0]+(0.158819942161)*x_ref[1]**o+(-0.175783146941)*x_ref[1]
        ref[0,2,0,1]=(0.31586435609)*x_ref[0]**o+(0.480823809413)*x_ref[0]+(-0.155590673671)*x_ref[1]**o+(-0.283774897002)*x_ref[1]
        ref[0,2,1,0]=(0.973677960338)*x_ref[0]**o+(-0.42119441385)*x_ref[0]+(0.936434159723)*x_ref[1]**o+(-0.271328279678)*x_ref[1]
        ref[0,2,1,1]=(0.995977746066)*x_ref[0]**o+(0.115566554312)*x_ref[0]+(0.6023712777)*x_ref[1]**o+(0.523708594385)*x_ref[1]
        ref[0,2,2,0]=(0.690331991941)*x_ref[0]**o+(0.708864222789)*x_ref[0]+(-0.0209681974418)*x_ref[1]**o+(-0.402664547241)*x_ref[1]
        ref[0,2,2,1]=(0.793403379571)*x_ref[0]**o+(-0.44195907862)*x_ref[0]+(-0.750943552401)*x_ref[1]**o+(0.87817105061)*x_ref[1]
        ref[0,3,0,0]=(-0.960650344452)*x_ref[0]**o+(0.918969286865)*x_ref[0]+(-0.788611475275)*x_ref[1]**o+(-0.917265819096)*x_ref[1]
        ref[0,3,0,1]=(0.379234695969)*x_ref[0]**o+(0.288784268667)*x_ref[0]+(-0.231788371542)*x_ref[1]**o+(-0.521488838978)*x_ref[1]
        ref[0,3,1,0]=(-0.482353410226)*x_ref[0]**o+(0.666269361063)*x_ref[0]+(0.207628082266)*x_ref[1]**o+(0.890536697637)*x_ref[1]
        ref[0,3,1,1]=(-0.256826971067)*x_ref[0]**o+(0.244573109457)*x_ref[0]+(0.435182409565)*x_ref[1]**o+(-0.98882467576)*x_ref[1]
        ref[0,3,2,0]=(0.727681501266)*x_ref[0]**o+(0.174117939756)*x_ref[0]+(0.626998345198)*x_ref[1]**o+(0.666518756047)*x_ref[1]
        ref[0,3,2,1]=(-0.125855300157)*x_ref[0]**o+(-0.750291117974)*x_ref[0]+(0.840722993443)*x_ref[1]**o+(0.0798848819291)*x_ref[1]
        ref[0,4,0,0]=(-0.156808480304)*x_ref[0]**o+(-0.864855000875)*x_ref[0]+(-0.825786767174)*x_ref[1]**o+(-0.563999207624)*x_ref[1]
        ref[0,4,0,1]=(-0.890116284779)*x_ref[0]**o+(0.795471417269)*x_ref[0]+(0.326454331959)*x_ref[1]**o+(-0.376644334859)*x_ref[1]
        ref[0,4,1,0]=(-0.453181887597)*x_ref[0]**o+(0.488149970589)*x_ref[0]+(0.506352954527)*x_ref[1]**o+(0.971016325918)*x_ref[1]
        ref[0,4,1,1]=(0.156976484311)*x_ref[0]**o+(0.988818166411)*x_ref[0]+(0.948579296281)*x_ref[1]**o+(-0.80303322078)*x_ref[1]
        ref[0,4,2,0]=(-0.0625333956516)*x_ref[0]**o+(0.357609677642)*x_ref[0]+(0.00245182913993)*x_ref[1]**o+(-0.162058851234)*x_ref[1]
        ref[0,4,2,1]=(0.645504631508)*x_ref[0]**o+(-0.0107543576265)*x_ref[0]+(0.15490632437)*x_ref[1]**o+(0.841985786269)*x_ref[1]
        ref[1,0,0,0]=(0.703923364197)*x_ref[0]**o+(-0.424818178123)*x_ref[0]+(-0.744399116071)*x_ref[1]**o+(0.303429987787)*x_ref[1]
        ref[1,0,0,1]=(-0.92849671168)*x_ref[0]**o+(0.050597602551)*x_ref[0]+(0.91000243103)*x_ref[1]**o+(-0.742775174195)*x_ref[1]
        ref[1,0,1,0]=(0.281218131435)*x_ref[0]**o+(0.171029842909)*x_ref[0]+(-0.65806576709)*x_ref[1]**o+(-0.718069456213)*x_ref[1]
        ref[1,0,1,1]=(0.312255067232)*x_ref[0]**o+(0.566401030665)*x_ref[0]+(-0.941709776271)*x_ref[1]**o+(0.0111896041808)*x_ref[1]
        ref[1,0,2,0]=(-0.886019321094)*x_ref[0]**o+(-0.688815756004)*x_ref[0]+(-0.595491288141)*x_ref[1]**o+(0.343402495621)*x_ref[1]
        ref[1,0,2,1]=(-0.681063845863)*x_ref[0]**o+(-0.676550403066)*x_ref[0]+(0.186106532969)*x_ref[1]**o+(0.181779537727)*x_ref[1]
        ref[1,1,0,0]=(0.128588939989)*x_ref[0]**o+(0.477711214734)*x_ref[0]+(-0.17697460919)*x_ref[1]**o+(0.0217208670244)*x_ref[1]
        ref[1,1,0,1]=(-0.652028522296)*x_ref[0]**o+(-0.503146425209)*x_ref[0]+(-0.427973128189)*x_ref[1]**o+(-0.182775422425)*x_ref[1]
        ref[1,1,1,0]=(-0.0499349942904)*x_ref[0]**o+(-0.0183804745613)*x_ref[0]+(-0.912604094211)*x_ref[1]**o+(-0.949181938689)*x_ref[1]
        ref[1,1,1,1]=(-0.472737626621)*x_ref[0]**o+(-0.142195314687)*x_ref[0]+(0.580775749383)*x_ref[1]**o+(-0.832549796746)*x_ref[1]
        ref[1,1,2,0]=(-0.288756468146)*x_ref[0]**o+(-0.50657056474)*x_ref[0]+(-0.809713268333)*x_ref[1]**o+(0.598450102365)*x_ref[1]
        ref[1,1,2,1]=(-0.259779636241)*x_ref[0]**o+(-0.227925001879)*x_ref[0]+(-0.0752800702413)*x_ref[1]**o+(-0.438841969161)*x_ref[1]
        ref[1,2,0,0]=(-0.617835522076)*x_ref[0]**o+(-0.969275786274)*x_ref[0]+(0.899717033555)*x_ref[1]**o+(0.77648944909)*x_ref[1]
        ref[1,2,0,1]=(-0.0619395581419)*x_ref[0]**o+(0.926394874574)*x_ref[0]+(0.984605961354)*x_ref[1]**o+(0.253347734181)*x_ref[1]
        ref[1,2,1,0]=(0.748474342269)*x_ref[0]**o+(0.0794003341104)*x_ref[0]+(-0.244607682532)*x_ref[1]**o+(-0.228787873564)*x_ref[1]
        ref[1,2,1,1]=(-0.266853205536)*x_ref[0]**o+(0.529714058158)*x_ref[0]+(-0.421599080872)*x_ref[1]**o+(-0.595557599245)*x_ref[1]
        ref[1,2,2,0]=(-0.377962443753)*x_ref[0]**o+(0.609550409525)*x_ref[0]+(0.951904771097)*x_ref[1]**o+(-0.123551801933)*x_ref[1]
        ref[1,2,2,1]=(0.3262671136)*x_ref[0]**o+(-0.574281078833)*x_ref[0]+(0.308261022622)*x_ref[1]**o+(-0.784762723843)*x_ref[1]
        ref[1,3,0,0]=(-0.419601410829)*x_ref[0]**o+(-0.576064062071)*x_ref[0]+(-0.431574278091)*x_ref[1]**o+(-0.264303418881)*x_ref[1]
        ref[1,3,0,1]=(0.176058414693)*x_ref[0]**o+(0.672752046411)*x_ref[0]+(-0.8579992002)*x_ref[1]**o+(-0.135904959185)*x_ref[1]
        ref[1,3,1,0]=(0.84480795105)*x_ref[0]**o+(0.846846799225)*x_ref[0]+(0.134417432017)*x_ref[1]**o+(-0.725316162128)*x_ref[1]
        ref[1,3,1,1]=(-0.102162279538)*x_ref[0]**o+(0.451180031789)*x_ref[0]+(0.490175190736)*x_ref[1]**o+(0.811823281177)*x_ref[1]
        ref[1,3,2,0]=(-0.582177348293)*x_ref[0]**o+(-0.74631626712)*x_ref[0]+(-0.275524221012)*x_ref[1]**o+(-0.658276720144)*x_ref[1]
        ref[1,3,2,1]=(0.304511057297)*x_ref[0]**o+(-0.354531411968)*x_ref[0]+(0.0944491594546)*x_ref[1]**o+(0.619039121597)*x_ref[1]
        ref[1,4,0,0]=(0.754898520324)*x_ref[0]**o+(-0.28914218119)*x_ref[0]+(0.8660675945)*x_ref[1]**o+(0.526777161228)*x_ref[1]
        ref[1,4,0,1]=(-0.820150637429)*x_ref[0]**o+(-0.399203443857)*x_ref[0]+(0.118863382167)*x_ref[1]**o+(0.269335037373)*x_ref[1]
        ref[1,4,1,0]=(-0.669389232967)*x_ref[0]**o+(-0.119080399209)*x_ref[0]+(0.999919873616)*x_ref[1]**o+(0.733390578378)*x_ref[1]
        ref[1,4,1,1]=(-0.85147101348)*x_ref[0]**o+(-0.0453996523642)*x_ref[0]+(0.00891698330334)*x_ref[1]**o+(-0.27128670805)*x_ref[1]
        ref[1,4,2,0]=(-0.946558097166)*x_ref[0]**o+(0.0141300604882)*x_ref[0]+(0.04075090617)*x_ref[1]**o+(-0.141095386138)*x_ref[1]
        ref[1,4,2,1]=(0.613700159605)*x_ref[0]**o+(0.115107829201)*x_ref[0]+(0.907429968641)*x_ref[1]**o+(0.901820725424)*x_ref[1]
        ref[2,0,0,0]=(-0.629486588984)*x_ref[0]**o+(-0.590908868493)*x_ref[0]+(0.689333052519)*x_ref[1]**o+(0.22096464697)*x_ref[1]
        ref[2,0,0,1]=(-0.475364708492)*x_ref[0]**o+(-0.3195126467)*x_ref[0]+(-0.688082519853)*x_ref[1]**o+(-0.889722517035)*x_ref[1]
        ref[2,0,1,0]=(0.931261765284)*x_ref[0]**o+(0.399101548845)*x_ref[0]+(-0.185342345742)*x_ref[1]**o+(0.188927724838)*x_ref[1]
        ref[2,0,1,1]=(-0.195889209768)*x_ref[0]**o+(-0.0878952476483)*x_ref[0]+(0.636980893598)*x_ref[1]**o+(0.827709469156)*x_ref[1]
        ref[2,0,2,0]=(0.644658731705)*x_ref[0]**o+(0.275771474382)*x_ref[0]+(0.0802903363405)*x_ref[1]**o+(-0.157785300253)*x_ref[1]
        ref[2,0,2,1]=(0.324153739483)*x_ref[0]**o+(-0.978043993375)*x_ref[0]+(0.440060321393)*x_ref[1]**o+(0.103952166755)*x_ref[1]
        ref[2,1,0,0]=(-0.233567312787)*x_ref[0]**o+(0.139529308019)*x_ref[0]+(-0.200168759944)*x_ref[1]**o+(-0.629057762953)*x_ref[1]
        ref[2,1,0,1]=(-0.407808993281)*x_ref[0]**o+(-0.208049914815)*x_ref[0]+(0.676817990818)*x_ref[1]**o+(-0.27174220248)*x_ref[1]
        ref[2,1,1,0]=(-0.159113456499)*x_ref[0]**o+(0.586749846436)*x_ref[0]+(0.112658884741)*x_ref[1]**o+(0.521917298039)*x_ref[1]
        ref[2,1,1,1]=(-0.377986619436)*x_ref[0]**o+(0.920912593255)*x_ref[0]+(0.61285598265)*x_ref[1]**o+(0.491316381102)*x_ref[1]
        ref[2,1,2,0]=(-0.594573349393)*x_ref[0]**o+(-0.784940398408)*x_ref[0]+(0.0602427709094)*x_ref[1]**o+(0.381334187216)*x_ref[1]
        ref[2,1,2,1]=(-0.359402310466)*x_ref[0]**o+(0.243014104221)*x_ref[0]+(0.960587701667)*x_ref[1]**o+(0.653124687952)*x_ref[1]
        ref[2,2,0,0]=(0.667031182369)*x_ref[0]**o+(0.099374903559)*x_ref[0]+(0.00279841725989)*x_ref[1]**o+(-0.413751543412)*x_ref[1]
        ref[2,2,0,1]=(0.793557898241)*x_ref[0]**o+(0.146572317359)*x_ref[0]+(0.812964170008)*x_ref[1]**o+(-0.988110793397)*x_ref[1]
        ref[2,2,1,0]=(0.900661246137)*x_ref[0]**o+(-0.997975978586)*x_ref[0]+(0.63300171387)*x_ref[1]**o+(-0.103878548778)*x_ref[1]
        ref[2,2,1,1]=(-0.456464050469)*x_ref[0]**o+(0.830699811714)*x_ref[0]+(-0.186689927425)*x_ref[1]**o+(-0.100170151164)*x_ref[1]
        ref[2,2,2,0]=(0.732282428164)*x_ref[0]**o+(0.928947589228)*x_ref[0]+(-0.282773641579)*x_ref[1]**o+(-0.927533112027)*x_ref[1]
        ref[2,2,2,1]=(0.768111868708)*x_ref[0]**o+(-0.308231472858)*x_ref[0]+(0.890807565689)*x_ref[1]**o+(-0.731196891492)*x_ref[1]
        ref[2,3,0,0]=(0.559658559574)*x_ref[0]**o+(0.104536615096)*x_ref[0]+(-0.162594824125)*x_ref[1]**o+(-0.172859200721)*x_ref[1]
        ref[2,3,0,1]=(0.769297853152)*x_ref[0]**o+(0.346598673904)*x_ref[0]+(0.0313124779153)*x_ref[1]**o+(-0.751767979251)*x_ref[1]
        ref[2,3,1,0]=(0.870887050816)*x_ref[0]**o+(0.593904859463)*x_ref[0]+(-0.108083837899)*x_ref[1]**o+(0.5467383682)*x_ref[1]
        ref[2,3,1,1]=(0.244253227645)*x_ref[0]**o+(0.911176738654)*x_ref[0]+(-0.497855487714)*x_ref[1]**o+(0.226443212585)*x_ref[1]
        ref[2,3,2,0]=(-0.599893560596)*x_ref[0]**o+(-0.893030581717)*x_ref[0]+(-0.745249903413)*x_ref[1]**o+(0.563740068619)*x_ref[1]
        ref[2,3,2,1]=(-0.198088241103)*x_ref[0]**o+(-0.171306730647)*x_ref[0]+(-0.917417138299)*x_ref[1]**o+(0.233920617125)*x_ref[1]
        ref[2,4,0,0]=(0.804585270618)*x_ref[0]**o+(0.3784879961)*x_ref[0]+(0.621076476928)*x_ref[1]**o+(-0.534671153952)*x_ref[1]
        ref[2,4,0,1]=(0.380555159947)*x_ref[0]**o+(-0.630487045841)*x_ref[0]+(0.917413683072)*x_ref[1]**o+(0.491925783834)*x_ref[1]
        ref[2,4,1,0]=(0.946642644964)*x_ref[0]**o+(-0.236438579482)*x_ref[0]+(0.525585953739)*x_ref[1]**o+(0.499940951133)*x_ref[1]
        ref[2,4,1,1]=(0.827450889548)*x_ref[0]**o+(-0.73264543626)*x_ref[0]+(0.0626492901072)*x_ref[1]**o+(-0.882759632862)*x_ref[1]
        ref[2,4,2,0]=(-0.0671814496498)*x_ref[0]**o+(0.108634446052)*x_ref[0]+(-0.421081640868)*x_ref[1]**o+(-0.546455336395)*x_ref[1]
        ref[2,4,2,1]=(-0.601339784479)*x_ref[0]**o+(-0.206937965916)*x_ref[0]+(-0.849120792543)*x_ref[1]**o+(0.615842037441)*x_ref[1]
        ref[3,0,0,0]=(0.640466650913)*x_ref[0]**o+(-0.936176165762)*x_ref[0]+(0.12758686525)*x_ref[1]**o+(-0.104771286099)*x_ref[1]
        ref[3,0,0,1]=(-0.868109993006)*x_ref[0]**o+(0.470815003671)*x_ref[0]+(-0.301208743474)*x_ref[1]**o+(-0.296271408095)*x_ref[1]
        ref[3,0,1,0]=(0.66584252787)*x_ref[0]**o+(-0.889549573986)*x_ref[0]+(-0.0876076820512)*x_ref[1]**o+(-0.249000029178)*x_ref[1]
        ref[3,0,1,1]=(0.289388044606)*x_ref[0]**o+(0.674519856916)*x_ref[0]+(-0.721031236925)*x_ref[1]**o+(0.093281544052)*x_ref[1]
        ref[3,0,2,0]=(-0.0558863543119)*x_ref[0]**o+(-0.367544430651)*x_ref[0]+(-0.85580152565)*x_ref[1]**o+(-0.140166367203)*x_ref[1]
        ref[3,0,2,1]=(0.600709511675)*x_ref[0]**o+(0.286322916564)*x_ref[0]+(-0.544094007372)*x_ref[1]**o+(0.470305833347)*x_ref[1]
        ref[3,1,0,0]=(-0.828780197168)*x_ref[0]**o+(-0.568096532278)*x_ref[0]+(0.435127414299)*x_ref[1]**o+(0.96890877255)*x_ref[1]
        ref[3,1,0,1]=(0.191455614175)*x_ref[0]**o+(-0.99507613827)*x_ref[0]+(0.0735829936134)*x_ref[1]**o+(0.635849313153)*x_ref[1]
        ref[3,1,1,0]=(-0.839049013591)*x_ref[0]**o+(-0.718731394184)*x_ref[0]+(0.575853740085)*x_ref[1]**o+(-0.57924857477)*x_ref[1]
        ref[3,1,1,1]=(0.116876810453)*x_ref[0]**o+(0.145727923885)*x_ref[0]+(0.0155069936368)*x_ref[1]**o+(0.223214885433)*x_ref[1]
        ref[3,1,2,0]=(-0.674811720223)*x_ref[0]**o+(0.651640807063)*x_ref[0]+(-0.652022625703)*x_ref[1]**o+(-0.119110788344)*x_ref[1]
        ref[3,1,2,1]=(-0.544748794193)*x_ref[0]**o+(-0.689440165483)*x_ref[0]+(0.936900319026)*x_ref[1]**o+(-0.337287311564)*x_ref[1]
        ref[3,2,0,0]=(-0.355059804554)*x_ref[0]**o+(0.458306329535)*x_ref[0]+(0.172261907494)*x_ref[1]**o+(0.498548138594)*x_ref[1]
        ref[3,2,0,1]=(0.624420187129)*x_ref[0]**o+(0.808931844722)*x_ref[0]+(-0.690635894194)*x_ref[1]**o+(-0.156122444554)*x_ref[1]
        ref[3,2,1,0]=(-0.190029529776)*x_ref[0]**o+(0.925600418137)*x_ref[0]+(0.601040324953)*x_ref[1]**o+(0.615485749173)*x_ref[1]
        ref[3,2,1,1]=(0.679796674662)*x_ref[0]**o+(0.879285148869)*x_ref[0]+(0.0476373514288)*x_ref[1]**o+(0.37655760501)*x_ref[1]
        ref[3,2,2,0]=(0.231748550393)*x_ref[0]**o+(0.209858138221)*x_ref[0]+(-0.88372579866)*x_ref[1]**o+(0.30534744967)*x_ref[1]
        ref[3,2,2,1]=(-0.772177697671)*x_ref[0]**o+(0.012439044435)*x_ref[0]+(-0.293218976284)*x_ref[1]**o+(0.306955066595)*x_ref[1]
        ref[3,3,0,0]=(0.387256929042)*x_ref[0]**o+(-0.434015953267)*x_ref[0]+(-0.098736946868)*x_ref[1]**o+(-0.0837712534186)*x_ref[1]
        ref[3,3,0,1]=(0.396246252967)*x_ref[0]**o+(-0.338955181605)*x_ref[0]+(-0.353504371554)*x_ref[1]**o+(-0.969811402293)*x_ref[1]
        ref[3,3,1,0]=(-0.336508964723)*x_ref[0]**o+(-0.411882233342)*x_ref[0]+(-0.603965321132)*x_ref[1]**o+(0.177221705416)*x_ref[1]
        ref[3,3,1,1]=(0.786971174387)*x_ref[0]**o+(0.759003522166)*x_ref[0]+(-0.0961289050409)*x_ref[1]**o+(-0.0811258682982)*x_ref[1]
        ref[3,3,2,0]=(0.526266925685)*x_ref[0]**o+(-0.113145219611)*x_ref[0]+(-0.208427816884)*x_ref[1]**o+(0.291484710079)*x_ref[1]
        ref[3,3,2,1]=(-0.591153907468)*x_ref[0]**o+(-0.342892047641)*x_ref[0]+(0.865954283634)*x_ref[1]**o+(-0.22112030903)*x_ref[1]
        ref[3,4,0,0]=(0.755162128186)*x_ref[0]**o+(0.494457469003)*x_ref[0]+(0.889254517137)*x_ref[1]**o+(0.880654056969)*x_ref[1]
        ref[3,4,0,1]=(0.450983401629)*x_ref[0]**o+(-0.0855489110466)*x_ref[0]+(0.0601625202676)*x_ref[1]**o+(-0.0730383237708)*x_ref[1]
        ref[3,4,1,0]=(0.495514588456)*x_ref[0]**o+(-0.0895269876334)*x_ref[0]+(-0.89729630673)*x_ref[1]**o+(0.759465768614)*x_ref[1]
        ref[3,4,1,1]=(-0.666472982468)*x_ref[0]**o+(-0.251715028171)*x_ref[0]+(-0.0623287375223)*x_ref[1]**o+(-0.895663326995)*x_ref[1]
        ref[3,4,2,0]=(0.489737923112)*x_ref[0]**o+(0.868066981889)*x_ref[0]+(0.0371747456704)*x_ref[1]**o+(-0.528951636028)*x_ref[1]
        ref[3,4,2,1]=(0.556312598455)*x_ref[0]**o+(-0.0979773350236)*x_ref[0]+(-0.82785928691)*x_ref[1]**o+(-0.567596765858)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.952750275738)*x[0]**o+(0.971489204087)*x[0]+(-0.320706487584)*x[1]**o+(-0.853260471578)*x[1]+(-0.696434862068)*x[2]**o+(0.641824397433)*x[2]
        arg[0,0,0,1]=(0.115484192786)*x[0]**o+(0.147813687309)*x[0]+(-0.998010548186)*x[1]**o+(0.0274712088882)*x[1]+(0.136937971489)*x[2]**o+(-0.657986135444)*x[2]
        arg[0,0,1,0]=(0.805857504779)*x[0]**o+(0.196278916938)*x[0]+(-0.969340785263)*x[1]**o+(-0.543890174888)*x[1]+(0.49118423357)*x[2]**o+(-0.883493800593)*x[2]
        arg[0,0,1,1]=(-0.808802136474)*x[0]**o+(-0.866039395649)*x[0]+(0.0339207025941)*x[1]**o+(-0.441368773534)*x[1]+(0.673688211653)*x[2]**o+(0.917894389398)*x[2]
        arg[0,0,2,0]=(0.0459902210078)*x[0]**o+(0.862332666528)*x[0]+(-0.547792306389)*x[1]**o+(0.722176431732)*x[1]+(-0.511278989032)*x[2]**o+(-0.692775336577)*x[2]
        arg[0,0,2,1]=(-0.641904406513)*x[0]**o+(-0.854450656063)*x[0]+(0.81138894849)*x[1]**o+(0.328886625074)*x[1]+(0.400935831363)*x[2]**o+(-0.25184059766)*x[2]
        arg[0,1,0,0]=(0.69922840942)*x[0]**o+(-0.798085508443)*x[0]+(-0.85191383431)*x[1]**o+(-0.162940829331)*x[1]+(0.504020642931)*x[2]**o+(0.0950957436173)*x[2]
        arg[0,1,0,1]=(0.28901429037)*x[0]**o+(-0.624508745729)*x[0]+(0.921288077478)*x[1]**o+(-0.769207901119)*x[1]+(0.0852374599763)*x[2]**o+(-0.402583033986)*x[2]
        arg[0,1,1,0]=(0.715817282215)*x[0]**o+(0.68735256665)*x[0]+(0.610212058649)*x[1]**o+(0.53560824609)*x[1]+(-0.238908897264)*x[2]**o+(0.619799158474)*x[2]
        arg[0,1,1,1]=(-0.371797414191)*x[0]**o+(0.106263350514)*x[0]+(0.583768355362)*x[1]**o+(0.818797667369)*x[1]+(0.0567571152493)*x[2]**o+(-0.242244867169)*x[2]
        arg[0,1,2,0]=(0.190480659729)*x[0]**o+(-0.632415604771)*x[0]+(0.252513826138)*x[1]**o+(0.874732840469)*x[1]+(-0.367769326078)*x[2]**o+(0.364453292404)*x[2]
        arg[0,1,2,1]=(-0.704507136581)*x[0]**o+(-0.614211737547)*x[0]+(0.533383507313)*x[1]**o+(-0.940007583478)*x[1]+(0.558260002746)*x[2]**o+(0.742565692751)*x[2]
        arg[0,2,0,0]=(0.132581156766)*x[0]**o+(0.671793183183)*x[0]+(-0.308255110769)*x[1]**o+(-0.956002957061)*x[1]+(0.300091541301)*x[2]**o+(-0.714260746791)*x[2]
        arg[0,2,0,1]=(-0.836648080859)*x[0]**o+(0.123813861017)*x[0]+(0.866569934894)*x[1]**o+(-0.957824572351)*x[1]+(-0.904478410952)*x[2]**o+(0.691673758521)*x[2]
        arg[0,2,1,0]=(0.411624721878)*x[0]**o+(0.594054722352)*x[0]+(0.150177701135)*x[1]**o+(0.945855261084)*x[1]+(0.295930158817)*x[2]**o+(-0.607980954424)*x[2]
        arg[0,2,1,1]=(-0.0777741308373)*x[0]**o+(0.697501295655)*x[0]+(0.981797797329)*x[1]**o+(-0.671827077949)*x[1]+(0.292310146229)*x[2]**o+(0.484652357857)*x[2]
        arg[0,2,2,0]=(0.194910469807)*x[0]**o+(-0.70008515623)*x[0]+(0.192808336639)*x[1]**o+(-0.962764477371)*x[1]+(-0.34292240753)*x[2]**o+(0.525470829202)*x[2]
        arg[0,2,2,1]=(0.850368420228)*x[0]**o+(-0.516355386368)*x[0]+(0.395764367873)*x[1]**o+(0.320830752607)*x[1]+(-0.26288679044)*x[2]**o+(0.174457723191)*x[2]
        arg[0,3,0,0]=(-0.713987556845)*x[0]**o+(0.0391066248253)*x[0]+(-0.63911733479)*x[1]**o+(-0.119953032121)*x[1]+(0.697879836447)*x[2]**o+(-0.474957057702)*x[2]
        arg[0,3,0,1]=(-0.128614897423)*x[0]**o+(0.7166036403)*x[0]+(0.678415978447)*x[1]**o+(-0.191453169618)*x[1]+(-0.755261361956)*x[2]**o+(-0.839956004196)*x[2]
        arg[0,3,1,0]=(0.258158740934)*x[0]**o+(-0.412199346967)*x[0]+(-0.781106341272)*x[1]**o+(-0.561162823101)*x[1]+(-0.59549864377)*x[2]**o+(-0.748038415721)*x[2]
        arg[0,3,1,1]=(0.68950222305)*x[0]**o+(-0.904027774445)*x[0]+(0.769744453441)*x[1]**o+(-0.265549152752)*x[1]+(-0.375978837131)*x[2]**o+(-0.387010981761)*x[2]
        arg[0,3,2,0]=(0.46852681319)*x[0]**o+(-0.887650258615)*x[0]+(-0.780275624319)*x[1]**o+(0.527897372599)*x[1]+(0.889281778411)*x[2]**o+(0.82538563877)*x[2]
        arg[0,3,2,1]=(-0.817514448184)*x[0]**o+(-0.230190768519)*x[0]+(-0.101550849344)*x[1]**o+(0.1654885857)*x[1]+(-0.231317749498)*x[2]**o+(0.580153306395)*x[2]
        arg[0,4,0,0]=(0.394993007242)*x[0]**o+(0.305948764395)*x[0]+(-0.66853441368)*x[1]**o+(0.279027498094)*x[1]+(0.161366934709)*x[2]**o+(-0.742002471552)*x[2]
        arg[0,4,0,1]=(-0.661214709471)*x[0]**o+(0.548880317935)*x[0]+(0.472682748503)*x[1]**o+(0.266797505699)*x[1]+(0.405618413683)*x[2]**o+(-0.0539677542667)*x[2]
        arg[0,4,1,0]=(-0.713892424429)*x[0]**o+(0.568075831592)*x[0]+(-0.956180895599)*x[1]**o+(0.63618957695)*x[1]+(0.983696160047)*x[2]**o+(-0.308157929002)*x[2]
        arg[0,4,1,1]=(-0.666582919353)*x[0]**o+(0.0816884254484)*x[0]+(0.190810048896)*x[1]**o+(0.482322905132)*x[1]+(-0.576597070954)*x[2]**o+(-0.722381611553)*x[2]
        arg[0,4,2,0]=(-0.0229650731041)*x[0]**o+(-0.0867816754979)*x[0]+(0.0905042102365)*x[1]**o+(-0.994963424765)*x[1]+(0.74504757815)*x[2]**o+(0.71462549644)*x[2]
        arg[0,4,2,1]=(-0.0464278496147)*x[0]**o+(0.122501026996)*x[0]+(-0.049450141901)*x[1]**o+(-0.388739464086)*x[1]+(-0.151389885067)*x[2]**o+(-0.0316948088803)*x[2]
        arg[1,0,0,0]=(-0.303653386851)*x[0]**o+(-0.435866753659)*x[0]+(-0.419078175082)*x[1]**o+(-0.907595244917)*x[1]+(0.442156528837)*x[2]**o+(-0.76481785523)*x[2]
        arg[1,0,0,1]=(0.879454551618)*x[0]**o+(-0.530789858588)*x[0]+(0.942203776542)*x[1]**o+(0.864362935436)*x[1]+(-0.252832164588)*x[2]**o+(0.00758252581015)*x[2]
        arg[1,0,1,0]=(0.522433142898)*x[0]**o+(0.564002257569)*x[0]+(0.149811689377)*x[1]**o+(-0.757726948165)*x[1]+(-0.767052869719)*x[2]**o+(0.743541806001)*x[2]
        arg[1,0,1,1]=(0.535612198186)*x[0]**o+(0.0874480088401)*x[0]+(0.913151292296)*x[1]**o+(-0.495024655308)*x[1]+(-0.42308905182)*x[2]**o+(-0.983623551153)*x[2]
        arg[1,0,2,0]=(0.587221456352)*x[0]**o+(0.0384438535009)*x[0]+(0.301554087927)*x[1]**o+(-0.944074061806)*x[1]+(0.808195432991)*x[2]**o+(-0.46545743768)*x[2]
        arg[1,0,2,1]=(0.136251169669)*x[0]**o+(0.772190271875)*x[0]+(-0.490981224022)*x[1]**o+(-0.191915588448)*x[1]+(0.792617864405)*x[2]**o+(0.139425359773)*x[2]
        arg[1,1,0,0]=(0.0440493146329)*x[0]**o+(-0.948951787109)*x[0]+(-0.564104026105)*x[1]**o+(0.177457883377)*x[1]+(0.578169345377)*x[2]**o+(0.35574407049)*x[2]
        arg[1,1,0,1]=(-0.740305651819)*x[0]**o+(-0.340976006932)*x[0]+(-0.657257828033)*x[1]**o+(-0.456240473558)*x[1]+(0.526236681039)*x[2]**o+(0.804762801115)*x[2]
        arg[1,1,1,0]=(-0.753480536304)*x[0]**o+(0.424080516297)*x[0]+(-0.97529611006)*x[1]**o+(-0.511450162624)*x[1]+(-0.248982226385)*x[2]**o+(0.497803766242)*x[2]
        arg[1,1,1,1]=(-0.273280180629)*x[0]**o+(0.798945176097)*x[0]+(0.552152546131)*x[1]**o+(0.737594404744)*x[1]+(0.0339269287984)*x[2]**o+(-0.773791176515)*x[2]
        arg[1,1,2,0]=(0.553404075378)*x[0]**o+(-0.630969888585)*x[0]+(0.455072287913)*x[1]**o+(0.145928503781)*x[1]+(0.632480470504)*x[2]**o+(0.363815684296)*x[2]
        arg[1,1,2,1]=(-0.426756293656)*x[0]**o+(-0.449867410734)*x[0]+(-0.875891527959)*x[1]**o+(-0.125243159763)*x[1]+(0.887344501669)*x[2]**o+(-0.0569005112264)*x[2]
        arg[1,2,0,0]=(0.762313320036)*x[0]**o+(-0.514266376348)*x[0]+(-0.915248032301)*x[1]**o+(0.50418081892)*x[1]+(-0.867662577318)*x[2]**o+(0.731635970592)*x[2]
        arg[1,2,0,1]=(-0.0866106252969)*x[0]**o+(0.063250309716)*x[0]+(0.598723588976)*x[1]**o+(0.317235331686)*x[1]+(-0.412892601258)*x[2]**o+(0.771147670726)*x[2]
        arg[1,2,1,0]=(0.57639148624)*x[0]**o+(0.773916349481)*x[0]+(0.401319543026)*x[1]**o+(-0.510387398986)*x[1]+(-0.796773375436)*x[2]**o+(-0.765544756429)*x[2]
        arg[1,2,1,1]=(-0.686952109389)*x[0]**o+(0.428892331428)*x[0]+(-0.139971762047)*x[1]**o+(-0.747581454837)*x[1]+(0.152196734461)*x[2]**o+(-0.778424993847)*x[2]
        arg[1,2,2,0]=(0.729199771391)*x[0]**o+(0.185870258363)*x[0]+(0.996198489286)*x[1]**o+(0.0343703739018)*x[1]+(-0.6637833208)*x[2]**o+(-0.738510210409)*x[2]
        arg[1,2,2,1]=(0.705773326438)*x[0]**o+(-0.633406525448)*x[0]+(0.252620923822)*x[1]**o+(-0.760785834478)*x[1]+(0.995411456622)*x[2]**o+(-0.365718663741)*x[2]
        arg[1,3,0,0]=(0.826142335301)*x[0]**o+(-0.869335317945)*x[0]+(0.660360312658)*x[1]**o+(0.96157015332)*x[1]+(-0.268246324116)*x[2]**o+(-0.575288823295)*x[2]
        arg[1,3,0,1]=(0.620117943072)*x[0]**o+(0.51267609202)*x[0]+(-0.882468485772)*x[1]**o+(0.771021246475)*x[1]+(0.398548927211)*x[2]**o+(0.275983192446)*x[2]
        arg[1,3,1,0]=(0.716523839796)*x[0]**o+(-0.0868143348547)*x[0]+(-0.456384710114)*x[1]**o+(-0.646489063979)*x[1]+(0.594500214771)*x[2]**o+(-0.299956296467)*x[2]
        arg[1,3,1,1]=(-0.453257208906)*x[0]**o+(0.583260425119)*x[0]+(0.263463648634)*x[1]**o+(0.493693446665)*x[1]+(-0.529884773858)*x[2]**o+(0.220733486264)*x[2]
        arg[1,3,2,0]=(0.669052014654)*x[0]**o+(-0.247902780475)*x[0]+(0.44071786167)*x[1]**o+(-0.21199653848)*x[1]+(-0.0713603448391)*x[2]**o+(-0.66681439981)*x[2]
        arg[1,3,2,1]=(-0.341533813758)*x[0]**o+(0.770581995272)*x[0]+(0.03376913861)*x[1]**o+(0.000270719073091)*x[1]+(-0.774936686899)*x[2]**o+(-0.0196300167016)*x[2]
        arg[1,4,0,0]=(-0.869695667108)*x[0]**o+(0.515676789402)*x[0]+(0.67731822534)*x[1]**o+(0.126887846346)*x[1]+(0.843200966455)*x[2]**o+(-0.59930363301)*x[2]
        arg[1,4,0,1]=(0.909369352728)*x[0]**o+(-0.110828876636)*x[0]+(-0.65561953478)*x[1]**o+(0.598491879954)*x[1]+(-0.557567770163)*x[2]**o+(-0.391108926502)*x[2]
        arg[1,4,1,0]=(0.320646291934)*x[0]**o+(0.164062007697)*x[0]+(-0.546323998218)*x[1]**o+(0.72651711346)*x[1]+(-0.613496738898)*x[2]**o+(0.430761082137)*x[2]
        arg[1,4,1,1]=(-0.721651561555)*x[0]**o+(-0.0647310624509)*x[0]+(-0.360525345485)*x[1]**o+(0.408239277468)*x[1]+(0.600551705094)*x[2]**o+(0.506237401392)*x[2]
        arg[1,4,2,0]=(-0.391702965605)*x[0]**o+(-0.574470582913)*x[0]+(-0.729724609696)*x[1]**o+(0.91381180302)*x[1]+(-0.544328104929)*x[2]**o+(-0.516992534607)*x[2]
        arg[1,4,2,1]=(-0.0409666962711)*x[0]**o+(-0.524913112823)*x[0]+(-0.677030607856)*x[1]**o+(-0.530859496807)*x[1]+(0.137890059737)*x[2]**o+(0.0669856870322)*x[2]
        arg[2,0,0,0]=(0.10369318023)*x[0]**o+(0.345691611083)*x[0]+(-0.222795950561)*x[1]**o+(-0.797690270935)*x[1]+(-0.703457203234)*x[2]**o+(-0.751193209871)*x[2]
        arg[2,0,0,1]=(0.166457278787)*x[0]**o+(-0.0304766240065)*x[0]+(0.159987374581)*x[1]**o+(0.828012214748)*x[1]+(0.0670696566332)*x[2]**o+(-0.498252024306)*x[2]
        arg[2,0,1,0]=(0.901095992383)*x[0]**o+(0.400296735981)*x[0]+(0.394687295672)*x[1]**o+(-0.986496611224)*x[1]+(-0.768945602918)*x[2]**o+(-0.782762026735)*x[2]
        arg[2,0,1,1]=(0.47969549703)*x[0]**o+(0.981251232699)*x[0]+(-0.69807700503)*x[1]**o+(-0.813549271424)*x[1]+(-0.135970806025)*x[2]**o+(0.288738861745)*x[2]
        arg[2,0,2,0]=(0.315262113284)*x[0]**o+(0.0380265323361)*x[0]+(0.259894600802)*x[1]**o+(0.538712801332)*x[1]+(-0.385783597257)*x[2]**o+(0.666670257561)*x[2]
        arg[2,0,2,1]=(-0.756715539091)*x[0]**o+(-0.823809574081)*x[0]+(-0.283842954238)*x[1]**o+(-0.640166311159)*x[1]+(-0.39947011053)*x[2]**o+(-0.393543798757)*x[2]
        arg[2,1,0,0]=(-0.292033523499)*x[0]**o+(-0.940246412652)*x[0]+(-0.327571437608)*x[1]**o+(0.837077706655)*x[1]+(0.446471499818)*x[2]**o+(-0.747249149016)*x[2]
        arg[2,1,0,1]=(-0.867504469067)*x[0]**o+(0.635684140727)*x[0]+(0.488043128661)*x[1]**o+(-0.163389002402)*x[1]+(0.0440527250714)*x[2]**o+(0.605766722432)*x[2]
        arg[2,1,1,0]=(-0.0622304802635)*x[0]**o+(0.607299318698)*x[0]+(0.974933906045)*x[1]**o+(0.0632442428777)*x[1]+(0.0672808299851)*x[2]**o+(-0.555700803374)*x[2]
        arg[2,1,1,1]=(-0.350438453721)*x[0]**o+(-0.826289294567)*x[0]+(-0.473764081208)*x[1]**o+(-0.247804269494)*x[1]+(0.834301955859)*x[2]**o+(0.443058363661)*x[2]
        arg[2,1,2,0]=(0.5046026606)*x[0]**o+(0.305502438052)*x[0]+(-0.201816724695)*x[1]**o+(-0.101544576856)*x[1]+(0.729817329051)*x[2]**o+(-0.851479834334)*x[2]
        arg[2,1,2,1]=(0.917039424926)*x[0]**o+(0.260565966637)*x[0]+(-0.595846674057)*x[1]**o+(0.654773238108)*x[1]+(0.544664592565)*x[2]**o+(0.363343885358)*x[2]
        arg[2,2,0,0]=(0.243879200917)*x[0]**o+(0.335612613868)*x[0]+(-0.287066716076)*x[1]**o+(0.506206687739)*x[1]+(-0.902513036448)*x[2]**o+(-0.427017746025)*x[2]
        arg[2,2,0,1]=(0.965102871162)*x[0]**o+(0.161476912964)*x[0]+(0.97007160306)*x[1]**o+(0.870104939402)*x[1]+(0.552404991536)*x[2]**o+(-0.319082395552)*x[2]
        arg[2,2,1,0]=(0.17611459545)*x[0]**o+(-0.457419791602)*x[0]+(-0.00217752259743)*x[1]**o+(-0.632177103931)*x[1]+(0.521123797224)*x[2]**o+(0.427935307605)*x[2]
        arg[2,2,1,1]=(-0.993298850316)*x[0]**o+(0.277763727112)*x[0]+(0.849464768611)*x[1]**o+(0.890919639607)*x[1]+(0.01030894969)*x[2]**o+(-0.76051772768)*x[2]
        arg[2,2,2,0]=(-0.622887788117)*x[0]**o+(0.40187802561)*x[0]+(0.836685367934)*x[1]**o+(0.700379534201)*x[1]+(0.683627950354)*x[2]**o+(0.445351804296)*x[2]
        arg[2,2,2,1]=(0.124965536288)*x[0]**o+(-0.198462168996)*x[0]+(-0.657084126986)*x[1]**o+(-0.420469996385)*x[1]+(0.237164091597)*x[2]**o+(0.161926160295)*x[2]
        arg[2,3,0,0]=(-0.622682941897)*x[0]**o+(-0.6279729398)*x[0]+(0.189086870792)*x[1]**o+(-0.350822257394)*x[1]+(-0.505481371359)*x[2]**o+(0.812944216714)*x[2]
        arg[2,3,0,1]=(-0.246704719057)*x[0]**o+(0.513716384141)*x[0]+(-0.264914192193)*x[1]**o+(-0.32984984663)*x[1]+(-0.80409813849)*x[2]**o+(-0.825568736974)*x[2]
        arg[2,3,1,0]=(-0.534782640399)*x[0]**o+(-0.917823259622)*x[0]+(0.691827074405)*x[1]**o+(-0.818629872154)*x[1]+(0.870046992385)*x[2]**o+(0.00423218787803)*x[2]
        arg[2,3,1,1]=(0.633068676613)*x[0]**o+(-0.895613811402)*x[0]+(0.62242069294)*x[1]**o+(-0.272136050955)*x[1]+(-0.199008776357)*x[2]**o+(0.836119124508)*x[2]
        arg[2,3,2,0]=(0.335884106464)*x[0]**o+(-0.857917873801)*x[0]+(-0.42653278071)*x[1]**o+(0.375200781448)*x[1]+(0.378427453722)*x[2]**o+(0.398603623445)*x[2]
        arg[2,3,2,1]=(0.0602499604703)*x[0]**o+(-0.921459358089)*x[0]+(0.558180675902)*x[1]**o+(0.0863734870743)*x[1]+(0.215167095483)*x[2]**o+(-0.37908830738)*x[2]
        arg[2,4,0,0]=(0.866321987007)*x[0]**o+(0.81950553744)*x[0]+(0.0422552872889)*x[1]**o+(0.369328458453)*x[1]+(0.0547909452885)*x[2]**o+(-0.396367133883)*x[2]
        arg[2,4,0,1]=(0.438824337571)*x[0]**o+(0.211435787811)*x[0]+(-0.801273340861)*x[1]**o+(-0.181654271172)*x[1]+(0.374171148243)*x[2]**o+(0.945633108881)*x[2]
        arg[2,4,1,0]=(0.588044951473)*x[0]**o+(0.401384824924)*x[0]+(-0.191996429616)*x[1]**o+(0.0472579697166)*x[1]+(0.470962958517)*x[2]**o+(0.52718875619)*x[2]
        arg[2,4,1,1]=(0.371613549998)*x[0]**o+(-0.414998926249)*x[0]+(-0.762013182749)*x[1]**o+(0.796144230249)*x[1]+(-0.555041053148)*x[2]**o+(-0.595298275916)*x[2]
        arg[2,4,2,0]=(-0.992527136862)*x[0]**o+(-0.995041267483)*x[0]+(-0.730305990844)*x[1]**o+(-0.52470798566)*x[1]+(-0.938945944067)*x[2]**o+(0.33396135548)*x[2]
        arg[2,4,2,1]=(0.0333921786723)*x[0]**o+(0.707990374515)*x[0]+(-0.106667413755)*x[1]**o+(0.282989655082)*x[1]+(-0.152535982788)*x[2]**o+(-0.75130096727)*x[2]
        arg[3,0,0,0]=(-0.822675771493)*x[0]**o+(0.726468488157)*x[0]+(0.972567560814)*x[1]**o+(0.308266449093)*x[1]+(0.71069864654)*x[2]**o+(-0.683195491271)*x[2]
        arg[3,0,0,1]=(-0.505155072199)*x[0]**o+(0.339013729552)*x[0]+(-0.81565931012)*x[1]**o+(0.362458680494)*x[1]+(0.960221320064)*x[2]**o+(-0.17234580875)*x[2]
        arg[3,0,1,0]=(0.67972657926)*x[0]**o+(0.143085543693)*x[0]+(-0.699516563264)*x[1]**o+(-0.892226133751)*x[1]+(-0.898541541049)*x[2]**o+(-0.736249681714)*x[2]
        arg[3,0,1,1]=(-0.144423319215)*x[0]**o+(-0.509713525676)*x[0]+(0.347584762861)*x[1]**o+(0.666169972794)*x[1]+(0.691734080163)*x[2]**o+(0.9338062148)*x[2]
        arg[3,0,2,0]=(-0.47323843589)*x[0]**o+(0.0819655059363)*x[0]+(0.0863030211133)*x[1]**o+(0.127803089307)*x[1]+(-0.155852489756)*x[2]**o+(-0.00470398386987)*x[2]
        arg[3,0,2,1]=(0.372452726232)*x[0]**o+(0.133808609857)*x[0]+(0.962945642209)*x[1]**o+(0.0446609213929)*x[1]+(-0.446207951293)*x[2]**o+(0.0400733439868)*x[2]
        arg[3,1,0,0]=(-0.668575734235)*x[0]**o+(0.905913572459)*x[0]+(-0.321327880802)*x[1]**o+(0.0263209456188)*x[1]+(-0.0355325780368)*x[2]**o+(-0.624130395105)*x[2]
        arg[3,1,0,1]=(0.859810828483)*x[0]**o+(-0.235372125226)*x[0]+(-0.16715244612)*x[1]**o+(-0.793399872713)*x[1]+(0.303973795556)*x[2]**o+(0.176740655638)*x[2]
        arg[3,1,1,0]=(0.736448874258)*x[0]**o+(-0.448956585466)*x[0]+(0.782095902992)*x[1]**o+(0.796524196866)*x[1]+(-0.732812289541)*x[2]**o+(0.687089667291)*x[2]
        arg[3,1,1,1]=(-0.797708068194)*x[0]**o+(-0.691308116296)*x[0]+(-0.644065410151)*x[1]**o+(-0.929624672017)*x[1]+(0.265040940279)*x[2]**o+(0.424929324889)*x[2]
        arg[3,1,2,0]=(0.555032482777)*x[0]**o+(-0.564617314931)*x[0]+(-0.612721853869)*x[1]**o+(-0.048466699593)*x[1]+(0.753265037897)*x[2]**o+(0.913086239874)*x[2]
        arg[3,1,2,1]=(0.580881807771)*x[0]**o+(0.970973128604)*x[0]+(-0.603610563753)*x[1]**o+(0.979985814438)*x[1]+(-0.639531987694)*x[2]**o+(-0.773094138073)*x[2]
        arg[3,2,0,0]=(0.574306840544)*x[0]**o+(0.449094769988)*x[0]+(-0.406303076614)*x[1]**o+(-0.776648781906)*x[1]+(-0.59829476647)*x[2]**o+(0.257632446666)*x[2]
        arg[3,2,0,1]=(0.689106127051)*x[0]**o+(0.717184851708)*x[0]+(-0.495581943615)*x[1]**o+(-0.30329139788)*x[1]+(0.901753128535)*x[2]**o+(0.48496505727)*x[2]
        arg[3,2,1,0]=(0.252050902256)*x[0]**o+(-0.0414281706943)*x[0]+(-0.284125627654)*x[1]**o+(0.477450554269)*x[1]+(0.849202242148)*x[2]**o+(0.211826565889)*x[2]
        arg[3,2,1,1]=(-0.319900673571)*x[0]**o+(0.965993723739)*x[0]+(0.671360092314)*x[1]**o+(-0.652805407934)*x[1]+(-0.0352316646853)*x[2]**o+(-0.442493499306)*x[2]
        arg[3,2,2,0]=(0.525577660439)*x[0]**o+(0.660882378915)*x[0]+(-0.419826738105)*x[1]**o+(-0.798251889077)*x[1]+(-0.430382393461)*x[2]**o+(0.714357242496)*x[2]
        arg[3,2,2,1]=(-0.200634944034)*x[0]**o+(0.211086366034)*x[0]+(0.116289218188)*x[1]**o+(-0.785792860501)*x[1]+(0.763320327438)*x[2]**o+(-0.457355826175)*x[2]
        arg[3,3,0,0]=(-0.890873783782)*x[0]**o+(0.322610697657)*x[0]+(-0.351437390332)*x[1]**o+(-0.720698169708)*x[1]+(-0.664444440194)*x[2]**o+(0.274865394829)*x[2]
        arg[3,3,0,1]=(-0.384885652342)*x[0]**o+(-0.409799972422)*x[0]+(0.679315041587)*x[1]**o+(0.0859788488759)*x[1]+(-0.136173937432)*x[2]**o+(-0.136591429212)*x[2]
        arg[3,3,1,0]=(0.147934741321)*x[0]**o+(0.246941956666)*x[0]+(0.696831123708)*x[1]**o+(-0.905966322958)*x[1]+(0.672705695255)*x[2]**o+(-0.776017262395)*x[2]
        arg[3,3,1,1]=(0.933982654679)*x[0]**o+(-0.367246348547)*x[0]+(-0.254076834401)*x[1]**o+(0.148617981234)*x[1]+(-0.486881004358)*x[2]**o+(0.636371570082)*x[2]
        arg[3,3,2,0]=(-0.536125392071)*x[0]**o+(-0.663380779479)*x[0]+(0.0655694689011)*x[1]**o+(0.0390002326894)*x[1]+(-0.731730129742)*x[2]**o+(-0.578791485067)*x[2]
        arg[3,3,2,1]=(0.705873902409)*x[0]**o+(-0.0444303506955)*x[0]+(-0.978519191225)*x[1]**o+(-0.0280878186958)*x[1]+(0.182378726121)*x[2]**o+(-0.555345919696)*x[2]
        arg[3,4,0,0]=(0.945945862395)*x[0]**o+(-0.448487950632)*x[0]+(-0.477570573229)*x[1]**o+(0.313871838624)*x[1]+(0.237944220287)*x[2]**o+(-0.218095130482)*x[2]
        arg[3,4,0,1]=(-0.758647277269)*x[0]**o+(-0.550604667208)*x[0]+(-0.262944399053)*x[1]**o+(0.994302652859)*x[1]+(0.363440310008)*x[2]**o+(-0.127863970613)*x[2]
        arg[3,4,1,0]=(0.667202433965)*x[0]**o+(0.954901068821)*x[0]+(0.719745510318)*x[1]**o+(-0.763882742007)*x[1]+(0.880927476123)*x[2]**o+(0.102034630893)*x[2]
        arg[3,4,1,1]=(-0.684644024456)*x[0]**o+(0.88037275093)*x[0]+(-0.772213766289)*x[1]**o+(-0.702615214094)*x[1]+(-0.25661708053)*x[2]**o+(0.126779040694)*x[2]
        arg[3,4,2,0]=(-0.589611385303)*x[0]**o+(0.638712635252)*x[0]+(0.745450741856)*x[1]**o+(0.399494092815)*x[1]+(-0.717293809038)*x[2]**o+(0.305662737751)*x[2]
        arg[3,4,2,1]=(0.0572426997796)*x[0]**o+(-0.764600144578)*x[0]+(0.482545219565)*x[1]**o+(0.525883069411)*x[1]+(-0.612224327195)*x[2]**o+(0.354344820256)*x[2]
        ref[0,0,0,0]=(0.952750275738)*x_ref[0]**o+(0.971489204087)*x_ref[0]+(-0.320706487584)*x_ref[1]**o+(-0.853260471578)*x_ref[1]+(-0.696434862068)*x_ref[2]**o+(0.641824397433)*x_ref[2]
        ref[0,0,0,1]=(0.115484192786)*x_ref[0]**o+(0.147813687309)*x_ref[0]+(-0.998010548186)*x_ref[1]**o+(0.0274712088882)*x_ref[1]+(0.136937971489)*x_ref[2]**o+(-0.657986135444)*x_ref[2]
        ref[0,0,1,0]=(0.805857504779)*x_ref[0]**o+(0.196278916938)*x_ref[0]+(-0.969340785263)*x_ref[1]**o+(-0.543890174888)*x_ref[1]+(0.49118423357)*x_ref[2]**o+(-0.883493800593)*x_ref[2]
        ref[0,0,1,1]=(-0.808802136474)*x_ref[0]**o+(-0.866039395649)*x_ref[0]+(0.0339207025941)*x_ref[1]**o+(-0.441368773534)*x_ref[1]+(0.673688211653)*x_ref[2]**o+(0.917894389398)*x_ref[2]
        ref[0,0,2,0]=(0.0459902210078)*x_ref[0]**o+(0.862332666528)*x_ref[0]+(-0.547792306389)*x_ref[1]**o+(0.722176431732)*x_ref[1]+(-0.511278989032)*x_ref[2]**o+(-0.692775336577)*x_ref[2]
        ref[0,0,2,1]=(-0.641904406513)*x_ref[0]**o+(-0.854450656063)*x_ref[0]+(0.81138894849)*x_ref[1]**o+(0.328886625074)*x_ref[1]+(0.400935831363)*x_ref[2]**o+(-0.25184059766)*x_ref[2]
        ref[0,1,0,0]=(0.69922840942)*x_ref[0]**o+(-0.798085508443)*x_ref[0]+(-0.85191383431)*x_ref[1]**o+(-0.162940829331)*x_ref[1]+(0.504020642931)*x_ref[2]**o+(0.0950957436173)*x_ref[2]
        ref[0,1,0,1]=(0.28901429037)*x_ref[0]**o+(-0.624508745729)*x_ref[0]+(0.921288077478)*x_ref[1]**o+(-0.769207901119)*x_ref[1]+(0.0852374599763)*x_ref[2]**o+(-0.402583033986)*x_ref[2]
        ref[0,1,1,0]=(0.715817282215)*x_ref[0]**o+(0.68735256665)*x_ref[0]+(0.610212058649)*x_ref[1]**o+(0.53560824609)*x_ref[1]+(-0.238908897264)*x_ref[2]**o+(0.619799158474)*x_ref[2]
        ref[0,1,1,1]=(-0.371797414191)*x_ref[0]**o+(0.106263350514)*x_ref[0]+(0.583768355362)*x_ref[1]**o+(0.818797667369)*x_ref[1]+(0.0567571152493)*x_ref[2]**o+(-0.242244867169)*x_ref[2]
        ref[0,1,2,0]=(0.190480659729)*x_ref[0]**o+(-0.632415604771)*x_ref[0]+(0.252513826138)*x_ref[1]**o+(0.874732840469)*x_ref[1]+(-0.367769326078)*x_ref[2]**o+(0.364453292404)*x_ref[2]
        ref[0,1,2,1]=(-0.704507136581)*x_ref[0]**o+(-0.614211737547)*x_ref[0]+(0.533383507313)*x_ref[1]**o+(-0.940007583478)*x_ref[1]+(0.558260002746)*x_ref[2]**o+(0.742565692751)*x_ref[2]
        ref[0,2,0,0]=(0.132581156766)*x_ref[0]**o+(0.671793183183)*x_ref[0]+(-0.308255110769)*x_ref[1]**o+(-0.956002957061)*x_ref[1]+(0.300091541301)*x_ref[2]**o+(-0.714260746791)*x_ref[2]
        ref[0,2,0,1]=(-0.836648080859)*x_ref[0]**o+(0.123813861017)*x_ref[0]+(0.866569934894)*x_ref[1]**o+(-0.957824572351)*x_ref[1]+(-0.904478410952)*x_ref[2]**o+(0.691673758521)*x_ref[2]
        ref[0,2,1,0]=(0.411624721878)*x_ref[0]**o+(0.594054722352)*x_ref[0]+(0.150177701135)*x_ref[1]**o+(0.945855261084)*x_ref[1]+(0.295930158817)*x_ref[2]**o+(-0.607980954424)*x_ref[2]
        ref[0,2,1,1]=(-0.0777741308373)*x_ref[0]**o+(0.697501295655)*x_ref[0]+(0.981797797329)*x_ref[1]**o+(-0.671827077949)*x_ref[1]+(0.292310146229)*x_ref[2]**o+(0.484652357857)*x_ref[2]
        ref[0,2,2,0]=(0.194910469807)*x_ref[0]**o+(-0.70008515623)*x_ref[0]+(0.192808336639)*x_ref[1]**o+(-0.962764477371)*x_ref[1]+(-0.34292240753)*x_ref[2]**o+(0.525470829202)*x_ref[2]
        ref[0,2,2,1]=(0.850368420228)*x_ref[0]**o+(-0.516355386368)*x_ref[0]+(0.395764367873)*x_ref[1]**o+(0.320830752607)*x_ref[1]+(-0.26288679044)*x_ref[2]**o+(0.174457723191)*x_ref[2]
        ref[0,3,0,0]=(-0.713987556845)*x_ref[0]**o+(0.0391066248253)*x_ref[0]+(-0.63911733479)*x_ref[1]**o+(-0.119953032121)*x_ref[1]+(0.697879836447)*x_ref[2]**o+(-0.474957057702)*x_ref[2]
        ref[0,3,0,1]=(-0.128614897423)*x_ref[0]**o+(0.7166036403)*x_ref[0]+(0.678415978447)*x_ref[1]**o+(-0.191453169618)*x_ref[1]+(-0.755261361956)*x_ref[2]**o+(-0.839956004196)*x_ref[2]
        ref[0,3,1,0]=(0.258158740934)*x_ref[0]**o+(-0.412199346967)*x_ref[0]+(-0.781106341272)*x_ref[1]**o+(-0.561162823101)*x_ref[1]+(-0.59549864377)*x_ref[2]**o+(-0.748038415721)*x_ref[2]
        ref[0,3,1,1]=(0.68950222305)*x_ref[0]**o+(-0.904027774445)*x_ref[0]+(0.769744453441)*x_ref[1]**o+(-0.265549152752)*x_ref[1]+(-0.375978837131)*x_ref[2]**o+(-0.387010981761)*x_ref[2]
        ref[0,3,2,0]=(0.46852681319)*x_ref[0]**o+(-0.887650258615)*x_ref[0]+(-0.780275624319)*x_ref[1]**o+(0.527897372599)*x_ref[1]+(0.889281778411)*x_ref[2]**o+(0.82538563877)*x_ref[2]
        ref[0,3,2,1]=(-0.817514448184)*x_ref[0]**o+(-0.230190768519)*x_ref[0]+(-0.101550849344)*x_ref[1]**o+(0.1654885857)*x_ref[1]+(-0.231317749498)*x_ref[2]**o+(0.580153306395)*x_ref[2]
        ref[0,4,0,0]=(0.394993007242)*x_ref[0]**o+(0.305948764395)*x_ref[0]+(-0.66853441368)*x_ref[1]**o+(0.279027498094)*x_ref[1]+(0.161366934709)*x_ref[2]**o+(-0.742002471552)*x_ref[2]
        ref[0,4,0,1]=(-0.661214709471)*x_ref[0]**o+(0.548880317935)*x_ref[0]+(0.472682748503)*x_ref[1]**o+(0.266797505699)*x_ref[1]+(0.405618413683)*x_ref[2]**o+(-0.0539677542667)*x_ref[2]
        ref[0,4,1,0]=(-0.713892424429)*x_ref[0]**o+(0.568075831592)*x_ref[0]+(-0.956180895599)*x_ref[1]**o+(0.63618957695)*x_ref[1]+(0.983696160047)*x_ref[2]**o+(-0.308157929002)*x_ref[2]
        ref[0,4,1,1]=(-0.666582919353)*x_ref[0]**o+(0.0816884254484)*x_ref[0]+(0.190810048896)*x_ref[1]**o+(0.482322905132)*x_ref[1]+(-0.576597070954)*x_ref[2]**o+(-0.722381611553)*x_ref[2]
        ref[0,4,2,0]=(-0.0229650731041)*x_ref[0]**o+(-0.0867816754979)*x_ref[0]+(0.0905042102365)*x_ref[1]**o+(-0.994963424765)*x_ref[1]+(0.74504757815)*x_ref[2]**o+(0.71462549644)*x_ref[2]
        ref[0,4,2,1]=(-0.0464278496147)*x_ref[0]**o+(0.122501026996)*x_ref[0]+(-0.049450141901)*x_ref[1]**o+(-0.388739464086)*x_ref[1]+(-0.151389885067)*x_ref[2]**o+(-0.0316948088803)*x_ref[2]
        ref[1,0,0,0]=(-0.303653386851)*x_ref[0]**o+(-0.435866753659)*x_ref[0]+(-0.419078175082)*x_ref[1]**o+(-0.907595244917)*x_ref[1]+(0.442156528837)*x_ref[2]**o+(-0.76481785523)*x_ref[2]
        ref[1,0,0,1]=(0.879454551618)*x_ref[0]**o+(-0.530789858588)*x_ref[0]+(0.942203776542)*x_ref[1]**o+(0.864362935436)*x_ref[1]+(-0.252832164588)*x_ref[2]**o+(0.00758252581015)*x_ref[2]
        ref[1,0,1,0]=(0.522433142898)*x_ref[0]**o+(0.564002257569)*x_ref[0]+(0.149811689377)*x_ref[1]**o+(-0.757726948165)*x_ref[1]+(-0.767052869719)*x_ref[2]**o+(0.743541806001)*x_ref[2]
        ref[1,0,1,1]=(0.535612198186)*x_ref[0]**o+(0.0874480088401)*x_ref[0]+(0.913151292296)*x_ref[1]**o+(-0.495024655308)*x_ref[1]+(-0.42308905182)*x_ref[2]**o+(-0.983623551153)*x_ref[2]
        ref[1,0,2,0]=(0.587221456352)*x_ref[0]**o+(0.0384438535009)*x_ref[0]+(0.301554087927)*x_ref[1]**o+(-0.944074061806)*x_ref[1]+(0.808195432991)*x_ref[2]**o+(-0.46545743768)*x_ref[2]
        ref[1,0,2,1]=(0.136251169669)*x_ref[0]**o+(0.772190271875)*x_ref[0]+(-0.490981224022)*x_ref[1]**o+(-0.191915588448)*x_ref[1]+(0.792617864405)*x_ref[2]**o+(0.139425359773)*x_ref[2]
        ref[1,1,0,0]=(0.0440493146329)*x_ref[0]**o+(-0.948951787109)*x_ref[0]+(-0.564104026105)*x_ref[1]**o+(0.177457883377)*x_ref[1]+(0.578169345377)*x_ref[2]**o+(0.35574407049)*x_ref[2]
        ref[1,1,0,1]=(-0.740305651819)*x_ref[0]**o+(-0.340976006932)*x_ref[0]+(-0.657257828033)*x_ref[1]**o+(-0.456240473558)*x_ref[1]+(0.526236681039)*x_ref[2]**o+(0.804762801115)*x_ref[2]
        ref[1,1,1,0]=(-0.753480536304)*x_ref[0]**o+(0.424080516297)*x_ref[0]+(-0.97529611006)*x_ref[1]**o+(-0.511450162624)*x_ref[1]+(-0.248982226385)*x_ref[2]**o+(0.497803766242)*x_ref[2]
        ref[1,1,1,1]=(-0.273280180629)*x_ref[0]**o+(0.798945176097)*x_ref[0]+(0.552152546131)*x_ref[1]**o+(0.737594404744)*x_ref[1]+(0.0339269287984)*x_ref[2]**o+(-0.773791176515)*x_ref[2]
        ref[1,1,2,0]=(0.553404075378)*x_ref[0]**o+(-0.630969888585)*x_ref[0]+(0.455072287913)*x_ref[1]**o+(0.145928503781)*x_ref[1]+(0.632480470504)*x_ref[2]**o+(0.363815684296)*x_ref[2]
        ref[1,1,2,1]=(-0.426756293656)*x_ref[0]**o+(-0.449867410734)*x_ref[0]+(-0.875891527959)*x_ref[1]**o+(-0.125243159763)*x_ref[1]+(0.887344501669)*x_ref[2]**o+(-0.0569005112264)*x_ref[2]
        ref[1,2,0,0]=(0.762313320036)*x_ref[0]**o+(-0.514266376348)*x_ref[0]+(-0.915248032301)*x_ref[1]**o+(0.50418081892)*x_ref[1]+(-0.867662577318)*x_ref[2]**o+(0.731635970592)*x_ref[2]
        ref[1,2,0,1]=(-0.0866106252969)*x_ref[0]**o+(0.063250309716)*x_ref[0]+(0.598723588976)*x_ref[1]**o+(0.317235331686)*x_ref[1]+(-0.412892601258)*x_ref[2]**o+(0.771147670726)*x_ref[2]
        ref[1,2,1,0]=(0.57639148624)*x_ref[0]**o+(0.773916349481)*x_ref[0]+(0.401319543026)*x_ref[1]**o+(-0.510387398986)*x_ref[1]+(-0.796773375436)*x_ref[2]**o+(-0.765544756429)*x_ref[2]
        ref[1,2,1,1]=(-0.686952109389)*x_ref[0]**o+(0.428892331428)*x_ref[0]+(-0.139971762047)*x_ref[1]**o+(-0.747581454837)*x_ref[1]+(0.152196734461)*x_ref[2]**o+(-0.778424993847)*x_ref[2]
        ref[1,2,2,0]=(0.729199771391)*x_ref[0]**o+(0.185870258363)*x_ref[0]+(0.996198489286)*x_ref[1]**o+(0.0343703739018)*x_ref[1]+(-0.6637833208)*x_ref[2]**o+(-0.738510210409)*x_ref[2]
        ref[1,2,2,1]=(0.705773326438)*x_ref[0]**o+(-0.633406525448)*x_ref[0]+(0.252620923822)*x_ref[1]**o+(-0.760785834478)*x_ref[1]+(0.995411456622)*x_ref[2]**o+(-0.365718663741)*x_ref[2]
        ref[1,3,0,0]=(0.826142335301)*x_ref[0]**o+(-0.869335317945)*x_ref[0]+(0.660360312658)*x_ref[1]**o+(0.96157015332)*x_ref[1]+(-0.268246324116)*x_ref[2]**o+(-0.575288823295)*x_ref[2]
        ref[1,3,0,1]=(0.620117943072)*x_ref[0]**o+(0.51267609202)*x_ref[0]+(-0.882468485772)*x_ref[1]**o+(0.771021246475)*x_ref[1]+(0.398548927211)*x_ref[2]**o+(0.275983192446)*x_ref[2]
        ref[1,3,1,0]=(0.716523839796)*x_ref[0]**o+(-0.0868143348547)*x_ref[0]+(-0.456384710114)*x_ref[1]**o+(-0.646489063979)*x_ref[1]+(0.594500214771)*x_ref[2]**o+(-0.299956296467)*x_ref[2]
        ref[1,3,1,1]=(-0.453257208906)*x_ref[0]**o+(0.583260425119)*x_ref[0]+(0.263463648634)*x_ref[1]**o+(0.493693446665)*x_ref[1]+(-0.529884773858)*x_ref[2]**o+(0.220733486264)*x_ref[2]
        ref[1,3,2,0]=(0.669052014654)*x_ref[0]**o+(-0.247902780475)*x_ref[0]+(0.44071786167)*x_ref[1]**o+(-0.21199653848)*x_ref[1]+(-0.0713603448391)*x_ref[2]**o+(-0.66681439981)*x_ref[2]
        ref[1,3,2,1]=(-0.341533813758)*x_ref[0]**o+(0.770581995272)*x_ref[0]+(0.03376913861)*x_ref[1]**o+(0.000270719073091)*x_ref[1]+(-0.774936686899)*x_ref[2]**o+(-0.0196300167016)*x_ref[2]
        ref[1,4,0,0]=(-0.869695667108)*x_ref[0]**o+(0.515676789402)*x_ref[0]+(0.67731822534)*x_ref[1]**o+(0.126887846346)*x_ref[1]+(0.843200966455)*x_ref[2]**o+(-0.59930363301)*x_ref[2]
        ref[1,4,0,1]=(0.909369352728)*x_ref[0]**o+(-0.110828876636)*x_ref[0]+(-0.65561953478)*x_ref[1]**o+(0.598491879954)*x_ref[1]+(-0.557567770163)*x_ref[2]**o+(-0.391108926502)*x_ref[2]
        ref[1,4,1,0]=(0.320646291934)*x_ref[0]**o+(0.164062007697)*x_ref[0]+(-0.546323998218)*x_ref[1]**o+(0.72651711346)*x_ref[1]+(-0.613496738898)*x_ref[2]**o+(0.430761082137)*x_ref[2]
        ref[1,4,1,1]=(-0.721651561555)*x_ref[0]**o+(-0.0647310624509)*x_ref[0]+(-0.360525345485)*x_ref[1]**o+(0.408239277468)*x_ref[1]+(0.600551705094)*x_ref[2]**o+(0.506237401392)*x_ref[2]
        ref[1,4,2,0]=(-0.391702965605)*x_ref[0]**o+(-0.574470582913)*x_ref[0]+(-0.729724609696)*x_ref[1]**o+(0.91381180302)*x_ref[1]+(-0.544328104929)*x_ref[2]**o+(-0.516992534607)*x_ref[2]
        ref[1,4,2,1]=(-0.0409666962711)*x_ref[0]**o+(-0.524913112823)*x_ref[0]+(-0.677030607856)*x_ref[1]**o+(-0.530859496807)*x_ref[1]+(0.137890059737)*x_ref[2]**o+(0.0669856870322)*x_ref[2]
        ref[2,0,0,0]=(0.10369318023)*x_ref[0]**o+(0.345691611083)*x_ref[0]+(-0.222795950561)*x_ref[1]**o+(-0.797690270935)*x_ref[1]+(-0.703457203234)*x_ref[2]**o+(-0.751193209871)*x_ref[2]
        ref[2,0,0,1]=(0.166457278787)*x_ref[0]**o+(-0.0304766240065)*x_ref[0]+(0.159987374581)*x_ref[1]**o+(0.828012214748)*x_ref[1]+(0.0670696566332)*x_ref[2]**o+(-0.498252024306)*x_ref[2]
        ref[2,0,1,0]=(0.901095992383)*x_ref[0]**o+(0.400296735981)*x_ref[0]+(0.394687295672)*x_ref[1]**o+(-0.986496611224)*x_ref[1]+(-0.768945602918)*x_ref[2]**o+(-0.782762026735)*x_ref[2]
        ref[2,0,1,1]=(0.47969549703)*x_ref[0]**o+(0.981251232699)*x_ref[0]+(-0.69807700503)*x_ref[1]**o+(-0.813549271424)*x_ref[1]+(-0.135970806025)*x_ref[2]**o+(0.288738861745)*x_ref[2]
        ref[2,0,2,0]=(0.315262113284)*x_ref[0]**o+(0.0380265323361)*x_ref[0]+(0.259894600802)*x_ref[1]**o+(0.538712801332)*x_ref[1]+(-0.385783597257)*x_ref[2]**o+(0.666670257561)*x_ref[2]
        ref[2,0,2,1]=(-0.756715539091)*x_ref[0]**o+(-0.823809574081)*x_ref[0]+(-0.283842954238)*x_ref[1]**o+(-0.640166311159)*x_ref[1]+(-0.39947011053)*x_ref[2]**o+(-0.393543798757)*x_ref[2]
        ref[2,1,0,0]=(-0.292033523499)*x_ref[0]**o+(-0.940246412652)*x_ref[0]+(-0.327571437608)*x_ref[1]**o+(0.837077706655)*x_ref[1]+(0.446471499818)*x_ref[2]**o+(-0.747249149016)*x_ref[2]
        ref[2,1,0,1]=(-0.867504469067)*x_ref[0]**o+(0.635684140727)*x_ref[0]+(0.488043128661)*x_ref[1]**o+(-0.163389002402)*x_ref[1]+(0.0440527250714)*x_ref[2]**o+(0.605766722432)*x_ref[2]
        ref[2,1,1,0]=(-0.0622304802635)*x_ref[0]**o+(0.607299318698)*x_ref[0]+(0.974933906045)*x_ref[1]**o+(0.0632442428777)*x_ref[1]+(0.0672808299851)*x_ref[2]**o+(-0.555700803374)*x_ref[2]
        ref[2,1,1,1]=(-0.350438453721)*x_ref[0]**o+(-0.826289294567)*x_ref[0]+(-0.473764081208)*x_ref[1]**o+(-0.247804269494)*x_ref[1]+(0.834301955859)*x_ref[2]**o+(0.443058363661)*x_ref[2]
        ref[2,1,2,0]=(0.5046026606)*x_ref[0]**o+(0.305502438052)*x_ref[0]+(-0.201816724695)*x_ref[1]**o+(-0.101544576856)*x_ref[1]+(0.729817329051)*x_ref[2]**o+(-0.851479834334)*x_ref[2]
        ref[2,1,2,1]=(0.917039424926)*x_ref[0]**o+(0.260565966637)*x_ref[0]+(-0.595846674057)*x_ref[1]**o+(0.654773238108)*x_ref[1]+(0.544664592565)*x_ref[2]**o+(0.363343885358)*x_ref[2]
        ref[2,2,0,0]=(0.243879200917)*x_ref[0]**o+(0.335612613868)*x_ref[0]+(-0.287066716076)*x_ref[1]**o+(0.506206687739)*x_ref[1]+(-0.902513036448)*x_ref[2]**o+(-0.427017746025)*x_ref[2]
        ref[2,2,0,1]=(0.965102871162)*x_ref[0]**o+(0.161476912964)*x_ref[0]+(0.97007160306)*x_ref[1]**o+(0.870104939402)*x_ref[1]+(0.552404991536)*x_ref[2]**o+(-0.319082395552)*x_ref[2]
        ref[2,2,1,0]=(0.17611459545)*x_ref[0]**o+(-0.457419791602)*x_ref[0]+(-0.00217752259743)*x_ref[1]**o+(-0.632177103931)*x_ref[1]+(0.521123797224)*x_ref[2]**o+(0.427935307605)*x_ref[2]
        ref[2,2,1,1]=(-0.993298850316)*x_ref[0]**o+(0.277763727112)*x_ref[0]+(0.849464768611)*x_ref[1]**o+(0.890919639607)*x_ref[1]+(0.01030894969)*x_ref[2]**o+(-0.76051772768)*x_ref[2]
        ref[2,2,2,0]=(-0.622887788117)*x_ref[0]**o+(0.40187802561)*x_ref[0]+(0.836685367934)*x_ref[1]**o+(0.700379534201)*x_ref[1]+(0.683627950354)*x_ref[2]**o+(0.445351804296)*x_ref[2]
        ref[2,2,2,1]=(0.124965536288)*x_ref[0]**o+(-0.198462168996)*x_ref[0]+(-0.657084126986)*x_ref[1]**o+(-0.420469996385)*x_ref[1]+(0.237164091597)*x_ref[2]**o+(0.161926160295)*x_ref[2]
        ref[2,3,0,0]=(-0.622682941897)*x_ref[0]**o+(-0.6279729398)*x_ref[0]+(0.189086870792)*x_ref[1]**o+(-0.350822257394)*x_ref[1]+(-0.505481371359)*x_ref[2]**o+(0.812944216714)*x_ref[2]
        ref[2,3,0,1]=(-0.246704719057)*x_ref[0]**o+(0.513716384141)*x_ref[0]+(-0.264914192193)*x_ref[1]**o+(-0.32984984663)*x_ref[1]+(-0.80409813849)*x_ref[2]**o+(-0.825568736974)*x_ref[2]
        ref[2,3,1,0]=(-0.534782640399)*x_ref[0]**o+(-0.917823259622)*x_ref[0]+(0.691827074405)*x_ref[1]**o+(-0.818629872154)*x_ref[1]+(0.870046992385)*x_ref[2]**o+(0.00423218787803)*x_ref[2]
        ref[2,3,1,1]=(0.633068676613)*x_ref[0]**o+(-0.895613811402)*x_ref[0]+(0.62242069294)*x_ref[1]**o+(-0.272136050955)*x_ref[1]+(-0.199008776357)*x_ref[2]**o+(0.836119124508)*x_ref[2]
        ref[2,3,2,0]=(0.335884106464)*x_ref[0]**o+(-0.857917873801)*x_ref[0]+(-0.42653278071)*x_ref[1]**o+(0.375200781448)*x_ref[1]+(0.378427453722)*x_ref[2]**o+(0.398603623445)*x_ref[2]
        ref[2,3,2,1]=(0.0602499604703)*x_ref[0]**o+(-0.921459358089)*x_ref[0]+(0.558180675902)*x_ref[1]**o+(0.0863734870743)*x_ref[1]+(0.215167095483)*x_ref[2]**o+(-0.37908830738)*x_ref[2]
        ref[2,4,0,0]=(0.866321987007)*x_ref[0]**o+(0.81950553744)*x_ref[0]+(0.0422552872889)*x_ref[1]**o+(0.369328458453)*x_ref[1]+(0.0547909452885)*x_ref[2]**o+(-0.396367133883)*x_ref[2]
        ref[2,4,0,1]=(0.438824337571)*x_ref[0]**o+(0.211435787811)*x_ref[0]+(-0.801273340861)*x_ref[1]**o+(-0.181654271172)*x_ref[1]+(0.374171148243)*x_ref[2]**o+(0.945633108881)*x_ref[2]
        ref[2,4,1,0]=(0.588044951473)*x_ref[0]**o+(0.401384824924)*x_ref[0]+(-0.191996429616)*x_ref[1]**o+(0.0472579697166)*x_ref[1]+(0.470962958517)*x_ref[2]**o+(0.52718875619)*x_ref[2]
        ref[2,4,1,1]=(0.371613549998)*x_ref[0]**o+(-0.414998926249)*x_ref[0]+(-0.762013182749)*x_ref[1]**o+(0.796144230249)*x_ref[1]+(-0.555041053148)*x_ref[2]**o+(-0.595298275916)*x_ref[2]
        ref[2,4,2,0]=(-0.992527136862)*x_ref[0]**o+(-0.995041267483)*x_ref[0]+(-0.730305990844)*x_ref[1]**o+(-0.52470798566)*x_ref[1]+(-0.938945944067)*x_ref[2]**o+(0.33396135548)*x_ref[2]
        ref[2,4,2,1]=(0.0333921786723)*x_ref[0]**o+(0.707990374515)*x_ref[0]+(-0.106667413755)*x_ref[1]**o+(0.282989655082)*x_ref[1]+(-0.152535982788)*x_ref[2]**o+(-0.75130096727)*x_ref[2]
        ref[3,0,0,0]=(-0.822675771493)*x_ref[0]**o+(0.726468488157)*x_ref[0]+(0.972567560814)*x_ref[1]**o+(0.308266449093)*x_ref[1]+(0.71069864654)*x_ref[2]**o+(-0.683195491271)*x_ref[2]
        ref[3,0,0,1]=(-0.505155072199)*x_ref[0]**o+(0.339013729552)*x_ref[0]+(-0.81565931012)*x_ref[1]**o+(0.362458680494)*x_ref[1]+(0.960221320064)*x_ref[2]**o+(-0.17234580875)*x_ref[2]
        ref[3,0,1,0]=(0.67972657926)*x_ref[0]**o+(0.143085543693)*x_ref[0]+(-0.699516563264)*x_ref[1]**o+(-0.892226133751)*x_ref[1]+(-0.898541541049)*x_ref[2]**o+(-0.736249681714)*x_ref[2]
        ref[3,0,1,1]=(-0.144423319215)*x_ref[0]**o+(-0.509713525676)*x_ref[0]+(0.347584762861)*x_ref[1]**o+(0.666169972794)*x_ref[1]+(0.691734080163)*x_ref[2]**o+(0.9338062148)*x_ref[2]
        ref[3,0,2,0]=(-0.47323843589)*x_ref[0]**o+(0.0819655059363)*x_ref[0]+(0.0863030211133)*x_ref[1]**o+(0.127803089307)*x_ref[1]+(-0.155852489756)*x_ref[2]**o+(-0.00470398386987)*x_ref[2]
        ref[3,0,2,1]=(0.372452726232)*x_ref[0]**o+(0.133808609857)*x_ref[0]+(0.962945642209)*x_ref[1]**o+(0.0446609213929)*x_ref[1]+(-0.446207951293)*x_ref[2]**o+(0.0400733439868)*x_ref[2]
        ref[3,1,0,0]=(-0.668575734235)*x_ref[0]**o+(0.905913572459)*x_ref[0]+(-0.321327880802)*x_ref[1]**o+(0.0263209456188)*x_ref[1]+(-0.0355325780368)*x_ref[2]**o+(-0.624130395105)*x_ref[2]
        ref[3,1,0,1]=(0.859810828483)*x_ref[0]**o+(-0.235372125226)*x_ref[0]+(-0.16715244612)*x_ref[1]**o+(-0.793399872713)*x_ref[1]+(0.303973795556)*x_ref[2]**o+(0.176740655638)*x_ref[2]
        ref[3,1,1,0]=(0.736448874258)*x_ref[0]**o+(-0.448956585466)*x_ref[0]+(0.782095902992)*x_ref[1]**o+(0.796524196866)*x_ref[1]+(-0.732812289541)*x_ref[2]**o+(0.687089667291)*x_ref[2]
        ref[3,1,1,1]=(-0.797708068194)*x_ref[0]**o+(-0.691308116296)*x_ref[0]+(-0.644065410151)*x_ref[1]**o+(-0.929624672017)*x_ref[1]+(0.265040940279)*x_ref[2]**o+(0.424929324889)*x_ref[2]
        ref[3,1,2,0]=(0.555032482777)*x_ref[0]**o+(-0.564617314931)*x_ref[0]+(-0.612721853869)*x_ref[1]**o+(-0.048466699593)*x_ref[1]+(0.753265037897)*x_ref[2]**o+(0.913086239874)*x_ref[2]
        ref[3,1,2,1]=(0.580881807771)*x_ref[0]**o+(0.970973128604)*x_ref[0]+(-0.603610563753)*x_ref[1]**o+(0.979985814438)*x_ref[1]+(-0.639531987694)*x_ref[2]**o+(-0.773094138073)*x_ref[2]
        ref[3,2,0,0]=(0.574306840544)*x_ref[0]**o+(0.449094769988)*x_ref[0]+(-0.406303076614)*x_ref[1]**o+(-0.776648781906)*x_ref[1]+(-0.59829476647)*x_ref[2]**o+(0.257632446666)*x_ref[2]
        ref[3,2,0,1]=(0.689106127051)*x_ref[0]**o+(0.717184851708)*x_ref[0]+(-0.495581943615)*x_ref[1]**o+(-0.30329139788)*x_ref[1]+(0.901753128535)*x_ref[2]**o+(0.48496505727)*x_ref[2]
        ref[3,2,1,0]=(0.252050902256)*x_ref[0]**o+(-0.0414281706943)*x_ref[0]+(-0.284125627654)*x_ref[1]**o+(0.477450554269)*x_ref[1]+(0.849202242148)*x_ref[2]**o+(0.211826565889)*x_ref[2]
        ref[3,2,1,1]=(-0.319900673571)*x_ref[0]**o+(0.965993723739)*x_ref[0]+(0.671360092314)*x_ref[1]**o+(-0.652805407934)*x_ref[1]+(-0.0352316646853)*x_ref[2]**o+(-0.442493499306)*x_ref[2]
        ref[3,2,2,0]=(0.525577660439)*x_ref[0]**o+(0.660882378915)*x_ref[0]+(-0.419826738105)*x_ref[1]**o+(-0.798251889077)*x_ref[1]+(-0.430382393461)*x_ref[2]**o+(0.714357242496)*x_ref[2]
        ref[3,2,2,1]=(-0.200634944034)*x_ref[0]**o+(0.211086366034)*x_ref[0]+(0.116289218188)*x_ref[1]**o+(-0.785792860501)*x_ref[1]+(0.763320327438)*x_ref[2]**o+(-0.457355826175)*x_ref[2]
        ref[3,3,0,0]=(-0.890873783782)*x_ref[0]**o+(0.322610697657)*x_ref[0]+(-0.351437390332)*x_ref[1]**o+(-0.720698169708)*x_ref[1]+(-0.664444440194)*x_ref[2]**o+(0.274865394829)*x_ref[2]
        ref[3,3,0,1]=(-0.384885652342)*x_ref[0]**o+(-0.409799972422)*x_ref[0]+(0.679315041587)*x_ref[1]**o+(0.0859788488759)*x_ref[1]+(-0.136173937432)*x_ref[2]**o+(-0.136591429212)*x_ref[2]
        ref[3,3,1,0]=(0.147934741321)*x_ref[0]**o+(0.246941956666)*x_ref[0]+(0.696831123708)*x_ref[1]**o+(-0.905966322958)*x_ref[1]+(0.672705695255)*x_ref[2]**o+(-0.776017262395)*x_ref[2]
        ref[3,3,1,1]=(0.933982654679)*x_ref[0]**o+(-0.367246348547)*x_ref[0]+(-0.254076834401)*x_ref[1]**o+(0.148617981234)*x_ref[1]+(-0.486881004358)*x_ref[2]**o+(0.636371570082)*x_ref[2]
        ref[3,3,2,0]=(-0.536125392071)*x_ref[0]**o+(-0.663380779479)*x_ref[0]+(0.0655694689011)*x_ref[1]**o+(0.0390002326894)*x_ref[1]+(-0.731730129742)*x_ref[2]**o+(-0.578791485067)*x_ref[2]
        ref[3,3,2,1]=(0.705873902409)*x_ref[0]**o+(-0.0444303506955)*x_ref[0]+(-0.978519191225)*x_ref[1]**o+(-0.0280878186958)*x_ref[1]+(0.182378726121)*x_ref[2]**o+(-0.555345919696)*x_ref[2]
        ref[3,4,0,0]=(0.945945862395)*x_ref[0]**o+(-0.448487950632)*x_ref[0]+(-0.477570573229)*x_ref[1]**o+(0.313871838624)*x_ref[1]+(0.237944220287)*x_ref[2]**o+(-0.218095130482)*x_ref[2]
        ref[3,4,0,1]=(-0.758647277269)*x_ref[0]**o+(-0.550604667208)*x_ref[0]+(-0.262944399053)*x_ref[1]**o+(0.994302652859)*x_ref[1]+(0.363440310008)*x_ref[2]**o+(-0.127863970613)*x_ref[2]
        ref[3,4,1,0]=(0.667202433965)*x_ref[0]**o+(0.954901068821)*x_ref[0]+(0.719745510318)*x_ref[1]**o+(-0.763882742007)*x_ref[1]+(0.880927476123)*x_ref[2]**o+(0.102034630893)*x_ref[2]
        ref[3,4,1,1]=(-0.684644024456)*x_ref[0]**o+(0.88037275093)*x_ref[0]+(-0.772213766289)*x_ref[1]**o+(-0.702615214094)*x_ref[1]+(-0.25661708053)*x_ref[2]**o+(0.126779040694)*x_ref[2]
        ref[3,4,2,0]=(-0.589611385303)*x_ref[0]**o+(0.638712635252)*x_ref[0]+(0.745450741856)*x_ref[1]**o+(0.399494092815)*x_ref[1]+(-0.717293809038)*x_ref[2]**o+(0.305662737751)*x_ref[2]
        ref[3,4,2,1]=(0.0572426997796)*x_ref[0]**o+(-0.764600144578)*x_ref[0]+(0.482545219565)*x_ref[1]**o+(0.525883069411)*x_ref[1]+(-0.612224327195)*x_ref[2]**o+(0.354344820256)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.100067148208)*x[0]+(-0.8375743725)*x[1]
        ref=(-0.100067148208)*x_ref[0]+(-0.8375743725)*x_ref[1]
      else:
        arg=(-0.247343416726)*x[0]+(0.604193484164)*x[1]+(-0.219899438287)*x[2]
        ref=(-0.247343416726)*x_ref[0]+(0.604193484164)*x_ref[1]+(-0.219899438287)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        arg[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
        ref[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        ref[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
      else:
        arg[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        arg[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
        ref[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        ref[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.00153698933322)*x[0]+(0.392281645764)*x[1]
        arg[0,1]=(-1.2794655544)*x[0]+(-0.373519639818)*x[1]
        arg[0,2]=(0.30127033321)*x[0]+(-0.0996865853738)*x[1]
        arg[0,3]=(1.23405799349)*x[0]+(0.498558200339)*x[1]
        arg[0,4]=(0.85519350742)*x[0]+(-0.19188046135)*x[1]
        arg[1,0]=(0.417367834962)*x[0]+(0.992318201639)*x[1]
        arg[1,1]=(-0.646432128833)*x[0]+(-0.872750321582)*x[1]
        arg[1,2]=(-0.806117185061)*x[0]+(-0.89396408325)*x[1]
        arg[1,3]=(0.768017237627)*x[0]+(-1.2904111912)*x[1]
        arg[1,4]=(-0.0159685523188)*x[0]+(0.401410193169)*x[1]
        arg[2,0]=(-0.844501716181)*x[0]+(-0.123115343819)*x[1]
        arg[2,1]=(1.08441909409)*x[0]+(0.152432282539)*x[1]
        arg[2,2]=(-0.60551037413)*x[0]+(1.15692828826)*x[1]
        arg[2,3]=(-1.21509168541)*x[0]+(1.03683236377)*x[1]
        arg[2,4]=(1.16850730134)*x[0]+(0.162202615909)*x[1]
        arg[3,0]=(0.727772166897)*x[0]+(-0.296815840935)*x[1]
        arg[3,1]=(-0.0880577151489)*x[0]+(1.66427691426)*x[1]
        arg[3,2]=(0.415144534458)*x[0]+(-0.555833723295)*x[1]
        arg[3,3]=(1.09444635829)*x[0]+(-0.335209417536)*x[1]
        arg[3,4]=(-1.05047769585)*x[0]+(0.684427039986)*x[1]
        ref[0,0]=(0.00153698933322)*x_ref[0]+(0.392281645764)*x_ref[1]
        ref[0,1]=(-1.2794655544)*x_ref[0]+(-0.373519639818)*x_ref[1]
        ref[0,2]=(0.30127033321)*x_ref[0]+(-0.0996865853738)*x_ref[1]
        ref[0,3]=(1.23405799349)*x_ref[0]+(0.498558200339)*x_ref[1]
        ref[0,4]=(0.85519350742)*x_ref[0]+(-0.19188046135)*x_ref[1]
        ref[1,0]=(0.417367834962)*x_ref[0]+(0.992318201639)*x_ref[1]
        ref[1,1]=(-0.646432128833)*x_ref[0]+(-0.872750321582)*x_ref[1]
        ref[1,2]=(-0.806117185061)*x_ref[0]+(-0.89396408325)*x_ref[1]
        ref[1,3]=(0.768017237627)*x_ref[0]+(-1.2904111912)*x_ref[1]
        ref[1,4]=(-0.0159685523188)*x_ref[0]+(0.401410193169)*x_ref[1]
        ref[2,0]=(-0.844501716181)*x_ref[0]+(-0.123115343819)*x_ref[1]
        ref[2,1]=(1.08441909409)*x_ref[0]+(0.152432282539)*x_ref[1]
        ref[2,2]=(-0.60551037413)*x_ref[0]+(1.15692828826)*x_ref[1]
        ref[2,3]=(-1.21509168541)*x_ref[0]+(1.03683236377)*x_ref[1]
        ref[2,4]=(1.16850730134)*x_ref[0]+(0.162202615909)*x_ref[1]
        ref[3,0]=(0.727772166897)*x_ref[0]+(-0.296815840935)*x_ref[1]
        ref[3,1]=(-0.0880577151489)*x_ref[0]+(1.66427691426)*x_ref[1]
        ref[3,2]=(0.415144534458)*x_ref[0]+(-0.555833723295)*x_ref[1]
        ref[3,3]=(1.09444635829)*x_ref[0]+(-0.335209417536)*x_ref[1]
        ref[3,4]=(-1.05047769585)*x_ref[0]+(0.684427039986)*x_ref[1]
      else:
        arg[0,0]=(-0.611607640773)*x[0]+(0.584485971628)*x[1]+(0.0090316337643)*x[2]
        arg[0,1]=(0.905278285857)*x[0]+(0.332523151242)*x[1]+(0.9909841807)*x[2]
        arg[0,2]=(-1.06629333132)*x[0]+(-0.38478180064)*x[1]+(-0.63309211256)*x[2]
        arg[0,3]=(0.122436386344)*x[0]+(-0.218752671204)*x[1]+(-0.746695172216)*x[2]
        arg[0,4]=(1.03152124874)*x[0]+(-1.11327271563)*x[1]+(-0.257591112135)*x[2]
        arg[1,0]=(0.969019600299)*x[0]+(-0.320279309689)*x[1]+(0.327767746545)*x[2]
        arg[1,1]=(0.291023069052)*x[0]+(0.78112901064)*x[1]+(0.879257239042)*x[2]
        arg[1,2]=(-0.131262891536)*x[0]+(-0.0109690229153)*x[1]+(-0.171900332406)*x[2]
        arg[1,3]=(-1.32272684486)*x[0]+(-0.256869536085)*x[1]+(0.0308524567462)*x[2]
        arg[1,4]=(1.30707061279)*x[0]+(0.972866282634)*x[1]+(-0.474948537882)*x[2]
        arg[2,0]=(0.169034210421)*x[0]+(-1.08714896226)*x[1]+(-0.636838450708)*x[2]
        arg[2,1]=(-0.470752780773)*x[0]+(-0.695286088059)*x[1]+(-1.19197058712)*x[2]
        arg[2,2]=(0.701403721228)*x[0]+(0.551002482436)*x[1]+(-0.295853851234)*x[2]
        arg[2,3]=(-0.796431361745)*x[0]+(0.0581884766856)*x[1]+(-1.13668563089)*x[2]
        arg[2,4]=(-0.733185798017)*x[0]+(0.178900103735)*x[1]+(-0.845671526781)*x[2]
        arg[3,0]=(-1.47064927107)*x[0]+(1.33393324965)*x[1]+(-0.767393460937)*x[2]
        arg[3,1]=(1.36784513226)*x[0]+(0.100008910431)*x[1]+(-1.09210256713)*x[2]
        arg[3,2]=(-0.873076669977)*x[0]+(0.367213631778)*x[1]+(0.0698296920039)*x[2]
        arg[3,3]=(-1.76609214158)*x[0]+(-1.26186286566)*x[1]+(0.377817035113)*x[2]
        arg[3,4]=(-1.28261709396)*x[0]+(0.240480767533)*x[1]+(0.307813279357)*x[2]
        ref[0,0]=(-0.611607640773)*x_ref[0]+(0.584485971628)*x_ref[1]+(0.0090316337643)*x_ref[2]
        ref[0,1]=(0.905278285857)*x_ref[0]+(0.332523151242)*x_ref[1]+(0.9909841807)*x_ref[2]
        ref[0,2]=(-1.06629333132)*x_ref[0]+(-0.38478180064)*x_ref[1]+(-0.63309211256)*x_ref[2]
        ref[0,3]=(0.122436386344)*x_ref[0]+(-0.218752671204)*x_ref[1]+(-0.746695172216)*x_ref[2]
        ref[0,4]=(1.03152124874)*x_ref[0]+(-1.11327271563)*x_ref[1]+(-0.257591112135)*x_ref[2]
        ref[1,0]=(0.969019600299)*x_ref[0]+(-0.320279309689)*x_ref[1]+(0.327767746545)*x_ref[2]
        ref[1,1]=(0.291023069052)*x_ref[0]+(0.78112901064)*x_ref[1]+(0.879257239042)*x_ref[2]
        ref[1,2]=(-0.131262891536)*x_ref[0]+(-0.0109690229153)*x_ref[1]+(-0.171900332406)*x_ref[2]
        ref[1,3]=(-1.32272684486)*x_ref[0]+(-0.256869536085)*x_ref[1]+(0.0308524567462)*x_ref[2]
        ref[1,4]=(1.30707061279)*x_ref[0]+(0.972866282634)*x_ref[1]+(-0.474948537882)*x_ref[2]
        ref[2,0]=(0.169034210421)*x_ref[0]+(-1.08714896226)*x_ref[1]+(-0.636838450708)*x_ref[2]
        ref[2,1]=(-0.470752780773)*x_ref[0]+(-0.695286088059)*x_ref[1]+(-1.19197058712)*x_ref[2]
        ref[2,2]=(0.701403721228)*x_ref[0]+(0.551002482436)*x_ref[1]+(-0.295853851234)*x_ref[2]
        ref[2,3]=(-0.796431361745)*x_ref[0]+(0.0581884766856)*x_ref[1]+(-1.13668563089)*x_ref[2]
        ref[2,4]=(-0.733185798017)*x_ref[0]+(0.178900103735)*x_ref[1]+(-0.845671526781)*x_ref[2]
        ref[3,0]=(-1.47064927107)*x_ref[0]+(1.33393324965)*x_ref[1]+(-0.767393460937)*x_ref[2]
        ref[3,1]=(1.36784513226)*x_ref[0]+(0.100008910431)*x_ref[1]+(-1.09210256713)*x_ref[2]
        ref[3,2]=(-0.873076669977)*x_ref[0]+(0.367213631778)*x_ref[1]+(0.0698296920039)*x_ref[2]
        ref[3,3]=(-1.76609214158)*x_ref[0]+(-1.26186286566)*x_ref[1]+(0.377817035113)*x_ref[2]
        ref[3,4]=(-1.28261709396)*x_ref[0]+(0.240480767533)*x_ref[1]+(0.307813279357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.423505169404)*x[0]+(-1.10390334264)*x[1]
        arg[0,0,1]=(-0.181820089254)*x[0]+(-0.374602268791)*x[1]
        arg[0,1,0]=(-1.36906917075)*x[0]+(0.26554385908)*x[1]
        arg[0,1,1]=(-0.280661572374)*x[0]+(-0.239232524287)*x[1]
        arg[1,0,0]=(-1.25670251934)*x[0]+(-0.429353977735)*x[1]
        arg[1,0,1]=(-0.747928970735)*x[0]+(1.17874317611)*x[1]
        arg[1,1,0]=(-0.548814611802)*x[0]+(-0.0249315764285)*x[1]
        arg[1,1,1]=(-0.481123117522)*x[0]+(1.23402203879)*x[1]
        arg[2,0,0]=(-0.0639512689884)*x[0]+(-0.0278213371376)*x[1]
        arg[2,0,1]=(-1.17292797621)*x[0]+(0.42140600304)*x[1]
        arg[2,1,0]=(-0.358052601083)*x[0]+(-1.66807880791)*x[1]
        arg[2,1,1]=(-0.291433251349)*x[0]+(0.111473162108)*x[1]
        arg[3,0,0]=(0.742643625989)*x[0]+(-1.29214106228)*x[1]
        arg[3,0,1]=(0.319887321826)*x[0]+(-0.40213369238)*x[1]
        arg[3,1,0]=(-0.449412064097)*x[0]+(-1.10654262009)*x[1]
        arg[3,1,1]=(1.19722227632)*x[0]+(1.3455799026)*x[1]
        arg[4,0,0]=(-1.77453188487)*x[0]+(0.634838793872)*x[1]
        arg[4,0,1]=(-0.215461066907)*x[0]+(1.11635233844)*x[1]
        arg[4,1,0]=(0.835816652592)*x[0]+(0.668378554214)*x[1]
        arg[4,1,1]=(-1.22133225406)*x[0]+(-0.148780252398)*x[1]
        arg[5,0,0]=(0.320245409639)*x[0]+(-1.08828513776)*x[1]
        arg[5,0,1]=(-0.1865835417)*x[0]+(-0.353282061874)*x[1]
        arg[5,1,0]=(1.38858040685)*x[0]+(0.166398261098)*x[1]
        arg[5,1,1]=(0.947718875755)*x[0]+(-1.06171934302)*x[1]
        ref[0,0,0]=(-0.423505169404)*x_ref[0]+(-1.10390334264)*x_ref[1]
        ref[0,0,1]=(-0.181820089254)*x_ref[0]+(-0.374602268791)*x_ref[1]
        ref[0,1,0]=(-1.36906917075)*x_ref[0]+(0.26554385908)*x_ref[1]
        ref[0,1,1]=(-0.280661572374)*x_ref[0]+(-0.239232524287)*x_ref[1]
        ref[1,0,0]=(-1.25670251934)*x_ref[0]+(-0.429353977735)*x_ref[1]
        ref[1,0,1]=(-0.747928970735)*x_ref[0]+(1.17874317611)*x_ref[1]
        ref[1,1,0]=(-0.548814611802)*x_ref[0]+(-0.0249315764285)*x_ref[1]
        ref[1,1,1]=(-0.481123117522)*x_ref[0]+(1.23402203879)*x_ref[1]
        ref[2,0,0]=(-0.0639512689884)*x_ref[0]+(-0.0278213371376)*x_ref[1]
        ref[2,0,1]=(-1.17292797621)*x_ref[0]+(0.42140600304)*x_ref[1]
        ref[2,1,0]=(-0.358052601083)*x_ref[0]+(-1.66807880791)*x_ref[1]
        ref[2,1,1]=(-0.291433251349)*x_ref[0]+(0.111473162108)*x_ref[1]
        ref[3,0,0]=(0.742643625989)*x_ref[0]+(-1.29214106228)*x_ref[1]
        ref[3,0,1]=(0.319887321826)*x_ref[0]+(-0.40213369238)*x_ref[1]
        ref[3,1,0]=(-0.449412064097)*x_ref[0]+(-1.10654262009)*x_ref[1]
        ref[3,1,1]=(1.19722227632)*x_ref[0]+(1.3455799026)*x_ref[1]
        ref[4,0,0]=(-1.77453188487)*x_ref[0]+(0.634838793872)*x_ref[1]
        ref[4,0,1]=(-0.215461066907)*x_ref[0]+(1.11635233844)*x_ref[1]
        ref[4,1,0]=(0.835816652592)*x_ref[0]+(0.668378554214)*x_ref[1]
        ref[4,1,1]=(-1.22133225406)*x_ref[0]+(-0.148780252398)*x_ref[1]
        ref[5,0,0]=(0.320245409639)*x_ref[0]+(-1.08828513776)*x_ref[1]
        ref[5,0,1]=(-0.1865835417)*x_ref[0]+(-0.353282061874)*x_ref[1]
        ref[5,1,0]=(1.38858040685)*x_ref[0]+(0.166398261098)*x_ref[1]
        ref[5,1,1]=(0.947718875755)*x_ref[0]+(-1.06171934302)*x_ref[1]
      else:
        arg[0,0,0]=(-0.892096322478)*x[0]+(0.720363183451)*x[1]+(0.600609272191)*x[2]
        arg[0,0,1]=(-0.418024461935)*x[0]+(-0.819275292035)*x[1]+(0.134962557692)*x[2]
        arg[0,1,0]=(-0.449783838361)*x[0]+(0.943677173533)*x[1]+(0.209424190991)*x[2]
        arg[0,1,1]=(-0.424386775004)*x[0]+(-0.213024388094)*x[1]+(-1.05927943384)*x[2]
        arg[1,0,0]=(-0.699504096539)*x[0]+(-0.153570308377)*x[1]+(0.576902392658)*x[2]
        arg[1,0,1]=(-0.426914055661)*x[0]+(-0.330405267059)*x[1]+(1.07656632155)*x[2]
        arg[1,1,0]=(0.775238005352)*x[0]+(1.11012072575)*x[1]+(0.764627098603)*x[2]
        arg[1,1,1]=(0.0130424708408)*x[0]+(-0.671778102107)*x[1]+(0.780684368983)*x[2]
        arg[2,0,0]=(-0.0512224258498)*x[0]+(-1.50515803984)*x[1]+(0.342824143946)*x[2]
        arg[2,0,1]=(0.117541624176)*x[0]+(0.372415524552)*x[1]+(0.255218095016)*x[2]
        arg[2,1,0]=(-0.574508895703)*x[0]+(0.345217292359)*x[1]+(0.209211956821)*x[2]
        arg[2,1,1]=(0.93643587147)*x[0]+(-0.096209009902)*x[1]+(0.205150859527)*x[2]
        arg[3,0,0]=(1.7541929884)*x[0]+(0.169061478447)*x[1]+(0.534746695515)*x[2]
        arg[3,0,1]=(0.442329266782)*x[0]+(0.334971710414)*x[1]+(1.66892557698)*x[2]
        arg[3,1,0]=(0.316436716758)*x[0]+(-1.25442030987)*x[1]+(0.372183105374)*x[2]
        arg[3,1,1]=(-0.254157820265)*x[0]+(0.287381300023)*x[1]+(0.0574117423352)*x[2]
        arg[4,0,0]=(-1.08380257521)*x[0]+(-1.31461890032)*x[1]+(-0.503802775099)*x[2]
        arg[4,0,1]=(-0.558293486798)*x[0]+(0.702317635925)*x[1]+(0.463814829137)*x[2]
        arg[4,1,0]=(-0.677160971974)*x[0]+(-0.14921275312)*x[1]+(-1.6501459435)*x[2]
        arg[4,1,1]=(-0.176672532474)*x[0]+(0.945017605811)*x[1]+(-0.790342939788)*x[2]
        arg[5,0,0]=(-1.45668070675)*x[0]+(-0.404957129204)*x[1]+(1.37383065954)*x[2]
        arg[5,0,1]=(0.53829326756)*x[0]+(0.543333031479)*x[1]+(-0.854057296362)*x[2]
        arg[5,1,0]=(0.805864283921)*x[0]+(-0.00186634953513)*x[1]+(-1.97111362539)*x[2]
        arg[5,1,1]=(-0.387746238967)*x[0]+(-0.739158353882)*x[1]+(-0.134082808118)*x[2]
        ref[0,0,0]=(-0.892096322478)*x_ref[0]+(0.720363183451)*x_ref[1]+(0.600609272191)*x_ref[2]
        ref[0,0,1]=(-0.418024461935)*x_ref[0]+(-0.819275292035)*x_ref[1]+(0.134962557692)*x_ref[2]
        ref[0,1,0]=(-0.449783838361)*x_ref[0]+(0.943677173533)*x_ref[1]+(0.209424190991)*x_ref[2]
        ref[0,1,1]=(-0.424386775004)*x_ref[0]+(-0.213024388094)*x_ref[1]+(-1.05927943384)*x_ref[2]
        ref[1,0,0]=(-0.699504096539)*x_ref[0]+(-0.153570308377)*x_ref[1]+(0.576902392658)*x_ref[2]
        ref[1,0,1]=(-0.426914055661)*x_ref[0]+(-0.330405267059)*x_ref[1]+(1.07656632155)*x_ref[2]
        ref[1,1,0]=(0.775238005352)*x_ref[0]+(1.11012072575)*x_ref[1]+(0.764627098603)*x_ref[2]
        ref[1,1,1]=(0.0130424708408)*x_ref[0]+(-0.671778102107)*x_ref[1]+(0.780684368983)*x_ref[2]
        ref[2,0,0]=(-0.0512224258498)*x_ref[0]+(-1.50515803984)*x_ref[1]+(0.342824143946)*x_ref[2]
        ref[2,0,1]=(0.117541624176)*x_ref[0]+(0.372415524552)*x_ref[1]+(0.255218095016)*x_ref[2]
        ref[2,1,0]=(-0.574508895703)*x_ref[0]+(0.345217292359)*x_ref[1]+(0.209211956821)*x_ref[2]
        ref[2,1,1]=(0.93643587147)*x_ref[0]+(-0.096209009902)*x_ref[1]+(0.205150859527)*x_ref[2]
        ref[3,0,0]=(1.7541929884)*x_ref[0]+(0.169061478447)*x_ref[1]+(0.534746695515)*x_ref[2]
        ref[3,0,1]=(0.442329266782)*x_ref[0]+(0.334971710414)*x_ref[1]+(1.66892557698)*x_ref[2]
        ref[3,1,0]=(0.316436716758)*x_ref[0]+(-1.25442030987)*x_ref[1]+(0.372183105374)*x_ref[2]
        ref[3,1,1]=(-0.254157820265)*x_ref[0]+(0.287381300023)*x_ref[1]+(0.0574117423352)*x_ref[2]
        ref[4,0,0]=(-1.08380257521)*x_ref[0]+(-1.31461890032)*x_ref[1]+(-0.503802775099)*x_ref[2]
        ref[4,0,1]=(-0.558293486798)*x_ref[0]+(0.702317635925)*x_ref[1]+(0.463814829137)*x_ref[2]
        ref[4,1,0]=(-0.677160971974)*x_ref[0]+(-0.14921275312)*x_ref[1]+(-1.6501459435)*x_ref[2]
        ref[4,1,1]=(-0.176672532474)*x_ref[0]+(0.945017605811)*x_ref[1]+(-0.790342939788)*x_ref[2]
        ref[5,0,0]=(-1.45668070675)*x_ref[0]+(-0.404957129204)*x_ref[1]+(1.37383065954)*x_ref[2]
        ref[5,0,1]=(0.53829326756)*x_ref[0]+(0.543333031479)*x_ref[1]+(-0.854057296362)*x_ref[2]
        ref[5,1,0]=(0.805864283921)*x_ref[0]+(-0.00186634953513)*x_ref[1]+(-1.97111362539)*x_ref[2]
        ref[5,1,1]=(-0.387746238967)*x_ref[0]+(-0.739158353882)*x_ref[1]+(-0.134082808118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.754311658086)*x[0]+(0.235157034342)*x[1]
        arg[0,0,0,1]=(-1.33786315024)*x[0]+(0.243349279742)*x[1]
        arg[0,0,1,0]=(1.48657750088)*x[0]+(1.09853108305)*x[1]
        arg[0,0,1,1]=(1.4501362266)*x[0]+(0.824811567155)*x[1]
        arg[0,0,2,0]=(1.12567566082)*x[0]+(0.663525681174)*x[1]
        arg[0,0,2,1]=(1.48991565019)*x[0]+(-0.680795365011)*x[1]
        arg[0,1,0,0]=(0.268571449528)*x[0]+(0.818595834672)*x[1]
        arg[0,1,0,1]=(-1.55490945555)*x[0]+(0.783595197888)*x[1]
        arg[0,1,1,0]=(0.913201382585)*x[0]+(0.855491379376)*x[1]
        arg[0,1,1,1]=(-0.179192075116)*x[0]+(-0.690415079523)*x[1]
        arg[0,1,2,0]=(-0.0393574502912)*x[0]+(-0.690157650634)*x[1]
        arg[0,1,2,1]=(0.784216863217)*x[0]+(0.498735486159)*x[1]
        arg[0,2,0,0]=(0.137939554165)*x[0]+(-1.33961519113)*x[1]
        arg[0,2,0,1]=(-1.03897311736)*x[0]+(-1.37797405312)*x[1]
        arg[0,2,1,0]=(0.199812003756)*x[0]+(-0.0388499674037)*x[1]
        arg[0,2,1,1]=(1.15530534202)*x[0]+(-1.28280869078)*x[1]
        arg[0,2,2,0]=(0.836447791155)*x[0]+(1.2627066026)*x[1]
        arg[0,2,2,1]=(-0.584608492688)*x[0]+(-1.55305974305)*x[1]
        arg[0,3,0,0]=(-0.661952381799)*x[0]+(-1.27062403778)*x[1]
        arg[0,3,0,1]=(-1.47789532259)*x[0]+(-0.279583524936)*x[1]
        arg[0,3,1,0]=(1.40729771357)*x[0]+(0.0971284144528)*x[1]
        arg[0,3,1,1]=(-0.459167955911)*x[0]+(-1.29033281524)*x[1]
        arg[0,3,2,0]=(-1.63189436824)*x[0]+(0.75729808422)*x[1]
        arg[0,3,2,1]=(0.522479828718)*x[0]+(-0.0231299515383)*x[1]
        arg[0,4,0,0]=(0.241891446615)*x[0]+(0.803327107651)*x[1]
        arg[0,4,0,1]=(0.273613502836)*x[0]+(-0.0115180464171)*x[1]
        arg[0,4,1,0]=(1.96219818696)*x[0]+(-0.349137094743)*x[1]
        arg[0,4,1,1]=(0.365690821719)*x[0]+(-1.6943347181)*x[1]
        arg[0,4,2,0]=(-0.283970779872)*x[0]+(-0.258537179804)*x[1]
        arg[0,4,2,1]=(-0.973333430598)*x[0]+(1.48706276212)*x[1]
        arg[1,0,0,0]=(1.18757466971)*x[0]+(0.838804656357)*x[1]
        arg[1,0,0,1]=(0.840859904702)*x[0]+(1.09955195682)*x[1]
        arg[1,0,1,0]=(1.46190513227)*x[0]+(-0.596959830422)*x[1]
        arg[1,0,1,1]=(0.718689296745)*x[0]+(-0.628074620222)*x[1]
        arg[1,0,2,0]=(0.440761293041)*x[0]+(0.115800854194)*x[1]
        arg[1,0,2,1]=(1.34891146904)*x[0]+(0.813690260652)*x[1]
        arg[1,1,0,0]=(0.183071638098)*x[0]+(-1.37996210755)*x[1]
        arg[1,1,0,1]=(0.258142115988)*x[0]+(-0.757709381529)*x[1]
        arg[1,1,1,0]=(1.72392829309)*x[0]+(0.444995102781)*x[1]
        arg[1,1,1,1]=(-0.380898209988)*x[0]+(-0.0950246860347)*x[1]
        arg[1,1,2,0]=(0.203007226016)*x[0]+(1.09786001832)*x[1]
        arg[1,1,2,1]=(-0.502500865341)*x[0]+(-0.629980527995)*x[1]
        arg[1,2,0,0]=(0.482549518436)*x[0]+(0.565840824146)*x[1]
        arg[1,2,0,1]=(0.355818565068)*x[0]+(-0.78273295574)*x[1]
        arg[1,2,1,0]=(1.79204033071)*x[0]+(-0.698040851585)*x[1]
        arg[1,2,1,1]=(-0.534827741044)*x[0]+(0.00943131627498)*x[1]
        arg[1,2,2,0]=(-1.00663524444)*x[0]+(0.0207854553456)*x[1]
        arg[1,2,2,1]=(-0.338048040159)*x[0]+(-0.380607406217)*x[1]
        arg[1,3,0,0]=(0.586310549044)*x[0]+(0.276326798805)*x[1]
        arg[1,3,0,1]=(-0.260482580702)*x[0]+(-0.349279395321)*x[1]
        arg[1,3,1,0]=(-0.196533829895)*x[0]+(0.614805272738)*x[1]
        arg[1,3,1,1]=(-0.363176483895)*x[0]+(-0.925817017731)*x[1]
        arg[1,3,2,0]=(-1.24182075421)*x[0]+(0.215966446013)*x[1]
        arg[1,3,2,1]=(1.07546652642)*x[0]+(0.383803991974)*x[1]
        arg[1,4,0,0]=(0.199312303525)*x[0]+(-0.314520883859)*x[1]
        arg[1,4,0,1]=(-1.56890826152)*x[0]+(-0.133086662241)*x[1]
        arg[1,4,1,0]=(1.1307241936)*x[0]+(-1.32905744984)*x[1]
        arg[1,4,1,1]=(-0.385991455298)*x[0]+(0.292913024507)*x[1]
        arg[1,4,2,0]=(1.14561678056)*x[0]+(-0.662696162276)*x[1]
        arg[1,4,2,1]=(-1.49533948954)*x[0]+(1.43396138744)*x[1]
        arg[2,0,0,0]=(-0.63160396493)*x[0]+(0.813120998776)*x[1]
        arg[2,0,0,1]=(0.65598998434)*x[0]+(-1.02445537553)*x[1]
        arg[2,0,1,0]=(-1.13473104075)*x[0]+(0.057331183449)*x[1]
        arg[2,0,1,1]=(-0.154892482107)*x[0]+(0.481969957579)*x[1]
        arg[2,0,2,0]=(0.521916671496)*x[0]+(0.546676943059)*x[1]
        arg[2,0,2,1]=(-1.68919379028)*x[0]+(1.08292038981)*x[1]
        arg[2,1,0,0]=(-0.643846595888)*x[0]+(0.528306097755)*x[1]
        arg[2,1,0,1]=(1.3745090497)*x[0]+(-0.0801648509184)*x[1]
        arg[2,1,1,0]=(0.129096817054)*x[0]+(0.90750370855)*x[1]
        arg[2,1,1,1]=(0.539309513305)*x[0]+(1.34949470961)*x[1]
        arg[2,1,2,0]=(1.4189703896)*x[0]+(-1.75011201893)*x[1]
        arg[2,1,2,1]=(-1.09325548386)*x[0]+(0.148279301716)*x[1]
        arg[2,2,0,0]=(-1.25785133743)*x[0]+(0.275071306594)*x[1]
        arg[2,2,0,1]=(-0.550152068761)*x[0]+(-1.20422484275)*x[1]
        arg[2,2,1,0]=(-0.102062863668)*x[0]+(-1.34431887541)*x[1]
        arg[2,2,1,1]=(0.103354724169)*x[0]+(-0.941952180446)*x[1]
        arg[2,2,2,0]=(1.2090339604)*x[0]+(-0.330357976796)*x[1]
        arg[2,2,2,1]=(0.203162449188)*x[0]+(-1.03640702014)*x[1]
        arg[2,3,0,0]=(-0.324576353514)*x[0]+(0.582614539267)*x[1]
        arg[2,3,0,1]=(-0.213823136215)*x[0]+(-0.473199105214)*x[1]
        arg[2,3,1,0]=(0.395144702599)*x[0]+(0.0358194467152)*x[1]
        arg[2,3,1,1]=(1.32860563396)*x[0]+(-0.577308310732)*x[1]
        arg[2,3,2,0]=(0.994205029432)*x[0]+(1.13433517636)*x[1]
        arg[2,3,2,1]=(-0.488499286742)*x[0]+(0.224253820738)*x[1]
        arg[2,4,0,0]=(1.48035663008)*x[0]+(-1.09697986914)*x[1]
        arg[2,4,0,1]=(-1.09991915151)*x[0]+(-0.557951199985)*x[1]
        arg[2,4,1,0]=(-1.23862699804)*x[0]+(-0.230602187713)*x[1]
        arg[2,4,1,1]=(-0.922609370448)*x[0]+(1.76673953413)*x[1]
        arg[2,4,2,0]=(0.731769419709)*x[0]+(-1.1094516608)*x[1]
        arg[2,4,2,1]=(-0.145672504805)*x[0]+(0.168194421355)*x[1]
        arg[3,0,0,0]=(1.02559707602)*x[0]+(0.260431368523)*x[1]
        arg[3,0,0,1]=(0.111899568824)*x[0]+(-0.90725475423)*x[1]
        arg[3,0,1,0]=(-0.226547745683)*x[0]+(-0.232830393372)*x[1]
        arg[3,0,1,1]=(0.347004318471)*x[0]+(-0.196130970001)*x[1]
        arg[3,0,2,0]=(-1.72155871442)*x[0]+(-1.28372407452)*x[1]
        arg[3,0,2,1]=(1.43817914693)*x[0]+(-0.354093253684)*x[1]
        arg[3,1,0,0]=(-0.293939245102)*x[0]+(0.0255562139103)*x[1]
        arg[3,1,0,1]=(0.139306213018)*x[0]+(0.595062226642)*x[1]
        arg[3,1,1,0]=(-0.670339920672)*x[0]+(0.387520583192)*x[1]
        arg[3,1,1,1]=(-0.0323325470954)*x[0]+(-0.681162999279)*x[1]
        arg[3,1,2,0]=(0.235275048739)*x[0]+(0.144725473773)*x[1]
        arg[3,1,2,1]=(0.132211795214)*x[0]+(0.262780216845)*x[1]
        arg[3,2,0,0]=(0.590514326163)*x[0]+(-0.84273677918)*x[1]
        arg[3,2,0,1]=(-0.925181556185)*x[0]+(-0.300244447001)*x[1]
        arg[3,2,1,0]=(-0.0509700795112)*x[0]+(-1.31919389616)*x[1]
        arg[3,2,1,1]=(1.50577291846)*x[0]+(-0.219284607769)*x[1]
        arg[3,2,2,0]=(-0.0460616673346)*x[0]+(-0.690312685466)*x[1]
        arg[3,2,2,1]=(-0.673765477936)*x[0]+(-0.0965860475641)*x[1]
        arg[3,3,0,0]=(-0.6577594174)*x[0]+(0.171886277016)*x[1]
        arg[3,3,0,1]=(-1.08810434718)*x[0]+(0.333153250183)*x[1]
        arg[3,3,1,0]=(0.150968371983)*x[0]+(0.290859317391)*x[1]
        arg[3,3,1,1]=(1.62096265131)*x[0]+(-1.10309602741)*x[1]
        arg[3,3,2,0]=(-0.932063639972)*x[0]+(-0.897131561126)*x[1]
        arg[3,3,2,1]=(-0.819110834345)*x[0]+(1.59159109562)*x[1]
        arg[3,4,0,0]=(0.272662931155)*x[0]+(0.69026285756)*x[1]
        arg[3,4,0,1]=(-0.600556370757)*x[0]+(-0.0853533546302)*x[1]
        arg[3,4,1,0]=(1.18544735281)*x[0]+(0.217751189034)*x[1]
        arg[3,4,1,1]=(-0.594609531358)*x[0]+(0.882158349788)*x[1]
        arg[3,4,2,0]=(1.18876765552)*x[0]+(-0.411481586979)*x[1]
        arg[3,4,2,1]=(-1.093067792)*x[0]+(-0.266744260768)*x[1]
        ref[0,0,0,0]=(-0.754311658086)*x_ref[0]+(0.235157034342)*x_ref[1]
        ref[0,0,0,1]=(-1.33786315024)*x_ref[0]+(0.243349279742)*x_ref[1]
        ref[0,0,1,0]=(1.48657750088)*x_ref[0]+(1.09853108305)*x_ref[1]
        ref[0,0,1,1]=(1.4501362266)*x_ref[0]+(0.824811567155)*x_ref[1]
        ref[0,0,2,0]=(1.12567566082)*x_ref[0]+(0.663525681174)*x_ref[1]
        ref[0,0,2,1]=(1.48991565019)*x_ref[0]+(-0.680795365011)*x_ref[1]
        ref[0,1,0,0]=(0.268571449528)*x_ref[0]+(0.818595834672)*x_ref[1]
        ref[0,1,0,1]=(-1.55490945555)*x_ref[0]+(0.783595197888)*x_ref[1]
        ref[0,1,1,0]=(0.913201382585)*x_ref[0]+(0.855491379376)*x_ref[1]
        ref[0,1,1,1]=(-0.179192075116)*x_ref[0]+(-0.690415079523)*x_ref[1]
        ref[0,1,2,0]=(-0.0393574502912)*x_ref[0]+(-0.690157650634)*x_ref[1]
        ref[0,1,2,1]=(0.784216863217)*x_ref[0]+(0.498735486159)*x_ref[1]
        ref[0,2,0,0]=(0.137939554165)*x_ref[0]+(-1.33961519113)*x_ref[1]
        ref[0,2,0,1]=(-1.03897311736)*x_ref[0]+(-1.37797405312)*x_ref[1]
        ref[0,2,1,0]=(0.199812003756)*x_ref[0]+(-0.0388499674037)*x_ref[1]
        ref[0,2,1,1]=(1.15530534202)*x_ref[0]+(-1.28280869078)*x_ref[1]
        ref[0,2,2,0]=(0.836447791155)*x_ref[0]+(1.2627066026)*x_ref[1]
        ref[0,2,2,1]=(-0.584608492688)*x_ref[0]+(-1.55305974305)*x_ref[1]
        ref[0,3,0,0]=(-0.661952381799)*x_ref[0]+(-1.27062403778)*x_ref[1]
        ref[0,3,0,1]=(-1.47789532259)*x_ref[0]+(-0.279583524936)*x_ref[1]
        ref[0,3,1,0]=(1.40729771357)*x_ref[0]+(0.0971284144528)*x_ref[1]
        ref[0,3,1,1]=(-0.459167955911)*x_ref[0]+(-1.29033281524)*x_ref[1]
        ref[0,3,2,0]=(-1.63189436824)*x_ref[0]+(0.75729808422)*x_ref[1]
        ref[0,3,2,1]=(0.522479828718)*x_ref[0]+(-0.0231299515383)*x_ref[1]
        ref[0,4,0,0]=(0.241891446615)*x_ref[0]+(0.803327107651)*x_ref[1]
        ref[0,4,0,1]=(0.273613502836)*x_ref[0]+(-0.0115180464171)*x_ref[1]
        ref[0,4,1,0]=(1.96219818696)*x_ref[0]+(-0.349137094743)*x_ref[1]
        ref[0,4,1,1]=(0.365690821719)*x_ref[0]+(-1.6943347181)*x_ref[1]
        ref[0,4,2,0]=(-0.283970779872)*x_ref[0]+(-0.258537179804)*x_ref[1]
        ref[0,4,2,1]=(-0.973333430598)*x_ref[0]+(1.48706276212)*x_ref[1]
        ref[1,0,0,0]=(1.18757466971)*x_ref[0]+(0.838804656357)*x_ref[1]
        ref[1,0,0,1]=(0.840859904702)*x_ref[0]+(1.09955195682)*x_ref[1]
        ref[1,0,1,0]=(1.46190513227)*x_ref[0]+(-0.596959830422)*x_ref[1]
        ref[1,0,1,1]=(0.718689296745)*x_ref[0]+(-0.628074620222)*x_ref[1]
        ref[1,0,2,0]=(0.440761293041)*x_ref[0]+(0.115800854194)*x_ref[1]
        ref[1,0,2,1]=(1.34891146904)*x_ref[0]+(0.813690260652)*x_ref[1]
        ref[1,1,0,0]=(0.183071638098)*x_ref[0]+(-1.37996210755)*x_ref[1]
        ref[1,1,0,1]=(0.258142115988)*x_ref[0]+(-0.757709381529)*x_ref[1]
        ref[1,1,1,0]=(1.72392829309)*x_ref[0]+(0.444995102781)*x_ref[1]
        ref[1,1,1,1]=(-0.380898209988)*x_ref[0]+(-0.0950246860347)*x_ref[1]
        ref[1,1,2,0]=(0.203007226016)*x_ref[0]+(1.09786001832)*x_ref[1]
        ref[1,1,2,1]=(-0.502500865341)*x_ref[0]+(-0.629980527995)*x_ref[1]
        ref[1,2,0,0]=(0.482549518436)*x_ref[0]+(0.565840824146)*x_ref[1]
        ref[1,2,0,1]=(0.355818565068)*x_ref[0]+(-0.78273295574)*x_ref[1]
        ref[1,2,1,0]=(1.79204033071)*x_ref[0]+(-0.698040851585)*x_ref[1]
        ref[1,2,1,1]=(-0.534827741044)*x_ref[0]+(0.00943131627498)*x_ref[1]
        ref[1,2,2,0]=(-1.00663524444)*x_ref[0]+(0.0207854553456)*x_ref[1]
        ref[1,2,2,1]=(-0.338048040159)*x_ref[0]+(-0.380607406217)*x_ref[1]
        ref[1,3,0,0]=(0.586310549044)*x_ref[0]+(0.276326798805)*x_ref[1]
        ref[1,3,0,1]=(-0.260482580702)*x_ref[0]+(-0.349279395321)*x_ref[1]
        ref[1,3,1,0]=(-0.196533829895)*x_ref[0]+(0.614805272738)*x_ref[1]
        ref[1,3,1,1]=(-0.363176483895)*x_ref[0]+(-0.925817017731)*x_ref[1]
        ref[1,3,2,0]=(-1.24182075421)*x_ref[0]+(0.215966446013)*x_ref[1]
        ref[1,3,2,1]=(1.07546652642)*x_ref[0]+(0.383803991974)*x_ref[1]
        ref[1,4,0,0]=(0.199312303525)*x_ref[0]+(-0.314520883859)*x_ref[1]
        ref[1,4,0,1]=(-1.56890826152)*x_ref[0]+(-0.133086662241)*x_ref[1]
        ref[1,4,1,0]=(1.1307241936)*x_ref[0]+(-1.32905744984)*x_ref[1]
        ref[1,4,1,1]=(-0.385991455298)*x_ref[0]+(0.292913024507)*x_ref[1]
        ref[1,4,2,0]=(1.14561678056)*x_ref[0]+(-0.662696162276)*x_ref[1]
        ref[1,4,2,1]=(-1.49533948954)*x_ref[0]+(1.43396138744)*x_ref[1]
        ref[2,0,0,0]=(-0.63160396493)*x_ref[0]+(0.813120998776)*x_ref[1]
        ref[2,0,0,1]=(0.65598998434)*x_ref[0]+(-1.02445537553)*x_ref[1]
        ref[2,0,1,0]=(-1.13473104075)*x_ref[0]+(0.057331183449)*x_ref[1]
        ref[2,0,1,1]=(-0.154892482107)*x_ref[0]+(0.481969957579)*x_ref[1]
        ref[2,0,2,0]=(0.521916671496)*x_ref[0]+(0.546676943059)*x_ref[1]
        ref[2,0,2,1]=(-1.68919379028)*x_ref[0]+(1.08292038981)*x_ref[1]
        ref[2,1,0,0]=(-0.643846595888)*x_ref[0]+(0.528306097755)*x_ref[1]
        ref[2,1,0,1]=(1.3745090497)*x_ref[0]+(-0.0801648509184)*x_ref[1]
        ref[2,1,1,0]=(0.129096817054)*x_ref[0]+(0.90750370855)*x_ref[1]
        ref[2,1,1,1]=(0.539309513305)*x_ref[0]+(1.34949470961)*x_ref[1]
        ref[2,1,2,0]=(1.4189703896)*x_ref[0]+(-1.75011201893)*x_ref[1]
        ref[2,1,2,1]=(-1.09325548386)*x_ref[0]+(0.148279301716)*x_ref[1]
        ref[2,2,0,0]=(-1.25785133743)*x_ref[0]+(0.275071306594)*x_ref[1]
        ref[2,2,0,1]=(-0.550152068761)*x_ref[0]+(-1.20422484275)*x_ref[1]
        ref[2,2,1,0]=(-0.102062863668)*x_ref[0]+(-1.34431887541)*x_ref[1]
        ref[2,2,1,1]=(0.103354724169)*x_ref[0]+(-0.941952180446)*x_ref[1]
        ref[2,2,2,0]=(1.2090339604)*x_ref[0]+(-0.330357976796)*x_ref[1]
        ref[2,2,2,1]=(0.203162449188)*x_ref[0]+(-1.03640702014)*x_ref[1]
        ref[2,3,0,0]=(-0.324576353514)*x_ref[0]+(0.582614539267)*x_ref[1]
        ref[2,3,0,1]=(-0.213823136215)*x_ref[0]+(-0.473199105214)*x_ref[1]
        ref[2,3,1,0]=(0.395144702599)*x_ref[0]+(0.0358194467152)*x_ref[1]
        ref[2,3,1,1]=(1.32860563396)*x_ref[0]+(-0.577308310732)*x_ref[1]
        ref[2,3,2,0]=(0.994205029432)*x_ref[0]+(1.13433517636)*x_ref[1]
        ref[2,3,2,1]=(-0.488499286742)*x_ref[0]+(0.224253820738)*x_ref[1]
        ref[2,4,0,0]=(1.48035663008)*x_ref[0]+(-1.09697986914)*x_ref[1]
        ref[2,4,0,1]=(-1.09991915151)*x_ref[0]+(-0.557951199985)*x_ref[1]
        ref[2,4,1,0]=(-1.23862699804)*x_ref[0]+(-0.230602187713)*x_ref[1]
        ref[2,4,1,1]=(-0.922609370448)*x_ref[0]+(1.76673953413)*x_ref[1]
        ref[2,4,2,0]=(0.731769419709)*x_ref[0]+(-1.1094516608)*x_ref[1]
        ref[2,4,2,1]=(-0.145672504805)*x_ref[0]+(0.168194421355)*x_ref[1]
        ref[3,0,0,0]=(1.02559707602)*x_ref[0]+(0.260431368523)*x_ref[1]
        ref[3,0,0,1]=(0.111899568824)*x_ref[0]+(-0.90725475423)*x_ref[1]
        ref[3,0,1,0]=(-0.226547745683)*x_ref[0]+(-0.232830393372)*x_ref[1]
        ref[3,0,1,1]=(0.347004318471)*x_ref[0]+(-0.196130970001)*x_ref[1]
        ref[3,0,2,0]=(-1.72155871442)*x_ref[0]+(-1.28372407452)*x_ref[1]
        ref[3,0,2,1]=(1.43817914693)*x_ref[0]+(-0.354093253684)*x_ref[1]
        ref[3,1,0,0]=(-0.293939245102)*x_ref[0]+(0.0255562139103)*x_ref[1]
        ref[3,1,0,1]=(0.139306213018)*x_ref[0]+(0.595062226642)*x_ref[1]
        ref[3,1,1,0]=(-0.670339920672)*x_ref[0]+(0.387520583192)*x_ref[1]
        ref[3,1,1,1]=(-0.0323325470954)*x_ref[0]+(-0.681162999279)*x_ref[1]
        ref[3,1,2,0]=(0.235275048739)*x_ref[0]+(0.144725473773)*x_ref[1]
        ref[3,1,2,1]=(0.132211795214)*x_ref[0]+(0.262780216845)*x_ref[1]
        ref[3,2,0,0]=(0.590514326163)*x_ref[0]+(-0.84273677918)*x_ref[1]
        ref[3,2,0,1]=(-0.925181556185)*x_ref[0]+(-0.300244447001)*x_ref[1]
        ref[3,2,1,0]=(-0.0509700795112)*x_ref[0]+(-1.31919389616)*x_ref[1]
        ref[3,2,1,1]=(1.50577291846)*x_ref[0]+(-0.219284607769)*x_ref[1]
        ref[3,2,2,0]=(-0.0460616673346)*x_ref[0]+(-0.690312685466)*x_ref[1]
        ref[3,2,2,1]=(-0.673765477936)*x_ref[0]+(-0.0965860475641)*x_ref[1]
        ref[3,3,0,0]=(-0.6577594174)*x_ref[0]+(0.171886277016)*x_ref[1]
        ref[3,3,0,1]=(-1.08810434718)*x_ref[0]+(0.333153250183)*x_ref[1]
        ref[3,3,1,0]=(0.150968371983)*x_ref[0]+(0.290859317391)*x_ref[1]
        ref[3,3,1,1]=(1.62096265131)*x_ref[0]+(-1.10309602741)*x_ref[1]
        ref[3,3,2,0]=(-0.932063639972)*x_ref[0]+(-0.897131561126)*x_ref[1]
        ref[3,3,2,1]=(-0.819110834345)*x_ref[0]+(1.59159109562)*x_ref[1]
        ref[3,4,0,0]=(0.272662931155)*x_ref[0]+(0.69026285756)*x_ref[1]
        ref[3,4,0,1]=(-0.600556370757)*x_ref[0]+(-0.0853533546302)*x_ref[1]
        ref[3,4,1,0]=(1.18544735281)*x_ref[0]+(0.217751189034)*x_ref[1]
        ref[3,4,1,1]=(-0.594609531358)*x_ref[0]+(0.882158349788)*x_ref[1]
        ref[3,4,2,0]=(1.18876765552)*x_ref[0]+(-0.411481586979)*x_ref[1]
        ref[3,4,2,1]=(-1.093067792)*x_ref[0]+(-0.266744260768)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.29218217256)*x[0]+(-0.382937892213)*x[1]+(0.14363896447)*x[2]
        arg[0,0,0,1]=(1.39856942421)*x[0]+(0.302082267835)*x[1]+(-0.40870736584)*x[2]
        arg[0,0,1,0]=(0.523846108043)*x[0]+(1.19961531204)*x[1]+(0.838881761349)*x[2]
        arg[0,0,1,1]=(0.634542282764)*x[0]+(-0.475514686466)*x[1]+(-1.68745428486)*x[2]
        arg[0,0,2,0]=(1.16379585731)*x[0]+(-0.928037595416)*x[1]+(0.998857178353)*x[2]
        arg[0,0,2,1]=(-0.804515607022)*x[0]+(-0.915469493439)*x[1]+(-0.143888000623)*x[2]
        arg[0,1,0,0]=(0.925843155486)*x[0]+(-0.168243418953)*x[1]+(-0.887839555187)*x[2]
        arg[0,1,0,1]=(-0.913103105329)*x[0]+(1.38011681732)*x[1]+(-0.992478793291)*x[2]
        arg[0,1,1,0]=(0.759685289903)*x[0]+(-0.579481939982)*x[1]+(0.577682929138)*x[2]
        arg[0,1,1,1]=(0.813155548228)*x[0]+(-0.178317843217)*x[1]+(-0.0460113194526)*x[2]
        arg[0,1,2,0]=(-0.195764860141)*x[0]+(1.44256979628)*x[1]+(0.283992441343)*x[2]
        arg[0,1,2,1]=(-0.921229459072)*x[0]+(-1.19295761547)*x[1]+(0.215736801944)*x[2]
        arg[0,2,0,0]=(-0.741726331752)*x[0]+(-0.225585937382)*x[1]+(1.12654556466)*x[2]
        arg[0,2,0,1]=(-0.709656162421)*x[0]+(0.298862010425)*x[1]+(-0.502443284827)*x[2]
        arg[0,2,1,0]=(-1.01166188669)*x[0]+(-1.43763408953)*x[1]+(-0.835160398488)*x[2]
        arg[0,2,1,1]=(1.80968424545)*x[0]+(-0.802666103495)*x[1]+(0.290283606115)*x[2]
        arg[0,2,2,0]=(0.346481981585)*x[0]+(-0.444176962215)*x[1]+(0.689127612227)*x[2]
        arg[0,2,2,1]=(1.06925494095)*x[0]+(0.452854988449)*x[1]+(-1.19974690364)*x[2]
        arg[0,3,0,0]=(-0.534896036328)*x[0]+(-0.312000770615)*x[1]+(-0.440033746693)*x[2]
        arg[0,3,0,1]=(0.580351431838)*x[0]+(-0.36623848451)*x[1]+(-0.507837490037)*x[2]
        arg[0,3,1,0]=(0.383180910522)*x[0]+(-0.0907152443563)*x[1]+(-0.200103743943)*x[2]
        arg[0,3,1,1]=(0.0160194687513)*x[0]+(0.862509612939)*x[1]+(-1.14643594474)*x[2]
        arg[0,3,2,0]=(-0.224390850199)*x[0]+(-0.176927715447)*x[1]+(0.213480525831)*x[2]
        arg[0,3,2,1]=(-0.0508511448148)*x[0]+(-1.28943467389)*x[1]+(0.902285302487)*x[2]
        arg[0,4,0,0]=(-0.633106625232)*x[0]+(-0.430651917485)*x[1]+(0.583149841952)*x[2]
        arg[0,4,0,1]=(0.579035528956)*x[0]+(0.159439682519)*x[1]+(-0.0926277063021)*x[2]
        arg[0,4,1,0]=(0.00484756366973)*x[0]+(0.0324822270755)*x[1]+(1.20052413378)*x[2]
        arg[0,4,1,1]=(-0.337383994688)*x[0]+(0.902076367981)*x[1]+(-0.1956377395)*x[2]
        arg[0,4,2,0]=(0.928434343456)*x[0]+(0.0376569884355)*x[1]+(-0.155274118568)*x[2]
        arg[0,4,2,1]=(-0.541990213238)*x[0]+(-0.0165076697829)*x[1]+(-0.210319224859)*x[2]
        arg[1,0,0,0]=(-1.67762025181)*x[0]+(1.61434755579)*x[1]+(0.357272049999)*x[2]
        arg[1,0,0,1]=(-1.62532544549)*x[0]+(-0.271426231027)*x[1]+(0.338680991778)*x[2]
        arg[1,0,1,0]=(1.73523408033)*x[0]+(1.11987092368)*x[1]+(0.126769364273)*x[2]
        arg[1,0,1,1]=(-1.49220786558)*x[0]+(1.51915566092)*x[1]+(1.24788192157)*x[2]
        arg[1,0,2,0]=(-0.687808766765)*x[0]+(-1.02318754737)*x[1]+(0.590653972416)*x[2]
        arg[1,0,2,1]=(0.756067354802)*x[0]+(-0.66964706965)*x[1]+(-1.18986499158)*x[2]
        arg[1,1,0,0]=(0.371345732177)*x[0]+(0.212739617115)*x[1]+(-1.13439047405)*x[2]
        arg[1,1,0,1]=(0.525761804111)*x[0]+(0.144130399691)*x[1]+(0.715517587677)*x[2]
        arg[1,1,1,0]=(-0.376160293215)*x[0]+(-0.469964935263)*x[1]+(0.413518194707)*x[2]
        arg[1,1,1,1]=(-0.787748398144)*x[0]+(-0.810090402237)*x[1]+(-0.50127109256)*x[2]
        arg[1,1,2,0]=(0.654555083692)*x[0]+(-0.142134700515)*x[1]+(-1.55891930521)*x[2]
        arg[1,1,2,1]=(-0.423058885456)*x[0]+(-0.65930158641)*x[1]+(0.94440057744)*x[2]
        arg[1,2,0,0]=(-0.924832007053)*x[0]+(1.0154890382)*x[1]+(0.11938079748)*x[2]
        arg[1,2,0,1]=(0.134642686931)*x[0]+(-1.30908067998)*x[1]+(0.367206139074)*x[2]
        arg[1,2,1,0]=(1.55524055201)*x[0]+(0.727208660637)*x[1]+(0.809810917377)*x[2]
        arg[1,2,1,1]=(0.94813391462)*x[0]+(-1.49402524308)*x[1]+(0.367824542434)*x[2]
        arg[1,2,2,0]=(-0.608295798731)*x[0]+(-0.142302252814)*x[1]+(-0.986529251308)*x[2]
        arg[1,2,2,1]=(0.0259394619607)*x[0]+(-1.60728202467)*x[1]+(0.312534656619)*x[2]
        arg[1,3,0,0]=(-1.34063373729)*x[0]+(-0.268499839329)*x[1]+(-0.335125886299)*x[2]
        arg[1,3,0,1]=(1.53280698072)*x[0]+(1.24369684711)*x[1]+(0.766369156132)*x[2]
        arg[1,3,1,0]=(0.902067176876)*x[0]+(0.791102994814)*x[1]+(-0.89266430493)*x[2]
        arg[1,3,1,1]=(0.704997629232)*x[0]+(-0.551611360332)*x[1]+(1.39134537491)*x[2]
        arg[1,3,2,0]=(-0.151690675211)*x[0]+(-1.01762975893)*x[1]+(-0.627493214284)*x[2]
        arg[1,3,2,1]=(1.00936645922)*x[0]+(0.0322458997248)*x[1]+(1.45039938194)*x[2]
        arg[1,4,0,0]=(-0.0916390388904)*x[0]+(-0.783111746469)*x[1]+(0.389725342225)*x[2]
        arg[1,4,0,1]=(1.18859328243)*x[0]+(0.871689661558)*x[1]+(-0.775464713448)*x[2]
        arg[1,4,1,0]=(0.913118342488)*x[0]+(1.53285648884)*x[1]+(0.0947365101243)*x[2]
        arg[1,4,1,1]=(-1.00735841224)*x[0]+(-1.12544193607)*x[1]+(0.267347229372)*x[2]
        arg[1,4,2,0]=(0.285943438284)*x[0]+(-1.0889389244)*x[1]+(-1.03054842327)*x[2]
        arg[1,4,2,1]=(-1.5768940139)*x[0]+(0.91237754587)*x[1]+(1.30220394199)*x[2]
        arg[2,0,0,0]=(-0.0634935883277)*x[0]+(0.157304131712)*x[1]+(-1.2995042074)*x[2]
        arg[2,0,0,1]=(-0.802705266871)*x[0]+(0.238233162774)*x[1]+(1.29463414013)*x[2]
        arg[2,0,1,0]=(-0.925468462671)*x[0]+(0.870960795966)*x[1]+(0.156522765713)*x[2]
        arg[2,0,1,1]=(-0.546666833309)*x[0]+(-1.14649607995)*x[1]+(-1.53906505048)*x[2]
        arg[2,0,2,0]=(0.464952910857)*x[0]+(0.408605786027)*x[1]+(-0.886974881114)*x[2]
        arg[2,0,2,1]=(0.491723247948)*x[0]+(0.194616420077)*x[1]+(0.409049782749)*x[2]
        arg[2,1,0,0]=(-1.19836883114)*x[0]+(1.0622722337)*x[1]+(-0.109193867727)*x[2]
        arg[2,1,0,1]=(1.01973600383)*x[0]+(-0.154482220539)*x[1]+(0.378696371956)*x[2]
        arg[2,1,1,0]=(-0.0812015296246)*x[0]+(-1.10252794568)*x[1]+(-0.00585426581857)*x[2]
        arg[2,1,1,1]=(0.536042665272)*x[0]+(0.685507117255)*x[1]+(-0.179652683139)*x[2]
        arg[2,1,2,0]=(-0.36750467811)*x[0]+(-1.23674550869)*x[1]+(-0.762905702801)*x[2]
        arg[2,1,2,1]=(0.329665965651)*x[0]+(0.101195810737)*x[1]+(-0.610015052375)*x[2]
        arg[2,2,0,0]=(-0.81644635185)*x[0]+(-0.604578476306)*x[1]+(0.0440925706991)*x[2]
        arg[2,2,0,1]=(-0.958241437971)*x[0]+(0.38563498206)*x[1]+(0.913406762318)*x[2]
        arg[2,2,1,0]=(1.22027492224)*x[0]+(-1.00726209939)*x[1]+(-1.05083068303)*x[2]
        arg[2,2,1,1]=(-1.47561192645)*x[0]+(-0.903549355608)*x[1]+(0.236771324671)*x[2]
        arg[2,2,2,0]=(0.284995864449)*x[0]+(-0.175760829921)*x[1]+(-0.720607220001)*x[2]
        arg[2,2,2,1]=(-0.535894515963)*x[0]+(-0.821302957162)*x[1]+(-1.75141540241)*x[2]
        arg[2,3,0,0]=(-0.333082106674)*x[0]+(0.579471826412)*x[1]+(0.561184138108)*x[2]
        arg[2,3,0,1]=(1.34803519557)*x[0]+(-0.441299088849)*x[1]+(-1.04945951616)*x[2]
        arg[2,3,1,0]=(0.299011096602)*x[0]+(1.32726458235)*x[1]+(-0.196298687359)*x[2]
        arg[2,3,1,1]=(1.48698834204)*x[0]+(0.527221496971)*x[1]+(0.0819813519422)*x[2]
        arg[2,3,2,0]=(0.681680955153)*x[0]+(0.503653972357)*x[1]+(0.434669347891)*x[2]
        arg[2,3,2,1]=(-0.152586789311)*x[0]+(-0.765242635579)*x[1]+(0.712663994387)*x[2]
        arg[2,4,0,0]=(0.630152300463)*x[0]+(1.12366733588)*x[1]+(-0.662966700638)*x[2]
        arg[2,4,0,1]=(1.02796709755)*x[0]+(-0.24093881584)*x[1]+(1.79445417019)*x[2]
        arg[2,4,1,0]=(-0.954575760758)*x[0]+(0.16331782608)*x[1]+(1.38309921277)*x[2]
        arg[2,4,1,1]=(0.536092514281)*x[0]+(-0.276364952673)*x[1]+(0.779619701528)*x[2]
        arg[2,4,2,0]=(0.267270444141)*x[0]+(-0.810746424331)*x[1]+(-0.733877853264)*x[2]
        arg[2,4,2,1]=(-0.0641688847923)*x[0]+(-0.386279280667)*x[1]+(0.0637571595255)*x[2]
        arg[3,0,0,0]=(0.469903180687)*x[0]+(-0.793812397342)*x[1]+(1.60841921779)*x[2]
        arg[3,0,0,1]=(0.720057600102)*x[0]+(0.502582253261)*x[1]+(0.460147791003)*x[2]
        arg[3,0,1,0]=(0.0327862564137)*x[0]+(-1.142962965)*x[1]+(0.114799131894)*x[2]
        arg[3,0,1,1]=(-0.997412656978)*x[0]+(-0.678175468576)*x[1]+(-0.152663147682)*x[2]
        arg[3,0,2,0]=(0.810424599078)*x[0]+(0.75522727786)*x[1]+(-0.205880941501)*x[2]
        arg[3,0,2,1]=(0.353601949476)*x[0]+(0.0963200813379)*x[1]+(0.0578829287146)*x[2]
        arg[3,1,0,0]=(0.346590730062)*x[0]+(1.07953790092)*x[1]+(0.647586851217)*x[2]
        arg[3,1,0,1]=(0.889879195334)*x[0]+(-0.352788186249)*x[1]+(0.434900135803)*x[2]
        arg[3,1,1,0]=(-0.721742734959)*x[0]+(0.166066672134)*x[1]+(-0.489266924809)*x[2]
        arg[3,1,1,1]=(-0.810487755181)*x[0]+(0.0354916979253)*x[1]+(0.0735520654395)*x[2]
        arg[3,1,2,0]=(1.02414338939)*x[0]+(-0.851076284556)*x[1]+(-0.232174494405)*x[2]
        arg[3,1,2,1]=(-0.327246566026)*x[0]+(-0.0499396023605)*x[1]+(0.611585402379)*x[2]
        arg[3,2,0,0]=(-0.365757628557)*x[0]+(0.868697342115)*x[1]+(-1.102545972)*x[2]
        arg[3,2,0,1]=(0.628592129227)*x[0]+(1.25545515311)*x[1]+(-1.0577476264)*x[2]
        arg[3,2,1,0]=(0.160995978044)*x[0]+(-0.947389499556)*x[1]+(-0.870476160656)*x[2]
        arg[3,2,1,1]=(0.910787139938)*x[0]+(-0.0681493345139)*x[1]+(-0.561910153781)*x[2]
        arg[3,2,2,0]=(0.0761644634428)*x[0]+(-0.700940967343)*x[1]+(1.19407210904)*x[2]
        arg[3,2,2,1]=(1.75717437065)*x[0]+(-0.144457317494)*x[1]+(-0.523233453079)*x[2]
        arg[3,3,0,0]=(0.448782872606)*x[0]+(0.650464256547)*x[1]+(1.19476547456)*x[2]
        arg[3,3,0,1]=(0.118586027201)*x[0]+(0.449971816918)*x[1]+(-0.510812733389)*x[2]
        arg[3,3,1,0]=(-0.330032265081)*x[0]+(-0.13657534891)*x[1]+(0.456365289556)*x[2]
        arg[3,3,1,1]=(0.697584212635)*x[0]+(0.897604241724)*x[1]+(1.31517116663)*x[2]
        arg[3,3,2,0]=(-0.940538707365)*x[0]+(-0.334862221665)*x[1]+(0.0969742096754)*x[2]
        arg[3,3,2,1]=(-0.985788813516)*x[0]+(0.348389725548)*x[1]+(-0.260816590361)*x[2]
        arg[3,4,0,0]=(0.754692421417)*x[0]+(0.775985781769)*x[1]+(-0.0866159222457)*x[2]
        arg[3,4,0,1]=(-1.17670502429)*x[0]+(1.12268279393)*x[1]+(-1.1279822095)*x[2]
        arg[3,4,1,0]=(-1.48674370666)*x[0]+(1.26685445713)*x[1]+(-0.0522828809306)*x[2]
        arg[3,4,1,1]=(0.401760689843)*x[0]+(-1.35707797971)*x[1]+(-0.0580654273828)*x[2]
        arg[3,4,2,0]=(-1.40697830852)*x[0]+(0.933923369958)*x[1]+(0.230152169159)*x[2]
        arg[3,4,2,1]=(-1.21242130555)*x[0]+(0.858316339663)*x[1]+(-0.839747078526)*x[2]
        ref[0,0,0,0]=(-1.29218217256)*x_ref[0]+(-0.382937892213)*x_ref[1]+(0.14363896447)*x_ref[2]
        ref[0,0,0,1]=(1.39856942421)*x_ref[0]+(0.302082267835)*x_ref[1]+(-0.40870736584)*x_ref[2]
        ref[0,0,1,0]=(0.523846108043)*x_ref[0]+(1.19961531204)*x_ref[1]+(0.838881761349)*x_ref[2]
        ref[0,0,1,1]=(0.634542282764)*x_ref[0]+(-0.475514686466)*x_ref[1]+(-1.68745428486)*x_ref[2]
        ref[0,0,2,0]=(1.16379585731)*x_ref[0]+(-0.928037595416)*x_ref[1]+(0.998857178353)*x_ref[2]
        ref[0,0,2,1]=(-0.804515607022)*x_ref[0]+(-0.915469493439)*x_ref[1]+(-0.143888000623)*x_ref[2]
        ref[0,1,0,0]=(0.925843155486)*x_ref[0]+(-0.168243418953)*x_ref[1]+(-0.887839555187)*x_ref[2]
        ref[0,1,0,1]=(-0.913103105329)*x_ref[0]+(1.38011681732)*x_ref[1]+(-0.992478793291)*x_ref[2]
        ref[0,1,1,0]=(0.759685289903)*x_ref[0]+(-0.579481939982)*x_ref[1]+(0.577682929138)*x_ref[2]
        ref[0,1,1,1]=(0.813155548228)*x_ref[0]+(-0.178317843217)*x_ref[1]+(-0.0460113194526)*x_ref[2]
        ref[0,1,2,0]=(-0.195764860141)*x_ref[0]+(1.44256979628)*x_ref[1]+(0.283992441343)*x_ref[2]
        ref[0,1,2,1]=(-0.921229459072)*x_ref[0]+(-1.19295761547)*x_ref[1]+(0.215736801944)*x_ref[2]
        ref[0,2,0,0]=(-0.741726331752)*x_ref[0]+(-0.225585937382)*x_ref[1]+(1.12654556466)*x_ref[2]
        ref[0,2,0,1]=(-0.709656162421)*x_ref[0]+(0.298862010425)*x_ref[1]+(-0.502443284827)*x_ref[2]
        ref[0,2,1,0]=(-1.01166188669)*x_ref[0]+(-1.43763408953)*x_ref[1]+(-0.835160398488)*x_ref[2]
        ref[0,2,1,1]=(1.80968424545)*x_ref[0]+(-0.802666103495)*x_ref[1]+(0.290283606115)*x_ref[2]
        ref[0,2,2,0]=(0.346481981585)*x_ref[0]+(-0.444176962215)*x_ref[1]+(0.689127612227)*x_ref[2]
        ref[0,2,2,1]=(1.06925494095)*x_ref[0]+(0.452854988449)*x_ref[1]+(-1.19974690364)*x_ref[2]
        ref[0,3,0,0]=(-0.534896036328)*x_ref[0]+(-0.312000770615)*x_ref[1]+(-0.440033746693)*x_ref[2]
        ref[0,3,0,1]=(0.580351431838)*x_ref[0]+(-0.36623848451)*x_ref[1]+(-0.507837490037)*x_ref[2]
        ref[0,3,1,0]=(0.383180910522)*x_ref[0]+(-0.0907152443563)*x_ref[1]+(-0.200103743943)*x_ref[2]
        ref[0,3,1,1]=(0.0160194687513)*x_ref[0]+(0.862509612939)*x_ref[1]+(-1.14643594474)*x_ref[2]
        ref[0,3,2,0]=(-0.224390850199)*x_ref[0]+(-0.176927715447)*x_ref[1]+(0.213480525831)*x_ref[2]
        ref[0,3,2,1]=(-0.0508511448148)*x_ref[0]+(-1.28943467389)*x_ref[1]+(0.902285302487)*x_ref[2]
        ref[0,4,0,0]=(-0.633106625232)*x_ref[0]+(-0.430651917485)*x_ref[1]+(0.583149841952)*x_ref[2]
        ref[0,4,0,1]=(0.579035528956)*x_ref[0]+(0.159439682519)*x_ref[1]+(-0.0926277063021)*x_ref[2]
        ref[0,4,1,0]=(0.00484756366973)*x_ref[0]+(0.0324822270755)*x_ref[1]+(1.20052413378)*x_ref[2]
        ref[0,4,1,1]=(-0.337383994688)*x_ref[0]+(0.902076367981)*x_ref[1]+(-0.1956377395)*x_ref[2]
        ref[0,4,2,0]=(0.928434343456)*x_ref[0]+(0.0376569884355)*x_ref[1]+(-0.155274118568)*x_ref[2]
        ref[0,4,2,1]=(-0.541990213238)*x_ref[0]+(-0.0165076697829)*x_ref[1]+(-0.210319224859)*x_ref[2]
        ref[1,0,0,0]=(-1.67762025181)*x_ref[0]+(1.61434755579)*x_ref[1]+(0.357272049999)*x_ref[2]
        ref[1,0,0,1]=(-1.62532544549)*x_ref[0]+(-0.271426231027)*x_ref[1]+(0.338680991778)*x_ref[2]
        ref[1,0,1,0]=(1.73523408033)*x_ref[0]+(1.11987092368)*x_ref[1]+(0.126769364273)*x_ref[2]
        ref[1,0,1,1]=(-1.49220786558)*x_ref[0]+(1.51915566092)*x_ref[1]+(1.24788192157)*x_ref[2]
        ref[1,0,2,0]=(-0.687808766765)*x_ref[0]+(-1.02318754737)*x_ref[1]+(0.590653972416)*x_ref[2]
        ref[1,0,2,1]=(0.756067354802)*x_ref[0]+(-0.66964706965)*x_ref[1]+(-1.18986499158)*x_ref[2]
        ref[1,1,0,0]=(0.371345732177)*x_ref[0]+(0.212739617115)*x_ref[1]+(-1.13439047405)*x_ref[2]
        ref[1,1,0,1]=(0.525761804111)*x_ref[0]+(0.144130399691)*x_ref[1]+(0.715517587677)*x_ref[2]
        ref[1,1,1,0]=(-0.376160293215)*x_ref[0]+(-0.469964935263)*x_ref[1]+(0.413518194707)*x_ref[2]
        ref[1,1,1,1]=(-0.787748398144)*x_ref[0]+(-0.810090402237)*x_ref[1]+(-0.50127109256)*x_ref[2]
        ref[1,1,2,0]=(0.654555083692)*x_ref[0]+(-0.142134700515)*x_ref[1]+(-1.55891930521)*x_ref[2]
        ref[1,1,2,1]=(-0.423058885456)*x_ref[0]+(-0.65930158641)*x_ref[1]+(0.94440057744)*x_ref[2]
        ref[1,2,0,0]=(-0.924832007053)*x_ref[0]+(1.0154890382)*x_ref[1]+(0.11938079748)*x_ref[2]
        ref[1,2,0,1]=(0.134642686931)*x_ref[0]+(-1.30908067998)*x_ref[1]+(0.367206139074)*x_ref[2]
        ref[1,2,1,0]=(1.55524055201)*x_ref[0]+(0.727208660637)*x_ref[1]+(0.809810917377)*x_ref[2]
        ref[1,2,1,1]=(0.94813391462)*x_ref[0]+(-1.49402524308)*x_ref[1]+(0.367824542434)*x_ref[2]
        ref[1,2,2,0]=(-0.608295798731)*x_ref[0]+(-0.142302252814)*x_ref[1]+(-0.986529251308)*x_ref[2]
        ref[1,2,2,1]=(0.0259394619607)*x_ref[0]+(-1.60728202467)*x_ref[1]+(0.312534656619)*x_ref[2]
        ref[1,3,0,0]=(-1.34063373729)*x_ref[0]+(-0.268499839329)*x_ref[1]+(-0.335125886299)*x_ref[2]
        ref[1,3,0,1]=(1.53280698072)*x_ref[0]+(1.24369684711)*x_ref[1]+(0.766369156132)*x_ref[2]
        ref[1,3,1,0]=(0.902067176876)*x_ref[0]+(0.791102994814)*x_ref[1]+(-0.89266430493)*x_ref[2]
        ref[1,3,1,1]=(0.704997629232)*x_ref[0]+(-0.551611360332)*x_ref[1]+(1.39134537491)*x_ref[2]
        ref[1,3,2,0]=(-0.151690675211)*x_ref[0]+(-1.01762975893)*x_ref[1]+(-0.627493214284)*x_ref[2]
        ref[1,3,2,1]=(1.00936645922)*x_ref[0]+(0.0322458997248)*x_ref[1]+(1.45039938194)*x_ref[2]
        ref[1,4,0,0]=(-0.0916390388904)*x_ref[0]+(-0.783111746469)*x_ref[1]+(0.389725342225)*x_ref[2]
        ref[1,4,0,1]=(1.18859328243)*x_ref[0]+(0.871689661558)*x_ref[1]+(-0.775464713448)*x_ref[2]
        ref[1,4,1,0]=(0.913118342488)*x_ref[0]+(1.53285648884)*x_ref[1]+(0.0947365101243)*x_ref[2]
        ref[1,4,1,1]=(-1.00735841224)*x_ref[0]+(-1.12544193607)*x_ref[1]+(0.267347229372)*x_ref[2]
        ref[1,4,2,0]=(0.285943438284)*x_ref[0]+(-1.0889389244)*x_ref[1]+(-1.03054842327)*x_ref[2]
        ref[1,4,2,1]=(-1.5768940139)*x_ref[0]+(0.91237754587)*x_ref[1]+(1.30220394199)*x_ref[2]
        ref[2,0,0,0]=(-0.0634935883277)*x_ref[0]+(0.157304131712)*x_ref[1]+(-1.2995042074)*x_ref[2]
        ref[2,0,0,1]=(-0.802705266871)*x_ref[0]+(0.238233162774)*x_ref[1]+(1.29463414013)*x_ref[2]
        ref[2,0,1,0]=(-0.925468462671)*x_ref[0]+(0.870960795966)*x_ref[1]+(0.156522765713)*x_ref[2]
        ref[2,0,1,1]=(-0.546666833309)*x_ref[0]+(-1.14649607995)*x_ref[1]+(-1.53906505048)*x_ref[2]
        ref[2,0,2,0]=(0.464952910857)*x_ref[0]+(0.408605786027)*x_ref[1]+(-0.886974881114)*x_ref[2]
        ref[2,0,2,1]=(0.491723247948)*x_ref[0]+(0.194616420077)*x_ref[1]+(0.409049782749)*x_ref[2]
        ref[2,1,0,0]=(-1.19836883114)*x_ref[0]+(1.0622722337)*x_ref[1]+(-0.109193867727)*x_ref[2]
        ref[2,1,0,1]=(1.01973600383)*x_ref[0]+(-0.154482220539)*x_ref[1]+(0.378696371956)*x_ref[2]
        ref[2,1,1,0]=(-0.0812015296246)*x_ref[0]+(-1.10252794568)*x_ref[1]+(-0.00585426581857)*x_ref[2]
        ref[2,1,1,1]=(0.536042665272)*x_ref[0]+(0.685507117255)*x_ref[1]+(-0.179652683139)*x_ref[2]
        ref[2,1,2,0]=(-0.36750467811)*x_ref[0]+(-1.23674550869)*x_ref[1]+(-0.762905702801)*x_ref[2]
        ref[2,1,2,1]=(0.329665965651)*x_ref[0]+(0.101195810737)*x_ref[1]+(-0.610015052375)*x_ref[2]
        ref[2,2,0,0]=(-0.81644635185)*x_ref[0]+(-0.604578476306)*x_ref[1]+(0.0440925706991)*x_ref[2]
        ref[2,2,0,1]=(-0.958241437971)*x_ref[0]+(0.38563498206)*x_ref[1]+(0.913406762318)*x_ref[2]
        ref[2,2,1,0]=(1.22027492224)*x_ref[0]+(-1.00726209939)*x_ref[1]+(-1.05083068303)*x_ref[2]
        ref[2,2,1,1]=(-1.47561192645)*x_ref[0]+(-0.903549355608)*x_ref[1]+(0.236771324671)*x_ref[2]
        ref[2,2,2,0]=(0.284995864449)*x_ref[0]+(-0.175760829921)*x_ref[1]+(-0.720607220001)*x_ref[2]
        ref[2,2,2,1]=(-0.535894515963)*x_ref[0]+(-0.821302957162)*x_ref[1]+(-1.75141540241)*x_ref[2]
        ref[2,3,0,0]=(-0.333082106674)*x_ref[0]+(0.579471826412)*x_ref[1]+(0.561184138108)*x_ref[2]
        ref[2,3,0,1]=(1.34803519557)*x_ref[0]+(-0.441299088849)*x_ref[1]+(-1.04945951616)*x_ref[2]
        ref[2,3,1,0]=(0.299011096602)*x_ref[0]+(1.32726458235)*x_ref[1]+(-0.196298687359)*x_ref[2]
        ref[2,3,1,1]=(1.48698834204)*x_ref[0]+(0.527221496971)*x_ref[1]+(0.0819813519422)*x_ref[2]
        ref[2,3,2,0]=(0.681680955153)*x_ref[0]+(0.503653972357)*x_ref[1]+(0.434669347891)*x_ref[2]
        ref[2,3,2,1]=(-0.152586789311)*x_ref[0]+(-0.765242635579)*x_ref[1]+(0.712663994387)*x_ref[2]
        ref[2,4,0,0]=(0.630152300463)*x_ref[0]+(1.12366733588)*x_ref[1]+(-0.662966700638)*x_ref[2]
        ref[2,4,0,1]=(1.02796709755)*x_ref[0]+(-0.24093881584)*x_ref[1]+(1.79445417019)*x_ref[2]
        ref[2,4,1,0]=(-0.954575760758)*x_ref[0]+(0.16331782608)*x_ref[1]+(1.38309921277)*x_ref[2]
        ref[2,4,1,1]=(0.536092514281)*x_ref[0]+(-0.276364952673)*x_ref[1]+(0.779619701528)*x_ref[2]
        ref[2,4,2,0]=(0.267270444141)*x_ref[0]+(-0.810746424331)*x_ref[1]+(-0.733877853264)*x_ref[2]
        ref[2,4,2,1]=(-0.0641688847923)*x_ref[0]+(-0.386279280667)*x_ref[1]+(0.0637571595255)*x_ref[2]
        ref[3,0,0,0]=(0.469903180687)*x_ref[0]+(-0.793812397342)*x_ref[1]+(1.60841921779)*x_ref[2]
        ref[3,0,0,1]=(0.720057600102)*x_ref[0]+(0.502582253261)*x_ref[1]+(0.460147791003)*x_ref[2]
        ref[3,0,1,0]=(0.0327862564137)*x_ref[0]+(-1.142962965)*x_ref[1]+(0.114799131894)*x_ref[2]
        ref[3,0,1,1]=(-0.997412656978)*x_ref[0]+(-0.678175468576)*x_ref[1]+(-0.152663147682)*x_ref[2]
        ref[3,0,2,0]=(0.810424599078)*x_ref[0]+(0.75522727786)*x_ref[1]+(-0.205880941501)*x_ref[2]
        ref[3,0,2,1]=(0.353601949476)*x_ref[0]+(0.0963200813379)*x_ref[1]+(0.0578829287146)*x_ref[2]
        ref[3,1,0,0]=(0.346590730062)*x_ref[0]+(1.07953790092)*x_ref[1]+(0.647586851217)*x_ref[2]
        ref[3,1,0,1]=(0.889879195334)*x_ref[0]+(-0.352788186249)*x_ref[1]+(0.434900135803)*x_ref[2]
        ref[3,1,1,0]=(-0.721742734959)*x_ref[0]+(0.166066672134)*x_ref[1]+(-0.489266924809)*x_ref[2]
        ref[3,1,1,1]=(-0.810487755181)*x_ref[0]+(0.0354916979253)*x_ref[1]+(0.0735520654395)*x_ref[2]
        ref[3,1,2,0]=(1.02414338939)*x_ref[0]+(-0.851076284556)*x_ref[1]+(-0.232174494405)*x_ref[2]
        ref[3,1,2,1]=(-0.327246566026)*x_ref[0]+(-0.0499396023605)*x_ref[1]+(0.611585402379)*x_ref[2]
        ref[3,2,0,0]=(-0.365757628557)*x_ref[0]+(0.868697342115)*x_ref[1]+(-1.102545972)*x_ref[2]
        ref[3,2,0,1]=(0.628592129227)*x_ref[0]+(1.25545515311)*x_ref[1]+(-1.0577476264)*x_ref[2]
        ref[3,2,1,0]=(0.160995978044)*x_ref[0]+(-0.947389499556)*x_ref[1]+(-0.870476160656)*x_ref[2]
        ref[3,2,1,1]=(0.910787139938)*x_ref[0]+(-0.0681493345139)*x_ref[1]+(-0.561910153781)*x_ref[2]
        ref[3,2,2,0]=(0.0761644634428)*x_ref[0]+(-0.700940967343)*x_ref[1]+(1.19407210904)*x_ref[2]
        ref[3,2,2,1]=(1.75717437065)*x_ref[0]+(-0.144457317494)*x_ref[1]+(-0.523233453079)*x_ref[2]
        ref[3,3,0,0]=(0.448782872606)*x_ref[0]+(0.650464256547)*x_ref[1]+(1.19476547456)*x_ref[2]
        ref[3,3,0,1]=(0.118586027201)*x_ref[0]+(0.449971816918)*x_ref[1]+(-0.510812733389)*x_ref[2]
        ref[3,3,1,0]=(-0.330032265081)*x_ref[0]+(-0.13657534891)*x_ref[1]+(0.456365289556)*x_ref[2]
        ref[3,3,1,1]=(0.697584212635)*x_ref[0]+(0.897604241724)*x_ref[1]+(1.31517116663)*x_ref[2]
        ref[3,3,2,0]=(-0.940538707365)*x_ref[0]+(-0.334862221665)*x_ref[1]+(0.0969742096754)*x_ref[2]
        ref[3,3,2,1]=(-0.985788813516)*x_ref[0]+(0.348389725548)*x_ref[1]+(-0.260816590361)*x_ref[2]
        ref[3,4,0,0]=(0.754692421417)*x_ref[0]+(0.775985781769)*x_ref[1]+(-0.0866159222457)*x_ref[2]
        ref[3,4,0,1]=(-1.17670502429)*x_ref[0]+(1.12268279393)*x_ref[1]+(-1.1279822095)*x_ref[2]
        ref[3,4,1,0]=(-1.48674370666)*x_ref[0]+(1.26685445713)*x_ref[1]+(-0.0522828809306)*x_ref[2]
        ref[3,4,1,1]=(0.401760689843)*x_ref[0]+(-1.35707797971)*x_ref[1]+(-0.0580654273828)*x_ref[2]
        ref[3,4,2,0]=(-1.40697830852)*x_ref[0]+(0.933923369958)*x_ref[1]+(0.230152169159)*x_ref[2]
        ref[3,4,2,1]=(-1.21242130555)*x_ref[0]+(0.858316339663)*x_ref[1]+(-0.839747078526)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.100067148208)*x[0]+(-0.8375743725)*x[1]
        ref=(-0.100067148208)*x_ref[0]+(-0.8375743725)*x_ref[1]
      else:
        arg=(-0.247343416726)*x[0]+(0.604193484164)*x[1]+(-0.219899438287)*x[2]
        ref=(-0.247343416726)*x_ref[0]+(0.604193484164)*x_ref[1]+(-0.219899438287)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        arg[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
        ref[0]=(-1.66196653104)*x[0]+(1.29891335227)*x[1]
        ref[1]=(-0.347844771248)*x[0]+(-0.404726300584)*x[1]
      else:
        arg[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        arg[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
        ref[0]=(0.334094742067)*x[0]+(-0.619063646554)*x[1]+(0.538572422001)*x[2]
        ref[1]=(0.673290433004)*x[0]+(0.169501972843)*x[1]+(-0.833663493893)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.00153698933322)*x[0]+(0.392281645764)*x[1]
        arg[0,1]=(-1.2794655544)*x[0]+(-0.373519639818)*x[1]
        arg[0,2]=(0.30127033321)*x[0]+(-0.0996865853738)*x[1]
        arg[0,3]=(1.23405799349)*x[0]+(0.498558200339)*x[1]
        arg[0,4]=(0.85519350742)*x[0]+(-0.19188046135)*x[1]
        arg[1,0]=(0.417367834962)*x[0]+(0.992318201639)*x[1]
        arg[1,1]=(-0.646432128833)*x[0]+(-0.872750321582)*x[1]
        arg[1,2]=(-0.806117185061)*x[0]+(-0.89396408325)*x[1]
        arg[1,3]=(0.768017237627)*x[0]+(-1.2904111912)*x[1]
        arg[1,4]=(-0.0159685523188)*x[0]+(0.401410193169)*x[1]
        arg[2,0]=(-0.844501716181)*x[0]+(-0.123115343819)*x[1]
        arg[2,1]=(1.08441909409)*x[0]+(0.152432282539)*x[1]
        arg[2,2]=(-0.60551037413)*x[0]+(1.15692828826)*x[1]
        arg[2,3]=(-1.21509168541)*x[0]+(1.03683236377)*x[1]
        arg[2,4]=(1.16850730134)*x[0]+(0.162202615909)*x[1]
        arg[3,0]=(0.727772166897)*x[0]+(-0.296815840935)*x[1]
        arg[3,1]=(-0.0880577151489)*x[0]+(1.66427691426)*x[1]
        arg[3,2]=(0.415144534458)*x[0]+(-0.555833723295)*x[1]
        arg[3,3]=(1.09444635829)*x[0]+(-0.335209417536)*x[1]
        arg[3,4]=(-1.05047769585)*x[0]+(0.684427039986)*x[1]
        ref[0,0]=(0.00153698933322)*x_ref[0]+(0.392281645764)*x_ref[1]
        ref[0,1]=(-1.2794655544)*x_ref[0]+(-0.373519639818)*x_ref[1]
        ref[0,2]=(0.30127033321)*x_ref[0]+(-0.0996865853738)*x_ref[1]
        ref[0,3]=(1.23405799349)*x_ref[0]+(0.498558200339)*x_ref[1]
        ref[0,4]=(0.85519350742)*x_ref[0]+(-0.19188046135)*x_ref[1]
        ref[1,0]=(0.417367834962)*x_ref[0]+(0.992318201639)*x_ref[1]
        ref[1,1]=(-0.646432128833)*x_ref[0]+(-0.872750321582)*x_ref[1]
        ref[1,2]=(-0.806117185061)*x_ref[0]+(-0.89396408325)*x_ref[1]
        ref[1,3]=(0.768017237627)*x_ref[0]+(-1.2904111912)*x_ref[1]
        ref[1,4]=(-0.0159685523188)*x_ref[0]+(0.401410193169)*x_ref[1]
        ref[2,0]=(-0.844501716181)*x_ref[0]+(-0.123115343819)*x_ref[1]
        ref[2,1]=(1.08441909409)*x_ref[0]+(0.152432282539)*x_ref[1]
        ref[2,2]=(-0.60551037413)*x_ref[0]+(1.15692828826)*x_ref[1]
        ref[2,3]=(-1.21509168541)*x_ref[0]+(1.03683236377)*x_ref[1]
        ref[2,4]=(1.16850730134)*x_ref[0]+(0.162202615909)*x_ref[1]
        ref[3,0]=(0.727772166897)*x_ref[0]+(-0.296815840935)*x_ref[1]
        ref[3,1]=(-0.0880577151489)*x_ref[0]+(1.66427691426)*x_ref[1]
        ref[3,2]=(0.415144534458)*x_ref[0]+(-0.555833723295)*x_ref[1]
        ref[3,3]=(1.09444635829)*x_ref[0]+(-0.335209417536)*x_ref[1]
        ref[3,4]=(-1.05047769585)*x_ref[0]+(0.684427039986)*x_ref[1]
      else:
        arg[0,0]=(-0.611607640773)*x[0]+(0.584485971628)*x[1]+(0.0090316337643)*x[2]
        arg[0,1]=(0.905278285857)*x[0]+(0.332523151242)*x[1]+(0.9909841807)*x[2]
        arg[0,2]=(-1.06629333132)*x[0]+(-0.38478180064)*x[1]+(-0.63309211256)*x[2]
        arg[0,3]=(0.122436386344)*x[0]+(-0.218752671204)*x[1]+(-0.746695172216)*x[2]
        arg[0,4]=(1.03152124874)*x[0]+(-1.11327271563)*x[1]+(-0.257591112135)*x[2]
        arg[1,0]=(0.969019600299)*x[0]+(-0.320279309689)*x[1]+(0.327767746545)*x[2]
        arg[1,1]=(0.291023069052)*x[0]+(0.78112901064)*x[1]+(0.879257239042)*x[2]
        arg[1,2]=(-0.131262891536)*x[0]+(-0.0109690229153)*x[1]+(-0.171900332406)*x[2]
        arg[1,3]=(-1.32272684486)*x[0]+(-0.256869536085)*x[1]+(0.0308524567462)*x[2]
        arg[1,4]=(1.30707061279)*x[0]+(0.972866282634)*x[1]+(-0.474948537882)*x[2]
        arg[2,0]=(0.169034210421)*x[0]+(-1.08714896226)*x[1]+(-0.636838450708)*x[2]
        arg[2,1]=(-0.470752780773)*x[0]+(-0.695286088059)*x[1]+(-1.19197058712)*x[2]
        arg[2,2]=(0.701403721228)*x[0]+(0.551002482436)*x[1]+(-0.295853851234)*x[2]
        arg[2,3]=(-0.796431361745)*x[0]+(0.0581884766856)*x[1]+(-1.13668563089)*x[2]
        arg[2,4]=(-0.733185798017)*x[0]+(0.178900103735)*x[1]+(-0.845671526781)*x[2]
        arg[3,0]=(-1.47064927107)*x[0]+(1.33393324965)*x[1]+(-0.767393460937)*x[2]
        arg[3,1]=(1.36784513226)*x[0]+(0.100008910431)*x[1]+(-1.09210256713)*x[2]
        arg[3,2]=(-0.873076669977)*x[0]+(0.367213631778)*x[1]+(0.0698296920039)*x[2]
        arg[3,3]=(-1.76609214158)*x[0]+(-1.26186286566)*x[1]+(0.377817035113)*x[2]
        arg[3,4]=(-1.28261709396)*x[0]+(0.240480767533)*x[1]+(0.307813279357)*x[2]
        ref[0,0]=(-0.611607640773)*x_ref[0]+(0.584485971628)*x_ref[1]+(0.0090316337643)*x_ref[2]
        ref[0,1]=(0.905278285857)*x_ref[0]+(0.332523151242)*x_ref[1]+(0.9909841807)*x_ref[2]
        ref[0,2]=(-1.06629333132)*x_ref[0]+(-0.38478180064)*x_ref[1]+(-0.63309211256)*x_ref[2]
        ref[0,3]=(0.122436386344)*x_ref[0]+(-0.218752671204)*x_ref[1]+(-0.746695172216)*x_ref[2]
        ref[0,4]=(1.03152124874)*x_ref[0]+(-1.11327271563)*x_ref[1]+(-0.257591112135)*x_ref[2]
        ref[1,0]=(0.969019600299)*x_ref[0]+(-0.320279309689)*x_ref[1]+(0.327767746545)*x_ref[2]
        ref[1,1]=(0.291023069052)*x_ref[0]+(0.78112901064)*x_ref[1]+(0.879257239042)*x_ref[2]
        ref[1,2]=(-0.131262891536)*x_ref[0]+(-0.0109690229153)*x_ref[1]+(-0.171900332406)*x_ref[2]
        ref[1,3]=(-1.32272684486)*x_ref[0]+(-0.256869536085)*x_ref[1]+(0.0308524567462)*x_ref[2]
        ref[1,4]=(1.30707061279)*x_ref[0]+(0.972866282634)*x_ref[1]+(-0.474948537882)*x_ref[2]
        ref[2,0]=(0.169034210421)*x_ref[0]+(-1.08714896226)*x_ref[1]+(-0.636838450708)*x_ref[2]
        ref[2,1]=(-0.470752780773)*x_ref[0]+(-0.695286088059)*x_ref[1]+(-1.19197058712)*x_ref[2]
        ref[2,2]=(0.701403721228)*x_ref[0]+(0.551002482436)*x_ref[1]+(-0.295853851234)*x_ref[2]
        ref[2,3]=(-0.796431361745)*x_ref[0]+(0.0581884766856)*x_ref[1]+(-1.13668563089)*x_ref[2]
        ref[2,4]=(-0.733185798017)*x_ref[0]+(0.178900103735)*x_ref[1]+(-0.845671526781)*x_ref[2]
        ref[3,0]=(-1.47064927107)*x_ref[0]+(1.33393324965)*x_ref[1]+(-0.767393460937)*x_ref[2]
        ref[3,1]=(1.36784513226)*x_ref[0]+(0.100008910431)*x_ref[1]+(-1.09210256713)*x_ref[2]
        ref[3,2]=(-0.873076669977)*x_ref[0]+(0.367213631778)*x_ref[1]+(0.0698296920039)*x_ref[2]
        ref[3,3]=(-1.76609214158)*x_ref[0]+(-1.26186286566)*x_ref[1]+(0.377817035113)*x_ref[2]
        ref[3,4]=(-1.28261709396)*x_ref[0]+(0.240480767533)*x_ref[1]+(0.307813279357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.423505169404)*x[0]+(-1.10390334264)*x[1]
        arg[0,0,1]=(-0.181820089254)*x[0]+(-0.374602268791)*x[1]
        arg[0,1,0]=(-1.36906917075)*x[0]+(0.26554385908)*x[1]
        arg[0,1,1]=(-0.280661572374)*x[0]+(-0.239232524287)*x[1]
        arg[1,0,0]=(-1.25670251934)*x[0]+(-0.429353977735)*x[1]
        arg[1,0,1]=(-0.747928970735)*x[0]+(1.17874317611)*x[1]
        arg[1,1,0]=(-0.548814611802)*x[0]+(-0.0249315764285)*x[1]
        arg[1,1,1]=(-0.481123117522)*x[0]+(1.23402203879)*x[1]
        arg[2,0,0]=(-0.0639512689884)*x[0]+(-0.0278213371376)*x[1]
        arg[2,0,1]=(-1.17292797621)*x[0]+(0.42140600304)*x[1]
        arg[2,1,0]=(-0.358052601083)*x[0]+(-1.66807880791)*x[1]
        arg[2,1,1]=(-0.291433251349)*x[0]+(0.111473162108)*x[1]
        arg[3,0,0]=(0.742643625989)*x[0]+(-1.29214106228)*x[1]
        arg[3,0,1]=(0.319887321826)*x[0]+(-0.40213369238)*x[1]
        arg[3,1,0]=(-0.449412064097)*x[0]+(-1.10654262009)*x[1]
        arg[3,1,1]=(1.19722227632)*x[0]+(1.3455799026)*x[1]
        arg[4,0,0]=(-1.77453188487)*x[0]+(0.634838793872)*x[1]
        arg[4,0,1]=(-0.215461066907)*x[0]+(1.11635233844)*x[1]
        arg[4,1,0]=(0.835816652592)*x[0]+(0.668378554214)*x[1]
        arg[4,1,1]=(-1.22133225406)*x[0]+(-0.148780252398)*x[1]
        arg[5,0,0]=(0.320245409639)*x[0]+(-1.08828513776)*x[1]
        arg[5,0,1]=(-0.1865835417)*x[0]+(-0.353282061874)*x[1]
        arg[5,1,0]=(1.38858040685)*x[0]+(0.166398261098)*x[1]
        arg[5,1,1]=(0.947718875755)*x[0]+(-1.06171934302)*x[1]
        ref[0,0,0]=(-0.423505169404)*x_ref[0]+(-1.10390334264)*x_ref[1]
        ref[0,0,1]=(-0.181820089254)*x_ref[0]+(-0.374602268791)*x_ref[1]
        ref[0,1,0]=(-1.36906917075)*x_ref[0]+(0.26554385908)*x_ref[1]
        ref[0,1,1]=(-0.280661572374)*x_ref[0]+(-0.239232524287)*x_ref[1]
        ref[1,0,0]=(-1.25670251934)*x_ref[0]+(-0.429353977735)*x_ref[1]
        ref[1,0,1]=(-0.747928970735)*x_ref[0]+(1.17874317611)*x_ref[1]
        ref[1,1,0]=(-0.548814611802)*x_ref[0]+(-0.0249315764285)*x_ref[1]
        ref[1,1,1]=(-0.481123117522)*x_ref[0]+(1.23402203879)*x_ref[1]
        ref[2,0,0]=(-0.0639512689884)*x_ref[0]+(-0.0278213371376)*x_ref[1]
        ref[2,0,1]=(-1.17292797621)*x_ref[0]+(0.42140600304)*x_ref[1]
        ref[2,1,0]=(-0.358052601083)*x_ref[0]+(-1.66807880791)*x_ref[1]
        ref[2,1,1]=(-0.291433251349)*x_ref[0]+(0.111473162108)*x_ref[1]
        ref[3,0,0]=(0.742643625989)*x_ref[0]+(-1.29214106228)*x_ref[1]
        ref[3,0,1]=(0.319887321826)*x_ref[0]+(-0.40213369238)*x_ref[1]
        ref[3,1,0]=(-0.449412064097)*x_ref[0]+(-1.10654262009)*x_ref[1]
        ref[3,1,1]=(1.19722227632)*x_ref[0]+(1.3455799026)*x_ref[1]
        ref[4,0,0]=(-1.77453188487)*x_ref[0]+(0.634838793872)*x_ref[1]
        ref[4,0,1]=(-0.215461066907)*x_ref[0]+(1.11635233844)*x_ref[1]
        ref[4,1,0]=(0.835816652592)*x_ref[0]+(0.668378554214)*x_ref[1]
        ref[4,1,1]=(-1.22133225406)*x_ref[0]+(-0.148780252398)*x_ref[1]
        ref[5,0,0]=(0.320245409639)*x_ref[0]+(-1.08828513776)*x_ref[1]
        ref[5,0,1]=(-0.1865835417)*x_ref[0]+(-0.353282061874)*x_ref[1]
        ref[5,1,0]=(1.38858040685)*x_ref[0]+(0.166398261098)*x_ref[1]
        ref[5,1,1]=(0.947718875755)*x_ref[0]+(-1.06171934302)*x_ref[1]
      else:
        arg[0,0,0]=(-0.892096322478)*x[0]+(0.720363183451)*x[1]+(0.600609272191)*x[2]
        arg[0,0,1]=(-0.418024461935)*x[0]+(-0.819275292035)*x[1]+(0.134962557692)*x[2]
        arg[0,1,0]=(-0.449783838361)*x[0]+(0.943677173533)*x[1]+(0.209424190991)*x[2]
        arg[0,1,1]=(-0.424386775004)*x[0]+(-0.213024388094)*x[1]+(-1.05927943384)*x[2]
        arg[1,0,0]=(-0.699504096539)*x[0]+(-0.153570308377)*x[1]+(0.576902392658)*x[2]
        arg[1,0,1]=(-0.426914055661)*x[0]+(-0.330405267059)*x[1]+(1.07656632155)*x[2]
        arg[1,1,0]=(0.775238005352)*x[0]+(1.11012072575)*x[1]+(0.764627098603)*x[2]
        arg[1,1,1]=(0.0130424708408)*x[0]+(-0.671778102107)*x[1]+(0.780684368983)*x[2]
        arg[2,0,0]=(-0.0512224258498)*x[0]+(-1.50515803984)*x[1]+(0.342824143946)*x[2]
        arg[2,0,1]=(0.117541624176)*x[0]+(0.372415524552)*x[1]+(0.255218095016)*x[2]
        arg[2,1,0]=(-0.574508895703)*x[0]+(0.345217292359)*x[1]+(0.209211956821)*x[2]
        arg[2,1,1]=(0.93643587147)*x[0]+(-0.096209009902)*x[1]+(0.205150859527)*x[2]
        arg[3,0,0]=(1.7541929884)*x[0]+(0.169061478447)*x[1]+(0.534746695515)*x[2]
        arg[3,0,1]=(0.442329266782)*x[0]+(0.334971710414)*x[1]+(1.66892557698)*x[2]
        arg[3,1,0]=(0.316436716758)*x[0]+(-1.25442030987)*x[1]+(0.372183105374)*x[2]
        arg[3,1,1]=(-0.254157820265)*x[0]+(0.287381300023)*x[1]+(0.0574117423352)*x[2]
        arg[4,0,0]=(-1.08380257521)*x[0]+(-1.31461890032)*x[1]+(-0.503802775099)*x[2]
        arg[4,0,1]=(-0.558293486798)*x[0]+(0.702317635925)*x[1]+(0.463814829137)*x[2]
        arg[4,1,0]=(-0.677160971974)*x[0]+(-0.14921275312)*x[1]+(-1.6501459435)*x[2]
        arg[4,1,1]=(-0.176672532474)*x[0]+(0.945017605811)*x[1]+(-0.790342939788)*x[2]
        arg[5,0,0]=(-1.45668070675)*x[0]+(-0.404957129204)*x[1]+(1.37383065954)*x[2]
        arg[5,0,1]=(0.53829326756)*x[0]+(0.543333031479)*x[1]+(-0.854057296362)*x[2]
        arg[5,1,0]=(0.805864283921)*x[0]+(-0.00186634953513)*x[1]+(-1.97111362539)*x[2]
        arg[5,1,1]=(-0.387746238967)*x[0]+(-0.739158353882)*x[1]+(-0.134082808118)*x[2]
        ref[0,0,0]=(-0.892096322478)*x_ref[0]+(0.720363183451)*x_ref[1]+(0.600609272191)*x_ref[2]
        ref[0,0,1]=(-0.418024461935)*x_ref[0]+(-0.819275292035)*x_ref[1]+(0.134962557692)*x_ref[2]
        ref[0,1,0]=(-0.449783838361)*x_ref[0]+(0.943677173533)*x_ref[1]+(0.209424190991)*x_ref[2]
        ref[0,1,1]=(-0.424386775004)*x_ref[0]+(-0.213024388094)*x_ref[1]+(-1.05927943384)*x_ref[2]
        ref[1,0,0]=(-0.699504096539)*x_ref[0]+(-0.153570308377)*x_ref[1]+(0.576902392658)*x_ref[2]
        ref[1,0,1]=(-0.426914055661)*x_ref[0]+(-0.330405267059)*x_ref[1]+(1.07656632155)*x_ref[2]
        ref[1,1,0]=(0.775238005352)*x_ref[0]+(1.11012072575)*x_ref[1]+(0.764627098603)*x_ref[2]
        ref[1,1,1]=(0.0130424708408)*x_ref[0]+(-0.671778102107)*x_ref[1]+(0.780684368983)*x_ref[2]
        ref[2,0,0]=(-0.0512224258498)*x_ref[0]+(-1.50515803984)*x_ref[1]+(0.342824143946)*x_ref[2]
        ref[2,0,1]=(0.117541624176)*x_ref[0]+(0.372415524552)*x_ref[1]+(0.255218095016)*x_ref[2]
        ref[2,1,0]=(-0.574508895703)*x_ref[0]+(0.345217292359)*x_ref[1]+(0.209211956821)*x_ref[2]
        ref[2,1,1]=(0.93643587147)*x_ref[0]+(-0.096209009902)*x_ref[1]+(0.205150859527)*x_ref[2]
        ref[3,0,0]=(1.7541929884)*x_ref[0]+(0.169061478447)*x_ref[1]+(0.534746695515)*x_ref[2]
        ref[3,0,1]=(0.442329266782)*x_ref[0]+(0.334971710414)*x_ref[1]+(1.66892557698)*x_ref[2]
        ref[3,1,0]=(0.316436716758)*x_ref[0]+(-1.25442030987)*x_ref[1]+(0.372183105374)*x_ref[2]
        ref[3,1,1]=(-0.254157820265)*x_ref[0]+(0.287381300023)*x_ref[1]+(0.0574117423352)*x_ref[2]
        ref[4,0,0]=(-1.08380257521)*x_ref[0]+(-1.31461890032)*x_ref[1]+(-0.503802775099)*x_ref[2]
        ref[4,0,1]=(-0.558293486798)*x_ref[0]+(0.702317635925)*x_ref[1]+(0.463814829137)*x_ref[2]
        ref[4,1,0]=(-0.677160971974)*x_ref[0]+(-0.14921275312)*x_ref[1]+(-1.6501459435)*x_ref[2]
        ref[4,1,1]=(-0.176672532474)*x_ref[0]+(0.945017605811)*x_ref[1]+(-0.790342939788)*x_ref[2]
        ref[5,0,0]=(-1.45668070675)*x_ref[0]+(-0.404957129204)*x_ref[1]+(1.37383065954)*x_ref[2]
        ref[5,0,1]=(0.53829326756)*x_ref[0]+(0.543333031479)*x_ref[1]+(-0.854057296362)*x_ref[2]
        ref[5,1,0]=(0.805864283921)*x_ref[0]+(-0.00186634953513)*x_ref[1]+(-1.97111362539)*x_ref[2]
        ref[5,1,1]=(-0.387746238967)*x_ref[0]+(-0.739158353882)*x_ref[1]+(-0.134082808118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnContactOne
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.754311658086)*x[0]+(0.235157034342)*x[1]
        arg[0,0,0,1]=(-1.33786315024)*x[0]+(0.243349279742)*x[1]
        arg[0,0,1,0]=(1.48657750088)*x[0]+(1.09853108305)*x[1]
        arg[0,0,1,1]=(1.4501362266)*x[0]+(0.824811567155)*x[1]
        arg[0,0,2,0]=(1.12567566082)*x[0]+(0.663525681174)*x[1]
        arg[0,0,2,1]=(1.48991565019)*x[0]+(-0.680795365011)*x[1]
        arg[0,1,0,0]=(0.268571449528)*x[0]+(0.818595834672)*x[1]
        arg[0,1,0,1]=(-1.55490945555)*x[0]+(0.783595197888)*x[1]
        arg[0,1,1,0]=(0.913201382585)*x[0]+(0.855491379376)*x[1]
        arg[0,1,1,1]=(-0.179192075116)*x[0]+(-0.690415079523)*x[1]
        arg[0,1,2,0]=(-0.0393574502912)*x[0]+(-0.690157650634)*x[1]
        arg[0,1,2,1]=(0.784216863217)*x[0]+(0.498735486159)*x[1]
        arg[0,2,0,0]=(0.137939554165)*x[0]+(-1.33961519113)*x[1]
        arg[0,2,0,1]=(-1.03897311736)*x[0]+(-1.37797405312)*x[1]
        arg[0,2,1,0]=(0.199812003756)*x[0]+(-0.0388499674037)*x[1]
        arg[0,2,1,1]=(1.15530534202)*x[0]+(-1.28280869078)*x[1]
        arg[0,2,2,0]=(0.836447791155)*x[0]+(1.2627066026)*x[1]
        arg[0,2,2,1]=(-0.584608492688)*x[0]+(-1.55305974305)*x[1]
        arg[0,3,0,0]=(-0.661952381799)*x[0]+(-1.27062403778)*x[1]
        arg[0,3,0,1]=(-1.47789532259)*x[0]+(-0.279583524936)*x[1]
        arg[0,3,1,0]=(1.40729771357)*x[0]+(0.0971284144528)*x[1]
        arg[0,3,1,1]=(-0.459167955911)*x[0]+(-1.29033281524)*x[1]
        arg[0,3,2,0]=(-1.63189436824)*x[0]+(0.75729808422)*x[1]
        arg[0,3,2,1]=(0.522479828718)*x[0]+(-0.0231299515383)*x[1]
        arg[0,4,0,0]=(0.241891446615)*x[0]+(0.803327107651)*x[1]
        arg[0,4,0,1]=(0.273613502836)*x[0]+(-0.0115180464171)*x[1]
        arg[0,4,1,0]=(1.96219818696)*x[0]+(-0.349137094743)*x[1]
        arg[0,4,1,1]=(0.365690821719)*x[0]+(-1.6943347181)*x[1]
        arg[0,4,2,0]=(-0.283970779872)*x[0]+(-0.258537179804)*x[1]
        arg[0,4,2,1]=(-0.973333430598)*x[0]+(1.48706276212)*x[1]
        arg[1,0,0,0]=(1.18757466971)*x[0]+(0.838804656357)*x[1]
        arg[1,0,0,1]=(0.840859904702)*x[0]+(1.09955195682)*x[1]
        arg[1,0,1,0]=(1.46190513227)*x[0]+(-0.596959830422)*x[1]
        arg[1,0,1,1]=(0.718689296745)*x[0]+(-0.628074620222)*x[1]
        arg[1,0,2,0]=(0.440761293041)*x[0]+(0.115800854194)*x[1]
        arg[1,0,2,1]=(1.34891146904)*x[0]+(0.813690260652)*x[1]
        arg[1,1,0,0]=(0.183071638098)*x[0]+(-1.37996210755)*x[1]
        arg[1,1,0,1]=(0.258142115988)*x[0]+(-0.757709381529)*x[1]
        arg[1,1,1,0]=(1.72392829309)*x[0]+(0.444995102781)*x[1]
        arg[1,1,1,1]=(-0.380898209988)*x[0]+(-0.0950246860347)*x[1]
        arg[1,1,2,0]=(0.203007226016)*x[0]+(1.09786001832)*x[1]
        arg[1,1,2,1]=(-0.502500865341)*x[0]+(-0.629980527995)*x[1]
        arg[1,2,0,0]=(0.482549518436)*x[0]+(0.565840824146)*x[1]
        arg[1,2,0,1]=(0.355818565068)*x[0]+(-0.78273295574)*x[1]
        arg[1,2,1,0]=(1.79204033071)*x[0]+(-0.698040851585)*x[1]
        arg[1,2,1,1]=(-0.534827741044)*x[0]+(0.00943131627498)*x[1]
        arg[1,2,2,0]=(-1.00663524444)*x[0]+(0.0207854553456)*x[1]
        arg[1,2,2,1]=(-0.338048040159)*x[0]+(-0.380607406217)*x[1]
        arg[1,3,0,0]=(0.586310549044)*x[0]+(0.276326798805)*x[1]
        arg[1,3,0,1]=(-0.260482580702)*x[0]+(-0.349279395321)*x[1]
        arg[1,3,1,0]=(-0.196533829895)*x[0]+(0.614805272738)*x[1]
        arg[1,3,1,1]=(-0.363176483895)*x[0]+(-0.925817017731)*x[1]
        arg[1,3,2,0]=(-1.24182075421)*x[0]+(0.215966446013)*x[1]
        arg[1,3,2,1]=(1.07546652642)*x[0]+(0.383803991974)*x[1]
        arg[1,4,0,0]=(0.199312303525)*x[0]+(-0.314520883859)*x[1]
        arg[1,4,0,1]=(-1.56890826152)*x[0]+(-0.133086662241)*x[1]
        arg[1,4,1,0]=(1.1307241936)*x[0]+(-1.32905744984)*x[1]
        arg[1,4,1,1]=(-0.385991455298)*x[0]+(0.292913024507)*x[1]
        arg[1,4,2,0]=(1.14561678056)*x[0]+(-0.662696162276)*x[1]
        arg[1,4,2,1]=(-1.49533948954)*x[0]+(1.43396138744)*x[1]
        arg[2,0,0,0]=(-0.63160396493)*x[0]+(0.813120998776)*x[1]
        arg[2,0,0,1]=(0.65598998434)*x[0]+(-1.02445537553)*x[1]
        arg[2,0,1,0]=(-1.13473104075)*x[0]+(0.057331183449)*x[1]
        arg[2,0,1,1]=(-0.154892482107)*x[0]+(0.481969957579)*x[1]
        arg[2,0,2,0]=(0.521916671496)*x[0]+(0.546676943059)*x[1]
        arg[2,0,2,1]=(-1.68919379028)*x[0]+(1.08292038981)*x[1]
        arg[2,1,0,0]=(-0.643846595888)*x[0]+(0.528306097755)*x[1]
        arg[2,1,0,1]=(1.3745090497)*x[0]+(-0.0801648509184)*x[1]
        arg[2,1,1,0]=(0.129096817054)*x[0]+(0.90750370855)*x[1]
        arg[2,1,1,1]=(0.539309513305)*x[0]+(1.34949470961)*x[1]
        arg[2,1,2,0]=(1.4189703896)*x[0]+(-1.75011201893)*x[1]
        arg[2,1,2,1]=(-1.09325548386)*x[0]+(0.148279301716)*x[1]
        arg[2,2,0,0]=(-1.25785133743)*x[0]+(0.275071306594)*x[1]
        arg[2,2,0,1]=(-0.550152068761)*x[0]+(-1.20422484275)*x[1]
        arg[2,2,1,0]=(-0.102062863668)*x[0]+(-1.34431887541)*x[1]
        arg[2,2,1,1]=(0.103354724169)*x[0]+(-0.941952180446)*x[1]
        arg[2,2,2,0]=(1.2090339604)*x[0]+(-0.330357976796)*x[1]
        arg[2,2,2,1]=(0.203162449188)*x[0]+(-1.03640702014)*x[1]
        arg[2,3,0,0]=(-0.324576353514)*x[0]+(0.582614539267)*x[1]
        arg[2,3,0,1]=(-0.213823136215)*x[0]+(-0.473199105214)*x[1]
        arg[2,3,1,0]=(0.395144702599)*x[0]+(0.0358194467152)*x[1]
        arg[2,3,1,1]=(1.32860563396)*x[0]+(-0.577308310732)*x[1]
        arg[2,3,2,0]=(0.994205029432)*x[0]+(1.13433517636)*x[1]
        arg[2,3,2,1]=(-0.488499286742)*x[0]+(0.224253820738)*x[1]
        arg[2,4,0,0]=(1.48035663008)*x[0]+(-1.09697986914)*x[1]
        arg[2,4,0,1]=(-1.09991915151)*x[0]+(-0.557951199985)*x[1]
        arg[2,4,1,0]=(-1.23862699804)*x[0]+(-0.230602187713)*x[1]
        arg[2,4,1,1]=(-0.922609370448)*x[0]+(1.76673953413)*x[1]
        arg[2,4,2,0]=(0.731769419709)*x[0]+(-1.1094516608)*x[1]
        arg[2,4,2,1]=(-0.145672504805)*x[0]+(0.168194421355)*x[1]
        arg[3,0,0,0]=(1.02559707602)*x[0]+(0.260431368523)*x[1]
        arg[3,0,0,1]=(0.111899568824)*x[0]+(-0.90725475423)*x[1]
        arg[3,0,1,0]=(-0.226547745683)*x[0]+(-0.232830393372)*x[1]
        arg[3,0,1,1]=(0.347004318471)*x[0]+(-0.196130970001)*x[1]
        arg[3,0,2,0]=(-1.72155871442)*x[0]+(-1.28372407452)*x[1]
        arg[3,0,2,1]=(1.43817914693)*x[0]+(-0.354093253684)*x[1]
        arg[3,1,0,0]=(-0.293939245102)*x[0]+(0.0255562139103)*x[1]
        arg[3,1,0,1]=(0.139306213018)*x[0]+(0.595062226642)*x[1]
        arg[3,1,1,0]=(-0.670339920672)*x[0]+(0.387520583192)*x[1]
        arg[3,1,1,1]=(-0.0323325470954)*x[0]+(-0.681162999279)*x[1]
        arg[3,1,2,0]=(0.235275048739)*x[0]+(0.144725473773)*x[1]
        arg[3,1,2,1]=(0.132211795214)*x[0]+(0.262780216845)*x[1]
        arg[3,2,0,0]=(0.590514326163)*x[0]+(-0.84273677918)*x[1]
        arg[3,2,0,1]=(-0.925181556185)*x[0]+(-0.300244447001)*x[1]
        arg[3,2,1,0]=(-0.0509700795112)*x[0]+(-1.31919389616)*x[1]
        arg[3,2,1,1]=(1.50577291846)*x[0]+(-0.219284607769)*x[1]
        arg[3,2,2,0]=(-0.0460616673346)*x[0]+(-0.690312685466)*x[1]
        arg[3,2,2,1]=(-0.673765477936)*x[0]+(-0.0965860475641)*x[1]
        arg[3,3,0,0]=(-0.6577594174)*x[0]+(0.171886277016)*x[1]
        arg[3,3,0,1]=(-1.08810434718)*x[0]+(0.333153250183)*x[1]
        arg[3,3,1,0]=(0.150968371983)*x[0]+(0.290859317391)*x[1]
        arg[3,3,1,1]=(1.62096265131)*x[0]+(-1.10309602741)*x[1]
        arg[3,3,2,0]=(-0.932063639972)*x[0]+(-0.897131561126)*x[1]
        arg[3,3,2,1]=(-0.819110834345)*x[0]+(1.59159109562)*x[1]
        arg[3,4,0,0]=(0.272662931155)*x[0]+(0.69026285756)*x[1]
        arg[3,4,0,1]=(-0.600556370757)*x[0]+(-0.0853533546302)*x[1]
        arg[3,4,1,0]=(1.18544735281)*x[0]+(0.217751189034)*x[1]
        arg[3,4,1,1]=(-0.594609531358)*x[0]+(0.882158349788)*x[1]
        arg[3,4,2,0]=(1.18876765552)*x[0]+(-0.411481586979)*x[1]
        arg[3,4,2,1]=(-1.093067792)*x[0]+(-0.266744260768)*x[1]
        ref[0,0,0,0]=(-0.754311658086)*x_ref[0]+(0.235157034342)*x_ref[1]
        ref[0,0,0,1]=(-1.33786315024)*x_ref[0]+(0.243349279742)*x_ref[1]
        ref[0,0,1,0]=(1.48657750088)*x_ref[0]+(1.09853108305)*x_ref[1]
        ref[0,0,1,1]=(1.4501362266)*x_ref[0]+(0.824811567155)*x_ref[1]
        ref[0,0,2,0]=(1.12567566082)*x_ref[0]+(0.663525681174)*x_ref[1]
        ref[0,0,2,1]=(1.48991565019)*x_ref[0]+(-0.680795365011)*x_ref[1]
        ref[0,1,0,0]=(0.268571449528)*x_ref[0]+(0.818595834672)*x_ref[1]
        ref[0,1,0,1]=(-1.55490945555)*x_ref[0]+(0.783595197888)*x_ref[1]
        ref[0,1,1,0]=(0.913201382585)*x_ref[0]+(0.855491379376)*x_ref[1]
        ref[0,1,1,1]=(-0.179192075116)*x_ref[0]+(-0.690415079523)*x_ref[1]
        ref[0,1,2,0]=(-0.0393574502912)*x_ref[0]+(-0.690157650634)*x_ref[1]
        ref[0,1,2,1]=(0.784216863217)*x_ref[0]+(0.498735486159)*x_ref[1]
        ref[0,2,0,0]=(0.137939554165)*x_ref[0]+(-1.33961519113)*x_ref[1]
        ref[0,2,0,1]=(-1.03897311736)*x_ref[0]+(-1.37797405312)*x_ref[1]
        ref[0,2,1,0]=(0.199812003756)*x_ref[0]+(-0.0388499674037)*x_ref[1]
        ref[0,2,1,1]=(1.15530534202)*x_ref[0]+(-1.28280869078)*x_ref[1]
        ref[0,2,2,0]=(0.836447791155)*x_ref[0]+(1.2627066026)*x_ref[1]
        ref[0,2,2,1]=(-0.584608492688)*x_ref[0]+(-1.55305974305)*x_ref[1]
        ref[0,3,0,0]=(-0.661952381799)*x_ref[0]+(-1.27062403778)*x_ref[1]
        ref[0,3,0,1]=(-1.47789532259)*x_ref[0]+(-0.279583524936)*x_ref[1]
        ref[0,3,1,0]=(1.40729771357)*x_ref[0]+(0.0971284144528)*x_ref[1]
        ref[0,3,1,1]=(-0.459167955911)*x_ref[0]+(-1.29033281524)*x_ref[1]
        ref[0,3,2,0]=(-1.63189436824)*x_ref[0]+(0.75729808422)*x_ref[1]
        ref[0,3,2,1]=(0.522479828718)*x_ref[0]+(-0.0231299515383)*x_ref[1]
        ref[0,4,0,0]=(0.241891446615)*x_ref[0]+(0.803327107651)*x_ref[1]
        ref[0,4,0,1]=(0.273613502836)*x_ref[0]+(-0.0115180464171)*x_ref[1]
        ref[0,4,1,0]=(1.96219818696)*x_ref[0]+(-0.349137094743)*x_ref[1]
        ref[0,4,1,1]=(0.365690821719)*x_ref[0]+(-1.6943347181)*x_ref[1]
        ref[0,4,2,0]=(-0.283970779872)*x_ref[0]+(-0.258537179804)*x_ref[1]
        ref[0,4,2,1]=(-0.973333430598)*x_ref[0]+(1.48706276212)*x_ref[1]
        ref[1,0,0,0]=(1.18757466971)*x_ref[0]+(0.838804656357)*x_ref[1]
        ref[1,0,0,1]=(0.840859904702)*x_ref[0]+(1.09955195682)*x_ref[1]
        ref[1,0,1,0]=(1.46190513227)*x_ref[0]+(-0.596959830422)*x_ref[1]
        ref[1,0,1,1]=(0.718689296745)*x_ref[0]+(-0.628074620222)*x_ref[1]
        ref[1,0,2,0]=(0.440761293041)*x_ref[0]+(0.115800854194)*x_ref[1]
        ref[1,0,2,1]=(1.34891146904)*x_ref[0]+(0.813690260652)*x_ref[1]
        ref[1,1,0,0]=(0.183071638098)*x_ref[0]+(-1.37996210755)*x_ref[1]
        ref[1,1,0,1]=(0.258142115988)*x_ref[0]+(-0.757709381529)*x_ref[1]
        ref[1,1,1,0]=(1.72392829309)*x_ref[0]+(0.444995102781)*x_ref[1]
        ref[1,1,1,1]=(-0.380898209988)*x_ref[0]+(-0.0950246860347)*x_ref[1]
        ref[1,1,2,0]=(0.203007226016)*x_ref[0]+(1.09786001832)*x_ref[1]
        ref[1,1,2,1]=(-0.502500865341)*x_ref[0]+(-0.629980527995)*x_ref[1]
        ref[1,2,0,0]=(0.482549518436)*x_ref[0]+(0.565840824146)*x_ref[1]
        ref[1,2,0,1]=(0.355818565068)*x_ref[0]+(-0.78273295574)*x_ref[1]
        ref[1,2,1,0]=(1.79204033071)*x_ref[0]+(-0.698040851585)*x_ref[1]
        ref[1,2,1,1]=(-0.534827741044)*x_ref[0]+(0.00943131627498)*x_ref[1]
        ref[1,2,2,0]=(-1.00663524444)*x_ref[0]+(0.0207854553456)*x_ref[1]
        ref[1,2,2,1]=(-0.338048040159)*x_ref[0]+(-0.380607406217)*x_ref[1]
        ref[1,3,0,0]=(0.586310549044)*x_ref[0]+(0.276326798805)*x_ref[1]
        ref[1,3,0,1]=(-0.260482580702)*x_ref[0]+(-0.349279395321)*x_ref[1]
        ref[1,3,1,0]=(-0.196533829895)*x_ref[0]+(0.614805272738)*x_ref[1]
        ref[1,3,1,1]=(-0.363176483895)*x_ref[0]+(-0.925817017731)*x_ref[1]
        ref[1,3,2,0]=(-1.24182075421)*x_ref[0]+(0.215966446013)*x_ref[1]
        ref[1,3,2,1]=(1.07546652642)*x_ref[0]+(0.383803991974)*x_ref[1]
        ref[1,4,0,0]=(0.199312303525)*x_ref[0]+(-0.314520883859)*x_ref[1]
        ref[1,4,0,1]=(-1.56890826152)*x_ref[0]+(-0.133086662241)*x_ref[1]
        ref[1,4,1,0]=(1.1307241936)*x_ref[0]+(-1.32905744984)*x_ref[1]
        ref[1,4,1,1]=(-0.385991455298)*x_ref[0]+(0.292913024507)*x_ref[1]
        ref[1,4,2,0]=(1.14561678056)*x_ref[0]+(-0.662696162276)*x_ref[1]
        ref[1,4,2,1]=(-1.49533948954)*x_ref[0]+(1.43396138744)*x_ref[1]
        ref[2,0,0,0]=(-0.63160396493)*x_ref[0]+(0.813120998776)*x_ref[1]
        ref[2,0,0,1]=(0.65598998434)*x_ref[0]+(-1.02445537553)*x_ref[1]
        ref[2,0,1,0]=(-1.13473104075)*x_ref[0]+(0.057331183449)*x_ref[1]
        ref[2,0,1,1]=(-0.154892482107)*x_ref[0]+(0.481969957579)*x_ref[1]
        ref[2,0,2,0]=(0.521916671496)*x_ref[0]+(0.546676943059)*x_ref[1]
        ref[2,0,2,1]=(-1.68919379028)*x_ref[0]+(1.08292038981)*x_ref[1]
        ref[2,1,0,0]=(-0.643846595888)*x_ref[0]+(0.528306097755)*x_ref[1]
        ref[2,1,0,1]=(1.3745090497)*x_ref[0]+(-0.0801648509184)*x_ref[1]
        ref[2,1,1,0]=(0.129096817054)*x_ref[0]+(0.90750370855)*x_ref[1]
        ref[2,1,1,1]=(0.539309513305)*x_ref[0]+(1.34949470961)*x_ref[1]
        ref[2,1,2,0]=(1.4189703896)*x_ref[0]+(-1.75011201893)*x_ref[1]
        ref[2,1,2,1]=(-1.09325548386)*x_ref[0]+(0.148279301716)*x_ref[1]
        ref[2,2,0,0]=(-1.25785133743)*x_ref[0]+(0.275071306594)*x_ref[1]
        ref[2,2,0,1]=(-0.550152068761)*x_ref[0]+(-1.20422484275)*x_ref[1]
        ref[2,2,1,0]=(-0.102062863668)*x_ref[0]+(-1.34431887541)*x_ref[1]
        ref[2,2,1,1]=(0.103354724169)*x_ref[0]+(-0.941952180446)*x_ref[1]
        ref[2,2,2,0]=(1.2090339604)*x_ref[0]+(-0.330357976796)*x_ref[1]
        ref[2,2,2,1]=(0.203162449188)*x_ref[0]+(-1.03640702014)*x_ref[1]
        ref[2,3,0,0]=(-0.324576353514)*x_ref[0]+(0.582614539267)*x_ref[1]
        ref[2,3,0,1]=(-0.213823136215)*x_ref[0]+(-0.473199105214)*x_ref[1]
        ref[2,3,1,0]=(0.395144702599)*x_ref[0]+(0.0358194467152)*x_ref[1]
        ref[2,3,1,1]=(1.32860563396)*x_ref[0]+(-0.577308310732)*x_ref[1]
        ref[2,3,2,0]=(0.994205029432)*x_ref[0]+(1.13433517636)*x_ref[1]
        ref[2,3,2,1]=(-0.488499286742)*x_ref[0]+(0.224253820738)*x_ref[1]
        ref[2,4,0,0]=(1.48035663008)*x_ref[0]+(-1.09697986914)*x_ref[1]
        ref[2,4,0,1]=(-1.09991915151)*x_ref[0]+(-0.557951199985)*x_ref[1]
        ref[2,4,1,0]=(-1.23862699804)*x_ref[0]+(-0.230602187713)*x_ref[1]
        ref[2,4,1,1]=(-0.922609370448)*x_ref[0]+(1.76673953413)*x_ref[1]
        ref[2,4,2,0]=(0.731769419709)*x_ref[0]+(-1.1094516608)*x_ref[1]
        ref[2,4,2,1]=(-0.145672504805)*x_ref[0]+(0.168194421355)*x_ref[1]
        ref[3,0,0,0]=(1.02559707602)*x_ref[0]+(0.260431368523)*x_ref[1]
        ref[3,0,0,1]=(0.111899568824)*x_ref[0]+(-0.90725475423)*x_ref[1]
        ref[3,0,1,0]=(-0.226547745683)*x_ref[0]+(-0.232830393372)*x_ref[1]
        ref[3,0,1,1]=(0.347004318471)*x_ref[0]+(-0.196130970001)*x_ref[1]
        ref[3,0,2,0]=(-1.72155871442)*x_ref[0]+(-1.28372407452)*x_ref[1]
        ref[3,0,2,1]=(1.43817914693)*x_ref[0]+(-0.354093253684)*x_ref[1]
        ref[3,1,0,0]=(-0.293939245102)*x_ref[0]+(0.0255562139103)*x_ref[1]
        ref[3,1,0,1]=(0.139306213018)*x_ref[0]+(0.595062226642)*x_ref[1]
        ref[3,1,1,0]=(-0.670339920672)*x_ref[0]+(0.387520583192)*x_ref[1]
        ref[3,1,1,1]=(-0.0323325470954)*x_ref[0]+(-0.681162999279)*x_ref[1]
        ref[3,1,2,0]=(0.235275048739)*x_ref[0]+(0.144725473773)*x_ref[1]
        ref[3,1,2,1]=(0.132211795214)*x_ref[0]+(0.262780216845)*x_ref[1]
        ref[3,2,0,0]=(0.590514326163)*x_ref[0]+(-0.84273677918)*x_ref[1]
        ref[3,2,0,1]=(-0.925181556185)*x_ref[0]+(-0.300244447001)*x_ref[1]
        ref[3,2,1,0]=(-0.0509700795112)*x_ref[0]+(-1.31919389616)*x_ref[1]
        ref[3,2,1,1]=(1.50577291846)*x_ref[0]+(-0.219284607769)*x_ref[1]
        ref[3,2,2,0]=(-0.0460616673346)*x_ref[0]+(-0.690312685466)*x_ref[1]
        ref[3,2,2,1]=(-0.673765477936)*x_ref[0]+(-0.0965860475641)*x_ref[1]
        ref[3,3,0,0]=(-0.6577594174)*x_ref[0]+(0.171886277016)*x_ref[1]
        ref[3,3,0,1]=(-1.08810434718)*x_ref[0]+(0.333153250183)*x_ref[1]
        ref[3,3,1,0]=(0.150968371983)*x_ref[0]+(0.290859317391)*x_ref[1]
        ref[3,3,1,1]=(1.62096265131)*x_ref[0]+(-1.10309602741)*x_ref[1]
        ref[3,3,2,0]=(-0.932063639972)*x_ref[0]+(-0.897131561126)*x_ref[1]
        ref[3,3,2,1]=(-0.819110834345)*x_ref[0]+(1.59159109562)*x_ref[1]
        ref[3,4,0,0]=(0.272662931155)*x_ref[0]+(0.69026285756)*x_ref[1]
        ref[3,4,0,1]=(-0.600556370757)*x_ref[0]+(-0.0853533546302)*x_ref[1]
        ref[3,4,1,0]=(1.18544735281)*x_ref[0]+(0.217751189034)*x_ref[1]
        ref[3,4,1,1]=(-0.594609531358)*x_ref[0]+(0.882158349788)*x_ref[1]
        ref[3,4,2,0]=(1.18876765552)*x_ref[0]+(-0.411481586979)*x_ref[1]
        ref[3,4,2,1]=(-1.093067792)*x_ref[0]+(-0.266744260768)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.29218217256)*x[0]+(-0.382937892213)*x[1]+(0.14363896447)*x[2]
        arg[0,0,0,1]=(1.39856942421)*x[0]+(0.302082267835)*x[1]+(-0.40870736584)*x[2]
        arg[0,0,1,0]=(0.523846108043)*x[0]+(1.19961531204)*x[1]+(0.838881761349)*x[2]
        arg[0,0,1,1]=(0.634542282764)*x[0]+(-0.475514686466)*x[1]+(-1.68745428486)*x[2]
        arg[0,0,2,0]=(1.16379585731)*x[0]+(-0.928037595416)*x[1]+(0.998857178353)*x[2]
        arg[0,0,2,1]=(-0.804515607022)*x[0]+(-0.915469493439)*x[1]+(-0.143888000623)*x[2]
        arg[0,1,0,0]=(0.925843155486)*x[0]+(-0.168243418953)*x[1]+(-0.887839555187)*x[2]
        arg[0,1,0,1]=(-0.913103105329)*x[0]+(1.38011681732)*x[1]+(-0.992478793291)*x[2]
        arg[0,1,1,0]=(0.759685289903)*x[0]+(-0.579481939982)*x[1]+(0.577682929138)*x[2]
        arg[0,1,1,1]=(0.813155548228)*x[0]+(-0.178317843217)*x[1]+(-0.0460113194526)*x[2]
        arg[0,1,2,0]=(-0.195764860141)*x[0]+(1.44256979628)*x[1]+(0.283992441343)*x[2]
        arg[0,1,2,1]=(-0.921229459072)*x[0]+(-1.19295761547)*x[1]+(0.215736801944)*x[2]
        arg[0,2,0,0]=(-0.741726331752)*x[0]+(-0.225585937382)*x[1]+(1.12654556466)*x[2]
        arg[0,2,0,1]=(-0.709656162421)*x[0]+(0.298862010425)*x[1]+(-0.502443284827)*x[2]
        arg[0,2,1,0]=(-1.01166188669)*x[0]+(-1.43763408953)*x[1]+(-0.835160398488)*x[2]
        arg[0,2,1,1]=(1.80968424545)*x[0]+(-0.802666103495)*x[1]+(0.290283606115)*x[2]
        arg[0,2,2,0]=(0.346481981585)*x[0]+(-0.444176962215)*x[1]+(0.689127612227)*x[2]
        arg[0,2,2,1]=(1.06925494095)*x[0]+(0.452854988449)*x[1]+(-1.19974690364)*x[2]
        arg[0,3,0,0]=(-0.534896036328)*x[0]+(-0.312000770615)*x[1]+(-0.440033746693)*x[2]
        arg[0,3,0,1]=(0.580351431838)*x[0]+(-0.36623848451)*x[1]+(-0.507837490037)*x[2]
        arg[0,3,1,0]=(0.383180910522)*x[0]+(-0.0907152443563)*x[1]+(-0.200103743943)*x[2]
        arg[0,3,1,1]=(0.0160194687513)*x[0]+(0.862509612939)*x[1]+(-1.14643594474)*x[2]
        arg[0,3,2,0]=(-0.224390850199)*x[0]+(-0.176927715447)*x[1]+(0.213480525831)*x[2]
        arg[0,3,2,1]=(-0.0508511448148)*x[0]+(-1.28943467389)*x[1]+(0.902285302487)*x[2]
        arg[0,4,0,0]=(-0.633106625232)*x[0]+(-0.430651917485)*x[1]+(0.583149841952)*x[2]
        arg[0,4,0,1]=(0.579035528956)*x[0]+(0.159439682519)*x[1]+(-0.0926277063021)*x[2]
        arg[0,4,1,0]=(0.00484756366973)*x[0]+(0.0324822270755)*x[1]+(1.20052413378)*x[2]
        arg[0,4,1,1]=(-0.337383994688)*x[0]+(0.902076367981)*x[1]+(-0.1956377395)*x[2]
        arg[0,4,2,0]=(0.928434343456)*x[0]+(0.0376569884355)*x[1]+(-0.155274118568)*x[2]
        arg[0,4,2,1]=(-0.541990213238)*x[0]+(-0.0165076697829)*x[1]+(-0.210319224859)*x[2]
        arg[1,0,0,0]=(-1.67762025181)*x[0]+(1.61434755579)*x[1]+(0.357272049999)*x[2]
        arg[1,0,0,1]=(-1.62532544549)*x[0]+(-0.271426231027)*x[1]+(0.338680991778)*x[2]
        arg[1,0,1,0]=(1.73523408033)*x[0]+(1.11987092368)*x[1]+(0.126769364273)*x[2]
        arg[1,0,1,1]=(-1.49220786558)*x[0]+(1.51915566092)*x[1]+(1.24788192157)*x[2]
        arg[1,0,2,0]=(-0.687808766765)*x[0]+(-1.02318754737)*x[1]+(0.590653972416)*x[2]
        arg[1,0,2,1]=(0.756067354802)*x[0]+(-0.66964706965)*x[1]+(-1.18986499158)*x[2]
        arg[1,1,0,0]=(0.371345732177)*x[0]+(0.212739617115)*x[1]+(-1.13439047405)*x[2]
        arg[1,1,0,1]=(0.525761804111)*x[0]+(0.144130399691)*x[1]+(0.715517587677)*x[2]
        arg[1,1,1,0]=(-0.376160293215)*x[0]+(-0.469964935263)*x[1]+(0.413518194707)*x[2]
        arg[1,1,1,1]=(-0.787748398144)*x[0]+(-0.810090402237)*x[1]+(-0.50127109256)*x[2]
        arg[1,1,2,0]=(0.654555083692)*x[0]+(-0.142134700515)*x[1]+(-1.55891930521)*x[2]
        arg[1,1,2,1]=(-0.423058885456)*x[0]+(-0.65930158641)*x[1]+(0.94440057744)*x[2]
        arg[1,2,0,0]=(-0.924832007053)*x[0]+(1.0154890382)*x[1]+(0.11938079748)*x[2]
        arg[1,2,0,1]=(0.134642686931)*x[0]+(-1.30908067998)*x[1]+(0.367206139074)*x[2]
        arg[1,2,1,0]=(1.55524055201)*x[0]+(0.727208660637)*x[1]+(0.809810917377)*x[2]
        arg[1,2,1,1]=(0.94813391462)*x[0]+(-1.49402524308)*x[1]+(0.367824542434)*x[2]
        arg[1,2,2,0]=(-0.608295798731)*x[0]+(-0.142302252814)*x[1]+(-0.986529251308)*x[2]
        arg[1,2,2,1]=(0.0259394619607)*x[0]+(-1.60728202467)*x[1]+(0.312534656619)*x[2]
        arg[1,3,0,0]=(-1.34063373729)*x[0]+(-0.268499839329)*x[1]+(-0.335125886299)*x[2]
        arg[1,3,0,1]=(1.53280698072)*x[0]+(1.24369684711)*x[1]+(0.766369156132)*x[2]
        arg[1,3,1,0]=(0.902067176876)*x[0]+(0.791102994814)*x[1]+(-0.89266430493)*x[2]
        arg[1,3,1,1]=(0.704997629232)*x[0]+(-0.551611360332)*x[1]+(1.39134537491)*x[2]
        arg[1,3,2,0]=(-0.151690675211)*x[0]+(-1.01762975893)*x[1]+(-0.627493214284)*x[2]
        arg[1,3,2,1]=(1.00936645922)*x[0]+(0.0322458997248)*x[1]+(1.45039938194)*x[2]
        arg[1,4,0,0]=(-0.0916390388904)*x[0]+(-0.783111746469)*x[1]+(0.389725342225)*x[2]
        arg[1,4,0,1]=(1.18859328243)*x[0]+(0.871689661558)*x[1]+(-0.775464713448)*x[2]
        arg[1,4,1,0]=(0.913118342488)*x[0]+(1.53285648884)*x[1]+(0.0947365101243)*x[2]
        arg[1,4,1,1]=(-1.00735841224)*x[0]+(-1.12544193607)*x[1]+(0.267347229372)*x[2]
        arg[1,4,2,0]=(0.285943438284)*x[0]+(-1.0889389244)*x[1]+(-1.03054842327)*x[2]
        arg[1,4,2,1]=(-1.5768940139)*x[0]+(0.91237754587)*x[1]+(1.30220394199)*x[2]
        arg[2,0,0,0]=(-0.0634935883277)*x[0]+(0.157304131712)*x[1]+(-1.2995042074)*x[2]
        arg[2,0,0,1]=(-0.802705266871)*x[0]+(0.238233162774)*x[1]+(1.29463414013)*x[2]
        arg[2,0,1,0]=(-0.925468462671)*x[0]+(0.870960795966)*x[1]+(0.156522765713)*x[2]
        arg[2,0,1,1]=(-0.546666833309)*x[0]+(-1.14649607995)*x[1]+(-1.53906505048)*x[2]
        arg[2,0,2,0]=(0.464952910857)*x[0]+(0.408605786027)*x[1]+(-0.886974881114)*x[2]
        arg[2,0,2,1]=(0.491723247948)*x[0]+(0.194616420077)*x[1]+(0.409049782749)*x[2]
        arg[2,1,0,0]=(-1.19836883114)*x[0]+(1.0622722337)*x[1]+(-0.109193867727)*x[2]
        arg[2,1,0,1]=(1.01973600383)*x[0]+(-0.154482220539)*x[1]+(0.378696371956)*x[2]
        arg[2,1,1,0]=(-0.0812015296246)*x[0]+(-1.10252794568)*x[1]+(-0.00585426581857)*x[2]
        arg[2,1,1,1]=(0.536042665272)*x[0]+(0.685507117255)*x[1]+(-0.179652683139)*x[2]
        arg[2,1,2,0]=(-0.36750467811)*x[0]+(-1.23674550869)*x[1]+(-0.762905702801)*x[2]
        arg[2,1,2,1]=(0.329665965651)*x[0]+(0.101195810737)*x[1]+(-0.610015052375)*x[2]
        arg[2,2,0,0]=(-0.81644635185)*x[0]+(-0.604578476306)*x[1]+(0.0440925706991)*x[2]
        arg[2,2,0,1]=(-0.958241437971)*x[0]+(0.38563498206)*x[1]+(0.913406762318)*x[2]
        arg[2,2,1,0]=(1.22027492224)*x[0]+(-1.00726209939)*x[1]+(-1.05083068303)*x[2]
        arg[2,2,1,1]=(-1.47561192645)*x[0]+(-0.903549355608)*x[1]+(0.236771324671)*x[2]
        arg[2,2,2,0]=(0.284995864449)*x[0]+(-0.175760829921)*x[1]+(-0.720607220001)*x[2]
        arg[2,2,2,1]=(-0.535894515963)*x[0]+(-0.821302957162)*x[1]+(-1.75141540241)*x[2]
        arg[2,3,0,0]=(-0.333082106674)*x[0]+(0.579471826412)*x[1]+(0.561184138108)*x[2]
        arg[2,3,0,1]=(1.34803519557)*x[0]+(-0.441299088849)*x[1]+(-1.04945951616)*x[2]
        arg[2,3,1,0]=(0.299011096602)*x[0]+(1.32726458235)*x[1]+(-0.196298687359)*x[2]
        arg[2,3,1,1]=(1.48698834204)*x[0]+(0.527221496971)*x[1]+(0.0819813519422)*x[2]
        arg[2,3,2,0]=(0.681680955153)*x[0]+(0.503653972357)*x[1]+(0.434669347891)*x[2]
        arg[2,3,2,1]=(-0.152586789311)*x[0]+(-0.765242635579)*x[1]+(0.712663994387)*x[2]
        arg[2,4,0,0]=(0.630152300463)*x[0]+(1.12366733588)*x[1]+(-0.662966700638)*x[2]
        arg[2,4,0,1]=(1.02796709755)*x[0]+(-0.24093881584)*x[1]+(1.79445417019)*x[2]
        arg[2,4,1,0]=(-0.954575760758)*x[0]+(0.16331782608)*x[1]+(1.38309921277)*x[2]
        arg[2,4,1,1]=(0.536092514281)*x[0]+(-0.276364952673)*x[1]+(0.779619701528)*x[2]
        arg[2,4,2,0]=(0.267270444141)*x[0]+(-0.810746424331)*x[1]+(-0.733877853264)*x[2]
        arg[2,4,2,1]=(-0.0641688847923)*x[0]+(-0.386279280667)*x[1]+(0.0637571595255)*x[2]
        arg[3,0,0,0]=(0.469903180687)*x[0]+(-0.793812397342)*x[1]+(1.60841921779)*x[2]
        arg[3,0,0,1]=(0.720057600102)*x[0]+(0.502582253261)*x[1]+(0.460147791003)*x[2]
        arg[3,0,1,0]=(0.0327862564137)*x[0]+(-1.142962965)*x[1]+(0.114799131894)*x[2]
        arg[3,0,1,1]=(-0.997412656978)*x[0]+(-0.678175468576)*x[1]+(-0.152663147682)*x[2]
        arg[3,0,2,0]=(0.810424599078)*x[0]+(0.75522727786)*x[1]+(-0.205880941501)*x[2]
        arg[3,0,2,1]=(0.353601949476)*x[0]+(0.0963200813379)*x[1]+(0.0578829287146)*x[2]
        arg[3,1,0,0]=(0.346590730062)*x[0]+(1.07953790092)*x[1]+(0.647586851217)*x[2]
        arg[3,1,0,1]=(0.889879195334)*x[0]+(-0.352788186249)*x[1]+(0.434900135803)*x[2]
        arg[3,1,1,0]=(-0.721742734959)*x[0]+(0.166066672134)*x[1]+(-0.489266924809)*x[2]
        arg[3,1,1,1]=(-0.810487755181)*x[0]+(0.0354916979253)*x[1]+(0.0735520654395)*x[2]
        arg[3,1,2,0]=(1.02414338939)*x[0]+(-0.851076284556)*x[1]+(-0.232174494405)*x[2]
        arg[3,1,2,1]=(-0.327246566026)*x[0]+(-0.0499396023605)*x[1]+(0.611585402379)*x[2]
        arg[3,2,0,0]=(-0.365757628557)*x[0]+(0.868697342115)*x[1]+(-1.102545972)*x[2]
        arg[3,2,0,1]=(0.628592129227)*x[0]+(1.25545515311)*x[1]+(-1.0577476264)*x[2]
        arg[3,2,1,0]=(0.160995978044)*x[0]+(-0.947389499556)*x[1]+(-0.870476160656)*x[2]
        arg[3,2,1,1]=(0.910787139938)*x[0]+(-0.0681493345139)*x[1]+(-0.561910153781)*x[2]
        arg[3,2,2,0]=(0.0761644634428)*x[0]+(-0.700940967343)*x[1]+(1.19407210904)*x[2]
        arg[3,2,2,1]=(1.75717437065)*x[0]+(-0.144457317494)*x[1]+(-0.523233453079)*x[2]
        arg[3,3,0,0]=(0.448782872606)*x[0]+(0.650464256547)*x[1]+(1.19476547456)*x[2]
        arg[3,3,0,1]=(0.118586027201)*x[0]+(0.449971816918)*x[1]+(-0.510812733389)*x[2]
        arg[3,3,1,0]=(-0.330032265081)*x[0]+(-0.13657534891)*x[1]+(0.456365289556)*x[2]
        arg[3,3,1,1]=(0.697584212635)*x[0]+(0.897604241724)*x[1]+(1.31517116663)*x[2]
        arg[3,3,2,0]=(-0.940538707365)*x[0]+(-0.334862221665)*x[1]+(0.0969742096754)*x[2]
        arg[3,3,2,1]=(-0.985788813516)*x[0]+(0.348389725548)*x[1]+(-0.260816590361)*x[2]
        arg[3,4,0,0]=(0.754692421417)*x[0]+(0.775985781769)*x[1]+(-0.0866159222457)*x[2]
        arg[3,4,0,1]=(-1.17670502429)*x[0]+(1.12268279393)*x[1]+(-1.1279822095)*x[2]
        arg[3,4,1,0]=(-1.48674370666)*x[0]+(1.26685445713)*x[1]+(-0.0522828809306)*x[2]
        arg[3,4,1,1]=(0.401760689843)*x[0]+(-1.35707797971)*x[1]+(-0.0580654273828)*x[2]
        arg[3,4,2,0]=(-1.40697830852)*x[0]+(0.933923369958)*x[1]+(0.230152169159)*x[2]
        arg[3,4,2,1]=(-1.21242130555)*x[0]+(0.858316339663)*x[1]+(-0.839747078526)*x[2]
        ref[0,0,0,0]=(-1.29218217256)*x_ref[0]+(-0.382937892213)*x_ref[1]+(0.14363896447)*x_ref[2]
        ref[0,0,0,1]=(1.39856942421)*x_ref[0]+(0.302082267835)*x_ref[1]+(-0.40870736584)*x_ref[2]
        ref[0,0,1,0]=(0.523846108043)*x_ref[0]+(1.19961531204)*x_ref[1]+(0.838881761349)*x_ref[2]
        ref[0,0,1,1]=(0.634542282764)*x_ref[0]+(-0.475514686466)*x_ref[1]+(-1.68745428486)*x_ref[2]
        ref[0,0,2,0]=(1.16379585731)*x_ref[0]+(-0.928037595416)*x_ref[1]+(0.998857178353)*x_ref[2]
        ref[0,0,2,1]=(-0.804515607022)*x_ref[0]+(-0.915469493439)*x_ref[1]+(-0.143888000623)*x_ref[2]
        ref[0,1,0,0]=(0.925843155486)*x_ref[0]+(-0.168243418953)*x_ref[1]+(-0.887839555187)*x_ref[2]
        ref[0,1,0,1]=(-0.913103105329)*x_ref[0]+(1.38011681732)*x_ref[1]+(-0.992478793291)*x_ref[2]
        ref[0,1,1,0]=(0.759685289903)*x_ref[0]+(-0.579481939982)*x_ref[1]+(0.577682929138)*x_ref[2]
        ref[0,1,1,1]=(0.813155548228)*x_ref[0]+(-0.178317843217)*x_ref[1]+(-0.0460113194526)*x_ref[2]
        ref[0,1,2,0]=(-0.195764860141)*x_ref[0]+(1.44256979628)*x_ref[1]+(0.283992441343)*x_ref[2]
        ref[0,1,2,1]=(-0.921229459072)*x_ref[0]+(-1.19295761547)*x_ref[1]+(0.215736801944)*x_ref[2]
        ref[0,2,0,0]=(-0.741726331752)*x_ref[0]+(-0.225585937382)*x_ref[1]+(1.12654556466)*x_ref[2]
        ref[0,2,0,1]=(-0.709656162421)*x_ref[0]+(0.298862010425)*x_ref[1]+(-0.502443284827)*x_ref[2]
        ref[0,2,1,0]=(-1.01166188669)*x_ref[0]+(-1.43763408953)*x_ref[1]+(-0.835160398488)*x_ref[2]
        ref[0,2,1,1]=(1.80968424545)*x_ref[0]+(-0.802666103495)*x_ref[1]+(0.290283606115)*x_ref[2]
        ref[0,2,2,0]=(0.346481981585)*x_ref[0]+(-0.444176962215)*x_ref[1]+(0.689127612227)*x_ref[2]
        ref[0,2,2,1]=(1.06925494095)*x_ref[0]+(0.452854988449)*x_ref[1]+(-1.19974690364)*x_ref[2]
        ref[0,3,0,0]=(-0.534896036328)*x_ref[0]+(-0.312000770615)*x_ref[1]+(-0.440033746693)*x_ref[2]
        ref[0,3,0,1]=(0.580351431838)*x_ref[0]+(-0.36623848451)*x_ref[1]+(-0.507837490037)*x_ref[2]
        ref[0,3,1,0]=(0.383180910522)*x_ref[0]+(-0.0907152443563)*x_ref[1]+(-0.200103743943)*x_ref[2]
        ref[0,3,1,1]=(0.0160194687513)*x_ref[0]+(0.862509612939)*x_ref[1]+(-1.14643594474)*x_ref[2]
        ref[0,3,2,0]=(-0.224390850199)*x_ref[0]+(-0.176927715447)*x_ref[1]+(0.213480525831)*x_ref[2]
        ref[0,3,2,1]=(-0.0508511448148)*x_ref[0]+(-1.28943467389)*x_ref[1]+(0.902285302487)*x_ref[2]
        ref[0,4,0,0]=(-0.633106625232)*x_ref[0]+(-0.430651917485)*x_ref[1]+(0.583149841952)*x_ref[2]
        ref[0,4,0,1]=(0.579035528956)*x_ref[0]+(0.159439682519)*x_ref[1]+(-0.0926277063021)*x_ref[2]
        ref[0,4,1,0]=(0.00484756366973)*x_ref[0]+(0.0324822270755)*x_ref[1]+(1.20052413378)*x_ref[2]
        ref[0,4,1,1]=(-0.337383994688)*x_ref[0]+(0.902076367981)*x_ref[1]+(-0.1956377395)*x_ref[2]
        ref[0,4,2,0]=(0.928434343456)*x_ref[0]+(0.0376569884355)*x_ref[1]+(-0.155274118568)*x_ref[2]
        ref[0,4,2,1]=(-0.541990213238)*x_ref[0]+(-0.0165076697829)*x_ref[1]+(-0.210319224859)*x_ref[2]
        ref[1,0,0,0]=(-1.67762025181)*x_ref[0]+(1.61434755579)*x_ref[1]+(0.357272049999)*x_ref[2]
        ref[1,0,0,1]=(-1.62532544549)*x_ref[0]+(-0.271426231027)*x_ref[1]+(0.338680991778)*x_ref[2]
        ref[1,0,1,0]=(1.73523408033)*x_ref[0]+(1.11987092368)*x_ref[1]+(0.126769364273)*x_ref[2]
        ref[1,0,1,1]=(-1.49220786558)*x_ref[0]+(1.51915566092)*x_ref[1]+(1.24788192157)*x_ref[2]
        ref[1,0,2,0]=(-0.687808766765)*x_ref[0]+(-1.02318754737)*x_ref[1]+(0.590653972416)*x_ref[2]
        ref[1,0,2,1]=(0.756067354802)*x_ref[0]+(-0.66964706965)*x_ref[1]+(-1.18986499158)*x_ref[2]
        ref[1,1,0,0]=(0.371345732177)*x_ref[0]+(0.212739617115)*x_ref[1]+(-1.13439047405)*x_ref[2]
        ref[1,1,0,1]=(0.525761804111)*x_ref[0]+(0.144130399691)*x_ref[1]+(0.715517587677)*x_ref[2]
        ref[1,1,1,0]=(-0.376160293215)*x_ref[0]+(-0.469964935263)*x_ref[1]+(0.413518194707)*x_ref[2]
        ref[1,1,1,1]=(-0.787748398144)*x_ref[0]+(-0.810090402237)*x_ref[1]+(-0.50127109256)*x_ref[2]
        ref[1,1,2,0]=(0.654555083692)*x_ref[0]+(-0.142134700515)*x_ref[1]+(-1.55891930521)*x_ref[2]
        ref[1,1,2,1]=(-0.423058885456)*x_ref[0]+(-0.65930158641)*x_ref[1]+(0.94440057744)*x_ref[2]
        ref[1,2,0,0]=(-0.924832007053)*x_ref[0]+(1.0154890382)*x_ref[1]+(0.11938079748)*x_ref[2]
        ref[1,2,0,1]=(0.134642686931)*x_ref[0]+(-1.30908067998)*x_ref[1]+(0.367206139074)*x_ref[2]
        ref[1,2,1,0]=(1.55524055201)*x_ref[0]+(0.727208660637)*x_ref[1]+(0.809810917377)*x_ref[2]
        ref[1,2,1,1]=(0.94813391462)*x_ref[0]+(-1.49402524308)*x_ref[1]+(0.367824542434)*x_ref[2]
        ref[1,2,2,0]=(-0.608295798731)*x_ref[0]+(-0.142302252814)*x_ref[1]+(-0.986529251308)*x_ref[2]
        ref[1,2,2,1]=(0.0259394619607)*x_ref[0]+(-1.60728202467)*x_ref[1]+(0.312534656619)*x_ref[2]
        ref[1,3,0,0]=(-1.34063373729)*x_ref[0]+(-0.268499839329)*x_ref[1]+(-0.335125886299)*x_ref[2]
        ref[1,3,0,1]=(1.53280698072)*x_ref[0]+(1.24369684711)*x_ref[1]+(0.766369156132)*x_ref[2]
        ref[1,3,1,0]=(0.902067176876)*x_ref[0]+(0.791102994814)*x_ref[1]+(-0.89266430493)*x_ref[2]
        ref[1,3,1,1]=(0.704997629232)*x_ref[0]+(-0.551611360332)*x_ref[1]+(1.39134537491)*x_ref[2]
        ref[1,3,2,0]=(-0.151690675211)*x_ref[0]+(-1.01762975893)*x_ref[1]+(-0.627493214284)*x_ref[2]
        ref[1,3,2,1]=(1.00936645922)*x_ref[0]+(0.0322458997248)*x_ref[1]+(1.45039938194)*x_ref[2]
        ref[1,4,0,0]=(-0.0916390388904)*x_ref[0]+(-0.783111746469)*x_ref[1]+(0.389725342225)*x_ref[2]
        ref[1,4,0,1]=(1.18859328243)*x_ref[0]+(0.871689661558)*x_ref[1]+(-0.775464713448)*x_ref[2]
        ref[1,4,1,0]=(0.913118342488)*x_ref[0]+(1.53285648884)*x_ref[1]+(0.0947365101243)*x_ref[2]
        ref[1,4,1,1]=(-1.00735841224)*x_ref[0]+(-1.12544193607)*x_ref[1]+(0.267347229372)*x_ref[2]
        ref[1,4,2,0]=(0.285943438284)*x_ref[0]+(-1.0889389244)*x_ref[1]+(-1.03054842327)*x_ref[2]
        ref[1,4,2,1]=(-1.5768940139)*x_ref[0]+(0.91237754587)*x_ref[1]+(1.30220394199)*x_ref[2]
        ref[2,0,0,0]=(-0.0634935883277)*x_ref[0]+(0.157304131712)*x_ref[1]+(-1.2995042074)*x_ref[2]
        ref[2,0,0,1]=(-0.802705266871)*x_ref[0]+(0.238233162774)*x_ref[1]+(1.29463414013)*x_ref[2]
        ref[2,0,1,0]=(-0.925468462671)*x_ref[0]+(0.870960795966)*x_ref[1]+(0.156522765713)*x_ref[2]
        ref[2,0,1,1]=(-0.546666833309)*x_ref[0]+(-1.14649607995)*x_ref[1]+(-1.53906505048)*x_ref[2]
        ref[2,0,2,0]=(0.464952910857)*x_ref[0]+(0.408605786027)*x_ref[1]+(-0.886974881114)*x_ref[2]
        ref[2,0,2,1]=(0.491723247948)*x_ref[0]+(0.194616420077)*x_ref[1]+(0.409049782749)*x_ref[2]
        ref[2,1,0,0]=(-1.19836883114)*x_ref[0]+(1.0622722337)*x_ref[1]+(-0.109193867727)*x_ref[2]
        ref[2,1,0,1]=(1.01973600383)*x_ref[0]+(-0.154482220539)*x_ref[1]+(0.378696371956)*x_ref[2]
        ref[2,1,1,0]=(-0.0812015296246)*x_ref[0]+(-1.10252794568)*x_ref[1]+(-0.00585426581857)*x_ref[2]
        ref[2,1,1,1]=(0.536042665272)*x_ref[0]+(0.685507117255)*x_ref[1]+(-0.179652683139)*x_ref[2]
        ref[2,1,2,0]=(-0.36750467811)*x_ref[0]+(-1.23674550869)*x_ref[1]+(-0.762905702801)*x_ref[2]
        ref[2,1,2,1]=(0.329665965651)*x_ref[0]+(0.101195810737)*x_ref[1]+(-0.610015052375)*x_ref[2]
        ref[2,2,0,0]=(-0.81644635185)*x_ref[0]+(-0.604578476306)*x_ref[1]+(0.0440925706991)*x_ref[2]
        ref[2,2,0,1]=(-0.958241437971)*x_ref[0]+(0.38563498206)*x_ref[1]+(0.913406762318)*x_ref[2]
        ref[2,2,1,0]=(1.22027492224)*x_ref[0]+(-1.00726209939)*x_ref[1]+(-1.05083068303)*x_ref[2]
        ref[2,2,1,1]=(-1.47561192645)*x_ref[0]+(-0.903549355608)*x_ref[1]+(0.236771324671)*x_ref[2]
        ref[2,2,2,0]=(0.284995864449)*x_ref[0]+(-0.175760829921)*x_ref[1]+(-0.720607220001)*x_ref[2]
        ref[2,2,2,1]=(-0.535894515963)*x_ref[0]+(-0.821302957162)*x_ref[1]+(-1.75141540241)*x_ref[2]
        ref[2,3,0,0]=(-0.333082106674)*x_ref[0]+(0.579471826412)*x_ref[1]+(0.561184138108)*x_ref[2]
        ref[2,3,0,1]=(1.34803519557)*x_ref[0]+(-0.441299088849)*x_ref[1]+(-1.04945951616)*x_ref[2]
        ref[2,3,1,0]=(0.299011096602)*x_ref[0]+(1.32726458235)*x_ref[1]+(-0.196298687359)*x_ref[2]
        ref[2,3,1,1]=(1.48698834204)*x_ref[0]+(0.527221496971)*x_ref[1]+(0.0819813519422)*x_ref[2]
        ref[2,3,2,0]=(0.681680955153)*x_ref[0]+(0.503653972357)*x_ref[1]+(0.434669347891)*x_ref[2]
        ref[2,3,2,1]=(-0.152586789311)*x_ref[0]+(-0.765242635579)*x_ref[1]+(0.712663994387)*x_ref[2]
        ref[2,4,0,0]=(0.630152300463)*x_ref[0]+(1.12366733588)*x_ref[1]+(-0.662966700638)*x_ref[2]
        ref[2,4,0,1]=(1.02796709755)*x_ref[0]+(-0.24093881584)*x_ref[1]+(1.79445417019)*x_ref[2]
        ref[2,4,1,0]=(-0.954575760758)*x_ref[0]+(0.16331782608)*x_ref[1]+(1.38309921277)*x_ref[2]
        ref[2,4,1,1]=(0.536092514281)*x_ref[0]+(-0.276364952673)*x_ref[1]+(0.779619701528)*x_ref[2]
        ref[2,4,2,0]=(0.267270444141)*x_ref[0]+(-0.810746424331)*x_ref[1]+(-0.733877853264)*x_ref[2]
        ref[2,4,2,1]=(-0.0641688847923)*x_ref[0]+(-0.386279280667)*x_ref[1]+(0.0637571595255)*x_ref[2]
        ref[3,0,0,0]=(0.469903180687)*x_ref[0]+(-0.793812397342)*x_ref[1]+(1.60841921779)*x_ref[2]
        ref[3,0,0,1]=(0.720057600102)*x_ref[0]+(0.502582253261)*x_ref[1]+(0.460147791003)*x_ref[2]
        ref[3,0,1,0]=(0.0327862564137)*x_ref[0]+(-1.142962965)*x_ref[1]+(0.114799131894)*x_ref[2]
        ref[3,0,1,1]=(-0.997412656978)*x_ref[0]+(-0.678175468576)*x_ref[1]+(-0.152663147682)*x_ref[2]
        ref[3,0,2,0]=(0.810424599078)*x_ref[0]+(0.75522727786)*x_ref[1]+(-0.205880941501)*x_ref[2]
        ref[3,0,2,1]=(0.353601949476)*x_ref[0]+(0.0963200813379)*x_ref[1]+(0.0578829287146)*x_ref[2]
        ref[3,1,0,0]=(0.346590730062)*x_ref[0]+(1.07953790092)*x_ref[1]+(0.647586851217)*x_ref[2]
        ref[3,1,0,1]=(0.889879195334)*x_ref[0]+(-0.352788186249)*x_ref[1]+(0.434900135803)*x_ref[2]
        ref[3,1,1,0]=(-0.721742734959)*x_ref[0]+(0.166066672134)*x_ref[1]+(-0.489266924809)*x_ref[2]
        ref[3,1,1,1]=(-0.810487755181)*x_ref[0]+(0.0354916979253)*x_ref[1]+(0.0735520654395)*x_ref[2]
        ref[3,1,2,0]=(1.02414338939)*x_ref[0]+(-0.851076284556)*x_ref[1]+(-0.232174494405)*x_ref[2]
        ref[3,1,2,1]=(-0.327246566026)*x_ref[0]+(-0.0499396023605)*x_ref[1]+(0.611585402379)*x_ref[2]
        ref[3,2,0,0]=(-0.365757628557)*x_ref[0]+(0.868697342115)*x_ref[1]+(-1.102545972)*x_ref[2]
        ref[3,2,0,1]=(0.628592129227)*x_ref[0]+(1.25545515311)*x_ref[1]+(-1.0577476264)*x_ref[2]
        ref[3,2,1,0]=(0.160995978044)*x_ref[0]+(-0.947389499556)*x_ref[1]+(-0.870476160656)*x_ref[2]
        ref[3,2,1,1]=(0.910787139938)*x_ref[0]+(-0.0681493345139)*x_ref[1]+(-0.561910153781)*x_ref[2]
        ref[3,2,2,0]=(0.0761644634428)*x_ref[0]+(-0.700940967343)*x_ref[1]+(1.19407210904)*x_ref[2]
        ref[3,2,2,1]=(1.75717437065)*x_ref[0]+(-0.144457317494)*x_ref[1]+(-0.523233453079)*x_ref[2]
        ref[3,3,0,0]=(0.448782872606)*x_ref[0]+(0.650464256547)*x_ref[1]+(1.19476547456)*x_ref[2]
        ref[3,3,0,1]=(0.118586027201)*x_ref[0]+(0.449971816918)*x_ref[1]+(-0.510812733389)*x_ref[2]
        ref[3,3,1,0]=(-0.330032265081)*x_ref[0]+(-0.13657534891)*x_ref[1]+(0.456365289556)*x_ref[2]
        ref[3,3,1,1]=(0.697584212635)*x_ref[0]+(0.897604241724)*x_ref[1]+(1.31517116663)*x_ref[2]
        ref[3,3,2,0]=(-0.940538707365)*x_ref[0]+(-0.334862221665)*x_ref[1]+(0.0969742096754)*x_ref[2]
        ref[3,3,2,1]=(-0.985788813516)*x_ref[0]+(0.348389725548)*x_ref[1]+(-0.260816590361)*x_ref[2]
        ref[3,4,0,0]=(0.754692421417)*x_ref[0]+(0.775985781769)*x_ref[1]+(-0.0866159222457)*x_ref[2]
        ref[3,4,0,1]=(-1.17670502429)*x_ref[0]+(1.12268279393)*x_ref[1]+(-1.1279822095)*x_ref[2]
        ref[3,4,1,0]=(-1.48674370666)*x_ref[0]+(1.26685445713)*x_ref[1]+(-0.0522828809306)*x_ref[2]
        ref[3,4,1,1]=(0.401760689843)*x_ref[0]+(-1.35707797971)*x_ref[1]+(-0.0580654273828)*x_ref[2]
        ref[3,4,2,0]=(-1.40697830852)*x_ref[0]+(0.933923369958)*x_ref[1]+(0.230152169159)*x_ref[2]
        ref[3,4,2,1]=(-1.21242130555)*x_ref[0]+(0.858316339663)*x_ref[1]+(-0.839747078526)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.485106931655)*x[0]**o+(-0.45911691249)*x[0]+(-0.780654139819)*x[1]**o+(-0.496275334303)*x[1]
        ref=(-0.485106931655)*x_ref[0]**o+(-0.45911691249)*x_ref[0]+(-0.780654139819)*x_ref[1]**o+(-0.496275334303)*x_ref[1]
      else:
        arg=(-0.312233545328)*x[0]**o+(0.386652391857)*x[0]+(0.893938213685)*x[1]**o+(-0.109836308832)*x[1]+(0.293419600866)*x[2]**o+(-0.802363407879)*x[2]
        ref=(-0.312233545328)*x_ref[0]**o+(0.386652391857)*x_ref[0]+(0.893938213685)*x_ref[1]**o+(-0.109836308832)*x_ref[1]+(0.293419600866)*x_ref[2]**o+(-0.802363407879)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.383147346126)*x[0]**o+(-0.103098043084)*x[0]+(0.389581571033)*x[1]**o+(0.647441523744)*x[1]
        arg[1]=(-0.0718078186662)*x[0]**o+(-0.975209175853)*x[0]+(-0.62913306056)*x[1]**o+(0.648038487393)*x[1]
        ref[0]=(0.383147346126)*x_ref[0]**o+(-0.103098043084)*x_ref[0]+(0.389581571033)*x_ref[1]**o+(0.647441523744)*x_ref[1]
        ref[1]=(-0.0718078186662)*x_ref[0]**o+(-0.975209175853)*x_ref[0]+(-0.62913306056)*x_ref[1]**o+(0.648038487393)*x_ref[1]
      else:
        arg[0]=(-0.392958184101)*x[0]**o+(-0.529709692826)*x[0]+(-0.834422900831)*x[1]**o+(0.52562214773)*x[1]+(-0.726762801989)*x[2]**o+(-0.329225625652)*x[2]
        arg[1]=(-0.388486489727)*x[0]**o+(0.62353995706)*x[0]+(0.0198506797091)*x[1]**o+(0.744336375443)*x[1]+(0.115966573066)*x[2]**o+(-0.195838697299)*x[2]
        ref[0]=(-0.392958184101)*x_ref[0]**o+(-0.529709692826)*x_ref[0]+(-0.834422900831)*x_ref[1]**o+(0.52562214773)*x_ref[1]+(-0.726762801989)*x_ref[2]**o+(-0.329225625652)*x_ref[2]
        ref[1]=(-0.388486489727)*x_ref[0]**o+(0.62353995706)*x_ref[0]+(0.0198506797091)*x_ref[1]**o+(0.744336375443)*x_ref[1]+(0.115966573066)*x_ref[2]**o+(-0.195838697299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.657915437562)*x[0]**o+(-0.545113889022)*x[0]+(0.531900702747)*x[1]**o+(0.593791195717)*x[1]
        arg[0,1]=(0.051080053648)*x[0]**o+(-0.693135614033)*x[0]+(0.132124625293)*x[1]**o+(0.790120778986)*x[1]
        arg[0,2]=(-0.474326324304)*x[0]**o+(0.844119715746)*x[0]+(-0.107139623866)*x[1]**o+(0.456623369855)*x[1]
        arg[0,3]=(-0.582421376463)*x[0]**o+(0.972810074845)*x[0]+(0.875278994526)*x[1]**o+(-0.382656553829)*x[1]
        arg[0,4]=(-0.904956623057)*x[0]**o+(0.701968065151)*x[0]+(-0.710231782855)*x[1]**o+(-0.123456684979)*x[1]
        arg[1,0]=(0.361017539866)*x[0]**o+(0.865249575899)*x[0]+(0.529515655309)*x[1]**o+(0.161402780958)*x[1]
        arg[1,1]=(0.813945406612)*x[0]**o+(0.801153720706)*x[0]+(-0.254227034266)*x[1]**o+(-0.422848409284)*x[1]
        arg[1,2]=(0.56207473833)*x[0]**o+(-0.831638855484)*x[0]+(-0.584619648965)*x[1]**o+(0.453944553892)*x[1]
        arg[1,3]=(0.25526821881)*x[0]**o+(-0.436135120045)*x[0]+(0.524842439137)*x[1]**o+(0.568007117728)*x[1]
        arg[1,4]=(-0.0587551077421)*x[0]**o+(-0.434635762185)*x[0]+(-0.592094375986)*x[1]**o+(0.873123434395)*x[1]
        arg[2,0]=(0.266819439236)*x[0]**o+(-0.45156140915)*x[0]+(0.436756372775)*x[1]**o+(0.12295128383)*x[1]
        arg[2,1]=(0.347128677451)*x[0]**o+(-0.0983169467916)*x[0]+(-0.414573209624)*x[1]**o+(-0.171300108487)*x[1]
        arg[2,2]=(-0.279695154296)*x[0]**o+(-0.911450144779)*x[0]+(0.5055340448)*x[1]**o+(-0.858058864152)*x[1]
        arg[2,3]=(-0.53768906863)*x[0]**o+(-0.0626923004303)*x[0]+(0.72295545269)*x[1]**o+(-0.77070425804)*x[1]
        arg[2,4]=(0.18215032008)*x[0]**o+(-0.965014882526)*x[0]+(0.596564813119)*x[1]**o+(0.0817024840554)*x[1]
        arg[3,0]=(-0.572404596368)*x[0]**o+(0.488365635829)*x[0]+(0.145439800826)*x[1]**o+(-0.337589667899)*x[1]
        arg[3,1]=(-0.628977426654)*x[0]**o+(0.22731945375)*x[0]+(0.989224773975)*x[1]**o+(0.0691553279499)*x[1]
        arg[3,2]=(-0.43609079916)*x[0]**o+(0.966113801416)*x[0]+(-0.853136416278)*x[1]**o+(-0.0125151911906)*x[1]
        arg[3,3]=(0.340159347855)*x[0]**o+(-0.254775013843)*x[0]+(0.795261224144)*x[1]**o+(-0.692255020659)*x[1]
        arg[3,4]=(-0.153521724504)*x[0]**o+(-0.681779299881)*x[0]+(0.693860779041)*x[1]**o+(0.882012604794)*x[1]
        ref[0,0]=(-0.657915437562)*x_ref[0]**o+(-0.545113889022)*x_ref[0]+(0.531900702747)*x_ref[1]**o+(0.593791195717)*x_ref[1]
        ref[0,1]=(0.051080053648)*x_ref[0]**o+(-0.693135614033)*x_ref[0]+(0.132124625293)*x_ref[1]**o+(0.790120778986)*x_ref[1]
        ref[0,2]=(-0.474326324304)*x_ref[0]**o+(0.844119715746)*x_ref[0]+(-0.107139623866)*x_ref[1]**o+(0.456623369855)*x_ref[1]
        ref[0,3]=(-0.582421376463)*x_ref[0]**o+(0.972810074845)*x_ref[0]+(0.875278994526)*x_ref[1]**o+(-0.382656553829)*x_ref[1]
        ref[0,4]=(-0.904956623057)*x_ref[0]**o+(0.701968065151)*x_ref[0]+(-0.710231782855)*x_ref[1]**o+(-0.123456684979)*x_ref[1]
        ref[1,0]=(0.361017539866)*x_ref[0]**o+(0.865249575899)*x_ref[0]+(0.529515655309)*x_ref[1]**o+(0.161402780958)*x_ref[1]
        ref[1,1]=(0.813945406612)*x_ref[0]**o+(0.801153720706)*x_ref[0]+(-0.254227034266)*x_ref[1]**o+(-0.422848409284)*x_ref[1]
        ref[1,2]=(0.56207473833)*x_ref[0]**o+(-0.831638855484)*x_ref[0]+(-0.584619648965)*x_ref[1]**o+(0.453944553892)*x_ref[1]
        ref[1,3]=(0.25526821881)*x_ref[0]**o+(-0.436135120045)*x_ref[0]+(0.524842439137)*x_ref[1]**o+(0.568007117728)*x_ref[1]
        ref[1,4]=(-0.0587551077421)*x_ref[0]**o+(-0.434635762185)*x_ref[0]+(-0.592094375986)*x_ref[1]**o+(0.873123434395)*x_ref[1]
        ref[2,0]=(0.266819439236)*x_ref[0]**o+(-0.45156140915)*x_ref[0]+(0.436756372775)*x_ref[1]**o+(0.12295128383)*x_ref[1]
        ref[2,1]=(0.347128677451)*x_ref[0]**o+(-0.0983169467916)*x_ref[0]+(-0.414573209624)*x_ref[1]**o+(-0.171300108487)*x_ref[1]
        ref[2,2]=(-0.279695154296)*x_ref[0]**o+(-0.911450144779)*x_ref[0]+(0.5055340448)*x_ref[1]**o+(-0.858058864152)*x_ref[1]
        ref[2,3]=(-0.53768906863)*x_ref[0]**o+(-0.0626923004303)*x_ref[0]+(0.72295545269)*x_ref[1]**o+(-0.77070425804)*x_ref[1]
        ref[2,4]=(0.18215032008)*x_ref[0]**o+(-0.965014882526)*x_ref[0]+(0.596564813119)*x_ref[1]**o+(0.0817024840554)*x_ref[1]
        ref[3,0]=(-0.572404596368)*x_ref[0]**o+(0.488365635829)*x_ref[0]+(0.145439800826)*x_ref[1]**o+(-0.337589667899)*x_ref[1]
        ref[3,1]=(-0.628977426654)*x_ref[0]**o+(0.22731945375)*x_ref[0]+(0.989224773975)*x_ref[1]**o+(0.0691553279499)*x_ref[1]
        ref[3,2]=(-0.43609079916)*x_ref[0]**o+(0.966113801416)*x_ref[0]+(-0.853136416278)*x_ref[1]**o+(-0.0125151911906)*x_ref[1]
        ref[3,3]=(0.340159347855)*x_ref[0]**o+(-0.254775013843)*x_ref[0]+(0.795261224144)*x_ref[1]**o+(-0.692255020659)*x_ref[1]
        ref[3,4]=(-0.153521724504)*x_ref[0]**o+(-0.681779299881)*x_ref[0]+(0.693860779041)*x_ref[1]**o+(0.882012604794)*x_ref[1]
      else:
        arg[0,0]=(-0.0201751669244)*x[0]**o+(0.337967826628)*x[0]+(0.262482967798)*x[1]**o+(-0.192863458488)*x[1]+(-0.696822606165)*x[2]**o+(-0.476556518002)*x[2]
        arg[0,1]=(-0.546277895941)*x[0]**o+(-0.118793125394)*x[0]+(0.987786322041)*x[1]**o+(-0.241472898602)*x[1]+(-0.454308453322)*x[2]**o+(-0.27057260464)*x[2]
        arg[0,2]=(0.383631846628)*x[0]**o+(0.979909217026)*x[0]+(-0.0503694684178)*x[1]**o+(-0.748153155067)*x[1]+(-0.721832849635)*x[2]**o+(0.144937710994)*x[2]
        arg[0,3]=(-0.540144273345)*x[0]**o+(-0.986853732479)*x[0]+(-0.14527387031)*x[1]**o+(0.329815126012)*x[1]+(0.566431907689)*x[2]**o+(-0.323179211167)*x[2]
        arg[0,4]=(-0.448624394254)*x[0]**o+(-0.272742080619)*x[0]+(-0.451405020267)*x[1]**o+(0.173204890481)*x[1]+(-0.0122463571841)*x[2]**o+(-0.595932249794)*x[2]
        arg[1,0]=(0.984160677776)*x[0]**o+(-0.243512460216)*x[0]+(0.82590496761)*x[1]**o+(-0.448270326787)*x[1]+(0.286008153265)*x[2]**o+(0.804229889769)*x[2]
        arg[1,1]=(0.531762548093)*x[0]**o+(0.211455783615)*x[0]+(0.836058347792)*x[1]**o+(0.806904570574)*x[1]+(-0.052512287992)*x[2]**o+(0.998679219355)*x[2]
        arg[1,2]=(0.462615968956)*x[0]**o+(0.960862689218)*x[0]+(-0.693932866844)*x[1]**o+(0.881110014782)*x[1]+(0.976329091354)*x[2]**o+(-0.647483316997)*x[2]
        arg[1,3]=(-0.764181748932)*x[0]**o+(-0.962197871121)*x[0]+(-0.143747029438)*x[1]**o+(0.913962833322)*x[1]+(0.165013597835)*x[2]**o+(0.166877976184)*x[2]
        arg[1,4]=(0.687688905493)*x[0]**o+(-0.152816071208)*x[0]+(-0.277974702822)*x[1]**o+(-0.389918839086)*x[1]+(0.229658530129)*x[2]**o+(-0.69924627348)*x[2]
        arg[2,0]=(0.0598215047949)*x[0]**o+(0.592389000293)*x[0]+(0.00886573296364)*x[1]**o+(0.339086619934)*x[1]+(0.476600446349)*x[2]**o+(0.767524427684)*x[2]
        arg[2,1]=(0.525936212445)*x[0]**o+(0.44607860423)*x[0]+(0.0397184541609)*x[1]**o+(-0.424118244789)*x[1]+(-0.273077104391)*x[2]**o+(0.435027185784)*x[2]
        arg[2,2]=(0.672992314995)*x[0]**o+(0.993339100191)*x[0]+(0.488983819898)*x[1]**o+(0.568579529177)*x[1]+(-0.321282052211)*x[2]**o+(-0.268847753649)*x[2]
        arg[2,3]=(-0.0536748741467)*x[0]**o+(-0.195658514433)*x[0]+(-0.931652264383)*x[1]**o+(-0.89957282695)*x[1]+(-0.811088135231)*x[2]**o+(0.0699177832759)*x[2]
        arg[2,4]=(-0.707875141752)*x[0]**o+(-0.00297935832343)*x[0]+(-0.959580991547)*x[1]**o+(0.0787110832272)*x[1]+(-0.754646316678)*x[2]**o+(-0.432731727467)*x[2]
        arg[3,0]=(0.728853463111)*x[0]**o+(-0.681375546727)*x[0]+(-0.283722278591)*x[1]**o+(0.950522890527)*x[1]+(0.608194817491)*x[2]**o+(0.571156129477)*x[2]
        arg[3,1]=(-0.0208969501228)*x[0]**o+(0.572596028151)*x[0]+(0.811179009214)*x[1]**o+(0.519098352581)*x[1]+(0.154527510954)*x[2]**o+(-0.21413742337)*x[2]
        arg[3,2]=(0.347209642455)*x[0]**o+(-0.114911324326)*x[0]+(-0.104561779357)*x[1]**o+(0.848703366841)*x[1]+(0.680312939858)*x[2]**o+(-0.680566404979)*x[2]
        arg[3,3]=(0.342687985669)*x[0]**o+(0.443470953221)*x[0]+(-0.849668173915)*x[1]**o+(0.337689225078)*x[1]+(-0.35083505383)*x[2]**o+(-0.284627678739)*x[2]
        arg[3,4]=(0.661690111131)*x[0]**o+(-0.0918926979359)*x[0]+(-0.587753600932)*x[1]**o+(-0.383115718043)*x[1]+(-0.0696588695623)*x[2]**o+(-0.799630628972)*x[2]
        ref[0,0]=(-0.0201751669244)*x_ref[0]**o+(0.337967826628)*x_ref[0]+(0.262482967798)*x_ref[1]**o+(-0.192863458488)*x_ref[1]+(-0.696822606165)*x_ref[2]**o+(-0.476556518002)*x_ref[2]
        ref[0,1]=(-0.546277895941)*x_ref[0]**o+(-0.118793125394)*x_ref[0]+(0.987786322041)*x_ref[1]**o+(-0.241472898602)*x_ref[1]+(-0.454308453322)*x_ref[2]**o+(-0.27057260464)*x_ref[2]
        ref[0,2]=(0.383631846628)*x_ref[0]**o+(0.979909217026)*x_ref[0]+(-0.0503694684178)*x_ref[1]**o+(-0.748153155067)*x_ref[1]+(-0.721832849635)*x_ref[2]**o+(0.144937710994)*x_ref[2]
        ref[0,3]=(-0.540144273345)*x_ref[0]**o+(-0.986853732479)*x_ref[0]+(-0.14527387031)*x_ref[1]**o+(0.329815126012)*x_ref[1]+(0.566431907689)*x_ref[2]**o+(-0.323179211167)*x_ref[2]
        ref[0,4]=(-0.448624394254)*x_ref[0]**o+(-0.272742080619)*x_ref[0]+(-0.451405020267)*x_ref[1]**o+(0.173204890481)*x_ref[1]+(-0.0122463571841)*x_ref[2]**o+(-0.595932249794)*x_ref[2]
        ref[1,0]=(0.984160677776)*x_ref[0]**o+(-0.243512460216)*x_ref[0]+(0.82590496761)*x_ref[1]**o+(-0.448270326787)*x_ref[1]+(0.286008153265)*x_ref[2]**o+(0.804229889769)*x_ref[2]
        ref[1,1]=(0.531762548093)*x_ref[0]**o+(0.211455783615)*x_ref[0]+(0.836058347792)*x_ref[1]**o+(0.806904570574)*x_ref[1]+(-0.052512287992)*x_ref[2]**o+(0.998679219355)*x_ref[2]
        ref[1,2]=(0.462615968956)*x_ref[0]**o+(0.960862689218)*x_ref[0]+(-0.693932866844)*x_ref[1]**o+(0.881110014782)*x_ref[1]+(0.976329091354)*x_ref[2]**o+(-0.647483316997)*x_ref[2]
        ref[1,3]=(-0.764181748932)*x_ref[0]**o+(-0.962197871121)*x_ref[0]+(-0.143747029438)*x_ref[1]**o+(0.913962833322)*x_ref[1]+(0.165013597835)*x_ref[2]**o+(0.166877976184)*x_ref[2]
        ref[1,4]=(0.687688905493)*x_ref[0]**o+(-0.152816071208)*x_ref[0]+(-0.277974702822)*x_ref[1]**o+(-0.389918839086)*x_ref[1]+(0.229658530129)*x_ref[2]**o+(-0.69924627348)*x_ref[2]
        ref[2,0]=(0.0598215047949)*x_ref[0]**o+(0.592389000293)*x_ref[0]+(0.00886573296364)*x_ref[1]**o+(0.339086619934)*x_ref[1]+(0.476600446349)*x_ref[2]**o+(0.767524427684)*x_ref[2]
        ref[2,1]=(0.525936212445)*x_ref[0]**o+(0.44607860423)*x_ref[0]+(0.0397184541609)*x_ref[1]**o+(-0.424118244789)*x_ref[1]+(-0.273077104391)*x_ref[2]**o+(0.435027185784)*x_ref[2]
        ref[2,2]=(0.672992314995)*x_ref[0]**o+(0.993339100191)*x_ref[0]+(0.488983819898)*x_ref[1]**o+(0.568579529177)*x_ref[1]+(-0.321282052211)*x_ref[2]**o+(-0.268847753649)*x_ref[2]
        ref[2,3]=(-0.0536748741467)*x_ref[0]**o+(-0.195658514433)*x_ref[0]+(-0.931652264383)*x_ref[1]**o+(-0.89957282695)*x_ref[1]+(-0.811088135231)*x_ref[2]**o+(0.0699177832759)*x_ref[2]
        ref[2,4]=(-0.707875141752)*x_ref[0]**o+(-0.00297935832343)*x_ref[0]+(-0.959580991547)*x_ref[1]**o+(0.0787110832272)*x_ref[1]+(-0.754646316678)*x_ref[2]**o+(-0.432731727467)*x_ref[2]
        ref[3,0]=(0.728853463111)*x_ref[0]**o+(-0.681375546727)*x_ref[0]+(-0.283722278591)*x_ref[1]**o+(0.950522890527)*x_ref[1]+(0.608194817491)*x_ref[2]**o+(0.571156129477)*x_ref[2]
        ref[3,1]=(-0.0208969501228)*x_ref[0]**o+(0.572596028151)*x_ref[0]+(0.811179009214)*x_ref[1]**o+(0.519098352581)*x_ref[1]+(0.154527510954)*x_ref[2]**o+(-0.21413742337)*x_ref[2]
        ref[3,2]=(0.347209642455)*x_ref[0]**o+(-0.114911324326)*x_ref[0]+(-0.104561779357)*x_ref[1]**o+(0.848703366841)*x_ref[1]+(0.680312939858)*x_ref[2]**o+(-0.680566404979)*x_ref[2]
        ref[3,3]=(0.342687985669)*x_ref[0]**o+(0.443470953221)*x_ref[0]+(-0.849668173915)*x_ref[1]**o+(0.337689225078)*x_ref[1]+(-0.35083505383)*x_ref[2]**o+(-0.284627678739)*x_ref[2]
        ref[3,4]=(0.661690111131)*x_ref[0]**o+(-0.0918926979359)*x_ref[0]+(-0.587753600932)*x_ref[1]**o+(-0.383115718043)*x_ref[1]+(-0.0696588695623)*x_ref[2]**o+(-0.799630628972)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.906705618142)*x[0]**o+(0.458982439103)*x[0]+(0.556464353903)*x[1]**o+(0.733236686234)*x[1]
        arg[0,0,1]=(-0.192889954523)*x[0]**o+(-0.814332539133)*x[0]+(0.446164214282)*x[1]**o+(0.032291315446)*x[1]
        arg[0,1,0]=(0.757081929199)*x[0]**o+(-0.3938524931)*x[0]+(-0.458060132115)*x[1]**o+(0.471831491182)*x[1]
        arg[0,1,1]=(0.459868785407)*x[0]**o+(0.513707183316)*x[0]+(-0.9682413377)*x[1]**o+(-0.49496265187)*x[1]
        arg[1,0,0]=(0.546655212697)*x[0]**o+(-0.296125788768)*x[0]+(0.570033305927)*x[1]**o+(-0.818919094317)*x[1]
        arg[1,0,1]=(-0.954422252892)*x[0]**o+(-0.925035286684)*x[0]+(0.171311127546)*x[1]**o+(-0.584606597796)*x[1]
        arg[1,1,0]=(-0.203956024883)*x[0]**o+(-0.933117428128)*x[0]+(-0.745101180472)*x[1]**o+(0.0390208509919)*x[1]
        arg[1,1,1]=(-0.67467593628)*x[0]**o+(0.42281003355)*x[0]+(-0.495401176467)*x[1]**o+(0.993257306842)*x[1]
        arg[2,0,0]=(-0.204011890956)*x[0]**o+(0.730452092671)*x[0]+(-0.0486718200506)*x[1]**o+(0.691928726795)*x[1]
        arg[2,0,1]=(0.381571380631)*x[0]**o+(0.504896621541)*x[0]+(-0.229805983176)*x[1]**o+(0.281401365705)*x[1]
        arg[2,1,0]=(0.578852751326)*x[0]**o+(-0.38588684985)*x[0]+(-0.630908284751)*x[1]**o+(0.237972319424)*x[1]
        arg[2,1,1]=(-0.928366568501)*x[0]**o+(0.254484026833)*x[0]+(0.903931940538)*x[1]**o+(-0.370818194785)*x[1]
        arg[3,0,0]=(-0.216012465475)*x[0]**o+(-0.938274177001)*x[0]+(0.0674408710328)*x[1]**o+(-0.731594784098)*x[1]
        arg[3,0,1]=(-0.156790905207)*x[0]**o+(0.284403081253)*x[0]+(0.3754228052)*x[1]**o+(-0.772758808958)*x[1]
        arg[3,1,0]=(0.89987768188)*x[0]**o+(0.0177886564195)*x[0]+(0.102329016274)*x[1]**o+(0.474039949682)*x[1]
        arg[3,1,1]=(0.0416243959618)*x[0]**o+(0.739593496294)*x[0]+(-0.0215969068381)*x[1]**o+(0.203954403681)*x[1]
        arg[4,0,0]=(-0.479380091881)*x[0]**o+(0.973630227471)*x[0]+(0.372617682727)*x[1]**o+(0.721993565094)*x[1]
        arg[4,0,1]=(0.360211306746)*x[0]**o+(0.260993313325)*x[0]+(0.252635752916)*x[1]**o+(0.534875843061)*x[1]
        arg[4,1,0]=(-0.609695114069)*x[0]**o+(0.283865161572)*x[0]+(-0.607441168016)*x[1]**o+(-0.838279309054)*x[1]
        arg[4,1,1]=(0.984921561958)*x[0]**o+(-0.352255390774)*x[0]+(-0.770706841576)*x[1]**o+(0.456415771191)*x[1]
        arg[5,0,0]=(0.903410389973)*x[0]**o+(0.609651342058)*x[0]+(-0.0890705788805)*x[1]**o+(0.613493629434)*x[1]
        arg[5,0,1]=(-0.977979577058)*x[0]**o+(-0.121665631763)*x[0]+(0.466897174395)*x[1]**o+(-0.438819940752)*x[1]
        arg[5,1,0]=(-0.537313389905)*x[0]**o+(-0.0401503571218)*x[0]+(-0.454563705252)*x[1]**o+(-0.728276295145)*x[1]
        arg[5,1,1]=(0.858855332454)*x[0]**o+(0.378938106473)*x[0]+(0.908957690039)*x[1]**o+(0.989040980713)*x[1]
        ref[0,0,0]=(-0.906705618142)*x_ref[0]**o+(0.458982439103)*x_ref[0]+(0.556464353903)*x_ref[1]**o+(0.733236686234)*x_ref[1]
        ref[0,0,1]=(-0.192889954523)*x_ref[0]**o+(-0.814332539133)*x_ref[0]+(0.446164214282)*x_ref[1]**o+(0.032291315446)*x_ref[1]
        ref[0,1,0]=(0.757081929199)*x_ref[0]**o+(-0.3938524931)*x_ref[0]+(-0.458060132115)*x_ref[1]**o+(0.471831491182)*x_ref[1]
        ref[0,1,1]=(0.459868785407)*x_ref[0]**o+(0.513707183316)*x_ref[0]+(-0.9682413377)*x_ref[1]**o+(-0.49496265187)*x_ref[1]
        ref[1,0,0]=(0.546655212697)*x_ref[0]**o+(-0.296125788768)*x_ref[0]+(0.570033305927)*x_ref[1]**o+(-0.818919094317)*x_ref[1]
        ref[1,0,1]=(-0.954422252892)*x_ref[0]**o+(-0.925035286684)*x_ref[0]+(0.171311127546)*x_ref[1]**o+(-0.584606597796)*x_ref[1]
        ref[1,1,0]=(-0.203956024883)*x_ref[0]**o+(-0.933117428128)*x_ref[0]+(-0.745101180472)*x_ref[1]**o+(0.0390208509919)*x_ref[1]
        ref[1,1,1]=(-0.67467593628)*x_ref[0]**o+(0.42281003355)*x_ref[0]+(-0.495401176467)*x_ref[1]**o+(0.993257306842)*x_ref[1]
        ref[2,0,0]=(-0.204011890956)*x_ref[0]**o+(0.730452092671)*x_ref[0]+(-0.0486718200506)*x_ref[1]**o+(0.691928726795)*x_ref[1]
        ref[2,0,1]=(0.381571380631)*x_ref[0]**o+(0.504896621541)*x_ref[0]+(-0.229805983176)*x_ref[1]**o+(0.281401365705)*x_ref[1]
        ref[2,1,0]=(0.578852751326)*x_ref[0]**o+(-0.38588684985)*x_ref[0]+(-0.630908284751)*x_ref[1]**o+(0.237972319424)*x_ref[1]
        ref[2,1,1]=(-0.928366568501)*x_ref[0]**o+(0.254484026833)*x_ref[0]+(0.903931940538)*x_ref[1]**o+(-0.370818194785)*x_ref[1]
        ref[3,0,0]=(-0.216012465475)*x_ref[0]**o+(-0.938274177001)*x_ref[0]+(0.0674408710328)*x_ref[1]**o+(-0.731594784098)*x_ref[1]
        ref[3,0,1]=(-0.156790905207)*x_ref[0]**o+(0.284403081253)*x_ref[0]+(0.3754228052)*x_ref[1]**o+(-0.772758808958)*x_ref[1]
        ref[3,1,0]=(0.89987768188)*x_ref[0]**o+(0.0177886564195)*x_ref[0]+(0.102329016274)*x_ref[1]**o+(0.474039949682)*x_ref[1]
        ref[3,1,1]=(0.0416243959618)*x_ref[0]**o+(0.739593496294)*x_ref[0]+(-0.0215969068381)*x_ref[1]**o+(0.203954403681)*x_ref[1]
        ref[4,0,0]=(-0.479380091881)*x_ref[0]**o+(0.973630227471)*x_ref[0]+(0.372617682727)*x_ref[1]**o+(0.721993565094)*x_ref[1]
        ref[4,0,1]=(0.360211306746)*x_ref[0]**o+(0.260993313325)*x_ref[0]+(0.252635752916)*x_ref[1]**o+(0.534875843061)*x_ref[1]
        ref[4,1,0]=(-0.609695114069)*x_ref[0]**o+(0.283865161572)*x_ref[0]+(-0.607441168016)*x_ref[1]**o+(-0.838279309054)*x_ref[1]
        ref[4,1,1]=(0.984921561958)*x_ref[0]**o+(-0.352255390774)*x_ref[0]+(-0.770706841576)*x_ref[1]**o+(0.456415771191)*x_ref[1]
        ref[5,0,0]=(0.903410389973)*x_ref[0]**o+(0.609651342058)*x_ref[0]+(-0.0890705788805)*x_ref[1]**o+(0.613493629434)*x_ref[1]
        ref[5,0,1]=(-0.977979577058)*x_ref[0]**o+(-0.121665631763)*x_ref[0]+(0.466897174395)*x_ref[1]**o+(-0.438819940752)*x_ref[1]
        ref[5,1,0]=(-0.537313389905)*x_ref[0]**o+(-0.0401503571218)*x_ref[0]+(-0.454563705252)*x_ref[1]**o+(-0.728276295145)*x_ref[1]
        ref[5,1,1]=(0.858855332454)*x_ref[0]**o+(0.378938106473)*x_ref[0]+(0.908957690039)*x_ref[1]**o+(0.989040980713)*x_ref[1]
      else:
        arg[0,0,0]=(-0.550852990025)*x[0]**o+(0.578705971531)*x[0]+(0.9533401901)*x[1]**o+(0.233181910221)*x[1]+(-0.143544272432)*x[2]**o+(-0.986513253736)*x[2]
        arg[0,0,1]=(-0.326417636931)*x[0]**o+(0.0910236969351)*x[0]+(-0.8240642668)*x[1]**o+(-0.242359389714)*x[1]+(0.266813299647)*x[2]**o+(-0.364956839435)*x[2]
        arg[0,1,0]=(-0.777107068999)*x[0]**o+(0.201120303703)*x[0]+(-0.684906707463)*x[1]**o+(-0.9620762843)*x[1]+(-0.314789913974)*x[2]**o+(-0.70682592319)*x[2]
        arg[0,1,1]=(-0.980292769796)*x[0]**o+(-0.827763529178)*x[0]+(0.256770279946)*x[1]**o+(0.619665292369)*x[1]+(0.014752087873)*x[2]**o+(-0.810044916486)*x[2]
        arg[1,0,0]=(-0.51164541516)*x[0]**o+(-0.429646605692)*x[0]+(-0.844516540906)*x[1]**o+(-0.0334629941654)*x[1]+(0.942164301071)*x[2]**o+(0.129008643344)*x[2]
        arg[1,0,1]=(-0.545329352463)*x[0]**o+(-0.0601985885543)*x[0]+(0.959752895562)*x[1]**o+(-0.617790988468)*x[1]+(-0.352323774737)*x[2]**o+(0.0854549755648)*x[2]
        arg[1,1,0]=(0.163270290732)*x[0]**o+(0.715418689409)*x[0]+(-0.892736094727)*x[1]**o+(-0.250085969902)*x[1]+(-0.323686607391)*x[2]**o+(-0.808822044566)*x[2]
        arg[1,1,1]=(-0.582858228848)*x[0]**o+(0.245356494646)*x[0]+(0.253867860186)*x[1]**o+(0.294533212339)*x[1]+(0.802766375942)*x[2]**o+(-0.142246487978)*x[2]
        arg[2,0,0]=(0.559682931749)*x[0]**o+(0.349781768104)*x[0]+(-0.903942828138)*x[1]**o+(0.285790428402)*x[1]+(0.334765408827)*x[2]**o+(-0.319553167552)*x[2]
        arg[2,0,1]=(-0.33958138616)*x[0]**o+(-0.476356251551)*x[0]+(0.222685919591)*x[1]**o+(0.947677379274)*x[1]+(-0.238198083505)*x[2]**o+(-0.158835521934)*x[2]
        arg[2,1,0]=(0.233005925554)*x[0]**o+(-0.0885827512025)*x[0]+(0.99298235997)*x[1]**o+(-0.972864413346)*x[1]+(-0.328337965331)*x[2]**o+(0.548449226427)*x[2]
        arg[2,1,1]=(-0.682434615287)*x[0]**o+(-0.681560497743)*x[0]+(-0.0427998840708)*x[1]**o+(-0.640532507401)*x[1]+(0.0600572641903)*x[2]**o+(0.580127097835)*x[2]
        arg[3,0,0]=(0.627556962303)*x[0]**o+(-0.2282646452)*x[0]+(0.201159933134)*x[1]**o+(-0.0874921144181)*x[1]+(0.487026890349)*x[2]**o+(-0.044256665504)*x[2]
        arg[3,0,1]=(-0.202969525043)*x[0]**o+(0.0445621391034)*x[0]+(-0.763479658456)*x[1]**o+(-0.600818445386)*x[1]+(-0.442800171339)*x[2]**o+(0.0422443400421)*x[2]
        arg[3,1,0]=(-0.422348404642)*x[0]**o+(0.723277196982)*x[0]+(0.882658191301)*x[1]**o+(-0.415218341669)*x[1]+(0.834911332561)*x[2]**o+(-0.843638850779)*x[2]
        arg[3,1,1]=(-0.685037113836)*x[0]**o+(0.119526696821)*x[0]+(0.732784440424)*x[1]**o+(0.156993935952)*x[1]+(-0.565837827825)*x[2]**o+(0.667635603478)*x[2]
        arg[4,0,0]=(-0.18210082527)*x[0]**o+(-0.00639006601355)*x[0]+(0.633771301081)*x[1]**o+(0.765450011643)*x[1]+(-0.323701472535)*x[2]**o+(0.841246809469)*x[2]
        arg[4,0,1]=(0.345847635107)*x[0]**o+(0.910917657816)*x[0]+(0.329667803527)*x[1]**o+(-0.0638797846725)*x[1]+(-0.705758696202)*x[2]**o+(0.39567870447)*x[2]
        arg[4,1,0]=(0.700805557734)*x[0]**o+(0.215376719066)*x[0]+(0.057212023986)*x[1]**o+(-0.601972647211)*x[1]+(-0.0163323501628)*x[2]**o+(-0.182314452762)*x[2]
        arg[4,1,1]=(-0.25633784052)*x[0]**o+(-0.36128209808)*x[0]+(-0.580660119759)*x[1]**o+(-0.718361413439)*x[1]+(0.294752937469)*x[2]**o+(0.479425008353)*x[2]
        arg[5,0,0]=(0.0250768411622)*x[0]**o+(-0.954437161073)*x[0]+(0.314623130846)*x[1]**o+(-0.915911802082)*x[1]+(-0.293481557155)*x[2]**o+(0.193040417512)*x[2]
        arg[5,0,1]=(-0.260552760825)*x[0]**o+(-0.993318746791)*x[0]+(-0.0888187740235)*x[1]**o+(0.228899511722)*x[1]+(0.667467747688)*x[2]**o+(0.210051895803)*x[2]
        arg[5,1,0]=(-0.925551425372)*x[0]**o+(-0.051063099903)*x[0]+(0.0607054862843)*x[1]**o+(0.778043999849)*x[1]+(-0.393429559505)*x[2]**o+(0.235480458425)*x[2]
        arg[5,1,1]=(0.00188621198193)*x[0]**o+(-0.529444982511)*x[0]+(0.766032106394)*x[1]**o+(-0.554844276584)*x[1]+(0.793956883291)*x[2]**o+(0.221253426455)*x[2]
        ref[0,0,0]=(-0.550852990025)*x_ref[0]**o+(0.578705971531)*x_ref[0]+(0.9533401901)*x_ref[1]**o+(0.233181910221)*x_ref[1]+(-0.143544272432)*x_ref[2]**o+(-0.986513253736)*x_ref[2]
        ref[0,0,1]=(-0.326417636931)*x_ref[0]**o+(0.0910236969351)*x_ref[0]+(-0.8240642668)*x_ref[1]**o+(-0.242359389714)*x_ref[1]+(0.266813299647)*x_ref[2]**o+(-0.364956839435)*x_ref[2]
        ref[0,1,0]=(-0.777107068999)*x_ref[0]**o+(0.201120303703)*x_ref[0]+(-0.684906707463)*x_ref[1]**o+(-0.9620762843)*x_ref[1]+(-0.314789913974)*x_ref[2]**o+(-0.70682592319)*x_ref[2]
        ref[0,1,1]=(-0.980292769796)*x_ref[0]**o+(-0.827763529178)*x_ref[0]+(0.256770279946)*x_ref[1]**o+(0.619665292369)*x_ref[1]+(0.014752087873)*x_ref[2]**o+(-0.810044916486)*x_ref[2]
        ref[1,0,0]=(-0.51164541516)*x_ref[0]**o+(-0.429646605692)*x_ref[0]+(-0.844516540906)*x_ref[1]**o+(-0.0334629941654)*x_ref[1]+(0.942164301071)*x_ref[2]**o+(0.129008643344)*x_ref[2]
        ref[1,0,1]=(-0.545329352463)*x_ref[0]**o+(-0.0601985885543)*x_ref[0]+(0.959752895562)*x_ref[1]**o+(-0.617790988468)*x_ref[1]+(-0.352323774737)*x_ref[2]**o+(0.0854549755648)*x_ref[2]
        ref[1,1,0]=(0.163270290732)*x_ref[0]**o+(0.715418689409)*x_ref[0]+(-0.892736094727)*x_ref[1]**o+(-0.250085969902)*x_ref[1]+(-0.323686607391)*x_ref[2]**o+(-0.808822044566)*x_ref[2]
        ref[1,1,1]=(-0.582858228848)*x_ref[0]**o+(0.245356494646)*x_ref[0]+(0.253867860186)*x_ref[1]**o+(0.294533212339)*x_ref[1]+(0.802766375942)*x_ref[2]**o+(-0.142246487978)*x_ref[2]
        ref[2,0,0]=(0.559682931749)*x_ref[0]**o+(0.349781768104)*x_ref[0]+(-0.903942828138)*x_ref[1]**o+(0.285790428402)*x_ref[1]+(0.334765408827)*x_ref[2]**o+(-0.319553167552)*x_ref[2]
        ref[2,0,1]=(-0.33958138616)*x_ref[0]**o+(-0.476356251551)*x_ref[0]+(0.222685919591)*x_ref[1]**o+(0.947677379274)*x_ref[1]+(-0.238198083505)*x_ref[2]**o+(-0.158835521934)*x_ref[2]
        ref[2,1,0]=(0.233005925554)*x_ref[0]**o+(-0.0885827512025)*x_ref[0]+(0.99298235997)*x_ref[1]**o+(-0.972864413346)*x_ref[1]+(-0.328337965331)*x_ref[2]**o+(0.548449226427)*x_ref[2]
        ref[2,1,1]=(-0.682434615287)*x_ref[0]**o+(-0.681560497743)*x_ref[0]+(-0.0427998840708)*x_ref[1]**o+(-0.640532507401)*x_ref[1]+(0.0600572641903)*x_ref[2]**o+(0.580127097835)*x_ref[2]
        ref[3,0,0]=(0.627556962303)*x_ref[0]**o+(-0.2282646452)*x_ref[0]+(0.201159933134)*x_ref[1]**o+(-0.0874921144181)*x_ref[1]+(0.487026890349)*x_ref[2]**o+(-0.044256665504)*x_ref[2]
        ref[3,0,1]=(-0.202969525043)*x_ref[0]**o+(0.0445621391034)*x_ref[0]+(-0.763479658456)*x_ref[1]**o+(-0.600818445386)*x_ref[1]+(-0.442800171339)*x_ref[2]**o+(0.0422443400421)*x_ref[2]
        ref[3,1,0]=(-0.422348404642)*x_ref[0]**o+(0.723277196982)*x_ref[0]+(0.882658191301)*x_ref[1]**o+(-0.415218341669)*x_ref[1]+(0.834911332561)*x_ref[2]**o+(-0.843638850779)*x_ref[2]
        ref[3,1,1]=(-0.685037113836)*x_ref[0]**o+(0.119526696821)*x_ref[0]+(0.732784440424)*x_ref[1]**o+(0.156993935952)*x_ref[1]+(-0.565837827825)*x_ref[2]**o+(0.667635603478)*x_ref[2]
        ref[4,0,0]=(-0.18210082527)*x_ref[0]**o+(-0.00639006601355)*x_ref[0]+(0.633771301081)*x_ref[1]**o+(0.765450011643)*x_ref[1]+(-0.323701472535)*x_ref[2]**o+(0.841246809469)*x_ref[2]
        ref[4,0,1]=(0.345847635107)*x_ref[0]**o+(0.910917657816)*x_ref[0]+(0.329667803527)*x_ref[1]**o+(-0.0638797846725)*x_ref[1]+(-0.705758696202)*x_ref[2]**o+(0.39567870447)*x_ref[2]
        ref[4,1,0]=(0.700805557734)*x_ref[0]**o+(0.215376719066)*x_ref[0]+(0.057212023986)*x_ref[1]**o+(-0.601972647211)*x_ref[1]+(-0.0163323501628)*x_ref[2]**o+(-0.182314452762)*x_ref[2]
        ref[4,1,1]=(-0.25633784052)*x_ref[0]**o+(-0.36128209808)*x_ref[0]+(-0.580660119759)*x_ref[1]**o+(-0.718361413439)*x_ref[1]+(0.294752937469)*x_ref[2]**o+(0.479425008353)*x_ref[2]
        ref[5,0,0]=(0.0250768411622)*x_ref[0]**o+(-0.954437161073)*x_ref[0]+(0.314623130846)*x_ref[1]**o+(-0.915911802082)*x_ref[1]+(-0.293481557155)*x_ref[2]**o+(0.193040417512)*x_ref[2]
        ref[5,0,1]=(-0.260552760825)*x_ref[0]**o+(-0.993318746791)*x_ref[0]+(-0.0888187740235)*x_ref[1]**o+(0.228899511722)*x_ref[1]+(0.667467747688)*x_ref[2]**o+(0.210051895803)*x_ref[2]
        ref[5,1,0]=(-0.925551425372)*x_ref[0]**o+(-0.051063099903)*x_ref[0]+(0.0607054862843)*x_ref[1]**o+(0.778043999849)*x_ref[1]+(-0.393429559505)*x_ref[2]**o+(0.235480458425)*x_ref[2]
        ref[5,1,1]=(0.00188621198193)*x_ref[0]**o+(-0.529444982511)*x_ref[0]+(0.766032106394)*x_ref[1]**o+(-0.554844276584)*x_ref[1]+(0.793956883291)*x_ref[2]**o+(0.221253426455)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactZero_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactZero onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.358768976574)*x[0]**o+(-0.231637053165)*x[0]+(0.798891506345)*x[1]**o+(0.292325388353)*x[1]
        arg[0,0,0,1]=(0.53301689922)*x[0]**o+(-0.761786870664)*x[0]+(0.284067900826)*x[1]**o+(-0.688662082205)*x[1]
        arg[0,0,1,0]=(-0.721988036113)*x[0]**o+(-0.743892757815)*x[0]+(-0.796787022375)*x[1]**o+(-0.499468412523)*x[1]
        arg[0,0,1,1]=(0.0710691488306)*x[0]**o+(0.0210385141456)*x[0]+(-0.709342417683)*x[1]**o+(0.901794071115)*x[1]
        arg[0,0,2,0]=(-0.800537263339)*x[0]**o+(0.583114786539)*x[0]+(0.422752622522)*x[1]**o+(-0.0954981460748)*x[1]
        arg[0,0,2,1]=(-0.0805357065445)*x[0]**o+(0.476377895547)*x[0]+(0.242203912566)*x[1]**o+(-0.479971540043)*x[1]
        arg[0,1,0,0]=(-0.476356823162)*x[0]**o+(-0.438853622252)*x[0]+(-0.937228141746)*x[1]**o+(0.366496742588)*x[1]
        arg[0,1,0,1]=(-0.496918253956)*x[0]**o+(-0.0227549061268)*x[0]+(0.830195584768)*x[1]**o+(-0.114624902423)*x[1]
        arg[0,1,1,0]=(0.431291036997)*x[0]**o+(0.807713720258)*x[0]+(0.730057543333)*x[1]**o+(0.635171725232)*x[1]
        arg[0,1,1,1]=(0.784425995411)*x[0]**o+(-0.397510494899)*x[0]+(-0.579911740115)*x[1]**o+(0.130426014607)*x[1]
        arg[0,1,2,0]=(0.473613650641)*x[0]**o+(-0.155617741177)*x[0]+(0.873237852103)*x[1]**o+(-0.749395152414)*x[1]
        arg[0,1,2,1]=(0.367649559431)*x[0]**o+(0.830181309644)*x[0]+(0.87726968147)*x[1]**o+(-0.450304793682)*x[1]
        arg[0,2,0,0]=(-0.700522876654)*x[0]**o+(-0.998187126003)*x[0]+(0.732491610173)*x[1]**o+(0.648679267)*x[1]
        arg[0,2,0,1]=(0.499746141285)*x[0]**o+(-0.730090975806)*x[0]+(-0.525566438654)*x[1]**o+(0.212469850421)*x[1]
        arg[0,2,1,0]=(-0.847752916298)*x[0]**o+(0.43510472619)*x[0]+(0.815305936548)*x[1]**o+(-0.143802676552)*x[1]
        arg[0,2,1,1]=(-0.453572376958)*x[0]**o+(-0.913023531929)*x[0]+(0.743328127665)*x[1]**o+(0.152586122591)*x[1]
        arg[0,2,2,0]=(0.526939334508)*x[0]**o+(0.247268722057)*x[0]+(-0.176432621481)*x[1]**o+(-0.510110202196)*x[1]
        arg[0,2,2,1]=(0.139685732003)*x[0]**o+(0.278304455867)*x[0]+(0.517953769616)*x[1]**o+(-0.272384261701)*x[1]
        arg[0,3,0,0]=(-0.38867907688)*x[0]**o+(-0.73043975055)*x[0]+(-0.14293243901)*x[1]**o+(0.0601042854043)*x[1]
        arg[0,3,0,1]=(-0.163055949188)*x[0]**o+(-0.035877062482)*x[0]+(0.0396615423106)*x[1]**o+(-0.870062303275)*x[1]
        arg[0,3,1,0]=(0.283839779914)*x[0]**o+(-0.519316925197)*x[0]+(-0.437401973959)*x[1]**o+(-0.823707232663)*x[1]
        arg[0,3,1,1]=(0.700305425892)*x[0]**o+(0.0299761521094)*x[0]+(-0.993032151025)*x[1]**o+(0.633293413221)*x[1]
        arg[0,3,2,0]=(-0.745357256365)*x[0]**o+(-0.451553456961)*x[0]+(0.05789646302)*x[1]**o+(-0.755901155251)*x[1]
        arg[0,3,2,1]=(-0.57330840403)*x[0]**o+(-0.589361483751)*x[0]+(-0.038677123513)*x[1]**o+(0.417332032168)*x[1]
        arg[0,4,0,0]=(0.759655161531)*x[0]**o+(0.737483461328)*x[0]+(0.573158629937)*x[1]**o+(-0.454191109106)*x[1]
        arg[0,4,0,1]=(-0.232982406645)*x[0]**o+(-0.926030038179)*x[0]+(-0.504256834465)*x[1]**o+(-0.107703527058)*x[1]
        arg[0,4,1,0]=(0.898411697016)*x[0]**o+(0.642045536742)*x[0]+(-0.953059791895)*x[1]**o+(-0.67501710093)*x[1]
        arg[0,4,1,1]=(0.300380574165)*x[0]**o+(0.204591960115)*x[0]+(-0.308849697562)*x[1]**o+(-0.489495887357)*x[1]
        arg[0,4,2,0]=(-0.814560286716)*x[0]**o+(0.154235599953)*x[0]+(0.279329273498)*x[1]**o+(-0.674063978926)*x[1]
        arg[0,4,2,1]=(-0.718018870365)*x[0]**o+(-0.735982858424)*x[0]+(-0.0132197132926)*x[1]**o+(0.398939587543)*x[1]
        arg[1,0,0,0]=(0.0559995351725)*x[0]**o+(-0.164549619892)*x[0]+(-0.794242529408)*x[1]**o+(0.0927822314785)*x[1]
        arg[1,0,0,1]=(-0.894238542341)*x[0]**o+(0.378559247473)*x[0]+(-0.183957272639)*x[1]**o+(0.486358968942)*x[1]
        arg[1,0,1,0]=(0.931288801001)*x[0]**o+(0.223815769968)*x[0]+(-0.719591906328)*x[1]**o+(-0.743644741105)*x[1]
        arg[1,0,1,1]=(0.156646029123)*x[0]**o+(0.306942879419)*x[0]+(0.747006830674)*x[1]**o+(-0.876534912274)*x[1]
        arg[1,0,2,0]=(-0.496819081999)*x[0]**o+(-0.939227402425)*x[0]+(0.0566331712182)*x[1]**o+(0.565724364661)*x[1]
        arg[1,0,2,1]=(0.903706429954)*x[0]**o+(0.667947320307)*x[0]+(0.97140855373)*x[1]**o+(0.924095235752)*x[1]
        arg[1,1,0,0]=(0.526180197299)*x[0]**o+(0.435493351674)*x[0]+(0.400107077848)*x[1]**o+(-0.704760833758)*x[1]
        arg[1,1,0,1]=(0.0723740015004)*x[0]**o+(-0.405859051371)*x[0]+(0.200787757511)*x[1]**o+(0.41819184563)*x[1]
        arg[1,1,1,0]=(-0.0320951655627)*x[0]**o+(0.235068445821)*x[0]+(0.450383349599)*x[1]**o+(-0.801487424765)*x[1]
        arg[1,1,1,1]=(-0.52478892744)*x[0]**o+(-0.449111297125)*x[0]+(0.731062182014)*x[1]**o+(0.533619608577)*x[1]
        arg[1,1,2,0]=(0.0889794057057)*x[0]**o+(-0.971856981571)*x[0]+(0.662216408922)*x[1]**o+(0.164089024468)*x[1]
        arg[1,1,2,1]=(-0.196374963647)*x[0]**o+(-0.739360211802)*x[0]+(0.916611650753)*x[1]**o+(-0.260102318115)*x[1]
        arg[1,2,0,0]=(0.627763572103)*x[0]**o+(-0.893900847319)*x[0]+(0.907548446417)*x[1]**o+(0.0687906203663)*x[1]
        arg[1,2,0,1]=(0.59755498682)*x[0]**o+(-0.457108009787)*x[0]+(0.71562139654)*x[1]**o+(-0.300872832559)*x[1]
        arg[1,2,1,0]=(-0.479641829834)*x[0]**o+(0.297188513722)*x[0]+(0.0227730978493)*x[1]**o+(0.199329665406)*x[1]
        arg[1,2,1,1]=(0.918974940101)*x[0]**o+(0.595194981816)*x[0]+(0.760030152554)*x[1]**o+(-0.798899878486)*x[1]
        arg[1,2,2,0]=(-0.878334511842)*x[0]**o+(0.910857236287)*x[0]+(0.0164747567997)*x[1]**o+(-0.876970888445)*x[1]
        arg[1,2,2,1]=(0.29445809181)*x[0]**o+(0.837849454931)*x[0]+(0.0759054371211)*x[1]**o+(0.501243080372)*x[1]
        arg[1,3,0,0]=(-0.529006436258)*x[0]**o+(-0.932325855538)*x[0]+(-0.561472972374)*x[1]**o+(0.845125736656)*x[1]
        arg[1,3,0,1]=(-0.843008257488)*x[0]**o+(0.614118623012)*x[0]+(-0.167015137447)*x[1]**o+(0.158701952625)*x[1]
        arg[1,3,1,0]=(-0.662966916148)*x[0]**o+(-0.443297576512)*x[0]+(0.295988608077)*x[1]**o+(0.059506908615)*x[1]
        arg[1,3,1,1]=(0.864199978014)*x[0]**o+(0.138723519602)*x[0]+(0.144964839302)*x[1]**o+(0.239863709389)*x[1]
        arg[1,3,2,0]=(0.479006097431)*x[0]**o+(0.542168797821)*x[0]+(0.561074562163)*x[1]**o+(-0.133365787275)*x[1]
        arg[1,3,2,1]=(-0.901033449452)*x[0]**o+(-0.620245150074)*x[0]+(0.598461569697)*x[1]**o+(-0.2213246854)*x[1]
        arg[1,4,0,0]=(-0.577067401348)*x[0]**o+(0.486154324874)*x[0]+(0.809205150442)*x[1]**o+(-0.218206418459)*x[1]
        arg[1,4,0,1]=(0.726556862066)*x[0]**o+(0.0293695509867)*x[0]+(0.601938348193)*x[1]**o+(0.967332697531)*x[1]
        arg[1,4,1,0]=(-0.675624388054)*x[0]**o+(-0.871613723214)*x[0]+(0.605519387693)*x[1]**o+(0.287186894411)*x[1]
        arg[1,4,1,1]=(-0.689904898214)*x[0]**o+(-0.918236705665)*x[0]+(-0.892168622177)*x[1]**o+(0.507441662612)*x[1]
        arg[1,4,2,0]=(0.845768980712)*x[0]**o+(-0.458467508114)*x[0]+(-0.518185249391)*x[1]**o+(0.829945330327)*x[1]
        arg[1,4,2,1]=(0.776852312178)*x[0]**o+(-0.268810999469)*x[0]+(-0.576757148841)*x[1]**o+(-0.50628294802)*x[1]
        arg[2,0,0,0]=(-0.686840219939)*x[0]**o+(0.6861172847)*x[0]+(0.240362263052)*x[1]**o+(0.962361523884)*x[1]
        arg[2,0,0,1]=(-0.943837188119)*x[0]**o+(0.829832768268)*x[0]+(0.377590124133)*x[1]**o+(-0.171714144962)*x[1]
        arg[2,0,1,0]=(0.862809148447)*x[0]**o+(0.0573402136489)*x[0]+(-0.743311826013)*x[1]**o+(0.427809313437)*x[1]
        arg[2,0,1,1]=(0.377380819576)*x[0]**o+(-0.223280285041)*x[0]+(0.178628658562)*x[1]**o+(0.121964611877)*x[1]
        arg[2,0,2,0]=(0.41723758689)*x[0]**o+(0.256039203883)*x[0]+(0.255358776676)*x[1]**o+(-0.881665964974)*x[1]
        arg[2,0,2,1]=(-0.0797103024499)*x[0]**o+(0.128068931414)*x[0]+(-0.758317834409)*x[1]**o+(0.654657809183)*x[1]
        arg[2,1,0,0]=(0.603648218133)*x[0]**o+(-0.286780270729)*x[0]+(-0.341108067214)*x[1]**o+(-0.769667356087)*x[1]
        arg[2,1,0,1]=(0.844044270035)*x[0]**o+(0.0585907857764)*x[0]+(-0.578232785252)*x[1]**o+(-0.40959017859)*x[1]
        arg[2,1,1,0]=(0.599709567393)*x[0]**o+(0.307092449739)*x[0]+(-0.402950586215)*x[1]**o+(0.0674239470451)*x[1]
        arg[2,1,1,1]=(0.182608249961)*x[0]**o+(0.373533584601)*x[0]+(0.40949869082)*x[1]**o+(0.696333617699)*x[1]
        arg[2,1,2,0]=(0.0462891111054)*x[0]**o+(0.371258293292)*x[0]+(0.782976187716)*x[1]**o+(0.117976017999)*x[1]
        arg[2,1,2,1]=(0.891841121526)*x[0]**o+(-0.998971173204)*x[0]+(0.0286712507928)*x[1]**o+(-0.0198507022516)*x[1]
        arg[2,2,0,0]=(0.284620902703)*x[0]**o+(0.936231203073)*x[0]+(0.829209752046)*x[1]**o+(-0.94940982125)*x[1]
        arg[2,2,0,1]=(-0.647956027376)*x[0]**o+(0.923690718922)*x[0]+(-0.0322423837346)*x[1]**o+(0.124962508606)*x[1]
        arg[2,2,1,0]=(-0.0696010411705)*x[0]**o+(-0.161672936739)*x[0]+(-0.997445762186)*x[1]**o+(0.177962302014)*x[1]
        arg[2,2,1,1]=(-0.897657700296)*x[0]**o+(-0.467866721175)*x[0]+(-0.560034332007)*x[1]**o+(-0.861509898295)*x[1]
        arg[2,2,2,0]=(-0.100880174048)*x[0]**o+(0.678568254695)*x[0]+(-0.83289411684)*x[1]**o+(0.680564393149)*x[1]
        arg[2,2,2,1]=(0.546159181491)*x[0]**o+(0.720296622966)*x[0]+(0.939046941118)*x[1]**o+(-0.716147554733)*x[1]
        arg[2,3,0,0]=(-0.921073162235)*x[0]**o+(0.979925882289)*x[0]+(-0.0938642966949)*x[1]**o+(0.482093047223)*x[1]
        arg[2,3,0,1]=(0.0709940956822)*x[0]**o+(-0.691401017243)*x[0]+(0.447000814029)*x[1]**o+(0.0575856240533)*x[1]
        arg[2,3,1,0]=(0.562800393365)*x[0]**o+(0.657646549915)*x[0]+(-0.835326361689)*x[1]**o+(0.3487960687)*x[1]
        arg[2,3,1,1]=(0.106668867409)*x[0]**o+(0.436327240765)*x[0]+(-0.585051261259)*x[1]**o+(0.616655338667)*x[1]
        arg[2,3,2,0]=(-0.501849628176)*x[0]**o+(-0.284722503916)*x[0]+(-0.892470199837)*x[1]**o+(-0.979942541795)*x[1]
        arg[2,3,2,1]=(0.236759540073)*x[0]**o+(-0.806931113769)*x[0]+(-0.225220003475)*x[1]**o+(-0.95183239844)*x[1]
        arg[2,4,0,0]=(-0.684823909581)*x[0]**o+(0.445615228059)*x[0]+(-0.638879895767)*x[1]**o+(-0.335505874088)*x[1]
        arg[2,4,0,1]=(-0.196708460612)*x[0]**o+(0.981592996809)*x[0]+(0.435774317028)*x[1]**o+(-0.0952200663109)*x[1]
        arg[2,4,1,0]=(-0.552759474764)*x[0]**o+(-0.614031173545)*x[0]+(0.304375542871)*x[1]**o+(-0.285131901533)*x[1]
        arg[2,4,1,1]=(0.147812951421)*x[0]**o+(-0.78533487681)*x[0]+(0.942205524401)*x[1]**o+(0.166280215075)*x[1]
        arg[2,4,2,0]=(0.843067982563)*x[0]**o+(0.0116520710289)*x[0]+(0.927948889059)*x[1]**o+(-0.395251919577)*x[1]
        arg[2,4,2,1]=(-0.316538010007)*x[0]**o+(-0.139262497126)*x[0]+(0.645409249217)*x[1]**o+(0.192949521992)*x[1]
        arg[3,0,0,0]=(0.88179290666)*x[0]**o+(0.337853287365)*x[0]+(-0.590219587193)*x[1]**o+(-0.827051504196)*x[1]
        arg[3,0,0,1]=(-0.104526336997)*x[0]**o+(0.299097012282)*x[0]+(-0.348493995853)*x[1]**o+(-0.780034322304)*x[1]
        arg[3,0,1,0]=(0.358724159445)*x[0]**o+(-0.0338089482488)*x[0]+(-0.0423054506456)*x[1]**o+(0.67180489557)*x[1]
        arg[3,0,1,1]=(-0.778880095665)*x[0]**o+(-0.448812173928)*x[0]+(-0.823686122382)*x[1]**o+(0.636028922656)*x[1]
        arg[3,0,2,0]=(0.146891002069)*x[0]**o+(-0.368903199628)*x[0]+(-0.596272077557)*x[1]**o+(0.703510683225)*x[1]
        arg[3,0,2,1]=(0.357416423547)*x[0]**o+(-0.103754905142)*x[0]+(0.32026011189)*x[1]**o+(0.755648597627)*x[1]
        arg[3,1,0,0]=(0.399082464238)*x[0]**o+(-0.12611921797)*x[0]+(0.426552225948)*x[1]**o+(-0.228714297286)*x[1]
        arg[3,1,0,1]=(-0.638411026717)*x[0]**o+(-0.342143922306)*x[0]+(-0.502107802626)*x[1]**o+(-0.881045304848)*x[1]
        arg[3,1,1,0]=(0.411134247295)*x[0]**o+(-0.216585727086)*x[0]+(0.270013339156)*x[1]**o+(0.251581243494)*x[1]
        arg[3,1,1,1]=(0.788758848294)*x[0]**o+(0.345181621902)*x[0]+(0.102843710598)*x[1]**o+(0.0561071077516)*x[1]
        arg[3,1,2,0]=(0.232441162347)*x[0]**o+(-0.687544250386)*x[0]+(0.369065746372)*x[1]**o+(0.628109564412)*x[1]
        arg[3,1,2,1]=(-0.912295044312)*x[0]**o+(0.847830497468)*x[0]+(0.776648618003)*x[1]**o+(0.522638388084)*x[1]
        arg[3,2,0,0]=(0.779333120942)*x[0]**o+(-0.0747119967046)*x[0]+(-0.594363282198)*x[1]**o+(0.513922159444)*x[1]
        arg[3,2,0,1]=(0.877492427404)*x[0]**o+(-0.191727825399)*x[0]+(0.0881296554818)*x[1]**o+(-0.408407503067)*x[1]
        arg[3,2,1,0]=(0.29696771215)*x[0]**o+(-0.914882727138)*x[0]+(0.684156831385)*x[1]**o+(0.142293411327)*x[1]
        arg[3,2,1,1]=(0.105927449304)*x[0]**o+(-0.863339661598)*x[0]+(-0.815550662607)*x[1]**o+(-0.261034715274)*x[1]
        arg[3,2,2,0]=(0.944082529871)*x[0]**o+(-0.449259627627)*x[0]+(-0.560526159016)*x[1]**o+(0.609050500798)*x[1]
        arg[3,2,2,1]=(0.529508870267)*x[0]**o+(0.627216712239)*x[0]+(-0.0148122228382)*x[1]**o+(0.0743274653452)*x[1]
        arg[3,3,0,0]=(-0.30609697448)*x[0]**o+(0.910972095956)*x[0]+(0.256274444924)*x[1]**o+(0.229096821635)*x[1]
        arg[3,3,0,1]=(-0.510323627078)*x[0]**o+(0.308418491165)*x[0]+(0.786418296678)*x[1]**o+(-0.248788086479)*x[1]
        arg[3,3,1,0]=(-0.981107944688)*x[0]**o+(-0.331637054173)*x[0]+(-0.972981902152)*x[1]**o+(-0.389364980465)*x[1]
        arg[3,3,1,1]=(-0.839880147367)*x[0]**o+(-0.911569972467)*x[0]+(0.209642332468)*x[1]**o+(0.0926875442137)*x[1]
        arg[3,3,2,0]=(-0.22724314378)*x[0]**o+(-0.0396604383619)*x[0]+(0.906829234649)*x[1]**o+(-0.862410550066)*x[1]
        arg[3,3,2,1]=(-0.412816401935)*x[0]**o+(-0.848442294241)*x[0]+(-0.599302206188)*x[1]**o+(0.719403009819)*x[1]
        arg[3,4,0,0]=(0.259283054664)*x[0]**o+(0.0983272120339)*x[0]+(-0.538373082787)*x[1]**o+(0.950164435617)*x[1]
        arg[3,4,0,1]=(0.633594236794)*x[0]**o+(0.120984469179)*x[0]+(0.472771297982)*x[1]**o+(0.873015785724)*x[1]
        arg[3,4,1,0]=(0.913524970057)*x[0]**o+(0.106590018424)*x[0]+(0.124026494061)*x[1]**o+(-0.254989594971)*x[1]
        arg[3,4,1,1]=(0.945528885492)*x[0]**o+(-0.323205519576)*x[0]+(-0.467294290307)*x[1]**o+(-0.557033486306)*x[1]
        arg[3,4,2,0]=(-0.268364682764)*x[0]**o+(0.0289685280025)*x[0]+(0.0493964636013)*x[1]**o+(0.872120137365)*x[1]
        arg[3,4,2,1]=(-0.266698796897)*x[0]**o+(-0.679148300702)*x[0]+(0.337653394795)*x[1]**o+(0.186866149716)*x[1]
        ref[0,0,0,0]=(0.358768976574)*x_ref[0]**o+(-0.231637053165)*x_ref[0]+(0.798891506345)*x_ref[1]**o+(0.292325388353)*x_ref[1]
        ref[0,0,0,1]=(0.53301689922)*x_ref[0]**o+(-0.761786870664)*x_ref[0]+(0.284067900826)*x_ref[1]**o+(-0.688662082205)*x_ref[1]
        ref[0,0,1,0]=(-0.721988036113)*x_ref[0]**o+(-0.743892757815)*x_ref[0]+(-0.796787022375)*x_ref[1]**o+(-0.499468412523)*x_ref[1]
        ref[0,0,1,1]=(0.0710691488306)*x_ref[0]**o+(0.0210385141456)*x_ref[0]+(-0.709342417683)*x_ref[1]**o+(0.901794071115)*x_ref[1]
        ref[0,0,2,0]=(-0.800537263339)*x_ref[0]**o+(0.583114786539)*x_ref[0]+(0.422752622522)*x_ref[1]**o+(-0.0954981460748)*x_ref[1]
        ref[0,0,2,1]=(-0.0805357065445)*x_ref[0]**o+(0.476377895547)*x_ref[0]+(0.242203912566)*x_ref[1]**o+(-0.479971540043)*x_ref[1]
        ref[0,1,0,0]=(-0.476356823162)*x_ref[0]**o+(-0.438853622252)*x_ref[0]+(-0.937228141746)*x_ref[1]**o+(0.366496742588)*x_ref[1]
        ref[0,1,0,1]=(-0.496918253956)*x_ref[0]**o+(-0.0227549061268)*x_ref[0]+(0.830195584768)*x_ref[1]**o+(-0.114624902423)*x_ref[1]
        ref[0,1,1,0]=(0.431291036997)*x_ref[0]**o+(0.807713720258)*x_ref[0]+(0.730057543333)*x_ref[1]**o+(0.635171725232)*x_ref[1]
        ref[0,1,1,1]=(0.784425995411)*x_ref[0]**o+(-0.397510494899)*x_ref[0]+(-0.579911740115)*x_ref[1]**o+(0.130426014607)*x_ref[1]
        ref[0,1,2,0]=(0.473613650641)*x_ref[0]**o+(-0.155617741177)*x_ref[0]+(0.873237852103)*x_ref[1]**o+(-0.749395152414)*x_ref[1]
        ref[0,1,2,1]=(0.367649559431)*x_ref[0]**o+(0.830181309644)*x_ref[0]+(0.87726968147)*x_ref[1]**o+(-0.450304793682)*x_ref[1]
        ref[0,2,0,0]=(-0.700522876654)*x_ref[0]**o+(-0.998187126003)*x_ref[0]+(0.732491610173)*x_ref[1]**o+(0.648679267)*x_ref[1]
        ref[0,2,0,1]=(0.499746141285)*x_ref[0]**o+(-0.730090975806)*x_ref[0]+(-0.525566438654)*x_ref[1]**o+(0.212469850421)*x_ref[1]
        ref[0,2,1,0]=(-0.847752916298)*x_ref[0]**o+(0.43510472619)*x_ref[0]+(0.815305936548)*x_ref[1]**o+(-0.143802676552)*x_ref[1]
        ref[0,2,1,1]=(-0.453572376958)*x_ref[0]**o+(-0.913023531929)*x_ref[0]+(0.743328127665)*x_ref[1]**o+(0.152586122591)*x_ref[1]
        ref[0,2,2,0]=(0.526939334508)*x_ref[0]**o+(0.247268722057)*x_ref[0]+(-0.176432621481)*x_ref[1]**o+(-0.510110202196)*x_ref[1]
        ref[0,2,2,1]=(0.139685732003)*x_ref[0]**o+(0.278304455867)*x_ref[0]+(0.517953769616)*x_ref[1]**o+(-0.272384261701)*x_ref[1]
        ref[0,3,0,0]=(-0.38867907688)*x_ref[0]**o+(-0.73043975055)*x_ref[0]+(-0.14293243901)*x_ref[1]**o+(0.0601042854043)*x_ref[1]
        ref[0,3,0,1]=(-0.163055949188)*x_ref[0]**o+(-0.035877062482)*x_ref[0]+(0.0396615423106)*x_ref[1]**o+(-0.870062303275)*x_ref[1]
        ref[0,3,1,0]=(0.283839779914)*x_ref[0]**o+(-0.519316925197)*x_ref[0]+(-0.437401973959)*x_ref[1]**o+(-0.823707232663)*x_ref[1]
        ref[0,3,1,1]=(0.700305425892)*x_ref[0]**o+(0.0299761521094)*x_ref[0]+(-0.993032151025)*x_ref[1]**o+(0.633293413221)*x_ref[1]
        ref[0,3,2,0]=(-0.745357256365)*x_ref[0]**o+(-0.451553456961)*x_ref[0]+(0.05789646302)*x_ref[1]**o+(-0.755901155251)*x_ref[1]
        ref[0,3,2,1]=(-0.57330840403)*x_ref[0]**o+(-0.589361483751)*x_ref[0]+(-0.038677123513)*x_ref[1]**o+(0.417332032168)*x_ref[1]
        ref[0,4,0,0]=(0.759655161531)*x_ref[0]**o+(0.737483461328)*x_ref[0]+(0.573158629937)*x_ref[1]**o+(-0.454191109106)*x_ref[1]
        ref[0,4,0,1]=(-0.232982406645)*x_ref[0]**o+(-0.926030038179)*x_ref[0]+(-0.504256834465)*x_ref[1]**o+(-0.107703527058)*x_ref[1]
        ref[0,4,1,0]=(0.898411697016)*x_ref[0]**o+(0.642045536742)*x_ref[0]+(-0.953059791895)*x_ref[1]**o+(-0.67501710093)*x_ref[1]
        ref[0,4,1,1]=(0.300380574165)*x_ref[0]**o+(0.204591960115)*x_ref[0]+(-0.308849697562)*x_ref[1]**o+(-0.489495887357)*x_ref[1]
        ref[0,4,2,0]=(-0.814560286716)*x_ref[0]**o+(0.154235599953)*x_ref[0]+(0.279329273498)*x_ref[1]**o+(-0.674063978926)*x_ref[1]
        ref[0,4,2,1]=(-0.718018870365)*x_ref[0]**o+(-0.735982858424)*x_ref[0]+(-0.0132197132926)*x_ref[1]**o+(0.398939587543)*x_ref[1]
        ref[1,0,0,0]=(0.0559995351725)*x_ref[0]**o+(-0.164549619892)*x_ref[0]+(-0.794242529408)*x_ref[1]**o+(0.0927822314785)*x_ref[1]
        ref[1,0,0,1]=(-0.894238542341)*x_ref[0]**o+(0.378559247473)*x_ref[0]+(-0.183957272639)*x_ref[1]**o+(0.486358968942)*x_ref[1]
        ref[1,0,1,0]=(0.931288801001)*x_ref[0]**o+(0.223815769968)*x_ref[0]+(-0.719591906328)*x_ref[1]**o+(-0.743644741105)*x_ref[1]
        ref[1,0,1,1]=(0.156646029123)*x_ref[0]**o+(0.306942879419)*x_ref[0]+(0.747006830674)*x_ref[1]**o+(-0.876534912274)*x_ref[1]
        ref[1,0,2,0]=(-0.496819081999)*x_ref[0]**o+(-0.939227402425)*x_ref[0]+(0.0566331712182)*x_ref[1]**o+(0.565724364661)*x_ref[1]
        ref[1,0,2,1]=(0.903706429954)*x_ref[0]**o+(0.667947320307)*x_ref[0]+(0.97140855373)*x_ref[1]**o+(0.924095235752)*x_ref[1]
        ref[1,1,0,0]=(0.526180197299)*x_ref[0]**o+(0.435493351674)*x_ref[0]+(0.400107077848)*x_ref[1]**o+(-0.704760833758)*x_ref[1]
        ref[1,1,0,1]=(0.0723740015004)*x_ref[0]**o+(-0.405859051371)*x_ref[0]+(0.200787757511)*x_ref[1]**o+(0.41819184563)*x_ref[1]
        ref[1,1,1,0]=(-0.0320951655627)*x_ref[0]**o+(0.235068445821)*x_ref[0]+(0.450383349599)*x_ref[1]**o+(-0.801487424765)*x_ref[1]
        ref[1,1,1,1]=(-0.52478892744)*x_ref[0]**o+(-0.449111297125)*x_ref[0]+(0.731062182014)*x_ref[1]**o+(0.533619608577)*x_ref[1]
        ref[1,1,2,0]=(0.0889794057057)*x_ref[0]**o+(-0.971856981571)*x_ref[0]+(0.662216408922)*x_ref[1]**o+(0.164089024468)*x_ref[1]
        ref[1,1,2,1]=(-0.196374963647)*x_ref[0]**o+(-0.739360211802)*x_ref[0]+(0.916611650753)*x_ref[1]**o+(-0.260102318115)*x_ref[1]
        ref[1,2,0,0]=(0.627763572103)*x_ref[0]**o+(-0.893900847319)*x_ref[0]+(0.907548446417)*x_ref[1]**o+(0.0687906203663)*x_ref[1]
        ref[1,2,0,1]=(0.59755498682)*x_ref[0]**o+(-0.457108009787)*x_ref[0]+(0.71562139654)*x_ref[1]**o+(-0.300872832559)*x_ref[1]
        ref[1,2,1,0]=(-0.479641829834)*x_ref[0]**o+(0.297188513722)*x_ref[0]+(0.0227730978493)*x_ref[1]**o+(0.199329665406)*x_ref[1]
        ref[1,2,1,1]=(0.918974940101)*x_ref[0]**o+(0.595194981816)*x_ref[0]+(0.760030152554)*x_ref[1]**o+(-0.798899878486)*x_ref[1]
        ref[1,2,2,0]=(-0.878334511842)*x_ref[0]**o+(0.910857236287)*x_ref[0]+(0.0164747567997)*x_ref[1]**o+(-0.876970888445)*x_ref[1]
        ref[1,2,2,1]=(0.29445809181)*x_ref[0]**o+(0.837849454931)*x_ref[0]+(0.0759054371211)*x_ref[1]**o+(0.501243080372)*x_ref[1]
        ref[1,3,0,0]=(-0.529006436258)*x_ref[0]**o+(-0.932325855538)*x_ref[0]+(-0.561472972374)*x_ref[1]**o+(0.845125736656)*x_ref[1]
        ref[1,3,0,1]=(-0.843008257488)*x_ref[0]**o+(0.614118623012)*x_ref[0]+(-0.167015137447)*x_ref[1]**o+(0.158701952625)*x_ref[1]
        ref[1,3,1,0]=(-0.662966916148)*x_ref[0]**o+(-0.443297576512)*x_ref[0]+(0.295988608077)*x_ref[1]**o+(0.059506908615)*x_ref[1]
        ref[1,3,1,1]=(0.864199978014)*x_ref[0]**o+(0.138723519602)*x_ref[0]+(0.144964839302)*x_ref[1]**o+(0.239863709389)*x_ref[1]
        ref[1,3,2,0]=(0.479006097431)*x_ref[0]**o+(0.542168797821)*x_ref[0]+(0.561074562163)*x_ref[1]**o+(-0.133365787275)*x_ref[1]
        ref[1,3,2,1]=(-0.901033449452)*x_ref[0]**o+(-0.620245150074)*x_ref[0]+(0.598461569697)*x_ref[1]**o+(-0.2213246854)*x_ref[1]
        ref[1,4,0,0]=(-0.577067401348)*x_ref[0]**o+(0.486154324874)*x_ref[0]+(0.809205150442)*x_ref[1]**o+(-0.218206418459)*x_ref[1]
        ref[1,4,0,1]=(0.726556862066)*x_ref[0]**o+(0.0293695509867)*x_ref[0]+(0.601938348193)*x_ref[1]**o+(0.967332697531)*x_ref[1]
        ref[1,4,1,0]=(-0.675624388054)*x_ref[0]**o+(-0.871613723214)*x_ref[0]+(0.605519387693)*x_ref[1]**o+(0.287186894411)*x_ref[1]
        ref[1,4,1,1]=(-0.689904898214)*x_ref[0]**o+(-0.918236705665)*x_ref[0]+(-0.892168622177)*x_ref[1]**o+(0.507441662612)*x_ref[1]
        ref[1,4,2,0]=(0.845768980712)*x_ref[0]**o+(-0.458467508114)*x_ref[0]+(-0.518185249391)*x_ref[1]**o+(0.829945330327)*x_ref[1]
        ref[1,4,2,1]=(0.776852312178)*x_ref[0]**o+(-0.268810999469)*x_ref[0]+(-0.576757148841)*x_ref[1]**o+(-0.50628294802)*x_ref[1]
        ref[2,0,0,0]=(-0.686840219939)*x_ref[0]**o+(0.6861172847)*x_ref[0]+(0.240362263052)*x_ref[1]**o+(0.962361523884)*x_ref[1]
        ref[2,0,0,1]=(-0.943837188119)*x_ref[0]**o+(0.829832768268)*x_ref[0]+(0.377590124133)*x_ref[1]**o+(-0.171714144962)*x_ref[1]
        ref[2,0,1,0]=(0.862809148447)*x_ref[0]**o+(0.0573402136489)*x_ref[0]+(-0.743311826013)*x_ref[1]**o+(0.427809313437)*x_ref[1]
        ref[2,0,1,1]=(0.377380819576)*x_ref[0]**o+(-0.223280285041)*x_ref[0]+(0.178628658562)*x_ref[1]**o+(0.121964611877)*x_ref[1]
        ref[2,0,2,0]=(0.41723758689)*x_ref[0]**o+(0.256039203883)*x_ref[0]+(0.255358776676)*x_ref[1]**o+(-0.881665964974)*x_ref[1]
        ref[2,0,2,1]=(-0.0797103024499)*x_ref[0]**o+(0.128068931414)*x_ref[0]+(-0.758317834409)*x_ref[1]**o+(0.654657809183)*x_ref[1]
        ref[2,1,0,0]=(0.603648218133)*x_ref[0]**o+(-0.286780270729)*x_ref[0]+(-0.341108067214)*x_ref[1]**o+(-0.769667356087)*x_ref[1]
        ref[2,1,0,1]=(0.844044270035)*x_ref[0]**o+(0.0585907857764)*x_ref[0]+(-0.578232785252)*x_ref[1]**o+(-0.40959017859)*x_ref[1]
        ref[2,1,1,0]=(0.599709567393)*x_ref[0]**o+(0.307092449739)*x_ref[0]+(-0.402950586215)*x_ref[1]**o+(0.0674239470451)*x_ref[1]
        ref[2,1,1,1]=(0.182608249961)*x_ref[0]**o+(0.373533584601)*x_ref[0]+(0.40949869082)*x_ref[1]**o+(0.696333617699)*x_ref[1]
        ref[2,1,2,0]=(0.0462891111054)*x_ref[0]**o+(0.371258293292)*x_ref[0]+(0.782976187716)*x_ref[1]**o+(0.117976017999)*x_ref[1]
        ref[2,1,2,1]=(0.891841121526)*x_ref[0]**o+(-0.998971173204)*x_ref[0]+(0.0286712507928)*x_ref[1]**o+(-0.0198507022516)*x_ref[1]
        ref[2,2,0,0]=(0.284620902703)*x_ref[0]**o+(0.936231203073)*x_ref[0]+(0.829209752046)*x_ref[1]**o+(-0.94940982125)*x_ref[1]
        ref[2,2,0,1]=(-0.647956027376)*x_ref[0]**o+(0.923690718922)*x_ref[0]+(-0.0322423837346)*x_ref[1]**o+(0.124962508606)*x_ref[1]
        ref[2,2,1,0]=(-0.0696010411705)*x_ref[0]**o+(-0.161672936739)*x_ref[0]+(-0.997445762186)*x_ref[1]**o+(0.177962302014)*x_ref[1]
        ref[2,2,1,1]=(-0.897657700296)*x_ref[0]**o+(-0.467866721175)*x_ref[0]+(-0.560034332007)*x_ref[1]**o+(-0.861509898295)*x_ref[1]
        ref[2,2,2,0]=(-0.100880174048)*x_ref[0]**o+(0.678568254695)*x_ref[0]+(-0.83289411684)*x_ref[1]**o+(0.680564393149)*x_ref[1]
        ref[2,2,2,1]=(0.546159181491)*x_ref[0]**o+(0.720296622966)*x_ref[0]+(0.939046941118)*x_ref[1]**o+(-0.716147554733)*x_ref[1]
        ref[2,3,0,0]=(-0.921073162235)*x_ref[0]**o+(0.979925882289)*x_ref[0]+(-0.0938642966949)*x_ref[1]**o+(0.482093047223)*x_ref[1]
        ref[2,3,0,1]=(0.0709940956822)*x_ref[0]**o+(-0.691401017243)*x_ref[0]+(0.447000814029)*x_ref[1]**o+(0.0575856240533)*x_ref[1]
        ref[2,3,1,0]=(0.562800393365)*x_ref[0]**o+(0.657646549915)*x_ref[0]+(-0.835326361689)*x_ref[1]**o+(0.3487960687)*x_ref[1]
        ref[2,3,1,1]=(0.106668867409)*x_ref[0]**o+(0.436327240765)*x_ref[0]+(-0.585051261259)*x_ref[1]**o+(0.616655338667)*x_ref[1]
        ref[2,3,2,0]=(-0.501849628176)*x_ref[0]**o+(-0.284722503916)*x_ref[0]+(-0.892470199837)*x_ref[1]**o+(-0.979942541795)*x_ref[1]
        ref[2,3,2,1]=(0.236759540073)*x_ref[0]**o+(-0.806931113769)*x_ref[0]+(-0.225220003475)*x_ref[1]**o+(-0.95183239844)*x_ref[1]
        ref[2,4,0,0]=(-0.684823909581)*x_ref[0]**o+(0.445615228059)*x_ref[0]+(-0.638879895767)*x_ref[1]**o+(-0.335505874088)*x_ref[1]
        ref[2,4,0,1]=(-0.196708460612)*x_ref[0]**o+(0.981592996809)*x_ref[0]+(0.435774317028)*x_ref[1]**o+(-0.0952200663109)*x_ref[1]
        ref[2,4,1,0]=(-0.552759474764)*x_ref[0]**o+(-0.614031173545)*x_ref[0]+(0.304375542871)*x_ref[1]**o+(-0.285131901533)*x_ref[1]
        ref[2,4,1,1]=(0.147812951421)*x_ref[0]**o+(-0.78533487681)*x_ref[0]+(0.942205524401)*x_ref[1]**o+(0.166280215075)*x_ref[1]
        ref[2,4,2,0]=(0.843067982563)*x_ref[0]**o+(0.0116520710289)*x_ref[0]+(0.927948889059)*x_ref[1]**o+(-0.395251919577)*x_ref[1]
        ref[2,4,2,1]=(-0.316538010007)*x_ref[0]**o+(-0.139262497126)*x_ref[0]+(0.645409249217)*x_ref[1]**o+(0.192949521992)*x_ref[1]
        ref[3,0,0,0]=(0.88179290666)*x_ref[0]**o+(0.337853287365)*x_ref[0]+(-0.590219587193)*x_ref[1]**o+(-0.827051504196)*x_ref[1]
        ref[3,0,0,1]=(-0.104526336997)*x_ref[0]**o+(0.299097012282)*x_ref[0]+(-0.348493995853)*x_ref[1]**o+(-0.780034322304)*x_ref[1]
        ref[3,0,1,0]=(0.358724159445)*x_ref[0]**o+(-0.0338089482488)*x_ref[0]+(-0.0423054506456)*x_ref[1]**o+(0.67180489557)*x_ref[1]
        ref[3,0,1,1]=(-0.778880095665)*x_ref[0]**o+(-0.448812173928)*x_ref[0]+(-0.823686122382)*x_ref[1]**o+(0.636028922656)*x_ref[1]
        ref[3,0,2,0]=(0.146891002069)*x_ref[0]**o+(-0.368903199628)*x_ref[0]+(-0.596272077557)*x_ref[1]**o+(0.703510683225)*x_ref[1]
        ref[3,0,2,1]=(0.357416423547)*x_ref[0]**o+(-0.103754905142)*x_ref[0]+(0.32026011189)*x_ref[1]**o+(0.755648597627)*x_ref[1]
        ref[3,1,0,0]=(0.399082464238)*x_ref[0]**o+(-0.12611921797)*x_ref[0]+(0.426552225948)*x_ref[1]**o+(-0.228714297286)*x_ref[1]
        ref[3,1,0,1]=(-0.638411026717)*x_ref[0]**o+(-0.342143922306)*x_ref[0]+(-0.502107802626)*x_ref[1]**o+(-0.881045304848)*x_ref[1]
        ref[3,1,1,0]=(0.411134247295)*x_ref[0]**o+(-0.216585727086)*x_ref[0]+(0.270013339156)*x_ref[1]**o+(0.251581243494)*x_ref[1]
        ref[3,1,1,1]=(0.788758848294)*x_ref[0]**o+(0.345181621902)*x_ref[0]+(0.102843710598)*x_ref[1]**o+(0.0561071077516)*x_ref[1]
        ref[3,1,2,0]=(0.232441162347)*x_ref[0]**o+(-0.687544250386)*x_ref[0]+(0.369065746372)*x_ref[1]**o+(0.628109564412)*x_ref[1]
        ref[3,1,2,1]=(-0.912295044312)*x_ref[0]**o+(0.847830497468)*x_ref[0]+(0.776648618003)*x_ref[1]**o+(0.522638388084)*x_ref[1]
        ref[3,2,0,0]=(0.779333120942)*x_ref[0]**o+(-0.0747119967046)*x_ref[0]+(-0.594363282198)*x_ref[1]**o+(0.513922159444)*x_ref[1]
        ref[3,2,0,1]=(0.877492427404)*x_ref[0]**o+(-0.191727825399)*x_ref[0]+(0.0881296554818)*x_ref[1]**o+(-0.408407503067)*x_ref[1]
        ref[3,2,1,0]=(0.29696771215)*x_ref[0]**o+(-0.914882727138)*x_ref[0]+(0.684156831385)*x_ref[1]**o+(0.142293411327)*x_ref[1]
        ref[3,2,1,1]=(0.105927449304)*x_ref[0]**o+(-0.863339661598)*x_ref[0]+(-0.815550662607)*x_ref[1]**o+(-0.261034715274)*x_ref[1]
        ref[3,2,2,0]=(0.944082529871)*x_ref[0]**o+(-0.449259627627)*x_ref[0]+(-0.560526159016)*x_ref[1]**o+(0.609050500798)*x_ref[1]
        ref[3,2,2,1]=(0.529508870267)*x_ref[0]**o+(0.627216712239)*x_ref[0]+(-0.0148122228382)*x_ref[1]**o+(0.0743274653452)*x_ref[1]
        ref[3,3,0,0]=(-0.30609697448)*x_ref[0]**o+(0.910972095956)*x_ref[0]+(0.256274444924)*x_ref[1]**o+(0.229096821635)*x_ref[1]
        ref[3,3,0,1]=(-0.510323627078)*x_ref[0]**o+(0.308418491165)*x_ref[0]+(0.786418296678)*x_ref[1]**o+(-0.248788086479)*x_ref[1]
        ref[3,3,1,0]=(-0.981107944688)*x_ref[0]**o+(-0.331637054173)*x_ref[0]+(-0.972981902152)*x_ref[1]**o+(-0.389364980465)*x_ref[1]
        ref[3,3,1,1]=(-0.839880147367)*x_ref[0]**o+(-0.911569972467)*x_ref[0]+(0.209642332468)*x_ref[1]**o+(0.0926875442137)*x_ref[1]
        ref[3,3,2,0]=(-0.22724314378)*x_ref[0]**o+(-0.0396604383619)*x_ref[0]+(0.906829234649)*x_ref[1]**o+(-0.862410550066)*x_ref[1]
        ref[3,3,2,1]=(-0.412816401935)*x_ref[0]**o+(-0.848442294241)*x_ref[0]+(-0.599302206188)*x_ref[1]**o+(0.719403009819)*x_ref[1]
        ref[3,4,0,0]=(0.259283054664)*x_ref[0]**o+(0.0983272120339)*x_ref[0]+(-0.538373082787)*x_ref[1]**o+(0.950164435617)*x_ref[1]
        ref[3,4,0,1]=(0.633594236794)*x_ref[0]**o+(0.120984469179)*x_ref[0]+(0.472771297982)*x_ref[1]**o+(0.873015785724)*x_ref[1]
        ref[3,4,1,0]=(0.913524970057)*x_ref[0]**o+(0.106590018424)*x_ref[0]+(0.124026494061)*x_ref[1]**o+(-0.254989594971)*x_ref[1]
        ref[3,4,1,1]=(0.945528885492)*x_ref[0]**o+(-0.323205519576)*x_ref[0]+(-0.467294290307)*x_ref[1]**o+(-0.557033486306)*x_ref[1]
        ref[3,4,2,0]=(-0.268364682764)*x_ref[0]**o+(0.0289685280025)*x_ref[0]+(0.0493964636013)*x_ref[1]**o+(0.872120137365)*x_ref[1]
        ref[3,4,2,1]=(-0.266698796897)*x_ref[0]**o+(-0.679148300702)*x_ref[0]+(0.337653394795)*x_ref[1]**o+(0.186866149716)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.928996629952)*x[0]**o+(-0.171383074114)*x[0]+(-0.324423798901)*x[1]**o+(-0.801354053962)*x[1]+(0.365256142808)*x[2]**o+(0.126262208857)*x[2]
        arg[0,0,0,1]=(0.326024843626)*x[0]**o+(-0.0091566240426)*x[0]+(-0.0396235158589)*x[1]**o+(-0.753302577837)*x[1]+(-0.38709916587)*x[2]**o+(-0.28660316595)*x[2]
        arg[0,0,1,0]=(0.755338255506)*x[0]**o+(0.851895434343)*x[0]+(0.419119464402)*x[1]**o+(0.670733085023)*x[1]+(0.819398310958)*x[2]**o+(0.476185335773)*x[2]
        arg[0,0,1,1]=(-0.981416259131)*x[0]**o+(-0.97000252267)*x[0]+(-0.132624451699)*x[1]**o+(-0.261428370649)*x[1]+(-0.349245951727)*x[2]**o+(-0.717343195877)*x[2]
        arg[0,0,2,0]=(-0.874826075266)*x[0]**o+(-0.399307905707)*x[0]+(0.350873599317)*x[1]**o+(0.15861651992)*x[1]+(0.396279933216)*x[2]**o+(0.92332515708)*x[2]
        arg[0,0,2,1]=(-0.183596683527)*x[0]**o+(-0.115194639348)*x[0]+(0.616111723618)*x[1]**o+(-0.402157657669)*x[1]+(0.44145594193)*x[2]**o+(0.686052407784)*x[2]
        arg[0,1,0,0]=(-0.669169178244)*x[0]**o+(0.835937801609)*x[0]+(0.468805302181)*x[1]**o+(-0.627292634023)*x[1]+(-0.674361231979)*x[2]**o+(-0.00929181785129)*x[2]
        arg[0,1,0,1]=(-0.568131655705)*x[0]**o+(0.690366535852)*x[0]+(-0.212524396914)*x[1]**o+(-0.767983769407)*x[1]+(0.477500128569)*x[2]**o+(-0.0190887598644)*x[2]
        arg[0,1,1,0]=(-0.374170635822)*x[0]**o+(0.0169009974558)*x[0]+(-0.162519103357)*x[1]**o+(0.786991144449)*x[1]+(0.0590294918131)*x[2]**o+(-0.213442509309)*x[2]
        arg[0,1,1,1]=(-0.558268657439)*x[0]**o+(0.584734933669)*x[0]+(-0.650790013689)*x[1]**o+(-0.192849048407)*x[1]+(0.505458636714)*x[2]**o+(-0.585174281371)*x[2]
        arg[0,1,2,0]=(-0.933243590775)*x[0]**o+(0.528060249767)*x[0]+(-0.763255746252)*x[1]**o+(0.771589118163)*x[1]+(-0.641287604338)*x[2]**o+(0.379566401585)*x[2]
        arg[0,1,2,1]=(-0.836976373935)*x[0]**o+(-0.969434110974)*x[0]+(0.758148372888)*x[1]**o+(-0.89212062535)*x[1]+(0.788693820015)*x[2]**o+(-0.620704367414)*x[2]
        arg[0,2,0,0]=(0.0163117360814)*x[0]**o+(-0.390867752068)*x[0]+(-0.725068297982)*x[1]**o+(-0.680312496646)*x[1]+(0.069273494969)*x[2]**o+(-0.576122275696)*x[2]
        arg[0,2,0,1]=(0.461771197614)*x[0]**o+(-0.767541404109)*x[0]+(-0.694279263622)*x[1]**o+(-0.721448812749)*x[1]+(0.248787896584)*x[2]**o+(0.237075401327)*x[2]
        arg[0,2,1,0]=(0.381987012965)*x[0]**o+(-0.288236233238)*x[0]+(-0.519012025084)*x[1]**o+(-0.256658293082)*x[1]+(0.807322167368)*x[2]**o+(-0.261916179859)*x[2]
        arg[0,2,1,1]=(0.257241267407)*x[0]**o+(-0.470612990721)*x[0]+(-0.208612656448)*x[1]**o+(0.839646962914)*x[1]+(0.916675150494)*x[2]**o+(-0.165468351005)*x[2]
        arg[0,2,2,0]=(0.92569232521)*x[0]**o+(0.809982458308)*x[0]+(-0.860530305336)*x[1]**o+(0.261687622817)*x[1]+(0.948003684651)*x[2]**o+(-0.473133742361)*x[2]
        arg[0,2,2,1]=(-0.113846814874)*x[0]**o+(0.680846459377)*x[0]+(0.920219149741)*x[1]**o+(0.172027554503)*x[1]+(0.475330041115)*x[2]**o+(-0.176218891429)*x[2]
        arg[0,3,0,0]=(-0.948638294297)*x[0]**o+(-0.0411621070703)*x[0]+(-0.294759271027)*x[1]**o+(-0.903733455086)*x[1]+(-0.995253733593)*x[2]**o+(0.106726960462)*x[2]
        arg[0,3,0,1]=(0.1592709614)*x[0]**o+(0.0683266445293)*x[0]+(0.336092506745)*x[1]**o+(-0.209047739881)*x[1]+(-0.330417243459)*x[2]**o+(-0.809242647755)*x[2]
        arg[0,3,1,0]=(-0.680637516359)*x[0]**o+(-0.713405815404)*x[0]+(-0.484624435043)*x[1]**o+(-0.279314462332)*x[1]+(-0.458856066504)*x[2]**o+(0.0588111305227)*x[2]
        arg[0,3,1,1]=(-0.853676338655)*x[0]**o+(-0.651539814195)*x[0]+(-0.109727277725)*x[1]**o+(0.0139006407942)*x[1]+(0.231943245426)*x[2]**o+(0.42972517233)*x[2]
        arg[0,3,2,0]=(-0.60799269765)*x[0]**o+(-0.802413269128)*x[0]+(0.151362498863)*x[1]**o+(-0.678226252875)*x[1]+(-0.867117685659)*x[2]**o+(-0.711982116071)*x[2]
        arg[0,3,2,1]=(-0.243012743941)*x[0]**o+(-0.692614579016)*x[0]+(0.0986817641914)*x[1]**o+(-0.495735862695)*x[1]+(-0.430040681653)*x[2]**o+(0.0108928722284)*x[2]
        arg[0,4,0,0]=(0.0989402181558)*x[0]**o+(-0.60096719882)*x[0]+(-0.324253928432)*x[1]**o+(-0.862723687559)*x[1]+(-0.34745416427)*x[2]**o+(0.936559074873)*x[2]
        arg[0,4,0,1]=(0.711021967566)*x[0]**o+(-0.939408401113)*x[0]+(-0.218100921633)*x[1]**o+(0.92835194875)*x[1]+(-0.353936913865)*x[2]**o+(-0.547266090928)*x[2]
        arg[0,4,1,0]=(0.172231614188)*x[0]**o+(-0.734622090671)*x[0]+(0.432686916647)*x[1]**o+(0.552595350492)*x[1]+(0.779662641076)*x[2]**o+(-0.0933220492779)*x[2]
        arg[0,4,1,1]=(0.996776716866)*x[0]**o+(-0.340786520686)*x[0]+(0.260075214171)*x[1]**o+(0.00385504852496)*x[1]+(0.377516772127)*x[2]**o+(-0.341751634181)*x[2]
        arg[0,4,2,0]=(-0.655642896664)*x[0]**o+(-0.840525173737)*x[0]+(-0.855343721564)*x[1]**o+(0.253558571524)*x[1]+(0.0463733059925)*x[2]**o+(-0.116359866912)*x[2]
        arg[0,4,2,1]=(0.131619661414)*x[0]**o+(-0.162067255954)*x[0]+(0.562536831367)*x[1]**o+(-0.940412472035)*x[1]+(-0.282265762783)*x[2]**o+(0.634711607923)*x[2]
        arg[1,0,0,0]=(0.484071808097)*x[0]**o+(0.566763524691)*x[0]+(-0.658614173653)*x[1]**o+(-0.140192232895)*x[1]+(-0.310114236409)*x[2]**o+(-0.918532401595)*x[2]
        arg[1,0,0,1]=(-0.81190380441)*x[0]**o+(0.644439878497)*x[0]+(0.479583403289)*x[1]**o+(-0.493318882779)*x[1]+(-0.424704492175)*x[2]**o+(0.700624317732)*x[2]
        arg[1,0,1,0]=(0.766650182435)*x[0]**o+(0.226365545129)*x[0]+(0.510828945971)*x[1]**o+(0.283221954498)*x[1]+(0.817303877886)*x[2]**o+(-0.245709044628)*x[2]
        arg[1,0,1,1]=(0.194787112664)*x[0]**o+(-0.879306736922)*x[0]+(0.185381291627)*x[1]**o+(-0.510692743975)*x[1]+(-0.422555014055)*x[2]**o+(0.556629919959)*x[2]
        arg[1,0,2,0]=(-0.958345828895)*x[0]**o+(-0.667735551013)*x[0]+(0.278271284225)*x[1]**o+(-0.0847995194757)*x[1]+(0.902184171441)*x[2]**o+(0.763125168854)*x[2]
        arg[1,0,2,1]=(-0.0766429670195)*x[0]**o+(0.894421139029)*x[0]+(-0.331560314101)*x[1]**o+(-0.0253297623997)*x[1]+(0.13406045967)*x[2]**o+(-0.326681111625)*x[2]
        arg[1,1,0,0]=(-0.979406855838)*x[0]**o+(0.528849909402)*x[0]+(-0.0486159476737)*x[1]**o+(-0.898392512695)*x[1]+(-0.678328327407)*x[2]**o+(0.644201802057)*x[2]
        arg[1,1,0,1]=(-0.977485554605)*x[0]**o+(0.794799941538)*x[0]+(-0.00237183848504)*x[1]**o+(-0.633497483149)*x[1]+(0.746593853568)*x[2]**o+(0.197741126365)*x[2]
        arg[1,1,1,0]=(0.379985506182)*x[0]**o+(0.69889144234)*x[0]+(-0.507293887409)*x[1]**o+(-0.727675545291)*x[1]+(-0.467064531693)*x[2]**o+(0.130273930089)*x[2]
        arg[1,1,1,1]=(-0.330627830819)*x[0]**o+(-0.0311542935889)*x[0]+(-0.0671494148749)*x[1]**o+(0.583605469163)*x[1]+(-0.172083996069)*x[2]**o+(0.666718011074)*x[2]
        arg[1,1,2,0]=(0.539618381063)*x[0]**o+(-0.546796237888)*x[0]+(0.857911750392)*x[1]**o+(0.119552731463)*x[1]+(0.329286820109)*x[2]**o+(0.637240649087)*x[2]
        arg[1,1,2,1]=(-0.828603071334)*x[0]**o+(0.861799149065)*x[0]+(0.615224040538)*x[1]**o+(-0.0868985543263)*x[1]+(-0.206168997809)*x[2]**o+(0.311103719286)*x[2]
        arg[1,2,0,0]=(0.695106867715)*x[0]**o+(-0.533303418727)*x[0]+(-0.393936955207)*x[1]**o+(-0.407635261686)*x[1]+(0.604346269147)*x[2]**o+(-0.10213061533)*x[2]
        arg[1,2,0,1]=(0.338035309929)*x[0]**o+(0.517135412671)*x[0]+(0.277863735829)*x[1]**o+(-0.703417079398)*x[1]+(0.369398635984)*x[2]**o+(-0.952905070601)*x[2]
        arg[1,2,1,0]=(0.937866603297)*x[0]**o+(0.436121631708)*x[0]+(0.568362098461)*x[1]**o+(0.223837722984)*x[1]+(-0.236896910941)*x[2]**o+(0.316114408308)*x[2]
        arg[1,2,1,1]=(0.285714491932)*x[0]**o+(0.633321537987)*x[0]+(0.953737237202)*x[1]**o+(0.732065555924)*x[1]+(-0.515759068197)*x[2]**o+(-0.519775334889)*x[2]
        arg[1,2,2,0]=(-0.0610222303479)*x[0]**o+(0.135974983944)*x[0]+(0.450213393447)*x[1]**o+(-0.44747541715)*x[1]+(-0.640660207187)*x[2]**o+(0.348457261119)*x[2]
        arg[1,2,2,1]=(0.131753033369)*x[0]**o+(-0.311359549998)*x[0]+(0.445707767235)*x[1]**o+(-0.710795569938)*x[1]+(-0.746038648477)*x[2]**o+(-0.210414773188)*x[2]
        arg[1,3,0,0]=(-0.323684061289)*x[0]**o+(-0.0937091220282)*x[0]+(-0.827884898323)*x[1]**o+(0.604442907568)*x[1]+(0.620923596379)*x[2]**o+(0.851758170763)*x[2]
        arg[1,3,0,1]=(-0.269239872573)*x[0]**o+(-0.603248280988)*x[0]+(0.603804530326)*x[1]**o+(-0.331910095046)*x[1]+(-0.653877236308)*x[2]**o+(0.115434865132)*x[2]
        arg[1,3,1,0]=(0.0531798412869)*x[0]**o+(0.292950346899)*x[0]+(-0.634376081823)*x[1]**o+(-0.775427771578)*x[1]+(-0.987113467149)*x[2]**o+(0.757980444447)*x[2]
        arg[1,3,1,1]=(0.757807292214)*x[0]**o+(0.81778478971)*x[0]+(-0.619873148405)*x[1]**o+(0.0129855659138)*x[1]+(-0.282861093681)*x[2]**o+(0.667211104045)*x[2]
        arg[1,3,2,0]=(-0.915205606853)*x[0]**o+(0.918704643459)*x[0]+(0.143238155504)*x[1]**o+(0.0285686377214)*x[1]+(-0.848712843243)*x[2]**o+(0.058647124862)*x[2]
        arg[1,3,2,1]=(0.441487712831)*x[0]**o+(-0.626878272507)*x[0]+(-0.862579662295)*x[1]**o+(0.204629243832)*x[1]+(0.964143347265)*x[2]**o+(-0.399125844282)*x[2]
        arg[1,4,0,0]=(-0.349791275126)*x[0]**o+(-0.140293844776)*x[0]+(-0.693879671138)*x[1]**o+(0.429290435637)*x[1]+(-0.892670314048)*x[2]**o+(-0.380777380409)*x[2]
        arg[1,4,0,1]=(0.528712424647)*x[0]**o+(0.449268002671)*x[0]+(-0.202472214961)*x[1]**o+(-0.374762208851)*x[1]+(-0.625590920648)*x[2]**o+(-0.000776205555531)*x[2]
        arg[1,4,1,0]=(0.00425881397184)*x[0]**o+(0.265831686248)*x[0]+(0.757626547866)*x[1]**o+(-0.934207064756)*x[1]+(0.541356932538)*x[2]**o+(0.217841597089)*x[2]
        arg[1,4,1,1]=(-0.02422207106)*x[0]**o+(-0.807018149115)*x[0]+(-0.402347046418)*x[1]**o+(-0.0548851919048)*x[1]+(0.850202044077)*x[2]**o+(-0.889576261979)*x[2]
        arg[1,4,2,0]=(0.487059068561)*x[0]**o+(0.338254525003)*x[0]+(-0.387019663154)*x[1]**o+(-0.436969552236)*x[1]+(-0.747690766603)*x[2]**o+(0.265854907003)*x[2]
        arg[1,4,2,1]=(0.687562221132)*x[0]**o+(0.4755782976)*x[0]+(-0.456287111804)*x[1]**o+(0.412138123928)*x[1]+(0.637920334647)*x[2]**o+(0.817007725525)*x[2]
        arg[2,0,0,0]=(-0.728166422776)*x[0]**o+(0.108904177195)*x[0]+(0.383848108635)*x[1]**o+(-0.19252707292)*x[1]+(-0.279319122495)*x[2]**o+(-0.0530727745585)*x[2]
        arg[2,0,0,1]=(0.380827452272)*x[0]**o+(-0.133776280687)*x[0]+(-0.174731551283)*x[1]**o+(-0.369231086652)*x[1]+(-0.0190771818046)*x[2]**o+(-0.331873068919)*x[2]
        arg[2,0,1,0]=(0.653644298826)*x[0]**o+(-0.916110182689)*x[0]+(0.0510855252492)*x[1]**o+(-0.934192949208)*x[1]+(0.572048462784)*x[2]**o+(-0.451401856863)*x[2]
        arg[2,0,1,1]=(-0.394958497502)*x[0]**o+(-0.690752972227)*x[0]+(-0.436352069349)*x[1]**o+(0.933524576573)*x[1]+(-0.524344161699)*x[2]**o+(0.526748792776)*x[2]
        arg[2,0,2,0]=(0.845649523782)*x[0]**o+(0.63927747558)*x[0]+(-0.534192218216)*x[1]**o+(0.714500925459)*x[1]+(0.156077814898)*x[2]**o+(-0.193712778792)*x[2]
        arg[2,0,2,1]=(0.809636640413)*x[0]**o+(-0.930654037481)*x[0]+(0.509441692942)*x[1]**o+(0.881563824585)*x[1]+(0.101333749057)*x[2]**o+(0.418791722704)*x[2]
        arg[2,1,0,0]=(0.289481281745)*x[0]**o+(-0.98279193802)*x[0]+(-0.443382747319)*x[1]**o+(-0.281710992472)*x[1]+(0.426349350391)*x[2]**o+(-0.870192121251)*x[2]
        arg[2,1,0,1]=(0.837046002753)*x[0]**o+(-0.0254166983371)*x[0]+(0.30178635814)*x[1]**o+(0.973915444924)*x[1]+(0.298631524734)*x[2]**o+(-0.422490970248)*x[2]
        arg[2,1,1,0]=(0.869799082394)*x[0]**o+(0.376436513628)*x[0]+(-0.74735679118)*x[1]**o+(0.477184335901)*x[1]+(-0.797472924916)*x[2]**o+(-0.568827873964)*x[2]
        arg[2,1,1,1]=(0.993441009779)*x[0]**o+(-0.264888299493)*x[0]+(0.861197560367)*x[1]**o+(-0.769058084144)*x[1]+(-0.86221947415)*x[2]**o+(-0.617024169672)*x[2]
        arg[2,1,2,0]=(-0.89303691749)*x[0]**o+(-0.390287412175)*x[0]+(-0.312883084291)*x[1]**o+(-0.432368502641)*x[1]+(-0.121476666357)*x[2]**o+(0.416361314288)*x[2]
        arg[2,1,2,1]=(-0.622214426504)*x[0]**o+(-0.595279134338)*x[0]+(0.000927173498715)*x[1]**o+(0.783001179117)*x[1]+(-0.956003738061)*x[2]**o+(0.482321214492)*x[2]
        arg[2,2,0,0]=(-0.0047994639068)*x[0]**o+(0.478805093492)*x[0]+(0.747825846519)*x[1]**o+(0.0756064430807)*x[1]+(-0.470534826525)*x[2]**o+(-0.421524571831)*x[2]
        arg[2,2,0,1]=(-0.0569198078126)*x[0]**o+(-0.36284856463)*x[0]+(-0.374343856016)*x[1]**o+(0.193470734451)*x[1]+(0.425350990791)*x[2]**o+(-0.903309757487)*x[2]
        arg[2,2,1,0]=(0.602075251229)*x[0]**o+(-0.558163518701)*x[0]+(-0.00146699467501)*x[1]**o+(0.437116756605)*x[1]+(-0.252696658318)*x[2]**o+(0.154017654456)*x[2]
        arg[2,2,1,1]=(0.226970288166)*x[0]**o+(0.844076458869)*x[0]+(0.231874137232)*x[1]**o+(-0.270233945335)*x[1]+(0.365979082276)*x[2]**o+(-0.937144802776)*x[2]
        arg[2,2,2,0]=(0.662440813722)*x[0]**o+(-0.600830429698)*x[0]+(0.248436486724)*x[1]**o+(0.73439963812)*x[1]+(-0.406317484259)*x[2]**o+(-0.406641354653)*x[2]
        arg[2,2,2,1]=(0.948630940404)*x[0]**o+(-0.695098513788)*x[0]+(-0.819843883615)*x[1]**o+(0.0479609386435)*x[1]+(-0.968033045837)*x[2]**o+(0.199490071238)*x[2]
        arg[2,3,0,0]=(0.967996364186)*x[0]**o+(0.562562580668)*x[0]+(0.660658005197)*x[1]**o+(-0.00126241998808)*x[1]+(-0.0188273819622)*x[2]**o+(-0.798370566647)*x[2]
        arg[2,3,0,1]=(0.108137383906)*x[0]**o+(0.230815090379)*x[0]+(-0.0817434914408)*x[1]**o+(0.256977068172)*x[1]+(0.323454379512)*x[2]**o+(0.0382049771532)*x[2]
        arg[2,3,1,0]=(-0.279443973124)*x[0]**o+(-0.948014830608)*x[0]+(-0.237787814838)*x[1]**o+(-0.880632864876)*x[1]+(0.305230465688)*x[2]**o+(0.722139326346)*x[2]
        arg[2,3,1,1]=(-0.192752669063)*x[0]**o+(0.763575125553)*x[0]+(-0.317096602795)*x[1]**o+(-0.241479822867)*x[1]+(0.643154065804)*x[2]**o+(0.903250562165)*x[2]
        arg[2,3,2,0]=(0.58153852366)*x[0]**o+(-0.131284078682)*x[0]+(0.803022288461)*x[1]**o+(0.862074208938)*x[1]+(0.934648335783)*x[2]**o+(0.720337200812)*x[2]
        arg[2,3,2,1]=(-0.866755142782)*x[0]**o+(0.299995782189)*x[0]+(0.830022439718)*x[1]**o+(0.787775630269)*x[1]+(-0.436980900353)*x[2]**o+(0.436719598692)*x[2]
        arg[2,4,0,0]=(-0.416946111972)*x[0]**o+(-0.342692819134)*x[0]+(-0.265712927184)*x[1]**o+(-0.503154497528)*x[1]+(0.436447837569)*x[2]**o+(-0.514226068117)*x[2]
        arg[2,4,0,1]=(-0.915995315837)*x[0]**o+(-0.562870475868)*x[0]+(0.158881525561)*x[1]**o+(-0.458461594412)*x[1]+(-0.117927681987)*x[2]**o+(-0.697751714189)*x[2]
        arg[2,4,1,0]=(-0.961291698815)*x[0]**o+(0.457487466485)*x[0]+(0.170987701633)*x[1]**o+(-0.14226156627)*x[1]+(-0.663462703057)*x[2]**o+(-0.0307934454283)*x[2]
        arg[2,4,1,1]=(0.588162542138)*x[0]**o+(-0.761783611043)*x[0]+(0.510241246326)*x[1]**o+(0.805053446677)*x[1]+(-0.851391511147)*x[2]**o+(-0.0976037736598)*x[2]
        arg[2,4,2,0]=(0.649047985858)*x[0]**o+(0.677435498282)*x[0]+(-0.620777977748)*x[1]**o+(-0.209440794016)*x[1]+(0.58009322661)*x[2]**o+(0.56317180284)*x[2]
        arg[2,4,2,1]=(-0.975596464331)*x[0]**o+(0.759364271658)*x[0]+(0.43142780149)*x[1]**o+(-0.701043203786)*x[1]+(-0.648159388134)*x[2]**o+(-0.129618261144)*x[2]
        arg[3,0,0,0]=(-0.0207100058532)*x[0]**o+(-0.829637327952)*x[0]+(-0.0326040806041)*x[1]**o+(0.652320712227)*x[1]+(-0.509435869246)*x[2]**o+(-0.66044392521)*x[2]
        arg[3,0,0,1]=(-0.312677151974)*x[0]**o+(0.0204494340128)*x[0]+(-0.989457740146)*x[1]**o+(-0.0606077687298)*x[1]+(-0.527267756606)*x[2]**o+(0.496006274201)*x[2]
        arg[3,0,1,0]=(0.842871894549)*x[0]**o+(-0.577687338895)*x[0]+(0.414019506131)*x[1]**o+(0.0410521331015)*x[1]+(-0.830482660184)*x[2]**o+(0.688692187619)*x[2]
        arg[3,0,1,1]=(-0.920070939376)*x[0]**o+(0.632972952755)*x[0]+(0.944186412149)*x[1]**o+(0.620319712728)*x[1]+(-0.171433727161)*x[2]**o+(0.267125744191)*x[2]
        arg[3,0,2,0]=(-0.0669535338492)*x[0]**o+(0.249181230746)*x[0]+(-0.147448553328)*x[1]**o+(-0.459701695839)*x[1]+(0.378458172384)*x[2]**o+(0.916594147889)*x[2]
        arg[3,0,2,1]=(0.0720111269432)*x[0]**o+(-0.390032127319)*x[0]+(-0.115565740734)*x[1]**o+(0.501210832815)*x[1]+(0.568129023658)*x[2]**o+(-0.303909080042)*x[2]
        arg[3,1,0,0]=(0.205675480731)*x[0]**o+(-0.843093328795)*x[0]+(-0.812415573635)*x[1]**o+(-0.706806177394)*x[1]+(-0.158442718286)*x[2]**o+(0.15187176613)*x[2]
        arg[3,1,0,1]=(0.149669308017)*x[0]**o+(0.487445503149)*x[0]+(-0.27973905833)*x[1]**o+(-0.504992462225)*x[1]+(-0.728166691522)*x[2]**o+(0.734846267391)*x[2]
        arg[3,1,1,0]=(0.0410578015251)*x[0]**o+(-0.525001495676)*x[0]+(-0.512685697472)*x[1]**o+(-0.284314686987)*x[1]+(0.888112049081)*x[2]**o+(0.425466793802)*x[2]
        arg[3,1,1,1]=(-0.374237047043)*x[0]**o+(-0.00078724690461)*x[0]+(0.849952487348)*x[1]**o+(-0.654945893799)*x[1]+(-0.536739132177)*x[2]**o+(0.20919294066)*x[2]
        arg[3,1,2,0]=(0.685749084068)*x[0]**o+(-0.730378478543)*x[0]+(-0.918072746761)*x[1]**o+(-0.706713872062)*x[1]+(-0.486100957526)*x[2]**o+(0.95736939257)*x[2]
        arg[3,1,2,1]=(-0.287859736321)*x[0]**o+(-0.271882215207)*x[0]+(-0.673357024623)*x[1]**o+(0.702867457627)*x[1]+(0.78445294416)*x[2]**o+(0.613888365259)*x[2]
        arg[3,2,0,0]=(0.568506008135)*x[0]**o+(-0.825947283734)*x[0]+(0.687369924271)*x[1]**o+(-0.866597782569)*x[1]+(-0.332810902355)*x[2]**o+(0.499502049139)*x[2]
        arg[3,2,0,1]=(0.757251773732)*x[0]**o+(0.293337812465)*x[0]+(0.686709285382)*x[1]**o+(-0.084997189402)*x[1]+(-0.338903817381)*x[2]**o+(-0.787646799698)*x[2]
        arg[3,2,1,0]=(-0.580327419983)*x[0]**o+(-0.603952498996)*x[0]+(0.523057988618)*x[1]**o+(-0.0835187394107)*x[1]+(0.459827887147)*x[2]**o+(0.689021201042)*x[2]
        arg[3,2,1,1]=(-0.41923416619)*x[0]**o+(0.0406426957237)*x[0]+(-0.908041008186)*x[1]**o+(0.382097138462)*x[1]+(0.66835706849)*x[2]**o+(0.355849275378)*x[2]
        arg[3,2,2,0]=(-0.133756767108)*x[0]**o+(-0.233743496333)*x[0]+(-0.661420120856)*x[1]**o+(0.788722576584)*x[1]+(-0.131764585707)*x[2]**o+(-0.781004479871)*x[2]
        arg[3,2,2,1]=(0.0308645139784)*x[0]**o+(-0.194986246344)*x[0]+(-0.267418187303)*x[1]**o+(0.332143733168)*x[1]+(0.513266431894)*x[2]**o+(0.594016981473)*x[2]
        arg[3,3,0,0]=(0.644314102613)*x[0]**o+(-0.977355279476)*x[0]+(0.516452323667)*x[1]**o+(0.172873579919)*x[1]+(0.778904819969)*x[2]**o+(0.559689750425)*x[2]
        arg[3,3,0,1]=(0.736135986201)*x[0]**o+(0.366918250826)*x[0]+(-0.0547007988818)*x[1]**o+(-0.830174436433)*x[1]+(0.613440124338)*x[2]**o+(0.574992469591)*x[2]
        arg[3,3,1,0]=(0.324504938146)*x[0]**o+(-0.915240176689)*x[0]+(-0.398105336249)*x[1]**o+(-0.526498343533)*x[1]+(0.43264573336)*x[2]**o+(0.393008086529)*x[2]
        arg[3,3,1,1]=(0.926367438323)*x[0]**o+(0.526364305462)*x[0]+(0.15596582236)*x[1]**o+(0.165158439711)*x[1]+(-0.508505348973)*x[2]**o+(-0.603637833992)*x[2]
        arg[3,3,2,0]=(0.661333046004)*x[0]**o+(0.425320373512)*x[0]+(-0.810809051016)*x[1]**o+(0.20446843023)*x[1]+(0.685136412413)*x[2]**o+(-0.0281127019894)*x[2]
        arg[3,3,2,1]=(0.851742499007)*x[0]**o+(-0.405506974975)*x[0]+(-0.296768983731)*x[1]**o+(0.389557354929)*x[1]+(0.335108341463)*x[2]**o+(0.0313094433351)*x[2]
        arg[3,4,0,0]=(0.0415035498649)*x[0]**o+(0.827106496882)*x[0]+(-0.481831487377)*x[1]**o+(-0.14935685817)*x[1]+(0.354765095566)*x[2]**o+(0.521898910137)*x[2]
        arg[3,4,0,1]=(-0.36328099035)*x[0]**o+(-0.243342386595)*x[0]+(0.0475935188475)*x[1]**o+(0.845308417697)*x[1]+(0.315259620082)*x[2]**o+(0.541313742759)*x[2]
        arg[3,4,1,0]=(-0.66364149672)*x[0]**o+(0.987505978353)*x[0]+(-0.337659775957)*x[1]**o+(0.197277151989)*x[1]+(-0.431260870605)*x[2]**o+(-0.803375909471)*x[2]
        arg[3,4,1,1]=(0.722104365428)*x[0]**o+(0.168416682879)*x[0]+(-0.415740544748)*x[1]**o+(-0.437836696476)*x[1]+(-0.280581752598)*x[2]**o+(0.434326371762)*x[2]
        arg[3,4,2,0]=(0.673144889684)*x[0]**o+(-0.0628529591764)*x[0]+(0.0161458704408)*x[1]**o+(-0.439966404278)*x[1]+(-0.963402070447)*x[2]**o+(0.308312287021)*x[2]
        arg[3,4,2,1]=(0.649131384887)*x[0]**o+(-0.825086043001)*x[0]+(0.204615530143)*x[1]**o+(0.00801105023644)*x[1]+(-0.234754758061)*x[2]**o+(0.659650083864)*x[2]
        ref[0,0,0,0]=(-0.928996629952)*x_ref[0]**o+(-0.171383074114)*x_ref[0]+(-0.324423798901)*x_ref[1]**o+(-0.801354053962)*x_ref[1]+(0.365256142808)*x_ref[2]**o+(0.126262208857)*x_ref[2]
        ref[0,0,0,1]=(0.326024843626)*x_ref[0]**o+(-0.0091566240426)*x_ref[0]+(-0.0396235158589)*x_ref[1]**o+(-0.753302577837)*x_ref[1]+(-0.38709916587)*x_ref[2]**o+(-0.28660316595)*x_ref[2]
        ref[0,0,1,0]=(0.755338255506)*x_ref[0]**o+(0.851895434343)*x_ref[0]+(0.419119464402)*x_ref[1]**o+(0.670733085023)*x_ref[1]+(0.819398310958)*x_ref[2]**o+(0.476185335773)*x_ref[2]
        ref[0,0,1,1]=(-0.981416259131)*x_ref[0]**o+(-0.97000252267)*x_ref[0]+(-0.132624451699)*x_ref[1]**o+(-0.261428370649)*x_ref[1]+(-0.349245951727)*x_ref[2]**o+(-0.717343195877)*x_ref[2]
        ref[0,0,2,0]=(-0.874826075266)*x_ref[0]**o+(-0.399307905707)*x_ref[0]+(0.350873599317)*x_ref[1]**o+(0.15861651992)*x_ref[1]+(0.396279933216)*x_ref[2]**o+(0.92332515708)*x_ref[2]
        ref[0,0,2,1]=(-0.183596683527)*x_ref[0]**o+(-0.115194639348)*x_ref[0]+(0.616111723618)*x_ref[1]**o+(-0.402157657669)*x_ref[1]+(0.44145594193)*x_ref[2]**o+(0.686052407784)*x_ref[2]
        ref[0,1,0,0]=(-0.669169178244)*x_ref[0]**o+(0.835937801609)*x_ref[0]+(0.468805302181)*x_ref[1]**o+(-0.627292634023)*x_ref[1]+(-0.674361231979)*x_ref[2]**o+(-0.00929181785129)*x_ref[2]
        ref[0,1,0,1]=(-0.568131655705)*x_ref[0]**o+(0.690366535852)*x_ref[0]+(-0.212524396914)*x_ref[1]**o+(-0.767983769407)*x_ref[1]+(0.477500128569)*x_ref[2]**o+(-0.0190887598644)*x_ref[2]
        ref[0,1,1,0]=(-0.374170635822)*x_ref[0]**o+(0.0169009974558)*x_ref[0]+(-0.162519103357)*x_ref[1]**o+(0.786991144449)*x_ref[1]+(0.0590294918131)*x_ref[2]**o+(-0.213442509309)*x_ref[2]
        ref[0,1,1,1]=(-0.558268657439)*x_ref[0]**o+(0.584734933669)*x_ref[0]+(-0.650790013689)*x_ref[1]**o+(-0.192849048407)*x_ref[1]+(0.505458636714)*x_ref[2]**o+(-0.585174281371)*x_ref[2]
        ref[0,1,2,0]=(-0.933243590775)*x_ref[0]**o+(0.528060249767)*x_ref[0]+(-0.763255746252)*x_ref[1]**o+(0.771589118163)*x_ref[1]+(-0.641287604338)*x_ref[2]**o+(0.379566401585)*x_ref[2]
        ref[0,1,2,1]=(-0.836976373935)*x_ref[0]**o+(-0.969434110974)*x_ref[0]+(0.758148372888)*x_ref[1]**o+(-0.89212062535)*x_ref[1]+(0.788693820015)*x_ref[2]**o+(-0.620704367414)*x_ref[2]
        ref[0,2,0,0]=(0.0163117360814)*x_ref[0]**o+(-0.390867752068)*x_ref[0]+(-0.725068297982)*x_ref[1]**o+(-0.680312496646)*x_ref[1]+(0.069273494969)*x_ref[2]**o+(-0.576122275696)*x_ref[2]
        ref[0,2,0,1]=(0.461771197614)*x_ref[0]**o+(-0.767541404109)*x_ref[0]+(-0.694279263622)*x_ref[1]**o+(-0.721448812749)*x_ref[1]+(0.248787896584)*x_ref[2]**o+(0.237075401327)*x_ref[2]
        ref[0,2,1,0]=(0.381987012965)*x_ref[0]**o+(-0.288236233238)*x_ref[0]+(-0.519012025084)*x_ref[1]**o+(-0.256658293082)*x_ref[1]+(0.807322167368)*x_ref[2]**o+(-0.261916179859)*x_ref[2]
        ref[0,2,1,1]=(0.257241267407)*x_ref[0]**o+(-0.470612990721)*x_ref[0]+(-0.208612656448)*x_ref[1]**o+(0.839646962914)*x_ref[1]+(0.916675150494)*x_ref[2]**o+(-0.165468351005)*x_ref[2]
        ref[0,2,2,0]=(0.92569232521)*x_ref[0]**o+(0.809982458308)*x_ref[0]+(-0.860530305336)*x_ref[1]**o+(0.261687622817)*x_ref[1]+(0.948003684651)*x_ref[2]**o+(-0.473133742361)*x_ref[2]
        ref[0,2,2,1]=(-0.113846814874)*x_ref[0]**o+(0.680846459377)*x_ref[0]+(0.920219149741)*x_ref[1]**o+(0.172027554503)*x_ref[1]+(0.475330041115)*x_ref[2]**o+(-0.176218891429)*x_ref[2]
        ref[0,3,0,0]=(-0.948638294297)*x_ref[0]**o+(-0.0411621070703)*x_ref[0]+(-0.294759271027)*x_ref[1]**o+(-0.903733455086)*x_ref[1]+(-0.995253733593)*x_ref[2]**o+(0.106726960462)*x_ref[2]
        ref[0,3,0,1]=(0.1592709614)*x_ref[0]**o+(0.0683266445293)*x_ref[0]+(0.336092506745)*x_ref[1]**o+(-0.209047739881)*x_ref[1]+(-0.330417243459)*x_ref[2]**o+(-0.809242647755)*x_ref[2]
        ref[0,3,1,0]=(-0.680637516359)*x_ref[0]**o+(-0.713405815404)*x_ref[0]+(-0.484624435043)*x_ref[1]**o+(-0.279314462332)*x_ref[1]+(-0.458856066504)*x_ref[2]**o+(0.0588111305227)*x_ref[2]
        ref[0,3,1,1]=(-0.853676338655)*x_ref[0]**o+(-0.651539814195)*x_ref[0]+(-0.109727277725)*x_ref[1]**o+(0.0139006407942)*x_ref[1]+(0.231943245426)*x_ref[2]**o+(0.42972517233)*x_ref[2]
        ref[0,3,2,0]=(-0.60799269765)*x_ref[0]**o+(-0.802413269128)*x_ref[0]+(0.151362498863)*x_ref[1]**o+(-0.678226252875)*x_ref[1]+(-0.867117685659)*x_ref[2]**o+(-0.711982116071)*x_ref[2]
        ref[0,3,2,1]=(-0.243012743941)*x_ref[0]**o+(-0.692614579016)*x_ref[0]+(0.0986817641914)*x_ref[1]**o+(-0.495735862695)*x_ref[1]+(-0.430040681653)*x_ref[2]**o+(0.0108928722284)*x_ref[2]
        ref[0,4,0,0]=(0.0989402181558)*x_ref[0]**o+(-0.60096719882)*x_ref[0]+(-0.324253928432)*x_ref[1]**o+(-0.862723687559)*x_ref[1]+(-0.34745416427)*x_ref[2]**o+(0.936559074873)*x_ref[2]
        ref[0,4,0,1]=(0.711021967566)*x_ref[0]**o+(-0.939408401113)*x_ref[0]+(-0.218100921633)*x_ref[1]**o+(0.92835194875)*x_ref[1]+(-0.353936913865)*x_ref[2]**o+(-0.547266090928)*x_ref[2]
        ref[0,4,1,0]=(0.172231614188)*x_ref[0]**o+(-0.734622090671)*x_ref[0]+(0.432686916647)*x_ref[1]**o+(0.552595350492)*x_ref[1]+(0.779662641076)*x_ref[2]**o+(-0.0933220492779)*x_ref[2]
        ref[0,4,1,1]=(0.996776716866)*x_ref[0]**o+(-0.340786520686)*x_ref[0]+(0.260075214171)*x_ref[1]**o+(0.00385504852496)*x_ref[1]+(0.377516772127)*x_ref[2]**o+(-0.341751634181)*x_ref[2]
        ref[0,4,2,0]=(-0.655642896664)*x_ref[0]**o+(-0.840525173737)*x_ref[0]+(-0.855343721564)*x_ref[1]**o+(0.253558571524)*x_ref[1]+(0.0463733059925)*x_ref[2]**o+(-0.116359866912)*x_ref[2]
        ref[0,4,2,1]=(0.131619661414)*x_ref[0]**o+(-0.162067255954)*x_ref[0]+(0.562536831367)*x_ref[1]**o+(-0.940412472035)*x_ref[1]+(-0.282265762783)*x_ref[2]**o+(0.634711607923)*x_ref[2]
        ref[1,0,0,0]=(0.484071808097)*x_ref[0]**o+(0.566763524691)*x_ref[0]+(-0.658614173653)*x_ref[1]**o+(-0.140192232895)*x_ref[1]+(-0.310114236409)*x_ref[2]**o+(-0.918532401595)*x_ref[2]
        ref[1,0,0,1]=(-0.81190380441)*x_ref[0]**o+(0.644439878497)*x_ref[0]+(0.479583403289)*x_ref[1]**o+(-0.493318882779)*x_ref[1]+(-0.424704492175)*x_ref[2]**o+(0.700624317732)*x_ref[2]
        ref[1,0,1,0]=(0.766650182435)*x_ref[0]**o+(0.226365545129)*x_ref[0]+(0.510828945971)*x_ref[1]**o+(0.283221954498)*x_ref[1]+(0.817303877886)*x_ref[2]**o+(-0.245709044628)*x_ref[2]
        ref[1,0,1,1]=(0.194787112664)*x_ref[0]**o+(-0.879306736922)*x_ref[0]+(0.185381291627)*x_ref[1]**o+(-0.510692743975)*x_ref[1]+(-0.422555014055)*x_ref[2]**o+(0.556629919959)*x_ref[2]
        ref[1,0,2,0]=(-0.958345828895)*x_ref[0]**o+(-0.667735551013)*x_ref[0]+(0.278271284225)*x_ref[1]**o+(-0.0847995194757)*x_ref[1]+(0.902184171441)*x_ref[2]**o+(0.763125168854)*x_ref[2]
        ref[1,0,2,1]=(-0.0766429670195)*x_ref[0]**o+(0.894421139029)*x_ref[0]+(-0.331560314101)*x_ref[1]**o+(-0.0253297623997)*x_ref[1]+(0.13406045967)*x_ref[2]**o+(-0.326681111625)*x_ref[2]
        ref[1,1,0,0]=(-0.979406855838)*x_ref[0]**o+(0.528849909402)*x_ref[0]+(-0.0486159476737)*x_ref[1]**o+(-0.898392512695)*x_ref[1]+(-0.678328327407)*x_ref[2]**o+(0.644201802057)*x_ref[2]
        ref[1,1,0,1]=(-0.977485554605)*x_ref[0]**o+(0.794799941538)*x_ref[0]+(-0.00237183848504)*x_ref[1]**o+(-0.633497483149)*x_ref[1]+(0.746593853568)*x_ref[2]**o+(0.197741126365)*x_ref[2]
        ref[1,1,1,0]=(0.379985506182)*x_ref[0]**o+(0.69889144234)*x_ref[0]+(-0.507293887409)*x_ref[1]**o+(-0.727675545291)*x_ref[1]+(-0.467064531693)*x_ref[2]**o+(0.130273930089)*x_ref[2]
        ref[1,1,1,1]=(-0.330627830819)*x_ref[0]**o+(-0.0311542935889)*x_ref[0]+(-0.0671494148749)*x_ref[1]**o+(0.583605469163)*x_ref[1]+(-0.172083996069)*x_ref[2]**o+(0.666718011074)*x_ref[2]
        ref[1,1,2,0]=(0.539618381063)*x_ref[0]**o+(-0.546796237888)*x_ref[0]+(0.857911750392)*x_ref[1]**o+(0.119552731463)*x_ref[1]+(0.329286820109)*x_ref[2]**o+(0.637240649087)*x_ref[2]
        ref[1,1,2,1]=(-0.828603071334)*x_ref[0]**o+(0.861799149065)*x_ref[0]+(0.615224040538)*x_ref[1]**o+(-0.0868985543263)*x_ref[1]+(-0.206168997809)*x_ref[2]**o+(0.311103719286)*x_ref[2]
        ref[1,2,0,0]=(0.695106867715)*x_ref[0]**o+(-0.533303418727)*x_ref[0]+(-0.393936955207)*x_ref[1]**o+(-0.407635261686)*x_ref[1]+(0.604346269147)*x_ref[2]**o+(-0.10213061533)*x_ref[2]
        ref[1,2,0,1]=(0.338035309929)*x_ref[0]**o+(0.517135412671)*x_ref[0]+(0.277863735829)*x_ref[1]**o+(-0.703417079398)*x_ref[1]+(0.369398635984)*x_ref[2]**o+(-0.952905070601)*x_ref[2]
        ref[1,2,1,0]=(0.937866603297)*x_ref[0]**o+(0.436121631708)*x_ref[0]+(0.568362098461)*x_ref[1]**o+(0.223837722984)*x_ref[1]+(-0.236896910941)*x_ref[2]**o+(0.316114408308)*x_ref[2]
        ref[1,2,1,1]=(0.285714491932)*x_ref[0]**o+(0.633321537987)*x_ref[0]+(0.953737237202)*x_ref[1]**o+(0.732065555924)*x_ref[1]+(-0.515759068197)*x_ref[2]**o+(-0.519775334889)*x_ref[2]
        ref[1,2,2,0]=(-0.0610222303479)*x_ref[0]**o+(0.135974983944)*x_ref[0]+(0.450213393447)*x_ref[1]**o+(-0.44747541715)*x_ref[1]+(-0.640660207187)*x_ref[2]**o+(0.348457261119)*x_ref[2]
        ref[1,2,2,1]=(0.131753033369)*x_ref[0]**o+(-0.311359549998)*x_ref[0]+(0.445707767235)*x_ref[1]**o+(-0.710795569938)*x_ref[1]+(-0.746038648477)*x_ref[2]**o+(-0.210414773188)*x_ref[2]
        ref[1,3,0,0]=(-0.323684061289)*x_ref[0]**o+(-0.0937091220282)*x_ref[0]+(-0.827884898323)*x_ref[1]**o+(0.604442907568)*x_ref[1]+(0.620923596379)*x_ref[2]**o+(0.851758170763)*x_ref[2]
        ref[1,3,0,1]=(-0.269239872573)*x_ref[0]**o+(-0.603248280988)*x_ref[0]+(0.603804530326)*x_ref[1]**o+(-0.331910095046)*x_ref[1]+(-0.653877236308)*x_ref[2]**o+(0.115434865132)*x_ref[2]
        ref[1,3,1,0]=(0.0531798412869)*x_ref[0]**o+(0.292950346899)*x_ref[0]+(-0.634376081823)*x_ref[1]**o+(-0.775427771578)*x_ref[1]+(-0.987113467149)*x_ref[2]**o+(0.757980444447)*x_ref[2]
        ref[1,3,1,1]=(0.757807292214)*x_ref[0]**o+(0.81778478971)*x_ref[0]+(-0.619873148405)*x_ref[1]**o+(0.0129855659138)*x_ref[1]+(-0.282861093681)*x_ref[2]**o+(0.667211104045)*x_ref[2]
        ref[1,3,2,0]=(-0.915205606853)*x_ref[0]**o+(0.918704643459)*x_ref[0]+(0.143238155504)*x_ref[1]**o+(0.0285686377214)*x_ref[1]+(-0.848712843243)*x_ref[2]**o+(0.058647124862)*x_ref[2]
        ref[1,3,2,1]=(0.441487712831)*x_ref[0]**o+(-0.626878272507)*x_ref[0]+(-0.862579662295)*x_ref[1]**o+(0.204629243832)*x_ref[1]+(0.964143347265)*x_ref[2]**o+(-0.399125844282)*x_ref[2]
        ref[1,4,0,0]=(-0.349791275126)*x_ref[0]**o+(-0.140293844776)*x_ref[0]+(-0.693879671138)*x_ref[1]**o+(0.429290435637)*x_ref[1]+(-0.892670314048)*x_ref[2]**o+(-0.380777380409)*x_ref[2]
        ref[1,4,0,1]=(0.528712424647)*x_ref[0]**o+(0.449268002671)*x_ref[0]+(-0.202472214961)*x_ref[1]**o+(-0.374762208851)*x_ref[1]+(-0.625590920648)*x_ref[2]**o+(-0.000776205555531)*x_ref[2]
        ref[1,4,1,0]=(0.00425881397184)*x_ref[0]**o+(0.265831686248)*x_ref[0]+(0.757626547866)*x_ref[1]**o+(-0.934207064756)*x_ref[1]+(0.541356932538)*x_ref[2]**o+(0.217841597089)*x_ref[2]
        ref[1,4,1,1]=(-0.02422207106)*x_ref[0]**o+(-0.807018149115)*x_ref[0]+(-0.402347046418)*x_ref[1]**o+(-0.0548851919048)*x_ref[1]+(0.850202044077)*x_ref[2]**o+(-0.889576261979)*x_ref[2]
        ref[1,4,2,0]=(0.487059068561)*x_ref[0]**o+(0.338254525003)*x_ref[0]+(-0.387019663154)*x_ref[1]**o+(-0.436969552236)*x_ref[1]+(-0.747690766603)*x_ref[2]**o+(0.265854907003)*x_ref[2]
        ref[1,4,2,1]=(0.687562221132)*x_ref[0]**o+(0.4755782976)*x_ref[0]+(-0.456287111804)*x_ref[1]**o+(0.412138123928)*x_ref[1]+(0.637920334647)*x_ref[2]**o+(0.817007725525)*x_ref[2]
        ref[2,0,0,0]=(-0.728166422776)*x_ref[0]**o+(0.108904177195)*x_ref[0]+(0.383848108635)*x_ref[1]**o+(-0.19252707292)*x_ref[1]+(-0.279319122495)*x_ref[2]**o+(-0.0530727745585)*x_ref[2]
        ref[2,0,0,1]=(0.380827452272)*x_ref[0]**o+(-0.133776280687)*x_ref[0]+(-0.174731551283)*x_ref[1]**o+(-0.369231086652)*x_ref[1]+(-0.0190771818046)*x_ref[2]**o+(-0.331873068919)*x_ref[2]
        ref[2,0,1,0]=(0.653644298826)*x_ref[0]**o+(-0.916110182689)*x_ref[0]+(0.0510855252492)*x_ref[1]**o+(-0.934192949208)*x_ref[1]+(0.572048462784)*x_ref[2]**o+(-0.451401856863)*x_ref[2]
        ref[2,0,1,1]=(-0.394958497502)*x_ref[0]**o+(-0.690752972227)*x_ref[0]+(-0.436352069349)*x_ref[1]**o+(0.933524576573)*x_ref[1]+(-0.524344161699)*x_ref[2]**o+(0.526748792776)*x_ref[2]
        ref[2,0,2,0]=(0.845649523782)*x_ref[0]**o+(0.63927747558)*x_ref[0]+(-0.534192218216)*x_ref[1]**o+(0.714500925459)*x_ref[1]+(0.156077814898)*x_ref[2]**o+(-0.193712778792)*x_ref[2]
        ref[2,0,2,1]=(0.809636640413)*x_ref[0]**o+(-0.930654037481)*x_ref[0]+(0.509441692942)*x_ref[1]**o+(0.881563824585)*x_ref[1]+(0.101333749057)*x_ref[2]**o+(0.418791722704)*x_ref[2]
        ref[2,1,0,0]=(0.289481281745)*x_ref[0]**o+(-0.98279193802)*x_ref[0]+(-0.443382747319)*x_ref[1]**o+(-0.281710992472)*x_ref[1]+(0.426349350391)*x_ref[2]**o+(-0.870192121251)*x_ref[2]
        ref[2,1,0,1]=(0.837046002753)*x_ref[0]**o+(-0.0254166983371)*x_ref[0]+(0.30178635814)*x_ref[1]**o+(0.973915444924)*x_ref[1]+(0.298631524734)*x_ref[2]**o+(-0.422490970248)*x_ref[2]
        ref[2,1,1,0]=(0.869799082394)*x_ref[0]**o+(0.376436513628)*x_ref[0]+(-0.74735679118)*x_ref[1]**o+(0.477184335901)*x_ref[1]+(-0.797472924916)*x_ref[2]**o+(-0.568827873964)*x_ref[2]
        ref[2,1,1,1]=(0.993441009779)*x_ref[0]**o+(-0.264888299493)*x_ref[0]+(0.861197560367)*x_ref[1]**o+(-0.769058084144)*x_ref[1]+(-0.86221947415)*x_ref[2]**o+(-0.617024169672)*x_ref[2]
        ref[2,1,2,0]=(-0.89303691749)*x_ref[0]**o+(-0.390287412175)*x_ref[0]+(-0.312883084291)*x_ref[1]**o+(-0.432368502641)*x_ref[1]+(-0.121476666357)*x_ref[2]**o+(0.416361314288)*x_ref[2]
        ref[2,1,2,1]=(-0.622214426504)*x_ref[0]**o+(-0.595279134338)*x_ref[0]+(0.000927173498715)*x_ref[1]**o+(0.783001179117)*x_ref[1]+(-0.956003738061)*x_ref[2]**o+(0.482321214492)*x_ref[2]
        ref[2,2,0,0]=(-0.0047994639068)*x_ref[0]**o+(0.478805093492)*x_ref[0]+(0.747825846519)*x_ref[1]**o+(0.0756064430807)*x_ref[1]+(-0.470534826525)*x_ref[2]**o+(-0.421524571831)*x_ref[2]
        ref[2,2,0,1]=(-0.0569198078126)*x_ref[0]**o+(-0.36284856463)*x_ref[0]+(-0.374343856016)*x_ref[1]**o+(0.193470734451)*x_ref[1]+(0.425350990791)*x_ref[2]**o+(-0.903309757487)*x_ref[2]
        ref[2,2,1,0]=(0.602075251229)*x_ref[0]**o+(-0.558163518701)*x_ref[0]+(-0.00146699467501)*x_ref[1]**o+(0.437116756605)*x_ref[1]+(-0.252696658318)*x_ref[2]**o+(0.154017654456)*x_ref[2]
        ref[2,2,1,1]=(0.226970288166)*x_ref[0]**o+(0.844076458869)*x_ref[0]+(0.231874137232)*x_ref[1]**o+(-0.270233945335)*x_ref[1]+(0.365979082276)*x_ref[2]**o+(-0.937144802776)*x_ref[2]
        ref[2,2,2,0]=(0.662440813722)*x_ref[0]**o+(-0.600830429698)*x_ref[0]+(0.248436486724)*x_ref[1]**o+(0.73439963812)*x_ref[1]+(-0.406317484259)*x_ref[2]**o+(-0.406641354653)*x_ref[2]
        ref[2,2,2,1]=(0.948630940404)*x_ref[0]**o+(-0.695098513788)*x_ref[0]+(-0.819843883615)*x_ref[1]**o+(0.0479609386435)*x_ref[1]+(-0.968033045837)*x_ref[2]**o+(0.199490071238)*x_ref[2]
        ref[2,3,0,0]=(0.967996364186)*x_ref[0]**o+(0.562562580668)*x_ref[0]+(0.660658005197)*x_ref[1]**o+(-0.00126241998808)*x_ref[1]+(-0.0188273819622)*x_ref[2]**o+(-0.798370566647)*x_ref[2]
        ref[2,3,0,1]=(0.108137383906)*x_ref[0]**o+(0.230815090379)*x_ref[0]+(-0.0817434914408)*x_ref[1]**o+(0.256977068172)*x_ref[1]+(0.323454379512)*x_ref[2]**o+(0.0382049771532)*x_ref[2]
        ref[2,3,1,0]=(-0.279443973124)*x_ref[0]**o+(-0.948014830608)*x_ref[0]+(-0.237787814838)*x_ref[1]**o+(-0.880632864876)*x_ref[1]+(0.305230465688)*x_ref[2]**o+(0.722139326346)*x_ref[2]
        ref[2,3,1,1]=(-0.192752669063)*x_ref[0]**o+(0.763575125553)*x_ref[0]+(-0.317096602795)*x_ref[1]**o+(-0.241479822867)*x_ref[1]+(0.643154065804)*x_ref[2]**o+(0.903250562165)*x_ref[2]
        ref[2,3,2,0]=(0.58153852366)*x_ref[0]**o+(-0.131284078682)*x_ref[0]+(0.803022288461)*x_ref[1]**o+(0.862074208938)*x_ref[1]+(0.934648335783)*x_ref[2]**o+(0.720337200812)*x_ref[2]
        ref[2,3,2,1]=(-0.866755142782)*x_ref[0]**o+(0.299995782189)*x_ref[0]+(0.830022439718)*x_ref[1]**o+(0.787775630269)*x_ref[1]+(-0.436980900353)*x_ref[2]**o+(0.436719598692)*x_ref[2]
        ref[2,4,0,0]=(-0.416946111972)*x_ref[0]**o+(-0.342692819134)*x_ref[0]+(-0.265712927184)*x_ref[1]**o+(-0.503154497528)*x_ref[1]+(0.436447837569)*x_ref[2]**o+(-0.514226068117)*x_ref[2]
        ref[2,4,0,1]=(-0.915995315837)*x_ref[0]**o+(-0.562870475868)*x_ref[0]+(0.158881525561)*x_ref[1]**o+(-0.458461594412)*x_ref[1]+(-0.117927681987)*x_ref[2]**o+(-0.697751714189)*x_ref[2]
        ref[2,4,1,0]=(-0.961291698815)*x_ref[0]**o+(0.457487466485)*x_ref[0]+(0.170987701633)*x_ref[1]**o+(-0.14226156627)*x_ref[1]+(-0.663462703057)*x_ref[2]**o+(-0.0307934454283)*x_ref[2]
        ref[2,4,1,1]=(0.588162542138)*x_ref[0]**o+(-0.761783611043)*x_ref[0]+(0.510241246326)*x_ref[1]**o+(0.805053446677)*x_ref[1]+(-0.851391511147)*x_ref[2]**o+(-0.0976037736598)*x_ref[2]
        ref[2,4,2,0]=(0.649047985858)*x_ref[0]**o+(0.677435498282)*x_ref[0]+(-0.620777977748)*x_ref[1]**o+(-0.209440794016)*x_ref[1]+(0.58009322661)*x_ref[2]**o+(0.56317180284)*x_ref[2]
        ref[2,4,2,1]=(-0.975596464331)*x_ref[0]**o+(0.759364271658)*x_ref[0]+(0.43142780149)*x_ref[1]**o+(-0.701043203786)*x_ref[1]+(-0.648159388134)*x_ref[2]**o+(-0.129618261144)*x_ref[2]
        ref[3,0,0,0]=(-0.0207100058532)*x_ref[0]**o+(-0.829637327952)*x_ref[0]+(-0.0326040806041)*x_ref[1]**o+(0.652320712227)*x_ref[1]+(-0.509435869246)*x_ref[2]**o+(-0.66044392521)*x_ref[2]
        ref[3,0,0,1]=(-0.312677151974)*x_ref[0]**o+(0.0204494340128)*x_ref[0]+(-0.989457740146)*x_ref[1]**o+(-0.0606077687298)*x_ref[1]+(-0.527267756606)*x_ref[2]**o+(0.496006274201)*x_ref[2]
        ref[3,0,1,0]=(0.842871894549)*x_ref[0]**o+(-0.577687338895)*x_ref[0]+(0.414019506131)*x_ref[1]**o+(0.0410521331015)*x_ref[1]+(-0.830482660184)*x_ref[2]**o+(0.688692187619)*x_ref[2]
        ref[3,0,1,1]=(-0.920070939376)*x_ref[0]**o+(0.632972952755)*x_ref[0]+(0.944186412149)*x_ref[1]**o+(0.620319712728)*x_ref[1]+(-0.171433727161)*x_ref[2]**o+(0.267125744191)*x_ref[2]
        ref[3,0,2,0]=(-0.0669535338492)*x_ref[0]**o+(0.249181230746)*x_ref[0]+(-0.147448553328)*x_ref[1]**o+(-0.459701695839)*x_ref[1]+(0.378458172384)*x_ref[2]**o+(0.916594147889)*x_ref[2]
        ref[3,0,2,1]=(0.0720111269432)*x_ref[0]**o+(-0.390032127319)*x_ref[0]+(-0.115565740734)*x_ref[1]**o+(0.501210832815)*x_ref[1]+(0.568129023658)*x_ref[2]**o+(-0.303909080042)*x_ref[2]
        ref[3,1,0,0]=(0.205675480731)*x_ref[0]**o+(-0.843093328795)*x_ref[0]+(-0.812415573635)*x_ref[1]**o+(-0.706806177394)*x_ref[1]+(-0.158442718286)*x_ref[2]**o+(0.15187176613)*x_ref[2]
        ref[3,1,0,1]=(0.149669308017)*x_ref[0]**o+(0.487445503149)*x_ref[0]+(-0.27973905833)*x_ref[1]**o+(-0.504992462225)*x_ref[1]+(-0.728166691522)*x_ref[2]**o+(0.734846267391)*x_ref[2]
        ref[3,1,1,0]=(0.0410578015251)*x_ref[0]**o+(-0.525001495676)*x_ref[0]+(-0.512685697472)*x_ref[1]**o+(-0.284314686987)*x_ref[1]+(0.888112049081)*x_ref[2]**o+(0.425466793802)*x_ref[2]
        ref[3,1,1,1]=(-0.374237047043)*x_ref[0]**o+(-0.00078724690461)*x_ref[0]+(0.849952487348)*x_ref[1]**o+(-0.654945893799)*x_ref[1]+(-0.536739132177)*x_ref[2]**o+(0.20919294066)*x_ref[2]
        ref[3,1,2,0]=(0.685749084068)*x_ref[0]**o+(-0.730378478543)*x_ref[0]+(-0.918072746761)*x_ref[1]**o+(-0.706713872062)*x_ref[1]+(-0.486100957526)*x_ref[2]**o+(0.95736939257)*x_ref[2]
        ref[3,1,2,1]=(-0.287859736321)*x_ref[0]**o+(-0.271882215207)*x_ref[0]+(-0.673357024623)*x_ref[1]**o+(0.702867457627)*x_ref[1]+(0.78445294416)*x_ref[2]**o+(0.613888365259)*x_ref[2]
        ref[3,2,0,0]=(0.568506008135)*x_ref[0]**o+(-0.825947283734)*x_ref[0]+(0.687369924271)*x_ref[1]**o+(-0.866597782569)*x_ref[1]+(-0.332810902355)*x_ref[2]**o+(0.499502049139)*x_ref[2]
        ref[3,2,0,1]=(0.757251773732)*x_ref[0]**o+(0.293337812465)*x_ref[0]+(0.686709285382)*x_ref[1]**o+(-0.084997189402)*x_ref[1]+(-0.338903817381)*x_ref[2]**o+(-0.787646799698)*x_ref[2]
        ref[3,2,1,0]=(-0.580327419983)*x_ref[0]**o+(-0.603952498996)*x_ref[0]+(0.523057988618)*x_ref[1]**o+(-0.0835187394107)*x_ref[1]+(0.459827887147)*x_ref[2]**o+(0.689021201042)*x_ref[2]
        ref[3,2,1,1]=(-0.41923416619)*x_ref[0]**o+(0.0406426957237)*x_ref[0]+(-0.908041008186)*x_ref[1]**o+(0.382097138462)*x_ref[1]+(0.66835706849)*x_ref[2]**o+(0.355849275378)*x_ref[2]
        ref[3,2,2,0]=(-0.133756767108)*x_ref[0]**o+(-0.233743496333)*x_ref[0]+(-0.661420120856)*x_ref[1]**o+(0.788722576584)*x_ref[1]+(-0.131764585707)*x_ref[2]**o+(-0.781004479871)*x_ref[2]
        ref[3,2,2,1]=(0.0308645139784)*x_ref[0]**o+(-0.194986246344)*x_ref[0]+(-0.267418187303)*x_ref[1]**o+(0.332143733168)*x_ref[1]+(0.513266431894)*x_ref[2]**o+(0.594016981473)*x_ref[2]
        ref[3,3,0,0]=(0.644314102613)*x_ref[0]**o+(-0.977355279476)*x_ref[0]+(0.516452323667)*x_ref[1]**o+(0.172873579919)*x_ref[1]+(0.778904819969)*x_ref[2]**o+(0.559689750425)*x_ref[2]
        ref[3,3,0,1]=(0.736135986201)*x_ref[0]**o+(0.366918250826)*x_ref[0]+(-0.0547007988818)*x_ref[1]**o+(-0.830174436433)*x_ref[1]+(0.613440124338)*x_ref[2]**o+(0.574992469591)*x_ref[2]
        ref[3,3,1,0]=(0.324504938146)*x_ref[0]**o+(-0.915240176689)*x_ref[0]+(-0.398105336249)*x_ref[1]**o+(-0.526498343533)*x_ref[1]+(0.43264573336)*x_ref[2]**o+(0.393008086529)*x_ref[2]
        ref[3,3,1,1]=(0.926367438323)*x_ref[0]**o+(0.526364305462)*x_ref[0]+(0.15596582236)*x_ref[1]**o+(0.165158439711)*x_ref[1]+(-0.508505348973)*x_ref[2]**o+(-0.603637833992)*x_ref[2]
        ref[3,3,2,0]=(0.661333046004)*x_ref[0]**o+(0.425320373512)*x_ref[0]+(-0.810809051016)*x_ref[1]**o+(0.20446843023)*x_ref[1]+(0.685136412413)*x_ref[2]**o+(-0.0281127019894)*x_ref[2]
        ref[3,3,2,1]=(0.851742499007)*x_ref[0]**o+(-0.405506974975)*x_ref[0]+(-0.296768983731)*x_ref[1]**o+(0.389557354929)*x_ref[1]+(0.335108341463)*x_ref[2]**o+(0.0313094433351)*x_ref[2]
        ref[3,4,0,0]=(0.0415035498649)*x_ref[0]**o+(0.827106496882)*x_ref[0]+(-0.481831487377)*x_ref[1]**o+(-0.14935685817)*x_ref[1]+(0.354765095566)*x_ref[2]**o+(0.521898910137)*x_ref[2]
        ref[3,4,0,1]=(-0.36328099035)*x_ref[0]**o+(-0.243342386595)*x_ref[0]+(0.0475935188475)*x_ref[1]**o+(0.845308417697)*x_ref[1]+(0.315259620082)*x_ref[2]**o+(0.541313742759)*x_ref[2]
        ref[3,4,1,0]=(-0.66364149672)*x_ref[0]**o+(0.987505978353)*x_ref[0]+(-0.337659775957)*x_ref[1]**o+(0.197277151989)*x_ref[1]+(-0.431260870605)*x_ref[2]**o+(-0.803375909471)*x_ref[2]
        ref[3,4,1,1]=(0.722104365428)*x_ref[0]**o+(0.168416682879)*x_ref[0]+(-0.415740544748)*x_ref[1]**o+(-0.437836696476)*x_ref[1]+(-0.280581752598)*x_ref[2]**o+(0.434326371762)*x_ref[2]
        ref[3,4,2,0]=(0.673144889684)*x_ref[0]**o+(-0.0628529591764)*x_ref[0]+(0.0161458704408)*x_ref[1]**o+(-0.439966404278)*x_ref[1]+(-0.963402070447)*x_ref[2]**o+(0.308312287021)*x_ref[2]
        ref[3,4,2,1]=(0.649131384887)*x_ref[0]**o+(-0.825086043001)*x_ref[0]+(0.204615530143)*x_ref[1]**o+(0.00801105023644)*x_ref[1]+(-0.234754758061)*x_ref[2]**o+(0.659650083864)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.935068705503)*x[0]**o+(-0.843768568975)*x[0]+(0.220151879481)*x[1]**o+(0.629247769917)*x[1]
        ref=(-0.935068705503)*x_ref[0]**o+(-0.843768568975)*x_ref[0]+(0.220151879481)*x_ref[1]**o+(0.629247769917)*x_ref[1]
      else:
        arg=(0.204715133488)*x[0]**o+(-0.933589384176)*x[0]+(-0.651477613601)*x[1]**o+(-0.128099401086)*x[1]+(0.0507915334433)*x[2]**o+(0.63225604515)*x[2]
        ref=(0.204715133488)*x_ref[0]**o+(-0.933589384176)*x_ref[0]+(-0.651477613601)*x_ref[1]**o+(-0.128099401086)*x_ref[1]+(0.0507915334433)*x_ref[2]**o+(0.63225604515)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.693671772608)*x[0]**o+(0.260200526395)*x[0]+(-0.99202032752)*x[1]**o+(0.698503413133)*x[1]
        arg[1]=(-0.549448913905)*x[0]**o+(-0.728769679965)*x[0]+(0.415132041969)*x[1]**o+(-0.106391482954)*x[1]
        ref[0]=(0.693671772608)*x_ref[0]**o+(0.260200526395)*x_ref[0]+(-0.99202032752)*x_ref[1]**o+(0.698503413133)*x_ref[1]
        ref[1]=(-0.549448913905)*x_ref[0]**o+(-0.728769679965)*x_ref[0]+(0.415132041969)*x_ref[1]**o+(-0.106391482954)*x_ref[1]
      else:
        arg[0]=(0.479813279739)*x[0]**o+(0.940795516779)*x[0]+(-0.671143421639)*x[1]**o+(-0.485412476854)*x[1]+(-0.962114397463)*x[2]**o+(0.0250734884375)*x[2]
        arg[1]=(0.703481492279)*x[0]**o+(0.298972367312)*x[0]+(0.22386003834)*x[1]**o+(-0.726486516907)*x[1]+(0.0556568111911)*x[2]**o+(-0.603711754869)*x[2]
        ref[0]=(0.479813279739)*x_ref[0]**o+(0.940795516779)*x_ref[0]+(-0.671143421639)*x_ref[1]**o+(-0.485412476854)*x_ref[1]+(-0.962114397463)*x_ref[2]**o+(0.0250734884375)*x_ref[2]
        ref[1]=(0.703481492279)*x_ref[0]**o+(0.298972367312)*x_ref[0]+(0.22386003834)*x_ref[1]**o+(-0.726486516907)*x_ref[1]+(0.0556568111911)*x_ref[2]**o+(-0.603711754869)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.436439283545)*x[0]**o+(0.705849609672)*x[0]+(-0.261476237133)*x[1]**o+(-0.10504570237)*x[1]
        arg[0,1]=(0.035208834696)*x[0]**o+(0.590058421274)*x[0]+(-0.532905126944)*x[1]**o+(0.126572388319)*x[1]
        arg[0,2]=(0.985266120604)*x[0]**o+(-0.465951670003)*x[0]+(0.855881157118)*x[1]**o+(0.00932326399668)*x[1]
        arg[0,3]=(0.867525798771)*x[0]**o+(-0.480556561488)*x[0]+(-0.958556900472)*x[1]**o+(-0.764939529718)*x[1]
        arg[0,4]=(-0.51062937871)*x[0]**o+(-0.135415437913)*x[0]+(0.340077694039)*x[1]**o+(0.874347806916)*x[1]
        arg[1,0]=(0.636810684661)*x[0]**o+(0.774651780754)*x[0]+(-0.314130997812)*x[1]**o+(-0.422706240721)*x[1]
        arg[1,1]=(0.495122166216)*x[0]**o+(0.757099858142)*x[0]+(-0.25861054515)*x[1]**o+(0.612180732826)*x[1]
        arg[1,2]=(-0.909636123949)*x[0]**o+(-0.0434375137409)*x[0]+(0.100612174358)*x[1]**o+(0.787104124194)*x[1]
        arg[1,3]=(-0.499307726848)*x[0]**o+(0.013205984499)*x[0]+(0.858074302957)*x[1]**o+(-0.630926261217)*x[1]
        arg[1,4]=(-0.410815849264)*x[0]**o+(0.820939326325)*x[0]+(-0.415370205615)*x[1]**o+(0.0602438669371)*x[1]
        arg[2,0]=(0.642545275281)*x[0]**o+(0.872239548588)*x[0]+(0.446294659843)*x[1]**o+(0.929350720888)*x[1]
        arg[2,1]=(0.182218495486)*x[0]**o+(-0.393184556589)*x[0]+(0.42311545961)*x[1]**o+(-0.479636250878)*x[1]
        arg[2,2]=(0.874574618473)*x[0]**o+(0.879031610251)*x[0]+(0.686157301617)*x[1]**o+(0.706917527697)*x[1]
        arg[2,3]=(0.614189029029)*x[0]**o+(0.0973950074653)*x[0]+(0.626052996278)*x[1]**o+(0.75043535006)*x[1]
        arg[2,4]=(0.386209196978)*x[0]**o+(0.644169312279)*x[0]+(-0.554562346917)*x[1]**o+(-0.474292025255)*x[1]
        arg[3,0]=(0.015012073686)*x[0]**o+(-0.665804159938)*x[0]+(-0.849030247328)*x[1]**o+(-0.936779648988)*x[1]
        arg[3,1]=(-0.818194334698)*x[0]**o+(-0.569993362998)*x[0]+(-0.806235611766)*x[1]**o+(0.35347475408)*x[1]
        arg[3,2]=(-0.227117779134)*x[0]**o+(-0.981285175536)*x[0]+(-0.226804205345)*x[1]**o+(-0.0134258740334)*x[1]
        arg[3,3]=(0.126716245369)*x[0]**o+(0.135476076345)*x[0]+(-0.641421845681)*x[1]**o+(0.0595388928019)*x[1]
        arg[3,4]=(-0.658260624128)*x[0]**o+(0.124989215683)*x[0]+(-0.949146047463)*x[1]**o+(0.805277938593)*x[1]
        ref[0,0]=(0.436439283545)*x_ref[0]**o+(0.705849609672)*x_ref[0]+(-0.261476237133)*x_ref[1]**o+(-0.10504570237)*x_ref[1]
        ref[0,1]=(0.035208834696)*x_ref[0]**o+(0.590058421274)*x_ref[0]+(-0.532905126944)*x_ref[1]**o+(0.126572388319)*x_ref[1]
        ref[0,2]=(0.985266120604)*x_ref[0]**o+(-0.465951670003)*x_ref[0]+(0.855881157118)*x_ref[1]**o+(0.00932326399668)*x_ref[1]
        ref[0,3]=(0.867525798771)*x_ref[0]**o+(-0.480556561488)*x_ref[0]+(-0.958556900472)*x_ref[1]**o+(-0.764939529718)*x_ref[1]
        ref[0,4]=(-0.51062937871)*x_ref[0]**o+(-0.135415437913)*x_ref[0]+(0.340077694039)*x_ref[1]**o+(0.874347806916)*x_ref[1]
        ref[1,0]=(0.636810684661)*x_ref[0]**o+(0.774651780754)*x_ref[0]+(-0.314130997812)*x_ref[1]**o+(-0.422706240721)*x_ref[1]
        ref[1,1]=(0.495122166216)*x_ref[0]**o+(0.757099858142)*x_ref[0]+(-0.25861054515)*x_ref[1]**o+(0.612180732826)*x_ref[1]
        ref[1,2]=(-0.909636123949)*x_ref[0]**o+(-0.0434375137409)*x_ref[0]+(0.100612174358)*x_ref[1]**o+(0.787104124194)*x_ref[1]
        ref[1,3]=(-0.499307726848)*x_ref[0]**o+(0.013205984499)*x_ref[0]+(0.858074302957)*x_ref[1]**o+(-0.630926261217)*x_ref[1]
        ref[1,4]=(-0.410815849264)*x_ref[0]**o+(0.820939326325)*x_ref[0]+(-0.415370205615)*x_ref[1]**o+(0.0602438669371)*x_ref[1]
        ref[2,0]=(0.642545275281)*x_ref[0]**o+(0.872239548588)*x_ref[0]+(0.446294659843)*x_ref[1]**o+(0.929350720888)*x_ref[1]
        ref[2,1]=(0.182218495486)*x_ref[0]**o+(-0.393184556589)*x_ref[0]+(0.42311545961)*x_ref[1]**o+(-0.479636250878)*x_ref[1]
        ref[2,2]=(0.874574618473)*x_ref[0]**o+(0.879031610251)*x_ref[0]+(0.686157301617)*x_ref[1]**o+(0.706917527697)*x_ref[1]
        ref[2,3]=(0.614189029029)*x_ref[0]**o+(0.0973950074653)*x_ref[0]+(0.626052996278)*x_ref[1]**o+(0.75043535006)*x_ref[1]
        ref[2,4]=(0.386209196978)*x_ref[0]**o+(0.644169312279)*x_ref[0]+(-0.554562346917)*x_ref[1]**o+(-0.474292025255)*x_ref[1]
        ref[3,0]=(0.015012073686)*x_ref[0]**o+(-0.665804159938)*x_ref[0]+(-0.849030247328)*x_ref[1]**o+(-0.936779648988)*x_ref[1]
        ref[3,1]=(-0.818194334698)*x_ref[0]**o+(-0.569993362998)*x_ref[0]+(-0.806235611766)*x_ref[1]**o+(0.35347475408)*x_ref[1]
        ref[3,2]=(-0.227117779134)*x_ref[0]**o+(-0.981285175536)*x_ref[0]+(-0.226804205345)*x_ref[1]**o+(-0.0134258740334)*x_ref[1]
        ref[3,3]=(0.126716245369)*x_ref[0]**o+(0.135476076345)*x_ref[0]+(-0.641421845681)*x_ref[1]**o+(0.0595388928019)*x_ref[1]
        ref[3,4]=(-0.658260624128)*x_ref[0]**o+(0.124989215683)*x_ref[0]+(-0.949146047463)*x_ref[1]**o+(0.805277938593)*x_ref[1]
      else:
        arg[0,0]=(-0.859332605724)*x[0]**o+(-0.239005430756)*x[0]+(-0.52959833033)*x[1]**o+(0.725877930655)*x[1]+(-0.0348436114161)*x[2]**o+(0.47287744745)*x[2]
        arg[0,1]=(-0.0816886854463)*x[0]**o+(-0.626507446322)*x[0]+(-0.93804868959)*x[1]**o+(-0.450134311929)*x[1]+(0.0957946938715)*x[2]**o+(-0.236994496198)*x[2]
        arg[0,2]=(-0.669441310039)*x[0]**o+(-0.406451428536)*x[0]+(-0.292718530458)*x[1]**o+(-0.0948016380336)*x[1]+(-0.0943083566957)*x[2]**o+(0.747978324024)*x[2]
        arg[0,3]=(-0.0350228116485)*x[0]**o+(-0.967728170529)*x[0]+(-0.16425180976)*x[1]**o+(-0.141334783297)*x[1]+(0.179458630309)*x[2]**o+(0.45080506919)*x[2]
        arg[0,4]=(0.909578721407)*x[0]**o+(-0.341028350711)*x[0]+(-0.937703389458)*x[1]**o+(-0.845025015446)*x[1]+(-0.762836324997)*x[2]**o+(0.772817997464)*x[2]
        arg[1,0]=(0.75675349915)*x[0]**o+(-0.208827138169)*x[0]+(-0.0146742840091)*x[1]**o+(-0.1486215734)*x[1]+(0.205344647252)*x[2]**o+(-0.409293867732)*x[2]
        arg[1,1]=(-0.215359027624)*x[0]**o+(0.220661567012)*x[0]+(0.122954614324)*x[1]**o+(-0.888828539035)*x[1]+(-0.860535003713)*x[2]**o+(-0.582396229154)*x[2]
        arg[1,2]=(-0.252703089432)*x[0]**o+(0.294168447756)*x[0]+(0.323290819691)*x[1]**o+(-0.0440370164755)*x[1]+(0.487110700327)*x[2]**o+(-0.269109129573)*x[2]
        arg[1,3]=(0.0653521784751)*x[0]**o+(0.328048279094)*x[0]+(-0.414467594535)*x[1]**o+(0.436061182815)*x[1]+(0.483513283452)*x[2]**o+(-0.744560862926)*x[2]
        arg[1,4]=(0.0102955000926)*x[0]**o+(-0.636994970624)*x[0]+(0.83732393089)*x[1]**o+(0.914683806871)*x[1]+(0.574439856221)*x[2]**o+(-0.096821482591)*x[2]
        arg[2,0]=(-0.129692996812)*x[0]**o+(-0.617534967242)*x[0]+(0.299770969796)*x[1]**o+(0.93870530376)*x[1]+(-0.864448551767)*x[2]**o+(0.755129965839)*x[2]
        arg[2,1]=(-0.354122778719)*x[0]**o+(0.576654786045)*x[0]+(-0.965116346607)*x[1]**o+(0.0705856628847)*x[1]+(0.817059960545)*x[2]**o+(-0.597731384232)*x[2]
        arg[2,2]=(-0.705190107591)*x[0]**o+(0.876019062019)*x[0]+(-0.732733986501)*x[1]**o+(0.505334922081)*x[1]+(-0.616601928994)*x[2]**o+(-0.332268983863)*x[2]
        arg[2,3]=(0.663768017414)*x[0]**o+(-0.431682444208)*x[0]+(0.186159158088)*x[1]**o+(-0.689766933274)*x[1]+(-0.765854811539)*x[2]**o+(-0.402774599632)*x[2]
        arg[2,4]=(-0.698031877485)*x[0]**o+(0.914418269242)*x[0]+(-0.693564630363)*x[1]**o+(-0.18289238868)*x[1]+(0.988158500287)*x[2]**o+(-0.33952288549)*x[2]
        arg[3,0]=(-0.188499634873)*x[0]**o+(-0.10427816897)*x[0]+(-0.772866404794)*x[1]**o+(0.841928986479)*x[1]+(-0.866955639695)*x[2]**o+(-0.152515560713)*x[2]
        arg[3,1]=(0.937695460564)*x[0]**o+(-0.551079526492)*x[0]+(0.353043727721)*x[1]**o+(-0.604814667118)*x[1]+(0.200913251642)*x[2]**o+(-0.0497427852682)*x[2]
        arg[3,2]=(-0.49130784927)*x[0]**o+(0.537445967432)*x[0]+(-0.408371200391)*x[1]**o+(0.945593173362)*x[1]+(-0.0956969348675)*x[2]**o+(-0.174133549441)*x[2]
        arg[3,3]=(0.847790369559)*x[0]**o+(0.743993523287)*x[0]+(-0.810126540693)*x[1]**o+(-0.775699501886)*x[1]+(-0.479687101313)*x[2]**o+(0.627116825116)*x[2]
        arg[3,4]=(0.352455163407)*x[0]**o+(0.86343850667)*x[0]+(0.126742974996)*x[1]**o+(0.0175699888323)*x[1]+(0.964899127449)*x[2]**o+(-0.206749046822)*x[2]
        ref[0,0]=(-0.859332605724)*x_ref[0]**o+(-0.239005430756)*x_ref[0]+(-0.52959833033)*x_ref[1]**o+(0.725877930655)*x_ref[1]+(-0.0348436114161)*x_ref[2]**o+(0.47287744745)*x_ref[2]
        ref[0,1]=(-0.0816886854463)*x_ref[0]**o+(-0.626507446322)*x_ref[0]+(-0.93804868959)*x_ref[1]**o+(-0.450134311929)*x_ref[1]+(0.0957946938715)*x_ref[2]**o+(-0.236994496198)*x_ref[2]
        ref[0,2]=(-0.669441310039)*x_ref[0]**o+(-0.406451428536)*x_ref[0]+(-0.292718530458)*x_ref[1]**o+(-0.0948016380336)*x_ref[1]+(-0.0943083566957)*x_ref[2]**o+(0.747978324024)*x_ref[2]
        ref[0,3]=(-0.0350228116485)*x_ref[0]**o+(-0.967728170529)*x_ref[0]+(-0.16425180976)*x_ref[1]**o+(-0.141334783297)*x_ref[1]+(0.179458630309)*x_ref[2]**o+(0.45080506919)*x_ref[2]
        ref[0,4]=(0.909578721407)*x_ref[0]**o+(-0.341028350711)*x_ref[0]+(-0.937703389458)*x_ref[1]**o+(-0.845025015446)*x_ref[1]+(-0.762836324997)*x_ref[2]**o+(0.772817997464)*x_ref[2]
        ref[1,0]=(0.75675349915)*x_ref[0]**o+(-0.208827138169)*x_ref[0]+(-0.0146742840091)*x_ref[1]**o+(-0.1486215734)*x_ref[1]+(0.205344647252)*x_ref[2]**o+(-0.409293867732)*x_ref[2]
        ref[1,1]=(-0.215359027624)*x_ref[0]**o+(0.220661567012)*x_ref[0]+(0.122954614324)*x_ref[1]**o+(-0.888828539035)*x_ref[1]+(-0.860535003713)*x_ref[2]**o+(-0.582396229154)*x_ref[2]
        ref[1,2]=(-0.252703089432)*x_ref[0]**o+(0.294168447756)*x_ref[0]+(0.323290819691)*x_ref[1]**o+(-0.0440370164755)*x_ref[1]+(0.487110700327)*x_ref[2]**o+(-0.269109129573)*x_ref[2]
        ref[1,3]=(0.0653521784751)*x_ref[0]**o+(0.328048279094)*x_ref[0]+(-0.414467594535)*x_ref[1]**o+(0.436061182815)*x_ref[1]+(0.483513283452)*x_ref[2]**o+(-0.744560862926)*x_ref[2]
        ref[1,4]=(0.0102955000926)*x_ref[0]**o+(-0.636994970624)*x_ref[0]+(0.83732393089)*x_ref[1]**o+(0.914683806871)*x_ref[1]+(0.574439856221)*x_ref[2]**o+(-0.096821482591)*x_ref[2]
        ref[2,0]=(-0.129692996812)*x_ref[0]**o+(-0.617534967242)*x_ref[0]+(0.299770969796)*x_ref[1]**o+(0.93870530376)*x_ref[1]+(-0.864448551767)*x_ref[2]**o+(0.755129965839)*x_ref[2]
        ref[2,1]=(-0.354122778719)*x_ref[0]**o+(0.576654786045)*x_ref[0]+(-0.965116346607)*x_ref[1]**o+(0.0705856628847)*x_ref[1]+(0.817059960545)*x_ref[2]**o+(-0.597731384232)*x_ref[2]
        ref[2,2]=(-0.705190107591)*x_ref[0]**o+(0.876019062019)*x_ref[0]+(-0.732733986501)*x_ref[1]**o+(0.505334922081)*x_ref[1]+(-0.616601928994)*x_ref[2]**o+(-0.332268983863)*x_ref[2]
        ref[2,3]=(0.663768017414)*x_ref[0]**o+(-0.431682444208)*x_ref[0]+(0.186159158088)*x_ref[1]**o+(-0.689766933274)*x_ref[1]+(-0.765854811539)*x_ref[2]**o+(-0.402774599632)*x_ref[2]
        ref[2,4]=(-0.698031877485)*x_ref[0]**o+(0.914418269242)*x_ref[0]+(-0.693564630363)*x_ref[1]**o+(-0.18289238868)*x_ref[1]+(0.988158500287)*x_ref[2]**o+(-0.33952288549)*x_ref[2]
        ref[3,0]=(-0.188499634873)*x_ref[0]**o+(-0.10427816897)*x_ref[0]+(-0.772866404794)*x_ref[1]**o+(0.841928986479)*x_ref[1]+(-0.866955639695)*x_ref[2]**o+(-0.152515560713)*x_ref[2]
        ref[3,1]=(0.937695460564)*x_ref[0]**o+(-0.551079526492)*x_ref[0]+(0.353043727721)*x_ref[1]**o+(-0.604814667118)*x_ref[1]+(0.200913251642)*x_ref[2]**o+(-0.0497427852682)*x_ref[2]
        ref[3,2]=(-0.49130784927)*x_ref[0]**o+(0.537445967432)*x_ref[0]+(-0.408371200391)*x_ref[1]**o+(0.945593173362)*x_ref[1]+(-0.0956969348675)*x_ref[2]**o+(-0.174133549441)*x_ref[2]
        ref[3,3]=(0.847790369559)*x_ref[0]**o+(0.743993523287)*x_ref[0]+(-0.810126540693)*x_ref[1]**o+(-0.775699501886)*x_ref[1]+(-0.479687101313)*x_ref[2]**o+(0.627116825116)*x_ref[2]
        ref[3,4]=(0.352455163407)*x_ref[0]**o+(0.86343850667)*x_ref[0]+(0.126742974996)*x_ref[1]**o+(0.0175699888323)*x_ref[1]+(0.964899127449)*x_ref[2]**o+(-0.206749046822)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.396745847367)*x[0]**o+(-0.700719490691)*x[0]+(0.117558242545)*x[1]**o+(-0.161976142849)*x[1]
        arg[0,0,1]=(-0.490732277107)*x[0]**o+(0.526833149216)*x[0]+(-0.136309303122)*x[1]**o+(0.629465649533)*x[1]
        arg[0,1,0]=(-0.242474968121)*x[0]**o+(-0.76270698332)*x[0]+(-0.0979504547153)*x[1]**o+(-0.0743521177641)*x[1]
        arg[0,1,1]=(0.641246476843)*x[0]**o+(0.0158147797433)*x[0]+(-0.213898399703)*x[1]**o+(-0.378401519675)*x[1]
        arg[1,0,0]=(0.603700544573)*x[0]**o+(-0.993751144334)*x[0]+(0.466222025692)*x[1]**o+(0.457689495241)*x[1]
        arg[1,0,1]=(-0.727287858841)*x[0]**o+(-0.177841994955)*x[0]+(0.720237448692)*x[1]**o+(0.716175629178)*x[1]
        arg[1,1,0]=(0.784514682242)*x[0]**o+(0.681214612353)*x[0]+(0.404266860166)*x[1]**o+(0.871218157219)*x[1]
        arg[1,1,1]=(-0.918634586367)*x[0]**o+(-0.722365105889)*x[0]+(-0.0337202979587)*x[1]**o+(-0.487790376172)*x[1]
        arg[2,0,0]=(0.159515902499)*x[0]**o+(-0.0650417831188)*x[0]+(0.107260481333)*x[1]**o+(0.469096212511)*x[1]
        arg[2,0,1]=(0.0436003865409)*x[0]**o+(-0.67488837592)*x[0]+(0.547761443272)*x[1]**o+(-0.69222031017)*x[1]
        arg[2,1,0]=(-0.153636036391)*x[0]**o+(0.0415911821723)*x[0]+(0.421990370678)*x[1]**o+(-0.128883832414)*x[1]
        arg[2,1,1]=(0.513420393955)*x[0]**o+(-0.769636964711)*x[0]+(0.874383588901)*x[1]**o+(-0.357923227767)*x[1]
        arg[3,0,0]=(0.943275907357)*x[0]**o+(0.372378913231)*x[0]+(0.559656674007)*x[1]**o+(-0.878007664293)*x[1]
        arg[3,0,1]=(-0.541535388034)*x[0]**o+(-0.0510200932939)*x[0]+(-0.448618747155)*x[1]**o+(0.693514427768)*x[1]
        arg[3,1,0]=(-0.287622342268)*x[0]**o+(0.8333134988)*x[0]+(-0.691930501782)*x[1]**o+(-0.659618596505)*x[1]
        arg[3,1,1]=(0.44446133556)*x[0]**o+(0.697704081094)*x[0]+(0.0679863572322)*x[1]**o+(-0.949443102916)*x[1]
        arg[4,0,0]=(0.546245770076)*x[0]**o+(-0.466891879347)*x[0]+(-0.608261901126)*x[1]**o+(0.669561643667)*x[1]
        arg[4,0,1]=(-0.522497485476)*x[0]**o+(0.70886716903)*x[0]+(-0.296815361132)*x[1]**o+(0.500540583082)*x[1]
        arg[4,1,0]=(-0.841653490079)*x[0]**o+(0.400589086976)*x[0]+(0.865994660457)*x[1]**o+(0.0205817210969)*x[1]
        arg[4,1,1]=(-0.64369513787)*x[0]**o+(0.883032618928)*x[0]+(-0.178766897224)*x[1]**o+(0.748041632322)*x[1]
        arg[5,0,0]=(0.471810615096)*x[0]**o+(-0.140340964274)*x[0]+(-0.7063314968)*x[1]**o+(0.0524231879837)*x[1]
        arg[5,0,1]=(-0.51339482213)*x[0]**o+(-0.81447661092)*x[0]+(0.437085868301)*x[1]**o+(0.862678017842)*x[1]
        arg[5,1,0]=(-0.752858172473)*x[0]**o+(0.325827525243)*x[0]+(-0.702286757187)*x[1]**o+(-0.467810951898)*x[1]
        arg[5,1,1]=(-0.256879797442)*x[0]**o+(0.257067666229)*x[0]+(0.912308244411)*x[1]**o+(0.0620337856352)*x[1]
        ref[0,0,0]=(-0.396745847367)*x_ref[0]**o+(-0.700719490691)*x_ref[0]+(0.117558242545)*x_ref[1]**o+(-0.161976142849)*x_ref[1]
        ref[0,0,1]=(-0.490732277107)*x_ref[0]**o+(0.526833149216)*x_ref[0]+(-0.136309303122)*x_ref[1]**o+(0.629465649533)*x_ref[1]
        ref[0,1,0]=(-0.242474968121)*x_ref[0]**o+(-0.76270698332)*x_ref[0]+(-0.0979504547153)*x_ref[1]**o+(-0.0743521177641)*x_ref[1]
        ref[0,1,1]=(0.641246476843)*x_ref[0]**o+(0.0158147797433)*x_ref[0]+(-0.213898399703)*x_ref[1]**o+(-0.378401519675)*x_ref[1]
        ref[1,0,0]=(0.603700544573)*x_ref[0]**o+(-0.993751144334)*x_ref[0]+(0.466222025692)*x_ref[1]**o+(0.457689495241)*x_ref[1]
        ref[1,0,1]=(-0.727287858841)*x_ref[0]**o+(-0.177841994955)*x_ref[0]+(0.720237448692)*x_ref[1]**o+(0.716175629178)*x_ref[1]
        ref[1,1,0]=(0.784514682242)*x_ref[0]**o+(0.681214612353)*x_ref[0]+(0.404266860166)*x_ref[1]**o+(0.871218157219)*x_ref[1]
        ref[1,1,1]=(-0.918634586367)*x_ref[0]**o+(-0.722365105889)*x_ref[0]+(-0.0337202979587)*x_ref[1]**o+(-0.487790376172)*x_ref[1]
        ref[2,0,0]=(0.159515902499)*x_ref[0]**o+(-0.0650417831188)*x_ref[0]+(0.107260481333)*x_ref[1]**o+(0.469096212511)*x_ref[1]
        ref[2,0,1]=(0.0436003865409)*x_ref[0]**o+(-0.67488837592)*x_ref[0]+(0.547761443272)*x_ref[1]**o+(-0.69222031017)*x_ref[1]
        ref[2,1,0]=(-0.153636036391)*x_ref[0]**o+(0.0415911821723)*x_ref[0]+(0.421990370678)*x_ref[1]**o+(-0.128883832414)*x_ref[1]
        ref[2,1,1]=(0.513420393955)*x_ref[0]**o+(-0.769636964711)*x_ref[0]+(0.874383588901)*x_ref[1]**o+(-0.357923227767)*x_ref[1]
        ref[3,0,0]=(0.943275907357)*x_ref[0]**o+(0.372378913231)*x_ref[0]+(0.559656674007)*x_ref[1]**o+(-0.878007664293)*x_ref[1]
        ref[3,0,1]=(-0.541535388034)*x_ref[0]**o+(-0.0510200932939)*x_ref[0]+(-0.448618747155)*x_ref[1]**o+(0.693514427768)*x_ref[1]
        ref[3,1,0]=(-0.287622342268)*x_ref[0]**o+(0.8333134988)*x_ref[0]+(-0.691930501782)*x_ref[1]**o+(-0.659618596505)*x_ref[1]
        ref[3,1,1]=(0.44446133556)*x_ref[0]**o+(0.697704081094)*x_ref[0]+(0.0679863572322)*x_ref[1]**o+(-0.949443102916)*x_ref[1]
        ref[4,0,0]=(0.546245770076)*x_ref[0]**o+(-0.466891879347)*x_ref[0]+(-0.608261901126)*x_ref[1]**o+(0.669561643667)*x_ref[1]
        ref[4,0,1]=(-0.522497485476)*x_ref[0]**o+(0.70886716903)*x_ref[0]+(-0.296815361132)*x_ref[1]**o+(0.500540583082)*x_ref[1]
        ref[4,1,0]=(-0.841653490079)*x_ref[0]**o+(0.400589086976)*x_ref[0]+(0.865994660457)*x_ref[1]**o+(0.0205817210969)*x_ref[1]
        ref[4,1,1]=(-0.64369513787)*x_ref[0]**o+(0.883032618928)*x_ref[0]+(-0.178766897224)*x_ref[1]**o+(0.748041632322)*x_ref[1]
        ref[5,0,0]=(0.471810615096)*x_ref[0]**o+(-0.140340964274)*x_ref[0]+(-0.7063314968)*x_ref[1]**o+(0.0524231879837)*x_ref[1]
        ref[5,0,1]=(-0.51339482213)*x_ref[0]**o+(-0.81447661092)*x_ref[0]+(0.437085868301)*x_ref[1]**o+(0.862678017842)*x_ref[1]
        ref[5,1,0]=(-0.752858172473)*x_ref[0]**o+(0.325827525243)*x_ref[0]+(-0.702286757187)*x_ref[1]**o+(-0.467810951898)*x_ref[1]
        ref[5,1,1]=(-0.256879797442)*x_ref[0]**o+(0.257067666229)*x_ref[0]+(0.912308244411)*x_ref[1]**o+(0.0620337856352)*x_ref[1]
      else:
        arg[0,0,0]=(0.564502533309)*x[0]**o+(-0.237721559752)*x[0]+(0.225601658644)*x[1]**o+(0.967190188292)*x[1]+(0.471760897542)*x[2]**o+(0.79404420728)*x[2]
        arg[0,0,1]=(-0.373762346136)*x[0]**o+(-0.643766021731)*x[0]+(0.088028383792)*x[1]**o+(-0.541163410313)*x[1]+(0.463315911637)*x[2]**o+(-0.596740807595)*x[2]
        arg[0,1,0]=(-0.526694453842)*x[0]**o+(0.405010889931)*x[0]+(-0.0153807962972)*x[1]**o+(-0.0446620712949)*x[1]+(0.00553918668197)*x[2]**o+(0.72170514564)*x[2]
        arg[0,1,1]=(-0.00750073616089)*x[0]**o+(-0.60431903208)*x[0]+(0.790032415931)*x[1]**o+(0.272709398336)*x[1]+(0.228539600712)*x[2]**o+(0.62284140325)*x[2]
        arg[1,0,0]=(-0.892797595418)*x[0]**o+(-0.151608141828)*x[0]+(0.313142074823)*x[1]**o+(0.80280439466)*x[1]+(-0.593056273594)*x[2]**o+(-0.41551679707)*x[2]
        arg[1,0,1]=(-0.385715092974)*x[0]**o+(-0.77884107065)*x[0]+(-0.0804019137225)*x[1]**o+(-0.240129932101)*x[1]+(0.238145788042)*x[2]**o+(-0.806337666628)*x[2]
        arg[1,1,0]=(0.940492459417)*x[0]**o+(-0.0054460073194)*x[0]+(-0.697916565168)*x[1]**o+(-0.949456184725)*x[1]+(-0.115461214789)*x[2]**o+(-0.403742409948)*x[2]
        arg[1,1,1]=(-0.711716124415)*x[0]**o+(0.586131021871)*x[0]+(0.625988666837)*x[1]**o+(0.638433219635)*x[1]+(0.971169174015)*x[2]**o+(-0.064949626884)*x[2]
        arg[2,0,0]=(-0.891973333325)*x[0]**o+(0.76090478364)*x[0]+(0.439956136239)*x[1]**o+(-0.74807071296)*x[1]+(-0.209999803467)*x[2]**o+(0.404396254503)*x[2]
        arg[2,0,1]=(0.559776288735)*x[0]**o+(0.284643618707)*x[0]+(0.0527799359967)*x[1]**o+(-0.201660204363)*x[1]+(0.590722655898)*x[2]**o+(-0.24940757474)*x[2]
        arg[2,1,0]=(-0.747588234432)*x[0]**o+(0.0196702046746)*x[0]+(0.0454238912973)*x[1]**o+(0.174053915643)*x[1]+(-0.868554688657)*x[2]**o+(-0.127425849813)*x[2]
        arg[2,1,1]=(-0.324276984008)*x[0]**o+(-0.494912096003)*x[0]+(-0.780953487415)*x[1]**o+(0.225787179491)*x[1]+(-0.760330660035)*x[2]**o+(0.0440722700759)*x[2]
        arg[3,0,0]=(0.37479028021)*x[0]**o+(-0.869009335783)*x[0]+(0.811673580204)*x[1]**o+(0.521877505351)*x[1]+(-0.508710979256)*x[2]**o+(0.558085136638)*x[2]
        arg[3,0,1]=(0.815143535066)*x[0]**o+(0.867977183071)*x[0]+(0.0315543942614)*x[1]**o+(0.786379597601)*x[1]+(0.0433279381339)*x[2]**o+(-0.56554445624)*x[2]
        arg[3,1,0]=(-0.948224238625)*x[0]**o+(-0.759282528831)*x[0]+(0.511802336753)*x[1]**o+(0.331073342499)*x[1]+(0.295241697427)*x[2]**o+(0.154864664771)*x[2]
        arg[3,1,1]=(-0.982634132939)*x[0]**o+(-0.0762014431695)*x[0]+(-0.0797462613351)*x[1]**o+(-0.512305271957)*x[1]+(-0.348883505418)*x[2]**o+(0.00580412526927)*x[2]
        arg[4,0,0]=(0.186597109908)*x[0]**o+(0.051957206238)*x[0]+(0.572882754553)*x[1]**o+(-0.114338954084)*x[1]+(-0.517239439578)*x[2]**o+(-0.298287404607)*x[2]
        arg[4,0,1]=(0.637628569107)*x[0]**o+(0.228562669958)*x[0]+(0.053657321561)*x[1]**o+(0.762909608846)*x[1]+(0.873936819574)*x[2]**o+(-0.20000969514)*x[2]
        arg[4,1,0]=(-0.717459489066)*x[0]**o+(-0.462174191129)*x[0]+(-0.70491180102)*x[1]**o+(-0.406072010622)*x[1]+(-0.903872259447)*x[2]**o+(0.487491537826)*x[2]
        arg[4,1,1]=(0.947654286923)*x[0]**o+(-0.228830597313)*x[0]+(-0.501442713655)*x[1]**o+(0.282437368749)*x[1]+(0.543592950002)*x[2]**o+(0.5400198247)*x[2]
        arg[5,0,0]=(0.144349494033)*x[0]**o+(-0.635328703132)*x[0]+(-0.797851162499)*x[1]**o+(0.356874780185)*x[1]+(-0.205725457569)*x[2]**o+(-0.674962512018)*x[2]
        arg[5,0,1]=(-0.728948504419)*x[0]**o+(0.944337871242)*x[0]+(-0.790388944678)*x[1]**o+(0.59119693076)*x[1]+(0.9569152246)*x[2]**o+(-0.205404245177)*x[2]
        arg[5,1,0]=(0.477597678348)*x[0]**o+(-0.318993929605)*x[0]+(0.481029970286)*x[1]**o+(0.708696918395)*x[1]+(0.0531716082887)*x[2]**o+(-0.847697491061)*x[2]
        arg[5,1,1]=(-0.458572001979)*x[0]**o+(0.582048244354)*x[0]+(-0.040756132547)*x[1]**o+(-0.953201198076)*x[1]+(-0.971859723822)*x[2]**o+(-0.796426143181)*x[2]
        ref[0,0,0]=(0.564502533309)*x_ref[0]**o+(-0.237721559752)*x_ref[0]+(0.225601658644)*x_ref[1]**o+(0.967190188292)*x_ref[1]+(0.471760897542)*x_ref[2]**o+(0.79404420728)*x_ref[2]
        ref[0,0,1]=(-0.373762346136)*x_ref[0]**o+(-0.643766021731)*x_ref[0]+(0.088028383792)*x_ref[1]**o+(-0.541163410313)*x_ref[1]+(0.463315911637)*x_ref[2]**o+(-0.596740807595)*x_ref[2]
        ref[0,1,0]=(-0.526694453842)*x_ref[0]**o+(0.405010889931)*x_ref[0]+(-0.0153807962972)*x_ref[1]**o+(-0.0446620712949)*x_ref[1]+(0.00553918668197)*x_ref[2]**o+(0.72170514564)*x_ref[2]
        ref[0,1,1]=(-0.00750073616089)*x_ref[0]**o+(-0.60431903208)*x_ref[0]+(0.790032415931)*x_ref[1]**o+(0.272709398336)*x_ref[1]+(0.228539600712)*x_ref[2]**o+(0.62284140325)*x_ref[2]
        ref[1,0,0]=(-0.892797595418)*x_ref[0]**o+(-0.151608141828)*x_ref[0]+(0.313142074823)*x_ref[1]**o+(0.80280439466)*x_ref[1]+(-0.593056273594)*x_ref[2]**o+(-0.41551679707)*x_ref[2]
        ref[1,0,1]=(-0.385715092974)*x_ref[0]**o+(-0.77884107065)*x_ref[0]+(-0.0804019137225)*x_ref[1]**o+(-0.240129932101)*x_ref[1]+(0.238145788042)*x_ref[2]**o+(-0.806337666628)*x_ref[2]
        ref[1,1,0]=(0.940492459417)*x_ref[0]**o+(-0.0054460073194)*x_ref[0]+(-0.697916565168)*x_ref[1]**o+(-0.949456184725)*x_ref[1]+(-0.115461214789)*x_ref[2]**o+(-0.403742409948)*x_ref[2]
        ref[1,1,1]=(-0.711716124415)*x_ref[0]**o+(0.586131021871)*x_ref[0]+(0.625988666837)*x_ref[1]**o+(0.638433219635)*x_ref[1]+(0.971169174015)*x_ref[2]**o+(-0.064949626884)*x_ref[2]
        ref[2,0,0]=(-0.891973333325)*x_ref[0]**o+(0.76090478364)*x_ref[0]+(0.439956136239)*x_ref[1]**o+(-0.74807071296)*x_ref[1]+(-0.209999803467)*x_ref[2]**o+(0.404396254503)*x_ref[2]
        ref[2,0,1]=(0.559776288735)*x_ref[0]**o+(0.284643618707)*x_ref[0]+(0.0527799359967)*x_ref[1]**o+(-0.201660204363)*x_ref[1]+(0.590722655898)*x_ref[2]**o+(-0.24940757474)*x_ref[2]
        ref[2,1,0]=(-0.747588234432)*x_ref[0]**o+(0.0196702046746)*x_ref[0]+(0.0454238912973)*x_ref[1]**o+(0.174053915643)*x_ref[1]+(-0.868554688657)*x_ref[2]**o+(-0.127425849813)*x_ref[2]
        ref[2,1,1]=(-0.324276984008)*x_ref[0]**o+(-0.494912096003)*x_ref[0]+(-0.780953487415)*x_ref[1]**o+(0.225787179491)*x_ref[1]+(-0.760330660035)*x_ref[2]**o+(0.0440722700759)*x_ref[2]
        ref[3,0,0]=(0.37479028021)*x_ref[0]**o+(-0.869009335783)*x_ref[0]+(0.811673580204)*x_ref[1]**o+(0.521877505351)*x_ref[1]+(-0.508710979256)*x_ref[2]**o+(0.558085136638)*x_ref[2]
        ref[3,0,1]=(0.815143535066)*x_ref[0]**o+(0.867977183071)*x_ref[0]+(0.0315543942614)*x_ref[1]**o+(0.786379597601)*x_ref[1]+(0.0433279381339)*x_ref[2]**o+(-0.56554445624)*x_ref[2]
        ref[3,1,0]=(-0.948224238625)*x_ref[0]**o+(-0.759282528831)*x_ref[0]+(0.511802336753)*x_ref[1]**o+(0.331073342499)*x_ref[1]+(0.295241697427)*x_ref[2]**o+(0.154864664771)*x_ref[2]
        ref[3,1,1]=(-0.982634132939)*x_ref[0]**o+(-0.0762014431695)*x_ref[0]+(-0.0797462613351)*x_ref[1]**o+(-0.512305271957)*x_ref[1]+(-0.348883505418)*x_ref[2]**o+(0.00580412526927)*x_ref[2]
        ref[4,0,0]=(0.186597109908)*x_ref[0]**o+(0.051957206238)*x_ref[0]+(0.572882754553)*x_ref[1]**o+(-0.114338954084)*x_ref[1]+(-0.517239439578)*x_ref[2]**o+(-0.298287404607)*x_ref[2]
        ref[4,0,1]=(0.637628569107)*x_ref[0]**o+(0.228562669958)*x_ref[0]+(0.053657321561)*x_ref[1]**o+(0.762909608846)*x_ref[1]+(0.873936819574)*x_ref[2]**o+(-0.20000969514)*x_ref[2]
        ref[4,1,0]=(-0.717459489066)*x_ref[0]**o+(-0.462174191129)*x_ref[0]+(-0.70491180102)*x_ref[1]**o+(-0.406072010622)*x_ref[1]+(-0.903872259447)*x_ref[2]**o+(0.487491537826)*x_ref[2]
        ref[4,1,1]=(0.947654286923)*x_ref[0]**o+(-0.228830597313)*x_ref[0]+(-0.501442713655)*x_ref[1]**o+(0.282437368749)*x_ref[1]+(0.543592950002)*x_ref[2]**o+(0.5400198247)*x_ref[2]
        ref[5,0,0]=(0.144349494033)*x_ref[0]**o+(-0.635328703132)*x_ref[0]+(-0.797851162499)*x_ref[1]**o+(0.356874780185)*x_ref[1]+(-0.205725457569)*x_ref[2]**o+(-0.674962512018)*x_ref[2]
        ref[5,0,1]=(-0.728948504419)*x_ref[0]**o+(0.944337871242)*x_ref[0]+(-0.790388944678)*x_ref[1]**o+(0.59119693076)*x_ref[1]+(0.9569152246)*x_ref[2]**o+(-0.205404245177)*x_ref[2]
        ref[5,1,0]=(0.477597678348)*x_ref[0]**o+(-0.318993929605)*x_ref[0]+(0.481029970286)*x_ref[1]**o+(0.708696918395)*x_ref[1]+(0.0531716082887)*x_ref[2]**o+(-0.847697491061)*x_ref[2]
        ref[5,1,1]=(-0.458572001979)*x_ref[0]**o+(0.582048244354)*x_ref[0]+(-0.040756132547)*x_ref[1]**o+(-0.953201198076)*x_ref[1]+(-0.971859723822)*x_ref[2]**o+(-0.796426143181)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnContactOne_fromData_ReducedFunctionOnContactOne_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnContactOne

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnContactOne onto ReducedFunctionOnContactOne
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnContactOne(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.937777610663)*x[0]**o+(-0.0228520946632)*x[0]+(0.642385396583)*x[1]**o+(-0.803821945812)*x[1]
        arg[0,0,0,1]=(-0.398785307492)*x[0]**o+(0.0450288952883)*x[0]+(-0.777780548133)*x[1]**o+(0.0273772099414)*x[1]
        arg[0,0,1,0]=(-0.667280086925)*x[0]**o+(-0.934642710274)*x[0]+(0.565327494491)*x[1]**o+(0.988521769056)*x[1]
        arg[0,0,1,1]=(-0.225562383498)*x[0]**o+(-0.280566031134)*x[0]+(-0.295768954831)*x[1]**o+(0.102759013802)*x[1]
        arg[0,0,2,0]=(-0.179577502546)*x[0]**o+(-0.156932777074)*x[0]+(-0.145616153633)*x[1]**o+(0.971876899135)*x[1]
        arg[0,0,2,1]=(-0.840266409118)*x[0]**o+(0.704356785022)*x[0]+(0.740091470711)*x[1]**o+(0.708478499023)*x[1]
        arg[0,1,0,0]=(-0.487161662806)*x[0]**o+(0.565313908399)*x[0]+(-0.33708697293)*x[1]**o+(0.731675966546)*x[1]
        arg[0,1,0,1]=(-0.135615857963)*x[0]**o+(0.265408777025)*x[0]+(-0.168207260227)*x[1]**o+(-0.598974659317)*x[1]
        arg[0,1,1,0]=(0.0533944010258)*x[0]**o+(0.819248906738)*x[0]+(-0.837005251524)*x[1]**o+(0.119613618595)*x[1]
        arg[0,1,1,1]=(-0.257520627047)*x[0]**o+(0.339264545713)*x[0]+(0.137721228188)*x[1]**o+(-0.512251030979)*x[1]
        arg[0,1,2,0]=(-0.797070396749)*x[0]**o+(0.690173285077)*x[0]+(-0.213206187271)*x[1]**o+(-0.873999125945)*x[1]
        arg[0,1,2,1]=(-0.847662300367)*x[0]**o+(0.785781114266)*x[0]+(-0.456918398811)*x[1]**o+(0.458881079947)*x[1]
        arg[0,2,0,0]=(-0.385269163334)*x[0]**o+(0.135941228964)*x[0]+(0.233245620648)*x[1]**o+(-0.246282209626)*x[1]
        arg[0,2,0,1]=(0.372914636218)*x[0]**o+(-0.160283657003)*x[0]+(0.0565043889781)*x[1]**o+(0.199715522021)*x[1]
        arg[0,2,1,0]=(0.676198390598)*x[0]**o+(0.938025122014)*x[0]+(0.161074696651)*x[1]**o+(-0.155495646964)*x[1]
        arg[0,2,1,1]=(-0.709286924155)*x[0]**o+(-0.173716284776)*x[0]+(0.0576435484694)*x[1]**o+(0.673924697832)*x[1]
        arg[0,2,2,0]=(-0.436468652969)*x[0]**o+(0.117279009082)*x[0]+(-0.731436369191)*x[1]**o+(0.919578503469)*x[1]
        arg[0,2,2,1]=(-0.691042646151)*x[0]**o+(-0.606293725574)*x[0]+(0.657291692759)*x[1]**o+(-0.0146398994729)*x[1]
        arg[0,3,0,0]=(0.365939313719)*x[0]**o+(-0.221236913792)*x[0]+(-0.729594967883)*x[1]**o+(0.0179024719114)*x[1]
        arg[0,3,0,1]=(0.135892133258)*x[0]**o+(0.728069461936)*x[0]+(-0.365691795279)*x[1]**o+(0.877708599922)*x[1]
        arg[0,3,1,0]=(-0.853312456817)*x[0]**o+(-0.713031969049)*x[0]+(0.274755827464)*x[1]**o+(-0.527330699033)*x[1]
        arg[0,3,1,1]=(-0.30358584687)*x[0]**o+(0.871629391968)*x[0]+(0.906915162645)*x[1]**o+(0.499955429041)*x[1]
        arg[0,3,2,0]=(-0.880532587427)*x[0]**o+(0.265729672258)*x[0]+(0.528677136349)*x[1]**o+(-0.224602964925)*x[1]
        arg[0,3,2,1]=(0.158632463532)*x[0]**o+(-0.665418329643)*x[0]+(0.957661069067)*x[1]**o+(-0.23771309434)*x[1]
        arg[0,4,0,0]=(-0.128073514905)*x[0]**o+(-0.452543798055)*x[0]+(0.303808626277)*x[1]**o+(0.841041779759)*x[1]
        arg[0,4,0,1]=(0.0219309964781)*x[0]**o+(-0.949488393443)*x[0]+(0.0436905404663)*x[1]**o+(-0.304527242795)*x[1]
        arg[0,4,1,0]=(-0.792275137775)*x[0]**o+(0.450079829644)*x[0]+(0.162879981637)*x[1]**o+(-0.185839015805)*x[1]
        arg[0,4,1,1]=(0.761932713893)*x[0]**o+(0.932601232945)*x[0]+(-0.606328305783)*x[1]**o+(-0.123973545682)*x[1]
        arg[0,4,2,0]=(-0.514088304908)*x[0]**o+(0.987668774807)*x[0]+(0.866078887168)*x[1]**o+(-0.608518224484)*x[1]
        arg[0,4,2,1]=(-0.427564608028)*x[0]**o+(0.231599614445)*x[0]+(-0.406862229819)*x[1]**o+(0.358456864265)*x[1]
        arg[1,0,0,0]=(0.342092171801)*x[0]**o+(-0.264522675185)*x[0]+(-0.37976917867)*x[1]**o+(0.778596602561)*x[1]
        arg[1,0,0,1]=(0.672156473235)*x[0]**o+(0.320317500856)*x[0]+(0.514206881296)*x[1]**o+(0.602594631705)*x[1]
        arg[1,0,1,0]=(0.914472920293)*x[0]**o+(0.525647943447)*x[0]+(0.685613621597)*x[1]**o+(0.482308039231)*x[1]
        arg[1,0,1,1]=(-0.727982874425)*x[0]**o+(-0.417964463065)*x[0]+(0.825853988063)*x[1]**o+(0.236134760252)*x[1]
        arg[1,0,2,0]=(0.922186381756)*x[0]**o+(-0.140522762044)*x[0]+(-0.282169431666)*x[1]**o+(0.778383988397)*x[1]
        arg[1,0,2,1]=(0.307296736289)*x[0]**o+(0.589844868569)*x[0]+(0.64571457752)*x[1]**o+(0.749333079485)*x[1]
        arg[1,1,0,0]=(0.908345176703)*x[0]**o+(-0.476053147164)*x[0]+(0.0828501189289)*x[1]**o+(0.280657117705)*x[1]
        arg[1,1,0,1]=(0.396128814097)*x[0]**o+(0.368821575482)*x[0]+(-0.916584602334)*x[1]**o+(-0.149377992159)*x[1]
        arg[1,1,1,0]=(-0.38468474313)*x[0]**o+(0.304742499301)*x[0]+(0.638083865846)*x[1]**o+(-0.332046460663)*x[1]
        arg[1,1,1,1]=(0.845097787138)*x[0]**o+(-0.34949498164)*x[0]+(0.229157202572)*x[1]**o+(0.677584672856)*x[1]
        arg[1,1,2,0]=(0.0280692825358)*x[0]**o+(0.567451370107)*x[0]+(0.66934188305)*x[1]**o+(-0.477197213188)*x[1]
        arg[1,1,2,1]=(0.375411487264)*x[0]**o+(0.55086548839)*x[0]+(-0.392135351366)*x[1]**o+(0.91971334589)*x[1]
        arg[1,2,0,0]=(-0.7659270368)*x[0]**o+(-0.338082792509)*x[0]+(-0.26272975232)*x[1]**o+(0.267936454568)*x[1]
        arg[1,2,0,1]=(0.918125455771)*x[0]**o+(0.182896655906)*x[0]+(0.525396835337)*x[1]**o+(0.201918789801)*x[1]
        arg[1,2,1,0]=(-0.54472490266)*x[0]**o+(0.693714832775)*x[0]+(0.898278265794)*x[1]**o+(0.345095187159)*x[1]
        arg[1,2,1,1]=(-0.498607752195)*x[0]**o+(0.84678344253)*x[0]+(0.678992455984)*x[1]**o+(-0.597396262326)*x[1]
        arg[1,2,2,0]=(0.215267998554)*x[0]**o+(0.384479388838)*x[0]+(-0.09078862662)*x[1]**o+(-0.0162185695524)*x[1]
        arg[1,2,2,1]=(-0.335574750173)*x[0]**o+(-0.439933690512)*x[0]+(0.194601817488)*x[1]**o+(0.795181445633)*x[1]
        arg[1,3,0,0]=(0.862180273604)*x[0]**o+(0.276662657248)*x[0]+(-0.908827123328)*x[1]**o+(0.290406540669)*x[1]
        arg[1,3,0,1]=(-0.620441636176)*x[0]**o+(0.448236189204)*x[0]+(-0.503817375869)*x[1]**o+(0.689447469467)*x[1]
        arg[1,3,1,0]=(0.126327572417)*x[0]**o+(-0.476552913848)*x[0]+(0.362273273226)*x[1]**o+(-0.816228830372)*x[1]
        arg[1,3,1,1]=(-0.65091977023)*x[0]**o+(0.0316560718061)*x[0]+(-0.922906273452)*x[1]**o+(-0.584811809954)*x[1]
        arg[1,3,2,0]=(-0.675982262546)*x[0]**o+(0.23518662745)*x[0]+(-0.270229269162)*x[1]**o+(-0.00300656227686)*x[1]
        arg[1,3,2,1]=(0.55691633063)*x[0]**o+(-0.0643123929496)*x[0]+(0.212825067376)*x[1]**o+(-0.0807098593265)*x[1]
        arg[1,4,0,0]=(0.199100268435)*x[0]**o+(-0.358028564112)*x[0]+(0.130301315926)*x[1]**o+(-0.255752658178)*x[1]
        arg[1,4,0,1]=(0.821720568968)*x[0]**o+(0.193634063182)*x[0]+(0.862353589302)*x[1]**o+(-0.42971617129)*x[1]
        arg[1,4,1,0]=(0.51160977767)*x[0]**o+(0.326743318954)*x[0]+(-0.866878934278)*x[1]**o+(-0.927585776234)*x[1]
        arg[1,4,1,1]=(0.299835915861)*x[0]**o+(-0.969027084981)*x[0]+(-0.589772230832)*x[1]**o+(0.382433302029)*x[1]
        arg[1,4,2,0]=(0.473101160384)*x[0]**o+(-0.591124830283)*x[0]+(0.0929141413982)*x[1]**o+(-0.529811769054)*x[1]
        arg[1,4,2,1]=(-0.712363897962)*x[0]**o+(-0.259229026597)*x[0]+(0.00983651162685)*x[1]**o+(-0.329312308231)*x[1]
        arg[2,0,0,0]=(-0.507414918232)*x[0]**o+(0.668167664289)*x[0]+(0.654870826953)*x[1]**o+(-0.411963068213)*x[1]
        arg[2,0,0,1]=(-0.894915678458)*x[0]**o+(0.647579368695)*x[0]+(0.222274441569)*x[1]**o+(0.737972296743)*x[1]
        arg[2,0,1,0]=(-0.457176106717)*x[0]**o+(0.99997698232)*x[0]+(0.524381951875)*x[1]**o+(0.796302331418)*x[1]
        arg[2,0,1,1]=(-0.761912711581)*x[0]**o+(-0.995235985398)*x[0]+(-0.371926591709)*x[1]**o+(-0.37949019966)*x[1]
        arg[2,0,2,0]=(0.402826798397)*x[0]**o+(0.473547871218)*x[0]+(0.801636706721)*x[1]**o+(0.314221308918)*x[1]
        arg[2,0,2,1]=(0.0416846902183)*x[0]**o+(0.815306853966)*x[0]+(-0.526315166764)*x[1]**o+(-0.327935702076)*x[1]
        arg[2,1,0,0]=(-0.53829655239)*x[0]**o+(-0.467914397053)*x[0]+(0.392659650792)*x[1]**o+(0.309531884089)*x[1]
        arg[2,1,0,1]=(-0.768324371377)*x[0]**o+(0.136336455054)*x[0]+(0.0653485752909)*x[1]**o+(-0.0181923467911)*x[1]
        arg[2,1,1,0]=(0.121561771322)*x[0]**o+(-0.794575616334)*x[0]+(0.038194779569)*x[1]**o+(-0.695864614551)*x[1]
        arg[2,1,1,1]=(0.941872575674)*x[0]**o+(0.322067879395)*x[0]+(0.176873583908)*x[1]**o+(0.266964433323)*x[1]
        arg[2,1,2,0]=(0.324335817101)*x[0]**o+(0.609048475305)*x[0]+(0.350973789003)*x[1]**o+(-0.284995830372)*x[1]
        arg[2,1,2,1]=(0.939941132515)*x[0]**o+(0.782468124836)*x[0]+(-0.238539378134)*x[1]**o+(-0.515203802574)*x[1]
        arg[2,2,0,0]=(0.080179317168)*x[0]**o+(-0.555769741953)*x[0]+(0.283304726007)*x[1]**o+(0.844935300942)*x[1]
        arg[2,2,0,1]=(0.52001819752)*x[0]**o+(-0.105912273152)*x[0]+(-0.360523424917)*x[1]**o+(-0.179135541646)*x[1]
        arg[2,2,1,0]=(0.719593478448)*x[0]**o+(-0.722945454687)*x[0]+(0.0316404625439)*x[1]**o+(0.154290328891)*x[1]
        arg[2,2,1,1]=(0.906382302919)*x[0]**o+(-0.137753840588)*x[0]+(-0.626417865101)*x[1]**o+(-0.890642745149)*x[1]
        arg[2,2,2,0]=(0.618763086671)*x[0]**o+(-0.267876013516)*x[0]+(0.582353922702)*x[1]**o+(-0.190278788501)*x[1]
        arg[2,2,2,1]=(0.778824546516)*x[0]**o+(-0.00837218512008)*x[0]+(0.576376019084)*x[1]**o+(-0.173470710629)*x[1]
        arg[2,3,0,0]=(0.351742111948)*x[0]**o+(-0.641801256932)*x[0]+(-0.979232109277)*x[1]**o+(0.973197715414)*x[1]
        arg[2,3,0,1]=(-0.194493570182)*x[0]**o+(-0.24088269914)*x[0]+(-0.594521666623)*x[1]**o+(0.442357460052)*x[1]
        arg[2,3,1,0]=(-0.0677209625916)*x[0]**o+(-0.329246651504)*x[0]+(-0.200109031557)*x[1]**o+(0.0666064469254)*x[1]
        arg[2,3,1,1]=(-0.31026797404)*x[0]**o+(-0.0420778789656)*x[0]+(-0.534104800904)*x[1]**o+(0.940011883171)*x[1]
        arg[2,3,2,0]=(0.572825859789)*x[0]**o+(-0.439819027899)*x[0]+(-0.641214796724)*x[1]**o+(0.281317897435)*x[1]
        arg[2,3,2,1]=(0.410920070755)*x[0]**o+(-0.517259918184)*x[0]+(0.436420469387)*x[1]**o+(0.212276585342)*x[1]
        arg[2,4,0,0]=(0.0537540090315)*x[0]**o+(-0.362577123782)*x[0]+(-0.0134992480023)*x[1]**o+(0.59055317928)*x[1]
        arg[2,4,0,1]=(0.894896469792)*x[0]**o+(-0.594219811149)*x[0]+(0.755490579022)*x[1]**o+(-0.222913251691)*x[1]
        arg[2,4,1,0]=(-0.95471806813)*x[0]**o+(0.364260591437)*x[0]+(-0.51232745259)*x[1]**o+(-0.874795288001)*x[1]
        arg[2,4,1,1]=(-0.146483465086)*x[0]**o+(0.182495641134)*x[0]+(-0.132211746968)*x[1]**o+(0.39915899917)*x[1]
        arg[2,4,2,0]=(0.630856642463)*x[0]**o+(0.568438094437)*x[0]+(0.283670966088)*x[1]**o+(0.200206672418)*x[1]
        arg[2,4,2,1]=(-0.181938802717)*x[0]**o+(0.850684926475)*x[0]+(-0.765889361148)*x[1]**o+(0.750762858672)*x[1]
        arg[3,0,0,0]=(0.618955933151)*x[0]**o+(0.0280502885803)*x[0]+(-0.518282623282)*x[1]**o+(-0.741257199588)*x[1]
        arg[3,0,0,1]=(0.122462861306)*x[0]**o+(-0.103190806023)*x[0]+(0.647959918767)*x[1]**o+(0.531727405032)*x[1]
        arg[3,0,1,0]=(-0.129468664704)*x[0]**o+(0.342288003693)*x[0]+(-0.0268246756133)*x[1]**o+(0.671147920273)*x[1]
        arg[3,0,1,1]=(-0.99153395274)*x[0]**o+(0.584874556271)*x[0]+(-0.29273067599)*x[1]**o+(-0.802100630404)*x[1]
        arg[3,0,2,0]=(0.406033817569)*x[0]**o+(-0.664112618189)*x[0]+(0.970253482906)*x[1]**o+(0.979709129918)*x[1]
        arg[3,0,2,1]=(-0.708798148641)*x[0]**o+(-0.574259437301)*x[0]+(0.305353886124)*x[1]**o+(0.90520799029)*x[1]
        arg[3,1,0,0]=(-0.296777292529)*x[0]**o+(0.788817860427)*x[0]+(-0.817934426475)*x[1]**o+(-0.480405944642)*x[1]
        arg[3,1,0,1]=(0.382866653892)*x[0]**o+(-0.0911340124698)*x[0]+(0.449555911305)*x[1]**o+(0.972590525904)*x[1]
        arg[3,1,1,0]=(0.150349675921)*x[0]**o+(0.316401916978)*x[0]+(0.176755455021)*x[1]**o+(-0.895165552576)*x[1]
        arg[3,1,1,1]=(-0.334194603044)*x[0]**o+(0.40353334527)*x[0]+(0.510619331296)*x[1]**o+(-0.423857458769)*x[1]
        arg[3,1,2,0]=(0.876266380991)*x[0]**o+(-0.1146693887)*x[0]+(0.0388933985336)*x[1]**o+(0.537276050219)*x[1]
        arg[3,1,2,1]=(0.953489127278)*x[0]**o+(-0.40088917811)*x[0]+(0.322112015322)*x[1]**o+(0.570451651396)*x[1]
        arg[3,2,0,0]=(-0.392674815471)*x[0]**o+(0.805725364633)*x[0]+(-0.550131544344)*x[1]**o+(-0.902042986324)*x[1]
        arg[3,2,0,1]=(0.996009652754)*x[0]**o+(0.534022968708)*x[0]+(-0.251810035763)*x[1]**o+(-0.324231235105)*x[1]
        arg[3,2,1,0]=(-0.280691594637)*x[0]**o+(0.716441336919)*x[0]+(0.444717298591)*x[1]**o+(-0.417682576675)*x[1]
        arg[3,2,1,1]=(-0.528033696021)*x[0]**o+(0.105118392268)*x[0]+(0.219345332231)*x[1]**o+(-0.554011705858)*x[1]
        arg[3,2,2,0]=(0.869967952526)*x[0]**o+(-0.285204973916)*x[0]+(0.676958471189)*x[1]**o+(-0.109813007662)*x[1]
        arg[3,2,2,1]=(-0.183338615818)*x[0]**o+(0.124214512553)*x[0]+(-0.00262420395052)*x[1]**o+(0.505099322191)*x[1]
        arg[3,3,0,0]=(-0.77019083751)*x[0]**o+(-0.288304317474)*x[0]+(-0.674724525333)*x[1]**o+(0.6230658909)*x[1]
        arg[3,3,0,1]=(0.4027332007)*x[0]**o+(0.0181383553658)*x[0]+(0.0346834190431)*x[1]**o+(-0.343302420746)*x[1]
        arg[3,3,1,0]=(0.0907452191096)*x[0]**o+(-0.734197819823)*x[0]+(-0.262536956938)*x[1]**o+(-0.838964523281)*x[1]
        arg[3,3,1,1]=(0.362001847336)*x[0]**o+(-0.99196319473)*x[0]+(-0.544565811529)*x[1]**o+(0.0460139495686)*x[1]
        arg[3,3,2,0]=(0.43659913886)*x[0]**o+(-0.42986865456)*x[0]+(-0.298749218949)*x[1]**o+(0.927334871957)*x[1]
        arg[3,3,2,1]=(0.400272592051)*x[0]**o+(0.332756541456)*x[0]+(0.971127734155)*x[1]**o+(0.653965425034)*x[1]
        arg[3,4,0,0]=(0.663781122153)*x[0]**o+(0.697898766724)*x[0]+(0.157432061303)*x[1]**o+(-0.0526003145497)*x[1]
        arg[3,4,0,1]=(-0.463791084222)*x[0]**o+(0.384914785855)*x[0]+(-0.996485418809)*x[1]**o+(-0.519106290889)*x[1]
        arg[3,4,1,0]=(-0.598068247392)*x[0]**o+(0.553482523952)*x[0]+(0.202485340343)*x[1]**o+(-0.540515409196)*x[1]
        arg[3,4,1,1]=(0.134764979841)*x[0]**o+(0.0141136152488)*x[0]+(-0.536476467366)*x[1]**o+(0.323150057577)*x[1]
        arg[3,4,2,0]=(-0.73096888708)*x[0]**o+(0.485529166207)*x[0]+(-0.541219829073)*x[1]**o+(0.471443044424)*x[1]
        arg[3,4,2,1]=(-0.815808733094)*x[0]**o+(0.694490460113)*x[0]+(0.292301816313)*x[1]**o+(0.700014734893)*x[1]
        ref[0,0,0,0]=(-0.937777610663)*x_ref[0]**o+(-0.0228520946632)*x_ref[0]+(0.642385396583)*x_ref[1]**o+(-0.803821945812)*x_ref[1]
        ref[0,0,0,1]=(-0.398785307492)*x_ref[0]**o+(0.0450288952883)*x_ref[0]+(-0.777780548133)*x_ref[1]**o+(0.0273772099414)*x_ref[1]
        ref[0,0,1,0]=(-0.667280086925)*x_ref[0]**o+(-0.934642710274)*x_ref[0]+(0.565327494491)*x_ref[1]**o+(0.988521769056)*x_ref[1]
        ref[0,0,1,1]=(-0.225562383498)*x_ref[0]**o+(-0.280566031134)*x_ref[0]+(-0.295768954831)*x_ref[1]**o+(0.102759013802)*x_ref[1]
        ref[0,0,2,0]=(-0.179577502546)*x_ref[0]**o+(-0.156932777074)*x_ref[0]+(-0.145616153633)*x_ref[1]**o+(0.971876899135)*x_ref[1]
        ref[0,0,2,1]=(-0.840266409118)*x_ref[0]**o+(0.704356785022)*x_ref[0]+(0.740091470711)*x_ref[1]**o+(0.708478499023)*x_ref[1]
        ref[0,1,0,0]=(-0.487161662806)*x_ref[0]**o+(0.565313908399)*x_ref[0]+(-0.33708697293)*x_ref[1]**o+(0.731675966546)*x_ref[1]
        ref[0,1,0,1]=(-0.135615857963)*x_ref[0]**o+(0.265408777025)*x_ref[0]+(-0.168207260227)*x_ref[1]**o+(-0.598974659317)*x_ref[1]
        ref[0,1,1,0]=(0.0533944010258)*x_ref[0]**o+(0.819248906738)*x_ref[0]+(-0.837005251524)*x_ref[1]**o+(0.119613618595)*x_ref[1]
        ref[0,1,1,1]=(-0.257520627047)*x_ref[0]**o+(0.339264545713)*x_ref[0]+(0.137721228188)*x_ref[1]**o+(-0.512251030979)*x_ref[1]
        ref[0,1,2,0]=(-0.797070396749)*x_ref[0]**o+(0.690173285077)*x_ref[0]+(-0.213206187271)*x_ref[1]**o+(-0.873999125945)*x_ref[1]
        ref[0,1,2,1]=(-0.847662300367)*x_ref[0]**o+(0.785781114266)*x_ref[0]+(-0.456918398811)*x_ref[1]**o+(0.458881079947)*x_ref[1]
        ref[0,2,0,0]=(-0.385269163334)*x_ref[0]**o+(0.135941228964)*x_ref[0]+(0.233245620648)*x_ref[1]**o+(-0.246282209626)*x_ref[1]
        ref[0,2,0,1]=(0.372914636218)*x_ref[0]**o+(-0.160283657003)*x_ref[0]+(0.0565043889781)*x_ref[1]**o+(0.199715522021)*x_ref[1]
        ref[0,2,1,0]=(0.676198390598)*x_ref[0]**o+(0.938025122014)*x_ref[0]+(0.161074696651)*x_ref[1]**o+(-0.155495646964)*x_ref[1]
        ref[0,2,1,1]=(-0.709286924155)*x_ref[0]**o+(-0.173716284776)*x_ref[0]+(0.0576435484694)*x_ref[1]**o+(0.673924697832)*x_ref[1]
        ref[0,2,2,0]=(-0.436468652969)*x_ref[0]**o+(0.117279009082)*x_ref[0]+(-0.731436369191)*x_ref[1]**o+(0.919578503469)*x_ref[1]
        ref[0,2,2,1]=(-0.691042646151)*x_ref[0]**o+(-0.606293725574)*x_ref[0]+(0.657291692759)*x_ref[1]**o+(-0.0146398994729)*x_ref[1]
        ref[0,3,0,0]=(0.365939313719)*x_ref[0]**o+(-0.221236913792)*x_ref[0]+(-0.729594967883)*x_ref[1]**o+(0.0179024719114)*x_ref[1]
        ref[0,3,0,1]=(0.135892133258)*x_ref[0]**o+(0.728069461936)*x_ref[0]+(-0.365691795279)*x_ref[1]**o+(0.877708599922)*x_ref[1]
        ref[0,3,1,0]=(-0.853312456817)*x_ref[0]**o+(-0.713031969049)*x_ref[0]+(0.274755827464)*x_ref[1]**o+(-0.527330699033)*x_ref[1]
        ref[0,3,1,1]=(-0.30358584687)*x_ref[0]**o+(0.871629391968)*x_ref[0]+(0.906915162645)*x_ref[1]**o+(0.499955429041)*x_ref[1]
        ref[0,3,2,0]=(-0.880532587427)*x_ref[0]**o+(0.265729672258)*x_ref[0]+(0.528677136349)*x_ref[1]**o+(-0.224602964925)*x_ref[1]
        ref[0,3,2,1]=(0.158632463532)*x_ref[0]**o+(-0.665418329643)*x_ref[0]+(0.957661069067)*x_ref[1]**o+(-0.23771309434)*x_ref[1]
        ref[0,4,0,0]=(-0.128073514905)*x_ref[0]**o+(-0.452543798055)*x_ref[0]+(0.303808626277)*x_ref[1]**o+(0.841041779759)*x_ref[1]
        ref[0,4,0,1]=(0.0219309964781)*x_ref[0]**o+(-0.949488393443)*x_ref[0]+(0.0436905404663)*x_ref[1]**o+(-0.304527242795)*x_ref[1]
        ref[0,4,1,0]=(-0.792275137775)*x_ref[0]**o+(0.450079829644)*x_ref[0]+(0.162879981637)*x_ref[1]**o+(-0.185839015805)*x_ref[1]
        ref[0,4,1,1]=(0.761932713893)*x_ref[0]**o+(0.932601232945)*x_ref[0]+(-0.606328305783)*x_ref[1]**o+(-0.123973545682)*x_ref[1]
        ref[0,4,2,0]=(-0.514088304908)*x_ref[0]**o+(0.987668774807)*x_ref[0]+(0.866078887168)*x_ref[1]**o+(-0.608518224484)*x_ref[1]
        ref[0,4,2,1]=(-0.427564608028)*x_ref[0]**o+(0.231599614445)*x_ref[0]+(-0.406862229819)*x_ref[1]**o+(0.358456864265)*x_ref[1]
        ref[1,0,0,0]=(0.342092171801)*x_ref[0]**o+(-0.264522675185)*x_ref[0]+(-0.37976917867)*x_ref[1]**o+(0.778596602561)*x_ref[1]
        ref[1,0,0,1]=(0.672156473235)*x_ref[0]**o+(0.320317500856)*x_ref[0]+(0.514206881296)*x_ref[1]**o+(0.602594631705)*x_ref[1]
        ref[1,0,1,0]=(0.914472920293)*x_ref[0]**o+(0.525647943447)*x_ref[0]+(0.685613621597)*x_ref[1]**o+(0.482308039231)*x_ref[1]
        ref[1,0,1,1]=(-0.727982874425)*x_ref[0]**o+(-0.417964463065)*x_ref[0]+(0.825853988063)*x_ref[1]**o+(0.236134760252)*x_ref[1]
        ref[1,0,2,0]=(0.922186381756)*x_ref[0]**o+(-0.140522762044)*x_ref[0]+(-0.282169431666)*x_ref[1]**o+(0.778383988397)*x_ref[1]
        ref[1,0,2,1]=(0.307296736289)*x_ref[0]**o+(0.589844868569)*x_ref[0]+(0.64571457752)*x_ref[1]**o+(0.749333079485)*x_ref[1]
        ref[1,1,0,0]=(0.908345176703)*x_ref[0]**o+(-0.476053147164)*x_ref[0]+(0.0828501189289)*x_ref[1]**o+(0.280657117705)*x_ref[1]
        ref[1,1,0,1]=(0.396128814097)*x_ref[0]**o+(0.368821575482)*x_ref[0]+(-0.916584602334)*x_ref[1]**o+(-0.149377992159)*x_ref[1]
        ref[1,1,1,0]=(-0.38468474313)*x_ref[0]**o+(0.304742499301)*x_ref[0]+(0.638083865846)*x_ref[1]**o+(-0.332046460663)*x_ref[1]
        ref[1,1,1,1]=(0.845097787138)*x_ref[0]**o+(-0.34949498164)*x_ref[0]+(0.229157202572)*x_ref[1]**o+(0.677584672856)*x_ref[1]
        ref[1,1,2,0]=(0.0280692825358)*x_ref[0]**o+(0.567451370107)*x_ref[0]+(0.66934188305)*x_ref[1]**o+(-0.477197213188)*x_ref[1]
        ref[1,1,2,1]=(0.375411487264)*x_ref[0]**o+(0.55086548839)*x_ref[0]+(-0.392135351366)*x_ref[1]**o+(0.91971334589)*x_ref[1]
        ref[1,2,0,0]=(-0.7659270368)*x_ref[0]**o+(-0.338082792509)*x_ref[0]+(-0.26272975232)*x_ref[1]**o+(0.267936454568)*x_ref[1]
        ref[1,2,0,1]=(0.918125455771)*x_ref[0]**o+(0.182896655906)*x_ref[0]+(0.525396835337)*x_ref[1]**o+(0.201918789801)*x_ref[1]
        ref[1,2,1,0]=(-0.54472490266)*x_ref[0]**o+(0.693714832775)*x_ref[0]+(0.898278265794)*x_ref[1]**o+(0.345095187159)*x_ref[1]
        ref[1,2,1,1]=(-0.498607752195)*x_ref[0]**o+(0.84678344253)*x_ref[0]+(0.678992455984)*x_ref[1]**o+(-0.597396262326)*x_ref[1]
        ref[1,2,2,0]=(0.215267998554)*x_ref[0]**o+(0.384479388838)*x_ref[0]+(-0.09078862662)*x_ref[1]**o+(-0.0162185695524)*x_ref[1]
        ref[1,2,2,1]=(-0.335574750173)*x_ref[0]**o+(-0.439933690512)*x_ref[0]+(0.194601817488)*x_ref[1]**o+(0.795181445633)*x_ref[1]
        ref[1,3,0,0]=(0.862180273604)*x_ref[0]**o+(0.276662657248)*x_ref[0]+(-0.908827123328)*x_ref[1]**o+(0.290406540669)*x_ref[1]
        ref[1,3,0,1]=(-0.620441636176)*x_ref[0]**o+(0.448236189204)*x_ref[0]+(-0.503817375869)*x_ref[1]**o+(0.689447469467)*x_ref[1]
        ref[1,3,1,0]=(0.126327572417)*x_ref[0]**o+(-0.476552913848)*x_ref[0]+(0.362273273226)*x_ref[1]**o+(-0.816228830372)*x_ref[1]
        ref[1,3,1,1]=(-0.65091977023)*x_ref[0]**o+(0.0316560718061)*x_ref[0]+(-0.922906273452)*x_ref[1]**o+(-0.584811809954)*x_ref[1]
        ref[1,3,2,0]=(-0.675982262546)*x_ref[0]**o+(0.23518662745)*x_ref[0]+(-0.270229269162)*x_ref[1]**o+(-0.00300656227686)*x_ref[1]
        ref[1,3,2,1]=(0.55691633063)*x_ref[0]**o+(-0.0643123929496)*x_ref[0]+(0.212825067376)*x_ref[1]**o+(-0.0807098593265)*x_ref[1]
        ref[1,4,0,0]=(0.199100268435)*x_ref[0]**o+(-0.358028564112)*x_ref[0]+(0.130301315926)*x_ref[1]**o+(-0.255752658178)*x_ref[1]
        ref[1,4,0,1]=(0.821720568968)*x_ref[0]**o+(0.193634063182)*x_ref[0]+(0.862353589302)*x_ref[1]**o+(-0.42971617129)*x_ref[1]
        ref[1,4,1,0]=(0.51160977767)*x_ref[0]**o+(0.326743318954)*x_ref[0]+(-0.866878934278)*x_ref[1]**o+(-0.927585776234)*x_ref[1]
        ref[1,4,1,1]=(0.299835915861)*x_ref[0]**o+(-0.969027084981)*x_ref[0]+(-0.589772230832)*x_ref[1]**o+(0.382433302029)*x_ref[1]
        ref[1,4,2,0]=(0.473101160384)*x_ref[0]**o+(-0.591124830283)*x_ref[0]+(0.0929141413982)*x_ref[1]**o+(-0.529811769054)*x_ref[1]
        ref[1,4,2,1]=(-0.712363897962)*x_ref[0]**o+(-0.259229026597)*x_ref[0]+(0.00983651162685)*x_ref[1]**o+(-0.329312308231)*x_ref[1]
        ref[2,0,0,0]=(-0.507414918232)*x_ref[0]**o+(0.668167664289)*x_ref[0]+(0.654870826953)*x_ref[1]**o+(-0.411963068213)*x_ref[1]
        ref[2,0,0,1]=(-0.894915678458)*x_ref[0]**o+(0.647579368695)*x_ref[0]+(0.222274441569)*x_ref[1]**o+(0.737972296743)*x_ref[1]
        ref[2,0,1,0]=(-0.457176106717)*x_ref[0]**o+(0.99997698232)*x_ref[0]+(0.524381951875)*x_ref[1]**o+(0.796302331418)*x_ref[1]
        ref[2,0,1,1]=(-0.761912711581)*x_ref[0]**o+(-0.995235985398)*x_ref[0]+(-0.371926591709)*x_ref[1]**o+(-0.37949019966)*x_ref[1]
        ref[2,0,2,0]=(0.402826798397)*x_ref[0]**o+(0.473547871218)*x_ref[0]+(0.801636706721)*x_ref[1]**o+(0.314221308918)*x_ref[1]
        ref[2,0,2,1]=(0.0416846902183)*x_ref[0]**o+(0.815306853966)*x_ref[0]+(-0.526315166764)*x_ref[1]**o+(-0.327935702076)*x_ref[1]
        ref[2,1,0,0]=(-0.53829655239)*x_ref[0]**o+(-0.467914397053)*x_ref[0]+(0.392659650792)*x_ref[1]**o+(0.309531884089)*x_ref[1]
        ref[2,1,0,1]=(-0.768324371377)*x_ref[0]**o+(0.136336455054)*x_ref[0]+(0.0653485752909)*x_ref[1]**o+(-0.0181923467911)*x_ref[1]
        ref[2,1,1,0]=(0.121561771322)*x_ref[0]**o+(-0.794575616334)*x_ref[0]+(0.038194779569)*x_ref[1]**o+(-0.695864614551)*x_ref[1]
        ref[2,1,1,1]=(0.941872575674)*x_ref[0]**o+(0.322067879395)*x_ref[0]+(0.176873583908)*x_ref[1]**o+(0.266964433323)*x_ref[1]
        ref[2,1,2,0]=(0.324335817101)*x_ref[0]**o+(0.609048475305)*x_ref[0]+(0.350973789003)*x_ref[1]**o+(-0.284995830372)*x_ref[1]
        ref[2,1,2,1]=(0.939941132515)*x_ref[0]**o+(0.782468124836)*x_ref[0]+(-0.238539378134)*x_ref[1]**o+(-0.515203802574)*x_ref[1]
        ref[2,2,0,0]=(0.080179317168)*x_ref[0]**o+(-0.555769741953)*x_ref[0]+(0.283304726007)*x_ref[1]**o+(0.844935300942)*x_ref[1]
        ref[2,2,0,1]=(0.52001819752)*x_ref[0]**o+(-0.105912273152)*x_ref[0]+(-0.360523424917)*x_ref[1]**o+(-0.179135541646)*x_ref[1]
        ref[2,2,1,0]=(0.719593478448)*x_ref[0]**o+(-0.722945454687)*x_ref[0]+(0.0316404625439)*x_ref[1]**o+(0.154290328891)*x_ref[1]
        ref[2,2,1,1]=(0.906382302919)*x_ref[0]**o+(-0.137753840588)*x_ref[0]+(-0.626417865101)*x_ref[1]**o+(-0.890642745149)*x_ref[1]
        ref[2,2,2,0]=(0.618763086671)*x_ref[0]**o+(-0.267876013516)*x_ref[0]+(0.582353922702)*x_ref[1]**o+(-0.190278788501)*x_ref[1]
        ref[2,2,2,1]=(0.778824546516)*x_ref[0]**o+(-0.00837218512008)*x_ref[0]+(0.576376019084)*x_ref[1]**o+(-0.173470710629)*x_ref[1]
        ref[2,3,0,0]=(0.351742111948)*x_ref[0]**o+(-0.641801256932)*x_ref[0]+(-0.979232109277)*x_ref[1]**o+(0.973197715414)*x_ref[1]
        ref[2,3,0,1]=(-0.194493570182)*x_ref[0]**o+(-0.24088269914)*x_ref[0]+(-0.594521666623)*x_ref[1]**o+(0.442357460052)*x_ref[1]
        ref[2,3,1,0]=(-0.0677209625916)*x_ref[0]**o+(-0.329246651504)*x_ref[0]+(-0.200109031557)*x_ref[1]**o+(0.0666064469254)*x_ref[1]
        ref[2,3,1,1]=(-0.31026797404)*x_ref[0]**o+(-0.0420778789656)*x_ref[0]+(-0.534104800904)*x_ref[1]**o+(0.940011883171)*x_ref[1]
        ref[2,3,2,0]=(0.572825859789)*x_ref[0]**o+(-0.439819027899)*x_ref[0]+(-0.641214796724)*x_ref[1]**o+(0.281317897435)*x_ref[1]
        ref[2,3,2,1]=(0.410920070755)*x_ref[0]**o+(-0.517259918184)*x_ref[0]+(0.436420469387)*x_ref[1]**o+(0.212276585342)*x_ref[1]
        ref[2,4,0,0]=(0.0537540090315)*x_ref[0]**o+(-0.362577123782)*x_ref[0]+(-0.0134992480023)*x_ref[1]**o+(0.59055317928)*x_ref[1]
        ref[2,4,0,1]=(0.894896469792)*x_ref[0]**o+(-0.594219811149)*x_ref[0]+(0.755490579022)*x_ref[1]**o+(-0.222913251691)*x_ref[1]
        ref[2,4,1,0]=(-0.95471806813)*x_ref[0]**o+(0.364260591437)*x_ref[0]+(-0.51232745259)*x_ref[1]**o+(-0.874795288001)*x_ref[1]
        ref[2,4,1,1]=(-0.146483465086)*x_ref[0]**o+(0.182495641134)*x_ref[0]+(-0.132211746968)*x_ref[1]**o+(0.39915899917)*x_ref[1]
        ref[2,4,2,0]=(0.630856642463)*x_ref[0]**o+(0.568438094437)*x_ref[0]+(0.283670966088)*x_ref[1]**o+(0.200206672418)*x_ref[1]
        ref[2,4,2,1]=(-0.181938802717)*x_ref[0]**o+(0.850684926475)*x_ref[0]+(-0.765889361148)*x_ref[1]**o+(0.750762858672)*x_ref[1]
        ref[3,0,0,0]=(0.618955933151)*x_ref[0]**o+(0.0280502885803)*x_ref[0]+(-0.518282623282)*x_ref[1]**o+(-0.741257199588)*x_ref[1]
        ref[3,0,0,1]=(0.122462861306)*x_ref[0]**o+(-0.103190806023)*x_ref[0]+(0.647959918767)*x_ref[1]**o+(0.531727405032)*x_ref[1]
        ref[3,0,1,0]=(-0.129468664704)*x_ref[0]**o+(0.342288003693)*x_ref[0]+(-0.0268246756133)*x_ref[1]**o+(0.671147920273)*x_ref[1]
        ref[3,0,1,1]=(-0.99153395274)*x_ref[0]**o+(0.584874556271)*x_ref[0]+(-0.29273067599)*x_ref[1]**o+(-0.802100630404)*x_ref[1]
        ref[3,0,2,0]=(0.406033817569)*x_ref[0]**o+(-0.664112618189)*x_ref[0]+(0.970253482906)*x_ref[1]**o+(0.979709129918)*x_ref[1]
        ref[3,0,2,1]=(-0.708798148641)*x_ref[0]**o+(-0.574259437301)*x_ref[0]+(0.305353886124)*x_ref[1]**o+(0.90520799029)*x_ref[1]
        ref[3,1,0,0]=(-0.296777292529)*x_ref[0]**o+(0.788817860427)*x_ref[0]+(-0.817934426475)*x_ref[1]**o+(-0.480405944642)*x_ref[1]
        ref[3,1,0,1]=(0.382866653892)*x_ref[0]**o+(-0.0911340124698)*x_ref[0]+(0.449555911305)*x_ref[1]**o+(0.972590525904)*x_ref[1]
        ref[3,1,1,0]=(0.150349675921)*x_ref[0]**o+(0.316401916978)*x_ref[0]+(0.176755455021)*x_ref[1]**o+(-0.895165552576)*x_ref[1]
        ref[3,1,1,1]=(-0.334194603044)*x_ref[0]**o+(0.40353334527)*x_ref[0]+(0.510619331296)*x_ref[1]**o+(-0.423857458769)*x_ref[1]
        ref[3,1,2,0]=(0.876266380991)*x_ref[0]**o+(-0.1146693887)*x_ref[0]+(0.0388933985336)*x_ref[1]**o+(0.537276050219)*x_ref[1]
        ref[3,1,2,1]=(0.953489127278)*x_ref[0]**o+(-0.40088917811)*x_ref[0]+(0.322112015322)*x_ref[1]**o+(0.570451651396)*x_ref[1]
        ref[3,2,0,0]=(-0.392674815471)*x_ref[0]**o+(0.805725364633)*x_ref[0]+(-0.550131544344)*x_ref[1]**o+(-0.902042986324)*x_ref[1]
        ref[3,2,0,1]=(0.996009652754)*x_ref[0]**o+(0.534022968708)*x_ref[0]+(-0.251810035763)*x_ref[1]**o+(-0.324231235105)*x_ref[1]
        ref[3,2,1,0]=(-0.280691594637)*x_ref[0]**o+(0.716441336919)*x_ref[0]+(0.444717298591)*x_ref[1]**o+(-0.417682576675)*x_ref[1]
        ref[3,2,1,1]=(-0.528033696021)*x_ref[0]**o+(0.105118392268)*x_ref[0]+(0.219345332231)*x_ref[1]**o+(-0.554011705858)*x_ref[1]
        ref[3,2,2,0]=(0.869967952526)*x_ref[0]**o+(-0.285204973916)*x_ref[0]+(0.676958471189)*x_ref[1]**o+(-0.109813007662)*x_ref[1]
        ref[3,2,2,1]=(-0.183338615818)*x_ref[0]**o+(0.124214512553)*x_ref[0]+(-0.00262420395052)*x_ref[1]**o+(0.505099322191)*x_ref[1]
        ref[3,3,0,0]=(-0.77019083751)*x_ref[0]**o+(-0.288304317474)*x_ref[0]+(-0.674724525333)*x_ref[1]**o+(0.6230658909)*x_ref[1]
        ref[3,3,0,1]=(0.4027332007)*x_ref[0]**o+(0.0181383553658)*x_ref[0]+(0.0346834190431)*x_ref[1]**o+(-0.343302420746)*x_ref[1]
        ref[3,3,1,0]=(0.0907452191096)*x_ref[0]**o+(-0.734197819823)*x_ref[0]+(-0.262536956938)*x_ref[1]**o+(-0.838964523281)*x_ref[1]
        ref[3,3,1,1]=(0.362001847336)*x_ref[0]**o+(-0.99196319473)*x_ref[0]+(-0.544565811529)*x_ref[1]**o+(0.0460139495686)*x_ref[1]
        ref[3,3,2,0]=(0.43659913886)*x_ref[0]**o+(-0.42986865456)*x_ref[0]+(-0.298749218949)*x_ref[1]**o+(0.927334871957)*x_ref[1]
        ref[3,3,2,1]=(0.400272592051)*x_ref[0]**o+(0.332756541456)*x_ref[0]+(0.971127734155)*x_ref[1]**o+(0.653965425034)*x_ref[1]
        ref[3,4,0,0]=(0.663781122153)*x_ref[0]**o+(0.697898766724)*x_ref[0]+(0.157432061303)*x_ref[1]**o+(-0.0526003145497)*x_ref[1]
        ref[3,4,0,1]=(-0.463791084222)*x_ref[0]**o+(0.384914785855)*x_ref[0]+(-0.996485418809)*x_ref[1]**o+(-0.519106290889)*x_ref[1]
        ref[3,4,1,0]=(-0.598068247392)*x_ref[0]**o+(0.553482523952)*x_ref[0]+(0.202485340343)*x_ref[1]**o+(-0.540515409196)*x_ref[1]
        ref[3,4,1,1]=(0.134764979841)*x_ref[0]**o+(0.0141136152488)*x_ref[0]+(-0.536476467366)*x_ref[1]**o+(0.323150057577)*x_ref[1]
        ref[3,4,2,0]=(-0.73096888708)*x_ref[0]**o+(0.485529166207)*x_ref[0]+(-0.541219829073)*x_ref[1]**o+(0.471443044424)*x_ref[1]
        ref[3,4,2,1]=(-0.815808733094)*x_ref[0]**o+(0.694490460113)*x_ref[0]+(0.292301816313)*x_ref[1]**o+(0.700014734893)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.244520699379)*x[0]**o+(-0.125771252842)*x[0]+(0.654557798928)*x[1]**o+(0.438110948503)*x[1]+(0.399996169353)*x[2]**o+(0.37037141324)*x[2]
        arg[0,0,0,1]=(0.019693348021)*x[0]**o+(-0.509938412509)*x[0]+(-0.425852907243)*x[1]**o+(0.351462945072)*x[1]+(0.628410872426)*x[2]**o+(0.4769252017)*x[2]
        arg[0,0,1,0]=(0.623813767602)*x[0]**o+(-0.940818977456)*x[0]+(-0.00111063869285)*x[1]**o+(-0.577558081394)*x[1]+(-0.30810971068)*x[2]**o+(-0.660149007373)*x[2]
        arg[0,0,1,1]=(0.994105796371)*x[0]**o+(0.199970906966)*x[0]+(0.961509001434)*x[1]**o+(-0.783069602127)*x[1]+(-0.179922863021)*x[2]**o+(-0.386974503089)*x[2]
        arg[0,0,2,0]=(0.932745137122)*x[0]**o+(0.627995830114)*x[0]+(-0.334344490736)*x[1]**o+(0.733210590815)*x[1]+(-0.23424959051)*x[2]**o+(0.140644214958)*x[2]
        arg[0,0,2,1]=(-0.282836519247)*x[0]**o+(-0.744077669768)*x[0]+(0.123898574243)*x[1]**o+(0.562384289004)*x[1]+(0.0757927394637)*x[2]**o+(0.767684067921)*x[2]
        arg[0,1,0,0]=(0.212366202901)*x[0]**o+(-0.679228304269)*x[0]+(0.829706652419)*x[1]**o+(0.24389996073)*x[1]+(0.909367025978)*x[2]**o+(0.280617790407)*x[2]
        arg[0,1,0,1]=(0.760070952632)*x[0]**o+(-0.678309252372)*x[0]+(0.0372281910363)*x[1]**o+(0.0284582814297)*x[1]+(-0.092349250941)*x[2]**o+(0.347713645921)*x[2]
        arg[0,1,1,0]=(0.248410230191)*x[0]**o+(-0.833983202295)*x[0]+(-0.759357278899)*x[1]**o+(0.880038285893)*x[1]+(-0.734799540652)*x[2]**o+(0.493330231104)*x[2]
        arg[0,1,1,1]=(0.0359919724228)*x[0]**o+(-0.0393899014057)*x[0]+(0.185155782488)*x[1]**o+(-0.92692925719)*x[1]+(0.74718037404)*x[2]**o+(-0.283900681763)*x[2]
        arg[0,1,2,0]=(-0.208643544667)*x[0]**o+(0.368200149096)*x[0]+(-0.327032725172)*x[1]**o+(-0.245547905792)*x[1]+(0.0886488214242)*x[2]**o+(0.179139897691)*x[2]
        arg[0,1,2,1]=(-0.0618762861459)*x[0]**o+(0.253419217897)*x[0]+(0.840270125508)*x[1]**o+(0.0223556878387)*x[1]+(0.0560904700176)*x[2]**o+(-0.646231049424)*x[2]
        arg[0,2,0,0]=(0.568966969865)*x[0]**o+(0.786776511348)*x[0]+(0.150595926116)*x[1]**o+(-0.780164417762)*x[1]+(0.510149831242)*x[2]**o+(-0.902871929903)*x[2]
        arg[0,2,0,1]=(-0.817062669957)*x[0]**o+(0.626805486868)*x[0]+(-0.566129786829)*x[1]**o+(0.248418548905)*x[1]+(-0.273995326019)*x[2]**o+(0.402023521018)*x[2]
        arg[0,2,1,0]=(-0.160276235113)*x[0]**o+(-0.0732543937053)*x[0]+(-0.0651595982698)*x[1]**o+(-0.529463476237)*x[1]+(0.346098812945)*x[2]**o+(0.418367016439)*x[2]
        arg[0,2,1,1]=(0.736907408498)*x[0]**o+(-0.123612748157)*x[0]+(0.000834659108651)*x[1]**o+(0.568833761363)*x[1]+(-0.542676953582)*x[2]**o+(0.026182553484)*x[2]
        arg[0,2,2,0]=(0.320985918151)*x[0]**o+(0.0512654311644)*x[0]+(-0.0313412913713)*x[1]**o+(0.295659930913)*x[1]+(-0.989560405329)*x[2]**o+(-0.107344246298)*x[2]
        arg[0,2,2,1]=(-0.957793939949)*x[0]**o+(-0.177591880789)*x[0]+(0.0785829772689)*x[1]**o+(0.121445956612)*x[1]+(-0.384134176508)*x[2]**o+(-0.775433914622)*x[2]
        arg[0,3,0,0]=(-0.130620609021)*x[0]**o+(0.268283814564)*x[0]+(-0.192835222037)*x[1]**o+(0.235341688541)*x[1]+(0.957069315916)*x[2]**o+(-0.708853567529)*x[2]
        arg[0,3,0,1]=(-0.377028600614)*x[0]**o+(-0.858446162753)*x[0]+(-0.850630580491)*x[1]**o+(-0.810676898206)*x[1]+(-0.613330062817)*x[2]**o+(0.714335738947)*x[2]
        arg[0,3,1,0]=(-0.602336649415)*x[0]**o+(-0.166667559558)*x[0]+(0.962948866081)*x[1]**o+(0.939531483467)*x[1]+(0.267767294244)*x[2]**o+(0.247155872428)*x[2]
        arg[0,3,1,1]=(0.838505391049)*x[0]**o+(0.115204438919)*x[0]+(0.989010581013)*x[1]**o+(-0.649789818418)*x[1]+(0.757328444925)*x[2]**o+(0.825158129449)*x[2]
        arg[0,3,2,0]=(0.724488034506)*x[0]**o+(0.0333093325925)*x[0]+(0.864643652652)*x[1]**o+(-0.557517585925)*x[1]+(-0.340534672063)*x[2]**o+(-0.00425705176271)*x[2]
        arg[0,3,2,1]=(-0.366459483792)*x[0]**o+(0.257706980646)*x[0]+(0.821731531023)*x[1]**o+(-0.212501303617)*x[1]+(-0.128977067795)*x[2]**o+(-0.931439847041)*x[2]
        arg[0,4,0,0]=(0.398430239547)*x[0]**o+(-0.00991917699933)*x[0]+(-0.182471763172)*x[1]**o+(0.218327764946)*x[1]+(-0.484510127696)*x[2]**o+(-0.360678193996)*x[2]
        arg[0,4,0,1]=(0.0640680415135)*x[0]**o+(0.671150063087)*x[0]+(0.208610130381)*x[1]**o+(-0.441252393083)*x[1]+(0.365150146582)*x[2]**o+(-0.228346558142)*x[2]
        arg[0,4,1,0]=(-0.852694824303)*x[0]**o+(0.181041494463)*x[0]+(0.898413617086)*x[1]**o+(-0.918988862595)*x[1]+(0.169010143324)*x[2]**o+(0.540458809834)*x[2]
        arg[0,4,1,1]=(-0.213782550366)*x[0]**o+(-0.180557891041)*x[0]+(0.23144059382)*x[1]**o+(0.0609482143429)*x[1]+(0.89256221946)*x[2]**o+(0.977857158982)*x[2]
        arg[0,4,2,0]=(0.320772898841)*x[0]**o+(0.371610189518)*x[0]+(0.692264150983)*x[1]**o+(-0.751154935415)*x[1]+(-0.163272902772)*x[2]**o+(0.0635386628764)*x[2]
        arg[0,4,2,1]=(-0.316885908577)*x[0]**o+(0.476104431241)*x[0]+(-0.647713072489)*x[1]**o+(0.683813449908)*x[1]+(0.699293342945)*x[2]**o+(-0.0373906770097)*x[2]
        arg[1,0,0,0]=(-0.030329265733)*x[0]**o+(-0.0115761157375)*x[0]+(-0.30993823088)*x[1]**o+(0.628877975389)*x[1]+(-0.582940961417)*x[2]**o+(0.154786570457)*x[2]
        arg[1,0,0,1]=(-0.593610753495)*x[0]**o+(-0.822232097764)*x[0]+(-0.465144000698)*x[1]**o+(-0.62576868204)*x[1]+(-0.53588531407)*x[2]**o+(0.119119578421)*x[2]
        arg[1,0,1,0]=(-0.15668734059)*x[0]**o+(-0.950658811047)*x[0]+(-0.267969935748)*x[1]**o+(0.0919415340304)*x[1]+(-0.0445382452662)*x[2]**o+(0.949984936459)*x[2]
        arg[1,0,1,1]=(-0.762474809436)*x[0]**o+(0.794314442281)*x[0]+(0.617742930372)*x[1]**o+(-0.502057530669)*x[1]+(0.42532154189)*x[2]**o+(0.343417687179)*x[2]
        arg[1,0,2,0]=(0.259583418084)*x[0]**o+(0.460122326434)*x[0]+(0.875185242587)*x[1]**o+(-0.621232091699)*x[1]+(0.922536397008)*x[2]**o+(-0.494729334323)*x[2]
        arg[1,0,2,1]=(0.79625863156)*x[0]**o+(-0.980360881506)*x[0]+(0.947721778185)*x[1]**o+(-0.234201025597)*x[1]+(-0.974251961452)*x[2]**o+(-0.300484605111)*x[2]
        arg[1,1,0,0]=(0.719927358834)*x[0]**o+(-0.296944527139)*x[0]+(0.970290716632)*x[1]**o+(0.45656092862)*x[1]+(0.657023751979)*x[2]**o+(-0.703898365316)*x[2]
        arg[1,1,0,1]=(-0.0600633107738)*x[0]**o+(0.525293730667)*x[0]+(0.17955637601)*x[1]**o+(-0.625782540859)*x[1]+(0.205255859488)*x[2]**o+(-0.963128832741)*x[2]
        arg[1,1,1,0]=(0.119721921725)*x[0]**o+(0.957918016103)*x[0]+(-0.574883179285)*x[1]**o+(0.915580622743)*x[1]+(-0.432538225757)*x[2]**o+(-0.937539574986)*x[2]
        arg[1,1,1,1]=(-0.295639299942)*x[0]**o+(-0.703954000058)*x[0]+(-0.185001166055)*x[1]**o+(-0.943933820897)*x[1]+(-0.12692007135)*x[2]**o+(0.476481616585)*x[2]
        arg[1,1,2,0]=(0.247892010246)*x[0]**o+(0.901198196531)*x[0]+(-0.586162224974)*x[1]**o+(-0.934177552379)*x[1]+(0.704265144051)*x[2]**o+(-0.688632953989)*x[2]
        arg[1,1,2,1]=(-0.116039842334)*x[0]**o+(-0.300373025101)*x[0]+(-0.286684372551)*x[1]**o+(-0.452337760059)*x[1]+(0.883210026596)*x[2]**o+(-0.319739612634)*x[2]
        arg[1,2,0,0]=(-0.0326555626249)*x[0]**o+(-0.0933559852573)*x[0]+(0.807443822382)*x[1]**o+(-0.148842350195)*x[1]+(-0.118904471983)*x[2]**o+(-0.740440145747)*x[2]
        arg[1,2,0,1]=(0.313623768488)*x[0]**o+(0.524958399998)*x[0]+(0.876747001712)*x[1]**o+(-0.468458894694)*x[1]+(0.52118935613)*x[2]**o+(-0.776405429712)*x[2]
        arg[1,2,1,0]=(0.1328367779)*x[0]**o+(-0.959628068869)*x[0]+(0.746000325891)*x[1]**o+(-0.454705589649)*x[1]+(0.08534697763)*x[2]**o+(0.174892567303)*x[2]
        arg[1,2,1,1]=(-0.482594997843)*x[0]**o+(0.765628170342)*x[0]+(0.103333249513)*x[1]**o+(0.840778499017)*x[1]+(0.136748193255)*x[2]**o+(0.771613240623)*x[2]
        arg[1,2,2,0]=(-0.305745376076)*x[0]**o+(-0.0464234361529)*x[0]+(-0.163528204043)*x[1]**o+(0.80782719974)*x[1]+(0.131046315567)*x[2]**o+(0.0849217124905)*x[2]
        arg[1,2,2,1]=(0.564915635291)*x[0]**o+(-0.162467524139)*x[0]+(-0.696721556806)*x[1]**o+(0.0172839552576)*x[1]+(-0.630040402586)*x[2]**o+(-0.181232689117)*x[2]
        arg[1,3,0,0]=(-0.768631958549)*x[0]**o+(0.537076196504)*x[0]+(-0.403489871013)*x[1]**o+(0.831966318632)*x[1]+(0.602387061638)*x[2]**o+(0.330495470463)*x[2]
        arg[1,3,0,1]=(0.770878177942)*x[0]**o+(-0.678609489017)*x[0]+(-0.895200519099)*x[1]**o+(0.935093561205)*x[1]+(0.0497551733469)*x[2]**o+(0.820432808383)*x[2]
        arg[1,3,1,0]=(0.052867263447)*x[0]**o+(0.566167926663)*x[0]+(-0.00397870350455)*x[1]**o+(-0.780374460006)*x[1]+(-0.166645199933)*x[2]**o+(-0.867274057327)*x[2]
        arg[1,3,1,1]=(0.0573446476183)*x[0]**o+(-0.699343104682)*x[0]+(0.428624411932)*x[1]**o+(-0.400834963806)*x[1]+(-0.722750021776)*x[2]**o+(0.386821944027)*x[2]
        arg[1,3,2,0]=(0.390777333115)*x[0]**o+(0.191564882434)*x[0]+(-0.774547811812)*x[1]**o+(0.747678394063)*x[1]+(0.662273739757)*x[2]**o+(-0.295629568405)*x[2]
        arg[1,3,2,1]=(-0.160077661734)*x[0]**o+(-0.512286879466)*x[0]+(0.616824388099)*x[1]**o+(0.0257995114324)*x[1]+(0.644270949659)*x[2]**o+(0.243520087896)*x[2]
        arg[1,4,0,0]=(-0.980501571755)*x[0]**o+(0.844184850883)*x[0]+(-0.608509998911)*x[1]**o+(-0.347687895783)*x[1]+(-0.536534228926)*x[2]**o+(0.632821444072)*x[2]
        arg[1,4,0,1]=(-0.881676821515)*x[0]**o+(-0.858728104026)*x[0]+(-0.576516592622)*x[1]**o+(0.695748958418)*x[1]+(0.856547670877)*x[2]**o+(-0.932960091743)*x[2]
        arg[1,4,1,0]=(-0.451435812217)*x[0]**o+(0.963886330308)*x[0]+(0.888532771273)*x[1]**o+(0.807924978471)*x[1]+(0.789117121354)*x[2]**o+(0.577647188972)*x[2]
        arg[1,4,1,1]=(-0.315320308371)*x[0]**o+(0.182798184972)*x[0]+(0.936963228212)*x[1]**o+(-0.634429372225)*x[1]+(0.634343556662)*x[2]**o+(0.975851394002)*x[2]
        arg[1,4,2,0]=(0.858246326961)*x[0]**o+(0.0955812000864)*x[0]+(0.16143064785)*x[1]**o+(0.397013086552)*x[1]+(-0.507188929963)*x[2]**o+(0.49789827039)*x[2]
        arg[1,4,2,1]=(-0.524634529078)*x[0]**o+(0.675491034226)*x[0]+(0.858857975014)*x[1]**o+(-0.57874207047)*x[1]+(0.476551353232)*x[2]**o+(-0.622601386038)*x[2]
        arg[2,0,0,0]=(0.976563609)*x[0]**o+(0.127133211238)*x[0]+(-0.410155025204)*x[1]**o+(-0.909032700787)*x[1]+(-0.197014506331)*x[2]**o+(0.557171133933)*x[2]
        arg[2,0,0,1]=(0.434874625301)*x[0]**o+(-0.576955025374)*x[0]+(-0.466686136319)*x[1]**o+(-0.359616702831)*x[1]+(-0.916050768111)*x[2]**o+(0.900510057507)*x[2]
        arg[2,0,1,0]=(0.615259852986)*x[0]**o+(-0.685399646912)*x[0]+(-0.222742007887)*x[1]**o+(-0.487739486778)*x[1]+(-0.675225266903)*x[2]**o+(0.494623157154)*x[2]
        arg[2,0,1,1]=(-0.89129010918)*x[0]**o+(0.0051571255592)*x[0]+(0.681003954208)*x[1]**o+(0.634041720719)*x[1]+(-0.308587458722)*x[2]**o+(-0.107801539193)*x[2]
        arg[2,0,2,0]=(-0.660496698398)*x[0]**o+(-0.889315862375)*x[0]+(0.64248433679)*x[1]**o+(-0.941990511222)*x[1]+(-0.173668816201)*x[2]**o+(0.237812802184)*x[2]
        arg[2,0,2,1]=(0.847365004226)*x[0]**o+(0.802449170085)*x[0]+(0.504737931993)*x[1]**o+(-0.695546680342)*x[1]+(0.0742668699455)*x[2]**o+(0.396565506083)*x[2]
        arg[2,1,0,0]=(0.812115742561)*x[0]**o+(-0.528091349418)*x[0]+(-0.442834134691)*x[1]**o+(-0.574414379092)*x[1]+(-0.734077338221)*x[2]**o+(0.159306515061)*x[2]
        arg[2,1,0,1]=(-0.259422255041)*x[0]**o+(-0.104016445541)*x[0]+(0.3560854635)*x[1]**o+(0.997015276785)*x[1]+(-0.707278312293)*x[2]**o+(-0.975235879836)*x[2]
        arg[2,1,1,0]=(0.402099188227)*x[0]**o+(0.0320354114274)*x[0]+(0.203384927114)*x[1]**o+(-0.847804170327)*x[1]+(-0.898459903546)*x[2]**o+(0.639388498254)*x[2]
        arg[2,1,1,1]=(-0.0477209380768)*x[0]**o+(0.531574667575)*x[0]+(0.280889682833)*x[1]**o+(-0.444963671116)*x[1]+(0.33935324652)*x[2]**o+(0.943089036339)*x[2]
        arg[2,1,2,0]=(0.396038211375)*x[0]**o+(-0.54994373568)*x[0]+(0.851017234687)*x[1]**o+(-0.739787661885)*x[1]+(-0.368325897478)*x[2]**o+(-0.449096507837)*x[2]
        arg[2,1,2,1]=(0.857714720302)*x[0]**o+(-0.750668366832)*x[0]+(-0.587901076381)*x[1]**o+(-0.24906126763)*x[1]+(0.131707530536)*x[2]**o+(0.900975507373)*x[2]
        arg[2,2,0,0]=(0.783919932392)*x[0]**o+(-0.967669966647)*x[0]+(0.82052580972)*x[1]**o+(-0.711919131449)*x[1]+(-0.396059626351)*x[2]**o+(-0.265805391355)*x[2]
        arg[2,2,0,1]=(-0.0398160197307)*x[0]**o+(-0.238193541)*x[0]+(0.282277679599)*x[1]**o+(-0.0526835304963)*x[1]+(0.590871489195)*x[2]**o+(-0.925835585742)*x[2]
        arg[2,2,1,0]=(-0.825757767738)*x[0]**o+(0.238488063696)*x[0]+(0.109616380527)*x[1]**o+(-0.942941715764)*x[1]+(0.632013079754)*x[2]**o+(0.0517997920543)*x[2]
        arg[2,2,1,1]=(-0.85343230326)*x[0]**o+(0.225247381624)*x[0]+(0.43171428489)*x[1]**o+(0.0937487532231)*x[1]+(0.894064129869)*x[2]**o+(-0.0876265498234)*x[2]
        arg[2,2,2,0]=(0.349294373111)*x[0]**o+(-0.734159680014)*x[0]+(0.0489227393033)*x[1]**o+(-0.245284468757)*x[1]+(-0.346469825649)*x[2]**o+(0.547054093324)*x[2]
        arg[2,2,2,1]=(-0.0478095998047)*x[0]**o+(0.440261344405)*x[0]+(-0.590330499679)*x[1]**o+(-0.961009986328)*x[1]+(0.546190002905)*x[2]**o+(-0.027965996519)*x[2]
        arg[2,3,0,0]=(0.148563550644)*x[0]**o+(0.576704219545)*x[0]+(-0.782168581109)*x[1]**o+(-0.8301245678)*x[1]+(0.937936245682)*x[2]**o+(0.846740948371)*x[2]
        arg[2,3,0,1]=(-0.447832058524)*x[0]**o+(0.385431388609)*x[0]+(0.193985676964)*x[1]**o+(-0.91717180275)*x[1]+(-0.408732230908)*x[2]**o+(0.548173152118)*x[2]
        arg[2,3,1,0]=(0.197592567864)*x[0]**o+(-0.77657127384)*x[0]+(0.266355135001)*x[1]**o+(0.604009981584)*x[1]+(-0.379930450459)*x[2]**o+(0.46952844479)*x[2]
        arg[2,3,1,1]=(0.996803794962)*x[0]**o+(0.319115676981)*x[0]+(0.255552997875)*x[1]**o+(0.21834578367)*x[1]+(0.389966036243)*x[2]**o+(-0.115308080726)*x[2]
        arg[2,3,2,0]=(-0.0244303880675)*x[0]**o+(-0.128929049017)*x[0]+(-0.350933414406)*x[1]**o+(-0.341716440852)*x[1]+(-0.724579076377)*x[2]**o+(0.21963465519)*x[2]
        arg[2,3,2,1]=(-0.92899971478)*x[0]**o+(-0.304496074717)*x[0]+(0.568151427024)*x[1]**o+(0.504629517922)*x[1]+(0.846158711477)*x[2]**o+(0.35936792568)*x[2]
        arg[2,4,0,0]=(-0.350964834352)*x[0]**o+(0.387667833884)*x[0]+(0.547101274223)*x[1]**o+(0.257218757373)*x[1]+(-0.00316748527841)*x[2]**o+(0.314561238615)*x[2]
        arg[2,4,0,1]=(0.408941739151)*x[0]**o+(0.819113792537)*x[0]+(0.532891848781)*x[1]**o+(-0.30688659725)*x[1]+(-0.736611937618)*x[2]**o+(-0.875783706051)*x[2]
        arg[2,4,1,0]=(-0.730885165985)*x[0]**o+(0.453908506302)*x[0]+(-0.617385752505)*x[1]**o+(0.357459789592)*x[1]+(0.286323175805)*x[2]**o+(-0.448879127031)*x[2]
        arg[2,4,1,1]=(-0.295377341281)*x[0]**o+(0.423424371051)*x[0]+(0.524695076927)*x[1]**o+(-0.698499625038)*x[1]+(0.617308008395)*x[2]**o+(-0.423469866581)*x[2]
        arg[2,4,2,0]=(0.104251379635)*x[0]**o+(0.989709098795)*x[0]+(-0.422075396624)*x[1]**o+(-0.254328561231)*x[1]+(-0.502052795074)*x[2]**o+(-0.0851411652168)*x[2]
        arg[2,4,2,1]=(0.379332842082)*x[0]**o+(-0.707504861783)*x[0]+(0.00855075647146)*x[1]**o+(0.792851159033)*x[1]+(-0.692042160791)*x[2]**o+(-0.18327165243)*x[2]
        arg[3,0,0,0]=(0.310700158718)*x[0]**o+(0.285647240187)*x[0]+(0.817947571838)*x[1]**o+(0.808698040916)*x[1]+(0.844184006443)*x[2]**o+(-0.310383215636)*x[2]
        arg[3,0,0,1]=(-0.0173109952616)*x[0]**o+(0.672034452037)*x[0]+(-0.464572256298)*x[1]**o+(0.205799523299)*x[1]+(-0.971979841338)*x[2]**o+(0.261515966001)*x[2]
        arg[3,0,1,0]=(0.0329384377028)*x[0]**o+(0.571501185072)*x[0]+(-0.902222556642)*x[1]**o+(-0.528317784373)*x[1]+(0.331989640881)*x[2]**o+(0.0427505624318)*x[2]
        arg[3,0,1,1]=(0.964350410761)*x[0]**o+(-0.327198989999)*x[0]+(-0.195271523191)*x[1]**o+(0.534663014601)*x[1]+(-0.0208261208706)*x[2]**o+(-0.158019249052)*x[2]
        arg[3,0,2,0]=(-0.622082726008)*x[0]**o+(0.559463883442)*x[0]+(0.841439693313)*x[1]**o+(-0.196363987516)*x[1]+(-0.0143926547396)*x[2]**o+(0.144732603689)*x[2]
        arg[3,0,2,1]=(-0.149531653278)*x[0]**o+(0.354229138322)*x[0]+(-0.507916088829)*x[1]**o+(0.221251580047)*x[1]+(0.197213706274)*x[2]**o+(-0.555645149276)*x[2]
        arg[3,1,0,0]=(0.331020785374)*x[0]**o+(-0.377647773513)*x[0]+(0.563355928949)*x[1]**o+(0.789309216711)*x[1]+(0.203663383875)*x[2]**o+(-0.0333894540645)*x[2]
        arg[3,1,0,1]=(-0.0965613913882)*x[0]**o+(0.18656484323)*x[0]+(-0.138245941339)*x[1]**o+(-0.34549538745)*x[1]+(0.827173403167)*x[2]**o+(-0.873380503939)*x[2]
        arg[3,1,1,0]=(0.932940890033)*x[0]**o+(-0.564883830288)*x[0]+(-0.505052725295)*x[1]**o+(-0.669420210206)*x[1]+(-0.799638096039)*x[2]**o+(0.469400233749)*x[2]
        arg[3,1,1,1]=(-0.9741761069)*x[0]**o+(-0.748125328781)*x[0]+(0.664448241931)*x[1]**o+(0.0957643296979)*x[1]+(0.112211322019)*x[2]**o+(-0.131870290434)*x[2]
        arg[3,1,2,0]=(-0.438609127704)*x[0]**o+(-0.36618286515)*x[0]+(-0.376420011115)*x[1]**o+(-0.67084624599)*x[1]+(0.821679796644)*x[2]**o+(0.80111319897)*x[2]
        arg[3,1,2,1]=(-0.409859432085)*x[0]**o+(0.848767816929)*x[0]+(0.382358459835)*x[1]**o+(-0.750139843631)*x[1]+(0.587152586559)*x[2]**o+(-0.26162952001)*x[2]
        arg[3,2,0,0]=(-0.506329785074)*x[0]**o+(-0.0967745752004)*x[0]+(-0.163316795268)*x[1]**o+(0.0628481921732)*x[1]+(0.395355494367)*x[2]**o+(0.0978901916332)*x[2]
        arg[3,2,0,1]=(0.803590309587)*x[0]**o+(0.470388053389)*x[0]+(0.674942922168)*x[1]**o+(0.223472834762)*x[1]+(0.552332209028)*x[2]**o+(-0.396443498843)*x[2]
        arg[3,2,1,0]=(-0.312384723839)*x[0]**o+(0.392213828608)*x[0]+(0.449465316671)*x[1]**o+(-0.249103089401)*x[1]+(0.601412913458)*x[2]**o+(0.209899368039)*x[2]
        arg[3,2,1,1]=(-0.583776179557)*x[0]**o+(0.485007047754)*x[0]+(-0.0354153633589)*x[1]**o+(-0.898630969916)*x[1]+(-0.266581492795)*x[2]**o+(0.822086680281)*x[2]
        arg[3,2,2,0]=(0.944617094397)*x[0]**o+(-0.173057994833)*x[0]+(0.927832160048)*x[1]**o+(-0.218599594252)*x[1]+(0.349957755011)*x[2]**o+(0.392602744545)*x[2]
        arg[3,2,2,1]=(0.461648084674)*x[0]**o+(0.826213376949)*x[0]+(0.143429054731)*x[1]**o+(-0.214265128007)*x[1]+(0.0317208127607)*x[2]**o+(-0.858361078381)*x[2]
        arg[3,3,0,0]=(0.306607920368)*x[0]**o+(0.147692891688)*x[0]+(-0.562661766456)*x[1]**o+(0.448595564727)*x[1]+(-0.707178218373)*x[2]**o+(0.0655554587866)*x[2]
        arg[3,3,0,1]=(0.135475820362)*x[0]**o+(0.51420718793)*x[0]+(0.343291220736)*x[1]**o+(-0.797532312602)*x[1]+(-0.273431011273)*x[2]**o+(0.135307779253)*x[2]
        arg[3,3,1,0]=(0.488028231073)*x[0]**o+(0.542021806829)*x[0]+(0.91916268609)*x[1]**o+(0.317368273365)*x[1]+(0.314807039395)*x[2]**o+(-0.841583308025)*x[2]
        arg[3,3,1,1]=(-0.182783814917)*x[0]**o+(0.250863481201)*x[0]+(0.755551457003)*x[1]**o+(0.946171548529)*x[1]+(0.0350225436682)*x[2]**o+(0.776059830052)*x[2]
        arg[3,3,2,0]=(-0.116707124994)*x[0]**o+(-0.727010280588)*x[0]+(0.172065389268)*x[1]**o+(0.754164564639)*x[1]+(0.803777635621)*x[2]**o+(0.293490264417)*x[2]
        arg[3,3,2,1]=(0.154235982797)*x[0]**o+(-0.124286021486)*x[0]+(-0.737324748116)*x[1]**o+(-0.136803541581)*x[1]+(-0.608370191604)*x[2]**o+(-0.933599034128)*x[2]
        arg[3,4,0,0]=(0.818061677328)*x[0]**o+(-0.771719613166)*x[0]+(0.742772795251)*x[1]**o+(0.0359285091312)*x[1]+(-0.985656851837)*x[2]**o+(-0.884598206512)*x[2]
        arg[3,4,0,1]=(0.198371732741)*x[0]**o+(0.646496618681)*x[0]+(0.232912446494)*x[1]**o+(0.945619492023)*x[1]+(-0.344370341731)*x[2]**o+(0.31156840357)*x[2]
        arg[3,4,1,0]=(-0.609298834606)*x[0]**o+(0.784697392412)*x[0]+(0.996065619079)*x[1]**o+(-0.284498300043)*x[1]+(-0.652795706803)*x[2]**o+(-0.254746840396)*x[2]
        arg[3,4,1,1]=(-0.0083632822507)*x[0]**o+(0.200923618933)*x[0]+(0.353883930064)*x[1]**o+(0.153257380179)*x[1]+(0.843965464205)*x[2]**o+(0.0418501400703)*x[2]
        arg[3,4,2,0]=(-0.226055622191)*x[0]**o+(-0.701789166696)*x[0]+(0.274080957143)*x[1]**o+(0.0596744660556)*x[1]+(-0.342083152987)*x[2]**o+(0.593518464697)*x[2]
        arg[3,4,2,1]=(-0.288168945023)*x[0]**o+(-0.742841678774)*x[0]+(0.972629532424)*x[1]**o+(0.892100560733)*x[1]+(0.803880474721)*x[2]**o+(-0.0638336211224)*x[2]
        ref[0,0,0,0]=(0.244520699379)*x_ref[0]**o+(-0.125771252842)*x_ref[0]+(0.654557798928)*x_ref[1]**o+(0.438110948503)*x_ref[1]+(0.399996169353)*x_ref[2]**o+(0.37037141324)*x_ref[2]
        ref[0,0,0,1]=(0.019693348021)*x_ref[0]**o+(-0.509938412509)*x_ref[0]+(-0.425852907243)*x_ref[1]**o+(0.351462945072)*x_ref[1]+(0.628410872426)*x_ref[2]**o+(0.4769252017)*x_ref[2]
        ref[0,0,1,0]=(0.623813767602)*x_ref[0]**o+(-0.940818977456)*x_ref[0]+(-0.00111063869285)*x_ref[1]**o+(-0.577558081394)*x_ref[1]+(-0.30810971068)*x_ref[2]**o+(-0.660149007373)*x_ref[2]
        ref[0,0,1,1]=(0.994105796371)*x_ref[0]**o+(0.199970906966)*x_ref[0]+(0.961509001434)*x_ref[1]**o+(-0.783069602127)*x_ref[1]+(-0.179922863021)*x_ref[2]**o+(-0.386974503089)*x_ref[2]
        ref[0,0,2,0]=(0.932745137122)*x_ref[0]**o+(0.627995830114)*x_ref[0]+(-0.334344490736)*x_ref[1]**o+(0.733210590815)*x_ref[1]+(-0.23424959051)*x_ref[2]**o+(0.140644214958)*x_ref[2]
        ref[0,0,2,1]=(-0.282836519247)*x_ref[0]**o+(-0.744077669768)*x_ref[0]+(0.123898574243)*x_ref[1]**o+(0.562384289004)*x_ref[1]+(0.0757927394637)*x_ref[2]**o+(0.767684067921)*x_ref[2]
        ref[0,1,0,0]=(0.212366202901)*x_ref[0]**o+(-0.679228304269)*x_ref[0]+(0.829706652419)*x_ref[1]**o+(0.24389996073)*x_ref[1]+(0.909367025978)*x_ref[2]**o+(0.280617790407)*x_ref[2]
        ref[0,1,0,1]=(0.760070952632)*x_ref[0]**o+(-0.678309252372)*x_ref[0]+(0.0372281910363)*x_ref[1]**o+(0.0284582814297)*x_ref[1]+(-0.092349250941)*x_ref[2]**o+(0.347713645921)*x_ref[2]
        ref[0,1,1,0]=(0.248410230191)*x_ref[0]**o+(-0.833983202295)*x_ref[0]+(-0.759357278899)*x_ref[1]**o+(0.880038285893)*x_ref[1]+(-0.734799540652)*x_ref[2]**o+(0.493330231104)*x_ref[2]
        ref[0,1,1,1]=(0.0359919724228)*x_ref[0]**o+(-0.0393899014057)*x_ref[0]+(0.185155782488)*x_ref[1]**o+(-0.92692925719)*x_ref[1]+(0.74718037404)*x_ref[2]**o+(-0.283900681763)*x_ref[2]
        ref[0,1,2,0]=(-0.208643544667)*x_ref[0]**o+(0.368200149096)*x_ref[0]+(-0.327032725172)*x_ref[1]**o+(-0.245547905792)*x_ref[1]+(0.0886488214242)*x_ref[2]**o+(0.179139897691)*x_ref[2]
        ref[0,1,2,1]=(-0.0618762861459)*x_ref[0]**o+(0.253419217897)*x_ref[0]+(0.840270125508)*x_ref[1]**o+(0.0223556878387)*x_ref[1]+(0.0560904700176)*x_ref[2]**o+(-0.646231049424)*x_ref[2]
        ref[0,2,0,0]=(0.568966969865)*x_ref[0]**o+(0.786776511348)*x_ref[0]+(0.150595926116)*x_ref[1]**o+(-0.780164417762)*x_ref[1]+(0.510149831242)*x_ref[2]**o+(-0.902871929903)*x_ref[2]
        ref[0,2,0,1]=(-0.817062669957)*x_ref[0]**o+(0.626805486868)*x_ref[0]+(-0.566129786829)*x_ref[1]**o+(0.248418548905)*x_ref[1]+(-0.273995326019)*x_ref[2]**o+(0.402023521018)*x_ref[2]
        ref[0,2,1,0]=(-0.160276235113)*x_ref[0]**o+(-0.0732543937053)*x_ref[0]+(-0.0651595982698)*x_ref[1]**o+(-0.529463476237)*x_ref[1]+(0.346098812945)*x_ref[2]**o+(0.418367016439)*x_ref[2]
        ref[0,2,1,1]=(0.736907408498)*x_ref[0]**o+(-0.123612748157)*x_ref[0]+(0.000834659108651)*x_ref[1]**o+(0.568833761363)*x_ref[1]+(-0.542676953582)*x_ref[2]**o+(0.026182553484)*x_ref[2]
        ref[0,2,2,0]=(0.320985918151)*x_ref[0]**o+(0.0512654311644)*x_ref[0]+(-0.0313412913713)*x_ref[1]**o+(0.295659930913)*x_ref[1]+(-0.989560405329)*x_ref[2]**o+(-0.107344246298)*x_ref[2]
        ref[0,2,2,1]=(-0.957793939949)*x_ref[0]**o+(-0.177591880789)*x_ref[0]+(0.0785829772689)*x_ref[1]**o+(0.121445956612)*x_ref[1]+(-0.384134176508)*x_ref[2]**o+(-0.775433914622)*x_ref[2]
        ref[0,3,0,0]=(-0.130620609021)*x_ref[0]**o+(0.268283814564)*x_ref[0]+(-0.192835222037)*x_ref[1]**o+(0.235341688541)*x_ref[1]+(0.957069315916)*x_ref[2]**o+(-0.708853567529)*x_ref[2]
        ref[0,3,0,1]=(-0.377028600614)*x_ref[0]**o+(-0.858446162753)*x_ref[0]+(-0.850630580491)*x_ref[1]**o+(-0.810676898206)*x_ref[1]+(-0.613330062817)*x_ref[2]**o+(0.714335738947)*x_ref[2]
        ref[0,3,1,0]=(-0.602336649415)*x_ref[0]**o+(-0.166667559558)*x_ref[0]+(0.962948866081)*x_ref[1]**o+(0.939531483467)*x_ref[1]+(0.267767294244)*x_ref[2]**o+(0.247155872428)*x_ref[2]
        ref[0,3,1,1]=(0.838505391049)*x_ref[0]**o+(0.115204438919)*x_ref[0]+(0.989010581013)*x_ref[1]**o+(-0.649789818418)*x_ref[1]+(0.757328444925)*x_ref[2]**o+(0.825158129449)*x_ref[2]
        ref[0,3,2,0]=(0.724488034506)*x_ref[0]**o+(0.0333093325925)*x_ref[0]+(0.864643652652)*x_ref[1]**o+(-0.557517585925)*x_ref[1]+(-0.340534672063)*x_ref[2]**o+(-0.00425705176271)*x_ref[2]
        ref[0,3,2,1]=(-0.366459483792)*x_ref[0]**o+(0.257706980646)*x_ref[0]+(0.821731531023)*x_ref[1]**o+(-0.212501303617)*x_ref[1]+(-0.128977067795)*x_ref[2]**o+(-0.931439847041)*x_ref[2]
        ref[0,4,0,0]=(0.398430239547)*x_ref[0]**o+(-0.00991917699933)*x_ref[0]+(-0.182471763172)*x_ref[1]**o+(0.218327764946)*x_ref[1]+(-0.484510127696)*x_ref[2]**o+(-0.360678193996)*x_ref[2]
        ref[0,4,0,1]=(0.0640680415135)*x_ref[0]**o+(0.671150063087)*x_ref[0]+(0.208610130381)*x_ref[1]**o+(-0.441252393083)*x_ref[1]+(0.365150146582)*x_ref[2]**o+(-0.228346558142)*x_ref[2]
        ref[0,4,1,0]=(-0.852694824303)*x_ref[0]**o+(0.181041494463)*x_ref[0]+(0.898413617086)*x_ref[1]**o+(-0.918988862595)*x_ref[1]+(0.169010143324)*x_ref[2]**o+(0.540458809834)*x_ref[2]
        ref[0,4,1,1]=(-0.213782550366)*x_ref[0]**o+(-0.180557891041)*x_ref[0]+(0.23144059382)*x_ref[1]**o+(0.0609482143429)*x_ref[1]+(0.89256221946)*x_ref[2]**o+(0.977857158982)*x_ref[2]
        ref[0,4,2,0]=(0.320772898841)*x_ref[0]**o+(0.371610189518)*x_ref[0]+(0.692264150983)*x_ref[1]**o+(-0.751154935415)*x_ref[1]+(-0.163272902772)*x_ref[2]**o+(0.0635386628764)*x_ref[2]
        ref[0,4,2,1]=(-0.316885908577)*x_ref[0]**o+(0.476104431241)*x_ref[0]+(-0.647713072489)*x_ref[1]**o+(0.683813449908)*x_ref[1]+(0.699293342945)*x_ref[2]**o+(-0.0373906770097)*x_ref[2]
        ref[1,0,0,0]=(-0.030329265733)*x_ref[0]**o+(-0.0115761157375)*x_ref[0]+(-0.30993823088)*x_ref[1]**o+(0.628877975389)*x_ref[1]+(-0.582940961417)*x_ref[2]**o+(0.154786570457)*x_ref[2]
        ref[1,0,0,1]=(-0.593610753495)*x_ref[0]**o+(-0.822232097764)*x_ref[0]+(-0.465144000698)*x_ref[1]**o+(-0.62576868204)*x_ref[1]+(-0.53588531407)*x_ref[2]**o+(0.119119578421)*x_ref[2]
        ref[1,0,1,0]=(-0.15668734059)*x_ref[0]**o+(-0.950658811047)*x_ref[0]+(-0.267969935748)*x_ref[1]**o+(0.0919415340304)*x_ref[1]+(-0.0445382452662)*x_ref[2]**o+(0.949984936459)*x_ref[2]
        ref[1,0,1,1]=(-0.762474809436)*x_ref[0]**o+(0.794314442281)*x_ref[0]+(0.617742930372)*x_ref[1]**o+(-0.502057530669)*x_ref[1]+(0.42532154189)*x_ref[2]**o+(0.343417687179)*x_ref[2]
        ref[1,0,2,0]=(0.259583418084)*x_ref[0]**o+(0.460122326434)*x_ref[0]+(0.875185242587)*x_ref[1]**o+(-0.621232091699)*x_ref[1]+(0.922536397008)*x_ref[2]**o+(-0.494729334323)*x_ref[2]
        ref[1,0,2,1]=(0.79625863156)*x_ref[0]**o+(-0.980360881506)*x_ref[0]+(0.947721778185)*x_ref[1]**o+(-0.234201025597)*x_ref[1]+(-0.974251961452)*x_ref[2]**o+(-0.300484605111)*x_ref[2]
        ref[1,1,0,0]=(0.719927358834)*x_ref[0]**o+(-0.296944527139)*x_ref[0]+(0.970290716632)*x_ref[1]**o+(0.45656092862)*x_ref[1]+(0.657023751979)*x_ref[2]**o+(-0.703898365316)*x_ref[2]
        ref[1,1,0,1]=(-0.0600633107738)*x_ref[0]**o+(0.525293730667)*x_ref[0]+(0.17955637601)*x_ref[1]**o+(-0.625782540859)*x_ref[1]+(0.205255859488)*x_ref[2]**o+(-0.963128832741)*x_ref[2]
        ref[1,1,1,0]=(0.119721921725)*x_ref[0]**o+(0.957918016103)*x_ref[0]+(-0.574883179285)*x_ref[1]**o+(0.915580622743)*x_ref[1]+(-0.432538225757)*x_ref[2]**o+(-0.937539574986)*x_ref[2]
        ref[1,1,1,1]=(-0.295639299942)*x_ref[0]**o+(-0.703954000058)*x_ref[0]+(-0.185001166055)*x_ref[1]**o+(-0.943933820897)*x_ref[1]+(-0.12692007135)*x_ref[2]**o+(0.476481616585)*x_ref[2]
        ref[1,1,2,0]=(0.247892010246)*x_ref[0]**o+(0.901198196531)*x_ref[0]+(-0.586162224974)*x_ref[1]**o+(-0.934177552379)*x_ref[1]+(0.704265144051)*x_ref[2]**o+(-0.688632953989)*x_ref[2]
        ref[1,1,2,1]=(-0.116039842334)*x_ref[0]**o+(-0.300373025101)*x_ref[0]+(-0.286684372551)*x_ref[1]**o+(-0.452337760059)*x_ref[1]+(0.883210026596)*x_ref[2]**o+(-0.319739612634)*x_ref[2]
        ref[1,2,0,0]=(-0.0326555626249)*x_ref[0]**o+(-0.0933559852573)*x_ref[0]+(0.807443822382)*x_ref[1]**o+(-0.148842350195)*x_ref[1]+(-0.118904471983)*x_ref[2]**o+(-0.740440145747)*x_ref[2]
        ref[1,2,0,1]=(0.313623768488)*x_ref[0]**o+(0.524958399998)*x_ref[0]+(0.876747001712)*x_ref[1]**o+(-0.468458894694)*x_ref[1]+(0.52118935613)*x_ref[2]**o+(-0.776405429712)*x_ref[2]
        ref[1,2,1,0]=(0.1328367779)*x_ref[0]**o+(-0.959628068869)*x_ref[0]+(0.746000325891)*x_ref[1]**o+(-0.454705589649)*x_ref[1]+(0.08534697763)*x_ref[2]**o+(0.174892567303)*x_ref[2]
        ref[1,2,1,1]=(-0.482594997843)*x_ref[0]**o+(0.765628170342)*x_ref[0]+(0.103333249513)*x_ref[1]**o+(0.840778499017)*x_ref[1]+(0.136748193255)*x_ref[2]**o+(0.771613240623)*x_ref[2]
        ref[1,2,2,0]=(-0.305745376076)*x_ref[0]**o+(-0.0464234361529)*x_ref[0]+(-0.163528204043)*x_ref[1]**o+(0.80782719974)*x_ref[1]+(0.131046315567)*x_ref[2]**o+(0.0849217124905)*x_ref[2]
        ref[1,2,2,1]=(0.564915635291)*x_ref[0]**o+(-0.162467524139)*x_ref[0]+(-0.696721556806)*x_ref[1]**o+(0.0172839552576)*x_ref[1]+(-0.630040402586)*x_ref[2]**o+(-0.181232689117)*x_ref[2]
        ref[1,3,0,0]=(-0.768631958549)*x_ref[0]**o+(0.537076196504)*x_ref[0]+(-0.403489871013)*x_ref[1]**o+(0.831966318632)*x_ref[1]+(0.602387061638)*x_ref[2]**o+(0.330495470463)*x_ref[2]
        ref[1,3,0,1]=(0.770878177942)*x_ref[0]**o+(-0.678609489017)*x_ref[0]+(-0.895200519099)*x_ref[1]**o+(0.935093561205)*x_ref[1]+(0.0497551733469)*x_ref[2]**o+(0.820432808383)*x_ref[2]
        ref[1,3,1,0]=(0.052867263447)*x_ref[0]**o+(0.566167926663)*x_ref[0]+(-0.00397870350455)*x_ref[1]**o+(-0.780374460006)*x_ref[1]+(-0.166645199933)*x_ref[2]**o+(-0.867274057327)*x_ref[2]
        ref[1,3,1,1]=(0.0573446476183)*x_ref[0]**o+(-0.699343104682)*x_ref[0]+(0.428624411932)*x_ref[1]**o+(-0.400834963806)*x_ref[1]+(-0.722750021776)*x_ref[2]**o+(0.386821944027)*x_ref[2]
        ref[1,3,2,0]=(0.390777333115)*x_ref[0]**o+(0.191564882434)*x_ref[0]+(-0.774547811812)*x_ref[1]**o+(0.747678394063)*x_ref[1]+(0.662273739757)*x_ref[2]**o+(-0.295629568405)*x_ref[2]
        ref[1,3,2,1]=(-0.160077661734)*x_ref[0]**o+(-0.512286879466)*x_ref[0]+(0.616824388099)*x_ref[1]**o+(0.0257995114324)*x_ref[1]+(0.644270949659)*x_ref[2]**o+(0.243520087896)*x_ref[2]
        ref[1,4,0,0]=(-0.980501571755)*x_ref[0]**o+(0.844184850883)*x_ref[0]+(-0.608509998911)*x_ref[1]**o+(-0.347687895783)*x_ref[1]+(-0.536534228926)*x_ref[2]**o+(0.632821444072)*x_ref[2]
        ref[1,4,0,1]=(-0.881676821515)*x_ref[0]**o+(-0.858728104026)*x_ref[0]+(-0.576516592622)*x_ref[1]**o+(0.695748958418)*x_ref[1]+(0.856547670877)*x_ref[2]**o+(-0.932960091743)*x_ref[2]
        ref[1,4,1,0]=(-0.451435812217)*x_ref[0]**o+(0.963886330308)*x_ref[0]+(0.888532771273)*x_ref[1]**o+(0.807924978471)*x_ref[1]+(0.789117121354)*x_ref[2]**o+(0.577647188972)*x_ref[2]
        ref[1,4,1,1]=(-0.315320308371)*x_ref[0]**o+(0.182798184972)*x_ref[0]+(0.936963228212)*x_ref[1]**o+(-0.634429372225)*x_ref[1]+(0.634343556662)*x_ref[2]**o+(0.975851394002)*x_ref[2]
        ref[1,4,2,0]=(0.858246326961)*x_ref[0]**o+(0.0955812000864)*x_ref[0]+(0.16143064785)*x_ref[1]**o+(0.397013086552)*x_ref[1]+(-0.507188929963)*x_ref[2]**o+(0.49789827039)*x_ref[2]
        ref[1,4,2,1]=(-0.524634529078)*x_ref[0]**o+(0.675491034226)*x_ref[0]+(0.858857975014)*x_ref[1]**o+(-0.57874207047)*x_ref[1]+(0.476551353232)*x_ref[2]**o+(-0.622601386038)*x_ref[2]
        ref[2,0,0,0]=(0.976563609)*x_ref[0]**o+(0.127133211238)*x_ref[0]+(-0.410155025204)*x_ref[1]**o+(-0.909032700787)*x_ref[1]+(-0.197014506331)*x_ref[2]**o+(0.557171133933)*x_ref[2]
        ref[2,0,0,1]=(0.434874625301)*x_ref[0]**o+(-0.576955025374)*x_ref[0]+(-0.466686136319)*x_ref[1]**o+(-0.359616702831)*x_ref[1]+(-0.916050768111)*x_ref[2]**o+(0.900510057507)*x_ref[2]
        ref[2,0,1,0]=(0.615259852986)*x_ref[0]**o+(-0.685399646912)*x_ref[0]+(-0.222742007887)*x_ref[1]**o+(-0.487739486778)*x_ref[1]+(-0.675225266903)*x_ref[2]**o+(0.494623157154)*x_ref[2]
        ref[2,0,1,1]=(-0.89129010918)*x_ref[0]**o+(0.0051571255592)*x_ref[0]+(0.681003954208)*x_ref[1]**o+(0.634041720719)*x_ref[1]+(-0.308587458722)*x_ref[2]**o+(-0.107801539193)*x_ref[2]
        ref[2,0,2,0]=(-0.660496698398)*x_ref[0]**o+(-0.889315862375)*x_ref[0]+(0.64248433679)*x_ref[1]**o+(-0.941990511222)*x_ref[1]+(-0.173668816201)*x_ref[2]**o+(0.237812802184)*x_ref[2]
        ref[2,0,2,1]=(0.847365004226)*x_ref[0]**o+(0.802449170085)*x_ref[0]+(0.504737931993)*x_ref[1]**o+(-0.695546680342)*x_ref[1]+(0.0742668699455)*x_ref[2]**o+(0.396565506083)*x_ref[2]
        ref[2,1,0,0]=(0.812115742561)*x_ref[0]**o+(-0.528091349418)*x_ref[0]+(-0.442834134691)*x_ref[1]**o+(-0.574414379092)*x_ref[1]+(-0.734077338221)*x_ref[2]**o+(0.159306515061)*x_ref[2]
        ref[2,1,0,1]=(-0.259422255041)*x_ref[0]**o+(-0.104016445541)*x_ref[0]+(0.3560854635)*x_ref[1]**o+(0.997015276785)*x_ref[1]+(-0.707278312293)*x_ref[2]**o+(-0.975235879836)*x_ref[2]
        ref[2,1,1,0]=(0.402099188227)*x_ref[0]**o+(0.0320354114274)*x_ref[0]+(0.203384927114)*x_ref[1]**o+(-0.847804170327)*x_ref[1]+(-0.898459903546)*x_ref[2]**o+(0.639388498254)*x_ref[2]
        ref[2,1,1,1]=(-0.0477209380768)*x_ref[0]**o+(0.531574667575)*x_ref[0]+(0.280889682833)*x_ref[1]**o+(-0.444963671116)*x_ref[1]+(0.33935324652)*x_ref[2]**o+(0.943089036339)*x_ref[2]
        ref[2,1,2,0]=(0.396038211375)*x_ref[0]**o+(-0.54994373568)*x_ref[0]+(0.851017234687)*x_ref[1]**o+(-0.739787661885)*x_ref[1]+(-0.368325897478)*x_ref[2]**o+(-0.449096507837)*x_ref[2]
        ref[2,1,2,1]=(0.857714720302)*x_ref[0]**o+(-0.750668366832)*x_ref[0]+(-0.587901076381)*x_ref[1]**o+(-0.24906126763)*x_ref[1]+(0.131707530536)*x_ref[2]**o+(0.900975507373)*x_ref[2]
        ref[2,2,0,0]=(0.783919932392)*x_ref[0]**o+(-0.967669966647)*x_ref[0]+(0.82052580972)*x_ref[1]**o+(-0.711919131449)*x_ref[1]+(-0.396059626351)*x_ref[2]**o+(-0.265805391355)*x_ref[2]
        ref[2,2,0,1]=(-0.0398160197307)*x_ref[0]**o+(-0.238193541)*x_ref[0]+(0.282277679599)*x_ref[1]**o+(-0.0526835304963)*x_ref[1]+(0.590871489195)*x_ref[2]**o+(-0.925835585742)*x_ref[2]
        ref[2,2,1,0]=(-0.825757767738)*x_ref[0]**o+(0.238488063696)*x_ref[0]+(0.109616380527)*x_ref[1]**o+(-0.942941715764)*x_ref[1]+(0.632013079754)*x_ref[2]**o+(0.0517997920543)*x_ref[2]
        ref[2,2,1,1]=(-0.85343230326)*x_ref[0]**o+(0.225247381624)*x_ref[0]+(0.43171428489)*x_ref[1]**o+(0.0937487532231)*x_ref[1]+(0.894064129869)*x_ref[2]**o+(-0.0876265498234)*x_ref[2]
        ref[2,2,2,0]=(0.349294373111)*x_ref[0]**o+(-0.734159680014)*x_ref[0]+(0.0489227393033)*x_ref[1]**o+(-0.245284468757)*x_ref[1]+(-0.346469825649)*x_ref[2]**o+(0.547054093324)*x_ref[2]
        ref[2,2,2,1]=(-0.0478095998047)*x_ref[0]**o+(0.440261344405)*x_ref[0]+(-0.590330499679)*x_ref[1]**o+(-0.961009986328)*x_ref[1]+(0.546190002905)*x_ref[2]**o+(-0.027965996519)*x_ref[2]
        ref[2,3,0,0]=(0.148563550644)*x_ref[0]**o+(0.576704219545)*x_ref[0]+(-0.782168581109)*x_ref[1]**o+(-0.8301245678)*x_ref[1]+(0.937936245682)*x_ref[2]**o+(0.846740948371)*x_ref[2]
        ref[2,3,0,1]=(-0.447832058524)*x_ref[0]**o+(0.385431388609)*x_ref[0]+(0.193985676964)*x_ref[1]**o+(-0.91717180275)*x_ref[1]+(-0.408732230908)*x_ref[2]**o+(0.548173152118)*x_ref[2]
        ref[2,3,1,0]=(0.197592567864)*x_ref[0]**o+(-0.77657127384)*x_ref[0]+(0.266355135001)*x_ref[1]**o+(0.604009981584)*x_ref[1]+(-0.379930450459)*x_ref[2]**o+(0.46952844479)*x_ref[2]
        ref[2,3,1,1]=(0.996803794962)*x_ref[0]**o+(0.319115676981)*x_ref[0]+(0.255552997875)*x_ref[1]**o+(0.21834578367)*x_ref[1]+(0.389966036243)*x_ref[2]**o+(-0.115308080726)*x_ref[2]
        ref[2,3,2,0]=(-0.0244303880675)*x_ref[0]**o+(-0.128929049017)*x_ref[0]+(-0.350933414406)*x_ref[1]**o+(-0.341716440852)*x_ref[1]+(-0.724579076377)*x_ref[2]**o+(0.21963465519)*x_ref[2]
        ref[2,3,2,1]=(-0.92899971478)*x_ref[0]**o+(-0.304496074717)*x_ref[0]+(0.568151427024)*x_ref[1]**o+(0.504629517922)*x_ref[1]+(0.846158711477)*x_ref[2]**o+(0.35936792568)*x_ref[2]
        ref[2,4,0,0]=(-0.350964834352)*x_ref[0]**o+(0.387667833884)*x_ref[0]+(0.547101274223)*x_ref[1]**o+(0.257218757373)*x_ref[1]+(-0.00316748527841)*x_ref[2]**o+(0.314561238615)*x_ref[2]
        ref[2,4,0,1]=(0.408941739151)*x_ref[0]**o+(0.819113792537)*x_ref[0]+(0.532891848781)*x_ref[1]**o+(-0.30688659725)*x_ref[1]+(-0.736611937618)*x_ref[2]**o+(-0.875783706051)*x_ref[2]
        ref[2,4,1,0]=(-0.730885165985)*x_ref[0]**o+(0.453908506302)*x_ref[0]+(-0.617385752505)*x_ref[1]**o+(0.357459789592)*x_ref[1]+(0.286323175805)*x_ref[2]**o+(-0.448879127031)*x_ref[2]
        ref[2,4,1,1]=(-0.295377341281)*x_ref[0]**o+(0.423424371051)*x_ref[0]+(0.524695076927)*x_ref[1]**o+(-0.698499625038)*x_ref[1]+(0.617308008395)*x_ref[2]**o+(-0.423469866581)*x_ref[2]
        ref[2,4,2,0]=(0.104251379635)*x_ref[0]**o+(0.989709098795)*x_ref[0]+(-0.422075396624)*x_ref[1]**o+(-0.254328561231)*x_ref[1]+(-0.502052795074)*x_ref[2]**o+(-0.0851411652168)*x_ref[2]
        ref[2,4,2,1]=(0.379332842082)*x_ref[0]**o+(-0.707504861783)*x_ref[0]+(0.00855075647146)*x_ref[1]**o+(0.792851159033)*x_ref[1]+(-0.692042160791)*x_ref[2]**o+(-0.18327165243)*x_ref[2]
        ref[3,0,0,0]=(0.310700158718)*x_ref[0]**o+(0.285647240187)*x_ref[0]+(0.817947571838)*x_ref[1]**o+(0.808698040916)*x_ref[1]+(0.844184006443)*x_ref[2]**o+(-0.310383215636)*x_ref[2]
        ref[3,0,0,1]=(-0.0173109952616)*x_ref[0]**o+(0.672034452037)*x_ref[0]+(-0.464572256298)*x_ref[1]**o+(0.205799523299)*x_ref[1]+(-0.971979841338)*x_ref[2]**o+(0.261515966001)*x_ref[2]
        ref[3,0,1,0]=(0.0329384377028)*x_ref[0]**o+(0.571501185072)*x_ref[0]+(-0.902222556642)*x_ref[1]**o+(-0.528317784373)*x_ref[1]+(0.331989640881)*x_ref[2]**o+(0.0427505624318)*x_ref[2]
        ref[3,0,1,1]=(0.964350410761)*x_ref[0]**o+(-0.327198989999)*x_ref[0]+(-0.195271523191)*x_ref[1]**o+(0.534663014601)*x_ref[1]+(-0.0208261208706)*x_ref[2]**o+(-0.158019249052)*x_ref[2]
        ref[3,0,2,0]=(-0.622082726008)*x_ref[0]**o+(0.559463883442)*x_ref[0]+(0.841439693313)*x_ref[1]**o+(-0.196363987516)*x_ref[1]+(-0.0143926547396)*x_ref[2]**o+(0.144732603689)*x_ref[2]
        ref[3,0,2,1]=(-0.149531653278)*x_ref[0]**o+(0.354229138322)*x_ref[0]+(-0.507916088829)*x_ref[1]**o+(0.221251580047)*x_ref[1]+(0.197213706274)*x_ref[2]**o+(-0.555645149276)*x_ref[2]
        ref[3,1,0,0]=(0.331020785374)*x_ref[0]**o+(-0.377647773513)*x_ref[0]+(0.563355928949)*x_ref[1]**o+(0.789309216711)*x_ref[1]+(0.203663383875)*x_ref[2]**o+(-0.0333894540645)*x_ref[2]
        ref[3,1,0,1]=(-0.0965613913882)*x_ref[0]**o+(0.18656484323)*x_ref[0]+(-0.138245941339)*x_ref[1]**o+(-0.34549538745)*x_ref[1]+(0.827173403167)*x_ref[2]**o+(-0.873380503939)*x_ref[2]
        ref[3,1,1,0]=(0.932940890033)*x_ref[0]**o+(-0.564883830288)*x_ref[0]+(-0.505052725295)*x_ref[1]**o+(-0.669420210206)*x_ref[1]+(-0.799638096039)*x_ref[2]**o+(0.469400233749)*x_ref[2]
        ref[3,1,1,1]=(-0.9741761069)*x_ref[0]**o+(-0.748125328781)*x_ref[0]+(0.664448241931)*x_ref[1]**o+(0.0957643296979)*x_ref[1]+(0.112211322019)*x_ref[2]**o+(-0.131870290434)*x_ref[2]
        ref[3,1,2,0]=(-0.438609127704)*x_ref[0]**o+(-0.36618286515)*x_ref[0]+(-0.376420011115)*x_ref[1]**o+(-0.67084624599)*x_ref[1]+(0.821679796644)*x_ref[2]**o+(0.80111319897)*x_ref[2]
        ref[3,1,2,1]=(-0.409859432085)*x_ref[0]**o+(0.848767816929)*x_ref[0]+(0.382358459835)*x_ref[1]**o+(-0.750139843631)*x_ref[1]+(0.587152586559)*x_ref[2]**o+(-0.26162952001)*x_ref[2]
        ref[3,2,0,0]=(-0.506329785074)*x_ref[0]**o+(-0.0967745752004)*x_ref[0]+(-0.163316795268)*x_ref[1]**o+(0.0628481921732)*x_ref[1]+(0.395355494367)*x_ref[2]**o+(0.0978901916332)*x_ref[2]
        ref[3,2,0,1]=(0.803590309587)*x_ref[0]**o+(0.470388053389)*x_ref[0]+(0.674942922168)*x_ref[1]**o+(0.223472834762)*x_ref[1]+(0.552332209028)*x_ref[2]**o+(-0.396443498843)*x_ref[2]
        ref[3,2,1,0]=(-0.312384723839)*x_ref[0]**o+(0.392213828608)*x_ref[0]+(0.449465316671)*x_ref[1]**o+(-0.249103089401)*x_ref[1]+(0.601412913458)*x_ref[2]**o+(0.209899368039)*x_ref[2]
        ref[3,2,1,1]=(-0.583776179557)*x_ref[0]**o+(0.485007047754)*x_ref[0]+(-0.0354153633589)*x_ref[1]**o+(-0.898630969916)*x_ref[1]+(-0.266581492795)*x_ref[2]**o+(0.822086680281)*x_ref[2]
        ref[3,2,2,0]=(0.944617094397)*x_ref[0]**o+(-0.173057994833)*x_ref[0]+(0.927832160048)*x_ref[1]**o+(-0.218599594252)*x_ref[1]+(0.349957755011)*x_ref[2]**o+(0.392602744545)*x_ref[2]
        ref[3,2,2,1]=(0.461648084674)*x_ref[0]**o+(0.826213376949)*x_ref[0]+(0.143429054731)*x_ref[1]**o+(-0.214265128007)*x_ref[1]+(0.0317208127607)*x_ref[2]**o+(-0.858361078381)*x_ref[2]
        ref[3,3,0,0]=(0.306607920368)*x_ref[0]**o+(0.147692891688)*x_ref[0]+(-0.562661766456)*x_ref[1]**o+(0.448595564727)*x_ref[1]+(-0.707178218373)*x_ref[2]**o+(0.0655554587866)*x_ref[2]
        ref[3,3,0,1]=(0.135475820362)*x_ref[0]**o+(0.51420718793)*x_ref[0]+(0.343291220736)*x_ref[1]**o+(-0.797532312602)*x_ref[1]+(-0.273431011273)*x_ref[2]**o+(0.135307779253)*x_ref[2]
        ref[3,3,1,0]=(0.488028231073)*x_ref[0]**o+(0.542021806829)*x_ref[0]+(0.91916268609)*x_ref[1]**o+(0.317368273365)*x_ref[1]+(0.314807039395)*x_ref[2]**o+(-0.841583308025)*x_ref[2]
        ref[3,3,1,1]=(-0.182783814917)*x_ref[0]**o+(0.250863481201)*x_ref[0]+(0.755551457003)*x_ref[1]**o+(0.946171548529)*x_ref[1]+(0.0350225436682)*x_ref[2]**o+(0.776059830052)*x_ref[2]
        ref[3,3,2,0]=(-0.116707124994)*x_ref[0]**o+(-0.727010280588)*x_ref[0]+(0.172065389268)*x_ref[1]**o+(0.754164564639)*x_ref[1]+(0.803777635621)*x_ref[2]**o+(0.293490264417)*x_ref[2]
        ref[3,3,2,1]=(0.154235982797)*x_ref[0]**o+(-0.124286021486)*x_ref[0]+(-0.737324748116)*x_ref[1]**o+(-0.136803541581)*x_ref[1]+(-0.608370191604)*x_ref[2]**o+(-0.933599034128)*x_ref[2]
        ref[3,4,0,0]=(0.818061677328)*x_ref[0]**o+(-0.771719613166)*x_ref[0]+(0.742772795251)*x_ref[1]**o+(0.0359285091312)*x_ref[1]+(-0.985656851837)*x_ref[2]**o+(-0.884598206512)*x_ref[2]
        ref[3,4,0,1]=(0.198371732741)*x_ref[0]**o+(0.646496618681)*x_ref[0]+(0.232912446494)*x_ref[1]**o+(0.945619492023)*x_ref[1]+(-0.344370341731)*x_ref[2]**o+(0.31156840357)*x_ref[2]
        ref[3,4,1,0]=(-0.609298834606)*x_ref[0]**o+(0.784697392412)*x_ref[0]+(0.996065619079)*x_ref[1]**o+(-0.284498300043)*x_ref[1]+(-0.652795706803)*x_ref[2]**o+(-0.254746840396)*x_ref[2]
        ref[3,4,1,1]=(-0.0083632822507)*x_ref[0]**o+(0.200923618933)*x_ref[0]+(0.353883930064)*x_ref[1]**o+(0.153257380179)*x_ref[1]+(0.843965464205)*x_ref[2]**o+(0.0418501400703)*x_ref[2]
        ref[3,4,2,0]=(-0.226055622191)*x_ref[0]**o+(-0.701789166696)*x_ref[0]+(0.274080957143)*x_ref[1]**o+(0.0596744660556)*x_ref[1]+(-0.342083152987)*x_ref[2]**o+(0.593518464697)*x_ref[2]
        ref[3,4,2,1]=(-0.288168945023)*x_ref[0]**o+(-0.742841678774)*x_ref[0]+(0.972629532424)*x_ref[1]**o+(0.892100560733)*x_ref[1]+(0.803880474721)*x_ref[2]**o+(-0.0638336211224)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactZero_fromData_rank0(self):
      """
      tests L2-norm of Data on the FunctionOnContactZero

      assumptions: self.domain supports integration on FunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.908822474879)*x[0]
        ref=sqrt((0.0)+(0.206489572711))

      else:
        arg=Data(0,(),w)
        arg=(-0.893168521371)*x[0]
        ref=sqrt((0.0)+(0.199437501892))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactZero_fromData_rank1(self):
      """
      tests L2-norm of Data on the FunctionOnContactZero

      assumptions: self.domain supports integration on FunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(-0.30839170903)*x[0]
        arg[1]=(0.108352504587)*x[1]
        ref=sqrt((0.00391342175007)+(0.0237763615497))

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.260572084359)*x[0]
        arg[1]=(0.250696802346)*x[1]
        arg[2]=(-0.498857619202)*x[2]
        ref=sqrt((0.103902603647)+(0.0169744527868))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactZero_fromData_rank2(self):
      """
      tests L2-norm of Data on the FunctionOnContactZero

      assumptions: self.domain supports integration on FunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.0107777943992)*x[0]
        arg[0,1]=(0.0188929341636)*x[1]
        arg[1,0]=(-0.00268267834861)*x[0]
        arg[1,1]=(-0.656039971477)*x[1]
        arg[2,0]=(-0.997645982047)*x[0]
        arg[2,1]=(0.993049748017)*x[1]
        arg[3,0]=(0.938636169945)*x[0]
        arg[3,1]=(-0.609892243528)*x[1]
        ref=sqrt((0.596287245963)+(0.46911468066))

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(0.128987155564)*x[0]
        arg[0,1]=(-0.582660210551)*x[1]
        arg[0,2]=(-0.605040149822)*x[2]
        arg[1,0]=(0.494540007291)*x[0]
        arg[1,1]=(0.563900397935)*x[1]
        arg[1,2]=(-0.806214220737)*x[2]
        arg[2,0]=(0.878027993704)*x[0]
        arg[2,1]=(0.597165310427)*x[1]
        arg[2,2]=(-0.444798130742)*x[2]
        arg[3,0]=(0.982067213342)*x[0]
        arg[3,1]=(-0.160013302763)*x[1]
        arg[3,2]=(-0.490830124654)*x[2]
        ref=sqrt((0.831500595262)+(0.49914916859))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactZero_fromData_rank3(self):
      """
      tests L2-norm of Data on the FunctionOnContactZero

      assumptions: self.domain supports integration on FunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.83264345248)*x[0]
        arg[0,0,1]=(-0.29697614845)*x[1]
        arg[0,1,0]=(-0.903150797089)*x[0]
        arg[0,1,1]=(-0.569048463149)*x[1]
        arg[1,0,0]=(0.317087339385)*x[0]
        arg[1,0,1]=(0.0120740865344)*x[1]
        arg[1,1,0]=(-0.221145803457)*x[0]
        arg[1,1,1]=(-0.15362409313)*x[1]
        arg[2,0,0]=(0.459428065932)*x[0]
        arg[2,0,1]=(-0.495488213623)*x[1]
        arg[2,1,0]=(0.919114285084)*x[0]
        arg[2,1,1]=(0.54511942052)*x[1]
        arg[3,0,0]=(0.442207797529)*x[0]
        arg[3,0,1]=(0.958341188449)*x[1]
        arg[3,1,0]=(-0.0915205976682)*x[0]
        arg[3,1,1]=(-0.410964460336)*x[1]
        arg[4,0,0]=(-0.243030095673)*x[0]
        arg[4,0,1]=(-0.413661756935)*x[1]
        arg[4,1,0]=(-0.0973014277464)*x[0]
        arg[4,1,1]=(0.204130676874)*x[1]
        arg[5,0,0]=(0.38083750131)*x[0]
        arg[5,0,1]=(0.286331574579)*x[1]
        arg[5,1,0]=(0.583320730408)*x[0]
        arg[5,1,1]=(-0.0899419806181)*x[1]
        ref=sqrt((0.789530406063)+(0.868006693351))

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(-0.98411764268)*x[0]
        arg[0,0,1]=(-0.180549567137)*x[1]
        arg[0,0,2]=(0.357923974709)*x[2]
        arg[0,1,0]=(0.67691778218)*x[0]
        arg[0,1,1]=(-0.643267669153)*x[1]
        arg[0,1,2]=(0.452374328301)*x[2]
        arg[1,0,0]=(0.494303694439)*x[0]
        arg[1,0,1]=(0.00767629822214)*x[1]
        arg[1,0,2]=(0.232436713422)*x[2]
        arg[1,1,0]=(-0.698568975983)*x[0]
        arg[1,1,1]=(-0.08483720353)*x[1]
        arg[1,1,2]=(-0.723935799719)*x[2]
        arg[2,0,0]=(0.931313239646)*x[0]
        arg[2,0,1]=(-0.0988070442341)*x[1]
        arg[2,0,2]=(-0.166201798278)*x[2]
        arg[2,1,0]=(-0.733172171406)*x[0]
        arg[2,1,1]=(0.24542706584)*x[1]
        arg[2,1,2]=(0.49170717592)*x[2]
        arg[3,0,0]=(-0.718346096869)*x[0]
        arg[3,0,1]=(0.297442410539)*x[1]
        arg[3,0,2]=(0.0277708718174)*x[2]
        arg[3,1,0]=(0.735100702194)*x[0]
        arg[3,1,1]=(-0.808007526296)*x[1]
        arg[3,1,2]=(-0.423023238313)*x[2]
        arg[4,0,0]=(0.383188529876)*x[0]
        arg[4,0,1]=(-0.190101601814)*x[1]
        arg[4,0,2]=(-0.404380990215)*x[2]
        arg[4,1,0]=(-0.31098204217)*x[0]
        arg[4,1,1]=(-0.715247945432)*x[1]
        arg[4,1,2]=(0.435774772982)*x[2]
        arg[5,0,0]=(-0.0662056096206)*x[0]
        arg[5,0,1]=(0.746710247367)*x[1]
        arg[5,0,2]=(0.753267249301)*x[2]
        arg[5,1,0]=(0.258102962518)*x[0]
        arg[5,1,1]=(-0.0270028483187)*x[1]
        arg[5,1,2]=(-0.971060107954)*x[2]
        ref=sqrt((1.86493016409)+(1.23371587938))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactZero_fromData_rank4(self):
      """
      tests L2-norm of Data on the FunctionOnContactZero

      assumptions: self.domain supports integration on FunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=FunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.144741217139)*x[0]
        arg[0,0,0,1]=(-0.188187241012)*x[1]
        arg[0,0,1,0]=(0.056339373733)*x[0]
        arg[0,0,1,1]=(0.222177130759)*x[1]
        arg[0,0,2,0]=(-0.437240978497)*x[0]
        arg[0,0,2,1]=(-0.479213082014)*x[1]
        arg[0,1,0,0]=(0.343204141181)*x[0]
        arg[0,1,0,1]=(0.099683148208)*x[1]
        arg[0,1,1,0]=(0.167657906331)*x[0]
        arg[0,1,1,1]=(0.371064917656)*x[1]
        arg[0,1,2,0]=(-0.175796118619)*x[0]
        arg[0,1,2,1]=(-0.523477567832)*x[1]
        arg[0,2,0,0]=(0.188832282277)*x[0]
        arg[0,2,0,1]=(0.948618356423)*x[1]
        arg[0,2,1,0]=(-0.3835066594)*x[0]
        arg[0,2,1,1]=(0.872875257641)*x[1]
        arg[0,2,2,0]=(-0.98414676194)*x[0]
        arg[0,2,2,1]=(0.59170993756)*x[1]
        arg[0,3,0,0]=(0.875820385471)*x[0]
        arg[0,3,0,1]=(-0.209139794169)*x[1]
        arg[0,3,1,0]=(0.783312979835)*x[0]
        arg[0,3,1,1]=(-0.440048708153)*x[1]
        arg[0,3,2,0]=(-0.906741480158)*x[0]
        arg[0,3,2,1]=(0.63834196559)*x[1]
        arg[0,4,0,0]=(0.0142616166715)*x[0]
        arg[0,4,0,1]=(-0.0745581409622)*x[1]
        arg[0,4,1,0]=(-0.779966728505)*x[0]
        arg[0,4,1,1]=(0.734744791584)*x[1]
        arg[0,4,2,0]=(0.0407627699446)*x[0]
        arg[0,4,2,1]=(-0.629666423098)*x[1]
        arg[1,0,0,0]=(-0.782291768803)*x[0]
        arg[1,0,0,1]=(-0.197799867335)*x[1]
        arg[1,0,1,0]=(-0.487323212409)*x[0]
        arg[1,0,1,1]=(-0.789940409633)*x[1]
        arg[1,0,2,0]=(-0.523871448728)*x[0]
        arg[1,0,2,1]=(-0.622272488658)*x[1]
        arg[1,1,0,0]=(0.370095832941)*x[0]
        arg[1,1,0,1]=(-0.0683624906905)*x[1]
        arg[1,1,1,0]=(-0.180583695959)*x[0]
        arg[1,1,1,1]=(-0.0529097308814)*x[1]
        arg[1,1,2,0]=(0.440088045764)*x[0]
        arg[1,1,2,1]=(-0.207665675854)*x[1]
        arg[1,2,0,0]=(-0.360120735448)*x[0]
        arg[1,2,0,1]=(0.877797020694)*x[1]
        arg[1,2,1,0]=(0.655192721419)*x[0]
        arg[1,2,1,1]=(0.978781227463)*x[1]
        arg[1,2,2,0]=(-0.205886258554)*x[0]
        arg[1,2,2,1]=(-0.0050396100363)*x[1]
        arg[1,3,0,0]=(-0.484332943177)*x[0]
        arg[1,3,0,1]=(0.964697361018)*x[1]
        arg[1,3,1,0]=(0.0815652906824)*x[0]
        arg[1,3,1,1]=(-0.531163025802)*x[1]
        arg[1,3,2,0]=(-0.229615479833)*x[0]
        arg[1,3,2,1]=(0.139971291609)*x[1]
        arg[1,4,0,0]=(0.960839193726)*x[0]
        arg[1,4,0,1]=(-0.9744930023)*x[1]
        arg[1,4,1,0]=(-0.0919000363634)*x[0]
        arg[1,4,1,1]=(0.483909186003)*x[1]
        arg[1,4,2,0]=(0.248597552464)*x[0]
        arg[1,4,2,1]=(0.965872402486)*x[1]
        arg[2,0,0,0]=(-0.935049188033)*x[0]
        arg[2,0,0,1]=(-0.823239964421)*x[1]
        arg[2,0,1,0]=(0.30901193713)*x[0]
        arg[2,0,1,1]=(-0.0538396976182)*x[1]
        arg[2,0,2,0]=(-0.449416473099)*x[0]
        arg[2,0,2,1]=(-0.857785267819)*x[1]
        arg[2,1,0,0]=(-0.167802578334)*x[0]
        arg[2,1,0,1]=(-0.7016101186)*x[1]
        arg[2,1,1,0]=(0.995848317929)*x[0]
        arg[2,1,1,1]=(-0.824595018011)*x[1]
        arg[2,1,2,0]=(-0.826030527686)*x[0]
        arg[2,1,2,1]=(-0.856151725651)*x[1]
        arg[2,2,0,0]=(0.342047605028)*x[0]
        arg[2,2,0,1]=(-0.342370307931)*x[1]
        arg[2,2,1,0]=(-0.540693890831)*x[0]
        arg[2,2,1,1]=(-0.720754897785)*x[1]
        arg[2,2,2,0]=(-0.0941215154824)*x[0]
        arg[2,2,2,1]=(0.115887914141)*x[1]
        arg[2,3,0,0]=(-0.0920059639279)*x[0]
        arg[2,3,0,1]=(0.910948619784)*x[1]
        arg[2,3,1,0]=(-0.40846609126)*x[0]
        arg[2,3,1,1]=(0.542728733954)*x[1]
        arg[2,3,2,0]=(0.451361830368)*x[0]
        arg[2,3,2,1]=(0.0767538617936)*x[1]
        arg[2,4,0,0]=(-0.431757008766)*x[0]
        arg[2,4,0,1]=(0.258988103513)*x[1]
        arg[2,4,1,0]=(-0.584559859464)*x[0]
        arg[2,4,1,1]=(-0.776827606804)*x[1]
        arg[2,4,2,0]=(-0.867078960637)*x[0]
        arg[2,4,2,1]=(-0.0972960135129)*x[1]
        arg[3,0,0,0]=(0.999511727022)*x[0]
        arg[3,0,0,1]=(-0.753218704622)*x[1]
        arg[3,0,1,0]=(0.506439354885)*x[0]
        arg[3,0,1,1]=(-0.0845853282297)*x[1]
        arg[3,0,2,0]=(-0.871475028801)*x[0]
        arg[3,0,2,1]=(-0.69261642657)*x[1]
        arg[3,1,0,0]=(0.549778399933)*x[0]
        arg[3,1,0,1]=(0.246187536149)*x[1]
        arg[3,1,1,0]=(-0.620454200676)*x[0]
        arg[3,1,1,1]=(0.372738564315)*x[1]
        arg[3,1,2,0]=(0.0544097075138)*x[0]
        arg[3,1,2,1]=(-0.0883198944676)*x[1]
        arg[3,2,0,0]=(-0.671057180545)*x[0]
        arg[3,2,0,1]=(-0.118991797907)*x[1]
        arg[3,2,1,0]=(-0.27196730081)*x[0]
        arg[3,2,1,1]=(-0.458807068968)*x[1]
        arg[3,2,2,0]=(-0.31485399083)*x[0]
        arg[3,2,2,1]=(0.487291960328)*x[1]
        arg[3,3,0,0]=(-0.750302723531)*x[0]
        arg[3,3,0,1]=(-0.486428789771)*x[1]
        arg[3,3,1,0]=(-0.295909324594)*x[0]
        arg[3,3,1,1]=(0.325702372596)*x[1]
        arg[3,3,2,0]=(-0.512109540631)*x[0]
        arg[3,3,2,1]=(0.563284721908)*x[1]
        arg[3,4,0,0]=(0.53579406798)*x[0]
        arg[3,4,0,1]=(-0.468431927975)*x[1]
        arg[3,4,1,0]=(0.570851401329)*x[0]
        arg[3,4,1,1]=(0.107791149865)*x[1]
        arg[3,4,2,0]=(-0.211543670157)*x[0]
        arg[3,4,2,1]=(0.848189550468)*x[1]
        ref=sqrt((6.54316712884)+(4.39564064326))

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(-0.392256494872)*x[0]
        arg[0,0,0,1]=(0.271488478841)*x[1]
        arg[0,0,0,2]=(-0.878268531987)*x[2]
        arg[0,0,1,0]=(0.106903787643)*x[0]
        arg[0,0,1,1]=(-0.117986434516)*x[1]
        arg[0,0,1,2]=(0.912673598522)*x[2]
        arg[0,0,2,0]=(-0.652598945824)*x[0]
        arg[0,0,2,1]=(-0.984195895647)*x[1]
        arg[0,0,2,2]=(-0.137845459589)*x[2]
        arg[0,1,0,0]=(0.696713200721)*x[0]
        arg[0,1,0,1]=(0.480337929223)*x[1]
        arg[0,1,0,2]=(-0.628726114325)*x[2]
        arg[0,1,1,0]=(-0.356503461347)*x[0]
        arg[0,1,1,1]=(-0.64670584045)*x[1]
        arg[0,1,1,2]=(-0.737787618908)*x[2]
        arg[0,1,2,0]=(-0.14207595302)*x[0]
        arg[0,1,2,1]=(-0.573485525334)*x[1]
        arg[0,1,2,2]=(-0.955542178481)*x[2]
        arg[0,2,0,0]=(0.371952883975)*x[0]
        arg[0,2,0,1]=(0.114016178762)*x[1]
        arg[0,2,0,2]=(0.552721218169)*x[2]
        arg[0,2,1,0]=(0.318426742113)*x[0]
        arg[0,2,1,1]=(0.862220239384)*x[1]
        arg[0,2,1,2]=(0.887376889978)*x[2]
        arg[0,2,2,0]=(0.203656742981)*x[0]
        arg[0,2,2,1]=(0.350541335144)*x[1]
        arg[0,2,2,2]=(-0.448401957447)*x[2]
        arg[0,3,0,0]=(-0.349200084042)*x[0]
        arg[0,3,0,1]=(-0.546012602125)*x[1]
        arg[0,3,0,2]=(-0.931488270979)*x[2]
        arg[0,3,1,0]=(0.207457831058)*x[0]
        arg[0,3,1,1]=(0.557715840549)*x[1]
        arg[0,3,1,2]=(-0.978130146744)*x[2]
        arg[0,3,2,0]=(-0.55088967957)*x[0]
        arg[0,3,2,1]=(0.0490328838057)*x[1]
        arg[0,3,2,2]=(0.150209929122)*x[2]
        arg[0,4,0,0]=(-0.484145614698)*x[0]
        arg[0,4,0,1]=(0.393078411279)*x[1]
        arg[0,4,0,2]=(0.0678631863917)*x[2]
        arg[0,4,1,0]=(-0.350980464628)*x[0]
        arg[0,4,1,1]=(-0.784203839564)*x[1]
        arg[0,4,1,2]=(0.636960296147)*x[2]
        arg[0,4,2,0]=(0.592799503581)*x[0]
        arg[0,4,2,1]=(-0.672104833683)*x[1]
        arg[0,4,2,2]=(0.0366914082467)*x[2]
        arg[1,0,0,0]=(-0.147168019774)*x[0]
        arg[1,0,0,1]=(-0.0823637938956)*x[1]
        arg[1,0,0,2]=(-0.852729690176)*x[2]
        arg[1,0,1,0]=(-0.612338430408)*x[0]
        arg[1,0,1,1]=(-0.85820035747)*x[1]
        arg[1,0,1,2]=(-0.463664966162)*x[2]
        arg[1,0,2,0]=(0.274600720491)*x[0]
        arg[1,0,2,1]=(-0.488508234093)*x[1]
        arg[1,0,2,2]=(-0.28251466519)*x[2]
        arg[1,1,0,0]=(-0.0196532462794)*x[0]
        arg[1,1,0,1]=(0.239352528871)*x[1]
        arg[1,1,0,2]=(-0.17054773873)*x[2]
        arg[1,1,1,0]=(-0.2396627789)*x[0]
        arg[1,1,1,1]=(0.868970323003)*x[1]
        arg[1,1,1,2]=(0.401487430312)*x[2]
        arg[1,1,2,0]=(-0.624411449783)*x[0]
        arg[1,1,2,1]=(0.0036634266684)*x[1]
        arg[1,1,2,2]=(0.736129120967)*x[2]
        arg[1,2,0,0]=(0.183756511707)*x[0]
        arg[1,2,0,1]=(-0.288651848639)*x[1]
        arg[1,2,0,2]=(-0.0672121537447)*x[2]
        arg[1,2,1,0]=(-0.323274725936)*x[0]
        arg[1,2,1,1]=(0.298001016025)*x[1]
        arg[1,2,1,2]=(-0.976052460675)*x[2]
        arg[1,2,2,0]=(0.596504441096)*x[0]
        arg[1,2,2,1]=(0.873776068983)*x[1]
        arg[1,2,2,2]=(-0.994068273196)*x[2]
        arg[1,3,0,0]=(-0.495387299681)*x[0]
        arg[1,3,0,1]=(-0.123674756551)*x[1]
        arg[1,3,0,2]=(0.581213818577)*x[2]
        arg[1,3,1,0]=(0.146405749701)*x[0]
        arg[1,3,1,1]=(-0.594994686675)*x[1]
        arg[1,3,1,2]=(-0.059093568436)*x[2]
        arg[1,3,2,0]=(0.651004255104)*x[0]
        arg[1,3,2,1]=(-0.977880706193)*x[1]
        arg[1,3,2,2]=(0.370344651319)*x[2]
        arg[1,4,0,0]=(-0.503657215247)*x[0]
        arg[1,4,0,1]=(-0.170885297253)*x[1]
        arg[1,4,0,2]=(0.533424480956)*x[2]
        arg[1,4,1,0]=(-0.220533193308)*x[0]
        arg[1,4,1,1]=(0.344537611882)*x[1]
        arg[1,4,1,2]=(0.861473877282)*x[2]
        arg[1,4,2,0]=(-0.0923010438884)*x[0]
        arg[1,4,2,1]=(-0.338256780498)*x[1]
        arg[1,4,2,2]=(0.528567959345)*x[2]
        arg[2,0,0,0]=(-0.0423053381485)*x[0]
        arg[2,0,0,1]=(0.856798579151)*x[1]
        arg[2,0,0,2]=(0.383258153853)*x[2]
        arg[2,0,1,0]=(0.350994872736)*x[0]
        arg[2,0,1,1]=(-0.78055158106)*x[1]
        arg[2,0,1,2]=(-0.770876699915)*x[2]
        arg[2,0,2,0]=(-0.935133287106)*x[0]
        arg[2,0,2,1]=(0.618238076989)*x[1]
        arg[2,0,2,2]=(-0.846783087949)*x[2]
        arg[2,1,0,0]=(-0.657394511405)*x[0]
        arg[2,1,0,1]=(0.576218821654)*x[1]
        arg[2,1,0,2]=(0.0269446356493)*x[2]
        arg[2,1,1,0]=(-0.310710230949)*x[0]
        arg[2,1,1,1]=(0.425412515598)*x[1]
        arg[2,1,1,2]=(-0.225225290862)*x[2]
        arg[2,1,2,0]=(-0.539928589495)*x[0]
        arg[2,1,2,1]=(-0.348082121765)*x[1]
        arg[2,1,2,2]=(-0.0287274646233)*x[2]
        arg[2,2,0,0]=(0.746132865923)*x[0]
        arg[2,2,0,1]=(-0.0234203693548)*x[1]
        arg[2,2,0,2]=(0.517411821941)*x[2]
        arg[2,2,1,0]=(-0.183204217349)*x[0]
        arg[2,2,1,1]=(0.714988861836)*x[1]
        arg[2,2,1,2]=(0.829083318937)*x[2]
        arg[2,2,2,0]=(0.458067555841)*x[0]
        arg[2,2,2,1]=(0.639317125869)*x[1]
        arg[2,2,2,2]=(0.104611520408)*x[2]
        arg[2,3,0,0]=(-0.420513461135)*x[0]
        arg[2,3,0,1]=(0.888686162754)*x[1]
        arg[2,3,0,2]=(0.939305777879)*x[2]
        arg[2,3,1,0]=(0.856795132015)*x[0]
        arg[2,3,1,1]=(0.817593141895)*x[1]
        arg[2,3,1,2]=(0.962503342535)*x[2]
        arg[2,3,2,0]=(-0.334586700245)*x[0]
        arg[2,3,2,1]=(0.182696129528)*x[1]
        arg[2,3,2,2]=(-0.707271571206)*x[2]
        arg[2,4,0,0]=(-0.594795981069)*x[0]
        arg[2,4,0,1]=(0.451239168073)*x[1]
        arg[2,4,0,2]=(0.191212211556)*x[2]
        arg[2,4,1,0]=(-0.503465984944)*x[0]
        arg[2,4,1,1]=(0.725377884208)*x[1]
        arg[2,4,1,2]=(-0.40719255752)*x[2]
        arg[2,4,2,0]=(0.268071476451)*x[0]
        arg[2,4,2,1]=(0.85066639942)*x[1]
        arg[2,4,2,2]=(-0.906021406945)*x[2]
        arg[3,0,0,0]=(-0.922179152122)*x[0]
        arg[3,0,0,1]=(-0.0903841240007)*x[1]
        arg[3,0,0,2]=(-0.751482803516)*x[2]
        arg[3,0,1,0]=(0.960697809119)*x[0]
        arg[3,0,1,1]=(0.638878873158)*x[1]
        arg[3,0,1,2]=(0.390932234724)*x[2]
        arg[3,0,2,0]=(-0.925078301694)*x[0]
        arg[3,0,2,1]=(0.793590580665)*x[1]
        arg[3,0,2,2]=(0.535478366911)*x[2]
        arg[3,1,0,0]=(-0.431951993217)*x[0]
        arg[3,1,0,1]=(0.211750261417)*x[1]
        arg[3,1,0,2]=(-0.930706580442)*x[2]
        arg[3,1,1,0]=(0.330979313323)*x[0]
        arg[3,1,1,1]=(-0.838919076081)*x[1]
        arg[3,1,1,2]=(0.134250050168)*x[2]
        arg[3,1,2,0]=(0.414922811301)*x[0]
        arg[3,1,2,1]=(-0.663692878121)*x[1]
        arg[3,1,2,2]=(0.88499278543)*x[2]
        arg[3,2,0,0]=(0.0884662742233)*x[0]
        arg[3,2,0,1]=(-0.412630722821)*x[1]
        arg[3,2,0,2]=(-0.730850884928)*x[2]
        arg[3,2,1,0]=(0.722207547366)*x[0]
        arg[3,2,1,1]=(-0.260067950749)*x[1]
        arg[3,2,1,2]=(0.426259201494)*x[2]
        arg[3,2,2,0]=(0.0516111795322)*x[0]
        arg[3,2,2,1]=(0.922853710048)*x[1]
        arg[3,2,2,2]=(-0.991912758116)*x[2]
        arg[3,3,0,0]=(0.263933965905)*x[0]
        arg[3,3,0,1]=(0.840541758799)*x[1]
        arg[3,3,0,2]=(0.417658511125)*x[2]
        arg[3,3,1,0]=(-0.901745614723)*x[0]
        arg[3,3,1,1]=(-0.623608908699)*x[1]
        arg[3,3,1,2]=(0.0522167208784)*x[2]
        arg[3,3,2,0]=(-0.549431264931)*x[0]
        arg[3,3,2,1]=(0.919971855457)*x[1]
        arg[3,3,2,2]=(0.142757773397)*x[2]
        arg[3,4,0,0]=(-0.258875259824)*x[0]
        arg[3,4,0,1]=(-0.0373872187041)*x[1]
        arg[3,4,0,2]=(0.445989164864)*x[2]
        arg[3,4,1,0]=(0.0273711397038)*x[0]
        arg[3,4,1,1]=(-0.85522629)*x[1]
        arg[3,4,1,2]=(-0.835392581226)*x[2]
        arg[3,4,2,0]=(0.402929450189)*x[0]
        arg[3,4,2,1]=(0.549623033221)*x[1]
        arg[3,4,2,2]=(0.237803212109)*x[2]
        ref=sqrt((15.1448100099)+(3.61598249819))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactZero_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactZero

      assumptions: self.domain supports integration on ReducedFunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.908822474879)*sqrt(x[0])
        ref=sqrt((0.0)*3./2.+(0.206489572711)/0.5)

      else:
        arg=Data(0,(),w)
        arg=(-0.893168521371)*sqrt(x[0])
        ref=sqrt((0.0)*3./2.+(0.199437501892)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactZero_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactZero

      assumptions: self.domain supports integration on ReducedFunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(-0.30839170903)*sqrt(x[0])
        arg[1]=(0.108352504587)*sqrt(x[1])
        ref=sqrt((0.00391342175007)*3./2.+(0.0237763615497)/0.5)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.260572084359)*sqrt(x[0])
        arg[1]=(0.250696802346)*sqrt(x[1])
        arg[2]=(-0.498857619202)*sqrt(x[2])
        ref=sqrt((0.103902603647)*3./2.+(0.0169744527868)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactZero_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactZero

      assumptions: self.domain supports integration on ReducedFunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.0107777943992)*sqrt(x[0])
        arg[0,1]=(0.0188929341636)*sqrt(x[1])
        arg[1,0]=(-0.00268267834861)*sqrt(x[0])
        arg[1,1]=(-0.656039971477)*sqrt(x[1])
        arg[2,0]=(-0.997645982047)*sqrt(x[0])
        arg[2,1]=(0.993049748017)*sqrt(x[1])
        arg[3,0]=(0.938636169945)*sqrt(x[0])
        arg[3,1]=(-0.609892243528)*sqrt(x[1])
        ref=sqrt((0.596287245963)*3./2.+(0.46911468066)/0.5)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(0.128987155564)*sqrt(x[0])
        arg[0,1]=(-0.582660210551)*sqrt(x[1])
        arg[0,2]=(-0.605040149822)*sqrt(x[2])
        arg[1,0]=(0.494540007291)*sqrt(x[0])
        arg[1,1]=(0.563900397935)*sqrt(x[1])
        arg[1,2]=(-0.806214220737)*sqrt(x[2])
        arg[2,0]=(0.878027993704)*sqrt(x[0])
        arg[2,1]=(0.597165310427)*sqrt(x[1])
        arg[2,2]=(-0.444798130742)*sqrt(x[2])
        arg[3,0]=(0.982067213342)*sqrt(x[0])
        arg[3,1]=(-0.160013302763)*sqrt(x[1])
        arg[3,2]=(-0.490830124654)*sqrt(x[2])
        ref=sqrt((0.831500595262)*3./2.+(0.49914916859)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactZero_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactZero

      assumptions: self.domain supports integration on ReducedFunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.83264345248)*sqrt(x[0])
        arg[0,0,1]=(-0.29697614845)*sqrt(x[1])
        arg[0,1,0]=(-0.903150797089)*sqrt(x[0])
        arg[0,1,1]=(-0.569048463149)*sqrt(x[1])
        arg[1,0,0]=(0.317087339385)*sqrt(x[0])
        arg[1,0,1]=(0.0120740865344)*sqrt(x[1])
        arg[1,1,0]=(-0.221145803457)*sqrt(x[0])
        arg[1,1,1]=(-0.15362409313)*sqrt(x[1])
        arg[2,0,0]=(0.459428065932)*sqrt(x[0])
        arg[2,0,1]=(-0.495488213623)*sqrt(x[1])
        arg[2,1,0]=(0.919114285084)*sqrt(x[0])
        arg[2,1,1]=(0.54511942052)*sqrt(x[1])
        arg[3,0,0]=(0.442207797529)*sqrt(x[0])
        arg[3,0,1]=(0.958341188449)*sqrt(x[1])
        arg[3,1,0]=(-0.0915205976682)*sqrt(x[0])
        arg[3,1,1]=(-0.410964460336)*sqrt(x[1])
        arg[4,0,0]=(-0.243030095673)*sqrt(x[0])
        arg[4,0,1]=(-0.413661756935)*sqrt(x[1])
        arg[4,1,0]=(-0.0973014277464)*sqrt(x[0])
        arg[4,1,1]=(0.204130676874)*sqrt(x[1])
        arg[5,0,0]=(0.38083750131)*sqrt(x[0])
        arg[5,0,1]=(0.286331574579)*sqrt(x[1])
        arg[5,1,0]=(0.583320730408)*sqrt(x[0])
        arg[5,1,1]=(-0.0899419806181)*sqrt(x[1])
        ref=sqrt((0.789530406063)*3./2.+(0.868006693351)/0.5)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(-0.98411764268)*sqrt(x[0])
        arg[0,0,1]=(-0.180549567137)*sqrt(x[1])
        arg[0,0,2]=(0.357923974709)*sqrt(x[2])
        arg[0,1,0]=(0.67691778218)*sqrt(x[0])
        arg[0,1,1]=(-0.643267669153)*sqrt(x[1])
        arg[0,1,2]=(0.452374328301)*sqrt(x[2])
        arg[1,0,0]=(0.494303694439)*sqrt(x[0])
        arg[1,0,1]=(0.00767629822214)*sqrt(x[1])
        arg[1,0,2]=(0.232436713422)*sqrt(x[2])
        arg[1,1,0]=(-0.698568975983)*sqrt(x[0])
        arg[1,1,1]=(-0.08483720353)*sqrt(x[1])
        arg[1,1,2]=(-0.723935799719)*sqrt(x[2])
        arg[2,0,0]=(0.931313239646)*sqrt(x[0])
        arg[2,0,1]=(-0.0988070442341)*sqrt(x[1])
        arg[2,0,2]=(-0.166201798278)*sqrt(x[2])
        arg[2,1,0]=(-0.733172171406)*sqrt(x[0])
        arg[2,1,1]=(0.24542706584)*sqrt(x[1])
        arg[2,1,2]=(0.49170717592)*sqrt(x[2])
        arg[3,0,0]=(-0.718346096869)*sqrt(x[0])
        arg[3,0,1]=(0.297442410539)*sqrt(x[1])
        arg[3,0,2]=(0.0277708718174)*sqrt(x[2])
        arg[3,1,0]=(0.735100702194)*sqrt(x[0])
        arg[3,1,1]=(-0.808007526296)*sqrt(x[1])
        arg[3,1,2]=(-0.423023238313)*sqrt(x[2])
        arg[4,0,0]=(0.383188529876)*sqrt(x[0])
        arg[4,0,1]=(-0.190101601814)*sqrt(x[1])
        arg[4,0,2]=(-0.404380990215)*sqrt(x[2])
        arg[4,1,0]=(-0.31098204217)*sqrt(x[0])
        arg[4,1,1]=(-0.715247945432)*sqrt(x[1])
        arg[4,1,2]=(0.435774772982)*sqrt(x[2])
        arg[5,0,0]=(-0.0662056096206)*sqrt(x[0])
        arg[5,0,1]=(0.746710247367)*sqrt(x[1])
        arg[5,0,2]=(0.753267249301)*sqrt(x[2])
        arg[5,1,0]=(0.258102962518)*sqrt(x[0])
        arg[5,1,1]=(-0.0270028483187)*sqrt(x[1])
        arg[5,1,2]=(-0.971060107954)*sqrt(x[2])
        ref=sqrt((1.86493016409)*3./2.+(1.23371587938)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactZero_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactZero

      assumptions: self.domain supports integration on ReducedFunctionOnContactZero
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactZero(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.144741217139)*sqrt(x[0])
        arg[0,0,0,1]=(-0.188187241012)*sqrt(x[1])
        arg[0,0,1,0]=(0.056339373733)*sqrt(x[0])
        arg[0,0,1,1]=(0.222177130759)*sqrt(x[1])
        arg[0,0,2,0]=(-0.437240978497)*sqrt(x[0])
        arg[0,0,2,1]=(-0.479213082014)*sqrt(x[1])
        arg[0,1,0,0]=(0.343204141181)*sqrt(x[0])
        arg[0,1,0,1]=(0.099683148208)*sqrt(x[1])
        arg[0,1,1,0]=(0.167657906331)*sqrt(x[0])
        arg[0,1,1,1]=(0.371064917656)*sqrt(x[1])
        arg[0,1,2,0]=(-0.175796118619)*sqrt(x[0])
        arg[0,1,2,1]=(-0.523477567832)*sqrt(x[1])
        arg[0,2,0,0]=(0.188832282277)*sqrt(x[0])
        arg[0,2,0,1]=(0.948618356423)*sqrt(x[1])
        arg[0,2,1,0]=(-0.3835066594)*sqrt(x[0])
        arg[0,2,1,1]=(0.872875257641)*sqrt(x[1])
        arg[0,2,2,0]=(-0.98414676194)*sqrt(x[0])
        arg[0,2,2,1]=(0.59170993756)*sqrt(x[1])
        arg[0,3,0,0]=(0.875820385471)*sqrt(x[0])
        arg[0,3,0,1]=(-0.209139794169)*sqrt(x[1])
        arg[0,3,1,0]=(0.783312979835)*sqrt(x[0])
        arg[0,3,1,1]=(-0.440048708153)*sqrt(x[1])
        arg[0,3,2,0]=(-0.906741480158)*sqrt(x[0])
        arg[0,3,2,1]=(0.63834196559)*sqrt(x[1])
        arg[0,4,0,0]=(0.0142616166715)*sqrt(x[0])
        arg[0,4,0,1]=(-0.0745581409622)*sqrt(x[1])
        arg[0,4,1,0]=(-0.779966728505)*sqrt(x[0])
        arg[0,4,1,1]=(0.734744791584)*sqrt(x[1])
        arg[0,4,2,0]=(0.0407627699446)*sqrt(x[0])
        arg[0,4,2,1]=(-0.629666423098)*sqrt(x[1])
        arg[1,0,0,0]=(-0.782291768803)*sqrt(x[0])
        arg[1,0,0,1]=(-0.197799867335)*sqrt(x[1])
        arg[1,0,1,0]=(-0.487323212409)*sqrt(x[0])
        arg[1,0,1,1]=(-0.789940409633)*sqrt(x[1])
        arg[1,0,2,0]=(-0.523871448728)*sqrt(x[0])
        arg[1,0,2,1]=(-0.622272488658)*sqrt(x[1])
        arg[1,1,0,0]=(0.370095832941)*sqrt(x[0])
        arg[1,1,0,1]=(-0.0683624906905)*sqrt(x[1])
        arg[1,1,1,0]=(-0.180583695959)*sqrt(x[0])
        arg[1,1,1,1]=(-0.0529097308814)*sqrt(x[1])
        arg[1,1,2,0]=(0.440088045764)*sqrt(x[0])
        arg[1,1,2,1]=(-0.207665675854)*sqrt(x[1])
        arg[1,2,0,0]=(-0.360120735448)*sqrt(x[0])
        arg[1,2,0,1]=(0.877797020694)*sqrt(x[1])
        arg[1,2,1,0]=(0.655192721419)*sqrt(x[0])
        arg[1,2,1,1]=(0.978781227463)*sqrt(x[1])
        arg[1,2,2,0]=(-0.205886258554)*sqrt(x[0])
        arg[1,2,2,1]=(-0.0050396100363)*sqrt(x[1])
        arg[1,3,0,0]=(-0.484332943177)*sqrt(x[0])
        arg[1,3,0,1]=(0.964697361018)*sqrt(x[1])
        arg[1,3,1,0]=(0.0815652906824)*sqrt(x[0])
        arg[1,3,1,1]=(-0.531163025802)*sqrt(x[1])
        arg[1,3,2,0]=(-0.229615479833)*sqrt(x[0])
        arg[1,3,2,1]=(0.139971291609)*sqrt(x[1])
        arg[1,4,0,0]=(0.960839193726)*sqrt(x[0])
        arg[1,4,0,1]=(-0.9744930023)*sqrt(x[1])
        arg[1,4,1,0]=(-0.0919000363634)*sqrt(x[0])
        arg[1,4,1,1]=(0.483909186003)*sqrt(x[1])
        arg[1,4,2,0]=(0.248597552464)*sqrt(x[0])
        arg[1,4,2,1]=(0.965872402486)*sqrt(x[1])
        arg[2,0,0,0]=(-0.935049188033)*sqrt(x[0])
        arg[2,0,0,1]=(-0.823239964421)*sqrt(x[1])
        arg[2,0,1,0]=(0.30901193713)*sqrt(x[0])
        arg[2,0,1,1]=(-0.0538396976182)*sqrt(x[1])
        arg[2,0,2,0]=(-0.449416473099)*sqrt(x[0])
        arg[2,0,2,1]=(-0.857785267819)*sqrt(x[1])
        arg[2,1,0,0]=(-0.167802578334)*sqrt(x[0])
        arg[2,1,0,1]=(-0.7016101186)*sqrt(x[1])
        arg[2,1,1,0]=(0.995848317929)*sqrt(x[0])
        arg[2,1,1,1]=(-0.824595018011)*sqrt(x[1])
        arg[2,1,2,0]=(-0.826030527686)*sqrt(x[0])
        arg[2,1,2,1]=(-0.856151725651)*sqrt(x[1])
        arg[2,2,0,0]=(0.342047605028)*sqrt(x[0])
        arg[2,2,0,1]=(-0.342370307931)*sqrt(x[1])
        arg[2,2,1,0]=(-0.540693890831)*sqrt(x[0])
        arg[2,2,1,1]=(-0.720754897785)*sqrt(x[1])
        arg[2,2,2,0]=(-0.0941215154824)*sqrt(x[0])
        arg[2,2,2,1]=(0.115887914141)*sqrt(x[1])
        arg[2,3,0,0]=(-0.0920059639279)*sqrt(x[0])
        arg[2,3,0,1]=(0.910948619784)*sqrt(x[1])
        arg[2,3,1,0]=(-0.40846609126)*sqrt(x[0])
        arg[2,3,1,1]=(0.542728733954)*sqrt(x[1])
        arg[2,3,2,0]=(0.451361830368)*sqrt(x[0])
        arg[2,3,2,1]=(0.0767538617936)*sqrt(x[1])
        arg[2,4,0,0]=(-0.431757008766)*sqrt(x[0])
        arg[2,4,0,1]=(0.258988103513)*sqrt(x[1])
        arg[2,4,1,0]=(-0.584559859464)*sqrt(x[0])
        arg[2,4,1,1]=(-0.776827606804)*sqrt(x[1])
        arg[2,4,2,0]=(-0.867078960637)*sqrt(x[0])
        arg[2,4,2,1]=(-0.0972960135129)*sqrt(x[1])
        arg[3,0,0,0]=(0.999511727022)*sqrt(x[0])
        arg[3,0,0,1]=(-0.753218704622)*sqrt(x[1])
        arg[3,0,1,0]=(0.506439354885)*sqrt(x[0])
        arg[3,0,1,1]=(-0.0845853282297)*sqrt(x[1])
        arg[3,0,2,0]=(-0.871475028801)*sqrt(x[0])
        arg[3,0,2,1]=(-0.69261642657)*sqrt(x[1])
        arg[3,1,0,0]=(0.549778399933)*sqrt(x[0])
        arg[3,1,0,1]=(0.246187536149)*sqrt(x[1])
        arg[3,1,1,0]=(-0.620454200676)*sqrt(x[0])
        arg[3,1,1,1]=(0.372738564315)*sqrt(x[1])
        arg[3,1,2,0]=(0.0544097075138)*sqrt(x[0])
        arg[3,1,2,1]=(-0.0883198944676)*sqrt(x[1])
        arg[3,2,0,0]=(-0.671057180545)*sqrt(x[0])
        arg[3,2,0,1]=(-0.118991797907)*sqrt(x[1])
        arg[3,2,1,0]=(-0.27196730081)*sqrt(x[0])
        arg[3,2,1,1]=(-0.458807068968)*sqrt(x[1])
        arg[3,2,2,0]=(-0.31485399083)*sqrt(x[0])
        arg[3,2,2,1]=(0.487291960328)*sqrt(x[1])
        arg[3,3,0,0]=(-0.750302723531)*sqrt(x[0])
        arg[3,3,0,1]=(-0.486428789771)*sqrt(x[1])
        arg[3,3,1,0]=(-0.295909324594)*sqrt(x[0])
        arg[3,3,1,1]=(0.325702372596)*sqrt(x[1])
        arg[3,3,2,0]=(-0.512109540631)*sqrt(x[0])
        arg[3,3,2,1]=(0.563284721908)*sqrt(x[1])
        arg[3,4,0,0]=(0.53579406798)*sqrt(x[0])
        arg[3,4,0,1]=(-0.468431927975)*sqrt(x[1])
        arg[3,4,1,0]=(0.570851401329)*sqrt(x[0])
        arg[3,4,1,1]=(0.107791149865)*sqrt(x[1])
        arg[3,4,2,0]=(-0.211543670157)*sqrt(x[0])
        arg[3,4,2,1]=(0.848189550468)*sqrt(x[1])
        ref=sqrt((6.54316712884)*3./2.+(4.39564064326)/0.5)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(-0.392256494872)*sqrt(x[0])
        arg[0,0,0,1]=(0.271488478841)*sqrt(x[1])
        arg[0,0,0,2]=(-0.878268531987)*sqrt(x[2])
        arg[0,0,1,0]=(0.106903787643)*sqrt(x[0])
        arg[0,0,1,1]=(-0.117986434516)*sqrt(x[1])
        arg[0,0,1,2]=(0.912673598522)*sqrt(x[2])
        arg[0,0,2,0]=(-0.652598945824)*sqrt(x[0])
        arg[0,0,2,1]=(-0.984195895647)*sqrt(x[1])
        arg[0,0,2,2]=(-0.137845459589)*sqrt(x[2])
        arg[0,1,0,0]=(0.696713200721)*sqrt(x[0])
        arg[0,1,0,1]=(0.480337929223)*sqrt(x[1])
        arg[0,1,0,2]=(-0.628726114325)*sqrt(x[2])
        arg[0,1,1,0]=(-0.356503461347)*sqrt(x[0])
        arg[0,1,1,1]=(-0.64670584045)*sqrt(x[1])
        arg[0,1,1,2]=(-0.737787618908)*sqrt(x[2])
        arg[0,1,2,0]=(-0.14207595302)*sqrt(x[0])
        arg[0,1,2,1]=(-0.573485525334)*sqrt(x[1])
        arg[0,1,2,2]=(-0.955542178481)*sqrt(x[2])
        arg[0,2,0,0]=(0.371952883975)*sqrt(x[0])
        arg[0,2,0,1]=(0.114016178762)*sqrt(x[1])
        arg[0,2,0,2]=(0.552721218169)*sqrt(x[2])
        arg[0,2,1,0]=(0.318426742113)*sqrt(x[0])
        arg[0,2,1,1]=(0.862220239384)*sqrt(x[1])
        arg[0,2,1,2]=(0.887376889978)*sqrt(x[2])
        arg[0,2,2,0]=(0.203656742981)*sqrt(x[0])
        arg[0,2,2,1]=(0.350541335144)*sqrt(x[1])
        arg[0,2,2,2]=(-0.448401957447)*sqrt(x[2])
        arg[0,3,0,0]=(-0.349200084042)*sqrt(x[0])
        arg[0,3,0,1]=(-0.546012602125)*sqrt(x[1])
        arg[0,3,0,2]=(-0.931488270979)*sqrt(x[2])
        arg[0,3,1,0]=(0.207457831058)*sqrt(x[0])
        arg[0,3,1,1]=(0.557715840549)*sqrt(x[1])
        arg[0,3,1,2]=(-0.978130146744)*sqrt(x[2])
        arg[0,3,2,0]=(-0.55088967957)*sqrt(x[0])
        arg[0,3,2,1]=(0.0490328838057)*sqrt(x[1])
        arg[0,3,2,2]=(0.150209929122)*sqrt(x[2])
        arg[0,4,0,0]=(-0.484145614698)*sqrt(x[0])
        arg[0,4,0,1]=(0.393078411279)*sqrt(x[1])
        arg[0,4,0,2]=(0.0678631863917)*sqrt(x[2])
        arg[0,4,1,0]=(-0.350980464628)*sqrt(x[0])
        arg[0,4,1,1]=(-0.784203839564)*sqrt(x[1])
        arg[0,4,1,2]=(0.636960296147)*sqrt(x[2])
        arg[0,4,2,0]=(0.592799503581)*sqrt(x[0])
        arg[0,4,2,1]=(-0.672104833683)*sqrt(x[1])
        arg[0,4,2,2]=(0.0366914082467)*sqrt(x[2])
        arg[1,0,0,0]=(-0.147168019774)*sqrt(x[0])
        arg[1,0,0,1]=(-0.0823637938956)*sqrt(x[1])
        arg[1,0,0,2]=(-0.852729690176)*sqrt(x[2])
        arg[1,0,1,0]=(-0.612338430408)*sqrt(x[0])
        arg[1,0,1,1]=(-0.85820035747)*sqrt(x[1])
        arg[1,0,1,2]=(-0.463664966162)*sqrt(x[2])
        arg[1,0,2,0]=(0.274600720491)*sqrt(x[0])
        arg[1,0,2,1]=(-0.488508234093)*sqrt(x[1])
        arg[1,0,2,2]=(-0.28251466519)*sqrt(x[2])
        arg[1,1,0,0]=(-0.0196532462794)*sqrt(x[0])
        arg[1,1,0,1]=(0.239352528871)*sqrt(x[1])
        arg[1,1,0,2]=(-0.17054773873)*sqrt(x[2])
        arg[1,1,1,0]=(-0.2396627789)*sqrt(x[0])
        arg[1,1,1,1]=(0.868970323003)*sqrt(x[1])
        arg[1,1,1,2]=(0.401487430312)*sqrt(x[2])
        arg[1,1,2,0]=(-0.624411449783)*sqrt(x[0])
        arg[1,1,2,1]=(0.0036634266684)*sqrt(x[1])
        arg[1,1,2,2]=(0.736129120967)*sqrt(x[2])
        arg[1,2,0,0]=(0.183756511707)*sqrt(x[0])
        arg[1,2,0,1]=(-0.288651848639)*sqrt(x[1])
        arg[1,2,0,2]=(-0.0672121537447)*sqrt(x[2])
        arg[1,2,1,0]=(-0.323274725936)*sqrt(x[0])
        arg[1,2,1,1]=(0.298001016025)*sqrt(x[1])
        arg[1,2,1,2]=(-0.976052460675)*sqrt(x[2])
        arg[1,2,2,0]=(0.596504441096)*sqrt(x[0])
        arg[1,2,2,1]=(0.873776068983)*sqrt(x[1])
        arg[1,2,2,2]=(-0.994068273196)*sqrt(x[2])
        arg[1,3,0,0]=(-0.495387299681)*sqrt(x[0])
        arg[1,3,0,1]=(-0.123674756551)*sqrt(x[1])
        arg[1,3,0,2]=(0.581213818577)*sqrt(x[2])
        arg[1,3,1,0]=(0.146405749701)*sqrt(x[0])
        arg[1,3,1,1]=(-0.594994686675)*sqrt(x[1])
        arg[1,3,1,2]=(-0.059093568436)*sqrt(x[2])
        arg[1,3,2,0]=(0.651004255104)*sqrt(x[0])
        arg[1,3,2,1]=(-0.977880706193)*sqrt(x[1])
        arg[1,3,2,2]=(0.370344651319)*sqrt(x[2])
        arg[1,4,0,0]=(-0.503657215247)*sqrt(x[0])
        arg[1,4,0,1]=(-0.170885297253)*sqrt(x[1])
        arg[1,4,0,2]=(0.533424480956)*sqrt(x[2])
        arg[1,4,1,0]=(-0.220533193308)*sqrt(x[0])
        arg[1,4,1,1]=(0.344537611882)*sqrt(x[1])
        arg[1,4,1,2]=(0.861473877282)*sqrt(x[2])
        arg[1,4,2,0]=(-0.0923010438884)*sqrt(x[0])
        arg[1,4,2,1]=(-0.338256780498)*sqrt(x[1])
        arg[1,4,2,2]=(0.528567959345)*sqrt(x[2])
        arg[2,0,0,0]=(-0.0423053381485)*sqrt(x[0])
        arg[2,0,0,1]=(0.856798579151)*sqrt(x[1])
        arg[2,0,0,2]=(0.383258153853)*sqrt(x[2])
        arg[2,0,1,0]=(0.350994872736)*sqrt(x[0])
        arg[2,0,1,1]=(-0.78055158106)*sqrt(x[1])
        arg[2,0,1,2]=(-0.770876699915)*sqrt(x[2])
        arg[2,0,2,0]=(-0.935133287106)*sqrt(x[0])
        arg[2,0,2,1]=(0.618238076989)*sqrt(x[1])
        arg[2,0,2,2]=(-0.846783087949)*sqrt(x[2])
        arg[2,1,0,0]=(-0.657394511405)*sqrt(x[0])
        arg[2,1,0,1]=(0.576218821654)*sqrt(x[1])
        arg[2,1,0,2]=(0.0269446356493)*sqrt(x[2])
        arg[2,1,1,0]=(-0.310710230949)*sqrt(x[0])
        arg[2,1,1,1]=(0.425412515598)*sqrt(x[1])
        arg[2,1,1,2]=(-0.225225290862)*sqrt(x[2])
        arg[2,1,2,0]=(-0.539928589495)*sqrt(x[0])
        arg[2,1,2,1]=(-0.348082121765)*sqrt(x[1])
        arg[2,1,2,2]=(-0.0287274646233)*sqrt(x[2])
        arg[2,2,0,0]=(0.746132865923)*sqrt(x[0])
        arg[2,2,0,1]=(-0.0234203693548)*sqrt(x[1])
        arg[2,2,0,2]=(0.517411821941)*sqrt(x[2])
        arg[2,2,1,0]=(-0.183204217349)*sqrt(x[0])
        arg[2,2,1,1]=(0.714988861836)*sqrt(x[1])
        arg[2,2,1,2]=(0.829083318937)*sqrt(x[2])
        arg[2,2,2,0]=(0.458067555841)*sqrt(x[0])
        arg[2,2,2,1]=(0.639317125869)*sqrt(x[1])
        arg[2,2,2,2]=(0.104611520408)*sqrt(x[2])
        arg[2,3,0,0]=(-0.420513461135)*sqrt(x[0])
        arg[2,3,0,1]=(0.888686162754)*sqrt(x[1])
        arg[2,3,0,2]=(0.939305777879)*sqrt(x[2])
        arg[2,3,1,0]=(0.856795132015)*sqrt(x[0])
        arg[2,3,1,1]=(0.817593141895)*sqrt(x[1])
        arg[2,3,1,2]=(0.962503342535)*sqrt(x[2])
        arg[2,3,2,0]=(-0.334586700245)*sqrt(x[0])
        arg[2,3,2,1]=(0.182696129528)*sqrt(x[1])
        arg[2,3,2,2]=(-0.707271571206)*sqrt(x[2])
        arg[2,4,0,0]=(-0.594795981069)*sqrt(x[0])
        arg[2,4,0,1]=(0.451239168073)*sqrt(x[1])
        arg[2,4,0,2]=(0.191212211556)*sqrt(x[2])
        arg[2,4,1,0]=(-0.503465984944)*sqrt(x[0])
        arg[2,4,1,1]=(0.725377884208)*sqrt(x[1])
        arg[2,4,1,2]=(-0.40719255752)*sqrt(x[2])
        arg[2,4,2,0]=(0.268071476451)*sqrt(x[0])
        arg[2,4,2,1]=(0.85066639942)*sqrt(x[1])
        arg[2,4,2,2]=(-0.906021406945)*sqrt(x[2])
        arg[3,0,0,0]=(-0.922179152122)*sqrt(x[0])
        arg[3,0,0,1]=(-0.0903841240007)*sqrt(x[1])
        arg[3,0,0,2]=(-0.751482803516)*sqrt(x[2])
        arg[3,0,1,0]=(0.960697809119)*sqrt(x[0])
        arg[3,0,1,1]=(0.638878873158)*sqrt(x[1])
        arg[3,0,1,2]=(0.390932234724)*sqrt(x[2])
        arg[3,0,2,0]=(-0.925078301694)*sqrt(x[0])
        arg[3,0,2,1]=(0.793590580665)*sqrt(x[1])
        arg[3,0,2,2]=(0.535478366911)*sqrt(x[2])
        arg[3,1,0,0]=(-0.431951993217)*sqrt(x[0])
        arg[3,1,0,1]=(0.211750261417)*sqrt(x[1])
        arg[3,1,0,2]=(-0.930706580442)*sqrt(x[2])
        arg[3,1,1,0]=(0.330979313323)*sqrt(x[0])
        arg[3,1,1,1]=(-0.838919076081)*sqrt(x[1])
        arg[3,1,1,2]=(0.134250050168)*sqrt(x[2])
        arg[3,1,2,0]=(0.414922811301)*sqrt(x[0])
        arg[3,1,2,1]=(-0.663692878121)*sqrt(x[1])
        arg[3,1,2,2]=(0.88499278543)*sqrt(x[2])
        arg[3,2,0,0]=(0.0884662742233)*sqrt(x[0])
        arg[3,2,0,1]=(-0.412630722821)*sqrt(x[1])
        arg[3,2,0,2]=(-0.730850884928)*sqrt(x[2])
        arg[3,2,1,0]=(0.722207547366)*sqrt(x[0])
        arg[3,2,1,1]=(-0.260067950749)*sqrt(x[1])
        arg[3,2,1,2]=(0.426259201494)*sqrt(x[2])
        arg[3,2,2,0]=(0.0516111795322)*sqrt(x[0])
        arg[3,2,2,1]=(0.922853710048)*sqrt(x[1])
        arg[3,2,2,2]=(-0.991912758116)*sqrt(x[2])
        arg[3,3,0,0]=(0.263933965905)*sqrt(x[0])
        arg[3,3,0,1]=(0.840541758799)*sqrt(x[1])
        arg[3,3,0,2]=(0.417658511125)*sqrt(x[2])
        arg[3,3,1,0]=(-0.901745614723)*sqrt(x[0])
        arg[3,3,1,1]=(-0.623608908699)*sqrt(x[1])
        arg[3,3,1,2]=(0.0522167208784)*sqrt(x[2])
        arg[3,3,2,0]=(-0.549431264931)*sqrt(x[0])
        arg[3,3,2,1]=(0.919971855457)*sqrt(x[1])
        arg[3,3,2,2]=(0.142757773397)*sqrt(x[2])
        arg[3,4,0,0]=(-0.258875259824)*sqrt(x[0])
        arg[3,4,0,1]=(-0.0373872187041)*sqrt(x[1])
        arg[3,4,0,2]=(0.445989164864)*sqrt(x[2])
        arg[3,4,1,0]=(0.0273711397038)*sqrt(x[0])
        arg[3,4,1,1]=(-0.85522629)*sqrt(x[1])
        arg[3,4,1,2]=(-0.835392581226)*sqrt(x[2])
        arg[3,4,2,0]=(0.402929450189)*sqrt(x[0])
        arg[3,4,2,1]=(0.549623033221)*sqrt(x[1])
        arg[3,4,2,2]=(0.237803212109)*sqrt(x[2])
        ref=sqrt((15.1448100099)*3./2.+(3.61598249819)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactOne_fromData_rank0(self):
      """
      tests L2-norm of Data on the FunctionOnContactOne

      assumptions: self.domain supports integration on FunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(0.39039157828)*x[0]
        ref=sqrt((0.0)+(0.038101396098))

      else:
        arg=Data(0,(),w)
        arg=(0.893204658234)*x[0]
        ref=sqrt((0.0)+(0.199453640373))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactOne_fromData_rank1(self):
      """
      tests L2-norm of Data on the FunctionOnContactOne

      assumptions: self.domain supports integration on FunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.697576620054)*x[0]
        arg[1]=(-0.302037273777)*x[1]
        ref=sqrt((0.0304088382502)+(0.121653285212))

      else:
        arg=Data(0,(3,),w)
        arg[0]=(0.649308272675)*x[0]
        arg[1]=(-0.408829699836)*x[1]
        arg[2]=(-0.984892977949)*x[2]
        ref=sqrt((0.37905196716)+(0.105400308241))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactOne_fromData_rank2(self):
      """
      tests L2-norm of Data on the FunctionOnContactOne

      assumptions: self.domain supports integration on FunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.504133645896)*x[0]
        arg[0,1]=(0.728228500775)*x[1]
        arg[1,0]=(0.968237164633)*x[0]
        arg[1,1]=(-0.205351628957)*x[1]
        arg[2,0]=(-0.154008228702)*x[0]
        arg[2,1]=(-0.972088404566)*x[1]
        arg[3,0]=(0.585153521593)*x[0]
        arg[3,1]=(0.371861820001)*x[1]
        ref=sqrt((0.551907706774)+(0.389439279561))

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.4847970605)*x[0]
        arg[0,1]=(-0.664433021479)*x[1]
        arg[0,2]=(-0.93730914676)*x[2]
        arg[1,0]=(0.712624065023)*x[0]
        arg[1,1]=(0.105431976424)*x[1]
        arg[1,2]=(-0.452636806104)*x[2]
        arg[2,0]=(0.488406064176)*x[0]
        arg[2,1]=(0.0594879976584)*x[1]
        arg[2,2]=(0.596963883202)*x[2]
        arg[3,0]=(0.619366395328)*x[0]
        arg[3,1]=(0.719965901041)*x[1]
        arg[3,2]=(-0.568971277169)*x[2]
        ref=sqrt((0.912666523048)+(0.341254115776))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactOne_fromData_rank3(self):
      """
      tests L2-norm of Data on the FunctionOnContactOne

      assumptions: self.domain supports integration on FunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.481233278183)*x[0]
        arg[0,0,1]=(-0.883447576048)*x[1]
        arg[0,1,0]=(0.525938652633)*x[0]
        arg[0,1,1]=(-0.694971850651)*x[1]
        arg[1,0,0]=(-0.935880633291)*x[0]
        arg[1,0,1]=(0.970445286468)*x[1]
        arg[1,1,0]=(-0.302571977945)*x[0]
        arg[1,1,1]=(-0.237750319821)*x[1]
        arg[2,0,0]=(0.922940652883)*x[0]
        arg[2,0,1]=(-0.624756614078)*x[1]
        arg[2,1,0]=(-0.146839388764)*x[0]
        arg[2,1,1]=(-0.995551211092)*x[1]
        arg[3,0,0]=(0.719184541649)*x[0]
        arg[3,0,1]=(-0.629461063375)*x[1]
        arg[3,1,0]=(-0.925581215276)*x[0]
        arg[3,1,1]=(-0.564418658495)*x[1]
        arg[4,0,0]=(-0.724035884026)*x[0]
        arg[4,0,1]=(0.78393984221)*x[1]
        arg[4,1,0]=(-0.722375249317)*x[0]
        arg[4,1,1]=(0.27571990944)*x[1]
        arg[5,0,0]=(-0.197579392553)*x[0]
        arg[5,0,1]=(0.341717398275)*x[1]
        arg[5,1,0]=(0.926327822859)*x[0]
        arg[5,1,1]=(0.44114488285)*x[1]
        ref=sqrt((1.78665006238)+(1.41652558894))

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(-0.348448140437)*x[0]
        arg[0,0,1]=(-0.530793589782)*x[1]
        arg[0,0,2]=(0.386511493376)*x[2]
        arg[0,1,0]=(0.00776536299593)*x[0]
        arg[0,1,1]=(-0.0696129074064)*x[1]
        arg[0,1,2]=(0.114179504216)*x[2]
        arg[1,0,0]=(-0.681009917406)*x[0]
        arg[1,0,1]=(0.763595403719)*x[1]
        arg[1,0,2]=(-0.54459342544)*x[2]
        arg[1,1,0]=(0.0839255723157)*x[0]
        arg[1,1,1]=(0.269389515003)*x[1]
        arg[1,1,2]=(0.612446509453)*x[2]
        arg[2,0,0]=(0.600070579238)*x[0]
        arg[2,0,1]=(-0.49585867531)*x[1]
        arg[2,0,2]=(-0.766857395141)*x[2]
        arg[2,1,0]=(-0.84515816441)*x[0]
        arg[2,1,1]=(-0.594337789049)*x[1]
        arg[2,1,2]=(0.774596009811)*x[2]
        arg[3,0,0]=(-0.833982234688)*x[0]
        arg[3,0,1]=(0.747667281218)*x[1]
        arg[3,0,2]=(-0.77025764051)*x[2]
        arg[3,1,0]=(0.0367023286497)*x[0]
        arg[3,1,1]=(-0.503080137993)*x[1]
        arg[3,1,2]=(-0.129478323797)*x[2]
        arg[4,0,0]=(0.304123530695)*x[0]
        arg[4,0,1]=(-0.873023368289)*x[1]
        arg[4,0,2]=(-0.712740806672)*x[2]
        arg[4,1,0]=(-0.0629132979843)*x[0]
        arg[4,1,1]=(0.096843438525)*x[1]
        arg[4,1,2]=(-0.608716559621)*x[2]
        arg[5,0,0]=(-0.642902822517)*x[0]
        arg[5,0,1]=(0.478134785533)*x[1]
        arg[5,0,2]=(0.757891512092)*x[2]
        arg[5,1,0]=(-0.144492755065)*x[0]
        arg[5,1,1]=(0.286775195938)*x[1]
        arg[5,1,2]=(-0.063580488638)*x[2]
        ref=sqrt((2.50835195543)+(0.723549067206))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnContactOne_fromData_rank4(self):
      """
      tests L2-norm of Data on the FunctionOnContactOne

      assumptions: self.domain supports integration on FunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=FunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(0.523372259748)*x[0]
        arg[0,0,0,1]=(-0.732795389957)*x[1]
        arg[0,0,1,0]=(0.0571737128606)*x[0]
        arg[0,0,1,1]=(-0.627260438721)*x[1]
        arg[0,0,2,0]=(-0.945159375045)*x[0]
        arg[0,0,2,1]=(-0.145103328237)*x[1]
        arg[0,1,0,0]=(-0.0586985419455)*x[0]
        arg[0,1,0,1]=(-0.788101334389)*x[1]
        arg[0,1,1,0]=(0.615023802803)*x[0]
        arg[0,1,1,1]=(0.965381283357)*x[1]
        arg[0,1,2,0]=(0.173979432067)*x[0]
        arg[0,1,2,1]=(-0.980483156424)*x[1]
        arg[0,2,0,0]=(-0.345940909471)*x[0]
        arg[0,2,0,1]=(0.00162430279595)*x[1]
        arg[0,2,1,0]=(-0.0847413350082)*x[0]
        arg[0,2,1,1]=(-0.911248036753)*x[1]
        arg[0,2,2,0]=(-0.817295589942)*x[0]
        arg[0,2,2,1]=(-0.0514569401185)*x[1]
        arg[0,3,0,0]=(0.329704440159)*x[0]
        arg[0,3,0,1]=(0.580379057442)*x[1]
        arg[0,3,1,0]=(0.737014729605)*x[0]
        arg[0,3,1,1]=(0.535521740242)*x[1]
        arg[0,3,2,0]=(-0.943554721405)*x[0]
        arg[0,3,2,1]=(0.798294295959)*x[1]
        arg[0,4,0,0]=(-0.1814410847)*x[0]
        arg[0,4,0,1]=(0.968476629577)*x[1]
        arg[0,4,1,0]=(0.26595250332)*x[0]
        arg[0,4,1,1]=(-0.493516007909)*x[1]
        arg[0,4,2,0]=(0.374323739856)*x[0]
        arg[0,4,2,1]=(-0.914413601829)*x[1]
        arg[1,0,0,0]=(-0.701090812201)*x[0]
        arg[1,0,0,1]=(-0.403596866322)*x[1]
        arg[1,0,1,0]=(0.587183778075)*x[0]
        arg[1,0,1,1]=(-0.218129059558)*x[1]
        arg[1,0,2,0]=(-0.0205642762349)*x[0]
        arg[1,0,2,1]=(0.441382235525)*x[1]
        arg[1,1,0,0]=(0.909683859519)*x[0]
        arg[1,1,0,1]=(0.960088832754)*x[1]
        arg[1,1,1,0]=(0.0450355911985)*x[0]
        arg[1,1,1,1]=(0.366849576658)*x[1]
        arg[1,1,2,0]=(0.315490164497)*x[0]
        arg[1,1,2,1]=(0.533746946808)*x[1]
        arg[1,2,0,0]=(-0.552919155957)*x[0]
        arg[1,2,0,1]=(-0.647373009727)*x[1]
        arg[1,2,1,0]=(0.32617079311)*x[0]
        arg[1,2,1,1]=(-0.3424263412)*x[1]
        arg[1,2,2,0]=(0.848733118893)*x[0]
        arg[1,2,2,1]=(-0.65341712364)*x[1]
        arg[1,3,0,0]=(-0.698538235208)*x[0]
        arg[1,3,0,1]=(-0.176858090319)*x[1]
        arg[1,3,1,0]=(-0.956962782849)*x[0]
        arg[1,3,1,1]=(0.53041598553)*x[1]
        arg[1,3,2,0]=(-0.241715236701)*x[0]
        arg[1,3,2,1]=(0.0340608544103)*x[1]
        arg[1,4,0,0]=(-0.399279751219)*x[0]
        arg[1,4,0,1]=(0.383860128496)*x[1]
        arg[1,4,1,0]=(0.229189108307)*x[0]
        arg[1,4,1,1]=(0.963667605308)*x[1]
        arg[1,4,2,0]=(0.731912848998)*x[0]
        arg[1,4,2,1]=(0.0534970206048)*x[1]
        arg[2,0,0,0]=(0.422292915614)*x[0]
        arg[2,0,0,1]=(0.915607515426)*x[1]
        arg[2,0,1,0]=(0.451054723135)*x[0]
        arg[2,0,1,1]=(0.0400870534188)*x[1]
        arg[2,0,2,0]=(-0.851047453002)*x[0]
        arg[2,0,2,1]=(0.300957420969)*x[1]
        arg[2,1,0,0]=(0.636801515211)*x[0]
        arg[2,1,0,1]=(0.386798062533)*x[1]
        arg[2,1,1,0]=(-0.0501653606261)*x[0]
        arg[2,1,1,1]=(0.46515271421)*x[1]
        arg[2,1,2,0]=(-0.416436918527)*x[0]
        arg[2,1,2,1]=(0.214018855333)*x[1]
        arg[2,2,0,0]=(0.0516439357756)*x[0]
        arg[2,2,0,1]=(-0.237707451356)*x[1]
        arg[2,2,1,0]=(0.117559050121)*x[0]
        arg[2,2,1,1]=(0.164349348553)*x[1]
        arg[2,2,2,0]=(0.0524038362463)*x[0]
        arg[2,2,2,1]=(0.690751059081)*x[1]
        arg[2,3,0,0]=(-0.942079336547)*x[0]
        arg[2,3,0,1]=(0.340143899875)*x[1]
        arg[2,3,1,0]=(-0.630411753789)*x[0]
        arg[2,3,1,1]=(-0.722078100418)*x[1]
        arg[2,3,2,0]=(-0.310606406168)*x[0]
        arg[2,3,2,1]=(0.728069014178)*x[1]
        arg[2,4,0,0]=(0.981678768954)*x[0]
        arg[2,4,0,1]=(-0.273023706956)*x[1]
        arg[2,4,1,0]=(0.307187251211)*x[0]
        arg[2,4,1,1]=(-0.885309661176)*x[1]
        arg[2,4,2,0]=(-0.808022645995)*x[0]
        arg[2,4,2,1]=(-0.21488080753)*x[1]
        arg[3,0,0,0]=(-0.785416067695)*x[0]
        arg[3,0,0,1]=(0.344457091999)*x[1]
        arg[3,0,1,0]=(-0.845835855973)*x[0]
        arg[3,0,1,1]=(0.365474280353)*x[1]
        arg[3,0,2,0]=(-0.250535373488)*x[0]
        arg[3,0,2,1]=(-0.888004656815)*x[1]
        arg[3,1,0,0]=(-0.0652508248717)*x[0]
        arg[3,1,0,1]=(0.402069695186)*x[1]
        arg[3,1,1,0]=(-0.229714763188)*x[0]
        arg[3,1,1,1]=(-0.866938136156)*x[1]
        arg[3,1,2,0]=(-0.638603745536)*x[0]
        arg[3,1,2,1]=(-0.997108734664)*x[1]
        arg[3,2,0,0]=(-0.182784593825)*x[0]
        arg[3,2,0,1]=(0.426951207538)*x[1]
        arg[3,2,1,0]=(-0.958149736654)*x[0]
        arg[3,2,1,1]=(-0.58458370587)*x[1]
        arg[3,2,2,0]=(0.410243417149)*x[0]
        arg[3,2,2,1]=(0.764474507608)*x[1]
        arg[3,3,0,0]=(0.235623988334)*x[0]
        arg[3,3,0,1]=(-0.105137401903)*x[1]
        arg[3,3,1,0]=(-0.957743319072)*x[0]
        arg[3,3,1,1]=(0.285680328399)*x[1]
        arg[3,3,2,0]=(-0.525269717395)*x[0]
        arg[3,3,2,1]=(0.655974370871)*x[1]
        arg[3,4,0,0]=(0.926779313365)*x[0]
        arg[3,4,0,1]=(-0.111556488031)*x[1]
        arg[3,4,1,0]=(-0.672785588938)*x[0]
        arg[3,4,1,1]=(-0.687907843899)*x[1]
        arg[3,4,2,0]=(0.503984478085)*x[0]
        arg[3,4,2,1]=(0.213194091516)*x[1]
        ref=sqrt((6.92178561091)+(4.96610502195))

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(-0.641108959812)*x[0]
        arg[0,0,0,1]=(0.395837684631)*x[1]
        arg[0,0,0,2]=(0.36258240616)*x[2]
        arg[0,0,1,0]=(-0.0543293722691)*x[0]
        arg[0,0,1,1]=(0.850820297369)*x[1]
        arg[0,0,1,2]=(0.0910730209593)*x[2]
        arg[0,0,2,0]=(-0.831108730829)*x[0]
        arg[0,0,2,1]=(-0.653018722047)*x[1]
        arg[0,0,2,2]=(0.29052921548)*x[2]
        arg[0,1,0,0]=(-0.115846569271)*x[0]
        arg[0,1,0,1]=(-0.828080308748)*x[1]
        arg[0,1,0,2]=(0.715255255273)*x[2]
        arg[0,1,1,0]=(-0.0106077934044)*x[0]
        arg[0,1,1,1]=(0.358323295554)*x[1]
        arg[0,1,1,2]=(-0.916496266702)*x[2]
        arg[0,1,2,0]=(0.740969432327)*x[0]
        arg[0,1,2,1]=(0.0631507192754)*x[1]
        arg[0,1,2,2]=(0.440009320758)*x[2]
        arg[0,2,0,0]=(0.689964376385)*x[0]
        arg[0,2,0,1]=(0.58199412545)*x[1]
        arg[0,2,0,2]=(0.649058556936)*x[2]
        arg[0,2,1,0]=(0.811660544475)*x[0]
        arg[0,2,1,1]=(-0.121902248927)*x[1]
        arg[0,2,1,2]=(-0.0750232179247)*x[2]
        arg[0,2,2,0]=(0.779166450882)*x[0]
        arg[0,2,2,1]=(0.298560101475)*x[1]
        arg[0,2,2,2]=(-0.776198356206)*x[2]
        arg[0,3,0,0]=(0.79804128408)*x[0]
        arg[0,3,0,1]=(-0.378223308322)*x[1]
        arg[0,3,0,2]=(0.191534109359)*x[2]
        arg[0,3,1,0]=(-0.59973117885)*x[0]
        arg[0,3,1,1]=(0.149042766249)*x[1]
        arg[0,3,1,2]=(-0.357473233062)*x[2]
        arg[0,3,2,0]=(-0.343758339283)*x[0]
        arg[0,3,2,1]=(-0.140764743713)*x[1]
        arg[0,3,2,2]=(-0.506377878901)*x[2]
        arg[0,4,0,0]=(-0.967765607304)*x[0]
        arg[0,4,0,1]=(-0.121373931498)*x[1]
        arg[0,4,0,2]=(-0.300794608584)*x[2]
        arg[0,4,1,0]=(-0.604537150389)*x[0]
        arg[0,4,1,1]=(-0.290834102056)*x[1]
        arg[0,4,1,2]=(0.509331591828)*x[2]
        arg[0,4,2,0]=(0.643653985564)*x[0]
        arg[0,4,2,1]=(-0.599746831159)*x[1]
        arg[0,4,2,2]=(0.843193710828)*x[2]
        arg[1,0,0,0]=(-0.598249432719)*x[0]
        arg[1,0,0,1]=(0.840724468488)*x[1]
        arg[1,0,0,2]=(0.558893493847)*x[2]
        arg[1,0,1,0]=(0.843245438555)*x[0]
        arg[1,0,1,1]=(0.280919545563)*x[1]
        arg[1,0,1,2]=(-0.0549504830615)*x[2]
        arg[1,0,2,0]=(0.295812998973)*x[0]
        arg[1,0,2,1]=(0.513221249344)*x[1]
        arg[1,0,2,2]=(-0.432253600492)*x[2]
        arg[1,1,0,0]=(0.300696333449)*x[0]
        arg[1,1,0,1]=(0.964202717779)*x[1]
        arg[1,1,0,2]=(0.131730374204)*x[2]
        arg[1,1,1,0]=(0.209135189207)*x[0]
        arg[1,1,1,1]=(0.301389564979)*x[1]
        arg[1,1,1,2]=(0.378481289816)*x[2]
        arg[1,1,2,0]=(0.407413524533)*x[0]
        arg[1,1,2,1]=(0.184345610049)*x[1]
        arg[1,1,2,2]=(-0.884401323556)*x[2]
        arg[1,2,0,0]=(0.886883417187)*x[0]
        arg[1,2,0,1]=(-0.0524504232352)*x[1]
        arg[1,2,0,2]=(0.049116549681)*x[2]
        arg[1,2,1,0]=(-0.41240815391)*x[0]
        arg[1,2,1,1]=(-0.523033558623)*x[1]
        arg[1,2,1,2]=(-0.920238404199)*x[2]
        arg[1,2,2,0]=(-0.635169424461)*x[0]
        arg[1,2,2,1]=(0.773163888164)*x[1]
        arg[1,2,2,2]=(0.554400056145)*x[2]
        arg[1,3,0,0]=(0.573361031392)*x[0]
        arg[1,3,0,1]=(-0.107749750058)*x[1]
        arg[1,3,0,2]=(-0.971799077244)*x[2]
        arg[1,3,1,0]=(-0.103579714499)*x[0]
        arg[1,3,1,1]=(0.93329275867)*x[1]
        arg[1,3,1,2]=(0.994409555402)*x[2]
        arg[1,3,2,0]=(0.571961012279)*x[0]
        arg[1,3,2,1]=(-0.54961086377)*x[1]
        arg[1,3,2,2]=(0.563815465418)*x[2]
        arg[1,4,0,0]=(-0.820638948871)*x[0]
        arg[1,4,0,1]=(0.407750087109)*x[1]
        arg[1,4,0,2]=(0.0911636159786)*x[2]
        arg[1,4,1,0]=(0.371133349901)*x[0]
        arg[1,4,1,1]=(0.688167208194)*x[1]
        arg[1,4,1,2]=(0.188789217725)*x[2]
        arg[1,4,2,0]=(0.961439941063)*x[0]
        arg[1,4,2,1]=(0.524326385264)*x[1]
        arg[1,4,2,2]=(-0.0866064394717)*x[2]
        arg[2,0,0,0]=(0.392756517944)*x[0]
        arg[2,0,0,1]=(0.713723092455)*x[1]
        arg[2,0,0,2]=(0.68441988583)*x[2]
        arg[2,0,1,0]=(-0.471505369012)*x[0]
        arg[2,0,1,1]=(0.152712812211)*x[1]
        arg[2,0,1,2]=(-0.579106568625)*x[2]
        arg[2,0,2,0]=(0.637759469489)*x[0]
        arg[2,0,2,1]=(-0.986405270597)*x[1]
        arg[2,0,2,2]=(0.498109007475)*x[2]
        arg[2,1,0,0]=(0.870185247694)*x[0]
        arg[2,1,0,1]=(0.738095675991)*x[1]
        arg[2,1,0,2]=(-0.613836737976)*x[2]
        arg[2,1,1,0]=(-0.507009374688)*x[0]
        arg[2,1,1,1]=(0.00752595652646)*x[1]
        arg[2,1,1,2]=(-0.389118104751)*x[2]
        arg[2,1,2,0]=(-0.466733433552)*x[0]
        arg[2,1,2,1]=(-0.356829687828)*x[1]
        arg[2,1,2,2]=(0.691337110432)*x[2]
        arg[2,2,0,0]=(-0.684116912917)*x[0]
        arg[2,2,0,1]=(0.705774798227)*x[1]
        arg[2,2,0,2]=(0.522279733527)*x[2]
        arg[2,2,1,0]=(0.390315182318)*x[0]
        arg[2,2,1,1]=(-0.450165044241)*x[1]
        arg[2,2,1,2]=(0.436062806791)*x[2]
        arg[2,2,2,0]=(0.973600590724)*x[0]
        arg[2,2,2,1]=(-0.89116205465)*x[1]
        arg[2,2,2,2]=(-0.575262591791)*x[2]
        arg[2,3,0,0]=(0.842833949385)*x[0]
        arg[2,3,0,1]=(-0.69939817904)*x[1]
        arg[2,3,0,2]=(0.343679860971)*x[2]
        arg[2,3,1,0]=(-0.387030334358)*x[0]
        arg[2,3,1,1]=(0.901637622406)*x[1]
        arg[2,3,1,2]=(0.054184099786)*x[2]
        arg[2,3,2,0]=(0.652712575782)*x[0]
        arg[2,3,2,1]=(0.286474075274)*x[1]
        arg[2,3,2,2]=(-0.745166160145)*x[2]
        arg[2,4,0,0]=(-0.104591304669)*x[0]
        arg[2,4,0,1]=(0.417496371683)*x[1]
        arg[2,4,0,2]=(0.0579341684769)*x[2]
        arg[2,4,1,0]=(0.24486914915)*x[0]
        arg[2,4,1,1]=(-0.166211416717)*x[1]
        arg[2,4,1,2]=(0.651515760357)*x[2]
        arg[2,4,2,0]=(0.934693067557)*x[0]
        arg[2,4,2,1]=(0.628486602755)*x[1]
        arg[2,4,2,2]=(0.855313741494)*x[2]
        arg[3,0,0,0]=(0.2241978058)*x[0]
        arg[3,0,0,1]=(0.389284487657)*x[1]
        arg[3,0,0,2]=(0.711521372962)*x[2]
        arg[3,0,1,0]=(0.653184309521)*x[0]
        arg[3,0,1,1]=(0.270052840275)*x[1]
        arg[3,0,1,2]=(0.388881294549)*x[2]
        arg[3,0,2,0]=(0.973583452692)*x[0]
        arg[3,0,2,1]=(0.508175099792)*x[1]
        arg[3,0,2,2]=(-0.989999307019)*x[2]
        arg[3,1,0,0]=(0.935765996464)*x[0]
        arg[3,1,0,1]=(-0.186350920716)*x[1]
        arg[3,1,0,2]=(-0.809013452979)*x[2]
        arg[3,1,1,0]=(0.467966757364)*x[0]
        arg[3,1,1,1]=(-0.347420416508)*x[1]
        arg[3,1,1,2]=(0.686386725318)*x[2]
        arg[3,1,2,0]=(-0.8117641465)*x[0]
        arg[3,1,2,1]=(-0.357390874294)*x[1]
        arg[3,1,2,2]=(-0.0202262660381)*x[2]
        arg[3,2,0,0]=(-0.181594076044)*x[0]
        arg[3,2,0,1]=(0.0655758078072)*x[1]
        arg[3,2,0,2]=(0.926686460908)*x[2]
        arg[3,2,1,0]=(-0.700426183586)*x[0]
        arg[3,2,1,1]=(0.600692261192)*x[1]
        arg[3,2,1,2]=(0.757146807006)*x[2]
        arg[3,2,2,0]=(0.588097603826)*x[0]
        arg[3,2,2,1]=(-0.748732775295)*x[1]
        arg[3,2,2,2]=(0.628960378488)*x[2]
        arg[3,3,0,0]=(-0.0612899752684)*x[0]
        arg[3,3,0,1]=(0.567198903535)*x[1]
        arg[3,3,0,2]=(0.743996205698)*x[2]
        arg[3,3,1,0]=(0.333763157477)*x[0]
        arg[3,3,1,1]=(0.795773692731)*x[1]
        arg[3,3,1,2]=(-0.676301506816)*x[2]
        arg[3,3,2,0]=(-0.255504087739)*x[0]
        arg[3,3,2,1]=(-0.556263970561)*x[1]
        arg[3,3,2,2]=(0.721245074469)*x[2]
        arg[3,4,0,0]=(0.384551582302)*x[0]
        arg[3,4,0,1]=(0.849519144853)*x[1]
        arg[3,4,0,2]=(0.101881023262)*x[2]
        arg[3,4,1,0]=(-0.685891926653)*x[0]
        arg[3,4,1,1]=(0.390887672448)*x[1]
        arg[3,4,1,2]=(0.0911617386398)*x[2]
        arg[3,4,2,0]=(-0.889618312513)*x[0]
        arg[3,4,2,1]=(0.130310876981)*x[1]
        arg[3,4,2,2]=(0.958562278204)*x[2]
        ref=sqrt((12.9075635889)+(5.72789470359))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactOne_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactOne

      assumptions: self.domain supports integration on ReducedFunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(0.39039157828)*sqrt(x[0])
        ref=sqrt((0.0)*3./2.+(0.038101396098)/0.5)

      else:
        arg=Data(0,(),w)
        arg=(0.893204658234)*sqrt(x[0])
        ref=sqrt((0.0)*3./2.+(0.199453640373)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactOne_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactOne

      assumptions: self.domain supports integration on ReducedFunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.697576620054)*sqrt(x[0])
        arg[1]=(-0.302037273777)*sqrt(x[1])
        ref=sqrt((0.0304088382502)*3./2.+(0.121653285212)/0.5)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(0.649308272675)*sqrt(x[0])
        arg[1]=(-0.408829699836)*sqrt(x[1])
        arg[2]=(-0.984892977949)*sqrt(x[2])
        ref=sqrt((0.37905196716)*3./2.+(0.105400308241)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactOne_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactOne

      assumptions: self.domain supports integration on ReducedFunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.504133645896)*sqrt(x[0])
        arg[0,1]=(0.728228500775)*sqrt(x[1])
        arg[1,0]=(0.968237164633)*sqrt(x[0])
        arg[1,1]=(-0.205351628957)*sqrt(x[1])
        arg[2,0]=(-0.154008228702)*sqrt(x[0])
        arg[2,1]=(-0.972088404566)*sqrt(x[1])
        arg[3,0]=(0.585153521593)*sqrt(x[0])
        arg[3,1]=(0.371861820001)*sqrt(x[1])
        ref=sqrt((0.551907706774)*3./2.+(0.389439279561)/0.5)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.4847970605)*sqrt(x[0])
        arg[0,1]=(-0.664433021479)*sqrt(x[1])
        arg[0,2]=(-0.93730914676)*sqrt(x[2])
        arg[1,0]=(0.712624065023)*sqrt(x[0])
        arg[1,1]=(0.105431976424)*sqrt(x[1])
        arg[1,2]=(-0.452636806104)*sqrt(x[2])
        arg[2,0]=(0.488406064176)*sqrt(x[0])
        arg[2,1]=(0.0594879976584)*sqrt(x[1])
        arg[2,2]=(0.596963883202)*sqrt(x[2])
        arg[3,0]=(0.619366395328)*sqrt(x[0])
        arg[3,1]=(0.719965901041)*sqrt(x[1])
        arg[3,2]=(-0.568971277169)*sqrt(x[2])
        ref=sqrt((0.912666523048)*3./2.+(0.341254115776)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactOne_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactOne

      assumptions: self.domain supports integration on ReducedFunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.481233278183)*sqrt(x[0])
        arg[0,0,1]=(-0.883447576048)*sqrt(x[1])
        arg[0,1,0]=(0.525938652633)*sqrt(x[0])
        arg[0,1,1]=(-0.694971850651)*sqrt(x[1])
        arg[1,0,0]=(-0.935880633291)*sqrt(x[0])
        arg[1,0,1]=(0.970445286468)*sqrt(x[1])
        arg[1,1,0]=(-0.302571977945)*sqrt(x[0])
        arg[1,1,1]=(-0.237750319821)*sqrt(x[1])
        arg[2,0,0]=(0.922940652883)*sqrt(x[0])
        arg[2,0,1]=(-0.624756614078)*sqrt(x[1])
        arg[2,1,0]=(-0.146839388764)*sqrt(x[0])
        arg[2,1,1]=(-0.995551211092)*sqrt(x[1])
        arg[3,0,0]=(0.719184541649)*sqrt(x[0])
        arg[3,0,1]=(-0.629461063375)*sqrt(x[1])
        arg[3,1,0]=(-0.925581215276)*sqrt(x[0])
        arg[3,1,1]=(-0.564418658495)*sqrt(x[1])
        arg[4,0,0]=(-0.724035884026)*sqrt(x[0])
        arg[4,0,1]=(0.78393984221)*sqrt(x[1])
        arg[4,1,0]=(-0.722375249317)*sqrt(x[0])
        arg[4,1,1]=(0.27571990944)*sqrt(x[1])
        arg[5,0,0]=(-0.197579392553)*sqrt(x[0])
        arg[5,0,1]=(0.341717398275)*sqrt(x[1])
        arg[5,1,0]=(0.926327822859)*sqrt(x[0])
        arg[5,1,1]=(0.44114488285)*sqrt(x[1])
        ref=sqrt((1.78665006238)*3./2.+(1.41652558894)/0.5)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(-0.348448140437)*sqrt(x[0])
        arg[0,0,1]=(-0.530793589782)*sqrt(x[1])
        arg[0,0,2]=(0.386511493376)*sqrt(x[2])
        arg[0,1,0]=(0.00776536299593)*sqrt(x[0])
        arg[0,1,1]=(-0.0696129074064)*sqrt(x[1])
        arg[0,1,2]=(0.114179504216)*sqrt(x[2])
        arg[1,0,0]=(-0.681009917406)*sqrt(x[0])
        arg[1,0,1]=(0.763595403719)*sqrt(x[1])
        arg[1,0,2]=(-0.54459342544)*sqrt(x[2])
        arg[1,1,0]=(0.0839255723157)*sqrt(x[0])
        arg[1,1,1]=(0.269389515003)*sqrt(x[1])
        arg[1,1,2]=(0.612446509453)*sqrt(x[2])
        arg[2,0,0]=(0.600070579238)*sqrt(x[0])
        arg[2,0,1]=(-0.49585867531)*sqrt(x[1])
        arg[2,0,2]=(-0.766857395141)*sqrt(x[2])
        arg[2,1,0]=(-0.84515816441)*sqrt(x[0])
        arg[2,1,1]=(-0.594337789049)*sqrt(x[1])
        arg[2,1,2]=(0.774596009811)*sqrt(x[2])
        arg[3,0,0]=(-0.833982234688)*sqrt(x[0])
        arg[3,0,1]=(0.747667281218)*sqrt(x[1])
        arg[3,0,2]=(-0.77025764051)*sqrt(x[2])
        arg[3,1,0]=(0.0367023286497)*sqrt(x[0])
        arg[3,1,1]=(-0.503080137993)*sqrt(x[1])
        arg[3,1,2]=(-0.129478323797)*sqrt(x[2])
        arg[4,0,0]=(0.304123530695)*sqrt(x[0])
        arg[4,0,1]=(-0.873023368289)*sqrt(x[1])
        arg[4,0,2]=(-0.712740806672)*sqrt(x[2])
        arg[4,1,0]=(-0.0629132979843)*sqrt(x[0])
        arg[4,1,1]=(0.096843438525)*sqrt(x[1])
        arg[4,1,2]=(-0.608716559621)*sqrt(x[2])
        arg[5,0,0]=(-0.642902822517)*sqrt(x[0])
        arg[5,0,1]=(0.478134785533)*sqrt(x[1])
        arg[5,0,2]=(0.757891512092)*sqrt(x[2])
        arg[5,1,0]=(-0.144492755065)*sqrt(x[0])
        arg[5,1,1]=(0.286775195938)*sqrt(x[1])
        arg[5,1,2]=(-0.063580488638)*sqrt(x[2])
        ref=sqrt((2.50835195543)*3./2.+(0.723549067206)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnContactOne_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnContactOne

      assumptions: self.domain supports integration on ReducedFunctionOnContactOne
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnContactOne(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(0.523372259748)*sqrt(x[0])
        arg[0,0,0,1]=(-0.732795389957)*sqrt(x[1])
        arg[0,0,1,0]=(0.0571737128606)*sqrt(x[0])
        arg[0,0,1,1]=(-0.627260438721)*sqrt(x[1])
        arg[0,0,2,0]=(-0.945159375045)*sqrt(x[0])
        arg[0,0,2,1]=(-0.145103328237)*sqrt(x[1])
        arg[0,1,0,0]=(-0.0586985419455)*sqrt(x[0])
        arg[0,1,0,1]=(-0.788101334389)*sqrt(x[1])
        arg[0,1,1,0]=(0.615023802803)*sqrt(x[0])
        arg[0,1,1,1]=(0.965381283357)*sqrt(x[1])
        arg[0,1,2,0]=(0.173979432067)*sqrt(x[0])
        arg[0,1,2,1]=(-0.980483156424)*sqrt(x[1])
        arg[0,2,0,0]=(-0.345940909471)*sqrt(x[0])
        arg[0,2,0,1]=(0.00162430279595)*sqrt(x[1])
        arg[0,2,1,0]=(-0.0847413350082)*sqrt(x[0])
        arg[0,2,1,1]=(-0.911248036753)*sqrt(x[1])
        arg[0,2,2,0]=(-0.817295589942)*sqrt(x[0])
        arg[0,2,2,1]=(-0.0514569401185)*sqrt(x[1])
        arg[0,3,0,0]=(0.329704440159)*sqrt(x[0])
        arg[0,3,0,1]=(0.580379057442)*sqrt(x[1])
        arg[0,3,1,0]=(0.737014729605)*sqrt(x[0])
        arg[0,3,1,1]=(0.535521740242)*sqrt(x[1])
        arg[0,3,2,0]=(-0.943554721405)*sqrt(x[0])
        arg[0,3,2,1]=(0.798294295959)*sqrt(x[1])
        arg[0,4,0,0]=(-0.1814410847)*sqrt(x[0])
        arg[0,4,0,1]=(0.968476629577)*sqrt(x[1])
        arg[0,4,1,0]=(0.26595250332)*sqrt(x[0])
        arg[0,4,1,1]=(-0.493516007909)*sqrt(x[1])
        arg[0,4,2,0]=(0.374323739856)*sqrt(x[0])
        arg[0,4,2,1]=(-0.914413601829)*sqrt(x[1])
        arg[1,0,0,0]=(-0.701090812201)*sqrt(x[0])
        arg[1,0,0,1]=(-0.403596866322)*sqrt(x[1])
        arg[1,0,1,0]=(0.587183778075)*sqrt(x[0])
        arg[1,0,1,1]=(-0.218129059558)*sqrt(x[1])
        arg[1,0,2,0]=(-0.0205642762349)*sqrt(x[0])
        arg[1,0,2,1]=(0.441382235525)*sqrt(x[1])
        arg[1,1,0,0]=(0.909683859519)*sqrt(x[0])
        arg[1,1,0,1]=(0.960088832754)*sqrt(x[1])
        arg[1,1,1,0]=(0.0450355911985)*sqrt(x[0])
        arg[1,1,1,1]=(0.366849576658)*sqrt(x[1])
        arg[1,1,2,0]=(0.315490164497)*sqrt(x[0])
        arg[1,1,2,1]=(0.533746946808)*sqrt(x[1])
        arg[1,2,0,0]=(-0.552919155957)*sqrt(x[0])
        arg[1,2,0,1]=(-0.647373009727)*sqrt(x[1])
        arg[1,2,1,0]=(0.32617079311)*sqrt(x[0])
        arg[1,2,1,1]=(-0.3424263412)*sqrt(x[1])
        arg[1,2,2,0]=(0.848733118893)*sqrt(x[0])
        arg[1,2,2,1]=(-0.65341712364)*sqrt(x[1])
        arg[1,3,0,0]=(-0.698538235208)*sqrt(x[0])
        arg[1,3,0,1]=(-0.176858090319)*sqrt(x[1])
        arg[1,3,1,0]=(-0.956962782849)*sqrt(x[0])
        arg[1,3,1,1]=(0.53041598553)*sqrt(x[1])
        arg[1,3,2,0]=(-0.241715236701)*sqrt(x[0])
        arg[1,3,2,1]=(0.0340608544103)*sqrt(x[1])
        arg[1,4,0,0]=(-0.399279751219)*sqrt(x[0])
        arg[1,4,0,1]=(0.383860128496)*sqrt(x[1])
        arg[1,4,1,0]=(0.229189108307)*sqrt(x[0])
        arg[1,4,1,1]=(0.963667605308)*sqrt(x[1])
        arg[1,4,2,0]=(0.731912848998)*sqrt(x[0])
        arg[1,4,2,1]=(0.0534970206048)*sqrt(x[1])
        arg[2,0,0,0]=(0.422292915614)*sqrt(x[0])
        arg[2,0,0,1]=(0.915607515426)*sqrt(x[1])
        arg[2,0,1,0]=(0.451054723135)*sqrt(x[0])
        arg[2,0,1,1]=(0.0400870534188)*sqrt(x[1])
        arg[2,0,2,0]=(-0.851047453002)*sqrt(x[0])
        arg[2,0,2,1]=(0.300957420969)*sqrt(x[1])
        arg[2,1,0,0]=(0.636801515211)*sqrt(x[0])
        arg[2,1,0,1]=(0.386798062533)*sqrt(x[1])
        arg[2,1,1,0]=(-0.0501653606261)*sqrt(x[0])
        arg[2,1,1,1]=(0.46515271421)*sqrt(x[1])
        arg[2,1,2,0]=(-0.416436918527)*sqrt(x[0])
        arg[2,1,2,1]=(0.214018855333)*sqrt(x[1])
        arg[2,2,0,0]=(0.0516439357756)*sqrt(x[0])
        arg[2,2,0,1]=(-0.237707451356)*sqrt(x[1])
        arg[2,2,1,0]=(0.117559050121)*sqrt(x[0])
        arg[2,2,1,1]=(0.164349348553)*sqrt(x[1])
        arg[2,2,2,0]=(0.0524038362463)*sqrt(x[0])
        arg[2,2,2,1]=(0.690751059081)*sqrt(x[1])
        arg[2,3,0,0]=(-0.942079336547)*sqrt(x[0])
        arg[2,3,0,1]=(0.340143899875)*sqrt(x[1])
        arg[2,3,1,0]=(-0.630411753789)*sqrt(x[0])
        arg[2,3,1,1]=(-0.722078100418)*sqrt(x[1])
        arg[2,3,2,0]=(-0.310606406168)*sqrt(x[0])
        arg[2,3,2,1]=(0.728069014178)*sqrt(x[1])
        arg[2,4,0,0]=(0.981678768954)*sqrt(x[0])
        arg[2,4,0,1]=(-0.273023706956)*sqrt(x[1])
        arg[2,4,1,0]=(0.307187251211)*sqrt(x[0])
        arg[2,4,1,1]=(-0.885309661176)*sqrt(x[1])
        arg[2,4,2,0]=(-0.808022645995)*sqrt(x[0])
        arg[2,4,2,1]=(-0.21488080753)*sqrt(x[1])
        arg[3,0,0,0]=(-0.785416067695)*sqrt(x[0])
        arg[3,0,0,1]=(0.344457091999)*sqrt(x[1])
        arg[3,0,1,0]=(-0.845835855973)*sqrt(x[0])
        arg[3,0,1,1]=(0.365474280353)*sqrt(x[1])
        arg[3,0,2,0]=(-0.250535373488)*sqrt(x[0])
        arg[3,0,2,1]=(-0.888004656815)*sqrt(x[1])
        arg[3,1,0,0]=(-0.0652508248717)*sqrt(x[0])
        arg[3,1,0,1]=(0.402069695186)*sqrt(x[1])
        arg[3,1,1,0]=(-0.229714763188)*sqrt(x[0])
        arg[3,1,1,1]=(-0.866938136156)*sqrt(x[1])
        arg[3,1,2,0]=(-0.638603745536)*sqrt(x[0])
        arg[3,1,2,1]=(-0.997108734664)*sqrt(x[1])
        arg[3,2,0,0]=(-0.182784593825)*sqrt(x[0])
        arg[3,2,0,1]=(0.426951207538)*sqrt(x[1])
        arg[3,2,1,0]=(-0.958149736654)*sqrt(x[0])
        arg[3,2,1,1]=(-0.58458370587)*sqrt(x[1])
        arg[3,2,2,0]=(0.410243417149)*sqrt(x[0])
        arg[3,2,2,1]=(0.764474507608)*sqrt(x[1])
        arg[3,3,0,0]=(0.235623988334)*sqrt(x[0])
        arg[3,3,0,1]=(-0.105137401903)*sqrt(x[1])
        arg[3,3,1,0]=(-0.957743319072)*sqrt(x[0])
        arg[3,3,1,1]=(0.285680328399)*sqrt(x[1])
        arg[3,3,2,0]=(-0.525269717395)*sqrt(x[0])
        arg[3,3,2,1]=(0.655974370871)*sqrt(x[1])
        arg[3,4,0,0]=(0.926779313365)*sqrt(x[0])
        arg[3,4,0,1]=(-0.111556488031)*sqrt(x[1])
        arg[3,4,1,0]=(-0.672785588938)*sqrt(x[0])
        arg[3,4,1,1]=(-0.687907843899)*sqrt(x[1])
        arg[3,4,2,0]=(0.503984478085)*sqrt(x[0])
        arg[3,4,2,1]=(0.213194091516)*sqrt(x[1])
        ref=sqrt((6.92178561091)*3./2.+(4.96610502195)/0.5)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(-0.641108959812)*sqrt(x[0])
        arg[0,0,0,1]=(0.395837684631)*sqrt(x[1])
        arg[0,0,0,2]=(0.36258240616)*sqrt(x[2])
        arg[0,0,1,0]=(-0.0543293722691)*sqrt(x[0])
        arg[0,0,1,1]=(0.850820297369)*sqrt(x[1])
        arg[0,0,1,2]=(0.0910730209593)*sqrt(x[2])
        arg[0,0,2,0]=(-0.831108730829)*sqrt(x[0])
        arg[0,0,2,1]=(-0.653018722047)*sqrt(x[1])
        arg[0,0,2,2]=(0.29052921548)*sqrt(x[2])
        arg[0,1,0,0]=(-0.115846569271)*sqrt(x[0])
        arg[0,1,0,1]=(-0.828080308748)*sqrt(x[1])
        arg[0,1,0,2]=(0.715255255273)*sqrt(x[2])
        arg[0,1,1,0]=(-0.0106077934044)*sqrt(x[0])
        arg[0,1,1,1]=(0.358323295554)*sqrt(x[1])
        arg[0,1,1,2]=(-0.916496266702)*sqrt(x[2])
        arg[0,1,2,0]=(0.740969432327)*sqrt(x[0])
        arg[0,1,2,1]=(0.0631507192754)*sqrt(x[1])
        arg[0,1,2,2]=(0.440009320758)*sqrt(x[2])
        arg[0,2,0,0]=(0.689964376385)*sqrt(x[0])
        arg[0,2,0,1]=(0.58199412545)*sqrt(x[1])
        arg[0,2,0,2]=(0.649058556936)*sqrt(x[2])
        arg[0,2,1,0]=(0.811660544475)*sqrt(x[0])
        arg[0,2,1,1]=(-0.121902248927)*sqrt(x[1])
        arg[0,2,1,2]=(-0.0750232179247)*sqrt(x[2])
        arg[0,2,2,0]=(0.779166450882)*sqrt(x[0])
        arg[0,2,2,1]=(0.298560101475)*sqrt(x[1])
        arg[0,2,2,2]=(-0.776198356206)*sqrt(x[2])
        arg[0,3,0,0]=(0.79804128408)*sqrt(x[0])
        arg[0,3,0,1]=(-0.378223308322)*sqrt(x[1])
        arg[0,3,0,2]=(0.191534109359)*sqrt(x[2])
        arg[0,3,1,0]=(-0.59973117885)*sqrt(x[0])
        arg[0,3,1,1]=(0.149042766249)*sqrt(x[1])
        arg[0,3,1,2]=(-0.357473233062)*sqrt(x[2])
        arg[0,3,2,0]=(-0.343758339283)*sqrt(x[0])
        arg[0,3,2,1]=(-0.140764743713)*sqrt(x[1])
        arg[0,3,2,2]=(-0.506377878901)*sqrt(x[2])
        arg[0,4,0,0]=(-0.967765607304)*sqrt(x[0])
        arg[0,4,0,1]=(-0.121373931498)*sqrt(x[1])
        arg[0,4,0,2]=(-0.300794608584)*sqrt(x[2])
        arg[0,4,1,0]=(-0.604537150389)*sqrt(x[0])
        arg[0,4,1,1]=(-0.290834102056)*sqrt(x[1])
        arg[0,4,1,2]=(0.509331591828)*sqrt(x[2])
        arg[0,4,2,0]=(0.643653985564)*sqrt(x[0])
        arg[0,4,2,1]=(-0.599746831159)*sqrt(x[1])
        arg[0,4,2,2]=(0.843193710828)*sqrt(x[2])
        arg[1,0,0,0]=(-0.598249432719)*sqrt(x[0])
        arg[1,0,0,1]=(0.840724468488)*sqrt(x[1])
        arg[1,0,0,2]=(0.558893493847)*sqrt(x[2])
        arg[1,0,1,0]=(0.843245438555)*sqrt(x[0])
        arg[1,0,1,1]=(0.280919545563)*sqrt(x[1])
        arg[1,0,1,2]=(-0.0549504830615)*sqrt(x[2])
        arg[1,0,2,0]=(0.295812998973)*sqrt(x[0])
        arg[1,0,2,1]=(0.513221249344)*sqrt(x[1])
        arg[1,0,2,2]=(-0.432253600492)*sqrt(x[2])
        arg[1,1,0,0]=(0.300696333449)*sqrt(x[0])
        arg[1,1,0,1]=(0.964202717779)*sqrt(x[1])
        arg[1,1,0,2]=(0.131730374204)*sqrt(x[2])
        arg[1,1,1,0]=(0.209135189207)*sqrt(x[0])
        arg[1,1,1,1]=(0.301389564979)*sqrt(x[1])
        arg[1,1,1,2]=(0.378481289816)*sqrt(x[2])
        arg[1,1,2,0]=(0.407413524533)*sqrt(x[0])
        arg[1,1,2,1]=(0.184345610049)*sqrt(x[1])
        arg[1,1,2,2]=(-0.884401323556)*sqrt(x[2])
        arg[1,2,0,0]=(0.886883417187)*sqrt(x[0])
        arg[1,2,0,1]=(-0.0524504232352)*sqrt(x[1])
        arg[1,2,0,2]=(0.049116549681)*sqrt(x[2])
        arg[1,2,1,0]=(-0.41240815391)*sqrt(x[0])
        arg[1,2,1,1]=(-0.523033558623)*sqrt(x[1])
        arg[1,2,1,2]=(-0.920238404199)*sqrt(x[2])
        arg[1,2,2,0]=(-0.635169424461)*sqrt(x[0])
        arg[1,2,2,1]=(0.773163888164)*sqrt(x[1])
        arg[1,2,2,2]=(0.554400056145)*sqrt(x[2])
        arg[1,3,0,0]=(0.573361031392)*sqrt(x[0])
        arg[1,3,0,1]=(-0.107749750058)*sqrt(x[1])
        arg[1,3,0,2]=(-0.971799077244)*sqrt(x[2])
        arg[1,3,1,0]=(-0.103579714499)*sqrt(x[0])
        arg[1,3,1,1]=(0.93329275867)*sqrt(x[1])
        arg[1,3,1,2]=(0.994409555402)*sqrt(x[2])
        arg[1,3,2,0]=(0.571961012279)*sqrt(x[0])
        arg[1,3,2,1]=(-0.54961086377)*sqrt(x[1])
        arg[1,3,2,2]=(0.563815465418)*sqrt(x[2])
        arg[1,4,0,0]=(-0.820638948871)*sqrt(x[0])
        arg[1,4,0,1]=(0.407750087109)*sqrt(x[1])
        arg[1,4,0,2]=(0.0911636159786)*sqrt(x[2])
        arg[1,4,1,0]=(0.371133349901)*sqrt(x[0])
        arg[1,4,1,1]=(0.688167208194)*sqrt(x[1])
        arg[1,4,1,2]=(0.188789217725)*sqrt(x[2])
        arg[1,4,2,0]=(0.961439941063)*sqrt(x[0])
        arg[1,4,2,1]=(0.524326385264)*sqrt(x[1])
        arg[1,4,2,2]=(-0.0866064394717)*sqrt(x[2])
        arg[2,0,0,0]=(0.392756517944)*sqrt(x[0])
        arg[2,0,0,1]=(0.713723092455)*sqrt(x[1])
        arg[2,0,0,2]=(0.68441988583)*sqrt(x[2])
        arg[2,0,1,0]=(-0.471505369012)*sqrt(x[0])
        arg[2,0,1,1]=(0.152712812211)*sqrt(x[1])
        arg[2,0,1,2]=(-0.579106568625)*sqrt(x[2])
        arg[2,0,2,0]=(0.637759469489)*sqrt(x[0])
        arg[2,0,2,1]=(-0.986405270597)*sqrt(x[1])
        arg[2,0,2,2]=(0.498109007475)*sqrt(x[2])
        arg[2,1,0,0]=(0.870185247694)*sqrt(x[0])
        arg[2,1,0,1]=(0.738095675991)*sqrt(x[1])
        arg[2,1,0,2]=(-0.613836737976)*sqrt(x[2])
        arg[2,1,1,0]=(-0.507009374688)*sqrt(x[0])
        arg[2,1,1,1]=(0.00752595652646)*sqrt(x[1])
        arg[2,1,1,2]=(-0.389118104751)*sqrt(x[2])
        arg[2,1,2,0]=(-0.466733433552)*sqrt(x[0])
        arg[2,1,2,1]=(-0.356829687828)*sqrt(x[1])
        arg[2,1,2,2]=(0.691337110432)*sqrt(x[2])
        arg[2,2,0,0]=(-0.684116912917)*sqrt(x[0])
        arg[2,2,0,1]=(0.705774798227)*sqrt(x[1])
        arg[2,2,0,2]=(0.522279733527)*sqrt(x[2])
        arg[2,2,1,0]=(0.390315182318)*sqrt(x[0])
        arg[2,2,1,1]=(-0.450165044241)*sqrt(x[1])
        arg[2,2,1,2]=(0.436062806791)*sqrt(x[2])
        arg[2,2,2,0]=(0.973600590724)*sqrt(x[0])
        arg[2,2,2,1]=(-0.89116205465)*sqrt(x[1])
        arg[2,2,2,2]=(-0.575262591791)*sqrt(x[2])
        arg[2,3,0,0]=(0.842833949385)*sqrt(x[0])
        arg[2,3,0,1]=(-0.69939817904)*sqrt(x[1])
        arg[2,3,0,2]=(0.343679860971)*sqrt(x[2])
        arg[2,3,1,0]=(-0.387030334358)*sqrt(x[0])
        arg[2,3,1,1]=(0.901637622406)*sqrt(x[1])
        arg[2,3,1,2]=(0.054184099786)*sqrt(x[2])
        arg[2,3,2,0]=(0.652712575782)*sqrt(x[0])
        arg[2,3,2,1]=(0.286474075274)*sqrt(x[1])
        arg[2,3,2,2]=(-0.745166160145)*sqrt(x[2])
        arg[2,4,0,0]=(-0.104591304669)*sqrt(x[0])
        arg[2,4,0,1]=(0.417496371683)*sqrt(x[1])
        arg[2,4,0,2]=(0.0579341684769)*sqrt(x[2])
        arg[2,4,1,0]=(0.24486914915)*sqrt(x[0])
        arg[2,4,1,1]=(-0.166211416717)*sqrt(x[1])
        arg[2,4,1,2]=(0.651515760357)*sqrt(x[2])
        arg[2,4,2,0]=(0.934693067557)*sqrt(x[0])
        arg[2,4,2,1]=(0.628486602755)*sqrt(x[1])
        arg[2,4,2,2]=(0.855313741494)*sqrt(x[2])
        arg[3,0,0,0]=(0.2241978058)*sqrt(x[0])
        arg[3,0,0,1]=(0.389284487657)*sqrt(x[1])
        arg[3,0,0,2]=(0.711521372962)*sqrt(x[2])
        arg[3,0,1,0]=(0.653184309521)*sqrt(x[0])
        arg[3,0,1,1]=(0.270052840275)*sqrt(x[1])
        arg[3,0,1,2]=(0.388881294549)*sqrt(x[2])
        arg[3,0,2,0]=(0.973583452692)*sqrt(x[0])
        arg[3,0,2,1]=(0.508175099792)*sqrt(x[1])
        arg[3,0,2,2]=(-0.989999307019)*sqrt(x[2])
        arg[3,1,0,0]=(0.935765996464)*sqrt(x[0])
        arg[3,1,0,1]=(-0.186350920716)*sqrt(x[1])
        arg[3,1,0,2]=(-0.809013452979)*sqrt(x[2])
        arg[3,1,1,0]=(0.467966757364)*sqrt(x[0])
        arg[3,1,1,1]=(-0.347420416508)*sqrt(x[1])
        arg[3,1,1,2]=(0.686386725318)*sqrt(x[2])
        arg[3,1,2,0]=(-0.8117641465)*sqrt(x[0])
        arg[3,1,2,1]=(-0.357390874294)*sqrt(x[1])
        arg[3,1,2,2]=(-0.0202262660381)*sqrt(x[2])
        arg[3,2,0,0]=(-0.181594076044)*sqrt(x[0])
        arg[3,2,0,1]=(0.0655758078072)*sqrt(x[1])
        arg[3,2,0,2]=(0.926686460908)*sqrt(x[2])
        arg[3,2,1,0]=(-0.700426183586)*sqrt(x[0])
        arg[3,2,1,1]=(0.600692261192)*sqrt(x[1])
        arg[3,2,1,2]=(0.757146807006)*sqrt(x[2])
        arg[3,2,2,0]=(0.588097603826)*sqrt(x[0])
        arg[3,2,2,1]=(-0.748732775295)*sqrt(x[1])
        arg[3,2,2,2]=(0.628960378488)*sqrt(x[2])
        arg[3,3,0,0]=(-0.0612899752684)*sqrt(x[0])
        arg[3,3,0,1]=(0.567198903535)*sqrt(x[1])
        arg[3,3,0,2]=(0.743996205698)*sqrt(x[2])
        arg[3,3,1,0]=(0.333763157477)*sqrt(x[0])
        arg[3,3,1,1]=(0.795773692731)*sqrt(x[1])
        arg[3,3,1,2]=(-0.676301506816)*sqrt(x[2])
        arg[3,3,2,0]=(-0.255504087739)*sqrt(x[0])
        arg[3,3,2,1]=(-0.556263970561)*sqrt(x[1])
        arg[3,3,2,2]=(0.721245074469)*sqrt(x[2])
        arg[3,4,0,0]=(0.384551582302)*sqrt(x[0])
        arg[3,4,0,1]=(0.849519144853)*sqrt(x[1])
        arg[3,4,0,2]=(0.101881023262)*sqrt(x[2])
        arg[3,4,1,0]=(-0.685891926653)*sqrt(x[0])
        arg[3,4,1,1]=(0.390887672448)*sqrt(x[1])
        arg[3,4,1,2]=(0.0911617386398)*sqrt(x[2])
        arg[3,4,2,0]=(-0.889618312513)*sqrt(x[0])
        arg[3,4,2,1]=(0.130310876981)*sqrt(x[1])
        arg[3,4,2,2]=(0.958562278204)*sqrt(x[2])
        ref=sqrt((12.9075635889)*3./2.+(5.72789470359)/0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")

